package Geo::Coordinates::OSTN02;
require Exporter;
use strict;
use warnings;

our @ISA = qw(Exporter);
our $VERSION = '2.02';
our @EXPORT = qw();
our @EXPORT_OK = qw(
    OSGB36_to_ETRS89
    ETRS89_to_OSGB36
);

our @ostn_data = <DATA>; # load all the data from below
our %ostn_shift_for=();

sub ETRS89_to_OSGB36 {
    return unless defined wantarray;

    my ($x,  $y,  $z ) = (@_, 0);
    my ($dx, $dy, $dz) = _find_OSTN02_shifts_at($x,$y);

    ($x, $y, $z) = _round_to_nearest_mm($x+$dx, $y+$dy, $z-$dz);

    return ($x, $y, $z) if wantarray;
    return "$x $y $z";
}

sub OSGB36_to_ETRS89 {
    return unless defined wantarray;

    my $epsilon = 0.00001;
    my ($x0, $y0, $z0) = (@_, 0);
    my ($dx, $dy, $dz) = _find_OSTN02_shifts_at($x0,$y0);
    my ($x,  $y,  $z ) = ($x0-$dx, $y0-$dy, $z0+$dz);
    my ($last_dx, $last_dy) = ($dx, $dy);
    APPROX:
    while (1) {
        ($dx, $dy, $dz) = _find_OSTN02_shifts_at($x,$y);
        ($x, $y) = ($x0-$dx, $y0-$dy);
        last APPROX if abs($dx-$last_dx)<$epsilon && abs($dy-$last_dy)<$epsilon;
        ($last_dx, $last_dy) = ($dx, $dy);
    }

    ($x, $y, $z) = _round_to_nearest_mm($x0-$dx, $y0-$dy, $z0+$dz);

    return ($x, $y, $z) if wantarray;
    return "$x $y $z";
}

sub _round_to_nearest_mm {
    my ($x,  $y,  $z ) = @_;
    $x = int($x*1000+0.5)/1000;
    $y = int($y*1000+0.5)/1000;
    $z = int($z*1000+0.5)/1000;
    return ($x, $y, $z);
}

sub _find_OSTN02_shifts_at {

    my ($x, $y) = @_;

    my $e_index = int($x/1000);
    my $n_index = int($y/1000);

    my $x0 = $e_index * 1000;
    my $y0 = $n_index * 1000;

    my $s0_ref = _get_ostn_ref($e_index+0, $n_index+0) || die "OSTN not defined for $x $y\n";
    my $s1_ref = _get_ostn_ref($e_index+1, $n_index+0) || die "OSTN not defined for $x $y\n";
    my $s2_ref = _get_ostn_ref($e_index+0, $n_index+1) || die "OSTN not defined for $x $y\n";
    my $s3_ref = _get_ostn_ref($e_index+1, $n_index+1) || die "OSTN not defined for $x $y\n";

    my $dx = $x - $x0; # offset within square
    my $dy = $y - $y0;

    my $t = $dx/1000;
    my $u = $dy/1000;

    my $f0 = (1-$t)*(1-$u);
    my $f1 =    $t *(1-$u);
    my $f2 = (1-$t)*   $u ;
    my $f3 =    $t *   $u ;

    my $se = $f0*$s0_ref->[0] + $f1*$s1_ref->[0] + $f2*$s2_ref->[0] + $f3*$s3_ref->[0];
    my $sn = $f0*$s0_ref->[1] + $f1*$s1_ref->[1] + $f2*$s2_ref->[1] + $f3*$s3_ref->[1];
    my $sg = $f0*$s0_ref->[2] + $f1*$s1_ref->[2] + $f2*$s2_ref->[2] + $f3*$s3_ref->[2];

    return ($se, $sn, $sg);
}

sub _get_ostn_ref {
    my $x = shift;
    my $y = shift;
    my $key = $x + ($y*701) + 1;

    return $ostn_shift_for{$key} if exists $ostn_shift_for{$key};

    my $lo = 0;
    my $hi = @ostn_data;
    my $mid;
    while (1) {
        $mid = ($hi+$lo)/2;
        my ($k, @data) = split ' ', $ostn_data[$mid];
        $ostn_shift_for{$k} = \@data;
        last        if $k == $key;
        $lo = $mid  if $k  < $key;
        $hi = $mid  if $k  > $key;
    }

    return $ostn_shift_for{$key};
}

1;

=pod

=head1 NAME

Geo::Coordinates::OSTN02 -- Implement the OSTN02 transformation


=head1 SYNOPSIS

  use Geo::Coordinates::OSTN02 qw/OSGB36_to_ETRS89 ETRS89_to_OSGB36/;

  ($x, $y, $z) = OSGB36_to_ETRS89($e, $n, $h);
  ($e, $n, $h) = ETRS89_to_OSGB36($x, $y, $z);

=head1 DESCRIPTION

The purpose and use of these modules is described briefly in the companion
L<Geo::Coordinates::OSGB> modules.  In essence they implement in Perl the
Ordnance Survey's OSTN02 transformation which is part of the current
definition of the British National Grid.  You are strongly recommended to
read the official public guides and other introductory documents that are
published by the Ordnance Survey of Great Britain with a wealth of other
technical information about the OSTN02 transformation.  The best starting
point on the web is
http://www.ordnancesurvey.co.uk/oswebsite/gps/information/index.html

=head1 FUNCTIONS

The following functions can be exported from the C<Geo::Coordinates::OSTN02>
module:

    OSGB36_to_ETRS89
    ETRS89_to_OSGB36

None is exported by default.

=over 4

=item OSGB36_to_ETRS89(easting, northing, [elevation])

Transforms from normal OSGB36 grid references to a pseudo-grid based on the WGS84/ETRS89 geoid
model, which can then be translated to lat/long coordinates using C<grid_to_ll()> with the 'WGS84'
parameter.

   my $elevation = '100'; # in metres
   my ($e, $n) = parse_grid_trad('TQ304293');
   my ($x, $y, $z) = OSGB36_to_ETRS89($e, $n, $elevation);
   my ($lat, $lon) = grid_to_ll($x, $y, 'ETRS89'); # or you can say 'WGS84' if you like - same thing

Elevation will default to 0 metres if you omit it.

=item ETRS89_to_OSGB36(x, y, [z])

Transforms WGS84/ETRS89 pseudo-grid coordinates into OSGB36 grid references.

  my ($lat, $lon, $height) = (51.5, -1, 10); # Somewhere in London
  my ($x, $y) = ll_to_grid($lat, $lon, 'ETRS89'); # or 'WGS84'
  my ($e, $n, $elevation) = ETRS89_to_OSGB36($x, $y, $height);

=back

=head1 IMPLEMENTATION NOTES

Since 2002 the British Ordnance Survey have defined the UK national grid not
as a projection from their own model of the earth (the Airy 1830 geoid,
revised 1936, known as OSGB36), but as a simple table of calculated
differences from a projection based on the European standard geoid ETRS89
(which is for Europe a functional equivalent of the international WGS84
geoid).  The idea is that you project your WGS84 latitude and longitude
coordinates into WGS84 (or ETRS89) grid references, and then look up the
appropriate three dimensional correction in the OS table, and add the results
to your grid reference to give you a standard British National Grid reference
that you can use with British OS maps.  Going back the other way is slightly
more complicated, as you have to work backwards, but a simple iteration will
do the job.  This package implements these lookups and adjustments.  You give
it a three dimensional grid reference (easting, northing, and altitude, all
in metres) and the package returns it corrected from one system to the other.

The problem in the implementation is that the table is huge, and most of it
is empty as the OS have chosen to leave the transformation undefined for
areas that are off shore.  So this package only works properly for grid
references that are actually on one or more OS maps.  The complete table
(including all the 0 lines) contains nearly 1 million lines with four data
points and a key.  In text form as supplied by the OS that is about 36M bytes
of table.  Even storing the numbers in binary form it would still need about
9M bytes.  About 2/3 of the points are blank though, so leaving them out
brings the amount of text down to about 10M bytes and binary data to about 3M
bytes.  For simplicity this package includes the text of the table with the
blank lines omitted.

A future version might compress the table into binary form.  The ranges of the corrections are
as follows:

  Max k: 861292 e: 665 n: 1228 x: 103.443 y: -50.168 z: 57.668
  Min k:     80 e:   2 n:    0 x:  86.275 y: -81.603 z: 43.982
  Range  861212    663    1228     17.168     31.435    13.686
  Bits       20                    15         15        14
  Total bits needed 64 = 8 bytes x 309792 lines = 2478336 = 2.4M bytes

Since we omit the blank lines, we cannot simply get the nth row for index n
so we keep two structures:  @ostn_data holds all the data in a big array, and
%ostn_shift_for is a hash that caches lines in the table that we have already
looked up.  To get a line, we calculate the index number from an (x,y) pair,
look for that in the hash, and use it if we can.  If it is not there, we then
search through the array using a simple binary chop, and cache the result.

This works pretty quickly, the only slow bit is loading the array at the
beginning, but it is much faster and need *much* less memory than loading all
the data into the hash (which would be simpler, and is what the original
version did, but it was too slow to be usable and meant that the tests failed
on many smaller systems as part of CPAN testing).

Loading the array takes about 1.7 seconds on my Windows machine (a 1.8G Hz
Intel Pentium M processor with 1G byte of memory) and just under 0.5 seconds
on my Linux system (a 2.8G Hz Intel Pentium 4 processor with 512M bytes of
memory).  I think this probably says more about speed of the disks I have
(and probably the efficiency of Perl under Linux), but your results should be
comparable.  Once the data is loaded, calling the routines is reasonably
quick.

=head1 BUGS

Please report any to toby@cpan.org

=head1 EXAMPLES

See the test routines included in the distribution.

=head1 AUTHOR

Toby Thurston --- 22 Sep 2008

web: http://www.wildfire.dircon.co.uk

=head1 SEE ALSO

See L<Geo::Coordinates::OSGB> for the main routines.
=cut
__DATA__
80 91.888 -81.603 53.184
81 91.903 -81.597 53.178
82 91.917 -81.592 53.173
83 91.936 -81.582 53.168
84 91.956 -81.569 53.164
85 91.976 -81.557 53.160
86 91.996 -81.545 53.157
87 92.016 -81.534 53.154
88 92.035 -81.522 53.151
89 92.056 -81.510 53.148
90 92.076 -81.497 53.143
91 92.096 -81.485 53.139
92 92.116 -81.473 53.135
93 92.139 -81.462 53.131
94 92.163 -81.450 53.127
95 92.188 -81.439 53.123
96 92.212 -81.428 53.117
97 92.236 -81.416 53.112
98 92.261 -81.405 53.105
99 92.285 -81.394 53.098
100 92.310 -81.382 53.092
780 91.873 -81.579 53.192
781 91.887 -81.574 53.184
782 91.902 -81.569 53.178
783 91.916 -81.563 53.172
784 91.932 -81.557 53.167
785 91.951 -81.545 53.162
786 91.971 -81.533 53.158
787 91.992 -81.521 53.155
788 92.012 -81.509 53.152
789 92.031 -81.497 53.148
790 92.052 -81.484 53.145
791 92.072 -81.473 53.140
792 92.091 -81.461 53.136
793 92.115 -81.450 53.132
794 92.140 -81.437 53.128
795 92.164 -81.426 53.124
796 92.189 -81.415 53.119
797 92.213 -81.403 53.114
798 92.237 -81.392 53.108
799 92.262 -81.381 53.102
800 92.286 -81.369 53.094
801 92.311 -81.358 53.089
802 92.336 -81.347 53.082
1480 91.858 -81.556 53.201
1481 91.872 -81.550 53.193
1482 91.887 -81.545 53.185
1483 91.901 -81.540 53.179
1484 91.916 -81.535 53.172
1485 91.930 -81.529 53.166
1486 91.947 -81.520 53.161
1487 91.967 -81.508 53.157
1488 91.988 -81.496 53.154
1489 92.007 -81.484 53.150
1490 92.027 -81.472 53.146
1491 92.048 -81.460 53.143
1492 92.068 -81.448 53.137
1493 92.092 -81.437 53.133
1494 92.117 -81.425 53.129
1495 92.141 -81.414 53.124
1496 92.165 -81.403 53.120
1497 92.190 -81.391 53.116
1498 92.214 -81.380 53.110
1499 92.239 -81.368 53.105
1500 92.263 -81.356 53.098
1501 92.288 -81.345 53.091
1502 92.312 -81.334 53.085
1503 92.337 -81.322 53.079
1504 92.356 -81.315 53.072
1569 93.499 -80.739 53.060
1570 93.513 -80.729 53.056
1571 93.520 -80.719 53.050
1572 93.528 -80.709 53.045
2181 91.857 -81.527 53.204
2182 91.872 -81.522 53.195
2183 91.886 -81.516 53.187
2184 91.900 -81.511 53.180
2185 91.915 -81.506 53.173
2186 91.929 -81.500 53.167
2187 91.944 -81.495 53.161
2188 91.963 -81.484 53.157
2189 91.984 -81.472 53.152
2190 92.003 -81.459 53.148
2191 92.023 -81.447 53.144
2192 92.044 -81.435 53.140
2193 92.069 -81.424 53.135
2194 92.093 -81.412 53.130
2195 92.118 -81.401 53.126
2196 92.142 -81.390 53.121
2197 92.166 -81.378 53.117
2198 92.191 -81.367 53.112
2199 92.215 -81.356 53.106
2200 92.240 -81.344 53.101
2201 92.264 -81.333 53.094
2202 92.289 -81.321 53.088
2203 92.313 -81.310 53.082
2204 92.333 -81.302 53.075
2205 92.351 -81.294 53.069
2206 92.369 -81.287 53.061
2226 92.739 -81.134 53.425
2227 92.758 -81.127 53.415
2228 92.776 -81.118 53.405
2229 92.794 -81.111 53.394
2230 92.813 -81.103 53.384
2269 93.478 -80.729 53.068
2270 93.494 -80.718 53.063
2271 93.510 -80.707 53.058
2272 93.518 -80.698 53.053
2273 93.525 -80.688 53.048
2274 93.533 -80.678 53.042
2275 93.540 -80.667 53.037
2276 93.547 -80.657 53.032
2277 93.554 -80.648 53.026
2278 93.561 -80.638 53.019
2279 93.568 -80.628 53.011
2280 93.575 -80.618 53.003
2881 91.843 -81.503 53.219
2882 91.857 -81.497 53.208
2883 91.871 -81.492 53.198
2884 91.885 -81.487 53.189
2885 91.900 -81.481 53.182
2886 91.915 -81.476 53.174
2887 91.929 -81.472 53.168
2888 91.943 -81.466 53.162
2889 91.959 -81.459 53.157
2890 91.979 -81.447 53.152
2891 91.999 -81.434 53.147
2892 92.021 -81.423 53.143
2893 92.045 -81.412 53.139
2894 92.070 -81.400 53.133
2895 92.094 -81.389 53.128
2896 92.119 -81.377 53.123
2897 92.143 -81.365 53.118
2898 92.167 -81.354 53.113
2899 92.192 -81.343 53.108
2900 92.217 -81.331 53.103
2901 92.241 -81.320 53.097
2902 92.266 -81.309 53.091
2903 92.290 -81.297 53.084
2904 92.309 -81.289 53.078
2905 92.328 -81.281 53.072
2906 92.346 -81.274 53.065
2907 92.365 -81.266 53.058
2926 92.716 -81.121 53.429
2927 92.734 -81.114 53.419
2928 92.753 -81.106 53.409
2929 92.771 -81.099 53.397
2930 92.790 -81.091 53.387
2931 92.808 -81.083 53.377
2932 92.827 -81.075 53.366
2933 92.845 -81.068 53.357
2934 92.864 -81.060 53.347
2935 92.882 -81.052 53.338
2936 92.901 -81.045 53.329
2969 93.457 -80.718 53.075
2970 93.473 -80.708 53.070
2971 93.489 -80.698 53.065
2972 93.505 -80.687 53.060
2973 93.516 -80.676 53.056
2974 93.523 -80.667 53.051
2975 93.531 -80.657 53.046
2976 93.538 -80.647 53.040
2977 93.545 -80.636 53.035
2978 93.552 -80.626 53.029
2979 93.559 -80.616 53.022
2980 93.566 -80.607 53.015
2981 93.573 -80.597 53.007
2982 93.579 -80.586 52.999
2983 93.586 -80.576 52.991
3581 91.827 -81.479 53.239
3582 91.842 -81.474 53.226
3583 91.856 -81.469 53.214
3584 91.871 -81.463 53.203
3585 91.885 -81.458 53.193
3586 91.899 -81.453 53.185
3587 91.914 -81.447 53.177
3588 91.928 -81.442 53.170
3589 91.943 -81.437 53.163
3590 91.957 -81.431 53.157
3591 91.975 -81.422 53.152
3592 91.997 -81.410 53.147
3593 92.022 -81.399 53.142
3594 92.047 -81.387 53.137
3595 92.071 -81.376 53.132
3596 92.095 -81.365 53.127
3597 92.120 -81.353 53.121
3598 92.144 -81.342 53.116
3599 92.169 -81.331 53.110
3600 92.193 -81.319 53.105
3601 92.218 -81.308 53.099
3602 92.242 -81.296 53.094
3603 92.267 -81.284 53.087
3604 92.286 -81.276 53.080
3605 92.305 -81.268 53.074
3606 92.323 -81.261 53.068
3607 92.342 -81.253 53.061
3608 92.360 -81.246 53.054
3609 92.379 -81.238 53.047
3626 92.693 -81.108 53.434
3627 92.711 -81.101 53.423
3628 92.730 -81.093 53.413
3629 92.748 -81.086 53.402
3630 92.767 -81.078 53.391
3631 92.785 -81.071 53.380
3632 92.804 -81.063 53.369
3633 92.822 -81.055 53.359
3634 92.841 -81.048 53.349
3635 92.859 -81.039 53.339
3636 92.877 -81.032 53.329
3637 92.896 -81.024 53.320
3638 92.915 -81.017 53.310
3639 92.933 -81.009 53.301
3669 93.435 -80.708 53.079
3670 93.451 -80.698 53.075
3671 93.468 -80.687 53.072
3672 93.484 -80.676 53.067
3673 93.500 -80.666 53.062
3674 93.514 -80.655 53.058
3675 93.521 -80.645 53.054
3676 93.529 -80.635 53.049
3677 93.536 -80.626 53.043
3678 93.543 -80.615 53.037
3679 93.550 -80.605 53.031
3680 93.557 -80.595 53.024
3681 93.564 -80.585 53.017
3682 93.571 -80.576 53.010
3683 93.577 -80.565 53.002
3684 93.584 -80.555 52.994
3685 93.591 -80.545 52.985
4282 91.827 -81.450 53.247
4283 91.841 -81.445 53.233
4284 91.856 -81.440 53.221
4285 91.870 -81.434 53.209
4286 91.884 -81.429 53.199
4287 91.899 -81.424 53.189
4288 91.913 -81.419 53.180
4289 91.928 -81.413 53.173
4290 91.942 -81.408 53.166
4291 91.956 -81.403 53.159
4292 91.974 -81.398 53.153
4293 91.999 -81.386 53.147
4294 92.023 -81.375 53.142
4295 92.048 -81.363 53.137
4296 92.072 -81.352 53.131
4297 92.096 -81.340 53.125
4298 92.121 -81.329 53.119
4299 92.145 -81.318 53.114
4300 92.170 -81.306 53.108
4301 92.195 -81.295 53.102
4302 92.219 -81.284 53.096
4303 92.243 -81.272 53.090
4304 92.263 -81.263 53.084
4305 92.282 -81.256 53.076
4306 92.300 -81.248 53.070
4307 92.319 -81.240 53.064
4308 92.337 -81.233 53.058
4309 92.356 -81.225 53.051
4310 92.374 -81.218 53.044
4311 92.392 -81.210 53.037
4327 92.688 -81.088 53.429
4328 92.707 -81.080 53.417
4329 92.725 -81.073 53.406
4330 92.743 -81.065 53.395
4331 92.762 -81.058 53.384
4332 92.781 -81.050 53.373
4333 92.799 -81.043 53.361
4334 92.817 -81.035 53.351
4335 92.836 -81.027 53.340
4336 92.855 -81.020 53.329
4337 92.873 -81.012 53.320
4338 92.892 -81.004 53.309
4339 92.910 -80.996 53.300
4340 92.928 -80.989 53.291
4369 93.414 -80.698 53.081
4370 93.430 -80.687 53.078
4371 93.446 -80.676 53.075
4372 93.463 -80.666 53.071
4373 93.479 -80.655 53.068
4374 93.495 -80.645 53.064
4375 93.512 -80.634 53.060
4376 93.519 -80.624 53.055
4377 93.526 -80.614 53.051
4378 93.534 -80.604 53.045
4379 93.541 -80.594 53.039
4380 93.548 -80.584 53.033
4381 93.555 -80.574 53.026
4382 93.562 -80.564 53.019
4383 93.569 -80.554 53.011
4384 93.575 -80.544 53.003
4385 93.582 -80.534 52.995
4386 93.589 -80.524 52.986
4387 93.593 -80.519 52.977
4983 91.826 -81.422 53.257
4984 91.841 -81.416 53.242
4985 91.855 -81.411 53.228
4986 91.869 -81.406 53.216
4987 91.884 -81.400 53.205
4988 91.898 -81.395 53.195
4989 91.913 -81.390 53.185
4990 91.927 -81.384 53.177
4991 91.940 -81.378 53.169
4992 91.954 -81.373 53.162
4993 91.968 -81.367 53.155
4994 92.000 -81.363 53.149
4995 92.024 -81.351 53.142
4996 92.049 -81.340 53.137
4997 92.073 -81.328 53.131
4998 92.097 -81.316 53.124
4999 92.122 -81.305 53.118
5000 92.147 -81.293 53.112
5001 92.171 -81.282 53.106
5002 92.196 -81.271 53.100
5003 92.220 -81.259 53.093
5004 92.240 -81.251 53.086
5005 92.259 -81.243 53.080
5006 92.277 -81.236 53.072
5007 92.296 -81.228 53.065
5008 92.314 -81.220 53.059
5009 92.333 -81.212 53.054
5010 92.351 -81.205 53.047
5011 92.369 -81.197 53.040
5012 92.388 -81.190 53.033
5027 92.665 -81.075 53.435
5028 92.684 -81.068 53.423
5029 92.702 -81.060 53.411
5030 92.721 -81.052 53.400
5031 92.739 -81.045 53.388
5032 92.758 -81.037 53.376
5033 92.776 -81.030 53.364
5034 92.794 -81.022 53.352
5035 92.813 -81.014 53.341
5036 92.832 -81.007 53.330
5037 92.850 -80.999 53.320
5038 92.868 -80.992 53.309
5039 92.887 -80.984 53.299
5040 92.905 -80.977 53.289
5041 92.924 -80.968 53.280
5042 92.942 -80.961 53.271
5069 93.392 -80.687 53.082
5070 93.409 -80.676 53.078
5071 93.425 -80.666 53.076
5072 93.441 -80.655 53.073
5073 93.458 -80.645 53.070
5074 93.474 -80.634 53.067
5075 93.490 -80.624 53.065
5076 93.507 -80.613 53.061
5077 93.517 -80.603 53.056
5078 93.524 -80.593 53.052
5079 93.532 -80.583 53.046
5080 93.539 -80.573 53.041
5081 93.546 -80.563 53.035
5082 93.553 -80.553 53.028
5083 93.560 -80.543 53.021
5084 93.567 -80.533 53.013
5085 93.573 -80.523 53.005
5086 93.580 -80.514 52.997
5087 93.583 -80.510 52.988
5088 93.586 -80.504 52.979
5089 93.590 -80.500 52.970
5090 93.593 -80.495 52.961
5683 91.811 -81.398 53.286
5684 91.825 -81.393 53.269
5685 91.840 -81.387 53.253
5686 91.855 -81.382 53.238
5687 91.869 -81.377 53.225
5688 91.883 -81.371 53.213
5689 91.896 -81.365 53.202
5690 91.911 -81.359 53.192
5691 91.925 -81.354 53.182
5692 91.939 -81.349 53.174
5693 91.953 -81.343 53.165
5694 91.966 -81.337 53.158
5695 92.000 -81.339 53.151
5696 92.026 -81.328 53.144
5697 92.050 -81.315 53.137
5698 92.074 -81.304 53.131
5699 92.099 -81.293 53.124
5700 92.123 -81.281 53.117
5701 92.148 -81.270 53.110
5702 92.172 -81.259 53.104
5703 92.197 -81.247 53.098
5704 92.217 -81.238 53.090
5705 92.236 -81.231 53.083
5706 92.254 -81.223 53.076
5707 92.273 -81.215 53.069
5708 92.291 -81.208 53.062
5709 92.309 -81.200 53.055
5710 92.328 -81.192 53.049
5711 92.347 -81.184 53.042
5712 92.365 -81.177 53.036
5713 92.383 -81.169 53.030
5727 92.642 -81.062 53.441
5728 92.660 -81.055 53.429
5729 92.679 -81.047 53.416
5730 92.698 -81.039 53.404
5731 92.716 -81.032 53.392
5732 92.734 -81.024 53.380
5733 92.753 -81.017 53.368
5734 92.772 -81.009 53.355
5735 92.790 -81.002 53.343
5736 92.808 -80.994 53.331
5737 92.827 -80.986 53.320
5738 92.845 -80.979 53.309
5739 92.864 -80.971 53.298
5740 92.882 -80.964 53.288
5741 92.901 -80.956 53.278
5742 92.919 -80.949 53.268
5743 92.938 -80.941 53.260
5769 93.371 -80.676 53.081
5770 93.387 -80.667 53.077
5771 93.404 -80.656 53.074
5772 93.420 -80.645 53.071
5773 93.436 -80.635 53.069
5774 93.452 -80.624 53.067
5775 93.469 -80.613 53.064
5776 93.485 -80.603 53.063
5777 93.502 -80.592 53.061
5778 93.515 -80.582 53.057
5779 93.522 -80.572 53.052
5780 93.530 -80.562 53.047
5781 93.537 -80.552 53.042
5782 93.544 -80.541 53.035
5783 93.551 -80.532 53.029
5784 93.558 -80.522 53.022
5785 93.565 -80.512 53.015
5786 93.570 -80.504 53.006
5787 93.573 -80.500 52.998
5788 93.576 -80.495 52.989
5789 93.580 -80.490 52.980
5790 93.583 -80.485 52.971
5791 93.586 -80.481 52.962
5792 93.590 -80.476 52.953
6384 91.811 -81.369 53.300
6385 91.825 -81.364 53.282
6386 91.839 -81.358 53.265
6387 91.853 -81.353 53.249
6388 91.867 -81.347 53.235
6389 91.881 -81.341 53.222
6390 91.895 -81.335 53.210
6391 91.909 -81.330 53.199
6392 91.923 -81.325 53.189
6393 91.937 -81.318 53.179
6394 91.951 -81.313 53.170
6395 91.965 -81.307 53.161
6396 91.985 -81.304 53.154
6397 92.028 -81.306 53.146
6398 92.052 -81.293 53.139
6399 92.076 -81.281 53.132
6400 92.100 -81.268 53.124
6401 92.125 -81.257 53.116
6402 92.149 -81.246 53.109
6403 92.173 -81.234 53.103
6404 92.194 -81.225 53.095
6405 92.213 -81.218 53.088
6406 92.231 -81.210 53.080
6407 92.249 -81.202 53.073
6408 92.268 -81.195 53.065
6409 92.286 -81.187 53.058
6410 92.305 -81.180 53.051
6411 92.323 -81.172 53.045
6412 92.342 -81.165 53.039
6413 92.360 -81.156 53.032
6414 92.379 -81.149 53.026
6415 92.398 -81.141 53.018
6428 92.638 -81.043 53.436
6429 92.656 -81.035 53.423
6430 92.674 -81.027 53.409
6431 92.693 -81.019 53.397
6432 92.711 -81.011 53.384
6433 92.730 -81.004 53.371
6434 92.748 -80.996 53.359
6435 92.767 -80.989 53.346
6436 92.785 -80.981 53.333
6437 92.804 -80.974 53.321
6438 92.822 -80.966 53.309
6439 92.841 -80.958 53.298
6440 92.859 -80.951 53.287
6441 92.878 -80.943 53.277
6442 92.896 -80.936 53.267
6443 92.915 -80.928 53.257
6444 92.933 -80.920 53.248
6445 92.951 -80.913 53.240
6470 93.366 -80.656 53.074
6471 93.382 -80.645 53.071
6472 93.399 -80.635 53.069
6473 93.415 -80.624 53.066
6474 93.431 -80.613 53.065
6475 93.447 -80.603 53.063
6476 93.464 -80.592 53.062
6477 93.480 -80.582 53.061
6478 93.497 -80.571 53.058
6479 93.513 -80.560 53.055
6480 93.520 -80.551 53.052
6481 93.528 -80.541 53.047
6482 93.535 -80.531 53.042
6483 93.542 -80.520 53.036
6484 93.549 -80.510 53.029
6485 93.556 -80.500 53.022
6486 93.560 -80.494 53.015
6487 93.563 -80.490 53.007
6488 93.567 -80.485 52.998
6489 93.570 -80.481 52.990
6490 93.573 -80.475 52.981
6491 93.576 -80.471 52.972
6492 93.580 -80.466 52.963
6493 93.583 -80.462 52.954
6494 93.586 -80.457 52.944
7085 91.809 -81.340 53.315
7086 91.823 -81.334 53.297
7087 91.837 -81.328 53.279
7088 91.851 -81.322 53.262
7089 91.865 -81.317 53.247
7090 91.879 -81.311 53.233
7091 91.893 -81.306 53.220
7092 91.907 -81.300 53.208
7093 91.921 -81.294 53.197
7094 91.935 -81.289 53.186
7095 91.949 -81.283 53.176
7096 91.963 -81.278 53.166
7097 91.977 -81.271 53.158
7098 92.024 -81.279 53.150
7099 92.054 -81.271 53.142
7100 92.078 -81.259 53.134
7101 92.102 -81.246 53.125
7102 92.126 -81.234 53.117
7103 92.150 -81.221 53.109
7104 92.171 -81.212 53.101
7105 92.190 -81.205 53.094
7106 92.208 -81.197 53.086
7107 92.226 -81.190 53.078
7108 92.245 -81.182 53.070
7109 92.264 -81.174 53.062
7110 92.282 -81.167 53.055
7111 92.300 -81.159 53.048
7112 92.319 -81.152 53.042
7113 92.337 -81.144 53.035
7114 92.356 -81.137 53.028
7115 92.374 -81.129 53.022
7116 92.392 -81.121 53.015
7129 92.633 -81.022 53.430
7130 92.651 -81.014 53.417
7131 92.670 -81.007 53.403
7132 92.689 -80.999 53.390
7133 92.707 -80.991 53.377
7134 92.725 -80.983 53.363
7135 92.744 -80.976 53.350
7136 92.762 -80.968 53.336
7137 92.781 -80.961 53.322
7138 92.799 -80.953 53.311
7139 92.817 -80.945 53.299
7140 92.836 -80.938 53.287
7141 92.855 -80.930 53.276
7142 92.873 -80.923 53.265
7143 92.892 -80.915 53.255
7144 92.910 -80.908 53.245
7145 92.928 -80.900 53.236
7146 92.947 -80.892 53.227
7170 93.345 -80.645 53.069
7171 93.361 -80.635 53.067
7172 93.377 -80.624 53.065
7173 93.394 -80.613 53.063
7174 93.410 -80.603 53.061
7175 93.426 -80.592 53.060
7176 93.442 -80.582 53.059
7177 93.459 -80.571 53.059
7178 93.475 -80.560 53.057
7179 93.492 -80.550 53.056
7180 93.508 -80.540 53.054
7181 93.518 -80.529 53.050
7182 93.525 -80.519 53.046
7183 93.533 -80.510 53.041
7184 93.540 -80.499 53.035
7185 93.547 -80.489 53.029
7186 93.550 -80.485 53.022
7187 93.553 -80.480 53.015
7188 93.557 -80.475 53.007
7189 93.560 -80.471 52.999
7190 93.563 -80.466 52.990
7191 93.567 -80.461 52.981
7192 93.570 -80.457 52.972
7193 93.573 -80.452 52.963
7194 93.577 -80.447 52.955
7195 93.580 -80.443 52.945
7196 93.583 -80.438 52.935
7786 91.807 -81.309 53.333
7787 91.821 -81.304 53.314
7788 91.835 -81.298 53.295
7789 91.849 -81.293 53.277
7790 91.863 -81.287 53.261
7791 91.877 -81.281 53.246
7792 91.891 -81.275 53.232
7793 91.905 -81.270 53.219
7794 91.919 -81.265 53.206
7795 91.933 -81.259 53.194
7796 91.947 -81.253 53.183
7797 91.961 -81.247 53.173
7798 91.975 -81.242 53.164
7799 92.009 -81.244 53.155
7800 92.056 -81.249 53.146
7801 92.079 -81.236 53.137
7802 92.103 -81.224 53.128
7803 92.128 -81.212 53.120
7804 92.147 -81.199 53.111
7805 92.165 -81.191 53.103
7806 92.185 -81.184 53.094
7807 92.203 -81.177 53.085
7808 92.222 -81.169 53.077
7809 92.240 -81.162 53.068
7810 92.259 -81.154 53.060
7811 92.277 -81.146 53.053
7812 92.296 -81.139 53.046
7813 92.314 -81.131 53.039
7814 92.333 -81.124 53.032
7815 92.351 -81.116 53.026
7816 92.370 -81.108 53.019
7817 92.388 -81.101 53.012
7830 92.628 -81.002 53.425
7831 92.647 -80.994 53.411
7832 92.665 -80.986 53.396
7833 92.684 -80.979 53.382
7834 92.702 -80.971 53.369
7835 92.721 -80.964 53.355
7836 92.739 -80.955 53.341
7837 92.758 -80.948 53.327
7838 92.776 -80.940 53.313
7839 92.795 -80.933 53.300
7840 92.813 -80.925 53.288
7841 92.832 -80.917 53.275
7842 92.850 -80.910 53.264
7843 92.868 -80.902 53.253
7844 92.887 -80.895 53.243
7845 92.906 -80.887 53.233
7846 92.924 -80.880 53.223
7847 92.942 -80.872 53.214
7848 92.961 -80.864 53.206
7870 93.323 -80.635 53.064
7871 93.339 -80.624 53.062
7872 93.356 -80.613 53.059
7873 93.372 -80.603 53.057
7874 93.389 -80.592 53.056
7875 93.405 -80.582 53.055
7876 93.421 -80.572 53.054
7877 93.437 -80.561 53.054
7878 93.454 -80.551 53.054
7879 93.470 -80.540 53.053
7880 93.487 -80.529 53.053
7881 93.503 -80.519 53.052
7882 93.516 -80.508 53.048
7883 93.523 -80.498 53.044
7884 93.531 -80.488 53.039
7885 93.537 -80.479 53.034
7886 93.540 -80.475 53.028
7887 93.543 -80.470 53.021
7888 93.547 -80.466 53.014
7889 93.550 -80.461 53.006
7890 93.553 -80.457 52.998
7891 93.557 -80.452 52.990
7892 93.560 -80.447 52.981
7893 93.563 -80.442 52.972
7894 93.567 -80.438 52.964
7895 93.570 -80.433 52.955
7896 93.573 -80.428 52.945
7897 93.577 -80.424 52.936
7898 93.580 -80.419 52.927
8486 91.791 -81.285 53.374
8487 91.805 -81.280 53.353
8488 91.819 -81.274 53.332
8489 91.834 -81.268 53.313
8490 91.847 -81.262 53.294
8491 91.861 -81.257 53.277
8492 91.875 -81.251 53.261
8493 91.889 -81.246 53.245
8494 91.904 -81.240 53.231
8495 91.917 -81.234 53.218
8496 91.931 -81.229 53.204
8497 91.945 -81.223 53.193
8498 91.959 -81.218 53.182
8499 91.974 -81.212 53.171
8500 91.994 -81.209 53.161
8501 92.048 -81.219 53.151
8502 92.081 -81.214 53.141
8503 92.105 -81.201 53.132
8504 92.126 -81.190 53.123
8505 92.142 -81.179 53.114
8506 92.158 -81.168 53.105
8507 92.177 -81.162 53.094
8508 92.197 -81.155 53.085
8509 92.216 -81.148 53.076
8510 92.236 -81.141 53.067
8511 92.254 -81.133 53.059
8512 92.273 -81.126 53.051
8513 92.291 -81.118 53.044
8514 92.309 -81.111 53.037
8515 92.328 -81.103 53.030
8516 92.347 -81.096 53.023
8517 92.365 -81.088 53.016
8518 92.383 -81.080 53.009
8531 92.624 -80.981 53.419
8532 92.642 -80.974 53.405
8533 92.661 -80.966 53.389
8534 92.679 -80.958 53.375
8535 92.698 -80.951 53.362
8536 92.716 -80.943 53.347
8537 92.734 -80.936 53.332
8538 92.753 -80.928 53.318
8539 92.772 -80.920 53.304
8540 92.790 -80.912 53.290
8541 92.808 -80.905 53.277
8542 92.827 -80.897 53.265
8543 92.845 -80.889 53.253
8544 92.864 -80.882 53.241
8545 92.882 -80.874 53.231
8546 92.901 -80.867 53.220
8547 92.919 -80.859 53.210
8548 92.938 -80.851 53.200
8549 92.956 -80.844 53.193
8550 92.975 -80.836 53.184
8551 92.991 -80.826 53.176
8571 93.318 -80.614 53.055
8572 93.334 -80.604 53.053
8573 93.351 -80.593 53.051
8574 93.367 -80.582 53.049
8575 93.383 -80.572 53.048
8576 93.400 -80.561 53.047
8577 93.416 -80.551 53.047
8578 93.432 -80.540 53.048
8579 93.449 -80.529 53.048
8580 93.465 -80.519 53.048
8581 93.482 -80.508 53.048
8582 93.498 -80.497 53.048
8583 93.514 -80.487 53.047
8584 93.521 -80.477 53.044
8585 93.526 -80.469 53.038
8586 93.530 -80.465 53.032
8587 93.533 -80.460 53.026
8588 93.537 -80.456 53.019
8589 93.540 -80.451 53.012
8590 93.543 -80.447 53.005
8591 93.547 -80.442 52.998
8592 93.550 -80.438 52.990
8593 93.553 -80.432 52.981
8594 93.557 -80.428 52.972
8595 93.560 -80.423 52.964
8596 93.563 -80.419 52.954
8597 93.567 -80.414 52.945
8598 93.570 -80.410 52.937
8599 93.573 -80.405 52.928
8600 93.577 -80.400 52.919
9187 91.790 -81.255 53.395
9188 91.803 -81.249 53.373
9189 91.817 -81.244 53.352
9190 91.831 -81.238 53.332
9191 91.846 -81.233 53.312
9192 91.860 -81.227 53.294
9193 91.873 -81.221 53.277
9194 91.887 -81.215 53.260
9195 91.901 -81.210 53.245
9196 91.916 -81.204 53.231
9197 91.930 -81.199 53.217
9198 91.943 -81.193 53.204
9199 91.957 -81.187 53.191
9200 91.971 -81.182 53.180
9201 91.986 -81.176 53.169
9202 92.032 -81.184 53.158
9203 92.083 -81.191 53.147
9204 92.104 -81.180 53.137
9205 92.120 -81.168 53.127
9206 92.136 -81.158 53.117
9207 92.152 -81.147 53.106
9208 92.170 -81.138 53.096
9209 92.190 -81.131 53.086
9210 92.209 -81.124 53.076
9211 92.228 -81.118 53.067
9212 92.248 -81.112 53.059
9213 92.268 -81.105 53.050
9214 92.287 -81.098 53.043
9215 92.305 -81.090 53.035
9216 92.323 -81.083 53.027
9217 92.342 -81.075 53.020
9218 92.360 -81.068 53.014
9219 92.379 -81.060 53.007
9232 92.619 -80.961 53.414
9233 92.638 -80.953 53.399
9234 92.656 -80.945 53.384
9235 92.674 -80.938 53.370
9236 92.693 -80.930 53.355
9237 92.712 -80.923 53.339
9238 92.730 -80.915 53.325
9239 92.748 -80.908 53.310
9240 92.767 -80.900 53.296
9241 92.785 -80.892 53.281
9242 92.804 -80.884 53.267
9243 92.823 -80.877 53.254
9244 92.841 -80.869 53.241
9245 92.859 -80.861 53.230
9246 92.878 -80.854 53.218
9247 92.896 -80.846 53.207
9248 92.915 -80.839 53.197
9249 92.933 -80.831 53.187
9250 92.951 -80.823 53.179
9251 92.970 -80.815 53.171
9252 92.986 -80.804 53.163
9253 93.002 -80.794 53.155
9254 93.018 -80.783 53.148
9255 93.035 -80.773 53.141
9256 93.051 -80.762 53.135
9257 93.068 -80.751 53.129
9272 93.313 -80.593 53.044
9273 93.329 -80.582 53.042
9274 93.346 -80.572 53.041
9275 93.362 -80.561 53.040
9276 93.378 -80.551 53.040
9277 93.395 -80.540 53.039
9278 93.411 -80.529 53.039
9279 93.427 -80.519 53.040
9280 93.444 -80.508 53.041
9281 93.460 -80.497 53.043
9282 93.477 -80.487 53.044
9283 93.493 -80.476 53.045
9284 93.509 -80.466 53.045
9285 93.516 -80.460 53.042
9286 93.520 -80.455 53.036
9287 93.523 -80.450 53.029
9288 93.526 -80.446 53.022
9289 93.530 -80.441 53.016
9290 93.533 -80.437 53.009
9291 93.537 -80.432 53.003
9292 93.540 -80.428 52.996
9293 93.543 -80.423 52.989
9294 93.547 -80.418 52.980
9295 93.550 -80.413 52.971
9296 93.553 -80.409 52.964
9297 93.557 -80.404 52.954
9298 93.560 -80.400 52.946
9299 93.563 -80.395 52.937
9300 93.567 -80.391 52.928
9301 93.570 -80.386 52.919
9302 93.574 -80.381 52.910
9888 91.788 -81.225 53.418
9889 91.802 -81.219 53.395
9890 91.815 -81.214 53.373
9891 91.829 -81.209 53.352
9892 91.844 -81.202 53.331
9893 91.858 -81.197 53.313
9894 91.872 -81.191 53.294
9895 91.885 -81.186 53.277
9896 91.899 -81.180 53.260
9897 91.914 -81.174 53.245
9898 91.928 -81.169 53.230
9899 91.942 -81.163 53.216
9900 91.955 -81.158 53.202
9901 91.969 -81.152 53.190
9902 91.984 -81.146 53.178
9903 92.018 -81.149 53.166
9904 92.067 -81.160 53.154
9905 92.091 -81.157 53.143
9906 92.111 -81.147 53.132
9907 92.131 -81.137 53.121
9908 92.147 -81.126 53.109
9909 92.163 -81.115 53.099
9910 92.182 -81.108 53.088
9911 92.202 -81.102 53.078
9912 92.221 -81.095 53.068
9913 92.240 -81.088 53.059
9914 92.260 -81.082 53.050
9915 92.279 -81.075 53.042
9916 92.299 -81.068 53.034
9917 92.318 -81.061 53.026
9918 92.338 -81.055 53.019
9919 92.356 -81.047 53.011
9920 92.374 -81.039 53.004
9933 92.615 -80.940 53.409
9934 92.633 -80.933 53.394
9935 92.651 -80.925 53.379
9936 92.670 -80.917 53.363
9937 92.689 -80.910 53.348
9938 92.707 -80.902 53.333
9939 92.725 -80.895 53.317
9940 92.744 -80.887 53.302
9941 92.762 -80.880 53.286
9942 92.781 -80.872 53.271
9943 92.799 -80.864 53.257
9944 92.818 -80.857 53.244
9945 92.836 -80.848 53.231
9946 92.855 -80.841 53.218
9947 92.873 -80.833 53.206
9948 92.892 -80.826 53.195
9949 92.910 -80.818 53.184
9950 92.929 -80.811 53.174
9951 92.947 -80.803 53.165
9952 92.965 -80.794 53.158
9953 92.981 -80.783 53.150
9954 92.997 -80.773 53.141
9955 93.013 -80.762 53.135
9956 93.030 -80.751 53.128
9957 93.046 -80.741 53.121
9958 93.063 -80.730 53.115
9959 93.079 -80.720 53.109
9960 93.095 -80.710 53.103
9961 93.112 -80.699 53.097
9962 93.128 -80.688 53.090
9972 93.292 -80.582 53.036
9973 93.308 -80.572 53.033
9974 93.324 -80.561 53.032
9975 93.341 -80.551 53.030
9976 93.357 -80.540 53.029
9977 93.373 -80.529 53.029
9978 93.390 -80.519 53.029
9979 93.406 -80.508 53.030
9980 93.422 -80.498 53.032
9981 93.439 -80.488 53.034
9982 93.455 -80.477 53.037
9983 93.468 -80.466 53.040
9984 93.481 -80.457 53.042
9985 93.497 -80.450 53.042
9986 93.509 -80.445 53.039
9987 93.513 -80.441 53.032
9988 93.516 -80.436 53.025
9989 93.520 -80.431 53.018
9990 93.523 -80.427 53.012
9991 93.526 -80.422 53.006
9992 93.530 -80.418 53.000
9993 93.534 -80.413 52.994
9994 93.537 -80.409 52.987
9995 93.540 -80.403 52.979
9996 93.543 -80.399 52.971
9997 93.547 -80.394 52.963
9998 93.550 -80.390 52.954
9999 93.553 -80.385 52.945
10000 93.557 -80.381 52.936
10001 93.560 -80.376 52.927
10002 93.564 -80.371 52.919
10003 93.574 -80.363 52.910
10589 91.786 -81.195 53.441
10590 91.800 -81.190 53.418
10591 91.814 -81.184 53.395
10592 91.828 -81.178 53.373
10593 91.842 -81.173 53.352
10594 91.856 -81.167 53.332
10595 91.870 -81.162 53.313
10596 91.884 -81.155 53.294
10597 91.898 -81.150 53.277
10598 91.912 -81.144 53.261
10599 91.926 -81.139 53.245
10600 91.940 -81.133 53.229
10601 91.954 -81.127 53.215
10602 91.968 -81.122 53.201
10603 91.982 -81.116 53.187
10604 92.003 -81.113 53.174
10605 92.038 -81.129 53.162
10606 92.066 -81.133 53.150
10607 92.086 -81.123 53.138
10608 92.107 -81.112 53.125
10609 92.127 -81.102 53.114
10610 92.150 -81.093 53.103
10611 92.174 -81.085 53.092
10612 92.194 -81.078 53.081
10613 92.213 -81.071 53.070
10614 92.233 -81.065 53.060
10615 92.252 -81.058 53.051
10616 92.272 -81.052 53.042
10617 92.291 -81.045 53.034
10618 92.311 -81.039 53.026
10619 92.331 -81.032 53.018
10620 92.350 -81.025 53.011
10621 92.369 -81.019 53.003
10634 92.610 -80.920 53.405
10635 92.629 -80.912 53.390
10636 92.647 -80.905 53.373
10637 92.665 -80.897 53.357
10638 92.684 -80.889 53.342
10639 92.702 -80.882 53.327
10640 92.721 -80.874 53.310
10641 92.739 -80.867 53.294
10642 92.758 -80.859 53.278
10643 92.776 -80.851 53.263
10644 92.795 -80.844 53.248
10645 92.813 -80.836 53.234
10646 92.832 -80.829 53.220
10647 92.850 -80.821 53.207
10648 92.868 -80.813 53.195
10649 92.887 -80.805 53.183
10650 92.906 -80.798 53.172
10651 92.924 -80.790 53.162
10652 92.942 -80.782 53.152
10653 92.959 -80.773 53.145
10654 92.976 -80.762 53.137
10655 92.992 -80.752 53.128
10656 93.008 -80.742 53.121
10657 93.025 -80.731 53.114
10658 93.041 -80.720 53.107
10659 93.058 -80.710 53.102
10660 93.074 -80.699 53.096
10661 93.090 -80.688 53.090
10662 93.107 -80.678 53.083
10663 93.123 -80.667 53.077
10664 93.139 -80.657 53.071
10665 93.156 -80.646 53.064
10673 93.286 -80.561 53.023
10674 93.303 -80.551 53.021
10675 93.319 -80.541 53.020
10676 93.336 -80.530 53.018
10677 93.352 -80.519 53.017
10678 93.368 -80.509 53.017
10679 93.384 -80.498 53.018
10680 93.401 -80.488 53.020
10681 93.414 -80.476 53.022
10682 93.426 -80.465 53.025
10683 93.439 -80.454 53.030
10684 93.452 -80.449 53.034
10685 93.468 -80.443 53.037
10686 93.485 -80.435 53.037
10687 93.498 -80.430 53.034
10688 93.506 -80.426 53.029
10689 93.510 -80.422 53.021
10690 93.513 -80.417 53.014
10691 93.516 -80.413 53.009
10692 93.520 -80.408 53.003
10693 93.523 -80.403 52.998
10694 93.526 -80.399 52.993
10695 93.530 -80.394 52.985
10696 93.534 -80.389 52.977
10697 93.537 -80.385 52.970
10698 93.540 -80.380 52.962
10699 93.543 -80.375 52.953
10700 93.547 -80.371 52.944
10701 93.550 -80.366 52.935
10702 93.554 -80.361 52.927
10703 93.564 -80.353 52.919
10704 93.574 -80.344 52.910
10705 93.584 -80.335 52.901
11291 91.798 -81.159 53.441
11292 91.812 -81.154 53.417
11293 91.826 -81.149 53.394
11294 91.840 -81.143 53.373
11295 91.854 -81.137 53.352
11296 91.868 -81.131 53.332
11297 91.882 -81.126 53.312
11298 91.896 -81.120 53.295
11299 91.910 -81.115 53.278
11300 91.924 -81.108 53.260
11301 91.938 -81.103 53.244
11302 91.952 -81.098 53.229
11303 91.965 -81.096 53.214
11304 91.979 -81.095 53.199
11305 91.992 -81.095 53.186
11306 92.010 -81.098 53.173
11307 92.040 -81.108 53.159
11308 92.061 -81.098 53.145
11309 92.082 -81.088 53.133
11310 92.112 -81.079 53.121
11311 92.143 -81.070 53.108
11312 92.164 -81.063 53.096
11313 92.184 -81.055 53.084
11314 92.205 -81.049 53.073
11315 92.224 -81.042 53.062
11316 92.244 -81.035 53.052
11317 92.265 -81.028 53.044
11318 92.284 -81.022 53.035
11319 92.303 -81.015 53.026
11320 92.323 -81.008 53.018
11321 92.342 -81.002 53.011
11322 92.362 -80.996 53.003
11336 92.624 -80.892 53.384
11337 92.642 -80.884 53.368
11338 92.661 -80.877 53.352
11339 92.679 -80.869 53.336
11340 92.698 -80.861 53.320
11341 92.716 -80.854 53.303
11342 92.735 -80.846 53.287
11343 92.753 -80.839 53.271
11344 92.772 -80.831 53.255
11345 92.790 -80.823 53.239
11346 92.808 -80.816 53.225
11347 92.827 -80.808 53.210
11348 92.846 -80.801 53.197
11349 92.864 -80.793 53.184
11350 92.882 -80.786 53.172
11351 92.901 -80.777 53.160
11352 92.919 -80.770 53.150
11353 92.938 -80.762 53.140
11354 92.954 -80.752 53.132
11355 92.971 -80.742 53.124
11356 92.987 -80.731 53.116
11357 93.003 -80.720 53.108
11358 93.020 -80.710 53.101
11359 93.036 -80.699 53.094
11360 93.053 -80.688 53.088
11361 93.069 -80.678 53.082
11362 93.085 -80.667 53.076
11363 93.101 -80.657 53.070
11364 93.118 -80.646 53.064
11365 93.134 -80.635 53.058
11366 93.151 -80.625 53.052
11367 93.167 -80.614 53.046
11374 93.281 -80.541 53.011
11375 93.298 -80.530 53.009
11376 93.314 -80.519 53.007
11377 93.331 -80.509 53.006
11378 93.347 -80.498 53.005
11379 93.360 -80.487 53.004
11380 93.372 -80.476 53.006
11381 93.385 -80.465 53.008
11382 93.398 -80.453 53.011
11383 93.410 -80.444 53.016
11384 93.424 -80.441 53.021
11385 93.440 -80.435 53.027
11386 93.456 -80.428 53.029
11387 93.473 -80.420 53.029
11388 93.486 -80.415 53.027
11389 93.496 -80.412 53.023
11390 93.503 -80.407 53.017
11391 93.507 -80.403 53.011
11392 93.510 -80.398 53.007
11393 93.513 -80.394 53.002
11394 93.516 -80.389 52.997
11395 93.520 -80.385 52.991
11396 93.523 -80.380 52.984
11397 93.526 -80.375 52.977
11398 93.530 -80.370 52.969
11399 93.534 -80.366 52.961
11400 93.537 -80.361 52.952
11401 93.540 -80.356 52.943
11402 93.544 -80.352 52.934
11403 93.554 -80.343 52.926
11404 93.564 -80.334 52.918
11405 93.574 -80.325 52.909
11406 93.584 -80.317 52.901
11407 93.594 -80.309 52.892
11992 91.796 -81.130 53.465
11993 91.810 -81.124 53.441
11994 91.824 -81.118 53.417
11995 91.838 -81.113 53.395
11996 91.852 -81.107 53.373
11997 91.866 -81.102 53.352
11998 91.880 -81.096 53.332
11999 91.894 -81.090 53.314
12000 91.908 -81.089 53.296
12001 91.921 -81.088 53.277
12002 91.934 -81.087 53.261
12003 91.947 -81.086 53.245
12004 91.961 -81.086 53.229
12005 91.974 -81.085 53.213
12006 91.987 -81.084 53.198
12007 92.001 -81.083 53.184
12008 92.015 -81.083 53.169
12009 92.040 -81.075 53.155
12010 92.065 -81.067 53.141
12011 92.090 -81.059 53.128
12012 92.115 -81.051 53.115
12013 92.140 -81.043 53.102
12014 92.165 -81.035 53.090
12015 92.190 -81.027 53.078
12016 92.215 -81.020 53.067
12017 92.235 -81.013 53.056
12018 92.255 -81.006 53.047
12019 92.275 -80.999 53.037
12020 92.295 -80.992 53.028
12021 92.315 -80.986 53.020
12022 92.335 -80.979 53.011
12023 92.354 -80.972 53.003
12038 92.638 -80.864 53.364
12039 92.656 -80.856 53.348
12040 92.674 -80.848 53.332
12041 92.693 -80.841 53.315
12042 92.712 -80.833 53.297
12043 92.730 -80.826 53.281
12044 92.748 -80.818 53.264
12045 92.767 -80.811 53.247
12046 92.785 -80.803 53.231
12047 92.804 -80.795 53.216
12048 92.823 -80.788 53.201
12049 92.841 -80.780 53.187
12050 92.859 -80.773 53.173
12051 92.878 -80.765 53.161
12052 92.897 -80.757 53.149
12053 92.915 -80.750 53.137
12054 92.933 -80.742 53.128
12055 92.949 -80.731 53.120
12056 92.966 -80.720 53.111
12057 92.982 -80.710 53.103
12058 92.998 -80.699 53.096
12059 93.015 -80.688 53.089
12060 93.031 -80.678 53.082
12061 93.048 -80.667 53.075
12062 93.064 -80.657 53.069
12063 93.080 -80.646 53.063
12064 93.096 -80.635 53.057
12065 93.113 -80.626 53.051
12066 93.129 -80.615 53.045
12067 93.146 -80.604 53.039
12068 93.162 -80.594 53.034
12069 93.179 -80.583 53.028
12070 93.195 -80.573 53.022
12075 93.276 -80.519 52.999
12076 93.293 -80.509 52.996
12077 93.306 -80.498 52.994
12078 93.319 -80.487 52.992
12079 93.331 -80.475 52.991
12080 93.343 -80.464 52.991
12081 93.356 -80.453 52.993
12082 93.368 -80.442 52.996
12083 93.382 -80.437 53.000
12084 93.396 -80.433 53.006
12085 93.412 -80.427 53.012
12086 93.428 -80.420 53.018
12087 93.444 -80.413 53.021
12088 93.460 -80.406 53.021
12089 93.475 -80.400 53.019
12090 93.485 -80.396 53.016
12091 93.494 -80.393 53.012
12092 93.500 -80.388 53.009
12093 93.503 -80.384 53.005
12094 93.507 -80.379 53.001
12095 93.510 -80.375 52.996
12096 93.513 -80.370 52.991
12097 93.517 -80.366 52.983
12098 93.520 -80.361 52.976
12099 93.523 -80.356 52.968
12100 93.526 -80.351 52.960
12101 93.530 -80.347 52.951
12102 93.534 -80.342 52.943
12103 93.544 -80.333 52.935
12104 93.554 -80.325 52.926
12105 93.564 -80.316 52.917
12106 93.574 -80.307 52.910
12107 93.584 -80.299 52.901
12108 93.594 -80.290 52.892
12693 91.794 -81.099 53.489
12694 91.808 -81.094 53.465
12695 91.822 -81.089 53.442
12696 91.836 -81.083 53.418
12697 91.850 -81.081 53.396
12698 91.863 -81.080 53.375
12699 91.876 -81.080 53.354
12700 91.889 -81.079 53.335
12701 91.903 -81.078 53.316
12702 91.916 -81.077 53.297
12703 91.930 -81.077 53.279
12704 91.943 -81.076 53.261
12705 91.956 -81.075 53.244
12706 91.970 -81.075 53.228
12707 91.985 -81.074 53.212
12708 91.999 -81.074 53.196
12709 92.013 -81.069 53.181
12710 92.027 -81.060 53.165
12711 92.042 -81.050 53.151
12712 92.060 -81.042 53.137
12713 92.082 -81.033 53.123
12714 92.107 -81.026 53.109
12715 92.132 -81.017 53.097
12716 92.157 -81.010 53.085
12717 92.181 -81.002 53.073
12718 92.207 -80.994 53.062
12719 92.232 -80.986 53.051
12720 92.258 -80.978 53.041
12721 92.284 -80.971 53.032
12722 92.306 -80.963 53.023
12723 92.325 -80.956 53.014
12724 92.346 -80.949 53.005
12740 92.652 -80.836 53.344
12741 92.670 -80.828 53.328
12742 92.689 -80.820 53.310
12743 92.707 -80.813 53.293
12744 92.725 -80.805 53.276
12745 92.744 -80.798 53.258
12746 92.763 -80.790 53.242
12747 92.781 -80.782 53.225
12748 92.799 -80.775 53.209
12749 92.818 -80.767 53.194
12750 92.836 -80.760 53.179
12751 92.855 -80.752 53.164
12752 92.873 -80.745 53.152
12753 92.892 -80.737 53.140
12754 92.910 -80.729 53.128
12755 92.928 -80.720 53.118
12756 92.944 -80.710 53.108
12757 92.960 -80.699 53.100
12758 92.977 -80.688 53.092
12759 92.993 -80.678 53.084
12760 93.010 -80.668 53.077
12761 93.026 -80.657 53.070
12762 93.043 -80.647 53.063
12763 93.059 -80.636 53.057
12764 93.075 -80.626 53.050
12765 93.091 -80.615 53.045
12766 93.108 -80.604 53.039
12767 93.124 -80.594 53.033
12768 93.141 -80.583 53.027
12769 93.157 -80.573 53.022
12770 93.174 -80.562 53.017
12771 93.190 -80.551 53.011
12772 93.206 -80.541 53.005
12773 93.222 -80.530 52.999
12774 93.238 -80.519 52.994
12775 93.252 -80.509 52.990
12776 93.265 -80.497 52.986
12777 93.277 -80.486 52.983
12778 93.289 -80.475 52.981
12779 93.302 -80.464 52.979
12780 93.314 -80.453 52.979
12781 93.327 -80.441 52.979
12782 93.340 -80.433 52.981
12783 93.353 -80.429 52.983
12784 93.367 -80.425 52.986
12785 93.383 -80.419 52.992
12786 93.400 -80.412 52.999
12787 93.416 -80.405 53.006
12788 93.432 -80.397 53.010
12789 93.448 -80.391 53.012
12790 93.463 -80.385 53.011
12791 93.473 -80.381 53.009
12792 93.483 -80.378 53.008
12793 93.493 -80.374 53.007
12794 93.497 -80.369 53.006
12795 93.500 -80.365 53.002
12796 93.503 -80.360 52.997
12797 93.507 -80.356 52.992
12798 93.510 -80.351 52.983
12799 93.513 -80.347 52.975
12800 93.517 -80.341 52.966
12801 93.520 -80.337 52.959
12802 93.523 -80.332 52.951
12803 93.534 -80.324 52.942
12804 93.544 -80.315 52.934
12805 93.554 -80.306 52.925
12806 93.564 -80.297 52.916
12807 93.574 -80.289 52.909
12808 93.584 -80.280 52.900
12809 93.594 -80.272 52.892
13394 91.792 -81.074 53.513
13395 91.805 -81.073 53.490
13396 91.818 -81.072 53.466
13397 91.832 -81.071 53.442
13398 91.845 -81.071 53.420
13399 91.858 -81.070 53.398
13400 91.872 -81.069 53.377
13401 91.885 -81.068 53.356
13402 91.898 -81.068 53.336
13403 91.913 -81.068 53.316
13404 91.927 -81.067 53.298
13405 91.942 -81.066 53.280
13406 91.956 -81.066 53.262
13407 91.970 -81.065 53.245
13408 91.985 -81.065 53.228
13409 91.999 -81.060 53.211
13410 92.012 -81.055 53.194
13411 92.026 -81.050 53.178
13412 92.040 -81.038 53.162
13413 92.054 -81.026 53.148
13414 92.070 -81.017 53.133
13415 92.086 -81.009 53.119
13416 92.106 -81.001 53.105
13417 92.127 -80.992 53.093
13418 92.149 -80.984 53.080
13419 92.173 -80.976 53.068
13420 92.199 -80.968 53.057
13421 92.223 -80.960 53.047
13422 92.248 -80.952 53.037
13423 92.274 -80.944 53.027
13424 92.299 -80.936 53.017
13442 92.665 -80.808 53.325
13443 92.684 -80.800 53.308
13444 92.702 -80.792 53.290
13445 92.721 -80.785 53.273
13446 92.739 -80.777 53.254
13447 92.758 -80.770 53.237
13448 92.776 -80.762 53.220
13449 92.795 -80.754 53.204
13450 92.813 -80.747 53.188
13451 92.832 -80.739 53.173
13452 92.850 -80.732 53.157
13453 92.869 -80.724 53.144
13454 92.887 -80.717 53.132
13455 92.906 -80.709 53.120
13456 92.923 -80.700 53.109
13457 92.939 -80.689 53.100
13458 92.955 -80.679 53.091
13459 92.972 -80.668 53.083
13460 92.988 -80.657 53.075
13461 93.005 -80.647 53.067
13462 93.021 -80.636 53.060
13463 93.038 -80.626 53.053
13464 93.054 -80.615 53.047
13465 93.070 -80.604 53.041
13466 93.086 -80.594 53.034
13467 93.103 -80.583 53.028
13468 93.119 -80.573 53.022
13469 93.136 -80.562 53.016
13470 93.152 -80.551 53.010
13471 93.168 -80.541 53.005
13472 93.185 -80.530 52.999
13473 93.198 -80.519 52.993
13474 93.211 -80.508 52.987
13475 93.223 -80.497 52.982
13476 93.235 -80.486 52.977
13477 93.248 -80.475 52.973
13478 93.260 -80.463 52.970
13479 93.273 -80.452 52.967
13480 93.285 -80.441 52.966
13481 93.298 -80.430 52.965
13482 93.311 -80.425 52.966
13483 93.325 -80.422 52.967
13484 93.339 -80.418 52.969
13485 93.355 -80.411 52.973
13486 93.371 -80.404 52.979
13487 93.388 -80.397 52.986
13488 93.404 -80.390 52.993
13489 93.420 -80.383 52.998
13490 93.436 -80.375 53.001
13491 93.451 -80.369 53.001
13492 93.462 -80.366 53.002
13493 93.472 -80.362 53.002
13494 93.481 -80.359 53.002
13495 93.490 -80.355 53.002
13496 93.493 -80.350 52.999
13497 93.497 -80.346 52.996
13498 93.500 -80.341 52.992
13499 93.503 -80.337 52.983
13500 93.507 -80.332 52.974
13501 93.510 -80.327 52.965
13502 93.513 -80.322 52.956
13503 93.523 -80.314 52.948
13504 93.534 -80.305 52.940
13505 93.544 -80.297 52.931
13506 93.554 -80.288 52.923
13507 93.564 -80.279 52.915
13508 93.574 -80.271 52.906
13509 93.584 -80.262 52.899
13510 93.594 -80.253 52.891
13511 93.604 -80.244 52.884
14096 91.801 -81.062 53.514
14097 91.814 -81.061 53.490
14098 91.827 -81.061 53.467
14099 91.841 -81.060 53.445
14100 91.856 -81.060 53.422
14101 91.870 -81.059 53.399
14102 91.884 -81.059 53.378
14103 91.898 -81.058 53.357
14104 91.913 -81.058 53.336
14105 91.927 -81.058 53.318
14106 91.942 -81.057 53.299
14107 91.956 -81.056 53.280
14108 91.970 -81.055 53.262
14109 91.984 -81.051 53.245
14110 91.998 -81.046 53.226
14111 92.011 -81.041 53.209
14112 92.025 -81.036 53.192
14113 92.039 -81.027 53.176
14114 92.053 -81.015 53.160
14115 92.067 -81.003 53.146
14116 92.082 -80.992 53.131
14117 92.098 -80.984 53.116
14118 92.114 -80.976 53.103
14119 92.131 -80.967 53.089
14120 92.152 -80.959 53.077
14121 92.173 -80.951 53.065
14122 92.194 -80.942 53.054
14123 92.215 -80.934 53.043
14124 92.240 -80.927 53.032
14125 92.265 -80.918 53.022
14144 92.680 -80.780 53.307
14145 92.698 -80.773 53.289
14146 92.716 -80.764 53.272
14147 92.735 -80.757 53.253
14148 92.753 -80.749 53.235
14149 92.772 -80.742 53.218
14150 92.790 -80.734 53.200
14151 92.808 -80.726 53.183
14152 92.827 -80.719 53.168
14153 92.846 -80.711 53.152
14154 92.864 -80.704 53.138
14155 92.882 -80.696 53.126
14156 92.901 -80.688 53.114
14157 92.918 -80.679 53.103
14158 92.934 -80.668 53.093
14159 92.950 -80.657 53.084
14160 92.967 -80.647 53.075
14161 92.983 -80.636 53.067
14162 93.000 -80.626 53.059
14163 93.016 -80.615 53.051
14164 93.032 -80.604 53.045
14165 93.049 -80.594 53.038
14166 93.065 -80.583 53.031
14167 93.081 -80.573 53.024
14168 93.098 -80.562 53.018
14169 93.114 -80.551 53.012
14170 93.131 -80.541 53.005
14171 93.144 -80.530 52.999
14172 93.156 -80.519 52.993
14173 93.169 -80.508 52.987
14174 93.182 -80.497 52.981
14175 93.194 -80.485 52.975
14176 93.206 -80.474 52.969
14177 93.219 -80.463 52.964
14178 93.231 -80.452 52.960
14179 93.243 -80.441 52.956
14180 93.256 -80.429 52.953
14181 93.269 -80.421 52.951
14182 93.283 -80.417 52.951
14183 93.297 -80.413 52.952
14184 93.310 -80.410 52.952
14185 93.327 -80.403 52.954
14186 93.343 -80.397 52.959
14187 93.359 -80.389 52.965
14188 93.375 -80.382 52.971
14189 93.392 -80.375 52.978
14190 93.408 -80.368 52.984
14191 93.424 -80.360 52.988
14192 93.440 -80.353 52.991
14193 93.450 -80.350 52.993
14194 93.460 -80.347 52.994
14195 93.470 -80.344 52.995
14196 93.480 -80.340 52.995
14197 93.487 -80.336 52.995
14198 93.490 -80.331 52.992
14199 93.493 -80.327 52.989
14200 93.497 -80.322 52.980
14201 93.500 -80.318 52.971
14202 93.503 -80.312 52.961
14203 93.513 -80.304 52.952
14204 93.523 -80.296 52.944
14205 93.534 -80.287 52.935
14206 93.544 -80.278 52.927
14207 93.554 -80.269 52.919
14208 93.564 -80.261 52.911
14209 93.574 -80.252 52.903
14210 93.584 -80.243 52.897
14211 93.594 -80.235 52.890
14212 93.604 -80.226 52.882
14797 91.798 -81.052 53.539
14798 91.812 -81.052 53.515
14799 91.826 -81.052 53.492
14800 91.841 -81.051 53.469
14801 91.856 -81.050 53.446
14802 91.870 -81.050 53.424
14803 91.884 -81.049 53.401
14804 91.898 -81.049 53.380
14805 91.913 -81.049 53.359
14806 91.927 -81.048 53.339
14807 91.942 -81.048 53.319
14808 91.956 -81.046 53.299
14809 91.969 -81.042 53.280
14810 91.984 -81.036 53.262
14811 91.997 -81.032 53.243
14812 92.011 -81.027 53.225
14813 92.024 -81.023 53.208
14814 92.038 -81.017 53.190
14815 92.052 -81.005 53.175
14816 92.066 -80.992 53.159
14817 92.080 -80.980 53.144
14818 92.094 -80.968 53.129
14819 92.109 -80.959 53.114
14820 92.126 -80.951 53.100
14821 92.142 -80.943 53.087
14822 92.158 -80.935 53.074
14823 92.177 -80.927 53.063
14824 92.198 -80.918 53.051
14825 92.219 -80.910 53.040
14826 92.239 -80.902 53.029
14846 92.693 -80.752 53.289
14847 92.712 -80.745 53.271
14848 92.730 -80.737 53.253
14849 92.748 -80.729 53.234
14850 92.767 -80.721 53.216
14851 92.786 -80.714 53.199
14852 92.804 -80.706 53.182
14853 92.823 -80.698 53.165
14854 92.841 -80.691 53.150
14855 92.859 -80.683 53.136
14856 92.878 -80.676 53.122
14857 92.897 -80.668 53.109
14858 92.913 -80.657 53.099
14859 92.929 -80.647 53.088
14860 92.945 -80.636 53.079
14861 92.962 -80.626 53.069
14862 92.978 -80.615 53.060
14863 92.995 -80.604 53.052
14864 93.011 -80.594 53.044
14865 93.027 -80.584 53.037
14866 93.044 -80.573 53.029
14867 93.060 -80.563 53.022
14868 93.076 -80.552 53.015
14869 93.090 -80.541 53.009
14870 93.102 -80.529 53.002
14871 93.115 -80.519 52.995
14872 93.128 -80.507 52.988
14873 93.140 -80.496 52.982
14874 93.152 -80.485 52.976
14875 93.165 -80.474 52.969
14876 93.178 -80.463 52.962
14877 93.190 -80.451 52.956
14878 93.202 -80.441 52.951
14879 93.215 -80.429 52.946
14880 93.227 -80.418 52.943
14881 93.240 -80.413 52.940
14882 93.255 -80.410 52.938
14883 93.268 -80.406 52.936
14884 93.282 -80.402 52.937
14885 93.298 -80.396 52.938
14886 93.314 -80.388 52.941
14887 93.331 -80.381 52.945
14888 93.347 -80.375 52.951
14889 93.363 -80.367 52.956
14890 93.379 -80.360 52.961
14891 93.396 -80.353 52.968
14892 93.412 -80.346 52.973
14893 93.428 -80.338 52.978
14894 93.438 -80.335 52.981
14895 93.448 -80.331 52.983
14896 93.458 -80.328 52.985
14897 93.468 -80.325 52.986
14898 93.478 -80.322 52.987
14899 93.483 -80.317 52.986
14900 93.487 -80.312 52.983
14901 93.490 -80.308 52.975
14902 93.493 -80.303 52.966
14903 93.503 -80.294 52.955
14904 93.513 -80.286 52.946
14905 93.523 -80.277 52.938
14906 93.534 -80.269 52.930
14907 93.544 -80.260 52.922
14908 93.554 -80.251 52.915
14909 93.564 -80.243 52.908
14910 93.574 -80.234 52.900
14911 93.584 -80.225 52.894
14912 93.594 -80.216 52.887
14913 93.604 -80.208 52.879
15498 91.798 -81.043 53.563
15499 91.812 -81.043 53.540
15500 91.826 -81.042 53.516
15501 91.841 -81.042 53.493
15502 91.856 -81.041 53.470
15503 91.870 -81.040 53.447
15504 91.884 -81.040 53.425
15505 91.898 -81.039 53.402
15506 91.913 -81.039 53.381
15507 91.927 -81.039 53.360
15508 91.942 -81.037 53.339
15509 91.955 -81.033 53.319
15510 91.969 -81.027 53.300
15511 91.983 -81.023 53.281
15512 91.996 -81.018 53.262
15513 92.010 -81.013 53.243
15514 92.023 -81.008 53.225
15515 92.037 -81.004 53.207
15516 92.051 -80.995 53.190
15517 92.065 -80.983 53.174
15518 92.079 -80.971 53.158
15519 92.093 -80.958 53.142
15520 92.107 -80.945 53.128
15521 92.122 -80.934 53.113
15522 92.138 -80.926 53.098
15523 92.154 -80.918 53.085
15524 92.170 -80.910 53.073
15525 92.186 -80.902 53.061
15526 92.202 -80.894 53.050
15548 92.707 -80.723 53.272
15549 92.725 -80.717 53.253
15550 92.744 -80.709 53.235
15551 92.763 -80.701 53.216
15552 92.781 -80.694 53.199
15553 92.799 -80.685 53.182
15554 92.818 -80.678 53.165
15555 92.837 -80.670 53.149
15556 92.855 -80.663 53.135
15557 92.873 -80.655 53.121
15558 92.892 -80.647 53.108
15559 92.908 -80.636 53.098
15560 92.924 -80.626 53.088
15561 92.940 -80.616 53.078
15562 92.957 -80.605 53.067
15563 92.973 -80.594 53.057
15564 92.990 -80.584 53.049
15565 93.006 -80.573 53.040
15566 93.022 -80.563 53.032
15567 93.036 -80.551 53.023
15568 93.049 -80.541 53.016
15569 93.061 -80.529 53.008
15570 93.074 -80.518 53.001
15571 93.086 -80.507 52.993
15572 93.098 -80.496 52.986
15573 93.111 -80.485 52.979
15574 93.123 -80.473 52.972
15575 93.136 -80.462 52.966
15576 93.148 -80.451 52.958
15577 93.160 -80.440 52.951
15578 93.173 -80.428 52.944
15579 93.186 -80.418 52.938
15580 93.198 -80.409 52.933
15581 93.212 -80.405 52.930
15582 93.226 -80.401 52.927
15583 93.240 -80.398 52.924
15584 93.254 -80.394 52.922
15585 93.270 -80.388 52.923
15586 93.286 -80.381 52.924
15587 93.302 -80.374 52.927
15588 93.319 -80.366 52.931
15589 93.335 -80.359 52.936
15590 93.351 -80.352 52.941
15591 93.367 -80.345 52.946
15592 93.383 -80.337 52.952
15593 93.400 -80.331 52.958
15594 93.416 -80.324 52.962
15595 93.427 -80.319 52.966
15596 93.437 -80.316 52.969
15597 93.446 -80.312 52.970
15598 93.457 -80.309 52.972
15599 93.467 -80.306 52.972
15600 93.477 -80.303 52.971
15601 93.480 -80.298 52.969
15602 93.483 -80.294 52.963
15603 93.493 -80.284 52.956
15604 93.503 -80.276 52.947
15605 93.513 -80.268 52.938
15606 93.523 -80.259 52.930
15607 93.534 -80.250 52.923
15608 93.544 -80.241 52.916
15609 93.554 -80.233 52.910
15610 93.564 -80.224 52.904
15611 93.574 -80.215 52.896
15612 93.584 -80.207 52.890
15613 93.594 -80.198 52.884
15614 93.604 -80.190 52.876
16200 91.812 -81.033 53.563
16201 91.826 -81.033 53.540
16202 91.841 -81.032 53.516
16203 91.856 -81.032 53.493
16204 91.870 -81.031 53.470
16205 91.884 -81.030 53.448
16206 91.898 -81.030 53.425
16207 91.913 -81.030 53.404
16208 91.927 -81.028 53.382
16209 91.941 -81.024 53.361
16210 91.954 -81.018 53.341
16211 91.968 -81.014 53.321
16212 91.982 -81.009 53.301
16213 91.996 -81.004 53.282
16214 92.009 -80.999 53.263
16215 92.023 -80.995 53.244
16216 92.036 -80.990 53.226
16217 92.050 -80.984 53.208
16218 92.064 -80.972 53.191
16219 92.078 -80.960 53.175
16220 92.092 -80.948 53.159
16221 92.106 -80.936 53.143
16222 92.121 -80.924 53.128
16223 92.135 -80.911 53.112
16224 92.150 -80.901 53.099
16225 92.165 -80.893 53.086
16226 92.181 -80.885 53.073
16227 92.198 -80.877 53.061
16250 92.719 -80.694 53.256
16251 92.738 -80.687 53.237
16252 92.758 -80.680 53.218
16253 92.776 -80.673 53.200
16254 92.795 -80.666 53.184
16255 92.813 -80.658 53.167
16256 92.832 -80.650 53.150
16257 92.850 -80.642 53.136
16258 92.869 -80.635 53.122
16259 92.886 -80.626 53.110
16260 92.903 -80.616 53.099
16261 92.919 -80.605 53.090
16262 92.935 -80.594 53.080
16263 92.952 -80.584 53.068
16264 92.968 -80.573 53.058
16265 92.982 -80.563 53.048
16266 92.995 -80.551 53.040
16267 93.007 -80.540 53.030
16268 93.019 -80.529 53.021
16269 93.032 -80.518 53.012
16270 93.045 -80.507 53.003
16271 93.057 -80.495 52.995
16272 93.069 -80.484 52.986
16273 93.082 -80.473 52.979
16274 93.094 -80.462 52.971
16275 93.107 -80.450 52.963
16276 93.119 -80.439 52.955
16277 93.132 -80.428 52.948
16278 93.144 -80.417 52.941
16279 93.156 -80.406 52.934
16280 93.170 -80.401 52.927
16281 93.184 -80.397 52.922
16282 93.198 -80.394 52.917
16283 93.212 -80.390 52.913
16284 93.225 -80.387 52.911
16285 93.241 -80.380 52.909
16286 93.258 -80.373 52.909
16287 93.274 -80.366 52.911
16288 93.290 -80.359 52.913
16289 93.306 -80.352 52.916
16290 93.323 -80.344 52.921
16291 93.339 -80.337 52.926
16292 93.355 -80.330 52.932
16293 93.371 -80.323 52.937
16294 93.388 -80.316 52.942
16295 93.404 -80.309 52.946
16296 93.415 -80.304 52.949
16297 93.425 -80.301 52.952
16298 93.435 -80.297 52.954
16299 93.445 -80.294 52.954
16300 93.455 -80.290 52.955
16301 93.465 -80.287 52.954
16302 93.474 -80.284 52.953
16303 93.483 -80.275 52.949
16304 93.493 -80.266 52.942
16305 93.503 -80.258 52.935
16306 93.513 -80.249 52.928
16307 93.523 -80.240 52.922
16308 93.534 -80.232 52.916
16309 93.544 -80.223 52.910
16310 93.554 -80.215 52.904
16311 93.564 -80.206 52.899
16312 93.574 -80.197 52.893
16313 93.584 -80.188 52.887
16314 93.594 -80.180 52.880
16315 93.604 -80.171 52.873
16316 93.614 -80.162 52.867
16902 91.826 -81.024 53.563
16903 91.841 -81.023 53.540
16904 91.856 -81.022 53.517
16905 91.870 -81.022 53.494
16906 91.884 -81.021 53.472
16907 91.898 -81.021 53.449
16908 91.913 -81.019 53.427
16909 91.926 -81.014 53.406
16910 91.940 -81.009 53.384
16911 91.954 -81.005 53.363
16912 91.967 -81.000 53.343
16913 91.981 -80.995 53.322
16914 91.995 -80.990 53.303
16915 92.008 -80.986 53.284
16916 92.022 -80.980 53.265
16917 92.035 -80.976 53.246
16918 92.050 -80.971 53.228
16919 92.063 -80.962 53.210
16920 92.077 -80.949 53.193
16921 92.091 -80.937 53.176
16922 92.105 -80.925 53.160
16923 92.120 -80.913 53.144
16924 92.134 -80.901 53.129
16925 92.148 -80.889 53.114
16926 92.162 -80.877 53.100
16927 92.177 -80.868 53.086
16951 92.689 -80.674 53.260
16952 92.715 -80.667 53.241
16953 92.741 -80.659 53.223
16954 92.768 -80.651 53.205
16955 92.789 -80.644 53.188
16956 92.809 -80.638 53.171
16957 92.827 -80.630 53.154
16958 92.846 -80.623 53.139
16959 92.864 -80.614 53.125
16960 92.881 -80.605 53.113
16961 92.898 -80.594 53.104
16962 92.914 -80.584 53.095
16963 92.928 -80.573 53.085
16964 92.941 -80.562 53.074
16965 92.953 -80.551 53.063
16966 92.966 -80.540 53.053
16967 92.978 -80.529 53.042
16968 92.991 -80.517 53.033
16969 93.003 -80.506 53.023
16970 93.015 -80.495 53.011
16971 93.028 -80.484 53.000
16972 93.040 -80.472 52.991
16973 93.053 -80.461 52.981
16974 93.065 -80.450 52.973
16975 93.077 -80.439 52.964
16976 93.090 -80.428 52.956
16977 93.102 -80.416 52.947
16978 93.115 -80.406 52.939
16979 93.128 -80.397 52.932
16980 93.142 -80.393 52.924
16981 93.155 -80.390 52.917
16982 93.169 -80.386 52.911
16983 93.183 -80.382 52.905
16984 93.197 -80.378 52.901
16985 93.213 -80.372 52.898
16986 93.229 -80.365 52.897
16987 93.246 -80.358 52.897
16988 93.262 -80.351 52.898
16989 93.278 -80.344 52.900
16990 93.294 -80.337 52.905
16991 93.310 -80.329 52.909
16992 93.327 -80.322 52.914
16993 93.343 -80.316 52.918
16994 93.359 -80.308 52.922
16995 93.375 -80.301 52.925
16996 93.392 -80.294 52.929
16997 93.404 -80.289 52.931
16998 93.414 -80.286 52.932
16999 93.424 -80.282 52.934
17000 93.433 -80.279 52.935
17001 93.443 -80.275 52.935
17002 93.453 -80.272 52.936
17003 93.463 -80.256 52.936
17004 93.483 -80.256 52.932
17005 93.493 -80.248 52.927
17006 93.503 -80.240 52.922
17007 93.513 -80.231 52.917
17008 93.523 -80.222 52.913
17009 93.534 -80.213 52.909
17010 93.544 -80.205 52.904
17011 93.554 -80.196 52.898
17012 93.564 -80.187 52.894
17013 93.574 -80.179 52.889
17014 93.584 -80.170 52.883
17015 93.594 -80.162 52.877
17016 93.604 -80.153 52.871
17017 93.614 -80.144 52.865
17058 93.990 -79.964 52.391
17059 93.999 -79.962 52.375
17060 94.007 -79.961 52.361
17061 94.016 -79.959 52.346
17062 94.025 -79.958 52.330
17063 94.033 -79.956 52.315
17064 94.042 -79.955 52.299
17065 94.050 -79.953 52.284
17066 94.059 -79.952 52.268
17603 91.826 -81.014 53.587
17604 91.841 -81.013 53.564
17605 91.856 -81.013 53.541
17606 91.870 -81.012 53.518
17607 91.884 -81.012 53.496
17608 91.898 -81.010 53.473
17609 91.912 -81.005 53.451
17610 91.926 -81.000 53.429
17611 91.939 -80.996 53.407
17612 91.953 -80.991 53.387
17613 91.966 -80.986 53.366
17614 91.980 -80.981 53.346
17615 91.994 -80.977 53.326
17616 92.008 -80.971 53.307
17617 92.021 -80.967 53.287
17618 92.035 -80.962 53.268
17619 92.049 -80.958 53.249
17620 92.062 -80.952 53.230
17621 92.076 -80.939 53.213
17622 92.090 -80.927 53.196
17623 92.104 -80.915 53.179
17624 92.119 -80.903 53.162
17625 92.133 -80.890 53.147
17626 92.147 -80.878 53.131
17627 92.161 -80.866 53.116
17652 92.651 -80.655 53.266
17653 92.677 -80.648 53.247
17654 92.704 -80.640 53.230
17655 92.730 -80.632 53.211
17656 92.757 -80.625 53.193
17657 92.783 -80.617 53.176
17658 92.809 -80.610 53.160
17659 92.836 -80.601 53.145
17660 92.860 -80.594 53.131
17661 92.874 -80.584 53.120
17662 92.887 -80.573 53.110
17663 92.899 -80.561 53.101
17664 92.912 -80.551 53.091
17665 92.924 -80.539 53.080
17666 92.936 -80.528 53.070
17667 92.949 -80.517 53.059
17668 92.961 -80.506 53.048
17669 92.974 -80.494 53.038
17670 92.986 -80.483 53.027
17671 92.999 -80.472 53.014
17672 93.011 -80.461 53.001
17673 93.023 -80.450 52.989
17674 93.036 -80.438 52.978
17675 93.049 -80.428 52.968
17676 93.061 -80.416 52.958
17677 93.073 -80.405 52.950
17678 93.086 -80.394 52.941
17679 93.099 -80.389 52.932
17680 93.113 -80.385 52.924
17681 93.127 -80.381 52.915
17682 93.141 -80.378 52.908
17683 93.155 -80.375 52.901
17684 93.168 -80.371 52.895
17685 93.185 -80.364 52.891
17686 93.201 -80.357 52.888
17687 93.217 -80.350 52.887
17688 93.233 -80.343 52.886
17689 93.250 -80.336 52.888
17690 93.266 -80.329 52.890
17691 93.282 -80.322 52.895
17692 93.298 -80.315 52.899
17693 93.314 -80.307 52.902
17694 93.331 -80.300 52.905
17695 93.347 -80.293 52.907
17696 93.363 -80.286 52.909
17697 93.379 -80.278 52.911
17698 93.392 -80.274 52.912
17699 93.402 -80.270 52.914
17700 93.412 -80.267 52.916
17701 93.422 -80.263 52.917
17702 93.432 -80.260 52.917
17703 93.442 -80.243 52.917
17704 93.451 -80.225 52.917
17705 93.465 -80.226 52.916
17706 93.488 -80.227 52.913
17707 93.503 -80.221 52.909
17708 93.513 -80.212 52.906
17709 93.523 -80.204 52.903
17710 93.534 -80.195 52.901
17711 93.544 -80.187 52.897
17712 93.554 -80.178 52.892
17713 93.564 -80.169 52.889
17714 93.574 -80.160 52.884
17715 93.584 -80.152 52.879
17716 93.594 -80.143 52.874
17717 93.604 -80.134 52.869
17718 93.614 -80.126 52.863
17719 93.624 -80.117 52.857
17758 93.981 -79.951 52.416
17759 93.989 -79.949 52.401
17760 93.998 -79.948 52.386
17761 94.006 -79.946 52.372
17762 94.015 -79.945 52.357
17763 94.024 -79.943 52.341
17764 94.032 -79.942 52.326
17765 94.041 -79.940 52.310
17766 94.049 -79.939 52.295
17767 94.058 -79.938 52.279
17768 94.066 -79.936 52.263
18305 91.841 -81.004 53.588
18306 91.856 -81.003 53.565
18307 91.870 -81.003 53.542
18308 91.884 -81.001 53.520
18309 91.897 -80.996 53.497
18310 91.911 -80.991 53.476
18311 91.925 -80.986 53.454
18312 91.939 -80.982 53.432
18313 91.952 -80.977 53.411
18314 91.966 -80.972 53.390
18315 91.980 -80.967 53.370
18316 91.993 -80.962 53.350
18317 92.007 -80.958 53.330
18318 92.020 -80.953 53.310
18319 92.034 -80.948 53.290
18320 92.048 -80.943 53.271
18321 92.062 -80.939 53.252
18322 92.075 -80.929 53.234
18323 92.089 -80.917 53.217
18324 92.103 -80.905 53.199
18325 92.118 -80.892 53.182
18326 92.132 -80.880 53.166
18327 92.146 -80.868 53.149
18353 92.616 -80.638 53.274
18354 92.641 -80.629 53.256
18355 92.666 -80.622 53.237
18356 92.693 -80.613 53.219
18357 92.719 -80.606 53.201
18358 92.745 -80.598 53.183
18359 92.772 -80.591 53.167
18360 92.798 -80.583 53.152
18361 92.824 -80.576 53.139
18362 92.850 -80.568 53.129
18363 92.868 -80.554 53.119
18364 92.882 -80.539 53.108
18365 92.895 -80.528 53.098
18366 92.908 -80.516 53.087
18367 92.920 -80.505 53.077
18368 92.932 -80.494 53.067
18369 92.945 -80.483 53.055
18370 92.957 -80.472 53.043
18371 92.970 -80.460 53.031
18372 92.982 -80.450 53.018
18373 92.995 -80.438 53.005
18374 93.007 -80.427 52.991
18375 93.019 -80.416 52.978
18376 93.032 -80.405 52.967
18377 93.044 -80.394 52.956
18378 93.057 -80.385 52.946
18379 93.071 -80.381 52.937
18380 93.085 -80.378 52.928
18381 93.098 -80.374 52.919
18382 93.113 -80.370 52.910
18383 93.126 -80.367 52.902
18384 93.140 -80.363 52.894
18385 93.156 -80.356 52.888
18386 93.172 -80.350 52.884
18387 93.189 -80.342 52.880
18388 93.205 -80.335 52.878
18389 93.221 -80.328 52.879
18390 93.237 -80.321 52.880
18391 93.254 -80.314 52.882
18392 93.270 -80.306 52.885
18393 93.286 -80.300 52.888
18394 93.302 -80.293 52.891
18395 93.319 -80.285 52.892
18396 93.335 -80.278 52.892
18397 93.351 -80.271 52.893
18398 93.367 -80.264 52.893
18399 93.380 -80.259 52.894
18400 93.391 -80.255 52.896
18401 93.401 -80.252 52.897
18402 93.410 -80.248 52.898
18403 93.420 -80.233 52.899
18404 93.430 -80.211 52.899
18405 93.439 -80.194 52.900
18406 93.445 -80.196 52.899
18407 93.470 -80.196 52.898
18408 93.493 -80.196 52.897
18409 93.513 -80.194 52.895
18410 93.523 -80.185 52.894
18411 93.534 -80.177 52.892
18412 93.544 -80.168 52.890
18413 93.554 -80.159 52.887
18414 93.564 -80.151 52.884
18415 93.574 -80.142 52.881
18416 93.584 -80.134 52.876
18417 93.594 -80.124 52.871
18418 93.604 -80.116 52.867
18419 93.614 -80.107 52.861
18420 93.624 -80.099 52.856
18421 93.634 -80.090 52.851
18458 93.971 -79.938 52.440
18459 93.980 -79.936 52.425
18460 93.988 -79.935 52.410
18461 93.997 -79.933 52.396
18462 94.005 -79.932 52.382
18463 94.014 -79.930 52.367
18464 94.023 -79.929 52.352
18465 94.031 -79.927 52.336
18466 94.040 -79.926 52.321
18467 94.048 -79.925 52.305
18468 94.057 -79.924 52.290
18469 94.065 -79.922 52.275
18470 94.074 -79.921 52.259
18497 94.231 -79.587 51.699
18498 94.244 -79.567 51.671
18499 94.257 -79.548 51.644
18500 94.269 -79.528 51.616
18501 94.282 -79.508 51.586
18502 94.295 -79.488 51.558
18503 94.308 -79.469 51.530
18504 94.320 -79.449 51.501
18505 94.337 -79.471 51.473
18506 94.353 -79.503 51.443
18507 94.370 -79.534 51.413
18508 94.388 -79.565 51.383
18509 94.405 -79.596 51.353
19007 91.856 -80.994 53.588
19008 91.869 -80.992 53.566
19009 91.883 -80.986 53.544
19010 91.897 -80.982 53.522
19011 91.911 -80.977 53.501
19012 91.924 -80.972 53.479
19013 91.938 -80.967 53.456
19014 91.951 -80.963 53.436
19015 91.965 -80.958 53.416
19016 91.979 -80.953 53.395
19017 91.993 -80.949 53.374
19018 92.006 -80.944 53.353
19019 92.020 -80.939 53.333
19020 92.033 -80.934 53.313
19021 92.047 -80.930 53.294
19022 92.061 -80.925 53.275
19023 92.074 -80.919 53.257
19024 92.088 -80.907 53.238
19025 92.102 -80.894 53.220
19026 92.117 -80.882 53.203
19027 92.131 -80.870 53.186
19054 92.598 -80.621 53.285
19055 92.619 -80.613 53.266
19056 92.640 -80.604 53.248
19057 92.661 -80.596 53.229
19058 92.683 -80.588 53.210
19059 92.708 -80.580 53.192
19060 92.734 -80.572 53.176
19061 92.760 -80.564 53.161
19062 92.786 -80.557 53.147
19063 92.812 -80.549 53.136
19064 92.839 -80.541 53.126
19065 92.859 -80.529 53.117
19066 92.873 -80.514 53.106
19067 92.888 -80.500 53.096
19068 92.902 -80.485 53.086
19069 92.916 -80.471 53.075
19070 92.928 -80.460 53.062
19071 92.940 -80.449 53.049
19072 92.953 -80.438 53.035
19073 92.966 -80.426 53.022
19074 92.978 -80.416 53.009
19075 92.990 -80.404 52.994
19076 93.003 -80.393 52.979
19077 93.015 -80.382 52.967
19078 93.029 -80.377 52.955
19079 93.043 -80.373 52.945
19080 93.057 -80.370 52.935
19081 93.070 -80.366 52.925
19082 93.084 -80.363 52.915
19083 93.098 -80.359 52.906
19084 93.112 -80.356 52.898
19085 93.128 -80.349 52.890
19086 93.144 -80.341 52.884
19087 93.160 -80.334 52.879
19088 93.177 -80.328 52.876
19089 93.193 -80.320 52.873
19090 93.209 -80.313 52.873
19091 93.225 -80.306 52.874
19092 93.241 -80.299 52.875
19093 93.258 -80.292 52.877
19094 93.274 -80.284 52.878
19095 93.290 -80.278 52.878
19096 93.306 -80.270 52.879
19097 93.323 -80.263 52.879
19098 93.339 -80.256 52.878
19099 93.355 -80.249 52.877
19100 93.369 -80.243 52.878
19101 93.379 -80.240 52.878
19102 93.389 -80.237 52.878
19103 93.399 -80.222 52.879
19104 93.409 -80.201 52.881
19105 93.419 -80.179 52.883
19106 93.427 -80.163 52.883
19107 93.433 -80.165 52.883
19108 93.450 -80.165 52.884
19109 93.475 -80.165 52.885
19110 93.498 -80.166 52.885
19111 93.522 -80.166 52.884
19112 93.534 -80.159 52.883
19113 93.544 -80.149 52.883
19114 93.554 -80.141 52.881
19115 93.564 -80.132 52.879
19116 93.574 -80.124 52.876
19117 93.584 -80.115 52.872
19118 93.594 -80.106 52.868
19119 93.604 -80.098 52.864
19120 93.614 -80.089 52.859
19121 93.624 -80.080 52.853
19122 93.634 -80.071 52.848
19123 93.644 -80.063 52.841
19158 93.962 -79.925 52.462
19159 93.970 -79.924 52.448
19160 93.979 -79.922 52.434
19161 93.987 -79.921 52.419
19162 93.996 -79.919 52.405
19163 94.004 -79.918 52.391
19164 94.013 -79.916 52.377
19165 94.022 -79.914 52.361
19166 94.030 -79.913 52.346
19167 94.039 -79.912 52.330
19168 94.047 -79.911 52.315
19169 94.056 -79.909 52.299
19170 94.064 -79.908 52.284
19171 94.073 -79.906 52.268
19196 94.251 -79.560 51.765
19197 94.241 -79.567 51.738
19198 94.229 -79.573 51.712
19199 94.241 -79.554 51.685
19200 94.254 -79.535 51.657
19201 94.266 -79.515 51.629
19202 94.279 -79.495 51.600
19203 94.292 -79.476 51.572
19204 94.305 -79.456 51.543
19205 94.317 -79.436 51.514
19206 94.332 -79.439 51.486
19207 94.349 -79.471 51.456
19208 94.366 -79.502 51.426
19209 94.383 -79.533 51.395
19210 94.400 -79.565 51.365
19211 94.417 -79.596 51.334
19212 94.434 -79.627 51.304
19709 91.869 -80.977 53.589
19710 91.882 -80.973 53.567
19711 91.896 -80.968 53.546
19712 91.910 -80.963 53.524
19713 91.923 -80.958 53.503
19714 91.937 -80.954 53.482
19715 91.951 -80.949 53.461
19716 91.964 -80.944 53.440
19717 91.978 -80.939 53.419
19718 91.992 -80.935 53.398
19719 92.005 -80.930 53.377
19720 92.019 -80.925 53.358
19721 92.032 -80.920 53.337
19722 92.047 -80.915 53.317
19723 92.060 -80.911 53.298
19724 92.074 -80.906 53.279
19725 92.087 -80.896 53.261
19726 92.101 -80.884 53.242
19727 92.116 -80.872 53.224
19755 92.581 -80.605 53.296
19756 92.602 -80.597 53.278
19757 92.623 -80.588 53.259
19758 92.644 -80.580 53.240
19759 92.665 -80.572 53.221
19760 92.686 -80.563 53.203
19761 92.707 -80.555 53.187
19762 92.727 -80.547 53.171
19763 92.749 -80.538 53.157
19764 92.775 -80.531 53.145
19765 92.801 -80.522 53.136
19766 92.828 -80.515 53.127
19767 92.850 -80.505 53.117
19768 92.864 -80.490 53.107
19769 92.879 -80.475 53.096
19770 92.894 -80.460 53.084
19771 92.908 -80.444 53.070
19772 92.922 -80.429 53.056
19773 92.936 -80.415 53.042
19774 92.949 -80.403 53.028
19775 92.961 -80.393 53.013
19776 92.974 -80.381 52.998
19777 92.987 -80.373 52.982
19778 93.000 -80.369 52.968
19779 93.014 -80.366 52.956
19780 93.028 -80.362 52.945
19781 93.042 -80.358 52.935
19782 93.056 -80.355 52.924
19783 93.070 -80.351 52.913
19784 93.083 -80.347 52.904
19785 93.099 -80.341 52.895
19786 93.116 -80.334 52.888
19787 93.132 -80.327 52.882
19788 93.148 -80.319 52.877
19789 93.164 -80.312 52.874
19790 93.181 -80.306 52.872
19791 93.197 -80.298 52.870
19792 93.213 -80.291 52.869
19793 93.229 -80.284 52.869
19794 93.246 -80.277 52.869
19795 93.262 -80.270 52.869
19796 93.278 -80.262 52.868
19797 93.294 -80.256 52.869
19798 93.311 -80.248 52.868
19799 93.327 -80.241 52.867
19800 93.343 -80.234 52.866
19801 93.357 -80.228 52.865
19802 93.367 -80.225 52.864
19803 93.377 -80.212 52.863
19804 93.387 -80.190 52.864
19805 93.397 -80.169 52.866
19806 93.407 -80.148 52.869
19807 93.416 -80.132 52.870
19808 93.421 -80.134 52.871
19809 93.432 -80.134 52.873
19810 93.455 -80.135 52.874
19811 93.479 -80.135 52.875
19812 93.503 -80.136 52.876
19813 93.527 -80.136 52.876
19814 93.544 -80.131 52.876
19815 93.554 -80.123 52.875
19816 93.564 -80.114 52.875
19817 93.574 -80.106 52.871
19818 93.584 -80.096 52.868
19819 93.594 -80.088 52.865
19820 93.604 -80.079 52.861
19821 93.614 -80.071 52.856
19822 93.624 -80.062 52.851
19823 93.634 -80.053 52.846
19824 93.644 -80.045 52.839
19825 93.654 -80.036 52.832
19858 93.953 -79.912 52.484
19859 93.961 -79.911 52.470
19860 93.970 -79.909 52.457
19861 93.978 -79.908 52.442
19862 93.987 -79.906 52.428
19863 93.995 -79.905 52.414
19864 94.004 -79.903 52.400
19865 94.012 -79.902 52.386
19866 94.021 -79.900 52.370
19867 94.030 -79.899 52.355
19868 94.038 -79.898 52.340
19869 94.047 -79.896 52.325
19870 94.055 -79.895 52.309
19871 94.064 -79.893 52.293
19872 94.072 -79.892 52.277
19873 94.081 -79.890 52.260
19896 94.265 -79.535 51.804
19897 94.255 -79.541 51.778
19898 94.244 -79.546 51.752
19899 94.232 -79.552 51.725
19900 94.238 -79.541 51.699
19901 94.251 -79.521 51.672
19902 94.263 -79.501 51.643
19903 94.276 -79.482 51.614
19904 94.289 -79.462 51.586
19905 94.301 -79.442 51.558
19906 94.315 -79.422 51.528
19907 94.328 -79.408 51.500
19908 94.345 -79.439 51.470
19909 94.361 -79.470 51.439
19910 94.379 -79.502 51.408
19911 94.396 -79.533 51.377
19912 94.413 -79.564 51.347
19913 94.429 -79.595 51.316
19914 94.447 -79.627 51.285
20412 91.895 -80.954 53.569
20413 91.909 -80.949 53.548
20414 91.923 -80.945 53.527
20415 91.936 -80.939 53.506
20416 91.950 -80.935 53.485
20417 91.963 -80.930 53.464
20418 91.977 -80.926 53.443
20419 91.991 -80.920 53.422
20420 92.005 -80.916 53.402
20421 92.018 -80.911 53.382
20422 92.032 -80.906 53.361
20423 92.046 -80.902 53.341
20424 92.059 -80.897 53.322
20425 92.073 -80.892 53.302
20426 92.087 -80.886 53.284
20427 92.101 -80.873 53.265
20456 92.565 -80.588 53.309
20457 92.586 -80.580 53.290
20458 92.606 -80.572 53.272
20459 92.628 -80.563 53.253
20460 92.648 -80.555 53.234
20461 92.669 -80.547 53.216
20462 92.690 -80.538 53.199
20463 92.711 -80.530 53.183
20464 92.732 -80.522 53.167
20465 92.753 -80.513 53.156
20466 92.773 -80.505 53.146
20467 92.794 -80.497 53.138
20468 92.816 -80.489 53.129
20469 92.841 -80.480 53.119
20470 92.855 -80.465 53.108
20471 92.870 -80.450 53.094
20472 92.884 -80.435 53.079
20473 92.899 -80.419 53.064
20474 92.913 -80.404 53.049
20475 92.930 -80.388 53.033
20476 92.946 -80.371 53.018
20477 92.962 -80.365 53.002
20478 92.977 -80.364 52.985
20479 92.992 -80.362 52.971
20480 93.007 -80.359 52.958
20481 93.022 -80.356 52.947
20482 93.038 -80.353 52.936
20483 93.053 -80.350 52.926
20484 93.066 -80.344 52.915
20485 93.078 -80.337 52.907
20486 93.091 -80.328 52.898
20487 93.105 -80.319 52.890
20488 93.120 -80.312 52.884
20489 93.136 -80.305 52.879
20490 93.152 -80.298 52.875
20491 93.169 -80.290 52.873
20492 93.185 -80.284 52.869
20493 93.201 -80.276 52.866
20494 93.217 -80.269 52.864
20495 93.234 -80.262 52.862
20496 93.250 -80.255 52.862
20497 93.266 -80.247 52.861
20498 93.282 -80.240 52.860
20499 93.298 -80.233 52.859
20500 93.315 -80.226 52.857
20501 93.331 -80.218 52.856
20502 93.346 -80.212 52.854
20503 93.356 -80.202 52.853
20504 93.365 -80.181 52.852
20505 93.375 -80.159 52.853
20506 93.386 -80.137 52.855
20507 93.395 -80.116 52.857
20508 93.404 -80.101 52.859
20509 93.410 -80.102 52.860
20510 93.415 -80.104 52.861
20511 93.436 -80.104 52.863
20512 93.461 -80.105 52.865
20513 93.484 -80.105 52.867
20514 93.508 -80.106 52.869
20515 93.532 -80.106 52.869
20516 93.554 -80.104 52.870
20517 93.564 -80.096 52.870
20518 93.573 -80.087 52.868
20519 93.583 -80.078 52.866
20520 93.593 -80.070 52.862
20521 93.604 -80.061 52.858
20522 93.614 -80.052 52.853
20523 93.624 -80.043 52.848
20524 93.634 -80.035 52.843
20525 93.644 -80.026 52.837
20526 93.654 -80.018 52.830
20527 93.664 -80.009 52.823
20557 93.934 -79.901 52.517
20558 93.943 -79.899 52.505
20559 93.952 -79.898 52.492
20560 93.960 -79.896 52.479
20561 93.969 -79.895 52.465
20562 93.977 -79.893 52.451
20563 93.986 -79.892 52.438
20564 93.994 -79.890 52.425
20565 94.003 -79.889 52.410
20566 94.011 -79.887 52.395
20567 94.020 -79.886 52.381
20568 94.029 -79.885 52.365
20569 94.037 -79.883 52.349
20570 94.046 -79.882 52.334
20571 94.054 -79.880 52.318
20572 94.063 -79.879 52.302
20573 94.071 -79.877 52.286
20574 94.080 -79.876 52.270
20575 94.088 -79.874 52.253
20596 94.268 -79.528 51.841
20597 94.269 -79.515 51.816
20598 94.258 -79.520 51.790
20599 94.247 -79.526 51.765
20600 94.236 -79.532 51.739
20601 94.235 -79.527 51.713
20602 94.248 -79.507 51.686
20603 94.260 -79.488 51.657
20604 94.273 -79.468 51.629
20605 94.286 -79.448 51.601
20606 94.298 -79.429 51.572
20607 94.312 -79.409 51.543
20608 94.324 -79.389 51.515
20609 94.340 -79.407 51.485
20610 94.357 -79.439 51.453
20611 94.374 -79.470 51.423
20612 94.391 -79.501 51.392
20613 94.408 -79.532 51.360
20614 94.425 -79.564 51.329
20615 94.442 -79.595 51.297
20616 94.459 -79.626 51.266
20617 94.476 -79.658 51.234
21115 91.922 -80.930 53.551
21116 91.936 -80.926 53.530
21117 91.949 -80.921 53.509
21118 91.963 -80.917 53.488
21119 91.977 -80.911 53.468
21120 91.990 -80.907 53.447
21121 92.004 -80.902 53.427
21122 92.017 -80.897 53.406
21123 92.031 -80.892 53.385
21124 92.045 -80.888 53.366
21125 92.059 -80.883 53.346
21126 92.072 -80.878 53.326
21157 92.548 -80.573 53.324
21158 92.569 -80.564 53.305
21159 92.590 -80.556 53.286
21160 92.611 -80.547 53.267
21161 92.632 -80.539 53.249
21162 92.652 -80.531 53.231
21163 92.673 -80.522 53.214
21164 92.694 -80.514 53.197
21165 92.715 -80.506 53.181
21166 92.736 -80.497 53.169
21167 92.757 -80.489 53.159
21168 92.778 -80.481 53.150
21169 92.798 -80.472 53.142
21170 92.819 -80.464 53.132
21171 92.840 -80.456 53.120
21172 92.863 -80.432 53.106
21173 92.883 -80.408 53.090
21174 92.901 -80.391 53.073
21175 92.917 -80.374 53.056
21176 92.934 -80.357 53.039
21177 92.951 -80.341 53.022
21178 92.966 -80.336 53.005
21179 92.981 -80.339 52.989
21180 92.995 -80.342 52.975
21181 93.009 -80.346 52.962
21182 93.023 -80.349 52.950
21183 93.036 -80.347 52.939
21184 93.048 -80.341 52.929
21185 93.060 -80.334 52.919
21186 93.072 -80.326 52.910
21187 93.085 -80.318 52.901
21188 93.099 -80.309 52.893
21189 93.113 -80.300 52.886
21190 93.126 -80.291 52.881
21191 93.140 -80.283 52.876
21192 93.157 -80.275 52.872
21193 93.173 -80.269 52.868
21194 93.189 -80.261 52.863
21195 93.205 -80.254 52.860
21196 93.221 -80.247 52.859
21197 93.238 -80.240 52.857
21198 93.255 -80.233 52.856
21199 93.271 -80.225 52.854
21200 93.288 -80.218 52.852
21201 93.304 -80.211 52.849
21202 93.321 -80.204 52.848
21203 93.334 -80.192 52.847
21204 93.344 -80.170 52.846
21205 93.354 -80.149 52.845
21206 93.364 -80.127 52.845
21207 93.373 -80.106 52.845
21208 93.383 -80.084 52.846
21209 93.392 -80.069 52.848
21210 93.398 -80.071 52.850
21211 93.404 -80.073 52.851
21212 93.418 -80.074 52.854
21213 93.441 -80.074 52.857
21214 93.466 -80.074 52.860
21215 93.489 -80.074 52.863
21216 93.513 -80.075 52.864
21217 93.537 -80.075 52.865
21218 93.561 -80.076 52.866
21219 93.573 -80.068 52.865
21220 93.583 -80.060 52.863
21221 93.593 -80.051 52.860
21222 93.604 -80.043 52.856
21223 93.614 -80.034 52.851
21224 93.624 -80.025 52.846
21225 93.634 -80.017 52.841
21226 93.644 -80.008 52.835
21227 93.654 -79.999 52.828
21228 93.664 -79.990 52.822
21229 93.674 -79.982 52.814
21230 93.684 -79.973 52.805
21231 93.694 -79.965 52.796
21232 93.704 -79.956 52.788
21233 93.714 -79.947 52.779
21234 93.724 -79.939 52.769
21235 93.734 -79.930 52.760
21236 93.745 -79.921 52.750
21258 93.933 -79.886 52.524
21259 93.942 -79.885 52.512
21260 93.951 -79.883 52.500
21261 93.959 -79.882 52.487
21262 93.968 -79.880 52.474
21263 93.976 -79.879 52.460
21264 93.985 -79.877 52.447
21265 93.993 -79.876 52.434
21266 94.002 -79.874 52.420
21267 94.010 -79.874 52.405
21268 94.019 -79.872 52.390
21269 94.028 -79.871 52.375
21270 94.036 -79.869 52.359
21271 94.045 -79.867 52.344
21272 94.053 -79.866 52.328
21273 94.062 -79.864 52.312
21274 94.070 -79.863 52.295
21275 94.079 -79.861 52.279
21276 94.087 -79.860 52.262
21296 94.259 -79.541 51.878
21297 94.267 -79.514 51.853
21298 94.272 -79.494 51.828
21299 94.261 -79.501 51.803
21300 94.250 -79.507 51.778
21301 94.239 -79.512 51.753
21302 94.232 -79.514 51.727
21303 94.245 -79.494 51.699
21304 94.258 -79.475 51.671
21305 94.270 -79.455 51.643
21306 94.283 -79.435 51.615
21307 94.296 -79.416 51.587
21308 94.309 -79.396 51.558
21309 94.321 -79.376 51.529
21310 94.336 -79.376 51.499
21311 94.353 -79.407 51.468
21312 94.369 -79.438 51.437
21313 94.387 -79.469 51.405
21314 94.404 -79.501 51.373
21315 94.421 -79.532 51.342
21316 94.437 -79.563 51.310
21317 94.455 -79.595 51.277
21318 94.472 -79.626 51.245
21319 94.485 -79.664 51.212
21320 94.493 -79.707 51.179
21819 91.962 -80.902 53.513
21820 91.976 -80.898 53.492
21821 91.990 -80.893 53.472
21822 92.003 -80.888 53.451
21823 92.017 -80.883 53.430
21858 92.533 -80.556 53.339
21859 92.553 -80.548 53.321
21860 92.573 -80.540 53.301
21861 92.594 -80.532 53.283
21862 92.615 -80.523 53.265
21863 92.636 -80.515 53.247
21864 92.657 -80.507 53.229
21865 92.677 -80.498 53.212
21866 92.698 -80.490 53.196
21867 92.719 -80.482 53.184
21868 92.740 -80.473 53.174
21869 92.761 -80.465 53.165
21870 92.782 -80.457 53.156
21871 92.802 -80.448 53.146
21872 92.824 -80.440 53.133
21873 92.844 -80.431 53.117
21874 92.866 -80.415 53.101
21875 92.889 -80.388 53.084
21876 92.912 -80.360 53.066
21877 92.935 -80.333 53.048
21878 92.955 -80.309 53.030
21879 92.971 -80.306 53.013
21880 92.983 -80.311 52.996
21881 92.994 -80.316 52.981
21882 93.006 -80.321 52.968
21883 93.018 -80.326 52.955
21884 93.030 -80.331 52.944
21885 93.042 -80.330 52.933
21886 93.054 -80.323 52.924
21887 93.067 -80.316 52.914
21888 93.080 -80.306 52.904
21889 93.093 -80.298 52.895
21890 93.107 -80.289 52.888
21891 93.119 -80.279 52.882
21892 93.136 -80.270 52.877
21893 93.161 -80.262 52.872
21894 93.178 -80.254 52.867
21895 93.194 -80.247 52.862
21896 93.211 -80.240 52.858
21897 93.227 -80.233 52.856
21898 93.244 -80.225 52.855
21899 93.261 -80.218 52.853
21900 93.277 -80.211 52.851
21901 93.294 -80.204 52.848
21902 93.310 -80.196 52.846
21903 93.320 -80.186 52.844
21904 93.325 -80.165 52.843
21905 93.333 -80.140 52.842
21906 93.342 -80.117 52.840
21907 93.352 -80.096 52.839
21908 93.362 -80.074 52.840
21909 93.372 -80.052 52.840
21910 93.380 -80.038 52.841
21911 93.386 -80.040 52.843
21912 93.392 -80.042 52.845
21913 93.399 -80.043 52.847
21914 93.423 -80.043 52.851
21915 93.446 -80.043 52.855
21916 93.471 -80.044 52.859
21917 93.494 -80.044 52.860
21918 93.518 -80.045 52.861
21919 93.542 -80.045 52.863
21920 93.566 -80.046 52.862
21921 93.583 -80.042 52.861
21922 93.593 -80.033 52.859
21923 93.604 -80.024 52.854
21924 93.614 -80.015 52.850
21925 93.624 -80.007 52.845
21926 93.634 -79.998 52.840
21927 93.644 -79.990 52.834
21928 93.654 -79.981 52.827
21929 93.664 -79.972 52.820
21930 93.674 -79.964 52.813
21931 93.684 -79.955 52.804
21932 93.694 -79.946 52.795
21933 93.704 -79.937 52.787
21934 93.714 -79.929 52.778
21935 93.724 -79.920 52.769
21936 93.734 -79.911 52.760
21937 93.745 -79.904 52.751
21938 93.753 -79.902 52.741
21959 93.933 -79.872 52.532
21960 93.941 -79.871 52.520
21961 93.950 -79.869 52.508
21962 93.959 -79.867 52.495
21963 93.967 -79.866 52.483
21964 93.976 -79.864 52.469
21965 93.984 -79.863 52.455
21966 93.993 -79.861 52.442
21967 94.001 -79.861 52.429
21968 94.010 -79.859 52.415
21969 94.018 -79.858 52.400
21970 94.027 -79.856 52.384
21971 94.035 -79.855 52.369
21972 94.044 -79.853 52.353
21973 94.052 -79.852 52.337
21974 94.061 -79.850 52.321
21975 94.069 -79.849 52.304
21976 94.078 -79.847 52.288
21977 94.086 -79.845 52.270
21978 94.095 -79.844 52.253
21995 94.241 -79.579 51.934
21996 94.249 -79.553 51.912
21997 94.258 -79.527 51.888
21998 94.266 -79.501 51.864
21999 94.275 -79.476 51.840
22000 94.264 -79.480 51.815
22001 94.253 -79.486 51.790
22002 94.243 -79.492 51.765
22003 94.231 -79.498 51.739
22004 94.242 -79.481 51.712
22005 94.255 -79.461 51.685
22006 94.267 -79.441 51.656
22007 94.280 -79.422 51.628
22008 94.293 -79.402 51.600
22009 94.306 -79.382 51.572
22010 94.318 -79.363 51.543
22011 94.331 -79.344 51.513
22012 94.348 -79.375 51.482
22013 94.365 -79.407 51.451
22014 94.382 -79.438 51.419
22015 94.399 -79.469 51.386
22016 94.416 -79.501 51.354
22017 94.433 -79.532 51.322
22018 94.450 -79.563 51.290
22019 94.465 -79.598 51.255
22020 94.474 -79.641 51.222
22021 94.483 -79.684 51.189
22022 94.491 -79.727 51.156
22559 92.529 -80.539 53.357
22560 92.546 -80.531 53.338
22561 92.562 -80.523 53.319
22562 92.578 -80.515 53.301
22563 92.598 -80.507 53.282
22564 92.619 -80.498 53.263
22565 92.640 -80.490 53.244
22566 92.661 -80.482 53.227
22567 92.682 -80.473 53.212
22568 92.703 -80.465 53.200
22569 92.723 -80.457 53.190
22570 92.744 -80.448 53.180
22571 92.765 -80.440 53.170
22572 92.786 -80.431 53.159
22573 92.807 -80.423 53.146
22574 92.828 -80.415 53.130
22575 92.848 -80.406 53.113
22576 92.869 -80.398 53.096
22577 92.892 -80.371 53.078
22578 92.915 -80.344 53.060
22579 92.938 -80.316 53.042
22580 92.961 -80.289 53.025
22581 92.976 -80.288 53.007
22582 92.989 -80.294 52.991
22583 93.001 -80.299 52.976
22584 93.013 -80.303 52.963
22585 93.024 -80.309 52.951
22586 93.037 -80.314 52.940
22587 93.049 -80.312 52.929
22588 93.061 -80.305 52.918
22589 93.074 -80.296 52.907
22590 93.087 -80.287 52.896
22591 93.098 -80.277 52.889
22592 93.109 -80.266 52.882
22593 93.131 -80.257 52.876
22594 93.156 -80.249 52.871
22595 93.181 -80.240 52.866
22596 93.200 -80.233 52.862
22597 93.216 -80.225 52.858
22598 93.233 -80.218 52.856
22599 93.250 -80.212 52.854
22600 93.266 -80.204 52.851
22601 93.283 -80.197 52.849
22602 93.299 -80.190 52.847
22603 93.309 -80.178 52.844
22604 93.307 -80.162 52.842
22605 93.315 -80.137 52.841
22606 93.324 -80.113 52.839
22607 93.332 -80.089 52.837
22608 93.341 -80.065 52.836
22609 93.350 -80.043 52.836
22610 93.360 -80.021 52.836
22611 93.368 -80.007 52.836
22612 93.374 -80.008 52.838
22613 93.380 -80.011 52.840
22614 93.386 -80.012 52.844
22615 93.404 -80.013 52.847
22616 93.428 -80.013 52.852
22617 93.451 -80.013 52.856
22618 93.476 -80.014 52.859
22619 93.499 -80.014 52.860
22620 93.523 -80.015 52.861
22621 93.547 -80.015 52.861
22622 93.571 -80.015 52.860
22623 93.593 -80.015 52.857
22624 93.604 -80.005 52.853
22625 93.614 -79.997 52.848
22626 93.624 -79.989 52.844
22627 93.634 -79.980 52.839
22628 93.644 -79.971 52.833
22629 93.654 -79.962 52.826
22630 93.664 -79.954 52.819
22631 93.674 -79.945 52.812
22632 93.684 -79.936 52.804
22633 93.694 -79.928 52.795
22634 93.704 -79.919 52.786
22635 93.714 -79.911 52.778
22636 93.724 -79.902 52.770
22637 93.734 -79.893 52.760
22638 93.744 -79.889 52.751
22639 93.752 -79.889 52.742
22640 93.761 -79.887 52.733
22661 93.940 -79.856 52.527
22662 93.949 -79.855 52.514
22663 93.958 -79.853 52.502
22664 93.966 -79.852 52.490
22665 93.975 -79.850 52.477
22666 93.983 -79.849 52.464
22667 93.992 -79.848 52.451
22668 94.000 -79.846 52.437
22669 94.009 -79.845 52.423
22670 94.017 -79.843 52.409
22671 94.026 -79.842 52.393
22672 94.035 -79.840 52.377
22673 94.043 -79.839 52.362
22674 94.052 -79.837 52.345
22675 94.060 -79.836 52.329
22676 94.069 -79.834 52.313
22677 94.077 -79.833 52.296
22678 94.086 -79.831 52.278
22679 94.095 -79.830 52.260
22695 94.231 -79.592 51.964
22696 94.240 -79.565 51.943
22697 94.248 -79.539 51.921
22698 94.257 -79.513 51.898
22699 94.265 -79.488 51.874
22700 94.274 -79.462 51.851
22701 94.267 -79.460 51.826
22702 94.257 -79.466 51.801
22703 94.246 -79.472 51.776
22704 94.235 -79.478 51.751
22705 94.239 -79.467 51.725
22706 94.252 -79.447 51.697
22707 94.264 -79.428 51.669
22708 94.277 -79.408 51.641
22709 94.290 -79.388 51.613
22710 94.302 -79.369 51.585
22711 94.316 -79.349 51.556
22712 94.328 -79.329 51.527
22713 94.344 -79.344 51.495
22714 94.361 -79.375 51.463
22715 94.378 -79.407 51.430
22716 94.395 -79.438 51.398
22717 94.412 -79.469 51.365
22718 94.429 -79.501 51.333
22719 94.445 -79.532 51.301
22720 94.454 -79.576 51.266
22721 94.463 -79.619 51.232
22722 94.472 -79.662 51.198
22723 94.481 -79.705 51.166
22724 94.489 -79.748 51.132
22725 94.498 -79.792 51.099
23260 92.525 -80.522 53.375
23261 92.542 -80.514 53.357
23262 92.558 -80.506 53.338
23263 92.573 -80.498 53.319
23264 92.589 -80.490 53.300
23265 92.605 -80.482 53.281
23266 92.623 -80.474 53.263
23267 92.644 -80.466 53.245
23268 92.665 -80.457 53.229
23269 92.686 -80.449 53.217
23270 92.707 -80.441 53.206
23271 92.727 -80.432 53.195
23272 92.748 -80.424 53.184
23273 92.769 -80.416 53.172
23274 92.790 -80.407 53.159
23275 92.811 -80.399 53.144
23276 92.832 -80.391 53.127
23277 92.853 -80.382 53.110
23278 92.873 -80.374 53.092
23279 92.896 -80.353 53.074
23280 92.919 -80.326 53.056
23281 92.941 -80.299 53.038
23282 92.965 -80.272 53.021
23283 92.979 -80.270 53.004
23284 92.992 -80.275 52.987
23285 93.005 -80.281 52.972
23286 93.017 -80.286 52.958
23287 93.030 -80.291 52.946
23288 93.043 -80.296 52.934
23289 93.055 -80.294 52.922
23290 93.066 -80.284 52.910
23291 93.077 -80.274 52.900
23292 93.088 -80.263 52.892
23293 93.101 -80.253 52.884
23294 93.126 -80.245 52.876
23295 93.151 -80.237 52.871
23296 93.176 -80.228 52.866
23297 93.201 -80.219 52.862
23298 93.222 -80.212 52.859
23299 93.239 -80.204 52.857
23300 93.256 -80.197 52.855
23301 93.272 -80.190 52.851
23302 93.289 -80.183 52.848
23303 93.297 -80.171 52.845
23304 93.296 -80.154 52.843
23305 93.297 -80.134 52.842
23306 93.305 -80.110 52.840
23307 93.314 -80.086 52.839
23308 93.323 -80.062 52.836
23309 93.331 -80.037 52.834
23310 93.340 -80.013 52.833
23311 93.348 -79.990 52.832
23312 93.357 -79.976 52.833
23313 93.362 -79.977 52.834
23314 93.370 -79.982 52.836
23315 93.381 -79.992 52.840
23316 93.395 -79.988 52.844
23317 93.409 -79.983 52.849
23318 93.433 -79.983 52.854
23319 93.457 -79.983 52.857
23320 93.481 -79.983 52.859
23321 93.504 -79.983 52.860
23322 93.529 -79.984 52.860
23323 93.552 -79.984 52.858
23324 93.576 -79.985 52.856
23325 93.600 -79.985 52.851
23326 93.614 -79.979 52.847
23327 93.624 -79.970 52.842
23328 93.634 -79.961 52.838
23329 93.644 -79.953 52.831
23330 93.654 -79.944 52.824
23331 93.664 -79.936 52.818
23332 93.674 -79.927 52.810
23333 93.684 -79.918 52.802
23334 93.694 -79.909 52.794
23335 93.704 -79.901 52.785
23336 93.714 -79.892 52.777
23337 93.724 -79.883 52.770
23338 93.734 -79.877 52.760
23339 93.743 -79.876 52.751
23340 93.751 -79.874 52.742
23341 93.760 -79.873 52.733
23362 93.939 -79.842 52.533
23363 93.948 -79.840 52.521
23364 93.957 -79.839 52.509
23365 93.965 -79.837 52.496
23366 93.974 -79.836 52.483
23367 93.982 -79.835 52.470
23368 93.991 -79.833 52.457
23369 93.999 -79.832 52.443
23370 94.008 -79.830 52.429
23371 94.016 -79.829 52.414
23372 94.025 -79.827 52.399
23373 94.034 -79.826 52.383
23374 94.042 -79.824 52.367
23375 94.051 -79.823 52.351
23376 94.059 -79.821 52.335
23377 94.068 -79.820 52.319
23378 94.076 -79.818 52.302
23379 94.085 -79.817 52.285
23380 94.094 -79.816 52.267
23381 94.102 -79.814 52.250
23395 94.222 -79.604 51.993
23396 94.230 -79.578 51.971
23397 94.239 -79.552 51.951
23398 94.247 -79.526 51.930
23399 94.256 -79.501 51.907
23400 94.264 -79.474 51.883
23401 94.273 -79.448 51.860
23402 94.271 -79.440 51.835
23403 94.260 -79.446 51.810
23404 94.249 -79.452 51.786
23405 94.238 -79.457 51.762
23406 94.236 -79.454 51.736
23407 94.249 -79.435 51.709
23408 94.261 -79.415 51.681
23409 94.274 -79.394 51.653
23410 94.287 -79.375 51.625
23411 94.299 -79.355 51.597
23412 94.313 -79.335 51.569
23413 94.325 -79.316 51.539
23414 94.339 -79.313 51.508
23415 94.356 -79.344 51.475
23416 94.373 -79.375 51.442
23417 94.390 -79.406 51.409
23418 94.407 -79.438 51.376
23419 94.424 -79.469 51.343
23420 94.434 -79.510 51.311
23421 94.443 -79.553 51.276
23422 94.452 -79.596 51.241
23423 94.461 -79.639 51.207
23424 94.470 -79.683 51.174
23425 94.479 -79.726 51.140
23426 94.487 -79.770 51.107
23961 92.521 -80.505 53.395
23962 92.537 -80.497 53.377
23963 92.553 -80.489 53.358
23964 92.569 -80.481 53.339
23965 92.585 -80.473 53.320
23966 92.601 -80.465 53.301
23967 92.618 -80.457 53.282
23968 92.634 -80.449 53.264
23969 92.649 -80.441 53.247
23970 92.669 -80.433 53.233
23971 92.690 -80.425 53.221
23972 92.711 -80.416 53.210
23973 92.732 -80.408 53.199
23974 92.753 -80.400 53.186
23975 92.773 -80.391 53.173
23976 92.794 -80.383 53.157
23977 92.815 -80.375 53.141
23978 92.836 -80.366 53.124
23979 92.857 -80.358 53.108
23980 92.877 -80.350 53.090
23981 92.899 -80.337 53.072
23982 92.922 -80.309 53.053
23983 92.945 -80.282 53.036
23984 92.968 -80.255 53.019
23985 92.983 -80.252 53.001
23986 92.996 -80.257 52.983
23987 93.008 -80.262 52.966
23988 93.021 -80.268 52.952
23989 93.034 -80.273 52.939
23990 93.047 -80.278 52.926
23991 93.057 -80.271 52.914
23992 93.067 -80.261 52.904
23993 93.078 -80.250 52.895
23994 93.096 -80.241 52.887
23995 93.121 -80.232 52.879
23996 93.146 -80.224 52.873
23997 93.171 -80.215 52.868
23998 93.196 -80.207 52.864
23999 93.221 -80.199 52.861
24000 93.245 -80.190 52.858
24001 93.262 -80.183 52.855
24002 93.278 -80.176 52.852
24003 93.286 -80.164 52.848
24004 93.284 -80.146 52.845
24005 93.283 -80.129 52.842
24006 93.288 -80.108 52.840
24007 93.296 -80.083 52.838
24008 93.304 -80.059 52.836
24009 93.313 -80.034 52.833
24010 93.322 -80.010 52.831
24011 93.330 -79.986 52.830
24012 93.339 -79.964 52.827
24013 93.354 -79.958 52.828
24014 93.365 -79.968 52.830
24015 93.375 -79.977 52.831
24016 93.387 -79.987 52.836
24017 93.404 -79.974 52.840
24018 93.418 -79.968 52.845
24019 93.431 -79.962 52.852
24020 93.444 -79.956 52.856
24021 93.462 -79.953 52.858
24022 93.485 -79.953 52.860
24023 93.509 -79.953 52.859
24024 93.533 -79.954 52.857
24025 93.557 -79.954 52.854
24026 93.581 -79.955 52.849
24027 93.605 -79.955 52.845
24028 93.624 -79.952 52.841
24029 93.634 -79.943 52.835
24030 93.644 -79.934 52.829
24031 93.654 -79.926 52.823
24032 93.664 -79.917 52.816
24033 93.674 -79.908 52.808
24034 93.684 -79.900 52.800
24035 93.694 -79.891 52.792
24036 93.704 -79.883 52.784
24037 93.714 -79.874 52.776
24038 93.724 -79.865 52.769
24039 93.733 -79.863 52.759
24040 93.742 -79.861 52.750
24041 93.751 -79.860 52.742
24042 93.759 -79.858 52.733
24043 93.768 -79.857 52.724
24063 93.938 -79.827 52.538
24064 93.947 -79.826 52.527
24065 93.956 -79.824 52.514
24066 93.964 -79.823 52.501
24067 93.973 -79.822 52.488
24068 93.981 -79.820 52.475
24069 93.990 -79.819 52.462
24070 93.998 -79.817 52.449
24071 94.007 -79.816 52.435
24072 94.015 -79.814 52.419
24073 94.024 -79.813 52.404
24074 94.033 -79.811 52.388
24075 94.041 -79.810 52.372
24076 94.050 -79.808 52.356
24077 94.058 -79.807 52.340
24078 94.067 -79.805 52.324
24079 94.075 -79.804 52.307
24080 94.084 -79.803 52.290
24081 94.092 -79.802 52.272
24082 94.101 -79.800 52.254
24083 94.110 -79.798 52.237
24094 94.204 -79.642 52.038
24095 94.212 -79.616 52.018
24096 94.221 -79.590 51.997
24097 94.229 -79.564 51.977
24098 94.238 -79.538 51.957
24099 94.247 -79.513 51.936
24100 94.255 -79.487 51.914
24101 94.264 -79.461 51.891
24102 94.272 -79.435 51.867
24103 94.274 -79.420 51.843
24104 94.263 -79.426 51.819
24105 94.252 -79.432 51.795
24106 94.241 -79.438 51.772
24107 94.233 -79.441 51.746
24108 94.246 -79.421 51.719
24109 94.259 -79.401 51.692
24110 94.271 -79.382 51.664
24111 94.284 -79.362 51.636
24112 94.297 -79.342 51.609
24113 94.310 -79.322 51.581
24114 94.322 -79.303 51.551
24115 94.335 -79.283 51.519
24116 94.352 -79.312 51.486
24117 94.369 -79.343 51.453
24118 94.386 -79.375 51.420
24119 94.403 -79.406 51.386
24120 94.415 -79.444 51.353
24121 94.424 -79.488 51.319
24122 94.432 -79.531 51.284
24123 94.441 -79.574 51.250
24124 94.450 -79.617 51.216
24125 94.459 -79.661 51.182
24126 94.468 -79.704 51.148
24127 94.477 -79.747 51.114
24662 92.517 -80.488 53.417
24663 92.533 -80.480 53.398
24664 92.549 -80.472 53.379
24665 92.565 -80.464 53.359
24666 92.581 -80.457 53.342
24667 92.597 -80.448 53.322
24668 92.614 -80.440 53.303
24669 92.629 -80.432 53.284
24670 92.645 -80.424 53.266
24671 92.661 -80.416 53.250
24672 92.677 -80.408 53.235
24673 92.694 -80.400 53.224
24674 92.715 -80.391 53.211
24675 92.736 -80.383 53.199
24676 92.757 -80.375 53.185
24677 92.778 -80.366 53.170
24678 92.798 -80.358 53.154
24679 92.819 -80.350 53.138
24680 92.840 -80.341 53.122
24681 92.861 -80.333 53.105
24682 92.882 -80.325 53.087
24683 92.903 -80.316 53.068
24684 92.925 -80.292 53.051
24685 92.948 -80.265 53.033
24686 92.971 -80.237 53.014
24687 92.987 -80.234 52.995
24688 92.999 -80.239 52.978
24689 93.012 -80.244 52.961
24690 93.025 -80.250 52.947
24691 93.038 -80.255 52.933
24692 93.048 -80.256 52.921
24693 93.058 -80.248 52.910
24694 93.068 -80.237 52.901
24695 93.091 -80.228 52.892
24696 93.116 -80.220 52.884
24697 93.141 -80.212 52.877
24698 93.166 -80.203 52.871
24699 93.191 -80.194 52.866
24700 93.216 -80.186 52.862
24701 93.241 -80.178 52.859
24702 93.266 -80.169 52.857
24703 93.275 -80.156 52.853
24704 93.273 -80.139 52.848
24705 93.271 -80.122 52.844
24706 93.270 -80.105 52.842
24707 93.278 -80.080 52.838
24708 93.287 -80.056 52.835
24709 93.295 -80.032 52.833
24710 93.304 -80.008 52.830
24711 93.313 -79.990 52.828
24712 93.322 -79.971 52.826
24713 93.334 -79.957 52.823
24714 93.358 -79.952 52.823
24715 93.370 -79.962 52.825
24716 93.381 -79.971 52.828
24717 93.392 -79.981 52.833
24718 93.414 -79.961 52.837
24719 93.427 -79.955 52.842
24720 93.440 -79.949 52.848
24721 93.453 -79.943 52.852
24722 93.467 -79.936 52.856
24723 93.480 -79.930 52.857
24724 93.493 -79.924 52.856
24725 93.514 -79.923 52.854
24726 93.538 -79.924 52.850
24727 93.562 -79.924 52.846
24728 93.586 -79.924 52.842
24729 93.610 -79.924 52.837
24730 93.634 -79.924 52.832
24731 93.644 -79.916 52.826
24732 93.654 -79.908 52.821
24733 93.664 -79.899 52.815
24734 93.674 -79.890 52.807
24735 93.684 -79.881 52.798
24736 93.694 -79.873 52.791
24737 93.704 -79.864 52.783
24738 93.714 -79.855 52.775
24739 93.724 -79.850 52.767
24740 93.732 -79.849 52.758
24741 93.741 -79.847 52.749
24742 93.750 -79.845 52.741
24743 93.758 -79.844 52.732
24744 93.767 -79.842 52.723
24764 93.938 -79.813 52.541
24765 93.946 -79.811 52.530
24766 93.955 -79.811 52.517
24767 93.964 -79.809 52.505
24768 93.972 -79.808 52.493
24769 93.981 -79.806 52.479
24770 93.989 -79.805 52.466
24771 93.998 -79.803 52.453
24772 94.006 -79.802 52.439
24773 94.015 -79.800 52.423
24774 94.024 -79.798 52.408
24775 94.032 -79.797 52.392
24776 94.041 -79.795 52.376
24777 94.049 -79.794 52.360
24778 94.058 -79.792 52.344
24779 94.066 -79.791 52.328
24780 94.075 -79.790 52.310
24781 94.083 -79.789 52.293
24782 94.092 -79.787 52.275
24783 94.101 -79.786 52.258
24784 94.109 -79.784 52.241
24785 94.117 -79.783 52.223
24794 94.194 -79.654 52.064
24795 94.203 -79.629 52.043
24796 94.211 -79.603 52.023
24797 94.220 -79.577 52.003
24798 94.228 -79.551 51.983
24799 94.237 -79.525 51.963
24800 94.246 -79.499 51.942
24801 94.254 -79.473 51.921
24802 94.263 -79.447 51.898
24803 94.271 -79.422 51.875
24804 94.277 -79.400 51.851
24805 94.266 -79.406 51.828
24806 94.255 -79.412 51.805
24807 94.245 -79.418 51.782
24808 94.234 -79.423 51.757
24809 94.243 -79.407 51.730
24810 94.256 -79.388 51.702
24811 94.268 -79.368 51.676
24812 94.281 -79.348 51.649
24813 94.294 -79.328 51.622
24814 94.307 -79.309 51.594
24815 94.319 -79.289 51.563
24816 94.332 -79.269 51.530
24817 94.347 -79.280 51.496
24818 94.364 -79.312 51.463
24819 94.382 -79.343 51.430
24820 94.395 -79.379 51.396
24821 94.404 -79.422 51.363
24822 94.413 -79.465 51.329
24823 94.422 -79.508 51.294
24824 94.430 -79.551 51.259
24825 94.439 -79.595 51.225
24826 94.448 -79.639 51.190
24827 94.457 -79.682 51.156
24828 94.466 -79.725 51.122
25363 92.513 -80.472 53.438
25364 92.529 -80.463 53.420
25365 92.545 -80.455 53.401
25366 92.561 -80.447 53.382
25367 92.577 -80.439 53.363
25368 92.593 -80.431 53.343
25369 92.609 -80.423 53.323
25370 92.625 -80.415 53.304
25371 92.641 -80.407 53.285
25372 92.657 -80.399 53.267
25373 92.673 -80.391 53.250
25374 92.690 -80.383 53.234
25375 92.705 -80.375 53.221
25376 92.721 -80.367 53.208
25377 92.740 -80.359 53.195
25378 92.761 -80.350 53.182
25379 92.782 -80.342 53.167
25380 92.802 -80.334 53.152
25381 92.824 -80.325 53.136
25382 92.844 -80.317 53.119
25383 92.865 -80.309 53.101
25384 92.886 -80.300 53.083
25385 92.907 -80.292 53.065
25386 92.929 -80.275 53.047
25387 92.951 -80.248 53.026
25388 92.975 -80.221 53.007
25389 92.990 -80.216 52.988
25390 93.003 -80.222 52.970
25391 93.016 -80.226 52.954
25392 93.028 -80.231 52.940
25393 93.041 -80.236 52.927
25394 93.048 -80.233 52.915
25395 93.061 -80.225 52.906
25396 93.086 -80.216 52.896
25397 93.111 -80.208 52.888
25398 93.135 -80.199 52.880
25399 93.157 -80.189 52.874
25400 93.180 -80.178 52.869
25401 93.202 -80.168 52.865
25402 93.224 -80.159 52.862
25403 93.247 -80.148 52.859
25404 93.261 -80.137 52.855
25405 93.260 -80.115 52.850
25406 93.258 -80.097 52.845
25407 93.260 -80.077 52.841
25408 93.269 -80.053 52.837
25409 93.278 -80.034 52.833
25410 93.287 -80.015 52.829
25411 93.296 -79.996 52.827
25412 93.305 -79.977 52.825
25413 93.317 -79.961 52.822
25414 93.331 -79.952 52.819
25415 93.352 -79.947 52.819
25416 93.376 -79.957 52.821
25417 93.387 -79.966 52.825
25418 93.398 -79.976 52.829
25419 93.421 -79.949 52.834
25420 93.436 -79.940 52.839
25421 93.449 -79.934 52.844
25422 93.463 -79.928 52.849
25423 93.476 -79.923 52.853
25424 93.489 -79.917 52.854
25425 93.502 -79.911 52.853
25426 93.516 -79.905 52.851
25427 93.529 -79.899 52.846
25428 93.543 -79.893 52.843
25429 93.567 -79.893 52.840
25430 93.591 -79.894 52.835
25431 93.615 -79.894 52.829
25432 93.639 -79.895 52.823
25433 93.654 -79.889 52.817
25434 93.664 -79.880 52.810
25435 93.674 -79.872 52.803
25436 93.684 -79.863 52.796
25437 93.694 -79.855 52.789
25438 93.704 -79.845 52.781
25439 93.714 -79.837 52.772
25440 93.723 -79.836 52.764
25441 93.731 -79.834 52.757
25442 93.740 -79.833 52.748
25443 93.749 -79.831 52.739
25444 93.757 -79.830 52.731
25445 93.766 -79.828 52.722
25446 93.774 -79.827 52.713
25465 93.937 -79.798 52.544
25466 93.945 -79.798 52.533
25467 93.954 -79.796 52.520
25468 93.963 -79.795 52.509
25469 93.971 -79.793 52.496
25470 93.980 -79.792 52.483
25471 93.988 -79.790 52.469
25472 93.997 -79.789 52.456
25473 94.005 -79.787 52.442
25474 94.014 -79.786 52.427
25475 94.023 -79.784 52.411
25476 94.031 -79.783 52.395
25477 94.040 -79.781 52.379
25478 94.048 -79.780 52.363
25479 94.057 -79.778 52.346
25480 94.065 -79.777 52.331
25481 94.074 -79.776 52.313
25482 94.082 -79.774 52.295
25483 94.091 -79.773 52.278
25484 94.100 -79.771 52.261
25485 94.108 -79.770 52.244
25486 94.117 -79.768 52.228
25487 94.125 -79.767 52.211
25492 94.168 -79.719 52.122
25493 94.177 -79.692 52.104
25494 94.185 -79.667 52.087
25495 94.194 -79.641 52.068
25496 94.202 -79.615 52.048
25497 94.210 -79.589 52.028
25498 94.219 -79.563 52.008
25499 94.227 -79.538 51.988
25500 94.236 -79.512 51.968
25501 94.245 -79.486 51.948
25502 94.253 -79.460 51.927
25503 94.262 -79.434 51.904
25504 94.270 -79.408 51.882
25505 94.279 -79.382 51.860
25506 94.269 -79.386 51.838
25507 94.259 -79.391 51.816
25508 94.248 -79.397 51.794
25509 94.237 -79.404 51.769
25510 94.240 -79.394 51.742
25511 94.253 -79.374 51.714
25512 94.265 -79.354 51.688
25513 94.278 -79.335 51.663
25514 94.291 -79.315 51.637
25515 94.303 -79.295 51.609
25516 94.316 -79.275 51.577
25517 94.329 -79.256 51.542
25518 94.343 -79.249 51.508
25519 94.360 -79.280 51.475
25520 94.376 -79.313 51.441
25521 94.384 -79.357 51.406
25522 94.393 -79.400 51.372
25523 94.402 -79.443 51.338
25524 94.411 -79.486 51.303
25525 94.419 -79.529 51.268
25526 94.428 -79.573 51.233
25527 94.437 -79.616 51.198
25528 94.446 -79.659 51.164
25529 94.455 -79.702 51.129
26065 92.525 -80.447 53.442
26066 92.541 -80.438 53.423
26067 92.557 -80.430 53.404
26068 92.573 -80.422 53.385
26069 92.589 -80.414 53.365
26070 92.605 -80.406 53.345
26071 92.621 -80.398 53.325
26072 92.637 -80.390 53.305
26073 92.653 -80.382 53.285
26074 92.669 -80.374 53.266
26075 92.685 -80.366 53.248
26076 92.701 -80.358 53.232
26077 92.717 -80.350 53.216
26078 92.733 -80.342 53.201
26079 92.749 -80.334 53.187
26080 92.766 -80.326 53.174
26081 92.786 -80.318 53.160
26082 92.807 -80.309 53.146
26083 92.828 -80.301 53.131
26084 92.849 -80.293 53.115
26085 92.869 -80.284 53.097
26086 92.890 -80.276 53.077
26087 92.911 -80.268 53.058
26088 92.932 -80.259 53.037
26089 92.955 -80.231 53.017
26090 92.978 -80.204 52.998
26091 92.994 -80.198 52.978
26092 93.007 -80.203 52.961
26093 93.019 -80.209 52.946
26094 93.032 -80.214 52.933
26095 93.042 -80.213 52.921
26096 93.050 -80.208 52.910
26097 93.068 -80.196 52.901
26098 93.090 -80.186 52.892
26099 93.112 -80.176 52.884
26100 93.135 -80.166 52.877
26101 93.157 -80.156 52.871
26102 93.180 -80.146 52.867
26103 93.202 -80.136 52.863
26104 93.224 -80.125 52.859
26105 93.248 -80.120 52.855
26106 93.247 -80.099 52.850
26107 93.245 -80.078 52.844
26108 93.252 -80.060 52.840
26109 93.261 -80.041 52.835
26110 93.270 -80.022 52.830
26111 93.279 -80.002 52.824
26112 93.288 -79.983 52.822
26113 93.298 -79.967 52.819
26114 93.313 -79.958 52.817
26115 93.328 -79.949 52.815
26116 93.347 -79.942 52.815
26117 93.381 -79.952 52.817
26118 93.393 -79.961 52.820
26119 93.403 -79.971 52.824
26120 93.428 -79.940 52.830
26121 93.445 -79.927 52.834
26122 93.459 -79.921 52.840
26123 93.472 -79.914 52.845
26124 93.485 -79.908 52.849
26125 93.498 -79.902 52.851
26126 93.512 -79.896 52.851
26127 93.525 -79.891 52.849
26128 93.538 -79.885 52.844
26129 93.552 -79.879 52.841
26130 93.565 -79.873 52.838
26131 93.578 -79.867 52.833
26132 93.596 -79.864 52.827
26133 93.620 -79.864 52.821
26134 93.644 -79.864 52.814
26135 93.664 -79.862 52.807
26136 93.674 -79.853 52.800
26137 93.684 -79.845 52.793
26138 93.694 -79.836 52.787
26139 93.704 -79.827 52.778
26140 93.713 -79.823 52.770
26141 93.722 -79.821 52.762
26142 93.730 -79.820 52.755
26143 93.739 -79.818 52.746
26144 93.748 -79.817 52.738
26145 93.756 -79.815 52.730
26146 93.765 -79.814 52.721
26147 93.773 -79.812 52.712
26148 93.782 -79.811 52.704
26166 93.936 -79.785 52.546
26167 93.944 -79.783 52.535
26168 93.953 -79.782 52.523
26169 93.962 -79.780 52.511
26170 93.970 -79.779 52.499
26171 93.979 -79.777 52.486
26172 93.987 -79.776 52.472
26173 93.996 -79.774 52.458
26174 94.004 -79.773 52.445
26175 94.013 -79.771 52.430
26176 94.022 -79.770 52.414
26177 94.030 -79.768 52.398
26178 94.039 -79.767 52.383
26179 94.047 -79.765 52.366
26180 94.056 -79.764 52.349
26181 94.064 -79.763 52.334
26182 94.073 -79.761 52.316
26183 94.081 -79.760 52.298
26184 94.090 -79.758 52.281
26185 94.099 -79.757 52.264
26186 94.107 -79.755 52.248
26187 94.116 -79.754 52.231
26188 94.124 -79.752 52.214
26189 94.133 -79.751 52.196
26190 94.141 -79.749 52.178
26191 94.150 -79.748 52.160
26192 94.158 -79.731 52.142
26193 94.167 -79.705 52.125
26194 94.176 -79.679 52.108
26195 94.184 -79.654 52.091
26196 94.193 -79.628 52.071
26197 94.201 -79.601 52.051
26198 94.210 -79.576 52.032
26199 94.218 -79.550 52.012
26200 94.227 -79.524 51.992
26201 94.236 -79.498 51.973
26202 94.244 -79.473 51.953
26203 94.252 -79.447 51.932
26204 94.261 -79.421 51.910
26205 94.269 -79.395 51.889
26206 94.278 -79.369 51.868
26207 94.273 -79.366 51.848
26208 94.262 -79.372 51.829
26209 94.251 -79.378 51.807
26210 94.240 -79.383 51.783
26211 94.237 -79.381 51.754
26212 94.250 -79.361 51.727
26213 94.262 -79.341 51.702
26214 94.275 -79.322 51.678
26215 94.288 -79.302 51.653
26216 94.300 -79.282 51.624
26217 94.314 -79.263 51.593
26218 94.326 -79.243 51.556
26219 94.339 -79.223 51.521
26220 94.355 -79.248 51.487
26221 94.364 -79.291 51.451
26222 94.373 -79.334 51.415
26223 94.382 -79.377 51.381
26224 94.391 -79.420 51.348
26225 94.400 -79.463 51.312
26226 94.409 -79.507 51.276
26227 94.417 -79.551 51.241
26228 94.426 -79.594 51.206
26229 94.435 -79.637 51.171
26230 94.444 -79.680 51.137
26766 92.521 -80.429 53.465
26767 92.537 -80.422 53.446
26768 92.553 -80.413 53.426
26769 92.569 -80.405 53.407
26770 92.585 -80.397 53.387
26771 92.601 -80.389 53.367
26772 92.617 -80.381 53.347
26773 92.633 -80.373 53.325
26774 92.649 -80.365 53.305
26775 92.665 -80.357 53.284
26776 92.681 -80.349 53.265
26777 92.697 -80.341 53.245
26778 92.713 -80.333 53.227
26779 92.729 -80.325 53.210
26780 92.745 -80.317 53.195
26781 92.761 -80.309 53.180
26782 92.777 -80.301 53.166
26783 92.793 -80.293 53.153
26784 92.811 -80.284 53.139
26785 92.832 -80.276 53.124
26786 92.853 -80.268 53.107
26787 92.873 -80.259 53.088
26788 92.895 -80.251 53.068
26789 92.915 -80.243 53.047
26790 92.936 -80.234 53.027
26791 92.958 -80.214 53.007
26792 92.981 -80.187 52.987
26793 92.998 -80.180 52.968
26794 93.010 -80.185 52.952
26795 93.023 -80.190 52.938
26796 93.035 -80.193 52.926
26797 93.042 -80.190 52.915
26798 93.056 -80.177 52.905
26799 93.070 -80.164 52.896
26800 93.089 -80.153 52.888
26801 93.112 -80.143 52.880
26802 93.134 -80.133 52.873
26803 93.156 -80.123 52.867
26804 93.179 -80.113 52.862
26805 93.201 -80.103 52.859
26806 93.225 -80.096 52.854
26807 93.233 -80.083 52.849
26808 93.231 -80.062 52.844
26809 93.244 -80.047 52.839
26810 93.253 -80.028 52.833
26811 93.262 -80.009 52.827
26812 93.271 -79.990 52.821
26813 93.281 -79.971 52.817
26814 93.295 -79.962 52.814
26815 93.310 -79.953 52.811
26816 93.325 -79.944 52.810
26817 93.341 -79.936 52.811
26818 93.387 -79.946 52.813
26819 93.398 -79.955 52.816
26820 93.410 -79.961 52.821
26821 93.435 -79.930 52.826
26822 93.454 -79.913 52.830
26823 93.468 -79.907 52.837
26824 93.481 -79.901 52.842
26825 93.494 -79.895 52.846
26826 93.508 -79.889 52.849
26827 93.521 -79.883 52.849
26828 93.534 -79.877 52.848
26829 93.548 -79.871 52.844
26830 93.561 -79.864 52.841
26831 93.574 -79.858 52.838
26832 93.587 -79.853 52.832
26833 93.601 -79.847 52.825
26834 93.614 -79.841 52.818
26835 93.627 -79.835 52.812
26836 93.649 -79.834 52.805
26837 93.673 -79.834 52.798
26838 93.684 -79.827 52.791
26839 93.694 -79.817 52.784
26840 93.704 -79.810 52.776
26841 93.713 -79.808 52.768
26842 93.721 -79.807 52.761
26843 93.730 -79.805 52.753
26844 93.739 -79.804 52.745
26845 93.747 -79.802 52.737
26846 93.756 -79.801 52.729
26847 93.764 -79.799 52.720
26848 93.773 -79.798 52.712
26849 93.781 -79.796 52.704
26867 93.935 -79.770 52.549
26868 93.943 -79.769 52.538
26869 93.952 -79.767 52.526
26870 93.961 -79.766 52.514
26871 93.969 -79.764 52.501
26872 93.978 -79.763 52.488
26873 93.986 -79.761 52.475
26874 93.995 -79.760 52.461
26875 94.003 -79.758 52.447
26876 94.012 -79.757 52.432
26877 94.020 -79.755 52.416
26878 94.029 -79.754 52.400
26879 94.038 -79.752 52.385
26880 94.046 -79.751 52.369
26881 94.055 -79.750 52.352
26882 94.063 -79.748 52.336
26883 94.072 -79.747 52.318
26884 94.080 -79.745 52.300
26885 94.089 -79.744 52.284
26886 94.098 -79.742 52.267
26887 94.106 -79.741 52.250
26888 94.115 -79.739 52.233
26889 94.123 -79.738 52.216
26890 94.132 -79.736 52.198
26891 94.140 -79.735 52.180
26892 94.149 -79.733 52.162
26893 94.157 -79.717 52.144
26894 94.166 -79.692 52.127
26895 94.175 -79.666 52.111
26896 94.183 -79.640 52.093
26897 94.192 -79.614 52.074
26898 94.200 -79.588 52.054
26899 94.209 -79.563 52.034
26900 94.217 -79.537 52.014
26901 94.226 -79.510 51.995
26902 94.235 -79.485 51.976
26903 94.243 -79.459 51.956
26904 94.252 -79.433 51.936
26905 94.260 -79.407 51.914
26906 94.269 -79.382 51.895
26907 94.277 -79.356 51.876
26908 94.276 -79.346 51.858
26909 94.265 -79.352 51.839
26910 94.265 -79.353 51.817
26911 94.278 -79.352 51.792
26912 94.290 -79.350 51.764
26913 94.302 -79.348 51.737
26914 94.313 -79.338 51.712
26915 94.319 -79.316 51.688
26916 94.326 -79.296 51.662
26917 94.332 -79.275 51.634
26918 94.337 -79.254 51.604
26919 94.339 -79.233 51.569
26920 94.340 -79.212 51.534
26921 94.345 -79.225 51.498
26922 94.354 -79.269 51.460
26923 94.362 -79.312 51.424
26924 94.371 -79.355 51.389
26925 94.380 -79.398 51.355
26926 94.389 -79.441 51.319
26927 94.398 -79.485 51.284
26928 94.407 -79.528 51.249
26929 94.415 -79.571 51.215
26930 94.424 -79.614 51.179
26931 94.433 -79.657 51.145
27468 92.533 -80.404 53.469
27469 92.549 -80.397 53.449
27470 92.565 -80.388 53.429
27471 92.581 -80.380 53.410
27472 92.597 -80.372 53.389
27473 92.613 -80.364 53.368
27474 92.629 -80.356 53.347
27475 92.645 -80.348 53.325
27476 92.661 -80.340 53.304
27477 92.677 -80.332 53.283
27478 92.693 -80.324 53.262
27479 92.709 -80.316 53.243
27480 92.725 -80.308 53.223
27481 92.741 -80.300 53.204
27482 92.757 -80.292 53.187
27483 92.773 -80.284 53.172
27484 92.789 -80.276 53.159
27485 92.805 -80.268 53.145
27486 92.821 -80.259 53.131
27487 92.838 -80.252 53.115
27488 92.857 -80.243 53.096
27489 92.878 -80.235 53.077
27490 92.899 -80.227 53.056
27491 92.919 -80.218 53.035
27492 92.940 -80.210 53.015
27493 92.961 -80.197 52.995
27494 92.985 -80.170 52.975
27495 93.001 -80.162 52.957
27496 93.014 -80.167 52.943
27497 93.027 -80.172 52.932
27498 93.035 -80.171 52.920
27499 93.048 -80.159 52.910
27500 93.062 -80.146 52.901
27501 93.075 -80.134 52.893
27502 93.089 -80.121 52.884
27503 93.111 -80.111 52.876
27504 93.134 -80.100 52.869
27505 93.156 -80.090 52.863
27506 93.178 -80.080 52.859
27507 93.202 -80.073 52.855
27508 93.219 -80.066 52.850
27509 93.222 -80.047 52.845
27510 93.236 -80.035 52.838
27511 93.246 -80.015 52.831
27512 93.255 -79.996 52.825
27513 93.264 -79.977 52.819
27514 93.278 -79.968 52.813
27515 93.292 -79.958 52.808
27516 93.307 -79.949 52.805
27517 93.322 -79.940 52.805
27518 93.337 -79.931 52.807
27519 93.386 -79.940 52.809
27520 93.404 -79.950 52.814
27521 93.417 -79.952 52.819
27522 93.442 -79.921 52.823
27523 93.464 -79.899 52.828
27524 93.477 -79.892 52.834
27525 93.490 -79.886 52.839
27526 93.503 -79.880 52.844
27527 93.517 -79.875 52.846
27528 93.530 -79.869 52.846
27529 93.544 -79.863 52.846
27530 93.557 -79.857 52.844
27531 93.570 -79.851 52.842
27532 93.583 -79.845 52.839
27533 93.596 -79.839 52.834
27534 93.610 -79.833 52.826
27535 93.623 -79.827 52.818
27536 93.636 -79.820 52.811
27537 93.650 -79.815 52.805
27538 93.663 -79.809 52.798
27539 93.678 -79.804 52.790
27540 93.694 -79.799 52.783
27541 93.703 -79.795 52.776
27542 93.712 -79.794 52.768
27543 93.720 -79.792 52.759
27544 93.729 -79.791 52.752
27545 93.737 -79.789 52.744
27546 93.746 -79.788 52.736
27547 93.755 -79.786 52.728
27548 93.763 -79.785 52.720
27549 93.772 -79.783 52.712
27550 93.780 -79.782 52.705
27551 93.789 -79.780 52.697
27568 93.934 -79.756 52.550
27569 93.943 -79.755 52.539
27570 93.952 -79.753 52.527
27571 93.960 -79.751 52.514
27572 93.969 -79.750 52.501
27573 93.977 -79.748 52.489
27574 93.986 -79.747 52.475
27575 93.994 -79.745 52.461
27576 94.003 -79.744 52.447
27577 94.011 -79.742 52.432
27578 94.020 -79.741 52.417
27579 94.029 -79.739 52.401
27580 94.037 -79.739 52.385
27581 94.046 -79.737 52.368
27582 94.054 -79.736 52.352
27583 94.063 -79.734 52.336
27584 94.071 -79.733 52.318
27585 94.080 -79.731 52.301
27586 94.088 -79.730 52.285
27587 94.097 -79.728 52.267
27588 94.106 -79.726 52.250
27589 94.114 -79.725 52.234
27590 94.123 -79.723 52.217
27591 94.131 -79.722 52.199
27592 94.140 -79.720 52.181
27593 94.148 -79.720 52.163
27594 94.157 -79.704 52.146
27595 94.165 -79.678 52.129
27596 94.174 -79.652 52.112
27597 94.182 -79.626 52.094
27598 94.191 -79.601 52.075
27599 94.199 -79.575 52.055
27600 94.208 -79.549 52.034
27601 94.216 -79.523 52.016
27602 94.225 -79.498 51.997
27603 94.234 -79.472 51.977
27604 94.242 -79.446 51.959
27605 94.251 -79.419 51.939
27606 94.259 -79.394 51.918
27607 94.268 -79.368 51.900
27608 94.276 -79.342 51.882
27609 94.286 -79.325 51.865
27610 94.293 -79.325 51.847
27611 94.301 -79.326 51.825
27612 94.309 -79.326 51.799
27613 94.317 -79.326 51.771
27614 94.325 -79.326 51.743
27615 94.329 -79.330 51.718
27616 94.329 -79.316 51.693
27617 94.331 -79.294 51.667
27618 94.332 -79.273 51.640
27619 94.334 -79.253 51.611
27620 94.335 -79.231 51.577
27621 94.337 -79.210 51.542
27622 94.334 -79.209 51.505
27623 94.343 -79.246 51.467
27624 94.352 -79.289 51.431
27625 94.360 -79.333 51.396
27626 94.369 -79.376 51.362
27627 94.378 -79.419 51.327
27628 94.387 -79.463 51.291
27629 94.396 -79.506 51.256
27630 94.405 -79.549 51.222
27631 94.413 -79.592 51.186
27632 94.422 -79.636 51.151
28169 92.528 -80.388 53.493
28170 92.545 -80.379 53.473
28171 92.561 -80.372 53.453
28172 92.577 -80.363 53.433
28173 92.593 -80.355 53.412
28174 92.609 -80.347 53.390
28175 92.625 -80.339 53.369
28176 92.641 -80.331 53.347
28177 92.657 -80.323 53.325
28178 92.673 -80.315 53.304
28179 92.689 -80.307 53.283
28180 92.705 -80.299 53.261
28181 92.721 -80.291 53.240
28182 92.737 -80.283 53.220
28183 92.753 -80.275 53.200
28184 92.769 -80.267 53.182
28185 92.785 -80.259 53.165
28186 92.801 -80.251 53.149
28187 92.817 -80.243 53.135
28188 92.833 -80.234 53.120
28189 92.849 -80.227 53.104
28190 92.865 -80.218 53.085
28191 92.882 -80.211 53.065
28192 92.903 -80.203 53.044
28193 92.924 -80.194 53.022
28194 92.944 -80.186 53.000
28195 92.965 -80.178 52.981
28196 92.988 -80.153 52.964
28197 93.005 -80.144 52.949
28198 93.018 -80.149 52.937
28199 93.028 -80.151 52.926
28200 93.040 -80.141 52.915
28201 93.054 -80.128 52.907
28202 93.067 -80.115 52.898
28203 93.081 -80.102 52.888
28204 93.094 -80.090 52.879
28205 93.111 -80.077 52.872
28206 93.133 -80.068 52.865
28207 93.155 -80.058 52.859
28208 93.179 -80.049 52.855
28209 93.206 -80.049 52.850
28210 93.213 -80.034 52.845
28211 93.227 -80.021 52.838
28212 93.237 -80.003 52.831
28213 93.247 -79.984 52.823
28214 93.260 -79.973 52.816
28215 93.275 -79.964 52.811
28216 93.289 -79.955 52.807
28217 93.304 -79.946 52.804
28218 93.319 -79.936 52.805
28219 93.334 -79.927 52.806
28220 93.380 -79.935 52.808
28221 93.409 -79.945 52.813
28222 93.424 -79.943 52.818
28223 93.449 -79.912 52.823
28224 93.473 -79.885 52.827
28225 93.486 -79.879 52.832
28226 93.499 -79.873 52.836
28227 93.513 -79.867 52.841
28228 93.526 -79.861 52.842
28229 93.540 -79.855 52.843
28230 93.553 -79.849 52.843
28231 93.566 -79.842 52.844
28232 93.579 -79.837 52.843
28233 93.592 -79.831 52.840
28234 93.606 -79.825 52.835
28235 93.619 -79.819 52.827
28236 93.632 -79.813 52.819
28237 93.646 -79.807 52.810
28238 93.659 -79.801 52.804
28239 93.672 -79.795 52.797
28240 93.684 -79.787 52.790
28241 93.694 -79.777 52.784
28242 93.708 -79.770 52.776
28243 93.715 -79.772 52.767
28244 93.721 -79.775 52.758
28245 93.728 -79.776 52.750
28246 93.736 -79.775 52.742
28247 93.745 -79.773 52.734
28248 93.754 -79.772 52.726
28249 93.762 -79.770 52.718
28250 93.771 -79.769 52.710
28251 93.779 -79.767 52.702
28252 93.788 -79.766 52.695
28269 93.933 -79.742 52.550
28270 93.942 -79.740 52.539
28271 93.951 -79.739 52.526
28272 93.959 -79.737 52.513
28273 93.968 -79.736 52.500
28274 93.976 -79.734 52.489
28275 93.985 -79.733 52.475
28276 93.993 -79.731 52.461
28277 94.002 -79.730 52.446
28278 94.010 -79.728 52.432
28279 94.019 -79.726 52.416
28280 94.028 -79.726 52.401
28281 94.036 -79.724 52.385
28282 94.045 -79.723 52.368
28283 94.053 -79.721 52.351
28284 94.062 -79.720 52.336
28285 94.070 -79.718 52.318
28286 94.079 -79.717 52.301
28287 94.087 -79.715 52.284
28288 94.096 -79.714 52.267
28289 94.105 -79.712 52.250
28290 94.113 -79.711 52.233
28291 94.122 -79.709 52.217
28292 94.130 -79.708 52.199
28293 94.139 -79.707 52.181
28294 94.147 -79.705 52.163
28295 94.156 -79.691 52.145
28296 94.165 -79.665 52.128
28297 94.173 -79.639 52.111
28298 94.182 -79.614 52.094
28299 94.190 -79.588 52.074
28300 94.199 -79.561 52.054
28301 94.207 -79.535 52.034
28302 94.215 -79.510 52.016
28303 94.224 -79.484 51.997
28304 94.232 -79.458 51.978
28305 94.241 -79.432 51.959
28306 94.250 -79.407 51.940
28307 94.258 -79.381 51.921
28308 94.266 -79.354 51.903
28309 94.274 -79.328 51.886
28310 94.288 -79.313 51.869
28311 94.296 -79.313 51.850
28312 94.305 -79.313 51.827
28313 94.312 -79.313 51.802
28314 94.320 -79.313 51.774
28315 94.328 -79.313 51.745
28316 94.326 -79.322 51.718
28317 94.326 -79.313 51.693
28318 94.327 -79.293 51.667
28319 94.329 -79.272 51.641
28320 94.331 -79.250 51.613
28321 94.332 -79.229 51.581
28322 94.331 -79.214 51.546
28323 94.323 -79.217 51.509
28324 94.332 -79.224 51.472
28325 94.341 -79.267 51.437
28326 94.350 -79.310 51.403
28327 94.358 -79.353 51.370
28328 94.367 -79.397 51.334
28329 94.376 -79.440 51.298
28330 94.385 -79.483 51.262
28331 94.394 -79.526 51.228
28332 94.403 -79.570 51.193
28333 94.411 -79.613 51.157
28871 92.541 -80.363 53.498
28872 92.557 -80.354 53.478
28873 92.573 -80.347 53.457
28874 92.589 -80.338 53.435
28875 92.605 -80.330 53.413
28876 92.621 -80.322 53.392
28877 92.637 -80.314 53.370
28878 92.653 -80.306 53.347
28879 92.669 -80.298 53.326
28880 92.685 -80.290 53.304
28881 92.701 -80.282 53.282
28882 92.717 -80.274 53.259
28883 92.733 -80.266 53.238
28884 92.749 -80.258 53.217
28885 92.765 -80.250 53.197
28886 92.781 -80.242 53.178
28887 92.797 -80.234 53.159
28888 92.813 -80.226 53.142
28889 92.829 -80.218 53.126
28890 92.845 -80.209 53.109
28891 92.861 -80.202 53.091
28892 92.877 -80.193 53.071
28893 92.894 -80.186 53.050
28894 92.909 -80.178 53.029
28895 92.928 -80.169 53.007
28896 92.948 -80.161 52.988
28897 92.970 -80.153 52.972
28898 92.991 -80.136 52.956
28899 93.009 -80.126 52.943
28900 93.021 -80.131 52.932
28901 93.031 -80.124 52.922
28902 93.046 -80.111 52.914
28903 93.059 -80.098 52.902
28904 93.073 -80.085 52.891
28905 93.086 -80.072 52.883
28906 93.100 -80.059 52.874
28907 93.114 -80.046 52.867
28908 93.133 -80.035 52.860
28909 93.155 -80.025 52.855
28910 93.183 -80.026 52.850
28911 93.204 -80.019 52.844
28912 93.219 -80.007 52.838
28913 93.230 -79.990 52.830
28914 93.242 -79.977 52.822
28915 93.257 -79.968 52.816
28916 93.272 -79.959 52.811
28917 93.286 -79.950 52.807
28918 93.301 -79.941 52.805
28919 93.316 -79.932 52.805
28920 93.331 -79.923 52.806
28921 93.374 -79.930 52.808
28922 93.415 -79.939 52.814
28923 93.431 -79.933 52.818
28924 93.457 -79.902 52.822
28925 93.482 -79.871 52.825
28926 93.495 -79.864 52.828
28927 93.509 -79.859 52.831
28928 93.522 -79.853 52.835
28929 93.535 -79.847 52.836
28930 93.549 -79.841 52.838
28931 93.562 -79.835 52.839
28932 93.575 -79.829 52.841
28933 93.588 -79.823 52.841
28934 93.602 -79.817 52.838
28935 93.615 -79.811 52.835
28936 93.628 -79.805 52.828
28937 93.641 -79.799 52.820
28938 93.653 -79.791 52.811
28939 93.663 -79.781 52.804
28940 93.673 -79.771 52.797
28941 93.683 -79.761 52.790
28942 93.693 -79.751 52.785
28943 93.706 -79.743 52.776
28944 93.721 -79.737 52.765
28945 93.734 -79.733 52.756
28946 93.741 -79.735 52.746
28947 93.748 -79.737 52.738
28948 93.754 -79.740 52.730
28949 93.761 -79.742 52.723
28950 93.767 -79.745 52.715
28951 93.774 -79.747 52.707
28952 93.780 -79.750 52.700
28953 93.787 -79.752 52.693
28954 93.795 -79.751 52.686
28956 93.813 -79.748 52.671
28957 93.821 -79.746 52.663
28958 93.830 -79.745 52.655
28959 93.838 -79.743 52.648
28960 93.847 -79.742 52.640
28961 93.855 -79.740 52.632
28962 93.864 -79.739 52.623
28963 93.872 -79.737 52.615
28964 93.881 -79.736 52.607
28965 93.890 -79.734 52.597
28970 93.932 -79.727 52.548
28971 93.941 -79.726 52.536
28972 93.949 -79.724 52.524
28973 93.958 -79.723 52.511
28974 93.967 -79.721 52.499
28975 93.975 -79.720 52.486
28976 93.984 -79.718 52.472
28977 93.992 -79.717 52.459
28978 94.001 -79.715 52.445
28979 94.009 -79.714 52.430
28980 94.018 -79.713 52.415
28981 94.027 -79.711 52.399
28982 94.035 -79.710 52.384
28983 94.044 -79.708 52.367
28984 94.052 -79.707 52.350
28985 94.061 -79.705 52.334
28986 94.069 -79.704 52.317
28987 94.078 -79.702 52.299
28988 94.086 -79.701 52.283
28989 94.095 -79.699 52.266
28990 94.104 -79.698 52.249
28991 94.112 -79.696 52.232
28992 94.121 -79.695 52.216
28993 94.129 -79.694 52.198
28994 94.138 -79.692 52.180
28995 94.146 -79.691 52.162
28996 94.155 -79.677 52.144
28997 94.164 -79.651 52.127
28998 94.172 -79.626 52.110
28999 94.181 -79.600 52.092
29000 94.189 -79.574 52.073
29001 94.198 -79.548 52.053
29002 94.206 -79.523 52.033
29003 94.215 -79.497 52.014
29004 94.223 -79.470 51.996
29005 94.232 -79.444 51.977
29006 94.241 -79.419 51.959
29007 94.249 -79.393 51.940
29008 94.257 -79.366 51.921
29009 94.264 -79.340 51.903
29010 94.273 -79.314 51.886
29011 94.289 -79.304 51.868
29012 94.300 -79.300 51.848
29013 94.308 -79.300 51.826
29014 94.316 -79.300 51.801
29015 94.323 -79.300 51.773
29016 94.327 -79.303 51.745
29017 94.324 -79.315 51.719
29018 94.323 -79.312 51.694
29019 94.324 -79.291 51.668
29020 94.326 -79.270 51.642
29021 94.327 -79.249 51.614
29022 94.329 -79.228 51.584
29023 94.323 -79.219 51.550
29024 94.311 -79.225 51.514
29025 94.321 -79.219 51.478
29026 94.330 -79.245 51.443
29027 94.339 -79.288 51.409
29028 94.348 -79.332 51.376
29029 94.356 -79.375 51.340
29030 94.365 -79.418 51.304
29031 94.374 -79.461 51.269
29032 94.383 -79.504 51.235
29033 94.392 -79.548 51.199
29034 94.401 -79.591 51.164
29035 94.409 -79.634 51.128
29572 92.538 -80.349 53.524
29573 92.553 -80.337 53.503
29574 92.569 -80.329 53.482
29575 92.585 -80.322 53.460
29576 92.600 -80.313 53.439
29577 92.617 -80.305 53.417
29578 92.633 -80.297 53.394
29579 92.649 -80.289 53.372
29580 92.665 -80.281 53.350
29581 92.681 -80.273 53.327
29582 92.697 -80.265 53.304
29583 92.713 -80.257 53.280
29584 92.729 -80.249 53.258
29585 92.745 -80.241 53.236
29586 92.761 -80.233 53.215
29587 92.777 -80.225 53.194
29588 92.793 -80.217 53.173
29589 92.809 -80.209 53.153
29590 92.825 -80.201 53.134
29591 92.841 -80.193 53.116
29592 92.857 -80.184 53.098
29593 92.873 -80.177 53.078
29594 92.889 -80.168 53.058
29595 92.905 -80.161 53.037
29596 92.921 -80.153 53.016
29597 92.937 -80.144 52.998
29598 92.953 -80.137 52.981
29599 92.974 -80.128 52.965
29600 92.995 -80.119 52.951
29601 93.012 -80.108 52.939
29602 93.023 -80.106 52.928
29603 93.037 -80.093 52.918
29604 93.051 -80.080 52.907
29605 93.065 -80.067 52.896
29606 93.078 -80.054 52.886
29607 93.092 -80.041 52.877
29608 93.106 -80.028 52.869
29609 93.119 -80.015 52.861
29610 93.133 -80.002 52.855
29611 93.172 -80.017 52.849
29612 93.190 -80.005 52.842
29613 93.208 -79.993 52.836
29614 93.224 -79.983 52.829
29615 93.239 -79.974 52.823
29616 93.254 -79.965 52.816
29617 93.269 -79.955 52.811
29618 93.283 -79.946 52.807
29619 93.298 -79.937 52.805
29620 93.313 -79.928 52.805
29621 93.328 -79.919 52.806
29622 93.369 -79.924 52.809
29623 93.418 -79.934 52.815
29624 93.438 -79.924 52.818
29625 93.464 -79.893 52.821
29626 93.489 -79.862 52.822
29627 93.505 -79.851 52.824
29628 93.518 -79.845 52.827
29629 93.531 -79.839 52.831
29630 93.545 -79.833 52.832
29631 93.558 -79.827 52.833
29632 93.571 -79.821 52.834
29633 93.584 -79.815 52.835
29634 93.597 -79.809 52.836
29635 93.611 -79.803 52.834
29636 93.623 -79.795 52.831
29637 93.632 -79.785 52.825
29638 93.642 -79.775 52.819
29639 93.652 -79.765 52.810
29640 93.662 -79.755 52.803
29641 93.672 -79.745 52.797
29642 93.682 -79.735 52.791
29643 93.692 -79.725 52.784
29644 93.704 -79.717 52.774
29645 93.719 -79.710 52.762
29646 93.734 -79.704 52.752
29647 93.750 -79.698 52.742
29648 93.761 -79.695 52.734
29649 93.767 -79.698 52.725
29650 93.774 -79.700 52.718
29651 93.780 -79.702 52.710
29652 93.787 -79.705 52.703
29653 93.793 -79.708 52.697
29654 93.800 -79.710 52.690
29655 93.806 -79.713 52.683
29656 93.813 -79.715 52.675
29657 93.820 -79.717 52.668
29658 93.826 -79.720 52.661
29659 93.833 -79.723 52.653
29660 93.840 -79.725 52.645
29661 93.846 -79.727 52.638
29662 93.855 -79.726 52.629
29663 93.863 -79.724 52.621
29664 93.872 -79.723 52.613
29665 93.880 -79.721 52.604
29666 93.889 -79.720 52.595
29667 93.898 -79.719 52.586
29668 93.906 -79.717 52.576
29669 93.915 -79.716 52.566
29670 93.923 -79.714 52.556
29671 93.932 -79.713 52.545
29672 93.940 -79.711 52.533
29673 93.949 -79.710 52.521
29674 93.958 -79.708 52.508
29675 93.966 -79.707 52.495
29676 93.975 -79.705 52.482
29677 93.983 -79.704 52.469
29678 93.992 -79.702 52.455
29679 94.000 -79.701 52.442
29680 94.009 -79.700 52.427
29681 94.017 -79.698 52.411
29682 94.026 -79.697 52.395
29683 94.035 -79.695 52.379
29684 94.043 -79.694 52.363
29685 94.051 -79.692 52.347
29686 94.060 -79.691 52.332
29687 94.068 -79.689 52.315
29688 94.077 -79.688 52.297
29689 94.085 -79.686 52.280
29690 94.094 -79.685 52.264
29691 94.103 -79.683 52.247
29692 94.111 -79.682 52.230
29693 94.120 -79.681 52.214
29694 94.128 -79.679 52.197
29695 94.137 -79.678 52.179
29696 94.145 -79.676 52.160
29697 94.154 -79.664 52.142
29698 94.163 -79.638 52.124
29699 94.171 -79.612 52.107
29700 94.180 -79.586 52.091
29701 94.188 -79.560 52.072
29702 94.197 -79.535 52.051
29703 94.205 -79.509 52.032
29704 94.214 -79.483 52.013
29705 94.222 -79.457 51.994
29706 94.231 -79.431 51.975
29707 94.239 -79.405 51.957
29708 94.247 -79.379 51.939
29709 94.254 -79.352 51.920
29710 94.262 -79.325 51.903
29711 94.273 -79.305 51.885
29712 94.289 -79.294 51.866
29713 94.303 -79.286 51.847
29714 94.311 -79.287 51.825
29715 94.319 -79.287 51.801
29716 94.327 -79.287 51.773
29717 94.325 -79.296 51.746
29718 94.321 -79.307 51.721
29719 94.319 -79.310 51.697
29720 94.321 -79.289 51.671
29721 94.322 -79.268 51.645
29722 94.324 -79.247 51.617
29723 94.325 -79.226 51.587
29724 94.316 -79.225 51.554
29725 94.302 -79.231 51.519
29726 94.310 -79.226 51.483
29727 94.319 -79.222 51.449
29728 94.328 -79.266 51.415
29729 94.337 -79.309 51.382
29730 94.346 -79.352 51.347
29731 94.354 -79.395 51.311
29732 94.363 -79.438 51.276
29733 94.372 -79.482 51.241
29734 94.381 -79.525 51.206
29735 94.390 -79.569 51.170
29736 94.398 -79.612 51.135
29737 94.407 -79.655 51.098
29738 94.416 -79.698 51.061
30274 92.552 -80.327 53.528
30275 92.565 -80.315 53.507
30276 92.580 -80.304 53.485
30277 92.596 -80.297 53.464
30278 92.613 -80.288 53.442
30279 92.629 -80.280 53.419
30280 92.645 -80.272 53.396
30281 92.661 -80.264 53.374
30282 92.676 -80.256 53.351
30283 92.693 -80.248 53.327
30284 92.709 -80.240 53.304
30285 92.725 -80.232 53.281
30286 92.741 -80.224 53.258
30287 92.757 -80.216 53.235
30288 92.773 -80.208 53.213
30289 92.789 -80.200 53.191
30290 92.805 -80.192 53.169
30291 92.821 -80.184 53.148
30292 92.837 -80.176 53.127
30293 92.853 -80.168 53.106
30294 92.869 -80.159 53.086
30295 92.885 -80.152 53.066
30296 92.901 -80.143 53.047
30297 92.917 -80.136 53.027
30298 92.933 -80.127 53.009
30299 92.949 -80.119 52.991
30300 92.965 -80.112 52.975
30301 92.981 -80.103 52.960
30302 92.999 -80.095 52.946
30303 93.015 -80.088 52.935
30304 93.029 -80.075 52.923
30305 93.043 -80.062 52.911
30306 93.059 -80.050 52.900
30307 93.083 -80.040 52.890
30308 93.108 -80.030 52.880
30309 93.132 -80.020 52.871
30310 93.157 -80.009 52.863
30311 93.182 -79.999 52.855
30312 93.200 -79.990 52.848
30313 93.211 -79.983 52.841
30314 93.222 -79.975 52.835
30315 93.233 -79.968 52.829
30316 93.246 -79.960 52.822
30317 93.259 -79.953 52.816
30318 93.272 -79.946 52.811
30319 93.285 -79.938 52.808
30320 93.297 -79.931 52.807
30321 93.310 -79.924 52.808
30322 93.325 -79.914 52.810
30323 93.363 -79.918 52.812
30324 93.413 -79.929 52.814
30325 93.445 -79.914 52.816
30326 93.471 -79.883 52.818
30327 93.496 -79.853 52.820
30328 93.514 -79.837 52.821
30329 93.527 -79.831 52.823
30330 93.541 -79.825 52.827
30331 93.554 -79.819 52.829
30332 93.567 -79.813 52.830
30333 93.580 -79.807 52.829
30334 93.592 -79.799 52.829
30335 93.602 -79.789 52.829
30336 93.612 -79.779 52.829
30337 93.622 -79.769 52.825
30338 93.632 -79.759 52.820
30339 93.641 -79.749 52.815
30340 93.651 -79.739 52.807
30341 93.661 -79.729 52.800
30342 93.671 -79.719 52.794
30343 93.681 -79.709 52.788
30344 93.691 -79.699 52.779
30345 93.702 -79.690 52.769
30346 93.717 -79.683 52.758
30347 93.732 -79.677 52.746
30348 93.748 -79.670 52.736
30349 93.763 -79.664 52.727
30350 93.778 -79.657 52.720
30351 93.787 -79.657 52.713
30352 93.793 -79.661 52.705
30353 93.800 -79.663 52.698
30354 93.806 -79.665 52.692
30355 93.813 -79.668 52.685
30356 93.820 -79.670 52.678
30357 93.826 -79.673 52.671
30358 93.833 -79.675 52.664
30359 93.840 -79.678 52.657
30360 93.846 -79.680 52.650
30361 93.853 -79.683 52.641
30362 93.859 -79.686 52.634
30363 93.866 -79.688 52.626
30364 93.872 -79.690 52.618
30365 93.879 -79.692 52.610
30366 93.886 -79.695 52.601
30367 93.892 -79.698 52.591
30368 93.899 -79.700 52.582
30369 93.905 -79.703 52.573
30370 93.914 -79.701 52.562
30371 93.922 -79.700 52.552
30372 93.931 -79.698 52.541
30373 93.939 -79.697 52.530
30374 93.948 -79.695 52.517
30375 93.957 -79.694 52.505
30376 93.965 -79.692 52.491
30377 93.974 -79.691 52.478
30378 93.982 -79.689 52.465
30379 93.991 -79.688 52.452
30380 93.999 -79.687 52.437
30381 94.008 -79.686 52.423
30382 94.016 -79.684 52.408
30383 94.025 -79.683 52.392
30384 94.034 -79.681 52.376
30385 94.042 -79.679 52.360
30386 94.051 -79.678 52.344
30387 94.059 -79.676 52.329
30388 94.068 -79.675 52.312
30389 94.076 -79.673 52.295
30390 94.085 -79.672 52.279
30391 94.094 -79.670 52.262
30392 94.102 -79.669 52.245
30393 94.111 -79.668 52.229
30394 94.119 -79.667 52.213
30395 94.128 -79.665 52.195
30396 94.136 -79.664 52.177
30397 94.145 -79.662 52.159
30398 94.153 -79.651 52.141
30399 94.162 -79.625 52.123
30400 94.171 -79.599 52.106
30401 94.179 -79.573 52.088
30402 94.187 -79.547 52.069
30403 94.196 -79.521 52.049
30404 94.204 -79.495 52.030
30405 94.213 -79.469 52.011
30406 94.221 -79.443 51.992
30407 94.228 -79.416 51.974
30408 94.237 -79.390 51.957
30409 94.245 -79.363 51.939
30410 94.252 -79.338 51.920
30411 94.260 -79.311 51.902
30412 94.274 -79.295 51.884
30413 94.289 -79.285 51.866
30414 94.306 -79.275 51.848
30415 94.314 -79.273 51.827
30416 94.322 -79.273 51.803
30417 94.326 -79.278 51.775
30418 94.322 -79.288 51.749
30419 94.319 -79.300 51.724
30420 94.316 -79.308 51.700
30421 94.317 -79.288 51.675
30422 94.319 -79.266 51.649
30423 94.320 -79.245 51.621
30424 94.321 -79.225 51.592
30425 94.308 -79.231 51.559
30426 94.294 -79.237 51.525
30427 94.298 -79.234 51.489
30428 94.308 -79.228 51.456
30429 94.317 -79.244 51.422
30430 94.326 -79.287 51.388
30431 94.335 -79.330 51.354
30432 94.343 -79.373 51.319
30433 94.352 -79.416 51.284
30434 94.361 -79.460 51.249
30435 94.370 -79.503 51.213
30436 94.379 -79.546 51.177
30437 94.388 -79.589 51.142
30438 94.396 -79.632 51.105
30439 94.405 -79.676 51.068
30440 94.414 -79.719 51.030
30441 94.423 -79.762 50.993
30977 92.578 -80.292 53.510
30978 92.592 -80.280 53.489
30979 92.609 -80.272 53.467
30980 92.625 -80.263 53.444
30981 92.641 -80.255 53.421
30982 92.656 -80.247 53.398
30983 92.672 -80.239 53.375
30984 92.689 -80.231 53.351
30985 92.705 -80.223 53.328
30986 92.721 -80.215 53.304
30987 92.737 -80.207 53.281
30988 92.753 -80.199 53.257
30989 92.769 -80.191 53.234
30990 92.785 -80.183 53.211
30991 92.801 -80.174 53.188
30992 92.817 -80.167 53.166
30993 92.833 -80.159 53.144
30994 92.849 -80.151 53.122
30995 92.865 -80.143 53.101
30996 92.881 -80.134 53.081
30997 92.897 -80.127 53.061
30998 92.913 -80.118 53.040
30999 92.929 -80.111 53.021
31000 92.945 -80.102 53.003
31001 92.961 -80.094 52.986
31002 92.977 -80.087 52.970
31003 92.993 -80.078 52.955
31004 93.009 -80.071 52.942
31005 93.023 -80.062 52.928
31006 93.044 -80.052 52.916
31007 93.064 -80.041 52.904
31008 93.085 -80.030 52.893
31009 93.106 -80.020 52.883
31010 93.126 -80.009 52.874
31011 93.147 -79.999 52.864
31012 93.167 -79.988 52.855
31013 93.188 -79.977 52.847
31014 93.213 -79.968 52.841
31015 93.231 -79.960 52.834
31016 93.243 -79.952 52.828
31017 93.256 -79.945 52.822
31018 93.269 -79.937 52.817
31019 93.281 -79.930 52.813
31020 93.295 -79.922 52.811
31021 93.308 -79.914 52.810
31022 93.322 -79.906 52.811
31023 93.335 -79.898 52.811
31024 93.359 -79.895 52.811
31025 93.406 -79.902 52.813
31026 93.442 -79.889 52.814
31027 93.478 -79.874 52.815
31028 93.503 -79.843 52.817
31029 93.523 -79.823 52.818
31030 93.537 -79.817 52.820
31031 93.550 -79.811 52.822
31032 93.561 -79.803 52.825
31033 93.571 -79.793 52.827
31034 93.581 -79.783 52.827
31035 93.591 -79.773 52.824
31036 93.601 -79.763 52.823
31037 93.611 -79.753 52.821
31038 93.621 -79.743 52.817
31039 93.631 -79.733 52.813
31040 93.640 -79.723 52.809
31041 93.650 -79.713 52.802
31042 93.660 -79.703 52.796
31043 93.670 -79.693 52.790
31044 93.680 -79.683 52.782
31045 93.690 -79.673 52.773
31046 93.700 -79.664 52.763
31047 93.715 -79.657 52.751
31048 93.730 -79.651 52.740
31049 93.746 -79.644 52.730
31050 93.761 -79.638 52.721
31051 93.776 -79.631 52.713
31052 93.791 -79.625 52.707
31053 93.806 -79.618 52.700
31054 93.813 -79.620 52.693
31055 93.820 -79.623 52.687
31056 93.826 -79.626 52.680
31057 93.833 -79.628 52.673
31058 93.840 -79.630 52.667
31059 93.846 -79.633 52.660
31060 93.853 -79.636 52.653
31061 93.859 -79.638 52.646
31062 93.866 -79.640 52.638
31063 93.872 -79.643 52.630
31064 93.879 -79.645 52.623
31065 93.886 -79.648 52.615
31066 93.892 -79.651 52.607
31067 93.899 -79.653 52.598
31068 93.905 -79.655 52.588
31069 93.912 -79.657 52.578
31070 93.918 -79.661 52.569
31071 93.925 -79.663 52.558
31072 93.931 -79.665 52.547
31073 93.938 -79.668 52.537
31074 93.944 -79.670 52.526
31075 93.951 -79.673 52.513
31076 93.958 -79.675 52.501
31077 93.965 -79.678 52.488
31078 93.973 -79.676 52.475
31079 93.981 -79.675 52.462
31080 93.990 -79.674 52.448
31081 93.998 -79.673 52.434
31082 94.007 -79.671 52.419
31083 94.015 -79.670 52.404
31084 94.024 -79.668 52.388
31085 94.033 -79.667 52.373
31086 94.041 -79.665 52.357
31087 94.050 -79.664 52.342
31088 94.058 -79.662 52.327
31089 94.067 -79.661 52.310
31090 94.075 -79.659 52.293
31091 94.084 -79.657 52.277
31092 94.092 -79.656 52.259
31093 94.101 -79.655 52.243
31094 94.110 -79.654 52.227
31095 94.118 -79.652 52.211
31096 94.127 -79.651 52.193
31097 94.135 -79.649 52.175
31098 94.144 -79.648 52.157
31099 94.152 -79.637 52.140
31100 94.161 -79.611 52.122
31101 94.170 -79.585 52.105
31102 94.178 -79.560 52.087
31103 94.187 -79.534 52.068
31104 94.195 -79.508 52.048
31105 94.203 -79.482 52.029
31106 94.211 -79.455 52.010
31107 94.219 -79.429 51.991
31108 94.226 -79.402 51.974
31109 94.235 -79.375 51.958
31110 94.242 -79.349 51.940
31111 94.250 -79.322 51.922
31112 94.258 -79.297 51.904
31113 94.274 -79.286 51.886
31114 94.290 -79.275 51.868
31115 94.306 -79.265 51.850
31116 94.318 -79.260 51.829
31117 94.325 -79.260 51.805
31118 94.323 -79.270 51.779
31119 94.320 -79.281 51.754
31120 94.316 -79.292 51.729
31121 94.313 -79.303 51.706
31122 94.314 -79.285 51.682
31123 94.315 -79.265 51.657
31124 94.317 -79.244 51.630
31125 94.314 -79.231 51.599
31126 94.300 -79.237 51.567
31127 94.287 -79.243 51.533
31128 94.286 -79.242 51.497
31129 94.296 -79.235 51.463
31130 94.306 -79.228 51.430
31131 94.315 -79.264 51.397
31132 94.324 -79.307 51.362
31133 94.333 -79.350 51.328
31134 94.341 -79.394 51.293
31135 94.350 -79.438 51.256
31136 94.359 -79.481 51.220
31137 94.368 -79.524 51.185
31138 94.377 -79.567 51.149
31139 94.386 -79.610 51.112
31140 94.394 -79.654 51.074
31141 94.403 -79.697 51.037
31142 94.412 -79.740 50.999
31143 94.424 -79.760 50.962
31680 92.605 -80.257 53.492
31681 92.621 -80.247 53.469
31682 92.636 -80.238 53.447
31683 92.652 -80.231 53.424
31684 92.668 -80.222 53.401
31685 92.685 -80.214 53.377
31686 92.701 -80.206 53.351
31687 92.717 -80.198 53.327
31688 92.732 -80.190 53.304
31689 92.748 -80.182 53.279
31690 92.765 -80.174 53.255
31691 92.781 -80.166 53.232
31692 92.797 -80.158 53.208
31693 92.813 -80.149 53.185
31694 92.829 -80.142 53.162
31695 92.845 -80.134 53.140
31696 92.861 -80.126 53.118
31697 92.877 -80.118 53.097
31698 92.893 -80.109 53.076
31699 92.909 -80.102 53.055
31700 92.925 -80.093 53.035
31701 92.941 -80.086 53.017
31702 92.957 -80.077 52.998
31703 92.973 -80.069 52.981
31704 92.989 -80.062 52.965
31705 93.005 -80.053 52.949
31706 93.021 -80.046 52.934
31707 93.036 -80.036 52.920
31708 93.047 -80.030 52.908
31709 93.068 -80.019 52.895
31710 93.088 -80.008 52.885
31711 93.109 -79.998 52.876
31712 93.129 -79.987 52.866
31713 93.150 -79.977 52.856
31714 93.170 -79.966 52.847
31715 93.194 -79.956 52.840
31716 93.219 -79.947 52.834
31717 93.241 -79.938 52.828
31718 93.256 -79.930 52.823
31719 93.271 -79.921 52.820
31720 93.286 -79.913 52.817
31721 93.301 -79.905 52.814
31722 93.315 -79.896 52.814
31723 93.330 -79.888 52.813
31724 93.345 -79.880 52.812
31725 93.360 -79.871 52.812
31726 93.393 -79.859 52.813
31727 93.428 -79.846 52.813
31728 93.465 -79.833 52.814
31729 93.500 -79.820 52.815
31730 93.531 -79.807 52.817
31731 93.541 -79.797 52.818
31732 93.550 -79.787 52.819
31733 93.560 -79.777 52.821
31734 93.570 -79.767 52.823
31735 93.580 -79.757 52.823
31736 93.590 -79.747 52.820
31737 93.600 -79.737 52.817
31738 93.610 -79.727 52.813
31739 93.620 -79.717 52.810
31740 93.630 -79.707 52.806
31741 93.640 -79.697 52.802
31742 93.649 -79.687 52.797
31743 93.659 -79.677 52.790
31744 93.670 -79.667 52.783
31745 93.679 -79.657 52.776
31746 93.689 -79.647 52.765
31747 93.699 -79.637 52.754
31748 93.713 -79.630 52.742
31749 93.728 -79.624 52.733
31750 93.744 -79.617 52.723
31751 93.759 -79.611 52.714
31752 93.774 -79.604 52.707
31753 93.789 -79.598 52.700
31754 93.804 -79.592 52.693
31755 93.820 -79.585 52.688
31756 93.833 -79.581 52.682
31757 93.840 -79.583 52.676
31758 93.846 -79.585 52.669
31759 93.853 -79.588 52.663
31760 93.859 -79.591 52.657
31761 93.866 -79.593 52.649
31762 93.872 -79.595 52.642
31763 93.879 -79.598 52.635
31764 93.886 -79.601 52.627
31765 93.892 -79.603 52.620
31766 93.899 -79.606 52.612
31767 93.905 -79.608 52.604
31768 93.912 -79.610 52.595
31769 93.918 -79.613 52.585
31770 93.925 -79.616 52.575
31771 93.931 -79.618 52.565
31772 93.938 -79.620 52.554
31773 93.944 -79.623 52.544
31774 93.951 -79.626 52.533
31775 93.958 -79.628 52.522
31776 93.965 -79.630 52.509
31777 93.971 -79.633 52.497
31778 93.978 -79.636 52.485
31779 93.984 -79.638 52.471
31780 93.991 -79.641 52.458
31781 93.997 -79.643 52.444
31782 94.004 -79.645 52.431
31783 94.010 -79.648 52.416
31784 94.017 -79.651 52.401
31785 94.024 -79.653 52.386
31786 94.032 -79.652 52.370
31787 94.040 -79.651 52.354
31788 94.049 -79.649 52.339
31789 94.057 -79.648 52.324
31790 94.066 -79.646 52.308
31791 94.074 -79.645 52.291
31792 94.083 -79.643 52.274
31793 94.091 -79.642 52.257
31794 94.100 -79.641 52.241
31795 94.109 -79.639 52.225
31796 94.117 -79.638 52.209
31797 94.126 -79.636 52.192
31798 94.134 -79.635 52.174
31799 94.143 -79.633 52.156
31800 94.151 -79.624 52.139
31801 94.160 -79.598 52.122
31802 94.169 -79.572 52.104
31803 94.177 -79.546 52.086
31804 94.185 -79.520 52.067
31805 94.193 -79.494 52.047
31806 94.201 -79.467 52.029
31807 94.209 -79.441 52.010
31808 94.217 -79.414 51.991
31809 94.224 -79.388 51.975
31810 94.232 -79.361 51.959
31811 94.240 -79.335 51.943
31812 94.248 -79.308 51.925
31813 94.259 -79.287 51.906
31814 94.274 -79.276 51.888
31815 94.290 -79.266 51.870
31816 94.307 -79.256 51.851
31817 94.321 -79.247 51.830
31818 94.324 -79.251 51.808
31819 94.321 -79.263 51.784
31820 94.317 -79.273 51.760
31821 94.314 -79.285 51.736
31822 94.310 -79.296 51.713
31823 94.310 -79.284 51.690
31824 94.312 -79.263 51.665
31825 94.314 -79.242 51.639
31826 94.306 -79.237 51.608
31827 94.293 -79.243 51.575
31828 94.279 -79.249 51.542
31829 94.275 -79.250 51.507
31830 94.285 -79.244 51.472
31831 94.294 -79.237 51.439
31832 94.304 -79.242 51.407
31833 94.313 -79.285 51.372
31834 94.322 -79.328 51.337
31835 94.331 -79.372 51.302
31836 94.339 -79.415 51.265
31837 94.348 -79.458 51.229
31838 94.357 -79.501 51.193
31839 94.366 -79.545 51.157
31840 94.374 -79.588 51.119
31841 94.384 -79.631 51.082
31842 94.392 -79.674 51.044
31843 94.403 -79.708 51.006
31844 94.416 -79.720 50.968
31845 94.429 -79.734 50.931
31846 94.442 -79.747 50.894
32383 92.633 -80.222 53.472
32384 92.648 -80.213 53.448
32385 92.664 -80.206 53.425
32386 92.681 -80.197 53.400
32387 92.697 -80.189 53.376
32388 92.712 -80.181 53.352
32389 92.728 -80.173 53.327
32390 92.744 -80.165 53.303
32391 92.761 -80.157 53.278
32392 92.777 -80.149 53.254
32393 92.793 -80.141 53.229
32394 92.808 -80.133 53.205
32395 92.825 -80.124 53.182
32396 92.841 -80.117 53.160
32397 92.857 -80.109 53.137
32398 92.873 -80.101 53.115
32399 92.889 -80.093 53.093
32400 92.905 -80.084 53.071
32401 92.921 -80.077 53.051
32402 92.937 -80.068 53.031
32403 92.953 -80.061 53.012
32404 92.969 -80.052 52.993
32405 92.985 -80.044 52.975
32406 93.001 -80.037 52.959
32407 93.017 -80.028 52.942
32408 93.033 -80.021 52.927
32409 93.049 -80.008 52.913
32410 93.056 -80.005 52.900
32411 93.071 -79.997 52.889
32412 93.091 -79.987 52.878
32413 93.112 -79.976 52.868
32414 93.133 -79.965 52.859
32415 93.153 -79.955 52.851
32416 93.174 -79.944 52.842
32417 93.199 -79.935 52.836
32418 93.222 -79.926 52.831
32419 93.237 -79.918 52.826
32420 93.252 -79.909 52.824
32421 93.267 -79.901 52.821
32422 93.282 -79.892 52.819
32423 93.296 -79.884 52.818
32424 93.311 -79.876 52.817
32425 93.330 -79.867 52.815
32426 93.353 -79.857 52.814
32427 93.379 -79.846 52.814
32428 93.415 -79.834 52.814
32429 93.451 -79.821 52.815
32430 93.487 -79.808 52.816
32431 93.519 -79.795 52.817
32432 93.534 -79.778 52.818
32433 93.549 -79.762 52.817
32434 93.559 -79.751 52.818
32435 93.569 -79.741 52.818
32436 93.579 -79.731 52.818
32437 93.589 -79.721 52.815
32438 93.599 -79.711 52.811
32439 93.609 -79.701 52.807
32440 93.619 -79.691 52.803
32441 93.629 -79.681 52.799
32442 93.639 -79.671 52.795
32443 93.648 -79.661 52.789
32444 93.658 -79.651 52.782
32445 93.669 -79.642 52.774
32446 93.679 -79.631 52.767
32447 93.688 -79.621 52.756
32448 93.698 -79.611 52.745
32449 93.712 -79.604 52.733
32450 93.727 -79.597 52.724
32451 93.742 -79.591 52.715
32452 93.757 -79.585 52.708
32453 93.772 -79.578 52.701
32454 93.787 -79.572 52.694
32455 93.802 -79.565 52.688
32456 93.818 -79.559 52.683
32457 93.833 -79.552 52.678
32458 93.848 -79.546 52.672
32459 93.859 -79.544 52.665
32460 93.866 -79.546 52.660
32461 93.872 -79.548 52.653
32462 93.879 -79.551 52.646
32463 93.886 -79.554 52.640
32464 93.892 -79.556 52.632
32465 93.899 -79.558 52.625
32466 93.905 -79.561 52.617
32467 93.912 -79.563 52.610
32468 93.918 -79.566 52.601
32469 93.925 -79.568 52.592
32470 93.931 -79.571 52.582
32471 93.938 -79.573 52.572
32472 93.944 -79.576 52.562
32473 93.951 -79.579 52.552
32474 93.958 -79.581 52.541
32475 93.965 -79.583 52.530
32476 93.971 -79.585 52.518
32477 93.978 -79.588 52.506
32478 93.984 -79.591 52.493
32479 93.991 -79.593 52.480
32480 93.997 -79.596 52.467
32481 94.004 -79.598 52.454
32482 94.010 -79.601 52.440
32483 94.017 -79.603 52.427
32484 94.024 -79.606 52.412
32485 94.030 -79.608 52.397
32486 94.037 -79.610 52.383
32487 94.043 -79.614 52.367
32488 94.050 -79.616 52.352
32489 94.056 -79.618 52.336
32490 94.063 -79.620 52.321
32491 94.069 -79.623 52.305
32492 94.076 -79.626 52.288
32493 94.083 -79.628 52.272
32494 94.091 -79.628 52.255
32495 94.100 -79.626 52.239
32496 94.108 -79.625 52.223
32497 94.117 -79.623 52.208
32498 94.125 -79.622 52.191
32499 94.134 -79.620 52.174
32500 94.142 -79.619 52.156
32501 94.150 -79.610 52.140
32502 94.159 -79.585 52.124
32503 94.168 -79.558 52.106
32504 94.175 -79.532 52.087
32505 94.183 -79.506 52.068
32506 94.191 -79.479 52.048
32507 94.199 -79.453 52.029
32508 94.207 -79.426 52.010
32509 94.214 -79.400 51.994
32510 94.222 -79.373 51.978
32511 94.230 -79.347 51.962
32512 94.238 -79.320 51.946
32513 94.246 -79.294 51.927
32514 94.259 -79.278 51.908
32515 94.275 -79.267 51.889
32516 94.291 -79.256 51.871
32517 94.307 -79.246 51.852
32518 94.323 -79.235 51.832
32519 94.322 -79.244 51.811
32520 94.318 -79.255 51.789
32521 94.315 -79.266 51.767
32522 94.311 -79.277 51.744
32523 94.308 -79.288 51.721
32524 94.307 -79.282 51.698
32525 94.309 -79.261 51.674
32526 94.310 -79.240 51.648
32527 94.298 -79.243 51.617
32528 94.285 -79.249 51.585
32529 94.272 -79.254 51.552
32530 94.264 -79.258 51.518
32531 94.273 -79.251 51.484
32532 94.283 -79.244 51.450
32533 94.293 -79.238 51.417
32534 94.302 -79.263 51.382
32535 94.311 -79.306 51.346
32536 94.320 -79.350 51.311
32537 94.329 -79.393 51.275
32538 94.337 -79.436 51.239
32539 94.346 -79.479 51.202
32540 94.355 -79.523 51.165
32541 94.364 -79.566 51.128
32542 94.372 -79.609 51.090
32543 94.382 -79.652 51.051
32544 94.394 -79.668 51.013
32545 94.407 -79.682 50.974
32546 94.420 -79.695 50.937
32547 94.433 -79.708 50.899
32548 94.447 -79.720 50.861
33085 92.646 -80.200 53.473
33086 92.660 -80.188 53.449
33087 92.676 -80.181 53.425
33088 92.692 -80.172 53.400
33089 92.708 -80.164 53.376
33090 92.724 -80.156 53.351
33091 92.740 -80.148 53.326
33092 92.757 -80.140 53.301
33093 92.773 -80.132 53.276
33094 92.788 -80.124 53.252
33095 92.804 -80.116 53.228
33096 92.820 -80.108 53.204
33097 92.837 -80.100 53.180
33098 92.853 -80.092 53.157
33099 92.869 -80.084 53.134
33100 92.885 -80.076 53.111
33101 92.901 -80.068 53.089
33102 92.917 -80.059 53.068
33103 92.933 -80.052 53.046
33104 92.949 -80.043 53.026
33105 92.965 -80.036 53.006
33106 92.981 -80.027 52.987
33107 92.997 -80.019 52.970
33108 93.013 -80.012 52.952
33109 93.029 -80.003 52.936
33110 93.046 -79.994 52.920
33111 93.063 -79.981 52.907
33112 93.070 -79.977 52.893
33113 93.076 -79.974 52.880
33114 93.095 -79.965 52.871
33115 93.116 -79.954 52.863
33116 93.136 -79.943 52.854
33117 93.156 -79.933 52.846
33118 93.180 -79.923 52.840
33119 93.203 -79.914 52.835
33120 93.218 -79.905 52.831
33121 93.233 -79.897 52.829
33122 93.248 -79.889 52.827
33123 93.263 -79.880 52.824
33124 93.277 -79.872 52.823
33125 93.301 -79.862 52.821
33126 93.324 -79.852 52.819
33127 93.347 -79.843 52.818
33128 93.370 -79.833 52.818
33129 93.402 -79.821 52.818
33130 93.438 -79.808 52.818
33131 93.474 -79.795 52.819
33132 93.506 -79.783 52.819
33133 93.521 -79.766 52.819
33134 93.537 -79.750 52.817
33135 93.551 -79.734 52.815
33136 93.566 -79.718 52.814
33137 93.578 -79.705 52.812
33138 93.588 -79.695 52.809
33139 93.598 -79.686 52.805
33140 93.608 -79.675 52.801
33141 93.618 -79.665 52.796
33142 93.628 -79.655 52.791
33143 93.638 -79.645 52.786
33144 93.648 -79.636 52.779
33145 93.657 -79.625 52.772
33146 93.667 -79.615 52.764
33147 93.678 -79.605 52.756
33148 93.687 -79.595 52.746
33149 93.697 -79.585 52.736
33150 93.710 -79.577 52.725
33151 93.725 -79.570 52.716
33152 93.740 -79.564 52.708
33153 93.755 -79.557 52.701
33154 93.770 -79.551 52.694
33155 93.785 -79.545 52.688
33156 93.800 -79.538 52.682
33157 93.816 -79.532 52.678
33158 93.831 -79.526 52.672
33159 93.846 -79.520 52.667
33160 93.861 -79.513 52.661
33161 93.876 -79.507 52.656
33162 93.886 -79.506 52.649
33163 93.892 -79.509 52.642
33164 93.899 -79.511 52.636
33165 93.905 -79.513 52.629
33166 93.912 -79.516 52.622
33167 93.918 -79.519 52.614
33168 93.925 -79.521 52.606
33169 93.931 -79.523 52.598
33170 93.938 -79.526 52.590
33171 93.944 -79.529 52.580
33172 93.951 -79.531 52.570
33173 93.958 -79.533 52.559
33174 93.965 -79.536 52.548
33175 93.971 -79.538 52.537
33176 93.978 -79.541 52.525
33177 93.984 -79.544 52.514
33178 93.991 -79.546 52.501
33179 93.997 -79.548 52.488
33180 94.004 -79.551 52.475
33181 94.010 -79.554 52.463
33182 94.017 -79.556 52.450
33183 94.024 -79.558 52.436
33184 94.030 -79.561 52.423
33185 94.037 -79.563 52.408
33186 94.043 -79.566 52.393
33187 94.050 -79.569 52.379
33188 94.056 -79.571 52.364
33189 94.063 -79.573 52.349
33190 94.069 -79.576 52.333
33191 94.076 -79.579 52.319
33192 94.082 -79.581 52.302
33193 94.089 -79.583 52.285
33194 94.096 -79.585 52.269
33195 94.103 -79.588 52.253
33196 94.109 -79.591 52.237
33197 94.116 -79.593 52.222
33198 94.122 -79.596 52.206
33199 94.129 -79.598 52.190
33200 94.135 -79.601 52.174
33201 94.142 -79.604 52.158
33202 94.150 -79.597 52.142
33203 94.157 -79.570 52.125
33204 94.166 -79.544 52.107
33205 94.173 -79.517 52.089
33206 94.181 -79.491 52.070
33207 94.189 -79.464 52.050
33208 94.197 -79.438 52.030
33209 94.205 -79.412 52.012
33210 94.212 -79.385 51.996
33211 94.220 -79.359 51.980
33212 94.228 -79.332 51.965
33213 94.236 -79.306 51.949
33214 94.244 -79.279 51.930
33215 94.260 -79.268 51.910
33216 94.275 -79.257 51.890
33217 94.291 -79.247 51.870
33218 94.300 -79.244 51.851
33219 94.308 -79.242 51.832
33220 94.317 -79.239 51.814
33221 94.316 -79.247 51.794
33222 94.312 -79.258 51.774
33223 94.309 -79.269 51.753
33224 94.305 -79.281 51.730
33225 94.303 -79.280 51.707
33226 94.305 -79.259 51.684
33227 94.305 -79.242 51.660
33228 94.291 -79.248 51.630
33229 94.277 -79.254 51.599
33230 94.264 -79.260 51.566
33231 94.252 -79.266 51.533
33232 94.262 -79.259 51.498
33233 94.272 -79.253 51.464
33234 94.281 -79.246 51.430
33235 94.291 -79.241 51.393
33236 94.300 -79.284 51.357
33237 94.309 -79.327 51.321
33238 94.318 -79.370 51.284
33239 94.327 -79.413 51.249
33240 94.335 -79.457 51.213
33241 94.344 -79.500 51.176
33242 94.353 -79.543 51.137
33243 94.362 -79.586 51.099
33244 94.372 -79.616 51.059
33245 94.386 -79.629 51.020
33246 94.399 -79.642 50.981
33247 94.412 -79.655 50.943
33248 94.425 -79.668 50.905
33249 94.438 -79.681 50.867
33788 92.673 -80.165 53.450
33789 92.688 -80.156 53.425
33790 92.704 -80.147 53.400
33791 92.720 -80.139 53.375
33792 92.736 -80.131 53.350
33793 92.753 -80.123 53.325
33794 92.768 -80.115 53.299
33795 92.784 -80.107 53.275
33796 92.800 -80.099 53.251
33797 92.816 -80.091 53.226
33798 92.833 -80.083 53.202
33799 92.849 -80.075 53.178
33800 92.865 -80.067 53.154
33801 92.880 -80.059 53.130
33802 92.897 -80.051 53.107
33803 92.913 -80.043 53.085
33804 92.929 -80.035 53.064
33805 92.945 -80.027 53.043
33806 92.961 -80.018 53.022
33807 92.977 -80.011 53.002
33808 92.993 -80.002 52.982
33809 93.009 -79.994 52.964
33810 93.025 -79.987 52.947
33811 93.041 -79.978 52.930
33812 93.058 -79.968 52.915
33813 93.076 -79.954 52.900
33814 93.083 -79.950 52.887
33815 93.090 -79.946 52.875
33816 93.098 -79.943 52.866
33817 93.119 -79.932 52.857
33818 93.139 -79.921 52.851
33819 93.160 -79.911 52.846
33820 93.184 -79.902 52.841
33821 93.199 -79.893 52.837
33822 93.214 -79.885 52.835
33823 93.228 -79.877 52.833
33824 93.248 -79.867 52.831
33825 93.271 -79.858 52.829
33826 93.295 -79.848 52.826
33827 93.318 -79.839 52.824
33828 93.341 -79.829 52.822
33829 93.364 -79.819 52.821
33830 93.389 -79.809 52.821
33831 93.425 -79.796 52.821
33832 93.461 -79.783 52.822
33833 93.494 -79.770 52.822
33834 93.509 -79.754 52.820
33835 93.524 -79.738 52.817
33836 93.539 -79.722 52.814
33837 93.554 -79.706 52.810
33838 93.569 -79.689 52.807
33839 93.584 -79.673 52.804
33840 93.597 -79.659 52.800
33841 93.607 -79.649 52.795
33842 93.617 -79.639 52.788
33843 93.627 -79.629 52.782
33844 93.637 -79.620 52.777
33845 93.647 -79.609 52.770
33846 93.657 -79.599 52.762
33847 93.666 -79.589 52.754
33848 93.677 -79.579 52.746
33849 93.687 -79.570 52.736
33850 93.696 -79.559 52.727
33851 93.708 -79.551 52.717
33852 93.723 -79.544 52.708
33853 93.738 -79.538 52.700
33854 93.753 -79.531 52.692
33855 93.768 -79.525 52.686
33856 93.783 -79.518 52.680
33857 93.798 -79.512 52.675
33858 93.814 -79.505 52.672
33859 93.829 -79.499 52.667
33860 93.844 -79.492 52.662
33861 93.859 -79.486 52.657
33862 93.874 -79.479 52.652
33863 93.890 -79.473 52.646
33864 93.905 -79.466 52.639
33865 93.912 -79.469 52.633
33866 93.918 -79.471 52.627
33867 93.925 -79.474 52.620
33868 93.931 -79.476 52.612
33869 93.938 -79.479 52.604
33870 93.944 -79.482 52.595
33871 93.951 -79.484 52.587
33872 93.958 -79.486 52.577
33873 93.965 -79.488 52.566
33874 93.971 -79.491 52.555
33875 93.978 -79.494 52.544
33876 93.984 -79.496 52.533
33877 93.991 -79.499 52.522
33878 93.997 -79.501 52.510
33879 94.004 -79.504 52.497
33880 94.010 -79.506 52.484
33881 94.017 -79.509 52.471
33882 94.024 -79.511 52.458
33883 94.030 -79.513 52.445
33884 94.037 -79.516 52.432
33885 94.043 -79.519 52.419
33886 94.050 -79.521 52.404
33887 94.056 -79.523 52.390
33888 94.063 -79.526 52.375
33889 94.069 -79.529 52.361
33890 94.076 -79.531 52.346
33891 94.082 -79.534 52.332
33892 94.089 -79.536 52.317
33893 94.096 -79.538 52.300
33894 94.103 -79.541 52.283
33895 94.105 -79.546 52.267
33896 94.106 -79.551 52.250
33897 94.107 -79.557 52.234
33898 94.109 -79.562 52.219
33899 94.110 -79.568 52.205
33900 94.113 -79.573 52.190
33901 94.122 -79.574 52.174
33902 94.135 -79.571 52.159
33903 94.146 -79.558 52.143
33904 94.154 -79.550 52.126
33905 94.162 -79.524 52.109
33906 94.170 -79.499 52.092
33907 94.178 -79.474 52.073
33908 94.187 -79.448 52.053
33909 94.195 -79.423 52.032
33910 94.202 -79.397 52.014
33911 94.210 -79.370 51.997
33912 94.218 -79.344 51.982
33913 94.226 -79.318 51.967
33914 94.234 -79.291 51.951
33915 94.244 -79.269 51.932
33916 94.260 -79.259 51.911
33917 94.268 -79.256 51.890
33918 94.276 -79.254 51.870
33919 94.284 -79.251 51.851
33920 94.293 -79.249 51.834
33921 94.301 -79.247 51.817
33922 94.309 -79.244 51.798
33923 94.310 -79.250 51.780
33924 94.306 -79.262 51.761
33925 94.302 -79.273 51.739
33926 94.300 -79.278 51.717
33927 94.301 -79.257 51.694
33928 94.296 -79.248 51.671
33929 94.283 -79.254 51.642
33930 94.270 -79.260 51.611
33931 94.257 -79.266 51.579
33932 94.243 -79.272 51.547
33933 94.251 -79.267 51.514
33934 94.260 -79.260 51.479
33935 94.270 -79.253 51.445
33936 94.280 -79.247 51.408
33937 94.289 -79.262 51.370
33938 94.298 -79.305 51.333
33939 94.307 -79.348 51.296
33940 94.316 -79.391 51.259
33941 94.325 -79.435 51.223
33942 94.333 -79.478 51.185
33943 94.342 -79.521 51.147
33944 94.351 -79.563 51.108
33945 94.364 -79.577 51.068
33946 94.377 -79.590 51.028
33947 94.391 -79.603 50.989
33948 94.404 -79.616 50.951
33949 94.417 -79.629 50.912
33950 94.430 -79.642 50.873
34491 92.700 -80.131 53.424
34492 92.716 -80.122 53.399
34493 92.732 -80.114 53.374
34494 92.748 -80.106 53.348
34495 92.764 -80.098 53.324
34496 92.780 -80.090 53.299
34497 92.796 -80.082 53.274
34498 92.812 -80.074 53.249
34499 92.829 -80.066 53.223
34500 92.845 -80.058 53.199
34501 92.860 -80.050 53.175
34502 92.876 -80.042 53.150
34503 92.893 -80.034 53.126
34504 92.909 -80.026 53.103
34505 92.925 -80.018 53.081
34506 92.941 -80.010 53.059
34507 92.956 -80.002 53.038
34508 92.973 -79.993 53.017
34509 92.989 -79.986 52.996
34510 93.005 -79.977 52.978
34511 93.021 -79.970 52.959
34512 93.037 -79.961 52.942
34513 93.053 -79.953 52.925
34514 93.070 -79.942 52.909
34515 93.088 -79.928 52.894
34516 93.097 -79.923 52.881
34517 93.103 -79.919 52.870
34518 93.111 -79.915 52.860
34519 93.122 -79.909 52.854
34520 93.143 -79.899 52.850
34521 93.165 -79.889 52.846
34522 93.180 -79.881 52.842
34523 93.195 -79.872 52.840
34524 93.219 -79.863 52.838
34525 93.242 -79.853 52.835
34526 93.265 -79.843 52.834
34527 93.288 -79.834 52.831
34528 93.312 -79.824 52.828
34529 93.335 -79.814 52.825
34530 93.358 -79.805 52.825
34531 93.382 -79.795 52.825
34532 93.412 -79.784 52.824
34533 93.447 -79.771 52.824
34534 93.482 -79.758 52.823
34535 93.497 -79.742 52.822
34536 93.512 -79.726 52.818
34537 93.527 -79.710 52.813
34538 93.542 -79.694 52.808
34539 93.557 -79.677 52.804
34540 93.572 -79.661 52.800
34541 93.586 -79.645 52.796
34542 93.602 -79.629 52.789
34543 93.616 -79.614 52.781
34544 93.626 -79.604 52.773
34545 93.636 -79.593 52.768
34546 93.646 -79.583 52.760
34547 93.656 -79.573 52.752
34548 93.665 -79.563 52.744
34549 93.676 -79.554 52.735
34550 93.686 -79.543 52.727
34551 93.695 -79.533 52.718
34552 93.706 -79.524 52.710
34553 93.721 -79.517 52.701
34554 93.736 -79.511 52.692
34555 93.751 -79.504 52.684
34556 93.767 -79.498 52.678
34557 93.782 -79.491 52.673
34558 93.797 -79.485 52.669
34559 93.812 -79.479 52.666
34560 93.827 -79.473 52.662
34561 93.842 -79.466 52.657
34562 93.857 -79.460 52.653
34563 93.872 -79.453 52.648
34564 93.888 -79.447 52.643
34565 93.903 -79.440 52.637
34566 93.918 -79.434 52.631
34567 93.931 -79.429 52.625
34568 93.938 -79.432 52.618
34569 93.944 -79.434 52.610
34570 93.951 -79.437 52.602
34571 93.958 -79.439 52.593
34572 93.965 -79.441 52.584
34573 93.971 -79.444 52.574
34574 93.978 -79.447 52.562
34575 93.984 -79.449 52.551
34576 93.991 -79.451 52.540
34577 93.997 -79.454 52.529
34578 94.004 -79.457 52.518
34579 94.010 -79.459 52.505
34580 94.017 -79.461 52.493
34581 94.024 -79.464 52.480
34582 94.030 -79.466 52.467
34583 94.037 -79.469 52.454
34584 94.043 -79.472 52.440
34585 94.050 -79.474 52.427
34586 94.056 -79.476 52.414
34587 94.059 -79.481 52.401
34588 94.060 -79.487 52.386
34589 94.061 -79.492 52.371
34590 94.062 -79.498 52.357
34591 94.063 -79.504 52.344
34592 94.064 -79.509 52.331
34593 94.065 -79.514 52.316
34594 94.067 -79.520 52.299
34595 94.068 -79.526 52.281
34596 94.069 -79.531 52.265
34597 94.070 -79.536 52.248
34598 94.075 -79.540 52.232
34599 94.084 -79.542 52.219
34600 94.093 -79.544 52.205
34601 94.102 -79.546 52.190
34602 94.112 -79.548 52.174
34603 94.128 -79.507 52.158
34604 94.143 -79.506 52.140
34605 94.149 -79.506 52.123
34606 94.158 -79.496 52.107
34607 94.167 -79.476 52.090
34608 94.175 -79.451 52.073
34609 94.183 -79.426 52.054
34610 94.191 -79.400 52.034
34611 94.199 -79.375 52.014
34612 94.207 -79.350 51.996
34613 94.215 -79.325 51.981
34614 94.223 -79.299 51.966
34615 94.231 -79.274 51.950
34616 94.238 -79.267 51.931
34617 94.245 -79.266 51.911
34618 94.253 -79.263 51.890
34619 94.261 -79.261 51.870
34620 94.269 -79.259 51.853
34621 94.277 -79.256 51.837
34622 94.285 -79.254 51.821
34623 94.294 -79.251 51.802
34624 94.302 -79.249 51.785
34625 94.303 -79.254 51.768
34626 94.300 -79.266 51.748
34627 94.296 -79.276 51.727
34628 94.298 -79.256 51.704
34629 94.289 -79.253 51.681
34630 94.276 -79.259 51.652
34631 94.262 -79.266 51.622
34632 94.249 -79.272 51.591
34633 94.236 -79.278 51.561
34634 94.239 -79.275 51.529
34635 94.249 -79.268 51.495
34636 94.259 -79.262 51.461
34637 94.268 -79.255 51.423
34638 94.278 -79.248 51.385
34639 94.287 -79.282 51.347
34640 94.296 -79.325 51.309
34641 94.305 -79.369 51.272
34642 94.314 -79.412 51.234
34643 94.322 -79.455 51.196
34644 94.331 -79.499 51.157
34645 94.342 -79.525 51.118
34646 94.356 -79.538 51.077
34647 94.369 -79.551 51.037
34648 94.382 -79.563 50.998
34649 94.395 -79.576 50.959
34650 94.408 -79.589 50.919
34651 94.421 -79.602 50.880
34652 94.434 -79.616 50.840
35195 92.744 -80.089 53.372
35196 92.760 -80.081 53.347
35197 92.776 -80.073 53.321
35198 92.792 -80.065 53.296
35199 92.808 -80.057 53.271
35200 92.825 -80.049 53.245
35201 92.840 -80.041 53.221
35202 92.856 -80.033 53.196
35203 92.872 -80.025 53.171
35204 92.888 -80.017 53.147
35205 92.905 -80.008 53.123
35206 92.921 -80.001 53.099
35207 92.936 -79.993 53.076
35208 92.952 -79.985 53.054
35209 92.969 -79.977 53.033
35210 92.985 -79.968 53.011
35211 93.001 -79.961 52.992
35212 93.017 -79.952 52.973
35213 93.032 -79.945 52.953
35214 93.049 -79.936 52.935
35215 93.065 -79.928 52.918
35216 93.082 -79.916 52.901
35217 93.100 -79.902 52.888
35218 93.111 -79.896 52.875
35219 93.117 -79.892 52.864
35220 93.124 -79.888 52.858
35221 93.131 -79.884 52.854
35222 93.146 -79.877 52.851
35223 93.166 -79.868 52.847
35224 93.189 -79.858 52.844
35225 93.213 -79.849 52.841
35226 93.236 -79.839 52.838
35227 93.259 -79.830 52.836
35228 93.282 -79.820 52.833
35229 93.306 -79.810 52.830
35230 93.329 -79.801 52.828
35231 93.352 -79.791 52.827
35232 93.375 -79.781 52.826
35233 93.399 -79.772 52.826
35234 93.434 -79.759 52.826
35235 93.470 -79.746 52.825
35236 93.485 -79.730 52.823
35237 93.500 -79.714 52.818
35238 93.515 -79.698 52.813
35239 93.530 -79.682 52.807
35240 93.545 -79.665 52.802
35241 93.559 -79.649 52.796
35242 93.574 -79.633 52.790
35243 93.589 -79.617 52.782
35244 93.604 -79.601 52.773
35245 93.619 -79.585 52.765
35246 93.634 -79.569 52.758
35247 93.645 -79.557 52.750
35248 93.655 -79.548 52.742
35249 93.665 -79.538 52.734
35250 93.675 -79.527 52.725
35251 93.685 -79.517 52.718
35252 93.695 -79.507 52.710
35253 93.704 -79.498 52.701
35254 93.719 -79.491 52.692
35255 93.734 -79.484 52.684
35256 93.750 -79.478 52.676
35257 93.765 -79.472 52.670
35258 93.780 -79.465 52.665
35259 93.795 -79.459 52.662
35260 93.810 -79.452 52.660
35261 93.825 -79.446 52.657
35262 93.840 -79.439 52.653
35263 93.855 -79.433 52.649
35264 93.870 -79.426 52.646
35265 93.886 -79.420 52.642
35266 93.901 -79.413 52.636
35267 93.916 -79.407 52.630
35268 93.931 -79.400 52.625
35269 93.946 -79.394 52.619
35270 93.958 -79.392 52.612
35271 93.965 -79.394 52.603
35272 93.971 -79.397 52.593
35273 93.978 -79.399 52.584
35274 93.984 -79.402 52.572
35275 93.991 -79.404 52.561
35276 93.997 -79.407 52.549
35277 94.004 -79.410 52.537
35278 94.010 -79.412 52.525
35279 94.013 -79.416 52.513
35280 94.014 -79.422 52.500
35281 94.015 -79.428 52.488
35282 94.016 -79.433 52.476
35283 94.017 -79.438 52.463
35284 94.018 -79.444 52.449
35285 94.019 -79.450 52.436
35286 94.020 -79.455 52.423
35287 94.022 -79.460 52.411
35288 94.023 -79.466 52.397
35289 94.024 -79.472 52.383
35290 94.025 -79.477 52.368
35291 94.026 -79.483 52.354
35292 94.027 -79.488 52.341
35293 94.028 -79.494 52.331
35294 94.029 -79.500 52.316
35295 94.030 -79.505 52.298
35296 94.037 -79.508 52.279
35297 94.046 -79.510 52.262
35298 94.055 -79.512 52.246
35299 94.064 -79.513 52.230
35300 94.073 -79.516 52.218
35301 94.083 -79.517 52.205
35302 94.092 -79.519 52.190
35303 94.128 -79.478 52.173
35304 94.139 -79.432 52.156
35305 94.141 -79.453 52.137
35306 94.145 -79.463 52.119
35307 94.154 -79.453 52.104
35308 94.163 -79.443 52.087
35309 94.171 -79.429 52.070
35310 94.179 -79.403 52.052
35311 94.187 -79.378 52.033
35312 94.195 -79.352 52.013
35313 94.203 -79.327 51.993
35314 94.211 -79.302 51.976
35315 94.219 -79.276 51.962
35316 94.227 -79.257 51.947
35317 94.234 -79.267 51.929
35318 94.239 -79.266 51.910
35319 94.245 -79.266 51.892
35320 94.250 -79.266 51.874
35321 94.255 -79.265 51.858
35322 94.262 -79.263 51.843
35323 94.270 -79.261 51.826
35324 94.278 -79.259 51.807
35325 94.286 -79.256 51.790
35326 94.295 -79.254 51.773
35327 94.297 -79.258 51.755
35328 94.294 -79.269 51.734
35329 94.294 -79.254 51.712
35330 94.281 -79.259 51.687
35331 94.268 -79.266 51.660
35332 94.255 -79.272 51.632
35333 94.242 -79.278 51.603
35334 94.228 -79.284 51.573
35335 94.227 -79.283 51.542
35336 94.238 -79.276 51.510
35337 94.247 -79.269 51.478
35338 94.257 -79.263 51.441
35339 94.267 -79.256 51.403
35340 94.276 -79.260 51.363
35341 94.285 -79.303 51.326
35342 94.294 -79.347 51.287
35343 94.302 -79.390 51.247
35344 94.312 -79.433 51.208
35345 94.321 -79.473 51.168
35346 94.334 -79.485 51.127
35347 94.347 -79.498 51.086
35348 94.360 -79.511 51.047
35349 94.373 -79.524 51.007
35350 94.387 -79.537 50.967
35351 94.400 -79.550 50.928
35352 94.413 -79.563 50.887
35353 94.426 -79.576 50.847
35899 92.788 -80.048 53.318
35900 92.804 -80.040 53.292
35901 92.820 -80.032 53.267
35902 92.836 -80.024 53.242
35903 92.852 -80.016 53.216
35904 92.868 -80.008 53.192
35905 92.884 -80.000 53.166
35906 92.901 -79.992 53.141
35907 92.916 -79.983 53.117
35908 92.932 -79.976 53.094
35909 92.948 -79.968 53.071
35910 92.965 -79.960 53.048
35911 92.981 -79.952 53.027
35912 92.997 -79.943 53.007
35913 93.012 -79.936 52.986
35914 93.028 -79.927 52.965
35915 93.045 -79.920 52.947
35916 93.061 -79.911 52.929
35917 93.077 -79.903 52.911
35918 93.095 -79.889 52.896
35919 93.113 -79.876 52.883
35920 93.124 -79.868 52.871
35921 93.131 -79.864 52.863
35922 93.138 -79.861 52.859
35923 93.144 -79.857 52.856
35924 93.160 -79.851 52.851
35925 93.184 -79.843 52.847
35926 93.207 -79.835 52.843
35927 93.230 -79.825 52.839
35928 93.253 -79.815 52.837
35929 93.276 -79.805 52.834
35930 93.300 -79.796 52.831
35931 93.323 -79.786 52.830
35932 93.346 -79.776 52.829
35933 93.369 -79.767 52.828
35934 93.393 -79.758 52.828
35935 93.421 -79.746 52.828
35936 93.457 -79.734 52.826
35937 93.473 -79.717 52.823
35938 93.488 -79.701 52.819
35939 93.502 -79.686 52.813
35940 93.517 -79.670 52.806
35941 93.533 -79.653 52.799
35942 93.547 -79.637 52.792
35943 93.562 -79.621 52.785
35944 93.577 -79.605 52.777
35945 93.592 -79.588 52.767
35946 93.607 -79.573 52.758
35947 93.622 -79.557 52.751
35948 93.637 -79.541 52.743
35949 93.651 -79.524 52.735
35950 93.664 -79.511 52.727
35951 93.674 -79.501 52.717
35952 93.684 -79.491 52.708
35953 93.694 -79.482 52.699
35954 93.704 -79.472 52.690
35955 93.717 -79.464 52.682
35956 93.732 -79.457 52.674
35957 93.748 -79.451 52.668
35958 93.763 -79.444 52.662
35959 93.778 -79.438 52.658
35960 93.793 -79.432 52.656
35961 93.808 -79.426 52.655
35962 93.823 -79.419 52.653
35963 93.838 -79.413 52.650
35964 93.854 -79.407 52.647
35965 93.869 -79.400 52.645
35966 93.884 -79.394 52.642
35967 93.899 -79.387 52.637
35968 93.914 -79.381 52.631
35969 93.929 -79.374 52.627
35970 93.944 -79.368 52.623
35971 93.960 -79.363 52.616
35972 93.971 -79.360 52.607
35973 93.969 -79.363 52.597
35974 93.970 -79.369 52.586
35975 93.971 -79.374 52.574
35976 93.972 -79.379 52.561
35977 93.973 -79.385 52.548
35978 93.975 -79.391 52.536
35979 93.976 -79.396 52.523
35980 93.977 -79.402 52.510
35981 93.978 -79.407 52.498
35982 93.979 -79.413 52.486
35983 93.980 -79.418 52.473
35984 93.981 -79.424 52.460
35985 93.982 -79.429 52.447
35986 93.983 -79.435 52.434
35987 93.984 -79.441 52.421
35988 93.985 -79.446 52.408
35989 93.986 -79.451 52.394
35990 93.987 -79.457 52.379
35991 93.988 -79.463 52.366
35992 93.989 -79.468 52.353
35993 93.991 -79.474 52.342
35994 93.999 -79.476 52.330
35995 94.008 -79.478 52.315
35996 94.017 -79.479 52.297
35997 94.027 -79.482 52.278
35998 94.036 -79.483 52.260
35999 94.045 -79.485 52.243
36000 94.054 -79.487 52.228
36001 94.063 -79.488 52.216
36002 94.072 -79.491 52.203
36003 94.121 -79.447 52.189
36004 94.099 -79.427 52.171
36005 94.136 -79.379 52.153
36006 94.138 -79.400 52.134
36007 94.141 -79.419 52.114
36008 94.149 -79.409 52.099
36009 94.158 -79.399 52.083
36010 94.167 -79.390 52.067
36011 94.175 -79.380 52.048
36012 94.183 -79.355 52.030
36013 94.191 -79.329 52.010
36014 94.199 -79.304 51.990
36015 94.207 -79.279 51.972
36016 94.215 -79.253 51.957
36017 94.223 -79.253 51.943
36018 94.230 -79.263 51.925
36019 94.236 -79.266 51.909
36020 94.241 -79.266 51.893
36021 94.246 -79.265 51.877
36022 94.251 -79.265 51.863
36023 94.257 -79.264 51.848
36024 94.262 -79.264 51.832
36025 94.267 -79.263 51.812
36026 94.272 -79.263 51.795
36027 94.279 -79.261 51.779
36028 94.287 -79.259 51.761
36029 94.291 -79.261 51.742
36030 94.287 -79.259 51.720
36031 94.274 -79.266 51.696
36032 94.261 -79.272 51.670
36033 94.247 -79.278 51.643
36034 94.234 -79.284 51.615
36035 94.220 -79.290 51.585
36036 94.216 -79.291 51.555
36037 94.226 -79.284 51.523
36038 94.236 -79.278 51.490
36039 94.246 -79.271 51.454
36040 94.255 -79.264 51.417
36041 94.265 -79.257 51.379
36042 94.274 -79.281 51.343
36043 94.283 -79.324 51.304
36044 94.292 -79.368 51.263
36045 94.300 -79.411 51.222
36046 94.313 -79.433 51.181
36047 94.326 -79.446 51.139
36048 94.339 -79.459 51.098
36049 94.352 -79.472 51.057
36050 94.365 -79.485 51.016
36051 94.378 -79.498 50.976
36052 94.392 -79.511 50.936
36053 94.405 -79.524 50.896
36054 94.418 -79.537 50.855
36055 94.431 -79.550 50.815
36602 92.816 -80.015 53.289
36603 92.832 -80.007 53.263
36604 92.848 -79.999 53.237
36605 92.864 -79.991 53.212
36606 92.880 -79.983 53.186
36607 92.896 -79.975 53.160
36608 92.912 -79.967 53.136
36609 92.928 -79.958 53.112
36610 92.944 -79.951 53.089
36611 92.960 -79.943 53.065
36612 92.977 -79.935 53.043
36613 92.992 -79.927 53.022
36614 93.008 -79.918 53.000
36615 93.024 -79.911 52.979
36616 93.041 -79.902 52.959
36617 93.057 -79.895 52.940
36618 93.073 -79.886 52.922
36619 93.089 -79.877 52.906
36620 93.108 -79.863 52.892
36621 93.126 -79.849 52.879
36622 93.138 -79.841 52.870
36623 93.144 -79.837 52.863
36624 93.151 -79.833 52.857
36625 93.158 -79.830 52.852
36626 93.179 -79.823 52.848
36627 93.202 -79.814 52.844
36628 93.224 -79.807 52.840
36629 93.248 -79.798 52.837
36630 93.271 -79.791 52.835
36631 93.294 -79.782 52.832
36632 93.317 -79.772 52.830
36633 93.340 -79.763 52.830
36634 93.363 -79.753 52.830
36635 93.387 -79.743 52.830
36636 93.410 -79.734 52.829
36637 93.444 -79.721 52.827
36638 93.461 -79.705 52.824
36639 93.476 -79.689 52.820
36640 93.490 -79.673 52.813
36641 93.505 -79.657 52.805
36642 93.520 -79.641 52.797
36643 93.535 -79.625 52.789
36644 93.550 -79.609 52.781
36645 93.565 -79.593 52.772
36646 93.580 -79.576 52.762
36647 93.594 -79.560 52.753
36648 93.610 -79.545 52.744
36649 93.625 -79.529 52.737
36650 93.639 -79.512 52.729
36651 93.654 -79.496 52.721
36652 93.669 -79.480 52.711
36653 93.683 -79.466 52.702
36654 93.693 -79.456 52.693
36655 93.703 -79.445 52.684
36656 93.715 -79.438 52.676
36657 93.730 -79.431 52.669
36658 93.746 -79.425 52.662
36659 93.761 -79.418 52.656
36660 93.776 -79.412 52.652
36661 93.791 -79.405 52.650
36662 93.806 -79.399 52.651
36663 93.822 -79.392 52.650
36664 93.837 -79.386 52.650
36665 93.852 -79.380 52.647
36666 93.866 -79.378 52.645
36667 93.882 -79.375 52.643
36668 93.896 -79.372 52.640
36669 93.911 -79.369 52.636
36670 93.926 -79.366 52.633
36671 93.941 -79.363 52.630
36672 93.946 -79.360 52.623
36673 93.951 -79.357 52.614
36674 93.946 -79.356 52.603
36675 93.938 -79.357 52.591
36676 93.935 -79.359 52.579
36677 93.936 -79.365 52.565
36678 93.937 -79.370 52.550
36679 93.938 -79.375 52.536
36680 93.939 -79.382 52.523
36681 93.940 -79.387 52.509
36682 93.941 -79.392 52.498
36683 93.942 -79.398 52.485
36684 93.943 -79.404 52.472
36685 93.944 -79.409 52.459
36686 93.945 -79.415 52.447
36687 93.946 -79.420 52.433
36688 93.947 -79.426 52.420
36689 93.949 -79.432 52.407
36690 93.950 -79.437 52.393
36691 93.952 -79.442 52.379
36692 93.961 -79.444 52.366
36693 93.970 -79.446 52.354
36694 93.979 -79.447 52.342
36695 93.988 -79.449 52.329
36696 93.998 -79.451 52.313
36697 94.007 -79.453 52.296
36698 94.016 -79.455 52.277
36699 94.025 -79.457 52.258
36700 94.035 -79.458 52.241
36701 94.044 -79.460 52.225
36702 94.053 -79.462 52.212
36703 94.096 -79.425 52.199
36704 94.107 -79.390 52.184
36705 94.109 -79.345 52.167
36706 94.134 -79.326 52.148
36707 94.136 -79.347 52.129
36708 94.138 -79.369 52.110
36709 94.145 -79.366 52.093
36710 94.153 -79.356 52.077
36711 94.162 -79.346 52.061
36712 94.171 -79.336 52.042
36713 94.179 -79.326 52.024
36714 94.187 -79.306 52.006
36715 94.195 -79.281 51.987
36716 94.203 -79.256 51.968
36717 94.211 -79.238 51.951
36718 94.218 -79.249 51.937
36719 94.225 -79.259 51.921
36720 94.232 -79.266 51.906
36721 94.237 -79.265 51.893
36722 94.243 -79.265 51.880
36723 94.248 -79.264 51.867
36724 94.253 -79.264 51.854
36725 94.258 -79.263 51.838
36726 94.263 -79.263 51.818
36727 94.269 -79.263 51.801
36728 94.274 -79.262 51.786
36729 94.279 -79.262 51.767
36730 94.284 -79.261 51.749
36731 94.285 -79.261 51.728
36732 94.267 -79.271 51.705
36733 94.253 -79.277 51.679
36734 94.240 -79.283 51.653
36735 94.226 -79.289 51.626
36736 94.213 -79.295 51.597
36737 94.205 -79.298 51.566
36738 94.214 -79.292 51.534
36739 94.224 -79.285 51.501
36740 94.234 -79.278 51.464
36741 94.244 -79.272 51.429
36742 94.254 -79.266 51.394
36743 94.263 -79.259 51.359
36744 94.272 -79.302 51.320
36745 94.281 -79.345 51.278
36746 94.291 -79.381 51.235
36747 94.304 -79.394 51.193
36748 94.317 -79.407 51.151
36749 94.330 -79.419 51.109
36750 94.344 -79.432 51.068
36751 94.357 -79.446 51.026
36752 94.370 -79.459 50.985
36753 94.383 -79.472 50.945
36754 94.396 -79.485 50.904
36755 94.409 -79.498 50.864
36756 94.422 -79.510 50.824
37308 92.892 -79.958 53.180
37309 92.908 -79.950 53.156
37310 92.924 -79.942 53.131
37311 92.940 -79.933 53.106
37312 92.956 -79.926 53.083
37313 92.972 -79.918 53.060
37314 92.988 -79.910 53.037
37315 93.004 -79.902 53.015
37316 93.020 -79.893 52.993
37317 93.037 -79.886 52.973
37318 93.053 -79.877 52.953
37319 93.068 -79.870 52.935
37320 93.084 -79.861 52.918
37321 93.101 -79.851 52.902
37322 93.120 -79.837 52.889
37323 93.138 -79.823 52.877
37324 93.151 -79.814 52.868
37325 93.158 -79.810 52.860
37326 93.165 -79.806 52.854
37327 93.174 -79.802 52.850
37328 93.197 -79.794 52.847
37329 93.220 -79.786 52.843
37330 93.242 -79.778 52.839
37331 93.266 -79.770 52.836
37332 93.289 -79.762 52.834
37333 93.311 -79.755 52.834
37334 93.335 -79.746 52.834
37335 93.358 -79.739 52.834
37336 93.381 -79.730 52.834
37337 93.404 -79.720 52.832
37338 93.431 -79.709 52.830
37339 93.448 -79.693 52.826
37340 93.463 -79.677 52.821
37341 93.478 -79.661 52.814
37342 93.493 -79.645 52.806
37343 93.508 -79.629 52.796
37344 93.523 -79.613 52.787
37345 93.538 -79.597 52.778
37346 93.553 -79.581 52.768
37347 93.568 -79.564 52.758
37348 93.582 -79.548 52.748
37349 93.597 -79.532 52.740
37350 93.612 -79.516 52.732
37351 93.627 -79.500 52.725
37352 93.642 -79.484 52.717
37353 93.657 -79.468 52.708
37354 93.672 -79.452 52.699
37355 93.687 -79.435 52.690
37356 93.702 -79.419 52.681
37357 93.713 -79.411 52.673
37358 93.728 -79.404 52.665
37359 93.744 -79.398 52.658
37360 93.759 -79.395 52.652
37361 93.773 -79.393 52.649
37362 93.788 -79.390 52.647
37363 93.803 -79.387 52.648
37364 93.818 -79.385 52.649
37365 93.833 -79.382 52.649
37366 93.848 -79.379 52.647
37367 93.862 -79.375 52.645
37368 93.877 -79.373 52.643
37369 93.892 -79.370 52.641
37370 93.907 -79.367 52.638
37371 93.918 -79.364 52.636
37372 93.922 -79.360 52.634
37373 93.926 -79.357 52.628
37374 93.931 -79.353 52.619
37375 93.929 -79.352 52.609
37376 93.921 -79.352 52.597
37377 93.913 -79.353 52.585
37378 93.905 -79.353 52.571
37379 93.901 -79.356 52.555
37380 93.902 -79.361 52.539
37381 93.903 -79.366 52.525
37382 93.904 -79.372 52.512
37383 93.905 -79.378 52.499
37384 93.906 -79.383 52.487
37385 93.907 -79.389 52.473
37386 93.908 -79.394 52.460
37387 93.909 -79.400 52.447
37388 93.910 -79.406 52.434
37389 93.914 -79.410 52.420
37390 93.923 -79.412 52.408
37391 93.932 -79.413 52.394
37392 93.941 -79.416 52.381
37393 93.951 -79.417 52.368
37394 93.960 -79.419 52.356
37395 93.969 -79.421 52.342
37396 93.978 -79.422 52.327
37397 93.987 -79.425 52.311
37398 93.996 -79.426 52.294
37399 94.006 -79.428 52.275
37400 94.015 -79.430 52.256
37401 94.024 -79.432 52.238
37402 94.033 -79.434 52.221
37403 94.071 -79.402 52.207
37404 94.115 -79.353 52.193
37405 94.084 -79.346 52.178
37406 94.085 -79.346 52.161
37407 94.116 -79.328 52.142
37408 94.133 -79.295 52.124
37409 94.135 -79.316 52.106
37410 94.140 -79.322 52.088
37411 94.149 -79.313 52.072
37412 94.157 -79.303 52.056
37413 94.166 -79.293 52.037
37414 94.175 -79.283 52.019
37415 94.183 -79.273 52.002
37416 94.191 -79.259 51.984
37417 94.200 -79.234 51.964
37418 94.207 -79.234 51.947
37419 94.214 -79.244 51.932
37420 94.221 -79.255 51.917
37421 94.228 -79.265 51.904
37422 94.234 -79.265 51.892
37423 94.239 -79.264 51.881
37424 94.244 -79.264 51.868
37425 94.249 -79.263 51.855
37426 94.254 -79.263 51.841
37427 94.260 -79.263 51.822
37428 94.265 -79.262 51.805
37429 94.270 -79.262 51.790
37430 94.275 -79.261 51.772
37431 94.281 -79.259 51.754
37432 94.289 -79.254 51.733
37433 94.297 -79.250 51.711
37434 94.278 -79.259 51.686
37435 94.260 -79.269 51.660
37436 94.242 -79.278 51.633
37437 94.223 -79.288 51.605
37438 94.204 -79.298 51.575
37439 94.206 -79.297 51.543
37440 94.213 -79.293 51.510
37441 94.222 -79.287 51.474
37442 94.232 -79.280 51.439
37443 94.242 -79.273 51.407
37444 94.252 -79.267 51.372
37445 94.261 -79.280 51.333
37446 94.270 -79.323 51.291
37447 94.282 -79.341 51.247
37448 94.295 -79.354 51.204
37449 94.309 -79.367 51.162
37450 94.322 -79.380 51.120
37451 94.335 -79.394 51.078
37452 94.348 -79.407 51.036
37453 94.361 -79.419 50.995
37454 94.374 -79.432 50.954
37455 94.388 -79.445 50.913
37456 94.401 -79.458 50.873
37457 94.414 -79.471 50.833
38013 92.952 -79.908 53.102
38014 92.968 -79.901 53.078
38015 92.984 -79.892 53.054
38016 93.000 -79.885 53.031
38017 93.016 -79.877 53.009
38018 93.032 -79.868 52.989
38019 93.048 -79.861 52.969
38020 93.064 -79.852 52.948
38021 93.080 -79.845 52.930
38022 93.096 -79.836 52.914
38023 93.114 -79.824 52.899
38024 93.132 -79.811 52.886
38025 93.150 -79.797 52.875
38026 93.165 -79.786 52.865
38027 93.171 -79.783 52.858
38028 93.179 -79.779 52.853
38029 93.192 -79.773 52.850
38030 93.215 -79.766 52.847
38031 93.238 -79.758 52.843
38032 93.261 -79.750 52.840
38033 93.284 -79.742 52.838
38034 93.307 -79.734 52.838
38035 93.330 -79.726 52.839
38036 93.353 -79.718 52.839
38037 93.376 -79.710 52.838
38038 93.399 -79.702 52.836
38039 93.422 -79.694 52.833
38040 93.434 -79.685 52.829
38041 93.448 -79.669 52.823
38042 93.463 -79.654 52.816
38043 93.477 -79.638 52.808
38044 93.492 -79.622 52.797
38045 93.506 -79.606 52.787
38046 93.521 -79.590 52.777
38047 93.536 -79.575 52.767
38048 93.550 -79.559 52.757
38049 93.565 -79.543 52.747
38050 93.579 -79.527 52.740
38051 93.594 -79.511 52.734
38052 93.609 -79.496 52.727
38053 93.623 -79.480 52.720
38054 93.638 -79.464 52.710
38055 93.652 -79.448 52.701
38056 93.667 -79.432 52.692
38057 93.682 -79.417 52.683
38058 93.707 -79.404 52.673
38059 93.725 -79.399 52.664
38060 93.740 -79.397 52.656
38061 93.755 -79.394 52.650
38062 93.769 -79.391 52.646
38063 93.784 -79.388 52.644
38064 93.799 -79.385 52.646
38065 93.814 -79.382 52.648
38066 93.829 -79.380 52.648
38067 93.844 -79.377 52.647
38068 93.858 -79.374 52.645
38069 93.873 -79.372 52.642
38070 93.888 -79.369 52.640
38071 93.893 -79.365 52.639
38072 93.898 -79.361 52.638
38073 93.902 -79.358 52.636
38074 93.907 -79.354 52.630
38075 93.911 -79.350 52.622
38076 93.912 -79.348 52.613
38077 93.904 -79.348 52.601
38078 93.896 -79.349 52.588
38079 93.888 -79.350 52.574
38080 93.880 -79.350 52.558
38081 93.872 -79.350 52.543
38082 93.866 -79.352 52.529
38083 93.867 -79.357 52.516
38084 93.868 -79.363 52.504
38085 93.869 -79.369 52.491
38086 93.870 -79.374 52.478
38087 93.875 -79.378 52.466
38088 93.885 -79.379 52.453
38089 93.894 -79.382 52.439
38090 93.903 -79.383 52.425
38091 93.912 -79.385 52.412
38092 93.922 -79.387 52.398
38093 93.931 -79.388 52.383
38094 93.940 -79.391 52.369
38095 93.949 -79.392 52.355
38096 93.959 -79.394 52.340
38097 93.968 -79.396 52.323
38098 93.977 -79.397 52.306
38099 93.986 -79.400 52.288
38100 93.995 -79.401 52.270
38101 94.004 -79.404 52.252
38102 94.014 -79.405 52.234
38103 94.047 -79.379 52.217
38104 94.101 -79.328 52.200
38105 94.109 -79.325 52.186
38106 94.094 -79.332 52.170
38107 94.080 -79.338 52.152
38108 94.094 -79.334 52.135
38109 94.117 -79.293 52.117
38110 94.133 -79.263 52.100
38111 94.136 -79.278 52.084
38112 94.144 -79.269 52.068
38113 94.153 -79.259 52.051
38114 94.161 -79.249 52.032
38115 94.170 -79.239 52.014
38116 94.179 -79.229 51.997
38117 94.187 -79.219 51.980
38118 94.195 -79.219 51.960
38119 94.202 -79.230 51.943
38120 94.209 -79.240 51.928
38121 94.217 -79.250 51.914
38122 94.224 -79.260 51.902
38123 94.229 -79.264 51.891
38124 94.235 -79.264 51.881
38125 94.240 -79.263 51.868
38126 94.246 -79.263 51.855
38127 94.251 -79.263 51.842
38128 94.256 -79.262 51.824
38129 94.261 -79.262 51.808
38130 94.266 -79.261 51.792
38131 94.272 -79.260 51.774
38132 94.280 -79.255 51.755
38133 94.287 -79.250 51.735
38134 94.295 -79.245 51.713
38135 94.303 -79.241 51.689
38136 94.290 -79.247 51.663
38137 94.272 -79.257 51.636
38138 94.253 -79.267 51.609
38139 94.235 -79.276 51.579
38140 94.226 -79.281 51.548
38141 94.229 -79.280 51.516
38142 94.231 -79.278 51.482
38143 94.234 -79.278 51.448
38144 94.237 -79.276 51.417
38145 94.241 -79.275 51.382
38146 94.250 -79.268 51.343
38147 94.261 -79.289 51.302
38148 94.274 -79.302 51.259
38149 94.287 -79.315 51.215
38150 94.300 -79.328 51.172
38151 94.314 -79.341 51.131
38152 94.327 -79.354 51.089
38153 94.340 -79.367 51.047
38154 94.353 -79.380 51.005
38155 94.366 -79.393 50.964
38156 94.380 -79.406 50.922
38157 94.393 -79.419 50.881
38158 94.406 -79.432 50.841
38716 92.980 -79.876 53.072
38717 92.996 -79.867 53.049
38718 93.012 -79.860 53.027
38719 93.028 -79.852 53.006
38720 93.044 -79.843 52.985
38721 93.060 -79.836 52.965
38722 93.076 -79.827 52.946
38723 93.092 -79.820 52.929
38724 93.109 -79.811 52.913
38725 93.126 -79.798 52.899
38726 93.144 -79.785 52.887
38727 93.162 -79.770 52.875
38728 93.179 -79.759 52.866
38729 93.185 -79.755 52.859
38730 93.192 -79.751 52.853
38731 93.210 -79.745 52.851
38732 93.233 -79.737 52.848
38733 93.256 -79.729 52.846
38734 93.279 -79.721 52.844
38735 93.302 -79.714 52.844
38736 93.325 -79.705 52.844
38737 93.348 -79.698 52.844
38738 93.371 -79.689 52.842
38739 93.394 -79.682 52.840
38740 93.416 -79.675 52.837
38741 93.436 -79.667 52.833
38742 93.444 -79.657 52.826
38743 93.452 -79.645 52.818
38744 93.461 -79.635 52.810
38745 93.469 -79.623 52.799
38746 93.482 -79.608 52.789
38747 93.497 -79.592 52.778
38748 93.512 -79.576 52.768
38749 93.526 -79.561 52.758
38750 93.541 -79.545 52.748
38751 93.556 -79.529 52.742
38752 93.570 -79.513 52.737
38753 93.585 -79.498 52.731
38754 93.600 -79.482 52.724
38755 93.614 -79.466 52.715
38756 93.629 -79.451 52.706
38757 93.643 -79.435 52.696
38758 93.665 -79.421 52.687
38759 93.693 -79.410 52.677
38760 93.721 -79.397 52.667
38761 93.736 -79.395 52.657
38762 93.751 -79.392 52.651
38763 93.766 -79.389 52.645
38764 93.780 -79.387 52.641
38765 93.795 -79.384 52.643
38766 93.810 -79.381 52.645
38767 93.825 -79.379 52.645
38768 93.840 -79.375 52.643
38769 93.855 -79.372 52.642
38770 93.864 -79.369 52.641
38771 93.868 -79.366 52.638
38772 93.873 -79.362 52.637
38773 93.877 -79.358 52.636
38774 93.882 -79.355 52.636
38775 93.887 -79.351 52.630
38776 93.891 -79.347 52.623
38777 93.894 -79.344 52.614
38778 93.886 -79.344 52.602
38779 93.879 -79.345 52.589
38780 93.871 -79.345 52.576
38781 93.864 -79.346 52.560
38782 93.857 -79.346 52.548
38783 93.849 -79.346 52.535
38784 93.842 -79.347 52.522
38785 93.840 -79.346 52.511
38786 93.848 -79.347 52.498
38787 93.856 -79.349 52.485
38788 93.865 -79.351 52.473
38789 93.874 -79.353 52.460
38790 93.884 -79.355 52.445
38791 93.893 -79.357 52.430
38792 93.902 -79.358 52.415
38793 93.911 -79.360 52.400
38794 93.921 -79.362 52.384
38795 93.930 -79.364 52.368
38796 93.939 -79.366 52.353
38797 93.948 -79.367 52.337
38798 93.958 -79.369 52.320
38799 93.967 -79.371 52.303
38800 93.976 -79.373 52.285
38801 93.985 -79.375 52.267
38802 93.994 -79.376 52.250
38803 94.023 -79.357 52.232
38804 94.053 -79.338 52.215
38805 94.082 -79.335 52.198
38806 94.087 -79.318 52.181
38807 94.087 -79.319 52.164
38808 94.090 -79.325 52.146
38809 94.076 -79.332 52.128
38810 94.096 -79.303 52.111
38811 94.117 -79.256 52.093
38812 94.132 -79.232 52.077
38813 94.140 -79.225 52.061
38814 94.148 -79.216 52.044
38815 94.157 -79.212 52.025
38816 94.165 -79.222 52.007
38817 94.174 -79.231 51.989
38818 94.180 -79.228 51.973
38819 94.184 -79.222 51.955
38820 94.198 -79.225 51.939
38821 94.205 -79.236 51.924
38822 94.212 -79.246 51.912
38823 94.219 -79.256 51.900
38824 94.226 -79.264 51.890
38825 94.231 -79.263 51.881
38826 94.237 -79.263 51.869
38827 94.242 -79.263 51.855
38828 94.247 -79.262 51.841
38829 94.252 -79.262 51.825
38830 94.258 -79.261 51.809
38831 94.263 -79.261 51.793
38832 94.270 -79.256 51.774
38833 94.278 -79.252 51.754
38834 94.286 -79.247 51.734
38835 94.293 -79.241 51.713
38836 94.301 -79.237 51.690
38837 94.310 -79.231 51.665
38838 94.302 -79.235 51.638
38839 94.283 -79.245 51.610
38840 94.265 -79.255 51.581
38841 94.247 -79.265 51.551
38842 94.249 -79.263 51.520
38843 94.252 -79.263 51.487
38844 94.255 -79.261 51.455
38845 94.257 -79.260 51.424
38846 94.260 -79.259 51.389
38847 94.258 -79.262 51.351
38848 94.260 -79.259 51.311
38849 94.269 -79.266 51.269
38850 94.279 -79.275 51.226
38851 94.292 -79.289 51.182
38852 94.305 -79.302 51.141
38853 94.318 -79.315 51.099
38854 94.331 -79.328 51.057
38855 94.345 -79.341 51.015
38856 94.358 -79.353 50.973
38857 94.371 -79.366 50.932
38858 94.384 -79.380 50.890
38859 94.397 -79.393 50.849
39418 92.992 -79.851 53.069
39419 93.008 -79.842 53.046
39420 93.024 -79.835 53.024
39421 93.040 -79.827 53.003
39422 93.056 -79.818 52.982
39423 93.072 -79.811 52.963
39424 93.088 -79.802 52.945
39425 93.105 -79.795 52.929
39426 93.121 -79.786 52.914
39427 93.139 -79.772 52.900
39428 93.157 -79.758 52.888
39429 93.175 -79.745 52.877
39430 93.192 -79.732 52.868
39431 93.199 -79.728 52.862
39432 93.206 -79.724 52.858
39433 93.228 -79.717 52.856
39434 93.251 -79.708 52.854
39435 93.274 -79.701 52.852
39436 93.297 -79.692 52.850
39437 93.320 -79.685 52.848
39438 93.343 -79.676 52.847
39439 93.366 -79.669 52.845
39440 93.389 -79.661 52.842
39441 93.411 -79.655 52.838
39442 93.433 -79.648 52.832
39443 93.447 -79.640 52.826
39444 93.455 -79.629 52.819
39445 93.463 -79.618 52.811
39446 93.471 -79.607 52.801
39447 93.479 -79.596 52.791
39448 93.487 -79.585 52.780
39449 93.495 -79.574 52.771
39450 93.503 -79.563 52.760
39451 93.517 -79.548 52.750
39452 93.532 -79.532 52.745
39453 93.546 -79.516 52.741
39454 93.561 -79.500 52.736
39455 93.575 -79.485 52.729
39456 93.590 -79.469 52.720
39457 93.605 -79.453 52.711
39458 93.623 -79.438 52.701
39459 93.651 -79.426 52.691
39460 93.679 -79.414 52.681
39461 93.707 -79.403 52.671
39462 93.732 -79.393 52.660
39463 93.747 -79.391 52.652
39464 93.762 -79.388 52.646
39465 93.776 -79.385 52.640
39466 93.791 -79.382 52.641
39467 93.806 -79.379 52.642
39468 93.821 -79.376 52.642
39469 93.835 -79.374 52.640
39470 93.839 -79.370 52.639
39471 93.844 -79.366 52.638
39472 93.848 -79.363 52.637
39473 93.853 -79.359 52.636
39474 93.857 -79.356 52.635
39475 93.865 -79.350 52.634
39476 93.877 -79.343 52.629
39477 93.890 -79.335 52.622
39478 93.896 -79.332 52.613
39479 93.895 -79.333 52.601
39480 93.894 -79.334 52.590
39481 93.888 -79.335 52.577
39482 93.880 -79.335 52.563
39483 93.873 -79.335 52.552
39484 93.865 -79.335 52.541
39485 93.858 -79.335 52.531
39486 93.860 -79.337 52.519
39487 93.863 -79.339 52.507
39488 93.870 -79.336 52.494
39489 93.878 -79.332 52.481
39490 93.887 -79.330 52.467
39491 93.894 -79.332 52.452
39492 93.902 -79.333 52.435
39493 93.910 -79.335 52.418
39494 93.918 -79.337 52.401
39495 93.925 -79.338 52.383
39496 93.933 -79.340 52.366
39497 93.941 -79.341 52.350
39498 93.949 -79.343 52.334
39499 93.957 -79.344 52.318
39500 93.965 -79.346 52.301
39501 93.972 -79.347 52.283
39502 93.980 -79.349 52.264
39503 93.987 -79.355 52.248
39504 94.015 -79.351 52.230
39505 94.044 -79.347 52.213
39506 94.073 -79.344 52.196
39507 94.068 -79.321 52.177
39508 94.059 -79.306 52.159
39509 94.062 -79.313 52.139
39510 94.065 -79.319 52.122
39511 94.078 -79.320 52.104
39512 94.100 -79.275 52.086
39513 94.122 -79.229 52.070
39514 94.135 -79.218 52.054
39515 94.143 -79.228 52.037
39516 94.151 -79.237 52.018
39517 94.159 -79.246 52.000
39518 94.166 -79.237 51.982
39519 94.173 -79.229 51.967
39520 94.175 -79.226 51.950
39521 94.185 -79.227 51.935
39522 94.201 -79.231 51.921
39523 94.208 -79.242 51.910
39524 94.215 -79.252 51.900
39525 94.222 -79.263 51.890
39526 94.227 -79.263 51.880
39527 94.232 -79.263 51.868
39528 94.238 -79.262 51.855
39529 94.243 -79.262 51.842
39530 94.249 -79.261 51.826
39531 94.254 -79.261 51.809
39532 94.260 -79.258 51.792
39533 94.268 -79.253 51.773
39534 94.276 -79.248 51.752
39535 94.284 -79.243 51.731
39536 94.292 -79.238 51.711
39537 94.299 -79.233 51.688
39538 94.308 -79.228 51.663
39539 94.316 -79.223 51.637
39540 94.314 -79.224 51.609
39541 94.295 -79.234 51.580
39542 94.276 -79.243 51.549
39543 94.269 -79.247 51.520
39544 94.272 -79.246 51.490
39545 94.275 -79.244 51.459
39546 94.277 -79.244 51.428
39547 94.280 -79.242 51.394
39548 94.254 -79.263 51.357
39549 94.262 -79.253 51.318
39550 94.277 -79.247 51.278
39551 94.285 -79.253 51.236
39552 94.294 -79.261 51.193
39553 94.302 -79.269 51.150
39554 94.311 -79.276 51.108
39555 94.323 -79.288 51.066
39556 94.336 -79.301 51.024
39557 94.349 -79.314 50.983
39558 94.362 -79.328 50.941
39559 94.376 -79.341 50.899
39560 94.389 -79.353 50.857
39561 94.402 -79.366 50.817
40120 93.004 -79.826 53.066
40121 93.020 -79.817 53.043
40122 93.036 -79.810 53.022
40123 93.052 -79.802 53.000
40124 93.068 -79.793 52.980
40125 93.084 -79.786 52.962
40126 93.100 -79.777 52.946
40127 93.116 -79.770 52.929
40128 93.133 -79.760 52.915
40129 93.151 -79.746 52.902
40130 93.169 -79.732 52.890
40131 93.187 -79.718 52.880
40132 93.206 -79.704 52.872
40133 93.212 -79.701 52.868
40134 93.223 -79.696 52.864
40135 93.246 -79.688 52.862
40136 93.269 -79.680 52.859
40137 93.292 -79.672 52.857
40138 93.315 -79.664 52.854
40139 93.338 -79.656 52.851
40140 93.361 -79.648 52.848
40141 93.383 -79.642 52.844
40142 93.405 -79.636 52.839
40143 93.427 -79.629 52.834
40144 93.450 -79.623 52.827
40145 93.458 -79.612 52.820
40146 93.466 -79.601 52.812
40147 93.474 -79.590 52.803
40148 93.482 -79.579 52.794
40149 93.490 -79.568 52.783
40150 93.498 -79.557 52.774
40151 93.506 -79.546 52.765
40152 93.514 -79.535 52.757
40153 93.521 -79.524 52.751
40154 93.530 -79.513 52.746
40155 93.538 -79.502 52.741
40156 93.552 -79.486 52.735
40157 93.566 -79.471 52.726
40158 93.581 -79.455 52.716
40159 93.609 -79.443 52.705
40160 93.637 -79.432 52.695
40161 93.665 -79.419 52.685
40162 93.693 -79.408 52.674
40163 93.721 -79.396 52.663
40164 93.743 -79.389 52.654
40165 93.758 -79.386 52.647
40166 93.772 -79.383 52.642
40167 93.787 -79.381 52.639
40168 93.802 -79.378 52.639
40169 93.810 -79.375 52.639
40170 93.815 -79.371 52.637
40171 93.819 -79.367 52.637
40172 93.824 -79.363 52.636
40173 93.835 -79.357 52.636
40174 93.847 -79.349 52.634
40175 93.860 -79.342 52.633
40176 93.872 -79.335 52.633
40177 93.885 -79.327 52.628
40178 93.897 -79.319 52.621
40179 93.898 -79.319 52.612
40180 93.896 -79.320 52.601
40181 93.895 -79.322 52.590
40182 93.894 -79.322 52.579
40183 93.893 -79.323 52.567
40184 93.889 -79.324 52.556
40185 93.882 -79.325 52.546
40186 93.877 -79.325 52.536
40187 93.880 -79.327 52.524
40188 93.884 -79.328 52.511
40189 93.888 -79.330 52.497
40190 93.893 -79.329 52.483
40191 93.901 -79.325 52.468
40192 93.909 -79.322 52.452
40193 93.917 -79.319 52.434
40194 93.925 -79.315 52.416
40195 93.933 -79.314 52.399
40196 93.941 -79.316 52.381
40197 93.948 -79.317 52.365
40198 93.957 -79.319 52.349
40199 93.964 -79.321 52.332
40200 93.972 -79.322 52.316
40201 93.980 -79.324 52.299
40202 93.988 -79.325 52.281
40203 94.000 -79.329 52.262
40204 94.006 -79.342 52.246
40205 94.006 -79.360 52.229
40206 94.035 -79.357 52.211
40207 94.063 -79.353 52.194
40208 94.048 -79.325 52.175
40209 94.031 -79.294 52.155
40210 94.034 -79.300 52.134
40211 94.051 -79.301 52.115
40212 94.071 -79.301 52.097
40213 94.091 -79.300 52.079
40214 94.113 -79.269 52.064
40215 94.129 -79.243 52.047
40216 94.137 -79.253 52.030
40217 94.145 -79.256 52.012
40218 94.151 -79.247 51.993
40219 94.158 -79.239 51.975
40220 94.164 -79.232 51.961
40221 94.165 -79.230 51.946
40222 94.172 -79.229 51.933
40223 94.189 -79.232 51.921
40224 94.203 -79.237 51.911
40225 94.210 -79.247 51.901
40226 94.218 -79.258 51.890
40227 94.224 -79.263 51.879
40228 94.229 -79.262 51.867
40229 94.235 -79.262 51.855
40230 94.240 -79.261 51.842
40231 94.245 -79.261 51.826
40232 94.251 -79.259 51.809
40233 94.259 -79.254 51.791
40234 94.266 -79.249 51.772
40235 94.274 -79.244 51.752
40236 94.282 -79.239 51.732
40237 94.290 -79.234 51.712
40238 94.298 -79.229 51.688
40239 94.306 -79.225 51.663
40240 94.314 -79.219 51.637
40241 94.322 -79.215 51.609
40242 94.325 -79.212 51.581
40243 94.307 -79.222 51.552
40244 94.290 -79.231 51.524
40245 94.292 -79.229 51.494
40246 94.295 -79.228 51.464
40247 94.297 -79.227 51.432
40248 94.266 -79.257 51.397
40249 94.255 -79.259 51.361
40250 94.260 -79.251 51.324
40251 94.282 -79.238 51.285
40252 94.294 -79.234 51.244
40253 94.302 -79.241 51.203
40254 94.311 -79.248 51.161
40255 94.319 -79.256 51.118
40256 94.327 -79.263 51.075
40257 94.335 -79.271 51.034
40258 94.344 -79.278 50.992
40259 94.354 -79.288 50.949
40260 94.367 -79.301 50.907
40261 94.381 -79.314 50.866
40262 94.394 -79.327 50.825
40822 93.016 -79.801 53.063
40823 93.032 -79.792 53.041
40824 93.048 -79.785 53.020
40825 93.064 -79.776 52.999
40826 93.080 -79.768 52.980
40827 93.096 -79.761 52.963
40828 93.112 -79.752 52.947
40829 93.128 -79.745 52.932
40830 93.145 -79.733 52.918
40831 93.163 -79.720 52.906
40832 93.182 -79.706 52.895
40833 93.200 -79.692 52.885
40834 93.218 -79.678 52.880
40835 93.226 -79.673 52.875
40836 93.241 -79.667 52.871
40837 93.264 -79.660 52.866
40838 93.287 -79.651 52.863
40839 93.310 -79.644 52.861
40840 93.333 -79.636 52.857
40841 93.356 -79.629 52.853
40842 93.378 -79.623 52.847
40843 93.400 -79.616 52.841
40844 93.422 -79.610 52.836
40845 93.444 -79.604 52.829
40846 93.461 -79.595 52.821
40847 93.469 -79.584 52.814
40848 93.477 -79.573 52.805
40849 93.485 -79.562 52.796
40850 93.493 -79.551 52.786
40851 93.500 -79.540 52.778
40852 93.508 -79.529 52.769
40853 93.516 -79.518 52.762
40854 93.524 -79.507 52.755
40855 93.533 -79.496 52.749
40856 93.540 -79.485 52.744
40857 93.549 -79.475 52.739
40858 93.559 -79.467 52.730
40859 93.572 -79.463 52.720
40860 93.596 -79.450 52.708
40861 93.623 -79.436 52.696
40862 93.651 -79.425 52.685
40863 93.679 -79.413 52.674
40864 93.707 -79.401 52.665
40865 93.735 -79.389 52.657
40866 93.754 -79.385 52.651
40867 93.768 -79.382 52.646
40868 93.781 -79.379 52.641
40869 93.786 -79.375 52.638
40870 93.791 -79.371 52.637
40871 93.804 -79.363 52.635
40872 93.817 -79.356 52.635
40873 93.829 -79.348 52.635
40874 93.842 -79.341 52.634
40875 93.854 -79.334 52.634
40876 93.867 -79.326 52.633
40877 93.880 -79.319 52.632
40878 93.892 -79.311 52.627
40879 93.901 -79.306 52.621
40880 93.899 -79.307 52.612
40881 93.898 -79.308 52.601
40882 93.897 -79.309 52.591
40883 93.895 -79.310 52.581
40884 93.894 -79.311 52.570
40885 93.893 -79.312 52.560
40886 93.890 -79.315 52.550
40887 93.896 -79.316 52.540
40888 93.902 -79.316 52.528
40889 93.905 -79.319 52.514
40890 93.908 -79.320 52.498
40891 93.912 -79.322 52.484
40892 93.916 -79.323 52.468
40893 93.924 -79.319 52.451
40894 93.932 -79.316 52.432
40895 93.940 -79.312 52.416
40896 93.948 -79.308 52.399
40897 93.956 -79.304 52.381
40898 93.964 -79.300 52.366
40899 93.972 -79.297 52.349
40900 93.980 -79.298 52.332
40901 93.987 -79.300 52.315
40902 93.995 -79.302 52.297
40903 94.009 -79.307 52.280
40904 94.025 -79.313 52.262
40905 94.026 -79.329 52.245
40906 94.017 -79.353 52.228
40907 94.026 -79.366 52.211
40908 94.047 -79.358 52.193
40909 94.029 -79.328 52.173
40910 94.013 -79.297 52.153
40911 94.026 -79.287 52.132
40912 94.045 -79.284 52.113
40913 94.064 -79.281 52.094
40914 94.084 -79.281 52.075
40915 94.104 -79.281 52.058
40916 94.124 -79.268 52.042
40917 94.131 -79.266 52.025
40918 94.137 -79.257 52.006
40919 94.143 -79.249 51.988
40920 94.150 -79.241 51.971
40921 94.154 -79.235 51.958
40922 94.156 -79.233 51.946
40923 94.158 -79.231 51.936
40924 94.175 -79.234 51.926
40925 94.192 -79.238 51.914
40926 94.206 -79.244 51.902
40927 94.213 -79.253 51.891
40928 94.220 -79.262 51.879
40929 94.225 -79.262 51.867
40930 94.230 -79.261 51.855
40931 94.236 -79.261 51.841
40932 94.241 -79.260 51.826
40933 94.249 -79.255 51.810
40934 94.257 -79.250 51.792
40935 94.265 -79.245 51.772
40936 94.272 -79.241 51.752
40937 94.280 -79.235 51.734
40938 94.288 -79.231 51.715
40939 94.296 -79.225 51.691
40940 94.304 -79.221 51.665
40941 94.312 -79.216 51.639
40942 94.320 -79.211 51.612
40943 94.328 -79.206 51.584
40944 94.335 -79.201 51.557
40945 94.318 -79.209 51.530
40946 94.313 -79.213 51.500
40947 94.315 -79.212 51.468
40948 94.296 -79.231 51.435
40949 94.261 -79.261 51.400
40950 94.255 -79.256 51.364
40951 94.258 -79.248 51.329
40952 94.285 -79.231 51.291
40953 94.299 -79.225 51.252
40954 94.311 -79.222 51.211
40955 94.319 -79.228 51.169
40956 94.327 -79.235 51.127
40957 94.335 -79.243 51.085
40958 94.344 -79.250 51.043
40959 94.352 -79.258 51.001
40960 94.360 -79.266 50.958
40961 94.368 -79.273 50.916
40962 94.377 -79.280 50.875
40963 94.385 -79.288 50.834
41524 93.028 -79.776 53.061
41525 93.044 -79.767 53.040
41526 93.060 -79.760 53.019
41527 93.076 -79.751 53.000
41528 93.092 -79.744 52.982
41529 93.108 -79.736 52.965
41530 93.124 -79.727 52.949
41531 93.140 -79.720 52.935
41532 93.157 -79.708 52.921
41533 93.176 -79.693 52.910
41534 93.194 -79.679 52.900
41535 93.212 -79.666 52.893
41536 93.230 -79.652 52.887
41537 93.240 -79.646 52.881
41538 93.259 -79.639 52.876
41539 93.282 -79.631 52.872
41540 93.305 -79.623 52.868
41541 93.328 -79.616 52.862
41542 93.350 -79.609 52.856
41543 93.372 -79.603 52.850
41544 93.395 -79.597 52.843
41545 93.417 -79.590 52.838
41546 93.439 -79.584 52.831
41547 93.461 -79.577 52.824
41548 93.472 -79.567 52.815
41549 93.480 -79.556 52.807
41550 93.487 -79.545 52.799
41551 93.495 -79.534 52.790
41552 93.503 -79.523 52.782
41553 93.511 -79.512 52.773
41554 93.521 -79.504 52.765
41555 93.531 -79.497 52.758
41556 93.541 -79.489 52.752
41557 93.551 -79.482 52.747
41558 93.561 -79.474 52.741
41559 93.571 -79.467 52.732
41560 93.587 -79.464 52.722
41561 93.604 -79.461 52.711
41562 93.621 -79.457 52.697
41563 93.647 -79.441 52.686
41564 93.672 -79.426 52.676
41565 93.698 -79.411 52.668
41566 93.723 -79.396 52.662
41567 93.750 -79.383 52.656
41568 93.761 -79.377 52.650
41569 93.774 -79.369 52.644
41570 93.786 -79.363 52.639
41571 93.799 -79.355 52.636
41572 93.811 -79.347 52.636
41573 93.824 -79.340 52.636
41574 93.836 -79.332 52.636
41575 93.849 -79.325 52.635
41576 93.861 -79.318 52.634
41577 93.874 -79.310 52.634
41578 93.887 -79.303 52.634
41579 93.899 -79.295 52.630
41580 93.902 -79.294 52.623
41581 93.901 -79.294 52.614
41582 93.898 -79.297 52.605
41583 93.895 -79.300 52.595
41584 93.891 -79.303 52.584
41585 93.888 -79.307 52.572
41586 93.885 -79.310 52.563
41587 93.888 -79.310 52.553
41588 93.896 -79.309 52.543
41589 93.905 -79.310 52.530
41590 93.917 -79.310 52.515
41591 93.929 -79.310 52.500
41592 93.933 -79.312 52.484
41593 93.936 -79.313 52.467
41594 93.939 -79.315 52.450
41595 93.946 -79.313 52.432
41596 93.955 -79.309 52.415
41597 93.963 -79.306 52.399
41598 93.971 -79.302 52.383
41599 93.978 -79.298 52.368
41600 93.986 -79.294 52.351
41601 93.994 -79.291 52.332
41602 94.005 -79.290 52.314
41603 94.019 -79.294 52.296
41604 94.034 -79.297 52.279
41605 94.047 -79.300 52.261
41606 94.045 -79.315 52.244
41607 94.038 -79.341 52.228
41608 94.029 -79.365 52.211
41609 94.028 -79.361 52.192
41610 94.019 -79.328 52.173
41611 94.011 -79.296 52.153
41612 94.026 -79.289 52.131
41613 94.046 -79.286 52.111
41614 94.065 -79.284 52.092
41615 94.084 -79.281 52.073
41616 94.104 -79.278 52.057
41617 94.112 -79.268 52.040
41618 94.123 -79.267 52.024
41619 94.129 -79.259 52.005
41620 94.136 -79.250 51.987
41621 94.142 -79.243 51.972
41622 94.144 -79.239 51.960
41623 94.146 -79.237 51.949
41624 94.148 -79.234 51.940
41625 94.161 -79.237 51.931
41626 94.179 -79.240 51.919
41627 94.196 -79.244 51.906
41628 94.209 -79.250 51.893
41629 94.216 -79.259 51.879
41630 94.221 -79.261 51.867
41631 94.227 -79.261 51.855
41632 94.232 -79.260 51.843
41633 94.240 -79.256 51.827
41634 94.247 -79.251 51.811
41635 94.255 -79.247 51.794
41636 94.263 -79.241 51.774
41637 94.271 -79.237 51.755
41638 94.278 -79.231 51.737
41639 94.286 -79.227 51.718
41640 94.294 -79.222 51.694
41641 94.302 -79.217 51.669
41642 94.310 -79.212 51.642
41643 94.318 -79.207 51.616
41644 94.326 -79.202 51.590
41645 94.333 -79.197 51.565
41646 94.341 -79.192 51.536
41647 94.333 -79.197 51.506
41648 94.327 -79.203 51.473
41649 94.274 -79.253 51.438
41650 94.261 -79.257 51.403
41651 94.256 -79.252 51.368
41652 94.255 -79.245 51.333
41653 94.282 -79.228 51.296
41654 94.305 -79.216 51.258
41655 94.316 -79.212 51.219
41656 94.327 -79.209 51.178
41657 94.336 -79.216 51.136
41658 94.344 -79.222 51.095
41659 94.352 -79.230 51.052
41660 94.360 -79.237 51.010
41661 94.368 -79.245 50.968
41662 94.377 -79.253 50.925
41663 94.385 -79.260 50.884
41664 94.394 -79.268 50.843
41724 95.166 -80.028 48.741
41725 95.186 -80.034 48.705
41726 95.210 -80.035 48.668
41727 95.233 -80.037 48.631
41728 95.257 -80.038 48.595
41729 95.280 -80.040 48.560
41730 95.304 -80.041 48.524
41731 95.328 -80.043 48.489
42226 93.040 -79.751 53.061
42227 93.056 -79.742 53.040
42228 93.072 -79.735 53.020
42229 93.088 -79.726 53.001
42230 93.104 -79.719 52.984
42231 93.120 -79.711 52.967
42232 93.136 -79.702 52.952
42233 93.152 -79.695 52.938
42234 93.170 -79.681 52.926
42235 93.188 -79.667 52.915
42236 93.206 -79.654 52.906
42237 93.224 -79.639 52.899
42238 93.242 -79.626 52.893
42239 93.255 -79.618 52.887
42240 93.277 -79.610 52.881
42241 93.300 -79.602 52.875
42242 93.323 -79.596 52.868
42243 93.345 -79.590 52.861
42244 93.367 -79.583 52.855
42245 93.389 -79.577 52.848
42246 93.411 -79.570 52.843
42247 93.433 -79.564 52.837
42248 93.455 -79.558 52.829
42249 93.474 -79.551 52.822
42250 93.484 -79.541 52.813
42251 93.493 -79.534 52.804
42252 93.503 -79.526 52.796
42253 93.513 -79.519 52.787
42254 93.523 -79.511 52.778
42255 93.533 -79.504 52.769
42256 93.543 -79.496 52.761
42257 93.553 -79.488 52.754
42258 93.563 -79.481 52.748
42259 93.573 -79.473 52.741
42260 93.587 -79.469 52.732
42261 93.603 -79.466 52.722
42262 93.619 -79.463 52.710
42263 93.635 -79.460 52.699
42264 93.651 -79.457 52.689
42265 93.673 -79.447 52.679
42266 93.698 -79.432 52.673
42267 93.723 -79.418 52.667
42268 93.745 -79.404 52.661
42269 93.758 -79.396 52.654
42270 93.771 -79.387 52.648
42271 93.784 -79.379 52.642
42272 93.797 -79.369 52.638
42273 93.809 -79.361 52.638
42274 93.823 -79.353 52.639
42275 93.836 -79.344 52.639
42276 93.848 -79.335 52.637
42277 93.861 -79.326 52.637
42278 93.873 -79.310 52.637
42279 93.885 -79.295 52.637
42280 93.897 -79.286 52.633
42281 93.897 -79.289 52.626
42282 93.893 -79.293 52.618
42283 93.890 -79.296 52.609
42284 93.886 -79.299 52.599
42285 93.883 -79.303 52.587
42286 93.880 -79.306 52.574
42287 93.878 -79.308 52.564
42288 93.889 -79.303 52.554
42289 93.897 -79.301 52.543
42290 93.903 -79.303 52.530
42291 93.913 -79.303 52.515
42292 93.926 -79.303 52.501
42293 93.938 -79.303 52.484
42294 93.951 -79.304 52.468
42295 93.961 -79.305 52.451
42296 93.964 -79.306 52.434
42297 93.969 -79.306 52.417
42298 93.977 -79.303 52.401
42299 93.985 -79.299 52.385
42300 93.993 -79.295 52.369
42301 94.001 -79.291 52.351
42302 94.016 -79.295 52.332
42303 94.030 -79.298 52.313
42304 94.032 -79.299 52.295
42305 94.034 -79.300 52.277
42306 94.036 -79.300 52.260
42307 94.038 -79.300 52.243
42308 94.036 -79.316 52.227
42309 94.030 -79.343 52.210
42310 94.024 -79.360 52.192
42311 94.016 -79.327 52.174
42312 94.008 -79.294 52.154
42313 94.027 -79.291 52.131
42314 94.046 -79.288 52.111
42315 94.065 -79.286 52.092
42316 94.084 -79.284 52.073
42317 94.091 -79.272 52.057
42318 94.097 -79.259 52.041
42319 94.115 -79.269 52.024
42320 94.121 -79.261 52.006
42321 94.128 -79.253 51.989
42322 94.133 -79.245 51.974
42323 94.135 -79.243 51.963
42324 94.137 -79.241 51.954
42325 94.138 -79.238 51.945
42326 94.148 -79.238 51.937
42327 94.165 -79.242 51.924
42328 94.182 -79.245 51.911
42329 94.199 -79.249 51.896
42330 94.212 -79.255 51.881
42331 94.218 -79.261 51.868
42332 94.223 -79.260 51.856
42333 94.229 -79.258 51.845
42334 94.238 -79.253 51.829
42335 94.246 -79.248 51.812
42336 94.253 -79.243 51.795
42337 94.261 -79.238 51.776
42338 94.269 -79.233 51.757
42339 94.277 -79.228 51.738
42340 94.284 -79.223 51.718
42341 94.292 -79.218 51.695
42342 94.300 -79.213 51.670
42343 94.308 -79.208 51.644
42344 94.316 -79.203 51.621
42345 94.324 -79.198 51.597
42346 94.332 -79.194 51.572
42347 94.339 -79.189 51.544
42348 94.345 -79.187 51.512
42349 94.305 -79.226 51.479
42350 94.267 -79.258 51.443
42351 94.262 -79.253 51.409
42352 94.257 -79.248 51.374
42353 94.253 -79.243 51.340
42354 94.280 -79.226 51.304
42355 94.307 -79.209 51.266
42356 94.321 -79.203 51.227
42357 94.332 -79.200 51.187
42358 94.344 -79.196 51.146
42359 94.352 -79.203 51.104
42360 94.360 -79.209 51.062
42361 94.369 -79.217 51.019
42362 94.377 -79.225 50.977
42363 94.386 -79.232 50.935
42364 94.394 -79.240 50.893
42365 94.402 -79.247 50.852
42424 95.144 -79.976 48.779
42425 95.158 -79.989 48.743
42426 95.176 -79.996 48.707
42427 95.200 -79.997 48.671
42428 95.224 -79.999 48.634
42429 95.247 -80.000 48.597
42430 95.271 -80.002 48.561
42431 95.294 -80.003 48.526
42432 95.318 -80.005 48.490
42433 95.341 -80.006 48.455
42434 95.365 -80.008 48.419
42928 93.052 -79.726 53.062
42929 93.068 -79.717 53.041
42930 93.084 -79.710 53.022
42931 93.100 -79.701 53.004
42932 93.116 -79.694 52.987
42933 93.132 -79.686 52.971
42934 93.148 -79.677 52.957
42935 93.164 -79.669 52.944
42936 93.183 -79.655 52.932
42937 93.201 -79.641 52.922
42938 93.219 -79.627 52.912
42939 93.237 -79.614 52.904
42940 93.255 -79.600 52.898
42941 93.272 -79.590 52.891
42942 93.295 -79.583 52.883
42943 93.317 -79.576 52.875
42944 93.339 -79.570 52.868
42945 93.361 -79.564 52.862
42946 93.382 -79.558 52.856
42947 93.403 -79.553 52.850
42948 93.425 -79.547 52.844
42949 93.446 -79.541 52.836
42950 93.467 -79.535 52.829
42951 93.483 -79.532 52.820
42952 93.499 -79.529 52.811
42953 93.515 -79.526 52.803
42954 93.525 -79.518 52.794
42955 93.536 -79.510 52.785
42956 93.546 -79.503 52.776
42957 93.555 -79.495 52.767
42958 93.565 -79.488 52.760
42959 93.575 -79.480 52.752
42960 93.586 -79.473 52.746
42961 93.602 -79.470 52.737
42962 93.618 -79.468 52.726
42963 93.634 -79.465 52.713
42964 93.650 -79.462 52.704
42965 93.666 -79.459 52.695
42966 93.682 -79.457 52.686
42967 93.698 -79.454 52.679
42968 93.723 -79.439 52.672
42969 93.741 -79.429 52.666
42970 93.754 -79.419 52.660
42971 93.766 -79.411 52.653
42972 93.779 -79.403 52.648
42973 93.791 -79.387 52.644
42974 93.802 -79.371 52.643
42975 93.814 -79.356 52.643
42976 93.826 -79.340 52.642
42977 93.838 -79.324 52.641
42978 93.849 -79.309 52.641
42979 93.861 -79.293 52.641
42980 93.874 -79.286 52.640
42981 93.888 -79.285 52.636
42982 93.888 -79.288 52.630
42983 93.885 -79.291 52.621
42984 93.881 -79.295 52.613
42985 93.878 -79.298 52.602
42986 93.874 -79.301 52.589
42987 93.871 -79.305 52.577
42988 93.880 -79.301 52.566
42989 93.891 -79.296 52.555
42990 93.899 -79.294 52.544
42991 93.904 -79.296 52.530
42992 93.910 -79.297 52.516
42993 93.922 -79.297 52.502
42994 93.934 -79.297 52.485
42995 93.947 -79.297 52.469
42996 93.960 -79.298 52.452
42997 93.972 -79.298 52.436
42998 93.985 -79.299 52.419
42999 93.992 -79.300 52.403
43000 94.000 -79.296 52.386
43001 94.011 -79.297 52.368
43002 94.014 -79.297 52.350
43003 94.016 -79.298 52.331
43004 94.018 -79.299 52.311
43005 94.020 -79.299 52.293
43006 94.023 -79.300 52.275
43007 94.025 -79.300 52.257
43008 94.027 -79.301 52.240
43009 94.029 -79.301 52.224
43010 94.026 -79.318 52.208
43011 94.020 -79.344 52.192
43012 94.013 -79.325 52.175
43013 94.008 -79.296 52.155
43014 94.027 -79.293 52.132
43015 94.047 -79.291 52.111
43016 94.066 -79.288 52.092
43017 94.071 -79.275 52.073
43018 94.077 -79.263 52.056
43019 94.082 -79.251 52.041
43020 94.100 -79.259 52.027
43021 94.113 -79.263 52.009
43022 94.120 -79.254 51.993
43023 94.124 -79.249 51.979
43024 94.125 -79.247 51.969
43025 94.127 -79.244 51.960
43026 94.129 -79.241 51.951
43027 94.134 -79.241 51.942
43028 94.151 -79.244 51.929
43029 94.168 -79.247 51.914
43030 94.185 -79.251 51.900
43031 94.202 -79.254 51.884
43032 94.214 -79.260 51.871
43033 94.220 -79.259 51.858
43034 94.228 -79.254 51.847
43035 94.236 -79.249 51.831
43036 94.244 -79.244 51.813
43037 94.251 -79.239 51.795
43038 94.259 -79.234 51.777
43039 94.267 -79.229 51.757
43040 94.273 -79.226 51.737
43041 94.280 -79.222 51.717
43042 94.286 -79.219 51.694
43043 94.293 -79.216 51.670
43044 94.299 -79.212 51.647
43045 94.306 -79.208 51.625
43046 94.313 -79.205 51.603
43047 94.319 -79.201 51.578
43048 94.321 -79.203 51.551
43049 94.317 -79.213 51.520
43050 94.282 -79.249 51.485
43051 94.268 -79.254 51.449
43052 94.263 -79.250 51.415
43053 94.258 -79.244 51.381
43054 94.252 -79.239 51.347
43055 94.278 -79.223 51.312
43056 94.305 -79.206 51.275
43057 94.327 -79.194 51.235
43058 94.338 -79.190 51.196
43059 94.349 -79.187 51.155
43060 94.360 -79.183 51.113
43061 94.369 -79.190 51.071
43062 94.377 -79.197 51.029
43063 94.386 -79.204 50.987
43064 94.394 -79.212 50.945
43065 94.402 -79.219 50.903
43066 94.410 -79.227 50.862
43124 95.123 -79.924 48.818
43125 95.136 -79.936 48.781
43126 95.149 -79.949 48.745
43127 95.167 -79.958 48.709
43128 95.190 -79.959 48.672
43129 95.214 -79.961 48.636
43130 95.237 -79.962 48.599
43131 95.261 -79.964 48.563
43132 95.284 -79.965 48.527
43133 95.308 -79.967 48.492
43134 95.332 -79.968 48.456
43135 95.355 -79.970 48.421
43136 95.379 -79.971 48.386
43629 93.048 -79.709 53.084
43630 93.064 -79.701 53.063
43631 93.080 -79.692 53.043
43632 93.095 -79.685 53.025
43633 93.112 -79.676 53.008
43634 93.128 -79.669 52.992
43635 93.144 -79.661 52.976
43636 93.160 -79.652 52.962
43637 93.177 -79.642 52.950
43638 93.195 -79.629 52.939
43639 93.213 -79.615 52.928
43640 93.231 -79.601 52.918
43641 93.249 -79.587 52.910
43642 93.267 -79.573 52.901
43643 93.286 -79.566 52.892
43644 93.307 -79.560 52.886
43645 93.329 -79.554 52.879
43646 93.350 -79.548 52.872
43647 93.371 -79.542 52.866
43648 93.392 -79.536 52.859
43649 93.413 -79.531 52.852
43650 93.435 -79.525 52.845
43651 93.456 -79.520 52.837
43652 93.474 -79.515 52.829
43653 93.490 -79.512 52.819
43654 93.505 -79.508 52.811
43655 93.521 -79.505 52.802
43656 93.537 -79.502 52.793
43657 93.553 -79.499 52.785
43658 93.567 -79.494 52.776
43659 93.577 -79.487 52.768
43660 93.587 -79.479 52.760
43661 93.601 -79.475 52.753
43662 93.617 -79.473 52.743
43663 93.633 -79.469 52.732
43664 93.649 -79.466 52.720
43665 93.665 -79.463 52.710
43666 93.681 -79.460 52.702
43667 93.694 -79.456 52.694
43668 93.707 -79.451 52.686
43669 93.720 -79.446 52.679
43670 93.732 -79.432 52.673
43671 93.744 -79.416 52.666
43672 93.756 -79.400 52.659
43673 93.767 -79.385 52.654
43674 93.779 -79.369 52.650
43675 93.791 -79.353 52.649
43676 93.802 -79.338 52.649
43677 93.814 -79.322 52.649
43678 93.826 -79.306 52.647
43679 93.837 -79.291 52.647
43680 93.851 -79.286 52.646
43681 93.865 -79.285 52.644
43682 93.879 -79.285 52.639
43683 93.880 -79.287 52.633
43684 93.876 -79.291 52.625
43685 93.873 -79.294 52.616
43686 93.869 -79.297 52.605
43687 93.866 -79.300 52.592
43688 93.870 -79.299 52.579
43689 93.881 -79.294 52.568
43690 93.892 -79.288 52.558
43691 93.900 -79.288 52.545
43692 93.906 -79.288 52.533
43693 93.911 -79.290 52.519
43694 93.918 -79.291 52.504
43695 93.930 -79.291 52.487
43696 93.943 -79.291 52.470
43697 93.956 -79.292 52.453
43698 93.968 -79.292 52.438
43699 93.981 -79.292 52.422
43700 93.994 -79.292 52.404
43701 94.004 -79.292 52.385
43702 94.003 -79.295 52.367
43703 94.003 -79.297 52.348
43704 94.004 -79.298 52.329
43705 94.007 -79.299 52.310
43706 94.009 -79.300 52.291
43707 94.011 -79.300 52.273
43708 94.013 -79.300 52.255
43709 94.015 -79.301 52.238
43710 94.017 -79.302 52.223
43711 94.019 -79.302 52.207
43712 94.017 -79.319 52.192
43713 94.010 -79.324 52.175
43714 94.008 -79.298 52.155
43715 94.028 -79.295 52.133
43716 94.046 -79.292 52.113
43717 94.051 -79.280 52.094
43718 94.057 -79.267 52.077
43719 94.063 -79.255 52.062
43720 94.068 -79.242 52.048
43721 94.084 -79.249 52.035
43722 94.103 -79.259 52.019
43723 94.112 -79.256 52.004
43724 94.114 -79.252 51.992
43725 94.116 -79.250 51.982
43726 94.118 -79.247 51.973
43727 94.120 -79.245 51.964
43728 94.121 -79.243 51.953
43729 94.137 -79.246 51.939
43730 94.154 -79.250 51.923
43731 94.172 -79.253 51.907
43732 94.189 -79.256 51.891
43733 94.208 -79.258 51.876
43734 94.214 -79.259 51.862
43735 94.221 -79.256 51.850
43736 94.227 -79.252 51.833
43737 94.234 -79.248 51.815
43738 94.241 -79.245 51.796
43739 94.247 -79.241 51.776
43740 94.254 -79.237 51.755
43741 94.260 -79.234 51.735
43742 94.267 -79.231 51.715
43743 94.273 -79.227 51.692
43744 94.280 -79.224 51.670
43745 94.286 -79.220 51.650
43746 94.292 -79.216 51.630
43747 94.299 -79.213 51.608
43748 94.297 -79.221 51.585
43749 94.293 -79.231 51.558
43750 94.290 -79.241 51.527
43751 94.274 -79.256 51.493
43752 94.269 -79.250 51.458
43753 94.264 -79.245 51.423
43754 94.258 -79.241 51.389
43755 94.253 -79.235 51.355
43756 94.275 -79.220 51.321
43757 94.302 -79.203 51.284
43758 94.330 -79.187 51.245
43759 94.343 -79.181 51.204
43760 94.354 -79.178 51.164
43761 94.365 -79.174 51.123
43762 94.377 -79.170 51.081
43763 94.386 -79.177 51.039
43764 94.394 -79.184 50.997
43765 94.402 -79.192 50.954
43766 94.410 -79.199 50.913
43767 94.419 -79.206 50.873
43824 95.102 -79.871 48.858
43825 95.115 -79.884 48.821
43826 95.128 -79.897 48.784
43827 95.141 -79.910 48.748
43828 95.157 -79.921 48.711
43829 95.180 -79.922 48.674
43830 95.204 -79.924 48.637
43831 95.228 -79.925 48.601
43832 95.251 -79.926 48.565
43833 95.275 -79.927 48.529
43834 95.298 -79.929 48.493
43835 95.322 -79.930 48.457
43836 95.345 -79.932 48.423
43837 95.369 -79.933 48.387
44331 93.060 -79.684 53.086
44332 93.075 -79.676 53.066
44333 93.091 -79.667 53.047
44334 93.108 -79.660 53.030
44335 93.124 -79.651 53.013
44336 93.140 -79.644 52.997
44337 93.156 -79.636 52.982
44338 93.171 -79.627 52.969
44339 93.189 -79.617 52.956
44340 93.207 -79.602 52.944
44341 93.225 -79.588 52.933
44342 93.243 -79.575 52.924
44343 93.262 -79.561 52.914
44344 93.280 -79.550 52.904
44345 93.298 -79.544 52.898
44346 93.318 -79.538 52.890
44347 93.339 -79.532 52.884
44348 93.360 -79.526 52.878
44349 93.381 -79.520 52.872
44350 93.402 -79.515 52.865
44351 93.423 -79.509 52.858
44352 93.445 -79.503 52.850
44353 93.465 -79.498 52.842
44354 93.480 -79.494 52.833
44355 93.496 -79.491 52.824
44356 93.512 -79.488 52.814
44357 93.528 -79.485 52.805
44358 93.544 -79.482 52.796
44359 93.559 -79.479 52.789
44360 93.575 -79.476 52.780
44361 93.588 -79.472 52.772
44362 93.603 -79.467 52.761
44363 93.616 -79.462 52.750
44364 93.628 -79.457 52.739
44365 93.641 -79.452 52.729
44366 93.654 -79.447 52.719
44367 93.667 -79.443 52.711
44368 93.680 -79.438 52.704
44369 93.693 -79.433 52.695
44370 93.707 -79.426 52.688
44371 93.720 -79.414 52.681
44372 93.732 -79.398 52.673
44373 93.744 -79.382 52.666
44374 93.756 -79.367 52.660
44375 93.767 -79.351 52.657
44376 93.779 -79.336 52.656
44377 93.790 -79.320 52.656
44378 93.802 -79.304 52.655
44379 93.814 -79.289 52.654
44380 93.827 -79.285 52.653
44381 93.841 -79.285 52.651
44382 93.855 -79.285 52.648
44383 93.869 -79.284 52.643
44384 93.871 -79.286 52.636
44385 93.867 -79.289 52.627
44386 93.864 -79.293 52.619
44387 93.861 -79.296 52.608
44388 93.860 -79.297 52.595
44389 93.871 -79.292 52.583
44390 93.883 -79.287 52.572
44391 93.894 -79.281 52.561
44392 93.902 -79.280 52.548
44393 93.907 -79.281 52.535
44394 93.913 -79.283 52.521
44395 93.919 -79.284 52.506
44396 93.926 -79.285 52.489
44397 93.939 -79.285 52.472
44398 93.952 -79.285 52.456
44399 93.964 -79.286 52.439
44400 93.980 -79.285 52.421
44401 93.995 -79.283 52.403
44402 94.004 -79.283 52.385
44403 94.004 -79.286 52.367
44404 94.003 -79.288 52.348
44405 94.003 -79.291 52.329
44406 94.003 -79.294 52.310
44407 94.002 -79.296 52.291
44408 94.002 -79.298 52.272
44409 94.002 -79.300 52.254
44410 94.004 -79.301 52.238
44411 94.006 -79.302 52.222
44412 94.008 -79.302 52.207
44413 94.010 -79.303 52.192
44414 94.007 -79.320 52.175
44415 94.009 -79.300 52.156
44416 94.026 -79.296 52.135
44417 94.031 -79.284 52.116
44418 94.037 -79.271 52.098
44419 94.043 -79.259 52.083
44420 94.048 -79.247 52.070
44421 94.054 -79.234 52.057
44422 94.069 -79.239 52.045
44423 94.087 -79.250 52.031
44424 94.103 -79.258 52.018
44425 94.105 -79.256 52.007
44426 94.107 -79.253 51.997
44427 94.109 -79.250 51.987
44428 94.112 -79.247 51.977
44429 94.114 -79.243 51.965
44430 94.129 -79.243 51.949
44431 94.148 -79.244 51.932
44432 94.167 -79.246 51.914
44433 94.186 -79.247 51.898
44434 94.197 -79.253 51.882
44435 94.202 -79.265 51.866
44436 94.208 -79.264 51.852
44437 94.214 -79.260 51.834
44438 94.221 -79.256 51.816
44439 94.227 -79.253 51.797
44440 94.234 -79.250 51.776
44441 94.241 -79.246 51.755
44442 94.247 -79.242 51.735
44443 94.253 -79.239 51.716
44444 94.260 -79.235 51.694
44445 94.266 -79.231 51.672
44446 94.273 -79.228 51.652
44447 94.277 -79.228 51.632
44448 94.273 -79.238 51.611
44449 94.270 -79.248 51.588
44450 94.266 -79.257 51.564
44451 94.268 -79.251 51.536
44452 94.275 -79.252 51.503
44453 94.270 -79.247 51.468
44454 94.264 -79.241 51.433
44455 94.259 -79.237 51.399
44456 94.254 -79.231 51.366
44457 94.273 -79.218 51.331
44458 94.300 -79.201 51.294
44459 94.327 -79.184 51.255
44460 94.348 -79.172 51.214
44461 94.360 -79.169 51.174
44462 94.371 -79.165 51.132
44463 94.383 -79.161 51.091
44464 94.394 -79.158 51.049
44465 94.402 -79.164 51.007
44466 94.410 -79.172 50.964
44467 94.419 -79.179 50.923
44468 94.427 -79.186 50.883
44524 95.080 -79.819 48.897
44525 95.093 -79.832 48.861
44526 95.106 -79.845 48.824
44527 95.119 -79.858 48.787
44528 95.132 -79.871 48.750
44529 95.147 -79.883 48.713
44530 95.171 -79.884 48.676
44531 95.194 -79.886 48.639
44532 95.218 -79.887 48.603
44533 95.241 -79.889 48.567
44534 95.265 -79.890 48.531
44535 95.288 -79.892 48.495
44536 95.312 -79.893 48.459
44537 95.336 -79.895 48.424
44538 95.359 -79.896 48.389
44539 95.383 -79.897 48.353
45033 93.071 -79.659 53.089
45034 93.087 -79.651 53.070
45035 93.103 -79.642 53.051
45036 93.120 -79.635 53.034
45037 93.136 -79.626 53.017
45038 93.152 -79.619 53.002
45039 93.167 -79.610 52.988
45040 93.184 -79.602 52.975
45041 93.201 -79.590 52.962
45042 93.219 -79.576 52.950
45043 93.237 -79.563 52.939
45044 93.256 -79.548 52.928
45045 93.274 -79.535 52.919
45046 93.292 -79.529 52.910
45047 93.311 -79.523 52.903
45048 93.330 -79.516 52.896
45049 93.349 -79.510 52.892
45050 93.370 -79.504 52.886
45051 93.391 -79.498 52.880
45052 93.412 -79.493 52.873
45053 93.433 -79.487 52.866
45054 93.455 -79.482 52.857
45055 93.471 -79.478 52.848
45056 93.487 -79.475 52.839
45057 93.502 -79.472 52.829
45058 93.518 -79.468 52.819
45059 93.534 -79.465 52.810
45060 93.550 -79.462 52.801
45061 93.560 -79.457 52.792
45062 93.570 -79.453 52.782
45063 93.586 -79.448 52.770
45064 93.602 -79.444 52.759
45065 93.615 -79.439 52.749
45066 93.628 -79.435 52.739
45067 93.640 -79.429 52.729
45068 93.653 -79.425 52.721
45069 93.666 -79.419 52.713
45070 93.680 -79.415 52.705
45071 93.694 -79.407 52.698
45072 93.708 -79.396 52.690
45073 93.720 -79.381 52.682
45074 93.732 -79.365 52.674
45075 93.744 -79.349 52.667
45076 93.755 -79.334 52.665
45077 93.767 -79.318 52.664
45078 93.778 -79.303 52.664
45079 93.790 -79.287 52.662
45080 93.804 -79.285 52.660
45081 93.818 -79.285 52.658
45082 93.832 -79.284 52.656
45083 93.846 -79.283 52.651
45084 93.860 -79.282 52.645
45085 93.862 -79.285 52.638
45086 93.859 -79.288 52.629
45087 93.856 -79.291 52.621
45088 93.852 -79.294 52.610
45089 93.862 -79.291 52.598
45090 93.873 -79.285 52.587
45091 93.885 -79.279 52.576
45092 93.896 -79.274 52.565
45093 93.903 -79.273 52.552
45094 93.909 -79.275 52.538
45095 93.914 -79.275 52.524
45096 93.920 -79.277 52.508
45097 93.926 -79.278 52.492
45098 93.935 -79.279 52.475
45099 93.949 -79.278 52.458
45100 93.965 -79.277 52.441
45101 93.981 -79.275 52.422
45102 93.996 -79.274 52.404
45103 94.005 -79.274 52.386
45104 94.005 -79.277 52.368
45105 94.004 -79.279 52.350
45106 94.004 -79.281 52.331
45107 94.003 -79.285 52.311
45108 94.003 -79.287 52.293
45109 94.002 -79.289 52.275
45110 94.002 -79.292 52.256
45111 94.002 -79.294 52.239
45112 94.001 -79.297 52.223
45113 94.001 -79.300 52.209
45114 94.000 -79.302 52.194
45115 94.001 -79.303 52.178
45116 94.006 -79.300 52.159
45117 94.011 -79.288 52.138
45118 94.017 -79.275 52.120
45119 94.023 -79.263 52.103
45120 94.028 -79.250 52.090
45121 94.034 -79.237 52.079
45122 94.040 -79.225 52.069
45123 94.054 -79.229 52.058
45124 94.072 -79.240 52.047
45125 94.098 -79.256 52.036
45126 94.101 -79.254 52.025
45127 94.104 -79.250 52.013
45128 94.107 -79.245 52.002
45129 94.110 -79.240 51.990
45130 94.113 -79.234 51.976
45131 94.125 -79.234 51.960
45132 94.144 -79.235 51.942
45133 94.164 -79.237 51.923
45134 94.181 -79.239 51.905
45135 94.185 -79.250 51.887
45136 94.190 -79.262 51.870
45137 94.195 -79.272 51.853
45138 94.201 -79.269 51.835
45139 94.207 -79.265 51.816
45140 94.214 -79.261 51.796
45141 94.220 -79.258 51.775
45142 94.227 -79.254 51.755
45143 94.234 -79.250 51.736
45144 94.240 -79.247 51.718
45145 94.247 -79.244 51.697
45146 94.253 -79.240 51.675
45147 94.253 -79.246 51.655
45148 94.249 -79.255 51.634
45149 94.246 -79.265 51.614
45150 94.248 -79.256 51.592
45151 94.251 -79.244 51.569
45152 94.258 -79.227 51.542
45153 94.276 -79.206 51.511
45154 94.270 -79.243 51.477
45155 94.265 -79.237 51.443
45156 94.260 -79.233 51.409
45157 94.255 -79.228 51.377
45158 94.270 -79.215 51.342
45159 94.298 -79.198 51.304
45160 94.325 -79.181 51.264
45161 94.352 -79.165 51.224
45162 94.365 -79.159 51.183
45163 94.377 -79.156 51.142
45164 94.388 -79.152 51.100
45165 94.399 -79.149 51.059
45166 94.410 -79.145 51.018
45167 94.419 -79.151 50.975
45168 94.427 -79.159 50.934
45169 94.435 -79.166 50.894
45224 95.058 -79.767 48.938
45225 95.071 -79.780 48.901
45226 95.085 -79.792 48.864
45227 95.098 -79.805 48.827
45228 95.111 -79.818 48.791
45229 95.124 -79.832 48.753
45230 95.137 -79.845 48.715
45231 95.161 -79.846 48.678
45232 95.184 -79.848 48.642
45233 95.208 -79.849 48.605
45234 95.232 -79.851 48.568
45235 95.255 -79.852 48.533
45236 95.279 -79.854 48.498
45237 95.302 -79.855 48.462
45238 95.326 -79.857 48.426
45239 95.349 -79.858 48.390
45240 95.373 -79.860 48.355
45241 95.397 -79.861 48.320
45735 93.083 -79.634 53.092
45736 93.099 -79.626 53.074
45737 93.116 -79.617 53.056
45738 93.132 -79.610 53.039
45739 93.147 -79.601 53.023
45740 93.163 -79.594 53.008
45741 93.180 -79.585 52.995
45742 93.196 -79.577 52.982
45743 93.214 -79.563 52.970
45744 93.232 -79.550 52.958
45745 93.250 -79.536 52.947
45746 93.268 -79.523 52.936
45747 93.286 -79.513 52.928
45748 93.305 -79.507 52.920
45749 93.324 -79.501 52.913
45750 93.342 -79.494 52.908
45751 93.360 -79.488 52.904
45752 93.380 -79.482 52.898
45753 93.401 -79.477 52.892
45754 93.422 -79.471 52.883
45755 93.444 -79.465 52.874
45756 93.462 -79.461 52.864
45757 93.477 -79.458 52.855
45758 93.493 -79.454 52.845
45759 93.509 -79.451 52.835
45760 93.521 -79.447 52.825
45761 93.532 -79.443 52.815
45762 93.541 -79.438 52.805
45763 93.552 -79.434 52.793
45764 93.568 -79.429 52.781
45765 93.585 -79.426 52.770
45766 93.601 -79.421 52.760
45767 93.614 -79.416 52.749
45768 93.627 -79.411 52.740
45769 93.640 -79.407 52.732
45770 93.653 -79.401 52.723
45771 93.666 -79.397 52.715
45772 93.681 -79.386 52.707
45773 93.695 -79.376 52.700
45774 93.708 -79.363 52.692
45775 93.720 -79.347 52.685
45776 93.731 -79.332 52.680
45777 93.743 -79.316 52.675
45778 93.755 -79.300 52.672
45779 93.767 -79.285 52.670
45780 93.781 -79.285 52.667
45781 93.795 -79.284 52.665
45782 93.809 -79.284 52.662
45783 93.823 -79.283 52.658
45784 93.837 -79.282 52.653
45785 93.851 -79.281 52.647
45786 93.854 -79.284 52.640
45787 93.851 -79.287 52.632
45788 93.847 -79.291 52.623
45789 93.852 -79.288 52.612
45790 93.864 -79.283 52.600
45791 93.875 -79.278 52.590
45792 93.886 -79.272 52.579
45793 93.897 -79.267 52.567
45794 93.905 -79.266 52.555
45795 93.910 -79.267 52.541
45796 93.916 -79.269 52.526
45797 93.922 -79.270 52.510
45798 93.927 -79.271 52.494
45799 93.936 -79.270 52.478
45800 93.951 -79.269 52.461
45801 93.966 -79.268 52.443
45802 93.982 -79.266 52.424
45803 93.997 -79.265 52.407
45804 94.006 -79.266 52.389
45805 94.005 -79.268 52.371
45806 94.005 -79.270 52.352
45807 94.004 -79.273 52.333
45808 94.004 -79.275 52.313
45809 94.004 -79.278 52.295
45810 94.003 -79.281 52.278
45811 94.003 -79.283 52.260
45812 94.002 -79.285 52.243
45813 94.002 -79.288 52.226
45814 94.001 -79.291 52.211
45815 94.001 -79.293 52.196
45816 94.003 -79.292 52.180
45817 94.004 -79.288 52.163
45818 94.002 -79.278 52.143
45819 94.002 -79.266 52.126
45820 94.008 -79.254 52.111
45821 94.014 -79.241 52.098
45822 94.019 -79.229 52.088
45823 94.025 -79.217 52.079
45824 94.039 -79.219 52.072
45825 94.061 -79.233 52.062
45826 94.085 -79.246 52.052
45827 94.099 -79.247 52.042
45828 94.102 -79.242 52.029
45829 94.105 -79.237 52.017
45830 94.108 -79.231 52.004
45831 94.111 -79.226 51.990
45832 94.122 -79.225 51.972
45833 94.141 -79.226 51.953
45834 94.160 -79.228 51.933
45835 94.169 -79.236 51.913
45836 94.174 -79.247 51.893
45837 94.178 -79.259 51.874
45838 94.183 -79.269 51.855
45839 94.188 -79.277 51.835
45840 94.194 -79.273 51.814
45841 94.201 -79.269 51.795
45842 94.207 -79.266 51.775
45843 94.214 -79.263 51.756
45844 94.220 -79.259 51.738
45845 94.227 -79.256 51.719
45846 94.232 -79.253 51.700
45847 94.229 -79.263 51.679
45848 94.225 -79.272 51.658
45849 94.228 -79.263 51.637
45850 94.231 -79.249 51.617
45851 94.235 -79.236 51.597
45852 94.238 -79.222 51.573
45853 94.253 -79.199 51.546
45854 94.273 -79.169 51.517
45855 94.271 -79.226 51.484
45856 94.266 -79.234 51.451
45857 94.261 -79.228 51.419
45858 94.255 -79.224 51.387
45859 94.268 -79.212 51.352
45860 94.295 -79.195 51.314
45861 94.323 -79.178 51.273
45862 94.350 -79.162 51.233
45863 94.370 -79.150 51.192
45864 94.382 -79.147 51.150
45865 94.393 -79.143 51.110
45866 94.404 -79.140 51.069
45867 94.416 -79.136 51.028
45868 94.427 -79.132 50.985
45869 94.435 -79.138 50.945
45870 94.444 -79.146 50.906
45924 95.037 -79.714 48.978
45925 95.050 -79.727 48.941
45926 95.063 -79.740 48.904
45927 95.076 -79.753 48.867
45928 95.090 -79.767 48.830
45929 95.103 -79.780 48.793
45930 95.116 -79.792 48.756
45931 95.129 -79.805 48.718
45932 95.151 -79.808 48.682
45933 95.175 -79.810 48.645
45934 95.198 -79.811 48.608
45935 95.222 -79.813 48.571
45936 95.245 -79.814 48.537
45937 95.269 -79.816 48.501
45938 95.293 -79.817 48.465
45939 95.316 -79.819 48.429
45940 95.340 -79.820 48.393
45941 95.363 -79.822 48.358
45942 95.387 -79.823 48.324
46436 93.079 -79.617 53.115
46437 93.095 -79.609 53.096
46438 93.112 -79.601 53.078
46439 93.127 -79.592 53.061
46440 93.143 -79.585 53.045
46441 93.159 -79.576 53.030
46442 93.176 -79.569 53.016
46443 93.192 -79.560 53.002
46444 93.208 -79.551 52.990
46445 93.226 -79.538 52.978
46446 93.244 -79.524 52.966
46447 93.263 -79.510 52.955
46448 93.281 -79.498 52.946
46449 93.299 -79.491 52.938
46450 93.318 -79.485 52.931
46451 93.336 -79.479 52.927
46452 93.355 -79.473 52.922
46453 93.373 -79.467 52.916
46454 93.392 -79.461 52.909
46455 93.411 -79.455 52.901
46456 93.432 -79.449 52.893
46457 93.452 -79.444 52.885
46458 93.468 -79.441 52.875
46459 93.484 -79.438 52.865
46460 93.493 -79.433 52.854
46461 93.503 -79.429 52.843
46462 93.513 -79.424 52.832
46463 93.523 -79.419 52.820
46464 93.534 -79.415 52.807
46465 93.551 -79.411 52.795
46466 93.567 -79.407 52.784
46467 93.584 -79.402 52.773
46468 93.600 -79.398 52.761
46469 93.613 -79.393 52.752
46470 93.626 -79.388 52.743
46471 93.639 -79.383 52.734
46472 93.652 -79.376 52.725
46473 93.667 -79.366 52.717
46474 93.682 -79.357 52.710
46475 93.696 -79.345 52.704
46476 93.708 -79.329 52.699
46477 93.720 -79.314 52.694
46478 93.731 -79.298 52.688
46479 93.744 -79.285 52.683
46480 93.758 -79.285 52.678
46481 93.771 -79.284 52.674
46482 93.785 -79.283 52.671
46483 93.799 -79.282 52.667
46484 93.814 -79.282 52.663
46485 93.828 -79.281 52.657
46486 93.842 -79.281 52.651
46487 93.846 -79.282 52.644
46488 93.842 -79.286 52.636
46489 93.843 -79.287 52.627
46490 93.854 -79.281 52.616
46491 93.865 -79.276 52.605
46492 93.876 -79.270 52.595
46493 93.888 -79.265 52.584
46494 93.899 -79.259 52.573
46495 93.906 -79.259 52.560
46496 93.912 -79.260 52.547
46497 93.917 -79.262 52.532
46498 93.923 -79.263 52.517
46499 93.931 -79.263 52.500
46500 93.941 -79.261 52.482
46501 93.952 -79.260 52.464
46502 93.967 -79.259 52.446
46503 93.983 -79.257 52.429
46504 93.998 -79.256 52.412
46505 94.007 -79.256 52.394
46506 94.006 -79.259 52.375
46507 94.006 -79.261 52.356
46508 94.005 -79.264 52.335
46509 94.005 -79.266 52.316
46510 94.004 -79.269 52.299
46511 94.004 -79.272 52.283
46512 94.004 -79.274 52.266
46513 94.003 -79.276 52.248
46514 94.003 -79.279 52.230
46515 94.004 -79.279 52.213
46516 94.007 -79.278 52.197
46517 94.010 -79.277 52.182
46518 94.011 -79.273 52.167
46519 94.008 -79.263 52.149
46520 94.006 -79.253 52.133
46521 94.004 -79.243 52.119
46522 94.001 -79.232 52.107
46523 94.005 -79.221 52.097
46524 94.010 -79.208 52.089
46525 94.025 -79.210 52.085
46526 94.047 -79.220 52.076
46527 94.065 -79.225 52.066
46528 94.083 -79.230 52.056
46529 94.100 -79.234 52.043
46530 94.103 -79.228 52.031
46531 94.106 -79.223 52.017
46532 94.109 -79.218 52.002
46533 94.119 -79.216 51.983
46534 94.138 -79.217 51.963
46535 94.153 -79.222 51.942
46536 94.157 -79.232 51.920
46537 94.162 -79.244 51.899
46538 94.167 -79.255 51.877
46539 94.171 -79.266 51.857
46540 94.176 -79.278 51.835
46541 94.181 -79.281 51.815
46542 94.188 -79.278 51.796
46543 94.194 -79.274 51.778
46544 94.200 -79.271 51.761
46545 94.207 -79.267 51.742
46546 94.209 -79.270 51.724
46547 94.205 -79.280 51.705
46548 94.208 -79.268 51.685
46549 94.211 -79.255 51.664
46550 94.215 -79.241 51.643
46551 94.218 -79.228 51.622
46552 94.221 -79.215 51.601
46553 94.226 -79.200 51.576
46554 94.247 -79.170 51.550
46555 94.268 -79.140 51.522
46556 94.273 -79.171 51.491
46557 94.267 -79.230 51.460
46558 94.261 -79.225 51.429
46559 94.256 -79.220 51.396
46560 94.266 -79.209 51.359
46561 94.293 -79.193 51.320
46562 94.320 -79.176 51.280
46563 94.347 -79.159 51.240
46564 94.374 -79.142 51.199
46565 94.387 -79.137 51.159
46566 94.399 -79.134 51.119
46567 94.410 -79.130 51.079
46568 94.421 -79.127 51.038
46569 94.432 -79.123 50.996
46570 94.443 -79.119 50.957
46571 94.452 -79.125 50.917
46625 95.028 -79.675 48.981
46626 95.042 -79.688 48.944
46627 95.055 -79.701 48.908
46628 95.068 -79.714 48.871
46629 95.081 -79.727 48.834
46630 95.094 -79.740 48.796
46631 95.107 -79.753 48.759
46632 95.120 -79.766 48.722
46633 95.141 -79.770 48.685
46634 95.165 -79.772 48.648
46635 95.188 -79.773 48.612
46636 95.212 -79.775 48.575
46637 95.236 -79.776 48.540
46638 95.259 -79.778 48.504
46639 95.282 -79.780 48.467
46640 95.306 -79.781 48.432
46641 95.330 -79.783 48.396
46642 95.353 -79.784 48.361
46643 95.377 -79.786 48.326
46644 95.401 -79.787 48.291
46711 96.850 -79.882 46.251
46712 96.872 -79.886 46.227
46713 96.894 -79.891 46.203
47138 93.091 -79.592 53.120
47139 93.107 -79.584 53.102
47140 93.123 -79.576 53.084
47141 93.139 -79.567 53.067
47142 93.155 -79.560 53.052
47143 93.171 -79.551 53.038
47144 93.188 -79.544 53.024
47145 93.203 -79.535 53.010
47146 93.220 -79.525 52.998
47147 93.238 -79.511 52.987
47148 93.257 -79.498 52.976
47149 93.275 -79.484 52.966
47150 93.293 -79.476 52.957
47151 93.311 -79.469 52.951
47152 93.330 -79.463 52.946
47153 93.349 -79.457 52.941
47154 93.367 -79.451 52.935
47155 93.386 -79.445 52.928
47156 93.404 -79.439 52.921
47157 93.423 -79.433 52.914
47158 93.442 -79.427 52.907
47159 93.455 -79.423 52.897
47160 93.465 -79.419 52.887
47161 93.475 -79.414 52.876
47162 93.485 -79.410 52.864
47163 93.495 -79.405 52.852
47164 93.504 -79.400 52.840
47165 93.516 -79.396 52.827
47166 93.533 -79.392 52.814
47167 93.549 -79.388 52.801
47168 93.566 -79.384 52.789
47169 93.582 -79.379 52.777
47170 93.599 -79.375 52.766
47171 93.612 -79.370 52.757
47172 93.625 -79.366 52.746
47173 93.639 -79.357 52.739
47174 93.654 -79.347 52.734
47175 93.669 -79.337 52.729
47176 93.684 -79.326 52.722
47177 93.696 -79.312 52.716
47178 93.708 -79.296 52.710
47179 93.720 -79.285 52.704
47180 93.734 -79.284 52.698
47181 93.748 -79.284 52.692
47182 93.762 -79.283 52.686
47183 93.776 -79.282 52.680
47184 93.790 -79.281 52.675
47185 93.804 -79.281 52.670
47186 93.819 -79.280 52.663
47187 93.833 -79.280 52.656
47188 93.837 -79.281 52.649
47189 93.834 -79.285 52.642
47190 93.845 -79.279 52.632
47191 93.856 -79.274 52.622
47192 93.867 -79.269 52.612
47193 93.878 -79.263 52.601
47194 93.889 -79.258 52.590
47195 93.900 -79.253 52.579
47196 93.907 -79.252 52.566
47197 93.913 -79.253 52.553
47198 93.919 -79.254 52.538
47199 93.927 -79.254 52.523
47200 93.936 -79.253 52.506
47201 93.946 -79.252 52.488
47202 93.956 -79.250 52.469
47203 93.968 -79.250 52.452
47204 93.984 -79.248 52.436
47205 93.999 -79.247 52.420
47206 94.007 -79.247 52.402
47207 94.007 -79.250 52.383
47208 94.006 -79.253 52.362
47209 94.006 -79.255 52.341
47210 94.006 -79.257 52.322
47211 94.005 -79.260 52.304
47212 94.005 -79.263 52.288
47213 94.004 -79.265 52.272
47214 94.005 -79.266 52.253
47215 94.008 -79.264 52.235
47216 94.011 -79.263 52.217
47217 94.013 -79.263 52.201
47218 94.016 -79.261 52.187
47219 94.018 -79.259 52.172
47220 94.015 -79.249 52.155
47221 94.013 -79.238 52.140
47222 94.010 -79.228 52.126
47223 94.008 -79.218 52.116
47224 94.006 -79.208 52.108
47225 94.003 -79.199 52.101
47226 94.008 -79.195 52.096
47227 94.026 -79.200 52.087
47228 94.044 -79.204 52.078
47229 94.062 -79.209 52.068
47230 94.080 -79.213 52.056
47231 94.099 -79.219 52.042
47232 94.105 -79.215 52.028
47233 94.107 -79.209 52.012
47234 94.116 -79.206 51.993
47235 94.134 -79.208 51.972
47236 94.141 -79.218 51.951
47237 94.145 -79.229 51.927
47238 94.150 -79.241 51.904
47239 94.154 -79.252 51.881
47240 94.159 -79.263 51.858
47241 94.164 -79.275 51.837
47242 94.168 -79.286 51.817
47243 94.174 -79.286 51.799
47244 94.181 -79.282 51.783
47245 94.187 -79.279 51.766
47246 94.185 -79.287 51.748
47247 94.188 -79.274 51.731
47248 94.192 -79.260 51.711
47249 94.195 -79.247 51.691
47250 94.198 -79.234 51.670
47251 94.202 -79.221 51.649
47252 94.205 -79.207 51.627
47253 94.208 -79.194 51.605
47254 94.221 -79.172 51.580
47255 94.242 -79.142 51.554
47256 94.262 -79.112 51.526
47257 94.274 -79.116 51.497
47258 94.268 -79.190 51.468
47259 94.262 -79.221 51.438
47260 94.257 -79.216 51.405
47261 94.263 -79.206 51.367
47262 94.290 -79.190 51.327
47263 94.318 -79.173 51.287
47264 94.345 -79.156 51.248
47265 94.372 -79.140 51.208
47266 94.393 -79.128 51.168
47267 94.404 -79.125 51.129
47268 94.415 -79.121 51.089
47269 94.426 -79.118 51.049
47270 94.437 -79.114 51.008
47271 94.449 -79.110 50.968
47272 94.460 -79.106 50.929
47326 95.020 -79.636 48.985
47327 95.033 -79.648 48.948
47328 95.046 -79.662 48.912
47329 95.059 -79.675 48.875
47330 95.072 -79.688 48.837
47331 95.086 -79.701 48.800
47332 95.099 -79.714 48.762
47333 95.112 -79.726 48.725
47334 95.131 -79.733 48.689
47335 95.155 -79.734 48.652
47336 95.179 -79.736 48.615
47337 95.202 -79.737 48.579
47338 95.226 -79.739 48.542
47339 95.249 -79.740 48.506
47340 95.273 -79.742 48.471
47341 95.297 -79.743 48.435
47342 95.320 -79.745 48.399
47343 95.343 -79.746 48.364
47344 95.367 -79.748 48.329
47345 95.391 -79.749 48.294
47346 95.414 -79.751 48.259
47411 96.823 -79.859 46.274
47412 96.846 -79.864 46.251
47413 96.868 -79.868 46.227
47414 96.891 -79.874 46.202
47415 96.913 -79.878 46.177
47416 96.936 -79.883 46.153
47417 96.958 -79.887 46.128
47418 96.997 -79.887 46.105
47419 97.043 -79.884 46.081
47420 97.089 -79.882 46.058
47421 97.135 -79.880 46.035
47840 93.104 -79.569 53.126
47841 93.119 -79.559 53.108
47842 93.135 -79.551 53.091
47843 93.151 -79.542 53.076
47844 93.167 -79.535 53.061
47845 93.183 -79.526 53.047
47846 93.199 -79.519 53.034
47847 93.215 -79.510 53.021
47848 93.233 -79.499 53.009
47849 93.251 -79.485 52.998
47850 93.269 -79.471 52.989
47851 93.287 -79.460 52.980
47852 93.306 -79.454 52.972
47853 93.324 -79.448 52.966
47854 93.343 -79.442 52.960
47855 93.361 -79.436 52.955
47856 93.380 -79.430 52.950
47857 93.398 -79.424 52.943
47858 93.417 -79.418 52.937
47859 93.434 -79.412 52.931
47860 93.443 -79.407 52.921
47861 93.451 -79.402 52.911
47862 93.460 -79.397 52.899
47863 93.468 -79.391 52.887
47864 93.476 -79.386 52.875
47865 93.486 -79.382 52.862
47866 93.498 -79.377 52.849
47867 93.515 -79.373 52.836
47868 93.532 -79.369 52.822
47869 93.548 -79.365 52.808
47870 93.565 -79.360 52.795
47871 93.581 -79.357 52.784
47872 93.598 -79.352 52.774
47873 93.612 -79.347 52.764
47874 93.626 -79.337 52.758
47875 93.641 -79.327 52.753
47876 93.656 -79.316 52.749
47877 93.671 -79.306 52.742
47878 93.684 -79.294 52.736
47879 93.697 -79.285 52.729
47880 93.711 -79.284 52.722
47881 93.725 -79.283 52.715
47882 93.739 -79.282 52.707
47883 93.753 -79.281 52.699
47884 93.767 -79.281 52.692
47885 93.781 -79.281 52.685
47886 93.795 -79.280 52.679
47887 93.809 -79.279 52.672
47888 93.823 -79.278 52.665
47889 93.829 -79.280 52.657
47890 93.835 -79.278 52.649
47891 93.846 -79.272 52.641
47892 93.857 -79.267 52.631
47893 93.868 -79.262 52.621
47894 93.880 -79.256 52.610
47895 93.891 -79.250 52.598
47896 93.902 -79.245 52.586
47897 93.909 -79.245 52.573
47898 93.915 -79.246 52.559
47899 93.922 -79.247 52.545
47900 93.931 -79.245 52.531
47901 93.941 -79.244 52.514
47902 93.951 -79.243 52.495
47903 93.961 -79.241 52.476
47904 93.971 -79.241 52.460
47905 93.985 -79.239 52.445
47906 94.000 -79.237 52.429
47907 94.008 -79.238 52.411
47908 94.008 -79.241 52.391
47909 94.007 -79.244 52.370
47910 94.007 -79.246 52.348
47911 94.006 -79.248 52.328
47912 94.006 -79.251 52.309
47913 94.006 -79.252 52.294
47914 94.009 -79.251 52.277
47915 94.012 -79.250 52.259
47916 94.014 -79.249 52.240
47917 94.017 -79.248 52.223
47918 94.019 -79.247 52.208
47919 94.023 -79.246 52.194
47920 94.025 -79.244 52.180
47921 94.023 -79.234 52.165
47922 94.019 -79.228 52.150
47923 94.015 -79.220 52.137
47924 94.012 -79.213 52.127
47925 94.009 -79.206 52.119
47926 94.005 -79.200 52.112
47927 94.010 -79.195 52.106
47928 94.022 -79.194 52.097
47929 94.034 -79.194 52.088
47930 94.046 -79.192 52.079
47931 94.060 -79.193 52.067
47932 94.077 -79.197 52.054
47933 94.096 -79.202 52.040
47934 94.106 -79.201 52.023
47935 94.112 -79.197 52.003
47936 94.125 -79.203 51.980
47937 94.129 -79.215 51.956
47938 94.134 -79.226 51.932
47939 94.138 -79.237 51.908
47940 94.143 -79.249 51.883
47941 94.147 -79.260 51.861
47942 94.152 -79.272 51.840
47943 94.156 -79.282 51.821
47944 94.161 -79.294 51.804
47945 94.165 -79.293 51.788
47946 94.169 -79.279 51.772
47947 94.172 -79.266 51.756
47948 94.175 -79.253 51.738
47949 94.179 -79.240 51.719
47950 94.182 -79.226 51.698
47951 94.185 -79.213 51.677
47952 94.189 -79.200 51.655
47953 94.192 -79.187 51.632
47954 94.195 -79.173 51.609
47955 94.216 -79.143 51.583
47956 94.237 -79.114 51.557
47957 94.253 -79.093 51.530
47958 94.267 -79.089 51.503
47959 94.269 -79.136 51.475
47960 94.263 -79.210 51.446
47961 94.258 -79.212 51.413
47962 94.261 -79.204 51.374
47963 94.288 -79.187 51.334
47964 94.316 -79.170 51.295
47965 94.343 -79.153 51.256
47966 94.370 -79.137 51.217
47967 94.397 -79.120 51.178
47968 94.409 -79.115 51.138
47969 94.420 -79.112 51.099
47970 94.432 -79.109 51.060
47971 94.443 -79.105 51.020
47972 94.454 -79.101 50.979
47973 94.466 -79.097 50.940
48027 95.012 -79.596 48.990
48028 95.025 -79.610 48.952
48029 95.038 -79.623 48.915
48030 95.051 -79.636 48.879
48031 95.064 -79.648 48.841
48032 95.077 -79.661 48.804
48033 95.091 -79.674 48.767
48034 95.104 -79.687 48.731
48035 95.122 -79.695 48.694
48036 95.145 -79.696 48.657
48037 95.169 -79.698 48.620
48038 95.192 -79.699 48.583
48039 95.216 -79.701 48.547
48040 95.240 -79.702 48.511
48041 95.263 -79.704 48.475
48042 95.286 -79.705 48.440
48043 95.310 -79.707 48.403
48044 95.334 -79.708 48.367
48045 95.357 -79.710 48.332
48046 95.381 -79.711 48.297
48047 95.404 -79.713 48.262
48069 95.851 -79.581 47.504
48070 95.874 -79.594 47.472
48071 95.896 -79.607 47.440
48072 95.919 -79.620 47.409
48073 95.942 -79.633 47.376
48074 95.965 -79.646 47.345
48075 95.988 -79.659 47.313
48076 96.011 -79.673 47.281
48077 96.033 -79.679 47.251
48111 96.797 -79.836 46.298
48112 96.820 -79.841 46.273
48113 96.842 -79.846 46.250
48114 96.865 -79.851 46.226
48115 96.887 -79.855 46.202
48116 96.910 -79.860 46.177
48117 96.932 -79.864 46.152
48118 96.955 -79.869 46.128
48119 96.993 -79.869 46.105
48120 97.039 -79.866 46.081
48121 97.086 -79.864 46.058
48122 97.133 -79.861 46.035
48123 97.132 -79.873 46.013
48124 97.131 -79.884 45.989
48125 97.131 -79.897 45.965
48126 97.134 -79.926 45.941
48127 97.138 -79.954 45.920
48542 93.117 -79.546 53.133
48543 93.131 -79.534 53.115
48544 93.147 -79.526 53.100
48545 93.163 -79.518 53.085
48546 93.179 -79.510 53.071
48547 93.195 -79.501 53.058
48548 93.211 -79.494 53.045
48549 93.227 -79.485 53.033
48550 93.245 -79.473 53.022
48551 93.263 -79.459 53.012
48552 93.281 -79.445 53.003
48553 93.300 -79.438 52.995
48554 93.319 -79.432 52.988
48555 93.337 -79.426 52.983
48556 93.355 -79.420 52.979
48557 93.374 -79.414 52.972
48558 93.393 -79.408 52.966
48559 93.411 -79.402 52.959
48560 93.426 -79.397 52.951
48561 93.437 -79.392 52.941
48562 93.445 -79.387 52.930
48563 93.453 -79.382 52.919
48564 93.462 -79.377 52.909
48565 93.470 -79.372 52.897
48566 93.478 -79.366 52.885
48567 93.489 -79.361 52.872
48568 93.505 -79.357 52.860
48569 93.521 -79.352 52.846
48570 93.538 -79.347 52.831
48571 93.554 -79.344 52.817
48572 93.570 -79.339 52.804
48573 93.586 -79.335 52.794
48574 93.602 -79.329 52.786
48575 93.617 -79.319 52.779
48576 93.631 -79.309 52.774
48577 93.646 -79.298 52.770
48578 93.660 -79.288 52.763
48579 93.675 -79.280 52.756
48580 93.689 -79.279 52.749
48581 93.702 -79.279 52.740
48582 93.716 -79.279 52.732
48583 93.730 -79.279 52.724
48584 93.744 -79.279 52.715
48585 93.758 -79.279 52.707
48586 93.772 -79.279 52.699
48587 93.786 -79.279 52.692
48588 93.800 -79.278 52.685
48589 93.814 -79.278 52.677
48590 93.826 -79.275 52.669
48591 93.837 -79.270 52.661
48592 93.848 -79.265 52.652
48593 93.859 -79.259 52.643
48594 93.870 -79.254 52.633
48595 93.881 -79.249 52.622
48596 93.892 -79.244 52.610
48597 93.903 -79.238 52.598
48598 93.910 -79.237 52.585
48599 93.917 -79.238 52.571
48600 93.927 -79.237 52.557
48601 93.936 -79.236 52.541
48602 93.946 -79.234 52.523
48603 93.956 -79.234 52.504
48604 93.966 -79.232 52.485
48605 93.975 -79.231 52.471
48606 93.986 -79.230 52.455
48607 94.001 -79.228 52.437
48608 94.009 -79.229 52.419
48609 94.008 -79.232 52.398
48610 94.008 -79.234 52.377
48611 94.007 -79.237 52.354
48612 94.007 -79.239 52.333
48613 94.010 -79.237 52.314
48614 94.013 -79.237 52.297
48615 94.015 -79.235 52.281
48616 94.018 -79.234 52.264
48617 94.020 -79.234 52.246
48618 94.023 -79.232 52.230
48619 94.023 -79.232 52.217
48620 94.023 -79.231 52.204
48621 94.023 -79.231 52.191
48622 94.025 -79.234 52.176
48623 94.021 -79.228 52.162
48624 94.018 -79.221 52.149
48625 94.014 -79.214 52.139
48626 94.011 -79.206 52.131
48627 94.007 -79.200 52.124
48628 94.011 -79.196 52.118
48629 94.024 -79.194 52.109
48630 94.036 -79.194 52.100
48631 94.048 -79.192 52.091
48632 94.059 -79.191 52.079
48633 94.071 -79.190 52.066
48634 94.083 -79.189 52.051
48635 94.096 -79.188 52.034
48636 94.109 -79.189 52.012
48637 94.113 -79.200 51.989
48638 94.118 -79.212 51.963
48639 94.123 -79.223 51.938
48640 94.128 -79.234 51.913
48641 94.133 -79.247 51.889
48642 94.138 -79.258 51.867
48643 94.143 -79.269 51.846
48644 94.148 -79.281 51.827
48645 94.150 -79.289 51.810
48646 94.148 -79.288 51.795
48647 94.151 -79.273 51.780
48648 94.155 -79.257 51.764
48649 94.159 -79.241 51.746
48650 94.164 -79.225 51.727
48651 94.168 -79.209 51.705
48652 94.172 -79.194 51.683
48653 94.176 -79.179 51.661
48654 94.179 -79.165 51.638
48655 94.189 -79.146 51.614
48656 94.204 -79.131 51.587
48657 94.219 -79.115 51.561
48658 94.234 -79.100 51.534
48659 94.247 -79.100 51.510
48660 94.259 -79.118 51.483
48661 94.265 -79.156 51.454
48662 94.258 -79.208 51.420
48663 94.259 -79.201 51.382
48664 94.286 -79.184 51.342
48665 94.313 -79.168 51.302
48666 94.340 -79.151 51.264
48667 94.367 -79.134 51.225
48668 94.395 -79.117 51.186
48669 94.415 -79.106 51.148
48670 94.426 -79.103 51.110
48671 94.437 -79.100 51.071
48672 94.449 -79.096 51.031
48673 94.460 -79.092 50.991
48674 94.471 -79.088 50.953
48728 95.003 -79.557 48.994
48729 95.016 -79.570 48.956
48730 95.030 -79.583 48.920
48731 95.043 -79.596 48.883
48732 95.056 -79.609 48.846
48733 95.069 -79.622 48.808
48734 95.082 -79.635 48.772
48735 95.095 -79.648 48.736
48736 95.112 -79.657 48.699
48737 95.135 -79.659 48.662
48738 95.159 -79.661 48.626
48739 95.183 -79.662 48.589
48740 95.206 -79.664 48.552
48741 95.230 -79.664 48.516
48742 95.253 -79.666 48.480
48743 95.277 -79.667 48.444
48744 95.301 -79.669 48.408
48745 95.324 -79.670 48.372
48746 95.347 -79.672 48.337
48747 95.371 -79.673 48.301
48748 95.395 -79.675 48.265
48749 95.418 -79.676 48.230
48758 95.630 -79.690 47.913
48759 95.654 -79.692 47.879
48760 95.678 -79.693 47.845
48761 95.693 -79.673 47.810
48762 95.708 -79.651 47.776
48763 95.723 -79.630 47.742
48764 95.738 -79.609 47.709
48765 95.753 -79.588 47.675
48766 95.768 -79.566 47.641
48767 95.783 -79.545 47.608
48768 95.802 -79.540 47.575
48769 95.824 -79.553 47.542
48770 95.847 -79.566 47.509
48771 95.870 -79.579 47.477
48772 95.893 -79.592 47.445
48773 95.916 -79.605 47.413
48774 95.939 -79.618 47.380
48775 95.961 -79.631 47.348
48776 95.984 -79.645 47.317
48777 96.007 -79.656 47.285
48778 96.030 -79.661 47.254
48779 96.052 -79.665 47.224
48811 96.771 -79.814 46.324
48812 96.794 -79.818 46.298
48813 96.816 -79.823 46.273
48814 96.839 -79.828 46.250
48815 96.861 -79.833 46.226
48816 96.884 -79.837 46.201
48817 96.906 -79.842 46.177
48818 96.929 -79.846 46.152
48819 96.951 -79.851 46.128
48820 96.990 -79.851 46.104
48821 97.036 -79.849 46.081
48822 97.083 -79.845 46.058
48823 97.129 -79.843 46.035
48824 97.130 -79.854 46.013
48825 97.129 -79.865 45.989
48826 97.128 -79.877 45.965
48827 97.132 -79.906 45.942
48828 97.135 -79.935 45.920
48829 97.139 -79.964 45.899
49244 93.130 -79.523 53.141
49245 93.144 -79.511 53.125
49246 93.159 -79.501 53.110
49247 93.175 -79.493 53.096
49248 93.191 -79.485 53.083
49249 93.207 -79.476 53.071
49250 93.223 -79.469 53.059
49251 93.239 -79.460 53.046
49252 93.258 -79.447 53.036
49253 93.276 -79.432 53.027
49254 93.294 -79.423 53.018
49255 93.312 -79.417 53.012
49256 93.331 -79.411 53.007
49257 93.349 -79.405 53.002
49258 93.368 -79.399 52.994
49259 93.386 -79.393 52.987
49260 93.404 -79.387 52.979
49261 93.419 -79.382 52.970
49262 93.431 -79.378 52.959
49263 93.439 -79.372 52.949
49264 93.447 -79.367 52.938
49265 93.455 -79.362 52.927
49266 93.463 -79.357 52.916
49267 93.471 -79.352 52.904
49268 93.485 -79.346 52.891
49269 93.500 -79.341 52.878
49270 93.516 -79.337 52.864
49271 93.532 -79.332 52.849
49272 93.548 -79.328 52.836
49273 93.564 -79.323 52.826
49274 93.581 -79.319 52.817
49275 93.596 -79.314 52.810
49276 93.611 -79.304 52.802
49277 93.626 -79.294 52.796
49278 93.640 -79.284 52.790
49279 93.655 -79.273 52.784
49280 93.669 -79.267 52.777
49281 93.683 -79.263 52.769
49282 93.697 -79.263 52.759
49283 93.711 -79.263 52.750
49284 93.725 -79.263 52.741
49285 93.739 -79.263 52.733
49286 93.753 -79.263 52.724
49287 93.767 -79.263 52.716
49288 93.780 -79.263 52.707
49289 93.794 -79.263 52.700
49290 93.807 -79.264 52.692
49291 93.820 -79.263 52.684
49292 93.833 -79.259 52.675
49293 93.846 -79.256 52.666
49294 93.859 -79.252 52.657
49295 93.871 -79.247 52.647
49296 93.883 -79.241 52.636
49297 93.894 -79.236 52.625
49298 93.905 -79.231 52.613
49299 93.912 -79.231 52.599
49300 93.922 -79.229 52.585
49301 93.932 -79.228 52.569
49302 93.941 -79.227 52.552
49303 93.951 -79.225 52.534
49304 93.961 -79.225 52.515
49305 93.971 -79.223 52.498
49306 93.980 -79.222 52.483
49307 93.990 -79.221 52.467
49308 94.002 -79.219 52.449
49309 94.009 -79.220 52.429
49310 94.009 -79.223 52.408
49311 94.009 -79.225 52.386
49312 94.011 -79.225 52.364
49313 94.014 -79.223 52.344
49314 94.016 -79.222 52.324
49315 94.017 -79.222 52.307
49316 94.017 -79.221 52.290
49317 94.017 -79.220 52.274
49318 94.017 -79.220 52.258
49319 94.017 -79.219 52.243
49320 94.017 -79.219 52.229
49321 94.017 -79.219 52.216
49322 94.017 -79.218 52.203
49323 94.019 -79.222 52.189
49324 94.024 -79.228 52.176
49325 94.020 -79.221 52.164
49326 94.017 -79.214 52.155
49327 94.013 -79.207 52.147
49328 94.010 -79.200 52.139
49329 94.013 -79.196 52.131
49330 94.025 -79.194 52.123
49331 94.037 -79.194 52.113
49332 94.049 -79.193 52.102
49333 94.061 -79.191 52.091
49334 94.073 -79.190 52.077
49335 94.085 -79.189 52.062
49336 94.094 -79.187 52.044
49337 94.100 -79.188 52.022
49338 94.107 -79.200 51.998
49339 94.112 -79.211 51.971
49340 94.117 -79.223 51.945
49341 94.122 -79.234 51.919
49342 94.127 -79.246 51.896
49343 94.132 -79.257 51.875
49344 94.137 -79.269 51.855
49345 94.138 -79.276 51.835
49346 94.138 -79.282 51.818
49347 94.136 -79.280 51.803
49348 94.133 -79.277 51.788
49349 94.135 -79.265 51.772
49350 94.139 -79.249 51.754
49351 94.143 -79.233 51.734
49352 94.147 -79.217 51.711
49353 94.151 -79.201 51.689
49354 94.155 -79.184 51.666
49355 94.160 -79.169 51.642
49356 94.170 -79.152 51.618
49357 94.185 -79.137 51.591
49358 94.199 -79.121 51.565
49359 94.214 -79.106 51.539
49360 94.227 -79.112 51.515
49361 94.239 -79.130 51.489
49362 94.251 -79.147 51.460
49363 94.260 -79.175 51.427
49364 94.256 -79.198 51.390
49365 94.283 -79.181 51.351
49366 94.311 -79.165 51.311
49367 94.338 -79.148 51.273
49368 94.365 -79.131 51.235
49369 94.393 -79.115 51.197
49370 94.420 -79.097 51.159
49371 94.431 -79.093 51.121
49372 94.442 -79.090 51.083
49373 94.454 -79.087 51.043
49374 94.465 -79.083 51.004
49375 94.476 -79.079 50.966
49376 94.488 -79.075 50.928
49429 94.995 -79.518 48.999
49430 95.008 -79.531 48.961
49431 95.021 -79.544 48.924
49432 95.034 -79.557 48.887
49433 95.047 -79.570 48.850
49434 95.060 -79.582 48.814
49435 95.073 -79.596 48.777
49436 95.087 -79.609 48.741
49437 95.102 -79.620 48.704
49438 95.126 -79.621 48.668
49439 95.149 -79.623 48.632
49440 95.173 -79.624 48.596
49441 95.196 -79.626 48.559
49442 95.220 -79.627 48.522
49443 95.244 -79.629 48.486
49444 95.267 -79.630 48.449
49445 95.290 -79.632 48.413
49446 95.314 -79.633 48.377
49447 95.338 -79.634 48.341
49448 95.361 -79.636 48.305
49449 95.385 -79.637 48.269
49450 95.408 -79.639 48.234
49451 95.432 -79.640 48.199
49452 95.456 -79.642 48.164
49453 95.479 -79.643 48.129
49454 95.503 -79.645 48.093
49455 95.526 -79.646 48.057
49456 95.550 -79.648 48.022
49457 95.573 -79.649 47.987
49458 95.597 -79.651 47.952
49459 95.621 -79.652 47.918
49460 95.644 -79.654 47.883
49461 95.668 -79.655 47.849
49462 95.686 -79.642 47.815
49463 95.701 -79.621 47.781
49464 95.716 -79.600 47.748
49465 95.731 -79.579 47.714
49466 95.746 -79.557 47.680
49467 95.761 -79.536 47.647
49468 95.776 -79.515 47.614
49469 95.798 -79.525 47.580
49470 95.821 -79.538 47.547
49471 95.844 -79.551 47.514
49472 95.867 -79.564 47.481
49473 95.890 -79.577 47.450
49474 95.912 -79.590 47.418
49475 95.935 -79.603 47.385
49476 95.958 -79.617 47.353
49477 95.981 -79.629 47.321
49478 96.004 -79.638 47.289
49479 96.026 -79.643 47.258
49480 96.048 -79.648 47.228
49511 96.745 -79.791 46.349
49512 96.768 -79.796 46.323
49513 96.790 -79.801 46.297
49514 96.813 -79.805 46.273
49515 96.835 -79.810 46.249
49516 96.858 -79.814 46.225
49517 96.880 -79.819 46.201
49518 96.902 -79.823 46.177
49519 96.925 -79.828 46.152
49520 96.948 -79.833 46.128
49521 96.986 -79.833 46.104
49522 97.033 -79.830 46.081
49523 97.079 -79.827 46.058
49524 97.125 -79.825 46.035
49525 97.127 -79.836 46.013
49526 97.126 -79.847 45.989
49527 97.125 -79.858 45.965
49528 97.129 -79.887 45.943
49529 97.133 -79.915 45.921
49530 97.136 -79.944 45.899
49531 97.140 -79.973 45.878
49532 97.143 -80.002 45.855
49945 93.129 -79.513 53.166
49946 93.143 -79.501 53.151
49947 93.157 -79.489 53.137
49948 93.171 -79.477 53.122
49949 93.187 -79.468 53.108
49950 93.203 -79.460 53.096
49951 93.219 -79.451 53.084
49952 93.235 -79.444 53.072
49953 93.252 -79.434 53.061
49954 93.270 -79.420 53.052
49955 93.288 -79.407 53.042
49956 93.306 -79.401 53.036
49957 93.325 -79.395 53.029
49958 93.344 -79.389 53.022
49959 93.362 -79.383 53.015
49960 93.380 -79.377 53.007
49961 93.396 -79.372 52.998
49962 93.411 -79.368 52.988
49963 93.424 -79.363 52.977
49964 93.432 -79.358 52.966
49965 93.440 -79.353 52.955
49966 93.448 -79.347 52.944
49967 93.457 -79.342 52.933
49968 93.466 -79.337 52.921
49969 93.481 -79.332 52.908
49970 93.495 -79.326 52.894
49971 93.510 -79.321 52.880
49972 93.526 -79.316 52.867
49973 93.543 -79.313 52.855
49974 93.559 -79.308 52.846
49975 93.575 -79.303 52.838
49976 93.591 -79.300 52.832
49977 93.606 -79.289 52.824
49978 93.621 -79.279 52.816
49979 93.635 -79.269 52.810
49980 93.649 -79.259 52.804
49981 93.663 -79.254 52.796
49982 93.678 -79.249 52.787
49983 93.692 -79.246 52.778
49984 93.706 -79.246 52.769
49985 93.719 -79.246 52.760
49986 93.733 -79.246 52.751
49987 93.747 -79.246 52.743
49988 93.761 -79.246 52.735
49989 93.775 -79.246 52.726
49990 93.789 -79.246 52.718
49991 93.799 -79.253 52.709
49992 93.813 -79.250 52.701
49993 93.826 -79.246 52.692
49994 93.839 -79.242 52.683
49995 93.852 -79.238 52.674
49996 93.865 -79.234 52.664
49997 93.878 -79.230 52.654
49998 93.892 -79.226 52.643
49999 93.905 -79.222 52.631
50000 93.915 -79.221 52.616
50001 93.925 -79.220 52.600
50002 93.936 -79.219 52.582
50003 93.946 -79.218 52.565
50004 93.956 -79.216 52.547
50005 93.966 -79.216 52.529
50006 93.975 -79.214 52.513
50007 93.985 -79.213 52.497
50008 93.995 -79.212 52.481
50009 94.004 -79.210 52.464
50010 94.010 -79.211 52.442
50011 94.012 -79.211 52.421
50012 94.013 -79.210 52.399
50013 94.013 -79.210 52.377
50014 94.013 -79.209 52.357
50015 94.013 -79.209 52.338
50016 94.013 -79.208 52.320
50017 94.013 -79.208 52.303
50018 94.013 -79.207 52.287
50019 94.013 -79.206 52.272
50020 94.012 -79.206 52.258
50021 94.012 -79.206 52.245
50022 94.012 -79.205 52.232
50023 94.012 -79.205 52.219
50024 94.014 -79.206 52.205
50025 94.018 -79.213 52.193
50026 94.022 -79.220 52.182
50027 94.019 -79.215 52.174
50028 94.015 -79.207 52.165
50029 94.012 -79.200 52.157
50030 94.014 -79.196 52.149
50031 94.027 -79.195 52.139
50032 94.039 -79.194 52.128
50033 94.051 -79.193 52.116
50034 94.063 -79.191 52.102
50035 94.074 -79.190 52.088
50036 94.080 -79.187 52.072
50037 94.085 -79.184 52.053
50038 94.092 -79.187 52.031
50039 94.099 -79.200 52.007
50040 94.106 -79.211 51.980
50041 94.111 -79.222 51.953
50042 94.116 -79.234 51.929
50043 94.121 -79.245 51.906
50044 94.126 -79.257 51.885
50045 94.126 -79.263 51.864
50046 94.126 -79.269 51.845
50047 94.126 -79.275 51.827
50048 94.124 -79.272 51.811
50049 94.121 -79.269 51.795
50050 94.119 -79.266 51.779
50051 94.119 -79.256 51.761
50052 94.123 -79.241 51.740
50053 94.127 -79.224 51.717
50054 94.131 -79.208 51.694
50055 94.135 -79.192 51.671
50056 94.143 -79.175 51.646
50057 94.153 -79.159 51.620
50058 94.165 -79.142 51.595
50059 94.180 -79.127 51.570
50060 94.194 -79.112 51.545
50061 94.207 -79.125 51.520
50062 94.219 -79.142 51.494
50063 94.231 -79.159 51.466
50064 94.243 -79.177 51.435
50065 94.255 -79.195 51.399
50066 94.281 -79.179 51.361
50067 94.309 -79.162 51.321
50068 94.336 -79.145 51.284
50069 94.363 -79.128 51.246
50070 94.390 -79.112 51.208
50071 94.417 -79.095 51.170
50072 94.436 -79.084 51.133
50073 94.448 -79.081 51.095
50074 94.459 -79.078 51.056
50075 94.470 -79.074 51.018
50076 94.482 -79.070 50.980
50077 94.493 -79.066 50.941
50078 94.504 -79.062 50.904
50130 94.986 -79.479 49.004
50131 94.999 -79.491 48.967
50132 95.013 -79.504 48.930
50133 95.026 -79.517 48.894
50134 95.039 -79.530 48.857
50135 95.052 -79.544 48.820
50136 95.065 -79.557 48.783
50137 95.078 -79.570 48.746
50138 95.092 -79.582 48.710
50139 95.116 -79.583 48.675
50140 95.139 -79.585 48.640
50141 95.163 -79.586 48.604
50142 95.187 -79.588 48.565
50143 95.210 -79.589 48.528
50144 95.234 -79.591 48.492
50145 95.257 -79.592 48.455
50146 95.281 -79.594 48.419
50147 95.305 -79.595 48.383
50148 95.328 -79.597 48.346
50149 95.351 -79.598 48.310
50150 95.375 -79.600 48.274
50151 95.399 -79.601 48.238
50152 95.422 -79.603 48.203
50153 95.446 -79.604 48.168
50154 95.469 -79.605 48.133
50155 95.493 -79.607 48.097
50156 95.517 -79.608 48.062
50157 95.540 -79.610 48.027
50158 95.564 -79.611 47.992
50159 95.587 -79.613 47.957
50160 95.611 -79.614 47.922
50161 95.634 -79.616 47.888
50162 95.658 -79.617 47.854
50163 95.679 -79.612 47.821
50164 95.694 -79.591 47.787
50165 95.709 -79.570 47.754
50166 95.724 -79.548 47.720
50167 95.739 -79.527 47.687
50168 95.754 -79.506 47.653
50169 95.772 -79.497 47.619
50170 95.795 -79.510 47.586
50171 95.818 -79.523 47.553
50172 95.840 -79.536 47.520
50173 95.863 -79.549 47.487
50174 95.886 -79.562 47.456
50175 95.909 -79.575 47.424
50176 95.932 -79.588 47.390
50177 95.955 -79.601 47.358
50178 95.977 -79.614 47.326
50179 96.000 -79.620 47.294
50180 96.022 -79.625 47.263
50181 96.045 -79.629 47.231
50182 96.067 -79.634 47.200
50211 96.719 -79.769 46.375
50212 96.742 -79.773 46.350
50213 96.764 -79.778 46.324
50214 96.787 -79.783 46.298
50215 96.809 -79.787 46.272
50216 96.831 -79.792 46.249
50217 96.854 -79.796 46.225
50218 96.877 -79.801 46.201
50219 96.899 -79.806 46.176
50220 96.922 -79.811 46.152
50221 96.944 -79.815 46.127
50222 96.982 -79.814 46.104
50223 97.029 -79.812 46.081
50224 97.076 -79.809 46.058
50225 97.122 -79.807 46.035
50226 97.125 -79.817 46.013
50227 97.124 -79.829 45.989
50228 97.123 -79.840 45.965
50229 97.126 -79.867 45.942
50230 97.130 -79.896 45.920
50231 97.134 -79.925 45.899
50232 97.137 -79.953 45.877
50233 97.141 -79.982 45.855
50234 97.144 -80.010 45.834
50647 93.142 -79.491 53.178
50648 93.156 -79.479 53.163
50649 93.170 -79.466 53.149
50650 93.185 -79.454 53.135
50651 93.199 -79.443 53.122
50652 93.215 -79.435 53.110
50653 93.231 -79.426 53.098
50654 93.247 -79.419 53.087
50655 93.264 -79.408 53.077
50656 93.282 -79.394 53.067
50657 93.300 -79.385 53.059
50658 93.319 -79.379 53.051
50659 93.338 -79.374 53.044
50660 93.356 -79.368 53.036
50661 93.373 -79.362 53.027
50662 93.388 -79.357 53.018
50663 93.403 -79.353 53.007
50664 93.418 -79.349 52.996
50665 93.426 -79.344 52.985
50666 93.434 -79.338 52.972
50667 93.442 -79.333 52.962
50668 93.450 -79.328 52.950
50669 93.462 -79.322 52.937
50670 93.476 -79.317 52.924
50671 93.491 -79.312 52.910
50672 93.506 -79.306 52.896
50673 93.521 -79.301 52.883
50674 93.538 -79.297 52.874
50675 93.554 -79.293 52.866
50676 93.570 -79.288 52.859
50677 93.586 -79.284 52.852
50678 93.601 -79.274 52.844
50679 93.615 -79.264 52.837
50680 93.630 -79.254 52.829
50681 93.644 -79.247 52.822
50682 93.658 -79.241 52.814
50683 93.672 -79.236 52.805
50684 93.686 -79.231 52.796
50685 93.700 -79.229 52.787
50686 93.714 -79.229 52.779
50687 93.728 -79.229 52.770
50688 93.742 -79.229 52.762
50689 93.756 -79.229 52.754
50690 93.770 -79.229 52.745
50691 93.780 -79.237 52.736
50692 93.792 -79.240 52.728
50693 93.805 -79.236 52.719
50694 93.819 -79.232 52.710
50695 93.832 -79.228 52.701
50696 93.845 -79.224 52.691
50697 93.858 -79.220 52.683
50698 93.871 -79.216 52.672
50699 93.885 -79.213 52.660
50700 93.899 -79.213 52.647
50701 93.899 -79.210 52.630
50702 93.910 -79.209 52.613
50703 93.921 -79.209 52.596
50704 93.931 -79.207 52.579
50705 93.942 -79.206 52.562
50706 93.954 -79.205 52.546
50707 93.964 -79.204 52.530
50708 93.975 -79.203 52.515
50709 93.986 -79.202 52.498
50710 93.997 -79.200 52.481
50711 94.007 -79.203 52.460
50712 94.008 -79.200 52.438
50713 94.008 -79.197 52.417
50714 94.008 -79.197 52.395
50715 94.008 -79.196 52.375
50716 94.008 -79.196 52.355
50717 94.008 -79.195 52.337
50718 94.008 -79.194 52.321
50719 94.008 -79.194 52.305
50720 94.008 -79.194 52.290
50721 94.008 -79.193 52.275
50722 94.008 -79.193 52.262
50723 94.008 -79.192 52.250
50724 94.008 -79.191 52.238
50725 94.008 -79.191 52.224
50726 94.012 -79.198 52.212
50727 94.016 -79.205 52.202
50728 94.020 -79.212 52.194
50729 94.018 -79.208 52.185
50730 94.014 -79.201 52.176
50731 94.016 -79.196 52.168
50732 94.028 -79.195 52.157
50733 94.040 -79.194 52.144
50734 94.052 -79.193 52.130
50735 94.060 -79.190 52.114
50736 94.066 -79.187 52.098
50737 94.071 -79.184 52.081
50738 94.077 -79.181 52.062
50739 94.084 -79.187 52.040
50740 94.091 -79.199 52.015
50741 94.098 -79.210 51.989
50742 94.105 -79.222 51.963
50743 94.110 -79.234 51.939
50744 94.114 -79.244 51.918
50745 94.114 -79.250 51.898
50746 94.114 -79.256 51.878
50747 94.114 -79.261 51.860
50748 94.114 -79.266 51.842
50749 94.111 -79.263 51.824
50750 94.109 -79.260 51.807
50751 94.106 -79.257 51.789
50752 94.104 -79.254 51.769
50753 94.103 -79.247 51.746
50754 94.107 -79.231 51.722
50755 94.111 -79.216 51.698
50756 94.117 -79.200 51.673
50757 94.127 -79.182 51.648
50758 94.137 -79.165 51.623
50759 94.147 -79.149 51.599
50760 94.160 -79.133 51.575
50761 94.175 -79.118 51.551
50762 94.187 -79.136 51.525
50763 94.199 -79.153 51.499
50764 94.211 -79.172 51.473
50765 94.235 -79.179 51.442
50766 94.260 -79.175 51.408
50767 94.280 -79.171 51.372
50768 94.306 -79.159 51.333
50769 94.333 -79.143 51.295
50770 94.360 -79.126 51.257
50771 94.388 -79.109 51.219
50772 94.415 -79.092 51.182
50773 94.442 -79.075 51.145
50774 94.453 -79.072 51.107
50775 94.465 -79.068 51.069
50776 94.476 -79.065 51.031
50777 94.487 -79.061 50.992
50778 94.498 -79.057 50.955
50779 94.509 -79.053 50.917
50780 94.521 -79.049 50.880
50781 94.532 -79.046 50.843
50829 94.952 -79.413 49.083
50830 94.965 -79.426 49.046
50831 94.978 -79.439 49.010
50832 94.991 -79.452 48.973
50833 95.004 -79.465 48.937
50834 95.017 -79.478 48.900
50835 95.031 -79.491 48.863
50836 95.044 -79.504 48.826
50837 95.057 -79.517 48.789
50838 95.070 -79.530 48.753
50839 95.083 -79.543 48.717
50840 95.106 -79.545 48.683
50841 95.130 -79.547 48.648
50842 95.153 -79.548 48.614
50843 95.177 -79.550 48.575
50844 95.200 -79.551 48.537
50845 95.224 -79.553 48.499
50846 95.248 -79.554 48.461
50847 95.271 -79.556 48.425
50848 95.295 -79.557 48.388
50849 95.318 -79.559 48.352
50850 95.342 -79.560 48.315
50851 95.366 -79.562 48.279
50852 95.389 -79.563 48.243
50853 95.412 -79.565 48.209
50854 95.436 -79.567 48.173
50855 95.460 -79.568 48.138
50856 95.483 -79.570 48.102
50857 95.507 -79.571 48.067
50858 95.530 -79.573 48.032
50859 95.554 -79.574 47.997
50860 95.578 -79.575 47.962
50861 95.601 -79.576 47.928
50862 95.625 -79.578 47.894
50863 95.648 -79.579 47.860
50864 95.672 -79.581 47.827
50865 95.687 -79.561 47.793
50866 95.702 -79.540 47.760
50867 95.717 -79.518 47.727
50868 95.732 -79.497 47.693
50869 95.747 -79.476 47.660
50870 95.768 -79.482 47.626
50871 95.792 -79.495 47.593
50872 95.814 -79.508 47.560
50873 95.837 -79.521 47.527
50874 95.860 -79.534 47.494
50875 95.883 -79.548 47.462
50876 95.905 -79.560 47.430
50877 95.928 -79.573 47.397
50878 95.951 -79.586 47.364
50879 95.974 -79.598 47.331
50880 95.996 -79.602 47.299
50881 96.019 -79.607 47.268
50882 96.041 -79.611 47.236
50883 96.063 -79.616 47.205
50884 96.086 -79.620 47.174
50911 96.693 -79.746 46.401
50912 96.716 -79.751 46.375
50913 96.738 -79.755 46.350
50914 96.760 -79.760 46.324
50915 96.783 -79.764 46.298
50916 96.806 -79.769 46.273
50917 96.828 -79.773 46.249
50918 96.851 -79.779 46.226
50919 96.873 -79.783 46.201
50920 96.895 -79.788 46.177
50921 96.918 -79.792 46.152
50922 96.940 -79.797 46.127
50923 96.979 -79.797 46.104
50924 97.026 -79.794 46.082
50925 97.072 -79.792 46.059
50926 97.118 -79.789 46.036
50927 97.123 -79.798 46.013
50928 97.122 -79.811 45.990
50929 97.121 -79.822 45.966
50930 97.124 -79.849 45.942
50931 97.127 -79.877 45.921
50932 97.131 -79.905 45.899
50933 97.135 -79.934 45.877
50934 97.138 -79.963 45.855
50935 97.145 -79.975 45.834
50936 97.152 -79.983 45.812
51348 93.141 -79.481 53.204
51349 93.155 -79.469 53.190
51350 93.169 -79.456 53.176
51351 93.184 -79.444 53.163
51352 93.198 -79.432 53.149
51353 93.212 -79.419 53.136
51354 93.227 -79.410 53.124
51355 93.243 -79.401 53.113
51356 93.259 -79.394 53.103
51357 93.276 -79.382 53.093
51358 93.295 -79.370 53.083
51359 93.313 -79.364 53.074
51360 93.332 -79.358 53.066
51361 93.350 -79.352 53.057
51362 93.365 -79.347 53.047
51363 93.380 -79.343 53.037
51364 93.395 -79.338 53.026
51365 93.410 -79.334 53.015
51366 93.420 -79.328 53.002
51367 93.428 -79.324 52.990
51368 93.436 -79.319 52.979
51369 93.444 -79.313 52.967
51370 93.457 -79.308 52.954
51371 93.472 -79.303 52.942
51372 93.487 -79.297 52.928
51373 93.502 -79.291 52.915
51374 93.516 -79.286 52.904
51375 93.532 -79.281 52.894
51376 93.548 -79.277 52.886
51377 93.564 -79.272 52.879
51378 93.580 -79.269 52.872
51379 93.595 -79.259 52.865
51380 93.610 -79.249 52.857
51381 93.625 -79.239 52.848
51382 93.638 -79.234 52.839
51383 93.652 -79.228 52.830
51384 93.666 -79.223 52.822
51385 93.681 -79.218 52.813
51386 93.695 -79.212 52.805
51387 93.709 -79.212 52.798
51388 93.722 -79.212 52.789
51389 93.736 -79.212 52.780
51390 93.751 -79.212 52.772
51391 93.762 -79.219 52.764
51392 93.772 -79.228 52.755
51393 93.785 -79.226 52.747
51394 93.798 -79.222 52.738
51395 93.811 -79.219 52.728
51396 93.824 -79.215 52.719
51397 93.837 -79.210 52.709
51398 93.850 -79.206 52.701
51399 93.864 -79.204 52.689
51400 93.878 -79.204 52.676
51401 93.893 -79.204 52.663
51402 93.890 -79.201 52.645
51403 93.894 -79.199 52.628
51404 93.905 -79.198 52.611
51405 93.916 -79.197 52.596
51406 93.926 -79.196 52.580
51407 93.937 -79.194 52.565
51408 93.948 -79.194 52.549
51409 93.960 -79.192 52.532
51410 93.970 -79.191 52.515
51411 93.990 -79.192 52.496
51412 93.997 -79.195 52.473
51413 93.999 -79.199 52.451
51414 94.001 -79.198 52.429
51415 94.002 -79.192 52.407
51416 94.003 -79.186 52.388
51417 94.003 -79.182 52.370
51418 94.003 -79.182 52.355
51419 94.003 -79.181 52.339
51420 94.003 -79.181 52.323
51421 94.003 -79.181 52.308
51422 94.003 -79.180 52.295
51423 94.003 -79.179 52.283
51424 94.003 -79.178 52.271
51425 94.003 -79.178 52.260
51426 94.003 -79.178 52.247
51427 94.006 -79.183 52.235
51428 94.010 -79.190 52.224
51429 94.014 -79.197 52.216
51430 94.018 -79.203 52.207
51431 94.017 -79.201 52.197
51432 94.018 -79.196 52.188
51433 94.030 -79.195 52.176
51434 94.041 -79.194 52.162
51435 94.046 -79.190 52.145
51436 94.052 -79.187 52.127
51437 94.057 -79.184 52.109
51438 94.063 -79.181 52.090
51439 94.068 -79.177 52.070
51440 94.075 -79.187 52.048
51441 94.083 -79.199 52.025
51442 94.090 -79.210 51.999
51443 94.097 -79.222 51.975
51444 94.102 -79.231 51.952
51445 94.102 -79.237 51.931
51446 94.102 -79.242 51.912
51447 94.102 -79.248 51.894
51448 94.102 -79.254 51.876
51449 94.102 -79.258 51.857
51450 94.099 -79.255 51.838
51451 94.097 -79.252 51.819
51452 94.094 -79.249 51.799
51453 94.091 -79.246 51.777
51454 94.089 -79.243 51.753
51455 94.086 -79.239 51.727
51456 94.091 -79.223 51.702
51457 94.101 -79.206 51.676
51458 94.111 -79.190 51.651
51459 94.121 -79.172 51.627
51460 94.130 -79.156 51.604
51461 94.140 -79.139 51.581
51462 94.154 -79.131 51.558
51463 94.167 -79.148 51.534
51464 94.188 -79.158 51.508
51465 94.216 -79.163 51.483
51466 94.243 -79.169 51.453
51467 94.266 -79.153 51.420
51468 94.283 -79.128 51.385
51469 94.306 -79.147 51.346
51470 94.331 -79.140 51.308
51471 94.358 -79.123 51.270
51472 94.386 -79.106 51.232
51473 94.413 -79.090 51.195
51474 94.440 -79.072 51.158
51475 94.459 -79.063 51.120
51476 94.470 -79.059 51.081
51477 94.481 -79.056 51.043
51478 94.492 -79.052 51.005
51479 94.504 -79.048 50.967
51480 94.515 -79.044 50.930
51481 94.526 -79.040 50.893
51482 94.538 -79.037 50.856
51483 94.549 -79.034 50.819
51529 94.933 -79.365 49.126
51530 94.943 -79.374 49.089
51531 94.956 -79.387 49.052
51532 94.969 -79.400 49.015
51533 94.983 -79.413 48.979
51534 94.996 -79.426 48.943
51535 95.009 -79.439 48.907
51536 95.022 -79.452 48.870
51537 95.035 -79.465 48.833
51538 95.048 -79.478 48.797
51539 95.061 -79.491 48.760
51540 95.074 -79.504 48.724
51541 95.096 -79.507 48.690
51542 95.120 -79.509 48.657
51543 95.143 -79.510 48.624
51544 95.167 -79.512 48.586
51545 95.191 -79.513 48.546
51546 95.214 -79.515 48.506
51547 95.238 -79.516 48.468
51548 95.261 -79.518 48.430
51549 95.285 -79.520 48.394
51550 95.309 -79.521 48.357
51551 95.332 -79.523 48.321
51552 95.355 -79.524 48.285
51553 95.379 -79.526 48.249
51554 95.403 -79.527 48.214
51555 95.426 -79.529 48.178
51556 95.450 -79.530 48.143
51557 95.473 -79.532 48.107
51558 95.497 -79.533 48.072
51559 95.521 -79.535 48.037
51560 95.544 -79.536 48.002
51561 95.568 -79.538 47.968
51562 95.591 -79.539 47.934
51563 95.615 -79.541 47.900
51564 95.638 -79.542 47.867
51565 95.662 -79.544 47.833
51566 95.680 -79.531 47.800
51567 95.695 -79.510 47.767
51568 95.710 -79.488 47.734
51569 95.726 -79.467 47.701
51570 95.742 -79.454 47.667
51571 95.765 -79.467 47.634
51572 95.788 -79.480 47.601
51573 95.811 -79.493 47.568
51574 95.833 -79.506 47.534
51575 95.856 -79.520 47.502
51576 95.879 -79.532 47.469
51577 95.902 -79.545 47.436
51578 95.925 -79.558 47.403
51579 95.948 -79.571 47.370
51580 95.970 -79.579 47.337
51581 95.992 -79.584 47.304
51582 96.015 -79.588 47.273
51583 96.038 -79.593 47.241
51584 96.060 -79.598 47.210
51585 96.083 -79.603 47.178
51611 96.667 -79.723 46.426
51612 96.689 -79.728 46.401
51613 96.712 -79.733 46.376
51614 96.735 -79.737 46.350
51615 96.757 -79.742 46.323
51616 96.780 -79.746 46.299
51617 96.802 -79.751 46.274
51618 96.824 -79.756 46.250
51619 96.847 -79.761 46.226
51620 96.869 -79.765 46.202
51621 96.892 -79.770 46.177
51622 96.914 -79.774 46.152
51623 96.937 -79.779 46.128
51624 96.976 -79.779 46.105
51625 97.022 -79.776 46.082
51626 97.069 -79.773 46.060
51627 97.115 -79.770 46.037
51628 97.120 -79.780 46.013
51629 97.120 -79.792 45.990
51630 97.119 -79.803 45.966
51631 97.122 -79.829 45.943
51632 97.125 -79.858 45.922
51633 97.129 -79.886 45.899
51634 97.132 -79.915 45.878
51635 97.138 -79.933 45.856
51636 97.145 -79.941 45.834
51637 97.152 -79.949 45.812
51638 97.159 -79.956 45.791
51639 97.166 -79.964 45.771
52050 93.154 -79.458 53.217
52051 93.168 -79.446 53.203
52052 93.183 -79.434 53.190
52053 93.197 -79.422 53.177
52054 93.211 -79.410 53.164
52055 93.225 -79.397 53.151
52056 93.239 -79.385 53.140
52057 93.255 -79.376 53.129
52058 93.271 -79.369 53.117
52059 93.289 -79.356 53.107
52060 93.307 -79.348 53.097
52061 93.326 -79.342 53.088
52062 93.342 -79.338 53.077
52063 93.357 -79.333 53.067
52064 93.372 -79.328 53.057
52065 93.387 -79.324 53.045
52066 93.402 -79.319 53.033
52067 93.413 -79.314 53.020
52068 93.421 -79.309 53.008
52069 93.429 -79.303 52.996
52070 93.438 -79.299 52.984
52071 93.453 -79.293 52.972
52072 93.468 -79.288 52.959
52073 93.483 -79.282 52.946
52074 93.497 -79.277 52.934
52075 93.512 -79.272 52.923
52076 93.527 -79.266 52.915
52077 93.543 -79.262 52.907
52078 93.559 -79.257 52.899
52079 93.575 -79.253 52.893
52080 93.590 -79.244 52.885
52081 93.605 -79.234 52.877
52082 93.619 -79.227 52.868
52083 93.633 -79.222 52.858
52084 93.647 -79.216 52.848
52085 93.661 -79.210 52.839
52086 93.675 -79.205 52.831
52087 93.689 -79.200 52.824
52088 93.703 -79.196 52.817
52089 93.717 -79.196 52.809
52090 93.731 -79.196 52.800
52091 93.743 -79.200 52.792
52092 93.753 -79.209 52.783
52093 93.764 -79.216 52.775
52094 93.777 -79.212 52.767
52095 93.790 -79.209 52.757
52096 93.803 -79.205 52.748
52097 93.817 -79.201 52.738
52098 93.830 -79.197 52.730
52099 93.844 -79.196 52.720
52100 93.858 -79.195 52.707
52101 93.873 -79.195 52.693
52102 93.888 -79.195 52.679
52103 93.884 -79.192 52.661
52104 93.878 -79.188 52.644
52105 93.889 -79.187 52.629
52106 93.900 -79.186 52.614
52107 93.911 -79.185 52.600
52108 93.922 -79.184 52.585
52109 93.932 -79.183 52.569
52110 93.943 -79.181 52.551
52111 93.962 -79.182 52.531
52112 93.981 -79.184 52.510
52113 93.988 -79.186 52.486
52114 93.990 -79.190 52.463
52115 93.992 -79.195 52.441
52116 93.994 -79.197 52.422
52117 93.995 -79.190 52.405
52118 93.996 -79.184 52.389
52119 93.997 -79.178 52.376
52120 93.997 -79.172 52.360
52121 93.998 -79.168 52.345
52122 93.998 -79.167 52.330
52123 93.998 -79.166 52.317
52124 93.998 -79.166 52.306
52125 93.998 -79.165 52.295
52126 93.998 -79.165 52.284
52127 93.998 -79.165 52.272
52128 94.000 -79.169 52.260
52129 94.004 -79.175 52.249
52130 94.008 -79.182 52.241
52131 94.012 -79.189 52.232
52132 94.016 -79.196 52.222
52133 94.019 -79.196 52.211
52134 94.027 -79.194 52.197
52135 94.032 -79.190 52.180
52136 94.038 -79.187 52.161
52137 94.043 -79.184 52.142
52138 94.049 -79.180 52.122
52139 94.054 -79.177 52.102
52140 94.060 -79.175 52.081
52141 94.067 -79.187 52.060
52142 94.074 -79.198 52.036
52143 94.082 -79.209 52.013
52144 94.086 -79.218 51.990
52145 94.089 -79.225 51.969
52146 94.089 -79.230 51.949
52147 94.089 -79.235 51.931
52148 94.089 -79.241 51.913
52149 94.090 -79.247 51.894
52150 94.089 -79.250 51.874
52151 94.086 -79.247 51.853
52152 94.084 -79.244 51.832
52153 94.082 -79.241 51.809
52154 94.079 -79.237 51.784
52155 94.077 -79.234 51.759
52156 94.077 -79.231 51.732
52157 94.081 -79.223 51.706
52158 94.090 -79.207 51.680
52159 94.099 -79.191 51.655
52160 94.108 -79.175 51.631
52161 94.117 -79.159 51.610
52162 94.126 -79.144 51.589
52163 94.142 -79.136 51.568
52164 94.170 -79.142 51.545
52165 94.197 -79.147 51.521
52166 94.224 -79.153 51.496
52167 94.252 -79.159 51.466
52168 94.269 -79.131 51.434
52169 94.282 -79.103 51.400
52170 94.308 -79.107 51.361
52171 94.332 -79.122 51.322
52172 94.356 -79.120 51.284
52173 94.383 -79.103 51.245
52174 94.410 -79.087 51.208
52175 94.437 -79.070 51.170
52176 94.464 -79.053 51.133
52177 94.475 -79.050 51.094
52178 94.486 -79.046 51.055
52179 94.498 -79.043 51.018
52180 94.509 -79.039 50.980
52181 94.521 -79.035 50.943
52182 94.532 -79.031 50.906
52183 94.543 -79.028 50.869
52184 94.554 -79.024 50.832
52185 94.565 -79.021 50.795
52186 94.577 -79.017 50.757
52228 94.925 -79.329 49.206
52229 94.933 -79.337 49.169
52230 94.942 -79.344 49.132
52231 94.950 -79.352 49.095
52232 94.958 -79.360 49.059
52233 94.967 -79.366 49.023
52234 94.975 -79.374 48.987
52235 94.987 -79.387 48.950
52236 95.000 -79.400 48.914
52237 95.014 -79.413 48.878
52238 95.027 -79.426 48.842
52239 95.040 -79.438 48.805
52240 95.053 -79.451 48.768
52241 95.066 -79.464 48.732
52242 95.087 -79.469 48.697
52243 95.110 -79.471 48.664
52244 95.133 -79.473 48.630
52245 95.157 -79.474 48.591
52246 95.181 -79.476 48.552
52247 95.204 -79.477 48.513
52248 95.228 -79.479 48.475
52249 95.252 -79.480 48.438
52250 95.275 -79.482 48.401
52251 95.299 -79.483 48.364
52252 95.322 -79.485 48.327
52253 95.346 -79.486 48.291
52254 95.370 -79.488 48.255
52255 95.393 -79.489 48.219
52256 95.416 -79.491 48.184
52257 95.440 -79.492 48.148
52258 95.464 -79.494 48.112
52259 95.487 -79.495 48.077
52260 95.511 -79.497 48.042
52261 95.534 -79.498 48.007
52262 95.558 -79.500 47.974
52263 95.582 -79.501 47.940
52264 95.605 -79.503 47.907
52265 95.629 -79.504 47.874
52266 95.652 -79.506 47.841
52267 95.673 -79.501 47.808
52268 95.688 -79.479 47.775
52269 95.703 -79.458 47.741
52270 95.719 -79.437 47.708
52271 95.739 -79.439 47.674
52272 95.761 -79.452 47.641
52273 95.784 -79.465 47.609
52274 95.807 -79.478 47.575
52275 95.830 -79.491 47.542
52276 95.853 -79.504 47.509
52277 95.876 -79.517 47.475
52278 95.898 -79.530 47.442
52279 95.921 -79.543 47.410
52280 95.944 -79.557 47.376
52281 95.967 -79.561 47.343
52282 95.989 -79.566 47.310
52283 96.012 -79.570 47.277
52284 96.034 -79.576 47.245
52285 96.056 -79.580 47.214
52286 96.079 -79.585 47.182
52287 96.102 -79.589 47.150
52308 96.574 -79.687 46.532
52309 96.596 -79.692 46.505
52310 96.618 -79.696 46.477
52311 96.641 -79.701 46.452
52312 96.664 -79.705 46.427
52313 96.686 -79.710 46.403
52314 96.709 -79.714 46.376
52315 96.731 -79.719 46.350
52316 96.753 -79.723 46.325
52317 96.776 -79.729 46.300
52318 96.798 -79.733 46.276
52319 96.821 -79.738 46.252
52320 96.843 -79.742 46.227
52321 96.866 -79.747 46.203
52322 96.888 -79.751 46.179
52323 96.911 -79.756 46.154
52324 96.933 -79.761 46.130
52325 96.972 -79.761 46.106
52326 97.019 -79.758 46.084
52327 97.065 -79.755 46.062
52328 97.112 -79.752 46.039
52329 97.118 -79.762 46.016
52330 97.117 -79.773 45.992
52331 97.117 -79.785 45.968
52332 97.119 -79.810 45.944
52333 97.123 -79.839 45.922
52334 97.126 -79.867 45.900
52335 97.131 -79.891 45.878
52336 97.138 -79.899 45.857
52337 97.145 -79.906 45.834
52338 97.152 -79.914 45.812
52339 97.159 -79.921 45.791
52340 97.166 -79.929 45.771
52341 97.173 -79.936 45.750
52751 93.153 -79.447 53.244
52752 93.167 -79.435 53.230
52753 93.182 -79.423 53.217
52754 93.196 -79.411 53.204
52755 93.210 -79.399 53.191
52756 93.224 -79.387 53.178
52757 93.238 -79.375 53.166
52758 93.252 -79.363 53.155
52759 93.267 -79.351 53.142
52760 93.283 -79.343 53.131
52761 93.301 -79.333 53.121
52762 93.320 -79.327 53.110
52763 93.335 -79.322 53.099
52764 93.350 -79.318 53.088
52765 93.364 -79.313 53.077
52766 93.379 -79.309 53.066
52767 93.395 -79.304 53.054
52768 93.407 -79.300 53.040
52769 93.415 -79.294 53.026
52770 93.423 -79.289 53.013
52771 93.434 -79.284 53.000
52772 93.448 -79.278 52.988
52773 93.464 -79.273 52.976
52774 93.478 -79.268 52.964
52775 93.493 -79.263 52.953
52776 93.508 -79.257 52.943
52777 93.523 -79.252 52.935
52778 93.538 -79.247 52.927
52779 93.553 -79.242 52.919
52780 93.570 -79.237 52.913
52781 93.585 -79.230 52.905
52782 93.600 -79.219 52.897
52783 93.614 -79.214 52.887
52784 93.628 -79.209 52.877
52785 93.642 -79.203 52.867
52786 93.655 -79.198 52.858
52787 93.670 -79.193 52.850
52788 93.684 -79.187 52.843
52789 93.698 -79.182 52.836
52790 93.712 -79.178 52.829
52791 93.724 -79.182 52.821
52792 93.734 -79.191 52.812
52793 93.745 -79.200 52.804
52794 93.757 -79.203 52.796
52795 93.770 -79.199 52.787
52796 93.783 -79.195 52.778
52797 93.796 -79.191 52.768
52798 93.809 -79.187 52.759
52799 93.824 -79.187 52.751
52800 93.838 -79.187 52.741
52801 93.853 -79.186 52.728
52802 93.867 -79.186 52.713
52803 93.882 -79.186 52.698
52804 93.879 -79.183 52.682
52805 93.870 -79.179 52.666
52806 93.873 -79.177 52.651
52807 93.884 -79.175 52.637
52808 93.895 -79.175 52.623
52809 93.906 -79.173 52.607
52810 93.917 -79.172 52.589
52811 93.934 -79.172 52.570
52812 93.955 -79.174 52.549
52813 93.970 -79.175 52.527
52814 93.978 -79.178 52.503
52815 93.981 -79.181 52.480
52816 93.983 -79.186 52.459
52817 93.985 -79.190 52.442
52818 93.987 -79.194 52.426
52819 93.988 -79.190 52.412
52820 93.989 -79.184 52.399
52821 93.990 -79.177 52.384
52822 93.991 -79.171 52.369
52823 93.991 -79.165 52.356
52824 93.992 -79.159 52.342
52825 93.993 -79.153 52.331
52826 93.993 -79.153 52.321
52827 93.993 -79.152 52.310
52828 93.993 -79.151 52.299
52829 93.994 -79.153 52.287
52830 93.998 -79.160 52.276
52831 94.002 -79.167 52.269
52832 94.006 -79.174 52.259
52833 94.011 -79.184 52.247
52834 94.016 -79.193 52.235
52835 94.019 -79.190 52.219
52836 94.024 -79.187 52.201
52837 94.029 -79.184 52.180
52838 94.035 -79.180 52.160
52839 94.040 -79.177 52.139
52840 94.046 -79.174 52.118
52841 94.052 -79.175 52.098
52842 94.059 -79.187 52.076
52843 94.065 -79.197 52.053
52844 94.069 -79.204 52.030
52845 94.073 -79.212 52.008
52846 94.076 -79.219 51.987
52847 94.076 -79.224 51.967
52848 94.077 -79.228 51.950
52849 94.077 -79.234 51.932
52850 94.078 -79.239 51.912
52851 94.077 -79.241 51.891
52852 94.074 -79.238 51.868
52853 94.073 -79.235 51.844
52854 94.073 -79.231 51.818
52855 94.074 -79.228 51.792
52856 94.075 -79.225 51.765
52857 94.076 -79.221 51.736
52858 94.079 -79.213 51.709
52859 94.088 -79.197 51.684
52860 94.097 -79.182 51.661
52861 94.106 -79.166 51.640
52862 94.115 -79.150 51.620
52863 94.127 -79.132 51.602
52864 94.150 -79.112 51.583
52865 94.178 -79.123 51.563
52866 94.205 -79.134 51.540
52867 94.232 -79.143 51.516
52868 94.256 -79.137 51.487
52869 94.269 -79.109 51.455
52870 94.281 -79.081 51.420
52871 94.305 -79.078 51.379
52872 94.332 -79.086 51.338
52873 94.358 -79.096 51.298
52874 94.381 -79.100 51.259
52875 94.408 -79.084 51.221
52876 94.435 -79.067 51.183
52877 94.463 -79.050 51.146
52878 94.481 -79.041 51.107
52879 94.492 -79.037 51.069
52880 94.503 -79.034 51.030
52881 94.514 -79.030 50.993
52882 94.526 -79.026 50.956
52883 94.537 -79.022 50.919
52884 94.548 -79.019 50.883
52885 94.559 -79.015 50.845
52886 94.571 -79.012 50.807
52887 94.582 -79.008 50.770
52888 94.594 -79.004 50.733
52928 94.925 -79.302 49.249
52929 94.933 -79.309 49.213
52930 94.942 -79.316 49.177
52931 94.950 -79.324 49.140
52932 94.958 -79.332 49.103
52933 94.967 -79.339 49.067
52934 94.975 -79.347 49.030
52935 94.983 -79.354 48.994
52936 94.991 -79.361 48.958
52937 94.999 -79.369 48.921
52938 95.008 -79.376 48.886
52939 95.019 -79.386 48.850
52940 95.032 -79.399 48.813
52941 95.045 -79.412 48.777
52942 95.058 -79.426 48.741
52943 95.076 -79.432 48.705
52944 95.100 -79.433 48.670
52945 95.124 -79.435 48.635
52946 95.147 -79.436 48.597
52947 95.171 -79.438 48.559
52948 95.194 -79.439 48.520
52949 95.218 -79.441 48.483
52950 95.242 -79.442 48.445
52951 95.265 -79.444 48.408
52952 95.289 -79.445 48.370
52953 95.313 -79.447 48.334
52954 95.336 -79.448 48.297
52955 95.359 -79.450 48.260
52956 95.383 -79.451 48.225
52957 95.407 -79.453 48.190
52958 95.430 -79.454 48.154
52959 95.454 -79.456 48.118
52960 95.477 -79.457 48.083
52961 95.501 -79.459 48.048
52962 95.525 -79.460 48.014
52963 95.548 -79.462 47.980
52964 95.572 -79.463 47.946
52965 95.595 -79.465 47.913
52966 95.619 -79.466 47.880
52967 95.642 -79.468 47.847
52968 95.666 -79.469 47.815
52969 95.682 -79.449 47.782
52970 95.696 -79.428 47.750
52971 95.712 -79.411 47.717
52972 95.735 -79.424 47.684
52973 95.758 -79.437 47.650
52974 95.781 -79.450 47.618
52975 95.804 -79.463 47.584
52976 95.827 -79.476 47.550
52977 95.849 -79.489 47.517
52978 95.872 -79.502 47.483
52979 95.895 -79.516 47.450
52980 95.918 -79.529 47.416
52981 95.941 -79.538 47.382
52982 95.963 -79.543 47.348
52983 95.985 -79.548 47.315
52984 96.008 -79.553 47.282
52985 96.031 -79.557 47.250
52986 96.053 -79.562 47.218
52987 96.076 -79.567 47.186
52988 96.098 -79.571 47.154
53005 96.480 -79.650 46.640
53006 96.503 -79.654 46.613
53007 96.525 -79.659 46.586
53008 96.547 -79.664 46.559
53009 96.570 -79.669 46.532
53010 96.592 -79.673 46.505
53011 96.615 -79.678 46.478
53012 96.638 -79.683 46.452
53013 96.660 -79.687 46.428
53014 96.682 -79.692 46.403
53015 96.705 -79.696 46.378
53016 96.727 -79.701 46.351
53017 96.750 -79.706 46.326
53018 96.772 -79.711 46.301
53019 96.795 -79.715 46.277
53020 96.817 -79.720 46.253
53021 96.840 -79.724 46.229
53022 96.862 -79.729 46.205
53023 96.885 -79.733 46.181
53024 96.907 -79.739 46.157
53025 96.930 -79.743 46.132
53026 96.969 -79.742 46.109
53027 97.015 -79.740 46.087
53028 97.062 -79.737 46.065
53029 97.108 -79.735 46.042
53030 97.116 -79.743 46.018
53031 97.115 -79.755 45.995
53032 97.114 -79.767 45.970
53033 97.117 -79.791 45.946
53034 97.120 -79.819 45.924
53035 97.124 -79.848 45.901
53036 97.131 -79.857 45.879
53037 97.138 -79.864 45.857
53038 97.145 -79.872 45.834
53039 97.152 -79.880 45.812
53040 97.159 -79.887 45.791
53041 97.166 -79.895 45.770
53042 97.173 -79.902 45.749
53043 97.180 -79.910 45.729
53452 93.153 -79.438 53.271
53453 93.166 -79.425 53.257
53454 93.181 -79.413 53.244
53455 93.195 -79.400 53.231
53456 93.209 -79.388 53.218
53457 93.223 -79.376 53.205
53458 93.237 -79.364 53.193
53459 93.251 -79.352 53.181
53460 93.265 -79.340 53.167
53461 93.279 -79.328 53.156
53462 93.295 -79.317 53.144
53463 93.312 -79.313 53.132
53464 93.327 -79.308 53.121
53465 93.342 -79.303 53.109
53466 93.357 -79.299 53.098
53467 93.371 -79.294 53.087
53468 93.386 -79.290 53.074
53469 93.401 -79.285 53.061
53470 93.409 -79.280 53.047
53471 93.417 -79.275 53.034
53472 93.429 -79.269 53.022
53473 93.444 -79.264 53.010
53474 93.459 -79.259 52.997
53475 93.474 -79.253 52.985
53476 93.489 -79.248 52.974
53477 93.504 -79.242 52.963
53478 93.518 -79.237 52.955
53479 93.533 -79.231 52.947
53480 93.548 -79.226 52.940
53481 93.564 -79.222 52.932
53482 93.579 -79.215 52.924
53483 93.594 -79.206 52.916
53484 93.608 -79.201 52.907
53485 93.622 -79.196 52.897
53486 93.636 -79.190 52.888
53487 93.650 -79.185 52.879
53488 93.664 -79.180 52.871
53489 93.678 -79.175 52.864
53490 93.692 -79.170 52.856
53491 93.705 -79.167 52.849
53492 93.716 -79.173 52.842
53493 93.726 -79.182 52.834
53494 93.736 -79.190 52.826
53495 93.749 -79.189 52.817
53496 93.762 -79.185 52.808
53497 93.775 -79.181 52.799
53498 93.789 -79.178 52.790
53499 93.803 -79.178 52.781
53500 93.818 -79.178 52.772
53501 93.833 -79.178 52.761
53502 93.847 -79.177 52.749
53503 93.862 -79.177 52.735
53504 93.876 -79.177 52.719
53505 93.873 -79.174 52.703
53506 93.864 -79.170 52.688
53507 93.857 -79.166 52.674
53508 93.868 -79.165 52.660
53509 93.879 -79.164 52.644
53510 93.890 -79.162 52.628
53511 93.906 -79.162 52.609
53512 93.926 -79.164 52.589
53513 93.946 -79.165 52.569
53514 93.960 -79.167 52.548
53515 93.967 -79.170 52.526
53516 93.972 -79.173 52.504
53517 93.974 -79.178 52.484
53518 93.976 -79.181 52.468
53519 93.978 -79.186 52.453
53520 93.980 -79.190 52.440
53521 93.981 -79.188 52.427
53522 93.982 -79.182 52.412
53523 93.983 -79.176 52.398
53524 93.984 -79.170 52.385
53525 93.985 -79.164 52.371
53526 93.986 -79.157 52.359
53527 93.986 -79.151 52.348
53528 93.987 -79.145 52.339
53529 93.988 -79.139 52.328
53530 93.989 -79.138 52.318
53531 93.993 -79.147 52.307
53532 93.998 -79.156 52.297
53533 94.003 -79.166 52.287
53534 94.008 -79.176 52.274
53535 94.013 -79.186 52.261
53536 94.017 -79.185 52.244
53537 94.020 -79.182 52.224
53538 94.024 -79.179 52.201
53539 94.028 -79.176 52.180
53540 94.032 -79.174 52.159
53541 94.037 -79.170 52.138
53542 94.044 -79.175 52.117
53543 94.048 -79.183 52.095
53544 94.052 -79.190 52.073
53545 94.056 -79.198 52.050
53546 94.060 -79.206 52.028
53547 94.063 -79.213 52.007
53548 94.064 -79.218 51.986
53549 94.064 -79.223 51.969
53550 94.065 -79.228 51.950
53551 94.065 -79.233 51.930
53552 94.070 -79.233 51.908
53553 94.070 -79.229 51.883
53554 94.071 -79.225 51.857
53555 94.072 -79.222 51.830
53556 94.073 -79.218 51.802
53557 94.073 -79.215 51.773
53558 94.074 -79.211 51.743
53559 94.077 -79.204 51.716
53560 94.086 -79.188 51.691
53561 94.095 -79.172 51.670
53562 94.104 -79.156 51.651
53563 94.114 -79.140 51.634
53564 94.135 -79.115 51.618
53565 94.157 -79.090 51.600
53566 94.185 -79.099 51.580
53567 94.212 -79.109 51.559
53568 94.240 -79.120 51.535
53569 94.257 -79.112 51.507
53570 94.269 -79.087 51.474
53571 94.281 -79.059 51.438
53572 94.301 -79.048 51.396
53573 94.328 -79.057 51.354
53574 94.355 -79.065 51.313
53575 94.383 -79.074 51.273
53576 94.406 -79.081 51.234
53577 94.433 -79.065 51.197
53578 94.460 -79.047 51.159
53579 94.486 -79.032 51.121
53580 94.497 -79.028 51.082
53581 94.508 -79.024 51.044
53582 94.520 -79.021 51.007
53583 94.531 -79.017 50.970
53584 94.542 -79.013 50.933
53585 94.554 -79.010 50.896
53586 94.565 -79.006 50.858
53587 94.576 -79.002 50.821
53588 94.588 -78.999 50.783
53589 94.599 -78.995 50.746
53590 94.610 -78.991 50.709
53591 94.619 -78.997 50.672
53628 94.925 -79.274 49.293
53629 94.933 -79.281 49.256
53630 94.942 -79.289 49.220
53631 94.950 -79.296 49.184
53632 94.958 -79.303 49.149
53633 94.967 -79.311 49.111
53634 94.975 -79.319 49.075
53635 94.983 -79.326 49.039
53636 94.991 -79.334 49.002
53637 95.000 -79.341 48.966
53638 95.008 -79.349 48.930
53639 95.017 -79.356 48.894
53640 95.025 -79.363 48.858
53641 95.033 -79.371 48.822
53642 95.041 -79.379 48.786
53643 95.049 -79.386 48.750
53644 95.067 -79.394 48.713
53645 95.091 -79.396 48.677
53646 95.114 -79.397 48.640
53647 95.137 -79.399 48.603
53648 95.161 -79.400 48.566
53649 95.185 -79.402 48.528
53650 95.208 -79.403 48.490
53651 95.232 -79.404 48.452
53652 95.255 -79.406 48.415
53653 95.279 -79.407 48.378
53654 95.303 -79.409 48.341
53655 95.326 -79.410 48.303
53656 95.350 -79.412 48.266
53657 95.374 -79.413 48.232
53658 95.397 -79.415 48.195
53659 95.420 -79.416 48.159
53660 95.444 -79.418 48.124
53661 95.468 -79.419 48.089
53662 95.491 -79.421 48.055
53663 95.515 -79.422 48.020
53664 95.538 -79.424 47.986
53665 95.562 -79.425 47.952
53666 95.586 -79.427 47.920
53667 95.609 -79.429 47.887
53668 95.633 -79.430 47.854
53669 95.656 -79.432 47.822
53670 95.675 -79.419 47.790
53671 95.690 -79.398 47.759
53672 95.709 -79.396 47.727
53673 95.732 -79.409 47.694
53674 95.755 -79.422 47.662
53675 95.777 -79.435 47.628
53676 95.800 -79.448 47.594
53677 95.823 -79.461 47.559
53678 95.846 -79.474 47.525
53679 95.869 -79.488 47.491
53680 95.892 -79.501 47.457
53681 95.914 -79.513 47.423
53682 95.937 -79.521 47.388
53683 95.959 -79.526 47.353
53684 95.982 -79.530 47.319
53685 96.005 -79.535 47.286
53686 96.027 -79.539 47.254
53687 96.049 -79.544 47.222
53688 96.072 -79.548 47.190
53689 96.094 -79.553 47.157
53704 96.432 -79.623 46.698
53705 96.454 -79.627 46.669
53706 96.476 -79.632 46.640
53707 96.499 -79.637 46.613
53708 96.521 -79.642 46.586
53709 96.544 -79.646 46.559
53710 96.567 -79.651 46.532
53711 96.589 -79.655 46.506
53712 96.611 -79.660 46.479
53713 96.634 -79.664 46.453
53714 96.656 -79.669 46.429
53715 96.679 -79.674 46.404
53716 96.701 -79.679 46.379
53717 96.724 -79.683 46.353
53718 96.746 -79.688 46.327
53719 96.768 -79.692 46.302
53720 96.791 -79.697 46.278
53721 96.814 -79.701 46.254
53722 96.836 -79.706 46.231
53723 96.859 -79.711 46.207
53724 96.881 -79.716 46.183
53725 96.903 -79.720 46.161
53726 96.926 -79.725 46.137
53727 96.965 -79.724 46.114
53728 97.012 -79.722 46.093
53729 97.058 -79.719 46.069
53730 97.105 -79.717 46.045
53731 97.114 -79.725 46.021
53732 97.113 -79.736 45.998
53733 97.112 -79.748 45.973
53734 97.114 -79.771 45.949
53735 97.118 -79.800 45.925
53736 97.124 -79.814 45.903
53737 97.131 -79.822 45.880
53738 97.138 -79.830 45.856
53739 97.145 -79.837 45.834
53740 97.152 -79.845 45.812
53741 97.159 -79.852 45.790
53742 97.166 -79.860 45.770
53743 97.173 -79.867 45.749
53744 97.180 -79.875 45.728
53745 97.187 -79.883 45.707
54154 93.166 -79.415 53.285
54155 93.180 -79.403 53.272
54156 93.194 -79.391 53.258
54157 93.208 -79.379 53.245
54158 93.222 -79.366 53.232
54159 93.236 -79.354 53.219
54160 93.250 -79.341 53.206
54161 93.264 -79.329 53.193
54162 93.278 -79.317 53.181
54163 93.292 -79.305 53.168
54164 93.304 -79.297 53.156
54165 93.319 -79.293 53.142
54166 93.334 -79.288 53.130
54167 93.349 -79.284 53.119
54168 93.364 -79.279 53.108
54169 93.379 -79.275 53.095
54170 93.394 -79.270 53.081
54171 93.402 -79.266 53.068
54172 93.410 -79.260 53.056
54173 93.425 -79.255 53.045
54174 93.440 -79.250 53.033
54175 93.455 -79.244 53.020
54176 93.470 -79.238 53.009
54177 93.485 -79.233 52.997
54178 93.499 -79.228 52.987
54179 93.514 -79.222 52.977
54180 93.529 -79.217 52.969
54181 93.544 -79.212 52.962
54182 93.559 -79.206 52.954
54183 93.574 -79.200 52.946
54184 93.588 -79.194 52.938
54185 93.603 -79.188 52.928
54186 93.617 -79.183 52.918
54187 93.630 -79.178 52.909
54188 93.643 -79.175 52.901
54189 93.656 -79.172 52.893
54190 93.670 -79.169 52.886
54191 93.683 -79.165 52.880
54192 93.696 -79.162 52.873
54193 93.707 -79.164 52.865
54194 93.718 -79.172 52.857
54195 93.728 -79.179 52.848
54196 93.742 -79.175 52.840
54197 93.755 -79.172 52.831
54198 93.769 -79.169 52.822
54199 93.783 -79.169 52.812
54200 93.798 -79.169 52.802
54201 93.813 -79.169 52.793
54202 93.827 -79.169 52.781
54203 93.842 -79.168 52.768
54204 93.856 -79.168 52.755
54205 93.870 -79.168 52.741
54206 93.868 -79.165 52.726
54207 93.859 -79.161 52.712
54208 93.850 -79.157 52.698
54209 93.852 -79.154 52.682
54210 93.863 -79.153 52.666
54211 93.878 -79.153 52.648
54212 93.898 -79.154 52.629
54213 93.919 -79.156 52.610
54214 93.939 -79.156 52.591
54215 93.949 -79.159 52.572
54216 93.957 -79.162 52.550
54217 93.963 -79.165 52.530
54218 93.965 -79.169 52.512
54219 93.967 -79.173 52.497
54220 93.969 -79.178 52.483
54221 93.971 -79.181 52.470
54222 93.973 -79.186 52.457
54223 93.974 -79.187 52.443
54224 93.975 -79.181 52.429
54225 93.976 -79.175 52.416
54226 93.977 -79.169 52.402
54227 93.978 -79.162 52.389
54228 93.979 -79.156 52.377
54229 93.980 -79.150 52.368
54230 93.980 -79.144 52.358
54231 93.984 -79.139 52.348
54232 93.990 -79.139 52.338
54233 93.995 -79.149 52.327
54234 94.000 -79.159 52.316
54235 94.005 -79.169 52.302
54236 94.010 -79.178 52.289
54237 94.014 -79.181 52.271
54238 94.017 -79.178 52.250
54239 94.021 -79.175 52.227
54240 94.025 -79.172 52.205
54241 94.029 -79.169 52.183
54242 94.032 -79.166 52.161
54243 94.031 -79.169 52.139
54244 94.035 -79.177 52.117
54245 94.039 -79.184 52.093
54246 94.043 -79.192 52.071
54247 94.047 -79.200 52.048
54248 94.050 -79.207 52.026
54249 94.051 -79.212 52.005
54250 94.051 -79.217 51.986
54251 94.052 -79.222 51.967
54252 94.059 -79.223 51.947
54253 94.070 -79.222 51.924
54254 94.070 -79.219 51.898
54255 94.070 -79.216 51.871
54256 94.070 -79.212 51.843
54257 94.071 -79.208 51.814
54258 94.072 -79.204 51.783
54259 94.072 -79.201 51.752
54260 94.075 -79.194 51.726
54261 94.084 -79.179 51.704
54262 94.093 -79.163 51.683
54263 94.102 -79.147 51.665
54264 94.122 -79.123 51.651
54265 94.144 -79.098 51.636
54266 94.166 -79.072 51.618
54267 94.192 -79.074 51.599
54268 94.220 -79.084 51.577
54269 94.248 -79.095 51.554
54270 94.261 -79.080 51.525
54271 94.271 -79.059 51.492
54272 94.281 -79.037 51.454
54273 94.298 -79.019 51.412
54274 94.325 -79.028 51.370
54275 94.352 -79.036 51.328
54276 94.380 -79.044 51.287
54277 94.407 -79.053 51.248
54278 94.430 -79.062 51.210
54279 94.458 -79.045 51.173
54280 94.485 -79.028 51.135
54281 94.502 -79.019 51.096
54282 94.514 -79.015 51.058
54283 94.525 -79.012 51.020
54284 94.537 -79.008 50.983
54285 94.548 -79.004 50.946
54286 94.559 -79.001 50.909
54287 94.570 -78.997 50.871
54288 94.581 -78.993 50.834
54289 94.593 -78.990 50.796
54290 94.604 -78.986 50.759
54291 94.615 -78.982 50.722
54292 94.627 -78.978 50.685
54293 94.635 -78.984 50.647
54327 94.917 -79.238 49.375
54328 94.925 -79.246 49.338
54329 94.933 -79.253 49.301
54330 94.942 -79.261 49.264
54331 94.950 -79.269 49.228
54332 94.959 -79.276 49.193
54333 94.967 -79.284 49.157
54334 94.975 -79.291 49.120
54335 94.983 -79.298 49.083
54336 94.991 -79.306 49.048
54337 95.000 -79.313 49.012
54338 95.008 -79.321 48.975
54339 95.017 -79.328 48.939
54340 95.025 -79.336 48.903
54341 95.033 -79.343 48.867
54342 95.041 -79.350 48.832
54343 95.052 -79.366 48.796
54344 95.068 -79.385 48.760
54345 95.082 -79.385 48.723
54346 95.105 -79.389 48.686
54347 95.136 -79.394 48.649
54348 95.164 -79.394 48.611
54349 95.181 -79.390 48.574
54350 95.198 -79.385 48.537
54351 95.215 -79.381 48.499
54352 95.232 -79.376 48.460
54353 95.249 -79.372 48.423
54354 95.269 -79.370 48.386
54355 95.293 -79.372 48.348
54356 95.316 -79.372 48.310
54357 95.340 -79.374 48.272
54358 95.364 -79.375 48.237
54359 95.387 -79.377 48.201
54360 95.411 -79.379 48.165
54361 95.435 -79.380 48.130
54362 95.458 -79.382 48.095
54363 95.481 -79.383 48.059
54364 95.505 -79.385 48.025
54365 95.529 -79.386 47.991
54366 95.552 -79.388 47.958
54367 95.576 -79.389 47.927
54368 95.599 -79.391 47.895
54369 95.623 -79.392 47.863
54370 95.647 -79.394 47.831
54371 95.668 -79.389 47.798
54372 95.683 -79.368 47.769
54373 95.705 -79.381 47.739
54374 95.729 -79.394 47.708
54375 95.751 -79.407 47.676
54376 95.774 -79.420 47.641
54377 95.797 -79.433 47.605
54378 95.820 -79.446 47.569
54379 95.842 -79.460 47.534
54380 95.865 -79.473 47.499
54381 95.888 -79.485 47.464
54382 95.911 -79.498 47.429
54383 95.934 -79.503 47.393
54384 95.956 -79.507 47.357
54385 95.978 -79.512 47.324
54386 96.001 -79.516 47.290
54387 96.023 -79.521 47.257
54388 96.046 -79.526 47.225
54389 96.068 -79.530 47.192
54390 96.091 -79.535 47.160
54391 96.113 -79.540 47.128
54404 96.405 -79.600 46.729
54405 96.428 -79.604 46.699
54406 96.450 -79.610 46.668
54407 96.473 -79.614 46.640
54408 96.496 -79.619 46.613
54409 96.518 -79.623 46.586
54410 96.540 -79.628 46.559
54411 96.563 -79.632 46.532
54412 96.585 -79.637 46.505
54413 96.608 -79.642 46.479
54414 96.630 -79.647 46.454
54415 96.653 -79.651 46.429
54416 96.675 -79.656 46.405
54417 96.697 -79.661 46.380
54418 96.720 -79.665 46.355
54419 96.743 -79.670 46.330
54420 96.765 -79.674 46.304
54421 96.788 -79.679 46.279
54422 96.810 -79.684 46.256
54423 96.832 -79.689 46.232
54424 96.855 -79.693 46.209
54425 96.878 -79.698 46.187
54426 96.900 -79.702 46.167
54427 96.923 -79.707 46.148
54428 96.962 -79.706 46.128
54429 97.009 -79.704 46.105
54430 97.055 -79.701 46.082
54431 97.103 -79.698 46.057
54432 97.111 -79.707 46.030
54433 97.111 -79.718 46.003
54434 97.110 -79.730 45.976
54435 97.112 -79.752 45.951
54436 97.117 -79.772 45.927
54437 97.124 -79.780 45.905
54438 97.131 -79.787 45.881
54439 97.138 -79.795 45.857
54440 97.145 -79.803 45.834
54441 97.152 -79.811 45.812
54442 97.159 -79.818 45.791
54443 97.166 -79.826 45.770
54444 97.173 -79.833 45.748
54445 97.180 -79.841 45.728
54446 97.188 -79.849 45.707
54447 97.195 -79.856 45.686
54855 93.165 -79.404 53.312
54856 93.179 -79.392 53.299
54857 93.193 -79.380 53.285
54858 93.207 -79.368 53.271
54859 93.221 -79.356 53.258
54860 93.235 -79.344 53.245
54861 93.249 -79.332 53.232
54862 93.263 -79.319 53.219
54863 93.277 -79.307 53.206
54864 93.291 -79.295 53.192
54865 93.303 -79.290 53.179
54866 93.314 -79.283 53.165
54867 93.326 -79.274 53.152
54868 93.341 -79.269 53.141
54869 93.356 -79.265 53.129
54870 93.371 -79.260 53.116
54871 93.386 -79.256 53.103
54872 93.396 -79.250 53.090
54873 93.406 -79.245 53.079
54874 93.421 -79.240 53.068
54875 93.436 -79.234 53.057
54876 93.450 -79.229 53.045
54877 93.466 -79.224 53.034
54878 93.480 -79.219 53.022
54879 93.495 -79.213 53.013
54880 93.510 -79.208 53.003
54881 93.524 -79.203 52.993
54882 93.540 -79.197 52.986
54883 93.554 -79.191 52.978
54884 93.569 -79.187 52.970
54885 93.582 -79.184 52.961
54886 93.595 -79.180 52.950
54887 93.608 -79.177 52.940
54888 93.621 -79.174 52.931
54889 93.634 -79.170 52.924
54890 93.647 -79.167 52.916
54891 93.660 -79.164 52.910
54892 93.673 -79.161 52.904
54893 93.686 -79.157 52.897
54894 93.699 -79.154 52.889
54895 93.709 -79.163 52.882
54896 93.721 -79.165 52.873
54897 93.734 -79.162 52.864
54898 93.749 -79.160 52.855
54899 93.763 -79.160 52.846
54900 93.778 -79.159 52.836
54901 93.792 -79.159 52.826
54902 93.806 -79.159 52.816
54903 93.821 -79.159 52.803
54904 93.836 -79.159 52.791
54905 93.850 -79.159 52.778
54906 93.865 -79.159 52.764
54907 93.862 -79.156 52.750
54908 93.853 -79.152 52.737
54909 93.845 -79.148 52.724
54910 93.837 -79.143 52.706
54911 93.850 -79.143 52.689
54912 93.870 -79.144 52.671
54913 93.891 -79.146 52.651
54914 93.911 -79.147 52.632
54915 93.931 -79.148 52.613
54916 93.939 -79.150 52.595
54917 93.946 -79.153 52.576
54918 93.954 -79.156 52.558
54919 93.956 -79.160 52.542
54920 93.958 -79.164 52.529
54921 93.960 -79.169 52.515
54922 93.962 -79.173 52.502
54923 93.964 -79.177 52.489
54924 93.966 -79.181 52.476
54925 93.968 -79.186 52.462
54926 93.968 -79.180 52.450
54927 93.969 -79.174 52.435
54928 93.970 -79.167 52.421
54929 93.971 -79.161 52.408
54930 93.972 -79.155 52.398
54931 93.973 -79.149 52.387
54932 93.979 -79.145 52.378
54933 93.985 -79.142 52.368
54934 93.991 -79.141 52.357
54935 93.996 -79.151 52.345
54936 94.001 -79.161 52.331
54937 94.006 -79.171 52.317
54938 94.011 -79.176 52.299
54939 94.014 -79.173 52.277
54940 94.018 -79.170 52.255
54941 94.022 -79.168 52.232
54942 94.026 -79.165 52.209
54943 94.026 -79.166 52.186
54944 94.025 -79.169 52.163
54945 94.024 -79.173 52.140
54946 94.026 -79.178 52.116
54947 94.030 -79.186 52.093
54948 94.034 -79.194 52.070
54949 94.038 -79.202 52.047
54950 94.038 -79.206 52.025
54951 94.039 -79.212 52.006
54952 94.044 -79.213 51.986
54953 94.055 -79.212 51.963
54954 94.066 -79.212 51.939
54955 94.071 -79.209 51.913
54956 94.070 -79.206 51.885
54957 94.069 -79.202 51.856
54958 94.069 -79.198 51.826
54959 94.070 -79.194 51.796
54960 94.071 -79.191 51.766
54961 94.073 -79.185 51.743
54962 94.082 -79.169 51.722
54963 94.091 -79.153 51.703
54964 94.109 -79.131 51.687
54965 94.131 -79.106 51.671
54966 94.153 -79.081 51.656
54967 94.175 -79.055 51.638
54968 94.200 -79.049 51.618
54969 94.227 -79.060 51.597
54970 94.254 -79.068 51.572
54971 94.264 -79.048 51.541
54972 94.274 -79.027 51.507
54973 94.284 -79.006 51.469
54974 94.297 -78.995 51.427
54975 94.322 -78.998 51.384
54976 94.349 -79.006 51.343
54977 94.377 -79.015 51.302
54978 94.404 -79.024 51.263
54979 94.431 -79.032 51.224
54980 94.455 -79.042 51.186
54981 94.482 -79.025 51.148
54982 94.508 -79.010 51.110
54983 94.520 -79.006 51.072
54984 94.531 -79.002 51.034
54985 94.542 -78.999 50.997
54986 94.553 -78.995 50.959
54987 94.564 -78.991 50.922
54988 94.575 -78.988 50.884
54989 94.587 -78.984 50.847
54990 94.598 -78.981 50.809
54991 94.610 -78.977 50.772
54992 94.621 -78.973 50.734
54993 94.632 -78.969 50.697
54994 94.643 -78.966 50.660
54995 94.652 -78.971 50.622
54996 94.661 -78.979 50.584
54997 94.669 -78.987 50.547
54998 94.677 -78.994 50.508
55027 94.917 -79.211 49.421
55028 94.925 -79.219 49.384
55029 94.933 -79.225 49.346
55030 94.942 -79.233 49.310
55031 94.950 -79.241 49.274
55032 94.959 -79.248 49.238
55033 94.967 -79.256 49.202
55034 94.975 -79.263 49.166
55035 94.983 -79.271 49.129
55036 94.992 -79.278 49.093
55037 95.000 -79.285 49.057
55038 95.008 -79.293 49.022
55039 95.017 -79.300 48.986
55040 95.025 -79.308 48.950
55041 95.035 -79.321 48.914
55042 95.047 -79.339 48.878
55043 95.060 -79.357 48.843
55044 95.072 -79.375 48.808
55045 95.111 -79.400 48.772
55046 95.101 -79.382 48.734
55047 95.125 -79.385 48.697
55048 95.148 -79.389 48.659
55049 95.172 -79.393 48.622
55050 95.199 -79.397 48.584
55051 95.237 -79.403 48.547
55052 95.273 -79.408 48.509
55053 95.292 -79.404 48.470
55054 95.309 -79.400 48.432
55055 95.326 -79.395 48.394
55056 95.343 -79.391 48.356
55057 95.360 -79.386 48.318
55058 95.378 -79.382 48.281
55059 95.395 -79.377 48.245
55060 95.412 -79.372 48.208
55061 95.429 -79.368 48.173
55062 95.447 -79.363 48.138
55063 95.464 -79.359 48.102
55064 95.481 -79.354 48.066
55065 95.498 -79.350 48.031
55066 95.519 -79.348 47.998
55067 95.542 -79.350 47.966
55068 95.566 -79.351 47.934
55069 95.590 -79.353 47.903
55070 95.613 -79.354 47.872
55071 95.636 -79.356 47.841
55072 95.660 -79.357 47.812
55073 95.679 -79.353 47.781
55074 95.702 -79.366 47.752
55075 95.725 -79.379 47.722
55076 95.748 -79.392 47.689
55077 95.770 -79.405 47.654
55078 95.794 -79.418 47.618
55079 95.816 -79.432 47.583
55080 95.839 -79.444 47.547
55081 95.862 -79.457 47.510
55082 95.885 -79.470 47.472
55083 95.907 -79.480 47.434
55084 95.930 -79.485 47.398
55085 95.952 -79.489 47.362
55086 95.975 -79.494 47.327
55087 95.997 -79.498 47.293
55088 96.020 -79.503 47.260
55089 96.042 -79.508 47.228
55090 96.064 -79.513 47.194
55091 96.087 -79.517 47.162
55092 96.110 -79.522 47.129
55103 96.357 -79.573 46.788
55104 96.379 -79.577 46.758
55105 96.402 -79.582 46.728
55106 96.425 -79.587 46.698
55107 96.447 -79.592 46.668
55108 96.469 -79.596 46.640
55109 96.492 -79.601 46.612
55110 96.514 -79.605 46.585
55111 96.537 -79.610 46.559
55112 96.559 -79.614 46.532
55113 96.582 -79.620 46.505
55114 96.604 -79.624 46.479
55115 96.626 -79.629 46.454
55116 96.650 -79.632 46.430
55117 96.674 -79.636 46.407
55118 96.697 -79.640 46.382
55119 96.722 -79.644 46.357
55120 96.745 -79.648 46.332
55121 96.769 -79.651 46.306
55122 96.793 -79.655 46.281
55123 96.817 -79.660 46.258
55124 96.841 -79.664 46.235
55125 96.865 -79.667 46.213
55126 96.888 -79.671 46.192
55127 96.912 -79.675 46.173
55128 96.927 -79.685 46.156
55129 96.972 -79.683 46.136
55130 97.021 -79.679 46.114
55131 97.070 -79.676 46.091
55132 97.098 -79.672 46.067
55133 97.120 -79.668 46.042
55134 97.115 -79.690 46.014
55135 97.108 -79.711 45.986
55136 97.110 -79.730 45.958
55137 97.117 -79.738 45.932
55138 97.124 -79.745 45.908
55139 97.131 -79.753 45.883
55140 97.138 -79.761 45.859
55141 97.145 -79.768 45.835
55142 97.152 -79.776 45.813
55143 97.160 -79.783 45.791
55144 97.167 -79.791 45.769
55145 97.174 -79.798 45.748
55146 97.181 -79.806 45.727
55147 97.188 -79.814 45.707
55148 97.195 -79.822 45.685
55149 97.202 -79.830 45.664
55556 93.164 -79.394 53.339
55557 93.178 -79.382 53.325
55558 93.192 -79.370 53.312
55559 93.206 -79.358 53.298
55560 93.220 -79.345 53.284
55561 93.234 -79.333 53.271
55562 93.248 -79.321 53.257
55563 93.262 -79.309 53.244
55564 93.276 -79.297 53.231
55565 93.290 -79.285 53.217
55566 93.304 -79.279 53.204
55567 93.313 -79.277 53.189
55568 93.324 -79.269 53.176
55569 93.336 -79.259 53.164
55570 93.348 -79.250 53.151
55571 93.363 -79.246 53.138
55572 93.378 -79.241 53.125
55573 93.390 -79.236 53.113
55574 93.402 -79.231 53.103
55575 93.417 -79.225 53.092
55576 93.431 -79.220 53.081
55577 93.446 -79.215 53.070
55578 93.461 -79.209 53.059
55579 93.476 -79.204 53.049
55580 93.491 -79.199 53.040
55581 93.505 -79.194 53.030
55582 93.520 -79.188 53.021
55583 93.534 -79.184 53.013
55584 93.549 -79.179 53.004
55585 93.563 -79.174 52.994
55586 93.576 -79.172 52.984
55587 93.588 -79.170 52.973
55588 93.600 -79.169 52.963
55589 93.612 -79.167 52.955
55590 93.625 -79.165 52.947
55591 93.638 -79.162 52.940
55592 93.650 -79.159 52.934
55593 93.663 -79.156 52.929
55594 93.677 -79.153 52.922
55595 93.690 -79.150 52.915
55596 93.701 -79.154 52.908
55597 93.714 -79.152 52.900
55598 93.728 -79.151 52.890
55599 93.743 -79.151 52.881
55600 93.757 -79.151 52.871
55601 93.772 -79.150 52.861
55602 93.786 -79.150 52.851
55603 93.801 -79.150 52.840
55604 93.816 -79.150 52.827
55605 93.830 -79.150 52.815
55606 93.845 -79.150 52.802
55607 93.859 -79.150 52.789
55608 93.857 -79.147 52.776
55609 93.848 -79.143 52.763
55610 93.839 -79.139 52.748
55611 93.832 -79.135 52.732
55612 93.842 -79.134 52.714
55613 93.863 -79.136 52.695
55614 93.883 -79.137 52.676
55615 93.904 -79.138 52.656
55616 93.921 -79.140 52.638
55617 93.928 -79.142 52.621
55618 93.936 -79.145 52.604
55619 93.943 -79.147 52.588
55620 93.946 -79.151 52.575
55621 93.948 -79.156 52.562
55622 93.951 -79.160 52.550
55623 93.953 -79.164 52.537
55624 93.955 -79.169 52.525
55625 93.957 -79.173 52.512
55626 93.958 -79.177 52.498
55627 93.960 -79.181 52.485
55628 93.962 -79.178 52.470
55629 93.963 -79.172 52.455
55630 93.963 -79.166 52.440
55631 93.964 -79.160 52.428
55632 93.968 -79.156 52.417
55633 93.974 -79.152 52.407
55634 93.981 -79.148 52.396
55635 93.987 -79.145 52.385
55636 93.993 -79.143 52.373
55637 93.998 -79.154 52.360
55638 94.003 -79.164 52.345
55639 94.008 -79.172 52.327
55640 94.012 -79.169 52.306
55641 94.015 -79.166 52.283
55642 94.019 -79.163 52.260
55643 94.020 -79.163 52.236
55644 94.019 -79.166 52.213
55645 94.019 -79.169 52.188
55646 94.018 -79.173 52.163
55647 94.017 -79.176 52.139
55648 94.016 -79.180 52.116
55649 94.020 -79.188 52.092
55650 94.025 -79.196 52.069
55651 94.025 -79.200 52.047
55652 94.030 -79.203 52.026
55653 94.041 -79.203 52.004
55654 94.052 -79.201 51.978
55655 94.063 -79.200 51.952
55656 94.072 -79.199 51.926
55657 94.071 -79.195 51.898
55658 94.070 -79.192 51.868
55659 94.068 -79.188 51.839
55660 94.069 -79.184 51.810
55661 94.069 -79.181 51.784
55662 94.071 -79.176 51.763
55663 94.080 -79.160 51.743
55664 94.096 -79.140 51.725
55665 94.118 -79.114 51.709
55666 94.139 -79.089 51.692
55667 94.161 -79.063 51.675
55668 94.183 -79.038 51.657
55669 94.207 -79.025 51.637
55670 94.235 -79.036 51.615
55671 94.257 -79.037 51.589
55672 94.267 -79.015 51.556
55673 94.277 -78.995 51.521
55674 94.289 -78.981 51.482
55675 94.309 -78.991 51.440
55676 94.329 -79.000 51.398
55677 94.351 -78.997 51.356
55678 94.374 -78.990 51.315
55679 94.400 -78.994 51.276
55680 94.427 -79.002 51.237
55681 94.455 -79.011 51.200
55682 94.480 -79.022 51.162
55683 94.507 -79.006 51.124
55684 94.525 -78.997 51.086
55685 94.536 -78.993 51.049
55686 94.547 -78.990 51.011
55687 94.558 -78.986 50.974
55688 94.570 -78.982 50.936
55689 94.581 -78.979 50.898
55690 94.593 -78.975 50.860
55691 94.604 -78.971 50.822
55692 94.615 -78.968 50.785
55693 94.626 -78.964 50.747
55694 94.637 -78.960 50.710
55695 94.648 -78.957 50.672
55696 94.660 -78.953 50.634
55697 94.669 -78.959 50.597
55698 94.677 -78.966 50.559
55699 94.685 -78.974 50.521
55700 94.694 -78.981 50.483
55701 94.702 -78.989 50.445
55702 94.710 -78.996 50.407
55703 94.718 -79.003 50.369
55704 94.726 -79.011 50.331
55705 94.735 -79.018 50.294
55706 94.743 -79.026 50.257
55727 94.917 -79.183 49.467
55728 94.925 -79.190 49.430
55729 94.934 -79.198 49.392
55730 94.942 -79.206 49.356
55731 94.951 -79.213 49.319
55732 94.959 -79.220 49.283
55733 94.967 -79.228 49.247
55734 94.975 -79.235 49.211
55735 94.983 -79.243 49.175
55736 94.992 -79.250 49.139
55737 95.000 -79.258 49.103
55738 95.008 -79.266 49.067
55739 95.018 -79.276 49.032
55740 95.030 -79.294 48.996
55741 95.042 -79.313 48.961
55742 95.055 -79.331 48.925
55743 95.067 -79.349 48.890
55744 95.079 -79.367 48.855
55745 95.115 -79.391 48.820
55746 95.135 -79.400 48.783
55747 95.120 -79.378 48.746
55748 95.144 -79.382 48.708
55749 95.168 -79.385 48.670
55750 95.192 -79.389 48.633
55751 95.215 -79.393 48.595
55752 95.239 -79.397 48.557
55753 95.263 -79.401 48.519
55754 95.300 -79.407 48.481
55755 95.337 -79.412 48.442
55756 95.374 -79.417 48.403
55757 95.403 -79.419 48.365
55758 95.420 -79.414 48.328
55759 95.438 -79.410 48.291
55760 95.455 -79.405 48.253
55761 95.472 -79.400 48.217
55762 95.489 -79.396 48.182
55763 95.506 -79.391 48.146
55764 95.523 -79.387 48.110
55765 95.540 -79.382 48.074
55766 95.558 -79.378 48.039
55767 95.575 -79.373 48.005
55768 95.592 -79.369 47.974
55769 95.609 -79.364 47.942
55770 95.626 -79.360 47.911
55771 95.642 -79.354 47.881
55772 95.652 -79.346 47.852
55773 95.661 -79.338 47.825
55774 95.670 -79.340 47.794
55775 95.691 -79.352 47.763
55776 95.718 -79.364 47.732
55777 95.744 -79.377 47.699
55778 95.767 -79.390 47.664
55779 95.790 -79.404 47.628
55780 95.813 -79.416 47.592
55781 95.835 -79.429 47.555
55782 95.859 -79.442 47.517
55783 95.881 -79.456 47.477
55784 95.904 -79.462 47.438
55785 95.926 -79.466 47.401
55786 95.949 -79.471 47.365
55787 95.971 -79.476 47.329
55788 95.993 -79.481 47.295
55789 96.016 -79.485 47.262
55790 96.039 -79.490 47.229
55791 96.061 -79.494 47.196
55792 96.084 -79.499 47.163
55793 96.106 -79.504 47.130
55803 96.340 -79.545 46.818
55804 96.365 -79.548 46.788
55805 96.388 -79.552 46.757
55806 96.412 -79.556 46.727
55807 96.436 -79.560 46.697
55808 96.460 -79.563 46.667
55809 96.484 -79.567 46.639
55810 96.508 -79.571 46.612
55811 96.532 -79.575 46.584
55812 96.555 -79.579 46.558
55813 96.579 -79.582 46.532
55814 96.603 -79.586 46.505
55815 96.627 -79.590 46.479
55816 96.651 -79.594 46.455
55817 96.675 -79.598 46.431
55818 96.698 -79.601 46.408
55819 96.723 -79.605 46.384
55820 96.746 -79.609 46.359
55821 96.770 -79.613 46.333
55822 96.794 -79.617 46.308
55823 96.818 -79.620 46.283
55824 96.842 -79.624 46.260
55825 96.866 -79.628 46.238
55826 96.890 -79.632 46.217
55827 96.909 -79.639 46.197
55828 96.920 -79.651 46.180
55829 96.939 -79.661 46.161
55830 96.987 -79.657 46.141
55831 97.036 -79.653 46.119
55832 97.068 -79.649 46.096
55833 97.090 -79.645 46.075
55834 97.111 -79.642 46.054
55835 97.134 -79.638 46.028
55836 97.130 -79.659 45.999
55837 97.116 -79.690 45.970
55838 97.117 -79.703 45.940
55839 97.124 -79.711 45.912
55840 97.132 -79.718 45.886
55841 97.139 -79.726 45.861
55842 97.146 -79.733 45.836
55843 97.153 -79.742 45.813
55844 97.160 -79.749 45.790
55845 97.167 -79.757 45.769
55846 97.174 -79.764 45.747
55847 97.181 -79.772 45.726
55848 97.188 -79.780 45.706
55849 97.195 -79.787 45.684
55850 97.202 -79.795 45.663
55851 97.209 -79.802 45.642
56258 93.177 -79.372 53.352
56259 93.191 -79.360 53.338
56260 93.205 -79.347 53.324
56261 93.219 -79.335 53.311
56262 93.233 -79.323 53.297
56263 93.247 -79.311 53.283
56264 93.261 -79.299 53.269
56265 93.275 -79.286 53.256
56266 93.289 -79.274 53.243
56267 93.304 -79.269 53.229
56268 93.313 -79.266 53.214
56269 93.323 -79.263 53.200
56270 93.334 -79.255 53.187
56271 93.346 -79.244 53.174
56272 93.357 -79.234 53.161
56273 93.370 -79.226 53.148
56274 93.383 -79.222 53.137
56275 93.398 -79.216 53.126
56276 93.412 -79.211 53.115
56277 93.427 -79.206 53.103
56278 93.441 -79.200 53.093
56279 93.456 -79.196 53.083
56280 93.471 -79.190 53.073
56281 93.485 -79.186 53.064
56282 93.499 -79.181 53.055
56283 93.514 -79.176 53.046
56284 93.529 -79.172 53.038
56285 93.543 -79.166 53.029
56286 93.557 -79.162 53.019
56287 93.571 -79.158 53.008
56288 93.583 -79.156 52.997
56289 93.595 -79.155 52.988
56290 93.608 -79.153 52.979
56291 93.620 -79.152 52.972
56292 93.632 -79.150 52.966
56293 93.644 -79.149 52.960
56294 93.656 -79.147 52.954
56295 93.669 -79.146 52.948
56296 93.681 -79.144 52.941
56297 93.693 -79.143 52.935
56298 93.708 -79.142 52.927
56299 93.722 -79.142 52.918
56300 93.737 -79.142 52.908
56301 93.752 -79.142 52.898
56302 93.766 -79.141 52.887
56303 93.781 -79.141 52.876
56304 93.795 -79.141 52.865
56305 93.810 -79.141 52.853
56306 93.825 -79.141 52.840
56307 93.839 -79.140 52.827
56308 93.854 -79.140 52.815
56309 93.851 -79.137 52.802
56310 93.843 -79.134 52.788
56311 93.834 -79.130 52.773
56312 93.837 -79.128 52.757
56313 93.842 -79.127 52.740
56314 93.855 -79.128 52.720
56315 93.875 -79.128 52.701
56316 93.896 -79.130 52.682
56317 93.911 -79.131 52.664
56318 93.918 -79.134 52.648
56319 93.925 -79.137 52.633
56320 93.933 -79.139 52.619
56321 93.937 -79.143 52.608
56322 93.939 -79.147 52.595
56323 93.941 -79.151 52.583
56324 93.943 -79.156 52.571
56325 93.945 -79.160 52.559
56326 93.947 -79.164 52.546
56327 93.949 -79.169 52.532
56328 93.951 -79.173 52.518
56329 93.953 -79.177 52.503
56330 93.955 -79.178 52.487
56331 93.956 -79.171 52.472
56332 93.957 -79.165 52.459
56333 93.963 -79.162 52.446
56334 93.970 -79.158 52.434
56335 93.976 -79.155 52.423
56336 93.983 -79.151 52.412
56337 93.989 -79.147 52.400
56338 93.995 -79.147 52.387
56339 94.000 -79.156 52.372
56340 94.005 -79.166 52.355
56341 94.009 -79.164 52.335
56342 94.012 -79.162 52.311
56343 94.015 -79.159 52.287
56344 94.014 -79.163 52.262
56345 94.013 -79.166 52.237
56346 94.012 -79.170 52.213
56347 94.011 -79.173 52.188
56348 94.011 -79.176 52.162
56349 94.010 -79.180 52.138
56350 94.009 -79.183 52.115
56351 94.011 -79.190 52.092
56352 94.015 -79.194 52.068
56353 94.026 -79.193 52.045
56354 94.037 -79.191 52.021
56355 94.048 -79.190 51.992
56356 94.059 -79.190 51.965
56357 94.070 -79.188 51.936
56358 94.072 -79.185 51.908
56359 94.071 -79.182 51.879
56360 94.069 -79.178 51.852
56361 94.068 -79.175 51.826
56362 94.068 -79.171 51.803
56363 94.069 -79.166 51.786
56364 94.082 -79.147 51.768
56365 94.104 -79.122 51.749
56366 94.126 -79.096 51.730
56367 94.148 -79.071 51.713
56368 94.170 -79.046 51.695
56369 94.192 -79.021 51.676
56370 94.214 -79.000 51.656
56371 94.242 -79.011 51.633
56372 94.260 -79.005 51.605
56373 94.270 -78.984 51.571
56374 94.281 -78.967 51.535
56375 94.301 -78.977 51.495
56376 94.322 -78.987 51.452
56377 94.341 -78.996 51.410
56378 94.361 -79.006 51.369
56379 94.383 -79.006 51.328
56380 94.406 -78.998 51.289
56381 94.429 -78.990 51.251
56382 94.452 -78.983 51.215
56383 94.479 -78.990 51.178
56384 94.506 -78.999 51.140
56385 94.530 -78.988 51.102
56386 94.541 -78.984 51.064
56387 94.553 -78.981 51.026
56388 94.564 -78.977 50.988
56389 94.575 -78.973 50.950
56390 94.586 -78.970 50.912
56391 94.598 -78.966 50.873
56392 94.609 -78.962 50.835
56393 94.620 -78.959 50.798
56394 94.631 -78.955 50.760
56395 94.643 -78.951 50.723
56396 94.654 -78.948 50.685
56397 94.666 -78.944 50.647
56398 94.677 -78.940 50.609
56399 94.685 -78.946 50.571
56400 94.694 -78.953 50.533
56401 94.702 -78.961 50.495
56402 94.710 -78.968 50.457
56403 94.718 -78.976 50.419
56404 94.727 -78.984 50.381
56405 94.735 -78.991 50.343
56406 94.743 -78.998 50.305
56407 94.752 -79.006 50.268
56408 94.760 -79.013 50.230
56409 94.768 -79.021 50.192
56410 94.776 -79.028 50.154
56411 94.785 -79.036 50.117
56425 94.901 -79.140 49.590
56426 94.909 -79.148 49.553
56427 94.917 -79.156 49.514
56428 94.925 -79.162 49.477
56429 94.934 -79.170 49.440
56430 94.942 -79.178 49.403
56431 94.951 -79.185 49.366
56432 94.959 -79.193 49.329
56433 94.967 -79.200 49.293
56434 94.975 -79.208 49.256
56435 94.983 -79.215 49.221
56436 94.992 -79.222 49.185
56437 95.001 -79.231 49.149
56438 95.013 -79.250 49.114
56439 95.025 -79.268 49.078
56440 95.038 -79.286 49.043
56441 95.050 -79.304 49.008
56442 95.062 -79.322 48.973
56443 95.074 -79.341 48.938
56444 95.087 -79.359 48.903
56445 95.119 -79.382 48.868
56446 95.120 -79.374 48.832
56447 95.126 -79.370 48.795
56448 95.137 -79.372 48.758
56449 95.162 -79.376 48.720
56450 95.187 -79.382 48.682
56451 95.211 -79.385 48.644
56452 95.235 -79.389 48.605
56453 95.259 -79.393 48.567
56454 95.282 -79.397 48.530
56455 95.306 -79.401 48.492
56456 95.330 -79.405 48.453
56457 95.364 -79.410 48.415
56458 95.401 -79.415 48.376
56459 95.438 -79.421 48.339
56460 95.475 -79.426 48.302
56461 95.512 -79.432 48.264
56462 95.532 -79.428 48.228
56463 95.549 -79.423 48.192
56464 95.566 -79.419 48.155
56465 95.583 -79.414 48.119
56466 95.593 -79.407 48.083
56467 95.603 -79.398 48.047
56468 95.612 -79.390 48.015
56469 95.621 -79.382 47.983
56470 95.631 -79.373 47.951
56471 95.640 -79.365 47.921
56472 95.650 -79.357 47.891
56473 95.659 -79.349 47.863
56474 95.669 -79.341 47.835
56475 95.678 -79.336 47.804
56476 95.687 -79.346 47.772
56477 95.696 -79.357 47.739
56478 95.705 -79.367 47.704
56479 95.730 -79.379 47.668
56480 95.756 -79.391 47.631
56481 95.783 -79.404 47.592
56482 95.810 -79.417 47.554
56483 95.837 -79.429 47.516
56484 95.877 -79.434 47.478
56485 95.906 -79.441 47.440
56486 95.925 -79.452 47.403
56487 95.949 -79.457 47.367
56488 95.974 -79.460 47.331
56489 95.999 -79.462 47.296
56490 96.024 -79.464 47.262
56491 96.049 -79.466 47.230
56492 96.074 -79.469 47.197
56493 96.099 -79.472 47.164
56494 96.124 -79.474 47.131
56495 96.150 -79.476 47.099
56502 96.318 -79.501 46.877
56503 96.342 -79.505 46.848
56504 96.366 -79.509 46.818
56505 96.389 -79.513 46.788
56506 96.413 -79.516 46.758
56507 96.437 -79.520 46.727
56508 96.461 -79.524 46.696
56509 96.485 -79.528 46.665
56510 96.509 -79.532 46.637
56511 96.533 -79.535 46.610
56512 96.556 -79.539 46.583
56513 96.581 -79.543 46.557
56514 96.604 -79.547 46.531
56515 96.628 -79.551 46.506
56516 96.652 -79.554 46.480
56517 96.676 -79.558 46.456
56518 96.700 -79.562 46.432
56519 96.724 -79.567 46.408
56520 96.747 -79.570 46.384
56521 96.771 -79.574 46.360
56522 96.795 -79.578 46.335
56523 96.819 -79.582 46.310
56524 96.843 -79.585 46.286
56525 96.867 -79.589 46.263
56526 96.891 -79.593 46.242
56527 96.902 -79.605 46.221
56528 96.913 -79.617 46.201
56529 96.924 -79.630 46.182
56530 96.954 -79.635 46.161
56531 97.003 -79.631 46.140
56532 97.038 -79.627 46.117
56533 97.060 -79.623 46.097
56534 97.082 -79.620 46.077
56535 97.103 -79.616 46.056
56536 97.125 -79.612 46.029
56537 97.147 -79.608 46.004
56538 97.158 -79.616 45.977
56539 97.149 -79.639 45.947
56540 97.124 -79.676 45.916
56541 97.132 -79.684 45.888
56542 97.139 -79.692 45.862
56543 97.146 -79.699 45.837
56544 97.153 -79.707 45.813
56545 97.160 -79.714 45.790
56546 97.167 -79.722 45.769
56547 97.174 -79.730 45.747
56548 97.181 -79.737 45.726
56549 97.188 -79.745 45.705
56550 97.195 -79.752 45.684
56551 97.203 -79.760 45.662
56552 97.210 -79.768 45.642
56553 97.217 -79.776 45.623
56959 93.176 -79.362 53.379
56960 93.190 -79.350 53.365
56961 93.204 -79.337 53.351
56962 93.218 -79.325 53.337
56963 93.232 -79.313 53.323
56964 93.247 -79.300 53.310
56965 93.260 -79.288 53.296
56966 93.274 -79.276 53.282
56967 93.288 -79.264 53.268
56968 93.304 -79.257 53.253
56969 93.314 -79.255 53.239
56970 93.324 -79.253 53.224
56971 93.333 -79.250 53.212
56972 93.345 -79.241 53.198
56973 93.356 -79.230 53.185
56974 93.367 -79.220 53.173
56975 93.379 -79.209 53.160
56976 93.393 -79.203 53.148
56977 93.407 -79.198 53.136
56978 93.421 -79.193 53.125
56979 93.436 -79.188 53.114
56980 93.450 -79.183 53.104
56981 93.465 -79.178 53.094
56982 93.479 -79.174 53.086
56983 93.494 -79.169 53.077
56984 93.508 -79.164 53.069
56985 93.522 -79.159 53.062
56986 93.537 -79.154 53.053
56987 93.551 -79.149 53.044
56988 93.566 -79.144 53.033
56989 93.578 -79.142 53.022
56990 93.590 -79.140 53.013
56991 93.603 -79.139 53.004
56992 93.615 -79.137 52.998
56993 93.627 -79.136 52.991
56994 93.639 -79.134 52.985
56995 93.651 -79.133 52.980
56996 93.663 -79.131 52.974
56997 93.676 -79.131 52.968
56998 93.687 -79.134 52.962
56999 93.702 -79.134 52.954
57000 93.716 -79.133 52.945
57001 93.731 -79.133 52.934
57002 93.746 -79.133 52.924
57003 93.761 -79.132 52.913
57004 93.775 -79.132 52.902
57005 93.790 -79.132 52.890
57006 93.804 -79.132 52.879
57007 93.819 -79.132 52.866
57008 93.834 -79.131 52.854
57009 93.848 -79.131 52.842
57010 93.846 -79.128 52.830
57011 93.837 -79.125 52.816
57012 93.838 -79.123 52.800
57013 93.843 -79.122 52.783
57014 93.847 -79.121 52.766
57015 93.852 -79.119 52.747
57016 93.868 -79.120 52.728
57017 93.889 -79.121 52.709
57018 93.900 -79.123 52.692
57019 93.908 -79.126 52.676
57020 93.915 -79.128 52.662
57021 93.922 -79.131 52.650
57022 93.928 -79.134 52.639
57023 93.930 -79.138 52.627
57024 93.932 -79.143 52.615
57025 93.934 -79.147 52.603
57026 93.936 -79.151 52.591
57027 93.938 -79.156 52.578
57028 93.940 -79.159 52.564
57029 93.942 -79.164 52.549
57030 93.944 -79.169 52.534
57031 93.946 -79.172 52.519
57032 93.948 -79.176 52.504
57033 93.953 -79.172 52.489
57034 93.959 -79.168 52.475
57035 93.965 -79.165 52.460
57036 93.972 -79.161 52.449
57037 93.978 -79.157 52.438
57038 93.984 -79.154 52.426
57039 93.991 -79.150 52.414
57040 93.997 -79.149 52.400
57041 94.002 -79.159 52.382
57042 94.006 -79.159 52.362
57043 94.010 -79.157 52.338
57044 94.009 -79.160 52.314
57045 94.008 -79.163 52.288
57046 94.007 -79.166 52.262
57047 94.006 -79.170 52.238
57048 94.005 -79.173 52.213
57049 94.004 -79.177 52.186
57050 94.003 -79.180 52.161
57051 94.003 -79.182 52.136
57052 94.003 -79.184 52.112
57053 94.011 -79.183 52.086
57054 94.022 -79.181 52.061
57055 94.033 -79.181 52.034
57056 94.044 -79.180 52.005
57057 94.055 -79.178 51.976
57058 94.066 -79.178 51.947
57059 94.073 -79.175 51.918
57060 94.072 -79.172 51.892
57061 94.070 -79.169 51.868
57062 94.069 -79.165 51.847
57063 94.068 -79.161 51.827
57064 94.069 -79.155 51.810
57065 94.091 -79.131 51.792
57066 94.113 -79.105 51.770
57067 94.135 -79.080 51.752
57068 94.156 -79.054 51.733
57069 94.179 -79.029 51.714
57070 94.200 -79.003 51.694
57071 94.222 -78.978 51.673
57072 94.250 -78.987 51.649
57073 94.263 -78.972 51.621
57074 94.274 -78.953 51.586
57075 94.294 -78.963 51.548
57076 94.314 -78.973 51.507
57077 94.334 -78.982 51.465
57078 94.354 -78.992 51.423
57079 94.373 -79.002 51.381
57080 94.394 -79.012 51.341
57081 94.415 -79.015 51.303
57082 94.437 -79.006 51.266
57083 94.461 -78.999 51.231
57084 94.484 -78.991 51.195
57085 94.506 -78.984 51.158
57086 94.530 -78.977 51.119
57087 94.547 -78.975 51.081
57088 94.558 -78.971 51.042
57089 94.569 -78.968 51.003
57090 94.580 -78.964 50.965
57091 94.592 -78.961 50.926
57092 94.603 -78.957 50.887
57093 94.614 -78.953 50.849
57094 94.626 -78.949 50.811
57095 94.637 -78.946 50.773
57096 94.648 -78.942 50.736
57097 94.660 -78.939 50.697
57098 94.671 -78.935 50.659
57099 94.682 -78.931 50.621
57100 94.693 -78.927 50.583
57101 94.702 -78.933 50.545
57102 94.710 -78.940 50.507
57103 94.718 -78.948 50.469
57104 94.727 -78.955 50.431
57105 94.735 -78.963 50.393
57106 94.744 -78.971 50.355
57107 94.752 -78.978 50.317
57108 94.760 -78.985 50.280
57109 94.768 -78.993 50.242
57110 94.776 -79.000 50.204
57111 94.785 -79.008 50.166
57112 94.793 -79.015 50.128
57113 94.802 -79.023 50.090
57114 94.810 -79.031 50.053
57115 94.818 -79.038 50.015
57116 94.826 -79.045 49.978
57117 94.835 -79.053 49.939
57123 94.884 -79.097 49.714
57124 94.893 -79.105 49.676
57125 94.901 -79.112 49.638
57126 94.909 -79.120 49.601
57127 94.917 -79.128 49.564
57128 94.926 -79.135 49.526
57129 94.934 -79.143 49.489
57130 94.942 -79.150 49.451
57131 94.951 -79.157 49.413
57132 94.959 -79.165 49.376
57133 94.967 -79.172 49.340
57134 94.975 -79.180 49.304
57135 94.984 -79.187 49.267
57136 94.996 -79.205 49.232
57137 95.008 -79.223 49.196
57138 95.020 -79.241 49.160
57139 95.033 -79.259 49.125
57140 95.045 -79.278 49.090
57141 95.057 -79.296 49.056
57142 95.069 -79.315 49.021
57143 95.082 -79.333 48.986
57144 95.090 -79.347 48.952
57145 95.104 -79.357 48.917
57146 95.116 -79.358 48.882
57147 95.127 -79.360 48.846
57148 95.139 -79.361 48.809
57149 95.151 -79.363 48.772
57150 95.175 -79.366 48.734
57151 95.199 -79.372 48.695
57152 95.224 -79.376 48.657
57153 95.249 -79.381 48.619
57154 95.274 -79.386 48.581
57155 95.299 -79.391 48.543
57156 95.323 -79.395 48.505
57157 95.348 -79.400 48.467
57158 95.373 -79.405 48.429
57159 95.397 -79.409 48.392
57160 95.427 -79.413 48.356
57161 95.465 -79.419 48.320
57162 95.503 -79.423 48.281
57163 95.542 -79.428 48.244
57164 95.570 -79.425 48.207
57165 95.580 -79.419 48.169
57166 95.590 -79.413 48.131
57167 95.600 -79.408 48.094
57168 95.610 -79.401 48.058
57169 95.620 -79.393 48.027
57170 95.629 -79.385 47.994
57171 95.638 -79.376 47.961
57172 95.648 -79.368 47.930
57173 95.657 -79.360 47.898
57174 95.667 -79.352 47.867
57175 95.676 -79.344 47.837
57176 95.686 -79.335 47.804
57177 95.695 -79.342 47.770
57178 95.704 -79.353 47.735
57179 95.713 -79.363 47.701
57180 95.722 -79.373 47.666
57181 95.732 -79.383 47.630
57182 95.741 -79.394 47.594
57183 95.768 -79.407 47.557
57184 95.818 -79.404 47.519
57185 95.872 -79.401 47.481
57186 95.925 -79.397 47.442
57187 95.945 -79.408 47.403
57188 95.965 -79.419 47.366
57189 95.984 -79.430 47.330
57190 96.004 -79.441 47.296
57191 96.023 -79.452 47.262
57192 96.043 -79.463 47.229
57193 96.062 -79.474 47.196
57194 96.084 -79.482 47.163
57195 96.108 -79.485 47.130
57196 96.133 -79.488 47.098
57197 96.159 -79.490 47.066
57198 96.184 -79.492 47.034
57199 96.209 -79.494 47.002
57200 96.234 -79.497 46.970
57201 96.259 -79.499 46.939
57202 96.285 -79.501 46.908
57203 96.310 -79.504 46.875
57204 96.334 -79.507 46.845
57205 96.360 -79.509 46.816
57206 96.385 -79.511 46.785
57207 96.410 -79.513 46.755
57208 96.435 -79.516 46.725
57209 96.460 -79.518 46.694
57210 96.485 -79.521 46.663
57211 96.511 -79.523 46.635
57212 96.536 -79.526 46.608
57213 96.560 -79.528 46.581
57214 96.586 -79.530 46.556
57215 96.611 -79.532 46.530
57216 96.636 -79.535 46.504
57217 96.661 -79.538 46.479
57218 96.686 -79.540 46.455
57219 96.712 -79.542 46.432
57220 96.737 -79.545 46.409
57221 96.762 -79.547 46.385
57222 96.787 -79.549 46.361
57223 96.812 -79.551 46.337
57224 96.837 -79.554 46.311
57225 96.862 -79.557 46.288
57226 96.884 -79.559 46.267
57227 96.895 -79.571 46.246
57228 96.906 -79.584 46.225
57229 96.917 -79.596 46.204
57230 96.928 -79.609 46.182
57231 96.969 -79.608 46.160
57232 97.008 -79.604 46.138
57233 97.030 -79.601 46.115
57234 97.051 -79.598 46.095
57235 97.073 -79.594 46.075
57236 97.095 -79.590 46.053
57237 97.119 -79.585 46.027
57238 97.147 -79.579 46.003
57239 97.174 -79.568 45.977
57240 97.202 -79.557 45.948
57241 97.181 -79.589 45.917
57242 97.155 -79.628 45.889
57243 97.139 -79.657 45.863
57244 97.146 -79.664 45.838
57245 97.153 -79.672 45.813
57246 97.160 -79.679 45.790
57247 97.167 -79.688 45.769
57248 97.174 -79.695 45.747
57249 97.181 -79.703 45.725
57250 97.188 -79.711 45.704
57251 97.195 -79.718 45.682
57252 97.203 -79.726 45.661
57253 97.210 -79.733 45.641
57254 97.217 -79.741 45.623
57255 97.224 -79.748 45.605
57660 93.175 -79.351 53.405
57661 93.189 -79.339 53.392
57662 93.203 -79.327 53.377
57663 93.217 -79.315 53.363
57664 93.231 -79.303 53.349
57665 93.246 -79.291 53.336
57666 93.260 -79.278 53.322
57667 93.273 -79.266 53.307
57668 93.288 -79.253 53.293
57669 93.304 -79.247 53.278
57670 93.314 -79.244 53.263
57671 93.324 -79.241 53.249
57672 93.334 -79.239 53.236
57673 93.343 -79.237 53.223
57674 93.355 -79.226 53.209
57675 93.366 -79.216 53.196
57676 93.377 -79.206 53.182
57677 93.389 -79.196 53.169
57678 93.396 -79.194 53.156
57679 93.415 -79.181 53.144
57680 93.430 -79.175 53.134
57681 93.444 -79.171 53.123
57682 93.459 -79.166 53.111
57683 93.473 -79.161 53.102
57684 93.488 -79.156 53.093
57685 93.502 -79.151 53.086
57686 93.516 -79.147 53.081
57687 93.531 -79.141 53.075
57688 93.546 -79.137 53.068
57689 93.560 -79.131 53.058
57690 93.573 -79.128 53.047
57691 93.585 -79.127 53.038
57692 93.598 -79.125 53.030
57693 93.610 -79.124 53.023
57694 93.622 -79.122 53.016
57695 93.634 -79.121 53.010
57696 93.646 -79.119 53.007
57697 93.658 -79.118 53.001
57698 93.666 -79.123 52.995
57699 93.679 -79.125 52.988
57700 93.695 -79.125 52.980
57701 93.710 -79.125 52.971
57702 93.725 -79.125 52.960
57703 93.740 -79.124 52.950
57704 93.755 -79.124 52.939
57705 93.769 -79.123 52.928
57706 93.784 -79.123 52.916
57707 93.798 -79.123 52.905
57708 93.813 -79.123 52.892
57709 93.828 -79.122 52.880
57710 93.842 -79.122 52.869
57711 93.840 -79.119 52.856
57712 93.839 -79.117 52.843
57713 93.844 -79.116 52.827
57714 93.848 -79.115 52.810
57715 93.853 -79.114 52.792
57716 93.857 -79.112 52.774
57717 93.862 -79.112 52.755
57718 93.881 -79.112 52.736
57719 93.890 -79.115 52.718
57720 93.897 -79.118 52.701
57721 93.905 -79.120 52.688
57722 93.912 -79.122 52.675
57723 93.919 -79.125 52.665
57724 93.921 -79.130 52.652
57725 93.923 -79.134 52.639
57726 93.925 -79.138 52.626
57727 93.927 -79.143 52.614
57728 93.929 -79.147 52.601
57729 93.931 -79.151 52.587
57730 93.933 -79.156 52.574
57731 93.935 -79.158 52.561
57732 93.939 -79.160 52.547
57733 93.942 -79.163 52.532
57734 93.946 -79.169 52.518
57735 93.953 -79.168 52.503
57736 93.960 -79.167 52.488
57737 93.967 -79.165 52.476
57738 93.973 -79.164 52.463
57739 93.980 -79.160 52.451
57740 93.986 -79.157 52.441
57741 93.992 -79.153 52.427
57742 93.999 -79.151 52.409
57743 94.003 -79.156 52.388
57744 94.004 -79.156 52.365
57745 94.003 -79.160 52.341
57746 94.002 -79.163 52.314
57747 94.001 -79.166 52.288
57748 94.001 -79.168 52.263
57749 94.001 -79.169 52.237
57750 94.002 -79.170 52.209
57751 94.002 -79.172 52.183
57752 94.002 -79.173 52.156
57753 94.002 -79.175 52.129
57754 94.010 -79.173 52.103
57755 94.019 -79.171 52.076
57756 94.030 -79.169 52.048
57757 94.041 -79.169 52.017
57758 94.052 -79.168 51.989
57759 94.063 -79.166 51.960
57760 94.074 -79.165 51.932
57761 94.073 -79.162 51.908
57762 94.072 -79.158 51.888
57763 94.070 -79.155 51.869
57764 94.069 -79.151 51.851
57765 94.082 -79.131 51.834
57766 94.101 -79.106 51.814
57767 94.122 -79.084 51.792
57768 94.143 -79.062 51.772
57769 94.165 -79.037 51.753
57770 94.187 -79.012 51.734
57771 94.209 -78.987 51.713
57772 94.231 -78.961 51.691
57773 94.257 -78.961 51.665
57774 94.267 -78.940 51.636
57775 94.286 -78.949 51.599
57776 94.306 -78.959 51.560
57777 94.326 -78.968 51.519
57778 94.346 -78.978 51.476
57779 94.366 -78.988 51.435
57780 94.386 -78.998 51.394
57781 94.406 -79.007 51.354
57782 94.426 -79.017 51.317
57783 94.447 -79.023 51.281
57784 94.469 -79.015 51.247
57785 94.492 -79.007 51.212
57786 94.514 -78.996 51.176
57787 94.535 -78.970 51.139
57788 94.553 -78.971 51.100
57789 94.563 -78.964 51.060
57790 94.574 -78.959 51.020
57791 94.586 -78.955 50.980
57792 94.597 -78.952 50.941
57793 94.609 -78.948 50.902
57794 94.620 -78.944 50.864
57795 94.631 -78.940 50.825
57796 94.642 -78.937 50.787
57797 94.653 -78.933 50.749
57798 94.665 -78.930 50.710
57799 94.676 -78.926 50.672
57800 94.687 -78.922 50.634
57801 94.699 -78.918 50.596
57802 94.710 -78.915 50.558
57803 94.718 -78.921 50.521
57804 94.727 -78.927 50.482
57805 94.735 -78.935 50.443
57806 94.744 -78.943 50.405
57807 94.752 -78.950 50.367
57808 94.760 -78.958 50.329
57809 94.768 -78.965 50.292
57810 94.777 -78.973 50.253
57811 94.785 -78.980 50.215
57812 94.793 -78.987 50.178
57813 94.802 -78.995 50.139
57814 94.810 -79.002 50.101
57815 94.818 -79.010 50.065
57816 94.826 -79.018 50.027
57817 94.835 -79.025 49.990
57818 94.843 -79.032 49.953
57819 94.851 -79.040 49.915
57820 94.859 -79.047 49.877
57821 94.868 -79.055 49.840
57822 94.876 -79.062 49.802
57823 94.884 -79.070 49.765
57824 94.893 -79.078 49.727
57825 94.901 -79.085 49.689
57826 94.909 -79.092 49.650
57827 94.917 -79.100 49.612
57828 94.926 -79.107 49.575
57829 94.934 -79.115 49.538
57830 94.943 -79.122 49.501
57831 94.951 -79.130 49.463
57832 94.959 -79.137 49.424
57833 94.967 -79.144 49.387
57834 94.979 -79.160 49.351
57835 94.991 -79.178 49.315
57836 95.003 -79.197 49.279
57837 95.016 -79.215 49.243
57838 95.028 -79.233 49.208
57839 95.040 -79.251 49.173
57840 95.052 -79.270 49.138
57841 95.062 -79.286 49.104
57842 95.069 -79.299 49.069
57843 95.075 -79.311 49.035
57844 95.082 -79.324 49.001
57845 95.096 -79.337 48.967
57846 95.114 -79.345 48.933
57847 95.129 -79.348 48.897
57848 95.141 -79.350 48.861
57849 95.153 -79.351 48.823
57850 95.164 -79.353 48.785
57851 95.187 -79.357 48.748
57852 95.212 -79.362 48.710
57853 95.237 -79.366 48.672
57854 95.261 -79.371 48.634
57855 95.286 -79.376 48.597
57856 95.311 -79.381 48.558
57857 95.336 -79.385 48.520
57858 95.361 -79.390 48.483
57859 95.386 -79.395 48.445
57860 95.410 -79.400 48.409
57861 95.430 -79.400 48.373
57862 95.461 -79.408 48.337
57863 95.500 -79.413 48.298
57864 95.540 -79.416 48.259
57865 95.569 -79.412 48.221
57866 95.583 -79.404 48.182
57867 95.593 -79.399 48.143
57868 95.603 -79.394 48.105
57869 95.613 -79.388 48.068
57870 95.623 -79.382 48.035
57871 95.633 -79.377 48.002
57872 95.644 -79.372 47.968
57873 95.654 -79.366 47.935
57874 95.664 -79.360 47.902
57875 95.674 -79.355 47.869
57876 95.684 -79.347 47.836
57877 95.693 -79.338 47.802
57878 95.702 -79.338 47.767
57879 95.712 -79.349 47.731
57880 95.721 -79.359 47.697
57881 95.731 -79.369 47.663
57882 95.741 -79.379 47.628
57883 95.751 -79.389 47.593
57884 95.774 -79.395 47.557
57885 95.827 -79.391 47.520
57886 95.879 -79.386 47.482
57887 95.932 -79.382 47.442
57888 95.961 -79.378 47.403
57889 95.981 -79.386 47.365
57890 96.002 -79.394 47.329
57891 96.022 -79.402 47.295
57892 96.042 -79.410 47.261
57893 96.062 -79.419 47.228
57894 96.082 -79.431 47.194
57895 96.101 -79.441 47.161
57896 96.121 -79.453 47.128
57897 96.141 -79.463 47.096
57898 96.160 -79.475 47.063
57899 96.180 -79.485 47.031
57900 96.199 -79.497 47.000
57901 96.219 -79.507 46.968
57902 96.243 -79.510 46.937
57903 96.268 -79.513 46.905
57904 96.294 -79.515 46.873
57905 96.319 -79.517 46.843
57906 96.344 -79.520 46.813
57907 96.369 -79.523 46.782
57908 96.394 -79.525 46.752
57909 96.419 -79.527 46.722
57910 96.444 -79.529 46.691
57911 96.469 -79.532 46.660
57912 96.495 -79.534 46.632
57913 96.520 -79.537 46.605
57914 96.545 -79.539 46.578
57915 96.570 -79.541 46.553
57916 96.595 -79.544 46.528
57917 96.620 -79.546 46.502
57918 96.645 -79.548 46.477
57919 96.670 -79.551 46.454
57920 96.695 -79.554 46.432
57921 96.721 -79.556 46.409
57922 96.746 -79.558 46.386
57923 96.771 -79.560 46.362
57924 96.796 -79.563 46.338
57925 96.821 -79.565 46.313
57926 96.851 -79.563 46.291
57927 96.877 -79.558 46.270
57928 96.897 -79.554 46.249
57929 96.910 -79.563 46.228
57930 96.921 -79.576 46.206
57931 96.936 -79.586 46.182
57932 96.978 -79.582 46.159
57933 97.000 -79.579 46.135
57934 97.021 -79.575 46.113
57935 97.045 -79.570 46.091
57936 97.073 -79.564 46.069
57937 97.100 -79.557 46.047
57938 97.127 -79.551 46.023
57939 97.155 -79.544 45.998
57940 97.182 -79.532 45.972
57941 97.199 -79.530 45.944
57942 97.214 -79.536 45.915
57943 97.213 -79.541 45.888
57944 97.187 -79.578 45.862
57945 97.161 -79.616 45.838
57946 97.153 -79.638 45.813
57947 97.160 -79.645 45.789
57948 97.167 -79.653 45.766
57949 97.174 -79.661 45.745
57950 97.181 -79.668 45.724
57951 97.188 -79.676 45.702
57952 97.195 -79.683 45.681
57953 97.203 -79.691 45.660
57954 97.210 -79.698 45.640
57955 97.217 -79.707 45.622
57956 97.224 -79.714 45.604
57957 97.231 -79.722 45.585
58362 93.188 -79.328 53.417
58363 93.202 -79.316 53.404
58364 93.216 -79.304 53.389
58365 93.230 -79.292 53.375
58366 93.245 -79.280 53.361
58367 93.259 -79.268 53.347
58368 93.273 -79.256 53.332
58369 93.287 -79.244 53.317
58370 93.305 -79.235 53.302
58371 93.314 -79.233 53.288
58372 93.324 -79.231 53.273
58373 93.334 -79.228 53.260
58374 93.344 -79.225 53.247
58375 93.353 -79.222 53.233
58376 93.365 -79.212 53.219
58377 93.376 -79.202 53.205
58378 93.388 -79.191 53.191
58379 93.398 -79.182 53.178
58380 93.406 -79.178 53.165
58381 93.417 -79.175 53.153
58382 93.436 -79.162 53.142
58383 93.453 -79.153 53.130
58384 93.468 -79.149 53.120
58385 93.482 -79.143 53.111
58386 93.496 -79.139 53.104
58387 93.511 -79.134 53.099
58388 93.525 -79.129 53.095
58389 93.540 -79.125 53.091
58390 93.554 -79.119 53.083
58391 93.569 -79.115 53.074
58392 93.581 -79.113 53.065
58393 93.593 -79.112 53.058
58394 93.605 -79.110 53.051
58395 93.617 -79.109 53.045
58396 93.630 -79.107 53.039
58397 93.642 -79.106 53.034
58398 93.649 -79.110 53.029
58399 93.657 -79.115 53.023
58400 93.671 -79.117 53.016
58401 93.688 -79.116 53.007
58402 93.703 -79.116 52.997
58403 93.718 -79.116 52.987
58404 93.733 -79.115 52.976
58405 93.748 -79.115 52.966
58406 93.763 -79.115 52.955
58407 93.777 -79.115 52.943
58408 93.792 -79.114 52.931
58409 93.807 -79.114 52.919
58410 93.822 -79.113 52.906
58411 93.837 -79.113 52.894
58412 93.840 -79.112 52.882
58413 93.845 -79.110 52.869
58414 93.849 -79.109 52.855
58415 93.854 -79.108 52.838
58416 93.858 -79.107 52.819
58417 93.863 -79.106 52.800
58418 93.867 -79.105 52.781
58419 93.872 -79.104 52.762
58420 93.880 -79.106 52.743
58421 93.887 -79.109 52.727
58422 93.894 -79.112 52.713
58423 93.902 -79.114 52.701
58424 93.909 -79.117 52.689
58425 93.912 -79.121 52.677
58426 93.914 -79.125 52.663
58427 93.915 -79.130 52.649
58428 93.918 -79.133 52.637
58429 93.921 -79.135 52.624
58430 93.924 -79.137 52.611
58431 93.928 -79.140 52.599
58432 93.931 -79.142 52.586
58433 93.934 -79.144 52.573
58434 93.937 -79.147 52.559
58435 93.941 -79.155 52.546
58436 93.946 -79.160 52.531
58437 93.953 -79.159 52.516
58438 93.960 -79.158 52.504
58439 93.966 -79.157 52.492
58440 93.973 -79.156 52.479
58441 93.980 -79.154 52.468
58442 93.987 -79.153 52.453
58443 93.994 -79.152 52.436
58444 94.000 -79.151 52.414
58445 94.000 -79.153 52.391
58446 94.000 -79.154 52.366
58447 94.000 -79.156 52.339
58448 94.000 -79.157 52.312
58449 94.000 -79.159 52.285
58450 94.000 -79.159 52.258
58451 94.000 -79.161 52.228
58452 94.001 -79.162 52.201
58453 94.001 -79.164 52.172
58454 94.002 -79.165 52.143
58455 94.010 -79.163 52.116
58456 94.019 -79.161 52.089
58457 94.028 -79.159 52.060
58458 94.037 -79.158 52.030
58459 94.048 -79.156 52.003
58460 94.059 -79.156 51.976
58461 94.070 -79.155 51.949
58462 94.074 -79.152 51.928
58463 94.073 -79.148 51.910
58464 94.071 -79.144 51.893
58465 94.076 -79.134 51.875
58466 94.095 -79.108 51.856
58467 94.113 -79.081 51.836
58468 94.133 -79.057 51.813
58469 94.154 -79.035 51.792
58470 94.176 -79.013 51.773
58471 94.197 -78.991 51.754
58472 94.218 -78.969 51.732
58473 94.240 -78.944 51.708
58474 94.260 -78.929 51.681
58475 94.278 -78.935 51.649
58476 94.298 -78.945 51.611
58477 94.319 -78.954 51.571
58478 94.338 -78.964 51.531
58479 94.358 -78.974 51.488
58480 94.378 -78.984 51.447
58481 94.398 -78.993 51.406
58482 94.418 -79.003 51.367
58483 94.438 -79.013 51.331
58484 94.458 -79.023 51.297
58485 94.478 -79.030 51.264
58486 94.495 -78.984 51.229
58487 94.516 -78.968 51.194
58488 94.538 -78.959 51.157
58489 94.556 -78.961 51.118
58490 94.570 -78.962 51.078
58491 94.581 -78.955 51.037
58492 94.592 -78.949 50.998
58493 94.603 -78.943 50.958
58494 94.614 -78.939 50.918
58495 94.625 -78.935 50.879
58496 94.636 -78.931 50.840
58497 94.647 -78.927 50.801
58498 94.659 -78.924 50.763
58499 94.670 -78.920 50.725
58500 94.682 -78.917 50.687
58501 94.693 -78.913 50.648
58502 94.704 -78.909 50.610
58503 94.715 -78.905 50.572
58504 94.726 -78.902 50.535
58505 94.736 -78.908 50.496
58506 94.744 -78.915 50.456
58507 94.752 -78.922 50.418
58508 94.760 -78.930 50.380
58509 94.768 -78.937 50.342
58510 94.777 -78.945 50.305
58511 94.785 -78.952 50.267
58512 94.793 -78.960 50.229
58513 94.802 -78.968 50.191
58514 94.810 -78.974 50.153
58515 94.818 -78.982 50.115
58516 94.827 -78.990 50.078
58517 94.835 -78.997 50.041
58518 94.843 -79.005 50.003
58519 94.851 -79.012 49.966
58520 94.859 -79.020 49.928
58521 94.868 -79.027 49.890
58522 94.876 -79.034 49.852
58523 94.885 -79.042 49.816
58524 94.893 -79.049 49.778
58525 94.901 -79.057 49.740
58526 94.909 -79.065 49.702
58527 94.917 -79.072 49.663
58528 94.926 -79.079 49.625
58529 94.934 -79.087 49.588
58530 94.943 -79.094 49.551
58531 94.951 -79.102 49.512
58532 94.961 -79.115 49.474
58533 94.974 -79.134 49.437
58534 94.986 -79.152 49.400
58535 94.998 -79.170 49.364
58536 95.011 -79.188 49.328
58537 95.023 -79.207 49.292
58538 95.034 -79.225 49.256
58539 95.041 -79.237 49.221
58540 95.047 -79.250 49.187
58541 95.054 -79.263 49.153
58542 95.060 -79.275 49.119
58543 95.066 -79.288 49.085
58544 95.074 -79.300 49.051
58545 95.089 -79.313 49.017
58546 95.103 -79.325 48.983
58547 95.122 -79.333 48.948
58548 95.142 -79.338 48.912
58549 95.154 -79.340 48.875
58550 95.166 -79.341 48.837
58551 95.178 -79.343 48.801
58552 95.199 -79.347 48.764
58553 95.225 -79.352 48.726
58554 95.249 -79.357 48.689
58555 95.274 -79.361 48.652
58556 95.299 -79.366 48.614
58557 95.324 -79.371 48.576
58558 95.348 -79.375 48.539
58559 95.373 -79.381 48.502
58560 95.398 -79.385 48.465
58561 95.416 -79.380 48.427
58562 95.431 -79.388 48.389
58563 95.458 -79.397 48.350
58564 95.497 -79.401 48.311
58565 95.537 -79.406 48.271
58566 95.566 -79.400 48.232
58567 95.585 -79.390 48.193
58568 95.595 -79.384 48.153
58569 95.606 -79.379 48.113
58570 95.616 -79.373 48.074
58571 95.626 -79.368 48.039
58572 95.636 -79.362 48.004
58573 95.647 -79.357 47.969
58574 95.657 -79.351 47.935
58575 95.667 -79.346 47.903
58576 95.677 -79.340 47.870
58577 95.687 -79.335 47.836
58578 95.697 -79.329 47.800
58579 95.713 -79.332 47.764
58580 95.724 -79.344 47.728
58581 95.734 -79.354 47.695
58582 95.744 -79.365 47.660
58583 95.754 -79.375 47.625
58584 95.764 -79.385 47.589
58585 95.789 -79.390 47.553
58586 95.841 -79.385 47.516
58587 95.894 -79.381 47.478
58588 95.923 -79.360 47.439
58589 95.942 -79.344 47.401
58590 95.989 -79.366 47.364
58591 96.014 -79.375 47.328
58592 96.034 -79.384 47.293
58593 96.054 -79.392 47.259
58594 96.075 -79.400 47.226
58595 96.095 -79.408 47.192
58596 96.116 -79.416 47.159
58597 96.136 -79.425 47.126
58598 96.157 -79.432 47.093
58599 96.177 -79.441 47.061
58600 96.197 -79.449 47.029
58601 96.218 -79.457 46.997
58602 96.238 -79.465 46.966
58603 96.258 -79.475 46.934
58604 96.277 -79.485 46.902
58605 96.297 -79.497 46.871
58606 96.316 -79.507 46.840
58607 96.336 -79.519 46.809
58608 96.356 -79.529 46.779
58609 96.378 -79.535 46.748
58610 96.403 -79.538 46.717
58611 96.429 -79.541 46.687
58612 96.454 -79.543 46.656
58613 96.478 -79.545 46.629
58614 96.504 -79.548 46.602
58615 96.529 -79.550 46.575
58616 96.554 -79.552 46.550
58617 96.579 -79.555 46.525
58618 96.604 -79.557 46.500
58619 96.629 -79.560 46.476
58620 96.655 -79.562 46.453
58621 96.680 -79.564 46.431
58622 96.704 -79.567 46.408
58623 96.730 -79.569 46.386
58624 96.755 -79.572 46.363
58625 96.780 -79.574 46.341
58626 96.811 -79.571 46.316
58627 96.844 -79.566 46.294
58628 96.869 -79.561 46.274
58629 96.888 -79.557 46.252
58630 96.908 -79.553 46.230
58631 96.925 -79.554 46.208
58632 96.948 -79.560 46.184
58633 96.970 -79.556 46.158
58634 96.998 -79.550 46.133
58635 97.025 -79.544 46.110
58636 97.053 -79.537 46.086
58637 97.081 -79.531 46.063
58638 97.108 -79.525 46.040
58639 97.136 -79.518 46.016
58640 97.163 -79.507 45.991
58641 97.183 -79.502 45.966
58642 97.196 -79.505 45.940
58643 97.208 -79.508 45.913
58644 97.223 -79.517 45.887
58645 97.222 -79.522 45.862
58646 97.220 -79.527 45.837
58647 97.193 -79.566 45.812
58648 97.167 -79.604 45.788
58649 97.167 -79.618 45.765
58650 97.174 -79.626 45.743
58651 97.181 -79.634 45.721
58652 97.188 -79.642 45.700
58653 97.195 -79.649 45.679
58654 97.203 -79.657 45.658
58655 97.210 -79.664 45.638
58656 97.217 -79.672 45.620
58657 97.224 -79.679 45.602
58658 97.231 -79.687 45.584
58659 97.238 -79.695 45.566
58673 97.337 -79.802 45.313
59064 93.201 -79.306 53.429
59065 93.215 -79.294 53.415
59066 93.229 -79.281 53.401
59067 93.243 -79.269 53.386
59068 93.258 -79.257 53.372
59069 93.272 -79.245 53.357
59070 93.286 -79.233 53.342
59071 93.305 -79.225 53.327
59072 93.314 -79.222 53.312
59073 93.325 -79.219 53.297
59074 93.334 -79.217 53.284
59075 93.344 -79.215 53.270
59076 93.353 -79.212 53.257
59077 93.363 -79.208 53.243
59078 93.375 -79.198 53.228
59079 93.387 -79.187 53.214
59080 93.398 -79.177 53.200
59081 93.407 -79.169 53.187
59082 93.416 -79.162 53.174
59083 93.422 -79.165 53.161
59084 93.438 -79.156 53.150
59085 93.457 -79.143 53.140
59086 93.476 -79.131 53.130
59087 93.490 -79.126 53.122
59088 93.505 -79.122 53.117
59089 93.519 -79.117 53.115
59090 93.534 -79.112 53.113
59091 93.548 -79.107 53.109
59092 93.563 -79.102 53.101
59093 93.576 -79.099 53.094
59094 93.588 -79.097 53.086
59095 93.601 -79.096 53.080
59096 93.613 -79.094 53.074
59097 93.625 -79.093 53.068
59098 93.633 -79.097 53.063
59099 93.640 -79.102 53.057
59100 93.648 -79.107 53.051
59101 93.663 -79.108 53.043
59102 93.681 -79.108 53.034
59103 93.697 -79.108 53.025
59104 93.711 -79.107 53.015
59105 93.726 -79.107 53.005
59106 93.741 -79.106 52.994
59107 93.756 -79.106 52.984
59108 93.771 -79.106 52.972
59109 93.786 -79.106 52.959
59110 93.800 -79.106 52.947
59111 93.816 -79.105 52.934
59112 93.830 -79.103 52.921
59113 93.837 -79.101 52.908
59114 93.841 -79.100 52.894
59115 93.845 -79.099 52.880
59116 93.849 -79.097 52.864
59117 93.853 -79.096 52.845
59118 93.856 -79.094 52.827
59119 93.860 -79.093 52.808
59120 93.868 -79.090 52.788
59121 93.875 -79.092 52.769
59122 93.882 -79.096 52.753
59123 93.888 -79.100 52.739
59124 93.894 -79.103 52.726
59125 93.900 -79.106 52.714
59126 93.904 -79.109 52.700
59127 93.907 -79.112 52.686
59128 93.910 -79.114 52.672
59129 93.914 -79.116 52.661
59130 93.917 -79.118 52.648
59131 93.920 -79.122 52.636
59132 93.923 -79.124 52.623
59133 93.926 -79.126 52.611
59134 93.929 -79.128 52.599
59135 93.933 -79.131 52.586
59136 93.936 -79.140 52.573
59137 93.940 -79.148 52.560
59138 93.946 -79.152 52.545
59139 93.953 -79.150 52.533
59140 93.960 -79.150 52.521
59141 93.966 -79.148 52.508
59142 93.973 -79.147 52.495
59143 93.980 -79.146 52.480
59144 93.993 -79.140 52.461
59145 94.000 -79.139 52.440
59146 94.000 -79.141 52.416
59147 93.999 -79.143 52.390
59148 93.999 -79.146 52.362
59149 93.999 -79.148 52.334
59150 93.999 -79.149 52.306
59151 93.999 -79.150 52.276
59152 93.999 -79.152 52.244
59153 93.999 -79.153 52.215
59154 94.000 -79.155 52.185
59155 94.002 -79.156 52.155
59156 94.010 -79.153 52.127
59157 94.019 -79.152 52.100
59158 94.028 -79.150 52.072
59159 94.036 -79.148 52.043
59160 94.045 -79.146 52.017
59161 94.056 -79.144 51.993
59162 94.067 -79.143 51.970
59163 94.075 -79.142 51.951
59164 94.074 -79.138 51.934
59165 94.072 -79.134 51.917
59166 94.088 -79.111 51.899
59167 94.107 -79.084 51.879
59168 94.126 -79.057 51.856
59169 94.144 -79.031 51.832
59170 94.165 -79.009 51.811
59171 94.186 -78.987 51.791
59172 94.207 -78.965 51.772
59173 94.228 -78.943 51.750
59174 94.250 -78.921 51.724
59175 94.266 -78.911 51.694
59176 94.285 -78.919 51.660
59177 94.304 -78.927 51.620
59178 94.324 -78.937 51.580
59179 94.343 -78.945 51.541
59180 94.362 -78.953 51.500
59181 94.382 -78.962 51.460
59182 94.401 -78.971 51.419
59183 94.420 -78.979 51.381
59184 94.439 -78.987 51.346
59185 94.459 -78.996 51.313
59186 94.476 -78.971 51.281
59187 94.498 -78.965 51.247
59188 94.520 -78.956 51.212
59189 94.542 -78.949 51.175
59190 94.559 -78.951 51.136
59191 94.575 -78.956 51.095
59192 94.588 -78.953 51.055
59193 94.599 -78.947 51.016
59194 94.609 -78.940 50.976
59195 94.620 -78.934 50.936
59196 94.631 -78.927 50.895
59197 94.642 -78.922 50.856
59198 94.653 -78.918 50.817
59199 94.664 -78.915 50.778
59200 94.676 -78.911 50.740
59201 94.687 -78.908 50.702
59202 94.698 -78.904 50.663
59203 94.709 -78.900 50.625
59204 94.720 -78.896 50.587
59205 94.732 -78.893 50.549
59206 94.743 -78.889 50.511
59207 94.752 -78.895 50.471
59208 94.760 -78.902 50.433
59209 94.769 -78.909 50.395
59210 94.777 -78.917 50.356
59211 94.785 -78.924 50.318
59212 94.793 -78.932 50.281
59213 94.802 -78.940 50.243
59214 94.810 -78.947 50.204
59215 94.818 -78.955 50.167
59216 94.827 -78.962 50.129
59217 94.835 -78.969 50.092
59218 94.843 -78.977 50.054
59219 94.851 -78.984 50.017
59220 94.859 -78.992 49.979
59221 94.868 -78.999 49.942
59222 94.876 -79.007 49.904
59223 94.885 -79.015 49.866
59224 94.893 -79.021 49.830
59225 94.901 -79.029 49.792
59226 94.909 -79.037 49.754
59227 94.918 -79.044 49.715
59228 94.926 -79.052 49.677
59229 94.934 -79.059 49.639
59230 94.944 -79.071 49.601
59231 94.957 -79.089 49.564
59232 94.969 -79.107 49.525
59233 94.981 -79.125 49.487
59234 94.993 -79.143 49.450
59235 95.005 -79.162 49.413
59236 95.013 -79.176 49.377
59237 95.020 -79.188 49.341
59238 95.026 -79.201 49.305
59239 95.032 -79.214 49.271
59240 95.039 -79.227 49.237
59241 95.045 -79.239 49.202
59242 95.051 -79.252 49.169
59243 95.058 -79.265 49.136
59244 95.067 -79.277 49.102
59245 95.082 -79.290 49.068
59246 95.096 -79.303 49.033
59247 95.111 -79.315 48.999
59248 95.131 -79.321 48.963
59249 95.151 -79.326 48.926
59250 95.168 -79.331 48.889
59251 95.179 -79.332 48.852
59252 95.191 -79.333 48.816
59253 95.212 -79.338 48.780
59254 95.237 -79.342 48.743
59255 95.262 -79.347 48.705
59256 95.286 -79.351 48.667
59257 95.311 -79.357 48.630
59258 95.336 -79.361 48.592
59259 95.361 -79.366 48.553
59260 95.384 -79.368 48.515
59261 95.402 -79.362 48.476
59262 95.419 -79.360 48.436
59263 95.432 -79.375 48.396
59264 95.456 -79.386 48.356
59265 95.495 -79.391 48.316
59266 95.534 -79.394 48.276
59267 95.563 -79.389 48.237
59268 95.586 -79.375 48.197
59269 95.598 -79.369 48.156
59270 95.608 -79.364 48.116
59271 95.618 -79.358 48.075
59272 95.628 -79.353 48.038
59273 95.639 -79.347 48.002
59274 95.649 -79.342 47.966
59275 95.659 -79.336 47.932
59276 95.670 -79.331 47.899
59277 95.680 -79.325 47.865
59278 95.690 -79.320 47.831
59279 95.700 -79.314 47.794
59280 95.723 -79.317 47.757
59281 95.736 -79.340 47.721
59282 95.746 -79.350 47.687
59283 95.757 -79.360 47.652
59284 95.767 -79.370 47.616
59285 95.777 -79.381 47.581
59286 95.804 -79.385 47.544
59287 95.856 -79.380 47.508
59288 95.909 -79.375 47.470
59289 95.912 -79.337 47.433
59290 95.922 -79.310 47.396
59291 95.970 -79.332 47.359
59292 96.018 -79.353 47.324
59293 96.046 -79.366 47.289
59294 96.066 -79.373 47.255
59295 96.087 -79.382 47.222
59296 96.107 -79.390 47.188
59297 96.128 -79.398 47.155
59298 96.148 -79.407 47.122
59299 96.169 -79.414 47.088
59300 96.189 -79.422 47.056
59301 96.209 -79.431 47.024
59302 96.230 -79.439 46.993
59303 96.250 -79.447 46.962
59304 96.270 -79.455 46.929
59305 96.291 -79.463 46.897
59306 96.311 -79.472 46.865
59307 96.331 -79.479 46.835
59308 96.352 -79.488 46.804
59309 96.372 -79.496 46.774
59310 96.393 -79.504 46.743
59311 96.413 -79.512 46.713
59312 96.434 -79.520 46.682
59313 96.453 -79.530 46.652
59314 96.473 -79.541 46.625
59315 96.493 -79.552 46.598
59316 96.513 -79.561 46.571
59317 96.538 -79.563 46.546
59318 96.563 -79.566 46.521
59319 96.588 -79.568 46.497
59320 96.613 -79.571 46.473
59321 96.639 -79.573 46.450
59322 96.664 -79.576 46.428
59323 96.689 -79.578 46.407
59324 96.714 -79.580 46.386
59325 96.739 -79.582 46.365
59326 96.770 -79.579 46.344
59327 96.804 -79.574 46.321
59328 96.837 -79.568 46.300
59329 96.860 -79.564 46.279
59330 96.880 -79.560 46.256
59331 96.900 -79.554 46.233
59332 96.925 -79.543 46.209
59333 96.951 -79.529 46.185
59334 96.978 -79.523 46.156
59335 97.006 -79.516 46.130
59336 97.033 -79.510 46.107
59337 97.061 -79.504 46.081
59338 97.089 -79.498 46.057
59339 97.116 -79.491 46.034
59340 97.144 -79.483 46.010
59341 97.168 -79.475 45.986
59342 97.180 -79.477 45.962
59343 97.192 -79.480 45.937
59344 97.205 -79.483 45.911
59345 97.217 -79.486 45.886
59346 97.232 -79.498 45.860
59347 97.231 -79.503 45.835
59348 97.230 -79.508 45.810
59349 97.226 -79.516 45.786
59350 97.200 -79.554 45.764
59351 97.174 -79.592 45.741
59352 97.182 -79.599 45.719
59353 97.189 -79.607 45.697
59354 97.196 -79.614 45.677
59355 97.203 -79.622 45.656
59356 97.210 -79.629 45.637
59357 97.217 -79.637 45.618
59358 97.224 -79.645 45.601
59359 97.231 -79.653 45.583
59360 97.238 -79.661 45.564
59361 97.245 -79.668 45.547
59362 97.252 -79.676 45.528
59363 97.259 -79.683 45.509
59364 97.266 -79.691 45.492
59365 97.274 -79.698 45.474
59366 97.281 -79.706 45.456
59367 97.288 -79.714 45.438
59368 97.295 -79.721 45.419
59369 97.302 -79.729 45.400
59370 97.309 -79.736 45.382
59371 97.316 -79.744 45.365
59372 97.323 -79.751 45.348
59373 97.330 -79.760 45.331
59374 97.337 -79.767 45.314
59375 97.345 -79.775 45.296
59766 93.214 -79.284 53.440
59767 93.228 -79.272 53.426
59768 93.242 -79.259 53.411
59769 93.257 -79.247 53.397
59770 93.271 -79.235 53.382
59771 93.286 -79.223 53.368
59772 93.305 -79.214 53.353
59773 93.315 -79.211 53.337
59774 93.325 -79.209 53.322
59775 93.334 -79.206 53.308
59776 93.344 -79.204 53.295
59777 93.354 -79.201 53.281
59778 93.363 -79.199 53.267
59779 93.373 -79.194 53.252
59780 93.385 -79.184 53.238
59781 93.397 -79.173 53.224
59782 93.408 -79.163 53.210
59783 93.417 -79.156 53.197
59784 93.425 -79.148 53.184
59785 93.432 -79.148 53.172
59786 93.439 -79.150 53.161
59787 93.459 -79.138 53.150
59788 93.478 -79.125 53.141
59789 93.497 -79.112 53.137
59790 93.513 -79.104 53.135
59791 93.528 -79.100 53.134
59792 93.543 -79.094 53.133
59793 93.557 -79.090 53.129
59794 93.571 -79.085 53.123
59795 93.583 -79.084 53.116
59796 93.595 -79.082 53.110
59797 93.608 -79.081 53.104
59798 93.616 -79.084 53.097
59799 93.624 -79.090 53.092
59800 93.631 -79.094 53.086
59801 93.639 -79.099 53.079
59802 93.656 -79.100 53.071
59803 93.674 -79.100 53.062
59804 93.690 -79.100 53.053
59805 93.705 -79.099 53.043
59806 93.719 -79.099 53.034
59807 93.734 -79.098 53.023
59808 93.749 -79.098 53.012
59809 93.764 -79.097 53.000
59810 93.779 -79.097 52.987
59811 93.794 -79.096 52.974
59812 93.808 -79.095 52.960
59813 93.823 -79.091 52.946
59814 93.830 -79.090 52.932
59815 93.834 -79.088 52.918
59816 93.838 -79.087 52.904
59817 93.841 -79.085 52.888
59818 93.845 -79.084 52.870
59819 93.849 -79.083 52.850
59820 93.856 -79.081 52.831
59821 93.864 -79.078 52.813
59822 93.872 -79.075 52.794
59823 93.878 -79.079 52.778
59824 93.885 -79.083 52.765
59825 93.891 -79.087 52.752
59826 93.897 -79.090 52.739
59827 93.900 -79.093 52.726
59828 93.903 -79.096 52.713
59829 93.906 -79.098 52.701
59830 93.909 -79.100 52.689
59831 93.912 -79.103 52.676
59832 93.915 -79.105 52.662
59833 93.918 -79.107 52.650
59834 93.921 -79.110 52.638
59835 93.925 -79.112 52.625
59836 93.928 -79.116 52.612
59837 93.932 -79.125 52.600
59838 93.936 -79.132 52.588
59839 93.940 -79.140 52.574
59840 93.945 -79.143 52.562
59841 93.953 -79.142 52.550
59842 93.959 -79.140 52.536
59843 93.966 -79.140 52.522
59844 93.981 -79.134 52.506
59845 93.998 -79.126 52.486
59846 94.001 -79.127 52.464
59847 94.001 -79.129 52.440
59848 94.000 -79.131 52.413
59849 94.000 -79.134 52.385
59850 93.999 -79.137 52.354
59851 93.999 -79.139 52.324
59852 93.998 -79.141 52.293
59853 93.998 -79.143 52.260
59854 93.998 -79.144 52.228
59855 93.998 -79.146 52.196
59856 94.002 -79.146 52.166
59857 94.010 -79.144 52.138
59858 94.019 -79.142 52.111
59859 94.028 -79.140 52.084
59860 94.036 -79.138 52.057
59861 94.045 -79.136 52.033
59862 94.053 -79.134 52.011
59863 94.063 -79.133 51.992
59864 94.074 -79.131 51.975
59865 94.075 -79.128 51.958
59866 94.082 -79.114 51.941
59867 94.101 -79.087 51.921
59868 94.120 -79.060 51.899
59869 94.138 -79.034 51.875
59870 94.156 -79.007 51.850
59871 94.176 -78.982 51.829
59872 94.197 -78.960 51.809
59873 94.218 -78.938 51.789
59874 94.240 -78.916 51.765
59875 94.261 -78.894 51.736
59876 94.274 -78.895 51.704
59877 94.285 -78.891 51.668
59878 94.304 -78.900 51.630
59879 94.324 -78.908 51.591
59880 94.343 -78.917 51.552
59881 94.362 -78.926 51.512
59882 94.382 -78.934 51.472
59883 94.401 -78.943 51.433
59884 94.420 -78.951 51.395
59885 94.439 -78.960 51.361
59886 94.458 -78.954 51.329
59887 94.478 -78.942 51.298
59888 94.501 -78.952 51.265
59889 94.524 -78.946 51.230
59890 94.546 -78.937 51.193
59891 94.562 -78.942 51.154
59892 94.578 -78.946 51.114
59893 94.595 -78.952 51.074
59894 94.605 -78.945 51.033
59895 94.616 -78.939 50.995
59896 94.627 -78.932 50.955
59897 94.637 -78.926 50.914
59898 94.648 -78.919 50.874
59899 94.659 -78.913 50.834
59900 94.670 -78.906 50.795
59901 94.681 -78.902 50.757
59902 94.692 -78.899 50.718
59903 94.703 -78.895 50.680
59904 94.715 -78.891 50.641
59905 94.726 -78.887 50.603
59906 94.737 -78.883 50.565
59907 94.749 -78.880 50.526
59908 94.760 -78.877 50.486
59909 94.769 -78.882 50.448
59910 94.777 -78.890 50.409
59911 94.785 -78.897 50.371
59912 94.793 -78.904 50.332
59913 94.802 -78.912 50.294
59914 94.810 -78.919 50.257
59915 94.818 -78.927 50.218
59916 94.827 -78.934 50.180
59917 94.835 -78.942 50.143
59918 94.843 -78.949 50.106
59919 94.851 -78.956 50.068
59920 94.860 -78.964 50.031
59921 94.868 -78.971 49.993
59922 94.877 -78.979 49.956
59923 94.885 -78.987 49.919
59924 94.893 -78.994 49.881
59925 94.901 -79.002 49.844
59926 94.909 -79.009 49.806
59927 94.918 -79.016 49.768
59928 94.927 -79.026 49.729
59929 94.939 -79.044 49.691
59930 94.952 -79.062 49.653
59931 94.964 -79.081 49.615
59932 94.976 -79.100 49.577
59933 94.985 -79.115 49.539
59934 94.991 -79.127 49.501
59935 94.998 -79.140 49.463
59936 95.004 -79.153 49.427
59937 95.011 -79.165 49.391
59938 95.017 -79.178 49.355
59939 95.024 -79.190 49.320
59940 95.030 -79.203 49.287
59941 95.037 -79.216 49.254
59942 95.043 -79.228 49.220
59943 95.049 -79.241 49.188
59944 95.060 -79.254 49.155
59945 95.074 -79.266 49.121
59946 95.089 -79.279 49.086
59947 95.103 -79.291 49.050
59948 95.119 -79.303 49.015
59949 95.139 -79.308 48.979
59950 95.160 -79.314 48.942
59951 95.180 -79.320 48.905
59952 95.192 -79.322 48.868
59953 95.204 -79.324 48.832
59954 95.225 -79.328 48.796
59955 95.249 -79.332 48.760
59956 95.274 -79.337 48.721
59957 95.299 -79.341 48.682
59958 95.324 -79.347 48.644
59959 95.348 -79.351 48.605
59960 95.370 -79.350 48.565
59961 95.388 -79.344 48.524
59962 95.405 -79.338 48.484
59963 95.420 -79.347 48.442
59964 95.433 -79.362 48.401
59965 95.453 -79.375 48.360
59966 95.492 -79.379 48.320
59967 95.531 -79.384 48.279
59968 95.560 -79.378 48.239
59969 95.583 -79.365 48.199
59970 95.601 -79.354 48.158
59971 95.611 -79.349 48.117
59972 95.621 -79.344 48.077
59973 95.631 -79.338 48.038
59974 95.641 -79.332 48.000
59975 95.652 -79.327 47.964
59976 95.662 -79.322 47.930
59977 95.672 -79.316 47.895
59978 95.682 -79.310 47.860
59979 95.692 -79.305 47.825
59980 95.703 -79.300 47.788
59981 95.730 -79.305 47.752
59982 95.750 -79.325 47.715
59983 95.759 -79.346 47.679
59984 95.769 -79.356 47.644
59985 95.779 -79.366 47.609
59986 95.790 -79.376 47.573
59987 95.819 -79.380 47.536
59988 95.872 -79.375 47.500
59989 95.905 -79.357 47.463
59990 95.901 -79.313 47.427
59991 95.903 -79.277 47.391
59992 95.951 -79.298 47.355
59993 95.999 -79.319 47.320
59994 96.046 -79.341 47.285
59995 96.079 -79.355 47.251
59996 96.099 -79.363 47.218
59997 96.120 -79.372 47.185
59998 96.140 -79.380 47.151
59999 96.160 -79.388 47.117
60000 96.181 -79.396 47.084
60001 96.201 -79.404 47.051
60002 96.222 -79.413 47.020
60003 96.242 -79.420 46.988
60004 96.262 -79.429 46.956
60005 96.283 -79.437 46.924
60006 96.303 -79.445 46.891
60007 96.323 -79.453 46.859
60008 96.344 -79.461 46.829
60009 96.364 -79.469 46.799
60010 96.384 -79.478 46.769
60011 96.405 -79.485 46.738
60012 96.425 -79.494 46.708
60013 96.446 -79.502 46.677
60014 96.466 -79.510 46.647
60015 96.486 -79.518 46.620
60016 96.507 -79.526 46.595
60017 96.527 -79.535 46.569
60018 96.547 -79.543 46.543
60019 96.568 -79.551 46.518
60020 96.588 -79.559 46.493
60021 96.608 -79.567 46.469
60022 96.629 -79.576 46.447
60023 96.649 -79.585 46.426
60024 96.673 -79.589 46.406
60025 96.698 -79.592 46.387
60026 96.730 -79.588 46.367
60027 96.763 -79.582 46.346
60028 96.796 -79.573 46.324
60029 96.829 -79.563 46.303
60030 96.859 -79.555 46.282
60031 96.883 -79.544 46.258
60032 96.907 -79.532 46.234
60033 96.932 -79.521 46.209
60034 96.957 -79.507 46.183
60035 96.986 -79.490 46.153
60036 97.014 -79.483 46.126
60037 97.042 -79.477 46.102
60038 97.069 -79.470 46.077
60039 97.097 -79.464 46.052
60040 97.124 -79.458 46.028
60041 97.152 -79.447 46.005
60042 97.164 -79.450 45.982
60043 97.176 -79.452 45.958
60044 97.188 -79.455 45.933
60045 97.201 -79.458 45.909
60046 97.213 -79.461 45.885
60047 97.231 -79.469 45.859
60048 97.241 -79.479 45.833
60049 97.240 -79.484 45.808
60050 97.239 -79.488 45.784
60051 97.237 -79.494 45.761
60052 97.232 -79.504 45.740
60053 97.206 -79.542 45.717
60054 97.189 -79.573 45.695
60055 97.196 -79.580 45.675
60056 97.203 -79.588 45.654
60057 97.210 -79.595 45.634
60058 97.217 -79.603 45.617
60059 97.224 -79.610 45.598
60060 97.231 -79.618 45.580
60061 97.238 -79.626 45.563
60062 97.245 -79.633 45.545
60063 97.253 -79.641 45.527
60064 97.260 -79.648 45.509
60065 97.267 -79.656 45.491
60066 97.274 -79.664 45.474
60067 97.281 -79.672 45.456
60068 97.288 -79.679 45.437
60069 97.295 -79.687 45.419
60070 97.302 -79.695 45.401
60071 97.309 -79.702 45.382
60072 97.316 -79.710 45.366
60073 97.323 -79.717 45.349
60074 97.330 -79.725 45.332
60075 97.337 -79.733 45.315
60076 97.345 -79.740 45.297
60077 97.352 -79.748 45.280
60469 93.241 -79.249 53.436
60470 93.256 -79.237 53.421
60471 93.270 -79.225 53.407
60472 93.285 -79.212 53.393
60473 93.305 -79.203 53.378
60474 93.315 -79.200 53.363
60475 93.325 -79.198 53.348
60476 93.335 -79.195 53.333
60477 93.344 -79.193 53.319
60478 93.354 -79.190 53.305
60479 93.364 -79.188 53.290
60480 93.373 -79.185 53.277
60481 93.383 -79.180 53.262
60482 93.395 -79.169 53.248
60483 93.407 -79.159 53.234
60484 93.418 -79.149 53.220
60485 93.426 -79.142 53.207
60486 93.434 -79.135 53.195
60487 93.442 -79.131 53.183
60488 93.447 -79.136 53.172
60489 93.460 -79.132 53.163
60490 93.480 -79.119 53.157
60491 93.499 -79.106 53.155
60492 93.518 -79.093 53.154
60493 93.537 -79.082 53.155
60494 93.551 -79.077 53.153
60495 93.565 -79.072 53.150
60496 93.578 -79.069 53.145
60497 93.590 -79.068 53.139
60498 93.599 -79.071 53.134
60499 93.607 -79.077 53.128
60500 93.615 -79.081 53.122
60501 93.622 -79.087 53.116
60502 93.630 -79.091 53.109
60503 93.648 -79.091 53.101
60504 93.666 -79.091 53.092
60505 93.683 -79.090 53.082
60506 93.698 -79.090 53.072
60507 93.713 -79.090 53.063
60508 93.727 -79.090 53.053
60509 93.743 -79.089 53.041
60510 93.757 -79.089 53.029
60511 93.772 -79.089 53.015
60512 93.787 -79.087 53.001
60513 93.801 -79.083 52.985
60514 93.817 -79.079 52.970
60515 93.823 -79.078 52.956
60516 93.827 -79.076 52.941
60517 93.830 -79.075 52.927
60518 93.834 -79.074 52.911
60519 93.838 -79.072 52.894
60520 93.844 -79.070 52.874
60521 93.852 -79.068 52.855
60522 93.860 -79.065 52.837
60523 93.868 -79.062 52.820
60524 93.875 -79.063 52.805
60525 93.881 -79.067 52.792
60526 93.888 -79.071 52.779
60527 93.894 -79.075 52.765
60528 93.896 -79.077 52.752
60529 93.898 -79.080 52.740
60530 93.901 -79.082 52.730
60531 93.904 -79.084 52.718
60532 93.907 -79.087 52.705
60533 93.911 -79.089 52.691
60534 93.914 -79.091 52.679
60535 93.917 -79.093 52.666
60536 93.920 -79.096 52.653
60537 93.923 -79.101 52.640
60538 93.927 -79.109 52.627
60539 93.931 -79.117 52.615
60540 93.935 -79.125 52.603
60541 93.939 -79.134 52.591
60542 93.945 -79.134 52.579
60543 93.953 -79.133 52.565
60544 93.969 -79.126 52.549
60545 93.986 -79.119 52.531
60546 94.002 -79.112 52.511
60547 94.002 -79.115 52.488
60548 94.001 -79.117 52.462
60549 94.001 -79.119 52.434
60550 94.000 -79.122 52.405
60551 94.000 -79.125 52.375
60552 93.999 -79.127 52.343
60553 93.999 -79.129 52.309
60554 93.998 -79.132 52.275
60555 93.998 -79.134 52.241
60556 93.997 -79.137 52.209
60557 94.002 -79.136 52.179
60558 94.010 -79.134 52.150
60559 94.019 -79.132 52.123
60560 94.028 -79.131 52.098
60561 94.036 -79.128 52.074
60562 94.045 -79.127 52.051
60563 94.053 -79.125 52.031
60564 94.062 -79.122 52.014
60565 94.071 -79.121 51.998
60566 94.076 -79.118 51.981
60567 94.095 -79.090 51.964
60568 94.114 -79.064 51.942
60569 94.132 -79.037 51.918
60570 94.150 -79.010 51.894
60571 94.169 -78.984 51.869
60572 94.188 -78.956 51.847
60573 94.208 -78.933 51.826
60574 94.229 -78.911 51.806
60575 94.250 -78.889 51.778
60576 94.269 -78.876 51.747
60577 94.282 -78.878 51.712
60578 94.294 -78.882 51.676
60579 94.304 -78.872 51.639
60580 94.324 -78.880 51.601
60581 94.343 -78.890 51.564
60582 94.362 -78.898 51.524
60583 94.382 -78.906 51.485
60584 94.401 -78.915 51.446
60585 94.420 -78.924 51.410
60586 94.439 -78.932 51.377
60587 94.458 -78.912 51.345
60588 94.481 -78.921 51.314
60589 94.504 -78.931 51.282
60590 94.527 -78.935 51.248
60591 94.549 -78.927 51.211
60592 94.565 -78.932 51.172
60593 94.581 -78.937 51.132
60594 94.598 -78.942 51.092
60595 94.612 -78.943 51.053
60596 94.623 -78.937 51.014
60597 94.633 -78.930 50.975
60598 94.644 -78.924 50.934
60599 94.655 -78.918 50.893
60600 94.666 -78.911 50.853
60601 94.677 -78.904 50.814
60602 94.687 -78.898 50.775
60603 94.698 -78.891 50.736
60604 94.709 -78.886 50.697
60605 94.720 -78.882 50.658
60606 94.732 -78.878 50.620
60607 94.743 -78.874 50.581
60608 94.754 -78.871 50.542
60609 94.765 -78.868 50.502
60610 94.776 -78.864 50.464
60611 94.785 -78.869 50.425
60612 94.793 -78.877 50.387
60613 94.802 -78.884 50.348
60614 94.810 -78.891 50.310
60615 94.819 -78.899 50.271
60616 94.827 -78.906 50.233
60617 94.835 -78.914 50.196
60618 94.843 -78.921 50.159
60619 94.851 -78.929 50.122
60620 94.860 -78.937 50.084
60621 94.868 -78.944 50.047
60622 94.877 -78.951 50.009
60623 94.885 -78.959 49.971
60624 94.893 -78.966 49.933
60625 94.901 -78.974 49.896
60626 94.910 -78.981 49.858
60627 94.922 -78.999 49.821
60628 94.934 -79.018 49.783
60629 94.947 -79.036 49.743
60630 94.957 -79.053 49.705
60631 94.964 -79.065 49.667
60632 94.970 -79.078 49.629
60633 94.976 -79.091 49.592
60634 94.983 -79.104 49.554
60635 94.989 -79.116 49.516
60636 94.996 -79.129 49.479
60637 95.002 -79.142 49.442
60638 95.008 -79.155 49.406
60639 95.015 -79.167 49.371
60640 95.022 -79.180 49.338
60641 95.028 -79.193 49.305
60642 95.034 -79.205 49.274
60643 95.041 -79.218 49.244
60644 95.053 -79.231 49.211
60645 95.067 -79.243 49.177
60646 95.082 -79.256 49.142
60647 95.096 -79.268 49.106
60648 95.110 -79.281 49.070
60649 95.127 -79.290 49.035
60650 95.148 -79.296 48.997
60651 95.168 -79.302 48.960
60652 95.188 -79.308 48.923
60653 95.206 -79.313 48.886
60654 95.217 -79.314 48.850
60655 95.237 -79.318 48.814
60656 95.262 -79.322 48.777
60657 95.286 -79.327 48.738
60658 95.312 -79.332 48.697
60659 95.336 -79.337 48.656
60660 95.355 -79.332 48.614
60661 95.373 -79.326 48.573
60662 95.391 -79.320 48.531
60663 95.408 -79.318 48.488
60664 95.421 -79.334 48.446
60665 95.434 -79.350 48.404
60666 95.450 -79.364 48.363
60667 95.489 -79.368 48.321
60668 95.528 -79.372 48.280
60669 95.558 -79.366 48.239
60670 95.580 -79.353 48.199
60671 95.603 -79.340 48.158
60672 95.613 -79.335 48.118
60673 95.624 -79.328 48.079
60674 95.634 -79.323 48.038
60675 95.644 -79.318 48.001
60676 95.654 -79.313 47.964
60677 95.665 -79.306 47.928
60678 95.675 -79.301 47.891
60679 95.685 -79.296 47.855
60680 95.695 -79.291 47.819
60681 95.705 -79.285 47.783
60682 95.737 -79.293 47.748
60683 95.763 -79.308 47.711
60684 95.772 -79.336 47.674
60685 95.782 -79.351 47.638
60686 95.792 -79.362 47.602
60687 95.802 -79.372 47.566
60688 95.834 -79.375 47.529
60689 95.886 -79.370 47.493
60690 95.894 -79.334 47.458
60691 95.890 -79.290 47.421
60692 95.886 -79.246 47.386
60693 95.932 -79.265 47.351
60694 95.979 -79.286 47.316
60695 96.027 -79.307 47.281
60696 96.075 -79.329 47.248
60697 96.111 -79.345 47.215
60698 96.132 -79.353 47.181
60699 96.152 -79.361 47.147
60700 96.173 -79.369 47.112
60701 96.193 -79.378 47.079
60702 96.214 -79.386 47.047
60703 96.234 -79.394 47.015
60704 96.254 -79.402 46.982
60705 96.274 -79.410 46.950
60706 96.295 -79.419 46.918
60707 96.315 -79.427 46.886
60708 96.335 -79.435 46.854
60709 96.356 -79.443 46.823
60710 96.376 -79.451 46.793
60711 96.397 -79.460 46.763
60712 96.417 -79.467 46.733
60713 96.438 -79.476 46.702
60714 96.458 -79.484 46.672
60715 96.478 -79.492 46.642
60716 96.499 -79.500 46.616
60717 96.519 -79.508 46.592
60718 96.539 -79.516 46.568
60719 96.559 -79.525 46.542
60720 96.580 -79.532 46.515
60721 96.600 -79.541 46.490
60722 96.621 -79.549 46.466
60723 96.641 -79.557 46.445
60724 96.662 -79.566 46.425
60725 96.681 -79.576 46.407
60726 96.699 -79.576 46.386
60727 96.719 -79.555 46.365
60728 96.752 -79.545 46.342
60729 96.785 -79.535 46.320
60730 96.818 -79.525 46.299
60731 96.851 -79.515 46.278
60732 96.891 -79.522 46.253
60733 96.915 -79.510 46.228
60734 96.940 -79.498 46.202
60735 96.964 -79.487 46.178
60736 96.994 -79.463 46.148
60737 97.022 -79.450 46.121
60738 97.050 -79.444 46.097
60739 97.078 -79.438 46.072
60740 97.105 -79.431 46.048
60741 97.133 -79.422 46.024
60742 97.148 -79.422 46.000
60743 97.161 -79.425 45.976
60744 97.173 -79.427 45.953
60745 97.185 -79.430 45.928
60746 97.198 -79.435 45.905
60747 97.215 -79.439 45.882
60748 97.232 -79.446 45.858
60749 97.248 -79.454 45.833
60750 97.250 -79.460 45.808
60751 97.249 -79.464 45.783
60752 97.248 -79.469 45.759
60753 97.247 -79.475 45.737
60754 97.245 -79.479 45.715
60755 97.239 -79.492 45.693
60756 97.212 -79.531 45.672
60757 97.203 -79.553 45.652
60758 97.210 -79.560 45.632
60759 97.217 -79.568 45.615
60760 97.225 -79.576 45.597
60761 97.232 -79.583 45.579
60762 97.239 -79.592 45.561
60763 97.246 -79.599 45.544
60764 97.253 -79.607 45.526
60765 97.260 -79.614 45.508
60766 97.267 -79.622 45.491
60767 97.274 -79.629 45.473
60768 97.281 -79.637 45.455
60769 97.288 -79.645 45.437
60770 97.295 -79.652 45.419
60771 97.302 -79.660 45.401
60772 97.309 -79.667 45.383
60773 97.316 -79.675 45.367
60774 97.324 -79.683 45.349
60775 97.331 -79.690 45.332
60776 97.338 -79.698 45.316
60777 97.345 -79.706 45.298
60778 97.352 -79.714 45.281
60779 97.371 -79.729 45.265
60835 98.445 -80.585 44.578
60836 98.464 -80.600 44.572
60837 98.487 -80.607 44.565
60838 98.512 -80.609 44.558
61172 93.269 -79.214 53.433
61173 93.285 -79.202 53.418
61174 93.306 -79.192 53.402
61175 93.315 -79.190 53.388
61176 93.325 -79.187 53.373
61177 93.335 -79.184 53.358
61178 93.345 -79.182 53.344
61179 93.354 -79.180 53.330
61180 93.364 -79.177 53.315
61181 93.373 -79.175 53.300
61182 93.383 -79.172 53.286
61183 93.394 -79.165 53.272
61184 93.405 -79.155 53.258
61185 93.417 -79.145 53.244
61186 93.427 -79.135 53.231
61187 93.436 -79.128 53.219
61188 93.444 -79.122 53.207
61189 93.452 -79.115 53.195
61190 93.458 -79.119 53.185
61191 93.464 -79.125 53.178
61192 93.481 -79.113 53.174
61193 93.500 -79.100 53.173
61194 93.519 -79.088 53.175
61195 93.539 -79.075 53.176
61196 93.558 -79.062 53.175
61197 93.574 -79.056 53.172
61198 93.582 -79.059 53.167
61199 93.590 -79.064 53.162
61200 93.598 -79.068 53.157
61201 93.606 -79.074 53.152
61202 93.613 -79.078 53.145
61203 93.623 -79.082 53.139
61204 93.640 -79.082 53.130
61205 93.658 -79.082 53.121
61206 93.676 -79.082 53.112
61207 93.691 -79.082 53.102
61208 93.706 -79.081 53.091
61209 93.721 -79.081 53.080
61210 93.736 -79.081 53.068
61211 93.751 -79.081 53.054
61212 93.765 -79.078 53.039
61213 93.780 -79.075 53.024
61214 93.795 -79.071 53.008
61215 93.809 -79.068 52.993
61216 93.815 -79.065 52.978
61217 93.819 -79.065 52.964
61218 93.823 -79.063 52.949
61219 93.827 -79.062 52.934
61220 93.832 -79.060 52.917
61221 93.840 -79.057 52.898
61222 93.848 -79.054 52.879
61223 93.856 -79.052 52.862
61224 93.864 -79.049 52.847
61225 93.871 -79.047 52.834
61226 93.878 -79.051 52.821
61227 93.884 -79.054 52.808
61228 93.891 -79.058 52.792
61229 93.893 -79.061 52.780
61230 93.894 -79.063 52.769
61231 93.896 -79.065 52.758
61232 93.900 -79.068 52.747
61233 93.903 -79.071 52.735
61234 93.906 -79.073 52.720
61235 93.909 -79.075 52.707
61236 93.912 -79.078 52.694
61237 93.915 -79.080 52.680
61238 93.919 -79.085 52.666
61239 93.923 -79.093 52.653
61240 93.927 -79.102 52.641
61241 93.931 -79.110 52.629
61242 93.935 -79.118 52.619
61243 93.939 -79.125 52.606
61244 93.957 -79.119 52.591
61245 93.974 -79.112 52.574
61246 93.991 -79.105 52.556
61247 94.003 -79.100 52.535
61248 94.003 -79.103 52.510
61249 94.002 -79.105 52.483
61250 94.002 -79.108 52.454
61251 94.001 -79.110 52.424
61252 94.001 -79.112 52.392
61253 94.000 -79.115 52.359
61254 94.000 -79.117 52.324
61255 93.999 -79.120 52.288
61256 94.000 -79.117 52.254
61257 94.001 -79.114 52.222
61258 94.006 -79.112 52.191
61259 94.014 -79.110 52.164
61260 94.023 -79.109 52.138
61261 94.032 -79.107 52.114
61262 94.040 -79.106 52.091
61263 94.048 -79.104 52.070
61264 94.057 -79.103 52.051
61265 94.065 -79.102 52.035
61266 94.074 -79.100 52.018
61267 94.086 -79.095 52.001
61268 94.107 -79.068 51.982
61269 94.126 -79.040 51.958
61270 94.145 -79.013 51.933
61271 94.163 -78.987 51.908
61272 94.182 -78.959 51.883
61273 94.200 -78.933 51.862
61274 94.219 -78.906 51.841
61275 94.240 -78.884 51.820
61276 94.261 -78.862 51.790
61277 94.277 -78.858 51.756
61278 94.290 -78.861 51.721
61279 94.303 -78.864 51.686
61280 94.314 -78.867 51.649
61281 94.324 -78.853 51.612
61282 94.343 -78.861 51.575
61283 94.362 -78.870 51.536
61284 94.382 -78.879 51.497
61285 94.401 -78.887 51.460
61286 94.420 -78.896 51.425
61287 94.438 -78.880 51.392
61288 94.461 -78.890 51.360
61289 94.484 -78.901 51.330
61290 94.507 -78.911 51.298
61291 94.531 -78.921 51.264
61292 94.552 -78.918 51.230
61293 94.568 -78.922 51.190
61294 94.584 -78.927 51.151
61295 94.601 -78.932 51.111
61296 94.617 -78.937 51.073
61297 94.629 -78.935 51.035
61298 94.640 -78.928 50.996
61299 94.651 -78.922 50.957
61300 94.662 -78.915 50.915
61301 94.672 -78.909 50.874
61302 94.683 -78.902 50.835
61303 94.694 -78.896 50.795
61304 94.705 -78.890 50.756
61305 94.715 -78.883 50.717
61306 94.726 -78.877 50.678
61307 94.737 -78.870 50.639
61308 94.748 -78.865 50.600
61309 94.759 -78.861 50.560
61310 94.771 -78.858 50.521
61311 94.782 -78.855 50.481
61312 94.793 -78.851 50.442
61313 94.802 -78.856 50.404
61314 94.810 -78.864 50.365
61315 94.819 -78.871 50.326
61316 94.827 -78.879 50.287
61317 94.835 -78.886 50.249
61318 94.843 -78.893 50.212
61319 94.852 -78.901 50.175
61320 94.860 -78.908 50.138
61321 94.868 -78.916 50.100
61322 94.877 -78.924 50.063
61323 94.885 -78.931 50.025
61324 94.893 -78.938 49.987
61325 94.905 -78.955 49.949
61326 94.917 -78.973 49.912
61327 94.929 -78.992 49.874
61328 94.936 -79.004 49.836
61329 94.942 -79.017 49.798
61330 94.949 -79.030 49.759
61331 94.955 -79.043 49.720
61332 94.962 -79.055 49.682
61333 94.968 -79.068 49.644
61334 94.974 -79.081 49.607
61335 94.981 -79.093 49.569
61336 94.987 -79.106 49.532
61337 94.993 -79.118 49.495
61338 95.000 -79.131 49.459
61339 95.006 -79.143 49.423
61340 95.013 -79.156 49.390
61341 95.016 -79.165 49.359
61342 95.019 -79.174 49.328
61343 95.025 -79.193 49.298
61344 95.044 -79.203 49.267
61345 95.059 -79.218 49.234
61346 95.075 -79.232 49.201
61347 95.089 -79.245 49.166
61348 95.103 -79.257 49.130
61349 95.118 -79.270 49.093
61350 95.136 -79.278 49.056
61351 95.156 -79.284 49.018
61352 95.177 -79.290 48.979
61353 95.197 -79.296 48.941
61354 95.217 -79.302 48.903
61355 95.231 -79.304 48.865
61356 95.250 -79.308 48.827
61357 95.274 -79.313 48.788
61358 95.299 -79.317 48.747
61359 95.323 -79.321 48.705
61360 95.341 -79.315 48.663
61361 95.359 -79.309 48.620
61362 95.377 -79.303 48.577
61363 95.395 -79.297 48.535
61364 95.409 -79.306 48.492
61365 95.422 -79.321 48.448
61366 95.435 -79.337 48.406
61367 95.448 -79.353 48.364
61368 95.486 -79.357 48.323
61369 95.526 -79.362 48.282
61370 95.555 -79.355 48.240
61371 95.577 -79.342 48.200
61372 95.600 -79.328 48.159
61373 95.616 -79.319 48.119
61374 95.626 -79.314 48.079
61375 95.636 -79.309 48.040
61376 95.647 -79.303 48.002
61377 95.657 -79.297 47.965
61378 95.667 -79.292 47.928
61379 95.677 -79.287 47.891
61380 95.687 -79.281 47.854
61381 95.698 -79.275 47.817
61382 95.709 -79.271 47.780
61383 95.743 -79.281 47.744
61384 95.776 -79.291 47.708
61385 95.785 -79.319 47.671
61386 95.795 -79.347 47.634
61387 95.805 -79.357 47.598
61388 95.815 -79.367 47.562
61389 95.849 -79.369 47.525
61390 95.887 -79.355 47.490
61391 95.883 -79.311 47.455
61392 95.879 -79.267 47.419
61393 95.875 -79.223 47.383
61394 95.912 -79.231 47.347
61395 95.960 -79.253 47.312
61396 96.008 -79.274 47.278
61397 96.055 -79.295 47.244
61398 96.103 -79.317 47.210
61399 96.144 -79.335 47.176
61400 96.164 -79.343 47.142
61401 96.185 -79.351 47.108
61402 96.205 -79.360 47.074
61403 96.226 -79.368 47.041
61404 96.246 -79.375 47.009
61405 96.266 -79.384 46.977
61406 96.287 -79.392 46.945
61407 96.307 -79.400 46.913
61408 96.327 -79.408 46.881
61409 96.347 -79.416 46.849
61410 96.368 -79.425 46.818
61411 96.388 -79.433 46.787
61412 96.409 -79.441 46.757
61413 96.429 -79.449 46.727
61414 96.450 -79.457 46.697
61415 96.470 -79.466 46.667
61416 96.490 -79.473 46.637
61417 96.511 -79.482 46.610
61418 96.531 -79.490 46.585
61419 96.551 -79.498 46.560
61420 96.572 -79.506 46.535
61421 96.592 -79.514 46.510
61422 96.613 -79.523 46.484
61423 96.633 -79.534 46.460
61424 96.654 -79.544 46.439
61425 96.670 -79.549 46.419
61426 96.684 -79.548 46.399
61427 96.707 -79.551 46.377
61428 96.726 -79.531 46.354
61429 96.744 -79.510 46.331
61430 96.773 -79.497 46.309
61431 96.806 -79.487 46.289
61432 96.839 -79.476 46.269
61433 96.875 -79.472 46.245
61434 96.920 -79.485 46.220
61435 96.947 -79.476 46.194
61436 96.971 -79.464 46.169
61437 97.000 -79.444 46.141
61438 97.030 -79.418 46.116
61439 97.058 -79.410 46.093
61440 97.086 -79.404 46.069
61441 97.113 -79.397 46.045
61442 97.133 -79.394 46.021
61443 97.145 -79.397 45.997
61444 97.157 -79.399 45.973
61445 97.170 -79.404 45.950
61446 97.184 -79.408 45.927
61447 97.198 -79.413 45.904
61448 97.220 -79.416 45.880
61449 97.235 -79.424 45.855
61450 97.255 -79.432 45.830
61451 97.261 -79.441 45.805
61452 97.257 -79.446 45.781
61453 97.258 -79.445 45.756
61454 97.257 -79.451 45.734
61455 97.256 -79.455 45.713
61456 97.254 -79.460 45.692
61457 97.253 -79.466 45.671
61458 97.245 -79.480 45.650
61459 97.218 -79.519 45.631
61460 97.218 -79.534 45.613
61461 97.225 -79.541 45.596
61462 97.232 -79.549 45.578
61463 97.239 -79.557 45.560
61464 97.246 -79.564 45.543
61465 97.253 -79.572 45.525
61466 97.260 -79.579 45.507
61467 97.267 -79.587 45.490
61468 97.274 -79.595 45.473
61469 97.281 -79.602 45.455
61470 97.288 -79.610 45.438
61471 97.296 -79.618 45.419
61472 97.303 -79.626 45.401
61473 97.310 -79.633 45.384
61474 97.317 -79.641 45.368
61475 97.324 -79.648 45.350
61476 97.331 -79.656 45.333
61477 97.338 -79.664 45.317
61478 97.345 -79.671 45.299
61479 97.358 -79.683 45.282
61480 97.377 -79.698 45.266
61481 97.396 -79.713 45.251
61534 98.412 -80.522 44.601
61535 98.431 -80.538 44.592
61536 98.450 -80.554 44.586
61537 98.469 -80.569 44.579
61538 98.492 -80.576 44.572
61539 98.517 -80.578 44.566
61540 98.542 -80.581 44.559
61541 98.567 -80.583 44.551
61542 98.591 -80.585 44.543
61543 98.616 -80.588 44.535
61544 98.641 -80.591 44.528
61545 98.665 -80.593 44.522
61546 98.690 -80.596 44.515
61547 98.715 -80.599 44.507
61548 98.741 -80.604 44.500
61549 98.767 -80.607 44.495
61550 98.792 -80.611 44.488
61551 98.817 -80.615 44.481
61552 98.843 -80.620 44.476
61553 98.867 -80.627 44.471
61874 93.284 -79.192 53.443
61875 93.306 -79.181 53.427
61876 93.315 -79.178 53.413
61877 93.326 -79.176 53.398
61878 93.335 -79.174 53.384
61879 93.345 -79.171 53.369
61880 93.354 -79.169 53.355
61881 93.364 -79.166 53.340
61882 93.374 -79.164 53.326
61883 93.383 -79.161 53.311
61884 93.393 -79.159 53.296
61885 93.404 -79.151 53.282
61886 93.415 -79.141 53.268
61887 93.427 -79.131 53.255
61888 93.437 -79.122 53.242
61889 93.445 -79.115 53.229
61890 93.453 -79.108 53.217
61891 93.462 -79.101 53.207
61892 93.468 -79.102 53.199
61893 93.474 -79.107 53.195
61894 93.483 -79.108 53.192
61895 93.502 -79.095 53.195
61896 93.521 -79.082 53.197
61897 93.541 -79.069 53.198
61898 93.560 -79.056 53.196
61899 93.573 -79.053 53.193
61900 93.581 -79.056 53.190
61901 93.588 -79.061 53.185
61902 93.596 -79.065 53.180
61903 93.604 -79.071 53.174
61904 93.615 -79.074 53.168
61905 93.633 -79.074 53.159
61906 93.650 -79.074 53.150
61907 93.668 -79.074 53.140
61908 93.685 -79.073 53.130
61909 93.699 -79.073 53.118
61910 93.714 -79.073 53.105
61911 93.729 -79.072 53.093
61912 93.744 -79.069 53.078
61913 93.759 -79.066 53.063
61914 93.773 -79.062 53.047
61915 93.788 -79.059 53.031
61916 93.802 -79.056 53.016
61917 93.808 -79.054 53.001
61918 93.812 -79.053 52.986
61919 93.816 -79.051 52.972
61920 93.820 -79.049 52.957
61921 93.828 -79.047 52.940
61922 93.836 -79.044 52.922
61923 93.844 -79.041 52.905
61924 93.852 -79.039 52.889
61925 93.860 -79.036 52.874
61926 93.868 -79.033 52.862
61927 93.874 -79.034 52.850
61928 93.881 -79.038 52.837
61929 93.887 -79.042 52.823
61930 93.890 -79.045 52.809
61931 93.891 -79.047 52.797
61932 93.892 -79.049 52.785
61933 93.895 -79.052 52.775
61934 93.898 -79.054 52.763
61935 93.901 -79.057 52.749
61936 93.904 -79.059 52.734
61937 93.908 -79.062 52.719
61938 93.911 -79.064 52.704
61939 93.914 -79.070 52.689
61940 93.918 -79.078 52.677
61941 93.922 -79.087 52.666
61942 93.926 -79.094 52.654
61943 93.930 -79.103 52.643
61944 93.936 -79.103 52.630
61945 93.962 -79.105 52.614
61946 93.979 -79.098 52.598
61947 93.995 -79.090 52.579
61948 94.004 -79.088 52.557
61949 94.003 -79.090 52.530
61950 94.003 -79.093 52.502
61951 94.002 -79.096 52.472
61952 94.002 -79.098 52.441
61953 94.001 -79.100 52.408
61954 94.002 -79.099 52.374
61955 94.003 -79.095 52.339
61956 94.004 -79.092 52.302
61957 94.005 -79.088 52.268
61958 94.006 -79.085 52.235
61959 94.012 -79.083 52.204
61960 94.020 -79.081 52.178
61961 94.029 -79.080 52.154
61962 94.037 -79.078 52.131
61963 94.046 -79.077 52.110
61964 94.054 -79.075 52.090
61965 94.062 -79.074 52.071
61966 94.071 -79.072 52.054
61967 94.080 -79.071 52.036
61968 94.095 -79.063 52.017
61969 94.112 -79.053 51.997
61970 94.132 -79.025 51.972
61971 94.153 -78.995 51.947
61972 94.175 -78.965 51.923
61973 94.194 -78.936 51.899
61974 94.213 -78.909 51.878
61975 94.231 -78.882 51.857
61976 94.251 -78.858 51.833
61977 94.271 -78.839 51.802
61978 94.285 -78.841 51.768
61979 94.298 -78.843 51.733
61980 94.312 -78.846 51.697
61981 94.323 -78.851 51.660
61982 94.330 -78.840 51.623
61983 94.343 -78.834 51.587
61984 94.362 -78.843 51.548
61985 94.382 -78.851 51.510
61986 94.401 -78.859 51.474
61987 94.420 -78.861 51.439
61988 94.441 -78.860 51.407
61989 94.465 -78.870 51.376
61990 94.487 -78.880 51.346
61991 94.510 -78.890 51.314
61992 94.534 -78.900 51.282
61993 94.555 -78.908 51.248
61994 94.571 -78.912 51.209
61995 94.588 -78.918 51.170
61996 94.604 -78.923 51.132
61997 94.620 -78.927 51.094
61998 94.636 -78.933 51.057
61999 94.647 -78.927 51.019
62000 94.657 -78.920 50.982
62001 94.668 -78.914 50.939
62002 94.679 -78.907 50.899
62003 94.690 -78.901 50.859
62004 94.700 -78.894 50.821
62005 94.711 -78.887 50.781
62006 94.722 -78.881 50.741
62007 94.733 -78.874 50.701
62008 94.744 -78.868 50.661
62009 94.754 -78.861 50.622
62010 94.765 -78.855 50.583
62011 94.776 -78.849 50.542
62012 94.787 -78.846 50.501
62013 94.799 -78.842 50.461
62014 94.810 -78.838 50.422
62015 94.819 -78.843 50.382
62016 94.827 -78.851 50.343
62017 94.835 -78.858 50.305
62018 94.843 -78.866 50.266
62019 94.852 -78.874 50.228
62020 94.860 -78.880 50.191
62021 94.868 -78.888 50.154
62022 94.877 -78.896 50.116
62023 94.888 -78.910 50.079
62024 94.900 -78.929 50.041
62025 94.908 -78.943 50.003
62026 94.914 -78.955 49.965
62027 94.921 -78.968 49.928
62028 94.927 -78.981 49.890
62029 94.933 -78.993 49.853
62030 94.940 -79.006 49.814
62031 94.947 -79.019 49.775
62032 94.953 -79.031 49.736
62033 94.959 -79.044 49.698
62034 94.963 -79.053 49.660
62035 94.965 -79.062 49.622
62036 94.968 -79.070 49.585
62037 94.970 -79.078 49.548
62038 94.973 -79.087 49.511
62039 94.975 -79.095 49.476
62040 94.978 -79.103 49.442
62041 94.979 -79.117 49.410
62042 94.975 -79.158 49.380
62043 95.007 -79.165 49.350
62044 95.029 -79.173 49.320
62045 95.045 -79.182 49.288
62046 95.061 -79.197 49.255
62047 95.076 -79.211 49.221
62048 95.092 -79.225 49.187
62049 95.108 -79.241 49.152
62050 95.123 -79.255 49.114
62051 95.144 -79.265 49.075
62052 95.165 -79.272 49.036
62053 95.185 -79.278 48.996
62054 95.205 -79.284 48.957
62055 95.226 -79.290 48.917
62056 95.244 -79.295 48.877
62057 95.262 -79.298 48.837
62058 95.287 -79.303 48.796
62059 95.309 -79.303 48.754
62060 95.327 -79.297 48.712
62061 95.344 -79.291 48.669
62062 95.362 -79.285 48.625
62063 95.380 -79.279 48.582
62064 95.397 -79.277 48.539
62065 95.410 -79.293 48.496
62066 95.423 -79.309 48.452
62067 95.436 -79.324 48.409
62068 95.449 -79.340 48.368
62069 95.483 -79.346 48.327
62070 95.523 -79.350 48.285
62071 95.552 -79.344 48.243
62072 95.574 -79.331 48.202
62073 95.597 -79.318 48.162
62074 95.619 -79.305 48.122
62075 95.629 -79.299 48.082
62076 95.639 -79.294 48.043
62077 95.650 -79.288 48.005
62078 95.660 -79.283 47.967
62079 95.670 -79.277 47.929
62080 95.680 -79.272 47.892
62081 95.690 -79.266 47.854
62082 95.700 -79.261 47.817
62083 95.716 -79.259 47.779
62084 95.750 -79.268 47.742
62085 95.783 -79.278 47.706
62086 95.799 -79.301 47.669
62087 95.808 -79.329 47.632
62088 95.818 -79.353 47.596
62089 95.828 -79.363 47.559
62090 95.864 -79.364 47.522
62091 95.876 -79.332 47.487
62092 95.872 -79.288 47.451
62093 95.868 -79.244 47.416
62094 95.864 -79.200 47.380
62095 95.893 -79.197 47.344
62096 95.941 -79.219 47.309
62097 95.988 -79.241 47.274
62098 96.036 -79.262 47.240
62099 96.084 -79.283 47.205
62100 96.131 -79.305 47.171
62101 96.177 -79.325 47.137
62102 96.197 -79.333 47.103
62103 96.217 -79.341 47.069
62104 96.238 -79.349 47.036
62105 96.258 -79.357 47.003
62106 96.278 -79.366 46.971
62107 96.299 -79.374 46.939
62108 96.319 -79.382 46.907
62109 96.339 -79.390 46.875
62110 96.360 -79.398 46.843
62111 96.380 -79.407 46.812
62112 96.400 -79.414 46.781
62113 96.421 -79.422 46.751
62114 96.441 -79.431 46.721
62115 96.462 -79.439 46.691
62116 96.482 -79.447 46.661
62117 96.503 -79.455 46.631
62118 96.523 -79.463 46.604
62119 96.543 -79.472 46.578
62120 96.564 -79.481 46.552
62121 96.585 -79.491 46.528
62122 96.605 -79.501 46.503
62123 96.626 -79.512 46.477
62124 96.647 -79.523 46.453
62125 96.658 -79.523 46.431
62126 96.669 -79.521 46.410
62127 96.692 -79.523 46.388
62128 96.715 -79.526 46.365
62129 96.734 -79.507 46.341
62130 96.752 -79.486 46.317
62131 96.771 -79.465 46.295
62132 96.795 -79.448 46.275
62133 96.827 -79.438 46.256
62134 96.861 -79.428 46.234
62135 96.899 -79.429 46.210
62136 96.945 -79.442 46.185
62137 96.979 -79.442 46.159
62138 97.005 -79.427 46.133
62139 97.036 -79.400 46.109
62140 97.066 -79.377 46.087
62141 97.094 -79.371 46.065
62142 97.116 -79.366 46.042
62143 97.129 -79.369 46.018
62144 97.143 -79.373 45.995
62145 97.156 -79.378 45.971
62146 97.170 -79.382 45.948
62147 97.184 -79.386 45.927
62148 97.203 -79.391 45.904
62149 97.223 -79.395 45.877
62150 97.239 -79.403 45.850
62151 97.268 -79.410 45.825
62152 97.279 -79.419 45.801
62153 97.272 -79.429 45.777
62154 97.264 -79.438 45.755
62155 97.261 -79.441 45.733
62156 97.266 -79.432 45.711
62157 97.265 -79.436 45.690
62158 97.263 -79.441 45.670
62159 97.262 -79.447 45.650
62160 97.261 -79.451 45.631
62161 97.251 -79.469 45.614
62162 97.225 -79.507 45.596
62163 97.232 -79.514 45.578
62164 97.239 -79.522 45.561
62165 97.246 -79.529 45.543
62166 97.253 -79.538 45.525
62167 97.260 -79.545 45.507
62168 97.267 -79.553 45.490
62169 97.275 -79.560 45.473
62170 97.282 -79.568 45.456
62171 97.289 -79.576 45.438
62172 97.296 -79.583 45.420
62173 97.303 -79.591 45.402
62174 97.310 -79.598 45.385
62175 97.317 -79.606 45.369
62176 97.324 -79.614 45.352
62177 97.331 -79.621 45.335
62178 97.338 -79.629 45.319
62179 97.345 -79.637 45.301
62180 97.363 -79.651 45.285
62181 97.382 -79.667 45.269
62182 97.401 -79.682 45.254
62233 98.379 -80.461 44.623
62234 98.398 -80.476 44.616
62235 98.417 -80.491 44.609
62236 98.436 -80.507 44.601
62237 98.455 -80.522 44.594
62238 98.475 -80.538 44.587
62239 98.498 -80.544 44.581
62240 98.522 -80.547 44.574
62241 98.547 -80.549 44.566
62242 98.572 -80.552 44.559
62243 98.597 -80.554 44.552
62244 98.622 -80.557 44.543
62245 98.646 -80.560 44.536
62246 98.671 -80.562 44.530
62247 98.696 -80.565 44.523
62248 98.721 -80.568 44.516
62249 98.746 -80.572 44.508
62250 98.772 -80.576 44.501
62251 98.797 -80.580 44.495
62252 98.823 -80.584 44.489
62253 98.848 -80.589 44.484
62254 98.871 -80.600 44.478
62577 93.316 -79.168 53.438
62578 93.326 -79.165 53.423
62579 93.335 -79.162 53.409
62580 93.345 -79.160 53.395
62581 93.355 -79.158 53.379
62582 93.364 -79.156 53.365
62583 93.374 -79.153 53.351
62584 93.384 -79.150 53.336
62585 93.394 -79.148 53.321
62586 93.403 -79.146 53.306
62587 93.414 -79.137 53.292
62588 93.425 -79.127 53.278
62589 93.437 -79.116 53.265
62590 93.446 -79.109 53.252
62591 93.454 -79.101 53.239
62592 93.463 -79.094 53.229
62593 93.471 -79.087 53.221
62594 93.479 -79.085 53.214
62595 93.484 -79.090 53.210
62596 93.490 -79.095 53.211
62597 93.504 -79.089 53.212
62598 93.523 -79.076 53.213
62599 93.542 -79.063 53.213
62600 93.561 -79.051 53.212
62601 93.572 -79.051 53.210
62602 93.579 -79.056 53.207
62603 93.587 -79.059 53.203
62604 93.595 -79.064 53.198
62605 93.608 -79.065 53.192
62606 93.625 -79.065 53.185
62607 93.643 -79.065 53.176
62608 93.660 -79.065 53.166
62609 93.678 -79.065 53.156
62610 93.693 -79.065 53.143
62611 93.707 -79.064 53.129
62612 93.722 -79.061 53.115
62613 93.737 -79.057 53.101
62614 93.752 -79.054 53.085
62615 93.766 -79.050 53.069
62616 93.781 -79.047 53.054
62617 93.795 -79.043 53.039
62618 93.800 -79.042 53.024
62619 93.804 -79.040 53.009
62620 93.808 -79.039 52.995
62621 93.816 -79.037 52.980
62622 93.824 -79.034 52.964
62623 93.832 -79.031 52.947
62624 93.840 -79.028 52.931
62625 93.848 -79.025 52.916
62626 93.856 -79.023 52.902
62627 93.863 -79.020 52.891
62628 93.871 -79.018 52.880
62629 93.877 -79.022 52.867
62630 93.884 -79.026 52.853
62631 93.887 -79.028 52.839
62632 93.888 -79.031 52.825
62633 93.889 -79.034 52.813
62634 93.890 -79.036 52.802
62635 93.893 -79.038 52.789
62636 93.897 -79.040 52.774
62637 93.900 -79.043 52.759
62638 93.903 -79.045 52.743
62639 93.906 -79.048 52.727
62640 93.910 -79.055 52.712
62641 93.914 -79.063 52.700
62642 93.918 -79.071 52.687
62643 93.922 -79.079 52.674
62644 93.922 -79.080 52.662
62645 93.933 -79.080 52.649
62646 93.959 -79.082 52.634
62647 93.984 -79.084 52.616
62648 94.000 -79.076 52.596
62649 94.005 -79.076 52.573
62650 94.004 -79.078 52.547
62651 94.004 -79.081 52.519
62652 94.004 -79.081 52.489
62653 94.005 -79.077 52.457
62654 94.006 -79.074 52.423
62655 94.007 -79.070 52.389
62656 94.008 -79.066 52.355
62657 94.010 -79.063 52.318
62658 94.011 -79.059 52.282
62659 94.012 -79.056 52.249
62660 94.017 -79.054 52.218
62661 94.026 -79.053 52.193
62662 94.034 -79.051 52.171
62663 94.043 -79.049 52.150
62664 94.051 -79.048 52.130
62665 94.060 -79.046 52.111
62666 94.069 -79.045 52.092
62667 94.078 -79.043 52.072
62668 94.087 -79.041 52.052
62669 94.104 -79.031 52.031
62670 94.120 -79.021 52.009
62671 94.137 -79.012 51.984
62672 94.158 -78.983 51.960
62673 94.179 -78.952 51.937
62674 94.200 -78.921 51.916
62675 94.221 -78.891 51.894
62676 94.243 -78.861 51.872
62677 94.262 -78.832 51.847
62678 94.279 -78.821 51.815
62679 94.292 -78.824 51.780
62680 94.306 -78.826 51.745
62681 94.320 -78.829 51.708
62682 94.332 -78.833 51.671
62683 94.343 -78.838 51.635
62684 94.346 -78.814 51.600
62685 94.362 -78.814 51.562
62686 94.382 -78.824 51.525
62687 94.401 -78.832 51.489
62688 94.421 -78.829 51.454
62689 94.445 -78.839 51.422
62690 94.468 -78.849 51.391
62691 94.491 -78.859 51.361
62692 94.515 -78.876 51.330
62693 94.536 -78.893 51.298
62694 94.550 -78.901 51.265
62695 94.569 -78.906 51.228
62696 94.591 -78.908 51.190
62697 94.607 -78.913 51.151
62698 94.623 -78.918 51.114
62699 94.639 -78.923 51.077
62700 94.653 -78.924 51.041
62701 94.664 -78.918 51.005
62702 94.675 -78.912 50.964
62703 94.686 -78.905 50.924
62704 94.696 -78.899 50.886
62705 94.707 -78.893 50.849
62706 94.718 -78.886 50.811
62707 94.729 -78.879 50.771
62708 94.740 -78.873 50.731
62709 94.750 -78.866 50.690
62710 94.761 -78.860 50.650
62711 94.772 -78.853 50.609
62712 94.782 -78.847 50.568
62713 94.793 -78.840 50.525
62714 94.804 -78.833 50.483
62715 94.815 -78.829 50.443
62716 94.826 -78.825 50.402
62717 94.835 -78.830 50.363
62718 94.844 -78.838 50.325
62719 94.852 -78.846 50.286
62720 94.860 -78.853 50.248
62721 94.871 -78.865 50.210
62722 94.880 -78.881 50.172
62723 94.887 -78.894 50.134
62724 94.893 -78.907 50.096
62725 94.899 -78.919 50.058
62726 94.906 -78.932 50.020
62727 94.909 -78.942 49.982
62728 94.912 -78.950 49.945
62729 94.914 -78.959 49.907
62730 94.917 -78.966 49.869
62731 94.919 -78.974 49.831
62732 94.922 -78.983 49.791
62733 94.924 -78.991 49.753
62734 94.927 -78.999 49.715
62735 94.929 -79.008 49.677
62736 94.932 -79.016 49.639
62737 94.934 -79.024 49.602
62738 94.936 -79.032 49.566
62739 94.939 -79.040 49.530
62740 94.935 -79.076 49.496
62741 94.929 -79.118 49.464
62742 94.957 -79.129 49.432
62743 94.989 -79.136 49.400
62744 95.013 -79.143 49.368
62745 95.030 -79.153 49.335
62746 95.047 -79.162 49.302
62747 95.063 -79.175 49.269
62748 95.078 -79.190 49.235
62749 95.094 -79.204 49.201
62750 95.109 -79.219 49.166
62751 95.126 -79.231 49.129
62752 95.148 -79.241 49.090
62753 95.170 -79.250 49.050
62754 95.191 -79.259 49.009
62755 95.212 -79.269 48.968
62756 95.234 -79.278 48.928
62757 95.254 -79.284 48.887
62758 95.274 -79.288 48.845
62759 95.294 -79.285 48.802
62760 95.312 -79.279 48.759
62761 95.330 -79.273 48.717
62762 95.348 -79.267 48.674
62763 95.366 -79.261 48.632
62764 95.384 -79.255 48.588
62765 95.398 -79.265 48.544
62766 95.411 -79.280 48.501
62767 95.424 -79.296 48.458
62768 95.437 -79.312 48.416
62769 95.450 -79.327 48.374
62770 95.481 -79.335 48.331
62771 95.520 -79.339 48.289
62772 95.549 -79.332 48.247
62773 95.571 -79.319 48.205
62774 95.594 -79.306 48.166
62775 95.616 -79.293 48.127
62776 95.631 -79.285 48.087
62777 95.641 -79.279 48.049
62778 95.652 -79.273 48.010
62779 95.662 -79.268 47.971
62780 95.672 -79.263 47.932
62781 95.683 -79.257 47.895
62782 95.693 -79.251 47.857
62783 95.703 -79.246 47.819
62784 95.723 -79.247 47.780
62785 95.757 -79.256 47.742
62786 95.790 -79.266 47.705
62787 95.812 -79.284 47.668
62788 95.821 -79.312 47.632
62789 95.831 -79.341 47.595
62790 95.840 -79.359 47.557
62791 95.869 -79.352 47.521
62792 95.865 -79.308 47.485
62793 95.861 -79.264 47.449
62794 95.857 -79.220 47.413
62795 95.853 -79.176 47.376
62796 95.874 -79.164 47.341
62797 95.921 -79.185 47.305
62798 95.969 -79.206 47.270
62799 96.017 -79.228 47.236
62800 96.064 -79.250 47.201
62801 96.112 -79.271 47.166
62802 96.160 -79.293 47.132
62803 96.207 -79.314 47.097
62804 96.230 -79.322 47.062
62805 96.250 -79.331 47.030
62806 96.270 -79.339 46.997
62807 96.291 -79.347 46.965
62808 96.311 -79.356 46.932
62809 96.331 -79.363 46.900
62810 96.352 -79.372 46.868
62811 96.372 -79.380 46.835
62812 96.392 -79.388 46.805
62813 96.413 -79.396 46.775
62814 96.433 -79.404 46.745
62815 96.454 -79.413 46.714
62816 96.474 -79.421 46.684
62817 96.494 -79.429 46.655
62818 96.515 -79.438 46.626
62819 96.536 -79.448 46.598
62820 96.556 -79.459 46.571
62821 96.577 -79.469 46.545
62822 96.598 -79.479 46.519
62823 96.619 -79.490 46.494
62824 96.637 -79.498 46.470
62825 96.646 -79.496 46.445
62826 96.655 -79.494 46.422
62827 96.677 -79.496 46.399
62828 96.699 -79.498 46.376
62829 96.722 -79.501 46.352
62830 96.742 -79.484 46.328
62831 96.760 -79.463 46.305
62832 96.779 -79.442 46.281
62833 96.798 -79.421 46.259
62834 96.816 -79.400 46.239
62835 96.849 -79.390 46.217
62836 96.882 -79.379 46.195
62837 96.924 -79.387 46.173
62838 96.969 -79.400 46.148
62839 97.011 -79.408 46.122
62840 97.041 -79.382 46.098
62841 97.072 -79.356 46.077
62842 97.101 -79.338 46.056
62843 97.114 -79.342 46.034
62844 97.128 -79.347 46.011
62845 97.142 -79.351 45.987
62846 97.156 -79.356 45.963
62847 97.169 -79.360 45.941
62848 97.186 -79.364 45.919
62849 97.208 -79.368 45.894
62850 97.226 -79.374 45.870
62851 97.251 -79.382 45.845
62852 97.281 -79.389 45.820
62853 97.296 -79.398 45.797
62854 97.289 -79.407 45.774
62855 97.282 -79.417 45.752
62856 97.275 -79.427 45.731
62857 97.267 -79.437 45.710
62858 97.265 -79.437 45.689
62859 97.271 -79.424 45.669
62860 97.273 -79.422 45.649
62861 97.272 -79.427 45.630
62862 97.271 -79.432 45.613
62863 97.269 -79.438 45.596
62864 97.257 -79.457 45.578
62865 97.239 -79.488 45.560
62866 97.246 -79.495 45.543
62867 97.253 -79.503 45.525
62868 97.260 -79.510 45.507
62869 97.267 -79.518 45.490
62870 97.275 -79.526 45.473
62871 97.282 -79.533 45.456
62872 97.289 -79.541 45.438
62873 97.296 -79.548 45.421
62874 97.303 -79.557 45.403
62875 97.310 -79.564 45.386
62876 97.317 -79.572 45.369
62877 97.324 -79.579 45.353
62878 97.331 -79.587 45.337
62879 97.338 -79.595 45.320
62880 97.349 -79.604 45.304
62881 97.368 -79.620 45.288
62882 97.387 -79.635 45.272
62883 97.406 -79.651 45.256
62884 97.426 -79.666 45.241
62931 98.327 -80.384 44.654
62932 98.346 -80.399 44.646
62933 98.365 -80.415 44.639
62934 98.384 -80.430 44.632
62935 98.404 -80.445 44.625
62936 98.423 -80.460 44.618
62937 98.442 -80.475 44.610
62938 98.461 -80.491 44.603
62939 98.480 -80.506 44.596
62940 98.503 -80.513 44.590
62941 98.528 -80.516 44.583
62942 98.553 -80.518 44.576
62943 98.577 -80.520 44.568
62944 98.602 -80.523 44.560
62945 98.627 -80.526 44.551
62946 98.652 -80.529 44.544
62947 98.676 -80.531 44.538
62948 98.701 -80.533 44.531
62949 98.726 -80.537 44.524
62950 98.751 -80.541 44.516
62951 98.777 -80.544 44.509
62952 98.803 -80.548 44.503
62953 98.828 -80.553 44.498
62954 98.852 -80.562 44.492
62955 98.875 -80.573 44.486
62956 98.899 -80.583 44.481
63280 93.336 -79.152 53.433
63281 93.345 -79.150 53.418
63282 93.355 -79.147 53.403
63283 93.365 -79.144 53.389
63284 93.374 -79.142 53.374
63285 93.384 -79.140 53.359
63286 93.394 -79.137 53.343
63287 93.404 -79.134 53.330
63288 93.413 -79.132 53.316
63289 93.424 -79.123 53.301
63290 93.436 -79.112 53.289
63291 93.447 -79.102 53.275
63292 93.455 -79.095 53.262
63293 93.464 -79.088 53.251
63294 93.472 -79.081 53.241
63295 93.481 -79.074 53.234
63296 93.489 -79.068 53.228
63297 93.494 -79.073 53.226
63298 93.500 -79.078 53.225
63299 93.505 -79.083 53.226
63300 93.524 -79.070 53.227
63301 93.544 -79.058 53.227
63302 93.562 -79.046 53.227
63303 93.570 -79.050 53.225
63304 93.578 -79.054 53.222
63305 93.586 -79.059 53.218
63306 93.601 -79.059 53.213
63307 93.618 -79.058 53.206
63308 93.635 -79.057 53.197
63309 93.653 -79.056 53.187
63310 93.671 -79.056 53.178
63311 93.686 -79.056 53.164
63312 93.701 -79.053 53.150
63313 93.715 -79.049 53.135
63314 93.730 -79.046 53.121
63315 93.745 -79.042 53.106
63316 93.759 -79.038 53.090
63317 93.774 -79.035 53.077
63318 93.788 -79.031 53.063
63319 93.793 -79.030 53.049
63320 93.797 -79.029 53.034
63321 93.803 -79.027 53.021
63322 93.812 -79.024 53.006
63323 93.820 -79.021 52.990
63324 93.828 -79.018 52.974
63325 93.836 -79.015 52.958
63326 93.844 -79.012 52.944
63327 93.851 -79.010 52.931
63328 93.859 -79.007 52.921
63329 93.867 -79.004 52.910
63330 93.874 -79.006 52.897
63331 93.880 -79.009 52.883
63332 93.883 -79.012 52.869
63333 93.884 -79.015 52.855
63334 93.885 -79.017 52.842
63335 93.887 -79.020 52.828
63336 93.889 -79.022 52.814
63337 93.892 -79.024 52.799
63338 93.895 -79.027 52.783
63339 93.898 -79.029 52.767
63340 93.901 -79.031 52.751
63341 93.905 -79.040 52.735
63342 93.909 -79.047 52.721
63343 93.913 -79.056 52.708
63344 93.915 -79.059 52.693
63345 93.913 -79.056 52.680
63346 93.929 -79.057 52.666
63347 93.956 -79.059 52.651
63348 93.981 -79.062 52.632
63349 94.005 -79.062 52.612
63350 94.006 -79.062 52.589
63351 94.007 -79.059 52.562
63352 94.008 -79.055 52.533
63353 94.009 -79.052 52.504
63354 94.010 -79.048 52.473
63355 94.012 -79.044 52.439
63356 94.013 -79.041 52.406
63357 94.014 -79.037 52.372
63358 94.015 -79.034 52.334
63359 94.016 -79.031 52.300
63360 94.017 -79.028 52.268
63361 94.023 -79.024 52.238
63362 94.031 -79.024 52.213
63363 94.040 -79.022 52.191
63364 94.048 -79.021 52.171
63365 94.057 -79.018 52.149
63366 94.067 -79.017 52.129
63367 94.076 -79.015 52.109
63368 94.085 -79.013 52.088
63369 94.096 -79.009 52.066
63370 94.113 -78.999 52.045
63371 94.129 -78.990 52.024
63372 94.146 -78.981 52.001
63373 94.162 -78.971 51.977
63374 94.183 -78.940 51.954
63375 94.204 -78.909 51.932
63376 94.226 -78.879 51.911
63377 94.247 -78.849 51.888
63378 94.268 -78.818 51.862
63379 94.285 -78.808 51.828
63380 94.300 -78.806 51.793
63381 94.314 -78.809 51.757
63382 94.327 -78.811 51.720
63383 94.340 -78.814 51.684
63384 94.352 -78.820 51.648
63385 94.363 -78.824 51.613
63386 94.363 -78.788 51.577
63387 94.382 -78.796 51.540
63388 94.402 -78.798 51.504
63389 94.424 -78.808 51.470
63390 94.448 -78.820 51.437
63391 94.474 -78.845 51.406
63392 94.500 -78.869 51.377
63393 94.515 -78.877 51.346
63394 94.530 -78.886 51.314
63395 94.544 -78.894 51.281
63396 94.558 -78.902 51.247
63397 94.578 -78.906 51.210
63398 94.601 -78.908 51.171
63399 94.623 -78.910 51.134
63400 94.642 -78.914 51.097
63401 94.658 -78.918 51.062
63402 94.671 -78.916 51.027
63403 94.682 -78.910 50.988
63404 94.692 -78.903 50.950
63405 94.703 -78.897 50.912
63406 94.714 -78.890 50.877
63407 94.724 -78.884 50.839
63408 94.736 -78.877 50.800
63409 94.746 -78.871 50.760
63410 94.757 -78.865 50.719
63411 94.768 -78.858 50.677
63412 94.778 -78.852 50.635
63413 94.789 -78.845 50.593
63414 94.800 -78.838 50.549
63415 94.811 -78.832 50.505
63416 94.822 -78.825 50.464
63417 94.832 -78.819 50.425
63418 94.843 -78.812 50.387
63419 94.852 -78.820 50.349
63420 94.856 -78.830 50.311
63421 94.858 -78.838 50.272
63422 94.861 -78.846 50.233
63423 94.863 -78.855 50.194
63424 94.866 -78.863 50.154
63425 94.868 -78.871 50.115
63426 94.871 -78.879 50.076
63427 94.874 -78.887 50.037
63428 94.876 -78.896 49.999
63429 94.879 -78.904 49.962
63430 94.881 -78.912 49.924
63431 94.884 -78.921 49.886
63432 94.886 -78.929 49.848
63433 94.888 -78.937 49.809
63434 94.891 -78.945 49.770
63435 94.893 -78.953 49.732
63436 94.896 -78.962 49.694
63437 94.898 -78.970 49.657
63438 94.897 -78.992 49.620
63439 94.892 -79.034 49.585
63440 94.886 -79.077 49.549
63441 94.907 -79.093 49.515
63442 94.939 -79.100 49.482
63443 94.972 -79.107 49.449
63444 94.997 -79.115 49.415
63445 95.014 -79.124 49.382
63446 95.031 -79.134 49.349
63447 95.048 -79.143 49.315
63448 95.064 -79.153 49.282
63449 95.079 -79.168 49.249
63450 95.095 -79.182 49.215
63451 95.111 -79.197 49.178
63452 95.130 -79.208 49.141
63453 95.152 -79.217 49.102
63454 95.174 -79.226 49.062
63455 95.195 -79.235 49.021
63456 95.216 -79.245 48.979
63457 95.238 -79.254 48.938
63458 95.261 -79.262 48.895
63459 95.281 -79.266 48.853
63460 95.298 -79.261 48.810
63461 95.316 -79.255 48.766
63462 95.334 -79.250 48.723
63463 95.351 -79.244 48.680
63464 95.370 -79.237 48.637
63465 95.386 -79.236 48.593
63466 95.399 -79.252 48.550
63467 95.412 -79.267 48.507
63468 95.425 -79.283 48.463
63469 95.438 -79.299 48.420
63470 95.451 -79.314 48.378
63471 95.478 -79.324 48.335
63472 95.517 -79.328 48.293
63473 95.546 -79.322 48.252
63474 95.568 -79.308 48.212
63475 95.591 -79.295 48.172
63476 95.613 -79.281 48.133
63477 95.634 -79.269 48.093
63478 95.644 -79.264 48.053
63479 95.655 -79.259 48.014
63480 95.665 -79.253 47.975
63481 95.675 -79.247 47.936
63482 95.685 -79.242 47.898
63483 95.695 -79.237 47.859
63484 95.705 -79.231 47.821
63485 95.730 -79.234 47.782
63486 95.763 -79.244 47.744
63487 95.797 -79.253 47.707
63488 95.825 -79.266 47.669
63489 95.835 -79.294 47.632
63490 95.844 -79.323 47.594
63491 95.853 -79.351 47.556
63492 95.858 -79.328 47.520
63493 95.853 -79.285 47.483
63494 95.849 -79.241 47.447
63495 95.846 -79.197 47.411
63496 95.842 -79.153 47.375
63497 95.855 -79.130 47.339
63498 95.902 -79.152 47.303
63499 95.950 -79.173 47.267
63500 95.997 -79.194 47.232
63501 96.045 -79.216 47.197
63502 96.093 -79.237 47.162
63503 96.141 -79.259 47.128
63504 96.188 -79.280 47.093
63505 96.236 -79.302 47.058
63506 96.262 -79.313 47.025
63507 96.283 -79.321 46.993
63508 96.303 -79.329 46.960
63509 96.323 -79.337 46.927
63510 96.343 -79.345 46.895
63511 96.364 -79.353 46.863
63512 96.384 -79.362 46.831
63513 96.404 -79.369 46.800
63514 96.425 -79.378 46.770
63515 96.445 -79.386 46.740
63516 96.466 -79.395 46.710
63517 96.486 -79.406 46.680
63518 96.508 -79.416 46.650
63519 96.528 -79.426 46.621
63520 96.549 -79.437 46.593
63521 96.570 -79.447 46.565
63522 96.591 -79.457 46.538
63523 96.611 -79.468 46.512
63524 96.625 -79.471 46.487
63525 96.635 -79.469 46.462
63526 96.644 -79.467 46.435
63527 96.662 -79.468 46.411
63528 96.684 -79.471 46.389
63529 96.707 -79.473 46.365
63530 96.730 -79.476 46.341
63531 96.750 -79.460 46.316
63532 96.768 -79.439 46.292
63533 96.787 -79.419 46.268
63534 96.806 -79.397 46.246
63535 96.824 -79.376 46.225
63536 96.843 -79.355 46.203
63537 96.870 -79.341 46.181
63538 96.903 -79.332 46.159
63539 96.949 -79.344 46.136
63540 96.994 -79.357 46.111
63541 97.043 -79.347 46.088
63542 97.058 -79.307 46.067
63543 97.098 -79.323 46.047
63544 97.113 -79.324 46.025
63545 97.127 -79.325 46.002
63546 97.141 -79.329 45.978
63547 97.155 -79.334 45.954
63548 97.169 -79.338 45.932
63549 97.191 -79.342 45.912
63550 97.213 -79.346 45.887
63551 97.234 -79.353 45.864
63552 97.264 -79.360 45.841
63553 97.290 -79.371 45.817
63554 97.307 -79.375 45.793
63555 97.306 -79.386 45.771
63556 97.299 -79.396 45.749
63557 97.292 -79.406 45.729
63558 97.285 -79.415 45.708
63559 97.278 -79.425 45.688
63560 97.271 -79.435 45.668
63561 97.270 -79.432 45.648
63562 97.276 -79.419 45.630
63563 97.281 -79.408 45.613
63564 97.280 -79.413 45.595
63565 97.278 -79.418 45.577
63566 97.277 -79.423 45.559
63567 97.263 -79.445 45.542
63568 97.253 -79.468 45.524
63569 97.260 -79.476 45.507
63570 97.267 -79.484 45.490
63571 97.275 -79.491 45.473
63572 97.282 -79.499 45.456
63573 97.289 -79.507 45.439
63574 97.296 -79.514 45.422
63575 97.303 -79.522 45.404
63576 97.310 -79.529 45.386
63577 97.317 -79.537 45.370
63578 97.324 -79.545 45.354
63579 97.331 -79.552 45.338
63580 97.338 -79.560 45.322
63581 97.355 -79.573 45.306
63582 97.374 -79.588 45.290
63583 97.393 -79.604 45.274
63584 97.412 -79.619 45.259
63585 97.431 -79.635 45.244
63586 97.450 -79.650 45.229
63587 97.469 -79.665 45.213
63588 97.489 -79.680 45.198
63630 98.294 -80.322 44.680
63631 98.313 -80.337 44.672
63632 98.332 -80.353 44.662
63633 98.351 -80.368 44.655
63634 98.370 -80.383 44.648
63635 98.390 -80.398 44.641
63636 98.409 -80.414 44.634
63637 98.428 -80.429 44.627
63638 98.447 -80.444 44.619
63639 98.466 -80.460 44.612
63640 98.486 -80.475 44.605
63641 98.508 -80.482 44.599
63642 98.533 -80.484 44.592
63643 98.558 -80.487 44.585
63644 98.583 -80.489 44.578
63645 98.607 -80.492 44.569
63646 98.632 -80.494 44.560
63647 98.657 -80.497 44.553
63648 98.682 -80.500 44.547
63649 98.707 -80.502 44.539
63650 98.731 -80.505 44.532
63651 98.757 -80.509 44.525
63652 98.783 -80.513 44.518
63653 98.808 -80.517 44.512
63654 98.833 -80.524 44.506
63655 98.856 -80.535 44.502
63656 98.879 -80.545 44.496
63657 98.903 -80.556 44.491
63658 98.926 -80.566 44.485
63983 93.355 -79.136 53.426
63984 93.365 -79.134 53.412
63985 93.374 -79.131 53.397
63986 93.384 -79.128 53.382
63987 93.394 -79.126 53.367
63988 93.404 -79.124 53.353
63989 93.413 -79.122 53.338
63990 93.423 -79.118 53.325
63991 93.434 -79.109 53.311
63992 93.446 -79.098 53.297
63993 93.457 -79.088 53.284
63994 93.465 -79.081 53.272
63995 93.474 -79.075 53.261
63996 93.482 -79.068 53.252
63997 93.490 -79.061 53.246
63998 93.498 -79.053 53.240
63999 93.505 -79.056 53.237
64000 93.510 -79.061 53.237
64001 93.515 -79.066 53.238
64002 93.526 -79.065 53.239
64003 93.546 -79.052 53.240
64004 93.561 -79.044 53.240
64005 93.569 -79.049 53.238
64006 93.577 -79.053 53.235
64007 93.594 -79.052 53.232
64008 93.612 -79.051 53.225
64009 93.629 -79.050 53.217
64010 93.646 -79.049 53.207
64011 93.664 -79.046 53.198
64012 93.680 -79.043 53.185
64013 93.694 -79.040 53.171
64014 93.708 -79.037 53.155
64015 93.723 -79.034 53.142
64016 93.737 -79.030 53.128
64017 93.752 -79.027 53.113
64018 93.767 -79.023 53.101
64019 93.782 -79.020 53.088
64020 93.786 -79.018 53.075
64021 93.791 -79.016 53.060
64022 93.799 -79.014 53.046
64023 93.807 -79.011 53.032
64024 93.816 -79.008 53.017
64025 93.824 -79.005 53.001
64026 93.832 -79.002 52.987
64027 93.839 -78.999 52.974
64028 93.847 -78.996 52.961
64029 93.855 -78.994 52.951
64030 93.863 -78.991 52.941
64031 93.871 -78.990 52.928
64032 93.877 -78.993 52.914
64033 93.880 -78.996 52.900
64034 93.881 -78.999 52.885
64035 93.882 -79.001 52.872
64036 93.883 -79.003 52.858
64037 93.884 -79.006 52.843
64038 93.887 -79.009 52.826
64039 93.890 -79.011 52.809
64040 93.894 -79.013 52.791
64041 93.897 -79.016 52.775
64042 93.901 -79.024 52.758
64043 93.905 -79.032 52.743
64044 93.908 -79.039 52.727
64045 93.906 -79.036 52.712
64046 93.905 -79.033 52.697
64047 93.926 -79.034 52.681
64048 93.952 -79.037 52.665
64049 93.978 -79.039 52.646
64050 93.997 -79.043 52.626
64051 94.008 -79.044 52.603
64052 94.010 -79.038 52.576
64053 94.012 -79.032 52.546
64054 94.014 -79.026 52.517
64055 94.016 -79.020 52.489
64056 94.017 -79.015 52.456
64057 94.018 -79.012 52.423
64058 94.019 -79.009 52.388
64059 94.020 -79.006 52.352
64060 94.022 -79.002 52.319
64061 94.023 -78.999 52.289
64062 94.028 -78.996 52.260
64063 94.037 -78.995 52.236
64064 94.046 -78.993 52.214
64065 94.055 -78.991 52.193
64066 94.064 -78.989 52.170
64067 94.073 -78.987 52.148
64068 94.082 -78.986 52.126
64069 94.092 -78.984 52.105
64070 94.105 -78.977 52.083
64071 94.121 -78.968 52.062
64072 94.138 -78.959 52.042
64073 94.154 -78.949 52.018
64074 94.171 -78.939 51.995
64075 94.188 -78.927 51.973
64076 94.209 -78.897 51.951
64077 94.230 -78.867 51.928
64078 94.251 -78.836 51.903
64079 94.272 -78.805 51.875
64080 94.289 -78.800 51.842
64081 94.305 -78.797 51.807
64082 94.321 -78.795 51.770
64083 94.335 -78.794 51.733
64084 94.348 -78.797 51.698
64085 94.360 -78.802 51.663
64086 94.372 -78.807 51.628
64087 94.380 -78.799 51.593
64088 94.382 -78.767 51.557
64089 94.407 -78.789 51.520
64090 94.433 -78.814 51.487
64091 94.460 -78.838 51.454
64092 94.480 -78.854 51.422
64093 94.494 -78.862 51.392
64094 94.509 -78.871 51.361
64095 94.525 -78.877 51.330
64096 94.543 -78.881 51.298
64097 94.562 -78.885 51.265
64098 94.581 -78.890 51.228
64099 94.600 -78.895 51.190
64100 94.619 -78.899 51.153
64101 94.638 -78.904 51.118
64102 94.657 -78.909 51.083
64103 94.677 -78.914 51.050
64104 94.688 -78.907 51.012
64105 94.698 -78.899 50.973
64106 94.709 -78.892 50.934
64107 94.719 -78.884 50.898
64108 94.730 -78.877 50.862
64109 94.741 -78.869 50.824
64110 94.751 -78.861 50.784
64111 94.762 -78.854 50.743
64112 94.772 -78.846 50.701
64113 94.782 -78.839 50.660
64114 94.793 -78.831 50.617
64115 94.804 -78.824 50.573
64116 94.814 -78.816 50.529
64117 94.821 -78.812 50.487
64118 94.825 -78.811 50.450
64119 94.834 -78.812 50.414
64120 94.857 -78.814 50.378
64121 94.855 -78.815 50.339
64122 94.849 -78.814 50.299
64123 94.843 -78.814 50.259
64124 94.836 -78.814 50.217
64125 94.833 -78.817 50.176
64126 94.835 -78.825 50.135
64127 94.838 -78.833 50.095
64128 94.840 -78.842 50.056
64129 94.842 -78.849 50.017
64130 94.845 -78.858 49.980
64131 94.847 -78.866 49.943
64132 94.850 -78.874 49.904
64133 94.852 -78.883 49.865
64134 94.855 -78.891 49.827
64135 94.857 -78.899 49.788
64136 94.859 -78.908 49.750
64137 94.854 -78.950 49.712
64138 94.848 -78.993 49.675
64139 94.845 -79.032 49.638
64140 94.854 -79.059 49.603
64141 94.889 -79.065 49.569
64142 94.922 -79.071 49.534
64143 94.954 -79.078 49.499
64144 94.982 -79.085 49.465
64145 94.998 -79.095 49.431
64146 95.016 -79.104 49.398
64147 95.032 -79.114 49.365
64148 95.049 -79.124 49.331
64149 95.065 -79.133 49.297
64150 95.081 -79.146 49.263
64151 95.097 -79.161 49.229
64152 95.113 -79.175 49.192
64153 95.134 -79.184 49.153
64154 95.156 -79.194 49.114
64155 95.177 -79.203 49.074
64156 95.199 -79.212 49.032
64157 95.222 -79.220 48.990
64158 95.247 -79.225 48.947
64159 95.273 -79.228 48.904
64160 95.296 -79.233 48.861
64161 95.301 -79.237 48.817
64162 95.319 -79.231 48.773
64163 95.337 -79.225 48.730
64164 95.355 -79.219 48.686
64165 95.373 -79.213 48.643
64166 95.387 -79.223 48.599
64167 95.400 -79.239 48.555
64168 95.413 -79.255 48.512
64169 95.426 -79.270 48.468
64170 95.439 -79.286 48.425
64171 95.452 -79.302 48.383
64172 95.475 -79.313 48.340
64173 95.515 -79.317 48.299
64174 95.543 -79.310 48.258
64175 95.565 -79.297 48.219
64176 95.588 -79.284 48.178
64177 95.610 -79.271 48.138
64178 95.633 -79.257 48.097
64179 95.647 -79.250 48.057
64180 95.657 -79.244 48.018
64181 95.667 -79.238 47.979
64182 95.678 -79.233 47.940
64183 95.688 -79.228 47.901
64184 95.698 -79.222 47.862
64185 95.708 -79.216 47.823
64186 95.736 -79.222 47.785
64187 95.770 -79.231 47.746
64188 95.804 -79.241 47.708
64189 95.837 -79.250 47.670
64190 95.848 -79.278 47.632
64191 95.858 -79.306 47.594
64192 95.866 -79.332 47.556
64193 95.861 -79.310 47.519
64194 95.843 -79.262 47.483
64195 95.838 -79.218 47.446
64196 95.835 -79.174 47.411
64197 95.831 -79.130 47.375
64198 95.835 -79.096 47.339
64199 95.883 -79.118 47.302
64200 95.931 -79.140 47.266
64201 95.978 -79.161 47.230
64202 96.026 -79.182 47.195
64203 96.074 -79.204 47.159
64204 96.121 -79.225 47.124
64205 96.169 -79.247 47.089
64206 96.217 -79.268 47.055
64207 96.264 -79.290 47.021
64208 96.295 -79.303 46.988
64209 96.315 -79.310 46.955
64210 96.335 -79.319 46.922
64211 96.356 -79.327 46.890
64212 96.376 -79.335 46.858
64213 96.396 -79.343 46.826
64214 96.417 -79.353 46.795
64215 96.438 -79.363 46.765
64216 96.459 -79.373 46.735
64217 96.479 -79.384 46.705
64218 96.500 -79.394 46.675
64219 96.521 -79.404 46.645
64220 96.542 -79.415 46.616
64221 96.562 -79.426 46.588
64222 96.583 -79.435 46.560
64223 96.604 -79.446 46.532
64224 96.614 -79.444 46.506
64225 96.623 -79.443 46.479
64226 96.631 -79.441 46.454
64227 96.647 -79.441 46.426
64228 96.669 -79.443 46.402
64229 96.692 -79.445 46.379
64230 96.715 -79.448 46.355
64231 96.738 -79.451 46.331
64232 96.758 -79.437 46.306
64233 96.777 -79.416 46.280
64234 96.795 -79.395 46.256
64235 96.814 -79.374 46.233
64236 96.832 -79.353 46.212
64237 96.851 -79.332 46.190
64238 96.870 -79.311 46.168
64239 96.892 -79.293 46.147
64240 96.928 -79.289 46.125
64241 96.974 -79.301 46.100
64242 97.007 -79.277 46.078
64243 97.056 -79.294 46.058
64244 97.094 -79.312 46.037
64245 97.109 -79.313 46.015
64246 97.124 -79.313 45.992
64247 97.139 -79.314 45.968
64248 97.154 -79.315 45.945
64249 97.174 -79.316 45.924
64250 97.196 -79.320 45.904
64251 97.218 -79.324 45.880
64252 97.220 -79.350 45.859
64253 97.209 -79.381 45.837
64254 97.246 -79.363 45.814
64255 97.280 -79.353 45.790
64256 97.296 -79.357 45.768
64257 97.300 -79.370 45.747
64258 97.304 -79.382 45.726
64259 97.302 -79.394 45.707
64260 97.295 -79.404 45.687
64261 97.288 -79.413 45.667
64262 97.281 -79.422 45.647
64263 97.274 -79.432 45.629
64264 97.274 -79.428 45.613
64265 97.280 -79.415 45.595
64266 97.285 -79.403 45.577
64267 97.288 -79.399 45.559
64268 97.286 -79.404 45.542
64269 97.285 -79.409 45.524
64270 97.269 -79.434 45.506
64271 97.267 -79.449 45.490
64272 97.275 -79.457 45.473
64273 97.282 -79.464 45.456
64274 97.289 -79.472 45.439
64275 97.296 -79.479 45.422
64276 97.303 -79.487 45.405
64277 97.310 -79.494 45.387
64278 97.317 -79.503 45.372
64279 97.324 -79.510 45.355
64280 97.331 -79.518 45.339
64281 97.341 -79.527 45.325
64282 97.360 -79.542 45.308
64283 97.379 -79.557 45.292
64284 97.398 -79.573 45.277
64285 97.418 -79.588 45.261
64286 97.437 -79.603 45.247
64287 97.456 -79.618 45.232
64288 97.475 -79.634 45.216
64289 97.494 -79.649 45.202
64290 97.513 -79.664 45.187
64291 97.532 -79.679 45.172
64292 97.551 -79.695 45.158
64329 98.261 -80.260 44.706
64330 98.280 -80.275 44.697
64331 98.299 -80.291 44.689
64332 98.318 -80.306 44.680
64333 98.338 -80.322 44.671
64334 98.357 -80.337 44.664
64335 98.376 -80.352 44.657
64336 98.395 -80.367 44.650
64337 98.414 -80.382 44.643
64338 98.433 -80.398 44.636
64339 98.452 -80.413 44.629
64340 98.472 -80.428 44.621
64341 98.491 -80.444 44.615
64342 98.514 -80.450 44.609
64343 98.538 -80.453 44.602
64344 98.564 -80.455 44.595
64345 98.588 -80.458 44.587
64346 98.613 -80.460 44.578
64347 98.638 -80.463 44.569
64348 98.662 -80.466 44.562
64349 98.687 -80.468 44.556
64350 98.712 -80.471 44.549
64351 98.737 -80.474 44.542
64352 98.763 -80.478 44.535
64353 98.788 -80.482 44.528
64354 98.813 -80.486 44.521
64355 98.837 -80.497 44.515
64356 98.860 -80.507 44.510
64357 98.883 -80.518 44.505
64358 98.907 -80.529 44.499
64359 98.930 -80.539 44.493
64360 98.953 -80.549 44.488
64686 93.375 -79.120 53.420
64687 93.384 -79.118 53.405
64688 93.394 -79.115 53.390
64689 93.404 -79.112 53.376
64690 93.414 -79.110 53.362
64691 93.423 -79.108 53.348
64692 93.433 -79.105 53.333
64693 93.444 -79.094 53.319
64694 93.456 -79.084 53.306
64695 93.466 -79.075 53.294
64696 93.475 -79.068 53.283
64697 93.483 -79.061 53.273
64698 93.491 -79.054 53.265
64699 93.500 -79.047 53.258
64700 93.508 -79.040 53.253
64701 93.515 -79.039 53.250
64702 93.520 -79.044 53.249
64703 93.526 -79.049 53.249
64704 93.532 -79.054 53.251
64705 93.547 -79.046 53.253
64706 93.560 -79.043 53.254
64707 93.570 -79.046 53.253
64708 93.587 -79.045 53.250
64709 93.605 -79.044 53.243
64710 93.624 -79.040 53.235
64711 93.643 -79.034 53.226
64712 93.661 -79.030 53.217
64713 93.675 -79.028 53.205
64714 93.689 -79.025 53.191
64715 93.703 -79.023 53.176
64716 93.716 -79.021 53.164
64717 93.730 -79.018 53.151
64718 93.746 -79.015 53.138
64719 93.760 -79.011 53.126
64720 93.775 -79.008 53.114
64721 93.779 -79.006 53.101
64722 93.787 -79.003 53.087
64723 93.795 -79.000 53.073
64724 93.803 -78.998 53.058
64725 93.812 -78.995 53.044
64726 93.820 -78.992 53.029
64727 93.827 -78.990 53.016
64728 93.835 -78.987 53.004
64729 93.843 -78.984 52.992
64730 93.851 -78.981 52.982
64731 93.859 -78.978 52.971
64732 93.867 -78.975 52.959
64733 93.874 -78.977 52.945
64734 93.876 -78.980 52.931
64735 93.877 -78.983 52.916
64736 93.879 -78.985 52.901
64737 93.880 -78.987 52.887
64738 93.881 -78.990 52.872
64739 93.883 -78.992 52.855
64740 93.886 -78.995 52.837
64741 93.889 -78.997 52.819
64742 93.892 -79.001 52.802
64743 93.896 -79.009 52.784
64744 93.900 -79.017 52.766
64745 93.899 -79.015 52.749
64746 93.898 -79.012 52.731
64747 93.897 -79.009 52.714
64748 93.923 -79.012 52.696
64749 93.949 -79.014 52.677
64750 93.968 -79.019 52.658
64751 93.985 -79.024 52.637
64752 94.003 -79.029 52.615
64753 94.009 -79.027 52.588
64754 94.011 -79.021 52.559
64755 94.013 -79.015 52.530
64756 94.015 -79.009 52.502
64757 94.017 -79.002 52.471
64758 94.019 -78.996 52.439
64759 94.022 -78.990 52.406
64760 94.024 -78.984 52.372
64761 94.026 -78.978 52.341
64762 94.028 -78.972 52.311
64763 94.034 -78.967 52.284
64764 94.044 -78.965 52.261
64765 94.053 -78.963 52.239
64766 94.062 -78.962 52.216
64767 94.071 -78.959 52.192
64768 94.080 -78.958 52.168
64769 94.089 -78.956 52.144
64770 94.099 -78.954 52.123
64771 94.114 -78.946 52.102
64772 94.130 -78.937 52.081
64773 94.146 -78.927 52.061
64774 94.163 -78.917 52.037
64775 94.180 -78.907 52.014
64776 94.196 -78.897 51.993
64777 94.213 -78.885 51.969
64778 94.235 -78.855 51.945
64779 94.256 -78.824 51.919
64780 94.276 -78.796 51.889
64781 94.292 -78.793 51.856
64782 94.308 -78.789 51.821
64783 94.324 -78.786 51.784
64784 94.340 -78.784 51.748
64785 94.356 -78.781 51.714
64786 94.368 -78.786 51.680
64787 94.379 -78.793 51.646
64788 94.391 -78.802 51.610
64789 94.395 -78.786 51.574
64790 94.419 -78.807 51.538
64791 94.445 -78.830 51.503
64792 94.459 -78.839 51.470
64793 94.475 -78.845 51.439
64794 94.494 -78.849 51.408
64795 94.512 -78.853 51.377
64796 94.531 -78.857 51.346
64797 94.550 -78.861 51.315
64798 94.568 -78.865 51.282
64799 94.587 -78.870 51.246
64800 94.607 -78.874 51.209
64801 94.626 -78.880 51.172
64802 94.645 -78.884 51.137
64803 94.663 -78.888 51.103
64804 94.677 -78.889 51.070
64805 94.690 -78.885 51.034
64806 94.700 -78.877 50.995
64807 94.711 -78.870 50.957
64808 94.721 -78.862 50.921
64809 94.732 -78.855 50.885
64810 94.742 -78.847 50.847
64811 94.753 -78.839 50.807
64812 94.763 -78.832 50.765
64813 94.774 -78.824 50.723
64814 94.784 -78.817 50.681
64815 94.789 -78.815 50.638
64816 94.793 -78.814 50.594
64817 94.797 -78.814 50.550
64818 94.801 -78.813 50.507
64819 94.805 -78.813 50.469
64820 94.827 -78.815 50.434
64821 94.850 -78.818 50.399
64822 94.874 -78.821 50.362
64823 94.871 -78.821 50.324
64824 94.864 -78.821 50.285
64825 94.858 -78.821 50.243
64826 94.851 -78.821 50.200
64827 94.845 -78.820 50.158
64828 94.839 -78.820 50.116
64829 94.832 -78.820 50.076
64830 94.826 -78.820 50.037
64831 94.820 -78.820 49.999
64832 94.814 -78.820 49.961
64833 94.817 -78.828 49.924
64834 94.819 -78.836 49.886
64835 94.817 -78.866 49.847
64836 94.824 -78.895 49.809
64837 94.832 -78.923 49.771
64838 94.840 -78.951 49.733
64839 94.848 -78.979 49.694
64840 94.856 -79.007 49.658
64841 94.863 -79.035 49.623
64842 94.896 -79.045 49.588
64843 94.935 -79.049 49.553
64844 94.966 -79.056 49.518
64845 94.983 -79.065 49.485
64846 95.000 -79.075 49.452
64847 95.017 -79.085 49.419
64848 95.033 -79.094 49.386
64849 95.050 -79.104 49.352
64850 95.066 -79.113 49.316
64851 95.083 -79.125 49.280
64852 95.098 -79.139 49.243
64853 95.117 -79.151 49.206
64854 95.138 -79.160 49.167
64855 95.160 -79.169 49.126
64856 95.182 -79.178 49.085
64857 95.207 -79.183 49.043
64858 95.233 -79.187 49.000
64859 95.259 -79.191 48.957
64860 95.284 -79.196 48.913
64861 95.310 -79.200 48.869
64862 95.314 -79.206 48.824
64863 95.323 -79.208 48.781
64864 95.341 -79.202 48.737
64865 95.358 -79.196 48.694
64866 95.376 -79.195 48.650
64867 95.388 -79.211 48.606
64868 95.401 -79.226 48.562
64869 95.414 -79.242 48.517
64870 95.427 -79.258 48.474
64871 95.440 -79.273 48.431
64872 95.453 -79.289 48.389
64873 95.472 -79.302 48.348
64874 95.512 -79.306 48.306
64875 95.540 -79.299 48.265
64876 95.562 -79.285 48.225
64877 95.585 -79.272 48.184
64878 95.607 -79.259 48.142
64879 95.630 -79.246 48.100
64880 95.650 -79.234 48.060
64881 95.660 -79.229 48.020
64882 95.670 -79.224 47.981
64883 95.680 -79.218 47.943
64884 95.690 -79.212 47.904
64885 95.700 -79.207 47.865
64886 95.711 -79.202 47.826
64887 95.743 -79.210 47.787
64888 95.776 -79.219 47.748
64889 95.810 -79.228 47.710
64890 95.844 -79.238 47.672
64891 95.862 -79.260 47.634
64892 95.871 -79.288 47.595
64893 95.875 -79.297 47.557
64894 95.870 -79.278 47.520
64895 95.861 -79.247 47.484
64896 95.843 -79.200 47.448
64897 95.825 -79.151 47.412
64898 95.820 -79.106 47.376
64899 95.816 -79.063 47.340
64900 95.864 -79.084 47.303
64901 95.911 -79.106 47.266
64902 95.959 -79.128 47.230
64903 96.007 -79.149 47.194
64904 96.054 -79.170 47.158
64905 96.102 -79.192 47.122
64906 96.150 -79.213 47.087
64907 96.197 -79.234 47.052
64908 96.245 -79.256 47.017
64909 96.293 -79.278 46.984
64910 96.327 -79.292 46.951
64911 96.347 -79.300 46.918
64912 96.368 -79.310 46.885
64913 96.389 -79.320 46.853
64914 96.409 -79.331 46.821
64915 96.431 -79.341 46.791
64916 96.451 -79.351 46.760
64917 96.472 -79.362 46.730
64918 96.493 -79.372 46.700
64919 96.514 -79.382 46.671
64920 96.534 -79.393 46.641
64921 96.555 -79.404 46.612
64922 96.576 -79.414 46.584
64923 96.593 -79.420 46.556
64924 96.602 -79.418 46.527
64925 96.611 -79.416 46.500
64926 96.620 -79.414 46.473
64927 96.631 -79.413 46.445
64928 96.654 -79.416 46.418
64929 96.677 -79.418 46.393
64930 96.700 -79.420 46.369
64931 96.723 -79.423 46.345
64932 96.745 -79.426 46.321
64933 96.765 -79.413 46.297
64934 96.784 -79.393 46.271
64935 96.803 -79.372 46.246
64936 96.821 -79.350 46.223
64937 96.840 -79.329 46.201
64938 96.859 -79.309 46.179
64939 96.877 -79.288 46.159
64940 96.896 -79.266 46.137
64941 96.914 -79.245 46.115
64942 96.955 -79.245 46.091
64943 97.000 -79.255 46.069
64944 97.058 -79.287 46.047
64945 97.091 -79.300 46.026
64946 97.105 -79.301 46.005
64947 97.120 -79.302 45.983
64948 97.135 -79.303 45.961
64949 97.157 -79.304 45.939
64950 97.179 -79.306 45.919
64951 97.197 -79.307 45.898
64952 97.205 -79.313 45.875
64953 97.171 -79.351 45.854
64954 97.180 -79.355 45.832
64955 97.215 -79.345 45.809
64956 97.249 -79.335 45.786
64957 97.282 -79.327 45.764
64958 97.286 -79.340 45.744
64959 97.290 -79.352 45.724
64960 97.294 -79.365 45.705
64961 97.298 -79.378 45.685
64962 97.302 -79.391 45.666
64963 97.298 -79.401 45.647
64964 97.291 -79.411 45.629
64965 97.284 -79.420 45.612
64966 97.277 -79.430 45.594
64967 97.279 -79.423 45.576
64968 97.284 -79.410 45.560
64969 97.290 -79.398 45.542
64970 97.295 -79.385 45.524
64971 97.294 -79.390 45.507
64972 97.293 -79.395 45.489
64973 97.276 -79.422 45.474
64974 97.282 -79.430 45.457
64975 97.289 -79.438 45.440
64976 97.296 -79.445 45.423
64977 97.303 -79.453 45.405
64978 97.310 -79.460 45.389
64979 97.317 -79.468 45.373
64980 97.324 -79.476 45.357
64981 97.331 -79.483 45.341
64982 97.346 -79.495 45.326
64983 97.365 -79.511 45.311
64984 97.384 -79.526 45.295
64985 97.403 -79.541 45.280
64986 97.423 -79.557 45.265
64987 97.442 -79.572 45.250
64988 97.461 -79.587 45.235
64989 97.480 -79.602 45.220
64990 97.500 -79.618 45.205
64991 97.519 -79.633 45.191
64992 97.538 -79.648 45.176
64993 97.557 -79.664 45.163
64994 97.576 -79.679 45.149
64995 97.595 -79.694 45.135
64996 97.614 -79.710 45.121
64997 97.634 -79.725 45.107
65028 98.228 -80.199 44.732
65029 98.247 -80.214 44.724
65030 98.266 -80.229 44.714
65031 98.286 -80.244 44.706
65032 98.305 -80.259 44.697
65033 98.324 -80.275 44.689
65034 98.343 -80.290 44.681
65035 98.362 -80.306 44.673
65036 98.381 -80.321 44.666
65037 98.401 -80.336 44.659
65038 98.420 -80.351 44.653
65039 98.439 -80.366 44.646
65040 98.458 -80.382 44.638
65041 98.477 -80.397 44.631
65042 98.496 -80.413 44.625
65043 98.519 -80.419 44.618
65044 98.544 -80.422 44.611
65045 98.569 -80.424 44.604
65046 98.593 -80.427 44.597
65047 98.619 -80.429 44.588
65048 98.643 -80.431 44.579
65049 98.668 -80.435 44.572
65050 98.693 -80.437 44.565
65051 98.717 -80.440 44.558
65052 98.742 -80.442 44.551
65053 98.768 -80.447 44.544
65054 98.793 -80.450 44.538
65055 98.817 -80.459 44.531
65056 98.841 -80.469 44.525
65057 98.864 -80.480 44.519
65058 98.887 -80.491 44.514
65059 98.911 -80.501 44.508
65060 98.934 -80.512 44.503
65061 98.957 -80.522 44.497
65062 98.981 -80.532 44.491
65388 93.385 -79.106 53.428
65389 93.395 -79.104 53.413
65390 93.404 -79.102 53.399
65391 93.414 -79.100 53.385
65392 93.424 -79.096 53.371
65393 93.433 -79.094 53.356
65394 93.443 -79.090 53.342
65395 93.454 -79.080 53.329
65396 93.466 -79.070 53.316
65397 93.476 -79.062 53.305
65398 93.484 -79.055 53.294
65399 93.492 -79.047 53.285
65400 93.501 -79.040 53.277
65401 93.509 -79.034 53.270
65402 93.517 -79.027 53.266
65403 93.525 -79.022 53.263
65404 93.531 -79.027 53.263
65405 93.536 -79.032 53.264
65406 93.542 -79.037 53.268
65407 93.549 -79.040 53.270
65408 93.564 -79.039 53.271
65409 93.583 -79.034 53.268
65410 93.602 -79.028 53.262
65411 93.621 -79.024 53.254
65412 93.640 -79.018 53.244
65413 93.656 -79.015 53.235
65414 93.671 -79.012 53.223
65415 93.684 -79.010 53.211
65416 93.698 -79.008 53.199
65417 93.712 -79.006 53.187
65418 93.726 -79.003 53.176
65419 93.740 -79.001 53.164
65420 93.754 -78.999 53.153
65421 93.768 -78.996 53.141
65422 93.775 -78.993 53.128
65423 93.783 -78.990 53.114
65424 93.791 -78.987 53.101
65425 93.799 -78.985 53.087
65426 93.807 -78.982 53.072
65427 93.815 -78.979 53.058
65428 93.823 -78.977 53.046
65429 93.831 -78.974 53.035
65430 93.839 -78.971 53.024
65431 93.847 -78.968 53.013
65432 93.855 -78.965 53.001
65433 93.863 -78.962 52.989
65434 93.871 -78.961 52.976
65435 93.873 -78.964 52.962
65436 93.874 -78.966 52.947
65437 93.875 -78.968 52.931
65438 93.876 -78.971 52.917
65439 93.877 -78.974 52.902
65440 93.878 -78.976 52.884
65441 93.881 -78.978 52.866
65442 93.884 -78.981 52.848
65443 93.888 -78.985 52.830
65444 93.892 -78.993 52.810
65445 93.892 -78.994 52.791
65446 93.890 -78.991 52.771
65447 93.889 -78.988 52.751
65448 93.894 -78.987 52.731
65449 93.920 -78.989 52.710
65450 93.938 -78.994 52.689
65451 93.956 -78.999 52.668
65452 93.973 -79.004 52.647
65453 93.991 -79.009 52.624
65454 94.009 -79.015 52.598
65455 94.011 -79.009 52.571
65456 94.013 -79.002 52.542
65457 94.015 -78.997 52.514
65458 94.017 -78.991 52.484
65459 94.019 -78.985 52.454
65460 94.022 -78.979 52.424
65461 94.023 -78.973 52.392
65462 94.027 -78.965 52.363
65463 94.037 -78.955 52.335
65464 94.045 -78.944 52.310
65465 94.053 -78.940 52.286
65466 94.060 -78.934 52.263
65467 94.068 -78.932 52.241
65468 94.078 -78.930 52.216
65469 94.087 -78.928 52.191
65470 94.096 -78.927 52.168
65471 94.106 -78.924 52.145
65472 94.122 -78.915 52.123
65473 94.139 -78.905 52.101
65474 94.155 -78.895 52.080
65475 94.172 -78.885 52.056
65476 94.188 -78.875 52.033
65477 94.205 -78.866 52.010
65478 94.221 -78.856 51.986
65479 94.239 -78.843 51.962
65480 94.260 -78.811 51.935
65481 94.280 -78.788 51.905
65482 94.296 -78.785 51.871
65483 94.312 -78.782 51.836
65484 94.328 -78.780 51.801
65485 94.343 -78.780 51.765
65486 94.359 -78.781 51.731
65487 94.374 -78.776 51.699
65488 94.386 -78.782 51.665
65489 94.397 -78.791 51.629
65490 94.409 -78.805 51.592
65491 94.426 -78.813 51.557
65492 94.445 -78.817 51.521
65493 94.463 -78.821 51.487
65494 94.482 -78.825 51.455
65495 94.500 -78.829 51.424
65496 94.519 -78.833 51.393
65497 94.538 -78.837 51.361
65498 94.556 -78.841 51.331
65499 94.575 -78.846 51.298
65500 94.594 -78.849 51.264
65501 94.613 -78.855 51.228
65502 94.632 -78.859 51.190
65503 94.648 -78.861 51.155
65504 94.663 -78.862 51.121
65505 94.678 -78.863 51.089
65506 94.691 -78.862 51.054
65507 94.702 -78.855 51.017
65508 94.712 -78.847 50.979
65509 94.723 -78.839 50.943
65510 94.734 -78.832 50.907
65511 94.744 -78.824 50.869
65512 94.753 -78.818 50.827
65513 94.757 -78.818 50.786
65514 94.761 -78.818 50.743
65515 94.765 -78.817 50.699
65516 94.769 -78.816 50.656
65517 94.773 -78.815 50.612
65518 94.777 -78.814 50.569
65519 94.781 -78.814 50.525
65520 94.796 -78.816 50.487
65521 94.820 -78.818 50.451
65522 94.843 -78.821 50.417
65523 94.866 -78.824 50.382
65524 94.890 -78.827 50.345
65525 94.886 -78.827 50.307
65526 94.880 -78.827 50.266
65527 94.873 -78.826 50.223
65528 94.864 -78.826 50.181
65529 94.857 -78.825 50.139
65530 94.849 -78.825 50.097
65531 94.841 -78.824 50.058
65532 94.833 -78.824 50.020
65533 94.825 -78.824 49.983
65534 94.817 -78.823 49.946
65535 94.816 -78.824 49.908
65536 94.827 -78.842 49.867
65537 94.835 -78.870 49.829
65538 94.843 -78.898 49.790
65539 94.850 -78.926 49.750
65540 94.858 -78.954 49.711
65541 94.866 -78.982 49.674
65542 94.874 -79.010 49.638
65543 94.898 -79.026 49.604
65544 94.937 -79.031 49.569
65545 94.968 -79.037 49.534
65546 94.985 -79.046 49.501
65547 95.001 -79.056 49.469
65548 95.018 -79.065 49.436
65549 95.034 -79.075 49.404
65550 95.051 -79.084 49.370
65551 95.067 -79.093 49.333
65552 95.084 -79.103 49.296
65553 95.100 -79.117 49.258
65554 95.120 -79.128 49.219
65555 95.142 -79.137 49.178
65556 95.168 -79.141 49.137
65557 95.193 -79.146 49.095
65558 95.219 -79.150 49.052
65559 95.245 -79.154 49.009
65560 95.270 -79.159 48.965
65561 95.296 -79.163 48.922
65562 95.321 -79.167 48.878
65563 95.329 -79.173 48.835
65564 95.332 -79.179 48.791
65565 95.344 -79.178 48.747
65566 95.362 -79.172 48.703
65567 95.377 -79.182 48.660
65568 95.389 -79.198 48.616
65569 95.402 -79.214 48.571
65570 95.415 -79.229 48.526
65571 95.428 -79.245 48.483
65572 95.441 -79.261 48.441
65573 95.454 -79.276 48.398
65574 95.470 -79.291 48.356
65575 95.509 -79.295 48.314
65576 95.537 -79.288 48.272
65577 95.559 -79.275 48.230
65578 95.582 -79.261 48.189
65579 95.604 -79.248 48.147
65580 95.627 -79.234 48.105
65581 95.649 -79.222 48.065
65582 95.663 -79.214 48.025
65583 95.673 -79.209 47.986
65584 95.683 -79.203 47.947
65585 95.693 -79.198 47.907
65586 95.703 -79.192 47.868
65587 95.716 -79.188 47.829
65588 95.750 -79.198 47.790
65589 95.783 -79.207 47.751
65590 95.817 -79.216 47.712
65591 95.850 -79.225 47.673
65592 95.875 -79.243 47.635
65593 95.884 -79.271 47.596
65594 95.883 -79.262 47.558
65595 95.878 -79.242 47.520
65596 95.873 -79.222 47.484
65597 95.861 -79.185 47.449
65598 95.843 -79.137 47.413
65599 95.826 -79.089 47.377
65600 95.808 -79.040 47.340
65601 95.848 -79.058 47.303
65602 95.898 -79.082 47.266
65603 95.948 -79.106 47.229
65604 95.997 -79.131 47.192
65605 96.046 -79.156 47.157
65606 96.096 -79.180 47.120
65607 96.145 -79.204 47.084
65608 96.194 -79.229 47.049
65609 96.244 -79.253 47.014
65610 96.294 -79.278 46.980
65611 96.325 -79.235 46.947
65612 96.356 -79.283 46.915
65613 96.381 -79.298 46.882
65614 96.402 -79.309 46.850
65615 96.423 -79.319 46.819
65616 96.444 -79.329 46.789
65617 96.465 -79.340 46.758
65618 96.485 -79.350 46.728
65619 96.506 -79.361 46.699
65620 96.527 -79.371 46.669
65621 96.548 -79.382 46.639
65622 96.569 -79.392 46.609
65623 96.581 -79.394 46.580
65624 96.590 -79.391 46.552
65625 96.599 -79.390 46.524
65626 96.608 -79.388 46.496
65627 96.617 -79.385 46.468
65628 96.639 -79.388 46.440
65629 96.662 -79.391 46.413
65630 96.685 -79.393 46.387
65631 96.708 -79.395 46.362
65632 96.730 -79.398 46.337
65633 96.753 -79.400 46.313
65634 96.773 -79.391 46.288
65635 96.792 -79.369 46.262
65636 96.811 -79.348 46.237
65637 96.829 -79.327 46.215
65638 96.848 -79.306 46.193
65639 96.867 -79.285 46.172
65640 96.885 -79.264 46.150
65641 96.904 -79.243 46.129
65642 96.923 -79.222 46.107
65643 96.950 -79.216 46.084
65644 97.002 -79.247 46.061
65645 97.060 -79.279 46.040
65646 97.087 -79.289 46.019
65647 97.102 -79.290 45.998
65648 97.117 -79.291 45.976
65649 97.140 -79.292 45.954
65650 97.159 -79.294 45.932
65651 97.173 -79.294 45.911
65652 97.188 -79.294 45.890
65653 97.197 -79.298 45.869
65654 97.163 -79.338 45.848
65655 97.164 -79.336 45.826
65656 97.183 -79.327 45.804
65657 97.218 -79.317 45.782
65658 97.252 -79.307 45.761
65659 97.271 -79.310 45.742
65660 97.275 -79.322 45.723
65661 97.279 -79.335 45.703
65662 97.283 -79.347 45.684
65663 97.287 -79.360 45.665
65664 97.291 -79.373 45.647
65665 97.295 -79.385 45.629
65666 97.299 -79.398 45.611
65667 97.294 -79.409 45.594
65668 97.287 -79.418 45.577
65669 97.280 -79.428 45.560
65670 97.283 -79.419 45.542
65671 97.288 -79.407 45.524
65672 97.294 -79.394 45.507
65673 97.300 -79.381 45.490
65674 97.302 -79.376 45.474
65675 97.301 -79.381 45.458
65676 97.289 -79.403 45.441
65677 97.296 -79.410 45.424
65678 97.303 -79.418 45.407
65679 97.310 -79.426 45.390
65680 97.318 -79.433 45.374
65681 97.325 -79.441 45.359
65682 97.332 -79.449 45.343
65683 97.352 -79.464 45.328
65684 97.371 -79.479 45.313
65685 97.390 -79.495 45.298
65686 97.409 -79.510 45.283
65687 97.428 -79.526 45.268
65688 97.447 -79.541 45.253
65689 97.466 -79.556 45.238
65690 97.486 -79.571 45.224
65691 97.505 -79.586 45.209
65692 97.524 -79.602 45.195
65693 97.543 -79.617 45.181
65694 97.563 -79.632 45.167
65695 97.582 -79.648 45.153
65696 97.601 -79.663 45.139
65697 97.620 -79.678 45.126
65698 97.639 -79.694 45.112
65699 97.658 -79.709 45.099
65700 97.677 -79.724 45.085
65701 97.697 -79.739 45.073
65702 97.716 -79.755 45.060
65703 97.735 -79.770 45.046
65704 97.754 -79.786 45.034
65705 97.773 -79.801 45.021
65724 98.138 -80.091 44.788
65725 98.157 -80.106 44.779
65726 98.176 -80.121 44.769
65727 98.195 -80.137 44.759
65728 98.214 -80.152 44.750
65729 98.233 -80.168 44.741
65730 98.252 -80.183 44.732
65731 98.272 -80.198 44.723
65732 98.291 -80.213 44.715
65733 98.310 -80.228 44.707
65734 98.329 -80.243 44.698
65735 98.348 -80.259 44.690
65736 98.367 -80.275 44.683
65737 98.387 -80.290 44.676
65738 98.406 -80.305 44.669
65739 98.425 -80.320 44.662
65740 98.444 -80.335 44.655
65741 98.463 -80.350 44.648
65742 98.483 -80.366 44.641
65743 98.502 -80.381 44.635
65744 98.525 -80.388 44.628
65745 98.550 -80.391 44.621
65746 98.574 -80.393 44.614
65747 98.599 -80.395 44.607
65748 98.624 -80.398 44.598
65749 98.649 -80.400 44.590
65750 98.673 -80.403 44.582
65751 98.698 -80.406 44.575
65752 98.723 -80.408 44.568
65753 98.747 -80.411 44.561
65754 98.773 -80.415 44.554
65755 98.798 -80.421 44.547
65756 98.821 -80.431 44.540
65757 98.845 -80.442 44.534
65758 98.868 -80.453 44.528
65759 98.891 -80.463 44.523
65760 98.915 -80.474 44.518
65761 98.938 -80.485 44.513
65762 98.961 -80.495 44.507
65763 98.985 -80.505 44.501
65764 99.008 -80.516 44.496
66091 93.405 -79.091 53.422
66092 93.414 -79.088 53.408
66093 93.424 -79.086 53.393
66094 93.433 -79.084 53.379
66095 93.443 -79.081 53.366
66096 93.453 -79.076 53.352
66097 93.465 -79.066 53.340
66098 93.476 -79.056 53.328
66099 93.485 -79.048 53.317
66100 93.493 -79.041 53.307
66101 93.502 -79.034 53.298
66102 93.510 -79.027 53.290
66103 93.518 -79.020 53.285
66104 93.527 -79.013 53.280
66105 93.535 -79.006 53.279
66106 93.541 -79.010 53.279
66107 93.547 -79.015 53.283
66108 93.552 -79.020 53.286
66109 93.561 -79.022 53.287
66110 93.580 -79.017 53.285
66111 93.599 -79.012 53.279
66112 93.618 -79.007 53.272
66113 93.637 -79.002 53.263
66114 93.652 -78.999 53.253
66115 93.666 -78.997 53.243
66116 93.680 -78.995 53.232
66117 93.694 -78.993 53.222
66118 93.707 -78.990 53.211
66119 93.721 -78.988 53.199
66120 93.735 -78.986 53.188
66121 93.751 -78.983 53.178
66122 93.763 -78.981 53.166
66123 93.770 -78.981 53.154
66124 93.777 -78.981 53.141
66125 93.785 -78.978 53.128
66126 93.793 -78.974 53.114
66127 93.801 -78.971 53.099
66128 93.810 -78.968 53.085
66129 93.818 -78.965 53.074
66130 93.826 -78.962 53.064
66131 93.834 -78.959 53.053
66132 93.843 -78.956 53.042
66133 93.851 -78.952 53.031
66134 93.859 -78.949 53.019
66135 93.867 -78.946 53.006
66136 93.870 -78.948 52.992
66137 93.871 -78.950 52.978
66138 93.872 -78.952 52.962
66139 93.873 -78.955 52.947
66140 93.874 -78.958 52.932
66141 93.875 -78.960 52.915
66142 93.876 -78.962 52.896
66143 93.880 -78.965 52.877
66144 93.883 -78.970 52.858
66145 93.885 -78.974 52.837
66146 93.883 -78.971 52.816
66147 93.882 -78.968 52.794
66148 93.880 -78.965 52.772
66149 93.891 -78.965 52.749
66150 93.908 -78.969 52.725
66151 93.926 -78.974 52.701
66152 93.944 -78.980 52.679
66153 93.962 -78.984 52.656
66154 93.979 -78.990 52.633
66155 93.997 -78.995 52.607
66156 94.011 -78.997 52.580
66157 94.013 -78.991 52.553
66158 94.015 -78.985 52.525
66159 94.017 -78.979 52.496
66160 94.019 -78.973 52.468
66161 94.021 -78.967 52.440
66162 94.029 -78.957 52.411
66163 94.038 -78.946 52.384
66164 94.047 -78.937 52.358
66165 94.056 -78.926 52.333
66166 94.064 -78.918 52.309
66167 94.072 -78.913 52.285
66168 94.079 -78.908 52.263
66169 94.086 -78.903 52.237
66170 94.094 -78.899 52.213
66171 94.103 -78.897 52.189
66172 94.115 -78.893 52.166
66173 94.131 -78.883 52.142
66174 94.147 -78.873 52.119
66175 94.164 -78.863 52.097
66176 94.181 -78.854 52.072
66177 94.197 -78.844 52.048
66178 94.214 -78.834 52.025
66179 94.230 -78.824 52.002
66180 94.247 -78.814 51.977
66181 94.264 -78.799 51.950
66182 94.283 -78.780 51.920
66183 94.299 -78.777 51.886
66184 94.315 -78.778 51.851
66185 94.331 -78.779 51.816
66186 94.346 -78.780 51.781
66187 94.362 -78.781 51.749
66188 94.378 -78.782 51.717
66189 94.393 -78.771 51.683
66190 94.404 -78.780 51.647
66191 94.419 -78.788 51.610
66192 94.437 -78.793 51.573
66193 94.455 -78.796 51.538
66194 94.472 -78.801 51.503
66195 94.490 -78.805 51.470
66196 94.507 -78.809 51.440
66197 94.526 -78.813 51.408
66198 94.544 -78.818 51.376
66199 94.563 -78.821 51.345
66200 94.581 -78.826 51.313
66201 94.600 -78.830 51.280
66202 94.619 -78.834 51.245
66203 94.634 -78.835 51.207
66204 94.648 -78.836 51.171
66205 94.663 -78.837 51.137
66206 94.678 -78.838 51.105
66207 94.692 -78.839 51.073
66208 94.703 -78.833 51.037
66209 94.714 -78.825 50.999
66210 94.721 -78.821 50.964
66211 94.725 -78.821 50.927
66212 94.729 -78.821 50.887
66213 94.733 -78.820 50.846
66214 94.737 -78.819 50.803
66215 94.741 -78.818 50.759
66216 94.745 -78.818 50.714
66217 94.749 -78.817 50.672
66218 94.753 -78.817 50.629
66219 94.757 -78.816 50.585
66220 94.766 -78.816 50.541
66221 94.790 -78.819 50.503
66222 94.813 -78.821 50.467
66223 94.836 -78.824 50.433
66224 94.860 -78.827 50.398
66225 94.889 -78.821 50.363
66226 94.893 -78.830 50.327
66227 94.885 -78.829 50.287
66228 94.878 -78.829 50.245
66229 94.870 -78.828 50.203
66230 94.861 -78.828 50.162
66231 94.853 -78.827 50.119
66232 94.845 -78.827 50.078
66233 94.838 -78.827 50.041
66234 94.830 -78.826 50.005
66235 94.822 -78.825 49.968
66236 94.817 -78.825 49.929
66237 94.833 -78.828 49.887
66238 94.845 -78.845 49.847
66239 94.853 -78.873 49.807
66240 94.861 -78.901 49.766
66241 94.868 -78.929 49.727
66242 94.877 -78.957 49.689
66243 94.885 -78.985 49.653
66244 94.900 -79.007 49.618
66245 94.939 -79.012 49.583
66246 94.971 -79.018 49.549
66247 94.987 -79.028 49.516
66248 95.003 -79.037 49.484
66249 95.020 -79.046 49.452
66250 95.036 -79.056 49.420
66251 95.052 -79.065 49.385
66252 95.069 -79.075 49.348
66253 95.085 -79.084 49.310
66254 95.103 -79.094 49.271
66255 95.128 -79.100 49.230
66256 95.154 -79.104 49.188
66257 95.179 -79.108 49.146
66258 95.205 -79.112 49.104
66259 95.231 -79.117 49.061
66260 95.256 -79.122 49.018
66261 95.282 -79.125 48.975
66262 95.307 -79.130 48.931
66263 95.333 -79.134 48.888
66264 95.345 -79.140 48.845
66265 95.347 -79.146 48.801
66266 95.350 -79.153 48.757
66267 95.365 -79.154 48.713
66268 95.378 -79.169 48.670
66269 95.390 -79.185 48.625
66270 95.403 -79.201 48.580
66271 95.416 -79.217 48.535
66272 95.429 -79.232 48.492
66273 95.442 -79.248 48.449
66274 95.455 -79.264 48.406
66275 95.468 -79.279 48.364
66276 95.506 -79.285 48.322
66277 95.534 -79.276 48.280
66278 95.556 -79.263 48.237
66279 95.579 -79.250 48.195
66280 95.601 -79.237 48.153
66281 95.624 -79.224 48.112
66282 95.646 -79.210 48.071
66283 95.665 -79.200 48.031
66284 95.675 -79.194 47.991
66285 95.685 -79.188 47.952
66286 95.696 -79.183 47.912
66287 95.706 -79.178 47.873
66288 95.723 -79.176 47.834
66289 95.756 -79.185 47.794
66290 95.790 -79.195 47.755
66291 95.824 -79.204 47.716
66292 95.857 -79.213 47.676
66293 95.888 -79.225 47.637
66294 95.895 -79.247 47.599
66295 95.891 -79.227 47.560
66296 95.886 -79.207 47.522
66297 95.881 -79.187 47.486
66298 95.877 -79.167 47.450
66299 95.863 -79.130 47.414
66300 95.849 -79.093 47.377
66301 95.836 -79.057 47.340
66302 95.867 -79.069 47.303
66303 95.917 -79.103 47.265
66304 95.969 -79.136 47.228
66305 96.019 -79.169 47.191
66306 96.068 -79.193 47.155
66307 96.118 -79.217 47.119
66308 96.168 -79.241 47.082
66309 96.217 -79.266 47.046
66310 96.252 -79.239 47.012
66311 96.281 -79.188 46.978
66312 96.310 -79.137 46.945
66313 96.341 -79.176 46.913
66314 96.372 -79.235 46.881
66315 96.403 -79.285 46.849
66316 96.431 -79.303 46.818
66317 96.457 -79.318 46.788
66318 96.478 -79.328 46.758
66319 96.499 -79.339 46.728
66320 96.520 -79.349 46.698
66321 96.540 -79.360 46.667
66322 96.560 -79.369 46.637
66323 96.570 -79.367 46.606
66324 96.579 -79.365 46.577
66325 96.588 -79.363 46.549
66326 96.596 -79.361 46.521
66327 96.605 -79.359 46.492
66328 96.624 -79.360 46.464
66329 96.647 -79.363 46.436
66330 96.670 -79.365 46.408
66331 96.692 -79.368 46.382
66332 96.715 -79.370 46.356
66333 96.738 -79.372 46.331
66334 96.761 -79.375 46.306
66335 96.782 -79.367 46.280
66336 96.800 -79.346 46.255
66337 96.819 -79.325 46.230
66338 96.837 -79.303 46.207
66339 96.856 -79.283 46.186
66340 96.875 -79.262 46.164
66341 96.893 -79.241 46.142
66342 96.912 -79.219 46.121
66343 96.934 -79.203 46.099
66344 96.971 -79.214 46.077
66345 97.004 -79.240 46.055
66346 97.062 -79.272 46.034
66347 97.083 -79.278 46.014
66348 97.100 -79.278 45.992
66349 97.120 -79.280 45.970
66350 97.135 -79.280 45.948
66351 97.150 -79.281 45.925
66352 97.165 -79.281 45.904
66353 97.179 -79.281 45.883
66354 97.190 -79.284 45.864
66355 97.156 -79.323 45.842
66356 97.164 -79.315 45.821
66357 97.166 -79.308 45.800
66358 97.186 -79.300 45.779
66359 97.221 -79.289 45.759
66360 97.255 -79.279 45.740
66361 97.260 -79.291 45.721
66362 97.264 -79.304 45.702
66363 97.268 -79.317 45.683
66364 97.272 -79.330 45.665
66365 97.276 -79.343 45.646
66366 97.280 -79.355 45.629
66367 97.284 -79.368 45.612
66368 97.288 -79.381 45.595
66369 97.292 -79.394 45.578
66370 97.296 -79.407 45.560
66371 97.290 -79.416 45.543
66372 97.283 -79.425 45.525
66373 97.287 -79.414 45.508
66374 97.293 -79.402 45.491
66375 97.298 -79.389 45.476
66376 97.304 -79.376 45.460
66377 97.309 -79.363 45.443
66378 97.309 -79.367 45.426
66379 97.304 -79.384 45.410
66380 97.311 -79.391 45.393
66381 97.318 -79.399 45.377
66382 97.325 -79.407 45.362
66383 97.337 -79.418 45.347
66384 97.357 -79.433 45.331
66385 97.376 -79.448 45.316
66386 97.395 -79.463 45.301
66387 97.415 -79.479 45.286
66388 97.434 -79.494 45.272
66389 97.453 -79.510 45.258
66390 97.472 -79.525 45.243
66391 97.491 -79.540 45.228
66392 97.510 -79.555 45.214
66393 97.529 -79.570 45.200
66394 97.548 -79.586 45.186
66395 97.568 -79.601 45.172
66396 97.587 -79.617 45.158
66397 97.606 -79.632 45.145
66398 97.625 -79.647 45.132
66399 97.645 -79.662 45.119
66400 97.664 -79.678 45.105
66401 97.683 -79.693 45.092
66402 97.702 -79.708 45.080
66403 97.721 -79.723 45.067
66404 97.740 -79.739 45.054
66405 97.759 -79.754 45.041
66406 97.779 -79.770 45.028
66407 97.798 -79.785 45.016
66408 97.817 -79.800 45.003
66409 97.836 -79.815 44.990
66410 97.855 -79.830 44.977
66411 97.874 -79.845 44.963
66412 97.893 -79.861 44.950
66413 97.913 -79.877 44.936
66414 97.932 -79.892 44.923
66415 97.951 -79.907 44.910
66416 97.970 -79.922 44.898
66417 97.990 -79.937 44.886
66418 98.009 -79.952 44.874
66419 98.028 -79.968 44.862
66420 98.047 -79.983 44.850
66421 98.066 -79.999 44.839
66422 98.085 -80.014 44.828
66423 98.104 -80.029 44.818
66424 98.124 -80.044 44.807
66425 98.143 -80.060 44.797
66426 98.162 -80.075 44.788
66427 98.181 -80.090 44.778
66428 98.201 -80.106 44.768
66429 98.220 -80.121 44.759
66430 98.239 -80.136 44.750
66431 98.258 -80.152 44.742
66432 98.277 -80.167 44.733
66433 98.296 -80.182 44.725
66434 98.315 -80.197 44.716
66435 98.335 -80.212 44.708
66436 98.354 -80.228 44.700
66437 98.373 -80.243 44.692
66438 98.392 -80.259 44.685
66439 98.411 -80.274 44.678
66440 98.430 -80.289 44.672
66441 98.449 -80.304 44.665
66442 98.469 -80.319 44.658
66443 98.488 -80.334 44.652
66444 98.507 -80.350 44.645
66445 98.530 -80.356 44.639
66446 98.555 -80.359 44.633
66447 98.580 -80.362 44.625
66448 98.604 -80.364 44.617
66449 98.629 -80.366 44.609
66450 98.654 -80.369 44.600
66451 98.678 -80.372 44.593
66452 98.704 -80.375 44.585
66453 98.728 -80.377 44.578
66454 98.753 -80.380 44.571
66455 98.779 -80.384 44.564
66456 98.802 -80.394 44.557
66457 98.826 -80.404 44.550
66458 98.849 -80.415 44.544
66459 98.872 -80.425 44.539
66460 98.896 -80.436 44.533
66461 98.919 -80.447 44.527
66462 98.942 -80.457 44.522
66463 98.966 -80.468 44.517
66464 98.989 -80.478 44.510
66465 99.012 -80.488 44.505
66466 99.035 -80.499 44.501
66794 93.424 -79.075 53.416
66795 93.434 -79.072 53.402
66796 93.443 -79.070 53.389
66797 93.453 -79.068 53.376
66798 93.464 -79.062 53.363
66799 93.475 -79.052 53.351
66800 93.486 -79.041 53.340
66801 93.495 -79.034 53.329
66802 93.503 -79.028 53.320
66803 93.511 -79.021 53.312
66804 93.519 -79.014 53.304
66805 93.528 -79.007 53.299
66806 93.536 -78.999 53.296
66807 93.545 -78.993 53.294
66808 93.551 -78.993 53.295
66809 93.557 -78.998 53.298
66810 93.562 -79.002 53.300
66811 93.577 -79.001 53.299
66812 93.596 -78.996 53.295
66813 93.615 -78.991 53.289
66814 93.633 -78.986 53.282
66815 93.647 -78.984 53.273
66816 93.661 -78.981 53.263
66817 93.675 -78.979 53.253
66818 93.689 -78.977 53.245
66819 93.703 -78.974 53.234
66820 93.717 -78.972 53.222
66821 93.734 -78.970 53.212
66822 93.750 -78.968 53.202
66823 93.759 -78.967 53.191
66824 93.766 -78.967 53.179
66825 93.773 -78.967 53.166
66826 93.780 -78.967 53.153
66827 93.786 -78.967 53.139
66828 93.794 -78.964 53.124
66829 93.802 -78.961 53.111
66830 93.811 -78.958 53.100
66831 93.819 -78.954 53.090
66832 93.827 -78.951 53.080
66833 93.836 -78.948 53.069
66834 93.844 -78.945 53.057
66835 93.852 -78.942 53.045
66836 93.862 -78.933 53.032
66837 93.865 -78.933 53.018
66838 93.866 -78.936 53.004
66839 93.868 -78.937 52.989
66840 93.869 -78.939 52.974
66841 93.871 -78.941 52.959
66842 93.872 -78.943 52.942
66843 93.873 -78.946 52.923
66844 93.875 -78.949 52.904
66845 93.877 -78.953 52.884
66846 93.876 -78.950 52.861
66847 93.874 -78.947 52.838
66848 93.872 -78.944 52.815
66849 93.869 -78.940 52.791
66850 93.880 -78.943 52.765
66851 93.897 -78.949 52.740
66852 93.914 -78.955 52.715
66853 93.932 -78.960 52.690
66854 93.949 -78.965 52.665
66855 93.967 -78.970 52.640
66856 93.985 -78.975 52.614
66857 94.003 -78.981 52.588
66858 94.013 -78.980 52.561
66859 94.015 -78.974 52.534
66860 94.017 -78.968 52.506
66861 94.022 -78.960 52.479
66862 94.031 -78.949 52.454
66863 94.040 -78.939 52.428
66864 94.049 -78.928 52.402
66865 94.058 -78.918 52.376
66866 94.067 -78.907 52.353
66867 94.076 -78.896 52.329
66868 94.083 -78.892 52.306
66869 94.091 -78.886 52.283
66870 94.098 -78.882 52.257
66871 94.105 -78.877 52.232
66872 94.113 -78.872 52.209
66873 94.125 -78.866 52.185
66874 94.142 -78.856 52.161
66875 94.158 -78.846 52.137
66876 94.175 -78.836 52.113
66877 94.191 -78.826 52.087
66878 94.207 -78.816 52.062
66879 94.224 -78.806 52.039
66880 94.241 -78.796 52.015
66881 94.257 -78.786 51.990
66882 94.273 -78.777 51.963
66883 94.287 -78.774 51.933
66884 94.302 -78.776 51.899
66885 94.318 -78.778 51.865
66886 94.334 -78.779 51.829
66887 94.349 -78.780 51.796
66888 94.365 -78.781 51.764
66889 94.381 -78.782 51.733
66890 94.397 -78.774 51.700
66891 94.413 -78.765 51.664
66892 94.430 -78.770 51.626
66893 94.448 -78.774 51.589
66894 94.466 -78.778 51.553
66895 94.483 -78.783 51.518
66896 94.500 -78.787 51.485
66897 94.518 -78.791 51.454
66898 94.535 -78.796 51.422
66899 94.553 -78.800 51.390
66900 94.570 -78.805 51.359
66901 94.588 -78.808 51.327
66902 94.608 -78.814 51.295
66903 94.626 -78.816 51.260
66904 94.640 -78.818 51.223
66905 94.655 -78.819 51.186
66906 94.671 -78.820 51.151
66907 94.685 -78.821 51.119
66908 94.700 -78.822 51.088
66909 94.698 -78.826 51.054
66910 94.697 -78.823 51.018
66911 94.701 -78.823 50.983
66912 94.705 -78.822 50.944
66913 94.709 -78.821 50.904
66914 94.713 -78.821 50.861
66915 94.717 -78.820 50.817
66916 94.721 -78.819 50.773
66917 94.725 -78.819 50.729
66918 94.729 -78.818 50.687
66919 94.733 -78.818 50.643
66920 94.737 -78.817 50.599
66921 94.760 -78.819 50.558
66922 94.783 -78.822 50.518
66923 94.806 -78.825 50.481
66924 94.833 -78.822 50.447
66925 94.864 -78.814 50.413
66926 94.887 -78.809 50.379
66927 94.892 -78.818 50.344
66928 94.890 -78.832 50.305
66929 94.882 -78.831 50.264
66930 94.874 -78.831 50.222
66931 94.866 -78.830 50.182
66932 94.858 -78.830 50.139
66933 94.850 -78.830 50.097
66934 94.842 -78.829 50.058
66935 94.834 -78.828 50.020
66936 94.826 -78.828 49.983
66937 94.819 -78.827 49.944
66938 94.834 -78.830 49.902
66939 94.851 -78.833 49.863
66940 94.863 -78.848 49.822
66941 94.872 -78.876 49.779
66942 94.879 -78.904 49.741
66943 94.887 -78.932 49.703
66944 94.895 -78.961 49.666
66945 94.903 -78.989 49.631
66946 94.942 -78.993 49.596
66947 94.973 -78.999 49.561
66948 94.990 -79.009 49.528
66949 95.006 -79.018 49.498
66950 95.023 -79.028 49.466
66951 95.039 -79.037 49.434
66952 95.055 -79.046 49.398
66953 95.071 -79.056 49.360
66954 95.089 -79.065 49.321
66955 95.119 -79.071 49.281
66956 95.143 -79.074 49.240
66957 95.168 -79.076 49.198
66958 95.192 -79.078 49.155
66959 95.217 -79.081 49.112
66960 95.242 -79.084 49.070
66961 95.268 -79.088 49.027
66962 95.293 -79.093 48.985
66963 95.319 -79.097 48.942
66964 95.344 -79.101 48.898
66965 95.360 -79.106 48.854
66966 95.362 -79.113 48.811
66967 95.365 -79.125 48.768
66968 95.368 -79.141 48.724
66969 95.379 -79.157 48.679
66970 95.391 -79.172 48.633
66971 95.404 -79.188 48.589
66972 95.417 -79.204 48.545
66973 95.430 -79.219 48.501
66974 95.443 -79.235 48.457
66975 95.456 -79.251 48.413
66976 95.469 -79.266 48.372
66977 95.503 -79.273 48.329
66978 95.531 -79.265 48.287
66979 95.553 -79.252 48.245
66980 95.576 -79.238 48.203
66981 95.598 -79.225 48.162
66982 95.621 -79.212 48.120
66983 95.643 -79.199 48.079
66984 95.666 -79.186 48.038
66985 95.678 -79.179 47.998
66986 95.688 -79.174 47.957
66987 95.698 -79.169 47.918
66988 95.708 -79.162 47.879
66989 95.729 -79.164 47.840
66990 95.763 -79.173 47.800
66991 95.797 -79.183 47.761
66992 95.830 -79.192 47.722
66993 95.864 -79.201 47.682
66994 95.898 -79.210 47.641
66995 95.903 -79.211 47.602
66996 95.899 -79.191 47.563
66997 95.894 -79.182 47.525
66998 95.890 -79.173 47.487
66999 95.886 -79.164 47.451
67000 95.891 -79.147 47.414
67001 95.878 -79.110 47.378
67002 95.864 -79.074 47.341
67003 95.884 -79.074 47.303
67004 95.935 -79.108 47.265
67005 95.986 -79.141 47.227
67006 96.037 -79.175 47.190
67007 96.088 -79.209 47.153
67008 96.138 -79.243 47.117
67009 96.182 -79.247 47.080
67010 96.209 -79.194 47.044
67011 96.238 -79.141 47.010
67012 96.266 -79.090 46.977
67013 96.295 -79.039 46.945
67014 96.326 -79.068 46.912
67015 96.357 -79.128 46.880
67016 96.388 -79.187 46.849
67017 96.420 -79.246 46.818
67018 96.450 -79.288 46.788
67019 96.478 -79.304 46.758
67020 96.506 -79.322 46.729
67021 96.533 -79.338 46.698
67022 96.549 -79.343 46.667
67023 96.558 -79.341 46.635
67024 96.567 -79.338 46.605
67025 96.576 -79.337 46.576
67026 96.585 -79.335 46.547
67027 96.594 -79.332 46.517
67028 96.609 -79.332 46.489
67029 96.631 -79.335 46.461
67030 96.655 -79.338 46.432
67031 96.677 -79.340 46.404
67032 96.700 -79.343 46.378
67033 96.723 -79.345 46.352
67034 96.746 -79.347 46.327
67035 96.768 -79.350 46.301
67036 96.790 -79.344 46.274
67037 96.808 -79.322 46.248
67038 96.827 -79.301 46.224
67039 96.846 -79.280 46.201
67040 96.864 -79.259 46.179
67041 96.883 -79.238 46.156
67042 96.901 -79.217 46.135
67043 96.920 -79.196 46.114
67044 96.955 -79.203 46.092
67045 96.992 -79.212 46.071
67046 97.029 -79.223 46.050
67047 97.064 -79.265 46.029
67048 97.082 -79.266 46.010
67049 97.097 -79.266 45.988
67050 97.112 -79.267 45.965
67051 97.126 -79.267 45.943
67052 97.141 -79.267 45.921
67053 97.156 -79.267 45.900
67054 97.171 -79.267 45.880
67055 97.183 -79.269 45.861
67056 97.152 -79.306 45.840
67057 97.165 -79.294 45.819
67058 97.166 -79.287 45.798
67059 97.168 -79.280 45.777
67060 97.190 -79.272 45.758
67061 97.224 -79.261 45.740
67062 97.245 -79.261 45.722
67063 97.249 -79.274 45.703
67064 97.253 -79.287 45.685
67065 97.257 -79.300 45.666
67066 97.261 -79.313 45.647
67067 97.265 -79.325 45.629
67068 97.270 -79.338 45.613
67069 97.274 -79.350 45.597
67070 97.278 -79.363 45.580
67071 97.282 -79.376 45.562
67072 97.286 -79.388 45.544
67073 97.290 -79.401 45.527
67074 97.293 -79.414 45.510
67075 97.286 -79.422 45.494
67076 97.291 -79.410 45.478
67077 97.297 -79.397 45.462
67078 97.303 -79.385 45.446
67079 97.308 -79.372 45.429
67080 97.314 -79.359 45.413
67081 97.317 -79.353 45.397
67082 97.318 -79.364 45.381
67083 97.325 -79.372 45.366
67084 97.343 -79.387 45.351
67085 97.362 -79.402 45.336
67086 97.381 -79.417 45.320
67087 97.400 -79.432 45.305
67088 97.420 -79.447 45.290
67089 97.439 -79.463 45.276
67090 97.458 -79.479 45.262
67091 97.477 -79.494 45.248
67092 97.497 -79.509 45.233
67093 97.516 -79.524 45.219
67094 97.535 -79.539 45.206
67095 97.554 -79.554 45.192
67096 97.573 -79.570 45.178
67097 97.592 -79.585 45.165
67098 97.611 -79.601 45.152
67099 97.631 -79.616 45.139
67100 97.650 -79.631 45.125
67101 97.669 -79.646 45.112
67102 97.688 -79.662 45.100
67103 97.707 -79.677 45.087
67104 97.727 -79.692 45.074
67105 97.746 -79.708 45.061
67106 97.765 -79.723 45.049
67107 97.784 -79.738 45.036
67108 97.803 -79.754 45.024
67109 97.822 -79.769 45.011
67110 97.842 -79.784 44.999
67111 97.861 -79.799 44.986
67112 97.880 -79.814 44.973
67113 97.899 -79.830 44.959
67114 97.918 -79.845 44.945
67115 97.937 -79.861 44.932
67116 97.956 -79.876 44.919
67117 97.975 -79.891 44.907
67118 97.995 -79.906 44.894
67119 98.014 -79.921 44.882
67120 98.033 -79.936 44.870
67121 98.053 -79.952 44.859
67122 98.072 -79.968 44.848
67123 98.091 -79.983 44.837
67124 98.110 -79.998 44.827
67125 98.129 -80.013 44.816
67126 98.148 -80.028 44.806
67127 98.167 -80.044 44.797
67128 98.186 -80.059 44.787
67129 98.206 -80.074 44.778
67130 98.225 -80.090 44.769
67131 98.244 -80.105 44.761
67132 98.263 -80.120 44.752
67133 98.283 -80.135 44.743
67134 98.302 -80.151 44.735
67135 98.321 -80.166 44.727
67136 98.340 -80.181 44.719
67137 98.359 -80.196 44.711
67138 98.378 -80.212 44.703
67139 98.397 -80.227 44.696
67140 98.417 -80.243 44.690
67141 98.436 -80.258 44.683
67142 98.455 -80.273 44.676
67143 98.474 -80.288 44.669
67144 98.493 -80.303 44.662
67145 98.512 -80.319 44.656
67146 98.535 -80.325 44.650
67147 98.560 -80.328 44.644
67148 98.585 -80.330 44.636
67149 98.610 -80.333 44.629
67150 98.635 -80.335 44.621
67151 98.659 -80.338 44.612
67152 98.684 -80.341 44.605
67153 98.709 -80.343 44.598
67154 98.734 -80.346 44.591
67155 98.759 -80.348 44.583
67156 98.783 -80.356 44.576
67157 98.806 -80.366 44.568
67158 98.830 -80.377 44.560
67159 98.853 -80.388 44.554
67160 98.876 -80.398 44.549
67161 98.900 -80.409 44.543
67162 98.923 -80.419 44.537
67163 98.946 -80.430 44.532
67164 98.970 -80.441 44.526
67165 98.993 -80.451 44.520
67166 99.016 -80.461 44.515
67167 99.039 -80.472 44.510
67168 99.063 -80.482 44.506
67496 93.434 -79.062 53.425
67497 93.444 -79.059 53.412
67498 93.453 -79.057 53.399
67499 93.463 -79.054 53.386
67500 93.474 -79.048 53.374
67501 93.485 -79.037 53.363
67502 93.496 -79.028 53.352
67503 93.504 -79.021 53.342
67504 93.512 -79.014 53.333
67505 93.521 -79.007 53.325
67506 93.529 -79.000 53.319
67507 93.538 -78.993 53.314
67508 93.546 -78.987 53.310
67509 93.554 -78.979 53.309
67510 93.562 -78.976 53.310
67511 93.567 -78.981 53.313
67512 93.574 -78.985 53.313
67513 93.593 -78.980 53.309
67514 93.612 -78.974 53.304
67515 93.629 -78.971 53.299
67516 93.643 -78.968 53.291
67517 93.656 -78.966 53.281
67518 93.671 -78.964 53.272
67519 93.685 -78.962 53.264
67520 93.701 -78.959 53.254
67521 93.718 -78.957 53.244
67522 93.735 -78.955 53.233
67523 93.745 -78.953 53.223
67524 93.755 -78.952 53.212
67525 93.763 -78.952 53.200
67526 93.769 -78.952 53.188
67527 93.776 -78.952 53.176
67528 93.782 -78.952 53.162
67529 93.789 -78.952 53.147
67530 93.796 -78.952 53.135
67531 93.803 -78.949 53.124
67532 93.812 -78.946 53.114
67533 93.820 -78.943 53.104
67534 93.828 -78.940 53.093
67535 93.837 -78.937 53.081
67536 93.846 -78.929 53.068
67537 93.857 -78.918 53.055
67538 93.859 -78.920 53.041
67539 93.860 -78.922 53.028
67540 93.862 -78.924 53.014
67541 93.863 -78.926 52.999
67542 93.865 -78.927 52.983
67543 93.867 -78.929 52.966
67544 93.868 -78.931 52.947
67545 93.870 -78.933 52.927
67546 93.866 -78.929 52.906
67547 93.863 -78.925 52.881
67548 93.860 -78.922 52.857
67549 93.857 -78.919 52.833
67550 93.855 -78.916 52.807
67551 93.872 -78.922 52.780
67552 93.889 -78.927 52.754
67553 93.906 -78.933 52.728
67554 93.923 -78.938 52.701
67555 93.940 -78.944 52.675
67556 93.957 -78.949 52.650
67557 93.974 -78.955 52.623
67558 93.991 -78.961 52.596
67559 94.008 -78.965 52.570
67560 94.014 -78.962 52.544
67561 94.024 -78.952 52.517
67562 94.033 -78.941 52.491
67563 94.042 -78.930 52.468
67564 94.051 -78.920 52.443
67565 94.060 -78.909 52.418
67566 94.069 -78.899 52.393
67567 94.078 -78.888 52.370
67568 94.087 -78.878 52.346
67569 94.096 -78.871 52.323
67570 94.103 -78.865 52.299
67571 94.110 -78.861 52.273
67572 94.117 -78.855 52.248
67573 94.125 -78.851 52.223
67574 94.139 -78.848 52.199
67575 94.155 -78.838 52.175
67576 94.172 -78.828 52.150
67577 94.189 -78.818 52.126
67578 94.205 -78.808 52.100
67579 94.221 -78.798 52.074
67580 94.238 -78.788 52.049
67581 94.254 -78.778 52.025
67582 94.271 -78.768 52.000
67583 94.282 -78.763 51.972
67584 94.292 -78.761 51.943
67585 94.308 -78.763 51.909
67586 94.323 -78.765 51.875
67587 94.339 -78.767 51.840
67588 94.354 -78.770 51.807
67589 94.369 -78.771 51.777
67590 94.385 -78.774 51.747
67591 94.400 -78.773 51.715
67592 94.416 -78.757 51.679
67593 94.433 -78.761 51.641
67594 94.451 -78.765 51.603
67595 94.468 -78.770 51.567
67596 94.486 -78.774 51.533
67597 94.503 -78.778 51.499
67598 94.521 -78.783 51.468
67599 94.538 -78.787 51.436
67600 94.555 -78.791 51.403
67601 94.573 -78.796 51.372
67602 94.593 -78.801 51.341
67603 94.620 -78.808 51.309
67604 94.646 -78.814 51.275
67605 94.661 -78.815 51.238
67606 94.676 -78.817 51.200
67607 94.691 -78.818 51.164
67608 94.706 -78.819 51.132
67609 94.718 -78.821 51.102
67610 94.716 -78.830 51.070
67611 94.712 -78.837 51.035
67612 94.690 -78.827 50.999
67613 94.685 -78.823 50.961
67614 94.689 -78.822 50.920
67615 94.693 -78.821 50.876
67616 94.697 -78.821 50.831
67617 94.701 -78.821 50.787
67618 94.705 -78.820 50.744
67619 94.709 -78.819 50.702
67620 94.713 -78.818 50.658
67621 94.729 -78.820 50.614
67622 94.753 -78.823 50.574
67623 94.777 -78.824 50.534
67624 94.808 -78.816 50.495
67625 94.839 -78.808 50.460
67626 94.866 -78.802 50.426
67627 94.885 -78.798 50.393
67628 94.891 -78.807 50.359
67629 94.890 -78.822 50.321
67630 94.887 -78.834 50.281
67631 94.879 -78.833 50.239
67632 94.871 -78.833 50.198
67633 94.863 -78.833 50.155
67634 94.855 -78.832 50.113
67635 94.847 -78.832 50.073
67636 94.839 -78.831 50.033
67637 94.831 -78.830 49.995
67638 94.823 -78.830 49.956
67639 94.835 -78.833 49.916
67640 94.852 -78.836 49.875
67641 94.868 -78.839 49.834
67642 94.882 -78.851 49.792
67643 94.890 -78.879 49.755
67644 94.897 -78.908 49.717
67645 94.905 -78.936 49.680
67646 94.913 -78.964 49.644
67647 94.944 -78.974 49.609
67648 94.976 -78.981 49.574
67649 94.992 -78.990 49.541
67650 95.008 -78.999 49.510
67651 95.025 -79.009 49.477
67652 95.041 -79.018 49.444
67653 95.057 -79.028 49.407
67654 95.076 -79.036 49.369
67655 95.106 -79.043 49.329
67656 95.135 -79.049 49.289
67657 95.162 -79.055 49.248
67658 95.187 -79.057 49.206
67659 95.211 -79.059 49.164
67660 95.236 -79.062 49.122
67661 95.260 -79.064 49.080
67662 95.285 -79.066 49.038
67663 95.309 -79.068 48.996
67664 95.334 -79.071 48.952
67665 95.358 -79.073 48.909
67666 95.375 -79.080 48.865
67667 95.378 -79.096 48.822
67668 95.380 -79.112 48.778
67669 95.382 -79.128 48.733
67670 95.385 -79.144 48.688
67671 95.392 -79.160 48.643
67672 95.405 -79.175 48.598
67673 95.418 -79.191 48.554
67674 95.431 -79.207 48.510
67675 95.444 -79.222 48.467
67676 95.457 -79.238 48.424
67677 95.470 -79.254 48.381
67678 95.500 -79.262 48.338
67679 95.528 -79.253 48.296
67680 95.550 -79.241 48.254
67681 95.572 -79.228 48.211
67682 95.595 -79.214 48.169
67683 95.618 -79.201 48.127
67684 95.640 -79.187 48.086
67685 95.663 -79.175 48.045
67686 95.680 -79.165 48.004
67687 95.691 -79.159 47.964
67688 95.701 -79.153 47.925
67689 95.711 -79.148 47.886
67690 95.736 -79.152 47.847
67691 95.769 -79.161 47.807
67692 95.803 -79.170 47.767
67693 95.837 -79.180 47.727
67694 95.871 -79.189 47.687
67695 95.899 -79.197 47.647
67696 95.912 -79.202 47.608
67697 95.908 -79.193 47.567
67698 95.903 -79.184 47.529
67699 95.899 -79.175 47.491
67700 95.895 -79.166 47.453
67701 95.893 -79.157 47.415
67702 95.905 -79.128 47.377
67703 95.892 -79.091 47.339
67704 95.901 -79.079 47.301
67705 95.952 -79.113 47.263
67706 96.003 -79.147 47.225
67707 96.054 -79.181 47.188
67708 96.105 -79.214 47.151
67709 96.151 -79.231 47.114
67710 96.179 -79.178 47.079
67711 96.207 -79.125 47.044
67712 96.235 -79.071 47.010
67713 96.263 -79.021 46.977
67714 96.297 -79.008 46.944
67715 96.327 -79.037 46.911
67716 96.356 -79.083 46.880
67717 96.384 -79.128 46.849
67718 96.412 -79.175 46.818
67719 96.441 -79.220 46.788
67720 96.469 -79.266 46.758
67721 96.496 -79.288 46.728
67722 96.523 -79.303 46.698
67723 96.546 -79.314 46.667
67724 96.555 -79.312 46.635
67725 96.564 -79.310 46.604
67726 96.573 -79.308 46.574
67727 96.582 -79.306 46.545
67728 96.594 -79.305 46.516
67729 96.616 -79.307 46.487
67730 96.640 -79.310 46.458
67731 96.662 -79.313 46.429
67732 96.685 -79.315 46.401
67733 96.708 -79.318 46.374
67734 96.731 -79.320 46.348
67735 96.753 -79.322 46.322
67736 96.776 -79.325 46.296
67737 96.798 -79.320 46.271
67738 96.816 -79.299 46.245
67739 96.835 -79.278 46.219
67740 96.854 -79.257 46.196
67741 96.872 -79.236 46.174
67742 96.891 -79.215 46.152
67743 96.909 -79.194 46.130
67744 96.939 -79.190 46.108
67745 96.976 -79.201 46.087
67746 97.013 -79.211 46.066
67747 97.050 -79.222 46.045
67748 97.076 -79.228 46.024
67749 97.074 -79.253 46.004
67750 97.088 -79.253 45.982
67751 97.103 -79.253 45.960
67752 97.117 -79.253 45.938
67753 97.133 -79.254 45.917
67754 97.147 -79.254 45.897
67755 97.162 -79.254 45.878
67756 97.175 -79.256 45.858
67757 97.151 -79.285 45.838
67758 97.165 -79.273 45.818
67759 97.167 -79.266 45.797
67760 97.168 -79.259 45.777
67761 97.170 -79.252 45.758
67762 97.193 -79.244 45.740
67763 97.228 -79.233 45.722
67764 97.235 -79.244 45.704
67765 97.239 -79.256 45.686
67766 97.243 -79.269 45.667
67767 97.247 -79.282 45.648
67768 97.251 -79.294 45.631
67769 97.255 -79.307 45.615
67770 97.259 -79.320 45.598
67771 97.263 -79.333 45.581
67772 97.267 -79.346 45.564
67773 97.271 -79.358 45.546
67774 97.275 -79.371 45.529
67775 97.279 -79.384 45.512
67776 97.283 -79.397 45.496
67777 97.287 -79.410 45.480
67778 97.290 -79.418 45.464
67779 97.296 -79.406 45.449
67780 97.302 -79.392 45.433
67781 97.309 -79.379 45.417
67782 97.317 -79.364 45.401
67783 97.325 -79.350 45.386
67784 97.329 -79.341 45.371
67785 97.349 -79.355 45.356
67786 97.368 -79.370 45.341
67787 97.387 -79.386 45.325
67788 97.406 -79.401 45.310
67789 97.425 -79.416 45.295
67790 97.444 -79.432 45.281
67791 97.463 -79.447 45.267
67792 97.483 -79.463 45.253
67793 97.502 -79.478 45.239
67794 97.521 -79.493 45.225
67795 97.540 -79.508 45.212
67796 97.560 -79.523 45.198
67797 97.579 -79.538 45.185
67798 97.598 -79.554 45.172
67799 97.617 -79.570 45.159
67800 97.636 -79.585 45.145
67801 97.655 -79.600 45.132
67802 97.674 -79.615 45.119
67803 97.693 -79.630 45.107
67804 97.713 -79.646 45.095
67805 97.732 -79.661 45.082
67806 97.751 -79.676 45.070
67807 97.770 -79.692 45.058
67808 97.789 -79.707 45.045
67809 97.809 -79.722 45.033
67810 97.828 -79.737 45.020
67811 97.847 -79.753 45.008
67812 97.866 -79.768 44.995
67813 97.885 -79.783 44.982
67814 97.904 -79.798 44.968
67815 97.924 -79.814 44.955
67816 97.943 -79.829 44.942
67817 97.962 -79.845 44.930
67818 97.981 -79.860 44.917
67819 98.000 -79.875 44.904
67820 98.019 -79.890 44.891
67821 98.038 -79.905 44.880
67822 98.058 -79.921 44.869
67823 98.077 -79.936 44.858
67824 98.096 -79.952 44.846
67825 98.115 -79.967 44.837
67826 98.135 -79.982 44.827
67827 98.154 -79.997 44.817
67828 98.173 -80.012 44.808
67829 98.192 -80.028 44.798
67830 98.211 -80.043 44.788
67831 98.230 -80.059 44.779
67832 98.249 -80.074 44.771
67833 98.269 -80.089 44.762
67834 98.288 -80.104 44.754
67835 98.307 -80.119 44.745
67836 98.326 -80.135 44.738
67837 98.345 -80.150 44.730
67838 98.365 -80.165 44.723
67839 98.384 -80.181 44.715
67840 98.403 -80.196 44.708
67841 98.422 -80.211 44.701
67842 98.441 -80.227 44.695
67843 98.460 -80.242 44.688
67844 98.480 -80.257 44.681
67845 98.499 -80.272 44.674
67846 98.518 -80.287 44.668
67847 98.541 -80.294 44.662
67848 98.566 -80.297 44.656
67849 98.590 -80.299 44.648
67850 98.616 -80.302 44.641
67851 98.640 -80.304 44.634
67852 98.665 -80.306 44.626
67853 98.690 -80.309 44.619
67854 98.714 -80.312 44.613
67855 98.739 -80.314 44.606
67856 98.762 -80.319 44.597
67857 98.787 -80.328 44.589
67858 98.810 -80.339 44.581
67859 98.834 -80.350 44.572
67860 98.857 -80.360 44.564
67861 98.880 -80.371 44.558
67862 98.904 -80.381 44.554
67863 98.927 -80.392 44.548
67864 98.950 -80.403 44.542
67865 98.974 -80.413 44.535
67866 98.997 -80.424 44.530
67867 99.020 -80.435 44.525
67868 99.043 -80.444 44.519
67869 99.067 -80.455 44.514
67870 99.090 -80.466 44.510
67871 99.113 -80.476 44.506
68198 93.444 -79.048 53.434
68199 93.453 -79.046 53.421
68200 93.464 -79.043 53.408
68201 93.473 -79.041 53.396
68202 93.484 -79.034 53.385
68203 93.495 -79.023 53.375
68204 93.505 -79.015 53.364
68205 93.513 -79.008 53.355
68206 93.522 -79.001 53.346
68207 93.530 -78.993 53.340
68208 93.539 -78.987 53.333
68209 93.547 -78.980 53.328
68210 93.555 -78.973 53.324
68211 93.564 -78.966 53.324
68212 93.572 -78.959 53.325
68213 93.577 -78.964 53.325
68214 93.590 -78.963 53.323
68215 93.609 -78.959 53.319
68216 93.624 -78.955 53.314
68217 93.638 -78.953 53.306
68218 93.652 -78.951 53.297
68219 93.669 -78.949 53.288
68220 93.685 -78.946 53.280
68221 93.702 -78.943 53.272
68222 93.719 -78.941 53.262
68223 93.731 -78.940 53.251
68224 93.741 -78.939 53.242
68225 93.751 -78.937 53.231
68226 93.759 -78.937 53.219
68227 93.765 -78.937 53.207
68228 93.772 -78.937 53.195
68229 93.779 -78.937 53.182
68230 93.785 -78.937 53.168
68231 93.792 -78.937 53.157
68232 93.798 -78.937 53.146
68233 93.805 -78.937 53.136
68234 93.813 -78.935 53.126
68235 93.821 -78.932 53.115
68236 93.831 -78.924 53.102
68237 93.842 -78.913 53.089
68238 93.851 -78.905 53.075
68239 93.853 -78.906 53.062
68240 93.854 -78.908 53.048
68241 93.856 -78.910 53.034
68242 93.857 -78.912 53.020
68243 93.859 -78.914 53.005
68244 93.861 -78.916 52.987
68245 93.861 -78.916 52.967
68246 93.857 -78.911 52.946
68247 93.852 -78.905 52.922
68248 93.848 -78.901 52.897
68249 93.845 -78.897 52.872
68250 93.842 -78.894 52.848
68251 93.847 -78.894 52.820
68252 93.864 -78.900 52.793
68253 93.881 -78.905 52.767
68254 93.898 -78.911 52.741
68255 93.915 -78.917 52.714
68256 93.932 -78.922 52.687
68257 93.948 -78.927 52.661
68258 93.968 -78.931 52.632
68259 93.988 -78.933 52.605
68260 94.008 -78.936 52.580
68261 94.024 -78.936 52.554
68262 94.034 -78.932 52.529
68263 94.044 -78.923 52.505
68264 94.053 -78.912 52.482
68265 94.062 -78.902 52.458
68266 94.071 -78.891 52.434
68267 94.080 -78.880 52.409
68268 94.089 -78.870 52.386
68269 94.099 -78.859 52.363
68270 94.108 -78.849 52.339
68271 94.115 -78.844 52.314
68272 94.122 -78.839 52.289
68273 94.129 -78.834 52.262
68274 94.140 -78.830 52.236
68275 94.154 -78.828 52.211
68276 94.170 -78.820 52.187
68277 94.186 -78.810 52.162
68278 94.202 -78.800 52.137
68279 94.219 -78.790 52.110
68280 94.236 -78.780 52.084
68281 94.252 -78.770 52.057
68282 94.268 -78.760 52.032
68283 94.281 -78.754 52.005
68284 94.290 -78.752 51.977
68285 94.298 -78.749 51.947
68286 94.313 -78.750 51.915
68287 94.328 -78.752 51.882
68288 94.343 -78.755 51.847
68289 94.359 -78.757 51.815
68290 94.374 -78.758 51.784
68291 94.390 -78.761 51.754
68292 94.403 -78.766 51.722
68293 94.418 -78.748 51.688
68294 94.436 -78.752 51.652
68295 94.454 -78.756 51.615
68296 94.471 -78.761 51.581
68297 94.488 -78.765 51.547
68298 94.506 -78.770 51.514
68299 94.523 -78.774 51.483
68300 94.541 -78.778 51.450
68301 94.558 -78.783 51.418
68302 94.577 -78.787 51.386
68303 94.604 -78.794 51.354
68304 94.631 -78.801 51.322
68305 94.659 -78.808 51.288
68306 94.682 -78.813 51.251
68307 94.696 -78.814 51.214
68308 94.711 -78.815 51.178
68309 94.726 -78.817 51.145
68310 94.733 -78.822 51.114
68311 94.730 -78.831 51.082
68312 94.728 -78.840 51.049
68313 94.725 -78.849 51.014
68314 94.704 -78.839 50.975
68315 94.683 -78.828 50.934
68316 94.673 -78.823 50.890
68317 94.677 -78.822 50.847
68318 94.681 -78.821 50.804
68319 94.685 -78.821 50.761
68320 94.689 -78.820 50.718
68321 94.699 -78.821 50.675
68322 94.722 -78.823 50.632
68323 94.751 -78.818 50.590
68324 94.782 -78.811 50.549
68325 94.813 -78.802 50.509
68326 94.844 -78.795 50.471
68327 94.864 -78.790 50.438
68328 94.883 -78.787 50.405
68329 94.889 -78.795 50.372
68330 94.888 -78.811 50.335
68331 94.887 -78.826 50.296
68332 94.884 -78.836 50.255
68333 94.876 -78.836 50.211
68334 94.867 -78.835 50.170
68335 94.859 -78.835 50.129
68336 94.852 -78.834 50.088
68337 94.844 -78.833 50.046
68338 94.836 -78.833 50.006
68339 94.828 -78.833 49.967
68340 94.836 -78.834 49.928
68341 94.852 -78.837 49.887
68342 94.870 -78.840 49.846
68343 94.886 -78.843 49.806
68344 94.900 -78.855 49.768
68345 94.908 -78.883 49.730
68346 94.916 -78.911 49.692
68347 94.923 -78.939 49.657
68348 94.946 -78.955 49.622
68349 94.978 -78.962 49.587
68350 94.995 -78.971 49.554
68351 95.011 -78.981 49.521
68352 95.028 -78.990 49.487
68353 95.044 -78.999 49.453
68354 95.063 -79.008 49.415
68355 95.093 -79.015 49.376
68356 95.122 -79.021 49.336
68357 95.151 -79.028 49.297
68358 95.181 -79.035 49.256
68359 95.205 -79.037 49.215
68360 95.230 -79.040 49.174
68361 95.255 -79.042 49.132
68362 95.279 -79.044 49.091
68363 95.304 -79.046 49.049
68364 95.328 -79.049 49.006
68365 95.353 -79.051 48.963
68366 95.376 -79.054 48.919
68367 95.385 -79.059 48.876
68368 95.390 -79.078 48.832
68369 95.394 -79.099 48.787
68370 95.397 -79.115 48.742
68371 95.399 -79.131 48.698
68372 95.402 -79.147 48.653
68373 95.406 -79.163 48.608
68374 95.419 -79.178 48.564
68375 95.432 -79.194 48.520
68376 95.445 -79.210 48.477
68377 95.458 -79.225 48.434
68378 95.471 -79.241 48.391
68379 95.497 -79.251 48.347
68380 95.525 -79.242 48.304
68381 95.547 -79.229 48.261
68382 95.569 -79.216 48.219
68383 95.592 -79.203 48.177
68384 95.615 -79.190 48.135
68385 95.637 -79.177 48.093
68386 95.660 -79.163 48.051
68387 95.682 -79.150 48.010
68388 95.693 -79.144 47.970
68389 95.703 -79.139 47.930
68390 95.713 -79.133 47.891
68391 95.743 -79.140 47.852
68392 95.776 -79.149 47.811
68393 95.810 -79.158 47.771
68394 95.843 -79.168 47.731
68395 95.871 -79.174 47.690
68396 95.883 -79.175 47.651
68397 95.895 -79.175 47.611
68398 95.912 -79.189 47.571
68399 95.912 -79.186 47.531
68400 95.908 -79.178 47.492
68401 95.903 -79.169 47.454
68402 95.899 -79.159 47.415
68403 95.933 -79.144 47.376
68404 95.919 -79.108 47.337
68405 95.918 -79.084 47.298
68406 95.969 -79.118 47.260
68407 96.020 -79.152 47.223
68408 96.071 -79.186 47.186
68409 96.121 -79.216 47.149
68410 96.149 -79.162 47.113
68411 96.176 -79.109 47.078
68412 96.204 -79.056 47.043
68413 96.237 -79.032 47.009
68414 96.271 -79.019 46.976
68415 96.305 -79.006 46.943
68416 96.335 -79.037 46.911
68417 96.364 -79.082 46.880
68418 96.392 -79.128 46.849
68419 96.420 -79.174 46.818
68420 96.449 -79.219 46.788
68421 96.475 -79.256 46.758
68422 96.489 -79.257 46.727
68423 96.511 -79.265 46.697
68424 96.537 -79.280 46.666
68425 96.552 -79.284 46.634
68426 96.561 -79.281 46.603
68427 96.570 -79.279 46.573
68428 96.579 -79.278 46.543
68429 96.601 -79.280 46.514
68430 96.624 -79.282 46.485
68431 96.647 -79.285 46.456
68432 96.670 -79.288 46.427
68433 96.693 -79.290 46.399
68434 96.716 -79.292 46.372
68435 96.738 -79.295 46.345
68436 96.761 -79.297 46.319
68437 96.784 -79.300 46.294
68438 96.806 -79.297 46.268
68439 96.824 -79.275 46.242
68440 96.843 -79.254 46.216
68441 96.862 -79.234 46.193
68442 96.880 -79.212 46.171
68443 96.899 -79.191 46.149
68444 96.923 -79.179 46.126
68445 96.960 -79.190 46.104
68446 96.997 -79.200 46.082
68447 97.034 -79.210 46.061
68448 97.051 -79.213 46.040
68449 97.079 -79.218 46.020
68450 97.080 -79.219 45.999
68451 97.079 -79.240 45.978
68452 97.094 -79.241 45.957
68453 97.109 -79.241 45.935
68454 97.123 -79.241 45.914
68455 97.139 -79.241 45.895
68456 97.153 -79.241 45.876
68457 97.168 -79.241 45.857
68458 97.150 -79.265 45.837
68459 97.166 -79.252 45.817
68460 97.167 -79.245 45.796
68461 97.169 -79.238 45.777
68462 97.171 -79.231 45.759
68463 97.172 -79.225 45.741
68464 97.196 -79.216 45.723
68465 97.220 -79.213 45.705
68466 97.224 -79.226 45.687
68467 97.228 -79.239 45.669
68468 97.232 -79.252 45.650
68469 97.236 -79.264 45.632
68470 97.240 -79.277 45.616
68471 97.244 -79.290 45.600
68472 97.248 -79.303 45.583
68473 97.252 -79.316 45.566
68474 97.256 -79.328 45.549
68475 97.260 -79.341 45.531
68476 97.264 -79.353 45.514
68477 97.277 -79.370 45.499
68478 97.291 -79.387 45.484
68479 97.293 -79.405 45.468
68480 97.321 -79.393 45.452
68481 97.346 -79.379 45.437
68482 97.354 -79.364 45.421
68483 97.360 -79.352 45.405
68484 97.366 -79.339 45.390
68485 97.372 -79.327 45.375
68486 97.374 -79.325 45.360
68487 97.361 -79.356 45.346
68488 97.395 -79.351 45.330
68489 97.415 -79.366 45.315
68490 97.434 -79.382 45.300
68491 97.453 -79.397 45.286
68492 97.472 -79.413 45.273
68493 97.491 -79.428 45.259
68494 97.510 -79.444 45.244
68495 97.529 -79.459 45.231
68496 97.548 -79.474 45.217
68497 97.567 -79.490 45.204
68498 97.586 -79.505 45.191
68499 97.605 -79.520 45.178
68500 97.625 -79.536 45.165
68501 97.644 -79.551 45.152
68502 97.662 -79.567 45.139
68503 97.681 -79.582 45.127
68504 97.701 -79.598 45.115
68505 97.720 -79.613 45.102
68506 97.739 -79.628 45.090
68507 97.758 -79.644 45.078
68508 97.777 -79.659 45.066
68509 97.796 -79.674 45.054
68510 97.815 -79.690 45.042
68511 97.834 -79.705 45.029
68512 97.853 -79.720 45.017
68513 97.872 -79.736 45.005
68514 97.891 -79.751 44.991
68515 97.911 -79.767 44.978
68516 97.930 -79.783 44.965
68517 97.948 -79.798 44.952
68518 97.967 -79.813 44.939
68519 97.987 -79.828 44.926
68520 98.006 -79.844 44.913
68521 98.025 -79.859 44.901
68522 98.044 -79.874 44.890
68523 98.063 -79.889 44.879
68524 98.082 -79.905 44.868
68525 98.101 -79.921 44.856
68526 98.121 -79.936 44.847
68527 98.140 -79.951 44.838
68528 98.159 -79.966 44.828
68529 98.178 -79.981 44.818
68530 98.198 -79.996 44.808
68531 98.217 -80.012 44.799
68532 98.236 -80.027 44.789
68533 98.255 -80.043 44.781
68534 98.274 -80.058 44.773
68535 98.293 -80.073 44.765
68536 98.312 -80.088 44.757
68537 98.332 -80.103 44.749
68538 98.351 -80.119 44.742
68539 98.370 -80.134 44.734
68540 98.389 -80.149 44.728
68541 98.408 -80.165 44.721
68542 98.427 -80.180 44.713
68543 98.447 -80.195 44.707
68544 98.466 -80.211 44.700
68545 98.485 -80.226 44.693
68546 98.504 -80.241 44.686
68547 98.523 -80.256 44.680
68548 98.547 -80.262 44.674
68549 98.571 -80.265 44.668
68550 98.596 -80.268 44.662
68551 98.621 -80.270 44.655
68552 98.644 -80.274 44.649
68553 98.666 -80.278 44.643
68554 98.688 -80.282 44.638
68555 98.709 -80.287 44.633
68556 98.730 -80.291 44.626
68557 98.752 -80.297 44.617
68558 98.773 -80.307 44.606
68559 98.814 -80.312 44.594
68560 98.838 -80.322 44.584
68561 98.861 -80.333 44.576
68562 98.884 -80.344 44.570
68563 98.908 -80.354 44.564
68564 98.931 -80.365 44.558
68565 98.954 -80.375 44.552
68566 98.977 -80.386 44.546
68567 99.001 -80.397 44.540
68568 99.024 -80.407 44.534
68569 99.047 -80.417 44.529
68570 99.071 -80.428 44.524
68571 99.094 -80.438 44.520
68572 99.117 -80.449 44.515
68573 99.141 -80.460 44.510
68574 99.164 -80.470 44.505
68575 99.187 -80.481 44.501
68900 93.454 -79.035 53.442
68901 93.464 -79.033 53.429
68902 93.473 -79.030 53.417
68903 93.483 -79.028 53.406
68904 93.494 -79.019 53.395
68905 93.505 -79.009 53.385
68906 93.514 -79.001 53.375
68907 93.523 -78.994 53.365
68908 93.532 -78.987 53.358
68909 93.540 -78.981 53.352
68910 93.548 -78.973 53.345
68911 93.556 -78.966 53.341
68912 93.565 -78.959 53.338
68913 93.573 -78.952 53.337
68914 93.581 -78.946 53.336
68915 93.588 -78.946 53.335
68916 93.606 -78.942 53.333
68917 93.620 -78.940 53.327
68918 93.635 -78.937 53.319
68919 93.652 -78.935 53.311
68920 93.669 -78.933 53.303
68921 93.686 -78.930 53.294
68922 93.703 -78.928 53.286
68923 93.718 -78.926 53.278
68924 93.727 -78.925 53.268
68925 93.737 -78.924 53.259
68926 93.747 -78.923 53.249
68927 93.755 -78.922 53.237
68928 93.762 -78.922 53.225
68929 93.768 -78.922 53.213
68930 93.775 -78.922 53.201
68931 93.781 -78.922 53.189
68932 93.788 -78.922 53.177
68933 93.795 -78.922 53.167
68934 93.801 -78.922 53.156
68935 93.808 -78.922 53.146
68936 93.815 -78.921 53.135
68937 93.827 -78.908 53.123
68938 93.838 -78.897 53.109
68939 93.845 -78.891 53.095
68940 93.846 -78.893 53.081
68941 93.848 -78.895 53.067
68942 93.850 -78.896 53.053
68943 93.851 -78.899 53.038
68944 93.853 -78.900 53.023
68945 93.852 -78.899 53.003
68946 93.848 -78.893 52.982
68947 93.843 -78.889 52.959
68948 93.839 -78.883 52.935
68949 93.834 -78.878 52.910
68950 93.830 -78.873 52.885
68951 93.827 -78.869 52.860
68952 93.839 -78.873 52.832
68953 93.856 -78.878 52.805
68954 93.873 -78.883 52.779
68955 93.893 -78.886 52.753
68956 93.913 -78.888 52.726
68957 93.934 -78.890 52.699
68958 93.954 -78.893 52.672
68959 93.974 -78.895 52.644
68960 93.995 -78.897 52.617
68961 94.015 -78.899 52.591
68962 94.033 -78.899 52.567
68963 94.043 -78.896 52.543
68964 94.053 -78.892 52.520
68965 94.063 -78.888 52.497
68966 94.073 -78.883 52.473
68967 94.082 -78.872 52.448
68968 94.091 -78.861 52.424
68969 94.100 -78.851 52.401
68970 94.109 -78.840 52.378
68971 94.118 -78.830 52.354
68972 94.127 -78.823 52.329
68973 94.134 -78.818 52.303
68974 94.141 -78.813 52.276
68975 94.154 -78.810 52.247
68976 94.169 -78.808 52.222
68977 94.184 -78.802 52.197
68978 94.200 -78.792 52.172
68979 94.216 -78.782 52.147
68980 94.233 -78.772 52.120
68981 94.249 -78.762 52.092
68982 94.266 -78.752 52.064
68983 94.280 -78.744 52.037
68984 94.288 -78.742 52.009
68985 94.297 -78.740 51.981
68986 94.306 -78.738 51.951
68987 94.317 -78.737 51.920
68988 94.333 -78.739 51.888
68989 94.348 -78.742 51.855
68990 94.364 -78.743 51.823
68991 94.379 -78.745 51.792
68992 94.393 -78.749 51.761
68993 94.406 -78.755 51.729
68994 94.421 -78.741 51.696
68995 94.438 -78.743 51.663
68996 94.456 -78.748 51.627
68997 94.473 -78.752 51.595
68998 94.491 -78.756 51.562
68999 94.508 -78.761 51.530
69000 94.526 -78.765 51.498
69001 94.543 -78.770 51.465
69002 94.561 -78.774 51.433
69003 94.588 -78.781 51.400
69004 94.615 -78.788 51.367
69005 94.642 -78.795 51.334
69006 94.670 -78.802 51.300
69007 94.697 -78.808 51.264
69008 94.717 -78.812 51.227
69009 94.732 -78.813 51.191
69010 94.747 -78.814 51.157
69011 94.747 -78.823 51.125
69012 94.744 -78.831 51.092
69013 94.742 -78.840 51.059
69014 94.740 -78.849 51.025
69015 94.737 -78.858 50.988
69016 94.718 -78.851 50.949
69017 94.697 -78.840 50.907
69018 94.675 -78.830 50.864
69019 94.661 -78.822 50.821
69020 94.665 -78.821 50.778
69021 94.669 -78.821 50.735
69022 94.694 -78.821 50.692
69023 94.725 -78.813 50.649
69024 94.756 -78.805 50.608
69025 94.787 -78.797 50.566
69026 94.818 -78.789 50.525
69027 94.844 -78.783 50.484
69028 94.862 -78.779 50.448
69029 94.881 -78.776 50.413
69030 94.888 -78.783 50.379
69031 94.887 -78.799 50.344
69032 94.886 -78.814 50.307
69033 94.884 -78.830 50.268
69034 94.880 -78.838 50.227
69035 94.872 -78.838 50.186
69036 94.864 -78.837 50.145
69037 94.856 -78.836 50.103
69038 94.848 -78.836 50.060
69039 94.840 -78.836 50.019
69040 94.833 -78.835 49.978
69041 94.837 -78.836 49.939
69042 94.853 -78.839 49.899
69043 94.870 -78.843 49.859
69044 94.887 -78.846 49.821
69045 94.904 -78.849 49.781
69046 94.918 -78.858 49.743
69047 94.926 -78.886 49.705
69048 94.934 -78.914 49.670
69049 94.949 -78.937 49.635
69050 94.981 -78.943 49.600
69051 94.997 -78.952 49.567
69052 95.014 -78.962 49.531
69053 95.030 -78.971 49.496
69054 95.050 -78.980 49.461
69055 95.079 -78.987 49.422
69056 95.109 -78.993 49.383
69057 95.138 -79.000 49.343
69058 95.168 -79.007 49.303
69059 95.197 -79.014 49.263
69060 95.225 -79.018 49.224
69061 95.249 -79.021 49.185
69062 95.273 -79.023 49.143
69063 95.298 -79.025 49.101
69064 95.323 -79.028 49.059
69065 95.347 -79.030 49.017
69066 95.372 -79.032 48.974
69067 95.385 -79.037 48.931
69068 95.394 -79.043 48.886
69069 95.400 -79.057 48.842
69070 95.404 -79.077 48.796
69071 95.409 -79.097 48.750
69072 95.414 -79.118 48.707
69073 95.416 -79.134 48.663
69074 95.419 -79.150 48.619
69075 95.421 -79.166 48.575
69076 95.433 -79.181 48.531
69077 95.446 -79.197 48.487
69078 95.459 -79.213 48.444
69079 95.472 -79.228 48.401
69080 95.495 -79.240 48.357
69081 95.522 -79.231 48.312
69082 95.544 -79.218 48.267
69083 95.566 -79.205 48.227
69084 95.589 -79.191 48.185
69085 95.612 -79.178 48.142
69086 95.634 -79.165 48.099
69087 95.657 -79.152 48.057
69088 95.679 -79.139 48.016
69089 95.696 -79.129 47.976
69090 95.706 -79.124 47.935
69091 95.716 -79.118 47.894
69092 95.749 -79.128 47.855
69093 95.783 -79.137 47.813
69094 95.817 -79.146 47.772
69095 95.842 -79.152 47.732
69096 95.854 -79.153 47.692
69097 95.867 -79.153 47.652
69098 95.879 -79.154 47.612
69099 95.891 -79.155 47.573
69100 95.909 -79.174 47.532
69101 95.917 -79.179 47.492
69102 95.912 -79.171 47.453
69103 95.908 -79.162 47.414
69104 95.919 -79.150 47.374
69105 95.947 -79.125 47.335
69106 95.935 -79.090 47.296
69107 95.986 -79.124 47.258
69108 96.037 -79.157 47.221
69109 96.088 -79.190 47.184
69110 96.118 -79.147 47.148
69111 96.146 -79.093 47.112
69112 96.176 -79.056 47.077
69113 96.210 -79.043 47.042
69114 96.245 -79.031 47.008
69115 96.279 -79.018 46.975
69116 96.313 -79.005 46.943
69117 96.343 -79.036 46.911
69118 96.372 -79.082 46.880
69119 96.400 -79.128 46.849
69120 96.429 -79.173 46.819
69121 96.457 -79.219 46.788
69122 96.477 -79.236 46.758
69123 96.491 -79.237 46.727
69124 96.506 -79.237 46.696
69125 96.525 -79.243 46.664
69126 96.549 -79.255 46.632
69127 96.558 -79.253 46.601
69128 96.568 -79.251 46.572
69129 96.586 -79.252 46.542
69130 96.609 -79.255 46.512
69131 96.632 -79.257 46.483
69132 96.655 -79.259 46.454
69133 96.678 -79.263 46.426
69134 96.700 -79.265 46.397
69135 96.723 -79.267 46.370
69136 96.746 -79.270 46.343
69137 96.769 -79.272 46.318
69138 96.792 -79.275 46.292
69139 96.814 -79.273 46.266
69140 96.832 -79.252 46.240
69141 96.851 -79.231 46.215
69142 96.870 -79.210 46.191
69143 96.888 -79.189 46.168
69144 96.907 -79.168 46.145
69145 96.944 -79.178 46.123
69146 96.981 -79.188 46.101
69147 97.018 -79.198 46.078
69148 97.021 -79.213 46.056
69149 97.056 -79.179 46.035
69150 97.081 -79.208 46.015
69151 97.077 -79.209 45.996
69152 97.085 -79.209 45.975
69153 97.085 -79.227 45.955
69154 97.100 -79.227 45.934
69155 97.115 -79.227 45.913
69156 97.130 -79.228 45.893
69157 97.145 -79.228 45.874
69158 97.159 -79.228 45.856
69159 97.149 -79.244 45.837
69160 97.166 -79.231 45.817
69161 97.168 -79.225 45.797
69162 97.169 -79.217 45.778
69163 97.171 -79.210 45.759
69164 97.173 -79.203 45.742
69165 97.174 -79.197 45.724
69166 97.199 -79.187 45.706
69167 97.210 -79.196 45.688
69168 97.214 -79.209 45.669
69169 97.218 -79.222 45.652
69170 97.221 -79.234 45.635
69171 97.225 -79.247 45.619
69172 97.229 -79.259 45.602
69173 97.233 -79.272 45.585
69174 97.237 -79.285 45.568
69175 97.251 -79.302 45.551
69176 97.265 -79.319 45.534
69177 97.280 -79.335 45.518
69178 97.294 -79.353 45.503
69179 97.308 -79.369 45.488
69180 97.297 -79.388 45.472
69181 97.318 -79.380 45.456
69182 97.353 -79.366 45.441
69183 97.374 -79.353 45.425
69184 97.387 -79.341 45.409
69185 97.400 -79.329 45.394
69186 97.412 -79.317 45.379
69187 97.418 -79.305 45.365
69188 97.418 -79.307 45.350
69189 97.404 -79.339 45.335
69190 97.391 -79.370 45.320
69191 97.433 -79.357 45.306
69192 97.461 -79.363 45.292
69193 97.480 -79.378 45.278
69194 97.500 -79.394 45.265
69195 97.519 -79.409 45.250
69196 97.537 -79.424 45.237
69197 97.557 -79.440 45.224
69198 97.576 -79.455 45.210
69199 97.595 -79.470 45.198
69200 97.614 -79.486 45.185
69201 97.633 -79.501 45.172
69202 97.652 -79.516 45.160
69203 97.671 -79.532 45.147
69204 97.690 -79.548 45.135
69205 97.709 -79.563 45.122
69206 97.728 -79.578 45.110
69207 97.747 -79.594 45.098
69208 97.767 -79.609 45.086
69209 97.786 -79.624 45.074
69210 97.805 -79.640 45.062
69211 97.823 -79.655 45.050
69212 97.843 -79.670 45.038
69213 97.862 -79.686 45.027
69214 97.881 -79.701 45.014
69215 97.900 -79.717 45.001
69216 97.919 -79.733 44.987
69217 97.938 -79.748 44.974
69218 97.957 -79.763 44.961
69219 97.976 -79.778 44.948
69220 97.995 -79.794 44.935
69221 98.014 -79.809 44.923
69222 98.033 -79.824 44.911
69223 98.053 -79.840 44.899
69224 98.072 -79.855 44.888
69225 98.090 -79.871 44.877
69226 98.109 -79.886 44.867
69227 98.129 -79.902 44.857
69228 98.148 -79.917 44.848
69229 98.167 -79.933 44.839
69230 98.186 -79.948 44.829
69231 98.205 -79.963 44.819
69232 98.224 -79.978 44.810
69233 98.243 -79.994 44.800
69234 98.262 -80.009 44.793
69235 98.281 -80.024 44.785
69236 98.300 -80.040 44.777
69237 98.319 -80.055 44.769
69238 98.339 -80.071 44.762
69239 98.357 -80.087 44.754
69240 98.376 -80.102 44.746
69241 98.395 -80.117 44.739
69242 98.415 -80.133 44.732
69243 98.434 -80.148 44.725
69244 98.453 -80.163 44.719
69245 98.472 -80.178 44.713
69246 98.491 -80.194 44.706
69247 98.508 -80.211 44.700
69248 98.526 -80.227 44.695
69249 98.550 -80.233 44.689
69250 98.570 -80.237 44.683
69251 98.591 -80.242 44.678
69252 98.613 -80.247 44.673
69253 98.635 -80.251 44.668
69254 98.656 -80.255 44.662
69255 98.678 -80.259 44.656
69256 98.700 -80.264 44.650
69257 98.721 -80.269 44.644
69258 98.742 -80.275 44.633
69259 98.762 -80.286 44.621
69260 98.786 -80.294 44.607
69261 98.836 -80.293 44.596
69262 98.865 -80.306 44.588
69263 98.888 -80.316 44.581
69264 98.912 -80.327 44.574
69265 98.935 -80.337 44.568
69266 98.958 -80.348 44.562
69267 98.981 -80.359 44.557
69268 99.005 -80.369 44.550
69269 99.028 -80.380 44.544
69270 99.051 -80.391 44.539
69271 99.075 -80.400 44.535
69272 99.098 -80.411 44.530
69273 99.121 -80.422 44.525
69274 99.144 -80.432 44.520
69275 99.168 -80.443 44.514
69276 99.191 -80.453 44.510
69277 99.214 -80.464 44.507
69278 99.238 -80.475 44.503
69602 93.464 -79.021 53.450
69603 93.474 -79.019 53.438
69604 93.483 -79.017 53.426
69605 93.493 -79.014 53.415
69606 93.504 -79.006 53.405
69607 93.515 -78.995 53.395
69608 93.524 -78.987 53.385
69609 93.533 -78.981 53.377
69610 93.541 -78.974 53.370
69611 93.549 -78.967 53.363
69612 93.557 -78.960 53.357
69613 93.566 -78.953 53.353
69614 93.574 -78.946 53.349
69615 93.582 -78.939 53.347
69616 93.591 -78.932 53.345
69617 93.602 -78.927 53.342
69618 93.619 -78.924 53.337
69619 93.636 -78.922 53.330
69620 93.653 -78.920 53.323
69621 93.670 -78.918 53.315
69622 93.687 -78.915 53.307
69623 93.704 -78.912 53.299
69624 93.713 -78.912 53.291
69625 93.723 -78.910 53.282
69626 93.733 -78.909 53.274
69627 93.743 -78.908 53.265
69628 93.751 -78.907 53.255
69629 93.758 -78.907 53.243
69630 93.764 -78.907 53.231
69631 93.771 -78.907 53.219
69632 93.778 -78.907 53.209
69633 93.784 -78.907 53.198
69634 93.791 -78.908 53.187
69635 93.798 -78.907 53.175
69636 93.805 -78.905 53.165
69637 93.812 -78.902 53.155
69638 93.822 -78.893 53.144
69639 93.833 -78.881 53.129
69640 93.839 -78.877 53.115
69641 93.840 -78.879 53.100
69642 93.842 -78.881 53.085
69643 93.844 -78.883 53.069
69644 93.845 -78.885 53.054
69645 93.843 -78.882 53.038
69646 93.839 -78.877 53.017
69647 93.834 -78.871 52.994
69648 93.830 -78.866 52.971
69649 93.825 -78.861 52.947
69650 93.821 -78.855 52.921
69651 93.817 -78.850 52.896
69652 93.821 -78.846 52.871
69653 93.839 -78.843 52.843
69654 93.859 -78.845 52.816
69655 93.880 -78.847 52.790
69656 93.900 -78.849 52.764
69657 93.920 -78.852 52.737
69658 93.940 -78.854 52.710
69659 93.961 -78.856 52.683
69660 93.981 -78.858 52.656
69661 94.002 -78.860 52.629
69662 94.022 -78.862 52.604
69663 94.041 -78.864 52.580
69664 94.051 -78.860 52.556
69665 94.061 -78.856 52.533
69666 94.071 -78.852 52.510
69667 94.081 -78.849 52.486
69668 94.091 -78.845 52.462
69669 94.102 -78.840 52.438
69670 94.111 -78.833 52.414
69671 94.120 -78.822 52.390
69672 94.129 -78.811 52.366
69673 94.138 -78.801 52.342
69674 94.145 -78.796 52.315
69675 94.155 -78.793 52.288
69676 94.170 -78.789 52.260
69677 94.183 -78.787 52.233
69678 94.197 -78.783 52.207
69679 94.214 -78.774 52.182
69680 94.230 -78.764 52.156
69681 94.247 -78.754 52.128
69682 94.263 -78.744 52.099
69683 94.278 -78.735 52.071
69684 94.287 -78.733 52.042
69685 94.296 -78.730 52.013
69686 94.305 -78.728 51.984
69687 94.313 -78.727 51.955
69688 94.322 -78.724 51.924
69689 94.338 -78.727 51.893
69690 94.353 -78.729 51.862
69691 94.368 -78.730 51.831
69692 94.384 -78.733 51.800
69693 94.397 -78.738 51.769
69694 94.410 -78.742 51.736
69695 94.424 -78.737 51.705
69696 94.441 -78.735 51.674
69697 94.459 -78.739 51.642
69698 94.476 -78.743 51.609
69699 94.493 -78.748 51.578
69700 94.511 -78.752 51.548
69701 94.529 -78.756 51.514
69702 94.546 -78.761 51.481
69703 94.572 -78.767 51.447
69704 94.600 -78.774 51.413
69705 94.627 -78.781 51.380
69706 94.654 -78.788 51.347
69707 94.681 -78.795 51.313
69708 94.708 -78.802 51.277
69709 94.736 -78.809 51.241
69710 94.752 -78.811 51.205
69711 94.763 -78.814 51.170
69712 94.761 -78.823 51.136
69713 94.759 -78.832 51.102
69714 94.756 -78.841 51.069
69715 94.754 -78.850 51.035
69716 94.752 -78.859 50.999
69717 94.748 -78.863 50.962
69718 94.728 -78.853 50.922
69719 94.706 -78.843 50.879
69720 94.684 -78.832 50.836
69721 94.662 -78.821 50.793
69722 94.668 -78.815 50.750
69723 94.699 -78.807 50.708
69724 94.731 -78.799 50.665
69725 94.762 -78.791 50.623
69726 94.792 -78.783 50.581
69727 94.824 -78.775 50.539
69728 94.843 -78.771 50.497
69729 94.861 -78.768 50.458
69730 94.879 -78.764 50.421
69731 94.887 -78.771 50.386
69732 94.885 -78.787 50.352
69733 94.884 -78.802 50.316
69734 94.883 -78.818 50.279
69735 94.882 -78.833 50.240
69736 94.877 -78.840 50.200
69737 94.869 -78.839 50.159
69738 94.861 -78.839 50.117
69739 94.853 -78.839 50.074
69740 94.845 -78.838 50.032
69741 94.837 -78.838 49.991
69742 94.837 -78.839 49.952
69743 94.854 -78.842 49.913
69744 94.871 -78.845 49.873
69745 94.888 -78.848 49.835
69746 94.904 -78.851 49.795
69747 94.921 -78.854 49.756
69748 94.936 -78.861 49.719
69749 94.945 -78.889 49.683
69750 94.952 -78.917 49.648
69751 94.983 -78.924 49.612
69752 95.000 -78.933 49.578
69753 95.016 -78.943 49.544
69754 95.037 -78.951 49.508
69755 95.066 -78.958 49.471
69756 95.096 -78.965 49.432
69757 95.125 -78.971 49.392
69758 95.155 -78.978 49.351
69759 95.184 -78.985 49.311
69760 95.213 -78.992 49.272
69761 95.243 -78.999 49.233
69762 95.268 -79.002 49.194
69763 95.293 -79.004 49.152
69764 95.317 -79.006 49.111
69765 95.341 -79.009 49.069
69766 95.366 -79.011 49.026
69767 95.386 -79.014 48.983
69768 95.394 -79.020 48.939
69769 95.403 -79.026 48.895
69770 95.410 -79.035 48.851
69771 95.415 -79.056 48.807
69772 95.419 -79.076 48.762
69773 95.424 -79.096 48.718
69774 95.428 -79.116 48.675
69775 95.433 -79.136 48.631
69776 95.436 -79.153 48.585
69777 95.439 -79.169 48.542
69778 95.447 -79.184 48.498
69779 95.460 -79.200 48.455
69780 95.473 -79.216 48.412
69781 95.492 -79.229 48.368
69782 95.519 -79.220 48.324
69783 95.541 -79.206 48.279
69784 95.563 -79.194 48.237
69785 95.586 -79.181 48.194
69786 95.609 -79.167 48.151
69787 95.631 -79.154 48.108
69788 95.654 -79.140 48.066
69789 95.676 -79.128 48.023
69790 95.698 -79.115 47.981
69791 95.709 -79.109 47.939
69792 95.723 -79.106 47.898
69793 95.756 -79.115 47.856
69794 95.790 -79.125 47.815
69795 95.813 -79.130 47.774
69796 95.826 -79.131 47.732
69797 95.838 -79.131 47.692
69798 95.850 -79.132 47.651
69799 95.862 -79.133 47.611
69800 95.875 -79.134 47.572
69801 95.888 -79.137 47.531
69802 95.907 -79.158 47.491
69803 95.921 -79.172 47.451
69804 95.917 -79.164 47.412
69805 95.913 -79.155 47.373
69806 95.960 -79.137 47.335
69807 95.961 -79.105 47.297
69808 96.003 -79.129 47.258
69809 96.054 -79.162 47.221
69810 96.088 -79.131 47.184
69811 96.116 -79.080 47.148
69812 96.150 -79.068 47.111
69813 96.184 -79.055 47.075
69814 96.219 -79.042 47.041
69815 96.253 -79.029 47.008
69816 96.287 -79.016 46.975
69817 96.321 -79.003 46.942
69818 96.352 -79.035 46.912
69819 96.380 -79.081 46.881
69820 96.408 -79.127 46.850
69821 96.437 -79.172 46.819
69822 96.464 -79.215 46.789
69823 96.479 -79.216 46.758
69824 96.493 -79.216 46.728
69825 96.508 -79.216 46.695
69826 96.523 -79.217 46.663
69827 96.540 -79.220 46.631
69828 96.555 -79.225 46.601
69829 96.572 -79.225 46.571
69830 96.594 -79.227 46.541
69831 96.617 -79.230 46.511
69832 96.640 -79.232 46.482
69833 96.663 -79.234 46.454
69834 96.685 -79.237 46.427
69835 96.708 -79.240 46.398
69836 96.731 -79.242 46.370
69837 96.754 -79.245 46.344
69838 96.777 -79.247 46.318
69839 96.799 -79.250 46.292
69840 96.822 -79.250 46.266
69841 96.840 -79.228 46.240
69842 96.859 -79.208 46.214
69843 96.878 -79.187 46.191
69844 96.896 -79.165 46.168
69845 96.928 -79.166 46.145
69846 96.965 -79.177 46.122
69847 96.999 -79.189 46.099
69848 96.991 -79.213 46.076
69849 97.026 -79.177 46.054
69850 97.064 -79.174 46.033
69851 97.084 -79.198 46.014
69852 97.080 -79.199 45.994
69853 97.081 -79.200 45.974
69854 97.090 -79.200 45.954
69855 97.099 -79.200 45.933
69856 97.106 -79.214 45.913
69857 97.121 -79.214 45.894
69858 97.136 -79.214 45.875
69859 97.151 -79.215 45.856
69860 97.148 -79.224 45.837
69861 97.167 -79.210 45.818
69862 97.168 -79.203 45.799
69863 97.170 -79.197 45.779
69864 97.172 -79.190 45.761
69865 97.173 -79.182 45.743
69866 97.175 -79.175 45.726
69867 97.177 -79.169 45.708
69868 97.194 -79.165 45.690
69869 97.198 -79.178 45.672
69870 97.203 -79.191 45.654
69871 97.207 -79.203 45.637
69872 97.212 -79.217 45.621
69873 97.226 -79.234 45.604
69874 97.240 -79.250 45.588
69875 97.254 -79.268 45.571
69876 97.268 -79.285 45.554
69877 97.283 -79.301 45.537
69878 97.297 -79.319 45.522
69879 97.311 -79.335 45.507
69880 97.316 -79.356 45.491
69881 97.300 -79.372 45.476
69882 97.316 -79.366 45.460
69883 97.350 -79.352 45.445
69884 97.373 -79.339 45.429
69885 97.386 -79.327 45.413
69886 97.399 -79.316 45.398
69887 97.412 -79.303 45.384
69888 97.425 -79.292 45.370
69889 97.447 -79.312 45.355
69890 97.475 -79.402 45.340
69891 97.461 -79.400 45.325
69892 97.443 -79.396 45.312
69893 97.424 -79.392 45.297
69894 97.461 -79.374 45.284
69895 97.508 -79.359 45.270
69896 97.527 -79.374 45.257
69897 97.546 -79.390 45.244
69898 97.565 -79.405 45.230
69899 97.584 -79.420 45.217
69900 97.603 -79.436 45.204
69901 97.623 -79.451 45.192
69902 97.642 -79.466 45.180
69903 97.661 -79.482 45.167
69904 97.680 -79.498 45.155
69905 97.699 -79.513 45.142
69906 97.718 -79.529 45.130
69907 97.737 -79.544 45.118
69908 97.756 -79.559 45.106
69909 97.775 -79.574 45.094
69910 97.794 -79.590 45.082
69911 97.813 -79.605 45.071
69912 97.832 -79.620 45.059
69913 97.851 -79.636 45.047
69914 97.870 -79.651 45.035
69915 97.889 -79.667 45.022
69916 97.909 -79.683 45.009
69917 97.928 -79.698 44.996
69918 97.947 -79.713 44.983
69919 97.965 -79.729 44.970
69920 97.985 -79.744 44.957
69921 98.004 -79.759 44.945
69922 98.023 -79.774 44.933
69923 98.042 -79.790 44.921
69924 98.061 -79.805 44.909
69925 98.080 -79.820 44.899
69926 98.098 -79.837 44.888
69927 98.114 -79.855 44.878
69928 98.132 -79.871 44.869
69929 98.149 -79.889 44.859
69930 98.166 -79.905 44.849
69931 98.182 -79.923 44.840
69932 98.200 -79.940 44.831
69933 98.217 -79.957 44.822
69934 98.233 -79.974 44.813
69935 98.250 -79.991 44.805
69936 98.268 -80.008 44.797
69937 98.285 -80.025 44.789
69938 98.302 -80.042 44.781
69939 98.319 -80.058 44.774
69940 98.336 -80.074 44.767
69941 98.354 -80.090 44.758
69942 98.371 -80.106 44.751
69943 98.388 -80.121 44.744
69944 98.406 -80.137 44.738
69945 98.423 -80.153 44.732
69946 98.446 -80.168 44.726
69947 98.469 -80.182 44.719
69948 98.503 -80.179 44.713
69949 98.530 -80.194 44.709
69950 98.557 -80.206 44.704
69951 98.577 -80.211 44.698
69952 98.598 -80.215 44.693
69953 98.620 -80.220 44.689
69954 98.640 -80.225 44.685
69955 98.661 -80.229 44.679
69956 98.683 -80.234 44.673
69957 98.704 -80.237 44.666
69958 98.728 -80.240 44.659
69959 98.754 -80.249 44.648
69960 98.779 -80.260 44.634
69961 98.790 -80.269 44.620
69962 98.797 -80.275 44.608
69963 98.845 -80.275 44.600
69964 98.892 -80.289 44.592
69965 98.915 -80.300 44.585
69966 98.939 -80.310 44.579
69967 98.962 -80.321 44.572
69968 98.985 -80.331 44.566
69969 99.009 -80.342 44.559
69970 99.032 -80.353 44.554
69971 99.055 -80.363 44.549
69972 99.079 -80.374 44.544
69973 99.102 -80.384 44.539
69974 99.125 -80.394 44.534
69975 99.148 -80.405 44.529
69976 99.172 -80.416 44.524
69977 99.195 -80.426 44.520
69978 99.218 -80.437 44.516
69979 99.242 -80.447 44.512
69980 99.265 -80.458 44.509
69981 99.288 -80.469 44.505
69982 99.311 -80.479 44.501
70304 93.474 -79.009 53.458
70305 93.484 -79.006 53.446
70306 93.493 -79.003 53.435
70307 93.503 -79.001 53.425
70308 93.514 -78.991 53.415
70309 93.525 -78.981 53.406
70310 93.534 -78.974 53.397
70311 93.542 -78.967 53.388
70312 93.550 -78.960 53.380
70313 93.559 -78.953 53.374
70314 93.567 -78.946 53.368
70315 93.575 -78.940 53.361
70316 93.583 -78.933 53.358
70317 93.592 -78.925 53.355
70318 93.609 -78.915 53.351
70319 93.620 -78.908 53.346
70320 93.637 -78.906 53.340
70321 93.654 -78.904 53.333
70322 93.671 -78.902 53.326
70323 93.688 -78.899 53.318
70324 93.700 -78.898 53.311
70325 93.709 -78.896 53.304
70326 93.719 -78.896 53.296
70327 93.729 -78.894 53.289
70328 93.739 -78.893 53.280
70329 93.747 -78.893 53.271
70330 93.754 -78.893 53.260
70331 93.761 -78.893 53.249
70332 93.767 -78.893 53.239
70333 93.774 -78.893 53.227
70334 93.781 -78.893 53.216
70335 93.787 -78.890 53.206
70336 93.794 -78.888 53.195
70337 93.801 -78.886 53.184
70338 93.808 -78.883 53.173
70339 93.818 -78.877 53.161
70340 93.829 -78.865 53.147
70341 93.833 -78.864 53.133
70342 93.834 -78.866 53.118
70343 93.836 -78.868 53.101
70344 93.838 -78.869 53.084
70345 93.834 -78.865 53.068
70346 93.830 -78.860 53.050
70347 93.825 -78.855 53.028
70348 93.821 -78.849 53.005
70349 93.816 -78.844 52.981
70350 93.817 -78.839 52.955
70351 93.822 -78.834 52.930
70352 93.828 -78.830 52.906
70353 93.837 -78.825 52.882
70354 93.852 -78.819 52.854
70355 93.868 -78.812 52.826
70356 93.887 -78.811 52.800
70357 93.907 -78.813 52.774
70358 93.927 -78.814 52.747
70359 93.947 -78.817 52.721
70360 93.968 -78.819 52.694
70361 93.988 -78.821 52.667
70362 94.009 -78.824 52.642
70363 94.029 -78.826 52.618
70364 94.049 -78.828 52.593
70365 94.059 -78.824 52.569
70366 94.069 -78.821 52.546
70367 94.079 -78.816 52.523
70368 94.090 -78.812 52.500
70369 94.100 -78.808 52.475
70370 94.110 -78.805 52.451
70371 94.120 -78.801 52.426
70372 94.130 -78.797 52.401
70373 94.140 -78.793 52.378
70374 94.149 -78.782 52.354
70375 94.157 -78.775 52.327
70376 94.171 -78.772 52.300
70377 94.184 -78.769 52.273
70378 94.198 -78.767 52.245
70379 94.212 -78.764 52.219
70380 94.228 -78.755 52.194
70381 94.244 -78.745 52.168
70382 94.261 -78.736 52.139
70383 94.277 -78.726 52.110
70384 94.286 -78.724 52.081
70385 94.294 -78.721 52.052
70386 94.303 -78.719 52.023
70387 94.312 -78.717 51.994
70388 94.320 -78.714 51.965
70389 94.329 -78.712 51.935
70390 94.343 -78.714 51.904
70391 94.358 -78.715 51.873
70392 94.373 -78.717 51.842
70393 94.387 -78.722 51.812
70394 94.400 -78.727 51.781
70395 94.414 -78.731 51.750
70396 94.427 -78.733 51.718
70397 94.443 -78.726 51.687
70398 94.461 -78.730 51.656
70399 94.479 -78.735 51.625
70400 94.496 -78.739 51.595
70401 94.513 -78.743 51.565
70402 94.531 -78.748 51.533
70403 94.557 -78.755 51.498
70404 94.584 -78.761 51.462
70405 94.611 -78.768 51.427
70406 94.638 -78.775 51.393
70407 94.666 -78.782 51.359
70408 94.693 -78.789 51.325
70409 94.720 -78.796 51.290
70410 94.747 -78.802 51.255
70411 94.773 -78.809 51.219
70412 94.777 -78.814 51.183
70413 94.775 -78.824 51.148
70414 94.771 -78.826 51.113
70415 94.768 -78.827 51.079
70416 94.764 -78.828 51.044
70417 94.760 -78.830 51.009
70418 94.756 -78.830 50.973
70419 94.737 -78.821 50.934
70420 94.715 -78.811 50.893
70421 94.693 -78.800 50.851
70422 94.676 -78.801 50.807
70423 94.679 -78.803 50.765
70424 94.705 -78.793 50.723
70425 94.736 -78.785 50.681
70426 94.767 -78.777 50.637
70427 94.798 -78.769 50.594
70428 94.823 -78.764 50.551
70429 94.841 -78.760 50.510
70430 94.859 -78.757 50.470
70431 94.877 -78.754 50.431
70432 94.885 -78.760 50.394
70433 94.884 -78.775 50.360
70434 94.883 -78.791 50.324
70435 94.882 -78.806 50.287
70436 94.881 -78.822 50.251
70437 94.879 -78.837 50.211
70438 94.874 -78.843 50.171
70439 94.866 -78.842 50.130
70440 94.858 -78.841 50.088
70441 94.850 -78.841 50.047
70442 94.842 -78.840 50.006
70443 94.838 -78.840 49.967
70444 94.855 -78.843 49.928
70445 94.872 -78.846 49.888
70446 94.889 -78.849 49.849
70447 94.905 -78.852 49.808
70448 94.922 -78.855 49.769
70449 94.939 -78.858 49.732
70450 94.955 -78.864 49.696
70451 94.963 -78.892 49.660
70452 94.986 -78.905 49.625
70453 95.002 -78.915 49.590
70454 95.024 -78.923 49.555
70455 95.054 -78.930 49.518
70456 95.083 -78.937 49.481
70457 95.112 -78.943 49.441
70458 95.142 -78.950 49.400
70459 95.171 -78.957 49.359
70460 95.200 -78.964 49.318
70461 95.230 -78.971 49.280
70462 95.259 -78.977 49.242
70463 95.287 -78.983 49.202
70464 95.311 -78.985 49.160
70465 95.336 -78.987 49.120
70466 95.360 -78.990 49.079
70467 95.385 -78.992 49.036
70468 95.395 -78.997 48.992
70469 95.403 -79.003 48.949
70470 95.412 -79.009 48.905
70471 95.420 -79.015 48.861
70472 95.425 -79.034 48.817
70473 95.429 -79.054 48.773
70474 95.434 -79.075 48.729
70475 95.439 -79.094 48.686
70476 95.444 -79.115 48.642
70477 95.448 -79.134 48.596
70478 95.453 -79.155 48.553
70479 95.456 -79.172 48.510
70480 95.461 -79.187 48.467
70481 95.474 -79.203 48.423
70482 95.489 -79.218 48.379
70483 95.516 -79.209 48.335
70484 95.538 -79.195 48.291
70485 95.560 -79.182 48.247
70486 95.583 -79.169 48.203
70487 95.606 -79.156 48.161
70488 95.628 -79.143 48.117
70489 95.651 -79.130 48.074
70490 95.673 -79.116 48.031
70491 95.696 -79.103 47.988
70492 95.711 -79.094 47.944
70493 95.729 -79.093 47.901
70494 95.763 -79.103 47.859
70495 95.784 -79.108 47.817
70496 95.797 -79.109 47.776
70497 95.809 -79.109 47.734
70498 95.821 -79.110 47.693
70499 95.834 -79.111 47.652
70500 95.846 -79.112 47.611
70501 95.858 -79.112 47.571
70502 95.871 -79.113 47.532
70503 95.886 -79.122 47.491
70504 95.905 -79.143 47.451
70505 95.925 -79.165 47.412
70506 95.921 -79.157 47.373
70507 95.945 -79.143 47.335
70508 95.989 -79.122 47.297
70509 96.020 -79.134 47.259
70510 96.057 -79.115 47.221
70511 96.090 -79.091 47.184
70512 96.124 -79.079 47.148
70513 96.158 -79.066 47.111
70514 96.192 -79.053 47.076
70515 96.227 -79.040 47.042
70516 96.261 -79.028 47.009
70517 96.295 -79.015 46.976
70518 96.329 -79.002 46.943
70519 96.359 -79.035 46.913
70520 96.388 -79.081 46.882
70521 96.416 -79.126 46.851
70522 96.445 -79.172 46.819
70523 96.466 -79.194 46.788
70524 96.480 -79.195 46.757
70525 96.496 -79.195 46.726
70526 96.510 -79.196 46.694
70527 96.525 -79.196 46.662
70528 96.540 -79.197 46.631
70529 96.554 -79.197 46.602
70530 96.579 -79.200 46.571
70531 96.602 -79.202 46.541
70532 96.624 -79.204 46.512
70533 96.648 -79.207 46.483
70534 96.670 -79.209 46.455
70535 96.693 -79.212 46.427
70536 96.716 -79.215 46.399
70537 96.739 -79.217 46.372
70538 96.761 -79.219 46.345
70539 96.784 -79.222 46.318
70540 96.807 -79.225 46.293
70541 96.830 -79.226 46.267
70542 96.849 -79.205 46.241
70543 96.867 -79.184 46.215
70544 96.886 -79.163 46.191
70545 96.912 -79.155 46.168
70546 96.949 -79.165 46.145
70547 96.972 -79.184 46.122
70548 96.960 -79.212 46.098
70549 96.996 -79.177 46.075
70550 97.031 -79.140 46.053
70551 97.072 -79.169 46.032
70552 97.087 -79.188 46.013
70553 97.083 -79.189 45.993
70554 97.078 -79.190 45.974
70555 97.086 -79.190 45.954
70556 97.095 -79.191 45.933
70557 97.104 -79.191 45.913
70558 97.112 -79.201 45.895
70559 97.127 -79.201 45.877
70560 97.142 -79.201 45.857
70561 97.147 -79.203 45.838
70562 97.167 -79.190 45.820
70563 97.169 -79.182 45.801
70564 97.171 -79.175 45.781
70565 97.172 -79.169 45.763
70566 97.174 -79.162 45.745
70567 97.175 -79.155 45.728
70568 97.177 -79.147 45.710
70569 97.179 -79.140 45.693
70570 97.186 -79.149 45.675
70571 97.201 -79.165 45.657
70572 97.215 -79.183 45.640
70573 97.229 -79.200 45.624
70574 97.243 -79.216 45.607
70575 97.257 -79.234 45.591
70576 97.272 -79.250 45.574
70577 97.286 -79.267 45.557
70578 97.300 -79.284 45.541
70579 97.314 -79.301 45.526
70580 97.328 -79.318 45.511
70581 97.306 -79.350 45.495
70582 97.304 -79.357 45.479
70583 97.319 -79.350 45.464
70584 97.347 -79.338 45.449
70585 97.373 -79.325 45.433
70586 97.385 -79.313 45.418
70587 97.398 -79.302 45.404
70588 97.412 -79.290 45.389
70589 97.428 -79.285 45.374
70590 97.458 -79.342 45.360
70591 97.475 -79.388 45.345
70592 97.482 -79.397 45.331
70593 97.478 -79.400 45.317
70594 97.460 -79.397 45.303
70595 97.471 -79.390 45.290
70596 97.496 -79.382 45.276
70597 97.520 -79.375 45.263
70598 97.543 -79.368 45.250
70599 97.574 -79.370 45.237
70600 97.593 -79.386 45.224
70601 97.612 -79.401 45.212
70602 97.631 -79.416 45.199
70603 97.650 -79.432 45.187
70604 97.669 -79.447 45.174
70605 97.687 -79.463 45.162
70606 97.704 -79.481 45.150
70607 97.721 -79.498 45.138
70608 97.738 -79.515 45.126
70609 97.755 -79.532 45.114
70610 97.772 -79.549 45.103
70611 97.789 -79.567 45.091
70612 97.806 -79.583 45.080
70613 97.823 -79.601 45.068
70614 97.840 -79.617 45.056
70615 97.857 -79.635 45.044
70616 97.874 -79.652 45.031
70617 97.891 -79.669 45.018
70618 97.908 -79.686 45.006
70619 97.925 -79.703 44.992
70620 97.942 -79.720 44.980
70621 97.959 -79.737 44.967
70622 97.976 -79.755 44.955
70623 97.993 -79.771 44.942
70624 98.010 -79.789 44.931
70625 98.027 -79.806 44.919
70626 98.043 -79.823 44.909
70627 98.061 -79.840 44.899
70628 98.078 -79.856 44.889
70629 98.095 -79.872 44.879
70630 98.112 -79.888 44.870
70631 98.130 -79.905 44.860
70632 98.147 -79.921 44.852
70633 98.164 -79.936 44.843
70634 98.181 -79.952 44.834
70635 98.199 -79.968 44.825
70636 98.216 -79.984 44.817
70637 98.233 -80.000 44.809
70638 98.251 -80.016 44.801
70639 98.268 -80.032 44.794
70640 98.285 -80.048 44.787
70641 98.303 -80.064 44.780
70642 98.320 -80.080 44.772
70643 98.342 -80.095 44.764
70644 98.366 -80.109 44.758
70645 98.389 -80.123 44.751
70646 98.414 -80.137 44.745
70647 98.449 -80.125 44.739
70648 98.480 -80.133 44.733
70649 98.507 -80.148 44.726
70650 98.534 -80.162 44.722
70651 98.562 -80.178 44.717
70652 98.585 -80.184 44.711
70653 98.605 -80.189 44.707
70654 98.627 -80.193 44.702
70655 98.648 -80.198 44.697
70656 98.672 -80.201 44.692
70657 98.696 -80.203 44.688
70658 98.720 -80.206 44.682
70659 98.745 -80.213 44.674
70660 98.772 -80.224 44.659
70661 98.797 -80.234 44.645
70662 98.803 -80.240 44.632
70663 98.805 -80.243 44.619
70664 98.821 -80.253 44.611
70665 98.858 -80.262 44.603
70666 98.917 -80.271 44.595
70667 98.943 -80.283 44.589
70668 98.966 -80.294 44.582
70669 98.989 -80.304 44.575
70670 99.013 -80.315 44.569
70671 99.036 -80.325 44.564
70672 99.059 -80.336 44.559
70673 99.082 -80.347 44.554
70674 99.106 -80.356 44.549
70675 99.129 -80.367 44.544
70676 99.152 -80.378 44.538
70677 99.176 -80.388 44.533
70678 99.199 -80.399 44.529
70679 99.222 -80.410 44.525
70680 99.246 -80.420 44.521
70681 99.269 -80.431 44.518
70682 99.292 -80.441 44.514
70683 99.315 -80.452 44.510
70684 99.342 -80.461 44.507
70685 99.381 -80.465 44.503
71006 93.484 -78.995 53.465
71007 93.493 -78.993 53.454
71008 93.503 -78.990 53.444
71009 93.513 -78.987 53.434
71010 93.524 -78.977 53.424
71011 93.535 -78.968 53.415
71012 93.543 -78.961 53.406
71013 93.551 -78.954 53.397
71014 93.560 -78.946 53.390
71015 93.568 -78.940 53.383
71016 93.576 -78.933 53.377
71017 93.585 -78.926 53.371
71018 93.593 -78.919 53.365
71019 93.602 -78.912 53.359
71020 93.622 -78.893 53.353
71021 93.638 -78.891 53.348
71022 93.654 -78.889 53.343
71023 93.672 -78.886 53.338
71024 93.686 -78.884 53.329
71025 93.696 -78.883 53.322
71026 93.705 -78.882 53.316
71027 93.715 -78.880 53.311
71028 93.725 -78.880 53.304
71029 93.734 -78.878 53.296
71030 93.744 -78.877 53.287
71031 93.750 -78.877 53.278
71032 93.757 -78.877 53.268
71033 93.764 -78.877 53.258
71034 93.770 -78.877 53.246
71035 93.777 -78.874 53.235
71036 93.784 -78.872 53.224
71037 93.791 -78.870 53.213
71038 93.798 -78.868 53.202
71039 93.805 -78.865 53.190
71040 93.813 -78.861 53.178
71041 93.825 -78.849 53.163
71042 93.827 -78.850 53.148
71043 93.828 -78.852 53.133
71044 93.829 -78.853 53.116
71045 93.825 -78.848 53.098
71046 93.821 -78.843 53.080
71047 93.816 -78.837 53.061
71048 93.814 -78.832 53.038
71049 93.819 -78.827 53.014
71050 93.824 -78.824 52.990
71051 93.830 -78.819 52.964
71052 93.835 -78.814 52.939
71053 93.840 -78.810 52.914
71054 93.852 -78.805 52.889
71055 93.868 -78.798 52.861
71056 93.884 -78.792 52.834
71057 93.900 -78.786 52.808
71058 93.916 -78.780 52.783
71059 93.934 -78.778 52.756
71060 93.955 -78.780 52.730
71061 93.975 -78.783 52.704
71062 93.995 -78.785 52.677
71063 94.015 -78.787 52.653
71064 94.036 -78.789 52.630
71065 94.056 -78.792 52.606
71066 94.068 -78.788 52.583
71067 94.078 -78.784 52.560
71068 94.088 -78.780 52.537
71069 94.098 -78.777 52.513
71070 94.108 -78.773 52.489
71071 94.118 -78.769 52.463
71072 94.128 -78.765 52.437
71073 94.138 -78.761 52.413
71074 94.149 -78.758 52.389
71075 94.159 -78.753 52.366
71076 94.173 -78.754 52.339
71077 94.185 -78.752 52.312
71078 94.199 -78.749 52.286
71079 94.213 -78.746 52.258
71080 94.226 -78.744 52.233
71081 94.242 -78.737 52.207
71082 94.258 -78.727 52.181
71083 94.275 -78.717 52.152
71084 94.284 -78.714 52.123
71085 94.293 -78.712 52.094
71086 94.301 -78.710 52.064
71087 94.310 -78.708 52.035
71088 94.319 -78.705 52.007
71089 94.328 -78.703 51.979
71090 94.336 -78.701 51.949
71091 94.347 -78.701 51.918
71092 94.363 -78.702 51.888
71093 94.378 -78.705 51.857
71094 94.391 -78.710 51.826
71095 94.404 -78.715 51.795
71096 94.417 -78.720 51.764
71097 94.430 -78.724 51.733
71098 94.447 -78.717 51.702
71099 94.464 -78.721 51.671
71100 94.481 -78.726 51.641
71101 94.499 -78.730 51.611
71102 94.516 -78.735 51.581
71103 94.541 -78.741 51.549
71104 94.568 -78.748 51.514
71105 94.596 -78.755 51.477
71106 94.623 -78.761 51.441
71107 94.650 -78.768 51.406
71108 94.677 -78.775 51.372
71109 94.704 -78.783 51.338
71110 94.732 -78.789 51.304
71111 94.755 -78.795 51.270
71112 94.773 -78.796 51.234
71113 94.784 -78.795 51.198
71114 94.783 -78.792 51.162
71115 94.779 -78.793 51.126
71116 94.775 -78.794 51.091
71117 94.771 -78.796 51.055
71118 94.768 -78.796 51.020
71119 94.764 -78.798 50.984
71120 94.747 -78.789 50.946
71121 94.724 -78.779 50.906
71122 94.705 -78.773 50.865
71123 94.693 -78.790 50.822
71124 94.695 -78.793 50.781
71125 94.714 -78.781 50.739
71126 94.741 -78.771 50.696
71127 94.772 -78.763 50.653
71128 94.803 -78.755 50.610
71129 94.821 -78.752 50.567
71130 94.839 -78.749 50.527
71131 94.857 -78.745 50.485
71132 94.875 -78.742 50.444
71133 94.884 -78.749 50.404
71134 94.883 -78.764 50.367
71135 94.882 -78.779 50.330
71136 94.880 -78.795 50.293
71137 94.879 -78.810 50.256
71138 94.878 -78.826 50.217
71139 94.877 -78.841 50.179
71140 94.871 -78.845 50.141
71141 94.862 -78.844 50.102
71142 94.854 -78.843 50.062
71143 94.847 -78.843 50.023
71144 94.839 -78.843 49.984
71145 94.856 -78.846 49.944
71146 94.873 -78.849 49.903
71147 94.890 -78.852 49.863
71148 94.906 -78.855 49.821
71149 94.923 -78.858 49.781
71150 94.939 -78.861 49.744
71151 94.957 -78.864 49.708
71152 94.973 -78.868 49.674
71153 94.989 -78.886 49.639
71154 95.012 -78.895 49.603
71155 95.041 -78.902 49.565
71156 95.070 -78.908 49.527
71157 95.100 -78.915 49.490
71158 95.129 -78.922 49.449
71159 95.158 -78.929 49.408
71160 95.188 -78.936 49.366
71161 95.217 -78.943 49.325
71162 95.246 -78.949 49.288
71163 95.276 -78.956 49.250
71164 95.305 -78.963 49.210
71165 95.330 -78.965 49.168
71166 95.355 -78.968 49.127
71167 95.379 -78.970 49.086
71168 95.395 -78.974 49.044
71169 95.404 -78.981 49.001
71170 95.412 -78.986 48.958
71171 95.421 -78.992 48.915
71172 95.429 -78.998 48.871
71173 95.435 -79.012 48.827
71174 95.440 -79.033 48.784
71175 95.445 -79.053 48.740
71176 95.449 -79.073 48.697
71177 95.454 -79.093 48.653
71178 95.458 -79.113 48.608
71179 95.463 -79.133 48.566
71180 95.467 -79.153 48.523
71181 95.472 -79.173 48.479
71182 95.475 -79.190 48.435
71183 95.488 -79.206 48.391
71184 95.513 -79.197 48.347
71185 95.535 -79.184 48.302
71186 95.557 -79.171 48.257
71187 95.580 -79.158 48.213
71188 95.603 -79.144 48.170
71189 95.625 -79.131 48.127
71190 95.648 -79.118 48.083
71191 95.670 -79.105 48.039
71192 95.693 -79.092 47.995
71193 95.714 -79.080 47.949
71194 95.736 -79.081 47.905
71195 95.756 -79.085 47.862
71196 95.768 -79.086 47.820
71197 95.780 -79.087 47.778
71198 95.793 -79.087 47.736
71199 95.805 -79.088 47.695
71200 95.817 -79.089 47.654
71201 95.830 -79.090 47.613
71202 95.842 -79.090 47.572
71203 95.854 -79.091 47.533
71204 95.867 -79.092 47.494
71205 95.883 -79.106 47.453
71206 95.903 -79.128 47.414
71207 95.923 -79.149 47.375
71208 95.930 -79.150 47.336
71209 95.986 -79.131 47.298
71210 96.029 -79.115 47.260
71211 96.063 -79.103 47.222
71212 96.098 -79.090 47.184
71213 96.132 -79.078 47.148
71214 96.166 -79.065 47.112
71215 96.200 -79.052 47.077
71216 96.235 -79.039 47.043
71217 96.269 -79.026 47.010
71218 96.303 -79.014 46.977
71219 96.337 -79.001 46.944
71220 96.367 -79.034 46.914
71221 96.396 -79.080 46.883
71222 96.424 -79.126 46.851
71223 96.453 -79.172 46.819
71224 96.468 -79.174 46.788
71225 96.482 -79.174 46.756
71226 96.498 -79.175 46.724
71227 96.512 -79.175 46.693
71228 96.527 -79.175 46.662
71229 96.544 -79.180 46.632
71230 96.568 -79.191 46.602
71231 96.573 -79.181 46.571
71232 96.600 -79.178 46.541
71233 96.632 -79.179 46.513
71234 96.655 -79.182 46.483
71235 96.678 -79.184 46.455
71236 96.701 -79.187 46.428
71237 96.724 -79.190 46.400
71238 96.746 -79.192 46.374
71239 96.769 -79.194 46.348
71240 96.792 -79.197 46.320
71241 96.815 -79.200 46.295
71242 96.837 -79.202 46.269
71243 96.857 -79.182 46.243
71244 96.875 -79.161 46.218
71245 96.896 -79.143 46.193
71246 96.933 -79.153 46.169
71247 96.945 -79.178 46.146
71248 96.930 -79.212 46.123
71249 96.965 -79.176 46.099
71250 97.001 -79.140 46.075
71251 97.037 -79.118 46.053
71252 97.079 -79.163 46.033
71253 97.090 -79.178 46.014
71254 97.085 -79.179 45.994
71255 97.081 -79.180 45.974
71256 97.081 -79.181 45.955
71257 97.091 -79.181 45.935
71258 97.100 -79.181 45.916
71259 97.109 -79.182 45.897
71260 97.118 -79.187 45.879
71261 97.133 -79.187 45.859
71262 97.146 -79.184 45.840
71263 97.168 -79.169 45.822
71264 97.169 -79.162 45.804
71265 97.171 -79.155 45.784
71266 97.173 -79.147 45.766
71267 97.174 -79.145 45.749
71268 97.175 -79.144 45.731
71269 97.176 -79.143 45.714
71270 97.177 -79.142 45.696
71271 97.195 -79.138 45.679
71272 97.218 -79.148 45.660
71273 97.232 -79.165 45.643
71274 97.246 -79.182 45.627
71275 97.260 -79.199 45.611
71276 97.275 -79.216 45.595
71277 97.289 -79.233 45.578
71278 97.303 -79.250 45.562
71279 97.317 -79.266 45.546
71280 97.331 -79.284 45.530
71281 97.346 -79.300 45.515
71282 97.296 -79.344 45.500
71283 97.308 -79.340 45.485
71284 97.323 -79.334 45.469
71285 97.344 -79.325 45.454
71286 97.372 -79.311 45.438
71287 97.385 -79.300 45.424
71288 97.397 -79.288 45.409
71289 97.411 -79.277 45.395
71290 97.438 -79.288 45.381
71291 97.469 -79.368 45.367
71292 97.474 -79.375 45.352
71293 97.480 -79.383 45.338
71294 97.489 -79.392 45.324
71295 97.496 -79.400 45.310
71296 97.505 -79.394 45.297
71297 97.527 -79.385 45.284
71298 97.543 -79.373 45.270
71299 97.558 -79.361 45.258
71300 97.562 -79.357 45.245
71301 97.567 -79.367 45.232
71302 97.582 -79.398 45.220
71303 97.599 -79.415 45.207
71304 97.616 -79.432 45.195
71305 97.633 -79.449 45.182
71306 97.650 -79.466 45.170
71307 97.667 -79.483 45.158
71308 97.684 -79.501 45.146
71309 97.701 -79.518 45.135
71310 97.718 -79.535 45.123
71311 97.735 -79.552 45.112
71312 97.752 -79.569 45.101
71313 97.769 -79.586 45.089
71314 97.786 -79.603 45.077
71315 97.803 -79.620 45.064
71316 97.820 -79.637 45.052
71317 97.837 -79.654 45.040
71318 97.854 -79.671 45.027
71319 97.871 -79.687 45.016
71320 97.888 -79.703 45.002
71321 97.906 -79.719 44.989
71322 97.923 -79.735 44.977
71323 97.941 -79.751 44.964
71324 97.958 -79.767 44.952
71325 97.975 -79.783 44.941
71326 97.993 -79.798 44.930
71327 98.010 -79.814 44.920
71328 98.027 -79.830 44.910
71329 98.044 -79.846 44.900
71330 98.062 -79.863 44.890
71331 98.079 -79.879 44.881
71332 98.096 -79.895 44.872
71333 98.113 -79.911 44.863
71334 98.131 -79.927 44.854
71335 98.148 -79.943 44.845
71336 98.165 -79.958 44.837
71337 98.182 -79.974 44.830
71338 98.200 -79.990 44.823
71339 98.217 -80.006 44.815
71340 98.238 -80.021 44.807
71341 98.262 -80.036 44.800
71342 98.286 -80.050 44.793
71343 98.309 -80.065 44.786
71344 98.333 -80.078 44.780
71345 98.357 -80.093 44.774
71346 98.385 -80.098 44.768
71347 98.429 -80.071 44.760
71348 98.456 -80.087 44.753
71349 98.484 -80.102 44.747
71350 98.511 -80.116 44.741
71351 98.538 -80.131 44.737
71352 98.566 -80.146 44.731
71353 98.592 -80.158 44.725
71354 98.615 -80.161 44.719
71355 98.640 -80.164 44.715
71356 98.664 -80.167 44.710
71357 98.688 -80.169 44.705
71358 98.712 -80.172 44.700
71359 98.737 -80.178 44.694
71360 98.763 -80.187 44.685
71361 98.789 -80.198 44.670
71362 98.810 -80.209 44.655
71363 98.819 -80.210 44.642
71364 98.822 -80.212 44.630
71365 98.834 -80.221 44.622
71366 98.850 -80.232 44.614
71367 98.869 -80.243 44.606
71368 98.917 -80.243 44.599
71369 98.970 -80.266 44.592
71370 98.993 -80.277 44.585
71371 99.017 -80.287 44.580
71372 99.040 -80.298 44.574
71373 99.063 -80.309 44.568
71374 99.086 -80.319 44.564
71375 99.110 -80.330 44.558
71376 99.133 -80.340 44.553
71377 99.156 -80.350 44.548
71378 99.180 -80.361 44.542
71379 99.203 -80.372 44.538
71380 99.226 -80.382 44.534
71381 99.250 -80.393 44.530
71382 99.273 -80.403 44.528
71383 99.296 -80.414 44.525
71384 99.320 -80.425 44.521
71385 99.353 -80.431 44.516
71386 99.392 -80.435 44.513
71387 99.430 -80.438 44.509
71388 99.469 -80.442 44.506
71708 93.494 -78.981 53.473
71709 93.503 -78.979 53.462
71710 93.513 -78.977 53.452
71711 93.523 -78.973 53.442
71712 93.535 -78.962 53.433
71713 93.544 -78.954 53.424
71714 93.552 -78.947 53.415
71715 93.561 -78.940 53.407
71716 93.569 -78.933 53.399
71717 93.577 -78.927 53.392
71718 93.586 -78.919 53.385
71719 93.594 -78.912 53.378
71720 93.603 -78.905 53.371
71721 93.655 -78.879 53.365
71722 93.638 -78.875 53.359
71723 93.655 -78.873 53.354
71724 93.672 -78.871 53.349
71725 93.682 -78.869 53.342
71726 93.692 -78.868 53.336
71727 93.701 -78.868 53.331
71728 93.711 -78.866 53.325
71729 93.720 -78.865 53.320
71730 93.730 -78.864 53.314
71731 93.740 -78.863 53.305
71732 93.746 -78.863 53.297
71733 93.753 -78.863 53.287
71734 93.760 -78.861 53.277
71735 93.767 -78.858 53.265
71736 93.774 -78.855 53.254
71737 93.781 -78.853 53.242
71738 93.788 -78.851 53.230
71739 93.795 -78.849 53.219
71740 93.802 -78.846 53.207
71741 93.809 -78.844 53.194
71742 93.819 -78.835 53.178
71743 93.821 -78.836 53.162
71744 93.820 -78.836 53.147
71745 93.816 -78.831 53.130
71746 93.812 -78.826 53.111
71747 93.816 -78.821 53.091
71748 93.821 -78.817 53.071
71749 93.826 -78.812 53.048
71750 93.831 -78.808 53.023
71751 93.837 -78.803 52.999
71752 93.842 -78.799 52.973
71753 93.847 -78.795 52.947
71754 93.852 -78.790 52.922
71755 93.867 -78.784 52.897
71756 93.884 -78.777 52.869
71757 93.899 -78.771 52.842
71758 93.915 -78.765 52.815
71759 93.931 -78.758 52.790
71760 93.947 -78.752 52.764
71761 93.963 -78.746 52.738
71762 93.982 -78.746 52.714
71763 94.002 -78.748 52.688
71764 94.023 -78.750 52.664
71765 94.043 -78.752 52.642
71766 94.063 -78.755 52.618
71767 94.076 -78.752 52.595
71768 94.086 -78.749 52.573
71769 94.097 -78.745 52.551
71770 94.107 -78.741 52.527
71771 94.117 -78.737 52.503
71772 94.127 -78.733 52.477
71773 94.137 -78.729 52.451
71774 94.145 -78.730 52.426
71775 94.149 -78.737 52.402
71776 94.169 -78.730 52.379
71777 94.196 -78.734 52.352
71778 94.206 -78.731 52.325
71779 94.216 -78.729 52.298
71780 94.227 -78.726 52.272
71781 94.242 -78.724 52.247
71782 94.256 -78.720 52.222
71783 94.272 -78.709 52.195
71784 94.283 -78.705 52.167
71785 94.291 -78.703 52.139
71786 94.300 -78.701 52.109
71787 94.309 -78.698 52.079
71788 94.318 -78.696 52.051
71789 94.326 -78.694 52.023
71790 94.335 -78.692 51.996
71791 94.343 -78.689 51.966
71792 94.352 -78.687 51.936
71793 94.368 -78.689 51.905
71794 94.382 -78.694 51.874
71795 94.395 -78.698 51.842
71796 94.408 -78.703 51.811
71797 94.421 -78.708 51.781
71798 94.434 -78.713 51.750
71799 94.449 -78.708 51.718
71800 94.467 -78.713 51.687
71801 94.484 -78.717 51.656
71802 94.501 -78.721 51.626
71803 94.526 -78.727 51.596
71804 94.553 -78.734 51.564
71805 94.580 -78.742 51.528
71806 94.607 -78.749 51.491
71807 94.634 -78.755 51.454
71808 94.662 -78.762 51.419
71809 94.687 -78.768 51.385
71810 94.705 -78.771 51.352
71811 94.723 -78.772 51.318
71812 94.741 -78.774 51.284
71813 94.759 -78.777 51.250
71814 94.770 -78.779 51.215
71815 94.778 -78.774 51.178
71816 94.784 -78.764 51.142
71817 94.783 -78.761 51.106
71818 94.779 -78.762 51.069
71819 94.775 -78.764 51.032
71820 94.772 -78.764 50.995
71821 94.756 -78.758 50.957
71822 94.734 -78.747 50.918
71823 94.721 -78.762 50.878
71824 94.710 -78.779 50.836
71825 94.711 -78.783 50.796
71826 94.730 -78.771 50.753
71827 94.749 -78.758 50.709
71828 94.777 -78.749 50.667
71829 94.801 -78.745 50.625
71830 94.819 -78.741 50.582
71831 94.837 -78.738 50.541
71832 94.855 -78.735 50.499
71833 94.873 -78.731 50.457
71834 94.883 -78.736 50.415
71835 94.881 -78.752 50.375
71836 94.880 -78.767 50.336
71837 94.879 -78.783 50.299
71838 94.878 -78.799 50.263
71839 94.877 -78.814 50.226
71840 94.876 -78.830 50.188
71841 94.874 -78.845 50.152
71842 94.867 -78.846 50.114
71843 94.859 -78.846 50.076
71844 94.851 -78.846 50.037
71845 94.843 -78.845 49.998
71846 94.857 -78.848 49.958
71847 94.874 -78.851 49.917
71848 94.890 -78.854 49.876
71849 94.907 -78.857 49.834
71850 94.924 -78.860 49.793
71851 94.941 -78.863 49.755
71852 94.957 -78.866 49.719
71853 94.974 -78.869 49.685
71854 94.998 -78.868 49.650
71855 95.028 -78.874 49.614
71856 95.057 -78.880 49.575
71857 95.087 -78.886 49.537
71858 95.116 -78.893 49.498
71859 95.145 -78.900 49.456
71860 95.175 -78.907 49.415
71861 95.204 -78.914 49.374
71862 95.234 -78.921 49.334
71863 95.263 -78.927 49.295
71864 95.292 -78.934 49.257
71865 95.322 -78.941 49.217
71866 95.349 -78.946 49.176
71867 95.374 -78.949 49.135
71868 95.396 -78.952 49.095
71869 95.404 -78.958 49.051
71870 95.413 -78.963 49.009
71871 95.421 -78.969 48.967
71872 95.429 -78.975 48.924
71873 95.438 -78.981 48.880
71874 95.446 -78.991 48.837
71875 95.450 -79.011 48.795
71876 95.455 -79.031 48.753
71877 95.459 -79.051 48.710
71878 95.464 -79.071 48.666
71879 95.468 -79.091 48.622
71880 95.473 -79.112 48.578
71881 95.478 -79.131 48.535
71882 95.482 -79.152 48.492
71883 95.487 -79.172 48.447
71884 95.491 -79.192 48.403
71885 95.511 -79.187 48.358
71886 95.533 -79.173 48.313
71887 95.556 -79.160 48.268
71888 95.578 -79.147 48.223
71889 95.601 -79.134 48.179
71890 95.623 -79.120 48.134
71891 95.645 -79.107 48.090
71892 95.668 -79.094 48.045
71893 95.690 -79.081 48.000
71894 95.712 -79.068 47.955
71895 95.728 -79.063 47.911
71896 95.740 -79.064 47.867
71897 95.752 -79.065 47.824
71898 95.764 -79.065 47.782
71899 95.776 -79.066 47.741
71900 95.789 -79.067 47.699
71901 95.801 -79.068 47.657
71902 95.813 -79.068 47.616
71903 95.825 -79.069 47.575
71904 95.838 -79.070 47.537
71905 95.850 -79.071 47.497
71906 95.862 -79.071 47.457
71907 95.881 -79.090 47.418
71908 95.901 -79.112 47.378
71909 95.921 -79.134 47.338
71910 95.947 -79.106 47.300
71911 96.035 -79.106 47.261
71912 96.072 -79.101 47.223
71913 96.106 -79.089 47.186
71914 96.140 -79.076 47.149
71915 96.174 -79.063 47.114
71916 96.208 -79.050 47.079
71917 96.243 -79.037 47.043
71918 96.277 -79.024 47.011
71919 96.311 -79.012 46.978
71920 96.345 -78.999 46.945
71921 96.375 -79.034 46.914
71922 96.404 -79.079 46.883
71923 96.432 -79.125 46.852
71924 96.455 -79.153 46.820
71925 96.470 -79.153 46.789
71926 96.484 -79.153 46.758
71927 96.499 -79.154 46.726
71928 96.514 -79.154 46.696
71929 96.532 -79.159 46.664
71930 96.556 -79.172 46.632
71931 96.580 -79.183 46.601
71932 96.593 -79.184 46.570
71933 96.591 -79.166 46.541
71934 96.608 -79.159 46.514
71935 96.645 -79.159 46.484
71936 96.681 -79.160 46.457
71937 96.709 -79.162 46.430
71938 96.731 -79.165 46.402
71939 96.754 -79.167 46.376
71940 96.777 -79.169 46.350
71941 96.800 -79.172 46.323
71942 96.822 -79.175 46.297
71943 96.846 -79.177 46.271
71944 96.865 -79.159 46.245
71945 96.883 -79.137 46.220
71946 96.916 -79.142 46.196
71947 96.919 -79.173 46.172
71948 96.899 -79.212 46.149
71949 96.935 -79.176 46.126
71950 96.970 -79.140 46.103
71951 97.006 -79.103 46.079
71952 97.044 -79.112 46.056
71953 97.087 -79.158 46.036
71954 97.093 -79.169 46.016
71955 97.088 -79.169 45.996
71956 97.084 -79.170 45.977
71957 97.080 -79.171 45.958
71958 97.086 -79.172 45.939
71959 97.096 -79.172 45.919
71960 97.105 -79.172 45.901
71961 97.114 -79.173 45.883
71962 97.124 -79.175 45.864
71963 97.145 -79.163 45.845
71964 97.164 -79.151 45.826
71965 97.171 -79.142 45.808
71966 97.171 -79.143 45.788
71967 97.172 -79.143 45.770
71968 97.173 -79.143 45.753
71969 97.173 -79.144 45.737
71970 97.174 -79.145 45.719
71971 97.175 -79.148 45.702
71972 97.185 -79.143 45.684
71973 97.223 -79.141 45.665
71974 97.249 -79.148 45.648
71975 97.263 -79.165 45.633
71976 97.278 -79.181 45.617
71977 97.292 -79.199 45.600
71978 97.306 -79.216 45.584
71979 97.320 -79.232 45.568
71980 97.334 -79.250 45.552
71981 97.349 -79.266 45.536
71982 97.341 -79.292 45.521
71983 97.296 -79.330 45.506
71984 97.312 -79.324 45.491
71985 97.327 -79.317 45.475
71986 97.343 -79.311 45.460
71987 97.371 -79.297 45.445
71988 97.384 -79.286 45.430
71989 97.397 -79.274 45.416
71990 97.421 -79.281 45.402
71991 97.449 -79.298 45.388
71992 97.468 -79.354 45.374
71993 97.474 -79.363 45.359
71994 97.479 -79.370 45.345
71995 97.486 -79.379 45.331
71996 97.495 -79.388 45.317
71997 97.510 -79.380 45.305
71998 97.526 -79.368 45.292
71999 97.536 -79.359 45.278
72000 97.540 -79.354 45.266
72001 97.544 -79.350 45.253
72002 97.547 -79.345 45.240
72003 97.551 -79.341 45.228
72004 97.559 -79.438 45.215
72005 97.575 -79.462 45.203
72006 97.596 -79.452 45.191
72007 97.613 -79.469 45.178
72008 97.630 -79.485 45.167
72009 97.648 -79.501 45.155
72010 97.665 -79.517 45.144
72011 97.682 -79.534 45.134
72012 97.700 -79.550 45.122
72013 97.717 -79.566 45.110
72014 97.734 -79.582 45.098
72015 97.751 -79.598 45.086
72016 97.769 -79.614 45.074
72017 97.786 -79.629 45.062
72018 97.803 -79.645 45.049
72019 97.820 -79.661 45.037
72020 97.838 -79.677 45.025
72021 97.855 -79.693 45.011
72022 97.872 -79.709 44.999
72023 97.890 -79.725 44.988
72024 97.907 -79.741 44.975
72025 97.924 -79.757 44.963
72026 97.942 -79.773 44.952
72027 97.959 -79.789 44.941
72028 97.976 -79.805 44.931
72029 97.994 -79.821 44.921
72030 98.011 -79.837 44.912
72031 98.028 -79.853 44.903
72032 98.045 -79.869 44.893
72033 98.063 -79.885 44.884
72034 98.080 -79.901 44.877
72035 98.097 -79.917 44.868
72036 98.114 -79.933 44.860
72037 98.134 -79.948 44.852
72038 98.158 -79.962 44.845
72039 98.181 -79.977 44.838
72040 98.205 -79.991 44.831
72041 98.229 -80.005 44.823
72042 98.253 -80.020 44.817
72043 98.276 -80.034 44.810
72044 98.302 -80.047 44.804
72045 98.334 -80.046 44.796
72046 98.346 -80.048 44.790
72047 98.387 -80.033 44.783
72048 98.434 -80.044 44.775
72049 98.462 -80.059 44.768
72050 98.489 -80.074 44.762
72051 98.517 -80.089 44.757
72052 98.544 -80.104 44.752
72053 98.572 -80.119 44.746
72054 98.600 -80.127 44.739
72055 98.630 -80.130 44.731
72056 98.655 -80.133 44.727
72057 98.679 -80.135 44.723
72058 98.704 -80.138 44.717
72059 98.728 -80.142 44.711
72060 98.754 -80.152 44.704
72061 98.776 -80.163 44.694
72062 98.789 -80.177 44.679
72063 98.808 -80.188 44.665
72064 98.831 -80.187 44.652
72065 98.839 -80.182 44.641
72066 98.847 -80.188 44.632
72067 98.863 -80.200 44.624
72068 98.880 -80.211 44.616
72069 98.896 -80.222 44.609
72070 98.929 -80.231 44.602
72071 98.997 -80.250 44.596
72072 99.021 -80.260 44.590
72073 99.044 -80.271 44.584
72074 99.067 -80.281 44.579
72075 99.090 -80.292 44.573
72076 99.114 -80.303 44.568
72077 99.137 -80.312 44.563
72078 99.160 -80.323 44.557
72079 99.184 -80.334 44.552
72080 99.207 -80.344 44.548
72081 99.230 -80.355 44.544
72082 99.254 -80.366 44.540
72083 99.277 -80.376 44.537
72084 99.300 -80.387 44.534
72085 99.326 -80.397 44.530
72086 99.364 -80.400 44.526
72087 99.403 -80.404 44.523
72088 99.441 -80.408 44.520
72089 99.480 -80.413 44.516
72090 99.518 -80.416 44.512
72091 99.557 -80.420 44.509
72410 93.504 -78.968 53.481
72411 93.513 -78.965 53.470
72412 93.523 -78.963 53.460
72413 93.533 -78.959 53.450
72414 93.545 -78.949 53.441
72415 93.554 -78.940 53.432
72416 93.562 -78.933 53.423
72417 93.570 -78.927 53.415
72418 93.578 -78.920 53.408
72419 93.587 -78.913 53.399
72420 93.595 -78.906 53.391
72421 93.604 -78.899 53.384
72422 93.643 -78.878 53.377
72423 93.660 -78.860 53.371
72424 93.656 -78.858 53.366
72425 93.668 -78.856 53.361
72426 93.678 -78.855 53.356
72427 93.688 -78.854 53.351
72428 93.697 -78.852 53.346
72429 93.707 -78.852 53.341
72430 93.716 -78.850 53.336
72431 93.726 -78.849 53.330
72432 93.736 -78.848 53.322
72433 93.743 -78.846 53.313
72434 93.750 -78.844 53.304
72435 93.757 -78.842 53.295
72436 93.764 -78.839 53.283
72437 93.771 -78.837 53.271
72438 93.778 -78.835 53.259
72439 93.785 -78.833 53.247
72440 93.792 -78.830 53.236
72441 93.799 -78.827 53.223
72442 93.806 -78.825 53.209
72443 93.813 -78.821 53.193
72444 93.811 -78.819 53.176
72445 93.812 -78.814 53.157
72446 93.818 -78.810 53.140
72447 93.823 -78.805 53.120
72448 93.828 -78.801 53.101
72449 93.833 -78.796 53.081
72450 93.839 -78.793 53.057
72451 93.844 -78.788 53.033
72452 93.849 -78.783 53.007
72453 93.854 -78.779 52.981
72454 93.860 -78.774 52.954
72455 93.867 -78.770 52.929
72456 93.883 -78.764 52.904
72457 93.899 -78.757 52.876
72458 93.915 -78.751 52.849
72459 93.930 -78.745 52.822
72460 93.946 -78.738 52.797
72461 93.963 -78.732 52.772
72462 93.978 -78.726 52.747
72463 93.994 -78.719 52.723
72464 94.010 -78.713 52.699
72465 94.030 -78.714 52.675
72466 94.050 -78.716 52.653
72467 94.070 -78.718 52.630
72468 94.084 -78.717 52.607
72469 94.095 -78.713 52.585
72470 94.105 -78.709 52.563
72471 94.115 -78.705 52.540
72472 94.125 -78.701 52.516
72473 94.130 -78.707 52.491
72474 94.134 -78.715 52.465
72475 94.139 -78.723 52.439
72476 94.153 -78.702 52.414
72477 94.205 -78.715 52.391
72478 94.219 -78.714 52.365
72479 94.229 -78.711 52.338
72480 94.239 -78.708 52.311
72481 94.249 -78.706 52.286
72482 94.259 -78.703 52.262
72483 94.270 -78.701 52.238
72484 94.281 -78.696 52.212
72485 94.290 -78.694 52.183
72486 94.299 -78.692 52.154
72487 94.308 -78.689 52.125
72488 94.316 -78.687 52.096
72489 94.325 -78.685 52.069
72490 94.334 -78.683 52.042
72491 94.342 -78.680 52.014
72492 94.351 -78.678 51.986
72493 94.359 -78.676 51.956
72494 94.372 -78.677 51.927
72495 94.385 -78.683 51.895
72496 94.398 -78.687 51.863
72497 94.411 -78.692 51.831
72498 94.424 -78.696 51.799
72499 94.437 -78.702 51.769
72500 94.452 -78.700 51.739
72501 94.469 -78.704 51.708
72502 94.486 -78.708 51.675
72503 94.509 -78.714 51.642
72504 94.537 -78.721 51.609
72505 94.564 -78.728 51.576
72506 94.591 -78.735 51.541
72507 94.618 -78.742 51.503
72508 94.636 -78.744 51.467
72509 94.654 -78.745 51.432
72510 94.672 -78.748 51.398
72511 94.690 -78.750 51.364
72512 94.708 -78.752 51.332
72513 94.726 -78.755 51.298
72514 94.745 -78.757 51.265
72515 94.755 -78.759 51.232
72516 94.766 -78.761 51.196
72517 94.773 -78.754 51.159
72518 94.779 -78.744 51.124
72519 94.784 -78.733 51.085
72520 94.783 -78.730 51.046
72521 94.779 -78.732 51.006
72522 94.765 -78.726 50.968
72523 94.750 -78.734 50.930
72524 94.739 -78.751 50.891
72525 94.727 -78.767 50.850
72526 94.726 -78.774 50.810
72527 94.745 -78.761 50.767
72528 94.764 -78.749 50.722
72529 94.781 -78.736 50.680
72530 94.799 -78.733 50.638
72531 94.817 -78.730 50.596
72532 94.835 -78.727 50.555
72533 94.853 -78.724 50.513
72534 94.872 -78.720 50.471
72535 94.881 -78.725 50.427
72536 94.880 -78.740 50.384
72537 94.879 -78.756 50.344
72538 94.878 -78.771 50.306
72539 94.877 -78.786 50.271
72540 94.875 -78.802 50.234
72541 94.874 -78.818 50.198
72542 94.873 -78.833 50.161
72543 94.872 -78.849 50.124
72544 94.864 -78.849 50.087
72545 94.856 -78.848 50.048
72546 94.848 -78.848 50.008
72547 94.857 -78.849 49.969
72548 94.875 -78.852 49.929
72549 94.891 -78.855 49.888
72550 94.908 -78.858 49.846
72551 94.924 -78.861 49.805
72552 94.942 -78.865 49.767
72553 94.958 -78.868 49.731
72554 94.987 -78.863 49.696
72555 95.016 -78.858 49.660
72556 95.045 -78.858 49.623
72557 95.074 -78.865 49.584
72558 95.103 -78.870 49.544
72559 95.132 -78.876 49.505
72560 95.162 -78.881 49.462
72561 95.191 -78.887 49.421
72562 95.220 -78.893 49.380
72563 95.250 -78.899 49.341
72564 95.279 -78.906 49.302
72565 95.309 -78.913 49.263
72566 95.338 -78.920 49.224
72567 95.368 -78.927 49.183
72568 95.393 -78.930 49.142
72569 95.405 -78.935 49.101
72570 95.413 -78.940 49.057
72571 95.421 -78.946 49.016
72572 95.430 -78.952 48.975
72573 95.439 -78.958 48.932
72574 95.447 -78.964 48.889
72575 95.456 -78.970 48.847
72576 95.460 -78.990 48.805
72577 95.465 -79.009 48.763
72578 95.469 -79.030 48.721
72579 95.474 -79.049 48.677
72580 95.479 -79.070 48.633
72581 95.483 -79.090 48.589
72582 95.488 -79.110 48.546
72583 95.492 -79.130 48.503
72584 95.497 -79.150 48.458
72585 95.511 -79.165 48.414
72586 95.526 -79.178 48.370
72587 95.548 -79.166 48.324
72588 95.570 -79.153 48.278
72589 95.593 -79.140 48.232
72590 95.615 -79.127 48.186
72591 95.637 -79.114 48.141
72592 95.660 -79.100 48.095
72593 95.682 -79.087 48.050
72594 95.705 -79.074 48.004
72595 95.725 -79.062 47.960
72596 95.734 -79.059 47.916
72597 95.741 -79.056 47.872
72598 95.747 -79.052 47.830
72599 95.753 -79.048 47.788
72600 95.760 -79.045 47.747
72601 95.772 -79.046 47.704
72602 95.784 -79.046 47.662
72603 95.797 -79.047 47.620
72604 95.809 -79.048 47.579
72605 95.821 -79.049 47.540
72606 95.834 -79.049 47.500
72607 95.846 -79.050 47.460
72608 95.859 -79.054 47.420
72609 95.879 -79.075 47.380
72610 95.899 -79.093 47.340
72611 95.941 -79.084 47.302
72612 96.037 -79.088 47.263
72613 96.077 -79.091 47.225
72614 96.114 -79.087 47.188
72615 96.148 -79.075 47.151
72616 96.182 -79.062 47.116
72617 96.217 -79.049 47.081
72618 96.251 -79.036 47.044
72619 96.285 -79.024 47.012
72620 96.320 -79.011 46.979
72621 96.354 -78.998 46.947
72622 96.383 -79.033 46.915
72623 96.412 -79.079 46.884
72624 96.440 -79.125 46.853
72625 96.457 -79.132 46.822
72626 96.472 -79.132 46.790
72627 96.486 -79.133 46.759
72628 96.501 -79.134 46.728
72629 96.520 -79.139 46.697
72630 96.544 -79.151 46.665
72631 96.568 -79.162 46.633
72632 96.591 -79.175 46.601
72633 96.614 -79.185 46.571
72634 96.612 -79.169 46.541
72635 96.610 -79.152 46.514
72636 96.617 -79.138 46.486
72637 96.653 -79.139 46.458
72638 96.690 -79.140 46.431
72639 96.726 -79.141 46.403
72640 96.762 -79.142 46.377
72641 96.785 -79.144 46.351
72642 96.807 -79.147 46.326
72643 96.830 -79.150 46.299
72644 96.853 -79.152 46.274
72645 96.872 -79.135 46.249
72646 96.901 -79.131 46.224
72647 96.892 -79.168 46.200
72648 96.869 -79.212 46.177
72649 96.904 -79.175 46.154
72650 96.940 -79.140 46.130
72651 96.975 -79.103 46.106
72652 97.011 -79.067 46.082
72653 97.052 -79.107 46.060
72654 97.094 -79.152 46.040
72655 97.095 -79.158 46.020
72656 97.091 -79.159 46.000
72657 97.087 -79.160 45.981
72658 97.082 -79.161 45.962
72659 97.082 -79.162 45.944
72660 97.091 -79.162 45.924
72661 97.100 -79.163 45.905
72662 97.104 -79.171 45.887
72663 97.108 -79.174 45.869
72664 97.124 -79.161 45.850
72665 97.145 -79.146 45.831
72666 97.168 -79.127 45.812
72667 97.184 -79.219 45.793
72668 97.181 -79.128 45.775
72669 97.180 -79.098 45.759
72670 97.181 -79.099 45.743
72671 97.170 -79.112 45.725
72672 97.168 -79.148 45.708
72673 97.178 -79.147 45.690
72674 97.213 -79.146 45.671
72675 97.250 -79.144 45.655
72676 97.281 -79.147 45.639
72677 97.295 -79.165 45.623
72678 97.309 -79.181 45.607
72679 97.323 -79.198 45.591
72680 97.337 -79.215 45.575
72681 97.352 -79.232 45.559
72682 97.366 -79.249 45.544
72683 97.330 -79.286 45.529
72684 97.300 -79.314 45.513
72685 97.315 -79.307 45.498
72686 97.331 -79.301 45.483
72687 97.347 -79.294 45.468
72688 97.370 -79.284 45.452
72689 97.383 -79.272 45.438
72690 97.404 -79.273 45.425
72691 97.432 -79.285 45.410
72692 97.461 -79.328 45.396
72693 97.468 -79.341 45.382
72694 97.473 -79.349 45.368
72695 97.479 -79.357 45.353
72696 97.485 -79.365 45.339
72697 97.493 -79.374 45.325
72698 97.508 -79.363 45.313
72699 97.514 -79.357 45.300
72700 97.518 -79.352 45.287
72701 97.522 -79.347 45.274
72702 97.525 -79.343 45.262
72703 97.529 -79.338 45.250
72704 97.533 -79.334 45.237
72705 97.537 -79.329 45.225
72706 97.543 -79.410 45.213
72707 97.566 -79.400 45.200
72708 97.585 -79.435 45.188
72709 97.605 -79.469 45.177
72710 97.620 -79.523 45.166
72711 97.635 -79.524 45.154
72712 97.649 -79.525 45.143
72713 97.666 -79.540 45.131
72714 97.683 -79.556 45.120
72715 97.701 -79.572 45.108
72716 97.718 -79.588 45.097
72717 97.735 -79.604 45.085
72718 97.752 -79.620 45.073
72719 97.770 -79.636 45.061
72720 97.787 -79.651 45.048
72721 97.804 -79.667 45.036
72722 97.821 -79.683 45.023
72723 97.839 -79.699 45.010
72724 97.856 -79.715 44.998
72725 97.873 -79.731 44.987
72726 97.891 -79.747 44.975
72727 97.908 -79.763 44.964
72728 97.925 -79.779 44.954
72729 97.943 -79.795 44.944
72730 97.960 -79.811 44.934
72731 97.977 -79.827 44.925
72732 97.995 -79.843 44.916
72733 98.012 -79.859 44.907
72734 98.029 -79.875 44.898
72735 98.054 -79.889 44.891
72736 98.077 -79.903 44.883
72737 98.101 -79.918 44.875
72738 98.125 -79.932 44.868
72739 98.148 -79.947 44.861
72740 98.171 -79.961 44.855
72741 98.196 -79.974 44.849
72742 98.222 -79.987 44.842
72743 98.253 -79.991 44.836
72744 98.266 -79.993 44.829
72745 98.278 -79.994 44.821
72746 98.290 -79.996 44.813
72747 98.319 -80.002 44.805
72748 98.371 -80.012 44.797
72749 98.423 -80.021 44.790
72750 98.467 -80.032 44.783
72751 98.495 -80.047 44.778
72752 98.522 -80.062 44.774
72753 98.550 -80.077 44.767
72754 98.578 -80.091 44.760
72755 98.607 -80.094 44.752
72756 98.637 -80.098 44.744
72757 98.667 -80.101 44.739
72758 98.696 -80.104 44.736
72759 98.720 -80.107 44.731
72760 98.742 -80.117 44.722
72761 98.756 -80.131 44.712
72762 98.769 -80.144 44.702
72763 98.782 -80.157 44.687
72764 98.807 -80.168 44.674
72765 98.833 -80.175 44.662
72766 98.849 -80.159 44.652
72767 98.859 -80.156 44.643
72768 98.876 -80.167 44.634
72769 98.892 -80.178 44.627
72770 98.909 -80.190 44.620
72771 98.926 -80.201 44.613
72772 98.942 -80.212 44.606
72773 98.995 -80.219 44.600
72774 99.048 -80.243 44.594
72775 99.071 -80.254 44.589
72776 99.094 -80.265 44.584
72777 99.118 -80.275 44.578
72778 99.141 -80.286 44.572
72779 99.164 -80.296 44.567
72780 99.188 -80.306 44.562
72781 99.211 -80.317 44.558
72782 99.234 -80.328 44.554
72783 99.258 -80.338 44.550
72784 99.281 -80.349 44.546
72785 99.304 -80.359 44.543
72786 99.336 -80.366 44.540
72787 99.375 -80.370 44.536
72788 99.413 -80.374 44.533
72789 99.452 -80.378 44.529
72790 99.491 -80.382 44.525
72791 99.529 -80.386 44.521
72792 99.564 -80.391 44.518
72793 99.567 -80.414 44.514
72794 99.570 -80.436 44.510
72795 99.574 -80.458 44.506
73112 93.513 -78.955 53.486
73113 93.523 -78.952 53.475
73114 93.533 -78.949 53.465
73115 93.543 -78.944 53.455
73116 93.555 -78.934 53.447
73117 93.563 -78.927 53.439
73118 93.571 -78.921 53.430
73119 93.580 -78.913 53.421
73120 93.588 -78.906 53.413
73121 93.596 -78.899 53.405
73122 93.605 -78.893 53.397
73123 93.631 -78.877 53.389
73124 93.697 -78.845 53.383
73125 93.666 -78.841 53.378
73126 93.662 -78.836 53.374
73127 93.672 -78.833 53.369
73128 93.682 -78.832 53.365
73129 93.691 -78.830 53.362
73130 93.700 -78.828 53.357
73131 93.710 -78.827 53.352
73132 93.721 -78.830 53.345
73133 93.732 -78.839 53.337
73134 93.739 -78.836 53.329
73135 93.746 -78.833 53.320
73136 93.753 -78.830 53.311
73137 93.760 -78.827 53.298
73138 93.767 -78.825 53.286
73139 93.774 -78.822 53.274
73140 93.781 -78.819 53.262
73141 93.788 -78.816 53.250
73142 93.795 -78.814 53.237
73143 93.802 -78.811 53.223
73144 93.814 -78.803 53.206
73145 93.820 -78.799 53.187
73146 93.825 -78.794 53.168
73147 93.830 -78.789 53.149
73148 93.835 -78.786 53.128
73149 93.841 -78.781 53.107
73150 93.846 -78.777 53.087
73151 93.851 -78.772 53.063
73152 93.856 -78.767 53.039
73153 93.862 -78.763 53.014
73154 93.867 -78.759 52.988
73155 93.872 -78.755 52.962
73156 93.883 -78.749 52.936
73157 93.898 -78.742 52.910
73158 93.914 -78.736 52.883
73159 93.930 -78.730 52.856
73160 93.946 -78.724 52.830
73161 93.962 -78.717 52.805
73162 93.978 -78.711 52.779
73163 93.994 -78.705 52.755
73164 94.009 -78.698 52.732
73165 94.026 -78.692 52.708
73166 94.041 -78.686 52.685
73167 94.057 -78.680 52.664
73168 94.077 -78.681 52.641
73169 94.092 -78.681 52.619
73170 94.103 -78.677 52.596
73171 94.111 -78.677 52.573
73172 94.115 -78.685 52.551
73173 94.120 -78.692 52.528
73174 94.124 -78.700 52.504
73175 94.128 -78.708 52.478
73176 94.144 -78.684 52.452
73177 94.189 -78.686 52.426
73178 94.232 -78.696 52.402
73179 94.242 -78.694 52.377
73180 94.252 -78.691 52.351
73181 94.262 -78.689 52.325
73182 94.272 -78.686 52.300
73183 94.281 -78.683 52.277
73184 94.282 -78.688 52.254
73185 94.290 -78.685 52.228
73186 94.299 -78.683 52.199
73187 94.308 -78.680 52.170
73188 94.316 -78.678 52.142
73189 94.325 -78.676 52.114
73190 94.333 -78.673 52.087
73191 94.342 -78.671 52.061
73192 94.351 -78.669 52.034
73193 94.359 -78.667 52.006
73194 94.368 -78.664 51.977
73195 94.380 -78.665 51.948
73196 94.390 -78.670 51.917
73197 94.402 -78.676 51.885
73198 94.415 -78.680 51.853
73199 94.428 -78.685 51.820
73200 94.441 -78.690 51.790
73201 94.455 -78.694 51.760
73202 94.472 -78.695 51.729
73203 94.494 -78.701 51.696
73204 94.521 -78.708 51.661
73205 94.548 -78.714 51.625
73206 94.567 -78.717 51.590
73207 94.585 -78.720 51.553
73208 94.603 -78.721 51.515
73209 94.621 -78.724 51.479
73210 94.639 -78.726 51.444
73211 94.657 -78.728 51.411
73212 94.676 -78.730 51.378
73213 94.694 -78.733 51.346
73214 94.712 -78.735 51.314
73215 94.730 -78.737 51.282
73216 94.740 -78.739 51.250
73217 94.751 -78.741 51.215
73218 94.762 -78.743 51.178
73219 94.768 -78.733 51.141
73220 94.773 -78.724 51.102
73221 94.779 -78.713 51.062
73222 94.784 -78.702 51.022
73223 94.779 -78.706 50.983
73224 94.767 -78.723 50.944
73225 94.756 -78.739 50.905
73226 94.744 -78.756 50.865
73227 94.742 -78.764 50.823
73228 94.761 -78.751 50.780
73229 94.778 -78.739 50.735
73230 94.787 -78.730 50.692
73231 94.796 -78.722 50.651
73232 94.815 -78.719 50.610
73233 94.833 -78.715 50.569
73234 94.851 -78.712 50.528
73235 94.870 -78.709 50.485
73236 94.880 -78.713 50.440
73237 94.879 -78.729 50.397
73238 94.878 -78.744 50.355
73239 94.876 -78.760 50.315
73240 94.875 -78.775 50.278
73241 94.874 -78.790 50.241
73242 94.873 -78.806 50.204
73243 94.872 -78.821 50.167
73244 94.871 -78.837 50.130
73245 94.868 -78.851 50.092
73246 94.860 -78.851 50.052
73247 94.853 -78.850 50.014
73248 94.858 -78.852 49.975
73249 94.875 -78.855 49.935
73250 94.892 -78.858 49.895
73251 94.909 -78.861 49.855
73252 94.925 -78.864 49.817
73253 94.947 -78.864 49.781
73254 94.975 -78.859 49.745
73255 95.004 -78.855 49.708
73256 95.033 -78.850 49.670
73257 95.061 -78.846 49.632
73258 95.091 -78.849 49.591
73259 95.120 -78.855 49.550
73260 95.149 -78.861 49.509
73261 95.179 -78.867 49.467
73262 95.208 -78.872 49.426
73263 95.237 -78.878 49.386
73264 95.266 -78.883 49.346
73265 95.296 -78.890 49.307
73266 95.325 -78.895 49.268
73267 95.354 -78.901 49.230
73268 95.384 -78.906 49.190
73269 95.405 -78.912 49.149
73270 95.414 -78.918 49.106
73271 95.422 -78.924 49.063
73272 95.430 -78.930 49.022
73273 95.439 -78.935 48.981
73274 95.448 -78.941 48.939
73275 95.456 -78.947 48.897
73276 95.465 -78.952 48.855
73277 95.471 -78.968 48.813
73278 95.475 -78.988 48.771
73279 95.480 -79.008 48.729
73280 95.484 -79.028 48.686
73281 95.489 -79.048 48.640
73282 95.493 -79.068 48.598
73283 95.498 -79.088 48.556
73284 95.502 -79.109 48.513
73285 95.516 -79.123 48.469
73286 95.531 -79.137 48.425
73287 95.546 -79.150 48.380
73288 95.562 -79.160 48.334
73289 95.585 -79.147 48.286
73290 95.607 -79.134 48.239
73291 95.629 -79.120 48.192
73292 95.652 -79.107 48.146
73293 95.674 -79.094 48.100
73294 95.697 -79.081 48.055
73295 95.719 -79.068 48.009
73296 95.733 -79.061 47.964
73297 95.742 -79.057 47.920
73298 95.750 -79.053 47.877
73299 95.756 -79.049 47.836
73300 95.763 -79.046 47.795
73301 95.769 -79.043 47.753
73302 95.776 -79.039 47.711
73303 95.782 -79.035 47.667
73304 95.789 -79.031 47.625
73305 95.795 -79.028 47.583
73306 95.805 -79.027 47.543
73307 95.817 -79.028 47.502
73308 95.830 -79.028 47.462
73309 95.842 -79.029 47.422
73310 95.857 -79.038 47.382
73311 95.877 -79.043 47.343
73312 95.943 -79.061 47.304
73313 96.030 -79.071 47.266
73314 96.079 -79.073 47.228
73315 96.118 -79.076 47.190
73316 96.156 -79.073 47.153
73317 96.190 -79.060 47.118
73318 96.225 -79.047 47.083
73319 96.259 -79.035 47.047
73320 96.294 -79.022 47.013
73321 96.328 -79.009 46.981
73322 96.362 -78.996 46.949
73323 96.391 -79.033 46.918
73324 96.419 -79.078 46.887
73325 96.444 -79.111 46.856
73326 96.459 -79.112 46.824
73327 96.474 -79.112 46.792
73328 96.488 -79.112 46.760
73329 96.508 -79.119 46.729
73330 96.532 -79.131 46.697
73331 96.555 -79.143 46.666
73332 96.579 -79.154 46.635
73333 96.603 -79.166 46.603
73334 96.626 -79.178 46.572
73335 96.633 -79.170 46.543
73336 96.631 -79.153 46.515
73337 96.629 -79.137 46.488
73338 96.627 -79.120 46.460
73339 96.662 -79.119 46.433
73340 96.698 -79.120 46.406
73341 96.735 -79.121 46.379
73342 96.771 -79.122 46.353
73343 96.808 -79.123 46.328
73344 96.838 -79.125 46.304
73345 96.861 -79.127 46.279
73346 96.885 -79.118 46.254
73347 96.865 -79.162 46.229
73348 96.840 -79.209 46.206
73349 96.874 -79.175 46.183
73350 96.909 -79.139 46.160
73351 96.945 -79.103 46.135
73352 96.980 -79.067 46.110
73353 97.018 -79.057 46.086
73354 97.060 -79.102 46.065
73355 97.102 -79.147 46.045
73356 97.098 -79.148 46.026
73357 97.094 -79.150 46.006
73358 97.090 -79.150 45.986
73359 97.085 -79.151 45.967
73360 97.074 -79.146 45.948
73361 97.076 -79.167 45.929
73362 97.079 -79.175 45.910
73363 97.082 -79.184 45.892
73364 97.088 -79.173 45.874
73365 97.104 -79.159 45.856
73366 97.126 -79.140 45.837
73367 97.149 -79.118 45.818
73368 97.171 -79.127 45.800
73369 97.191 -79.244 45.782
73370 97.196 -79.235 45.766
73371 97.197 -79.145 45.750
73372 97.170 -79.119 45.732
73373 97.149 -79.137 45.714
73374 97.176 -79.153 45.696
73375 97.203 -79.150 45.679
73376 97.240 -79.148 45.662
73377 97.278 -79.147 45.646
73378 97.312 -79.147 45.630
73379 97.326 -79.164 45.615
73380 97.341 -79.181 45.599
73381 97.355 -79.198 45.583
73382 97.369 -79.215 45.567
73383 97.375 -79.235 45.551
73384 97.320 -79.280 45.537
73385 97.304 -79.297 45.521
73386 97.319 -79.291 45.506
73387 97.334 -79.285 45.491
73388 97.350 -79.278 45.477
73389 97.369 -79.270 45.462
73390 97.387 -79.266 45.447
73391 97.414 -79.277 45.434
73392 97.442 -79.289 45.419
73393 97.462 -79.320 45.405
73394 97.467 -79.328 45.391
73395 97.473 -79.336 45.376
73396 97.478 -79.344 45.362
73397 97.484 -79.352 45.349
73398 97.491 -79.360 45.335
73399 97.495 -79.356 45.322
73400 97.499 -79.352 45.309
73401 97.503 -79.347 45.297
73402 97.507 -79.343 45.284
73403 97.511 -79.338 45.271
73404 97.515 -79.334 45.259
73405 97.519 -79.330 45.247
73406 97.523 -79.325 45.234
73407 97.527 -79.321 45.222
73408 97.540 -79.342 45.210
73409 97.557 -79.372 45.198
73410 97.576 -79.407 45.187
73411 97.595 -79.441 45.176
73412 97.614 -79.476 45.164
73413 97.634 -79.520 45.153
73414 97.648 -79.607 45.142
73415 97.662 -79.607 45.130
73416 97.676 -79.609 45.119
73417 97.691 -79.610 45.107
73418 97.705 -79.611 45.096
73419 97.719 -79.612 45.084
73420 97.736 -79.626 45.072
73421 97.753 -79.642 45.060
73422 97.771 -79.657 45.047
73423 97.788 -79.673 45.035
73424 97.805 -79.689 45.022
73425 97.822 -79.705 45.010
73426 97.840 -79.721 44.999
73427 97.857 -79.737 44.988
73428 97.874 -79.754 44.977
73429 97.892 -79.770 44.967
73430 97.909 -79.786 44.958
73431 97.926 -79.802 44.948
73432 97.949 -79.816 44.939
73433 97.973 -79.830 44.931
73434 97.997 -79.845 44.922
73435 98.020 -79.859 44.913
73436 98.043 -79.874 44.906
73437 98.066 -79.889 44.899
73438 98.090 -79.902 44.892
73439 98.116 -79.914 44.886
73440 98.143 -79.927 44.880
73441 98.171 -79.935 44.874
73442 98.184 -79.936 44.868
73443 98.199 -79.939 44.862
73444 98.222 -79.948 44.855
73445 98.246 -79.955 44.847
73446 98.270 -79.964 44.838
73447 98.295 -79.969 44.828
73448 98.324 -79.977 44.819
73449 98.362 -79.989 44.811
73450 98.406 -79.999 44.804
73451 98.458 -80.009 44.798
73452 98.501 -80.021 44.794
73453 98.528 -80.036 44.790
73454 98.556 -80.051 44.783
73455 98.585 -80.059 44.775
73456 98.614 -80.062 44.765
73457 98.644 -80.065 44.755
73458 98.673 -80.068 44.748
73459 98.703 -80.073 44.743
73460 98.723 -80.083 44.737
73461 98.735 -80.098 44.727
73462 98.748 -80.112 44.717
73463 98.761 -80.124 44.706
73464 98.779 -80.137 44.693
73465 98.805 -80.147 44.682
73466 98.833 -80.156 44.672
73467 98.861 -80.145 44.662
73468 98.875 -80.131 44.653
73469 98.889 -80.135 44.645
73470 98.905 -80.146 44.638
73471 98.922 -80.157 44.631
73472 98.938 -80.168 44.624
73473 98.955 -80.180 44.617
73474 98.971 -80.191 44.610
73475 99.001 -80.199 44.604
73476 99.075 -80.227 44.599
73477 99.098 -80.237 44.594
73478 99.122 -80.248 44.588
73479 99.145 -80.259 44.582
73480 99.168 -80.269 44.577
73481 99.192 -80.279 44.572
73482 99.215 -80.290 44.568
73483 99.238 -80.300 44.563
73484 99.262 -80.311 44.559
73485 99.285 -80.322 44.556
73486 99.308 -80.332 44.552
73487 99.347 -80.336 44.549
73488 99.385 -80.340 44.546
73489 99.424 -80.344 44.542
73490 99.463 -80.348 44.538
73491 99.501 -80.352 44.535
73492 99.540 -80.356 44.531
73493 99.561 -80.369 44.527
73494 99.564 -80.391 44.524
73495 99.567 -80.413 44.519
73496 99.570 -80.435 44.515
73497 99.573 -80.457 44.510
73498 99.577 -80.479 44.505
73814 93.523 -78.941 53.490
73815 93.533 -78.939 53.479
73816 93.543 -78.937 53.470
73817 93.553 -78.930 53.461
73818 93.564 -78.921 53.452
73819 93.572 -78.914 53.443
73820 93.581 -78.907 53.435
73821 93.589 -78.900 53.426
73822 93.597 -78.893 53.418
73823 93.606 -78.886 53.410
73824 93.619 -78.877 53.403
73825 93.688 -78.843 53.397
73826 93.703 -78.826 53.391
73827 93.699 -78.820 53.386
73828 93.695 -78.814 53.383
73829 93.691 -78.808 53.380
73830 93.687 -78.802 53.377
73831 93.694 -78.805 53.372
73832 93.707 -78.818 53.367
73833 93.720 -78.830 53.361
73834 93.728 -78.831 53.352
73835 93.735 -78.829 53.343
73836 93.742 -78.826 53.335
73837 93.749 -78.823 53.325
73838 93.756 -78.821 53.313
73839 93.763 -78.818 53.300
73840 93.770 -78.814 53.288
73841 93.777 -78.811 53.275
73842 93.784 -78.809 53.262
73843 93.791 -78.806 53.248
73844 93.807 -78.797 53.235
73845 93.813 -78.789 53.216
73846 93.827 -78.781 53.197
73847 93.837 -78.774 53.177
73848 93.843 -78.770 53.157
73849 93.848 -78.765 53.136
73850 93.853 -78.761 53.114
73851 93.858 -78.757 53.093
73852 93.863 -78.752 53.069
73853 93.869 -78.748 53.045
73854 93.874 -78.743 53.020
73855 93.880 -78.739 52.995
73856 93.885 -78.734 52.969
73857 93.898 -78.729 52.943
73858 93.914 -78.722 52.918
73859 93.930 -78.716 52.890
73860 93.945 -78.710 52.864
73861 93.962 -78.703 52.839
73862 93.977 -78.697 52.813
73863 93.993 -78.691 52.788
73864 94.009 -78.685 52.764
73865 94.025 -78.678 52.742
73866 94.041 -78.672 52.718
73867 94.057 -78.666 52.696
73868 94.073 -78.659 52.675
73869 94.088 -78.653 52.653
73870 94.097 -78.655 52.631
73871 94.101 -78.662 52.608
73872 94.105 -78.670 52.585
73873 94.109 -78.678 52.561
73874 94.114 -78.686 52.537
73875 94.118 -78.694 52.512
73876 94.135 -78.665 52.488
73877 94.174 -78.658 52.463
73878 94.225 -78.671 52.438
73879 94.255 -78.677 52.414
73880 94.265 -78.673 52.389
73881 94.275 -78.671 52.364
73882 94.285 -78.668 52.339
73883 94.294 -78.666 52.315
73884 94.296 -78.670 52.292
73885 94.288 -78.682 52.269
73886 94.293 -78.681 52.244
73887 94.308 -78.671 52.215
73888 94.317 -78.669 52.186
73889 94.325 -78.667 52.159
73890 94.334 -78.664 52.132
73891 94.343 -78.662 52.105
73892 94.351 -78.660 52.079
73893 94.360 -78.658 52.054
73894 94.368 -78.655 52.026
73895 94.377 -78.653 51.998
73896 94.391 -78.652 51.969
73897 94.403 -78.658 51.939
73898 94.413 -78.662 51.907
73899 94.424 -78.667 51.874
73900 94.435 -78.673 51.842
73901 94.445 -78.678 51.811
73902 94.458 -78.683 51.781
73903 94.478 -78.688 51.750
73904 94.498 -78.691 51.716
73905 94.516 -78.693 51.680
73906 94.534 -78.695 51.642
73907 94.552 -78.697 51.604
73908 94.570 -78.699 51.566
73909 94.589 -78.702 51.528
73910 94.607 -78.704 51.492
73911 94.625 -78.706 51.458
73912 94.643 -78.708 51.425
73913 94.661 -78.711 51.392
73914 94.679 -78.712 51.360
73915 94.697 -78.714 51.330
73916 94.714 -78.717 51.299
73917 94.725 -78.719 51.266
73918 94.736 -78.721 51.232
73919 94.747 -78.724 51.197
73920 94.757 -78.724 51.159
73921 94.762 -78.714 51.119
73922 94.768 -78.703 51.079
73923 94.773 -78.696 51.038
73924 94.776 -78.705 50.998
73925 94.780 -78.714 50.957
73926 94.773 -78.728 50.918
73927 94.761 -78.745 50.879
73928 94.758 -78.754 50.837
73929 94.776 -78.741 50.793
73930 94.785 -78.733 50.748
73931 94.794 -78.724 50.705
73932 94.803 -78.716 50.663
73933 94.813 -78.708 50.623
73934 94.831 -78.705 50.583
73935 94.849 -78.701 50.540
73936 94.867 -78.698 50.497
73937 94.879 -78.702 50.452
73938 94.877 -78.717 50.408
73939 94.876 -78.733 50.365
73940 94.875 -78.748 50.323
73941 94.874 -78.764 50.285
73942 94.873 -78.779 50.247
73943 94.872 -78.794 50.209
73944 94.870 -78.810 50.172
73945 94.869 -78.825 50.134
73946 94.867 -78.841 50.095
73947 94.865 -78.853 50.056
73948 94.857 -78.853 50.017
73949 94.859 -78.854 49.977
73950 94.876 -78.857 49.938
73951 94.893 -78.860 49.899
73952 94.910 -78.863 49.860
73953 94.935 -78.860 49.822
73954 94.964 -78.855 49.785
73955 94.992 -78.851 49.748
73956 95.021 -78.846 49.710
73957 95.050 -78.842 49.672
73958 95.078 -78.837 49.632
73959 95.107 -78.835 49.591
73960 95.137 -78.840 49.550
73961 95.166 -78.846 49.510
73962 95.195 -78.852 49.470
73963 95.225 -78.858 49.430
73964 95.254 -78.863 49.389
73965 95.283 -78.869 49.349
73966 95.313 -78.874 49.308
73967 95.342 -78.880 49.268
73968 95.370 -78.886 49.227
73969 95.389 -78.899 49.189
73970 95.407 -78.905 49.149
73971 95.421 -78.903 49.109
73972 95.431 -78.907 49.067
73973 95.440 -78.912 49.026
73974 95.448 -78.918 48.984
73975 95.457 -78.924 48.941
73976 95.465 -78.930 48.900
73977 95.473 -78.936 48.857
73978 95.481 -78.946 48.815
73979 95.485 -78.966 48.774
73980 95.490 -78.987 48.732
73981 95.494 -79.006 48.689
73982 95.499 -79.027 48.646
73983 95.503 -79.046 48.606
73984 95.509 -79.067 48.565
73985 95.521 -79.082 48.522
73986 95.536 -79.095 48.478
73987 95.551 -79.109 48.433
73988 95.566 -79.122 48.387
73989 95.582 -79.134 48.341
73990 95.599 -79.140 48.293
73991 95.621 -79.127 48.245
73992 95.644 -79.114 48.199
73993 95.666 -79.100 48.153
73994 95.689 -79.087 48.107
73995 95.711 -79.074 48.061
73996 95.731 -79.062 48.014
73997 95.740 -79.059 47.970
73998 95.749 -79.055 47.927
73999 95.758 -79.051 47.884
74000 95.766 -79.048 47.842
74001 95.772 -79.044 47.800
74002 95.779 -79.040 47.758
74003 95.785 -79.037 47.715
74004 95.792 -79.033 47.672
74005 95.798 -79.030 47.630
74006 95.805 -79.026 47.588
74007 95.811 -79.022 47.547
74008 95.817 -79.018 47.507
74009 95.824 -79.015 47.466
74010 95.830 -79.012 47.426
74011 95.838 -79.009 47.386
74012 95.877 -79.019 47.347
74013 95.944 -79.037 47.306
74014 96.017 -79.053 47.268
74015 96.081 -79.055 47.230
74016 96.120 -79.058 47.193
74017 96.160 -79.060 47.157
74018 96.198 -79.059 47.122
74019 96.233 -79.046 47.086
74020 96.267 -79.034 47.051
74021 96.302 -79.021 47.016
74022 96.336 -79.008 46.984
74023 96.370 -78.995 46.953
74024 96.399 -79.032 46.922
74025 96.428 -79.078 46.891
74026 96.446 -79.090 46.859
74027 96.461 -79.091 46.827
74028 96.476 -79.091 46.795
74029 96.496 -79.099 46.763
74030 96.520 -79.110 46.732
74031 96.543 -79.122 46.699
74032 96.567 -79.134 46.668
74033 96.590 -79.146 46.637
74034 96.614 -79.158 46.605
74035 96.638 -79.169 46.575
74036 96.654 -79.172 46.547
74037 96.652 -79.156 46.519
74038 96.650 -79.138 46.491
74039 96.648 -79.122 46.463
74040 96.646 -79.105 46.435
74041 96.670 -79.100 46.408
74042 96.707 -79.100 46.381
74043 96.743 -79.101 46.355
74044 96.780 -79.102 46.331
74045 96.816 -79.103 46.307
74046 96.863 -79.106 46.282
74047 96.838 -79.157 46.258
74048 96.813 -79.204 46.233
74049 96.844 -79.175 46.210
74050 96.879 -79.139 46.187
74051 96.914 -79.103 46.162
74052 96.950 -79.066 46.138
74053 96.985 -79.030 46.114
74054 97.025 -79.052 46.092
74055 97.068 -79.096 46.070
74056 97.105 -79.137 46.050
74057 97.101 -79.138 46.031
74058 97.097 -79.140 46.013
74059 97.092 -79.140 45.993
74060 97.060 -79.115 45.974
74061 97.068 -79.181 45.954
74062 97.064 -79.186 45.935
74063 97.060 -79.190 45.917
74064 97.060 -79.197 45.899
74065 97.068 -79.172 45.881
74066 97.084 -79.156 45.863
74067 97.107 -79.134 45.844
74068 97.130 -79.113 45.825
74069 97.153 -79.092 45.807
74070 97.174 -79.126 45.790
74071 97.173 -79.229 45.774
74072 97.225 -79.238 45.757
74073 97.277 -79.247 45.740
74074 97.233 -79.188 45.722
74075 97.165 -79.149 45.705
74076 97.193 -79.154 45.688
74077 97.231 -79.153 45.671
74078 97.267 -79.151 45.654
74079 97.305 -79.150 45.638
74080 97.342 -79.148 45.622
74081 97.358 -79.163 45.606
74082 97.372 -79.181 45.591
74083 97.386 -79.197 45.575
74084 97.365 -79.229 45.559
74085 97.311 -79.274 45.545
74086 97.307 -79.281 45.530
74087 97.323 -79.275 45.515
74088 97.338 -79.269 45.501
74089 97.354 -79.262 45.486
74090 97.364 -79.258 45.471
74091 97.397 -79.269 45.457
74092 97.425 -79.281 45.443
74093 97.452 -79.293 45.429
74094 97.461 -79.307 45.415
74095 97.460 -79.318 45.401
74096 97.454 -79.331 45.385
74097 97.463 -79.338 45.371
74098 97.477 -79.344 45.358
74099 97.491 -79.350 45.345
74100 97.508 -79.361 45.332
74101 97.522 -79.369 45.319
74102 97.526 -79.365 45.306
74103 97.530 -79.360 45.294
74104 97.534 -79.356 45.281
74105 97.538 -79.351 45.268
74106 97.542 -79.347 45.256
74107 97.546 -79.343 45.243
74108 97.550 -79.338 45.231
74109 97.564 -79.362 45.219
74110 97.579 -79.390 45.208
74111 97.594 -79.417 45.197
74112 97.608 -79.445 45.186
74113 97.624 -79.473 45.174
74114 97.639 -79.501 45.165
74115 97.647 -79.513 45.154
74116 97.658 -79.598 45.141
74117 97.672 -79.648 45.129
74118 97.684 -79.654 45.117
74119 97.698 -79.662 45.106
74120 97.710 -79.669 45.094
74121 97.723 -79.676 45.082
74122 97.736 -79.683 45.070
74123 97.748 -79.691 45.058
74124 97.761 -79.697 45.046
74125 97.776 -79.698 45.034
74126 97.790 -79.699 45.023
74127 97.806 -79.712 45.013
74128 97.823 -79.728 45.002
74129 97.845 -79.742 44.991
74130 97.868 -79.758 44.981
74131 97.891 -79.772 44.972
74132 97.915 -79.786 44.962
74133 97.938 -79.801 44.954
74134 97.961 -79.816 44.946
74135 97.985 -79.830 44.938
74136 98.011 -79.842 44.931
74137 98.037 -79.855 44.925
74138 98.064 -79.867 44.920
74139 98.090 -79.879 44.914
74140 98.113 -79.886 44.908
74141 98.137 -79.894 44.901
74142 98.160 -79.902 44.894
74143 98.184 -79.911 44.887
74144 98.208 -79.919 44.880
74145 98.231 -79.927 44.872
74146 98.256 -79.933 44.862
74147 98.282 -79.939 44.852
74148 98.307 -79.944 44.841
74149 98.335 -79.951 44.832
74150 98.373 -79.963 44.825
74151 98.411 -79.974 44.820
74152 98.449 -79.987 44.815
74153 98.493 -79.997 44.809
74154 98.534 -80.009 44.804
74155 98.562 -80.024 44.797
74156 98.591 -80.027 44.787
74157 98.621 -80.030 44.776
74158 98.650 -80.035 44.764
74159 98.679 -80.040 44.756
74160 98.708 -80.045 44.749
74161 98.719 -80.060 44.742
74162 98.731 -80.074 44.732
74163 98.743 -80.090 44.721
74164 98.755 -80.105 44.711
74165 98.786 -80.108 44.700
74166 98.821 -80.108 44.691
74167 98.857 -80.107 44.682
74168 98.882 -80.117 44.673
74169 98.896 -80.118 44.664
74170 98.905 -80.116 44.656
74171 98.921 -80.121 44.649
74172 98.936 -80.127 44.641
74173 98.951 -80.136 44.634
74174 98.968 -80.147 44.627
74175 98.984 -80.158 44.620
74176 99.001 -80.169 44.615
74177 99.017 -80.181 44.610
74178 99.075 -80.197 44.604
74179 99.126 -80.221 44.599
74180 99.149 -80.231 44.593
74181 99.172 -80.242 44.587
74182 99.196 -80.252 44.582
74183 99.219 -80.262 44.579
74184 99.242 -80.273 44.575
74185 99.266 -80.284 44.570
74186 99.289 -80.294 44.566
74187 99.320 -80.302 44.562
74188 99.358 -80.306 44.559
74189 99.397 -80.309 44.555
74190 99.435 -80.314 44.551
74191 99.474 -80.318 44.548
74192 99.512 -80.322 44.545
74193 99.551 -80.325 44.541
74194 99.557 -80.346 44.537
74195 99.560 -80.368 44.532
74196 99.563 -80.391 44.528
74197 99.567 -80.413 44.523
74198 99.570 -80.435 44.519
74199 99.573 -80.457 44.513
74200 99.576 -80.478 44.508
74516 93.534 -78.928 53.494
74517 93.543 -78.925 53.484
74518 93.553 -78.923 53.474
74519 93.563 -78.916 53.465
74520 93.573 -78.907 53.456
74521 93.582 -78.900 53.448
74522 93.590 -78.893 53.439
74523 93.599 -78.886 53.431
74524 93.607 -78.880 53.423
74525 93.615 -78.873 53.416
74526 93.677 -78.842 53.410
74527 93.736 -78.812 53.404
74528 93.725 -78.809 53.400
74529 93.714 -78.805 53.397
74530 93.704 -78.802 53.394
74531 93.693 -78.799 53.392
74532 93.693 -78.804 53.388
74533 93.706 -78.817 53.383
74534 93.717 -78.824 53.376
74535 93.724 -78.824 53.367
74536 93.731 -78.821 53.358
74537 93.738 -78.818 53.349
74538 93.745 -78.816 53.338
74539 93.752 -78.813 53.327
74540 93.759 -78.810 53.314
74541 93.766 -78.808 53.300
74542 93.773 -78.805 53.288
74543 93.780 -78.802 53.273
74544 93.795 -78.794 53.258
74545 93.809 -78.783 53.244
74546 93.813 -78.774 53.225
74547 93.826 -78.767 53.205
74548 93.840 -78.758 53.184
74549 93.854 -78.750 53.165
74550 93.860 -78.745 53.143
74551 93.865 -78.741 53.122
74552 93.871 -78.736 53.101
74553 93.876 -78.732 53.076
74554 93.882 -78.727 53.051
74555 93.887 -78.724 53.025
74556 93.892 -78.719 53.000
74557 93.898 -78.714 52.976
74558 93.913 -78.708 52.951
74559 93.929 -78.702 52.926
74560 93.945 -78.695 52.898
74561 93.961 -78.689 52.872
74562 93.977 -78.683 52.848
74563 93.993 -78.677 52.822
74564 94.009 -78.670 52.798
74565 94.025 -78.664 52.774
74566 94.041 -78.658 52.752
74567 94.056 -78.651 52.730
74568 94.069 -78.648 52.708
74569 94.079 -78.646 52.687
74570 94.089 -78.645 52.665
74571 94.091 -78.648 52.642
74572 94.095 -78.655 52.620
74573 94.099 -78.664 52.596
74574 94.103 -78.671 52.572
74575 94.107 -78.679 52.547
74576 94.126 -78.647 52.522
74577 94.158 -78.630 52.498
74578 94.209 -78.643 52.474
74579 94.261 -78.656 52.450
74580 94.278 -78.656 52.427
74581 94.288 -78.654 52.403
74582 94.298 -78.651 52.378
74583 94.308 -78.648 52.355
74584 94.310 -78.652 52.332
74585 94.302 -78.664 52.308
74586 94.294 -78.676 52.285
74587 94.295 -78.680 52.260
74588 94.312 -78.667 52.233
74589 94.326 -78.658 52.204
74590 94.335 -78.655 52.178
74591 94.343 -78.653 52.151
74592 94.352 -78.651 52.125
74593 94.360 -78.648 52.098
74594 94.369 -78.646 52.072
74595 94.378 -78.644 52.045
74596 94.387 -78.642 52.017
74597 94.400 -78.642 51.987
74598 94.415 -78.644 51.957
74599 94.425 -78.649 51.925
74600 94.436 -78.655 51.892
74601 94.447 -78.660 51.859
74602 94.458 -78.665 51.829
74603 94.468 -78.672 51.799
74604 94.482 -78.683 51.767
74605 94.497 -78.683 51.733
74606 94.511 -78.683 51.696
74607 94.527 -78.682 51.658
74608 94.542 -78.681 51.619
74609 94.557 -78.680 51.580
74610 94.574 -78.682 51.542
74611 94.592 -78.684 51.506
74612 94.610 -78.686 51.472
74613 94.628 -78.688 51.439
74614 94.646 -78.690 51.406
74615 94.665 -78.692 51.376
74616 94.683 -78.695 51.345
74617 94.699 -78.697 51.314
74618 94.710 -78.699 51.282
74619 94.720 -78.702 51.249
74620 94.732 -78.704 51.214
74621 94.742 -78.706 51.177
74622 94.751 -78.704 51.136
74623 94.757 -78.693 51.095
74624 94.760 -78.702 51.053
74625 94.764 -78.710 51.012
74626 94.767 -78.718 50.971
74627 94.771 -78.727 50.930
74628 94.774 -78.736 50.892
74629 94.773 -78.744 50.850
74630 94.783 -78.735 50.807
74631 94.792 -78.727 50.762
74632 94.801 -78.717 50.718
74633 94.810 -78.709 50.676
74634 94.819 -78.701 50.636
74635 94.829 -78.693 50.596
74636 94.847 -78.690 50.552
74637 94.865 -78.686 50.508
74638 94.877 -78.690 50.463
74639 94.876 -78.705 50.419
74640 94.875 -78.720 50.375
74641 94.874 -78.736 50.332
74642 94.873 -78.752 50.292
74643 94.871 -78.767 50.253
74644 94.870 -78.783 50.214
74645 94.868 -78.798 50.177
74646 94.867 -78.814 50.138
74647 94.866 -78.829 50.099
74648 94.865 -78.845 50.059
74649 94.862 -78.855 50.019
74650 94.860 -78.856 49.980
74651 94.877 -78.859 49.941
74652 94.895 -78.861 49.902
74653 94.923 -78.856 49.863
74654 94.952 -78.852 49.824
74655 94.981 -78.847 49.786
74656 95.010 -78.843 49.749
74657 95.038 -78.838 49.710
74658 95.067 -78.833 49.670
74659 95.096 -78.829 49.629
74660 95.124 -78.824 49.588
74661 95.154 -78.826 49.548
74662 95.183 -78.831 49.509
74663 95.212 -78.837 49.469
74664 95.242 -78.843 49.430
74665 95.271 -78.849 49.389
74666 95.300 -78.854 49.349
74667 95.329 -78.860 49.307
74668 95.350 -78.871 49.266
74669 95.370 -78.882 49.224
74670 95.390 -78.894 49.186
74671 95.407 -78.902 49.148
74672 95.421 -78.899 49.108
74673 95.435 -78.897 49.067
74674 95.449 -78.896 49.026
74675 95.457 -78.902 48.984
74676 95.466 -78.907 48.941
74677 95.474 -78.913 48.900
74678 95.482 -78.919 48.858
74679 95.491 -78.925 48.816
74680 95.495 -78.945 48.775
74681 95.500 -78.965 48.734
74682 95.505 -78.985 48.692
74683 95.510 -79.005 48.651
74684 95.514 -79.025 48.612
74685 95.526 -79.040 48.571
74686 95.541 -79.053 48.529
74687 95.556 -79.067 48.484
74688 95.571 -79.080 48.439
74689 95.587 -79.093 48.393
74690 95.602 -79.106 48.346
74691 95.617 -79.120 48.298
74692 95.636 -79.120 48.251
74693 95.659 -79.107 48.204
74694 95.681 -79.093 48.158
74695 95.703 -79.081 48.111
74696 95.726 -79.068 48.065
74697 95.739 -79.061 48.020
74698 95.748 -79.057 47.977
74699 95.757 -79.053 47.934
74700 95.766 -79.049 47.891
74701 95.775 -79.046 47.849
74702 95.782 -79.042 47.806
74703 95.789 -79.038 47.763
74704 95.795 -79.035 47.720
74705 95.801 -79.031 47.677
74706 95.808 -79.028 47.635
74707 95.814 -79.024 47.594
74708 95.821 -79.021 47.553
74709 95.827 -79.017 47.512
74710 95.833 -79.013 47.472
74711 95.840 -79.009 47.431
74712 95.851 -79.006 47.391
74713 95.888 -79.002 47.351
74714 95.946 -79.014 47.310
74715 96.012 -79.031 47.272
74716 96.083 -79.037 47.234
74717 96.123 -79.040 47.197
74718 96.163 -79.042 47.161
74719 96.202 -79.045 47.125
74720 96.241 -79.045 47.090
74721 96.275 -79.032 47.054
74722 96.310 -79.019 47.020
74723 96.344 -79.006 46.988
74724 96.378 -78.993 46.957
74725 96.407 -79.031 46.927
74726 96.434 -79.069 46.896
74727 96.448 -79.070 46.864
74728 96.463 -79.070 46.832
74729 96.484 -79.078 46.799
74730 96.508 -79.090 46.767
74731 96.531 -79.102 46.736
74732 96.554 -79.114 46.703
74733 96.578 -79.125 46.672
74734 96.602 -79.137 46.641
74735 96.625 -79.149 46.609
74736 96.649 -79.161 46.580
74737 96.673 -79.172 46.551
74738 96.672 -79.157 46.524
74739 96.671 -79.140 46.495
74740 96.669 -79.123 46.467
74741 96.667 -79.106 46.439
74742 96.665 -79.090 46.411
74743 96.679 -79.080 46.384
74744 96.716 -79.081 46.358
74745 96.768 -79.085 46.334
74746 96.827 -79.090 46.310
74747 96.885 -79.095 46.285
74748 96.832 -79.167 46.261
74749 96.814 -79.175 46.236
74750 96.849 -79.138 46.213
74751 96.884 -79.103 46.189
74752 96.920 -79.066 46.165
74753 96.955 -79.030 46.141
74754 96.991 -79.001 46.119
74755 97.033 -79.046 46.097
74756 97.075 -79.091 46.076
74757 97.108 -79.128 46.056
74758 97.104 -79.128 46.038
74759 97.085 -79.115 46.020
74760 97.050 -79.087 46.001
74761 97.020 -79.062 45.981
74762 97.058 -79.201 45.961
74763 97.054 -79.206 45.942
74764 97.050 -79.210 45.924
74765 97.049 -79.194 45.906
74766 97.049 -79.171 45.889
74767 97.065 -79.150 45.871
74768 97.066 -79.140 45.853
74769 97.076 -79.130 45.834
74770 97.087 -79.118 45.816
74771 97.097 -79.107 45.799
74772 97.119 -79.096 45.783
74773 97.169 -79.121 45.767
74774 97.219 -79.146 45.749
74775 97.268 -79.171 45.732
74776 97.221 -79.158 45.715
74777 97.187 -79.158 45.697
74778 97.221 -79.157 45.680
74779 97.258 -79.156 45.664
74780 97.295 -79.154 45.647
74781 97.332 -79.153 45.631
74782 97.369 -79.151 45.615
74783 97.389 -79.163 45.599
74784 97.403 -79.180 45.583
74785 97.355 -79.222 45.568
74786 97.301 -79.268 45.554
74787 97.311 -79.265 45.539
74788 97.327 -79.259 45.525
74789 97.331 -79.256 45.510
74790 97.330 -79.258 45.495
74791 97.354 -79.259 45.480
74792 97.394 -79.263 45.466
74793 97.412 -79.279 45.452
74794 97.424 -79.300 45.439
74795 97.423 -79.311 45.424
74796 97.436 -79.317 45.410
74797 97.450 -79.322 45.395
74798 97.463 -79.328 45.380
74799 97.477 -79.334 45.367
74800 97.491 -79.340 45.354
74801 97.506 -79.348 45.341
74802 97.524 -79.362 45.328
74803 97.543 -79.376 45.315
74804 97.552 -79.378 45.303
74805 97.557 -79.373 45.290
74806 97.561 -79.369 45.277
74807 97.565 -79.365 45.264
74808 97.569 -79.360 45.252
74809 97.573 -79.356 45.241
74810 97.587 -79.382 45.229
74811 97.602 -79.410 45.217
74812 97.617 -79.438 45.206
74813 97.632 -79.465 45.196
74814 97.645 -79.488 45.185
74815 97.651 -79.494 45.175
74816 97.657 -79.499 45.165
74817 97.668 -79.578 45.154
74818 97.683 -79.623 45.142
74819 97.696 -79.630 45.130
74820 97.709 -79.638 45.119
74821 97.721 -79.645 45.108
74822 97.734 -79.651 45.096
74823 97.747 -79.659 45.085
74824 97.759 -79.666 45.073
74825 97.773 -79.673 45.061
74826 97.785 -79.680 45.049
74827 97.798 -79.688 45.037
74828 97.811 -79.697 45.026
74829 97.828 -79.709 45.016
74830 97.850 -79.723 45.006
74831 97.872 -79.738 44.997
74832 97.894 -79.753 44.987
74833 97.916 -79.768 44.978
74834 97.941 -79.789 44.970
74835 97.975 -79.818 44.962
74836 98.004 -79.833 44.956
74837 98.032 -79.849 44.949
74838 98.058 -79.856 44.943
74839 98.078 -79.852 44.938
74840 98.099 -79.853 44.933
74841 98.123 -79.863 44.928
74842 98.146 -79.873 44.921
74843 98.169 -79.882 44.913
74844 98.193 -79.889 44.905
74845 98.218 -79.896 44.897
74846 98.243 -79.902 44.887
74847 98.269 -79.908 44.876
74848 98.294 -79.913 44.865
74849 98.319 -79.919 44.852
74850 98.346 -79.926 44.845
74851 98.384 -79.937 44.839
74852 98.422 -79.949 44.835
74853 98.459 -79.961 44.831
74854 98.497 -79.972 44.824
74855 98.535 -79.984 44.816
74856 98.568 -79.992 44.808
74857 98.598 -79.996 44.797
74858 98.627 -80.002 44.784
74859 98.656 -80.006 44.771
74860 98.686 -80.011 44.761
74861 98.704 -80.021 44.752
74862 98.716 -80.037 44.745
74863 98.727 -80.052 44.737
74864 98.743 -80.065 44.727
74865 98.769 -80.073 44.717
74866 98.799 -80.077 44.708
74867 98.834 -80.077 44.700
74868 98.866 -80.080 44.691
74869 98.887 -80.089 44.683
74870 98.910 -80.099 44.674
74871 98.921 -80.098 44.666
74872 98.937 -80.102 44.659
74873 98.953 -80.108 44.652
74874 98.969 -80.112 44.645
74875 98.984 -80.120 44.638
74876 98.999 -80.128 44.632
74877 99.013 -80.137 44.626
74878 99.030 -80.148 44.620
74879 99.047 -80.159 44.615
74880 99.073 -80.167 44.610
74881 99.153 -80.204 44.605
74882 99.177 -80.215 44.599
74883 99.200 -80.225 44.593
74884 99.223 -80.235 44.589
74885 99.246 -80.246 44.585
74886 99.270 -80.256 44.580
74887 99.293 -80.267 44.577
74888 99.330 -80.272 44.573
74889 99.369 -80.275 44.569
74890 99.407 -80.279 44.565
74891 99.446 -80.284 44.562
74892 99.485 -80.287 44.558
74893 99.523 -80.291 44.555
74894 99.550 -80.301 44.550
74895 99.554 -80.323 44.546
74896 99.557 -80.345 44.541
74897 99.560 -80.367 44.536
74898 99.563 -80.390 44.531
74899 99.566 -80.412 44.526
74900 99.570 -80.434 44.522
74901 99.573 -80.456 44.516
74902 99.576 -80.478 44.510
75218 93.544 -78.915 53.497
75219 93.553 -78.912 53.488
75220 93.563 -78.910 53.478
75221 93.574 -78.902 53.469
75222 93.583 -78.894 53.461
75223 93.591 -78.886 53.452
75224 93.600 -78.880 53.443
75225 93.608 -78.873 53.437
75226 93.616 -78.866 53.429
75227 93.664 -78.842 53.423
75228 93.725 -78.810 53.417
75229 93.723 -78.808 53.414
75230 93.714 -78.805 53.412
75231 93.703 -78.802 53.410
75232 93.692 -78.799 53.407
75233 93.692 -78.803 53.404
75234 93.705 -78.816 53.398
75235 93.713 -78.817 53.391
75236 93.720 -78.816 53.382
75237 93.727 -78.814 53.373
75238 93.734 -78.811 53.363
75239 93.741 -78.808 53.352
75240 93.748 -78.805 53.341
75241 93.755 -78.803 53.327
75242 93.762 -78.800 53.314
75243 93.769 -78.797 53.301
75244 93.781 -78.791 53.286
75245 93.806 -78.777 53.270
75246 93.810 -78.768 53.254
75247 93.814 -78.760 53.235
75248 93.825 -78.752 53.214
75249 93.839 -78.744 53.193
75250 93.853 -78.736 53.172
75251 93.867 -78.728 53.151
75252 93.878 -78.721 53.130
75253 93.884 -78.717 53.109
75254 93.889 -78.712 53.085
75255 93.894 -78.708 53.060
75256 93.899 -78.703 53.034
75257 93.905 -78.698 53.009
75258 93.913 -78.694 52.984
75259 93.929 -78.687 52.958
75260 93.944 -78.681 52.933
75261 93.961 -78.675 52.907
75262 93.976 -78.669 52.881
75263 93.992 -78.662 52.855
75264 94.008 -78.656 52.831
75265 94.024 -78.650 52.808
75266 94.039 -78.645 52.787
75267 94.048 -78.643 52.765
75268 94.058 -78.642 52.743
75269 94.068 -78.639 52.721
75270 94.078 -78.638 52.700
75271 94.088 -78.636 52.677
75272 94.088 -78.642 52.654
75273 94.088 -78.649 52.631
75274 94.092 -78.657 52.608
75275 94.097 -78.664 52.583
75276 94.117 -78.629 52.558
75277 94.142 -78.601 52.533
75278 94.194 -78.614 52.509
75279 94.246 -78.628 52.486
75280 94.291 -78.639 52.463
75281 94.301 -78.636 52.440
75282 94.311 -78.633 52.417
75283 94.321 -78.631 52.393
75284 94.324 -78.634 52.371
75285 94.316 -78.646 52.349
75286 94.309 -78.658 52.326
75287 94.300 -78.670 52.302
75288 94.297 -78.677 52.278
75289 94.314 -78.665 52.251
75290 94.330 -78.653 52.224
75291 94.344 -78.644 52.197
75292 94.352 -78.642 52.172
75293 94.361 -78.639 52.145
75294 94.370 -78.637 52.117
75295 94.379 -78.636 52.091
75296 94.387 -78.633 52.064
75297 94.396 -78.631 52.036
75298 94.410 -78.630 52.005
75299 94.427 -78.631 51.974
75300 94.438 -78.636 51.942
75301 94.449 -78.642 51.909
75302 94.459 -78.647 51.876
75303 94.470 -78.651 51.846
75304 94.480 -78.675 51.815
75305 94.493 -78.688 51.781
75306 94.508 -78.687 51.746
75307 94.524 -78.686 51.710
75308 94.539 -78.686 51.672
75309 94.553 -78.686 51.632
75310 94.568 -78.685 51.593
75311 94.583 -78.684 51.555
75312 94.599 -78.683 51.519
75313 94.614 -78.683 51.484
75314 94.629 -78.682 51.452
75315 94.643 -78.681 51.420
75316 94.659 -78.680 51.390
75317 94.674 -78.680 51.361
75318 94.686 -78.679 51.331
75319 94.697 -78.681 51.297
75320 94.708 -78.683 51.262
75321 94.718 -78.686 51.226
75322 94.730 -78.688 51.189
75323 94.740 -78.690 51.149
75324 94.748 -78.695 51.109
75325 94.754 -78.703 51.067
75326 94.760 -78.711 51.026
75327 94.765 -78.718 50.984
75328 94.771 -78.726 50.942
75329 94.776 -78.733 50.903
75330 94.783 -78.733 50.862
75331 94.791 -78.724 50.819
75332 94.799 -78.720 50.774
75333 94.807 -78.711 50.730
75334 94.816 -78.703 50.689
75335 94.825 -78.695 50.649
75336 94.834 -78.686 50.608
75337 94.845 -78.679 50.564
75338 94.863 -78.676 50.519
75339 94.876 -78.678 50.474
75340 94.875 -78.694 50.429
75341 94.874 -78.709 50.385
75342 94.872 -78.724 50.342
75343 94.871 -78.740 50.302
75344 94.870 -78.755 50.261
75345 94.868 -78.771 50.221
75346 94.867 -78.786 50.183
75347 94.866 -78.802 50.144
75348 94.865 -78.818 50.105
75349 94.864 -78.833 50.064
75350 94.862 -78.849 50.023
75351 94.861 -78.858 49.984
75352 94.883 -78.858 49.945
75353 94.912 -78.852 49.905
75354 94.941 -78.848 49.863
75355 94.969 -78.843 49.823
75356 94.998 -78.839 49.784
75357 95.027 -78.834 49.747
75358 95.055 -78.830 49.707
75359 95.084 -78.825 49.666
75360 95.113 -78.821 49.625
75361 95.141 -78.816 49.585
75362 95.170 -78.811 49.545
75363 95.200 -78.817 49.504
75364 95.229 -78.822 49.466
75365 95.258 -78.828 49.426
75366 95.287 -78.833 49.386
75367 95.312 -78.843 49.346
75368 95.331 -78.854 49.304
75369 95.351 -78.866 49.263
75370 95.370 -78.877 49.221
75371 95.390 -78.890 49.183
75372 95.408 -78.898 49.144
75373 95.422 -78.896 49.104
75374 95.436 -78.893 49.064
75375 95.450 -78.891 49.023
75376 95.464 -78.888 48.982
75377 95.474 -78.890 48.941
75378 95.483 -78.896 48.899
75379 95.491 -78.902 48.858
75380 95.500 -78.908 48.817
75381 95.506 -78.923 48.776
75382 95.511 -78.943 48.734
75383 95.515 -78.963 48.694
75384 95.520 -78.984 48.655
75385 95.531 -78.999 48.615
75386 95.546 -79.012 48.576
75387 95.562 -79.025 48.535
75388 95.577 -79.039 48.490
75389 95.592 -79.052 48.444
75390 95.607 -79.065 48.397
75391 95.622 -79.078 48.350
75392 95.637 -79.092 48.303
75393 95.653 -79.105 48.256
75394 95.673 -79.100 48.209
75395 95.695 -79.087 48.161
75396 95.718 -79.074 48.115
75397 95.737 -79.062 48.070
75398 95.746 -79.059 48.026
75399 95.755 -79.055 47.984
75400 95.765 -79.051 47.942
75401 95.774 -79.047 47.899
75402 95.783 -79.044 47.856
75403 95.792 -79.040 47.812
75404 95.798 -79.037 47.768
75405 95.805 -79.033 47.726
75406 95.811 -79.029 47.683
75407 95.817 -79.025 47.641
75408 95.824 -79.022 47.599
75409 95.830 -79.018 47.558
75410 95.836 -79.015 47.518
75411 95.843 -79.011 47.478
75412 95.849 -79.008 47.437
75413 95.882 -79.004 47.396
75414 95.918 -78.999 47.355
75415 95.955 -78.996 47.316
75416 96.014 -79.008 47.277
75417 96.085 -79.018 47.239
75418 96.125 -79.021 47.201
75419 96.165 -79.024 47.165
75420 96.204 -79.027 47.129
75421 96.244 -79.030 47.094
75422 96.284 -79.031 47.059
75423 96.318 -79.018 47.025
75424 96.352 -79.005 46.992
75425 96.386 -78.992 46.962
75426 96.415 -79.031 46.932
75427 96.435 -79.049 46.901
75428 96.450 -79.049 46.870
75429 96.471 -79.058 46.839
75430 96.495 -79.070 46.805
75431 96.519 -79.082 46.772
75432 96.542 -79.093 46.741
75433 96.566 -79.106 46.709
75434 96.590 -79.117 46.678
75435 96.613 -79.129 46.647
75436 96.637 -79.140 46.615
75437 96.661 -79.153 46.585
75438 96.684 -79.164 46.557
75439 96.693 -79.159 46.529
75440 96.691 -79.142 46.499
75441 96.689 -79.125 46.471
75442 96.687 -79.109 46.443
75443 96.688 -79.090 46.415
75444 96.690 -79.071 46.388
75445 96.751 -79.039 46.362
75446 96.791 -79.075 46.336
75447 96.849 -79.080 46.312
75448 96.908 -79.085 46.288
75449 96.876 -79.131 46.264
75450 96.853 -79.125 46.240
75451 96.876 -79.093 46.216
75452 96.900 -79.062 46.192
75453 96.925 -79.030 46.168
75454 96.960 -78.993 46.146
75455 96.999 -78.996 46.124
75456 97.041 -79.041 46.102
75457 97.083 -79.086 46.083
75458 97.110 -79.117 46.064
75459 97.075 -79.088 46.045
75460 97.040 -79.060 46.028
75461 97.005 -79.032 46.009
75462 97.011 -79.048 45.989
75463 97.049 -79.222 45.969
75464 97.045 -79.226 45.950
75465 97.043 -79.211 45.932
75466 97.044 -79.191 45.915
75467 97.050 -79.182 45.898
75468 97.058 -79.172 45.881
75469 97.069 -79.161 45.864
75470 97.080 -79.150 45.848
75471 97.091 -79.138 45.830
75472 97.100 -79.124 45.814
75473 97.104 -79.101 45.798
75474 97.124 -79.086 45.781
75475 97.162 -79.079 45.763
75476 97.207 -79.092 45.745
75477 97.224 -79.104 45.727
75478 97.214 -79.122 45.708
75479 97.231 -79.159 45.690
75480 97.262 -79.162 45.674
75481 97.296 -79.159 45.657
75482 97.329 -79.158 45.640
75483 97.363 -79.156 45.624
75484 97.397 -79.153 45.609
75485 97.399 -79.172 45.593
75486 97.345 -79.216 45.578
75487 97.300 -79.255 45.564
75488 97.309 -79.255 45.550
75489 97.320 -79.255 45.536
75490 97.331 -79.256 45.522
75491 97.343 -79.256 45.507
75492 97.373 -79.259 45.493
75493 97.404 -79.266 45.478
75494 97.402 -79.291 45.465
75495 97.410 -79.297 45.451
75496 97.423 -79.301 45.436
75497 97.436 -79.307 45.422
75498 97.450 -79.313 45.407
75499 97.463 -79.319 45.393
75500 97.477 -79.324 45.379
75501 97.490 -79.330 45.366
75502 97.504 -79.335 45.353
75503 97.522 -79.349 45.340
75504 97.541 -79.363 45.327
75505 97.560 -79.377 45.314
75506 97.579 -79.391 45.301
75507 97.583 -79.387 45.288
75508 97.587 -79.382 45.275
75509 97.591 -79.378 45.263
75510 97.595 -79.375 45.251
75511 97.610 -79.402 45.240
75512 97.626 -79.430 45.228
75513 97.640 -79.457 45.217
75514 97.649 -79.469 45.207
75515 97.655 -79.474 45.197
75516 97.661 -79.479 45.187
75517 97.667 -79.485 45.178
75518 97.677 -79.557 45.168
75519 97.693 -79.589 45.157
75520 97.707 -79.606 45.145
75521 97.720 -79.613 45.134
75522 97.733 -79.620 45.123
75523 97.745 -79.627 45.111
75524 97.758 -79.635 45.100
75525 97.771 -79.642 45.088
75526 97.784 -79.649 45.076
75527 97.796 -79.656 45.064
75528 97.810 -79.666 45.052
75529 97.825 -79.677 45.040
75530 97.841 -79.689 45.031
75531 97.862 -79.702 45.022
75532 97.883 -79.717 45.013
75533 97.906 -79.732 45.004
75534 97.927 -79.746 44.995
75535 97.964 -79.786 44.987
75536 98.003 -79.831 44.980
75537 98.027 -79.835 44.974
75538 98.047 -79.831 44.967
75539 98.068 -79.827 44.959
75540 98.088 -79.823 44.952
75541 98.110 -79.828 44.948
75542 98.133 -79.838 44.944
75543 98.157 -79.848 44.939
75544 98.180 -79.860 44.932
75545 98.205 -79.866 44.923
75546 98.230 -79.871 44.912
75547 98.255 -79.877 44.902
75548 98.281 -79.883 44.889
75549 98.306 -79.888 44.875
75550 98.331 -79.894 44.862
75551 98.357 -79.900 44.853
75552 98.395 -79.911 44.847
75553 98.433 -79.924 44.842
75554 98.471 -79.935 44.836
75555 98.508 -79.947 44.830
75556 98.542 -79.956 44.821
75557 98.574 -79.963 44.812
75558 98.603 -79.968 44.801
75559 98.633 -79.973 44.790
75560 98.662 -79.977 44.778
75561 98.689 -79.983 44.768
75562 98.701 -79.999 44.759
75563 98.712 -80.014 44.752
75564 98.733 -80.024 44.745
75565 98.759 -80.033 44.736
75566 98.784 -80.041 44.726
75567 98.812 -80.047 44.718
75568 98.847 -80.047 44.709
75569 98.874 -80.053 44.701
75570 98.895 -80.062 44.692
75571 98.916 -80.072 44.684
75572 98.938 -80.082 44.677
75573 98.952 -80.085 44.669
75574 98.969 -80.090 44.663
75575 98.985 -80.094 44.657
75576 99.001 -80.099 44.650
75577 99.017 -80.104 44.643
75578 99.031 -80.112 44.638
75579 99.046 -80.121 44.632
75580 99.061 -80.129 44.627
75581 99.076 -80.138 44.622
75582 99.092 -80.149 44.616
75583 99.156 -80.176 44.611
75584 99.204 -80.198 44.605
75585 99.227 -80.208 44.601
75586 99.251 -80.218 44.596
75587 99.274 -80.229 44.591
75588 99.302 -80.237 44.588
75589 99.341 -80.241 44.584
75590 99.379 -80.245 44.580
75591 99.418 -80.250 44.576
75592 99.457 -80.253 44.573
75593 99.495 -80.257 44.569
75594 99.534 -80.261 44.564
75595 99.547 -80.278 44.560
75596 99.550 -80.300 44.556
75597 99.553 -80.322 44.551
75598 99.556 -80.344 44.546
75599 99.560 -80.367 44.541
75600 99.563 -80.389 44.535
75601 99.566 -80.411 44.530
75602 99.569 -80.433 44.524
75603 99.572 -80.455 44.518
75604 99.576 -80.478 44.512
75605 99.579 -80.500 44.505
75606 99.582 -80.522 44.498
75607 99.585 -80.544 44.491
75608 99.588 -80.566 44.483
75609 99.592 -80.588 44.476
75610 99.595 -80.610 44.468
75611 99.598 -80.632 44.460
75612 99.602 -80.654 44.452
75613 99.605 -80.677 44.443
75614 99.608 -80.699 44.434
75615 99.611 -80.721 44.426
75616 99.615 -80.743 44.416
75617 99.632 -80.760 44.407
75919 93.544 -78.904 53.510
75920 93.553 -78.901 53.501
75921 93.563 -78.899 53.491
75922 93.573 -78.896 53.481
75923 93.584 -78.887 53.474
75924 93.592 -78.880 53.465
75925 93.601 -78.874 53.456
75926 93.609 -78.866 53.450
75927 93.618 -78.859 53.442
75928 93.652 -78.841 53.437
75929 93.717 -78.808 53.433
75930 93.712 -78.806 53.430
75931 93.710 -78.805 53.428
75932 93.703 -78.802 53.427
75933 93.692 -78.799 53.426
75934 93.691 -78.802 53.420
75935 93.702 -78.809 53.414
75936 93.709 -78.809 53.407
75937 93.716 -78.808 53.398
75938 93.723 -78.807 53.387
75939 93.730 -78.804 53.377
75940 93.737 -78.801 53.368
75941 93.744 -78.799 53.355
75942 93.751 -78.796 53.342
75943 93.758 -78.793 53.328
75944 93.768 -78.788 53.314
75945 93.793 -78.774 53.298
75946 93.808 -78.763 53.282
75947 93.812 -78.754 53.265
75948 93.815 -78.745 53.245
75949 93.824 -78.737 53.224
75950 93.838 -78.730 53.202
75951 93.852 -78.721 53.181
75952 93.866 -78.714 53.160
75953 93.880 -78.705 53.140
75954 93.894 -78.697 53.118
75955 93.901 -78.692 53.095
75956 93.906 -78.688 53.070
75957 93.912 -78.683 53.044
75958 93.917 -78.679 53.018
75959 93.928 -78.673 52.993
75960 93.944 -78.667 52.967
75961 93.960 -78.661 52.941
75962 93.976 -78.655 52.916
75963 93.992 -78.648 52.891
75964 94.008 -78.642 52.866
75965 94.018 -78.640 52.844
75966 94.028 -78.639 52.822
75967 94.038 -78.636 52.800
75968 94.048 -78.635 52.781
75969 94.058 -78.633 52.758
75970 94.067 -78.632 52.735
75971 94.077 -78.630 52.713
75972 94.085 -78.630 52.690
75973 94.084 -78.636 52.668
75974 94.083 -78.642 52.645
75975 94.088 -78.646 52.622
75976 94.108 -78.611 52.597
75977 94.128 -78.574 52.572
75978 94.179 -78.586 52.547
75979 94.230 -78.599 52.522
75980 94.282 -78.613 52.499
75981 94.314 -78.619 52.476
75982 94.324 -78.616 52.455
75983 94.334 -78.614 52.432
75984 94.338 -78.616 52.409
75985 94.330 -78.628 52.387
75986 94.322 -78.640 52.366
75987 94.315 -78.652 52.344
75988 94.307 -78.664 52.321
75989 94.299 -78.676 52.297
75990 94.316 -78.664 52.271
75991 94.332 -78.651 52.244
75992 94.349 -78.639 52.218
75993 94.362 -78.631 52.193
75994 94.370 -78.629 52.167
75995 94.379 -78.626 52.139
75996 94.388 -78.624 52.112
75997 94.396 -78.622 52.084
75998 94.405 -78.620 52.055
75999 94.419 -78.619 52.024
76000 94.437 -78.620 51.992
76001 94.450 -78.623 51.959
76002 94.461 -78.628 51.926
76003 94.472 -78.633 51.893
76004 94.482 -78.644 51.862
76005 94.492 -78.677 51.830
76006 94.505 -78.692 51.795
76007 94.520 -78.692 51.759
76008 94.535 -78.691 51.723
76009 94.550 -78.690 51.684
76010 94.565 -78.690 51.644
76011 94.580 -78.689 51.605
76012 94.596 -78.689 51.566
76013 94.610 -78.688 51.531
76014 94.625 -78.687 51.496
76015 94.640 -78.686 51.464
76016 94.653 -78.683 51.433
76017 94.665 -78.680 51.403
76018 94.677 -78.676 51.374
76019 94.688 -78.673 51.344
76020 94.699 -78.675 51.310
76021 94.710 -78.677 51.274
76022 94.720 -78.680 51.237
76023 94.732 -78.682 51.199
76024 94.744 -78.684 51.160
76025 94.758 -78.689 51.120
76026 94.764 -78.697 51.080
76027 94.769 -78.705 51.038
76028 94.775 -78.712 50.996
76029 94.780 -78.720 50.954
76030 94.786 -78.727 50.914
76031 94.793 -78.724 50.873
76032 94.803 -78.711 50.832
76033 94.812 -78.698 50.787
76034 94.819 -78.695 50.742
76035 94.825 -78.692 50.700
76036 94.832 -78.688 50.660
76037 94.841 -78.680 50.618
76038 94.849 -78.671 50.574
76039 94.862 -78.664 50.529
76040 94.875 -78.667 50.484
76041 94.873 -78.682 50.438
76042 94.872 -78.698 50.393
76043 94.871 -78.713 50.349
76044 94.870 -78.728 50.310
76045 94.868 -78.744 50.270
76046 94.867 -78.759 50.231
76047 94.866 -78.775 50.192
76048 94.865 -78.790 50.153
76049 94.867 -78.804 50.113
76050 94.868 -78.818 50.071
76051 94.871 -78.833 50.028
76052 94.876 -78.844 49.987
76053 94.896 -78.846 49.946
76054 94.926 -78.842 49.904
76055 94.956 -78.838 49.862
76056 94.986 -78.834 49.821
76057 95.015 -78.830 49.781
76058 95.044 -78.826 49.743
76059 95.072 -78.821 49.704
76060 95.101 -78.817 49.663
76061 95.130 -78.812 49.622
76062 95.159 -78.808 49.583
76063 95.187 -78.803 49.543
76064 95.216 -78.802 49.504
76065 95.246 -78.808 49.464
76066 95.273 -78.814 49.424
76067 95.293 -78.826 49.383
76068 95.312 -78.838 49.344
76069 95.331 -78.849 49.303
76070 95.351 -78.861 49.261
76071 95.371 -78.873 49.220
76072 95.390 -78.885 49.181
76073 95.408 -78.893 49.142
76074 95.422 -78.891 49.102
76075 95.436 -78.889 49.062
76076 95.450 -78.886 49.021
76077 95.464 -78.884 48.981
76078 95.478 -78.882 48.942
76079 95.492 -78.880 48.901
76080 95.500 -78.885 48.859
76081 95.509 -78.891 48.817
76082 95.516 -78.902 48.776
76083 95.521 -78.921 48.735
76084 95.525 -78.942 48.694
76085 95.537 -78.958 48.654
76086 95.552 -78.971 48.614
76087 95.567 -78.984 48.574
76088 95.582 -78.997 48.534
76089 95.597 -79.011 48.492
76090 95.612 -79.024 48.447
76091 95.627 -79.037 48.401
76092 95.643 -79.050 48.356
76093 95.658 -79.064 48.309
76094 95.673 -79.077 48.261
76095 95.688 -79.090 48.212
76096 95.710 -79.081 48.164
76097 95.732 -79.068 48.119
76098 95.745 -79.061 48.076
76099 95.754 -79.057 48.032
76100 95.763 -79.053 47.991
76101 95.772 -79.049 47.949
76102 95.781 -79.046 47.906
76103 95.791 -79.042 47.862
76104 95.800 -79.038 47.819
76105 95.808 -79.034 47.776
76106 95.814 -79.031 47.734
76107 95.820 -79.028 47.691
76108 95.827 -79.024 47.648
76109 95.833 -79.020 47.605
76110 95.840 -79.016 47.565
76111 95.846 -79.013 47.524
76112 95.852 -79.009 47.483
76113 95.877 -79.006 47.442
76114 95.913 -79.002 47.401
76115 95.950 -78.997 47.361
76116 95.986 -78.993 47.321
76117 96.023 -78.990 47.282
76118 96.087 -79.000 47.245
76119 96.127 -79.003 47.208
76120 96.167 -79.006 47.171
76121 96.206 -79.009 47.135
76122 96.246 -79.012 47.099
76123 96.286 -79.014 47.063
76124 96.326 -79.016 47.031
76125 96.360 -79.003 46.998
76126 96.394 -78.991 46.967
76127 96.422 -79.028 46.937
76128 96.437 -79.028 46.906
76129 96.459 -79.038 46.875
76130 96.483 -79.049 46.844
76131 96.507 -79.062 46.813
76132 96.530 -79.073 46.780
76133 96.554 -79.085 46.748
76134 96.578 -79.096 46.716
76135 96.601 -79.109 46.684
76136 96.625 -79.120 46.653
76137 96.649 -79.132 46.621
76138 96.672 -79.143 46.592
76139 96.696 -79.156 46.563
76140 96.714 -79.161 46.534
76141 96.716 -79.142 46.505
76142 96.718 -79.123 46.476
76143 96.720 -79.104 46.447
76144 96.722 -79.085 46.419
76145 96.724 -79.066 46.393
76146 96.788 -79.032 46.367
76147 96.862 -78.996 46.342
76148 96.915 -79.001 46.318
76149 96.949 -79.045 46.293
76150 96.938 -79.068 46.269
76151 96.911 -79.065 46.244
76152 96.915 -79.045 46.220
76153 96.939 -79.012 46.197
76154 96.958 -78.983 46.174
76155 96.970 -78.958 46.153
76156 96.978 -78.976 46.131
76157 96.979 -78.987 46.110
76158 97.012 -79.026 46.091
76159 97.034 -79.056 46.073
76160 96.991 -79.025 46.055
76161 96.979 -79.006 46.037
76162 96.976 -78.987 46.018
76163 97.007 -79.090 45.999
76164 97.039 -79.241 45.979
76165 97.038 -79.228 45.960
76166 97.043 -79.219 45.942
76167 97.049 -79.211 45.925
76168 97.055 -79.202 45.908
76169 97.063 -79.193 45.892
76170 97.074 -79.181 45.876
76171 97.085 -79.170 45.859
76172 97.095 -79.159 45.841
76173 97.101 -79.139 45.824
76174 97.106 -79.117 45.808
76175 97.110 -79.094 45.792
76176 97.131 -79.078 45.774
76177 97.168 -79.071 45.756
76178 97.205 -79.064 45.738
76179 97.230 -79.094 45.719
76180 97.254 -79.125 45.701
76181 97.275 -79.159 45.685
76182 97.303 -79.166 45.669
76183 97.334 -79.161 45.652
76184 97.366 -79.156 45.636
76185 97.397 -79.150 45.620
76186 97.407 -79.162 45.605
76187 97.401 -79.182 45.591
76188 97.349 -79.227 45.577
76189 97.322 -79.253 45.562
76190 97.333 -79.253 45.548
76191 97.344 -79.253 45.534
76192 97.355 -79.254 45.520
76193 97.392 -79.257 45.506
76194 97.399 -79.277 45.492
76195 97.404 -79.288 45.478
76196 97.414 -79.291 45.464
76197 97.425 -79.294 45.449
76198 97.436 -79.297 45.435
76199 97.450 -79.303 45.421
76200 97.463 -79.309 45.406
76201 97.476 -79.314 45.393
76202 97.490 -79.320 45.380
76203 97.504 -79.325 45.367
76204 97.520 -79.336 45.354
76205 97.539 -79.350 45.340
76206 97.558 -79.364 45.326
76207 97.577 -79.378 45.313
76208 97.596 -79.392 45.300
76209 97.609 -79.400 45.288
76210 97.613 -79.395 45.274
76211 97.619 -79.394 45.262
76212 97.634 -79.422 45.250
76213 97.647 -79.444 45.239
76214 97.653 -79.450 45.229
76215 97.658 -79.455 45.220
76216 97.664 -79.460 45.211
76217 97.670 -79.466 45.201
76218 97.676 -79.472 45.192
76219 97.687 -79.525 45.183
76220 97.703 -79.556 45.174
76221 97.719 -79.582 45.162
76222 97.731 -79.588 45.150
76223 97.744 -79.596 45.139
76224 97.757 -79.603 45.127
76225 97.770 -79.610 45.115
76226 97.782 -79.617 45.103
76227 97.795 -79.624 45.090
76228 97.809 -79.635 45.078
76229 97.823 -79.647 45.067
76230 97.838 -79.658 45.056
76231 97.853 -79.670 45.048
76232 97.874 -79.683 45.040
76233 97.895 -79.696 45.032
76234 97.917 -79.710 45.021
76235 97.947 -79.739 45.013
76236 97.961 -79.744 45.005
76237 97.971 -79.741 44.998
76238 97.996 -79.752 44.991
76239 98.030 -79.766 44.984
76240 98.064 -79.780 44.977
76241 98.097 -79.794 44.971
76242 98.121 -79.805 44.963
76243 98.144 -79.817 44.957
76244 98.167 -79.829 44.951
76245 98.192 -79.839 44.943
76246 98.217 -79.843 44.934
76247 98.242 -79.848 44.923
76248 98.267 -79.852 44.910
76249 98.293 -79.858 44.896
76250 98.318 -79.863 44.882
76251 98.343 -79.869 44.870
76252 98.368 -79.875 44.860
76253 98.406 -79.886 44.853
76254 98.443 -79.898 44.846
76255 98.481 -79.910 44.840
76256 98.515 -79.918 44.832
76257 98.547 -79.925 44.824
76258 98.578 -79.933 44.815
76259 98.609 -79.939 44.805
76260 98.639 -79.944 44.795
76261 98.668 -79.949 44.785
76262 98.685 -79.961 44.776
76263 98.698 -79.975 44.767
76264 98.723 -79.983 44.759
76265 98.748 -79.992 44.753
76266 98.774 -80.000 44.745
76267 98.799 -80.008 44.736
76268 98.825 -80.017 44.728
76269 98.860 -80.018 44.719
76270 98.881 -80.027 44.710
76271 98.903 -80.036 44.701
76272 98.924 -80.044 44.694
76273 98.944 -80.043 44.687
76274 98.961 -80.047 44.680
76275 98.981 -80.061 44.675
76276 99.000 -80.075 44.669
76277 99.016 -80.081 44.662
76278 99.032 -80.086 44.655
76279 99.049 -80.091 44.651
76280 99.065 -80.096 44.646
76281 99.079 -80.105 44.641
76282 99.093 -80.113 44.635
76283 99.108 -80.121 44.629
76284 99.123 -80.130 44.624
76285 99.153 -80.145 44.619
76286 99.231 -80.181 44.613
76287 99.255 -80.191 44.608
76288 99.278 -80.202 44.604
76289 99.313 -80.207 44.600
76290 99.352 -80.211 44.596
76291 99.391 -80.215 44.592
76292 99.429 -80.219 44.587
76293 99.468 -80.223 44.583
76294 99.506 -80.227 44.579
76295 99.540 -80.233 44.575
76296 99.543 -80.256 44.570
76297 99.547 -80.278 44.566
76298 99.550 -80.300 44.561
76299 99.553 -80.322 44.556
76300 99.556 -80.344 44.550
76301 99.559 -80.366 44.544
76302 99.563 -80.388 44.538
76303 99.566 -80.410 44.532
76304 99.569 -80.432 44.526
76305 99.572 -80.455 44.520
76306 99.575 -80.477 44.513
76307 99.579 -80.499 44.505
76308 99.582 -80.521 44.497
76309 99.585 -80.543 44.490
76310 99.588 -80.566 44.482
76311 99.591 -80.588 44.474
76312 99.595 -80.610 44.466
76313 99.598 -80.632 44.458
76314 99.601 -80.654 44.449
76315 99.605 -80.676 44.440
76316 99.608 -80.698 44.431
76317 99.611 -80.720 44.421
76318 99.640 -80.733 44.412
76319 99.688 -80.739 44.402
76320 99.736 -80.744 44.391
76321 99.784 -80.750 44.381
76620 93.544 -78.893 53.523
76621 93.554 -78.890 53.513
76622 93.563 -78.888 53.503
76623 93.573 -78.885 53.493
76624 93.582 -78.883 53.486
76625 93.593 -78.874 53.478
76626 93.602 -78.867 53.470
76627 93.610 -78.860 53.463
76628 93.619 -78.853 53.456
76629 93.641 -78.840 53.451
76630 93.709 -78.806 53.448
76631 93.700 -78.805 53.447
76632 93.698 -78.802 53.446
76633 93.697 -78.801 53.445
76634 93.692 -78.799 53.443
76635 93.690 -78.802 53.437
76636 93.698 -78.802 53.430
76637 93.705 -78.802 53.422
76638 93.711 -78.801 53.412
76639 93.718 -78.799 53.402
76640 93.726 -78.797 53.393
76641 93.733 -78.794 53.382
76642 93.740 -78.791 53.370
76643 93.747 -78.788 53.356
76644 93.755 -78.785 53.342
76645 93.779 -78.771 53.327
76646 93.804 -78.757 53.310
76647 93.809 -78.749 53.293
76648 93.813 -78.739 53.275
76649 93.816 -78.731 53.255
76650 93.823 -78.723 53.234
76651 93.837 -78.715 53.213
76652 93.851 -78.707 53.191
76653 93.865 -78.699 53.172
76654 93.879 -78.691 53.152
76655 93.893 -78.683 53.130
76656 93.907 -78.675 53.106
76657 93.919 -78.667 53.080
76658 93.924 -78.663 53.054
76659 93.929 -78.659 53.029
76660 93.943 -78.653 53.003
76661 93.960 -78.646 52.977
76662 93.976 -78.640 52.951
76663 93.987 -78.637 52.926
76664 93.997 -78.636 52.901
76665 94.007 -78.633 52.878
76666 94.017 -78.632 52.857
76667 94.027 -78.630 52.836
76668 94.037 -78.629 52.815
76669 94.047 -78.627 52.796
76670 94.057 -78.626 52.773
76671 94.067 -78.623 52.750
76672 94.077 -78.622 52.728
76673 94.082 -78.623 52.705
76674 94.081 -78.630 52.682
76675 94.079 -78.636 52.659
76676 94.099 -78.592 52.636
76677 94.119 -78.556 52.611
76678 94.163 -78.557 52.586
76679 94.214 -78.571 52.560
76680 94.266 -78.584 52.536
76681 94.318 -78.598 52.513
76682 94.337 -78.598 52.492
76683 94.347 -78.596 52.472
76684 94.351 -78.598 52.450
76685 94.344 -78.610 52.427
76686 94.336 -78.622 52.406
76687 94.328 -78.634 52.384
76688 94.321 -78.646 52.362
76689 94.313 -78.658 52.339
76690 94.305 -78.670 52.316
76691 94.318 -78.661 52.292
76692 94.334 -78.649 52.266
76693 94.351 -78.637 52.240
76694 94.367 -78.624 52.214
76695 94.380 -78.617 52.188
76696 94.388 -78.615 52.161
76697 94.397 -78.613 52.135
76698 94.405 -78.611 52.107
76699 94.414 -78.608 52.077
76700 94.428 -78.608 52.046
76701 94.447 -78.608 52.012
76702 94.462 -78.610 51.978
76703 94.473 -78.615 51.945
76704 94.484 -78.620 51.911
76705 94.494 -78.646 51.878
76706 94.504 -78.680 51.845
76707 94.517 -78.697 51.810
76708 94.532 -78.696 51.773
76709 94.547 -78.695 51.735
76710 94.562 -78.695 51.696
76711 94.577 -78.694 51.656
76712 94.592 -78.693 51.617
76713 94.607 -78.693 51.577
76714 94.620 -78.689 51.541
76715 94.631 -78.686 51.507
76716 94.643 -78.682 51.475
76717 94.655 -78.678 51.444
76718 94.667 -78.674 51.414
76719 94.679 -78.670 51.384
76720 94.690 -78.667 51.354
76721 94.701 -78.669 51.320
76722 94.712 -78.671 51.283
76723 94.723 -78.673 51.245
76724 94.734 -78.676 51.208
76725 94.751 -78.679 51.169
76726 94.768 -78.683 51.130
76727 94.773 -78.691 51.091
76728 94.779 -78.698 51.050
76729 94.784 -78.705 51.008
76730 94.790 -78.713 50.966
76731 94.795 -78.720 50.924
76732 94.804 -78.714 50.884
76733 94.813 -78.701 50.843
76734 94.822 -78.688 50.799
76735 94.832 -78.675 50.754
76736 94.839 -78.670 50.711
76737 94.845 -78.666 50.669
76738 94.852 -78.663 50.628
76739 94.858 -78.659 50.583
76740 94.865 -78.655 50.538
76741 94.873 -78.655 50.493
76742 94.872 -78.670 50.445
76743 94.873 -78.685 50.400
76744 94.875 -78.698 50.356
76745 94.877 -78.713 50.316
76746 94.879 -78.727 50.278
76747 94.881 -78.741 50.240
76748 94.882 -78.755 50.200
76749 94.884 -78.769 50.161
76750 94.886 -78.783 50.120
76751 94.888 -78.797 50.076
76752 94.890 -78.811 50.032
76753 94.900 -78.819 49.989
76754 94.914 -78.824 49.946
76755 94.934 -78.825 49.903
76756 94.964 -78.822 49.860
76757 94.993 -78.818 49.819
76758 95.024 -78.814 49.779
76759 95.053 -78.811 49.740
76760 95.083 -78.807 49.701
76761 95.113 -78.804 49.662
76762 95.143 -78.800 49.622
76763 95.173 -78.796 49.582
76764 95.202 -78.793 49.543
76765 95.233 -78.789 49.504
76766 95.254 -78.798 49.463
76767 95.273 -78.810 49.423
76768 95.293 -78.821 49.382
76769 95.312 -78.833 49.342
76770 95.332 -78.845 49.302
76771 95.351 -78.857 49.261
76772 95.371 -78.868 49.220
76773 95.390 -78.880 49.182
76774 95.409 -78.890 49.142
76775 95.422 -78.887 49.102
76776 95.437 -78.885 49.061
76777 95.451 -78.883 49.021
76778 95.464 -78.880 48.982
76779 95.478 -78.878 48.943
76780 95.492 -78.876 48.902
76781 95.506 -78.873 48.860
76782 95.518 -78.874 48.818
76783 95.526 -78.880 48.777
76784 95.531 -78.900 48.736
76785 95.542 -78.916 48.695
76786 95.557 -78.930 48.654
76787 95.572 -78.943 48.613
76788 95.587 -78.956 48.573
76789 95.602 -78.969 48.532
76790 95.618 -78.983 48.492
76791 95.633 -78.996 48.450
76792 95.648 -79.009 48.406
76793 95.663 -79.022 48.360
76794 95.678 -79.035 48.312
76795 95.693 -79.049 48.264
76796 95.708 -79.062 48.216
76797 95.724 -79.074 48.170
76798 95.744 -79.063 48.126
76799 95.753 -79.059 48.083
76800 95.762 -79.056 48.039
76801 95.771 -79.052 47.998
76802 95.780 -79.048 47.956
76803 95.789 -79.044 47.912
76804 95.798 -79.040 47.869
76805 95.807 -79.037 47.827
76806 95.816 -79.033 47.785
76807 95.824 -79.029 47.742
76808 95.830 -79.025 47.699
76809 95.836 -79.021 47.656
76810 95.843 -79.018 47.612
76811 95.849 -79.015 47.572
76812 95.855 -79.011 47.530
76813 95.871 -79.007 47.489
76814 95.907 -79.003 47.448
76815 95.944 -78.999 47.408
76816 95.980 -78.996 47.368
76817 96.017 -78.991 47.328
76818 96.053 -78.987 47.289
76819 96.090 -78.984 47.251
76820 96.129 -78.985 47.214
76821 96.169 -78.987 47.177
76822 96.208 -78.990 47.141
76823 96.248 -78.993 47.105
76824 96.288 -78.996 47.070
76825 96.328 -78.999 47.037
76826 96.368 -79.001 47.005
76827 96.402 -78.990 46.974
76828 96.425 -79.007 46.943
76829 96.447 -79.018 46.913
76830 96.471 -79.030 46.882
76831 96.495 -79.041 46.851
76832 96.518 -79.053 46.820
76833 96.542 -79.065 46.789
76834 96.566 -79.077 46.757
76835 96.584 -79.086 46.724
76836 96.603 -79.096 46.691
76837 96.621 -79.105 46.660
76838 96.640 -79.115 46.629
76839 96.658 -79.124 46.599
76840 96.677 -79.134 46.570
76841 96.707 -79.138 46.539
76842 96.743 -79.134 46.510
76843 96.752 -79.118 46.481
76844 96.754 -79.099 46.452
76845 96.756 -79.080 46.425
76846 96.758 -79.061 46.398
76847 96.823 -79.027 46.373
76848 96.897 -78.990 46.348
76849 96.972 -78.954 46.324
76850 96.989 -78.949 46.299
76851 96.965 -78.977 46.274
76852 96.942 -79.006 46.249
76853 96.935 -79.006 46.226
76854 96.946 -78.981 46.204
76855 96.958 -78.955 46.182
76856 96.969 -78.943 46.161
76857 96.977 -78.963 46.140
76858 96.985 -78.993 46.119
76859 96.993 -79.021 46.100
76860 96.983 -79.019 46.082
76861 96.959 -79.006 46.065
76862 96.976 -78.987 46.046
76863 96.994 -78.979 46.027
76864 97.045 -79.140 46.008
76865 97.037 -79.256 45.989
76866 97.043 -79.247 45.970
76867 97.048 -79.239 45.953
76868 97.054 -79.231 45.936
76869 97.061 -79.222 45.918
76870 97.068 -79.213 45.902
76871 97.078 -79.202 45.886
76872 97.089 -79.190 45.869
76873 97.099 -79.177 45.850
76874 97.103 -79.154 45.833
76875 97.107 -79.132 45.817
76876 97.112 -79.109 45.801
76877 97.116 -79.087 45.785
76878 97.137 -79.071 45.767
76879 97.169 -79.078 45.749
76880 97.193 -79.109 45.730
76881 97.229 -79.134 45.713
76882 97.268 -79.156 45.697
76883 97.301 -79.159 45.682
76884 97.332 -79.153 45.666
76885 97.364 -79.148 45.650
76886 97.385 -79.151 45.633
76887 97.391 -79.165 45.619
76888 97.398 -79.179 45.605
76889 97.404 -79.193 45.590
76890 97.362 -79.232 45.576
76891 97.346 -79.251 45.562
76892 97.357 -79.252 45.548
76893 97.371 -79.253 45.533
76894 97.397 -79.263 45.520
76895 97.400 -79.276 45.505
76896 97.407 -79.282 45.491
76897 97.417 -79.286 45.478
76898 97.428 -79.288 45.464
76899 97.439 -79.291 45.450
76900 97.450 -79.294 45.436
76901 97.463 -79.299 45.422
76902 97.476 -79.304 45.409
76903 97.490 -79.310 45.396
76904 97.504 -79.316 45.383
76905 97.518 -79.322 45.369
76906 97.537 -79.337 45.356
76907 97.556 -79.350 45.342
76908 97.575 -79.365 45.329
76909 97.594 -79.379 45.315
76910 97.612 -79.393 45.301
76911 97.632 -79.407 45.288
76912 97.642 -79.414 45.275
76913 97.650 -79.425 45.262
76914 97.656 -79.431 45.251
76915 97.662 -79.436 45.243
76916 97.668 -79.441 45.234
76917 97.674 -79.447 45.225
76918 97.680 -79.452 45.214
76919 97.686 -79.457 45.205
76920 97.697 -79.491 45.196
76921 97.713 -79.523 45.185
76922 97.729 -79.554 45.174
76923 97.743 -79.564 45.163
76924 97.755 -79.571 45.153
76925 97.768 -79.579 45.141
76926 97.781 -79.585 45.129
76927 97.794 -79.593 45.117
76928 97.808 -79.604 45.104
76929 97.822 -79.616 45.091
76930 97.837 -79.627 45.080
76931 97.852 -79.639 45.072
76932 97.866 -79.651 45.065
76933 97.886 -79.664 45.057
76934 97.908 -79.676 45.048
76935 97.928 -79.689 45.039
76936 97.943 -79.695 45.031
76937 97.956 -79.701 45.024
76938 97.973 -79.709 45.016
76939 97.981 -79.733 45.007
76940 98.026 -79.735 45.000
76941 98.066 -79.752 44.992
76942 98.105 -79.773 44.985
76943 98.132 -79.786 44.977
76944 98.155 -79.798 44.969
76945 98.178 -79.811 44.961
76946 98.204 -79.817 44.952
76947 98.229 -79.820 44.943
76948 98.254 -79.825 44.932
76949 98.279 -79.829 44.917
76950 98.304 -79.833 44.902
76951 98.330 -79.838 44.889
76952 98.355 -79.844 44.877
76953 98.380 -79.850 44.868
76954 98.417 -79.861 44.859
76955 98.454 -79.872 44.851
76956 98.487 -79.880 44.843
76957 98.519 -79.888 44.835
76958 98.551 -79.895 44.826
76959 98.582 -79.902 44.818
76960 98.614 -79.909 44.809
76961 98.644 -79.915 44.800
76962 98.670 -79.922 44.793
76963 98.688 -79.935 44.784
76964 98.713 -79.943 44.775
76965 98.738 -79.952 44.767
76966 98.764 -79.960 44.761
76967 98.789 -79.968 44.754
76968 98.814 -79.977 44.746
76969 98.840 -79.985 44.738
76970 98.867 -79.991 44.729
76971 98.888 -79.996 44.721
76972 98.909 -79.996 44.712
76973 98.929 -79.994 44.705
76974 98.950 -79.993 44.698
76975 98.964 -79.988 44.692
76976 98.983 -80.002 44.687
76977 99.002 -80.016 44.681
76978 99.021 -80.030 44.675
76979 99.040 -80.044 44.668
76980 99.059 -80.059 44.664
76981 99.078 -80.072 44.659
76982 99.096 -80.083 44.654
76983 99.112 -80.088 44.649
76984 99.127 -80.097 44.643
76985 99.209 -80.128 44.638
76986 99.214 -80.128 44.632
76987 99.219 -80.129 44.627
76988 99.243 -80.153 44.622
76989 99.286 -80.173 44.617
76990 99.324 -80.177 44.613
76991 99.363 -80.181 44.609
76992 99.402 -80.185 44.605
76993 99.440 -80.189 44.599
76994 99.479 -80.193 44.595
76995 99.517 -80.196 44.591
76996 99.537 -80.210 44.586
76997 99.540 -80.232 44.581
76998 99.543 -80.255 44.576
76999 99.546 -80.277 44.571
77000 99.550 -80.299 44.565
77001 99.553 -80.321 44.559
77002 99.556 -80.344 44.553
77003 99.559 -80.366 44.546
77004 99.562 -80.388 44.540
77005 99.566 -80.410 44.534
77006 99.569 -80.431 44.527
77007 99.572 -80.454 44.520
77008 99.575 -80.476 44.512
77009 99.578 -80.498 44.504
77010 99.581 -80.520 44.496
77011 99.585 -80.542 44.489
77012 99.588 -80.565 44.481
77013 99.591 -80.587 44.472
77014 99.594 -80.609 44.464
77015 99.597 -80.631 44.455
77016 99.601 -80.654 44.445
77017 99.604 -80.676 44.436
77018 99.608 -80.698 44.426
77019 99.648 -80.706 44.416
77020 99.697 -80.711 44.406
77021 99.745 -80.717 44.396
77022 99.792 -80.723 44.386
77023 99.840 -80.728 44.375
77024 99.888 -80.734 44.363
77025 99.936 -80.739 44.353
77321 93.544 -78.882 53.534
77322 93.554 -78.880 53.525
77323 93.564 -78.877 53.515
77324 93.573 -78.874 53.506
77325 93.583 -78.872 53.498
77326 93.592 -78.870 53.490
77327 93.603 -78.860 53.483
77328 93.611 -78.853 53.476
77329 93.620 -78.846 53.469
77330 93.629 -78.839 53.465
77331 93.698 -78.805 53.463
77332 93.688 -78.802 53.464
77333 93.687 -78.801 53.464
77334 93.685 -78.799 53.462
77335 93.684 -78.797 53.459
77336 93.687 -78.795 53.453
77337 93.694 -78.795 53.446
77338 93.701 -78.794 53.438
77339 93.707 -78.793 53.427
77340 93.714 -78.793 53.417
77341 93.721 -78.789 53.407
77342 93.728 -78.786 53.396
77343 93.736 -78.783 53.383
77344 93.743 -78.781 53.370
77345 93.766 -78.768 53.355
77346 93.791 -78.754 53.338
77347 93.807 -78.742 53.321
77348 93.811 -78.734 53.303
77349 93.814 -78.725 53.285
77350 93.817 -78.717 53.265
77351 93.822 -78.708 53.244
77352 93.836 -78.701 53.223
77353 93.850 -78.692 53.202
77354 93.864 -78.684 53.182
77355 93.878 -78.677 53.163
77356 93.892 -78.668 53.141
77357 93.906 -78.661 53.117
77358 93.920 -78.652 53.091
77359 93.934 -78.645 53.065
77360 93.943 -78.639 53.040
77361 93.957 -78.634 53.013
77362 93.967 -78.633 52.986
77363 93.977 -78.631 52.960
77364 93.987 -78.629 52.936
77365 93.996 -78.627 52.913
77366 94.006 -78.626 52.890
77367 94.016 -78.624 52.869
77368 94.027 -78.623 52.849
77369 94.036 -78.620 52.830
77370 94.046 -78.619 52.810
77371 94.056 -78.617 52.788
77372 94.066 -78.616 52.765
77373 94.076 -78.614 52.742
77374 94.079 -78.617 52.720
77375 94.077 -78.624 52.697
77376 94.076 -78.606 52.674
77377 94.105 -78.551 52.650
77378 94.147 -78.529 52.625
77379 94.199 -78.542 52.599
77380 94.251 -78.556 52.573
77381 94.302 -78.570 52.551
77382 94.350 -78.582 52.529
77383 94.360 -78.579 52.508
77384 94.365 -78.580 52.489
77385 94.357 -78.592 52.467
77386 94.350 -78.604 52.445
77387 94.342 -78.617 52.424
77388 94.334 -78.628 52.402
77389 94.327 -78.640 52.380
77390 94.319 -78.652 52.357
77391 94.312 -78.664 52.335
77392 94.320 -78.660 52.312
77393 94.336 -78.647 52.288
77394 94.353 -78.635 52.261
77395 94.369 -78.623 52.235
77396 94.386 -78.611 52.209
77397 94.397 -78.604 52.184
77398 94.406 -78.601 52.158
77399 94.415 -78.599 52.129
77400 94.423 -78.597 52.098
77401 94.437 -78.596 52.065
77402 94.456 -78.597 52.030
77403 94.474 -78.598 51.996
77404 94.485 -78.602 51.964
77405 94.496 -78.616 51.929
77406 94.506 -78.649 51.894
77407 94.517 -78.683 51.859
77408 94.529 -78.702 51.825
77409 94.544 -78.701 51.787
77410 94.559 -78.700 51.748
77411 94.574 -78.699 51.708
77412 94.586 -78.696 51.668
77413 94.598 -78.692 51.629
77414 94.610 -78.689 51.589
77415 94.622 -78.685 51.554
77416 94.633 -78.680 51.520
77417 94.645 -78.677 51.486
77418 94.657 -78.673 51.455
77419 94.669 -78.669 51.423
77420 94.681 -78.665 51.392
77421 94.692 -78.661 51.361
77422 94.703 -78.663 51.325
77423 94.714 -78.665 51.288
77424 94.725 -78.667 51.251
77425 94.739 -78.670 51.214
77426 94.757 -78.673 51.176
77427 94.776 -78.677 51.138
77428 94.783 -78.684 51.099
77429 94.788 -78.692 51.059
77430 94.794 -78.699 51.018
77431 94.800 -78.707 50.977
77432 94.806 -78.714 50.935
77433 94.814 -78.704 50.894
77434 94.824 -78.691 50.853
77435 94.833 -78.678 50.809
77436 94.842 -78.665 50.764
77437 94.851 -78.652 50.720
77438 94.859 -78.645 50.677
77439 94.865 -78.641 50.636
77440 94.872 -78.637 50.591
77441 94.879 -78.634 50.546
77442 94.886 -78.638 50.500
77443 94.891 -78.649 50.452
77444 94.892 -78.664 50.405
77445 94.894 -78.677 50.361
77446 94.896 -78.692 50.322
77447 94.898 -78.705 50.284
77448 94.899 -78.720 50.246
77449 94.901 -78.733 50.206
77450 94.903 -78.748 50.166
77451 94.905 -78.762 50.124
77452 94.907 -78.776 50.080
77453 94.909 -78.789 50.034
77454 94.923 -78.794 49.989
77455 94.938 -78.799 49.945
77456 94.953 -78.803 49.902
77457 94.972 -78.805 49.860
77458 95.001 -78.802 49.819
77459 95.031 -78.799 49.779
77460 95.061 -78.795 49.738
77461 95.091 -78.791 49.700
77462 95.121 -78.787 49.662
77463 95.151 -78.783 49.623
77464 95.181 -78.780 49.584
77465 95.210 -78.777 49.544
77466 95.232 -78.774 49.504
77467 95.252 -78.793 49.464
77468 95.273 -78.805 49.423
77469 95.293 -78.817 49.383
77470 95.312 -78.829 49.341
77471 95.332 -78.840 49.301
77472 95.351 -78.852 49.262
77473 95.371 -78.864 49.222
77474 95.390 -78.876 49.183
77475 95.409 -78.886 49.143
77476 95.423 -78.883 49.103
77477 95.437 -78.881 49.062
77478 95.451 -78.878 49.023
77479 95.465 -78.876 48.983
77480 95.479 -78.874 48.943
77481 95.492 -78.871 48.903
77482 95.507 -78.869 48.862
77483 95.520 -78.867 48.820
77484 95.534 -78.865 48.779
77485 95.547 -78.875 48.737
77486 95.562 -78.888 48.696
77487 95.577 -78.902 48.655
77488 95.592 -78.915 48.613
77489 95.606 -78.928 48.572
77490 95.620 -78.941 48.531
77491 95.633 -78.955 48.492
77492 95.648 -78.968 48.451
77493 95.662 -78.981 48.408
77494 95.675 -78.994 48.362
77495 95.689 -79.008 48.313
77496 95.703 -79.021 48.265
77497 95.717 -79.034 48.219
77498 95.731 -79.048 48.176
77499 95.747 -79.056 48.133
77500 95.760 -79.057 48.090
77501 95.769 -79.053 48.046
77502 95.778 -79.049 48.005
77503 95.788 -79.046 47.963
77504 95.797 -79.042 47.919
77505 95.806 -79.038 47.877
77506 95.815 -79.034 47.835
77507 95.824 -79.031 47.793
77508 95.833 -79.027 47.750
77509 95.839 -79.023 47.707
77510 95.846 -79.020 47.664
77511 95.852 -79.016 47.620
77512 95.859 -79.012 47.578
77513 95.866 -79.009 47.537
77514 95.902 -79.005 47.496
77515 95.939 -79.001 47.456
77516 95.975 -78.997 47.416
77517 96.012 -78.993 47.376
77518 96.048 -78.990 47.336
77519 96.072 -78.984 47.297
77520 96.095 -78.977 47.259
77521 96.129 -78.976 47.222
77522 96.169 -78.977 47.184
77523 96.209 -78.979 47.148
77524 96.249 -78.981 47.112
77525 96.290 -78.982 47.078
77526 96.329 -78.984 47.044
77527 96.370 -78.985 47.012
77528 96.409 -78.987 46.981
77529 96.429 -78.995 46.950
77530 96.447 -79.004 46.920
77531 96.466 -79.014 46.890
77532 96.484 -79.023 46.858
77533 96.503 -79.033 46.828
77534 96.522 -79.042 46.797
77535 96.540 -79.052 46.766
77536 96.559 -79.061 46.732
77537 96.578 -79.071 46.700
77538 96.596 -79.080 46.668
77539 96.615 -79.090 46.637
77540 96.633 -79.100 46.607
77541 96.648 -79.106 46.576
77542 96.687 -79.109 46.545
77543 96.735 -79.111 46.516
77544 96.764 -79.103 46.487
77545 96.788 -79.093 46.459
77546 96.790 -79.075 46.431
77547 96.792 -79.056 46.405
77548 96.859 -79.021 46.380
77549 96.929 -78.984 46.355
77550 96.961 -78.938 46.330
77551 96.981 -78.954 46.305
77552 96.953 -78.981 46.280
77553 96.930 -79.009 46.255
77554 96.934 -78.979 46.233
77555 96.946 -78.954 46.212
77556 96.957 -78.928 46.191
77557 96.967 -78.933 46.170
77558 96.975 -78.952 46.149
77559 96.983 -78.980 46.129
77560 96.991 -79.009 46.108
77561 96.999 -79.037 46.091
77562 96.984 -79.002 46.074
77563 96.967 -78.962 46.056
77564 97.006 -78.972 46.037
77565 97.014 -79.100 46.019
77566 97.029 -79.228 46.000
77567 97.048 -79.267 45.982
77568 97.054 -79.259 45.965
77569 97.060 -79.250 45.948
77570 97.066 -79.242 45.928
77571 97.072 -79.234 45.911
77572 97.083 -79.222 45.895
77573 97.093 -79.211 45.878
77574 97.100 -79.192 45.859
77575 97.105 -79.170 45.843
77576 97.109 -79.147 45.828
77577 97.114 -79.125 45.811
77578 97.118 -79.102 45.796
77579 97.123 -79.079 45.779
77580 97.150 -79.084 45.761
77581 97.189 -79.106 45.742
77582 97.228 -79.129 45.726
77583 97.267 -79.152 45.711
77584 97.299 -79.151 45.695
77585 97.330 -79.146 45.680
77586 97.362 -79.140 45.664
77587 97.369 -79.154 45.649
77588 97.375 -79.169 45.634
77589 97.382 -79.182 45.621
77590 97.388 -79.197 45.609
77591 97.395 -79.210 45.596
77592 97.375 -79.237 45.582
77593 97.369 -79.250 45.567
77594 97.391 -79.251 45.552
77595 97.396 -79.265 45.537
77596 97.403 -79.271 45.523
77597 97.410 -79.278 45.509
77598 97.421 -79.280 45.496
77599 97.432 -79.283 45.483
77600 97.443 -79.286 45.469
77601 97.454 -79.288 45.455
77602 97.465 -79.291 45.441
77603 97.476 -79.294 45.427
77604 97.490 -79.300 45.414
77605 97.504 -79.306 45.400
77606 97.517 -79.312 45.387
77607 97.535 -79.324 45.373
77608 97.554 -79.338 45.358
77609 97.573 -79.352 45.344
77610 97.592 -79.366 45.330
77611 97.610 -79.380 45.317
77612 97.630 -79.393 45.303
77613 97.648 -79.400 45.288
77614 97.661 -79.407 45.275
77615 97.667 -79.413 45.264
77616 97.672 -79.419 45.256
77617 97.678 -79.424 45.247
77618 97.684 -79.430 45.239
77619 97.690 -79.436 45.229
77620 97.696 -79.442 45.219
77621 97.707 -79.457 45.209
77622 97.723 -79.488 45.198
77623 97.739 -79.520 45.187
77624 97.754 -79.540 45.176
77625 97.767 -79.547 45.168
77626 97.780 -79.554 45.155
77627 97.793 -79.562 45.142
77628 97.807 -79.573 45.129
77629 97.821 -79.585 45.115
77630 97.836 -79.597 45.102
77631 97.851 -79.608 45.092
77632 97.865 -79.620 45.087
77633 97.879 -79.632 45.081
77634 97.898 -79.644 45.074
77635 97.919 -79.656 45.065
77636 97.940 -79.667 45.057
77637 97.956 -79.675 45.049
77638 97.968 -79.679 45.041
77639 97.981 -79.684 45.033
77640 98.000 -79.693 45.023
77641 98.035 -79.711 45.013
77642 98.074 -79.732 45.005
77643 98.112 -79.752 44.997
77644 98.143 -79.767 44.989
77645 98.166 -79.780 44.980
77646 98.190 -79.789 44.970
77647 98.215 -79.793 44.960
77648 98.241 -79.798 44.949
77649 98.266 -79.802 44.937
77650 98.291 -79.806 44.923
77651 98.316 -79.811 44.910
77652 98.342 -79.815 44.898
77653 98.367 -79.819 44.887
77654 98.392 -79.825 44.880
77655 98.427 -79.835 44.871
77656 98.460 -79.843 44.862
77657 98.492 -79.850 44.851
77658 98.523 -79.858 44.841
77659 98.555 -79.864 44.832
77660 98.586 -79.872 44.824
77661 98.618 -79.879 44.816
77662 98.650 -79.886 44.809
77663 98.675 -79.895 44.802
77664 98.700 -79.903 44.794
77665 98.724 -79.912 44.785
77666 98.749 -79.921 44.777
77667 98.773 -79.930 44.770
77668 98.798 -79.938 44.764
77669 98.822 -79.947 44.757
77670 98.847 -79.955 44.749
77671 98.870 -79.956 44.740
77672 98.892 -79.952 44.731
77673 98.914 -79.947 44.724
77674 98.935 -79.943 44.717
77675 98.956 -79.942 44.711
77676 98.966 -79.930 44.705
77677 98.985 -79.943 44.700
77678 99.005 -79.960 44.694
77679 99.024 -79.977 44.688
77680 99.044 -79.994 44.682
77681 99.063 -80.012 44.677
77682 99.082 -80.028 44.672
77683 99.101 -80.046 44.667
77684 99.121 -80.062 44.663
77685 99.147 -80.077 44.657
77686 99.211 -80.102 44.652
77687 99.248 -80.121 44.646
77688 99.255 -80.126 44.640
77689 99.267 -80.126 44.635
77690 99.297 -80.150 44.631
77691 99.337 -80.151 44.628
77692 99.373 -80.151 44.623
77693 99.412 -80.155 44.617
77694 99.451 -80.159 44.613
77695 99.489 -80.162 44.608
77696 99.528 -80.166 44.602
77697 99.533 -80.187 44.597
77698 99.536 -80.209 44.592
77699 99.540 -80.231 44.586
77700 99.543 -80.254 44.581
77701 99.546 -80.276 44.575
77702 99.549 -80.298 44.567
77703 99.552 -80.320 44.561
77704 99.556 -80.343 44.555
77705 99.559 -80.365 44.548
77706 99.562 -80.387 44.540
77707 99.565 -80.409 44.533
77708 99.568 -80.431 44.527
77709 99.572 -80.453 44.519
77710 99.575 -80.475 44.510
77711 99.578 -80.497 44.503
77712 99.581 -80.519 44.495
77713 99.584 -80.542 44.487
77714 99.588 -80.564 44.478
77715 99.591 -80.586 44.469
77716 99.594 -80.608 44.460
77717 99.597 -80.630 44.451
77718 99.600 -80.653 44.441
77719 99.609 -80.673 44.431
77720 99.657 -80.679 44.421
77721 99.705 -80.684 44.411
77722 99.753 -80.690 44.402
77723 99.800 -80.695 44.390
77724 99.849 -80.701 44.379
77725 99.897 -80.707 44.367
77726 99.944 -80.712 44.356
77727 99.992 -80.718 44.346
77728 100.040 -80.723 44.335
78022 93.545 -78.871 53.545
78023 93.554 -78.868 53.536
78024 93.564 -78.866 53.527
78025 93.573 -78.864 53.518
78026 93.583 -78.861 53.509
78027 93.593 -78.858 53.501
78028 93.603 -78.855 53.494
78029 93.613 -78.847 53.488
78030 93.621 -78.840 53.481
78031 93.629 -78.833 53.478
78032 93.686 -78.805 53.477
78033 93.676 -78.801 53.478
78034 93.675 -78.799 53.478
78035 93.674 -78.797 53.477
78036 93.678 -78.795 53.474
78037 93.683 -78.790 53.468
78038 93.690 -78.787 53.461
78039 93.697 -78.786 53.453
78040 93.703 -78.786 53.443
78041 93.710 -78.785 53.432
78042 93.717 -78.782 53.421
78043 93.724 -78.780 53.410
78044 93.731 -78.777 53.397
78045 93.753 -78.765 53.382
78046 93.777 -78.751 53.366
78047 93.802 -78.737 53.349
78048 93.808 -78.728 53.331
78049 93.812 -78.720 53.313
78050 93.815 -78.711 53.295
78051 93.819 -78.702 53.274
78052 93.822 -78.694 53.254
78053 93.835 -78.686 53.233
78054 93.849 -78.678 53.212
78055 93.863 -78.670 53.192
78056 93.877 -78.662 53.172
78057 93.891 -78.654 53.150
78058 93.905 -78.646 53.126
78059 93.919 -78.638 53.101
78060 93.933 -78.630 53.075
78061 93.945 -78.619 53.048
78062 93.956 -78.625 53.021
78063 93.966 -78.624 52.995
78064 93.976 -78.623 52.970
78065 93.986 -78.621 52.946
78066 93.996 -78.620 52.923
78067 94.006 -78.617 52.900
78068 94.016 -78.616 52.881
78069 94.026 -78.614 52.862
78070 94.036 -78.613 52.843
78071 94.046 -78.611 52.824
78072 94.055 -78.610 52.802
78073 94.065 -78.608 52.778
78074 94.075 -78.606 52.754
78075 94.075 -78.611 52.731
78076 94.073 -78.616 52.708
78077 94.081 -78.567 52.685
78078 94.115 -78.522 52.661
78079 94.176 -78.524 52.636
78080 94.235 -78.527 52.610
78081 94.287 -78.541 52.586
78082 94.339 -78.554 52.565
78083 94.373 -78.561 52.544
78084 94.379 -78.562 52.525
78085 94.371 -78.574 52.506
78086 94.363 -78.586 52.485
78087 94.356 -78.598 52.463
78088 94.348 -78.611 52.441
78089 94.340 -78.623 52.419
78090 94.333 -78.634 52.397
78091 94.325 -78.646 52.374
78092 94.318 -78.658 52.355
78093 94.322 -78.658 52.332
78094 94.339 -78.645 52.308
78095 94.355 -78.633 52.282
78096 94.371 -78.620 52.256
78097 94.388 -78.608 52.231
78098 94.404 -78.596 52.207
78099 94.415 -78.590 52.179
78100 94.424 -78.588 52.150
78101 94.432 -78.586 52.118
78102 94.447 -78.585 52.083
78103 94.465 -78.586 52.049
78104 94.483 -78.586 52.015
78105 94.497 -78.589 51.982
78106 94.508 -78.618 51.946
78107 94.519 -78.651 51.909
78108 94.529 -78.685 51.872
78109 94.541 -78.706 51.839
78110 94.553 -78.702 51.801
78111 94.564 -78.698 51.761
78112 94.576 -78.695 51.720
78113 94.588 -78.691 51.681
78114 94.600 -78.687 51.641
78115 94.612 -78.683 51.602
78116 94.624 -78.680 51.566
78117 94.635 -78.676 51.532
78118 94.647 -78.672 51.498
78119 94.659 -78.668 51.465
78120 94.671 -78.664 51.432
78121 94.683 -78.660 51.399
78122 94.694 -78.656 51.366
78123 94.705 -78.657 51.331
78124 94.716 -78.659 51.295
78125 94.727 -78.661 51.258
78126 94.745 -78.664 51.221
78127 94.764 -78.668 51.183
78128 94.782 -78.672 51.145
78129 94.792 -78.678 51.105
78130 94.798 -78.686 51.066
78131 94.804 -78.693 51.026
78132 94.810 -78.701 50.986
78133 94.815 -78.706 50.944
78134 94.825 -78.693 50.903
78135 94.834 -78.680 50.861
78136 94.843 -78.667 50.818
78137 94.852 -78.655 50.773
78138 94.861 -78.642 50.728
78139 94.871 -78.629 50.684
78140 94.879 -78.619 50.642
78141 94.886 -78.616 50.597
78142 94.893 -78.612 50.550
78143 94.899 -78.622 50.503
78144 94.906 -78.631 50.455
78145 94.911 -78.642 50.409
78146 94.913 -78.656 50.364
78147 94.915 -78.670 50.324
78148 94.917 -78.684 50.285
78149 94.918 -78.698 50.246
78150 94.920 -78.713 50.205
78151 94.922 -78.727 50.164
78152 94.924 -78.741 50.121
78153 94.926 -78.755 50.078
78154 94.933 -78.765 50.032
78155 94.948 -78.769 49.988
78156 94.962 -78.774 49.945
78157 94.977 -78.778 49.902
78158 94.992 -78.783 49.860
78159 95.010 -78.786 49.819
78160 95.039 -78.782 49.779
78161 95.069 -78.778 49.739
78162 95.099 -78.775 49.700
78163 95.129 -78.771 49.662
78164 95.159 -78.767 49.624
78165 95.189 -78.764 49.586
78166 95.218 -78.760 49.546
78167 95.215 -78.761 49.505
78168 95.244 -78.780 49.465
78169 95.273 -78.800 49.424
78170 95.293 -78.812 49.384
78171 95.313 -78.824 49.343
78172 95.332 -78.836 49.303
78173 95.351 -78.848 49.263
78174 95.371 -78.859 49.223
78175 95.391 -78.871 49.185
78176 95.409 -78.881 49.146
78177 95.423 -78.879 49.106
78178 95.438 -78.877 49.066
78179 95.451 -78.874 49.025
78180 95.465 -78.872 48.986
78181 95.479 -78.870 48.946
78182 95.493 -78.868 48.905
78183 95.507 -78.865 48.864
78184 95.521 -78.863 48.823
78185 95.529 -78.855 48.781
78186 95.545 -78.852 48.739
78187 95.551 -78.863 48.698
78188 95.562 -78.875 48.657
78189 95.576 -78.888 48.614
78190 95.590 -78.902 48.572
78191 95.604 -78.915 48.530
78192 95.617 -78.928 48.488
78193 95.631 -78.942 48.445
78194 95.645 -78.955 48.401
78195 95.659 -78.968 48.357
78196 95.673 -78.981 48.310
78197 95.687 -78.995 48.265
78198 95.701 -79.008 48.222
78199 95.715 -79.021 48.181
78200 95.730 -79.031 48.140
78201 95.747 -79.034 48.097
78202 95.764 -79.038 48.054
78203 95.780 -79.042 48.013
78204 95.796 -79.044 47.970
78205 95.805 -79.040 47.928
78206 95.814 -79.037 47.884
78207 95.823 -79.033 47.842
78208 95.832 -79.029 47.800
78209 95.841 -79.025 47.757
78210 95.849 -79.021 47.714
78211 95.855 -79.018 47.672
78212 95.862 -79.014 47.628
78213 95.869 -79.010 47.586
78214 95.897 -79.006 47.545
78215 95.933 -79.002 47.504
78216 95.970 -78.999 47.465
78217 96.005 -78.995 47.425
78218 96.028 -78.989 47.385
78219 96.051 -78.983 47.345
78220 96.075 -78.976 47.307
78221 96.098 -78.970 47.268
78222 96.129 -78.968 47.231
78223 96.169 -78.969 47.193
78224 96.209 -78.971 47.157
78225 96.249 -78.972 47.120
78226 96.289 -78.974 47.085
78227 96.329 -78.975 47.050
78228 96.369 -78.977 47.018
78229 96.404 -78.971 46.988
78230 96.416 -78.971 46.957
78231 96.422 -78.979 46.926
78232 96.441 -78.989 46.896
78233 96.460 -78.998 46.865
78234 96.478 -79.008 46.835
78235 96.497 -79.018 46.804
78236 96.516 -79.027 46.773
78237 96.534 -79.037 46.741
78238 96.553 -79.046 46.709
78239 96.572 -79.056 46.677
78240 96.590 -79.065 46.645
78241 96.605 -79.072 46.614
78242 96.620 -79.079 46.583
78243 96.668 -79.081 46.552
78244 96.717 -79.083 46.522
78245 96.756 -79.081 46.494
78246 96.785 -79.072 46.466
78247 96.814 -79.065 46.438
78248 96.826 -79.050 46.412
78249 96.879 -79.011 46.386
78250 96.911 -78.963 46.361
78251 96.937 -78.947 46.336
78252 96.957 -78.963 46.311
78253 96.946 -78.985 46.287
78254 96.924 -78.996 46.263
78255 96.934 -78.953 46.242
78256 96.945 -78.927 46.221
78257 96.956 -78.905 46.201
78258 96.965 -78.924 46.181
78259 96.974 -78.942 46.161
78260 96.982 -78.967 46.141
78261 96.990 -78.996 46.120
78262 96.990 -79.002 46.101
78263 96.974 -78.964 46.083
78264 96.970 -78.967 46.065
78265 96.945 -79.078 46.046
78266 97.004 -79.047 46.028
78267 97.019 -79.192 46.009
78268 97.037 -79.231 45.992
78269 97.056 -79.269 45.976
78270 97.065 -79.270 45.959
78271 97.071 -79.262 45.941
78272 97.077 -79.253 45.924
78273 97.087 -79.242 45.907
78274 97.097 -79.230 45.890
78275 97.102 -79.207 45.871
78276 97.106 -79.185 45.855
78277 97.111 -79.162 45.839
78278 97.115 -79.140 45.823
78279 97.124 -79.109 45.807
78280 97.133 -79.078 45.791
78281 97.149 -79.079 45.774
78282 97.188 -79.102 45.757
78283 97.227 -79.125 45.741
78284 97.265 -79.147 45.726
78285 97.297 -79.143 45.710
78286 97.328 -79.138 45.696
78287 97.346 -79.144 45.680
78288 97.353 -79.158 45.664
78289 97.359 -79.172 45.650
78290 97.366 -79.186 45.637
78291 97.372 -79.200 45.625
78292 97.379 -79.214 45.613
78293 97.385 -79.228 45.600
78294 97.388 -79.244 45.586
78295 97.393 -79.253 45.571
78296 97.399 -79.259 45.555
78297 97.406 -79.266 45.540
78298 97.414 -79.272 45.527
78299 97.425 -79.275 45.515
78300 97.436 -79.278 45.502
78301 97.447 -79.280 45.489
78302 97.457 -79.283 45.475
78303 97.468 -79.285 45.460
78304 97.479 -79.288 45.446
78305 97.491 -79.291 45.433
78306 97.504 -79.296 45.419
78307 97.517 -79.302 45.405
78308 97.532 -79.310 45.390
78309 97.551 -79.325 45.374
78310 97.571 -79.338 45.359
78311 97.589 -79.353 45.346
78312 97.608 -79.363 45.333
78313 97.627 -79.370 45.317
78314 97.646 -79.378 45.301
78315 97.664 -79.385 45.287
78316 97.677 -79.391 45.277
78317 97.683 -79.397 45.268
78318 97.689 -79.404 45.260
78319 97.695 -79.410 45.252
78320 97.703 -79.418 45.243
78321 97.714 -79.433 45.233
78322 97.725 -79.448 45.221
78323 97.736 -79.463 45.211
78324 97.749 -79.486 45.200
78325 97.766 -79.515 45.191
78326 97.778 -79.523 45.182
78327 97.791 -79.531 45.171
78328 97.806 -79.542 45.156
78329 97.820 -79.554 45.140
78330 97.835 -79.566 45.125
78331 97.849 -79.577 45.112
78332 97.864 -79.589 45.103
78333 97.878 -79.601 45.099
78334 97.892 -79.613 45.095
78335 97.909 -79.625 45.089
78336 97.930 -79.636 45.081
78337 97.950 -79.647 45.074
78338 97.973 -79.657 45.065
78339 97.992 -79.667 45.056
78340 98.012 -79.676 45.046
78341 98.031 -79.686 45.037
78342 98.051 -79.695 45.028
78343 98.082 -79.711 45.020
78344 98.121 -79.732 45.012
78345 98.154 -79.749 45.003
78346 98.177 -79.761 44.993
78347 98.202 -79.767 44.981
78348 98.227 -79.770 44.972
78349 98.252 -79.775 44.960
78350 98.278 -79.780 44.948
78351 98.303 -79.783 44.935
78352 98.329 -79.788 44.922
78353 98.354 -79.792 44.908
78354 98.379 -79.796 44.898
78355 98.405 -79.802 44.890
78356 98.435 -79.811 44.882
78357 98.468 -79.820 44.871
78358 98.501 -79.830 44.860
78359 98.533 -79.839 44.849
78360 98.566 -79.848 44.840
78361 98.598 -79.857 44.832
78362 98.631 -79.867 44.825
78363 98.662 -79.874 44.818
78364 98.679 -79.872 44.811
78365 98.700 -79.880 44.803
78366 98.721 -79.887 44.795
78367 98.743 -79.896 44.788
78368 98.768 -79.904 44.781
78369 98.793 -79.913 44.774
78370 98.817 -79.921 44.768
78371 98.842 -79.930 44.760
78372 98.865 -79.931 44.751
78373 98.887 -79.927 44.743
78374 98.909 -79.922 44.736
78375 98.932 -79.918 44.730
78376 98.957 -79.919 44.724
78377 98.973 -79.914 44.718
78378 98.992 -79.926 44.713
78379 99.011 -79.943 44.707
78380 99.030 -79.960 44.701
78381 99.050 -79.977 44.696
78382 99.069 -79.994 44.691
78383 99.089 -80.011 44.686
78384 99.108 -80.028 44.681
78385 99.128 -80.045 44.675
78386 99.157 -80.059 44.669
78387 99.189 -80.071 44.664
78388 99.243 -80.093 44.658
78389 99.272 -80.109 44.653
78390 99.277 -80.118 44.650
78391 99.305 -80.127 44.647
78392 99.353 -80.137 44.644
78393 99.396 -80.142 44.639
78394 99.425 -80.128 44.634
78395 99.462 -80.128 44.631
78396 99.500 -80.132 44.626
78397 99.526 -80.143 44.620
78398 99.529 -80.165 44.612
78399 99.533 -80.187 44.604
78400 99.536 -80.209 44.597
78401 99.539 -80.231 44.590
78402 99.543 -80.253 44.583
78403 99.546 -80.275 44.575
78404 99.549 -80.297 44.568
78405 99.552 -80.319 44.562
78406 99.555 -80.342 44.555
78407 99.559 -80.364 44.547
78408 99.562 -80.386 44.539
78409 99.565 -80.408 44.533
78410 99.568 -80.431 44.524
78411 99.571 -80.453 44.516
78412 99.575 -80.475 44.509
78413 99.578 -80.497 44.501
78414 99.581 -80.519 44.492
78415 99.584 -80.541 44.484
78416 99.587 -80.563 44.475
78417 99.591 -80.585 44.466
78418 99.594 -80.607 44.457
78419 99.597 -80.630 44.446
78420 99.617 -80.646 44.436
78421 99.665 -80.651 44.427
78422 99.713 -80.657 44.417
78423 99.761 -80.663 44.406
78424 99.809 -80.668 44.394
78425 99.857 -80.674 44.383
78426 99.905 -80.679 44.372
78427 99.952 -80.685 44.361
78428 100.001 -80.691 44.350
78429 100.049 -80.696 44.339
78430 100.097 -80.701 44.328
78724 93.554 -78.858 53.546
78725 93.564 -78.855 53.537
78726 93.574 -78.852 53.528
78727 93.583 -78.850 53.520
78728 93.593 -78.848 53.512
78729 93.603 -78.846 53.504
78730 93.613 -78.841 53.499
78731 93.622 -78.833 53.493
78732 93.630 -78.827 53.490
78733 93.674 -78.804 53.490
78734 93.664 -78.799 53.492
78735 93.665 -78.796 53.491
78736 93.670 -78.794 53.490
78737 93.675 -78.791 53.487
78738 93.680 -78.786 53.482
78739 93.686 -78.782 53.475
78740 93.692 -78.779 53.467
78741 93.699 -78.779 53.457
78742 93.706 -78.777 53.446
78743 93.713 -78.775 53.435
78744 93.720 -78.772 53.423
78745 93.739 -78.762 53.409
78746 93.764 -78.748 53.393
78747 93.788 -78.734 53.377
78748 93.806 -78.722 53.360
78749 93.809 -78.714 53.342
78750 93.813 -78.705 53.324
78751 93.816 -78.697 53.306
78752 93.820 -78.688 53.285
78753 93.823 -78.680 53.263
78754 93.834 -78.671 53.242
78755 93.848 -78.664 53.221
78756 93.862 -78.655 53.200
78757 93.876 -78.648 53.180
78758 93.890 -78.639 53.159
78759 93.904 -78.632 53.135
78760 93.918 -78.622 53.110
78761 93.930 -78.609 53.083
78762 93.943 -78.596 53.056
78763 93.954 -78.601 53.029
78764 93.965 -78.608 53.003
78765 93.975 -78.614 52.979
78766 93.985 -78.613 52.956
78767 93.995 -78.611 52.933
78768 94.005 -78.610 52.911
78769 94.015 -78.608 52.892
78770 94.025 -78.607 52.874
78771 94.035 -78.604 52.855
78772 94.045 -78.603 52.837
78773 94.055 -78.601 52.815
78774 94.065 -78.600 52.791
78775 94.073 -78.599 52.766
78776 94.072 -78.606 52.743
78777 94.065 -78.582 52.720
78778 94.085 -78.529 52.696
78779 94.140 -78.520 52.672
78780 94.201 -78.523 52.647
78781 94.262 -78.525 52.622
78782 94.323 -78.526 52.599
78783 94.374 -78.539 52.579
78784 94.393 -78.544 52.560
78785 94.385 -78.556 52.541
78786 94.378 -78.568 52.522
78787 94.369 -78.580 52.502
78788 94.362 -78.592 52.480
78789 94.354 -78.604 52.458
78790 94.347 -78.617 52.435
78791 94.339 -78.629 52.414
78792 94.331 -78.640 52.392
78793 94.324 -78.652 52.372
78794 94.324 -78.656 52.350
78795 94.341 -78.644 52.325
78796 94.357 -78.631 52.301
78797 94.373 -78.619 52.276
78798 94.390 -78.607 52.251
78799 94.406 -78.594 52.227
78800 94.422 -78.582 52.199
78801 94.433 -78.576 52.169
78802 94.442 -78.575 52.137
78803 94.456 -78.574 52.102
78804 94.474 -78.574 52.068
78805 94.492 -78.575 52.034
78806 94.506 -78.580 51.998
78807 94.515 -78.611 51.961
78808 94.524 -78.642 51.924
78809 94.533 -78.673 51.886
78810 94.542 -78.692 51.850
78811 94.554 -78.692 51.813
78812 94.566 -78.691 51.774
78813 94.578 -78.689 51.733
78814 94.590 -78.686 51.693
78815 94.602 -78.682 51.654
78816 94.614 -78.678 51.615
78817 94.626 -78.674 51.578
78818 94.637 -78.670 51.543
78819 94.649 -78.667 51.509
78820 94.661 -78.663 51.474
78821 94.673 -78.659 51.440
78822 94.685 -78.655 51.406
78823 94.696 -78.651 51.372
78824 94.708 -78.651 51.338
78825 94.718 -78.653 51.302
78826 94.733 -78.656 51.265
78827 94.751 -78.659 51.227
78828 94.770 -78.663 51.190
78829 94.789 -78.667 51.151
78830 94.802 -78.672 51.111
78831 94.808 -78.680 51.073
78832 94.814 -78.687 51.034
78833 94.819 -78.694 50.993
78834 94.826 -78.696 50.953
78835 94.835 -78.683 50.911
78836 94.844 -78.670 50.869
78837 94.853 -78.658 50.827
78838 94.862 -78.645 50.781
78839 94.872 -78.632 50.735
78840 94.881 -78.619 50.691
78841 94.890 -78.606 50.647
78842 94.899 -78.594 50.601
78843 94.906 -78.596 50.554
78844 94.913 -78.605 50.506
78845 94.920 -78.615 50.458
78846 94.927 -78.624 50.411
78847 94.932 -78.635 50.366
78848 94.934 -78.649 50.325
78849 94.935 -78.664 50.284
78850 94.937 -78.677 50.244
78851 94.939 -78.692 50.202
78852 94.941 -78.705 50.160
78853 94.943 -78.720 50.118
78854 94.945 -78.733 50.074
78855 94.957 -78.740 50.030
78856 94.972 -78.745 49.987
78857 94.986 -78.749 49.944
78858 95.001 -78.753 49.902
78859 95.016 -78.758 49.860
78860 95.031 -78.762 49.819
78861 95.047 -78.766 49.779
78862 95.077 -78.762 49.741
78863 95.107 -78.758 49.702
78864 95.137 -78.755 49.664
78865 95.167 -78.751 49.627
78866 95.196 -78.748 49.589
78867 95.227 -78.744 49.549
78868 95.207 -78.749 49.507
78869 95.236 -78.769 49.466
78870 95.265 -78.789 49.426
78871 95.294 -78.808 49.386
78872 95.313 -78.820 49.345
78873 95.332 -78.831 49.306
78874 95.352 -78.843 49.267
78875 95.371 -78.855 49.228
78876 95.391 -78.866 49.189
78877 95.410 -78.877 49.149
78878 95.424 -78.875 49.109
78879 95.438 -78.873 49.069
78880 95.452 -78.871 49.029
78881 95.466 -78.868 48.989
78882 95.479 -78.866 48.949
78883 95.493 -78.864 48.908
78884 95.508 -78.861 48.867
78885 95.516 -78.854 48.825
78886 95.520 -78.843 48.783
78887 95.535 -78.839 48.741
78888 95.554 -78.843 48.699
78889 95.559 -78.854 48.657
78890 95.565 -78.864 48.615
78891 95.573 -78.875 48.572
78892 95.588 -78.889 48.530
78893 95.601 -78.902 48.487
78894 95.615 -78.915 48.443
78895 95.629 -78.928 48.399
78896 95.643 -78.942 48.355
78897 95.657 -78.955 48.312
78898 95.671 -78.968 48.270
78899 95.685 -78.982 48.228
78900 95.699 -78.995 48.188
78901 95.714 -79.006 48.146
78902 95.731 -79.009 48.105
78903 95.747 -79.013 48.063
78904 95.764 -79.017 48.021
78905 95.781 -79.021 47.979
78906 95.798 -79.024 47.935
78907 95.815 -79.028 47.893
78908 95.830 -79.031 47.850
78909 95.839 -79.027 47.807
78910 95.848 -79.023 47.764
78911 95.858 -79.019 47.721
78912 95.865 -79.015 47.679
78913 95.872 -79.012 47.636
78914 95.891 -79.009 47.595
78915 95.927 -79.005 47.554
78916 95.961 -79.000 47.513
78917 95.984 -78.994 47.474
78918 96.007 -78.987 47.433
78919 96.030 -78.981 47.393
78920 96.053 -78.975 47.354
78921 96.077 -78.969 47.315
78922 96.100 -78.963 47.277
78923 96.129 -78.960 47.240
78924 96.169 -78.962 47.201
78925 96.209 -78.963 47.165
78926 96.249 -78.965 47.129
78927 96.289 -78.966 47.092
78928 96.329 -78.968 47.058
78929 96.365 -78.964 47.026
78930 96.398 -78.956 46.995
78931 96.424 -78.951 46.964
78932 96.419 -78.957 46.933
78933 96.416 -78.964 46.902
78934 96.435 -78.974 46.872
78935 96.454 -78.983 46.842
78936 96.472 -78.993 46.812
78937 96.491 -79.002 46.780
78938 96.510 -79.012 46.749
78939 96.528 -79.021 46.717
78940 96.547 -79.031 46.685
78941 96.562 -79.038 46.653
78942 96.576 -79.045 46.621
78943 96.601 -79.050 46.589
78944 96.649 -79.052 46.558
78945 96.697 -79.054 46.529
78946 96.745 -79.056 46.500
78947 96.778 -79.049 46.472
78948 96.806 -79.042 46.444
78949 96.844 -79.036 46.418
78950 96.861 -78.990 46.392
78951 96.892 -78.943 46.367
78952 96.913 -78.956 46.342
78953 96.933 -78.972 46.318
78954 96.942 -78.990 46.294
78955 96.939 -78.993 46.271
78956 96.934 -78.929 46.251
78957 96.945 -78.899 46.231
78958 96.955 -78.895 46.212
78959 96.963 -78.914 46.193
78960 96.972 -78.932 46.173
78961 96.980 -78.954 46.153
78962 96.988 -78.983 46.131
78963 96.979 -78.964 46.112
78964 96.973 -78.959 46.093
78965 96.973 -78.974 46.074
78966 96.973 -78.990 46.056
78967 96.997 -79.008 46.038
78968 97.008 -79.155 46.020
78969 97.027 -79.194 46.003
78970 97.046 -79.233 45.987
78971 97.065 -79.272 45.971
78972 97.076 -79.281 45.955
78973 97.082 -79.273 45.937
78974 97.092 -79.263 45.919
78975 97.099 -79.245 45.901
78976 97.103 -79.223 45.883
78977 97.108 -79.200 45.867
78978 97.116 -79.171 45.852
78979 97.125 -79.140 45.836
78980 97.134 -79.109 45.820
78981 97.143 -79.078 45.804
78982 97.158 -79.081 45.788
78983 97.186 -79.097 45.771
78984 97.226 -79.120 45.756
78985 97.264 -79.141 45.741
78986 97.295 -79.136 45.726
78987 97.323 -79.134 45.712
78988 97.330 -79.147 45.697
78989 97.336 -79.162 45.681
78990 97.343 -79.175 45.666
78991 97.350 -79.190 45.653
78992 97.356 -79.203 45.639
78993 97.363 -79.217 45.626
78994 97.371 -79.231 45.616
78995 97.384 -79.240 45.604
78996 97.395 -79.248 45.590
78997 97.402 -79.255 45.572
78998 97.409 -79.261 45.557
78999 97.417 -79.266 45.546
79000 97.428 -79.269 45.536
79001 97.439 -79.272 45.523
79002 97.450 -79.275 45.509
79003 97.461 -79.278 45.495
79004 97.472 -79.280 45.481
79005 97.483 -79.283 45.467
79006 97.494 -79.285 45.452
79007 97.505 -79.288 45.437
79008 97.517 -79.292 45.422
79009 97.530 -79.297 45.407
79010 97.549 -79.311 45.393
79011 97.568 -79.325 45.376
79012 97.587 -79.334 45.361
79013 97.606 -79.341 45.345
79014 97.625 -79.348 45.328
79015 97.643 -79.355 45.312
79016 97.662 -79.363 45.299
79017 97.681 -79.370 45.291
79018 97.694 -79.376 45.282
79019 97.700 -79.382 45.274
79020 97.710 -79.395 45.266
79021 97.721 -79.410 45.257
79022 97.732 -79.426 45.247
79023 97.743 -79.441 45.236
79024 97.754 -79.456 45.224
79025 97.766 -79.471 45.212
79026 97.777 -79.486 45.202
79027 97.790 -79.500 45.193
79028 97.804 -79.513 45.177
79029 97.817 -79.527 45.161
79030 97.831 -79.541 45.145
79031 97.845 -79.554 45.131
79032 97.858 -79.569 45.118
79033 97.871 -79.582 45.109
79034 97.885 -79.596 45.104
79035 97.897 -79.608 45.098
79036 97.911 -79.620 45.092
79037 97.931 -79.631 45.086
79038 97.952 -79.642 45.079
79039 97.975 -79.652 45.073
79040 98.004 -79.661 45.066
79041 98.032 -79.670 45.057
79042 98.061 -79.679 45.049
79043 98.083 -79.689 45.041
79044 98.102 -79.698 45.033
79045 98.129 -79.711 45.025
79046 98.165 -79.730 45.016
79047 98.189 -79.739 45.006
79048 98.214 -79.744 44.995
79049 98.239 -79.748 44.984
79050 98.265 -79.752 44.972
79051 98.290 -79.757 44.959
79052 98.316 -79.763 44.946
79053 98.343 -79.770 44.931
79054 98.369 -79.776 44.918
79055 98.395 -79.783 44.907
79056 98.422 -79.789 44.899
79057 98.449 -79.800 44.890
79058 98.482 -79.809 44.879
79059 98.515 -79.818 44.869
79060 98.548 -79.828 44.858
79061 98.580 -79.837 44.849
79062 98.613 -79.846 44.841
79063 98.646 -79.855 44.834
79064 98.670 -79.858 44.827
79065 98.685 -79.852 44.821
79066 98.706 -79.858 44.812
79067 98.727 -79.866 44.806
79068 98.748 -79.874 44.799
79069 98.769 -79.880 44.793
79070 98.790 -79.888 44.785
79071 98.812 -79.896 44.779
79072 98.837 -79.905 44.772
79073 98.860 -79.906 44.764
79074 98.883 -79.904 44.757
79075 98.908 -79.904 44.751
79076 98.934 -79.905 44.744
79077 98.959 -79.905 44.739
79078 98.981 -79.904 44.734
79079 98.998 -79.909 44.728
79080 99.018 -79.926 44.721
79081 99.038 -79.943 44.714
79082 99.057 -79.960 44.708
79083 99.076 -79.977 44.704
79084 99.095 -79.994 44.699
79085 99.115 -80.011 44.694
79086 99.135 -80.027 44.689
79087 99.167 -80.040 44.682
79088 99.199 -80.052 44.676
79089 99.231 -80.065 44.671
79090 99.276 -80.083 44.667
79091 99.296 -80.096 44.664
79092 99.303 -80.098 44.661
79093 99.349 -80.108 44.658
79094 99.404 -80.122 44.654
79095 99.454 -80.133 44.650
79096 99.484 -80.119 44.647
79097 99.513 -80.106 44.642
79098 99.523 -80.120 44.635
79099 99.526 -80.142 44.627
79100 99.529 -80.164 44.617
79101 99.533 -80.186 44.608
79102 99.536 -80.208 44.599
79103 99.539 -80.231 44.591
79104 99.542 -80.253 44.583
79105 99.545 -80.275 44.576
79106 99.549 -80.297 44.569
79107 99.552 -80.319 44.561
79108 99.555 -80.341 44.554
79109 99.558 -80.363 44.546
79110 99.561 -80.385 44.539
79111 99.565 -80.407 44.530
79112 99.568 -80.430 44.521
79113 99.571 -80.452 44.513
79114 99.574 -80.474 44.506
79115 99.577 -80.496 44.498
79116 99.581 -80.519 44.489
79117 99.584 -80.541 44.481
79118 99.587 -80.563 44.471
79119 99.590 -80.585 44.462
79120 99.593 -80.607 44.451
79121 99.625 -80.619 44.441
79122 99.674 -80.624 44.431
79123 99.721 -80.629 44.421
79124 99.769 -80.635 44.410
79125 99.817 -80.641 44.399
79126 99.865 -80.647 44.387
79127 99.913 -80.652 44.376
79128 99.961 -80.657 44.365
79129 100.009 -80.663 44.353
79130 100.057 -80.669 44.342
79131 100.105 -80.675 44.331
79425 93.555 -78.846 53.555
79426 93.564 -78.844 53.545
79427 93.574 -78.842 53.537
79428 93.584 -78.839 53.529
79429 93.593 -78.836 53.521
79430 93.603 -78.834 53.514
79431 93.613 -78.832 53.510
79432 93.623 -78.827 53.505
79433 93.631 -78.820 53.502
79434 93.662 -78.803 53.503
79435 93.658 -78.796 53.505
79436 93.662 -78.794 53.504
79437 93.666 -78.791 53.502
79438 93.672 -78.788 53.499
79439 93.677 -78.783 53.495
79440 93.683 -78.778 53.489
79441 93.689 -78.774 53.481
79442 93.695 -78.771 53.471
79443 93.702 -78.771 53.460
79444 93.709 -78.767 53.448
79445 93.726 -78.759 53.435
79446 93.751 -78.745 53.420
79447 93.775 -78.731 53.404
79448 93.800 -78.717 53.388
79449 93.807 -78.708 53.370
79450 93.811 -78.699 53.352
79451 93.814 -78.691 53.333
79452 93.818 -78.683 53.315
79453 93.821 -78.673 53.294
79454 93.824 -78.665 53.272
79455 93.833 -78.657 53.251
79456 93.847 -78.649 53.230
79457 93.861 -78.641 53.209
79458 93.875 -78.633 53.188
79459 93.889 -78.625 53.167
79460 93.903 -78.614 53.143
79461 93.915 -78.601 53.118
79462 93.928 -78.588 53.091
79463 93.940 -78.575 53.064
79464 93.952 -78.579 53.038
79465 93.963 -78.585 53.013
79466 93.973 -78.592 52.990
79467 93.984 -78.598 52.967
79468 93.994 -78.604 52.944
79469 94.004 -78.601 52.923
79470 94.014 -78.600 52.904
79471 94.024 -78.598 52.886
79472 94.034 -78.597 52.868
79473 94.044 -78.595 52.848
79474 94.054 -78.594 52.827
79475 94.064 -78.592 52.804
79476 94.070 -78.593 52.779
79477 94.067 -78.590 52.755
79478 94.062 -78.546 52.731
79479 94.105 -78.517 52.707
79480 94.166 -78.520 52.682
79481 94.226 -78.521 52.658
79482 94.287 -78.523 52.634
79483 94.348 -78.526 52.612
79484 94.400 -78.521 52.594
79485 94.399 -78.539 52.575
79486 94.391 -78.550 52.556
79487 94.384 -78.562 52.538
79488 94.376 -78.574 52.518
79489 94.368 -78.586 52.496
79490 94.360 -78.598 52.474
79491 94.353 -78.611 52.451
79492 94.345 -78.623 52.430
79493 94.337 -78.635 52.409
79494 94.330 -78.646 52.389
79495 94.326 -78.655 52.366
79496 94.343 -78.642 52.343
79497 94.359 -78.630 52.319
79498 94.376 -78.617 52.295
79499 94.392 -78.604 52.271
79500 94.408 -78.592 52.245
79501 94.424 -78.580 52.217
79502 94.441 -78.567 52.187
79503 94.451 -78.564 52.155
79504 94.465 -78.563 52.121
79505 94.483 -78.563 52.086
79506 94.495 -78.570 52.051
79507 94.507 -78.578 52.014
79508 94.518 -78.594 51.976
79509 94.527 -78.624 51.937
79510 94.536 -78.655 51.899
79511 94.544 -78.673 51.862
79512 94.555 -78.675 51.824
79513 94.567 -78.674 51.786
79514 94.579 -78.673 51.747
79515 94.591 -78.673 51.708
79516 94.604 -78.673 51.670
79517 94.616 -78.672 51.631
79518 94.628 -78.669 51.593
79519 94.639 -78.665 51.557
79520 94.651 -78.661 51.521
79521 94.663 -78.658 51.486
79522 94.675 -78.654 51.450
79523 94.687 -78.650 51.414
79524 94.698 -78.646 51.378
79525 94.710 -78.645 51.344
79526 94.720 -78.647 51.308
79527 94.739 -78.651 51.271
79528 94.758 -78.654 51.232
79529 94.777 -78.658 51.194
79530 94.795 -78.661 51.156
79531 94.812 -78.665 51.116
79532 94.818 -78.673 51.078
79533 94.823 -78.680 51.039
79534 94.829 -78.688 50.999
79535 94.836 -78.686 50.959
79536 94.845 -78.673 50.918
79537 94.854 -78.661 50.876
79538 94.864 -78.648 50.833
79539 94.873 -78.635 50.787
79540 94.882 -78.622 50.742
79541 94.892 -78.609 50.697
79542 94.901 -78.596 50.652
79543 94.919 -78.587 50.606
79544 94.929 -78.586 50.558
79545 94.927 -78.589 50.510
79546 94.934 -78.598 50.461
79547 94.941 -78.608 50.413
79548 94.948 -78.617 50.365
79549 94.953 -78.628 50.322
79550 94.955 -78.642 50.281
79551 94.957 -78.656 50.240
79552 94.958 -78.670 50.199
79553 94.960 -78.684 50.157
79554 94.962 -78.698 50.114
79555 94.966 -78.711 50.071
79556 94.981 -78.715 50.028
79557 94.995 -78.720 49.985
79558 95.011 -78.724 49.943
79559 95.025 -78.728 49.900
79560 95.040 -78.733 49.860
79561 95.055 -78.737 49.820
79562 95.069 -78.742 49.781
79563 95.085 -78.745 49.744
79564 95.115 -78.742 49.706
79565 95.144 -78.739 49.668
79566 95.175 -78.735 49.631
79567 95.204 -78.731 49.592
79568 95.226 -78.729 49.551
79569 95.199 -78.738 49.510
79570 95.228 -78.758 49.468
79571 95.257 -78.777 49.428
79572 95.286 -78.797 49.388
79573 95.313 -78.815 49.348
79574 95.332 -78.827 49.310
79575 95.352 -78.838 49.272
79576 95.372 -78.850 49.234
79577 95.391 -78.861 49.194
79578 95.410 -78.874 49.154
79579 95.424 -78.871 49.114
79580 95.438 -78.868 49.073
79581 95.452 -78.866 49.034
79582 95.466 -78.864 48.994
79583 95.480 -78.861 48.953
79584 95.494 -78.859 48.912
79585 95.502 -78.852 48.870
79586 95.507 -78.841 48.828
79587 95.510 -78.830 48.786
79588 95.525 -78.825 48.743
79589 95.548 -78.827 48.700
79590 95.562 -78.834 48.658
79591 95.568 -78.844 48.616
79592 95.574 -78.855 48.573
79593 95.580 -78.865 48.530
79594 95.586 -78.876 48.486
79595 95.599 -78.889 48.443
79596 95.613 -78.902 48.399
79597 95.627 -78.915 48.356
79598 95.641 -78.929 48.316
79599 95.655 -78.942 48.276
79600 95.669 -78.955 48.235
79601 95.683 -78.968 48.194
79602 95.697 -78.980 48.154
79603 95.714 -78.984 48.113
79604 95.731 -78.987 48.072
79605 95.748 -78.991 48.030
79606 95.765 -78.995 47.987
79607 95.781 -78.999 47.943
79608 95.799 -79.002 47.901
79609 95.815 -79.006 47.858
79610 95.832 -79.010 47.814
79611 95.849 -79.014 47.772
79612 95.866 -79.018 47.730
79613 95.875 -79.014 47.687
79614 95.886 -79.010 47.645
79615 95.916 -79.006 47.605
79616 95.940 -78.999 47.564
79617 95.963 -78.993 47.523
79618 95.986 -78.987 47.482
79619 96.009 -78.981 47.442
79620 96.032 -78.974 47.402
79621 96.055 -78.968 47.362
79622 96.079 -78.962 47.323
79623 96.102 -78.956 47.285
79624 96.129 -78.952 47.248
79625 96.169 -78.953 47.209
79626 96.209 -78.955 47.172
79627 96.249 -78.956 47.136
79628 96.289 -78.958 47.100
79629 96.326 -78.956 47.066
79630 96.359 -78.948 47.033
79631 96.392 -78.940 47.003
79632 96.425 -78.933 46.971
79633 96.427 -78.937 46.939
79634 96.422 -78.943 46.909
79635 96.417 -78.949 46.879
79636 96.429 -78.959 46.849
79637 96.447 -78.968 46.819
79638 96.466 -78.977 46.788
79639 96.484 -78.987 46.757
79640 96.503 -78.996 46.724
79641 96.519 -79.004 46.692
79642 96.533 -79.011 46.660
79643 96.547 -79.018 46.627
79644 96.582 -79.021 46.595
79645 96.630 -79.024 46.564
79646 96.678 -79.025 46.535
79647 96.726 -79.028 46.506
79648 96.773 -79.030 46.477
79649 96.806 -79.049 46.450
79650 96.825 -79.057 46.423
79651 96.845 -78.996 46.397
79652 96.871 -78.946 46.373
79653 96.892 -78.961 46.349
79654 96.912 -78.976 46.325
79655 96.931 -78.986 46.302
79656 96.937 -78.968 46.280
79657 96.933 -78.905 46.260
79658 96.944 -78.872 46.241
79659 96.953 -78.885 46.223
79660 96.961 -78.904 46.204
79661 96.970 -78.923 46.185
79662 96.979 -78.941 46.164
79663 96.984 -78.965 46.142
79664 96.976 -78.951 46.122
79665 96.976 -78.967 46.103
79666 96.976 -78.983 46.084
79667 96.976 -78.999 46.066
79668 96.974 -79.024 46.048
79669 96.999 -79.103 46.031
79670 97.017 -79.159 46.014
79671 97.036 -79.197 45.997
79672 97.055 -79.235 45.982
79673 97.074 -79.274 45.968
79674 97.088 -79.293 45.950
79675 97.096 -79.283 45.932
79676 97.101 -79.260 45.912
79677 97.108 -79.232 45.895
79678 97.117 -79.201 45.880
79679 97.126 -79.170 45.865
79680 97.136 -79.139 45.849
79681 97.145 -79.109 45.833
79682 97.154 -79.078 45.818
79683 97.168 -79.084 45.803
79684 97.185 -79.093 45.786
79685 97.224 -79.115 45.770
79686 97.262 -79.134 45.755
79687 97.293 -79.129 45.741
79688 97.307 -79.137 45.727
79689 97.314 -79.151 45.713
79690 97.320 -79.165 45.698
79691 97.327 -79.179 45.684
79692 97.333 -79.193 45.669
79693 97.343 -79.205 45.656
79694 97.357 -79.215 45.643
79695 97.371 -79.224 45.630
79696 97.385 -79.234 45.617
79697 97.398 -79.244 45.603
79698 97.405 -79.250 45.587
79699 97.412 -79.256 45.572
79700 97.421 -79.261 45.559
79701 97.432 -79.263 45.547
79702 97.443 -79.266 45.534
79703 97.454 -79.269 45.521
79704 97.465 -79.272 45.507
79705 97.476 -79.275 45.494
79706 97.487 -79.277 45.480
79707 97.498 -79.280 45.467
79708 97.509 -79.283 45.454
79709 97.520 -79.285 45.440
79710 97.531 -79.288 45.425
79711 97.547 -79.297 45.409
79712 97.566 -79.304 45.392
79713 97.585 -79.311 45.374
79714 97.603 -79.319 45.355
79715 97.624 -79.327 45.336
79716 97.643 -79.335 45.320
79717 97.663 -79.344 45.307
79718 97.683 -79.353 45.299
79719 97.700 -79.360 45.291
79720 97.713 -79.374 45.283
79721 97.723 -79.390 45.274
79722 97.733 -79.405 45.264
79723 97.744 -79.420 45.254
79724 97.754 -79.435 45.242
79725 97.764 -79.451 45.230
79726 97.775 -79.466 45.218
79727 97.785 -79.482 45.208
79728 97.797 -79.497 45.197
79729 97.810 -79.510 45.181
79730 97.823 -79.524 45.166
79731 97.837 -79.538 45.151
79732 97.851 -79.551 45.137
79733 97.864 -79.566 45.125
79734 97.877 -79.579 45.115
79735 97.889 -79.592 45.107
79736 97.901 -79.604 45.100
79737 97.913 -79.615 45.093
79738 97.933 -79.626 45.087
79739 97.953 -79.637 45.080
79740 97.978 -79.647 45.076
79741 98.006 -79.656 45.072
79742 98.034 -79.665 45.066
79743 98.063 -79.674 45.058
79744 98.091 -79.683 45.051
79745 98.120 -79.692 45.045
79746 98.148 -79.701 45.036
79747 98.175 -79.711 45.027
79748 98.201 -79.717 45.018
79749 98.227 -79.724 45.009
79750 98.253 -79.730 44.997
79751 98.280 -79.737 44.984
79752 98.306 -79.743 44.969
79753 98.333 -79.750 44.953
79754 98.359 -79.756 44.938
79755 98.385 -79.763 44.924
79756 98.412 -79.770 44.913
79757 98.438 -79.778 44.904
79758 98.465 -79.789 44.895
79759 98.496 -79.798 44.887
79760 98.529 -79.808 44.877
79761 98.562 -79.817 44.867
79762 98.594 -79.826 44.858
79763 98.627 -79.836 44.851
79764 98.660 -79.845 44.845
79765 98.678 -79.842 44.839
79766 98.693 -79.835 44.833
79767 98.711 -79.836 44.826
79768 98.732 -79.844 44.819
79769 98.754 -79.852 44.812
79770 98.775 -79.859 44.806
79771 98.796 -79.867 44.799
79772 98.817 -79.874 44.792
79773 98.838 -79.882 44.785
79774 98.860 -79.889 44.779
79775 98.885 -79.890 44.773
79776 98.911 -79.890 44.766
79777 98.936 -79.891 44.759
79778 98.961 -79.892 44.753
79779 98.987 -79.892 44.748
79780 99.002 -79.893 44.741
79781 99.024 -79.908 44.734
79782 99.044 -79.925 44.727
79783 99.063 -79.943 44.721
79784 99.083 -79.959 44.717
79785 99.102 -79.977 44.713
79786 99.122 -79.993 44.708
79787 99.145 -80.008 44.702
79788 99.178 -80.021 44.695
79789 99.209 -80.034 44.689
79790 99.241 -80.046 44.684
79791 99.273 -80.059 44.681
79792 99.308 -80.073 44.678
79793 99.322 -80.077 44.675
79794 99.336 -80.074 44.671
79795 99.381 -80.087 44.668
79796 99.427 -80.099 44.664
79797 99.472 -80.111 44.660
79798 99.502 -80.103 44.655
79799 99.526 -80.090 44.648
79800 99.526 -80.113 44.639
79801 99.526 -80.141 44.628
79802 99.529 -80.163 44.618
79803 99.532 -80.185 44.608
79804 99.536 -80.208 44.599
79805 99.539 -80.230 44.590
79806 99.542 -80.252 44.583
79807 99.545 -80.274 44.575
79808 99.548 -80.296 44.567
79809 99.552 -80.319 44.559
79810 99.555 -80.341 44.552
79811 99.558 -80.363 44.545
79812 99.561 -80.385 44.536
79813 99.564 -80.407 44.527
79814 99.568 -80.429 44.519
79815 99.571 -80.451 44.511
79816 99.574 -80.473 44.503
79817 99.577 -80.495 44.494
79818 99.580 -80.518 44.486
79819 99.584 -80.540 44.476
79820 99.587 -80.562 44.467
79821 99.590 -80.584 44.456
79822 99.634 -80.591 44.445
79823 99.682 -80.597 44.435
79824 99.729 -80.603 44.425
79825 99.777 -80.608 44.415
79826 99.826 -80.613 44.403
79827 99.873 -80.620 44.391
79828 99.921 -80.625 44.380
79829 99.969 -80.631 44.368
79830 100.017 -80.636 44.357
79831 100.065 -80.641 44.345
79832 100.113 -80.648 44.334
80126 93.555 -78.836 53.563
80127 93.565 -78.833 53.554
80128 93.574 -78.831 53.546
80129 93.584 -78.828 53.538
80130 93.593 -78.826 53.531
80131 93.604 -78.824 53.524
80132 93.613 -78.821 53.519
80133 93.623 -78.818 53.515
80134 93.632 -78.814 53.513
80135 93.650 -78.802 53.515
80136 93.654 -78.794 53.518
80137 93.658 -78.791 53.518
80138 93.663 -78.789 53.517
80139 93.669 -78.784 53.514
80140 93.674 -78.780 53.510
80141 93.680 -78.775 53.503
80142 93.686 -78.770 53.494
80143 93.692 -78.765 53.485
80144 93.698 -78.763 53.474
80145 93.712 -78.756 53.461
80146 93.737 -78.742 53.448
80147 93.762 -78.728 53.432
80148 93.786 -78.714 53.415
80149 93.805 -78.702 53.399
80150 93.808 -78.693 53.381
80151 93.812 -78.685 53.362
80152 93.815 -78.677 53.343
80153 93.819 -78.668 53.323
80154 93.822 -78.659 53.302
80155 93.826 -78.651 53.281
80156 93.832 -78.642 53.261
80157 93.846 -78.635 53.240
80158 93.860 -78.626 53.218
80159 93.874 -78.618 53.196
80160 93.887 -78.606 53.175
80161 93.900 -78.593 53.151
80162 93.913 -78.580 53.126
80163 93.925 -78.567 53.099
80164 93.938 -78.554 53.073
80165 93.949 -78.555 53.048
80166 93.960 -78.562 53.023
80167 93.971 -78.568 53.001
80168 93.982 -78.575 52.978
80169 93.992 -78.581 52.955
80170 94.003 -78.588 52.935
80171 94.013 -78.592 52.917
80172 94.024 -78.591 52.899
80173 94.034 -78.589 52.881
80174 94.043 -78.587 52.861
80175 94.053 -78.586 52.839
80176 94.063 -78.584 52.815
80177 94.066 -78.587 52.791
80178 94.058 -78.557 52.766
80179 94.070 -78.514 52.741
80180 94.131 -78.517 52.717
80181 94.192 -78.518 52.692
80182 94.245 -78.515 52.668
80183 94.290 -78.506 52.646
80184 94.334 -78.496 52.626
80185 94.379 -78.487 52.607
80186 94.390 -78.535 52.589
80187 94.397 -78.545 52.571
80188 94.390 -78.557 52.553
80189 94.382 -78.568 52.533
80190 94.374 -78.580 52.511
80191 94.366 -78.592 52.488
80192 94.359 -78.604 52.466
80193 94.351 -78.617 52.445
80194 94.343 -78.629 52.425
80195 94.336 -78.641 52.404
80196 94.328 -78.652 52.383
80197 94.345 -78.640 52.360
80198 94.361 -78.627 52.336
80199 94.378 -78.615 52.313
80200 94.394 -78.603 52.289
80201 94.410 -78.590 52.263
80202 94.426 -78.578 52.233
80203 94.443 -78.566 52.203
80204 94.459 -78.554 52.172
80205 94.473 -78.552 52.138
80206 94.485 -78.560 52.103
80207 94.496 -78.568 52.067
80208 94.507 -78.576 52.029
80209 94.519 -78.583 51.990
80210 94.529 -78.608 51.951
80211 94.538 -78.638 51.913
80212 94.546 -78.652 51.875
80213 94.555 -78.658 51.838
80214 94.567 -78.658 51.801
80215 94.579 -78.657 51.764
80216 94.592 -78.656 51.726
80217 94.604 -78.656 51.687
80218 94.616 -78.655 51.649
80219 94.628 -78.655 51.609
80220 94.641 -78.655 51.572
80221 94.653 -78.654 51.535
80222 94.665 -78.652 51.499
80223 94.677 -78.648 51.461
80224 94.689 -78.645 51.424
80225 94.700 -78.641 51.386
80226 94.712 -78.639 51.351
80227 94.726 -78.642 51.314
80228 94.746 -78.645 51.276
80229 94.764 -78.648 51.237
80230 94.783 -78.652 51.199
80231 94.802 -78.656 51.160
80232 94.821 -78.659 51.121
80233 94.827 -78.667 51.083
80234 94.833 -78.674 51.044
80235 94.838 -78.682 51.004
80236 94.846 -78.676 50.965
80237 94.855 -78.664 50.924
80238 94.865 -78.651 50.881
80239 94.874 -78.638 50.838
80240 94.883 -78.625 50.793
80241 94.893 -78.612 50.747
80242 94.902 -78.599 50.702
80243 94.923 -78.592 50.656
80244 94.961 -78.590 50.608
80245 94.972 -78.589 50.559
80246 94.964 -78.589 50.509
80247 94.956 -78.588 50.460
80248 94.955 -78.592 50.413
80249 94.962 -78.601 50.366
80250 94.969 -78.610 50.323
80251 94.974 -78.621 50.281
80252 94.976 -78.635 50.239
80253 94.977 -78.649 50.197
80254 94.979 -78.663 50.155
80255 94.981 -78.677 50.112
80256 94.990 -78.686 50.069
80257 95.005 -78.690 50.026
80258 95.020 -78.695 49.984
80259 95.035 -78.699 49.941
80260 95.049 -78.704 49.900
80261 95.064 -78.708 49.861
80262 95.079 -78.712 49.823
80263 95.094 -78.717 49.786
80264 95.109 -78.721 49.749
80265 95.123 -78.726 49.712
80266 95.153 -78.722 49.674
80267 95.183 -78.718 49.635
80268 95.212 -78.715 49.595
80269 95.209 -78.715 49.554
80270 95.191 -78.726 49.513
80271 95.219 -78.745 49.472
80272 95.249 -78.766 49.431
80273 95.278 -78.786 49.391
80274 95.307 -78.805 49.351
80275 95.333 -78.822 49.313
80276 95.352 -78.833 49.276
80277 95.372 -78.846 49.239
80278 95.391 -78.857 49.200
80279 95.411 -78.869 49.160
80280 95.425 -78.867 49.120
80281 95.439 -78.865 49.080
80282 95.453 -78.862 49.039
80283 95.466 -78.860 48.999
80284 95.480 -78.858 48.959
80285 95.489 -78.851 48.917
80286 95.493 -78.839 48.874
80287 95.497 -78.828 48.831
80288 95.500 -78.817 48.789
80289 95.515 -78.812 48.745
80290 95.538 -78.814 48.702
80291 95.561 -78.816 48.661
80292 95.570 -78.824 48.617
80293 95.577 -78.835 48.574
80294 95.582 -78.846 48.530
80295 95.588 -78.856 48.488
80296 95.594 -78.867 48.445
80297 95.600 -78.877 48.402
80298 95.611 -78.889 48.360
80299 95.625 -78.902 48.320
80300 95.639 -78.916 48.281
80301 95.653 -78.929 48.241
80302 95.667 -78.943 48.202
80303 95.681 -78.954 48.163
80304 95.698 -78.958 48.124
80305 95.715 -78.962 48.082
80306 95.732 -78.965 48.040
80307 95.748 -78.970 47.997
80308 95.765 -78.974 47.953
80309 95.782 -78.977 47.910
80310 95.799 -78.981 47.867
80311 95.816 -78.985 47.824
80312 95.832 -78.989 47.781
80313 95.852 -78.993 47.739
80314 95.876 -78.999 47.698
80315 95.895 -79.003 47.656
80316 95.918 -78.998 47.615
80317 95.942 -78.992 47.574
80318 95.965 -78.986 47.532
80319 95.988 -78.980 47.491
80320 96.012 -78.974 47.451
80321 96.035 -78.968 47.411
80322 96.058 -78.962 47.370
80323 96.081 -78.955 47.332
80324 96.104 -78.949 47.294
80325 96.128 -78.943 47.255
80326 96.169 -78.945 47.217
80327 96.208 -78.946 47.180
80328 96.249 -78.948 47.143
80329 96.288 -78.949 47.107
80330 96.321 -78.940 47.073
80331 96.354 -78.933 47.041
80332 96.386 -78.925 47.009
80333 96.417 -78.917 46.979
80334 96.430 -78.916 46.948
80335 96.430 -78.922 46.917
80336 96.425 -78.929 46.886
80337 96.420 -78.936 46.856
80338 96.422 -78.943 46.826
80339 96.441 -78.952 46.795
80340 96.460 -78.962 46.762
80341 96.476 -78.970 46.730
80342 96.490 -78.977 46.697
80343 96.504 -78.984 46.664
80344 96.518 -78.990 46.632
80345 96.562 -78.993 46.602
80346 96.610 -78.995 46.572
80347 96.659 -78.996 46.542
80348 96.704 -79.028 46.514
80349 96.739 -79.048 46.485
80350 96.764 -79.046 46.457
80351 96.790 -79.045 46.431
80352 96.814 -79.043 46.405
80353 96.853 -78.967 46.381
80354 96.884 -78.946 46.358
80355 96.900 -78.949 46.335
80356 96.910 -78.937 46.313
80357 96.922 -78.929 46.290
80358 96.933 -78.917 46.270
80359 96.943 -78.857 46.252
80360 96.951 -78.876 46.234
80361 96.960 -78.894 46.216
80362 96.968 -78.913 46.196
80363 96.977 -78.931 46.176
80364 96.979 -78.943 46.153
80365 96.979 -78.959 46.133
80366 96.979 -78.974 46.114
80367 96.979 -78.990 46.096
80368 96.979 -79.006 46.078
80369 96.979 -79.022 46.061
80370 96.989 -79.049 46.043
80371 97.007 -79.122 46.026
80372 97.026 -79.161 46.010
80373 97.045 -79.199 45.995
80374 97.064 -79.237 45.978
80375 97.083 -79.276 45.962
80376 97.100 -79.294 45.945
80377 97.109 -79.263 45.926
80378 97.118 -79.231 45.909
80379 97.127 -79.200 45.895
80380 97.137 -79.169 45.880
80381 97.146 -79.139 45.864
80382 97.155 -79.108 45.850
80383 97.164 -79.079 45.835
80384 97.179 -79.087 45.819
80385 97.194 -79.095 45.803
80386 97.223 -79.112 45.787
80387 97.260 -79.127 45.772
80388 97.285 -79.126 45.757
80389 97.291 -79.140 45.742
80390 97.298 -79.154 45.728
80391 97.304 -79.169 45.714
80392 97.315 -79.179 45.699
80393 97.329 -79.189 45.685
80394 97.343 -79.199 45.671
80395 97.357 -79.209 45.658
80396 97.371 -79.218 45.644
80397 97.385 -79.228 45.630
80398 97.398 -79.237 45.616
80399 97.408 -79.245 45.601
80400 97.415 -79.251 45.586
80401 97.424 -79.255 45.572
80402 97.435 -79.258 45.559
80403 97.446 -79.261 45.546
80404 97.457 -79.263 45.533
80405 97.468 -79.266 45.520
80406 97.479 -79.269 45.506
80407 97.491 -79.272 45.493
80408 97.502 -79.275 45.481
80409 97.513 -79.277 45.468
80410 97.526 -79.280 45.454
80411 97.541 -79.283 45.438
80412 97.560 -79.290 45.421
80413 97.579 -79.299 45.402
80414 97.599 -79.307 45.383
80415 97.619 -79.316 45.362
80416 97.639 -79.324 45.342
80417 97.659 -79.332 45.326
80418 97.678 -79.341 45.314
80419 97.691 -79.346 45.304
80420 97.691 -79.346 45.296
80421 97.706 -79.362 45.287
80422 97.724 -79.382 45.277
80423 97.740 -79.401 45.267
80424 97.750 -79.416 45.255
80425 97.761 -79.432 45.243
80426 97.771 -79.447 45.231
80427 97.782 -79.463 45.220
80428 97.792 -79.478 45.208
80429 97.802 -79.494 45.196
80430 97.816 -79.507 45.183
80431 97.829 -79.521 45.170
80432 97.843 -79.535 45.158
80433 97.856 -79.548 45.144
80434 97.870 -79.563 45.131
80435 97.882 -79.575 45.120
80436 97.893 -79.586 45.112
80437 97.905 -79.598 45.103
80438 97.917 -79.610 45.096
80439 97.934 -79.621 45.090
80440 97.955 -79.632 45.084
80441 97.980 -79.642 45.078
80442 98.008 -79.651 45.074
80443 98.036 -79.660 45.070
80444 98.065 -79.669 45.064
80445 98.093 -79.678 45.057
80446 98.122 -79.686 45.049
80447 98.150 -79.695 45.040
80448 98.186 -79.689 45.030
80449 98.218 -79.704 45.020
80450 98.244 -79.711 45.009
80451 98.270 -79.717 44.997
80452 98.296 -79.724 44.985
80453 98.323 -79.730 44.972
80454 98.349 -79.737 44.957
80455 98.375 -79.743 44.944
80456 98.402 -79.750 44.931
80457 98.428 -79.757 44.921
80458 98.455 -79.767 44.911
80459 98.483 -79.777 44.903
80460 98.511 -79.787 44.895
80461 98.544 -79.797 44.886
80462 98.576 -79.806 44.878
80463 98.609 -79.815 44.870
80464 98.641 -79.824 44.863
80465 98.672 -79.832 44.857
80466 98.687 -79.825 44.852
80467 98.701 -79.818 44.846
80468 98.717 -79.815 44.839
80469 98.738 -79.823 44.832
80470 98.759 -79.830 44.825
80471 98.780 -79.837 44.820
80472 98.801 -79.845 44.813
80473 98.822 -79.852 44.806
80474 98.844 -79.860 44.800
80475 98.865 -79.868 44.794
80476 98.887 -79.876 44.788
80477 98.913 -79.877 44.782
80478 98.938 -79.877 44.775
80479 98.963 -79.878 44.768
80480 98.989 -79.879 44.761
80481 99.005 -79.878 44.753
80482 99.029 -79.892 44.746
80483 99.051 -79.908 44.741
80484 99.070 -79.925 44.736
80485 99.089 -79.942 44.730
80486 99.109 -79.959 44.726
80487 99.128 -79.976 44.721
80488 99.155 -79.990 44.714
80489 99.188 -80.002 44.707
80490 99.219 -80.015 44.703
80491 99.251 -80.027 44.699
80492 99.283 -80.040 44.696
80493 99.314 -80.052 44.692
80494 99.337 -80.057 44.687
80495 99.360 -80.062 44.681
80496 99.372 -80.060 44.675
80497 99.385 -80.059 44.669
80498 99.429 -80.070 44.663
80499 99.474 -80.082 44.657
80500 99.514 -80.087 44.651
80501 99.538 -80.072 44.643
80502 99.540 -80.095 44.634
80503 99.531 -80.133 44.624
80504 99.528 -80.162 44.615
80505 99.532 -80.184 44.605
80506 99.535 -80.207 44.597
80507 99.539 -80.229 44.589
80508 99.542 -80.251 44.581
80509 99.545 -80.273 44.573
80510 99.548 -80.295 44.564
80511 99.551 -80.318 44.557
80512 99.554 -80.340 44.549
80513 99.558 -80.362 44.541
80514 99.561 -80.384 44.533
80515 99.564 -80.406 44.525
80516 99.567 -80.428 44.516
80517 99.570 -80.450 44.508
80518 99.574 -80.472 44.499
80519 99.577 -80.494 44.491
80520 99.580 -80.517 44.481
80521 99.583 -80.539 44.472
80522 99.594 -80.559 44.461
80523 99.642 -80.564 44.450
80524 99.690 -80.569 44.440
80525 99.738 -80.576 44.430
80526 99.786 -80.581 44.419
80527 99.834 -80.586 44.407
80528 99.881 -80.592 44.395
80529 99.929 -80.598 44.383
80530 99.978 -80.604 44.372
80531 100.025 -80.609 44.361
80532 100.073 -80.614 44.350
80533 100.121 -80.620 44.338
80827 93.555 -78.825 53.571
80828 93.565 -78.822 53.563
80829 93.574 -78.820 53.555
80830 93.584 -78.818 53.547
80831 93.594 -78.815 53.540
80832 93.604 -78.812 53.534
80833 93.613 -78.810 53.529
80834 93.623 -78.808 53.525
80835 93.633 -78.805 53.524
80836 93.642 -78.800 53.526
80837 93.651 -78.792 53.529
80838 93.655 -78.789 53.530
80839 93.659 -78.785 53.529
80840 93.665 -78.780 53.527
80841 93.672 -78.776 53.522
80842 93.677 -78.771 53.516
80843 93.683 -78.767 53.508
80844 93.689 -78.762 53.498
80845 93.699 -78.755 53.486
80846 93.724 -78.739 53.474
80847 93.749 -78.725 53.460
80848 93.773 -78.711 53.444
80849 93.798 -78.697 53.427
80850 93.806 -78.687 53.410
80851 93.809 -78.679 53.391
80852 93.813 -78.670 53.371
80853 93.817 -78.662 53.351
80854 93.820 -78.654 53.331
80855 93.823 -78.645 53.311
80856 93.827 -78.636 53.290
80857 93.832 -78.628 53.270
80858 93.845 -78.620 53.249
80859 93.859 -78.611 53.227
80860 93.872 -78.598 53.204
80861 93.885 -78.586 53.182
80862 93.898 -78.573 53.159
80863 93.910 -78.560 53.134
80864 93.923 -78.547 53.108
80865 93.935 -78.534 53.083
80866 93.947 -78.532 53.059
80867 93.958 -78.539 53.033
80868 93.969 -78.545 53.011
80869 93.979 -78.551 52.988
80870 93.990 -78.557 52.966
80871 94.001 -78.564 52.947
80872 94.011 -78.570 52.929
80873 94.023 -78.577 52.911
80874 94.033 -78.581 52.893
80875 94.043 -78.579 52.873
80876 94.053 -78.577 52.851
80877 94.062 -78.576 52.826
80878 94.060 -78.565 52.802
80879 94.050 -78.523 52.777
80880 94.090 -78.509 52.752
80881 94.135 -78.500 52.727
80882 94.180 -78.491 52.703
80883 94.224 -78.481 52.680
80884 94.269 -78.472 52.660
80885 94.313 -78.462 52.641
80886 94.357 -78.454 52.622
80887 94.365 -78.537 52.604
80888 94.380 -78.544 52.587
80889 94.395 -78.551 52.569
80890 94.388 -78.563 52.549
80891 94.381 -78.574 52.527
80892 94.372 -78.586 52.505
80893 94.365 -78.598 52.483
80894 94.357 -78.611 52.461
80895 94.350 -78.623 52.440
80896 94.342 -78.635 52.420
80897 94.334 -78.647 52.399
80898 94.347 -78.638 52.376
80899 94.363 -78.626 52.352
80900 94.380 -78.614 52.329
80901 94.396 -78.601 52.305
80902 94.412 -78.589 52.279
80903 94.428 -78.576 52.251
80904 94.445 -78.564 52.222
80905 94.460 -78.551 52.190
80906 94.471 -78.552 52.156
80907 94.484 -78.558 52.120
80908 94.497 -78.566 52.083
80909 94.508 -78.573 52.044
80910 94.520 -78.581 52.006
80911 94.531 -78.590 51.968
80912 94.540 -78.620 51.930
80913 94.548 -78.632 51.893
80914 94.556 -78.641 51.857
80915 94.568 -78.641 51.820
80916 94.580 -78.640 51.782
80917 94.592 -78.639 51.745
80918 94.605 -78.639 51.706
80919 94.617 -78.639 51.668
80920 94.629 -78.638 51.628
80921 94.641 -78.637 51.588
80922 94.653 -78.637 51.550
80923 94.666 -78.636 51.512
80924 94.678 -78.636 51.474
80925 94.690 -78.635 51.435
80926 94.702 -78.635 51.396
80927 94.714 -78.633 51.359
80928 94.733 -78.636 51.321
80929 94.752 -78.640 51.282
80930 94.771 -78.643 51.243
80931 94.789 -78.647 51.204
80932 94.809 -78.651 51.165
80933 94.827 -78.654 51.127
80934 94.837 -78.661 51.088
80935 94.842 -78.668 51.049
80936 94.848 -78.676 51.009
80937 94.857 -78.666 50.970
80938 94.866 -78.653 50.929
80939 94.875 -78.640 50.886
80940 94.885 -78.627 50.842
80941 94.894 -78.614 50.798
80942 94.903 -78.601 50.752
80943 94.928 -78.595 50.706
80944 94.965 -78.595 50.658
80945 95.002 -78.593 50.608
80946 95.014 -78.592 50.558
80947 95.006 -78.592 50.507
80948 94.998 -78.591 50.459
80949 94.991 -78.591 50.412
80950 94.983 -78.590 50.366
80951 94.982 -78.594 50.323
80952 94.989 -78.603 50.280
80953 94.994 -78.614 50.237
80954 94.996 -78.628 50.196
80955 94.998 -78.642 50.154
80956 95.000 -78.656 50.111
80957 95.014 -78.661 50.068
80958 95.029 -78.665 50.026
80959 95.044 -78.670 49.984
80960 95.058 -78.674 49.941
80961 95.073 -78.679 49.901
80962 95.088 -78.683 49.863
80963 95.103 -78.687 49.826
80964 95.118 -78.692 49.790
80965 95.132 -78.696 49.753
80966 95.147 -78.701 49.716
80967 95.162 -78.705 49.678
80968 95.190 -78.702 49.638
80969 95.220 -78.698 49.598
80970 95.192 -78.702 49.557
80971 95.182 -78.714 49.516
80972 95.211 -78.734 49.476
80973 95.241 -78.754 49.436
80974 95.270 -78.774 49.396
80975 95.299 -78.793 49.357
80976 95.328 -78.813 49.318
80977 95.352 -78.829 49.280
80978 95.372 -78.841 49.242
80979 95.391 -78.852 49.204
80980 95.411 -78.865 49.165
80981 95.425 -78.863 49.126
80982 95.439 -78.861 49.086
80983 95.453 -78.858 49.046
80984 95.467 -78.856 49.005
80985 95.476 -78.849 48.965
80986 95.479 -78.838 48.923
80987 95.483 -78.827 48.880
80988 95.487 -78.815 48.836
80989 95.491 -78.804 48.793
80990 95.504 -78.799 48.749
80991 95.528 -78.800 48.705
80992 95.551 -78.802 48.664
80993 95.573 -78.805 48.620
80994 95.579 -78.815 48.576
80995 95.585 -78.826 48.533
80996 95.591 -78.836 48.492
80997 95.597 -78.847 48.449
80998 95.603 -78.857 48.406
80999 95.609 -78.868 48.365
81000 95.614 -78.878 48.325
81001 95.623 -78.890 48.286
81002 95.637 -78.902 48.247
81003 95.651 -78.916 48.209
81004 95.665 -78.929 48.171
81005 95.682 -78.933 48.133
81006 95.698 -78.937 48.092
81007 95.716 -78.940 48.050
81008 95.732 -78.944 48.008
81009 95.749 -78.948 47.964
81010 95.768 -78.952 47.920
81011 95.792 -78.958 47.877
81012 95.816 -78.963 47.834
81013 95.839 -78.968 47.791
81014 95.864 -78.974 47.749
81015 95.875 -78.988 47.708
81016 95.894 -78.991 47.666
81017 95.918 -78.987 47.625
81018 95.943 -78.983 47.584
81019 95.967 -78.979 47.543
81020 95.990 -78.973 47.501
81021 96.014 -78.967 47.461
81022 96.037 -78.961 47.421
81023 96.060 -78.954 47.381
81024 96.083 -78.948 47.342
81025 96.106 -78.942 47.303
81026 96.129 -78.936 47.263
81027 96.168 -78.937 47.225
81028 96.208 -78.939 47.188
81029 96.249 -78.940 47.151
81030 96.282 -78.933 47.115
81031 96.315 -78.925 47.081
81032 96.347 -78.917 47.048
81033 96.373 -78.908 47.017
81034 96.398 -78.899 46.986
81035 96.414 -78.896 46.956
81036 96.415 -78.902 46.925
81037 96.416 -78.908 46.894
81038 96.418 -78.915 46.864
81039 96.419 -78.921 46.833
81040 96.418 -78.928 46.802
81041 96.433 -78.936 46.769
81042 96.447 -78.943 46.737
81043 96.461 -78.949 46.704
81044 96.475 -78.956 46.670
81045 96.495 -78.963 46.639
81046 96.543 -78.965 46.609
81047 96.589 -78.986 46.579
81048 96.632 -78.986 46.551
81049 96.675 -78.978 46.522
81050 96.717 -78.971 46.494
81051 96.759 -78.962 46.465
81052 96.802 -78.955 46.440
81053 96.839 -78.949 46.415
81054 96.855 -78.940 46.392
81055 96.866 -78.943 46.369
81056 96.889 -78.932 46.347
81057 96.901 -78.926 46.325
81058 96.913 -78.919 46.302
81059 96.925 -78.910 46.282
81060 96.927 -78.857 46.263
81061 96.916 -78.865 46.245
81062 96.924 -78.882 46.227
81063 96.939 -78.902 46.207
81064 96.967 -78.924 46.186
81065 96.986 -78.945 46.165
81066 96.985 -78.962 46.145
81067 96.984 -78.980 46.126
81068 96.984 -78.996 46.108
81069 96.983 -79.014 46.091
81070 96.982 -79.031 46.073
81071 96.982 -79.046 46.056
81072 96.997 -79.086 46.039
81073 97.017 -79.129 46.023
81074 97.038 -79.176 46.007
81075 97.059 -79.222 45.990
81076 97.080 -79.269 45.975
81077 97.100 -79.269 45.959
81078 97.119 -79.262 45.942
81079 97.129 -79.231 45.926
81080 97.138 -79.200 45.911
81081 97.147 -79.169 45.896
81082 97.156 -79.138 45.881
81083 97.165 -79.107 45.867
81084 97.175 -79.082 45.852
81085 97.189 -79.090 45.837
81086 97.205 -79.097 45.820
81087 97.222 -79.107 45.805
81088 97.258 -79.119 45.790
81089 97.268 -79.130 45.773
81090 97.275 -79.143 45.757
81091 97.288 -79.154 45.743
81092 97.302 -79.164 45.730
81093 97.316 -79.173 45.714
81094 97.329 -79.183 45.700
81095 97.344 -79.193 45.687
81096 97.357 -79.202 45.673
81097 97.371 -79.212 45.659
81098 97.385 -79.222 45.644
81099 97.398 -79.231 45.629
81100 97.411 -79.240 45.614
81101 97.418 -79.247 45.600
81102 97.428 -79.250 45.587
81103 97.439 -79.253 45.573
81104 97.450 -79.255 45.558
81105 97.461 -79.258 45.545
81106 97.472 -79.261 45.533
81107 97.483 -79.263 45.519
81108 97.496 -79.266 45.507
81109 97.511 -79.269 45.495
81110 97.525 -79.272 45.479
81111 97.540 -79.275 45.462
81112 97.555 -79.278 45.444
81113 97.575 -79.287 45.424
81114 97.595 -79.295 45.404
81115 97.615 -79.304 45.384
81116 97.635 -79.313 45.366
81117 97.654 -79.321 45.346
81118 97.674 -79.329 45.332
81119 97.682 -79.332 45.319
81120 97.682 -79.332 45.307
81121 97.682 -79.332 45.296
81122 97.698 -79.349 45.286
81123 97.717 -79.369 45.277
81124 97.735 -79.390 45.268
81125 97.753 -79.410 45.257
81126 97.768 -79.429 45.246
81127 97.778 -79.444 45.236
81128 97.789 -79.459 45.224
81129 97.799 -79.475 45.212
81130 97.809 -79.490 45.199
81131 97.822 -79.504 45.187
81132 97.836 -79.518 45.175
81133 97.849 -79.532 45.162
81134 97.862 -79.545 45.149
81135 97.874 -79.558 45.137
81136 97.886 -79.570 45.127
81137 97.897 -79.582 45.118
81138 97.909 -79.593 45.109
81139 97.921 -79.605 45.102
81140 97.936 -79.617 45.095
81141 97.956 -79.627 45.088
81142 97.982 -79.637 45.081
81143 98.010 -79.646 45.075
81144 98.038 -79.655 45.071
81145 98.067 -79.664 45.065
81146 98.095 -79.673 45.058
81147 98.124 -79.682 45.050
81148 98.160 -79.674 45.041
81149 98.197 -79.665 45.031
81150 98.231 -79.686 45.021
81151 98.261 -79.698 45.010
81152 98.287 -79.704 44.999
81153 98.313 -79.711 44.987
81154 98.340 -79.717 44.975
81155 98.366 -79.724 44.962
81156 98.392 -79.730 44.951
81157 98.418 -79.737 44.940
81158 98.445 -79.745 44.929
81159 98.472 -79.755 44.920
81160 98.499 -79.766 44.912
81161 98.526 -79.776 44.904
81162 98.558 -79.786 44.896
81163 98.590 -79.795 44.889
81164 98.623 -79.805 44.883
81165 98.656 -79.814 44.876
81166 98.680 -79.816 44.871
81167 98.695 -79.809 44.865
81168 98.709 -79.802 44.860
81169 98.723 -79.795 44.853
81170 98.743 -79.801 44.846
81171 98.765 -79.808 44.840
81172 98.786 -79.816 44.834
81173 98.807 -79.824 44.828
81174 98.828 -79.831 44.822
81175 98.849 -79.839 44.815
81176 98.871 -79.847 44.810
81177 98.893 -79.855 44.804
81178 98.915 -79.863 44.797
81179 98.940 -79.864 44.789
81180 98.965 -79.864 44.782
81181 98.991 -79.865 44.775
81182 99.009 -79.863 44.768
81183 99.033 -79.877 44.761
81184 99.057 -79.891 44.755
81185 99.077 -79.908 44.749
81186 99.096 -79.925 44.745
81187 99.116 -79.942 44.740
81188 99.135 -79.958 44.734
81189 99.166 -79.971 44.727
81190 99.198 -79.984 44.720
81191 99.229 -79.996 44.714
81192 99.261 -80.009 44.709
81193 99.290 -80.019 44.705
81194 99.313 -80.024 44.699
81195 99.337 -80.028 44.693
81196 99.360 -80.034 44.687
81197 99.383 -80.038 44.679
81198 99.406 -80.043 44.673
81199 99.426 -80.046 44.667
81200 99.439 -80.044 44.660
81201 99.476 -80.053 44.654
81202 99.521 -80.065 44.647
81203 99.550 -80.055 44.638
81204 99.552 -80.077 44.630
81205 99.544 -80.115 44.621
81206 99.536 -80.152 44.611
81207 99.532 -80.184 44.602
81208 99.535 -80.206 44.594
81209 99.538 -80.228 44.586
81210 99.541 -80.250 44.578
81211 99.545 -80.272 44.570
81212 99.548 -80.295 44.561
81213 99.551 -80.317 44.553
81214 99.554 -80.339 44.546
81215 99.557 -80.361 44.538
81216 99.561 -80.383 44.530
81217 99.564 -80.406 44.522
81218 99.567 -80.428 44.513
81219 99.570 -80.450 44.505
81220 99.573 -80.472 44.496
81221 99.577 -80.494 44.486
81222 99.580 -80.516 44.476
81223 99.603 -80.532 44.466
81224 99.650 -80.537 44.456
81225 99.698 -80.542 44.445
81226 99.746 -80.548 44.435
81227 99.794 -80.554 44.424
81228 99.842 -80.560 44.412
81229 99.890 -80.565 44.400
81230 99.938 -80.570 44.388
81231 99.986 -80.576 44.376
81232 100.034 -80.582 44.365
81233 100.081 -80.588 44.353
81234 100.130 -80.593 44.341
81528 93.556 -78.814 53.581
81529 93.565 -78.811 53.571
81530 93.575 -78.809 53.563
81531 93.584 -78.807 53.556
81532 93.594 -78.804 53.549
81533 93.604 -78.802 53.543
81534 93.614 -78.799 53.539
81535 93.623 -78.797 53.535
81536 93.633 -78.794 53.534
81537 93.642 -78.792 53.535
81538 93.647 -78.789 53.539
81539 93.650 -78.784 53.541
81540 93.654 -78.780 53.540
81541 93.663 -78.778 53.538
81542 93.669 -78.774 53.534
81543 93.675 -78.769 53.528
81544 93.680 -78.764 53.519
81545 93.688 -78.759 53.509
81546 93.712 -78.742 53.498
81547 93.736 -78.727 53.485
81548 93.760 -78.710 53.470
81549 93.784 -78.694 53.454
81550 93.804 -78.682 53.437
81551 93.807 -78.673 53.419
81552 93.811 -78.664 53.399
81553 93.814 -78.656 53.379
81554 93.818 -78.648 53.359
81555 93.821 -78.639 53.339
81556 93.825 -78.630 53.319
81557 93.828 -78.622 53.298
81558 93.831 -78.614 53.277
81559 93.844 -78.604 53.255
81560 93.857 -78.591 53.232
81561 93.869 -78.578 53.210
81562 93.882 -78.565 53.187
81563 93.895 -78.552 53.164
81564 93.908 -78.539 53.140
81565 93.920 -78.526 53.115
81566 93.933 -78.514 53.090
81567 93.945 -78.509 53.066
81568 93.956 -78.515 53.041
81569 93.967 -78.522 53.018
81570 93.977 -78.528 52.996
81571 93.988 -78.535 52.975
81572 93.999 -78.541 52.957
81573 94.009 -78.548 52.939
81574 94.020 -78.554 52.921
81575 94.031 -78.561 52.903
81576 94.042 -78.567 52.885
81577 94.052 -78.570 52.863
81578 94.060 -78.569 52.838
81579 94.055 -78.535 52.813
81580 94.060 -78.511 52.788
81581 94.074 -78.480 52.762
81582 94.114 -78.466 52.736
81583 94.158 -78.456 52.714
81584 94.203 -78.447 52.693
81585 94.248 -78.438 52.674
81586 94.292 -78.428 52.656
81587 94.336 -78.421 52.637
81588 94.343 -78.511 52.619
81589 94.355 -78.545 52.602
81590 94.370 -78.552 52.584
81591 94.385 -78.559 52.565
81592 94.387 -78.569 52.544
81593 94.379 -78.580 52.521
81594 94.371 -78.592 52.499
81595 94.363 -78.604 52.478
81596 94.356 -78.617 52.456
81597 94.348 -78.629 52.436
81598 94.340 -78.641 52.414
81599 94.349 -78.636 52.390
81600 94.365 -78.624 52.367
81601 94.382 -78.611 52.343
81602 94.398 -78.599 52.319
81603 94.414 -78.587 52.294
81604 94.430 -78.574 52.268
81605 94.445 -78.561 52.239
81606 94.453 -78.547 52.208
81607 94.466 -78.553 52.173
81608 94.479 -78.560 52.137
81609 94.493 -78.566 52.100
81610 94.506 -78.573 52.062
81611 94.520 -78.579 52.024
81612 94.532 -78.586 51.987
81613 94.542 -78.602 51.951
81614 94.550 -78.611 51.915
81615 94.558 -78.620 51.878
81616 94.568 -78.623 51.841
81617 94.580 -78.623 51.804
81618 94.593 -78.623 51.767
81619 94.605 -78.622 51.729
81620 94.617 -78.621 51.690
81621 94.629 -78.621 51.648
81622 94.642 -78.620 51.606
81623 94.654 -78.620 51.566
81624 94.666 -78.620 51.527
81625 94.679 -78.620 51.488
81626 94.692 -78.623 51.448
81627 94.705 -78.624 51.409
81628 94.718 -78.626 51.371
81629 94.740 -78.631 51.332
81630 94.758 -78.634 51.293
81631 94.777 -78.638 51.253
81632 94.796 -78.642 51.214
81633 94.815 -78.645 51.174
81634 94.834 -78.648 51.133
81635 94.846 -78.654 51.093
81636 94.852 -78.661 51.053
81637 94.858 -78.669 51.013
81638 94.867 -78.656 50.974
81639 94.876 -78.643 50.933
81640 94.886 -78.630 50.891
81641 94.895 -78.617 50.847
81642 94.904 -78.604 50.802
81643 94.932 -78.600 50.756
81644 94.970 -78.598 50.708
81645 95.007 -78.598 50.659
81646 95.044 -78.596 50.608
81647 95.056 -78.595 50.556
81648 95.048 -78.595 50.507
81649 95.041 -78.595 50.459
81650 95.033 -78.594 50.412
81651 95.025 -78.593 50.366
81652 95.018 -78.592 50.323
81653 95.010 -78.592 50.280
81654 95.010 -78.596 50.237
81655 95.016 -78.607 50.194
81656 95.017 -78.620 50.153
81657 95.023 -78.632 50.111
81658 95.038 -78.636 50.069
81659 95.053 -78.640 50.027
81660 95.068 -78.645 49.985
81661 95.083 -78.649 49.944
81662 95.097 -78.654 49.903
81663 95.112 -78.658 49.865
81664 95.127 -78.663 49.829
81665 95.142 -78.667 49.793
81666 95.157 -78.671 49.757
81667 95.171 -78.676 49.719
81668 95.186 -78.680 49.680
81669 95.201 -78.685 49.640
81670 95.219 -78.683 49.600
81671 95.176 -78.689 49.560
81672 95.174 -78.703 49.521
81673 95.203 -78.723 49.480
81674 95.233 -78.742 49.440
81675 95.262 -78.762 49.401
81676 95.290 -78.782 49.363
81677 95.320 -78.802 49.324
81678 95.349 -78.821 49.284
81679 95.372 -78.836 49.245
81680 95.392 -78.848 49.207
81681 95.411 -78.860 49.169
81682 95.425 -78.858 49.130
81683 95.440 -78.856 49.091
81684 95.453 -78.854 49.051
81685 95.462 -78.847 49.011
81686 95.466 -78.836 48.971
81687 95.470 -78.824 48.929
81688 95.474 -78.813 48.886
81689 95.478 -78.802 48.841
81690 95.482 -78.790 48.798
81691 95.494 -78.785 48.754
81692 95.517 -78.787 48.710
81693 95.541 -78.789 48.667
81694 95.564 -78.791 48.624
81695 95.581 -78.796 48.581
81696 95.587 -78.806 48.538
81697 95.593 -78.817 48.496
81698 95.599 -78.827 48.454
81699 95.605 -78.837 48.411
81700 95.611 -78.848 48.371
81701 95.617 -78.858 48.331
81702 95.623 -78.869 48.292
81703 95.629 -78.879 48.254
81704 95.635 -78.890 48.217
81705 95.649 -78.903 48.180
81706 95.665 -78.907 48.143
81707 95.683 -78.912 48.102
81708 95.707 -78.917 48.061
81709 95.731 -78.922 48.019
81710 95.755 -78.927 47.976
81711 95.779 -78.933 47.932
81712 95.803 -78.938 47.888
81713 95.827 -78.943 47.845
81714 95.850 -78.949 47.802
81715 95.862 -78.963 47.760
81716 95.874 -78.978 47.718
81717 95.894 -78.980 47.676
81718 95.918 -78.975 47.636
81719 95.943 -78.971 47.595
81720 95.967 -78.967 47.554
81721 95.991 -78.963 47.513
81722 96.016 -78.959 47.473
81723 96.039 -78.953 47.433
81724 96.062 -78.947 47.392
81725 96.085 -78.941 47.352
81726 96.108 -78.935 47.312
81727 96.131 -78.929 47.273
81728 96.168 -78.929 47.234
81729 96.208 -78.930 47.197
81730 96.243 -78.925 47.160
81731 96.276 -78.918 47.124
81732 96.304 -78.908 47.090
81733 96.329 -78.899 47.057
81734 96.354 -78.890 47.025
81735 96.379 -78.881 46.994
81736 96.398 -78.877 46.963
81737 96.399 -78.883 46.932
81738 96.401 -78.889 46.902
81739 96.402 -78.896 46.871
81740 96.403 -78.902 46.841
81741 96.406 -78.910 46.810
81742 96.417 -78.921 46.777
81743 96.427 -78.924 46.744
81744 96.437 -78.927 46.712
81745 96.446 -78.930 46.678
81746 96.474 -78.935 46.647
81747 96.518 -78.930 46.617
81748 96.563 -78.930 46.587
81749 96.608 -78.930 46.560
81750 96.654 -78.930 46.531
81751 96.699 -78.930 46.503
81752 96.744 -78.930 46.475
81753 96.790 -78.930 46.450
81754 96.832 -78.927 46.427
81755 96.851 -78.923 46.404
81756 96.861 -78.922 46.382
81757 96.871 -78.926 46.360
81758 96.892 -78.916 46.338
81759 96.904 -78.909 46.316
81760 96.916 -78.902 46.295
81761 96.915 -78.864 46.275
81762 96.905 -78.871 46.256
81763 96.907 -78.878 46.237
81764 96.938 -78.899 46.217
81765 96.966 -78.921 46.196
81766 96.995 -78.943 46.176
81767 96.996 -78.961 46.157
81768 96.995 -78.978 46.138
81769 96.994 -78.995 46.121
81770 96.993 -79.012 46.103
81771 96.990 -79.028 46.086
81772 96.983 -79.043 46.069
81773 96.994 -79.077 46.051
81774 97.015 -79.123 46.035
81775 97.036 -79.170 46.019
81776 97.057 -79.216 46.003
81777 97.078 -79.245 45.989
81778 97.098 -79.240 45.974
81779 97.118 -79.234 45.959
81780 97.139 -79.229 45.944
81781 97.148 -79.200 45.928
81782 97.157 -79.169 45.913
81783 97.166 -79.137 45.900
81784 97.175 -79.107 45.885
81785 97.185 -79.085 45.869
81786 97.201 -79.093 45.854
81787 97.215 -79.100 45.837
81788 97.230 -79.109 45.823
81789 97.247 -79.119 45.808
81790 97.260 -79.128 45.791
81791 97.275 -79.138 45.774
81792 97.288 -79.148 45.759
81793 97.302 -79.157 45.745
81794 97.316 -79.167 45.728
81795 97.329 -79.177 45.714
81796 97.344 -79.187 45.701
81797 97.357 -79.196 45.687
81798 97.371 -79.206 45.673
81799 97.385 -79.216 45.658
81800 97.399 -79.225 45.642
81801 97.413 -79.234 45.627
81802 97.421 -79.241 45.613
81803 97.432 -79.244 45.600
81804 97.443 -79.247 45.585
81805 97.454 -79.250 45.572
81806 97.465 -79.253 45.559
81807 97.480 -79.256 45.546
81808 97.495 -79.258 45.532
81809 97.510 -79.261 45.519
81810 97.524 -79.264 45.505
81811 97.539 -79.267 45.488
81812 97.554 -79.270 45.470
81813 97.571 -79.275 45.450
81814 97.591 -79.284 45.428
81815 97.610 -79.292 45.407
81816 97.630 -79.300 45.388
81817 97.650 -79.309 45.370
81818 97.670 -79.318 45.353
81819 97.674 -79.319 45.339
81820 97.674 -79.319 45.325
81821 97.674 -79.318 45.312
81822 97.674 -79.318 45.301
81823 97.690 -79.336 45.290
81824 97.709 -79.357 45.280
81825 97.727 -79.377 45.270
81826 97.745 -79.397 45.261
81827 97.764 -79.418 45.251
81828 97.782 -79.438 45.241
81829 97.796 -79.456 45.229
81830 97.806 -79.471 45.216
81831 97.816 -79.486 45.205
81832 97.828 -79.501 45.192
81833 97.842 -79.515 45.180
81834 97.855 -79.529 45.168
81835 97.867 -79.541 45.155
81836 97.878 -79.553 45.144
81837 97.890 -79.564 45.134
81838 97.901 -79.576 45.125
81839 97.913 -79.588 45.116
81840 97.925 -79.600 45.109
81841 97.937 -79.611 45.101
81842 97.958 -79.623 45.093
81843 97.984 -79.632 45.085
81844 98.012 -79.641 45.077
81845 98.040 -79.650 45.071
81846 98.069 -79.659 45.065
81847 98.097 -79.667 45.057
81848 98.134 -79.659 45.049
81849 98.171 -79.650 45.041
81850 98.208 -79.641 45.032
81851 98.241 -79.661 45.022
81852 98.277 -79.685 45.012
81853 98.303 -79.692 45.002
81854 98.329 -79.698 44.991
81855 98.356 -79.704 44.981
81856 98.382 -79.711 44.969
81857 98.409 -79.717 44.960
81858 98.435 -79.724 44.950
81859 98.462 -79.734 44.940
81860 98.489 -79.745 44.930
81861 98.516 -79.755 44.922
81862 98.544 -79.765 44.914
81863 98.572 -79.775 44.908
81864 98.604 -79.784 44.902
81865 98.637 -79.794 44.896
81866 98.670 -79.803 44.889
81867 98.688 -79.799 44.884
81868 98.703 -79.792 44.879
81869 98.717 -79.786 44.873
81870 98.731 -79.779 44.867
81871 98.749 -79.780 44.861
81872 98.770 -79.787 44.854
81873 98.791 -79.794 44.849
81874 98.812 -79.802 44.843
81875 98.833 -79.809 44.837
81876 98.855 -79.817 44.830
81877 98.877 -79.825 44.824
81878 98.899 -79.833 44.818
81879 98.920 -79.842 44.812
81880 98.942 -79.849 44.804
81881 98.968 -79.851 44.797
81882 98.993 -79.852 44.790
81883 99.015 -79.851 44.783
81884 99.037 -79.861 44.776
81885 99.061 -79.875 44.769
81886 99.083 -79.890 44.763
81887 99.102 -79.908 44.758
81888 99.122 -79.924 44.752
81889 99.144 -79.940 44.746
81890 99.176 -79.953 44.739
81891 99.208 -79.965 44.731
81892 99.239 -79.978 44.724
81893 99.266 -79.986 44.718
81894 99.289 -79.990 44.711
81895 99.312 -79.996 44.704
81896 99.336 -80.000 44.697
81897 99.359 -80.005 44.689
81898 99.382 -80.009 44.683
81899 99.406 -80.015 44.676
81900 99.429 -80.019 44.670
81901 99.452 -80.024 44.664
81902 99.475 -80.028 44.658
81903 99.493 -80.030 44.650
81904 99.523 -80.036 44.642
81905 99.562 -80.039 44.634
81906 99.565 -80.059 44.625
81907 99.556 -80.096 44.616
81908 99.548 -80.134 44.608
81909 99.540 -80.171 44.598
81910 99.535 -80.206 44.590
81911 99.538 -80.228 44.583
81912 99.541 -80.250 44.574
81913 99.544 -80.272 44.566
81914 99.548 -80.294 44.558
81915 99.551 -80.316 44.551
81916 99.554 -80.338 44.543
81917 99.557 -80.360 44.535
81918 99.560 -80.383 44.527
81919 99.564 -80.405 44.518
81920 99.567 -80.427 44.510
81921 99.570 -80.449 44.501
81922 99.573 -80.471 44.491
81923 99.576 -80.494 44.481
81924 99.611 -80.504 44.471
81925 99.658 -80.510 44.462
81926 99.706 -80.516 44.451
81927 99.755 -80.521 44.439
81928 99.802 -80.526 44.428
81929 99.850 -80.532 44.416
81930 99.898 -80.538 44.404
81931 99.946 -80.543 44.393
81932 99.994 -80.549 44.380
81933 100.042 -80.554 44.369
81934 100.092 -80.565 44.357
81935 100.143 -80.578 44.344
82229 93.556 -78.803 53.589
82230 93.565 -78.801 53.581
82231 93.575 -78.798 53.572
82232 93.585 -78.796 53.565
82233 93.594 -78.793 53.558
82234 93.604 -78.791 53.552
82235 93.614 -78.788 53.547
82236 93.624 -78.786 53.544
82237 93.633 -78.783 53.542
82238 93.641 -78.780 53.541
82239 93.641 -78.783 53.545
82240 93.644 -78.778 53.547
82241 93.648 -78.774 53.547
82242 93.658 -78.774 53.548
82243 93.666 -78.772 53.544
82244 93.672 -78.767 53.537
82245 93.678 -78.763 53.529
82246 93.700 -78.747 53.520
82247 93.724 -78.731 53.508
82248 93.749 -78.714 53.494
82249 93.770 -78.698 53.478
82250 93.789 -78.681 53.462
82251 93.801 -78.667 53.444
82252 93.804 -78.659 53.425
82253 93.808 -78.651 53.406
82254 93.812 -78.642 53.387
82255 93.815 -78.634 53.367
82256 93.818 -78.625 53.347
82257 93.822 -78.617 53.326
82258 93.825 -78.608 53.304
82259 93.828 -78.596 53.282
82260 93.840 -78.589 53.259
82261 93.852 -78.577 53.236
82262 93.865 -78.564 53.214
82263 93.878 -78.551 53.191
82264 93.891 -78.539 53.167
82265 93.903 -78.526 53.143
82266 93.916 -78.514 53.119
82267 93.928 -78.501 53.094
82268 93.950 -78.498 53.070
82269 93.960 -78.504 53.046
82270 93.970 -78.510 53.025
82271 93.980 -78.516 53.004
82272 93.991 -78.521 52.983
82273 94.001 -78.527 52.965
82274 94.011 -78.532 52.946
82275 94.022 -78.539 52.928
82276 94.032 -78.545 52.911
82277 94.043 -78.550 52.892
82278 94.053 -78.556 52.871
82279 94.066 -78.556 52.847
82280 94.072 -78.532 52.822
82281 94.077 -78.509 52.797
82282 94.087 -78.481 52.771
82283 94.107 -78.445 52.745
82284 94.137 -78.422 52.724
82285 94.182 -78.413 52.704
82286 94.226 -78.403 52.685
82287 94.271 -78.394 52.668
82288 94.314 -78.388 52.650
82289 94.321 -78.479 52.632
82290 94.330 -78.547 52.616
82291 94.345 -78.554 52.598
82292 94.360 -78.560 52.578
82293 94.375 -78.567 52.557
82294 94.385 -78.575 52.536
82295 94.377 -78.586 52.515
82296 94.369 -78.598 52.494
82297 94.362 -78.611 52.473
82298 94.354 -78.623 52.451
82299 94.346 -78.635 52.428
82300 94.351 -78.634 52.404
82301 94.367 -78.622 52.381
82302 94.384 -78.610 52.357
82303 94.400 -78.598 52.333
82304 94.416 -78.585 52.309
82305 94.431 -78.572 52.284
82306 94.439 -78.557 52.255
82307 94.448 -78.548 52.224
82308 94.461 -78.554 52.189
82309 94.474 -78.561 52.154
82310 94.488 -78.567 52.118
82311 94.501 -78.574 52.080
82312 94.514 -78.580 52.043
82313 94.530 -78.585 52.006
82314 94.546 -78.589 51.973
82315 94.552 -78.592 51.937
82316 94.560 -78.600 51.901
82317 94.569 -78.607 51.864
82318 94.581 -78.606 51.826
82319 94.593 -78.606 51.789
82320 94.606 -78.605 51.751
82321 94.618 -78.604 51.712
82322 94.630 -78.606 51.668
82323 94.643 -78.608 51.624
82324 94.657 -78.611 51.581
82325 94.670 -78.612 51.541
82326 94.683 -78.614 51.500
82327 94.695 -78.617 51.460
82328 94.708 -78.618 51.419
82329 94.722 -78.620 51.380
82330 94.743 -78.625 51.341
82331 94.764 -78.629 51.301
82332 94.784 -78.633 51.260
82333 94.803 -78.636 51.219
82334 94.821 -78.640 51.179
82335 94.840 -78.643 51.137
82336 94.856 -78.648 51.095
82337 94.862 -78.655 51.054
82338 94.868 -78.659 51.013
82339 94.878 -78.646 50.973
82340 94.887 -78.633 50.931
82341 94.896 -78.620 50.889
82342 94.905 -78.608 50.846
82343 94.937 -78.604 50.800
82344 94.974 -78.603 50.752
82345 95.012 -78.601 50.704
82346 95.049 -78.601 50.654
82347 95.086 -78.600 50.604
82348 95.098 -78.598 50.555
82349 95.091 -78.598 50.506
82350 95.083 -78.598 50.459
82351 95.075 -78.597 50.413
82352 95.067 -78.596 50.367
82353 95.060 -78.595 50.323
82354 95.052 -78.595 50.279
82355 95.044 -78.594 50.236
82356 95.037 -78.593 50.195
82357 95.036 -78.600 50.154
82358 95.047 -78.607 50.112
82359 95.062 -78.611 50.069
82360 95.077 -78.616 50.028
82361 95.092 -78.620 49.986
82362 95.107 -78.624 49.945
82363 95.121 -78.629 49.905
82364 95.131 -78.632 49.867
82365 95.140 -78.636 49.830
82366 95.149 -78.639 49.795
82367 95.159 -78.642 49.758
82368 95.169 -78.645 49.720
82369 95.178 -78.648 49.680
82370 95.188 -78.652 49.642
82371 95.196 -78.657 49.603
82372 95.177 -78.664 49.564
82373 95.192 -78.686 49.524
82374 95.208 -78.708 49.483
82375 95.225 -78.731 49.443
82376 95.253 -78.751 49.404
82377 95.282 -78.771 49.367
82378 95.312 -78.790 49.329
82379 95.341 -78.810 49.291
82380 95.370 -78.830 49.252
82381 95.392 -78.843 49.213
82382 95.411 -78.855 49.173
82383 95.426 -78.855 49.134
82384 95.440 -78.852 49.095
82385 95.449 -78.846 49.056
82386 95.453 -78.834 49.016
82387 95.457 -78.823 48.975
82388 95.460 -78.811 48.934
82389 95.464 -78.800 48.892
82390 95.468 -78.789 48.848
82391 95.472 -78.777 48.804
82392 95.483 -78.771 48.760
82393 95.507 -78.774 48.716
82394 95.530 -78.775 48.672
82395 95.554 -78.777 48.629
82396 95.577 -78.779 48.587
82397 95.590 -78.786 48.544
82398 95.596 -78.797 48.502
82399 95.602 -78.808 48.460
82400 95.607 -78.818 48.419
82401 95.613 -78.828 48.378
82402 95.619 -78.839 48.339
82403 95.625 -78.849 48.300
82404 95.631 -78.860 48.263
82405 95.637 -78.870 48.226
82406 95.645 -78.878 48.191
82407 95.670 -78.886 48.154
82408 95.694 -78.892 48.114
82409 95.718 -78.897 48.073
82410 95.742 -78.902 48.031
82411 95.766 -78.908 47.988
82412 95.790 -78.913 47.945
82413 95.814 -78.918 47.901
82414 95.837 -78.924 47.858
82415 95.849 -78.938 47.815
82416 95.861 -78.953 47.772
82417 95.872 -78.968 47.730
82418 95.893 -78.968 47.687
82419 95.918 -78.964 47.648
82420 95.942 -78.959 47.608
82421 95.966 -78.955 47.567
82422 95.990 -78.951 47.527
82423 96.015 -78.947 47.487
82424 96.039 -78.943 47.447
82425 96.063 -78.939 47.405
82426 96.088 -78.934 47.364
82427 96.111 -78.928 47.323
82428 96.134 -78.922 47.283
82429 96.168 -78.921 47.245
82430 96.204 -78.918 47.208
82431 96.234 -78.909 47.171
82432 96.259 -78.900 47.133
82433 96.285 -78.891 47.099
82434 96.310 -78.882 47.066
82435 96.335 -78.873 47.033
82436 96.360 -78.864 47.002
82437 96.382 -78.856 46.971
82438 96.384 -78.863 46.939
82439 96.385 -78.869 46.909
82440 96.386 -78.876 46.880
82441 96.388 -78.882 46.849
82442 96.397 -78.897 46.818
82443 96.408 -78.914 46.785
82444 96.418 -78.916 46.752
82445 96.428 -78.919 46.719
82446 96.437 -78.921 46.686
82447 96.466 -78.923 46.655
82448 96.512 -78.922 46.625
82449 96.557 -78.922 46.596
82450 96.602 -78.922 46.568
82451 96.648 -78.922 46.540
82452 96.693 -78.922 46.512
82453 96.738 -78.921 46.485
82454 96.781 -78.919 46.461
82455 96.822 -78.915 46.438
82456 96.847 -78.911 46.417
82457 96.858 -78.908 46.395
82458 96.868 -78.905 46.373
82459 96.878 -78.906 46.351
82460 96.895 -78.899 46.331
82461 96.907 -78.893 46.309
82462 96.904 -78.870 46.288
82463 96.909 -78.872 46.268
82464 96.918 -78.878 46.247
82465 96.934 -78.900 46.228
82466 96.964 -78.919 46.208
82467 96.995 -78.940 46.188
82468 97.006 -78.959 46.170
82469 97.005 -78.976 46.152
82470 96.999 -78.991 46.135
82471 96.993 -79.006 46.118
82472 96.986 -79.021 46.100
82473 96.980 -79.036 46.082
82474 96.991 -79.071 46.065
82475 97.012 -79.117 46.050
82476 97.033 -79.164 46.035
82477 97.054 -79.210 46.020
82478 97.075 -79.215 46.006
82479 97.095 -79.210 45.990
82480 97.115 -79.205 45.975
82481 97.136 -79.200 45.960
82482 97.156 -79.194 45.944
82483 97.167 -79.169 45.930
82484 97.176 -79.137 45.916
82485 97.185 -79.106 45.901
82486 97.196 -79.088 45.886
82487 97.211 -79.096 45.870
82488 97.226 -79.103 45.855
82489 97.238 -79.109 45.840
82490 97.248 -79.115 45.825
82491 97.261 -79.125 45.808
82492 97.275 -79.134 45.791
82493 97.289 -79.143 45.776
82494 97.303 -79.153 45.761
82495 97.316 -79.162 45.744
82496 97.330 -79.172 45.731
82497 97.344 -79.181 45.717
82498 97.358 -79.190 45.702
82499 97.371 -79.200 45.688
82500 97.385 -79.209 45.673
82501 97.399 -79.219 45.658
82502 97.413 -79.228 45.642
82503 97.424 -79.236 45.627
82504 97.435 -79.238 45.612
82505 97.450 -79.241 45.596
82506 97.465 -79.244 45.582
82507 97.479 -79.247 45.569
82508 97.494 -79.250 45.554
82509 97.509 -79.253 45.539
82510 97.523 -79.256 45.524
82511 97.538 -79.259 45.509
82512 97.553 -79.262 45.492
82513 97.568 -79.265 45.474
82514 97.586 -79.272 45.453
82515 97.606 -79.281 45.431
82516 97.626 -79.289 45.411
82517 97.646 -79.297 45.393
82518 97.665 -79.306 45.376
82519 97.665 -79.306 45.361
82520 97.665 -79.305 45.346
82521 97.665 -79.304 45.332
82522 97.665 -79.304 45.319
82523 97.665 -79.303 45.307
82524 97.682 -79.323 45.296
82525 97.701 -79.344 45.286
82526 97.719 -79.364 45.275
82527 97.737 -79.385 45.265
82528 97.756 -79.405 45.256
82529 97.774 -79.426 45.246
82530 97.792 -79.446 45.234
82531 97.811 -79.466 45.222
82532 97.823 -79.483 45.211
82533 97.834 -79.498 45.198
82534 97.847 -79.512 45.186
82535 97.860 -79.524 45.175
82536 97.871 -79.536 45.163
82537 97.882 -79.548 45.152
82538 97.894 -79.560 45.142
82539 97.906 -79.571 45.133
82540 97.917 -79.582 45.123
82541 97.929 -79.595 45.116
82542 97.940 -79.606 45.108
82543 97.959 -79.617 45.099
82544 97.986 -79.627 45.091
82545 98.014 -79.636 45.083
82546 98.042 -79.645 45.076
82547 98.071 -79.653 45.069
82548 98.108 -79.644 45.061
82549 98.145 -79.635 45.053
82550 98.182 -79.626 45.046
82551 98.218 -79.616 45.037
82552 98.251 -79.636 45.027
82553 98.287 -79.661 45.018
82554 98.320 -79.679 45.009
82555 98.346 -79.685 44.999
82556 98.372 -79.692 44.989
82557 98.398 -79.698 44.978
82558 98.425 -79.704 44.968
82559 98.451 -79.712 44.959
82560 98.479 -79.723 44.949
82561 98.506 -79.733 44.941
82562 98.533 -79.743 44.933
82563 98.560 -79.754 44.925
82564 98.587 -79.764 44.920
82565 98.619 -79.773 44.914
82566 98.652 -79.783 44.908
82567 98.681 -79.789 44.902
82568 98.696 -79.783 44.897
82569 98.710 -79.776 44.892
82570 98.725 -79.770 44.888
82571 98.739 -79.763 44.882
82572 98.755 -79.758 44.875
82573 98.776 -79.765 44.869
82574 98.796 -79.773 44.864
82575 98.817 -79.780 44.858
82576 98.839 -79.788 44.852
82577 98.861 -79.796 44.845
82578 98.882 -79.804 44.839
82579 98.905 -79.812 44.833
82580 98.926 -79.820 44.826
82581 98.948 -79.828 44.819
82582 98.970 -79.836 44.813
82583 98.995 -79.838 44.806
82584 99.020 -79.839 44.799
82585 99.041 -79.846 44.792
82586 99.065 -79.860 44.784
82587 99.089 -79.874 44.776
82588 99.110 -79.890 44.770
82589 99.129 -79.907 44.764
82590 99.154 -79.922 44.758
82591 99.186 -79.934 44.750
82592 99.218 -79.947 44.742
82593 99.242 -79.952 44.733
82594 99.265 -79.958 44.724
82595 99.288 -79.962 44.717
82596 99.311 -79.967 44.710
82597 99.335 -79.971 44.702
82598 99.358 -79.977 44.694
82599 99.381 -79.981 44.687
82600 99.405 -79.986 44.681
82601 99.428 -79.990 44.674
82602 99.451 -79.995 44.668
82603 99.474 -80.000 44.661
82604 99.497 -80.005 44.654
82605 99.521 -80.009 44.647
82606 99.544 -80.014 44.639
82607 99.571 -80.019 44.630
82608 99.577 -80.040 44.621
82609 99.569 -80.078 44.612
82610 99.561 -80.115 44.603
82611 99.553 -80.153 44.595
82612 99.544 -80.191 44.587
82613 99.538 -80.227 44.579
82614 99.541 -80.249 44.571
82615 99.544 -80.272 44.563
82616 99.547 -80.294 44.556
82617 99.550 -80.316 44.548
82618 99.554 -80.337 44.540
82619 99.557 -80.359 44.532
82620 99.560 -80.382 44.523
82621 99.563 -80.404 44.515
82622 99.566 -80.426 44.507
82623 99.570 -80.448 44.497
82624 99.573 -80.471 44.487
82625 99.619 -80.477 44.477
82626 99.667 -80.482 44.467
82627 99.715 -80.488 44.456
82628 99.763 -80.494 44.444
82629 99.810 -80.499 44.432
82630 99.858 -80.505 44.420
82631 99.906 -80.510 44.409
82632 99.955 -80.516 44.397
82633 100.005 -80.527 44.384
82634 100.056 -80.540 44.373
82635 100.107 -80.553 44.361
82636 100.158 -80.566 44.349
82930 93.556 -78.793 53.597
82931 93.566 -78.789 53.589
82932 93.575 -78.787 53.581
82933 93.585 -78.785 53.573
82934 93.594 -78.783 53.566
82935 93.605 -78.780 53.560
82936 93.614 -78.777 53.555
82937 93.624 -78.775 53.551
82938 93.633 -78.773 53.548
82939 93.638 -78.767 53.546
82940 93.637 -78.758 53.549
82941 93.638 -78.773 53.552
82942 93.642 -78.769 53.554
82943 93.652 -78.768 53.554
82944 93.661 -78.767 53.551
82945 93.669 -78.765 53.545
82946 93.689 -78.752 53.536
82947 93.708 -78.735 53.525
82948 93.727 -78.718 53.513
82949 93.747 -78.701 53.499
82950 93.766 -78.684 53.484
82951 93.785 -78.667 53.466
82952 93.796 -78.655 53.448
82953 93.799 -78.645 53.429
82954 93.803 -78.637 53.410
82955 93.806 -78.629 53.391
82956 93.809 -78.620 53.372
82957 93.813 -78.612 53.351
82958 93.817 -78.603 53.331
82959 93.819 -78.589 53.309
82960 93.822 -78.575 53.286
82961 93.835 -78.579 53.262
82962 93.847 -78.566 53.238
82963 93.860 -78.554 53.216
82964 93.873 -78.541 53.193
82965 93.886 -78.529 53.169
82966 93.898 -78.516 53.143
82967 93.911 -78.503 53.120
82968 93.939 -78.488 53.096
82969 93.960 -78.486 53.073
82970 93.964 -78.495 53.051
82971 93.974 -78.501 53.031
82972 93.985 -78.507 53.011
82973 93.995 -78.512 52.990
82974 94.005 -78.518 52.972
82975 94.016 -78.523 52.953
82976 94.027 -78.529 52.935
82977 94.037 -78.535 52.917
82978 94.048 -78.540 52.898
82979 94.062 -78.539 52.876
82980 94.077 -78.537 52.853
82981 94.089 -78.529 52.827
82982 94.094 -78.506 52.802
82983 94.100 -78.482 52.777
82984 94.119 -78.447 52.752
82985 94.139 -78.411 52.731
82986 94.161 -78.379 52.711
82987 94.205 -78.369 52.694
82988 94.250 -78.360 52.677
82989 94.292 -78.356 52.660
82990 94.299 -78.445 52.643
82991 94.307 -78.535 52.626
82992 94.321 -78.555 52.608
82993 94.335 -78.561 52.589
82994 94.350 -78.568 52.568
82995 94.365 -78.575 52.547
82996 94.380 -78.582 52.527
82997 94.376 -78.593 52.507
82998 94.368 -78.604 52.486
82999 94.360 -78.617 52.464
83000 94.352 -78.629 52.442
83001 94.353 -78.633 52.418
83002 94.369 -78.620 52.395
83003 94.386 -78.608 52.371
83004 94.402 -78.595 52.347
83005 94.418 -78.582 52.323
83006 94.425 -78.568 52.298
83007 94.432 -78.553 52.269
83008 94.442 -78.549 52.238
83009 94.456 -78.555 52.204
83010 94.470 -78.561 52.170
83011 94.485 -78.566 52.135
83012 94.501 -78.570 52.097
83013 94.517 -78.574 52.061
83014 94.532 -78.579 52.026
83015 94.548 -78.583 51.992
83016 94.557 -78.586 51.958
83017 94.563 -78.588 51.923
83018 94.570 -78.590 51.886
83019 94.582 -78.592 51.848
83020 94.595 -78.594 51.810
83021 94.608 -78.596 51.771
83022 94.621 -78.598 51.731
83023 94.634 -78.600 51.688
83024 94.647 -78.602 51.643
83025 94.660 -78.604 51.596
83026 94.673 -78.606 51.554
83027 94.686 -78.608 51.512
83028 94.699 -78.610 51.470
83029 94.712 -78.612 51.428
83030 94.725 -78.614 51.387
83031 94.747 -78.619 51.346
83032 94.768 -78.623 51.305
83033 94.789 -78.627 51.262
83034 94.809 -78.631 51.221
83035 94.828 -78.635 51.180
83036 94.846 -78.638 51.138
83037 94.865 -78.642 51.095
83038 94.872 -78.649 51.052
83039 94.879 -78.648 51.009
83040 94.888 -78.636 50.967
83041 94.897 -78.623 50.925
83042 94.906 -78.611 50.882
83043 94.942 -78.608 50.840
83044 94.979 -78.608 50.794
83045 95.016 -78.606 50.746
83046 95.053 -78.605 50.698
83047 95.072 -78.605 50.649
83048 95.075 -78.607 50.601
83049 95.069 -78.606 50.553
83050 95.064 -78.605 50.505
83051 95.058 -78.604 50.459
83052 95.052 -78.604 50.413
83053 95.047 -78.603 50.367
83054 95.041 -78.602 50.324
83055 95.036 -78.601 50.279
83056 95.030 -78.601 50.236
83057 95.030 -78.595 50.194
83058 95.046 -78.580 50.154
83059 95.061 -78.585 50.112
83060 95.072 -78.589 50.069
83061 95.083 -78.592 50.027
83062 95.094 -78.596 49.986
83063 95.104 -78.599 49.945
83064 95.115 -78.603 49.906
83065 95.126 -78.607 49.867
83066 95.136 -78.611 49.830
83067 95.147 -78.614 49.794
83068 95.158 -78.618 49.757
83069 95.168 -78.622 49.719
83070 95.179 -78.625 49.681
83071 95.190 -78.629 49.643
83072 95.214 -78.641 49.606
83073 95.224 -78.652 49.567
83074 95.232 -78.670 49.527
83075 95.246 -78.691 49.486
83076 95.259 -78.712 49.446
83077 95.274 -78.733 49.407
83078 95.290 -78.756 49.370
83079 95.305 -78.778 49.333
83080 95.333 -78.799 49.297
83081 95.361 -78.818 49.259
83082 95.391 -78.838 49.220
83083 95.411 -78.850 49.181
83084 95.426 -78.851 49.140
83085 95.436 -78.844 49.100
83086 95.439 -78.833 49.061
83087 95.443 -78.821 49.020
83088 95.447 -78.810 48.980
83089 95.451 -78.799 48.939
83090 95.455 -78.787 48.897
83091 95.459 -78.776 48.854
83092 95.463 -78.764 48.809
83093 95.473 -78.758 48.765
83094 95.496 -78.760 48.721
83095 95.520 -78.761 48.677
83096 95.543 -78.764 48.635
83097 95.567 -78.766 48.593
83098 95.590 -78.767 48.550
83099 95.598 -78.777 48.508
83100 95.604 -78.788 48.467
83101 95.610 -78.798 48.427
83102 95.616 -78.808 48.387
83103 95.622 -78.819 48.348
83104 95.628 -78.830 48.310
83105 95.633 -78.840 48.273
83106 95.641 -78.849 48.237
83107 95.654 -78.852 48.202
83108 95.678 -78.862 48.166
83109 95.705 -78.872 48.127
83110 95.729 -78.877 48.086
83111 95.753 -78.883 48.045
83112 95.777 -78.888 48.001
83113 95.801 -78.893 47.958
83114 95.825 -78.899 47.915
83115 95.837 -78.913 47.872
83116 95.848 -78.928 47.828
83117 95.859 -78.943 47.786
83118 95.873 -78.953 47.743
83119 95.894 -78.952 47.700
83120 95.918 -78.949 47.660
83121 95.942 -78.945 47.619
83122 95.966 -78.942 47.579
83123 95.990 -78.939 47.540
83124 96.015 -78.935 47.500
83125 96.039 -78.931 47.459
83126 96.063 -78.927 47.416
83127 96.088 -78.923 47.374
83128 96.112 -78.919 47.334
83129 96.136 -78.915 47.294
83130 96.165 -78.910 47.256
83131 96.190 -78.901 47.219
83132 96.215 -78.892 47.181
83133 96.240 -78.883 47.144
83134 96.265 -78.874 47.108
83135 96.291 -78.865 47.074
83136 96.316 -78.855 47.042
83137 96.340 -78.846 47.011
83138 96.366 -78.837 46.980
83139 96.368 -78.843 46.948
83140 96.369 -78.849 46.916
83141 96.371 -78.855 46.886
83142 96.376 -78.868 46.854
83143 96.388 -78.885 46.822
83144 96.399 -78.902 46.791
83145 96.409 -78.908 46.758
83146 96.418 -78.911 46.726
83147 96.429 -78.914 46.694
83148 96.461 -78.915 46.664
83149 96.506 -78.915 46.634
83150 96.551 -78.914 46.606
83151 96.596 -78.914 46.577
83152 96.642 -78.914 46.550
83153 96.687 -78.914 46.523
83154 96.731 -78.912 46.497
83155 96.771 -78.907 46.473
83156 96.812 -78.902 46.451
83157 96.844 -78.898 46.430
83158 96.854 -78.895 46.408
83159 96.864 -78.893 46.386
83160 96.874 -78.890 46.365
83161 96.885 -78.887 46.345
83162 96.898 -78.883 46.322
83163 96.901 -78.873 46.301
83164 96.922 -78.869 46.280
83165 96.929 -78.877 46.259
83166 96.929 -78.902 46.239
83167 96.960 -78.921 46.220
83168 96.990 -78.946 46.201
83169 97.008 -78.954 46.184
83170 97.002 -78.969 46.167
83171 96.996 -78.984 46.150
83172 96.990 -78.999 46.132
83173 96.983 -79.013 46.115
83174 96.977 -79.028 46.097
83175 96.989 -79.065 46.080
83176 97.010 -79.111 46.066
83177 97.031 -79.158 46.052
83178 97.052 -79.190 46.038
83179 97.073 -79.185 46.022
83180 97.093 -79.180 46.006
83181 97.113 -79.175 45.991
83182 97.133 -79.170 45.976
83183 97.153 -79.165 45.962
83184 97.174 -79.159 45.948
83185 97.186 -79.137 45.933
83186 97.195 -79.106 45.918
83187 97.207 -79.091 45.903
83188 97.222 -79.099 45.888
83189 97.232 -79.104 45.873
83190 97.239 -79.107 45.857
83191 97.249 -79.112 45.842
83192 97.262 -79.122 45.825
83193 97.277 -79.131 45.809
83194 97.290 -79.140 45.793
83195 97.304 -79.150 45.777
83196 97.317 -79.159 45.761
83197 97.331 -79.169 45.746
83198 97.345 -79.178 45.733
83199 97.359 -79.187 45.717
83200 97.372 -79.197 45.702
83201 97.386 -79.206 45.687
83202 97.401 -79.215 45.673
83203 97.416 -79.223 45.657
83204 97.429 -79.225 45.640
83205 97.449 -79.234 45.622
83206 97.464 -79.237 45.604
83207 97.478 -79.239 45.589
83208 97.493 -79.242 45.574
83209 97.508 -79.245 45.558
83210 97.523 -79.248 45.541
83211 97.537 -79.251 45.523
83212 97.552 -79.254 45.506
83213 97.567 -79.257 45.489
83214 97.582 -79.260 45.471
83215 97.602 -79.269 45.453
83216 97.621 -79.278 45.435
83217 97.642 -79.286 45.417
83218 97.657 -79.292 45.401
83219 97.657 -79.292 45.385
83220 97.657 -79.291 45.370
83221 97.657 -79.291 45.355
83222 97.656 -79.291 45.341
83223 97.656 -79.290 45.328
83224 97.657 -79.291 45.315
83225 97.675 -79.310 45.303
83226 97.693 -79.331 45.293
83227 97.712 -79.351 45.282
83228 97.730 -79.372 45.273
83229 97.748 -79.392 45.263
83230 97.767 -79.413 45.252
83231 97.785 -79.433 45.241
83232 97.803 -79.454 45.230
83233 97.821 -79.473 45.219
83234 97.840 -79.494 45.207
83235 97.852 -79.507 45.195
83236 97.863 -79.519 45.183
83237 97.875 -79.531 45.172
83238 97.886 -79.542 45.161
83239 97.898 -79.554 45.150
83240 97.910 -79.566 45.141
83241 97.921 -79.578 45.131
83242 97.933 -79.589 45.123
83243 97.944 -79.601 45.115
83244 97.961 -79.612 45.106
83245 97.988 -79.622 45.097
83246 98.016 -79.631 45.089
83247 98.045 -79.638 45.082
83248 98.082 -79.629 45.075
83249 98.119 -79.620 45.067
83250 98.156 -79.611 45.060
83251 98.194 -79.602 45.053
83252 98.229 -79.592 45.044
83253 98.262 -79.610 45.035
83254 98.298 -79.636 45.027
83255 98.334 -79.660 45.017
83256 98.363 -79.672 45.008
83257 98.389 -79.679 44.998
83258 98.415 -79.685 44.988
83259 98.441 -79.692 44.978
83260 98.468 -79.701 44.968
83261 98.496 -79.711 44.960
83262 98.523 -79.722 44.952
83263 98.550 -79.732 44.944
83264 98.577 -79.742 44.937
83265 98.604 -79.752 44.931
83266 98.633 -79.763 44.925
83267 98.666 -79.772 44.920
83268 98.690 -79.773 44.915
83269 98.704 -79.767 44.910
83270 98.718 -79.760 44.907
83271 98.733 -79.753 44.903
83272 98.747 -79.746 44.897
83273 98.762 -79.739 44.890
83274 98.781 -79.744 44.884
83275 98.802 -79.751 44.878
83276 98.823 -79.758 44.873
83277 98.845 -79.767 44.867
83278 98.867 -79.774 44.860
83279 98.888 -79.783 44.854
83280 98.910 -79.791 44.847
83281 98.932 -79.798 44.841
83282 98.954 -79.807 44.835
83283 98.976 -79.815 44.828
83284 98.997 -79.823 44.821
83285 99.022 -79.825 44.814
83286 99.045 -79.830 44.807
83287 99.069 -79.845 44.798
83288 99.093 -79.858 44.789
83289 99.116 -79.873 44.783
83290 99.135 -79.889 44.776
83291 99.164 -79.903 44.769
83292 99.195 -79.914 44.760
83293 99.218 -79.920 44.750
83294 99.241 -79.924 44.740
83295 99.264 -79.929 44.730
83296 99.287 -79.933 44.722
83297 99.311 -79.938 44.714
83298 99.334 -79.943 44.706
83299 99.357 -79.948 44.699
83300 99.380 -79.952 44.692
83301 99.404 -79.957 44.686
83302 99.427 -79.962 44.679
83303 99.450 -79.967 44.672
83304 99.474 -79.971 44.665
83305 99.497 -79.976 44.658
83306 99.520 -79.982 44.651
83307 99.544 -79.988 44.643
83308 99.568 -79.995 44.634
83309 99.588 -79.998 44.625
83310 99.587 -80.026 44.616
83311 99.582 -80.060 44.607
83312 99.573 -80.097 44.599
83313 99.565 -80.135 44.591
83314 99.557 -80.173 44.583
83315 99.549 -80.210 44.576
83316 99.541 -80.248 44.569
83317 99.544 -80.271 44.561
83318 99.547 -80.293 44.553
83319 99.550 -80.315 44.545
83320 99.553 -80.337 44.537
83321 99.557 -80.359 44.529
83322 99.560 -80.381 44.520
83323 99.563 -80.403 44.512
83324 99.566 -80.425 44.502
83325 99.579 -80.444 44.493
83326 99.627 -80.450 44.483
83327 99.675 -80.455 44.471
83328 99.723 -80.461 44.461
83329 99.771 -80.466 44.450
83330 99.819 -80.472 44.439
83331 99.867 -80.478 44.426
83332 99.917 -80.489 44.414
83333 99.969 -80.501 44.403
83334 100.020 -80.514 44.390
83335 100.071 -80.527 44.378
83336 100.122 -80.540 44.366
83337 100.174 -80.553 44.353
83632 93.566 -78.779 53.597
83633 93.576 -78.777 53.588
83634 93.585 -78.774 53.579
83635 93.595 -78.771 53.573
83636 93.605 -78.769 53.568
83637 93.614 -78.767 53.562
83638 93.624 -78.764 53.557
83639 93.634 -78.761 53.553
83640 93.635 -78.754 53.551
83641 93.633 -78.745 53.554
83642 93.632 -78.767 53.557
83643 93.636 -78.763 53.560
83644 93.645 -78.762 53.559
83645 93.655 -78.761 53.556
83646 93.667 -78.749 53.550
83647 93.685 -78.738 53.541
83648 93.704 -78.720 53.530
83649 93.723 -78.704 53.517
83650 93.743 -78.686 53.503
83651 93.762 -78.670 53.488
83652 93.781 -78.653 53.470
83653 93.791 -78.641 53.452
83654 93.794 -78.633 53.433
83655 93.797 -78.623 53.414
83656 93.801 -78.615 53.395
83657 93.804 -78.607 53.375
83658 93.807 -78.596 53.355
83659 93.811 -78.582 53.333
83660 93.813 -78.568 53.310
83661 93.816 -78.554 53.286
83662 93.824 -78.550 53.262
83663 93.842 -78.555 53.238
83664 93.855 -78.542 53.215
83665 93.868 -78.530 53.192
83666 93.880 -78.517 53.168
83667 93.893 -78.505 53.143
83668 93.908 -78.492 53.119
83669 93.952 -78.474 53.096
83670 93.970 -78.474 53.075
83671 93.971 -78.484 53.055
83672 93.979 -78.492 53.038
83673 93.989 -78.497 53.019
83674 94.000 -78.503 52.998
83675 94.010 -78.508 52.978
83676 94.020 -78.514 52.959
83677 94.031 -78.520 52.941
83678 94.043 -78.523 52.922
83679 94.058 -78.521 52.902
83680 94.072 -78.520 52.880
83681 94.087 -78.518 52.856
83682 94.102 -78.517 52.830
83683 94.111 -78.503 52.804
83684 94.117 -78.479 52.780
83685 94.132 -78.448 52.757
83686 94.151 -78.412 52.736
83687 94.171 -78.376 52.717
83688 94.190 -78.341 52.701
83689 94.229 -78.326 52.686
83690 94.271 -78.322 52.669
83691 94.278 -78.413 52.652
83692 94.285 -78.503 52.635
83693 94.295 -78.556 52.617
83694 94.311 -78.563 52.598
83695 94.325 -78.570 52.577
83696 94.340 -78.576 52.557
83697 94.355 -78.583 52.538
83698 94.370 -78.589 52.519
83699 94.374 -78.599 52.498
83700 94.366 -78.611 52.477
83701 94.359 -78.623 52.455
83702 94.355 -78.631 52.432
83703 94.371 -78.618 52.409
83704 94.388 -78.606 52.385
83705 94.404 -78.594 52.361
83706 94.411 -78.579 52.337
83707 94.418 -78.564 52.310
83708 94.427 -78.549 52.281
83709 94.441 -78.547 52.249
83710 94.457 -78.551 52.216
83711 94.472 -78.556 52.183
83712 94.488 -78.560 52.148
83713 94.503 -78.564 52.112
83714 94.519 -78.569 52.076
83715 94.535 -78.573 52.041
83716 94.550 -78.577 52.008
83717 94.563 -78.581 51.976
83718 94.569 -78.582 51.942
83719 94.575 -78.585 51.905
83720 94.585 -78.583 51.866
83721 94.599 -78.587 51.828
83722 94.612 -78.590 51.789
83723 94.625 -78.592 51.748
83724 94.638 -78.594 51.705
83725 94.651 -78.596 51.658
83726 94.664 -78.598 51.611
83727 94.677 -78.600 51.566
83728 94.690 -78.602 51.522
83729 94.703 -78.604 51.478
83730 94.716 -78.606 51.434
83731 94.729 -78.608 51.389
83732 94.750 -78.612 51.345
83733 94.771 -78.617 51.301
83734 94.792 -78.621 51.258
83735 94.814 -78.625 51.215
83736 94.834 -78.630 51.173
83737 94.853 -78.633 51.130
83738 94.872 -78.636 51.085
83739 94.881 -78.643 51.042
83740 94.889 -78.639 50.999
83741 94.898 -78.626 50.955
83742 94.909 -78.614 50.913
83743 94.946 -78.613 50.871
83744 94.983 -78.611 50.829
83745 95.021 -78.611 50.784
83746 95.030 -78.612 50.738
83747 95.034 -78.614 50.692
83748 95.039 -78.616 50.644
83749 95.034 -78.615 50.597
83750 95.029 -78.614 50.551
83751 95.023 -78.614 50.504
83752 95.017 -78.613 50.459
83753 95.012 -78.612 50.414
83754 95.006 -78.611 50.369
83755 95.000 -78.611 50.324
83756 95.006 -78.599 50.280
83757 95.021 -78.581 50.237
83758 95.035 -78.564 50.194
83759 95.054 -78.561 50.152
83760 95.074 -78.566 50.110
83761 95.086 -78.570 50.068
83762 95.097 -78.573 50.026
83763 95.108 -78.576 49.984
83764 95.118 -78.580 49.943
83765 95.129 -78.584 49.905
83766 95.139 -78.588 49.866
83767 95.150 -78.592 49.828
83768 95.161 -78.595 49.791
83769 95.172 -78.598 49.754
83770 95.182 -78.602 49.717
83771 95.193 -78.606 49.680
83772 95.208 -78.612 49.644
83773 95.241 -78.630 49.606
83774 95.274 -78.646 49.568
83775 95.275 -78.656 49.528
83776 95.289 -78.677 49.488
83777 95.303 -78.698 49.448
83778 95.317 -78.720 49.409
83779 95.331 -78.741 49.372
83780 95.345 -78.762 49.337
83781 95.358 -78.783 49.302
83782 95.373 -78.804 49.266
83783 95.387 -78.825 49.228
83784 95.412 -78.846 49.189
83785 95.422 -78.843 49.148
83786 95.426 -78.831 49.107
83787 95.429 -78.820 49.066
83788 95.433 -78.808 49.026
83789 95.438 -78.797 48.985
83790 95.442 -78.786 48.944
83791 95.445 -78.774 48.902
83792 95.449 -78.763 48.859
83793 95.453 -78.752 48.815
83794 95.463 -78.744 48.770
83795 95.486 -78.746 48.725
83796 95.510 -78.749 48.682
83797 95.533 -78.750 48.641
83798 95.556 -78.752 48.599
83799 95.580 -78.754 48.557
83800 95.600 -78.758 48.515
83801 95.606 -78.768 48.475
83802 95.612 -78.779 48.436
83803 95.618 -78.789 48.397
83804 95.624 -78.799 48.358
83805 95.630 -78.810 48.320
83806 95.637 -78.819 48.283
83807 95.650 -78.824 48.248
83808 95.661 -78.827 48.213
83809 95.684 -78.836 48.178
83810 95.713 -78.849 48.139
83811 95.740 -78.858 48.099
83812 95.764 -78.863 48.058
83813 95.788 -78.868 48.014
83814 95.812 -78.874 47.971
83815 95.824 -78.888 47.929
83816 95.835 -78.903 47.886
83817 95.850 -78.908 47.843
83818 95.868 -78.910 47.800
83819 95.883 -78.914 47.757
83820 95.896 -78.917 47.714
83821 95.920 -78.914 47.673
83822 95.945 -78.911 47.631
83823 95.969 -78.907 47.590
83824 95.993 -78.904 47.551
83825 96.017 -78.901 47.511
83826 96.041 -78.898 47.469
83827 96.065 -78.894 47.427
83828 96.090 -78.891 47.385
83829 96.114 -78.888 47.344
83830 96.138 -78.885 47.305
83831 96.156 -78.896 47.267
83832 96.172 -78.883 47.230
83833 96.196 -78.874 47.193
83834 96.221 -78.865 47.155
83835 96.246 -78.855 47.118
83836 96.271 -78.846 47.083
83837 96.297 -78.837 47.051
83838 96.321 -78.828 47.020
83839 96.346 -78.820 46.989
83840 96.352 -78.824 46.956
83841 96.353 -78.830 46.923
83842 96.356 -78.837 46.891
83843 96.367 -78.855 46.858
83844 96.378 -78.872 46.826
83845 96.389 -78.890 46.796
83846 96.400 -78.900 46.764
83847 96.410 -78.903 46.732
83848 96.419 -78.905 46.702
83849 96.455 -78.906 46.672
83850 96.500 -78.906 46.643
83851 96.545 -78.906 46.616
83852 96.591 -78.905 46.587
83853 96.636 -78.905 46.560
83854 96.680 -78.905 46.534
83855 96.721 -78.899 46.508
83856 96.761 -78.895 46.485
83857 96.802 -78.890 46.463
83858 96.840 -78.885 46.443
83859 96.851 -78.883 46.421
83860 96.861 -78.880 46.399
83861 96.871 -78.877 46.379
83862 96.881 -78.874 46.357
83863 96.891 -78.871 46.335
83864 96.911 -78.870 46.313
83865 96.932 -78.866 46.292
83866 96.938 -78.876 46.271
83867 96.925 -78.905 46.251
83868 96.954 -78.952 46.232
83869 96.982 -78.993 46.214
83870 96.995 -78.968 46.198
83871 96.999 -78.962 46.181
83872 96.992 -78.976 46.165
83873 96.986 -78.991 46.147
83874 96.979 -79.006 46.129
83875 96.973 -79.021 46.112
83876 96.987 -79.059 46.096
83877 97.008 -79.105 46.082
83878 97.029 -79.152 46.069
83879 97.049 -79.160 46.056
83880 97.070 -79.156 46.038
83881 97.090 -79.150 46.022
83882 97.110 -79.145 46.008
83883 97.131 -79.140 45.993
83884 97.151 -79.134 45.980
83885 97.171 -79.129 45.966
83886 97.191 -79.125 45.953
83887 97.206 -79.106 45.938
83888 97.218 -79.094 45.924
83889 97.227 -79.099 45.909
83890 97.234 -79.101 45.892
83891 97.240 -79.104 45.875
83892 97.250 -79.110 45.859
83893 97.264 -79.119 45.843
83894 97.278 -79.128 45.827
83895 97.292 -79.138 45.810
83896 97.305 -79.147 45.793
83897 97.319 -79.156 45.774
83898 97.333 -79.165 45.759
83899 97.348 -79.175 45.744
83900 97.363 -79.183 45.729
83901 97.378 -79.192 45.711
83902 97.393 -79.200 45.693
83903 97.408 -79.209 45.674
83904 97.419 -79.208 45.657
83905 97.431 -79.207 45.641
83906 97.452 -79.217 45.624
83907 97.474 -79.228 45.607
83908 97.492 -79.234 45.591
83909 97.507 -79.237 45.575
83910 97.522 -79.240 45.559
83911 97.536 -79.243 45.542
83912 97.551 -79.246 45.526
83913 97.566 -79.249 45.509
83914 97.580 -79.252 45.493
83915 97.597 -79.257 45.476
83916 97.617 -79.266 45.459
83917 97.637 -79.274 45.441
83918 97.648 -79.278 45.425
83919 97.648 -79.278 45.409
83920 97.648 -79.278 45.394
83921 97.648 -79.278 45.380
83922 97.648 -79.277 45.366
83923 97.648 -79.276 45.351
83924 97.648 -79.276 45.338
83925 97.649 -79.278 45.325
83926 97.667 -79.297 45.313
83927 97.685 -79.318 45.303
83928 97.704 -79.338 45.293
83929 97.722 -79.359 45.283
83930 97.740 -79.379 45.272
83931 97.758 -79.400 45.260
83932 97.777 -79.420 45.248
83933 97.795 -79.441 45.237
83934 97.813 -79.461 45.226
83935 97.840 -79.480 45.215
83936 97.854 -79.492 45.204
83937 97.866 -79.507 45.192
83938 97.878 -79.523 45.180
83939 97.890 -79.537 45.169
83940 97.902 -79.549 45.159
83941 97.914 -79.560 45.149
83942 97.925 -79.573 45.140
83943 97.937 -79.584 45.131
83944 97.948 -79.596 45.123
83945 97.962 -79.607 45.114
83946 97.989 -79.617 45.105
83947 98.019 -79.623 45.097
83948 98.056 -79.614 45.090
83949 98.093 -79.605 45.082
83950 98.130 -79.596 45.075
83951 98.167 -79.587 45.068
83952 98.205 -79.578 45.062
83953 98.239 -79.567 45.053
83954 98.272 -79.585 45.045
83955 98.308 -79.610 45.036
83956 98.344 -79.635 45.027
83957 98.379 -79.659 45.017
83958 98.406 -79.666 45.008
83959 98.432 -79.672 44.999
83960 98.458 -79.679 44.988
83961 98.486 -79.689 44.979
83962 98.512 -79.700 44.971
83963 98.539 -79.711 44.963
83964 98.567 -79.720 44.956
83965 98.594 -79.731 44.949
83966 98.621 -79.742 44.942
83967 98.648 -79.751 44.937
83968 98.680 -79.761 44.932
83969 98.698 -79.757 44.928
83970 98.712 -79.750 44.923
83971 98.726 -79.744 44.919
83972 98.741 -79.737 44.915
83973 98.755 -79.730 44.910
83974 98.770 -79.723 44.904
83975 98.786 -79.722 44.898
83976 98.807 -79.730 44.893
83977 98.829 -79.737 44.888
83978 98.851 -79.745 44.882
83979 98.872 -79.753 44.876
83980 98.894 -79.761 44.869
83981 98.916 -79.769 44.863
83982 98.938 -79.777 44.857
83983 98.959 -79.786 44.850
83984 98.981 -79.793 44.843
83985 99.003 -79.802 44.837
83986 99.025 -79.810 44.829
83987 99.049 -79.815 44.821
83988 99.072 -79.829 44.811
83989 99.096 -79.843 44.802
83990 99.121 -79.857 44.795
83991 99.142 -79.872 44.788
83992 99.170 -79.881 44.778
83993 99.194 -79.886 44.768
83994 99.217 -79.891 44.757
83995 99.240 -79.896 44.746
83996 99.264 -79.900 44.736
83997 99.287 -79.905 44.727
83998 99.310 -79.912 44.719
83999 99.334 -79.918 44.711
84000 99.358 -79.924 44.704
84001 99.382 -79.930 44.697
84002 99.406 -79.937 44.691
84003 99.429 -79.943 44.685
84004 99.453 -79.949 44.676
84005 99.477 -79.955 44.668
84006 99.500 -79.962 44.661
84007 99.524 -79.968 44.654
84008 99.548 -79.974 44.646
84009 99.570 -79.980 44.637
84010 99.584 -79.977 44.628
84011 99.582 -80.005 44.618
84012 99.581 -80.034 44.610
84013 99.580 -80.062 44.602
84014 99.579 -80.090 44.595
84015 99.577 -80.118 44.587
84016 99.570 -80.155 44.580
84017 99.561 -80.192 44.572
84018 99.553 -80.230 44.564
84019 99.545 -80.268 44.557
84020 99.547 -80.292 44.549
84021 99.550 -80.314 44.541
84022 99.553 -80.336 44.533
84023 99.556 -80.359 44.525
84024 99.559 -80.381 44.517
84025 99.563 -80.403 44.507
84026 99.587 -80.417 44.498
84027 99.635 -80.422 44.487
84028 99.683 -80.428 44.476
84029 99.731 -80.434 44.466
84030 99.779 -80.439 44.455
84031 99.830 -80.450 44.444
84032 99.881 -80.463 44.432
84033 99.932 -80.476 44.420
84034 99.984 -80.489 44.408
84035 100.035 -80.502 44.396
84036 100.086 -80.515 44.383
84037 100.137 -80.527 44.370
84038 100.189 -80.540 44.358
84039 100.240 -80.553 44.346
84333 93.566 -78.767 53.603
84334 93.576 -78.765 53.595
84335 93.585 -78.763 53.587
84336 93.595 -78.761 53.579
84337 93.605 -78.758 53.573
84338 93.615 -78.755 53.568
84339 93.624 -78.753 53.562
84340 93.633 -78.750 53.558
84341 93.631 -78.740 53.555
84342 93.630 -78.731 53.556
84343 93.628 -78.739 53.559
84344 93.630 -78.757 53.562
84345 93.639 -78.756 53.562
84346 93.650 -78.739 53.559
84347 93.665 -78.727 53.553
84348 93.683 -78.717 53.544
84349 93.700 -78.706 53.533
84350 93.719 -78.689 53.521
84351 93.739 -78.673 53.506
84352 93.758 -78.655 53.491
84353 93.777 -78.639 53.474
84354 93.785 -78.627 53.455
84355 93.789 -78.619 53.436
84356 93.792 -78.611 53.417
84357 93.796 -78.601 53.397
84358 93.799 -78.589 53.377
84359 93.801 -78.575 53.357
84360 93.804 -78.561 53.334
84361 93.807 -78.547 53.310
84362 93.809 -78.532 53.285
84363 93.812 -78.518 53.260
84364 93.837 -78.544 53.236
84365 93.850 -78.532 53.213
84366 93.863 -78.519 53.190
84367 93.875 -78.507 53.166
84368 93.888 -78.494 53.141
84369 93.921 -78.478 53.118
84370 93.965 -78.460 53.097
84371 93.980 -78.462 53.078
84372 93.981 -78.472 53.060
84373 93.984 -78.482 53.044
84374 93.994 -78.488 53.025
84375 94.004 -78.494 53.004
84376 94.015 -78.499 52.984
84377 94.025 -78.505 52.964
84378 94.039 -78.504 52.944
84379 94.054 -78.503 52.923
84380 94.068 -78.501 52.903
84381 94.082 -78.501 52.880
84382 94.097 -78.499 52.855
84383 94.112 -78.498 52.828
84384 94.126 -78.496 52.804
84385 94.133 -78.476 52.782
84386 94.144 -78.448 52.762
84387 94.164 -78.413 52.743
84388 94.183 -78.377 52.725
84389 94.203 -78.341 52.708
84390 94.222 -78.306 52.692
84391 94.249 -78.290 52.676
84392 94.256 -78.380 52.659
84393 94.263 -78.470 52.641
84394 94.271 -78.557 52.623
84395 94.286 -78.564 52.604
84396 94.300 -78.571 52.584
84397 94.316 -78.578 52.565
84398 94.330 -78.584 52.546
84399 94.345 -78.591 52.527
84400 94.360 -78.598 52.508
84401 94.372 -78.605 52.489
84402 94.365 -78.617 52.468
84403 94.357 -78.629 52.444
84404 94.373 -78.617 52.422
84405 94.390 -78.604 52.399
84406 94.397 -78.589 52.375
84407 94.406 -78.574 52.349
84408 94.423 -78.560 52.322
84409 94.441 -78.545 52.292
84410 94.450 -78.545 52.259
84411 94.460 -78.545 52.227
84412 94.475 -78.550 52.194
84413 94.490 -78.554 52.160
84414 94.506 -78.558 52.124
84415 94.522 -78.563 52.088
84416 94.537 -78.567 52.054
84417 94.553 -78.571 52.022
84418 94.568 -78.576 51.991
84419 94.574 -78.577 51.957
84420 94.580 -78.579 51.921
84421 94.591 -78.578 51.882
84422 94.601 -78.577 51.844
84423 94.615 -78.581 51.804
84424 94.628 -78.585 51.763
84425 94.641 -78.588 51.719
84426 94.654 -78.590 51.672
84427 94.668 -78.592 51.623
84428 94.680 -78.594 51.577
84429 94.693 -78.596 51.530
84430 94.706 -78.598 51.485
84431 94.719 -78.600 51.438
84432 94.733 -78.602 51.391
84433 94.754 -78.606 51.345
84434 94.775 -78.611 51.299
84435 94.796 -78.615 51.255
84436 94.817 -78.619 51.211
84437 94.839 -78.623 51.166
84438 94.859 -78.627 51.121
84439 94.878 -78.631 51.076
84440 94.891 -78.636 51.032
84441 94.899 -78.629 50.989
84442 94.913 -78.618 50.945
84443 94.951 -78.617 50.903
84444 94.982 -78.616 50.861
84445 94.987 -78.618 50.818
84446 94.991 -78.620 50.775
84447 94.996 -78.622 50.732
84448 95.000 -78.623 50.688
84449 94.999 -78.624 50.641
84450 94.993 -78.623 50.595
84451 94.987 -78.623 50.550
84452 94.982 -78.622 50.505
84453 94.976 -78.620 50.460
84454 94.971 -78.620 50.415
84455 94.983 -78.604 50.370
84456 94.997 -78.586 50.325
84457 95.012 -78.568 50.281
84458 95.026 -78.550 50.238
84459 95.041 -78.535 50.194
84460 95.062 -78.540 50.151
84461 95.082 -78.545 50.109
84462 95.100 -78.550 50.067
84463 95.111 -78.554 50.025
84464 95.121 -78.557 49.984
84465 95.132 -78.561 49.942
84466 95.143 -78.565 49.903
84467 95.154 -78.569 49.865
84468 95.164 -78.572 49.827
84469 95.175 -78.576 49.790
84470 95.186 -78.579 49.753
84471 95.196 -78.583 49.716
84472 95.207 -78.587 49.679
84473 95.235 -78.601 49.641
84474 95.268 -78.617 49.602
84475 95.301 -78.635 49.564
84476 95.333 -78.651 49.526
84477 95.333 -78.664 49.487
84478 95.346 -78.686 49.448
84479 95.360 -78.706 49.410
84480 95.375 -78.727 49.374
84481 95.388 -78.749 49.340
84482 95.402 -78.770 49.306
84483 95.416 -78.791 49.271
84484 95.430 -78.811 49.234
84485 95.438 -78.825 49.196
84486 95.429 -78.824 49.156
84487 95.416 -78.818 49.115
84488 95.420 -78.807 49.074
84489 95.424 -78.796 49.033
84490 95.428 -78.784 48.992
84491 95.432 -78.773 48.951
84492 95.436 -78.761 48.908
84493 95.440 -78.750 48.864
84494 95.444 -78.739 48.819
84495 95.453 -78.731 48.775
84496 95.476 -78.733 48.731
84497 95.499 -78.735 48.688
84498 95.523 -78.736 48.647
84499 95.546 -78.739 48.606
84500 95.570 -78.741 48.564
84501 95.593 -78.742 48.523
84502 95.609 -78.749 48.483
84503 95.615 -78.759 48.444
84504 95.621 -78.769 48.406
84505 95.626 -78.780 48.367
84506 95.633 -78.789 48.330
84507 95.645 -78.794 48.294
84508 95.657 -78.798 48.260
84509 95.669 -78.802 48.225
84510 95.690 -78.811 48.189
84511 95.719 -78.822 48.150
84512 95.748 -78.834 48.111
84513 95.775 -78.843 48.070
84514 95.799 -78.849 48.026
84515 95.812 -78.863 47.984
84516 95.828 -78.865 47.942
84517 95.845 -78.866 47.900
84518 95.863 -78.867 47.858
84519 95.879 -78.870 47.815
84520 95.894 -78.875 47.772
84521 95.901 -78.881 47.729
84522 95.923 -78.879 47.686
84523 95.948 -78.876 47.645
84524 95.972 -78.873 47.603
84525 95.996 -78.869 47.562
84526 96.020 -78.866 47.520
84527 96.044 -78.863 47.478
84528 96.068 -78.860 47.436
84529 96.093 -78.856 47.395
84530 96.117 -78.853 47.355
84531 96.137 -78.861 47.317
84532 96.155 -78.879 47.278
84533 96.172 -78.873 47.242
84534 96.188 -78.861 47.205
84535 96.205 -78.849 47.167
84536 96.227 -78.838 47.130
84537 96.252 -78.829 47.094
84538 96.277 -78.820 47.060
84539 96.302 -78.811 47.028
84540 96.327 -78.802 46.997
84541 96.336 -78.803 46.965
84542 96.337 -78.810 46.930
84543 96.346 -78.824 46.897
84544 96.358 -78.843 46.864
84545 96.369 -78.860 46.832
84546 96.380 -78.877 46.801
84547 96.391 -78.893 46.771
84548 96.401 -78.895 46.740
84549 96.410 -78.898 46.710
84550 96.449 -78.898 46.681
84551 96.494 -78.898 46.653
84552 96.539 -78.898 46.626
84553 96.585 -78.898 46.598
84554 96.629 -78.897 46.572
84555 96.670 -78.893 46.546
84556 96.711 -78.887 46.520
84557 96.751 -78.883 46.496
84558 96.791 -78.878 46.474
84559 96.832 -78.873 46.454
84560 96.848 -78.870 46.432
84561 96.858 -78.867 46.410
84562 96.868 -78.865 46.387
84563 96.878 -78.861 46.366
84564 96.890 -78.862 46.345
84565 96.911 -78.865 46.324
84566 96.932 -78.861 46.304
84567 96.940 -78.869 46.284
84568 96.930 -78.900 46.265
84569 96.956 -78.957 46.246
84570 96.971 -78.996 46.228
84571 96.980 -78.984 46.212
84572 96.993 -78.959 46.195
84573 96.989 -78.968 46.178
84574 96.982 -78.984 46.162
84575 96.976 -78.998 46.146
84576 96.970 -79.013 46.129
84577 96.984 -79.053 46.113
84578 97.006 -79.099 46.098
84579 97.027 -79.136 46.084
84580 97.047 -79.131 46.069
84581 97.067 -79.125 46.054
84582 97.087 -79.120 46.038
84583 97.108 -79.115 46.023
84584 97.128 -79.110 46.009
84585 97.148 -79.105 45.996
84586 97.168 -79.100 45.983
84587 97.188 -79.094 45.971
84588 97.209 -79.089 45.957
84589 97.223 -79.093 45.942
84590 97.229 -79.096 45.926
84591 97.235 -79.099 45.909
84592 97.242 -79.102 45.892
84593 97.251 -79.107 45.875
84594 97.265 -79.117 45.858
84595 97.280 -79.125 45.839
84596 97.295 -79.134 45.821
84597 97.310 -79.143 45.803
84598 97.325 -79.152 45.783
84599 97.341 -79.160 45.766
84600 97.356 -79.169 45.750
84601 97.371 -79.178 45.734
84602 97.385 -79.187 45.716
84603 97.400 -79.193 45.697
84604 97.411 -79.192 45.678
84605 97.423 -79.191 45.660
84606 97.434 -79.190 45.644
84607 97.453 -79.198 45.628
84608 97.475 -79.209 45.612
84609 97.498 -79.220 45.595
84610 97.520 -79.231 45.578
84611 97.535 -79.235 45.562
84612 97.550 -79.238 45.546
84613 97.565 -79.241 45.530
84614 97.580 -79.244 45.514
84615 97.594 -79.247 45.499
84616 97.613 -79.254 45.482
84617 97.633 -79.263 45.465
84618 97.640 -79.266 45.449
84619 97.640 -79.265 45.434
84620 97.640 -79.264 45.419
84621 97.639 -79.264 45.405
84622 97.639 -79.263 45.392
84623 97.639 -79.263 45.377
84624 97.639 -79.263 45.363
84625 97.639 -79.262 45.349
84626 97.641 -79.265 45.335
84627 97.659 -79.285 45.325
84628 97.678 -79.305 45.315
84629 97.696 -79.325 45.303
84630 97.714 -79.346 45.293
84631 97.733 -79.366 45.281
84632 97.751 -79.387 45.268
84633 97.769 -79.407 45.255
84634 97.788 -79.428 45.244
84635 97.816 -79.450 45.233
84636 97.840 -79.463 45.223
84637 97.855 -79.473 45.212
84638 97.867 -79.488 45.201
84639 97.879 -79.504 45.190
84640 97.892 -79.519 45.178
84641 97.904 -79.534 45.168
84642 97.916 -79.549 45.158
84643 97.929 -79.564 45.150
84644 97.941 -79.579 45.140
84645 97.952 -79.591 45.131
84646 97.964 -79.602 45.123
84647 97.993 -79.608 45.115
84648 98.030 -79.599 45.107
84649 98.067 -79.590 45.099
84650 98.104 -79.581 45.092
84651 98.142 -79.572 45.085
84652 98.178 -79.563 45.077
84653 98.216 -79.554 45.070
84654 98.249 -79.542 45.062
84655 98.282 -79.560 45.054
84656 98.318 -79.585 45.046
84657 98.354 -79.610 45.037
84658 98.389 -79.635 45.027
84659 98.422 -79.653 45.018
84660 98.448 -79.659 45.009
84661 98.475 -79.668 45.000
84662 98.502 -79.679 44.991
84663 98.529 -79.689 44.983
84664 98.557 -79.699 44.976
84665 98.584 -79.709 44.968
84666 98.611 -79.720 44.961
84667 98.638 -79.730 44.954
84668 98.665 -79.740 44.949
84669 98.692 -79.748 44.944
84670 98.706 -79.741 44.940
84671 98.720 -79.734 44.936
84672 98.734 -79.727 44.932
84673 98.749 -79.720 44.928
84674 98.763 -79.714 44.923
84675 98.778 -79.707 44.919
84676 98.792 -79.701 44.913
84677 98.813 -79.708 44.908
84678 98.835 -79.716 44.903
84679 98.856 -79.723 44.897
84680 98.878 -79.732 44.891
84681 98.900 -79.740 44.884
84682 98.922 -79.748 44.879
84683 98.943 -79.756 44.872
84684 98.966 -79.764 44.865
84685 98.987 -79.772 44.858
84686 99.008 -79.781 44.851
84687 99.027 -79.792 44.843
84688 99.058 -79.799 44.834
84689 99.082 -79.813 44.824
84690 99.106 -79.827 44.814
84691 99.130 -79.841 44.806
84692 99.148 -79.855 44.797
84693 99.172 -79.861 44.786
84694 99.196 -79.867 44.774
84695 99.219 -79.874 44.763
84696 99.243 -79.880 44.752
84697 99.267 -79.886 44.743
84698 99.291 -79.892 44.735
84699 99.314 -79.899 44.727
84700 99.338 -79.905 44.719
84701 99.362 -79.911 44.711
84702 99.385 -79.917 44.703
84703 99.409 -79.924 44.697
84704 99.433 -79.930 44.690
84705 99.457 -79.936 44.681
84706 99.481 -79.942 44.673
84707 99.504 -79.949 44.665
84708 99.528 -79.955 44.657
84709 99.552 -79.961 44.648
84710 99.567 -79.956 44.641
84711 99.579 -79.957 44.632
84712 99.578 -79.985 44.622
84713 99.576 -80.013 44.614
84714 99.575 -80.041 44.605
84715 99.574 -80.069 44.597
84716 99.573 -80.097 44.590
84717 99.571 -80.125 44.583
84718 99.570 -80.153 44.575
84719 99.569 -80.181 44.567
84720 99.566 -80.212 44.560
84721 99.558 -80.250 44.553
84722 99.549 -80.287 44.545
84723 99.550 -80.313 44.538
84724 99.553 -80.335 44.530
84725 99.556 -80.358 44.521
84726 99.559 -80.380 44.512
84727 99.595 -80.390 44.502
84728 99.644 -80.395 44.492
84729 99.692 -80.401 44.481
84730 99.742 -80.413 44.471
84731 99.793 -80.425 44.460
84732 99.845 -80.438 44.449
84733 99.896 -80.450 44.437
84734 99.947 -80.463 44.425
84735 99.999 -80.476 44.413
84736 100.050 -80.489 44.401
84737 100.101 -80.502 44.387
84738 100.152 -80.515 44.375
84739 100.198 -80.530 44.363
84740 100.243 -80.547 44.349
85034 93.566 -78.757 53.610
85035 93.576 -78.755 53.602
85036 93.586 -78.752 53.594
85037 93.595 -78.749 53.586
85038 93.605 -78.747 53.578
85039 93.615 -78.745 53.573
85040 93.625 -78.742 53.567
85041 93.629 -78.736 53.562
85042 93.628 -78.727 53.559
85043 93.627 -78.718 53.559
85044 93.625 -78.709 53.561
85045 93.624 -78.749 53.565
85046 93.635 -78.732 53.564
85047 93.646 -78.715 53.560
85048 93.663 -78.705 53.555
85049 93.681 -78.694 53.546
85050 93.698 -78.683 53.535
85051 93.716 -78.673 53.522
85052 93.734 -78.658 53.508
85053 93.754 -78.642 53.492
85054 93.773 -78.624 53.475
85055 93.780 -78.614 53.456
85056 93.784 -78.605 53.437
85057 93.787 -78.596 53.418
85058 93.790 -78.582 53.398
85059 93.792 -78.568 53.377
85060 93.795 -78.554 53.355
85061 93.798 -78.540 53.332
85062 93.800 -78.526 53.308
85063 93.803 -78.511 53.283
85064 93.806 -78.497 53.258
85065 93.827 -78.517 53.233
85066 93.845 -78.520 53.209
85067 93.858 -78.508 53.186
85068 93.870 -78.495 53.162
85069 93.889 -78.482 53.137
85070 93.933 -78.464 53.117
85071 93.977 -78.447 53.098
85072 93.990 -78.450 53.081
85073 93.991 -78.460 53.065
85074 93.992 -78.470 53.048
85075 93.998 -78.479 53.030
85076 94.009 -78.484 53.009
85077 94.020 -78.489 52.988
85078 94.035 -78.487 52.966
85079 94.049 -78.485 52.945
85080 94.064 -78.484 52.923
85081 94.078 -78.482 52.902
85082 94.093 -78.481 52.879
85083 94.108 -78.479 52.853
85084 94.122 -78.478 52.827
85085 94.137 -78.476 52.805
85086 94.150 -78.474 52.785
85087 94.157 -78.449 52.767
85088 94.177 -78.413 52.749
85089 94.196 -78.379 52.731
85090 94.215 -78.343 52.714
85091 94.233 -78.310 52.697
85092 94.236 -78.304 52.680
85093 94.237 -78.360 52.662
85094 94.242 -78.438 52.643
85095 94.249 -78.528 52.626
85096 94.261 -78.566 52.608
85097 94.276 -78.573 52.589
85098 94.290 -78.579 52.571
85099 94.306 -78.586 52.552
85100 94.321 -78.592 52.534
85101 94.335 -78.599 52.515
85102 94.350 -78.606 52.497
85103 94.365 -78.612 52.478
85104 94.363 -78.623 52.455
85105 94.376 -78.614 52.433
85106 94.385 -78.600 52.410
85107 94.402 -78.586 52.385
85108 94.420 -78.570 52.360
85109 94.437 -78.556 52.332
85110 94.453 -78.545 52.302
85111 94.462 -78.545 52.269
85112 94.471 -78.545 52.236
85113 94.480 -78.546 52.202
85114 94.493 -78.548 52.167
85115 94.508 -78.552 52.132
85116 94.524 -78.557 52.098
85117 94.540 -78.561 52.064
85118 94.555 -78.565 52.032
85119 94.571 -78.570 52.002
85120 94.579 -78.572 51.969
85121 94.586 -78.574 51.932
85122 94.596 -78.573 51.895
85123 94.607 -78.572 51.856
85124 94.617 -78.571 51.816
85125 94.631 -78.574 51.774
85126 94.644 -78.578 51.730
85127 94.657 -78.582 51.684
85128 94.671 -78.586 51.635
85129 94.684 -78.588 51.587
85130 94.697 -78.590 51.538
85131 94.710 -78.592 51.489
85132 94.723 -78.594 51.442
85133 94.736 -78.596 51.393
85134 94.757 -78.600 51.345
85135 94.778 -78.604 51.298
85136 94.800 -78.608 51.252
85137 94.821 -78.613 51.206
85138 94.842 -78.617 51.160
85139 94.863 -78.621 51.114
85140 94.885 -78.626 51.069
85141 94.901 -78.630 51.025
85142 94.918 -78.622 50.982
85143 94.939 -78.623 50.937
85144 94.944 -78.624 50.894
85145 94.949 -78.626 50.852
85146 94.953 -78.628 50.809
85147 94.958 -78.630 50.767
85148 94.962 -78.632 50.725
85149 94.964 -78.633 50.684
85150 94.958 -78.632 50.639
85151 94.952 -78.631 50.594
85152 94.947 -78.630 50.549
85153 94.945 -78.626 50.505
85154 94.959 -78.608 50.461
85155 94.974 -78.591 50.416
85156 94.988 -78.573 50.371
85157 95.002 -78.554 50.327
85158 95.017 -78.537 50.283
85159 95.031 -78.519 50.238
85160 95.049 -78.515 50.195
85161 95.069 -78.520 50.152
85162 95.090 -78.526 50.109
85163 95.111 -78.530 50.067
85164 95.125 -78.535 50.025
85165 95.135 -78.539 49.983
85166 95.146 -78.542 49.942
85167 95.157 -78.545 49.902
85168 95.168 -78.549 49.864
85169 95.178 -78.553 49.827
85170 95.189 -78.557 49.789
85171 95.199 -78.561 49.752
85172 95.210 -78.564 49.715
85173 95.229 -78.573 49.677
85174 95.261 -78.589 49.639
85175 95.295 -78.606 49.600
85176 95.328 -78.623 49.562
85177 95.360 -78.640 49.524
85178 95.381 -78.654 49.487
85179 95.390 -78.672 49.449
85180 95.404 -78.693 49.410
85181 95.418 -78.714 49.376
85182 95.432 -78.736 49.342
85183 95.446 -78.756 49.308
85184 95.460 -78.777 49.273
85185 95.471 -78.796 49.238
85186 95.465 -78.795 49.201
85187 95.460 -78.794 49.163
85188 95.450 -78.792 49.122
85189 95.432 -78.786 49.082
85190 95.415 -78.782 49.040
85191 95.418 -78.771 48.999
85192 95.422 -78.759 48.957
85193 95.426 -78.748 48.913
85194 95.430 -78.737 48.868
85195 95.434 -78.726 48.824
85196 95.443 -78.717 48.780
85197 95.466 -78.719 48.736
85198 95.489 -78.721 48.694
85199 95.513 -78.724 48.653
85200 95.536 -78.725 48.612
85201 95.559 -78.727 48.572
85202 95.583 -78.729 48.531
85203 95.606 -78.731 48.492
85204 95.617 -78.739 48.453
85205 95.623 -78.749 48.414
85206 95.629 -78.760 48.376
85207 95.641 -78.764 48.339
85208 95.653 -78.768 48.303
85209 95.665 -78.773 48.268
85210 95.677 -78.777 48.233
85211 95.695 -78.784 48.197
85212 95.725 -78.796 48.159
85213 95.753 -78.808 48.120
85214 95.782 -78.820 48.079
85215 95.806 -78.821 48.038
85216 95.823 -78.822 47.998
85217 95.841 -78.823 47.956
85218 95.858 -78.824 47.914
85219 95.875 -78.826 47.872
85220 95.890 -78.831 47.829
85221 95.904 -78.836 47.787
85222 95.911 -78.843 47.744
85223 95.926 -78.844 47.702
85224 95.950 -78.841 47.659
85225 95.974 -78.838 47.616
85226 95.999 -78.835 47.573
85227 96.023 -78.831 47.531
85228 96.047 -78.828 47.489
85229 96.071 -78.825 47.447
85230 96.095 -78.822 47.407
85231 96.119 -78.821 47.367
85232 96.136 -78.839 47.329
85233 96.154 -78.856 47.291
85234 96.171 -78.862 47.254
85235 96.188 -78.850 47.217
85236 96.204 -78.838 47.179
85237 96.221 -78.825 47.142
85238 96.237 -78.813 47.106
85239 96.258 -78.802 47.071
85240 96.283 -78.793 47.039
85241 96.308 -78.784 47.007
85242 96.320 -78.783 46.974
85243 96.325 -78.794 46.939
85244 96.337 -78.812 46.906
85245 96.348 -78.830 46.874
85246 96.360 -78.847 46.842
85247 96.371 -78.865 46.811
85248 96.382 -78.883 46.781
85249 96.392 -78.887 46.750
85250 96.401 -78.890 46.720
85251 96.443 -78.890 46.691
85252 96.488 -78.890 46.664
85253 96.533 -78.890 46.637
85254 96.579 -78.890 46.610
85255 96.619 -78.885 46.584
85256 96.660 -78.880 46.558
85257 96.700 -78.875 46.531
85258 96.741 -78.871 46.508
85259 96.781 -78.866 46.486
85260 96.822 -78.861 46.466
85261 96.844 -78.857 46.444
85262 96.855 -78.855 46.421
85263 96.865 -78.852 46.398
85264 96.875 -78.849 46.378
85265 96.890 -78.853 46.357
85266 96.910 -78.861 46.337
85267 96.931 -78.857 46.317
85268 96.941 -78.862 46.297
85269 96.934 -78.896 46.278
85270 96.960 -78.952 46.260
85271 96.971 -78.978 46.242
85272 96.966 -78.993 46.226
85273 96.977 -78.976 46.210
85274 96.985 -78.961 46.193
85275 96.979 -78.976 46.177
85276 96.973 -78.990 46.161
85277 96.966 -79.006 46.145
85278 96.982 -79.046 46.129
85279 97.004 -79.093 46.114
85280 97.024 -79.106 46.099
85281 97.044 -79.101 46.084
85282 97.065 -79.096 46.069
85283 97.085 -79.090 46.053
85284 97.105 -79.085 46.037
85285 97.125 -79.080 46.023
85286 97.146 -79.075 46.011
85287 97.166 -79.070 45.999
85288 97.189 -79.073 45.986
85289 97.213 -79.078 45.972
85290 97.230 -79.087 45.956
85291 97.237 -79.091 45.938
85292 97.243 -79.095 45.922
85293 97.248 -79.098 45.906
85294 97.257 -79.103 45.889
85295 97.272 -79.112 45.870
85296 97.287 -79.120 45.849
85297 97.302 -79.129 45.829
85298 97.317 -79.137 45.810
85299 97.332 -79.147 45.791
85300 97.348 -79.155 45.772
85301 97.363 -79.164 45.755
85302 97.378 -79.173 45.738
85303 97.391 -79.178 45.720
85304 97.403 -79.177 45.702
85305 97.414 -79.176 45.683
85306 97.426 -79.175 45.666
85307 97.437 -79.175 45.650
85308 97.454 -79.179 45.634
85309 97.476 -79.190 45.618
85310 97.498 -79.201 45.602
85311 97.520 -79.212 45.585
85312 97.542 -79.223 45.568
85313 97.564 -79.233 45.552
85314 97.579 -79.236 45.536
85315 97.593 -79.239 45.521
85316 97.608 -79.243 45.506
85317 97.629 -79.251 45.489
85318 97.631 -79.252 45.473
85319 97.631 -79.251 45.458
85320 97.631 -79.250 45.444
85321 97.631 -79.250 45.430
85322 97.631 -79.250 45.417
85323 97.631 -79.250 45.404
85324 97.630 -79.249 45.389
85325 97.630 -79.249 45.375
85326 97.630 -79.248 45.360
85327 97.634 -79.252 45.346
85328 97.652 -79.272 45.335
85329 97.670 -79.292 45.325
85330 97.688 -79.313 45.312
85331 97.707 -79.333 45.300
85332 97.725 -79.353 45.288
85333 97.743 -79.374 45.276
85334 97.761 -79.394 45.263
85335 97.788 -79.416 45.252
85336 97.823 -79.436 45.241
85337 97.840 -79.444 45.231
85338 97.856 -79.454 45.220
85339 97.868 -79.469 45.210
85340 97.880 -79.485 45.200
85341 97.892 -79.500 45.189
85342 97.905 -79.515 45.179
85343 97.917 -79.530 45.170
85344 97.930 -79.545 45.160
85345 97.942 -79.561 45.152
85346 97.954 -79.576 45.143
85347 97.967 -79.591 45.134
85348 97.995 -79.595 45.125
85349 98.041 -79.575 45.117
85350 98.078 -79.566 45.109
85351 98.115 -79.557 45.102
85352 98.153 -79.548 45.095
85353 98.190 -79.538 45.087
85354 98.227 -79.529 45.080
85355 98.260 -79.517 45.071
85356 98.293 -79.535 45.063
85357 98.329 -79.560 45.055
85358 98.364 -79.585 45.047
85359 98.400 -79.610 45.038
85360 98.436 -79.634 45.029
85361 98.465 -79.646 45.020
85362 98.492 -79.657 45.011
85363 98.519 -79.667 45.003
85364 98.546 -79.677 44.996
85365 98.573 -79.688 44.988
85366 98.598 -79.697 44.980
85367 98.624 -79.706 44.973
85368 98.649 -79.715 44.966
85369 98.675 -79.724 44.960
85370 98.690 -79.738 44.956
85371 98.707 -79.729 44.952
85372 98.724 -79.720 44.948
85373 98.741 -79.712 44.944
85374 98.757 -79.704 44.940
85375 98.771 -79.698 44.937
85376 98.786 -79.691 44.933
85377 98.800 -79.684 44.929
85378 98.818 -79.686 44.923
85379 98.841 -79.695 44.919
85380 98.862 -79.702 44.913
85381 98.884 -79.711 44.907
85382 98.905 -79.720 44.900
85383 98.924 -79.731 44.894
85384 98.944 -79.742 44.887
85385 98.963 -79.754 44.879
85386 98.982 -79.765 44.870
85387 99.001 -79.776 44.862
85388 99.037 -79.780 44.853
85389 99.076 -79.783 44.844
85390 99.100 -79.796 44.834
85391 99.125 -79.811 44.822
85392 99.148 -79.824 44.810
85393 99.157 -79.839 44.800
85394 99.176 -79.847 44.789
85395 99.200 -79.853 44.778
85396 99.223 -79.859 44.768
85397 99.247 -79.866 44.758
85398 99.271 -79.872 44.749
85399 99.294 -79.878 44.742
85400 99.319 -79.884 44.734
85401 99.342 -79.891 44.726
85402 99.366 -79.897 44.717
85403 99.390 -79.903 44.709
85404 99.413 -79.909 44.701
85405 99.437 -79.916 44.694
85406 99.461 -79.922 44.685
85407 99.484 -79.928 44.677
85408 99.508 -79.935 44.669
85409 99.532 -79.941 44.660
85410 99.549 -79.938 44.651
85411 99.564 -79.932 44.644
85412 99.574 -79.936 44.636
85413 99.573 -79.965 44.626
85414 99.572 -79.993 44.617
85415 99.571 -80.021 44.608
85416 99.569 -80.049 44.600
85417 99.568 -80.077 44.593
85418 99.567 -80.105 44.585
85419 99.566 -80.133 44.577
85420 99.564 -80.161 44.571
85421 99.563 -80.189 44.563
85422 99.562 -80.216 44.555
85423 99.561 -80.244 44.549
85424 99.559 -80.272 44.541
85425 99.554 -80.306 44.533
85426 99.553 -80.335 44.525
85427 99.556 -80.357 44.517
85428 99.604 -80.363 44.507
85429 99.655 -80.374 44.497
85430 99.706 -80.387 44.487
85431 99.757 -80.400 44.476
85432 99.808 -80.413 44.465
85433 99.860 -80.425 44.453
85434 99.911 -80.438 44.441
85435 99.963 -80.451 44.430
85436 100.014 -80.463 44.417
85437 100.065 -80.476 44.405
85438 100.112 -80.492 44.392
85439 100.156 -80.508 44.379
85440 100.201 -80.525 44.367
85441 100.245 -80.541 44.354
85735 93.567 -78.746 53.617
85736 93.576 -78.743 53.609
85737 93.586 -78.741 53.600
85738 93.596 -78.739 53.592
85739 93.606 -78.736 53.585
85740 93.615 -78.733 53.578
85741 93.625 -78.731 53.571
85742 93.626 -78.724 53.565
85743 93.625 -78.714 53.563
85744 93.623 -78.705 53.561
85745 93.622 -78.695 53.562
85746 93.620 -78.721 53.565
85747 93.630 -78.707 53.563
85748 93.644 -78.692 53.560
85749 93.661 -78.682 53.554
85750 93.679 -78.671 53.545
85751 93.697 -78.661 53.534
85752 93.714 -78.651 53.522
85753 93.731 -78.640 53.508
85754 93.750 -78.627 53.491
85755 93.769 -78.611 53.474
85756 93.775 -78.600 53.454
85757 93.778 -78.589 53.434
85758 93.781 -78.576 53.414
85759 93.784 -78.561 53.393
85760 93.786 -78.547 53.372
85761 93.789 -78.532 53.350
85762 93.792 -78.519 53.326
85763 93.794 -78.504 53.302
85764 93.797 -78.490 53.277
85765 93.800 -78.476 53.252
85766 93.815 -78.485 53.227
85767 93.840 -78.510 53.203
85768 93.853 -78.497 53.180
85769 93.865 -78.485 53.158
85770 93.902 -78.468 53.135
85771 93.945 -78.451 53.116
85772 93.989 -78.433 53.098
85773 94.000 -78.438 53.081
85774 94.001 -78.448 53.066
85775 94.002 -78.458 53.051
85776 94.004 -78.469 53.033
85777 94.015 -78.471 53.012
85778 94.030 -78.470 52.989
85779 94.045 -78.468 52.966
85780 94.059 -78.467 52.943
85781 94.074 -78.465 52.922
85782 94.088 -78.463 52.900
85783 94.103 -78.462 52.876
85784 94.118 -78.460 52.850
85785 94.132 -78.459 52.827
85786 94.147 -78.457 52.808
85787 94.161 -78.456 52.790
85788 94.173 -78.448 52.772
85789 94.189 -78.415 52.755
85790 94.208 -78.379 52.737
85791 94.220 -78.357 52.718
85792 94.222 -78.351 52.701
85793 94.225 -78.344 52.683
85794 94.227 -78.384 52.665
85795 94.228 -78.445 52.646
85796 94.229 -78.507 52.629
85797 94.236 -78.567 52.612
85798 94.251 -78.574 52.594
85799 94.266 -78.580 52.576
85800 94.281 -78.587 52.558
85801 94.296 -78.594 52.540
85802 94.312 -78.599 52.522
85803 94.328 -78.605 52.505
85804 94.345 -78.611 52.486
85805 94.361 -78.617 52.464
85806 94.381 -78.611 52.442
85807 94.399 -78.596 52.419
85808 94.416 -78.582 52.395
85809 94.433 -78.567 52.370
85810 94.451 -78.552 52.342
85811 94.464 -78.545 52.312
85812 94.473 -78.545 52.279
85813 94.483 -78.546 52.245
85814 94.492 -78.546 52.211
85815 94.501 -78.546 52.176
85816 94.510 -78.546 52.141
85817 94.526 -78.551 52.108
85818 94.542 -78.555 52.075
85819 94.557 -78.559 52.042
85820 94.573 -78.564 52.010
85821 94.585 -78.567 51.976
85822 94.591 -78.569 51.941
85823 94.602 -78.567 51.904
85824 94.612 -78.567 51.866
85825 94.622 -78.565 51.826
85826 94.633 -78.564 51.784
85827 94.647 -78.568 51.739
85828 94.660 -78.572 51.693
85829 94.674 -78.576 51.644
85830 94.687 -78.579 51.594
85831 94.700 -78.582 51.545
85832 94.713 -78.586 51.496
85833 94.726 -78.588 51.446
85834 94.740 -78.589 51.396
85835 94.761 -78.594 51.347
85836 94.782 -78.598 51.299
85837 94.803 -78.602 51.251
85838 94.824 -78.607 51.204
85839 94.845 -78.611 51.157
85840 94.867 -78.615 51.110
85841 94.888 -78.619 51.065
85842 94.904 -78.622 51.020
85843 94.901 -78.630 50.976
85844 94.906 -78.633 50.931
85845 94.910 -78.634 50.888
85846 94.915 -78.636 50.845
85847 94.919 -78.638 50.803
85848 94.924 -78.640 50.761
85849 94.928 -78.642 50.720
85850 94.922 -78.641 50.679
85851 94.917 -78.640 50.636
85852 94.921 -78.631 50.592
85853 94.935 -78.613 50.549
85854 94.950 -78.595 50.506
85855 94.964 -78.577 50.462
85856 94.979 -78.559 50.417
85857 94.993 -78.542 50.372
85858 95.007 -78.523 50.328
85859 95.022 -78.506 50.284
85860 95.037 -78.490 50.239
85861 95.057 -78.495 50.195
85862 95.077 -78.500 50.152
85863 95.098 -78.505 50.109
85864 95.118 -78.510 50.065
85865 95.139 -78.516 50.024
85866 95.149 -78.519 49.982
85867 95.160 -78.523 49.941
85868 95.171 -78.526 49.902
85869 95.181 -78.530 49.864
85870 95.192 -78.534 49.826
85871 95.203 -78.538 49.788
85872 95.213 -78.542 49.751
85873 95.224 -78.545 49.713
85874 95.255 -78.561 49.674
85875 95.288 -78.577 49.637
85876 95.321 -78.595 49.599
85877 95.354 -78.611 49.561
85878 95.387 -78.628 49.523
85879 95.420 -78.645 49.486
85880 95.433 -78.659 49.449
85881 95.448 -78.680 49.412
85882 95.462 -78.701 49.377
85883 95.475 -78.722 49.343
85884 95.489 -78.743 49.309
85885 95.503 -78.764 49.276
85886 95.498 -78.764 49.242
85887 95.493 -78.764 49.206
85888 95.488 -78.764 49.169
85889 95.483 -78.763 49.129
85890 95.469 -78.759 49.089
85891 95.452 -78.755 49.047
85892 95.435 -78.749 49.006
85893 95.418 -78.745 48.963
85894 95.416 -78.735 48.918
85895 95.420 -78.724 48.872
85896 95.424 -78.712 48.828
85897 95.432 -78.704 48.785
85898 95.456 -78.706 48.742
85899 95.479 -78.708 48.700
85900 95.502 -78.710 48.659
85901 95.526 -78.711 48.619
85902 95.549 -78.714 48.580
85903 95.572 -78.716 48.540
85904 95.596 -78.717 48.501
85905 95.619 -78.720 48.462
85906 95.625 -78.730 48.423
85907 95.637 -78.735 48.386
85908 95.649 -78.739 48.350
85909 95.661 -78.743 48.313
85910 95.673 -78.748 48.277
85911 95.685 -78.752 48.241
85912 95.701 -78.758 48.204
85913 95.730 -78.770 48.167
85914 95.759 -78.782 48.127
85915 95.788 -78.786 48.087
85916 95.818 -78.780 48.047
85917 95.836 -78.780 48.007
85918 95.853 -78.781 47.967
85919 95.870 -78.782 47.925
85920 95.886 -78.787 47.883
85921 95.901 -78.793 47.841
85922 95.914 -78.798 47.799
85923 95.921 -78.803 47.757
85924 95.930 -78.809 47.715
85925 95.953 -78.806 47.672
85926 95.977 -78.803 47.629
85927 96.002 -78.800 47.585
85928 96.026 -78.797 47.544
85929 96.050 -78.793 47.502
85930 96.074 -78.790 47.460
85931 96.098 -78.787 47.420
85932 96.117 -78.799 47.380
85933 96.135 -78.816 47.341
85934 96.153 -78.834 47.303
85935 96.171 -78.852 47.266
85936 96.187 -78.839 47.229
85937 96.203 -78.827 47.192
85938 96.220 -78.815 47.155
85939 96.237 -78.802 47.118
85940 96.253 -78.790 47.083
85941 96.269 -78.777 47.049
85942 96.289 -78.767 47.017
85943 96.305 -78.764 46.984
85944 96.316 -78.782 46.950
85945 96.327 -78.799 46.917
85946 96.339 -78.818 46.885
85947 96.350 -78.835 46.854
85948 96.361 -78.852 46.823
85949 96.373 -78.870 46.792
85950 96.383 -78.879 46.761
85951 96.393 -78.882 46.731
85952 96.437 -78.882 46.703
85953 96.482 -78.882 46.676
85954 96.527 -78.881 46.649
85955 96.569 -78.877 46.623
85956 96.609 -78.873 46.597
85957 96.650 -78.868 46.570
85958 96.690 -78.863 46.542
85959 96.731 -78.858 46.520
85960 96.771 -78.853 46.499
85961 96.812 -78.849 46.478
85962 96.841 -78.845 46.456
85963 96.851 -78.842 46.433
85964 96.861 -78.839 46.411
85965 96.871 -78.836 46.391
85966 96.889 -78.844 46.371
85967 96.909 -78.855 46.351
85968 96.930 -78.852 46.332
85969 96.943 -78.855 46.311
85970 96.938 -78.890 46.292
85971 96.965 -78.947 46.273
85972 96.971 -78.959 46.256
85973 96.958 -78.977 46.240
85974 96.963 -78.988 46.224
85975 96.975 -78.968 46.209
85976 96.975 -78.968 46.193
85977 96.969 -78.983 46.177
85978 96.963 -78.998 46.161
85979 96.980 -79.040 46.144
85980 97.001 -79.081 46.130
85981 97.022 -79.076 46.115
85982 97.042 -79.071 46.101
85983 97.062 -79.066 46.083
85984 97.082 -79.061 46.066
85985 97.102 -79.056 46.050
85986 97.122 -79.050 46.035
85987 97.146 -79.053 46.025
85988 97.169 -79.057 46.012
85989 97.193 -79.062 45.996
85990 97.216 -79.066 45.982
85991 97.232 -79.078 45.965
85992 97.246 -79.087 45.947
85993 97.251 -79.090 45.931
85994 97.257 -79.093 45.915
85995 97.264 -79.097 45.897
85996 97.280 -79.106 45.877
85997 97.295 -79.115 45.856
85998 97.310 -79.124 45.836
85999 97.325 -79.132 45.817
86000 97.340 -79.141 45.798
86001 97.355 -79.150 45.781
86002 97.370 -79.159 45.763
86003 97.383 -79.162 45.745
86004 97.394 -79.161 45.727
86005 97.406 -79.160 45.709
86006 97.417 -79.159 45.691
86007 97.429 -79.159 45.675
86008 97.440 -79.158 45.659
86009 97.454 -79.160 45.644
86010 97.476 -79.172 45.628
86011 97.499 -79.182 45.612
86012 97.521 -79.194 45.595
86013 97.543 -79.204 45.578
86014 97.566 -79.215 45.562
86015 97.588 -79.226 45.546
86016 97.607 -79.234 45.531
86017 97.623 -79.238 45.516
86018 97.623 -79.238 45.501
86019 97.623 -79.237 45.486
86020 97.622 -79.237 45.471
86021 97.622 -79.237 45.456
86022 97.622 -79.236 45.444
86023 97.622 -79.236 45.431
86024 97.621 -79.235 45.418
86025 97.621 -79.235 45.402
86026 97.621 -79.234 45.386
86027 97.621 -79.234 45.370
86028 97.626 -79.239 45.355
86029 97.644 -79.259 45.343
86030 97.662 -79.279 45.332
86031 97.680 -79.300 45.320
86032 97.699 -79.320 45.308
86033 97.717 -79.341 45.296
86034 97.735 -79.361 45.284
86035 97.759 -79.382 45.272
86036 97.799 -79.406 45.260
86037 97.823 -79.419 45.249
86038 97.840 -79.427 45.239
86039 97.856 -79.435 45.229
86040 97.869 -79.451 45.219
86041 97.881 -79.466 45.209
86042 97.893 -79.481 45.200
86043 97.906 -79.496 45.191
86044 97.918 -79.511 45.181
86045 97.931 -79.526 45.171
86046 97.943 -79.541 45.163
86047 97.955 -79.557 45.155
86048 97.975 -79.569 45.145
86049 97.999 -79.576 45.136
86050 98.024 -79.584 45.128
86051 98.055 -79.583 45.120
86052 98.108 -79.555 45.113
86053 98.161 -79.527 45.106
86054 98.201 -79.514 45.099
86055 98.238 -79.505 45.091
86056 98.271 -79.493 45.082
86057 98.303 -79.510 45.073
86058 98.339 -79.535 45.066
86059 98.374 -79.560 45.057
86060 98.414 -79.591 45.049
86061 98.444 -79.620 45.041
86062 98.476 -79.632 45.033
86063 98.502 -79.641 45.024
86064 98.527 -79.650 45.016
86065 98.553 -79.659 45.007
86066 98.578 -79.668 44.999
86067 98.604 -79.677 44.991
86068 98.629 -79.686 44.985
86069 98.655 -79.695 44.978
86070 98.680 -79.704 44.972
86071 98.691 -79.723 44.968
86072 98.701 -79.723 44.964
86073 98.718 -79.714 44.959
86074 98.735 -79.705 44.956
86075 98.752 -79.696 44.952
86076 98.770 -79.688 44.949
86077 98.787 -79.679 44.946
86078 98.803 -79.670 44.942
86079 98.821 -79.670 44.938
86080 98.840 -79.681 44.933
86081 98.860 -79.692 44.927
86082 98.879 -79.704 44.921
86083 98.898 -79.715 44.913
86084 98.917 -79.726 44.905
86085 98.936 -79.739 44.897
86086 98.955 -79.750 44.888
86087 98.975 -79.761 44.878
86088 99.013 -79.764 44.868
86089 99.055 -79.764 44.860
86090 99.094 -79.766 44.851
86091 99.119 -79.780 44.840
86092 99.143 -79.794 44.828
86093 99.164 -79.808 44.815
86094 99.168 -79.822 44.803
86095 99.180 -79.833 44.792
86096 99.204 -79.839 44.781
86097 99.227 -79.845 44.773
86098 99.251 -79.852 44.764
86099 99.275 -79.858 44.755
86100 99.298 -79.864 44.748
86101 99.322 -79.871 44.740
86102 99.346 -79.877 44.732
86103 99.369 -79.883 44.723
86104 99.394 -79.889 44.713
86105 99.417 -79.896 44.706
86106 99.441 -79.902 44.697
86107 99.465 -79.908 44.688
86108 99.488 -79.914 44.681
86109 99.512 -79.921 44.673
86110 99.531 -79.921 44.663
86111 99.546 -79.914 44.655
86112 99.561 -79.908 44.647
86113 99.570 -79.916 44.638
86114 99.568 -79.944 44.628
86115 99.567 -79.972 44.620
86116 99.566 -80.000 44.612
86117 99.565 -80.028 44.603
86118 99.564 -80.056 44.595
86119 99.562 -80.084 44.587
86120 99.561 -80.112 44.580
86121 99.560 -80.140 44.574
86122 99.559 -80.168 44.566
86123 99.557 -80.196 44.559
86124 99.556 -80.224 44.552
86125 99.555 -80.252 44.544
86126 99.554 -80.280 44.537
86127 99.552 -80.308 44.529
86128 99.567 -80.337 44.521
86129 99.619 -80.349 44.511
86130 99.669 -80.362 44.501
86131 99.721 -80.375 44.491
86132 99.772 -80.387 44.481
86133 99.824 -80.400 44.470
86134 99.875 -80.413 44.458
86135 99.926 -80.425 44.446
86136 99.978 -80.438 44.434
86137 100.025 -80.453 44.422
86138 100.069 -80.469 44.409
86139 100.114 -80.486 44.396
86140 100.158 -80.502 44.384
86141 100.203 -80.519 44.371
86142 100.248 -80.535 44.358
86436 93.567 -78.735 53.624
86437 93.577 -78.733 53.615
86438 93.586 -78.730 53.606
86439 93.596 -78.728 53.597
86440 93.606 -78.725 53.590
86441 93.616 -78.723 53.582
86442 93.624 -78.720 53.575
86443 93.623 -78.711 53.569
86444 93.621 -78.701 53.565
86445 93.620 -78.692 53.561
86446 93.619 -78.683 53.560
86447 93.617 -78.688 53.564
86448 93.625 -78.681 53.562
86449 93.642 -78.670 53.557
86450 93.659 -78.660 53.550
86451 93.677 -78.649 53.541
86452 93.695 -78.639 53.532
86453 93.712 -78.628 53.520
86454 93.729 -78.617 53.505
86455 93.747 -78.607 53.488
86456 93.765 -78.596 53.469
86457 93.770 -78.582 53.449
86458 93.772 -78.568 53.427
86459 93.775 -78.554 53.406
86460 93.778 -78.540 53.384
86461 93.780 -78.526 53.363
86462 93.783 -78.512 53.341
86463 93.786 -78.498 53.318
86464 93.788 -78.483 53.295
86465 93.791 -78.469 53.270
86466 93.794 -78.455 53.245
86467 93.803 -78.453 53.221
86468 93.830 -78.485 53.198
86469 93.848 -78.486 53.175
86470 93.870 -78.472 53.154
86471 93.914 -78.454 53.132
86472 93.958 -78.437 53.115
86473 94.002 -78.420 53.098
86474 94.010 -78.426 53.081
86475 94.011 -78.436 53.065
86476 94.012 -78.447 53.050
86477 94.008 -78.459 53.032
86478 94.018 -78.463 53.010
86479 94.034 -78.460 52.986
86480 94.049 -78.457 52.963
86481 94.065 -78.454 52.940
86482 94.080 -78.451 52.918
86483 94.096 -78.448 52.896
86484 94.112 -78.445 52.873
86485 94.127 -78.443 52.849
86486 94.142 -78.440 52.829
86487 94.157 -78.438 52.811
86488 94.172 -78.437 52.795
86489 94.186 -78.435 52.778
86490 94.202 -78.416 52.760
86491 94.207 -78.404 52.741
86492 94.209 -78.398 52.721
86493 94.212 -78.391 52.703
86494 94.215 -78.385 52.685
86495 94.217 -78.408 52.668
86496 94.218 -78.470 52.649
86497 94.219 -78.532 52.633
86498 94.231 -78.557 52.617
86499 94.247 -78.562 52.599
86500 94.263 -78.568 52.582
86501 94.280 -78.573 52.564
86502 94.296 -78.579 52.546
86503 94.312 -78.586 52.529
86504 94.329 -78.591 52.511
86505 94.345 -78.597 52.493
86506 94.362 -78.600 52.471
86507 94.386 -78.593 52.450
86508 94.408 -78.587 52.428
86509 94.429 -78.578 52.404
86510 94.447 -78.564 52.379
86511 94.464 -78.548 52.351
86512 94.476 -78.545 52.321
86513 94.485 -78.545 52.289
86514 94.494 -78.546 52.255
86515 94.503 -78.546 52.220
86516 94.513 -78.546 52.185
86517 94.522 -78.546 52.151
86518 94.532 -78.547 52.119
86519 94.544 -78.549 52.086
86520 94.560 -78.553 52.051
86521 94.575 -78.557 52.017
86522 94.590 -78.561 51.983
86523 94.597 -78.564 51.948
86524 94.607 -78.562 51.912
86525 94.617 -78.561 51.874
86526 94.628 -78.560 51.835
86527 94.638 -78.558 51.793
86528 94.649 -78.558 51.748
86529 94.663 -78.562 51.702
86530 94.676 -78.566 51.653
86531 94.690 -78.569 51.602
86532 94.703 -78.573 51.552
86533 94.716 -78.576 51.502
86534 94.730 -78.580 51.451
86535 94.743 -78.583 51.401
86536 94.764 -78.587 51.351
86537 94.784 -78.590 51.301
86538 94.806 -78.594 51.252
86539 94.827 -78.597 51.204
86540 94.847 -78.601 51.155
86541 94.868 -78.604 51.108
86542 94.889 -78.608 51.062
86543 94.893 -78.611 51.017
86544 94.896 -78.616 50.972
86545 94.890 -78.622 50.928
86546 94.886 -78.634 50.884
86547 94.881 -78.646 50.842
86548 94.886 -78.648 50.800
86549 94.890 -78.650 50.758
86550 94.887 -78.650 50.717
86551 94.898 -78.636 50.677
86552 94.912 -78.617 50.635
86553 94.926 -78.600 50.592
86554 94.941 -78.582 50.549
86555 94.955 -78.564 50.506
86556 94.969 -78.546 50.463
86557 94.984 -78.528 50.419
86558 94.998 -78.510 50.375
86559 95.013 -78.492 50.330
86560 95.027 -78.475 50.285
86561 95.045 -78.470 50.240
86562 95.065 -78.475 50.196
86563 95.086 -78.480 50.153
86564 95.106 -78.485 50.110
86565 95.126 -78.490 50.066
86566 95.147 -78.495 50.024
86567 95.163 -78.500 49.982
86568 95.174 -78.504 49.941
86569 95.185 -78.507 49.903
86570 95.195 -78.511 49.864
86571 95.206 -78.515 49.826
86572 95.217 -78.518 49.787
86573 95.228 -78.522 49.750
86574 95.249 -78.532 49.712
86575 95.282 -78.549 49.675
86576 95.315 -78.566 49.638
86577 95.348 -78.582 49.600
86578 95.381 -78.600 49.562
86579 95.414 -78.617 49.524
86580 95.447 -78.633 49.488
86581 95.480 -78.650 49.452
86582 95.491 -78.667 49.416
86583 95.505 -78.688 49.380
86584 95.519 -78.709 49.346
86585 95.533 -78.730 49.312
86586 95.532 -78.734 49.280
86587 95.527 -78.733 49.247
86588 95.522 -78.733 49.213
86589 95.516 -78.733 49.176
86590 95.511 -78.732 49.136
86591 95.506 -78.732 49.096
86592 95.489 -78.727 49.054
86593 95.472 -78.722 49.012
86594 95.455 -78.717 48.969
86595 95.438 -78.712 48.923
86596 95.421 -78.708 48.877
86597 95.415 -78.699 48.834
86598 95.422 -78.690 48.792
86599 95.445 -78.692 48.749
86600 95.469 -78.694 48.709
86601 95.492 -78.696 48.668
86602 95.516 -78.698 48.629
86603 95.539 -78.700 48.589
86604 95.563 -78.702 48.550
86605 95.588 -78.702 48.511
86606 95.613 -78.702 48.472
86607 95.632 -78.705 48.435
86608 95.644 -78.710 48.397
86609 95.657 -78.714 48.361
86610 95.669 -78.718 48.325
86611 95.681 -78.723 48.289
86612 95.692 -78.727 48.251
86613 95.707 -78.732 48.212
86614 95.736 -78.744 48.172
86615 95.765 -78.755 48.132
86616 95.794 -78.748 48.091
86617 95.824 -78.742 48.051
86618 95.848 -78.738 48.011
86619 95.866 -78.739 47.971
86620 95.881 -78.743 47.931
86621 95.896 -78.749 47.889
86622 95.911 -78.753 47.848
86623 95.924 -78.759 47.807
86624 95.932 -78.764 47.766
86625 95.940 -78.771 47.725
86626 95.956 -78.772 47.684
86627 95.980 -78.769 47.642
86628 96.004 -78.765 47.599
86629 96.028 -78.762 47.557
86630 96.052 -78.759 47.516
86631 96.077 -78.755 47.474
86632 96.099 -78.759 47.434
86633 96.116 -78.777 47.393
86634 96.134 -78.794 47.354
86635 96.152 -78.811 47.316
86636 96.170 -78.829 47.279
86637 96.186 -78.829 47.241
86638 96.203 -78.817 47.204
86639 96.220 -78.805 47.167
86640 96.236 -78.792 47.130
86641 96.252 -78.780 47.094
86642 96.269 -78.767 47.061
86643 96.286 -78.755 47.028
86644 96.297 -78.759 46.995
86645 96.307 -78.769 46.962
86646 96.318 -78.787 46.930
86647 96.329 -78.805 46.898
86648 96.340 -78.822 46.866
86649 96.352 -78.840 46.836
86650 96.363 -78.858 46.805
86651 96.374 -78.871 46.775
86652 96.385 -78.874 46.745
86653 96.431 -78.874 46.717
86654 96.476 -78.874 46.690
86655 96.518 -78.871 46.665
86656 96.559 -78.865 46.638
86657 96.599 -78.861 46.612
86658 96.640 -78.856 46.585
86659 96.680 -78.851 46.557
86660 96.721 -78.846 46.534
86661 96.761 -78.841 46.512
86662 96.802 -78.836 46.490
86663 96.837 -78.832 46.469
86664 96.848 -78.829 46.448
86665 96.858 -78.827 46.426
86666 96.869 -78.824 46.405
86667 96.889 -78.835 46.386
86668 96.909 -78.846 46.367
86669 96.930 -78.847 46.349
86670 96.945 -78.849 46.328
86671 96.943 -78.886 46.308
86672 96.966 -78.935 46.289
86673 96.971 -78.941 46.272
86674 96.958 -78.959 46.257
86675 96.952 -78.973 46.241
86676 96.960 -78.984 46.226
86677 96.972 -78.961 46.209
86678 96.966 -78.976 46.194
86679 96.959 -78.990 46.178
86680 96.978 -79.034 46.161
86681 96.999 -79.052 46.146
86682 97.019 -79.046 46.132
86683 97.039 -79.041 46.117
86684 97.060 -79.036 46.100
86685 97.080 -79.031 46.082
86686 97.102 -79.033 46.064
86687 97.126 -79.037 46.049
86688 97.150 -79.042 46.037
86689 97.173 -79.046 46.024
86690 97.197 -79.051 46.009
86691 97.219 -79.057 45.995
86692 97.234 -79.068 45.978
86693 97.250 -79.079 45.961
86694 97.260 -79.086 45.943
86695 97.266 -79.089 45.924
86696 97.272 -79.092 45.905
86697 97.287 -79.101 45.885
86698 97.302 -79.109 45.866
86699 97.317 -79.118 45.846
86700 97.332 -79.127 45.827
86701 97.348 -79.136 45.808
86702 97.363 -79.145 45.791
86703 97.375 -79.147 45.773
86704 97.386 -79.146 45.756
86705 97.398 -79.145 45.738
86706 97.409 -79.144 45.720
86707 97.421 -79.143 45.703
86708 97.432 -79.143 45.686
86709 97.443 -79.141 45.671
86710 97.455 -79.141 45.655
86711 97.477 -79.153 45.639
86712 97.500 -79.163 45.624
86713 97.522 -79.175 45.608
86714 97.544 -79.185 45.591
86715 97.566 -79.196 45.576
86716 97.588 -79.207 45.560
86717 97.610 -79.219 45.544
86718 97.614 -79.225 45.528
86719 97.613 -79.224 45.514
86720 97.613 -79.224 45.499
86721 97.613 -79.223 45.483
86722 97.613 -79.223 45.468
86723 97.613 -79.222 45.455
86724 97.613 -79.222 45.443
86725 97.613 -79.222 45.430
86726 97.613 -79.221 45.413
86727 97.612 -79.221 45.396
86728 97.612 -79.220 45.379
86729 97.618 -79.226 45.364
86730 97.636 -79.247 45.351
86731 97.655 -79.266 45.339
86732 97.673 -79.287 45.327
86733 97.691 -79.307 45.316
86734 97.709 -79.328 45.305
86735 97.731 -79.349 45.293
86736 97.771 -79.372 45.281
86737 97.806 -79.393 45.269
86738 97.823 -79.401 45.258
86739 97.840 -79.410 45.248
86740 97.856 -79.418 45.239
86741 97.870 -79.432 45.229
86742 97.882 -79.447 45.220
86743 97.894 -79.462 45.211
86744 97.907 -79.477 45.202
86745 97.919 -79.492 45.193
86746 97.932 -79.507 45.182
86747 97.944 -79.523 45.174
86748 97.959 -79.537 45.166
86749 97.982 -79.547 45.158
86750 98.006 -79.554 45.149
86751 98.031 -79.562 45.142
86752 98.056 -79.569 45.135
86753 98.080 -79.576 45.128
86754 98.114 -79.572 45.120
86755 98.167 -79.544 45.111
86756 98.220 -79.516 45.103
86757 98.283 -79.508 45.095
86758 98.329 -79.528 45.087
86759 98.354 -79.551 45.078
86760 98.379 -79.574 45.069
86761 98.406 -79.595 45.060
86762 98.439 -79.604 45.051
86763 98.472 -79.613 45.042
86764 98.505 -79.621 45.033
86765 98.533 -79.630 45.025
86766 98.558 -79.639 45.018
86767 98.584 -79.648 45.010
86768 98.609 -79.657 45.003
86769 98.634 -79.667 44.997
86770 98.660 -79.676 44.990
86771 98.685 -79.686 44.984
86772 98.692 -79.705 44.980
86773 98.696 -79.716 44.975
86774 98.713 -79.708 44.972
86775 98.730 -79.698 44.968
86776 98.747 -79.690 44.964
86777 98.764 -79.681 44.961
86778 98.781 -79.673 44.957
86779 98.798 -79.665 44.953
86780 98.814 -79.665 44.949
86781 98.833 -79.676 44.944
86782 98.852 -79.689 44.938
86783 98.872 -79.700 44.931
86784 98.891 -79.711 44.923
86785 98.910 -79.723 44.914
86786 98.929 -79.734 44.905
86787 98.949 -79.745 44.895
86788 98.990 -79.747 44.885
86789 99.031 -79.748 44.874
86790 99.072 -79.748 44.864
86791 99.113 -79.750 44.856
86792 99.137 -79.764 44.844
86793 99.161 -79.777 44.831
86794 99.175 -79.792 44.819
86795 99.179 -79.806 44.806
86796 99.184 -79.820 44.795
86797 99.207 -79.826 44.784
86798 99.231 -79.832 44.776
86799 99.255 -79.839 44.768
86800 99.279 -79.845 44.760
86801 99.302 -79.851 44.752
86802 99.326 -79.857 44.745
86803 99.350 -79.864 44.736
86804 99.373 -79.870 44.727
86805 99.397 -79.876 44.718
86806 99.421 -79.882 44.709
86807 99.445 -79.889 44.701
86808 99.469 -79.895 44.692
86809 99.492 -79.901 44.684
86810 99.512 -79.902 44.676
86811 99.527 -79.897 44.667
86812 99.543 -79.891 44.659
86813 99.558 -79.885 44.651
86814 99.565 -79.896 44.642
86815 99.564 -79.924 44.632
86816 99.563 -79.952 44.624
86817 99.561 -79.980 44.616
86818 99.560 -80.008 44.607
86819 99.559 -80.036 44.599
86820 99.558 -80.064 44.591
86821 99.556 -80.092 44.584
86822 99.555 -80.120 44.577
86823 99.554 -80.148 44.570
86824 99.553 -80.175 44.562
86825 99.551 -80.203 44.555
86826 99.550 -80.231 44.548
86827 99.549 -80.259 44.541
86828 99.565 -80.284 44.533
86829 99.590 -80.314 44.524
86830 99.628 -80.355 44.515
86831 99.685 -80.349 44.505
86832 99.736 -80.362 44.495
86833 99.787 -80.375 44.485
86834 99.839 -80.388 44.474
86835 99.890 -80.400 44.463
86836 99.938 -80.415 44.451
86837 99.982 -80.431 44.439
86838 100.027 -80.447 44.427
86839 100.071 -80.464 44.414
86840 100.116 -80.480 44.401
86841 100.160 -80.497 44.389
86842 100.205 -80.513 44.375
86843 100.250 -80.529 44.362
87137 93.567 -78.724 53.631
87138 93.577 -78.722 53.621
87139 93.586 -78.719 53.611
87140 93.596 -78.717 53.602
87141 93.606 -78.714 53.594
87142 93.616 -78.712 53.586
87143 93.620 -78.706 53.579
87144 93.619 -78.697 53.571
87145 93.618 -78.688 53.566
87146 93.617 -78.679 53.561
87147 93.615 -78.670 53.559
87148 93.614 -78.661 53.560
87149 93.623 -78.658 53.557
87150 93.640 -78.648 53.553
87151 93.658 -78.637 53.545
87152 93.675 -78.626 53.536
87153 93.694 -78.616 53.524
87154 93.713 -78.606 53.510
87155 93.733 -78.595 53.495
87156 93.753 -78.585 53.476
87157 93.767 -78.579 53.456
87158 93.775 -78.570 53.434
87159 93.774 -78.553 53.412
87160 93.772 -78.535 53.390
87161 93.771 -78.519 53.368
87162 93.774 -78.504 53.348
87163 93.777 -78.491 53.328
87164 93.779 -78.476 53.307
87165 93.782 -78.462 53.286
87166 93.785 -78.448 53.262
87167 93.787 -78.434 53.238
87168 93.791 -78.420 53.215
87169 93.818 -78.452 53.192
87170 93.843 -78.476 53.171
87171 93.883 -78.458 53.150
87172 93.926 -78.441 53.129
87173 93.970 -78.423 53.113
87174 94.014 -78.406 53.097
87175 94.019 -78.413 53.081
87176 94.018 -78.426 53.063
87177 94.014 -78.438 53.046
87178 94.010 -78.451 53.028
87179 94.019 -78.458 53.006
87180 94.033 -78.458 52.981
87181 94.049 -78.455 52.958
87182 94.064 -78.452 52.935
87183 94.080 -78.449 52.913
87184 94.096 -78.446 52.892
87185 94.111 -78.443 52.870
87186 94.127 -78.440 52.847
87187 94.142 -78.437 52.830
87188 94.158 -78.434 52.814
87189 94.174 -78.431 52.798
87190 94.189 -78.428 52.782
87191 94.196 -78.433 52.764
87192 94.197 -78.445 52.745
87193 94.199 -78.438 52.725
87194 94.202 -78.432 52.706
87195 94.204 -78.425 52.689
87196 94.206 -78.432 52.673
87197 94.208 -78.495 52.655
87198 94.214 -78.537 52.639
87199 94.231 -78.542 52.623
87200 94.248 -78.548 52.605
87201 94.264 -78.554 52.587
87202 94.280 -78.560 52.570
87203 94.296 -78.565 52.552
87204 94.313 -78.571 52.535
87205 94.329 -78.577 52.518
87206 94.345 -78.582 52.498
87207 94.365 -78.581 52.477
87208 94.388 -78.574 52.456
87209 94.411 -78.567 52.434
87210 94.434 -78.561 52.412
87211 94.457 -78.554 52.386
87212 94.478 -78.545 52.359
87213 94.487 -78.545 52.330
87214 94.496 -78.545 52.297
87215 94.506 -78.546 52.263
87216 94.515 -78.546 52.229
87217 94.525 -78.546 52.194
87218 94.534 -78.546 52.162
87219 94.543 -78.547 52.129
87220 94.552 -78.547 52.096
87221 94.562 -78.547 52.060
87222 94.578 -78.551 52.025
87223 94.594 -78.556 51.990
87224 94.602 -78.558 51.954
87225 94.613 -78.557 51.919
87226 94.623 -78.555 51.882
87227 94.633 -78.554 51.845
87228 94.643 -78.553 51.803
87229 94.654 -78.551 51.758
87230 94.666 -78.552 51.710
87231 94.679 -78.555 51.663
87232 94.692 -78.559 51.612
87233 94.705 -78.563 51.560
87234 94.719 -78.567 51.509
87235 94.733 -78.570 51.458
87236 94.747 -78.572 51.407
87237 94.765 -78.576 51.355
87238 94.786 -78.579 51.304
87239 94.807 -78.582 51.254
87240 94.828 -78.586 51.204
87241 94.849 -78.589 51.155
87242 94.870 -78.592 51.107
87243 94.879 -78.596 51.061
87244 94.882 -78.601 51.015
87245 94.885 -78.604 50.970
87246 94.888 -78.609 50.926
87247 94.880 -78.612 50.883
87248 94.875 -78.624 50.841
87249 94.871 -78.636 50.799
87250 94.874 -78.640 50.757
87251 94.888 -78.622 50.716
87252 94.903 -78.604 50.676
87253 94.917 -78.586 50.634
87254 94.931 -78.569 50.591
87255 94.946 -78.551 50.549
87256 94.960 -78.532 50.505
87257 94.974 -78.515 50.463
87258 94.989 -78.497 50.419
87259 95.003 -78.479 50.375
87260 95.018 -78.461 50.331
87261 95.032 -78.445 50.286
87262 95.053 -78.450 50.242
87263 95.073 -78.454 50.198
87264 95.094 -78.460 50.155
87265 95.114 -78.465 50.111
87266 95.134 -78.470 50.068
87267 95.155 -78.476 50.025
87268 95.175 -78.480 49.983
87269 95.188 -78.485 49.942
87270 95.199 -78.489 49.904
87271 95.209 -78.492 49.865
87272 95.220 -78.495 49.827
87273 95.231 -78.499 49.788
87274 95.243 -78.504 49.750
87275 95.276 -78.520 49.713
87276 95.309 -78.538 49.676
87277 95.342 -78.554 49.639
87278 95.375 -78.571 49.602
87279 95.408 -78.588 49.564
87280 95.441 -78.605 49.528
87281 95.474 -78.622 49.492
87282 95.507 -78.639 49.456
87283 95.538 -78.655 49.419
87284 95.549 -78.674 49.384
87285 95.562 -78.695 49.350
87286 95.565 -78.704 49.316
87287 95.560 -78.704 49.284
87288 95.555 -78.703 49.252
87289 95.549 -78.702 49.218
87290 95.544 -78.702 49.183
87291 95.539 -78.702 49.143
87292 95.534 -78.701 49.103
87293 95.526 -78.699 49.061
87294 95.509 -78.695 49.018
87295 95.492 -78.690 48.973
87296 95.475 -78.685 48.928
87297 95.458 -78.680 48.883
87298 95.441 -78.675 48.840
87299 95.421 -78.672 48.798
87300 95.442 -78.673 48.757
87301 95.467 -78.673 48.717
87302 95.491 -78.673 48.678
87303 95.517 -78.674 48.638
87304 95.541 -78.675 48.597
87305 95.566 -78.675 48.558
87306 95.591 -78.676 48.520
87307 95.616 -78.677 48.481
87308 95.633 -78.680 48.443
87309 95.649 -78.684 48.406
87310 95.663 -78.689 48.370
87311 95.676 -78.693 48.335
87312 95.688 -78.697 48.299
87313 95.700 -78.702 48.261
87314 95.713 -78.706 48.220
87315 95.742 -78.717 48.180
87316 95.771 -78.717 48.139
87317 95.800 -78.710 48.099
87318 95.830 -78.704 48.058
87319 95.859 -78.697 48.019
87320 95.877 -78.699 47.979
87321 95.892 -78.704 47.937
87322 95.907 -78.709 47.895
87323 95.922 -78.714 47.854
87324 95.934 -78.720 47.812
87325 95.942 -78.726 47.772
87326 95.950 -78.732 47.731
87327 95.958 -78.737 47.691
87328 95.982 -78.734 47.651
87329 96.007 -78.731 47.610
87330 96.031 -78.727 47.570
87331 96.055 -78.724 47.530
87332 96.079 -78.721 47.489
87333 96.098 -78.736 47.449
87334 96.115 -78.754 47.409
87335 96.133 -78.771 47.370
87336 96.151 -78.789 47.331
87337 96.168 -78.806 47.293
87338 96.186 -78.819 47.254
87339 96.202 -78.806 47.218
87340 96.219 -78.794 47.180
87341 96.235 -78.782 47.143
87342 96.252 -78.769 47.107
87343 96.268 -78.757 47.073
87344 96.280 -78.759 47.041
87345 96.290 -78.767 47.009
87346 96.300 -78.776 46.977
87347 96.310 -78.783 46.944
87348 96.320 -78.793 46.913
87349 96.331 -78.810 46.882
87350 96.342 -78.827 46.851
87351 96.354 -78.845 46.821
87352 96.365 -78.863 46.791
87353 96.379 -78.865 46.761
87354 96.425 -78.865 46.734
87355 96.468 -78.863 46.707
87356 96.508 -78.858 46.680
87357 96.549 -78.853 46.654
87358 96.589 -78.849 46.627
87359 96.629 -78.844 46.599
87360 96.670 -78.839 46.573
87361 96.711 -78.834 46.549
87362 96.751 -78.829 46.527
87363 96.792 -78.824 46.505
87364 96.832 -78.820 46.484
87365 96.844 -78.817 46.464
87366 96.855 -78.814 46.443
87367 96.868 -78.815 46.421
87368 96.888 -78.826 46.402
87369 96.908 -78.836 46.383
87370 96.929 -78.843 46.364
87371 96.946 -78.842 46.343
87372 96.947 -78.880 46.324
87373 96.965 -78.918 46.305
87374 96.969 -78.924 46.289
87375 96.958 -78.940 46.274
87376 96.944 -78.957 46.258
87377 96.950 -78.968 46.242
87378 96.958 -78.977 46.226
87379 96.962 -78.968 46.210
87380 96.956 -78.983 46.194
87381 96.976 -79.027 46.178
87382 96.996 -79.021 46.164
87383 97.016 -79.016 46.149
87384 97.036 -79.012 46.133
87385 97.059 -79.012 46.116
87386 97.083 -79.017 46.098
87387 97.106 -79.021 46.079
87388 97.130 -79.026 46.062
87389 97.154 -79.031 46.049
87390 97.177 -79.035 46.035
87391 97.201 -79.040 46.022
87392 97.221 -79.047 46.007
87393 97.237 -79.059 45.992
87394 97.252 -79.069 45.976
87395 97.268 -79.081 45.957
87396 97.275 -79.084 45.936
87397 97.280 -79.087 45.916
87398 97.295 -79.096 45.896
87399 97.310 -79.104 45.878
87400 97.325 -79.113 45.859
87401 97.340 -79.122 45.839
87402 97.355 -79.131 45.821
87403 97.367 -79.131 45.803
87404 97.378 -79.131 45.786
87405 97.389 -79.129 45.769
87406 97.401 -79.128 45.752
87407 97.412 -79.128 45.735
87408 97.424 -79.127 45.719
87409 97.435 -79.126 45.700
87410 97.446 -79.125 45.684
87411 97.457 -79.125 45.669
87412 97.478 -79.134 45.653
87413 97.500 -79.144 45.638
87414 97.522 -79.156 45.623
87415 97.544 -79.166 45.608
87416 97.567 -79.178 45.593
87417 97.588 -79.190 45.576
87418 97.610 -79.206 45.559
87419 97.612 -79.211 45.543
87420 97.607 -79.210 45.529
87421 97.605 -79.209 45.512
87422 97.605 -79.209 45.494
87423 97.604 -79.209 45.478
87424 97.604 -79.208 45.465
87425 97.604 -79.208 45.451
87426 97.604 -79.207 45.437
87427 97.604 -79.207 45.420
87428 97.604 -79.206 45.403
87429 97.604 -79.206 45.387
87430 97.610 -79.213 45.373
87431 97.629 -79.234 45.360
87432 97.647 -79.254 45.348
87433 97.665 -79.274 45.337
87434 97.683 -79.294 45.326
87435 97.703 -79.315 45.314
87436 97.743 -79.339 45.303
87437 97.783 -79.363 45.291
87438 97.806 -79.375 45.280
87439 97.823 -79.384 45.270
87440 97.840 -79.392 45.260
87441 97.856 -79.400 45.251
87442 97.871 -79.412 45.242
87443 97.883 -79.428 45.232
87444 97.895 -79.443 45.223
87445 97.908 -79.458 45.215
87446 97.920 -79.473 45.206
87447 97.933 -79.488 45.197
87448 97.945 -79.504 45.189
87449 97.966 -79.514 45.181
87450 97.990 -79.524 45.173
87451 98.014 -79.532 45.166
87452 98.038 -79.540 45.158
87453 98.063 -79.547 45.151
87454 98.087 -79.554 45.143
87455 98.112 -79.561 45.135
87456 98.137 -79.569 45.126
87457 98.179 -79.564 45.117
87458 98.249 -79.545 45.108
87459 98.314 -79.528 45.099
87460 98.340 -79.551 45.089
87461 98.369 -79.567 45.080
87462 98.402 -79.576 45.071
87463 98.435 -79.585 45.062
87464 98.468 -79.594 45.053
87465 98.502 -79.602 45.044
87466 98.535 -79.611 45.036
87467 98.563 -79.620 45.029
87468 98.589 -79.629 45.022
87469 98.614 -79.639 45.016
87470 98.640 -79.648 45.009
87471 98.665 -79.657 45.003
87472 98.691 -79.666 44.997
87473 98.694 -79.688 44.992
87474 98.690 -79.709 44.987
87475 98.707 -79.701 44.984
87476 98.724 -79.692 44.980
87477 98.741 -79.683 44.976
87478 98.758 -79.675 44.973
87479 98.775 -79.666 44.968
87480 98.791 -79.663 44.963
87481 98.807 -79.662 44.958
87482 98.826 -79.673 44.950
87483 98.845 -79.684 44.943
87484 98.864 -79.695 44.935
87485 98.884 -79.708 44.928
87486 98.903 -79.719 44.919
87487 98.925 -79.729 44.910
87488 98.966 -79.730 44.902
87489 99.007 -79.731 44.892
87490 99.048 -79.732 44.879
87491 99.089 -79.733 44.867
87492 99.131 -79.734 44.855
87493 99.155 -79.747 44.843
87494 99.180 -79.761 44.832
87495 99.185 -79.775 44.822
87496 99.189 -79.789 44.810
87497 99.192 -79.804 44.799
87498 99.211 -79.812 44.789
87499 99.235 -79.818 44.780
87500 99.259 -79.824 44.770
87501 99.282 -79.831 44.761
87502 99.306 -79.837 44.754
87503 99.330 -79.843 44.747
87504 99.354 -79.849 44.739
87505 99.377 -79.856 44.730
87506 99.401 -79.862 44.721
87507 99.425 -79.868 44.712
87508 99.448 -79.874 44.704
87509 99.472 -79.881 44.695
87510 99.494 -79.885 44.687
87511 99.509 -79.879 44.678
87512 99.524 -79.874 44.670
87513 99.540 -79.867 44.662
87514 99.555 -79.861 44.653
87515 99.561 -79.875 44.645
87516 99.559 -79.903 44.636
87517 99.558 -79.931 44.628
87518 99.557 -79.959 44.619
87519 99.556 -79.987 44.611
87520 99.554 -80.015 44.603
87521 99.553 -80.043 44.595
87522 99.552 -80.071 44.588
87523 99.551 -80.099 44.581
87524 99.549 -80.127 44.573
87525 99.548 -80.155 44.566
87526 99.547 -80.183 44.559
87527 99.548 -80.211 44.552
87528 99.567 -80.235 44.545
87529 99.587 -80.260 44.537
87530 99.613 -80.290 44.529
87531 99.651 -80.331 44.520
87532 99.690 -80.369 44.511
87533 99.751 -80.349 44.501
87534 99.802 -80.362 44.490
87535 99.851 -80.376 44.480
87536 99.896 -80.392 44.468
87537 99.940 -80.409 44.456
87538 99.985 -80.425 44.444
87539 100.029 -80.441 44.432
87540 100.074 -80.458 44.419
87541 100.118 -80.474 44.407
87542 100.163 -80.491 44.394
87543 100.207 -80.507 44.380
87544 100.252 -80.523 44.366
87838 93.567 -78.714 53.638
87839 93.577 -78.711 53.628
87840 93.587 -78.708 53.617
87841 93.596 -78.706 53.607
87842 93.606 -78.704 53.598
87843 93.616 -78.701 53.590
87844 93.617 -78.693 53.582
87845 93.616 -78.684 53.574
87846 93.615 -78.675 53.567
87847 93.613 -78.666 53.561
87848 93.612 -78.656 53.557
87849 93.611 -78.647 53.556
87850 93.624 -78.636 53.552
87851 93.644 -78.626 53.547
87852 93.663 -78.615 53.539
87853 93.683 -78.604 53.528
87854 93.703 -78.595 53.514
87855 93.722 -78.584 53.498
87856 93.741 -78.575 53.481
87857 93.755 -78.570 53.460
87858 93.770 -78.564 53.439
87859 93.784 -78.559 53.416
87860 93.786 -78.545 53.393
87861 93.784 -78.527 53.371
87862 93.783 -78.510 53.350
87863 93.781 -78.493 53.329
87864 93.779 -78.476 53.310
87865 93.778 -78.458 53.292
87866 93.776 -78.442 53.273
87867 93.779 -78.427 53.252
87868 93.781 -78.413 53.231
87869 93.784 -78.398 53.209
87870 93.806 -78.420 53.187
87871 93.844 -78.444 53.166
87872 93.891 -78.432 53.145
87873 93.935 -78.418 53.126
87874 93.981 -78.404 53.111
87875 94.024 -78.391 53.095
87876 94.024 -78.404 53.079
87877 94.020 -78.417 53.059
87878 94.016 -78.430 53.041
87879 94.012 -78.442 53.021
87880 94.021 -78.450 52.999
87881 94.033 -78.455 52.974
87882 94.048 -78.452 52.951
87883 94.064 -78.449 52.929
87884 94.079 -78.446 52.907
87885 94.095 -78.443 52.886
87886 94.111 -78.440 52.867
87887 94.126 -78.438 52.846
87888 94.141 -78.435 52.831
87889 94.157 -78.432 52.816
87890 94.172 -78.429 52.801
87891 94.185 -78.428 52.786
87892 94.190 -78.435 52.769
87893 94.190 -78.447 52.750
87894 94.191 -78.459 52.729
87895 94.191 -78.470 52.712
87896 94.194 -78.466 52.695
87897 94.196 -78.459 52.679
87898 94.198 -78.518 52.662
87899 94.214 -78.523 52.646
87900 94.231 -78.529 52.630
87901 94.248 -78.534 52.612
87902 94.264 -78.539 52.594
87903 94.280 -78.545 52.576
87904 94.296 -78.551 52.559
87905 94.313 -78.557 52.542
87906 94.329 -78.563 52.525
87907 94.346 -78.568 52.506
87908 94.368 -78.561 52.485
87909 94.391 -78.554 52.463
87910 94.414 -78.548 52.441
87911 94.437 -78.542 52.418
87912 94.460 -78.535 52.393
87913 94.470 -78.548 52.367
87914 94.484 -78.542 52.338
87915 94.497 -78.538 52.305
87916 94.509 -78.540 52.272
87917 94.522 -78.543 52.238
87918 94.535 -78.545 52.203
87919 94.545 -78.546 52.171
87920 94.555 -78.547 52.140
87921 94.564 -78.547 52.107
87922 94.573 -78.547 52.072
87923 94.582 -78.547 52.037
87924 94.596 -78.550 52.002
87925 94.608 -78.553 51.967
87926 94.618 -78.551 51.931
87927 94.628 -78.550 51.894
87928 94.639 -78.549 51.856
87929 94.649 -78.548 51.815
87930 94.660 -78.546 51.769
87931 94.670 -78.545 51.722
87932 94.681 -78.545 51.674
87933 94.695 -78.549 51.622
87934 94.708 -78.553 51.570
87935 94.722 -78.555 51.518
87936 94.737 -78.558 51.466
87937 94.751 -78.561 51.414
87938 94.767 -78.564 51.361
87939 94.788 -78.567 51.308
87940 94.809 -78.570 51.258
87941 94.830 -78.574 51.208
87942 94.851 -78.577 51.158
87943 94.865 -78.581 51.109
87944 94.868 -78.586 51.061
87945 94.871 -78.589 51.015
87946 94.874 -78.594 50.970
87947 94.877 -78.598 50.926
87948 94.879 -78.602 50.883
87949 94.872 -78.604 50.841
87950 94.869 -78.618 50.799
87951 94.881 -78.608 50.757
87952 94.893 -78.591 50.716
87953 94.908 -78.573 50.676
87954 94.922 -78.555 50.633
87955 94.936 -78.538 50.590
87956 94.951 -78.520 50.547
87957 94.965 -78.501 50.505
87958 94.979 -78.484 50.461
87959 94.994 -78.466 50.417
87960 95.008 -78.448 50.373
87961 95.023 -78.430 50.330
87962 95.040 -78.424 50.287
87963 95.061 -78.429 50.244
87964 95.081 -78.435 50.200
87965 95.102 -78.440 50.157
87966 95.122 -78.445 50.113
87967 95.142 -78.450 50.069
87968 95.163 -78.455 50.027
87969 95.183 -78.460 49.986
87970 95.202 -78.465 49.944
87971 95.212 -78.469 49.906
87972 95.224 -78.473 49.867
87973 95.234 -78.476 49.829
87974 95.245 -78.480 49.790
87975 95.269 -78.492 49.752
87976 95.303 -78.509 49.714
87977 95.335 -78.526 49.677
87978 95.369 -78.542 49.642
87979 95.401 -78.560 49.604
87980 95.435 -78.576 49.567
87981 95.467 -78.593 49.532
87982 95.500 -78.610 49.496
87983 95.533 -78.627 49.459
87984 95.566 -78.644 49.423
87985 95.592 -78.661 49.389
87986 95.599 -78.673 49.354
87987 95.593 -78.673 49.320
87988 95.588 -78.673 49.288
87989 95.583 -78.672 49.256
87990 95.578 -78.672 49.222
87991 95.572 -78.671 49.187
87992 95.567 -78.670 49.148
87993 95.561 -78.670 49.108
87994 95.556 -78.670 49.068
87995 95.546 -78.667 49.024
87996 95.529 -78.662 48.979
87997 95.512 -78.658 48.934
87998 95.495 -78.653 48.890
87999 95.478 -78.648 48.847
88000 95.454 -78.646 48.806
88001 95.445 -78.646 48.766
88002 95.470 -78.647 48.726
88003 95.495 -78.648 48.688
88004 95.520 -78.648 48.648
88005 95.544 -78.648 48.606
88006 95.569 -78.649 48.567
88007 95.594 -78.650 48.528
88008 95.619 -78.650 48.489
88009 95.635 -78.654 48.451
88010 95.650 -78.658 48.414
88011 95.665 -78.663 48.378
88012 95.680 -78.667 48.343
88013 95.694 -78.672 48.308
88014 95.708 -78.677 48.269
88015 95.720 -78.680 48.228
88016 95.748 -78.685 48.187
88017 95.777 -78.678 48.147
88018 95.806 -78.673 48.106
88019 95.834 -78.669 48.066
88020 95.862 -78.667 48.027
88021 95.883 -78.666 47.986
88022 95.903 -78.665 47.944
88023 95.918 -78.670 47.903
88024 95.933 -78.676 47.862
88025 95.944 -78.681 47.820
88026 95.952 -78.687 47.779
88027 95.960 -78.693 47.738
88028 95.967 -78.699 47.698
88029 95.985 -78.699 47.659
88030 96.010 -78.696 47.619
88031 96.034 -78.693 47.580
88032 96.058 -78.689 47.541
88033 96.079 -78.696 47.500
88034 96.097 -78.714 47.462
88035 96.114 -78.732 47.422
88036 96.132 -78.749 47.383
88037 96.150 -78.767 47.344
88038 96.167 -78.784 47.305
88039 96.185 -78.802 47.267
88040 96.202 -78.796 47.231
88041 96.218 -78.783 47.193
88042 96.235 -78.771 47.156
88043 96.251 -78.758 47.119
88044 96.264 -78.759 47.085
88045 96.273 -78.767 47.053
88046 96.284 -78.775 47.022
88047 96.294 -78.783 46.992
88048 96.304 -78.792 46.960
88049 96.313 -78.800 46.929
88050 96.323 -78.808 46.899
88051 96.333 -78.817 46.868
88052 96.344 -78.833 46.838
88053 96.356 -78.850 46.808
88054 96.374 -78.858 46.779
88055 96.417 -78.855 46.752
88056 96.458 -78.851 46.724
88057 96.498 -78.846 46.695
88058 96.538 -78.841 46.670
88059 96.579 -78.836 46.642
88060 96.619 -78.831 46.614
88061 96.660 -78.827 46.589
88062 96.700 -78.822 46.565
88063 96.741 -78.817 46.543
88064 96.782 -78.812 46.522
88065 96.822 -78.808 46.501
88066 96.841 -78.804 46.481
88067 96.852 -78.802 46.460
88068 96.868 -78.807 46.438
88069 96.888 -78.817 46.419
88070 96.908 -78.827 46.399
88071 96.928 -78.838 46.379
88072 96.948 -78.835 46.359
88073 96.952 -78.876 46.340
88074 96.964 -78.899 46.322
88075 96.968 -78.906 46.305
88076 96.958 -78.921 46.290
88077 96.944 -78.938 46.275
88078 96.939 -78.953 46.259
88079 96.947 -78.964 46.243
88080 96.955 -78.968 46.227
88081 96.953 -78.975 46.210
88082 96.973 -78.997 46.195
88083 96.994 -78.992 46.181
88084 97.016 -78.993 46.165
88085 97.039 -78.997 46.148
88086 97.063 -79.002 46.131
88087 97.087 -79.006 46.113
88088 97.110 -79.011 46.094
88089 97.134 -79.015 46.076
88090 97.158 -79.020 46.061
88091 97.181 -79.024 46.047
88092 97.205 -79.029 46.033
88093 97.223 -79.038 46.021
88094 97.239 -79.049 46.007
88095 97.254 -79.060 45.991
88096 97.271 -79.071 45.973
88097 97.283 -79.080 45.953
88098 97.289 -79.083 45.933
88099 97.302 -79.090 45.911
88100 97.317 -79.100 45.893
88101 97.332 -79.108 45.874
88102 97.348 -79.116 45.855
88103 97.359 -79.115 45.837
88104 97.370 -79.115 45.820
88105 97.381 -79.114 45.803
88106 97.392 -79.113 45.787
88107 97.404 -79.112 45.770
88108 97.415 -79.112 45.754
88109 97.427 -79.111 45.737
88110 97.438 -79.109 45.720
88111 97.449 -79.109 45.704
88112 97.460 -79.108 45.690
88113 97.478 -79.115 45.675
88114 97.501 -79.125 45.660
88115 97.523 -79.137 45.645
88116 97.544 -79.150 45.630
88117 97.566 -79.162 45.613
88118 97.587 -79.176 45.594
88119 97.611 -79.194 45.574
88120 97.610 -79.197 45.555
88121 97.605 -79.197 45.540
88122 97.600 -79.196 45.523
88123 97.596 -79.195 45.506
88124 97.596 -79.195 45.490
88125 97.596 -79.194 45.474
88126 97.595 -79.194 45.459
88127 97.595 -79.194 45.444
88128 97.595 -79.193 45.429
88129 97.595 -79.193 45.414
88130 97.595 -79.192 45.399
88131 97.602 -79.200 45.383
88132 97.620 -79.221 45.370
88133 97.639 -79.241 45.358
88134 97.657 -79.261 45.347
88135 97.675 -79.281 45.336
88136 97.714 -79.305 45.325
88137 97.754 -79.328 45.314
88138 97.789 -79.349 45.302
88139 97.806 -79.358 45.292
88140 97.823 -79.366 45.283
88141 97.840 -79.375 45.273
88142 97.856 -79.383 45.264
88143 97.872 -79.393 45.254
88144 97.884 -79.409 45.245
88145 97.896 -79.424 45.236
88146 97.909 -79.439 45.228
88147 97.921 -79.454 45.220
88148 97.934 -79.469 45.213
88149 97.951 -79.482 45.205
88150 97.973 -79.492 45.197
88151 97.997 -79.502 45.190
88152 98.021 -79.510 45.183
88153 98.045 -79.518 45.175
88154 98.070 -79.525 45.167
88155 98.095 -79.532 45.158
88156 98.119 -79.539 45.150
88157 98.144 -79.547 45.141
88158 98.183 -79.543 45.131
88159 98.238 -79.538 45.120
88160 98.293 -79.533 45.110
88161 98.332 -79.539 45.100
88162 98.365 -79.548 45.092
88163 98.398 -79.557 45.083
88164 98.432 -79.566 45.074
88165 98.465 -79.574 45.065
88166 98.499 -79.583 45.056
88167 98.532 -79.592 45.049
88168 98.565 -79.601 45.042
88169 98.594 -79.610 45.036
88170 98.620 -79.619 45.029
88171 98.645 -79.628 45.023
88172 98.670 -79.637 45.018
88173 98.696 -79.647 45.011
88174 98.695 -79.670 45.006
88175 98.687 -79.699 45.000
88176 98.701 -79.694 44.995
88177 98.718 -79.686 44.991
88178 98.735 -79.676 44.988
88179 98.752 -79.668 44.984
88180 98.769 -79.662 44.978
88181 98.785 -79.661 44.971
88182 98.800 -79.661 44.965
88183 98.819 -79.669 44.957
88184 98.838 -79.680 44.947
88185 98.857 -79.692 44.938
88186 98.876 -79.703 44.929
88187 98.902 -79.712 44.920
88188 98.942 -79.713 44.911
88189 98.984 -79.714 44.902
88190 99.025 -79.715 44.893
88191 99.066 -79.716 44.881
88192 99.107 -79.717 44.868
88193 99.148 -79.718 44.857
88194 99.174 -79.730 44.845
88195 99.192 -79.745 44.835
88196 99.196 -79.759 44.825
88197 99.199 -79.773 44.814
88198 99.203 -79.788 44.804
88199 99.215 -79.798 44.794
88200 99.239 -79.805 44.784
88201 99.263 -79.811 44.774
88202 99.286 -79.817 44.764
88203 99.310 -79.823 44.756
88204 99.334 -79.830 44.749
88205 99.357 -79.836 44.741
88206 99.381 -79.842 44.731
88207 99.405 -79.849 44.723
88208 99.429 -79.855 44.715
88209 99.452 -79.861 44.707
88210 99.476 -79.867 44.698
88211 99.491 -79.861 44.689
88212 99.506 -79.855 44.680
88213 99.521 -79.850 44.672
88214 99.537 -79.844 44.664
88215 99.552 -79.838 44.655
88216 99.556 -79.855 44.647
88217 99.555 -79.883 44.640
88218 99.553 -79.911 44.632
88219 99.552 -79.939 44.623
88220 99.551 -79.967 44.614
88221 99.550 -79.995 44.606
88222 99.548 -80.023 44.599
88223 99.547 -80.051 44.592
88224 99.546 -80.079 44.585
88225 99.545 -80.107 44.577
88226 99.543 -80.135 44.570
88227 99.550 -80.162 44.563
88228 99.569 -80.186 44.556
88229 99.589 -80.211 44.549
88230 99.609 -80.235 44.542
88231 99.636 -80.267 44.534
88232 99.674 -80.308 44.525
88233 99.712 -80.349 44.516
88234 99.747 -80.375 44.506
88235 99.809 -80.354 44.496
88236 99.853 -80.370 44.485
88237 99.898 -80.387 44.474
88238 99.942 -80.403 44.462
88239 99.987 -80.419 44.450
88240 100.031 -80.435 44.437
88241 100.076 -80.452 44.425
88242 100.121 -80.469 44.412
88243 100.165 -80.485 44.399
88244 100.210 -80.501 44.386
88245 100.255 -80.517 44.371
88246 100.299 -80.534 44.357
88247 100.344 -80.550 44.344
88248 100.388 -80.566 44.330
88539 93.568 -78.702 53.644
88540 93.577 -78.700 53.633
88541 93.587 -78.698 53.623
88542 93.597 -78.695 53.612
88543 93.607 -78.692 53.602
88544 93.615 -78.689 53.593
88545 93.614 -78.680 53.584
88546 93.612 -78.671 53.575
88547 93.611 -78.661 53.567
88548 93.610 -78.652 53.560
88549 93.609 -78.643 53.555
88550 93.608 -78.633 53.552
88551 93.631 -78.618 53.547
88552 93.653 -78.604 53.540
88553 93.673 -78.594 53.530
88554 93.692 -78.583 53.518
88555 93.712 -78.573 53.502
88556 93.729 -78.565 53.484
88557 93.744 -78.560 53.464
88558 93.758 -78.554 53.442
88559 93.773 -78.549 53.419
88560 93.787 -78.544 53.397
88561 93.798 -78.535 53.374
88562 93.796 -78.519 53.353
88563 93.795 -78.501 53.332
88564 93.793 -78.484 53.313
88565 93.791 -78.467 53.294
88566 93.790 -78.450 53.277
88567 93.788 -78.433 53.259
88568 93.787 -78.416 53.241
88569 93.785 -78.398 53.222
88570 93.783 -78.382 53.201
88571 93.799 -78.346 53.180
88572 93.847 -78.399 53.159
88573 93.895 -78.395 53.139
88574 93.940 -78.381 53.121
88575 93.980 -78.373 53.105
88576 94.007 -78.385 53.089
88577 94.026 -78.396 53.073
88578 94.022 -78.409 53.054
88579 94.018 -78.422 53.035
88580 94.014 -78.435 53.014
88581 94.023 -78.442 52.992
88582 94.035 -78.447 52.969
88583 94.048 -78.450 52.947
88584 94.063 -78.447 52.924
88585 94.078 -78.444 52.904
88586 94.094 -78.441 52.886
88587 94.110 -78.438 52.868
88588 94.125 -78.435 52.850
88589 94.141 -78.432 52.836
88590 94.155 -78.429 52.821
88591 94.168 -78.429 52.805
88592 94.181 -78.427 52.790
88593 94.184 -78.437 52.773
88594 94.184 -78.449 52.755
88595 94.185 -78.460 52.735
88596 94.185 -78.473 52.718
88597 94.186 -78.484 52.702
88598 94.186 -78.496 52.687
88599 94.198 -78.504 52.670
88600 94.214 -78.509 52.653
88601 94.231 -78.514 52.637
88602 94.248 -78.520 52.619
88603 94.264 -78.525 52.601
88604 94.280 -78.531 52.584
88605 94.297 -78.537 52.567
88606 94.313 -78.542 52.550
88607 94.329 -78.548 52.533
88608 94.348 -78.548 52.514
88609 94.371 -78.542 52.493
88610 94.394 -78.535 52.470
88611 94.417 -78.529 52.447
88612 94.439 -78.523 52.425
88613 94.449 -78.537 52.400
88614 94.459 -78.550 52.374
88615 94.472 -78.545 52.347
88616 94.486 -78.539 52.313
88617 94.500 -78.532 52.281
88618 94.513 -78.529 52.247
88619 94.526 -78.531 52.213
88620 94.538 -78.533 52.181
88621 94.551 -78.535 52.150
88622 94.563 -78.539 52.120
88623 94.576 -78.541 52.086
88624 94.589 -78.543 52.052
88625 94.601 -78.545 52.017
88626 94.613 -78.548 51.982
88627 94.623 -78.546 51.945
88628 94.634 -78.545 51.907
88629 94.644 -78.543 51.869
88630 94.654 -78.542 51.827
88631 94.665 -78.540 51.781
88632 94.675 -78.539 51.734
88633 94.685 -78.537 51.685
88634 94.698 -78.539 51.632
88635 94.712 -78.542 51.578
88636 94.726 -78.544 51.525
88637 94.740 -78.547 51.472
88638 94.754 -78.550 51.420
88639 94.769 -78.552 51.367
88640 94.789 -78.556 51.313
88641 94.811 -78.559 51.262
88642 94.832 -78.563 51.211
88643 94.851 -78.566 51.160
88644 94.854 -78.570 51.111
88645 94.857 -78.574 51.063
88646 94.860 -78.579 51.017
88647 94.862 -78.582 50.972
88648 94.865 -78.587 50.929
88649 94.868 -78.591 50.886
88650 94.871 -78.595 50.845
88651 94.876 -78.594 50.803
88652 94.887 -78.576 50.761
88653 94.899 -78.560 50.719
88654 94.913 -78.542 50.677
88655 94.927 -78.524 50.634
88656 94.942 -78.506 50.591
88657 94.956 -78.489 50.547
88658 94.970 -78.470 50.504
88659 94.985 -78.453 50.461
88660 94.999 -78.435 50.416
88661 95.014 -78.416 50.373
88662 95.028 -78.399 50.329
88663 95.048 -78.404 50.285
88664 95.068 -78.410 50.242
88665 95.089 -78.414 50.201
88666 95.110 -78.420 50.159
88667 95.130 -78.425 50.115
88668 95.150 -78.430 50.071
88669 95.171 -78.435 50.029
88670 95.191 -78.440 49.988
88671 95.212 -78.445 49.946
88672 95.227 -78.450 49.908
88673 95.237 -78.454 49.869
88674 95.248 -78.457 49.831
88675 95.263 -78.463 49.795
88676 95.296 -78.480 49.756
88677 95.329 -78.498 49.719
88678 95.362 -78.514 49.682
88679 95.395 -78.531 49.645
88680 95.428 -78.548 49.607
88681 95.461 -78.565 49.570
88682 95.494 -78.582 49.535
88683 95.527 -78.598 49.499
88684 95.560 -78.615 49.464
88685 95.589 -78.630 49.429
88686 95.612 -78.642 49.394
88687 95.620 -78.643 49.359
88688 95.621 -78.642 49.324
88689 95.616 -78.642 49.291
88690 95.611 -78.642 49.258
88691 95.605 -78.641 49.223
88692 95.600 -78.640 49.188
88693 95.595 -78.640 49.150
88694 95.590 -78.639 49.112
88695 95.584 -78.639 49.072
88696 95.579 -78.639 49.029
88697 95.566 -78.635 48.985
88698 95.549 -78.630 48.941
88699 95.532 -78.625 48.898
88700 95.514 -78.620 48.856
88701 95.487 -78.620 48.816
88702 95.461 -78.620 48.776
88703 95.473 -78.620 48.737
88704 95.498 -78.621 48.697
88705 95.523 -78.622 48.657
88706 95.548 -78.623 48.615
88707 95.572 -78.623 48.575
88708 95.597 -78.623 48.535
88709 95.621 -78.624 48.496
88710 95.636 -78.628 48.458
88711 95.652 -78.632 48.421
88712 95.666 -78.636 48.384
88713 95.681 -78.641 48.349
88714 95.696 -78.646 48.314
88715 95.711 -78.651 48.277
88716 95.726 -78.655 48.237
88717 95.752 -78.653 48.195
88718 95.780 -78.649 48.154
88719 95.808 -78.646 48.113
88720 95.837 -78.642 48.073
88721 95.859 -78.642 48.034
88722 95.881 -78.644 47.994
88723 95.902 -78.641 47.953
88724 95.923 -78.639 47.913
88725 95.944 -78.637 47.873
88726 95.954 -78.642 47.831
88727 95.962 -78.648 47.787
88728 95.970 -78.655 47.747
88729 95.977 -78.660 47.706
88730 95.988 -78.664 47.666
88731 96.012 -78.661 47.626
88732 96.036 -78.658 47.586
88733 96.060 -78.657 47.546
88734 96.078 -78.674 47.507
88735 96.096 -78.692 47.470
88736 96.113 -78.709 47.431
88737 96.131 -78.727 47.392
88738 96.149 -78.744 47.354
88739 96.166 -78.761 47.316
88740 96.184 -78.779 47.279
88741 96.201 -78.786 47.242
88742 96.218 -78.773 47.205
88743 96.234 -78.761 47.168
88744 96.247 -78.758 47.132
88745 96.257 -78.767 47.097
88746 96.267 -78.775 47.066
88747 96.277 -78.783 47.037
88748 96.287 -78.792 47.007
88749 96.297 -78.800 46.978
88750 96.307 -78.808 46.948
88751 96.317 -78.817 46.918
88752 96.327 -78.824 46.888
88753 96.337 -78.833 46.857
88754 96.346 -78.841 46.827
88755 96.367 -78.849 46.797
88756 96.407 -78.843 46.768
88757 96.448 -78.839 46.739
88758 96.488 -78.834 46.710
88759 96.528 -78.829 46.683
88760 96.569 -78.824 46.656
88761 96.609 -78.819 46.629
88762 96.650 -78.814 46.605
88763 96.690 -78.810 46.583
88764 96.731 -78.805 46.562
88765 96.771 -78.800 46.540
88766 96.812 -78.796 46.519
88767 96.838 -78.791 46.498
88768 96.848 -78.789 46.477
88769 96.867 -78.798 46.455
88770 96.887 -78.808 46.436
88771 96.907 -78.818 46.417
88772 96.928 -78.829 46.398
88773 96.948 -78.830 46.378
88774 96.956 -78.871 46.357
88775 96.963 -78.882 46.338
88776 96.967 -78.888 46.319
88777 96.958 -78.902 46.304
88778 96.944 -78.920 46.289
88779 96.931 -78.937 46.273
88780 96.936 -78.949 46.257
88781 96.944 -78.959 46.241
88782 96.950 -78.969 46.225
88783 96.972 -78.973 46.208
88784 96.996 -78.977 46.192
88785 97.020 -78.982 46.176
88786 97.043 -78.987 46.159
88787 97.067 -78.991 46.142
88788 97.091 -78.996 46.125
88789 97.114 -79.000 46.107
88790 97.138 -79.005 46.090
88791 97.162 -79.009 46.075
88792 97.185 -79.014 46.061
88793 97.209 -79.018 46.046
88794 97.226 -79.028 46.034
88795 97.241 -79.040 46.020
88796 97.257 -79.050 46.005
88797 97.273 -79.062 45.988
88798 97.289 -79.072 45.971
88799 97.298 -79.078 45.952
88800 97.310 -79.085 45.932
88801 97.325 -79.094 45.911
88802 97.339 -79.101 45.890
88803 97.351 -79.100 45.871
88804 97.362 -79.100 45.854
88805 97.373 -79.099 45.838
88806 97.384 -79.097 45.822
88807 97.396 -79.096 45.806
88808 97.407 -79.096 45.788
88809 97.418 -79.095 45.771
88810 97.430 -79.094 45.753
88811 97.441 -79.093 45.736
88812 97.452 -79.093 45.721
88813 97.463 -79.092 45.708
88814 97.479 -79.096 45.695
88815 97.501 -79.108 45.679
88816 97.522 -79.121 45.663
88817 97.543 -79.134 45.645
88818 97.565 -79.147 45.627
88819 97.586 -79.161 45.607
88820 97.612 -79.184 45.587
88821 97.607 -79.183 45.568
88822 97.603 -79.182 45.552
88823 97.598 -79.182 45.536
88824 97.593 -79.181 45.519
88825 97.588 -79.181 45.502
88826 97.587 -79.181 45.485
88827 97.587 -79.180 45.469
88828 97.587 -79.179 45.453
88829 97.586 -79.179 45.439
88830 97.586 -79.178 45.425
88831 97.586 -79.178 45.409
88832 97.595 -79.187 45.394
88833 97.613 -79.208 45.381
88834 97.631 -79.228 45.369
88835 97.650 -79.248 45.358
88836 97.686 -79.272 45.347
88837 97.726 -79.295 45.336
88838 97.766 -79.319 45.325
88839 97.789 -79.332 45.315
88840 97.806 -79.341 45.305
88841 97.823 -79.349 45.296
88842 97.840 -79.357 45.287
88843 97.856 -79.366 45.277
88844 97.873 -79.374 45.267
88845 97.885 -79.389 45.258
88846 97.897 -79.405 45.250
88847 97.910 -79.420 45.242
88848 97.922 -79.435 45.235
88849 97.935 -79.451 45.228
88850 97.958 -79.460 45.220
88851 97.981 -79.470 45.213
88852 98.004 -79.480 45.206
88853 98.028 -79.488 45.199
88854 98.053 -79.496 45.191
88855 98.077 -79.503 45.183
88856 98.102 -79.510 45.174
88857 98.127 -79.517 45.165
88858 98.151 -79.519 45.155
88859 98.195 -79.514 45.144
88860 98.250 -79.509 45.132
88861 98.301 -79.505 45.121
88862 98.330 -79.512 45.111
88863 98.363 -79.523 45.103
88864 98.395 -79.535 45.094
88865 98.428 -79.547 45.086
88866 98.462 -79.555 45.078
88867 98.495 -79.564 45.070
88868 98.528 -79.573 45.064
88869 98.562 -79.582 45.057
88870 98.595 -79.591 45.050
88871 98.625 -79.600 45.044
88872 98.650 -79.609 45.038
88873 98.675 -79.618 45.032
88874 98.701 -79.627 45.026
88875 98.697 -79.653 45.020
88876 98.688 -79.682 45.014
88877 98.696 -79.687 45.008
88878 98.713 -79.679 45.002
88879 98.730 -79.670 44.997
88880 98.747 -79.661 44.993
88881 98.763 -79.661 44.986
88882 98.778 -79.660 44.979
88883 98.794 -79.659 44.972
88884 98.812 -79.665 44.964
88885 98.831 -79.676 44.953
88886 98.850 -79.688 44.942
88887 98.878 -79.695 44.933
88888 98.919 -79.696 44.923
88889 98.960 -79.697 44.912
88890 99.001 -79.698 44.902
88891 99.042 -79.699 44.893
88892 99.083 -79.700 44.882
88893 99.124 -79.701 44.870
88894 99.165 -79.702 44.860
88895 99.192 -79.714 44.849
88896 99.203 -79.728 44.838
88897 99.206 -79.742 44.827
88898 99.210 -79.757 44.818
88899 99.213 -79.771 44.808
88900 99.219 -79.785 44.798
88901 99.242 -79.791 44.790
88902 99.267 -79.797 44.781
88903 99.290 -79.803 44.771
88904 99.314 -79.810 44.761
88905 99.338 -79.816 44.752
88906 99.361 -79.822 44.743
88907 99.385 -79.828 44.733
88908 99.409 -79.835 44.725
88909 99.432 -79.841 44.717
88910 99.456 -79.847 44.708
88911 99.473 -79.844 44.700
88912 99.488 -79.838 44.691
88913 99.503 -79.832 44.683
88914 99.518 -79.827 44.675
88915 99.534 -79.820 44.666
88916 99.549 -79.814 44.658
88917 99.551 -79.834 44.651
88918 99.550 -79.862 44.643
88919 99.549 -79.890 44.635
88920 99.548 -79.918 44.627
88921 99.546 -79.946 44.619
88922 99.545 -79.974 44.610
88923 99.544 -80.002 44.604
88924 99.543 -80.030 44.596
88925 99.541 -80.059 44.589
88926 99.540 -80.087 44.582
88927 99.552 -80.112 44.574
88928 99.571 -80.137 44.567
88929 99.591 -80.162 44.561
88930 99.611 -80.186 44.554
88931 99.630 -80.210 44.547
88932 99.659 -80.243 44.539
88933 99.697 -80.284 44.531
88934 99.735 -80.325 44.521
88935 99.765 -80.362 44.511
88936 99.815 -80.346 44.502
88937 99.859 -80.359 44.490
88938 99.900 -80.381 44.479
88939 99.945 -80.397 44.468
88940 99.989 -80.413 44.456
88941 100.034 -80.430 44.444
88942 100.078 -80.446 44.431
88943 100.123 -80.463 44.417
88944 100.168 -80.478 44.404
88945 100.212 -80.495 44.391
88946 100.257 -80.512 44.377
88947 100.301 -80.528 44.362
88948 100.346 -80.544 44.348
88949 100.391 -80.560 44.334
88950 100.435 -80.577 44.321
88951 100.480 -80.593 44.307
89241 93.578 -78.689 53.639
89242 93.587 -78.686 53.628
89243 93.597 -78.684 53.617
89244 93.607 -78.682 53.606
89245 93.612 -78.677 53.596
89246 93.610 -78.667 53.586
89247 93.609 -78.658 53.575
89248 93.608 -78.648 53.566
89249 93.606 -78.639 53.558
89250 93.605 -78.630 53.550
89251 93.614 -78.618 53.545
89252 93.636 -78.603 53.538
89253 93.658 -78.588 53.530
89254 93.681 -78.573 53.519
89255 93.701 -78.561 53.504
89256 93.717 -78.555 53.487
89257 93.731 -78.550 53.467
89258 93.747 -78.545 53.446
89259 93.761 -78.539 53.425
89260 93.776 -78.534 53.404
89261 93.790 -78.529 53.382
89262 93.805 -78.523 53.360
89263 93.808 -78.510 53.340
89264 93.807 -78.493 53.319
89265 93.805 -78.476 53.300
89266 93.803 -78.459 53.282
89267 93.802 -78.442 53.264
89268 93.800 -78.424 53.246
89269 93.799 -78.407 53.229
89270 93.797 -78.390 53.210
89271 93.795 -78.373 53.191
89272 93.815 -78.358 53.171
89273 93.849 -78.354 53.151
89274 93.900 -78.358 53.132
89275 93.936 -78.355 53.115
89276 93.964 -78.366 53.098
89277 93.991 -78.377 53.082
89278 94.018 -78.388 53.066
89279 94.024 -78.401 53.048
89280 94.020 -78.413 53.029
89281 94.016 -78.426 53.009
89282 94.025 -78.433 52.987
89283 94.036 -78.438 52.966
89284 94.048 -78.444 52.943
89285 94.062 -78.444 52.921
89286 94.078 -78.441 52.903
89287 94.094 -78.438 52.887
89288 94.109 -78.435 52.871
89289 94.125 -78.432 52.855
89290 94.138 -78.430 52.840
89291 94.151 -78.429 52.826
89292 94.165 -78.428 52.811
89293 94.177 -78.428 52.795
89294 94.178 -78.439 52.779
89295 94.178 -78.451 52.762
89296 94.179 -78.463 52.743
89297 94.179 -78.475 52.727
89298 94.180 -78.486 52.711
89299 94.183 -78.492 52.696
89300 94.197 -78.492 52.679
89301 94.215 -78.495 52.663
89302 94.231 -78.501 52.647
89303 94.248 -78.506 52.629
89304 94.264 -78.511 52.611
89305 94.281 -78.517 52.593
89306 94.297 -78.523 52.577
89307 94.313 -78.529 52.559
89308 94.330 -78.534 52.541
89309 94.351 -78.529 52.523
89310 94.374 -78.523 52.502
89311 94.397 -78.517 52.480
89312 94.418 -78.512 52.457
89313 94.428 -78.526 52.435
89314 94.438 -78.539 52.411
89315 94.448 -78.552 52.385
89316 94.461 -78.548 52.357
89317 94.475 -78.542 52.324
89318 94.488 -78.535 52.291
89319 94.502 -78.529 52.258
89320 94.516 -78.523 52.225
89321 94.529 -78.519 52.193
89322 94.542 -78.521 52.163
89323 94.554 -78.523 52.134
89324 94.567 -78.526 52.102
89325 94.579 -78.529 52.069
89326 94.593 -78.531 52.034
89327 94.609 -78.532 51.999
89328 94.625 -78.534 51.961
89329 94.635 -78.532 51.922
89330 94.645 -78.531 51.882
89331 94.655 -78.529 51.839
89332 94.666 -78.528 51.793
89333 94.676 -78.526 51.745
89334 94.687 -78.525 51.696
89335 94.700 -78.526 51.641
89336 94.715 -78.529 51.585
89337 94.730 -78.532 51.530
89338 94.744 -78.535 51.477
89339 94.758 -78.539 51.425
89340 94.772 -78.541 51.372
89341 94.791 -78.545 51.318
89342 94.812 -78.548 51.266
89343 94.833 -78.551 51.214
89344 94.840 -78.555 51.163
89345 94.843 -78.559 51.114
89346 94.846 -78.564 51.066
89347 94.849 -78.567 51.020
89348 94.852 -78.572 50.976
89349 94.854 -78.576 50.933
89350 94.863 -78.578 50.891
89351 94.873 -78.578 50.850
89352 94.883 -78.562 50.808
89353 94.894 -78.545 50.766
89354 94.905 -78.528 50.723
89355 94.918 -78.510 50.680
89356 94.932 -78.493 50.636
89357 94.947 -78.475 50.592
89358 94.961 -78.457 50.547
89359 94.975 -78.439 50.504
89360 94.990 -78.422 50.460
89361 95.004 -78.404 50.416
89362 95.019 -78.385 50.372
89363 95.036 -78.379 50.328
89364 95.056 -78.384 50.284
89365 95.076 -78.389 50.241
89366 95.097 -78.395 50.200
89367 95.117 -78.400 50.158
89368 95.138 -78.405 50.115
89369 95.159 -78.410 50.072
89370 95.179 -78.415 50.031
89371 95.199 -78.420 49.990
89372 95.220 -78.426 49.950
89373 95.240 -78.431 49.912
89374 95.251 -78.434 49.874
89375 95.262 -78.438 49.837
89376 95.290 -78.452 49.800
89377 95.323 -78.469 49.762
89378 95.356 -78.485 49.724
89379 95.389 -78.503 49.686
89380 95.422 -78.520 49.649
89381 95.455 -78.536 49.611
89382 95.483 -78.551 49.574
89383 95.508 -78.563 49.539
89384 95.532 -78.575 49.503
89385 95.555 -78.586 49.467
89386 95.580 -78.598 49.433
89387 95.603 -78.611 49.397
89388 95.611 -78.612 49.362
89389 95.616 -78.611 49.328
89390 95.621 -78.611 49.294
89391 95.626 -78.611 49.260
89392 95.631 -78.610 49.225
89393 95.628 -78.610 49.190
89394 95.623 -78.609 49.154
89395 95.617 -78.608 49.117
89396 95.612 -78.608 49.078
89397 95.607 -78.608 49.036
89398 95.602 -78.607 48.993
89399 95.586 -78.603 48.950
89400 95.568 -78.598 48.908
89401 95.547 -78.595 48.867
89402 95.521 -78.595 48.827
89403 95.494 -78.595 48.788
89404 95.476 -78.595 48.749
89405 95.501 -78.595 48.709
89406 95.526 -78.595 48.668
89407 95.551 -78.596 48.626
89408 95.576 -78.597 48.585
89409 95.601 -78.598 48.544
89410 95.623 -78.599 48.505
89411 95.638 -78.602 48.467
89412 95.653 -78.606 48.429
89413 95.668 -78.611 48.392
89414 95.683 -78.615 48.355
89415 95.697 -78.620 48.319
89416 95.712 -78.624 48.281
89417 95.730 -78.627 48.241
89418 95.754 -78.626 48.200
89419 95.782 -78.623 48.158
89420 95.811 -78.619 48.118
89421 95.835 -78.618 48.079
89422 95.856 -78.619 48.040
89423 95.878 -78.620 48.000
89424 95.900 -78.619 47.959
89425 95.921 -78.616 47.918
89426 95.941 -78.614 47.878
89427 95.957 -78.613 47.838
89428 95.972 -78.610 47.797
89429 95.980 -78.615 47.757
89430 95.987 -78.621 47.717
89431 95.995 -78.627 47.677
89432 96.015 -78.626 47.636
89433 96.039 -78.623 47.596
89434 96.059 -78.634 47.556
89435 96.077 -78.651 47.516
89436 96.094 -78.669 47.478
89437 96.112 -78.686 47.439
89438 96.130 -78.704 47.400
89439 96.148 -78.722 47.362
89440 96.165 -78.739 47.324
89441 96.183 -78.757 47.286
89442 96.200 -78.774 47.250
89443 96.217 -78.762 47.214
89444 96.231 -78.758 47.178
89445 96.241 -78.767 47.143
89446 96.251 -78.775 47.109
89447 96.261 -78.783 47.078
89448 96.270 -78.792 47.048
89449 96.281 -78.800 47.019
89450 96.291 -78.808 46.988
89451 96.303 -78.812 46.958
89452 96.315 -78.816 46.929
89453 96.328 -78.819 46.899
89454 96.341 -78.823 46.869
89455 96.354 -78.826 46.839
89456 96.352 -78.815 46.810
89457 96.393 -78.817 46.781
89458 96.436 -78.818 46.752
89459 96.477 -78.820 46.725
89460 96.518 -78.817 46.698
89461 96.559 -78.812 46.672
89462 96.599 -78.807 46.648
89463 96.640 -78.802 46.625
89464 96.680 -78.798 46.603
89465 96.721 -78.793 46.581
89466 96.761 -78.788 46.561
89467 96.802 -78.783 46.539
89468 96.835 -78.779 46.518
89469 96.847 -78.778 46.496
89470 96.867 -78.789 46.475
89471 96.887 -78.799 46.455
89472 96.907 -78.810 46.435
89473 96.927 -78.820 46.415
89474 96.947 -78.824 46.395
89475 96.957 -78.858 46.374
89476 96.962 -78.864 46.354
89477 96.966 -78.871 46.337
89478 96.958 -78.884 46.321
89479 96.945 -78.901 46.305
89480 96.931 -78.918 46.288
89481 96.925 -78.933 46.273
89482 96.933 -78.944 46.257
89483 96.949 -78.955 46.240
89484 96.974 -78.961 46.221
89485 97.000 -78.966 46.204
89486 97.024 -78.971 46.186
89487 97.047 -78.975 46.169
89488 97.071 -78.980 46.153
89489 97.095 -78.984 46.136
89490 97.118 -78.989 46.119
89491 97.142 -78.993 46.102
89492 97.166 -78.998 46.088
89493 97.189 -79.002 46.074
89494 97.212 -79.008 46.060
89495 97.228 -79.019 46.048
89496 97.243 -79.030 46.034
89497 97.259 -79.041 46.019
89498 97.275 -79.052 46.003
89499 97.291 -79.063 45.988
89500 97.306 -79.074 45.971
89501 97.317 -79.080 45.952
89502 97.331 -79.085 45.929
89503 97.342 -79.084 45.907
89504 97.354 -79.084 45.886
89505 97.365 -79.083 45.869
89506 97.376 -79.082 45.854
89507 97.387 -79.081 45.839
89508 97.399 -79.081 45.823
89509 97.410 -79.080 45.805
89510 97.422 -79.079 45.787
89511 97.436 -79.078 45.768
89512 97.449 -79.078 45.750
89513 97.462 -79.078 45.736
89514 97.471 -79.075 45.724
89515 97.479 -79.079 45.712
89516 97.500 -79.093 45.696
89517 97.521 -79.106 45.677
89518 97.542 -79.119 45.657
89519 97.564 -79.132 45.638
89520 97.585 -79.146 45.618
89521 97.610 -79.169 45.599
89522 97.605 -79.169 45.581
89523 97.600 -79.169 45.565
89524 97.596 -79.168 45.549
89525 97.591 -79.168 45.532
89526 97.586 -79.167 45.514
89527 97.581 -79.166 45.497
89528 97.578 -79.165 45.480
89529 97.578 -79.165 45.464
89530 97.578 -79.165 45.449
89531 97.578 -79.165 45.434
89532 97.578 -79.164 45.419
89533 97.587 -79.175 45.405
89534 97.605 -79.195 45.392
89535 97.624 -79.216 45.380
89536 97.658 -79.238 45.369
89537 97.698 -79.262 45.359
89538 97.738 -79.285 45.348
89539 97.773 -79.306 45.337
89540 97.789 -79.314 45.327
89541 97.806 -79.322 45.317
89542 97.823 -79.332 45.308
89543 97.840 -79.340 45.299
89544 97.856 -79.348 45.289
89545 97.873 -79.357 45.279
89546 97.886 -79.370 45.271
89547 97.898 -79.385 45.264
89548 97.911 -79.401 45.256
89549 97.923 -79.416 45.249
89550 97.942 -79.429 45.242
89551 97.965 -79.438 45.234
89552 97.988 -79.448 45.227
89553 98.011 -79.458 45.219
89554 98.035 -79.466 45.210
89555 98.060 -79.473 45.202
89556 98.084 -79.481 45.193
89557 98.109 -79.488 45.184
89558 98.134 -79.494 45.174
89559 98.158 -79.490 45.164
89560 98.207 -79.485 45.154
89561 98.262 -79.480 45.144
89562 98.301 -79.479 45.133
89563 98.328 -79.484 45.123
89564 98.361 -79.496 45.116
89565 98.394 -79.507 45.107
89566 98.427 -79.520 45.099
89567 98.459 -79.531 45.092
89568 98.492 -79.543 45.085
89569 98.525 -79.554 45.078
89570 98.558 -79.563 45.072
89571 98.591 -79.572 45.066
89572 98.625 -79.581 45.060
89573 98.655 -79.589 45.053
89574 98.680 -79.598 45.047
89575 98.706 -79.607 45.041
89576 98.699 -79.636 45.035
89577 98.690 -79.665 45.029
89578 98.690 -79.681 45.022
89579 98.707 -79.672 45.014
89580 98.724 -79.664 45.007
89581 98.740 -79.660 45.000
89582 98.756 -79.659 44.994
89583 98.772 -79.658 44.987
89584 98.788 -79.657 44.980
89585 98.804 -79.661 44.971
89586 98.823 -79.672 44.960
89587 98.854 -79.678 44.950
89588 98.895 -79.679 44.940
89589 98.936 -79.680 44.928
89590 98.978 -79.681 44.917
89591 99.018 -79.683 44.906
89592 99.060 -79.683 44.896
89593 99.101 -79.684 44.884
89594 99.142 -79.685 44.873
89595 99.183 -79.686 44.863
89596 99.209 -79.698 44.853
89597 99.213 -79.712 44.842
89598 99.216 -79.726 44.832
89599 99.220 -79.740 44.823
89600 99.223 -79.755 44.814
89601 99.227 -79.769 44.805
89602 99.246 -79.777 44.796
89603 99.270 -79.783 44.786
89604 99.294 -79.789 44.776
89605 99.317 -79.796 44.766
89606 99.342 -79.802 44.756
89607 99.365 -79.808 44.746
89608 99.389 -79.814 44.737
89609 99.413 -79.821 44.728
89610 99.436 -79.827 44.720
89611 99.454 -79.826 44.711
89612 99.470 -79.820 44.702
89613 99.485 -79.814 44.694
89614 99.500 -79.808 44.685
89615 99.515 -79.803 44.677
89616 99.530 -79.797 44.669
89617 99.545 -79.791 44.662
89618 99.547 -79.814 44.655
89619 99.545 -79.842 44.647
89620 99.544 -79.870 44.639
89621 99.543 -79.898 44.632
89622 99.542 -79.926 44.624
89623 99.540 -79.954 44.615
89624 99.539 -79.982 44.608
89625 99.538 -80.010 44.601
89626 99.537 -80.038 44.593
89627 99.554 -80.063 44.587
89628 99.573 -80.087 44.579
89629 99.593 -80.112 44.572
89630 99.613 -80.137 44.566
89631 99.632 -80.161 44.558
89632 99.652 -80.186 44.551
89633 99.682 -80.220 44.543
89634 99.720 -80.261 44.534
89635 99.758 -80.302 44.525
89636 99.783 -80.336 44.516
89637 99.807 -80.356 44.506
89638 99.868 -80.350 44.495
89639 99.913 -80.361 44.484
89640 99.947 -80.391 44.473
89641 99.992 -80.407 44.461
89642 100.036 -80.424 44.448
89643 100.081 -80.441 44.436
89644 100.125 -80.457 44.423
89645 100.170 -80.473 44.410
89646 100.214 -80.489 44.396
89647 100.259 -80.506 44.382
89648 100.303 -80.522 44.367
89649 100.348 -80.538 44.353
89650 100.393 -80.555 44.339
89651 100.437 -80.571 44.325
89652 100.482 -80.588 44.311
89653 100.527 -80.604 44.296
89654 100.571 -80.620 44.282
89655 100.616 -80.636 44.267
89942 93.578 -78.678 53.644
89943 93.588 -78.676 53.632
89944 93.597 -78.673 53.620
89945 93.607 -78.670 53.609
89946 93.608 -78.663 53.598
89947 93.607 -78.654 53.587
89948 93.606 -78.645 53.575
89949 93.604 -78.636 53.565
89950 93.603 -78.626 53.555
89951 93.602 -78.617 53.546
89952 93.619 -78.602 53.538
89953 93.641 -78.587 53.528
89954 93.664 -78.572 53.518
89955 93.687 -78.557 53.505
89956 93.705 -78.545 53.490
89957 93.720 -78.539 53.471
89958 93.734 -78.535 53.451
89959 93.749 -78.529 53.431
89960 93.764 -78.524 53.412
89961 93.778 -78.519 53.392
89962 93.793 -78.514 53.371
89963 93.807 -78.508 53.350
89964 93.821 -78.501 53.330
89965 93.819 -78.485 53.310
89966 93.817 -78.467 53.291
89967 93.816 -78.451 53.272
89968 93.814 -78.433 53.254
89969 93.813 -78.416 53.235
89970 93.811 -78.399 53.218
89971 93.809 -78.382 53.201
89972 93.807 -78.364 53.182
89973 93.830 -78.350 53.163
89974 93.859 -78.336 53.144
89975 93.893 -78.337 53.125
89976 93.920 -78.348 53.108
89977 93.947 -78.359 53.092
89978 93.975 -78.370 53.076
89979 94.002 -78.381 53.060
89980 94.026 -78.392 53.043
89981 94.022 -78.405 53.025
89982 94.018 -78.418 53.005
89983 94.026 -78.425 52.985
89984 94.038 -78.430 52.963
89985 94.050 -78.435 52.942
89986 94.062 -78.441 52.921
89987 94.077 -78.438 52.905
89988 94.093 -78.435 52.890
89989 94.109 -78.432 52.875
89990 94.122 -78.431 52.860
89991 94.135 -78.430 52.845
89992 94.148 -78.429 52.831
89993 94.161 -78.428 52.817
89994 94.171 -78.429 52.802
89995 94.171 -78.442 52.786
89996 94.172 -78.453 52.770
89997 94.172 -78.465 52.753
89998 94.173 -78.477 52.736
89999 94.176 -78.482 52.721
90000 94.180 -78.486 52.706
90001 94.191 -78.488 52.690
90002 94.213 -78.483 52.674
90003 94.231 -78.486 52.658
90004 94.248 -78.492 52.641
90005 94.265 -78.497 52.623
90006 94.281 -78.503 52.605
90007 94.297 -78.508 52.589
90008 94.314 -78.514 52.570
90009 94.332 -78.517 52.553
90010 94.354 -78.510 52.535
90011 94.377 -78.504 52.514
90012 94.397 -78.501 52.492
90013 94.407 -78.514 52.470
90014 94.417 -78.528 52.447
90015 94.427 -78.542 52.422
90016 94.436 -78.554 52.396
90017 94.450 -78.552 52.369
90018 94.463 -78.545 52.336
90019 94.477 -78.539 52.303
90020 94.491 -78.532 52.270
90021 94.504 -78.526 52.238
90022 94.518 -78.520 52.206
90023 94.532 -78.513 52.175
90024 94.545 -78.509 52.145
90025 94.558 -78.512 52.114
90026 94.571 -78.514 52.081
90027 94.588 -78.516 52.046
90028 94.604 -78.517 52.011
90029 94.621 -78.519 51.973
90030 94.637 -78.520 51.934
90031 94.647 -78.518 51.893
90032 94.657 -78.517 51.850
90033 94.668 -78.515 51.802
90034 94.678 -78.514 51.752
90035 94.688 -78.512 51.702
90036 94.701 -78.514 51.646
90037 94.716 -78.517 51.591
90038 94.732 -78.520 51.535
90039 94.748 -78.524 51.481
90040 94.762 -78.527 51.429
90041 94.776 -78.529 51.376
90042 94.793 -78.532 51.323
90043 94.814 -78.536 51.270
90044 94.826 -78.540 51.219
90045 94.829 -78.544 51.168
90046 94.832 -78.548 51.120
90047 94.835 -78.552 51.071
90048 94.841 -78.555 51.024
90049 94.851 -78.557 50.981
90050 94.861 -78.558 50.939
90051 94.871 -78.560 50.899
90052 94.881 -78.546 50.858
90053 94.890 -78.529 50.816
90054 94.900 -78.514 50.773
90055 94.911 -78.497 50.729
90056 94.923 -78.479 50.685
90057 94.937 -78.462 50.640
90058 94.952 -78.444 50.594
90059 94.966 -78.426 50.548
90060 94.980 -78.408 50.504
90061 94.995 -78.390 50.460
90062 95.009 -78.373 50.415
90063 95.024 -78.354 50.371
90064 95.044 -78.359 50.327
90065 95.064 -78.364 50.284
90066 95.085 -78.369 50.240
90067 95.105 -78.374 50.198
90068 95.125 -78.379 50.156
90069 95.146 -78.385 50.114
90070 95.167 -78.390 50.073
90071 95.187 -78.395 50.033
90072 95.207 -78.400 49.993
90073 95.228 -78.405 49.953
90074 95.248 -78.410 49.916
90075 95.265 -78.415 49.879
90076 95.283 -78.423 49.842
90077 95.317 -78.440 49.804
90078 95.349 -78.457 49.766
90079 95.379 -78.472 49.729
90080 95.402 -78.484 49.691
90081 95.426 -78.495 49.653
90082 95.451 -78.507 49.616
90083 95.474 -78.520 49.579
90084 95.498 -78.532 49.543
90085 95.523 -78.543 49.506
90086 95.546 -78.555 49.471
90087 95.570 -78.567 49.436
90088 95.594 -78.579 49.400
90089 95.602 -78.580 49.366
90090 95.607 -78.580 49.332
90091 95.612 -78.579 49.298
90092 95.617 -78.579 49.263
90093 95.622 -78.579 49.228
90094 95.628 -78.579 49.194
90095 95.633 -78.578 49.159
90096 95.638 -78.578 49.123
90097 95.640 -78.577 49.085
90098 95.635 -78.576 49.045
90099 95.629 -78.576 49.003
90100 95.623 -78.575 48.961
90101 95.606 -78.570 48.919
90102 95.581 -78.570 48.878
90103 95.554 -78.569 48.839
90104 95.527 -78.569 48.800
90105 95.501 -78.569 48.762
90106 95.504 -78.569 48.721
90107 95.529 -78.570 48.680
90108 95.554 -78.570 48.637
90109 95.579 -78.571 48.595
90110 95.604 -78.571 48.554
90111 95.624 -78.573 48.515
90112 95.639 -78.577 48.476
90113 95.655 -78.581 48.438
90114 95.670 -78.585 48.400
90115 95.684 -78.589 48.362
90116 95.699 -78.594 48.325
90117 95.714 -78.598 48.286
90118 95.735 -78.599 48.246
90119 95.758 -78.598 48.206
90120 95.784 -78.596 48.165
90121 95.811 -78.594 48.124
90122 95.832 -78.595 48.085
90123 95.854 -78.595 48.046
90124 95.876 -78.596 48.006
90125 95.897 -78.597 47.965
90126 95.917 -78.595 47.925
90127 95.934 -78.594 47.885
90128 95.951 -78.593 47.845
90129 95.966 -78.590 47.805
90130 95.981 -78.587 47.766
90131 95.997 -78.583 47.727
90132 96.006 -78.589 47.687
90133 96.018 -78.592 47.647
90134 96.040 -78.594 47.607
90135 96.058 -78.611 47.567
90136 96.076 -78.629 47.528
90137 96.093 -78.647 47.488
90138 96.111 -78.664 47.449
90139 96.128 -78.682 47.410
90140 96.147 -78.699 47.372
90141 96.164 -78.717 47.334
90142 96.182 -78.734 47.296
90143 96.199 -78.752 47.260
90144 96.215 -78.758 47.224
90145 96.225 -78.766 47.189
90146 96.238 -78.769 47.155
90147 96.250 -78.773 47.123
90148 96.263 -78.776 47.092
90149 96.276 -78.779 47.061
90150 96.289 -78.783 47.030
90151 96.302 -78.786 47.000
90152 96.315 -78.789 46.970
90153 96.327 -78.793 46.940
90154 96.340 -78.796 46.911
90155 96.353 -78.799 46.882
90156 96.366 -78.802 46.852
90157 96.357 -78.788 46.823
90158 96.377 -78.777 46.795
90159 96.419 -78.779 46.767
90160 96.461 -78.780 46.741
90161 96.503 -78.782 46.714
90162 96.545 -78.784 46.690
90163 96.587 -78.786 46.667
90164 96.629 -78.787 46.645
90165 96.670 -78.786 46.622
90166 96.711 -78.780 46.601
90167 96.751 -78.776 46.581
90168 96.792 -78.771 46.561
90169 96.831 -78.766 46.538
90170 96.846 -78.769 46.516
90171 96.866 -78.780 46.494
90172 96.886 -78.790 46.474
90173 96.906 -78.801 46.453
90174 96.927 -78.811 46.432
90175 96.947 -78.820 46.412
90176 96.956 -78.840 46.391
90177 96.960 -78.846 46.370
90178 96.965 -78.852 46.353
90179 96.958 -78.865 46.337
90180 96.945 -78.883 46.320
90181 96.931 -78.900 46.304
90182 96.917 -78.918 46.289
90183 96.931 -78.931 46.272
90184 96.947 -78.943 46.253
90185 96.973 -78.948 46.234
90186 96.999 -78.953 46.215
90187 97.024 -78.959 46.197
90188 97.050 -78.964 46.179
90189 97.075 -78.969 46.163
90190 97.099 -78.974 46.147
90191 97.122 -78.978 46.131
90192 97.146 -78.983 46.114
90193 97.170 -78.987 46.100
90194 97.193 -78.992 46.087
90195 97.214 -78.998 46.075
90196 97.230 -79.009 46.062
90197 97.246 -79.020 46.047
90198 97.261 -79.031 46.032
90199 97.277 -79.042 46.018
90200 97.293 -79.053 46.003
90201 97.309 -79.065 45.987
90202 97.325 -79.071 45.970
90203 97.338 -79.070 45.947
90204 97.352 -79.070 45.924
90205 97.365 -79.069 45.901
90206 97.379 -79.069 45.882
90207 97.392 -79.069 45.867
90208 97.405 -79.068 45.852
90209 97.419 -79.068 45.836
90210 97.433 -79.068 45.819
90211 97.445 -79.067 45.802
90212 97.453 -79.065 45.783
90213 97.462 -79.062 45.765
90214 97.470 -79.060 45.750
90215 97.478 -79.058 45.736
90216 97.487 -79.056 45.723
90217 97.499 -79.078 45.706
90218 97.520 -79.091 45.688
90219 97.541 -79.104 45.668
90220 97.562 -79.118 45.651
90221 97.586 -79.135 45.632
90222 97.608 -79.156 45.614
90223 97.603 -79.155 45.597
90224 97.598 -79.155 45.579
90225 97.593 -79.154 45.561
90226 97.589 -79.153 45.542
90227 97.584 -79.153 45.524
90228 97.579 -79.153 45.508
90229 97.574 -79.152 45.492
90230 97.570 -79.151 45.476
90231 97.569 -79.151 45.461
90232 97.569 -79.150 45.445
90233 97.569 -79.150 45.430
90234 97.579 -79.162 45.417
90235 97.597 -79.182 45.404
90236 97.630 -79.204 45.392
90237 97.669 -79.228 45.380
90238 97.709 -79.252 45.371
90239 97.749 -79.275 45.360
90240 97.773 -79.288 45.350
90241 97.790 -79.297 45.339
90242 97.806 -79.305 45.329
90243 97.823 -79.313 45.320
90244 97.840 -79.322 45.310
90245 97.856 -79.331 45.301
90246 97.873 -79.339 45.292
90247 97.887 -79.351 45.284
90248 97.899 -79.366 45.276
90249 97.912 -79.382 45.269
90250 97.926 -79.396 45.261
90251 97.949 -79.406 45.255
90252 97.972 -79.416 45.247
90253 97.996 -79.426 45.239
90254 98.019 -79.435 45.231
90255 98.042 -79.444 45.222
90256 98.067 -79.451 45.212
90257 98.092 -79.459 45.203
90258 98.116 -79.466 45.194
90259 98.141 -79.465 45.184
90260 98.165 -79.461 45.174
90261 98.218 -79.456 45.164
90262 98.274 -79.451 45.155
90263 98.301 -79.454 45.145
90264 98.327 -79.457 45.136
90265 98.359 -79.468 45.130
90266 98.392 -79.479 45.121
90267 98.425 -79.491 45.114
90268 98.457 -79.503 45.107
90269 98.490 -79.515 45.100
90270 98.523 -79.527 45.094
90271 98.556 -79.538 45.088
90272 98.588 -79.551 45.083
90273 98.622 -79.561 45.077
90274 98.655 -79.570 45.070
90275 98.686 -79.579 45.063
90276 98.709 -79.590 45.057
90277 98.700 -79.619 45.050
90278 98.692 -79.648 45.044
90279 98.685 -79.674 45.037
90280 98.701 -79.666 45.028
90281 98.718 -79.659 45.020
90282 98.734 -79.658 45.014
90283 98.749 -79.657 45.007
90284 98.765 -79.657 45.000
90285 98.781 -79.655 44.991
90286 98.797 -79.657 44.982
90287 98.831 -79.661 44.971
90288 98.872 -79.663 44.960
90289 98.913 -79.664 44.949
90290 98.954 -79.664 44.936
90291 98.995 -79.665 44.924
90292 99.036 -79.667 44.913
90293 99.077 -79.667 44.903
90294 99.118 -79.668 44.891
90295 99.159 -79.670 44.880
90296 99.200 -79.670 44.870
90297 99.220 -79.681 44.861
90298 99.223 -79.695 44.850
90299 99.227 -79.710 44.840
90300 99.230 -79.724 44.830
90301 99.234 -79.739 44.821
90302 99.237 -79.753 44.812
90303 99.251 -79.764 44.802
90304 99.274 -79.770 44.792
90305 99.298 -79.776 44.782
90306 99.322 -79.783 44.772
90307 99.345 -79.789 44.761
90308 99.369 -79.795 44.751
90309 99.393 -79.801 44.742
90310 99.417 -79.808 44.733
90311 99.436 -79.808 44.724
90312 99.451 -79.802 44.714
90313 99.467 -79.796 44.705
90314 99.482 -79.791 44.697
90315 99.497 -79.785 44.688
90316 99.512 -79.779 44.680
90317 99.527 -79.773 44.674
90318 99.542 -79.767 44.667
90319 99.542 -79.793 44.661
90320 99.541 -79.821 44.654
90321 99.540 -79.849 44.647
90322 99.538 -79.877 44.639
90323 99.537 -79.905 44.629
90324 99.536 -79.933 44.620
90325 99.535 -79.961 44.612
90326 99.536 -79.989 44.604
90327 99.555 -80.014 44.597
90328 99.575 -80.038 44.591
90329 99.594 -80.062 44.584
90330 99.614 -80.087 44.577
90331 99.634 -80.112 44.571
90332 99.654 -80.137 44.563
90333 99.674 -80.161 44.556
90334 99.705 -80.196 44.548
90335 99.743 -80.237 44.539
90336 99.778 -80.277 44.530
90337 99.801 -80.309 44.521
90338 99.825 -80.342 44.511
90339 99.869 -80.347 44.500
90340 99.921 -80.354 44.490
90341 99.967 -80.362 44.478
90342 99.994 -80.402 44.466
90343 100.039 -80.418 44.454
90344 100.083 -80.435 44.441
90345 100.128 -80.450 44.429
90346 100.172 -80.467 44.415
90347 100.217 -80.484 44.401
90348 100.261 -80.500 44.386
90349 100.306 -80.516 44.372
90350 100.350 -80.532 44.358
90351 100.395 -80.549 44.343
90352 100.439 -80.565 44.329
90353 100.484 -80.582 44.314
90354 100.528 -80.598 44.300
90355 100.573 -80.614 44.285
90356 100.618 -80.631 44.270
90357 100.662 -80.647 44.256
90358 100.707 -80.663 44.241
90359 100.752 -80.679 44.227
90643 93.578 -78.667 53.649
90644 93.588 -78.665 53.636
90645 93.597 -78.662 53.624
90646 93.606 -78.659 53.611
90647 93.605 -78.650 53.599
90648 93.604 -78.641 53.587
90649 93.602 -78.632 53.575
90650 93.601 -78.622 53.564
90651 93.600 -78.613 53.553
90652 93.602 -78.601 53.541
90653 93.624 -78.586 53.530
90654 93.647 -78.572 53.517
90655 93.670 -78.557 53.504
90656 93.692 -78.542 53.490
90657 93.707 -78.532 53.474
90658 93.722 -78.526 53.456
90659 93.737 -78.519 53.438
90660 93.752 -78.514 53.419
90661 93.767 -78.509 53.400
90662 93.781 -78.504 53.381
90663 93.796 -78.498 53.362
90664 93.811 -78.493 53.341
90665 93.825 -78.488 53.321
90666 93.831 -78.476 53.301
90667 93.829 -78.459 53.283
90668 93.828 -78.442 53.264
90669 93.826 -78.425 53.245
90670 93.825 -78.407 53.226
90671 93.823 -78.391 53.209
90672 93.821 -78.373 53.191
90673 93.820 -78.356 53.173
90674 93.844 -78.341 53.155
90675 93.873 -78.328 53.137
90676 93.897 -78.334 53.119
90677 93.911 -78.342 53.101
90678 93.931 -78.352 53.086
90679 93.958 -78.363 53.071
90680 93.985 -78.374 53.056
90681 94.012 -78.385 53.039
90682 94.024 -78.397 53.021
90683 94.020 -78.410 53.003
90684 94.028 -78.416 52.982
90685 94.040 -78.422 52.961
90686 94.052 -78.427 52.941
90687 94.064 -78.432 52.922
90688 94.077 -78.435 52.908
90689 94.092 -78.433 52.894
90690 94.105 -78.432 52.880
90691 94.118 -78.431 52.865
90692 94.131 -78.429 52.850
90693 94.144 -78.429 52.836
90694 94.157 -78.427 52.825
90695 94.165 -78.432 52.809
90696 94.165 -78.444 52.794
90697 94.166 -78.455 52.779
90698 94.166 -78.467 52.762
90699 94.170 -78.472 52.747
90700 94.174 -78.476 52.732
90701 94.178 -78.481 52.718
90702 94.186 -78.483 52.702
90703 94.208 -78.479 52.687
90704 94.230 -78.475 52.672
90705 94.248 -78.478 52.654
90706 94.265 -78.483 52.637
90707 94.281 -78.489 52.620
90708 94.297 -78.495 52.604
90709 94.314 -78.500 52.586
90710 94.334 -78.498 52.568
90711 94.357 -78.491 52.550
90712 94.377 -78.490 52.529
90713 94.386 -78.504 52.507
90714 94.396 -78.517 52.484
90715 94.406 -78.530 52.460
90716 94.416 -78.544 52.434
90717 94.425 -78.557 52.408
90718 94.438 -78.555 52.380
90719 94.452 -78.549 52.348
90720 94.466 -78.542 52.316
90721 94.479 -78.535 52.283
90722 94.493 -78.529 52.250
90723 94.506 -78.523 52.219
90724 94.521 -78.517 52.188
90725 94.534 -78.510 52.156
90726 94.552 -78.499 52.124
90727 94.566 -78.498 52.090
90728 94.582 -78.500 52.056
90729 94.599 -78.501 52.021
90730 94.616 -78.503 51.983
90731 94.632 -78.504 51.943
90732 94.648 -78.506 51.902
90733 94.659 -78.504 51.857
90734 94.669 -78.503 51.808
90735 94.679 -78.501 51.757
90736 94.690 -78.500 51.706
90737 94.702 -78.501 51.651
90738 94.718 -78.504 51.596
90739 94.734 -78.508 51.540
90740 94.749 -78.511 51.486
90741 94.765 -78.515 51.433
90742 94.779 -78.519 51.381
90743 94.794 -78.521 51.328
90744 94.813 -78.525 51.276
90745 94.816 -78.529 51.224
90746 94.820 -78.532 51.174
90747 94.829 -78.534 51.126
90748 94.839 -78.535 51.079
90749 94.849 -78.537 51.033
90750 94.859 -78.539 50.990
90751 94.868 -78.540 50.948
90752 94.878 -78.530 50.908
90753 94.888 -78.514 50.868
90754 94.898 -78.498 50.825
90755 94.908 -78.482 50.782
90756 94.918 -78.465 50.737
90757 94.929 -78.448 50.691
90758 94.943 -78.431 50.644
90759 94.957 -78.413 50.597
90760 94.971 -78.395 50.549
90761 94.985 -78.377 50.504
90762 95.000 -78.359 50.459
90763 95.014 -78.341 50.414
90764 95.031 -78.334 50.370
90765 95.052 -78.339 50.326
90766 95.072 -78.344 50.283
90767 95.093 -78.349 50.241
90768 95.113 -78.354 50.198
90769 95.133 -78.360 50.156
90770 95.154 -78.365 50.115
90771 95.174 -78.369 50.076
90772 95.195 -78.375 50.037
90773 95.215 -78.380 49.997
90774 95.236 -78.385 49.957
90775 95.256 -78.391 49.919
90776 95.278 -78.398 49.882
90777 95.298 -78.404 49.844
90778 95.322 -78.416 49.807
90779 95.345 -78.429 49.768
90780 95.370 -78.441 49.730
90781 95.393 -78.452 49.693
90782 95.417 -78.464 49.657
90783 95.442 -78.476 49.620
90784 95.465 -78.489 49.583
90785 95.489 -78.500 49.547
90786 95.513 -78.512 49.512
90787 95.537 -78.524 49.476
90788 95.561 -78.535 49.440
90789 95.585 -78.548 49.405
90790 95.593 -78.549 49.370
90791 95.598 -78.549 49.336
90792 95.603 -78.548 49.303
90793 95.609 -78.548 49.268
90794 95.614 -78.548 49.234
90795 95.619 -78.548 49.201
90796 95.624 -78.547 49.167
90797 95.629 -78.547 49.132
90798 95.634 -78.546 49.095
90799 95.639 -78.545 49.055
90800 95.644 -78.545 49.014
90801 95.649 -78.545 48.972
90802 95.640 -78.544 48.931
90803 95.614 -78.544 48.893
90804 95.587 -78.544 48.853
90805 95.561 -78.543 48.812
90806 95.534 -78.543 48.775
90807 95.508 -78.543 48.735
90808 95.532 -78.543 48.692
90809 95.557 -78.544 48.649
90810 95.582 -78.545 48.606
90811 95.607 -78.545 48.565
90812 95.626 -78.548 48.526
90813 95.641 -78.551 48.486
90814 95.656 -78.555 48.448
90815 95.671 -78.559 48.410
90816 95.686 -78.564 48.372
90817 95.701 -78.568 48.333
90818 95.716 -78.572 48.294
90819 95.740 -78.571 48.254
90820 95.763 -78.570 48.214
90821 95.787 -78.569 48.174
90822 95.809 -78.570 48.133
90823 95.830 -78.570 48.092
90824 95.852 -78.572 48.052
90825 95.874 -78.573 48.012
90826 95.896 -78.573 47.972
90827 95.911 -78.573 47.933
90828 95.927 -78.573 47.892
90829 95.944 -78.573 47.851
90830 95.961 -78.571 47.810
90831 95.976 -78.568 47.773
90832 95.991 -78.564 47.735
90833 96.007 -78.561 47.697
90834 96.022 -78.558 47.659
90835 96.039 -78.571 47.620
90836 96.056 -78.588 47.581
90837 96.073 -78.604 47.542
90838 96.090 -78.622 47.502
90839 96.108 -78.639 47.463
90840 96.125 -78.656 47.424
90841 96.142 -78.673 47.384
90842 96.159 -78.689 47.346
90843 96.177 -78.707 47.308
90844 96.190 -78.720 47.272
90845 96.208 -78.738 47.236
90846 96.235 -78.742 47.202
90847 96.250 -78.745 47.169
90848 96.262 -78.749 47.138
90849 96.275 -78.752 47.107
90850 96.288 -78.756 47.076
90851 96.301 -78.759 47.044
90852 96.314 -78.762 47.014
90853 96.327 -78.766 46.985
90854 96.339 -78.769 46.955
90855 96.353 -78.772 46.925
90856 96.365 -78.776 46.897
90857 96.378 -78.779 46.868
90858 96.369 -78.765 46.839
90859 96.361 -78.738 46.811
90860 96.402 -78.739 46.784
90861 96.445 -78.742 46.758
90862 96.486 -78.743 46.732
90863 96.528 -78.745 46.707
90864 96.571 -78.746 46.684
90865 96.612 -78.748 46.662
90866 96.655 -78.749 46.642
90867 96.696 -78.751 46.622
90868 96.738 -78.752 46.602
90869 96.781 -78.754 46.581
90870 96.822 -78.754 46.558
90871 96.846 -78.761 46.535
90872 96.866 -78.771 46.513
90873 96.886 -78.781 46.492
90874 96.906 -78.792 46.472
90875 96.926 -78.802 46.451
90876 96.946 -78.812 46.429
90877 96.955 -78.822 46.406
90878 96.959 -78.829 46.383
90879 96.964 -78.835 46.363
90880 96.958 -78.847 46.347
90881 96.936 -78.869 46.332
90882 96.923 -78.890 46.317
90883 96.919 -78.905 46.301
90884 96.929 -78.919 46.283
90885 96.946 -78.930 46.264
90886 96.971 -78.935 46.244
90887 96.997 -78.940 46.226
90888 97.023 -78.946 46.208
90889 97.048 -78.951 46.190
90890 97.074 -78.956 46.174
90891 97.100 -78.962 46.157
90892 97.125 -78.967 46.142
90893 97.150 -78.971 46.127
90894 97.174 -78.976 46.114
90895 97.197 -78.981 46.101
90896 97.217 -78.989 46.089
90897 97.232 -78.999 46.074
90898 97.248 -79.011 46.060
90899 97.265 -79.021 46.046
90900 97.287 -79.032 46.031
90901 97.309 -79.043 46.016
90902 97.330 -79.053 46.000
90903 97.349 -79.059 45.981
90904 97.362 -79.059 45.959
90905 97.375 -79.059 45.937
90906 97.389 -79.058 45.915
90907 97.402 -79.058 45.895
90908 97.416 -79.058 45.879
90909 97.428 -79.056 45.864
90910 97.436 -79.054 45.849
90911 97.444 -79.052 45.834
90912 97.453 -79.049 45.817
90913 97.461 -79.047 45.800
90914 97.469 -79.045 45.783
90915 97.477 -79.043 45.768
90916 97.486 -79.040 45.752
90917 97.494 -79.038 45.736
90918 97.499 -79.060 45.720
90919 97.519 -79.076 45.703
90920 97.540 -79.089 45.685
90921 97.561 -79.103 45.669
90922 97.587 -79.124 45.651
90923 97.605 -79.142 45.632
90924 97.600 -79.141 45.613
90925 97.596 -79.140 45.593
90926 97.591 -79.140 45.573
90927 97.586 -79.140 45.554
90928 97.582 -79.139 45.536
90929 97.577 -79.139 45.520
90930 97.572 -79.138 45.504
90931 97.567 -79.137 45.489
90932 97.563 -79.137 45.473
90933 97.560 -79.137 45.457
90934 97.560 -79.136 45.442
90935 97.572 -79.149 45.429
90936 97.601 -79.171 45.415
90937 97.641 -79.194 45.403
90938 97.681 -79.218 45.391
90939 97.721 -79.242 45.382
90940 97.756 -79.263 45.372
90941 97.773 -79.271 45.361
90942 97.790 -79.279 45.352
90943 97.806 -79.288 45.341
90944 97.823 -79.296 45.331
90945 97.840 -79.305 45.321
90946 97.856 -79.313 45.312
90947 97.873 -79.322 45.305
90948 97.888 -79.332 45.297
90949 97.900 -79.347 45.288
90950 97.913 -79.363 45.280
90951 97.934 -79.374 45.273
90952 97.957 -79.384 45.266
90953 97.980 -79.394 45.258
90954 98.003 -79.404 45.251
90955 98.026 -79.413 45.242
90956 98.050 -79.422 45.233
90957 98.074 -79.429 45.224
90958 98.099 -79.437 45.214
90959 98.124 -79.440 45.204
90960 98.148 -79.436 45.195
90961 98.175 -79.432 45.185
90962 98.230 -79.427 45.175
90963 98.275 -79.425 45.166
90964 98.301 -79.428 45.157
90965 98.327 -79.432 45.149
90966 98.357 -79.440 45.142
90967 98.390 -79.452 45.134
90968 98.423 -79.463 45.127
90969 98.456 -79.476 45.121
90970 98.489 -79.487 45.115
90971 98.521 -79.499 45.109
90972 98.554 -79.510 45.103
90973 98.587 -79.523 45.098
90974 98.620 -79.534 45.092
90975 98.652 -79.546 45.086
90976 98.685 -79.558 45.080
90977 98.711 -79.573 45.073
90978 98.702 -79.601 45.066
90979 98.693 -79.630 45.059
90980 98.685 -79.659 45.052
90981 98.696 -79.659 45.043
90982 98.712 -79.657 45.035
90983 98.727 -79.657 45.026
90984 98.743 -79.655 45.019
90985 98.762 -79.645 45.011
90986 98.784 -79.629 45.002
90987 98.805 -79.643 44.992
90988 98.845 -79.645 44.981
90989 98.886 -79.646 44.969
90990 98.928 -79.647 44.957
90991 98.969 -79.648 44.945
90992 99.010 -79.649 44.933
90993 99.051 -79.650 44.922
90994 99.092 -79.651 44.911
90995 99.134 -79.652 44.900
90996 99.175 -79.653 44.890
90997 99.216 -79.654 44.880
90998 99.230 -79.664 44.871
90999 99.234 -79.679 44.860
91000 99.237 -79.693 44.849
91001 99.241 -79.708 44.839
91002 99.244 -79.722 44.829
91003 99.248 -79.736 44.819
91004 99.254 -79.750 44.809
91005 99.278 -79.756 44.799
91006 99.302 -79.762 44.788
91007 99.326 -79.768 44.778
91008 99.349 -79.775 44.769
91009 99.373 -79.781 44.759
91010 99.397 -79.787 44.749
91011 99.418 -79.790 44.739
91012 99.433 -79.785 44.730
91013 99.448 -79.779 44.720
91014 99.464 -79.773 44.710
91015 99.479 -79.767 44.702
91016 99.494 -79.761 44.694
91017 99.509 -79.755 44.685
91018 99.524 -79.750 44.678
91019 99.539 -79.745 44.671
91020 99.537 -79.773 44.665
91021 99.536 -79.801 44.658
91022 99.535 -79.829 44.651
91023 99.534 -79.857 44.642
91024 99.533 -79.885 44.634
91025 99.531 -79.913 44.625
91026 99.538 -79.939 44.617
91027 99.557 -79.965 44.610
91028 99.577 -79.989 44.603
91029 99.596 -80.013 44.597
91030 99.616 -80.038 44.589
91031 99.636 -80.062 44.583
91032 99.655 -80.087 44.576
91033 99.675 -80.112 44.569
91034 99.695 -80.137 44.562
91035 99.728 -80.173 44.553
91036 99.766 -80.214 44.545
91037 99.796 -80.250 44.535
91038 99.820 -80.284 44.526
91039 99.843 -80.316 44.517
91040 99.870 -80.343 44.506
91041 99.922 -80.350 44.495
91042 99.975 -80.358 44.484
91043 100.021 -80.363 44.472
91044 100.042 -80.411 44.459
91045 100.085 -80.428 44.447
91046 100.130 -80.445 44.434
91047 100.174 -80.461 44.420
91048 100.219 -80.478 44.406
91049 100.264 -80.494 44.391
91050 100.308 -80.510 44.377
91051 100.353 -80.526 44.362
91052 100.397 -80.543 44.348
91053 100.442 -80.560 44.333
91054 100.486 -80.576 44.317
91055 100.531 -80.592 44.302
91056 100.575 -80.608 44.287
91057 100.620 -80.625 44.272
91058 100.665 -80.641 44.258
91059 100.709 -80.657 44.243
91060 100.754 -80.674 44.229
91061 100.798 -80.690 44.215
91062 100.843 -80.707 44.201
91063 100.888 -80.723 44.187
91345 93.588 -78.654 53.641
91346 93.598 -78.651 53.628
91347 93.603 -78.646 53.614
91348 93.602 -78.637 53.601
91349 93.600 -78.627 53.588
91350 93.599 -78.618 53.575
91351 93.597 -78.609 53.563
91352 93.596 -78.600 53.550
91353 93.608 -78.586 53.537
91354 93.630 -78.571 53.524
91355 93.652 -78.556 53.511
91356 93.675 -78.541 53.496
91357 93.694 -78.528 53.481
91358 93.709 -78.520 53.464
91359 93.724 -78.513 53.446
91360 93.740 -78.505 53.428
91361 93.755 -78.498 53.410
91362 93.769 -78.494 53.392
91363 93.784 -78.489 53.373
91364 93.798 -78.483 53.353
91365 93.813 -78.478 53.333
91366 93.828 -78.473 53.314
91367 93.843 -78.467 53.295
91368 93.841 -78.451 53.276
91369 93.840 -78.433 53.257
91370 93.838 -78.416 53.237
91371 93.837 -78.399 53.219
91372 93.835 -78.382 53.200
91373 93.833 -78.365 53.183
91374 93.832 -78.347 53.165
91375 93.858 -78.334 53.148
91376 93.894 -78.322 53.130
91377 93.907 -78.331 53.112
91378 93.920 -78.339 53.095
91379 93.934 -78.348 53.080
91380 93.947 -78.357 53.066
91381 93.969 -78.366 53.051
91382 93.996 -78.378 53.034
91383 94.024 -78.388 53.017
91384 94.022 -78.401 52.999
91385 94.030 -78.408 52.978
91386 94.042 -78.413 52.959
91387 94.054 -78.419 52.940
91388 94.065 -78.424 52.922
91389 94.077 -78.429 52.909
91390 94.088 -78.432 52.896
91391 94.101 -78.432 52.882
91392 94.114 -78.430 52.868
91393 94.127 -78.429 52.855
91394 94.140 -78.428 52.842
91395 94.153 -78.427 52.831
91396 94.158 -78.434 52.817
91397 94.159 -78.446 52.803
91398 94.159 -78.457 52.790
91399 94.164 -78.462 52.773
91400 94.168 -78.467 52.758
91401 94.172 -78.470 52.743
91402 94.176 -78.475 52.729
91403 94.181 -78.479 52.714
91404 94.202 -78.475 52.699
91405 94.224 -78.470 52.684
91406 94.247 -78.467 52.667
91407 94.265 -78.469 52.650
91408 94.281 -78.474 52.634
91409 94.297 -78.480 52.618
91410 94.315 -78.485 52.601
91411 94.337 -78.478 52.584
91412 94.355 -78.479 52.567
91413 94.365 -78.492 52.545
91414 94.375 -78.506 52.521
91415 94.385 -78.519 52.497
91416 94.395 -78.532 52.472
91417 94.405 -78.546 52.445
91418 94.414 -78.559 52.417
91419 94.427 -78.559 52.388
91420 94.440 -78.552 52.358
91421 94.454 -78.546 52.327
91422 94.468 -78.539 52.295
91423 94.482 -78.533 52.262
91424 94.495 -78.526 52.230
91425 94.509 -78.520 52.198
91426 94.533 -78.504 52.166
91427 94.557 -78.488 52.132
91428 94.569 -78.482 52.098
91429 94.581 -78.484 52.063
91430 94.595 -78.486 52.028
91431 94.611 -78.488 51.990
91432 94.627 -78.489 51.950
91433 94.644 -78.491 51.907
91434 94.661 -78.492 51.860
91435 94.671 -78.491 51.811
91436 94.681 -78.489 51.761
91437 94.691 -78.488 51.709
91438 94.703 -78.488 51.655
91439 94.719 -78.492 51.600
91440 94.735 -78.495 51.545
91441 94.750 -78.499 51.490
91442 94.766 -78.502 51.437
91443 94.781 -78.506 51.385
91444 94.797 -78.510 51.334
91445 94.808 -78.511 51.282
91446 94.818 -78.513 51.231
91447 94.827 -78.514 51.181
91448 94.837 -78.516 51.133
91449 94.847 -78.517 51.087
91450 94.856 -78.519 51.042
91451 94.866 -78.520 51.000
91452 94.876 -78.514 50.958
91453 94.886 -78.498 50.918
91454 94.896 -78.482 50.878
91455 94.906 -78.466 50.835
91456 94.915 -78.449 50.790
91457 94.925 -78.433 50.744
91458 94.935 -78.416 50.696
91459 94.948 -78.400 50.648
91460 94.962 -78.382 50.599
91461 94.976 -78.363 50.550
91462 94.990 -78.346 50.503
91463 95.005 -78.328 50.457
91464 95.019 -78.310 50.413
91465 95.039 -78.313 50.369
91466 95.060 -78.319 50.325
91467 95.080 -78.324 50.282
91468 95.101 -78.329 50.240
91469 95.121 -78.335 50.198
91470 95.141 -78.339 50.157
91471 95.162 -78.344 50.117
91472 95.182 -78.350 50.079
91473 95.203 -78.355 50.040
91474 95.223 -78.360 50.000
91475 95.260 -78.390 49.959
91476 95.274 -78.383 49.921
91477 95.284 -78.374 49.883
91478 95.294 -78.376 49.845
91479 95.313 -78.385 49.808
91480 95.336 -78.398 49.771
91481 95.360 -78.409 49.734
91482 95.384 -78.421 49.697
91483 95.408 -78.433 49.662
91484 95.432 -78.445 49.625
91485 95.456 -78.457 49.588
91486 95.480 -78.469 49.552
91487 95.504 -78.481 49.516
91488 95.528 -78.493 49.480
91489 95.552 -78.504 49.444
91490 95.576 -78.517 49.409
91491 95.584 -78.518 49.375
91492 95.589 -78.517 49.341
91493 95.595 -78.517 49.308
91494 95.600 -78.517 49.275
91495 95.605 -78.517 49.242
91496 95.610 -78.516 49.208
91497 95.615 -78.516 49.175
91498 95.620 -78.515 49.140
91499 95.625 -78.515 49.103
91500 95.630 -78.514 49.064
91501 95.635 -78.514 49.024
91502 95.641 -78.515 48.984
91503 95.650 -78.522 48.944
91504 95.645 -78.518 48.905
91505 95.620 -78.518 48.866
91506 95.594 -78.517 48.827
91507 95.567 -78.517 48.788
91508 95.541 -78.517 48.747
91509 95.535 -78.517 48.705
91510 95.560 -78.518 48.661
91511 95.585 -78.519 48.619
91512 95.610 -78.519 48.578
91513 95.627 -78.522 48.538
91514 95.642 -78.526 48.498
91515 95.658 -78.529 48.459
91516 95.673 -78.533 48.421
91517 95.687 -78.538 48.382
91518 95.702 -78.542 48.344
91519 95.721 -78.544 48.304
91520 95.744 -78.543 48.265
91521 95.767 -78.543 48.226
91522 95.789 -78.547 48.186
91523 95.810 -78.546 48.145
91524 95.832 -78.547 48.103
91525 95.853 -78.548 48.061
91526 95.876 -78.548 48.020
91527 95.895 -78.549 47.980
91528 95.907 -78.548 47.940
91529 95.923 -78.549 47.899
91530 95.939 -78.549 47.858
91531 95.955 -78.549 47.818
91532 95.971 -78.551 47.781
91533 95.987 -78.548 47.743
91534 96.004 -78.546 47.704
91535 96.020 -78.544 47.666
91536 96.032 -78.543 47.627
91537 96.049 -78.557 47.588
91538 96.066 -78.575 47.550
91539 96.084 -78.592 47.511
91540 96.101 -78.609 47.472
91541 96.118 -78.626 47.433
91542 96.135 -78.642 47.393
91543 96.152 -78.659 47.355
91544 96.164 -78.670 47.319
91545 96.174 -78.695 47.283
91546 96.192 -78.713 47.249
91547 96.219 -78.717 47.217
91548 96.245 -78.721 47.185
91549 96.272 -78.726 47.156
91550 96.287 -78.729 47.125
91551 96.300 -78.733 47.094
91552 96.313 -78.736 47.061
91553 96.326 -78.739 47.032
91554 96.338 -78.742 47.003
91555 96.352 -78.745 46.974
91556 96.364 -78.749 46.945
91557 96.377 -78.752 46.916
91558 96.390 -78.755 46.887
91559 96.382 -78.742 46.858
91560 96.344 -78.698 46.829
91561 96.386 -78.701 46.802
91562 96.428 -78.702 46.776
91563 96.470 -78.704 46.750
91564 96.512 -78.705 46.724
91565 96.554 -78.707 46.701
91566 96.596 -78.708 46.679
91567 96.638 -78.710 46.658
91568 96.680 -78.711 46.638
91569 96.726 -78.717 46.618
91570 96.774 -78.727 46.596
91571 96.822 -78.737 46.574
91572 96.834 -78.748 46.552
91573 96.855 -78.758 46.529
91574 96.875 -78.769 46.509
91575 96.895 -78.780 46.489
91576 96.916 -78.790 46.468
91577 96.936 -78.801 46.445
91578 96.938 -78.819 46.421
91579 96.938 -78.837 46.397
91580 96.944 -78.843 46.377
91581 96.933 -78.845 46.360
91582 96.928 -78.861 46.344
91583 96.924 -78.877 46.329
91584 96.920 -78.893 46.311
91585 96.926 -78.906 46.293
91586 96.944 -78.917 46.274
91587 96.969 -78.922 46.255
91588 96.995 -78.927 46.237
91589 97.021 -78.933 46.220
91590 97.046 -78.938 46.202
91591 97.072 -78.943 46.186
91592 97.098 -78.949 46.171
91593 97.123 -78.954 46.156
91594 97.149 -78.959 46.141
91595 97.175 -78.965 46.128
91596 97.201 -78.970 46.116
91597 97.224 -78.979 46.103
91598 97.246 -78.990 46.088
91599 97.267 -78.999 46.073
91600 97.289 -79.010 46.058
91601 97.311 -79.021 46.043
91602 97.332 -79.031 46.028
91603 97.354 -79.042 46.012
91604 97.372 -79.048 45.993
91605 97.386 -79.047 45.971
91606 97.399 -79.047 45.950
91607 97.410 -79.046 45.929
91608 97.419 -79.043 45.908
91609 97.427 -79.041 45.891
91610 97.435 -79.039 45.876
91611 97.443 -79.037 45.862
91612 97.452 -79.034 45.848
91613 97.460 -79.032 45.832
91614 97.468 -79.030 45.816
91615 97.476 -79.028 45.800
91616 97.485 -79.025 45.786
91617 97.493 -79.023 45.770
91618 97.501 -79.020 45.752
91619 97.509 -79.024 45.737
91620 97.518 -79.061 45.721
91621 97.539 -79.075 45.705
91622 97.560 -79.087 45.690
91623 97.589 -79.113 45.671
91624 97.603 -79.128 45.651
91625 97.598 -79.128 45.629
91626 97.593 -79.127 45.606
91627 97.589 -79.126 45.586
91628 97.584 -79.126 45.567
91629 97.579 -79.125 45.549
91630 97.574 -79.125 45.533
91631 97.570 -79.124 45.517
91632 97.565 -79.124 45.500
91633 97.560 -79.123 45.484
91634 97.556 -79.122 45.469
91635 97.551 -79.122 45.454
91636 97.573 -79.137 45.440
91637 97.613 -79.161 45.427
91638 97.653 -79.184 45.414
91639 97.692 -79.208 45.402
91640 97.732 -79.232 45.394
91641 97.756 -79.245 45.383
91642 97.773 -79.253 45.372
91643 97.790 -79.262 45.360
91644 97.806 -79.270 45.350
91645 97.823 -79.278 45.340
91646 97.840 -79.287 45.330
91647 97.856 -79.296 45.321
91648 97.873 -79.304 45.313
91649 97.889 -79.313 45.305
91650 97.901 -79.328 45.296
91651 97.918 -79.342 45.289
91652 97.941 -79.352 45.282
91653 97.964 -79.362 45.275
91654 97.987 -79.372 45.269
91655 98.010 -79.382 45.262
91656 98.033 -79.391 45.254
91657 98.057 -79.400 45.246
91658 98.082 -79.407 45.237
91659 98.106 -79.415 45.227
91660 98.131 -79.411 45.217
91661 98.155 -79.407 45.208
91662 98.187 -79.403 45.198
91663 98.242 -79.397 45.188
91664 98.275 -79.399 45.178
91665 98.301 -79.402 45.170
91666 98.327 -79.406 45.162
91667 98.356 -79.412 45.155
91668 98.388 -79.424 45.147
91669 98.421 -79.435 45.140
91670 98.454 -79.447 45.134
91671 98.487 -79.459 45.128
91672 98.519 -79.471 45.123
91673 98.552 -79.483 45.118
91674 98.585 -79.494 45.113
91675 98.618 -79.507 45.105
91676 98.650 -79.518 45.099
91677 98.684 -79.530 45.094
91678 98.720 -79.542 45.087
91679 98.726 -79.546 45.080
91680 98.727 -79.548 45.073
91681 98.699 -79.616 45.067
91682 98.693 -79.645 45.058
91683 98.715 -79.629 45.048
91684 98.736 -79.612 45.038
91685 98.758 -79.595 45.030
91686 98.779 -79.591 45.021
91687 98.801 -79.607 45.012
91688 98.823 -79.622 45.002
91689 98.858 -79.628 44.989
91690 98.899 -79.629 44.978
91691 98.940 -79.630 44.966
91692 98.982 -79.631 44.953
91693 99.023 -79.632 44.943
91694 99.064 -79.633 44.932
91695 99.105 -79.634 44.921
91696 99.146 -79.635 44.911
91697 99.188 -79.636 44.902
91698 99.215 -79.636 44.891
91699 99.229 -79.648 44.880
91700 99.243 -79.663 44.869
91701 99.248 -79.677 44.859
91702 99.251 -79.692 44.849
91703 99.255 -79.706 44.838
91704 99.258 -79.720 44.828
91705 99.262 -79.735 44.818
91706 99.282 -79.742 44.807
91707 99.305 -79.748 44.796
91708 99.329 -79.755 44.786
91709 99.353 -79.761 44.777
91710 99.377 -79.767 44.766
91711 99.400 -79.773 44.756
91712 99.415 -79.767 44.746
91713 99.430 -79.761 44.736
91714 99.445 -79.755 44.727
91715 99.461 -79.749 44.718
91716 99.476 -79.744 44.709
91717 99.491 -79.738 44.700
91718 99.506 -79.732 44.691
91719 99.521 -79.726 44.682
91720 99.534 -79.724 44.677
91721 99.533 -79.752 44.669
91722 99.532 -79.780 44.661
91723 99.530 -79.808 44.655
91724 99.529 -79.836 44.646
91725 99.527 -79.864 44.638
91726 99.540 -79.890 44.630
91727 99.559 -79.915 44.624
91728 99.579 -79.939 44.617
91729 99.598 -79.964 44.610
91730 99.618 -79.989 44.603
91731 99.638 -80.013 44.595
91732 99.657 -80.038 44.588
91733 99.677 -80.062 44.582
91734 99.697 -80.087 44.575
91735 99.716 -80.112 44.568
91736 99.751 -80.149 44.560
91737 99.789 -80.190 44.551
91738 99.815 -80.225 44.541
91739 99.838 -80.257 44.532
91740 99.861 -80.290 44.523
91741 99.887 -80.321 44.512
91742 99.922 -80.347 44.501
91743 99.976 -80.355 44.490
91744 100.028 -80.363 44.478
91745 100.075 -80.364 44.465
91746 100.096 -80.412 44.453
91747 100.132 -80.439 44.439
91748 100.177 -80.455 44.425
91749 100.221 -80.472 44.411
91750 100.266 -80.488 44.397
91751 100.311 -80.504 44.382
91752 100.355 -80.521 44.367
91753 100.400 -80.537 44.352
91754 100.444 -80.554 44.337
91755 100.489 -80.569 44.321
91756 100.533 -80.586 44.305
91757 100.578 -80.603 44.290
91758 100.622 -80.619 44.275
91759 100.667 -80.635 44.260
91760 100.711 -80.651 44.245
91761 100.756 -80.668 44.231
91762 100.800 -80.685 44.217
91763 100.845 -80.701 44.202
91764 100.890 -80.717 44.188
91765 100.933 -80.735 44.175
91766 100.974 -80.757 44.161
91767 101.014 -80.779 44.147
92046 93.588 -78.643 53.644
92047 93.598 -78.640 53.631
92048 93.600 -78.633 53.617
92049 93.598 -78.623 53.603
92050 93.596 -78.614 53.590
92051 93.595 -78.605 53.575
92052 93.594 -78.596 53.562
92053 93.595 -78.586 53.548
92054 93.613 -78.570 53.534
92055 93.635 -78.555 53.520
92056 93.658 -78.540 53.505
92057 93.680 -78.525 53.490
92058 93.696 -78.514 53.474
92059 93.711 -78.507 53.456
92060 93.726 -78.500 53.439
92061 93.742 -78.492 53.421
92062 93.757 -78.485 53.403
92063 93.772 -78.478 53.385
92064 93.787 -78.473 53.366
92065 93.801 -78.468 53.347
92066 93.816 -78.463 53.327
92067 93.831 -78.457 53.308
92068 93.845 -78.452 53.288
92069 93.853 -78.442 53.269
92070 93.852 -78.425 53.250
92071 93.850 -78.407 53.232
92072 93.849 -78.391 53.212
92073 93.847 -78.373 53.193
92074 93.845 -78.357 53.175
92075 93.844 -78.339 53.158
92076 93.873 -78.326 53.140
92077 93.903 -78.319 53.123
92078 93.917 -78.328 53.106
92079 93.930 -78.337 53.089
92080 93.944 -78.345 53.073
92081 93.957 -78.354 53.059
92082 93.971 -78.363 53.045
92083 93.984 -78.371 53.028
92084 94.007 -78.381 53.011
92085 94.024 -78.393 52.994
92086 94.032 -78.400 52.974
92087 94.044 -78.405 52.956
92088 94.055 -78.410 52.938
92089 94.067 -78.415 52.923
92090 94.079 -78.420 52.909
92091 94.089 -78.425 52.896
92092 94.099 -78.429 52.883
92093 94.111 -78.430 52.871
92094 94.124 -78.429 52.859
92095 94.136 -78.428 52.848
92096 94.149 -78.426 52.837
92097 94.152 -78.436 52.825
92098 94.153 -78.448 52.813
92099 94.157 -78.452 52.800
92100 94.161 -78.456 52.784
92101 94.165 -78.460 52.769
92102 94.169 -78.464 52.754
92103 94.173 -78.469 52.739
92104 94.178 -78.473 52.725
92105 94.197 -78.470 52.710
92106 94.219 -78.466 52.695
92107 94.241 -78.462 52.679
92108 94.263 -78.458 52.664
92109 94.281 -78.460 52.649
92110 94.298 -78.466 52.633
92111 94.318 -78.466 52.616
92112 94.334 -78.468 52.599
92113 94.344 -78.482 52.581
92114 94.354 -78.495 52.558
92115 94.364 -78.508 52.534
92116 94.373 -78.521 52.508
92117 94.384 -78.535 52.481
92118 94.394 -78.548 52.453
92119 94.403 -78.561 52.425
92120 94.415 -78.562 52.397
92121 94.429 -78.556 52.368
92122 94.443 -78.549 52.336
92123 94.457 -78.543 52.303
92124 94.470 -78.536 52.268
92125 94.488 -78.526 52.236
92126 94.512 -78.510 52.203
92127 94.537 -78.493 52.171
92128 94.561 -78.476 52.137
92129 94.574 -78.469 52.101
92130 94.585 -78.468 52.066
92131 94.598 -78.470 52.030
92132 94.610 -78.472 51.992
92133 94.623 -78.474 51.952
92134 94.639 -78.476 51.909
92135 94.655 -78.477 51.862
92136 94.672 -78.479 51.814
92137 94.683 -78.477 51.764
92138 94.693 -78.476 51.713
92139 94.704 -78.475 51.659
92140 94.720 -78.479 51.603
92141 94.736 -78.482 51.548
92142 94.751 -78.486 51.494
92143 94.767 -78.490 51.442
92144 94.784 -78.489 51.390
92145 94.802 -78.483 51.339
92146 94.816 -78.493 51.288
92147 94.825 -78.495 51.237
92148 94.835 -78.496 51.188
92149 94.845 -78.498 51.140
92150 94.854 -78.499 51.095
92151 94.864 -78.501 51.052
92152 94.874 -78.498 51.011
92153 94.884 -78.482 50.970
92154 94.894 -78.466 50.929
92155 94.903 -78.450 50.889
92156 94.913 -78.433 50.844
92157 94.923 -78.417 50.798
92158 94.933 -78.401 50.750
92159 94.943 -78.385 50.701
92160 94.954 -78.368 50.650
92161 94.967 -78.351 50.600
92162 94.981 -78.332 50.550
92163 94.995 -78.315 50.502
92164 95.010 -78.297 50.456
92165 95.027 -78.288 50.412
92166 95.047 -78.294 50.367
92167 95.067 -78.299 50.324
92168 95.088 -78.304 50.281
92169 95.109 -78.309 50.239
92170 95.129 -78.314 50.197
92171 95.149 -78.319 50.158
92172 95.170 -78.325 50.119
92173 95.191 -78.331 50.082
92174 95.239 -78.376 50.042
92175 95.259 -78.380 50.001
92176 95.269 -78.369 49.960
92177 95.280 -78.357 49.921
92178 95.290 -78.352 49.883
92179 95.300 -78.354 49.846
92180 95.309 -78.357 49.808
92181 95.327 -78.366 49.773
92182 95.351 -78.378 49.738
92183 95.375 -78.390 49.703
92184 95.399 -78.402 49.667
92185 95.423 -78.413 49.630
92186 95.447 -78.426 49.592
92187 95.471 -78.438 49.556
92188 95.495 -78.450 49.520
92189 95.519 -78.461 49.484
92190 95.543 -78.473 49.448
92191 95.567 -78.485 49.414
92192 95.576 -78.487 49.379
92193 95.581 -78.486 49.345
92194 95.586 -78.486 49.313
92195 95.591 -78.485 49.282
92196 95.596 -78.485 49.250
92197 95.601 -78.485 49.216
92198 95.606 -78.485 49.182
92199 95.611 -78.484 49.147
92200 95.616 -78.484 49.110
92201 95.621 -78.483 49.072
92202 95.627 -78.484 49.034
92203 95.636 -78.492 48.996
92204 95.645 -78.498 48.956
92205 95.644 -78.498 48.918
92206 95.640 -78.494 48.880
92207 95.627 -78.492 48.841
92208 95.601 -78.492 48.802
92209 95.574 -78.492 48.761
92210 95.547 -78.492 48.718
92211 95.563 -78.492 48.675
92212 95.588 -78.492 48.633
92213 95.613 -78.493 48.592
92214 95.629 -78.497 48.552
92215 95.644 -78.500 48.511
92216 95.659 -78.504 48.472
92217 95.674 -78.507 48.433
92218 95.689 -78.511 48.394
92219 95.704 -78.516 48.356
92220 95.726 -78.516 48.317
92221 95.749 -78.517 48.278
92222 95.770 -78.521 48.239
92223 95.792 -78.525 48.198
92224 95.812 -78.523 48.157
92225 95.833 -78.523 48.115
92226 95.854 -78.523 48.072
92227 95.877 -78.524 48.031
92228 95.893 -78.525 47.990
92229 95.903 -78.524 47.949
92230 95.919 -78.524 47.908
92231 95.935 -78.524 47.868
92232 95.950 -78.526 47.828
92233 95.967 -78.529 47.789
92234 95.985 -78.533 47.750
92235 96.002 -78.532 47.711
92236 96.018 -78.529 47.672
92237 96.031 -78.529 47.633
92238 96.042 -78.529 47.595
92239 96.059 -78.545 47.558
92240 96.077 -78.562 47.519
92241 96.094 -78.579 47.481
92242 96.111 -78.595 47.442
92243 96.126 -78.609 47.404
92244 96.137 -78.620 47.368
92245 96.149 -78.636 47.332
92246 96.158 -78.670 47.296
92247 96.176 -78.688 47.264
92248 96.202 -78.692 47.233
92249 96.229 -78.696 47.203
92250 96.256 -78.701 47.174
92251 96.282 -78.705 47.144
92252 96.309 -78.709 47.112
92253 96.325 -78.712 47.081
92254 96.338 -78.716 47.052
92255 96.350 -78.719 47.023
92256 96.364 -78.722 46.994
92257 96.376 -78.726 46.965
92258 96.389 -78.729 46.936
92259 96.402 -78.733 46.907
92260 96.394 -78.718 46.877
92261 96.335 -78.664 46.849
92262 96.370 -78.661 46.822
92263 96.411 -78.663 46.795
92264 96.453 -78.664 46.768
92265 96.496 -78.666 46.742
92266 96.537 -78.667 46.717
92267 96.584 -78.675 46.695
92268 96.633 -78.685 46.673
92269 96.682 -78.695 46.652
92270 96.732 -78.705 46.632
92271 96.781 -78.714 46.610
92272 96.812 -78.724 46.589
92273 96.821 -78.736 46.568
92274 96.841 -78.746 46.545
92275 96.862 -78.757 46.525
92276 96.882 -78.767 46.505
92277 96.903 -78.777 46.484
92278 96.918 -78.790 46.460
92279 96.918 -78.808 46.436
92280 96.919 -78.827 46.412
92281 96.919 -78.846 46.391
92282 96.930 -78.855 46.374
92283 96.930 -78.847 46.357
92284 96.926 -78.863 46.340
92285 96.921 -78.879 46.322
92286 96.924 -78.894 46.303
92287 96.942 -78.904 46.284
92288 96.968 -78.909 46.266
92289 96.994 -78.915 46.249
92290 97.019 -78.920 46.232
92291 97.045 -78.925 46.216
92292 97.071 -78.930 46.200
92293 97.096 -78.936 46.186
92294 97.122 -78.941 46.170
92295 97.148 -78.946 46.154
92296 97.176 -78.956 46.142
92297 97.205 -78.967 46.130
92298 97.239 -78.982 46.117
92299 97.262 -78.990 46.101
92300 97.285 -78.999 46.086
92301 97.308 -79.008 46.070
92302 97.331 -79.016 46.054
92303 97.354 -79.024 46.039
92304 97.376 -79.033 46.023
92305 97.393 -79.035 46.004
92306 97.401 -79.033 45.981
92307 97.409 -79.031 45.961
92308 97.418 -79.028 45.941
92309 97.426 -79.026 45.921
92310 97.434 -79.024 45.905
92311 97.442 -79.021 45.890
92312 97.451 -79.019 45.875
92313 97.459 -79.017 45.862
92314 97.467 -79.015 45.846
92315 97.475 -79.012 45.830
92316 97.484 -79.009 45.814
92317 97.492 -79.007 45.802
92318 97.500 -79.005 45.787
92319 97.509 -79.002 45.771
92320 97.517 -79.000 45.756
92321 97.519 -79.040 45.742
92322 97.538 -79.059 45.726
92323 97.561 -79.075 45.708
92324 97.590 -79.103 45.690
92325 97.601 -79.114 45.669
92326 97.596 -79.113 45.644
92327 97.591 -79.113 45.621
92328 97.586 -79.112 45.601
92329 97.582 -79.112 45.581
92330 97.577 -79.112 45.563
92331 97.572 -79.111 45.547
92332 97.567 -79.110 45.531
92333 97.563 -79.109 45.515
92334 97.558 -79.109 45.497
92335 97.560 -79.106 45.481
92336 97.565 -79.106 45.466
92337 97.601 -79.129 45.452
92338 97.637 -79.153 45.438
92339 97.672 -79.175 45.425
92340 97.708 -79.199 45.414
92341 97.739 -79.219 45.404
92342 97.756 -79.227 45.393
92343 97.773 -79.236 45.382
92344 97.790 -79.244 45.369
92345 97.806 -79.253 45.359
92346 97.823 -79.261 45.349
92347 97.840 -79.269 45.339
92348 97.857 -79.278 45.330
92349 97.873 -79.287 45.322
92350 97.890 -79.295 45.314
92351 97.902 -79.310 45.305
92352 97.925 -79.319 45.299
92353 97.948 -79.330 45.292
92354 97.971 -79.340 45.285
92355 97.995 -79.350 45.279
92356 98.017 -79.360 45.272
92357 98.040 -79.369 45.265
92358 98.064 -79.379 45.258
92359 98.089 -79.385 45.249
92360 98.113 -79.387 45.240
92361 98.138 -79.382 45.230
92362 98.162 -79.378 45.220
92363 98.199 -79.373 45.210
92364 98.249 -79.370 45.200
92365 98.275 -79.373 45.191
92366 98.301 -79.376 45.184
92367 98.327 -79.380 45.176
92368 98.354 -79.384 45.168
92369 98.386 -79.396 45.160
92370 98.420 -79.407 45.153
92371 98.452 -79.419 45.146
92372 98.485 -79.432 45.141
92373 98.518 -79.443 45.136
92374 98.551 -79.455 45.132
92375 98.583 -79.466 45.127
92376 98.616 -79.479 45.118
92377 98.649 -79.490 45.111
92378 98.685 -79.503 45.105
92379 98.722 -79.515 45.098
92380 98.740 -79.522 45.090
92381 98.740 -79.524 45.082
92382 98.741 -79.526 45.076
92383 98.741 -79.529 45.068
92384 98.731 -79.562 45.058
92385 98.753 -79.545 45.047
92386 98.775 -79.554 45.036
92387 98.797 -79.570 45.026
92388 98.819 -79.585 45.017
92389 98.841 -79.601 45.009
92390 98.870 -79.611 44.998
92391 98.912 -79.612 44.987
92392 98.953 -79.613 44.976
92393 98.994 -79.614 44.963
92394 99.035 -79.615 44.953
92395 99.076 -79.616 44.942
92396 99.118 -79.617 44.931
92397 99.159 -79.618 44.922
92398 99.201 -79.614 44.912
92399 99.211 -79.620 44.901
92400 99.224 -79.632 44.890
92401 99.239 -79.646 44.879
92402 99.254 -79.661 44.868
92403 99.262 -79.675 44.858
92404 99.265 -79.689 44.848
92405 99.269 -79.704 44.837
92406 99.272 -79.718 44.827
92407 99.286 -79.729 44.816
92408 99.309 -79.735 44.804
92409 99.333 -79.741 44.794
92410 99.357 -79.747 44.784
92411 99.380 -79.754 44.773
92412 99.397 -79.749 44.763
92413 99.412 -79.743 44.754
92414 99.427 -79.738 44.744
92415 99.442 -79.732 44.735
92416 99.457 -79.726 44.726
92417 99.472 -79.720 44.717
92418 99.487 -79.714 44.706
92419 99.502 -79.708 44.697
92420 99.517 -79.703 44.688
92421 99.529 -79.704 44.682
92422 99.528 -79.732 44.674
92423 99.527 -79.760 44.666
92424 99.525 -79.788 44.660
92425 99.524 -79.816 44.652
92426 99.542 -79.841 44.644
92427 99.561 -79.865 44.637
92428 99.581 -79.890 44.631
92429 99.600 -79.914 44.624
92430 99.620 -79.939 44.616
92431 99.640 -79.964 44.609
92432 99.659 -79.989 44.602
92433 99.679 -80.013 44.596
92434 99.699 -80.037 44.589
92435 99.718 -80.062 44.582
92436 99.738 -80.087 44.574
92437 99.774 -80.126 44.566
92438 99.810 -80.165 44.558
92439 99.833 -80.198 44.548
92440 99.856 -80.231 44.539
92441 99.880 -80.263 44.530
92442 99.908 -80.292 44.519
92443 99.931 -80.320 44.508
92444 99.967 -80.345 44.496
92445 100.030 -80.359 44.484
92446 100.082 -80.366 44.472
92447 100.129 -80.365 44.459
92448 100.150 -80.413 44.445
92449 100.179 -80.450 44.431
92450 100.224 -80.466 44.417
92451 100.268 -80.482 44.403
92452 100.313 -80.498 44.387
92453 100.357 -80.515 44.372
92454 100.402 -80.532 44.356
92455 100.446 -80.547 44.341
92456 100.491 -80.564 44.326
92457 100.536 -80.580 44.310
92458 100.580 -80.597 44.293
92459 100.625 -80.613 44.278
92460 100.670 -80.629 44.263
92461 100.714 -80.646 44.247
92462 100.759 -80.662 44.233
92463 100.803 -80.679 44.219
92464 100.848 -80.695 44.204
92465 100.889 -80.715 44.190
92466 100.929 -80.737 44.176
92467 100.970 -80.759 44.163
92468 101.010 -80.781 44.150
92469 101.050 -80.803 44.137
92470 101.090 -80.826 44.123
92748 93.597 -78.629 53.633
92749 93.596 -78.620 53.620
92750 93.594 -78.611 53.605
92751 93.593 -78.601 53.591
92752 93.592 -78.592 53.575
92753 93.591 -78.582 53.561
92754 93.596 -78.570 53.546
92755 93.618 -78.554 53.531
92756 93.641 -78.540 53.516
92757 93.663 -78.525 53.501
92758 93.685 -78.510 53.485
92759 93.697 -78.502 53.468
92760 93.713 -78.495 53.451
92761 93.728 -78.487 53.433
92762 93.744 -78.479 53.416
92763 93.759 -78.473 53.397
92764 93.774 -78.465 53.378
92765 93.790 -78.457 53.359
92766 93.804 -78.453 53.341
92767 93.819 -78.448 53.321
92768 93.834 -78.442 53.302
92769 93.848 -78.437 53.283
92770 93.863 -78.432 53.263
92771 93.864 -78.416 53.245
92772 93.862 -78.399 53.226
92773 93.861 -78.382 53.206
92774 93.859 -78.365 53.187
92775 93.857 -78.347 53.169
92776 93.858 -78.331 53.151
92777 93.888 -78.317 53.134
92778 93.913 -78.316 53.117
92779 93.927 -78.326 53.100
92780 93.940 -78.334 53.083
92781 93.954 -78.342 53.067
92782 93.967 -78.351 53.052
92783 93.981 -78.360 53.037
92784 93.994 -78.368 53.020
92785 94.007 -78.377 53.004
92786 94.021 -78.385 52.987
92787 94.034 -78.391 52.969
92788 94.045 -78.396 52.952
92789 94.057 -78.401 52.936
92790 94.069 -78.406 52.922
92791 94.081 -78.410 52.909
92792 94.090 -78.415 52.898
92793 94.101 -78.420 52.887
92794 94.110 -78.424 52.875
92795 94.120 -78.428 52.866
92796 94.133 -78.427 52.856
92797 94.145 -78.426 52.846
92798 94.146 -78.438 52.834
92799 94.150 -78.442 52.823
92800 94.154 -78.446 52.811
92801 94.158 -78.450 52.795
92802 94.163 -78.454 52.779
92803 94.167 -78.458 52.763
92804 94.171 -78.463 52.749
92805 94.175 -78.467 52.735
92806 94.192 -78.466 52.721
92807 94.214 -78.461 52.706
92808 94.236 -78.457 52.691
92809 94.258 -78.454 52.677
92810 94.280 -78.449 52.664
92811 94.298 -78.451 52.649
92812 94.314 -78.457 52.632
92813 94.323 -78.470 52.613
92814 94.333 -78.484 52.593
92815 94.343 -78.497 52.569
92816 94.353 -78.510 52.544
92817 94.362 -78.523 52.517
92818 94.372 -78.537 52.489
92819 94.382 -78.550 52.460
92820 94.392 -78.564 52.432
92821 94.404 -78.566 52.405
92822 94.418 -78.560 52.375
92823 94.431 -78.553 52.342
92824 94.445 -78.546 52.307
92825 94.469 -78.531 52.272
92826 94.493 -78.514 52.239
92827 94.518 -78.498 52.207
92828 94.542 -78.482 52.174
92829 94.566 -78.465 52.139
92830 94.579 -78.457 52.103
92831 94.590 -78.451 52.066
92832 94.602 -78.454 52.030
92833 94.614 -78.456 51.992
92834 94.627 -78.457 51.952
92835 94.639 -78.460 51.910
92836 94.651 -78.461 51.863
92837 94.667 -78.463 51.814
92838 94.684 -78.464 51.764
92839 94.694 -78.463 51.714
92840 94.705 -78.463 51.660
92841 94.721 -78.466 51.605
92842 94.737 -78.470 51.549
92843 94.752 -78.473 51.495
92844 94.771 -78.465 51.443
92845 94.790 -78.457 51.391
92846 94.807 -78.458 51.341
92847 94.821 -78.471 51.290
92848 94.833 -78.476 51.241
92849 94.843 -78.478 51.194
92850 94.852 -78.479 51.148
92851 94.862 -78.481 51.105
92852 94.872 -78.482 51.063
92853 94.882 -78.466 51.022
92854 94.892 -78.450 50.981
92855 94.901 -78.434 50.938
92856 94.911 -78.417 50.894
92857 94.921 -78.401 50.848
92858 94.931 -78.385 50.800
92859 94.941 -78.369 50.751
92860 94.950 -78.353 50.700
92861 94.960 -78.336 50.649
92862 94.972 -78.319 50.599
92863 94.986 -78.301 50.549
92864 95.000 -78.284 50.501
92865 95.015 -78.266 50.455
92866 95.035 -78.269 50.410
92867 95.055 -78.273 50.366
92868 95.075 -78.279 50.323
92869 95.096 -78.284 50.281
92870 95.116 -78.289 50.239
92871 95.137 -78.294 50.198
92872 95.170 -78.317 50.157
92873 95.217 -78.363 50.118
92874 95.244 -78.376 50.079
92875 95.255 -78.366 50.040
92876 95.265 -78.354 50.000
92877 95.276 -78.343 49.959
92878 95.286 -78.332 49.921
92879 95.296 -78.330 49.883
92880 95.306 -78.333 49.846
92881 95.315 -78.335 49.810
92882 95.324 -78.338 49.776
92883 95.342 -78.347 49.742
92884 95.366 -78.359 49.708
92885 95.390 -78.370 49.672
92886 95.414 -78.382 49.634
92887 95.438 -78.395 49.597
92888 95.462 -78.407 49.560
92889 95.486 -78.418 49.523
92890 95.510 -78.430 49.487
92891 95.534 -78.442 49.452
92892 95.558 -78.454 49.418
92893 95.566 -78.456 49.384
92894 95.571 -78.455 49.351
92895 95.577 -78.455 49.319
92896 95.582 -78.454 49.289
92897 95.587 -78.454 49.258
92898 95.592 -78.454 49.225
92899 95.597 -78.454 49.190
92900 95.602 -78.453 49.155
92901 95.607 -78.452 49.118
92902 95.613 -78.454 49.081
92903 95.622 -78.460 49.044
92904 95.631 -78.467 49.007
92905 95.640 -78.475 48.969
92906 95.643 -78.477 48.931
92907 95.639 -78.473 48.893
92908 95.635 -78.470 48.856
92909 95.631 -78.467 48.817
92910 95.607 -78.466 48.775
92911 95.581 -78.466 48.732
92912 95.566 -78.466 48.689
92913 95.592 -78.467 48.648
92914 95.615 -78.467 48.607
92915 95.630 -78.471 48.566
92916 95.645 -78.475 48.526
92917 95.661 -78.479 48.487
92918 95.676 -78.482 48.448
92919 95.691 -78.485 48.409
92920 95.707 -78.489 48.371
92921 95.730 -78.491 48.332
92922 95.752 -78.495 48.292
92923 95.773 -78.500 48.252
92924 95.795 -78.503 48.211
92925 95.815 -78.501 48.169
92926 95.834 -78.499 48.127
92927 95.856 -78.499 48.085
92928 95.878 -78.500 48.043
92929 95.891 -78.500 48.002
92930 95.900 -78.499 47.960
92931 95.915 -78.499 47.920
92932 95.931 -78.500 47.879
92933 95.945 -78.502 47.837
92934 95.961 -78.506 47.797
92935 95.980 -78.510 47.758
92936 95.999 -78.514 47.719
92937 96.016 -78.515 47.680
92938 96.029 -78.515 47.641
92939 96.041 -78.515 47.602
92940 96.052 -78.515 47.565
92941 96.070 -78.532 47.527
92942 96.087 -78.548 47.490
92943 96.099 -78.560 47.453
92944 96.111 -78.570 47.417
92945 96.123 -78.580 47.383
92946 96.132 -78.611 47.349
92947 96.141 -78.646 47.313
92948 96.160 -78.663 47.281
92949 96.186 -78.667 47.252
92950 96.213 -78.671 47.223
92951 96.240 -78.676 47.192
92952 96.266 -78.680 47.162
92953 96.293 -78.684 47.132
92954 96.319 -78.688 47.101
92955 96.345 -78.692 47.074
92956 96.363 -78.695 47.044
92957 96.376 -78.698 47.014
92958 96.388 -78.702 46.987
92959 96.401 -78.705 46.958
92960 96.414 -78.709 46.928
92961 96.406 -78.695 46.897
92962 96.347 -78.641 46.870
92963 96.353 -78.622 46.843
92964 96.395 -78.623 46.815
92965 96.443 -78.632 46.787
92966 96.492 -78.642 46.760
92967 96.541 -78.651 46.734
92968 96.590 -78.661 46.710
92969 96.640 -78.671 46.688
92970 96.688 -78.681 46.666
92971 96.738 -78.691 46.645
92972 96.787 -78.701 46.624
92973 96.803 -78.712 46.604
92974 96.808 -78.723 46.582
92975 96.828 -78.733 46.560
92976 96.849 -78.744 46.539
92977 96.869 -78.755 46.518
92978 96.889 -78.765 46.496
92979 96.898 -78.780 46.473
92980 96.898 -78.799 46.450
92981 96.899 -78.817 46.427
92982 96.907 -78.833 46.405
92983 96.917 -78.843 46.387
92984 96.927 -78.854 46.369
92985 96.927 -78.850 46.351
92986 96.923 -78.866 46.333
92987 96.922 -78.881 46.315
92988 96.941 -78.891 46.297
92989 96.966 -78.896 46.278
92990 96.992 -78.902 46.261
92991 97.018 -78.907 46.245
92992 97.043 -78.912 46.229
92993 97.069 -78.918 46.213
92994 97.095 -78.923 46.197
92995 97.123 -78.933 46.181
92996 97.152 -78.944 46.165
92997 97.181 -78.955 46.153
92998 97.210 -78.965 46.141
92999 97.241 -78.978 46.128
93000 97.277 -78.995 46.114
93001 97.300 -79.003 46.098
93002 97.322 -79.012 46.082
93003 97.345 -79.021 46.066
93004 97.368 -79.029 46.051
93005 97.386 -79.031 46.034
93006 97.400 -79.029 46.016
93007 97.408 -79.024 45.997
93008 97.417 -79.018 45.977
93009 97.425 -79.012 45.956
93010 97.433 -79.009 45.936
93011 97.442 -79.006 45.920
93012 97.450 -79.003 45.904
93013 97.458 -79.001 45.888
93014 97.466 -78.999 45.874
93015 97.474 -78.996 45.858
93016 97.483 -78.994 45.843
93017 97.491 -78.992 45.828
93018 97.500 -78.990 45.815
93019 97.508 -78.987 45.801
93020 97.516 -78.985 45.786
93021 97.524 -78.983 45.772
93022 97.529 -79.005 45.758
93023 97.537 -79.044 45.741
93024 97.562 -79.065 45.723
93025 97.592 -79.092 45.704
93026 97.598 -79.100 45.682
93027 97.594 -79.100 45.658
93028 97.589 -79.099 45.636
93029 97.584 -79.099 45.615
93030 97.579 -79.098 45.596
93031 97.575 -79.097 45.577
93032 97.570 -79.096 45.561
93033 97.573 -79.094 45.545
93034 97.575 -79.091 45.528
93035 97.578 -79.088 45.510
93036 97.580 -79.086 45.494
93037 97.605 -79.100 45.479
93038 97.641 -79.123 45.464
93039 97.676 -79.147 45.450
93040 97.712 -79.169 45.437
93041 97.742 -79.189 45.425
93042 97.749 -79.194 45.414
93043 97.754 -79.206 45.403
93044 97.760 -79.219 45.392
93045 97.779 -79.227 45.379
93046 97.798 -79.235 45.368
93047 97.817 -79.244 45.358
93048 97.836 -79.252 45.349
93049 97.855 -79.260 45.340
93050 97.873 -79.269 45.332
93051 97.890 -79.278 45.324
93052 97.910 -79.288 45.316
93053 97.932 -79.297 45.310
93054 97.955 -79.307 45.303
93055 97.979 -79.317 45.296
93056 98.002 -79.327 45.289
93057 98.025 -79.337 45.282
93058 98.048 -79.347 45.276
93059 98.071 -79.356 45.269
93060 98.096 -79.362 45.260
93061 98.120 -79.357 45.251
93062 98.145 -79.353 45.241
93063 98.169 -79.349 45.230
93064 98.210 -79.344 45.220
93065 98.249 -79.344 45.210
93066 98.275 -79.347 45.203
93067 98.301 -79.350 45.196
93068 98.327 -79.354 45.188
93069 98.353 -79.357 45.180
93070 98.385 -79.368 45.173
93071 98.418 -79.380 45.166
93072 98.450 -79.391 45.159
93073 98.483 -79.404 45.153
93074 98.516 -79.415 45.147
93075 98.549 -79.427 45.141
93076 98.581 -79.439 45.134
93077 98.614 -79.451 45.124
93078 98.650 -79.463 45.115
93079 98.687 -79.475 45.108
93080 98.723 -79.487 45.102
93081 98.754 -79.498 45.095
93082 98.754 -79.500 45.086
93083 98.755 -79.502 45.078
93084 98.755 -79.504 45.071
93085 98.756 -79.507 45.061
93086 98.771 -79.519 45.051
93087 98.792 -79.534 45.039
93088 98.814 -79.549 45.030
93089 98.837 -79.565 45.021
93090 98.858 -79.580 45.014
93091 98.883 -79.594 45.004
93092 98.924 -79.595 44.994
93093 98.966 -79.596 44.983
93094 99.007 -79.597 44.972
93095 99.048 -79.598 44.962
93096 99.089 -79.599 44.951
93097 99.130 -79.600 44.941
93098 99.173 -79.599 44.931
93099 99.196 -79.598 44.920
93100 99.206 -79.604 44.908
93101 99.219 -79.615 44.899
93102 99.234 -79.629 44.888
93103 99.249 -79.644 44.877
93104 99.264 -79.658 44.868
93105 99.276 -79.673 44.857
93106 99.279 -79.688 44.847
93107 99.283 -79.701 44.836
93108 99.290 -79.715 44.824
93109 99.313 -79.721 44.812
93110 99.337 -79.727 44.801
93111 99.361 -79.733 44.790
93112 99.378 -79.732 44.781
93113 99.394 -79.726 44.773
93114 99.409 -79.720 44.764
93115 99.424 -79.714 44.754
93116 99.439 -79.708 44.745
93117 99.454 -79.702 44.735
93118 99.469 -79.697 44.726
93119 99.484 -79.691 44.715
93120 99.499 -79.685 44.706
93121 99.514 -79.679 44.698
93122 99.524 -79.683 44.692
93123 99.523 -79.711 44.684
93124 99.522 -79.739 44.675
93125 99.524 -79.767 44.667
93126 99.544 -79.792 44.659
93127 99.563 -79.816 44.652
93128 99.583 -79.841 44.645
93129 99.603 -79.865 44.640
93130 99.622 -79.890 44.632
93131 99.642 -79.914 44.624
93132 99.661 -79.939 44.617
93133 99.681 -79.964 44.611
93134 99.701 -79.988 44.604
93135 99.720 -80.013 44.596
93136 99.740 -80.037 44.590
93137 99.760 -80.062 44.582
93138 99.797 -80.102 44.574
93139 99.828 -80.140 44.565
93140 99.851 -80.172 44.556
93141 99.874 -80.205 44.547
93142 99.903 -80.234 44.538
93143 99.930 -80.262 44.527
93144 99.952 -80.290 44.516
93145 99.976 -80.319 44.504
93146 100.011 -80.344 44.492
93147 100.080 -80.362 44.479
93148 100.135 -80.371 44.466
93149 100.184 -80.368 44.453
93150 100.204 -80.414 44.439
93151 100.226 -80.460 44.424
93152 100.271 -80.476 44.409
93153 100.315 -80.493 44.394
93154 100.360 -80.509 44.378
93155 100.404 -80.526 44.363
93156 100.449 -80.541 44.346
93157 100.493 -80.558 44.331
93158 100.538 -80.575 44.315
93159 100.582 -80.591 44.299
93160 100.627 -80.607 44.283
93161 100.671 -80.623 44.267
93162 100.716 -80.640 44.252
93163 100.761 -80.656 44.237
93164 100.805 -80.673 44.223
93165 100.845 -80.695 44.208
93166 100.885 -80.717 44.193
93167 100.925 -80.739 44.179
93168 100.966 -80.761 44.165
93169 101.006 -80.784 44.152
93170 101.046 -80.806 44.139
93171 101.087 -80.828 44.125
93449 93.594 -78.616 53.636
93450 93.592 -78.607 53.621
93451 93.591 -78.598 53.606
93452 93.590 -78.588 53.591
93453 93.588 -78.579 53.576
93454 93.587 -78.570 53.560
93455 93.602 -78.554 53.545
93456 93.624 -78.539 53.529
93457 93.646 -78.524 53.513
93458 93.669 -78.509 53.497
93459 93.687 -78.497 53.481
93460 93.699 -78.489 53.464
93461 93.714 -78.482 53.446
93462 93.730 -78.474 53.428
93463 93.746 -78.467 53.411
93464 93.761 -78.460 53.392
93465 93.776 -78.452 53.373
93466 93.791 -78.445 53.354
93467 93.807 -78.438 53.335
93468 93.822 -78.432 53.315
93469 93.836 -78.427 53.296
93470 93.851 -78.422 53.278
93471 93.865 -78.416 53.259
93472 93.876 -78.408 53.240
93473 93.874 -78.391 53.221
93474 93.872 -78.374 53.200
93475 93.871 -78.357 53.182
93476 93.869 -78.339 53.164
93477 93.872 -78.322 53.146
93478 93.902 -78.309 53.129
93479 93.923 -78.314 53.112
93480 93.937 -78.322 53.095
93481 93.950 -78.331 53.078
93482 93.964 -78.340 53.062
93483 93.977 -78.348 53.046
93484 93.991 -78.357 53.030
93485 94.004 -78.366 53.013
93486 94.017 -78.374 52.998
93487 94.026 -78.378 52.982
93488 94.036 -78.382 52.965
93489 94.047 -78.387 52.949
93490 94.059 -78.391 52.935
93491 94.071 -78.397 52.922
93492 94.082 -78.401 52.911
93493 94.092 -78.406 52.901
93494 94.102 -78.410 52.892
93495 94.112 -78.414 52.881
93496 94.122 -78.419 52.871
93497 94.131 -78.423 52.862
93498 94.140 -78.427 52.853
93499 94.144 -78.432 52.842
93500 94.148 -78.435 52.831
93501 94.152 -78.440 52.819
93502 94.156 -78.444 52.802
93503 94.160 -78.448 52.786
93504 94.165 -78.453 52.771
93505 94.169 -78.457 52.759
93506 94.173 -78.461 52.745
93507 94.186 -78.461 52.730
93508 94.208 -78.457 52.716
93509 94.230 -78.453 52.702
93510 94.253 -78.449 52.689
93511 94.274 -78.445 52.675
93512 94.295 -78.438 52.660
93513 94.304 -78.454 52.642
93514 94.313 -78.471 52.623
93515 94.322 -78.486 52.602
93516 94.332 -78.499 52.577
93517 94.342 -78.513 52.550
93518 94.351 -78.526 52.522
93519 94.361 -78.539 52.493
93520 94.371 -78.552 52.464
93521 94.381 -78.566 52.435
93522 94.393 -78.570 52.408
93523 94.406 -78.563 52.376
93524 94.424 -78.553 52.343
93525 94.449 -78.536 52.308
93526 94.473 -78.520 52.272
93527 94.497 -78.503 52.239
93528 94.522 -78.487 52.207
93529 94.546 -78.470 52.174
93530 94.571 -78.454 52.138
93531 94.584 -78.445 52.102
93532 94.595 -78.439 52.065
93533 94.607 -78.438 52.027
93534 94.619 -78.439 51.990
93535 94.631 -78.442 51.950
93536 94.643 -78.443 51.908
93537 94.655 -78.445 51.861
93538 94.668 -78.447 51.812
93539 94.680 -78.449 51.762
93540 94.695 -78.451 51.712
93541 94.706 -78.450 51.659
93542 94.722 -78.454 51.604
93543 94.740 -78.451 51.548
93544 94.759 -78.442 51.494
93545 94.777 -78.434 51.442
93546 94.796 -78.426 51.391
93547 94.812 -78.434 51.341
93548 94.826 -78.446 51.291
93549 94.840 -78.458 51.244
93550 94.850 -78.460 51.198
93551 94.860 -78.461 51.155
93552 94.870 -78.463 51.113
93553 94.880 -78.451 51.072
93554 94.890 -78.434 51.030
93555 94.899 -78.418 50.987
93556 94.909 -78.401 50.943
93557 94.919 -78.385 50.896
93558 94.928 -78.369 50.847
93559 94.938 -78.353 50.798
93560 94.948 -78.337 50.748
93561 94.958 -78.320 50.697
93562 94.968 -78.304 50.646
93563 94.978 -78.288 50.595
93564 94.991 -78.270 50.546
93565 95.005 -78.252 50.498
93566 95.022 -78.243 50.452
93567 95.043 -78.248 50.407
93568 95.063 -78.253 50.364
93569 95.084 -78.259 50.322
93570 95.104 -78.263 50.280
93571 95.148 -78.304 50.238
93572 95.196 -78.350 50.197
93573 95.229 -78.373 50.156
93574 95.240 -78.363 50.116
93575 95.250 -78.351 50.076
93576 95.261 -78.340 50.037
93577 95.271 -78.329 49.998
93578 95.282 -78.318 49.958
93579 95.293 -78.307 49.920
93580 95.302 -78.308 49.883
93581 95.312 -78.311 49.846
93582 95.321 -78.313 49.811
93583 95.330 -78.316 49.777
93584 95.339 -78.319 49.743
93585 95.357 -78.327 49.709
93586 95.381 -78.339 49.672
93587 95.405 -78.351 49.635
93588 95.429 -78.363 49.598
93589 95.453 -78.375 49.561
93590 95.477 -78.387 49.526
93591 95.501 -78.399 49.492
93592 95.525 -78.410 49.456
93593 95.549 -78.423 49.423
93594 95.557 -78.425 49.390
93595 95.562 -78.424 49.357
93596 95.567 -78.423 49.326
93597 95.572 -78.423 49.295
93598 95.578 -78.423 49.264
93599 95.583 -78.423 49.233
93600 95.588 -78.422 49.199
93601 95.593 -78.422 49.163
93602 95.599 -78.423 49.128
93603 95.608 -78.430 49.092
93604 95.617 -78.437 49.057
93605 95.626 -78.444 49.021
93606 95.635 -78.451 48.984
93607 95.642 -78.457 48.946
93608 95.638 -78.453 48.909
93609 95.634 -78.449 48.872
93610 95.630 -78.446 48.833
93611 95.626 -78.442 48.791
93612 95.614 -78.440 48.749
93613 95.588 -78.440 48.706
93614 95.595 -78.440 48.665
93615 95.617 -78.442 48.623
93616 95.632 -78.445 48.581
93617 95.647 -78.449 48.542
93618 95.662 -78.453 48.502
93619 95.677 -78.457 48.463
93620 95.692 -78.460 48.424
93621 95.711 -78.465 48.386
93622 95.733 -78.470 48.346
93623 95.755 -78.473 48.304
93624 95.777 -78.478 48.262
93625 95.798 -78.481 48.220
93626 95.818 -78.479 48.179
93627 95.837 -78.477 48.137
93628 95.857 -78.476 48.095
93629 95.879 -78.476 48.055
93630 95.890 -78.476 48.015
93631 95.898 -78.475 47.973
93632 95.911 -78.475 47.932
93633 95.926 -78.476 47.892
93634 95.940 -78.479 47.850
93635 95.956 -78.482 47.811
93636 95.975 -78.487 47.771
93637 95.993 -78.491 47.732
93638 96.011 -78.496 47.695
93639 96.023 -78.498 47.657
93640 96.036 -78.499 47.619
93641 96.049 -78.500 47.582
93642 96.061 -78.502 47.544
93643 96.073 -78.510 47.507
93644 96.085 -78.520 47.469
93645 96.097 -78.530 47.434
93646 96.107 -78.552 47.400
93647 96.116 -78.587 47.365
93648 96.124 -78.622 47.330
93649 96.144 -78.638 47.300
93650 96.170 -78.642 47.269
93651 96.197 -78.646 47.239
93652 96.224 -78.651 47.210
93653 96.250 -78.655 47.181
93654 96.276 -78.658 47.151
93655 96.303 -78.663 47.120
93656 96.329 -78.667 47.093
93657 96.356 -78.671 47.065
93658 96.382 -78.675 47.036
93659 96.400 -78.679 47.007
93660 96.413 -78.682 46.978
93661 96.426 -78.686 46.949
93662 96.418 -78.672 46.919
93663 96.360 -78.617 46.891
93664 96.350 -78.599 46.862
93665 96.400 -78.609 46.833
93666 96.449 -78.619 46.806
93667 96.498 -78.629 46.779
93668 96.547 -78.639 46.753
93669 96.596 -78.648 46.728
93670 96.645 -78.658 46.706
93671 96.694 -78.668 46.683
93672 96.744 -78.678 46.661
93673 96.793 -78.688 46.639
93674 96.793 -78.699 46.618
93675 96.795 -78.710 46.596
93676 96.815 -78.720 46.574
93677 96.835 -78.731 46.553
93678 96.856 -78.742 46.532
93679 96.876 -78.752 46.510
93680 96.879 -78.770 46.487
93681 96.879 -78.788 46.464
93682 96.883 -78.805 46.441
93683 96.894 -78.819 46.420
93684 96.904 -78.830 46.401
93685 96.915 -78.840 46.382
93686 96.925 -78.851 46.364
93687 96.924 -78.853 46.345
93688 96.920 -78.869 46.328
93689 96.939 -78.878 46.310
93690 96.965 -78.883 46.291
93691 96.991 -78.889 46.273
93692 97.016 -78.894 46.257
93693 97.042 -78.899 46.242
93694 97.071 -78.910 46.226
93695 97.099 -78.921 46.210
93696 97.128 -78.931 46.195
93697 97.157 -78.943 46.180
93698 97.186 -78.953 46.167
93699 97.215 -78.964 46.154
93700 97.244 -78.975 46.141
93701 97.288 -78.997 46.127
93702 97.314 -79.008 46.111
93703 97.337 -79.017 46.095
93704 97.360 -79.025 46.078
93705 97.376 -79.025 46.062
93706 97.392 -79.024 46.045
93707 97.407 -79.024 46.028
93708 97.416 -79.018 46.010
93709 97.425 -79.012 45.991
93710 97.433 -79.006 45.970
93711 97.441 -79.000 45.949
93712 97.449 -78.994 45.933
93713 97.457 -78.988 45.917
93714 97.465 -78.984 45.901
93715 97.474 -78.981 45.885
93716 97.482 -78.979 45.870
93717 97.490 -78.977 45.855
93718 97.499 -78.974 45.841
93719 97.507 -78.972 45.828
93720 97.515 -78.970 45.815
93721 97.523 -78.968 45.801
93722 97.532 -78.965 45.787
93723 97.539 -78.969 45.771
93724 97.540 -79.021 45.754
93725 97.564 -79.053 45.736
93726 97.593 -79.081 45.715
93727 97.596 -79.086 45.693
93728 97.591 -79.086 45.671
93729 97.587 -79.085 45.650
93730 97.584 -79.084 45.629
93731 97.586 -79.081 45.610
93732 97.589 -79.078 45.592
93733 97.591 -79.076 45.575
93734 97.594 -79.073 45.558
93735 97.596 -79.070 45.541
93736 97.599 -79.068 45.523
93737 97.609 -79.071 45.507
93738 97.645 -79.093 45.492
93739 97.680 -79.117 45.478
93740 97.716 -79.140 45.463
93741 97.747 -79.160 45.450
93742 97.754 -79.165 45.438
93743 97.761 -79.170 45.425
93744 97.768 -79.178 45.413
93745 97.772 -79.192 45.401
93746 97.776 -79.206 45.389
93747 97.781 -79.219 45.380
93748 97.800 -79.227 45.370
93749 97.819 -79.235 45.360
93750 97.838 -79.244 45.351
93751 97.857 -79.252 45.342
93752 97.875 -79.247 45.333
93753 97.893 -79.240 45.326
93754 97.911 -79.232 45.320
93755 97.946 -79.259 45.314
93756 97.981 -79.287 45.308
93757 98.009 -79.305 45.300
93758 98.032 -79.315 45.293
93759 98.055 -79.325 45.285
93760 98.079 -79.334 45.277
93761 98.103 -79.333 45.269
93762 98.128 -79.328 45.260
93763 98.152 -79.324 45.250
93764 98.176 -79.320 45.241
93765 98.222 -79.316 45.230
93766 98.249 -79.319 45.221
93767 98.275 -79.322 45.214
93768 98.301 -79.325 45.207
93769 98.327 -79.328 45.200
93770 98.353 -79.332 45.192
93771 98.383 -79.340 45.184
93772 98.416 -79.352 45.176
93773 98.449 -79.363 45.169
93774 98.482 -79.375 45.162
93775 98.514 -79.388 45.154
93776 98.547 -79.399 45.146
93777 98.580 -79.411 45.137
93778 98.615 -79.423 45.128
93779 98.652 -79.435 45.118
93780 98.688 -79.447 45.109
93781 98.725 -79.460 45.102
93782 98.762 -79.472 45.095
93783 98.768 -79.476 45.087
93784 98.768 -79.478 45.077
93785 98.769 -79.480 45.070
93786 98.769 -79.482 45.061
93787 98.787 -79.496 45.051
93788 98.810 -79.513 45.041
93789 98.832 -79.529 45.031
93790 98.854 -79.545 45.023
93791 98.876 -79.560 45.017
93792 98.898 -79.575 45.009
93793 98.937 -79.578 44.999
93794 98.978 -79.579 44.988
93795 99.019 -79.580 44.979
93796 99.060 -79.581 44.969
93797 99.101 -79.582 44.959
93798 99.143 -79.583 44.950
93799 99.182 -79.576 44.937
93800 99.192 -79.582 44.924
93801 99.201 -79.587 44.913
93802 99.214 -79.598 44.903
93803 99.229 -79.613 44.893
93804 99.244 -79.627 44.883
93805 99.259 -79.642 44.874
93806 99.274 -79.657 44.864
93807 99.289 -79.671 44.853
93808 99.293 -79.686 44.842
93809 99.296 -79.700 44.832
93810 99.317 -79.708 44.820
93811 99.341 -79.714 44.809
93812 99.360 -79.714 44.799
93813 99.375 -79.708 44.790
93814 99.391 -79.702 44.781
93815 99.406 -79.696 44.772
93816 99.421 -79.691 44.762
93817 99.436 -79.685 44.752
93818 99.451 -79.679 44.742
93819 99.466 -79.673 44.732
93820 99.481 -79.667 44.723
93821 99.496 -79.661 44.714
93822 99.511 -79.656 44.707
93823 99.520 -79.663 44.700
93824 99.519 -79.691 44.692
93825 99.525 -79.718 44.683
93826 99.545 -79.742 44.675
93827 99.565 -79.767 44.667
93828 99.584 -79.792 44.660
93829 99.604 -79.816 44.653
93830 99.624 -79.841 44.648
93831 99.644 -79.865 44.641
93832 99.663 -79.889 44.633
93833 99.683 -79.914 44.625
93834 99.703 -79.939 44.620
93835 99.722 -79.964 44.613
93836 99.742 -79.988 44.605
93837 99.762 -80.013 44.598
93838 99.782 -80.038 44.590
93839 99.820 -80.079 44.582
93840 99.846 -80.113 44.574
93841 99.869 -80.146 44.565
93842 99.896 -80.176 44.556
93843 99.924 -80.205 44.547
93844 99.951 -80.234 44.537
93845 99.974 -80.262 44.526
93846 99.997 -80.289 44.514
93847 100.020 -80.317 44.501
93848 100.055 -80.343 44.489
93849 100.125 -80.360 44.475
93850 100.189 -80.375 44.462
93851 100.238 -80.374 44.447
93852 100.258 -80.416 44.432
93853 100.278 -80.463 44.417
93854 100.318 -80.487 44.402
93855 100.362 -80.504 44.386
93856 100.407 -80.519 44.369
93857 100.451 -80.536 44.353
93858 100.496 -80.552 44.336
93859 100.540 -80.569 44.320
93860 100.585 -80.585 44.304
93861 100.629 -80.601 44.288
93862 100.674 -80.618 44.272
93863 100.718 -80.634 44.257
93864 100.760 -80.654 44.242
93865 100.800 -80.676 44.227
93866 100.841 -80.698 44.212
93867 100.881 -80.720 44.197
93868 100.921 -80.742 44.183
93869 100.962 -80.764 44.169
93870 101.002 -80.786 44.155
93871 101.042 -80.808 44.142
93872 101.082 -80.830 44.128
93873 101.122 -80.852 44.114
94143 93.531 -78.625 53.747
94144 93.541 -78.623 53.732
94145 93.551 -78.620 53.717
94146 93.560 -78.617 53.702
94147 93.570 -78.615 53.686
94148 93.579 -78.613 53.671
94149 93.589 -78.611 53.655
94150 93.590 -78.602 53.638
94151 93.589 -78.593 53.623
94152 93.588 -78.584 53.607
94153 93.586 -78.575 53.591
94154 93.585 -78.566 53.576
94155 93.589 -78.554 53.559
94156 93.607 -78.539 53.543
94157 93.629 -78.523 53.527
94158 93.652 -78.508 53.510
94159 93.674 -78.494 53.494
94160 93.688 -78.484 53.477
94161 93.701 -78.476 53.459
94162 93.716 -78.469 53.441
94163 93.732 -78.461 53.423
94164 93.747 -78.454 53.405
94165 93.763 -78.447 53.386
94166 93.778 -78.439 53.368
94167 93.793 -78.432 53.349
94168 93.809 -78.425 53.330
94169 93.824 -78.417 53.311
94170 93.839 -78.412 53.292
94171 93.854 -78.407 53.273
94172 93.868 -78.401 53.254
94173 93.883 -78.396 53.235
94174 93.887 -78.382 53.216
94175 93.885 -78.365 53.197
94176 93.883 -78.348 53.178
94177 93.882 -78.331 53.159
94178 93.887 -78.314 53.142
94179 93.920 -78.303 53.125
94180 93.933 -78.311 53.108
94181 93.946 -78.319 53.092
94182 93.960 -78.329 53.074
94183 93.974 -78.337 53.057
94184 93.987 -78.345 53.041
94185 94.000 -78.354 53.024
94186 94.013 -78.362 53.009
94187 94.021 -78.366 52.994
94188 94.029 -78.369 52.979
94189 94.037 -78.373 52.963
94190 94.049 -78.378 52.948
94191 94.061 -78.382 52.934
94192 94.073 -78.388 52.924
94193 94.084 -78.392 52.914
94194 94.094 -78.396 52.905
94195 94.104 -78.401 52.896
94196 94.113 -78.405 52.886
94197 94.123 -78.410 52.876
94198 94.132 -78.414 52.867
94199 94.136 -78.423 52.859
94200 94.141 -78.428 52.849
94201 94.146 -78.430 52.838
94202 94.150 -78.434 52.825
94203 94.154 -78.438 52.808
94204 94.158 -78.442 52.793
94205 94.162 -78.447 52.779
94206 94.166 -78.451 52.767
94207 94.171 -78.455 52.754
94208 94.181 -78.457 52.741
94209 94.203 -78.452 52.727
94210 94.225 -78.448 52.713
94211 94.247 -78.445 52.699
94212 94.266 -78.433 52.684
94213 94.286 -78.425 52.668
94214 94.300 -78.435 52.649
94215 94.309 -78.452 52.628
94216 94.317 -78.470 52.606
94217 94.326 -78.487 52.580
94218 94.334 -78.504 52.552
94219 94.343 -78.521 52.523
94220 94.351 -78.539 52.493
94221 94.360 -78.554 52.464
94222 94.370 -78.568 52.435
94223 94.382 -78.573 52.405
94224 94.405 -78.557 52.372
94225 94.429 -78.542 52.337
94226 94.454 -78.525 52.301
94227 94.478 -78.508 52.266
94228 94.502 -78.492 52.234
94229 94.527 -78.476 52.202
94230 94.551 -78.459 52.169
94231 94.575 -78.442 52.132
94232 94.590 -78.433 52.095
94233 94.601 -78.428 52.057
94234 94.611 -78.422 52.020
94235 94.623 -78.423 51.981
94236 94.635 -78.425 51.941
94237 94.647 -78.427 51.897
94238 94.660 -78.429 51.851
94239 94.672 -78.431 51.803
94240 94.684 -78.433 51.755
94241 94.696 -78.435 51.706
94242 94.708 -78.437 51.653
94243 94.726 -78.430 51.599
94244 94.745 -78.422 51.544
94245 94.764 -78.414 51.490
94246 94.782 -78.407 51.437
94247 94.801 -78.399 51.386
94248 94.817 -78.409 51.337
94249 94.831 -78.421 51.290
94250 94.845 -78.433 51.244
94251 94.858 -78.442 51.201
94252 94.868 -78.443 51.160
94253 94.878 -78.435 51.119
94254 94.888 -78.418 51.077
94255 94.897 -78.402 51.032
94256 94.907 -78.385 50.987
94257 94.917 -78.369 50.941
94258 94.926 -78.353 50.892
94259 94.936 -78.337 50.842
94260 94.946 -78.321 50.792
94261 94.956 -78.304 50.741
94262 94.966 -78.288 50.690
94263 94.975 -78.272 50.639
94264 94.985 -78.256 50.589
94265 94.996 -78.239 50.541
94266 95.011 -78.221 50.494
94267 95.030 -78.223 50.448
94268 95.051 -78.228 50.403
94269 95.079 -78.244 50.361
94270 95.126 -78.291 50.320
94271 95.174 -78.336 50.278
94272 95.214 -78.370 50.236
94273 95.225 -78.359 50.195
94274 95.235 -78.348 50.155
94275 95.246 -78.337 50.114
94276 95.257 -78.326 50.073
94277 95.267 -78.314 50.034
94278 95.278 -78.304 49.995
94279 95.288 -78.292 49.957
94280 95.299 -78.283 49.920
94281 95.308 -78.286 49.883
94282 95.317 -78.289 49.847
94283 95.327 -78.291 49.812
94284 95.336 -78.294 49.778
94285 95.345 -78.297 49.743
94286 95.355 -78.300 49.709
94287 95.372 -78.308 49.672
94288 95.396 -78.320 49.635
94289 95.420 -78.332 49.599
94290 95.444 -78.344 49.563
94291 95.468 -78.356 49.530
94292 95.492 -78.367 49.497
94293 95.516 -78.379 49.462
94294 95.540 -78.391 49.429
94295 95.548 -78.393 49.397
94296 95.553 -78.393 49.365
94297 95.558 -78.392 49.333
94298 95.563 -78.392 49.302
94299 95.569 -78.391 49.271
94300 95.574 -78.391 49.240
94301 95.579 -78.391 49.207
94302 95.585 -78.392 49.173
94303 95.594 -78.399 49.138
94304 95.603 -78.406 49.104
94305 95.612 -78.413 49.070
94306 95.621 -78.420 49.034
94307 95.631 -78.427 48.997
94308 95.640 -78.435 48.961
94309 95.637 -78.432 48.924
94310 95.633 -78.429 48.888
94311 95.629 -78.425 48.847
94312 95.625 -78.422 48.807
94313 95.621 -78.418 48.765
94314 95.617 -78.415 48.724
94315 95.598 -78.414 48.682
94316 95.618 -78.416 48.640
94317 95.633 -78.420 48.598
94318 95.649 -78.424 48.559
94319 95.665 -78.430 48.519
94320 95.681 -78.435 48.480
94321 95.697 -78.442 48.440
94322 95.712 -78.444 48.400
94323 95.736 -78.448 48.359
94324 95.758 -78.452 48.317
94325 95.780 -78.456 48.275
94326 95.801 -78.459 48.232
94327 95.821 -78.457 48.190
94328 95.840 -78.455 48.149
94329 95.860 -78.454 48.107
94330 95.879 -78.452 48.067
94331 95.888 -78.451 48.027
94332 95.897 -78.451 47.986
94333 95.907 -78.450 47.946
94334 95.920 -78.453 47.906
94335 95.935 -78.455 47.865
94336 95.951 -78.459 47.826
94337 95.969 -78.463 47.786
94338 95.985 -78.470 47.748
94339 95.999 -78.476 47.711
94340 96.012 -78.479 47.673
94341 96.025 -78.480 47.636
94342 96.037 -78.483 47.598
94343 96.050 -78.492 47.561
94344 96.062 -78.497 47.523
94345 96.073 -78.498 47.486
94346 96.083 -78.499 47.452
94347 96.091 -78.528 47.417
94348 96.099 -78.562 47.382
94349 96.108 -78.597 47.348
94350 96.128 -78.613 47.317
94351 96.154 -78.617 47.286
94352 96.181 -78.621 47.256
94353 96.207 -78.626 47.228
94354 96.234 -78.630 47.200
94355 96.260 -78.633 47.171
94356 96.287 -78.638 47.140
94357 96.313 -78.642 47.112
94358 96.340 -78.646 47.084
94359 96.366 -78.650 47.056
94360 96.393 -78.655 47.027
94361 96.419 -78.658 46.998
94362 96.438 -78.662 46.969
94363 96.431 -78.649 46.940
94364 96.413 -78.624 46.910
94365 96.409 -78.610 46.880
94366 96.405 -78.595 46.851
94367 96.455 -78.605 46.824
94368 96.504 -78.615 46.798
94369 96.553 -78.626 46.772
94370 96.602 -78.636 46.746
94371 96.651 -78.645 46.724
94372 96.700 -78.655 46.701
94373 96.749 -78.665 46.678
94374 96.783 -78.675 46.656
94375 96.784 -78.686 46.633
94376 96.784 -78.697 46.611
94377 96.802 -78.708 46.589
94378 96.822 -78.718 46.568
94379 96.842 -78.729 46.547
94380 96.859 -78.741 46.526
94381 96.859 -78.760 46.502
94382 96.860 -78.778 46.479
94383 96.871 -78.793 46.457
94384 96.882 -78.805 46.436
94385 96.892 -78.816 46.416
94386 96.902 -78.827 46.397
94387 96.912 -78.837 46.378
94388 96.923 -78.849 46.360
94389 96.921 -78.856 46.341
94390 96.938 -78.865 46.323
94391 96.963 -78.871 46.305
94392 96.989 -78.876 46.287
94393 97.018 -78.886 46.270
94394 97.046 -78.897 46.255
94395 97.075 -78.908 46.240
94396 97.104 -78.919 46.225
94397 97.133 -78.930 46.211
94398 97.162 -78.941 46.196
94399 97.191 -78.952 46.182
94400 97.220 -78.962 46.169
94401 97.249 -78.974 46.156
94402 97.287 -78.991 46.140
94403 97.328 -79.012 46.125
94404 97.351 -79.021 46.109
94405 97.367 -79.020 46.091
94406 97.382 -79.018 46.074
94407 97.398 -79.018 46.058
94408 97.414 -79.017 46.041
94409 97.424 -79.012 46.021
94410 97.432 -79.007 46.003
94411 97.440 -79.001 45.983
94412 97.448 -78.995 45.962
94413 97.457 -78.989 45.944
94414 97.465 -78.983 45.928
94415 97.473 -78.977 45.913
94416 97.481 -78.971 45.897
94417 97.490 -78.965 45.881
94418 97.498 -78.959 45.867
94419 97.506 -78.956 45.854
94420 97.514 -78.954 45.842
94421 97.522 -78.952 45.830
94422 97.531 -78.949 45.816
94423 97.539 -78.947 45.800
94424 97.547 -78.945 45.784
94425 97.550 -78.985 45.766
94426 97.571 -79.029 45.747
94427 97.595 -79.070 45.726
94428 97.597 -79.071 45.704
94429 97.599 -79.068 45.683
94430 97.602 -79.066 45.663
94431 97.604 -79.063 45.643
94432 97.607 -79.060 45.623
94433 97.609 -79.058 45.604
94434 97.612 -79.055 45.587
94435 97.614 -79.052 45.569
94436 97.617 -79.049 45.553
94437 97.619 -79.046 45.536
94438 97.649 -79.065 45.521
94439 97.684 -79.087 45.506
94440 97.720 -79.111 45.493
94441 97.752 -79.131 45.478
94442 97.759 -79.137 45.465
94443 97.767 -79.141 45.452
94444 97.774 -79.147 45.439
94445 97.781 -79.152 45.426
94446 97.785 -79.165 45.414
94447 97.789 -79.178 45.402
94448 97.793 -79.192 45.392
94449 97.797 -79.206 45.382
94450 97.802 -79.219 45.372
94451 97.821 -79.224 45.363
94452 97.839 -79.216 45.353
94453 97.857 -79.208 45.344
94454 97.874 -79.200 45.336
94455 97.892 -79.191 45.329
94456 97.910 -79.183 45.322
94457 97.928 -79.175 45.316
94458 97.962 -79.202 45.308
94459 97.997 -79.229 45.300
94460 98.032 -79.256 45.292
94461 98.084 -79.300 45.283
94462 98.110 -79.303 45.275
94463 98.135 -79.300 45.266
94464 98.159 -79.295 45.257
94465 98.183 -79.291 45.250
94466 98.223 -79.289 45.240
94467 98.249 -79.293 45.231
94468 98.275 -79.296 45.224
94469 98.301 -79.300 45.217
94470 98.327 -79.303 45.210
94471 98.353 -79.306 45.203
94472 98.381 -79.313 45.193
94473 98.414 -79.324 45.184
94474 98.447 -79.336 45.175
94475 98.480 -79.347 45.167
94476 98.512 -79.360 45.158
94477 98.545 -79.371 45.148
94478 98.580 -79.383 45.138
94479 98.617 -79.395 45.129
94480 98.653 -79.407 45.118
94481 98.690 -79.420 45.107
94482 98.726 -79.432 45.099
94483 98.763 -79.444 45.091
94484 98.781 -79.451 45.083
94485 98.782 -79.454 45.074
94486 98.782 -79.456 45.067
94487 98.783 -79.458 45.058
94488 98.803 -79.473 45.049
94489 98.827 -79.491 45.039
94490 98.850 -79.508 45.030
94491 98.872 -79.523 45.023
94492 98.894 -79.539 45.018
94493 98.916 -79.554 45.011
94494 98.949 -79.561 45.002
94495 98.991 -79.562 44.991
94496 99.031 -79.563 44.980
94497 99.073 -79.564 44.971
94498 99.114 -79.565 44.961
94499 99.162 -79.557 44.952
94500 99.177 -79.560 44.940
94501 99.187 -79.565 44.928
94502 99.197 -79.570 44.917
94503 99.209 -79.582 44.906
94504 99.224 -79.596 44.896
94505 99.239 -79.610 44.886
94506 99.254 -79.626 44.877
94507 99.269 -79.640 44.867
94508 99.284 -79.654 44.857
94509 99.299 -79.669 44.848
94510 99.307 -79.683 44.838
94511 99.321 -79.694 44.827
94512 99.342 -79.696 44.817
94513 99.357 -79.690 44.807
94514 99.372 -79.685 44.798
94515 99.387 -79.679 44.789
94516 99.403 -79.673 44.779
94517 99.418 -79.667 44.770
94518 99.433 -79.661 44.759
94519 99.448 -79.655 44.748
94520 99.463 -79.650 44.738
94521 99.478 -79.644 44.729
94522 99.493 -79.638 44.721
94523 99.508 -79.632 44.714
94524 99.515 -79.642 44.706
94525 99.527 -79.668 44.698
94526 99.547 -79.693 44.690
94527 99.567 -79.717 44.681
94528 99.586 -79.742 44.674
94529 99.606 -79.767 44.668
94530 99.626 -79.792 44.661
94531 99.645 -79.816 44.656
94532 99.665 -79.840 44.650
94533 99.685 -79.865 44.643
94534 99.704 -79.889 44.636
94535 99.724 -79.914 44.630
94536 99.744 -79.939 44.623
94537 99.763 -79.964 44.616
94538 99.783 -79.988 44.608
94539 99.804 -80.015 44.600
94540 99.841 -80.055 44.591
94541 99.864 -80.087 44.582
94542 99.890 -80.118 44.574
94543 99.918 -80.147 44.565
94544 99.946 -80.176 44.556
94545 99.973 -80.204 44.547
94546 99.996 -80.232 44.536
94547 100.018 -80.260 44.525
94548 100.042 -80.288 44.512
94549 100.065 -80.316 44.499
94550 100.100 -80.341 44.486
94551 100.169 -80.359 44.472
94552 100.239 -80.377 44.457
94553 100.292 -80.380 44.442
94554 100.312 -80.416 44.426
94555 100.332 -80.465 44.410
94556 100.364 -80.497 44.393
94557 100.409 -80.513 44.376
94558 100.454 -80.530 44.359
94559 100.498 -80.546 44.343
94560 100.543 -80.563 44.326
94561 100.587 -80.579 44.309
94562 100.632 -80.595 44.293
94563 100.676 -80.612 44.277
94564 100.716 -80.634 44.262
94565 100.757 -80.656 44.248
94566 100.796 -80.678 44.232
94567 100.837 -80.701 44.218
94568 100.877 -80.723 44.203
94569 100.917 -80.745 44.189
94570 100.958 -80.767 44.174
94571 100.998 -80.789 44.159
94572 101.038 -80.811 44.145
94573 101.078 -80.833 44.131
94574 101.119 -80.855 44.118
94575 101.159 -80.877 44.105
94843 93.522 -78.617 53.766
94844 93.532 -78.614 53.752
94845 93.541 -78.611 53.737
94846 93.551 -78.609 53.721
94847 93.561 -78.607 53.706
94848 93.570 -78.604 53.690
94849 93.580 -78.601 53.674
94850 93.588 -78.598 53.658
94851 93.587 -78.589 53.641
94852 93.586 -78.580 53.625
94853 93.584 -78.571 53.609
94854 93.583 -78.562 53.592
94855 93.582 -78.553 53.575
94856 93.590 -78.538 53.557
94857 93.612 -78.523 53.541
94858 93.635 -78.508 53.525
94859 93.657 -78.493 53.508
94860 93.677 -78.479 53.491
94861 93.689 -78.471 53.473
94862 93.703 -78.463 53.455
94863 93.718 -78.456 53.437
94864 93.734 -78.448 53.418
94865 93.749 -78.442 53.399
94866 93.765 -78.434 53.381
94867 93.780 -78.426 53.363
94868 93.795 -78.420 53.344
94869 93.811 -78.412 53.326
94870 93.826 -78.404 53.307
94871 93.842 -78.398 53.288
94872 93.856 -78.391 53.268
94873 93.871 -78.386 53.249
94874 93.886 -78.381 53.230
94875 93.898 -78.374 53.211
94876 93.897 -78.357 53.193
94877 93.895 -78.340 53.174
94878 93.894 -78.322 53.155
94879 93.901 -78.307 53.138
94880 93.929 -78.300 53.121
94881 93.943 -78.308 53.104
94882 93.957 -78.317 53.089
94883 93.970 -78.326 53.072
94884 93.983 -78.334 53.055
94885 93.997 -78.343 53.039
94886 94.008 -78.349 53.023
94887 94.016 -78.353 53.009
94888 94.025 -78.357 52.995
94889 94.033 -78.360 52.981
94890 94.041 -78.364 52.967
94891 94.051 -78.368 52.953
94892 94.063 -78.373 52.938
94893 94.075 -78.378 52.927
94894 94.085 -78.382 52.917
94895 94.096 -78.387 52.907
94896 94.105 -78.391 52.897
94897 94.115 -78.395 52.887
94898 94.124 -78.401 52.877
94899 94.128 -78.410 52.869
94900 94.133 -78.419 52.861
94901 94.138 -78.426 52.852
94902 94.142 -78.429 52.842
94903 94.147 -78.431 52.831
94904 94.151 -78.433 52.814
94905 94.156 -78.436 52.799
94906 94.160 -78.441 52.786
94907 94.164 -78.445 52.775
94908 94.168 -78.449 52.764
94909 94.176 -78.452 52.752
94910 94.197 -78.448 52.738
94911 94.219 -78.444 52.724
94912 94.239 -78.432 52.708
94913 94.257 -78.418 52.691
94914 94.277 -78.412 52.672
94915 94.296 -78.416 52.652
94916 94.305 -78.433 52.630
94917 94.313 -78.451 52.607
94918 94.322 -78.468 52.579
94919 94.330 -78.485 52.550
94920 94.339 -78.502 52.520
94921 94.347 -78.520 52.490
94922 94.356 -78.537 52.461
94923 94.364 -78.554 52.431
94924 94.384 -78.561 52.400
94925 94.409 -78.546 52.364
94926 94.433 -78.530 52.328
94927 94.458 -78.514 52.292
94928 94.482 -78.497 52.257
94929 94.507 -78.480 52.226
94930 94.531 -78.464 52.194
94931 94.556 -78.448 52.161
94932 94.580 -78.431 52.124
94933 94.595 -78.422 52.084
94934 94.606 -78.416 52.045
94935 94.616 -78.410 52.007
94936 94.627 -78.407 51.967
94937 94.639 -78.409 51.925
94938 94.651 -78.410 51.880
94939 94.664 -78.413 51.833
94940 94.676 -78.415 51.786
94941 94.688 -78.416 51.738
94942 94.700 -78.419 51.690
94943 94.712 -78.419 51.639
94944 94.725 -78.420 51.586
94945 94.744 -78.411 51.532
94946 94.763 -78.404 51.478
94947 94.781 -78.396 51.427
94948 94.800 -78.388 51.377
94949 94.814 -78.393 51.330
94950 94.831 -78.402 51.286
94951 94.848 -78.410 51.243
94952 94.864 -78.420 51.202
94953 94.876 -78.419 51.162
94954 94.886 -78.402 51.119
94955 94.895 -78.386 51.073
94956 94.905 -78.369 51.025
94957 94.915 -78.354 50.977
94958 94.924 -78.337 50.929
94959 94.934 -78.321 50.880
94960 94.944 -78.305 50.829
94961 94.954 -78.288 50.778
94962 94.964 -78.272 50.727
94963 94.973 -78.256 50.677
94964 94.983 -78.240 50.628
94965 94.993 -78.224 50.580
94966 95.003 -78.207 50.533
94967 95.018 -78.197 50.488
94968 95.057 -78.232 50.443
94969 95.105 -78.277 50.399
94970 95.153 -78.322 50.356
94971 95.199 -78.367 50.315
94972 95.210 -78.356 50.274
94973 95.220 -78.344 50.232
94974 95.231 -78.334 50.192
94975 95.242 -78.322 50.152
94976 95.252 -78.311 50.113
94977 95.263 -78.300 50.073
94978 95.273 -78.289 50.034
94979 95.284 -78.278 49.996
94980 95.295 -78.266 49.957
94981 95.305 -78.261 49.920
94982 95.314 -78.264 49.885
94983 95.323 -78.267 49.849
94984 95.333 -78.270 49.814
94985 95.342 -78.272 49.779
94986 95.351 -78.275 49.744
94987 95.361 -78.279 49.708
94988 95.370 -78.282 49.672
94989 95.387 -78.288 49.636
94990 95.411 -78.300 49.602
94991 95.435 -78.313 49.567
94992 95.459 -78.324 49.535
94993 95.483 -78.336 49.503
94994 95.507 -78.348 49.469
94995 95.531 -78.360 49.436
94996 95.539 -78.362 49.405
94997 95.544 -78.362 49.374
94998 95.550 -78.361 49.342
94999 95.555 -78.361 49.311
95000 95.560 -78.360 49.279
95001 95.565 -78.360 49.247
95002 95.571 -78.361 49.215
95003 95.580 -78.368 49.183
95004 95.589 -78.376 49.150
95005 95.598 -78.382 49.116
95006 95.608 -78.390 49.082
95007 95.617 -78.397 49.047
95008 95.626 -78.404 49.010
95009 95.635 -78.411 48.974
95010 95.636 -78.411 48.938
95011 95.632 -78.408 48.902
95012 95.628 -78.404 48.862
95013 95.624 -78.401 48.822
95014 95.620 -78.398 48.781
95015 95.615 -78.394 48.740
95016 95.611 -78.391 48.699
95017 95.621 -78.393 48.659
95018 95.637 -78.399 48.618
95019 95.654 -78.405 48.578
95020 95.670 -78.411 48.538
95021 95.686 -78.417 48.497
95022 95.701 -78.422 48.455
95023 95.713 -78.423 48.414
95024 95.737 -78.427 48.372
95025 95.760 -78.431 48.331
95026 95.783 -78.435 48.289
95027 95.804 -78.437 48.246
95028 95.824 -78.435 48.204
95029 95.842 -78.434 48.161
95030 95.859 -78.433 48.120
95031 95.874 -78.432 48.080
95032 95.885 -78.432 48.041
95033 95.895 -78.431 48.001
95034 95.905 -78.430 47.961
95035 95.917 -78.431 47.921
95036 95.930 -78.432 47.881
95037 95.945 -78.435 47.842
95038 95.959 -78.442 47.803
95039 95.973 -78.449 47.765
95040 95.987 -78.456 47.729
95041 96.001 -78.460 47.691
95042 96.013 -78.464 47.653
95043 96.026 -78.472 47.614
95044 96.038 -78.480 47.576
95045 96.051 -78.488 47.539
95046 96.062 -78.492 47.505
95047 96.070 -78.500 47.470
95048 96.075 -78.511 47.435
95049 96.083 -78.538 47.400
95050 96.092 -78.573 47.366
95051 96.112 -78.588 47.334
95052 96.138 -78.592 47.303
95053 96.165 -78.596 47.274
95054 96.191 -78.600 47.246
95055 96.218 -78.604 47.218
95056 96.244 -78.608 47.190
95057 96.270 -78.613 47.161
95058 96.297 -78.617 47.131
95059 96.324 -78.621 47.102
95060 96.350 -78.625 47.074
95061 96.377 -78.630 47.048
95062 96.405 -78.634 47.018
95063 96.434 -78.640 46.988
95064 96.463 -78.644 46.958
95065 96.472 -78.636 46.927
95066 96.468 -78.621 46.898
95067 96.464 -78.607 46.870
95068 96.461 -78.592 46.843
95069 96.510 -78.602 46.818
95070 96.559 -78.612 46.791
95071 96.608 -78.622 46.764
95072 96.657 -78.632 46.743
95073 96.707 -78.642 46.721
95074 96.755 -78.651 46.698
95075 96.773 -78.662 46.675
95076 96.774 -78.673 46.653
95077 96.774 -78.684 46.630
95078 96.789 -78.695 46.607
95079 96.809 -78.705 46.584
95080 96.829 -78.716 46.563
95081 96.839 -78.731 46.542
95082 96.839 -78.749 46.519
95083 96.847 -78.766 46.497
95084 96.859 -78.780 46.475
95085 96.870 -78.792 46.452
95086 96.880 -78.802 46.433
95087 96.890 -78.813 46.414
95088 96.900 -78.824 46.396
95089 96.910 -78.834 46.378
95090 96.920 -78.846 46.359
95091 96.936 -78.852 46.339
95092 96.965 -78.863 46.320
95093 96.994 -78.874 46.302
95094 97.023 -78.885 46.286
95095 97.051 -78.896 46.270
95096 97.080 -78.907 46.255
95097 97.109 -78.918 46.240
95098 97.138 -78.928 46.224
95099 97.167 -78.940 46.209
95100 97.196 -78.950 46.197
95101 97.225 -78.961 46.184
95102 97.254 -78.972 46.170
95103 97.286 -78.985 46.154
95104 97.333 -79.010 46.139
95105 97.357 -79.014 46.122
95106 97.373 -79.013 46.103
95107 97.388 -79.012 46.085
95108 97.404 -79.012 46.068
95109 97.420 -79.011 46.050
95110 97.432 -79.007 46.031
95111 97.440 -79.001 46.013
95112 97.448 -78.996 45.993
95113 97.456 -78.990 45.972
95114 97.464 -78.984 45.956
95115 97.472 -78.977 45.940
95116 97.480 -78.971 45.924
95117 97.489 -78.966 45.908
95118 97.497 -78.960 45.893
95119 97.505 -78.954 45.879
95120 97.513 -78.948 45.866
95121 97.522 -78.942 45.855
95122 97.530 -78.936 45.842
95123 97.538 -78.932 45.829
95124 97.546 -78.930 45.814
95125 97.555 -78.927 45.798
95126 97.560 -78.949 45.779
95127 97.583 -78.993 45.759
95128 97.604 -79.031 45.737
95129 97.617 -79.050 45.715
95130 97.620 -79.048 45.694
95131 97.623 -79.045 45.674
95132 97.626 -79.042 45.655
95133 97.628 -79.040 45.636
95134 97.631 -79.037 45.616
95135 97.633 -79.034 45.598
95136 97.636 -79.031 45.581
95137 97.638 -79.028 45.565
95138 97.652 -79.035 45.549
95139 97.688 -79.058 45.534
95140 97.724 -79.081 45.521
95141 97.756 -79.103 45.507
95142 97.764 -79.107 45.493
95143 97.771 -79.112 45.479
95144 97.778 -79.118 45.466
95145 97.786 -79.123 45.454
95146 97.793 -79.128 45.441
95147 97.798 -79.137 45.428
95148 97.802 -79.151 45.416
95149 97.806 -79.165 45.405
95150 97.810 -79.178 45.395
95151 97.814 -79.192 45.385
95152 97.826 -79.200 45.374
95153 97.842 -79.193 45.364
95154 97.857 -79.182 45.355
95155 97.872 -79.172 45.346
95156 97.887 -79.161 45.337
95157 97.901 -79.151 45.329
95158 97.917 -79.140 45.321
95159 97.932 -79.130 45.313
95160 97.947 -79.120 45.305
95161 98.008 -79.172 45.297
95162 98.069 -79.225 45.287
95163 98.114 -79.258 45.279
95164 98.141 -79.267 45.271
95165 98.166 -79.266 45.263
95166 98.191 -79.262 45.257
95167 98.223 -79.263 45.249
95168 98.249 -79.267 45.240
95169 98.275 -79.270 45.232
95170 98.301 -79.274 45.225
95171 98.327 -79.277 45.217
95172 98.353 -79.280 45.208
95173 98.379 -79.285 45.197
95174 98.412 -79.296 45.186
95175 98.445 -79.308 45.176
95176 98.478 -79.319 45.166
95177 98.511 -79.332 45.156
95178 98.545 -79.344 45.146
95179 98.582 -79.356 45.136
95180 98.618 -79.368 45.126
95181 98.655 -79.380 45.116
95182 98.692 -79.392 45.106
95183 98.728 -79.404 45.098
95184 98.765 -79.417 45.088
95185 98.795 -79.427 45.079
95186 98.795 -79.429 45.071
95187 98.796 -79.432 45.064
95188 98.796 -79.434 45.056
95189 98.819 -79.451 45.047
95190 98.843 -79.468 45.038
95191 98.866 -79.485 45.030
95192 98.889 -79.503 45.024
95193 98.912 -79.518 45.018
95194 98.933 -79.534 45.012
95195 98.962 -79.545 45.004
95196 99.003 -79.545 44.993
95197 99.044 -79.546 44.982
95198 99.085 -79.548 44.972
95199 99.131 -79.542 44.962
95200 99.162 -79.537 44.953
95201 99.172 -79.543 44.942
95202 99.182 -79.548 44.931
95203 99.192 -79.554 44.919
95204 99.205 -79.565 44.907
95205 99.219 -79.579 44.898
95206 99.234 -79.594 44.888
95207 99.250 -79.609 44.879
95208 99.264 -79.623 44.868
95209 99.279 -79.638 44.860
95210 99.294 -79.652 44.852
95211 99.309 -79.667 44.842
95212 99.324 -79.679 44.833
95213 99.339 -79.673 44.824
95214 99.354 -79.667 44.814
95215 99.369 -79.661 44.805
95216 99.384 -79.656 44.795
95217 99.400 -79.650 44.785
95218 99.415 -79.645 44.775
95219 99.431 -79.639 44.765
95220 99.446 -79.634 44.755
95221 99.462 -79.629 44.745
95222 99.477 -79.623 44.735
95223 99.492 -79.617 44.728
95224 99.507 -79.612 44.720
95225 99.531 -79.619 44.712
95226 99.549 -79.644 44.704
95227 99.569 -79.668 44.696
95228 99.588 -79.693 44.688
95229 99.608 -79.717 44.681
95230 99.628 -79.742 44.676
95231 99.647 -79.767 44.670
95232 99.667 -79.791 44.665
95233 99.687 -79.816 44.659
95234 99.706 -79.840 44.653
95235 99.726 -79.865 44.646
95236 99.746 -79.889 44.640
95237 99.765 -79.914 44.633
95238 99.785 -79.939 44.626
95239 99.804 -79.963 44.618
95240 99.828 -79.991 44.610
95241 99.859 -80.028 44.601
95242 99.883 -80.061 44.592
95243 99.912 -80.089 44.584
95244 99.940 -80.118 44.576
95245 99.969 -80.146 44.568
95246 99.994 -80.175 44.558
95247 100.017 -80.203 44.548
95248 100.040 -80.231 44.537
95249 100.063 -80.259 44.525
95250 100.086 -80.287 44.511
95251 100.114 -80.313 44.497
95252 100.150 -80.338 44.483
95253 100.213 -80.358 44.468
95254 100.283 -80.375 44.453
95255 100.346 -80.386 44.437
95256 100.366 -80.418 44.419
95257 100.386 -80.466 44.402
95258 100.411 -80.508 44.385
95259 100.456 -80.524 44.368
95260 100.500 -80.541 44.351
95261 100.545 -80.557 44.333
95262 100.590 -80.573 44.317
95263 100.632 -80.592 44.301
95264 100.672 -80.614 44.285
95265 100.712 -80.636 44.269
95266 100.753 -80.659 44.254
95267 100.792 -80.681 44.238
95268 100.833 -80.703 44.224
95269 100.873 -80.725 44.209
95270 100.913 -80.747 44.194
95271 100.954 -80.769 44.179
95272 100.994 -80.791 44.164
95273 101.034 -80.813 44.149
95274 101.074 -80.835 44.135
95275 101.115 -80.857 44.122
95276 101.159 -80.872 44.108
95277 101.206 -80.885 44.094
95543 93.512 -78.608 53.784
95544 93.522 -78.606 53.770
95545 93.532 -78.603 53.757
95546 93.542 -78.601 53.741
95547 93.551 -78.598 53.725
95548 93.561 -78.596 53.710
95549 93.570 -78.593 53.694
95550 93.580 -78.591 53.677
95551 93.585 -78.586 53.661
95552 93.584 -78.576 53.643
95553 93.582 -78.567 53.626
95554 93.581 -78.557 53.609
95555 93.580 -78.548 53.592
95556 93.581 -78.538 53.574
95557 93.595 -78.523 53.556
95558 93.618 -78.507 53.539
95559 93.640 -78.492 53.522
95560 93.662 -78.477 53.504
95561 93.679 -78.467 53.487
95562 93.690 -78.458 53.469
95563 93.705 -78.451 53.451
95564 93.720 -78.443 53.433
95565 93.735 -78.436 53.415
95566 93.751 -78.429 53.395
95567 93.767 -78.421 53.377
95568 93.782 -78.414 53.358
95569 93.797 -78.407 53.339
95570 93.813 -78.399 53.321
95571 93.828 -78.391 53.302
95572 93.844 -78.385 53.283
95573 93.859 -78.377 53.264
95574 93.874 -78.371 53.245
95575 93.889 -78.366 53.226
95576 93.903 -78.360 53.207
95577 93.909 -78.348 53.189
95578 93.907 -78.331 53.170
95579 93.906 -78.314 53.152
95580 93.916 -78.298 53.135
95581 93.939 -78.297 53.118
95582 93.953 -78.306 53.102
95583 93.967 -78.314 53.087
95584 93.980 -78.322 53.070
95585 93.993 -78.332 53.054
95586 94.003 -78.337 53.039
95587 94.011 -78.341 53.025
95588 94.019 -78.344 53.011
95589 94.028 -78.347 52.998
95590 94.036 -78.351 52.984
95591 94.044 -78.355 52.972
95592 94.053 -78.359 52.958
95593 94.065 -78.364 52.944
95594 94.077 -78.369 52.931
95595 94.087 -78.373 52.921
95596 94.097 -78.378 52.910
95597 94.107 -78.382 52.899
95598 94.116 -78.387 52.887
95599 94.120 -78.396 52.877
95600 94.125 -78.404 52.869
95601 94.129 -78.413 52.861
95602 94.134 -78.423 52.854
95603 94.139 -78.428 52.845
95604 94.143 -78.430 52.835
95605 94.148 -78.432 52.819
95606 94.152 -78.435 52.805
95607 94.157 -78.437 52.792
95608 94.161 -78.439 52.782
95609 94.166 -78.443 52.772
95610 94.170 -78.448 52.762
95611 94.192 -78.443 52.749
95612 94.211 -78.431 52.733
95613 94.229 -78.417 52.716
95614 94.248 -78.404 52.696
95615 94.268 -78.400 52.674
95616 94.289 -78.401 52.651
95617 94.301 -78.414 52.627
95618 94.310 -78.432 52.603
95619 94.318 -78.449 52.574
95620 94.326 -78.467 52.543
95621 94.335 -78.483 52.513
95622 94.343 -78.501 52.483
95623 94.352 -78.518 52.453
95624 94.360 -78.535 52.421
95625 94.381 -78.529 52.389
95626 94.414 -78.535 52.353
95627 94.438 -78.519 52.316
95628 94.463 -78.502 52.281
95629 94.487 -78.485 52.246
95630 94.511 -78.469 52.215
95631 94.536 -78.453 52.184
95632 94.560 -78.436 52.151
95633 94.585 -78.420 52.111
95634 94.601 -78.410 52.070
95635 94.611 -78.404 52.029
95636 94.622 -78.398 51.988
95637 94.632 -78.392 51.946
95638 94.644 -78.392 51.903
95639 94.656 -78.395 51.857
95640 94.668 -78.396 51.809
95641 94.680 -78.398 51.762
95642 94.692 -78.401 51.715
95643 94.705 -78.402 51.667
95644 94.717 -78.401 51.618
95645 94.729 -78.402 51.567
95646 94.743 -78.401 51.515
95647 94.762 -78.393 51.463
95648 94.780 -78.385 51.413
95649 94.799 -78.378 51.365
95650 94.812 -78.381 51.322
95651 94.827 -78.388 51.280
95652 94.844 -78.397 51.240
95653 94.861 -78.405 51.201
95654 94.886 -78.407 51.159
95655 94.897 -78.408 51.113
95656 94.907 -78.410 51.065
95657 94.918 -78.411 51.014
95658 94.929 -78.410 50.965
95659 94.938 -78.393 50.915
95660 94.948 -78.376 50.866
95661 94.958 -78.359 50.815
95662 94.968 -78.342 50.764
95663 94.977 -78.325 50.714
95664 94.987 -78.308 50.666
95665 94.997 -78.291 50.618
95666 95.006 -78.274 50.571
95667 95.016 -78.257 50.525
95668 95.075 -78.271 50.481
95669 95.131 -78.310 50.438
95670 95.179 -78.355 50.395
95671 95.195 -78.353 50.352
95672 95.205 -78.341 50.311
95673 95.216 -78.330 50.270
95674 95.227 -78.319 50.230
95675 95.237 -78.308 50.190
95676 95.248 -78.297 50.150
95677 95.258 -78.286 50.112
95678 95.269 -78.275 50.073
95679 95.280 -78.263 50.035
95680 95.290 -78.252 49.997
95681 95.301 -78.241 49.959
95682 95.311 -78.239 49.922
95683 95.320 -78.242 49.886
95684 95.329 -78.245 49.851
95685 95.339 -78.248 49.815
95686 95.348 -78.250 49.780
95687 95.357 -78.253 49.745
95688 95.367 -78.257 49.709
95689 95.376 -78.260 49.674
95690 95.386 -78.262 49.639
95691 95.402 -78.269 49.606
95692 95.426 -78.281 49.573
95693 95.450 -78.293 49.542
95694 95.474 -78.305 49.510
95695 95.497 -78.317 49.477
95696 95.522 -78.329 49.446
95697 95.531 -78.331 49.415
95698 95.536 -78.331 49.385
95699 95.541 -78.330 49.353
95700 95.546 -78.329 49.322
95701 95.551 -78.329 49.291
95702 95.557 -78.330 49.260
95703 95.566 -78.338 49.229
95704 95.575 -78.344 49.197
95705 95.584 -78.352 49.164
95706 95.594 -78.359 49.129
95707 95.603 -78.366 49.094
95708 95.612 -78.373 49.058
95709 95.621 -78.380 49.022
95710 95.630 -78.387 48.986
95711 95.635 -78.391 48.951
95712 95.631 -78.388 48.916
95713 95.627 -78.384 48.878
95714 95.623 -78.380 48.837
95715 95.618 -78.377 48.796
95716 95.614 -78.373 48.754
95717 95.610 -78.370 48.714
95718 95.625 -78.379 48.674
95719 95.642 -78.381 48.633
95720 95.659 -78.387 48.591
95721 95.675 -78.393 48.551
95722 95.692 -78.398 48.509
95723 95.705 -78.401 48.467
95724 95.717 -78.403 48.426
95725 95.738 -78.407 48.385
95726 95.761 -78.410 48.343
95727 95.784 -78.413 48.302
95728 95.804 -78.416 48.260
95729 95.821 -78.416 48.219
95730 95.838 -78.415 48.177
95731 95.855 -78.414 48.137
95732 95.869 -78.413 48.097
95733 95.880 -78.413 48.057
95734 95.891 -78.412 48.016
95735 95.905 -78.413 47.976
95736 95.917 -78.413 47.936
95737 95.929 -78.414 47.896
95738 95.943 -78.417 47.856
95739 95.955 -78.424 47.818
95740 95.967 -78.430 47.780
95741 95.979 -78.437 47.741
95742 95.992 -78.444 47.704
95743 96.005 -78.452 47.666
95744 96.019 -78.460 47.628
95745 96.032 -78.467 47.590
95746 96.045 -78.475 47.555
95747 96.059 -78.485 47.522
95748 96.059 -78.499 47.487
95749 96.062 -78.512 47.452
95750 96.067 -78.523 47.418
95751 96.075 -78.548 47.385
95752 96.096 -78.563 47.354
95753 96.122 -78.567 47.322
95754 96.149 -78.571 47.291
95755 96.175 -78.575 47.262
95756 96.201 -78.579 47.234
95757 96.230 -78.585 47.206
95758 96.259 -78.590 47.178
95759 96.289 -78.595 47.150
95760 96.318 -78.601 47.121
95761 96.347 -78.607 47.093
95762 96.377 -78.613 47.065
95763 96.406 -78.618 47.035
95764 96.435 -78.623 47.004
95765 96.464 -78.625 46.974
95766 96.493 -78.628 46.945
95767 96.523 -78.630 46.917
95768 96.523 -78.617 46.889
95769 96.519 -78.603 46.862
95770 96.516 -78.589 46.836
95771 96.565 -78.599 46.809
95772 96.614 -78.609 46.782
95773 96.663 -78.619 46.760
95774 96.713 -78.629 46.738
95775 96.761 -78.639 46.716
95776 96.764 -78.649 46.693
95777 96.765 -78.661 46.670
95778 96.765 -78.671 46.647
95779 96.775 -78.683 46.623
95780 96.796 -78.693 46.601
95781 96.816 -78.704 46.579
95782 96.819 -78.720 46.557
95783 96.822 -78.739 46.536
95784 96.835 -78.753 46.513
95785 96.847 -78.767 46.491
95786 96.857 -78.777 46.469
95787 96.867 -78.789 46.449
95788 96.877 -78.799 46.430
95789 96.887 -78.810 46.412
95790 96.897 -78.821 46.394
95791 96.911 -78.836 46.375
95792 96.932 -78.854 46.355
95793 96.962 -78.865 46.335
95794 96.992 -78.875 46.318
95795 97.022 -78.886 46.302
95796 97.051 -78.896 46.286
95797 97.081 -78.907 46.270
95798 97.111 -78.918 46.255
95799 97.141 -78.928 46.239
95800 97.170 -78.939 46.224
95801 97.200 -78.949 46.212
95802 97.230 -78.960 46.198
95803 97.259 -78.971 46.184
95804 97.288 -78.981 46.168
95805 97.325 -78.995 46.152
95806 97.363 -79.007 46.134
95807 97.378 -79.006 46.114
95808 97.394 -79.006 46.096
95809 97.409 -79.005 46.078
95810 97.426 -79.004 46.060
95811 97.439 -79.002 46.042
95812 97.447 -78.996 46.023
95813 97.456 -78.990 46.003
95814 97.464 -78.984 45.983
95815 97.472 -78.978 45.967
95816 97.480 -78.972 45.950
95817 97.489 -78.966 45.934
95818 97.497 -78.960 45.918
95819 97.505 -78.955 45.903
95820 97.513 -78.949 45.890
95821 97.521 -78.943 45.877
95822 97.529 -78.937 45.866
95823 97.537 -78.930 45.854
95824 97.545 -78.925 45.840
95825 97.554 -78.919 45.827
95826 97.562 -78.913 45.811
95827 97.572 -78.912 45.792
95828 97.593 -78.949 45.770
95829 97.614 -78.987 45.748
95830 97.636 -79.024 45.726
95831 97.641 -79.027 45.706
95832 97.644 -79.024 45.685
95833 97.646 -79.021 45.666
95834 97.649 -79.018 45.647
95835 97.651 -79.015 45.628
95836 97.654 -79.013 45.610
95837 97.656 -79.010 45.593
95838 97.659 -79.007 45.578
95839 97.692 -79.028 45.562
95840 97.728 -79.052 45.548
95841 97.761 -79.073 45.534
95842 97.769 -79.078 45.519
95843 97.776 -79.084 45.506
95844 97.783 -79.089 45.494
95845 97.790 -79.093 45.481
95846 97.798 -79.099 45.469
95847 97.805 -79.104 45.455
95848 97.812 -79.110 45.442
95849 97.816 -79.124 45.429
95850 97.819 -79.137 45.417
95851 97.823 -79.151 45.406
95852 97.827 -79.165 45.396
95853 97.842 -79.172 45.385
95854 97.859 -79.179 45.374
95855 97.874 -79.169 45.365
95856 97.889 -79.159 45.357
95857 97.903 -79.149 45.347
95858 97.919 -79.138 45.338
95859 97.934 -79.128 45.329
95860 97.949 -79.118 45.319
95861 97.963 -79.107 45.309
95862 97.997 -79.111 45.300
95863 98.055 -79.149 45.290
95864 98.116 -79.202 45.281
95865 98.144 -79.214 45.273
95866 98.171 -79.223 45.266
95867 98.197 -79.233 45.259
95868 98.223 -79.238 45.252
95869 98.249 -79.241 45.245
95870 98.275 -79.244 45.237
95871 98.301 -79.248 45.230
95872 98.327 -79.251 45.221
95873 98.353 -79.255 45.211
95874 98.379 -79.258 45.200
95875 98.411 -79.269 45.190
95876 98.443 -79.280 45.179
95877 98.476 -79.292 45.167
95878 98.510 -79.303 45.157
95879 98.547 -79.316 45.147
95880 98.583 -79.328 45.137
95881 98.620 -79.341 45.126
95882 98.656 -79.353 45.116
95883 98.693 -79.365 45.106
95884 98.730 -79.377 45.097
95885 98.766 -79.389 45.086
95886 98.803 -79.401 45.077
95887 98.809 -79.405 45.069
95888 98.809 -79.407 45.062
95889 98.811 -79.411 45.054
95890 98.835 -79.429 45.045
95891 98.859 -79.446 45.036
95892 98.882 -79.463 45.030
95893 98.906 -79.480 45.024
95894 98.929 -79.498 45.018
95895 98.951 -79.513 45.012
95896 98.975 -79.527 45.005
95897 99.016 -79.529 44.996
95898 99.057 -79.529 44.983
95899 99.099 -79.529 44.973
95900 99.148 -79.515 44.962
95901 99.158 -79.521 44.952
95902 99.167 -79.526 44.942
95903 99.178 -79.532 44.931
95904 99.187 -79.538 44.920
95905 99.200 -79.548 44.908
95906 99.215 -79.563 44.898
95907 99.229 -79.577 44.889
95908 99.244 -79.592 44.880
95909 99.260 -79.607 44.870
95910 99.274 -79.621 44.863
95911 99.289 -79.636 44.855
95912 99.304 -79.651 44.846
95913 99.326 -79.659 44.838
95914 99.345 -79.661 44.828
95915 99.360 -79.656 44.818
95916 99.375 -79.651 44.808
95917 99.391 -79.645 44.798
95918 99.406 -79.639 44.788
95919 99.421 -79.634 44.779
95920 99.437 -79.629 44.769
95921 99.452 -79.623 44.760
95922 99.468 -79.617 44.750
95923 99.483 -79.612 44.741
95924 99.498 -79.607 44.733
95925 99.517 -79.604 44.726
95926 99.547 -79.607 44.718
95927 99.579 -79.617 44.710
95928 99.597 -79.642 44.702
95929 99.615 -79.667 44.695
95930 99.632 -79.692 44.689
95931 99.650 -79.717 44.684
95932 99.669 -79.742 44.679
95933 99.689 -79.767 44.674
95934 99.708 -79.791 44.669
95935 99.728 -79.816 44.663
95936 99.748 -79.840 44.657
95937 99.767 -79.864 44.650
95938 99.787 -79.889 44.643
95939 99.806 -79.914 44.636
95940 99.826 -79.939 44.629
95941 99.850 -79.968 44.620
95942 99.877 -80.002 44.612
95943 99.905 -80.031 44.604
95944 99.934 -80.060 44.598
95945 99.963 -80.089 44.590
95946 99.993 -80.118 44.582
95947 100.033 -80.143 44.574
95948 100.067 -80.167 44.564
95949 100.096 -80.193 44.552
95950 100.124 -80.220 44.540
95951 100.153 -80.247 44.526
95952 100.182 -80.273 44.512
95953 100.207 -80.300 44.497
95954 100.242 -80.326 44.482
95955 100.303 -80.346 44.467
95956 100.360 -80.361 44.451
95957 100.384 -80.368 44.433
95958 100.414 -80.418 44.414
95959 100.440 -80.467 44.396
95960 100.460 -80.515 44.378
95961 100.502 -80.535 44.360
95962 100.547 -80.551 44.342
95963 100.587 -80.573 44.325
95964 100.628 -80.594 44.308
95965 100.668 -80.617 44.292
95966 100.708 -80.639 44.276
95967 100.749 -80.661 44.260
95968 100.789 -80.683 44.245
95969 100.829 -80.705 44.230
95970 100.869 -80.727 44.215
95971 100.909 -80.749 44.200
95972 100.950 -80.771 44.185
95973 100.990 -80.793 44.170
95974 101.030 -80.815 44.154
95975 101.071 -80.836 44.139
95976 101.117 -80.849 44.126
95977 101.164 -80.862 44.112
95978 101.210 -80.875 44.099
95979 101.257 -80.888 44.085
96244 93.512 -78.597 53.788
96245 93.522 -78.595 53.774
96246 93.532 -78.592 53.760
96247 93.542 -78.590 53.745
96248 93.551 -78.587 53.729
96249 93.561 -78.585 53.714
96250 93.571 -78.582 53.697
96251 93.580 -78.580 53.680
96252 93.582 -78.573 53.663
96253 93.580 -78.563 53.645
96254 93.579 -78.554 53.627
96255 93.578 -78.545 53.609
96256 93.576 -78.535 53.591
96257 93.583 -78.522 53.573
96258 93.601 -78.507 53.555
96259 93.623 -78.492 53.537
96260 93.645 -78.476 53.518
96261 93.668 -78.462 53.500
96262 93.680 -78.454 53.483
96263 93.691 -78.445 53.465
96264 93.707 -78.438 53.446
96265 93.722 -78.430 53.428
96266 93.737 -78.423 53.410
96267 93.753 -78.416 53.391
96268 93.768 -78.408 53.372
96269 93.784 -78.401 53.353
96270 93.799 -78.394 53.335
96271 93.815 -78.386 53.315
96272 93.830 -78.379 53.296
96273 93.845 -78.372 53.277
96274 93.861 -78.364 53.259
96275 93.876 -78.357 53.240
96276 93.892 -78.350 53.221
96277 93.906 -78.345 53.203
96278 93.921 -78.340 53.184
96279 93.919 -78.322 53.165
96280 93.918 -78.306 53.148
96281 93.930 -78.290 53.132
96282 93.949 -78.294 53.117
96283 93.963 -78.303 53.102
96284 93.976 -78.311 53.087
96285 93.990 -78.320 53.071
96286 93.998 -78.324 53.056
96287 94.006 -78.328 53.041
96288 94.015 -78.332 53.028
96289 94.023 -78.335 53.016
96290 94.031 -78.339 53.003
96291 94.039 -78.342 52.989
96292 94.047 -78.346 52.977
96293 94.056 -78.350 52.964
96294 94.067 -78.354 52.950
96295 94.079 -78.360 52.935
96296 94.088 -78.363 52.924
96297 94.099 -78.368 52.913
96298 94.107 -78.373 52.901
96299 94.112 -78.382 52.888
96300 94.117 -78.391 52.877
96301 94.121 -78.400 52.868
96302 94.126 -78.409 52.861
96303 94.130 -78.417 52.855
96304 94.135 -78.426 52.847
96305 94.140 -78.429 52.837
96306 94.144 -78.432 52.822
96307 94.149 -78.433 52.809
96308 94.153 -78.435 52.798
96309 94.158 -78.438 52.788
96310 94.163 -78.440 52.779
96311 94.167 -78.442 52.769
96312 94.183 -78.429 52.756
96313 94.201 -78.416 52.738
96314 94.220 -78.402 52.718
96315 94.238 -78.388 52.696
96316 94.259 -78.387 52.672
96317 94.280 -78.389 52.647
96318 94.297 -78.395 52.621
96319 94.306 -78.413 52.596
96320 94.314 -78.430 52.566
96321 94.323 -78.448 52.535
96322 94.331 -78.464 52.503
96323 94.339 -78.482 52.473
96324 94.348 -78.499 52.441
96325 94.359 -78.510 52.408
96326 94.381 -78.502 52.375
96327 94.411 -78.507 52.338
96328 94.443 -78.507 52.302
96329 94.467 -78.491 52.266
96330 94.492 -78.474 52.232
96331 94.516 -78.457 52.199
96332 94.541 -78.442 52.165
96333 94.565 -78.425 52.131
96334 94.590 -78.408 52.090
96335 94.606 -78.398 52.047
96336 94.616 -78.392 52.003
96337 94.627 -78.386 51.960
96338 94.637 -78.380 51.917
96339 94.648 -78.376 51.872
96340 94.661 -78.378 51.825
96341 94.673 -78.380 51.776
96342 94.685 -78.382 51.729
96343 94.697 -78.384 51.683
96344 94.709 -78.386 51.637
96345 94.721 -78.384 51.589
96346 94.734 -78.382 51.541
96347 94.745 -78.386 51.492
96348 94.761 -78.382 51.442
96349 94.779 -78.375 51.395
96350 94.798 -78.367 51.351
96351 94.812 -78.369 51.310
96352 94.824 -78.373 51.272
96353 94.841 -78.382 51.234
96354 94.862 -78.387 51.195
96355 94.887 -78.388 51.150
96356 94.907 -78.389 51.101
96357 94.918 -78.391 51.050
96358 94.928 -78.392 51.000
96359 94.939 -78.394 50.950
96360 94.950 -78.395 50.900
96361 94.961 -78.397 50.850
96362 94.972 -78.398 50.799
96363 94.982 -78.400 50.750
96364 94.993 -78.401 50.702
96365 95.003 -78.397 50.655
96366 95.013 -78.379 50.608
96367 95.023 -78.363 50.563
96368 95.047 -78.354 50.518
96369 95.119 -78.373 50.475
96370 95.165 -78.327 50.432
96371 95.191 -78.338 50.390
96372 95.201 -78.327 50.348
96373 95.212 -78.316 50.307
96374 95.222 -78.305 50.267
96375 95.233 -78.294 50.228
96376 95.244 -78.282 50.189
96377 95.254 -78.272 50.150
96378 95.265 -78.260 50.112
96379 95.275 -78.249 50.074
96380 95.286 -78.238 50.036
96381 95.297 -78.227 49.999
96382 95.307 -78.216 49.962
96383 95.317 -78.217 49.924
96384 95.326 -78.220 49.888
96385 95.335 -78.223 49.852
96386 95.345 -78.226 49.817
96387 95.354 -78.228 49.782
96388 95.364 -78.232 49.746
96389 95.373 -78.235 49.712
96390 95.382 -78.238 49.678
96391 95.392 -78.241 49.644
96392 95.401 -78.243 49.612
96393 95.417 -78.250 49.581
96394 95.441 -78.262 49.550
96395 95.465 -78.274 49.519
96396 95.488 -78.285 49.488
96397 95.513 -78.297 49.458
96398 95.522 -78.300 49.428
96399 95.527 -78.299 49.398
96400 95.532 -78.299 49.367
96401 95.537 -78.298 49.335
96402 95.543 -78.300 49.304
96403 95.552 -78.307 49.273
96404 95.561 -78.314 49.243
96405 95.570 -78.321 49.211
96406 95.580 -78.328 49.178
96407 95.589 -78.335 49.144
96408 95.598 -78.342 49.107
96409 95.607 -78.350 49.071
96410 95.616 -78.357 49.034
96411 95.625 -78.363 48.998
96412 95.634 -78.370 48.963
96413 95.630 -78.366 48.928
96414 95.626 -78.363 48.891
96415 95.621 -78.360 48.852
96416 95.617 -78.357 48.811
96417 95.613 -78.353 48.769
96418 95.615 -78.353 48.728
96419 95.629 -78.362 48.687
96420 95.646 -78.366 48.646
96421 95.664 -78.368 48.604
96422 95.681 -78.374 48.562
96423 95.697 -78.380 48.520
96424 95.708 -78.382 48.478
96425 95.721 -78.383 48.438
96426 95.739 -78.385 48.397
96427 95.762 -78.389 48.355
96428 95.785 -78.396 48.314
96429 95.801 -78.397 48.273
96430 95.817 -78.396 48.232
96431 95.834 -78.396 48.190
96432 95.850 -78.395 48.150
96433 95.863 -78.395 48.109
96434 95.874 -78.394 48.069
96435 95.891 -78.395 48.027
96436 95.907 -78.398 47.986
96437 95.919 -78.398 47.946
96438 95.931 -78.399 47.906
96439 95.946 -78.403 47.867
96440 95.959 -78.407 47.829
96441 95.971 -78.414 47.791
96442 95.983 -78.421 47.752
96443 95.996 -78.428 47.715
96444 96.010 -78.435 47.678
96445 96.023 -78.444 47.640
96446 96.036 -78.451 47.605
96447 96.050 -78.463 47.571
96448 96.064 -78.475 47.539
96449 96.068 -78.486 47.504
96450 96.067 -78.501 47.469
96451 96.067 -78.517 47.435
96452 96.066 -78.532 47.404
96453 96.084 -78.541 47.373
96454 96.114 -78.546 47.341
96455 96.143 -78.551 47.309
96456 96.172 -78.557 47.280
96457 96.202 -78.563 47.251
96458 96.231 -78.569 47.222
96459 96.260 -78.574 47.194
96460 96.290 -78.580 47.168
96461 96.319 -78.586 47.140
96462 96.348 -78.591 47.111
96463 96.378 -78.597 47.083
96464 96.407 -78.601 47.052
96465 96.436 -78.604 47.021
96466 96.465 -78.607 46.992
96467 96.495 -78.609 46.963
96468 96.524 -78.612 46.935
96469 96.553 -78.614 46.908
96470 96.578 -78.614 46.881
96471 96.574 -78.600 46.854
96472 96.571 -78.586 46.828
96473 96.620 -78.595 46.801
96474 96.669 -78.605 46.777
96475 96.718 -78.615 46.754
96476 96.754 -78.626 46.732
96477 96.755 -78.636 46.709
96478 96.755 -78.648 46.686
96479 96.756 -78.659 46.663
96480 96.762 -78.670 46.639
96481 96.783 -78.680 46.617
96482 96.799 -78.692 46.594
96483 96.800 -78.711 46.572
96484 96.810 -78.726 46.551
96485 96.823 -78.740 46.529
96486 96.834 -78.753 46.506
96487 96.844 -78.764 46.485
96488 96.855 -78.774 46.465
96489 96.865 -78.786 46.445
96490 96.875 -78.796 46.426
96491 96.889 -78.812 46.408
96492 96.908 -78.836 46.390
96493 96.928 -78.859 46.370
96494 96.955 -78.868 46.350
96495 96.984 -78.879 46.334
96496 97.014 -78.890 46.318
96497 97.044 -78.900 46.302
96498 97.074 -78.910 46.285
96499 97.103 -78.921 46.270
96500 97.133 -78.931 46.255
96501 97.163 -78.942 46.240
96502 97.192 -78.952 46.226
96503 97.222 -78.963 46.213
96504 97.252 -78.974 46.198
96505 97.283 -78.981 46.182
96506 97.318 -78.981 46.164
96507 97.356 -78.993 46.145
96508 97.384 -79.000 46.125
96509 97.400 -78.999 46.107
96510 97.416 -78.998 46.089
96511 97.431 -78.997 46.070
96512 97.447 -78.996 46.052
96513 97.455 -78.990 46.034
96514 97.463 -78.984 46.014
96515 97.471 -78.979 45.993
96516 97.479 -78.973 45.977
96517 97.488 -78.967 45.960
96518 97.496 -78.961 45.943
96519 97.504 -78.955 45.927
96520 97.512 -78.949 45.913
96521 97.521 -78.943 45.900
96522 97.529 -78.937 45.886
96523 97.537 -78.931 45.876
96524 97.545 -78.925 45.864
96525 97.554 -78.919 45.850
96526 97.562 -78.914 45.838
96527 97.572 -78.908 45.821
96528 97.590 -78.905 45.801
96529 97.603 -78.905 45.780
96530 97.625 -78.943 45.756
96531 97.646 -78.980 45.735
96532 97.662 -79.006 45.715
96533 97.664 -79.003 45.696
96534 97.667 -79.000 45.676
96535 97.670 -78.997 45.657
96536 97.672 -78.995 45.639
96537 97.675 -78.992 45.623
96538 97.677 -78.989 45.606
96539 97.696 -78.999 45.590
96540 97.732 -79.022 45.575
96541 97.766 -79.044 45.561
96542 97.773 -79.049 45.546
96543 97.781 -79.055 45.532
96544 97.788 -79.059 45.518
96545 97.795 -79.065 45.505
96546 97.802 -79.070 45.493
96547 97.809 -79.075 45.480
96548 97.817 -79.080 45.467
96549 97.824 -79.085 45.454
96550 97.829 -79.096 45.442
96551 97.833 -79.110 45.429
96552 97.837 -79.124 45.417
96553 97.841 -79.137 45.406
96554 97.858 -79.143 45.395
96555 97.874 -79.150 45.385
96556 97.891 -79.157 45.375
96557 97.906 -79.147 45.366
96558 97.921 -79.136 45.356
96559 97.936 -79.126 45.345
96560 97.951 -79.115 45.335
96561 97.965 -79.105 45.324
96562 97.981 -79.095 45.313
96563 97.996 -79.099 45.303
96564 98.046 -79.101 45.293
96565 98.101 -79.126 45.284
96566 98.148 -79.161 45.276
96567 98.174 -79.170 45.268
96568 98.200 -79.182 45.260
96569 98.224 -79.196 45.253
96570 98.249 -79.209 45.247
96571 98.275 -79.219 45.241
96572 98.301 -79.222 45.233
96573 98.327 -79.225 45.224
96574 98.353 -79.229 45.214
96575 98.379 -79.232 45.203
96576 98.409 -79.241 45.194
96577 98.442 -79.252 45.183
96578 98.475 -79.264 45.170
96579 98.512 -79.276 45.161
96580 98.548 -79.288 45.151
96581 98.585 -79.300 45.140
96582 98.622 -79.313 45.130
96583 98.658 -79.325 45.118
96584 98.695 -79.338 45.106
96585 98.731 -79.350 45.096
96586 98.768 -79.362 45.086
96587 98.804 -79.374 45.076
96588 98.822 -79.381 45.068
96589 98.823 -79.383 45.061
96590 98.828 -79.388 45.051
96591 98.851 -79.406 45.043
96592 98.875 -79.423 45.036
96593 98.898 -79.441 45.031
96594 98.922 -79.458 45.025
96595 98.945 -79.475 45.019
96596 98.969 -79.492 45.013
96597 98.991 -79.507 45.005
96598 99.028 -79.511 44.996
96599 99.069 -79.513 44.986
96600 99.119 -79.500 44.975
96601 99.143 -79.499 44.965
96602 99.153 -79.504 44.954
96603 99.163 -79.510 44.942
96604 99.172 -79.516 44.931
96605 99.183 -79.521 44.921
96606 99.195 -79.532 44.910
96607 99.210 -79.546 44.899
96608 99.225 -79.561 44.890
96609 99.239 -79.576 44.880
96610 99.255 -79.590 44.872
96611 99.270 -79.604 44.864
96612 99.284 -79.619 44.857
96613 99.305 -79.629 44.849
96614 99.327 -79.637 44.840
96615 99.348 -79.645 44.831
96616 99.366 -79.645 44.821
96617 99.381 -79.639 44.810
96618 99.397 -79.634 44.800
96619 99.412 -79.629 44.791
96620 99.427 -79.623 44.782
96621 99.442 -79.617 44.773
96622 99.458 -79.612 44.764
96623 99.473 -79.607 44.755
96624 99.489 -79.601 44.747
96625 99.504 -79.595 44.739
96626 99.532 -79.598 44.732
96627 99.561 -79.601 44.725
96628 99.591 -79.604 44.717
96629 99.623 -79.612 44.710
96630 99.646 -79.640 44.704
96631 99.663 -79.665 44.699
96632 99.681 -79.690 44.694
96633 99.699 -79.715 44.689
96634 99.716 -79.740 44.684
96635 99.733 -79.765 44.678
96636 99.751 -79.790 44.671
96637 99.769 -79.815 44.664
96638 99.789 -79.840 44.658
96639 99.808 -79.864 44.651
96640 99.828 -79.889 44.644
96641 99.848 -79.914 44.638
96642 99.872 -79.943 44.631
96643 99.899 -79.974 44.623
96644 99.927 -80.002 44.616
96645 99.960 -80.033 44.609
96646 99.992 -80.064 44.601
96647 100.024 -80.096 44.593
96648 100.063 -80.120 44.585
96649 100.109 -80.139 44.575
96650 100.154 -80.157 44.564
96651 100.192 -80.180 44.552
96652 100.221 -80.206 44.539
96653 100.250 -80.233 44.526
96654 100.276 -80.260 44.512
96655 100.294 -80.290 44.497
96656 100.334 -80.314 44.481
96657 100.384 -80.331 44.464
96658 100.434 -80.365 44.447
96659 100.462 -80.410 44.428
96660 100.482 -80.450 44.409
96661 100.502 -80.491 44.390
96662 100.555 -80.554 44.371
96663 100.543 -80.553 44.352
96664 100.583 -80.576 44.333
96665 100.624 -80.598 44.316
96666 100.664 -80.620 44.300
96667 100.704 -80.641 44.284
96668 100.745 -80.663 44.268
96669 100.785 -80.685 44.253
96670 100.825 -80.707 44.238
96671 100.865 -80.729 44.222
96672 100.905 -80.751 44.206
96673 100.946 -80.773 44.190
96674 100.986 -80.795 44.175
96675 101.029 -80.814 44.159
96676 101.075 -80.827 44.144
96677 101.121 -80.840 44.130
96678 101.168 -80.853 44.116
96679 101.214 -80.866 44.103
96680 101.261 -80.879 44.089
96681 101.307 -80.892 44.075
96944 93.503 -78.589 53.807
96945 93.513 -78.586 53.793
96946 93.522 -78.584 53.779
96947 93.532 -78.582 53.764
96948 93.542 -78.579 53.749
96949 93.552 -78.576 53.733
96950 93.561 -78.574 53.716
96951 93.571 -78.572 53.699
96952 93.580 -78.568 53.681
96953 93.578 -78.559 53.664
96954 93.577 -78.550 53.646
96955 93.576 -78.541 53.627
96956 93.574 -78.532 53.608
96957 93.574 -78.522 53.590
96958 93.584 -78.507 53.571
96959 93.606 -78.492 53.552
96960 93.629 -78.476 53.534
96961 93.651 -78.461 53.515
96962 93.669 -78.449 53.497
96963 93.681 -78.441 53.479
96964 93.693 -78.432 53.460
96969 93.770 -78.395 53.367
96970 93.786 -78.388 53.348
96971 93.801 -78.381 53.330
96972 93.817 -78.373 53.309
96973 93.832 -78.366 53.290
96974 93.847 -78.359 53.271
96975 93.863 -78.351 53.253
96976 93.878 -78.344 53.234
96977 93.894 -78.337 53.215
96978 93.909 -78.329 53.198
96979 93.923 -78.325 53.179
96980 93.931 -78.314 53.161
96981 93.930 -78.297 53.144
96982 93.946 -78.283 53.129
96983 93.959 -78.291 53.116
96984 93.973 -78.300 53.102
96985 93.985 -78.308 53.088
96986 93.994 -78.312 53.074
96987 94.002 -78.316 53.059
96988 94.010 -78.319 53.044
96989 94.018 -78.322 53.032
96990 94.026 -78.326 53.020
96991 94.035 -78.330 53.007
96992 94.043 -78.334 52.994
96993 94.051 -78.338 52.981
96994 94.059 -78.341 52.967
96995 94.069 -78.345 52.953
96996 94.080 -78.350 52.940
96997 94.090 -78.354 52.928
96998 94.099 -78.360 52.915
96999 94.104 -78.369 52.902
97000 94.108 -78.377 52.889
97001 94.113 -78.386 52.877
97002 94.118 -78.395 52.868
97003 94.122 -78.404 52.859
97004 94.127 -78.413 52.853
97005 94.132 -78.422 52.846
97006 94.136 -78.428 52.837
97007 94.141 -78.430 52.824
97008 94.145 -78.432 52.813
97009 94.150 -78.435 52.803
97010 94.154 -78.437 52.793
97011 94.159 -78.439 52.783
97012 94.163 -78.432 52.772
97013 94.174 -78.414 52.757
97014 94.192 -78.401 52.737
97015 94.210 -78.387 52.715
97016 94.228 -78.373 52.691
97017 94.250 -78.374 52.666
97018 94.271 -78.376 52.640
97019 94.291 -78.379 52.613
97020 94.301 -78.394 52.586
97021 94.310 -78.411 52.555
97022 94.319 -78.429 52.524
97023 94.327 -78.446 52.492
97024 94.336 -78.463 52.461
97025 94.344 -78.480 52.428
97026 94.359 -78.483 52.394
97027 94.381 -78.475 52.360
97028 94.406 -78.473 52.323
97029 94.444 -78.487 52.286
97030 94.472 -78.479 52.249
97031 94.496 -78.463 52.214
97032 94.521 -78.446 52.178
97033 94.545 -78.430 52.143
97034 94.570 -78.413 52.107
97035 94.594 -78.397 52.064
97036 94.611 -78.386 52.021
97037 94.622 -78.380 51.976
97038 94.632 -78.375 51.932
97039 94.642 -78.369 51.887
97040 94.653 -78.363 51.840
97041 94.665 -78.362 51.791
97042 94.677 -78.364 51.743
97043 94.689 -78.366 51.696
97044 94.701 -78.368 51.651
97045 94.713 -78.369 51.605
97046 94.726 -78.366 51.558
97047 94.739 -78.363 51.512
97048 94.750 -78.367 51.466
97049 94.760 -78.371 51.420
97050 94.778 -78.364 51.377
97051 94.797 -78.357 51.337
97052 94.811 -78.358 51.300
97053 94.823 -78.362 51.264
97054 94.838 -78.366 51.225
97055 94.863 -78.368 51.184
97056 94.888 -78.369 51.136
97057 94.913 -78.371 51.086
97058 94.928 -78.373 51.034
97059 94.939 -78.374 50.983
97060 94.950 -78.376 50.934
97061 94.960 -78.377 50.885
97062 94.971 -78.379 50.836
97063 94.982 -78.380 50.787
97064 94.992 -78.382 50.739
97065 95.003 -78.383 50.692
97066 95.014 -78.385 50.645
97067 95.025 -78.385 50.599
97068 95.034 -78.382 50.555
97069 95.051 -78.340 50.510
97070 95.093 -78.288 50.467
97071 95.126 -78.244 50.426
97072 95.162 -78.266 50.385
97073 95.198 -78.289 50.345
97074 95.218 -78.291 50.305
97075 95.229 -78.279 50.265
97076 95.239 -78.268 50.227
97077 95.250 -78.257 50.189
97078 95.260 -78.246 50.152
97079 95.271 -78.235 50.116
97080 95.281 -78.224 50.079
97081 95.293 -78.213 50.040
97082 95.303 -78.201 50.002
97083 95.313 -78.193 49.965
97084 95.323 -78.195 49.928
97085 95.332 -78.198 49.890
97086 95.341 -78.201 49.854
97087 95.351 -78.204 49.819
97088 95.360 -78.207 49.784
97089 95.370 -78.210 49.749
97090 95.379 -78.213 49.715
97091 95.388 -78.216 49.684
97092 95.398 -78.219 49.652
97093 95.407 -78.221 49.621
97094 95.416 -78.224 49.590
97095 95.431 -78.231 49.561
97096 95.456 -78.242 49.532
97097 95.479 -78.254 49.502
97098 95.503 -78.266 49.472
97099 95.513 -78.269 49.444
97100 95.518 -78.268 49.415
97101 95.523 -78.268 49.384
97102 95.529 -78.269 49.351
97103 95.538 -78.276 49.318
97104 95.547 -78.283 49.287
97105 95.556 -78.290 49.256
97106 95.565 -78.297 49.225
97107 95.574 -78.304 49.193
97108 95.584 -78.312 49.160
97109 95.593 -78.319 49.123
97110 95.602 -78.326 49.085
97111 95.611 -78.333 49.047
97112 95.620 -78.340 49.009
97113 95.630 -78.347 48.973
97114 95.629 -78.346 48.938
97115 95.625 -78.343 48.903
97116 95.620 -78.339 48.865
97117 95.616 -78.335 48.825
97118 95.612 -78.332 48.784
97119 95.620 -78.336 48.742
97120 95.634 -78.344 48.700
97121 95.650 -78.353 48.657
97122 95.668 -78.354 48.615
97123 95.686 -78.356 48.572
97124 95.700 -78.360 48.530
97125 95.712 -78.361 48.489
97126 95.725 -78.363 48.449
97127 95.740 -78.366 48.408
97128 95.762 -78.373 48.367
97129 95.783 -78.376 48.325
97130 95.798 -78.377 48.284
97131 95.814 -78.377 48.242
97132 95.830 -78.376 48.202
97133 95.846 -78.376 48.160
97134 95.858 -78.376 48.118
97135 95.875 -78.378 48.077
97136 95.892 -78.379 48.035
97137 95.908 -78.382 47.995
97138 95.920 -78.382 47.955
97139 95.934 -78.385 47.915
97140 95.948 -78.388 47.876
97141 95.963 -78.392 47.836
97142 95.975 -78.398 47.798
97143 95.987 -78.404 47.761
97144 96.001 -78.412 47.724
97145 96.014 -78.420 47.688
97146 96.027 -78.429 47.652
97147 96.042 -78.441 47.620
97148 96.056 -78.453 47.588
97149 96.071 -78.465 47.555
97150 96.077 -78.473 47.520
97151 96.077 -78.489 47.486
97152 96.076 -78.504 47.453
97153 96.075 -78.520 47.422
97154 96.090 -78.525 47.391
97155 96.115 -78.530 47.359
97156 96.144 -78.536 47.328
97157 96.174 -78.542 47.299
97158 96.203 -78.547 47.269
97159 96.232 -78.553 47.239
97160 96.261 -78.558 47.211
97161 96.291 -78.564 47.184
97162 96.320 -78.570 47.157
97163 96.349 -78.576 47.130
97164 96.379 -78.579 47.101
97165 96.408 -78.582 47.070
97166 96.437 -78.586 47.039
97167 96.466 -78.589 47.010
97168 96.496 -78.591 46.981
97169 96.525 -78.593 46.953
97170 96.554 -78.596 46.926
97171 96.584 -78.598 46.901
97172 96.613 -78.601 46.874
97173 96.629 -78.597 46.847
97174 96.626 -78.582 46.820
97175 96.675 -78.592 46.793
97176 96.724 -78.602 46.769
97177 96.745 -78.613 46.747
97178 96.745 -78.624 46.725
97179 96.746 -78.635 46.702
97180 96.746 -78.646 46.678
97181 96.749 -78.657 46.654
97182 96.769 -78.667 46.630
97183 96.780 -78.682 46.608
97184 96.786 -78.698 46.586
97185 96.799 -78.713 46.565
97186 96.811 -78.727 46.543
97187 96.822 -78.739 46.521
97188 96.832 -78.750 46.499
97189 96.842 -78.761 46.480
97190 96.853 -78.772 46.461
97191 96.868 -78.789 46.441
97192 96.887 -78.812 46.422
97193 96.906 -78.836 46.403
97194 96.926 -78.859 46.385
97195 96.947 -78.872 46.366
97196 96.977 -78.882 46.349
97197 97.007 -78.893 46.333
97198 97.036 -78.903 46.319
97199 97.066 -78.914 46.303
97200 97.096 -78.924 46.286
97201 97.125 -78.935 46.271
97202 97.155 -78.946 46.257
97203 97.185 -78.956 46.243
97204 97.215 -78.966 46.228
97205 97.244 -78.977 46.213
97206 97.277 -78.979 46.196
97207 97.312 -78.975 46.177
97208 97.348 -78.977 46.157
97209 97.385 -78.990 46.137
97210 97.406 -78.993 46.118
97211 97.422 -78.992 46.099
97212 97.437 -78.991 46.080
97213 97.453 -78.990 46.063
97214 97.463 -78.985 46.043
97215 97.471 -78.979 46.023
97216 97.479 -78.973 46.003
97217 97.487 -78.968 45.986
97218 97.496 -78.962 45.970
97219 97.504 -78.955 45.953
97220 97.512 -78.949 45.937
97221 97.520 -78.943 45.922
97222 97.528 -78.938 45.908
97223 97.536 -78.932 45.894
97224 97.544 -78.926 45.882
97225 97.553 -78.920 45.869
97226 97.561 -78.914 45.856
97227 97.572 -78.908 45.842
97228 97.590 -78.905 45.825
97229 97.608 -78.902 45.807
97230 97.621 -78.899 45.786
97231 97.635 -78.899 45.765
97232 97.656 -78.936 45.745
97233 97.678 -78.974 45.726
97234 97.685 -78.982 45.707
97235 97.688 -78.979 45.689
97236 97.690 -78.977 45.671
97237 97.693 -78.974 45.653
97238 97.696 -78.971 45.637
97239 97.700 -78.970 45.621
97240 97.736 -78.993 45.605
97241 97.771 -79.015 45.589
97242 97.778 -79.021 45.573
97243 97.785 -79.026 45.558
97244 97.792 -79.031 45.544
97245 97.800 -79.036 45.531
97246 97.807 -79.041 45.518
97247 97.814 -79.046 45.505
97248 97.821 -79.051 45.492
97249 97.829 -79.056 45.478
97250 97.836 -79.062 45.466
97251 97.843 -79.069 45.454
97252 97.846 -79.083 45.442
97253 97.850 -79.096 45.428
97254 97.857 -79.109 45.416
97255 97.873 -79.115 45.405
97256 97.890 -79.122 45.395
97257 97.907 -79.129 45.385
97258 97.923 -79.134 45.374
97259 97.938 -79.124 45.364
97260 97.953 -79.113 45.352
97261 97.967 -79.103 45.340
97262 97.983 -79.093 45.328
97263 97.998 -79.082 45.317
97264 98.004 -79.087 45.307
97265 98.045 -79.089 45.297
97266 98.096 -79.091 45.287
97267 98.148 -79.103 45.278
97268 98.177 -79.118 45.270
97269 98.202 -79.132 45.262
97270 98.227 -79.145 45.254
97271 98.252 -79.159 45.247
97272 98.277 -79.172 45.242
97273 98.302 -79.185 45.234
97274 98.327 -79.198 45.226
97275 98.353 -79.203 45.217
97276 98.379 -79.206 45.207
97277 98.407 -79.212 45.197
97278 98.440 -79.225 45.187
97279 98.477 -79.237 45.175
97280 98.513 -79.249 45.166
97281 98.550 -79.261 45.155
97282 98.586 -79.273 45.143
97283 98.623 -79.285 45.132
97284 98.660 -79.297 45.121
97285 98.696 -79.310 45.110
97286 98.733 -79.322 45.098
97287 98.770 -79.335 45.089
97288 98.806 -79.347 45.079
97289 98.836 -79.357 45.070
97290 98.836 -79.359 45.061
97291 98.844 -79.366 45.053
97292 98.867 -79.383 45.045
97293 98.890 -79.400 45.038
97294 98.914 -79.418 45.032
97295 98.938 -79.435 45.026
97296 98.961 -79.453 45.020
97297 98.985 -79.470 45.014
97298 99.008 -79.487 45.006
97299 99.041 -79.494 44.998
97300 99.088 -79.486 44.988
97301 99.129 -79.476 44.977
97302 99.139 -79.482 44.967
97303 99.148 -79.488 44.956
97304 99.158 -79.494 44.944
97305 99.168 -79.499 44.933
97306 99.178 -79.505 44.922
97307 99.190 -79.515 44.913
97308 99.205 -79.529 44.901
97309 99.220 -79.545 44.890
97310 99.235 -79.559 44.881
97311 99.250 -79.573 44.873
97312 99.265 -79.588 44.866
97313 99.284 -79.598 44.859
97314 99.306 -79.607 44.851
97315 99.328 -79.615 44.842
97316 99.349 -79.623 44.833
97317 99.371 -79.632 44.824
97318 99.387 -79.629 44.813
97319 99.403 -79.623 44.803
97320 99.418 -79.617 44.793
97321 99.433 -79.612 44.784
97322 99.448 -79.607 44.775
97323 99.464 -79.601 44.768
97324 99.479 -79.595 44.760
97325 99.495 -79.590 44.752
97326 99.515 -79.588 44.745
97327 99.546 -79.592 44.738
97328 99.576 -79.595 44.730
97329 99.606 -79.598 44.723
97330 99.636 -79.601 44.717
97331 99.666 -79.607 44.712
97332 99.695 -79.638 44.706
97333 99.712 -79.663 44.701
97334 99.730 -79.688 44.696
97335 99.748 -79.713 44.691
97336 99.765 -79.739 44.685
97337 99.782 -79.764 44.679
97338 99.800 -79.789 44.673
97339 99.818 -79.814 44.666
97340 99.835 -79.839 44.659
97341 99.852 -79.864 44.653
97342 99.870 -79.889 44.648
97343 99.894 -79.916 44.640
97344 99.926 -79.947 44.633
97345 99.958 -79.978 44.626
97346 99.990 -80.009 44.618
97347 100.023 -80.041 44.611
97348 100.055 -80.072 44.603
97349 100.093 -80.098 44.594
97350 100.139 -80.116 44.584
97351 100.185 -80.134 44.574
97352 100.230 -80.153 44.563
97353 100.276 -80.171 44.552
97354 100.318 -80.192 44.540
97355 100.345 -80.219 44.526
97356 100.364 -80.252 44.512
97357 100.387 -80.295 44.495
97358 100.430 -80.329 44.478
97359 100.482 -80.362 44.461
97360 100.517 -80.411 44.443
97361 100.543 -80.455 44.425
97362 100.628 -80.530 44.407
97363 100.607 -80.585 44.388
97364 100.545 -80.563 44.367
97365 100.579 -80.578 44.347
97366 100.620 -80.600 44.328
97367 100.660 -80.622 44.312
97368 100.700 -80.644 44.295
97369 100.741 -80.666 44.279
97370 100.781 -80.688 44.263
97371 100.821 -80.710 44.246
97372 100.861 -80.732 44.230
97373 100.901 -80.754 44.213
97374 100.941 -80.776 44.197
97375 100.986 -80.792 44.181
97376 101.033 -80.804 44.165
97377 101.079 -80.817 44.150
97378 101.125 -80.830 44.136
97379 101.172 -80.843 44.121
97380 101.218 -80.856 44.107
97381 101.265 -80.869 44.093
97382 101.312 -80.882 44.079
97383 101.358 -80.895 44.065
97645 93.503 -78.578 53.812
97646 93.513 -78.576 53.798
97647 93.523 -78.573 53.783
97648 93.533 -78.570 53.768
97649 93.542 -78.568 53.753
97650 93.552 -78.566 53.736
97651 93.562 -78.563 53.718
97652 93.571 -78.561 53.700
97653 93.576 -78.555 53.682
97654 93.575 -78.546 53.664
97655 93.574 -78.537 53.645
97656 93.572 -78.528 53.626
97657 93.571 -78.518 53.607
97658 93.575 -78.506 53.588
97659 93.589 -78.491 53.569
97660 93.612 -78.476 53.549
97661 93.634 -78.460 53.530
97662 93.656 -78.445 53.511
97663 93.671 -78.435 53.493
97664 93.682 -78.428 53.474
97673 93.819 -78.360 53.304
97674 93.834 -78.354 53.285
97675 93.849 -78.346 53.266
97676 93.864 -78.338 53.246
97677 93.880 -78.332 53.227
97678 93.896 -78.324 53.209
97679 93.911 -78.316 53.193
97680 93.926 -78.310 53.174
97681 93.941 -78.304 53.157
97682 93.942 -78.288 53.140
97683 93.956 -78.280 53.125
97684 93.969 -78.289 53.113
97685 93.981 -78.295 53.102
97686 93.989 -78.299 53.089
97687 93.997 -78.303 53.076
97688 94.005 -78.307 53.061
97689 94.013 -78.310 53.047
97690 94.022 -78.314 53.035
97691 94.030 -78.317 53.023
97692 94.038 -78.321 53.011
97693 94.046 -78.325 52.998
97694 94.054 -78.329 52.985
97695 94.062 -78.332 52.971
97696 94.071 -78.336 52.957
97697 94.082 -78.341 52.944
97698 94.091 -78.346 52.931
97699 94.096 -78.355 52.918
97700 94.100 -78.363 52.903
97701 94.105 -78.373 52.890
97702 94.110 -78.382 52.878
97703 94.114 -78.390 52.868
97704 94.119 -78.399 52.857
97705 94.123 -78.408 52.850
97706 94.128 -78.416 52.843
97707 94.133 -78.426 52.834
97708 94.137 -78.429 52.822
97709 94.142 -78.432 52.811
97710 94.146 -78.434 52.801
97711 94.151 -78.436 52.792
97712 94.155 -78.438 52.780
97713 94.158 -78.420 52.766
97714 94.164 -78.399 52.750
97715 94.183 -78.385 52.728
97716 94.201 -78.372 52.705
97717 94.219 -78.360 52.681
97718 94.241 -78.362 52.656
97719 94.262 -78.363 52.629
97720 94.282 -78.366 52.601
97721 94.297 -78.375 52.574
97722 94.306 -78.392 52.542
97723 94.315 -78.410 52.510
97724 94.323 -78.427 52.478
97725 94.332 -78.444 52.446
97726 94.340 -78.461 52.413
97727 94.359 -78.456 52.379
97728 94.381 -78.448 52.345
97729 94.402 -78.439 52.307
97730 94.439 -78.453 52.268
97731 94.476 -78.467 52.230
97732 94.501 -78.451 52.191
97733 94.526 -78.435 52.153
97734 94.550 -78.419 52.115
97735 94.574 -78.402 52.077
97736 94.599 -78.385 52.035
97737 94.616 -78.374 51.992
97738 94.627 -78.369 51.948
97739 94.637 -78.363 51.903
97740 94.648 -78.357 51.856
97741 94.659 -78.351 51.807
97742 94.669 -78.346 51.757
97743 94.681 -78.347 51.710
97744 94.693 -78.350 51.664
97745 94.705 -78.351 51.618
97746 94.718 -78.352 51.572
97747 94.731 -78.349 51.527
97748 94.744 -78.346 51.483
97749 94.755 -78.348 51.439
97750 94.765 -78.352 51.398
97751 94.778 -78.354 51.359
97752 94.796 -78.346 51.323
97753 94.810 -78.347 51.289
97754 94.823 -78.351 51.252
97755 94.844 -78.351 51.213
97756 94.866 -78.351 51.170
97757 94.889 -78.351 51.121
97758 94.913 -78.352 51.070
97759 94.938 -78.354 51.018
97760 94.949 -78.355 50.967
97761 94.960 -78.357 50.918
97762 94.971 -78.358 50.870
97763 94.981 -78.360 50.823
97764 94.992 -78.361 50.775
97765 95.002 -78.361 50.728
97766 95.011 -78.357 50.681
97767 95.020 -78.354 50.635
97768 95.029 -78.349 50.591
97769 95.039 -78.338 50.547
97770 95.061 -78.278 50.503
97771 95.083 -78.219 50.461
97772 95.111 -78.183 50.421
97773 95.126 -78.184 50.381
97774 95.162 -78.207 50.342
97775 95.198 -78.229 50.302
97776 95.234 -78.252 50.264
97777 95.246 -78.243 50.227
97778 95.256 -78.232 50.190
97779 95.267 -78.220 50.153
97780 95.277 -78.210 50.117
97781 95.288 -78.198 50.079
97782 95.299 -78.187 50.041
97783 95.309 -78.175 50.004
97784 95.319 -78.171 49.967
97785 95.329 -78.174 49.929
97786 95.338 -78.176 49.893
97787 95.347 -78.179 49.857
97788 95.356 -78.182 49.823
97789 95.366 -78.185 49.790
97790 95.376 -78.188 49.757
97791 95.385 -78.191 49.726
97792 95.394 -78.194 49.693
97793 95.404 -78.197 49.661
97794 95.413 -78.199 49.629
97795 95.422 -78.202 49.600
97796 95.431 -78.205 49.572
97797 95.447 -78.211 49.543
97798 95.470 -78.223 49.514
97799 95.494 -78.235 49.484
97800 95.503 -78.238 49.455
97801 95.509 -78.237 49.427
97802 95.515 -78.238 49.397
97803 95.524 -78.245 49.365
97804 95.533 -78.253 49.333
97805 95.542 -78.260 49.303
97806 95.551 -78.266 49.272
97807 95.560 -78.274 49.240
97808 95.570 -78.281 49.208
97809 95.579 -78.288 49.174
97810 95.588 -78.295 49.138
97811 95.597 -78.302 49.100
97812 95.606 -78.310 49.060
97813 95.616 -78.316 49.022
97814 95.625 -78.323 48.984
97815 95.628 -78.326 48.946
97816 95.623 -78.322 48.911
97817 95.619 -78.319 48.874
97818 95.615 -78.315 48.835
97819 95.611 -78.312 48.795
97820 95.625 -78.319 48.753
97821 95.639 -78.328 48.712
97822 95.654 -78.336 48.670
97823 95.671 -78.341 48.629
97824 95.689 -78.342 48.586
97825 95.703 -78.343 48.543
97826 95.716 -78.344 48.500
97827 95.728 -78.347 48.459
97828 95.740 -78.349 48.419
97829 95.762 -78.355 48.377
97830 95.780 -78.357 48.335
97831 95.796 -78.357 48.293
97832 95.811 -78.357 48.251
97833 95.826 -78.357 48.209
97834 95.842 -78.358 48.168
97835 95.860 -78.360 48.126
97836 95.876 -78.362 48.085
97837 95.893 -78.363 48.044
97838 95.910 -78.366 48.003
97839 95.922 -78.366 47.963
97840 95.937 -78.370 47.925
97841 95.951 -78.374 47.888
97842 95.965 -78.378 47.850
97843 95.979 -78.382 47.813
97844 95.991 -78.388 47.776
97845 96.005 -78.396 47.740
97846 96.019 -78.406 47.704
97847 96.033 -78.418 47.667
97848 96.047 -78.430 47.635
97849 96.062 -78.442 47.602
97850 96.076 -78.454 47.569
97851 96.086 -78.461 47.535
97852 96.085 -78.476 47.501
97853 96.084 -78.492 47.468
97854 96.085 -78.506 47.437
97855 96.102 -78.510 47.407
97856 96.118 -78.514 47.376
97857 96.146 -78.520 47.346
97858 96.175 -78.526 47.317
97859 96.204 -78.531 47.287
97860 96.234 -78.537 47.257
97861 96.263 -78.542 47.229
97862 96.292 -78.548 47.201
97863 96.321 -78.554 47.174
97864 96.350 -78.558 47.147
97865 96.380 -78.561 47.118
97866 96.409 -78.564 47.087
97867 96.438 -78.568 47.057
97868 96.467 -78.570 47.026
97869 96.497 -78.573 46.998
97870 96.526 -78.575 46.971
97871 96.555 -78.578 46.944
97872 96.585 -78.580 46.918
97873 96.614 -78.582 46.891
97874 96.643 -78.585 46.864
97875 96.672 -78.587 46.837
97876 96.681 -78.579 46.810
97877 96.730 -78.589 46.785
97878 96.735 -78.600 46.762
97879 96.736 -78.611 46.739
97880 96.736 -78.622 46.716
97881 96.737 -78.633 46.692
97882 96.737 -78.644 46.669
97883 96.756 -78.655 46.645
97884 96.761 -78.671 46.623
97885 96.774 -78.686 46.601
97886 96.787 -78.700 46.579
97887 96.799 -78.715 46.557
97888 96.809 -78.726 46.535
97889 96.819 -78.736 46.513
97890 96.830 -78.747 46.494
97891 96.846 -78.766 46.474
97892 96.865 -78.789 46.455
97893 96.885 -78.812 46.436
97894 96.904 -78.836 46.418
97895 96.923 -78.859 46.399
97896 96.943 -78.880 46.381
97897 96.969 -78.886 46.364
97898 96.999 -78.896 46.348
97899 97.029 -78.906 46.333
97900 97.058 -78.917 46.317
97901 97.088 -78.927 46.301
97902 97.118 -78.938 46.286
97903 97.148 -78.949 46.271
97904 97.177 -78.959 46.256
97905 97.207 -78.970 46.241
97906 97.237 -78.980 46.226
97907 97.272 -78.976 46.208
97908 97.306 -78.973 46.189
97909 97.341 -78.970 46.169
97910 97.377 -78.975 46.148
97911 97.412 -78.986 46.129
97912 97.428 -78.985 46.110
97913 97.443 -78.984 46.092
97914 97.459 -78.984 46.074
97915 97.470 -78.980 46.055
97916 97.478 -78.974 46.035
97917 97.487 -78.968 46.015
97918 97.495 -78.962 45.997
97919 97.503 -78.956 45.980
97920 97.511 -78.950 45.963
97921 97.519 -78.944 45.946
97922 97.528 -78.938 45.930
97923 97.536 -78.932 45.915
97924 97.544 -78.927 45.901
97925 97.552 -78.921 45.886
97926 97.561 -78.915 45.873
97927 97.573 -78.910 45.860
97928 97.590 -78.906 45.845
97929 97.608 -78.903 45.829
97930 97.626 -78.899 45.811
97931 97.640 -78.897 45.792
97932 97.653 -78.895 45.774
97933 97.666 -78.893 45.755
97934 97.688 -78.930 45.737
97935 97.706 -78.961 45.719
97936 97.709 -78.959 45.702
97937 97.711 -78.955 45.685
97938 97.714 -78.952 45.667
97939 97.717 -78.950 45.653
97940 97.740 -78.964 45.637
97941 97.775 -78.987 45.621
97942 97.783 -78.992 45.603
97943 97.790 -78.996 45.586
97944 97.797 -79.002 45.571
97945 97.804 -79.007 45.556
97946 97.812 -79.012 45.544
97947 97.819 -79.017 45.531
97948 97.826 -79.022 45.518
97949 97.833 -79.028 45.505
97950 97.841 -79.033 45.492
97951 97.848 -79.037 45.479
97952 97.855 -79.043 45.466
97953 97.860 -79.056 45.453
97954 97.864 -79.069 45.439
97955 97.872 -79.080 45.427
97956 97.889 -79.087 45.416
97957 97.906 -79.093 45.405
97958 97.922 -79.100 45.394
97959 97.938 -79.107 45.383
97960 97.955 -79.111 45.373
97961 97.969 -79.101 45.361
97962 97.985 -79.090 45.348
97963 98.000 -79.080 45.336
97964 98.014 -79.070 45.323
97965 98.021 -79.074 45.311
97966 98.045 -79.078 45.300
97967 98.095 -79.080 45.291
97968 98.145 -79.081 45.282
97969 98.177 -79.096 45.273
97970 98.203 -79.105 45.265
97971 98.228 -79.115 45.256
97972 98.253 -79.124 45.249
97973 98.279 -79.134 45.242
97974 98.304 -79.143 45.235
97975 98.329 -79.153 45.227
97976 98.355 -79.162 45.218
97977 98.379 -79.174 45.209
97978 98.405 -79.184 45.200
97979 98.441 -79.197 45.190
97980 98.478 -79.209 45.179
97981 98.515 -79.221 45.169
97982 98.552 -79.234 45.158
97983 98.588 -79.246 45.147
97984 98.625 -79.258 45.136
97985 98.661 -79.270 45.125
97986 98.698 -79.282 45.113
97987 98.734 -79.294 45.102
97988 98.771 -79.306 45.093
97989 98.808 -79.319 45.083
97990 98.844 -79.332 45.073
97991 98.850 -79.335 45.063
97992 98.860 -79.343 45.055
97993 98.883 -79.360 45.048
97994 98.907 -79.378 45.041
97995 98.930 -79.395 45.034
97996 98.954 -79.413 45.027
97997 98.978 -79.430 45.021
97998 99.001 -79.447 45.015
97999 99.025 -79.465 45.007
98000 99.057 -79.473 44.999
98001 99.108 -79.457 44.989
98002 99.124 -79.460 44.979
98003 99.134 -79.466 44.969
98004 99.144 -79.472 44.959
98005 99.153 -79.477 44.949
98006 99.163 -79.483 44.938
98007 99.173 -79.488 44.926
98008 99.185 -79.498 44.915
98009 99.200 -79.513 44.903
98010 99.215 -79.528 44.892
98011 99.230 -79.542 44.882
98012 99.245 -79.557 44.874
98013 99.264 -79.568 44.866
98014 99.285 -79.576 44.858
98015 99.307 -79.585 44.850
98016 99.329 -79.593 44.841
98017 99.350 -79.602 44.832
98018 99.372 -79.610 44.823
98019 99.393 -79.617 44.814
98020 99.408 -79.612 44.804
98021 99.424 -79.607 44.795
98022 99.439 -79.601 44.786
98023 99.454 -79.595 44.778
98024 99.470 -79.590 44.771
98025 99.485 -79.585 44.764
98026 99.500 -79.579 44.756
98027 99.530 -79.582 44.749
98028 99.560 -79.585 44.743
98029 99.590 -79.588 44.736
98030 99.620 -79.592 44.729
98031 99.650 -79.595 44.724
98032 99.680 -79.598 44.719
98033 99.710 -79.604 44.713
98034 99.744 -79.636 44.708
98035 99.761 -79.661 44.703
98036 99.779 -79.686 44.698
98037 99.796 -79.711 44.692
98038 99.814 -79.736 44.687
98039 99.831 -79.761 44.681
98040 99.849 -79.786 44.674
98041 99.866 -79.811 44.668
98042 99.884 -79.836 44.661
98043 99.900 -79.862 44.656
98044 99.931 -79.899 44.649
98045 99.957 -79.924 44.641
98046 99.989 -79.955 44.633
98047 100.021 -79.987 44.626
98048 100.054 -80.018 44.618
98049 100.086 -80.049 44.610
98050 100.123 -80.075 44.603
98051 100.169 -80.093 44.594
98052 100.215 -80.112 44.584
98053 100.261 -80.131 44.574
98054 100.307 -80.149 44.563
98055 100.352 -80.168 44.552
98056 100.392 -80.201 44.539
98057 100.431 -80.239 44.525
98058 100.464 -80.279 44.511
98059 100.487 -80.322 44.495
98060 100.529 -80.359 44.478
98061 100.564 -80.408 44.460
98062 100.563 -80.458 44.442
98063 100.536 -80.509 44.424
98064 100.532 -80.554 44.406
98065 100.554 -80.582 44.385
98066 100.583 -80.591 44.365
98067 100.616 -80.602 44.346
98068 100.656 -80.624 44.327
98069 100.696 -80.646 44.309
98070 100.737 -80.668 44.291
98071 100.777 -80.690 44.273
98072 100.817 -80.712 44.256
98073 100.857 -80.734 44.239
98074 100.897 -80.756 44.222
98075 100.944 -80.770 44.204
98076 100.990 -80.782 44.188
98077 101.037 -80.795 44.172
98078 101.083 -80.808 44.157
98079 101.130 -80.821 44.142
98080 101.176 -80.834 44.127
98081 101.222 -80.847 44.113
98082 101.269 -80.860 44.098
98083 101.316 -80.873 44.084
98084 101.362 -80.886 44.070
98085 101.408 -80.898 44.055
98345 93.494 -78.570 53.829
98346 93.504 -78.567 53.816
98347 93.513 -78.564 53.802
98348 93.523 -78.562 53.787
98349 93.533 -78.560 53.771
98350 93.543 -78.557 53.755
98351 93.552 -78.554 53.738
98352 93.562 -78.552 53.720
98353 93.571 -78.550 53.701
98354 93.573 -78.542 53.682
98355 93.571 -78.533 53.663
98356 93.570 -78.523 53.644
98357 93.569 -78.514 53.624
98358 93.568 -78.505 53.605
98359 93.577 -78.490 53.586
98360 93.594 -78.475 53.566
98361 93.617 -78.460 53.546
98362 93.639 -78.445 53.526
98363 93.660 -78.431 53.507
98364 93.672 -78.423 53.488
98365 93.683 -78.414 53.469
98375 93.836 -78.341 53.280
98376 93.851 -78.333 53.260
98377 93.866 -78.326 53.241
98378 93.882 -78.319 53.222
98379 93.897 -78.311 53.204
98380 93.913 -78.304 53.187
98381 93.928 -78.297 53.169
98382 93.943 -78.289 53.152
98383 93.954 -78.280 53.135
98384 93.966 -78.278 53.121
98385 93.976 -78.283 53.109
98386 93.984 -78.287 53.098
98387 93.992 -78.290 53.088
98388 94.000 -78.294 53.076
98389 94.008 -78.297 53.063
98390 94.016 -78.301 53.049
98391 94.025 -78.305 53.036
98392 94.033 -78.309 53.025
98393 94.041 -78.313 53.013
98394 94.049 -78.316 53.001
98395 94.057 -78.319 52.988
98396 94.066 -78.323 52.975
98397 94.074 -78.327 52.961
98398 94.082 -78.332 52.948
98399 94.087 -78.341 52.934
98400 94.092 -78.350 52.920
98401 94.097 -78.359 52.904
98402 94.101 -78.367 52.890
98403 94.106 -78.376 52.878
98404 94.111 -78.385 52.866
98405 94.115 -78.394 52.854
98406 94.120 -78.403 52.846
98407 94.124 -78.412 52.838
98408 94.129 -78.420 52.828
98409 94.134 -78.429 52.816
98410 94.138 -78.431 52.805
98411 94.143 -78.432 52.795
98412 94.147 -78.435 52.786
98413 94.151 -78.429 52.772
98414 94.153 -78.407 52.757
98415 94.156 -78.385 52.739
98416 94.173 -78.371 52.717
98417 94.191 -78.357 52.694
98418 94.210 -78.347 52.669
98419 94.231 -78.349 52.643
98420 94.253 -78.351 52.615
98421 94.273 -78.353 52.588
98422 94.294 -78.356 52.560
98423 94.302 -78.373 52.528
98424 94.311 -78.391 52.496
98425 94.319 -78.408 52.464
98426 94.328 -78.425 52.431
98427 94.339 -78.438 52.398
98428 94.360 -78.429 52.364
98429 94.381 -78.421 52.328
98430 94.402 -78.413 52.289
98431 94.434 -78.420 52.248
98432 94.472 -78.433 52.207
98433 94.506 -78.440 52.164
98434 94.530 -78.423 52.123
98435 94.555 -78.407 52.083
98436 94.579 -78.391 52.044
98437 94.604 -78.374 52.002
98438 94.622 -78.363 51.960
98439 94.632 -78.357 51.918
98440 94.642 -78.351 51.874
98441 94.653 -78.345 51.826
98442 94.664 -78.339 51.776
98443 94.674 -78.333 51.725
98444 94.686 -78.332 51.676
98445 94.698 -78.333 51.630
98446 94.710 -78.335 51.585
98447 94.722 -78.335 51.540
98448 94.735 -78.332 51.497
98449 94.748 -78.329 51.456
98450 94.760 -78.329 51.414
98451 94.770 -78.332 51.375
98452 94.781 -78.336 51.340
98453 94.795 -78.335 51.308
98454 94.810 -78.335 51.274
98455 94.829 -78.336 51.237
98456 94.851 -78.336 51.196
98457 94.873 -78.336 51.152
98458 94.895 -78.335 51.102
98459 94.917 -78.335 51.051
98460 94.939 -78.335 51.001
98461 94.960 -78.336 50.952
98462 94.970 -78.338 50.903
98463 94.980 -78.337 50.856
98464 94.989 -78.333 50.810
98465 94.997 -78.329 50.763
98466 95.006 -78.325 50.717
98467 95.015 -78.321 50.670
98468 95.024 -78.317 50.625
98469 95.032 -78.313 50.582
98470 95.049 -78.276 50.539
98471 95.071 -78.216 50.497
98472 95.100 -78.175 50.456
98473 95.142 -78.169 50.416
98474 95.151 -78.164 50.376
98475 95.161 -78.160 50.339
98476 95.170 -78.156 50.300
98477 95.197 -78.170 50.263
98478 95.233 -78.193 50.227
98479 95.262 -78.206 50.190
98480 95.273 -78.195 50.153
98481 95.283 -78.184 50.116
98482 95.294 -78.172 50.078
98483 95.305 -78.162 50.041
98484 95.316 -78.150 50.005
98485 95.325 -78.149 49.968
98486 95.334 -78.152 49.932
98487 95.344 -78.154 49.896
98488 95.353 -78.157 49.862
98489 95.362 -78.160 49.829
98490 95.372 -78.163 49.797
98491 95.382 -78.166 49.766
98492 95.391 -78.169 49.735
98493 95.400 -78.172 49.702
98494 95.409 -78.175 49.670
98495 95.419 -78.178 49.638
98496 95.428 -78.180 49.611
98497 95.438 -78.183 49.583
98498 95.447 -78.186 49.555
98499 95.461 -78.192 49.526
98500 95.485 -78.203 49.497
98501 95.494 -78.206 49.469
98502 95.500 -78.207 49.441
98503 95.510 -78.215 49.412
98504 95.519 -78.222 49.382
98505 95.528 -78.228 49.351
98506 95.537 -78.236 49.321
98507 95.546 -78.243 49.290
98508 95.556 -78.250 49.257
98509 95.565 -78.257 49.223
98510 95.574 -78.264 49.189
98511 95.583 -78.272 49.153
98512 95.593 -78.279 49.115
98513 95.602 -78.285 49.074
98514 95.611 -78.293 49.034
98515 95.620 -78.300 48.994
98516 95.626 -78.305 48.955
98517 95.622 -78.301 48.918
98518 95.618 -78.298 48.881
98519 95.614 -78.294 48.843
98520 95.615 -78.294 48.803
98521 95.629 -78.303 48.763
98522 95.644 -78.311 48.722
98523 95.659 -78.319 48.681
98524 95.674 -78.328 48.640
98525 95.691 -78.329 48.596
98526 95.704 -78.329 48.552
98527 95.718 -78.330 48.509
98528 95.731 -78.331 48.468
98529 95.743 -78.332 48.427
98530 95.762 -78.337 48.386
98531 95.777 -78.337 48.344
98532 95.793 -78.337 48.303
98533 95.810 -78.333 48.260
98534 95.827 -78.332 48.217
98535 95.845 -78.340 48.176
98536 95.862 -78.343 48.135
98537 95.878 -78.345 48.094
98538 95.894 -78.347 48.054
98539 95.911 -78.350 48.015
98540 95.925 -78.352 47.976
98541 95.939 -78.356 47.938
98542 95.954 -78.360 47.901
98543 95.968 -78.363 47.864
98544 95.982 -78.366 47.828
98545 95.995 -78.373 47.791
98546 96.010 -78.384 47.755
98547 96.024 -78.396 47.719
98548 96.039 -78.408 47.681
98549 96.053 -78.420 47.649
98550 96.067 -78.432 47.616
98551 96.082 -78.445 47.583
98552 96.092 -78.451 47.549
98553 96.094 -78.463 47.516
98554 96.093 -78.479 47.483
98555 96.097 -78.491 47.451
98556 96.114 -78.495 47.422
98557 96.130 -78.500 47.392
98558 96.147 -78.504 47.363
98559 96.176 -78.510 47.334
98560 96.205 -78.516 47.304
98561 96.235 -78.521 47.275
98562 96.264 -78.526 47.245
98563 96.293 -78.532 47.218
98564 96.322 -78.536 47.190
98565 96.352 -78.540 47.163
98566 96.381 -78.543 47.134
98567 96.410 -78.547 47.105
98568 96.439 -78.550 47.074
98569 96.468 -78.552 47.044
98570 96.498 -78.554 47.016
98571 96.527 -78.557 46.989
98572 96.556 -78.559 46.963
98573 96.585 -78.562 46.936
98574 96.615 -78.564 46.909
98575 96.644 -78.567 46.881
98576 96.673 -78.569 46.853
98577 96.702 -78.571 46.828
98578 96.724 -78.576 46.802
98579 96.726 -78.587 46.777
98580 96.726 -78.598 46.754
98581 96.727 -78.609 46.730
98582 96.727 -78.620 46.706
98583 96.728 -78.631 46.683
98584 96.740 -78.643 46.660
98585 96.750 -78.658 46.638
98586 96.762 -78.673 46.616
98587 96.775 -78.688 46.593
98588 96.787 -78.701 46.571
98589 96.797 -78.712 46.549
98590 96.807 -78.723 46.528
98591 96.824 -78.742 46.508
98592 96.843 -78.766 46.488
98593 96.863 -78.789 46.470
98594 96.882 -78.812 46.451
98595 96.901 -78.836 46.432
98596 96.921 -78.859 46.414
98597 96.940 -78.882 46.395
98598 96.963 -78.891 46.379
98599 96.992 -78.899 46.362
98600 97.021 -78.910 46.345
98601 97.051 -78.921 46.329
98602 97.081 -78.931 46.314
98603 97.110 -78.942 46.299
98604 97.140 -78.952 46.283
98605 97.170 -78.962 46.268
98606 97.199 -78.973 46.253
98607 97.232 -78.977 46.237
98608 97.266 -78.974 46.219
98609 97.301 -78.971 46.200
98610 97.335 -78.967 46.181
98611 97.370 -78.964 46.159
98612 97.407 -78.973 46.141
98613 97.434 -78.978 46.123
98614 97.449 -78.977 46.105
98615 97.465 -78.977 46.087
98616 97.478 -78.974 46.068
98617 97.486 -78.968 46.048
98618 97.494 -78.962 46.028
98619 97.503 -78.956 46.010
98620 97.511 -78.951 45.991
98621 97.519 -78.945 45.973
98622 97.527 -78.939 45.956
98623 97.535 -78.933 45.939
98624 97.543 -78.927 45.923
98625 97.551 -78.921 45.906
98626 97.560 -78.915 45.891
98627 97.573 -78.911 45.877
98628 97.591 -78.907 45.862
98629 97.609 -78.904 45.847
98630 97.627 -78.900 45.832
98631 97.645 -78.897 45.816
98632 97.659 -78.894 45.797
98633 97.672 -78.892 45.781
98634 97.685 -78.890 45.764
98635 97.698 -78.887 45.747
98636 97.719 -78.924 45.729
98637 97.730 -78.937 45.713
98638 97.732 -78.934 45.698
98639 97.735 -78.932 45.683
98640 97.744 -78.934 45.668
98641 97.780 -78.957 45.652
98642 97.787 -78.962 45.636
98643 97.795 -78.968 45.617
98644 97.802 -78.973 45.600
98645 97.809 -78.978 45.584
98646 97.816 -78.983 45.569
98647 97.824 -78.988 45.557
98648 97.831 -78.993 45.544
98649 97.838 -78.999 45.532
98650 97.846 -79.003 45.520
98651 97.853 -79.009 45.506
98652 97.860 -79.014 45.493
98653 97.867 -79.019 45.480
98654 97.873 -79.028 45.466
98655 97.877 -79.042 45.452
98656 97.888 -79.051 45.439
98657 97.904 -79.058 45.428
98658 97.921 -79.065 45.416
98659 97.938 -79.071 45.404
98660 97.954 -79.078 45.391
98661 97.970 -79.085 45.380
98662 97.987 -79.088 45.368
98663 98.002 -79.078 45.356
98664 98.017 -79.068 45.343
98665 98.031 -79.057 45.329
98666 98.039 -79.060 45.317
98667 98.044 -79.065 45.305
98668 98.095 -79.068 45.296
98669 98.142 -79.077 45.286
98670 98.176 -79.095 45.276
98671 98.201 -79.104 45.268
98672 98.227 -79.114 45.259
98673 98.252 -79.123 45.250
98674 98.277 -79.133 45.243
98675 98.302 -79.142 45.236
98676 98.327 -79.152 45.227
98677 98.354 -79.160 45.219
98678 98.384 -79.164 45.210
98679 98.416 -79.169 45.202
98680 98.450 -79.178 45.193
98681 98.484 -79.187 45.182
98682 98.518 -79.196 45.172
98683 98.553 -79.206 45.161
98684 98.590 -79.218 45.151
98685 98.626 -79.231 45.140
98686 98.663 -79.243 45.128
98687 98.700 -79.255 45.117
98688 98.736 -79.267 45.107
98689 98.773 -79.279 45.098
98690 98.809 -79.291 45.087
98691 98.846 -79.303 45.076
98692 98.863 -79.311 45.066
98693 98.876 -79.321 45.059
98694 98.899 -79.338 45.053
98695 98.923 -79.355 45.046
98696 98.946 -79.372 45.038
98697 98.970 -79.390 45.031
98698 98.994 -79.407 45.023
98699 99.022 -79.419 45.017
98700 99.052 -79.429 45.009
98701 99.081 -79.438 45.000
98702 99.110 -79.438 44.991
98703 99.119 -79.444 44.981
98704 99.129 -79.450 44.971
98705 99.139 -79.455 44.961
98706 99.149 -79.460 44.951
98707 99.158 -79.466 44.940
98708 99.168 -79.472 44.928
98709 99.181 -79.482 44.916
98710 99.195 -79.496 44.904
98711 99.210 -79.511 44.893
98712 99.225 -79.526 44.882
98713 99.242 -79.538 44.873
98714 99.264 -79.546 44.865
98715 99.286 -79.554 44.857
98716 99.308 -79.563 44.849
98717 99.330 -79.572 44.840
98718 99.351 -79.580 44.831
98719 99.373 -79.588 44.822
98720 99.395 -79.597 44.814
98721 99.414 -79.601 44.805
98722 99.430 -79.595 44.796
98723 99.445 -79.590 44.788
98724 99.461 -79.585 44.781
98725 99.476 -79.579 44.774
98726 99.491 -79.574 44.767
98727 99.514 -79.573 44.760
98728 99.544 -79.576 44.753
98729 99.574 -79.579 44.747
98730 99.604 -79.582 44.741
98731 99.634 -79.585 44.735
98732 99.664 -79.588 44.730
98733 99.694 -79.592 44.725
98734 99.724 -79.595 44.720
98735 99.754 -79.599 44.715
98736 99.792 -79.634 44.710
98737 99.810 -79.659 44.705
98738 99.828 -79.684 44.699
98739 99.845 -79.709 44.694
98740 99.862 -79.734 44.688
98741 99.880 -79.759 44.681
98742 99.898 -79.784 44.675
98743 99.915 -79.810 44.669
98744 99.902 -79.846 44.664
98745 99.973 -79.886 44.657
98746 99.997 -79.909 44.650
98747 100.021 -79.933 44.644
98748 100.053 -79.963 44.636
98749 100.085 -79.994 44.628
98750 100.117 -80.025 44.620
98751 100.153 -80.052 44.612
98752 100.199 -80.071 44.604
98753 100.245 -80.090 44.595
98754 100.291 -80.108 44.586
98755 100.335 -80.129 44.575
98756 100.374 -80.167 44.565
98757 100.412 -80.206 44.553
98758 100.451 -80.243 44.539
98759 100.489 -80.281 44.526
98760 100.527 -80.316 44.510
98761 100.559 -80.344 44.494
98762 100.555 -80.390 44.477
98763 100.552 -80.435 44.459
98764 100.548 -80.479 44.441
98765 100.545 -80.524 44.424
98766 100.570 -80.559 44.405
98767 100.595 -80.593 44.387
98768 100.622 -80.618 44.368
98769 100.652 -80.626 44.347
98770 100.692 -80.648 44.326
98771 100.733 -80.670 44.306
98772 100.773 -80.692 44.285
98773 100.813 -80.714 44.267
98774 100.855 -80.734 44.249
98775 100.901 -80.747 44.231
98776 100.948 -80.760 44.213
98777 100.994 -80.773 44.196
98778 101.041 -80.786 44.180
98779 101.088 -80.798 44.164
98780 101.134 -80.811 44.148
98781 101.180 -80.824 44.133
98782 101.227 -80.837 44.118
98783 101.273 -80.850 44.104
98784 101.320 -80.863 44.089
98785 101.366 -80.876 44.075
98786 101.413 -80.889 44.060
98787 101.459 -80.902 44.047
99046 93.494 -78.558 53.833
99047 93.504 -78.556 53.820
99048 93.514 -78.554 53.806
99049 93.523 -78.551 53.791
99050 93.533 -78.548 53.775
99051 93.543 -78.546 53.758
99052 93.552 -78.544 53.740
99053 93.562 -78.542 53.722
99054 93.571 -78.539 53.703
99055 93.569 -78.529 53.683
99056 93.568 -78.520 53.663
99057 93.567 -78.510 53.643
99058 93.566 -78.501 53.623
99059 93.568 -78.490 53.603
99060 93.578 -78.475 53.584
99061 93.600 -78.460 53.563
99062 93.622 -78.445 53.542
99063 93.645 -78.429 53.522
99064 93.661 -78.418 53.502
99065 93.673 -78.410 53.483
99066 93.685 -78.401 53.464
99076 93.838 -78.328 53.274
99077 93.853 -78.320 53.255
99078 93.868 -78.313 53.235
99079 93.884 -78.306 53.216
99080 93.899 -78.298 53.198
99081 93.915 -78.291 53.180
99082 93.930 -78.284 53.163
99083 93.945 -78.276 53.146
99084 93.961 -78.269 53.130
99085 93.969 -78.272 53.116
99086 93.977 -78.276 53.104
99087 93.984 -78.280 53.093
99088 93.992 -78.285 53.086
99089 93.999 -78.288 53.075
99090 94.007 -78.293 53.063
99091 94.014 -78.297 53.050
99092 94.022 -78.301 53.037
99093 94.030 -78.305 53.025
99094 94.037 -78.310 53.014
99095 94.045 -78.313 53.003
99096 94.052 -78.318 52.990
99097 94.060 -78.322 52.975
99098 94.067 -78.326 52.961
99099 94.075 -78.331 52.947
99100 94.081 -78.338 52.932
99101 94.088 -78.345 52.916
99102 94.093 -78.354 52.900
99103 94.098 -78.363 52.885
99104 94.102 -78.372 52.872
99105 94.107 -78.380 52.861
99106 94.112 -78.389 52.849
99107 94.116 -78.398 52.839
99108 94.121 -78.407 52.830
99109 94.126 -78.416 52.820
99110 94.130 -78.425 52.808
99111 94.135 -78.429 52.796
99112 94.139 -78.432 52.785
99113 94.144 -78.434 52.775
99114 94.146 -78.416 52.760
99115 94.149 -78.395 52.744
99116 94.151 -78.373 52.726
99117 94.164 -78.356 52.704
99118 94.182 -78.342 52.680
99119 94.201 -78.334 52.654
99120 94.222 -78.336 52.627
99121 94.244 -78.338 52.600
99122 94.265 -78.340 52.572
99123 94.285 -78.342 52.544
99124 94.298 -78.354 52.512
99125 94.307 -78.372 52.480
99126 94.315 -78.389 52.448
99127 94.324 -78.406 52.415
99128 94.339 -78.410 52.381
99129 94.360 -78.402 52.347
99130 94.381 -78.394 52.310
99131 94.402 -78.385 52.269
99132 94.429 -78.385 52.227
99133 94.467 -78.399 52.183
99134 94.504 -78.413 52.137
99135 94.535 -78.412 52.093
99136 94.559 -78.396 52.050
99137 94.584 -78.379 52.009
99138 94.608 -78.363 51.967
99139 94.627 -78.351 51.925
99140 94.637 -78.344 51.883
99141 94.648 -78.339 51.839
99142 94.659 -78.333 51.791
99143 94.669 -78.327 51.741
99144 94.680 -78.322 51.691
99145 94.690 -78.316 51.643
99146 94.702 -78.317 51.598
99147 94.714 -78.319 51.554
99148 94.727 -78.316 51.510
99149 94.740 -78.314 51.469
99150 94.753 -78.311 51.428
99151 94.765 -78.310 51.389
99152 94.775 -78.313 51.354
99153 94.786 -78.317 51.321
99154 94.797 -78.321 51.289
99155 94.815 -78.322 51.255
99156 94.835 -78.322 51.215
99157 94.856 -78.322 51.172
99158 94.876 -78.322 51.128
99159 94.897 -78.322 51.079
99160 94.917 -78.322 51.030
99161 94.937 -78.322 50.982
99162 94.958 -78.320 50.935
99163 94.975 -78.304 50.888
99164 94.984 -78.300 50.841
99165 94.992 -78.297 50.796
99166 95.001 -78.293 50.751
99167 95.010 -78.289 50.705
99168 95.019 -78.285 50.660
99169 95.028 -78.281 50.617
99170 95.037 -78.274 50.575
99171 95.059 -78.215 50.533
99172 95.085 -78.166 50.492
99173 95.128 -78.160 50.452
99174 95.172 -78.154 50.413
99175 95.182 -78.150 50.375
99176 95.191 -78.145 50.338
99177 95.200 -78.141 50.300
99178 95.210 -78.137 50.262
99179 95.219 -78.133 50.225
99180 95.233 -78.133 50.189
99181 95.269 -78.156 50.151
99182 95.290 -78.158 50.114
99183 95.301 -78.147 50.077
99184 95.311 -78.136 50.042
99185 95.322 -78.125 50.007
99186 95.331 -78.127 49.971
99187 95.340 -78.130 49.936
99188 95.350 -78.132 49.902
99189 95.359 -78.135 49.869
99190 95.369 -78.138 49.836
99191 95.378 -78.141 49.804
99192 95.387 -78.144 49.772
99193 95.397 -78.147 49.741
99194 95.406 -78.150 49.710
99195 95.415 -78.153 49.679
99196 95.425 -78.156 49.649
99197 95.435 -78.158 49.620
99198 95.444 -78.161 49.592
99199 95.453 -78.164 49.565
99200 95.462 -78.167 49.538
99201 95.476 -78.172 49.512
99202 95.486 -78.177 49.486
99203 95.496 -78.184 49.459
99204 95.505 -78.191 49.430
99205 95.514 -78.198 49.402
99206 95.523 -78.205 49.373
99207 95.533 -78.213 49.342
99208 95.542 -78.219 49.309
99209 95.551 -78.226 49.275
99210 95.560 -78.234 49.240
99211 95.569 -78.241 49.205
99212 95.579 -78.247 49.169
99213 95.588 -78.255 49.130
99214 95.597 -78.262 49.088
99215 95.606 -78.269 49.045
99216 95.615 -78.276 49.003
99217 95.624 -78.283 48.963
99218 95.621 -78.281 48.924
99219 95.617 -78.277 48.887
99220 95.613 -78.274 48.849
99221 95.620 -78.277 48.809
99222 95.634 -78.285 48.771
99223 95.649 -78.294 48.731
99224 95.664 -78.303 48.690
99225 95.678 -78.311 48.649
99226 95.692 -78.315 48.604
99227 95.705 -78.316 48.560
99228 95.719 -78.316 48.517
99229 95.732 -78.317 48.476
99230 95.745 -78.318 48.435
99231 95.760 -78.315 48.394
99232 95.777 -78.311 48.352
99233 95.795 -78.307 48.311
99234 95.812 -78.305 48.269
99235 95.830 -78.313 48.226
99236 95.848 -78.321 48.185
99237 95.865 -78.326 48.144
99238 95.881 -78.328 48.104
99239 95.897 -78.330 48.066
99240 95.913 -78.333 48.029
99241 95.928 -78.337 47.991
99242 95.943 -78.341 47.952
99243 95.958 -78.344 47.915
99244 95.972 -78.347 47.879
99245 95.987 -78.351 47.842
99246 96.002 -78.360 47.806
99247 96.016 -78.374 47.770
99248 96.030 -78.386 47.733
99249 96.045 -78.398 47.695
99250 96.059 -78.410 47.662
99251 96.074 -78.423 47.630
99252 96.088 -78.435 47.597
99253 96.098 -78.440 47.564
99254 96.103 -78.451 47.531
99255 96.102 -78.467 47.498
99256 96.109 -78.476 47.466
99257 96.126 -78.481 47.436
99258 96.143 -78.485 47.407
99259 96.159 -78.489 47.379
99260 96.177 -78.494 47.351
99261 96.206 -78.500 47.321
99262 96.236 -78.505 47.291
99263 96.265 -78.511 47.261
99264 96.294 -78.515 47.233
99265 96.323 -78.518 47.206
99266 96.353 -78.522 47.179
99267 96.382 -78.525 47.151
99268 96.410 -78.529 47.122
99269 96.440 -78.532 47.093
99270 96.469 -78.534 47.064
99271 96.499 -78.536 47.035
99272 96.528 -78.539 47.008
99273 96.557 -78.541 46.982
99274 96.586 -78.543 46.954
99275 96.613 -78.546 46.927
99276 96.640 -78.549 46.899
99277 96.665 -78.552 46.871
99278 96.691 -78.555 46.846
99279 96.703 -78.564 46.820
99280 96.707 -78.574 46.793
99281 96.710 -78.586 46.769
99282 96.713 -78.596 46.745
99283 96.716 -78.608 46.721
99284 96.718 -78.619 46.698
99285 96.725 -78.631 46.675
99286 96.738 -78.646 46.653
99287 96.750 -78.661 46.632
99288 96.763 -78.675 46.609
99289 96.774 -78.687 46.586
99290 96.785 -78.698 46.565
99291 96.802 -78.720 46.544
99292 96.822 -78.742 46.524
99293 96.841 -78.766 46.505
99294 96.861 -78.789 46.485
99295 96.880 -78.812 46.466
99296 96.899 -78.836 46.447
99297 96.919 -78.859 46.428
99298 96.938 -78.882 46.409
99299 96.959 -78.900 46.393
99300 96.984 -78.902 46.375
99301 97.014 -78.913 46.358
99302 97.043 -78.924 46.342
99303 97.073 -78.934 46.326
99304 97.103 -78.945 46.310
99305 97.133 -78.955 46.295
99306 97.162 -78.966 46.279
99307 97.192 -78.977 46.264
99308 97.226 -78.974 46.247
99309 97.261 -78.971 46.230
99310 97.296 -78.968 46.211
99311 97.330 -78.965 46.192
99312 97.365 -78.962 46.173
99313 97.399 -78.959 46.154
99314 97.437 -78.971 46.134
99315 97.455 -78.971 46.115
99316 97.471 -78.970 46.097
99317 97.486 -78.969 46.078
99318 97.494 -78.963 46.059
99319 97.502 -78.957 46.040
99320 97.510 -78.951 46.023
99321 97.518 -78.945 46.004
99322 97.526 -78.940 45.984
99323 97.535 -78.933 45.967
99324 97.543 -78.927 45.950
99325 97.551 -78.921 45.932
99326 97.560 -78.915 45.914
99327 97.574 -78.912 45.898
99328 97.591 -78.908 45.883
99329 97.609 -78.905 45.867
99330 97.627 -78.901 45.852
99331 97.645 -78.898 45.836
99332 97.663 -78.894 45.820
99333 97.677 -78.892 45.803
99334 97.690 -78.890 45.787
99335 97.704 -78.886 45.771
99336 97.717 -78.884 45.755
99337 97.730 -78.882 45.738
99338 97.750 -78.916 45.722
99339 97.753 -78.914 45.708
99340 97.755 -78.911 45.695
99341 97.784 -78.927 45.679
99342 97.792 -78.933 45.664
99343 97.799 -78.939 45.648
99344 97.807 -78.944 45.629
99345 97.814 -78.949 45.612
99346 97.821 -78.954 45.597
99347 97.828 -78.959 45.582
99348 97.836 -78.965 45.571
99349 97.843 -78.969 45.559
99350 97.850 -78.974 45.547
99351 97.858 -78.980 45.535
99352 97.865 -78.985 45.521
99353 97.872 -78.990 45.507
99354 97.879 -78.995 45.494
99355 97.886 -79.001 45.479
99356 97.890 -79.015 45.466
99357 97.903 -79.023 45.453
99358 97.920 -79.030 45.440
99359 97.937 -79.037 45.427
99360 97.953 -79.043 45.413
99361 97.969 -79.050 45.399
99362 97.986 -79.057 45.387
99363 98.003 -79.064 45.375
99364 98.019 -79.065 45.363
99365 98.033 -79.055 45.350
99366 98.049 -79.045 45.336
99367 98.056 -79.047 45.323
99368 98.062 -79.053 45.311
99369 98.094 -79.056 45.301
99370 98.138 -79.074 45.290
99371 98.174 -79.093 45.280
99372 98.200 -79.103 45.271
99373 98.225 -79.112 45.262
99374 98.250 -79.122 45.253
99375 98.276 -79.132 45.244
99376 98.305 -79.137 45.237
99377 98.335 -79.141 45.229
99378 98.366 -79.146 45.220
99379 98.396 -79.150 45.210
99380 98.428 -79.156 45.201
99381 98.462 -79.165 45.192
99382 98.496 -79.174 45.183
99383 98.530 -79.182 45.174
99384 98.564 -79.191 45.165
99385 98.598 -79.200 45.155
99386 98.632 -79.209 45.143
99387 98.666 -79.218 45.132
99388 98.701 -79.228 45.120
99389 98.738 -79.240 45.109
99390 98.774 -79.252 45.098
99391 98.811 -79.264 45.088
99392 98.848 -79.276 45.077
99393 98.877 -79.287 45.067
99394 98.891 -79.298 45.060
99395 98.915 -79.316 45.053
99396 98.939 -79.333 45.046
99397 98.965 -79.347 45.039
99398 98.994 -79.357 45.031
99399 99.024 -79.367 45.023
99400 99.047 -79.379 45.017
99401 99.066 -79.391 45.008
99402 99.086 -79.404 45.000
99403 99.101 -79.416 44.991
99404 99.115 -79.427 44.982
99405 99.124 -79.433 44.971
99406 99.134 -79.438 44.961
99407 99.144 -79.444 44.951
99408 99.154 -79.450 44.939
99409 99.163 -79.456 44.927
99410 99.175 -79.465 44.915
99411 99.191 -79.480 44.903
99412 99.205 -79.494 44.892
99413 99.222 -79.507 44.881
99414 99.243 -79.516 44.871
99415 99.265 -79.524 44.863
99416 99.287 -79.532 44.855
99417 99.309 -79.541 44.847
99418 99.331 -79.550 44.839
99419 99.352 -79.558 44.830
99420 99.374 -79.567 44.821
99421 99.396 -79.575 44.813
99422 99.417 -79.583 44.804
99423 99.435 -79.585 44.797
99424 99.451 -79.579 44.790
99425 99.466 -79.574 44.783
99426 99.482 -79.568 44.776
99427 99.498 -79.563 44.769
99428 99.528 -79.567 44.763
99429 99.558 -79.570 44.757
99430 99.588 -79.573 44.751
99431 99.618 -79.576 44.745
99432 99.648 -79.579 44.740
99433 99.678 -79.582 44.735
99434 99.708 -79.586 44.731
99435 99.738 -79.589 44.726
99436 99.768 -79.592 44.722
99437 99.798 -79.595 44.717
99438 99.837 -79.629 44.712
99439 99.859 -79.657 44.706
99440 99.876 -79.682 44.701
99441 99.894 -79.707 44.696
99442 99.911 -79.733 44.689
99443 99.929 -79.758 44.683
99444 99.930 -79.789 44.677
99445 99.930 -79.833 44.671
99446 100.015 -79.874 44.665
99447 100.039 -79.896 44.659
99448 100.063 -79.920 44.653
99449 100.087 -79.943 44.645
99450 100.116 -79.971 44.638
99451 100.148 -80.002 44.630
99452 100.184 -80.030 44.622
99453 100.229 -80.048 44.614
99454 100.275 -80.067 44.606
99455 100.317 -80.096 44.598
99456 100.355 -80.134 44.588
99457 100.394 -80.171 44.577
99458 100.431 -80.206 44.565
99459 100.466 -80.233 44.553
99460 100.500 -80.261 44.539
99461 100.535 -80.288 44.523
99462 100.569 -80.316 44.507
99463 100.567 -80.360 44.491
99464 100.563 -80.405 44.473
99465 100.560 -80.450 44.455
99466 100.562 -80.493 44.436
99467 100.587 -80.527 44.418
99468 100.612 -80.562 44.400
99469 100.637 -80.597 44.381
99470 100.662 -80.632 44.362
99471 100.690 -80.653 44.342
99472 100.728 -80.673 44.320
99473 100.769 -80.695 44.298
99474 100.813 -80.711 44.278
99475 100.859 -80.724 44.259
99476 100.906 -80.737 44.241
99477 100.952 -80.750 44.223
99478 100.998 -80.763 44.205
99479 101.045 -80.776 44.189
99480 101.092 -80.789 44.172
99481 101.138 -80.801 44.156
99482 101.184 -80.814 44.140
99483 101.231 -80.827 44.125
99484 101.277 -80.840 44.110
99485 101.324 -80.853 44.095
99486 101.370 -80.866 44.080
99487 101.417 -80.879 44.066
99488 101.463 -80.892 44.052
99489 101.510 -80.905 44.037
99747 93.495 -78.548 53.838
99748 93.504 -78.545 53.824
99749 93.514 -78.542 53.809
99750 93.523 -78.540 53.794
99751 93.534 -78.538 53.777
99752 93.543 -78.535 53.760
99753 93.553 -78.532 53.741
99754 93.562 -78.530 53.723
99755 93.567 -78.525 53.703
99756 93.566 -78.516 53.683
99757 93.565 -78.507 53.663
99758 93.563 -78.498 53.642
99759 93.562 -78.489 53.620
99760 93.569 -78.474 53.600
99761 93.583 -78.459 53.580
99762 93.606 -78.444 53.559
99763 93.628 -78.429 53.538
99764 93.650 -78.413 53.518
99765 93.662 -78.405 53.498
99766 93.674 -78.397 53.478
99767 93.686 -78.388 53.459
99777 93.840 -78.315 53.267
99778 93.855 -78.307 53.249
99779 93.870 -78.300 53.229
99780 93.886 -78.293 53.210
99781 93.901 -78.285 53.191
99782 93.917 -78.279 53.173
99783 93.932 -78.271 53.155
99784 93.947 -78.263 53.138
99785 93.962 -78.257 53.124
99786 93.967 -78.262 53.110
99787 93.970 -78.272 53.097
99788 93.973 -78.282 53.087
99789 93.979 -78.288 53.082
99790 93.987 -78.293 53.073
99791 93.994 -78.297 53.062
99792 94.002 -78.301 53.049
99793 94.010 -78.305 53.036
99794 94.017 -78.310 53.024
99795 94.025 -78.313 53.013
99796 94.033 -78.318 53.001
99797 94.040 -78.322 52.987
99798 94.048 -78.326 52.972
99799 94.055 -78.330 52.956
99800 94.063 -78.335 52.942
99801 94.070 -78.340 52.926
99802 94.076 -78.347 52.908
99803 94.083 -78.354 52.891
99804 94.089 -78.362 52.875
99805 94.096 -78.369 52.861
99806 94.103 -78.376 52.850
99807 94.108 -78.385 52.839
99808 94.113 -78.393 52.828
99809 94.117 -78.402 52.818
99810 94.122 -78.411 52.807
99811 94.127 -78.420 52.794
99812 94.131 -78.429 52.782
99813 94.136 -78.431 52.770
99814 94.140 -78.425 52.757
99815 94.142 -78.403 52.743
99816 94.144 -78.382 52.726
99817 94.147 -78.360 52.707
99818 94.154 -78.341 52.685
99819 94.173 -78.327 52.660
99820 94.192 -78.322 52.635
99821 94.213 -78.323 52.608
99822 94.235 -78.326 52.580
99823 94.256 -78.328 52.552
99824 94.276 -78.329 52.524
99825 94.294 -78.335 52.492
99826 94.303 -78.353 52.461
99827 94.312 -78.370 52.429
99828 94.320 -78.387 52.395
99829 94.339 -78.383 52.361
99830 94.360 -78.375 52.325
99831 94.381 -78.366 52.288
99832 94.402 -78.358 52.246
99833 94.424 -78.351 52.202
99834 94.462 -78.366 52.157
99835 94.499 -78.379 52.109
99836 94.537 -78.394 52.062
99837 94.564 -78.385 52.016
99838 94.589 -78.368 51.971
99839 94.613 -78.351 51.929
99840 94.632 -78.339 51.887
99841 94.642 -78.333 51.845
99842 94.653 -78.327 51.801
99843 94.664 -78.321 51.753
99844 94.674 -78.316 51.705
99845 94.685 -78.310 51.657
99846 94.695 -78.304 51.611
99847 94.706 -78.300 51.566
99848 94.718 -78.302 51.523
99849 94.732 -78.299 51.481
99850 94.744 -78.296 51.441
99851 94.757 -78.294 51.401
99852 94.770 -78.291 51.364
99853 94.780 -78.294 51.331
99854 94.791 -78.299 51.299
99855 94.801 -78.304 51.267
99856 94.812 -78.308 51.231
99857 94.825 -78.312 51.190
99858 94.845 -78.312 51.147
99859 94.865 -78.312 51.103
99860 94.886 -78.312 51.057
99861 94.906 -78.312 51.010
99862 94.927 -78.313 50.964
99863 94.947 -78.305 50.918
99864 94.968 -78.284 50.873
99865 94.988 -78.264 50.828
99866 94.996 -78.260 50.783
99867 95.005 -78.257 50.739
99868 95.014 -78.253 50.695
99869 95.023 -78.249 50.651
99870 95.031 -78.244 50.609
99871 95.047 -78.213 50.568
99872 95.070 -78.157 50.527
99873 95.114 -78.151 50.488
99874 95.157 -78.146 50.448
99875 95.201 -78.140 50.410
99876 95.213 -78.135 50.372
99877 95.222 -78.131 50.335
99878 95.232 -78.127 50.297
99879 95.241 -78.122 50.260
99880 95.250 -78.118 50.224
99881 95.259 -78.114 50.187
99882 95.269 -78.109 50.151
99883 95.278 -78.106 50.115
99884 95.305 -78.119 50.079
99885 95.317 -78.110 50.044
99886 95.328 -78.102 50.010
99887 95.337 -78.105 49.975
99888 95.346 -78.108 49.942
99889 95.356 -78.111 49.909
99890 95.365 -78.113 49.876
99891 95.375 -78.116 49.844
99892 95.384 -78.119 49.812
99893 95.393 -78.122 49.781
99894 95.403 -78.125 49.751
99895 95.412 -78.128 49.722
99896 95.421 -78.131 49.692
99897 95.431 -78.134 49.663
99898 95.442 -78.138 49.634
99899 95.452 -78.142 49.607
99900 95.462 -78.147 49.581
99901 95.471 -78.150 49.556
99902 95.477 -78.154 49.531
99903 95.483 -78.155 49.506
99904 95.491 -78.160 49.480
99905 95.500 -78.167 49.453
99906 95.510 -78.175 49.424
99907 95.519 -78.181 49.394
99908 95.528 -78.188 49.362
99909 95.537 -78.196 49.328
99910 95.546 -78.203 49.293
99911 95.555 -78.210 49.257
99912 95.564 -78.217 49.220
99913 95.573 -78.224 49.181
99914 95.583 -78.232 49.141
99915 95.592 -78.238 49.099
99916 95.601 -78.245 49.057
99917 95.610 -78.253 49.015
99918 95.619 -78.260 48.972
99919 95.620 -78.260 48.931
99920 95.616 -78.257 48.891
99921 95.612 -78.253 48.852
99922 95.624 -78.260 48.813
99923 95.639 -78.269 48.775
99924 95.654 -78.277 48.736
99925 95.669 -78.286 48.696
99926 95.686 -78.298 48.655
99927 95.703 -78.304 48.613
99928 95.706 -78.302 48.570
99929 95.719 -78.303 48.526
99930 95.732 -78.304 48.484
99931 95.745 -78.304 48.442
99932 95.762 -78.285 48.400
99933 95.779 -78.282 48.359
99934 95.797 -78.278 48.317
99935 95.815 -78.285 48.275
99936 95.833 -78.294 48.234
99937 95.851 -78.302 48.195
99938 95.868 -78.308 48.156
99939 95.884 -78.310 48.117
99940 95.900 -78.313 48.079
99941 95.917 -78.316 48.042
99942 95.934 -78.320 48.005
99943 95.948 -78.324 47.966
99944 95.963 -78.327 47.929
99945 95.979 -78.330 47.893
99946 95.995 -78.332 47.856
99947 96.010 -78.344 47.820
99948 96.022 -78.362 47.783
99949 96.036 -78.376 47.746
99950 96.050 -78.388 47.709
99951 96.065 -78.401 47.676
99952 96.079 -78.413 47.643
99953 96.094 -78.425 47.611
99954 96.104 -78.430 47.578
99955 96.112 -78.438 47.545
99956 96.111 -78.454 47.512
99957 96.121 -78.462 47.481
99958 96.138 -78.466 47.451
99959 96.155 -78.470 47.422
99960 96.171 -78.475 47.394
99961 96.187 -78.479 47.367
99962 96.207 -78.484 47.337
99963 96.237 -78.489 47.307
99964 96.266 -78.493 47.278
99965 96.295 -78.497 47.249
99966 96.324 -78.500 47.221
99967 96.354 -78.504 47.194
99968 96.383 -78.507 47.167
99969 96.411 -78.510 47.140
99970 96.441 -78.513 47.111
99971 96.470 -78.516 47.082
99972 96.498 -78.518 47.054
99973 96.523 -78.521 47.027
99974 96.549 -78.524 47.000
99975 96.576 -78.527 46.972
99976 96.601 -78.530 46.944
99977 96.627 -78.533 46.916
99978 96.654 -78.536 46.889
99979 96.680 -78.540 46.863
99980 96.683 -78.551 46.838
99981 96.686 -78.562 46.813
99982 96.689 -78.573 46.789
99983 96.692 -78.584 46.763
99984 96.700 -78.598 46.738
99985 96.708 -78.608 46.714
99986 96.719 -78.616 46.691
99987 96.726 -78.633 46.669
99988 96.739 -78.648 46.648
99989 96.751 -78.662 46.625
99990 96.762 -78.673 46.603
99991 96.781 -78.696 46.582
99992 96.800 -78.720 46.562
99993 96.819 -78.742 46.542
99994 96.839 -78.766 46.522
99995 96.858 -78.789 46.503
99996 96.878 -78.812 46.484
99997 96.897 -78.836 46.464
99998 96.916 -78.859 46.445
99999 96.936 -78.882 46.426
100000 96.955 -78.905 46.409
100001 96.979 -78.911 46.391
100002 97.006 -78.917 46.374
100003 97.036 -78.927 46.357
100004 97.066 -78.938 46.339
100005 97.095 -78.949 46.323
100006 97.125 -78.959 46.307
100007 97.155 -78.969 46.291
100008 97.186 -78.975 46.275
100009 97.221 -78.972 46.258
100010 97.255 -78.968 46.241
100011 97.290 -78.965 46.223
100012 97.325 -78.962 46.204
100013 97.360 -78.959 46.186
100014 97.394 -78.955 46.168
100015 97.430 -78.955 46.148
100016 97.461 -78.965 46.128
100017 97.476 -78.964 46.109
100018 97.492 -78.963 46.090
100019 97.502 -78.958 46.070
100020 97.510 -78.952 46.051
100021 97.518 -78.946 46.034
100022 97.526 -78.940 46.015
100023 97.534 -78.933 45.996
100024 97.542 -78.928 45.979
100025 97.550 -78.922 45.961
100026 97.559 -78.916 45.942
100027 97.574 -78.912 45.923
100028 97.592 -78.908 45.906
100029 97.609 -78.905 45.890
100030 97.628 -78.902 45.873
100031 97.645 -78.899 45.857
100032 97.663 -78.895 45.842
100033 97.681 -78.892 45.826
100034 97.696 -78.889 45.809
100035 97.709 -78.886 45.794
100036 97.722 -78.884 45.777
100037 97.735 -78.882 45.761
100038 97.749 -78.880 45.746
100039 97.762 -78.877 45.731
100040 97.774 -78.893 45.717
100041 97.788 -78.899 45.704
100042 97.797 -78.905 45.689
100043 97.804 -78.910 45.673
100044 97.811 -78.915 45.658
100045 97.818 -78.920 45.640
100046 97.826 -78.925 45.624
100047 97.833 -78.930 45.610
100048 97.841 -78.936 45.597
100049 97.848 -78.940 45.586
100050 97.855 -78.946 45.575
100051 97.862 -78.951 45.563
100052 97.869 -78.956 45.550
100053 97.877 -78.961 45.536
100054 97.884 -78.966 45.523
100055 97.891 -78.971 45.509
100056 97.898 -78.977 45.494
100057 97.903 -78.987 45.480
100058 97.919 -78.994 45.466
100059 97.936 -79.001 45.452
100060 97.952 -79.008 45.438
100061 97.968 -79.015 45.423
100062 97.985 -79.021 45.408
100063 98.002 -79.028 45.394
100064 98.018 -79.035 45.381
100065 98.034 -79.042 45.368
100066 98.051 -79.043 45.356
100067 98.066 -79.032 45.342
100068 98.073 -79.034 45.329
100069 98.079 -79.040 45.317
100070 98.093 -79.046 45.306
100071 98.134 -79.071 45.295
100072 98.173 -79.093 45.285
100073 98.199 -79.103 45.274
100074 98.225 -79.110 45.265
100075 98.256 -79.114 45.256
100076 98.286 -79.118 45.246
100077 98.316 -79.122 45.238
100078 98.347 -79.127 45.229
100079 98.377 -79.131 45.220
100080 98.407 -79.135 45.210
100081 98.440 -79.143 45.201
100082 98.474 -79.152 45.192
100083 98.508 -79.160 45.183
100084 98.543 -79.169 45.176
100085 98.576 -79.178 45.167
100086 98.610 -79.187 45.157
100087 98.645 -79.196 45.147
100088 98.678 -79.204 45.135
100089 98.713 -79.213 45.123
100090 98.746 -79.222 45.112
100091 98.781 -79.231 45.100
100092 98.815 -79.240 45.090
100093 98.849 -79.249 45.080
100094 98.886 -79.261 45.069
100095 98.908 -79.275 45.060
100096 98.934 -79.287 45.052
100097 98.954 -79.299 45.044
100098 98.975 -79.311 45.037
100099 98.995 -79.323 45.030
100100 99.015 -79.335 45.022
100101 99.034 -79.347 45.015
100102 99.055 -79.360 45.007
100103 99.071 -79.375 44.998
100104 99.083 -79.380 44.990
100105 99.100 -79.397 44.981
100106 99.118 -79.414 44.970
100107 99.130 -79.422 44.959
100108 99.139 -79.428 44.949
100109 99.149 -79.433 44.936
100110 99.159 -79.439 44.924
100111 99.171 -79.448 44.912
100112 99.186 -79.463 44.900
100113 99.201 -79.477 44.890
100114 99.223 -79.485 44.880
100115 99.244 -79.494 44.870
100116 99.266 -79.503 44.861
100117 99.288 -79.511 44.852
100118 99.309 -79.520 44.845
100119 99.331 -79.528 44.836
100120 99.353 -79.536 44.828
100121 99.375 -79.545 44.820
100122 99.397 -79.553 44.812
100123 99.418 -79.562 44.804
100124 99.440 -79.570 44.797
100125 99.457 -79.568 44.791
100126 99.472 -79.563 44.784
100127 99.488 -79.557 44.778
100128 99.512 -79.557 44.772
100129 99.542 -79.560 44.767
100130 99.572 -79.564 44.761
100131 99.602 -79.567 44.756
100132 99.632 -79.570 44.750
100133 99.662 -79.573 44.745
100134 99.692 -79.577 44.741
100135 99.722 -79.580 44.737
100136 99.752 -79.583 44.732
100137 99.782 -79.586 44.728
100138 99.812 -79.589 44.724
100139 99.842 -79.593 44.719
100140 99.880 -79.626 44.713
100141 99.907 -79.655 44.708
100142 99.925 -79.680 44.703
100143 99.942 -79.705 44.696
100144 99.958 -79.731 44.690
100145 99.932 -79.773 44.684
100146 99.981 -79.834 44.678
100147 100.057 -79.861 44.672
100148 100.081 -79.883 44.666
100149 100.105 -79.907 44.659
100150 100.129 -79.930 44.652
100151 100.153 -79.953 44.646
100152 100.180 -79.978 44.640
100153 100.214 -80.007 44.633
100154 100.260 -80.026 44.624
100155 100.299 -80.062 44.616
100156 100.335 -80.095 44.608
100157 100.370 -80.123 44.598
100158 100.405 -80.150 44.586
100159 100.439 -80.178 44.573
100160 100.474 -80.206 44.560
100161 100.508 -80.233 44.547
100162 100.543 -80.261 44.532
100163 100.577 -80.288 44.516
100164 100.579 -80.331 44.500
100165 100.575 -80.375 44.483
100166 100.572 -80.420 44.465
100167 100.579 -80.461 44.447
100168 100.605 -80.496 44.428
100169 100.630 -80.531 44.410
100170 100.654 -80.566 44.392
100171 100.680 -80.600 44.374
100172 100.705 -80.635 44.354
100173 100.730 -80.670 44.333
100174 100.770 -80.689 44.311
100175 100.817 -80.702 44.291
100176 100.863 -80.715 44.271
100177 100.910 -80.728 44.251
100178 100.956 -80.741 44.233
100179 101.002 -80.754 44.215
100180 101.049 -80.767 44.198
100181 101.096 -80.779 44.181
100182 101.142 -80.792 44.164
100183 101.189 -80.805 44.148
100184 101.235 -80.818 44.133
100185 101.281 -80.831 44.117
100186 101.328 -80.844 44.101
100187 101.375 -80.857 44.086
100188 101.421 -80.870 44.071
100189 101.467 -80.883 44.057
100190 101.514 -80.895 44.042
100191 101.548 -80.911 44.027
100448 93.495 -78.537 53.841
100449 93.504 -78.534 53.828
100450 93.514 -78.532 53.813
100451 93.524 -78.529 53.797
100452 93.534 -78.527 53.779
100453 93.543 -78.524 53.761
100454 93.553 -78.522 53.742
100455 93.563 -78.520 53.723
100456 93.564 -78.512 53.703
100457 93.563 -78.503 53.683
100458 93.561 -78.494 53.663
100459 93.560 -78.484 53.640
100460 93.561 -78.474 53.618
100461 93.571 -78.459 53.596
100462 93.588 -78.443 53.576
100463 93.611 -78.429 53.554
100464 93.633 -78.413 53.533
100465 93.652 -78.401 53.514
100466 93.663 -78.392 53.492
100467 93.675 -78.384 53.472
100479 93.857 -78.295 53.242
100480 93.872 -78.288 53.222
100481 93.888 -78.280 53.202
100482 93.903 -78.272 53.184
100483 93.918 -78.266 53.165
100484 93.934 -78.258 53.146
100485 93.949 -78.250 53.129
100486 93.963 -78.246 53.114
100487 93.966 -78.248 53.100
100488 93.969 -78.259 53.088
100489 93.973 -78.269 53.078
100490 93.976 -78.279 53.074
100491 93.980 -78.290 53.067
100492 93.983 -78.300 53.058
100493 93.990 -78.305 53.047
100494 93.997 -78.310 53.033
100495 94.005 -78.313 53.021
100496 94.012 -78.318 53.009
100497 94.020 -78.322 52.997
100498 94.028 -78.326 52.982
100499 94.035 -78.330 52.965
100500 94.043 -78.335 52.948
100501 94.050 -78.338 52.932
100502 94.058 -78.343 52.915
100503 94.065 -78.349 52.897
100504 94.071 -78.357 52.878
100505 94.078 -78.363 52.861
100506 94.084 -78.371 52.847
100507 94.091 -78.378 52.837
100508 94.098 -78.385 52.825
100509 94.104 -78.392 52.813
100510 94.111 -78.400 52.801
100511 94.117 -78.407 52.789
100512 94.123 -78.415 52.777
100513 94.128 -78.423 52.764
100514 94.132 -78.430 52.751
100515 94.135 -78.412 52.737
100516 94.137 -78.391 52.721
100517 94.140 -78.369 52.703
100518 94.142 -78.347 52.683
100519 94.145 -78.326 52.660
100520 94.163 -78.312 52.636
100521 94.183 -78.309 52.610
100522 94.204 -78.311 52.582
100523 94.225 -78.313 52.555
100524 94.247 -78.315 52.527
100525 94.267 -78.317 52.498
100526 94.288 -78.319 52.468
100527 94.299 -78.334 52.436
100528 94.308 -78.351 52.404
100529 94.318 -78.365 52.369
100530 94.339 -78.357 52.333
100531 94.360 -78.348 52.295
100532 94.381 -78.340 52.255
100533 94.402 -78.332 52.212
100534 94.423 -78.323 52.168
100535 94.457 -78.332 52.122
100536 94.494 -78.346 52.074
100537 94.532 -78.360 52.026
100538 94.569 -78.373 51.979
100539 94.593 -78.357 51.935
100540 94.618 -78.340 51.892
100541 94.637 -78.327 51.849
100542 94.648 -78.321 51.805
100543 94.659 -78.316 51.761
100544 94.669 -78.310 51.715
100545 94.680 -78.304 51.668
100546 94.690 -78.297 51.623
100547 94.700 -78.292 51.579
100548 94.711 -78.286 51.536
100549 94.723 -78.285 51.496
100550 94.736 -78.282 51.454
100551 94.749 -78.279 51.415
100552 94.760 -78.276 51.377
100553 94.772 -78.273 51.340
100554 94.784 -78.277 51.306
100555 94.795 -78.282 51.273
100556 94.806 -78.286 51.240
100557 94.816 -78.291 51.202
100558 94.826 -78.296 51.162
100559 94.836 -78.300 51.121
100560 94.854 -78.302 51.079
100561 94.875 -78.302 51.035
100562 94.895 -78.302 50.991
100563 94.916 -78.302 50.946
100564 94.936 -78.289 50.902
100565 94.957 -78.268 50.859
100566 94.978 -78.247 50.815
100567 94.999 -78.225 50.772
100568 95.009 -78.220 50.729
100569 95.018 -78.216 50.686
100570 95.027 -78.213 50.643
100571 95.035 -78.209 50.602
100572 95.056 -78.151 50.561
100573 95.099 -78.143 50.521
100574 95.142 -78.137 50.483
100575 95.186 -78.131 50.444
100576 95.230 -78.125 50.406
100577 95.244 -78.121 50.368
100578 95.253 -78.116 50.332
100579 95.263 -78.112 50.295
100580 95.272 -78.108 50.258
100581 95.281 -78.103 50.222
100582 95.290 -78.100 50.187
100583 95.300 -78.095 50.151
100584 95.309 -78.091 50.117
100585 95.318 -78.087 50.082
100586 95.327 -78.082 50.048
100587 95.335 -78.082 50.014
100588 95.346 -78.087 49.981
100589 95.356 -78.091 49.949
100590 95.366 -78.094 49.917
100591 95.377 -78.099 49.885
100592 95.387 -78.103 49.853
100593 95.397 -78.106 49.822
100594 95.407 -78.111 49.792
100595 95.417 -78.115 49.762
100596 95.427 -78.119 49.735
100597 95.438 -78.123 49.706
100598 95.448 -78.127 49.678
100599 95.456 -78.131 49.650
100600 95.462 -78.135 49.624
100601 95.468 -78.139 49.599
100602 95.473 -78.143 49.574
100603 95.478 -78.144 49.550
100604 95.482 -78.140 49.525
100605 95.490 -78.144 49.500
100606 95.498 -78.149 49.472
100607 95.506 -78.154 49.443
100608 95.514 -78.159 49.413
100609 95.523 -78.165 49.380
100610 95.532 -78.172 49.345
100611 95.541 -78.179 49.309
100612 95.550 -78.186 49.272
100613 95.559 -78.194 49.234
100614 95.568 -78.200 49.193
100615 95.578 -78.207 49.151
100616 95.587 -78.215 49.109
100617 95.596 -78.222 49.067
100618 95.605 -78.228 49.024
100619 95.615 -78.236 48.981
100620 95.619 -78.239 48.938
100621 95.615 -78.236 48.897
100622 95.615 -78.235 48.856
100623 95.629 -78.244 48.817
100624 95.644 -78.252 48.779
100625 95.659 -78.260 48.741
100626 95.678 -78.274 48.702
100627 95.697 -78.288 48.663
100628 95.714 -78.294 48.621
100629 95.729 -78.291 48.579
100630 95.725 -78.290 48.535
100631 95.733 -78.290 48.492
100632 95.747 -78.276 48.450
100633 95.764 -78.255 48.408
100634 95.781 -78.251 48.366
100635 95.799 -78.258 48.326
100636 95.817 -78.266 48.285
100637 95.835 -78.275 48.244
100638 95.853 -78.282 48.206
100639 95.872 -78.291 48.168
100640 95.888 -78.293 48.130
100641 95.904 -78.296 48.091
100642 95.920 -78.300 48.054
100643 95.937 -78.304 48.017
100644 95.956 -78.307 47.979
100645 95.972 -78.309 47.942
100646 95.988 -78.312 47.905
100647 96.004 -78.314 47.869
100648 96.017 -78.329 47.832
100649 96.030 -78.346 47.795
100650 96.043 -78.363 47.759
100651 96.056 -78.379 47.722
100652 96.071 -78.391 47.689
100653 96.085 -78.403 47.656
100654 96.099 -78.415 47.624
100655 96.110 -78.420 47.592
100656 96.120 -78.426 47.559
100657 96.119 -78.441 47.527
100658 96.133 -78.447 47.497
100659 96.150 -78.451 47.466
100660 96.167 -78.456 47.437
100661 96.183 -78.460 47.409
100662 96.199 -78.464 47.381
100663 96.216 -78.469 47.353
100664 96.238 -78.474 47.325
100665 96.267 -78.478 47.296
100666 96.296 -78.482 47.266
100667 96.325 -78.485 47.238
100668 96.355 -78.489 47.211
100669 96.383 -78.493 47.184
100670 96.414 -78.498 47.157
100671 96.449 -78.501 47.128
100672 96.464 -78.501 47.099
100673 96.485 -78.502 47.072
100674 96.512 -78.506 47.045
100675 96.538 -78.509 47.017
100676 96.564 -78.512 46.989
100677 96.590 -78.515 46.961
100678 96.616 -78.518 46.934
100679 96.642 -78.521 46.906
100680 96.660 -78.527 46.880
100681 96.663 -78.538 46.856
100682 96.666 -78.549 46.833
100683 96.677 -78.565 46.808
100684 96.689 -78.579 46.781
100685 96.695 -78.589 46.755
100686 96.702 -78.598 46.731
100687 96.716 -78.604 46.708
100688 96.725 -78.616 46.686
100689 96.731 -78.633 46.664
100690 96.740 -78.650 46.642
100691 96.759 -78.673 46.620
100692 96.779 -78.696 46.600
100693 96.798 -78.720 46.580
100694 96.817 -78.742 46.561
100695 96.836 -78.766 46.541
100696 96.856 -78.789 46.521
100697 96.875 -78.812 46.503
100698 96.895 -78.836 46.483
100699 96.914 -78.859 46.463
100700 96.934 -78.882 46.444
100701 96.953 -78.905 46.426
100702 96.974 -78.919 46.409
100703 96.999 -78.921 46.391
100704 97.028 -78.930 46.372
100705 97.058 -78.941 46.354
100706 97.088 -78.952 46.336
100707 97.117 -78.962 46.320
100708 97.147 -78.973 46.303
100709 97.181 -78.973 46.287
100710 97.216 -78.969 46.270
100711 97.250 -78.966 46.252
100712 97.285 -78.962 46.235
100713 97.319 -78.959 46.218
100714 97.354 -78.956 46.200
100715 97.389 -78.952 46.182
100716 97.423 -78.949 46.163
100717 97.459 -78.953 46.143
100718 97.483 -78.957 46.122
100719 97.498 -78.956 46.102
100720 97.509 -78.952 46.081
100721 97.517 -78.946 46.061
100722 97.525 -78.940 46.043
100723 97.534 -78.934 46.026
100724 97.542 -78.928 46.009
100725 97.550 -78.922 45.989
100726 97.558 -78.917 45.971
100727 97.574 -78.913 45.952
100728 97.592 -78.909 45.933
100729 97.610 -78.906 45.916
100730 97.628 -78.902 45.899
100731 97.646 -78.899 45.882
100732 97.663 -78.896 45.865
100733 97.681 -78.893 45.849
100734 97.699 -78.889 45.832
100735 97.715 -78.886 45.817
100736 97.728 -78.883 45.801
100737 97.741 -78.881 45.786
100738 97.754 -78.879 45.770
100739 97.768 -78.877 45.755
100740 97.781 -78.874 45.740
100741 97.794 -78.872 45.727
100742 97.801 -78.876 45.713
100743 97.809 -78.881 45.698
100744 97.816 -78.886 45.683
100745 97.823 -78.891 45.667
100746 97.830 -78.896 45.651
100747 97.838 -78.902 45.636
100748 97.845 -78.906 45.623
100749 97.852 -78.912 45.611
100750 97.860 -78.917 45.603
100751 97.867 -78.922 45.594
100752 97.874 -78.927 45.584
100753 97.881 -78.932 45.570
100754 97.889 -78.937 45.556
100755 97.896 -78.943 45.540
100756 97.903 -78.947 45.525
100757 97.911 -78.952 45.512
100758 97.918 -78.959 45.498
100759 97.935 -78.966 45.482
100760 97.951 -78.973 45.466
100761 97.968 -78.980 45.449
100762 97.984 -78.987 45.432
100763 98.001 -78.993 45.417
100764 98.017 -79.000 45.402
100765 98.034 -79.007 45.388
100766 98.050 -79.014 45.374
100767 98.067 -79.021 45.362
100768 98.082 -79.020 45.348
100769 98.091 -79.021 45.335
100770 98.108 -79.035 45.322
100771 98.136 -79.056 45.311
100772 98.159 -79.075 45.300
100773 98.179 -79.090 45.290
100774 98.207 -79.091 45.279
100775 98.237 -79.096 45.269
100776 98.267 -79.100 45.259
100777 98.297 -79.104 45.249
100778 98.327 -79.109 45.240
100779 98.358 -79.112 45.231
100780 98.388 -79.117 45.221
100781 98.419 -79.121 45.211
100782 98.452 -79.130 45.201
100783 98.487 -79.138 45.192
100784 98.520 -79.147 45.184
100785 98.555 -79.156 45.176
100786 98.589 -79.165 45.167
100787 98.623 -79.174 45.158
100788 98.657 -79.182 45.150
100789 98.691 -79.191 45.137
100790 98.725 -79.200 45.126
100791 98.759 -79.209 45.116
100792 98.793 -79.218 45.104
100793 98.827 -79.226 45.094
100794 98.861 -79.235 45.083
100795 98.888 -79.242 45.071
100796 98.906 -79.247 45.060
100797 98.923 -79.256 45.050
100798 98.943 -79.268 45.041
100799 98.963 -79.280 45.035
100800 98.983 -79.292 45.028
100801 99.003 -79.304 45.021
100802 99.023 -79.316 45.014
100803 99.043 -79.328 45.004
100804 99.052 -79.337 44.995
100805 99.065 -79.344 44.985
100806 99.083 -79.361 44.976
100807 99.100 -79.379 44.965
100808 99.118 -79.396 44.954
100809 99.135 -79.411 44.944
100810 99.144 -79.417 44.932
100811 99.154 -79.422 44.921
100812 99.166 -79.432 44.910
100813 99.181 -79.447 44.898
100814 99.202 -79.455 44.887
100815 99.224 -79.463 44.877
100816 99.245 -79.473 44.867
100817 99.267 -79.481 44.858
100818 99.289 -79.489 44.850
100819 99.310 -79.498 44.842
100820 99.332 -79.506 44.834
100821 99.354 -79.514 44.827
100822 99.375 -79.523 44.819
100823 99.398 -79.532 44.812
100824 99.419 -79.540 44.805
100825 99.441 -79.548 44.799
100826 99.463 -79.557 44.792
100827 99.478 -79.552 44.785
100828 99.497 -79.548 44.780
100829 99.526 -79.551 44.775
100830 99.557 -79.554 44.771
100831 99.586 -79.558 44.765
100832 99.617 -79.561 44.760
100833 99.646 -79.564 44.756
100834 99.677 -79.567 44.751
100835 99.706 -79.571 44.746
100836 99.736 -79.574 44.743
100837 99.766 -79.577 44.739
100838 99.796 -79.580 44.734
100839 99.826 -79.584 44.729
100840 99.856 -79.587 44.725
100841 99.886 -79.590 44.721
100842 99.924 -79.621 44.715
100843 99.956 -79.653 44.709
100844 99.974 -79.678 44.702
100845 99.967 -79.734 44.696
100846 99.971 -79.829 44.690
100847 99.989 -79.848 44.684
100848 100.025 -79.867 44.678
100849 100.062 -79.886 44.673
100850 100.099 -79.905 44.666
100851 100.135 -79.924 44.659
100852 100.171 -79.944 44.655
100853 100.208 -79.963 44.649
100854 100.240 -79.986 44.644
100855 100.274 -80.012 44.635
100856 100.309 -80.040 44.626
100857 100.344 -80.068 44.617
100858 100.379 -80.095 44.608
100859 100.413 -80.123 44.594
100860 100.447 -80.150 44.581
100861 100.482 -80.178 44.567
100862 100.516 -80.206 44.554
100863 100.551 -80.233 44.538
100864 100.586 -80.261 44.523
100865 100.592 -80.301 44.507
100866 100.587 -80.346 44.491
100867 100.584 -80.391 44.474
100868 100.597 -80.430 44.455
100869 100.622 -80.465 44.437
100870 100.647 -80.500 44.419
100871 100.672 -80.534 44.402
100872 100.697 -80.569 44.385
100873 100.722 -80.604 44.364
100874 100.749 -80.639 44.344
100875 100.777 -80.676 44.323
100876 100.821 -80.692 44.305
100877 100.867 -80.705 44.285
100878 100.914 -80.718 44.265
100879 100.960 -80.731 44.245
100880 101.007 -80.744 44.225
100881 101.053 -80.757 44.206
100882 101.100 -80.770 44.189
100883 101.146 -80.782 44.172
100884 101.193 -80.795 44.156
100885 101.239 -80.808 44.140
100886 101.285 -80.821 44.124
100887 101.332 -80.834 44.108
100888 101.379 -80.847 44.092
100889 101.425 -80.860 44.077
100890 101.471 -80.873 44.062
100891 101.504 -80.889 44.047
100892 101.537 -80.905 44.032
101149 93.495 -78.526 53.844
101150 93.505 -78.523 53.830
101151 93.514 -78.521 53.815
101152 93.524 -78.518 53.799
101153 93.534 -78.516 53.780
101154 93.544 -78.514 53.762
101155 93.553 -78.511 53.743
101156 93.562 -78.508 53.723
101157 93.561 -78.499 53.703
101158 93.559 -78.489 53.683
101159 93.558 -78.480 53.663
101160 93.557 -78.471 53.639
101161 93.562 -78.458 53.615
101162 93.572 -78.443 53.593
101163 93.594 -78.428 53.571
101164 93.616 -78.413 53.549
101165 93.639 -78.398 53.528
101166 93.653 -78.388 53.508
101167 93.665 -78.379 53.486
101168 93.677 -78.371 53.465
101180 93.859 -78.282 53.234
101181 93.874 -78.275 53.214
101182 93.890 -78.267 53.194
101183 93.905 -78.260 53.175
101184 93.920 -78.253 53.157
101185 93.936 -78.245 53.138
101186 93.951 -78.239 53.120
101187 93.963 -78.235 53.104
101188 93.965 -78.235 53.090
101189 93.969 -78.245 53.077
101190 93.972 -78.256 53.067
101191 93.976 -78.266 53.063
101192 93.979 -78.276 53.056
101193 93.982 -78.287 53.047
101194 93.986 -78.297 53.036
101195 93.989 -78.307 53.022
101196 93.993 -78.317 53.010
101197 94.000 -78.322 52.997
101198 94.007 -78.326 52.983
101199 94.015 -78.330 52.967
101200 94.023 -78.335 52.950
101201 94.030 -78.338 52.932
101202 94.038 -78.343 52.915
101203 94.045 -78.347 52.897
101204 94.053 -78.351 52.878
101205 94.060 -78.358 52.860
101206 94.066 -78.366 52.844
101207 94.073 -78.373 52.830
101208 94.079 -78.380 52.819
101209 94.086 -78.387 52.805
101210 94.092 -78.395 52.792
101211 94.099 -78.401 52.779
101212 94.106 -78.409 52.767
101213 94.112 -78.416 52.755
101214 94.119 -78.423 52.743
101215 94.130 -78.420 52.728
101216 94.130 -78.400 52.712
101217 94.133 -78.378 52.694
101218 94.135 -78.357 52.675
101219 94.137 -78.335 52.655
101220 94.140 -78.313 52.632
101221 94.154 -78.297 52.607
101222 94.174 -78.296 52.581
101223 94.195 -78.298 52.553
101224 94.216 -78.300 52.525
101225 94.238 -78.302 52.498
101226 94.258 -78.304 52.469
101227 94.279 -78.307 52.439
101228 94.295 -78.315 52.408
101229 94.303 -78.332 52.374
101230 94.318 -78.338 52.339
101231 94.339 -78.329 52.300
101232 94.360 -78.321 52.259
101233 94.381 -78.313 52.217
101234 94.402 -78.304 52.173
101235 94.423 -78.296 52.128
101236 94.452 -78.298 52.082
101237 94.489 -78.312 52.035
101238 94.527 -78.326 51.990
101239 94.564 -78.340 51.946
101240 94.598 -78.345 51.903
101241 94.622 -78.329 51.860
101242 94.642 -78.315 51.815
101243 94.653 -78.310 51.770
101244 94.664 -78.304 51.725
101245 94.674 -78.297 51.679
101246 94.685 -78.292 51.635
101247 94.695 -78.286 51.591
101248 94.706 -78.280 51.549
101249 94.714 -78.272 51.508
101250 94.723 -78.267 51.469
101251 94.735 -78.264 51.428
101252 94.746 -78.262 51.389
101253 94.758 -78.259 51.352
101254 94.770 -78.257 51.315
101255 94.781 -78.260 51.280
101256 94.793 -78.264 51.245
101257 94.806 -78.269 51.211
101258 94.818 -78.274 51.174
101259 94.830 -78.279 51.135
101260 94.840 -78.284 51.096
101261 94.850 -78.288 51.056
101262 94.864 -78.291 51.015
101263 94.885 -78.292 50.973
101264 94.905 -78.292 50.930
101265 94.926 -78.273 50.888
101266 94.947 -78.253 50.846
101267 94.968 -78.232 50.804
101268 94.988 -78.210 50.761
101269 95.010 -78.189 50.719
101270 95.022 -78.180 50.678
101271 95.030 -78.176 50.637
101272 95.044 -78.149 50.596
101273 95.085 -78.134 50.556
101274 95.128 -78.128 50.518
101275 95.171 -78.122 50.480
101276 95.215 -78.116 50.442
101277 95.258 -78.111 50.404
101278 95.275 -78.106 50.366
101279 95.284 -78.102 50.330
101280 95.294 -78.097 50.294
101281 95.303 -78.094 50.258
101282 95.312 -78.089 50.222
101283 95.321 -78.084 50.187
101284 95.328 -78.083 50.153
101285 95.334 -78.081 50.118
101286 95.341 -78.080 50.085
101287 95.347 -78.078 50.053
101288 95.353 -78.078 50.020
101289 95.362 -78.079 49.987
101290 95.373 -78.083 49.956
101291 95.383 -78.087 49.925
101292 95.393 -78.091 49.894
101293 95.403 -78.095 49.864
101294 95.413 -78.100 49.834
101295 95.424 -78.103 49.805
101296 95.434 -78.108 49.776
101297 95.442 -78.112 49.748
101298 95.447 -78.116 49.721
101299 95.453 -78.119 49.695
101300 95.458 -78.124 49.668
101301 95.464 -78.128 49.643
101302 95.470 -78.131 49.618
101303 95.475 -78.136 49.592
101304 95.480 -78.133 49.568
101305 95.484 -78.128 49.543
101306 95.490 -78.129 49.517
101307 95.498 -78.134 49.489
101308 95.506 -78.138 49.460
101309 95.514 -78.143 49.429
101310 95.522 -78.148 49.397
101311 95.529 -78.153 49.361
101312 95.537 -78.157 49.324
101313 95.545 -78.163 49.286
101314 95.555 -78.169 49.246
101315 95.564 -78.177 49.204
101316 95.573 -78.184 49.162
101317 95.582 -78.191 49.119
101318 95.591 -78.198 49.076
101319 95.601 -78.205 49.033
101320 95.610 -78.213 48.989
101321 95.618 -78.219 48.945
101322 95.614 -78.216 48.903
101323 95.619 -78.218 48.861
101324 95.634 -78.226 48.822
101325 95.650 -78.236 48.783
101326 95.669 -78.250 48.746
101327 95.688 -78.263 48.709
101328 95.707 -78.278 48.672
101329 95.725 -78.285 48.630
101330 95.740 -78.282 48.587
101331 95.750 -78.279 48.544
101332 95.747 -78.274 48.501
101333 95.749 -78.248 48.459
101334 95.765 -78.225 48.417
101335 95.783 -78.231 48.375
101336 95.802 -78.239 48.336
101337 95.820 -78.247 48.296
101338 95.838 -78.255 48.256
101339 95.856 -78.263 48.218
101340 95.874 -78.271 48.181
101341 95.891 -78.275 48.143
101342 95.907 -78.279 48.103
101343 95.925 -78.282 48.065
101344 95.944 -78.285 48.028
101345 95.962 -78.288 47.991
101346 95.980 -78.291 47.954
101347 95.996 -78.294 47.917
101348 96.012 -78.297 47.880
101349 96.025 -78.313 47.844
101350 96.038 -78.331 47.807
101351 96.050 -78.347 47.771
101352 96.063 -78.365 47.734
101353 96.077 -78.380 47.702
101354 96.091 -78.392 47.669
101355 96.105 -78.404 47.637
101356 96.116 -78.410 47.606
101357 96.126 -78.416 47.574
101358 96.129 -78.428 47.543
101359 96.146 -78.432 47.512
101360 96.162 -78.436 47.482
101361 96.179 -78.441 47.452
101362 96.191 -78.447 47.423
101363 96.202 -78.454 47.396
101364 96.212 -78.462 47.369
101365 96.239 -78.467 47.342
101366 96.268 -78.470 47.314
101367 96.297 -78.474 47.285
101368 96.326 -78.478 47.257
101369 96.357 -78.482 47.229
101370 96.388 -78.486 47.201
101371 96.422 -78.491 47.173
101372 96.457 -78.495 47.145
101373 96.491 -78.498 47.117
101374 96.526 -78.502 47.088
101375 96.539 -78.501 47.061
101376 96.551 -78.500 47.034
101377 96.562 -78.499 47.006
101378 96.578 -78.499 46.979
101379 96.604 -78.502 46.952
101380 96.630 -78.505 46.924
101381 96.639 -78.514 46.899
101382 96.654 -78.532 46.875
101383 96.669 -78.550 46.851
101384 96.676 -78.559 46.825
101385 96.682 -78.569 46.800
101386 96.689 -78.578 46.774
101387 96.698 -78.586 46.749
101388 96.711 -78.592 46.726
101389 96.724 -78.598 46.703
101390 96.730 -78.615 46.681
101391 96.737 -78.633 46.661
101392 96.757 -78.664 46.640
101393 96.776 -78.695 46.619
101394 96.796 -78.720 46.598
101395 96.815 -78.742 46.580
101396 96.834 -78.766 46.560
101397 96.854 -78.789 46.541
101398 96.873 -78.812 46.522
101399 96.892 -78.836 46.502
101400 96.912 -78.859 46.482
101401 96.931 -78.882 46.463
101402 96.950 -78.905 46.445
101403 96.970 -78.927 46.427
101404 96.995 -78.930 46.407
101405 97.021 -78.934 46.386
101406 97.050 -78.945 46.368
101407 97.080 -78.955 46.351
101408 97.110 -78.965 46.333
101409 97.141 -78.974 46.318
101410 97.175 -78.970 46.301
101411 97.210 -78.967 46.284
101412 97.245 -78.963 46.266
101413 97.279 -78.960 46.249
101414 97.314 -78.957 46.231
101415 97.349 -78.953 46.213
101416 97.383 -78.950 46.195
101417 97.418 -78.947 46.177
101418 97.453 -78.943 46.157
101419 97.489 -78.950 46.136
101420 97.504 -78.949 46.115
101421 97.517 -78.946 46.093
101422 97.525 -78.941 46.073
101423 97.533 -78.935 46.053
101424 97.541 -78.929 46.036
101425 97.549 -78.923 46.018
101426 97.558 -78.917 46.000
101427 97.575 -78.914 45.982
101428 97.592 -78.910 45.964
101429 97.610 -78.907 45.946
101430 97.628 -78.903 45.928
101431 97.646 -78.900 45.910
101432 97.664 -78.896 45.893
101433 97.682 -78.893 45.876
101434 97.700 -78.890 45.858
101435 97.717 -78.886 45.842
101436 97.733 -78.883 45.826
101437 97.746 -78.881 45.810
101438 97.759 -78.879 45.795
101439 97.773 -78.877 45.779
101440 97.786 -78.874 45.764
101441 97.799 -78.871 45.749
101442 97.801 -78.849 45.735
101443 97.802 -78.864 45.721
101444 97.809 -78.872 45.708
101445 97.816 -78.877 45.692
101446 97.823 -78.882 45.677
101447 97.831 -78.886 45.662
101448 97.839 -78.892 45.648
101449 97.846 -78.896 45.636
101450 97.854 -78.902 45.626
101451 97.861 -78.906 45.618
101452 97.868 -78.911 45.609
101453 97.876 -78.916 45.599
101454 97.883 -78.921 45.586
101455 97.891 -78.926 45.572
101456 97.898 -78.930 45.557
101457 97.906 -78.936 45.541
101458 97.919 -78.942 45.529
101459 97.948 -78.950 45.515
101460 97.966 -78.957 45.498
101461 97.982 -78.962 45.479
101462 97.997 -78.968 45.460
101463 98.012 -78.973 45.442
101464 98.028 -78.979 45.425
101465 98.043 -78.984 45.409
101466 98.059 -78.990 45.395
101467 98.074 -78.995 45.381
101468 98.089 -79.000 45.369
101469 98.103 -79.007 45.355
101470 98.120 -79.024 45.341
101471 98.140 -79.043 45.327
101472 98.160 -79.062 45.316
101473 98.180 -79.082 45.305
101474 98.199 -79.084 45.295
101475 98.218 -79.077 45.286
101476 98.248 -79.081 45.274
101477 98.278 -79.085 45.264
101478 98.308 -79.090 45.253
101479 98.339 -79.094 45.243
101480 98.369 -79.098 45.233
101481 98.400 -79.103 45.223
101482 98.431 -79.108 45.213
101483 98.464 -79.116 45.203
101484 98.499 -79.125 45.193
101485 98.533 -79.134 45.183
101486 98.567 -79.143 45.173
101487 98.601 -79.152 45.164
101488 98.635 -79.160 45.154
101489 98.669 -79.169 45.145
101490 98.703 -79.178 45.135
101491 98.737 -79.187 45.125
101492 98.771 -79.195 45.115
101493 98.805 -79.204 45.104
101494 98.839 -79.213 45.093
101495 98.863 -79.219 45.082
101496 98.887 -79.226 45.070
101497 98.907 -79.231 45.059
101498 98.921 -79.235 45.049
101499 98.934 -79.240 45.040
101500 98.951 -79.248 45.034
101501 98.972 -79.261 45.027
101502 98.991 -79.273 45.019
101503 99.011 -79.285 45.011
101504 99.031 -79.301 45.000
101505 99.033 -79.300 44.991
101506 99.047 -79.308 44.981
101507 99.065 -79.325 44.972
101508 99.082 -79.343 44.961
101509 99.100 -79.360 44.949
101510 99.118 -79.378 44.938
101511 99.136 -79.394 44.927
101512 99.149 -79.406 44.916
101513 99.161 -79.416 44.906
101514 99.182 -79.425 44.895
101515 99.203 -79.433 44.884
101516 99.225 -79.442 44.874
101517 99.246 -79.451 44.864
101518 99.268 -79.459 44.855
101519 99.290 -79.467 44.848
101520 99.311 -79.476 44.841
101521 99.333 -79.484 44.833
101522 99.355 -79.492 44.827
101523 99.376 -79.501 44.820
101524 99.398 -79.510 44.813
101525 99.420 -79.518 44.806
101526 99.442 -79.526 44.801
101527 99.464 -79.535 44.795
101528 99.484 -79.541 44.789
101529 99.511 -79.542 44.784
101530 99.541 -79.545 44.779
101531 99.571 -79.548 44.775
101532 99.600 -79.552 44.770
101533 99.631 -79.555 44.766
101534 99.660 -79.558 44.763
101535 99.691 -79.561 44.758
101536 99.720 -79.565 44.753
101537 99.751 -79.568 44.749
101538 99.780 -79.571 44.745
101539 99.810 -79.574 44.742
101540 99.840 -79.578 44.737
101541 99.870 -79.581 44.733
101542 99.900 -79.584 44.728
101543 99.930 -79.587 44.723
101544 99.968 -79.617 44.717
101545 100.027 -79.673 44.710
101546 99.978 -79.760 44.703
101547 99.988 -79.771 44.697
101548 100.032 -79.810 44.691
101549 100.068 -79.830 44.684
101550 100.105 -79.849 44.679
101551 100.142 -79.867 44.673
101552 100.174 -79.889 44.667
101553 100.199 -79.917 44.663
101554 100.224 -79.943 44.659
101555 100.248 -79.975 44.654
101556 100.283 -79.986 44.645
101557 100.317 -80.012 44.636
101558 100.352 -80.040 44.626
101559 100.387 -80.068 44.616
101560 100.421 -80.095 44.602
101561 100.456 -80.122 44.588
101562 100.490 -80.150 44.574
101563 100.525 -80.178 44.559
101564 100.559 -80.206 44.543
101565 100.594 -80.233 44.528
101566 100.604 -80.272 44.513
101567 100.600 -80.317 44.497
101568 100.596 -80.362 44.480
101569 100.614 -80.399 44.463
101570 100.639 -80.434 44.445
101571 100.665 -80.468 44.427
101572 100.689 -80.503 44.408
101573 100.714 -80.538 44.391
101574 100.741 -80.573 44.372
101575 100.769 -80.609 44.353
101576 100.797 -80.646 44.335
101577 100.826 -80.682 44.318
101578 100.872 -80.696 44.299
101579 100.918 -80.709 44.279
101580 100.965 -80.722 44.258
101581 101.011 -80.735 44.237
101582 101.057 -80.748 44.217
101583 101.104 -80.761 44.198
101584 101.150 -80.773 44.180
101585 101.197 -80.786 44.163
101586 101.243 -80.799 44.147
101587 101.290 -80.812 44.130
101588 101.336 -80.825 44.114
101589 101.383 -80.838 44.099
101590 101.428 -80.851 44.083
101591 101.461 -80.867 44.067
101592 101.493 -80.883 44.052
101593 101.526 -80.899 44.037
101850 93.495 -78.515 53.846
101851 93.505 -78.513 53.832
101852 93.515 -78.510 53.816
101853 93.524 -78.507 53.800
101854 93.534 -78.505 53.781
101855 93.544 -78.503 53.762
101856 93.554 -78.500 53.742
101857 93.559 -78.495 53.722
101858 93.557 -78.485 53.703
101859 93.556 -78.476 53.682
101860 93.555 -78.467 53.661
101861 93.553 -78.458 53.636
101862 93.563 -78.442 53.611
101863 93.577 -78.427 53.588
101864 93.599 -78.412 53.565
101865 93.622 -78.397 53.543
101866 93.643 -78.383 53.522
101867 93.654 -78.375 53.501
101868 93.666 -78.366 53.480
101869 93.678 -78.358 53.459
101882 93.876 -78.262 53.205
101883 93.892 -78.254 53.185
101884 93.907 -78.247 53.166
101885 93.922 -78.240 53.147
101886 93.938 -78.232 53.129
101887 93.951 -78.228 53.111
101888 93.964 -78.223 53.093
101889 93.965 -78.222 53.078
101890 93.968 -78.232 53.065
101891 93.972 -78.242 53.055
101892 93.975 -78.253 53.049
101893 93.978 -78.263 53.041
101894 93.982 -78.273 53.031
101895 93.985 -78.284 53.019
101896 93.989 -78.294 53.007
101897 93.992 -78.304 52.993
101898 93.995 -78.314 52.978
101899 93.999 -78.325 52.963
101900 94.003 -78.335 52.946
101901 94.010 -78.338 52.928
101902 94.018 -78.343 52.907
101903 94.026 -78.347 52.889
101904 94.033 -78.351 52.871
101905 94.041 -78.355 52.853
101906 94.048 -78.360 52.835
101907 94.055 -78.367 52.820
101908 94.061 -78.375 52.806
101909 94.068 -78.382 52.792
101910 94.074 -78.389 52.777
101911 94.081 -78.396 52.763
101912 94.087 -78.404 52.749
101913 94.094 -78.411 52.736
101914 94.105 -78.410 52.723
101915 94.117 -78.406 52.709
101916 94.129 -78.403 52.694
101917 94.130 -78.385 52.677
101918 94.131 -78.364 52.658
101919 94.131 -78.343 52.637
101920 94.133 -78.322 52.617
101921 94.135 -78.300 52.596
101922 94.144 -78.282 52.573
101923 94.165 -78.284 52.547
101924 94.186 -78.285 52.519
101925 94.207 -78.288 52.492
101926 94.228 -78.290 52.464
101927 94.250 -78.291 52.436
101928 94.270 -78.294 52.406
101929 94.291 -78.296 52.374
101930 94.299 -78.313 52.340
101931 94.318 -78.310 52.303
101932 94.339 -78.302 52.263
101933 94.360 -78.294 52.220
101934 94.381 -78.285 52.176
101935 94.402 -78.278 52.129
101936 94.423 -78.269 52.083
101937 94.447 -78.264 52.038
101938 94.484 -78.278 51.994
101939 94.522 -78.292 51.954
101940 94.559 -78.306 51.914
101941 94.597 -78.319 51.875
101942 94.627 -78.317 51.830
101943 94.648 -78.304 51.784
101944 94.659 -78.297 51.737
101945 94.669 -78.291 51.691
101946 94.680 -78.286 51.647
101947 94.690 -78.280 51.604
101948 94.698 -78.271 51.562
101949 94.706 -78.263 51.521
101950 94.714 -78.253 51.482
101951 94.721 -78.250 51.443
101952 94.733 -78.247 51.404
101953 94.744 -78.244 51.366
101954 94.756 -78.242 51.329
101955 94.767 -78.239 51.293
101956 94.779 -78.242 51.257
101957 94.791 -78.247 51.221
101958 94.804 -78.252 51.185
101959 94.816 -78.257 51.149
101960 94.828 -78.262 51.112
101961 94.840 -78.266 51.074
101962 94.852 -78.271 51.036
101963 94.864 -78.276 50.997
101964 94.875 -78.281 50.957
101965 94.894 -78.279 50.915
101966 94.915 -78.258 50.875
101967 94.936 -78.237 50.834
101968 94.957 -78.216 50.793
101969 94.978 -78.194 50.752
101970 94.998 -78.174 50.711
101971 95.020 -78.153 50.671
101972 95.034 -78.140 50.631
101973 95.069 -78.126 50.591
101974 95.113 -78.120 50.552
101975 95.157 -78.114 50.515
101976 95.200 -78.108 50.477
101977 95.244 -78.102 50.440
101978 95.287 -78.097 50.403
101979 95.306 -78.091 50.365
101980 95.315 -78.087 50.330
101981 95.322 -78.085 50.294
101982 95.329 -78.084 50.259
101983 95.335 -78.082 50.224
101984 95.341 -78.081 50.190
101985 95.347 -78.080 50.156
101986 95.354 -78.078 50.122
101987 95.360 -78.077 50.091
101988 95.367 -78.075 50.059
101989 95.373 -78.074 50.028
101990 95.379 -78.072 49.996
101991 95.389 -78.076 49.966
101992 95.399 -78.080 49.936
101993 95.410 -78.084 49.905
101994 95.420 -78.088 49.876
101995 95.426 -78.092 49.848
101996 95.432 -78.096 49.820
101997 95.438 -78.100 49.792
101998 95.444 -78.104 49.765
101999 95.449 -78.108 49.738
102000 95.455 -78.112 49.711
102001 95.461 -78.116 49.685
102002 95.466 -78.120 49.660
102003 95.472 -78.124 49.635
102004 95.477 -78.127 49.610
102005 95.481 -78.122 49.586
102006 95.485 -78.118 49.561
102007 95.489 -78.113 49.535
102008 95.497 -78.119 49.507
102009 95.505 -78.123 49.477
102010 95.513 -78.128 49.445
102011 95.521 -78.132 49.412
102012 95.529 -78.138 49.375
102013 95.537 -78.142 49.338
102014 95.545 -78.147 49.299
102015 95.553 -78.151 49.258
102016 95.560 -78.156 49.215
102017 95.568 -78.161 49.172
102018 95.578 -78.167 49.128
102019 95.587 -78.175 49.084
102020 95.596 -78.181 49.040
102021 95.605 -78.188 48.996
102022 95.614 -78.196 48.952
102023 95.613 -78.194 48.909
102024 95.624 -78.201 48.867
102025 95.641 -78.212 48.828
102026 95.660 -78.225 48.789
102027 95.679 -78.239 48.751
102028 95.698 -78.253 48.713
102029 95.717 -78.267 48.675
102030 95.735 -78.275 48.635
102031 95.739 -78.268 48.593
102032 95.745 -78.261 48.551
102033 95.761 -78.260 48.508
102034 95.772 -78.252 48.466
102035 95.773 -78.232 48.425
102036 95.786 -78.229 48.385
102037 95.804 -78.235 48.347
102038 95.822 -78.241 48.308
102039 95.840 -78.247 48.268
102040 95.859 -78.252 48.230
102041 95.877 -78.257 48.194
102042 95.895 -78.258 48.157
102043 95.913 -78.261 48.116
102044 95.932 -78.264 48.077
102045 95.950 -78.267 48.039
102046 95.968 -78.270 48.002
102047 95.987 -78.272 47.964
102048 96.004 -78.275 47.928
102049 96.020 -78.280 47.892
102050 96.033 -78.297 47.855
102051 96.045 -78.315 47.819
102052 96.058 -78.332 47.783
102053 96.071 -78.349 47.749
102054 96.084 -78.366 47.716
102055 96.097 -78.382 47.684
102056 96.111 -78.395 47.652
102057 96.121 -78.401 47.621
102058 96.131 -78.406 47.590
102059 96.141 -78.413 47.560
102060 96.157 -78.418 47.529
102061 96.167 -78.426 47.498
102062 96.177 -78.433 47.468
102063 96.188 -78.440 47.439
102064 96.198 -78.448 47.411
102065 96.211 -78.455 47.384
102066 96.240 -78.459 47.357
102067 96.269 -78.463 47.330
102068 96.299 -78.467 47.302
102069 96.330 -78.471 47.273
102070 96.363 -78.476 47.246
102071 96.395 -78.480 47.219
102072 96.430 -78.484 47.191
102073 96.464 -78.488 47.162
102074 96.499 -78.492 47.133
102075 96.534 -78.495 47.105
102076 96.568 -78.499 47.077
102077 96.602 -78.503 47.050
102078 96.614 -78.502 47.022
102079 96.625 -78.501 46.996
102080 96.638 -78.500 46.969
102081 96.634 -78.501 46.942
102082 96.649 -78.520 46.918
102083 96.656 -78.530 46.894
102084 96.663 -78.539 46.871
102085 96.670 -78.549 46.845
102086 96.676 -78.558 46.820
102087 96.683 -78.567 46.794
102088 96.694 -78.574 46.768
102089 96.707 -78.580 46.744
102090 96.720 -78.586 46.722
102091 96.729 -78.598 46.700
102092 96.736 -78.615 46.680
102093 96.754 -78.645 46.659
102094 96.773 -78.676 46.638
102095 96.793 -78.707 46.616
102096 96.813 -78.738 46.597
102097 96.832 -78.766 46.578
102098 96.852 -78.789 46.559
102099 96.871 -78.812 46.540
102100 96.890 -78.836 46.520
102101 96.909 -78.859 46.500
102102 96.929 -78.882 46.480
102103 96.948 -78.905 46.462
102104 96.967 -78.929 46.443
102105 96.991 -78.938 46.423
102106 97.015 -78.940 46.404
102107 97.043 -78.948 46.385
102108 97.073 -78.959 46.366
102109 97.102 -78.969 46.348
102110 97.135 -78.971 46.331
102111 97.170 -78.968 46.314
102112 97.205 -78.964 46.296
102113 97.239 -78.961 46.278
102114 97.274 -78.958 46.260
102115 97.308 -78.954 46.242
102116 97.343 -78.951 46.223
102117 97.378 -78.948 46.205
102118 97.413 -78.944 46.186
102119 97.447 -78.941 46.167
102120 97.482 -78.937 46.148
102121 97.510 -78.943 46.127
102122 97.524 -78.941 46.106
102123 97.532 -78.935 46.084
102124 97.541 -78.930 46.065
102125 97.549 -78.924 46.046
102126 97.557 -78.918 46.028
102127 97.575 -78.915 46.011
102128 97.593 -78.911 45.993
102129 97.611 -78.908 45.975
102130 97.629 -78.904 45.957
102131 97.646 -78.901 45.939
102132 97.664 -78.897 45.921
102133 97.682 -78.894 45.904
102134 97.700 -78.890 45.888
102135 97.718 -78.887 45.870
102136 97.736 -78.883 45.853
102137 97.752 -78.880 45.837
102138 97.765 -78.878 45.820
102139 97.778 -78.876 45.805
102140 97.791 -78.874 45.790
102141 97.802 -78.868 45.775
102142 97.805 -78.846 45.759
102143 97.807 -78.825 45.743
102144 97.808 -78.826 45.729
102145 97.809 -78.861 45.716
102146 97.816 -78.868 45.702
102147 97.823 -78.873 45.687
102148 97.831 -78.878 45.673
102149 97.839 -78.883 45.660
102150 97.846 -78.888 45.648
102151 97.853 -78.893 45.639
102152 97.861 -78.897 45.630
102153 97.868 -78.902 45.619
102154 97.876 -78.907 45.608
102155 97.883 -78.912 45.597
102156 97.891 -78.917 45.584
102157 97.898 -78.922 45.570
102158 97.921 -78.930 45.557
102159 97.950 -78.939 45.544
102160 97.979 -78.948 45.529
102161 98.004 -78.956 45.512
102162 98.020 -78.962 45.492
102163 98.035 -78.967 45.471
102164 98.051 -78.972 45.451
102165 98.066 -78.977 45.433
102166 98.081 -78.984 45.416
102167 98.096 -78.989 45.401
102168 98.112 -78.994 45.387
102169 98.121 -79.003 45.374
102170 98.126 -79.015 45.361
102171 98.140 -79.031 45.348
102172 98.160 -79.049 45.335
102173 98.180 -79.069 45.323
102174 98.201 -79.084 45.311
102175 98.218 -79.078 45.301
102176 98.229 -79.062 45.291
102177 98.260 -79.067 45.281
102178 98.290 -79.071 45.271
102179 98.320 -79.075 45.261
102180 98.350 -79.080 45.249
102181 98.380 -79.084 45.237
102182 98.411 -79.088 45.226
102183 98.443 -79.094 45.216
102184 98.477 -79.103 45.205
102185 98.511 -79.112 45.194
102186 98.545 -79.121 45.183
102187 98.579 -79.129 45.173
102188 98.613 -79.138 45.163
102189 98.647 -79.147 45.152
102190 98.681 -79.156 45.142
102191 98.715 -79.165 45.133
102192 98.749 -79.173 45.123
102193 98.783 -79.182 45.113
102194 98.814 -79.190 45.103
102195 98.838 -79.197 45.092
102196 98.862 -79.203 45.081
102197 98.886 -79.210 45.070
102198 98.908 -79.216 45.059
102199 98.922 -79.220 45.049
102200 98.935 -79.224 45.040
102201 98.948 -79.228 45.033
102202 98.961 -79.231 45.025
102203 98.980 -79.241 45.017
102204 99.000 -79.254 45.007
102205 99.016 -79.285 44.996
102206 99.014 -79.262 44.987
102207 99.029 -79.272 44.978
102208 99.047 -79.290 44.968
102209 99.065 -79.307 44.956
102210 99.082 -79.324 44.944
102211 99.100 -79.341 44.933
102212 99.120 -79.359 44.921
102213 99.144 -79.376 44.911
102214 99.163 -79.392 44.900
102215 99.185 -79.400 44.890
102216 99.207 -79.409 44.880
102217 99.228 -79.417 44.871
102218 99.251 -79.426 44.863
102219 99.272 -79.434 44.855
102220 99.294 -79.442 44.848
102221 99.316 -79.451 44.841
102222 99.338 -79.458 44.834
102223 99.360 -79.466 44.828
102224 99.381 -79.475 44.822
102225 99.404 -79.483 44.816
102226 99.425 -79.491 44.810
102227 99.447 -79.500 44.804
102228 99.469 -79.508 44.799
102229 99.492 -79.523 44.794
102230 99.522 -79.527 44.789
102231 99.552 -79.531 44.784
102232 99.582 -79.535 44.780
102233 99.613 -79.538 44.776
102234 99.642 -79.541 44.773
102235 99.673 -79.545 44.769
102236 99.703 -79.549 44.765
102237 99.733 -79.553 44.761
102238 99.763 -79.557 44.758
102239 99.793 -79.560 44.755
102240 99.823 -79.563 44.751
102241 99.853 -79.567 44.746
102242 99.884 -79.571 44.741
102243 99.914 -79.575 44.736
102244 99.943 -79.579 44.731
102245 99.972 -79.584 44.725
102246 99.974 -79.683 44.719
102247 99.989 -79.713 44.711
102248 100.039 -79.753 44.704
102249 100.075 -79.772 44.698
102250 100.109 -79.793 44.690
102251 100.134 -79.820 44.685
102252 100.159 -79.847 44.680
102253 100.184 -79.874 44.674
102254 100.209 -79.901 44.671
102255 100.234 -79.928 44.667
102256 100.257 -79.963 44.662
102257 100.288 -79.980 44.653
102258 100.326 -79.986 44.644
102259 100.360 -80.012 44.633
102260 100.395 -80.040 44.620
102261 100.429 -80.068 44.607
102262 100.464 -80.095 44.594
102263 100.498 -80.122 44.579
102264 100.533 -80.150 44.565
102265 100.568 -80.178 44.549
102266 100.602 -80.206 44.534
102267 100.616 -80.243 44.518
102268 100.612 -80.287 44.502
102269 100.608 -80.332 44.486
102270 100.632 -80.368 44.469
102271 100.656 -80.402 44.451
102272 100.682 -80.437 44.433
102273 100.707 -80.472 44.415
102274 100.732 -80.507 44.397
102275 100.760 -80.543 44.379
102276 100.788 -80.579 44.362
102277 100.817 -80.616 44.346
102278 100.845 -80.652 44.328
102279 100.876 -80.686 44.310
102280 100.922 -80.699 44.291
102281 100.969 -80.712 44.270
102282 101.015 -80.725 44.249
102283 101.061 -80.738 44.228
102284 101.108 -80.751 44.207
102285 101.155 -80.764 44.188
102286 101.201 -80.776 44.170
102287 101.247 -80.789 44.154
102288 101.294 -80.802 44.137
102289 101.340 -80.815 44.121
102290 101.384 -80.829 44.105
102291 101.417 -80.845 44.089
102292 101.450 -80.861 44.073
102293 101.482 -80.877 44.057
102294 101.515 -80.893 44.042
102551 93.496 -78.504 53.848
102552 93.505 -78.501 53.833
102553 93.515 -78.499 53.817
102554 93.524 -78.497 53.800
102555 93.535 -78.494 53.781
102556 93.544 -78.492 53.762
102557 93.554 -78.489 53.742
102558 93.555 -78.482 53.721
102559 93.554 -78.473 53.701
102560 93.553 -78.463 53.680
102561 93.551 -78.454 53.657
102562 93.555 -78.442 53.632
102563 93.565 -78.427 53.607
102564 93.582 -78.411 53.582
102565 93.605 -78.397 53.560
102566 93.627 -78.382 53.537
102567 93.644 -78.369 53.515
102568 93.656 -78.362 53.494
102569 93.667 -78.354 53.473
102570 93.679 -78.345 53.453
102583 93.878 -78.249 53.195
102584 93.894 -78.242 53.175
102585 93.909 -78.235 53.156
102586 93.924 -78.227 53.137
102587 93.939 -78.221 53.118
102588 93.952 -78.216 53.099
102589 93.965 -78.213 53.081
102590 93.965 -78.210 53.065
102591 93.968 -78.219 53.051
102592 93.971 -78.229 53.040
102593 93.974 -78.239 53.031
102594 93.978 -78.250 53.021
102595 93.981 -78.260 53.010
102596 93.985 -78.270 52.998
102597 93.988 -78.281 52.984
102598 93.991 -78.291 52.969
102599 93.995 -78.301 52.953
102600 93.998 -78.312 52.937
102601 94.002 -78.322 52.919
102602 94.005 -78.332 52.899
102603 94.008 -78.342 52.877
102604 94.013 -78.351 52.859
102605 94.020 -78.356 52.841
102606 94.028 -78.360 52.823
102607 94.036 -78.364 52.807
102608 94.043 -78.369 52.792
102609 94.049 -78.376 52.778
102610 94.056 -78.384 52.763
102611 94.062 -78.391 52.747
102612 94.069 -78.398 52.731
102613 94.079 -78.398 52.716
102614 94.091 -78.395 52.702
102615 94.104 -78.391 52.689
102616 94.115 -78.388 52.674
102617 94.127 -78.385 52.657
102618 94.130 -78.369 52.639
102619 94.131 -78.349 52.620
102620 94.131 -78.328 52.599
102621 94.132 -78.307 52.578
102622 94.132 -78.287 52.557
102623 94.135 -78.269 52.535
102624 94.156 -78.271 52.509
102625 94.177 -78.273 52.482
102626 94.198 -78.275 52.454
102627 94.219 -78.277 52.426
102628 94.241 -78.279 52.397
102629 94.261 -78.281 52.367
102630 94.282 -78.283 52.334
102631 94.297 -78.292 52.300
102632 94.318 -78.284 52.262
102633 94.339 -78.275 52.220
102634 94.360 -78.267 52.175
102635 94.382 -78.259 52.130
102636 94.403 -78.250 52.084
102637 94.423 -78.242 52.039
102638 94.445 -78.234 51.997
102639 94.479 -78.244 51.958
102640 94.517 -78.258 51.920
102641 94.554 -78.272 51.881
102642 94.592 -78.286 51.840
102643 94.629 -78.300 51.797
102644 94.653 -78.291 51.752
102645 94.664 -78.285 51.705
102646 94.674 -78.279 51.661
102647 94.682 -78.270 51.619
102648 94.690 -78.261 51.577
102649 94.698 -78.253 51.538
102650 94.706 -78.244 51.499
102651 94.714 -78.235 51.460
102652 94.720 -78.232 51.422
102653 94.730 -78.230 51.384
102654 94.742 -78.228 51.347
102655 94.753 -78.225 51.311
102656 94.765 -78.222 51.275
102657 94.777 -78.225 51.239
102658 94.789 -78.229 51.203
102659 94.801 -78.235 51.166
102660 94.813 -78.239 51.130
102661 94.825 -78.244 51.093
102662 94.837 -78.249 51.057
102663 94.849 -78.253 51.020
102664 94.862 -78.259 50.981
102665 94.874 -78.263 50.942
102666 94.890 -78.263 50.903
102667 94.910 -78.244 50.863
102668 94.931 -78.222 50.823
102669 94.952 -78.201 50.783
102670 94.973 -78.180 50.744
102671 94.993 -78.159 50.704
102672 95.015 -78.138 50.665
102673 95.035 -78.117 50.625
102674 95.075 -78.108 50.587
102675 95.119 -78.103 50.550
102676 95.163 -78.097 50.512
102677 95.207 -78.091 50.475
102678 95.251 -78.085 50.439
102679 95.295 -78.079 50.402
102680 95.318 -78.079 50.366
102681 95.325 -78.078 50.331
102682 95.332 -78.078 50.296
102683 95.340 -78.077 50.261
102684 95.347 -78.075 50.227
102685 95.354 -78.075 50.194
102686 95.362 -78.074 50.161
102687 95.370 -78.073 50.128
102688 95.377 -78.072 50.097
102689 95.384 -78.071 50.067
102690 95.392 -78.070 50.037
102691 95.399 -78.069 50.006
102692 95.406 -78.069 49.977
102693 95.411 -78.073 49.947
102694 95.417 -78.077 49.917
102695 95.423 -78.081 49.889
102696 95.428 -78.084 49.862
102697 95.435 -78.089 49.835
102698 95.440 -78.093 49.808
102699 95.446 -78.097 49.782
102700 95.451 -78.101 49.756
102701 95.457 -78.105 49.730
102702 95.463 -78.109 49.704
102703 95.468 -78.112 49.679
102704 95.474 -78.117 49.654
102705 95.479 -78.116 49.629
102706 95.483 -78.112 49.605
102707 95.487 -78.106 49.579
102708 95.491 -78.102 49.552
102709 95.497 -78.103 49.523
102710 95.505 -78.107 49.492
102711 95.513 -78.112 49.459
102712 95.521 -78.117 49.423
102713 95.529 -78.122 49.386
102714 95.536 -78.126 49.348
102715 95.544 -78.131 49.308
102716 95.552 -78.136 49.266
102717 95.560 -78.141 49.223
102718 95.568 -78.145 49.179
102719 95.576 -78.150 49.134
102720 95.584 -78.155 49.091
102721 95.592 -78.159 49.047
102722 95.600 -78.165 49.003
102723 95.609 -78.172 48.959
102724 95.614 -78.175 48.916
102725 95.632 -78.188 48.874
102726 95.651 -78.201 48.833
102727 95.671 -78.210 48.794
102728 95.690 -78.216 48.755
102729 95.710 -78.222 48.716
102730 95.730 -78.237 48.678
102731 95.746 -78.253 48.638
102732 95.758 -78.260 48.598
102733 95.764 -78.253 48.556
102734 95.771 -78.247 48.516
102735 95.778 -78.242 48.475
102736 95.785 -78.238 48.434
102737 95.790 -78.231 48.396
102738 95.807 -78.232 48.357
102739 95.825 -78.232 48.319
102740 95.843 -78.234 48.280
102741 95.861 -78.235 48.243
102742 95.881 -78.238 48.206
102743 95.900 -78.239 48.168
102744 95.919 -78.242 48.128
102745 95.938 -78.245 48.089
102746 95.956 -78.248 48.050
102747 95.974 -78.251 48.013
102748 95.993 -78.254 47.976
102749 96.012 -78.257 47.940
102750 96.028 -78.264 47.904
102751 96.040 -78.282 47.868
102752 96.053 -78.299 47.832
102753 96.066 -78.316 47.797
102754 96.079 -78.334 47.764
102755 96.091 -78.351 47.731
102756 96.104 -78.368 47.700
102757 96.117 -78.385 47.668
102758 96.127 -78.388 47.637
102759 96.138 -78.390 47.607
102760 96.154 -78.399 47.576
102761 96.160 -78.410 47.545
102762 96.163 -78.419 47.515
102763 96.174 -78.426 47.485
102764 96.184 -78.434 47.456
102765 96.195 -78.441 47.428
102766 96.212 -78.448 47.400
102767 96.241 -78.451 47.373
102768 96.272 -78.456 47.346
102769 96.305 -78.460 47.318
102770 96.337 -78.465 47.291
102771 96.369 -78.470 47.264
102772 96.403 -78.473 47.238
102773 96.438 -78.477 47.209
102774 96.472 -78.481 47.179
102775 96.507 -78.485 47.150
102776 96.541 -78.489 47.122
102777 96.576 -78.492 47.095
102778 96.610 -78.496 47.068
102779 96.645 -78.500 47.041
102780 96.677 -78.504 47.015
102781 96.670 -78.504 46.988
102782 96.648 -78.504 46.961
102783 96.644 -78.510 46.937
102784 96.650 -78.520 46.913
102785 96.657 -78.529 46.889
102786 96.663 -78.539 46.865
102787 96.670 -78.548 46.840
102788 96.677 -78.557 46.814
102789 96.690 -78.563 46.788
102790 96.702 -78.568 46.764
102791 96.716 -78.574 46.741
102792 96.728 -78.580 46.719
102793 96.735 -78.598 46.699
102794 96.752 -78.626 46.677
102795 96.771 -78.658 46.656
102796 96.791 -78.688 46.634
102797 96.810 -78.719 46.615
102798 96.829 -78.750 46.595
102799 96.849 -78.781 46.576
102800 96.868 -78.812 46.556
102801 96.888 -78.836 46.537
102802 96.907 -78.859 46.517
102803 96.927 -78.882 46.496
102804 96.946 -78.905 46.478
102805 96.965 -78.929 46.458
102806 96.986 -78.946 46.438
102807 97.011 -78.949 46.421
102808 97.035 -78.951 46.402
102809 97.065 -78.962 46.382
102810 97.095 -78.971 46.363
102811 97.130 -78.968 46.345
102812 97.164 -78.965 46.327
102813 97.199 -78.962 46.309
102814 97.234 -78.959 46.289
102815 97.268 -78.955 46.270
102816 97.303 -78.952 46.250
102817 97.338 -78.949 46.230
102818 97.372 -78.945 46.212
102819 97.407 -78.942 46.194
102820 97.442 -78.938 46.176
102821 97.476 -78.935 46.158
102822 97.511 -78.933 46.138
102823 97.532 -78.936 46.117
102824 97.540 -78.930 46.097
102825 97.548 -78.924 46.077
102826 97.558 -78.918 46.057
102827 97.575 -78.915 46.038
102828 97.593 -78.912 46.020
102829 97.611 -78.908 46.002
102830 97.629 -78.905 45.984
102831 97.647 -78.902 45.967
102832 97.665 -78.898 45.950
102833 97.682 -78.895 45.933
102834 97.701 -78.891 45.917
102835 97.718 -78.888 45.902
102836 97.736 -78.884 45.884
102837 97.754 -78.881 45.866
102838 97.771 -78.878 45.849
102839 97.784 -78.876 45.834
102840 97.797 -78.873 45.818
102841 97.806 -78.865 45.803
102842 97.809 -78.843 45.789
102843 97.811 -78.822 45.772
102844 97.813 -78.801 45.755
102845 97.815 -78.788 45.738
102846 97.816 -78.823 45.724
102847 97.816 -78.858 45.710
102848 97.823 -78.865 45.696
102849 97.831 -78.869 45.683
102850 97.839 -78.874 45.671
102851 97.846 -78.879 45.659
102852 97.853 -78.883 45.647
102853 97.861 -78.889 45.636
102854 97.868 -78.893 45.625
102855 97.876 -78.899 45.613
102856 97.883 -78.903 45.603
102857 97.895 -78.909 45.590
102858 97.924 -78.918 45.577
102859 97.952 -78.927 45.564
102860 97.981 -78.937 45.550
102861 98.010 -78.945 45.535
102862 98.038 -78.954 45.519
102863 98.058 -78.961 45.500
102864 98.073 -78.966 45.478
102865 98.088 -78.971 45.458
102866 98.104 -78.977 45.439
102867 98.119 -78.983 45.423
102868 98.133 -78.989 45.407
102869 98.139 -78.999 45.393
102870 98.144 -79.011 45.380
102871 98.149 -79.022 45.367
102872 98.161 -79.037 45.354
102873 98.181 -79.056 45.341
102874 98.202 -79.076 45.330
102875 98.220 -79.079 45.318
102876 98.238 -79.073 45.308
102877 98.248 -79.057 45.298
102878 98.271 -79.053 45.288
102879 98.301 -79.057 45.278
102880 98.331 -79.061 45.268
102881 98.361 -79.065 45.256
102882 98.392 -79.069 45.243
102883 98.422 -79.074 45.231
102884 98.455 -79.081 45.220
102885 98.489 -79.090 45.208
102886 98.523 -79.099 45.196
102887 98.557 -79.107 45.184
102888 98.591 -79.116 45.174
102889 98.625 -79.125 45.163
102890 98.659 -79.134 45.152
102891 98.694 -79.143 45.141
102892 98.727 -79.151 45.131
102893 98.762 -79.160 45.121
102894 98.789 -79.168 45.110
102895 98.813 -79.174 45.100
102896 98.837 -79.181 45.090
102897 98.861 -79.187 45.079
102898 98.885 -79.194 45.069
102899 98.909 -79.200 45.060
102900 98.923 -79.204 45.050
102901 98.936 -79.208 45.040
102902 98.949 -79.212 45.029
102903 98.962 -79.216 45.019
102904 98.976 -79.220 45.010
102905 98.989 -79.224 45.002
102906 98.997 -79.248 44.992
102907 98.989 -79.222 44.982
102908 99.020 -79.237 44.972
102909 99.044 -79.254 44.961
102910 99.068 -79.272 44.950
102911 99.092 -79.290 44.939
102912 99.116 -79.307 44.928
102913 99.140 -79.325 44.918
102914 99.163 -79.342 44.908
102915 99.180 -79.359 44.898
102916 99.193 -79.372 44.888
102917 99.214 -79.380 44.879
102918 99.236 -79.388 44.871
102919 99.258 -79.397 44.863
102920 99.280 -79.405 44.856
102921 99.302 -79.413 44.849
102922 99.324 -79.421 44.843
102923 99.346 -79.429 44.836
102924 99.367 -79.438 44.830
102925 99.389 -79.446 44.825
102926 99.411 -79.454 44.819
102927 99.433 -79.463 44.814
102928 99.455 -79.474 44.808
102929 99.479 -79.489 44.803
102930 99.502 -79.504 44.800
102931 99.532 -79.508 44.795
102932 99.562 -79.511 44.791
102933 99.592 -79.515 44.787
102934 99.623 -79.519 44.784
102935 99.652 -79.523 44.780
102936 99.683 -79.526 44.777
102937 99.713 -79.530 44.773
102938 99.743 -79.533 44.770
102939 99.773 -79.537 44.768
102940 99.803 -79.541 44.767
102941 99.833 -79.545 44.762
102942 99.863 -79.548 44.757
102943 99.892 -79.552 44.751
102944 99.917 -79.557 44.746
102945 99.943 -79.561 44.740
102946 99.974 -79.582 44.734
102947 100.007 -79.634 44.728
102948 100.040 -79.689 44.721
102949 100.069 -79.723 44.713
102950 100.094 -79.751 44.705
102951 100.119 -79.777 44.698
102952 100.144 -79.805 44.691
102953 100.169 -79.831 44.686
102954 100.194 -79.858 44.682
102955 100.219 -79.886 44.680
102956 100.244 -79.912 44.675
102957 100.266 -79.950 44.668
102958 100.294 -79.976 44.658
102959 100.331 -79.981 44.649
102960 100.368 -79.986 44.638
102961 100.403 -80.012 44.625
102962 100.437 -80.040 44.612
102963 100.472 -80.067 44.599
102964 100.507 -80.095 44.585
102965 100.541 -80.122 44.569
102966 100.576 -80.150 44.554
102967 100.611 -80.178 44.539
102968 100.628 -80.213 44.522
102969 100.624 -80.258 44.506
102970 100.624 -80.302 44.490
102971 100.649 -80.336 44.474
102972 100.674 -80.371 44.455
102973 100.699 -80.406 44.438
102974 100.724 -80.441 44.421
102975 100.752 -80.476 44.405
102976 100.780 -80.513 44.387
102977 100.809 -80.549 44.371
102978 100.837 -80.585 44.354
102979 100.865 -80.622 44.336
102980 100.894 -80.659 44.316
102981 100.926 -80.690 44.297
102982 100.973 -80.702 44.279
102983 101.019 -80.715 44.257
102984 101.066 -80.728 44.236
102985 101.112 -80.741 44.216
102986 101.159 -80.754 44.196
102987 101.205 -80.767 44.178
102988 101.252 -80.779 44.161
102989 101.298 -80.792 44.145
102990 101.340 -80.807 44.127
102991 101.373 -80.823 44.111
102992 101.405 -80.839 44.094
102993 101.438 -80.855 44.078
102994 101.471 -80.871 44.062
102995 101.503 -80.887 44.047
103252 93.496 -78.493 53.848
103253 93.506 -78.491 53.833
103254 93.515 -78.489 53.817
103255 93.525 -78.485 53.798
103256 93.535 -78.483 53.779
103257 93.544 -78.481 53.760
103258 93.553 -78.478 53.739
103259 93.552 -78.469 53.718
103260 93.551 -78.460 53.697
103261 93.549 -78.450 53.675
103262 93.548 -78.441 53.651
103263 93.556 -78.426 53.626
103264 93.566 -78.411 53.601
103265 93.587 -78.396 53.577
103266 93.610 -78.381 53.553
103267 93.633 -78.366 53.531
103268 93.645 -78.357 53.509
103269 93.657 -78.348 53.488
103270 93.669 -78.341 53.466
103271 93.680 -78.332 53.446
103284 93.880 -78.236 53.186
103285 93.895 -78.229 53.165
103286 93.911 -78.222 53.145
103287 93.926 -78.214 53.125
103288 93.939 -78.210 53.106
103289 93.952 -78.205 53.087
103290 93.965 -78.201 53.068
103291 93.966 -78.199 53.050
103292 93.967 -78.206 53.035
103293 93.970 -78.216 53.021
103294 93.974 -78.226 53.009
103295 93.977 -78.236 52.996
103296 93.981 -78.247 52.982
103297 93.984 -78.257 52.969
103298 93.987 -78.267 52.954
103299 93.991 -78.278 52.937
103300 93.994 -78.288 52.919
103301 93.998 -78.298 52.902
103302 94.001 -78.309 52.884
103303 94.004 -78.319 52.864
103304 94.008 -78.329 52.843
103305 94.011 -78.339 52.823
103306 94.015 -78.350 52.806
103307 94.018 -78.360 52.790
103308 94.024 -78.368 52.776
103309 94.031 -78.373 52.762
103310 94.038 -78.379 52.747
103311 94.044 -78.386 52.731
103312 94.054 -78.386 52.714
103313 94.066 -78.383 52.697
103314 94.078 -78.380 52.680
103315 94.090 -78.377 52.667
103316 94.102 -78.374 52.652
103317 94.114 -78.371 52.636
103318 94.125 -78.368 52.619
103319 94.130 -78.354 52.599
103320 94.131 -78.334 52.579
103321 94.131 -78.313 52.560
103322 94.132 -78.292 52.539
103323 94.149 -78.291 52.517
103324 94.167 -78.290 52.493
103325 94.184 -78.289 52.468
103326 94.201 -78.288 52.441
103327 94.218 -78.287 52.413
103328 94.236 -78.285 52.384
103329 94.253 -78.285 52.355
103330 94.270 -78.282 52.325
103331 94.283 -78.275 52.292
103332 94.297 -78.267 52.257
103333 94.318 -78.257 52.216
103334 94.339 -78.248 52.173
103335 94.360 -78.240 52.127
103336 94.382 -78.232 52.083
103337 94.403 -78.223 52.040
103338 94.424 -78.215 51.999
103339 94.445 -78.206 51.962
103340 94.475 -78.210 51.927
103341 94.512 -78.224 51.890
103342 94.549 -78.238 51.852
103343 94.587 -78.252 51.811
103344 94.624 -78.266 51.768
103345 94.657 -78.279 51.723
103346 94.666 -78.269 51.678
103347 94.674 -78.260 51.636
103348 94.682 -78.252 51.595
103349 94.690 -78.243 51.556
103350 94.698 -78.234 51.518
103351 94.706 -78.225 51.479
103352 94.714 -78.216 51.441
103353 94.720 -78.213 51.404
103354 94.728 -78.213 51.368
103355 94.739 -78.210 51.332
103356 94.751 -78.208 51.297
103357 94.762 -78.205 51.262
103358 94.774 -78.207 51.226
103359 94.786 -78.212 51.190
103360 94.799 -78.217 51.154
103361 94.811 -78.222 51.117
103362 94.823 -78.227 51.081
103363 94.835 -78.232 51.045
103364 94.847 -78.237 51.009
103365 94.859 -78.241 50.970
103366 94.872 -78.246 50.932
103367 94.892 -78.244 50.894
103368 94.910 -78.227 50.856
103369 94.929 -78.209 50.816
103370 94.951 -78.188 50.777
103371 94.971 -78.166 50.739
103372 94.992 -78.145 50.700
103373 95.014 -78.125 50.661
103374 95.034 -78.103 50.622
103375 95.065 -78.089 50.585
103376 95.109 -78.083 50.548
103377 95.153 -78.077 50.512
103378 95.197 -78.071 50.476
103379 95.241 -78.065 50.439
103380 95.281 -78.061 50.404
103381 95.306 -78.063 50.369
103382 95.322 -78.064 50.333
103383 95.329 -78.062 50.298
103384 95.336 -78.062 50.264
103385 95.344 -78.061 50.232
103386 95.351 -78.060 50.199
103387 95.358 -78.059 50.167
103388 95.366 -78.059 50.135
103389 95.374 -78.057 50.106
103390 95.381 -78.056 50.076
103391 95.388 -78.056 50.047
103392 95.396 -78.055 50.018
103393 95.403 -78.052 49.989
103394 95.414 -78.062 49.959
103395 95.419 -78.068 49.930
103396 95.425 -78.072 49.903
103397 95.431 -78.078 49.876
103398 95.437 -78.081 49.850
103399 95.442 -78.085 49.824
103400 95.448 -78.089 49.800
103401 95.454 -78.094 49.776
103402 95.459 -78.097 49.751
103403 95.465 -78.101 49.725
103404 95.471 -78.106 49.700
103405 95.476 -78.109 49.675
103406 95.480 -78.106 49.649
103407 95.484 -78.100 49.624
103408 95.488 -78.096 49.597
103409 95.492 -78.091 49.569
103410 95.499 -78.088 49.539
103411 95.509 -78.094 49.506
103412 95.519 -78.100 49.470
103413 95.528 -78.105 49.432
103414 95.536 -78.109 49.393
103415 95.544 -78.115 49.354
103416 95.553 -78.119 49.313
103417 95.561 -78.125 49.270
103418 95.569 -78.129 49.227
103419 95.578 -78.134 49.184
103420 95.587 -78.140 49.139
103421 95.595 -78.144 49.096
103422 95.604 -78.150 49.053
103423 95.612 -78.154 49.009
103424 95.620 -78.159 48.966
103425 95.626 -78.152 48.923
103426 95.644 -78.152 48.880
103427 95.665 -78.158 48.838
103428 95.684 -78.164 48.798
103429 95.704 -78.170 48.759
103430 95.724 -78.189 48.720
103431 95.740 -78.207 48.681
103432 95.756 -78.223 48.641
103433 95.772 -78.239 48.601
103434 95.784 -78.244 48.561
103435 95.791 -78.238 48.523
103436 95.798 -78.234 48.483
103437 95.805 -78.228 48.444
103438 95.806 -78.219 48.406
103439 95.809 -78.212 48.367
103440 95.827 -78.213 48.330
103441 95.846 -78.215 48.294
103442 95.866 -78.217 48.255
103443 95.885 -78.219 48.218
103444 95.905 -78.221 48.179
103445 95.924 -78.223 48.139
103446 95.943 -78.226 48.101
103447 95.961 -78.229 48.063
103448 95.979 -78.232 48.024
103449 95.998 -78.235 47.988
103450 96.016 -78.239 47.953
103451 96.031 -78.247 47.917
103452 96.046 -78.260 47.881
103453 96.059 -78.278 47.846
103454 96.072 -78.295 47.811
103455 96.085 -78.312 47.779
103456 96.097 -78.329 47.747
103457 96.110 -78.347 47.715
103458 96.124 -78.360 47.685
103459 96.136 -78.364 47.653
103460 96.147 -78.366 47.623
103461 96.164 -78.378 47.592
103462 96.184 -78.397 47.561
103463 96.190 -78.407 47.531
103464 96.192 -78.416 47.502
103465 96.194 -78.425 47.474
103466 96.196 -78.434 47.447
103467 96.215 -78.440 47.418
103468 96.247 -78.445 47.389
103469 96.279 -78.449 47.362
103470 96.311 -78.454 47.336
103471 96.343 -78.459 47.309
103472 96.377 -78.463 47.283
103473 96.411 -78.467 47.257
103474 96.446 -78.470 47.227
103475 96.480 -78.474 47.197
103476 96.515 -78.478 47.168
103477 96.549 -78.482 47.140
103478 96.584 -78.485 47.113
103479 96.618 -78.489 47.087
103480 96.650 -78.489 47.061
103481 96.676 -78.490 47.035
103482 96.680 -78.499 47.008
103483 96.661 -78.501 46.981
103484 96.639 -78.501 46.956
103485 96.644 -78.509 46.932
103486 96.650 -78.518 46.908
103487 96.657 -78.528 46.884
103488 96.664 -78.537 46.860
103489 96.673 -78.545 46.834
103490 96.685 -78.551 46.809
103491 96.698 -78.557 46.786
103492 96.711 -78.563 46.762
103493 96.724 -78.569 46.738
103494 96.734 -78.580 46.717
103495 96.749 -78.608 46.696
103496 96.769 -78.639 46.673
103497 96.788 -78.670 46.651
103498 96.808 -78.700 46.632
103499 96.827 -78.731 46.611
103500 96.847 -78.762 46.590
103501 96.866 -78.793 46.571
103502 96.885 -78.824 46.552
103503 96.905 -78.855 46.531
103504 96.924 -78.882 46.511
103505 96.944 -78.905 46.492
103506 96.963 -78.929 46.473
103507 96.982 -78.952 46.454
103508 97.006 -78.957 46.436
103509 97.031 -78.959 46.418
103510 97.058 -78.965 46.398
103511 97.090 -78.969 46.378
103512 97.124 -78.965 46.359
103513 97.159 -78.962 46.340
103514 97.193 -78.959 46.320
103515 97.228 -78.955 46.299
103516 97.263 -78.952 46.278
103517 97.298 -78.949 46.258
103518 97.332 -78.946 46.238
103519 97.367 -78.943 46.219
103520 97.401 -78.939 46.200
103521 97.436 -78.936 46.182
103522 97.471 -78.933 46.164
103523 97.506 -78.929 46.146
103524 97.537 -78.929 46.126
103525 97.548 -78.924 46.106
103526 97.558 -78.919 46.088
103527 97.576 -78.916 46.069
103528 97.594 -78.912 46.051
103529 97.611 -78.909 46.032
103530 97.630 -78.905 46.013
103531 97.647 -78.902 45.995
103532 97.665 -78.899 45.977
103533 97.683 -78.896 45.961
103534 97.701 -78.892 45.944
103535 97.719 -78.889 45.928
103536 97.736 -78.885 45.913
103537 97.754 -78.882 45.897
103538 97.772 -78.878 45.880
103539 97.789 -78.875 45.863
103540 97.802 -78.873 45.848
103541 97.810 -78.861 45.831
103542 97.813 -78.840 45.815
103543 97.815 -78.819 45.800
103544 97.817 -78.798 45.783
103545 97.820 -78.777 45.766
103546 97.821 -78.755 45.749
103547 97.822 -78.785 45.733
103548 97.823 -78.820 45.719
103549 97.823 -78.854 45.705
103550 97.831 -78.861 45.693
103551 97.838 -78.865 45.680
103552 97.846 -78.871 45.668
103553 97.853 -78.875 45.656
103554 97.861 -78.880 45.644
103555 97.868 -78.885 45.632
103556 97.876 -78.890 45.620
103557 97.897 -78.897 45.609
103558 97.926 -78.906 45.596
103559 97.955 -78.915 45.582
103560 97.984 -78.924 45.569
103561 98.012 -78.933 45.555
103562 98.041 -78.943 45.539
103563 98.070 -78.952 45.523
103564 98.096 -78.960 45.505
103565 98.111 -78.965 45.484
103566 98.127 -78.971 45.464
103567 98.142 -78.976 45.446
103568 98.151 -78.985 45.429
103569 98.156 -78.996 45.414
103570 98.161 -79.007 45.400
103571 98.166 -79.018 45.386
103572 98.171 -79.029 45.373
103573 98.182 -79.043 45.360
103574 98.202 -79.063 45.348
103575 98.222 -79.079 45.336
103576 98.240 -79.073 45.325
103577 98.257 -79.067 45.315
103578 98.270 -79.054 45.305
103579 98.282 -79.038 45.295
103580 98.312 -79.043 45.285
103581 98.343 -79.046 45.274
103582 98.373 -79.051 45.263
103583 98.403 -79.055 45.249
103584 98.433 -79.059 45.236
103585 98.467 -79.068 45.224
103586 98.501 -79.077 45.212
103587 98.535 -79.085 45.200
103588 98.570 -79.094 45.188
103589 98.603 -79.103 45.177
103590 98.638 -79.112 45.166
103591 98.671 -79.121 45.155
103592 98.706 -79.129 45.144
103593 98.740 -79.138 45.133
103594 98.764 -79.145 45.122
103595 98.788 -79.151 45.112
103596 98.812 -79.158 45.101
103597 98.836 -79.164 45.091
103598 98.860 -79.171 45.081
103599 98.884 -79.178 45.070
103600 98.908 -79.184 45.061
103601 98.924 -79.188 45.050
103602 98.937 -79.192 45.040
103603 98.950 -79.197 45.028
103604 98.963 -79.200 45.018
103605 98.977 -79.204 45.008
103606 98.999 -79.219 44.998
103607 98.961 -79.212 44.988
103608 99.025 -79.227 44.978
103609 99.042 -79.214 44.968
103610 99.064 -79.220 44.957
103611 99.088 -79.237 44.946
103612 99.112 -79.256 44.935
103613 99.136 -79.273 44.925
103614 99.159 -79.291 44.915
103615 99.184 -79.308 44.906
103616 99.199 -79.324 44.897
103617 99.208 -79.339 44.888
103618 99.222 -79.351 44.879
103619 99.244 -79.359 44.871
103620 99.266 -79.367 44.864
103621 99.288 -79.375 44.858
103622 99.309 -79.384 44.852
103623 99.332 -79.392 44.846
103624 99.353 -79.400 44.839
103625 99.375 -79.409 44.834
103626 99.397 -79.417 44.828
103627 99.419 -79.426 44.823
103628 99.442 -79.440 44.818
103629 99.465 -79.455 44.814
103630 99.488 -79.470 44.809
103631 99.512 -79.485 44.806
103632 99.542 -79.488 44.802
103633 99.572 -79.492 44.799
103634 99.603 -79.496 44.796
103635 99.633 -79.500 44.792
103636 99.662 -79.503 44.789
103637 99.693 -79.507 44.787
103638 99.723 -79.510 44.783
103639 99.753 -79.514 44.781
103640 99.783 -79.518 44.779
103641 99.811 -79.522 44.776
103642 99.837 -79.526 44.771
103643 99.863 -79.531 44.766
103644 99.889 -79.535 44.761
103645 99.915 -79.541 44.756
103646 99.941 -79.548 44.751
103647 99.976 -79.581 44.745
103648 100.009 -79.631 44.738
103649 100.042 -79.686 44.731
103650 100.079 -79.708 44.722
103651 100.104 -79.735 44.714
103652 100.129 -79.762 44.706
103653 100.154 -79.789 44.699
103654 100.179 -79.816 44.693
103655 100.204 -79.842 44.690
103656 100.229 -79.870 44.687
103657 100.253 -79.899 44.681
103658 100.275 -79.937 44.674
103659 100.300 -79.971 44.663
103660 100.337 -79.976 44.652
103661 100.374 -79.981 44.641
103662 100.411 -79.986 44.629
103663 100.446 -80.012 44.616
103664 100.480 -80.040 44.603
103665 100.515 -80.067 44.589
103666 100.550 -80.095 44.572
103667 100.584 -80.122 44.557
103668 100.619 -80.150 44.542
103669 100.639 -80.184 44.526
103670 100.636 -80.228 44.510
103671 100.641 -80.270 44.494
103672 100.666 -80.305 44.477
103673 100.691 -80.340 44.461
103674 100.716 -80.375 44.444
103675 100.743 -80.410 44.428
103676 100.771 -80.446 44.411
103677 100.799 -80.483 44.393
103678 100.828 -80.519 44.376
103679 100.856 -80.556 44.358
103680 100.885 -80.592 44.340
103681 100.913 -80.629 44.322
103682 100.942 -80.665 44.303
103683 100.977 -80.693 44.284
103684 101.023 -80.706 44.264
103685 101.070 -80.719 44.244
103686 101.116 -80.732 44.224
103687 101.162 -80.745 44.203
103688 101.209 -80.757 44.185
103689 101.256 -80.770 44.168
103690 101.296 -80.785 44.151
103691 101.329 -80.801 44.133
103692 101.361 -80.817 44.116
103693 101.394 -80.833 44.100
103694 101.427 -80.848 44.084
103695 101.460 -80.865 44.067
103696 101.492 -80.881 44.051
103953 93.496 -78.482 53.848
103954 93.506 -78.479 53.833
103955 93.515 -78.477 53.816
103956 93.525 -78.475 53.797
103957 93.535 -78.473 53.777
103958 93.545 -78.470 53.757
103959 93.550 -78.464 53.736
103960 93.549 -78.455 53.715
103961 93.547 -78.446 53.693
103962 93.546 -78.437 53.670
103963 93.547 -78.426 53.645
103964 93.558 -78.410 53.620
103965 93.571 -78.395 53.595
103966 93.593 -78.380 53.571
103967 93.616 -78.365 53.547
103968 93.635 -78.352 53.524
103969 93.646 -78.344 53.502
103970 93.658 -78.335 53.480
103971 93.670 -78.327 53.459
103986 93.897 -78.216 53.155
103987 93.913 -78.209 53.136
103988 93.927 -78.203 53.114
103989 93.940 -78.198 53.094
103990 93.953 -78.194 53.074
103991 93.966 -78.190 53.054
103992 93.966 -78.188 53.035
103993 93.966 -78.192 53.018
103994 93.970 -78.203 53.001
103995 93.973 -78.213 52.986
103996 93.977 -78.223 52.970
103997 93.980 -78.234 52.954
103998 93.983 -78.244 52.939
103999 93.987 -78.254 52.922
104000 93.990 -78.264 52.904
104001 93.994 -78.275 52.885
104002 93.997 -78.285 52.867
104003 94.000 -78.295 52.849
104004 94.004 -78.306 52.829
104005 94.007 -78.316 52.807
104006 94.011 -78.326 52.787
104007 94.014 -78.336 52.769
104008 94.017 -78.347 52.754
104009 94.021 -78.357 52.741
104010 94.025 -78.367 52.727
104011 94.032 -78.374 52.711
104012 94.042 -78.371 52.693
104013 94.053 -78.369 52.676
104014 94.065 -78.366 52.658
104015 94.076 -78.363 52.641
104016 94.088 -78.360 52.627
104017 94.100 -78.357 52.611
104018 94.112 -78.354 52.594
104019 94.124 -78.350 52.575
104020 94.130 -78.339 52.556
104021 94.131 -78.319 52.536
104022 94.146 -78.315 52.515
104023 94.164 -78.313 52.494
104024 94.181 -78.313 52.471
104025 94.197 -78.309 52.447
104026 94.210 -78.301 52.422
104027 94.224 -78.294 52.395
104028 94.238 -78.286 52.368
104029 94.252 -78.279 52.340
104030 94.265 -78.271 52.310
104031 94.279 -78.263 52.279
104032 94.292 -78.256 52.245
104033 94.307 -78.249 52.208
104034 94.321 -78.241 52.166
104035 94.341 -78.231 52.121
104036 94.362 -78.220 52.075
104037 94.383 -78.210 52.034
104038 94.404 -78.200 51.996
104039 94.424 -78.191 51.962
104040 94.445 -78.181 51.932
104041 94.470 -78.176 51.900
104042 94.507 -78.192 51.865
104043 94.545 -78.209 51.826
104044 94.582 -78.225 51.786
104045 94.620 -78.241 51.742
104046 94.656 -78.257 51.698
104047 94.666 -78.250 51.654
104048 94.674 -78.242 51.613
104049 94.682 -78.233 51.573
104050 94.690 -78.224 51.536
104051 94.698 -78.216 51.500
104052 94.706 -78.206 51.463
104053 94.715 -78.197 51.426
104054 94.721 -78.193 51.389
104055 94.725 -78.196 51.354
104056 94.737 -78.194 51.319
104057 94.749 -78.191 51.284
104058 94.760 -78.188 51.250
104059 94.772 -78.190 51.215
104060 94.784 -78.194 51.179
104061 94.796 -78.200 51.143
104062 94.809 -78.204 51.107
104063 94.821 -78.210 51.071
104064 94.833 -78.214 51.035
104065 94.845 -78.219 51.000
104066 94.857 -78.224 50.962
104067 94.873 -78.229 50.924
104068 94.895 -78.226 50.886
104069 94.913 -78.209 50.849
104070 94.931 -78.191 50.811
104071 94.950 -78.174 50.772
104072 94.970 -78.153 50.734
104073 94.991 -78.131 50.696
104074 95.012 -78.110 50.658
104075 95.033 -78.090 50.620
104076 95.055 -78.069 50.583
104077 95.099 -78.063 50.547
104078 95.143 -78.057 50.512
104079 95.187 -78.052 50.477
104080 95.231 -78.046 50.441
104081 95.264 -78.044 50.405
104082 95.289 -78.046 50.371
104083 95.313 -78.048 50.336
104084 95.326 -78.048 50.302
104085 95.333 -78.047 50.269
104086 95.340 -78.047 50.237
104087 95.348 -78.046 50.205
104088 95.355 -78.044 50.173
104089 95.362 -78.044 50.143
104090 95.370 -78.043 50.114
104091 95.378 -78.042 50.085
104092 95.385 -78.041 50.057
104093 95.392 -78.040 50.029
104094 95.400 -78.036 50.000
104095 95.411 -78.044 49.971
104096 95.422 -78.056 49.943
104097 95.427 -78.061 49.917
104098 95.433 -78.065 49.891
104099 95.439 -78.071 49.866
104100 95.445 -78.076 49.843
104101 95.450 -78.081 49.821
104102 95.456 -78.086 49.799
104103 95.461 -78.090 49.775
104104 95.467 -78.094 49.750
104105 95.473 -78.098 49.724
104106 95.478 -78.099 49.698
104107 95.482 -78.094 49.670
104108 95.488 -78.089 49.642
104109 95.499 -78.081 49.612
104110 95.511 -78.074 49.581
104111 95.522 -78.067 49.547
104112 95.520 -78.084 49.511
104113 95.531 -78.090 49.473
104114 95.541 -78.096 49.433
104115 95.551 -78.101 49.392
104116 95.562 -78.107 49.352
104117 95.572 -78.113 49.312
104118 95.583 -78.119 49.270
104119 95.593 -78.125 49.228
104120 95.604 -78.131 49.185
104121 95.614 -78.136 49.142
104122 95.624 -78.142 49.099
104123 95.635 -78.148 49.056
104124 95.644 -78.153 49.013
104125 95.653 -78.158 48.970
104126 95.650 -78.140 48.928
104127 95.662 -78.119 48.886
104128 95.678 -78.112 48.843
104129 95.698 -78.118 48.803
104130 95.718 -78.139 48.764
104131 95.735 -78.161 48.725
104132 95.751 -78.177 48.685
104133 95.767 -78.193 48.646
104134 95.783 -78.209 48.606
104135 95.799 -78.225 48.567
104136 95.811 -78.230 48.529
104137 95.818 -78.225 48.492
104138 95.821 -78.218 48.455
104139 95.823 -78.208 48.417
104140 95.820 -78.204 48.380
104141 95.834 -78.196 48.343
104142 95.851 -78.196 48.305
104143 95.871 -78.198 48.266
104144 95.890 -78.200 48.228
104145 95.909 -78.202 48.190
104146 95.929 -78.204 48.151
104147 95.948 -78.207 48.112
104148 95.966 -78.210 48.074
104149 95.984 -78.213 48.036
104150 96.002 -78.217 48.001
104151 96.015 -78.224 47.965
104152 96.030 -78.232 47.929
104153 96.050 -78.240 47.894
104154 96.065 -78.255 47.861
104155 96.078 -78.272 47.827
104156 96.091 -78.290 47.794
104157 96.103 -78.307 47.762
104158 96.116 -78.324 47.731
104159 96.130 -78.332 47.700
104160 96.145 -78.339 47.668
104161 96.156 -78.341 47.638
104162 96.174 -78.356 47.608
104163 96.194 -78.375 47.577
104164 96.214 -78.394 47.547
104165 96.221 -78.405 47.518
104166 96.223 -78.414 47.491
104167 96.225 -78.423 47.463
104168 96.225 -78.433 47.434
104169 96.253 -78.438 47.407
104170 96.286 -78.443 47.380
104171 96.317 -78.448 47.354
104172 96.350 -78.452 47.328
104173 96.384 -78.456 47.302
104174 96.419 -78.460 47.275
104175 96.453 -78.463 47.246
104176 96.488 -78.467 47.217
104177 96.523 -78.472 47.187
104178 96.555 -78.473 47.160
104179 96.586 -78.473 47.133
104180 96.616 -78.472 47.107
104181 96.647 -78.471 47.080
104182 96.661 -78.476 47.054
104183 96.664 -78.485 47.027
104184 96.668 -78.494 47.000
104185 96.653 -78.496 46.976
104186 96.638 -78.498 46.952
104187 96.644 -78.508 46.928
104188 96.651 -78.517 46.904
104189 96.657 -78.526 46.880
104190 96.669 -78.533 46.854
104191 96.681 -78.539 46.829
104192 96.694 -78.545 46.806
104193 96.707 -78.551 46.782
104194 96.720 -78.557 46.759
104195 96.732 -78.563 46.735
104196 96.747 -78.589 46.713
104197 96.766 -78.620 46.690
104198 96.786 -78.651 46.668
104199 96.805 -78.682 46.649
104200 96.825 -78.712 46.629
104201 96.844 -78.743 46.608
104202 96.864 -78.774 46.587
104203 96.883 -78.805 46.567
104204 96.902 -78.836 46.546
104205 96.922 -78.868 46.525
104206 96.941 -78.898 46.506
104207 96.961 -78.929 46.487
104208 96.980 -78.952 46.467
104209 97.002 -78.966 46.449
104210 97.026 -78.968 46.430
104211 97.048 -78.962 46.410
104212 97.084 -78.965 46.389
104213 97.119 -78.962 46.370
104214 97.154 -78.959 46.350
104215 97.188 -78.956 46.329
104216 97.223 -78.953 46.308
104217 97.257 -78.949 46.286
104218 97.292 -78.946 46.265
104219 97.327 -78.943 46.244
104220 97.362 -78.940 46.226
104221 97.396 -78.937 46.206
104222 97.431 -78.933 46.187
104223 97.465 -78.930 46.170
104224 97.500 -78.927 46.152
104225 97.535 -78.924 46.133
104226 97.559 -78.920 46.114
104227 97.576 -78.917 46.096
104228 97.594 -78.913 46.079
104229 97.612 -78.910 46.061
104230 97.630 -78.906 46.043
104231 97.648 -78.903 46.024
104232 97.665 -78.899 46.006
104233 97.683 -78.896 45.988
104234 97.701 -78.893 45.972
104235 97.719 -78.890 45.955
104236 97.737 -78.886 45.939
104237 97.754 -78.883 45.924
104238 97.773 -78.879 45.908
104239 97.790 -78.876 45.892
104240 97.808 -78.873 45.876
104241 97.814 -78.858 45.860
104242 97.816 -78.837 45.844
104243 97.819 -78.816 45.827
104244 97.821 -78.795 45.811
104245 97.823 -78.774 45.794
104246 97.826 -78.752 45.777
104247 97.827 -78.731 45.760
104248 97.828 -78.747 45.744
104249 97.829 -78.782 45.730
104250 97.830 -78.817 45.716
104251 97.831 -78.851 45.703
104252 97.838 -78.857 45.690
104253 97.846 -78.861 45.678
104254 97.853 -78.866 45.665
104255 97.861 -78.871 45.654
104256 97.871 -78.877 45.641
104257 97.900 -78.886 45.628
104258 97.929 -78.895 45.614
104259 97.957 -78.904 45.601
104260 97.986 -78.913 45.587
104261 98.015 -78.922 45.573
104262 98.043 -78.931 45.559
104263 98.072 -78.940 45.542
104264 98.101 -78.949 45.524
104265 98.130 -78.958 45.506
104266 98.149 -78.965 45.487
104267 98.163 -78.971 45.469
104268 98.168 -78.982 45.452
104269 98.173 -78.993 45.437
104270 98.178 -79.004 45.422
104271 98.184 -79.015 45.407
104272 98.189 -79.026 45.392
104273 98.194 -79.037 45.380
104274 98.203 -79.050 45.367
104275 98.223 -79.069 45.355
104276 98.242 -79.073 45.344
104277 98.260 -79.067 45.333
104278 98.277 -79.062 45.322
104279 98.292 -79.052 45.312
104280 98.295 -79.026 45.301
104281 98.323 -79.028 45.291
104282 98.354 -79.032 45.280
104283 98.384 -79.037 45.268
104284 98.414 -79.040 45.255
104285 98.445 -79.046 45.242
104286 98.480 -79.055 45.230
104287 98.514 -79.063 45.218
104288 98.548 -79.072 45.205
104289 98.582 -79.081 45.193
104290 98.616 -79.090 45.182
104291 98.650 -79.099 45.171
104292 98.684 -79.107 45.160
104293 98.715 -79.115 45.149
104294 98.739 -79.122 45.137
104295 98.763 -79.128 45.126
104296 98.787 -79.135 45.116
104297 98.811 -79.141 45.105
104298 98.835 -79.148 45.095
104299 98.859 -79.155 45.084
104300 98.882 -79.161 45.073
104301 98.907 -79.168 45.062
104302 98.925 -79.173 45.051
104303 98.938 -79.177 45.040
104304 98.951 -79.181 45.029
104305 98.965 -79.184 45.018
104306 98.983 -79.195 45.007
104307 99.010 -79.217 44.996
104308 99.034 -79.237 44.986
104309 99.052 -79.228 44.975
104310 99.070 -79.214 44.965
104311 99.087 -79.200 44.955
104312 99.108 -79.203 44.944
104313 99.132 -79.222 44.934
104314 99.156 -79.239 44.925
104315 99.180 -79.256 44.915
104316 99.204 -79.274 44.906
104317 99.217 -79.290 44.897
104318 99.226 -79.304 44.888
104319 99.235 -79.319 44.880
104320 99.252 -79.330 44.873
104321 99.274 -79.338 44.867
104322 99.295 -79.347 44.861
104323 99.317 -79.355 44.855
104324 99.339 -79.363 44.849
104325 99.361 -79.372 44.843
104326 99.383 -79.380 44.838
104327 99.405 -79.390 44.834
104328 99.428 -79.406 44.829
104329 99.451 -79.421 44.826
104330 99.475 -79.437 44.822
104331 99.498 -79.452 44.818
104332 99.522 -79.466 44.815
104333 99.552 -79.469 44.812
104334 99.582 -79.473 44.809
104335 99.613 -79.476 44.805
104336 99.643 -79.480 44.803
104337 99.673 -79.484 44.800
104338 99.703 -79.488 44.798
104339 99.731 -79.491 44.795
104340 99.757 -79.496 44.793
104341 99.783 -79.501 44.789
104342 99.808 -79.505 44.786
104343 99.835 -79.510 44.781
104344 99.860 -79.515 44.777
104345 99.886 -79.520 44.772
104346 99.912 -79.524 44.767
104347 99.943 -79.547 44.762
104348 99.977 -79.579 44.756
104349 100.010 -79.629 44.750
104350 100.044 -79.683 44.741
104351 100.081 -79.701 44.733
104352 100.114 -79.720 44.724
104353 100.139 -79.746 44.715
104354 100.164 -79.773 44.707
104355 100.189 -79.800 44.701
104356 100.214 -79.827 44.698
104357 100.239 -79.854 44.692
104358 100.263 -79.886 44.685
104359 100.285 -79.925 44.677
104360 100.307 -79.964 44.666
104361 100.343 -79.971 44.656
104362 100.380 -79.976 44.644
104363 100.417 -79.981 44.632
104364 100.454 -79.986 44.620
104365 100.489 -80.012 44.607
104366 100.523 -80.040 44.592
104367 100.558 -80.067 44.576
104368 100.593 -80.095 44.560
104369 100.627 -80.122 44.544
104370 100.651 -80.154 44.528
104371 100.648 -80.199 44.512
104372 100.658 -80.239 44.496
104373 100.684 -80.274 44.480
104374 100.709 -80.309 44.464
104375 100.734 -80.344 44.448
104376 100.763 -80.380 44.432
104377 100.791 -80.416 44.415
104378 100.820 -80.453 44.397
104379 100.848 -80.489 44.380
104380 100.876 -80.526 44.362
104381 100.905 -80.562 44.344
104382 100.933 -80.598 44.326
104383 100.962 -80.635 44.308
104384 100.990 -80.672 44.289
104385 101.028 -80.696 44.270
104386 101.074 -80.709 44.251
104387 101.120 -80.722 44.231
104388 101.167 -80.735 44.210
104389 101.213 -80.748 44.192
104390 101.252 -80.763 44.174
104391 101.285 -80.779 44.156
104392 101.318 -80.795 44.139
104393 101.350 -80.811 44.122
104394 101.383 -80.826 44.106
104395 101.416 -80.843 44.090
104396 101.449 -80.859 44.073
104397 101.481 -80.875 44.057
104398 101.514 -80.891 44.041
104654 93.496 -78.471 53.847
104655 93.506 -78.469 53.831
104656 93.516 -78.467 53.814
104657 93.525 -78.464 53.796
104658 93.535 -78.461 53.775
104659 93.545 -78.459 53.754
104660 93.547 -78.451 53.733
104661 93.545 -78.442 53.711
104662 93.544 -78.433 53.688
104663 93.543 -78.424 53.664
104664 93.549 -78.410 53.640
104665 93.559 -78.395 53.615
104666 93.576 -78.380 53.590
104667 93.599 -78.365 53.565
104668 93.621 -78.350 53.541
104669 93.636 -78.339 53.518
104670 93.648 -78.331 53.495
104671 93.659 -78.322 53.473
104672 93.671 -78.314 53.452
104687 93.899 -78.203 53.145
104688 93.915 -78.196 53.123
104689 93.928 -78.191 53.102
104690 93.940 -78.187 53.081
104691 93.954 -78.183 53.060
104692 93.966 -78.178 53.040
104693 93.967 -78.177 53.020
104694 93.966 -78.179 53.001
104695 93.969 -78.189 52.983
104696 93.973 -78.200 52.964
104697 93.976 -78.210 52.946
104698 93.979 -78.220 52.928
104699 93.983 -78.231 52.908
104700 93.986 -78.241 52.889
104701 93.990 -78.251 52.870
104702 93.993 -78.261 52.851
104703 93.996 -78.272 52.833
104704 94.000 -78.282 52.814
104705 94.003 -78.292 52.794
104706 94.007 -78.303 52.772
104707 94.010 -78.313 52.751
104708 94.013 -78.323 52.732
104709 94.017 -78.333 52.717
104710 94.020 -78.344 52.703
104711 94.024 -78.354 52.687
104712 94.033 -78.357 52.670
104713 94.052 -78.351 52.652
104714 94.063 -78.348 52.633
104715 94.073 -78.345 52.615
104716 94.083 -78.342 52.598
104717 94.094 -78.340 52.582
104718 94.105 -78.337 52.565
104719 94.115 -78.335 52.546
104720 94.126 -78.332 52.527
104721 94.137 -78.328 52.508
104722 94.151 -78.320 52.487
104723 94.165 -78.313 52.465
104724 94.178 -78.305 52.445
104725 94.192 -78.297 52.422
104726 94.206 -78.290 52.398
104727 94.219 -78.282 52.373
104728 94.233 -78.275 52.346
104729 94.247 -78.267 52.318
104730 94.261 -78.260 52.288
104731 94.274 -78.252 52.256
104732 94.288 -78.244 52.222
104733 94.301 -78.237 52.187
104734 94.316 -78.229 52.149
104735 94.323 -78.224 52.109
104736 94.344 -78.214 52.068
104737 94.365 -78.203 52.027
104738 94.386 -78.194 51.992
104739 94.406 -78.184 51.958
104740 94.427 -78.174 51.927
104741 94.448 -78.163 51.900
104742 94.468 -78.153 51.868
104743 94.503 -78.168 51.833
104744 94.540 -78.184 51.794
104745 94.578 -78.200 51.753
104746 94.615 -78.216 51.712
104747 94.653 -78.232 51.671
104748 94.666 -78.232 51.630
104749 94.674 -78.223 51.592
104750 94.682 -78.215 51.556
104751 94.690 -78.206 51.520
104752 94.698 -78.197 51.484
104753 94.706 -78.188 51.448
104754 94.715 -78.179 51.413
104755 94.721 -78.173 51.378
104756 94.723 -78.179 51.343
104757 94.735 -78.176 51.309
104758 94.746 -78.174 51.275
104759 94.758 -78.171 51.240
104760 94.770 -78.172 51.205
104761 94.782 -78.177 51.170
104762 94.794 -78.182 51.134
104763 94.806 -78.187 51.099
104764 94.818 -78.192 51.064
104765 94.830 -78.197 51.028
104766 94.842 -78.202 50.993
104767 94.854 -78.206 50.956
104768 94.876 -78.213 50.919
104769 94.897 -78.207 50.880
104770 94.915 -78.191 50.843
104771 94.933 -78.173 50.806
104772 94.951 -78.156 50.769
104773 94.970 -78.139 50.731
104774 94.990 -78.118 50.693
104775 95.011 -78.097 50.656
104776 95.032 -78.076 50.618
104777 95.053 -78.055 50.582
104778 95.089 -78.044 50.547
104779 95.133 -78.037 50.512
104780 95.177 -78.032 50.478
104781 95.221 -78.026 50.442
104782 95.248 -78.028 50.407
104783 95.272 -78.029 50.373
104784 95.297 -78.031 50.339
104785 95.321 -78.033 50.306
104786 95.330 -78.033 50.275
104787 95.337 -78.031 50.242
104788 95.344 -78.031 50.211
104789 95.352 -78.030 50.180
104790 95.359 -78.029 50.151
104791 95.367 -78.028 50.123
104792 95.374 -78.027 50.095
104793 95.382 -78.026 50.067
104794 95.389 -78.024 50.039
104795 95.397 -78.019 50.010
104796 95.407 -78.025 49.983
104797 95.418 -78.037 49.956
104798 95.430 -78.049 49.931
104799 95.436 -78.054 49.908
104800 95.441 -78.059 49.885
104801 95.447 -78.064 49.865
104802 95.453 -78.069 49.845
104803 95.458 -78.075 49.823
104804 95.464 -78.079 49.800
104805 95.469 -78.084 49.774
104806 95.474 -78.089 49.746
104807 95.487 -78.085 49.718
104808 95.498 -78.078 49.687
104809 95.510 -78.071 49.656
104810 95.521 -78.063 49.623
104811 95.532 -78.056 49.589
104812 95.543 -78.049 49.552
104813 95.543 -78.054 49.514
104814 95.541 -78.079 49.475
104815 95.552 -78.085 49.433
104816 95.562 -78.091 49.391
104817 95.572 -78.097 49.350
104818 95.583 -78.103 49.309
104819 95.594 -78.109 49.268
104820 95.604 -78.114 49.226
104821 95.614 -78.120 49.184
104822 95.625 -78.126 49.142
104823 95.635 -78.131 49.100
104824 95.645 -78.138 49.058
104825 95.656 -78.144 49.016
104826 95.664 -78.128 48.974
104827 95.670 -78.126 48.932
104828 95.684 -78.109 48.890
104829 95.696 -78.088 48.848
104830 95.711 -78.089 48.808
104831 95.729 -78.116 48.768
104832 95.745 -78.131 48.728
104833 95.761 -78.147 48.688
104834 95.777 -78.163 48.649
104835 95.794 -78.178 48.610
104836 95.809 -78.194 48.571
104837 95.825 -78.210 48.535
104838 95.836 -78.216 48.499
104839 95.833 -78.213 48.463
104840 95.827 -78.216 48.426
104841 95.826 -78.216 48.389
104842 95.852 -78.201 48.351
104843 95.871 -78.190 48.313
104844 95.879 -78.183 48.275
104845 95.895 -78.181 48.237
104846 95.914 -78.184 48.199
104847 95.934 -78.186 48.161
104848 95.953 -78.188 48.124
104849 95.971 -78.191 48.086
104850 95.989 -78.195 48.049
104851 96.001 -78.202 48.013
104852 96.014 -78.210 47.978
104853 96.030 -78.217 47.943
104854 96.050 -78.225 47.909
104855 96.070 -78.234 47.876
104856 96.084 -78.250 47.843
104857 96.096 -78.267 47.810
104858 96.109 -78.285 47.778
104859 96.123 -78.297 47.747
104860 96.137 -78.305 47.715
104861 96.153 -78.313 47.684
104862 96.165 -78.316 47.654
104863 96.184 -78.335 47.623
104864 96.204 -78.354 47.593
104865 96.225 -78.373 47.564
104866 96.244 -78.391 47.536
104867 96.251 -78.403 47.508
104868 96.253 -78.411 47.480
104869 96.253 -78.423 47.453
104870 96.259 -78.432 47.425
104871 96.292 -78.437 47.398
104872 96.324 -78.442 47.373
104873 96.358 -78.445 47.347
104874 96.392 -78.449 47.321
104875 96.427 -78.454 47.293
104876 96.461 -78.457 47.265
104877 96.491 -78.457 47.236
104878 96.522 -78.456 47.208
104879 96.552 -78.455 47.181
104880 96.583 -78.454 47.154
104881 96.613 -78.454 47.128
104882 96.641 -78.454 47.102
104883 96.645 -78.462 47.075
104884 96.649 -78.471 47.048
104885 96.652 -78.480 47.022
104886 96.656 -78.489 46.998
104887 96.644 -78.492 46.973
104888 96.638 -78.498 46.949
104889 96.644 -78.507 46.925
104890 96.652 -78.516 46.901
104891 96.664 -78.522 46.876
104892 96.677 -78.528 46.850
104893 96.690 -78.533 46.826
104894 96.702 -78.539 46.803
104895 96.716 -78.545 46.779
104896 96.728 -78.551 46.754
104897 96.742 -78.570 46.731
104898 96.761 -78.601 46.708
104899 96.781 -78.633 46.686
104900 96.800 -78.664 46.666
104901 96.819 -78.695 46.645
104902 96.838 -78.726 46.625
104903 96.857 -78.757 46.603
104904 96.876 -78.788 46.581
104905 96.895 -78.819 46.560
104906 96.914 -78.850 46.540
104907 96.934 -78.881 46.520
104908 96.952 -78.912 46.501
104909 96.971 -78.943 46.480
104910 96.996 -78.964 46.460
104911 96.989 -78.888 46.440
104912 97.034 -78.937 46.419
104913 97.072 -78.952 46.398
104914 97.107 -78.949 46.378
104915 97.142 -78.946 46.357
104916 97.176 -78.943 46.336
104917 97.212 -78.940 46.315
104918 97.246 -78.937 46.294
104919 97.281 -78.933 46.272
104920 97.316 -78.930 46.251
104921 97.351 -78.927 46.232
104922 97.383 -78.924 46.212
104923 97.414 -78.921 46.193
104924 97.445 -78.919 46.176
104925 97.475 -78.917 46.157
104926 97.529 -78.915 46.138
104927 97.565 -78.903 46.120
104928 97.566 -78.904 46.103
104929 97.584 -78.901 46.086
104930 97.602 -78.897 46.068
104931 97.621 -78.894 46.051
104932 97.640 -78.891 46.035
104933 97.658 -78.888 46.018
104934 97.676 -78.884 45.999
104935 97.695 -78.881 45.981
104936 97.713 -78.878 45.964
104937 97.731 -78.875 45.948
104938 97.750 -78.872 45.933
104939 97.768 -78.868 45.918
104940 97.781 -78.858 45.902
104941 97.805 -78.844 45.886
104942 97.813 -78.827 45.870
104943 97.820 -78.811 45.855
104944 97.825 -78.792 45.839
104945 97.827 -78.771 45.822
104946 97.830 -78.749 45.806
104947 97.832 -78.728 45.790
104948 97.834 -78.707 45.774
104949 97.835 -78.710 45.757
104950 97.836 -78.744 45.742
104951 97.837 -78.779 45.729
104952 97.837 -78.813 45.717
104953 97.838 -78.848 45.704
104954 97.846 -78.852 45.691
104955 97.853 -78.858 45.678
104956 97.874 -78.865 45.666
104957 97.902 -78.874 45.653
104958 97.931 -78.883 45.640
104959 97.960 -78.892 45.627
104960 97.989 -78.901 45.613
104961 98.017 -78.910 45.598
104962 98.046 -78.919 45.581
104963 98.075 -78.928 45.563
104964 98.103 -78.937 45.546
104965 98.131 -78.945 45.528
104966 98.158 -78.952 45.511
104967 98.178 -78.962 45.495
104968 98.186 -78.978 45.477
104969 98.191 -78.989 45.459
104970 98.196 -79.000 45.444
104971 98.201 -79.012 45.429
104972 98.207 -79.022 45.415
104973 98.212 -79.034 45.401
104974 98.217 -79.044 45.388
104975 98.224 -79.056 45.376
104976 98.244 -79.073 45.364
104977 98.262 -79.068 45.353
104978 98.279 -79.062 45.341
104979 98.297 -79.056 45.330
104980 98.313 -79.049 45.319
104981 98.317 -79.023 45.308
104982 98.335 -79.014 45.297
104983 98.365 -79.018 45.286
104984 98.395 -79.022 45.275
104985 98.426 -79.026 45.262
104986 98.458 -79.033 45.249
104987 98.492 -79.041 45.237
104988 98.526 -79.050 45.224
104989 98.560 -79.059 45.212
104990 98.594 -79.068 45.199
104991 98.628 -79.077 45.189
104992 98.662 -79.085 45.177
104993 98.690 -79.093 45.166
104994 98.714 -79.100 45.155
104995 98.738 -79.106 45.143
104996 98.762 -79.112 45.132
104997 98.786 -79.118 45.121
104998 98.809 -79.125 45.110
104999 98.834 -79.132 45.100
105000 98.857 -79.138 45.090
105001 98.881 -79.145 45.078
105002 98.905 -79.151 45.066
105003 98.926 -79.157 45.054
105004 98.939 -79.161 45.043
105005 98.952 -79.165 45.032
105006 98.967 -79.170 45.021
105007 98.994 -79.192 45.010
105008 99.021 -79.215 44.999
105009 99.047 -79.216 44.988
105010 99.073 -79.217 44.977
105011 99.097 -79.215 44.966
105012 99.114 -79.201 44.956
105013 99.132 -79.187 44.945
105014 99.152 -79.187 44.935
105015 99.175 -79.205 44.925
105016 99.200 -79.222 44.915
105017 99.223 -79.240 44.907
105018 99.236 -79.255 44.899
105019 99.244 -79.270 44.890
105020 99.254 -79.285 44.883
105021 99.263 -79.300 44.876
105022 99.281 -79.310 44.870
105023 99.303 -79.318 44.865
105024 99.325 -79.326 44.860
105025 99.347 -79.335 44.855
105026 99.369 -79.343 44.849
105027 99.392 -79.357 44.845
105028 99.415 -79.372 44.841
105029 99.438 -79.387 44.838
105030 99.461 -79.403 44.835
105031 99.484 -79.418 44.832
105032 99.507 -79.433 44.829
105033 99.533 -79.446 44.826
105034 99.562 -79.450 44.823
105035 99.592 -79.454 44.820
105036 99.623 -79.457 44.817
105037 99.650 -79.461 44.815
105038 99.677 -79.466 44.812
105039 99.702 -79.470 44.810
105040 99.728 -79.475 44.807
105041 99.754 -79.480 44.804
105042 99.780 -79.485 44.800
105043 99.806 -79.489 44.796
105044 99.832 -79.494 44.792
105045 99.858 -79.498 44.788
105046 99.884 -79.503 44.784
105047 99.911 -79.513 44.779
105048 99.945 -79.545 44.775
105049 99.979 -79.578 44.769
105050 100.012 -79.626 44.762
105051 100.045 -79.680 44.753
105052 100.083 -79.695 44.744
105053 100.123 -79.705 44.734
105054 100.149 -79.730 44.725
105055 100.174 -79.758 44.713
105056 100.199 -79.784 44.706
105057 100.223 -79.811 44.701
105058 100.249 -79.838 44.694
105059 100.272 -79.874 44.688
105060 100.294 -79.912 44.680
105061 100.316 -79.951 44.669
105062 100.349 -79.966 44.658
105063 100.385 -79.971 44.647
105064 100.423 -79.976 44.635
105065 100.460 -79.981 44.622
105066 100.497 -79.987 44.609
105067 100.531 -80.012 44.594
105068 100.566 -80.040 44.578
105069 100.601 -80.067 44.561
105070 100.635 -80.094 44.545
105071 100.664 -80.125 44.529
105072 100.660 -80.170 44.513
105073 100.676 -80.208 44.498
105074 100.701 -80.243 44.482
105075 100.726 -80.277 44.466
105076 100.754 -80.313 44.450
105077 100.782 -80.350 44.433
105078 100.811 -80.386 44.416
105079 100.839 -80.422 44.400
105080 100.868 -80.459 44.383
105081 100.896 -80.496 44.366
105082 100.924 -80.532 44.347
105083 100.953 -80.569 44.330
105084 100.981 -80.605 44.313
105085 101.010 -80.641 44.295
105086 101.038 -80.678 44.276
105087 101.078 -80.700 44.256
105088 101.124 -80.713 44.236
105089 101.171 -80.726 44.216
105090 101.208 -80.741 44.198
105091 101.241 -80.757 44.180
105092 101.273 -80.773 44.162
105093 101.306 -80.789 44.145
105094 101.339 -80.804 44.128
105095 101.372 -80.820 44.112
105096 101.405 -80.837 44.095
105097 101.438 -80.853 44.079
105098 101.470 -80.869 44.062
105099 101.503 -80.885 44.046
105355 93.497 -78.460 53.844
105356 93.506 -78.458 53.828
105357 93.516 -78.455 53.810
105358 93.526 -78.453 53.791
105359 93.536 -78.451 53.771
105360 93.544 -78.448 53.750
105361 93.543 -78.438 53.729
105362 93.542 -78.429 53.706
105363 93.541 -78.420 53.682
105364 93.540 -78.410 53.658
105365 93.550 -78.395 53.633
105366 93.560 -78.379 53.608
105367 93.581 -78.364 53.583
105368 93.604 -78.349 53.559
105369 93.626 -78.335 53.535
105370 93.637 -78.326 53.512
105371 93.649 -78.318 53.489
105372 93.660 -78.310 53.467
105373 93.672 -78.301 53.445
105388 93.901 -78.191 53.133
105389 93.915 -78.184 53.111
105390 93.928 -78.180 53.089
105391 93.941 -78.176 53.068
105392 93.954 -78.172 53.046
105393 93.967 -78.167 53.025
105394 93.967 -78.166 53.004
105395 93.965 -78.166 52.984
105396 93.969 -78.176 52.964
105397 93.972 -78.186 52.945
105398 93.976 -78.197 52.924
105399 93.979 -78.206 52.904
105400 93.982 -78.217 52.883
105401 93.986 -78.228 52.863
105402 93.989 -78.238 52.842
105403 93.993 -78.248 52.823
105404 93.996 -78.258 52.802
105405 93.999 -78.269 52.781
105406 94.003 -78.279 52.760
105407 94.006 -78.289 52.739
105408 94.010 -78.300 52.717
105409 94.013 -78.310 52.697
105410 94.016 -78.320 52.678
105411 94.020 -78.331 52.660
105412 94.024 -78.341 52.641
105413 94.035 -78.341 52.622
105414 94.059 -78.331 52.604
105415 94.073 -78.327 52.586
105416 94.083 -78.325 52.568
105417 94.094 -78.322 52.550
105418 94.104 -78.319 52.532
105419 94.115 -78.316 52.512
105420 94.125 -78.313 52.492
105421 94.136 -78.311 52.475
105422 94.146 -78.308 52.455
105423 94.160 -78.300 52.434
105424 94.174 -78.293 52.412
105425 94.187 -78.285 52.393
105426 94.201 -78.278 52.371
105427 94.215 -78.270 52.348
105428 94.228 -78.263 52.321
105429 94.242 -78.255 52.294
105430 94.256 -78.247 52.265
105431 94.270 -78.240 52.233
105432 94.283 -78.232 52.200
105433 94.297 -78.225 52.165
105434 94.311 -78.218 52.130
105435 94.325 -78.210 52.093
105436 94.333 -78.205 52.057
105437 94.346 -78.197 52.020
105438 94.367 -78.187 51.984
105439 94.388 -78.177 51.953
105440 94.409 -78.167 51.922
105441 94.429 -78.157 51.892
105442 94.450 -78.147 51.865
105443 94.471 -78.137 51.833
105444 94.498 -78.144 51.799
105445 94.536 -78.159 51.762
105446 94.573 -78.175 51.724
105447 94.611 -78.191 51.686
105448 94.648 -78.208 51.648
105449 94.665 -78.211 51.610
105450 94.674 -78.205 51.574
105451 94.682 -78.196 51.539
105452 94.690 -78.187 51.505
105453 94.698 -78.178 51.471
105454 94.707 -78.169 51.436
105455 94.715 -78.160 51.402
105456 94.722 -78.153 51.367
105457 94.722 -78.161 51.333
105458 94.732 -78.159 51.300
105459 94.744 -78.156 51.266
105460 94.755 -78.154 51.233
105461 94.767 -78.155 51.198
105462 94.779 -78.159 51.163
105463 94.791 -78.165 51.127
105464 94.804 -78.169 51.093
105465 94.816 -78.175 51.058
105466 94.828 -78.179 51.023
105467 94.840 -78.184 50.987
105468 94.853 -78.189 50.951
105469 94.878 -78.197 50.914
105470 94.899 -78.189 50.876
105471 94.917 -78.172 50.840
105472 94.935 -78.155 50.803
105473 94.954 -78.138 50.766
105474 94.972 -78.120 50.729
105475 94.990 -78.103 50.692
105476 95.010 -78.083 50.655
105477 95.031 -78.062 50.618
105478 95.052 -78.041 50.583
105479 95.079 -78.024 50.548
105480 95.123 -78.018 50.513
105481 95.167 -78.012 50.479
105482 95.207 -78.009 50.445
105483 95.231 -78.010 50.411
105484 95.255 -78.012 50.377
105485 95.279 -78.014 50.343
105486 95.304 -78.016 50.311
105487 95.326 -78.018 50.279
105488 95.334 -78.017 50.248
105489 95.341 -78.016 50.216
105490 95.348 -78.015 50.186
105491 95.356 -78.014 50.158
105492 95.364 -78.013 50.131
105493 95.371 -78.012 50.104
105494 95.378 -78.012 50.077
105495 95.386 -78.008 50.049
105496 95.394 -78.003 50.022
105497 95.403 -78.006 49.995
105498 95.414 -78.018 49.971
105499 95.426 -78.030 49.947
105500 95.438 -78.042 49.927
105501 95.444 -78.047 49.908
105502 95.449 -78.053 49.891
105503 95.455 -78.057 49.872
105504 95.459 -78.062 49.851
105505 95.460 -78.065 49.826
105506 95.461 -78.067 49.798
105507 95.463 -78.069 49.767
105508 95.494 -78.065 49.735
105509 95.520 -78.060 49.701
105510 95.531 -78.053 49.666
105511 95.542 -78.045 49.631
105512 95.553 -78.038 49.595
105513 95.564 -78.031 49.555
105514 95.570 -78.025 49.515
105515 95.561 -78.055 49.475
105516 95.562 -78.075 49.433
105517 95.573 -78.081 49.390
105518 95.584 -78.087 49.348
105519 95.594 -78.092 49.307
105520 95.604 -78.098 49.265
105521 95.615 -78.104 49.224
105522 95.625 -78.109 49.182
105523 95.635 -78.116 49.141
105524 95.646 -78.122 49.100
105525 95.662 -78.102 49.059
105526 95.666 -78.094 49.018
105527 95.669 -78.106 48.977
105528 95.677 -78.105 48.935
105529 95.692 -78.091 48.893
105530 95.707 -78.078 48.852
105531 95.726 -78.072 48.812
105532 95.743 -78.090 48.771
105533 95.760 -78.106 48.731
105534 95.777 -78.124 48.691
105535 95.794 -78.141 48.652
105536 95.811 -78.158 48.613
105537 95.828 -78.175 48.574
105538 95.844 -78.192 48.539
105539 95.843 -78.213 48.504
105540 95.834 -78.228 48.469
105541 95.828 -78.232 48.433
105542 95.839 -78.225 48.396
105543 95.866 -78.210 48.358
105544 95.893 -78.194 48.320
105545 95.907 -78.184 48.283
105546 95.916 -78.177 48.245
105547 95.925 -78.170 48.208
105548 95.938 -78.167 48.172
105549 95.958 -78.169 48.136
105550 95.975 -78.173 48.099
105551 95.988 -78.180 48.062
105552 96.000 -78.188 48.025
105553 96.013 -78.194 47.990
105554 96.029 -78.202 47.957
105555 96.049 -78.210 47.923
105556 96.070 -78.219 47.891
105557 96.090 -78.227 47.859
105558 96.102 -78.244 47.826
105559 96.115 -78.262 47.795
105560 96.130 -78.269 47.763
105561 96.145 -78.277 47.731
105562 96.160 -78.285 47.701
105563 96.175 -78.294 47.670
105564 96.195 -78.313 47.639
105565 96.215 -78.332 47.609
105566 96.235 -78.351 47.581
105567 96.255 -78.370 47.554
105568 96.275 -78.389 47.526
105569 96.281 -78.401 47.499
105570 96.282 -78.412 47.472
105571 96.282 -78.423 47.445
105572 96.298 -78.431 47.417
105573 96.331 -78.435 47.392
105574 96.366 -78.439 47.366
105575 96.397 -78.440 47.340
105576 96.428 -78.439 47.312
105577 96.458 -78.438 47.283
105578 96.489 -78.438 47.255
105579 96.519 -78.437 47.229
105580 96.550 -78.436 47.202
105581 96.580 -78.435 47.176
105582 96.610 -78.435 47.151
105583 96.625 -78.439 47.124
105584 96.629 -78.448 47.098
105585 96.633 -78.457 47.072
105586 96.637 -78.466 47.045
105587 96.641 -78.475 47.021
105588 96.644 -78.484 46.996
105589 96.635 -78.488 46.971
105590 96.638 -78.496 46.947
105591 96.647 -78.504 46.923
105592 96.660 -78.510 46.899
105593 96.673 -78.516 46.873
105594 96.685 -78.522 46.847
105595 96.696 -78.530 46.823
105596 96.705 -78.539 46.799
105597 96.715 -78.549 46.776
105598 96.716 -78.566 46.751
105599 96.728 -78.595 46.727
105600 96.747 -78.626 46.704
105601 96.766 -78.657 46.682
105602 96.785 -78.688 46.661
105603 96.804 -78.719 46.641
105604 96.823 -78.750 46.619
105605 96.842 -78.781 46.596
105606 96.861 -78.812 46.574
105607 96.880 -78.843 46.553
105608 96.899 -78.874 46.533
105609 96.919 -78.905 46.513
105610 96.941 -78.931 46.492
105611 96.975 -78.943 46.469
105612 96.984 -78.866 46.448
105613 97.020 -78.911 46.426
105614 97.061 -78.939 46.404
105615 97.095 -78.936 46.384
105616 97.130 -78.933 46.362
105617 97.163 -78.930 46.341
105618 97.194 -78.927 46.319
105619 97.225 -78.924 46.298
105620 97.255 -78.922 46.277
105621 97.286 -78.919 46.257
105622 97.317 -78.917 46.237
105623 97.348 -78.915 46.218
105624 97.378 -78.912 46.199
105625 97.441 -78.917 46.181
105626 97.487 -78.910 46.163
105627 97.525 -78.900 46.144
105628 97.560 -78.889 46.126
105629 97.582 -78.881 46.109
105630 97.568 -78.886 46.092
105631 97.586 -78.883 46.076
105632 97.605 -78.880 46.061
105633 97.623 -78.876 46.047
105634 97.642 -78.873 46.030
105635 97.660 -78.870 46.011
105636 97.678 -78.867 45.993
105637 97.694 -78.860 45.975
105638 97.707 -78.851 45.958
105639 97.721 -78.841 45.941
105640 97.734 -78.832 45.925
105641 97.772 -78.818 45.909
105642 97.791 -78.801 45.894
105643 97.799 -78.785 45.878
105644 97.806 -78.768 45.862
105645 97.814 -78.752 45.846
105646 97.821 -78.735 45.830
105647 97.829 -78.719 45.814
105648 97.837 -78.702 45.798
105649 97.840 -78.683 45.782
105650 97.842 -78.672 45.765
105651 97.842 -78.706 45.751
105652 97.843 -78.741 45.738
105653 97.844 -78.776 45.727
105654 97.845 -78.810 45.714
105655 97.847 -78.844 45.702
105656 97.874 -78.850 45.689
105657 97.900 -78.857 45.677
105658 97.927 -78.864 45.664
105659 97.954 -78.870 45.651
105660 97.981 -78.877 45.638
105661 98.008 -78.883 45.624
105662 98.035 -78.890 45.608
105663 98.062 -78.897 45.591
105664 98.088 -78.903 45.571
105665 98.115 -78.910 45.553
105666 98.142 -78.917 45.535
105667 98.169 -78.924 45.518
105668 98.183 -78.939 45.502
105669 98.192 -78.957 45.484
105670 98.201 -78.975 45.467
105671 98.210 -78.993 45.452
105672 98.220 -79.012 45.437
105673 98.229 -79.029 45.424
105674 98.234 -79.041 45.411
105675 98.239 -79.052 45.398
105676 98.245 -79.063 45.385
105677 98.264 -79.068 45.373
105678 98.281 -79.062 45.362
105679 98.299 -79.056 45.350
105680 98.316 -79.050 45.339
105681 98.334 -79.044 45.327
105682 98.339 -79.021 45.317
105683 98.346 -78.999 45.305
105684 98.376 -79.003 45.293
105685 98.407 -79.008 45.284
105686 98.437 -79.012 45.271
105687 98.470 -79.019 45.258
105688 98.504 -79.028 45.245
105689 98.538 -79.037 45.231
105690 98.572 -79.046 45.218
105691 98.606 -79.055 45.207
105692 98.640 -79.063 45.196
105693 98.665 -79.070 45.184
105694 98.689 -79.076 45.172
105695 98.713 -79.083 45.161
105696 98.736 -79.090 45.150
105697 98.761 -79.096 45.139
105698 98.784 -79.103 45.128
105699 98.808 -79.109 45.118
105700 98.832 -79.115 45.107
105701 98.856 -79.122 45.095
105702 98.880 -79.128 45.084
105703 98.904 -79.135 45.073
105704 98.927 -79.141 45.060
105705 98.940 -79.145 45.049
105706 98.953 -79.150 45.038
105707 98.978 -79.168 45.026
105708 99.005 -79.183 45.014
105709 99.029 -79.184 45.002
105710 99.055 -79.185 44.991
105711 99.080 -79.187 44.981
105712 99.106 -79.188 44.970
105713 99.134 -79.190 44.959
105714 99.159 -79.188 44.949
105715 99.177 -79.175 44.938
105716 99.196 -79.170 44.928
105717 99.219 -79.188 44.919
105718 99.243 -79.206 44.911
105719 99.254 -79.221 44.903
105720 99.263 -79.235 44.894
105721 99.272 -79.250 44.887
105722 99.281 -79.265 44.881
105723 99.290 -79.280 44.875
105724 99.311 -79.289 44.870
105725 99.333 -79.297 44.866
105726 99.355 -79.306 44.862
105727 99.378 -79.322 44.858
105728 99.401 -79.338 44.853
105729 99.424 -79.353 44.850
105730 99.447 -79.369 44.847
105731 99.471 -79.384 44.845
105732 99.494 -79.399 44.844
105733 99.517 -79.415 44.841
105734 99.543 -79.427 44.839
105735 99.570 -79.431 44.835
105736 99.596 -79.435 44.833
105737 99.622 -79.440 44.831
105738 99.648 -79.444 44.828
105739 99.674 -79.450 44.825
105740 99.700 -79.454 44.821
105741 99.726 -79.459 44.817
105742 99.752 -79.463 44.814
105743 99.777 -79.468 44.810
105744 99.803 -79.473 44.806
105745 99.829 -79.477 44.803
105746 99.855 -79.482 44.799
105747 99.881 -79.486 44.796
105748 99.913 -79.511 44.792
105749 99.947 -79.544 44.786
105750 99.981 -79.576 44.780
105751 100.014 -79.623 44.772
105752 100.047 -79.677 44.764
105753 100.086 -79.689 44.755
105754 100.125 -79.698 44.745
105755 100.158 -79.715 44.735
105756 100.184 -79.742 44.722
105757 100.208 -79.769 44.712
105758 100.233 -79.795 44.704
105759 100.259 -79.823 44.697
105760 100.281 -79.861 44.690
105761 100.303 -79.899 44.682
105762 100.326 -79.938 44.672
105763 100.354 -79.961 44.660
105764 100.392 -79.967 44.648
105765 100.428 -79.971 44.636
105766 100.466 -79.977 44.623
105767 100.502 -79.981 44.609
105768 100.540 -79.987 44.594
105769 100.574 -80.012 44.578
105770 100.609 -80.040 44.562
105771 100.643 -80.067 44.545
105772 100.676 -80.096 44.529
105773 100.672 -80.140 44.515
105774 100.693 -80.177 44.499
105775 100.718 -80.211 44.483
105776 100.745 -80.247 44.467
105777 100.774 -80.284 44.451
105778 100.802 -80.320 44.434
105779 100.831 -80.356 44.418
105780 100.859 -80.393 44.401
105781 100.888 -80.429 44.385
105782 100.916 -80.466 44.368
105783 100.945 -80.502 44.350
105784 100.973 -80.538 44.333
105785 101.001 -80.575 44.316
105786 101.030 -80.611 44.298
105787 101.058 -80.648 44.280
105788 101.087 -80.685 44.261
105789 101.129 -80.703 44.241
105790 101.164 -80.718 44.222
105791 101.197 -80.735 44.204
105792 101.230 -80.751 44.186
105793 101.262 -80.767 44.168
105794 101.295 -80.782 44.151
105795 101.328 -80.798 44.134
105796 101.360 -80.815 44.118
105797 101.393 -80.831 44.101
105798 101.426 -80.847 44.084
105799 101.459 -80.863 44.067
105800 101.491 -80.879 44.052
106056 93.497 -78.449 53.840
106057 93.507 -78.447 53.824
106058 93.516 -78.445 53.806
106059 93.526 -78.442 53.786
106060 93.536 -78.439 53.766
106061 93.541 -78.435 53.745
106062 93.540 -78.425 53.723
106063 93.538 -78.416 53.700
106064 93.537 -78.407 53.676
106065 93.541 -78.395 53.651
106066 93.552 -78.379 53.626
106067 93.564 -78.363 53.601
106068 93.587 -78.348 53.577
106069 93.610 -78.334 53.552
106070 93.627 -78.322 53.529
106071 93.639 -78.313 53.505
106072 93.650 -78.305 53.482
106073 93.662 -78.297 53.459
106074 93.674 -78.288 53.437
106090 93.916 -78.173 53.098
106091 93.929 -78.169 53.076
106092 93.942 -78.165 53.054
106093 93.955 -78.160 53.032
106094 93.968 -78.156 53.010
106095 93.968 -78.154 52.988
106096 93.965 -78.153 52.967
106097 93.968 -78.163 52.946
106098 93.972 -78.173 52.926
106099 93.975 -78.183 52.904
106100 93.978 -78.194 52.881
106101 93.982 -78.204 52.860
106102 93.985 -78.214 52.838
106103 93.989 -78.225 52.817
106104 93.992 -78.235 52.796
106105 93.995 -78.245 52.775
106106 93.999 -78.255 52.753
106107 94.002 -78.266 52.732
106108 94.006 -78.276 52.709
106109 94.009 -78.286 52.688
106110 94.012 -78.297 52.666
106111 94.016 -78.307 52.646
106112 94.019 -78.317 52.625
106113 94.024 -78.327 52.605
106114 94.036 -78.323 52.585
106115 94.061 -78.314 52.565
106116 94.083 -78.307 52.545
106117 94.094 -78.304 52.525
106118 94.104 -78.301 52.505
106119 94.114 -78.298 52.485
106120 94.125 -78.295 52.464
106121 94.135 -78.293 52.443
106122 94.146 -78.290 52.423
106123 94.156 -78.288 52.401
106124 94.169 -78.282 52.380
106125 94.183 -78.274 52.358
106126 94.196 -78.266 52.337
106127 94.210 -78.259 52.315
106128 94.224 -78.251 52.292
106129 94.238 -78.244 52.263
106130 94.251 -78.236 52.234
106131 94.265 -78.228 52.204
106132 94.279 -78.221 52.173
106133 94.292 -78.213 52.142
106134 94.306 -78.206 52.111
106135 94.320 -78.198 52.078
106136 94.334 -78.191 52.045
106137 94.346 -78.185 52.012
106138 94.349 -78.181 51.980
106139 94.370 -78.170 51.947
106140 94.391 -78.160 51.918
106141 94.411 -78.150 51.888
106142 94.432 -78.141 51.858
106143 94.453 -78.130 51.827
106144 94.473 -78.120 51.796
106145 94.494 -78.119 51.764
106146 94.532 -78.134 51.731
106147 94.569 -78.151 51.698
106148 94.607 -78.167 51.663
106149 94.644 -78.183 51.628
106150 94.664 -78.188 51.593
106151 94.674 -78.186 51.558
106152 94.682 -78.178 51.524
106153 94.690 -78.169 51.491
106154 94.698 -78.159 51.459
106155 94.707 -78.151 51.426
106156 94.715 -78.142 51.391
106157 94.722 -78.134 51.358
106158 94.722 -78.141 51.324
106159 94.730 -78.142 51.291
106160 94.741 -78.140 51.259
106161 94.753 -78.137 51.226
106162 94.765 -78.138 51.192
106163 94.777 -78.143 51.157
106164 94.789 -78.147 51.122
106165 94.801 -78.152 51.088
106166 94.813 -78.157 51.054
106167 94.826 -78.162 51.019
106168 94.838 -78.167 50.983
106169 94.856 -78.173 50.947
106170 94.881 -78.180 50.910
106171 94.901 -78.171 50.873
106172 94.919 -78.153 50.837
106173 94.937 -78.136 50.801
106174 94.956 -78.119 50.764
106175 94.974 -78.102 50.728
106176 94.992 -78.084 50.691
106177 95.011 -78.068 50.655
106178 95.030 -78.048 50.619
106179 95.051 -78.027 50.584
106180 95.072 -78.006 50.549
106181 95.113 -77.999 50.515
106182 95.157 -77.993 50.481
106183 95.190 -77.991 50.447
106184 95.214 -77.993 50.414
106185 95.239 -77.995 50.380
106186 95.263 -77.997 50.348
106187 95.287 -78.000 50.315
106188 95.312 -78.001 50.284
106189 95.330 -78.003 50.253
106190 95.338 -78.001 50.223
106191 95.345 -78.000 50.193
106192 95.352 -78.000 50.166
106193 95.360 -77.999 50.140
106194 95.368 -77.998 50.113
106195 95.375 -77.996 50.085
106196 95.383 -77.991 50.059
106197 95.390 -77.987 50.032
106198 95.401 -77.994 50.007
106199 95.412 -78.006 49.985
106200 95.423 -78.016 49.966
106201 95.435 -78.027 49.949
106202 95.443 -78.034 49.932
106203 95.444 -78.037 49.913
106204 95.445 -78.040 49.891
106205 95.446 -78.042 49.867
106206 95.447 -78.045 49.840
106207 95.447 -78.047 49.810
106208 95.460 -78.047 49.777
106209 95.492 -78.044 49.743
106210 95.524 -78.039 49.707
106211 95.552 -78.034 49.670
106212 95.563 -78.027 49.633
106213 95.574 -78.020 49.596
106214 95.586 -78.012 49.555
106215 95.597 -78.005 49.515
106216 95.588 -78.019 49.474
106217 95.579 -78.056 49.432
106218 95.584 -78.070 49.390
106219 95.594 -78.076 49.348
106220 95.605 -78.082 49.306
106221 95.615 -78.087 49.265
106222 95.625 -78.094 49.224
106223 95.636 -78.100 49.183
106224 95.649 -78.094 49.142
106225 95.664 -78.066 49.102
106226 95.676 -78.058 49.062
106227 95.673 -78.078 49.021
106228 95.669 -78.097 48.979
106229 95.684 -78.084 48.938
106230 95.699 -78.071 48.896
106231 95.714 -78.057 48.855
106232 95.741 -78.066 48.816
106233 95.766 -78.077 48.775
106234 95.783 -78.094 48.734
106235 95.800 -78.111 48.694
106236 95.817 -78.128 48.654
106237 95.834 -78.145 48.615
106238 95.851 -78.162 48.577
106239 95.860 -78.181 48.542
106240 95.848 -78.205 48.508
106241 95.838 -78.228 48.474
106242 95.829 -78.246 48.440
106243 95.852 -78.232 48.404
106244 95.879 -78.217 48.366
106245 95.906 -78.202 48.328
106246 95.933 -78.187 48.291
106247 95.944 -78.178 48.255
106248 95.953 -78.172 48.218
106249 95.962 -78.164 48.183
106250 95.969 -78.158 48.147
106251 95.974 -78.158 48.110
106252 95.986 -78.166 48.074
106253 95.999 -78.172 48.038
106254 96.011 -78.180 48.004
106255 96.028 -78.188 47.970
106256 96.048 -78.196 47.937
106257 96.068 -78.203 47.905
106258 96.089 -78.212 47.873
106259 96.108 -78.222 47.841
106260 96.122 -78.234 47.811
106261 96.137 -78.242 47.780
106262 96.153 -78.251 47.748
106263 96.168 -78.261 47.719
106264 96.183 -78.272 47.688
106265 96.201 -78.289 47.657
106266 96.222 -78.308 47.627
106267 96.242 -78.327 47.599
106268 96.262 -78.347 47.572
106269 96.283 -78.366 47.545
106270 96.296 -78.382 47.518
106271 96.300 -78.395 47.491
106272 96.303 -78.408 47.464
106273 96.307 -78.421 47.439
106274 96.334 -78.423 47.412
106275 96.364 -78.423 47.386
106276 96.395 -78.422 47.360
106277 96.425 -78.421 47.332
106278 96.456 -78.420 47.305
106279 96.486 -78.420 47.277
106280 96.517 -78.419 47.250
106281 96.547 -78.418 47.224
106282 96.577 -78.417 47.199
106283 96.606 -78.416 47.173
106284 96.610 -78.426 47.148
106285 96.614 -78.435 47.122
106286 96.617 -78.443 47.095
106287 96.621 -78.452 47.069
106288 96.625 -78.461 47.044
106289 96.628 -78.470 47.020
106290 96.632 -78.479 46.996
106291 96.631 -78.485 46.973
106292 96.643 -78.492 46.947
106293 96.654 -78.501 46.920
106294 96.663 -78.510 46.893
106295 96.673 -78.520 46.868
106296 96.682 -78.529 46.843
106297 96.692 -78.539 46.820
106298 96.701 -78.548 46.796
106299 96.700 -78.566 46.772
106300 96.695 -78.588 46.748
106301 96.713 -78.618 46.724
106302 96.732 -78.649 46.701
106303 96.751 -78.680 46.679
106304 96.770 -78.711 46.657
106305 96.789 -78.742 46.636
106306 96.808 -78.774 46.612
106307 96.827 -78.805 46.589
106308 96.847 -78.836 46.568
106309 96.866 -78.867 46.546
106310 96.887 -78.895 46.525
106311 96.923 -78.905 46.503
106312 96.955 -78.916 46.480
106313 97.003 -78.909 46.457
106314 97.019 -78.892 46.434
106315 97.039 -78.926 46.411
106316 97.067 -78.925 46.390
106317 97.097 -78.922 46.369
106318 97.128 -78.920 46.347
106319 97.159 -78.917 46.326
106320 97.189 -78.915 46.304
106321 97.220 -78.912 46.283
106322 97.251 -78.909 46.263
106323 97.282 -78.907 46.243
106324 97.329 -78.908 46.224
106325 97.400 -78.914 46.205
106326 97.443 -78.906 46.186
106327 97.485 -78.897 46.168
106328 97.520 -78.886 46.150
106329 97.555 -78.874 46.131
106330 97.589 -78.863 46.114
106331 97.588 -78.862 46.098
106332 97.570 -78.868 46.082
106333 97.588 -78.865 46.068
106334 97.607 -78.861 46.054
106335 97.620 -78.852 46.038
106336 97.634 -78.843 46.020
106337 97.647 -78.833 46.003
106338 97.661 -78.824 45.985
106339 97.674 -78.815 45.967
106340 97.687 -78.805 45.949
106341 97.711 -78.794 45.933
106342 97.769 -78.775 45.917
106343 97.777 -78.758 45.901
106344 97.786 -78.743 45.885
106345 97.802 -78.735 45.869
106346 97.818 -78.726 45.853
106347 97.834 -78.717 45.837
106348 97.850 -78.708 45.821
106349 97.866 -78.699 45.805
106350 97.852 -78.684 45.788
106351 97.849 -78.642 45.772
106352 97.861 -78.652 45.758
106353 97.878 -78.661 45.746
106354 97.895 -78.668 45.735
106355 97.895 -78.710 45.722
106356 97.868 -78.813 45.710
106357 97.877 -78.833 45.699
106358 97.907 -78.836 45.687
106359 97.936 -78.839 45.675
106360 97.965 -78.843 45.662
106361 97.992 -78.849 45.648
106362 98.019 -78.855 45.634
106363 98.046 -78.862 45.619
106364 98.073 -78.869 45.602
106365 98.099 -78.875 45.584
106366 98.127 -78.882 45.566
106367 98.153 -78.889 45.547
106368 98.178 -78.896 45.527
106369 98.187 -78.915 45.509
106370 98.197 -78.933 45.491
106371 98.206 -78.951 45.475
106372 98.215 -78.969 45.461
106373 98.224 -78.987 45.447
106374 98.234 -79.006 45.435
106375 98.243 -79.024 45.422
106376 98.251 -79.040 45.408
106377 98.260 -79.058 45.395
106378 98.283 -79.062 45.383
106379 98.301 -79.056 45.372
106380 98.318 -79.050 45.361
106381 98.336 -79.044 45.349
106382 98.353 -79.038 45.337
106383 98.361 -79.018 45.327
106384 98.364 -78.992 45.315
106385 98.387 -78.989 45.304
106386 98.418 -78.993 45.293
106387 98.448 -78.997 45.280
106388 98.482 -79.006 45.266
106389 98.516 -79.015 45.253
106390 98.550 -79.024 45.240
106391 98.584 -79.032 45.227
106392 98.616 -79.040 45.215
106393 98.640 -79.047 45.204
106394 98.663 -79.053 45.193
106395 98.688 -79.060 45.182
106396 98.711 -79.067 45.171
106397 98.735 -79.073 45.161
106398 98.759 -79.080 45.150
106399 98.783 -79.086 45.139
106400 98.807 -79.093 45.128
106401 98.831 -79.100 45.116
106402 98.855 -79.106 45.104
106403 98.879 -79.112 45.092
106404 98.903 -79.119 45.080
106405 98.927 -79.125 45.068
106406 98.941 -79.130 45.056
106407 98.962 -79.143 45.045
106408 98.988 -79.152 45.032
106409 99.013 -79.153 45.020
106410 99.038 -79.154 45.008
106411 99.063 -79.155 44.997
106412 99.088 -79.156 44.986
106413 99.114 -79.158 44.976
106414 99.141 -79.159 44.964
106415 99.168 -79.161 44.954
106416 99.195 -79.162 44.943
106417 99.221 -79.162 44.933
106418 99.239 -79.154 44.924
106419 99.263 -79.172 44.916
106420 99.273 -79.186 44.907
106421 99.282 -79.201 44.900
106422 99.291 -79.216 44.894
106423 99.300 -79.231 44.889
106424 99.309 -79.245 44.884
106425 99.319 -79.259 44.880
106426 99.341 -79.272 44.877
106427 99.364 -79.288 44.874
106428 99.386 -79.301 44.870
106429 99.408 -79.315 44.865
106430 99.430 -79.328 44.862
106431 99.452 -79.341 44.859
106432 99.474 -79.355 44.856
106433 99.496 -79.368 44.854
106434 99.519 -79.383 44.852
106435 99.542 -79.408 44.849
106436 99.568 -79.415 44.847
106437 99.593 -79.419 44.844
106438 99.620 -79.424 44.842
106439 99.645 -79.429 44.839
106440 99.671 -79.433 44.835
106441 99.697 -79.438 44.831
106442 99.723 -79.442 44.826
106443 99.749 -79.447 44.823
106444 99.775 -79.451 44.820
106445 99.801 -79.456 44.816
106446 99.827 -79.461 44.813
106447 99.853 -79.466 44.810
106448 99.880 -79.477 44.806
106449 99.915 -79.510 44.802
106450 99.948 -79.542 44.795
106451 99.983 -79.576 44.788
106452 100.016 -79.620 44.780
106453 100.049 -79.673 44.773
106454 100.088 -79.683 44.764
106455 100.127 -79.692 44.754
106456 100.167 -79.701 44.744
106457 100.194 -79.726 44.730
106458 100.218 -79.753 44.718
106459 100.244 -79.780 44.708
106460 100.268 -79.809 44.700
106461 100.290 -79.848 44.693
106462 100.313 -79.886 44.684
106463 100.335 -79.926 44.673
106464 100.360 -79.957 44.660
106465 100.397 -79.961 44.648
106466 100.434 -79.967 44.636
106467 100.471 -79.971 44.623
106468 100.508 -79.977 44.608
106469 100.545 -79.981 44.593
106470 100.582 -79.987 44.578
106471 100.617 -80.012 44.561
106472 100.652 -80.039 44.545
106473 100.686 -80.067 44.529
106474 100.686 -80.110 44.515
106475 100.710 -80.145 44.500
106476 100.737 -80.181 44.484
106477 100.765 -80.217 44.468
106478 100.793 -80.253 44.451
106479 100.822 -80.290 44.435
106480 100.850 -80.326 44.419
106481 100.879 -80.363 44.402
106482 100.907 -80.399 44.387
106483 100.936 -80.435 44.370
106484 100.964 -80.472 44.353
106485 100.993 -80.509 44.336
106486 101.021 -80.545 44.319
106487 101.050 -80.582 44.301
106488 101.081 -80.616 44.284
106489 101.115 -80.649 44.265
106490 101.149 -80.676 44.246
106491 101.182 -80.692 44.227
106492 101.215 -80.707 44.210
106493 101.248 -80.723 44.193
106494 101.281 -80.739 44.175
106495 101.314 -80.754 44.158
106496 101.347 -80.770 44.140
106497 101.380 -80.786 44.123
106498 101.413 -80.802 44.107
106499 101.446 -80.818 44.090
106500 101.479 -80.834 44.074
106501 101.512 -80.850 44.058
106757 93.497 -78.438 53.835
106758 93.507 -78.436 53.819
106759 93.516 -78.434 53.800
106760 93.526 -78.431 53.781
106761 93.536 -78.429 53.760
106762 93.538 -78.421 53.738
106763 93.536 -78.412 53.716
106764 93.535 -78.403 53.693
106765 93.534 -78.394 53.669
106766 93.543 -78.379 53.644
106767 93.553 -78.363 53.618
106768 93.570 -78.348 53.594
106769 93.592 -78.333 53.569
106770 93.615 -78.318 53.545
106771 93.628 -78.309 53.521
106772 93.640 -78.300 53.498
106773 93.651 -78.292 53.474
106774 93.663 -78.284 53.451
106775 93.675 -78.275 53.429
106791 93.917 -78.162 53.085
106792 93.929 -78.158 53.063
106793 93.942 -78.153 53.040
106794 93.955 -78.149 53.018
106795 93.968 -78.145 52.995
106796 93.968 -78.144 52.973
106797 93.966 -78.142 52.951
106798 93.968 -78.150 52.929
106799 93.971 -78.159 52.907
106800 93.974 -78.170 52.884
106801 93.978 -78.180 52.860
106802 93.981 -78.191 52.838
106803 93.985 -78.201 52.817
106804 93.988 -78.211 52.794
106805 93.991 -78.222 52.772
106806 93.995 -78.232 52.750
106807 93.998 -78.242 52.728
106808 94.002 -78.252 52.705
106809 94.005 -78.263 52.683
106810 94.008 -78.273 52.661
106811 94.012 -78.283 52.639
106812 94.015 -78.294 52.616
106813 94.019 -78.304 52.595
106814 94.028 -78.312 52.573
106815 94.037 -78.307 52.552
106816 94.062 -78.297 52.530
106817 94.087 -78.288 52.508
106818 94.104 -78.283 52.486
106819 94.114 -78.280 52.466
106820 94.125 -78.278 52.445
106821 94.135 -78.275 52.422
106822 94.145 -78.272 52.400
106823 94.156 -78.269 52.378
106824 94.167 -78.266 52.356
106825 94.178 -78.263 52.333
106826 94.192 -78.255 52.311
106827 94.205 -78.247 52.288
106828 94.219 -78.240 52.266
106829 94.233 -78.232 52.241
106830 94.247 -78.225 52.214
106831 94.260 -78.217 52.185
106832 94.274 -78.210 52.156
106833 94.288 -78.202 52.128
106834 94.301 -78.194 52.099
106835 94.315 -78.187 52.069
106836 94.329 -78.179 52.039
106837 94.343 -78.172 52.008
106838 94.357 -78.166 51.979
106839 94.359 -78.162 51.949
106840 94.372 -78.153 51.919
106841 94.393 -78.144 51.891
106842 94.414 -78.134 51.861
106843 94.434 -78.124 51.831
106844 94.455 -78.113 51.802
106845 94.476 -78.103 51.771
106846 94.494 -78.097 51.740
106847 94.527 -78.110 51.709
106848 94.565 -78.126 51.677
106849 94.602 -78.143 51.645
106850 94.640 -78.159 51.611
106851 94.662 -78.166 51.578
106852 94.674 -78.166 51.545
106853 94.682 -78.159 51.512
106854 94.690 -78.150 51.479
106855 94.699 -78.141 51.448
106856 94.707 -78.132 51.415
106857 94.715 -78.123 51.382
106858 94.723 -78.115 51.349
106859 94.723 -78.122 51.316
106860 94.728 -78.125 51.284
106861 94.739 -78.122 51.253
106862 94.751 -78.120 51.220
106863 94.762 -78.120 51.187
106864 94.774 -78.125 51.153
106865 94.786 -78.130 51.118
106866 94.799 -78.134 51.084
106867 94.811 -78.140 51.050
106868 94.823 -78.144 51.016
106869 94.835 -78.150 50.980
106870 94.859 -78.156 50.945
106871 94.884 -78.164 50.908
106872 94.903 -78.153 50.871
106873 94.921 -78.135 50.835
106874 94.940 -78.118 50.799
106875 94.958 -78.100 50.763
106876 94.976 -78.084 50.728
106877 94.994 -78.066 50.691
106878 95.013 -78.049 50.655
106879 95.031 -78.032 50.620
106880 95.050 -78.013 50.585
106881 95.070 -77.993 50.551
106882 95.103 -77.979 50.517
106883 95.147 -77.973 50.483
106884 95.173 -77.975 50.450
106885 95.197 -77.977 50.416
106886 95.222 -77.978 50.384
106887 95.246 -77.981 50.352
106888 95.270 -77.982 50.320
106889 95.295 -77.984 50.289
106890 95.319 -77.987 50.259
106891 95.334 -77.987 50.229
106892 95.342 -77.986 50.201
106893 95.349 -77.985 50.173
106894 95.356 -77.984 50.148
106895 95.364 -77.983 50.121
106896 95.372 -77.980 50.094
106897 95.379 -77.975 50.068
106898 95.390 -77.981 50.043
106899 95.401 -77.993 50.020
106900 95.412 -78.003 50.000
106901 95.422 -78.012 49.985
106902 95.424 -78.014 49.969
106903 95.427 -78.016 49.952
106904 95.429 -78.018 49.931
106905 95.431 -78.020 49.906
106906 95.432 -78.023 49.880
106907 95.433 -78.025 49.851
106908 95.434 -78.028 49.819
106909 95.458 -78.025 49.784
106910 95.490 -78.022 49.748
106911 95.522 -78.017 49.710
106912 95.553 -78.013 49.671
106913 95.585 -78.009 49.632
106914 95.596 -78.002 49.593
106915 95.607 -77.994 49.553
106916 95.618 -77.987 49.512
106917 95.614 -77.986 49.471
106918 95.606 -78.020 49.429
106919 95.597 -78.056 49.388
106920 95.605 -78.065 49.347
106921 95.616 -78.072 49.305
106922 95.626 -78.078 49.264
106923 95.636 -78.083 49.224
106924 95.654 -78.053 49.184
106925 95.665 -78.032 49.145
106926 95.680 -78.030 49.105
106927 95.676 -78.050 49.064
106928 95.672 -78.069 49.023
106929 95.675 -78.078 48.981
106930 95.690 -78.064 48.940
106931 95.706 -78.050 48.899
106932 95.729 -78.052 48.858
106933 95.756 -78.061 48.819
106934 95.782 -78.070 48.780
106935 95.807 -78.081 48.739
106936 95.824 -78.098 48.698
106937 95.841 -78.115 48.658
106938 95.858 -78.132 48.618
106939 95.875 -78.149 48.580
106940 95.865 -78.173 48.543
106941 95.854 -78.197 48.509
106942 95.844 -78.219 48.475
106943 95.841 -78.228 48.441
106944 95.866 -78.223 48.406
106945 95.893 -78.216 48.370
106946 95.919 -78.210 48.334
106947 95.946 -78.194 48.299
106948 95.972 -78.179 48.264
106949 95.980 -78.172 48.227
106950 95.985 -78.167 48.192
106951 95.987 -78.163 48.155
106952 95.988 -78.159 48.118
106953 95.990 -78.155 48.082
106954 95.998 -78.158 48.049
106955 96.010 -78.165 48.017
106956 96.028 -78.172 47.984
106957 96.048 -78.181 47.952
106958 96.068 -78.189 47.920
106959 96.091 -78.197 47.888
106960 96.115 -78.204 47.856
106961 96.131 -78.211 47.826
106962 96.147 -78.220 47.795
106963 96.162 -78.229 47.764
106964 96.178 -78.239 47.734
106965 96.193 -78.252 47.704
106966 96.207 -78.265 47.673
106967 96.227 -78.283 47.644
106968 96.247 -78.302 47.617
106969 96.267 -78.322 47.590
106970 96.288 -78.341 47.564
106971 96.294 -78.354 47.537
106972 96.297 -78.367 47.510
106973 96.304 -78.381 47.484
106974 96.318 -78.395 47.459
106975 96.333 -78.404 47.432
106976 96.362 -78.404 47.405
106977 96.392 -78.403 47.379
106978 96.422 -78.402 47.353
106979 96.453 -78.401 47.326
106980 96.483 -78.401 47.299
106981 96.514 -78.400 47.272
106982 96.544 -78.399 47.247
106983 96.575 -78.398 47.222
106984 96.591 -78.403 47.197
106985 96.594 -78.412 47.172
106986 96.598 -78.420 47.146
106987 96.602 -78.429 47.120
106988 96.605 -78.438 47.094
106989 96.609 -78.447 47.069
106990 96.613 -78.456 47.045
106991 96.617 -78.464 47.021
106992 96.624 -78.470 46.996
106993 96.630 -78.490 46.968
106994 96.640 -78.500 46.941
106995 96.650 -78.509 46.914
106996 96.659 -78.519 46.889
106997 96.668 -78.528 46.865
106998 96.678 -78.538 46.841
106999 96.687 -78.547 46.817
107000 96.685 -78.566 46.793
107001 96.680 -78.588 46.769
107002 96.679 -78.611 46.744
107003 96.698 -78.642 46.722
107004 96.717 -78.673 46.698
107005 96.736 -78.705 46.676
107006 96.755 -78.736 46.654
107007 96.774 -78.767 46.630
107008 96.793 -78.798 46.606
107009 96.812 -78.829 46.583
107010 96.833 -78.858 46.560
107011 96.868 -78.868 46.538
107012 96.903 -78.879 46.515
107013 96.935 -78.890 46.492
107014 96.972 -78.894 46.467
107015 97.019 -78.872 46.442
107016 97.026 -78.899 46.419
107017 97.032 -78.915 46.397
107018 97.062 -78.912 46.375
107019 97.093 -78.909 46.354
107020 97.123 -78.907 46.332
107021 97.154 -78.905 46.310
107022 97.185 -78.902 46.289
107023 97.218 -78.899 46.269
107024 97.299 -78.910 46.248
107025 97.357 -78.909 46.229
107026 97.400 -78.902 46.209
107027 97.443 -78.893 46.190
107028 97.480 -78.883 46.172
107029 97.515 -78.871 46.154
107030 97.549 -78.860 46.137
107031 97.584 -78.849 46.120
107032 97.618 -78.837 46.104
107033 97.594 -78.843 46.088
107034 97.583 -78.858 46.074
107035 97.597 -78.850 46.059
107036 97.611 -78.841 46.043
107037 97.625 -78.832 46.026
107038 97.639 -78.824 46.010
107039 97.652 -78.814 45.993
107040 97.666 -78.805 45.976
107041 97.681 -78.798 45.958
107042 97.742 -78.796 45.942
107043 97.786 -78.762 45.926
107044 97.802 -78.752 45.909
107045 97.818 -78.742 45.891
107046 97.833 -78.734 45.875
107047 97.850 -78.725 45.859
107048 97.866 -78.716 45.842
107049 97.881 -78.707 45.827
107050 97.891 -78.697 45.811
107051 97.875 -78.681 45.795
107052 97.863 -78.657 45.779
107053 97.844 -78.682 45.764
107054 97.856 -78.692 45.751
107055 97.874 -78.682 45.739
107056 97.919 -78.675 45.726
107057 97.891 -78.778 45.714
107058 97.885 -78.817 45.702
107059 97.908 -78.824 45.691
107060 97.937 -78.827 45.679
107061 97.966 -78.830 45.666
107062 97.995 -78.833 45.652
107063 98.025 -78.836 45.639
107064 98.054 -78.839 45.624
107065 98.083 -78.842 45.607
107066 98.110 -78.847 45.590
107067 98.138 -78.854 45.572
107068 98.164 -78.861 45.554
107069 98.183 -78.873 45.536
107070 98.193 -78.891 45.519
107071 98.202 -78.909 45.505
107072 98.211 -78.927 45.491
107073 98.218 -78.942 45.477
107074 98.226 -78.957 45.462
107075 98.233 -78.972 45.448
107076 98.241 -78.987 45.434
107077 98.248 -79.002 45.422
107078 98.265 -79.039 45.409
107079 98.290 -79.053 45.397
107080 98.319 -79.049 45.384
107081 98.338 -79.044 45.371
107082 98.355 -79.038 45.359
107083 98.373 -79.033 45.346
107084 98.382 -79.015 45.335
107085 98.386 -78.990 45.324
107086 98.399 -78.974 45.312
107087 98.429 -78.979 45.301
107088 98.460 -78.984 45.288
107089 98.494 -78.993 45.274
107090 98.528 -79.002 45.261
107091 98.563 -79.010 45.249
107092 98.590 -79.018 45.236
107093 98.615 -79.024 45.224
107094 98.638 -79.031 45.213
107095 98.662 -79.037 45.203
107096 98.686 -79.044 45.193
107097 98.710 -79.050 45.184
107098 98.734 -79.057 45.173
107099 98.758 -79.063 45.162
107100 98.782 -79.070 45.151
107101 98.806 -79.077 45.138
107102 98.830 -79.083 45.127
107103 98.854 -79.090 45.115
107104 98.878 -79.096 45.104
107105 98.902 -79.103 45.090
107106 98.926 -79.109 45.077
107107 98.946 -79.118 45.064
107108 98.971 -79.120 45.052
107109 98.996 -79.122 45.040
107110 99.021 -79.122 45.028
107111 99.046 -79.124 45.016
107112 99.071 -79.125 45.004
107113 99.096 -79.126 44.992
107114 99.121 -79.127 44.981
107115 99.147 -79.128 44.970
107116 99.174 -79.130 44.960
107117 99.202 -79.131 44.950
107118 99.229 -79.133 44.941
107119 99.256 -79.134 44.932
107120 99.282 -79.137 44.924
107121 99.288 -79.155 44.915
107122 99.294 -79.172 44.907
107123 99.300 -79.190 44.902
107124 99.307 -79.207 44.898
107125 99.313 -79.225 44.893
107126 99.316 -79.234 44.889
107127 99.336 -79.248 44.885
107128 99.361 -79.248 44.881
107129 99.383 -79.262 44.877
107130 99.405 -79.275 44.873
107131 99.427 -79.288 44.870
107132 99.449 -79.302 44.868
107133 99.471 -79.315 44.864
107134 99.494 -79.331 44.863
107135 99.518 -79.353 44.860
107136 99.541 -79.379 44.858
107137 99.565 -79.398 44.855
107138 99.591 -79.403 44.853
107139 99.617 -79.407 44.850
107140 99.643 -79.412 44.847
107141 99.668 -79.416 44.843
107142 99.695 -79.421 44.839
107143 99.720 -79.426 44.835
107144 99.747 -79.431 44.832
107145 99.772 -79.435 44.829
107146 99.798 -79.440 44.826
107147 99.824 -79.444 44.823
107148 99.850 -79.449 44.820
107149 99.883 -79.476 44.815
107150 99.916 -79.508 44.809
107151 99.950 -79.541 44.801
107152 99.984 -79.574 44.792
107153 100.018 -79.617 44.784
107154 100.052 -79.667 44.776
107155 100.090 -79.676 44.766
107156 100.130 -79.686 44.757
107157 100.168 -79.695 44.746
107158 100.203 -79.711 44.734
107159 100.228 -79.737 44.721
107160 100.253 -79.764 44.710
107161 100.277 -79.796 44.701
107162 100.299 -79.836 44.694
107163 100.322 -79.874 44.684
107164 100.344 -79.913 44.672
107165 100.366 -79.952 44.658
107166 100.403 -79.957 44.646
107167 100.440 -79.961 44.635
107168 100.477 -79.967 44.622
107169 100.514 -79.971 44.607
107170 100.551 -79.977 44.592
107171 100.588 -79.982 44.577
107172 100.625 -79.987 44.560
107173 100.660 -80.012 44.544
107174 100.695 -80.039 44.529
107175 100.703 -80.079 44.515
107176 100.728 -80.114 44.500
107177 100.757 -80.150 44.484
107178 100.785 -80.187 44.468
107179 100.814 -80.223 44.451
107180 100.842 -80.259 44.435
107181 100.870 -80.297 44.419
107182 100.899 -80.333 44.403
107183 100.927 -80.369 44.388
107184 100.956 -80.406 44.373
107185 100.989 -80.439 44.356
107186 101.024 -80.472 44.340
107187 101.057 -80.505 44.323
107188 101.092 -80.538 44.305
107189 101.125 -80.572 44.288
107190 101.159 -80.605 44.269
107191 101.193 -80.629 44.250
107192 101.226 -80.645 44.233
107193 101.259 -80.661 44.216
107194 101.292 -80.676 44.199
107195 101.325 -80.692 44.181
107196 101.358 -80.708 44.163
107197 101.391 -80.724 44.146
107198 101.424 -80.740 44.129
107199 101.457 -80.756 44.112
107200 101.490 -80.772 44.096
107201 101.523 -80.788 44.080
107202 101.556 -80.804 44.064
107458 93.497 -78.428 53.827
107459 93.507 -78.425 53.811
107460 93.517 -78.423 53.792
107461 93.526 -78.420 53.773
107462 93.536 -78.417 53.752
107463 93.534 -78.408 53.731
107464 93.533 -78.399 53.709
107465 93.532 -78.390 53.685
107466 93.534 -78.379 53.661
107467 93.544 -78.363 53.636
107468 93.554 -78.347 53.611
107469 93.575 -78.332 53.586
107470 93.598 -78.317 53.561
107471 93.618 -78.304 53.537
107472 93.629 -78.296 53.513
107473 93.641 -78.288 53.490
107474 93.652 -78.279 53.467
107475 93.664 -78.271 53.443
107492 93.917 -78.150 53.072
107493 93.930 -78.147 53.049
107494 93.943 -78.142 53.026
107495 93.956 -78.138 53.003
107496 93.969 -78.134 52.980
107497 93.969 -78.132 52.956
107498 93.966 -78.131 52.934
107499 93.967 -78.136 52.911
107500 93.970 -78.147 52.889
107501 93.974 -78.156 52.865
107502 93.977 -78.167 52.841
107503 93.981 -78.177 52.820
107504 93.984 -78.188 52.796
107505 93.987 -78.198 52.773
107506 93.991 -78.208 52.751
107507 93.994 -78.219 52.728
107508 93.998 -78.228 52.705
107509 94.001 -78.239 52.682
107510 94.004 -78.250 52.659
107511 94.008 -78.260 52.636
107512 94.011 -78.270 52.614
107513 94.015 -78.280 52.591
107514 94.018 -78.291 52.569
107515 94.031 -78.297 52.546
107516 94.039 -78.290 52.523
107517 94.064 -78.281 52.500
107518 94.088 -78.272 52.477
107519 94.114 -78.263 52.454
107520 94.124 -78.260 52.433
107521 94.135 -78.257 52.410
107522 94.145 -78.254 52.386
107523 94.156 -78.251 52.363
107524 94.166 -78.249 52.341
107525 94.177 -78.246 52.317
107526 94.187 -78.243 52.294
107527 94.201 -78.236 52.271
107528 94.214 -78.228 52.247
107529 94.228 -78.221 52.222
107530 94.242 -78.213 52.197
107531 94.256 -78.206 52.171
107532 94.269 -78.198 52.144
107533 94.283 -78.191 52.116
107534 94.297 -78.183 52.089
107535 94.311 -78.175 52.062
107536 94.325 -78.168 52.034
107537 94.338 -78.160 52.006
107538 94.352 -78.153 51.978
107539 94.366 -78.147 51.950
107540 94.372 -78.142 51.922
107541 94.374 -78.137 51.894
107542 94.396 -78.127 51.866
107543 94.416 -78.117 51.838
107544 94.437 -78.107 51.808
107545 94.458 -78.097 51.780
107546 94.478 -78.087 51.751
107547 94.498 -78.078 51.721
107548 94.523 -78.086 51.690
107549 94.560 -78.102 51.661
107550 94.598 -78.118 51.630
107551 94.635 -78.134 51.598
107552 94.661 -78.143 51.565
107553 94.672 -78.144 51.533
107554 94.682 -78.141 51.502
107555 94.690 -78.131 51.471
107556 94.699 -78.122 51.440
107557 94.707 -78.114 51.408
107558 94.715 -78.105 51.375
107559 94.723 -78.096 51.343
107560 94.723 -78.102 51.311
107561 94.725 -78.108 51.280
107562 94.737 -78.106 51.248
107563 94.748 -78.103 51.216
107564 94.760 -78.103 51.183
107565 94.772 -78.108 51.150
107566 94.784 -78.112 51.115
107567 94.796 -78.117 51.082
107568 94.809 -78.122 51.048
107569 94.821 -78.127 51.014
107570 94.837 -78.133 50.979
107571 94.862 -78.140 50.944
107572 94.886 -78.147 50.907
107573 94.906 -78.134 50.870
107574 94.924 -78.117 50.834
107575 94.942 -78.100 50.799
107576 94.960 -78.082 50.763
107577 94.978 -78.065 50.728
107578 94.996 -78.048 50.692
107579 95.015 -78.031 50.656
107580 95.033 -78.013 50.621
107581 95.051 -77.996 50.587
107582 95.069 -77.978 50.553
107583 95.093 -77.959 50.519
107584 95.132 -77.956 50.486
107585 95.157 -77.958 50.453
107586 95.181 -77.959 50.420
107587 95.205 -77.962 50.388
107588 95.230 -77.963 50.356
107589 95.254 -77.965 50.324
107590 95.278 -77.968 50.293
107591 95.303 -77.969 50.264
107592 95.327 -77.971 50.236
107593 95.338 -77.971 50.208
107594 95.346 -77.970 50.180
107595 95.353 -77.969 50.154
107596 95.360 -77.968 50.128
107597 95.369 -77.963 50.102
107598 95.379 -77.968 50.077
107599 95.390 -77.980 50.054
107600 95.401 -77.990 50.034
107601 95.404 -77.992 50.016
107602 95.406 -77.994 50.004
107603 95.409 -77.996 49.988
107604 95.411 -77.998 49.968
107605 95.414 -78.000 49.944
107606 95.416 -78.002 49.917
107607 95.418 -78.003 49.889
107608 95.420 -78.006 49.860
107609 95.424 -78.008 49.825
107610 95.456 -78.003 49.788
107611 95.487 -78.000 49.750
107612 95.519 -77.996 49.710
107613 95.551 -77.991 49.669
107614 95.583 -77.987 49.628
107615 95.615 -77.983 49.588
107616 95.628 -77.976 49.548
107617 95.640 -77.968 49.508
107618 95.642 -77.965 49.468
107619 95.632 -77.985 49.426
107620 95.623 -78.021 49.385
107621 95.616 -78.056 49.344
107622 95.626 -78.061 49.303
107623 95.641 -78.046 49.263
107624 95.660 -78.011 49.223
107625 95.677 -77.994 49.185
107626 95.684 -78.002 49.148
107627 95.680 -78.022 49.107
107628 95.676 -78.041 49.066
107629 95.672 -78.061 49.024
107630 95.682 -78.056 48.983
107631 95.697 -78.043 48.943
107632 95.716 -78.037 48.902
107633 95.743 -78.046 48.862
107634 95.770 -78.055 48.823
107635 95.797 -78.064 48.784
107636 95.824 -78.073 48.745
107637 95.847 -78.085 48.704
107638 95.865 -78.102 48.663
107639 95.881 -78.119 48.622
107640 95.883 -78.141 48.583
107641 95.870 -78.165 48.546
107642 95.860 -78.188 48.510
107643 95.856 -78.200 48.475
107644 95.858 -78.202 48.441
107645 95.880 -78.197 48.406
107646 95.906 -78.191 48.371
107647 95.933 -78.184 48.338
107648 95.960 -78.178 48.305
107649 95.986 -78.172 48.270
107650 96.002 -78.170 48.235
107651 96.004 -78.167 48.199
107652 96.005 -78.165 48.162
107653 96.007 -78.162 48.125
107654 96.008 -78.159 48.090
107655 96.010 -78.156 48.059
107656 96.011 -78.152 48.028
107657 96.029 -78.158 47.997
107658 96.054 -78.166 47.966
107659 96.079 -78.173 47.934
107660 96.104 -78.181 47.903
107661 96.125 -78.187 47.872
107662 96.141 -78.191 47.841
107663 96.156 -78.198 47.811
107664 96.172 -78.208 47.780
107665 96.187 -78.218 47.750
107666 96.202 -78.232 47.720
107667 96.217 -78.244 47.690
107668 96.232 -78.258 47.662
107669 96.252 -78.277 47.635
107670 96.272 -78.296 47.608
107671 96.288 -78.313 47.582
107672 96.291 -78.326 47.556
107673 96.300 -78.340 47.530
107674 96.315 -78.354 47.504
107675 96.329 -78.368 47.477
107676 96.344 -78.379 47.450
107677 96.360 -78.385 47.424
107678 96.389 -78.385 47.398
107679 96.419 -78.384 47.374
107680 96.450 -78.383 47.347
107681 96.480 -78.382 47.321
107682 96.511 -78.382 47.294
107683 96.541 -78.381 47.270
107684 96.571 -78.380 47.246
107685 96.575 -78.388 47.222
107686 96.579 -78.398 47.196
107687 96.582 -78.407 47.171
107688 96.586 -78.415 47.146
107689 96.590 -78.424 47.119
107690 96.594 -78.433 47.094
107691 96.602 -78.442 47.069
107692 96.609 -78.448 47.044
107693 96.614 -78.453 47.017
107694 96.620 -78.477 46.989
107695 96.626 -78.498 46.961
107696 96.636 -78.508 46.935
107697 96.645 -78.517 46.910
107698 96.655 -78.527 46.886
107699 96.664 -78.536 46.862
107700 96.674 -78.546 46.838
107701 96.670 -78.567 46.813
107702 96.665 -78.588 46.789
107703 96.660 -78.610 46.766
107704 96.664 -78.636 46.743
107705 96.683 -78.667 46.719
107706 96.702 -78.698 46.696
107707 96.721 -78.729 46.672
107708 96.740 -78.760 46.649
107709 96.759 -78.791 46.625
107710 96.779 -78.821 46.599
107711 96.814 -78.831 46.576
107712 96.850 -78.842 46.552
107713 96.885 -78.852 46.529
107714 96.915 -78.864 46.504
107715 96.945 -78.876 46.479
107716 97.005 -78.832 46.453
107717 97.014 -78.872 46.428
107718 97.015 -78.895 46.405
107719 97.027 -78.902 46.382
107720 97.058 -78.899 46.360
107721 97.088 -78.897 46.338
107722 97.119 -78.894 46.316
107723 97.188 -78.902 46.294
107724 97.270 -78.912 46.273
107725 97.314 -78.905 46.253
107726 97.357 -78.897 46.233
107727 97.400 -78.889 46.213
107728 97.440 -78.880 46.195
107729 97.475 -78.868 46.177
107730 97.510 -78.857 46.159
107731 97.544 -78.846 46.142
107732 97.579 -78.834 46.125
107733 97.613 -78.823 46.109
107734 97.644 -78.813 46.093
107735 97.621 -78.830 46.079
107736 97.614 -78.865 46.064
107737 97.628 -78.857 46.048
107738 97.641 -78.848 46.032
107739 97.655 -78.839 46.015
107740 97.669 -78.830 45.999
107741 97.684 -78.823 45.983
107742 97.728 -78.831 45.967
107743 97.779 -78.836 45.951
107744 97.812 -78.798 45.935
107745 97.833 -78.754 45.919
107746 97.850 -78.742 45.901
107747 97.865 -78.733 45.884
107748 97.881 -78.724 45.867
107749 97.897 -78.715 45.850
107750 97.913 -78.706 45.835
107751 97.914 -78.694 45.819
107752 97.898 -78.678 45.803
107753 97.881 -78.663 45.787
107754 97.864 -78.670 45.772
107755 97.858 -78.688 45.758
107756 97.876 -78.680 45.744
107757 97.928 -78.663 45.731
107758 97.915 -78.742 45.718
107759 97.905 -78.793 45.706
107760 97.909 -78.811 45.695
107761 97.938 -78.814 45.682
107762 97.967 -78.817 45.669
107763 97.997 -78.820 45.656
107764 98.026 -78.823 45.642
107765 98.055 -78.826 45.627
107766 98.084 -78.829 45.611
107767 98.113 -78.832 45.595
107768 98.143 -78.835 45.578
107769 98.172 -78.838 45.561
107770 98.185 -78.843 45.544
107771 98.193 -78.858 45.529
107772 98.200 -78.873 45.517
107773 98.208 -78.888 45.505
107774 98.215 -78.903 45.492
107775 98.223 -78.919 45.476
107776 98.230 -78.934 45.462
107777 98.238 -78.949 45.448
107778 98.245 -78.965 45.436
107779 98.269 -79.019 45.424
107780 98.293 -79.043 45.411
107781 98.322 -79.040 45.397
107782 98.350 -79.037 45.381
107783 98.375 -79.033 45.367
107784 98.392 -79.027 45.354
107785 98.405 -79.012 45.343
107786 98.408 -78.987 45.330
107787 98.411 -78.961 45.318
107788 98.440 -78.965 45.307
107789 98.473 -78.971 45.295
107790 98.507 -78.980 45.283
107791 98.541 -78.988 45.270
107792 98.565 -78.995 45.259
107793 98.589 -79.002 45.246
107794 98.613 -79.008 45.234
107795 98.637 -79.015 45.224
107796 98.661 -79.021 45.214
107797 98.685 -79.028 45.204
107798 98.709 -79.034 45.193
107799 98.733 -79.040 45.183
107800 98.757 -79.047 45.172
107801 98.781 -79.054 45.161
107802 98.805 -79.060 45.150
107803 98.829 -79.067 45.137
107804 98.852 -79.073 45.125
107805 98.874 -79.080 45.114
107806 98.897 -79.087 45.101
107807 98.919 -79.093 45.089
107808 98.956 -79.111 45.076
107809 98.980 -79.100 45.064
107810 99.004 -79.091 45.051
107811 99.029 -79.092 45.038
107812 99.054 -79.093 45.025
107813 99.079 -79.094 45.012
107814 99.104 -79.096 45.000
107815 99.129 -79.096 44.989
107816 99.154 -79.098 44.979
107817 99.182 -79.100 44.969
107818 99.209 -79.101 44.960
107819 99.236 -79.103 44.950
107820 99.258 -79.112 44.940
107821 99.277 -79.124 44.931
107822 99.285 -79.140 44.922
107823 99.292 -79.158 44.914
107824 99.298 -79.175 44.909
107825 99.303 -79.191 44.904
107826 99.306 -79.200 44.900
107827 99.309 -79.209 44.896
107828 99.323 -79.244 44.891
107829 99.349 -79.240 44.886
107830 99.376 -79.235 44.881
107831 99.402 -79.235 44.878
107832 99.423 -79.248 44.876
107833 99.445 -79.262 44.873
107834 99.470 -79.279 44.870
107835 99.495 -79.297 44.868
107836 99.517 -79.324 44.866
107837 99.540 -79.351 44.864
107838 99.563 -79.379 44.860
107839 99.588 -79.386 44.857
107840 99.614 -79.391 44.854
107841 99.640 -79.395 44.851
107842 99.666 -79.400 44.850
107843 99.692 -79.405 44.847
107844 99.718 -79.410 44.844
107845 99.744 -79.414 44.841
107846 99.770 -79.419 44.838
107847 99.795 -79.423 44.834
107848 99.822 -79.428 44.830
107849 99.850 -79.441 44.825
107850 99.884 -79.474 44.818
107851 99.918 -79.507 44.810
107852 99.952 -79.540 44.802
107853 99.986 -79.573 44.793
107854 100.020 -79.614 44.784
107855 100.054 -79.661 44.776
107856 100.093 -79.670 44.767
107857 100.132 -79.679 44.758
107858 100.171 -79.689 44.748
107859 100.210 -79.698 44.736
107860 100.238 -79.722 44.723
107861 100.263 -79.748 44.712
107862 100.286 -79.784 44.701
107863 100.309 -79.823 44.693
107864 100.331 -79.861 44.683
107865 100.353 -79.900 44.671
107866 100.375 -79.939 44.657
107867 100.409 -79.952 44.645
107868 100.446 -79.957 44.633
107869 100.483 -79.962 44.621
107870 100.520 -79.967 44.605
107871 100.557 -79.972 44.590
107872 100.594 -79.977 44.575
107873 100.631 -79.982 44.558
107874 100.668 -79.987 44.543
107875 100.703 -80.012 44.529
107876 100.720 -80.048 44.515
107877 100.748 -80.084 44.499
107878 100.776 -80.121 44.484
107879 100.805 -80.157 44.468
107880 100.833 -80.193 44.451
107881 100.863 -80.228 44.436
107882 100.897 -80.262 44.421
107883 100.931 -80.295 44.405
107884 100.966 -80.328 44.390
107885 100.999 -80.362 44.375
107886 101.034 -80.394 44.359
107887 101.067 -80.428 44.343
107888 101.102 -80.461 44.325
107889 101.135 -80.494 44.308
107890 101.170 -80.527 44.291
107891 101.204 -80.560 44.273
107892 101.237 -80.583 44.256
107893 101.270 -80.599 44.239
107894 101.303 -80.615 44.222
107895 101.336 -80.630 44.204
107896 101.369 -80.646 44.187
107897 101.402 -80.662 44.169
107898 101.435 -80.678 44.152
107899 101.468 -80.694 44.135
107900 101.501 -80.710 44.119
107901 101.534 -80.726 44.102
107902 101.567 -80.742 44.086
107903 101.600 -80.757 44.070
107904 101.633 -80.773 44.052
108160 93.507 -78.414 53.802
108161 93.517 -78.412 53.784
108162 93.527 -78.410 53.764
108163 93.532 -78.404 53.743
108164 93.531 -78.395 53.721
108165 93.530 -78.385 53.700
108166 93.528 -78.376 53.676
108167 93.536 -78.363 53.652
108168 93.546 -78.347 53.626
108169 93.558 -78.332 53.602
108170 93.581 -78.316 53.577
108171 93.604 -78.302 53.552
108172 93.619 -78.291 53.528
108173 93.631 -78.282 53.505
108174 93.642 -78.275 53.481
108175 93.654 -78.266 53.457
108176 93.665 -78.258 53.434
108194 93.931 -78.135 53.036
108195 93.943 -78.131 53.013
108196 93.957 -78.127 52.989
108197 93.969 -78.122 52.965
108198 93.969 -78.121 52.941
108199 93.967 -78.120 52.917
108200 93.966 -78.123 52.894
108201 93.970 -78.133 52.871
108202 93.973 -78.144 52.848
108203 93.977 -78.153 52.824
108204 93.980 -78.164 52.802
108205 93.983 -78.174 52.778
108206 93.987 -78.184 52.754
108207 93.990 -78.195 52.731
108208 93.994 -78.205 52.708
108209 93.997 -78.216 52.685
108210 94.000 -78.225 52.662
108211 94.004 -78.236 52.638
108212 94.007 -78.247 52.615
108213 94.011 -78.257 52.592
108214 94.014 -78.267 52.569
108215 94.018 -78.277 52.546
108216 94.034 -78.282 52.523
108217 94.040 -78.273 52.499
108218 94.065 -78.264 52.475
108219 94.090 -78.254 52.452
108220 94.115 -78.245 52.430
108221 94.134 -78.238 52.407
108222 94.145 -78.236 52.384
108223 94.155 -78.233 52.360
108224 94.166 -78.231 52.335
108225 94.177 -78.228 52.311
108226 94.187 -78.225 52.287
108227 94.197 -78.222 52.264
108228 94.210 -78.216 52.239
108229 94.223 -78.209 52.215
108230 94.238 -78.201 52.190
108231 94.251 -78.194 52.164
108232 94.265 -78.186 52.138
108233 94.279 -78.178 52.112
108234 94.292 -78.171 52.086
108235 94.306 -78.163 52.059
108236 94.320 -78.156 52.034
108237 94.334 -78.148 52.007
108238 94.347 -78.141 51.980
108239 94.361 -78.134 51.953
108240 94.376 -78.128 51.926
108241 94.385 -78.122 51.899
108242 94.386 -78.118 51.872
108243 94.398 -78.110 51.845
108244 94.419 -78.100 51.817
108245 94.439 -78.091 51.789
108246 94.460 -78.080 51.762
108247 94.481 -78.070 51.734
108248 94.501 -78.060 51.705
108249 94.519 -78.061 51.676
108250 94.556 -78.077 51.647
108251 94.594 -78.094 51.617
108252 94.631 -78.109 51.587
108253 94.659 -78.120 51.555
108254 94.671 -78.121 51.524
108255 94.682 -78.121 51.493
108256 94.690 -78.112 51.464
108257 94.699 -78.104 51.433
108258 94.707 -78.095 51.402
108259 94.715 -78.086 51.370
108260 94.723 -78.078 51.338
108261 94.724 -78.082 51.307
108262 94.724 -78.090 51.276
108263 94.734 -78.088 51.244
108264 94.746 -78.086 51.213
108265 94.758 -78.085 51.181
108266 94.770 -78.091 51.148
108267 94.782 -78.095 51.113
108268 94.794 -78.100 51.080
108269 94.806 -78.105 51.046
108270 94.818 -78.109 51.012
108271 94.840 -78.116 50.978
108272 94.864 -78.124 50.943
108273 94.889 -78.131 50.907
108274 94.908 -78.116 50.870
108275 94.926 -78.098 50.834
108276 94.944 -78.081 50.799
108277 94.962 -78.064 50.764
108278 94.981 -78.047 50.729
108279 94.999 -78.029 50.694
108280 95.017 -78.012 50.658
108281 95.035 -77.995 50.624
108282 95.053 -77.978 50.589
108283 95.071 -77.961 50.556
108284 95.090 -77.943 50.522
108285 95.115 -77.939 50.490
108286 95.140 -77.940 50.457
108287 95.164 -77.943 50.424
108288 95.188 -77.944 50.392
108289 95.213 -77.946 50.361
108290 95.237 -77.949 50.329
108291 95.261 -77.950 50.299
108292 95.285 -77.953 50.271
108293 95.310 -77.954 50.242
108294 95.334 -77.956 50.214
108295 95.342 -77.956 50.188
108296 95.350 -77.955 50.162
108297 95.357 -77.952 50.137
108298 95.367 -77.956 50.112
108299 95.379 -77.967 50.089
108300 95.383 -77.970 50.067
108301 95.385 -77.972 50.047
108302 95.388 -77.974 50.031
108303 95.390 -77.976 50.016
108304 95.393 -77.978 49.998
108305 95.395 -77.980 49.976
108306 95.398 -77.981 49.951
108307 95.400 -77.984 49.923
108308 95.403 -77.985 49.893
108309 95.409 -77.987 49.861
108310 95.422 -77.992 49.825
108311 95.453 -77.982 49.787
108312 95.485 -77.978 49.747
108313 95.517 -77.974 49.704
108314 95.549 -77.969 49.663
108315 95.581 -77.965 49.622
108316 95.612 -77.961 49.582
108317 95.644 -77.957 49.542
108318 95.661 -77.950 49.502
108319 95.671 -77.943 49.462
108320 95.659 -77.949 49.422
108321 95.650 -77.986 49.382
108322 95.656 -78.012 49.342
108323 95.662 -77.995 49.302
108324 95.665 -77.970 49.263
108325 95.689 -77.956 49.225
108326 95.687 -77.974 49.188
108327 95.683 -77.993 49.150
108328 95.680 -78.013 49.111
108329 95.676 -78.033 49.070
108330 95.674 -78.050 49.028
108331 95.689 -78.036 48.988
108332 95.704 -78.022 48.946
108333 95.731 -78.031 48.905
108334 95.758 -78.040 48.864
108335 95.784 -78.050 48.826
108336 95.812 -78.059 48.787
108337 95.838 -78.068 48.748
108338 95.865 -78.077 48.708
108339 95.888 -78.089 48.668
108340 95.901 -78.107 48.627
108341 95.888 -78.132 48.587
108342 95.876 -78.156 48.550
108343 95.870 -78.172 48.513
108344 95.872 -78.174 48.477
108345 95.874 -78.175 48.441
108346 95.894 -78.172 48.406
108347 95.920 -78.166 48.371
108348 95.947 -78.159 48.337
108349 95.974 -78.153 48.305
108350 95.990 -78.151 48.271
108351 96.007 -78.150 48.235
108352 96.021 -78.149 48.201
108353 96.022 -78.147 48.165
108354 96.024 -78.144 48.130
108355 96.025 -78.141 48.096
108356 96.027 -78.138 48.066
108357 96.028 -78.135 48.036
108358 96.043 -78.140 48.008
108359 96.067 -78.149 47.977
108360 96.093 -78.156 47.947
108361 96.118 -78.164 47.917
108362 96.135 -78.169 47.887
108363 96.151 -78.174 47.855
108364 96.166 -78.178 47.825
108365 96.181 -78.186 47.796
108366 96.196 -78.197 47.768
108367 96.212 -78.211 47.738
108368 96.226 -78.224 47.709
108369 96.241 -78.238 47.680
108370 96.257 -78.252 47.653
108371 96.277 -78.271 47.627
108372 96.286 -78.285 47.602
108373 96.297 -78.299 47.575
108374 96.312 -78.313 47.549
108375 96.326 -78.327 47.523
108376 96.340 -78.341 47.497
108377 96.355 -78.354 47.471
108378 96.371 -78.360 47.444
108379 96.387 -78.366 47.419
108380 96.416 -78.365 47.396
108381 96.447 -78.364 47.370
108382 96.477 -78.363 47.344
108383 96.508 -78.363 47.319
108384 96.538 -78.362 47.295
108385 96.555 -78.366 47.270
108386 96.559 -78.375 47.245
108387 96.563 -78.384 47.222
108388 96.567 -78.392 47.197
108389 96.571 -78.401 47.171
108390 96.578 -78.410 47.144
108391 96.588 -78.420 47.118
108392 96.596 -78.427 47.090
108393 96.596 -78.428 47.063
108394 96.605 -78.436 47.036
108395 96.610 -78.460 47.010
108396 96.616 -78.484 46.984
108397 96.622 -78.507 46.958
108398 96.631 -78.517 46.933
108399 96.641 -78.526 46.908
108400 96.650 -78.535 46.884
108401 96.660 -78.545 46.860
108402 96.655 -78.567 46.835
108403 96.650 -78.588 46.810
108404 96.645 -78.610 46.786
108405 96.639 -78.632 46.764
108406 96.649 -78.659 46.740
108407 96.668 -78.690 46.715
108408 96.687 -78.721 46.691
108409 96.706 -78.752 46.667
108410 96.725 -78.783 46.642
108411 96.760 -78.795 46.616
108412 96.795 -78.805 46.592
108413 96.830 -78.815 46.568
108414 96.865 -78.826 46.543
108415 96.895 -78.838 46.518
108416 96.970 -78.849 46.492
108417 97.011 -78.861 46.465
108418 97.018 -78.844 46.439
108419 97.001 -78.865 46.414
108420 97.006 -78.886 46.390
108421 97.022 -78.890 46.367
108422 97.076 -78.893 46.344
108423 97.158 -78.903 46.321
108424 97.228 -78.908 46.299
108425 97.271 -78.900 46.278
108426 97.314 -78.893 46.259
108427 97.357 -78.884 46.239
108428 97.400 -78.877 46.219
108429 97.435 -78.865 46.200
108430 97.470 -78.854 46.182
108431 97.505 -78.843 46.165
108432 97.539 -78.831 46.148
108433 97.574 -78.820 46.131
108434 97.608 -78.808 46.115
108435 97.643 -78.797 46.100
108436 97.664 -78.789 46.086
108437 97.657 -78.825 46.070
108438 97.650 -78.860 46.054
108439 97.658 -78.865 46.038
108440 97.672 -78.855 46.022
108441 97.687 -78.847 46.006
108442 97.731 -78.856 45.991
108443 97.774 -78.865 45.976
108444 97.817 -78.874 45.960
108445 97.840 -78.833 45.945
108446 97.860 -78.789 45.929
108447 97.880 -78.745 45.911
108448 97.897 -78.732 45.894
108449 97.913 -78.724 45.877
108450 97.929 -78.714 45.859
108451 97.945 -78.705 45.844
108452 97.937 -78.692 45.827
108453 97.921 -78.676 45.811
108454 97.904 -78.660 45.796
108455 97.892 -78.668 45.780
108456 97.885 -78.706 45.765
108457 97.879 -78.679 45.750
108458 97.929 -78.661 45.737
108459 97.938 -78.708 45.724
108460 97.925 -78.768 45.711
108461 97.925 -78.789 45.698
108462 97.939 -78.801 45.685
108463 97.968 -78.804 45.673
108464 97.998 -78.807 45.660
108465 98.027 -78.809 45.645
108466 98.056 -78.812 45.629
108467 98.086 -78.815 45.613
108468 98.115 -78.818 45.599
108469 98.144 -78.821 45.585
108470 98.170 -78.829 45.569
108471 98.184 -78.831 45.553
108472 98.190 -78.820 45.537
108473 98.197 -78.835 45.525
108474 98.205 -78.850 45.513
108475 98.212 -78.865 45.502
108476 98.220 -78.880 45.489
108477 98.228 -78.896 45.476
108478 98.235 -78.911 45.463
108479 98.253 -78.934 45.451
108480 98.273 -78.999 45.437
108481 98.296 -79.034 45.422
108482 98.324 -79.031 45.408
108483 98.353 -79.028 45.392
108484 98.381 -79.024 45.378
108485 98.410 -79.021 45.366
108486 98.426 -79.010 45.354
108487 98.429 -78.984 45.343
108488 98.433 -78.959 45.332
108489 98.452 -78.950 45.322
108490 98.485 -78.958 45.309
108491 98.516 -78.965 45.296
108492 98.540 -78.972 45.283
108493 98.564 -78.979 45.268
108494 98.588 -78.985 45.254
108495 98.612 -78.992 45.242
108496 98.636 -78.999 45.232
108497 98.660 -79.005 45.221
108498 98.684 -79.012 45.210
108499 98.706 -79.018 45.199
108500 98.728 -79.024 45.188
108501 98.750 -79.031 45.177
108502 98.773 -79.038 45.166
108503 98.795 -79.044 45.155
108504 98.817 -79.051 45.143
108505 98.839 -79.058 45.131
108506 98.861 -79.064 45.120
108507 98.884 -79.071 45.108
108508 98.939 -79.129 45.097
108509 98.972 -79.134 45.086
108510 98.995 -79.125 45.074
108511 99.019 -79.114 45.061
108512 99.043 -79.104 45.048
108513 99.066 -79.093 45.035
108514 99.090 -79.084 45.021
108515 99.114 -79.073 45.009
108516 99.137 -79.066 44.998
108517 99.162 -79.068 44.988
108518 99.188 -79.069 44.980
108519 99.213 -79.075 44.971
108520 99.233 -79.087 44.960
108521 99.253 -79.099 44.948
108522 99.272 -79.111 44.938
108523 99.283 -79.127 44.929
108524 99.289 -79.144 44.920
108525 99.294 -79.157 44.913
108526 99.296 -79.166 44.906
108527 99.299 -79.175 44.901
108528 99.302 -79.184 44.896
108529 99.315 -79.219 44.892
108530 99.336 -79.235 44.888
108531 99.363 -79.231 44.884
108532 99.390 -79.227 44.881
108533 99.427 -79.221 44.878
108534 99.445 -79.227 44.875
108535 99.470 -79.245 44.872
108536 99.494 -79.269 44.870
108537 99.516 -79.296 44.868
108538 99.539 -79.323 44.866
108539 99.562 -79.350 44.864
108540 99.586 -79.370 44.861
108541 99.612 -79.375 44.859
108542 99.638 -79.379 44.858
108543 99.663 -79.384 44.857
108544 99.689 -79.388 44.855
108545 99.715 -79.393 44.852
108546 99.741 -79.397 44.849
108547 99.767 -79.402 44.844
108548 99.793 -79.407 44.839
108549 99.819 -79.411 44.833
108550 99.852 -79.440 44.827
108551 99.886 -79.473 44.819
108552 99.920 -79.506 44.810
108553 99.954 -79.538 44.802
108554 99.988 -79.571 44.793
108555 100.021 -79.612 44.784
108556 100.056 -79.654 44.776
108557 100.096 -79.664 44.767
108558 100.134 -79.673 44.758
108559 100.174 -79.683 44.749
108560 100.212 -79.692 44.737
108561 100.248 -79.706 44.725
108562 100.273 -79.733 44.713
108563 100.295 -79.771 44.702
108564 100.318 -79.810 44.691
108565 100.340 -79.849 44.680
108566 100.362 -79.888 44.669
108567 100.385 -79.927 44.657
108568 100.415 -79.947 44.644
108569 100.452 -79.952 44.631
108570 100.489 -79.957 44.618
108571 100.525 -79.962 44.604
108572 100.563 -79.967 44.588
108573 100.600 -79.972 44.572
108574 100.637 -79.977 44.557
108575 100.674 -79.982 44.543
108576 100.711 -79.987 44.528
108577 100.739 -80.018 44.514
108578 100.771 -80.052 44.499
108579 100.805 -80.085 44.484
108580 100.839 -80.118 44.469
108581 100.874 -80.151 44.453
108582 100.908 -80.184 44.438
108583 100.941 -80.218 44.423
108584 100.976 -80.251 44.408
108585 101.009 -80.284 44.393
108586 101.042 -80.318 44.377
108587 101.075 -80.352 44.361
108588 101.108 -80.385 44.346
108589 101.141 -80.419 44.328
108590 101.174 -80.453 44.311
108591 101.206 -80.487 44.294
108592 101.240 -80.521 44.278
108593 101.275 -80.538 44.261
108594 101.307 -80.554 44.244
108595 101.339 -80.571 44.228
108596 101.371 -80.587 44.210
108597 101.403 -80.603 44.193
108598 101.435 -80.620 44.176
108599 101.467 -80.635 44.159
108600 101.499 -80.651 44.142
108601 101.531 -80.668 44.125
108602 101.563 -80.684 44.108
108603 101.596 -80.700 44.092
108604 101.628 -80.716 44.075
108605 101.659 -80.732 44.057
108862 93.517 -78.401 53.775
108863 93.527 -78.398 53.755
108864 93.529 -78.391 53.733
108865 93.528 -78.382 53.711
108866 93.526 -78.373 53.689
108867 93.526 -78.363 53.667
108868 93.537 -78.347 53.642
108869 93.547 -78.332 53.616
108870 93.564 -78.316 53.592
108871 93.586 -78.301 53.568
108872 93.609 -78.286 53.543
108873 93.620 -78.278 53.520
108874 93.632 -78.269 53.496
108875 93.643 -78.261 53.473
108876 93.655 -78.253 53.448
108877 93.666 -78.245 53.425
108895 93.931 -78.124 53.022
108896 93.944 -78.120 52.999
108897 93.957 -78.116 52.975
108898 93.970 -78.111 52.951
108899 93.970 -78.109 52.927
108900 93.967 -78.109 52.903
108901 93.966 -78.109 52.879
108902 93.969 -78.120 52.856
108903 93.973 -78.130 52.832
108904 93.976 -78.141 52.808
108905 93.979 -78.150 52.784
108906 93.983 -78.161 52.761
108907 93.986 -78.172 52.737
108908 93.990 -78.181 52.714
108909 93.993 -78.192 52.691
108910 93.996 -78.202 52.667
108911 94.000 -78.213 52.644
108912 94.003 -78.222 52.619
108913 94.007 -78.233 52.595
108914 94.010 -78.244 52.572
108915 94.013 -78.253 52.549
108916 94.021 -78.262 52.526
108917 94.036 -78.266 52.503
108918 94.042 -78.256 52.479
108919 94.066 -78.247 52.454
108920 94.091 -78.238 52.431
108921 94.117 -78.228 52.407
108922 94.141 -78.219 52.384
108923 94.155 -78.216 52.360
108924 94.166 -78.213 52.336
108925 94.176 -78.210 52.312
108926 94.187 -78.207 52.287
108927 94.197 -78.204 52.263
108928 94.208 -78.202 52.238
108929 94.219 -78.197 52.214
108930 94.233 -78.190 52.190
108931 94.247 -78.182 52.165
108932 94.260 -78.175 52.139
108933 94.274 -78.167 52.113
108934 94.288 -78.159 52.088
108935 94.301 -78.152 52.063
108936 94.315 -78.144 52.037
108937 94.329 -78.137 52.011
108938 94.343 -78.129 51.986
108939 94.356 -78.122 51.960
108940 94.370 -78.116 51.934
108941 94.385 -78.109 51.908
108942 94.397 -78.103 51.882
108943 94.399 -78.098 51.856
108944 94.401 -78.094 51.830
108945 94.421 -78.084 51.804
108946 94.442 -78.074 51.776
108947 94.463 -78.063 51.749
108948 94.483 -78.053 51.721
108949 94.504 -78.044 51.694
108950 94.521 -78.041 51.666
108951 94.551 -78.053 51.638
108952 94.589 -78.069 51.608
108953 94.626 -78.085 51.577
108954 94.657 -78.097 51.547
108955 94.669 -78.098 51.517
108956 94.681 -78.098 51.487
108957 94.691 -78.094 51.458
108958 94.699 -78.085 51.428
108959 94.707 -78.077 51.397
108960 94.715 -78.068 51.365
108961 94.723 -78.059 51.334
108962 94.725 -78.062 51.303
108963 94.724 -78.070 51.272
108964 94.732 -78.072 51.242
108965 94.744 -78.069 51.211
108966 94.755 -78.068 51.179
108967 94.767 -78.073 51.146
108968 94.779 -78.078 51.113
108969 94.791 -78.083 51.079
108970 94.804 -78.087 51.045
108971 94.818 -78.093 51.011
108972 94.842 -78.100 50.978
108973 94.867 -78.107 50.943
108974 94.892 -78.114 50.907
108975 94.910 -78.097 50.870
108976 94.928 -78.080 50.835
108977 94.947 -78.062 50.800
108978 94.965 -78.046 50.765
108979 94.983 -78.028 50.731
108980 95.001 -78.011 50.696
108981 95.019 -77.993 50.660
108982 95.037 -77.977 50.626
108983 95.055 -77.959 50.592
108984 95.074 -77.942 50.558
108985 95.088 -77.927 50.525
108986 95.102 -77.913 50.493
108987 95.125 -77.917 50.461
108988 95.149 -77.921 50.428
108989 95.173 -77.924 50.396
108990 95.196 -77.928 50.365
108991 95.220 -77.931 50.335
108992 95.245 -77.934 50.305
108993 95.269 -77.935 50.277
108994 95.294 -77.937 50.249
108995 95.318 -77.940 50.222
108996 95.339 -77.941 50.196
108997 95.346 -77.940 50.171
108998 95.356 -77.943 50.147
108999 95.362 -77.948 50.124
109000 95.365 -77.950 50.102
109001 95.367 -77.952 50.081
109002 95.369 -77.954 50.061
109003 95.373 -77.956 50.044
109004 95.379 -77.959 50.027
109005 95.386 -77.961 50.006
109006 95.392 -77.963 49.982
109007 95.399 -77.965 49.955
109008 95.405 -77.968 49.925
109009 95.411 -77.970 49.893
109010 95.418 -77.972 49.859
109011 95.429 -77.981 49.822
109012 95.454 -77.990 49.782
109013 95.485 -77.979 49.740
109014 95.516 -77.968 49.696
109015 95.547 -77.957 49.656
109016 95.579 -77.946 49.616
109017 95.610 -77.939 49.577
109018 95.641 -77.935 49.538
109019 95.674 -77.931 49.498
109020 95.694 -77.925 49.459
109021 95.686 -77.925 49.420
109022 95.677 -77.958 49.380
109023 95.679 -77.995 49.341
109024 95.674 -77.993 49.302
109025 95.734 -77.946 49.264
109026 95.691 -77.945 49.226
109027 95.687 -77.965 49.189
109028 95.683 -77.985 49.152
109029 95.679 -78.005 49.114
109030 95.676 -78.025 49.074
109031 95.681 -78.028 49.033
109032 95.696 -78.015 48.993
109033 95.719 -78.016 48.952
109034 95.746 -78.025 48.910
109035 95.772 -78.034 48.870
109036 95.799 -78.044 48.830
109037 95.826 -78.053 48.790
109038 95.853 -78.062 48.751
109039 95.880 -78.072 48.712
109040 95.906 -78.081 48.672
109041 95.906 -78.100 48.632
109042 95.893 -78.125 48.592
109043 95.885 -78.144 48.555
109044 95.886 -78.145 48.519
109045 95.888 -78.147 48.483
109046 95.890 -78.150 48.446
109047 95.908 -78.146 48.409
109048 95.935 -78.140 48.373
109049 95.961 -78.134 48.338
109050 95.977 -78.132 48.305
109051 95.993 -78.131 48.271
109052 96.009 -78.131 48.236
109053 96.025 -78.129 48.203
109054 96.039 -78.128 48.169
109055 96.041 -78.125 48.135
109056 96.042 -78.122 48.103
109057 96.044 -78.120 48.073
109058 96.045 -78.117 48.045
109059 96.057 -78.120 48.017
109060 96.082 -78.129 47.989
109061 96.106 -78.139 47.961
109062 96.130 -78.147 47.931
109063 96.146 -78.152 47.901
109064 96.161 -78.157 47.871
109065 96.176 -78.163 47.841
109066 96.191 -78.169 47.812
109067 96.205 -78.179 47.784
109068 96.221 -78.192 47.754
109069 96.236 -78.204 47.725
109070 96.250 -78.217 47.697
109071 96.265 -78.230 47.672
109072 96.279 -78.244 47.647
109073 96.294 -78.259 47.622
109074 96.308 -78.272 47.596
109075 96.323 -78.286 47.570
109076 96.337 -78.300 47.543
109077 96.352 -78.314 47.517
109078 96.366 -78.328 47.491
109079 96.382 -78.335 47.465
109080 96.398 -78.341 47.440
109081 96.414 -78.347 47.417
109082 96.444 -78.346 47.392
109083 96.475 -78.345 47.367
109084 96.505 -78.344 47.343
109085 96.536 -78.344 47.320
109086 96.540 -78.352 47.295
109087 96.543 -78.361 47.270
109088 96.547 -78.369 47.246
109089 96.553 -78.379 47.220
109090 96.563 -78.388 47.193
109091 96.573 -78.397 47.165
109092 96.583 -78.405 47.138
109093 96.583 -78.407 47.110
109094 96.586 -78.407 47.082
109095 96.596 -78.420 47.056
109096 96.601 -78.443 47.030
109097 96.607 -78.467 47.005
109098 96.612 -78.490 46.980
109099 96.618 -78.514 46.955
109100 96.627 -78.525 46.931
109101 96.637 -78.534 46.906
109102 96.645 -78.545 46.881
109103 96.640 -78.567 46.856
109104 96.635 -78.588 46.831
109105 96.629 -78.610 46.807
109106 96.624 -78.632 46.784
109107 96.618 -78.654 46.759
109108 96.634 -78.683 46.734
109109 96.653 -78.714 46.710
109110 96.672 -78.745 46.685
109111 96.706 -78.758 46.659
109112 96.741 -78.768 46.633
109113 96.777 -78.779 46.608
109114 96.812 -78.789 46.583
109115 96.845 -78.800 46.557
109116 96.890 -78.816 46.531
109117 96.924 -78.833 46.504
109118 96.957 -78.850 46.476
109119 96.991 -78.868 46.450
109120 97.056 -78.913 46.423
109121 97.042 -78.908 46.398
109122 97.051 -78.890 46.374
109123 97.127 -78.905 46.351
109124 97.185 -78.904 46.327
109125 97.228 -78.896 46.305
109126 97.271 -78.888 46.284
109127 97.314 -78.880 46.264
109128 97.357 -78.872 46.245
109129 97.396 -78.862 46.225
109130 97.430 -78.851 46.206
109131 97.465 -78.839 46.189
109132 97.500 -78.828 46.171
109133 97.534 -78.817 46.154
109134 97.569 -78.805 46.138
109135 97.604 -78.795 46.122
109136 97.643 -78.794 46.107
109137 97.697 -78.792 46.092
109138 97.689 -78.849 46.078
109139 97.688 -78.863 46.062
109140 97.690 -78.858 46.046
109141 97.690 -78.872 46.030
109142 97.681 -78.864 46.015
109143 97.732 -78.864 46.000
109144 97.783 -78.864 45.985
109145 97.820 -78.846 45.969
109146 97.850 -78.827 45.954
109147 97.886 -78.823 45.939
109148 97.908 -78.780 45.922
109149 97.928 -78.737 45.905
109150 97.945 -78.723 45.887
109151 97.961 -78.714 45.869
109152 97.977 -78.705 45.854
109153 97.960 -78.689 45.837
109154 97.943 -78.673 45.821
109155 97.924 -78.665 45.805
109156 97.920 -78.686 45.790
109157 97.911 -78.718 45.775
109158 97.900 -78.683 45.760
109159 97.931 -78.659 45.746
109160 97.962 -78.673 45.732
109161 97.945 -78.745 45.718
109162 97.945 -78.764 45.704
109163 97.945 -78.785 45.691
109164 97.969 -78.790 45.678
109165 97.999 -78.793 45.664
109166 98.028 -78.796 45.649
109167 98.058 -78.799 45.634
109168 98.087 -78.802 45.619
109169 98.116 -78.805 45.604
109170 98.145 -78.808 45.590
109171 98.167 -78.822 45.576
109172 98.184 -78.829 45.560
109173 98.190 -78.818 45.545
109174 98.195 -78.805 45.533
109175 98.202 -78.812 45.522
109176 98.210 -78.827 45.512
109177 98.217 -78.843 45.499
109178 98.225 -78.858 45.486
109179 98.232 -78.873 45.473
109180 98.257 -78.915 45.461
109181 98.277 -78.981 45.446
109182 98.299 -79.024 45.432
109183 98.327 -79.021 45.418
109184 98.356 -79.018 45.403
109185 98.384 -79.015 45.390
109186 98.406 -78.975 45.377
109187 98.394 -78.987 45.366
109188 98.420 -78.977 45.354
109189 98.454 -78.955 45.343
109190 98.463 -78.936 45.332
109191 98.491 -78.943 45.320
109192 98.515 -78.949 45.308
109193 98.538 -78.956 45.294
109194 98.560 -78.962 45.279
109195 98.582 -78.969 45.264
109196 98.604 -78.976 45.251
109197 98.627 -78.983 45.239
109198 98.649 -78.989 45.228
109199 98.671 -78.996 45.216
109200 98.693 -79.002 45.203
109201 98.715 -79.009 45.190
109202 98.737 -79.015 45.179
109203 98.760 -79.022 45.167
109204 98.782 -79.029 45.156
109205 98.804 -79.035 45.145
109206 98.826 -79.042 45.133
109207 98.849 -79.049 45.121
109208 98.871 -79.055 45.110
109209 98.949 -79.101 45.099
109210 98.972 -79.088 45.090
109211 98.997 -79.081 45.080
109212 99.021 -79.075 45.069
109213 99.046 -79.068 45.057
109214 99.070 -79.062 45.045
109215 99.094 -79.056 45.032
109216 99.119 -79.049 45.019
109217 99.143 -79.044 45.007
109218 99.168 -79.046 44.997
109219 99.188 -79.052 44.990
109220 99.197 -79.062 44.980
109221 99.222 -79.073 44.969
109222 99.248 -79.085 44.958
109223 99.267 -79.097 44.948
109224 99.280 -79.112 44.938
109225 99.284 -79.123 44.929
109226 99.287 -79.132 44.922
109227 99.289 -79.141 44.914
109228 99.292 -79.150 44.907
109229 99.294 -79.159 44.901
109230 99.307 -79.194 44.895
109231 99.323 -79.231 44.890
109232 99.350 -79.227 44.886
109233 99.414 -79.216 44.882
109234 99.494 -79.203 44.879
109235 99.471 -79.200 44.877
109236 99.470 -79.213 44.875
109237 99.493 -79.241 44.873
109238 99.515 -79.267 44.870
109239 99.538 -79.294 44.868
109240 99.561 -79.322 44.867
109241 99.584 -79.349 44.866
109242 99.609 -79.358 44.866
109243 99.635 -79.363 44.866
109244 99.661 -79.367 44.864
109245 99.687 -79.372 44.861
109246 99.713 -79.376 44.858
109247 99.738 -79.381 44.854
109248 99.765 -79.386 44.847
109249 99.790 -79.391 44.841
109250 99.820 -79.406 44.834
109251 99.853 -79.438 44.827
109252 99.888 -79.472 44.818
109253 99.922 -79.504 44.809
109254 99.956 -79.537 44.800
109255 99.990 -79.570 44.792
109256 100.023 -79.609 44.784
109257 100.059 -79.648 44.775
109258 100.098 -79.657 44.767
109259 100.137 -79.667 44.758
109260 100.176 -79.676 44.748
109261 100.215 -79.686 44.737
109262 100.254 -79.695 44.726
109263 100.282 -79.720 44.713
109264 100.304 -79.758 44.701
109265 100.327 -79.797 44.690
109266 100.349 -79.836 44.678
109267 100.371 -79.875 44.667
109268 100.394 -79.914 44.655
109269 100.420 -79.943 44.643
109270 100.458 -79.947 44.630
109271 100.494 -79.952 44.616
109272 100.531 -79.957 44.602
109273 100.568 -79.962 44.586
109274 100.606 -79.967 44.570
109275 100.643 -79.970 44.556
109276 100.680 -79.969 44.542
109277 100.717 -79.968 44.528
109278 100.752 -79.992 44.513
109279 100.783 -80.025 44.498
109280 100.816 -80.059 44.483
109281 100.848 -80.093 44.469
109282 100.882 -80.127 44.455
109283 100.914 -80.160 44.440
109284 100.948 -80.194 44.425
109285 100.980 -80.228 44.411
109286 101.013 -80.262 44.395
109287 101.046 -80.296 44.380
109288 101.079 -80.329 44.364
109289 101.112 -80.363 44.348
109290 101.145 -80.397 44.332
109291 101.178 -80.431 44.315
109292 101.211 -80.464 44.299
109293 101.251 -80.482 44.282
109294 101.292 -80.500 44.266
109295 101.324 -80.516 44.250
109296 101.356 -80.532 44.233
109297 101.388 -80.549 44.216
109298 101.420 -80.565 44.199
109299 101.453 -80.581 44.182
109300 101.484 -80.597 44.165
109301 101.516 -80.613 44.148
109302 101.549 -80.629 44.131
109303 101.581 -80.645 44.114
109304 101.613 -80.662 44.098
109305 101.645 -80.678 44.080
109306 101.677 -80.694 44.063
109564 93.526 -78.387 53.744
109565 93.525 -78.378 53.723
109566 93.524 -78.369 53.701
109567 93.523 -78.360 53.678
109568 93.528 -78.347 53.656
109569 93.538 -78.332 53.632
109570 93.548 -78.316 53.607
109571 93.569 -78.300 53.583
109572 93.592 -78.285 53.559
109573 93.610 -78.273 53.534
109574 93.621 -78.265 53.511
109575 93.633 -78.257 53.488
109576 93.644 -78.248 53.464
109577 93.656 -78.240 53.439
109578 93.668 -78.232 53.417
109597 93.945 -78.109 52.986
109598 93.958 -78.104 52.963
109599 93.971 -78.100 52.938
109600 93.970 -78.098 52.914
109601 93.968 -78.097 52.890
109602 93.965 -78.096 52.866
109603 93.969 -78.106 52.842
109604 93.972 -78.117 52.818
109605 93.975 -78.127 52.794
109606 93.979 -78.138 52.770
109607 93.982 -78.147 52.746
109608 93.986 -78.158 52.722
109609 93.989 -78.169 52.699
109610 93.992 -78.178 52.675
109611 93.996 -78.189 52.651
109612 93.999 -78.199 52.628
109613 94.003 -78.210 52.603
109614 94.006 -78.220 52.580
109615 94.009 -78.230 52.556
109616 94.013 -78.241 52.532
109617 94.024 -78.247 52.509
109618 94.037 -78.249 52.486
109619 94.043 -78.239 52.461
109620 94.068 -78.230 52.438
109621 94.093 -78.221 52.413
109622 94.118 -78.212 52.388
109623 94.143 -78.203 52.364
109624 94.165 -78.194 52.339
109625 94.176 -78.192 52.315
109626 94.186 -78.189 52.291
109627 94.197 -78.186 52.267
109628 94.207 -78.184 52.242
109629 94.218 -78.181 52.216
109630 94.228 -78.178 52.192
109631 94.242 -78.171 52.168
109632 94.256 -78.163 52.143
109633 94.269 -78.156 52.118
109634 94.283 -78.148 52.092
109635 94.297 -78.141 52.067
109636 94.311 -78.133 52.043
109637 94.324 -78.125 52.018
109638 94.338 -78.118 51.993
109639 94.352 -78.110 51.968
109640 94.365 -78.103 51.942
109641 94.380 -78.097 51.918
109642 94.394 -78.090 51.893
109643 94.408 -78.083 51.867
109644 94.411 -78.078 51.842
109645 94.412 -78.074 51.818
109646 94.424 -78.067 51.792
109647 94.445 -78.057 51.765
109648 94.465 -78.047 51.738
109649 94.486 -78.037 51.711
109650 94.506 -78.027 51.684
109651 94.524 -78.022 51.657
109652 94.547 -78.028 51.630
109653 94.584 -78.044 51.601
109654 94.622 -78.060 51.571
109655 94.656 -78.075 51.541
109656 94.668 -78.075 51.512
109657 94.679 -78.075 51.482
109658 94.691 -78.075 51.453
109659 94.699 -78.067 51.424
109660 94.707 -78.058 51.394
109661 94.715 -78.049 51.362
109662 94.723 -78.040 51.332
109663 94.725 -78.044 51.301
109664 94.725 -78.050 51.270
109665 94.730 -78.054 51.240
109666 94.741 -78.052 51.209
109667 94.753 -78.050 51.178
109668 94.765 -78.056 51.146
109669 94.777 -78.060 51.112
109670 94.789 -78.065 51.079
109671 94.802 -78.070 51.045
109672 94.821 -78.076 51.012
109673 94.845 -78.084 50.978
109674 94.870 -78.091 50.943
109675 94.894 -78.096 50.907
109676 94.912 -78.078 50.872
109677 94.930 -78.062 50.836
109678 94.949 -78.044 50.801
109679 94.967 -78.027 50.767
109680 94.985 -78.010 50.732
109681 95.003 -77.993 50.697
109682 95.022 -77.975 50.662
109683 95.040 -77.958 50.627
109684 95.058 -77.941 50.594
109685 95.073 -77.925 50.560
109686 95.082 -77.912 50.527
109687 95.091 -77.899 50.496
109688 95.113 -77.887 50.464
109689 95.137 -77.891 50.431
109690 95.161 -77.896 50.400
109691 95.184 -77.899 50.368
109692 95.207 -77.903 50.339
109693 95.231 -77.907 50.311
109694 95.255 -77.911 50.284
109695 95.278 -77.915 50.258
109696 95.302 -77.919 50.232
109697 95.326 -77.923 50.206
109698 95.343 -77.927 50.182
109699 95.349 -77.929 50.160
109700 95.356 -77.931 50.139
109701 95.362 -77.934 50.117
109702 95.369 -77.936 50.097
109703 95.375 -77.938 50.077
109704 95.382 -77.940 50.057
109705 95.388 -77.943 50.036
109706 95.394 -77.946 50.012
109707 95.401 -77.948 49.985
109708 95.407 -77.950 49.957
109709 95.414 -77.953 49.924
109710 95.420 -77.955 49.890
109711 95.426 -77.957 49.854
109712 95.437 -77.964 49.814
109713 95.459 -77.986 49.772
109714 95.486 -77.989 49.730
109715 95.517 -77.978 49.687
109716 95.548 -77.967 49.648
109717 95.579 -77.956 49.609
109718 95.610 -77.944 49.572
109719 95.641 -77.933 49.534
109720 95.672 -77.922 49.496
109721 95.703 -77.911 49.458
109722 95.705 -77.923 49.418
109723 95.680 -77.964 49.380
109724 95.654 -78.005 49.342
109725 95.671 -78.003 49.303
109726 95.722 -77.960 49.265
109727 95.690 -77.937 49.228
109728 95.687 -77.956 49.190
109729 95.683 -77.977 49.153
109730 95.679 -77.996 49.116
109731 95.675 -78.016 49.078
109732 95.688 -78.008 49.038
109733 95.707 -78.002 48.999
109734 95.734 -78.011 48.958
109735 95.760 -78.020 48.917
109736 95.787 -78.029 48.876
109737 95.814 -78.038 48.835
109738 95.840 -78.047 48.795
109739 95.868 -78.056 48.755
109740 95.890 -78.059 48.717
109741 95.908 -78.067 48.678
109742 95.905 -78.090 48.638
109743 95.896 -78.116 48.598
109744 95.901 -78.117 48.561
109745 95.902 -78.119 48.525
109746 95.904 -78.121 48.490
109747 95.906 -78.123 48.452
109748 95.922 -78.120 48.414
109749 95.947 -78.115 48.377
109750 95.963 -78.114 48.341
109751 95.979 -78.112 48.307
109752 95.995 -78.112 48.272
109753 96.012 -78.111 48.238
109754 96.028 -78.109 48.206
109755 96.044 -78.109 48.173
109756 96.058 -78.107 48.141
109757 96.060 -78.105 48.109
109758 96.061 -78.102 48.081
109759 96.063 -78.099 48.054
109760 96.072 -78.100 48.026
109761 96.096 -78.110 48.000
109762 96.121 -78.119 47.973
109763 96.139 -78.131 47.944
109764 96.154 -78.142 47.913
109765 96.168 -78.148 47.885
109766 96.183 -78.153 47.856
109767 96.198 -78.159 47.827
109768 96.214 -78.168 47.798
109769 96.229 -78.180 47.768
109770 96.243 -78.192 47.740
109771 96.258 -78.205 47.713
109772 96.273 -78.217 47.690
109773 96.287 -78.227 47.666
109774 96.301 -78.238 47.641
109775 96.317 -78.249 47.616
109776 96.333 -78.260 47.590
109777 96.348 -78.273 47.563
109778 96.363 -78.288 47.536
109779 96.377 -78.301 47.511
109780 96.393 -78.310 47.486
109781 96.409 -78.316 47.462
109782 96.425 -78.322 47.438
109783 96.442 -78.327 47.414
109784 96.472 -78.326 47.390
109785 96.502 -78.326 47.366
109786 96.521 -78.329 47.344
109787 96.524 -78.338 47.320
109788 96.528 -78.347 47.295
109789 96.539 -78.356 47.268
109790 96.549 -78.365 47.241
109791 96.559 -78.374 47.213
109792 96.569 -78.383 47.183
109793 96.570 -78.385 47.155
109794 96.570 -78.385 47.129
109795 96.578 -78.385 47.103
109796 96.586 -78.402 47.077
109797 96.592 -78.426 47.051
109798 96.597 -78.450 47.026
109799 96.603 -78.473 47.001
109800 96.608 -78.497 46.977
109801 96.614 -78.520 46.953
109802 96.622 -78.533 46.928
109803 96.630 -78.545 46.903
109804 96.624 -78.567 46.878
109805 96.619 -78.588 46.853
109806 96.614 -78.610 46.827
109807 96.609 -78.632 46.804
109808 96.603 -78.654 46.778
109809 96.600 -78.677 46.752
109810 96.619 -78.708 46.729
109811 96.652 -78.721 46.703
109812 96.687 -78.731 46.677
109813 96.725 -78.742 46.650
109814 96.765 -78.752 46.624
109815 96.806 -78.764 46.597
109816 96.842 -78.777 46.569
109817 96.871 -78.795 46.542
109818 96.901 -78.812 46.513
109819 96.930 -78.830 46.485
109820 96.960 -78.848 46.458
109821 96.989 -78.865 46.432
109822 97.019 -78.883 46.407
109823 97.073 -78.893 46.381
109824 97.142 -78.899 46.356
109825 97.185 -78.892 46.334
109826 97.228 -78.883 46.311
109827 97.271 -78.875 46.289
109828 97.314 -78.868 46.271
109829 97.356 -78.859 46.251
109830 97.391 -78.848 46.232
109831 97.425 -78.836 46.213
109832 97.460 -78.825 46.195
109833 97.495 -78.814 46.179
109834 97.530 -78.805 46.162
109835 97.569 -78.804 46.146
109836 97.607 -78.803 46.130
109837 97.646 -78.802 46.115
109838 97.699 -78.798 46.099
109839 97.691 -78.855 46.084
109840 97.684 -78.912 46.069
109841 97.655 -78.890 46.054
109842 97.648 -78.854 46.039
109843 97.698 -78.852 46.024
109844 97.749 -78.853 46.009
109845 97.800 -78.853 45.994
109846 97.821 -78.814 45.979
109847 97.845 -78.783 45.965
109848 97.882 -78.779 45.949
109849 97.918 -78.774 45.932
109850 97.954 -78.770 45.915
109851 97.975 -78.729 45.898
109852 97.993 -78.713 45.880
109853 97.999 -78.702 45.865
109854 97.979 -78.689 45.848
109855 97.946 -78.688 45.831
109856 97.953 -78.678 45.816
109857 97.947 -78.705 45.800
109858 97.935 -78.723 45.785
109859 97.921 -78.686 45.770
109860 97.932 -78.657 45.755
109861 97.984 -78.639 45.741
109862 97.964 -78.720 45.726
109863 97.964 -78.741 45.712
109864 97.965 -78.761 45.698
109865 97.971 -78.777 45.684
109866 98.000 -78.780 45.670
109867 98.029 -78.783 45.655
109868 98.059 -78.786 45.640
109869 98.088 -78.789 45.625
109870 98.117 -78.793 45.611
109871 98.143 -78.801 45.596
109872 98.164 -78.815 45.582
109873 98.183 -78.827 45.567
109874 98.189 -78.815 45.553
109875 98.195 -78.803 45.541
109876 98.201 -78.792 45.531
109877 98.207 -78.789 45.521
109878 98.214 -78.805 45.507
109879 98.222 -78.820 45.494
109880 98.232 -78.836 45.481
109881 98.262 -78.896 45.469
109882 98.281 -78.961 45.454
109883 98.302 -79.015 45.441
109884 98.325 -78.986 45.428
109885 98.346 -78.943 45.414
109886 98.366 -78.900 45.401
109887 98.354 -78.913 45.388
109888 98.336 -78.936 45.376
109889 98.318 -78.959 45.363
109890 98.358 -78.939 45.351
109891 98.416 -78.930 45.339
109892 98.476 -78.948 45.329
109893 98.499 -78.951 45.317
109894 98.522 -78.953 45.304
109895 98.545 -78.955 45.292
109896 98.568 -78.959 45.279
109897 98.591 -78.961 45.265
109898 98.614 -78.967 45.251
109899 98.636 -78.974 45.237
109900 98.658 -78.980 45.223
109901 98.680 -78.987 45.209
109902 98.702 -78.993 45.194
109903 98.724 -78.999 45.181
109904 98.746 -79.006 45.169
109905 98.769 -79.013 45.156
109906 98.791 -79.020 45.144
109907 98.813 -79.026 45.132
109908 98.836 -79.033 45.121
109909 98.883 -79.062 45.109
109910 98.938 -79.053 45.100
109911 98.973 -79.042 45.091
109912 98.997 -79.035 45.081
109913 99.021 -79.028 45.072
109914 99.046 -79.023 45.062
109915 99.071 -79.024 45.051
109916 99.097 -79.027 45.038
109917 99.122 -79.028 45.026
109918 99.147 -79.030 45.014
109919 99.171 -79.033 45.005
109920 99.185 -79.043 44.997
109921 99.199 -79.054 44.987
109922 99.206 -79.064 44.976
109923 99.212 -79.074 44.966
109924 99.217 -79.084 44.955
109925 99.271 -79.090 44.945
109926 99.265 -79.109 44.936
109927 99.268 -79.118 44.928
109928 99.271 -79.127 44.920
109929 99.274 -79.135 44.912
109930 99.300 -79.135 44.906
109931 99.317 -79.167 44.898
109932 99.331 -79.206 44.892
109933 99.400 -79.211 44.888
109934 99.481 -79.198 44.884
109935 99.562 -79.185 44.881
109936 99.525 -79.184 44.879
109937 99.479 -79.189 44.877
109938 99.492 -79.212 44.876
109939 99.514 -79.239 44.874
109940 99.537 -79.266 44.872
109941 99.560 -79.294 44.871
109942 99.583 -79.321 44.872
109943 99.607 -79.341 44.872
109944 99.632 -79.346 44.871
109945 99.658 -79.350 44.868
109946 99.684 -79.356 44.863
109947 99.710 -79.360 44.858
109948 99.736 -79.365 44.852
109949 99.762 -79.369 44.845
109950 99.788 -79.374 44.838
109951 99.821 -79.404 44.831
109952 99.855 -79.438 44.823
109953 99.890 -79.470 44.815
109954 99.923 -79.503 44.807
109955 99.958 -79.535 44.798
109956 99.991 -79.568 44.790
109957 100.025 -79.606 44.782
109958 100.061 -79.641 44.773
109959 100.100 -79.651 44.765
109960 100.139 -79.661 44.757
109961 100.178 -79.670 44.747
109962 100.217 -79.679 44.737
109963 100.256 -79.689 44.726
109964 100.288 -79.707 44.714
109965 100.312 -79.746 44.702
109966 100.336 -79.785 44.689
109967 100.358 -79.824 44.678
109968 100.381 -79.862 44.666
109969 100.403 -79.902 44.653
109970 100.426 -79.937 44.641
109971 100.463 -79.943 44.628
109972 100.500 -79.947 44.615
109973 100.538 -79.946 44.601
109974 100.575 -79.946 44.586
109975 100.613 -79.945 44.570
109976 100.650 -79.944 44.555
109977 100.687 -79.943 44.541
109978 100.724 -79.943 44.526
109979 100.759 -79.968 44.512
109980 100.792 -80.004 44.499
109981 100.825 -80.039 44.484
109982 100.854 -80.071 44.469
109983 100.886 -80.104 44.456
109984 100.919 -80.138 44.441
109985 100.952 -80.171 44.427
109986 100.984 -80.206 44.412
109987 101.018 -80.239 44.397
109988 101.050 -80.273 44.382
109989 101.084 -80.306 44.366
109990 101.116 -80.341 44.351
109991 101.149 -80.375 44.335
109992 101.182 -80.408 44.319
109993 101.218 -80.428 44.304
109994 101.264 -80.441 44.287
109995 101.309 -80.462 44.271
109996 101.341 -80.478 44.255
109997 101.374 -80.494 44.239
109998 101.405 -80.510 44.222
109999 101.438 -80.526 44.205
110000 101.470 -80.543 44.189
110001 101.502 -80.559 44.171
110002 101.534 -80.575 44.154
110003 101.566 -80.591 44.137
110004 101.598 -80.607 44.120
110005 101.630 -80.623 44.104
110006 101.662 -80.640 44.086
110007 101.694 -80.656 44.069
110266 93.522 -78.365 53.712
110267 93.520 -78.356 53.691
110268 93.519 -78.346 53.668
110269 93.530 -78.331 53.645
110270 93.540 -78.316 53.622
110271 93.552 -78.300 53.597
110272 93.575 -78.285 53.573
110273 93.597 -78.270 53.549
110274 93.611 -78.260 53.526
110275 93.623 -78.252 53.502
110276 93.634 -78.244 53.479
110277 93.646 -78.235 53.455
110278 93.657 -78.227 53.431
110302 93.968 -78.086 52.878
110303 93.966 -78.085 52.854
110304 93.968 -78.094 52.830
110305 93.971 -78.103 52.806
110306 93.975 -78.114 52.782
110307 93.978 -78.124 52.757
110308 93.982 -78.134 52.733
110309 93.985 -78.144 52.709
110310 93.988 -78.155 52.685
110311 93.992 -78.166 52.661
110312 93.995 -78.175 52.637
110313 93.999 -78.186 52.613
110314 94.002 -78.196 52.589
110315 94.005 -78.206 52.566
110316 94.009 -78.217 52.542
110317 94.012 -78.227 52.517
110318 94.028 -78.232 52.494
110319 94.039 -78.232 52.470
110320 94.045 -78.222 52.445
110321 94.069 -78.213 52.422
110322 94.095 -78.204 52.398
110323 94.119 -78.195 52.373
110324 94.144 -78.186 52.349
110325 94.169 -78.176 52.324
110326 94.186 -78.171 52.299
110327 94.196 -78.169 52.274
110328 94.207 -78.166 52.250
110329 94.217 -78.163 52.225
110330 94.228 -78.160 52.200
110331 94.239 -78.157 52.176
110332 94.251 -78.152 52.152
110333 94.265 -78.144 52.127
110334 94.278 -78.137 52.101
110335 94.292 -78.129 52.076
110336 94.306 -78.122 52.051
110337 94.320 -78.114 52.027
110338 94.333 -78.106 52.002
110339 94.347 -78.099 51.978
110340 94.361 -78.091 51.953
110341 94.374 -78.084 51.928
110342 94.389 -78.078 51.904
110343 94.403 -78.071 51.879
110344 94.417 -78.064 51.854
110345 94.424 -78.059 51.830
110346 94.425 -78.055 51.807
110347 94.426 -78.050 51.782
110348 94.447 -78.040 51.756
110349 94.468 -78.030 51.729
110350 94.488 -78.020 51.703
110351 94.509 -78.010 51.677
110352 94.528 -78.003 51.650
110353 94.543 -78.004 51.623
110354 94.580 -78.019 51.595
110355 94.618 -78.036 51.565
110356 94.654 -78.052 51.536
110357 94.666 -78.052 51.507
110358 94.678 -78.053 51.478
110359 94.689 -78.053 51.449
110360 94.699 -78.048 51.421
110361 94.707 -78.040 51.391
110362 94.715 -78.031 51.360
110363 94.723 -78.022 51.330
110364 94.726 -78.024 51.300
110365 94.726 -78.031 51.269
110366 94.727 -78.037 51.239
110367 94.739 -78.034 51.208
110368 94.750 -78.033 51.177
110369 94.762 -78.038 51.145
110370 94.775 -78.043 51.112
110371 94.787 -78.048 51.079
110372 94.799 -78.053 51.046
110373 94.823 -78.060 51.013
110374 94.848 -78.067 50.979
110375 94.873 -78.075 50.943
110376 94.896 -78.078 50.908
110377 94.914 -78.060 50.873
110378 94.932 -78.043 50.837
110379 94.951 -78.026 50.802
110380 94.969 -78.009 50.768
110381 94.987 -77.991 50.733
110382 95.005 -77.975 50.698
110383 95.024 -77.957 50.664
110384 95.042 -77.940 50.629
110385 95.058 -77.923 50.596
110386 95.067 -77.910 50.563
110387 95.076 -77.897 50.530
110388 95.085 -77.884 50.499
110389 95.099 -77.863 50.467
110390 95.124 -77.862 50.435
110391 95.148 -77.866 50.405
110392 95.172 -77.870 50.374
110393 95.195 -77.874 50.346
110394 95.219 -77.877 50.318
110395 95.243 -77.882 50.291
110396 95.266 -77.886 50.263
110397 95.293 -77.899 50.236
110398 95.305 -77.918 50.210
110399 95.341 -77.915 50.186
110400 95.358 -77.914 50.163
110401 95.365 -77.916 50.143
110402 95.371 -77.918 50.123
110403 95.378 -77.921 50.103
110404 95.384 -77.923 50.083
110405 95.390 -77.925 50.062
110406 95.397 -77.928 50.038
110407 95.403 -77.930 50.012
110408 95.409 -77.933 49.983
110409 95.416 -77.935 49.952
110410 95.422 -77.937 49.918
110411 95.429 -77.940 49.881
110412 95.435 -77.942 49.843
110413 95.445 -77.948 49.802
110414 95.467 -77.970 49.761
110415 95.488 -77.992 49.719
110416 95.518 -77.987 49.680
110417 95.549 -77.976 49.643
110418 95.580 -77.965 49.606
110419 95.611 -77.954 49.568
110420 95.642 -77.943 49.531
110421 95.667 -77.928 49.494
110422 95.686 -77.911 49.457
110423 95.697 -77.902 49.420
110424 95.672 -77.943 49.383
110425 95.646 -77.984 49.347
110426 95.652 -77.997 49.310
110427 95.682 -77.984 49.272
110428 95.733 -77.940 49.234
110429 95.687 -77.948 49.195
110430 95.683 -77.968 49.156
110431 95.679 -77.988 49.119
110432 95.679 -78.000 49.081
110433 95.694 -77.987 49.043
110434 95.722 -77.996 49.003
110435 95.748 -78.005 48.964
110436 95.775 -78.014 48.925
110437 95.802 -78.023 48.884
110438 95.828 -78.033 48.843
110439 95.851 -78.036 48.802
110440 95.873 -78.038 48.761
110441 95.895 -78.041 48.722
110442 95.905 -78.056 48.683
110443 95.903 -78.080 48.644
110444 95.895 -78.103 48.605
110445 95.902 -78.102 48.568
110446 95.909 -78.100 48.531
110447 95.915 -78.098 48.496
110448 95.922 -78.097 48.459
110449 95.933 -78.096 48.421
110450 95.949 -78.095 48.384
110451 95.965 -78.094 48.347
110452 95.981 -78.093 48.312
110453 95.998 -78.092 48.277
110454 96.014 -78.091 48.244
110455 96.030 -78.090 48.212
110456 96.046 -78.089 48.179
110457 96.062 -78.087 48.148
110458 96.077 -78.087 48.117
110459 96.079 -78.084 48.090
110460 96.080 -78.081 48.063
110461 96.087 -78.081 48.035
110462 96.111 -78.091 48.009
110463 96.134 -78.100 47.981
110464 96.148 -78.113 47.952
110465 96.162 -78.127 47.921
110466 96.176 -78.138 47.893
110467 96.191 -78.144 47.865
110468 96.206 -78.150 47.836
110469 96.222 -78.156 47.809
110470 96.236 -78.169 47.781
110471 96.251 -78.181 47.754
110472 96.266 -78.193 47.728
110473 96.281 -78.204 47.705
110474 96.293 -78.214 47.681
110475 96.305 -78.223 47.657
110476 96.319 -78.234 47.634
110477 96.336 -78.245 47.609
110478 96.352 -78.257 47.582
110479 96.368 -78.268 47.556
110480 96.385 -78.279 47.531
110481 96.403 -78.287 47.506
110482 96.419 -78.291 47.482
110483 96.436 -78.297 47.458
110484 96.452 -78.303 47.435
110485 96.469 -78.308 47.412
110486 96.500 -78.307 47.389
110487 96.505 -78.316 47.365
110488 96.514 -78.324 47.341
110489 96.524 -78.333 47.315
110490 96.534 -78.342 47.288
110491 96.545 -78.351 47.258
110492 96.555 -78.360 47.229
110493 96.557 -78.363 47.201
110494 96.556 -78.364 47.175
110495 96.556 -78.365 47.149
110496 96.571 -78.364 47.124
110497 96.577 -78.385 47.099
110498 96.583 -78.409 47.072
110499 96.588 -78.433 47.047
110500 96.594 -78.457 47.022
110501 96.599 -78.480 46.999
110502 96.605 -78.504 46.974
110503 96.610 -78.527 46.948
110504 96.615 -78.545 46.921
110505 96.609 -78.567 46.896
110506 96.604 -78.588 46.871
110507 96.599 -78.610 46.846
110508 96.595 -78.631 46.821
110509 96.592 -78.650 46.794
110510 96.589 -78.670 46.768
110511 96.607 -78.685 46.744
110512 96.648 -78.695 46.717
110513 96.688 -78.706 46.690
110514 96.729 -78.717 46.662
110515 96.769 -78.728 46.635
110516 96.810 -78.739 46.607
110517 96.847 -78.752 46.579
110518 96.876 -78.770 46.551
110519 96.906 -78.787 46.522
110520 96.935 -78.805 46.494
110521 96.965 -78.823 46.466
110522 96.994 -78.840 46.440
110523 97.021 -78.852 46.414
110524 97.066 -78.862 46.389
110525 97.116 -78.872 46.364
110526 97.184 -78.879 46.342
110527 97.228 -78.871 46.319
110528 97.271 -78.863 46.297
110529 97.314 -78.855 46.278
110530 97.351 -78.845 46.258
110531 97.386 -78.833 46.239
110532 97.420 -78.822 46.221
110533 97.456 -78.814 46.203
110534 97.495 -78.814 46.186
110535 97.534 -78.813 46.170
110536 97.572 -78.812 46.155
110537 97.611 -78.811 46.139
110538 97.650 -78.811 46.123
110539 97.691 -78.808 46.107
110540 97.694 -78.861 46.092
110541 97.675 -78.896 46.078
110542 97.656 -78.888 46.063
110543 97.676 -78.858 46.048
110544 97.714 -78.842 46.033
110545 97.765 -78.843 46.019
110546 97.800 -78.821 46.004
110547 97.821 -78.782 45.990
110548 97.842 -78.742 45.975
110549 97.877 -78.734 45.960
110550 97.913 -78.730 45.943
110551 97.949 -78.725 45.926
110552 97.986 -78.721 45.909
110553 98.022 -78.717 45.892
110554 98.015 -78.705 45.876
110555 97.982 -78.705 45.859
110556 97.981 -78.686 45.843
110557 97.981 -78.698 45.827
110558 97.974 -78.724 45.811
110559 97.959 -78.728 45.795
110560 97.941 -78.689 45.780
110561 97.953 -78.661 45.765
110562 97.986 -78.638 45.750
110563 97.984 -78.696 45.735
110564 97.984 -78.717 45.721
110565 97.985 -78.736 45.706
110566 97.985 -78.757 45.692
110567 98.001 -78.767 45.677
110568 98.030 -78.771 45.662
110569 98.060 -78.774 45.646
110570 98.089 -78.777 45.632
110571 98.119 -78.780 45.618
110572 98.140 -78.794 45.602
110573 98.161 -78.808 45.587
110574 98.182 -78.824 45.573
110575 98.189 -78.814 45.561
110576 98.195 -78.802 45.549
110577 98.200 -78.789 45.538
110578 98.206 -78.777 45.527
110579 98.211 -78.767 45.514
110580 98.219 -78.782 45.502
110581 98.246 -78.811 45.489
110582 98.266 -78.876 45.476
110583 98.262 -78.913 45.461
110584 98.285 -78.899 45.447
110585 98.306 -78.869 45.434
110586 98.327 -78.826 45.420
110587 98.314 -78.839 45.407
110588 98.296 -78.862 45.394
110589 98.283 -78.886 45.381
110590 98.302 -78.915 45.369
110591 98.313 -78.918 45.357
110592 98.334 -78.929 45.344
110593 98.428 -78.957 45.332
110594 98.479 -78.974 45.321
110595 98.502 -78.976 45.309
110596 98.524 -78.979 45.299
110597 98.548 -78.981 45.287
110598 98.570 -78.984 45.274
110599 98.593 -78.987 45.261
110600 98.617 -78.989 45.247
110601 98.639 -78.992 45.233
110602 98.662 -78.994 45.218
110603 98.686 -78.997 45.203
110604 98.708 -78.999 45.188
110605 98.731 -79.002 45.173
110606 98.755 -79.005 45.159
110607 98.777 -79.007 45.145
110608 98.800 -79.011 45.133
110609 98.822 -79.017 45.121
110610 98.873 -79.013 45.110
110611 98.929 -79.006 45.100
110612 98.975 -79.004 45.091
110613 99.000 -79.006 45.082
110614 99.025 -79.007 45.073
110615 99.051 -79.009 45.064
110616 99.076 -79.011 45.054
110617 99.101 -79.012 45.043
110618 99.127 -79.015 45.030
110619 99.152 -79.016 45.020
110620 99.168 -79.024 45.011
110621 99.182 -79.035 45.003
110622 99.195 -79.046 44.992
110623 99.209 -79.056 44.981
110624 99.215 -79.066 44.970
110625 99.226 -79.073 44.960
110626 99.246 -79.075 44.949
110627 99.273 -79.077 44.938
110628 99.236 -79.116 44.930
110629 99.259 -79.118 44.922
110630 99.299 -79.112 44.915
110631 99.340 -79.106 44.909
110632 99.359 -79.137 44.902
110633 99.428 -79.142 44.896
110634 99.497 -79.147 44.891
110635 99.565 -79.153 44.888
110636 99.612 -79.169 44.886
110637 99.562 -79.171 44.884
110638 99.505 -79.173 44.881
110639 99.505 -79.190 44.879
110640 99.513 -79.211 44.878
110641 99.536 -79.238 44.878
110642 99.559 -79.266 44.878
110643 99.582 -79.293 44.877
110644 99.605 -79.320 44.875
110645 99.630 -79.330 44.872
110646 99.656 -79.335 44.868
110647 99.682 -79.339 44.862
110648 99.707 -79.344 44.856
110649 99.733 -79.348 44.850
110650 99.759 -79.353 44.842
110651 99.789 -79.370 44.834
110652 99.823 -79.404 44.826
110653 99.857 -79.436 44.819
110654 99.891 -79.469 44.811
110655 99.925 -79.501 44.803
110656 99.959 -79.534 44.795
110657 99.993 -79.567 44.787
110658 100.027 -79.604 44.779
110659 100.063 -79.635 44.771
110660 100.102 -79.645 44.763
110661 100.141 -79.654 44.755
110662 100.181 -79.664 44.745
110663 100.219 -79.673 44.736
110664 100.254 -79.683 44.726
110665 100.286 -79.695 44.715
110666 100.311 -79.733 44.703
110667 100.334 -79.773 44.691
110668 100.358 -79.811 44.679
110669 100.382 -79.850 44.666
110670 100.406 -79.889 44.651
110671 100.428 -79.920 44.638
110672 100.470 -79.923 44.625
110673 100.507 -79.922 44.611
110674 100.545 -79.921 44.598
110675 100.582 -79.921 44.584
110676 100.620 -79.920 44.570
110677 100.657 -79.919 44.555
110678 100.694 -79.918 44.540
110679 100.732 -79.918 44.526
110680 100.766 -79.946 44.512
110681 100.799 -79.980 44.498
110682 100.832 -80.015 44.484
110683 100.865 -80.051 44.469
110684 100.898 -80.086 44.456
110685 100.925 -80.117 44.442
110686 100.956 -80.149 44.428
110687 100.989 -80.183 44.414
110688 101.022 -80.217 44.400
110689 101.055 -80.250 44.385
110690 101.088 -80.284 44.369
110691 101.120 -80.318 44.355
110692 101.153 -80.352 44.339
110693 101.185 -80.375 44.323
110694 101.231 -80.388 44.309
110695 101.277 -80.402 44.293
110696 101.326 -80.424 44.278
110697 101.358 -80.440 44.261
110698 101.391 -80.456 44.245
110699 101.423 -80.472 44.229
110700 101.455 -80.488 44.212
110701 101.487 -80.504 44.195
110702 101.519 -80.521 44.177
110703 101.551 -80.537 44.160
110704 101.583 -80.553 44.144
110705 101.615 -80.569 44.127
110706 101.647 -80.585 44.110
110707 101.679 -80.601 44.092
110708 101.711 -80.617 44.075
111023 94.096 -78.188 52.385
111024 94.121 -78.178 52.360
111025 94.146 -78.169 52.335
111026 94.171 -78.159 52.311
111027 94.196 -78.150 52.286
111028 94.207 -78.147 52.261
111029 94.217 -78.145 52.236
111030 94.227 -78.142 52.211
111031 94.238 -78.139 52.186
111032 94.249 -78.137 52.162
111033 94.260 -78.132 52.138
111034 94.274 -78.125 52.113
111035 94.287 -78.117 52.087
111036 94.301 -78.109 52.063
111037 94.315 -78.102 52.039
111038 94.329 -78.094 52.014
111039 94.342 -78.087 51.990
111040 94.356 -78.079 51.967
111041 94.370 -78.072 51.942
111042 94.384 -78.065 51.918
111043 94.398 -78.058 51.894
111044 94.412 -78.052 51.870
111045 94.426 -78.045 51.846
111046 94.436 -78.039 51.823
111047 94.437 -78.035 51.800
111048 94.438 -78.031 51.775
111049 94.450 -78.023 51.750
111050 94.470 -78.013 51.723
111051 94.491 -78.003 51.698
111052 94.511 -77.993 51.672
111053 94.532 -77.984 51.645
111054 94.547 -77.985 51.619
111055 94.576 -77.995 51.591
111056 94.613 -78.011 51.562
111057 94.651 -78.028 51.534
111058 94.665 -78.029 51.505
111059 94.676 -78.030 51.476
111060 94.688 -78.031 51.447
111061 94.699 -78.030 51.420
111062 94.707 -78.021 51.390
111063 94.715 -78.012 51.360
111064 94.723 -78.003 51.330
111065 94.726 -78.004 51.299
111066 94.726 -78.011 51.269
111067 94.726 -78.018 51.239
111068 94.736 -78.017 51.208
111069 94.748 -78.015 51.177
111070 94.760 -78.021 51.145
111071 94.772 -78.025 51.111
111072 94.784 -78.031 51.079
111073 94.802 -78.036 51.046
111074 94.826 -78.044 51.013
111075 94.851 -78.050 50.979
111076 94.876 -78.058 50.944
111077 94.898 -78.059 50.909
111078 94.917 -78.042 50.874
111079 94.935 -78.025 50.839
111080 94.953 -78.007 50.804
111081 94.971 -77.990 50.769
111082 94.989 -77.973 50.735
111083 95.007 -77.956 50.701
111084 95.026 -77.939 50.666
111085 95.043 -77.921 50.632
111086 95.052 -77.909 50.598
111087 95.061 -77.895 50.565
111088 95.070 -77.882 50.532
111089 95.078 -77.868 50.501
111090 95.088 -77.855 50.469
111091 95.112 -77.833 50.438
111092 95.136 -77.837 50.408
111093 95.159 -77.840 50.379
111094 95.183 -77.845 50.351
111095 95.206 -77.849 50.324
111096 95.230 -77.852 50.296
111097 95.268 -77.902 50.266
111098 95.270 -77.912 50.237
111099 95.291 -77.915 50.210
111100 95.328 -77.908 50.186
111101 95.357 -77.902 50.163
111102 95.371 -77.902 50.141
111103 95.380 -77.903 50.122
111104 95.386 -77.906 50.102
111105 95.392 -77.908 50.081
111106 95.399 -77.910 50.059
111107 95.405 -77.912 50.034
111108 95.412 -77.915 50.006
111109 95.418 -77.917 49.976
111110 95.424 -77.920 49.942
111111 95.431 -77.922 49.906
111112 95.438 -77.924 49.868
111113 95.444 -77.927 49.829
111114 95.453 -77.932 49.789
111115 95.475 -77.954 49.750
111116 95.496 -77.976 49.711
111117 95.519 -77.997 49.675
111118 95.550 -77.986 49.640
111119 95.581 -77.975 49.605
111120 95.607 -77.961 49.568
111121 95.626 -77.943 49.531
111122 95.645 -77.925 49.494
111123 95.664 -77.908 49.457
111124 95.683 -77.890 49.421
111125 95.664 -77.923 49.385
111126 95.638 -77.964 49.349
111127 95.643 -77.978 49.313
111128 95.659 -77.983 49.276
111129 95.692 -77.963 49.238
111130 95.686 -77.940 49.201
111131 95.683 -77.959 49.162
111132 95.679 -77.979 49.124
111133 95.686 -77.980 49.085
111134 95.709 -77.981 49.048
111135 95.736 -77.990 49.010
111136 95.763 -78.000 48.972
111137 95.790 -78.009 48.933
111138 95.813 -78.012 48.892
111139 95.835 -78.015 48.851
111140 95.856 -78.018 48.810
111141 95.879 -78.020 48.769
111142 95.894 -78.030 48.730
111143 95.903 -78.047 48.691
111144 95.900 -78.070 48.651
111145 95.895 -78.090 48.612
111146 95.902 -78.087 48.575
111147 95.909 -78.086 48.539
111148 95.915 -78.084 48.502
111149 95.922 -78.082 48.466
111150 95.936 -78.082 48.429
111151 95.952 -78.079 48.392
111152 95.968 -78.076 48.354
111153 95.984 -78.073 48.320
111154 96.000 -78.072 48.286
111155 96.016 -78.071 48.253
111156 96.032 -78.070 48.220
111157 96.048 -78.069 48.189
111158 96.065 -78.068 48.158
111159 96.081 -78.067 48.128
111160 96.096 -78.065 48.100
111161 96.097 -78.063 48.073
111162 96.101 -78.061 48.045
111163 96.125 -78.071 48.019
111164 96.143 -78.082 47.991
111165 96.157 -78.096 47.962
111166 96.171 -78.109 47.932
111167 96.185 -78.122 47.905
111168 96.199 -78.134 47.877
111169 96.215 -78.141 47.849
111170 96.229 -78.146 47.822
111171 96.244 -78.156 47.795
111172 96.259 -78.169 47.768
111173 96.274 -78.181 47.743
111174 96.287 -78.191 47.720
111175 96.299 -78.200 47.696
111176 96.312 -78.210 47.673
111177 96.324 -78.219 47.651
111178 96.338 -78.229 47.627
111179 96.355 -78.241 47.601
111180 96.374 -78.253 47.574
111181 96.393 -78.264 47.550
111182 96.412 -78.272 47.524
111183 96.429 -78.276 47.499
111184 96.445 -78.279 47.475
111185 96.461 -78.283 47.453
111186 96.478 -78.287 47.430
111187 96.490 -78.293 47.407
111188 96.500 -78.301 47.383
111189 96.510 -78.310 47.358
111190 96.520 -78.319 47.332
111191 96.530 -78.329 47.305
111192 96.540 -78.338 47.276
111193 96.544 -78.341 47.248
111194 96.543 -78.342 47.220
111195 96.543 -78.344 47.195
111196 96.547 -78.344 47.169
111197 96.562 -78.345 47.144
111198 96.568 -78.369 47.119
111199 96.573 -78.392 47.092
111200 96.579 -78.416 47.067
111201 96.584 -78.439 47.042
111202 96.590 -78.463 47.017
111203 96.595 -78.487 46.992
111204 96.603 -78.507 46.965
111205 96.613 -78.530 46.937
111206 96.613 -78.548 46.911
111207 96.610 -78.567 46.885
111208 96.607 -78.586 46.859
111209 96.604 -78.606 46.834
111210 96.601 -78.626 46.807
111211 96.599 -78.645 46.781
111212 96.621 -78.659 46.756
111213 96.652 -78.670 46.728
111214 96.693 -78.682 46.700
111215 96.734 -78.692 46.672
111216 96.774 -78.703 46.644
111217 96.815 -78.714 46.616
111218 96.852 -78.727 46.588
111219 96.882 -78.745 46.560
111220 96.911 -78.762 46.531
111221 96.941 -78.780 46.503
111222 96.970 -78.798 46.475
111223 96.997 -78.809 46.448
111224 97.024 -78.820 46.423
111225 97.068 -78.830 46.397
111226 97.117 -78.841 46.373
111227 97.168 -78.851 46.350
111228 97.228 -78.858 46.327
111229 97.271 -78.851 46.305
111230 97.311 -78.842 46.285
111231 97.346 -78.830 46.265
111232 97.383 -78.824 46.246
111233 97.421 -78.824 46.228
111234 97.460 -78.822 46.211
111235 97.499 -78.821 46.194
111236 97.537 -78.821 46.178
111237 97.576 -78.820 46.164
111238 97.614 -78.819 46.148
111239 97.653 -78.818 46.133
111240 97.692 -78.818 46.117
111241 97.696 -78.867 46.102
111242 97.674 -78.865 46.088
111243 97.676 -78.854 46.074
111244 97.704 -78.861 46.059
111245 97.731 -78.831 46.044
111246 97.780 -78.830 46.029
111247 97.801 -78.789 46.015
111248 97.822 -78.750 46.001
111249 97.843 -78.711 45.986
111250 97.870 -78.690 45.971
111251 97.889 -78.696 45.956
111252 97.911 -78.702 45.938
111253 97.935 -78.709 45.921
111254 97.959 -78.715 45.904
111255 97.986 -78.713 45.888
111256 98.001 -78.711 45.871
111257 98.014 -78.691 45.855
111258 98.008 -78.717 45.839
111259 98.001 -78.743 45.822
111260 97.984 -78.733 45.806
111261 97.962 -78.693 45.791
111262 97.973 -78.664 45.776
111263 97.987 -78.636 45.760
111264 98.005 -78.670 45.745
111265 98.004 -78.692 45.730
111266 98.004 -78.713 45.717
111267 98.005 -78.733 45.702
111268 98.005 -78.753 45.687
111269 98.032 -78.758 45.672
111270 98.061 -78.760 45.656
111271 98.090 -78.763 45.641
111272 98.115 -78.772 45.626
111273 98.137 -78.787 45.611
111274 98.158 -78.802 45.596
111275 98.179 -78.817 45.582
111276 98.189 -78.811 45.569
111277 98.195 -78.800 45.557
111278 98.200 -78.788 45.545
111279 98.206 -78.776 45.534
111280 98.211 -78.764 45.522
111281 98.217 -78.753 45.510
111282 98.235 -78.790 45.498
111283 98.228 -78.821 45.484
111284 98.249 -78.800 45.469
111285 98.269 -78.774 45.455
111286 98.288 -78.749 45.441
111287 98.307 -78.772 45.427
111288 98.325 -78.801 45.413
111289 98.344 -78.830 45.400
111290 98.363 -78.859 45.387
111291 98.382 -78.889 45.374
111292 98.391 -78.883 45.362
111293 98.376 -78.888 45.349
111294 98.381 -78.927 45.337
111295 98.446 -78.990 45.325
111296 98.481 -78.999 45.314
111297 98.504 -79.002 45.304
111298 98.527 -79.004 45.292
111299 98.550 -79.007 45.280
111300 98.573 -79.009 45.268
111301 98.596 -79.012 45.255
111302 98.619 -79.015 45.241
111303 98.642 -79.017 45.227
111304 98.665 -79.020 45.213
111305 98.688 -79.022 45.196
111306 98.711 -79.025 45.180
111307 98.734 -79.028 45.163
111308 98.757 -79.030 45.148
111309 98.797 -79.026 45.135
111310 98.833 -79.006 45.122
111311 98.878 -78.988 45.111
111312 98.934 -78.981 45.101
111313 98.979 -78.980 45.091
111314 99.004 -78.987 45.082
111315 99.029 -78.993 45.074
111316 99.055 -78.996 45.065
111317 99.080 -78.997 45.055
111318 99.106 -78.999 45.045
111319 99.131 -79.001 45.033
111320 99.151 -79.006 45.023
111321 99.165 -79.017 45.015
111322 99.179 -79.028 45.006
111323 99.192 -79.038 44.995
111324 99.205 -79.049 44.983
111325 99.217 -79.059 44.972
111326 99.237 -79.060 44.960
111327 99.256 -79.061 44.950
111328 99.274 -79.062 44.940
111329 99.282 -79.079 44.931
111330 99.309 -79.082 44.923
111331 99.345 -79.079 44.916
111332 99.375 -79.081 44.909
111333 99.415 -79.090 44.904
111334 99.462 -79.105 44.899
111335 99.508 -79.120 44.895
111336 99.554 -79.134 44.892
111337 99.604 -79.159 44.891
111338 99.599 -79.157 44.889
111339 99.542 -79.159 44.887
111340 99.525 -79.172 44.885
111341 99.531 -79.191 44.884
111342 99.537 -79.211 44.884
111343 99.558 -79.237 44.883
111344 99.581 -79.264 44.880
111345 99.604 -79.291 44.876
111346 99.627 -79.313 44.871
111347 99.653 -79.318 44.866
111348 99.679 -79.322 44.860
111349 99.705 -79.327 44.853
111350 99.731 -79.332 44.846
111351 99.757 -79.337 44.839
111352 99.791 -79.369 44.830
111353 99.825 -79.402 44.822
111354 99.859 -79.435 44.815
111355 99.893 -79.467 44.807
111356 99.927 -79.500 44.799
111357 99.961 -79.533 44.791
111358 99.995 -79.566 44.784
111359 100.029 -79.601 44.776
111360 100.065 -79.629 44.768
111361 100.105 -79.638 44.760
111362 100.144 -79.648 44.751
111363 100.183 -79.657 44.742
111364 100.216 -79.667 44.732
111365 100.250 -79.677 44.724
111366 100.283 -79.687 44.714
111367 100.309 -79.721 44.704
111368 100.333 -79.760 44.691
111369 100.358 -79.797 44.678
111370 100.387 -79.830 44.665
111371 100.415 -79.862 44.650
111372 100.420 -79.889 44.636
111373 100.477 -79.898 44.623
111374 100.514 -79.897 44.609
111375 100.552 -79.896 44.597
111376 100.590 -79.896 44.583
111377 100.627 -79.895 44.570
111378 100.664 -79.894 44.556
111379 100.701 -79.893 44.541
111380 100.739 -79.892 44.526
111381 100.772 -79.922 44.512
111382 100.805 -79.957 44.498
111383 100.839 -79.993 44.484
111384 100.872 -80.027 44.471
111385 100.905 -80.062 44.457
111386 100.938 -80.097 44.443
111387 100.971 -80.133 44.430
111388 100.996 -80.162 44.417
111389 101.026 -80.194 44.404
111390 101.059 -80.228 44.388
111391 101.092 -80.262 44.373
111392 101.125 -80.296 44.358
111393 101.151 -80.322 44.343
111394 101.198 -80.334 44.328
111395 101.244 -80.346 44.313
111396 101.291 -80.369 44.299
111397 101.344 -80.385 44.283
111398 101.376 -80.402 44.267
111399 101.408 -80.418 44.251
111400 101.440 -80.434 44.235
111401 101.472 -80.450 44.219
111402 101.504 -80.466 44.202
111403 101.536 -80.482 44.184
111404 101.568 -80.498 44.167
111405 101.601 -80.515 44.150
111406 101.632 -80.531 44.134
111407 101.665 -80.547 44.116
111408 101.697 -80.563 44.099
111409 101.729 -80.579 44.081
111745 94.421 -78.033 51.863
111746 94.435 -78.026 51.840
111747 94.449 -78.019 51.817
111748 94.450 -78.015 51.794
111749 94.451 -78.011 51.770
111750 94.452 -78.006 51.745
111751 94.473 -77.996 51.720
111752 94.493 -77.987 51.694
111753 94.514 -77.977 51.669
111754 94.535 -77.967 51.643
111755 94.551 -77.966 51.617
111756 94.571 -77.970 51.589
111757 94.609 -77.987 51.562
111758 94.646 -78.003 51.533
111759 94.664 -78.006 51.504
111760 94.675 -78.007 51.476
111761 94.686 -78.008 51.447
111762 94.698 -78.008 51.420
111763 94.707 -78.002 51.390
111764 94.715 -77.993 51.360
111765 94.723 -77.984 51.329
111766 94.728 -77.984 51.300
111767 94.727 -77.992 51.270
111768 94.727 -77.999 51.240
111769 94.734 -78.000 51.210
111770 94.746 -77.998 51.179
111771 94.758 -78.003 51.146
111772 94.770 -78.008 51.113
111773 94.782 -78.013 51.081
111774 94.804 -78.020 51.048
111775 94.829 -78.027 51.014
111776 94.853 -78.034 50.980
111777 94.878 -78.041 50.946
111778 94.901 -78.040 50.910
111779 94.919 -78.024 50.876
111780 94.937 -78.006 50.840
111781 94.955 -77.989 50.806
111782 94.973 -77.971 50.772
111783 94.992 -77.955 50.738
111784 95.010 -77.937 50.704
111785 95.028 -77.920 50.669
111786 95.037 -77.907 50.634
111787 95.046 -77.893 50.601
111788 95.055 -77.881 50.568
111789 95.064 -77.867 50.534
111790 95.072 -77.854 50.503
111791 95.082 -77.840 50.472
111792 95.094 -77.823 50.441
111793 95.123 -77.807 50.410
111794 95.147 -77.812 50.382
111795 95.171 -77.815 50.354
111796 95.205 -77.852 50.327
111797 95.244 -77.904 50.296
111798 95.239 -77.907 50.266
111799 95.242 -77.906 50.236
111800 95.278 -77.908 50.207
111801 95.315 -77.900 50.182
111802 95.352 -77.893 50.158
111803 95.370 -77.891 50.135
111804 95.384 -77.890 50.114
111805 95.395 -77.890 50.094
111806 95.401 -77.893 50.073
111807 95.407 -77.895 50.049
111808 95.414 -77.897 50.024
111809 95.420 -77.899 49.995
111810 95.427 -77.902 49.963
111811 95.433 -77.904 49.929
111812 95.440 -77.907 49.892
111813 95.446 -77.909 49.854
111814 95.453 -77.912 49.815
111815 95.461 -77.916 49.777
111816 95.482 -77.938 49.740
111817 95.505 -77.959 49.704
111818 95.527 -77.981 49.671
111819 95.547 -77.993 49.639
111820 95.566 -77.976 49.605
111821 95.585 -77.959 49.569
111822 95.604 -77.940 49.532
111823 95.623 -77.923 49.495
111824 95.642 -77.906 49.460
111825 95.662 -77.887 49.423
111826 95.656 -77.902 49.388
111827 95.630 -77.943 49.352
111828 95.634 -77.958 49.315
111829 95.650 -77.963 49.279
111830 95.665 -77.968 49.243
111831 95.703 -77.943 49.207
111832 95.682 -77.951 49.169
111833 95.679 -77.971 49.129
111834 95.697 -77.966 49.091
111835 95.724 -77.975 49.054
111836 95.750 -77.984 49.017
111837 95.774 -77.990 48.980
111838 95.796 -77.992 48.941
111839 95.818 -77.994 48.901
111840 95.840 -77.996 48.860
111841 95.862 -77.999 48.819
111842 95.883 -78.003 48.779
111843 95.892 -78.019 48.739
111844 95.901 -78.037 48.699
111845 95.898 -78.060 48.659
111846 95.895 -78.075 48.619
111847 95.902 -78.074 48.583
111848 95.909 -78.072 48.547
111849 95.915 -78.070 48.510
111850 95.925 -78.070 48.474
111851 95.938 -78.071 48.437
111852 95.954 -78.068 48.400
111853 95.970 -78.065 48.362
111854 95.986 -78.062 48.328
111855 96.003 -78.059 48.295
111856 96.019 -78.056 48.262
111857 96.035 -78.053 48.230
111858 96.051 -78.049 48.200
111859 96.067 -78.048 48.169
111860 96.083 -78.047 48.139
111861 96.099 -78.046 48.112
111862 96.115 -78.045 48.085
111863 96.116 -78.042 48.057
111864 96.138 -78.052 48.030
111865 96.152 -78.065 48.002
111866 96.166 -78.078 47.973
111867 96.180 -78.091 47.945
111868 96.194 -78.104 47.918
111869 96.208 -78.118 47.891
111870 96.222 -78.131 47.864
111871 96.237 -78.137 47.837
111872 96.252 -78.145 47.810
111873 96.267 -78.157 47.784
111874 96.281 -78.169 47.758
111875 96.294 -78.178 47.734
111876 96.306 -78.187 47.711
111877 96.318 -78.197 47.689
111878 96.331 -78.206 47.667
111879 96.343 -78.214 47.644
111880 96.363 -78.225 47.619
111881 96.382 -78.237 47.592
111882 96.402 -78.248 47.567
111883 96.421 -78.257 47.541
111884 96.437 -78.261 47.516
111885 96.454 -78.265 47.492
111886 96.470 -78.269 47.469
111887 96.484 -78.273 47.446
111888 96.489 -78.280 47.423
111889 96.496 -78.288 47.399
111890 96.506 -78.297 47.374
111891 96.516 -78.306 47.347
111892 96.526 -78.315 47.321
111893 96.531 -78.319 47.294
111894 96.530 -78.321 47.267
111895 96.530 -78.322 47.240
111896 96.529 -78.322 47.215
111897 96.540 -78.322 47.188
111898 96.553 -78.328 47.162
111899 96.558 -78.352 47.137
111900 96.564 -78.376 47.111
111901 96.569 -78.399 47.086
111902 96.575 -78.423 47.060
111903 96.582 -78.442 47.034
111904 96.591 -78.458 47.007
111905 96.601 -78.473 46.980
111906 96.614 -78.517 46.951
111907 96.619 -78.525 46.924
111908 96.620 -78.542 46.897
111909 96.617 -78.561 46.870
111910 96.614 -78.581 46.844
111911 96.611 -78.601 46.817
111912 96.608 -78.620 46.791
111913 96.638 -78.633 46.764
111914 96.668 -78.645 46.736
111915 96.699 -78.657 46.708
111916 96.738 -78.667 46.681
111917 96.779 -78.678 46.652
111918 96.819 -78.689 46.624
111919 96.857 -78.702 46.597
111920 96.887 -78.719 46.569
111921 96.916 -78.737 46.541
111922 96.946 -78.755 46.512
111923 96.973 -78.766 46.484
111924 97.000 -78.777 46.457
111925 97.027 -78.789 46.432
111926 97.070 -78.799 46.406
111927 97.120 -78.809 46.380
111928 97.170 -78.819 46.357
111929 97.220 -78.830 46.335
111930 97.271 -78.838 46.313
111931 97.309 -78.833 46.292
111932 97.348 -78.833 46.273
111933 97.386 -78.832 46.255
111934 97.425 -78.831 46.237
111935 97.463 -78.830 46.221
111936 97.502 -78.830 46.204
111937 97.541 -78.829 46.188
111938 97.579 -78.828 46.174
111939 97.618 -78.827 46.159
111940 97.657 -78.827 46.146
111941 97.696 -78.826 46.131
111942 97.694 -78.843 46.116
111943 97.696 -78.832 46.101
111944 97.699 -78.821 46.086
111945 97.721 -78.827 46.071
111946 97.749 -78.826 46.056
111947 97.781 -78.798 46.042
111948 97.805 -78.757 46.028
111949 97.831 -78.715 46.014
111950 97.850 -78.691 45.999
111951 97.867 -78.691 45.984
111952 97.892 -78.695 45.967
111953 97.919 -78.703 45.950
111954 97.945 -78.711 45.933
111955 97.972 -78.719 45.916
111956 97.999 -78.727 45.899
111957 98.018 -78.710 45.883
111958 98.031 -78.702 45.867
111959 98.042 -78.727 45.851
111960 98.054 -78.754 45.834
111961 98.008 -78.737 45.818
111962 97.983 -78.697 45.803
111963 97.994 -78.668 45.787
111964 98.005 -78.639 45.772
111965 98.028 -78.635 45.756
111966 98.024 -78.669 45.741
111967 98.024 -78.689 45.727
111968 98.024 -78.709 45.712
111969 98.024 -78.729 45.696
111970 98.033 -78.744 45.681
111971 98.062 -78.747 45.666
111972 98.091 -78.750 45.650
111973 98.112 -78.765 45.635
111974 98.134 -78.780 45.621
111975 98.155 -78.796 45.607
111976 98.177 -78.809 45.592
111977 98.191 -78.815 45.579
111978 98.197 -78.810 45.566
111979 98.203 -78.802 45.553
111980 98.210 -78.793 45.540
111981 98.216 -78.783 45.529
111982 98.232 -78.786 45.518
111983 98.228 -78.819 45.505
111984 98.250 -78.808 45.491
111985 98.274 -78.796 45.476
111986 98.295 -78.777 45.462
111987 98.314 -78.752 45.447
111988 98.339 -78.755 45.433
111989 98.367 -78.782 45.419
111990 98.396 -78.810 45.405
111991 98.424 -78.837 45.392
111992 98.445 -78.871 45.379
111993 98.412 -78.891 45.366
111994 98.416 -78.893 45.354
111995 98.440 -78.885 45.343
111996 98.406 -78.959 45.331
111997 98.460 -79.021 45.319
111998 98.483 -79.024 45.308
111999 98.506 -79.027 45.295
112000 98.529 -79.030 45.283
112001 98.552 -79.032 45.271
112002 98.575 -79.034 45.257
112003 98.598 -79.037 45.245
112004 98.621 -79.040 45.233
112005 98.644 -79.043 45.218
112006 98.667 -79.045 45.204
112007 98.690 -79.048 45.188
112008 98.734 -79.043 45.173
112009 98.779 -79.037 45.157
112010 98.825 -79.031 45.142
112011 98.846 -78.990 45.128
112012 98.883 -78.963 45.116
112013 98.939 -78.956 45.104
112014 98.982 -78.955 45.095
112015 99.008 -78.963 45.086
112016 99.033 -78.971 45.079
112017 99.059 -78.979 45.069
112018 99.085 -78.984 45.059
112019 99.110 -78.985 45.049
112020 99.135 -78.987 45.039
112021 99.148 -78.998 45.029
112022 99.161 -79.009 45.019
112023 99.175 -79.019 45.009
112024 99.189 -79.026 44.998
112025 99.203 -79.032 44.987
112026 99.219 -79.041 44.975
112027 99.242 -79.060 44.962
112028 99.263 -79.066 44.952
112029 99.282 -79.065 44.942
112030 99.300 -79.065 44.933
112031 99.325 -79.080 44.924
112032 99.354 -79.083 44.917
112033 99.372 -79.071 44.911
112034 99.405 -79.072 44.905
112035 99.451 -79.087 44.902
112036 99.497 -79.101 44.899
112037 99.548 -79.128 44.898
112038 99.596 -79.150 44.896
112039 99.633 -79.143 44.895
112040 99.579 -79.146 44.895
112041 99.546 -79.154 44.893
112042 99.552 -79.173 44.891
112043 99.558 -79.193 44.889
112044 99.563 -79.212 44.886
112045 99.580 -79.236 44.881
112046 99.603 -79.263 44.875
112047 99.626 -79.291 44.868
112048 99.650 -79.301 44.861
112049 99.677 -79.306 44.855
112050 99.702 -79.311 44.847
112051 99.728 -79.316 44.840
112052 99.759 -79.335 44.832
112053 99.792 -79.368 44.824
112054 99.827 -79.400 44.817
112055 99.860 -79.433 44.811
112056 99.895 -79.466 44.803
112057 99.929 -79.499 44.795
112058 99.963 -79.532 44.787
112059 99.997 -79.564 44.779
112060 100.031 -79.598 44.771
112061 100.068 -79.622 44.764
112062 100.107 -79.632 44.755
112063 100.145 -79.641 44.746
112064 100.179 -79.651 44.738
112065 100.213 -79.661 44.729
112066 100.246 -79.671 44.720
112067 100.280 -79.681 44.712
112068 100.311 -79.705 44.704
112069 100.339 -79.738 44.692
112070 100.368 -79.770 44.678
112071 100.397 -79.802 44.664
112072 100.425 -79.835 44.650
112073 100.414 -79.857 44.636
112074 100.472 -79.866 44.623
112075 100.522 -79.871 44.609
112076 100.559 -79.871 44.596
112077 100.597 -79.870 44.583
112078 100.634 -79.869 44.570
112079 100.671 -79.868 44.556
112080 100.708 -79.867 44.542
112081 100.746 -79.867 44.528
112082 100.779 -79.899 44.513
112083 100.813 -79.933 44.500
112084 100.846 -79.969 44.487
112085 100.879 -80.004 44.474
112086 100.912 -80.039 44.461
112087 100.945 -80.074 44.448
112088 100.978 -80.109 44.434
112089 101.011 -80.144 44.421
112090 101.044 -80.180 44.406
112091 101.068 -80.208 44.392
112092 101.096 -80.239 44.377
112093 101.123 -80.267 44.363
112094 101.165 -80.281 44.348
112095 101.211 -80.293 44.333
112096 101.257 -80.305 44.319
112097 101.305 -80.334 44.305
112098 101.361 -80.347 44.289
112099 101.393 -80.363 44.274
112100 101.425 -80.379 44.258
112101 101.457 -80.396 44.243
112102 101.489 -80.412 44.226
112103 101.522 -80.428 44.209
112104 101.553 -80.444 44.191
112105 101.586 -80.460 44.174
112106 101.618 -80.476 44.157
112107 101.650 -80.493 44.140
112108 101.682 -80.509 44.124
112109 101.714 -80.525 44.106
112110 101.746 -80.541 44.088
112111 101.778 -80.557 44.071
112448 94.459 -78.000 51.813
112449 94.463 -77.996 51.791
112450 94.464 -77.991 51.767
112451 94.465 -77.987 51.742
112452 94.475 -77.980 51.718
112453 94.496 -77.970 51.692
112454 94.517 -77.960 51.667
112455 94.537 -77.949 51.643
112456 94.554 -77.947 51.616
112457 94.569 -77.947 51.590
112458 94.605 -77.962 51.562
112459 94.642 -77.978 51.533
112460 94.662 -77.984 51.505
112461 94.674 -77.984 51.477
112462 94.685 -77.984 51.449
112463 94.696 -77.985 51.421
112464 94.707 -77.984 51.392
112465 94.715 -77.975 51.362
112466 94.723 -77.966 51.331
112467 94.728 -77.965 51.302
112468 94.728 -77.972 51.272
112469 94.728 -77.979 51.241
112470 94.732 -77.983 51.212
112471 94.743 -77.981 51.181
112472 94.755 -77.986 51.148
112473 94.767 -77.990 51.115
112474 94.782 -77.996 51.083
112475 94.807 -78.003 51.049
112476 94.832 -78.010 51.015
112477 94.856 -78.018 50.982
112478 94.881 -78.025 50.948
112479 94.903 -78.022 50.913
112480 94.921 -78.005 50.878
112481 94.939 -77.988 50.843
112482 94.958 -77.971 50.809
112483 94.976 -77.953 50.775
112484 94.994 -77.936 50.741
112485 95.012 -77.919 50.706
112486 95.022 -77.906 50.671
112487 95.031 -77.892 50.637
112488 95.040 -77.879 50.603
112489 95.049 -77.865 50.570
112490 95.057 -77.852 50.536
112491 95.066 -77.839 50.504
112492 95.075 -77.826 50.473
112493 95.084 -77.812 50.441
112494 95.111 -77.778 50.409
112495 95.141 -77.804 50.380
112496 95.181 -77.855 50.351
112497 95.219 -77.907 50.322
112498 95.209 -77.902 50.291
112499 95.193 -77.891 50.262
112500 95.228 -77.908 50.232
112501 95.264 -77.900 50.203
112502 95.302 -77.893 50.176
112503 95.339 -77.886 50.152
112504 95.370 -77.880 50.129
112505 95.383 -77.879 50.106
112506 95.396 -77.878 50.082
112507 95.409 -77.877 50.058
112508 95.416 -77.880 50.034
112509 95.422 -77.882 50.006
112510 95.429 -77.884 49.978
112511 95.436 -77.887 49.948
112512 95.442 -77.889 49.914
112513 95.448 -77.891 49.880
112514 95.455 -77.894 49.846
112515 95.461 -77.896 49.811
112516 95.469 -77.899 49.773
112517 95.490 -77.921 49.736
112518 95.513 -77.943 49.701
112519 95.534 -77.965 49.668
112520 95.542 -77.982 49.636
112521 95.546 -77.972 49.603
112522 95.563 -77.956 49.570
112523 95.583 -77.938 49.535
112524 95.602 -77.921 49.500
112525 95.620 -77.902 49.465
112526 95.639 -77.885 49.429
112527 95.648 -77.882 49.394
112528 95.622 -77.922 49.358
112529 95.625 -77.938 49.322
112530 95.640 -77.943 49.285
112531 95.656 -77.949 49.247
112532 95.672 -77.954 49.210
112533 95.682 -77.943 49.172
112534 95.685 -77.952 49.136
112535 95.711 -77.961 49.099
112536 95.736 -77.966 49.063
112537 95.757 -77.968 49.026
112538 95.779 -77.971 48.989
112539 95.802 -77.973 48.950
112540 95.823 -77.975 48.910
112541 95.845 -77.978 48.871
112542 95.868 -77.980 48.831
112543 95.881 -77.993 48.790
112544 95.890 -78.009 48.749
112545 95.898 -78.027 48.708
112546 95.896 -78.050 48.668
112547 95.895 -78.062 48.629
112548 95.902 -78.059 48.592
112549 95.909 -78.058 48.556
112550 95.916 -78.056 48.520
112551 95.929 -78.057 48.482
112552 95.942 -78.059 48.446
112553 95.956 -78.057 48.409
112554 95.972 -78.053 48.372
112555 95.988 -78.050 48.338
112556 96.005 -78.047 48.305
112557 96.021 -78.044 48.273
112558 96.037 -78.041 48.242
112559 96.053 -78.038 48.212
112560 96.070 -78.035 48.182
112561 96.086 -78.032 48.153
112562 96.102 -78.028 48.127
112563 96.118 -78.025 48.099
112564 96.133 -78.024 48.071
112565 96.147 -78.034 48.044
112566 96.161 -78.047 48.015
112567 96.175 -78.060 47.987
112568 96.189 -78.074 47.959
112569 96.203 -78.087 47.933
112570 96.217 -78.100 47.906
112571 96.231 -78.113 47.880
112572 96.245 -78.126 47.854
112573 96.260 -78.133 47.828
112574 96.275 -78.145 47.802
112575 96.288 -78.155 47.776
112576 96.300 -78.164 47.751
112577 96.314 -78.173 47.727
112578 96.329 -78.181 47.703
112579 96.344 -78.189 47.680
112580 96.352 -78.197 47.657
112581 96.371 -78.210 47.634
112582 96.391 -78.222 47.609
112583 96.410 -78.233 47.583
112584 96.430 -78.243 47.557
112585 96.446 -78.247 47.531
112586 96.463 -78.250 47.506
112587 96.479 -78.253 47.483
112588 96.485 -78.261 47.460
112589 96.490 -78.268 47.436
112590 96.495 -78.275 47.413
112591 96.502 -78.283 47.388
112592 96.512 -78.292 47.363
112593 96.518 -78.298 47.337
112594 96.518 -78.299 47.311
112595 96.517 -78.300 47.285
112596 96.517 -78.301 47.259
112597 96.517 -78.302 47.233
112598 96.532 -78.301 47.207
112599 96.543 -78.311 47.180
112600 96.549 -78.335 47.155
112601 96.554 -78.358 47.129
112602 96.561 -78.379 47.101
112603 96.571 -78.395 47.072
112604 96.580 -78.410 47.044
112605 96.589 -78.425 47.018
112606 96.599 -78.441 46.991
112607 96.616 -78.506 46.964
112608 96.619 -78.506 46.936
112609 96.625 -78.520 46.908
112610 96.627 -78.536 46.881
112611 96.624 -78.556 46.854
112612 96.621 -78.576 46.827
112613 96.622 -78.594 46.800
112614 96.654 -78.606 46.773
112615 96.684 -78.618 46.746
112616 96.715 -78.630 46.718
112617 96.746 -78.642 46.690
112618 96.783 -78.654 46.662
112619 96.824 -78.664 46.634
112620 96.863 -78.677 46.607
112621 96.892 -78.694 46.580
112622 96.922 -78.711 46.551
112623 96.949 -78.724 46.522
112624 96.976 -78.734 46.494
112625 97.003 -78.745 46.467
112626 97.030 -78.757 46.442
112627 97.072 -78.767 46.416
112628 97.122 -78.777 46.391
112629 97.172 -78.788 46.367
112630 97.222 -78.798 46.344
112631 97.246 -78.815 46.323
112632 97.281 -78.819 46.302
112633 97.321 -78.819 46.283
112634 97.361 -78.820 46.264
112635 97.401 -78.820 46.246
112636 97.442 -78.821 46.230
112637 97.482 -78.821 46.214
112638 97.522 -78.821 46.199
112639 97.563 -78.821 46.185
112640 97.603 -78.822 46.171
112641 97.643 -78.822 46.157
112642 97.689 -78.818 46.143
112643 97.720 -78.811 46.128
112644 97.718 -78.799 46.112
112645 97.721 -78.789 46.097
112646 97.738 -78.790 46.083
112647 97.772 -78.799 46.068
112648 97.798 -78.760 46.054
112649 97.824 -78.718 46.040
112650 97.843 -78.692 46.027
112651 97.850 -78.697 46.012
112652 97.871 -78.695 45.997
112653 97.892 -78.693 45.980
112654 97.917 -78.698 45.962
112655 97.944 -78.705 45.945
112656 97.970 -78.714 45.928
112657 97.997 -78.721 45.912
112658 98.018 -78.710 45.895
112659 98.031 -78.706 45.878
112660 98.042 -78.732 45.862
112661 98.034 -78.743 45.846
112662 98.044 -78.741 45.831
112663 98.006 -78.702 45.815
112664 98.014 -78.672 45.799
112665 98.026 -78.642 45.784
112666 98.041 -78.614 45.768
112667 98.043 -78.645 45.752
112668 98.044 -78.664 45.737
112669 98.044 -78.685 45.722
112670 98.044 -78.705 45.706
112671 98.044 -78.725 45.691
112672 98.063 -78.734 45.675
112673 98.088 -78.744 45.661
112674 98.109 -78.758 45.646
112675 98.134 -78.772 45.633
112676 98.158 -78.785 45.618
112677 98.176 -78.771 45.603
112678 98.200 -78.767 45.589
112679 98.204 -78.786 45.575
112680 98.209 -78.807 45.561
112681 98.215 -78.814 45.547
112682 98.221 -78.805 45.536
112683 98.227 -78.818 45.523
112684 98.250 -78.806 45.510
112685 98.273 -78.794 45.498
112686 98.296 -78.783 45.484
112687 98.319 -78.771 45.469
112688 98.341 -78.755 45.454
112689 98.362 -78.739 45.439
112690 98.391 -78.767 45.423
112691 98.419 -78.794 45.409
112692 98.431 -78.839 45.396
112693 98.440 -78.887 45.382
112694 98.431 -78.905 45.370
112695 98.455 -78.898 45.358
112696 98.480 -78.890 45.347
112697 98.504 -78.882 45.335
112698 98.430 -78.990 45.323
112699 98.462 -79.047 45.311
112700 98.486 -79.049 45.297
112701 98.509 -79.053 45.285
112702 98.531 -79.055 45.272
112703 98.555 -79.058 45.259
112704 98.577 -79.060 45.246
112705 98.600 -79.062 45.234
112706 98.626 -79.065 45.221
112707 98.671 -79.059 45.207
112708 98.716 -79.053 45.193
112709 98.762 -79.047 45.179
112710 98.807 -79.042 45.163
112711 98.842 -79.020 45.148
112712 98.860 -78.974 45.134
112713 98.888 -78.938 45.121
112714 98.945 -78.931 45.110
112715 98.986 -78.931 45.101
112716 99.012 -78.939 45.094
112717 99.038 -78.947 45.086
112718 99.063 -78.955 45.076
112719 99.089 -78.963 45.066
112720 99.115 -78.971 45.057
112721 99.132 -78.977 45.047
112722 99.146 -78.984 45.036
112723 99.160 -78.990 45.025
112724 99.174 -78.997 45.013
112725 99.188 -79.003 45.002
112726 99.202 -79.010 44.990
112727 99.220 -79.021 44.978
112728 99.243 -79.040 44.965
112729 99.266 -79.059 44.955
112730 99.289 -79.077 44.945
112731 99.308 -79.076 44.934
112732 99.334 -79.084 44.926
112733 99.352 -79.072 44.919
112734 99.370 -79.060 44.914
112735 99.394 -79.053 44.909
112736 99.440 -79.068 44.906
112737 99.487 -79.083 44.904
112738 99.546 -79.132 44.904
112739 99.589 -79.141 44.904
112740 99.626 -79.134 44.903
112741 99.617 -79.132 44.901
112742 99.566 -79.136 44.897
112743 99.572 -79.155 44.893
112744 99.578 -79.175 44.888
112745 99.584 -79.194 44.883
112746 99.589 -79.212 44.877
112747 99.602 -79.235 44.871
112748 99.625 -79.263 44.864
112749 99.648 -79.285 44.856
112750 99.674 -79.290 44.848
112751 99.700 -79.294 44.841
112752 99.726 -79.300 44.833
112753 99.760 -79.334 44.826
112754 99.794 -79.366 44.818
112755 99.828 -79.399 44.811
112756 99.862 -79.432 44.805
112757 99.897 -79.465 44.798
112758 99.930 -79.498 44.790
112759 99.965 -79.530 44.783
112760 99.998 -79.563 44.774
112761 100.033 -79.595 44.767
112762 100.070 -79.616 44.759
112763 100.108 -79.625 44.751
112764 100.142 -79.636 44.741
112765 100.176 -79.645 44.732
112766 100.207 -79.648 44.724
112767 100.244 -79.654 44.715
112768 100.269 -79.671 44.709
112769 100.320 -79.678 44.702
112770 100.349 -79.711 44.692
112771 100.378 -79.743 44.679
112772 100.406 -79.776 44.666
112773 100.417 -79.804 44.653
112774 100.409 -79.825 44.639
112775 100.467 -79.834 44.625
112776 100.524 -79.844 44.610
112777 100.566 -79.845 44.596
112778 100.604 -79.845 44.583
112779 100.641 -79.844 44.570
112780 100.678 -79.843 44.557
112781 100.716 -79.842 44.543
112782 100.753 -79.842 44.530
112783 100.786 -79.875 44.516
112784 100.820 -79.911 44.502
112785 100.852 -79.946 44.490
112786 100.886 -79.980 44.478
112787 100.919 -80.016 44.464
112788 100.952 -80.051 44.451
112789 100.985 -80.086 44.438
112790 101.018 -80.121 44.424
112791 101.051 -80.156 44.410
112792 101.085 -80.191 44.396
112793 101.117 -80.226 44.382
112794 101.146 -80.232 44.368
112795 101.178 -80.240 44.354
112796 101.224 -80.252 44.339
112797 101.270 -80.263 44.325
112798 101.318 -80.300 44.312
112799 101.378 -80.309 44.296
112800 101.410 -80.325 44.281
112801 101.443 -80.341 44.266
112802 101.474 -80.357 44.250
112803 101.506 -80.374 44.233
112804 101.539 -80.390 44.216
112805 101.571 -80.406 44.198
112806 101.603 -80.422 44.181
112807 101.635 -80.438 44.165
112808 101.667 -80.454 44.148
112809 101.699 -80.470 44.130
112810 101.731 -80.487 44.112
112811 101.763 -80.503 44.094
113148 94.454 -77.988 51.835
113149 94.468 -77.981 51.813
113150 94.475 -77.976 51.790
113151 94.476 -77.971 51.767
113152 94.477 -77.968 51.743
113153 94.478 -77.963 51.718
113154 94.498 -77.953 51.694
113155 94.519 -77.943 51.669
113156 94.540 -77.933 51.644
113157 94.558 -77.928 51.618
113158 94.573 -77.929 51.591
113159 94.600 -77.937 51.564
113160 94.638 -77.953 51.537
113161 94.661 -77.961 51.508
113162 94.672 -77.962 51.480
113163 94.684 -77.962 51.452
113164 94.695 -77.962 51.424
113165 94.706 -77.963 51.395
113166 94.715 -77.956 51.365
113167 94.723 -77.947 51.334
113168 94.729 -77.945 51.305
113169 94.729 -77.953 51.276
113170 94.728 -77.960 51.245
113171 94.729 -77.966 51.215
113172 94.741 -77.963 51.183
113173 94.753 -77.968 51.150
113174 94.765 -77.973 51.118
113175 94.785 -77.980 51.085
113176 94.810 -77.987 51.051
113177 94.834 -77.994 51.018
113178 94.859 -78.001 50.985
113179 94.884 -78.009 50.951
113180 94.905 -78.004 50.917
113181 94.923 -77.987 50.881
113182 94.942 -77.969 50.846
113183 94.960 -77.953 50.812
113184 94.978 -77.935 50.777
113185 94.996 -77.918 50.743
113186 95.007 -77.904 50.709
113187 95.016 -77.890 50.674
113188 95.025 -77.877 50.639
113189 95.034 -77.864 50.605
113190 95.042 -77.851 50.572
113191 95.051 -77.837 50.538
113192 95.060 -77.824 50.505
113193 95.069 -77.811 50.473
113194 95.078 -77.798 50.442
113195 95.089 -77.783 50.409
113196 95.095 -77.853 50.379
113197 95.195 -77.909 50.349
113198 95.178 -77.899 50.319
113199 95.159 -77.884 50.288
113200 95.178 -77.895 50.258
113201 95.214 -77.900 50.228
113202 95.251 -77.893 50.199
113203 95.288 -77.886 50.172
113204 95.325 -77.878 50.147
113205 95.362 -77.871 50.122
113206 95.382 -77.868 50.098
113207 95.396 -77.867 50.073
113208 95.409 -77.866 50.048
113209 95.422 -77.865 50.022
113210 95.431 -77.867 49.995
113211 95.438 -77.869 49.967
113212 95.444 -77.871 49.939
113213 95.451 -77.874 49.908
113214 95.457 -77.876 49.877
113215 95.463 -77.878 49.844
113216 95.470 -77.881 49.810
113217 95.476 -77.884 49.773
113218 95.498 -77.906 49.736
113219 95.521 -77.928 49.701
113220 95.536 -77.952 49.667
113221 95.528 -77.976 49.635
113222 95.546 -77.958 49.602
113223 95.549 -77.949 49.570
113224 95.560 -77.935 49.536
113225 95.580 -77.918 49.502
113226 95.599 -77.900 49.468
113227 95.618 -77.883 49.432
113228 95.637 -77.865 49.397
113229 95.614 -77.902 49.362
113230 95.616 -77.918 49.325
113231 95.631 -77.924 49.288
113232 95.647 -77.929 49.252
113233 95.663 -77.934 49.217
113234 95.678 -77.940 49.181
113235 95.697 -77.943 49.145
113236 95.719 -77.945 49.109
113237 95.741 -77.947 49.073
113238 95.763 -77.949 49.036
113239 95.785 -77.952 48.999
113240 95.807 -77.954 48.962
113241 95.829 -77.956 48.923
113242 95.851 -77.959 48.884
113243 95.870 -77.965 48.844
113244 95.878 -77.982 48.803
113245 95.887 -78.000 48.761
113246 95.896 -78.017 48.720
113247 95.893 -78.040 48.679
113248 95.895 -78.047 48.639
113249 95.902 -78.046 48.602
113250 95.908 -78.044 48.565
113251 95.919 -78.044 48.528
113252 95.932 -78.045 48.491
113253 95.945 -78.046 48.455
113254 95.959 -78.046 48.419
113255 95.975 -78.043 48.384
113256 95.991 -78.040 48.351
113257 96.007 -78.036 48.318
113258 96.023 -78.033 48.286
113259 96.039 -78.030 48.255
113260 96.055 -78.027 48.225
113261 96.072 -78.024 48.197
113262 96.088 -78.021 48.168
113263 96.104 -78.018 48.142
113264 96.120 -78.015 48.115
113265 96.133 -78.016 48.088
113266 96.148 -78.024 48.060
113267 96.163 -78.036 48.031
113268 96.178 -78.048 48.003
113269 96.193 -78.060 47.975
113270 96.209 -78.072 47.950
113271 96.224 -78.084 47.924
113272 96.239 -78.097 47.898
113273 96.254 -78.109 47.873
113274 96.268 -78.122 47.847
113275 96.282 -78.132 47.821
113276 96.298 -78.140 47.795
113277 96.313 -78.148 47.769
113278 96.328 -78.156 47.745
113279 96.343 -78.165 47.720
113280 96.354 -78.172 47.696
113281 96.361 -78.181 47.672
113282 96.380 -78.194 47.647
113283 96.399 -78.206 47.622
113284 96.419 -78.217 47.598
113285 96.439 -78.228 47.572
113286 96.455 -78.232 47.547
113287 96.471 -78.235 47.521
113288 96.480 -78.241 47.497
113289 96.485 -78.249 47.473
113290 96.490 -78.256 47.448
113291 96.496 -78.263 47.426
113292 96.501 -78.271 47.402
113293 96.505 -78.276 47.378
113294 96.505 -78.277 47.353
113295 96.504 -78.279 47.328
113296 96.504 -78.279 47.303
113297 96.503 -78.280 47.276
113298 96.509 -78.281 47.250
113299 96.525 -78.280 47.224
113300 96.534 -78.294 47.197
113301 96.541 -78.315 47.171
113302 96.550 -78.330 47.143
113303 96.559 -78.346 47.113
113304 96.569 -78.361 47.084
113305 96.578 -78.377 47.055
113306 96.587 -78.392 47.029
113307 96.597 -78.408 47.003
113308 96.615 -78.482 46.976
113309 96.620 -78.486 46.948
113310 96.625 -78.501 46.919
113311 96.631 -78.515 46.892
113312 96.634 -78.531 46.865
113313 96.631 -78.551 46.838
113314 96.639 -78.568 46.810
113315 96.670 -78.580 46.783
113316 96.700 -78.592 46.756
113317 96.731 -78.604 46.728
113318 96.762 -78.616 46.700
113319 96.793 -78.628 46.672
113320 96.828 -78.639 46.645
113321 96.868 -78.651 46.618
113322 96.897 -78.669 46.591
113323 96.925 -78.680 46.562
113324 96.952 -78.692 46.534
113325 96.979 -78.703 46.506
113326 97.006 -78.714 46.478
113327 97.033 -78.725 46.453
113328 97.075 -78.736 46.427
113329 97.124 -78.746 46.401
113330 97.174 -78.756 46.377
113331 97.205 -78.771 46.354
113332 97.214 -78.793 46.332
113333 97.246 -78.799 46.312
113334 97.286 -78.799 46.293
113335 97.326 -78.799 46.275
113336 97.366 -78.799 46.257
113337 97.406 -78.800 46.240
113338 97.447 -78.801 46.225
113339 97.487 -78.801 46.211
113340 97.527 -78.802 46.196
113341 97.568 -78.802 46.182
113342 97.619 -78.793 46.168
113343 97.679 -78.779 46.153
113344 97.727 -78.787 46.139
113345 97.748 -78.777 46.124
113346 97.758 -78.762 46.109
113347 97.773 -78.757 46.094
113348 97.790 -78.763 46.081
113349 97.817 -78.721 46.068
113350 97.837 -78.693 46.054
113351 97.841 -78.699 46.040
113352 97.853 -78.702 46.026
113353 97.874 -78.699 46.010
113354 97.896 -78.698 45.992
113355 97.917 -78.696 45.975
113356 97.942 -78.700 45.958
113357 97.969 -78.708 45.941
113358 97.996 -78.716 45.924
113359 98.018 -78.710 45.907
113360 98.031 -78.711 45.890
113361 98.041 -78.736 45.875
113362 98.016 -78.724 45.860
113363 98.040 -78.728 45.844
113364 98.022 -78.702 45.828
113365 98.030 -78.673 45.812
113366 98.044 -78.645 45.795
113367 98.058 -78.617 45.779
113368 98.063 -78.620 45.764
113369 98.064 -78.641 45.748
113370 98.064 -78.661 45.733
113371 98.064 -78.681 45.717
113372 98.065 -78.702 45.702
113373 98.068 -78.720 45.688
113374 98.090 -78.734 45.673
113375 98.114 -78.747 45.658
113376 98.140 -78.760 45.644
113377 98.161 -78.761 45.629
113378 98.171 -78.726 45.614
113379 98.199 -78.717 45.600
113380 98.213 -78.738 45.585
113381 98.218 -78.758 45.570
113382 98.223 -78.778 45.556
113383 98.228 -78.799 45.542
113384 98.246 -78.805 45.528
113385 98.270 -78.793 45.516
113386 98.295 -78.781 45.504
113387 98.319 -78.769 45.490
113388 98.342 -78.758 45.475
113389 98.365 -78.745 45.461
113390 98.387 -78.732 45.445
113391 98.408 -78.758 45.429
113392 98.417 -78.806 45.414
113393 98.429 -78.852 45.400
113394 98.441 -78.897 45.386
113395 98.460 -78.902 45.374
113396 98.495 -78.902 45.363
113397 98.519 -78.894 45.351
113398 98.544 -78.886 45.339
113399 98.551 -78.905 45.326
113400 98.453 -79.021 45.313
113401 98.465 -79.072 45.300
113402 98.488 -79.075 45.287
113403 98.511 -79.078 45.273
113404 98.534 -79.081 45.261
113405 98.563 -79.081 45.247
113406 98.608 -79.075 45.234
113407 98.654 -79.069 45.222
113408 98.699 -79.064 45.208
113409 98.744 -79.058 45.195
113410 98.790 -79.053 45.182
113411 98.835 -79.046 45.168
113412 98.855 -79.003 45.153
113413 98.873 -78.957 45.139
113414 98.894 -78.913 45.126
113415 98.950 -78.906 45.117
113416 98.990 -78.907 45.110
113417 99.016 -78.915 45.104
113418 99.042 -78.923 45.095
113419 99.067 -78.930 45.086
113420 99.093 -78.939 45.077
113421 99.112 -78.949 45.069
113422 99.130 -78.955 45.059
113423 99.144 -78.962 45.048
113424 99.158 -78.968 45.036
113425 99.172 -78.975 45.023
113426 99.187 -78.981 45.009
113427 99.201 -78.988 44.996
113428 99.220 -79.002 44.983
113429 99.243 -79.020 44.970
113430 99.267 -79.038 44.958
113431 99.290 -79.056 44.948
113432 99.312 -79.075 44.937
113433 99.331 -79.073 44.929
113434 99.349 -79.061 44.923
113435 99.367 -79.049 44.918
113436 99.385 -79.037 44.914
113437 99.430 -79.050 44.913
113438 99.483 -79.084 44.912
113439 99.543 -79.137 44.911
113440 99.581 -79.132 44.910
113441 99.618 -79.125 44.907
113442 99.653 -79.118 44.902
113443 99.596 -79.121 44.897
113444 99.592 -79.137 44.892
113445 99.598 -79.156 44.886
113446 99.604 -79.175 44.878
113447 99.610 -79.195 44.872
113448 99.616 -79.214 44.865
113449 99.624 -79.234 44.858
113450 99.646 -79.261 44.849
113451 99.671 -79.273 44.841
113452 99.697 -79.278 44.833
113453 99.728 -79.300 44.825
113454 99.762 -79.332 44.819
113455 99.796 -79.365 44.812
113456 99.830 -79.397 44.805
113457 99.864 -79.430 44.798
113458 99.898 -79.463 44.792
113459 99.932 -79.496 44.785
113460 99.966 -79.529 44.778
113461 100.000 -79.561 44.770
113462 100.034 -79.595 44.762
113463 100.071 -79.609 44.753
113464 100.100 -79.614 44.745
113465 100.140 -79.610 44.736
113466 100.180 -79.614 44.727
113467 100.205 -79.630 44.718
113468 100.230 -79.647 44.710
113469 100.255 -79.664 44.703
113470 100.307 -79.666 44.696
113471 100.359 -79.683 44.687
113472 100.388 -79.716 44.676
113473 100.416 -79.748 44.665
113474 100.408 -79.773 44.653
113475 100.403 -79.793 44.640
113476 100.461 -79.802 44.627
113477 100.519 -79.812 44.613
113478 100.573 -79.820 44.599
113479 100.611 -79.819 44.586
113480 100.648 -79.818 44.572
113481 100.686 -79.817 44.559
113482 100.723 -79.817 44.545
113483 100.760 -79.817 44.532
113484 100.793 -79.852 44.518
113485 100.826 -79.887 44.505
113486 100.859 -79.922 44.493
113487 100.893 -79.958 44.481
113488 100.926 -79.993 44.467
113489 100.959 -80.027 44.455
113490 100.992 -80.063 44.442
113491 101.025 -80.098 44.428
113492 101.058 -80.133 44.414
113493 101.094 -80.164 44.401
113494 101.135 -80.187 44.387
113495 101.169 -80.193 44.373
113496 101.198 -80.200 44.360
113497 101.237 -80.210 44.345
113498 101.283 -80.222 44.331
113499 101.332 -80.266 44.318
113500 101.396 -80.271 44.302
113501 101.427 -80.287 44.287
113502 101.460 -80.303 44.272
113503 101.492 -80.319 44.257
113504 101.524 -80.335 44.240
113505 101.556 -80.351 44.223
113506 101.588 -80.368 44.206
113507 101.620 -80.384 44.189
113508 101.652 -80.400 44.171
113509 101.684 -80.416 44.154
113510 101.716 -80.432 44.136
113511 101.749 -80.448 44.119
113512 101.780 -80.464 44.101
113848 94.449 -77.975 51.859
113849 94.463 -77.969 51.837
113850 94.477 -77.962 51.815
113851 94.488 -77.956 51.793
113852 94.489 -77.952 51.770
113853 94.490 -77.948 51.746
113854 94.491 -77.943 51.722
113855 94.501 -77.937 51.697
113856 94.522 -77.927 51.673
113857 94.542 -77.916 51.649
113858 94.561 -77.909 51.623
113859 94.577 -77.910 51.596
113860 94.596 -77.912 51.569
113861 94.633 -77.929 51.542
113862 94.659 -77.938 51.514
113863 94.671 -77.938 51.486
113864 94.682 -77.939 51.458
113865 94.694 -77.940 51.430
113866 94.705 -77.940 51.401
113867 94.715 -77.937 51.371
113868 94.724 -77.929 51.341
113869 94.729 -77.925 51.311
113870 94.729 -77.933 51.281
113871 94.729 -77.940 51.250
113872 94.729 -77.947 51.220
113873 94.738 -77.946 51.189
113874 94.751 -77.951 51.156
113875 94.763 -77.956 51.123
113876 94.788 -77.963 51.090
113877 94.813 -77.971 51.056
113878 94.837 -77.978 51.024
113879 94.861 -77.985 50.990
113880 94.886 -77.992 50.955
113881 94.907 -77.985 50.920
113882 94.925 -77.968 50.884
113883 94.944 -77.951 50.850
113884 94.962 -77.934 50.816
113885 94.980 -77.917 50.781
113886 94.992 -77.902 50.746
113887 95.001 -77.889 50.712
113888 95.010 -77.876 50.677
113889 95.019 -77.862 50.641
113890 95.028 -77.849 50.607
113891 95.036 -77.836 50.574
113892 95.045 -77.823 50.541
113893 95.054 -77.809 50.508
113894 95.063 -77.796 50.475
113895 95.075 -77.789 50.443
113896 95.089 -77.782 50.411
113897 95.109 -77.795 50.380
113898 95.100 -77.852 50.349
113899 95.109 -77.865 50.317
113900 95.130 -77.880 50.285
113901 95.164 -77.898 50.255
113902 95.201 -77.893 50.225
113903 95.238 -77.886 50.196
113904 95.275 -77.878 50.169
113905 95.312 -77.871 50.142
113906 95.349 -77.863 50.115
113907 95.382 -77.857 50.090
113908 95.395 -77.856 50.066
113909 95.408 -77.855 50.041
113910 95.421 -77.855 50.015
113911 95.435 -77.854 49.988
113912 95.446 -77.854 49.962
113913 95.453 -77.856 49.936
113914 95.459 -77.859 49.909
113915 95.466 -77.861 49.879
113916 95.472 -77.863 49.847
113917 95.478 -77.865 49.812
113918 95.485 -77.868 49.775
113919 95.507 -77.890 49.738
113920 95.528 -77.912 49.702
113921 95.526 -77.942 49.668
113922 95.515 -77.970 49.634
113923 95.545 -77.943 49.602
113924 95.548 -77.934 49.570
113925 95.551 -77.926 49.536
113926 95.558 -77.915 49.503
113927 95.577 -77.897 49.469
113928 95.596 -77.880 49.434
113929 95.615 -77.862 49.399
113930 95.606 -77.881 49.364
113931 95.607 -77.899 49.328
113932 95.622 -77.904 49.292
113933 95.638 -77.909 49.258
113934 95.654 -77.915 49.225
113935 95.691 -77.907 49.190
113936 95.702 -77.924 49.156
113937 95.724 -77.926 49.120
113938 95.746 -77.928 49.083
113939 95.768 -77.931 49.047
113940 95.790 -77.933 49.011
113941 95.812 -77.935 48.974
113942 95.834 -77.938 48.936
113943 95.856 -77.940 48.898
113944 95.867 -77.956 48.858
113945 95.876 -77.972 48.817
113946 95.885 -77.990 48.775
113947 95.894 -78.007 48.733
113948 95.891 -78.031 48.691
113949 95.895 -78.034 48.650
113950 95.902 -78.031 48.612
113951 95.910 -78.031 48.575
113952 95.923 -78.031 48.538
113953 95.936 -78.033 48.501
113954 95.948 -78.034 48.465
113955 95.961 -78.034 48.430
113956 95.977 -78.031 48.396
113957 95.993 -78.028 48.364
113958 96.010 -78.025 48.332
113959 96.026 -78.022 48.299
113960 96.042 -78.019 48.268
113961 96.058 -78.015 48.239
113962 96.074 -78.012 48.211
113963 96.090 -78.009 48.184
113964 96.106 -78.006 48.158
113965 96.119 -78.009 48.131
113966 96.131 -78.014 48.104
113967 96.145 -78.021 48.076
113968 96.160 -78.028 48.048
113969 96.175 -78.040 48.020
113970 96.190 -78.053 47.993
113971 96.205 -78.065 47.967
113972 96.221 -78.077 47.942
113973 96.236 -78.089 47.917
113974 96.251 -78.101 47.892
113975 96.268 -78.108 47.867
113976 96.290 -78.115 47.841
113977 96.311 -78.123 47.814
113978 96.327 -78.131 47.789
113979 96.342 -78.140 47.763
113980 96.357 -78.148 47.738
113981 96.357 -78.155 47.712
113982 96.370 -78.166 47.688
113983 96.389 -78.178 47.662
113984 96.408 -78.190 47.636
113985 96.428 -78.202 47.613
113986 96.447 -78.213 47.588
113987 96.464 -78.217 47.563
113988 96.476 -78.222 47.536
113989 96.481 -78.229 47.511
113990 96.486 -78.237 47.486
113991 96.491 -78.244 47.461
113992 96.494 -78.250 47.438
113993 96.496 -78.255 47.415
113994 96.498 -78.261 47.392
113995 96.497 -78.261 47.367
113996 96.495 -78.261 47.343
113997 96.493 -78.261 47.318
113998 96.491 -78.261 47.291
113999 96.503 -78.260 47.266
114000 96.519 -78.260 47.239
114001 96.530 -78.266 47.212
114002 96.539 -78.282 47.185
114003 96.548 -78.297 47.154
114004 96.557 -78.313 47.124
114005 96.567 -78.329 47.094
114006 96.576 -78.344 47.066
114007 96.585 -78.360 47.040
114008 96.594 -78.375 47.014
114009 96.615 -78.458 46.987
114010 96.621 -78.467 46.959
114011 96.626 -78.481 46.931
114012 96.631 -78.496 46.903
114013 96.637 -78.510 46.876
114014 96.641 -78.526 46.848
114015 96.655 -78.542 46.821
114016 96.686 -78.554 46.793
114017 96.717 -78.566 46.766
114018 96.748 -78.578 46.738
114019 96.779 -78.589 46.711
114020 96.809 -78.601 46.684
114021 96.840 -78.614 46.657
114022 96.873 -78.626 46.629
114023 96.900 -78.638 46.602
114024 96.928 -78.649 46.574
114025 96.954 -78.660 46.546
114026 96.981 -78.671 46.518
114027 97.009 -78.683 46.490
114028 97.036 -78.694 46.464
114029 97.077 -78.705 46.438
114030 97.126 -78.714 46.413
114031 97.163 -78.728 46.388
114032 97.173 -78.749 46.364
114033 97.182 -78.770 46.342
114034 97.210 -78.778 46.322
114035 97.250 -78.779 46.304
114036 97.291 -78.779 46.286
114037 97.331 -78.780 46.268
114038 97.371 -78.780 46.251
114039 97.412 -78.780 46.236
114040 97.452 -78.780 46.221
114041 97.492 -78.781 46.205
114042 97.549 -78.768 46.192
114043 97.609 -78.754 46.178
114044 97.669 -78.739 46.164
114045 97.716 -78.749 46.150
114046 97.763 -78.761 46.135
114047 97.775 -78.742 46.121
114048 97.788 -78.741 46.108
114049 97.809 -78.724 46.095
114050 97.830 -78.694 46.082
114051 97.835 -78.700 46.069
114052 97.839 -78.706 46.056
114053 97.857 -78.705 46.040
114054 97.878 -78.704 46.024
114055 97.899 -78.702 46.006
114056 97.921 -78.700 45.989
114057 97.942 -78.698 45.972
114058 97.967 -78.703 45.954
114059 97.994 -78.711 45.937
114060 98.018 -78.710 45.920
114061 98.030 -78.715 45.903
114062 97.991 -78.709 45.887
114063 98.012 -78.711 45.872
114064 98.035 -78.715 45.856
114065 98.035 -78.702 45.840
114066 98.042 -78.673 45.824
114067 98.056 -78.645 45.807
114068 98.087 -78.608 45.791
114069 98.130 -78.636 45.775
114070 98.153 -78.667 45.760
114071 98.153 -78.686 45.745
114072 98.133 -78.686 45.729
114073 98.112 -78.687 45.713
114074 98.131 -78.695 45.698
114075 98.153 -78.703 45.684
114076 98.173 -78.713 45.670
114077 98.164 -78.717 45.655
114078 98.157 -78.715 45.639
114079 98.167 -78.680 45.625
114080 98.193 -78.669 45.610
114081 98.222 -78.689 45.594
114082 98.226 -78.709 45.579
114083 98.233 -78.730 45.564
114084 98.245 -78.752 45.548
114085 98.256 -78.773 45.534
114086 98.275 -78.777 45.521
114087 98.299 -78.765 45.509
114088 98.323 -78.754 45.495
114089 98.348 -78.742 45.481
114090 98.372 -78.730 45.467
114091 98.397 -78.719 45.451
114092 98.421 -78.759 45.436
114093 98.433 -78.805 45.420
114094 98.445 -78.849 45.404
114095 98.454 -78.874 45.390
114096 98.460 -78.868 45.379
114097 98.484 -78.867 45.368
114098 98.528 -78.875 45.354
114099 98.573 -78.883 45.341
114100 98.608 -78.883 45.328
114101 98.575 -78.937 45.314
114102 98.478 -79.053 45.302
114103 98.467 -79.098 45.289
114104 98.500 -79.097 45.276
114105 98.545 -79.091 45.263
114106 98.591 -79.086 45.249
114107 98.636 -79.080 45.235
114108 98.681 -79.075 45.221
114109 98.727 -79.068 45.208
114110 98.772 -79.063 45.196
114111 98.817 -79.057 45.183
114112 98.851 -79.034 45.171
114113 98.869 -78.987 45.157
114114 98.887 -78.940 45.143
114115 98.905 -78.894 45.129
114116 98.955 -78.881 45.122
114117 98.994 -78.883 45.114
114118 99.020 -78.890 45.107
114119 99.046 -78.899 45.099
114120 99.071 -78.906 45.090
114121 99.092 -78.916 45.081
114122 99.107 -78.928 45.073
114123 99.125 -78.934 45.063
114124 99.143 -78.940 45.053
114125 99.157 -78.946 45.043
114126 99.171 -78.952 45.031
114127 99.186 -78.959 45.018
114128 99.200 -78.966 45.004
114129 99.221 -78.982 44.991
114130 99.244 -79.000 44.978
114131 99.267 -79.018 44.965
114132 99.290 -79.037 44.952
114133 99.319 -79.051 44.941
114134 99.341 -79.057 44.933
114135 99.353 -79.048 44.927
114136 99.366 -79.038 44.923
114137 99.383 -79.027 44.921
114138 99.420 -79.035 44.921
114139 99.481 -79.088 44.920
114140 99.537 -79.131 44.917
114141 99.573 -79.123 44.913
114142 99.610 -79.116 44.907
114143 99.647 -79.109 44.901
114144 99.634 -79.107 44.894
114145 99.613 -79.119 44.889
114146 99.619 -79.138 44.882
114147 99.624 -79.157 44.873
114148 99.630 -79.177 44.865
114149 99.636 -79.196 44.858
114150 99.642 -79.215 44.850
114151 99.648 -79.234 44.842
114152 99.668 -79.257 44.834
114153 99.696 -79.266 44.826
114154 99.730 -79.298 44.819
114155 99.764 -79.331 44.812
114156 99.798 -79.363 44.805
114157 99.832 -79.396 44.798
114158 99.866 -79.429 44.791
114159 99.900 -79.462 44.784
114160 99.936 -79.496 44.778
114161 99.976 -79.532 44.771
114162 100.008 -79.560 44.764
114163 100.023 -79.574 44.755
114164 100.072 -79.573 44.747
114165 100.115 -79.573 44.739
114166 100.140 -79.589 44.731
114167 100.165 -79.606 44.722
114168 100.191 -79.623 44.713
114169 100.216 -79.639 44.704
114170 100.241 -79.656 44.697
114171 100.266 -79.673 44.689
114172 100.368 -79.657 44.682
114173 100.397 -79.689 44.673
114174 100.416 -79.719 44.662
114175 100.398 -79.742 44.651
114176 100.397 -79.761 44.641
114177 100.455 -79.770 44.628
114178 100.512 -79.780 44.615
114179 100.570 -79.789 44.602
114180 100.618 -79.794 44.588
114181 100.655 -79.793 44.575
114182 100.693 -79.792 44.561
114183 100.730 -79.792 44.547
114184 100.767 -79.794 44.533
114185 100.800 -79.829 44.520
114186 100.833 -79.864 44.508
114187 100.866 -79.899 44.496
114188 100.900 -79.934 44.484
114189 100.932 -79.969 44.471
114190 100.966 -80.005 44.458
114191 100.999 -80.040 44.446
114192 101.032 -80.074 44.433
114193 101.070 -80.101 44.420
114194 101.111 -80.124 44.406
114195 101.151 -80.148 44.393
114196 101.192 -80.171 44.379
114197 101.220 -80.162 44.366
114198 101.250 -80.169 44.352
114199 101.296 -80.181 44.337
114200 101.346 -80.233 44.324
114201 101.413 -80.232 44.309
114202 101.445 -80.249 44.294
114203 101.477 -80.265 44.279
114204 101.509 -80.281 44.263
114205 101.541 -80.297 44.246
114206 101.573 -80.313 44.229
114207 101.605 -80.329 44.213
114208 101.637 -80.345 44.196
114209 101.670 -80.362 44.178
114210 101.701 -80.378 44.160
114211 101.734 -80.394 44.142
114212 101.766 -80.410 44.125
114213 101.798 -80.426 44.107
114548 94.444 -77.963 51.885
114549 94.458 -77.956 51.863
114550 94.472 -77.949 51.841
114551 94.486 -77.943 51.820
114552 94.500 -77.937 51.798
114553 94.501 -77.932 51.775
114554 94.502 -77.928 51.752
114555 94.503 -77.924 51.727
114556 94.505 -77.919 51.703
114557 94.524 -77.910 51.678
114558 94.545 -77.899 51.655
114559 94.565 -77.890 51.630
114560 94.580 -77.891 51.603
114561 94.596 -77.891 51.575
114562 94.629 -77.904 51.548
114563 94.657 -77.915 51.521
114564 94.669 -77.915 51.493
114565 94.681 -77.916 51.466
114566 94.692 -77.917 51.437
114567 94.704 -77.918 51.408
114568 94.715 -77.918 51.378
114569 94.724 -77.910 51.349
114570 94.730 -77.906 51.318
114571 94.730 -77.913 51.288
114572 94.730 -77.921 51.256
114573 94.729 -77.928 51.226
114574 94.736 -77.929 51.194
114575 94.748 -77.934 51.162
114576 94.766 -77.940 51.129
114577 94.790 -77.946 51.095
114578 94.815 -77.954 51.061
114579 94.840 -77.961 51.028
114580 94.864 -77.968 50.994
114581 94.889 -77.975 50.959
114582 94.909 -77.967 50.924
114583 94.928 -77.949 50.889
114584 94.946 -77.933 50.854
114585 94.964 -77.915 50.820
114586 94.977 -77.900 50.784
114587 94.986 -77.887 50.749
114588 94.994 -77.874 50.715
114589 95.003 -77.861 50.680
114590 95.013 -77.847 50.645
114591 95.021 -77.834 50.611
114592 95.030 -77.821 50.577
114593 95.040 -77.810 50.543
114594 95.053 -77.803 50.510
114595 95.066 -77.796 50.478
114596 95.078 -77.789 50.446
114597 95.093 -77.783 50.415
114598 95.119 -77.783 50.383
114599 95.136 -77.801 50.351
114600 95.140 -77.827 50.318
114601 95.120 -77.840 50.286
114602 95.153 -77.891 50.255
114603 95.187 -77.886 50.225
114604 95.225 -77.878 50.196
114605 95.262 -77.871 50.168
114606 95.299 -77.863 50.142
114607 95.336 -77.855 50.115
114608 95.373 -77.848 50.090
114609 95.394 -77.845 50.065
114610 95.408 -77.844 50.041
114611 95.421 -77.843 50.015
114612 95.434 -77.843 49.991
114613 95.448 -77.842 49.966
114614 95.461 -77.841 49.941
114615 95.468 -77.843 49.915
114616 95.474 -77.846 49.885
114617 95.481 -77.848 49.852
114618 95.487 -77.850 49.817
114619 95.493 -77.852 49.781
114620 95.515 -77.873 49.743
114621 95.534 -77.896 49.705
114622 95.516 -77.932 49.670
114623 95.501 -77.964 49.635
114624 95.545 -77.928 49.602
114625 95.548 -77.920 49.569
114626 95.551 -77.911 49.534
114627 95.554 -77.902 49.501
114628 95.557 -77.894 49.468
114629 95.574 -77.877 49.434
114630 95.593 -77.859 49.401
114631 95.598 -77.861 49.367
114632 95.598 -77.880 49.333
114633 95.613 -77.884 49.298
114634 95.629 -77.890 49.263
114635 95.670 -77.881 49.228
114636 95.716 -77.868 49.196
114637 95.725 -77.887 49.164
114638 95.730 -77.907 49.131
114639 95.752 -77.909 49.097
114640 95.773 -77.912 49.062
114641 95.796 -77.915 49.025
114642 95.818 -77.917 48.987
114643 95.839 -77.919 48.949
114644 95.856 -77.928 48.912
114645 95.865 -77.945 48.873
114646 95.874 -77.962 48.832
114647 95.883 -77.980 48.790
114648 95.891 -77.997 48.748
114649 95.888 -78.021 48.706
114650 95.895 -78.019 48.664
114651 95.902 -78.018 48.625
114652 95.913 -78.018 48.586
114653 95.926 -78.019 48.548
114654 95.939 -78.020 48.512
114655 95.952 -78.022 48.477
114656 95.964 -78.022 48.442
114657 95.979 -78.020 48.407
114658 95.995 -78.017 48.376
114659 96.012 -78.014 48.344
114660 96.028 -78.011 48.313
114661 96.044 -78.008 48.282
114662 96.060 -78.005 48.253
114663 96.077 -78.002 48.226
114664 96.093 -77.999 48.199
114665 96.105 -78.002 48.172
114666 96.117 -78.006 48.146
114667 96.128 -78.012 48.119
114668 96.143 -78.018 48.092
114669 96.157 -78.026 48.066
114670 96.172 -78.034 48.039
114671 96.187 -78.045 48.011
114672 96.202 -78.057 47.986
114673 96.218 -78.069 47.961
114674 96.233 -78.081 47.937
114675 96.250 -78.086 47.913
114676 96.268 -78.090 47.888
114677 96.291 -78.098 47.861
114678 96.312 -78.106 47.833
114679 96.334 -78.115 47.808
114680 96.356 -78.123 47.782
114681 96.364 -78.131 47.756
114682 96.360 -78.138 47.731
114683 96.379 -78.150 47.705
114684 96.398 -78.162 47.679
114685 96.417 -78.174 47.653
114686 96.437 -78.186 47.629
114687 96.456 -78.197 47.604
114688 96.471 -78.203 47.577
114689 96.476 -78.210 47.551
114690 96.480 -78.216 47.526
114691 96.482 -78.222 47.500
114692 96.484 -78.228 47.474
114693 96.487 -78.233 47.450
114694 96.489 -78.238 47.426
114695 96.491 -78.244 47.401
114696 96.491 -78.246 47.375
114697 96.489 -78.246 47.351
114698 96.487 -78.246 47.326
114699 96.487 -78.246 47.301
114700 96.503 -78.246 47.277
114701 96.521 -78.250 47.250
114702 96.535 -78.257 47.222
114703 96.540 -78.261 47.193
114704 96.546 -78.266 47.163
114705 96.555 -78.280 47.133
114706 96.565 -78.296 47.104
114707 96.574 -78.311 47.077
114708 96.583 -78.326 47.051
114709 96.592 -78.342 47.025
114710 96.614 -78.434 46.998
114711 96.622 -78.449 46.971
114712 96.627 -78.462 46.943
114713 96.632 -78.476 46.914
114714 96.638 -78.491 46.887
114715 96.640 -78.504 46.860
114716 96.664 -78.517 46.832
114717 96.695 -78.529 46.804
114718 96.725 -78.541 46.777
114719 96.755 -78.553 46.750
114720 96.786 -78.565 46.723
114721 96.816 -78.576 46.697
114722 96.847 -78.589 46.671
114723 96.876 -78.600 46.644
114724 96.903 -78.609 46.616
114725 96.930 -78.617 46.587
114726 96.957 -78.629 46.560
114727 96.984 -78.639 46.533
114728 97.012 -78.651 46.504
114729 97.039 -78.662 46.477
114730 97.079 -78.673 46.451
114731 97.121 -78.685 46.425
114732 97.132 -78.705 46.400
114733 97.141 -78.726 46.376
114734 97.151 -78.747 46.354
114735 97.175 -78.758 46.334
114736 97.215 -78.758 46.316
114737 97.255 -78.758 46.297
114738 97.296 -78.759 46.278
114739 97.336 -78.759 46.262
114740 97.376 -78.760 46.247
114741 97.420 -78.758 46.232
114742 97.479 -78.743 46.217
114743 97.539 -78.729 46.204
114744 97.599 -78.714 46.191
114745 97.658 -78.700 46.177
114746 97.705 -78.711 46.164
114747 97.752 -78.723 46.150
114748 97.790 -78.727 46.136
114749 97.803 -78.725 46.123
114750 97.824 -78.695 46.110
114751 97.828 -78.701 46.097
114752 97.832 -78.707 46.083
114753 97.839 -78.712 46.070
114754 97.860 -78.710 46.054
114755 97.882 -78.708 46.037
114756 97.903 -78.706 46.019
114757 97.924 -78.705 46.003
114758 97.946 -78.703 45.986
114759 97.967 -78.701 45.969
114760 97.992 -78.706 45.951
114761 98.017 -78.709 45.934
114762 98.030 -78.719 45.917
114763 97.984 -78.695 45.900
114764 98.007 -78.699 45.885
114765 98.031 -78.702 45.869
114766 98.048 -78.702 45.852
114767 98.054 -78.672 45.836
114768 98.093 -78.632 45.820
114769 98.144 -78.609 45.803
114770 98.184 -78.649 45.787
114771 98.187 -78.659 45.772
114772 98.176 -78.656 45.757
114773 98.160 -78.654 45.742
114774 98.169 -78.661 45.724
114775 98.177 -78.672 45.709
114776 98.186 -78.683 45.694
114777 98.181 -78.677 45.680
114778 98.173 -78.668 45.664
114779 98.164 -78.659 45.649
114780 98.168 -78.630 45.634
114781 98.187 -78.620 45.619
114782 98.230 -78.640 45.602
114783 98.238 -78.661 45.587
114784 98.249 -78.683 45.572
114785 98.261 -78.705 45.556
114786 98.273 -78.726 45.542
114787 98.284 -78.748 45.527
114788 98.303 -78.749 45.513
114789 98.328 -78.738 45.500
114790 98.352 -78.726 45.486
114791 98.377 -78.714 45.471
114792 98.402 -78.723 45.455
114793 98.427 -78.761 45.440
114794 98.449 -78.802 45.425
114795 98.461 -78.847 45.408
114796 98.467 -78.844 45.395
114797 98.473 -78.837 45.383
114798 98.478 -78.830 45.371
114799 98.507 -78.832 45.358
114800 98.552 -78.839 45.344
114801 98.596 -78.847 45.330
114802 98.592 -78.904 45.317
114803 98.492 -79.022 45.304
114804 98.487 -79.076 45.290
114805 98.530 -79.086 45.276
114806 98.574 -79.093 45.263
114807 98.619 -79.090 45.249
114808 98.664 -79.084 45.235
114809 98.709 -79.079 45.221
114810 98.755 -79.073 45.208
114811 98.800 -79.068 45.196
114812 98.846 -79.062 45.184
114813 98.864 -79.017 45.172
114814 98.882 -78.971 45.159
114815 98.900 -78.924 45.145
114816 98.918 -78.877 45.132
114817 98.960 -78.856 45.125
114818 98.998 -78.858 45.116
114819 99.024 -78.866 45.108
114820 99.050 -78.874 45.100
114821 99.072 -78.883 45.091
114822 99.086 -78.896 45.083
114823 99.101 -78.907 45.075
114824 99.120 -78.913 45.065
114825 99.139 -78.918 45.056
114826 99.156 -78.924 45.046
114827 99.170 -78.930 45.037
114828 99.184 -78.937 45.025
114829 99.199 -78.943 45.013
114830 99.221 -78.962 45.001
114831 99.244 -78.981 44.988
114832 99.268 -78.998 44.973
114833 99.296 -79.012 44.958
114834 99.324 -79.028 44.945
114835 99.352 -79.041 44.938
114836 99.364 -79.031 44.933
114837 99.376 -79.022 44.929
114838 99.389 -79.012 44.929
114839 99.418 -79.038 44.927
114840 99.478 -79.092 44.922
114841 99.529 -79.122 44.916
114842 99.566 -79.114 44.910
114843 99.603 -79.107 44.904
114844 99.639 -79.100 44.897
114845 99.671 -79.093 44.889
114846 99.633 -79.101 44.882
114847 99.639 -79.120 44.874
114848 99.645 -79.139 44.866
114849 99.650 -79.159 44.857
114850 99.656 -79.178 44.850
114851 99.662 -79.197 44.842
114852 99.668 -79.216 44.834
114853 99.674 -79.235 44.826
114854 99.703 -79.268 44.819
114855 99.744 -79.304 44.812
114856 99.784 -79.341 44.805
114857 99.825 -79.378 44.798
114858 99.865 -79.415 44.791
114859 99.906 -79.452 44.784
114860 99.932 -79.474 44.778
114861 99.950 -79.494 44.771
114862 99.969 -79.514 44.765
114863 99.987 -79.534 44.757
114864 100.034 -79.551 44.749
114865 100.087 -79.545 44.742
114866 100.104 -79.564 44.734
114867 100.126 -79.582 44.725
114868 100.151 -79.598 44.717
114869 100.177 -79.615 44.708
114870 100.202 -79.632 44.698
114871 100.227 -79.648 44.690
114872 100.253 -79.665 44.683
114873 100.379 -79.629 44.677
114874 100.407 -79.662 44.668
114875 100.406 -79.688 44.659
114876 100.388 -79.711 44.649
114877 100.392 -79.730 44.640
114878 100.449 -79.739 44.628
114879 100.507 -79.748 44.615
114880 100.564 -79.758 44.603
114881 100.622 -79.767 44.590
114882 100.662 -79.767 44.577
114883 100.700 -79.767 44.564
114884 100.737 -79.766 44.551
114885 100.774 -79.770 44.537
114886 100.807 -79.805 44.524
114887 100.840 -79.841 44.512
114888 100.873 -79.876 44.500
114889 100.906 -79.911 44.488
114890 100.939 -79.946 44.476
114891 100.973 -79.981 44.463
114892 101.006 -80.015 44.451
114893 101.047 -80.038 44.438
114894 101.088 -80.062 44.425
114895 101.128 -80.086 44.412
114896 101.169 -80.109 44.399
114897 101.209 -80.133 44.386
114898 101.250 -80.156 44.373
114899 101.271 -80.130 44.359
114900 101.310 -80.140 44.345
114901 101.359 -80.199 44.331
114902 101.430 -80.194 44.316
114903 101.462 -80.210 44.301
114904 101.494 -80.226 44.286
114905 101.526 -80.243 44.271
114906 101.558 -80.259 44.253
114907 101.591 -80.275 44.236
114908 101.623 -80.291 44.219
114909 101.655 -80.307 44.202
114910 101.687 -80.323 44.184
114911 101.719 -80.340 44.166
114912 101.751 -80.356 44.148
114913 101.783 -80.372 44.130
114914 101.815 -80.388 44.112
115248 94.439 -77.950 51.913
115249 94.453 -77.944 51.891
115250 94.467 -77.937 51.869
115251 94.481 -77.931 51.848
115252 94.495 -77.924 51.825
115253 94.509 -77.918 51.803
115254 94.514 -77.912 51.781
115255 94.516 -77.909 51.758
115256 94.517 -77.904 51.735
115257 94.518 -77.899 51.711
115258 94.527 -77.893 51.686
115259 94.547 -77.883 51.662
115260 94.568 -77.873 51.637
115261 94.584 -77.872 51.611
115262 94.600 -77.872 51.585
115263 94.624 -77.880 51.556
115264 94.656 -77.892 51.529
115265 94.668 -77.893 51.501
115266 94.679 -77.893 51.474
115267 94.691 -77.894 51.446
115268 94.702 -77.894 51.416
115269 94.714 -77.895 51.386
115270 94.724 -77.892 51.356
115271 94.730 -77.887 51.326
115272 94.730 -77.893 51.295
115273 94.730 -77.901 51.264
115274 94.730 -77.909 51.232
115275 94.734 -77.912 51.201
115276 94.746 -77.916 51.168
115277 94.769 -77.923 51.135
115278 94.793 -77.931 51.101
115279 94.818 -77.937 51.067
115280 94.842 -77.945 51.033
115281 94.867 -77.952 50.999
115282 94.892 -77.959 50.964
115283 94.912 -77.949 50.929
115284 94.930 -77.931 50.894
115285 94.948 -77.914 50.859
115286 94.962 -77.899 50.824
115287 94.971 -77.886 50.789
115288 94.980 -77.872 50.754
115289 94.988 -77.859 50.719
115290 94.997 -77.846 50.684
115291 95.006 -77.833 50.649
115292 95.018 -77.824 50.615
115293 95.031 -77.817 50.580
115294 95.043 -77.810 50.546
115295 95.056 -77.802 50.513
115296 95.068 -77.796 50.481
115297 95.081 -77.789 50.448
115298 95.097 -77.783 50.417
115299 95.123 -77.783 50.385
115300 95.150 -77.784 50.352
115301 95.154 -77.801 50.319
115302 95.158 -77.843 50.287
115303 95.163 -77.868 50.256
115304 95.182 -77.874 50.225
115305 95.211 -77.871 50.197
115306 95.249 -77.863 50.169
115307 95.285 -77.855 50.142
115308 95.323 -77.848 50.116
115309 95.360 -77.841 50.091
115310 95.394 -77.834 50.068
115311 95.407 -77.834 50.045
115312 95.420 -77.833 50.021
115313 95.433 -77.832 49.998
115314 95.447 -77.831 49.974
115315 95.460 -77.830 49.949
115316 95.473 -77.830 49.922
115317 95.483 -77.830 49.892
115318 95.489 -77.833 49.859
115319 95.496 -77.835 49.823
115320 95.502 -77.837 49.786
115321 95.522 -77.857 49.748
115322 95.524 -77.886 49.710
115323 95.506 -77.922 49.674
115324 95.488 -77.958 49.639
115325 95.537 -77.919 49.604
115326 95.548 -77.906 49.569
115327 95.551 -77.896 49.534
115328 95.554 -77.888 49.500
115329 95.557 -77.879 49.467
115330 95.560 -77.871 49.434
115331 95.571 -77.857 49.402
115332 95.590 -77.840 49.370
115333 95.589 -77.860 49.336
115334 95.604 -77.865 49.302
115335 95.649 -77.853 49.267
115336 95.694 -77.842 49.234
115337 95.725 -77.843 49.202
115338 95.754 -77.845 49.170
115339 95.740 -77.884 49.138
115340 95.757 -77.891 49.105
115341 95.779 -77.893 49.072
115342 95.801 -77.896 49.037
115343 95.823 -77.898 49.000
115344 95.844 -77.901 48.962
115345 95.853 -77.918 48.924
115346 95.863 -77.935 48.885
115347 95.871 -77.953 48.845
115348 95.880 -77.970 48.802
115349 95.889 -77.987 48.760
115350 95.888 -78.007 48.718
115351 95.895 -78.006 48.678
115352 95.904 -78.005 48.638
115353 95.917 -78.006 48.599
115354 95.930 -78.007 48.561
115355 95.942 -78.008 48.524
115356 95.955 -78.009 48.489
115357 95.968 -78.010 48.455
115358 95.982 -78.009 48.421
115359 95.998 -78.006 48.389
115360 96.014 -78.003 48.358
115361 96.030 -78.000 48.327
115362 96.046 -77.996 48.297
115363 96.062 -77.993 48.268
115364 96.079 -77.990 48.241
115365 96.091 -77.994 48.214
115366 96.102 -77.999 48.189
115367 96.114 -78.004 48.163
115368 96.125 -78.009 48.136
115369 96.140 -78.016 48.110
115370 96.155 -78.024 48.083
115371 96.169 -78.031 48.056
115372 96.184 -78.039 48.031
115373 96.199 -78.050 48.007
115374 96.215 -78.062 47.983
115375 96.232 -78.065 47.959
115376 96.250 -78.068 47.936
115377 96.269 -78.072 47.910
115378 96.291 -78.081 47.883
115379 96.313 -78.089 47.856
115380 96.335 -78.097 47.830
115381 96.353 -78.106 47.804
115382 96.359 -78.112 47.779
115383 96.368 -78.122 47.752
115384 96.387 -78.134 47.724
115385 96.406 -78.147 47.697
115386 96.425 -78.159 47.670
115387 96.444 -78.171 47.645
115388 96.463 -78.183 47.619
115389 96.469 -78.188 47.592
115390 96.471 -78.194 47.565
115391 96.473 -78.200 47.539
115392 96.475 -78.205 47.513
115393 96.477 -78.210 47.487
115394 96.479 -78.216 47.462
115395 96.481 -78.222 47.437
115396 96.484 -78.227 47.411
115397 96.485 -78.232 47.384
115398 96.483 -78.232 47.360
115399 96.481 -78.232 47.335
115400 96.487 -78.232 47.309
115401 96.504 -78.232 47.284
115402 96.523 -78.241 47.258
115403 96.540 -78.247 47.231
115404 96.546 -78.252 47.202
115405 96.552 -78.257 47.173
115406 96.557 -78.262 47.144
115407 96.564 -78.266 47.116
115408 96.572 -78.279 47.089
115409 96.581 -78.294 47.063
115410 96.590 -78.310 47.036
115411 96.614 -78.410 47.010
115412 96.624 -78.442 46.983
115413 96.628 -78.442 46.955
115414 96.633 -78.457 46.927
115415 96.629 -78.469 46.900
115416 96.630 -78.483 46.872
115417 96.663 -78.495 46.845
115418 96.698 -78.504 46.817
115419 96.729 -78.516 46.789
115420 96.759 -78.528 46.762
115421 96.789 -78.540 46.735
115422 96.819 -78.552 46.710
115423 96.850 -78.564 46.684
115424 96.879 -78.575 46.657
115425 96.906 -78.584 46.630
115426 96.933 -78.592 46.602
115427 96.960 -78.601 46.574
115428 96.987 -78.610 46.547
115429 97.015 -78.619 46.519
115430 97.041 -78.630 46.491
115431 97.080 -78.642 46.464
115432 97.090 -78.662 46.438
115433 97.100 -78.683 46.413
115434 97.110 -78.703 46.389
115435 97.119 -78.724 46.367
115436 97.140 -78.738 46.347
115437 97.180 -78.738 46.328
115438 97.220 -78.739 46.309
115439 97.260 -78.739 46.291
115440 97.301 -78.739 46.274
115441 97.351 -78.733 46.258
115442 97.410 -78.718 46.243
115443 97.469 -78.703 46.230
115444 97.529 -78.689 46.217
115445 97.589 -78.674 46.205
115446 97.647 -78.662 46.192
115447 97.694 -78.673 46.179
115448 97.741 -78.685 46.166
115449 97.788 -78.696 46.152
115450 97.818 -78.696 46.139
115451 97.822 -78.702 46.125
115452 97.826 -78.708 46.111
115453 97.830 -78.714 46.097
115454 97.843 -78.716 46.084
115455 97.864 -78.714 46.067
115456 97.885 -78.712 46.049
115457 97.907 -78.711 46.033
115458 97.928 -78.709 46.016
115459 97.949 -78.707 46.000
115460 97.970 -78.705 45.983
115461 97.992 -78.704 45.966
115462 98.017 -78.708 45.949
115463 97.955 -78.679 45.932
115464 97.979 -78.683 45.914
115465 98.003 -78.686 45.899
115466 98.027 -78.690 45.883
115467 98.051 -78.694 45.867
115468 98.098 -78.655 45.850
115469 98.154 -78.608 45.833
115470 98.199 -78.623 45.816
115471 98.196 -78.629 45.799
115472 98.192 -78.633 45.784
115473 98.182 -78.631 45.769
115474 98.179 -78.633 45.754
115475 98.189 -78.644 45.736
115476 98.198 -78.655 45.720
115477 98.196 -78.653 45.704
115478 98.187 -78.644 45.689
115479 98.179 -78.635 45.674
115480 98.171 -78.626 45.658
115481 98.163 -78.617 45.643
115482 98.158 -78.610 45.628
115483 98.245 -78.625 45.612
115484 98.257 -78.626 45.596
115485 98.266 -78.636 45.581
115486 98.278 -78.657 45.565
115487 98.289 -78.679 45.550
115488 98.301 -78.701 45.535
115489 98.312 -78.722 45.519
115490 98.333 -78.722 45.505
115491 98.357 -78.710 45.490
115492 98.381 -78.698 45.475
115493 98.407 -78.726 45.459
115494 98.432 -78.764 45.444
115495 98.457 -78.804 45.429
115496 98.475 -78.821 45.413
115497 98.480 -78.814 45.399
115498 98.485 -78.807 45.386
115499 98.491 -78.800 45.374
115500 98.496 -78.793 45.361
115501 98.539 -78.799 45.347
115502 98.606 -78.810 45.333
115503 98.536 -78.909 45.320
115504 98.497 -78.987 45.306
115505 98.529 -79.015 45.291
115506 98.565 -79.038 45.277
115507 98.608 -79.045 45.264
115508 98.652 -79.052 45.249
115509 98.696 -79.059 45.235
115510 98.739 -79.066 45.220
115511 98.783 -79.073 45.208
115512 98.828 -79.072 45.196
115513 98.860 -79.047 45.184
115514 98.878 -79.000 45.173
115515 98.896 -78.954 45.160
115516 98.914 -78.908 45.147
115517 98.932 -78.861 45.134
115518 98.966 -78.831 45.125
115519 99.002 -78.834 45.115
115520 99.028 -78.843 45.106
115521 99.051 -78.851 45.098
115522 99.066 -78.862 45.090
115523 99.081 -78.874 45.082
115524 99.096 -78.886 45.074
115525 99.115 -78.892 45.065
115526 99.134 -78.897 45.057
115527 99.153 -78.902 45.048
115528 99.169 -78.908 45.038
115529 99.183 -78.915 45.028
115530 99.199 -78.924 45.016
115531 99.222 -78.942 45.004
115532 99.246 -78.960 44.992
115533 99.274 -78.974 44.977
115534 99.302 -78.989 44.963
115535 99.330 -79.003 44.950
115536 99.358 -79.018 44.942
115537 99.375 -79.015 44.937
115538 99.387 -79.006 44.933
115539 99.400 -78.996 44.930
115540 99.423 -79.018 44.925
115541 99.478 -79.089 44.919
115542 99.521 -79.112 44.913
115543 99.558 -79.105 44.906
115544 99.595 -79.098 44.899
115545 99.632 -79.091 44.892
115546 99.668 -79.084 44.884
115547 99.653 -79.083 44.875
115548 99.659 -79.102 44.866
115549 99.665 -79.121 44.858
115550 99.671 -79.140 44.850
115551 99.677 -79.159 44.842
115552 99.683 -79.178 44.835
115553 99.707 -79.206 44.827
115554 99.744 -79.240 44.819
115555 99.771 -79.284 44.812
115556 99.794 -79.334 44.804
115557 99.835 -79.371 44.797
115558 99.858 -79.394 44.791
115559 99.876 -79.414 44.784
115560 99.895 -79.434 44.778
115561 99.913 -79.454 44.771
115562 99.932 -79.474 44.765
115563 99.950 -79.494 44.759
115564 99.994 -79.511 44.751
115565 100.048 -79.528 44.744
115566 100.102 -79.529 44.737
115567 100.119 -79.548 44.729
115568 100.135 -79.567 44.720
115569 100.152 -79.586 44.712
115570 100.169 -79.605 44.703
115571 100.188 -79.624 44.695
115572 100.213 -79.641 44.687
115573 100.239 -79.657 44.680
115574 100.343 -79.631 44.673
115575 100.415 -79.635 44.665
115576 100.397 -79.657 44.656
115577 100.379 -79.679 44.646
115578 100.385 -79.697 44.638
115579 100.443 -79.707 44.627
115580 100.501 -79.717 44.615
115581 100.559 -79.726 44.604
115582 100.616 -79.736 44.592
115583 100.670 -79.742 44.580
115584 100.707 -79.742 44.568
115585 100.744 -79.741 44.555
115586 100.781 -79.747 44.542
115587 100.814 -79.783 44.529
115588 100.847 -79.817 44.517
115589 100.880 -79.852 44.505
115590 100.913 -79.887 44.493
115591 100.947 -79.923 44.481
115592 100.983 -79.952 44.468
115593 101.024 -79.976 44.457
115594 101.064 -79.999 44.444
115595 101.105 -80.023 44.431
115596 101.145 -80.046 44.419
115597 101.186 -80.070 44.407
115598 101.226 -80.093 44.395
115599 101.266 -80.117 44.382
115600 101.306 -80.138 44.368
115601 101.323 -80.098 44.353
115602 101.373 -80.165 44.339
115603 101.447 -80.156 44.324
115604 101.479 -80.172 44.309
115605 101.512 -80.188 44.293
115606 101.544 -80.204 44.277
115607 101.575 -80.221 44.260
115608 101.608 -80.237 44.242
115609 101.640 -80.253 44.225
115610 101.672 -80.269 44.208
115611 101.704 -80.285 44.189
115612 101.736 -80.301 44.171
115613 101.768 -80.317 44.154
115614 101.800 -80.334 44.135
115948 94.433 -77.938 51.942
115949 94.448 -77.931 51.920
115950 94.462 -77.925 51.898
115951 94.476 -77.918 51.876
115952 94.490 -77.912 51.855
115953 94.504 -77.905 51.833
115954 94.519 -77.899 51.812
115955 94.527 -77.893 51.790
115956 94.528 -77.889 51.767
115957 94.529 -77.884 51.743
115958 94.530 -77.881 51.719
115959 94.531 -77.876 51.695
115960 94.550 -77.866 51.671
115961 94.570 -77.856 51.647
115962 94.588 -77.853 51.621
115963 94.603 -77.853 51.595
115964 94.620 -77.855 51.567
115965 94.654 -77.869 51.539
115966 94.666 -77.870 51.512
115967 94.678 -77.871 51.484
115968 94.689 -77.871 51.456
115969 94.701 -77.871 51.427
115970 94.712 -77.872 51.397
115971 94.724 -77.873 51.367
115972 94.731 -77.867 51.336
115973 94.731 -77.874 51.305
115974 94.731 -77.881 51.274
115975 94.731 -77.889 51.242
115976 94.731 -77.895 51.210
115977 94.747 -77.899 51.175
115978 94.771 -77.906 51.142
115979 94.796 -77.914 51.108
115980 94.821 -77.921 51.074
115981 94.845 -77.928 51.040
115982 94.870 -77.936 51.006
115983 94.895 -77.943 50.971
115984 94.914 -77.931 50.935
115985 94.932 -77.913 50.899
115986 94.947 -77.897 50.863
115987 94.956 -77.884 50.828
115988 94.965 -77.871 50.794
115989 94.973 -77.858 50.759
115990 94.982 -77.845 50.724
115991 94.995 -77.837 50.688
115992 95.008 -77.830 50.653
115993 95.021 -77.824 50.617
115994 95.033 -77.817 50.583
115995 95.046 -77.809 50.548
115996 95.059 -77.802 50.516
115997 95.071 -77.796 50.483
115998 95.084 -77.788 50.449
115999 95.101 -77.783 50.416
116000 95.127 -77.784 50.384
116001 95.138 -77.799 50.352
116002 95.158 -77.800 50.320
116003 95.169 -77.815 50.289
116004 95.174 -77.840 50.257
116005 95.188 -77.854 50.227
116006 95.210 -77.855 50.198
116007 95.235 -77.855 50.170
116008 95.272 -77.848 50.144
116009 95.310 -77.841 50.118
116010 95.346 -77.834 50.094
116011 95.384 -77.826 50.073
116012 95.406 -77.822 50.052
116013 95.420 -77.821 50.031
116014 95.433 -77.821 50.007
116015 95.446 -77.820 49.983
116016 95.460 -77.819 49.958
116017 95.473 -77.818 49.931
116018 95.486 -77.818 49.899
116019 95.498 -77.818 49.865
116020 95.505 -77.820 49.828
116021 95.511 -77.822 49.791
116022 95.530 -77.841 49.754
116023 95.514 -77.876 49.716
116024 95.496 -77.912 49.680
116025 95.478 -77.948 49.644
116026 95.524 -77.913 49.608
116027 95.548 -77.890 49.572
116028 95.550 -77.882 49.536
116029 95.553 -77.873 49.502
116030 95.556 -77.865 49.467
116031 95.559 -77.856 49.435
116032 95.562 -77.847 49.404
116033 95.568 -77.837 49.373
116034 95.581 -77.840 49.340
116035 95.620 -77.833 49.306
116036 95.651 -77.834 49.272
116037 95.681 -77.835 49.239
116038 95.711 -77.836 49.208
116039 95.742 -77.837 49.175
116040 95.769 -77.841 49.143
116041 95.762 -77.872 49.112
116042 95.784 -77.874 49.081
116043 95.806 -77.877 49.048
116044 95.828 -77.879 49.013
116045 95.842 -77.891 48.974
116046 95.851 -77.908 48.936
116047 95.860 -77.925 48.898
116048 95.869 -77.943 48.858
116049 95.878 -77.960 48.815
116050 95.887 -77.978 48.773
116051 95.888 -77.993 48.732
116052 95.895 -77.991 48.691
116053 95.907 -77.993 48.652
116054 95.920 -77.993 48.613
116055 95.933 -77.994 48.575
116056 95.946 -77.996 48.537
116057 95.958 -77.996 48.502
116058 95.971 -77.998 48.468
116059 95.984 -77.998 48.435
116060 96.001 -77.995 48.403
116061 96.017 -77.992 48.372
116062 96.033 -77.989 48.343
116063 96.049 -77.986 48.314
116064 96.065 -77.982 48.284
116065 96.077 -77.987 48.257
116066 96.088 -77.992 48.232
116067 96.100 -77.996 48.207
116068 96.111 -78.001 48.181
116069 96.123 -78.006 48.154
116070 96.137 -78.014 48.128
116071 96.152 -78.022 48.101
116072 96.167 -78.029 48.075
116073 96.182 -78.037 48.051
116074 96.197 -78.041 48.028
116075 96.215 -78.044 48.004
116076 96.232 -78.047 47.981
116077 96.250 -78.050 47.957
116078 96.270 -78.056 47.930
116079 96.292 -78.064 47.904
116080 96.314 -78.072 47.877
116081 96.331 -78.080 47.851
116082 96.337 -78.087 47.825
116083 96.350 -78.099 47.800
116084 96.367 -78.113 47.772
116085 96.376 -78.133 47.742
116086 96.395 -78.145 47.714
116087 96.414 -78.157 47.687
116088 96.433 -78.169 47.661
116089 96.452 -78.181 47.634
116090 96.462 -78.187 47.607
116091 96.464 -78.184 47.580
116092 96.466 -78.183 47.553
116093 96.468 -78.188 47.526
116094 96.470 -78.194 47.499
116095 96.472 -78.200 47.474
116096 96.474 -78.205 47.449
116097 96.476 -78.210 47.422
116098 96.479 -78.216 47.395
116099 96.477 -78.216 47.369
116100 96.476 -78.216 47.343
116101 96.487 -78.216 47.317
116102 96.506 -78.222 47.291
116103 96.524 -78.231 47.266
116104 96.543 -78.238 47.240
116105 96.551 -78.244 47.212
116106 96.557 -78.248 47.184
116107 96.563 -78.253 47.156
116108 96.569 -78.257 47.128
116109 96.575 -78.262 47.101
116110 96.581 -78.266 47.075
116111 96.588 -78.276 47.049
116112 96.613 -78.386 47.022
116113 96.627 -78.434 46.995
116114 96.623 -78.421 46.968
116115 96.618 -78.433 46.941
116116 96.619 -78.448 46.913
116117 96.625 -78.461 46.886
116118 96.662 -78.474 46.859
116119 96.697 -78.483 46.830
116120 96.732 -78.492 46.802
116121 96.762 -78.504 46.775
116122 96.793 -78.516 46.748
116123 96.823 -78.528 46.722
116124 96.853 -78.540 46.696
116125 96.881 -78.550 46.670
116126 96.908 -78.559 46.642
116127 96.936 -78.567 46.616
116128 96.963 -78.576 46.588
116129 96.990 -78.585 46.561
116130 97.017 -78.594 46.533
116131 97.044 -78.605 46.505
116132 97.052 -78.620 46.479
116133 97.058 -78.639 46.453
116134 97.068 -78.660 46.427
116135 97.078 -78.680 46.404
116136 97.088 -78.702 46.382
116137 97.104 -78.717 46.360
116138 97.145 -78.717 46.341
116139 97.185 -78.718 46.323
116140 97.225 -78.718 46.306
116141 97.281 -78.708 46.290
116142 97.340 -78.692 46.273
116143 97.400 -78.678 46.258
116144 97.459 -78.664 46.245
116145 97.519 -78.649 46.233
116146 97.579 -78.635 46.220
116147 97.636 -78.624 46.206
116148 97.683 -78.636 46.193
116149 97.730 -78.647 46.180
116150 97.779 -78.661 46.167
116151 97.816 -78.699 46.155
116152 97.820 -78.709 46.141
116153 97.824 -78.714 46.128
116154 97.828 -78.720 46.113
116155 97.846 -78.720 46.097
116156 97.867 -78.718 46.080
116157 97.889 -78.717 46.064
116158 97.910 -78.715 46.048
116159 97.931 -78.713 46.031
116160 97.953 -78.711 46.015
116161 97.974 -78.710 45.999
116162 97.996 -78.708 45.981
116163 98.018 -78.702 45.964
116164 97.957 -78.662 45.947
116165 97.985 -78.664 45.929
116166 98.013 -78.665 45.913
116167 98.040 -78.667 45.897
116168 98.068 -78.668 45.880
116169 98.144 -78.627 45.863
116170 98.208 -78.593 45.845
116171 98.205 -78.598 45.828
116172 98.201 -78.603 45.812
116173 98.197 -78.608 45.795
116174 98.188 -78.607 45.780
116175 98.200 -78.616 45.764
116176 98.209 -78.626 45.747
116177 98.210 -78.629 45.730
116178 98.202 -78.620 45.714
116179 98.194 -78.611 45.699
116180 98.185 -78.601 45.684
116181 98.189 -78.601 45.668
116182 98.197 -78.605 45.653
116183 98.204 -78.608 45.637
116184 98.271 -78.623 45.621
116185 98.283 -78.623 45.606
116186 98.291 -78.630 45.590
116187 98.300 -78.637 45.575
116188 98.308 -78.644 45.559
116189 98.317 -78.654 45.543
116190 98.329 -78.675 45.527
116191 98.341 -78.697 45.511
116192 98.361 -78.694 45.496
116193 98.386 -78.690 45.481
116194 98.412 -78.730 45.464
116195 98.437 -78.760 45.449
116196 98.462 -78.777 45.433
116197 98.490 -78.802 45.417
116198 98.506 -78.811 45.403
116199 98.511 -78.803 45.389
116200 98.516 -78.795 45.377
116201 98.518 -78.781 45.365
116202 98.550 -78.761 45.350
116203 98.580 -78.796 45.336
116204 98.506 -78.896 45.320
116205 98.538 -78.925 45.305
116206 98.571 -78.954 45.291
116207 98.603 -78.983 45.278
116208 98.643 -78.998 45.265
116209 98.687 -79.005 45.251
116210 98.730 -79.012 45.236
116211 98.774 -79.018 45.221
116212 98.818 -79.025 45.208
116213 98.861 -79.031 45.196
116214 98.883 -79.022 45.185
116215 98.891 -78.984 45.174
116216 98.909 -78.937 45.161
116217 98.927 -78.891 45.148
116218 98.945 -78.844 45.135
116219 98.972 -78.806 45.123
116220 99.006 -78.810 45.114
116221 99.031 -78.818 45.106
116222 99.046 -78.830 45.097
116223 99.061 -78.842 45.090
116224 99.076 -78.854 45.081
116225 99.090 -78.865 45.073
116226 99.110 -78.871 45.065
116227 99.129 -78.876 45.056
116228 99.147 -78.881 45.046
116229 99.166 -78.886 45.037
116230 99.182 -78.893 45.027
116231 99.200 -78.904 45.016
116232 99.224 -78.921 45.005
116233 99.252 -78.936 44.994
116234 99.280 -78.950 44.980
116235 99.308 -78.965 44.966
116236 99.336 -78.980 44.953
116237 99.364 -78.994 44.944
116238 99.385 -78.999 44.938
116239 99.398 -78.990 44.934
116240 99.394 -78.958 44.928
116241 99.428 -78.997 44.922
116242 99.491 -79.048 44.915
116243 99.518 -79.071 44.907
116244 99.553 -79.081 44.900
116245 99.587 -79.089 44.893
116246 99.624 -79.082 44.886
116247 99.660 -79.075 44.878
116248 99.688 -79.068 44.868
116249 99.680 -79.084 44.859
116250 99.686 -79.103 44.851
116251 99.691 -79.122 44.843
116252 99.712 -79.148 44.836
116253 99.749 -79.181 44.828
116254 99.785 -79.215 44.820
116255 99.821 -79.248 44.812
116256 99.845 -79.284 44.805
116257 99.835 -79.302 44.797
116258 99.822 -79.353 44.790
116259 99.840 -79.374 44.784
116260 99.858 -79.394 44.778
116261 99.876 -79.413 44.772
116262 99.895 -79.434 44.766
116263 99.913 -79.454 44.759
116264 99.954 -79.471 44.752
116265 100.008 -79.488 44.746
116266 100.062 -79.504 44.739
116267 100.116 -79.513 44.732
116268 100.133 -79.532 44.725
116269 100.150 -79.551 44.717
116270 100.167 -79.570 44.708
116271 100.184 -79.589 44.700
116272 100.200 -79.608 44.691
116273 100.217 -79.628 44.684
116274 100.234 -79.647 44.676
116275 100.260 -79.659 44.669
116276 100.405 -79.603 44.662
116277 100.388 -79.626 44.653
116278 100.369 -79.648 44.644
116279 100.380 -79.665 44.635
116280 100.437 -79.675 44.625
116281 100.495 -79.685 44.615
116282 100.553 -79.694 44.605
116283 100.611 -79.704 44.594
116284 100.668 -79.713 44.583
116285 100.714 -79.717 44.571
116286 100.751 -79.716 44.558
116287 100.787 -79.724 44.547
116288 100.821 -79.759 44.535
116289 100.854 -79.794 44.522
116290 100.887 -79.829 44.511
116291 100.920 -79.864 44.499
116292 100.960 -79.889 44.488
116293 101.000 -79.913 44.475
116294 101.041 -79.936 44.464
116295 101.081 -79.960 44.452
116296 101.121 -79.983 44.441
116297 101.162 -80.008 44.429
116298 101.202 -80.031 44.417
116299 101.243 -80.055 44.405
116300 101.283 -80.078 44.392
116301 101.324 -80.102 44.378
116302 101.356 -80.106 44.364
116303 101.387 -80.131 44.348
116304 101.465 -80.118 44.333
116305 101.496 -80.134 44.317
116306 101.529 -80.150 44.300
116307 101.561 -80.166 44.284
116308 101.593 -80.182 44.266
116309 101.625 -80.198 44.248
116310 101.657 -80.215 44.230
116311 101.689 -80.231 44.213
116312 101.722 -80.247 44.195
116313 101.753 -80.263 44.176
116314 101.785 -80.279 44.159
116315 101.818 -80.295 44.140
116648 94.429 -77.926 51.973
116649 94.443 -77.919 51.951
116650 94.457 -77.912 51.928
116651 94.471 -77.906 51.906
116652 94.485 -77.899 51.884
116653 94.499 -77.893 51.862
116654 94.514 -77.886 51.842
116655 94.528 -77.880 51.821
116656 94.540 -77.874 51.799
116657 94.541 -77.869 51.776
116658 94.542 -77.865 51.752
116659 94.543 -77.861 51.729
116660 94.544 -77.856 51.706
116661 94.552 -77.849 51.682
116662 94.573 -77.840 51.658
116663 94.592 -77.834 51.633
116664 94.607 -77.834 51.606
116665 94.622 -77.835 51.579
116666 94.653 -77.846 51.551
116667 94.665 -77.847 51.524
116668 94.676 -77.848 51.496
116669 94.688 -77.848 51.467
116670 94.699 -77.849 51.439
116671 94.711 -77.849 51.410
116672 94.722 -77.850 51.379
116673 94.732 -77.847 51.348
116674 94.731 -77.855 51.317
116675 94.731 -77.862 51.285
116676 94.731 -77.869 51.253
116677 94.731 -77.877 51.220
116678 94.749 -77.883 51.185
116679 94.774 -77.890 51.151
116680 94.799 -77.897 51.117
116681 94.823 -77.905 51.082
116682 94.848 -77.912 51.047
116683 94.873 -77.919 51.013
116684 94.897 -77.926 50.978
116685 94.916 -77.912 50.941
116686 94.932 -77.896 50.905
116687 94.941 -77.883 50.869
116688 94.950 -77.869 50.833
116689 94.960 -77.859 50.798
116690 94.973 -77.852 50.764
116691 94.985 -77.844 50.728
116692 94.998 -77.837 50.692
116693 95.011 -77.830 50.656
116694 95.023 -77.823 50.622
116695 95.036 -77.816 50.587
116696 95.049 -77.809 50.551
116697 95.061 -77.802 50.519
116698 95.074 -77.795 50.485
116699 95.087 -77.788 50.451
116700 95.102 -77.786 50.418
116701 95.122 -77.787 50.386
116702 95.141 -77.787 50.354
116703 95.161 -77.788 50.322
116704 95.180 -77.789 50.291
116705 95.185 -77.814 50.260
116706 95.194 -77.834 50.230
116707 95.216 -77.836 50.201
116708 95.240 -77.837 50.173
116709 95.262 -77.839 50.146
116710 95.296 -77.834 50.121
116711 95.333 -77.826 50.096
116712 95.371 -77.818 50.074
116713 95.406 -77.812 50.054
116714 95.419 -77.811 50.033
116715 95.432 -77.810 50.010
116716 95.446 -77.809 49.986
116717 95.459 -77.808 49.961
116718 95.472 -77.808 49.935
116719 95.485 -77.807 49.903
116720 95.498 -77.806 49.869
116721 95.512 -77.805 49.834
116722 95.520 -77.807 49.797
116723 95.522 -77.830 49.761
116724 95.504 -77.866 49.725
116725 95.486 -77.902 49.688
116726 95.468 -77.938 49.650
116727 95.511 -77.907 49.614
116728 95.547 -77.876 49.579
116729 95.550 -77.868 49.543
116730 95.553 -77.859 49.508
116731 95.556 -77.850 49.473
116732 95.559 -77.841 49.439
116733 95.562 -77.833 49.407
116734 95.571 -77.825 49.376
116735 95.591 -77.824 49.345
116736 95.612 -77.826 49.311
116737 95.637 -77.827 49.278
116738 95.668 -77.828 49.245
116739 95.698 -77.830 49.213
116740 95.728 -77.830 49.182
116741 95.758 -77.832 49.151
116742 95.784 -77.837 49.120
116743 95.790 -77.855 49.090
116744 95.812 -77.858 49.057
116745 95.831 -77.864 49.023
116746 95.840 -77.881 48.985
116747 95.849 -77.898 48.950
116748 95.858 -77.915 48.912
116749 95.867 -77.933 48.871
116750 95.876 -77.950 48.830
116751 95.884 -77.967 48.789
116752 95.888 -77.979 48.748
116753 95.898 -77.979 48.706
116754 95.911 -77.980 48.667
116755 95.923 -77.981 48.628
116756 95.936 -77.982 48.589
116757 95.949 -77.984 48.553
116758 95.962 -77.984 48.517
116759 95.974 -77.985 48.482
116760 95.987 -77.987 48.449
116761 96.003 -77.984 48.418
116762 96.019 -77.981 48.389
116763 96.035 -77.978 48.359
116764 96.051 -77.975 48.330
116765 96.062 -77.979 48.301
116766 96.074 -77.984 48.275
116767 96.086 -77.989 48.252
116768 96.097 -77.994 48.227
116769 96.109 -77.999 48.200
116770 96.120 -78.004 48.173
116771 96.135 -78.012 48.146
116772 96.150 -78.019 48.120
116773 96.166 -78.021 48.095
116774 96.182 -78.022 48.071
116775 96.198 -78.023 48.048
116776 96.215 -78.025 48.024
116777 96.232 -78.028 47.999
116778 96.250 -78.031 47.974
116779 96.271 -78.038 47.948
116780 96.293 -78.047 47.922
116781 96.310 -78.055 47.896
116782 96.317 -78.062 47.870
116783 96.333 -78.076 47.843
116784 96.350 -78.090 47.817
116785 96.366 -78.103 47.789
116786 96.372 -78.126 47.760
116787 96.384 -78.144 47.732
116788 96.402 -78.156 47.703
116789 96.421 -78.169 47.676
116790 96.441 -78.181 47.649
116791 96.456 -78.191 47.622
116792 96.458 -78.188 47.595
116793 96.460 -78.184 47.567
116794 96.461 -78.182 47.539
116795 96.463 -78.179 47.512
116796 96.465 -78.182 47.486
116797 96.467 -78.188 47.461
116798 96.469 -78.194 47.434
116799 96.471 -78.199 47.407
116800 96.472 -78.202 47.380
116801 96.471 -78.202 47.353
116802 96.488 -78.205 47.327
116803 96.507 -78.213 47.301
116804 96.526 -78.221 47.275
116805 96.544 -78.228 47.249
116806 96.556 -78.235 47.221
116807 96.562 -78.239 47.194
116808 96.568 -78.244 47.167
116809 96.574 -78.248 47.140
116810 96.580 -78.253 47.115
116811 96.586 -78.258 47.088
116812 96.588 -78.270 47.062
116813 96.604 -78.351 47.036
116814 96.614 -78.411 47.009
116815 96.607 -78.395 46.982
116816 96.609 -78.412 46.955
116817 96.610 -78.426 46.928
116818 96.624 -78.440 46.900
116819 96.661 -78.453 46.872
116820 96.695 -78.462 46.844
116821 96.731 -78.471 46.816
116822 96.765 -78.479 46.788
116823 96.796 -78.492 46.761
116824 96.826 -78.504 46.734
116825 96.857 -78.516 46.707
116826 96.884 -78.525 46.680
116827 96.911 -78.534 46.654
116828 96.939 -78.542 46.629
116829 96.966 -78.551 46.603
116830 96.992 -78.563 46.576
116831 97.017 -78.579 46.549
116832 97.042 -78.594 46.522
116833 97.051 -78.609 46.495
116834 97.045 -78.623 46.468
116835 97.040 -78.639 46.443
116836 97.046 -78.658 46.420
116837 97.056 -78.678 46.398
116838 97.069 -78.697 46.376
116839 97.109 -78.698 46.356
116840 97.151 -78.697 46.338
116841 97.211 -78.682 46.321
116842 97.271 -78.667 46.305
116843 97.330 -78.653 46.288
116844 97.389 -78.639 46.273
116845 97.449 -78.624 46.259
116846 97.509 -78.610 46.246
116847 97.568 -78.595 46.233
116848 97.625 -78.586 46.220
116849 97.672 -78.598 46.206
116850 97.720 -78.611 46.193
116851 97.780 -78.643 46.180
116852 97.817 -78.682 46.167
116853 97.818 -78.716 46.155
116854 97.822 -78.721 46.141
116855 97.828 -78.727 46.126
116856 97.850 -78.725 46.110
116857 97.871 -78.723 46.094
116858 97.892 -78.721 46.078
116859 97.914 -78.719 46.062
116860 97.935 -78.717 46.045
116861 97.956 -78.716 46.029
116862 97.984 -78.667 46.013
116863 98.010 -78.678 45.995
116864 98.038 -78.689 45.978
116865 97.976 -78.636 45.961
116866 98.004 -78.638 45.944
116867 98.031 -78.639 45.927
116868 98.059 -78.641 45.910
116869 98.092 -78.638 45.894
116870 98.170 -78.595 45.876
116871 98.201 -78.580 45.857
116872 98.205 -78.576 45.840
116873 98.206 -78.577 45.824
116874 98.202 -78.582 45.807
116875 98.195 -78.582 45.791
116876 98.220 -78.598 45.775
116877 98.225 -78.604 45.758
116878 98.217 -78.595 45.741
116879 98.213 -78.590 45.724
116880 98.220 -78.593 45.709
116881 98.228 -78.597 45.694
116882 98.235 -78.601 45.678
116883 98.243 -78.604 45.662
116884 98.250 -78.608 45.646
116885 98.296 -78.621 45.630
116886 98.308 -78.620 45.615
116887 98.317 -78.627 45.600
116888 98.325 -78.634 45.584
116889 98.334 -78.641 45.568
116890 98.343 -78.648 45.552
116891 98.351 -78.655 45.535
116892 98.359 -78.662 45.519
116893 98.368 -78.671 45.504
116894 98.391 -78.689 45.488
116895 98.416 -78.706 45.470
116896 98.441 -78.724 45.454
116897 98.466 -78.742 45.438
116898 98.495 -78.768 45.423
116899 98.524 -78.796 45.408
116900 98.537 -78.801 45.395
116901 98.542 -78.793 45.382
116902 98.533 -78.748 45.368
116903 98.560 -78.724 45.354
116904 98.523 -78.801 45.339
116905 98.548 -78.835 45.323
116906 98.580 -78.864 45.308
116907 98.613 -78.893 45.294
116908 98.645 -78.921 45.280
116909 98.677 -78.950 45.266
116910 98.721 -78.957 45.252
116911 98.765 -78.964 45.238
116912 98.809 -78.971 45.223
116913 98.853 -78.977 45.208
116914 98.857 -78.962 45.197
116915 98.886 -78.956 45.185
116916 98.926 -78.968 45.174
116917 98.934 -78.921 45.162
116918 98.942 -78.874 45.149
116919 98.958 -78.828 45.136
116920 98.977 -78.782 45.124
116921 99.010 -78.786 45.114
116922 99.025 -78.797 45.105
116923 99.041 -78.809 45.096
116924 99.055 -78.821 45.089
116925 99.070 -78.833 45.080
116926 99.085 -78.844 45.071
116927 99.104 -78.849 45.064
116928 99.123 -78.855 45.054
116929 99.142 -78.860 45.044
116930 99.161 -78.865 45.035
116931 99.180 -78.871 45.025
116932 99.202 -78.883 45.015
116933 99.229 -78.897 45.004
116934 99.258 -78.912 44.993
116935 99.285 -78.927 44.982
116936 99.313 -78.941 44.969
116937 99.342 -78.955 44.955
116938 99.369 -78.970 44.944
116939 99.397 -78.984 44.937
116940 99.409 -78.974 44.932
116941 99.386 -78.917 44.925
116942 99.434 -78.976 44.918
116943 99.512 -78.976 44.910
116944 99.517 -79.020 44.902
116945 99.551 -79.030 44.893
116946 99.585 -79.040 44.887
116947 99.620 -79.050 44.879
116948 99.653 -79.060 44.871
116949 99.690 -79.059 44.860
116950 99.700 -79.065 44.851
116951 99.717 -79.090 44.842
116952 99.754 -79.123 44.834
116953 99.781 -79.158 44.827
116954 99.803 -79.193 44.821
116955 99.823 -79.223 44.814
116956 99.836 -79.229 44.807
116957 99.849 -79.234 44.799
116958 99.846 -79.260 44.791
116959 99.825 -79.315 44.785
116960 99.822 -79.353 44.779
116961 99.840 -79.373 44.773
116962 99.858 -79.394 44.767
116963 99.876 -79.413 44.760
116964 99.914 -79.432 44.754
116965 99.968 -79.448 44.747
116966 100.021 -79.465 44.741
116967 100.075 -79.481 44.735
116968 100.129 -79.498 44.729
116969 100.148 -79.516 44.722
116970 100.163 -79.535 44.714
116971 100.179 -79.555 44.705
116972 100.195 -79.576 44.697
116973 100.210 -79.595 44.689
116974 100.226 -79.615 44.681
116975 100.254 -79.624 44.673
116976 100.292 -79.624 44.665
116977 100.331 -79.624 44.658
116978 100.439 -79.589 44.650
116979 100.392 -79.612 44.642
116980 100.374 -79.633 44.632
116981 100.432 -79.643 44.623
116982 100.489 -79.652 44.614
116983 100.547 -79.662 44.604
116984 100.605 -79.672 44.594
116985 100.662 -79.681 44.583
116986 100.720 -79.691 44.573
116987 100.759 -79.690 44.562
116988 100.794 -79.701 44.551
116989 100.828 -79.736 44.539
116990 100.861 -79.770 44.527
116991 100.895 -79.804 44.516
116992 100.936 -79.827 44.505
116993 100.976 -79.851 44.494
116994 101.017 -79.874 44.483
116995 101.057 -79.898 44.472
116996 101.098 -79.921 44.461
116997 101.138 -79.945 44.451
116998 101.179 -79.968 44.440
116999 101.219 -79.993 44.427
117000 101.258 -80.021 44.415
117001 101.298 -80.049 44.402
117002 101.337 -80.078 44.389
117003 101.376 -80.106 44.374
117004 101.401 -80.097 44.358
117005 101.481 -80.082 44.342
117006 101.514 -80.096 44.325
117007 101.546 -80.112 44.307
117008 101.578 -80.127 44.289
117009 101.610 -80.144 44.272
117010 101.642 -80.160 44.253
117011 101.674 -80.176 44.235
117012 101.706 -80.193 44.218
117013 101.739 -80.209 44.200
117014 101.771 -80.225 44.182
117015 101.802 -80.240 44.163
117016 101.835 -80.257 44.144
117348 94.424 -77.915 52.003
117349 94.438 -77.907 51.981
117350 94.452 -77.900 51.960
117351 94.466 -77.893 51.938
117352 94.480 -77.887 51.916
117353 94.494 -77.881 51.894
117354 94.509 -77.874 51.873
117355 94.523 -77.867 51.852
117356 94.537 -77.861 51.831
117357 94.551 -77.854 51.810
117358 94.553 -77.849 51.787
117359 94.554 -77.845 51.764
117360 94.555 -77.841 51.742
117361 94.556 -77.837 51.719
117362 94.557 -77.832 51.694
117363 94.575 -77.823 51.670
117364 94.595 -77.815 51.645
117365 94.610 -77.816 51.619
117366 94.626 -77.816 51.593
117367 94.649 -77.822 51.565
117368 94.663 -77.824 51.538
117369 94.675 -77.825 51.509
117370 94.686 -77.825 51.480
117371 94.698 -77.826 51.451
117372 94.709 -77.827 51.424
117373 94.721 -77.827 51.394
117374 94.732 -77.827 51.363
117375 94.745 -77.830 51.331
117376 94.777 -77.825 51.299
117377 94.810 -77.821 51.265
117378 94.819 -77.862 51.231
117379 94.810 -77.834 51.197
117380 94.805 -77.847 51.162
117381 94.806 -77.877 51.126
117382 94.826 -77.888 51.090
117383 94.851 -77.896 51.055
117384 94.876 -77.902 51.021
117385 94.900 -77.910 50.985
117386 94.917 -77.894 50.949
117387 94.925 -77.881 50.913
117388 94.937 -77.872 50.876
117389 94.950 -77.865 50.839
117390 94.963 -77.859 50.804
117391 94.975 -77.851 50.769
117392 94.988 -77.844 50.733
117393 95.001 -77.837 50.697
117394 95.014 -77.829 50.661
117395 95.026 -77.820 50.626
117396 95.038 -77.812 50.591
117397 95.051 -77.803 50.555
117398 95.063 -77.794 50.522
117399 95.077 -77.789 50.488
117400 95.092 -77.783 50.454
117401 95.107 -77.777 50.422
117402 95.124 -77.774 50.389
117403 95.144 -77.775 50.356
117404 95.164 -77.776 50.324
117405 95.183 -77.777 50.294
117406 95.197 -77.787 50.263
117407 95.202 -77.812 50.233
117408 95.222 -77.816 50.204
117409 95.246 -77.818 50.176
117410 95.268 -77.819 50.149
117411 95.291 -77.821 50.123
117412 95.320 -77.818 50.098
117413 95.357 -77.811 50.075
117414 95.394 -77.803 50.053
117415 95.419 -77.799 50.032
117416 95.432 -77.799 50.010
117417 95.445 -77.798 49.987
117418 95.458 -77.797 49.963
117419 95.472 -77.796 49.936
117420 95.485 -77.796 49.904
117421 95.498 -77.795 49.871
117422 95.512 -77.794 49.836
117423 95.523 -77.793 49.801
117424 95.512 -77.821 49.765
117425 95.494 -77.856 49.729
117426 95.476 -77.892 49.692
117427 95.458 -77.928 49.655
117428 95.497 -77.901 49.619
117429 95.546 -77.862 49.583
117430 95.550 -77.852 49.548
117431 95.553 -77.844 49.512
117432 95.556 -77.835 49.478
117433 95.559 -77.827 49.443
117434 95.576 -77.821 49.411
117435 95.602 -77.812 49.380
117436 95.615 -77.814 49.349
117437 95.635 -77.816 49.317
117438 95.656 -77.818 49.287
117439 95.677 -77.820 49.258
117440 95.697 -77.822 49.228
117441 95.718 -77.824 49.199
117442 95.745 -77.825 49.168
117443 95.775 -77.826 49.136
117444 95.799 -77.833 49.101
117445 95.817 -77.840 49.066
117446 95.829 -77.854 49.031
117447 95.837 -77.871 48.994
117448 95.846 -77.888 48.960
117449 95.855 -77.906 48.925
117450 95.864 -77.923 48.886
117451 95.873 -77.940 48.845
117452 95.882 -77.957 48.804
117453 95.889 -77.965 48.764
117454 95.902 -77.967 48.722
117455 95.914 -77.968 48.683
117456 95.927 -77.968 48.644
117457 95.940 -77.970 48.605
117458 95.952 -77.971 48.568
117459 95.965 -77.972 48.532
117460 95.978 -77.973 48.497
117461 95.990 -77.974 48.464
117462 96.005 -77.973 48.434
117463 96.021 -77.970 48.404
117464 96.037 -77.967 48.375
117465 96.048 -77.972 48.346
117466 96.060 -77.977 48.316
117467 96.072 -77.981 48.288
117468 96.083 -77.987 48.262
117469 96.095 -77.991 48.239
117470 96.106 -77.996 48.215
117471 96.118 -78.000 48.189
117472 96.134 -78.001 48.162
117473 96.151 -78.002 48.136
117474 96.166 -78.003 48.110
117475 96.182 -78.004 48.086
117476 96.198 -78.006 48.061
117477 96.215 -78.006 48.037
117478 96.232 -78.009 48.011
117479 96.250 -78.012 47.986
117480 96.272 -78.021 47.960
117481 96.289 -78.029 47.934
117482 96.300 -78.040 47.908
117483 96.316 -78.053 47.882
117484 96.333 -78.066 47.856
117485 96.349 -78.080 47.829
117486 96.366 -78.094 47.802
117487 96.374 -78.115 47.774
117488 96.377 -78.139 47.745
117489 96.391 -78.155 47.716
117490 96.410 -78.167 47.690
117491 96.429 -78.179 47.663
117492 96.448 -78.191 47.635
117493 96.452 -78.191 47.608
117494 96.454 -78.188 47.581
117495 96.455 -78.185 47.554
117496 96.457 -78.182 47.527
117497 96.459 -78.180 47.500
117498 96.460 -78.177 47.474
117499 96.462 -78.177 47.447
117500 96.464 -78.182 47.419
117501 96.466 -78.187 47.392
117502 96.471 -78.188 47.365
117503 96.490 -78.195 47.338
117504 96.509 -78.203 47.312
117505 96.528 -78.211 47.286
117506 96.546 -78.219 47.260
117507 96.561 -78.225 47.233
117508 96.568 -78.231 47.206
117509 96.574 -78.235 47.179
117510 96.580 -78.240 47.153
117511 96.585 -78.244 47.127
117512 96.589 -78.254 47.101
117513 96.587 -78.273 47.075
117514 96.591 -78.312 47.049
117515 96.601 -78.370 47.022
117516 96.594 -78.351 46.995
117517 96.598 -78.380 46.968
117518 96.601 -78.404 46.941
117519 96.622 -78.419 46.913
117520 96.660 -78.432 46.885
117521 96.694 -78.441 46.858
117522 96.729 -78.450 46.831
117523 96.764 -78.459 46.803
117524 96.799 -78.467 46.776
117525 96.830 -78.479 46.749
117526 96.860 -78.491 46.721
117527 96.887 -78.500 46.694
117528 96.914 -78.508 46.668
117529 96.940 -78.521 46.642
117530 96.966 -78.536 46.617
117531 96.991 -78.551 46.591
117532 97.016 -78.567 46.564
117533 97.041 -78.582 46.538
117534 97.049 -78.598 46.511
117535 97.044 -78.613 46.484
117536 97.038 -78.627 46.459
117537 97.033 -78.642 46.436
117538 97.028 -78.657 46.413
117539 97.034 -78.677 46.392
117540 97.081 -78.671 46.371
117541 97.141 -78.657 46.353
117542 97.201 -78.642 46.336
117543 97.260 -78.628 46.319
117544 97.320 -78.614 46.303
117545 97.379 -78.599 46.288
117546 97.439 -78.585 46.273
117547 97.499 -78.570 46.259
117548 97.558 -78.555 46.245
117549 97.614 -78.548 46.232
117550 97.662 -78.562 46.219
117551 97.721 -78.593 46.205
117552 97.781 -78.625 46.192
117553 97.818 -78.664 46.179
117554 97.817 -78.714 46.166
117555 97.820 -78.728 46.152
117556 97.832 -78.731 46.138
117557 97.853 -78.729 46.123
117558 97.874 -78.727 46.107
117559 97.896 -78.725 46.092
117560 97.928 -78.683 46.076
117561 97.954 -78.636 46.059
117562 97.974 -78.642 46.042
117563 98.002 -78.654 46.026
117564 98.029 -78.665 46.009
117565 98.058 -78.677 45.991
117566 97.992 -78.620 45.974
117567 98.023 -78.611 45.958
117568 98.051 -78.613 45.940
117569 98.078 -78.614 45.923
117570 98.118 -78.605 45.906
117571 98.187 -78.568 45.888
117572 98.190 -78.569 45.869
117573 98.195 -78.566 45.853
117574 98.200 -78.563 45.836
117575 98.204 -78.560 45.818
117576 98.215 -78.565 45.802
117577 98.239 -78.580 45.786
117578 98.244 -78.582 45.769
117579 98.251 -78.586 45.752
117580 98.259 -78.589 45.736
117581 98.267 -78.592 45.720
117582 98.274 -78.595 45.704
117583 98.282 -78.599 45.688
117584 98.289 -78.603 45.672
117585 98.296 -78.606 45.655
117586 98.320 -78.619 45.641
117587 98.334 -78.617 45.625
117588 98.343 -78.625 45.609
117589 98.351 -78.632 45.594
117590 98.360 -78.639 45.577
117591 98.367 -78.645 45.561
117592 98.372 -78.647 45.544
117593 98.376 -78.649 45.528
117594 98.380 -78.651 45.512
117595 98.385 -78.649 45.496
117596 98.413 -78.667 45.479
117597 98.440 -78.686 45.462
117598 98.465 -78.710 45.446
117599 98.498 -78.736 45.430
117600 98.529 -78.762 45.415
117601 98.558 -78.791 45.401
117602 98.563 -78.776 45.387
117603 98.548 -78.714 45.372
117604 98.567 -78.688 45.358
117605 98.558 -78.745 45.343
117606 98.590 -78.774 45.327
117607 98.622 -78.802 45.311
117608 98.654 -78.832 45.297
117609 98.687 -78.861 45.283
117610 98.719 -78.890 45.268
117611 98.756 -78.909 45.254
117612 98.800 -78.916 45.240
117613 98.844 -78.924 45.226
117614 98.881 -78.927 45.210
117615 98.853 -78.893 45.198
117616 98.889 -78.890 45.186
117617 98.935 -78.893 45.175
117618 98.968 -78.906 45.163
117619 98.976 -78.859 45.151
117620 98.984 -78.812 45.138
117621 98.992 -78.765 45.126
117622 99.005 -78.764 45.114
117623 99.020 -78.777 45.104
117624 99.035 -78.788 45.096
117625 99.050 -78.800 45.087
117626 99.065 -78.812 45.078
117627 99.080 -78.823 45.069
117628 99.099 -78.828 45.061
117629 99.118 -78.833 45.052
117630 99.136 -78.839 45.042
117631 99.153 -78.843 45.033
117632 99.170 -78.848 45.024
117633 99.205 -78.866 45.013
117634 99.233 -78.878 45.002
117635 99.263 -78.890 44.990
117636 99.291 -78.902 44.978
117637 99.319 -78.918 44.966
117638 99.347 -78.932 44.954
117639 99.375 -78.946 44.943
117640 99.403 -78.961 44.935
117641 99.417 -78.955 44.927
117642 99.384 -78.884 44.921
117643 99.455 -78.912 44.913
117644 99.517 -78.915 44.904
117645 99.532 -78.955 44.895
117646 99.559 -78.977 44.887
117647 99.592 -78.987 44.880
117648 99.626 -78.998 44.872
117649 99.660 -79.008 44.862
117650 99.689 -79.028 44.852
117651 99.715 -79.071 44.843
117652 99.736 -79.107 44.834
117653 99.758 -79.142 44.826
117654 99.776 -79.168 44.820
117655 99.789 -79.173 44.814
117656 99.802 -79.176 44.808
117657 99.825 -79.197 44.801
117658 99.848 -79.219 44.793
117659 99.858 -79.243 44.786
117660 99.835 -79.272 44.780
117661 99.814 -79.324 44.773
117662 99.822 -79.353 44.767
117663 99.840 -79.373 44.762
117664 99.874 -79.391 44.756
117665 99.928 -79.408 44.749
117666 99.981 -79.425 44.743
117667 100.035 -79.443 44.738
117668 100.089 -79.463 44.732
117669 100.139 -79.484 44.726
117670 100.155 -79.504 44.719
117671 100.171 -79.524 44.711
117672 100.186 -79.544 44.703
117673 100.202 -79.564 44.695
117674 100.217 -79.584 44.687
117675 100.252 -79.587 44.679
117676 100.290 -79.587 44.671
117677 100.329 -79.587 44.663
117678 100.367 -79.587 44.655
117679 100.405 -79.587 44.647
117680 100.443 -79.586 44.639
117681 100.457 -79.589 44.629
117682 100.485 -79.603 44.621
117683 100.513 -79.617 44.612
117684 100.542 -79.630 44.603
117685 100.599 -79.640 44.593
117686 100.656 -79.649 44.584
117687 100.714 -79.659 44.574
117688 100.766 -79.665 44.565
117689 100.802 -79.677 44.554
117690 100.835 -79.712 44.542
117691 100.872 -79.741 44.531
117692 100.912 -79.764 44.521
117693 100.953 -79.788 44.510
117694 100.993 -79.811 44.500
117695 101.034 -79.835 44.489
117696 101.073 -79.863 44.480
117697 101.112 -79.892 44.470
117698 101.151 -79.920 44.459
117699 101.191 -79.949 44.449
117700 101.230 -79.977 44.437
117701 101.269 -80.005 44.425
117702 101.308 -80.034 44.412
117703 101.347 -80.062 44.399
117704 101.386 -80.090 44.383
117705 101.425 -80.098 44.366
117706 101.492 -80.054 44.350
117707 101.531 -80.057 44.333
117708 101.563 -80.074 44.314
117709 101.595 -80.090 44.295
117710 101.627 -80.106 44.277
117711 101.659 -80.121 44.259
117712 101.692 -80.138 44.240
117713 101.724 -80.154 44.222
117714 101.756 -80.170 44.204
117715 101.788 -80.187 44.186
117716 101.820 -80.203 44.167
117717 101.852 -80.218 44.147
118048 94.420 -77.903 52.034
118049 94.433 -77.896 52.012
118050 94.447 -77.888 51.991
118051 94.461 -77.881 51.970
118052 94.475 -77.874 51.948
118053 94.489 -77.868 51.927
118054 94.504 -77.861 51.906
118055 94.518 -77.855 51.885
118056 94.532 -77.848 51.863
118057 94.546 -77.841 51.843
118058 94.560 -77.835 51.822
118059 94.566 -77.830 51.801
118060 94.567 -77.825 51.779
118061 94.568 -77.821 51.757
118062 94.569 -77.817 51.734
118063 94.570 -77.812 51.711
118064 94.578 -77.806 51.687
118065 94.599 -77.796 51.662
118066 94.614 -77.797 51.637
118067 94.629 -77.797 51.610
118068 94.645 -77.798 51.582
118069 94.662 -77.802 51.554
118070 94.673 -77.802 51.526
118071 94.685 -77.802 51.497
118072 94.697 -77.804 51.467
118073 94.709 -77.808 51.438
118074 94.722 -77.811 51.409
118075 94.737 -77.815 51.378
118076 94.764 -77.815 51.346
118077 94.795 -77.811 51.315
118078 94.826 -77.807 51.282
118079 94.813 -77.838 51.247
118080 94.822 -77.840 51.212
118081 94.859 -77.799 51.175
118082 94.877 -77.811 51.139
118083 94.883 -77.832 51.102
118084 94.889 -77.853 51.066
118085 94.895 -77.875 51.031
118086 94.902 -77.893 50.995
118087 94.915 -77.884 50.957
118088 94.927 -77.875 50.920
118089 94.940 -77.867 50.883
118090 94.952 -77.859 50.847
118091 94.965 -77.849 50.810
118092 94.977 -77.841 50.775
118093 94.990 -77.832 50.739
118094 95.002 -77.824 50.703
118095 95.015 -77.815 50.667
118096 95.028 -77.806 50.631
118097 95.040 -77.798 50.596
118098 95.055 -77.796 50.561
118099 95.071 -77.793 50.527
118100 95.086 -77.787 50.493
118101 95.098 -77.774 50.458
118102 95.114 -77.771 50.425
118103 95.131 -77.767 50.392
118104 95.147 -77.764 50.359
118105 95.166 -77.764 50.327
118106 95.186 -77.765 50.297
118107 95.205 -77.765 50.267
118108 95.213 -77.785 50.236
118109 95.229 -77.796 50.208
118110 95.252 -77.798 50.179
118111 95.274 -77.799 50.152
118112 95.297 -77.801 50.125
118113 95.320 -77.802 50.100
118114 95.344 -77.803 50.075
118115 95.381 -77.796 50.051
118116 95.418 -77.789 50.029
118117 95.431 -77.787 50.008
118118 95.445 -77.787 49.986
118119 95.458 -77.787 49.962
118120 95.471 -77.786 49.935
118121 95.484 -77.785 49.904
118122 95.497 -77.784 49.871
118123 95.498 -77.778 49.835
118124 95.496 -77.770 49.802
118125 95.594 -77.823 49.767
118126 95.504 -77.849 49.730
118127 95.466 -77.882 49.693
118128 95.448 -77.918 49.658
118129 95.483 -77.895 49.622
118130 95.533 -77.855 49.586
118131 95.550 -77.838 49.551
118132 95.553 -77.830 49.516
118133 95.556 -77.821 49.481
118134 95.580 -77.818 49.447
118135 95.604 -77.815 49.415
118136 95.629 -77.796 49.384
118137 95.639 -77.805 49.353
118138 95.660 -77.806 49.324
118139 95.680 -77.808 49.298
118140 95.701 -77.810 49.273
118141 95.722 -77.812 49.245
118142 95.742 -77.814 49.217
118143 95.763 -77.816 49.186
118144 95.783 -77.818 49.151
118145 95.804 -77.820 49.114
118146 95.818 -77.825 49.077
118147 95.827 -77.843 49.041
118148 95.836 -77.859 49.005
118149 95.845 -77.877 48.970
118150 95.854 -77.894 48.935
118151 95.863 -77.911 48.899
118152 95.872 -77.928 48.858
118153 95.881 -77.946 48.819
118154 95.894 -77.955 48.779
118155 95.905 -77.955 48.738
118156 95.917 -77.956 48.699
118157 95.931 -77.956 48.661
118158 95.943 -77.958 48.622
118159 95.956 -77.959 48.584
118160 95.968 -77.959 48.547
118161 95.981 -77.961 48.512
118162 95.994 -77.962 48.479
118163 96.008 -77.962 48.450
118164 96.023 -77.960 48.421
118165 96.034 -77.965 48.392
118166 96.046 -77.969 48.361
118167 96.057 -77.974 48.330
118168 96.071 -77.978 48.300
118169 96.084 -77.981 48.272
118170 96.098 -77.984 48.249
118171 96.111 -77.987 48.225
118172 96.121 -77.985 48.201
118173 96.134 -77.984 48.175
118174 96.151 -77.984 48.149
118175 96.167 -77.986 48.123
118176 96.183 -77.987 48.098
118177 96.199 -77.988 48.073
118178 96.216 -77.989 48.048
118179 96.232 -77.990 48.022
118180 96.251 -77.996 47.996
118181 96.267 -78.003 47.971
118182 96.283 -78.016 47.945
118183 96.299 -78.030 47.918
118184 96.315 -78.044 47.893
118185 96.332 -78.057 47.867
118186 96.348 -78.071 47.841
118187 96.365 -78.084 47.814
118188 96.375 -78.103 47.786
118189 96.379 -78.128 47.758
118190 96.383 -78.152 47.730
118191 96.399 -78.166 47.704
118192 96.418 -78.178 47.677
118193 96.437 -78.191 47.650
118194 96.446 -78.194 47.623
118195 96.448 -78.191 47.595
118196 96.450 -78.188 47.568
118197 96.451 -78.186 47.542
118198 96.453 -78.183 47.515
118199 96.454 -78.181 47.487
118200 96.456 -78.178 47.460
118201 96.458 -78.175 47.432
118202 96.459 -78.172 47.405
118203 96.473 -78.178 47.378
118204 96.492 -78.186 47.350
118205 96.511 -78.194 47.323
118206 96.529 -78.201 47.297
118207 96.548 -78.209 47.271
118208 96.567 -78.217 47.245
118209 96.573 -78.222 47.219
118210 96.579 -78.226 47.193
118211 96.585 -78.231 47.167
118212 96.590 -78.238 47.141
118213 96.588 -78.257 47.115
118214 96.586 -78.275 47.089
118215 96.584 -78.294 47.062
118216 96.589 -78.329 47.035
118217 96.581 -78.327 47.008
118218 96.584 -78.344 46.981
118219 96.599 -78.373 46.955
118220 96.621 -78.397 46.928
118221 96.658 -78.410 46.900
118222 96.693 -78.420 46.873
118223 96.728 -78.429 46.846
118224 96.763 -78.438 46.820
118225 96.798 -78.447 46.792
118226 96.832 -78.455 46.765
118227 96.863 -78.466 46.737
118228 96.889 -78.479 46.710
118229 96.914 -78.495 46.682
118230 96.939 -78.510 46.656
118231 96.964 -78.525 46.631
118232 96.990 -78.541 46.605
118233 97.015 -78.556 46.580
118234 97.040 -78.572 46.553
118235 97.047 -78.587 46.527
118236 97.042 -78.601 46.501
118237 97.037 -78.617 46.475
118238 97.031 -78.631 46.451
118239 97.026 -78.645 46.429
118240 97.035 -78.651 46.408
118241 97.118 -78.640 46.387
118242 97.169 -78.624 46.368
118243 97.219 -78.608 46.351
118244 97.269 -78.592 46.335
118245 97.319 -78.576 46.319
118246 97.370 -78.560 46.303
118247 97.429 -78.545 46.289
118248 97.489 -78.530 46.274
118249 97.548 -78.516 46.260
118250 97.603 -78.512 46.246
118251 97.663 -78.544 46.233
118252 97.722 -78.576 46.218
118253 97.782 -78.608 46.205
118254 97.820 -78.645 46.193
118255 97.818 -78.695 46.179
118256 97.818 -78.735 46.165
118257 97.836 -78.735 46.151
118258 97.864 -78.705 46.136
118259 97.899 -78.649 46.121
118260 97.924 -78.604 46.106
118261 97.938 -78.607 46.090
118262 97.966 -78.618 46.073
118263 97.994 -78.630 46.056
118264 98.021 -78.641 46.040
118265 98.049 -78.653 46.023
118266 98.013 -78.634 46.005
118267 98.018 -78.609 45.987
118268 98.031 -78.588 45.970
118269 98.069 -78.587 45.953
118270 98.097 -78.589 45.936
118271 98.144 -78.573 45.918
118272 98.177 -78.558 45.900
118273 98.180 -78.558 45.882
118274 98.184 -78.555 45.865
118275 98.189 -78.552 45.848
118276 98.195 -78.548 45.831
118277 98.202 -78.570 45.815
118278 98.227 -78.597 45.798
118279 98.249 -78.600 45.781
118280 98.272 -78.603 45.765
118281 98.293 -78.605 45.748
118282 98.315 -78.608 45.731
118283 98.322 -78.608 45.715
118284 98.329 -78.608 45.698
118285 98.336 -78.608 45.682
118286 98.343 -78.608 45.665
118287 98.350 -78.608 45.650
118288 98.357 -78.612 45.634
118289 98.362 -78.615 45.618
118290 98.366 -78.617 45.603
118291 98.370 -78.620 45.586
118292 98.375 -78.622 45.570
118293 98.379 -78.624 45.553
118294 98.383 -78.627 45.537
118295 98.387 -78.630 45.521
118296 98.390 -78.625 45.505
118297 98.403 -78.626 45.488
118298 98.429 -78.648 45.471
118299 98.439 -78.705 45.454
118300 98.470 -78.734 45.439
118301 98.505 -78.758 45.424
118302 98.546 -78.762 45.408
118303 98.564 -78.734 45.392
118304 98.557 -78.677 45.376
118305 98.567 -78.655 45.362
118306 98.599 -78.683 45.347
118307 98.632 -78.713 45.331
118308 98.664 -78.742 45.315
118309 98.696 -78.771 45.301
118310 98.728 -78.799 45.286
118311 98.761 -78.828 45.271
118312 98.793 -78.857 45.257
118313 98.835 -78.869 45.243
118314 98.878 -78.876 45.228
118315 98.878 -78.858 45.214
118316 98.849 -78.823 45.201
118317 98.893 -78.824 45.188
118318 98.939 -78.827 45.176
118319 98.985 -78.830 45.164
118320 99.010 -78.844 45.151
118321 99.018 -78.796 45.139
118322 99.002 -78.777 45.127
118323 99.003 -78.775 45.116
118324 99.017 -78.783 45.104
118325 99.031 -78.791 45.094
118326 99.046 -78.799 45.086
118327 99.061 -78.806 45.077
118328 99.075 -78.802 45.068
118329 99.089 -78.805 45.059
118330 99.106 -78.810 45.049
118331 99.123 -78.814 45.040
118332 99.140 -78.819 45.031
118333 99.157 -78.824 45.022
118334 99.206 -78.855 45.011
118335 99.234 -78.867 45.001
118336 99.264 -78.879 44.989
118337 99.293 -78.890 44.978
118338 99.322 -78.902 44.967
118339 99.351 -78.915 44.955
118340 99.379 -78.927 44.944
118341 99.409 -78.938 44.933
118342 99.410 -78.913 44.924
118343 99.394 -78.853 44.915
118344 99.486 -78.852 44.907
118345 99.534 -78.851 44.898
118346 99.556 -78.885 44.889
118347 99.574 -78.924 44.881
118348 99.607 -78.936 44.873
118349 99.633 -78.962 44.863
118350 99.656 -78.996 44.853
118351 99.679 -79.030 44.843
118352 99.713 -79.091 44.835
118353 99.728 -79.109 44.827
118354 99.739 -79.106 44.820
118355 99.762 -79.128 44.813
118356 99.785 -79.149 44.807
118357 99.808 -79.171 44.801
118358 99.832 -79.193 44.794
118359 99.855 -79.214 44.787
118360 99.869 -79.237 44.781
118361 99.847 -79.266 44.775
118362 99.825 -79.295 44.769
118363 99.803 -79.333 44.763
118364 99.834 -79.352 44.758
118365 99.888 -79.370 44.751
118366 99.941 -79.391 44.746
118367 99.995 -79.411 44.741
118368 100.049 -79.432 44.735
118369 100.103 -79.452 44.730
118370 100.146 -79.473 44.724
118371 100.162 -79.493 44.717
118372 100.177 -79.513 44.708
118373 100.193 -79.533 44.701
118374 100.211 -79.551 44.693
118375 100.250 -79.551 44.685
118376 100.288 -79.550 44.677
118377 100.326 -79.550 44.669
118378 100.365 -79.550 44.662
118379 100.403 -79.550 44.653
118380 100.441 -79.550 44.645
118381 100.474 -79.552 44.637
118382 100.503 -79.558 44.628
118383 100.541 -79.569 44.619
118384 100.580 -79.579 44.611
118385 100.618 -79.589 44.602
118386 100.656 -79.600 44.593
118387 100.690 -79.612 44.584
118388 100.718 -79.626 44.576
118389 100.766 -79.636 44.567
118390 100.808 -79.654 44.557
118391 100.848 -79.678 44.546
118392 100.888 -79.705 44.534
118393 100.927 -79.733 44.524
118394 100.966 -79.762 44.514
118395 101.005 -79.790 44.505
118396 101.045 -79.819 44.494
118397 101.084 -79.847 44.485
118398 101.123 -79.874 44.474
118399 101.162 -79.899 44.464
118400 101.201 -79.924 44.453
118401 101.240 -79.949 44.442
118402 101.279 -79.974 44.430
118403 101.318 -79.999 44.417
118404 101.357 -80.024 44.404
118405 101.396 -80.049 44.388
118406 101.435 -80.068 44.372
118407 101.504 -80.027 44.357
118408 101.548 -80.019 44.339
118409 101.581 -80.035 44.320
118410 101.613 -80.051 44.300
118411 101.644 -80.068 44.281
118412 101.677 -80.084 44.263
118413 101.709 -80.099 44.244
118414 101.741 -80.116 44.225
118415 101.773 -80.132 44.207
118416 101.805 -80.148 44.189
118417 101.837 -80.165 44.170
118418 101.870 -80.181 44.151
118748 94.415 -77.892 52.067
118749 94.429 -77.884 52.045
118750 94.442 -77.877 52.023
118751 94.456 -77.869 52.002
118752 94.470 -77.862 51.981
118753 94.484 -77.855 51.959
118754 94.498 -77.849 51.939
118755 94.513 -77.842 51.918
118756 94.527 -77.836 51.898
118757 94.541 -77.829 51.876
118758 94.555 -77.822 51.856
118759 94.569 -77.816 51.836
118760 94.578 -77.810 51.816
118761 94.579 -77.805 51.795
118762 94.580 -77.802 51.774
118763 94.581 -77.797 51.752
118764 94.582 -77.793 51.729
118765 94.583 -77.789 51.706
118766 94.601 -77.780 51.682
118767 94.618 -77.778 51.656
118768 94.633 -77.778 51.628
118769 94.648 -77.779 51.600
118770 94.662 -77.782 51.571
118771 94.675 -77.785 51.542
118772 94.687 -77.788 51.513
118773 94.700 -77.792 51.484
118774 94.713 -77.795 51.454
118775 94.725 -77.798 51.424
118776 94.741 -77.805 51.393
118777 94.764 -77.808 51.361
118778 94.795 -77.804 51.330
118779 94.810 -77.812 51.297
118780 94.797 -77.843 51.263
118781 94.780 -77.909 51.227
118782 94.848 -77.824 51.190
118783 94.868 -77.866 51.152
118784 94.875 -77.888 51.115
118785 94.882 -77.902 51.078
118786 94.893 -77.891 51.043
118787 94.904 -77.880 51.006
118788 94.916 -77.871 50.968
118789 94.929 -77.862 50.930
118790 94.942 -77.853 50.893
118791 94.954 -77.844 50.856
118792 94.967 -77.836 50.819
118793 94.979 -77.827 50.783
118794 94.992 -77.818 50.747
118795 95.004 -77.810 50.710
118796 95.018 -77.803 50.674
118797 95.033 -77.802 50.638
118798 95.049 -77.800 50.602
118799 95.065 -77.798 50.567
118800 95.079 -77.792 50.533
118801 95.089 -77.771 50.498
118802 95.105 -77.768 50.464
118803 95.121 -77.764 50.431
118804 95.138 -77.761 50.397
118805 95.154 -77.757 50.363
118806 95.171 -77.754 50.332
118807 95.188 -77.752 50.302
118808 95.208 -77.753 50.272
118809 95.225 -77.758 50.241
118810 95.235 -77.777 50.212
118811 95.258 -77.779 50.184
118812 95.280 -77.780 50.156
118813 95.303 -77.782 50.128
118814 95.326 -77.783 50.103
118815 95.348 -77.784 50.077
118816 95.372 -77.786 50.052
118817 95.405 -77.781 50.027
118818 95.430 -77.777 50.004
118819 95.444 -77.776 49.981
118820 95.457 -77.775 49.958
118821 95.470 -77.774 49.930
118822 95.475 -77.771 49.899
118823 95.473 -77.763 49.867
118824 95.471 -77.755 49.834
118825 95.478 -77.753 49.799
118826 95.588 -77.811 49.763
118827 95.615 -77.855 49.726
118828 95.526 -77.882 49.691
118829 95.438 -77.909 49.656
118830 95.470 -77.889 49.621
118831 95.520 -77.849 49.586
118832 95.549 -77.824 49.554
118833 95.560 -77.817 49.520
118834 95.584 -77.814 49.485
118835 95.608 -77.811 49.451
118836 95.631 -77.808 49.419
118837 95.643 -77.792 49.388
118838 95.663 -77.795 49.358
118839 95.684 -77.796 49.332
118840 95.704 -77.799 49.310
118841 95.725 -77.800 49.288
118842 95.745 -77.802 49.262
118843 95.766 -77.805 49.234
118844 95.787 -77.806 49.202
118845 95.800 -77.805 49.165
118846 95.804 -77.805 49.130
118847 95.814 -77.812 49.092
118848 95.829 -77.823 49.054
118849 95.838 -77.840 49.018
118850 95.847 -77.857 48.981
118851 95.856 -77.874 48.944
118852 95.865 -77.892 48.907
118853 95.874 -77.909 48.870
118854 95.883 -77.926 48.832
118855 95.899 -77.941 48.793
118856 95.919 -77.952 48.754
118857 95.931 -77.952 48.716
118858 95.942 -77.952 48.677
118859 95.954 -77.952 48.638
118860 95.965 -77.953 48.600
118861 95.977 -77.953 48.562
118862 95.988 -77.953 48.527
118863 96.000 -77.953 48.495
118864 96.011 -77.952 48.469
118865 96.024 -77.954 48.440
118866 96.037 -77.957 48.409
118867 96.050 -77.960 48.376
118868 96.064 -77.963 48.344
118869 96.077 -77.966 48.313
118870 96.091 -77.969 48.283
118871 96.104 -77.973 48.258
118872 96.117 -77.976 48.233
118873 96.125 -77.971 48.208
118874 96.134 -77.966 48.184
118875 96.151 -77.968 48.160
118876 96.167 -77.968 48.134
118877 96.183 -77.969 48.107
118878 96.199 -77.971 48.083
118879 96.216 -77.971 48.058
118880 96.232 -77.972 48.032
118881 96.249 -77.980 48.006
118882 96.265 -77.993 47.980
118883 96.282 -78.007 47.954
118884 96.298 -78.021 47.928
118885 96.315 -78.034 47.902
118886 96.331 -78.048 47.877
118887 96.347 -78.061 47.852
118888 96.364 -78.075 47.825
118889 96.377 -78.091 47.799
118890 96.381 -78.116 47.772
118891 96.384 -78.141 47.745
118892 96.388 -78.165 47.718
118893 96.407 -78.177 47.692
118894 96.426 -78.189 47.665
118895 96.440 -78.197 47.638
118896 96.442 -78.194 47.609
118897 96.444 -78.192 47.582
118898 96.445 -78.189 47.555
118899 96.447 -78.187 47.529
118900 96.448 -78.184 47.502
118901 96.450 -78.181 47.474
118902 96.452 -78.178 47.445
118903 96.463 -78.170 47.419
118904 96.477 -78.167 47.391
118905 96.494 -78.176 47.363
118906 96.512 -78.184 47.337
118907 96.531 -78.191 47.311
118908 96.549 -78.200 47.285
118909 96.568 -78.207 47.259
118910 96.578 -78.213 47.233
118911 96.584 -78.217 47.208
118912 96.590 -78.222 47.183
118913 96.589 -78.240 47.158
118914 96.587 -78.259 47.133
118915 96.585 -78.278 47.106
118916 96.583 -78.297 47.079
118917 96.586 -78.303 47.051
118918 96.588 -78.312 47.025
118919 96.599 -78.322 46.999
118920 96.626 -78.335 46.973
118921 96.644 -78.362 46.946
118922 96.670 -78.381 46.920
118923 96.706 -78.389 46.893
118924 96.742 -78.397 46.867
118925 96.779 -78.405 46.840
118926 96.815 -78.413 46.813
118927 96.839 -78.431 46.785
118928 96.862 -78.451 46.757
118929 96.888 -78.467 46.728
118930 96.913 -78.483 46.699
118931 96.938 -78.498 46.672
118932 96.963 -78.514 46.646
118933 96.989 -78.529 46.620
118934 97.014 -78.545 46.595
118935 97.039 -78.561 46.570
118936 97.045 -78.576 46.544
118937 97.040 -78.590 46.517
118938 97.035 -78.605 46.492
118939 97.030 -78.620 46.468
118940 97.041 -78.621 46.445
118941 97.070 -78.625 46.424
118942 97.151 -78.566 46.404
118943 97.257 -78.601 46.385
118944 97.307 -78.586 46.367
118945 97.358 -78.569 46.350
118946 97.407 -78.553 46.334
118947 97.458 -78.537 46.318
118948 97.508 -78.521 46.303
118949 97.534 -78.553 46.288
118950 97.538 -78.605 46.274
118951 97.625 -78.561 46.260
118952 97.672 -78.551 46.247
118953 97.726 -78.566 46.233
118954 97.783 -78.589 46.219
118955 97.821 -78.627 46.206
118956 97.837 -78.589 46.191
118957 97.855 -78.579 46.177
118958 97.871 -78.615 46.163
118959 97.893 -78.573 46.148
118960 97.902 -78.571 46.134
118961 97.930 -78.582 46.119
118962 97.958 -78.595 46.104
118963 97.986 -78.606 46.086
118964 98.013 -78.617 46.069
118965 98.041 -78.629 46.052
118966 98.069 -78.640 46.035
118967 98.039 -78.623 46.018
118968 98.045 -78.598 46.000
118969 98.050 -78.573 45.982
118970 98.083 -78.562 45.965
118971 98.116 -78.562 45.948
118972 98.165 -78.543 45.931
118973 98.166 -78.548 45.913
118974 98.170 -78.548 45.894
118975 98.174 -78.545 45.877
118976 98.180 -78.540 45.861
118977 98.187 -78.534 45.845
118978 98.195 -78.549 45.828
118979 98.204 -78.615 45.811
118980 98.229 -78.623 45.793
118981 98.251 -78.625 45.777
118982 98.273 -78.628 45.760
118983 98.295 -78.630 45.743
118984 98.316 -78.633 45.726
118985 98.339 -78.636 45.710
118986 98.360 -78.638 45.693
118987 98.382 -78.640 45.675
118988 98.401 -78.642 45.659
118989 98.398 -78.636 45.644
118990 98.397 -78.630 45.628
118991 98.395 -78.625 45.611
118992 98.393 -78.619 45.595
118993 98.392 -78.613 45.579
118994 98.390 -78.608 45.563
118995 98.390 -78.604 45.545
118996 98.395 -78.607 45.529
118997 98.397 -78.603 45.513
118998 98.400 -78.598 45.496
118999 98.408 -78.645 45.480
119000 98.417 -78.702 45.464
119001 98.450 -78.711 45.447
119002 98.492 -78.712 45.433
119003 98.534 -78.714 45.417
119004 98.552 -78.686 45.401
119005 98.552 -78.639 45.385
119006 98.575 -78.637 45.367
119007 98.597 -78.655 45.352
119008 98.640 -78.670 45.337
119009 98.687 -78.691 45.322
119010 98.733 -78.712 45.306
119011 98.770 -78.738 45.292
119012 98.803 -78.767 45.277
119013 98.835 -78.796 45.263
119014 98.869 -78.821 45.248
119015 98.903 -78.822 45.234
119016 98.874 -78.788 45.219
119017 98.864 -78.765 45.204
119018 98.903 -78.756 45.191
119019 98.941 -78.748 45.178
119020 98.980 -78.739 45.165
119021 99.016 -78.733 45.153
119022 99.025 -78.811 45.141
119023 99.006 -78.795 45.128
119024 99.000 -78.788 45.117
119025 99.015 -78.796 45.105
119026 99.029 -78.803 45.093
119027 99.044 -78.811 45.085
119028 99.056 -78.793 45.076
119029 99.070 -78.783 45.066
119030 99.079 -78.783 45.056
119031 99.093 -78.786 45.047
119032 99.110 -78.790 45.038
119033 99.127 -78.795 45.029
119034 99.144 -78.799 45.019
119035 99.202 -78.839 45.009
119036 99.236 -78.855 44.999
119037 99.265 -78.868 44.989
119038 99.294 -78.879 44.979
119039 99.323 -78.891 44.969
119040 99.352 -78.903 44.958
119041 99.380 -78.915 44.947
119042 99.420 -78.906 44.933
119043 99.460 -78.885 44.921
119044 99.449 -78.864 44.910
119045 99.528 -78.950 44.901
119046 99.501 -78.850 44.892
119047 99.531 -78.874 44.884
119048 99.576 -78.880 44.875
119049 99.599 -78.915 44.865
119050 99.622 -78.949 44.855
119051 99.645 -78.984 44.845
119052 99.674 -79.024 44.837
119053 99.700 -79.057 44.829
119054 99.722 -79.079 44.821
119055 99.746 -79.101 44.813
119056 99.769 -79.122 44.807
119057 99.792 -79.144 44.800
119058 99.816 -79.165 44.794
119059 99.839 -79.187 44.788
119060 99.862 -79.209 44.782
119061 99.881 -79.231 44.776
119062 99.859 -79.260 44.770
119063 99.836 -79.289 44.764
119064 99.814 -79.319 44.759
119065 99.856 -79.338 44.754
119066 99.901 -79.359 44.748
119067 99.955 -79.379 44.743
119068 100.009 -79.400 44.738
119069 100.063 -79.421 44.732
119070 100.117 -79.441 44.727
119071 100.153 -79.462 44.721
119072 100.169 -79.482 44.715
119073 100.185 -79.501 44.706
119074 100.209 -79.513 44.699
119075 100.248 -79.513 44.691
119076 100.286 -79.513 44.684
119077 100.324 -79.513 44.676
119078 100.363 -79.513 44.668
119079 100.395 -79.516 44.660
119080 100.422 -79.521 44.652
119081 100.450 -79.527 44.644
119082 100.477 -79.532 44.635
119083 100.509 -79.540 44.627
119084 100.548 -79.551 44.618
119085 100.586 -79.561 44.610
119086 100.625 -79.571 44.601
119087 100.664 -79.582 44.592
119088 100.702 -79.592 44.584
119089 100.741 -79.602 44.576
119090 100.779 -79.613 44.567
119091 100.818 -79.629 44.558
119092 100.857 -79.645 44.547
119093 100.897 -79.662 44.537
119094 100.936 -79.679 44.527
119095 100.975 -79.695 44.518
119096 101.014 -79.712 44.508
119097 101.053 -79.732 44.499
119098 101.093 -79.756 44.488
119099 101.131 -79.781 44.478
119100 101.171 -79.806 44.468
119101 101.209 -79.831 44.457
119102 101.249 -79.856 44.446
119103 101.288 -79.880 44.433
119104 101.327 -79.905 44.420
119105 101.366 -79.930 44.407
119106 101.405 -79.955 44.392
119107 101.444 -79.979 44.376
119108 101.516 -79.999 44.361
119109 101.566 -79.980 44.343
119110 101.598 -79.997 44.323
119111 101.630 -80.013 44.304
119112 101.662 -80.029 44.285
119113 101.694 -80.046 44.266
119114 101.726 -80.062 44.247
119115 101.758 -80.077 44.228
119116 101.785 -80.095 44.209
119117 101.812 -80.112 44.191
119118 101.839 -80.131 44.173
119448 94.413 -77.877 52.102
119449 94.424 -77.872 52.081
119450 94.438 -77.865 52.059
119451 94.452 -77.857 52.038
119452 94.465 -77.849 52.017
119453 94.479 -77.843 51.995
119454 94.493 -77.837 51.974
119455 94.508 -77.830 51.953
119456 94.522 -77.823 51.933
119457 94.536 -77.817 51.912
119458 94.550 -77.810 51.892
119459 94.564 -77.803 51.872
119460 94.578 -77.797 51.852
119461 94.591 -77.790 51.833
119462 94.592 -77.787 51.813
119463 94.593 -77.782 51.793
119464 94.594 -77.777 51.773
119465 94.595 -77.774 51.751
119466 94.596 -77.769 51.727
119467 94.604 -77.763 51.702
119468 94.621 -77.759 51.674
119469 94.636 -77.759 51.646
119470 94.651 -77.762 51.617
119471 94.665 -77.768 51.589
119472 94.678 -77.773 51.559
119473 94.690 -77.776 51.530
119474 94.703 -77.779 51.501
119475 94.716 -77.782 51.471
119476 94.729 -77.785 51.440
119477 94.746 -77.794 51.408
119478 94.765 -77.800 51.377
119479 94.795 -77.796 51.345
119480 94.795 -77.818 51.312
119481 94.780 -77.872 51.278
119482 94.805 -77.855 51.243
119483 94.845 -77.845 51.205
119484 94.911 -77.874 51.167
119485 94.873 -77.900 51.128
119486 94.884 -77.889 51.092
119487 94.895 -77.878 51.056
119488 94.906 -77.867 51.019
119489 94.918 -77.856 50.979
119490 94.930 -77.848 50.940
119491 94.943 -77.840 50.902
119492 94.956 -77.830 50.865
119493 94.968 -77.822 50.828
119494 94.981 -77.813 50.792
119495 94.995 -77.809 50.755
119496 95.012 -77.808 50.718
119497 95.027 -77.806 50.682
119498 95.043 -77.805 50.645
119499 95.058 -77.803 50.609
119500 95.073 -77.796 50.574
119501 95.083 -77.776 50.539
119502 95.096 -77.764 50.504
119503 95.112 -77.761 50.469
119504 95.128 -77.758 50.436
119505 95.145 -77.754 50.402
119506 95.161 -77.751 50.369
119507 95.178 -77.748 50.337
119508 95.194 -77.744 50.308
119509 95.210 -77.741 50.277
119510 95.230 -77.742 50.247
119511 95.242 -77.756 50.218
119512 95.263 -77.759 50.189
119513 95.286 -77.761 50.161
119514 95.309 -77.762 50.134
119515 95.332 -77.764 50.107
119516 95.354 -77.765 50.081
119517 95.378 -77.767 50.055
119518 95.400 -77.768 50.029
119519 95.428 -77.766 50.004
119520 95.444 -77.765 49.980
119521 95.452 -77.764 49.956
119522 95.450 -77.755 49.928
119523 95.449 -77.749 49.898
119524 95.453 -77.744 49.866
119525 95.457 -77.740 49.833
119526 95.474 -77.735 49.798
119527 95.573 -77.794 49.762
119528 95.687 -77.855 49.726
119529 95.636 -77.888 49.691
119530 95.554 -77.881 49.657
119531 95.496 -77.869 49.623
119532 95.535 -77.836 49.589
119533 95.565 -77.813 49.557
119534 95.588 -77.810 49.524
119535 95.612 -77.807 49.490
119536 95.636 -77.804 49.456
119537 95.665 -77.787 49.426
119538 95.666 -77.783 49.396
119539 95.687 -77.785 49.367
119540 95.707 -77.787 49.343
119541 95.728 -77.789 49.322
119542 95.749 -77.791 49.301
119543 95.769 -77.793 49.276
119544 95.785 -77.793 49.246
119545 95.787 -77.787 49.214
119546 95.792 -77.787 49.179
119547 95.799 -77.795 49.146
119548 95.809 -77.802 49.109
119549 95.826 -77.809 49.071
119550 95.840 -77.821 49.034
119551 95.849 -77.837 48.996
119552 95.859 -77.855 48.957
119553 95.867 -77.872 48.919
119554 95.876 -77.889 48.882
119555 95.885 -77.906 48.844
119556 95.901 -77.921 48.806
119557 95.922 -77.935 48.766
119558 95.944 -77.949 48.728
119559 95.956 -77.949 48.690
119560 95.967 -77.949 48.651
119561 95.979 -77.949 48.613
119562 95.991 -77.949 48.576
119563 96.003 -77.949 48.541
119564 96.014 -77.952 48.508
119565 96.022 -77.948 48.481
119566 96.030 -77.943 48.452
119567 96.043 -77.946 48.420
119568 96.057 -77.949 48.387
119569 96.070 -77.953 48.355
119570 96.084 -77.956 48.324
119571 96.097 -77.959 48.294
119572 96.110 -77.962 48.267
119573 96.122 -77.964 48.242
119574 96.129 -77.957 48.217
119575 96.136 -77.951 48.194
119576 96.150 -77.950 48.170
119577 96.167 -77.951 48.145
119578 96.183 -77.953 48.118
119579 96.199 -77.953 48.094
119580 96.216 -77.954 48.068
119581 96.232 -77.956 48.043
119582 96.249 -77.971 48.018
119583 96.264 -77.984 47.994
119584 96.281 -77.998 47.968
119585 96.298 -78.011 47.942
119586 96.314 -78.025 47.916
119587 96.330 -78.038 47.891
119588 96.347 -78.052 47.865
119589 96.364 -78.065 47.839
119590 96.379 -78.080 47.812
119591 96.383 -78.104 47.785
119592 96.386 -78.128 47.758
119593 96.390 -78.153 47.731
119594 96.396 -78.176 47.706
119595 96.414 -78.188 47.680
119596 96.434 -78.200 47.654
119597 96.436 -78.198 47.627
119598 96.438 -78.195 47.599
119599 96.439 -78.193 47.572
119600 96.441 -78.190 47.545
119601 96.443 -78.187 47.518
119602 96.444 -78.184 47.490
119603 96.458 -78.175 47.462
119604 96.471 -78.166 47.435
119605 96.484 -78.157 47.407
119606 96.500 -78.164 47.380
119607 96.516 -78.173 47.354
119608 96.533 -78.181 47.328
119609 96.551 -78.190 47.301
119610 96.570 -78.197 47.275
119611 96.583 -78.204 47.250
119612 96.589 -78.209 47.225
119613 96.590 -78.223 47.200
119614 96.588 -78.242 47.174
119615 96.588 -78.257 47.149
119616 96.596 -78.260 47.123
119617 96.599 -78.269 47.097
119618 96.611 -78.275 47.070
119619 96.628 -78.279 47.043
119620 96.633 -78.291 47.017
119621 96.653 -78.297 46.991
119622 96.670 -78.325 46.965
119623 96.703 -78.337 46.938
119624 96.739 -78.345 46.911
119625 96.776 -78.353 46.884
119626 96.796 -78.374 46.857
119627 96.816 -78.395 46.830
119628 96.836 -78.416 46.802
119629 96.859 -78.435 46.774
119630 96.885 -78.453 46.746
119631 96.911 -78.470 46.717
119632 96.937 -78.487 46.688
119633 96.962 -78.503 46.662
119634 96.987 -78.519 46.636
119635 97.013 -78.534 46.611
119636 97.038 -78.549 46.586
119637 97.043 -78.564 46.560
119638 97.038 -78.579 46.534
119639 97.040 -78.592 46.509
119640 97.051 -78.593 46.485
119641 97.060 -78.589 46.462
119642 97.105 -78.598 46.441
119643 97.186 -78.555 46.422
119644 97.273 -78.551 46.403
119645 97.387 -78.555 46.384
119646 97.446 -78.547 46.365
119647 97.468 -78.563 46.349
119648 97.476 -78.570 46.334
119649 97.485 -78.576 46.318
119650 97.506 -78.584 46.303
119651 97.588 -78.600 46.288
119652 97.653 -78.615 46.275
119653 97.698 -78.595 46.261
119654 97.748 -78.596 46.247
119655 97.799 -78.600 46.234
119656 97.798 -78.613 46.219
119657 97.861 -78.608 46.204
119658 97.905 -78.589 46.190
119659 97.871 -78.539 46.176
119660 97.894 -78.548 46.161
119661 97.922 -78.559 46.146
119662 97.949 -78.570 46.132
119663 97.978 -78.582 46.116
119664 98.005 -78.593 46.099
119665 98.033 -78.604 46.082
119666 98.061 -78.616 46.064
119667 98.088 -78.627 46.047
119668 98.066 -78.611 46.030
119669 98.071 -78.587 46.013
119670 98.077 -78.562 45.995
119671 98.086 -78.539 45.978
119672 98.135 -78.535 45.960
119673 98.154 -78.534 45.943
119674 98.155 -78.539 45.925
119675 98.159 -78.537 45.907
119676 98.165 -78.532 45.891
119677 98.173 -78.526 45.875
119678 98.180 -78.520 45.859
119679 98.187 -78.529 45.841
119680 98.192 -78.607 45.824
119681 98.209 -78.645 45.807
119682 98.230 -78.648 45.790
119683 98.252 -78.651 45.772
119684 98.274 -78.653 45.755
119685 98.300 -78.647 45.738
119686 98.327 -78.636 45.721
119687 98.355 -78.626 45.704
119688 98.382 -78.615 45.687
119689 98.384 -78.608 45.671
119690 98.384 -78.603 45.655
119691 98.384 -78.599 45.639
119692 98.384 -78.595 45.623
119693 98.384 -78.592 45.606
119694 98.384 -78.588 45.589
119695 98.384 -78.584 45.572
119696 98.385 -78.581 45.555
119697 98.402 -78.595 45.539
119698 98.405 -78.609 45.522
119699 98.407 -78.608 45.505
119700 98.414 -78.648 45.489
119701 98.425 -78.698 45.473
119702 98.455 -78.683 45.457
119703 98.484 -78.668 45.443
119704 98.523 -78.665 45.427
119705 98.541 -78.637 45.410
119706 98.564 -78.630 45.393
119707 98.588 -78.623 45.375
119708 98.609 -78.643 45.359
119709 98.631 -78.665 45.344
119710 98.664 -78.667 45.328
119711 98.711 -78.688 45.313
119712 98.757 -78.708 45.298
119713 98.803 -78.730 45.283
119714 98.850 -78.750 45.269
119715 98.896 -78.771 45.254
119716 98.912 -78.764 45.239
119717 98.899 -78.742 45.224
119718 98.902 -78.725 45.208
119719 98.940 -78.716 45.194
119720 98.967 -78.720 45.181
119721 98.994 -78.726 45.168
119722 99.020 -78.730 45.156
119723 99.016 -78.798 45.143
119724 99.011 -78.813 45.129
119725 98.998 -78.800 45.118
119726 99.012 -78.808 45.105
119727 99.027 -78.816 45.093
119728 99.039 -78.796 45.083
119729 99.048 -78.754 45.074
119730 99.066 -78.763 45.065
119731 99.077 -78.765 45.054
119732 99.084 -78.764 45.045
119733 99.097 -78.767 45.035
119734 99.114 -78.771 45.025
119735 99.136 -78.780 45.017
119736 99.194 -78.819 45.007
119737 99.237 -78.844 44.997
119738 99.266 -78.856 44.988
119739 99.295 -78.868 44.979
119740 99.324 -78.880 44.969
119741 99.353 -78.892 44.957
119742 99.396 -78.874 44.946
119743 99.440 -78.853 44.933
119744 99.485 -78.833 44.920
119745 99.473 -78.824 44.908
119746 99.488 -78.836 44.897
119747 99.495 -78.844 44.888
119748 99.525 -78.838 44.878
119749 99.564 -78.852 44.869
119750 99.592 -78.892 44.859
119751 99.624 -78.935 44.849
119752 99.655 -78.978 44.840
119753 99.687 -79.021 44.832
119754 99.711 -79.051 44.824
119755 99.731 -79.074 44.816
119756 99.753 -79.096 44.808
119757 99.776 -79.118 44.801
119758 99.799 -79.139 44.794
119759 99.822 -79.161 44.788
119760 99.845 -79.182 44.782
119761 99.868 -79.204 44.777
119762 99.892 -79.225 44.771
119763 99.870 -79.254 44.765
119764 99.848 -79.284 44.760
119765 99.861 -79.294 44.755
119766 99.916 -79.312 44.750
119767 99.945 -79.344 44.745
119768 99.969 -79.369 44.741
119769 100.023 -79.389 44.736
119770 100.077 -79.410 44.730
119771 100.130 -79.430 44.724
119772 100.160 -79.451 44.718
119773 100.176 -79.470 44.712
119774 100.207 -79.476 44.705
119775 100.246 -79.476 44.698
119776 100.284 -79.476 44.691
119777 100.316 -79.479 44.683
119778 100.343 -79.485 44.675
119779 100.370 -79.491 44.668
119780 100.398 -79.496 44.660
119781 100.425 -79.501 44.652
119782 100.452 -79.507 44.643
119783 100.479 -79.513 44.635
119784 100.516 -79.522 44.626
119785 100.555 -79.532 44.618
119786 100.594 -79.543 44.610
119787 100.632 -79.553 44.601
119788 100.671 -79.563 44.593
119789 100.709 -79.574 44.584
119790 100.748 -79.584 44.576
119791 100.786 -79.598 44.567
119792 100.826 -79.615 44.559
119793 100.865 -79.632 44.549
119794 100.904 -79.648 44.539
119795 100.943 -79.665 44.529
119796 100.982 -79.682 44.520
119797 101.022 -79.698 44.511
119798 101.061 -79.715 44.502
119799 101.100 -79.732 44.491
119800 101.139 -79.748 44.481
119801 101.178 -79.765 44.471
119802 101.217 -79.782 44.459
119803 101.257 -79.798 44.448
119804 101.296 -79.815 44.435
119805 101.335 -79.832 44.422
119806 101.374 -79.849 44.409
119807 101.413 -79.865 44.393
119808 101.453 -79.886 44.378
119809 101.534 -79.928 44.363
119810 101.583 -79.943 44.344
119811 101.613 -79.959 44.324
119812 101.639 -79.977 44.305
119813 101.666 -79.995 44.286
119814 101.693 -80.012 44.267
119815 101.719 -80.030 44.248
119816 101.746 -80.047 44.229
119817 101.772 -80.065 44.211
119818 101.799 -80.083 44.192
119819 101.826 -80.100 44.173
120148 94.412 -77.859 52.137
120149 94.423 -77.856 52.115
120150 94.433 -77.853 52.094
120151 94.447 -77.846 52.074
120152 94.461 -77.838 52.052
120153 94.475 -77.830 52.031
120154 94.488 -77.824 52.010
120155 94.503 -77.817 51.989
120156 94.517 -77.811 51.969
120157 94.531 -77.804 51.948
120158 94.545 -77.797 51.928
120159 94.559 -77.791 51.908
120160 94.573 -77.784 51.888
120161 94.588 -77.777 51.869
120162 94.602 -77.771 51.851
120163 94.605 -77.767 51.832
120164 94.606 -77.762 51.813
120165 94.607 -77.758 51.792
120166 94.608 -77.754 51.770
120167 94.609 -77.749 51.746
120168 94.610 -77.745 51.720
120169 94.625 -77.740 51.693
120170 94.640 -77.742 51.665
120171 94.653 -77.748 51.636
120172 94.667 -77.754 51.606
120173 94.681 -77.760 51.576
120174 94.693 -77.763 51.547
120175 94.706 -77.766 51.518
120176 94.719 -77.769 51.488
120177 94.733 -77.774 51.457
120178 94.750 -77.783 51.426
120179 94.767 -77.791 51.394
120180 94.793 -77.792 51.361
120181 94.784 -77.829 51.327
120182 94.809 -77.813 51.293
120183 94.820 -77.817 51.258
120184 94.856 -77.832 51.220
120185 94.903 -77.852 51.182
120186 94.918 -77.861 51.143
120187 94.887 -77.881 51.106
120188 94.898 -77.866 51.069
120189 94.909 -77.854 51.031
120190 94.920 -77.843 50.991
120191 94.932 -77.834 50.951
120192 94.945 -77.825 50.912
120193 94.958 -77.818 50.874
120194 94.973 -77.816 50.837
120195 94.989 -77.814 50.800
120196 95.005 -77.812 50.763
120197 95.021 -77.811 50.726
120198 95.036 -77.809 50.690
120199 95.052 -77.808 50.653
120200 95.066 -77.802 50.617
120201 95.076 -77.781 50.581
120202 95.086 -77.761 50.546
120203 95.103 -77.758 50.511
120204 95.119 -77.755 50.476
120205 95.135 -77.751 50.442
120206 95.151 -77.748 50.408
120207 95.168 -77.744 50.375
120208 95.185 -77.741 50.344
120209 95.201 -77.738 50.314
120210 95.217 -77.734 50.283
120211 95.234 -77.731 50.253
120212 95.252 -77.731 50.224
120213 95.269 -77.740 50.195
120214 95.293 -77.741 50.168
120215 95.315 -77.743 50.141
120216 95.338 -77.744 50.113
120217 95.360 -77.746 50.086
120218 95.384 -77.747 50.059
120219 95.401 -77.745 50.033
120220 95.418 -77.743 50.007
120221 95.445 -77.759 49.981
120222 95.449 -77.755 49.956
120223 95.453 -77.751 49.929
120224 95.457 -77.747 49.900
120225 95.461 -77.743 49.869
120226 95.465 -77.739 49.835
120227 95.498 -77.736 49.800
120228 95.513 -77.716 49.764
120229 95.635 -77.805 49.728
120230 95.676 -77.834 49.693
120231 95.671 -77.848 49.660
120232 95.588 -77.841 49.627
120233 95.580 -77.818 49.594
120234 95.603 -77.805 49.561
120235 95.619 -77.804 49.529
120236 95.640 -77.800 49.497
120237 95.664 -77.797 49.464
120238 95.700 -77.764 49.434
120239 95.690 -77.774 49.405
120240 95.711 -77.775 49.377
120241 95.732 -77.777 49.354
120242 95.752 -77.779 49.332
120243 95.770 -77.780 49.311
120244 95.772 -77.775 49.287
120245 95.774 -77.769 49.257
120246 95.779 -77.770 49.225
120247 95.786 -77.777 49.192
120248 95.794 -77.784 49.159
120249 95.804 -77.791 49.123
120250 95.821 -77.799 49.085
120251 95.838 -77.806 49.049
120252 95.851 -77.818 49.011
120253 95.861 -77.835 48.972
120254 95.870 -77.852 48.932
120255 95.879 -77.870 48.893
120256 95.888 -77.887 48.856
120257 95.903 -77.902 48.819
120258 95.924 -77.915 48.778
120259 95.946 -77.929 48.740
120260 95.967 -77.943 48.702
120261 95.981 -77.947 48.664
120262 95.993 -77.947 48.626
120263 96.005 -77.947 48.589
120264 96.016 -77.949 48.554
120265 96.027 -77.954 48.519
120266 96.032 -77.943 48.491
120267 96.037 -77.934 48.462
120268 96.049 -77.935 48.430
120269 96.063 -77.939 48.397
120270 96.076 -77.942 48.366
120271 96.090 -77.945 48.335
120272 96.103 -77.948 48.304
120273 96.116 -77.951 48.277
120274 96.126 -77.949 48.251
120275 96.133 -77.943 48.227
120276 96.141 -77.937 48.204
120277 96.150 -77.933 48.180
120278 96.166 -77.934 48.156
120279 96.183 -77.935 48.129
120280 96.199 -77.936 48.105
120281 96.216 -77.937 48.079
120282 96.238 -77.953 48.054
120283 96.261 -77.979 48.030
120284 96.264 -77.975 48.008
120285 96.281 -77.988 47.983
120286 96.297 -78.002 47.957
120287 96.313 -78.015 47.932
120288 96.330 -78.029 47.906
120289 96.346 -78.043 47.880
120290 96.363 -78.056 47.853
120291 96.379 -78.069 47.826
120292 96.384 -78.093 47.799
120293 96.388 -78.117 47.773
120294 96.391 -78.141 47.747
120295 96.395 -78.166 47.721
120296 96.403 -78.187 47.696
120297 96.422 -78.199 47.669
120298 96.430 -78.201 47.643
120299 96.432 -78.198 47.616
120300 96.433 -78.196 47.588
120301 96.435 -78.193 47.561
120302 96.439 -78.189 47.533
120303 96.452 -78.180 47.506
120304 96.465 -78.172 47.479
120305 96.479 -78.163 47.451
120306 96.492 -78.153 47.424
120307 96.507 -78.151 47.397
120308 96.523 -78.160 47.371
120309 96.539 -78.169 47.345
120310 96.555 -78.178 47.318
120311 96.572 -78.188 47.293
120312 96.589 -78.195 47.268
120313 96.591 -78.207 47.243
120314 96.594 -78.217 47.217
120315 96.597 -78.227 47.190
120316 96.600 -78.237 47.165
120317 96.603 -78.247 47.140
120318 96.607 -78.257 47.114
120319 96.622 -78.260 47.089
120320 96.643 -78.262 47.062
120321 96.664 -78.263 47.035
120322 96.677 -78.269 47.009
120323 96.699 -78.285 46.983
120324 96.733 -78.296 46.956
120325 96.753 -78.317 46.928
120326 96.773 -78.338 46.900
120327 96.793 -78.360 46.872
120328 96.813 -78.381 46.845
120329 96.833 -78.402 46.818
120330 96.856 -78.421 46.791
120331 96.882 -78.438 46.763
120332 96.908 -78.455 46.735
120333 96.934 -78.473 46.706
120334 96.960 -78.489 46.679
120335 96.986 -78.507 46.652
120336 97.012 -78.523 46.627
120337 97.037 -78.539 46.601
120338 97.041 -78.553 46.575
120339 97.051 -78.565 46.550
120340 97.062 -78.565 46.527
120341 97.070 -78.561 46.503
120342 97.079 -78.557 46.480
120343 97.140 -78.572 46.458
120344 97.223 -78.565 46.439
120345 97.304 -78.559 46.421
120346 97.363 -78.551 46.402
120347 97.389 -78.551 46.382
120348 97.430 -78.547 46.365
120349 97.441 -78.552 46.349
120350 97.483 -78.564 46.333
120351 97.545 -78.578 46.317
120352 97.608 -78.592 46.303
120353 97.671 -78.607 46.288
120354 97.721 -78.599 46.274
120355 97.747 -78.602 46.260
120356 97.753 -78.600 46.247
120357 97.777 -78.599 46.233
120358 97.811 -78.582 46.217
120359 97.853 -78.549 46.202
120360 97.888 -78.522 46.188
120361 97.935 -78.549 46.174
120362 97.941 -78.546 46.159
120363 97.969 -78.557 46.144
120364 97.997 -78.569 46.128
120365 98.025 -78.581 46.111
120366 98.053 -78.592 46.094
120367 98.080 -78.604 46.077
120368 98.108 -78.615 46.060
120369 98.092 -78.600 46.044
120370 98.098 -78.576 46.026
120371 98.103 -78.551 46.009
120372 98.108 -78.526 45.991
120373 98.134 -78.515 45.973
120374 98.142 -78.523 45.956
120375 98.145 -78.529 45.938
120376 98.151 -78.525 45.921
120377 98.158 -78.518 45.904
120378 98.166 -78.512 45.888
120379 98.173 -78.505 45.872
120380 98.180 -78.507 45.854
120381 98.185 -78.586 45.837
120382 98.197 -78.646 45.820
120383 98.223 -78.640 45.803
120384 98.251 -78.630 45.786
120385 98.279 -78.619 45.769
120386 98.306 -78.608 45.751
120387 98.334 -78.598 45.734
120388 98.361 -78.587 45.716
120389 98.376 -78.578 45.699
120390 98.367 -78.571 45.682
120391 98.365 -78.567 45.667
120392 98.366 -78.563 45.651
120393 98.366 -78.559 45.634
120394 98.366 -78.555 45.618
120395 98.366 -78.551 45.601
120396 98.383 -78.565 45.583
120397 98.400 -78.579 45.565
120398 98.413 -78.595 45.549
120399 98.412 -78.614 45.532
120400 98.413 -78.618 45.515
120401 98.420 -78.653 45.499
120402 98.437 -78.680 45.483
120403 98.467 -78.665 45.468
120404 98.497 -78.650 45.453
120405 98.526 -78.636 45.437
120406 98.553 -78.627 45.421
120407 98.577 -78.620 45.402
120408 98.601 -78.614 45.383
120409 98.623 -78.626 45.367
120410 98.645 -78.648 45.351
120411 98.666 -78.671 45.334
120412 98.692 -78.672 45.320
120413 98.734 -78.684 45.305
120414 98.781 -78.705 45.290
120415 98.827 -78.726 45.275
120416 98.866 -78.740 45.261
120417 98.883 -78.731 45.246
120418 98.887 -78.707 45.229
120419 98.917 -78.709 45.213
120420 98.945 -78.714 45.198
120421 98.971 -78.718 45.185
120422 98.998 -78.724 45.173
120423 99.024 -78.728 45.159
120424 99.020 -78.783 45.145
120425 99.008 -78.811 45.132
120426 98.997 -78.814 45.120
120427 99.010 -78.821 45.107
120428 99.022 -78.800 45.095
120429 99.030 -78.754 45.083
120430 99.044 -78.733 45.074
120431 99.062 -78.742 45.064
120432 99.074 -78.747 45.053
120433 99.082 -78.746 45.043
120434 99.089 -78.745 45.033
120435 99.101 -78.747 45.023
120436 99.128 -78.760 45.014
120437 99.186 -78.799 45.005
120438 99.238 -78.833 44.996
120439 99.267 -78.845 44.987
120440 99.296 -78.857 44.978
120441 99.327 -78.865 44.967
120442 99.371 -78.843 44.956
120443 99.416 -78.822 44.945
120444 99.454 -78.809 44.932
120445 99.487 -78.803 44.919
120446 99.527 -78.814 44.907
120447 99.534 -78.813 44.895
120448 99.540 -78.812 44.884
120449 99.547 -78.811 44.874
120450 99.574 -78.847 44.864
120451 99.606 -78.890 44.854
120452 99.637 -78.933 44.845
120453 99.669 -78.977 44.836
120454 99.701 -79.020 44.828
120455 99.722 -79.044 44.819
120456 99.742 -79.067 44.812
120457 99.762 -79.090 44.805
120458 99.782 -79.112 44.797
120459 99.805 -79.134 44.790
120460 99.829 -79.156 44.784
120461 99.852 -79.178 44.778
120462 99.875 -79.199 44.772
120463 99.898 -79.221 44.767
120464 99.883 -79.249 44.761
120465 99.860 -79.278 44.756
120466 99.915 -79.266 44.752
120467 99.972 -79.279 44.748
120468 100.010 -79.325 44.744
120469 100.034 -79.350 44.739
120470 100.058 -79.375 44.735
120471 100.090 -79.398 44.730
120472 100.144 -79.419 44.724
120473 100.167 -79.439 44.718
120474 100.205 -79.440 44.711
120475 100.236 -79.443 44.704
120476 100.264 -79.448 44.696
120477 100.291 -79.454 44.689
120478 100.318 -79.460 44.682
120479 100.345 -79.465 44.675
120480 100.373 -79.470 44.668
120481 100.400 -79.476 44.660
120482 100.427 -79.482 44.652
120483 100.455 -79.487 44.644
120484 100.485 -79.494 44.635
120485 100.524 -79.504 44.627
120486 100.562 -79.514 44.619
120487 100.600 -79.525 44.610
120488 100.639 -79.535 44.602
120489 100.677 -79.545 44.594
120490 100.716 -79.556 44.585
120491 100.755 -79.568 44.576
120492 100.793 -79.585 44.568
120493 100.833 -79.601 44.559
120494 100.872 -79.618 44.550
120495 100.911 -79.635 44.541
120496 100.951 -79.651 44.531
120497 100.990 -79.668 44.522
120498 101.029 -79.685 44.513
120499 101.068 -79.701 44.503
120500 101.107 -79.718 44.492
120501 101.146 -79.735 44.482
120502 101.186 -79.751 44.471
120503 101.225 -79.768 44.459
120504 101.264 -79.785 44.446
120505 101.303 -79.802 44.433
120506 101.342 -79.818 44.420
120507 101.382 -79.835 44.407
120508 101.420 -79.852 44.392
120509 101.465 -79.874 44.376
120510 101.554 -79.907 44.361
120511 101.576 -79.915 44.343
120512 101.600 -79.929 44.324
120513 101.626 -79.947 44.305
120514 101.653 -79.965 44.286
120515 101.680 -79.982 44.267
120516 101.706 -80.000 44.248
120517 101.733 -80.018 44.230
120518 101.759 -80.035 44.211
120519 101.786 -80.052 44.192
120520 101.813 -80.071 44.173
120848 94.390 -77.850 52.170
120849 94.415 -77.841 52.150
120850 94.433 -77.835 52.129
120851 94.443 -77.833 52.108
120852 94.456 -77.827 52.088
120853 94.470 -77.819 52.067
120854 94.484 -77.812 52.046
120855 94.498 -77.805 52.026
120856 94.512 -77.798 52.004
120857 94.526 -77.792 51.984
120858 94.540 -77.785 51.964
120859 94.554 -77.778 51.944
120860 94.568 -77.772 51.924
120861 94.583 -77.765 51.905
120862 94.597 -77.758 51.887
120863 94.611 -77.752 51.870
120864 94.617 -77.747 51.851
120865 94.618 -77.743 51.831
120866 94.619 -77.738 51.810
120867 94.620 -77.734 51.788
120868 94.622 -77.730 51.763
120869 94.623 -77.725 51.738
120870 94.628 -77.721 51.710
120871 94.642 -77.727 51.682
120872 94.656 -77.733 51.653
120873 94.670 -77.739 51.622
120874 94.683 -77.745 51.593
120875 94.697 -77.750 51.564
120876 94.709 -77.754 51.535
120877 94.722 -77.757 51.506
120878 94.738 -77.764 51.475
120879 94.755 -77.772 51.443
120880 94.772 -77.780 51.411
120881 94.787 -77.788 51.378
120882 94.799 -77.792 51.344
120883 94.810 -77.796 51.309
120884 94.822 -77.799 51.273
120885 94.865 -77.818 51.235
120886 94.889 -77.827 51.197
120887 94.906 -77.834 51.159
120888 94.916 -77.848 51.121
120889 94.901 -77.861 51.082
120890 94.912 -77.846 51.043
120891 94.922 -77.832 51.003
120892 94.935 -77.824 50.962
120893 94.951 -77.822 50.923
120894 94.967 -77.821 50.885
120895 94.983 -77.819 50.847
120896 94.998 -77.817 50.810
120897 95.014 -77.815 50.773
120898 95.030 -77.814 50.736
120899 95.046 -77.812 50.699
120900 95.060 -77.806 50.662
120901 95.070 -77.786 50.626
120902 95.079 -77.765 50.590
120903 95.093 -77.755 50.555
120904 95.110 -77.751 50.519
120905 95.126 -77.748 50.484
120906 95.142 -77.745 50.450
120907 95.159 -77.741 50.416
120908 95.175 -77.738 50.383
120909 95.191 -77.735 50.351
120910 95.208 -77.731 50.321
120911 95.225 -77.728 50.290
120912 95.241 -77.724 50.260
120913 95.257 -77.721 50.232
120914 95.275 -77.720 50.204
120915 95.292 -77.717 50.176
120916 95.309 -77.714 50.148
120917 95.326 -77.712 50.119
120918 95.343 -77.710 50.092
120919 95.359 -77.708 50.066
120920 95.377 -77.705 50.039
120921 95.394 -77.703 50.013
120922 95.427 -77.730 49.988
120923 95.457 -77.754 49.964
120924 95.461 -77.749 49.938
120925 95.465 -77.746 49.909
120926 95.469 -77.741 49.878
120927 95.473 -77.737 49.844
120928 95.518 -77.734 49.806
120929 95.547 -77.719 49.768
120930 95.539 -77.715 49.732
120931 95.560 -77.734 49.698
120932 95.586 -77.754 49.664
120933 95.612 -77.776 49.631
120934 95.624 -77.801 49.599
120935 95.601 -77.787 49.568
120936 95.551 -77.771 49.536
120937 95.603 -77.777 49.505
120938 95.689 -77.781 49.474
120939 95.735 -77.740 49.444
120940 95.714 -77.764 49.416
120941 95.735 -77.766 49.388
120942 95.756 -77.768 49.363
120943 95.758 -77.762 49.340
120944 95.760 -77.757 49.319
120945 95.762 -77.752 49.295
120946 95.766 -77.752 49.266
120947 95.773 -77.759 49.236
120948 95.781 -77.766 49.204
120949 95.789 -77.773 49.169
120950 95.799 -77.780 49.133
120951 95.816 -77.788 49.097
120952 95.833 -77.796 49.061
120953 95.850 -77.802 49.025
120954 95.863 -77.815 48.987
120955 95.872 -77.833 48.947
120956 95.881 -77.850 48.906
120957 95.890 -77.868 48.867
120958 95.905 -77.883 48.829
120959 95.926 -77.896 48.791
120960 95.947 -77.909 48.753
120961 95.968 -77.923 48.715
120962 95.989 -77.937 48.677
120963 96.007 -77.944 48.638
120964 96.018 -77.946 48.601
120965 96.031 -77.956 48.566
120966 96.037 -77.949 48.530
120967 96.042 -77.940 48.500
120968 96.047 -77.930 48.470
120969 96.056 -77.924 48.439
120970 96.069 -77.928 48.407
120971 96.083 -77.931 48.375
120972 96.096 -77.934 48.344
120973 96.109 -77.937 48.314
120974 96.122 -77.940 48.287
120975 96.130 -77.936 48.262
120976 96.137 -77.930 48.238
120977 96.145 -77.924 48.215
120978 96.151 -77.917 48.190
120979 96.166 -77.917 48.165
120980 96.182 -77.918 48.139
120981 96.198 -77.918 48.116
120982 96.215 -77.920 48.091
120983 96.244 -77.951 48.066
120984 96.274 -77.986 48.043
120985 96.277 -77.983 48.021
120986 96.281 -77.980 47.997
120987 96.296 -77.992 47.973
120988 96.313 -78.006 47.947
120989 96.329 -78.019 47.921
120990 96.345 -78.033 47.894
120991 96.362 -78.047 47.868
120992 96.378 -78.060 47.843
120993 96.386 -78.081 47.817
120994 96.390 -78.106 47.791
120995 96.393 -78.130 47.765
120996 96.397 -78.154 47.739
120997 96.400 -78.179 47.712
120998 96.411 -78.197 47.683
120999 96.424 -78.204 47.656
121000 96.426 -78.202 47.629
121001 96.427 -78.199 47.603
121002 96.433 -78.194 47.575
121003 96.446 -78.185 47.548
121004 96.460 -78.176 47.521
121005 96.473 -78.168 47.495
121006 96.486 -78.159 47.469
121007 96.500 -78.150 47.441
121008 96.513 -78.141 47.413
121009 96.529 -78.148 47.387
121010 96.545 -78.157 47.361
121011 96.561 -78.166 47.336
121012 96.578 -78.175 47.311
121013 96.593 -78.184 47.285
121014 96.598 -78.194 47.258
121015 96.601 -78.204 47.232
121016 96.604 -78.214 47.207
121017 96.607 -78.224 47.182
121018 96.611 -78.233 47.158
121019 96.614 -78.243 47.132
121020 96.635 -78.245 47.106
121021 96.655 -78.247 47.079
121022 96.675 -78.248 47.052
121023 96.696 -78.250 47.027
121024 96.710 -78.260 47.001
121025 96.730 -78.281 46.973
121026 96.750 -78.302 46.945
121027 96.770 -78.323 46.916
121028 96.790 -78.344 46.889
121029 96.810 -78.366 46.861
121030 96.830 -78.387 46.834
121031 96.853 -78.406 46.807
121032 96.879 -78.423 46.780
121033 96.905 -78.441 46.753
121034 96.932 -78.457 46.725
121035 96.957 -78.475 46.697
121036 96.983 -78.492 46.670
121037 97.010 -78.509 46.644
121038 97.035 -78.526 46.619
121039 97.061 -78.537 46.594
121040 97.072 -78.536 46.569
121041 97.080 -78.532 46.545
121042 97.089 -78.529 46.521
121043 97.097 -78.524 46.498
121044 97.131 -78.545 46.477
121045 97.213 -78.530 46.457
121046 97.282 -78.514 46.438
121047 97.362 -78.523 46.419
121048 97.383 -78.525 46.399
121049 97.404 -78.527 46.381
121050 97.450 -78.538 46.364
121051 97.511 -78.552 46.348
121052 97.571 -78.567 46.332
121053 97.632 -78.582 46.317
121054 97.691 -78.598 46.302
121055 97.722 -78.603 46.288
121056 97.747 -78.604 46.274
121057 97.772 -78.604 46.260
121058 97.797 -78.605 46.245
121059 97.825 -78.598 46.230
121060 97.867 -78.567 46.215
121061 97.899 -78.539 46.201
121062 97.917 -78.526 46.186
121063 97.971 -78.535 46.171
121064 97.992 -78.547 46.156
121065 98.017 -78.557 46.140
121066 98.044 -78.568 46.124
121067 98.072 -78.579 46.108
121068 98.100 -78.591 46.091
121069 98.128 -78.602 46.074
121070 98.119 -78.589 46.057
121071 98.124 -78.564 46.040
121072 98.129 -78.540 46.023
121073 98.135 -78.515 46.005
121074 98.131 -78.507 45.987
121075 98.137 -78.515 45.969
121076 98.145 -78.523 45.952
121077 98.148 -78.514 45.934
121078 98.151 -78.504 45.918
121079 98.158 -78.498 45.902
121080 98.166 -78.491 45.885
121081 98.173 -78.487 45.868
121082 98.180 -78.561 45.851
121083 98.211 -78.578 45.834
121084 98.230 -78.601 45.816
121085 98.257 -78.591 45.800
121086 98.285 -78.580 45.783
121087 98.312 -78.570 45.764
121088 98.340 -78.559 45.747
121089 98.367 -78.549 45.729
121090 98.359 -78.542 45.711
121091 98.350 -78.535 45.694
121092 98.347 -78.529 45.679
121093 98.347 -78.526 45.663
121094 98.347 -78.522 45.645
121095 98.364 -78.535 45.629
121096 98.381 -78.550 45.613
121097 98.398 -78.564 45.595
121098 98.415 -78.577 45.576
121099 98.420 -78.595 45.559
121100 98.420 -78.614 45.543
121101 98.420 -78.628 45.526
121102 98.426 -78.657 45.510
121103 98.450 -78.661 45.494
121104 98.479 -78.647 45.479
121105 98.507 -78.636 45.463
121106 98.533 -78.631 45.447
121107 98.562 -78.622 45.430
121108 98.590 -78.611 45.411
121109 98.614 -78.604 45.392
121110 98.637 -78.611 45.375
121111 98.658 -78.633 45.358
121112 98.679 -78.655 45.342
121113 98.701 -78.677 45.327
121114 98.725 -78.686 45.312
121115 98.758 -78.681 45.297
121116 98.801 -78.699 45.282
121117 98.831 -78.705 45.267
121118 98.851 -78.699 45.251
121119 98.882 -78.704 45.234
121120 98.914 -78.708 45.218
121121 98.945 -78.711 45.204
121122 98.975 -78.716 45.191
121123 99.002 -78.721 45.178
121124 99.028 -78.726 45.163
121125 99.025 -78.768 45.149
121126 99.012 -78.796 45.135
121127 99.000 -78.824 45.123
121128 99.005 -78.804 45.110
121129 99.014 -78.757 45.097
121130 99.022 -78.711 45.086
121131 99.040 -78.714 45.076
121132 99.057 -78.723 45.065
121133 99.072 -78.729 45.054
121134 99.079 -78.728 45.043
121135 99.086 -78.727 45.032
121136 99.094 -78.727 45.021
121137 99.120 -78.740 45.012
121138 99.178 -78.780 45.003
121139 99.235 -78.818 44.993
121140 99.268 -78.833 44.984
121141 99.303 -78.833 44.974
121142 99.346 -78.814 44.964
121143 99.378 -78.808 44.954
121144 99.411 -78.802 44.943
121145 99.443 -78.796 44.931
121146 99.482 -78.796 44.918
121147 99.537 -78.815 44.905
121148 99.551 -78.818 44.894
121149 99.557 -78.816 44.884
121150 99.563 -78.815 44.873
121151 99.587 -78.845 44.862
121152 99.619 -78.888 44.852
121153 99.651 -78.931 44.843
121154 99.683 -78.974 44.834
121155 99.713 -79.015 44.825
121156 99.733 -79.037 44.817
121157 99.753 -79.060 44.809
121158 99.773 -79.083 44.801
121159 99.793 -79.106 44.793
121160 99.813 -79.129 44.787
121161 99.835 -79.151 44.781
121162 99.858 -79.172 44.774
121163 99.881 -79.194 44.769
121164 99.905 -79.216 44.764
121165 99.895 -79.243 44.759
121166 99.886 -79.265 44.754
121167 99.968 -79.237 44.751
121168 100.028 -79.246 44.747
121169 100.066 -79.294 44.743
121170 100.099 -79.332 44.739
121171 100.123 -79.357 44.735
121172 100.148 -79.382 44.731
121173 100.157 -79.405 44.724
121174 100.184 -79.412 44.718
121175 100.211 -79.418 44.711
121176 100.239 -79.423 44.703
121177 100.266 -79.429 44.696
121178 100.293 -79.434 44.689
121179 100.321 -79.440 44.682
121180 100.348 -79.445 44.676
121181 100.375 -79.451 44.669
121182 100.403 -79.456 44.661
121183 100.430 -79.462 44.652
121184 100.457 -79.467 44.645
121185 100.492 -79.476 44.637
121186 100.530 -79.486 44.628
121187 100.569 -79.496 44.620
121188 100.607 -79.507 44.612
121189 100.645 -79.517 44.603
121190 100.684 -79.527 44.595
121191 100.722 -79.538 44.586
121192 100.762 -79.554 44.578
121193 100.801 -79.571 44.570
121194 100.840 -79.588 44.561
121195 100.879 -79.604 44.552
121196 100.919 -79.621 44.543
121197 100.958 -79.638 44.534
121198 100.997 -79.654 44.525
121199 101.036 -79.671 44.515
121200 101.075 -79.688 44.504
121201 101.115 -79.704 44.494
121202 101.154 -79.721 44.482
121203 101.193 -79.738 44.471
121204 101.232 -79.755 44.458
121205 101.271 -79.771 44.445
121206 101.311 -79.788 44.432
121207 101.350 -79.805 44.418
121208 101.385 -79.823 44.405
121209 101.417 -79.842 44.390
121210 101.489 -79.871 44.374
121211 101.555 -79.892 44.358
121212 101.577 -79.902 44.341
121213 101.594 -79.908 44.323
121214 101.613 -79.917 44.304
121215 101.640 -79.934 44.285
121216 101.667 -79.952 44.267
121217 101.693 -79.970 44.248
121218 101.720 -79.987 44.230
121219 101.746 -80.005 44.211
121220 101.773 -80.023 44.192
121221 101.799 -80.040 44.173
121548 94.367 -77.843 52.206
121549 94.392 -77.834 52.184
121550 94.417 -77.824 52.164
121551 94.441 -77.815 52.143
121552 94.454 -77.812 52.122
121553 94.465 -77.808 52.102
121554 94.479 -77.800 52.082
121555 94.493 -77.793 52.062
121556 94.507 -77.786 52.042
121557 94.521 -77.779 52.021
121558 94.535 -77.773 52.001
121559 94.549 -77.766 51.981
121560 94.563 -77.759 51.961
121561 94.578 -77.752 51.942
121562 94.592 -77.746 51.924
121563 94.606 -77.740 51.906
121564 94.620 -77.733 51.888
121565 94.630 -77.727 51.870
121566 94.631 -77.723 51.850
121567 94.632 -77.718 51.828
121568 94.633 -77.714 51.805
121569 94.634 -77.710 51.780
121570 94.635 -77.705 51.754
121571 94.632 -77.707 51.727
121572 94.645 -77.712 51.697
121573 94.658 -77.718 51.668
121574 94.672 -77.724 51.638
121575 94.685 -77.730 51.609
121576 94.699 -77.736 51.581
121577 94.712 -77.741 51.553
121578 94.725 -77.744 51.524
121579 94.742 -77.753 51.493
121580 94.759 -77.761 51.461
121581 94.774 -77.766 51.428
121582 94.788 -77.771 51.394
121583 94.800 -77.774 51.360
121584 94.812 -77.778 51.325
121585 94.824 -77.782 51.289
121586 94.858 -77.796 51.251
121587 94.875 -77.802 51.213
121588 94.891 -77.808 51.175
121589 94.904 -77.819 51.135
121590 94.915 -77.834 51.095
121591 94.915 -77.841 51.055
121592 94.921 -77.817 51.014
121593 94.945 -77.827 50.974
121594 94.961 -77.825 50.935
121595 94.976 -77.824 50.896
121596 94.992 -77.821 50.858
121597 95.008 -77.820 50.820
121598 95.024 -77.818 50.783
121599 95.039 -77.817 50.746
121600 95.054 -77.811 50.709
121601 95.063 -77.790 50.672
121602 95.073 -77.770 50.635
121603 95.084 -77.752 50.599
121604 95.100 -77.748 50.564
121605 95.117 -77.745 50.529
121606 95.133 -77.742 50.492
121607 95.149 -77.738 50.458
121608 95.166 -77.735 50.424
121609 95.182 -77.731 50.391
121610 95.198 -77.728 50.359
121611 95.215 -77.725 50.328
121612 95.232 -77.721 50.298
121613 95.248 -77.718 50.268
121614 95.264 -77.718 50.239
121615 95.282 -77.708 50.212
121616 95.289 -77.700 50.184
121617 95.300 -77.693 50.155
121618 95.315 -77.689 50.128
121619 95.330 -77.683 50.101
121620 95.345 -77.679 50.074
121621 95.359 -77.674 50.048
121622 95.375 -77.669 50.023
121623 95.405 -77.695 49.998
121624 95.440 -77.724 49.974
121625 95.469 -77.748 49.946
121626 95.473 -77.744 49.916
121627 95.477 -77.740 49.885
121628 95.481 -77.736 49.851
121629 95.537 -77.733 49.815
121630 95.582 -77.722 49.778
121631 95.568 -77.727 49.741
121632 95.547 -77.742 49.705
121633 95.565 -77.761 49.671
121634 95.590 -77.780 49.637
121635 95.606 -77.795 49.605
121636 95.556 -77.780 49.574
121637 95.548 -77.749 49.545
121638 95.593 -77.751 49.516
121639 95.671 -77.741 49.486
121640 95.759 -77.727 49.456
121641 95.738 -77.754 49.427
121642 95.743 -77.750 49.400
121643 95.745 -77.745 49.374
121644 95.747 -77.739 49.350
121645 95.749 -77.733 49.327
121646 95.754 -77.734 49.303
121647 95.761 -77.741 49.277
121648 95.768 -77.749 49.248
121649 95.776 -77.755 49.216
121650 95.783 -77.762 49.181
121651 95.793 -77.770 49.145
121652 95.810 -77.777 49.110
121653 95.828 -77.784 49.072
121654 95.845 -77.792 49.035
121655 95.863 -77.799 48.998
121656 95.874 -77.813 48.959
121657 95.883 -77.830 48.919
121658 95.892 -77.848 48.879
121659 95.907 -77.863 48.841
121660 95.928 -77.877 48.803
121661 95.949 -77.890 48.765
121662 95.970 -77.904 48.727
121663 95.991 -77.917 48.690
121664 96.012 -77.930 48.651
121665 96.027 -77.937 48.613
121666 96.037 -77.940 48.577
121667 96.047 -77.945 48.542
121668 96.053 -77.936 48.511
121669 96.058 -77.926 48.480
121670 96.063 -77.916 48.448
121671 96.076 -77.917 48.417
121672 96.089 -77.920 48.385
121673 96.102 -77.924 48.354
121674 96.115 -77.927 48.324
121675 96.128 -77.928 48.298
121676 96.135 -77.922 48.272
121677 96.142 -77.915 48.248
121678 96.149 -77.909 48.225
121679 96.155 -77.903 48.201
121680 96.166 -77.899 48.176
121681 96.182 -77.900 48.150
121682 96.198 -77.902 48.126
121683 96.219 -77.912 48.102
121684 96.250 -77.949 48.078
121685 96.281 -77.985 48.056
121686 96.290 -77.990 48.034
121687 96.293 -77.987 48.011
121688 96.296 -77.984 47.986
121689 96.312 -77.996 47.960
121690 96.328 -78.010 47.934
121691 96.344 -78.024 47.908
121692 96.361 -78.037 47.882
121693 96.378 -78.050 47.857
121694 96.388 -78.069 47.832
121695 96.392 -78.094 47.808
121696 96.395 -78.119 47.782
121697 96.399 -78.143 47.756
121698 96.402 -78.167 47.730
121699 96.406 -78.192 47.703
121700 96.418 -78.208 47.675
121701 96.419 -78.205 47.647
121702 96.428 -78.199 47.620
121703 96.441 -78.191 47.592
121704 96.454 -78.181 47.566
121705 96.467 -78.172 47.540
121706 96.481 -78.164 47.513
121707 96.494 -78.155 47.486
121708 96.508 -78.146 47.458
121709 96.521 -78.138 47.431
121710 96.535 -78.136 47.405
121711 96.551 -78.145 47.378
121712 96.568 -78.154 47.352
121713 96.584 -78.163 47.327
121714 96.598 -78.172 47.300
121715 96.605 -78.181 47.274
121716 96.608 -78.191 47.248
121717 96.611 -78.200 47.223
121718 96.615 -78.210 47.199
121719 96.618 -78.220 47.174
121720 96.625 -78.228 47.148
121721 96.646 -78.229 47.123
121722 96.667 -78.231 47.096
121723 96.687 -78.232 47.069
121724 96.704 -78.238 47.043
121725 96.719 -78.248 47.018
121726 96.728 -78.267 46.990
121727 96.747 -78.288 46.962
121728 96.767 -78.309 46.934
121729 96.787 -78.330 46.907
121730 96.807 -78.351 46.880
121731 96.827 -78.373 46.852
121732 96.851 -78.391 46.826
121733 96.877 -78.408 46.799
121734 96.902 -78.426 46.771
121735 96.929 -78.443 46.743
121736 96.955 -78.460 46.716
121737 96.980 -78.477 46.689
121738 97.008 -78.489 46.663
121739 97.036 -78.489 46.639
121740 97.065 -78.490 46.614
121741 97.080 -78.495 46.590
121742 97.092 -78.501 46.566
121743 97.104 -78.507 46.541
121744 97.116 -78.514 46.518
121745 97.124 -78.505 46.496
121746 97.187 -78.484 46.475
121747 97.268 -78.477 46.456
121748 97.360 -78.496 46.436
121749 97.382 -78.498 46.417
121750 97.429 -78.509 46.398
121751 97.489 -78.523 46.380
121752 97.547 -78.539 46.363
121753 97.590 -78.551 46.347
121754 97.632 -78.564 46.331
121755 97.674 -78.576 46.317
121756 97.721 -78.592 46.302
121757 97.767 -78.609 46.288
121758 97.792 -78.610 46.273
121759 97.816 -78.611 46.257
121760 97.841 -78.611 46.243
121761 97.880 -78.584 46.228
121762 97.911 -78.557 46.213
121763 97.922 -78.541 46.198
121764 97.946 -78.531 46.182
121765 97.996 -78.535 46.167
121766 98.055 -78.557 46.152
121767 98.064 -78.555 46.136
121768 98.092 -78.567 46.121
121769 98.120 -78.579 46.104
121770 98.148 -78.590 46.087
121771 98.145 -78.578 46.069
121772 98.150 -78.553 46.053
121773 98.156 -78.529 46.036
121774 98.148 -78.516 46.019
121775 98.141 -78.506 46.000
121776 98.148 -78.514 45.982
121777 98.156 -78.521 45.966
121778 98.169 -78.525 45.949
121779 98.168 -78.511 45.933
121780 98.166 -78.497 45.916
121781 98.164 -78.482 45.900
121782 98.166 -78.470 45.883
121783 98.193 -78.493 45.866
121784 98.224 -78.510 45.848
121785 98.242 -78.537 45.831
121786 98.264 -78.552 45.813
121787 98.291 -78.542 45.796
121788 98.318 -78.531 45.778
121789 98.346 -78.521 45.759
121790 98.352 -78.513 45.742
121791 98.342 -78.505 45.725
121792 98.333 -78.498 45.708
121793 98.329 -78.493 45.691
121794 98.345 -78.506 45.674
121795 98.362 -78.520 45.656
121796 98.379 -78.534 45.639
121797 98.396 -78.548 45.622
121798 98.413 -78.561 45.605
121799 98.429 -78.576 45.586
121800 98.428 -78.595 45.569
121801 98.428 -78.614 45.552
121802 98.427 -78.633 45.536
121803 98.432 -78.658 45.520
121804 98.462 -78.643 45.504
121805 98.488 -78.638 45.488
121806 98.514 -78.632 45.471
121807 98.539 -78.626 45.455
121808 98.568 -78.617 45.438
121809 98.600 -78.604 45.420
121810 98.627 -78.594 45.400
121811 98.650 -78.595 45.383
121812 98.671 -78.617 45.366
121813 98.693 -78.639 45.350
121814 98.714 -78.661 45.335
121815 98.735 -78.683 45.319
121816 98.758 -78.702 45.303
121817 98.788 -78.693 45.289
121818 98.818 -78.686 45.273
121819 98.839 -78.685 45.257
121820 98.875 -78.688 45.240
121821 98.908 -78.699 45.224
121822 98.941 -78.710 45.209
121823 98.973 -78.714 45.196
121824 99.004 -78.719 45.182
121825 99.032 -78.724 45.167
121826 99.030 -78.753 45.152
121827 99.012 -78.773 45.139
121828 98.992 -78.776 45.126
121829 98.998 -78.748 45.113
121830 99.006 -78.714 45.100
121831 99.017 -78.684 45.090
121832 99.035 -78.693 45.080
121833 99.069 -78.695 45.069
121834 99.076 -78.705 45.056
121835 99.085 -78.703 45.043
121836 99.094 -78.701 45.032
121837 99.106 -78.717 45.021
121838 99.126 -78.752 45.011
121839 99.178 -78.779 45.000
121840 99.230 -78.805 44.990
121841 99.271 -78.811 44.981
121842 99.303 -78.805 44.970
121843 99.336 -78.799 44.960
121844 99.368 -78.794 44.950
121845 99.401 -78.788 44.939
121846 99.433 -78.782 44.927
121847 99.481 -78.793 44.916
121848 99.535 -78.811 44.905
121849 99.567 -78.821 44.895
121850 99.574 -78.820 44.884
121851 99.580 -78.819 44.874
121852 99.601 -78.843 44.863
121853 99.633 -78.886 44.852
121854 99.664 -78.930 44.843
121855 99.696 -78.972 44.834
121856 99.724 -79.009 44.824
121857 99.744 -79.031 44.815
121858 99.764 -79.054 44.806
121859 99.784 -79.077 44.798
121860 99.804 -79.100 44.790
121861 99.824 -79.122 44.784
121862 99.844 -79.145 44.778
121863 99.865 -79.168 44.771
121864 99.889 -79.189 44.766
121865 99.912 -79.211 44.761
121866 99.906 -79.237 44.756
121867 99.938 -79.237 44.752
121868 100.020 -79.209 44.749
121869 100.064 -79.239 44.746
121870 100.095 -79.297 44.742
121871 100.127 -79.313 44.739
121872 100.133 -79.338 44.735
121873 100.140 -79.363 44.731
121874 100.159 -79.384 44.725
121875 100.187 -79.392 44.718
121876 100.214 -79.397 44.711
121877 100.242 -79.404 44.704
121878 100.269 -79.409 44.697
121879 100.296 -79.414 44.690
121880 100.323 -79.419 44.683
121881 100.351 -79.425 44.677
121882 100.378 -79.431 44.670
121883 100.405 -79.436 44.662
121884 100.432 -79.441 44.655
121885 100.460 -79.447 44.647
121886 100.498 -79.457 44.639
121887 100.537 -79.468 44.631
121888 100.575 -79.479 44.621
121889 100.614 -79.488 44.613
121890 100.652 -79.499 44.605
121891 100.691 -79.509 44.597
121892 100.730 -79.524 44.589
121893 100.769 -79.541 44.580
121894 100.809 -79.557 44.572
121895 100.847 -79.574 44.564
121896 100.887 -79.591 44.555
121897 100.926 -79.607 44.546
121898 100.965 -79.624 44.537
121899 101.004 -79.641 44.527
121900 101.044 -79.657 44.516
121901 101.082 -79.674 44.505
121902 101.122 -79.691 44.494
121903 101.161 -79.708 44.483
121904 101.200 -79.724 44.471
121905 101.240 -79.741 44.458
121906 101.279 -79.758 44.445
121907 101.317 -79.774 44.431
121908 101.350 -79.793 44.417
121909 101.382 -79.813 44.402
121910 101.421 -79.833 44.386
121911 101.512 -79.867 44.371
121912 101.554 -79.877 44.355
121913 101.578 -79.889 44.338
121914 101.595 -79.895 44.320
121915 101.612 -79.901 44.302
121916 101.628 -79.906 44.283
121917 101.653 -79.922 44.264
121918 101.680 -79.939 44.246
121919 101.707 -79.958 44.229
121920 101.734 -79.975 44.210
121921 101.760 -79.993 44.191
122248 94.343 -77.835 52.241
122249 94.368 -77.826 52.220
122250 94.393 -77.817 52.199
122251 94.418 -77.808 52.178
122252 94.443 -77.799 52.158
122253 94.464 -77.791 52.137
122254 94.474 -77.788 52.117
122255 94.488 -77.780 52.097
122256 94.502 -77.773 52.078
122257 94.516 -77.767 52.058
122258 94.530 -77.760 52.037
122259 94.544 -77.753 52.017
122260 94.558 -77.747 51.998
122261 94.573 -77.740 51.979
122262 94.587 -77.733 51.961
122263 94.601 -77.727 51.943
122264 94.615 -77.721 51.925
122265 94.629 -77.714 51.906
122266 94.642 -77.708 51.886
122267 94.644 -77.703 51.865
122268 94.645 -77.699 51.842
122269 94.646 -77.695 51.818
122270 94.646 -77.692 51.794
122271 94.642 -77.693 51.769
122272 94.639 -77.695 51.742
122273 94.648 -77.699 51.713
122274 94.662 -77.705 51.685
122275 94.675 -77.710 51.656
122276 94.688 -77.716 51.629
122277 94.702 -77.721 51.602
122278 94.715 -77.727 51.574
122279 94.729 -77.733 51.544
122280 94.744 -77.739 51.511
122281 94.759 -77.744 51.478
122282 94.774 -77.749 51.445
122283 94.789 -77.755 51.411
122284 94.802 -77.758 51.377
122285 94.813 -77.761 51.342
122286 94.825 -77.765 51.306
122287 94.843 -77.771 51.267
122288 94.860 -77.777 51.228
122289 94.877 -77.783 51.190
122290 94.892 -77.791 51.150
122291 94.903 -77.806 51.110
122292 94.913 -77.819 51.069
122293 94.920 -77.800 51.026
122294 94.949 -77.820 50.986
122295 94.970 -77.828 50.946
122296 94.986 -77.827 50.907
122297 95.001 -77.824 50.869
122298 95.017 -77.823 50.832
122299 95.033 -77.821 50.795
122300 95.047 -77.815 50.757
122301 95.057 -77.795 50.720
122302 95.067 -77.774 50.683
122303 95.076 -77.755 50.646
122304 95.091 -77.745 50.610
122305 95.107 -77.742 50.575
122306 95.123 -77.739 50.538
122307 95.140 -77.735 50.501
122308 95.157 -77.732 50.467
122309 95.173 -77.728 50.433
122310 95.189 -77.725 50.400
122311 95.205 -77.722 50.367
122312 95.222 -77.718 50.336
122313 95.241 -77.718 50.305
122314 95.258 -77.718 50.276
122315 95.266 -77.713 50.246
122316 95.291 -77.700 50.218
122317 95.301 -77.691 50.191
122318 95.308 -77.683 50.162
122319 95.316 -77.675 50.136
122320 95.327 -77.669 50.109
122321 95.342 -77.664 50.082
122322 95.357 -77.659 50.057
122323 95.372 -77.654 50.032
122324 95.395 -77.671 50.007
122325 95.419 -77.691 49.981
122326 95.451 -77.718 49.952
122327 95.482 -77.743 49.921
122328 95.486 -77.738 49.889
122329 95.490 -77.734 49.857
122330 95.557 -77.732 49.823
122331 95.617 -77.725 49.787
122332 95.597 -77.738 49.750
122333 95.577 -77.753 49.715
122334 95.555 -77.769 49.679
122335 95.570 -77.787 49.645
122336 95.567 -77.793 49.613
122337 95.551 -77.752 49.582
122338 95.584 -77.735 49.554
122339 95.628 -77.736 49.526
122340 95.673 -77.738 49.498
122341 95.732 -77.735 49.467
122342 95.731 -77.732 49.438
122343 95.733 -77.727 49.411
122344 95.735 -77.721 49.386
122345 95.737 -77.716 49.360
122346 95.741 -77.716 49.335
122347 95.749 -77.724 49.311
122348 95.756 -77.730 49.285
122349 95.763 -77.737 49.258
122350 95.771 -77.745 49.228
122351 95.778 -77.752 49.192
122352 95.788 -77.759 49.157
122353 95.805 -77.766 49.121
122354 95.822 -77.774 49.082
122355 95.840 -77.781 49.045
122356 95.857 -77.789 49.007
122357 95.875 -77.796 48.969
122358 95.886 -77.811 48.929
122359 95.894 -77.828 48.890
122360 95.909 -77.843 48.852
122361 95.930 -77.857 48.814
122362 95.951 -77.871 48.776
122363 95.972 -77.884 48.739
122364 95.991 -77.895 48.702
122365 96.010 -77.905 48.663
122366 96.028 -77.914 48.626
122367 96.041 -77.920 48.590
122368 96.052 -77.925 48.555
122369 96.062 -77.930 48.522
122370 96.068 -77.922 48.490
122371 96.074 -77.912 48.458
122372 96.082 -77.906 48.428
122373 96.095 -77.909 48.396
122374 96.108 -77.912 48.365
122375 96.122 -77.915 48.334
122376 96.132 -77.915 48.308
122377 96.139 -77.908 48.283
122378 96.146 -77.902 48.257
122379 96.153 -77.896 48.234
122380 96.159 -77.890 48.211
122381 96.166 -77.883 48.186
122382 96.182 -77.884 48.161
122383 96.198 -77.884 48.136
122384 96.225 -77.910 48.113
122385 96.256 -77.946 48.090
122386 96.287 -77.983 48.068
122387 96.302 -77.998 48.046
122388 96.306 -77.994 48.022
122389 96.309 -77.991 47.997
122390 96.312 -77.988 47.972
122391 96.327 -78.000 47.947
122392 96.344 -78.014 47.922
122393 96.361 -78.028 47.896
122394 96.377 -78.041 47.871
122395 96.390 -78.058 47.847
122396 96.393 -78.082 47.823
122397 96.397 -78.106 47.798
122398 96.400 -78.131 47.773
122399 96.404 -78.156 47.747
122400 96.407 -78.180 47.721
122401 96.411 -78.205 47.693
122402 96.421 -78.204 47.665
122403 96.435 -78.195 47.636
122404 96.448 -78.187 47.609
122405 96.462 -78.178 47.584
122406 96.475 -78.169 47.558
122407 96.488 -78.160 47.531
122408 96.502 -78.151 47.503
122409 96.515 -78.142 47.475
122410 96.528 -78.134 47.449
122411 96.542 -78.125 47.422
122412 96.558 -78.133 47.395
122413 96.574 -78.142 47.368
122414 96.590 -78.151 47.342
122415 96.602 -78.159 47.314
122416 96.612 -78.168 47.289
122417 96.615 -78.178 47.264
122418 96.619 -78.188 47.239
122419 96.622 -78.197 47.215
122420 96.625 -78.207 47.190
122421 96.638 -78.213 47.164
122422 96.658 -78.214 47.140
122423 96.678 -78.216 47.114
122424 96.698 -78.218 47.087
122425 96.712 -78.228 47.060
122426 96.726 -78.238 47.034
122427 96.737 -78.255 47.008
122428 96.746 -78.273 46.981
122429 96.764 -78.294 46.953
122430 96.784 -78.316 46.926
122431 96.804 -78.337 46.899
122432 96.824 -78.358 46.872
122433 96.848 -78.376 46.846
122434 96.874 -78.394 46.818
122435 96.900 -78.411 46.789
122436 96.926 -78.428 46.761
122437 96.952 -78.445 46.734
122438 96.980 -78.451 46.708
122439 97.009 -78.452 46.683
122440 97.036 -78.460 46.659
122441 97.063 -78.471 46.636
122442 97.090 -78.485 46.612
122443 97.112 -78.501 46.588
122444 97.124 -78.507 46.562
122445 97.136 -78.514 46.538
122446 97.157 -78.502 46.515
122447 97.198 -78.470 46.493
122448 97.266 -78.456 46.473
122449 97.357 -78.468 46.454
122450 97.404 -78.479 46.435
122451 97.446 -78.492 46.416
122452 97.488 -78.504 46.397
122453 97.530 -78.517 46.379
122454 97.573 -78.529 46.363
122455 97.615 -78.542 46.347
122456 97.657 -78.554 46.332
122457 97.699 -78.566 46.317
122458 97.754 -78.589 46.302
122459 97.811 -78.616 46.286
122460 97.836 -78.617 46.271
122461 97.861 -78.617 46.256
122462 97.894 -78.601 46.241
122463 97.923 -78.576 46.226
122464 97.934 -78.559 46.211
122465 97.945 -78.542 46.195
122466 97.974 -78.535 46.179
122467 98.024 -78.540 46.163
122468 98.075 -78.545 46.148
122469 98.112 -78.554 46.132
122470 98.139 -78.566 46.115
122471 98.167 -78.577 46.098
122472 98.171 -78.567 46.081
122473 98.177 -78.542 46.064
122474 98.169 -78.529 46.048
122475 98.161 -78.517 46.030
122476 98.152 -78.506 46.013
122477 98.158 -78.513 45.996
122478 98.167 -78.519 45.981
122479 98.180 -78.523 45.964
122480 98.192 -78.524 45.948
122481 98.189 -78.509 45.932
122482 98.187 -78.494 45.916
122483 98.188 -78.475 45.899
122484 98.212 -78.460 45.881
122485 98.238 -78.465 45.864
122486 98.256 -78.469 45.846
122487 98.274 -78.496 45.828
122488 98.297 -78.503 45.810
122489 98.325 -78.493 45.793
122490 98.344 -78.483 45.774
122491 98.335 -78.476 45.756
122492 98.325 -78.469 45.738
122493 98.333 -78.476 45.721
122494 98.344 -78.485 45.704
122495 98.360 -78.504 45.686
122496 98.377 -78.518 45.668
122497 98.394 -78.532 45.650
122498 98.411 -78.546 45.633
122499 98.428 -78.560 45.615
122500 98.437 -78.576 45.596
122501 98.436 -78.595 45.578
122502 98.436 -78.614 45.561
122503 98.435 -78.633 45.545
122504 98.443 -78.644 45.529
122505 98.468 -78.639 45.512
122506 98.495 -78.633 45.496
122507 98.520 -78.628 45.478
122508 98.546 -78.622 45.462
122509 98.575 -78.613 45.445
122510 98.606 -78.599 45.428
122511 98.638 -78.586 45.408
122512 98.663 -78.578 45.391
122513 98.685 -78.600 45.374
122514 98.706 -78.622 45.357
122515 98.728 -78.644 45.342
122516 98.750 -78.665 45.326
122517 98.761 -78.677 45.310
122518 98.787 -78.669 45.294
122519 98.817 -78.661 45.278
122520 98.834 -78.664 45.262
122521 98.852 -78.669 45.246
122522 98.891 -78.670 45.230
122523 98.929 -78.673 45.215
122524 98.962 -78.685 45.200
122525 98.996 -78.695 45.186
122526 99.029 -78.707 45.171
122527 99.020 -78.713 45.157
122528 98.991 -78.717 45.143
122529 98.991 -78.714 45.130
122530 99.007 -78.715 45.118
122531 99.018 -78.695 45.106
122532 99.041 -78.652 45.094
122533 99.088 -78.646 45.082
122534 99.125 -78.646 45.070
122535 99.093 -78.677 45.058
122536 99.095 -78.674 45.045
122537 99.107 -78.695 45.033
122538 99.120 -78.732 45.022
122539 99.134 -78.769 45.010
122540 99.184 -78.802 44.999
122541 99.221 -78.814 44.989
122542 99.260 -78.798 44.978
122543 99.292 -78.792 44.968
122544 99.325 -78.786 44.957
122545 99.357 -78.780 44.947
122546 99.390 -78.774 44.937
122547 99.424 -78.771 44.925
122548 99.479 -78.789 44.915
122549 99.534 -78.808 44.906
122550 99.584 -78.825 44.895
122551 99.591 -78.824 44.885
122552 99.597 -78.823 44.875
122553 99.615 -78.841 44.865
122554 99.646 -78.884 44.853
122555 99.678 -78.927 44.843
122556 99.710 -78.971 44.833
122557 99.734 -79.002 44.823
122558 99.755 -79.024 44.813
122559 99.775 -79.047 44.804
122560 99.795 -79.070 44.796
122561 99.815 -79.093 44.788
122562 99.835 -79.115 44.781
122563 99.855 -79.138 44.775
122564 99.875 -79.161 44.769
122565 99.895 -79.184 44.763
122566 99.918 -79.206 44.758
122567 99.918 -79.231 44.754
122568 99.991 -79.209 44.750
122569 100.017 -79.243 44.746
122570 100.056 -79.263 44.743
122571 100.095 -79.276 44.739
122572 100.118 -79.294 44.735
122573 100.124 -79.319 44.730
122574 100.132 -79.344 44.725
122575 100.161 -79.362 44.721
122576 100.190 -79.372 44.715
122577 100.217 -79.378 44.709
122578 100.244 -79.383 44.703
122579 100.271 -79.388 44.698
122580 100.298 -79.394 44.691
122581 100.326 -79.400 44.685
122582 100.353 -79.405 44.678
122583 100.381 -79.410 44.671
122584 100.408 -79.416 44.664
122585 100.435 -79.422 44.657
122586 100.467 -79.429 44.649
122587 100.505 -79.439 44.641
122588 100.544 -79.450 44.633
122589 100.582 -79.460 44.624
122590 100.621 -79.470 44.615
122591 100.659 -79.481 44.608
122592 100.698 -79.494 44.600
122593 100.738 -79.510 44.591
122594 100.776 -79.527 44.583
122595 100.816 -79.544 44.575
122596 100.855 -79.560 44.566
122597 100.894 -79.577 44.557
122598 100.933 -79.594 44.548
122599 100.973 -79.610 44.539
122600 101.011 -79.627 44.529
122601 101.051 -79.644 44.518
122602 101.090 -79.661 44.507
122603 101.129 -79.677 44.495
122604 101.169 -79.694 44.484
122605 101.208 -79.711 44.471
122606 101.247 -79.727 44.458
122607 101.282 -79.745 44.444
122608 101.315 -79.764 44.430
122609 101.347 -79.783 44.415
122610 101.379 -79.802 44.400
122611 101.445 -79.830 44.384
122612 101.517 -79.851 44.368
122613 101.553 -79.864 44.352
122614 101.579 -79.876 44.335
122615 101.596 -79.882 44.318
122616 101.612 -79.887 44.300
122617 101.629 -79.893 44.281
122618 101.646 -79.899 44.263
122619 101.667 -79.910 44.245
122620 101.694 -79.927 44.227
122621 101.721 -79.945 44.208
122622 101.747 -79.963 44.189
122948 94.320 -77.827 52.276
122949 94.345 -77.818 52.255
122950 94.369 -77.809 52.234
122951 94.395 -77.800 52.213
122952 94.419 -77.791 52.192
122953 94.445 -77.781 52.172
122954 94.469 -77.772 52.153
122955 94.484 -77.768 52.133
122956 94.497 -77.761 52.113
122957 94.511 -77.754 52.094
122958 94.525 -77.748 52.074
122959 94.539 -77.741 52.053
122960 94.553 -77.734 52.033
122961 94.568 -77.728 52.015
122962 94.582 -77.721 51.997
122963 94.596 -77.714 51.980
122964 94.610 -77.708 51.963
122965 94.624 -77.702 51.945
122966 94.638 -77.695 51.923
122967 94.653 -77.689 51.902
122968 94.657 -77.683 51.880
122969 94.657 -77.680 51.857
122970 94.656 -77.678 51.833
122971 94.653 -77.679 51.809
122972 94.649 -77.680 51.785
122973 94.646 -77.682 51.758
122974 94.651 -77.686 51.730
122975 94.665 -77.691 51.701
122976 94.678 -77.696 51.673
122977 94.691 -77.702 51.647
122978 94.700 -77.710 51.621
122979 94.709 -77.717 51.593
122980 94.725 -77.721 51.562
122981 94.741 -77.726 51.529
122982 94.757 -77.730 51.495
122983 94.772 -77.736 51.461
122984 94.788 -77.740 51.427
122985 94.809 -77.740 51.393
122986 94.825 -77.741 51.357
122987 94.841 -77.743 51.320
122988 94.837 -77.759 51.281
122989 94.837 -77.773 51.242
122990 94.858 -77.768 51.203
122991 94.879 -77.765 51.164
122992 94.891 -77.772 51.124
122993 94.902 -77.779 51.082
122994 94.924 -77.794 51.040
122995 94.952 -77.813 50.999
122996 94.979 -77.831 50.960
122997 94.995 -77.830 50.921
122998 95.011 -77.827 50.883
122999 95.027 -77.826 50.846
123000 95.041 -77.821 50.808
123001 95.051 -77.800 50.770
123002 95.060 -77.780 50.732
123003 95.070 -77.759 50.695
123004 95.082 -77.742 50.658
123005 95.098 -77.739 50.622
123006 95.114 -77.735 50.586
123007 95.130 -77.732 50.550
123008 95.147 -77.729 50.513
123009 95.164 -77.725 50.477
123010 95.180 -77.722 50.443
123011 95.196 -77.718 50.409
123012 95.214 -77.718 50.376
123013 95.234 -77.718 50.345
123014 95.251 -77.718 50.314
123015 95.257 -77.714 50.283
123016 95.273 -77.705 50.252
123017 95.299 -77.692 50.224
123018 95.312 -77.683 50.197
123019 95.319 -77.674 50.170
123020 95.327 -77.667 50.144
123021 95.334 -77.658 50.116
123022 95.342 -77.650 50.086
123023 95.354 -77.644 50.059
123024 95.369 -77.639 50.033
123025 95.391 -77.653 50.006
123026 95.415 -77.673 49.979
123027 95.440 -77.692 49.950
123028 95.464 -77.712 49.920
123029 95.494 -77.736 49.890
123030 95.498 -77.733 49.859
123031 95.577 -77.731 49.827
123032 95.639 -77.731 49.794
123033 95.621 -77.747 49.761
123034 95.602 -77.764 49.725
123035 95.584 -77.780 49.690
123036 95.563 -77.796 49.655
123037 95.561 -77.767 49.623
123038 95.576 -77.719 49.594
123039 95.620 -77.721 49.565
123040 95.665 -77.722 49.537
123041 95.709 -77.724 49.509
123042 95.732 -77.721 49.479
123043 95.720 -77.708 49.450
123044 95.722 -77.703 49.422
123045 95.724 -77.698 49.397
123046 95.729 -77.699 49.371
123047 95.736 -77.705 49.345
123048 95.743 -77.713 49.318
123049 95.751 -77.720 49.292
123050 95.758 -77.727 49.265
123051 95.766 -77.734 49.237
123052 95.773 -77.741 49.202
123053 95.782 -77.749 49.166
123054 95.800 -77.755 49.130
123055 95.817 -77.763 49.092
123056 95.834 -77.771 49.053
123057 95.852 -77.778 49.014
123058 95.869 -77.786 48.975
123059 95.886 -77.793 48.936
123060 95.897 -77.808 48.898
123061 95.911 -77.824 48.861
123062 95.932 -77.837 48.824
123063 95.953 -77.851 48.788
123064 95.971 -77.860 48.751
123065 95.989 -77.870 48.713
123066 96.007 -77.879 48.675
123067 96.026 -77.889 48.638
123068 96.045 -77.902 48.603
123069 96.056 -77.905 48.567
123070 96.067 -77.910 48.532
123071 96.078 -77.915 48.499
123072 96.084 -77.908 48.468
123073 96.089 -77.898 48.438
123074 96.101 -77.899 48.407
123075 96.115 -77.902 48.376
123076 96.128 -77.905 48.345
123077 96.136 -77.900 48.319
123078 96.143 -77.894 48.293
123079 96.150 -77.888 48.267
123080 96.157 -77.882 48.243
123081 96.163 -77.875 48.220
123082 96.170 -77.869 48.196
123083 96.181 -77.866 48.171
123084 96.199 -77.871 48.147
123085 96.231 -77.908 48.124
123086 96.261 -77.944 48.101
123087 96.293 -77.981 48.079
123088 96.315 -78.005 48.056
123089 96.318 -78.002 48.033
123090 96.321 -77.999 48.008
123091 96.324 -77.995 47.984
123092 96.327 -77.992 47.960
123093 96.343 -78.005 47.935
123094 96.360 -78.018 47.910
123095 96.376 -78.031 47.885
123096 96.392 -78.046 47.861
123097 96.395 -78.071 47.837
123098 96.399 -78.095 47.813
123099 96.402 -78.119 47.789
123100 96.406 -78.144 47.763
123101 96.409 -78.169 47.738
123102 96.417 -78.184 47.710
123103 96.424 -78.182 47.680
123104 96.436 -78.176 47.651
123105 96.453 -78.175 47.627
123106 96.469 -78.174 47.601
123107 96.483 -78.165 47.574
123108 96.496 -78.156 47.548
123109 96.510 -78.147 47.521
123110 96.523 -78.138 47.494
123111 96.536 -78.130 47.466
123112 96.549 -78.121 47.439
123113 96.564 -78.121 47.411
123114 96.581 -78.130 47.383
123115 96.594 -78.138 47.356
123116 96.606 -78.147 47.330
123117 96.618 -78.155 47.306
123118 96.623 -78.164 47.282
123119 96.626 -78.174 47.256
123120 96.629 -78.184 47.232
123121 96.632 -78.194 47.207
123122 96.649 -78.197 47.181
123123 96.670 -78.199 47.156
123124 96.690 -78.200 47.131
123125 96.705 -78.209 47.105
123126 96.719 -78.219 47.079
123127 96.733 -78.229 47.053
123128 96.745 -78.243 47.026
123129 96.755 -78.262 47.000
123130 96.764 -78.280 46.972
123131 96.781 -78.300 46.945
123132 96.801 -78.322 46.919
123133 96.821 -78.343 46.893
123134 96.845 -78.362 46.865
123135 96.871 -78.379 46.837
123136 96.897 -78.396 46.809
123137 96.923 -78.413 46.782
123138 96.952 -78.413 46.755
123139 96.979 -78.420 46.728
123140 97.006 -78.432 46.703
123141 97.033 -78.443 46.679
123142 97.060 -78.455 46.655
123143 97.087 -78.470 46.631
123144 97.114 -78.483 46.605
123145 97.140 -78.498 46.579
123146 97.156 -78.513 46.555
123147 97.171 -78.514 46.533
123148 97.220 -78.479 46.511
123149 97.283 -78.460 46.490
123150 97.346 -78.458 46.471
123151 97.391 -78.459 46.452
123152 97.429 -78.470 46.433
123153 97.471 -78.482 46.414
123154 97.513 -78.495 46.396
123155 97.556 -78.507 46.379
123156 97.598 -78.519 46.363
123157 97.640 -78.532 46.347
123158 97.682 -78.544 46.332
123159 97.727 -78.559 46.316
123160 97.787 -78.587 46.300
123161 97.847 -78.615 46.285
123162 97.881 -78.623 46.269
123163 97.908 -78.618 46.254
123164 97.935 -78.594 46.238
123165 97.946 -78.577 46.223
123166 97.957 -78.560 46.207
123167 97.967 -78.543 46.191
123168 98.003 -78.540 46.175
123169 98.054 -78.545 46.160
123170 98.103 -78.550 46.143
123171 98.154 -78.554 46.126
123172 98.187 -78.565 46.109
123173 98.198 -78.556 46.092
123174 98.191 -78.542 46.076
123175 98.182 -78.531 46.059
123176 98.173 -78.519 46.042
123177 98.164 -78.507 46.026
123178 98.168 -78.511 46.011
123179 98.179 -78.517 45.996
123180 98.192 -78.521 45.979
123181 98.205 -78.525 45.963
123182 98.212 -78.522 45.947
123183 98.210 -78.507 45.931
123184 98.216 -78.480 45.915
123185 98.234 -78.447 45.897
123186 98.252 -78.454 45.880
123187 98.263 -78.460 45.862
123188 98.285 -78.450 45.844
123189 98.306 -78.454 45.826
123190 98.331 -78.454 45.808
123191 98.327 -78.448 45.789
123192 98.335 -78.454 45.770
123193 98.343 -78.460 45.752
123194 98.352 -78.467 45.733
123195 98.364 -78.473 45.716
123196 98.376 -78.499 45.698
123197 98.392 -78.517 45.679
123198 98.409 -78.530 45.662
123199 98.426 -78.544 45.644
123200 98.443 -78.558 45.626
123201 98.444 -78.576 45.605
123202 98.444 -78.595 45.588
123203 98.444 -78.614 45.571
123204 98.453 -78.623 45.554
123205 98.463 -78.630 45.537
123206 98.476 -78.634 45.520
123207 98.501 -78.629 45.502
123208 98.527 -78.623 45.485
123209 98.553 -78.617 45.469
123210 98.581 -78.609 45.452
123211 98.613 -78.595 45.435
123212 98.645 -78.582 45.418
123213 98.676 -78.568 45.399
123214 98.698 -78.584 45.381
123215 98.721 -78.604 45.363
123216 98.744 -78.626 45.348
123217 98.767 -78.647 45.333
123218 98.768 -78.651 45.317
123219 98.786 -78.645 45.303
123220 98.816 -78.637 45.286
123221 98.829 -78.642 45.270
123222 98.840 -78.648 45.254
123223 98.867 -78.651 45.239
123224 98.907 -78.653 45.223
123225 98.947 -78.654 45.207
123226 98.984 -78.659 45.192
123227 99.017 -78.670 45.178
123228 99.017 -78.677 45.162
123229 99.014 -78.724 45.147
123230 99.038 -78.808 45.135
123231 99.041 -78.745 45.122
123232 99.035 -78.645 45.109
123233 99.048 -78.634 45.096
123234 99.111 -78.618 45.084
123235 99.174 -78.602 45.072
123236 99.124 -78.634 45.059
123237 99.102 -78.683 45.047
123238 99.110 -78.731 45.035
123239 99.116 -78.779 45.023
123240 99.125 -78.838 45.011
123241 99.166 -78.827 45.000
123242 99.209 -78.803 44.989
123243 99.249 -78.784 44.977
123244 99.282 -78.778 44.967
123245 99.314 -78.772 44.957
123246 99.347 -78.766 44.946
123247 99.379 -78.760 44.935
123248 99.423 -78.767 44.925
123249 99.478 -78.786 44.915
123250 99.533 -78.804 44.905
123251 99.587 -78.823 44.895
123252 99.608 -78.827 44.887
123253 99.614 -78.827 44.877
123254 99.628 -78.839 44.867
123255 99.660 -78.883 44.856
123256 99.692 -78.925 44.845
123257 99.723 -78.968 44.834
123258 99.746 -78.995 44.824
123259 99.766 -79.018 44.813
123260 99.786 -79.040 44.803
123261 99.805 -79.063 44.794
123262 99.826 -79.086 44.787
123263 99.846 -79.109 44.780
123264 99.866 -79.132 44.773
123265 99.886 -79.155 44.767
123266 99.906 -79.178 44.761
123267 99.926 -79.200 44.756
123268 99.946 -79.203 44.752
123269 99.977 -79.216 44.748
123270 100.017 -79.228 44.744
123271 100.056 -79.241 44.740
123272 100.096 -79.253 44.736
123273 100.109 -79.276 44.732
123274 100.115 -79.300 44.727
123275 100.134 -79.322 44.722
123276 100.163 -79.340 44.718
123277 100.192 -79.352 44.713
123278 100.219 -79.358 44.708
123279 100.247 -79.363 44.703
123280 100.274 -79.369 44.698
123281 100.301 -79.374 44.692
123282 100.329 -79.380 44.686
123283 100.356 -79.385 44.679
123284 100.383 -79.391 44.673
123285 100.410 -79.396 44.666
123286 100.438 -79.402 44.659
123287 100.474 -79.411 44.651
123288 100.512 -79.422 44.643
123289 100.551 -79.432 44.635
123290 100.589 -79.442 44.626
123291 100.628 -79.452 44.618
123292 100.666 -79.463 44.610
123293 100.705 -79.480 44.602
123294 100.745 -79.497 44.593
123295 100.784 -79.513 44.585
123296 100.823 -79.530 44.576
123297 100.862 -79.547 44.567
123298 100.901 -79.563 44.559
123299 100.940 -79.580 44.550
123300 100.980 -79.597 44.540
123301 101.019 -79.614 44.530
123302 101.058 -79.630 44.519
123303 101.098 -79.647 44.508
123304 101.137 -79.664 44.496
123305 101.176 -79.680 44.483
123306 101.215 -79.697 44.470
123307 101.247 -79.716 44.457
123308 101.280 -79.735 44.443
123309 101.312 -79.755 44.428
123310 101.345 -79.773 44.413
123311 101.377 -79.792 44.398
123312 101.470 -79.828 44.382
123313 101.522 -79.835 44.365
123314 101.551 -79.851 44.349
123315 101.580 -79.863 44.332
123316 101.596 -79.868 44.315
123317 101.613 -79.874 44.297
123318 101.630 -79.880 44.279
123319 101.646 -79.886 44.261
123320 101.664 -79.891 44.243
123321 101.681 -79.898 44.226
123322 101.707 -79.915 44.206
123323 101.734 -79.933 44.188
123619 93.982 -77.875 52.947
123620 93.980 -77.874 52.928
123621 93.978 -77.873 52.907
123622 93.975 -77.872 52.887
123623 93.973 -77.871 52.867
123624 93.970 -77.870 52.846
123625 93.968 -77.868 52.825
123626 93.967 -77.871 52.804
123627 93.971 -77.882 52.782
123628 93.974 -77.892 52.760
123629 93.977 -77.902 52.739
123630 93.981 -77.912 52.717
123631 93.992 -77.919 52.695
123648 94.296 -77.820 52.311
123649 94.321 -77.811 52.291
123650 94.346 -77.802 52.270
123651 94.371 -77.793 52.249
123652 94.396 -77.783 52.229
123653 94.421 -77.774 52.208
123654 94.446 -77.765 52.188
123655 94.471 -77.755 52.169
123656 94.495 -77.747 52.150
123657 94.506 -77.743 52.130
123658 94.520 -77.735 52.111
123659 94.534 -77.728 52.092
123660 94.548 -77.722 52.072
123661 94.563 -77.715 52.052
123662 94.577 -77.708 52.033
123663 94.591 -77.702 52.016
123664 94.605 -77.696 52.001
123665 94.619 -77.689 51.983
123666 94.633 -77.683 51.964
123667 94.648 -77.676 51.941
123668 94.662 -77.669 51.919
123669 94.669 -77.664 51.897
123670 94.667 -77.664 51.874
123671 94.664 -77.665 51.850
123672 94.660 -77.667 51.826
123673 94.657 -77.668 51.802
123674 94.653 -77.670 51.775
123675 94.654 -77.672 51.746
123676 94.664 -77.679 51.717
123677 94.673 -77.686 51.689
123678 94.682 -77.693 51.663
123679 94.691 -77.701 51.638
123680 94.705 -77.706 51.611
123681 94.720 -77.711 51.579
123682 94.736 -77.715 51.547
123683 94.752 -77.721 51.512
123684 94.774 -77.727 51.476
123685 94.793 -77.732 51.442
123686 94.802 -77.720 51.407
123687 94.828 -77.720 51.371
123688 94.854 -77.720 51.333
123689 94.856 -77.738 51.294
123690 94.849 -77.756 51.256
123691 94.842 -77.774 51.217
123692 94.843 -77.787 51.178
123693 94.868 -77.781 51.137
123694 94.894 -77.774 51.096
123695 94.913 -77.771 51.054
123696 94.956 -77.806 51.014
123697 94.984 -77.826 50.975
123698 95.004 -77.833 50.936
123699 95.020 -77.830 50.898
123700 95.035 -77.825 50.860
123701 95.044 -77.805 50.822
123702 95.054 -77.784 50.784
123703 95.064 -77.764 50.746
123704 95.074 -77.743 50.709
123705 95.089 -77.736 50.672
123706 95.105 -77.732 50.635
123707 95.121 -77.729 50.599
123708 95.137 -77.726 50.562
123709 95.154 -77.722 50.525
123710 95.170 -77.719 50.489
123711 95.188 -77.718 50.454
123712 95.208 -77.718 50.420
123713 95.227 -77.718 50.387
123714 95.245 -77.718 50.355
123715 95.250 -77.714 50.323
123716 95.256 -77.710 50.292
123717 95.281 -77.698 50.261
123718 95.306 -77.685 50.232
123719 95.323 -77.674 50.203
123720 95.330 -77.666 50.176
123721 95.338 -77.658 50.150
123722 95.345 -77.650 50.122
123723 95.353 -77.642 50.092
123724 95.360 -77.633 50.063
123725 95.368 -77.625 50.035
123726 95.387 -77.635 50.007
123727 95.411 -77.655 49.979
123728 95.436 -77.674 49.950
123729 95.460 -77.694 49.921
123730 95.484 -77.714 49.892
123731 95.523 -77.727 49.861
123732 95.631 -77.705 49.831
123733 95.635 -77.708 49.800
123734 95.626 -77.708 49.770
123735 95.617 -77.708 49.735
123736 95.606 -77.711 49.701
123737 95.592 -77.716 49.667
123738 95.578 -77.721 49.634
123739 95.610 -77.705 49.604
123740 95.633 -77.702 49.576
123741 95.656 -77.699 49.547
123742 95.679 -77.696 49.518
123743 95.702 -77.693 49.489
123744 95.721 -77.689 49.461
123745 95.711 -77.680 49.434
123746 95.716 -77.680 49.408
123747 95.724 -77.688 49.382
123748 95.731 -77.695 49.355
123749 95.738 -77.702 49.327
123750 95.746 -77.709 49.300
123751 95.753 -77.716 49.273
123752 95.760 -77.724 49.244
123753 95.768 -77.730 49.210
123754 95.777 -77.737 49.174
123755 95.795 -77.745 49.137
123756 95.812 -77.752 49.099
123757 95.829 -77.760 49.062
123758 95.846 -77.768 49.023
123759 95.864 -77.774 48.984
123760 95.881 -77.782 48.945
123761 95.898 -77.790 48.907
123762 95.913 -77.805 48.871
123763 95.933 -77.816 48.835
123764 95.951 -77.826 48.798
123765 95.971 -77.838 48.762
123766 95.990 -77.851 48.725
123767 96.011 -77.863 48.688
123768 96.031 -77.876 48.652
123769 96.051 -77.888 48.615
123770 96.066 -77.896 48.578
123771 96.073 -77.893 48.544
123772 96.082 -77.895 48.510
123773 96.093 -77.900 48.478
123774 96.100 -77.894 48.447
123775 96.107 -77.888 48.417
123776 96.121 -77.891 48.388
123777 96.133 -77.893 48.357
123778 96.140 -77.887 48.329
123779 96.147 -77.881 48.302
123780 96.154 -77.874 48.275
123781 96.161 -77.868 48.251
123782 96.168 -77.862 48.228
123783 96.174 -77.855 48.205
123784 96.181 -77.849 48.181
123785 96.205 -77.869 48.157
123786 96.236 -77.906 48.133
123787 96.267 -77.942 48.109
123788 96.299 -77.979 48.086
123789 96.327 -78.012 48.064
123790 96.331 -78.009 48.042
123791 96.334 -78.006 48.019
123792 96.337 -78.003 47.996
123793 96.340 -78.000 47.972
123794 96.343 -77.996 47.948
123795 96.359 -78.009 47.923
123796 96.375 -78.022 47.898
123797 96.392 -78.036 47.874
123798 96.397 -78.059 47.851
123799 96.400 -78.084 47.828
123800 96.404 -78.108 47.803
123801 96.408 -78.132 47.777
123802 96.414 -78.149 47.750
123803 96.425 -78.159 47.723
123804 96.433 -78.165 47.695
123805 96.436 -78.149 47.667
123806 96.446 -78.141 47.643
123807 96.463 -78.140 47.617
123808 96.480 -78.139 47.590
123809 96.497 -78.138 47.563
123810 96.514 -78.138 47.536
123811 96.530 -78.134 47.508
123812 96.544 -78.126 47.480
123813 96.557 -78.117 47.453
123814 96.571 -78.109 47.425
123815 96.586 -78.118 47.398
123816 96.599 -78.126 47.373
123817 96.611 -78.134 47.347
123818 96.623 -78.142 47.323
123819 96.630 -78.151 47.299
123820 96.633 -78.161 47.273
123821 96.637 -78.171 47.249
123822 96.641 -78.180 47.224
123823 96.661 -78.181 47.200
123824 96.681 -78.183 47.174
123825 96.699 -78.189 47.149
123826 96.713 -78.200 47.123
123827 96.726 -78.210 47.097
123828 96.740 -78.220 47.071
123829 96.753 -78.231 47.045
123830 96.763 -78.250 47.018
123831 96.772 -78.268 46.991
123832 96.783 -78.287 46.965
123833 96.798 -78.307 46.938
123834 96.818 -78.329 46.912
123835 96.842 -78.347 46.884
123836 96.868 -78.364 46.857
123837 96.895 -78.376 46.830
123838 96.923 -78.381 46.802
123839 96.950 -78.392 46.775
123840 96.977 -78.404 46.748
123841 97.004 -78.416 46.721
123842 97.030 -78.427 46.698
123843 97.058 -78.439 46.673
123844 97.084 -78.454 46.648
123845 97.111 -78.467 46.623
123846 97.138 -78.482 46.599
123847 97.165 -78.496 46.575
123848 97.185 -78.514 46.552
123849 97.234 -78.476 46.530
123850 97.276 -78.459 46.509
123851 97.332 -78.452 46.488
123852 97.394 -78.450 46.469
123853 97.426 -78.453 46.450
123854 97.454 -78.460 46.431
123855 97.496 -78.473 46.413
123856 97.538 -78.485 46.396
123857 97.581 -78.497 46.379
123858 97.623 -78.510 46.363
123859 97.665 -78.522 46.347
123860 97.707 -78.534 46.331
123861 97.760 -78.556 46.315
123862 97.820 -78.584 46.299
123863 97.880 -78.613 46.283
123864 97.925 -78.629 46.267
123865 97.947 -78.612 46.250
123866 97.958 -78.595 46.234
123867 97.968 -78.578 46.218
123868 97.980 -78.561 46.202
123869 97.991 -78.545 46.186
123870 98.032 -78.545 46.171
123871 98.082 -78.550 46.154
123872 98.133 -78.554 46.136
123873 98.182 -78.557 46.120
123874 98.212 -78.555 46.104
123875 98.203 -78.544 46.088
123876 98.195 -78.532 46.072
123877 98.185 -78.520 46.055
123878 98.177 -78.509 46.041
123879 98.178 -78.510 46.026
123880 98.190 -78.515 46.010
123881 98.203 -78.519 45.994
123882 98.217 -78.523 45.978
123883 98.230 -78.527 45.963
123884 98.233 -78.519 45.947
123885 98.243 -78.485 45.930
123886 98.260 -78.444 45.913
123887 98.267 -78.445 45.895
123888 98.271 -78.457 45.878
123889 98.291 -78.451 45.860
123890 98.313 -78.442 45.841
123891 98.334 -78.433 45.822
123892 98.345 -78.438 45.803
123893 98.353 -78.445 45.783
123894 98.361 -78.451 45.764
123895 98.371 -78.458 45.746
123896 98.383 -78.465 45.728
123897 98.395 -78.480 45.709
123898 98.408 -78.512 45.690
123899 98.424 -78.529 45.672
123900 98.441 -78.542 45.653
123901 98.453 -78.557 45.635
123902 98.452 -78.576 45.615
123903 98.455 -78.592 45.597
123904 98.465 -78.598 45.580
123905 98.476 -78.605 45.562
123906 98.486 -78.612 45.545
123907 98.497 -78.619 45.527
123908 98.508 -78.624 45.509
123909 98.534 -78.619 45.492
123910 98.560 -78.614 45.476
123911 98.588 -78.604 45.459
123912 98.620 -78.591 45.442
123913 98.651 -78.577 45.425
123914 98.675 -78.564 45.407
123915 98.710 -78.567 45.389
123916 98.737 -78.586 45.371
123917 98.760 -78.608 45.356
123918 98.783 -78.629 45.341
123919 98.775 -78.627 45.326
123920 98.785 -78.620 45.312
123921 98.812 -78.614 45.296
123922 98.823 -78.621 45.279
123923 98.835 -78.627 45.263
123924 98.846 -78.633 45.248
123925 98.883 -78.635 45.232
123926 98.923 -78.636 45.216
123927 98.962 -78.636 45.200
123928 99.002 -78.638 45.185
123929 99.028 -78.665 45.168
123930 99.043 -78.737 45.152
123931 99.043 -78.682 45.140
123932 99.038 -78.590 45.126
123933 99.025 -78.602 45.112
123934 99.057 -78.620 45.098
123935 99.094 -78.617 45.086
123936 99.147 -78.608 45.073
123937 99.186 -78.604 45.060
123938 99.119 -78.651 45.049
123939 99.118 -78.717 45.037
123940 99.118 -78.783 45.026
123941 99.114 -78.838 45.014
123942 99.153 -78.816 45.002
123943 99.196 -78.793 44.991
123944 99.238 -78.770 44.980
123945 99.271 -78.764 44.970
123946 99.304 -78.758 44.959
123947 99.336 -78.752 44.948
123948 99.369 -78.746 44.938
123949 99.422 -78.763 44.928
123950 99.476 -78.782 44.917
123951 99.531 -78.801 44.907
123952 99.585 -78.819 44.896
123953 99.625 -78.831 44.887
123954 99.631 -78.830 44.878
123955 99.642 -78.837 44.868
123956 99.674 -78.880 44.857
123957 99.705 -78.924 44.846
123958 99.736 -78.965 44.835
123959 99.756 -78.988 44.823
123960 99.776 -79.011 44.812
123961 99.796 -79.034 44.803
123962 99.817 -79.057 44.794
123963 99.837 -79.080 44.786
123964 99.857 -79.103 44.779
123965 99.877 -79.125 44.772
123966 99.897 -79.148 44.766
123967 99.917 -79.171 44.760
123968 99.933 -79.187 44.755
123969 99.938 -79.180 44.750
123970 99.978 -79.193 44.746
123971 100.018 -79.206 44.742
123972 100.057 -79.219 44.738
123973 100.093 -79.233 44.734
123974 100.099 -79.257 44.730
123975 100.107 -79.281 44.726
123976 100.136 -79.300 44.721
123977 100.166 -79.318 44.717
123978 100.195 -79.332 44.713
123979 100.222 -79.338 44.708
123980 100.249 -79.344 44.703
123981 100.276 -79.349 44.697
123982 100.304 -79.354 44.692
123983 100.331 -79.360 44.687
123984 100.358 -79.366 44.681
123985 100.385 -79.371 44.674
123986 100.413 -79.376 44.668
123987 100.442 -79.382 44.660
123988 100.481 -79.393 44.653
123989 100.519 -79.404 44.645
123990 100.558 -79.413 44.637
123991 100.596 -79.424 44.630
123992 100.635 -79.435 44.621
123993 100.674 -79.450 44.612
123994 100.713 -79.466 44.604
123995 100.752 -79.483 44.595
123996 100.791 -79.500 44.587
123997 100.830 -79.516 44.578
123998 100.869 -79.533 44.569
123999 100.909 -79.550 44.560
124000 100.948 -79.567 44.551
124001 100.987 -79.583 44.541
124002 101.027 -79.600 44.530
124003 101.065 -79.617 44.519
124004 101.105 -79.633 44.508
124005 101.144 -79.650 44.495
124006 101.180 -79.667 44.482
124007 101.212 -79.687 44.469
124008 101.245 -79.706 44.455
124009 101.277 -79.725 44.441
124010 101.310 -79.744 44.426
124011 101.342 -79.764 44.411
124012 101.401 -79.789 44.396
124013 101.475 -79.811 44.379
124014 101.527 -79.818 44.362
124015 101.550 -79.837 44.345
124016 101.581 -79.850 44.328
124017 101.597 -79.855 44.312
124018 101.614 -79.861 44.294
124019 101.631 -79.867 44.276
124020 101.647 -79.872 44.258
124021 101.664 -79.878 44.241
124022 101.681 -79.883 44.223
124023 101.698 -79.889 44.205
124024 101.721 -79.902 44.186
124317 93.944 -77.877 53.018
124318 93.958 -77.873 52.998
124319 93.970 -77.868 52.978
124320 93.983 -77.864 52.958
124321 93.980 -77.863 52.938
124322 93.978 -77.862 52.918
124323 93.976 -77.861 52.898
124324 93.973 -77.860 52.878
124325 93.971 -77.859 52.857
124326 93.968 -77.858 52.837
124327 93.967 -77.859 52.816
124328 93.970 -77.868 52.793
124329 93.973 -77.879 52.771
124330 93.977 -77.889 52.750
124331 93.980 -77.899 52.729
124332 93.995 -77.904 52.707
124333 94.011 -77.909 52.685
124334 94.027 -77.912 52.663
124335 94.043 -77.917 52.639
124336 94.059 -77.921 52.617
124337 94.062 -77.915 52.595
124338 94.063 -77.906 52.571
124348 94.273 -77.812 52.347
124349 94.297 -77.803 52.327
124350 94.323 -77.794 52.306
124351 94.348 -77.784 52.285
124352 94.372 -77.775 52.265
124353 94.398 -77.766 52.245
124354 94.422 -77.757 52.225
124355 94.448 -77.748 52.205
124356 94.472 -77.739 52.186
124357 94.497 -77.730 52.167
124358 94.516 -77.724 52.149
124359 94.529 -77.716 52.130
124360 94.543 -77.709 52.111
124361 94.558 -77.703 52.092
124362 94.572 -77.696 52.073
124363 94.586 -77.689 52.055
124364 94.600 -77.683 52.037
124365 94.614 -77.677 52.019
124366 94.628 -77.670 52.001
124367 94.643 -77.664 51.981
124368 94.663 -77.655 51.958
124369 94.684 -77.646 51.936
124370 94.669 -77.651 51.914
124371 94.666 -77.653 51.892
124372 94.662 -77.655 51.868
124373 94.657 -77.656 51.844
124374 94.653 -77.658 51.821
124375 94.649 -77.659 51.793
124376 94.649 -77.664 51.765
124377 94.654 -77.670 51.736
124378 94.664 -77.677 51.708
124379 94.672 -77.685 51.682
124380 94.684 -77.691 51.655
124381 94.700 -77.696 51.627
124382 94.716 -77.701 51.595
124383 94.739 -77.708 51.560
124384 94.761 -77.714 51.525
124385 94.782 -77.721 51.489
124386 94.792 -77.716 51.454
124387 94.793 -77.700 51.419
124388 94.819 -77.700 51.383
124389 94.846 -77.701 51.345
124390 94.873 -77.732 51.307
124391 94.873 -77.752 51.269
124392 94.836 -77.767 51.230
124393 94.855 -77.778 51.191
124394 94.866 -77.771 51.151
124395 94.891 -77.765 51.111
124396 94.912 -77.761 51.071
124397 94.924 -77.759 51.031
124398 94.981 -77.812 50.992
124399 95.014 -77.835 50.953
124400 95.028 -77.830 50.914
124401 95.038 -77.809 50.876
124402 95.048 -77.789 50.838
124403 95.057 -77.768 50.799
124404 95.067 -77.748 50.761
124405 95.079 -77.733 50.723
124406 95.096 -77.729 50.686
124407 95.112 -77.726 50.649
124408 95.128 -77.722 50.612
124409 95.144 -77.719 50.575
124410 95.162 -77.717 50.537
124411 95.181 -77.717 50.501
124412 95.201 -77.718 50.466
124413 95.220 -77.718 50.432
124414 95.238 -77.718 50.398
124415 95.244 -77.714 50.365
124416 95.249 -77.710 50.333
124417 95.263 -77.702 50.301
124418 95.288 -77.690 50.270
124419 95.314 -77.677 50.240
124420 95.334 -77.665 50.210
124421 95.341 -77.658 50.182
124422 95.349 -77.649 50.155
124423 95.356 -77.641 50.128
124424 95.364 -77.633 50.099
124425 95.372 -77.625 50.068
124426 95.379 -77.617 50.038
124427 95.420 -77.592 50.009
124428 95.453 -77.603 49.979
124429 95.483 -77.617 49.951
124430 95.513 -77.632 49.923
124431 95.543 -77.646 49.895
124432 95.570 -77.661 49.865
124433 95.605 -77.671 49.836
124434 95.618 -77.677 49.807
124435 95.609 -77.677 49.778
124436 95.600 -77.677 49.745
124437 95.591 -77.677 49.712
124438 95.582 -77.677 49.678
124439 95.596 -77.680 49.645
124440 95.608 -77.688 49.615
124441 95.624 -77.683 49.585
124442 95.640 -77.677 49.556
124443 95.657 -77.671 49.527
124444 95.673 -77.666 49.498
124445 95.694 -77.662 49.471
124446 95.703 -77.663 49.444
124447 95.711 -77.670 49.419
124448 95.719 -77.677 49.392
124449 95.726 -77.684 49.365
124450 95.733 -77.691 49.337
124451 95.741 -77.699 49.310
124452 95.748 -77.705 49.282
124453 95.755 -77.712 49.252
124454 95.763 -77.720 49.218
124455 95.772 -77.727 49.182
124456 95.789 -77.734 49.145
124457 95.807 -77.742 49.107
124458 95.824 -77.749 49.069
124459 95.841 -77.756 49.031
124460 95.859 -77.764 48.993
124461 95.876 -77.771 48.955
124462 95.893 -77.779 48.917
124463 95.915 -77.788 48.881
124464 95.936 -77.800 48.845
124465 95.956 -77.812 48.808
124466 95.976 -77.825 48.772
124467 95.996 -77.837 48.736
124468 96.016 -77.850 48.700
124469 96.036 -77.862 48.663
124470 96.056 -77.875 48.627
124471 96.077 -77.887 48.589
124472 96.085 -77.887 48.555
124473 96.092 -77.884 48.521
124474 96.098 -77.881 48.488
124475 96.108 -77.885 48.457
124476 96.115 -77.880 48.427
124477 96.127 -77.881 48.398
124478 96.137 -77.879 48.368
124479 96.144 -77.873 48.340
124480 96.151 -77.866 48.312
124481 96.158 -77.860 48.285
124482 96.165 -77.854 48.260
124483 96.172 -77.848 48.235
124484 96.178 -77.846 48.212
124485 96.185 -77.845 48.188
124486 96.212 -77.869 48.165
124487 96.242 -77.903 48.141
124488 96.273 -77.940 48.118
124489 96.304 -77.977 48.094
124490 96.335 -78.013 48.072
124491 96.343 -78.016 48.049
124492 96.346 -78.013 48.027
124493 96.349 -78.010 48.005
124494 96.353 -78.006 47.982
124495 96.356 -78.003 47.958
124496 96.359 -78.000 47.933
124497 96.375 -78.013 47.909
124498 96.391 -78.026 47.887
124499 96.399 -78.047 47.865
124500 96.402 -78.072 47.842
124501 96.406 -78.097 47.817
124502 96.412 -78.114 47.791
124503 96.422 -78.124 47.763
124504 96.432 -78.134 47.736
124505 96.442 -78.144 47.709
124506 96.446 -78.131 47.683
124507 96.449 -78.116 47.658
124508 96.456 -78.104 47.632
124509 96.473 -78.103 47.605
124510 96.490 -78.103 47.578
124511 96.507 -78.103 47.550
124512 96.524 -78.102 47.521
124513 96.541 -78.101 47.493
124514 96.558 -78.100 47.466
124515 96.576 -78.100 47.439
124516 96.591 -78.105 47.414
124517 96.603 -78.113 47.389
124518 96.615 -78.122 47.365
124519 96.627 -78.129 47.340
124520 96.637 -78.138 47.315
124521 96.641 -78.148 47.290
124522 96.644 -78.158 47.266
124523 96.653 -78.165 47.241
124524 96.673 -78.166 47.217
124525 96.692 -78.169 47.192
124526 96.706 -78.180 47.167
124527 96.720 -78.190 47.142
124528 96.733 -78.200 47.116
124529 96.747 -78.210 47.090
124530 96.761 -78.221 47.063
124531 96.771 -78.238 47.036
124532 96.781 -78.257 47.010
124533 96.791 -78.275 46.984
124534 96.800 -78.294 46.957
124535 96.815 -78.314 46.930
124536 96.839 -78.332 46.903
124537 96.867 -78.341 46.877
124538 96.894 -78.353 46.850
124539 96.921 -78.364 46.822
124540 96.947 -78.376 46.795
124541 96.974 -78.388 46.768
124542 97.001 -78.399 46.739
124543 97.028 -78.411 46.715
124544 97.054 -78.424 46.691
124545 97.082 -78.438 46.666
124546 97.108 -78.452 46.642
124547 97.136 -78.466 46.620
124548 97.163 -78.480 46.597
124549 97.195 -78.492 46.574
124550 97.237 -78.481 46.552
124551 97.279 -78.464 46.529
124552 97.320 -78.448 46.507
124553 97.381 -78.444 46.485
124554 97.438 -78.443 46.467
124555 97.461 -78.448 46.448
124556 97.484 -78.452 46.429
124557 97.521 -78.463 46.413
124558 97.563 -78.475 46.396
124559 97.605 -78.487 46.379
124560 97.648 -78.500 46.363
124561 97.690 -78.512 46.347
124562 97.733 -78.526 46.331
124563 97.793 -78.554 46.314
124564 97.853 -78.582 46.297
124565 97.913 -78.610 46.281
124566 97.949 -78.620 46.264
124567 97.962 -78.605 46.247
124568 97.975 -78.590 46.231
124569 97.988 -78.576 46.214
124570 98.000 -78.561 46.197
124571 98.013 -78.546 46.181
124572 98.059 -78.544 46.164
124573 98.107 -78.543 46.147
124574 98.156 -78.540 46.131
124575 98.205 -78.516 46.116
124576 98.209 -78.528 46.100
124577 98.207 -78.534 46.085
124578 98.198 -78.522 46.071
124579 98.190 -78.510 46.055
124580 98.189 -78.509 46.040
124581 98.202 -78.514 46.025
124582 98.215 -78.517 46.010
124583 98.228 -78.521 45.994
124584 98.241 -78.526 45.979
124585 98.253 -78.529 45.964
124586 98.265 -78.493 45.947
124587 98.276 -78.453 45.929
124588 98.282 -78.438 45.912
124589 98.285 -78.448 45.894
124590 98.296 -78.452 45.876
124591 98.319 -78.435 45.858
124592 98.339 -78.431 45.838
124593 98.357 -78.432 45.819
124594 98.371 -78.436 45.798
124595 98.379 -78.443 45.778
124596 98.391 -78.449 45.759
124597 98.402 -78.467 45.740
124598 98.415 -78.480 45.721
124599 98.427 -78.493 45.702
124600 98.439 -78.525 45.682
124601 98.455 -78.540 45.663
124602 98.461 -78.557 45.645
124603 98.468 -78.568 45.625
124604 98.479 -78.574 45.607
124605 98.489 -78.581 45.589
124606 98.499 -78.588 45.571
124607 98.509 -78.595 45.553
124608 98.520 -78.601 45.535
124609 98.530 -78.608 45.517
124610 98.541 -78.614 45.499
124611 98.567 -78.609 45.482
124612 98.594 -78.601 45.465
124613 98.626 -78.586 45.448
124614 98.647 -78.574 45.431
124615 98.661 -78.562 45.413
124616 98.691 -78.556 45.396
124617 98.739 -78.571 45.380
124618 98.777 -78.589 45.365
124619 98.796 -78.608 45.350
124620 98.783 -78.602 45.335
124621 98.784 -78.596 45.321
124622 98.807 -78.593 45.306
124623 98.818 -78.599 45.289
124624 98.830 -78.605 45.272
124625 98.841 -78.612 45.256
124626 98.859 -78.617 45.240
124627 98.899 -78.617 45.224
124628 98.938 -78.619 45.208
124629 98.978 -78.620 45.192
124630 99.039 -78.643 45.176
124631 99.037 -78.583 45.159
124632 99.019 -78.593 45.146
124633 99.001 -78.602 45.131
124634 99.020 -78.623 45.118
124635 99.064 -78.612 45.105
124636 99.083 -78.617 45.092
124637 99.120 -78.615 45.079
124638 99.174 -78.605 45.067
124639 99.171 -78.619 45.054
124640 99.138 -78.675 45.043
124641 99.138 -78.741 45.032
124642 99.136 -78.805 45.020
124643 99.141 -78.805 45.008
124644 99.184 -78.782 44.997
124645 99.227 -78.758 44.986
124646 99.261 -78.750 44.975
124647 99.293 -78.744 44.964
124648 99.326 -78.739 44.953
124649 99.365 -78.741 44.942
124650 99.420 -78.759 44.931
124651 99.475 -78.778 44.920
124652 99.529 -78.797 44.909
124653 99.584 -78.815 44.898
124654 99.639 -78.834 44.888
124655 99.648 -78.834 44.878
124656 99.655 -78.835 44.868
124657 99.687 -78.878 44.858
124658 99.719 -78.921 44.847
124659 99.747 -78.959 44.835
124660 99.767 -78.982 44.822
124661 99.787 -79.005 44.812
124662 99.807 -79.028 44.803
124663 99.828 -79.050 44.793
124664 99.848 -79.073 44.785
124665 99.867 -79.096 44.779
124666 99.888 -79.118 44.772
124667 99.908 -79.141 44.765
124668 99.922 -79.153 44.760
124669 99.935 -79.163 44.755
124670 99.941 -79.159 44.750
124671 99.979 -79.171 44.745
124672 100.018 -79.183 44.741
124673 100.058 -79.196 44.737
124674 100.083 -79.214 44.733
124675 100.090 -79.238 44.729
124676 100.109 -79.259 44.725
124677 100.139 -79.278 44.721
124678 100.168 -79.296 44.717
124679 100.197 -79.313 44.713
124680 100.224 -79.318 44.708
124681 100.252 -79.323 44.704
124682 100.279 -79.329 44.698
124683 100.307 -79.335 44.693
124684 100.334 -79.340 44.688
124685 100.361 -79.345 44.682
124686 100.388 -79.351 44.676
124687 100.416 -79.357 44.669
124688 100.449 -79.365 44.662
124689 100.487 -79.375 44.655
124690 100.526 -79.385 44.647
124691 100.564 -79.395 44.639
124692 100.603 -79.406 44.631
124693 100.642 -79.419 44.623
124694 100.681 -79.436 44.614
124695 100.720 -79.453 44.606
124696 100.759 -79.469 44.597
124697 100.798 -79.486 44.588
124698 100.838 -79.503 44.579
124699 100.877 -79.520 44.571
124700 100.916 -79.536 44.560
124701 100.956 -79.553 44.550
124702 100.994 -79.570 44.540
124703 101.034 -79.586 44.530
124704 101.073 -79.603 44.518
124705 101.112 -79.620 44.506
124706 101.144 -79.639 44.494
124707 101.177 -79.657 44.480
124708 101.210 -79.676 44.466
124709 101.242 -79.696 44.452
124710 101.275 -79.715 44.438
124711 101.307 -79.734 44.423
124712 101.340 -79.753 44.408
124713 101.426 -79.787 44.392
124714 101.481 -79.795 44.375
124715 101.526 -79.803 44.358
124716 101.550 -79.823 44.342
124717 101.581 -79.836 44.325
124718 101.598 -79.842 44.308
124719 101.615 -79.848 44.291
124720 101.631 -79.854 44.273
124721 101.648 -79.859 44.255
124722 101.665 -79.864 44.238
124723 101.682 -79.871 44.220
124724 101.698 -79.876 44.202
124725 101.715 -79.882 44.184
125015 93.907 -77.878 53.087
125016 93.919 -77.874 53.067
125017 93.932 -77.870 53.048
125018 93.945 -77.866 53.029
125019 93.958 -77.862 53.009
125020 93.971 -77.857 52.989
125021 93.983 -77.853 52.969
125022 93.981 -77.852 52.949
125023 93.978 -77.851 52.930
125024 93.976 -77.849 52.910
125025 93.974 -77.849 52.890
125026 93.971 -77.847 52.870
125027 93.969 -77.846 52.849
125028 93.966 -77.845 52.828
125029 93.969 -77.855 52.806
125030 93.973 -77.865 52.785
125031 93.976 -77.876 52.764
125032 93.983 -77.884 52.743
125033 93.998 -77.889 52.721
125034 94.014 -77.893 52.699
125035 94.030 -77.897 52.677
125036 94.046 -77.902 52.653
125037 94.062 -77.906 52.631
125038 94.063 -77.898 52.608
125039 94.064 -77.890 52.586
125040 94.065 -77.882 52.564
125041 94.066 -77.874 52.542
125042 94.100 -77.859 52.519
125043 94.125 -77.850 52.496
125048 94.249 -77.805 52.385
125049 94.274 -77.796 52.363
125050 94.299 -77.786 52.342
125051 94.324 -77.777 52.322
125052 94.349 -77.768 52.301
125053 94.374 -77.758 52.280
125054 94.399 -77.749 52.261
125055 94.424 -77.740 52.242
125056 94.449 -77.731 52.223
125057 94.474 -77.722 52.203
125058 94.499 -77.713 52.185
125059 94.524 -77.703 52.167
125060 94.538 -77.697 52.150
125061 94.553 -77.690 52.132
125062 94.567 -77.684 52.114
125063 94.581 -77.677 52.096
125064 94.602 -77.669 52.077
125065 94.623 -77.660 52.058
125066 94.644 -77.652 52.039
125067 94.665 -77.643 52.019
125068 94.686 -77.634 51.999
125069 94.707 -77.626 51.976
125070 94.674 -77.643 51.954
125071 94.655 -77.640 51.932
125072 94.655 -77.642 51.909
125073 94.655 -77.644 51.885
125074 94.651 -77.645 51.862
125075 94.647 -77.647 51.839
125076 94.644 -77.649 51.812
125077 94.646 -77.656 51.784
125078 94.649 -77.662 51.756
125079 94.654 -77.669 51.729
125080 94.664 -77.676 51.701
125081 94.681 -77.681 51.673
125082 94.703 -77.688 51.644
125083 94.724 -77.695 51.611
125084 94.747 -77.702 51.575
125085 94.769 -77.708 51.540
125086 94.790 -77.715 51.505
125087 94.790 -77.700 51.470
125088 94.786 -77.681 51.435
125089 94.811 -77.680 51.400
125090 94.837 -77.684 51.362
125091 94.865 -77.715 51.324
125092 94.831 -77.741 51.286
125093 94.858 -77.762 51.247
125094 94.874 -77.768 51.209
125095 94.875 -77.761 51.170
125096 94.889 -77.755 51.130
125097 94.910 -77.750 51.090
125098 94.923 -77.749 51.050
125099 94.942 -77.755 51.010
125100 94.922 -77.761 50.971
125101 95.032 -77.814 50.933
125102 95.041 -77.793 50.894
125103 95.051 -77.774 50.855
125104 95.061 -77.753 50.816
125105 95.071 -77.733 50.777
125106 95.086 -77.726 50.740
125107 95.102 -77.723 50.702
125108 95.119 -77.719 50.664
125109 95.135 -77.716 50.627
125110 95.155 -77.717 50.590
125111 95.174 -77.717 50.552
125112 95.194 -77.718 50.516
125113 95.213 -77.718 50.480
125114 95.232 -77.718 50.445
125115 95.238 -77.714 50.411
125116 95.243 -77.710 50.378
125117 95.249 -77.707 50.344
125118 95.271 -77.695 50.312
125119 95.296 -77.683 50.280
125120 95.321 -77.670 50.249
125121 95.345 -77.657 50.219
125122 95.353 -77.649 50.189
125123 95.360 -77.641 50.161
125124 95.368 -77.633 50.132
125125 95.375 -77.624 50.102
125126 95.383 -77.617 50.072
125127 95.399 -77.595 50.043
125128 95.444 -77.574 50.014
125129 95.487 -77.558 49.984
125130 95.518 -77.573 49.956
125131 95.548 -77.587 49.928
125132 95.578 -77.602 49.900
125133 95.574 -77.621 49.872
125134 95.579 -77.638 49.843
125135 95.601 -77.646 49.814
125136 95.592 -77.646 49.786
125137 95.583 -77.646 49.755
125138 95.573 -77.646 49.723
125139 95.593 -77.650 49.691
125140 95.618 -77.654 49.658
125141 95.625 -77.675 49.626
125142 95.639 -77.673 49.594
125143 95.655 -77.667 49.564
125144 95.672 -77.661 49.536
125145 95.686 -77.656 49.507
125146 95.689 -77.654 49.480
125147 95.698 -77.655 49.452
125148 95.706 -77.659 49.428
125149 95.713 -77.666 49.402
125150 95.721 -77.674 49.375
125151 95.728 -77.680 49.346
125152 95.736 -77.688 49.319
125153 95.743 -77.695 49.291
125154 95.750 -77.702 49.261
125155 95.758 -77.709 49.227
125156 95.766 -77.716 49.190
125157 95.784 -77.724 49.153
125158 95.801 -77.731 49.114
125159 95.819 -77.739 49.076
125160 95.836 -77.746 49.038
125161 95.853 -77.753 49.000
125162 95.871 -77.761 48.964
125163 95.894 -77.772 48.927
125164 95.917 -77.781 48.891
125165 95.940 -77.790 48.854
125166 95.961 -77.799 48.818
125167 95.981 -77.812 48.783
125168 96.002 -77.824 48.747
125169 96.022 -77.837 48.711
125170 96.042 -77.849 48.674
125171 96.061 -77.862 48.637
125172 96.082 -77.874 48.600
125173 96.097 -77.881 48.566
125174 96.103 -77.878 48.533
125175 96.110 -77.875 48.499
125176 96.117 -77.872 48.467
125177 96.123 -77.870 48.437
125178 96.133 -77.870 48.407
125179 96.142 -77.865 48.376
125180 96.148 -77.859 48.351
125181 96.155 -77.857 48.324
125182 96.162 -77.855 48.295
125183 96.169 -77.855 48.269
125184 96.176 -77.854 48.241
125185 96.183 -77.853 48.216
125186 96.198 -77.863 48.192
125187 96.225 -77.888 48.170
125188 96.252 -77.913 48.149
125189 96.279 -77.938 48.126
125190 96.310 -77.975 48.102
125191 96.341 -78.011 48.079
125192 96.356 -78.024 48.056
125193 96.359 -78.021 48.033
125194 96.362 -78.018 48.012
125195 96.365 -78.014 47.989
125196 96.368 -78.011 47.965
125197 96.372 -78.008 47.941
125198 96.375 -78.004 47.919
125199 96.390 -78.017 47.897
125200 96.401 -78.036 47.876
125201 96.404 -78.060 47.855
125202 96.410 -78.079 47.830
125203 96.420 -78.089 47.804
125204 96.430 -78.099 47.776
125205 96.439 -78.109 47.750
125206 96.449 -78.119 47.723
125207 96.455 -78.115 47.697
125208 96.458 -78.099 47.672
125209 96.461 -78.083 47.645
125210 96.466 -78.069 47.618
125211 96.483 -78.068 47.590
125212 96.500 -78.067 47.562
125213 96.517 -78.066 47.533
125214 96.535 -78.065 47.504
125215 96.552 -78.065 47.478
125216 96.576 -78.067 47.454
125217 96.600 -78.081 47.430
125218 96.607 -78.100 47.405
125219 96.619 -78.109 47.381
125220 96.631 -78.117 47.356
125221 96.644 -78.125 47.331
125222 96.648 -78.134 47.307
125223 96.651 -78.144 47.283
125224 96.664 -78.150 47.259
125225 96.685 -78.151 47.234
125226 96.699 -78.160 47.210
125227 96.713 -78.170 47.185
125228 96.727 -78.181 47.160
125229 96.741 -78.191 47.134
125230 96.754 -78.201 47.107
125231 96.768 -78.211 47.081
125232 96.780 -78.225 47.054
125233 96.790 -78.244 47.029
125234 96.799 -78.263 47.003
125235 96.809 -78.282 46.976
125236 96.818 -78.300 46.950
125237 96.837 -78.313 46.922
125238 96.864 -78.325 46.895
125239 96.891 -78.337 46.867
125240 96.918 -78.348 46.839
125241 96.945 -78.360 46.812
125242 96.971 -78.372 46.785
125243 96.998 -78.383 46.758
125244 97.025 -78.395 46.734
125245 97.052 -78.408 46.710
125246 97.079 -78.422 46.687
125247 97.106 -78.436 46.664
125248 97.133 -78.450 46.640
125249 97.162 -78.464 46.617
125250 97.198 -78.474 46.593
125251 97.240 -78.479 46.572
125252 97.282 -78.470 46.550
125253 97.323 -78.454 46.527
125254 97.367 -78.438 46.506
125255 97.430 -78.436 46.485
125256 97.472 -78.437 46.466
125257 97.496 -78.442 46.447
125258 97.519 -78.447 46.431
125259 97.546 -78.453 46.414
125260 97.588 -78.465 46.397
125261 97.631 -78.478 46.381
125262 97.673 -78.490 46.364
125263 97.715 -78.502 46.347
125264 97.767 -78.523 46.330
125265 97.826 -78.551 46.312
125266 97.886 -78.579 46.295
125267 97.930 -78.605 46.278
125268 97.943 -78.590 46.260
125269 97.956 -78.576 46.244
125270 97.969 -78.561 46.227
125271 97.974 -78.529 46.210
125272 98.013 -78.531 46.194
125273 98.055 -78.529 46.177
125274 98.095 -78.528 46.161
125275 98.132 -78.516 46.145
125276 98.180 -78.491 46.130
125277 98.206 -78.485 46.115
125278 98.207 -78.501 46.100
125279 98.208 -78.515 46.086
125280 98.202 -78.512 46.071
125281 98.200 -78.507 46.056
125282 98.213 -78.511 46.041
125283 98.226 -78.515 46.026
125284 98.228 -78.514 46.010
125285 98.229 -78.510 45.995
125286 98.240 -78.507 45.980
125287 98.279 -78.502 45.963
125288 98.290 -78.463 45.946
125289 98.298 -78.433 45.928
125290 98.301 -78.443 45.911
125291 98.304 -78.445 45.893
125292 98.324 -78.427 45.874
125293 98.343 -78.428 45.854
125294 98.361 -78.430 45.834
125295 98.379 -78.432 45.814
125296 98.392 -78.460 45.793
125297 98.405 -78.473 45.772
125298 98.418 -78.486 45.752
125299 98.430 -78.499 45.733
125300 98.443 -78.512 45.713
125301 98.456 -78.525 45.694
125302 98.468 -78.538 45.675
125303 98.481 -78.544 45.656
125304 98.492 -78.551 45.636
125305 98.502 -78.557 45.617
125306 98.512 -78.564 45.598
125307 98.522 -78.570 45.580
125308 98.533 -78.577 45.562
125309 98.543 -78.584 45.544
125310 98.554 -78.590 45.525
125311 98.564 -78.597 45.507
125312 98.574 -78.604 45.490
125313 98.601 -78.596 45.473
125314 98.620 -78.584 45.456
125315 98.634 -78.572 45.438
125316 98.647 -78.560 45.421
125317 98.671 -78.546 45.404
125318 98.720 -78.561 45.388
125319 98.769 -78.576 45.373
125320 98.803 -78.583 45.357
125321 98.790 -78.577 45.342
125322 98.783 -78.571 45.328
125323 98.802 -78.572 45.312
125324 98.813 -78.578 45.295
125325 98.825 -78.584 45.278
125326 98.836 -78.590 45.263
125327 98.847 -78.596 45.248
125328 98.875 -78.600 45.233
125329 98.914 -78.601 45.216
125330 99.002 -78.581 45.199
125331 99.018 -78.586 45.182
125332 99.000 -78.596 45.166
125333 98.982 -78.605 45.153
125334 98.982 -78.625 45.138
125335 99.041 -78.607 45.124
125336 99.071 -78.604 45.112
125337 99.091 -78.610 45.099
125338 99.110 -78.615 45.086
125339 99.147 -78.612 45.074
125340 99.201 -78.603 45.062
125341 99.158 -78.633 45.051
125342 99.158 -78.699 45.039
125343 99.157 -78.765 45.027
125344 99.142 -78.786 45.016
125345 99.171 -78.771 45.005
125346 99.214 -78.748 44.993
125347 99.250 -78.736 44.981
125348 99.283 -78.730 44.970
125349 99.315 -78.724 44.959
125350 99.364 -78.737 44.947
125351 99.419 -78.756 44.936
125352 99.474 -78.774 44.924
125353 99.528 -78.793 44.912
125354 99.583 -78.812 44.902
125355 99.637 -78.830 44.890
125356 99.664 -78.838 44.879
125357 99.671 -78.837 44.869
125358 99.701 -78.877 44.857
125359 99.732 -78.920 44.846
125360 99.758 -78.952 44.835
125361 99.778 -78.975 44.824
125362 99.798 -78.998 44.814
125363 99.818 -79.021 44.804
125364 99.838 -79.043 44.794
125365 99.858 -79.066 44.787
125366 99.878 -79.089 44.779
125367 99.898 -79.110 44.772
125368 99.911 -79.120 44.766
125369 99.924 -79.130 44.761
125370 99.937 -79.140 44.756
125371 99.945 -79.141 44.751
125372 99.980 -79.148 44.747
125373 100.019 -79.161 44.742
125374 100.059 -79.174 44.738
125375 100.074 -79.195 44.734
125376 100.082 -79.219 44.730
125377 100.112 -79.237 44.726
125378 100.141 -79.256 44.722
125379 100.171 -79.274 44.718
125380 100.200 -79.292 44.714
125381 100.227 -79.298 44.710
125382 100.255 -79.303 44.705
125383 100.282 -79.309 44.700
125384 100.309 -79.315 44.694
125385 100.336 -79.320 44.689
125386 100.363 -79.325 44.684
125387 100.391 -79.331 44.678
125388 100.418 -79.337 44.671
125389 100.456 -79.347 44.665
125390 100.494 -79.357 44.657
125391 100.533 -79.367 44.648
125392 100.571 -79.378 44.640
125393 100.610 -79.389 44.632
125394 100.649 -79.406 44.624
125395 100.688 -79.422 44.615
125396 100.727 -79.439 44.607
125397 100.767 -79.456 44.598
125398 100.806 -79.473 44.589
125399 100.845 -79.489 44.579
125400 100.884 -79.506 44.570
125401 100.923 -79.523 44.559
125402 100.963 -79.539 44.550
125403 101.002 -79.556 44.540
125404 101.041 -79.573 44.529
125405 101.077 -79.590 44.517
125406 101.110 -79.609 44.504
125407 101.142 -79.629 44.492
125408 101.175 -79.648 44.478
125409 101.207 -79.667 44.464
125410 101.240 -79.686 44.450
125411 101.272 -79.704 44.435
125412 101.305 -79.724 44.419
125413 101.357 -79.748 44.403
125414 101.434 -79.772 44.387
125415 101.486 -79.779 44.371
125416 101.525 -79.788 44.355
125417 101.549 -79.809 44.338
125418 101.582 -79.823 44.321
125419 101.599 -79.829 44.305
125420 101.615 -79.835 44.287
125421 101.632 -79.840 44.269
125422 101.649 -79.846 44.252
125423 101.666 -79.852 44.234
125424 101.683 -79.857 44.216
125425 101.699 -79.863 44.198
125715 93.894 -77.871 53.118
125716 93.907 -77.868 53.099
125717 93.920 -77.863 53.079
125718 93.933 -77.859 53.060
125719 93.946 -77.855 53.041
125720 93.959 -77.850 53.021
125721 93.972 -77.846 53.001
125722 93.984 -77.842 52.982
125723 93.981 -77.840 52.961
125724 93.979 -77.840 52.942
125725 93.977 -77.839 52.923
125726 93.974 -77.837 52.903
125727 93.972 -77.837 52.883
125728 93.969 -77.835 52.862
125729 93.967 -77.834 52.841
125730 93.969 -77.842 52.820
125731 93.972 -77.852 52.799
125732 93.976 -77.862 52.777
125733 93.986 -77.870 52.756
125734 94.001 -77.874 52.734
125735 94.017 -77.878 52.713
125736 94.033 -77.882 52.691
125737 94.049 -77.887 52.668
125738 94.063 -77.890 52.645
125739 94.064 -77.881 52.623
125740 94.066 -77.873 52.600
125741 94.067 -77.865 52.579
125742 94.068 -77.857 52.556
125743 94.101 -77.843 52.533
125744 94.126 -77.834 52.510
125745 94.151 -77.824 52.488
125746 94.176 -77.815 52.466
125747 94.201 -77.806 52.444
125748 94.226 -77.797 52.423
125749 94.251 -77.787 52.400
125750 94.276 -77.778 52.379
125751 94.300 -77.769 52.358
125752 94.326 -77.760 52.338
125753 94.350 -77.751 52.317
125754 94.376 -77.742 52.297
125755 94.400 -77.733 52.278
125756 94.425 -77.724 52.258
125757 94.450 -77.714 52.239
125758 94.475 -77.705 52.221
125759 94.500 -77.696 52.204
125760 94.524 -77.688 52.187
125761 94.537 -77.685 52.171
125762 94.558 -77.677 52.154
125763 94.578 -77.667 52.137
125764 94.600 -77.659 52.120
125765 94.620 -77.651 52.102
125766 94.641 -77.642 52.083
125767 94.662 -77.633 52.062
125768 94.683 -77.625 52.040
125769 94.704 -77.617 52.018
125770 94.694 -77.633 51.996
125771 94.665 -77.633 51.972
125772 94.644 -77.630 51.948
125773 94.642 -77.632 51.925
125774 94.642 -77.633 51.902
125775 94.641 -77.636 51.879
125776 94.640 -77.637 51.856
125777 94.641 -77.642 51.831
125778 94.643 -77.649 51.804
125779 94.649 -77.655 51.776
125780 94.656 -77.661 51.749
125781 94.667 -77.668 51.721
125782 94.689 -77.675 51.691
125783 94.711 -77.682 51.660
125784 94.733 -77.689 51.627
125785 94.755 -77.696 51.591
125786 94.777 -77.702 51.556
125787 94.792 -77.703 51.522
125788 94.788 -77.684 51.488
125789 94.784 -77.665 51.454
125790 94.803 -77.660 51.419
125791 94.829 -77.667 51.381
125792 94.857 -77.699 51.344
125793 94.877 -77.730 51.306
125794 94.891 -77.765 51.267
125795 94.892 -77.759 51.228
125796 94.893 -77.752 51.190
125797 94.894 -77.746 51.151
125798 94.909 -77.740 51.111
125799 94.920 -77.736 51.071
125800 95.028 -77.732 51.031
125801 95.129 -77.727 50.991
125802 94.966 -77.752 50.952
125803 95.045 -77.778 50.913
125804 95.055 -77.758 50.874
125805 95.064 -77.737 50.834
125806 95.076 -77.723 50.795
125807 95.093 -77.720 50.757
125808 95.109 -77.716 50.718
125809 95.128 -77.716 50.680
125810 95.148 -77.716 50.643
125811 95.168 -77.717 50.605
125812 95.187 -77.717 50.567
125813 95.207 -77.718 50.531
125814 95.226 -77.718 50.495
125815 95.231 -77.714 50.459
125816 95.237 -77.710 50.425
125817 95.242 -77.707 50.391
125818 95.253 -77.700 50.357
125819 95.278 -77.687 50.324
125820 95.304 -77.675 50.291
125821 95.328 -77.662 50.259
125822 95.353 -77.649 50.228
125823 95.364 -77.640 50.198
125824 95.372 -77.632 50.168
125825 95.379 -77.624 50.139
125826 95.387 -77.616 50.109
125827 95.394 -77.608 50.079
125828 95.422 -77.577 50.049
125829 95.467 -77.556 50.020
125830 95.513 -77.536 49.991
125831 95.552 -77.528 49.962
125832 95.583 -77.542 49.934
125833 95.607 -77.558 49.906
125834 95.586 -77.580 49.878
125835 95.564 -77.603 49.850
125836 95.584 -77.615 49.822
125837 95.574 -77.615 49.793
125838 95.565 -77.615 49.764
125839 95.590 -77.620 49.733
125840 95.615 -77.624 49.701
125841 95.640 -77.629 49.668
125842 95.656 -77.642 49.636
125843 95.654 -77.663 49.604
125844 95.670 -77.658 49.574
125845 95.670 -77.652 49.544
125846 95.673 -77.651 49.516
125847 95.682 -77.652 49.488
125848 95.691 -77.652 49.461
125849 95.699 -77.653 49.437
125850 95.708 -77.655 49.411
125851 95.716 -77.663 49.385
125852 95.723 -77.670 49.356
125853 95.730 -77.677 49.328
125854 95.738 -77.684 49.299
125855 95.745 -77.691 49.270
125856 95.753 -77.699 49.235
125857 95.761 -77.705 49.199
125858 95.778 -77.713 49.162
125859 95.796 -77.721 49.123
125860 95.813 -77.727 49.085
125861 95.831 -77.735 49.047
125862 95.848 -77.743 49.010
125863 95.871 -77.753 48.973
125864 95.895 -77.765 48.936
125865 95.918 -77.774 48.900
125866 95.941 -77.782 48.864
125867 95.964 -77.790 48.828
125868 95.986 -77.799 48.793
125869 96.007 -77.811 48.757
125870 96.027 -77.824 48.720
125871 96.047 -77.836 48.684
125872 96.067 -77.849 48.649
125873 96.087 -77.861 48.612
125874 96.107 -77.874 48.577
125875 96.115 -77.872 48.542
125876 96.122 -77.870 48.509
125877 96.128 -77.867 48.478
125878 96.135 -77.864 48.448
125879 96.140 -77.865 48.419
125880 96.146 -77.866 48.390
125881 96.152 -77.865 48.363
125882 96.159 -77.864 48.335
125883 96.166 -77.863 48.306
125884 96.173 -77.862 48.277
125885 96.180 -77.862 48.250
125886 96.187 -77.860 48.224
125887 96.212 -77.883 48.197
125888 96.239 -77.908 48.174
125889 96.265 -77.932 48.152
125890 96.293 -77.957 48.131
125891 96.320 -77.982 48.110
125892 96.347 -78.009 48.087
125893 96.368 -78.031 48.064
125894 96.372 -78.028 48.041
125895 96.375 -78.025 48.019
125896 96.378 -78.022 47.996
125897 96.381 -78.018 47.973
125898 96.384 -78.015 47.950
125899 96.387 -78.012 47.929
125900 96.390 -78.009 47.908
125901 96.402 -78.024 47.886
125902 96.408 -78.044 47.865
125903 96.418 -78.053 47.841
125904 96.428 -78.064 47.815
125905 96.437 -78.074 47.788
125906 96.447 -78.084 47.763
125907 96.456 -78.094 47.737
125908 96.464 -78.097 47.711
125909 96.467 -78.081 47.684
125910 96.470 -78.065 47.656
125911 96.473 -78.050 47.628
125912 96.477 -78.034 47.600
125913 96.493 -78.032 47.572
125914 96.511 -78.031 47.546
125915 96.528 -78.031 47.520
125916 96.549 -78.031 47.494
125917 96.583 -78.037 47.470
125918 96.612 -78.046 47.446
125919 96.617 -78.073 47.420
125920 96.624 -78.096 47.397
125921 96.636 -78.104 47.373
125922 96.648 -78.112 47.349
125923 96.655 -78.122 47.325
125924 96.658 -78.131 47.300
125925 96.676 -78.134 47.276
125926 96.693 -78.141 47.251
125927 96.707 -78.150 47.227
125928 96.720 -78.161 47.202
125929 96.734 -78.171 47.177
125930 96.748 -78.181 47.150
125931 96.761 -78.191 47.124
125932 96.776 -78.202 47.098
125933 96.789 -78.213 47.071
125934 96.798 -78.232 47.046
125935 96.808 -78.251 47.020
125936 96.817 -78.269 46.994
125937 96.832 -78.285 46.967
125938 96.851 -78.297 46.939
125939 96.869 -78.308 46.911
125940 96.888 -78.320 46.883
125941 96.915 -78.332 46.855
125942 96.942 -78.344 46.830
125943 96.969 -78.355 46.805
125944 96.996 -78.366 46.780
125945 97.022 -78.379 46.755
125946 97.049 -78.392 46.732
125947 97.076 -78.407 46.708
125948 97.104 -78.420 46.684
125949 97.133 -78.435 46.661
125950 97.162 -78.449 46.637
125951 97.202 -78.456 46.613
125952 97.243 -78.462 46.591
125953 97.285 -78.468 46.569
125954 97.326 -78.459 46.547
125955 97.368 -78.442 46.526
125956 97.416 -78.430 46.505
125957 97.478 -78.428 46.485
125958 97.508 -78.432 46.466
125959 97.531 -78.436 46.449
125960 97.554 -78.441 46.433
125961 97.578 -78.446 46.416
125962 97.613 -78.455 46.400
125963 97.655 -78.468 46.382
125964 97.698 -78.480 46.364
125965 97.740 -78.493 46.346
125966 97.800 -78.520 46.329
125967 97.860 -78.549 46.310
125968 97.912 -78.591 46.293
125969 97.924 -78.576 46.275
125970 97.931 -78.547 46.258
125971 97.934 -78.511 46.241
125972 97.970 -78.510 46.224
125973 98.014 -78.518 46.208
125974 98.055 -78.520 46.191
125975 98.098 -78.512 46.175
125976 98.134 -78.498 46.160
125977 98.164 -78.469 46.144
125978 98.202 -78.444 46.130
125979 98.203 -78.458 46.115
125980 98.204 -78.473 46.101
125981 98.207 -78.489 46.087
125982 98.203 -78.501 46.073
125983 98.204 -78.498 46.058
125984 98.204 -78.495 46.042
125985 98.205 -78.492 46.027
125986 98.206 -78.489 46.012
125987 98.229 -78.485 45.997
125988 98.268 -78.480 45.980
125989 98.303 -78.473 45.963
125990 98.314 -78.432 45.945
125991 98.317 -78.438 45.927
125992 98.319 -78.433 45.908
125993 98.329 -78.423 45.889
125994 98.343 -78.431 45.870
125995 98.355 -78.435 45.849
125996 98.367 -78.439 45.829
125997 98.383 -78.449 45.808
125998 98.400 -78.466 45.788
125999 98.412 -78.479 45.768
126000 98.425 -78.492 45.748
126001 98.438 -78.504 45.728
126002 98.450 -78.517 45.708
126003 98.457 -78.532 45.689
126004 98.471 -78.538 45.669
126005 98.484 -78.544 45.650
126006 98.496 -78.550 45.630
126007 98.508 -78.556 45.611
126008 98.520 -78.562 45.592
126009 98.533 -78.567 45.573
126010 98.545 -78.573 45.554
126011 98.558 -78.579 45.535
126012 98.570 -78.586 45.517
126013 98.582 -78.592 45.499
126014 98.593 -78.589 45.482
126015 98.606 -78.582 45.464
126016 98.620 -78.570 45.446
126017 98.634 -78.557 45.429
126018 98.651 -78.535 45.412
126019 98.701 -78.551 45.395
126020 98.741 -78.558 45.380
126021 98.775 -78.559 45.364
126022 98.785 -78.553 45.348
126023 98.784 -78.547 45.334
126024 98.797 -78.551 45.317
126025 98.808 -78.557 45.301
126026 98.820 -78.563 45.284
126027 98.831 -78.569 45.269
126028 98.842 -78.575 45.254
126029 98.854 -78.581 45.239
126030 98.909 -78.575 45.223
126031 98.962 -78.587 45.206
126032 98.981 -78.598 45.189
126033 98.963 -78.608 45.173
126034 98.949 -78.620 45.160
126035 98.999 -78.614 45.146
126036 99.060 -78.592 45.132
126037 99.079 -78.597 45.120
126038 99.098 -78.602 45.107
126039 99.117 -78.607 45.095
126040 99.137 -78.612 45.083
126041 99.173 -78.610 45.071
126042 99.219 -78.604 45.060
126043 99.178 -78.657 45.048
126044 99.177 -78.724 45.037
126045 99.165 -78.752 45.025
126046 99.159 -78.761 45.013
126047 99.202 -78.737 45.000
126048 99.239 -78.723 44.989
126049 99.272 -78.717 44.977
126050 99.308 -78.714 44.965
126051 99.363 -78.733 44.954
126052 99.417 -78.752 44.942
126053 99.472 -78.771 44.930
126054 99.526 -78.789 44.918
126055 99.581 -78.808 44.906
126056 99.636 -78.827 44.894
126057 99.682 -78.842 44.883
126058 99.688 -78.841 44.871
126059 99.714 -78.874 44.859
126060 99.746 -78.918 44.848
126061 99.769 -78.946 44.836
126062 99.789 -78.968 44.825
126063 99.809 -78.991 44.816
126064 99.829 -79.014 44.805
126065 99.849 -79.037 44.796
126066 99.869 -79.060 44.788
126067 99.886 -79.076 44.781
126068 99.899 -79.086 44.774
126069 99.912 -79.096 44.768
126070 99.925 -79.106 44.763
126071 99.938 -79.115 44.758
126072 99.950 -79.122 44.754
126073 99.981 -79.126 44.749
126074 100.020 -79.138 44.745
126075 100.059 -79.152 44.740
126076 100.065 -79.176 44.736
126077 100.084 -79.197 44.733
126078 100.114 -79.216 44.728
126079 100.143 -79.234 44.724
126080 100.173 -79.252 44.720
126081 100.202 -79.270 44.717
126082 100.230 -79.278 44.712
126083 100.257 -79.284 44.708
126084 100.284 -79.289 44.703
126085 100.312 -79.294 44.697
126086 100.339 -79.300 44.691
126087 100.366 -79.306 44.686
126088 100.394 -79.311 44.680
126089 100.424 -79.318 44.673
126090 100.463 -79.328 44.666
126091 100.501 -79.338 44.658
126097 100.735 -79.426 44.607
126098 100.774 -79.442 44.598
126099 100.813 -79.459 44.589
126100 100.852 -79.476 44.579
126101 100.892 -79.492 44.570
126102 100.931 -79.509 44.559
126103 100.970 -79.526 44.549
126104 101.009 -79.542 44.538
126105 101.042 -79.561 44.527
126106 101.074 -79.580 44.515
126107 101.107 -79.599 44.502
126108 101.139 -79.618 44.489
126109 101.172 -79.638 44.475
126110 101.204 -79.657 44.461
126111 101.237 -79.676 44.446
126112 101.269 -79.695 44.431
126113 101.302 -79.714 44.415
126114 101.382 -79.746 44.399
126115 101.439 -79.756 44.383
126116 101.491 -79.762 44.367
126117 101.524 -79.773 44.351
126118 101.549 -79.795 44.335
126119 101.583 -79.811 44.319
126120 101.600 -79.816 44.300
126121 101.616 -79.821 44.282
126122 101.633 -79.827 44.265
126123 101.650 -79.833 44.247
126124 101.667 -79.839 44.230
126415 93.882 -77.865 53.149
126416 93.895 -77.860 53.129
126417 93.908 -77.856 53.110
126418 93.921 -77.852 53.090
126419 93.933 -77.848 53.071
126420 93.946 -77.843 53.052
126421 93.959 -77.839 53.033
126422 93.972 -77.835 53.014
126423 93.984 -77.830 52.994
126424 93.982 -77.830 52.974
126425 93.979 -77.828 52.954
126426 93.977 -77.827 52.935
126427 93.975 -77.826 52.915
126428 93.972 -77.825 52.895
126429 93.970 -77.824 52.874
126430 93.967 -77.823 52.853
126431 93.968 -77.829 52.833
126432 93.972 -77.839 52.812
126433 93.975 -77.849 52.791
126434 93.989 -77.855 52.770
126435 94.005 -77.859 52.748
126436 94.020 -77.863 52.726
126437 94.037 -77.868 52.704
126438 94.052 -77.871 52.683
126439 94.065 -77.873 52.660
126440 94.066 -77.865 52.637
126441 94.067 -77.856 52.615
126442 94.068 -77.849 52.593
126443 94.069 -77.840 52.570
126444 94.103 -77.826 52.547
126445 94.127 -77.817 52.524
126446 94.152 -77.808 52.503
126447 94.177 -77.799 52.481
126448 94.202 -77.789 52.459
126449 94.227 -77.780 52.438
126450 94.252 -77.771 52.416
126451 94.277 -77.761 52.395
126452 94.302 -77.752 52.374
126453 94.327 -77.743 52.354
126454 94.352 -77.734 52.334
126455 94.377 -77.725 52.314
126456 94.402 -77.716 52.295
126457 94.427 -77.707 52.275
126458 94.452 -77.697 52.257
126459 94.475 -77.690 52.239
126460 94.492 -77.688 52.222
126461 94.511 -77.681 52.207
126462 94.520 -77.677 52.192
126463 94.524 -77.681 52.176
126464 94.545 -77.673 52.160
126465 94.566 -77.664 52.144
126466 94.586 -77.656 52.126
126467 94.608 -77.647 52.106
126468 94.628 -77.639 52.085
126469 94.651 -77.630 52.063
126470 94.689 -77.617 52.041
126471 94.692 -77.624 52.018
126472 94.661 -77.623 51.994
126473 94.635 -77.620 51.971
126474 94.629 -77.620 51.948
126475 94.629 -77.623 51.924
126476 94.628 -77.629 51.900
126477 94.639 -77.664 51.877
126478 94.654 -77.651 51.852
126479 94.667 -77.642 51.825
126480 94.673 -77.649 51.797
126481 94.679 -77.655 51.771
126482 94.685 -77.662 51.742
126483 94.697 -77.669 51.711
126484 94.719 -77.676 51.679
126485 94.741 -77.683 51.645
126486 94.763 -77.689 51.610
126487 94.785 -77.696 51.575
126488 94.790 -77.687 51.542
126489 94.786 -77.668 51.508
126490 94.782 -77.649 51.474
126491 94.794 -77.640 51.440
126492 94.821 -77.651 51.403
126493 94.846 -77.683 51.366
126494 94.864 -77.713 51.328
126495 94.910 -77.755 51.288
126496 94.909 -77.750 51.249
126497 94.904 -77.743 51.211
126498 94.900 -77.735 51.173
126499 94.898 -77.724 51.133
126500 94.962 -77.711 51.094
126501 95.035 -77.701 51.054
126502 95.107 -77.690 51.013
126503 95.089 -77.727 50.973
126504 94.998 -77.755 50.933
126505 95.058 -77.742 50.893
126506 95.068 -77.721 50.853
126507 95.084 -77.717 50.814
126508 95.102 -77.715 50.775
126509 95.121 -77.716 50.735
126510 95.141 -77.716 50.697
126511 95.161 -77.717 50.659
126512 95.180 -77.717 50.622
126513 95.200 -77.718 50.583
126514 95.219 -77.718 50.546
126515 95.225 -77.714 50.510
126516 95.230 -77.710 50.474
126517 95.236 -77.706 50.439
126518 95.241 -77.703 50.404
126519 95.261 -77.692 50.370
126520 95.286 -77.680 50.337
126521 95.311 -77.667 50.303
126522 95.336 -77.655 50.271
126523 95.361 -77.642 50.239
126524 95.375 -77.632 50.208
126525 95.383 -77.624 50.177
126526 95.390 -77.615 50.147
126527 95.398 -77.608 50.117
126528 95.409 -77.592 50.086
126529 95.446 -77.558 50.057
126530 95.490 -77.539 50.028
126531 95.536 -77.518 49.999
126532 95.581 -77.498 49.969
126533 95.617 -77.498 49.941
126534 95.618 -77.517 49.913
126535 95.597 -77.539 49.885
126536 95.576 -77.562 49.858
126537 95.561 -77.583 49.830
126538 95.562 -77.585 49.801
126539 95.588 -77.595 49.773
126540 95.613 -77.630 49.742
126541 95.636 -77.645 49.710
126542 95.660 -77.660 49.678
126543 95.684 -77.635 49.646
126544 95.658 -77.652 49.615
126545 95.653 -77.649 49.584
126546 95.657 -77.648 49.554
126547 95.666 -77.649 49.525
126548 95.675 -77.649 49.497
126549 95.684 -77.650 49.471
126550 95.692 -77.651 49.446
126551 95.701 -77.652 49.420
126552 95.710 -77.652 49.394
126553 95.718 -77.659 49.366
126554 95.725 -77.666 49.338
126555 95.733 -77.674 49.308
126556 95.740 -77.680 49.278
126557 95.747 -77.687 49.244
126558 95.756 -77.695 49.208
126559 95.773 -77.702 49.170
126560 95.791 -77.709 49.132
126561 95.808 -77.717 49.094
126562 95.825 -77.724 49.056
126563 95.847 -77.735 49.019
126564 95.873 -77.748 48.982
126565 95.897 -77.758 48.945
126566 95.919 -77.767 48.909
126567 95.943 -77.775 48.873
126568 95.965 -77.783 48.838
126569 95.988 -77.792 48.802
126570 96.011 -77.800 48.766
126571 96.032 -77.810 48.730
126572 96.052 -77.823 48.695
126573 96.073 -77.835 48.659
126574 96.093 -77.848 48.622
126575 96.113 -77.861 48.587
126576 96.127 -77.867 48.553
126577 96.134 -77.864 48.520
126578 96.141 -77.861 48.488
126579 96.147 -77.858 48.458
126580 96.148 -77.865 48.429
126581 96.150 -77.874 48.400
126582 96.157 -77.872 48.372
126583 96.163 -77.871 48.344
126584 96.170 -77.871 48.315
126585 96.177 -77.870 48.286
126586 96.184 -77.868 48.258
126587 96.198 -77.877 48.231
126588 96.225 -77.902 48.203
126589 96.252 -77.927 48.179
126590 96.279 -77.952 48.156
126591 96.306 -77.977 48.134
126592 96.333 -78.001 48.115
126593 96.360 -78.026 48.092
126594 96.365 -78.025 48.069
126595 96.364 -78.022 48.045
126596 96.369 -78.020 48.024
126597 96.374 -78.018 48.002
126598 96.379 -78.016 47.980
126599 96.384 -78.014 47.959
126600 96.389 -78.012 47.938
126601 96.393 -78.010 47.917
126602 96.404 -78.009 47.896
126603 96.415 -78.018 47.874
126604 96.425 -78.028 47.850
126605 96.435 -78.038 47.826
126606 96.445 -78.049 47.801
126607 96.454 -78.059 47.776
126608 96.464 -78.069 47.750
126609 96.473 -78.078 47.723
126610 96.477 -78.064 47.696
126611 96.480 -78.048 47.667
126612 96.483 -78.032 47.639
126613 96.486 -78.016 47.611
126614 96.489 -78.000 47.584
126615 96.504 -77.996 47.559
126616 96.522 -77.996 47.534
126617 96.555 -78.002 47.509
126618 96.589 -78.007 47.486
126619 96.622 -78.013 47.461
126620 96.630 -78.038 47.436
126621 96.635 -78.065 47.412
126622 96.641 -78.091 47.389
126623 96.653 -78.100 47.365
126624 96.662 -78.108 47.341
126625 96.667 -78.117 47.316
126626 96.686 -78.121 47.292
126627 96.700 -78.131 47.269
126628 96.714 -78.141 47.244
126629 96.728 -78.151 47.218
126630 96.741 -78.162 47.192
126631 96.755 -78.172 47.166
126632 96.769 -78.182 47.140
126633 96.783 -78.192 47.114
126634 96.796 -78.203 47.087
126635 96.807 -78.220 47.062
126636 96.816 -78.239 47.037
126637 96.829 -78.255 47.011
126638 96.847 -78.267 46.983
126639 96.865 -78.279 46.955
126640 96.883 -78.291 46.926
126641 96.901 -78.304 46.898
126642 96.920 -78.316 46.871
126643 96.939 -78.327 46.848
126644 96.966 -78.339 46.825
126645 96.993 -78.351 46.802
126646 97.020 -78.363 46.778
126647 97.046 -78.376 46.754
126648 97.075 -78.391 46.730
126649 97.104 -78.405 46.705
126650 97.133 -78.420 46.681
126651 97.163 -78.432 46.657
126652 97.205 -78.438 46.633
126653 97.246 -78.445 46.609
126654 97.288 -78.451 46.587
126655 97.329 -78.456 46.566
126656 97.371 -78.448 46.546
126657 97.412 -78.432 46.525
126658 97.464 -78.423 46.505
126659 97.520 -78.421 46.486
126660 97.543 -78.426 46.470
126661 97.566 -78.431 46.453
126662 97.589 -78.435 46.437
126663 97.612 -78.440 46.420
126664 97.638 -78.446 46.401
126665 97.680 -78.458 46.383
126666 97.723 -78.470 46.365
126667 97.773 -78.490 46.346
126668 97.833 -78.518 46.328
126669 97.889 -78.563 46.310
126670 97.891 -78.529 46.293
126671 97.894 -78.493 46.275
126672 97.927 -78.488 46.257
126673 97.971 -78.496 46.240
126674 98.015 -78.505 46.225
126675 98.053 -78.519 46.208
126676 98.101 -78.496 46.192
126677 98.136 -78.479 46.176
126678 98.166 -78.452 46.161
126679 98.184 -78.434 46.147
126680 98.169 -78.435 46.133
126681 98.167 -78.454 46.118
126682 98.166 -78.473 46.104
126683 98.189 -78.501 46.090
126684 98.187 -78.495 46.076
126685 98.181 -78.476 46.060
126686 98.181 -78.471 46.045
126687 98.182 -78.469 46.029
126688 98.219 -78.463 46.013
126689 98.256 -78.458 45.996
126690 98.295 -78.454 45.979
126691 98.328 -78.442 45.962
126692 98.334 -78.432 45.944
126693 98.329 -78.419 45.924
126694 98.321 -78.413 45.905
126695 98.331 -78.417 45.886
126696 98.343 -78.421 45.866
126697 98.355 -78.426 45.845
126698 98.372 -78.437 45.824
126699 98.393 -78.458 45.804
126700 98.407 -78.471 45.784
126701 98.419 -78.484 45.764
126702 98.432 -78.497 45.743
126703 98.445 -78.510 45.723
126704 98.447 -78.526 45.703
126705 98.461 -78.532 45.683
126706 98.474 -78.538 45.664
126707 98.486 -78.544 45.644
126708 98.499 -78.550 45.625
126709 98.511 -78.556 45.605
126710 98.523 -78.561 45.585
126711 98.535 -78.567 45.565
126712 98.548 -78.573 45.546
126713 98.560 -78.579 45.528
126714 98.576 -78.589 45.509
126715 98.583 -78.570 45.491
126716 98.594 -78.568 45.473
126717 98.606 -78.567 45.455
126718 98.620 -78.555 45.438
126719 98.638 -78.532 45.420
126720 98.671 -78.531 45.402
126721 98.705 -78.532 45.386
126722 98.739 -78.533 45.370
126723 98.755 -78.529 45.353
126724 98.775 -78.529 45.338
126725 98.798 -78.532 45.320
126726 98.792 -78.516 45.304
126727 98.803 -78.523 45.289
126728 98.815 -78.529 45.274
126729 98.827 -78.536 45.259
126730 98.840 -78.549 45.243
126731 98.862 -78.586 45.227
126732 98.915 -78.599 45.212
126733 98.944 -78.611 45.196
126734 98.926 -78.620 45.181
126735 98.958 -78.620 45.168
126736 99.020 -78.597 45.155
126737 99.067 -78.584 45.142
126738 99.086 -78.589 45.129
126739 99.106 -78.594 45.116
126740 99.125 -78.599 45.104
126741 99.144 -78.604 45.092
126742 99.163 -78.610 45.081
126743 99.200 -78.608 45.069
126744 99.205 -78.617 45.058
126745 99.197 -78.681 45.047
126746 99.188 -78.720 45.034
126747 99.171 -78.734 45.021
126748 99.190 -78.727 45.009
126749 99.229 -78.709 44.997
126750 99.262 -78.703 44.985
126751 99.306 -78.711 44.973
126752 99.361 -78.730 44.961
126753 99.416 -78.749 44.948
126754 99.471 -78.767 44.936
126755 99.525 -78.786 44.924
126756 99.580 -78.805 44.911
126757 99.634 -78.824 44.899
126758 99.689 -78.842 44.887
126759 99.705 -78.845 44.875
126760 99.727 -78.873 44.862
126761 99.759 -78.916 44.850
126762 99.780 -78.939 44.839
126763 99.800 -78.962 44.827
126764 99.820 -78.985 44.817
126765 99.840 -79.008 44.808
126766 99.860 -79.031 44.798
126767 99.874 -79.043 44.791
126768 99.888 -79.053 44.784
126769 99.901 -79.062 44.777
126770 99.914 -79.072 44.772
126771 99.927 -79.082 44.766
126772 99.940 -79.092 44.762
126773 99.954 -79.102 44.758
126774 99.982 -79.103 44.753
126775 100.021 -79.116 44.748
126776 100.049 -79.133 44.744
126777 100.057 -79.157 44.739
126778 100.086 -79.175 44.736
126779 100.116 -79.194 44.733
126780 100.145 -79.212 44.728
126781 100.175 -79.230 44.723
126782 100.204 -79.248 44.720
126783 100.232 -79.258 44.716
126784 100.260 -79.263 44.711
126785 100.287 -79.269 44.706
126786 100.314 -79.275 44.700
126787 100.342 -79.280 44.695
126788 100.369 -79.285 44.689
126789 100.396 -79.291 44.682
126790 100.431 -79.300 44.675
126791 100.470 -79.310 44.668
126805 101.007 -79.532 44.537
126806 101.040 -79.551 44.526
126807 101.072 -79.570 44.513
126808 101.105 -79.589 44.500
126809 101.137 -79.608 44.486
126810 101.170 -79.627 44.473
126811 101.202 -79.646 44.458
126812 101.235 -79.666 44.443
126813 101.267 -79.685 44.427
126814 101.314 -79.708 44.412
126815 101.392 -79.733 44.396
126816 101.444 -79.739 44.379
126817 101.496 -79.745 44.363
126818 101.523 -79.758 44.347
126819 101.549 -79.782 44.331
126820 101.584 -79.797 44.314
126821 101.600 -79.803 44.296
126822 101.617 -79.808 44.278
127115 93.870 -77.858 53.179
127116 93.883 -77.853 53.160
127117 93.896 -77.849 53.140
127118 93.908 -77.845 53.121
127119 93.921 -77.840 53.102
127120 93.934 -77.837 53.082
127121 93.947 -77.832 53.063
127122 93.960 -77.827 53.045
127123 93.973 -77.824 53.026
127124 93.985 -77.819 53.007
127125 93.982 -77.818 52.987
127126 93.980 -77.817 52.967
127127 93.978 -77.816 52.947
127128 93.975 -77.815 52.928
127129 93.973 -77.814 52.908
127130 93.970 -77.813 52.888
127131 93.968 -77.812 52.867
127132 93.968 -77.815 52.847
127133 93.971 -77.826 52.826
127134 93.976 -77.835 52.805
127135 93.992 -77.840 52.783
127136 94.008 -77.843 52.762
127137 94.024 -77.848 52.740
127138 94.040 -77.852 52.718
127139 94.055 -77.856 52.695
127140 94.066 -77.856 52.673
127141 94.067 -77.848 52.651
127142 94.068 -77.840 52.630
127143 94.069 -77.832 52.607
127144 94.070 -77.824 52.584
127145 94.104 -77.809 52.562
127146 94.129 -77.800 52.539
127147 94.154 -77.791 52.517
127148 94.179 -77.781 52.495
127149 94.204 -77.772 52.473
127150 94.228 -77.763 52.453
127151 94.254 -77.754 52.432
127152 94.278 -77.745 52.411
127153 94.303 -77.736 52.390
127154 94.328 -77.727 52.369
127155 94.353 -77.718 52.349
127156 94.378 -77.708 52.330
127157 94.403 -77.699 52.312
127158 94.425 -77.692 52.292
127159 94.442 -77.691 52.274
127160 94.459 -77.689 52.257
127161 94.476 -77.687 52.241
127162 94.505 -77.659 52.226
127163 94.506 -77.666 52.212
127164 94.504 -77.676 52.198
127165 94.511 -77.678 52.184
127166 94.532 -77.670 52.167
127167 94.553 -77.661 52.149
127168 94.574 -77.652 52.129
127169 94.603 -77.642 52.109
127170 94.640 -77.629 52.087
127171 94.679 -77.615 52.065
127172 94.695 -77.613 52.041
127173 94.663 -77.611 52.018
127174 94.632 -77.610 51.995
127175 94.616 -77.615 51.972
127176 94.615 -77.629 51.948
127177 94.627 -77.669 51.923
127178 94.644 -77.700 51.899
127179 94.660 -77.686 51.873
127180 94.674 -77.673 51.846
127181 94.689 -77.659 51.819
127182 94.702 -77.650 51.792
127183 94.708 -77.656 51.763
127184 94.714 -77.663 51.732
127185 94.727 -77.670 51.699
127186 94.749 -77.677 51.666
127187 94.771 -77.683 51.631
127188 94.793 -77.690 51.598
127189 94.789 -77.671 51.565
127190 94.784 -77.652 51.531
127191 94.780 -77.633 51.498
127192 94.786 -77.620 51.465
127193 94.813 -77.635 51.428
127194 94.834 -77.666 51.391
127195 94.874 -77.705 51.353
127196 94.909 -77.727 51.313
127197 94.943 -77.751 51.274
127198 94.938 -77.743 51.235
127199 94.934 -77.736 51.197
127200 94.939 -77.719 51.158
127201 94.968 -77.689 51.118
127202 95.012 -77.665 51.078
127203 95.084 -77.655 51.036
127204 95.082 -77.683 50.997
127205 95.046 -77.729 50.956
127206 95.048 -77.746 50.915
127207 95.075 -77.714 50.874
127208 95.095 -77.715 50.834
127209 95.115 -77.715 50.794
127210 95.134 -77.716 50.754
127211 95.154 -77.716 50.716
127212 95.173 -77.717 50.678
127213 95.193 -77.717 50.639
127214 95.212 -77.718 50.600
127215 95.218 -77.714 50.563
127216 95.224 -77.710 50.526
127217 95.230 -77.706 50.489
127218 95.235 -77.703 50.454
127219 95.243 -77.698 50.420
127220 95.268 -77.685 50.385
127221 95.294 -77.672 50.351
127222 95.318 -77.660 50.317
127223 95.343 -77.647 50.284
127224 95.369 -77.634 50.251
127225 95.386 -77.623 50.220
127226 95.394 -77.615 50.189
127227 95.401 -77.607 50.158
127228 95.409 -77.599 50.126
127229 95.428 -77.567 50.096
127230 95.469 -77.541 50.066
127231 95.514 -77.520 50.036
127232 95.559 -77.500 50.007
127233 95.604 -77.479 49.978
127234 95.644 -77.461 49.950
127235 95.643 -77.482 49.922
127236 95.625 -77.506 49.893
127237 95.608 -77.530 49.865
127238 95.590 -77.555 49.838
127239 95.580 -77.563 49.810
127240 95.603 -77.578 49.780
127241 95.627 -77.593 49.750
127242 95.651 -77.608 49.719
127243 95.682 -77.592 49.688
127244 95.684 -77.604 49.656
127245 95.649 -77.633 49.625
127246 95.641 -77.644 49.595
127247 95.650 -77.645 49.565
127248 95.659 -77.645 49.536
127249 95.668 -77.646 49.508
127250 95.677 -77.647 49.481
127251 95.685 -77.648 49.454
127252 95.694 -77.649 49.428
127253 95.703 -77.649 49.401
127254 95.712 -77.650 49.376
127255 95.720 -77.655 49.348
127256 95.728 -77.663 49.318
127257 95.735 -77.670 49.286
127258 95.742 -77.677 49.253
127259 95.751 -77.684 49.217
127260 95.768 -77.691 49.179
127261 95.785 -77.699 49.140
127262 95.803 -77.706 49.103
127263 95.824 -77.717 49.066
127264 95.850 -77.729 49.028
127265 95.875 -77.742 48.992
127266 95.898 -77.751 48.955
127267 95.921 -77.759 48.918
127268 95.944 -77.768 48.881
127269 95.967 -77.776 48.846
127270 95.989 -77.785 48.811
127271 96.012 -77.793 48.776
127272 96.035 -77.802 48.740
127273 96.057 -77.810 48.704
127274 96.078 -77.822 48.669
127275 96.098 -77.835 48.632
127276 96.118 -77.847 48.598
127277 96.138 -77.860 48.564
127278 96.146 -77.858 48.530
127279 96.152 -77.855 48.498
127280 96.159 -77.852 48.467
127281 96.156 -77.867 48.437
127282 96.154 -77.881 48.407
127283 96.161 -77.880 48.379
127284 96.168 -77.879 48.351
127285 96.174 -77.878 48.323
127286 96.181 -77.877 48.294
127287 96.188 -77.876 48.266
127288 96.212 -77.896 48.237
127289 96.238 -77.921 48.209
127290 96.265 -77.946 48.184
127291 96.292 -77.971 48.160
127292 96.319 -77.996 48.137
127293 96.343 -78.018 48.116
127294 96.346 -78.012 48.094
127295 96.348 -78.007 48.071
127296 96.345 -78.003 48.048
127297 96.342 -78.000 48.027
127298 96.340 -77.996 48.007
127299 96.345 -77.995 47.987
127300 96.350 -77.993 47.966
127301 96.356 -77.990 47.944
127302 96.361 -77.988 47.924
127303 96.386 -77.992 47.904
127304 96.414 -77.996 47.882
127305 96.433 -78.003 47.859
127306 96.443 -78.013 47.836
127307 96.452 -78.023 47.813
127308 96.462 -78.034 47.787
127309 96.471 -78.044 47.760
127310 96.481 -78.053 47.732
127311 96.486 -78.047 47.705
127312 96.489 -78.031 47.678
127313 96.493 -78.015 47.650
127314 96.496 -77.999 47.623
127315 96.499 -77.983 47.598
127316 96.502 -77.967 47.572
127317 96.529 -77.967 47.547
127318 96.561 -77.971 47.524
127319 96.595 -77.978 47.500
127320 96.628 -77.983 47.476
127321 96.643 -78.003 47.452
127322 96.648 -78.031 47.427
127323 96.653 -78.058 47.404
127324 96.658 -78.084 47.381
127325 96.669 -78.095 47.357
127326 96.679 -78.102 47.333
127327 96.693 -78.111 47.309
127328 96.707 -78.122 47.284
127329 96.721 -78.131 47.259
127330 96.735 -78.142 47.233
127331 96.748 -78.152 47.206
127332 96.762 -78.163 47.178
127333 96.776 -78.172 47.153
127334 96.790 -78.183 47.127
127335 96.804 -78.193 47.101
127336 96.815 -78.208 47.077
127337 96.825 -78.226 47.052
127338 96.843 -78.238 47.025
127339 96.862 -78.250 46.997
127340 96.880 -78.263 46.970
127341 96.898 -78.274 46.941
127342 96.916 -78.286 46.914
127343 96.934 -78.298 46.889
127344 96.952 -78.310 46.867
127345 96.970 -78.322 46.844
127346 96.990 -78.335 46.822
127347 97.017 -78.347 46.797
127348 97.045 -78.361 46.773
127349 97.075 -78.376 46.749
127350 97.103 -78.389 46.725
127351 97.133 -78.404 46.700
127352 97.166 -78.415 46.676
127353 97.208 -78.421 46.652
127354 97.249 -78.426 46.630
127355 97.291 -78.432 46.607
127356 97.332 -78.438 46.585
127357 97.374 -78.445 46.564
127358 97.415 -78.438 46.543
127359 97.457 -78.421 46.524
127360 97.513 -78.414 46.505
127361 97.555 -78.416 46.490
127362 97.578 -78.420 46.474
127363 97.601 -78.425 46.458
127364 97.625 -78.430 46.442
127365 97.648 -78.435 46.423
127366 97.671 -78.439 46.404
127367 97.706 -78.448 46.385
127368 97.747 -78.461 46.366
127369 97.806 -78.488 46.348
127370 97.847 -78.506 46.329
127371 97.855 -78.475 46.311
127372 97.884 -78.467 46.293
127373 97.929 -78.474 46.275
127374 97.973 -78.482 46.258
127375 98.017 -78.493 46.242
127376 98.050 -78.517 46.226
127377 98.099 -78.494 46.210
127378 98.132 -78.473 46.194
127379 98.157 -78.451 46.179
127380 98.150 -78.442 46.165
127381 98.132 -78.438 46.150
127382 98.114 -78.437 46.136
127383 98.131 -78.453 46.122
127384 98.154 -78.464 46.108
127385 98.176 -78.475 46.094
127386 98.170 -78.465 46.078
127387 98.165 -78.455 46.063
127388 98.180 -78.444 46.047
127389 98.218 -78.438 46.030
127390 98.255 -78.434 46.013
127391 98.293 -78.429 45.996
127392 98.331 -78.424 45.979
127393 98.338 -78.416 45.961
127394 98.319 -78.397 45.941
127395 98.318 -78.399 45.922
127396 98.319 -78.403 45.902
127397 98.331 -78.407 45.882
127398 98.343 -78.412 45.862
127399 98.361 -78.426 45.841
127400 98.385 -78.448 45.820
127401 98.401 -78.463 45.800
127402 98.414 -78.476 45.780
127403 98.426 -78.489 45.759
127404 98.437 -78.504 45.738
127405 98.436 -78.521 45.718
127406 98.451 -78.526 45.698
127407 98.464 -78.532 45.678
127408 98.477 -78.538 45.659
127409 98.489 -78.544 45.640
127410 98.502 -78.550 45.620
127411 98.514 -78.555 45.599
127412 98.526 -78.561 45.579
127413 98.539 -78.568 45.559
127414 98.555 -78.578 45.539
127415 98.566 -78.575 45.520
127416 98.573 -78.554 45.502
127417 98.582 -78.545 45.484
127418 98.594 -78.547 45.465
127419 98.606 -78.557 45.448
127420 98.621 -78.555 45.430
127421 98.645 -78.542 45.412
127422 98.674 -78.526 45.395
127423 98.703 -78.509 45.378
127424 98.732 -78.510 45.361
127425 98.766 -78.519 45.344
127426 98.798 -78.527 45.328
127427 98.822 -78.530 45.312
127428 98.814 -78.513 45.296
127429 98.806 -78.495 45.281
127430 98.806 -78.496 45.265
127431 98.826 -78.535 45.249
127432 98.846 -78.573 45.234
127433 98.868 -78.611 45.219
127434 98.908 -78.623 45.204
127435 98.917 -78.627 45.190
127436 98.979 -78.604 45.176
127437 99.041 -78.580 45.164
127438 99.075 -78.576 45.152
127439 99.094 -78.581 45.139
127440 99.113 -78.586 45.126
127441 99.132 -78.592 45.113
127442 99.151 -78.597 45.101
127443 99.170 -78.601 45.090
127444 99.190 -78.607 45.078
127445 99.226 -78.604 45.067
127446 99.217 -78.639 45.056
127447 99.211 -78.686 45.043
127448 99.194 -78.701 45.031
127449 99.177 -78.715 45.019
127450 99.218 -78.695 45.006
127451 99.251 -78.689 44.993
127452 99.305 -78.708 44.980
127453 99.360 -78.727 44.967
127454 99.414 -78.745 44.956
127455 99.469 -78.764 44.942
127456 99.524 -78.783 44.930
127457 99.578 -78.801 44.918
127458 99.633 -78.820 44.905
127459 99.688 -78.839 44.892
127460 99.722 -78.849 44.879
127461 99.741 -78.871 44.867
127462 99.771 -78.910 44.854
127463 99.791 -78.933 44.842
127464 99.811 -78.955 44.831
127465 99.831 -78.978 44.821
127466 99.850 -78.999 44.812
127467 99.863 -79.009 44.803
127468 99.876 -79.018 44.795
127469 99.890 -79.028 44.788
127470 99.903 -79.038 44.782
127471 99.916 -79.048 44.777
127472 99.929 -79.058 44.771
127473 99.942 -79.068 44.766
127474 99.955 -79.078 44.761
127475 99.983 -79.081 44.756
127476 100.022 -79.093 44.752
127477 100.040 -79.114 44.748
127478 100.060 -79.135 44.744
127479 100.089 -79.153 44.740
127480 100.118 -79.172 44.736
127481 100.148 -79.190 44.732
127482 100.177 -79.208 44.727
127483 100.207 -79.226 44.723
127484 100.235 -79.238 44.719
127485 100.262 -79.244 44.714
127486 100.289 -79.249 44.709
127487 100.317 -79.255 44.704
127488 100.344 -79.260 44.698
127489 100.371 -79.266 44.691
127490 100.400 -79.272 44.684
127491 100.438 -79.281 44.677
127513 101.232 -79.655 44.440
127514 101.264 -79.675 44.425
127515 101.338 -79.705 44.409
127516 101.397 -79.717 44.393
127517 101.450 -79.723 44.376
127518 101.498 -79.730 44.360
127519 101.522 -79.743 44.344
127520 101.548 -79.768 44.327
127815 93.857 -77.851 53.210
127816 93.870 -77.846 53.191
127817 93.884 -77.842 53.172
127818 93.896 -77.838 53.152
127819 93.909 -77.834 53.132
127820 93.922 -77.830 53.113
127821 93.935 -77.825 53.094
127822 93.947 -77.821 53.075
127823 93.961 -77.817 53.057
127824 93.973 -77.812 53.038
127825 93.985 -77.808 53.019
127826 93.983 -77.807 52.999
127827 93.980 -77.806 52.980
127828 93.978 -77.805 52.961
127829 93.976 -77.804 52.941
127830 93.973 -77.802 52.921
127831 93.971 -77.802 52.902
127832 93.968 -77.800 52.881
127833 93.967 -77.802 52.860
127834 93.971 -77.812 52.840
127835 93.979 -77.820 52.818
127836 93.995 -77.824 52.796
127837 94.011 -77.828 52.775
127838 94.027 -77.833 52.754
127839 94.043 -77.837 52.732
127840 94.059 -77.841 52.709
127841 94.067 -77.840 52.687
127842 94.068 -77.831 52.665
127843 94.069 -77.823 52.644
127844 94.070 -77.815 52.621
127845 94.072 -77.807 52.598
127846 94.105 -77.793 52.576
127847 94.130 -77.783 52.554
127848 94.155 -77.774 52.532
127849 94.180 -77.765 52.510
127850 94.205 -77.755 52.488
127851 94.230 -77.746 52.468
127852 94.255 -77.737 52.447
127853 94.280 -77.728 52.426
127854 94.305 -77.719 52.405
127855 94.330 -77.710 52.385
127856 94.355 -77.700 52.365
127857 94.376 -77.695 52.346
127858 94.393 -77.693 52.328
127859 94.410 -77.691 52.309
127860 94.427 -77.689 52.292
127861 94.444 -77.688 52.275
127862 94.464 -77.679 52.260
127863 94.494 -77.645 52.245
127864 94.492 -77.655 52.231
127865 94.490 -77.664 52.217
127866 94.489 -77.674 52.204
127867 94.498 -77.675 52.188
127868 94.520 -77.666 52.171
127869 94.554 -77.654 52.153
127870 94.592 -77.641 52.132
127871 94.630 -77.627 52.111
127872 94.668 -77.614 52.090
127873 94.702 -77.597 52.066
127874 94.728 -77.596 52.043
127875 94.636 -77.613 52.020
127876 94.602 -77.628 51.997
127877 94.616 -77.672 51.974
127878 94.634 -77.722 51.948
127879 94.650 -77.736 51.922
127880 94.665 -77.722 51.896
127881 94.680 -77.708 51.869
127882 94.695 -77.695 51.843
127883 94.709 -77.681 51.814
127884 94.724 -77.667 51.785
127885 94.738 -77.658 51.754
127886 94.744 -77.664 51.722
127887 94.757 -77.671 51.688
127888 94.779 -77.677 51.654
127889 94.791 -77.674 51.622
127890 94.787 -77.655 51.590
127891 94.783 -77.636 51.556
127892 94.779 -77.617 51.524
127893 94.777 -77.600 51.491
127894 94.804 -77.618 51.456
127895 94.836 -77.654 51.418
127896 94.872 -77.675 51.379
127897 94.906 -77.696 51.340
127898 94.941 -77.719 51.301
127899 94.973 -77.745 51.262
127900 94.952 -77.745 51.224
127901 94.980 -77.714 51.184
127902 95.008 -77.683 51.143
127903 95.037 -77.653 51.102
127904 95.065 -77.622 51.061
127905 95.080 -77.634 51.021
127906 95.073 -77.672 50.980
127907 95.060 -77.731 50.939
127908 95.077 -77.743 50.897
127909 95.108 -77.715 50.855
127910 95.127 -77.716 50.815
127911 95.147 -77.716 50.775
127912 95.167 -77.717 50.736
127913 95.186 -77.717 50.697
127914 95.205 -77.718 50.658
127915 95.212 -77.714 50.619
127916 95.217 -77.710 50.581
127917 95.223 -77.706 50.544
127918 95.229 -77.703 50.507
127919 95.234 -77.699 50.471
127920 95.251 -77.690 50.436
127921 95.276 -77.677 50.401
127922 95.301 -77.664 50.366
127923 95.326 -77.652 50.331
127924 95.351 -77.639 50.298
127925 95.376 -77.627 50.265
127926 95.398 -77.614 50.233
127927 95.405 -77.606 50.201
127928 95.412 -77.599 50.170
127929 95.420 -77.590 50.138
127930 95.447 -77.543 50.107
127931 95.492 -77.523 50.077
127932 95.537 -77.502 50.047
127933 95.582 -77.482 50.017
127934 95.619 -77.464 49.988
127935 95.635 -77.453 49.959
127936 95.672 -77.462 49.931
127937 95.692 -77.491 49.901
127938 95.674 -77.515 49.873
127939 95.658 -77.535 49.845
127940 95.645 -77.541 49.818
127941 95.633 -77.546 49.789
127942 95.641 -77.557 49.760
127943 95.668 -77.558 49.730
127944 95.679 -77.584 49.699
127945 95.665 -77.613 49.667
127946 95.658 -77.623 49.636
127947 95.659 -77.628 49.605
127948 95.660 -77.633 49.575
127949 95.661 -77.638 49.546
127950 95.662 -77.643 49.518
127951 95.670 -77.645 49.491
127952 95.678 -77.645 49.464
127953 95.687 -77.646 49.437
127954 95.696 -77.647 49.411
127955 95.705 -77.648 49.384
127956 95.714 -77.649 49.355
127957 95.722 -77.652 49.326
127958 95.730 -77.659 49.294
127959 95.737 -77.666 49.262
127960 95.745 -77.674 49.227
127961 95.763 -77.680 49.189
127962 95.780 -77.688 49.150
127963 95.802 -77.698 49.112
127964 95.827 -77.711 49.075
127965 95.852 -77.724 49.037
127966 95.877 -77.736 49.001
127967 95.900 -77.744 48.965
127968 95.922 -77.752 48.928
127969 95.945 -77.761 48.891
127970 95.968 -77.769 48.855
127971 95.991 -77.777 48.820
127972 96.014 -77.786 48.785
127973 96.036 -77.794 48.749
127974 96.059 -77.802 48.714
127975 96.082 -77.811 48.678
127976 96.103 -77.821 48.642
127977 96.123 -77.834 48.608
127978 96.144 -77.846 48.574
127979 96.158 -77.852 48.540
127980 96.164 -77.849 48.507
127981 96.170 -77.848 48.475
127982 96.164 -77.868 48.444
127983 96.158 -77.887 48.413
127984 96.165 -77.887 48.385
127985 96.172 -77.887 48.357
127986 96.179 -77.886 48.330
127987 96.185 -77.884 48.301
127988 96.198 -77.890 48.272
127989 96.225 -77.915 48.243
127990 96.252 -77.940 48.215
127991 96.278 -77.965 48.190
127992 96.306 -77.990 48.166
127993 96.324 -78.004 48.143
127994 96.326 -77.999 48.120
127995 96.329 -77.993 48.096
127996 96.331 -77.988 48.073
127997 96.328 -77.985 48.050
127998 96.326 -77.981 48.029
127999 96.323 -77.978 48.010
128000 96.320 -77.975 47.990
128001 96.318 -77.971 47.969
128002 96.322 -77.969 47.948
128003 96.340 -77.971 47.927
128004 96.368 -77.975 47.908
128005 96.396 -77.979 47.888
128006 96.424 -77.983 47.867
128007 96.450 -77.988 47.845
128008 96.460 -77.998 47.822
128009 96.469 -78.009 47.796
128010 96.479 -78.018 47.769
128011 96.488 -78.028 47.742
128012 96.496 -78.029 47.717
128013 96.499 -78.013 47.689
128014 96.502 -77.997 47.662
128015 96.505 -77.981 47.636
128016 96.508 -77.965 47.611
128017 96.528 -77.961 47.586
128018 96.559 -77.963 47.561
128019 96.590 -77.965 47.537
128020 96.620 -77.968 47.514
128021 96.639 -77.970 47.490
128022 96.651 -77.978 47.467
128023 96.661 -77.999 47.442
128024 96.666 -78.025 47.418
128025 96.671 -78.051 47.395
128026 96.676 -78.077 47.372
128027 96.687 -78.092 47.348
128028 96.700 -78.102 47.324
128029 96.715 -78.112 47.299
128030 96.728 -78.122 47.272
128031 96.742 -78.133 47.245
128032 96.756 -78.143 47.218
128033 96.769 -78.153 47.192
128034 96.783 -78.163 47.165
128035 96.797 -78.174 47.139
128036 96.811 -78.184 47.114
128037 96.823 -78.196 47.089
128038 96.840 -78.209 47.064
128039 96.859 -78.221 47.037
128040 96.877 -78.233 47.009
128041 96.895 -78.245 46.981
128042 96.912 -78.259 46.953
128043 96.930 -78.272 46.926
128044 96.948 -78.286 46.903
128045 96.965 -78.300 46.882
128046 96.983 -78.313 46.861
128047 96.999 -78.322 46.840
128048 97.016 -78.332 46.816
128049 97.045 -78.345 46.793
128050 97.074 -78.360 46.769
128051 97.103 -78.374 46.745
128052 97.132 -78.388 46.720
128053 97.169 -78.397 46.696
128054 97.211 -78.403 46.672
128055 97.252 -78.409 46.649
128056 97.294 -78.415 46.626
128057 97.335 -78.421 46.603
128058 97.377 -78.427 46.582
128059 97.418 -78.432 46.562
128060 97.460 -78.426 46.543
128061 97.501 -78.410 46.526
128062 97.562 -78.406 46.511
128063 97.590 -78.410 46.495
128064 97.613 -78.415 46.479
128065 97.636 -78.420 46.462
128066 97.660 -78.424 46.444
128067 97.683 -78.429 46.425
128068 97.706 -78.434 46.406
128069 97.730 -78.438 46.387
128070 97.779 -78.457 46.368
128071 97.805 -78.449 46.349
128072 97.847 -78.463 46.330
128073 97.886 -78.453 46.312
128074 97.931 -78.460 46.294
128075 97.975 -78.469 46.277
128076 98.018 -78.480 46.261
128077 98.051 -78.507 46.244
128078 98.089 -78.493 46.229
128079 98.123 -78.471 46.213
128080 98.136 -78.455 46.198
128081 98.117 -78.451 46.183
128082 98.109 -78.440 46.168
128083 98.110 -78.423 46.154
128084 98.113 -78.422 46.141
128085 98.119 -78.427 46.127
128086 98.142 -78.438 46.111
128087 98.159 -78.445 46.096
128088 98.153 -78.435 46.080
128089 98.209 -78.413 46.064
128090 98.271 -78.397 46.048
128091 98.306 -78.392 46.031
128092 98.309 -78.389 46.014
128093 98.311 -78.386 45.996
128094 98.314 -78.384 45.977
128095 98.315 -78.382 45.958
128096 98.314 -78.386 45.938
128097 98.313 -78.390 45.918
128098 98.319 -78.394 45.898
128099 98.331 -78.398 45.878
128100 98.351 -78.413 45.858
128101 98.374 -78.436 45.838
128102 98.395 -78.456 45.817
128103 98.408 -78.469 45.796
128104 98.421 -78.482 45.775
128105 98.421 -78.502 45.754
128106 98.425 -78.515 45.733
128107 98.440 -78.520 45.713
128108 98.455 -78.526 45.694
128109 98.467 -78.532 45.674
128110 98.480 -78.538 45.654
128111 98.492 -78.544 45.634
128112 98.504 -78.549 45.614
128113 98.518 -78.557 45.593
128114 98.533 -78.567 45.573
128115 98.549 -78.577 45.553
128116 98.557 -78.560 45.533
128117 98.563 -78.539 45.515
128118 98.571 -78.522 45.496
128119 98.582 -78.535 45.477
128120 98.593 -78.556 45.459
128121 98.604 -78.577 45.441
128122 98.624 -78.576 45.423
128123 98.663 -78.545 45.405
128124 98.704 -78.510 45.387
128125 98.739 -78.494 45.370
128126 98.764 -78.505 45.353
128127 98.788 -78.517 45.336
128128 98.821 -78.526 45.320
128129 98.846 -78.529 45.304
128130 98.838 -78.511 45.288
128131 98.837 -78.498 45.273
128132 98.846 -78.532 45.257
128133 98.855 -78.567 45.242
128134 98.864 -78.602 45.228
128135 98.875 -78.634 45.213
128136 98.937 -78.611 45.199
128137 99.000 -78.587 45.186
128138 99.062 -78.564 45.174
128139 99.082 -78.568 45.162
128140 99.101 -78.573 45.150
128141 99.121 -78.579 45.136
128142 99.140 -78.584 45.124
128143 99.159 -78.589 45.112
128144 99.178 -78.594 45.100
128145 99.197 -78.599 45.087
128146 99.217 -78.604 45.075
128147 99.253 -78.602 45.064
128148 99.233 -78.653 45.052
128149 99.217 -78.667 45.041
128150 99.200 -78.683 45.029
128151 99.208 -78.681 45.016
128152 99.249 -78.685 45.003
128153 99.304 -78.704 44.989
128154 99.358 -78.723 44.976
128155 99.413 -78.741 44.964
128156 99.468 -78.760 44.950
128157 99.522 -78.779 44.937
128158 99.577 -78.797 44.924
128159 99.632 -78.816 44.912
128160 99.686 -78.835 44.899
128161 99.738 -78.852 44.886
128162 99.755 -78.868 44.873
128163 99.781 -78.903 44.861
128164 99.801 -78.926 44.849
128165 99.822 -78.949 44.838
128166 99.838 -78.965 44.827
128167 99.852 -78.975 44.818
128168 99.865 -78.985 44.809
128169 99.878 -78.995 44.801
128170 99.891 -79.005 44.794
128171 99.904 -79.015 44.788
128172 99.917 -79.024 44.782
128173 99.930 -79.034 44.776
128174 99.943 -79.044 44.771
128175 99.957 -79.054 44.765
128176 99.984 -79.059 44.761
128177 100.023 -79.071 44.756
128178 100.033 -79.095 44.752
128179 100.062 -79.113 44.748
128180 100.091 -79.131 44.744
128181 100.121 -79.150 44.740
128182 100.150 -79.168 44.735
128183 100.180 -79.186 44.731
128184 100.209 -79.205 44.726
128185 100.238 -79.219 44.722
128186 100.265 -79.224 44.717
128187 100.292 -79.229 44.712
128188 100.320 -79.234 44.706
128189 100.347 -79.241 44.700
128190 100.374 -79.246 44.694
128191 100.406 -79.253 44.687
128478 93.409 -78.125 53.895
128479 93.418 -78.122 53.882
128515 93.845 -77.844 53.239
128516 93.858 -77.840 53.220
128517 93.871 -77.835 53.202
128518 93.884 -77.831 53.183
128519 93.897 -77.827 53.164
128520 93.910 -77.822 53.144
128521 93.922 -77.818 53.126
128522 93.935 -77.814 53.107
128523 93.948 -77.809 53.088
128524 93.961 -77.805 53.070
128525 93.974 -77.801 53.051
128526 93.986 -77.797 53.032
128527 93.983 -77.796 53.013
128528 93.981 -77.795 52.994
128529 93.978 -77.793 52.974
128530 93.976 -77.793 52.955
128531 93.974 -77.791 52.936
128532 93.971 -77.790 52.915
128533 93.969 -77.790 52.895
128534 93.967 -77.789 52.874
128535 93.970 -77.799 52.853
128536 93.983 -77.805 52.832
128537 93.998 -77.809 52.810
128538 94.014 -77.813 52.789
128539 94.030 -77.818 52.768
128540 94.046 -77.822 52.746
128541 94.062 -77.826 52.724
128542 94.068 -77.823 52.702
128543 94.070 -77.815 52.679
128544 94.071 -77.806 52.657
128545 94.072 -77.799 52.635
128546 94.073 -77.790 52.612
128547 94.107 -77.775 52.591
128548 94.132 -77.766 52.569
128549 94.156 -77.757 52.547
128550 94.182 -77.748 52.525
128551 94.206 -77.739 52.503
128552 94.231 -77.730 52.483
128553 94.257 -77.721 52.462
128554 94.281 -77.711 52.441
128555 94.307 -77.702 52.420
128556 94.326 -77.697 52.401
128557 94.343 -77.696 52.382
128558 94.360 -77.694 52.363
128559 94.378 -77.692 52.345
128560 94.395 -77.690 52.328
128561 94.412 -77.688 52.311
128562 94.429 -77.686 52.295
128563 94.458 -77.657 52.280
128564 94.480 -77.633 52.266
128565 94.478 -77.643 52.252
128566 94.477 -77.653 52.239
128567 94.475 -77.663 52.225
128568 94.475 -77.673 52.211
128569 94.488 -77.660 52.196
128570 94.504 -77.640 52.178
128571 94.520 -77.620 52.158
128572 94.536 -77.600 52.137
128573 94.552 -77.580 52.116
128574 94.699 -77.587 52.093
128575 94.680 -77.607 52.071
128576 94.597 -77.627 52.050
128577 94.604 -77.675 52.027
128578 94.622 -77.725 52.003
128579 94.640 -77.776 51.977
128580 94.655 -77.771 51.950
128581 94.670 -77.757 51.924
128582 94.685 -77.744 51.897
128583 94.700 -77.730 51.869
128584 94.715 -77.717 51.841
128585 94.730 -77.703 51.812
128586 94.745 -77.689 51.781
128587 94.759 -77.676 51.748
128588 94.773 -77.664 51.714
128589 94.787 -77.671 51.680
128590 94.789 -77.658 51.647
128591 94.785 -77.639 51.615
128592 94.781 -77.620 51.584
128593 94.777 -77.602 51.553
128594 94.773 -77.583 51.521
128595 94.799 -77.602 51.484
128596 94.834 -77.624 51.446
128597 94.868 -77.644 51.408
128598 94.904 -77.665 51.369
128599 94.938 -77.688 51.331
128600 94.971 -77.713 51.293
128601 94.995 -77.700 51.253
128602 95.018 -77.672 51.211
128603 95.041 -77.644 51.170
128604 95.064 -77.616 51.128
128605 95.087 -77.588 51.087
128606 95.099 -77.588 51.046
128607 95.090 -77.645 51.005
128608 95.076 -77.705 50.964
128609 95.088 -77.729 50.922
128610 95.112 -77.735 50.879
128611 95.140 -77.716 50.837
128612 95.160 -77.716 50.797
128613 95.179 -77.717 50.757
128614 95.199 -77.717 50.718
128615 95.205 -77.714 50.679
128616 95.211 -77.710 50.639
128617 95.216 -77.706 50.601
128618 95.222 -77.703 50.563
128619 95.228 -77.699 50.526
128620 95.233 -77.696 50.490
128621 95.258 -77.683 50.454
128622 95.283 -77.670 50.419
128623 95.308 -77.657 50.382
128624 95.333 -77.645 50.347
128625 95.358 -77.632 50.313
128626 95.383 -77.619 50.280
128627 95.408 -77.606 50.247
128628 95.416 -77.598 50.215
128629 95.424 -77.589 50.183
128630 95.439 -77.567 50.151
128631 95.470 -77.525 50.120
128632 95.515 -77.504 50.089
128633 95.560 -77.484 50.058
128634 95.594 -77.467 50.028
128635 95.610 -77.457 49.999
128636 95.627 -77.445 49.969
128637 95.644 -77.436 49.940
128638 95.706 -77.469 49.910
128639 95.740 -77.500 49.881
128640 95.726 -77.514 49.854
128641 95.713 -77.519 49.826
128642 95.703 -77.530 49.797
128643 95.694 -77.548 49.769
128644 95.685 -77.566 49.741
128645 95.680 -77.591 49.710
128646 95.678 -77.602 49.679
128647 95.681 -77.605 49.647
128648 95.681 -77.610 49.616
128649 95.682 -77.614 49.586
128650 95.683 -77.620 49.557
128651 95.684 -77.625 49.529
128652 95.685 -77.630 49.500
128653 95.686 -77.635 49.474
128654 95.687 -77.640 49.447
128655 95.689 -77.644 49.420
128656 95.698 -77.645 49.392
128657 95.707 -77.645 49.364
128658 95.716 -77.646 49.334
128659 95.725 -77.649 49.303
128660 95.732 -77.655 49.271
128661 95.740 -77.662 49.237
128662 95.757 -77.670 49.199
128663 95.778 -77.680 49.160
128664 95.804 -77.692 49.122
128665 95.829 -77.705 49.084
128666 95.854 -77.718 49.047
128667 95.879 -77.728 49.011
128668 95.901 -77.736 48.974
128669 95.924 -77.745 48.938
128670 95.947 -77.753 48.901
128671 95.969 -77.762 48.865
128672 95.992 -77.771 48.829
128673 96.015 -77.779 48.793
128674 96.038 -77.787 48.757
128675 96.060 -77.796 48.722
128676 96.083 -77.804 48.687
128677 96.106 -77.812 48.652
128678 96.129 -77.821 48.618
128679 96.149 -77.834 48.584
128680 96.169 -77.846 48.550
128681 96.176 -77.843 48.516
128682 96.178 -77.849 48.484
128683 96.172 -77.869 48.452
128684 96.166 -77.889 48.421
128685 96.169 -77.895 48.393
128686 96.176 -77.894 48.365
128687 96.183 -77.893 48.337
128688 96.190 -77.892 48.308
128689 96.211 -77.909 48.280
128690 96.238 -77.934 48.251
128691 96.265 -77.959 48.223
128692 96.292 -77.984 48.198
128693 96.304 -77.991 48.174
128694 96.307 -77.986 48.149
128695 96.309 -77.981 48.125
128696 96.311 -77.975 48.101
128697 96.314 -77.970 48.077
128698 96.311 -77.966 48.055
128699 96.309 -77.962 48.034
128700 96.306 -77.959 48.014
128701 96.303 -77.956 47.994
128702 96.301 -77.953 47.974
128703 96.303 -77.949 47.952
128704 96.325 -77.953 47.931
128705 96.350 -77.957 47.912
128706 96.378 -77.962 47.892
128707 96.406 -77.966 47.873
128708 96.435 -77.970 47.851
128709 96.463 -77.975 47.829
128710 96.477 -77.983 47.803
128711 96.486 -77.993 47.777
128712 96.496 -78.003 47.751
128713 96.506 -78.012 47.727
128714 96.509 -77.996 47.700
128715 96.512 -77.980 47.674
128716 96.515 -77.964 47.648
128717 96.528 -77.955 47.623
128718 96.559 -77.958 47.599
128719 96.589 -77.960 47.575
128720 96.615 -77.962 47.550
128721 96.625 -77.963 47.528
128722 96.637 -77.965 47.504
128723 96.647 -77.967 47.481
128724 96.662 -77.984 47.455
128725 96.676 -78.002 47.432
128726 96.685 -78.025 47.409
128727 96.690 -78.050 47.386
128728 96.695 -78.076 47.361
128729 96.709 -78.087 47.336
128730 96.722 -78.097 47.311
128731 96.736 -78.107 47.284
128732 96.750 -78.118 47.257
128733 96.763 -78.128 47.230
128734 96.777 -78.139 47.204
128735 96.791 -78.149 47.177
128736 96.805 -78.159 47.151
128737 96.822 -78.173 47.125
128738 96.833 -78.191 47.100
128739 96.851 -78.204 47.074
128740 96.869 -78.218 47.047
128741 96.887 -78.232 47.019
128742 96.904 -78.245 46.991
128743 96.922 -78.259 46.963
128744 96.940 -78.272 46.938
128745 96.957 -78.286 46.915
128746 96.973 -78.294 46.895
128747 96.987 -78.300 46.875
128748 97.004 -78.308 46.856
128749 97.021 -78.318 46.834
128750 97.045 -78.330 46.811
128751 97.074 -78.344 46.788
128752 97.103 -78.359 46.764
128753 97.132 -78.373 46.740
128754 97.172 -78.379 46.716
128755 97.214 -78.385 46.691
128756 97.255 -78.391 46.667
128757 97.297 -78.398 46.645
128758 97.338 -78.403 46.623
128759 97.380 -78.409 46.602
128760 97.421 -78.415 46.581
128761 97.463 -78.421 46.564
128762 97.504 -78.416 46.549
128763 97.548 -78.401 46.532
128764 97.602 -78.400 46.514
128765 97.625 -78.404 46.497
128766 97.648 -78.409 46.480
128767 97.671 -78.414 46.463
128768 97.695 -78.419 46.445
128769 97.718 -78.423 46.426
128770 97.741 -78.428 46.408
128771 97.770 -78.442 46.389
128772 97.831 -78.507 46.370
128773 97.869 -78.510 46.351
128774 97.906 -78.496 46.332
128775 97.943 -78.483 46.315
128776 97.981 -78.470 46.298
128777 98.017 -78.460 46.281
128778 98.050 -78.464 46.265
128779 98.082 -78.470 46.249
128780 98.114 -78.470 46.234
128781 98.115 -78.455 46.218
128782 98.117 -78.438 46.203
128783 98.119 -78.420 46.189
128784 98.120 -78.403 46.175
128785 98.123 -78.398 46.161
128786 98.126 -78.403 46.146
128787 98.130 -78.408 46.131
128788 98.133 -78.413 46.114
128789 98.144 -78.415 46.098
128790 98.217 -78.391 46.082
128791 98.269 -78.375 46.066
128792 98.301 -78.365 46.049
128793 98.307 -78.363 46.031
128794 98.313 -78.363 46.013
128795 98.319 -78.361 45.994
128796 98.327 -78.357 45.974
128797 98.315 -78.369 45.955
128798 98.310 -78.376 45.934
128799 98.309 -78.380 45.914
128800 98.319 -78.384 45.894
128801 98.340 -78.402 45.874
128802 98.364 -78.424 45.854
128803 98.387 -78.447 45.833
128804 98.402 -78.461 45.813
128805 98.415 -78.475 45.792
128806 98.404 -78.501 45.771
128807 98.415 -78.510 45.749
128808 98.430 -78.515 45.729
128809 98.445 -78.520 45.710
128810 98.458 -78.526 45.689
128811 98.470 -78.532 45.669
128812 98.483 -78.538 45.649
128813 98.497 -78.545 45.629
128814 98.512 -78.556 45.608
128815 98.527 -78.566 45.587
128816 98.540 -78.565 45.567
128817 98.547 -78.544 45.547
128818 98.554 -78.523 45.528
128819 98.557 -78.482 45.508
128820 98.568 -78.501 45.489
128821 98.581 -78.526 45.470
128822 98.598 -78.558 45.452
128823 98.632 -78.564 45.434
128824 98.674 -78.532 45.416
128825 98.715 -78.497 45.398
128826 98.752 -78.474 45.381
128827 98.777 -78.486 45.363
128828 98.801 -78.498 45.346
128829 98.826 -78.510 45.329
128830 98.851 -78.521 45.313
128831 98.869 -78.514 45.297
128832 98.869 -78.495 45.282
128833 98.877 -78.507 45.267
128834 98.886 -78.519 45.251
128835 98.895 -78.531 45.237
128836 98.903 -78.543 45.223
128837 98.958 -78.588 45.209
128838 99.020 -78.570 45.196
128839 99.070 -78.555 45.184
128840 99.089 -78.561 45.171
128841 99.109 -78.566 45.159
128842 99.128 -78.571 45.147
128843 99.147 -78.576 45.134
128844 99.166 -78.581 45.122
128845 99.186 -78.586 45.109
128846 99.205 -78.592 45.097
128847 99.224 -78.596 45.085
128848 99.243 -78.601 45.073
128849 99.256 -78.620 45.062
128850 99.239 -78.635 45.051
128851 99.223 -78.649 45.039
128852 99.206 -78.664 45.026
128853 99.248 -78.682 45.013
128854 99.302 -78.700 45.000
128855 99.357 -78.719 44.986
128856 99.412 -78.738 44.973
128857 99.466 -78.756 44.959
128858 99.521 -78.775 44.946
128859 99.575 -78.794 44.932
128860 99.630 -78.812 44.919
128861 99.685 -78.831 44.906
128862 99.739 -78.850 44.894
128863 99.768 -78.867 44.881
128864 99.792 -78.896 44.868
128865 99.812 -78.919 44.857
128866 99.827 -78.931 44.845
128867 99.840 -78.941 44.834
128868 99.853 -78.951 44.825
128869 99.866 -78.961 44.816
128870 99.879 -78.971 44.808
128871 99.893 -78.981 44.800
128872 99.906 -78.991 44.794
128873 99.919 -79.001 44.788
128874 99.932 -79.011 44.782
128875 99.945 -79.021 44.776
128876 99.959 -79.031 44.770
128877 99.985 -79.036 44.766
128878 100.015 -79.052 44.762
128879 100.034 -79.073 44.758
128880 100.064 -79.091 44.753
128881 100.093 -79.109 44.749
128882 100.123 -79.128 44.745
128883 100.152 -79.146 44.740
128884 100.182 -79.165 44.736
128885 100.211 -79.183 44.731
128886 100.240 -79.198 44.727
128887 100.267 -79.204 44.722
128888 100.295 -79.209 44.716
128889 100.322 -79.215 44.710
128890 100.349 -79.220 44.704
128891 100.376 -79.226 44.698
128892 100.413 -79.235 44.690
129175 93.340 -78.145 53.951
129176 93.363 -78.133 53.941
129177 93.387 -78.121 53.931
129178 93.399 -78.116 53.918
129179 93.409 -78.114 53.905
129180 93.419 -78.112 53.892
129181 93.428 -78.109 53.880
129182 93.438 -78.106 53.866
129183 93.444 -78.102 53.853
129184 93.445 -78.091 53.840
129215 93.833 -77.837 53.270
129216 93.846 -77.833 53.251
129217 93.859 -77.828 53.231
129218 93.871 -77.824 53.212
129219 93.885 -77.820 53.194
129220 93.897 -77.815 53.176
129221 93.910 -77.812 53.157
129222 93.923 -77.807 53.138
129223 93.936 -77.802 53.120
129224 93.949 -77.799 53.101
129225 93.962 -77.794 53.083
129226 93.975 -77.790 53.065
129227 93.986 -77.786 53.046
129228 93.984 -77.785 53.028
129229 93.981 -77.783 53.009
129230 93.979 -77.783 52.990
129231 93.977 -77.781 52.971
129232 93.974 -77.780 52.951
129233 93.972 -77.779 52.931
129234 93.969 -77.778 52.911
129235 93.967 -77.777 52.890
129236 93.970 -77.786 52.869
129237 93.986 -77.790 52.847
129238 94.001 -77.794 52.826
129239 94.017 -77.799 52.804
129240 94.033 -77.802 52.783
129241 94.049 -77.807 52.761
129242 94.065 -77.811 52.739
129243 94.070 -77.806 52.717
129244 94.071 -77.798 52.695
129245 94.072 -77.790 52.673
129246 94.073 -77.782 52.650
129247 94.074 -77.774 52.628
129248 94.108 -77.758 52.606
129249 94.133 -77.749 52.584
129250 94.158 -77.740 52.562
129251 94.183 -77.731 52.541
129252 94.208 -77.722 52.519
129253 94.233 -77.713 52.499
129254 94.258 -77.703 52.478
129255 94.277 -77.699 52.458
129256 94.294 -77.698 52.436
129257 94.311 -77.696 52.417
129258 94.328 -77.694 52.399
129259 94.345 -77.692 52.382
129260 94.362 -77.690 52.364
129261 94.380 -77.689 52.347
129262 94.397 -77.687 52.331
129263 94.417 -77.677 52.315
129264 94.452 -77.636 52.301
129265 94.466 -77.622 52.288
129266 94.465 -77.632 52.275
129267 94.463 -77.642 52.262
129268 94.465 -77.651 52.249
129269 94.473 -77.659 52.235
129270 94.485 -77.652 52.222
129271 94.501 -77.633 52.204
129272 94.517 -77.613 52.185
129273 94.533 -77.593 52.165
129274 94.590 -77.582 52.144
129275 94.666 -77.602 52.121
129276 94.582 -77.622 52.100
129277 94.591 -77.676 52.080
129278 94.611 -77.728 52.057
129279 94.628 -77.779 52.033
129280 94.646 -77.820 52.008
129281 94.661 -77.806 51.982
129282 94.676 -77.793 51.955
129283 94.690 -77.779 51.928
129284 94.705 -77.765 51.901
129285 94.720 -77.752 51.873
129286 94.735 -77.739 51.844
129287 94.750 -77.725 51.812
129288 94.765 -77.711 51.778
129289 94.780 -77.698 51.743
129290 94.786 -77.658 51.709
129291 94.803 -77.667 51.675
129292 94.799 -77.648 51.642
129293 94.794 -77.629 51.612
129294 94.790 -77.609 51.581
129295 94.786 -77.575 51.549
129296 94.805 -77.575 51.512
129297 94.831 -77.592 51.474
129298 94.866 -77.614 51.437
129299 94.901 -77.634 51.401
129300 94.936 -77.656 51.364
129301 94.968 -77.680 51.326
129302 95.001 -77.670 51.287
129303 95.024 -77.641 51.244
129304 95.047 -77.613 51.201
129305 95.069 -77.585 51.159
129306 95.076 -77.572 51.117
129307 95.078 -77.576 51.074
129308 95.105 -77.619 51.033
129309 95.092 -77.678 50.992
129310 95.084 -77.730 50.948
129311 95.124 -77.716 50.905
129312 95.147 -77.721 50.862
129313 95.172 -77.717 50.821
129314 95.192 -77.717 50.781
129315 95.199 -77.714 50.740
129316 95.204 -77.710 50.701
129317 95.210 -77.706 50.661
129318 95.215 -77.702 50.622
129319 95.221 -77.699 50.584
129320 95.227 -77.696 50.546
129321 95.241 -77.688 50.509
129322 95.266 -77.675 50.473
129323 95.290 -77.662 50.437
129324 95.316 -77.649 50.401
129325 95.341 -77.637 50.366
129326 95.366 -77.624 50.331
129327 95.391 -77.611 50.296
129328 95.416 -77.599 50.263
129329 95.427 -77.589 50.230
129330 95.435 -77.581 50.198
129331 95.458 -77.542 50.166
129332 95.493 -77.507 50.134
129333 95.539 -77.486 50.102
129334 95.569 -77.471 50.071
129335 95.585 -77.460 50.041
129336 95.602 -77.448 50.011
129337 95.618 -77.438 49.981
129338 95.634 -77.426 49.951
129339 95.678 -77.442 49.921
129340 95.750 -77.486 49.892
129341 95.763 -77.515 49.863
129342 95.749 -77.531 49.834
129343 95.734 -77.547 49.806
129344 95.719 -77.563 49.779
129345 95.702 -77.573 49.751
129346 95.678 -77.595 49.722
129347 95.678 -77.603 49.691
129348 95.681 -77.606 49.659
129349 95.685 -77.608 49.627
129350 95.688 -77.611 49.597
129351 95.692 -77.614 49.568
129352 95.695 -77.617 49.539
129353 95.699 -77.619 49.510
129354 95.702 -77.622 49.484
129355 95.706 -77.625 49.457
129356 95.709 -77.627 49.430
129357 95.711 -77.632 49.402
129358 95.712 -77.636 49.373
129359 95.713 -77.642 49.343
129360 95.718 -77.645 49.313
129361 95.727 -77.645 49.281
129362 95.735 -77.652 49.247
129363 95.755 -77.661 49.210
129364 95.780 -77.674 49.171
129365 95.806 -77.686 49.133
129366 95.831 -77.699 49.095
129367 95.856 -77.711 49.058
129368 95.880 -77.721 49.021
129369 95.903 -77.730 48.984
129370 95.925 -77.738 48.948
129371 95.948 -77.746 48.911
129372 95.971 -77.755 48.875
129373 95.993 -77.763 48.839
129374 96.017 -77.771 48.802
129375 96.039 -77.780 48.766
129376 96.062 -77.788 48.730
129377 96.085 -77.797 48.695
129378 96.107 -77.805 48.661
129379 96.130 -77.814 48.627
129380 96.153 -77.822 48.594
129381 96.174 -77.833 48.560
129382 96.188 -77.837 48.526
129383 96.186 -77.850 48.494
129384 96.180 -77.870 48.462
129385 96.174 -77.890 48.430
129386 96.173 -77.902 48.402
129387 96.180 -77.901 48.373
129388 96.187 -77.900 48.345
129389 96.197 -77.904 48.316
129390 96.225 -77.929 48.288
129391 96.251 -77.954 48.261
129392 96.278 -77.978 48.233
129393 96.285 -77.978 48.209
129394 96.287 -77.973 48.183
129395 96.290 -77.967 48.157
129396 96.292 -77.962 48.132
129397 96.294 -77.956 48.108
129398 96.297 -77.951 48.084
129399 96.294 -77.947 48.061
129400 96.292 -77.944 48.041
129401 96.289 -77.940 48.020
129402 96.287 -77.937 47.999
129403 96.284 -77.934 47.979
129404 96.303 -77.930 47.957
129405 96.324 -77.933 47.936
129406 96.346 -77.938 47.915
129407 96.368 -77.943 47.896
129408 96.390 -77.949 47.875
129409 96.416 -77.953 47.854
129410 96.445 -77.957 47.833
129411 96.473 -77.962 47.809
129412 96.494 -77.968 47.785
129413 96.504 -77.978 47.761
129414 96.513 -77.988 47.736
129415 96.518 -77.978 47.711
129416 96.521 -77.962 47.686
129417 96.528 -77.949 47.660
129418 96.558 -77.952 47.635
129419 96.589 -77.954 47.611
129420 96.601 -77.956 47.587
129421 96.612 -77.957 47.563
129422 96.623 -77.959 47.541
129423 96.633 -77.960 47.518
129424 96.644 -77.962 47.494
129425 96.658 -77.975 47.469
129426 96.672 -77.992 47.445
129427 96.687 -78.009 47.422
129428 96.701 -78.026 47.398
129429 96.709 -78.050 47.372
129430 96.717 -78.071 47.346
129431 96.730 -78.081 47.320
129432 96.744 -78.092 47.295
129433 96.758 -78.102 47.269
129434 96.771 -78.112 47.243
129435 96.789 -78.126 47.215
129436 96.807 -78.139 47.188
129437 96.825 -78.153 47.162
129438 96.836 -78.170 47.136
129439 96.845 -78.189 47.109
129440 96.861 -78.204 47.083
129441 96.878 -78.218 47.056
129442 96.896 -78.231 47.028
129443 96.913 -78.244 47.000
129444 96.932 -78.258 46.973
129445 96.947 -78.266 46.949
129446 96.961 -78.272 46.925
129447 96.976 -78.279 46.906
129448 96.992 -78.286 46.886
129449 97.009 -78.295 46.867
129450 97.027 -78.305 46.848
129451 97.045 -78.315 46.827
129452 97.074 -78.329 46.804
129453 97.103 -78.343 46.782
129454 97.134 -78.356 46.759
129455 97.175 -78.362 46.734
129456 97.217 -78.368 46.710
129457 97.258 -78.373 46.688
129458 97.300 -78.379 46.667
129459 97.342 -78.385 46.644
129460 97.383 -78.391 46.622
129461 97.424 -78.398 46.601
129462 97.466 -78.404 46.584
129463 97.507 -78.409 46.567
129464 97.548 -78.405 46.550
129465 97.596 -78.392 46.532
129466 97.637 -78.394 46.514
129467 97.660 -78.399 46.497
129468 97.683 -78.404 46.480
129469 97.707 -78.408 46.463
129470 97.730 -78.413 46.446
129471 97.753 -78.418 46.429
129472 97.802 -78.465 46.410
129473 97.832 -78.492 46.390
129474 97.857 -78.486 46.370
129475 97.891 -78.476 46.352
129476 97.929 -78.462 46.336
129477 97.966 -78.449 46.319
129478 98.003 -78.436 46.302
129479 98.037 -78.437 46.286
129480 98.070 -78.441 46.270
129481 98.101 -78.445 46.255
129482 98.115 -78.435 46.240
129483 98.127 -78.418 46.225
129484 98.128 -78.400 46.210
129485 98.130 -78.382 46.195
129486 98.132 -78.375 46.181
129487 98.135 -78.380 46.165
129488 98.139 -78.385 46.149
129489 98.141 -78.391 46.133
129490 98.145 -78.400 46.116
129491 98.176 -78.388 46.099
129492 98.209 -78.376 46.083
129493 98.262 -78.360 46.065
129494 98.314 -78.343 46.047
129495 98.321 -78.341 46.029
129496 98.327 -78.340 46.010
129497 98.339 -78.332 45.991
129498 98.331 -78.340 45.971
129499 98.319 -78.352 45.951
129500 98.308 -78.364 45.930
129501 98.307 -78.370 45.911
129502 98.329 -78.390 45.890
129503 98.353 -78.413 45.869
129504 98.376 -78.435 45.849
129505 98.396 -78.454 45.829
129506 98.408 -78.468 45.809
129507 98.388 -78.499 45.788
129508 98.404 -78.504 45.767
129509 98.419 -78.509 45.746
129510 98.434 -78.514 45.725
129511 98.448 -78.520 45.706
129512 98.461 -78.526 45.686
129513 98.475 -78.534 45.665
129514 98.491 -78.545 45.644
129515 98.506 -78.554 45.622
129516 98.521 -78.564 45.601
129517 98.530 -78.550 45.581
129518 98.536 -78.527 45.561
129519 98.537 -78.467 45.541
129520 98.548 -78.446 45.522
129521 98.565 -78.479 45.502
129522 98.582 -78.511 45.482
129523 98.599 -78.543 45.463
129524 98.639 -78.538 45.445
129525 98.686 -78.518 45.427
129526 98.727 -78.483 45.410
129527 98.766 -78.455 45.392
129528 98.790 -78.467 45.374
129529 98.815 -78.479 45.357
129530 98.840 -78.490 45.341
129531 98.865 -78.500 45.324
129532 98.884 -78.476 45.308
129533 98.894 -78.485 45.293
129534 98.902 -78.498 45.278
129535 98.910 -78.512 45.263
129536 98.918 -78.526 45.248
129537 98.933 -78.538 45.234
129538 98.988 -78.540 45.221
129539 99.043 -78.543 45.208
129540 99.078 -78.548 45.195
129541 99.097 -78.553 45.183
129542 99.116 -78.558 45.170
129543 99.136 -78.563 45.158
129544 99.155 -78.568 45.145
129545 99.174 -78.573 45.133
129546 99.194 -78.578 45.119
129547 99.213 -78.582 45.107
129548 99.232 -78.588 45.095
129549 99.252 -78.592 45.083
129550 99.268 -78.601 45.072
129551 99.257 -78.621 45.060
129552 99.238 -78.641 45.049
129553 99.219 -78.661 45.036
129554 99.265 -78.670 45.024
129555 99.323 -78.703 45.010
129556 99.373 -78.720 44.997
129557 99.423 -78.738 44.983
129558 99.473 -78.755 44.969
129559 99.522 -78.772 44.955
129560 99.574 -78.790 44.942
129561 99.629 -78.809 44.928
129562 99.683 -78.827 44.914
129563 99.738 -78.846 44.901
129564 99.782 -78.865 44.889
129565 99.802 -78.888 44.877
129566 99.816 -78.898 44.866
129567 99.829 -78.908 44.854
129568 99.842 -78.918 44.843
129569 99.855 -78.927 44.834
129570 99.868 -78.937 44.825
129571 99.881 -78.947 44.816
129572 99.895 -78.957 44.808
129573 99.908 -78.967 44.801
129574 99.921 -78.977 44.794
129575 99.934 -78.987 44.788
129576 99.947 -78.996 44.783
129577 99.960 -79.006 44.777
129578 99.986 -79.014 44.772
129579 100.007 -79.033 44.768
129580 100.037 -79.051 44.763
129581 100.066 -79.069 44.758
129582 100.096 -79.087 44.754
129583 100.125 -79.106 44.749
129584 100.154 -79.124 44.745
129585 100.184 -79.143 44.740
129586 100.213 -79.161 44.736
129587 100.243 -79.178 44.731
129588 100.270 -79.184 44.726
129589 100.297 -79.190 44.720
129590 100.325 -79.195 44.714
129591 100.352 -79.200 44.708
129592 100.382 -79.207 44.701
129593 100.420 -79.217 44.694
129874 93.292 -78.158 53.979
129875 93.315 -78.147 53.970
129876 93.339 -78.134 53.960
129877 93.362 -78.122 53.950
129878 93.386 -78.111 53.941
129879 93.400 -78.106 53.928
129880 93.409 -78.103 53.915
129881 93.419 -78.101 53.902
129882 93.429 -78.098 53.890
129883 93.438 -78.096 53.876
129884 93.440 -78.089 53.863
129885 93.446 -78.076 53.850
129886 93.457 -78.060 53.835
129887 93.467 -78.045 53.819
129888 93.477 -78.029 53.804
129889 93.487 -78.014 53.787
129916 93.834 -77.826 53.281
129917 93.847 -77.821 53.262
129918 93.859 -77.817 53.243
129919 93.872 -77.813 53.224
129920 93.885 -77.808 53.205
129921 93.898 -77.804 53.187
129922 93.911 -77.800 53.169
129923 93.924 -77.796 53.151
129924 93.936 -77.791 53.133
129925 93.949 -77.787 53.115
129926 93.962 -77.783 53.098
129927 93.975 -77.779 53.079
129928 93.987 -77.774 53.060
129929 93.984 -77.774 53.043
129930 93.982 -77.772 53.025
129931 93.979 -77.771 53.008
129932 93.977 -77.770 52.988
129933 93.975 -77.769 52.968
129934 93.972 -77.768 52.948
129935 93.970 -77.767 52.928
129936 93.967 -77.766 52.906
129937 93.973 -77.771 52.885
129938 93.989 -77.774 52.863
129939 94.005 -77.779 52.842
129940 94.020 -77.783 52.820
129941 94.036 -77.787 52.799
129942 94.052 -77.792 52.777
129943 94.068 -77.796 52.755
129944 94.071 -77.790 52.733
129945 94.072 -77.781 52.711
129946 94.073 -77.773 52.689
129947 94.074 -77.765 52.666
129948 94.075 -77.757 52.644
129949 94.110 -77.742 52.622
129950 94.135 -77.733 52.600
129951 94.159 -77.724 52.578
129952 94.185 -77.714 52.557
129953 94.209 -77.705 52.536
129954 94.227 -77.702 52.516
129955 94.245 -77.700 52.495
129956 94.262 -77.699 52.475
129957 94.279 -77.696 52.455
129958 94.296 -77.695 52.436
129959 94.313 -77.693 52.418
129960 94.330 -77.691 52.401
129961 94.347 -77.689 52.385
129962 94.364 -77.687 52.368
129963 94.382 -77.686 52.352
129964 94.411 -77.655 52.338
129965 94.445 -77.614 52.325
129966 94.453 -77.611 52.313
129967 94.451 -77.621 52.301
129968 94.456 -77.630 52.290
129969 94.464 -77.637 52.277
129970 94.472 -77.645 52.263
129971 94.482 -77.645 52.249
129972 94.498 -77.626 52.233
129973 94.513 -77.605 52.214
129974 94.530 -77.586 52.194
129975 94.650 -77.596 52.172
129976 94.567 -77.617 52.150
129977 94.574 -77.670 52.129
129978 94.597 -77.728 52.108
129979 94.617 -77.782 52.086
129980 94.635 -77.833 52.064
129981 94.651 -77.855 52.041
129982 94.666 -77.842 52.017
129983 94.681 -77.828 51.989
129984 94.696 -77.815 51.963
129985 94.711 -77.801 51.936
129986 94.725 -77.787 51.909
129987 94.741 -77.774 51.879
129988 94.755 -77.760 51.846
129989 94.762 -77.718 51.810
129990 94.764 -77.660 51.774
129991 94.773 -77.635 51.739
129992 94.813 -77.674 51.705
129993 94.832 -77.687 51.670
129994 94.828 -77.668 51.639
129995 94.825 -77.617 51.608
129996 94.822 -77.555 51.576
129997 94.845 -77.551 51.539
129998 94.884 -77.586 51.502
129999 94.886 -77.592 51.466
130000 94.898 -77.604 51.433
130001 94.933 -77.625 51.396
130002 94.973 -77.645 51.359
130003 95.007 -77.639 51.320
130004 95.030 -77.611 51.277
130005 95.052 -77.583 51.234
130006 95.060 -77.568 51.191
130007 95.057 -77.564 51.147
130008 95.057 -77.564 51.103
130009 95.084 -77.608 51.061
130010 95.108 -77.652 51.020
130011 95.095 -77.711 50.977
130012 95.136 -77.696 50.933
130013 95.159 -77.702 50.890
130014 95.183 -77.708 50.847
130015 95.189 -77.710 50.805
130016 95.198 -77.710 50.764
130017 95.204 -77.706 50.723
130018 95.209 -77.702 50.684
130019 95.215 -77.699 50.644
130020 95.220 -77.696 50.605
130021 95.226 -77.692 50.567
130022 95.248 -77.680 50.530
130023 95.273 -77.667 50.493
130024 95.298 -77.655 50.457
130025 95.323 -77.642 50.421
130026 95.348 -77.629 50.385
130027 95.373 -77.617 50.350
130028 95.398 -77.604 50.315
130029 95.423 -77.591 50.282
130030 95.439 -77.581 50.248
130031 95.450 -77.566 50.215
130032 95.477 -77.517 50.182
130033 95.517 -77.489 50.150
130034 95.544 -77.474 50.118
130035 95.560 -77.463 50.086
130036 95.577 -77.451 50.055
130037 95.593 -77.441 50.024
130038 95.609 -77.429 49.994
130039 95.625 -77.418 49.964
130040 95.665 -77.429 49.933
130041 95.766 -77.500 49.903
130042 95.772 -77.526 49.874
130043 95.758 -77.542 49.845
130044 95.743 -77.558 49.816
130045 95.729 -77.571 49.788
130046 95.700 -77.577 49.759
130047 95.677 -77.599 49.732
130048 95.678 -77.605 49.703
130049 95.682 -77.608 49.671
130050 95.685 -77.610 49.640
130051 95.689 -77.613 49.609
130052 95.692 -77.615 49.580
130053 95.696 -77.618 49.551
130054 95.699 -77.620 49.522
130055 95.703 -77.624 49.494
130056 95.706 -77.627 49.467
130057 95.710 -77.629 49.439
130058 95.713 -77.632 49.411
130059 95.717 -77.634 49.382
130060 95.721 -77.637 49.353
130061 95.724 -77.639 49.322
130062 95.728 -77.642 49.290
130063 95.734 -77.644 49.257
130064 95.758 -77.655 49.221
130065 95.783 -77.668 49.182
130066 95.808 -77.680 49.144
130067 95.833 -77.693 49.106
130068 95.859 -77.705 49.069
130069 95.881 -77.714 49.032
130070 95.904 -77.722 48.995
130071 95.927 -77.730 48.958
130072 95.950 -77.740 48.920
130073 95.972 -77.748 48.884
130074 95.995 -77.756 48.848
130075 96.018 -77.765 48.812
130076 96.041 -77.773 48.777
130077 96.063 -77.781 48.741
130078 96.086 -77.790 48.705
130079 96.109 -77.798 48.669
130080 96.131 -77.806 48.637
130081 96.155 -77.815 48.604
130082 96.177 -77.823 48.571
130083 96.199 -77.832 48.538
130084 96.194 -77.851 48.505
130085 96.188 -77.871 48.473
130086 96.182 -77.891 48.442
130087 96.177 -77.909 48.413
130088 96.184 -77.909 48.385
130089 96.191 -77.908 48.356
130090 96.210 -77.923 48.327
130091 96.238 -77.948 48.299
130092 96.262 -77.970 48.271
130093 96.265 -77.965 48.243
130094 96.267 -77.959 48.218
130095 96.269 -77.954 48.191
130096 96.272 -77.949 48.165
130097 96.274 -77.943 48.140
130098 96.277 -77.938 48.115
130099 96.279 -77.933 48.091
130100 96.277 -77.929 48.068
130101 96.275 -77.925 48.048
130102 96.272 -77.922 48.027
130103 96.269 -77.918 48.005
130104 96.282 -77.915 47.984
130105 96.303 -77.910 47.962
130106 96.324 -77.913 47.940
130107 96.346 -77.918 47.919
130108 96.368 -77.924 47.899
130109 96.389 -77.929 47.877
130110 96.411 -77.934 47.856
130111 96.433 -77.939 47.834
130112 96.455 -77.944 47.814
130113 96.483 -77.949 47.792
130114 96.511 -77.953 47.769
130115 96.521 -77.963 47.745
130116 96.527 -77.961 47.720
130117 96.530 -77.945 47.696
130118 96.558 -77.946 47.670
130119 96.577 -77.948 47.646
130120 96.588 -77.949 47.622
130121 96.599 -77.951 47.598
130122 96.609 -77.952 47.573
130123 96.620 -77.953 47.549
130124 96.631 -77.955 47.526
130125 96.642 -77.956 47.502
130126 96.654 -77.965 47.479
130127 96.668 -77.981 47.455
130128 96.683 -77.999 47.431
130129 96.697 -78.016 47.407
130130 96.712 -78.034 47.382
130131 96.727 -78.050 47.357
130132 96.738 -78.065 47.331
130133 96.756 -78.078 47.304
130134 96.773 -78.092 47.277
130135 96.792 -78.106 47.250
130136 96.809 -78.119 47.223
130137 96.827 -78.132 47.196
130138 96.839 -78.149 47.170
130139 96.848 -78.169 47.144
130140 96.856 -78.188 47.117
130141 96.870 -78.203 47.089
130142 96.888 -78.217 47.062
130143 96.905 -78.231 47.035
130144 96.921 -78.238 47.008
130145 96.936 -78.244 46.983
130146 96.950 -78.251 46.959
130147 96.965 -78.257 46.937
130148 96.980 -78.263 46.916
130149 96.997 -78.272 46.897
130150 97.015 -78.282 46.879
130151 97.032 -78.291 46.861
130152 97.049 -78.301 46.841
130153 97.074 -78.313 46.820
130154 97.102 -78.328 46.798
130155 97.137 -78.338 46.776
130156 97.178 -78.344 46.753
130157 97.220 -78.350 46.730
130158 97.261 -78.356 46.708
130159 97.303 -78.362 46.688
130160 97.345 -78.368 46.666
130161 97.386 -78.374 46.643
130162 97.427 -78.379 46.622
130163 97.469 -78.385 46.604
130164 97.510 -78.391 46.586
130165 97.552 -78.398 46.569
130166 97.593 -78.394 46.552
130167 97.645 -78.385 46.534
130168 97.672 -78.388 46.516
130169 97.695 -78.393 46.500
130170 97.718 -78.398 46.483
130171 97.741 -78.403 46.466
130172 97.774 -78.422 46.450
130173 97.816 -78.463 46.430
130174 97.825 -78.464 46.410
130175 97.850 -78.459 46.391
130176 97.880 -78.451 46.373
130177 97.914 -78.441 46.357
130178 97.951 -78.428 46.341
130179 97.989 -78.415 46.324
130180 98.024 -78.410 46.308
130181 98.057 -78.413 46.293
130182 98.084 -78.418 46.277
130183 98.097 -78.417 46.262
130184 98.114 -78.399 46.246
130185 98.132 -78.381 46.231
130186 98.139 -78.363 46.216
130187 98.141 -78.351 46.200
130188 98.144 -78.357 46.184
130189 98.143 -78.366 46.167
130190 98.139 -78.376 46.151
130191 98.148 -78.388 46.133
130192 98.178 -78.376 46.116
130193 98.209 -78.364 46.099
130194 98.240 -78.352 46.081
130195 98.271 -78.341 46.063
130196 98.306 -78.328 46.045
130197 98.336 -78.319 46.027
130198 98.348 -78.309 46.008
130199 98.347 -78.310 45.988
130200 98.337 -78.326 45.968
130201 98.333 -78.347 45.947
130202 98.329 -78.370 45.927
130203 98.334 -78.387 45.907
130204 98.343 -78.401 45.886
130205 98.366 -78.423 45.865
130206 98.389 -78.445 45.845
130207 98.391 -78.467 45.824
130208 98.378 -78.494 45.804
130209 98.393 -78.498 45.783
130210 98.408 -78.504 45.763
130211 98.424 -78.508 45.742
130212 98.439 -78.514 45.722
130213 98.454 -78.523 45.701
130214 98.469 -78.533 45.680
130215 98.485 -78.543 45.659
130216 98.500 -78.553 45.637
130217 98.513 -78.555 45.616
130218 98.519 -78.523 45.595
130219 98.535 -78.480 45.575
130220 98.563 -78.448 45.554
130221 98.575 -78.432 45.534
130222 98.566 -78.463 45.515
130223 98.583 -78.496 45.494
130224 98.600 -78.529 45.475
130225 98.646 -78.513 45.457
130226 98.693 -78.495 45.440
130227 98.738 -78.470 45.422
130228 98.779 -78.435 45.404
130229 98.804 -78.448 45.386
130230 98.829 -78.459 45.369
130231 98.854 -78.468 45.353
130232 98.881 -78.465 45.336
130233 98.900 -78.438 45.320
130234 98.913 -78.468 45.305
130235 98.925 -78.497 45.290
130236 98.934 -78.510 45.275
130237 98.942 -78.524 45.260
130238 98.966 -78.534 45.246
130239 99.021 -78.536 45.232
130240 99.074 -78.534 45.220
130241 99.088 -78.538 45.207
130242 99.108 -78.542 45.195
130243 99.127 -78.548 45.183
130244 99.146 -78.552 45.170
130245 99.166 -78.557 45.157
130246 99.185 -78.562 45.144
130247 99.205 -78.567 45.131
130248 99.224 -78.573 45.118
130249 99.243 -78.577 45.106
130250 99.263 -78.582 45.094
130251 99.268 -78.600 45.083
130252 99.270 -78.620 45.071
130253 99.251 -78.640 45.059
130254 99.233 -78.660 45.047
130255 99.274 -78.656 45.035
130256 99.339 -78.681 45.022
130257 99.401 -78.722 45.008
130258 99.452 -78.743 44.994
130259 99.502 -78.761 44.980
130260 99.552 -78.778 44.966
130261 99.602 -78.795 44.952
130262 99.652 -78.812 44.939
130263 99.702 -78.830 44.926
130264 99.752 -78.847 44.913
130265 99.797 -78.858 44.900
130266 99.804 -78.871 44.888
130267 99.817 -78.876 44.877
130268 99.830 -78.884 44.865
130269 99.843 -78.894 44.854
130270 99.857 -78.904 44.844
130271 99.870 -78.914 44.835
130272 99.883 -78.924 44.826
130273 99.896 -78.933 44.818
130274 99.909 -78.943 44.810
130275 99.922 -78.953 44.803
130276 99.935 -78.963 44.796
130277 99.948 -78.973 44.790
130278 99.962 -78.983 44.785
130279 99.987 -78.991 44.779
130280 100.010 -79.011 44.774
130281 100.039 -79.029 44.769
130282 100.069 -79.047 44.764
130283 100.098 -79.065 44.759
130284 100.127 -79.084 44.755
130285 100.157 -79.103 44.750
130286 100.186 -79.121 44.746
130287 100.216 -79.139 44.740
130288 100.245 -79.157 44.735
130289 100.273 -79.164 44.730
130290 100.300 -79.169 44.724
130291 100.327 -79.175 44.717
130292 100.354 -79.181 44.710
130293 100.388 -79.189 44.703
130294 100.427 -79.199 44.695
130295 100.465 -79.209 44.687
130573 93.243 -78.172 54.008
130574 93.267 -78.159 53.999
130575 93.291 -78.148 53.989
130576 93.314 -78.136 53.980
130577 93.338 -78.124 53.969
130578 93.362 -78.112 53.959
130579 93.386 -78.100 53.950
130580 93.400 -78.095 53.936
130581 93.410 -78.092 53.923
130582 93.419 -78.090 53.911
130583 93.429 -78.087 53.898
130584 93.438 -78.085 53.885
130585 93.437 -78.075 53.872
130586 93.447 -78.060 53.859
130587 93.458 -78.044 53.844
130588 93.468 -78.029 53.829
130589 93.478 -78.013 53.813
130590 93.489 -77.998 53.798
130591 93.507 -77.984 53.783
130616 93.822 -77.818 53.311
130617 93.834 -77.815 53.292
130618 93.847 -77.810 53.274
130619 93.860 -77.806 53.255
130620 93.873 -77.802 53.236
130621 93.886 -77.797 53.218
130622 93.899 -77.793 53.200
130623 93.911 -77.789 53.181
130624 93.924 -77.784 53.164
130625 93.937 -77.780 53.146
130626 93.950 -77.776 53.129
130627 93.963 -77.771 53.112
130628 93.976 -77.768 53.094
130629 93.987 -77.763 53.077
130630 93.985 -77.762 53.060
130631 93.982 -77.761 53.045
130632 93.980 -77.760 53.028
130633 93.978 -77.759 53.009
130634 93.975 -77.758 52.989
130635 93.973 -77.757 52.968
130636 93.970 -77.755 52.947
130637 93.968 -77.754 52.925
130638 93.976 -77.755 52.903
130639 93.992 -77.759 52.880
130640 94.008 -77.764 52.858
130641 94.024 -77.768 52.836
130642 94.040 -77.772 52.815
130643 94.055 -77.777 52.794
130644 94.071 -77.781 52.772
130645 94.072 -77.773 52.750
130646 94.073 -77.765 52.728
130647 94.074 -77.756 52.706
130648 94.076 -77.749 52.683
130649 94.077 -77.740 52.660
130650 94.111 -77.725 52.638
130651 94.136 -77.716 52.616
130652 94.161 -77.706 52.594
130653 94.178 -77.705 52.573
130654 94.195 -77.702 52.553
130655 94.212 -77.701 52.532
130656 94.229 -77.699 52.512
130657 94.247 -77.697 52.492
130658 94.264 -77.696 52.473
130659 94.281 -77.693 52.455
130660 94.298 -77.692 52.438
130661 94.315 -77.690 52.421
130662 94.332 -77.688 52.405
130663 94.349 -77.686 52.390
130664 94.370 -77.675 52.375
130665 94.405 -77.633 52.362
130666 94.439 -77.592 52.352
130667 94.439 -77.600 52.342
130668 94.447 -77.608 52.332
130669 94.455 -77.616 52.323
130670 94.463 -77.624 52.311
130671 94.471 -77.631 52.298
130672 94.479 -77.638 52.283
130673 94.494 -77.618 52.266
130674 94.510 -77.599 52.248
130675 94.587 -77.591 52.227
130676 94.552 -77.611 52.204
130677 94.557 -77.663 52.182
130678 94.579 -77.722 52.161
130679 94.603 -77.780 52.142
130680 94.623 -77.836 52.122
130681 94.641 -77.887 52.102
130682 94.657 -77.890 52.080
130683 94.672 -77.877 52.057
130684 94.686 -77.863 52.030
130685 94.701 -77.850 52.003
130686 94.716 -77.837 51.975
130687 94.731 -77.823 51.944
130688 94.739 -77.787 51.911
130689 94.741 -77.729 51.877
130690 94.743 -77.672 51.841
130691 94.747 -77.620 51.804
130692 94.761 -77.611 51.769
130693 94.795 -77.641 51.735
130694 94.843 -77.696 51.701
130695 94.863 -77.659 51.667
130696 94.861 -77.597 51.633
130697 94.858 -77.536 51.599
130698 94.876 -77.520 51.562
130699 94.915 -77.555 51.528
130700 94.955 -77.591 51.494
130701 94.965 -77.604 51.461
130702 94.958 -77.610 51.426
130703 94.982 -77.620 51.389
130704 95.013 -77.608 51.351
130705 95.036 -77.580 51.310
130706 95.044 -77.564 51.268
130707 95.041 -77.560 51.224
130708 95.038 -77.555 51.179
130709 95.036 -77.552 51.134
130710 95.062 -77.595 51.090
130711 95.089 -77.639 51.048
130712 95.111 -77.684 51.006
130713 95.138 -77.689 50.963
130714 95.171 -77.683 50.919
130715 95.145 -77.673 50.874
130716 95.164 -77.680 50.831
130717 95.176 -77.684 50.788
130718 95.188 -77.686 50.746
130719 95.200 -77.689 50.706
130720 95.211 -77.692 50.667
130721 95.219 -77.692 50.627
130722 95.231 -77.686 50.589
130723 95.255 -77.673 50.551
130724 95.280 -77.660 50.514
130725 95.306 -77.647 50.478
130726 95.331 -77.634 50.442
130727 95.355 -77.621 50.406
130728 95.381 -77.609 50.371
130729 95.406 -77.596 50.336
130730 95.430 -77.583 50.302
130731 95.450 -77.572 50.268
130732 95.469 -77.541 50.234
130733 95.496 -77.493 50.200
130734 95.519 -77.477 50.167
130735 95.536 -77.468 50.135
130736 95.554 -77.461 50.103
130737 95.572 -77.452 50.071
130738 95.590 -77.444 50.039
130739 95.608 -77.436 50.008
130740 95.628 -77.429 49.977
130741 95.681 -77.444 49.946
130742 95.783 -77.514 49.916
130743 95.781 -77.536 49.886
130744 95.767 -77.552 49.856
130745 95.752 -77.568 49.827
130746 95.727 -77.574 49.798
130747 95.698 -77.580 49.769
130748 95.675 -77.604 49.743
130749 95.679 -77.606 49.714
130750 95.682 -77.609 49.684
130751 95.686 -77.611 49.653
130752 95.689 -77.614 49.623
130753 95.693 -77.617 49.593
130754 95.696 -77.620 49.563
130755 95.700 -77.622 49.534
130756 95.703 -77.625 49.505
130757 95.707 -77.628 49.477
130758 95.710 -77.630 49.449
130759 95.714 -77.633 49.420
130760 95.718 -77.636 49.392
130761 95.721 -77.639 49.362
130762 95.724 -77.641 49.331
130763 95.732 -77.642 49.299
130764 95.742 -77.643 49.266
130765 95.760 -77.650 49.232
130766 95.785 -77.662 49.193
130767 95.811 -77.675 49.155
130768 95.836 -77.687 49.118
130769 95.860 -77.699 49.080
130770 95.883 -77.707 49.043
130771 95.905 -77.715 49.006
130772 95.928 -77.724 48.969
130773 95.951 -77.732 48.931
130774 95.974 -77.740 48.894
130775 95.996 -77.749 48.858
130776 96.019 -77.757 48.822
130777 96.042 -77.765 48.787
130778 96.065 -77.774 48.751
130779 96.088 -77.783 48.716
130780 96.110 -77.791 48.679
130781 96.133 -77.799 48.647
130782 96.158 -77.808 48.615
130783 96.182 -77.818 48.582
130784 96.192 -77.828 48.549
130785 96.197 -77.851 48.517
130786 96.196 -77.872 48.485
130787 96.190 -77.892 48.454
130788 96.184 -77.912 48.425
130789 96.188 -77.916 48.396
130790 96.197 -77.918 48.367
130791 96.224 -77.943 48.338
130792 96.243 -77.957 48.309
130793 96.245 -77.952 48.281
130794 96.248 -77.946 48.253
130795 96.250 -77.941 48.227
130796 96.252 -77.936 48.201
130797 96.255 -77.931 48.174
130798 96.257 -77.924 48.149
130799 96.259 -77.919 48.123
130800 96.262 -77.914 48.099
130801 96.260 -77.910 48.075
130802 96.258 -77.907 48.055
130803 96.255 -77.903 48.034
130804 96.261 -77.899 48.012
130805 96.282 -77.895 47.990
130806 96.303 -77.890 47.968
130807 96.324 -77.893 47.946
130808 96.346 -77.899 47.924
130809 96.368 -77.904 47.903
130810 96.389 -77.909 47.881
130811 96.411 -77.915 47.860
130812 96.433 -77.919 47.838
130813 96.455 -77.924 47.818
130814 96.476 -77.930 47.797
130815 96.498 -77.935 47.776
130816 96.521 -77.940 47.754
130817 96.537 -77.944 47.730
130818 96.553 -77.940 47.706
130819 96.564 -77.941 47.681
130820 96.575 -77.943 47.656
130821 96.585 -77.944 47.631
130822 96.596 -77.946 47.606
130823 96.607 -77.947 47.580
130824 96.617 -77.949 47.555
130825 96.628 -77.950 47.532
130826 96.639 -77.952 47.508
130827 96.650 -77.955 47.485
130828 96.664 -77.971 47.462
130829 96.679 -77.989 47.437
130830 96.693 -78.006 47.413
130831 96.708 -78.023 47.389
130832 96.723 -78.040 47.364
130833 96.746 -78.056 47.339
130834 96.773 -78.072 47.312
130835 96.794 -78.085 47.284
130836 96.811 -78.099 47.257
130837 96.829 -78.112 47.231
130838 96.842 -78.128 47.204
130839 96.851 -78.147 47.177
130840 96.859 -78.167 47.151
130841 96.868 -78.186 47.124
130842 96.880 -78.203 47.096
130843 96.895 -78.210 47.068
130844 96.909 -78.216 47.041
130845 96.925 -78.223 47.016
130846 96.939 -78.229 46.992
130847 96.954 -78.236 46.969
130848 96.969 -78.242 46.948
130849 96.985 -78.250 46.927
130850 97.003 -78.260 46.908
130851 97.020 -78.269 46.891
130852 97.037 -78.279 46.874
130853 97.055 -78.288 46.855
130854 97.073 -78.298 46.834
130855 97.102 -78.312 46.813
130856 97.140 -78.320 46.793
130857 97.182 -78.326 46.772
130858 97.223 -78.332 46.751
130859 97.264 -78.338 46.729
130860 97.306 -78.344 46.708
130861 97.348 -78.350 46.686
130862 97.389 -78.356 46.665
130863 97.431 -78.362 46.644
130864 97.472 -78.368 46.624
130865 97.513 -78.374 46.606
130866 97.555 -78.380 46.590
130867 97.596 -78.385 46.575
130868 97.638 -78.383 46.558
130869 97.683 -78.378 46.540
130870 97.707 -78.383 46.522
130871 97.730 -78.388 46.503
130872 97.753 -78.392 46.486
130873 97.799 -78.436 46.471
130874 97.809 -78.437 46.452
130875 97.818 -78.438 46.432
130876 97.843 -78.432 46.412
130877 97.873 -78.425 46.393
130878 97.903 -78.416 46.380
130879 97.937 -78.407 46.364
130880 97.974 -78.394 46.347
130881 98.011 -78.382 46.331
130882 98.043 -78.385 46.315
130883 98.068 -78.391 46.300
130884 98.078 -78.399 46.283
130885 98.096 -78.381 46.267
130886 98.113 -78.363 46.252
130887 98.131 -78.344 46.236
130888 98.148 -78.329 46.219
130889 98.144 -78.340 46.202
130890 98.141 -78.351 46.185
130891 98.137 -78.362 46.168
130892 98.150 -78.375 46.151
130893 98.181 -78.363 46.134
130894 98.212 -78.351 46.116
130895 98.242 -78.339 46.098
130896 98.273 -78.328 46.079
130897 98.303 -78.316 46.061
130898 98.334 -78.305 46.044
130899 98.344 -78.306 46.025
130900 98.355 -78.298 46.006
130901 98.372 -78.328 45.986
130902 98.382 -78.363 45.966
130903 98.378 -78.385 45.946
130904 98.377 -78.407 45.925
130905 98.387 -78.420 45.904
130906 98.396 -78.435 45.882
130907 98.392 -78.433 45.861
130908 98.373 -78.458 45.841
130909 98.371 -78.495 45.820
130910 98.410 -78.505 45.800
130911 98.432 -78.513 45.780
130912 98.445 -78.511 45.760
130913 98.449 -78.499 45.739
130914 98.457 -78.506 45.718
130915 98.469 -78.523 45.696
130916 98.481 -78.539 45.674
130917 98.494 -78.551 45.652
130918 98.511 -78.530 45.631
130919 98.538 -78.498 45.610
130920 98.566 -78.467 45.589
130921 98.594 -78.434 45.568
130922 98.605 -78.418 45.547
130923 98.570 -78.448 45.527
130924 98.584 -78.481 45.507
130925 98.605 -78.506 45.488
130926 98.653 -78.488 45.470
130927 98.700 -78.470 45.452
130928 98.748 -78.451 45.434
130929 98.791 -78.421 45.417
130930 98.817 -78.429 45.399
130931 98.843 -78.437 45.383
130932 98.870 -78.434 45.366
130933 98.898 -78.431 45.349
130934 98.915 -78.401 45.333
130935 98.932 -78.447 45.318
130936 98.948 -78.492 45.303
130937 98.957 -78.508 45.288
130938 98.966 -78.522 45.273
130939 98.999 -78.529 45.259
130940 99.054 -78.532 45.246
130941 99.102 -78.517 45.233
130942 99.109 -78.523 45.220
130943 99.118 -78.532 45.208
130944 99.138 -78.537 45.196
130945 99.157 -78.542 45.183
130946 99.177 -78.547 45.170
130947 99.196 -78.552 45.157
130948 99.215 -78.557 45.144
130949 99.234 -78.562 45.131
130950 99.254 -78.567 45.119
130951 99.264 -78.579 45.107
130952 99.267 -78.599 45.095
130953 99.270 -78.619 45.085
130954 99.265 -78.639 45.073
130955 99.246 -78.658 45.061
130956 99.280 -78.658 45.049
130957 99.351 -78.652 45.034
130958 99.413 -78.692 45.020
130959 99.474 -78.733 45.006
130960 99.532 -78.766 44.992
130961 99.582 -78.783 44.978
130962 99.632 -78.801 44.964
130963 99.682 -78.818 44.951
130964 99.731 -78.835 44.939
130965 99.785 -78.849 44.926
130966 99.822 -78.846 44.914
130967 99.808 -78.874 44.902
130968 99.818 -78.888 44.890
130969 99.831 -78.893 44.878
130970 99.844 -78.897 44.867
130971 99.858 -78.902 44.856
130972 99.871 -78.907 44.846
130973 99.884 -78.912 44.837
130974 99.898 -78.916 44.830
130975 99.911 -78.921 44.822
130976 99.924 -78.929 44.814
130977 99.937 -78.939 44.805
130978 99.950 -78.949 44.799
130979 99.964 -78.959 44.793
130980 99.983 -78.971 44.787
130981 100.012 -78.989 44.781
130982 100.042 -79.007 44.776
130983 100.071 -79.025 44.770
130984 100.101 -79.043 44.765
130985 100.130 -79.062 44.761
130986 100.159 -79.081 44.757
130987 100.189 -79.099 44.752
130988 100.218 -79.117 44.746
130989 100.248 -79.135 44.740
130990 100.275 -79.144 44.735
130991 100.302 -79.150 44.729
130992 100.330 -79.155 44.722
130993 100.357 -79.160 44.715
130994 100.395 -79.171 44.707
130995 100.434 -79.181 44.699
130996 100.473 -79.196 44.690
130997 100.512 -79.212 44.682
131273 93.219 -78.173 54.027
131274 93.243 -78.161 54.018
131275 93.267 -78.150 54.008
131276 93.290 -78.138 53.998
131277 93.314 -78.125 53.988
131278 93.338 -78.114 53.979
131279 93.361 -78.102 53.969
131280 93.385 -78.090 53.959
131281 93.400 -78.084 53.945
131282 93.410 -78.081 53.932
131283 93.419 -78.079 53.920
131284 93.429 -78.077 53.907
131285 93.435 -78.072 53.894
131286 93.439 -78.059 53.882
131287 93.449 -78.044 53.868
131288 93.460 -78.029 53.853
131289 93.470 -78.013 53.838
131290 93.480 -77.998 53.822
131291 93.491 -77.982 53.807
131292 93.508 -77.971 53.792
131293 93.520 -77.962 53.777
131294 93.532 -77.954 53.762
131298 93.578 -77.921 53.692
131299 93.589 -77.912 53.674
131300 93.601 -77.905 53.657
131301 93.613 -77.896 53.640
131302 93.624 -77.888 53.620
131303 93.636 -77.880 53.600
131304 93.648 -77.871 53.581
131317 93.822 -77.808 53.322
131318 93.835 -77.803 53.304
131319 93.848 -77.799 53.285
131320 93.860 -77.795 53.266
131321 93.873 -77.790 53.248
131322 93.887 -77.786 53.230
131323 93.899 -77.782 53.212
131324 93.912 -77.777 53.194
131325 93.925 -77.774 53.177
131326 93.938 -77.769 53.160
131327 93.951 -77.765 53.143
131328 93.964 -77.761 53.126
131329 93.976 -77.756 53.110
131330 93.988 -77.752 53.095
131331 93.985 -77.751 53.080
131332 93.983 -77.750 53.064
131333 93.980 -77.749 53.047
131334 93.978 -77.748 53.028
131335 93.976 -77.746 53.009
131336 93.975 -77.743 52.988
131337 93.975 -77.740 52.966
131338 93.975 -77.737 52.943
131339 93.977 -77.735 52.920
131340 93.993 -77.740 52.897
131341 94.009 -77.745 52.876
131342 94.025 -77.750 52.855
131343 94.041 -77.755 52.834
131344 94.058 -77.760 52.813
131345 94.073 -77.764 52.791
131346 94.074 -77.756 52.768
131347 94.075 -77.748 52.745
131348 94.076 -77.740 52.723
131349 94.077 -77.732 52.700
131350 94.078 -77.724 52.677
131351 94.112 -77.708 52.655
131352 94.129 -77.707 52.632
131353 94.146 -77.705 52.611
131354 94.163 -77.703 52.590
131355 94.180 -77.702 52.569
131356 94.197 -77.699 52.549
131357 94.214 -77.698 52.529
131358 94.231 -77.696 52.510
131359 94.248 -77.694 52.491
131360 94.265 -77.692 52.474
131361 94.282 -77.690 52.457
131362 94.299 -77.689 52.441
131363 94.317 -77.687 52.427
131364 94.334 -77.685 52.413
131365 94.364 -77.653 52.401
131366 94.399 -77.611 52.389
131367 94.430 -77.577 52.380
131368 94.437 -77.586 52.372
131369 94.445 -77.594 52.366
131370 94.453 -77.602 52.359
131371 94.461 -77.610 52.348
131372 94.469 -77.617 52.335
131373 94.477 -77.625 52.318
131374 94.491 -77.611 52.300
131375 94.507 -77.592 52.280
131376 94.538 -77.605 52.260
131377 94.540 -77.657 52.237
131378 94.562 -77.716 52.216
131379 94.585 -77.774 52.196
131380 94.608 -77.834 52.177
131381 94.629 -77.890 52.157
131382 94.647 -77.940 52.137
131383 94.662 -77.926 52.116
131384 94.677 -77.912 52.093
131385 94.692 -77.899 52.066
131386 94.707 -77.885 52.038
131387 94.717 -77.855 52.009
131388 94.719 -77.799 51.977
131389 94.721 -77.741 51.943
131390 94.723 -77.683 51.908
131391 94.724 -77.626 51.874
131392 94.734 -77.598 51.837
131393 94.748 -77.589 51.803
131394 94.777 -77.608 51.769
131395 94.833 -77.657 51.734
131396 94.888 -77.615 51.697
131397 94.897 -77.577 51.661
131398 94.895 -77.516 51.626
131399 94.907 -77.489 51.591
131400 94.947 -77.524 51.557
131401 94.986 -77.560 51.523
131402 95.012 -77.599 51.486
131403 94.984 -77.610 51.449
131404 94.990 -77.595 51.412
131405 95.019 -77.577 51.375
131406 95.029 -77.561 51.335
131407 95.026 -77.556 51.294
131408 95.023 -77.551 51.251
131409 95.020 -77.548 51.208
131410 95.016 -77.543 51.164
131411 95.041 -77.584 51.119
131412 95.067 -77.627 51.076
131413 95.094 -77.671 51.034
131414 95.125 -77.702 50.990
131415 95.137 -77.649 50.946
131416 95.118 -77.642 50.901
131417 95.136 -77.649 50.858
131418 95.152 -77.655 50.815
131419 95.164 -77.658 50.773
131420 95.176 -77.661 50.731
131421 95.187 -77.664 50.690
131422 95.210 -77.677 50.651
131423 95.234 -77.667 50.612
131424 95.259 -77.657 50.574
131425 95.285 -77.645 50.538
131426 95.311 -77.634 50.501
131427 95.336 -77.623 50.465
131428 95.362 -77.611 50.429
131429 95.388 -77.601 50.394
131430 95.413 -77.589 50.359
131431 95.438 -77.576 50.324
131432 95.461 -77.564 50.290
131433 95.483 -77.532 50.256
131434 95.501 -77.510 50.221
131435 95.519 -77.488 50.187
131436 95.537 -77.478 50.153
131437 95.555 -77.470 50.119
131438 95.573 -77.462 50.088
131439 95.591 -77.454 50.056
131440 95.609 -77.446 50.024
131441 95.655 -77.455 49.991
131442 95.706 -77.468 49.960
131443 95.800 -77.528 49.930
131444 95.791 -77.547 49.900
131445 95.776 -77.563 49.869
131446 95.754 -77.572 49.839
131447 95.725 -77.577 49.810
131448 95.696 -77.583 49.781
131449 95.676 -77.605 49.753
131450 95.679 -77.608 49.727
131451 95.683 -77.611 49.698
131452 95.686 -77.613 49.667
131453 95.690 -77.616 49.637
131454 95.693 -77.618 49.606
131455 95.697 -77.621 49.575
131456 95.700 -77.624 49.545
131457 95.703 -77.627 49.516
131458 95.707 -77.630 49.487
131459 95.710 -77.632 49.459
131460 95.714 -77.635 49.430
131461 95.718 -77.637 49.401
131462 95.721 -77.640 49.371
131463 95.731 -77.641 49.339
131464 95.740 -77.642 49.308
131465 95.750 -77.642 49.275
131466 95.762 -77.644 49.241
131467 95.788 -77.656 49.203
131468 95.813 -77.669 49.166
131469 95.838 -77.682 49.129
131470 95.862 -77.691 49.093
131471 95.884 -77.699 49.055
131472 95.907 -77.708 49.017
131473 95.930 -77.717 48.980
131474 95.953 -77.725 48.942
131475 95.975 -77.733 48.905
131476 95.998 -77.742 48.869
131477 96.021 -77.750 48.832
131478 96.043 -77.758 48.796
131479 96.066 -77.767 48.761
131480 96.089 -77.775 48.726
131481 96.116 -77.784 48.690
131482 96.143 -77.794 48.658
131483 96.164 -77.803 48.626
131484 96.174 -77.814 48.594
131485 96.183 -77.825 48.560
131486 96.188 -77.848 48.528
131487 96.193 -77.871 48.497
131488 96.198 -77.893 48.466
131489 96.192 -77.913 48.437
131490 96.192 -77.924 48.408
131491 96.210 -77.937 48.378
131492 96.223 -77.944 48.349
131493 96.226 -77.939 48.319
131494 96.228 -77.934 48.290
131495 96.230 -77.928 48.263
131496 96.233 -77.922 48.236
131497 96.235 -77.917 48.210
131498 96.237 -77.912 48.184
131499 96.240 -77.906 48.157
131500 96.242 -77.901 48.131
131501 96.244 -77.896 48.107
131502 96.243 -77.891 48.084
131503 96.241 -77.888 48.062
131504 96.240 -77.884 48.041
131505 96.261 -77.880 48.020
131506 96.282 -77.875 47.998
131507 96.302 -77.871 47.976
131508 96.324 -77.874 47.953
131509 96.345 -77.879 47.930
131510 96.368 -77.884 47.909
131511 96.389 -77.890 47.887
131512 96.411 -77.895 47.865
131513 96.433 -77.899 47.843
131514 96.454 -77.905 47.822
131515 96.476 -77.910 47.801
131516 96.498 -77.915 47.782
131517 96.519 -77.925 47.763
131518 96.542 -77.936 47.740
131519 96.550 -77.935 47.716
131520 96.561 -77.937 47.691
131521 96.572 -77.938 47.665
131522 96.583 -77.940 47.638
131523 96.593 -77.941 47.611
131524 96.604 -77.942 47.584
131525 96.615 -77.943 47.559
131526 96.625 -77.945 47.535
131527 96.636 -77.946 47.511
131528 96.647 -77.948 47.487
131529 96.661 -77.962 47.464
131530 96.675 -77.979 47.440
131531 96.689 -77.996 47.416
131532 96.704 -78.013 47.393
131533 96.719 -78.031 47.368
131534 96.744 -78.046 47.343
131535 96.771 -78.061 47.317
131536 96.798 -78.075 47.291
131537 96.826 -78.091 47.265
131538 96.845 -78.106 47.237
131539 96.852 -78.124 47.211
131540 96.858 -78.141 47.183
131541 96.864 -78.158 47.156
131542 96.870 -78.175 47.129
131543 96.883 -78.188 47.102
131544 96.898 -78.195 47.076
131545 96.913 -78.201 47.049
131546 96.928 -78.208 47.024
131547 96.943 -78.214 47.001
131548 96.958 -78.221 46.979
131549 96.973 -78.228 46.958
131550 96.991 -78.237 46.939
131551 97.008 -78.247 46.921
131552 97.025 -78.256 46.905
131553 97.043 -78.266 46.889
131554 97.061 -78.275 46.871
131555 97.078 -78.285 46.852
131556 97.102 -78.297 46.831
131557 97.143 -78.303 46.811
131558 97.185 -78.309 46.791
131559 97.226 -78.315 46.771
131560 97.267 -78.321 46.749
131561 97.309 -78.326 46.727
131562 97.351 -78.332 46.707
131563 97.392 -78.338 46.688
131564 97.434 -78.344 46.670
131565 97.475 -78.351 46.651
131566 97.516 -78.356 46.632
131567 97.558 -78.362 46.616
131568 97.599 -78.368 46.601
131569 97.641 -78.374 46.584
131570 97.682 -78.372 46.565
131571 97.719 -78.373 46.547
131572 97.742 -78.377 46.527
131573 97.777 -78.401 46.508
131574 97.792 -78.409 46.490
131575 97.801 -78.410 46.471
131576 97.811 -78.410 46.453
131577 97.835 -78.405 46.435
131578 97.866 -78.398 46.417
131579 97.896 -78.390 46.403
131580 97.927 -78.382 46.388
131581 97.959 -78.373 46.371
131582 97.997 -78.360 46.355
131583 98.030 -78.358 46.338
131584 98.052 -78.365 46.322
131585 98.062 -78.374 46.305
131586 98.077 -78.363 46.288
131587 98.095 -78.344 46.272
131588 98.114 -78.327 46.256
131589 98.142 -78.318 46.239
131590 98.143 -78.326 46.221
131591 98.139 -78.337 46.203
131592 98.135 -78.347 46.187
131593 98.153 -78.361 46.169
131594 98.183 -78.351 46.151
131595 98.214 -78.338 46.134
131596 98.244 -78.327 46.115
131597 98.275 -78.315 46.096
131598 98.306 -78.304 46.078
131599 98.328 -78.329 46.061
131600 98.336 -78.319 46.042
131601 98.344 -78.309 46.022
131602 98.354 -78.300 46.003
131603 98.369 -78.322 45.984
131604 98.389 -78.366 45.963
131605 98.408 -78.410 45.942
131606 98.402 -78.406 45.920
131607 98.389 -78.394 45.899
131608 98.374 -78.407 45.878
131609 98.357 -78.454 45.857
131610 98.342 -78.501 45.837
131611 98.391 -78.514 45.816
131612 98.443 -78.509 45.796
131613 98.455 -78.495 45.777
131614 98.458 -78.483 45.757
131615 98.465 -78.483 45.736
131616 98.477 -78.499 45.713
131617 98.488 -78.516 45.690
131618 98.503 -78.531 45.666
131619 98.539 -78.514 45.644
131620 98.569 -78.484 45.623
131621 98.597 -78.452 45.601
131622 98.625 -78.420 45.581
131623 98.636 -78.404 45.560
131624 98.600 -78.435 45.539
131625 98.584 -78.466 45.520
131626 98.612 -78.481 45.501
131627 98.659 -78.463 45.483
131628 98.707 -78.445 45.466
131629 98.755 -78.426 45.448
131630 98.802 -78.407 45.430
131631 98.832 -78.405 45.412
131632 98.859 -78.402 45.396
131633 98.887 -78.399 45.379
131634 98.915 -78.396 45.363
131635 98.934 -78.381 45.347
131636 98.950 -78.426 45.331
131637 98.967 -78.471 45.317
131638 98.981 -78.507 45.302
131639 98.989 -78.520 45.287
131640 99.032 -78.526 45.274
131641 99.082 -78.515 45.261
131642 99.130 -78.500 45.248
131643 99.143 -78.503 45.235
131644 99.129 -78.522 45.222
131645 99.148 -78.526 45.209
131646 99.167 -78.532 45.196
131647 99.187 -78.536 45.183
131648 99.206 -78.542 45.170
131649 99.226 -78.546 45.157
131650 99.245 -78.551 45.145
131651 99.261 -78.559 45.132
131652 99.264 -78.579 45.120
131653 99.267 -78.598 45.108
131654 99.270 -78.617 45.098
131655 99.273 -78.637 45.087
131656 99.259 -78.658 45.076
131657 99.286 -78.659 45.063
131658 99.359 -78.643 45.048
131659 99.424 -78.663 45.033
131660 99.486 -78.704 45.018
131661 99.548 -78.744 45.004
131662 99.610 -78.785 44.990
131663 99.661 -78.806 44.977
131664 99.711 -78.824 44.964
131665 99.761 -78.841 44.952
131666 99.860 -78.803 44.940
131667 99.847 -78.832 44.930
131668 99.833 -78.861 44.918
131669 99.820 -78.896 44.906
131670 99.832 -78.905 44.894
131671 99.845 -78.910 44.883
131672 99.858 -78.915 44.872
131673 99.871 -78.919 44.862
131674 99.885 -78.924 44.852
131675 99.898 -78.928 44.844
131676 99.912 -78.933 44.835
131677 99.925 -78.937 44.826
131678 99.938 -78.943 44.818
131679 99.951 -78.947 44.810
131680 99.968 -78.950 44.803
131681 99.986 -78.950 44.797
131682 100.014 -78.967 44.791
131683 100.044 -78.985 44.785
131684 100.073 -79.003 44.779
131685 100.103 -79.022 44.773
131686 100.132 -79.040 44.769
131687 100.161 -79.059 44.764
131688 100.191 -79.077 44.758
131689 100.220 -79.095 44.752
131690 100.250 -79.113 44.746
131691 100.278 -79.124 44.740
131692 100.305 -79.130 44.734
131693 100.332 -79.135 44.727
131694 100.363 -79.142 44.719
131695 100.402 -79.153 44.711
131696 100.441 -79.165 44.703
131697 100.480 -79.182 44.694
131698 100.519 -79.199 44.685
131973 93.195 -78.175 54.043
131974 93.219 -78.163 54.036
131975 93.242 -78.151 54.026
131976 93.266 -78.139 54.016
131977 93.290 -78.128 54.006
131978 93.313 -78.116 53.996
131979 93.337 -78.103 53.987
131980 93.360 -78.092 53.977
131981 93.384 -78.080 53.967
131982 93.400 -78.073 53.954
131983 93.410 -78.071 53.942
131984 93.420 -78.068 53.930
131985 93.429 -78.065 53.917
131986 93.432 -78.059 53.904
131987 93.440 -78.044 53.891
131988 93.450 -78.028 53.878
131989 93.461 -78.013 53.863
131990 93.471 -77.997 53.847
131991 93.481 -77.982 53.832
131992 93.496 -77.967 53.816
131993 93.509 -77.958 53.801
131994 93.521 -77.949 53.786
131995 93.533 -77.941 53.770
131996 93.544 -77.933 53.753
131997 93.556 -77.924 53.735
131998 93.567 -77.916 53.718
131999 93.579 -77.908 53.701
132000 93.590 -77.899 53.682
132001 93.602 -77.891 53.665
132002 93.614 -77.884 53.649
132003 93.625 -77.875 53.629
132004 93.637 -77.867 53.610
132005 93.650 -77.859 53.590
132006 93.665 -77.852 53.572
132007 93.681 -77.844 53.552
132008 93.694 -77.839 53.532
132009 93.707 -77.835 53.512
132018 93.823 -77.796 53.334
132019 93.836 -77.792 53.315
132020 93.848 -77.788 53.296
132021 93.861 -77.783 53.278
132022 93.874 -77.779 53.260
132023 93.887 -77.775 53.242
132024 93.900 -77.771 53.224
132025 93.913 -77.766 53.207
132026 93.925 -77.762 53.190
132027 93.938 -77.758 53.173
132028 93.951 -77.753 53.157
132029 93.964 -77.749 53.142
132030 93.977 -77.745 53.127
132031 93.988 -77.741 53.113
132032 93.986 -77.740 53.098
132033 93.983 -77.739 53.082
132034 93.982 -77.736 53.065
132035 93.982 -77.733 53.045
132036 93.982 -77.730 53.026
132037 93.982 -77.727 53.006
132038 93.982 -77.724 52.985
132039 93.982 -77.721 52.961
132040 93.982 -77.718 52.938
132041 93.993 -77.721 52.916
132042 94.010 -77.726 52.894
132043 94.026 -77.730 52.874
132044 94.041 -77.736 52.853
132045 94.055 -77.741 52.832
132046 94.070 -77.745 52.810
132047 94.082 -77.737 52.788
132048 94.082 -77.730 52.765
132049 94.083 -77.721 52.742
132050 94.083 -77.714 52.718
132051 94.122 -77.687 52.695
132052 94.134 -77.684 52.672
132053 94.147 -77.685 52.648
132054 94.161 -77.686 52.627
132055 94.169 -77.690 52.606
132056 94.166 -77.699 52.585
132057 94.182 -77.699 52.565
132058 94.199 -77.696 52.546
132059 94.216 -77.695 52.528
132060 94.233 -77.693 52.510
132061 94.250 -77.691 52.493
132062 94.267 -77.689 52.477
132063 94.284 -77.687 52.462
132064 94.301 -77.686 52.449
132065 94.323 -77.674 52.437
132066 94.357 -77.631 52.426
132067 94.378 -77.589 52.419
132068 94.414 -77.572 52.414
132069 94.451 -77.555 52.409
132070 94.459 -77.564 52.402
132071 94.452 -77.588 52.396
132072 94.460 -77.595 52.386
132073 94.468 -77.604 52.372
132074 94.476 -77.611 52.353
132075 94.488 -77.604 52.332
132076 94.523 -77.600 52.311
132077 94.522 -77.651 52.290
132078 94.545 -77.709 52.269
132079 94.568 -77.768 52.251
132080 94.591 -77.827 52.232
132081 94.614 -77.886 52.213
132082 94.636 -77.943 52.192
132083 94.652 -77.975 52.170
132084 94.668 -77.962 52.148
132085 94.682 -77.948 52.125
132086 94.695 -77.925 52.100
132087 94.696 -77.868 52.071
132088 94.698 -77.810 52.040
132089 94.700 -77.752 52.008
132090 94.702 -77.696 51.975
132091 94.704 -77.638 51.941
132092 94.707 -77.583 51.908
132093 94.721 -77.574 51.872
132094 94.735 -77.565 51.838
132095 94.759 -77.575 51.803
132096 94.843 -77.599 51.766
132097 94.905 -77.557 51.729
132098 94.917 -77.523 51.692
132099 94.928 -77.491 51.656
132100 94.938 -77.457 51.623
132101 94.970 -77.495 51.589
132102 94.983 -77.538 51.553
132103 94.997 -77.580 51.514
132104 94.998 -77.605 51.474
132105 94.999 -77.570 51.436
132106 95.013 -77.557 51.399
132107 95.010 -77.552 51.360
132108 95.006 -77.548 51.319
132109 95.003 -77.544 51.278
132110 95.000 -77.539 51.237
132111 94.997 -77.535 51.193
132112 95.020 -77.572 51.149
132113 95.046 -77.615 51.106
132114 95.077 -77.658 51.064
132115 95.053 -77.705 51.018
132116 95.063 -77.636 50.973
132117 95.092 -77.614 50.929
132118 95.109 -77.619 50.886
132119 95.128 -77.627 50.842
132120 95.139 -77.630 50.799
132121 95.160 -77.636 50.757
132122 95.189 -77.656 50.716
132123 95.214 -77.659 50.677
132124 95.235 -77.645 50.638
132125 95.260 -77.633 50.600
132126 95.286 -77.622 50.564
132127 95.312 -77.611 50.527
132128 95.337 -77.600 50.490
132129 95.363 -77.589 50.454
132130 95.389 -77.577 50.418
132131 95.414 -77.567 50.384
132132 95.440 -77.555 50.348
132133 95.471 -77.554 50.314
132134 95.486 -77.545 50.279
132135 95.502 -77.536 50.244
132136 95.520 -77.517 50.209
132137 95.538 -77.495 50.174
132138 95.556 -77.480 50.141
132139 95.574 -77.472 50.107
132140 95.592 -77.464 50.074
132141 95.629 -77.467 50.041
132142 95.681 -77.480 50.008
132143 95.733 -77.493 49.976
132144 95.815 -77.542 49.944
132145 95.800 -77.558 49.913
132146 95.781 -77.569 49.882
132147 95.752 -77.575 49.852
132148 95.723 -77.580 49.823
132149 95.694 -77.586 49.795
132150 95.676 -77.607 49.768
132151 95.680 -77.609 49.740
132152 95.683 -77.612 49.711
132153 95.686 -77.614 49.682
132154 95.690 -77.617 49.650
132155 95.693 -77.620 49.619
132156 95.697 -77.623 49.587
132157 95.700 -77.626 49.556
132158 95.704 -77.628 49.527
132159 95.707 -77.631 49.497
132160 95.711 -77.633 49.468
132161 95.714 -77.636 49.439
132162 95.719 -77.639 49.409
132163 95.729 -77.639 49.379
132164 95.739 -77.639 49.347
132165 95.749 -77.640 49.316
132166 95.758 -77.641 49.284
132167 95.768 -77.641 49.251
132168 95.790 -77.651 49.214
132169 95.815 -77.663 49.178
132170 95.840 -77.676 49.142
132171 95.863 -77.684 49.105
132172 95.886 -77.692 49.067
132173 95.908 -77.701 49.030
132174 95.931 -77.709 48.992
132175 95.954 -77.718 48.954
132176 95.977 -77.726 48.918
132177 96.000 -77.734 48.881
132178 96.022 -77.743 48.845
132179 96.047 -77.752 48.809
132180 96.074 -77.761 48.773
132181 96.101 -77.771 48.737
132182 96.127 -77.780 48.702
132183 96.145 -77.790 48.669
132184 96.155 -77.799 48.636
132185 96.165 -77.809 48.604
132186 96.174 -77.822 48.571
132187 96.179 -77.845 48.540
132188 96.184 -77.868 48.508
132189 96.189 -77.890 48.477
132190 96.194 -77.913 48.448
132191 96.197 -77.931 48.419
132192 96.203 -77.931 48.388
132193 96.206 -77.925 48.358
132194 96.208 -77.920 48.328
132195 96.210 -77.915 48.298
132196 96.213 -77.909 48.271
132197 96.216 -77.904 48.244
132198 96.218 -77.899 48.218
132199 96.220 -77.893 48.191
132200 96.223 -77.888 48.165
132201 96.225 -77.882 48.140
132202 96.227 -77.877 48.115
132203 96.227 -77.873 48.092
132204 96.224 -77.869 48.071
132205 96.240 -77.865 48.051
132206 96.261 -77.861 48.030
132207 96.282 -77.856 48.008
132208 96.302 -77.852 47.984
132209 96.324 -77.854 47.961
132210 96.345 -77.859 47.939
132211 96.367 -77.865 47.917
132212 96.389 -77.870 47.895
132213 96.411 -77.875 47.873
132214 96.433 -77.880 47.850
132215 96.454 -77.890 47.827
132216 96.475 -77.902 47.805
132217 96.496 -77.915 47.784
132218 96.517 -77.927 47.762
132219 96.543 -77.935 47.739
132220 96.551 -77.934 47.716
132221 96.558 -77.931 47.692
132222 96.569 -77.933 47.666
132223 96.580 -77.934 47.640
132224 96.591 -77.936 47.614
132225 96.601 -77.937 47.589
132226 96.612 -77.939 47.564
132227 96.623 -77.940 47.540
132228 96.633 -77.942 47.517
132229 96.644 -77.943 47.493
132230 96.657 -77.952 47.469
132231 96.671 -77.969 47.445
132232 96.685 -77.986 47.420
132233 96.700 -78.003 47.397
132234 96.715 -78.021 47.373
132235 96.742 -78.035 47.348
132236 96.769 -78.050 47.323
132237 96.797 -78.065 47.298
132238 96.824 -78.080 47.271
132239 96.848 -78.095 47.244
132240 96.850 -78.109 47.217
132241 96.856 -78.125 47.189
132242 96.862 -78.143 47.161
132243 96.868 -78.160 47.136
132244 96.882 -78.173 47.110
132245 96.902 -78.180 47.084
132246 96.917 -78.186 47.059
132247 96.932 -78.193 47.034
132248 96.947 -78.199 47.011
132249 96.961 -78.206 46.989
132250 96.978 -78.214 46.969
132251 96.996 -78.224 46.951
132252 97.014 -78.233 46.933
132253 97.031 -78.243 46.916
132254 97.048 -78.253 46.901
132255 97.066 -78.262 46.885
132256 97.084 -78.272 46.867
132257 97.106 -78.282 46.847
132258 97.147 -78.288 46.828
132259 97.188 -78.292 46.809
132260 97.229 -78.297 46.789
132261 97.271 -78.303 46.768
132262 97.312 -78.309 46.748
132263 97.354 -78.315 46.729
132264 97.395 -78.321 46.711
132265 97.437 -78.326 46.695
132266 97.478 -78.332 46.676
132267 97.520 -78.338 46.658
132268 97.561 -78.344 46.641
132269 97.602 -78.351 46.626
132270 97.644 -78.357 46.609
132271 97.685 -78.362 46.591
132272 97.729 -78.363 46.572
132273 97.754 -78.367 46.552
132274 97.776 -78.382 46.532
132275 97.785 -78.382 46.511
132276 97.794 -78.383 46.493
132277 97.804 -78.384 46.475
132278 97.828 -78.379 46.458
132279 97.858 -78.371 46.441
132280 97.889 -78.363 46.426
132281 97.919 -78.355 46.411
132282 97.949 -78.347 46.395
132283 97.982 -78.338 46.379
132284 98.018 -78.331 46.361
132285 98.036 -78.338 46.344
132286 98.045 -78.347 46.327
132287 98.059 -78.344 46.309
132288 98.077 -78.326 46.292
132289 98.105 -78.317 46.276
132290 98.133 -78.308 46.259
132291 98.141 -78.311 46.241
132292 98.137 -78.322 46.223
132293 98.137 -78.334 46.205
132294 98.156 -78.347 46.187
132295 98.185 -78.338 46.169
132296 98.216 -78.326 46.152
132297 98.247 -78.314 46.132
132298 98.278 -78.302 46.114
132299 98.305 -78.311 46.096
132300 98.320 -78.341 46.077
132301 98.328 -78.332 46.058
132302 98.336 -78.322 46.039
132303 98.344 -78.313 46.020
132304 98.352 -78.303 46.000
132305 98.367 -78.317 45.979
132306 98.386 -78.360 45.958
132307 98.390 -78.376 45.937
132308 98.377 -78.357 45.915
132309 98.359 -78.406 45.893
132310 98.341 -78.454 45.872
132311 98.324 -78.502 45.851
132312 98.363 -78.520 45.830
132313 98.416 -78.499 45.810
132314 98.464 -78.479 45.791
132315 98.468 -78.467 45.772
132316 98.473 -78.460 45.752
132317 98.484 -78.476 45.729
132318 98.496 -78.492 45.706
132319 98.521 -78.501 45.681
132320 98.558 -78.485 45.658
132321 98.594 -78.463 45.636
132322 98.628 -78.438 45.614
132323 98.655 -78.406 45.594
132324 98.666 -78.390 45.573
132325 98.631 -78.421 45.552
132326 98.595 -78.451 45.533
132327 98.619 -78.456 45.514
132328 98.666 -78.438 45.496
132329 98.718 -78.416 45.479
132330 98.771 -78.392 45.461
132331 98.816 -78.376 45.443
132332 98.842 -78.378 45.426
132333 98.876 -78.368 45.410
132334 98.904 -78.365 45.393
132335 98.932 -78.362 45.377
132336 98.953 -78.360 45.361
132337 98.969 -78.405 45.345
132338 98.986 -78.450 45.331
132339 99.002 -78.495 45.316
132340 99.013 -78.519 45.302
132341 99.062 -78.514 45.288
132342 99.110 -78.498 45.275
132343 99.158 -78.482 45.263
132344 99.177 -78.482 45.250
132345 99.145 -78.509 45.237
132346 99.159 -78.516 45.223
132347 99.178 -78.521 45.210
132348 99.198 -78.526 45.196
132349 99.217 -78.531 45.183
132350 99.236 -78.536 45.170
132351 99.256 -78.541 45.158
132352 99.261 -78.558 45.145
132353 99.264 -78.578 45.133
132354 99.267 -78.597 45.122
132355 99.270 -78.617 45.110
132356 99.273 -78.636 45.100
132357 99.273 -78.656 45.089
132358 99.292 -78.661 45.076
132359 99.365 -78.645 45.062
132360 99.436 -78.634 45.048
132361 99.498 -78.674 45.033
132362 99.560 -78.714 45.018
132363 99.622 -78.755 45.004
132364 99.683 -78.796 44.991
132365 99.741 -78.829 44.979
132366 99.842 -78.798 44.968
132367 99.885 -78.790 44.957
132368 99.871 -78.819 44.945
132369 99.857 -78.847 44.934
132370 99.844 -78.876 44.922
132371 99.832 -78.918 44.911
132372 99.845 -78.923 44.899
132373 99.859 -78.927 44.887
132374 99.872 -78.932 44.876
132375 99.886 -78.937 44.866
132376 99.899 -78.941 44.856
132377 99.912 -78.946 44.847
132378 99.930 -78.949 44.838
132379 99.950 -78.952 44.830
132380 99.970 -78.955 44.821
132381 99.979 -78.942 44.814
132382 99.994 -78.939 44.807
132383 100.020 -78.951 44.800
132384 100.046 -78.963 44.794
132385 100.075 -78.982 44.788
132386 100.105 -79.000 44.782
132387 100.134 -79.018 44.776
132388 100.163 -79.037 44.771
132389 100.193 -79.055 44.765
132390 100.222 -79.073 44.759
132391 100.252 -79.091 44.753
132392 100.280 -79.104 44.747
132393 100.308 -79.109 44.740
132394 100.335 -79.115 44.732
132395 100.370 -79.125 44.725
132396 100.409 -79.135 44.717
132397 100.448 -79.152 44.708
132398 100.487 -79.169 44.698
132399 100.527 -79.185 44.689
132400 100.564 -79.203 44.678
132673 93.171 -78.176 54.059
132674 93.195 -78.165 54.051
132675 93.218 -78.153 54.044
132676 93.241 -78.141 54.034
132677 93.265 -78.129 54.025
132678 93.289 -78.117 54.015
132679 93.312 -78.105 54.005
132680 93.336 -78.094 53.996
132681 93.360 -78.081 53.986
132682 93.383 -78.069 53.976
132683 93.401 -78.062 53.963
132684 93.410 -78.059 53.951
132685 93.420 -78.057 53.939
132686 93.430 -78.055 53.926
132687 93.431 -78.044 53.913
132688 93.442 -78.028 53.900
132689 93.452 -78.012 53.888
132690 93.462 -77.997 53.872
132691 93.473 -77.982 53.856
132692 93.483 -77.966 53.841
132693 93.499 -77.953 53.826
132694 93.510 -77.945 53.810
132695 93.522 -77.937 53.795
132696 93.534 -77.928 53.779
132697 93.545 -77.920 53.761
132698 93.557 -77.912 53.745
132699 93.568 -77.903 53.728
132700 93.580 -77.895 53.709
132701 93.591 -77.887 53.692
132702 93.603 -77.878 53.674
132703 93.615 -77.870 53.657
132704 93.626 -77.862 53.638
132705 93.638 -77.854 53.619
132706 93.652 -77.846 53.600
132707 93.667 -77.839 53.582
132708 93.682 -77.832 53.562
132709 93.695 -77.827 53.542
132710 93.708 -77.824 53.522
132719 93.823 -77.785 53.345
132720 93.836 -77.780 53.326
132721 93.849 -77.777 53.307
132722 93.862 -77.772 53.290
132723 93.874 -77.768 53.272
132724 93.888 -77.764 53.254
132725 93.900 -77.759 53.237
132726 93.913 -77.755 53.219
132727 93.926 -77.751 53.202
132728 93.939 -77.746 53.185
132729 93.952 -77.743 53.170
132730 93.965 -77.738 53.155
132731 93.978 -77.733 53.141
132732 93.989 -77.730 53.129
132733 93.989 -77.727 53.114
132734 93.989 -77.724 53.097
132735 93.989 -77.721 53.078
132736 93.989 -77.718 53.058
132737 93.989 -77.714 53.040
132738 93.989 -77.711 53.021
132739 93.989 -77.708 53.000
132740 93.989 -77.705 52.980
132741 93.989 -77.702 52.959
132742 93.994 -77.702 52.937
132743 94.008 -77.707 52.915
132744 94.023 -77.711 52.893
132745 94.037 -77.717 52.872
132746 94.052 -77.722 52.851
132747 94.066 -77.726 52.830
132748 94.083 -77.718 52.808
132749 94.090 -77.711 52.785
132750 94.091 -77.702 52.762
132751 94.091 -77.695 52.737
132752 94.121 -77.667 52.712
132753 94.124 -77.672 52.688
132754 94.137 -77.673 52.665
132755 94.151 -77.674 52.643
132756 94.165 -77.675 52.622
132757 94.178 -77.676 52.602
132758 94.192 -77.677 52.583
132759 94.205 -77.678 52.564
132760 94.219 -77.679 52.546
132761 94.232 -77.680 52.529
132762 94.246 -77.681 52.513
132763 94.260 -77.682 52.498
132764 94.273 -77.683 52.484
132765 94.287 -77.684 52.473
132766 94.298 -77.638 52.462
132767 94.323 -77.602 52.454
132768 94.360 -77.585 52.449
132769 94.396 -77.568 52.447
132770 94.433 -77.551 52.444
132771 94.469 -77.535 52.439
132772 94.500 -77.524 52.432
132773 94.488 -77.551 52.421
132774 94.477 -77.579 52.405
132775 94.474 -77.598 52.385
132776 94.485 -77.597 52.364
132777 94.505 -77.644 52.344
132778 94.528 -77.703 52.325
132779 94.551 -77.761 52.308
132780 94.573 -77.821 52.291
132781 94.596 -77.880 52.273
132782 94.619 -77.938 52.254
132783 94.642 -77.997 52.233
132784 94.658 -78.010 52.211
132785 94.672 -77.994 52.189
132786 94.674 -77.937 52.166
132787 94.676 -77.880 52.138
132788 94.678 -77.822 52.109
132789 94.680 -77.765 52.077
132790 94.682 -77.707 52.044
132791 94.684 -77.649 52.010
132792 94.686 -77.592 51.977
132793 94.694 -77.561 51.945
132794 94.708 -77.551 51.910
132795 94.722 -77.542 51.874
132796 94.741 -77.542 51.837
132797 94.853 -77.542 51.798
132798 94.923 -77.498 51.760
132799 94.934 -77.465 51.724
132800 94.943 -77.477 51.690
132801 94.945 -77.451 51.659
132802 94.961 -77.465 51.624
132803 94.982 -77.496 51.585
132804 95.002 -77.527 51.544
132805 95.007 -77.545 51.502
132806 95.001 -77.553 51.462
132807 95.023 -77.569 51.423
132808 94.991 -77.545 51.383
132809 94.988 -77.540 51.344
132810 94.985 -77.536 51.306
132811 94.982 -77.531 51.266
132812 94.978 -77.526 51.222
132813 94.999 -77.561 51.181
132814 95.035 -77.603 51.139
132815 95.085 -77.645 51.095
132816 94.991 -77.711 51.048
132817 95.079 -77.662 51.001
132818 95.104 -77.646 50.958
132819 95.113 -77.630 50.914
132820 95.121 -77.614 50.870
132821 95.137 -77.609 50.826
132822 95.167 -77.636 50.783
132823 95.195 -77.652 50.743
132824 95.216 -77.637 50.704
132825 95.237 -77.623 50.665
132826 95.261 -77.611 50.627
132827 95.287 -77.599 50.591
132828 95.313 -77.588 50.554
132829 95.338 -77.577 50.516
132830 95.364 -77.566 50.479
132831 95.390 -77.555 50.444
132832 95.415 -77.543 50.409
132833 95.447 -77.542 50.374
132834 95.479 -77.543 50.339
132835 95.495 -77.534 50.304
132836 95.511 -77.526 50.268
132837 95.526 -77.517 50.232
132838 95.542 -77.508 50.197
132839 95.557 -77.498 50.163
132840 95.575 -77.482 50.128
132841 95.604 -77.480 50.093
132842 95.655 -77.493 50.059
132843 95.707 -77.506 50.025
132844 95.759 -77.519 49.992
132845 95.824 -77.552 49.959
132846 95.808 -77.567 49.927
132847 95.779 -77.573 49.897
132848 95.750 -77.578 49.866
132849 95.722 -77.584 49.836
132850 95.693 -77.589 49.807
132851 95.676 -77.608 49.780
132852 95.680 -77.611 49.752
132853 95.683 -77.614 49.723
132854 95.687 -77.616 49.693
132855 95.690 -77.619 49.663
132856 95.694 -77.622 49.631
132857 95.697 -77.624 49.600
132858 95.701 -77.627 49.569
132859 95.704 -77.630 49.538
132860 95.708 -77.633 49.508
132861 95.711 -77.635 49.478
132862 95.717 -77.637 49.448
132863 95.727 -77.638 49.417
132864 95.737 -77.638 49.386
132865 95.747 -77.639 49.354
132866 95.757 -77.639 49.324
132867 95.767 -77.639 49.292
132868 95.776 -77.640 49.259
132869 95.792 -77.645 49.223
132870 95.817 -77.658 49.189
132871 95.842 -77.668 49.154
132872 95.865 -77.677 49.117
132873 95.887 -77.685 49.080
132874 95.910 -77.694 49.043
132875 95.933 -77.702 49.005
132876 95.955 -77.711 48.967
132877 95.978 -77.719 48.931
132878 96.005 -77.728 48.895
132879 96.032 -77.738 48.858
132880 96.059 -77.747 48.823
132881 96.086 -77.756 48.786
132882 96.112 -77.766 48.750
132883 96.126 -77.776 48.714
132884 96.136 -77.786 48.681
132885 96.147 -77.796 48.648
132886 96.156 -77.805 48.615
132887 96.165 -77.819 48.583
132888 96.170 -77.842 48.551
132889 96.175 -77.865 48.520
132890 96.180 -77.887 48.489
132891 96.185 -77.910 48.459
132892 96.197 -77.925 48.429
132893 96.197 -77.920 48.399
132894 96.196 -77.912 48.368
132895 96.195 -77.905 48.337
132896 96.194 -77.897 48.307
132897 96.195 -77.890 48.279
132898 96.198 -77.885 48.252
132899 96.200 -77.880 48.226
132900 96.203 -77.874 48.200
132901 96.205 -77.869 48.174
132902 96.207 -77.864 48.149
132903 96.210 -77.859 48.125
132904 96.209 -77.854 48.102
132905 96.220 -77.850 48.082
132906 96.240 -77.846 48.061
132907 96.261 -77.841 48.040
132908 96.282 -77.837 48.017
132909 96.302 -77.832 47.994
132910 96.324 -77.835 47.971
132911 96.345 -77.840 47.947
132912 96.367 -77.845 47.925
132913 96.388 -77.854 47.903
132914 96.409 -77.867 47.880
132915 96.431 -77.879 47.857
132916 96.452 -77.891 47.833
132917 96.473 -77.904 47.811
132918 96.494 -77.916 47.788
132919 96.517 -77.928 47.765
132920 96.544 -77.934 47.742
132921 96.552 -77.933 47.718
132922 96.559 -77.931 47.694
132923 96.568 -77.929 47.669
132924 96.577 -77.930 47.643
132925 96.588 -77.931 47.618
132926 96.599 -77.933 47.594
132927 96.609 -77.934 47.570
132928 96.620 -77.935 47.546
132929 96.631 -77.937 47.522
132930 96.642 -77.938 47.498
132931 96.653 -77.942 47.474
132932 96.667 -77.959 47.450
132933 96.682 -77.976 47.426
132934 96.696 -77.993 47.402
132935 96.713 -78.010 47.378
132936 96.741 -78.025 47.353
132937 96.768 -78.040 47.329
132938 96.795 -78.055 47.303
132939 96.838 -78.072 47.278
132940 96.854 -78.085 47.251
132941 96.853 -78.097 47.223
132942 96.854 -78.110 47.195
132943 96.860 -78.128 47.168
132944 96.866 -78.145 47.144
132945 96.881 -78.157 47.118
132946 96.905 -78.165 47.093
132947 96.921 -78.171 47.068
132948 96.936 -78.178 47.044
132949 96.950 -78.184 47.021
132950 96.966 -78.192 46.999
132951 96.984 -78.201 46.979
132952 97.002 -78.211 46.962
132953 97.019 -78.220 46.945
132954 97.036 -78.230 46.928
132955 97.053 -78.240 46.912
132956 97.070 -78.253 46.897
132957 97.087 -78.265 46.880
132958 97.103 -78.278 46.864
132959 97.156 -78.283 46.845
132960 97.197 -78.288 46.826
132961 97.238 -78.293 46.807
132962 97.279 -78.298 46.788
132963 97.320 -78.303 46.768
132964 97.361 -78.308 46.751
132965 97.402 -78.313 46.735
132966 97.443 -78.318 46.719
132967 97.485 -78.322 46.700
132968 97.525 -78.328 46.683
132969 97.567 -78.332 46.667
132970 97.607 -78.338 46.651
132971 97.649 -78.342 46.634
132972 97.689 -78.347 46.616
132973 97.731 -78.352 46.598
132974 97.759 -78.354 46.578
132975 97.769 -78.354 46.558
132976 97.778 -78.355 46.537
132977 97.787 -78.356 46.519
132978 97.797 -78.357 46.500
132979 97.820 -78.351 46.482
132980 97.851 -78.344 46.465
132981 97.881 -78.336 46.449
132982 97.912 -78.329 46.434
132983 97.942 -78.321 46.418
132984 97.972 -78.313 46.401
132985 98.005 -78.304 46.384
132986 98.020 -78.311 46.367
132987 98.029 -78.320 46.349
132988 98.040 -78.326 46.331
132989 98.068 -78.317 46.313
132990 98.095 -78.307 46.296
132991 98.123 -78.298 46.279
132992 98.139 -78.297 46.261
132993 98.135 -78.308 46.243
132994 98.140 -78.319 46.224
132995 98.159 -78.333 46.205
132996 98.188 -78.325 46.187
132997 98.219 -78.313 46.169
132998 98.249 -78.301 46.150
132999 98.280 -78.290 46.131
133000 98.299 -78.341 46.112
133001 98.311 -78.355 46.093
133002 98.319 -78.345 46.074
133003 98.327 -78.335 46.054
133004 98.336 -78.326 46.035
133005 98.344 -78.316 46.015
133006 98.352 -78.306 45.993
133007 98.364 -78.311 45.972
133008 98.380 -78.347 45.951
133009 98.365 -78.345 45.930
133010 98.346 -78.398 45.909
133011 98.327 -78.450 45.888
133012 98.308 -78.501 45.866
133013 98.336 -78.511 45.846
133014 98.388 -78.490 45.825
133015 98.441 -78.469 45.805
133016 98.478 -78.451 45.786
133017 98.481 -78.439 45.766
133018 98.492 -78.453 45.743
133019 98.503 -78.470 45.720
133020 98.540 -78.473 45.695
133021 98.577 -78.457 45.671
133022 98.613 -78.435 45.649
133023 98.649 -78.412 45.627
133024 98.685 -78.389 45.606
133025 98.697 -78.376 45.585
133026 98.662 -78.401 45.565
133027 98.630 -78.416 45.546
133028 98.647 -78.408 45.528
133029 98.700 -78.385 45.510
133030 98.752 -78.360 45.493
133031 98.801 -78.341 45.475
133032 98.820 -78.354 45.458
133033 98.839 -78.367 45.441
133034 98.873 -78.358 45.423
133035 98.911 -78.342 45.407
133036 98.948 -78.327 45.391
133037 98.972 -78.338 45.375
133038 98.988 -78.384 45.359
133039 99.004 -78.429 45.344
133040 99.021 -78.475 45.330
133041 99.042 -78.512 45.316
133042 99.090 -78.496 45.303
133043 99.138 -78.481 45.290
133044 99.186 -78.466 45.277
133045 99.210 -78.462 45.265
133046 99.179 -78.489 45.252
133047 99.169 -78.505 45.238
133048 99.189 -78.510 45.226
133049 99.208 -78.516 45.212
133050 99.228 -78.520 45.198
133051 99.247 -78.526 45.185
133052 99.258 -78.538 45.173
133053 99.261 -78.557 45.160
133054 99.264 -78.576 45.148
133055 99.267 -78.596 45.136
133056 99.270 -78.616 45.124
133057 99.273 -78.636 45.113
133058 99.276 -78.655 45.101
133059 99.298 -78.662 45.088
133060 99.371 -78.646 45.074
133061 99.443 -78.630 45.060
133062 99.510 -78.645 45.046
133063 99.572 -78.686 45.031
133064 99.634 -78.726 45.018
133065 99.693 -78.764 45.005
133066 99.746 -78.808 44.993
133067 99.800 -78.843 44.982
133068 99.882 -78.818 44.971
133069 99.887 -78.827 44.961
133070 99.878 -78.843 44.949
133071 99.868 -78.862 44.938
133072 99.854 -78.891 44.926
133073 99.846 -78.935 44.914
133074 99.859 -78.940 44.901
133075 99.873 -78.945 44.889
133076 99.891 -78.948 44.879
133077 99.911 -78.950 44.867
133078 99.931 -78.953 44.858
133079 99.951 -78.956 44.849
133080 99.971 -78.959 44.841
133081 99.981 -78.947 44.832
133082 99.987 -78.928 44.824
133083 100.003 -78.927 44.816
133084 100.029 -78.939 44.809
133085 100.054 -78.951 44.802
133086 100.080 -78.963 44.796
133087 100.107 -78.978 44.790
133088 100.136 -78.996 44.784
133089 100.166 -79.015 44.778
133090 100.195 -79.033 44.772
133091 100.225 -79.051 44.766
133092 100.254 -79.069 44.760
133093 100.283 -79.084 44.753
133094 100.311 -79.090 44.745
133095 100.339 -79.096 44.738
133096 100.378 -79.106 44.730
133097 100.416 -79.122 44.721
133098 100.456 -79.138 44.712
133099 100.495 -79.155 44.702
133100 100.530 -79.173 44.692
133101 100.562 -79.193 44.681
133102 100.595 -79.212 44.670
133373 93.146 -78.178 54.075
133374 93.170 -78.166 54.068
133375 93.194 -78.154 54.060
133376 93.217 -78.142 54.053
133377 93.241 -78.131 54.043
133378 93.265 -78.119 54.034
133379 93.288 -78.106 54.024
133380 93.312 -78.095 54.015
133381 93.336 -78.083 54.006
133382 93.359 -78.071 53.996
133383 93.383 -78.059 53.986
133384 93.401 -78.051 53.973
133385 93.411 -78.049 53.960
133386 93.420 -78.047 53.948
133387 93.426 -78.041 53.935
133388 93.433 -78.028 53.922
133389 93.443 -78.012 53.910
133390 93.453 -77.997 53.897
133391 93.464 -77.981 53.882
133392 93.474 -77.966 53.866
133393 93.485 -77.950 53.850
133394 93.500 -77.940 53.836
133395 93.512 -77.931 53.820
133396 93.523 -77.924 53.805
133397 93.535 -77.915 53.789
133398 93.547 -77.907 53.771
133399 93.558 -77.899 53.754
133400 93.570 -77.890 53.737
133401 93.581 -77.882 53.719
133402 93.593 -77.874 53.702
133403 93.605 -77.865 53.684
133404 93.616 -77.857 53.667
133405 93.628 -77.849 53.648
133406 93.639 -77.841 53.628
133407 93.653 -77.834 53.609
133408 93.669 -77.826 53.591
133409 93.683 -77.821 53.572
133410 93.695 -77.817 53.552
133411 93.708 -77.812 53.532
133412 93.721 -77.808 53.512
133420 93.824 -77.774 53.356
133421 93.837 -77.770 53.337
133422 93.850 -77.765 53.319
133423 93.862 -77.761 53.302
133424 93.875 -77.757 53.284
133425 93.888 -77.752 53.266
133426 93.901 -77.748 53.249
133427 93.914 -77.744 53.232
133428 93.927 -77.740 53.215
133429 93.939 -77.735 53.198
133430 93.949 -77.730 53.182
133431 93.961 -77.725 53.167
133432 93.971 -77.721 53.154
133433 93.985 -77.716 53.142
133434 93.990 -77.718 53.127
133435 93.995 -77.719 53.110
133436 93.998 -77.718 53.091
133437 93.998 -77.715 53.072
133438 93.998 -77.712 53.054
133439 93.998 -77.708 53.036
133440 93.997 -77.705 53.017
133441 93.997 -77.702 52.997
133442 93.997 -77.699 52.976
133443 94.001 -77.695 52.955
133444 94.007 -77.689 52.932
133445 94.019 -77.692 52.910
133446 94.034 -77.698 52.889
133447 94.049 -77.702 52.868
133448 94.063 -77.707 52.847
133449 94.080 -77.699 52.826
133450 94.097 -77.692 52.803
133451 94.099 -77.684 52.781
133452 94.114 -77.665 52.755
133453 94.116 -77.648 52.729
133454 94.120 -77.655 52.704
133455 94.127 -77.661 52.681
133456 94.141 -77.662 52.659
133457 94.154 -77.663 52.638
133458 94.168 -77.664 52.619
133459 94.182 -77.665 52.600
133460 94.195 -77.666 52.582
133461 94.209 -77.667 52.565
133462 94.222 -77.668 52.548
133463 94.236 -77.669 52.533
133464 94.259 -77.659 52.520
133465 94.298 -77.631 52.507
133466 94.344 -77.625 52.497
133467 94.341 -77.620 52.489
133468 94.362 -77.602 52.482
133469 94.384 -77.584 52.480
133470 94.405 -77.567 52.479
133471 94.426 -77.548 52.477
133472 94.451 -77.531 52.473
133473 94.488 -77.514 52.466
133474 94.524 -77.498 52.453
133475 94.529 -77.511 52.436
133476 94.518 -77.538 52.416
133477 94.490 -77.574 52.398
133478 94.488 -77.666 52.379
133479 94.509 -77.755 52.363
133480 94.536 -77.815 52.349
133481 94.564 -77.873 52.332
133482 94.591 -77.932 52.314
133483 94.618 -77.990 52.296
133484 94.645 -78.050 52.274
133485 94.652 -78.006 52.253
133486 94.654 -77.949 52.231
133487 94.655 -77.891 52.207
133488 94.657 -77.834 52.179
133489 94.660 -77.776 52.148
133490 94.662 -77.719 52.116
133491 94.664 -77.661 52.080
133492 94.666 -77.604 52.046
133493 94.668 -77.546 52.012
133494 94.681 -77.537 51.981
133495 94.695 -77.528 51.945
133496 94.709 -77.519 51.907
133497 94.723 -77.510 51.868
133498 94.861 -77.489 51.828
133499 94.932 -77.553 51.792
133500 94.940 -77.572 51.759
133501 94.942 -77.539 51.728
133502 94.943 -77.495 51.695
133503 94.943 -77.452 51.659
133504 94.964 -77.483 51.621
133505 94.984 -77.514 51.579
133506 94.993 -77.535 51.538
133507 94.992 -77.545 51.498
133508 95.019 -77.564 51.460
133509 95.045 -77.583 51.421
133510 95.072 -77.602 51.382
133511 95.037 -77.577 51.342
133512 94.997 -77.546 51.302
133513 94.960 -77.519 51.260
133514 95.006 -77.560 51.217
133515 95.091 -77.632 51.172
133516 95.116 -77.650 51.125
133517 95.005 -77.694 51.078
133518 95.093 -77.643 51.032
133519 95.114 -77.639 50.986
133520 95.125 -77.642 50.941
133521 95.135 -77.629 50.897
133522 95.145 -77.615 50.853
133523 95.176 -77.644 50.810
133524 95.197 -77.630 50.770
133525 95.218 -77.614 50.731
133526 95.239 -77.600 50.693
133527 95.262 -77.588 50.654
133528 95.287 -77.577 50.617
133529 95.313 -77.565 50.580
133530 95.339 -77.554 50.543
133531 95.365 -77.542 50.508
133532 95.390 -77.532 50.472
133533 95.422 -77.532 50.436
133534 95.455 -77.532 50.401
133535 95.486 -77.532 50.366
133536 95.504 -77.524 50.330
133537 95.519 -77.515 50.293
133538 95.535 -77.506 50.256
133539 95.550 -77.497 50.220
133540 95.566 -77.489 50.186
133541 95.582 -77.479 50.151
133542 95.629 -77.494 50.114
133543 95.681 -77.512 50.078
133544 95.733 -77.530 50.043
133545 95.785 -77.545 50.009
133546 95.833 -77.563 49.975
133547 95.806 -77.570 49.943
133548 95.777 -77.576 49.912
133549 95.748 -77.581 49.880
133550 95.720 -77.587 49.849
133551 95.691 -77.592 49.820
133552 95.677 -77.610 49.791
133553 95.680 -77.612 49.763
133554 95.684 -77.615 49.733
133555 95.687 -77.617 49.703
133556 95.691 -77.620 49.673
133557 95.694 -77.624 49.644
133558 95.698 -77.626 49.613
133559 95.701 -77.629 49.581
133560 95.705 -77.631 49.550
133561 95.708 -77.634 49.518
133562 95.716 -77.636 49.487
133563 95.725 -77.636 49.456
133564 95.735 -77.636 49.425
133565 95.745 -77.637 49.394
133566 95.755 -77.637 49.362
133567 95.765 -77.638 49.331
133568 95.775 -77.639 49.300
133569 95.785 -77.639 49.267
133570 95.795 -77.639 49.232
133571 95.820 -77.652 49.200
133572 95.843 -77.661 49.166
133573 95.866 -77.670 49.130
133574 95.889 -77.678 49.094
133575 95.911 -77.686 49.057
133576 95.937 -77.696 49.019
133577 95.963 -77.705 48.981
133578 95.990 -77.714 48.944
133579 96.017 -77.724 48.908
133580 96.044 -77.733 48.871
133581 96.071 -77.743 48.836
133582 96.098 -77.752 48.800
133583 96.108 -77.761 48.763
133584 96.118 -77.771 48.727
133585 96.128 -77.782 48.694
133586 96.138 -77.792 48.661
133587 96.148 -77.802 48.627
133588 96.156 -77.817 48.595
133589 96.161 -77.839 48.563
133590 96.166 -77.862 48.531
133591 96.171 -77.884 48.501
133592 96.180 -77.902 48.470
133593 96.194 -77.914 48.440
133594 96.199 -77.914 48.409
133595 96.198 -77.906 48.378
133596 96.197 -77.899 48.347
133597 96.196 -77.891 48.317
133598 96.195 -77.884 48.287
133599 96.194 -77.876 48.261
133600 96.194 -77.868 48.234
133601 96.193 -77.861 48.209
133602 96.192 -77.853 48.183
133603 96.191 -77.846 48.159
133604 96.192 -77.840 48.135
133605 96.198 -77.835 48.113
133606 96.219 -77.830 48.092
133607 96.240 -77.826 48.071
133608 96.260 -77.821 48.050
133609 96.281 -77.817 48.027
133610 96.302 -77.812 48.004
133611 96.323 -77.818 47.980
133612 96.344 -77.831 47.956
133613 96.365 -77.843 47.933
133614 96.386 -77.856 47.911
133615 96.407 -77.868 47.888
133616 96.429 -77.881 47.864
133617 96.450 -77.893 47.840
133618 96.471 -77.906 47.817
133619 96.492 -77.918 47.794
133620 96.519 -77.928 47.771
133621 96.545 -77.934 47.746
133622 96.553 -77.932 47.722
133623 96.561 -77.931 47.697
133624 96.569 -77.928 47.673
133625 96.576 -77.927 47.648
133626 96.585 -77.926 47.623
133627 96.596 -77.928 47.599
133628 96.607 -77.929 47.575
133629 96.617 -77.931 47.551
133630 96.628 -77.932 47.527
133631 96.639 -77.934 47.502
133632 96.650 -77.935 47.478
133633 96.663 -77.949 47.455
133634 96.678 -77.966 47.432
133635 96.692 -77.984 47.408
133636 96.712 -78.000 47.383
133637 96.739 -78.015 47.358
133638 96.766 -78.029 47.333
133639 96.800 -78.045 47.308
133640 96.853 -78.064 47.283
133641 96.861 -78.076 47.256
133642 96.859 -78.087 47.229
133643 96.857 -78.099 47.202
133644 96.858 -78.112 47.176
133645 96.864 -78.129 47.151
133646 96.880 -78.141 47.127
133647 96.904 -78.149 47.102
133648 96.924 -78.156 47.078
133649 96.939 -78.163 47.054
133650 96.954 -78.169 47.031
133651 96.972 -78.178 47.009
133652 96.990 -78.188 46.990
133653 97.007 -78.197 46.972
133654 97.024 -78.210 46.955
133655 97.040 -78.222 46.938
133656 97.056 -78.234 46.923
133657 97.073 -78.247 46.908
133658 97.090 -78.259 46.893
133659 97.106 -78.272 46.879
133660 97.165 -78.279 46.861
133661 97.206 -78.284 46.843
133662 97.247 -78.289 46.825
133663 97.288 -78.294 46.807
133664 97.329 -78.299 46.788
133665 97.370 -78.304 46.771
133666 97.410 -78.307 46.756
133667 97.449 -78.308 46.740
133668 97.488 -78.309 46.723
133669 97.527 -78.310 46.706
133670 97.566 -78.311 46.691
133671 97.605 -78.313 46.675
133672 97.644 -78.313 46.658
133673 97.681 -78.314 46.640
133674 97.710 -78.313 46.623
133675 97.740 -78.311 46.604
133676 97.760 -78.326 46.584
133677 97.771 -78.329 46.565
133678 97.780 -78.329 46.546
133679 97.789 -78.329 46.528
133680 97.813 -78.325 46.509
133681 97.844 -78.317 46.491
133682 97.874 -78.309 46.474
133683 97.904 -78.301 46.457
133684 97.935 -78.294 46.441
133685 97.968 -78.286 46.424
133686 97.996 -78.280 46.407
133687 98.008 -78.285 46.389
133688 98.018 -78.295 46.371
133689 98.032 -78.304 46.353
133690 98.059 -78.307 46.334
133691 98.086 -78.298 46.316
133692 98.114 -78.288 46.298
133693 98.137 -78.282 46.280
133694 98.133 -78.293 46.261
133695 98.143 -78.306 46.242
133696 98.162 -78.319 46.223
133697 98.191 -78.312 46.205
133698 98.221 -78.300 46.187
133699 98.252 -78.288 46.168
133700 98.275 -78.318 46.148
133701 98.294 -78.370 46.129
133702 98.303 -78.368 46.109
133703 98.311 -78.358 46.090
133704 98.319 -78.348 46.070
133705 98.327 -78.338 46.050
133706 98.336 -78.329 46.030
133707 98.344 -78.319 46.008
133708 98.352 -78.309 45.987
133709 98.361 -78.306 45.966
133710 98.353 -78.338 45.946
133711 98.336 -78.388 45.925
133712 98.332 -78.435 45.904
133713 98.327 -78.482 45.882
133714 98.338 -78.493 45.861
133715 98.361 -78.481 45.840
133716 98.414 -78.460 45.819
133717 98.466 -78.438 45.800
133718 98.491 -78.423 45.778
133719 98.499 -78.429 45.755
133720 98.521 -78.441 45.732
133721 98.559 -78.444 45.707
133722 98.595 -78.429 45.684
133723 98.631 -78.407 45.661
133724 98.667 -78.384 45.640
133725 98.702 -78.360 45.618
133726 98.725 -78.331 45.597
133727 98.701 -78.344 45.578
133728 98.671 -78.357 45.560
133729 98.681 -78.353 45.542
133730 98.734 -78.329 45.524
133731 98.786 -78.307 45.506
133732 98.805 -78.319 45.488
133733 98.824 -78.332 45.471
133734 98.843 -78.345 45.455
133735 98.865 -78.354 45.437
133736 98.904 -78.338 45.420
133737 98.941 -78.322 45.405
133738 98.962 -78.351 45.390
133739 98.976 -78.401 45.374
133740 98.989 -78.449 45.360
133741 99.011 -78.478 45.346
133742 99.033 -78.504 45.331
133743 99.114 -78.481 45.317
133744 99.166 -78.463 45.304
133745 99.214 -78.448 45.292
133746 99.244 -78.442 45.280
133747 99.212 -78.469 45.267
133748 99.180 -78.495 45.255
133749 99.200 -78.500 45.243
133750 99.219 -78.505 45.229
133751 99.238 -78.510 45.215
133752 99.255 -78.517 45.202
133753 99.258 -78.537 45.189
133754 99.261 -78.556 45.176
133755 99.264 -78.576 45.163
133756 99.267 -78.595 45.151
133757 99.270 -78.615 45.138
133758 99.273 -78.634 45.125
133759 99.286 -78.645 45.113
133760 99.323 -78.648 45.101
133761 99.385 -78.640 45.086
133762 99.449 -78.632 45.072
133763 99.522 -78.615 45.059
133764 99.581 -78.653 45.044
133765 99.630 -78.677 45.031
133766 99.663 -78.718 45.019
133767 99.752 -78.786 45.006
133768 99.783 -78.815 44.995
133769 99.841 -78.862 44.985
133770 99.898 -78.849 44.976
133771 99.890 -78.865 44.963
133772 99.880 -78.881 44.951
133773 99.871 -78.896 44.939
133774 99.868 -78.906 44.926
133775 99.873 -78.949 44.913
133776 99.893 -78.952 44.902
133777 99.913 -78.955 44.891
133778 99.933 -78.958 44.880
133779 99.953 -78.960 44.870
133780 99.973 -78.963 44.860
133781 99.983 -78.952 44.851
133782 99.989 -78.933 44.842
133783 99.994 -78.915 44.834
133784 100.012 -78.915 44.826
133785 100.037 -78.927 44.818
133786 100.063 -78.940 44.812
133787 100.088 -78.952 44.805
133788 100.114 -78.964 44.799
133789 100.140 -78.976 44.793
133790 100.168 -78.993 44.786
133791 100.198 -79.011 44.780
133792 100.227 -79.029 44.773
133793 100.257 -79.047 44.766
133794 100.285 -79.065 44.759
133795 100.313 -79.070 44.751
133796 100.346 -79.078 44.743
133797 100.385 -79.091 44.734
133798 100.424 -79.108 44.725
133799 100.462 -79.125 44.716
133800 100.494 -79.144 44.706
133801 100.527 -79.163 44.695
133802 100.559 -79.182 44.683
133803 100.592 -79.202 44.671
133804 100.624 -79.221 44.660
134073 93.127 -78.180 54.092
134074 93.146 -78.168 54.085
134075 93.169 -78.156 54.077
134076 93.193 -78.144 54.070
134077 93.217 -78.132 54.062
134078 93.240 -78.120 54.053
134079 93.264 -78.108 54.044
134080 93.288 -78.097 54.034
134081 93.311 -78.084 54.024
134082 93.335 -78.073 54.015
134083 93.359 -78.061 54.005
134084 93.382 -78.049 53.996
134085 93.401 -78.040 53.983
134086 93.411 -78.038 53.970
134087 93.421 -78.035 53.958
134088 93.424 -78.028 53.945
134089 93.434 -78.012 53.933
134090 93.444 -77.996 53.920
134091 93.455 -77.981 53.908
134092 93.465 -77.965 53.892
134093 93.475 -77.950 53.876
134094 93.490 -77.935 53.861
134095 93.501 -77.927 53.846
134096 93.513 -77.918 53.830
134097 93.524 -77.910 53.815
134098 93.536 -77.902 53.799
134099 93.548 -77.894 53.781
134100 93.559 -77.886 53.763
134101 93.571 -77.877 53.746
134102 93.582 -77.869 53.729
134103 93.594 -77.861 53.711
134104 93.606 -77.852 53.694
134105 93.617 -77.844 53.676
134106 93.629 -77.836 53.658
134107 93.640 -77.827 53.639
134108 93.655 -77.821 53.619
134109 93.670 -77.814 53.601
134110 93.683 -77.809 53.582
134111 93.696 -77.805 53.562
134112 93.709 -77.801 53.543
134113 93.721 -77.797 53.523
134121 93.825 -77.762 53.367
134122 93.837 -77.758 53.349
134123 93.850 -77.754 53.332
134124 93.863 -77.749 53.313
134125 93.876 -77.746 53.296
134126 93.889 -77.741 53.278
134127 93.899 -77.736 53.260
134128 93.910 -77.731 53.244
134129 93.921 -77.727 53.227
134130 93.931 -77.721 53.210
134131 93.942 -77.717 53.194
134132 93.953 -77.712 53.180
134133 93.965 -77.708 53.167
134134 93.979 -77.703 53.154
134135 93.986 -77.704 53.139
134136 93.991 -77.705 53.122
134137 93.996 -77.706 53.103
134138 94.000 -77.708 53.085
134139 94.005 -77.709 53.067
134140 94.006 -77.706 53.049
134141 94.006 -77.703 53.031
134142 94.006 -77.700 53.011
134143 94.006 -77.697 52.991
134144 94.011 -77.692 52.970
134145 94.017 -77.686 52.947
134146 94.023 -77.681 52.926
134147 94.031 -77.679 52.905
134148 94.045 -77.683 52.885
134149 94.060 -77.688 52.864
134150 94.076 -77.680 52.843
134151 94.094 -77.673 52.821
134152 94.105 -77.664 52.798
134153 94.108 -77.646 52.772
134154 94.111 -77.629 52.747
134155 94.115 -77.637 52.721
134156 94.118 -77.649 52.699
134157 94.131 -77.650 52.677
134158 94.144 -77.652 52.657
134159 94.158 -77.652 52.637
134160 94.172 -77.653 52.619
134161 94.185 -77.655 52.602
134162 94.199 -77.655 52.586
134163 94.238 -77.627 52.569
134164 94.276 -77.599 52.555
134165 94.316 -77.570 52.543
134166 94.354 -77.542 52.532
134167 94.391 -77.589 52.524
134168 94.382 -77.620 52.518
134169 94.403 -77.602 52.514
134170 94.424 -77.584 52.512
134171 94.445 -77.567 52.509
134172 94.466 -77.548 52.504
134173 94.487 -77.531 52.499
134174 94.508 -77.513 52.493
134175 94.529 -77.495 52.481
134176 94.550 -77.477 52.466
134177 94.557 -77.477 52.449
134178 94.521 -77.518 52.431
134179 94.485 -77.560 52.416
134180 94.483 -77.652 52.403
134181 94.498 -77.772 52.390
134182 94.514 -77.892 52.374
134183 94.535 -77.985 52.356
134184 94.562 -78.044 52.337
134185 94.590 -78.103 52.316
134186 94.610 -78.024 52.296
134187 94.630 -77.918 52.275
134188 94.637 -77.846 52.250
134189 94.639 -77.788 52.220
134190 94.641 -77.730 52.189
134191 94.643 -77.673 52.155
134192 94.645 -77.616 52.117
134193 94.647 -77.558 52.081
134194 94.654 -77.523 52.047
134195 94.669 -77.514 52.013
134196 94.683 -77.505 51.977
134197 94.697 -77.496 51.939
134198 94.710 -77.487 51.898
134199 94.713 -77.486 51.859
134200 94.902 -77.614 51.825
134201 94.938 -77.627 51.794
134202 94.939 -77.583 51.766
134203 94.940 -77.540 51.731
134204 94.941 -77.497 51.694
134205 94.945 -77.470 51.656
134206 94.966 -77.500 51.615
134207 94.979 -77.525 51.573
134208 94.983 -77.536 51.534
134209 95.010 -77.555 51.498
134210 95.036 -77.575 51.458
134211 95.063 -77.594 51.418
134212 95.090 -77.613 51.378
134213 95.117 -77.632 51.337
134214 95.092 -77.614 51.295
134215 95.136 -77.652 51.251
134216 95.130 -77.643 51.206
134217 95.124 -77.635 51.158
134218 95.092 -77.639 51.111
134219 95.107 -77.624 51.064
134220 95.123 -77.624 51.016
134221 95.134 -77.627 50.969
134222 95.146 -77.630 50.924
134223 95.157 -77.628 50.881
134224 95.178 -77.621 50.838
134225 95.199 -77.607 50.797
134226 95.220 -77.592 50.757
134227 95.241 -77.578 50.719
134228 95.263 -77.564 50.682
134229 95.288 -77.553 50.646
134230 95.314 -77.542 50.610
134231 95.340 -77.531 50.573
134232 95.366 -77.520 50.537
134233 95.398 -77.520 50.501
134234 95.430 -77.520 50.465
134235 95.462 -77.520 50.428
134236 95.494 -77.521 50.391
134237 95.513 -77.514 50.354
134238 95.528 -77.504 50.317
134239 95.544 -77.495 50.281
134240 95.559 -77.487 50.244
134241 95.574 -77.478 50.209
134242 95.600 -77.476 50.173
134243 95.653 -77.495 50.136
134244 95.705 -77.514 50.099
134245 95.757 -77.532 50.063
134246 95.810 -77.550 50.027
134247 95.826 -77.566 49.993
134248 95.804 -77.573 49.959
134249 95.775 -77.579 49.925
134250 95.746 -77.585 49.893
134251 95.718 -77.590 49.862
134252 95.689 -77.596 49.831
134253 95.677 -77.611 49.801
134254 95.681 -77.614 49.770
134255 95.684 -77.617 49.740
134256 95.688 -77.620 49.710
134257 95.691 -77.622 49.681
134258 95.695 -77.625 49.652
134259 95.698 -77.627 49.622
134260 95.702 -77.630 49.590
134261 95.705 -77.633 49.559
134262 95.713 -77.634 49.529
134263 95.724 -77.634 49.498
134264 95.734 -77.635 49.466
134265 95.743 -77.636 49.434
134266 95.753 -77.636 49.402
134267 95.763 -77.636 49.369
134268 95.773 -77.636 49.337
134269 95.783 -77.637 49.307
134270 95.793 -77.638 49.276
134271 95.803 -77.638 49.245
134272 95.822 -77.645 49.212
134273 95.844 -77.654 49.179
134274 95.868 -77.663 49.144
134275 95.895 -77.672 49.108
134276 95.921 -77.681 49.071
134277 95.948 -77.691 49.033
134278 95.975 -77.700 48.995
134279 96.002 -77.709 48.958
134280 96.029 -77.719 48.922
134281 96.055 -77.728 48.885
134282 96.080 -77.738 48.849
134283 96.090 -77.748 48.813
134284 96.100 -77.758 48.777
134285 96.109 -77.768 48.741
134286 96.119 -77.777 48.707
134287 96.129 -77.788 48.674
134288 96.139 -77.798 48.640
134289 96.147 -77.814 48.607
134290 96.152 -77.837 48.574
134291 96.157 -77.859 48.542
134292 96.164 -77.880 48.510
134293 96.178 -77.891 48.480
134294 96.192 -77.902 48.450
134295 96.201 -77.908 48.419
134296 96.200 -77.900 48.387
134297 96.199 -77.893 48.357
134298 96.198 -77.885 48.327
134299 96.197 -77.877 48.296
134300 96.196 -77.870 48.269
134301 96.195 -77.862 48.243
134302 96.195 -77.855 48.217
134303 96.194 -77.847 48.192
134304 96.193 -77.840 48.168
134305 96.193 -77.833 48.145
134306 96.208 -77.829 48.123
134307 96.230 -77.825 48.103
134308 96.250 -77.821 48.081
134309 96.271 -77.817 48.059
134310 96.293 -77.813 48.037
134311 96.310 -77.812 48.014
134312 96.323 -77.821 47.990
134313 96.342 -77.833 47.966
134314 96.363 -77.846 47.943
134315 96.384 -77.858 47.920
134316 96.405 -77.871 47.897
134317 96.427 -77.883 47.874
134318 96.448 -77.896 47.849
134319 96.469 -77.908 47.825
134320 96.491 -77.920 47.801
134321 96.521 -77.928 47.777
134322 96.547 -77.933 47.754
134323 96.554 -77.931 47.730
134324 96.562 -77.930 47.705
134325 96.570 -77.928 47.680
134326 96.578 -77.926 47.655
134327 96.585 -77.924 47.631
134328 96.593 -77.923 47.607
134329 96.604 -77.924 47.582
134330 96.615 -77.925 47.558
134331 96.625 -77.927 47.533
134332 96.636 -77.928 47.509
134333 96.647 -77.930 47.485
134334 96.659 -77.939 47.460
134335 96.674 -77.956 47.437
134336 96.688 -77.973 47.413
134337 96.710 -77.989 47.388
134338 96.737 -78.004 47.364
134339 96.764 -78.019 47.339
134340 96.815 -78.037 47.314
134341 96.868 -78.056 47.288
134342 96.867 -78.067 47.262
134343 96.865 -78.078 47.236
134344 96.863 -78.089 47.210
134345 96.861 -78.100 47.185
134346 96.862 -78.114 47.160
134347 96.879 -78.126 47.136
134348 96.903 -78.133 47.112
134349 96.928 -78.141 47.087
134350 96.943 -78.147 47.065
134351 96.960 -78.156 47.042
134352 96.977 -78.166 47.021
134353 96.994 -78.178 47.000
134354 97.010 -78.191 46.982
134355 97.027 -78.203 46.965
134356 97.043 -78.216 46.948
134357 97.060 -78.228 46.933
134358 97.076 -78.241 46.918
134359 97.093 -78.253 46.904
134360 97.109 -78.266 46.891
134361 97.151 -78.275 46.876
134362 97.207 -78.265 46.860
134363 97.246 -78.266 46.842
134364 97.285 -78.267 46.824
134365 97.324 -78.268 46.806
134366 97.363 -78.269 46.791
134367 97.402 -78.270 46.776
134368 97.441 -78.272 46.760
134369 97.480 -78.272 46.743
134370 97.519 -78.274 46.727
134371 97.558 -78.275 46.714
134372 97.596 -78.276 46.698
134373 97.626 -78.275 46.682
134374 97.656 -78.273 46.665
134375 97.685 -78.272 46.648
134376 97.715 -78.270 46.629
134377 97.739 -78.279 46.611
134378 97.758 -78.294 46.592
134379 97.773 -78.302 46.573
134380 97.782 -78.303 46.554
134381 97.806 -78.297 46.535
134382 97.836 -78.290 46.516
134383 97.867 -78.282 46.498
134384 97.905 -78.275 46.480
134385 97.949 -78.266 46.464
134386 97.967 -78.262 46.447
134387 97.984 -78.258 46.429
134388 97.998 -78.260 46.411
134389 98.008 -78.269 46.392
134390 98.024 -78.278 46.374
134391 98.051 -78.282 46.355
134392 98.077 -78.288 46.336
134393 98.104 -78.279 46.318
134394 98.132 -78.269 46.298
134395 98.131 -78.279 46.279
134396 98.146 -78.291 46.260
134397 98.165 -78.305 46.242
134398 98.193 -78.300 46.224
134399 98.224 -78.288 46.205
134400 98.251 -78.295 46.187
134401 98.270 -78.347 46.166
134402 98.287 -78.391 46.146
134403 98.295 -78.381 46.126
134404 98.303 -78.371 46.107
134405 98.311 -78.361 46.087
134406 98.319 -78.351 46.066
134407 98.327 -78.341 46.046
134408 98.336 -78.332 46.025
134409 98.344 -78.322 46.003
134410 98.351 -78.310 45.982
134411 98.394 -78.304 45.961
134412 98.384 -78.354 45.940
134413 98.374 -78.404 45.918
134414 98.365 -78.454 45.897
134415 98.376 -78.465 45.876
134416 98.397 -78.453 45.855
134417 98.418 -78.442 45.834
134418 98.439 -78.429 45.813
134419 98.492 -78.408 45.791
134420 98.507 -78.406 45.768
134421 98.539 -78.412 45.745
134422 98.577 -78.415 45.721
134423 98.614 -78.401 45.696
134424 98.650 -78.379 45.673
134425 98.680 -78.353 45.652
134426 98.704 -78.322 45.630
134427 98.727 -78.293 45.609
134428 98.730 -78.282 45.590
134429 98.714 -78.298 45.572
134430 98.716 -78.298 45.554
134431 98.769 -78.274 45.537
134432 98.790 -78.285 45.519
134433 98.809 -78.297 45.501
134434 98.828 -78.311 45.484
134435 98.847 -78.324 45.468
134436 98.866 -78.337 45.451
134437 98.896 -78.333 45.434
134438 98.931 -78.326 45.419
134439 98.944 -78.376 45.405
134440 98.962 -78.413 45.390
134441 98.985 -78.438 45.375
134442 99.007 -78.464 45.361
134443 99.030 -78.490 45.346
134444 99.087 -78.487 45.332
134445 99.182 -78.451 45.319
134446 99.242 -78.431 45.307
134447 99.278 -78.422 45.294
134448 99.246 -78.448 45.283
134449 99.214 -78.475 45.272
134450 99.210 -78.489 45.261
134451 99.229 -78.495 45.247
134452 99.249 -78.499 45.233
134453 99.255 -78.517 45.220
134454 99.258 -78.535 45.207
134455 99.261 -78.555 45.194
134456 99.264 -78.575 45.180
134457 99.271 -78.592 45.166
134458 99.287 -78.600 45.152
134459 99.303 -78.609 45.139
134460 99.320 -78.617 45.126
134461 99.365 -78.617 45.113
134462 99.428 -78.609 45.099
134463 99.491 -78.601 45.085
134464 99.540 -78.597 45.072
134465 99.570 -78.596 45.057
134466 99.599 -78.633 45.044
134467 99.621 -78.688 45.032
134468 99.745 -78.757 45.020
134469 99.769 -78.781 45.007
134470 99.826 -78.828 44.997
134471 99.884 -78.874 44.987
134472 99.901 -78.886 44.975
134473 99.899 -78.893 44.964
134474 99.901 -78.896 44.952
134475 99.903 -78.899 44.938
134476 99.905 -78.933 44.926
134477 99.914 -78.959 44.914
134478 99.934 -78.962 44.903
134479 99.955 -78.965 44.891
134480 99.974 -78.968 44.881
134481 99.986 -78.957 44.871
134482 99.991 -78.938 44.861
134483 99.996 -78.920 44.853
134484 100.001 -78.901 44.845
134485 100.020 -78.904 44.837
134486 100.046 -78.916 44.828
134487 100.072 -78.928 44.822
134488 100.098 -78.940 44.815
134489 100.123 -78.952 44.808
134490 100.148 -78.965 44.801
134491 100.174 -78.977 44.794
134492 100.200 -78.989 44.787
134493 100.229 -79.007 44.780
134494 100.259 -79.025 44.772
134495 100.288 -79.044 44.764
134496 100.316 -79.050 44.757
134497 100.353 -79.061 44.748
134498 100.392 -79.078 44.739
134499 100.427 -79.096 44.730
134500 100.459 -79.115 44.720
134501 100.492 -79.134 44.709
134502 100.524 -79.153 44.698
134503 100.557 -79.172 44.686
134504 100.589 -79.191 44.673
134505 100.622 -79.211 44.661
134772 93.097 -78.194 54.118
134773 93.112 -78.181 54.110
134774 93.126 -78.169 54.102
134775 93.145 -78.157 54.094
134776 93.169 -78.146 54.087
134777 93.193 -78.134 54.079
134778 93.216 -78.122 54.072
134779 93.240 -78.110 54.063
134780 93.264 -78.098 54.054
134781 93.287 -78.086 54.044
134782 93.310 -78.075 54.035
134783 93.334 -78.062 54.025
134784 93.358 -78.050 54.016
134785 93.381 -78.039 54.006
134786 93.402 -78.029 53.994
134787 93.411 -78.027 53.982
134788 93.421 -78.025 53.969
134789 93.425 -78.012 53.957
134790 93.436 -77.996 53.945
134791 93.446 -77.981 53.932
134792 93.456 -77.965 53.918
134793 93.467 -77.950 53.903
134794 93.477 -77.934 53.887
134795 93.491 -77.922 53.871
134796 93.502 -77.914 53.856
134797 93.514 -77.906 53.840
134798 93.525 -77.897 53.825
134799 93.537 -77.889 53.809
134800 93.549 -77.881 53.791
134801 93.561 -77.873 53.774
134802 93.572 -77.865 53.756
134803 93.584 -77.856 53.739
134804 93.595 -77.848 53.721
134805 93.607 -77.840 53.703
134806 93.619 -77.831 53.686
134807 93.630 -77.823 53.668
134808 93.642 -77.815 53.649
134809 93.657 -77.808 53.629
134810 93.671 -77.802 53.611
134811 93.684 -77.799 53.591
134812 93.697 -77.794 53.572
134813 93.709 -77.790 53.554
134814 93.722 -77.786 53.534
134815 93.735 -77.781 53.514
134822 93.825 -77.752 53.379
134823 93.838 -77.747 53.360
134824 93.848 -77.742 53.343
134825 93.859 -77.737 53.325
134826 93.870 -77.733 53.308
134827 93.880 -77.727 53.291
134828 93.892 -77.723 53.273
134829 93.902 -77.718 53.256
134830 93.913 -77.713 53.239
134831 93.924 -77.708 53.222
134832 93.934 -77.703 53.207
134833 93.946 -77.699 53.192
134834 93.960 -77.695 53.177
134835 93.974 -77.690 53.163
134836 93.982 -77.690 53.149
134837 93.986 -77.691 53.133
134838 93.991 -77.692 53.115
134839 93.996 -77.694 53.096
134840 94.001 -77.696 53.078
134841 94.006 -77.696 53.059
134842 94.011 -77.698 53.041
134843 94.014 -77.698 53.022
134844 94.015 -77.694 53.003
134845 94.022 -77.689 52.983
134846 94.028 -77.683 52.962
134847 94.033 -77.678 52.942
134848 94.039 -77.673 52.922
134849 94.045 -77.667 52.901
134850 94.056 -77.670 52.881
134851 94.070 -77.663 52.860
134852 94.072 -77.661 52.838
134853 94.075 -77.658 52.815
134854 94.078 -77.645 52.790
134855 94.088 -77.626 52.765
134856 94.119 -77.614 52.741
134857 94.114 -77.632 52.718
134858 94.121 -77.639 52.697
134859 94.135 -77.639 52.677
134860 94.148 -77.641 52.658
134861 94.177 -77.624 52.640
134862 94.216 -77.595 52.623
134863 94.255 -77.567 52.607
134864 94.289 -77.545 52.592
134865 94.321 -77.526 52.579
134866 94.345 -77.517 52.568
134867 94.346 -77.527 52.559
134868 94.328 -77.579 52.553
134869 94.347 -77.606 52.549
134870 94.368 -77.588 52.546
134871 94.389 -77.570 52.541
134872 94.410 -77.552 52.536
134873 94.431 -77.534 52.532
134874 94.452 -77.517 52.527
134875 94.473 -77.498 52.521
134876 94.497 -77.483 52.511
134877 94.540 -77.469 52.499
134878 94.546 -77.458 52.483
134879 94.539 -77.466 52.469
134880 94.516 -77.504 52.456
134881 94.480 -77.545 52.444
134882 94.478 -77.637 52.431
134883 94.493 -77.758 52.415
134884 94.509 -77.877 52.398
134885 94.525 -77.998 52.379
134886 94.541 -78.118 52.357
134887 94.555 -78.081 52.336
134888 94.574 -77.975 52.316
134889 94.594 -77.869 52.291
134890 94.613 -77.764 52.259
134891 94.623 -77.685 52.225
134892 94.625 -77.627 52.188
134893 94.626 -77.570 52.149
134894 94.628 -77.513 52.112
134895 94.642 -77.500 52.076
134896 94.655 -77.491 52.041
134897 94.676 -77.487 52.005
134898 94.716 -77.500 51.968
134899 94.757 -77.514 51.929
134900 94.801 -77.530 51.892
134901 94.904 -77.600 51.858
134902 94.965 -77.635 51.828
134903 94.951 -77.611 51.799
134904 94.937 -77.586 51.765
134905 94.938 -77.542 51.728
134906 94.938 -77.499 51.691
134907 94.948 -77.487 51.651
134908 94.965 -77.515 51.609
134909 94.974 -77.529 51.570
134910 95.001 -77.548 51.535
134911 95.028 -77.567 51.495
134912 95.054 -77.586 51.453
134913 95.081 -77.605 51.411
134914 95.105 -77.623 51.370
134915 95.140 -77.616 51.328
134916 95.142 -77.634 51.285
134917 95.139 -77.628 51.240
134918 95.133 -77.619 51.195
134919 95.127 -77.611 51.146
134920 95.120 -77.605 51.096
134921 95.132 -77.608 51.047
134922 95.144 -77.611 50.999
134923 95.155 -77.614 50.953
134924 95.179 -77.617 50.909
134925 95.188 -77.602 50.867
134926 95.201 -77.585 50.824
134927 95.222 -77.570 50.783
134928 95.243 -77.556 50.745
134929 95.264 -77.542 50.709
134930 95.289 -77.530 50.675
134931 95.315 -77.519 50.640
134932 95.341 -77.509 50.603
134933 95.374 -77.509 50.567
134934 95.406 -77.509 50.531
134935 95.438 -77.510 50.494
134936 95.470 -77.510 50.456
134937 95.502 -77.510 50.419
134938 95.521 -77.503 50.381
134939 95.537 -77.495 50.343
134940 95.552 -77.486 50.307
134941 95.568 -77.476 50.270
134942 95.584 -77.467 50.234
134943 95.624 -77.477 50.197
134944 95.677 -77.496 50.159
134945 95.729 -77.514 50.122
134946 95.781 -77.533 50.085
134947 95.808 -77.549 50.047
134948 95.807 -77.564 50.011
134949 95.802 -77.577 49.976
134950 95.773 -77.582 49.940
134951 95.745 -77.588 49.907
134952 95.716 -77.593 49.874
134953 95.687 -77.599 49.842
134954 95.678 -77.613 49.811
134955 95.681 -77.615 49.779
134956 95.685 -77.618 49.748
134957 95.688 -77.621 49.718
134958 95.692 -77.624 49.689
134959 95.695 -77.627 49.660
134960 95.698 -77.629 49.631
134961 95.702 -77.632 49.600
134962 95.712 -77.633 49.569
134963 95.722 -77.633 49.538
134964 95.732 -77.633 49.507
134965 95.742 -77.633 49.475
134966 95.752 -77.634 49.442
134967 95.761 -77.635 49.410
134968 95.771 -77.635 49.377
134969 95.781 -77.636 49.346
134970 95.792 -77.636 49.316
134971 95.801 -77.636 49.286
134972 95.811 -77.637 49.255
134973 95.826 -77.639 49.224
134974 95.853 -77.649 49.192
134975 95.880 -77.658 49.158
134976 95.906 -77.667 49.122
134977 95.933 -77.677 49.085
134978 95.960 -77.686 49.047
134979 95.987 -77.696 49.009
134980 96.014 -77.705 48.972
134981 96.041 -77.714 48.935
134982 96.061 -77.724 48.899
134983 96.071 -77.733 48.863
134984 96.081 -77.744 48.827
134985 96.091 -77.754 48.791
134986 96.101 -77.764 48.755
134987 96.111 -77.774 48.721
134988 96.120 -77.783 48.687
134989 96.130 -77.793 48.653
134990 96.138 -77.811 48.620
134991 96.143 -77.834 48.587
134992 96.148 -77.856 48.554
134993 96.161 -77.868 48.521
134994 96.176 -77.881 48.491
134995 96.190 -77.892 48.460
134996 96.203 -77.902 48.429
134997 96.202 -77.894 48.398
134998 96.201 -77.887 48.367
134999 96.200 -77.879 48.337
135000 96.199 -77.871 48.306
135001 96.198 -77.864 48.280
135002 96.197 -77.856 48.253
135003 96.196 -77.849 48.227
135004 96.196 -77.841 48.202
135005 96.195 -77.834 48.177
135006 96.203 -77.831 48.154
135007 96.221 -77.827 48.131
135008 96.242 -77.823 48.111
135009 96.263 -77.819 48.089
135010 96.285 -77.815 48.067
135011 96.305 -77.811 48.045
135012 96.314 -77.815 48.022
135013 96.326 -77.823 47.999
135014 96.344 -77.835 47.975
135015 96.362 -77.847 47.952
135016 96.382 -77.860 47.929
135017 96.403 -77.872 47.906
135018 96.425 -77.885 47.883
135019 96.446 -77.897 47.859
135020 96.467 -77.910 47.835
135021 96.493 -77.919 47.810
135022 96.523 -77.927 47.786
135023 96.548 -77.932 47.763
135024 96.556 -77.931 47.738
135025 96.564 -77.929 47.713
135026 96.571 -77.927 47.688
135027 96.579 -77.925 47.663
135028 96.587 -77.924 47.638
135029 96.594 -77.921 47.614
135030 96.602 -77.920 47.589
135031 96.612 -77.921 47.565
135032 96.622 -77.922 47.540
135033 96.633 -77.924 47.516
135034 96.644 -77.925 47.491
135035 96.655 -77.929 47.467
135036 96.670 -77.946 47.443
135037 96.684 -77.963 47.419
135038 96.708 -77.979 47.394
135039 96.735 -77.993 47.369
135040 96.777 -78.010 47.345
135041 96.829 -78.029 47.320
135042 96.875 -78.047 47.294
135043 96.873 -78.058 47.269
135044 96.871 -78.069 47.243
135045 96.869 -78.080 47.218
135046 96.867 -78.091 47.193
135047 96.865 -78.103 47.169
135048 96.879 -78.112 47.146
135049 96.903 -78.119 47.122
135050 96.927 -78.125 47.098
135051 96.947 -78.134 47.076
135052 96.964 -78.147 47.054
135053 96.980 -78.159 47.031
135054 96.997 -78.171 47.010
135055 97.014 -78.183 46.992
135056 97.031 -78.195 46.975
135057 97.047 -78.207 46.958
135058 97.064 -78.219 46.943
135059 97.081 -78.232 46.929
135060 97.097 -78.244 46.915
135061 97.114 -78.256 46.903
135062 97.134 -78.260 46.889
135063 97.194 -78.233 46.875
135064 97.237 -78.229 46.859
135065 97.277 -78.231 46.841
135066 97.315 -78.232 46.824
135067 97.355 -78.233 46.810
135068 97.393 -78.234 46.796
135069 97.433 -78.235 46.779
135070 97.472 -78.237 46.762
135071 97.511 -78.238 46.747
135072 97.542 -78.237 46.735
135073 97.572 -78.235 46.720
135074 97.601 -78.234 46.705
135075 97.631 -78.232 46.689
135076 97.662 -78.234 46.672
135077 97.695 -78.240 46.654
135078 97.732 -78.244 46.636
135079 97.750 -78.252 46.619
135080 97.758 -78.263 46.599
135081 97.775 -78.275 46.579
135082 97.799 -78.271 46.559
135083 97.842 -78.263 46.539
135084 97.885 -78.255 46.521
135085 97.923 -78.248 46.503
135086 97.939 -78.244 46.486
135087 97.956 -78.239 46.469
135088 97.973 -78.235 46.450
135089 97.988 -78.235 46.432
135090 97.998 -78.244 46.413
135091 98.017 -78.251 46.394
135092 98.043 -78.257 46.375
135093 98.069 -78.262 46.356
135094 98.095 -78.267 46.337
135095 98.123 -78.260 46.317
135096 98.131 -78.264 46.296
135097 98.149 -78.278 46.276
135098 98.168 -78.291 46.258
135099 98.196 -78.287 46.241
135100 98.226 -78.275 46.222
135101 98.245 -78.324 46.203
135102 98.265 -78.376 46.183
135103 98.279 -78.404 46.163
135104 98.287 -78.394 46.144
135105 98.295 -78.384 46.123
135106 98.303 -78.374 46.104
135107 98.311 -78.364 46.084
135108 98.319 -78.355 46.063
135109 98.321 -78.342 46.041
135110 98.326 -78.330 46.019
135111 98.350 -78.321 45.998
135112 98.413 -78.298 45.976
135113 98.456 -78.306 45.955
135114 98.439 -78.357 45.934
135115 98.388 -78.416 45.913
135116 98.390 -78.427 45.892
135117 98.415 -78.417 45.870
135118 98.441 -78.407 45.848
135119 98.466 -78.397 45.827
135120 98.492 -78.387 45.804
135121 98.516 -78.380 45.780
135122 98.558 -78.383 45.756
135123 98.596 -78.386 45.733
135124 98.632 -78.373 45.709
135125 98.659 -78.345 45.686
135126 98.683 -78.316 45.664
135127 98.710 -78.294 45.642
135128 98.737 -78.272 45.622
135129 98.730 -78.259 45.603
135130 98.693 -78.253 45.584
135131 98.751 -78.243 45.567
135132 98.776 -78.250 45.550
135133 98.794 -78.263 45.533
135134 98.813 -78.276 45.515
135135 98.832 -78.289 45.498
135136 98.851 -78.302 45.482
135137 98.870 -78.315 45.466
135138 98.889 -78.328 45.449
135139 98.914 -78.347 45.434
135140 98.936 -78.373 45.420
135141 98.959 -78.399 45.405
135142 98.982 -78.425 45.390
135143 99.004 -78.451 45.375
135144 99.026 -78.476 45.361
135145 99.061 -78.492 45.348
135146 99.156 -78.457 45.335
135147 99.251 -78.422 45.322
135148 99.312 -78.401 45.309
135149 99.280 -78.428 45.297
135150 99.248 -78.454 45.286
135151 99.221 -78.479 45.276
135152 99.240 -78.484 45.264
135153 99.252 -78.496 45.251
135154 99.255 -78.515 45.238
135155 99.258 -78.535 45.224
135156 99.271 -78.546 45.211
135157 99.287 -78.554 45.197
135158 99.303 -78.564 45.182
135159 99.320 -78.572 45.168
135160 99.336 -78.580 45.154
135161 99.352 -78.589 45.140
135162 99.407 -78.586 45.126
135163 99.471 -78.579 45.112
135164 99.523 -78.573 45.098
135165 99.553 -78.573 45.085
135166 99.559 -78.592 45.071
135167 99.562 -78.614 45.058
135168 99.646 -78.677 45.045
135169 99.738 -78.728 45.033
135170 99.755 -78.747 45.020
135171 99.812 -78.794 45.008
135172 99.869 -78.839 44.998
135173 99.916 -78.873 44.987
135174 99.933 -78.886 44.976
135175 99.935 -78.888 44.964
135176 99.937 -78.890 44.951
135177 99.939 -78.912 44.940
135178 99.945 -78.945 44.928
135179 99.956 -78.968 44.915
135180 99.976 -78.971 44.903
135181 99.988 -78.962 44.893
135182 99.993 -78.943 44.883
135183 99.998 -78.924 44.873
135184 100.003 -78.905 44.864
135185 100.008 -78.887 44.855
135186 100.029 -78.892 44.847
135187 100.055 -78.904 44.839
135188 100.080 -78.916 44.832
135189 100.106 -78.928 44.825
135190 100.132 -78.940 44.817
135191 100.157 -78.952 44.810
135192 100.183 -78.965 44.802
135193 100.208 -78.977 44.795
135194 100.234 -78.989 44.788
135195 100.261 -79.004 44.779
135196 100.290 -79.022 44.771
135197 100.321 -79.031 44.762
135198 100.359 -79.048 44.753
135199 100.392 -79.067 44.743
135200 100.424 -79.086 44.733
135201 100.457 -79.105 44.723
135202 100.489 -79.124 44.712
135203 100.522 -79.143 44.700
135204 100.554 -79.162 44.687
135205 100.586 -79.181 44.675
135206 100.619 -79.200 44.663
135207 100.651 -79.219 44.649
135472 93.082 -78.196 54.134
135473 93.096 -78.184 54.127
135474 93.111 -78.172 54.120
135475 93.125 -78.159 54.111
135476 93.145 -78.147 54.104
135477 93.168 -78.135 54.097
135478 93.192 -78.123 54.089
135479 93.215 -78.112 54.081
135480 93.239 -78.100 54.072
135481 93.263 -78.087 54.063
135482 93.286 -78.076 54.055
135483 93.310 -78.064 54.045
135484 93.334 -78.052 54.036
135485 93.357 -78.040 54.027
135486 93.381 -78.028 54.017
135487 93.402 -78.018 54.006
135488 93.411 -78.016 53.994
135489 93.417 -78.012 53.981
135490 93.427 -77.996 53.969
135491 93.437 -77.981 53.957
135492 93.447 -77.965 53.944
135493 93.458 -77.949 53.930
135494 93.468 -77.934 53.914
135495 93.479 -77.918 53.898
135496 93.492 -77.909 53.882
135497 93.504 -77.901 53.866
135498 93.515 -77.893 53.851
135499 93.527 -77.884 53.835
135500 93.539 -77.876 53.820
135501 93.550 -77.868 53.802
135502 93.562 -77.860 53.784
135503 93.573 -77.852 53.766
135504 93.585 -77.843 53.748
135505 93.596 -77.835 53.731
135506 93.608 -77.827 53.714
135507 93.620 -77.818 53.696
135508 93.631 -77.810 53.678
135509 93.644 -77.802 53.659
135510 93.658 -77.796 53.640
135511 93.672 -77.791 53.620
135512 93.684 -77.787 53.601
135513 93.697 -77.783 53.582
135514 93.710 -77.779 53.564
135515 93.723 -77.774 53.545
135516 93.735 -77.770 53.525
135523 93.820 -77.738 53.390
135524 93.830 -77.733 53.372
135525 93.841 -77.729 53.355
135526 93.852 -77.724 53.337
135527 93.862 -77.719 53.321
135528 93.873 -77.714 53.304
135529 93.884 -77.709 53.285
135530 93.895 -77.705 53.268
135531 93.905 -77.699 53.252
135532 93.916 -77.695 53.235
135533 93.927 -77.689 53.219
135534 93.940 -77.686 53.204
135535 93.954 -77.682 53.189
135536 93.968 -77.677 53.174
135537 93.977 -77.676 53.159
135538 93.982 -77.677 53.143
135539 93.987 -77.679 53.126
135540 93.992 -77.680 53.106
135541 93.997 -77.682 53.087
135542 94.002 -77.683 53.068
135543 94.007 -77.684 53.049
135544 94.011 -77.686 53.030
135545 94.027 -77.676 53.012
135546 94.037 -77.669 52.994
135547 94.043 -77.664 52.975
135548 94.049 -77.659 52.957
135549 94.055 -77.654 52.937
135550 94.060 -77.649 52.917
135551 94.062 -77.644 52.897
135552 94.052 -77.646 52.876
135553 94.043 -77.651 52.855
135554 94.045 -77.648 52.832
135555 94.048 -77.643 52.810
135556 94.068 -77.622 52.787
135557 94.119 -77.597 52.763
135558 94.126 -77.604 52.740
135559 94.117 -77.623 52.719
135560 94.144 -77.608 52.699
135561 94.176 -77.589 52.680
135562 94.208 -77.570 52.663
135563 94.240 -77.551 52.646
135564 94.270 -77.534 52.630
135565 94.284 -77.538 52.617
135566 94.297 -77.542 52.604
135567 94.302 -77.550 52.594
135568 94.292 -77.555 52.586
135569 94.293 -77.551 52.583
135570 94.312 -77.592 52.581
135571 94.333 -77.573 52.578
135572 94.354 -77.556 52.571
135573 94.375 -77.538 52.565
135574 94.396 -77.520 52.560
135575 94.417 -77.502 52.556
135576 94.451 -77.490 52.550
135577 94.494 -77.476 52.542
135578 94.537 -77.463 52.532
135579 94.530 -77.453 52.519
135580 94.523 -77.443 52.508
135581 94.515 -77.455 52.497
135582 94.507 -77.490 52.486
135583 94.475 -77.530 52.473
135584 94.473 -77.623 52.457
135585 94.489 -77.743 52.440
135586 94.504 -77.863 52.421
135587 94.518 -77.890 52.398
135588 94.531 -77.848 52.376
135589 94.543 -77.806 52.353
135590 94.555 -77.765 52.326
135591 94.568 -77.723 52.293
135592 94.580 -77.680 52.256
135593 94.596 -77.609 52.219
135594 94.608 -77.524 52.181
135595 94.615 -77.486 52.145
135596 94.645 -77.491 52.110
135597 94.686 -77.504 52.076
135598 94.727 -77.517 52.039
135599 94.768 -77.531 52.001
135600 94.809 -77.545 51.961
135601 94.849 -77.558 51.922
135602 94.907 -77.586 51.887
135603 94.992 -77.642 51.855
135604 94.978 -77.617 51.826
135605 94.964 -77.592 51.793
135606 94.950 -77.568 51.757
135607 94.936 -77.544 51.720
135608 94.936 -77.501 51.678
135609 94.950 -77.504 51.638
135610 94.965 -77.520 51.600
135611 94.992 -77.539 51.566
135612 95.019 -77.558 51.527
135613 95.035 -77.573 51.488
135614 95.049 -77.587 51.449
135615 95.090 -77.575 51.409
135616 95.125 -77.561 51.367
135617 95.112 -77.586 51.324
135618 95.127 -77.597 51.282
135619 95.141 -77.604 51.235
135620 95.135 -77.595 51.183
135621 95.130 -77.589 51.130
135622 95.142 -77.592 51.078
135623 95.158 -77.595 51.030
135624 95.182 -77.599 50.983
135625 95.206 -77.604 50.938
135626 95.215 -77.589 50.893
135627 95.221 -77.569 50.851
135628 95.227 -77.549 50.810
135629 95.245 -77.533 50.771
135630 95.265 -77.519 50.733
135631 95.290 -77.508 50.699
135632 95.317 -77.498 50.666
135633 95.349 -77.498 50.633
135634 95.381 -77.498 50.597
135635 95.413 -77.498 50.561
135636 95.446 -77.498 50.525
135637 95.477 -77.499 50.488
135638 95.510 -77.499 50.449
135639 95.530 -77.493 50.411
135640 95.545 -77.484 50.372
135641 95.561 -77.475 50.335
135642 95.577 -77.466 50.299
135643 95.596 -77.461 50.262
135644 95.649 -77.479 50.223
135645 95.701 -77.497 50.184
135646 95.753 -77.515 50.145
135647 95.790 -77.533 50.107
135648 95.789 -77.547 50.068
135649 95.788 -77.561 50.031
135650 95.788 -77.576 49.993
135651 95.771 -77.586 49.957
135652 95.743 -77.591 49.921
135653 95.714 -77.597 49.887
135654 95.685 -77.602 49.854
135655 95.678 -77.614 49.822
135656 95.681 -77.617 49.790
135657 95.685 -77.620 49.758
135658 95.688 -77.623 49.728
135659 95.692 -77.625 49.698
135660 95.695 -77.628 49.669
135661 95.700 -77.630 49.640
135662 95.710 -77.630 49.609
135663 95.720 -77.631 49.578
135664 95.730 -77.632 49.548
135665 95.740 -77.632 49.516
135666 95.750 -77.633 49.482
135667 95.760 -77.633 49.450
135668 95.770 -77.633 49.419
135669 95.779 -77.634 49.387
135670 95.790 -77.634 49.357
135671 95.802 -77.633 49.327
135672 95.816 -77.632 49.297
135673 95.830 -77.630 49.266
135674 95.845 -77.633 49.235
135675 95.865 -77.644 49.204
135676 95.891 -77.653 49.171
135677 95.918 -77.662 49.136
135678 95.945 -77.672 49.099
135679 95.972 -77.681 49.061
135680 95.999 -77.691 49.023
135681 96.026 -77.700 48.986
135682 96.043 -77.710 48.949
135683 96.053 -77.720 48.913
135684 96.062 -77.730 48.878
135685 96.073 -77.740 48.841
135686 96.082 -77.749 48.805
135687 96.092 -77.760 48.770
135688 96.102 -77.770 48.735
135689 96.112 -77.780 48.701
135690 96.122 -77.790 48.667
135691 96.129 -77.808 48.633
135692 96.134 -77.830 48.599
135693 96.145 -77.846 48.566
135694 96.159 -77.858 48.532
135695 96.173 -77.869 48.501
135696 96.188 -77.881 48.471
135697 96.202 -77.892 48.440
135698 96.204 -77.888 48.409
135699 96.203 -77.881 48.378
135700 96.202 -77.873 48.347
135701 96.201 -77.865 48.317
135702 96.200 -77.858 48.291
135703 96.199 -77.850 48.264
135704 96.198 -77.843 48.237
135705 96.198 -77.836 48.212
135706 96.206 -77.832 48.187
135707 96.214 -77.829 48.163
135708 96.234 -77.825 48.139
135709 96.255 -77.821 48.117
135710 96.276 -77.817 48.096
135711 96.297 -77.813 48.074
135712 96.312 -77.812 48.052
135713 96.319 -77.818 48.030
135714 96.329 -77.825 48.007
135715 96.347 -77.837 47.984
135716 96.365 -77.849 47.962
135717 96.383 -77.862 47.938
135718 96.401 -77.874 47.915
135719 96.422 -77.887 47.893
135720 96.444 -77.899 47.869
135721 96.465 -77.912 47.845
135722 96.495 -77.919 47.821
135723 96.525 -77.927 47.797
135724 96.549 -77.932 47.772
135725 96.557 -77.930 47.747
135726 96.565 -77.928 47.722
135727 96.573 -77.926 47.696
135728 96.580 -77.924 47.671
135729 96.588 -77.923 47.645
135730 96.595 -77.921 47.621
135731 96.603 -77.919 47.597
135732 96.611 -77.918 47.572
135733 96.620 -77.917 47.547
135734 96.630 -77.918 47.522
135735 96.642 -77.920 47.498
135736 96.652 -77.921 47.475
135737 96.666 -77.936 47.450
135738 96.680 -77.953 47.425
135739 96.707 -77.968 47.401
135740 96.738 -77.984 47.376
135741 96.791 -78.003 47.351
135742 96.844 -78.021 47.327
135743 96.881 -78.037 47.302
135744 96.879 -78.048 47.276
135745 96.877 -78.059 47.251
135746 96.875 -78.071 47.227
135747 96.873 -78.082 47.203
135748 96.874 -78.096 47.178
135749 96.890 -78.106 47.156
135750 96.914 -78.112 47.134
135751 96.938 -78.119 47.111
135752 96.954 -78.119 47.088
135753 96.971 -78.129 47.064
135754 96.988 -78.141 47.042
135755 97.005 -78.153 47.022
135756 97.022 -78.166 47.003
135757 97.038 -78.178 46.986
135758 97.055 -78.190 46.970
135759 97.072 -78.202 46.954
135760 97.088 -78.214 46.940
135761 97.105 -78.226 46.926
135762 97.121 -78.238 46.914
135763 97.144 -78.234 46.901
135764 97.172 -78.215 46.888
135765 97.229 -78.194 46.874
135766 97.268 -78.194 46.859
135767 97.307 -78.196 46.843
135768 97.347 -78.197 46.828
135769 97.385 -78.198 46.813
135770 97.425 -78.199 46.798
135771 97.458 -78.199 46.784
135772 97.488 -78.197 46.770
135773 97.517 -78.196 46.757
135774 97.548 -78.198 46.742
135775 97.581 -78.204 46.727
135776 97.614 -78.210 46.710
135777 97.647 -78.216 46.692
135778 97.679 -78.224 46.674
135779 97.718 -78.227 46.656
135780 97.758 -78.228 46.638
135781 97.783 -78.235 46.619
135782 97.791 -78.245 46.600
135783 97.821 -78.244 46.582
135784 97.865 -78.236 46.564
135785 97.894 -78.230 46.547
135786 97.912 -78.225 46.528
135787 97.928 -78.221 46.510
135788 97.945 -78.217 46.491
135789 97.961 -78.213 46.471
135790 97.978 -78.210 46.452
135791 97.988 -78.219 46.434
135792 98.009 -78.225 46.414
135793 98.035 -78.231 46.395
135794 98.062 -78.236 46.375
135795 98.088 -78.241 46.355
135796 98.114 -78.247 46.335
135797 98.134 -78.250 46.314
135798 98.153 -78.264 46.292
135799 98.171 -78.277 46.274
135800 98.198 -78.274 46.256
135801 98.221 -78.301 46.238
135802 98.240 -78.354 46.219
135803 98.259 -78.406 46.198
135804 98.271 -78.416 46.179
135805 98.279 -78.407 46.159
135806 98.287 -78.397 46.138
135807 98.295 -78.388 46.119
135808 98.297 -78.375 46.100
135809 98.298 -78.362 46.080
135810 98.303 -78.350 46.058
135811 98.326 -78.340 46.036
135812 98.350 -78.331 46.013
135813 98.411 -78.309 45.991
135814 98.484 -78.279 45.970
135815 98.442 -78.322 45.950
135816 98.391 -78.382 45.929
135817 98.399 -78.387 45.908
135818 98.424 -78.377 45.885
135819 98.450 -78.367 45.862
135820 98.475 -78.357 45.840
135821 98.491 -78.354 45.816
135822 98.504 -78.353 45.792
135823 98.530 -78.353 45.768
135824 98.586 -78.357 45.744
135825 98.633 -78.348 45.721
135826 98.671 -78.331 45.698
135827 98.698 -78.309 45.677
135828 98.725 -78.287 45.655
135829 98.752 -78.265 45.635
135830 98.725 -78.257 45.616
135831 98.695 -78.234 45.597
135832 98.757 -78.209 45.580
135833 98.783 -78.235 45.563
135834 98.802 -78.248 45.547
135835 98.820 -78.261 45.529
135836 98.840 -78.274 45.512
135837 98.858 -78.287 45.496
135838 98.877 -78.300 45.481
135839 98.897 -78.313 45.465
135840 98.915 -78.335 45.450
135841 98.933 -78.360 45.435
135842 98.955 -78.385 45.420
135843 98.978 -78.411 45.405
135844 99.001 -78.437 45.391
135845 99.023 -78.463 45.379
135846 99.046 -78.489 45.365
135847 99.129 -78.463 45.351
135848 99.224 -78.428 45.338
135849 99.279 -78.413 45.325
135850 99.283 -78.423 45.313
135851 99.282 -78.435 45.301
135852 99.250 -78.461 45.290
135853 99.249 -78.476 45.279
135854 99.255 -78.492 45.267
135855 99.272 -78.501 45.253
135856 99.288 -78.509 45.240
135857 99.304 -78.518 45.226
135858 99.321 -78.526 45.212
135859 99.337 -78.535 45.197
135860 99.353 -78.544 45.184
135861 99.369 -78.552 45.170
135862 99.385 -78.561 45.156
135863 99.450 -78.555 45.140
135864 99.505 -78.549 45.126
135865 99.528 -78.554 45.112
135866 99.532 -78.576 45.099
135867 99.545 -78.597 45.085
135868 99.582 -78.617 45.071
135869 99.699 -78.678 45.058
135870 99.737 -78.702 45.045
135871 99.741 -78.711 45.032
135872 99.798 -78.755 45.021
135873 99.856 -78.799 45.009
135874 99.911 -78.840 44.998
135875 99.932 -78.855 44.988
135876 99.954 -78.870 44.977
135877 99.972 -78.883 44.966
135878 99.973 -78.890 44.954
135879 99.980 -78.924 44.941
135880 99.986 -78.957 44.929
135881 99.990 -78.967 44.918
135882 99.995 -78.949 44.906
135883 100.000 -78.930 44.896
135884 100.005 -78.911 44.886
135885 100.010 -78.892 44.877
135886 100.015 -78.873 44.868
135887 100.038 -78.880 44.859
135888 100.064 -78.893 44.850
135889 100.089 -78.905 44.843
135890 100.115 -78.917 44.835
135891 100.140 -78.929 44.827
135892 100.166 -78.941 44.819
135893 100.192 -78.953 44.812
135894 100.217 -78.965 44.804
135895 100.243 -78.977 44.795
135896 100.268 -78.990 44.786
135897 100.294 -79.002 44.777
135898 100.324 -79.018 44.768
135899 100.357 -79.037 44.758
135900 100.389 -79.056 44.748
135901 100.422 -79.076 44.738
135902 100.454 -79.095 44.726
135903 100.487 -79.114 44.714
135904 100.519 -79.133 44.702
135905 100.552 -79.152 44.689
135906 100.584 -79.172 44.677
135907 100.617 -79.190 44.664
135908 100.649 -79.209 44.650
136170 93.040 -78.214 54.161
136171 93.054 -78.210 54.155
136172 93.068 -78.197 54.148
136173 93.081 -78.185 54.142
136174 93.095 -78.173 54.135
136175 93.110 -78.161 54.128
136176 93.124 -78.149 54.120
136177 93.144 -78.137 54.113
136178 93.167 -78.125 54.106
136179 93.191 -78.113 54.097
136180 93.215 -78.101 54.091
136181 93.238 -78.089 54.082
136182 93.262 -78.078 54.072
136183 93.286 -78.065 54.064
136184 93.309 -78.053 54.055
136185 93.333 -78.042 54.046
136186 93.357 -78.030 54.037
136187 93.380 -78.018 54.029
136188 93.402 -78.008 54.018
136189 93.412 -78.005 54.006
136190 93.418 -77.996 53.995
136191 93.428 -77.981 53.982
136192 93.438 -77.965 53.969
136193 93.449 -77.949 53.956
136194 93.459 -77.934 53.941
136195 93.469 -77.918 53.925
136196 93.482 -77.905 53.909
136197 93.493 -77.896 53.892
136198 93.505 -77.888 53.877
136199 93.516 -77.880 53.861
136200 93.528 -77.871 53.845
136201 93.540 -77.863 53.830
136202 93.551 -77.855 53.813
136203 93.563 -77.846 53.795
136204 93.574 -77.839 53.777
136205 93.586 -77.830 53.759
136206 93.597 -77.822 53.741
136207 93.609 -77.814 53.724
136208 93.621 -77.805 53.707
136209 93.632 -77.797 53.688
136210 93.646 -77.790 53.669
136211 93.659 -77.784 53.650
136212 93.672 -77.780 53.632
136213 93.685 -77.776 53.613
136214 93.698 -77.771 53.594
136215 93.711 -77.768 53.575
136216 93.723 -77.763 53.556
136217 93.736 -77.758 53.536
136218 93.747 -77.754 53.516
136224 93.812 -77.725 53.404
136225 93.823 -77.720 53.385
136226 93.833 -77.715 53.367
136227 93.844 -77.710 53.350
136228 93.855 -77.705 53.333
136229 93.865 -77.701 53.316
136230 93.876 -77.696 53.298
136231 93.887 -77.691 53.281
136232 93.898 -77.686 53.264
136233 93.908 -77.681 53.247
136234 93.921 -77.677 53.231
136235 93.935 -77.673 53.215
136236 93.948 -77.668 53.200
136237 93.963 -77.664 53.184
136238 93.973 -77.662 53.169
136239 93.978 -77.664 53.152
136240 93.983 -77.665 53.135
136241 93.988 -77.667 53.115
136242 93.993 -77.667 53.095
136243 93.998 -77.669 53.076
136244 94.002 -77.671 53.056
136245 94.012 -77.666 53.038
136246 94.028 -77.654 53.021
136247 94.048 -77.642 53.004
136248 94.062 -77.633 52.987
136249 94.068 -77.628 52.969
136250 94.074 -77.624 52.951
136251 94.076 -77.618 52.933
136252 94.075 -77.613 52.913
136253 94.073 -77.608 52.892
136254 94.060 -77.613 52.872
136255 94.041 -77.623 52.852
136256 94.023 -77.633 52.831
136257 94.049 -77.619 52.810
136258 94.099 -77.594 52.788
136259 94.145 -77.572 52.767
136260 94.163 -77.569 52.745
136261 94.175 -77.564 52.725
136262 94.190 -77.558 52.707
136263 94.209 -77.555 52.690
136264 94.222 -77.559 52.672
136265 94.236 -77.562 52.655
136266 94.250 -77.566 52.642
136267 94.262 -77.569 52.630
136268 94.263 -77.563 52.621
136269 94.264 -77.557 52.614
136270 94.266 -77.550 52.615
136271 94.277 -77.577 52.614
136272 94.298 -77.559 52.612
136273 94.319 -77.542 52.605
136274 94.340 -77.523 52.598
136275 94.369 -77.509 52.592
136276 94.404 -77.498 52.588
136277 94.449 -77.484 52.583
136278 94.495 -77.470 52.576
136279 94.522 -77.457 52.567
136280 94.514 -77.448 52.556
136281 94.507 -77.438 52.547
136282 94.499 -77.428 52.539
136283 94.492 -77.445 52.529
136284 94.484 -77.479 52.516
136285 94.470 -77.515 52.500
136286 94.468 -77.608 52.482
136287 94.483 -77.689 52.463
136288 94.495 -77.646 52.438
136289 94.508 -77.605 52.413
136290 94.520 -77.563 52.388
136291 94.533 -77.521 52.358
136292 94.547 -77.494 52.323
136293 94.564 -77.488 52.286
136294 94.580 -77.483 52.247
136295 94.597 -77.476 52.213
136296 94.635 -77.494 52.179
136297 94.683 -77.513 52.145
136298 94.732 -77.531 52.112
136299 94.778 -77.548 52.074
136300 94.819 -77.562 52.034
136301 94.860 -77.576 51.993
136302 94.901 -77.589 51.953
136303 94.942 -77.602 51.916
136304 95.014 -77.644 51.882
136305 95.005 -77.624 51.852
136306 94.991 -77.599 51.819
136307 94.977 -77.575 51.784
136308 94.963 -77.550 51.747
136309 94.949 -77.526 51.705
136310 94.935 -77.501 51.666
136311 94.949 -77.510 51.629
136312 94.964 -77.523 51.595
136313 94.979 -77.538 51.558
136314 94.993 -77.552 51.521
136315 95.039 -77.535 51.483
136316 95.117 -77.483 51.443
136317 95.109 -77.504 51.402
136318 95.097 -77.530 51.361
136319 95.094 -77.550 51.320
136320 95.112 -77.560 51.273
136321 95.130 -77.570 51.221
136322 95.139 -77.573 51.167
136323 95.162 -77.577 51.113
136324 95.185 -77.581 51.062
136325 95.209 -77.586 51.014
136326 95.233 -77.590 50.967
136327 95.243 -77.575 50.923
136328 95.249 -77.555 50.880
136329 95.255 -77.536 50.838
136330 95.261 -77.516 50.797
136331 95.267 -77.497 50.760
136332 95.293 -77.486 50.724
136333 95.325 -77.487 50.690
136334 95.357 -77.487 50.658
136335 95.389 -77.487 50.626
136336 95.421 -77.488 50.593
136337 95.453 -77.488 50.557
136338 95.485 -77.488 50.521
136339 95.517 -77.488 50.482
136340 95.539 -77.483 50.443
136341 95.554 -77.473 50.404
136342 95.570 -77.464 50.365
136343 95.585 -77.456 50.327
136344 95.620 -77.461 50.289
136345 95.672 -77.479 50.248
136346 95.725 -77.498 50.209
136347 95.771 -77.516 50.170
136348 95.770 -77.530 50.130
136349 95.770 -77.545 50.090
136350 95.769 -77.559 50.051
136351 95.768 -77.573 50.013
136352 95.767 -77.588 49.975
136353 95.741 -77.594 49.938
136354 95.712 -77.600 49.903
136355 95.683 -77.605 49.868
136356 95.678 -77.616 49.836
136357 95.682 -77.619 49.802
136358 95.685 -77.621 49.770
136359 95.689 -77.624 49.738
136360 95.692 -77.627 49.707
136361 95.698 -77.629 49.677
136362 95.708 -77.629 49.648
136363 95.719 -77.630 49.617
136364 95.728 -77.630 49.586
136365 95.738 -77.630 49.555
136366 95.748 -77.631 49.523
136367 95.758 -77.631 49.490
136368 95.768 -77.632 49.459
136369 95.780 -77.631 49.429
136370 95.795 -77.630 49.399
136371 95.809 -77.628 49.370
136372 95.823 -77.627 49.340
136373 95.837 -77.625 49.311
136374 95.851 -77.623 49.279
136375 95.866 -77.627 49.248
136376 95.882 -77.639 49.217
136377 95.903 -77.649 49.183
136378 95.930 -77.658 49.149
136379 95.957 -77.667 49.112
136380 95.984 -77.677 49.074
136381 96.011 -77.686 49.036
136382 96.024 -77.696 48.999
136383 96.034 -77.705 48.963
136384 96.044 -77.716 48.927
136385 96.054 -77.726 48.892
136386 96.064 -77.736 48.856
136387 96.074 -77.746 48.819
136388 96.084 -77.755 48.783
136389 96.094 -77.766 48.748
136390 96.103 -77.776 48.714
136391 96.113 -77.786 48.680
136392 96.120 -77.805 48.646
136393 96.128 -77.824 48.612
136394 96.143 -77.835 48.578
136395 96.157 -77.846 48.544
136396 96.171 -77.858 48.512
136397 96.185 -77.869 48.481
136398 96.200 -77.881 48.451
136399 96.206 -77.882 48.420
136400 96.205 -77.874 48.389
136401 96.204 -77.867 48.358
136402 96.203 -77.859 48.329
136403 96.202 -77.852 48.302
136404 96.201 -77.845 48.275
136405 96.200 -77.837 48.249
136406 96.208 -77.834 48.223
136407 96.216 -77.830 48.197
136408 96.226 -77.827 48.173
136409 96.247 -77.823 48.149
136410 96.268 -77.818 48.127
136411 96.289 -77.815 48.105
136412 96.310 -77.811 48.083
136413 96.317 -77.815 48.060
136414 96.323 -77.821 48.038
136415 96.332 -77.827 48.015
136416 96.350 -77.840 47.992
136417 96.368 -77.852 47.970
136418 96.386 -77.865 47.948
136419 96.403 -77.877 47.926
136420 96.421 -77.889 47.903
136421 96.442 -77.901 47.879
136422 96.467 -77.912 47.856
136423 96.497 -77.919 47.831
136424 96.527 -77.926 47.808
136425 96.550 -77.931 47.782
136426 96.558 -77.929 47.756
136427 96.566 -77.928 47.731
136428 96.574 -77.926 47.706
136429 96.581 -77.924 47.680
136430 96.589 -77.922 47.654
136431 96.597 -77.920 47.630
136432 96.604 -77.918 47.605
136433 96.612 -77.917 47.581
136434 96.619 -77.915 47.556
136435 96.628 -77.914 47.531
136436 96.639 -77.915 47.506
136437 96.650 -77.917 47.482
136438 96.662 -77.926 47.458
136439 96.678 -77.943 47.433
136440 96.705 -77.958 47.409
136441 96.753 -77.976 47.384
136442 96.806 -77.994 47.359
136443 96.859 -78.012 47.335
136444 96.887 -78.028 47.310
136445 96.885 -78.039 47.284
136446 96.883 -78.050 47.260
136447 96.881 -78.062 47.236
136448 96.882 -78.076 47.212
136449 96.883 -78.091 47.188
136450 96.901 -78.100 47.166
136451 96.925 -78.106 47.143
136452 96.945 -78.109 47.120
136453 96.958 -78.107 47.098
136454 96.979 -78.112 47.076
136455 96.996 -78.123 47.054
136456 97.013 -78.135 47.034
136457 97.029 -78.147 47.015
136458 97.046 -78.159 46.998
136459 97.063 -78.172 46.981
136460 97.079 -78.184 46.966
136461 97.096 -78.196 46.951
136462 97.112 -78.208 46.938
136463 97.129 -78.220 46.924
136464 97.155 -78.207 46.911
136465 97.182 -78.188 46.900
136466 97.210 -78.169 46.888
136467 97.260 -78.158 46.875
136468 97.299 -78.159 46.861
136469 97.338 -78.160 46.846
136470 97.373 -78.161 46.832
136471 97.403 -78.159 46.818
136472 97.435 -78.163 46.804
136473 97.467 -78.169 46.791
136474 97.500 -78.175 46.778
136475 97.533 -78.181 46.763
136476 97.566 -78.187 46.747
136477 97.598 -78.193 46.730
136478 97.632 -78.200 46.712
136479 97.664 -78.207 46.694
136480 97.704 -78.210 46.676
136481 97.744 -78.211 46.658
136482 97.785 -78.213 46.641
136483 97.816 -78.217 46.623
136484 97.846 -78.216 46.606
136485 97.867 -78.212 46.589
136486 97.883 -78.207 46.572
136487 97.900 -78.203 46.552
136488 97.917 -78.199 46.532
136489 97.933 -78.194 46.513
136490 97.950 -78.190 46.493
136491 97.967 -78.185 46.473
136492 97.978 -78.194 46.454
136493 98.001 -78.200 46.434
136494 98.027 -78.205 46.414
136495 98.054 -78.210 46.394
136496 98.080 -78.216 46.373
136497 98.097 -78.240 46.353
136498 98.125 -78.246 46.331
136499 98.153 -78.252 46.309
136500 98.174 -78.263 46.290
136501 98.197 -78.279 46.271
136502 98.216 -78.331 46.252
136503 98.235 -78.383 46.233
136504 98.253 -78.435 46.212
136505 98.263 -78.429 46.192
136506 98.271 -78.420 46.172
136507 98.273 -78.407 46.152
136508 98.274 -78.395 46.134
136509 98.275 -78.382 46.116
136510 98.279 -78.369 46.097
136511 98.303 -78.360 46.075
136512 98.326 -78.351 46.053
136513 98.350 -78.341 46.029
136514 98.409 -78.319 46.007
136515 98.456 -78.294 45.986
136516 98.445 -78.288 45.965
136517 98.393 -78.347 45.944
136518 98.408 -78.347 45.922
136519 98.433 -78.337 45.899
136520 98.456 -78.329 45.876
136521 98.468 -78.328 45.853
136522 98.482 -78.327 45.829
136523 98.495 -78.326 45.805
136524 98.508 -78.325 45.781
136525 98.543 -78.326 45.757
136526 98.592 -78.320 45.734
136527 98.631 -78.303 45.711
136528 98.669 -78.285 45.689
136529 98.708 -78.268 45.668
136530 98.744 -78.258 45.648
136531 98.737 -78.209 45.629
136532 98.708 -78.178 45.610
136533 98.753 -78.175 45.593
136534 98.780 -78.202 45.577
136535 98.806 -78.228 45.560
136536 98.830 -78.250 45.543
136537 98.849 -78.263 45.526
136538 98.867 -78.275 45.510
136539 98.886 -78.288 45.494
136540 98.906 -78.301 45.479
136541 98.924 -78.326 45.464
136542 98.941 -78.350 45.449
136543 98.959 -78.374 45.435
136544 98.978 -78.398 45.420
136545 98.997 -78.423 45.407
136546 99.020 -78.449 45.395
136547 99.043 -78.474 45.381
136548 99.103 -78.468 45.367
136549 99.179 -78.443 45.353
136550 99.183 -78.453 45.341
136551 99.186 -78.463 45.328
136552 99.196 -78.470 45.315
136553 99.232 -78.462 45.304
136554 99.264 -78.455 45.293
136555 99.291 -78.449 45.281
136556 99.305 -78.468 45.268
136557 99.321 -78.481 45.254
136558 99.337 -78.490 45.240
136559 99.353 -78.498 45.226
136560 99.369 -78.507 45.212
136561 99.386 -78.516 45.199
136562 99.402 -78.524 45.186
136563 99.426 -78.532 45.172
136564 99.488 -78.526 45.156
136565 99.502 -78.539 45.141
136566 99.524 -78.559 45.127
136567 99.562 -78.579 45.113
136568 99.600 -78.600 45.100
136569 99.637 -78.620 45.086
136570 99.766 -78.688 45.072
136571 99.752 -78.686 45.059
136572 99.759 -78.693 45.046
136573 99.785 -78.715 45.035
136574 99.843 -78.759 45.024
136575 99.901 -78.803 45.012
136576 99.927 -78.822 45.001
136577 99.949 -78.837 44.990
136578 99.971 -78.852 44.978
136579 99.992 -78.867 44.966
136580 100.000 -78.887 44.954
136581 99.999 -78.912 44.943
136582 99.999 -78.938 44.931
136583 100.002 -78.934 44.920
136584 100.007 -78.916 44.910
136585 100.012 -78.897 44.900
136586 100.017 -78.878 44.891
136587 100.023 -78.859 44.881
136588 100.047 -78.869 44.871
136589 100.072 -78.881 44.862
136590 100.098 -78.893 44.854
136591 100.123 -78.905 44.846
136592 100.149 -78.918 44.838
136593 100.174 -78.930 44.829
136594 100.200 -78.942 44.822
136595 100.226 -78.954 44.813
136596 100.251 -78.966 44.803
136597 100.283 -78.979 44.794
136598 100.309 -78.992 44.784
136599 100.322 -79.009 44.773
136600 100.354 -79.028 44.763
136601 100.387 -79.046 44.753
136602 100.419 -79.065 44.742
136603 100.452 -79.085 44.730
136604 100.484 -79.104 44.718
136605 100.516 -79.123 44.705
136606 100.549 -79.142 44.692
136607 100.581 -79.161 44.679
136608 100.614 -79.181 44.665
136609 100.646 -79.200 44.651
136867 92.985 -78.219 54.195
136868 92.998 -78.215 54.189
136869 93.012 -78.210 54.181
136870 93.025 -78.205 54.174
136871 93.039 -78.200 54.168
136872 93.053 -78.195 54.162
136873 93.067 -78.188 54.156
136874 93.081 -78.175 54.150
136875 93.095 -78.163 54.143
136876 93.109 -78.150 54.135
136877 93.123 -78.138 54.128
136878 93.143 -78.126 54.121
136879 93.167 -78.115 54.114
136880 93.191 -78.103 54.105
136881 93.214 -78.091 54.099
136882 93.238 -78.079 54.090
136883 93.262 -78.067 54.081
136884 93.285 -78.056 54.073
136885 93.309 -78.044 54.065
136886 93.332 -78.031 54.057
136887 93.356 -78.020 54.049
136888 93.379 -78.008 54.041
136889 93.402 -77.996 54.031
136890 93.412 -77.994 54.019
136891 93.420 -77.980 54.007
136892 93.430 -77.965 53.995
136893 93.440 -77.949 53.982
136894 93.450 -77.934 53.968
136895 93.461 -77.918 53.953
136896 93.471 -77.902 53.936
136897 93.483 -77.892 53.919
136898 93.495 -77.884 53.902
136899 93.506 -77.875 53.887
136900 93.518 -77.867 53.871
136901 93.530 -77.859 53.856
136902 93.541 -77.850 53.840
136903 93.553 -77.842 53.823
136904 93.564 -77.834 53.805
136905 93.576 -77.825 53.788
136906 93.587 -77.818 53.770
136907 93.599 -77.809 53.752
136908 93.611 -77.801 53.735
136909 93.622 -77.793 53.718
136910 93.634 -77.784 53.699
136911 93.647 -77.777 53.680
136912 93.660 -77.774 53.662
136913 93.673 -77.769 53.643
136914 93.686 -77.765 53.624
136915 93.697 -77.760 53.605
136916 93.707 -77.755 53.587
136917 93.718 -77.750 53.567
136918 93.729 -77.746 53.547
136919 93.740 -77.740 53.528
136926 93.815 -77.706 53.399
136927 93.826 -77.702 53.381
136928 93.836 -77.697 53.363
136929 93.847 -77.692 53.346
136930 93.858 -77.687 53.329
136931 93.868 -77.682 53.311
136932 93.879 -77.677 53.294
136933 93.890 -77.672 53.277
136934 93.902 -77.667 53.260
136935 93.915 -77.664 53.243
136936 93.929 -77.660 53.227
136937 93.943 -77.655 53.211
136938 93.957 -77.652 53.193
136939 93.969 -77.649 53.176
136940 93.974 -77.650 53.159
136941 93.979 -77.652 53.141
136942 93.984 -77.653 53.121
136943 93.989 -77.654 53.103
136944 93.993 -77.655 53.084
136945 93.999 -77.656 53.064
136946 94.014 -77.643 53.046
136947 94.030 -77.631 53.028
136948 94.050 -77.620 53.012
136949 94.069 -77.608 52.996
136950 94.087 -77.598 52.980
136951 94.090 -77.592 52.964
136952 94.089 -77.588 52.947
136953 94.087 -77.583 52.928
136954 94.086 -77.577 52.908
136955 94.084 -77.573 52.888
136956 94.069 -77.579 52.870
136957 94.050 -77.589 52.851
136958 94.045 -77.600 52.832
136959 94.110 -77.568 52.812
136960 94.184 -77.530 52.792
136961 94.175 -77.544 52.770
136962 94.165 -77.558 52.750
136963 94.187 -77.549 52.733
136964 94.199 -77.551 52.715
136965 94.213 -77.555 52.698
136966 94.227 -77.558 52.680
136967 94.233 -77.564 52.668
136968 94.242 -77.551 52.656
136969 94.249 -77.549 52.646
136970 94.238 -77.557 52.640
136971 94.238 -77.551 52.640
136972 94.242 -77.558 52.640
136973 94.263 -77.545 52.638
136974 94.287 -77.529 52.633
136975 94.322 -77.517 52.626
136976 94.357 -77.505 52.620
136977 94.403 -77.492 52.616
136978 94.450 -77.477 52.612
136979 94.496 -77.463 52.606
136980 94.505 -77.452 52.598
136981 94.498 -77.442 52.588
136982 94.491 -77.432 52.580
136983 94.484 -77.423 52.572
136984 94.476 -77.414 52.564
136985 94.469 -77.434 52.553
136986 94.461 -77.469 52.538
136987 94.453 -77.503 52.521
136988 94.468 -77.510 52.500
136989 94.484 -77.504 52.474
136990 94.501 -77.498 52.446
136991 94.518 -77.492 52.417
136992 94.535 -77.486 52.384
136993 94.551 -77.480 52.349
136994 94.568 -77.475 52.313
136995 94.585 -77.469 52.277
136996 94.602 -77.463 52.244
136997 94.644 -77.496 52.211
136998 94.692 -77.515 52.177
136999 94.740 -77.533 52.145
137000 94.788 -77.551 52.107
137001 94.837 -77.570 52.066
137002 94.885 -77.589 52.024
137003 94.933 -77.607 51.984
137004 94.982 -77.626 51.947
137005 95.030 -77.644 51.911
137006 95.026 -77.633 51.878
137007 95.004 -77.613 51.844
137008 94.982 -77.592 51.808
137009 94.960 -77.572 51.771
137010 94.938 -77.551 51.730
137011 94.964 -77.490 51.692
137012 94.908 -77.489 51.657
137013 94.923 -77.502 51.623
137014 94.937 -77.517 51.587
137015 94.988 -77.495 51.551
137016 95.065 -77.443 51.514
137017 95.106 -77.424 51.473
137018 95.094 -77.449 51.434
137019 95.081 -77.474 51.395
137020 95.068 -77.499 51.356
137021 95.079 -77.513 51.310
137022 95.098 -77.523 51.259
137023 95.128 -77.536 51.204
137024 95.168 -77.548 51.148
137025 95.190 -77.551 51.095
137026 95.213 -77.555 51.044
137027 95.236 -77.558 50.996
137028 95.244 -77.541 50.952
137029 95.252 -77.523 50.908
137030 95.259 -77.504 50.865
137031 95.267 -77.486 50.823
137032 95.280 -77.475 50.786
137033 95.302 -77.476 50.749
137034 95.332 -77.476 50.713
137035 95.365 -77.476 50.680
137036 95.397 -77.476 50.651
137037 95.428 -77.476 50.620
137038 95.461 -77.476 50.586
137039 95.493 -77.477 50.550
137040 95.525 -77.477 50.512
137041 95.547 -77.472 50.473
137042 95.563 -77.464 50.434
137043 95.579 -77.454 50.394
137044 95.594 -77.445 50.355
137045 95.644 -77.462 50.315
137046 95.696 -77.480 50.274
137047 95.749 -77.499 50.234
137048 95.752 -77.514 50.195
137049 95.751 -77.528 50.154
137050 95.751 -77.542 50.112
137051 95.750 -77.557 50.072
137052 95.749 -77.571 50.033
137053 95.749 -77.586 49.993
137054 95.739 -77.598 49.956
137055 95.710 -77.603 49.919
137056 95.681 -77.609 49.884
137057 95.679 -77.617 49.850
137058 95.682 -77.620 49.816
137059 95.686 -77.623 49.782
137060 95.689 -77.626 49.750
137061 95.697 -77.627 49.719
137062 95.707 -77.627 49.688
137063 95.716 -77.628 49.657
137064 95.727 -77.628 49.626
137065 95.737 -77.629 49.595
137066 95.746 -77.630 49.564
137067 95.758 -77.629 49.533
137068 95.772 -77.627 49.500
137069 95.787 -77.626 49.470
137070 95.801 -77.624 49.440
137071 95.815 -77.622 49.413
137072 95.829 -77.620 49.383
137073 95.843 -77.619 49.354
137074 95.858 -77.617 49.324
137075 95.872 -77.616 49.292
137076 95.887 -77.621 49.261
137077 95.902 -77.633 49.229
137078 95.917 -77.643 49.196
137079 95.942 -77.653 49.162
137080 95.969 -77.662 49.125
137081 95.995 -77.672 49.088
137082 96.006 -77.682 49.049
137083 96.016 -77.692 49.012
137084 96.026 -77.702 48.976
137085 96.035 -77.711 48.940
137086 96.045 -77.722 48.905
137087 96.055 -77.732 48.869
137088 96.065 -77.742 48.833
137089 96.075 -77.752 48.797
137090 96.085 -77.761 48.762
137091 96.095 -77.772 48.728
137092 96.105 -77.782 48.694
137093 96.112 -77.801 48.660
137094 96.126 -77.812 48.625
137095 96.141 -77.824 48.591
137096 96.155 -77.835 48.557
137097 96.169 -77.846 48.524
137098 96.183 -77.858 48.493
137099 96.197 -77.870 48.462
137100 96.208 -77.876 48.431
137101 96.207 -77.868 48.400
137102 96.206 -77.862 48.370
137103 96.205 -77.854 48.340
137104 96.204 -77.846 48.313
137105 96.203 -77.839 48.286
137106 96.210 -77.835 48.260
137107 96.219 -77.831 48.233
137108 96.227 -77.828 48.208
137109 96.239 -77.824 48.183
137110 96.260 -77.821 48.158
137111 96.281 -77.817 48.136
137112 96.302 -77.812 48.114
137113 96.315 -77.813 48.092
137114 96.321 -77.818 48.069
137115 96.328 -77.824 48.047
137116 96.335 -77.830 48.024
137117 96.353 -77.842 48.000
137118 96.371 -77.854 47.978
137119 96.389 -77.867 47.957
137120 96.406 -77.879 47.935
137121 96.425 -77.891 47.913
137122 96.442 -77.903 47.890
137123 96.470 -77.911 47.866
137124 96.499 -77.918 47.842
137125 96.529 -77.926 47.818
137126 96.552 -77.931 47.792
137127 96.559 -77.928 47.767
137128 96.567 -77.927 47.741
137129 96.575 -77.925 47.716
137130 96.583 -77.923 47.689
137131 96.590 -77.921 47.664
137132 96.598 -77.919 47.639
137133 96.605 -77.918 47.614
137134 96.613 -77.916 47.590
137135 96.621 -77.914 47.565
137136 96.628 -77.912 47.541
137137 96.636 -77.911 47.516
137138 96.647 -77.912 47.491
137139 96.658 -77.916 47.467
137140 96.676 -77.933 47.442
137141 96.715 -77.949 47.418
137142 96.767 -77.968 47.394
137143 96.821 -77.986 47.369
137144 96.874 -78.005 47.344
137145 96.893 -78.018 47.319
137146 96.891 -78.030 47.294
137147 96.890 -78.042 47.270
137148 96.891 -78.056 47.246
137149 96.891 -78.071 47.223
137150 96.892 -78.085 47.199
137151 96.912 -78.094 47.177
137152 96.936 -78.100 47.154
137153 96.950 -78.097 47.131
137154 96.962 -78.095 47.109
137155 96.983 -78.100 47.088
137156 97.003 -78.106 47.067
137157 97.020 -78.118 47.047
137158 97.037 -78.130 47.028
137159 97.053 -78.142 47.010
137160 97.070 -78.154 46.994
137161 97.087 -78.166 46.978
137162 97.103 -78.178 46.963
137163 97.120 -78.191 46.949
137164 97.138 -78.200 46.935
137165 97.165 -78.181 46.922
137166 97.193 -78.162 46.912
137167 97.221 -78.143 46.902
137168 97.250 -78.123 46.890
137169 97.289 -78.122 46.877
137170 97.321 -78.127 46.864
137171 97.354 -78.133 46.851
137172 97.387 -78.139 46.838
137173 97.419 -78.145 46.825
137174 97.452 -78.151 46.811
137175 97.485 -78.157 46.796
137176 97.517 -78.163 46.781
137177 97.550 -78.169 46.766
137178 97.583 -78.176 46.749
137179 97.616 -78.184 46.731
137180 97.650 -78.191 46.715
137181 97.690 -78.192 46.698
137182 97.731 -78.194 46.681
137183 97.771 -78.196 46.664
137184 97.811 -78.197 46.648
137185 97.832 -78.207 46.632
137186 97.855 -78.189 46.614
137187 97.872 -78.184 46.594
137188 97.888 -78.180 46.574
137189 97.905 -78.176 46.554
137190 97.922 -78.172 46.535
137191 97.939 -78.167 46.515
137192 97.955 -78.163 46.495
137193 97.967 -78.169 46.474
137194 97.994 -78.174 46.454
137195 98.020 -78.179 46.433
137196 98.046 -78.184 46.412
137197 98.049 -78.238 46.391
137198 98.077 -78.244 46.371
137199 98.105 -78.250 46.349
137200 98.133 -78.256 46.327
137201 98.169 -78.263 46.306
137202 98.197 -78.297 46.285
137203 98.213 -78.356 46.265
137204 98.229 -78.412 46.245
137205 98.246 -78.453 46.224
137206 98.249 -78.440 46.204
137207 98.250 -78.427 46.184
137208 98.251 -78.414 46.165
137209 98.252 -78.401 46.147
137210 98.255 -78.388 46.129
137211 98.279 -78.379 46.110
137212 98.303 -78.369 46.089
137213 98.326 -78.360 46.066
137214 98.350 -78.351 46.044
137215 98.407 -78.330 46.022
137216 98.428 -78.308 46.001
137217 98.426 -78.288 45.979
137218 98.396 -78.313 45.958
137219 98.417 -78.307 45.935
137220 98.434 -78.304 45.912
137221 98.446 -78.302 45.888
137222 98.459 -78.301 45.865
137223 98.473 -78.300 45.841
137224 98.486 -78.299 45.817
137225 98.499 -78.298 45.794
137226 98.512 -78.297 45.770
137227 98.551 -78.293 45.746
137228 98.589 -78.275 45.723
137229 98.628 -78.274 45.702
137230 98.665 -78.274 45.681
137231 98.698 -78.255 45.660
137232 98.716 -78.184 45.641
137233 98.722 -78.123 45.623
137234 98.750 -78.141 45.606
137235 98.776 -78.168 45.589
137236 98.803 -78.195 45.573
137237 98.830 -78.222 45.557
137238 98.856 -78.248 45.539
137239 98.877 -78.264 45.523
137240 98.896 -78.277 45.507
137241 98.915 -78.291 45.492
137242 98.932 -78.315 45.478
137243 98.950 -78.340 45.463
137244 98.969 -78.364 45.450
137245 98.987 -78.388 45.435
137246 99.005 -78.413 45.421
137247 99.022 -78.437 45.409
137248 99.041 -78.461 45.395
137249 99.077 -78.474 45.382
137250 99.097 -78.475 45.368
137251 99.134 -78.467 45.354
137252 99.170 -78.460 45.341
137253 99.207 -78.452 45.330
137254 99.243 -78.445 45.319
137255 99.272 -78.438 45.308
137256 99.299 -78.432 45.295
137257 99.327 -78.426 45.281
137258 99.354 -78.420 45.268
137259 99.377 -78.423 45.254
137260 99.392 -78.442 45.239
137261 99.406 -78.460 45.225
137262 99.420 -78.479 45.212
137263 99.435 -78.496 45.200
137264 99.471 -78.502 45.187
137265 99.504 -78.522 45.173
137266 99.542 -78.542 45.158
137267 99.579 -78.562 45.143
137268 99.617 -78.582 45.129
137269 99.655 -78.603 45.115
137270 99.693 -78.623 45.101
137271 99.781 -78.670 45.087
137272 99.778 -78.674 45.074
137273 99.785 -78.682 45.061
137274 99.791 -78.689 45.049
137275 99.830 -78.720 45.037
137276 99.888 -78.764 45.025
137277 99.922 -78.789 45.013
137278 99.944 -78.804 45.002
137279 99.966 -78.819 44.991
137280 99.984 -78.834 44.979
137281 99.999 -78.849 44.967
137282 100.011 -78.866 44.956
137283 100.010 -78.892 44.945
137284 100.010 -78.917 44.934
137285 100.014 -78.902 44.924
137286 100.019 -78.883 44.914
137287 100.025 -78.865 44.904
137288 100.030 -78.846 44.894
137289 100.055 -78.857 44.884
137290 100.081 -78.869 44.875
137291 100.107 -78.881 44.867
137292 100.132 -78.893 44.858
137293 100.157 -78.905 44.849
137294 100.183 -78.918 44.840
137295 100.209 -78.930 44.831
137296 100.234 -78.942 44.822
137297 100.280 -78.957 44.813
137298 100.308 -78.970 44.802
137299 100.318 -78.984 44.792
137300 100.329 -79.002 44.781
137301 100.352 -79.018 44.769
137302 100.384 -79.037 44.758
137303 100.417 -79.056 44.746
137304 100.449 -79.075 44.734
137305 100.482 -79.093 44.721
137306 100.514 -79.113 44.708
137307 100.547 -79.132 44.695
137308 100.579 -79.151 44.681
137309 100.612 -79.170 44.667
137310 100.644 -79.190 44.653
137565 92.943 -78.220 54.217
137566 92.956 -78.215 54.212
137567 92.970 -78.210 54.207
137568 92.984 -78.206 54.201
137569 92.997 -78.200 54.195
137570 93.011 -78.196 54.188
137571 93.024 -78.191 54.181
137572 93.038 -78.186 54.175
137573 93.052 -78.181 54.169
137574 93.066 -78.177 54.163
137575 93.080 -78.165 54.157
137576 93.094 -78.153 54.150
137577 93.108 -78.141 54.143
137578 93.122 -78.128 54.135
137579 93.143 -78.116 54.128
137580 93.166 -78.104 54.121
137581 93.190 -78.093 54.114
137582 93.213 -78.081 54.107
137583 93.237 -78.069 54.099
137584 93.261 -78.057 54.090
137585 93.284 -78.045 54.082
137586 93.308 -78.033 54.075
137587 93.332 -78.022 54.068
137588 93.355 -78.009 54.060
137589 93.379 -77.997 54.053
137590 93.403 -77.986 54.043
137591 93.411 -77.980 54.031
137592 93.421 -77.965 54.020
137593 93.431 -77.949 54.007
137594 93.441 -77.934 53.994
137595 93.451 -77.918 53.981
137596 93.462 -77.902 53.966
137597 93.473 -77.887 53.949
137598 93.484 -77.879 53.932
137599 93.496 -77.871 53.915
137600 93.507 -77.862 53.899
137601 93.519 -77.854 53.883
137602 93.531 -77.846 53.868
137603 93.542 -77.837 53.852
137604 93.554 -77.829 53.835
137605 93.565 -77.821 53.818
137606 93.577 -77.812 53.799
137607 93.588 -77.804 53.782
137608 93.600 -77.796 53.764
137609 93.612 -77.788 53.747
137610 93.623 -77.780 53.731
137611 93.635 -77.771 53.711
137612 93.646 -77.766 53.692
137613 93.657 -77.761 53.674
137614 93.667 -77.756 53.655
137615 93.678 -77.752 53.636
137616 93.689 -77.746 53.618
137617 93.700 -77.742 53.599
137618 93.710 -77.736 53.579
137619 93.721 -77.732 53.559
137620 93.732 -77.727 53.540
137621 93.743 -77.722 53.521
137627 93.807 -77.693 53.412
137628 93.818 -77.688 53.395
137629 93.829 -77.683 53.377
137630 93.839 -77.678 53.360
137631 93.850 -77.674 53.343
137632 93.861 -77.668 53.326
137633 93.871 -77.664 53.308
137634 93.883 -77.659 53.291
137635 93.896 -77.655 53.274
137636 93.910 -77.651 53.257
137637 93.924 -77.647 53.240
137638 93.937 -77.642 53.223
137639 93.952 -77.639 53.204
137640 93.965 -77.635 53.187
137641 93.970 -77.636 53.169
137642 93.975 -77.638 53.151
137643 93.980 -77.639 53.133
137644 93.984 -77.640 53.115
137645 93.989 -77.642 53.096
137646 94.001 -77.634 53.077
137647 94.016 -77.621 53.058
137648 94.032 -77.608 53.039
137649 94.051 -77.597 53.022
137650 94.071 -77.586 53.006
137651 94.088 -77.575 52.990
137652 94.094 -77.567 52.974
137653 94.099 -77.558 52.958
137654 94.100 -77.552 52.940
137655 94.099 -77.547 52.921
137656 94.097 -77.542 52.902
137657 94.095 -77.537 52.884
137658 94.078 -77.545 52.866
137659 94.072 -77.556 52.849
137660 94.077 -77.567 52.832
137661 94.143 -77.551 52.813
137662 94.139 -77.561 52.793
137663 94.170 -77.545 52.774
137664 94.183 -77.544 52.755
137665 94.192 -77.545 52.739
137666 94.205 -77.548 52.723
137667 94.211 -77.555 52.706
137668 94.185 -77.557 52.694
137669 94.216 -77.536 52.682
137670 94.250 -77.520 52.671
137671 94.239 -77.527 52.664
137672 94.227 -77.535 52.663
137673 94.215 -77.542 52.662
137674 94.240 -77.536 52.660
137675 94.275 -77.524 52.657
137676 94.311 -77.513 52.652
137677 94.357 -77.498 52.648
137678 94.404 -77.485 52.643
137679 94.451 -77.470 52.640
137680 94.497 -77.456 52.634
137681 94.489 -77.446 52.626
137682 94.482 -77.437 52.618
137683 94.475 -77.427 52.610
137684 94.468 -77.417 52.602
137685 94.460 -77.407 52.595
137686 94.453 -77.398 52.585
137687 94.445 -77.423 52.571
137688 94.441 -77.457 52.554
137689 94.468 -77.482 52.532
137690 94.489 -77.491 52.504
137691 94.505 -77.485 52.474
137692 94.522 -77.479 52.442
137693 94.539 -77.473 52.409
137694 94.556 -77.467 52.376
137695 94.572 -77.461 52.343
137696 94.589 -77.455 52.311
137697 94.609 -77.456 52.278
137698 94.652 -77.498 52.245
137699 94.701 -77.517 52.211
137700 94.749 -77.536 52.177
137701 94.797 -77.554 52.138
137702 94.846 -77.573 52.098
137703 94.894 -77.591 52.056
137704 94.943 -77.609 52.015
137705 94.991 -77.628 51.977
137706 95.024 -77.636 51.940
137707 95.044 -77.636 51.905
137708 95.022 -77.616 51.870
137709 95.000 -77.595 51.835
137710 95.008 -77.544 51.798
137711 94.979 -77.482 51.760
137712 94.957 -77.445 51.722
137713 94.894 -77.479 51.685
137714 94.882 -77.481 51.650
137715 94.936 -77.454 51.613
137716 95.015 -77.403 51.577
137717 95.093 -77.352 51.540
137718 95.091 -77.368 51.501
137719 95.078 -77.393 51.463
137720 95.066 -77.418 51.428
137721 95.053 -77.444 51.391
137722 95.046 -77.466 51.346
137723 95.064 -77.476 51.296
137724 95.106 -77.492 51.241
137725 95.163 -77.512 51.183
137726 95.192 -77.520 51.127
137727 95.215 -77.523 51.075
137728 95.237 -77.525 51.027
137729 95.245 -77.506 50.979
137730 95.252 -77.488 50.934
137731 95.260 -77.470 50.891
137732 95.276 -77.464 50.849
137733 95.298 -77.464 50.808
137734 95.320 -77.464 50.770
137735 95.342 -77.465 50.733
137736 95.372 -77.465 50.699
137737 95.404 -77.465 50.670
137738 95.437 -77.466 50.640
137739 95.468 -77.466 50.610
137740 95.500 -77.466 50.575
137741 95.533 -77.467 50.538
137742 95.556 -77.462 50.500
137743 95.571 -77.453 50.462
137744 95.587 -77.444 50.421
137745 95.616 -77.445 50.380
137746 95.668 -77.463 50.339
137747 95.721 -77.482 50.299
137748 95.734 -77.497 50.259
137749 95.733 -77.511 50.219
137750 95.732 -77.526 50.177
137751 95.732 -77.540 50.135
137752 95.731 -77.555 50.094
137753 95.730 -77.569 50.053
137754 95.730 -77.584 50.013
137755 95.729 -77.599 49.974
137756 95.708 -77.606 49.937
137757 95.679 -77.612 49.900
137758 95.679 -77.619 49.865
137759 95.683 -77.622 49.830
137760 95.686 -77.624 49.796
137761 95.695 -77.625 49.763
137762 95.705 -77.626 49.732
137763 95.715 -77.627 49.700
137764 95.725 -77.627 49.668
137765 95.736 -77.627 49.636
137766 95.750 -77.625 49.604
137767 95.765 -77.624 49.573
137768 95.779 -77.621 49.543
137769 95.793 -77.620 49.511
137770 95.807 -77.618 49.481
137771 95.821 -77.617 49.452
137772 95.836 -77.615 49.425
137773 95.850 -77.614 49.396
137774 95.864 -77.611 49.366
137775 95.878 -77.610 49.336
137776 95.892 -77.608 49.305
137777 95.907 -77.615 49.273
137778 95.923 -77.627 49.241
137779 95.939 -77.637 49.209
137780 95.954 -77.649 49.174
137781 95.977 -77.658 49.137
137782 95.987 -77.667 49.100
137783 95.997 -77.678 49.062
137784 96.007 -77.688 49.025
137785 96.017 -77.698 48.989
137786 96.027 -77.708 48.954
137787 96.037 -77.718 48.918
137788 96.046 -77.728 48.882
137789 96.056 -77.738 48.846
137790 96.066 -77.748 48.810
137791 96.076 -77.758 48.776
137792 96.086 -77.768 48.741
137793 96.096 -77.778 48.707
137794 96.109 -77.790 48.673
137795 96.124 -77.801 48.639
137796 96.138 -77.812 48.605
137797 96.153 -77.824 48.570
137798 96.167 -77.836 48.537
137799 96.181 -77.847 48.505
137800 96.195 -77.859 48.474
137801 96.209 -77.870 48.443
137802 96.209 -77.863 48.412
137803 96.208 -77.855 48.382
137804 96.207 -77.848 48.351
137805 96.206 -77.840 48.324
137806 96.213 -77.836 48.298
137807 96.221 -77.833 48.271
137808 96.229 -77.830 48.245
137809 96.237 -77.827 48.219
137810 96.251 -77.823 48.194
137811 96.272 -77.818 48.169
137812 96.294 -77.815 48.146
137813 96.313 -77.812 48.124
137814 96.320 -77.817 48.102
137815 96.326 -77.822 48.078
137816 96.332 -77.827 48.056
137817 96.338 -77.833 48.033
137818 96.356 -77.844 48.010
137819 96.374 -77.856 47.987
137820 96.391 -77.868 47.966
137821 96.409 -77.881 47.944
137822 96.428 -77.893 47.922
137823 96.449 -77.904 47.900
137824 96.477 -77.912 47.876
137825 96.506 -77.918 47.852
137826 96.534 -77.926 47.828
137827 96.556 -77.929 47.802
137828 96.565 -77.929 47.777
137829 96.571 -77.927 47.752
137830 96.576 -77.924 47.726
137831 96.584 -77.922 47.699
137832 96.591 -77.921 47.673
137833 96.599 -77.919 47.648
137834 96.607 -77.917 47.624
137835 96.614 -77.915 47.599
137836 96.622 -77.914 47.575
137837 96.630 -77.912 47.551
137838 96.638 -77.910 47.526
137839 96.645 -77.908 47.501
137840 96.655 -77.909 47.477
137841 96.676 -77.922 47.452
137842 96.729 -77.941 47.428
137843 96.783 -77.959 47.404
137844 96.835 -77.978 47.379
137845 96.889 -77.996 47.354
137846 96.899 -78.009 47.330
137847 96.898 -78.022 47.305
137848 96.899 -78.037 47.281
137849 96.900 -78.051 47.258
137850 96.901 -78.065 47.234
137851 96.902 -78.080 47.210
137852 96.924 -78.087 47.188
137853 96.947 -78.094 47.166
137854 96.954 -78.086 47.144
137855 96.967 -78.083 47.121
137856 96.987 -78.087 47.100
137857 97.008 -78.093 47.079
137858 97.028 -78.100 47.060
137859 97.044 -78.112 47.041
137860 97.061 -78.124 47.023
137861 97.078 -78.136 47.006
137862 97.094 -78.148 46.989
137863 97.111 -78.160 46.974
137864 97.127 -78.172 46.960
137865 97.148 -78.174 46.947
137866 97.176 -78.155 46.934
137867 97.204 -78.135 46.923
137868 97.231 -78.116 46.913
137869 97.252 -78.103 46.904
137870 97.273 -78.103 46.893
137871 97.306 -78.109 46.882
137872 97.338 -78.116 46.870
137873 97.371 -78.122 46.857
137874 97.404 -78.128 46.844
137875 97.437 -78.134 46.830
137876 97.469 -78.140 46.816
137877 97.502 -78.146 46.800
137878 97.535 -78.153 46.785
137879 97.568 -78.160 46.769
137880 97.601 -78.168 46.753
137881 97.636 -78.173 46.737
137882 97.676 -78.175 46.721
137883 97.717 -78.177 46.704
137884 97.757 -78.178 46.687
137885 97.788 -78.193 46.671
137886 97.807 -78.206 46.654
137887 97.822 -78.209 46.636
137888 97.837 -78.213 46.616
137889 97.858 -78.202 46.596
137890 97.891 -78.159 46.576
137891 97.911 -78.149 46.556
137892 97.927 -78.144 46.535
137893 97.944 -78.141 46.514
137894 97.959 -78.142 46.493
137895 97.986 -78.147 46.472
137896 97.994 -78.189 46.451
137897 98.000 -78.237 46.431
137898 98.028 -78.242 46.409
137899 98.056 -78.248 46.388
137900 98.094 -78.255 46.366
137901 98.134 -78.263 46.344
137902 98.173 -78.269 46.322
137903 98.203 -78.298 46.300
137904 98.219 -78.357 46.279
137905 98.235 -78.416 46.258
137906 98.238 -78.449 46.237
137907 98.227 -78.447 46.217
137908 98.228 -78.434 46.197
137909 98.229 -78.421 46.177
137910 98.231 -78.408 46.159
137911 98.255 -78.398 46.140
137912 98.279 -78.389 46.121
137913 98.302 -78.380 46.101
137914 98.326 -78.370 46.079
137915 98.350 -78.361 46.057
137916 98.402 -78.341 46.036
137917 98.400 -78.322 46.015
137918 98.398 -78.303 45.993
137919 98.402 -78.281 45.971
137920 98.416 -78.280 45.948
137921 98.430 -78.279 45.924
137922 98.444 -78.279 45.901
137923 98.458 -78.279 45.877
137924 98.472 -78.278 45.853
137925 98.486 -78.277 45.830
137926 98.500 -78.276 45.807
137927 98.514 -78.276 45.783
137928 98.525 -78.269 45.759
137929 98.554 -78.277 45.736
137930 98.585 -78.283 45.713
137931 98.623 -78.283 45.693
137932 98.652 -78.252 45.672
137933 98.687 -78.184 45.653
137934 98.727 -78.116 45.634
137935 98.746 -78.143 45.618
137936 98.773 -78.159 45.602
137937 98.800 -78.176 45.586
137938 98.826 -78.193 45.569
137939 98.853 -78.215 45.551
137940 98.879 -78.241 45.534
137941 98.905 -78.266 45.518
137942 98.923 -78.281 45.504
137943 98.941 -78.305 45.490
137944 98.959 -78.329 45.477
137945 98.978 -78.354 45.464
137946 98.999 -78.378 45.449
137947 99.022 -78.402 45.434
137948 99.044 -78.426 45.420
137949 99.067 -78.450 45.407
137950 99.087 -78.458 45.394
137951 99.108 -78.458 45.382
137952 99.144 -78.451 45.368
137953 99.181 -78.443 45.356
137954 99.217 -78.435 45.344
137955 99.253 -78.428 45.333
137956 99.280 -78.422 45.320
137957 99.308 -78.416 45.307
137958 99.335 -78.410 45.294
137959 99.363 -78.403 45.280
137960 99.390 -78.397 45.267
137961 99.418 -78.391 45.252
137962 99.445 -78.385 45.238
137963 99.464 -78.396 45.225
137964 99.478 -78.415 45.214
137965 99.521 -78.454 45.201
137966 99.559 -78.496 45.188
137967 99.597 -78.538 45.174
137968 99.635 -78.565 45.160
137969 99.673 -78.586 45.145
137970 99.710 -78.606 45.130
137971 99.748 -78.626 45.116
137972 99.797 -78.655 45.102
137973 99.804 -78.662 45.089
137974 99.810 -78.670 45.077
137975 99.817 -78.677 45.063
137976 99.824 -78.685 45.050
137977 99.874 -78.724 45.038
137978 99.917 -78.756 45.027
137979 99.939 -78.771 45.015
137980 99.959 -78.786 45.003
137981 99.975 -78.801 44.990
137982 99.991 -78.815 44.980
137983 100.007 -78.830 44.969
137984 100.022 -78.845 44.958
137985 100.021 -78.871 44.948
137986 100.022 -78.888 44.938
137987 100.027 -78.869 44.927
137988 100.032 -78.851 44.917
137989 100.039 -78.833 44.906
137990 100.064 -78.846 44.897
137991 100.089 -78.858 44.888
137992 100.115 -78.870 44.879
137993 100.141 -78.882 44.870
137994 100.167 -78.894 44.860
137995 100.192 -78.906 44.850
137996 100.231 -78.921 44.841
137997 100.277 -78.936 44.831
137998 100.308 -78.949 44.822
137999 100.318 -78.959 44.810
138000 100.328 -78.976 44.799
138001 100.338 -78.995 44.788
138002 100.352 -79.009 44.776
138003 100.382 -79.027 44.764
138004 100.414 -79.046 44.751
138005 100.446 -79.065 44.738
138006 100.479 -79.084 44.724
138007 100.511 -79.103 44.711
138008 100.544 -79.122 44.698
138009 100.576 -79.141 44.684
138010 100.609 -79.160 44.669
138011 100.641 -79.179 44.655
138265 92.928 -78.210 54.229
138266 92.942 -78.206 54.222
138267 92.955 -78.201 54.217
138268 92.969 -78.197 54.212
138269 92.983 -78.191 54.207
138270 92.997 -78.187 54.201
138271 93.010 -78.182 54.194
138272 93.024 -78.177 54.187
138273 93.038 -78.172 54.181
138274 93.051 -78.168 54.175
138275 93.065 -78.163 54.169
138276 93.079 -78.154 54.163
138277 93.093 -78.142 54.157
138278 93.107 -78.130 54.150
138279 93.121 -78.118 54.142
138280 93.142 -78.106 54.135
138281 93.165 -78.094 54.128
138282 93.189 -78.082 54.121
138283 93.213 -78.070 54.115
138284 93.236 -78.059 54.108
138285 93.260 -78.047 54.099
138286 93.284 -78.034 54.092
138287 93.307 -78.023 54.086
138288 93.331 -78.011 54.079
138289 93.355 -77.999 54.071
138290 93.378 -77.987 54.064
138291 93.402 -77.975 54.055
138292 93.412 -77.964 54.044
138293 93.422 -77.949 54.032
138294 93.432 -77.933 54.019
138295 93.443 -77.918 54.006
138296 93.452 -77.903 53.994
138297 93.461 -77.891 53.980
138298 93.471 -77.878 53.963
138299 93.480 -77.866 53.946
138300 93.491 -77.859 53.929
138301 93.502 -77.850 53.912
138302 93.513 -77.842 53.897
138303 93.524 -77.834 53.881
138304 93.535 -77.825 53.866
138305 93.546 -77.818 53.850
138306 93.557 -77.809 53.832
138307 93.568 -77.801 53.814
138308 93.579 -77.793 53.796
138309 93.590 -77.785 53.778
138310 93.601 -77.777 53.761
138311 93.609 -77.759 53.743
138312 93.626 -77.760 53.724
138313 93.638 -77.752 53.705
138314 93.649 -77.748 53.686
138315 93.660 -77.743 53.667
138316 93.671 -77.738 53.649
138317 93.681 -77.733 53.631
138318 93.692 -77.728 53.613
138319 93.703 -77.723 53.593
138320 93.713 -77.718 53.573
138321 93.724 -77.714 53.554
138322 93.735 -77.708 53.535
138323 93.746 -77.704 53.517
138326 93.778 -77.689 53.463
138327 93.789 -77.684 53.444
138328 93.799 -77.680 53.426
138329 93.810 -77.674 53.408
138330 93.821 -77.670 53.391
138331 93.832 -77.664 53.374
138332 93.842 -77.660 53.358
138333 93.853 -77.655 53.341
138334 93.864 -77.650 53.323
138335 93.876 -77.646 53.305
138336 93.891 -77.642 53.288
138337 93.904 -77.638 53.271
138338 93.918 -77.634 53.254
138339 93.932 -77.630 53.236
138340 93.946 -77.626 53.217
138341 93.960 -77.622 53.199
138342 93.966 -77.623 53.181
138343 93.971 -77.624 53.163
138344 93.975 -77.625 53.145
138345 93.980 -77.627 53.128
138346 93.988 -77.624 53.109
138347 94.003 -77.611 53.090
138348 94.018 -77.599 53.071
138349 94.034 -77.586 53.053
138350 94.053 -77.575 53.035
138351 94.072 -77.564 53.018
138352 94.083 -77.555 53.002
138353 94.088 -77.547 52.985
138354 94.092 -77.539 52.969
138355 94.097 -77.530 52.951
138356 94.102 -77.521 52.934
138357 94.106 -77.513 52.916
138358 94.108 -77.506 52.898
138359 94.106 -77.501 52.882
138360 94.099 -77.512 52.867
138361 94.104 -77.523 52.852
138362 94.113 -77.559 52.836
138363 94.143 -77.548 52.818
138364 94.168 -77.537 52.798
138365 94.178 -77.539 52.779
138366 94.187 -77.539 52.763
138367 94.195 -77.542 52.748
138368 94.146 -77.567 52.733
138369 94.150 -77.538 52.720
138370 94.191 -77.519 52.708
138371 94.231 -77.500 52.696
138372 94.240 -77.498 52.687
138373 94.228 -77.505 52.684
138374 94.214 -77.526 52.682
138375 94.228 -77.532 52.681
138376 94.264 -77.520 52.680
138377 94.311 -77.506 52.677
138378 94.358 -77.492 52.674
138379 94.405 -77.478 52.672
138380 94.452 -77.464 52.669
138381 94.481 -77.451 52.663
138382 94.474 -77.441 52.654
138383 94.466 -77.432 52.645
138384 94.459 -77.422 52.637
138385 94.452 -77.412 52.629
138386 94.445 -77.402 52.621
138387 94.437 -77.393 52.612
138388 94.430 -77.383 52.600
138389 94.435 -77.409 52.583
138390 94.461 -77.433 52.560
138391 94.487 -77.458 52.531
138392 94.510 -77.471 52.499
138393 94.527 -77.466 52.466
138394 94.543 -77.460 52.435
138395 94.560 -77.454 52.405
138396 94.577 -77.448 52.375
138397 94.594 -77.442 52.345
138398 94.619 -77.454 52.311
138399 94.661 -77.501 52.278
138400 94.709 -77.519 52.245
138401 94.758 -77.538 52.208
138402 94.806 -77.556 52.169
138403 94.854 -77.575 52.129
138404 94.903 -77.593 52.087
138405 94.948 -77.610 52.046
138406 94.981 -77.617 52.008
138407 95.014 -77.624 51.971
138408 95.046 -77.631 51.936
138409 95.042 -77.599 51.900
138410 95.026 -77.558 51.865
138411 95.009 -77.516 51.829
138412 94.993 -77.474 51.791
138413 94.966 -77.441 51.753
138414 94.931 -77.415 51.715
138415 94.948 -77.427 51.680
138416 94.963 -77.363 51.643
138417 95.041 -77.312 51.606
138418 95.088 -77.287 51.569
138419 95.075 -77.313 51.530
138420 95.063 -77.338 51.495
138421 95.050 -77.363 51.460
138422 95.038 -77.388 51.425
138423 95.025 -77.413 51.381
138424 95.032 -77.429 51.331
138425 95.083 -77.448 51.277
138426 95.140 -77.468 51.218
138427 95.195 -77.488 51.160
138428 95.217 -77.491 51.107
138429 95.237 -77.490 51.057
138430 95.245 -77.471 51.008
138431 95.252 -77.453 50.963
138432 95.273 -77.451 50.919
138433 95.295 -77.452 50.876
138434 95.316 -77.452 50.833
138435 95.338 -77.453 50.792
138436 95.360 -77.454 50.753
138437 95.382 -77.454 50.718
138438 95.412 -77.454 50.687
138439 95.444 -77.454 50.657
138440 95.476 -77.455 50.629
138441 95.508 -77.455 50.595
138442 95.540 -77.455 50.559
138443 95.565 -77.451 50.522
138444 95.581 -77.442 50.484
138445 95.596 -77.433 50.445
138446 95.640 -77.446 50.406
138447 95.692 -77.464 50.368
138448 95.716 -77.480 50.328
138449 95.714 -77.495 50.286
138450 95.714 -77.509 50.244
138451 95.713 -77.523 50.200
138452 95.712 -77.538 50.160
138453 95.711 -77.552 50.119
138454 95.711 -77.567 50.077
138455 95.710 -77.582 50.035
138456 95.709 -77.596 49.995
138457 95.706 -77.610 49.956
138458 95.678 -77.616 49.918
138459 95.680 -77.620 49.881
138460 95.683 -77.624 49.845
138461 95.693 -77.624 49.811
138462 95.703 -77.624 49.777
138463 95.714 -77.624 49.745
138464 95.729 -77.623 49.713
138465 95.743 -77.620 49.680
138466 95.757 -77.619 49.647
138467 95.771 -77.617 49.616
138468 95.785 -77.616 49.584
138469 95.800 -77.614 49.551
138470 95.814 -77.613 49.519
138471 95.828 -77.611 49.489
138472 95.842 -77.609 49.460
138473 95.856 -77.608 49.431
138474 95.871 -77.606 49.403
138475 95.885 -77.605 49.374
138476 95.899 -77.603 49.345
138477 95.913 -77.601 49.315
138478 95.928 -77.610 49.284
138479 95.944 -77.620 49.253
138480 95.959 -77.631 49.219
138481 95.968 -77.645 49.183
138482 95.978 -77.655 49.147
138483 95.989 -77.664 49.110
138484 96.000 -77.674 49.072
138485 96.010 -77.684 49.037
138486 96.020 -77.694 49.002
138487 96.031 -77.704 48.967
138488 96.041 -77.714 48.931
138489 96.051 -77.724 48.896
138490 96.061 -77.733 48.861
138491 96.072 -77.744 48.825
138492 96.082 -77.754 48.790
138493 96.093 -77.764 48.756
138494 96.096 -77.773 48.722
138495 96.109 -77.782 48.688
138496 96.122 -77.791 48.653
138497 96.136 -77.801 48.618
138498 96.150 -77.813 48.583
138499 96.165 -77.824 48.551
138500 96.179 -77.836 48.519
138501 96.193 -77.847 48.487
138502 96.207 -77.859 48.456
138503 96.211 -77.857 48.425
138504 96.210 -77.849 48.394
138505 96.209 -77.842 48.363
138506 96.215 -77.837 48.336
138507 96.223 -77.834 48.309
138508 96.231 -77.831 48.283
138509 96.239 -77.827 48.257
138510 96.247 -77.824 48.231
138511 96.264 -77.821 48.205
138512 96.285 -77.817 48.181
138513 96.306 -77.812 48.157
138514 96.318 -77.815 48.135
138515 96.324 -77.820 48.112
138516 96.330 -77.825 48.090
138517 96.337 -77.830 48.067
138518 96.343 -77.836 48.043
138519 96.359 -77.846 48.020
138520 96.376 -77.859 47.998
138521 96.394 -77.871 47.976
138522 96.412 -77.884 47.954
138523 96.430 -77.896 47.932
138524 96.456 -77.904 47.908
138525 96.484 -77.912 47.884
138526 96.513 -77.918 47.859
138527 96.538 -77.924 47.835
138528 96.556 -77.920 47.812
138529 96.573 -77.921 47.787
138530 96.592 -77.928 47.762
138531 96.602 -77.930 47.736
138532 96.608 -77.928 47.710
138533 96.614 -77.925 47.684
138534 96.619 -77.923 47.658
138535 96.625 -77.921 47.634
138536 96.631 -77.918 47.609
138537 96.637 -77.916 47.585
138538 96.643 -77.914 47.561
138539 96.648 -77.912 47.537
138540 96.654 -77.909 47.512
138541 96.660 -77.907 47.488
138542 96.691 -77.915 47.463
138543 96.744 -77.933 47.439
138544 96.797 -77.951 47.414
138545 96.850 -77.970 47.390
138546 96.903 -77.988 47.365
138547 96.907 -78.002 47.341
138548 96.907 -78.016 47.317
138549 96.908 -78.031 47.294
138550 96.909 -78.046 47.270
138551 96.910 -78.060 47.247
138552 96.911 -78.075 47.224
138553 96.935 -78.081 47.202
138554 96.951 -78.082 47.180
138555 96.959 -78.075 47.158
138556 96.971 -78.071 47.136
138557 96.992 -78.075 47.114
138558 97.013 -78.081 47.093
138559 97.033 -78.085 47.074
138560 97.052 -78.094 47.055
138561 97.069 -78.106 47.036
138562 97.085 -78.119 47.019
138563 97.102 -78.131 47.002
138564 97.118 -78.143 46.986
138565 97.135 -78.155 46.972
138566 97.159 -78.147 46.958
138567 97.186 -78.128 46.945
138568 97.214 -78.109 46.935
138569 97.238 -78.094 46.925
138570 97.246 -78.090 46.917
138571 97.257 -78.086 46.909
138572 97.290 -78.092 46.898
138573 97.323 -78.098 46.887
138574 97.356 -78.104 46.874
138575 97.388 -78.110 46.861
138576 97.421 -78.116 46.847
138577 97.454 -78.122 46.833
138578 97.487 -78.129 46.818
138579 97.520 -78.137 46.803
138580 97.553 -78.144 46.789
138581 97.586 -78.151 46.773
138582 97.623 -78.156 46.758
138583 97.663 -78.157 46.742
138584 97.703 -78.159 46.725
138585 97.739 -78.166 46.708
138586 97.764 -78.191 46.692
138587 97.783 -78.204 46.675
138588 97.798 -78.207 46.656
138589 97.812 -78.211 46.637
138590 97.827 -78.214 46.617
138591 97.842 -78.218 46.597
138592 97.857 -78.221 46.575
138593 97.891 -78.178 46.554
138594 97.925 -78.135 46.532
138595 97.946 -78.128 46.511
138596 97.942 -78.194 46.490
138597 97.951 -78.235 46.469
138598 97.980 -78.241 46.449
138599 98.020 -78.248 46.427
138600 98.059 -78.255 46.405
138601 98.098 -78.263 46.382
138602 98.138 -78.269 46.359
138603 98.177 -78.276 46.337
138604 98.210 -78.300 46.315
138605 98.226 -78.358 46.293
138606 98.242 -78.416 46.272
138607 98.238 -78.436 46.251
138608 98.224 -78.436 46.230
138609 98.210 -78.436 46.209
138610 98.208 -78.428 46.188
138611 98.231 -78.418 46.170
138612 98.255 -78.409 46.151
138613 98.279 -78.399 46.131
138614 98.302 -78.390 46.110
138615 98.326 -78.380 46.089
138616 98.350 -78.371 46.067
138617 98.374 -78.356 46.047
138618 98.372 -78.330 46.026
138619 98.371 -78.297 46.005
138620 98.425 -78.272 45.983
138621 98.441 -78.270 45.961
138622 98.455 -78.269 45.937
138623 98.469 -78.269 45.913
138624 98.483 -78.269 45.889
138625 98.497 -78.268 45.865
138626 98.511 -78.267 45.842
138627 98.525 -78.266 45.818
138628 98.538 -78.260 45.795
138629 98.548 -78.244 45.771
138630 98.560 -78.244 45.747
138631 98.589 -78.252 45.725
138632 98.618 -78.260 45.705
138633 98.654 -78.256 45.685
138634 98.695 -78.188 45.665
138635 98.733 -78.130 45.646
138636 98.745 -78.167 45.630
138637 98.770 -78.185 45.615
138638 98.797 -78.202 45.599
138639 98.823 -78.219 45.583
138640 98.850 -78.235 45.565
138641 98.876 -78.252 45.548
138642 98.903 -78.269 45.531
138643 98.937 -78.279 45.516
138644 98.968 -78.294 45.502
138645 98.990 -78.318 45.488
138646 99.012 -78.342 45.475
138647 99.034 -78.366 45.460
138648 99.057 -78.390 45.446
138649 99.079 -78.414 45.431
138650 99.100 -78.431 45.418
138651 99.120 -78.432 45.406
138652 99.139 -78.432 45.394
138653 99.158 -78.432 45.382
138654 99.192 -78.426 45.370
138655 99.228 -78.419 45.358
138656 99.261 -78.411 45.346
138657 99.289 -78.405 45.332
138658 99.316 -78.399 45.319
138659 99.344 -78.393 45.306
138660 99.371 -78.387 45.292
138661 99.399 -78.380 45.279
138662 99.426 -78.374 45.265
138663 99.456 -78.373 45.252
138664 99.487 -78.375 45.239
138665 99.518 -78.382 45.227
138666 99.560 -78.416 45.215
138667 99.603 -78.451 45.202
138668 99.645 -78.487 45.190
138669 99.688 -78.523 45.176
138670 99.727 -78.563 45.162
138671 99.766 -78.604 45.147
138672 99.803 -78.630 45.133
138673 99.824 -78.643 45.119
138674 99.830 -78.651 45.106
138675 99.837 -78.658 45.093
138676 99.843 -78.666 45.079
138677 99.850 -78.673 45.066
138678 99.861 -78.684 45.053
138679 99.912 -78.723 45.040
138680 99.934 -78.738 45.029
138681 99.950 -78.752 45.017
138682 99.967 -78.767 45.005
138683 99.982 -78.782 44.994
138684 99.998 -78.797 44.983
138685 100.014 -78.811 44.972
138686 100.030 -78.826 44.961
138687 100.032 -78.850 44.951
138688 100.034 -78.855 44.941
138689 100.040 -78.836 44.930
138690 100.047 -78.822 44.920
138691 100.073 -78.834 44.911
138692 100.099 -78.846 44.901
138693 100.124 -78.858 44.892
138694 100.149 -78.871 44.882
138695 100.183 -78.883 44.871
138696 100.228 -78.899 44.861
138697 100.274 -78.914 44.851
138698 100.307 -78.927 44.841
138699 100.317 -78.938 44.830
138700 100.327 -78.949 44.818
138701 100.337 -78.968 44.806
138702 100.348 -78.987 44.794
138703 100.360 -79.004 44.782
138704 100.379 -79.016 44.769
138705 100.412 -79.035 44.756
138706 100.444 -79.055 44.742
138707 100.477 -79.074 44.728
138708 100.509 -79.093 44.715
138709 100.542 -79.112 44.701
138710 100.574 -79.131 44.686
138711 100.607 -79.150 44.671
138712 100.639 -79.169 44.656
138966 92.927 -78.197 54.233
138967 92.941 -78.192 54.227
138968 92.955 -78.187 54.222
138969 92.969 -78.182 54.217
138970 92.982 -78.178 54.212
138971 92.996 -78.173 54.207
138972 93.009 -78.168 54.200
138973 93.023 -78.163 54.193
138974 93.037 -78.159 54.187
138975 93.051 -78.153 54.181
138976 93.064 -78.149 54.175
138977 93.078 -78.144 54.169
138978 93.092 -78.132 54.163
138979 93.106 -78.120 54.156
138980 93.120 -78.107 54.149
138981 93.141 -78.096 54.142
138982 93.165 -78.084 54.135
138983 93.189 -78.072 54.129
138984 93.212 -78.060 54.123
138985 93.236 -78.048 54.116
138986 93.260 -78.036 54.107
138987 93.283 -78.025 54.101
138988 93.307 -78.012 54.095
138989 93.331 -78.000 54.089
138990 93.354 -77.989 54.082
138991 93.377 -77.977 54.075
138992 93.400 -77.962 54.066
138993 93.408 -77.957 54.056
138994 93.416 -77.945 54.044
138995 93.424 -77.932 54.033
138996 93.432 -77.920 54.021
138997 93.440 -77.908 54.009
138998 93.448 -77.896 53.994
138999 93.466 -77.884 53.978
139000 93.473 -77.868 53.961
139001 93.481 -77.852 53.945
139002 93.490 -77.839 53.928
139003 93.501 -77.830 53.912
139004 93.512 -77.823 53.897
139005 93.523 -77.815 53.882
139006 93.534 -77.806 53.866
139007 93.545 -77.798 53.848
139008 93.556 -77.790 53.831
139009 93.567 -77.782 53.811
139010 93.576 -77.771 53.793
139011 93.580 -77.743 53.775
139012 93.599 -77.745 53.757
139013 93.617 -77.748 53.738
139014 93.632 -77.738 53.719
139015 93.641 -77.734 53.700
139016 93.652 -77.729 53.680
139017 93.663 -77.724 53.662
139018 93.674 -77.719 53.644
139019 93.684 -77.714 53.626
139020 93.695 -77.710 53.607
139021 93.706 -77.705 53.588
139022 93.716 -77.700 53.569
139023 93.727 -77.695 53.550
139024 93.738 -77.690 53.532
139025 93.749 -77.686 53.514
139026 93.760 -77.680 53.496
139027 93.770 -77.676 53.478
139028 93.781 -77.671 53.460
139029 93.792 -77.666 53.442
139030 93.802 -77.661 53.424
139031 93.814 -77.656 53.407
139032 93.824 -77.652 53.390
139033 93.835 -77.646 53.373
139034 93.846 -77.642 53.355
139035 93.857 -77.637 53.338
139036 93.871 -77.633 53.321
139037 93.885 -77.629 53.303
139038 93.899 -77.625 53.286
139039 93.913 -77.621 53.268
139040 93.927 -77.617 53.250
139041 93.940 -77.613 53.231
139042 93.955 -77.609 53.213
139043 93.962 -77.609 53.196
139044 93.966 -77.610 53.178
139045 93.971 -77.611 53.160
139046 93.976 -77.613 53.142
139047 93.990 -77.602 53.123
139048 94.005 -77.589 53.105
139049 94.020 -77.577 53.086
139050 94.035 -77.564 53.068
139051 94.055 -77.552 53.050
139052 94.068 -77.543 53.032
139053 94.077 -77.535 53.015
139054 94.082 -77.526 52.999
139055 94.086 -77.518 52.981
139056 94.091 -77.510 52.964
139057 94.096 -77.501 52.948
139058 94.100 -77.492 52.932
139059 94.105 -77.484 52.916
139060 94.112 -77.476 52.901
139061 94.140 -77.476 52.887
139062 94.128 -77.500 52.873
139063 94.133 -77.520 52.859
139064 94.148 -77.535 52.842
139065 94.163 -77.532 52.823
139066 94.179 -77.531 52.804
139067 94.198 -77.529 52.788
139068 94.188 -77.534 52.773
139069 94.155 -77.530 52.760
139070 94.133 -77.530 52.749
139071 94.165 -77.503 52.737
139072 94.206 -77.484 52.725
139073 94.240 -77.468 52.714
139074 94.227 -77.486 52.707
139075 94.213 -77.514 52.704
139076 94.218 -77.528 52.702
139077 94.265 -77.514 52.703
139078 94.312 -77.499 52.702
139079 94.359 -77.485 52.700
139080 94.406 -77.471 52.699
139081 94.453 -77.457 52.696
139082 94.465 -77.445 52.691
139083 94.458 -77.436 52.684
139084 94.451 -77.426 52.676
139085 94.443 -77.417 52.668
139086 94.435 -77.407 52.659
139087 94.428 -77.397 52.650
139088 94.421 -77.387 52.642
139089 94.415 -77.375 52.628
139090 94.427 -77.370 52.610
139091 94.455 -77.385 52.586
139092 94.481 -77.410 52.556
139093 94.507 -77.435 52.525
139094 94.531 -77.452 52.493
139095 94.548 -77.446 52.465
139096 94.564 -77.441 52.436
139097 94.581 -77.435 52.407
139098 94.598 -77.429 52.376
139099 94.628 -77.452 52.343
139100 94.670 -77.501 52.310
139101 94.718 -77.521 52.276
139102 94.767 -77.540 52.239
139103 94.815 -77.558 52.199
139104 94.863 -77.577 52.157
139105 94.906 -77.591 52.116
139106 94.938 -77.599 52.078
139107 94.971 -77.606 52.041
139108 95.005 -77.603 52.005
139109 95.036 -77.598 51.969
139110 95.046 -77.581 51.934
139111 95.030 -77.539 51.899
139112 95.013 -77.497 51.862
139113 94.996 -77.455 51.824
139114 94.970 -77.422 51.786
139115 94.939 -77.395 51.748
139116 94.939 -77.357 51.713
139117 95.068 -77.367 51.675
139118 95.100 -77.276 51.637
139119 95.068 -77.246 51.599
139120 95.055 -77.273 51.562
139121 95.042 -77.299 51.527
139122 95.029 -77.325 51.493
139123 95.017 -77.351 51.458
139124 95.003 -77.378 51.414
139125 95.006 -77.386 51.365
139126 95.060 -77.404 51.312
139127 95.117 -77.424 51.253
139128 95.174 -77.444 51.193
139129 95.220 -77.459 51.137
139130 95.238 -77.455 51.086
139131 95.247 -77.439 51.037
139132 95.269 -77.439 50.990
139133 95.290 -77.440 50.945
139134 95.313 -77.441 50.903
139135 95.334 -77.442 50.858
139136 95.356 -77.442 50.814
139137 95.378 -77.442 50.773
139138 95.400 -77.443 50.737
139139 95.422 -77.444 50.704
139140 95.452 -77.444 50.674
139141 95.484 -77.444 50.645
139142 95.516 -77.444 50.612
139143 95.548 -77.445 50.577
139144 95.573 -77.441 50.542
139145 95.589 -77.432 50.504
139146 95.612 -77.429 50.467
139147 95.664 -77.447 50.431
139148 95.697 -77.464 50.394
139149 95.696 -77.478 50.355
139150 95.695 -77.492 50.314
139151 95.695 -77.507 50.271
139152 95.694 -77.521 50.227
139153 95.693 -77.536 50.186
139154 95.692 -77.551 50.145
139155 95.692 -77.565 50.104
139156 95.691 -77.580 50.059
139157 95.690 -77.594 50.017
139158 95.690 -77.608 49.977
139159 95.676 -77.620 49.937
139160 95.682 -77.621 49.899
139161 95.692 -77.622 49.862
139162 95.706 -77.620 49.827
139163 95.721 -77.618 49.792
139164 95.735 -77.617 49.759
139165 95.749 -77.615 49.726
139166 95.763 -77.614 49.693
139167 95.777 -77.612 49.660
139168 95.792 -77.610 49.627
139169 95.806 -77.608 49.595
139170 95.820 -77.607 49.561
139171 95.834 -77.605 49.529
139172 95.848 -77.604 49.499
139173 95.863 -77.602 49.469
139174 95.877 -77.600 49.439
139175 95.891 -77.599 49.409
139176 95.905 -77.597 49.379
139177 95.919 -77.595 49.350
139178 95.934 -77.594 49.319
139179 95.949 -77.604 49.290
139180 95.964 -77.614 49.260
139181 95.970 -77.630 49.227
139182 95.970 -77.649 49.191
139183 95.983 -77.655 49.155
139184 95.996 -77.662 49.119
139185 96.008 -77.670 49.083
139186 96.018 -77.680 49.048
139187 96.028 -77.689 49.013
139188 96.039 -77.700 48.979
139189 96.049 -77.710 48.944
139190 96.059 -77.720 48.910
139191 96.070 -77.730 48.875
139192 96.080 -77.740 48.839
139193 96.090 -77.749 48.805
139194 96.093 -77.758 48.771
139195 96.096 -77.767 48.737
139196 96.109 -77.777 48.702
139197 96.122 -77.786 48.667
139198 96.136 -77.795 48.632
139199 96.149 -77.804 48.596
139200 96.163 -77.814 48.564
139201 96.177 -77.824 48.533
139202 96.191 -77.836 48.501
139203 96.205 -77.847 48.469
139204 96.213 -77.851 48.438
139205 96.212 -77.843 48.406
139206 96.218 -77.839 48.376
139207 96.226 -77.836 48.349
139208 96.234 -77.832 48.322
139209 96.242 -77.829 48.296
139210 96.250 -77.826 48.270
139211 96.258 -77.822 48.244
139212 96.277 -77.818 48.218
139213 96.298 -77.815 48.193
139214 96.316 -77.812 48.169
139215 96.322 -77.818 48.146
139216 96.328 -77.823 48.124
139217 96.335 -77.828 48.102
139218 96.341 -77.834 48.078
139219 96.347 -77.839 48.054
139220 96.362 -77.849 48.031
139221 96.379 -77.861 48.008
139222 96.397 -77.873 47.986
139223 96.415 -77.886 47.964
139224 96.435 -77.897 47.941
139225 96.463 -77.904 47.917
139226 96.491 -77.912 47.892
139227 96.519 -77.918 47.868
139228 96.539 -77.918 47.843
139229 96.557 -77.916 47.820
139230 96.572 -77.906 47.796
139231 96.590 -77.913 47.771
139232 96.609 -77.920 47.746
139233 96.627 -77.928 47.720
139234 96.640 -77.931 47.695
139235 96.646 -77.928 47.669
139236 96.651 -77.926 47.645
139237 96.657 -77.924 47.620
139238 96.663 -77.921 47.596
139239 96.668 -77.919 47.572
139240 96.674 -77.917 47.547
139241 96.680 -77.915 47.523
139242 96.686 -77.912 47.500
139243 96.708 -77.917 47.475
139244 96.760 -77.934 47.450
139245 96.813 -77.952 47.426
139246 96.866 -77.969 47.402
139247 96.904 -77.981 47.378
139248 96.916 -77.996 47.353
139249 96.917 -78.011 47.329
139250 96.918 -78.025 47.306
139251 96.919 -78.040 47.283
139252 96.920 -78.054 47.260
139253 96.922 -78.069 47.238
139254 96.946 -78.075 47.216
139255 96.956 -78.070 47.194
139256 96.964 -78.062 47.172
139257 96.975 -78.059 47.150
139258 96.996 -78.063 47.128
139259 97.017 -78.069 47.107
139260 97.037 -78.073 47.088
139261 97.058 -78.078 47.068
139262 97.076 -78.089 47.049
139263 97.093 -78.101 47.031
139264 97.109 -78.113 47.015
139265 97.126 -78.125 46.999
139266 97.143 -78.138 46.984
139267 97.169 -78.121 46.970
139268 97.197 -78.102 46.957
139269 97.223 -78.085 46.946
139270 97.232 -78.081 46.937
139271 97.241 -78.076 46.928
139272 97.249 -78.072 46.922
139273 97.275 -78.075 46.913
139274 97.308 -78.081 46.902
139275 97.341 -78.087 46.890
139276 97.373 -78.093 46.877
139277 97.406 -78.099 46.863
139278 97.439 -78.106 46.849
139279 97.472 -78.113 46.836
139280 97.505 -78.120 46.821
139281 97.538 -78.128 46.806
139282 97.571 -78.134 46.791
139283 97.609 -78.138 46.778
139284 97.649 -78.141 46.763
139285 97.689 -78.142 46.745
139286 97.715 -78.164 46.728
139287 97.739 -78.188 46.711
139288 97.758 -78.203 46.694
139289 97.773 -78.206 46.677
139290 97.788 -78.210 46.656
139291 97.803 -78.213 46.635
139292 97.818 -78.216 46.614
139293 97.832 -78.219 46.592
139294 97.848 -78.222 46.569
139295 97.862 -78.226 46.548
139296 97.889 -78.200 46.528
139297 97.905 -78.234 46.507
139298 97.945 -78.241 46.486
139299 97.984 -78.247 46.464
139300 98.024 -78.255 46.442
139301 98.063 -78.262 46.420
139302 98.102 -78.269 46.397
139303 98.142 -78.276 46.374
139304 98.181 -78.283 46.353
139305 98.216 -78.300 46.330
139306 98.232 -78.360 46.308
139307 98.248 -78.418 46.286
139308 98.237 -78.424 46.264
139309 98.224 -78.424 46.243
139310 98.210 -78.424 46.222
139311 98.224 -78.410 46.200
139312 98.253 -78.402 46.181
139313 98.276 -78.394 46.161
139314 98.298 -78.385 46.141
139315 98.321 -78.377 46.120
139316 98.344 -78.369 46.099
139317 98.351 -78.364 46.077
139318 98.347 -78.348 46.057
139319 98.345 -78.316 46.036
139320 98.377 -78.287 46.015
139321 98.448 -78.263 45.994
139322 98.466 -78.260 45.972
139323 98.480 -78.260 45.948
139324 98.494 -78.260 45.925
139325 98.508 -78.259 45.901
139326 98.522 -78.258 45.877
139327 98.536 -78.257 45.853
139328 98.550 -78.254 45.830
139329 98.560 -78.238 45.806
139330 98.570 -78.221 45.782
139331 98.567 -78.211 45.759
139332 98.596 -78.219 45.737
139333 98.625 -78.227 45.717
139334 98.662 -78.226 45.697
139335 98.703 -78.194 45.677
139336 98.739 -78.144 45.658
139337 98.750 -78.181 45.642
139338 98.768 -78.212 45.627
139339 98.794 -78.228 45.612
139340 98.820 -78.244 45.597
139341 98.847 -78.261 45.580
139342 98.874 -78.277 45.562
139343 98.912 -78.284 45.545
139344 98.947 -78.294 45.528
139345 98.982 -78.304 45.513
139346 99.017 -78.314 45.498
139347 99.047 -78.330 45.484
139348 99.069 -78.354 45.470
139349 99.091 -78.378 45.455
139350 99.114 -78.402 45.440
139351 99.133 -78.404 45.428
139352 99.152 -78.404 45.417
139353 99.171 -78.405 45.406
139354 99.191 -78.405 45.394
139355 99.210 -78.406 45.383
139356 99.238 -78.401 45.371
139357 99.269 -78.395 45.358
139358 99.297 -78.388 45.345
139359 99.325 -78.383 45.332
139360 99.356 -78.385 45.318
139361 99.386 -78.385 45.305
139362 99.417 -78.387 45.292
139363 99.448 -78.388 45.278
139364 99.479 -78.389 45.264
139365 99.510 -78.391 45.251
139366 99.541 -78.392 45.240
139367 99.573 -78.413 45.228
139368 99.615 -78.447 45.215
139369 99.657 -78.482 45.203
139370 99.701 -78.518 45.190
139371 99.744 -78.554 45.176
139372 99.786 -78.589 45.162
139373 99.829 -78.625 45.149
139374 99.840 -78.633 45.134
139375 99.850 -78.640 45.120
139376 99.860 -78.647 45.107
139377 99.869 -78.655 45.094
139378 99.875 -78.662 45.081
139379 99.882 -78.669 45.067
139380 99.906 -78.689 45.054
139381 99.926 -78.705 45.043
139382 99.942 -78.720 45.031
139383 99.958 -78.734 45.020
139384 99.974 -78.749 45.008
139385 99.990 -78.764 44.997
139386 100.005 -78.778 44.986
139387 100.021 -78.793 44.975
139388 100.037 -78.808 44.965
139389 100.043 -78.829 44.954
139390 100.047 -78.823 44.944
139391 100.056 -78.810 44.935
139392 100.081 -78.822 44.926
139393 100.107 -78.834 44.916
139394 100.135 -78.847 44.905
139395 100.180 -78.862 44.894
139396 100.225 -78.877 44.882
139397 100.271 -78.893 44.872
139398 100.306 -78.906 44.860
139399 100.316 -78.917 44.849
139400 100.326 -78.927 44.838
139401 100.336 -78.941 44.826
139402 100.347 -78.960 44.813
139403 100.357 -78.979 44.801
139404 100.368 -78.999 44.788
139405 100.384 -79.010 44.774
139406 100.409 -79.025 44.761
139407 100.441 -79.044 44.747
139408 100.474 -79.064 44.733
139409 100.506 -79.083 44.718
139410 100.539 -79.102 44.703
139411 100.571 -79.121 44.689
139412 100.604 -79.140 44.674
139413 100.636 -79.159 44.659
139666 92.913 -78.188 54.243
139667 92.927 -78.183 54.237
139668 92.940 -78.178 54.231
139669 92.954 -78.173 54.226
139670 92.968 -78.169 54.221
139671 92.981 -78.164 54.216
139672 92.995 -78.159 54.210
139673 93.009 -78.154 54.204
139674 93.022 -78.150 54.197
139675 93.036 -78.144 54.192
139676 93.050 -78.140 54.186
139677 93.063 -78.135 54.180
139678 93.077 -78.130 54.175
139679 93.091 -78.122 54.169
139680 93.105 -78.109 54.162
139681 93.119 -78.097 54.155
139682 93.141 -78.085 54.148
139683 93.164 -78.073 54.142
139684 93.188 -78.062 54.137
139685 93.212 -78.050 54.131
139686 93.235 -78.038 54.124
139687 93.259 -78.026 54.116
139688 93.282 -78.014 54.109
139689 93.306 -78.003 54.104
139690 93.330 -77.990 54.098
139691 93.353 -77.977 54.092
139692 93.373 -77.959 54.085
139693 93.393 -77.943 54.077
139694 93.398 -77.947 54.067
139695 93.404 -77.949 54.056
139696 93.413 -77.937 54.045
139697 93.421 -77.924 54.034
139698 93.429 -77.912 54.022
139699 93.437 -77.899 54.007
139700 93.461 -77.890 53.992
139701 93.468 -77.874 53.976
139702 93.476 -77.858 53.960
139703 93.483 -77.842 53.944
139704 93.491 -77.825 53.929
139705 93.499 -77.812 53.914
139706 93.510 -77.803 53.900
139707 93.521 -77.795 53.884
139708 93.533 -77.787 53.866
139709 93.544 -77.779 53.847
139710 93.549 -77.755 53.828
139711 93.552 -77.727 53.809
139712 93.570 -77.730 53.790
139713 93.590 -77.736 53.772
139714 93.609 -77.736 53.753
139715 93.624 -77.727 53.733
139716 93.636 -77.719 53.714
139717 93.644 -77.715 53.695
139718 93.655 -77.711 53.676
139719 93.666 -77.706 53.657
139720 93.677 -77.701 53.640
139721 93.688 -77.696 53.622
139722 93.698 -77.691 53.603
139723 93.709 -77.686 53.584
139724 93.720 -77.682 53.565
139725 93.730 -77.677 53.547
139726 93.741 -77.672 53.530
139727 93.752 -77.667 53.511
139728 93.763 -77.662 53.494
139729 93.773 -77.657 53.476
139730 93.784 -77.652 53.458
139731 93.795 -77.648 53.440
139732 93.805 -77.642 53.423
139733 93.817 -77.638 53.406
139734 93.827 -77.633 53.389
139735 93.838 -77.628 53.371
139736 93.852 -77.624 53.354
139737 93.865 -77.620 53.337
139738 93.880 -77.616 53.319
139739 93.894 -77.612 53.302
139740 93.907 -77.608 53.284
139741 93.921 -77.604 53.267
139742 93.935 -77.600 53.249
139743 93.949 -77.595 53.231
139744 93.957 -77.595 53.213
139745 93.962 -77.596 53.195
139746 93.967 -77.598 53.178
139747 93.977 -77.592 53.160
139748 93.992 -77.580 53.142
139749 94.007 -77.567 53.124
139750 94.022 -77.554 53.106
139751 94.037 -77.541 53.087
139752 94.054 -77.530 53.068
139753 94.063 -77.523 53.051
139754 94.071 -77.515 53.034
139755 94.075 -77.506 53.018
139756 94.080 -77.498 53.000
139757 94.084 -77.489 52.984
139758 94.089 -77.481 52.968
139759 94.098 -77.474 52.952
139760 94.119 -77.473 52.938
139761 94.146 -77.472 52.925
139762 94.178 -77.472 52.911
139763 94.210 -77.473 52.896
139764 94.243 -77.473 52.882
139765 94.220 -77.500 52.866
139766 94.222 -77.511 52.849
139767 94.245 -77.508 52.829
139768 94.247 -77.501 52.812
139769 94.211 -77.510 52.798
139770 94.198 -77.526 52.788
139771 94.173 -77.530 52.780
139772 94.169 -77.520 52.769
139773 94.196 -77.491 52.756
139774 94.207 -77.436 52.743
139775 94.197 -77.448 52.734
139776 94.189 -77.480 52.727
139777 94.217 -77.511 52.724
139778 94.266 -77.507 52.724
139779 94.313 -77.492 52.724
139780 94.360 -77.478 52.723
139781 94.407 -77.464 52.722
139782 94.454 -77.450 52.720
139783 94.449 -77.440 52.717
139784 94.441 -77.430 52.712
139785 94.434 -77.420 52.706
139786 94.427 -77.411 52.697
139787 94.420 -77.401 52.688
139788 94.412 -77.392 52.678
139789 94.408 -77.374 52.669
139790 94.404 -77.357 52.655
139791 94.416 -77.353 52.635
139792 94.445 -77.363 52.610
139793 94.473 -77.373 52.581
139794 94.501 -77.387 52.552
139795 94.527 -77.412 52.524
139796 94.552 -77.433 52.496
139797 94.569 -77.427 52.467
139798 94.585 -77.421 52.436
139799 94.602 -77.416 52.403
139800 94.638 -77.450 52.369
139801 94.680 -77.499 52.336
139802 94.727 -77.523 52.302
139803 94.775 -77.542 52.265
139804 94.824 -77.561 52.225
139805 94.863 -77.573 52.186
139806 94.896 -77.580 52.148
139807 94.930 -77.578 52.112
139808 94.964 -77.574 52.077
139809 94.998 -77.570 52.043
139810 95.024 -77.564 52.009
139811 95.043 -77.555 51.974
139812 95.033 -77.520 51.938
139813 95.017 -77.478 51.901
139814 95.000 -77.436 51.862
139815 94.974 -77.404 51.823
139816 94.949 -77.374 51.784
139817 94.950 -77.336 51.748
139818 94.950 -77.298 51.710
139819 95.042 -77.296 51.671
139820 95.016 -77.332 51.633
139821 95.023 -77.343 51.594
139822 95.023 -77.347 51.559
139823 95.023 -77.351 51.524
139824 95.024 -77.355 51.488
139825 95.024 -77.359 51.445
139826 95.033 -77.371 51.397
139827 95.062 -77.387 51.343
139828 95.111 -77.398 51.284
139829 95.160 -77.410 51.224
139830 95.209 -77.421 51.167
139831 95.243 -77.427 51.114
139832 95.265 -77.428 51.064
139833 95.287 -77.429 51.016
139834 95.309 -77.429 50.969
139835 95.331 -77.429 50.925
139836 95.352 -77.430 50.879
139837 95.374 -77.430 50.835
139838 95.396 -77.431 50.793
139839 95.418 -77.432 50.756
139840 95.440 -77.432 50.722
139841 95.462 -77.432 50.690
139842 95.491 -77.433 50.658
139843 95.524 -77.433 50.626
139844 95.556 -77.433 50.593
139845 95.582 -77.431 50.558
139846 95.598 -77.422 50.522
139847 95.635 -77.429 50.488
139848 95.679 -77.447 50.454
139849 95.678 -77.461 50.417
139850 95.677 -77.476 50.380
139851 95.676 -77.490 50.340
139852 95.676 -77.505 50.299
139853 95.675 -77.520 50.256
139854 95.674 -77.534 50.213
139855 95.673 -77.548 50.172
139856 95.673 -77.563 50.131
139857 95.672 -77.577 50.085
139858 95.671 -77.592 50.041
139859 95.672 -77.604 49.998
139860 95.678 -77.613 49.958
139861 95.695 -77.617 49.918
139862 95.710 -77.615 49.880
139863 95.726 -77.614 49.844
139864 95.741 -77.611 49.807
139865 95.756 -77.609 49.773
139866 95.770 -77.608 49.739
139867 95.784 -77.606 49.706
139868 95.798 -77.605 49.672
139869 95.812 -77.603 49.639
139870 95.826 -77.602 49.605
139871 95.841 -77.599 49.572
139872 95.855 -77.598 49.540
139873 95.869 -77.596 49.509
139874 95.883 -77.595 49.479
139875 95.897 -77.593 49.447
139876 95.912 -77.592 49.415
139877 95.926 -77.589 49.384
139878 95.940 -77.588 49.354
139879 95.954 -77.586 49.323
139880 95.970 -77.598 49.295
139881 95.972 -77.615 49.266
139882 95.972 -77.634 49.233
139883 95.975 -77.650 49.197
139884 95.987 -77.657 49.162
139885 96.001 -77.663 49.127
139886 96.013 -77.670 49.092
139887 96.026 -77.676 49.058
139888 96.036 -77.686 49.024
139889 96.046 -77.696 48.990
139890 96.057 -77.705 48.957
139891 96.067 -77.716 48.923
139892 96.078 -77.726 48.888
139893 96.088 -77.736 48.853
139894 96.091 -77.744 48.819
139895 96.093 -77.753 48.785
139896 96.096 -77.761 48.751
139897 96.110 -77.771 48.716
139898 96.123 -77.780 48.681
139899 96.136 -77.790 48.646
139900 96.150 -77.799 48.610
139901 96.163 -77.808 48.578
139902 96.176 -77.818 48.546
139903 96.189 -77.827 48.515
139904 96.203 -77.837 48.483
139905 96.215 -77.845 48.451
139906 96.220 -77.840 48.420
139907 96.228 -77.837 48.390
139908 96.236 -77.834 48.363
139909 96.244 -77.830 48.336
139910 96.252 -77.827 48.310
139911 96.260 -77.824 48.284
139912 96.269 -77.821 48.257
139913 96.290 -77.816 48.232
139914 96.311 -77.812 48.207
139915 96.320 -77.815 48.183
139916 96.327 -77.821 48.159
139917 96.333 -77.826 48.136
139918 96.339 -77.831 48.113
139919 96.346 -77.837 48.090
139920 96.352 -77.842 48.066
139921 96.364 -77.851 48.041
139922 96.382 -77.863 48.019
139923 96.400 -77.875 47.996
139924 96.418 -77.887 47.973
139925 96.442 -77.897 47.950
139926 96.471 -77.904 47.926
139927 96.499 -77.912 47.902
139928 96.519 -77.912 47.877
139929 96.539 -77.913 47.852
139930 96.559 -77.913 47.828
139931 96.572 -77.896 47.805
139932 96.589 -77.898 47.781
139933 96.607 -77.905 47.757
139934 96.626 -77.912 47.731
139935 96.645 -77.919 47.705
139936 96.663 -77.927 47.681
139937 96.677 -77.931 47.657
139938 96.683 -77.929 47.632
139939 96.689 -77.927 47.608
139940 96.694 -77.924 47.583
139941 96.700 -77.922 47.558
139942 96.706 -77.920 47.535
139943 96.712 -77.918 47.512
139944 96.724 -77.918 47.487
139945 96.777 -77.935 47.464
139946 96.830 -77.953 47.440
139947 96.881 -77.970 47.415
139948 96.897 -77.974 47.391
139949 96.911 -77.990 47.367
139950 96.925 -78.006 47.342
139951 96.928 -78.020 47.319
139952 96.929 -78.034 47.296
139953 96.929 -78.049 47.273
139954 96.933 -78.062 47.252
139955 96.953 -78.066 47.230
139956 96.961 -78.059 47.208
139957 96.969 -78.051 47.187
139958 96.979 -78.047 47.166
139959 97.001 -78.051 47.142
139960 97.021 -78.056 47.121
139961 97.042 -78.061 47.101
139962 97.063 -78.066 47.082
139963 97.083 -78.072 47.062
139964 97.100 -78.083 47.044
139965 97.117 -78.095 47.027
139966 97.134 -78.107 47.012
139967 97.152 -78.114 46.997
139968 97.180 -78.095 46.983
139969 97.208 -78.076 46.970
139970 97.217 -78.072 46.959
139971 97.226 -78.067 46.949
139972 97.235 -78.062 46.939
139973 97.243 -78.058 46.930
139974 97.259 -78.057 46.921
139975 97.292 -78.063 46.914
139976 97.325 -78.069 46.906
139977 97.358 -78.075 46.893
139978 97.391 -78.082 46.880
139979 97.424 -78.089 46.866
139980 97.457 -78.097 46.851
139981 97.490 -78.104 46.837
139982 97.523 -78.111 46.823
139983 97.556 -78.119 46.810
139984 97.595 -78.121 46.796
139985 97.635 -78.123 46.781
139986 97.666 -78.138 46.764
139987 97.690 -78.162 46.748
139988 97.714 -78.186 46.730
139989 97.734 -78.202 46.712
139990 97.749 -78.205 46.693
139991 97.764 -78.208 46.672
139992 97.779 -78.212 46.650
139993 97.793 -78.215 46.628
139994 97.808 -78.218 46.608
139995 97.823 -78.222 46.586
139996 97.838 -78.225 46.566
139997 97.865 -78.231 46.545
139998 97.909 -78.241 46.522
139999 97.949 -78.247 46.500
140000 97.988 -78.254 46.478
140001 98.028 -78.262 46.456
140002 98.067 -78.269 46.433
140003 98.106 -78.276 46.411
140004 98.146 -78.283 46.389
140005 98.185 -78.290 46.368
140006 98.223 -78.302 46.345
140007 98.239 -78.360 46.322
140008 98.251 -78.411 46.299
140009 98.237 -78.411 46.277
140010 98.223 -78.411 46.256
140011 98.210 -78.397 46.234
140012 98.212 -78.373 46.212
140013 98.322 -78.387 46.192
140014 98.334 -78.375 46.172
140015 98.329 -78.359 46.151
140016 98.337 -78.354 46.130
140017 98.331 -78.353 46.109
140018 98.323 -78.352 46.088
140019 98.319 -78.333 46.068
140020 98.329 -78.302 46.046
140021 98.401 -78.278 46.026
140022 98.472 -78.253 46.004
140023 98.491 -78.250 45.981
140024 98.505 -78.250 45.959
140025 98.519 -78.250 45.936
140026 98.533 -78.249 45.912
140027 98.547 -78.248 45.888
140028 98.561 -78.247 45.865
140029 98.571 -78.232 45.841
140030 98.581 -78.215 45.816
140031 98.591 -78.197 45.794
140032 98.590 -78.184 45.771
140033 98.602 -78.186 45.749
140034 98.631 -78.194 45.729
140035 98.670 -78.191 45.709
140036 98.711 -78.186 45.689
140037 98.744 -78.158 45.670
140038 98.756 -78.196 45.653
140039 98.768 -78.233 45.638
140040 98.791 -78.254 45.623
140041 98.818 -78.271 45.609
140042 98.848 -78.284 45.593
140043 98.886 -78.291 45.577
140044 98.923 -78.298 45.559
140045 98.957 -78.309 45.543
140046 98.992 -78.319 45.527
140047 99.027 -78.329 45.511
140048 99.062 -78.339 45.496
140049 99.097 -78.350 45.482
140050 99.127 -78.366 45.468
140051 99.147 -78.377 45.454
140052 99.166 -78.377 45.440
140053 99.185 -78.377 45.428
140054 99.202 -78.379 45.417
140055 99.217 -78.382 45.405
140056 99.232 -78.385 45.395
140057 99.248 -78.388 45.384
140058 99.286 -78.397 45.372
140059 99.317 -78.398 45.358
140060 99.348 -78.399 45.345
140061 99.379 -78.401 45.331
140062 99.410 -78.402 45.318
140063 99.441 -78.403 45.305
140064 99.472 -78.404 45.291
140065 99.503 -78.406 45.277
140066 99.534 -78.407 45.263
140067 99.564 -78.408 45.252
140068 99.596 -78.423 45.240
140069 99.629 -78.445 45.228
140070 99.670 -78.478 45.217
140071 99.713 -78.514 45.203
140072 99.756 -78.550 45.190
140073 99.798 -78.586 45.176
140074 99.837 -78.617 45.163
140075 99.847 -78.624 45.149
140076 99.857 -78.631 45.134
140077 99.867 -78.638 45.121
140078 99.877 -78.645 45.108
140079 99.888 -78.652 45.095
140080 99.898 -78.659 45.082
140081 99.907 -78.667 45.070
140082 99.923 -78.677 45.058
140083 99.936 -78.689 45.046
140084 99.949 -78.701 45.034
140085 99.965 -78.716 45.023
140086 99.981 -78.730 45.011
140087 99.997 -78.745 45.000
140088 100.013 -78.760 44.990
140089 100.028 -78.774 44.980
140090 100.045 -78.789 44.969
140091 100.054 -78.808 44.959
140092 100.064 -78.799 44.949
140093 100.090 -78.811 44.939
140094 100.132 -78.825 44.928
140095 100.177 -78.840 44.917
140096 100.223 -78.855 44.905
140097 100.268 -78.871 44.893
140098 100.305 -78.885 44.881
140099 100.316 -78.896 44.869
140100 100.326 -78.906 44.857
140101 100.335 -78.917 44.845
140102 100.346 -78.933 44.832
140103 100.356 -78.952 44.820
140104 100.367 -78.971 44.807
140105 100.378 -78.990 44.794
140106 100.392 -79.005 44.779
140107 100.408 -79.016 44.765
140108 100.439 -79.034 44.751
140109 100.472 -79.053 44.736
140110 100.504 -79.072 44.722
140111 100.537 -79.092 44.707
140112 100.569 -79.111 44.692
140113 100.602 -79.130 44.677
140114 100.634 -79.149 44.661
140366 92.899 -78.178 54.252
140367 92.912 -78.174 54.245
140368 92.926 -78.169 54.239
140369 92.940 -78.164 54.234
140370 92.953 -78.159 54.229
140371 92.967 -78.154 54.224
140372 92.981 -78.150 54.219
140373 92.994 -78.145 54.214
140374 93.008 -78.141 54.208
140375 93.021 -78.135 54.202
140376 93.035 -78.131 54.196
140377 93.049 -78.126 54.190
140378 93.063 -78.121 54.186
140379 93.076 -78.116 54.181
140380 93.090 -78.112 54.175
140381 93.104 -78.099 54.167
140382 93.118 -78.087 54.161
140383 93.140 -78.075 54.155
140384 93.163 -78.063 54.150
140385 93.187 -78.051 54.144
140386 93.211 -78.040 54.139
140387 93.234 -78.028 54.133
140388 93.258 -78.015 54.125
140389 93.282 -78.004 54.119
140390 93.305 -77.991 54.113
140391 93.326 -77.973 54.108
140392 93.347 -77.955 54.101
140393 93.368 -77.937 54.095
140394 93.384 -77.925 54.087
140395 93.390 -77.930 54.077
140396 93.396 -77.934 54.067
140397 93.401 -77.938 54.056
140398 93.409 -77.928 54.044
140399 93.417 -77.916 54.033
140400 93.441 -77.910 54.019
140401 93.455 -77.896 54.005
140402 93.463 -77.881 53.990
140403 93.471 -77.864 53.974
140404 93.478 -77.848 53.959
140405 93.486 -77.831 53.945
140406 93.493 -77.815 53.930
140407 93.501 -77.799 53.916
140408 93.509 -77.784 53.900
140409 93.516 -77.766 53.882
140410 93.523 -77.745 53.864
140411 93.531 -77.726 53.845
140412 93.543 -77.714 53.826
140413 93.563 -77.721 53.807
140414 93.583 -77.727 53.789
140415 93.600 -77.724 53.769
140416 93.615 -77.714 53.749
140417 93.630 -77.705 53.729
140418 93.639 -77.701 53.710
140419 93.647 -77.697 53.692
140420 93.658 -77.692 53.674
140421 93.669 -77.687 53.655
140422 93.680 -77.683 53.637
140423 93.691 -77.678 53.618
140424 93.701 -77.673 53.600
140425 93.712 -77.668 53.582
140426 93.723 -77.663 53.564
140427 93.733 -77.658 53.546
140428 93.744 -77.653 53.528
140429 93.755 -77.649 53.511
140430 93.766 -77.644 53.493
140431 93.776 -77.639 53.475
140432 93.787 -77.634 53.457
140433 93.798 -77.629 53.440
140434 93.808 -77.624 53.423
140435 93.820 -77.620 53.406
140436 93.832 -77.615 53.388
140437 93.846 -77.611 53.371
140438 93.860 -77.607 53.354
140439 93.874 -77.603 53.337
140440 93.888 -77.599 53.319
140441 93.902 -77.595 53.302
140442 93.916 -77.591 53.285
140443 93.929 -77.587 53.267
140444 93.943 -77.583 53.250
140445 93.953 -77.581 53.233
140446 93.958 -77.583 53.215
140447 93.964 -77.583 53.198
140448 93.979 -77.570 53.181
140449 93.994 -77.557 53.163
140450 94.009 -77.544 53.145
140451 94.024 -77.531 53.127
140452 94.038 -77.517 53.109
140453 94.048 -77.511 53.091
140454 94.058 -77.503 53.074
140455 94.065 -77.495 53.057
140456 94.069 -77.486 53.040
140457 94.074 -77.478 53.022
140458 94.084 -77.472 53.006
140459 94.106 -77.470 52.991
140460 94.127 -77.468 52.976
140461 94.151 -77.467 52.962
140462 94.183 -77.467 52.948
140463 94.216 -77.467 52.935
140464 94.248 -77.468 52.921
140465 94.280 -77.468 52.907
140466 94.313 -77.469 52.892
140467 94.345 -77.469 52.875
140468 94.339 -77.474 52.857
140469 94.228 -77.451 52.840
140470 94.215 -77.469 52.827
140471 94.208 -77.490 52.817
140472 94.200 -77.511 52.810
140473 94.190 -77.529 52.801
140474 94.191 -77.507 52.789
140475 94.182 -77.495 52.776
140476 94.174 -77.479 52.766
140477 94.168 -77.458 52.756
140478 94.210 -77.473 52.748
140479 94.267 -77.500 52.745
140480 94.314 -77.486 52.742
140481 94.361 -77.471 52.741
140482 94.408 -77.457 52.742
140483 94.440 -77.445 52.741
140484 94.433 -77.435 52.739
140485 94.426 -77.425 52.737
140486 94.418 -77.415 52.733
140487 94.411 -77.406 52.725
140488 94.406 -77.391 52.716
140489 94.402 -77.374 52.706
140490 94.398 -77.357 52.695
140491 94.394 -77.339 52.680
140492 94.404 -77.335 52.660
140493 94.433 -77.345 52.634
140494 94.462 -77.354 52.607
140495 94.490 -77.364 52.581
140496 94.519 -77.374 52.555
140497 94.547 -77.389 52.527
140498 94.573 -77.413 52.496
140499 94.590 -77.408 52.463
140500 94.607 -77.402 52.428
140501 94.647 -77.448 52.393
140502 94.690 -77.497 52.358
140503 94.736 -77.526 52.324
140504 94.784 -77.545 52.288
140505 94.821 -77.555 52.251
140506 94.854 -77.553 52.215
140507 94.889 -77.549 52.181
140508 94.923 -77.545 52.146
140509 94.958 -77.542 52.113
140510 94.992 -77.538 52.081
140511 95.012 -77.530 52.047
140512 95.031 -77.521 52.013
140513 95.037 -77.501 51.977
140514 95.021 -77.460 51.938
140515 95.004 -77.417 51.899
140516 94.977 -77.385 51.859
140517 94.959 -77.354 51.821
140518 94.960 -77.316 51.784
140519 94.960 -77.278 51.745
140520 94.965 -77.250 51.706
140521 94.996 -77.269 51.667
140522 95.044 -77.322 51.628
140523 95.045 -77.326 51.591
140524 95.045 -77.329 51.555
140525 95.046 -77.334 51.518
140526 95.046 -77.338 51.472
140527 95.061 -77.356 51.422
140528 95.077 -77.376 51.369
140529 95.119 -77.389 51.310
140530 95.168 -77.400 51.252
140531 95.207 -77.405 51.195
140532 95.235 -77.407 51.141
140533 95.264 -77.409 51.090
140534 95.295 -77.413 51.040
140535 95.325 -77.417 50.992
140536 95.348 -77.418 50.946
140537 95.371 -77.419 50.900
140538 95.392 -77.420 50.855
140539 95.414 -77.420 50.813
140540 95.436 -77.420 50.774
140541 95.458 -77.421 50.739
140542 95.480 -77.422 50.705
140543 95.501 -77.422 50.670
140544 95.531 -77.423 50.636
140545 95.563 -77.423 50.603
140546 95.591 -77.420 50.569
140547 95.607 -77.412 50.534
140548 95.660 -77.430 50.501
140549 95.660 -77.445 50.468
140550 95.659 -77.459 50.434
140551 95.658 -77.474 50.396
140552 95.657 -77.489 50.357
140553 95.657 -77.503 50.316
140554 95.656 -77.517 50.275
140555 95.655 -77.532 50.236
140556 95.655 -77.545 50.197
140557 95.658 -77.553 50.156
140558 95.662 -77.561 50.111
140559 95.665 -77.569 50.065
140560 95.668 -77.577 50.022
140561 95.677 -77.588 49.980
140562 95.700 -77.608 49.938
140563 95.719 -77.615 49.899
140564 95.734 -77.613 49.861
140565 95.749 -77.611 49.824
140566 95.764 -77.608 49.788
140567 95.779 -77.606 49.754
140568 95.795 -77.604 49.720
140569 95.810 -77.602 49.685
140570 95.825 -77.600 49.650
140571 95.840 -77.598 49.616
140572 95.855 -77.595 49.581
140573 95.870 -77.593 49.547
140574 95.885 -77.591 49.515
140575 95.901 -77.589 49.484
140576 95.916 -77.587 49.451
140577 95.931 -77.585 49.420
140578 95.946 -77.583 49.391
140579 95.961 -77.581 49.362
140580 95.975 -77.581 49.332
140581 95.975 -77.600 49.304
140582 95.974 -77.619 49.273
140583 95.974 -77.638 49.240
140584 95.979 -77.652 49.204
140585 95.992 -77.658 49.169
140586 96.005 -77.664 49.133
140587 96.018 -77.671 49.098
140588 96.030 -77.677 49.066
140589 96.043 -77.683 49.035
140590 96.054 -77.692 49.002
140591 96.064 -77.702 48.969
140592 96.075 -77.711 48.936
140593 96.086 -77.721 48.902
140594 96.088 -77.730 48.866
140595 96.090 -77.739 48.833
140596 96.092 -77.747 48.800
140597 96.097 -77.756 48.766
140598 96.110 -77.765 48.730
140599 96.123 -77.774 48.695
140600 96.136 -77.784 48.661
140601 96.150 -77.793 48.626
140602 96.163 -77.802 48.593
140603 96.177 -77.812 48.561
140604 96.190 -77.821 48.529
140605 96.202 -77.830 48.497
140606 96.219 -77.837 48.466
140607 96.230 -77.838 48.435
140608 96.238 -77.835 48.404
140609 96.246 -77.831 48.377
140610 96.254 -77.828 48.350
140611 96.262 -77.825 48.324
140612 96.270 -77.822 48.298
140613 96.282 -77.818 48.271
140614 96.303 -77.815 48.246
140615 96.318 -77.814 48.221
140616 96.325 -77.819 48.197
140617 96.331 -77.824 48.173
140618 96.337 -77.830 48.149
140619 96.344 -77.834 48.126
140620 96.350 -77.840 48.102
140621 96.357 -77.845 48.078
140622 96.367 -77.853 48.055
140623 96.385 -77.865 48.032
140624 96.403 -77.877 48.008
140625 96.421 -77.890 47.984
140626 96.450 -77.897 47.961
140627 96.478 -77.904 47.936
140628 96.500 -77.906 47.912
140629 96.520 -77.907 47.887
140630 96.540 -77.908 47.863
140631 96.560 -77.909 47.838
140632 96.573 -77.893 47.814
140633 96.587 -77.882 47.789
140634 96.606 -77.890 47.765
140635 96.624 -77.897 47.740
140636 96.643 -77.904 47.715
140637 96.661 -77.912 47.691
140638 96.680 -77.918 47.668
140639 96.698 -77.926 47.644
140640 96.715 -77.931 47.620
140641 96.721 -77.930 47.596
140642 96.726 -77.928 47.572
140643 96.732 -77.925 47.549
140644 96.738 -77.923 47.525
140645 96.743 -77.921 47.501
140646 96.793 -77.937 47.477
140647 96.846 -77.954 47.454
140648 96.874 -77.962 47.429
140649 96.891 -77.966 47.405
140650 96.904 -77.982 47.381
140651 96.919 -77.998 47.357
140652 96.932 -78.014 47.333
140653 96.938 -78.029 47.311
140654 96.939 -78.044 47.288
140655 96.944 -78.056 47.266
140656 96.958 -78.055 47.245
140657 96.966 -78.047 47.224
140658 96.973 -78.039 47.202
140659 96.984 -78.034 47.180
140660 97.005 -78.039 47.158
140661 97.025 -78.044 47.136
140662 97.046 -78.049 47.116
140663 97.067 -78.054 47.096
140664 97.088 -78.059 47.076
140665 97.108 -78.065 47.058
140666 97.124 -78.078 47.041
140667 97.141 -78.090 47.024
140668 97.163 -78.087 47.010
140669 97.190 -78.069 46.996
140670 97.202 -78.062 46.983
140671 97.211 -78.058 46.971
140672 97.220 -78.053 46.960
140673 97.229 -78.050 46.950
140674 97.238 -78.045 46.940
140675 97.246 -78.040 46.931
140676 97.277 -78.045 46.924
140677 97.309 -78.051 46.918
140678 97.343 -78.059 46.906
140679 97.376 -78.065 46.894
140680 97.408 -78.073 46.881
140681 97.442 -78.080 46.866
140682 97.475 -78.087 46.853
140683 97.508 -78.095 46.841
140684 97.541 -78.102 46.830
140685 97.581 -78.104 46.815
140686 97.617 -78.111 46.800
140687 97.641 -78.135 46.783
140688 97.665 -78.160 46.766
140689 97.689 -78.184 46.748
140690 97.710 -78.200 46.729
140691 97.724 -78.203 46.709
140692 97.739 -78.207 46.688
140693 97.754 -78.210 46.665
140694 97.769 -78.213 46.644
140695 97.784 -78.217 46.624
140696 97.798 -78.220 46.603
140697 97.816 -78.224 46.582
140698 97.862 -78.235 46.560
140699 97.909 -78.247 46.536
140700 97.951 -78.254 46.514
140701 97.992 -78.262 46.492
140702 98.032 -78.269 46.469
140703 98.071 -78.275 46.447
140704 98.110 -78.283 46.425
140705 98.150 -78.290 46.403
140706 98.190 -78.297 46.382
140707 98.229 -78.304 46.359
140708 98.246 -78.362 46.336
140709 98.251 -78.399 46.313
140710 98.237 -78.399 46.291
140711 98.223 -78.393 46.268
140712 98.211 -78.364 46.246
140713 98.199 -78.335 46.223
140714 98.309 -78.350 46.203
140715 98.419 -78.364 46.183
140716 98.413 -78.350 46.162
140717 98.324 -78.343 46.141
140718 98.302 -78.341 46.120
140719 98.296 -78.339 46.099
140720 98.292 -78.318 46.079
140721 98.353 -78.293 46.058
140722 98.424 -78.269 46.036
140723 98.494 -78.244 46.014
140724 98.516 -78.241 45.992
140725 98.529 -78.241 45.970
140726 98.544 -78.240 45.947
140727 98.558 -78.239 45.923
140728 98.572 -78.238 45.900
140729 98.583 -78.225 45.876
140730 98.593 -78.209 45.852
140731 98.603 -78.191 45.828
140732 98.613 -78.175 45.806
140733 98.621 -78.158 45.783
140734 98.609 -78.153 45.761
140735 98.638 -78.162 45.740
140736 98.678 -78.156 45.720
140737 98.719 -78.151 45.700
140738 98.748 -78.176 45.681
140739 98.762 -78.210 45.664
140740 98.774 -78.247 45.648
140741 98.788 -78.280 45.633
140742 98.822 -78.291 45.620
140743 98.859 -78.297 45.606
140744 98.897 -78.304 45.591
140745 98.933 -78.313 45.574
140746 98.968 -78.324 45.557
140747 99.003 -78.334 45.541
140748 99.038 -78.344 45.525
140749 99.072 -78.354 45.510
140750 99.107 -78.365 45.496
140751 99.133 -78.365 45.482
140752 99.149 -78.367 45.468
140753 99.164 -78.370 45.453
140754 99.179 -78.373 45.440
140755 99.194 -78.376 45.428
140756 99.208 -78.379 45.417
140757 99.223 -78.382 45.407
140758 99.254 -78.389 45.396
140759 99.287 -78.397 45.384
140760 99.323 -78.406 45.371
140761 99.371 -78.415 45.357
140762 99.402 -78.416 45.344
140763 99.433 -78.418 45.331
140764 99.464 -78.419 45.317
140765 99.495 -78.420 45.304
140766 99.526 -78.422 45.291
140767 99.557 -78.423 45.276
140768 99.588 -78.424 45.265
140769 99.619 -78.432 45.254
140770 99.652 -78.454 45.242
140771 99.684 -78.476 45.231
140772 99.725 -78.510 45.217
140773 99.768 -78.545 45.203
140774 99.811 -78.581 45.190
140775 99.844 -78.608 45.177
140776 99.854 -78.615 45.163
140777 99.864 -78.622 45.149
140778 99.874 -78.629 45.137
140779 99.885 -78.636 45.124
140780 99.894 -78.644 45.112
140781 99.904 -78.653 45.100
140782 99.911 -78.660 45.088
140783 99.924 -78.667 45.076
140784 99.951 -78.677 45.064
140785 99.965 -78.689 45.051
140786 99.978 -78.702 45.039
140787 99.991 -78.714 45.027
140788 100.004 -78.726 45.015
140789 100.017 -78.738 45.004
140790 100.031 -78.750 44.993
140791 100.044 -78.762 44.983
140792 100.057 -78.775 44.972
140793 100.084 -78.789 44.963
140794 100.129 -78.804 44.952
140795 100.174 -78.819 44.940
140796 100.220 -78.834 44.928
140797 100.265 -78.849 44.916
140798 100.305 -78.863 44.903
140799 100.315 -78.874 44.890
140800 100.325 -78.884 44.878
140801 100.335 -78.895 44.865
140802 100.345 -78.905 44.852
140803 100.355 -78.925 44.839
140804 100.366 -78.944 44.826
140805 100.376 -78.964 44.813
140806 100.387 -78.983 44.798
140807 100.399 -78.999 44.784
140808 100.415 -79.011 44.770
140809 100.436 -79.024 44.755
140810 100.469 -79.043 44.740
140811 100.501 -79.062 44.725
140812 100.534 -79.081 44.710
140813 100.566 -79.101 44.695
140814 100.599 -79.120 44.680
140815 100.631 -79.139 44.664
141067 92.892 -78.166 54.254
141068 92.908 -78.160 54.247
141069 92.925 -78.155 54.241
141070 92.939 -78.150 54.236
141071 92.952 -78.145 54.233
141072 92.966 -78.141 54.227
141073 92.980 -78.136 54.222
141074 92.994 -78.131 54.218
141075 93.007 -78.126 54.211
141076 93.021 -78.122 54.206
141077 93.035 -78.117 54.201
141078 93.048 -78.112 54.195
141079 93.062 -78.107 54.190
141080 93.075 -78.103 54.186
141081 93.089 -78.097 54.180
141082 93.103 -78.089 54.173
141083 93.117 -78.077 54.167
141084 93.139 -78.065 54.161
141085 93.163 -78.053 54.156
141086 93.187 -78.041 54.151
141087 93.210 -78.029 54.146
141088 93.234 -78.017 54.141
141089 93.258 -78.006 54.134
141090 93.279 -77.987 54.128
141091 93.299 -77.969 54.122
141092 93.321 -77.951 54.116
141093 93.342 -77.933 54.111
141094 93.362 -77.915 54.104
141095 93.376 -77.907 54.096
141096 93.381 -77.912 54.086
141097 93.387 -77.915 54.075
141098 93.393 -77.920 54.064
141099 93.399 -77.924 54.053
141100 93.405 -77.921 54.042
141101 93.443 -77.918 54.030
141102 93.450 -77.902 54.016
141103 93.458 -77.887 54.001
141104 93.462 -77.869 53.986
141105 93.465 -77.851 53.973
141106 93.468 -77.834 53.959
141107 93.478 -77.814 53.946
141108 93.488 -77.793 53.931
141109 93.498 -77.774 53.915
141110 93.512 -77.764 53.898
141111 93.522 -77.750 53.880
141112 93.530 -77.731 53.862
141113 93.538 -77.711 53.843
141114 93.555 -77.712 53.825
141115 93.575 -77.718 53.806
141116 93.591 -77.713 53.786
141117 93.607 -77.703 53.766
141118 93.621 -77.693 53.745
141119 93.634 -77.686 53.726
141120 93.642 -77.683 53.708
141121 93.651 -77.679 53.690
141122 93.661 -77.674 53.672
141123 93.672 -77.669 53.653
141124 93.683 -77.664 53.635
141125 93.694 -77.659 53.616
141126 93.704 -77.655 53.598
141127 93.715 -77.649 53.581
141128 93.726 -77.645 53.564
141129 93.736 -77.640 53.546
141130 93.748 -77.635 53.528
141131 93.758 -77.630 53.511
141132 93.769 -77.625 53.493
141133 93.780 -77.620 53.475
141134 93.790 -77.616 53.458
141135 93.801 -77.611 53.441
141136 93.813 -77.606 53.425
141137 93.827 -77.602 53.407
141138 93.841 -77.599 53.390
141139 93.855 -77.594 53.373
141140 93.868 -77.590 53.355
141141 93.883 -77.586 53.338
141142 93.896 -77.582 53.322
141143 93.910 -77.578 53.306
141144 93.924 -77.574 53.288
141145 93.938 -77.570 53.272
141146 93.949 -77.567 53.257
141147 93.954 -77.569 53.239
141148 93.966 -77.561 53.222
141149 93.981 -77.548 53.205
141150 93.996 -77.535 53.188
141151 94.009 -77.520 53.170
141152 94.020 -77.502 53.151
141153 94.031 -77.495 53.134
141154 94.042 -77.489 53.117
141155 94.053 -77.483 53.100
141156 94.058 -77.474 53.082
141157 94.070 -77.469 53.065
141158 94.092 -77.467 53.047
141159 94.113 -77.465 53.030
141160 94.134 -77.464 53.014
141161 94.156 -77.462 53.000
141162 94.189 -77.462 52.986
141163 94.221 -77.463 52.973
141164 94.253 -77.463 52.961
141165 94.286 -77.464 52.948
141166 94.318 -77.464 52.934
141167 94.350 -77.464 52.919
141168 94.353 -77.457 52.904
141169 94.323 -77.434 52.887
141170 94.227 -77.407 52.871
141171 94.220 -77.428 52.856
141172 94.211 -77.452 52.846
141173 94.202 -77.482 52.839
141174 94.196 -77.512 52.830
141175 94.193 -77.504 52.820
141176 94.187 -77.484 52.807
141177 94.182 -77.463 52.797
141178 94.179 -77.444 52.788
141179 94.216 -77.448 52.778
141180 94.261 -77.462 52.770
141181 94.316 -77.479 52.764
141182 94.362 -77.464 52.759
141183 94.409 -77.451 52.758
141184 94.424 -77.439 52.757
141185 94.417 -77.429 52.756
141186 94.410 -77.420 52.755
141187 94.403 -77.408 52.752
141188 94.399 -77.391 52.747
141189 94.395 -77.373 52.740
141190 94.391 -77.357 52.730
141191 94.387 -77.339 52.719
141192 94.383 -77.322 52.705
141193 94.393 -77.316 52.686
141194 94.422 -77.326 52.663
141195 94.451 -77.336 52.638
141196 94.479 -77.346 52.612
141197 94.507 -77.356 52.585
141198 94.536 -77.366 52.554
141199 94.565 -77.376 52.521
141200 94.593 -77.390 52.487
141201 94.615 -77.397 52.451
141202 94.657 -77.446 52.414
141203 94.699 -77.495 52.377
141204 94.745 -77.528 52.341
141205 94.779 -77.528 52.308
141206 94.814 -77.524 52.278
141207 94.848 -77.520 52.247
141208 94.882 -77.517 52.215
141209 94.917 -77.513 52.180
141210 94.951 -77.509 52.148
141211 94.980 -77.504 52.117
141212 94.999 -77.495 52.084
141213 95.018 -77.487 52.050
141214 95.037 -77.479 52.014
141215 95.024 -77.441 51.975
141216 95.007 -77.399 51.935
141217 94.981 -77.366 51.896
141218 94.969 -77.332 51.858
141219 94.970 -77.294 51.822
141220 94.970 -77.257 51.781
141221 94.987 -77.254 51.741
141222 95.004 -77.253 51.702
141223 95.030 -77.264 51.663
141224 95.067 -77.304 51.624
141225 95.067 -77.309 51.586
141226 95.068 -77.313 51.547
141227 95.073 -77.321 51.498
141228 95.090 -77.341 51.447
141229 95.106 -77.361 51.392
141230 95.129 -77.378 51.336
141231 95.166 -77.382 51.278
141232 95.193 -77.383 51.222
141233 95.220 -77.384 51.168
141234 95.248 -77.385 51.115
141235 95.276 -77.388 51.064
141236 95.307 -77.392 51.014
141237 95.338 -77.396 50.967
141238 95.369 -77.400 50.920
141239 95.399 -77.404 50.875
141240 95.429 -77.407 50.832
141241 95.454 -77.409 50.792
141242 95.476 -77.410 50.756
141243 95.498 -77.410 50.720
141244 95.520 -77.410 50.684
141245 95.542 -77.411 50.648
141246 95.571 -77.412 50.614
141247 95.600 -77.410 50.580
141248 95.631 -77.413 50.547
141249 95.641 -77.428 50.515
141250 95.640 -77.443 50.483
141251 95.639 -77.457 50.450
141252 95.639 -77.472 50.413
141253 95.638 -77.486 50.375
141254 95.641 -77.494 50.335
141255 95.644 -77.502 50.294
141256 95.648 -77.511 50.257
141257 95.651 -77.518 50.219
141258 95.654 -77.526 50.180
141259 95.658 -77.534 50.136
141260 95.661 -77.542 50.090
141261 95.664 -77.551 50.046
141262 95.675 -77.563 50.002
141263 95.699 -77.583 49.959
141264 95.723 -77.602 49.919
141265 95.742 -77.612 49.880
141266 95.757 -77.610 49.842
141267 95.772 -77.608 49.805
141268 95.788 -77.605 49.769
141269 95.803 -77.604 49.734
141270 95.818 -77.602 49.699
141271 95.833 -77.599 49.663
141272 95.848 -77.597 49.626
141273 95.864 -77.595 49.590
141274 95.879 -77.592 49.554
141275 95.894 -77.591 49.521
141276 95.909 -77.589 49.488
141277 95.924 -77.586 49.455
141278 95.939 -77.584 49.424
141279 95.954 -77.582 49.394
141280 95.969 -77.580 49.365
141281 95.977 -77.584 49.339
141282 95.977 -77.603 49.310
141283 95.976 -77.623 49.279
141284 95.976 -77.642 49.246
141285 95.984 -77.652 49.210
141286 95.996 -77.659 49.176
141287 96.009 -77.665 49.143
141288 96.022 -77.672 49.110
141289 96.035 -77.678 49.079
141290 96.047 -77.684 49.048
141291 96.060 -77.691 49.016
141292 96.073 -77.698 48.983
141293 96.083 -77.708 48.949
141294 96.086 -77.716 48.915
141295 96.088 -77.724 48.880
141296 96.089 -77.733 48.848
141297 96.091 -77.742 48.815
141298 96.097 -77.750 48.780
141299 96.110 -77.760 48.745
141300 96.124 -77.769 48.711
141301 96.137 -77.778 48.676
141302 96.150 -77.788 48.641
141303 96.164 -77.797 48.609
141304 96.176 -77.806 48.576
141305 96.188 -77.815 48.544
141306 96.203 -77.824 48.512
141307 96.222 -77.827 48.481
141308 96.236 -77.831 48.450
141309 96.249 -77.833 48.419
141310 96.257 -77.830 48.392
141311 96.265 -77.827 48.365
141312 96.273 -77.823 48.338
141313 96.281 -77.820 48.312
141314 96.295 -77.816 48.286
141315 96.315 -77.812 48.261
141316 96.323 -77.817 48.236
141317 96.329 -77.822 48.212
141318 96.335 -77.827 48.188
141319 96.342 -77.833 48.164
141320 96.348 -77.838 48.140
141321 96.355 -77.843 48.116
141322 96.361 -77.849 48.092
141323 96.370 -77.855 48.068
141324 96.388 -77.868 48.045
141325 96.406 -77.880 48.021
141326 96.429 -77.890 47.997
141327 96.457 -77.897 47.973
141328 96.480 -77.900 47.949
141329 96.501 -77.901 47.923
141330 96.521 -77.902 47.898
141331 96.541 -77.902 47.875
141332 96.561 -77.903 47.850
141333 96.575 -77.889 47.825
141334 96.587 -77.873 47.801
141335 96.604 -77.874 47.775
141336 96.623 -77.881 47.751
141337 96.642 -77.889 47.727
141338 96.660 -77.896 47.703
141339 96.678 -77.903 47.680
141340 96.697 -77.911 47.656
141341 96.716 -77.918 47.633
141342 96.734 -77.925 47.609
141343 96.752 -77.932 47.585
141344 96.758 -77.930 47.561
141345 96.763 -77.928 47.537
141346 96.769 -77.926 47.514
141347 96.810 -77.938 47.491
141348 96.851 -77.951 47.468
141349 96.868 -77.953 47.444
141350 96.884 -77.959 47.420
141351 96.898 -77.975 47.396
141352 96.912 -77.991 47.372
141353 96.926 -78.007 47.349
141354 96.940 -78.023 47.326
141355 96.948 -78.038 47.303
141356 96.955 -78.050 47.281
141357 96.963 -78.043 47.260
141358 96.970 -78.035 47.239
141359 96.978 -78.028 47.218
141360 96.989 -78.022 47.195
141361 97.009 -78.027 47.173
141362 97.030 -78.032 47.152
141363 97.050 -78.037 47.131
141364 97.071 -78.042 47.111
141365 97.092 -78.047 47.091
141366 97.113 -78.052 47.073
141367 97.132 -78.059 47.055
141368 97.149 -78.072 47.038
141369 97.173 -78.061 47.023
141370 97.186 -78.053 47.009
141371 97.195 -78.050 46.996
141372 97.205 -78.045 46.984
141373 97.214 -78.040 46.972
141374 97.223 -78.036 46.960
141375 97.232 -78.031 46.950
141376 97.241 -78.027 46.942
141377 97.261 -78.028 46.934
141378 97.294 -78.034 46.928
141379 97.327 -78.042 46.918
141380 97.361 -78.050 46.907
141381 97.393 -78.056 46.894
141382 97.427 -78.064 46.881
141383 97.460 -78.072 46.869
141384 97.493 -78.078 46.859
141385 97.527 -78.084 46.849
141386 97.568 -78.087 46.834
141387 97.593 -78.109 46.818
141388 97.616 -78.134 46.802
141389 97.641 -78.158 46.785
141390 97.664 -78.182 46.765
141391 97.685 -78.199 46.745
141392 97.700 -78.202 46.725
141393 97.715 -78.206 46.704
141394 97.730 -78.209 46.682
141395 97.744 -78.213 46.661
141396 97.759 -78.216 46.639
141397 97.774 -78.219 46.618
141398 97.813 -78.228 46.596
141399 97.860 -78.239 46.573
141400 97.906 -78.250 46.551
141401 97.952 -78.261 46.529
141402 97.994 -78.269 46.507
141403 98.035 -78.275 46.484
141404 98.075 -78.282 46.463
141405 98.115 -78.290 46.441
141406 98.154 -78.297 46.418
141407 98.194 -78.304 46.395
141408 98.233 -78.311 46.372
141409 98.252 -78.363 46.349
141410 98.250 -78.386 46.326
141411 98.237 -78.386 46.303
141412 98.224 -78.360 46.281
141413 98.211 -78.332 46.258
141414 98.199 -78.303 46.235
141415 98.296 -78.313 46.214
141416 98.406 -78.327 46.193
141417 98.501 -78.343 46.172
141418 98.426 -78.341 46.150
141419 98.340 -78.333 46.130
141420 98.268 -78.327 46.109
141421 98.304 -78.308 46.088
141422 98.375 -78.284 46.067
141423 98.446 -78.260 46.046
141424 98.517 -78.235 46.024
141425 98.541 -78.231 46.001
141426 98.555 -78.231 45.980
141427 98.569 -78.230 45.957
141428 98.583 -78.229 45.934
141429 98.594 -78.219 45.911
141430 98.604 -78.203 45.887
141431 98.615 -78.185 45.863
141432 98.625 -78.169 45.840
141433 98.635 -78.151 45.817
141434 98.645 -78.134 45.795
141435 98.629 -78.125 45.773
141436 98.646 -78.128 45.751
141437 98.687 -78.122 45.731
141438 98.725 -78.173 45.712
141439 98.742 -78.211 45.692
141440 98.756 -78.244 45.675
141441 98.773 -78.275 45.658
141442 98.798 -78.295 45.643
141443 98.833 -78.304 45.630
141444 98.871 -78.311 45.618
141445 98.908 -78.319 45.604
141446 98.943 -78.329 45.588
141447 98.978 -78.339 45.570
141448 99.013 -78.349 45.555
141449 99.048 -78.360 45.539
141450 99.077 -78.363 45.525
141451 99.101 -78.362 45.510
141452 99.126 -78.361 45.497
141453 99.141 -78.364 45.483
141454 99.155 -78.367 45.468
141455 99.170 -78.369 45.454
141456 99.185 -78.373 45.441
141457 99.200 -78.376 45.428
141458 99.228 -78.382 45.418
141459 99.260 -78.390 45.407
141460 99.293 -78.398 45.395
141461 99.325 -78.406 45.384
141462 99.358 -78.414 45.370
141463 99.411 -78.426 45.358
141464 99.456 -78.434 45.345
141465 99.488 -78.435 45.332
141466 99.518 -78.436 45.319
141467 99.549 -78.438 45.305
141468 99.580 -78.439 45.290
141469 99.611 -78.440 45.278
141470 99.642 -78.442 45.266
141471 99.675 -78.463 45.254
141472 99.707 -78.485 45.243
141473 99.739 -78.507 45.229
141474 99.780 -78.541 45.216
141475 99.823 -78.576 45.203
141476 99.851 -78.599 45.191
141477 99.861 -78.606 45.178
141478 99.871 -78.613 45.165
141479 99.881 -78.620 45.152
141480 99.891 -78.630 45.141
141481 99.900 -78.639 45.129
141482 99.907 -78.645 45.117
141483 99.914 -78.652 45.105
141484 99.921 -78.659 45.093
141485 99.948 -78.668 45.081
141486 99.980 -78.677 45.068
141487 99.993 -78.689 45.054
141488 100.006 -78.702 45.042
141489 100.019 -78.714 45.030
141490 100.032 -78.727 45.017
141491 100.045 -78.739 45.007
141492 100.059 -78.751 44.996
141493 100.077 -78.764 44.985
141494 100.125 -78.780 44.975
141495 100.171 -78.797 44.964
141496 100.217 -78.812 44.952
141497 100.262 -78.827 44.939
141498 100.304 -78.842 44.927
141499 100.314 -78.852 44.913
141500 100.324 -78.863 44.900
141501 100.334 -78.874 44.887
141502 100.344 -78.884 44.874
141503 100.354 -78.898 44.860
141504 100.365 -78.917 44.847
141505 100.375 -78.937 44.833
141506 100.386 -78.955 44.819
141507 100.396 -78.975 44.805
141508 100.407 -78.994 44.791
141509 100.423 -79.006 44.775
141510 100.439 -79.017 44.760
141511 100.466 -79.034 44.744
141512 100.499 -79.053 44.729
141513 100.531 -79.071 44.714
141514 100.564 -79.090 44.699
141515 100.596 -79.110 44.683
141516 100.629 -79.129 44.667
141767 92.862 -78.160 54.259
141768 92.879 -78.155 54.254
141769 92.896 -78.150 54.248
141770 92.912 -78.144 54.243
141771 92.928 -78.138 54.238
141772 92.945 -78.133 54.233
141773 92.961 -78.128 54.228
141774 92.978 -78.122 54.224
141775 92.993 -78.117 54.219
141776 93.006 -78.112 54.214
141777 93.020 -78.108 54.209
141778 93.034 -78.103 54.203
141779 93.048 -78.098 54.198
141780 93.061 -78.094 54.193
141781 93.075 -78.088 54.189
141782 93.088 -78.084 54.184
141783 93.102 -78.078 54.178
141784 93.116 -78.066 54.172
141785 93.139 -78.054 54.166
141786 93.162 -78.043 54.161
141787 93.186 -78.031 54.157
141788 93.210 -78.018 54.153
141789 93.231 -78.002 54.148
141790 93.252 -77.984 54.142
141791 93.273 -77.965 54.136
141792 93.294 -77.947 54.130
141793 93.315 -77.929 54.124
141794 93.336 -77.917 54.118
141795 93.355 -77.910 54.112
141796 93.358 -77.899 54.103
141797 93.353 -77.907 54.093
141798 93.358 -77.912 54.082
141799 93.363 -77.916 54.072
141800 93.369 -77.921 54.062
141801 93.376 -77.924 54.050
141802 93.414 -77.917 54.039
141803 93.417 -77.899 54.027
141804 93.425 -77.880 54.013
141805 93.435 -77.860 53.999
141806 93.446 -77.840 53.986
141807 93.457 -77.821 53.972
141808 93.467 -77.800 53.958
141809 93.477 -77.780 53.944
141810 93.488 -77.761 53.928
141811 93.501 -77.751 53.911
141812 93.515 -77.743 53.895
141813 93.529 -77.736 53.878
141814 93.537 -77.717 53.860
141815 93.548 -77.703 53.842
141816 93.568 -77.710 53.824
141817 93.583 -77.701 53.803
141818 93.597 -77.692 53.782
141819 93.613 -77.682 53.761
141820 93.628 -77.672 53.743
141821 93.637 -77.668 53.724
141822 93.644 -77.664 53.706
141823 93.654 -77.661 53.689
141824 93.664 -77.655 53.671
141825 93.675 -77.651 53.652
141826 93.686 -77.645 53.634
141827 93.697 -77.641 53.616
141828 93.707 -77.636 53.599
141829 93.718 -77.631 53.582
141830 93.729 -77.627 53.565
141831 93.740 -77.621 53.547
141832 93.751 -77.617 53.529
141833 93.761 -77.612 53.512
141834 93.772 -77.607 53.494
141835 93.783 -77.602 53.477
141836 93.793 -77.597 53.460
141837 93.807 -77.593 53.443
141838 93.822 -77.589 53.426
141839 93.835 -77.585 53.409
141840 93.849 -77.581 53.392
141841 93.863 -77.577 53.375
141842 93.877 -77.573 53.359
141843 93.891 -77.569 53.343
141844 93.905 -77.565 53.327
141845 93.918 -77.561 53.311
141846 93.932 -77.557 53.297
141847 93.945 -77.554 53.283
141848 93.953 -77.551 53.268
141849 93.968 -77.538 53.253
141850 93.980 -77.522 53.237
141851 93.991 -77.504 53.220
141852 94.001 -77.486 53.202
141853 94.012 -77.479 53.183
141854 94.024 -77.473 53.164
141855 94.035 -77.467 53.146
141856 94.055 -77.466 53.127
141857 94.078 -77.464 53.109
141858 94.100 -77.463 53.091
141859 94.121 -77.461 53.073
141860 94.142 -77.459 53.056
141861 94.163 -77.457 53.040
141862 94.194 -77.457 53.026
141863 94.226 -77.458 53.012
141864 94.259 -77.458 53.000
141865 94.291 -77.459 52.989
141866 94.316 -77.457 52.976
141867 94.318 -77.450 52.963
141868 94.319 -77.442 52.949
141869 94.320 -77.435 52.935
141870 94.316 -77.425 52.920
141871 94.227 -77.385 52.904
141872 94.217 -77.415 52.888
141873 94.207 -77.445 52.875
141874 94.201 -77.475 52.867
141875 94.203 -77.496 52.858
141876 94.201 -77.489 52.849
141877 94.195 -77.467 52.836
141878 94.189 -77.447 52.828
141879 94.197 -77.434 52.819
141880 94.234 -77.439 52.809
141881 94.270 -77.442 52.798
141882 94.312 -77.452 52.788
141883 94.363 -77.457 52.781
141884 94.411 -77.444 52.776
141885 94.408 -77.433 52.773
141886 94.401 -77.424 52.770
141887 94.397 -77.407 52.768
141888 94.393 -77.390 52.767
141889 94.389 -77.373 52.764
141890 94.384 -77.356 52.760
141891 94.380 -77.338 52.752
141892 94.376 -77.322 52.742
141893 94.372 -77.304 52.728
141894 94.382 -77.298 52.710
141895 94.410 -77.308 52.688
141896 94.439 -77.318 52.663
141897 94.468 -77.328 52.637
141898 94.496 -77.338 52.609
141899 94.525 -77.348 52.577
141900 94.550 -77.359 52.543
141901 94.573 -77.371 52.507
141902 94.616 -77.376 52.470
141903 94.664 -77.413 52.432
141904 94.706 -77.460 52.396
141905 94.738 -77.500 52.363
141906 94.772 -77.496 52.336
141907 94.807 -77.492 52.309
141908 94.841 -77.488 52.280
141909 94.876 -77.484 52.246
141910 94.910 -77.480 52.212
141911 94.945 -77.476 52.180
141912 94.968 -77.470 52.149
141913 94.987 -77.461 52.119
141914 95.006 -77.453 52.086
141915 95.025 -77.445 52.050
141916 95.028 -77.422 52.012
141917 95.012 -77.380 51.972
141918 94.985 -77.348 51.933
141919 94.979 -77.311 51.895
141920 94.980 -77.273 51.857
141921 94.992 -77.261 51.817
141922 95.009 -77.259 51.778
141923 95.026 -77.257 51.738
141924 95.043 -77.255 51.697
141925 95.063 -77.258 51.658
141926 95.090 -77.288 51.618
141927 95.090 -77.291 51.578
141928 95.101 -77.306 51.528
141929 95.117 -77.326 51.474
141930 95.128 -77.348 51.419
141931 95.134 -77.366 51.362
141932 95.151 -77.360 51.306
141933 95.178 -77.360 51.250
141934 95.206 -77.362 51.194
141935 95.233 -77.363 51.140
141936 95.261 -77.365 51.088
141937 95.289 -77.367 51.037
141938 95.320 -77.371 50.986
141939 95.350 -77.375 50.939
141940 95.381 -77.379 50.895
141941 95.412 -77.382 50.851
141942 95.442 -77.386 50.809
141943 95.473 -77.390 50.771
141944 95.503 -77.395 50.734
141945 95.534 -77.398 50.698
141946 95.560 -77.400 50.662
141947 95.582 -77.401 50.626
141948 95.608 -77.400 50.593
141949 95.623 -77.412 50.562
141950 95.622 -77.426 50.531
141951 95.624 -77.436 50.499
141952 95.627 -77.443 50.466
141953 95.631 -77.451 50.431
141954 95.634 -77.460 50.395
141955 95.637 -77.467 50.356
141956 95.640 -77.476 50.314
141957 95.643 -77.483 50.275
141958 95.647 -77.492 50.239
141959 95.650 -77.500 50.202
141960 95.654 -77.508 50.160
141961 95.657 -77.516 50.115
141962 95.660 -77.524 50.070
141963 95.674 -77.539 50.025
141964 95.698 -77.558 49.981
141965 95.721 -77.577 49.940
141966 95.745 -77.597 49.901
141967 95.765 -77.609 49.861
141968 95.780 -77.607 49.823
141969 95.796 -77.605 49.785
141970 95.811 -77.603 49.748
141971 95.826 -77.601 49.713
141972 95.841 -77.599 49.675
141973 95.856 -77.596 49.637
141974 95.872 -77.594 49.599
141975 95.887 -77.592 49.561
141976 95.902 -77.590 49.526
141977 95.917 -77.588 49.491
141978 95.932 -77.586 49.458
141979 95.947 -77.583 49.427
141980 95.962 -77.581 49.396
141981 95.975 -77.581 49.366
141982 95.978 -77.589 49.342
141983 95.979 -77.602 49.314
141984 95.979 -77.624 49.284
141985 95.978 -77.645 49.252
141986 95.988 -77.654 49.217
141987 96.001 -77.661 49.183
141988 96.013 -77.668 49.152
141989 96.026 -77.676 49.121
141990 96.038 -77.683 49.090
141991 96.051 -77.690 49.059
141992 96.063 -77.698 49.027
141993 96.076 -77.705 48.995
141994 96.085 -77.702 48.962
141995 96.085 -77.711 48.928
141996 96.087 -77.719 48.894
141997 96.089 -77.727 48.862
141998 96.090 -77.736 48.829
141999 96.098 -77.745 48.795
142000 96.111 -77.754 48.760
142001 96.124 -77.764 48.726
142002 96.137 -77.773 48.691
142003 96.149 -77.782 48.657
142004 96.161 -77.791 48.624
142005 96.173 -77.800 48.591
142006 96.185 -77.809 48.559
142007 96.211 -77.814 48.528
142008 96.226 -77.818 48.497
142009 96.240 -77.821 48.466
142010 96.254 -77.826 48.434
142011 96.267 -77.827 48.407
142012 96.275 -77.824 48.380
142013 96.283 -77.821 48.353
142014 96.291 -77.818 48.326
142015 96.307 -77.814 48.302
142016 96.321 -77.815 48.277
142017 96.327 -77.820 48.251
142018 96.334 -77.825 48.228
142019 96.340 -77.830 48.204
142020 96.346 -77.836 48.179
142021 96.353 -77.841 48.155
142022 96.359 -77.846 48.131
142023 96.366 -77.852 48.106
142024 96.373 -77.858 48.082
142025 96.391 -77.870 48.059
142026 96.409 -77.882 48.035
142027 96.436 -77.890 48.010
142028 96.461 -77.894 47.985
142029 96.481 -77.895 47.961
142030 96.501 -77.896 47.936
142031 96.521 -77.896 47.911
142032 96.541 -77.897 47.888
142033 96.561 -77.898 47.863
142034 96.576 -77.886 47.838
142035 96.589 -77.869 47.813
142036 96.603 -77.859 47.788
142037 96.621 -77.866 47.763
142038 96.640 -77.874 47.739
142039 96.658 -77.881 47.717
142040 96.677 -77.888 47.693
142041 96.695 -77.896 47.669
142042 96.714 -77.902 47.645
142043 96.733 -77.910 47.621
142044 96.751 -77.918 47.597
142045 96.769 -77.924 47.573
142046 96.788 -77.932 47.549
142047 96.795 -77.931 47.527
142048 96.826 -77.939 47.505
142049 96.845 -77.943 47.482
142050 96.862 -77.946 47.459
142051 96.878 -77.952 47.435
142052 96.892 -77.968 47.412
142053 96.905 -77.984 47.389
142054 96.919 -78.000 47.365
142055 96.933 -78.015 47.341
142056 96.947 -78.031 47.318
142057 96.959 -78.039 47.297
142058 96.968 -78.031 47.276
142059 96.975 -78.024 47.255
142060 96.983 -78.015 47.233
142061 96.993 -78.010 47.210
142062 97.013 -78.015 47.188
142063 97.034 -78.020 47.167
142064 97.055 -78.025 47.146
142065 97.076 -78.030 47.127
142066 97.096 -78.035 47.108
142067 97.117 -78.040 47.089
142068 97.138 -78.045 47.071
142069 97.156 -78.054 47.054
142070 97.171 -78.045 47.039
142071 97.180 -78.040 47.024
142072 97.189 -78.036 47.010
142073 97.198 -78.031 46.997
142074 97.208 -78.027 46.985
142075 97.218 -78.024 46.974
142076 97.228 -78.021 46.962
142077 97.238 -78.018 46.952
142078 97.255 -78.018 46.944
142079 97.281 -78.021 46.936
142080 97.312 -78.026 46.927
142081 97.346 -78.033 46.916
142082 97.378 -78.040 46.904
142083 97.412 -78.047 46.893
142084 97.445 -78.055 46.884
142085 97.478 -78.062 46.874
142086 97.513 -78.067 46.864
142087 97.544 -78.083 46.850
142088 97.568 -78.107 46.835
142089 97.592 -78.131 46.818
142090 97.615 -78.156 46.800
142091 97.640 -78.180 46.780
142092 97.661 -78.197 46.760
142093 97.675 -78.201 46.740
142094 97.690 -78.204 46.720
142095 97.705 -78.207 46.699
142096 97.720 -78.211 46.677
142097 97.735 -78.214 46.654
142098 97.764 -78.221 46.631
142099 97.810 -78.232 46.609
142100 97.857 -78.244 46.587
142101 97.903 -78.254 46.565
142102 97.950 -78.266 46.544
142103 97.995 -78.275 46.522
142104 98.036 -78.282 46.500
142105 98.078 -78.289 46.477
142106 98.119 -78.297 46.455
142107 98.158 -78.304 46.432
142108 98.198 -78.311 46.409
142109 98.237 -78.318 46.385
142110 98.259 -78.364 46.362
142111 98.250 -78.374 46.339
142112 98.237 -78.357 46.316
142113 98.224 -78.329 46.293
142114 98.212 -78.299 46.270
142115 98.199 -78.270 46.248
142116 98.283 -78.275 46.225
142117 98.372 -78.291 46.203
142118 98.450 -78.308 46.182
142119 98.468 -78.331 46.159
142120 98.405 -78.344 46.139
142121 98.322 -78.334 46.118
142122 98.327 -78.299 46.097
142123 98.398 -78.275 46.076
142124 98.469 -78.250 46.055
142125 98.541 -78.226 46.033
142126 98.565 -78.222 46.010
142127 98.579 -78.221 45.989
142128 98.593 -78.220 45.968
142129 98.606 -78.213 45.945
142130 98.616 -78.197 45.922
142131 98.626 -78.179 45.898
142132 98.636 -78.163 45.874
142133 98.646 -78.146 45.851
142134 98.656 -78.128 45.828
142135 98.666 -78.112 45.807
142136 98.659 -78.100 45.786
142137 98.653 -78.100 45.764
142138 98.692 -78.159 45.744
142139 98.724 -78.212 45.723
142140 98.737 -78.245 45.703
142141 98.763 -78.265 45.686
142142 98.788 -78.285 45.670
142143 98.814 -78.304 45.655
142144 98.844 -78.318 45.640
142145 98.882 -78.325 45.627
142146 98.918 -78.334 45.614
142147 98.953 -78.344 45.599
142148 98.988 -78.354 45.584
142149 99.020 -78.361 45.568
142150 99.045 -78.360 45.553
142151 99.070 -78.359 45.539
142152 99.095 -78.358 45.525
142153 99.118 -78.358 45.511
142154 99.132 -78.361 45.496
142155 99.147 -78.363 45.482
142156 99.162 -78.366 45.468
142157 99.177 -78.369 45.454
142158 99.202 -78.375 45.441
142159 99.234 -78.383 45.430
142160 99.267 -78.391 45.419
142161 99.299 -78.399 45.407
142162 99.331 -78.407 45.395
142163 99.364 -78.415 45.383
142164 99.396 -78.423 45.371
142165 99.446 -78.435 45.358
142166 99.500 -78.446 45.345
142167 99.542 -78.452 45.332
142168 99.573 -78.454 45.318
142169 99.604 -78.455 45.303
142170 99.635 -78.456 45.290
142171 99.665 -78.457 45.278
142172 99.697 -78.473 45.266
142173 99.730 -78.495 45.254
142174 99.762 -78.517 45.241
142175 99.795 -78.539 45.228
142176 99.836 -78.573 45.216
142177 99.858 -78.590 45.204
142178 99.868 -78.597 45.192
142179 99.877 -78.606 45.180
142180 99.887 -78.615 45.168
142181 99.896 -78.624 45.157
142182 99.904 -78.632 45.145
142183 99.911 -78.639 45.134
142184 99.918 -78.645 45.122
142185 99.925 -78.651 45.110
142186 99.941 -78.660 45.097
142187 99.972 -78.669 45.083
142188 100.004 -78.679 45.069
142189 100.021 -78.690 45.056
142190 100.034 -78.702 45.044
142191 100.047 -78.714 45.031
142192 100.060 -78.727 45.020
142193 100.074 -78.739 45.008
142194 100.118 -78.755 44.997
142195 100.165 -78.772 44.986
142196 100.212 -78.789 44.975
142197 100.259 -78.805 44.963
142198 100.303 -78.821 44.950
142199 100.314 -78.831 44.937
142200 100.324 -78.842 44.924
142201 100.334 -78.852 44.911
142202 100.344 -78.863 44.897
142203 100.353 -78.874 44.883
142204 100.364 -78.890 44.869
142205 100.374 -78.909 44.855
142206 100.385 -78.929 44.841
142207 100.395 -78.948 44.826
142208 100.406 -78.967 44.811
142209 100.416 -78.987 44.797
142210 100.430 -79.000 44.781
142211 100.446 -79.012 44.765
142212 100.464 -79.023 44.749
142213 100.496 -79.043 44.733
142214 100.528 -79.062 44.718
142215 100.561 -79.081 44.703
142216 100.594 -79.100 44.686
142217 100.626 -79.118 44.670
142467 92.833 -78.154 54.262
142468 92.850 -78.149 54.258
142469 92.866 -78.144 54.254
142470 92.882 -78.138 54.248
142471 92.899 -78.133 54.243
142472 92.916 -78.128 54.239
142473 92.932 -78.122 54.233
142474 92.948 -78.116 54.229
142475 92.965 -78.111 54.225
142476 92.982 -78.106 54.221
142477 92.998 -78.100 54.215
142478 93.014 -78.095 54.210
142479 93.031 -78.090 54.205
142480 93.047 -78.084 54.200
142481 93.060 -78.079 54.196
142482 93.074 -78.075 54.192
142483 93.087 -78.070 54.187
142484 93.101 -78.065 54.182
142485 93.115 -78.056 54.177
142486 93.138 -78.044 54.171
142487 93.162 -78.032 54.166
142488 93.184 -78.016 54.162
142489 93.205 -77.998 54.158
142490 93.225 -77.980 54.154
142491 93.246 -77.970 54.147
142492 93.266 -77.964 54.141
142493 93.286 -77.957 54.135
142494 93.306 -77.951 54.131
142495 93.326 -77.945 54.124
142496 93.358 -77.883 54.117
142497 93.348 -77.893 54.109
142498 93.338 -77.903 54.099
142499 93.328 -77.915 54.090
142500 93.318 -77.925 54.080
142501 93.316 -77.933 54.070
142502 93.342 -77.926 54.059
142503 93.382 -77.907 54.049
142504 93.404 -77.887 54.038
142505 93.415 -77.867 54.025
142506 93.425 -77.847 54.012
142507 93.435 -77.827 53.998
142508 93.446 -77.808 53.984
142509 93.457 -77.787 53.970
142510 93.467 -77.768 53.957
142511 93.477 -77.748 53.942
142512 93.491 -77.739 53.926
142513 93.505 -77.731 53.911
142514 93.518 -77.723 53.895
142515 93.533 -77.715 53.878
142516 93.544 -77.702 53.861
142517 93.559 -77.699 53.843
142518 93.574 -77.689 53.822
142519 93.589 -77.680 53.801
142520 93.604 -77.670 53.780
142521 93.619 -77.661 53.761
142522 93.632 -77.653 53.742
142523 93.639 -77.649 53.724
142524 93.647 -77.646 53.707
142525 93.657 -77.642 53.689
142526 93.667 -77.637 53.671
142527 93.679 -77.633 53.653
142528 93.689 -77.627 53.635
142529 93.700 -77.623 53.618
142530 93.711 -77.617 53.601
142531 93.721 -77.613 53.584
142532 93.732 -77.608 53.567
142533 93.743 -77.603 53.549
142534 93.754 -77.599 53.532
142535 93.764 -77.593 53.514
142536 93.775 -77.589 53.497
142537 93.788 -77.584 53.480
142538 93.802 -77.580 53.463
142539 93.816 -77.577 53.446
142540 93.830 -77.572 53.429
142541 93.844 -77.568 53.412
142542 93.857 -77.564 53.396
142543 93.871 -77.560 53.381
142544 93.885 -77.556 53.365
142545 93.899 -77.552 53.349
142546 93.913 -77.548 53.335
142547 93.927 -77.544 53.322
142548 93.940 -77.540 53.311
142549 93.952 -77.525 53.297
142550 93.962 -77.507 53.283
142551 93.973 -77.489 53.268
142552 93.983 -77.471 53.252
142553 93.994 -77.463 53.234
142554 94.005 -77.457 53.216
142555 94.030 -77.457 53.196
142556 94.055 -77.457 53.177
142557 94.081 -77.457 53.158
142558 94.107 -77.458 53.140
142559 94.128 -77.457 53.121
142560 94.149 -77.455 53.102
142561 94.171 -77.453 53.084
142562 94.199 -77.453 53.067
142563 94.231 -77.453 53.052
142564 94.264 -77.454 53.041
142565 94.281 -77.450 53.030
142566 94.282 -77.442 53.020
142567 94.283 -77.436 53.008
142568 94.285 -77.428 52.995
142569 94.286 -77.420 52.982
142570 94.288 -77.413 52.968
142571 94.289 -77.405 52.953
142572 94.223 -77.379 52.937
142573 94.213 -77.408 52.922
142574 94.206 -77.439 52.907
142575 94.208 -77.467 52.894
142576 94.210 -77.478 52.883
142577 94.208 -77.473 52.874
142578 94.202 -77.451 52.865
142579 94.197 -77.430 52.858
142580 94.215 -77.423 52.850
142581 94.252 -77.428 52.839
142582 94.288 -77.432 52.828
142583 94.325 -77.437 52.819
142584 94.362 -77.442 52.810
142585 94.400 -77.438 52.799
142586 94.394 -77.424 52.789
142587 94.390 -77.407 52.782
142588 94.386 -77.390 52.779
142589 94.382 -77.373 52.778
142590 94.378 -77.356 52.778
142591 94.373 -77.338 52.776
142592 94.369 -77.321 52.770
142593 94.365 -77.304 52.761
142594 94.361 -77.287 52.748
142595 94.370 -77.281 52.730
142596 94.397 -77.291 52.709
142597 94.420 -77.304 52.684
142598 94.442 -77.316 52.658
142599 94.465 -77.329 52.629
142600 94.488 -77.341 52.597
142601 94.510 -77.353 52.561
142602 94.533 -77.365 52.524
142603 94.591 -77.364 52.488
142604 94.664 -77.357 52.451
142605 94.721 -77.363 52.418
142606 94.739 -77.432 52.389
142607 94.766 -77.464 52.366
142608 94.801 -77.459 52.341
142609 94.835 -77.455 52.311
142610 94.869 -77.451 52.276
142611 94.904 -77.448 52.243
142612 94.936 -77.443 52.212
142613 94.956 -77.436 52.183
142614 94.974 -77.427 52.153
142615 94.993 -77.419 52.121
142616 95.013 -77.411 52.085
142617 95.031 -77.403 52.047
142618 95.015 -77.361 52.008
142619 94.989 -77.329 51.969
142620 94.989 -77.291 51.930
142621 94.996 -77.267 51.893
142622 95.014 -77.265 51.853
142623 95.031 -77.263 51.813
142624 95.048 -77.261 51.772
142625 95.065 -77.260 51.731
142626 95.082 -77.257 51.692
142627 95.099 -77.256 51.651
142628 95.113 -77.271 51.610
142629 95.126 -77.292 51.558
142630 95.129 -77.316 51.505
142631 95.131 -77.341 51.449
142632 95.138 -77.356 51.393
142633 95.155 -77.350 51.336
142634 95.171 -77.344 51.279
142635 95.191 -77.340 51.220
142636 95.219 -77.341 51.164
142637 95.246 -77.342 51.110
142638 95.274 -77.344 51.058
142639 95.302 -77.346 51.008
142640 95.332 -77.350 50.961
142641 95.363 -77.354 50.916
142642 95.394 -77.357 50.873
142643 95.424 -77.361 50.830
142644 95.455 -77.365 50.789
142645 95.485 -77.370 50.750
142646 95.516 -77.373 50.713
142647 95.547 -77.377 50.678
142648 95.578 -77.381 50.643
142649 95.607 -77.389 50.610
142650 95.616 -77.398 50.580
142651 95.617 -77.402 50.551
142652 95.620 -77.409 50.520
142653 95.623 -77.417 50.486
142654 95.627 -77.425 50.453
142655 95.630 -77.433 50.417
142656 95.633 -77.441 50.378
142657 95.636 -77.449 50.336
142658 95.639 -77.457 50.299
142659 95.643 -77.465 50.262
142660 95.646 -77.473 50.225
142661 95.649 -77.481 50.183
142662 95.653 -77.489 50.139
142663 95.656 -77.497 50.094
142664 95.673 -77.514 50.049
142665 95.696 -77.533 50.006
142666 95.720 -77.552 49.964
142667 95.744 -77.572 49.922
142668 95.767 -77.592 49.882
142669 95.789 -77.606 49.842
142670 95.804 -77.604 49.803
142671 95.819 -77.602 49.765
142672 95.834 -77.600 49.727
142673 95.849 -77.598 49.688
142674 95.865 -77.595 49.648
142675 95.880 -77.593 49.607
142676 95.895 -77.592 49.568
142677 95.910 -77.589 49.531
142678 95.925 -77.587 49.495
142679 95.940 -77.585 49.461
142680 95.955 -77.583 49.429
142681 95.970 -77.580 49.399
142682 95.976 -77.586 49.369
142683 95.979 -77.593 49.341
142684 95.982 -77.602 49.313
142685 95.981 -77.624 49.286
142686 95.980 -77.645 49.256
142687 95.990 -77.663 49.223
142688 96.003 -77.677 49.190
142689 96.015 -77.683 49.158
142690 96.027 -77.691 49.126
142691 96.040 -77.699 49.095
142692 96.052 -77.705 49.064
142693 96.065 -77.713 49.035
142694 96.080 -77.699 49.004
142695 96.087 -77.695 48.972
142696 96.088 -77.704 48.940
142697 96.088 -77.713 48.908
142698 96.088 -77.722 48.877
142699 96.090 -77.730 48.844
142700 96.098 -77.740 48.810
142701 96.110 -77.749 48.776
142702 96.122 -77.758 48.741
142703 96.134 -77.767 48.706
142704 96.146 -77.777 48.671
142705 96.158 -77.786 48.639
142706 96.170 -77.795 48.606
142707 96.194 -77.799 48.574
142708 96.215 -77.804 48.544
142709 96.229 -77.808 48.512
142710 96.244 -77.812 48.481
142711 96.258 -77.816 48.450
142712 96.272 -77.820 48.422
142713 96.286 -77.822 48.396
142714 96.294 -77.819 48.369
142715 96.302 -77.816 48.343
142716 96.319 -77.812 48.317
142717 96.325 -77.818 48.292
142718 96.332 -77.823 48.267
142719 96.338 -77.828 48.244
142720 96.344 -77.834 48.220
142721 96.351 -77.839 48.195
142722 96.358 -77.844 48.171
142723 96.364 -77.849 48.146
142724 96.370 -77.855 48.121
142725 96.376 -77.860 48.096
142726 96.394 -77.872 48.072
142727 96.415 -77.883 48.048
142728 96.442 -77.889 48.024
142729 96.462 -77.890 47.999
142730 96.482 -77.890 47.975
142731 96.502 -77.891 47.950
142732 96.522 -77.891 47.925
142733 96.542 -77.892 47.902
142734 96.562 -77.893 47.876
142735 96.577 -77.882 47.851
142736 96.590 -77.865 47.827
142737 96.603 -77.849 47.803
142738 96.620 -77.851 47.777
142739 96.639 -77.859 47.752
142740 96.657 -77.865 47.729
142741 96.675 -77.873 47.705
142742 96.694 -77.881 47.681
142743 96.713 -77.887 47.657
142744 96.731 -77.895 47.633
142745 96.749 -77.902 47.610
142746 96.768 -77.909 47.587
142747 96.787 -77.917 47.564
142748 96.805 -77.924 47.542
142749 96.822 -77.931 47.520
142750 96.838 -77.934 47.498
142751 96.856 -77.937 47.475
142752 96.872 -77.945 47.453
142753 96.885 -77.961 47.429
142754 96.899 -77.977 47.405
142755 96.911 -77.990 47.381
142756 96.920 -77.997 47.357
142757 96.929 -78.004 47.334
142758 96.938 -78.011 47.314
142759 96.953 -78.010 47.291
142760 96.970 -78.007 47.270
142761 96.987 -78.003 47.249
142762 96.997 -77.998 47.229
142763 97.018 -78.003 47.207
142764 97.038 -78.008 47.184
142765 97.059 -78.012 47.163
142766 97.080 -78.017 47.145
142767 97.100 -78.022 47.125
142768 97.120 -78.026 47.105
142769 97.141 -78.031 47.088
142770 97.161 -78.041 47.071
142771 97.171 -78.038 47.055
142772 97.181 -78.035 47.039
142773 97.191 -78.032 47.025
142774 97.202 -78.029 47.012
142775 97.212 -78.026 47.000
142776 97.222 -78.023 46.987
142777 97.232 -78.020 46.976
142778 97.244 -78.015 46.965
142779 97.267 -78.012 46.955
142780 97.289 -78.025 46.946
142781 97.313 -78.033 46.936
142782 97.342 -78.033 46.926
142783 97.369 -78.033 46.915
142784 97.397 -78.033 46.905
142785 97.430 -78.038 46.896
142786 97.463 -78.046 46.888
142787 97.495 -78.056 46.878
142788 97.519 -78.081 46.865
142789 97.543 -78.105 46.849
142790 97.567 -78.129 46.832
142791 97.591 -78.153 46.815
142792 97.615 -78.178 46.794
142793 97.636 -78.197 46.775
142794 97.651 -78.200 46.756
142795 97.666 -78.203 46.736
142796 97.680 -78.206 46.714
142797 97.696 -78.210 46.691
142798 97.715 -78.214 46.668
142799 97.761 -78.225 46.645
142800 97.808 -78.236 46.623
142801 97.855 -78.247 46.602
142802 97.901 -78.259 46.580
142803 97.948 -78.270 46.558
142804 97.994 -78.281 46.536
142805 98.037 -78.289 46.514
142806 98.079 -78.297 46.491
142807 98.121 -78.304 46.469
142808 98.162 -78.310 46.446
142809 98.202 -78.318 46.422
142810 98.241 -78.325 46.399
142811 98.264 -78.362 46.375
142812 98.250 -78.354 46.351
142813 98.238 -78.325 46.328
142814 98.225 -78.296 46.305
142815 98.212 -78.267 46.283
142816 98.201 -78.238 46.260
142817 98.243 -78.241 46.237
142818 98.321 -78.257 46.214
142819 98.399 -78.274 46.192
142820 98.477 -78.291 46.168
142821 98.437 -78.319 46.146
142822 98.384 -78.345 46.125
142823 98.370 -78.312 46.105
142824 98.422 -78.266 46.085
142825 98.493 -78.241 46.065
142826 98.564 -78.217 46.043
142827 98.590 -78.212 46.020
142828 98.604 -78.211 45.998
142829 98.618 -78.207 45.976
142830 98.628 -78.191 45.954
142831 98.638 -78.173 45.932
142832 98.648 -78.156 45.908
142833 98.658 -78.140 45.885
142834 98.668 -78.122 45.863
142835 98.678 -78.106 45.840
142836 98.688 -78.088 45.819
142837 98.689 -78.074 45.797
142838 98.689 -78.114 45.775
142839 98.706 -78.196 45.754
142840 98.728 -78.235 45.734
142841 98.753 -78.254 45.716
142842 98.779 -78.274 45.698
142843 98.804 -78.294 45.682
142844 98.830 -78.313 45.667
142845 98.856 -78.332 45.652
142846 98.893 -78.339 45.638
142847 98.928 -78.348 45.624
142848 98.963 -78.359 45.610
142849 98.989 -78.359 45.595
142850 99.014 -78.357 45.582
142851 99.039 -78.357 45.568
142852 99.064 -78.355 45.553
142853 99.088 -78.354 45.540
142854 99.109 -78.355 45.525
142855 99.124 -78.357 45.510
142856 99.139 -78.360 45.495
142857 99.153 -78.363 45.481
142858 99.175 -78.368 45.467
142859 99.208 -78.376 45.454
142860 99.240 -78.384 45.441
142861 99.273 -78.392 45.430
142862 99.305 -78.401 45.419
142863 99.337 -78.408 45.407
142864 99.370 -78.416 45.396
142865 99.402 -78.424 45.383
142866 99.435 -78.432 45.371
142867 99.482 -78.443 45.359
142868 99.535 -78.454 45.345
142869 99.588 -78.467 45.330
142870 99.627 -78.471 45.315
142871 99.658 -78.473 45.303
142872 99.689 -78.473 45.290
142873 99.720 -78.482 45.277
142874 99.753 -78.504 45.265
142875 99.785 -78.526 45.253
142876 99.818 -78.548 45.241
142877 99.850 -78.571 45.230
142878 99.864 -78.582 45.217
142879 99.873 -78.592 45.205
142880 99.883 -78.601 45.194
142881 99.892 -78.610 45.184
142882 99.900 -78.618 45.171
142883 99.907 -78.624 45.159
142884 99.914 -78.631 45.148
142885 99.921 -78.638 45.136
142886 99.928 -78.645 45.123
142887 99.935 -78.651 45.110
142888 99.965 -78.661 45.097
142889 99.997 -78.670 45.082
142890 100.028 -78.680 45.069
142891 100.049 -78.691 45.057
142892 100.062 -78.703 45.044
142893 100.075 -78.715 45.032
142894 100.111 -78.730 45.020
142895 100.158 -78.747 45.008
142896 100.206 -78.764 44.997
142897 100.253 -78.781 44.985
142898 100.292 -78.799 44.973
142899 100.312 -78.811 44.961
142900 100.322 -78.821 44.948
142901 100.333 -78.832 44.934
142902 100.343 -78.842 44.921
142903 100.353 -78.852 44.907
142904 100.363 -78.863 44.891
142905 100.373 -78.882 44.877
142906 100.384 -78.902 44.863
142907 100.395 -78.921 44.848
142908 100.405 -78.940 44.833
142909 100.415 -78.959 44.818
142910 100.426 -78.978 44.803
142911 100.438 -78.996 44.788
142912 100.454 -79.006 44.772
142913 100.471 -79.018 44.755
142914 100.494 -79.032 44.739
142915 100.526 -79.051 44.723
142916 100.559 -79.071 44.707
142917 100.591 -79.090 44.690
142918 100.624 -79.109 44.675
142919 100.656 -79.128 44.659
143168 92.820 -78.144 54.260
143169 92.837 -78.138 54.256
143170 92.853 -78.132 54.253
143171 92.870 -78.127 54.247
143172 92.886 -78.122 54.241
143173 92.903 -78.116 54.237
143174 92.919 -78.111 54.234
143175 92.936 -78.106 54.229
143176 92.952 -78.100 54.225
143177 92.969 -78.094 54.221
143178 92.985 -78.089 54.216
143179 93.002 -78.084 54.211
143180 93.018 -78.078 54.207
143181 93.035 -78.073 54.202
143182 93.051 -78.068 54.198
143183 93.068 -78.062 54.194
143184 93.084 -78.056 54.190
143185 93.100 -78.051 54.187
143186 93.114 -78.046 54.182
143187 93.136 -78.031 54.176
143188 93.156 -78.023 54.170
143189 93.176 -78.017 54.166
143190 93.196 -78.011 54.162
143191 93.216 -78.004 54.158
143192 93.236 -77.998 54.153
143193 93.256 -77.992 54.146
143194 93.276 -77.985 54.141
143195 93.296 -77.979 54.136
143196 93.319 -77.962 54.131
143197 93.349 -77.876 54.125
143198 93.339 -77.887 54.117
143199 93.329 -77.898 54.108
143200 93.319 -77.909 54.098
143201 93.303 -77.938 54.088
143202 93.301 -77.927 54.079
143203 93.308 -77.912 54.070
143204 93.346 -77.893 54.060
143205 93.386 -77.874 54.049
143206 93.404 -77.854 54.036
143207 93.414 -77.834 54.024
143208 93.425 -77.815 54.009
143209 93.435 -77.795 53.995
143210 93.446 -77.774 53.983
143211 93.456 -77.755 53.970
143212 93.467 -77.735 53.956
143213 93.480 -77.726 53.942
143214 93.494 -77.718 53.927
143215 93.508 -77.711 53.912
143216 93.522 -77.702 53.897
143217 93.536 -77.695 53.880
143218 93.551 -77.687 53.862
143219 93.565 -77.677 53.842
143220 93.580 -77.668 53.820
143221 93.595 -77.658 53.799
143222 93.610 -77.649 53.780
143223 93.625 -77.639 53.761
143224 93.634 -77.635 53.744
143225 93.642 -77.631 53.727
143226 93.650 -77.628 53.709
143227 93.660 -77.624 53.691
143228 93.671 -77.619 53.674
143229 93.682 -77.614 53.656
143230 93.692 -77.609 53.639
143231 93.703 -77.605 53.621
143232 93.714 -77.599 53.604
143233 93.724 -77.595 53.587
143234 93.735 -77.589 53.570
143235 93.746 -77.585 53.552
143236 93.757 -77.580 53.535
143237 93.769 -77.576 53.517
143238 93.783 -77.571 53.501
143239 93.796 -77.567 53.484
143240 93.811 -77.564 53.468
143241 93.824 -77.559 53.451
143242 93.838 -77.555 53.435
143243 93.852 -77.551 53.418
143244 93.866 -77.547 53.402
143245 93.880 -77.543 53.387
143246 93.894 -77.539 53.372
143247 93.908 -77.535 53.359
143248 93.923 -77.526 53.347
143249 93.939 -77.508 53.338
143250 93.947 -77.490 53.327
143251 93.954 -77.473 53.313
143252 93.965 -77.456 53.299
143253 93.978 -77.448 53.284
143254 94.004 -77.448 53.267
143255 94.030 -77.448 53.249
143256 94.055 -77.449 53.230
143257 94.081 -77.449 53.211
143258 94.106 -77.450 53.191
143259 94.132 -77.450 53.172
143260 94.157 -77.451 53.153
143261 94.179 -77.449 53.133
143262 94.204 -77.448 53.113
143263 94.237 -77.448 53.095
143264 94.245 -77.442 53.080
143265 94.247 -77.436 53.069
143266 94.248 -77.428 53.060
143267 94.250 -77.420 53.051
143268 94.251 -77.413 53.040
143269 94.252 -77.406 53.028
143270 94.254 -77.398 53.015
143271 94.255 -77.391 53.002
143272 94.257 -77.383 52.988
143273 94.219 -77.372 52.972
143274 94.211 -77.401 52.956
143275 94.214 -77.431 52.941
143276 94.216 -77.457 52.925
143277 94.217 -77.460 52.911
143278 94.216 -77.456 52.900
143279 94.210 -77.435 52.892
143280 94.204 -77.414 52.887
143281 94.233 -77.414 52.881
143282 94.270 -77.418 52.870
143283 94.307 -77.423 52.857
143284 94.343 -77.427 52.847
143285 94.380 -77.432 52.839
143286 94.393 -77.423 52.829
143287 94.389 -77.405 52.816
143288 94.386 -77.388 52.804
143289 94.382 -77.370 52.793
143290 94.378 -77.353 52.790
143291 94.374 -77.336 52.787
143292 94.370 -77.319 52.784
143293 94.367 -77.301 52.780
143294 94.363 -77.284 52.772
143295 94.372 -77.273 52.760
143296 94.391 -77.268 52.744
143297 94.413 -77.280 52.722
143298 94.435 -77.292 52.698
143299 94.458 -77.304 52.672
143300 94.481 -77.317 52.644
143301 94.503 -77.329 52.613
143302 94.526 -77.341 52.578
143303 94.547 -77.350 52.541
143304 94.567 -77.351 52.505
143305 94.637 -77.345 52.472
143306 94.710 -77.334 52.444
143307 94.743 -77.355 52.419
143308 94.761 -77.424 52.397
143309 94.794 -77.426 52.369
143310 94.828 -77.423 52.338
143311 94.862 -77.419 52.302
143312 94.897 -77.415 52.273
143313 94.924 -77.409 52.245
143314 94.943 -77.401 52.216
143315 94.962 -77.393 52.188
143316 94.981 -77.385 52.154
143317 95.000 -77.377 52.118
143318 95.019 -77.369 52.080
143319 95.019 -77.342 52.041
143320 94.999 -77.307 52.003
143321 95.001 -77.273 51.965
143322 95.019 -77.272 51.928
143323 95.036 -77.269 51.888
143324 95.053 -77.268 51.847
143325 95.070 -77.266 51.805
143326 95.087 -77.264 51.765
143327 95.104 -77.262 51.725
143328 95.113 -77.260 51.685
143329 95.126 -77.261 51.642
143330 95.129 -77.285 51.590
143331 95.131 -77.310 51.536
143332 95.134 -77.335 51.480
143333 95.142 -77.347 51.425
143334 95.158 -77.341 51.368
143335 95.175 -77.334 51.308
143336 95.191 -77.328 51.246
143337 95.207 -77.321 51.188
143338 95.232 -77.320 51.133
143339 95.259 -77.322 51.080
143340 95.286 -77.323 51.030
143341 95.315 -77.325 50.982
143342 95.345 -77.329 50.938
143343 95.376 -77.332 50.895
143344 95.406 -77.336 50.852
143345 95.437 -77.340 50.809
143346 95.468 -77.345 50.768
143347 95.498 -77.348 50.730
143348 95.529 -77.352 50.695
143349 95.559 -77.356 50.661
143350 95.588 -77.370 50.628
143351 95.614 -77.392 50.598
143352 95.621 -77.400 50.569
143353 95.623 -77.404 50.540
143354 95.625 -77.408 50.507
143355 95.627 -77.412 50.474
143356 95.629 -77.416 50.438
143357 95.631 -77.420 50.398
143358 95.632 -77.423 50.355
143359 95.635 -77.430 50.319
143360 95.638 -77.439 50.283
143361 95.642 -77.446 50.246
143362 95.645 -77.454 50.205
143363 95.649 -77.463 50.162
143364 95.652 -77.470 50.118
143365 95.672 -77.489 50.074
143366 95.695 -77.508 50.030
143367 95.719 -77.527 49.987
143368 95.742 -77.547 49.945
143369 95.766 -77.567 49.904
143370 95.790 -77.586 49.863
143371 95.812 -77.604 49.822
143372 95.827 -77.602 49.782
143373 95.842 -77.599 49.741
143374 95.858 -77.597 49.700
143375 95.873 -77.595 49.659
143376 95.888 -77.592 49.616
143377 95.903 -77.591 49.576
143378 95.918 -77.589 49.538
143379 95.933 -77.586 49.501
143380 95.949 -77.584 49.467
143381 95.964 -77.582 49.435
143382 95.974 -77.583 49.404
143383 95.977 -77.591 49.373
143384 95.981 -77.599 49.343
143385 95.984 -77.605 49.314
143386 95.984 -77.624 49.285
143387 95.983 -77.645 49.256
143388 95.993 -77.663 49.226
143389 96.005 -77.680 49.195
143390 96.016 -77.699 49.165
143391 96.029 -77.706 49.134
143392 96.041 -77.714 49.106
143393 96.054 -77.721 49.077
143394 96.068 -77.711 49.049
143395 96.085 -77.688 49.018
143396 96.090 -77.687 48.986
143397 96.090 -77.696 48.954
143398 96.091 -77.705 48.922
143399 96.091 -77.714 48.890
143400 96.095 -77.724 48.858
143401 96.095 -77.733 48.826
143402 96.107 -77.743 48.792
143403 96.119 -77.752 48.757
143404 96.131 -77.761 48.722
143405 96.143 -77.771 48.686
143406 96.155 -77.780 48.654
143407 96.176 -77.786 48.622
143408 96.203 -77.790 48.590
143409 96.218 -77.794 48.559
143410 96.233 -77.798 48.527
143411 96.247 -77.802 48.496
143412 96.261 -77.806 48.465
143413 96.276 -77.811 48.438
143414 96.291 -77.815 48.412
143415 96.304 -77.818 48.386
143416 96.312 -77.814 48.360
143417 96.324 -77.816 48.334
143418 96.330 -77.821 48.309
143419 96.336 -77.826 48.283
143420 96.343 -77.831 48.260
143421 96.349 -77.837 48.237
143422 96.356 -77.842 48.212
143423 96.362 -77.847 48.188
143424 96.368 -77.852 48.163
143425 96.375 -77.858 48.137
143426 96.381 -77.863 48.112
143427 96.396 -77.874 48.088
143428 96.422 -77.883 48.064
143429 96.442 -77.884 48.039
143430 96.462 -77.884 48.015
143431 96.482 -77.885 47.989
143432 96.503 -77.886 47.964
143433 96.523 -77.887 47.940
143434 96.543 -77.887 47.916
143435 96.563 -77.887 47.891
143436 96.579 -77.878 47.866
143437 96.591 -77.862 47.841
143438 96.604 -77.846 47.817
143439 96.618 -77.836 47.792
143440 96.637 -77.843 47.767
143441 96.656 -77.850 47.744
143442 96.674 -77.858 47.720
143443 96.692 -77.865 47.695
143444 96.711 -77.872 47.671
143445 96.730 -77.880 47.647
143446 96.748 -77.887 47.624
143447 96.766 -77.894 47.601
143448 96.785 -77.901 47.578
143449 96.803 -77.909 47.556
143450 96.818 -77.915 47.535
143451 96.834 -77.922 47.513
143452 96.850 -77.929 47.491
143453 96.865 -77.937 47.469
143454 96.874 -77.946 47.445
143455 96.883 -77.953 47.420
143456 96.892 -77.960 47.397
143457 96.900 -77.967 47.372
143458 96.909 -77.974 47.350
143459 96.918 -77.981 47.329
143460 96.929 -77.985 47.307
143461 96.951 -77.984 47.286
143462 96.974 -77.983 47.266
143463 96.998 -77.982 47.246
143464 97.018 -77.987 47.225
143465 97.038 -77.991 47.202
143466 97.058 -77.996 47.180
143467 97.079 -78.000 47.162
143468 97.099 -78.006 47.141
143469 97.119 -78.010 47.122
143470 97.140 -78.016 47.104
143471 97.161 -78.034 47.087
143472 97.175 -78.037 47.071
143473 97.185 -78.034 47.055
143474 97.195 -78.031 47.041
143475 97.205 -78.028 47.027
143476 97.215 -78.025 47.014
143477 97.225 -78.022 47.001
143478 97.235 -78.019 46.990
143479 97.257 -78.008 46.978
143480 97.281 -78.007 46.967
143481 97.302 -78.020 46.957
143482 97.323 -78.034 46.947
143483 97.346 -78.044 46.936
143484 97.373 -78.044 46.925
143485 97.401 -78.044 46.917
143486 97.430 -78.044 46.908
143487 97.458 -78.044 46.900
143488 97.470 -78.054 46.892
143489 97.495 -78.078 46.878
143490 97.518 -78.103 46.862
143491 97.542 -78.127 46.845
143492 97.566 -78.152 46.828
143493 97.590 -78.176 46.809
143494 97.611 -78.195 46.790
143495 97.627 -78.198 46.772
143496 97.641 -78.202 46.751
143497 97.656 -78.205 46.728
143498 97.671 -78.208 46.704
143499 97.712 -78.218 46.682
143500 97.759 -78.229 46.659
143501 97.805 -78.241 46.637
143502 97.852 -78.252 46.616
143503 97.898 -78.263 46.594
143504 97.945 -78.274 46.572
143505 97.992 -78.285 46.551
143506 98.038 -78.296 46.529
143507 98.080 -78.304 46.505
143508 98.122 -78.310 46.482
143509 98.163 -78.317 46.459
143510 98.205 -78.325 46.434
143511 98.245 -78.332 46.411
143512 98.264 -78.349 46.387
143513 98.251 -78.321 46.363
143514 98.243 -78.294 46.341
143515 98.255 -78.272 46.318
143516 98.267 -78.251 46.295
143517 98.278 -78.231 46.272
143518 98.258 -78.233 46.249
143519 98.292 -78.235 46.225
143520 98.348 -78.240 46.202
143521 98.426 -78.257 46.179
143522 98.454 -78.279 46.156
143523 98.405 -78.308 46.134
143524 98.423 -78.310 46.114
143525 98.459 -78.274 46.093
143526 98.515 -78.232 46.072
143527 98.586 -78.208 46.051
143528 98.615 -78.202 46.029
143529 98.629 -78.201 46.007
143530 98.639 -78.184 45.986
143531 98.649 -78.167 45.964
143532 98.659 -78.150 45.943
143533 98.669 -78.134 45.919
143534 98.679 -78.116 45.896
143535 98.690 -78.100 45.874
143536 98.700 -78.082 45.852
143537 98.710 -78.065 45.830
143538 98.719 -78.049 45.809
143539 98.724 -78.113 45.787
143540 98.727 -78.184 45.765
143541 98.740 -78.234 45.746
143542 98.767 -78.257 45.727
143543 98.793 -78.279 45.710
143544 98.819 -78.301 45.695
143545 98.845 -78.322 45.680
143546 98.871 -78.341 45.665
143547 98.904 -78.354 45.650
143548 98.933 -78.357 45.635
143549 98.957 -78.356 45.620
143550 98.982 -78.354 45.607
143551 99.007 -78.354 45.594
143552 99.032 -78.352 45.581
143553 99.057 -78.351 45.567
143554 99.082 -78.350 45.553
143555 99.100 -78.351 45.539
143556 99.115 -78.354 45.524
143557 99.130 -78.357 45.509
143558 99.149 -78.361 45.495
143559 99.181 -78.369 45.480
143560 99.214 -78.377 45.466
143561 99.246 -78.385 45.454
143562 99.279 -78.393 45.442
143563 99.311 -78.401 45.431
143564 99.343 -78.410 45.420
143565 99.376 -78.417 45.408
143566 99.408 -78.426 45.395
143567 99.441 -78.433 45.382
143568 99.473 -78.442 45.369
143569 99.517 -78.451 45.356
143570 99.570 -78.463 45.342
143571 99.624 -78.475 45.328
143572 99.677 -78.486 45.314
143573 99.712 -78.489 45.302
143574 99.743 -78.491 45.289
143575 99.776 -78.514 45.277
143576 99.808 -78.535 45.265
143577 99.833 -78.557 45.254
143578 99.855 -78.579 45.242
143579 99.869 -78.578 45.229
143580 99.878 -78.587 45.218
143581 99.888 -78.596 45.207
143582 99.896 -78.604 45.197
143583 99.903 -78.611 45.185
143584 99.910 -78.617 45.173
143585 99.917 -78.624 45.161
143586 99.924 -78.631 45.150
143587 99.931 -78.637 45.137
143588 99.939 -78.644 45.124
143589 99.958 -78.652 45.110
143590 99.990 -78.661 45.095
143591 100.021 -78.671 45.082
143592 100.053 -78.681 45.070
143593 100.077 -78.691 45.057
143594 100.104 -78.705 45.044
143595 100.150 -78.722 45.032
143596 100.188 -78.741 45.020
143597 100.226 -78.759 45.007
143598 100.264 -78.777 44.995
143599 100.296 -78.794 44.983
143600 100.315 -78.805 44.971
143601 100.328 -78.816 44.958
143602 100.339 -78.827 44.944
143603 100.349 -78.836 44.929
143604 100.359 -78.846 44.915
143605 100.370 -78.856 44.900
143606 100.383 -78.874 44.885
143607 100.394 -78.893 44.870
143608 100.404 -78.913 44.855
143609 100.414 -78.932 44.840
143610 100.425 -78.952 44.825
143611 100.435 -78.971 44.810
143612 100.446 -78.990 44.794
143613 100.462 -79.002 44.778
143614 100.478 -79.012 44.762
143615 100.494 -79.024 44.746
143616 100.524 -79.041 44.729
143617 100.556 -79.060 44.712
143618 100.588 -79.080 44.696
143619 100.621 -79.099 44.679
143620 100.653 -79.118 44.664
143621 100.686 -79.137 44.647
143622 100.718 -79.156 44.631
143868 92.786 -78.116 54.262
143869 92.804 -78.119 54.258
143870 92.823 -78.122 54.253
143871 92.840 -78.122 54.250
143872 92.857 -78.116 54.245
143873 92.873 -78.110 54.240
143874 92.890 -78.105 54.236
143875 92.906 -78.100 54.232
143876 92.923 -78.094 54.228
143877 92.939 -78.089 54.224
143878 92.956 -78.084 54.221
143879 92.972 -78.078 54.216
143880 92.989 -78.072 54.212
143881 93.005 -78.067 54.208
143882 93.022 -78.062 54.203
143883 93.038 -78.056 54.199
143884 93.055 -78.051 54.196
143885 93.071 -78.045 54.194
143886 93.088 -78.040 54.192
143887 93.106 -78.038 54.186
143888 93.126 -78.040 54.180
143889 93.146 -78.044 54.175
143890 93.166 -78.045 54.170
143891 93.187 -78.039 54.166
143892 93.207 -78.032 54.163
143893 93.227 -78.026 54.158
143894 93.247 -78.020 54.153
143895 93.267 -78.013 54.148
143896 93.287 -78.007 54.143
143897 93.325 -77.918 54.139
143898 93.339 -77.871 54.134
143899 93.324 -77.897 54.127
143900 93.307 -77.932 54.119
143901 93.300 -77.936 54.109
143902 93.299 -77.921 54.100
143903 93.298 -77.906 54.091
143904 93.300 -77.892 54.082
143905 93.316 -77.879 54.072
143906 93.350 -77.861 54.061
143907 93.390 -77.841 54.049
143908 93.404 -77.821 54.035
143909 93.414 -77.802 54.022
143910 93.425 -77.781 54.009
143911 93.435 -77.761 53.997
143912 93.445 -77.742 53.986
143913 93.456 -77.722 53.973
143914 93.470 -77.713 53.959
143915 93.484 -77.705 53.945
143916 93.497 -77.698 53.930
143917 93.511 -77.690 53.914
143918 93.525 -77.683 53.899
143919 93.542 -77.675 53.881
143920 93.557 -77.665 53.861
143921 93.572 -77.656 53.839
143922 93.587 -77.646 53.819
143923 93.602 -77.637 53.800
143924 93.617 -77.627 53.782
143925 93.630 -77.620 53.764
143926 93.637 -77.617 53.748
143927 93.645 -77.613 53.731
143928 93.653 -77.610 53.713
143929 93.663 -77.605 53.695
143930 93.674 -77.601 53.678
143931 93.685 -77.595 53.661
143932 93.695 -77.591 53.644
143933 93.706 -77.586 53.626
143934 93.717 -77.581 53.609
143935 93.727 -77.577 53.592
143936 93.738 -77.571 53.574
143937 93.749 -77.567 53.556
143938 93.763 -77.562 53.539
143939 93.777 -77.558 53.523
143940 93.791 -77.555 53.507
143941 93.805 -77.551 53.490
143942 93.819 -77.546 53.474
143943 93.833 -77.542 53.457
143944 93.846 -77.539 53.441
143945 93.860 -77.534 53.425
143946 93.874 -77.529 53.410
143947 93.890 -77.521 53.396
143948 93.905 -77.510 53.382
143949 93.921 -77.494 53.370
143950 93.939 -77.473 53.359
143951 93.948 -77.454 53.348
143952 93.955 -77.438 53.337
143953 93.960 -77.434 53.325
143954 93.988 -77.436 53.312
143955 94.015 -77.437 53.296
143956 94.044 -77.438 53.278
143957 94.072 -77.439 53.260
143958 94.100 -77.440 53.243
143959 94.128 -77.441 53.224
143960 94.156 -77.442 53.205
143961 94.183 -77.443 53.185
143962 94.208 -77.443 53.164
143963 94.210 -77.436 53.144
143964 94.211 -77.428 53.125
143965 94.213 -77.420 53.109
143966 94.214 -77.414 53.097
143967 94.215 -77.406 53.089
143968 94.217 -77.398 53.083
143969 94.218 -77.391 53.073
143970 94.220 -77.384 53.062
143971 94.221 -77.376 53.050
143972 94.222 -77.369 53.037
143973 94.224 -77.361 53.023
143974 94.216 -77.365 53.008
143975 94.219 -77.394 52.991
143976 94.221 -77.423 52.976
143977 94.223 -77.439 52.958
143978 94.224 -77.442 52.942
143979 94.223 -77.440 52.928
143980 94.218 -77.419 52.920
143981 94.214 -77.399 52.915
143982 94.251 -77.404 52.909
143983 94.288 -77.408 52.900
143984 94.324 -77.413 52.886
143985 94.361 -77.417 52.875
143986 94.376 -77.415 52.866
143987 94.385 -77.401 52.857
143988 94.391 -77.384 52.845
143989 94.387 -77.367 52.830
143990 94.383 -77.350 52.813
143991 94.380 -77.332 52.804
143992 94.376 -77.315 52.797
143993 94.372 -77.298 52.790
143994 94.370 -77.282 52.785
143995 94.389 -77.276 52.776
143996 94.407 -77.270 52.765
143997 94.425 -77.265 52.749
143998 94.445 -77.266 52.729
143999 94.468 -77.279 52.707
144000 94.491 -77.291 52.682
144001 94.513 -77.304 52.655
144002 94.536 -77.316 52.625
144003 94.557 -77.326 52.593
144004 94.574 -77.328 52.560
144005 94.591 -77.329 52.529
144006 94.617 -77.329 52.501
144007 94.687 -77.323 52.476
144008 94.742 -77.314 52.452
144009 94.764 -77.348 52.426
144010 94.787 -77.395 52.395
144011 94.822 -77.390 52.363
144012 94.856 -77.386 52.331
144013 94.891 -77.382 52.304
144014 94.912 -77.375 52.276
144015 94.930 -77.367 52.248
144016 94.950 -77.359 52.219
144017 94.969 -77.351 52.185
144018 94.987 -77.342 52.148
144019 95.006 -77.335 52.110
144020 95.023 -77.324 52.072
144021 95.009 -77.287 52.035
144022 95.023 -77.278 51.999
144023 95.040 -77.276 51.961
144024 95.057 -77.274 51.921
144025 95.069 -77.273 51.880
144026 95.076 -77.271 51.838
144027 95.084 -77.270 51.798
144028 95.090 -77.260 51.758
144029 95.097 -77.240 51.718
144030 95.123 -77.256 51.675
144031 95.131 -77.279 51.622
144032 95.134 -77.304 51.567
144033 95.136 -77.328 51.512
144034 95.145 -77.338 51.458
144035 95.162 -77.331 51.399
144036 95.179 -77.325 51.338
144037 95.195 -77.319 51.275
144038 95.211 -77.312 51.214
144039 95.228 -77.306 51.158
144040 95.245 -77.300 51.105
144041 95.272 -77.301 51.051
144042 95.300 -77.302 51.005
144043 95.327 -77.304 50.961
144044 95.358 -77.307 50.919
144045 95.388 -77.311 50.874
144046 95.419 -77.316 50.831
144047 95.450 -77.320 50.789
144048 95.480 -77.323 50.748
144049 95.511 -77.327 50.712
144050 95.542 -77.331 50.678
144051 95.568 -77.350 50.647
144052 95.595 -77.373 50.616
144053 95.621 -77.395 50.588
144054 95.627 -77.402 50.559
144055 95.629 -77.406 50.528
144056 95.631 -77.410 50.495
144057 95.633 -77.414 50.459
144058 95.634 -77.418 50.418
144059 95.636 -77.422 50.374
144060 95.638 -77.426 50.336
144061 95.640 -77.429 50.301
144062 95.642 -77.433 50.266
144063 95.644 -77.437 50.226
144064 95.646 -77.442 50.184
144065 95.648 -77.445 50.142
144066 95.671 -77.464 50.099
144067 95.694 -77.483 50.054
144068 95.718 -77.502 50.010
144069 95.741 -77.522 49.968
144070 95.765 -77.542 49.926
144071 95.788 -77.561 49.885
144072 95.812 -77.580 49.842
144073 95.835 -77.600 49.799
144074 95.850 -77.599 49.756
144075 95.866 -77.596 49.713
144076 95.881 -77.594 49.670
144077 95.896 -77.592 49.627
144078 95.911 -77.590 49.585
144079 95.926 -77.588 49.546
144080 95.942 -77.586 49.509
144081 95.957 -77.583 49.474
144082 95.972 -77.581 49.442
144083 95.975 -77.589 49.410
144084 95.979 -77.595 49.378
144085 95.982 -77.603 49.347
144086 95.985 -77.611 49.316
144087 95.987 -77.624 49.287
144088 95.986 -77.645 49.258
144089 95.997 -77.663 49.229
144090 96.008 -77.680 49.201
144091 96.020 -77.698 49.172
144092 96.031 -77.716 49.144
144093 96.043 -77.729 49.118
144094 96.057 -77.723 49.091
144095 96.073 -77.700 49.063
144096 96.090 -77.677 49.033
144097 96.093 -77.680 49.002
144098 96.093 -77.689 48.970
144099 96.094 -77.699 48.936
144100 96.101 -77.707 48.904
144101 96.102 -77.716 48.873
144102 96.093 -77.728 48.842
144103 96.104 -77.737 48.808
144104 96.116 -77.746 48.773
144105 96.128 -77.755 48.738
144106 96.141 -77.765 48.702
144107 96.157 -77.772 48.670
144108 96.185 -77.776 48.638
144109 96.207 -77.780 48.606
144110 96.222 -77.784 48.574
144111 96.236 -77.788 48.543
144112 96.251 -77.793 48.512
144113 96.265 -77.796 48.481
144114 96.279 -77.801 48.455
144115 96.294 -77.805 48.429
144116 96.309 -77.809 48.403
144117 96.322 -77.813 48.377
144118 96.328 -77.819 48.351
144119 96.334 -77.824 48.325
144120 96.341 -77.830 48.300
144121 96.347 -77.835 48.277
144122 96.354 -77.840 48.253
144123 96.360 -77.846 48.230
144124 96.366 -77.851 48.205
144125 96.373 -77.856 48.180
144126 96.379 -77.862 48.154
144127 96.385 -77.866 48.129
144128 96.401 -77.876 48.105
144129 96.422 -77.877 48.080
144130 96.443 -77.878 48.056
144131 96.463 -77.879 48.031
144132 96.483 -77.880 48.005
144133 96.503 -77.881 47.979
144134 96.523 -77.881 47.954
144135 96.543 -77.882 47.930
144136 96.564 -77.883 47.906
144137 96.580 -77.875 47.882
144138 96.593 -77.859 47.857
144139 96.605 -77.842 47.833
144140 96.618 -77.825 47.807
144141 96.636 -77.827 47.783
144142 96.654 -77.835 47.759
144143 96.672 -77.843 47.735
144144 96.691 -77.849 47.711
144145 96.710 -77.857 47.686
144146 96.728 -77.864 47.662
144147 96.747 -77.871 47.639
144148 96.765 -77.879 47.615
144149 96.784 -77.886 47.592
144150 96.800 -77.893 47.570
144151 96.815 -77.899 47.549
144152 96.831 -77.906 47.528
144153 96.848 -77.915 47.506
144154 96.867 -77.926 47.484
144155 96.876 -77.934 47.460
144156 96.885 -77.940 47.435
144157 96.894 -77.948 47.411
144158 96.903 -77.956 47.387
144159 96.912 -77.962 47.365
144160 96.921 -77.970 47.345
144161 96.930 -77.969 47.324
144162 96.951 -77.968 47.303
144163 96.974 -77.967 47.283
144164 96.997 -77.966 47.262
144165 97.018 -77.970 47.242
144166 97.038 -77.975 47.221
144167 97.058 -77.980 47.198
144168 97.079 -77.984 47.177
144169 97.099 -77.989 47.158
144170 97.119 -77.993 47.139
144171 97.140 -78.008 47.121
144172 97.161 -78.025 47.104
144173 97.179 -78.037 47.087
144174 97.189 -78.034 47.072
144175 97.199 -78.031 47.057
144176 97.209 -78.028 47.042
144177 97.219 -78.025 47.028
144178 97.229 -78.022 47.015
144179 97.246 -78.013 47.003
144180 97.271 -78.000 46.991
144181 97.293 -78.002 46.980
144182 97.315 -78.015 46.970
144183 97.336 -78.028 46.959
144184 97.357 -78.042 46.948
144185 97.378 -78.055 46.937
144186 97.405 -78.056 46.929
144187 97.434 -78.056 46.919
144188 97.462 -78.056 46.910
144189 97.470 -78.062 46.902
144190 97.469 -78.076 46.890
144191 97.494 -78.101 46.875
144192 97.517 -78.125 46.859
144193 97.541 -78.150 46.842
144194 97.565 -78.174 46.823
144195 97.587 -78.194 46.805
144196 97.602 -78.197 46.785
144197 97.617 -78.200 46.764
144198 97.632 -78.203 46.741
144199 97.663 -78.211 46.718
144200 97.710 -78.222 46.696
144201 97.756 -78.233 46.673
144202 97.803 -78.244 46.651
144203 97.850 -78.256 46.629
144204 97.896 -78.267 46.606
144205 97.943 -78.278 46.584
144206 97.989 -78.289 46.562
144207 98.036 -78.300 46.539
144208 98.081 -78.310 46.517
144209 98.123 -78.317 46.494
144210 98.164 -78.324 46.471
144211 98.206 -78.332 46.446
144212 98.247 -78.338 46.422
144213 98.271 -78.332 46.399
144214 98.291 -78.310 46.376
144215 98.311 -78.287 46.353
144216 98.331 -78.265 46.330
144217 98.343 -78.244 46.307
144218 98.346 -78.229 46.283
144219 98.326 -78.232 46.260
144220 98.307 -78.234 46.237
144221 98.353 -78.236 46.213
144222 98.405 -78.239 46.190
144223 98.455 -78.241 46.166
144224 98.480 -78.244 46.144
144225 98.501 -78.247 46.123
144226 98.522 -78.248 46.101
144227 98.550 -78.236 46.080
144228 98.609 -78.200 46.059
144229 98.640 -78.192 46.038
144230 98.651 -78.178 46.016
144231 98.661 -78.161 45.995
144232 98.671 -78.144 45.974
144233 98.681 -78.128 45.954
144234 98.691 -78.110 45.931
144235 98.698 -78.096 45.908
144236 98.700 -78.085 45.885
144237 98.702 -78.075 45.864
144238 98.704 -78.065 45.842
144239 98.707 -78.057 45.821
144240 98.741 -78.099 45.799
144241 98.744 -78.169 45.777
144242 98.751 -78.231 45.757
144243 98.778 -78.253 45.738
144244 98.804 -78.275 45.721
144245 98.831 -78.298 45.706
144246 98.858 -78.321 45.692
144247 98.878 -78.342 45.678
144248 98.901 -78.351 45.664
144249 98.926 -78.353 45.648
144250 98.951 -78.352 45.633
144251 98.976 -78.351 45.618
144252 99.001 -78.350 45.605
144253 99.025 -78.348 45.592
144254 99.051 -78.347 45.579
144255 99.075 -78.346 45.564
144256 99.092 -78.348 45.550
144257 99.107 -78.351 45.536
144258 99.123 -78.354 45.521
144259 99.155 -78.362 45.506
144260 99.187 -78.370 45.492
144261 99.220 -78.379 45.479
144262 99.252 -78.386 45.466
144263 99.285 -78.395 45.454
144264 99.317 -78.402 45.443
144265 99.349 -78.410 45.431
144266 99.382 -78.419 45.419
144267 99.414 -78.426 45.406
144268 99.447 -78.435 45.393
144269 99.479 -78.442 45.379
144270 99.511 -78.451 45.366
144271 99.553 -78.460 45.353
144272 99.606 -78.472 45.340
144273 99.659 -78.483 45.325
144274 99.712 -78.495 45.313
144275 99.766 -78.507 45.301
144276 99.794 -78.523 45.288
144277 99.813 -78.544 45.277
144278 99.832 -78.566 45.265
144279 99.859 -78.576 45.253
144280 99.875 -78.573 45.241
144281 99.884 -78.582 45.230
144282 99.893 -78.590 45.220
144283 99.900 -78.597 45.209
144284 99.907 -78.604 45.199
144285 99.914 -78.610 45.187
144286 99.921 -78.617 45.175
144287 99.928 -78.623 45.164
144288 99.935 -78.630 45.151
144289 99.942 -78.637 45.137
144290 99.951 -78.644 45.122
144291 99.982 -78.653 45.108
144292 100.014 -78.663 45.095
144293 100.046 -78.672 45.082
144294 100.081 -78.683 45.069
144295 100.122 -78.701 45.056
144296 100.159 -78.719 45.044
144297 100.197 -78.737 45.031
144298 100.235 -78.755 45.017
144299 100.273 -78.774 45.005
144300 100.297 -78.787 44.992
144301 100.315 -78.799 44.979
144302 100.333 -78.811 44.966
144303 100.345 -78.821 44.952
144304 100.355 -78.831 44.937
144305 100.365 -78.841 44.923
144306 100.376 -78.851 44.907
144307 100.389 -78.867 44.892
144308 100.403 -78.886 44.877
144309 100.413 -78.905 44.862
144310 100.424 -78.924 44.847
144311 100.434 -78.943 44.832
144312 100.445 -78.963 44.817
144313 100.456 -78.982 44.800
144314 100.470 -78.996 44.784
144315 100.486 -79.008 44.768
144316 100.502 -79.018 44.752
144317 100.521 -79.031 44.735
144318 100.554 -79.050 44.718
144319 100.586 -79.069 44.702
144320 100.619 -79.089 44.684
144321 100.651 -79.108 44.668
144322 100.684 -79.127 44.652
144323 100.716 -79.146 44.636
144324 100.749 -79.165 44.620
144325 100.781 -79.184 44.604
144568 92.747 -78.069 54.263
144569 92.766 -78.072 54.259
144570 92.784 -78.075 54.255
144571 92.803 -78.078 54.251
144572 92.821 -78.081 54.247
144573 92.839 -78.084 54.242
144574 92.858 -78.087 54.238
144575 92.876 -78.090 54.234
144576 92.894 -78.088 54.231
144577 92.910 -78.083 54.227
144578 92.926 -78.078 54.223
144579 92.943 -78.072 54.221
144580 92.959 -78.067 54.217
144581 92.976 -78.061 54.212
144582 92.992 -78.056 54.208
144583 93.009 -78.050 54.204
144584 93.025 -78.045 54.201
144585 93.042 -78.040 54.199
144586 93.058 -78.034 54.197
144587 93.075 -78.029 54.195
144588 93.095 -78.032 54.190
144589 93.116 -78.034 54.185
144590 93.136 -78.037 54.180
144591 93.156 -78.040 54.176
144592 93.177 -78.043 54.172
144593 93.197 -78.045 54.169
144594 93.217 -78.048 54.165
144595 93.237 -78.048 54.161
144596 93.258 -78.041 54.157
144597 93.285 -77.997 54.152
144598 93.323 -77.902 54.148
144599 93.311 -77.925 54.143
144600 93.299 -77.945 54.139
144601 93.298 -77.931 54.131
144602 93.297 -77.916 54.122
144603 93.296 -77.901 54.113
144604 93.295 -77.887 54.104
144605 93.294 -77.872 54.095
144606 93.307 -77.859 54.085
144607 93.323 -77.845 54.074
144608 93.354 -77.828 54.061
144609 93.393 -77.808 54.048
144610 93.404 -77.789 54.035
144611 93.414 -77.768 54.023
144612 93.425 -77.749 54.012
144613 93.435 -77.729 54.002
144614 93.445 -77.709 53.990
144615 93.459 -77.701 53.976
144616 93.473 -77.692 53.962
144617 93.487 -77.685 53.948
144618 93.501 -77.677 53.933
144619 93.517 -77.670 53.918
144620 93.533 -77.662 53.901
144621 93.548 -77.653 53.882
144622 93.563 -77.644 53.862
144623 93.578 -77.635 53.843
144624 93.593 -77.625 53.824
144625 93.608 -77.616 53.806
144626 93.623 -77.606 53.788
144627 93.632 -77.602 53.771
144628 93.640 -77.599 53.754
144629 93.648 -77.595 53.736
144630 93.655 -77.592 53.719
144631 93.666 -77.587 53.701
144632 93.677 -77.582 53.684
144633 93.688 -77.577 53.667
144634 93.698 -77.573 53.650
144635 93.709 -77.567 53.633
144636 93.720 -77.563 53.615
144637 93.730 -77.558 53.597
144638 93.744 -77.554 53.580
144639 93.758 -77.549 53.563
144640 93.772 -77.545 53.546
144641 93.785 -77.542 53.530
144642 93.799 -77.537 53.513
144643 93.813 -77.533 53.497
144644 93.827 -77.530 53.480
144645 93.842 -77.523 53.465
144646 93.857 -77.515 53.450
144647 93.872 -77.507 53.435
144648 93.888 -77.494 53.422
144649 93.903 -77.478 53.407
144650 93.920 -77.459 53.393
144651 93.938 -77.437 53.382
144652 93.949 -77.419 53.369
144653 93.956 -77.402 53.359
144654 93.959 -77.415 53.349
144655 93.982 -77.420 53.338
144656 94.010 -77.421 53.323
144657 94.039 -77.423 53.307
144658 94.067 -77.423 53.291
144659 94.095 -77.424 53.274
144660 94.123 -77.426 53.255
144661 94.151 -77.426 53.235
144662 94.178 -77.428 53.215
144663 94.188 -77.432 53.194
144664 94.188 -77.428 53.174
144665 94.189 -77.420 53.155
144666 94.190 -77.413 53.138
144667 94.192 -77.405 53.127
144668 94.193 -77.398 53.119
144669 94.194 -77.390 53.113
144670 94.196 -77.382 53.104
144671 94.197 -77.375 53.094
144672 94.198 -77.367 53.084
144673 94.201 -77.360 53.071
144674 94.219 -77.352 53.057
144675 94.224 -77.357 53.041
144676 94.226 -77.387 53.025
144677 94.228 -77.416 53.008
144678 94.230 -77.421 52.991
144679 94.230 -77.423 52.974
144680 94.231 -77.423 52.959
144681 94.225 -77.402 52.949
144682 94.232 -77.389 52.942
144683 94.269 -77.394 52.936
144684 94.306 -77.398 52.927
144685 94.342 -77.402 52.914
144686 94.359 -77.401 52.902
144687 94.367 -77.394 52.892
144688 94.376 -77.378 52.884
144689 94.385 -77.363 52.874
144690 94.388 -77.346 52.860
144691 94.385 -77.329 52.841
144692 94.381 -77.311 52.822
144693 94.377 -77.294 52.807
144694 94.387 -77.284 52.796
144695 94.405 -77.279 52.787
144696 94.423 -77.273 52.778
144697 94.442 -77.268 52.766
144698 94.460 -77.262 52.752
144699 94.479 -77.257 52.733
144700 94.500 -77.266 52.714
144701 94.523 -77.279 52.691
144702 94.546 -77.291 52.666
144703 94.568 -77.303 52.639
144704 94.585 -77.304 52.611
144705 94.601 -77.305 52.583
144706 94.618 -77.307 52.556
144707 94.634 -77.307 52.532
144708 94.667 -77.307 52.509
144709 94.738 -77.296 52.485
144710 94.763 -77.302 52.455
144711 94.785 -77.340 52.422
144712 94.815 -77.357 52.390
144713 94.849 -77.354 52.361
144714 94.881 -77.349 52.335
144715 94.899 -77.341 52.308
144716 94.918 -77.333 52.280
144717 94.937 -77.325 52.249
144718 94.956 -77.316 52.214
144719 94.975 -77.308 52.177
144720 94.994 -77.301 52.139
144721 95.013 -77.292 52.103
144722 95.026 -77.285 52.067
144723 95.032 -77.283 52.030
144724 95.039 -77.282 51.992
144725 95.046 -77.281 51.951
144726 95.053 -77.279 51.911
144727 95.060 -77.279 51.871
144728 95.067 -77.259 51.831
144729 95.074 -77.238 51.791
144730 95.090 -77.235 51.751
144731 95.116 -77.252 51.707
144732 95.134 -77.273 51.654
144733 95.136 -77.297 51.599
144734 95.139 -77.322 51.543
144735 95.149 -77.328 51.488
144736 95.166 -77.322 51.429
144737 95.182 -77.316 51.366
144738 95.199 -77.309 51.301
144739 95.215 -77.303 51.239
144740 95.232 -77.296 51.181
144741 95.248 -77.290 51.126
144742 95.265 -77.284 51.075
144743 95.285 -77.280 51.029
144744 95.313 -77.282 50.985
144745 95.340 -77.282 50.941
144746 95.371 -77.286 50.896
144747 95.401 -77.291 50.851
144748 95.431 -77.294 50.808
144749 95.462 -77.298 50.769
144750 95.493 -77.302 50.733
144751 95.523 -77.308 50.699
144752 95.549 -77.331 50.666
144753 95.576 -77.354 50.636
144754 95.602 -77.376 50.608
144755 95.628 -77.398 50.578
144756 95.633 -77.404 50.547
144757 95.634 -77.408 50.513
144758 95.636 -77.412 50.477
144759 95.638 -77.416 50.438
144760 95.640 -77.420 50.396
144761 95.642 -77.424 50.354
144762 95.644 -77.428 50.316
144763 95.646 -77.432 50.281
144764 95.648 -77.436 50.244
144765 95.650 -77.439 50.204
144766 95.660 -77.449 50.163
144767 95.677 -77.459 50.119
144768 95.696 -77.467 50.074
144769 95.716 -77.477 50.031
144770 95.740 -77.497 49.990
144771 95.763 -77.517 49.950
144772 95.787 -77.536 49.907
144773 95.811 -77.555 49.862
144774 95.834 -77.575 49.817
144775 95.858 -77.595 49.772
144776 95.874 -77.595 49.726
144777 95.889 -77.593 49.681
144778 95.904 -77.592 49.637
144779 95.919 -77.589 49.595
144780 95.935 -77.587 49.555
144781 95.950 -77.585 49.518
144782 95.965 -77.583 49.483
144783 95.973 -77.586 49.451
144784 95.977 -77.593 49.418
144785 95.980 -77.601 49.385
144786 95.983 -77.608 49.352
144787 95.987 -77.615 49.322
144788 95.990 -77.624 49.294
144789 95.989 -77.645 49.267
144790 96.000 -77.663 49.236
144791 96.012 -77.680 49.207
144792 96.023 -77.698 49.178
144793 96.034 -77.715 49.153
144794 96.048 -77.725 49.130
144795 96.061 -77.711 49.105
144796 96.078 -77.689 49.078
144797 96.094 -77.666 49.049
144798 96.096 -77.674 49.018
144799 96.097 -77.683 48.985
144800 96.106 -77.691 48.952
144801 96.114 -77.699 48.921
144802 96.106 -77.710 48.889
144803 96.090 -77.722 48.856
144804 96.102 -77.731 48.823
144805 96.113 -77.740 48.788
144806 96.125 -77.750 48.754
144807 96.138 -77.758 48.719
144808 96.166 -77.762 48.687
144809 96.194 -77.766 48.654
144810 96.210 -77.771 48.622
144811 96.225 -77.774 48.592
144812 96.240 -77.779 48.561
144813 96.254 -77.783 48.529
144814 96.269 -77.787 48.498
144815 96.283 -77.791 48.471
144816 96.298 -77.795 48.445
144817 96.308 -77.802 48.419
144818 96.318 -77.808 48.393
144819 96.336 -77.808 48.367
144820 96.354 -77.807 48.341
144821 96.363 -77.812 48.315
144822 96.367 -77.820 48.293
144823 96.371 -77.827 48.269
144824 96.376 -77.835 48.246
144825 96.380 -77.843 48.222
144826 96.384 -77.850 48.198
144827 96.389 -77.858 48.172
144828 96.393 -77.865 48.146
144829 96.403 -77.871 48.122
144830 96.424 -77.872 48.098
144831 96.444 -77.873 48.073
144832 96.464 -77.874 48.048
144833 96.484 -77.874 48.022
144834 96.504 -77.875 47.997
144835 96.524 -77.876 47.972
144836 96.544 -77.877 47.947
144837 96.564 -77.877 47.922
144838 96.582 -77.871 47.898
144839 96.594 -77.855 47.875
144840 96.607 -77.838 47.850
144841 96.619 -77.821 47.824
144842 96.634 -77.812 47.799
144843 96.653 -77.820 47.776
144844 96.671 -77.827 47.752
144845 96.689 -77.834 47.727
144846 96.708 -77.842 47.703
144847 96.727 -77.849 47.679
144848 96.745 -77.856 47.654
144849 96.763 -77.864 47.630
144850 96.781 -77.871 47.607
144851 96.796 -77.877 47.584
144852 96.812 -77.884 47.562
144853 96.829 -77.893 47.540
144854 96.852 -77.906 47.519
144855 96.873 -77.918 47.496
144856 96.882 -77.924 47.472
144857 96.891 -77.932 47.448
144858 96.900 -77.940 47.425
144859 96.909 -77.946 47.403
144860 96.918 -77.954 47.382
144861 96.927 -77.962 47.362
144862 96.936 -77.956 47.342
144863 96.950 -77.952 47.323
144864 96.973 -77.951 47.302
144865 96.997 -77.949 47.281
144866 97.017 -77.953 47.260
144867 97.037 -77.959 47.237
144868 97.058 -77.963 47.214
144869 97.078 -77.968 47.192
144870 97.098 -77.972 47.173
144871 97.119 -77.983 47.154
144872 97.140 -78.000 47.136
144873 97.161 -78.017 47.119
144874 97.181 -78.034 47.103
144875 97.192 -78.033 47.087
144876 97.203 -78.030 47.073
144877 97.213 -78.027 47.057
144878 97.223 -78.024 47.043
144879 97.235 -78.019 47.030
144880 97.259 -78.006 47.018
144881 97.284 -77.993 47.005
144882 97.306 -77.996 46.994
144883 97.327 -78.010 46.983
144884 97.349 -78.023 46.971
144885 97.370 -78.037 46.960
144886 97.391 -78.050 46.949
144887 97.413 -78.063 46.939
144888 97.438 -78.069 46.930
144889 97.466 -78.069 46.920
144890 97.477 -78.074 46.911
144891 97.470 -78.085 46.901
144892 97.469 -78.099 46.887
144893 97.493 -78.123 46.871
144894 97.516 -78.147 46.854
144895 97.540 -78.172 46.836
144896 97.563 -78.192 46.818
144897 97.578 -78.196 46.798
144898 97.592 -78.199 46.776
144899 97.614 -78.204 46.754
144900 97.661 -78.215 46.731
144901 97.707 -78.226 46.709
144902 97.754 -78.238 46.686
144903 97.800 -78.249 46.664
144904 97.847 -78.260 46.641
144905 97.893 -78.271 46.618
144906 97.940 -78.282 46.595
144907 97.987 -78.294 46.572
144908 98.033 -78.304 46.549
144909 98.080 -78.316 46.527
144910 98.123 -78.324 46.504
144911 98.165 -78.331 46.481
144912 98.207 -78.338 46.457
144913 98.230 -78.347 46.433
144914 98.244 -78.339 46.410
144915 98.264 -78.316 46.388
144916 98.284 -78.294 46.365
144917 98.304 -78.276 46.342
144918 98.323 -78.265 46.318
144919 98.343 -78.253 46.295
144920 98.363 -78.241 46.271
144921 98.375 -78.232 46.248
144922 98.363 -78.235 46.224
144923 98.414 -78.238 46.201
144924 98.428 -78.221 46.177
144925 98.472 -78.214 46.154
144926 98.515 -78.207 46.132
144927 98.559 -78.200 46.109
144928 98.603 -78.193 46.089
144929 98.643 -78.188 46.068
144930 98.662 -78.173 46.046
144931 98.663 -78.163 46.025
144932 98.665 -78.153 46.004
144933 98.667 -78.142 45.984
144934 98.669 -78.132 45.963
144935 98.670 -78.122 45.941
144936 98.672 -78.111 45.919
144937 98.674 -78.101 45.897
144938 98.676 -78.091 45.875
144939 98.675 -78.079 45.854
144940 98.684 -78.078 45.832
144941 98.724 -78.116 45.810
144942 98.762 -78.155 45.789
144943 98.765 -78.225 45.769
144944 98.790 -78.250 45.749
144945 98.816 -78.273 45.732
144946 98.840 -78.295 45.717
144947 98.848 -78.314 45.702
144948 98.865 -78.326 45.688
144949 98.896 -78.327 45.674
144950 98.920 -78.340 45.661
144951 98.944 -78.347 45.647
144952 98.969 -78.347 45.632
144953 98.994 -78.345 45.618
144954 99.019 -78.344 45.604
144955 99.044 -78.343 45.590
144956 99.069 -78.342 45.576
144957 99.083 -78.345 45.561
144958 99.098 -78.347 45.547
144959 99.129 -78.355 45.532
144960 99.161 -78.363 45.518
144961 99.193 -78.371 45.505
144962 99.226 -78.379 45.492
144963 99.258 -78.388 45.479
144964 99.291 -78.395 45.467
144965 99.323 -78.404 45.454
144966 99.355 -78.411 45.442
144967 99.388 -78.420 45.431
144968 99.420 -78.428 45.418
144969 99.453 -78.435 45.404
144970 99.485 -78.444 45.390
144971 99.518 -78.451 45.377
144972 99.550 -78.460 45.364
144973 99.588 -78.469 45.350
144974 99.641 -78.480 45.336
144975 99.694 -78.492 45.323
144976 99.748 -78.504 45.311
144977 99.784 -78.525 45.298
144978 99.811 -78.552 45.287
144979 99.831 -78.574 45.275
144980 99.863 -78.575 45.263
144981 99.880 -78.567 45.253
144982 99.889 -78.576 45.242
144983 99.896 -78.582 45.232
144984 99.903 -78.589 45.223
144985 99.910 -78.596 45.214
144986 99.917 -78.603 45.202
144987 99.924 -78.610 45.190
144988 99.931 -78.617 45.177
144989 99.938 -78.623 45.163
144990 99.945 -78.630 45.150
144991 99.952 -78.636 45.135
144992 99.975 -78.645 45.120
144993 100.007 -78.655 45.107
144994 100.043 -78.665 45.094
144995 100.091 -78.680 45.080
144996 100.131 -78.697 45.067
144997 100.168 -78.715 45.054
144998 100.206 -78.733 45.041
144999 100.245 -78.752 45.028
145000 100.279 -78.769 45.015
145001 100.297 -78.781 45.001
145002 100.315 -78.793 44.987
145003 100.333 -78.805 44.973
145004 100.351 -78.816 44.959
145005 100.361 -78.826 44.945
145006 100.371 -78.836 44.930
145007 100.382 -78.846 44.915
145008 100.394 -78.858 44.900
145009 100.409 -78.878 44.884
145010 100.423 -78.897 44.869
145011 100.433 -78.916 44.854
145012 100.444 -78.936 44.839
145013 100.455 -78.955 44.823
145014 100.465 -78.974 44.806
145015 100.477 -78.991 44.790
145016 100.493 -79.002 44.774
145017 100.509 -79.013 44.758
145018 100.526 -79.024 44.740
145019 100.551 -79.040 44.723
145020 100.583 -79.059 44.707
145021 100.616 -79.078 44.690
145022 100.648 -79.097 44.673
145023 100.681 -79.117 44.656
145024 100.713 -79.136 44.640
145025 100.746 -79.155 44.625
145026 100.778 -79.174 44.609
145027 100.811 -79.193 44.593
145028 100.843 -79.212 44.577
145269 92.728 -78.025 54.260
145270 92.746 -78.028 54.256
145271 92.764 -78.031 54.252
145272 92.783 -78.034 54.248
145273 92.801 -78.037 54.245
145274 92.819 -78.040 54.240
145275 92.838 -78.043 54.235
145276 92.856 -78.046 54.232
145277 92.874 -78.049 54.229
145278 92.893 -78.052 54.225
145279 92.911 -78.054 54.222
145280 92.929 -78.057 54.220
145281 92.946 -78.056 54.216
145282 92.963 -78.050 54.212
145283 92.980 -78.045 54.208
145284 92.996 -78.039 54.205
145285 93.012 -78.034 54.203
145286 93.029 -78.028 54.202
145287 93.046 -78.023 54.200
145288 93.065 -78.023 54.198
145289 93.085 -78.025 54.194
145290 93.105 -78.028 54.191
145291 93.126 -78.031 54.187
145292 93.146 -78.034 54.183
145293 93.166 -78.037 54.180
145294 93.186 -78.039 54.178
145295 93.206 -78.040 54.174
145296 93.226 -78.039 54.170
145297 93.246 -78.038 54.167
145298 93.277 -77.999 54.163
145299 93.298 -77.953 54.159
145300 93.297 -77.940 54.155
145301 93.296 -77.925 54.150
145302 93.295 -77.911 54.144
145303 93.294 -77.896 54.136
145304 93.293 -77.881 54.127
145305 93.292 -77.867 54.119
145306 93.290 -77.852 54.109
145307 93.299 -77.838 54.098
145308 93.314 -77.824 54.087
145309 93.331 -77.812 54.073
145310 93.357 -77.795 54.061
145311 93.393 -77.775 54.050
145312 93.404 -77.755 54.039
145313 93.414 -77.736 54.028
145314 93.424 -77.716 54.017
145315 93.435 -77.696 54.006
145316 93.449 -77.688 53.993
145317 93.463 -77.680 53.980
145318 93.477 -77.673 53.966
145319 93.492 -77.665 53.953
145320 93.508 -77.658 53.938
145321 93.523 -77.650 53.922
145322 93.539 -77.641 53.905
145323 93.554 -77.632 53.887
145324 93.569 -77.623 53.868
145325 93.584 -77.614 53.850
145326 93.600 -77.604 53.832
145327 93.614 -77.595 53.815
145328 93.627 -77.586 53.798
145329 93.635 -77.583 53.781
145330 93.643 -77.580 53.764
145331 93.650 -77.577 53.745
145332 93.658 -77.573 53.727
145333 93.669 -77.569 53.709
145334 93.680 -77.564 53.692
145335 93.691 -77.559 53.675
145336 93.701 -77.554 53.657
145337 93.712 -77.549 53.639
145338 93.724 -77.545 53.621
145339 93.739 -77.541 53.603
145340 93.752 -77.536 53.586
145341 93.766 -77.533 53.570
145342 93.780 -77.529 53.554
145343 93.794 -77.524 53.537
145344 93.809 -77.517 53.522
145345 93.824 -77.509 53.506
145346 93.839 -77.501 53.492
145347 93.854 -77.492 53.477
145348 93.870 -77.478 53.463
145349 93.885 -77.462 53.449
145350 93.901 -77.445 53.433
145351 93.919 -77.423 53.418
145352 93.937 -77.402 53.405
145353 93.949 -77.383 53.392
145354 93.957 -77.367 53.381
145355 93.957 -77.395 53.371
145356 93.977 -77.404 53.362
145357 94.005 -77.406 53.349
145358 94.033 -77.407 53.335
145359 94.061 -77.408 53.322
145360 94.089 -77.409 53.305
145361 94.118 -77.410 53.285
145362 94.145 -77.411 53.265
145363 94.165 -77.418 53.245
145364 94.171 -77.421 53.225
145365 94.168 -77.421 53.204
145366 94.169 -77.414 53.184
145367 94.170 -77.407 53.167
145368 94.171 -77.399 53.155
145369 94.172 -77.392 53.147
145370 94.174 -77.384 53.142
145371 94.175 -77.376 53.133
145372 94.182 -77.369 53.125
145373 94.199 -77.361 53.115
145374 94.217 -77.354 53.103
145375 94.235 -77.347 53.089
145376 94.233 -77.355 53.074
145377 94.234 -77.379 53.058
145378 94.236 -77.401 53.041
145379 94.237 -77.403 53.024
145380 94.238 -77.404 53.007
145381 94.239 -77.407 52.992
145382 94.233 -77.386 52.979
145383 94.250 -77.379 52.969
145384 94.287 -77.383 52.960
145385 94.324 -77.388 52.951
145386 94.342 -77.387 52.938
145387 94.350 -77.383 52.925
145388 94.358 -77.372 52.915
145389 94.367 -77.356 52.908
145390 94.377 -77.340 52.900
145391 94.386 -77.325 52.889
145392 94.386 -77.307 52.873
145393 94.385 -77.292 52.853
145394 94.403 -77.286 52.834
145395 94.422 -77.281 52.815
145396 94.440 -77.276 52.800
145397 94.458 -77.270 52.787
145398 94.477 -77.265 52.774
145399 94.495 -77.259 52.760
145400 94.513 -77.254 52.744
145401 94.533 -77.253 52.725
145402 94.555 -77.266 52.705
145403 94.578 -77.278 52.684
145404 94.596 -77.281 52.661
145405 94.612 -77.282 52.637
145406 94.629 -77.283 52.612
145407 94.645 -77.285 52.589
145408 94.662 -77.285 52.565
145409 94.678 -77.286 52.541
145410 94.717 -77.285 52.516
145411 94.759 -77.283 52.483
145412 94.785 -77.289 52.450
145413 94.809 -77.326 52.419
145414 94.843 -77.321 52.393
145415 94.868 -77.315 52.367
145416 94.887 -77.307 52.340
145417 94.906 -77.298 52.311
145418 94.925 -77.291 52.279
145419 94.944 -77.282 52.243
145420 94.962 -77.286 52.205
145421 94.979 -77.293 52.167
145422 94.997 -77.300 52.134
145423 95.010 -77.291 52.097
145424 95.017 -77.291 52.059
145425 95.024 -77.289 52.021
145426 95.031 -77.288 51.981
145427 95.037 -77.277 51.941
145428 95.044 -77.257 51.902
145429 95.051 -77.236 51.863
145430 95.057 -77.216 51.824
145431 95.083 -77.231 51.783
145432 95.110 -77.248 51.738
145433 95.136 -77.266 51.685
145434 95.139 -77.291 51.630
145435 95.141 -77.315 51.572
145436 95.154 -77.319 51.516
145437 95.170 -77.313 51.455
145438 95.186 -77.306 51.391
145439 95.203 -77.300 51.324
145440 95.219 -77.293 51.262
145441 95.236 -77.287 51.203
145442 95.252 -77.281 51.148
145443 95.268 -77.274 51.097
145444 95.285 -77.268 51.052
145445 95.302 -77.261 51.008
145446 95.326 -77.260 50.963
145447 95.353 -77.262 50.917
145448 95.383 -77.266 50.872
145449 95.414 -77.269 50.829
145450 95.445 -77.273 50.789
145451 95.475 -77.277 50.753
145452 95.504 -77.288 50.719
145453 95.530 -77.311 50.686
145454 95.556 -77.334 50.653
145455 95.583 -77.357 50.623
145456 95.609 -77.379 50.593
145457 95.635 -77.402 50.562
145458 95.638 -77.407 50.529
145459 95.640 -77.410 50.493
145460 95.642 -77.414 50.455
145461 95.644 -77.418 50.413
145462 95.646 -77.422 50.371
145463 95.648 -77.426 50.333
145464 95.650 -77.430 50.298
145465 95.652 -77.434 50.262
145466 95.658 -77.441 50.222
145467 95.673 -77.453 50.180
145468 95.687 -77.464 50.137
145469 95.706 -77.473 50.092
145470 95.725 -77.480 50.050
145471 95.745 -77.488 50.010
145472 95.764 -77.496 49.971
145473 95.786 -77.511 49.927
145474 95.809 -77.530 49.881
145475 95.833 -77.550 49.834
145476 95.856 -77.570 49.787
145477 95.880 -77.589 49.740
145478 95.897 -77.593 49.694
145479 95.912 -77.591 49.649
145480 95.927 -77.589 49.606
145481 95.943 -77.586 49.566
145482 95.958 -77.584 49.529
145483 95.971 -77.583 49.493
145484 95.975 -77.591 49.460
145485 95.978 -77.598 49.427
145486 95.981 -77.605 49.393
145487 95.985 -77.613 49.361
145488 95.988 -77.620 49.332
145489 95.992 -77.627 49.304
145490 95.992 -77.645 49.278
145491 96.004 -77.662 49.248
145492 96.015 -77.680 49.219
145493 96.027 -77.696 49.191
145494 96.042 -77.701 49.166
145495 96.058 -77.706 49.142
145496 96.071 -77.692 49.117
145497 96.084 -77.676 49.092
145498 96.098 -77.658 49.063
145499 96.102 -77.666 49.033
145500 96.111 -77.674 49.000
145501 96.120 -77.683 48.967
145502 96.127 -77.691 48.936
145503 96.110 -77.703 48.905
145504 96.093 -77.716 48.872
145505 96.099 -77.726 48.838
145506 96.111 -77.735 48.804
145507 96.122 -77.744 48.769
145508 96.148 -77.749 48.734
145509 96.175 -77.752 48.702
145510 96.200 -77.756 48.670
145511 96.215 -77.761 48.639
145512 96.229 -77.765 48.608
145513 96.243 -77.769 48.577
145514 96.258 -77.773 48.546
145515 96.272 -77.777 48.515
145516 96.285 -77.782 48.488
145517 96.294 -77.791 48.461
145518 96.302 -77.799 48.435
145519 96.313 -77.805 48.408
145520 96.331 -77.804 48.382
145521 96.349 -77.803 48.357
145522 96.367 -77.802 48.332
145523 96.385 -77.802 48.309
145524 96.398 -77.804 48.286
145525 96.403 -77.812 48.263
145526 96.407 -77.819 48.239
145527 96.411 -77.827 48.215
145528 96.416 -77.834 48.190
145529 96.419 -77.843 48.165
145530 96.410 -77.863 48.140
145531 96.424 -77.867 48.115
145532 96.444 -77.868 48.090
145533 96.464 -77.868 48.066
145534 96.484 -77.869 48.040
145535 96.505 -77.870 48.015
145536 96.525 -77.871 47.990
145537 96.545 -77.871 47.964
145538 96.565 -77.872 47.940
145539 96.583 -77.868 47.916
145540 96.595 -77.851 47.891
145541 96.608 -77.834 47.867
145542 96.621 -77.818 47.842
145543 96.633 -77.802 47.818
145544 96.651 -77.805 47.794
145545 96.669 -77.812 47.769
145546 96.688 -77.819 47.745
145547 96.707 -77.826 47.720
145548 96.725 -77.834 47.696
145549 96.744 -77.841 47.671
145550 96.762 -77.848 47.647
145551 96.778 -77.855 47.622
145552 96.793 -77.862 47.599
145553 96.811 -77.871 47.577
145554 96.833 -77.884 47.555
145555 96.856 -77.896 47.534
145556 96.877 -77.908 47.511
145557 96.888 -77.916 47.487
145558 96.897 -77.924 47.463
145559 96.906 -77.931 47.440
145560 96.915 -77.938 47.419
145561 96.925 -77.946 47.399
145562 96.934 -77.949 47.380
145563 96.942 -77.943 47.360
145564 96.950 -77.936 47.341
145565 96.972 -77.934 47.320
145566 96.996 -77.934 47.299
145567 97.017 -77.937 47.277
145568 97.037 -77.942 47.254
145569 97.058 -77.946 47.231
145570 97.078 -77.951 47.209
145571 97.098 -77.958 47.190
145572 97.119 -77.975 47.171
145573 97.140 -77.992 47.152
145574 97.161 -78.009 47.135
145575 97.181 -78.026 47.119
145576 97.196 -78.032 47.103
145577 97.207 -78.029 47.089
145578 97.217 -78.026 47.073
145579 97.227 -78.023 47.059
145580 97.248 -78.012 47.046
145581 97.273 -77.999 47.034
145582 97.297 -77.985 47.020
145583 97.319 -77.991 47.008
145584 97.341 -78.005 46.997
145585 97.362 -78.018 46.985
145586 97.383 -78.031 46.973
145587 97.404 -78.045 46.961
145588 97.425 -78.058 46.950
145589 97.446 -78.072 46.940
145590 97.470 -78.080 46.930
145591 97.484 -78.084 46.920
145592 97.477 -78.096 46.910
145593 97.470 -78.107 46.897
145594 97.468 -78.121 46.882
145595 97.492 -78.145 46.864
145596 97.516 -78.169 46.848
145597 97.538 -78.191 46.830
145598 97.553 -78.194 46.810
145599 97.568 -78.197 46.789
145600 97.611 -78.208 46.766
145601 97.658 -78.219 46.744
145602 97.705 -78.231 46.721
145603 97.751 -78.241 46.698
145604 97.798 -78.253 46.675
145605 97.845 -78.264 46.652
145606 97.891 -78.275 46.628
145607 97.938 -78.286 46.605
145608 97.984 -78.297 46.581
145609 98.031 -78.309 46.557
145610 98.077 -78.319 46.534
145611 98.124 -78.331 46.511
145612 98.166 -78.338 46.488
145613 98.196 -78.347 46.465
145614 98.202 -78.357 46.443
145615 98.217 -78.346 46.422
145616 98.237 -78.332 46.399
145617 98.256 -78.320 46.376
145618 98.276 -78.309 46.352
145619 98.296 -78.297 46.330
145620 98.316 -78.282 46.307
145621 98.338 -78.263 46.283
145622 98.359 -78.247 46.259
145623 98.380 -78.235 46.235
145624 98.367 -78.210 46.211
145625 98.393 -78.207 46.187
145626 98.435 -78.204 46.165
145627 98.477 -78.202 46.142
145628 98.519 -78.199 46.119
145629 98.561 -78.196 46.098
145630 98.602 -78.193 46.077
145631 98.636 -78.189 46.056
145632 98.637 -78.178 46.035
145633 98.639 -78.168 46.014
145634 98.641 -78.158 45.993
145635 98.643 -78.147 45.972
145636 98.644 -78.138 45.950
145637 98.646 -78.127 45.929
145638 98.648 -78.116 45.908
145639 98.647 -78.105 45.886
145640 98.646 -78.093 45.865
145641 98.662 -78.100 45.843
145642 98.702 -78.137 45.822
145643 98.742 -78.173 45.801
145644 98.782 -78.211 45.780
145645 98.801 -78.247 45.761
145646 98.810 -78.266 45.743
145647 98.818 -78.285 45.728
145648 98.826 -78.304 45.712
145649 98.858 -78.304 45.698
145650 98.890 -78.304 45.685
145651 98.914 -78.316 45.672
145652 98.938 -78.329 45.659
145653 98.962 -78.342 45.645
145654 98.988 -78.341 45.631
145655 99.012 -78.341 45.616
145656 99.037 -78.339 45.602
145657 99.060 -78.339 45.588
145658 99.075 -78.341 45.573
145659 99.102 -78.348 45.558
145660 99.135 -78.356 45.544
145661 99.167 -78.364 45.532
145662 99.199 -78.373 45.519
145663 99.232 -78.380 45.506
145664 99.264 -78.388 45.493
145665 99.297 -78.397 45.479
145666 99.329 -78.404 45.465
145667 99.362 -78.413 45.453
145668 99.394 -78.420 45.442
145669 99.426 -78.429 45.428
145670 99.459 -78.437 45.414
145671 99.491 -78.445 45.400
145672 99.524 -78.453 45.387
145673 99.556 -78.460 45.374
145674 99.588 -78.469 45.360
145675 99.623 -78.477 45.347
145676 99.677 -78.489 45.335
145677 99.729 -78.501 45.322
145678 99.767 -78.522 45.309
145679 99.793 -78.549 45.298
145680 99.834 -78.586 45.286
145681 99.867 -78.573 45.275
145682 99.886 -78.562 45.265
145683 99.893 -78.569 45.253
145684 99.900 -78.576 45.244
145685 99.907 -78.582 45.235
145686 99.914 -78.589 45.227
145687 99.921 -78.595 45.215
145688 99.928 -78.602 45.202
145689 99.935 -78.609 45.189
145690 99.942 -78.616 45.175
145691 99.949 -78.623 45.162
145692 99.956 -78.630 45.148
145693 99.968 -78.636 45.133
145694 100.003 -78.647 45.119
145695 100.052 -78.661 45.106
145696 100.101 -78.676 45.091
145697 100.140 -78.693 45.079
145698 100.178 -78.712 45.065
145699 100.216 -78.730 45.051
145700 100.254 -78.749 45.039
145701 100.279 -78.763 45.024
145702 100.297 -78.774 45.009
145703 100.316 -78.786 44.994
145704 100.334 -78.798 44.981
145705 100.352 -78.809 44.966
145706 100.367 -78.821 44.952
145707 100.378 -78.830 44.937
145708 100.388 -78.840 44.922
145709 100.398 -78.851 44.907
145710 100.413 -78.870 44.891
145711 100.428 -78.890 44.876
145712 100.443 -78.908 44.861
145713 100.454 -78.927 44.846
145714 100.464 -78.947 44.830
145715 100.475 -78.966 44.814
145716 100.485 -78.986 44.797
145717 100.501 -78.997 44.780
145718 100.517 -79.009 44.763
145719 100.534 -79.019 44.746
145720 100.550 -79.031 44.729
145721 100.581 -79.049 44.712
145722 100.614 -79.068 44.696
145723 100.646 -79.087 44.678
145724 100.679 -79.106 44.662
145725 100.711 -79.126 44.645
145726 100.744 -79.145 44.629
145727 100.776 -79.164 44.614
145728 100.809 -79.183 44.598
145729 100.855 -79.204 44.583
145730 100.908 -79.210 44.568
145731 100.960 -79.216 44.554
145971 92.726 -77.984 54.252
145972 92.744 -77.987 54.248
145973 92.763 -77.990 54.245
145974 92.781 -77.993 54.241
145975 92.799 -77.996 54.237
145976 92.817 -77.999 54.233
145977 92.836 -78.002 54.230
145978 92.854 -78.005 54.227
145979 92.873 -78.007 54.224
145980 92.891 -78.010 54.222
145981 92.909 -78.013 54.219
145982 92.928 -78.016 54.216
145983 92.946 -78.019 54.212
145984 92.965 -78.022 54.209
145985 92.983 -78.025 54.207
145986 93.000 -78.023 54.206
145987 93.016 -78.017 54.205
145988 93.034 -78.014 54.204
145989 93.054 -78.017 54.202
145990 93.074 -78.019 54.199
145991 93.094 -78.019 54.197
145992 93.114 -78.018 54.193
145993 93.134 -78.018 54.191
145994 93.153 -78.016 54.188
145995 93.174 -78.015 54.186
145996 93.193 -78.015 54.182
145997 93.213 -78.013 54.179
145998 93.232 -78.012 54.176
145999 93.266 -77.964 54.173
146000 93.287 -77.934 54.170
146001 93.294 -77.920 54.166
146002 93.292 -77.906 54.162
146003 93.291 -77.890 54.155
146004 93.290 -77.876 54.147
146005 93.289 -77.862 54.139
146006 93.288 -77.846 54.131
146007 93.287 -77.832 54.121
146008 93.290 -77.818 54.111
146009 93.306 -77.805 54.099
146010 93.322 -77.791 54.086
146011 93.338 -77.777 54.074
146012 93.361 -77.762 54.063
146013 93.393 -77.743 54.053
146014 93.404 -77.723 54.043
146015 93.414 -77.703 54.032
146016 93.424 -77.683 54.021
146017 93.438 -77.675 54.008
146018 93.452 -77.667 53.996
146019 93.467 -77.660 53.984
146020 93.483 -77.652 53.972
146021 93.499 -77.645 53.959
146022 93.514 -77.637 53.946
146023 93.530 -77.630 53.930
146024 93.545 -77.620 53.913
146025 93.560 -77.611 53.894
146026 93.575 -77.602 53.877
146027 93.590 -77.592 53.861
146028 93.606 -77.583 53.844
146029 93.621 -77.573 53.828
146030 93.630 -77.568 53.811
146031 93.638 -77.565 53.794
146032 93.645 -77.562 53.774
146033 93.653 -77.558 53.755
146034 93.661 -77.555 53.737
146035 93.673 -77.550 53.718
146036 93.683 -77.545 53.701
146037 93.694 -77.540 53.682
146038 93.705 -77.536 53.664
146039 93.719 -77.532 53.645
146040 93.733 -77.528 53.628
146041 93.747 -77.523 53.611
146042 93.761 -77.519 53.595
146043 93.776 -77.511 53.579
146044 93.791 -77.503 53.563
146045 93.806 -77.495 53.548
146046 93.822 -77.486 53.535
146047 93.837 -77.477 53.521
146048 93.852 -77.462 53.508
146049 93.867 -77.446 53.494
146050 93.883 -77.430 53.478
146051 93.901 -77.410 53.462
146052 93.919 -77.389 53.447
146053 93.937 -77.367 53.433
146054 93.951 -77.348 53.418
146055 93.955 -77.347 53.405
146056 93.955 -77.376 53.394
146057 93.972 -77.389 53.384
146058 94.000 -77.390 53.373
146059 94.028 -77.392 53.361
146060 94.056 -77.392 53.349
146061 94.084 -77.394 53.334
146062 94.112 -77.395 53.315
146063 94.132 -77.401 53.295
146064 94.150 -77.407 53.275
146065 94.152 -77.410 53.254
146066 94.150 -77.410 53.233
146067 94.147 -77.409 53.212
146068 94.149 -77.401 53.196
146069 94.150 -77.394 53.182
146070 94.151 -77.386 53.172
146071 94.162 -77.379 53.168
146072 94.180 -77.371 53.161
146073 94.197 -77.363 53.153
146074 94.215 -77.356 53.143
146075 94.233 -77.348 53.133
146076 94.245 -77.346 53.120
146077 94.242 -77.355 53.105
146078 94.242 -77.372 53.090
146079 94.243 -77.382 53.073
146080 94.244 -77.384 53.056
146081 94.245 -77.386 53.040
146082 94.246 -77.388 53.025
146083 94.241 -77.370 53.010
146084 94.268 -77.369 52.996
146085 94.305 -77.373 52.984
146086 94.326 -77.373 52.974
146087 94.333 -77.370 52.960
146088 94.340 -77.365 52.948
146089 94.349 -77.350 52.937
146090 94.359 -77.334 52.930
146091 94.368 -77.318 52.923
146092 94.377 -77.302 52.914
146093 94.391 -77.291 52.901
146094 94.413 -77.287 52.883
146095 94.435 -77.282 52.863
146096 94.457 -77.278 52.842
146097 94.475 -77.273 52.823
146098 94.493 -77.267 52.807
146099 94.511 -77.262 52.792
146100 94.530 -77.257 52.778
146101 94.548 -77.251 52.763
146102 94.566 -77.245 52.746
146103 94.588 -77.252 52.728
146104 94.607 -77.258 52.709
146105 94.623 -77.259 52.690
146106 94.639 -77.260 52.669
146107 94.656 -77.261 52.646
146108 94.673 -77.262 52.623
146109 94.689 -77.263 52.599
146110 94.705 -77.264 52.572
146111 94.722 -77.266 52.545
146112 94.756 -77.265 52.513
146113 94.781 -77.271 52.482
146114 94.808 -77.277 52.452
146115 94.836 -77.288 52.425
146116 94.855 -77.281 52.399
146117 94.874 -77.277 52.372
146118 94.892 -77.285 52.342
146119 94.909 -77.291 52.309
146120 94.927 -77.298 52.272
146121 94.945 -77.305 52.234
146122 94.963 -77.312 52.198
146123 94.980 -77.319 52.162
146124 94.991 -77.307 52.125
146125 95.000 -77.298 52.086
146126 95.007 -77.295 52.049
146127 95.014 -77.275 52.009
146128 95.021 -77.254 51.970
146129 95.028 -77.234 51.932
146130 95.035 -77.213 51.893
146131 95.050 -77.210 51.853
146132 95.076 -77.227 51.811
146133 95.103 -77.244 51.765
146134 95.130 -77.262 51.712
146135 95.142 -77.284 51.657
146136 95.144 -77.308 51.598
146137 95.157 -77.310 51.539
146138 95.174 -77.303 51.475
146139 95.190 -77.297 51.409
146140 95.206 -77.291 51.344
146141 95.223 -77.284 51.282
146142 95.240 -77.278 51.225
146143 95.256 -77.271 51.170
146144 95.272 -77.265 51.118
146145 95.289 -77.259 51.073
146146 95.305 -77.252 51.029
146147 95.322 -77.246 50.984
146148 95.338 -77.240 50.937
146149 95.366 -77.241 50.892
146150 95.396 -77.244 50.849
146151 95.426 -77.248 50.808
146152 95.457 -77.252 50.772
146153 95.485 -77.269 50.738
146154 95.511 -77.291 50.704
146155 95.537 -77.314 50.669
146156 95.564 -77.337 50.637
146157 95.590 -77.360 50.605
146158 95.616 -77.382 50.576
146159 95.642 -77.405 50.543
146160 95.644 -77.409 50.508
146161 95.646 -77.413 50.469
146162 95.648 -77.417 50.427
146163 95.650 -77.420 50.388
146164 95.652 -77.424 50.351
146165 95.654 -77.428 50.317
146166 95.656 -77.432 50.279
146167 95.670 -77.445 50.239
146168 95.685 -77.456 50.197
146169 95.700 -77.468 50.155
146170 95.716 -77.479 50.110
146171 95.735 -77.486 50.069
146172 95.754 -77.494 50.029
146173 95.774 -77.502 49.991
146174 95.793 -77.510 49.946
146175 95.812 -77.517 49.900
146176 95.832 -77.525 49.852
146177 95.855 -77.545 49.803
146178 95.879 -77.564 49.755
146179 95.902 -77.584 49.707
146180 95.920 -77.590 49.662
146181 95.936 -77.588 49.617
146182 95.951 -77.586 49.577
146183 95.966 -77.583 49.540
146184 95.972 -77.589 49.504
146185 95.976 -77.595 49.470
146186 95.979 -77.603 49.436
146187 95.983 -77.610 49.403
146188 95.986 -77.617 49.371
146189 95.990 -77.625 49.341
146190 95.993 -77.632 49.314
146191 95.995 -77.645 49.289
146192 96.007 -77.662 49.261
146193 96.021 -77.672 49.232
146194 96.036 -77.677 49.203
146195 96.052 -77.683 49.177
146196 96.068 -77.688 49.152
146197 96.081 -77.672 49.127
146198 96.093 -77.656 49.103
146199 96.098 -77.653 49.076
146200 96.112 -77.659 49.046
146201 96.125 -77.666 49.014
146202 96.133 -77.674 48.982
146203 96.130 -77.684 48.951
146204 96.113 -77.697 48.919
146205 96.096 -77.709 48.887
146206 96.096 -77.720 48.853
146207 96.108 -77.729 48.819
146208 96.129 -77.735 48.784
146209 96.157 -77.739 48.749
146210 96.184 -77.743 48.717
146211 96.203 -77.747 48.686
146212 96.218 -77.751 48.654
146213 96.232 -77.755 48.624
146214 96.247 -77.759 48.592
146215 96.261 -77.763 48.561
146216 96.271 -77.771 48.531
146217 96.279 -77.779 48.503
146218 96.288 -77.787 48.477
146219 96.296 -77.796 48.450
146220 96.309 -77.800 48.423
146221 96.326 -77.800 48.399
146222 96.344 -77.799 48.374
146223 96.362 -77.799 48.349
146224 96.380 -77.798 48.326
146225 96.398 -77.797 48.303
146226 96.415 -77.797 48.280
146227 96.434 -77.796 48.256
146228 96.439 -77.803 48.232
146229 96.443 -77.811 48.208
146230 96.439 -77.826 48.183
146231 96.427 -77.848 48.158
146232 96.432 -77.858 48.133
146233 96.445 -77.862 48.108
146234 96.465 -77.863 48.084
146235 96.485 -77.864 48.059
146236 96.505 -77.865 48.034
146237 96.525 -77.865 48.008
146238 96.545 -77.866 47.983
146239 96.566 -77.867 47.959
146240 96.584 -77.864 47.934
146241 96.597 -77.848 47.909
146242 96.609 -77.831 47.885
146243 96.622 -77.815 47.861
146244 96.635 -77.798 47.837
146245 96.650 -77.789 47.812
146246 96.668 -77.796 47.786
146247 96.686 -77.804 47.762
146248 96.705 -77.811 47.738
146249 96.724 -77.818 47.714
146250 96.742 -77.826 47.689
146251 96.758 -77.833 47.664
146252 96.774 -77.839 47.639
146253 96.793 -77.849 47.616
146254 96.815 -77.862 47.594
146255 96.837 -77.874 47.572
146256 96.859 -77.886 47.551
146257 96.881 -77.899 47.527
146258 96.894 -77.908 47.503
146259 96.903 -77.915 47.479
146260 96.912 -77.922 47.457
146261 96.922 -77.930 47.436
146262 96.931 -77.937 47.417
146263 96.939 -77.937 47.398
146264 96.948 -77.929 47.378
146265 96.956 -77.922 47.358
146266 96.972 -77.918 47.337
146267 96.995 -77.918 47.315
146268 97.016 -77.921 47.293
146269 97.037 -77.925 47.271
146270 97.057 -77.930 47.249
146271 97.077 -77.934 47.227
146272 97.098 -77.950 47.208
146273 97.119 -77.967 47.189
146274 97.140 -77.984 47.170
146275 97.161 -78.001 47.152
146276 97.181 -78.018 47.137
146277 97.201 -78.031 47.122
146278 97.211 -78.028 47.106
146279 97.221 -78.025 47.090
146280 97.237 -78.018 47.076
146281 97.261 -78.004 47.063
146282 97.286 -77.991 47.050
146283 97.310 -77.978 47.036
146284 97.332 -77.986 47.023
146285 97.353 -78.000 47.012
146286 97.374 -78.012 46.999
146287 97.395 -78.026 46.987
146288 97.417 -78.040 46.974
146289 97.438 -78.053 46.962
146290 97.459 -78.066 46.950
146291 97.480 -78.080 46.939
146292 97.490 -78.095 46.928
146293 97.484 -78.107 46.916
146294 97.476 -78.119 46.903
146295 97.469 -78.130 46.888
146296 97.467 -78.144 46.872
146297 97.491 -78.168 46.857
146298 97.514 -78.190 46.840
146299 97.529 -78.193 46.821
146300 97.563 -78.201 46.801
146301 97.609 -78.213 46.779
146302 97.656 -78.223 46.756
146303 97.702 -78.235 46.733
146304 97.749 -78.246 46.710
146305 97.795 -78.257 46.687
146306 97.842 -78.268 46.663
146307 97.888 -78.279 46.639
146308 97.935 -78.291 46.615
146309 97.982 -78.301 46.590
146310 98.028 -78.313 46.567
146311 98.075 -78.324 46.545
146312 98.122 -78.335 46.522
146313 98.161 -78.345 46.498
146314 98.170 -78.344 46.475
146315 98.181 -78.339 46.454
146316 98.203 -78.321 46.432
146317 98.224 -78.302 46.411
146318 98.245 -78.289 46.388
146319 98.267 -78.278 46.363
146320 98.288 -78.266 46.340
146321 98.309 -78.254 46.318
146322 98.330 -78.243 46.293
146323 98.352 -78.232 46.270
146324 98.358 -78.213 46.246
146325 98.362 -78.200 46.221
146326 98.392 -78.202 46.199
146327 98.406 -78.210 46.175
146328 98.431 -78.213 46.152
146329 98.473 -78.211 46.129
146330 98.515 -78.208 46.108
146331 98.557 -78.205 46.087
146332 98.598 -78.202 46.065
146333 98.611 -78.194 46.045
146334 98.613 -78.184 46.023
146335 98.615 -78.173 46.001
146336 98.617 -78.163 45.980
146337 98.618 -78.153 45.959
146338 98.620 -78.143 45.938
146339 98.619 -78.131 45.917
146340 98.618 -78.119 45.896
146341 98.617 -78.106 45.875
146342 98.640 -78.122 45.855
146343 98.680 -78.158 45.835
146344 98.720 -78.195 45.813
146345 98.758 -78.227 45.792
146346 98.781 -78.238 45.773
146347 98.789 -78.257 45.754
146348 98.796 -78.276 45.738
146349 98.819 -78.283 45.723
146350 98.851 -78.282 45.708
146351 98.883 -78.282 45.696
146352 98.909 -78.291 45.682
146353 98.933 -78.304 45.669
146354 98.957 -78.318 45.656
146355 98.981 -78.331 45.642
146356 99.006 -78.336 45.629
146357 99.030 -78.335 45.615
146358 99.052 -78.335 45.600
146359 99.076 -78.341 45.586
146360 99.108 -78.349 45.571
146361 99.141 -78.357 45.557
146362 99.173 -78.365 45.544
146363 99.205 -78.373 45.531
146364 99.238 -78.382 45.518
146365 99.270 -78.389 45.505
146366 99.303 -78.398 45.491
146367 99.335 -78.406 45.477
146368 99.368 -78.413 45.464
146369 99.400 -78.422 45.452
146370 99.432 -78.429 45.439
146371 99.465 -78.438 45.425
146372 99.497 -78.446 45.411
146373 99.530 -78.454 45.398
146374 99.562 -78.462 45.384
146375 99.594 -78.470 45.371
146376 99.627 -78.478 45.358
146377 99.659 -78.486 45.346
146378 99.712 -78.498 45.334
146379 99.749 -78.519 45.322
146380 99.784 -78.554 45.310
146381 99.830 -78.595 45.298
146382 99.853 -78.556 45.287
146383 99.893 -78.557 45.276
146384 99.899 -78.564 45.265
146385 99.906 -78.571 45.255
146386 99.913 -78.577 45.246
146387 99.920 -78.584 45.238
146388 99.927 -78.591 45.226
146389 99.934 -78.597 45.214
146390 99.941 -78.604 45.200
146391 99.948 -78.611 45.186
146392 99.955 -78.617 45.173
146393 99.963 -78.623 45.159
146394 99.972 -78.630 45.145
146395 100.013 -78.643 45.131
146396 100.062 -78.658 45.117
146397 100.111 -78.672 45.103
146398 100.149 -78.690 45.090
146399 100.187 -78.708 45.076
146400 100.225 -78.727 45.062
146401 100.262 -78.745 45.049
146402 100.280 -78.756 45.034
146403 100.298 -78.768 45.019
146404 100.316 -78.780 45.004
146405 100.334 -78.791 44.990
146406 100.352 -78.803 44.975
146407 100.370 -78.814 44.960
146408 100.384 -78.825 44.945
146409 100.394 -78.835 44.929
146410 100.404 -78.846 44.914
146411 100.418 -78.861 44.899
146412 100.433 -78.881 44.884
146413 100.448 -78.901 44.869
146414 100.463 -78.920 44.853
146415 100.474 -78.940 44.837
146416 100.484 -78.959 44.820
146417 100.494 -78.977 44.804
146418 100.508 -78.992 44.787
146419 100.525 -79.003 44.769
146420 100.541 -79.015 44.752
146421 100.557 -79.025 44.735
146422 100.578 -79.039 44.718
146423 100.611 -79.059 44.701
146424 100.643 -79.078 44.684
146425 100.676 -79.096 44.668
146426 100.708 -79.115 44.651
146427 100.741 -79.134 44.635
146428 100.773 -79.154 44.619
146429 100.806 -79.173 44.603
146430 100.861 -79.187 44.589
146431 100.913 -79.194 44.574
146432 100.965 -79.200 44.559
146433 101.017 -79.206 44.545
146434 101.069 -79.212 44.532
146672 92.706 -77.940 54.248
146673 92.724 -77.943 54.244
146674 92.742 -77.946 54.241
146675 92.761 -77.949 54.237
146676 92.779 -77.952 54.234
146677 92.798 -77.955 54.230
146678 92.816 -77.958 54.228
146679 92.834 -77.960 54.224
146680 92.853 -77.963 54.222
146681 92.871 -77.966 54.220
146682 92.889 -77.969 54.218
146683 92.908 -77.972 54.215
146684 92.926 -77.975 54.212
146685 92.944 -77.978 54.209
146686 92.963 -77.982 54.207
146687 92.983 -77.990 54.206
146688 93.003 -77.996 54.206
146689 93.021 -77.996 54.206
146690 93.042 -77.996 54.204
146691 93.061 -77.994 54.202
146692 93.081 -77.993 54.200
146693 93.100 -77.993 54.199
146694 93.121 -77.992 54.196
146695 93.140 -77.990 54.194
146696 93.160 -77.990 54.192
146697 93.180 -77.989 54.190
146698 93.200 -77.987 54.187
146699 93.220 -77.985 54.184
146700 93.256 -77.930 54.182
146701 93.269 -77.912 54.179
146702 93.280 -77.899 54.176
146703 93.289 -77.885 54.173
146704 93.288 -77.871 54.165
146705 93.287 -77.856 54.158
146706 93.286 -77.841 54.150
146707 93.285 -77.827 54.142
146708 93.284 -77.812 54.132
146709 93.283 -77.797 54.121
146710 93.298 -77.784 54.110
146711 93.313 -77.771 54.097
146712 93.330 -77.757 54.086
146713 93.345 -77.744 54.076
146714 93.365 -77.730 54.067
146715 93.393 -77.710 54.058
146716 93.403 -77.690 54.048
146717 93.414 -77.671 54.037
146718 93.428 -77.663 54.026
146719 93.442 -77.655 54.014
146720 93.458 -77.648 54.003
146721 93.474 -77.640 53.992
146722 93.490 -77.633 53.980
146723 93.505 -77.625 53.968
146724 93.521 -77.617 53.954
146725 93.537 -77.608 53.939
146726 93.552 -77.599 53.924
146727 93.567 -77.589 53.908
146728 93.581 -77.580 53.892
146729 93.597 -77.570 53.875
146730 93.619 -77.558 53.861
146731 93.626 -77.559 53.845
146732 93.629 -77.561 53.827
146733 93.641 -77.547 53.808
146734 93.648 -77.544 53.788
146735 93.656 -77.540 53.768
146736 93.664 -77.536 53.748
146737 93.676 -77.532 53.729
146738 93.686 -77.527 53.710
146739 93.700 -77.523 53.690
146740 93.713 -77.519 53.672
146741 93.728 -77.513 53.654
146742 93.743 -77.505 53.637
146743 93.758 -77.497 53.621
146744 93.773 -77.489 53.605
146745 93.788 -77.480 53.589
146746 93.803 -77.473 53.575
146747 93.819 -77.461 53.563
146748 93.834 -77.445 53.550
146749 93.850 -77.430 53.538
146750 93.865 -77.414 53.524
146751 93.882 -77.396 53.507
146752 93.900 -77.375 53.491
146753 93.918 -77.354 53.475
146754 93.936 -77.332 53.462
146755 93.951 -77.312 53.446
146756 93.954 -77.328 53.432
146757 93.954 -77.357 53.418
146758 93.967 -77.373 53.406
146759 93.994 -77.375 53.396
146760 94.023 -77.376 53.385
146761 94.051 -77.377 53.373
146762 94.079 -77.378 53.358
146763 94.099 -77.384 53.341
146764 94.118 -77.391 53.323
146765 94.136 -77.398 53.304
146766 94.135 -77.398 53.282
146767 94.132 -77.398 53.261
146768 94.129 -77.398 53.240
146769 94.128 -77.395 53.223
146770 94.143 -77.388 53.207
146771 94.160 -77.380 53.195
146772 94.178 -77.373 53.187
146773 94.196 -77.365 53.181
146774 94.213 -77.358 53.175
146775 94.231 -77.351 53.169
146776 94.249 -77.343 53.160
146777 94.254 -77.346 53.148
146778 94.252 -77.355 53.134
146779 94.250 -77.362 53.120
146780 94.250 -77.363 53.103
146781 94.251 -77.366 53.087
146782 94.252 -77.368 53.072
146783 94.253 -77.370 53.057
146784 94.250 -77.354 53.040
146785 94.286 -77.359 53.025
146786 94.309 -77.360 53.010
146787 94.316 -77.356 52.994
146788 94.324 -77.352 52.981
146789 94.332 -77.343 52.969
146790 94.341 -77.327 52.959
146791 94.350 -77.312 52.950
146792 94.360 -77.297 52.942
146793 94.373 -77.285 52.936
146794 94.389 -77.275 52.927
146795 94.411 -77.270 52.911
146796 94.434 -77.266 52.893
146797 94.456 -77.263 52.873
146798 94.479 -77.259 52.853
146799 94.501 -77.254 52.834
146800 94.524 -77.251 52.817
146801 94.546 -77.247 52.801
146802 94.568 -77.243 52.786
146803 94.591 -77.238 52.771
146804 94.611 -77.235 52.755
146805 94.621 -77.235 52.738
146806 94.636 -77.236 52.721
146807 94.651 -77.237 52.701
146808 94.666 -77.238 52.680
146809 94.680 -77.238 52.657
146810 94.695 -77.240 52.631
146811 94.710 -77.241 52.604
146812 94.724 -77.241 52.575
146813 94.738 -77.254 52.544
146814 94.765 -77.257 52.514
146815 94.794 -77.256 52.485
146816 94.824 -77.265 52.458
146817 94.835 -77.274 52.432
146818 94.854 -77.285 52.404
146819 94.873 -77.296 52.373
146820 94.892 -77.307 52.339
146821 94.910 -77.316 52.302
146822 94.928 -77.323 52.265
146823 94.946 -77.330 52.228
146824 94.963 -77.335 52.190
146825 94.972 -77.320 52.152
146826 94.974 -77.296 52.113
146827 94.991 -77.273 52.075
146828 94.998 -77.252 52.036
146829 95.004 -77.232 51.997
146830 95.012 -77.211 51.957
146831 95.018 -77.191 51.918
146832 95.044 -77.206 51.877
146833 95.070 -77.223 51.833
146834 95.097 -77.241 51.786
146835 95.123 -77.258 51.733
146836 95.144 -77.277 51.678
146837 95.147 -77.302 51.621
146838 95.161 -77.301 51.560
146839 95.178 -77.294 51.494
146840 95.194 -77.288 51.427
146841 95.210 -77.281 51.360
146842 95.227 -77.275 51.299
146843 95.243 -77.268 51.242
146844 95.260 -77.262 51.188
146845 95.276 -77.256 51.137
146846 95.293 -77.249 51.091
146847 95.309 -77.243 51.046
146848 95.326 -77.236 51.002
146849 95.342 -77.230 50.956
146850 95.358 -77.224 50.912
146851 95.379 -77.220 50.870
146852 95.408 -77.223 50.829
146853 95.439 -77.227 50.791
146854 95.466 -77.250 50.755
146855 95.492 -77.273 50.720
146856 95.518 -77.295 50.685
146857 95.544 -77.317 50.651
146858 95.571 -77.340 50.620
146859 95.597 -77.363 50.591
146860 95.623 -77.385 50.558
146861 95.648 -77.407 50.523
146862 95.650 -77.411 50.484
146863 95.652 -77.415 50.442
146864 95.654 -77.419 50.403
146865 95.656 -77.423 50.367
146866 95.658 -77.426 50.332
146867 95.668 -77.436 50.295
146868 95.683 -77.448 50.256
146869 95.697 -77.460 50.216
146870 95.712 -77.472 50.174
146871 95.727 -77.484 50.129
146872 95.745 -77.492 50.088
146873 95.764 -77.500 50.048
146874 95.783 -77.508 50.008
146875 95.803 -77.516 49.963
146876 95.822 -77.523 49.917
146877 95.841 -77.531 49.870
146878 95.861 -77.539 49.821
146879 95.880 -77.546 49.771
146880 95.901 -77.559 49.722
146881 95.924 -77.578 49.675
146882 95.944 -77.587 49.631
146883 95.959 -77.585 49.589
146884 95.970 -77.586 49.550
146885 95.974 -77.593 49.515
146886 95.977 -77.601 49.480
146887 95.981 -77.608 49.445
146888 95.984 -77.615 49.412
146889 95.987 -77.622 49.380
146890 95.991 -77.630 49.350
146891 95.994 -77.637 49.322
146892 95.999 -77.643 49.298
146893 96.015 -77.648 49.273
146894 96.031 -77.653 49.245
146895 96.046 -77.658 49.214
146896 96.062 -77.664 49.184
146897 96.078 -77.667 49.159
146898 96.091 -77.652 49.135
146899 96.100 -77.640 49.111
146900 96.097 -77.650 49.085
146901 96.103 -77.658 49.057
146902 96.118 -77.664 49.027
146903 96.133 -77.671 48.996
146904 96.130 -77.680 48.965
146905 96.117 -77.690 48.934
146906 96.100 -77.703 48.901
146907 96.093 -77.714 48.868
146908 96.111 -77.721 48.835
146909 96.138 -77.725 48.800
146910 96.166 -77.729 48.763
146911 96.192 -77.733 48.732
146912 96.207 -77.737 48.701
146913 96.222 -77.741 48.670
146914 96.236 -77.745 48.639
146915 96.249 -77.751 48.607
146916 96.257 -77.759 48.575
146917 96.265 -77.768 48.545
146918 96.273 -77.776 48.518
146919 96.282 -77.784 48.491
146920 96.290 -77.793 48.465
146921 96.304 -77.796 48.441
146922 96.321 -77.796 48.416
146923 96.339 -77.796 48.391
146924 96.357 -77.795 48.365
146925 96.375 -77.794 48.343
146926 96.393 -77.793 48.320
146927 96.411 -77.793 48.297
146928 96.429 -77.793 48.274
146929 96.446 -77.792 48.250
146930 96.465 -77.791 48.226
146931 96.459 -77.808 48.201
146932 96.447 -77.831 48.178
146933 96.443 -77.846 48.152
146934 96.453 -77.853 48.128
146935 96.466 -77.858 48.104
146936 96.486 -77.859 48.078
146937 96.506 -77.859 48.053
146938 96.526 -77.860 48.028
146939 96.546 -77.861 48.003
146940 96.566 -77.862 47.979
146941 96.586 -77.861 47.954
146942 96.598 -77.844 47.929
146943 96.611 -77.827 47.904
146944 96.623 -77.811 47.879
146945 96.637 -77.794 47.855
146946 96.649 -77.777 47.831
146947 96.667 -77.781 47.806
146948 96.685 -77.788 47.781
146949 96.703 -77.796 47.756
146950 96.722 -77.803 47.732
146951 96.740 -77.810 47.707
146952 96.755 -77.817 47.682
146953 96.774 -77.827 47.658
146954 96.797 -77.840 47.634
146955 96.818 -77.852 47.612
146956 96.840 -77.864 47.590
146957 96.863 -77.877 47.567
146958 96.885 -77.889 47.543
146959 96.901 -77.899 47.519
146960 96.910 -77.906 47.495
146961 96.919 -77.914 47.474
146962 96.928 -77.921 47.453
146963 96.937 -77.928 47.433
146964 96.945 -77.923 47.414
146965 96.954 -77.916 47.394
146966 96.962 -77.909 47.374
146967 96.971 -77.902 47.353
146968 96.995 -77.901 47.333
146969 97.016 -77.904 47.311
146970 97.036 -77.909 47.288
146971 97.056 -77.913 47.267
146972 97.077 -77.924 47.246
146973 97.098 -77.942 47.227
146974 97.119 -77.959 47.208
146975 97.140 -77.976 47.190
146976 97.161 -77.993 47.172
146977 97.181 -78.010 47.156
146978 97.202 -78.027 47.141
146979 97.214 -78.028 47.125
146980 97.226 -78.023 47.110
146981 97.250 -78.010 47.095
146982 97.275 -77.996 47.080
146983 97.299 -77.984 47.067
146984 97.323 -77.970 47.053
146985 97.345 -77.981 47.039
146986 97.366 -77.994 47.027
146987 97.387 -78.007 47.014
146988 97.408 -78.021 47.000
146989 97.430 -78.034 46.987
146990 97.451 -78.047 46.975
146991 97.473 -78.062 46.962
146992 97.496 -78.080 46.949
146993 97.490 -78.094 46.936
146994 97.485 -78.109 46.923
146995 97.480 -78.125 46.909
146996 97.476 -78.141 46.895
146997 97.469 -78.153 46.879
146998 97.466 -78.166 46.865
146999 97.490 -78.188 46.849
147000 97.513 -78.194 46.831
147001 97.560 -78.205 46.811
147002 97.606 -78.216 46.790
147003 97.653 -78.228 46.767
147004 97.700 -78.238 46.744
147005 97.746 -78.250 46.721
147006 97.793 -78.261 46.697
147007 97.840 -78.272 46.672
147008 97.888 -78.284 46.649
147009 97.938 -78.295 46.624
147010 97.987 -78.307 46.600
147011 98.036 -78.318 46.578
147012 98.074 -78.324 46.557
147013 98.110 -78.329 46.533
147014 98.132 -78.329 46.510
147015 98.145 -78.319 46.486
147016 98.174 -78.308 46.465
147017 98.195 -78.297 46.443
147018 98.216 -78.285 46.421
147019 98.238 -78.274 46.399
147020 98.259 -78.262 46.374
147021 98.280 -78.250 46.350
147022 98.301 -78.239 46.327
147023 98.323 -78.228 46.303
147024 98.344 -78.216 46.280
147025 98.333 -78.193 46.257
147026 98.363 -78.194 46.233
147027 98.393 -78.196 46.210
147028 98.423 -78.197 46.186
147029 98.443 -78.202 46.162
147030 98.432 -78.221 46.140
147031 98.469 -78.220 46.118
147032 98.511 -78.217 46.096
147033 98.553 -78.214 46.073
147034 98.585 -78.210 46.053
147035 98.587 -78.200 46.031
147036 98.588 -78.189 46.009
147037 98.590 -78.178 45.987
147038 98.592 -78.169 45.966
147039 98.591 -78.156 45.945
147040 98.590 -78.144 45.924
147041 98.589 -78.131 45.904
147042 98.588 -78.119 45.883
147043 98.617 -78.143 45.863
147044 98.657 -78.180 45.844
147045 98.697 -78.214 45.823
147046 98.720 -78.225 45.803
147047 98.744 -78.237 45.784
147048 98.767 -78.248 45.765
147049 98.780 -78.263 45.748
147050 98.812 -78.262 45.734
147051 98.844 -78.261 45.720
147052 98.875 -78.260 45.706
147053 98.904 -78.267 45.692
147054 98.928 -78.280 45.679
147055 98.952 -78.294 45.665
147056 98.976 -78.307 45.652
147057 99.000 -78.319 45.639
147058 99.024 -78.332 45.626
147059 99.049 -78.334 45.612
147060 99.082 -78.342 45.598
147061 99.114 -78.351 45.584
147062 99.147 -78.358 45.569
147063 99.179 -78.366 45.556
147064 99.212 -78.374 45.544
147065 99.244 -78.382 45.531
147066 99.276 -78.391 45.517
147067 99.309 -78.398 45.504
147068 99.341 -78.407 45.490
147069 99.374 -78.415 45.475
147070 99.406 -78.423 45.463
147071 99.438 -78.431 45.450
147072 99.471 -78.438 45.437
147073 99.503 -78.447 45.423
147074 99.536 -78.455 45.409
147075 99.568 -78.463 45.395
147076 99.600 -78.471 45.382
147077 99.633 -78.479 45.369
147078 99.665 -78.487 45.357
147079 99.698 -78.495 45.346
147080 99.738 -78.522 45.335
147081 99.780 -78.562 45.323
147082 99.825 -78.603 45.312
147083 99.859 -78.560 45.300
147084 99.902 -78.555 45.288
147085 99.909 -78.562 45.277
147086 99.916 -78.568 45.267
147087 99.923 -78.575 45.256
147088 99.929 -78.582 45.246
147089 99.938 -78.589 45.235
147090 99.947 -78.595 45.222
147091 99.957 -78.601 45.209
147092 99.966 -78.608 45.196
147093 99.975 -78.615 45.183
147094 99.984 -78.622 45.170
147095 99.993 -78.629 45.157
147096 100.023 -78.640 45.143
147097 100.072 -78.654 45.130
147098 100.120 -78.668 45.116
147099 100.159 -78.686 45.102
147100 100.197 -78.705 45.088
147101 100.234 -78.724 45.074
147102 100.262 -78.738 45.060
147103 100.280 -78.750 45.044
147104 100.298 -78.761 45.029
147105 100.316 -78.773 45.014
147106 100.334 -78.785 44.999
147107 100.352 -78.796 44.983
147108 100.370 -78.808 44.968
147109 100.388 -78.820 44.953
147110 100.400 -78.830 44.937
147111 100.410 -78.840 44.922
147112 100.423 -78.854 44.907
147113 100.438 -78.873 44.892
147114 100.453 -78.893 44.877
147115 100.468 -78.912 44.861
147116 100.483 -78.931 44.844
147117 100.493 -78.951 44.827
147118 100.504 -78.970 44.810
147119 100.516 -78.987 44.793
147120 100.532 -78.998 44.776
147121 100.549 -79.009 44.758
147122 100.565 -79.021 44.741
147123 100.581 -79.031 44.725
147124 100.609 -79.048 44.708
147125 100.641 -79.068 44.691
147126 100.674 -79.087 44.674
147127 100.706 -79.106 44.658
147128 100.739 -79.125 44.642
147129 100.771 -79.143 44.625
147130 100.814 -79.165 44.610
147131 100.866 -79.171 44.595
147132 100.918 -79.177 44.580
147133 100.970 -79.183 44.566
147134 101.022 -79.190 44.551
147135 101.074 -79.196 44.537
147136 101.126 -79.202 44.523
147137 101.178 -79.208 44.508
147373 92.686 -77.896 54.244
147374 92.704 -77.899 54.240
147375 92.723 -77.902 54.237
147376 92.741 -77.905 54.233
147377 92.760 -77.908 54.230
147378 92.778 -77.911 54.227
147379 92.796 -77.914 54.225
147380 92.814 -77.916 54.221
147381 92.833 -77.919 54.219
147382 92.852 -77.925 54.218
147383 92.872 -77.933 54.217
147384 92.892 -77.940 54.214
147385 92.912 -77.947 54.211
147386 92.932 -77.955 54.208
147387 92.952 -77.962 54.206
147388 92.972 -77.969 54.205
147389 92.992 -77.977 54.205
147390 93.012 -77.976 54.206
147391 93.028 -77.971 54.204
147392 93.048 -77.968 54.203
147393 93.068 -77.968 54.203
147394 93.087 -77.967 54.203
147395 93.107 -77.966 54.202
147396 93.127 -77.965 54.200
147397 93.147 -77.964 54.199
147398 93.166 -77.963 54.196
147399 93.186 -77.962 54.194
147400 93.209 -77.951 54.191
147401 93.239 -77.905 54.189
147402 93.251 -77.891 54.187
147403 93.262 -77.877 54.184
147404 93.273 -77.864 54.181
147405 93.284 -77.851 54.174
147406 93.284 -77.836 54.168
147407 93.283 -77.821 54.160
147408 93.282 -77.807 54.151
147409 93.281 -77.792 54.140
147410 93.280 -77.777 54.130
147411 93.289 -77.764 54.119
147412 93.305 -77.750 54.108
147413 93.321 -77.737 54.099
147414 93.337 -77.724 54.091
147415 93.353 -77.710 54.082
147416 93.369 -77.696 54.074
147417 93.393 -77.677 54.064
147418 93.403 -77.658 54.055
147419 93.417 -77.650 54.044
147420 93.433 -77.642 54.033
147421 93.449 -77.635 54.023
147422 93.465 -77.627 54.012
147423 93.481 -77.620 54.001
147424 93.496 -77.612 53.990
147425 93.512 -77.605 53.978
147426 93.528 -77.595 53.965
147427 93.543 -77.586 53.951
147428 93.558 -77.577 53.936
147429 93.577 -77.566 53.921
147430 93.599 -77.554 53.906
147431 93.621 -77.541 53.891
147432 93.627 -77.543 53.876
147433 93.628 -77.548 53.858
147434 93.630 -77.554 53.840
147435 93.636 -77.549 53.821
147436 93.648 -77.533 53.800
147437 93.659 -77.522 53.781
147438 93.667 -77.518 53.760
147439 93.680 -77.514 53.740
147440 93.695 -77.507 53.719
147441 93.710 -77.499 53.699
147442 93.725 -77.491 53.680
147443 93.741 -77.483 53.663
147444 93.756 -77.474 53.649
147445 93.771 -77.467 53.633
147446 93.786 -77.458 53.617
147447 93.801 -77.445 53.601
147448 93.817 -77.429 53.589
147449 93.832 -77.415 53.577
147450 93.848 -77.404 53.564
147451 93.864 -77.394 53.550
147452 93.877 -77.384 53.536
147453 93.897 -77.355 53.520
147454 93.916 -77.326 53.504
147455 93.935 -77.297 53.491
147456 93.952 -77.279 53.476
147457 93.952 -77.308 53.460
147458 93.952 -77.338 53.444
147459 93.961 -77.358 53.429
147460 93.989 -77.359 53.417
147461 94.017 -77.360 53.406
147462 94.045 -77.361 53.394
147463 94.066 -77.367 53.380
147464 94.085 -77.373 53.366
147465 94.104 -77.380 53.350
147466 94.120 -77.386 53.331
147467 94.117 -77.387 53.309
147468 94.114 -77.387 53.287
147469 94.111 -77.387 53.267
147470 94.131 -77.382 53.249
147471 94.153 -77.378 53.232
147472 94.173 -77.372 53.217
147473 94.193 -77.367 53.206
147474 94.211 -77.360 53.200
147475 94.229 -77.352 53.196
147476 94.247 -77.345 53.191
147477 94.264 -77.337 53.184
147478 94.263 -77.345 53.173
147479 94.262 -77.346 53.161
147480 94.261 -77.347 53.147
147481 94.259 -77.347 53.131
147482 94.258 -77.348 53.116
147483 94.259 -77.349 53.102
147484 94.260 -77.351 53.088
147485 94.268 -77.345 53.070
147486 94.292 -77.345 53.053
147487 94.299 -77.341 53.036
147488 94.307 -77.338 53.018
147489 94.314 -77.334 53.004
147490 94.323 -77.321 52.992
147491 94.332 -77.305 52.984
147492 94.345 -77.292 52.975
147493 94.358 -77.280 52.968
147494 94.370 -77.267 52.960
147495 94.388 -77.258 52.950
147496 94.410 -77.254 52.936
147497 94.432 -77.251 52.919
147498 94.455 -77.247 52.901
147499 94.477 -77.242 52.882
147500 94.500 -77.238 52.864
147501 94.522 -77.235 52.847
147502 94.545 -77.231 52.830
147503 94.567 -77.226 52.815
147504 94.589 -77.222 52.800
147505 94.609 -77.218 52.785
147506 94.620 -77.217 52.769
147507 94.629 -77.216 52.752
147508 94.639 -77.216 52.734
147509 94.649 -77.216 52.713
147510 94.660 -77.215 52.690
147511 94.669 -77.214 52.664
147512 94.682 -77.215 52.636
147513 94.697 -77.216 52.607
147514 94.718 -77.219 52.578
147515 94.751 -77.229 52.549
147516 94.785 -77.238 52.520
147517 94.813 -77.248 52.492
147518 94.824 -77.256 52.465
147519 94.835 -77.264 52.436
147520 94.846 -77.273 52.403
147521 94.862 -77.283 52.369
147522 94.881 -77.294 52.333
147523 94.900 -77.305 52.296
147524 94.919 -77.316 52.257
147525 94.943 -77.302 52.217
147526 94.929 -77.304 52.178
147527 94.931 -77.281 52.139
147528 94.949 -77.257 52.100
147529 94.970 -77.233 52.061
147530 94.988 -77.210 52.022
147531 94.995 -77.189 51.981
147532 95.011 -77.185 51.941
147533 95.037 -77.202 51.898
147534 95.063 -77.219 51.853
147535 95.090 -77.237 51.805
147536 95.117 -77.254 51.753
147537 95.143 -77.272 51.698
147538 95.149 -77.295 51.641
147539 95.165 -77.291 51.580
147540 95.181 -77.285 51.512
147541 95.198 -77.278 51.444
147542 95.214 -77.272 51.377
147543 95.231 -77.266 51.317
147544 95.247 -77.259 51.261
147545 95.264 -77.253 51.208
147546 95.280 -77.246 51.157
147547 95.297 -77.240 51.110
147548 95.313 -77.234 51.064
147549 95.329 -77.227 51.020
147550 95.346 -77.221 50.976
147551 95.362 -77.214 50.933
147552 95.379 -77.208 50.891
147553 95.395 -77.201 50.850
147554 95.420 -77.207 50.812
147555 95.446 -77.230 50.774
147556 95.473 -77.253 50.736
147557 95.499 -77.276 50.699
147558 95.525 -77.298 50.664
147559 95.551 -77.321 50.631
147560 95.578 -77.344 50.600
147561 95.604 -77.367 50.568
147562 95.630 -77.389 50.534
147563 95.654 -77.409 50.497
147564 95.656 -77.413 50.456
147565 95.658 -77.417 50.419
147566 95.660 -77.421 50.383
147567 95.666 -77.428 50.348
147568 95.681 -77.439 50.312
147569 95.695 -77.451 50.274
147570 95.710 -77.464 50.234
147571 95.725 -77.476 50.192
147572 95.739 -77.488 50.147
147573 95.755 -77.498 50.106
147574 95.774 -77.506 50.067
147575 95.794 -77.514 50.027
147576 95.813 -77.522 49.982
147577 95.832 -77.530 49.936
147578 95.851 -77.537 49.888
147579 95.871 -77.545 49.836
147580 95.890 -77.552 49.784
147581 95.909 -77.560 49.734
147582 95.928 -77.567 49.687
147583 95.948 -77.576 49.643
147584 95.967 -77.584 49.602
147585 95.972 -77.591 49.563
147586 95.975 -77.598 49.526
147587 95.980 -77.602 49.490
147588 95.984 -77.607 49.455
147589 95.988 -77.612 49.420
147590 95.992 -77.617 49.387
147591 95.997 -77.621 49.357
147592 96.001 -77.627 49.329
147593 96.009 -77.630 49.303
147594 96.025 -77.631 49.277
147595 96.040 -77.635 49.250
147596 96.056 -77.640 49.221
147597 96.072 -77.645 49.191
147598 96.088 -77.648 49.166
147599 96.100 -77.632 49.142
147600 96.099 -77.637 49.119
147601 96.096 -77.647 49.094
147602 96.095 -77.657 49.067
147603 96.110 -77.663 49.037
147604 96.125 -77.669 49.007
147605 96.117 -77.679 48.977
147606 96.107 -77.689 48.946
147607 96.097 -77.699 48.914
147608 96.092 -77.708 48.882
147609 96.120 -77.711 48.849
147610 96.148 -77.715 48.814
147611 96.175 -77.719 48.779
147612 96.196 -77.723 48.747
147613 96.210 -77.727 48.716
147614 96.225 -77.731 48.685
147615 96.235 -77.739 48.652
147616 96.243 -77.747 48.621
147617 96.251 -77.756 48.590
147618 96.259 -77.765 48.559
147619 96.267 -77.773 48.533
147620 96.276 -77.781 48.506
147621 96.284 -77.790 48.480
147622 96.299 -77.793 48.456
147623 96.316 -77.792 48.431
147624 96.334 -77.791 48.406
147625 96.352 -77.791 48.381
147626 96.370 -77.790 48.358
147627 96.388 -77.790 48.336
147628 96.406 -77.789 48.314
147629 96.425 -77.788 48.291
147630 96.451 -77.787 48.268
147631 96.471 -77.787 48.244
147632 96.478 -77.792 48.219
147633 96.466 -77.814 48.196
147634 96.455 -77.836 48.171
147635 96.465 -77.842 48.146
147636 96.474 -77.849 48.122
147637 96.486 -77.854 48.098
147638 96.507 -77.854 48.073
147639 96.527 -77.855 48.048
147640 96.547 -77.855 48.024
147641 96.567 -77.856 48.000
147642 96.587 -77.857 47.975
147643 96.600 -77.840 47.950
147644 96.612 -77.824 47.925
147645 96.625 -77.807 47.899
147646 96.638 -77.791 47.874
147647 96.650 -77.774 47.850
147648 96.665 -77.766 47.825
147649 96.683 -77.773 47.800
147650 96.702 -77.780 47.775
147651 96.721 -77.788 47.751
147652 96.736 -77.794 47.725
147653 96.756 -77.805 47.701
147654 96.778 -77.818 47.677
147655 96.800 -77.830 47.652
147656 96.822 -77.842 47.630
147657 96.844 -77.855 47.607
147658 96.866 -77.867 47.583
147659 96.888 -77.880 47.559
147660 96.907 -77.890 47.536
147661 96.916 -77.898 47.514
147662 96.925 -77.906 47.492
147663 96.934 -77.913 47.471
147664 96.943 -77.917 47.451
147665 96.951 -77.909 47.432
147666 96.960 -77.902 47.413
147667 96.968 -77.896 47.392
147668 96.976 -77.888 47.372
147669 96.994 -77.885 47.351
147670 97.016 -77.887 47.329
147671 97.036 -77.892 47.307
147672 97.056 -77.899 47.286
147673 97.077 -77.917 47.266
147674 97.098 -77.934 47.247
147675 97.119 -77.951 47.228
147676 97.140 -77.968 47.210
147677 97.161 -77.985 47.192
147678 97.181 -78.002 47.176
147679 97.202 -78.019 47.161
147680 97.218 -78.027 47.146
147681 97.239 -78.016 47.130
147682 97.263 -78.003 47.114
147683 97.288 -77.989 47.099
147684 97.312 -77.976 47.084
147685 97.336 -77.962 47.070
147686 97.358 -77.975 47.057
147687 97.379 -77.989 47.043
147688 97.400 -78.003 47.029
147689 97.422 -78.015 47.015
147690 97.443 -78.029 47.001
147691 97.466 -78.047 46.988
147692 97.489 -78.064 46.973
147693 97.501 -78.079 46.959
147694 97.490 -78.093 46.945
147695 97.485 -78.109 46.930
147696 97.480 -78.125 46.916
147697 97.476 -78.140 46.901
147698 97.471 -78.156 46.886
147699 97.467 -78.172 46.872
147700 97.465 -78.187 46.857
147701 97.512 -78.198 46.840
147702 97.561 -78.210 46.821
147703 97.610 -78.221 46.800
147704 97.659 -78.232 46.778
147705 97.709 -78.244 46.754
147706 97.757 -78.256 46.730
147707 97.807 -78.267 46.705
147708 97.856 -78.279 46.681
147709 97.905 -78.290 46.658
147710 97.945 -78.297 46.635
147711 97.982 -78.302 46.612
147712 98.018 -78.307 46.590
147713 98.054 -78.312 46.568
147714 98.083 -78.319 46.544
147715 98.100 -78.313 46.520
147716 98.122 -78.304 46.497
147717 98.166 -78.293 46.476
147718 98.187 -78.282 46.453
147719 98.209 -78.269 46.431
147720 98.230 -78.258 46.408
147721 98.251 -78.247 46.385
147722 98.273 -78.235 46.361
147723 98.294 -78.224 46.338
147724 98.315 -78.212 46.314
147725 98.313 -78.191 46.292
147726 98.334 -78.187 46.268
147727 98.364 -78.188 46.244
147728 98.393 -78.190 46.221
147729 98.423 -78.191 46.196
147730 98.453 -78.193 46.173
147731 98.481 -78.194 46.150
147732 98.470 -78.213 46.127
147733 98.464 -78.229 46.104
147734 98.507 -78.226 46.082
147735 98.549 -78.223 46.060
147736 98.561 -78.215 46.038
147737 98.563 -78.205 46.016
147738 98.564 -78.194 45.995
147739 98.563 -78.182 45.973
147740 98.562 -78.170 45.952
147741 98.561 -78.158 45.930
147742 98.560 -78.145 45.910
147743 98.559 -78.133 45.891
147744 98.595 -78.165 45.872
147745 98.635 -78.201 45.853
147746 98.659 -78.213 45.833
147747 98.683 -78.224 45.814
147748 98.707 -78.235 45.794
147749 98.742 -78.244 45.775
147750 98.782 -78.248 45.759
147751 98.811 -78.246 45.744
147752 98.840 -78.243 45.730
147753 98.869 -78.240 45.717
147754 98.898 -78.243 45.702
147755 98.925 -78.257 45.688
147756 98.952 -78.272 45.674
147757 98.979 -78.288 45.661
147758 99.006 -78.303 45.649
147759 99.033 -78.317 45.637
147760 99.050 -78.323 45.623
147761 99.081 -78.330 45.610
147762 99.113 -78.338 45.596
147763 99.145 -78.346 45.582
147764 99.178 -78.354 45.568
147765 99.210 -78.362 45.556
147766 99.242 -78.369 45.544
147767 99.275 -78.377 45.530
147768 99.307 -78.385 45.516
147769 99.339 -78.393 45.502
147770 99.371 -78.401 45.488
147771 99.404 -78.409 45.474
147772 99.436 -78.416 45.462
147773 99.469 -78.425 45.448
147774 99.501 -78.432 45.434
147775 99.534 -78.440 45.421
147776 99.566 -78.448 45.407
147777 99.598 -78.456 45.394
147778 99.630 -78.464 45.382
147779 99.662 -78.472 45.371
147780 99.693 -78.481 45.359
147781 99.711 -78.517 45.348
147782 99.761 -78.562 45.336
147783 99.810 -78.594 45.325
147784 99.853 -78.570 45.314
147785 99.902 -78.551 45.303
147786 99.923 -78.560 45.290
147787 99.932 -78.567 45.279
147788 99.941 -78.573 45.269
147789 99.950 -78.580 45.257
147790 99.960 -78.587 45.245
147791 99.969 -78.594 45.233
147792 99.977 -78.600 45.221
147793 99.986 -78.607 45.208
147794 99.995 -78.614 45.195
147795 100.004 -78.620 45.183
147796 100.013 -78.627 45.170
147797 100.033 -78.636 45.155
147798 100.081 -78.651 45.142
147799 100.130 -78.665 45.129
147800 100.168 -78.683 45.114
147801 100.206 -78.702 45.100
147802 100.244 -78.720 45.086
147803 100.262 -78.732 45.071
147804 100.280 -78.743 45.055
147805 100.298 -78.755 45.040
147806 100.317 -78.767 45.024
147807 100.335 -78.778 45.009
147808 100.352 -78.790 44.992
147809 100.370 -78.802 44.977
147810 100.388 -78.813 44.961
147811 100.406 -78.825 44.945
147812 100.416 -78.835 44.930
147813 100.427 -78.846 44.915
147814 100.442 -78.865 44.900
147815 100.458 -78.884 44.884
147816 100.473 -78.904 44.869
147817 100.488 -78.924 44.852
147818 100.503 -78.943 44.834
147819 100.513 -78.962 44.818
147820 100.524 -78.981 44.801
147821 100.540 -78.993 44.784
147822 100.556 -79.004 44.766
147823 100.572 -79.015 44.749
147824 100.588 -79.027 44.732
147825 100.606 -79.038 44.715
147826 100.638 -79.057 44.697
147827 100.671 -79.076 44.681
147828 100.703 -79.096 44.664
147829 100.736 -79.115 44.648
147830 100.768 -79.134 44.632
147831 100.819 -79.148 44.616
147832 100.871 -79.154 44.601
147833 100.923 -79.160 44.586
147834 100.975 -79.167 44.572
147835 101.027 -79.173 44.558
147836 101.079 -79.179 44.543
147837 101.131 -79.185 44.529
147838 101.183 -79.191 44.515
147839 101.236 -79.198 44.500
147840 101.287 -79.204 44.488
148074 92.666 -77.852 54.239
148075 92.685 -77.855 54.235
148076 92.703 -77.858 54.231
148077 92.721 -77.862 54.229
148078 92.741 -77.868 54.226
148079 92.761 -77.876 54.222
148080 92.781 -77.884 54.220
148081 92.801 -77.890 54.217
148082 92.821 -77.898 54.216
148083 92.841 -77.906 54.214
148084 92.861 -77.912 54.213
148085 92.881 -77.920 54.210
148086 92.901 -77.928 54.207
148087 92.921 -77.934 54.206
148088 92.941 -77.942 54.205
148089 92.961 -77.949 54.203
148090 92.981 -77.957 54.202
148091 93.002 -77.955 54.203
148092 93.019 -77.949 54.204
148093 93.036 -77.944 54.204
148094 93.054 -77.942 54.204
148095 93.074 -77.941 54.206
148096 93.094 -77.940 54.206
148097 93.114 -77.939 54.206
148098 93.133 -77.938 54.205
148099 93.153 -77.937 54.202
148100 93.173 -77.936 54.200
148101 93.199 -77.916 54.199
148102 93.221 -77.883 54.197
148103 93.232 -77.870 54.194
148104 93.243 -77.856 54.191
148105 93.254 -77.843 54.188
148106 93.266 -77.829 54.182
148107 93.277 -77.816 54.176
148108 93.280 -77.802 54.168
148109 93.279 -77.787 54.157
148110 93.278 -77.772 54.148
148111 93.277 -77.758 54.138
148112 93.281 -77.743 54.128
148113 93.296 -77.730 54.118
148114 93.312 -77.717 54.110
148115 93.328 -77.703 54.102
148116 93.344 -77.689 54.094
148117 93.360 -77.677 54.087
148118 93.376 -77.663 54.080
148119 93.393 -77.645 54.072
148120 93.408 -77.637 54.062
148121 93.424 -77.630 54.052
148122 93.440 -77.623 54.042
148123 93.455 -77.615 54.033
148124 93.472 -77.608 54.023
148125 93.487 -77.600 54.013
148126 93.503 -77.592 54.004
148127 93.518 -77.583 53.991
148128 93.535 -77.574 53.977
148129 93.557 -77.561 53.963
148130 93.579 -77.549 53.949
148131 93.601 -77.537 53.934
148132 93.623 -77.524 53.919
148133 93.623 -77.530 53.903
148134 93.617 -77.540 53.886
148135 93.617 -77.545 53.868
148136 93.628 -77.534 53.850
148137 93.638 -77.524 53.830
148138 93.649 -77.514 53.810
148139 93.657 -77.506 53.789
148140 93.672 -77.497 53.768
148141 93.688 -77.488 53.746
148142 93.705 -77.478 53.726
148143 93.722 -77.469 53.707
148144 93.738 -77.461 53.690
148145 93.753 -77.452 53.677
148146 93.768 -77.444 53.661
148147 93.784 -77.431 53.644
148148 93.799 -77.420 53.628
148149 93.816 -77.410 53.614
148150 93.832 -77.399 53.601
148151 93.847 -77.389 53.589
148152 93.863 -77.378 53.575
148153 93.872 -77.377 53.559
148154 93.892 -77.348 53.543
148155 93.911 -77.320 53.528
148156 93.930 -77.291 53.515
148157 93.945 -77.275 53.502
148158 93.950 -77.289 53.487
148159 93.951 -77.318 53.472
148160 93.956 -77.342 53.457
148161 93.984 -77.344 53.441
148162 94.012 -77.345 53.426
148163 94.034 -77.350 53.413
148164 94.052 -77.357 53.400
148165 94.071 -77.363 53.386
148166 94.090 -77.370 53.371
148167 94.102 -77.375 53.355
148168 94.099 -77.375 53.333
148169 94.097 -77.376 53.312
148170 94.101 -77.365 53.292
148171 94.130 -77.364 53.273
148172 94.158 -77.363 53.256
148173 94.178 -77.357 53.239
148174 94.198 -77.351 53.226
148175 94.217 -77.346 53.219
148176 94.237 -77.340 53.215
148177 94.257 -77.335 53.211
148178 94.273 -77.329 53.206
148179 94.273 -77.328 53.197
148180 94.272 -77.329 53.186
148181 94.271 -77.330 53.171
148182 94.270 -77.332 53.157
148183 94.269 -77.332 53.143
148184 94.268 -77.332 53.130
148185 94.267 -77.333 53.117
148186 94.275 -77.332 53.100
148187 94.282 -77.328 53.081
148188 94.290 -77.324 53.063
148189 94.297 -77.320 53.045
148190 94.305 -77.314 53.030
148191 94.316 -77.301 53.018
148192 94.329 -77.288 53.010
148193 94.342 -77.276 53.003
148194 94.355 -77.263 52.995
148195 94.368 -77.251 52.986
148196 94.386 -77.242 52.975
148197 94.408 -77.238 52.961
148198 94.431 -77.235 52.947
148199 94.453 -77.230 52.931
148200 94.476 -77.226 52.914
148201 94.498 -77.222 52.898
148202 94.521 -77.219 52.881
148203 94.543 -77.214 52.863
148204 94.565 -77.210 52.847
148205 94.585 -77.206 52.832
148206 94.606 -77.201 52.817
148207 94.619 -77.199 52.801
148208 94.628 -77.199 52.785
148209 94.638 -77.198 52.766
148210 94.648 -77.197 52.745
148211 94.659 -77.197 52.721
148212 94.668 -77.197 52.694
148213 94.678 -77.196 52.667
148214 94.688 -77.196 52.639
148215 94.708 -77.201 52.611
148216 94.742 -77.210 52.582
148217 94.775 -77.220 52.553
148218 94.803 -77.229 52.525
148219 94.814 -77.238 52.499
148220 94.825 -77.246 52.470
148221 94.835 -77.254 52.438
148222 94.846 -77.263 52.401
148223 94.857 -77.272 52.363
148224 94.871 -77.281 52.324
148225 94.890 -77.291 52.284
148226 94.919 -77.257 52.244
148227 94.901 -77.273 52.204
148228 94.887 -77.278 52.164
148229 94.914 -77.249 52.124
148230 94.935 -77.225 52.083
148231 94.957 -77.201 52.043
148232 94.980 -77.172 52.003
148233 95.005 -77.183 51.961
148234 95.031 -77.198 51.917
148235 95.057 -77.216 51.870
148236 95.084 -77.233 51.823
148237 95.110 -77.251 51.771
148238 95.136 -77.268 51.717
148239 95.153 -77.288 51.658
148240 95.169 -77.282 51.595
148241 95.185 -77.275 51.528
148242 95.202 -77.269 51.460
148243 95.218 -77.263 51.395
148244 95.235 -77.256 51.336
148245 95.251 -77.250 51.281
148246 95.267 -77.243 51.229
148247 95.284 -77.237 51.179
148248 95.300 -77.230 51.131
148249 95.316 -77.223 51.084
148250 95.332 -77.216 51.040
148251 95.348 -77.210 50.997
148252 95.364 -77.203 50.955
148253 95.380 -77.197 50.913
148254 95.396 -77.190 50.872
148255 95.415 -77.200 50.833
148256 95.436 -77.217 50.793
148257 95.459 -77.238 50.753
148258 95.482 -77.258 50.714
148259 95.506 -77.279 50.677
148260 95.532 -77.301 50.641
148261 95.559 -77.324 50.607
148262 95.585 -77.347 50.575
148263 95.611 -77.370 50.542
148264 95.637 -77.392 50.508
148265 95.660 -77.411 50.469
148266 95.662 -77.415 50.434
148267 95.664 -77.420 50.400
148268 95.678 -77.432 50.365
148269 95.693 -77.443 50.330
148270 95.707 -77.455 50.292
148271 95.722 -77.467 50.252
148272 95.737 -77.479 50.211
148273 95.751 -77.492 50.166
148274 95.766 -77.504 50.125
148275 95.784 -77.513 50.086
148276 95.803 -77.520 50.046
148277 95.823 -77.528 50.002
148278 95.842 -77.536 49.955
148279 95.861 -77.543 49.906
148280 95.880 -77.551 49.852
148281 95.900 -77.558 49.799
148282 95.919 -77.566 49.748
148283 95.938 -77.573 49.700
148284 95.953 -77.577 49.656
148285 95.969 -77.581 49.614
148286 95.976 -77.583 49.575
148287 95.983 -77.586 49.537
148288 95.987 -77.590 49.500
148289 95.991 -77.595 49.464
148290 95.995 -77.600 49.429
148291 96.000 -77.605 49.394
148292 96.004 -77.609 49.363
148293 96.009 -77.614 49.334
148294 96.019 -77.617 49.307
148295 96.035 -77.618 49.280
148296 96.050 -77.620 49.254
148297 96.066 -77.621 49.228
148298 96.082 -77.627 49.199
148299 96.097 -77.627 49.174
148300 96.102 -77.624 49.150
148301 96.098 -77.634 49.127
148302 96.095 -77.645 49.102
148303 96.091 -77.655 49.076
148304 96.101 -77.661 49.047
148305 96.115 -77.668 49.017
148306 96.105 -77.678 48.988
148307 96.095 -77.688 48.958
148308 96.091 -77.696 48.927
148309 96.109 -77.699 48.896
148310 96.129 -77.702 48.862
148311 96.157 -77.705 48.827
148312 96.184 -77.709 48.794
148313 96.199 -77.713 48.761
148314 96.212 -77.719 48.730
148315 96.221 -77.727 48.699
148316 96.229 -77.736 48.667
148317 96.237 -77.744 48.636
148318 96.245 -77.753 48.605
148319 96.253 -77.761 48.574
148320 96.262 -77.770 48.547
148321 96.270 -77.778 48.521
148322 96.278 -77.787 48.495
148323 96.294 -77.789 48.471
148324 96.311 -77.788 48.446
148325 96.329 -77.787 48.421
148326 96.347 -77.787 48.397
148327 96.365 -77.787 48.375
148328 96.384 -77.786 48.353
148329 96.410 -77.783 48.331
148330 96.437 -77.781 48.308
148331 96.464 -77.780 48.286
148332 96.475 -77.783 48.262
148333 96.484 -77.787 48.237
148334 96.486 -77.797 48.213
148335 96.473 -77.819 48.189
148336 96.476 -77.830 48.165
148337 96.486 -77.837 48.140
148338 96.496 -77.844 48.117
148339 96.507 -77.849 48.093
148340 96.527 -77.850 48.068
148341 96.547 -77.851 48.044
148342 96.568 -77.851 48.020
148343 96.588 -77.852 47.996
148344 96.601 -77.837 47.971
148345 96.614 -77.821 47.946
148346 96.626 -77.804 47.920
148347 96.639 -77.787 47.894
148348 96.652 -77.771 47.870
148349 96.664 -77.754 47.846
148350 96.682 -77.758 47.821
148351 96.700 -77.765 47.795
148352 96.717 -77.772 47.770
148353 96.738 -77.783 47.745
148354 96.759 -77.796 47.721
148355 96.782 -77.808 47.696
148356 96.804 -77.820 47.670
148357 96.826 -77.833 47.647
148358 96.848 -77.845 47.623
148359 96.870 -77.858 47.600
148360 96.892 -77.870 47.577
148361 96.913 -77.882 47.555
148362 96.923 -77.890 47.533
148363 96.932 -77.897 47.512
148364 96.941 -77.904 47.490
148365 96.949 -77.903 47.470
148366 96.957 -77.896 47.451
148367 96.966 -77.889 47.433
148368 96.974 -77.882 47.412
148369 96.982 -77.875 47.391
148370 96.993 -77.869 47.370
148371 97.015 -77.871 47.347
148372 97.035 -77.875 47.326
148373 97.056 -77.892 47.306
148374 97.077 -77.909 47.286
148375 97.098 -77.926 47.267
148376 97.119 -77.943 47.248
148377 97.140 -77.960 47.230
148378 97.160 -77.977 47.212
148379 97.181 -77.994 47.196
148380 97.202 -78.009 47.180
148381 97.226 -78.015 47.166
148382 97.252 -78.009 47.150
148383 97.277 -77.995 47.134
148384 97.301 -77.982 47.118
148385 97.325 -77.968 47.102
148386 97.350 -77.957 47.089
148387 97.371 -77.970 47.075
148388 97.392 -77.984 47.060
148389 97.413 -77.997 47.045
148390 97.436 -78.013 47.030
148391 97.459 -78.031 47.015
148392 97.482 -78.048 47.000
148393 97.505 -78.065 46.986
148394 97.500 -78.078 46.971
148395 97.490 -78.091 46.956
148396 97.485 -78.107 46.940
148397 97.480 -78.123 46.925
148398 97.476 -78.139 46.909
148399 97.471 -78.155 46.893
148400 97.486 -78.172 46.879
148401 97.503 -78.189 46.863
148402 97.528 -78.204 46.846
148403 97.578 -78.216 46.826
148404 97.627 -78.227 46.806
148405 97.676 -78.239 46.785
148406 97.725 -78.250 46.762
148407 97.774 -78.262 46.738
148408 97.816 -78.269 46.715
148409 97.853 -78.275 46.692
148410 97.888 -78.280 46.669
148411 97.925 -78.285 46.646
148412 97.961 -78.290 46.623
148413 97.997 -78.294 46.601
148414 98.027 -78.301 46.578
148415 98.055 -78.308 46.554
148416 98.068 -78.299 46.532
148417 98.111 -78.288 46.509
148418 98.159 -78.278 46.487
148419 98.180 -78.266 46.464
148420 98.202 -78.254 46.441
148421 98.223 -78.243 46.418
148422 98.244 -78.232 46.394
148423 98.265 -78.220 46.371
148424 98.286 -78.208 46.348
148425 98.304 -78.195 46.325
148426 98.304 -78.179 46.303
148427 98.334 -78.181 46.279
148428 98.364 -78.182 46.255
148429 98.394 -78.184 46.231
148430 98.424 -78.185 46.207
148431 98.453 -78.188 46.184
148432 98.484 -78.189 46.161
148433 98.519 -78.186 46.137
148434 98.508 -78.205 46.114
148435 98.496 -78.223 46.090
148436 98.502 -78.235 46.068
148437 98.534 -78.231 46.046
148438 98.536 -78.220 46.024
148439 98.535 -78.208 46.003
148440 98.534 -78.196 45.981
148441 98.533 -78.184 45.959
148442 98.532 -78.171 45.938
148443 98.531 -78.159 45.917
148444 98.532 -78.150 45.899
148445 98.573 -78.186 45.880
148446 98.598 -78.200 45.862
148447 98.622 -78.211 45.844
148448 98.661 -78.219 45.824
148449 98.705 -78.226 45.805
148450 98.748 -78.233 45.786
148451 98.791 -78.241 45.770
148452 98.828 -78.243 45.755
148453 98.857 -78.240 45.740
148454 98.886 -78.238 45.726
148455 98.916 -78.235 45.712
148456 98.943 -78.247 45.697
148457 98.970 -78.263 45.684
148458 98.997 -78.278 45.672
148459 99.024 -78.292 45.659
148460 99.051 -78.307 45.646
148461 99.058 -78.311 45.634
148462 99.080 -78.314 45.621
148463 99.108 -78.319 45.607
148464 99.141 -78.328 45.593
148465 99.173 -78.335 45.580
148466 99.205 -78.343 45.568
148467 99.237 -78.351 45.555
148468 99.270 -78.359 45.542
148469 99.302 -78.366 45.528
148470 99.335 -78.375 45.513
148471 99.367 -78.382 45.499
148472 99.399 -78.391 45.486
148473 99.432 -78.398 45.473
148474 99.464 -78.406 45.460
148475 99.496 -78.414 45.446
148476 99.528 -78.422 45.434
148477 99.561 -78.430 45.420
148478 99.593 -78.438 45.408
148479 99.626 -78.445 45.395
148480 99.658 -78.454 45.384
148481 99.664 -78.472 45.372
148482 99.673 -78.506 45.361
148483 99.723 -78.551 45.349
148484 99.771 -78.564 45.338
148485 99.814 -78.541 45.327
148486 99.889 -78.539 45.316
148487 99.919 -78.554 45.304
148488 99.946 -78.564 45.293
148489 99.962 -78.572 45.281
148490 99.971 -78.579 45.269
148491 99.980 -78.586 45.257
148492 99.989 -78.592 45.246
148493 99.998 -78.598 45.235
148494 100.007 -78.605 45.221
148495 100.016 -78.612 45.208
148496 100.025 -78.619 45.196
148497 100.034 -78.626 45.183
148498 100.043 -78.633 45.168
148499 100.091 -78.647 45.154
148500 100.139 -78.661 45.140
148501 100.177 -78.680 45.127
148502 100.215 -78.698 45.112
148503 100.245 -78.714 45.097
148504 100.263 -78.725 45.081
148505 100.281 -78.736 45.066
148506 100.298 -78.749 45.050
148507 100.317 -78.760 45.035
148508 100.335 -78.772 45.019
148509 100.353 -78.783 45.002
148510 100.371 -78.795 44.986
148511 100.389 -78.807 44.970
148512 100.407 -78.818 44.954
148513 100.422 -78.830 44.938
148514 100.433 -78.839 44.923
148515 100.447 -78.857 44.908
148516 100.462 -78.877 44.892
148517 100.477 -78.896 44.876
148518 100.492 -78.915 44.859
148519 100.507 -78.935 44.842
148520 100.522 -78.954 44.826
148521 100.534 -78.974 44.809
148522 100.548 -78.988 44.792
148523 100.564 -78.999 44.775
148524 100.580 -79.010 44.757
148525 100.596 -79.021 44.739
148526 100.613 -79.033 44.722
148527 100.636 -79.047 44.705
148528 100.668 -79.066 44.688
148529 100.701 -79.085 44.671
148530 100.734 -79.105 44.655
148531 100.772 -79.125 44.639
148532 100.824 -79.131 44.623
148533 100.876 -79.137 44.607
148534 100.928 -79.143 44.593
148535 100.980 -79.150 44.578
148536 101.033 -79.156 44.564
148537 101.085 -79.162 44.550
148538 101.136 -79.169 44.535
148539 101.189 -79.175 44.521
148540 101.241 -79.181 44.508
148541 101.292 -79.187 44.494
148542 101.322 -79.200 44.481
148776 92.670 -77.827 54.230
148777 92.690 -77.834 54.226
148778 92.710 -77.841 54.224
148779 92.730 -77.849 54.221
148780 92.750 -77.856 54.218
148781 92.770 -77.863 54.215
148782 92.790 -77.871 54.213
148783 92.810 -77.878 54.212
148784 92.830 -77.885 54.210
148785 92.850 -77.893 54.209
148786 92.870 -77.900 54.207
148787 92.890 -77.907 54.204
148788 92.910 -77.915 54.203
148789 92.930 -77.922 54.202
148790 92.950 -77.929 54.201
148791 92.970 -77.937 54.200
148792 92.992 -77.934 54.202
148793 93.009 -77.929 54.202
148794 93.026 -77.924 54.203
148795 93.043 -77.918 54.205
148796 93.061 -77.915 54.207
148797 93.081 -77.915 54.208
148798 93.100 -77.913 54.209
148799 93.120 -77.912 54.208
148800 93.140 -77.912 54.206
148801 93.160 -77.910 54.206
148802 93.188 -77.881 54.205
148803 93.203 -77.862 54.204
148804 93.214 -77.848 54.201
148805 93.225 -77.834 54.198
148806 93.236 -77.821 54.195
148807 93.247 -77.808 54.190
148808 93.258 -77.794 54.183
148809 93.269 -77.780 54.174
148810 93.276 -77.767 54.164
148811 93.275 -77.752 54.156
148812 93.274 -77.738 54.147
148813 93.273 -77.723 54.138
148814 93.288 -77.709 54.129
148815 93.304 -77.696 54.121
148816 93.320 -77.683 54.114
148817 93.336 -77.670 54.106
148818 93.352 -77.656 54.100
148819 93.367 -77.642 54.094
148820 93.382 -77.632 54.087
148821 93.398 -77.624 54.078
148822 93.414 -77.617 54.069
148823 93.430 -77.609 54.060
148824 93.446 -77.602 54.052
148825 93.462 -77.594 54.042
148826 93.478 -77.587 54.033
148827 93.494 -77.580 54.024
148828 93.509 -77.573 54.012
148829 93.526 -77.565 54.000
148830 93.544 -77.556 53.986
148831 93.561 -77.546 53.972
148832 93.579 -77.537 53.958
148833 93.597 -77.528 53.944
148834 93.605 -77.526 53.928
148835 93.599 -77.537 53.910
148836 93.608 -77.530 53.891
148837 93.619 -77.520 53.873
148838 93.627 -77.511 53.854
148839 93.635 -77.503 53.834
148840 93.643 -77.495 53.814
148841 93.652 -77.486 53.793
148842 93.670 -77.476 53.771
148843 93.687 -77.467 53.752
148844 93.704 -77.457 53.734
148845 93.721 -77.448 53.718
148846 93.740 -77.439 53.704
148847 93.765 -77.426 53.688
148848 93.783 -77.415 53.672
148849 93.799 -77.404 53.654
148850 93.815 -77.394 53.638
148851 93.831 -77.383 53.625
148852 93.846 -77.373 53.612
148853 93.862 -77.362 53.599
148854 93.867 -77.371 53.582
148855 93.887 -77.341 53.566
148856 93.906 -77.313 53.551
148857 93.925 -77.284 53.538
148858 93.936 -77.279 53.525
148859 93.943 -77.286 53.512
148860 93.949 -77.299 53.498
148861 93.951 -77.327 53.483
148862 93.979 -77.328 53.466
148863 94.001 -77.332 53.449
148864 94.019 -77.339 53.433
148865 94.039 -77.346 53.419
148866 94.057 -77.353 53.405
148867 94.076 -77.360 53.391
148868 94.084 -77.363 53.376
148869 94.081 -77.363 53.356
148870 94.084 -77.356 53.337
148871 94.100 -77.348 53.318
148872 94.129 -77.347 53.298
148873 94.159 -77.346 53.279
148874 94.183 -77.343 53.261
148875 94.203 -77.337 53.247
148876 94.222 -77.332 53.238
148877 94.243 -77.326 53.232
148878 94.262 -77.320 53.227
148879 94.278 -77.314 53.223
148880 94.279 -77.307 53.216
148881 94.279 -77.306 53.206
148882 94.278 -77.307 53.194
148883 94.277 -77.308 53.179
148884 94.276 -77.309 53.167
148885 94.275 -77.310 53.155
148886 94.274 -77.311 53.143
148887 94.277 -77.310 53.128
148888 94.283 -77.307 53.111
148889 94.289 -77.304 53.093
148890 94.295 -77.300 53.076
148891 94.300 -77.296 53.061
148892 94.314 -77.283 53.049
148893 94.327 -77.271 53.038
148894 94.340 -77.258 53.032
148895 94.353 -77.246 53.024
148896 94.366 -77.234 53.015
148897 94.385 -77.226 53.001
148898 94.407 -77.222 52.988
148899 94.429 -77.218 52.976
148900 94.452 -77.214 52.963
148901 94.474 -77.210 52.948
148902 94.497 -77.207 52.933
148903 94.519 -77.202 52.917
148904 94.542 -77.198 52.898
148905 94.562 -77.194 52.881
148906 94.582 -77.189 52.865
148907 94.602 -77.185 52.850
148908 94.618 -77.182 52.834
148909 94.628 -77.181 52.817
148910 94.637 -77.181 52.798
148911 94.647 -77.180 52.776
148912 94.657 -77.179 52.750
148913 94.667 -77.179 52.724
148914 94.677 -77.179 52.696
148915 94.690 -77.181 52.670
148916 94.703 -77.185 52.644
148917 94.732 -77.193 52.615
148918 94.766 -77.202 52.586
148919 94.792 -77.211 52.558
148920 94.804 -77.219 52.531
148921 94.814 -77.228 52.502
148922 94.825 -77.237 52.470
148923 94.836 -77.245 52.433
148924 94.847 -77.254 52.394
148925 94.857 -77.262 52.353
148926 94.868 -77.270 52.312
148927 94.885 -77.272 52.271
148928 94.889 -77.258 52.230
148929 94.885 -77.224 52.189
148930 94.900 -77.242 52.148
148931 94.926 -77.232 52.106
148932 94.949 -77.205 52.063
148933 94.985 -77.166 52.020
148934 95.008 -77.188 51.975
148935 95.033 -77.204 51.929
148936 95.058 -77.219 51.882
148937 95.083 -77.235 51.835
148938 95.108 -77.251 51.783
148939 95.132 -77.266 51.729
148940 95.156 -77.279 51.672
148941 95.172 -77.272 51.609
148942 95.188 -77.265 51.542
148943 95.204 -77.258 51.476
148944 95.219 -77.251 51.416
148945 95.236 -77.244 51.361
148946 95.252 -77.238 51.307
148947 95.268 -77.231 51.255
148948 95.283 -77.224 51.205
148949 95.300 -77.217 51.156
148950 95.316 -77.210 51.109
148951 95.331 -77.204 51.064
148952 95.347 -77.197 51.020
148953 95.364 -77.190 50.977
148954 95.379 -77.183 50.935
148955 95.398 -77.185 50.894
148956 95.419 -77.199 50.853
148957 95.437 -77.214 50.813
148958 95.456 -77.231 50.773
148959 95.480 -77.251 50.734
148960 95.503 -77.271 50.696
148961 95.526 -77.291 50.659
148962 95.549 -77.312 50.621
148963 95.572 -77.332 50.585
148964 95.596 -77.353 50.550
148965 95.619 -77.373 50.517
148966 95.645 -77.395 50.484
148967 95.665 -77.414 50.451
148968 95.676 -77.423 50.418
148969 95.691 -77.435 50.384
148970 95.705 -77.447 50.348
148971 95.720 -77.459 50.310
148972 95.735 -77.471 50.270
148973 95.749 -77.483 50.228
148974 95.764 -77.495 50.185
148975 95.778 -77.508 50.145
148976 95.794 -77.519 50.105
148977 95.813 -77.526 50.064
148978 95.832 -77.534 50.020
148979 95.851 -77.542 49.974
148980 95.871 -77.549 49.925
148981 95.890 -77.556 49.870
148982 95.905 -77.560 49.815
148983 95.921 -77.564 49.763
148984 95.937 -77.568 49.714
148985 95.952 -77.572 49.669
148986 95.968 -77.576 49.626
148987 95.978 -77.579 49.586
148988 95.984 -77.580 49.547
148989 95.991 -77.583 49.511
148990 95.997 -77.584 49.474
148991 96.003 -77.588 49.438
148992 96.007 -77.592 49.402
148993 96.012 -77.597 49.369
148994 96.016 -77.602 49.338
148995 96.030 -77.604 49.309
148996 96.045 -77.605 49.282
148997 96.061 -77.607 49.257
148998 96.076 -77.608 49.234
148999 96.092 -77.610 49.208
149000 96.104 -77.611 49.182
149001 96.101 -77.622 49.157
149002 96.097 -77.632 49.135
149003 96.093 -77.642 49.110
149004 96.090 -77.652 49.084
149005 96.093 -77.661 49.057
149006 96.103 -77.667 49.027
149007 96.093 -77.677 48.999
149008 96.090 -77.686 48.970
149009 96.109 -77.688 48.941
149010 96.127 -77.690 48.908
149011 96.147 -77.693 48.874
149012 96.166 -77.696 48.840
149013 96.189 -77.699 48.807
149014 96.198 -77.707 48.774
149015 96.206 -77.716 48.743
149016 96.215 -77.724 48.713
149017 96.223 -77.733 48.683
149018 96.231 -77.741 48.652
149019 96.239 -77.749 48.620
149020 96.248 -77.758 48.589
149021 96.256 -77.767 48.562
149022 96.264 -77.775 48.535
149023 96.272 -77.783 48.509
149024 96.289 -77.785 48.486
149025 96.306 -77.784 48.461
149026 96.324 -77.783 48.437
149027 96.343 -77.783 48.414
149028 96.370 -77.780 48.392
149029 96.396 -77.779 48.370
149030 96.423 -77.777 48.348
149031 96.450 -77.775 48.326
149032 96.471 -77.775 48.303
149033 96.479 -77.779 48.279
149034 96.488 -77.783 48.255
149035 96.497 -77.787 48.232
149036 96.493 -77.802 48.209
149037 96.488 -77.819 48.186
149038 96.498 -77.826 48.161
149039 96.507 -77.833 48.137
149040 96.517 -77.839 48.112
149041 96.528 -77.845 48.088
149042 96.548 -77.846 48.064
149043 96.568 -77.846 48.041
149044 96.588 -77.847 48.016
149045 96.602 -77.834 47.991
149046 96.615 -77.817 47.966
149047 96.628 -77.800 47.941
149048 96.641 -77.783 47.916
149049 96.653 -77.767 47.891
149050 96.666 -77.751 47.867
149051 96.680 -77.743 47.842
149052 96.698 -77.749 47.817
149053 96.719 -77.761 47.791
149054 96.741 -77.774 47.766
149055 96.763 -77.786 47.741
149056 96.786 -77.798 47.716
149057 96.807 -77.811 47.691
149058 96.829 -77.823 47.667
149059 96.852 -77.836 47.643
149060 96.874 -77.848 47.619
149061 96.896 -77.861 47.596
149062 96.918 -77.873 47.574
149063 96.929 -77.881 47.551
149064 96.938 -77.888 47.530
149065 96.947 -77.896 47.508
149066 96.955 -77.890 47.488
149067 96.963 -77.883 47.469
149068 96.971 -77.876 47.449
149069 96.980 -77.868 47.430
149070 96.988 -77.862 47.409
149071 96.997 -77.855 47.386
149072 97.015 -77.854 47.365
149073 97.035 -77.866 47.346
149074 97.056 -77.884 47.327
149075 97.077 -77.900 47.307
149076 97.098 -77.918 47.288
149077 97.118 -77.930 47.269
149078 97.138 -77.942 47.251
149079 97.158 -77.953 47.234
149080 97.178 -77.965 47.217
149081 97.202 -77.972 47.201
149082 97.231 -77.975 47.186
149083 97.259 -77.978 47.169
149084 97.288 -77.981 47.153
149085 97.314 -77.975 47.137
149086 97.338 -77.961 47.121
149087 97.362 -77.952 47.106
149088 97.383 -77.965 47.091
149089 97.405 -77.980 47.076
149090 97.429 -77.997 47.061
149091 97.452 -78.015 47.045
149092 97.475 -78.032 47.029
149093 97.498 -78.050 47.014
149094 97.519 -78.066 46.998
149095 97.499 -78.076 46.982
149096 97.490 -78.091 46.967
149097 97.486 -78.106 46.950
149098 97.480 -78.122 46.933
149099 97.494 -78.139 46.917
149100 97.511 -78.156 46.900
149101 97.528 -78.173 46.884
149102 97.545 -78.191 46.867
149103 97.562 -78.207 46.850
149104 97.594 -78.222 46.830
149105 97.643 -78.234 46.811
149106 97.687 -78.242 46.790
149107 97.723 -78.247 46.769
149108 97.759 -78.253 46.746
149109 97.796 -78.257 46.724
149110 97.832 -78.263 46.703
149111 97.868 -78.267 46.680
149112 97.904 -78.272 46.657
149113 97.940 -78.278 46.634
149114 97.971 -78.284 46.612
149115 97.999 -78.291 46.588
149116 98.023 -78.294 46.565
149117 98.038 -78.284 46.543
149118 98.101 -78.273 46.521
149119 98.151 -78.262 46.498
149120 98.172 -78.250 46.475
149121 98.194 -78.239 46.452
149122 98.215 -78.228 46.428
149123 98.236 -78.216 46.404
149124 98.257 -78.204 46.381
149125 98.279 -78.193 46.358
149126 98.275 -78.172 46.336
149127 98.305 -78.173 46.313
149128 98.335 -78.175 46.289
149129 98.365 -78.176 46.265
149130 98.394 -78.178 46.241
149131 98.424 -78.180 46.218
149132 98.454 -78.181 46.194
149133 98.484 -78.183 46.172
149134 98.514 -78.184 46.148
149135 98.557 -78.178 46.124
149136 98.546 -78.197 46.101
149137 98.534 -78.216 46.078
149138 98.523 -78.235 46.056
149139 98.513 -78.230 46.034
149140 98.506 -78.222 46.013
149141 98.505 -78.210 45.991
149142 98.504 -78.197 45.969
149143 98.503 -78.184 45.947
149144 98.502 -78.172 45.927
149145 98.510 -78.171 45.907
149146 98.538 -78.187 45.889
149147 98.581 -78.194 45.871
149148 98.625 -78.201 45.853
149149 98.668 -78.209 45.834
149150 98.711 -78.216 45.815
149151 98.754 -78.222 45.796
149152 98.797 -78.230 45.780
149153 98.841 -78.237 45.765
149154 98.874 -78.238 45.751
149155 98.904 -78.235 45.737
149156 98.933 -78.232 45.722
149157 98.960 -78.238 45.707
149158 98.988 -78.253 45.693
149159 99.015 -78.268 45.680
149160 99.042 -78.282 45.667
149161 99.069 -78.297 45.655
149162 99.065 -78.299 45.643
149163 99.087 -78.301 45.630
149164 99.110 -78.304 45.617
149165 99.136 -78.309 45.604
149166 99.168 -78.317 45.590
149167 99.201 -78.325 45.578
149168 99.233 -78.332 45.566
149169 99.265 -78.341 45.553
149170 99.297 -78.348 45.539
149171 99.330 -78.357 45.524
149172 99.362 -78.364 45.510
149173 99.395 -78.372 45.498
149174 99.427 -78.380 45.485
149175 99.459 -78.388 45.472
149176 99.492 -78.395 45.459
149177 99.524 -78.404 45.446
149178 99.556 -78.411 45.434
149179 99.588 -78.420 45.421
149180 99.621 -78.427 45.408
149181 99.650 -78.436 45.397
149182 99.636 -78.463 45.386
149183 99.635 -78.495 45.375
149184 99.686 -78.540 45.362
149185 99.731 -78.534 45.351
149186 99.792 -78.523 45.340
149187 99.875 -78.529 45.329
149188 99.906 -78.547 45.318
149189 99.933 -78.557 45.306
149190 99.961 -78.567 45.294
149191 99.988 -78.576 45.282
149192 100.001 -78.584 45.270
149193 100.010 -78.590 45.258
149194 100.019 -78.597 45.246
149195 100.028 -78.604 45.233
149196 100.037 -78.611 45.222
149197 100.046 -78.617 45.209
149198 100.055 -78.624 45.197
149199 100.064 -78.631 45.182
149200 100.101 -78.643 45.167
149201 100.149 -78.658 45.153
149202 100.187 -78.676 45.138
149203 100.224 -78.695 45.123
149204 100.245 -78.707 45.107
149205 100.263 -78.718 45.092
149206 100.281 -78.730 45.076
149207 100.299 -78.742 45.060
149208 100.317 -78.754 45.044
149209 100.335 -78.765 45.028
149210 100.353 -78.777 45.011
149211 100.371 -78.789 44.994
149212 100.389 -78.800 44.978
149213 100.407 -78.812 44.962
149214 100.425 -78.824 44.946
149215 100.439 -78.834 44.931
149216 100.452 -78.849 44.915
149217 100.467 -78.868 44.899
149218 100.482 -78.888 44.883
149219 100.497 -78.907 44.867
149220 100.512 -78.927 44.851
149221 100.527 -78.946 44.834
149222 100.542 -78.965 44.816
149223 100.555 -78.983 44.799
149224 100.571 -78.994 44.783
149225 100.587 -79.005 44.765
149226 100.604 -79.016 44.747
149227 100.620 -79.028 44.729
149228 100.636 -79.039 44.712
149229 100.666 -79.056 44.695
149230 100.698 -79.075 44.678
149231 100.731 -79.094 44.662
149232 100.777 -79.109 44.646
149233 100.829 -79.115 44.630
149234 100.882 -79.121 44.615
149235 100.933 -79.128 44.600
149236 100.986 -79.134 44.586
149237 101.038 -79.140 44.572
149238 101.090 -79.146 44.557
149239 101.142 -79.152 44.543
149240 101.194 -79.159 44.529
149241 101.246 -79.165 44.515
149242 101.297 -79.171 44.501
149243 101.322 -79.184 44.488
149477 92.659 -77.806 54.225
149478 92.680 -77.814 54.221
149479 92.699 -77.821 54.219
149480 92.719 -77.828 54.217
149481 92.739 -77.836 54.214
149482 92.759 -77.843 54.211
149483 92.779 -77.851 54.210
149484 92.799 -77.858 54.208
149485 92.819 -77.865 54.207
149486 92.839 -77.873 54.206
149487 92.859 -77.880 54.204
149488 92.879 -77.887 54.202
149489 92.899 -77.895 54.200
149490 92.919 -77.902 54.200
149491 92.939 -77.909 54.199
149492 92.958 -77.917 54.199
149493 92.982 -77.914 54.201
149494 92.999 -77.909 54.201
149495 93.016 -77.903 54.202
149496 93.032 -77.898 54.205
149497 93.050 -77.893 54.206
149498 93.067 -77.889 54.207
149499 93.087 -77.887 54.209
149500 93.107 -77.887 54.211
149501 93.127 -77.886 54.211
149502 93.146 -77.884 54.212
149503 93.173 -77.853 54.212
149504 93.184 -77.840 54.210
149505 93.195 -77.827 54.208
149506 93.206 -77.813 54.205
149507 93.217 -77.799 54.201
149508 93.228 -77.787 54.196
149509 93.239 -77.773 54.189
149510 93.251 -77.759 54.181
149511 93.262 -77.746 54.173
149512 93.272 -77.733 54.165
149513 93.271 -77.718 54.157
149514 93.270 -77.703 54.150
149515 93.280 -77.689 54.141
149516 93.295 -77.676 54.133
149517 93.311 -77.662 54.126
149518 93.327 -77.649 54.119
149519 93.343 -77.636 54.113
149520 93.357 -77.626 54.108
149521 93.371 -77.617 54.102
149522 93.387 -77.609 54.094
149523 93.403 -77.602 54.085
149524 93.419 -77.594 54.077
149525 93.435 -77.586 54.068
149526 93.450 -77.580 54.059
149527 93.467 -77.572 54.051
149528 93.483 -77.564 54.042
149529 93.498 -77.555 54.030
149530 93.512 -77.551 54.018
149531 93.525 -77.548 54.005
149532 93.538 -77.545 53.992
149533 93.551 -77.541 53.978
149534 93.568 -77.533 53.964
149535 93.585 -77.524 53.948
149536 93.588 -77.524 53.930
149537 93.597 -77.516 53.911
149538 93.605 -77.508 53.891
149539 93.613 -77.499 53.872
149540 93.621 -77.492 53.853
149541 93.630 -77.483 53.834
149542 93.638 -77.475 53.814
149543 93.651 -77.466 53.792
149544 93.668 -77.457 53.774
149545 93.685 -77.447 53.758
149546 93.706 -77.436 53.742
149547 93.732 -77.423 53.727
149548 93.758 -77.411 53.714
149549 93.782 -77.398 53.698
149550 93.798 -77.388 53.681
149551 93.814 -77.377 53.663
149552 93.830 -77.367 53.648
149553 93.845 -77.356 53.634
149554 93.861 -77.345 53.622
149555 93.864 -77.360 53.607
149556 93.882 -77.335 53.590
149557 93.901 -77.306 53.573
149558 93.920 -77.277 53.560
149559 93.927 -77.284 53.547
149560 93.934 -77.291 53.535
149561 93.942 -77.298 53.523
149562 93.948 -77.309 53.508
149563 93.968 -77.316 53.491
149564 93.987 -77.323 53.473
149565 94.006 -77.329 53.454
149566 94.025 -77.336 53.438
149567 94.043 -77.343 53.424
149568 94.062 -77.350 53.411
149569 94.066 -77.352 53.396
149570 94.068 -77.346 53.379
149571 94.074 -77.333 53.361
149572 94.099 -77.330 53.343
149573 94.128 -77.329 53.323
149574 94.158 -77.329 53.304
149575 94.188 -77.328 53.285
149576 94.208 -77.323 53.269
149577 94.227 -77.317 53.259
149578 94.248 -77.311 53.251
149579 94.267 -77.306 53.245
149580 94.283 -77.300 53.241
149581 94.284 -77.293 53.234
149582 94.285 -77.286 53.226
149583 94.285 -77.284 53.215
149584 94.284 -77.285 53.203
149585 94.283 -77.285 53.192
149586 94.282 -77.287 53.181
149587 94.281 -77.288 53.170
149588 94.283 -77.287 53.156
149589 94.289 -77.284 53.142
149590 94.296 -77.280 53.127
149591 94.301 -77.277 53.113
149592 94.301 -77.276 53.099
149593 94.311 -77.266 53.086
149594 94.324 -77.254 53.075
149595 94.337 -77.241 53.065
149596 94.350 -77.229 53.055
149597 94.363 -77.216 53.045
149598 94.383 -77.210 53.032
149599 94.405 -77.206 53.022
149600 94.428 -77.202 53.011
149601 94.450 -77.198 52.999
149602 94.473 -77.194 52.986
149603 94.495 -77.190 52.971
149604 94.518 -77.186 52.954
149605 94.538 -77.182 52.935
149606 94.558 -77.177 52.917
149607 94.578 -77.172 52.900
149608 94.598 -77.168 52.884
149609 94.617 -77.164 52.868
149610 94.627 -77.163 52.850
149611 94.636 -77.163 52.829
149612 94.646 -77.163 52.806
149613 94.657 -77.162 52.780
149614 94.667 -77.161 52.752
149615 94.678 -77.163 52.726
149616 94.691 -77.166 52.701
149617 94.704 -77.170 52.675
149618 94.722 -77.175 52.647
149619 94.756 -77.184 52.618
149620 94.782 -77.194 52.589
149621 94.793 -77.202 52.561
149622 94.804 -77.210 52.531
149623 94.815 -77.219 52.499
149624 94.826 -77.227 52.464
149625 94.836 -77.235 52.424
149626 94.852 -77.243 52.382
149627 94.871 -77.250 52.340
149628 94.889 -77.253 52.298
149629 94.905 -77.248 52.257
149630 94.921 -77.242 52.215
149631 94.931 -77.235 52.172
149632 94.937 -77.226 52.128
149633 94.953 -77.204 52.082
149634 94.988 -77.172 52.036
149635 95.012 -77.194 51.989
149636 95.036 -77.210 51.940
149637 95.061 -77.225 51.891
149638 95.086 -77.241 51.843
149639 95.110 -77.248 51.790
149640 95.134 -77.254 51.735
149641 95.157 -77.259 51.680
149642 95.171 -77.258 51.618
149643 95.187 -77.251 51.554
149644 95.203 -77.244 51.492
149645 95.219 -77.238 51.437
149646 95.235 -77.231 51.385
149647 95.251 -77.225 51.333
149648 95.267 -77.218 51.281
149649 95.283 -77.211 51.231
149650 95.299 -77.204 51.182
149651 95.315 -77.197 51.134
149652 95.331 -77.191 51.088
149653 95.347 -77.184 51.042
149654 95.363 -77.177 50.999
149655 95.379 -77.170 50.957
149656 95.401 -77.184 50.916
149657 95.422 -77.198 50.874
149658 95.440 -77.213 50.833
149659 95.458 -77.228 50.792
149660 95.477 -77.244 50.754
149661 95.500 -77.264 50.716
149662 95.523 -77.285 50.677
149663 95.547 -77.305 50.637
149664 95.570 -77.326 50.598
149665 95.593 -77.346 50.563
149666 95.616 -77.366 50.530
149667 95.640 -77.386 50.500
149668 95.663 -77.407 50.467
149669 95.680 -77.424 50.433
149670 95.696 -77.438 50.399
149671 95.713 -77.450 50.363
149672 95.729 -77.463 50.325
149673 95.745 -77.475 50.287
149674 95.761 -77.487 50.246
149675 95.776 -77.499 50.203
149676 95.791 -77.511 50.162
149677 95.806 -77.523 50.123
149678 95.823 -77.533 50.082
149679 95.841 -77.539 50.039
149680 95.858 -77.543 49.993
149681 95.873 -77.547 49.944
149682 95.889 -77.551 49.888
149683 95.904 -77.555 49.833
149684 95.920 -77.559 49.780
149685 95.936 -77.563 49.729
149686 95.952 -77.567 49.683
149687 95.967 -77.571 49.639
149688 95.979 -77.574 49.598
149689 95.986 -77.576 49.559
149690 95.992 -77.578 49.522
149691 95.999 -77.580 49.485
149692 96.006 -77.582 49.448
149693 96.012 -77.583 49.411
149694 96.019 -77.585 49.376
149695 96.024 -77.589 49.343
149696 96.040 -77.591 49.313
149697 96.054 -77.592 49.285
149698 96.068 -77.593 49.260
149699 96.083 -77.594 49.236
149700 96.097 -77.595 49.211
149701 96.102 -77.606 49.187
149702 96.099 -77.617 49.163
149703 96.096 -77.629 49.138
149704 96.092 -77.639 49.116
149705 96.089 -77.649 49.092
149706 96.085 -77.659 49.066
149707 96.091 -77.667 49.036
149708 96.089 -77.674 49.008
149709 96.108 -77.677 48.980
149710 96.127 -77.680 48.951
149711 96.146 -77.682 48.919
149712 96.162 -77.683 48.886
149713 96.177 -77.683 48.851
149714 96.190 -77.691 48.817
149715 96.197 -77.701 48.786
149716 96.203 -77.711 48.756
149717 96.209 -77.721 48.727
149718 96.217 -77.730 48.698
149719 96.225 -77.738 48.666
149720 96.234 -77.746 48.634
149721 96.242 -77.755 48.603
149722 96.250 -77.764 48.577
149723 96.258 -77.772 48.551
149724 96.266 -77.780 48.526
149725 96.284 -77.780 48.503
149726 96.302 -77.780 48.478
149727 96.329 -77.778 48.454
149728 96.356 -77.776 48.430
149729 96.382 -77.774 48.409
149730 96.409 -77.772 48.387
149731 96.436 -77.771 48.366
149732 96.462 -77.768 48.344
149733 96.475 -77.771 48.320
149734 96.483 -77.775 48.296
149735 96.493 -77.780 48.272
149736 96.501 -77.783 48.250
149737 96.510 -77.788 48.227
149738 96.501 -77.808 48.204
149739 96.509 -77.815 48.181
149740 96.519 -77.821 48.158
149741 96.528 -77.827 48.134
149742 96.538 -77.834 48.110
149743 96.549 -77.840 48.087
149744 96.569 -77.841 48.062
149745 96.589 -77.842 48.038
149746 96.604 -77.830 48.013
149747 96.616 -77.813 47.989
149748 96.629 -77.796 47.963
149749 96.642 -77.780 47.938
149750 96.655 -77.764 47.914
149751 96.667 -77.747 47.889
149752 96.680 -77.730 47.864
149753 96.700 -77.739 47.839
149754 96.723 -77.752 47.813
149755 96.745 -77.764 47.787
149756 96.767 -77.777 47.762
149757 96.789 -77.789 47.737
149758 96.811 -77.801 47.711
149759 96.833 -77.814 47.687
149760 96.855 -77.826 47.663
149761 96.877 -77.839 47.640
149762 96.899 -77.851 47.616
149763 96.922 -77.864 47.593
149764 96.935 -77.872 47.571
149765 96.944 -77.880 47.549
149766 96.953 -77.884 47.527
149767 96.961 -77.877 47.507
149768 96.969 -77.869 47.487
149769 96.977 -77.862 47.466
149770 96.986 -77.855 47.446
149771 96.994 -77.848 47.425
149772 97.003 -77.841 47.404
149773 97.014 -77.840 47.384
149774 97.034 -77.852 47.365
149775 97.054 -77.863 47.347
149776 97.074 -77.875 47.329
149777 97.094 -77.887 47.310
149778 97.114 -77.899 47.291
149779 97.134 -77.910 47.274
149780 97.154 -77.922 47.257
149781 97.178 -77.929 47.239
149782 97.207 -77.932 47.222
149783 97.235 -77.934 47.206
149784 97.264 -77.937 47.188
149785 97.293 -77.940 47.172
149786 97.321 -77.943 47.156
149787 97.350 -77.946 47.139
149788 97.375 -77.946 47.124
149789 97.398 -77.964 47.108
149790 97.422 -77.981 47.092
149791 97.445 -77.999 47.077
149792 97.468 -78.016 47.060
149793 97.491 -78.033 47.044
149794 97.511 -78.050 47.028
149795 97.517 -78.065 47.011
149796 97.498 -78.075 46.994
149797 97.490 -78.090 46.977
149798 97.503 -78.106 46.959
149799 97.520 -78.124 46.942
149800 97.537 -78.141 46.924
149801 97.554 -78.158 46.906
149802 97.567 -78.172 46.888
149803 97.580 -78.187 46.871
149804 97.594 -78.201 46.853
149805 97.603 -78.212 46.835
149806 97.631 -78.225 46.815
149807 97.667 -78.230 46.794
149808 97.703 -78.235 46.774
149809 97.739 -78.240 46.753
149810 97.775 -78.245 46.733
149811 97.811 -78.250 46.713
149812 97.848 -78.255 46.691
149813 97.883 -78.260 46.669
149814 97.915 -78.266 46.646
149815 97.943 -78.273 46.623
149816 97.971 -78.280 46.599
149817 97.992 -78.279 46.577
149818 98.027 -78.269 46.555
149819 98.091 -78.258 46.532
149820 98.144 -78.247 46.509
149821 98.165 -78.235 46.486
149822 98.186 -78.224 46.462
149823 98.208 -78.212 46.439
149824 98.229 -78.200 46.415
149825 98.250 -78.189 46.391
149826 98.260 -78.172 46.368
149827 98.276 -78.166 46.347
149828 98.305 -78.167 46.323
149829 98.335 -78.169 46.299
149830 98.365 -78.171 46.275
149831 98.395 -78.172 46.252
149832 98.425 -78.174 46.228
149833 98.454 -78.175 46.205
149834 98.485 -78.177 46.183
149835 98.514 -78.178 46.158
149836 98.544 -78.180 46.134
149837 98.595 -78.170 46.111
149838 98.584 -78.189 46.089
149839 98.573 -78.200 46.067
149840 98.563 -78.193 46.045
149841 98.553 -78.186 46.023
149842 98.543 -78.180 46.000
149843 98.533 -78.173 45.979
149844 98.512 -78.172 45.958
149845 98.499 -78.168 45.937
149846 98.496 -78.169 45.917
149847 98.544 -78.176 45.898
149848 98.587 -78.184 45.880
149849 98.631 -78.191 45.862
149850 98.674 -78.198 45.844
149851 98.717 -78.205 45.825
149852 98.761 -78.213 45.807
149853 98.804 -78.219 45.792
149854 98.847 -78.227 45.777
149855 98.890 -78.234 45.763
149856 98.921 -78.232 45.749
149857 98.950 -78.229 45.733
149858 98.979 -78.228 45.718
149859 99.006 -78.243 45.703
149860 99.032 -78.258 45.690
149861 99.060 -78.272 45.676
149862 99.085 -78.287 45.663
149863 99.078 -78.287 45.651
149864 99.095 -78.289 45.639
149865 99.118 -78.292 45.627
149866 99.140 -78.295 45.614
149867 99.163 -78.298 45.600
149868 99.196 -78.307 45.588
149869 99.228 -78.314 45.576
149870 99.261 -78.322 45.564
149871 99.293 -78.330 45.549
149872 99.325 -78.338 45.535
149873 99.357 -78.346 45.521
149874 99.390 -78.354 45.509
149875 99.422 -78.361 45.496
149876 99.454 -78.369 45.483
149877 99.487 -78.377 45.471
149878 99.519 -78.385 45.459
149879 99.552 -78.393 45.447
149880 99.584 -78.401 45.435
149881 99.616 -78.409 45.422
149882 99.622 -78.428 45.410
149883 99.608 -78.455 45.399
149884 99.598 -78.483 45.387
149885 99.648 -78.528 45.375
149886 99.695 -78.507 45.364
149887 99.779 -78.512 45.353
149888 99.860 -78.518 45.342
149889 99.893 -78.541 45.331
149890 99.921 -78.550 45.319
149891 99.948 -78.560 45.306
149892 99.975 -78.570 45.294
149893 100.002 -78.579 45.283
149894 100.030 -78.589 45.271
149895 100.040 -78.595 45.258
149896 100.049 -78.602 45.246
149897 100.058 -78.609 45.234
149898 100.067 -78.616 45.222
149899 100.076 -78.623 45.210
149900 100.085 -78.630 45.195
149901 100.111 -78.639 45.181
149902 100.158 -78.655 45.165
149903 100.196 -78.673 45.149
149904 100.227 -78.689 45.134
149905 100.245 -78.700 45.118
149906 100.263 -78.712 45.102
149907 100.281 -78.724 45.087
149908 100.299 -78.736 45.070
149909 100.317 -78.747 45.053
149910 100.335 -78.758 45.036
149911 100.353 -78.771 45.019
149912 100.371 -78.782 45.003
149913 100.390 -78.794 44.987
149914 100.408 -78.805 44.971
149915 100.426 -78.817 44.955
149916 100.443 -78.829 44.939
149917 100.457 -78.841 44.923
149918 100.472 -78.861 44.907
149919 100.487 -78.880 44.891
149920 100.501 -78.899 44.874
149921 100.516 -78.918 44.858
149922 100.531 -78.938 44.842
149923 100.547 -78.957 44.825
149924 100.562 -78.975 44.807
149925 100.579 -78.989 44.790
149926 100.595 -79.000 44.772
149927 100.612 -79.011 44.754
149928 100.628 -79.022 44.737
149929 100.644 -79.034 44.720
149930 100.663 -79.046 44.703
149931 100.696 -79.065 44.686
149932 100.729 -79.084 44.670
149933 100.782 -79.092 44.653
149934 100.835 -79.098 44.637
149935 100.887 -79.104 44.622
149936 100.938 -79.111 44.607
149937 100.991 -79.117 44.593
149938 101.043 -79.123 44.579
149939 101.095 -79.129 44.564
149940 101.147 -79.136 44.550
149941 101.199 -79.142 44.536
149942 101.251 -79.148 44.522
149943 101.296 -79.156 44.508
149944 101.321 -79.169 44.495
149945 101.345 -79.183 44.482
150178 92.648 -77.787 54.220
150179 92.668 -77.794 54.217
150180 92.688 -77.801 54.215
150181 92.708 -77.808 54.213
150182 92.728 -77.816 54.210
150183 92.748 -77.823 54.208
150184 92.768 -77.830 54.206
150185 92.788 -77.838 54.205
150186 92.808 -77.846 54.204
150187 92.828 -77.852 54.203
150188 92.848 -77.860 54.202
150189 92.868 -77.868 54.201
150190 92.888 -77.874 54.199
150191 92.908 -77.882 54.198
150192 92.928 -77.890 54.198
150193 92.948 -77.896 54.198
150194 92.972 -77.893 54.200
150195 92.989 -77.887 54.201
150196 93.006 -77.883 54.202
150197 93.023 -77.877 54.204
150198 93.040 -77.872 54.205
150199 93.057 -77.867 54.206
150200 93.074 -77.862 54.209
150201 93.093 -77.861 54.213
150202 93.114 -77.860 54.216
150203 93.133 -77.859 54.217
150204 93.154 -77.832 54.218
150205 93.165 -77.818 54.217
150206 93.177 -77.805 54.215
150207 93.188 -77.792 54.211
150208 93.199 -77.778 54.207
150209 93.210 -77.765 54.202
150210 93.221 -77.752 54.196
150211 93.232 -77.738 54.189
150212 93.243 -77.724 54.182
150213 93.255 -77.711 54.175
150214 93.266 -77.698 54.168
150215 93.267 -77.683 54.161
150216 93.271 -77.669 54.153
150217 93.287 -77.655 54.146
150218 93.303 -77.642 54.139
150219 93.318 -77.630 54.132
150220 93.332 -77.620 54.125
150221 93.346 -77.611 54.119
150222 93.359 -77.602 54.112
150223 93.375 -77.594 54.105
150224 93.392 -77.586 54.098
150225 93.407 -77.579 54.090
150226 93.423 -77.571 54.082
150227 93.439 -77.564 54.073
150228 93.455 -77.556 54.064
150229 93.471 -77.548 54.055
150230 93.487 -77.539 54.044
150231 93.501 -77.533 54.032
150232 93.515 -77.530 54.020
150233 93.528 -77.526 54.008
150234 93.541 -77.523 53.994
150235 93.554 -77.519 53.980
150236 93.567 -77.516 53.966
150237 93.579 -77.510 53.948
150238 93.586 -77.502 53.928
150239 93.592 -77.495 53.908
150240 93.600 -77.488 53.889
150241 93.608 -77.479 53.871
150242 93.616 -77.472 53.853
150243 93.624 -77.464 53.834
150244 93.632 -77.455 53.813
150245 93.649 -77.446 53.796
150246 93.673 -77.434 53.780
150247 93.699 -77.421 53.765
150248 93.724 -77.409 53.752
150249 93.750 -77.396 53.739
150250 93.776 -77.384 53.723
150251 93.797 -77.372 53.706
150252 93.813 -77.361 53.687
150253 93.829 -77.351 53.671
150254 93.845 -77.340 53.656
150255 93.859 -77.332 53.642
150256 93.861 -77.347 53.629
150257 93.876 -77.328 53.613
150258 93.896 -77.299 53.597
150259 93.911 -77.282 53.582
150260 93.918 -77.288 53.568
150261 93.926 -77.295 53.554
150262 93.933 -77.301 53.542
150263 93.941 -77.302 53.528
150264 93.957 -77.306 53.512
150265 93.974 -77.313 53.495
150266 93.992 -77.319 53.476
150267 94.010 -77.326 53.459
150268 94.029 -77.332 53.444
150269 94.048 -77.339 53.430
150270 94.051 -77.337 53.417
150271 94.057 -77.323 53.402
150272 94.068 -77.313 53.386
150273 94.098 -77.312 53.368
150274 94.128 -77.312 53.349
150275 94.157 -77.311 53.329
150276 94.187 -77.310 53.310
150277 94.213 -77.308 53.293
150278 94.232 -77.303 53.281
150279 94.253 -77.297 53.273
150280 94.272 -77.291 53.266
150281 94.288 -77.285 53.260
150282 94.289 -77.279 53.253
150283 94.290 -77.271 53.246
150284 94.291 -77.264 53.238
150285 94.291 -77.261 53.229
150286 94.290 -77.263 53.219
150287 94.289 -77.263 53.209
150288 94.288 -77.264 53.199
150289 94.290 -77.264 53.187
150290 94.296 -77.260 53.175
150291 94.302 -77.257 53.164
150292 94.304 -77.255 53.152
150293 94.305 -77.254 53.139
150294 94.309 -77.250 53.126
150295 94.322 -77.237 53.115
150296 94.335 -77.225 53.104
150297 94.348 -77.212 53.093
150298 94.361 -77.200 53.082
150299 94.382 -77.194 53.068
150300 94.404 -77.190 53.058
150301 94.426 -77.186 53.049
150302 94.449 -77.182 53.038
150303 94.471 -77.178 53.025
150304 94.493 -77.174 53.010
150305 94.514 -77.169 52.992
150306 94.535 -77.165 52.972
150307 94.554 -77.160 52.953
150308 94.574 -77.156 52.935
150309 94.595 -77.151 52.917
150310 94.615 -77.147 52.900
150311 94.626 -77.146 52.881
150312 94.636 -77.145 52.859
150313 94.645 -77.144 52.835
150314 94.655 -77.144 52.809
150315 94.666 -77.144 52.782
150316 94.679 -77.148 52.756
150317 94.693 -77.151 52.732
150318 94.706 -77.155 52.706
150319 94.719 -77.159 52.677
150320 94.746 -77.166 52.645
150321 94.772 -77.175 52.616
150322 94.782 -77.184 52.587
150323 94.793 -77.192 52.557
150324 94.804 -77.201 52.524
150325 94.818 -77.208 52.488
150326 94.837 -77.215 52.449
150327 94.856 -77.222 52.407
150328 94.875 -77.229 52.366
150329 94.893 -77.234 52.325
150330 94.910 -77.229 52.282
150331 94.926 -77.223 52.239
150332 94.942 -77.218 52.196
150333 94.958 -77.213 52.150
150334 94.972 -77.207 52.103
150335 94.990 -77.180 52.055
150336 95.014 -77.199 52.005
150337 95.039 -77.211 51.952
150338 95.063 -77.216 51.897
150339 95.087 -77.222 51.846
150340 95.111 -77.228 51.791
150341 95.135 -77.234 51.735
150342 95.159 -77.239 51.680
150343 95.172 -77.240 51.621
150344 95.187 -77.238 51.563
150345 95.203 -77.232 51.507
150346 95.218 -77.225 51.457
150347 95.235 -77.218 51.409
150348 95.251 -77.211 51.359
150349 95.266 -77.204 51.308
150350 95.282 -77.197 51.257
150351 95.299 -77.191 51.207
150352 95.315 -77.184 51.158
150353 95.330 -77.177 51.112
150354 95.346 -77.171 51.065
150355 95.362 -77.164 51.021
150356 95.382 -77.168 50.979
150357 95.404 -77.183 50.938
150358 95.425 -77.197 50.895
150359 95.443 -77.213 50.853
150360 95.461 -77.227 50.812
150361 95.479 -77.241 50.774
150362 95.497 -77.257 50.736
150363 95.521 -77.278 50.696
150364 95.544 -77.298 50.654
150365 95.567 -77.318 50.615
150366 95.591 -77.338 50.579
150367 95.614 -77.359 50.545
150368 95.637 -77.379 50.514
150369 95.660 -77.400 50.480
150370 95.678 -77.417 50.446
150371 95.694 -77.434 50.412
150372 95.710 -77.451 50.376
150373 95.727 -77.464 50.339
150374 95.743 -77.477 50.301
150375 95.759 -77.489 50.262
150376 95.775 -77.501 50.220
150377 95.792 -77.514 50.180
150378 95.808 -77.526 50.141
150379 95.824 -77.533 50.100
150380 95.840 -77.536 50.057
150381 95.856 -77.539 50.011
150382 95.872 -77.542 49.962
150383 95.888 -77.546 49.906
150384 95.903 -77.550 49.851
150385 95.919 -77.555 49.797
150386 95.935 -77.558 49.744
150387 95.951 -77.562 49.697
150388 95.966 -77.567 49.653
150389 95.980 -77.570 49.612
150390 95.987 -77.572 49.571
150391 95.994 -77.573 49.533
150392 96.001 -77.575 49.495
150393 96.007 -77.577 49.457
150394 96.014 -77.579 49.419
150395 96.022 -77.579 49.383
150396 96.033 -77.578 49.349
150397 96.043 -77.577 49.318
150398 96.057 -77.577 49.289
150399 96.072 -77.579 49.262
150400 96.086 -77.580 49.236
150401 96.100 -77.572 49.212
150402 96.100 -77.596 49.189
150403 96.096 -77.611 49.166
150404 96.093 -77.622 49.141
150405 96.090 -77.633 49.119
150406 96.087 -77.645 49.095
150407 96.084 -77.656 49.070
150408 96.089 -77.664 49.044
150409 96.104 -77.664 49.016
150410 96.119 -77.664 48.988
150411 96.133 -77.664 48.959
150412 96.149 -77.664 48.928
150413 96.163 -77.664 48.895
150414 96.178 -77.667 48.861
150415 96.192 -77.679 48.828
150416 96.205 -77.688 48.798
150417 96.216 -77.697 48.769
150418 96.221 -77.707 48.740
150419 96.227 -77.718 48.711
150420 96.233 -77.727 48.679
150421 96.238 -77.737 48.648
150422 96.244 -77.747 48.617
150423 96.250 -77.757 48.592
150424 96.255 -77.768 48.567
150425 96.261 -77.777 48.543
150426 96.288 -77.775 48.519
150427 96.315 -77.774 48.494
150428 96.341 -77.771 48.470
150429 96.368 -77.770 48.446
150430 96.395 -77.768 48.425
150431 96.421 -77.765 48.404
150432 96.448 -77.764 48.382
150433 96.470 -77.763 48.360
150434 96.479 -77.768 48.337
150435 96.488 -77.771 48.313
150436 96.497 -77.776 48.290
150437 96.506 -77.780 48.268
150438 96.514 -77.784 48.245
150439 96.520 -77.790 48.223
150440 96.521 -77.803 48.202
150441 96.530 -77.810 48.179
150442 96.540 -77.817 48.156
150443 96.550 -77.823 48.133
150444 96.559 -77.830 48.110
150445 96.570 -77.836 48.085
150446 96.590 -77.837 48.061
150447 96.605 -77.826 48.037
150448 96.618 -77.809 48.012
150449 96.630 -77.793 47.987
150450 96.643 -77.777 47.961
150451 96.657 -77.762 47.936
150452 96.672 -77.750 47.911
150453 96.687 -77.738 47.886
150454 96.704 -77.733 47.860
150455 96.727 -77.743 47.834
150456 96.748 -77.755 47.808
150457 96.770 -77.767 47.782
150458 96.793 -77.779 47.757
150459 96.815 -77.792 47.732
150460 96.837 -77.804 47.708
150461 96.859 -77.817 47.684
150462 96.881 -77.829 47.660
150463 96.903 -77.842 47.636
150464 96.925 -77.854 47.613
150465 96.941 -77.864 47.591
150466 96.950 -77.871 47.568
150467 96.959 -77.870 47.547
150468 96.967 -77.863 47.526
150469 96.975 -77.856 47.505
150470 96.983 -77.849 47.485
150471 96.992 -77.842 47.463
150472 97.001 -77.836 47.442
150473 97.012 -77.834 47.422
150474 97.023 -77.832 47.403
150475 97.036 -77.832 47.385
150476 97.050 -77.832 47.367
150477 97.070 -77.843 47.350
150478 97.090 -77.855 47.332
150479 97.110 -77.867 47.314
150480 97.130 -77.879 47.297
150481 97.155 -77.886 47.280
150482 97.183 -77.888 47.262
150483 97.212 -77.891 47.243
150484 97.240 -77.894 47.225
150485 97.269 -77.897 47.207
150486 97.300 -77.909 47.190
150487 97.331 -77.921 47.174
150488 97.360 -77.929 47.157
150489 97.381 -77.929 47.142
150490 97.393 -77.941 47.126
150491 97.409 -77.959 47.109
150492 97.426 -77.978 47.092
150493 97.442 -77.996 47.074
150494 97.458 -78.015 47.057
150495 97.470 -78.034 47.041
150496 97.476 -78.029 47.023
150497 97.489 -78.040 47.006
150498 97.509 -78.077 46.988
150499 97.522 -78.091 46.970
150500 97.536 -78.106 46.952
150501 97.549 -78.120 46.933
150502 97.563 -78.134 46.914
150503 97.576 -78.149 46.895
150504 97.590 -78.163 46.877
150505 97.603 -78.178 46.858
150506 97.603 -78.180 46.840
150507 97.618 -78.200 46.820
150508 97.646 -78.218 46.800
150509 97.682 -78.223 46.781
150510 97.718 -78.228 46.762
150511 97.754 -78.233 46.742
150512 97.791 -78.238 46.722
150513 97.827 -78.243 46.702
150514 97.859 -78.249 46.679
150515 97.887 -78.256 46.656
150516 97.916 -78.263 46.633
150517 97.944 -78.270 46.610
150518 97.960 -78.264 46.587
150519 98.017 -78.253 46.565
150520 98.080 -78.242 46.543
150521 98.136 -78.232 46.520
150522 98.158 -78.220 46.496
150523 98.179 -78.208 46.473
150524 98.200 -78.197 46.449
150525 98.221 -78.185 46.425
150526 98.242 -78.174 46.401
150527 98.246 -78.159 46.378
150528 98.276 -78.160 46.356
150529 98.306 -78.162 46.332
150530 98.336 -78.163 46.308
150531 98.366 -78.165 46.284
150532 98.395 -78.166 46.261
150533 98.425 -78.168 46.237
150534 98.455 -78.169 46.215
150535 98.485 -78.171 46.193
150536 98.515 -78.172 46.169
150537 98.545 -78.174 46.145
150538 98.575 -78.175 46.122
150539 98.633 -78.162 46.100
150540 98.623 -78.162 46.077
150541 98.576 -78.173 46.055
150542 98.563 -78.169 46.033
150543 98.550 -78.165 46.010
150544 98.537 -78.161 45.990
150545 98.524 -78.156 45.970
150546 98.511 -78.152 45.948
150547 98.510 -78.158 45.927
150548 98.537 -78.165 45.907
150549 98.590 -78.173 45.888
150550 98.637 -78.181 45.871
150551 98.680 -78.188 45.853
150552 98.723 -78.195 45.835
150553 98.767 -78.202 45.818
150554 98.810 -78.210 45.804
150555 98.853 -78.216 45.789
150556 98.897 -78.223 45.775
150557 98.937 -78.229 45.760
150558 98.967 -78.226 45.745
150559 98.996 -78.223 45.730
150560 99.023 -78.233 45.715
150561 99.051 -78.248 45.701
150562 99.077 -78.263 45.687
150563 99.099 -78.275 45.673
150564 99.092 -78.275 45.660
150565 99.102 -78.277 45.648
150566 99.125 -78.280 45.635
150567 99.147 -78.283 45.622
150568 99.169 -78.285 45.610
150569 99.192 -78.288 45.598
150570 99.223 -78.296 45.585
150571 99.256 -78.304 45.574
150572 99.288 -78.312 45.560
150573 99.321 -78.319 45.546
150574 99.353 -78.327 45.532
150575 99.385 -78.335 45.520
150576 99.417 -78.343 45.507
150577 99.450 -78.351 45.495
150578 99.482 -78.359 45.483
150579 99.514 -78.366 45.472
150580 99.547 -78.375 45.460
150581 99.579 -78.382 45.447
150582 99.607 -78.392 45.435
150583 99.593 -78.420 45.422
150584 99.580 -78.447 45.411
150585 99.567 -78.474 45.399
150586 99.609 -78.498 45.389
150587 99.682 -78.495 45.377
150588 99.766 -78.501 45.366
150589 99.846 -78.508 45.355
150590 99.878 -78.536 45.343
150591 99.908 -78.543 45.331
150592 99.935 -78.553 45.318
150593 99.963 -78.563 45.306
150594 99.990 -78.572 45.296
150595 100.017 -78.582 45.283
150596 100.044 -78.592 45.271
150597 100.069 -78.601 45.258
150598 100.078 -78.608 45.245
150599 100.087 -78.614 45.233
150600 100.097 -78.621 45.222
150601 100.106 -78.628 45.208
150602 100.120 -78.636 45.193
150603 100.168 -78.651 45.178
150604 100.205 -78.669 45.162
150605 100.227 -78.682 45.146
150606 100.246 -78.694 45.130
150607 100.264 -78.705 45.113
150608 100.281 -78.717 45.096
150609 100.299 -78.729 45.079
150610 100.318 -78.740 45.062
150611 100.336 -78.752 45.045
150612 100.354 -78.764 45.028
150613 100.372 -78.776 45.011
150614 100.390 -78.787 44.995
150615 100.408 -78.799 44.979
150616 100.426 -78.811 44.964
150617 100.444 -78.822 44.948
150618 100.462 -78.834 44.932
150619 100.476 -78.852 44.916
150620 100.492 -78.871 44.899
150621 100.508 -78.888 44.882
150622 100.525 -78.905 44.866
150623 100.541 -78.921 44.849
150624 100.557 -78.939 44.832
150625 100.568 -78.960 44.815
150626 100.572 -78.987 44.797
150627 100.588 -79.012 44.780
150628 100.609 -79.026 44.763
150629 100.635 -79.017 44.745
150630 100.651 -79.028 44.728
150631 100.668 -79.040 44.711
150632 100.693 -79.055 44.694
150633 100.736 -79.069 44.678
150634 100.788 -79.075 44.661
150635 100.840 -79.081 44.645
150636 100.892 -79.088 44.630
150637 100.944 -79.094 44.616
150638 100.996 -79.100 44.600
150639 101.048 -79.106 44.585
150640 101.100 -79.112 44.571
150641 101.152 -79.119 44.558
150642 101.204 -79.125 44.543
150643 101.256 -79.131 44.529
150644 101.295 -79.141 44.516
150645 101.320 -79.155 44.503
150646 101.344 -79.168 44.490
150647 101.369 -79.181 44.477
150879 92.637 -77.767 54.215
150880 92.657 -77.774 54.213
150881 92.677 -77.781 54.211
150882 92.697 -77.789 54.209
150883 92.717 -77.796 54.207
150884 92.737 -77.803 54.205
150885 92.757 -77.811 54.204
150886 92.777 -77.818 54.202
150887 92.797 -77.825 54.201
150888 92.817 -77.833 54.201
150889 92.837 -77.840 54.200
150890 92.857 -77.847 54.199
150891 92.877 -77.855 54.198
150892 92.897 -77.862 54.198
150893 92.917 -77.869 54.198
150894 92.937 -77.877 54.199
150895 92.961 -77.873 54.200
150896 92.979 -77.867 54.200
150897 92.996 -77.862 54.202
150898 93.013 -77.856 54.204
150899 93.030 -77.851 54.205
150900 93.047 -77.846 54.207
150901 93.064 -77.840 54.211
150902 93.080 -77.836 54.216
150903 93.100 -77.834 54.220
150904 93.120 -77.833 54.223
150905 93.136 -77.811 54.224
150906 93.147 -77.797 54.223
150907 93.158 -77.783 54.221
150908 93.169 -77.771 54.218
150909 93.181 -77.757 54.214
150910 93.192 -77.743 54.209
150911 93.203 -77.730 54.203
150912 93.214 -77.717 54.196
150913 93.225 -77.703 54.191
150914 93.236 -77.689 54.185
150915 93.248 -77.676 54.179
150916 93.259 -77.663 54.173
150917 93.263 -77.649 54.166
150918 93.279 -77.635 54.160
150919 93.292 -77.625 54.153
150920 93.306 -77.615 54.146
150921 93.320 -77.605 54.138
150922 93.334 -77.596 54.130
150923 93.348 -77.586 54.124
150924 93.364 -77.579 54.116
150925 93.380 -77.571 54.109
150926 93.396 -77.564 54.102
150927 93.412 -77.556 54.094
150928 93.428 -77.548 54.086
150929 93.444 -77.539 54.077
150930 93.460 -77.531 54.066
150931 93.476 -77.523 54.055
150932 93.491 -77.514 54.042
150933 93.504 -77.511 54.029
150934 93.517 -77.508 54.018
150935 93.531 -77.504 54.005
150936 93.544 -77.501 53.993
150937 93.558 -77.497 53.979
150938 93.575 -77.492 53.962
150939 93.583 -77.485 53.943
150940 93.590 -77.477 53.924
150941 93.596 -77.470 53.905
150942 93.603 -77.463 53.887
150943 93.609 -77.456 53.869
150944 93.616 -77.448 53.852
150945 93.622 -77.442 53.834
150946 93.640 -77.432 53.817
150947 93.665 -77.419 53.802
150948 93.691 -77.407 53.788
150949 93.717 -77.394 53.776
150950 93.743 -77.382 53.763
150951 93.768 -77.369 53.748
150952 93.794 -77.356 53.731
150953 93.812 -77.345 53.711
150954 93.828 -77.335 53.694
150955 93.844 -77.324 53.678
150956 93.856 -77.320 53.663
150957 93.858 -77.335 53.650
150958 93.871 -77.321 53.635
150959 93.891 -77.292 53.620
150960 93.902 -77.286 53.605
150961 93.910 -77.293 53.590
150962 93.917 -77.299 53.575
150963 93.925 -77.300 53.560
150964 93.934 -77.294 53.545
150965 93.949 -77.298 53.529
150966 93.967 -77.304 53.512
150967 93.984 -77.310 53.494
150968 94.001 -77.316 53.478
150969 94.018 -77.323 53.464
150970 94.035 -77.327 53.452
150971 94.041 -77.314 53.439
150972 94.047 -77.301 53.425
150973 94.067 -77.295 53.410
150974 94.097 -77.294 53.393
150975 94.127 -77.294 53.373
150976 94.156 -77.294 53.354
150977 94.186 -77.293 53.335
150978 94.215 -77.292 53.318
150979 94.238 -77.288 53.308
150980 94.258 -77.282 53.300
150981 94.277 -77.276 53.294
150982 94.293 -77.271 53.288
150983 94.294 -77.263 53.282
150984 94.295 -77.257 53.276
150985 94.295 -77.250 53.270
150986 94.296 -77.243 53.261
150987 94.297 -77.239 53.253
150988 94.296 -77.240 53.243
150989 94.295 -77.241 53.233
150990 94.296 -77.241 53.223
150991 94.302 -77.238 53.213
150992 94.307 -77.235 53.204
150993 94.308 -77.233 53.193
150994 94.308 -77.232 53.182
150995 94.309 -77.230 53.170
150996 94.319 -77.220 53.158
150997 94.332 -77.207 53.148
150998 94.345 -77.195 53.137
150999 94.358 -77.183 53.125
151000 94.380 -77.178 53.108
151001 94.402 -77.174 53.097
151002 94.425 -77.170 53.087
151003 94.447 -77.166 53.077
151004 94.470 -77.162 53.064
151005 94.491 -77.157 53.048
151006 94.511 -77.153 53.030
151007 94.531 -77.148 53.010
151008 94.551 -77.144 52.989
151009 94.571 -77.139 52.969
151010 94.591 -77.135 52.950
151011 94.611 -77.130 52.933
151012 94.625 -77.128 52.912
151013 94.635 -77.128 52.890
151014 94.644 -77.127 52.865
151015 94.654 -77.126 52.839
151016 94.668 -77.129 52.813
151017 94.681 -77.133 52.787
151018 94.694 -77.137 52.763
151019 94.707 -77.141 52.736
151020 94.720 -77.144 52.705
151021 94.737 -77.148 52.672
151022 94.761 -77.157 52.641
151023 94.772 -77.166 52.610
151024 94.784 -77.174 52.577
151025 94.803 -77.181 52.543
151026 94.822 -77.188 52.507
151027 94.841 -77.194 52.468
151028 94.860 -77.201 52.427
151029 94.879 -77.207 52.387
151030 94.898 -77.214 52.346
151031 94.914 -77.209 52.303
151032 94.930 -77.204 52.259
151033 94.946 -77.198 52.216
151034 94.962 -77.193 52.169
151035 94.978 -77.188 52.120
151036 94.994 -77.182 52.069
151037 95.016 -77.185 52.017
151038 95.040 -77.191 51.961
151039 95.064 -77.197 51.904
151040 95.088 -77.202 51.850
151041 95.112 -77.208 51.794
151042 95.136 -77.213 51.740
151043 95.160 -77.219 51.686
151044 95.174 -77.220 51.632
151045 95.187 -77.222 51.579
151046 95.202 -77.219 51.528
151047 95.218 -77.212 51.481
151048 95.234 -77.205 51.433
151049 95.250 -77.198 51.385
151050 95.266 -77.191 51.334
151051 95.282 -77.185 51.283
151052 95.298 -77.178 51.233
151053 95.314 -77.171 51.184
151054 95.330 -77.164 51.137
151055 95.346 -77.157 51.091
151056 95.362 -77.153 51.045
151057 95.385 -77.167 51.001
151058 95.407 -77.182 50.958
151059 95.429 -77.197 50.916
151060 95.447 -77.212 50.875
151061 95.464 -77.226 50.835
151062 95.482 -77.241 50.795
151063 95.500 -77.256 50.755
151064 95.518 -77.270 50.714
151065 95.541 -77.291 50.672
151066 95.564 -77.311 50.635
151067 95.588 -77.332 50.598
151068 95.611 -77.352 50.561
151069 95.634 -77.373 50.525
151070 95.658 -77.393 50.492
151071 95.676 -77.410 50.457
151072 95.692 -77.427 50.421
151073 95.708 -77.444 50.387
151074 95.725 -77.461 50.351
151075 95.741 -77.477 50.314
151076 95.757 -77.492 50.276
151077 95.773 -77.504 50.236
151078 95.790 -77.516 50.197
151079 95.806 -77.528 50.158
151080 95.822 -77.536 50.118
151081 95.838 -77.537 50.075
151082 95.854 -77.539 50.029
151083 95.871 -77.542 49.980
151084 95.887 -77.544 49.925
151085 95.903 -77.546 49.869
151086 95.918 -77.549 49.814
151087 95.934 -77.554 49.760
151088 95.950 -77.558 49.712
151089 95.966 -77.561 49.668
151090 95.981 -77.565 49.626
151091 95.988 -77.567 49.584
151092 95.995 -77.569 49.544
151093 96.002 -77.571 49.505
151094 96.011 -77.570 49.466
151095 96.021 -77.570 49.427
151096 96.032 -77.568 49.391
151097 96.042 -77.567 49.356
151098 96.052 -77.565 49.322
151099 96.063 -77.564 49.293
151100 96.074 -77.564 49.264
151101 96.089 -77.557 49.236
151102 96.104 -77.549 49.209
151103 96.098 -77.584 49.186
151104 96.094 -77.605 49.164
151105 96.091 -77.616 49.143
151106 96.088 -77.627 49.118
151107 96.085 -77.639 49.094
151108 96.089 -77.631 49.071
151109 96.091 -77.645 49.049
151110 96.106 -77.645 49.023
151111 96.120 -77.646 48.996
151112 96.135 -77.646 48.966
151113 96.150 -77.646 48.934
151114 96.165 -77.646 48.901
151115 96.179 -77.654 48.869
151116 96.193 -77.665 48.838
151117 96.207 -77.676 48.809
151118 96.220 -77.685 48.781
151119 96.232 -77.694 48.752
151120 96.239 -77.703 48.721
151121 96.245 -77.714 48.690
151122 96.250 -77.724 48.660
151123 96.256 -77.733 48.632
151124 96.262 -77.743 48.608
151125 96.267 -77.753 48.584
151126 96.284 -77.760 48.559
151127 96.304 -77.765 48.534
151128 96.327 -77.767 48.510
151129 96.354 -77.765 48.486
151130 96.381 -77.763 48.462
151131 96.407 -77.761 48.441
151132 96.434 -77.759 48.420
151133 96.461 -77.757 48.397
151134 96.475 -77.759 48.374
151135 96.483 -77.764 48.352
151136 96.492 -77.768 48.330
151137 96.501 -77.772 48.309
151138 96.510 -77.776 48.288
151139 96.518 -77.780 48.265
151140 96.527 -77.784 48.244
151141 96.532 -77.792 48.224
151142 96.542 -77.799 48.202
151143 96.551 -77.805 48.180
151144 96.561 -77.812 48.156
151145 96.571 -77.818 48.132
151146 96.581 -77.824 48.108
151147 96.590 -77.831 48.085
151148 96.606 -77.823 48.060
151149 96.619 -77.807 48.036
151150 96.635 -77.795 48.010
151151 96.650 -77.783 47.984
151152 96.665 -77.771 47.959
151153 96.679 -77.758 47.932
151154 96.694 -77.746 47.907
151155 96.710 -77.734 47.881
151156 96.730 -77.740 47.854
151157 96.752 -77.750 47.828
151158 96.774 -77.761 47.802
151159 96.796 -77.771 47.777
151160 96.818 -77.782 47.753
151161 96.840 -77.795 47.729
151162 96.863 -77.807 47.704
151163 96.885 -77.820 47.680
151164 96.907 -77.832 47.657
151165 96.929 -77.844 47.634
151166 96.947 -77.855 47.611
151167 96.956 -77.863 47.589
151168 96.965 -77.857 47.567
151169 96.973 -77.849 47.546
151170 96.981 -77.843 47.525
151171 96.990 -77.836 47.504
151172 97.001 -77.834 47.483
151173 97.012 -77.832 47.462
151174 97.022 -77.830 47.442
151175 97.034 -77.828 47.422
151176 97.047 -77.827 47.404
151177 97.061 -77.827 47.386
151178 97.074 -77.827 47.370
151179 97.087 -77.827 47.353
151180 97.106 -77.836 47.335
151181 97.131 -77.843 47.317
151182 97.161 -77.849 47.300
151183 97.192 -77.861 47.280
151184 97.223 -77.873 47.261
151185 97.255 -77.885 47.243
151186 97.286 -77.897 47.226
151187 97.317 -77.909 47.209
151188 97.344 -77.914 47.192
151189 97.367 -77.912 47.175
151190 97.389 -77.912 47.160
151191 97.397 -77.917 47.144
151192 97.410 -77.931 47.127
151193 97.425 -77.950 47.108
151194 97.437 -77.969 47.090
151195 97.449 -77.989 47.072
151196 97.461 -78.009 47.054
151197 97.473 -78.013 47.036
151198 97.492 -78.017 47.019
151199 97.513 -78.038 47.000
151200 97.532 -78.068 46.981
151201 97.545 -78.082 46.963
151202 97.559 -78.097 46.943
151203 97.572 -78.112 46.923
151204 97.585 -78.126 46.903
151205 97.599 -78.141 46.884
151206 97.604 -78.147 46.865
151207 97.603 -78.148 46.846
151208 97.614 -78.162 46.826
151209 97.636 -78.190 46.807
151210 97.662 -78.211 46.788
151211 97.698 -78.216 46.770
151212 97.734 -78.221 46.750
151213 97.770 -78.225 46.731
151214 97.803 -78.232 46.711
151215 97.831 -78.238 46.688
151216 97.860 -78.245 46.665
151217 97.888 -78.253 46.642
151218 97.916 -78.259 46.620
151219 97.944 -78.249 46.598
151220 98.007 -78.238 46.576
151221 98.070 -78.227 46.554
151222 98.129 -78.216 46.530
151223 98.150 -78.204 46.507
151224 98.171 -78.193 46.482
151225 98.193 -78.181 46.458
151226 98.214 -78.170 46.435
151227 98.217 -78.151 46.412
151228 98.247 -78.153 46.388
151229 98.277 -78.154 46.366
151230 98.306 -78.156 46.341
151231 98.337 -78.157 46.317
151232 98.366 -78.159 46.293
151233 98.396 -78.160 46.269
151234 98.426 -78.162 46.246
151235 98.456 -78.163 46.224
151236 98.486 -78.165 46.202
151237 98.515 -78.166 46.180
151238 98.545 -78.168 46.157
151239 98.576 -78.168 46.134
151240 98.609 -78.160 46.111
151241 98.600 -78.163 46.087
151242 98.588 -78.158 46.065
151243 98.575 -78.153 46.043
151244 98.562 -78.150 46.020
151245 98.549 -78.145 46.001
151246 98.536 -78.141 45.980
151247 98.526 -78.139 45.958
151248 98.527 -78.147 45.939
151249 98.528 -78.153 45.919
151250 98.579 -78.161 45.900
151251 98.632 -78.169 45.881
151252 98.684 -78.177 45.864
151253 98.729 -78.184 45.846
151254 98.773 -78.191 45.829
151255 98.816 -78.199 45.814
151256 98.859 -78.206 45.799
151257 98.903 -78.213 45.785
151258 98.946 -78.220 45.771
151259 98.984 -78.224 45.756
151260 99.013 -78.221 45.741
151261 99.042 -78.223 45.726
151262 99.068 -78.238 45.712
151263 99.095 -78.253 45.697
151264 99.113 -78.264 45.683
151265 99.106 -78.264 45.670
151266 99.110 -78.265 45.657
151267 99.133 -78.268 45.644
151268 99.155 -78.271 45.632
151269 99.177 -78.273 45.620
151270 99.200 -78.276 45.607
151271 99.222 -78.279 45.595
151272 99.251 -78.285 45.583
151273 99.283 -78.293 45.570
151274 99.315 -78.301 45.556
151275 99.348 -78.309 45.543
151276 99.380 -78.317 45.532
151277 99.413 -78.325 45.519
151278 99.445 -78.332 45.507
151279 99.478 -78.341 45.496
151280 99.510 -78.348 45.484
151281 99.542 -78.356 45.472
151282 99.574 -78.364 45.460
151283 99.579 -78.384 45.448
151284 99.565 -78.411 45.434
151285 99.552 -78.438 45.422
151286 99.539 -78.465 45.411
151287 99.584 -78.479 45.401
151288 99.668 -78.485 45.390
151289 99.753 -78.490 45.378
151290 99.832 -78.498 45.367
151291 99.863 -78.526 45.355
151292 99.895 -78.536 45.343
151293 99.922 -78.546 45.331
151294 99.949 -78.556 45.319
151295 99.977 -78.565 45.307
151296 100.004 -78.575 45.294
151297 100.031 -78.585 45.282
151298 100.059 -78.595 45.268
151299 100.086 -78.604 45.254
151300 100.108 -78.613 45.242
151301 100.117 -78.620 45.231
151302 100.126 -78.626 45.219
151303 100.135 -78.633 45.205
151304 100.177 -78.647 45.189
151305 100.210 -78.664 45.173
151306 100.228 -78.676 45.157
151307 100.246 -78.687 45.141
151308 100.264 -78.699 45.124
151309 100.282 -78.711 45.106
151310 100.300 -78.722 45.088
151311 100.318 -78.734 45.071
151312 100.336 -78.745 45.054
151313 100.354 -78.758 45.038
151314 100.372 -78.769 45.021
151315 100.390 -78.780 45.004
151316 100.409 -78.793 44.987
151317 100.427 -78.804 44.972
151318 100.446 -78.815 44.956
151319 100.466 -78.827 44.940
151320 100.486 -78.836 44.924
151321 100.502 -78.852 44.907
151322 100.518 -78.869 44.890
151323 100.535 -78.886 44.874
151324 100.551 -78.903 44.857
151325 100.567 -78.920 44.840
151326 100.573 -78.946 44.823
151327 100.578 -78.971 44.806
151328 100.583 -78.998 44.790
151329 100.599 -79.023 44.772
151330 100.617 -79.049 44.755
151331 100.636 -79.068 44.738
151332 100.664 -79.056 44.720
151333 100.692 -79.046 44.703
151334 100.741 -79.053 44.687
151335 100.793 -79.059 44.670
151336 100.845 -79.065 44.654
151337 100.897 -79.071 44.638
151338 100.949 -79.078 44.623
151339 101.001 -79.084 44.608
151340 101.053 -79.090 44.594
151341 101.105 -79.096 44.579
151342 101.157 -79.103 44.565
151343 101.209 -79.109 44.550
151344 101.261 -79.115 44.536
151345 101.294 -79.126 44.524
151346 101.319 -79.140 44.511
151347 101.343 -79.153 44.498
151348 101.368 -79.166 44.484
151580 92.626 -77.746 54.210
151581 92.646 -77.754 54.209
151582 92.666 -77.761 54.207
151583 92.686 -77.768 54.205
151584 92.706 -77.776 54.204
151585 92.726 -77.783 54.202
151586 92.746 -77.790 54.201
151587 92.766 -77.798 54.200
151588 92.786 -77.805 54.199
151589 92.806 -77.812 54.199
151590 92.826 -77.820 54.198
151591 92.846 -77.827 54.198
151592 92.866 -77.834 54.198
151593 92.886 -77.842 54.198
151594 92.906 -77.849 54.198
151595 92.926 -77.857 54.199
151596 92.950 -77.853 54.200
151597 92.970 -77.846 54.201
151598 92.986 -77.841 54.202
151599 93.003 -77.836 54.205
151600 93.020 -77.830 54.206
151601 93.037 -77.825 54.209
151602 93.054 -77.820 54.213
151603 93.071 -77.815 54.220
151604 93.087 -77.809 54.225
151605 93.107 -77.802 54.228
151606 93.118 -77.789 54.229
151607 93.129 -77.776 54.228
151608 93.140 -77.762 54.226
151609 93.151 -77.749 54.223
151610 93.162 -77.735 54.220
151611 93.174 -77.722 54.215
151612 93.185 -77.708 54.210
151613 93.196 -77.696 54.204
151614 93.208 -77.683 54.200
151615 93.219 -77.671 54.195
151616 93.231 -77.658 54.190
151617 93.242 -77.645 54.185
151618 93.254 -77.632 54.179
151619 93.267 -77.620 54.174
151620 93.281 -77.610 54.168
151621 93.295 -77.600 54.160
151622 93.309 -77.591 54.151
151623 93.323 -77.581 54.143
151624 93.337 -77.571 54.136
151625 93.353 -77.564 54.128
151626 93.368 -77.556 54.121
151627 93.384 -77.548 54.114
151628 93.401 -77.540 54.107
151629 93.416 -77.532 54.099
151630 93.432 -77.523 54.089
151631 93.448 -77.515 54.079
151632 93.464 -77.507 54.067
151633 93.480 -77.498 54.053
151634 93.494 -77.492 54.040
151635 93.507 -77.489 54.029
151636 93.520 -77.486 54.016
151637 93.534 -77.483 54.003
151638 93.548 -77.478 53.991
151639 93.566 -77.473 53.975
151640 93.580 -77.467 53.958
151641 93.587 -77.460 53.940
151642 93.593 -77.452 53.922
151643 93.600 -77.445 53.905
151644 93.606 -77.438 53.888
151645 93.613 -77.431 53.872
151646 93.626 -77.423 53.855
151647 93.641 -77.414 53.838
151648 93.662 -77.403 53.824
151649 93.684 -77.392 53.811
151650 93.709 -77.379 53.799
151651 93.735 -77.367 53.788
151652 93.761 -77.354 53.774
151653 93.786 -77.341 53.756
151654 93.811 -77.329 53.736
151655 93.827 -77.319 53.718
151656 93.843 -77.308 53.701
151657 93.853 -77.307 53.683
151658 93.855 -77.322 53.670
151659 93.866 -77.314 53.656
151660 93.886 -77.285 53.642
151661 93.893 -77.291 53.628
151662 93.901 -77.297 53.612
151663 93.909 -77.298 53.595
151664 93.918 -77.291 53.579
151665 93.927 -77.286 53.562
151666 93.942 -77.289 53.545
151667 93.959 -77.295 53.528
151668 93.976 -77.302 53.511
151669 93.993 -77.308 53.497
151670 94.010 -77.315 53.484
151671 94.023 -77.304 53.472
151672 94.031 -77.291 53.460
151673 94.037 -77.278 53.447
151674 94.067 -77.278 53.433
151675 94.096 -77.277 53.417
151676 94.126 -77.276 53.397
151677 94.155 -77.276 53.378
151678 94.185 -77.276 53.359
151679 94.214 -77.275 53.343
151680 94.243 -77.273 53.333
151681 94.263 -77.268 53.326
151682 94.282 -77.262 53.321
151683 94.298 -77.256 53.318
151684 94.298 -77.249 53.314
151685 94.299 -77.242 53.311
151686 94.300 -77.235 53.306
151687 94.301 -77.229 53.299
151688 94.302 -77.222 53.291
151689 94.303 -77.216 53.283
151690 94.302 -77.218 53.274
151691 94.302 -77.218 53.266
151692 94.309 -77.214 53.258
151693 94.310 -77.213 53.249
151694 94.311 -77.211 53.239
151695 94.312 -77.210 53.227
151696 94.312 -77.208 53.215
151697 94.317 -77.203 53.204
151698 94.330 -77.191 53.193
151699 94.343 -77.179 53.181
151700 94.356 -77.166 53.166
151701 94.378 -77.162 53.148
151702 94.401 -77.158 53.135
151703 94.423 -77.153 53.125
151704 94.445 -77.150 53.115
151705 94.467 -77.145 53.102
151706 94.487 -77.141 53.085
151707 94.507 -77.136 53.067
151708 94.527 -77.132 53.045
151709 94.547 -77.127 53.023
151710 94.567 -77.122 53.002
151711 94.588 -77.118 52.983
151712 94.608 -77.113 52.965
151713 94.624 -77.110 52.945
151714 94.634 -77.110 52.923
151715 94.643 -77.110 52.898
151716 94.655 -77.111 52.872
151717 94.669 -77.115 52.845
151718 94.682 -77.119 52.819
151719 94.695 -77.122 52.794
151720 94.709 -77.125 52.766
151721 94.722 -77.129 52.733
151722 94.735 -77.133 52.697
151723 94.751 -77.139 52.663
151724 94.769 -77.146 52.629
151725 94.788 -77.153 52.594
151726 94.807 -77.160 52.557
151727 94.826 -77.166 52.520
151728 94.845 -77.173 52.481
151729 94.864 -77.180 52.442
151730 94.883 -77.187 52.403
151731 94.902 -77.194 52.362
151732 94.918 -77.190 52.319
151733 94.934 -77.185 52.275
151734 94.951 -77.179 52.230
151735 94.967 -77.174 52.182
151736 94.983 -77.169 52.132
151737 95.000 -77.166 52.079
151738 95.022 -77.169 52.024
151739 95.042 -77.172 51.968
151740 95.065 -77.177 51.912
151741 95.089 -77.182 51.856
151742 95.113 -77.188 51.805
151743 95.137 -77.194 51.754
151744 95.161 -77.199 51.703
151745 95.175 -77.201 51.654
151746 95.189 -77.202 51.604
151747 95.202 -77.204 51.555
151748 95.217 -77.199 51.507
151749 95.234 -77.192 51.460
151750 95.250 -77.185 51.411
151751 95.265 -77.179 51.361
151752 95.281 -77.172 51.311
151753 95.298 -77.165 51.260
151754 95.313 -77.158 51.210
151755 95.329 -77.151 51.163
151756 95.345 -77.144 51.117
151757 95.366 -77.152 51.071
151758 95.388 -77.166 51.026
151759 95.410 -77.182 50.982
151760 95.432 -77.196 50.940
151761 95.450 -77.211 50.899
151762 95.468 -77.226 50.859
151763 95.485 -77.240 50.818
151764 95.503 -77.255 50.776
151765 95.521 -77.269 50.734
151766 95.539 -77.285 50.693
151767 95.561 -77.304 50.657
151768 95.585 -77.325 50.620
151769 95.608 -77.345 50.582
151770 95.631 -77.365 50.543
151771 95.655 -77.385 50.506
151772 95.673 -77.404 50.470
151773 95.689 -77.420 50.433
151774 95.706 -77.437 50.397
151775 95.722 -77.454 50.360
151776 95.738 -77.470 50.324
151777 95.755 -77.487 50.288
151778 95.771 -77.504 50.252
151779 95.788 -77.518 50.215
151780 95.804 -77.530 50.176
151781 95.820 -77.537 50.136
151782 95.836 -77.539 50.093
151783 95.852 -77.541 50.047
151784 95.868 -77.543 49.997
151785 95.884 -77.545 49.942
151786 95.900 -77.548 49.886
151787 95.916 -77.550 49.831
151788 95.933 -77.552 49.778
151789 95.949 -77.555 49.729
151790 95.965 -77.557 49.682
151791 95.980 -77.561 49.638
151792 95.990 -77.563 49.596
151793 96.000 -77.562 49.555
151794 96.010 -77.561 49.515
151795 96.020 -77.561 49.475
151796 96.029 -77.561 49.435
151797 96.038 -77.560 49.397
151798 96.047 -77.560 49.361
151799 96.057 -77.560 49.327
151800 96.066 -77.560 49.298
151801 96.076 -77.558 49.267
151802 96.092 -77.534 49.237
151803 96.103 -77.534 49.208
151804 96.096 -77.572 49.182
151805 96.091 -77.598 49.159
151806 96.088 -77.609 49.136
151807 96.091 -77.606 49.112
151808 96.096 -77.598 49.090
151809 96.098 -77.601 49.069
151810 96.094 -77.626 49.049
151811 96.107 -77.627 49.025
151812 96.122 -77.627 48.999
151813 96.137 -77.628 48.971
151814 96.152 -77.628 48.941
151815 96.167 -77.629 48.911
151816 96.180 -77.640 48.880
151817 96.194 -77.652 48.848
151818 96.208 -77.663 48.821
151819 96.222 -77.673 48.792
151820 96.234 -77.682 48.761
151821 96.247 -77.690 48.731
151822 96.257 -77.700 48.703
151823 96.263 -77.710 48.674
151824 96.268 -77.720 48.647
151825 96.274 -77.730 48.623
151826 96.287 -77.737 48.598
151827 96.307 -77.743 48.574
151828 96.326 -77.748 48.549
151829 96.346 -77.753 48.525
151830 96.367 -77.758 48.501
151831 96.393 -77.756 48.477
151832 96.419 -77.755 48.456
151833 96.447 -77.752 48.434
151834 96.470 -77.752 48.413
151835 96.479 -77.755 48.391
151836 96.487 -77.760 48.369
151837 96.497 -77.764 48.349
151838 96.505 -77.768 48.329
151839 96.514 -77.772 48.309
151840 96.523 -77.776 48.287
151841 96.534 -77.780 48.266
151842 96.543 -77.782 48.245
151843 96.552 -77.788 48.223
151844 96.562 -77.794 48.200
151845 96.572 -77.801 48.177
151846 96.581 -77.808 48.154
151847 96.591 -77.814 48.131
151848 96.600 -77.821 48.107
151849 96.612 -77.825 48.082
151850 96.627 -77.815 48.058
151851 96.642 -77.802 48.032
151852 96.657 -77.790 48.006
151853 96.672 -77.778 47.980
151854 96.687 -77.766 47.953
151855 96.702 -77.754 47.927
151856 96.717 -77.742 47.901
151857 96.734 -77.736 47.874
151858 96.756 -77.746 47.849
151859 96.778 -77.757 47.823
151860 96.800 -77.767 47.798
151861 96.822 -77.777 47.774
151862 96.844 -77.787 47.750
151863 96.866 -77.798 47.725
151864 96.888 -77.810 47.701
151865 96.910 -77.822 47.678
151866 96.933 -77.835 47.655
151867 96.954 -77.847 47.632
151868 96.962 -77.850 47.610
151869 96.971 -77.843 47.587
151870 96.979 -77.837 47.565
151871 96.990 -77.834 47.544
151872 97.001 -77.832 47.523
151873 97.011 -77.830 47.502
151874 97.022 -77.827 47.482
151875 97.033 -77.825 47.461
151876 97.044 -77.824 47.442
151877 97.058 -77.824 47.423
151878 97.071 -77.823 47.406
151879 97.085 -77.823 47.389
151880 97.098 -77.822 47.372
151881 97.115 -77.825 47.355
151882 97.147 -77.837 47.336
151883 97.178 -77.849 47.318
151884 97.209 -77.862 47.298
151885 97.241 -77.874 47.279
151886 97.272 -77.886 47.261
151887 97.303 -77.898 47.244
151888 97.327 -77.898 47.226
151889 97.350 -77.897 47.210
151890 97.373 -77.896 47.193
151891 97.396 -77.895 47.177
151892 97.406 -77.899 47.161
151893 97.415 -77.910 47.144
151894 97.428 -77.930 47.123
151895 97.440 -77.949 47.105
151896 97.453 -77.970 47.087
151897 97.466 -77.988 47.068
151898 97.481 -78.002 47.050
151899 97.498 -78.012 47.032
151900 97.517 -78.015 47.013
151901 97.538 -78.037 46.993
151902 97.555 -78.059 46.974
151903 97.568 -78.074 46.954
151904 97.581 -78.088 46.933
151905 97.595 -78.103 46.912
151906 97.605 -78.115 46.892
151907 97.604 -78.116 46.872
151908 97.604 -78.116 46.852
151909 97.609 -78.125 46.833
151910 97.631 -78.152 46.814
151911 97.653 -78.180 46.795
151912 97.677 -78.203 46.777
151913 97.713 -78.209 46.759
151914 97.747 -78.214 46.739
151915 97.776 -78.221 46.719
151916 97.804 -78.228 46.697
151917 97.832 -78.235 46.674
151918 97.860 -78.242 46.652
151919 97.884 -78.244 46.631
151920 97.933 -78.234 46.609
151921 97.997 -78.222 46.586
151922 98.060 -78.212 46.563
151923 98.121 -78.200 46.540
151924 98.143 -78.189 46.517
151925 98.164 -78.178 46.493
151926 98.185 -78.166 46.469
151927 98.206 -78.154 46.446
151928 98.217 -78.145 46.423
151929 98.247 -78.147 46.399
151930 98.277 -78.148 46.376
151931 98.307 -78.150 46.351
151932 98.337 -78.151 46.327
151933 98.367 -78.153 46.304
151934 98.396 -78.154 46.280
151935 98.426 -78.156 46.257
151936 98.456 -78.157 46.235
151937 98.486 -78.159 46.212
151938 98.519 -78.154 46.190
151939 98.556 -78.140 46.167
151940 98.592 -78.125 46.144
151941 98.617 -78.137 46.121
151942 98.613 -78.147 46.099
151943 98.600 -78.142 46.077
151944 98.587 -78.138 46.055
151945 98.574 -78.134 46.033
151946 98.561 -78.129 46.013
151947 98.548 -78.125 45.992
151948 98.544 -78.128 45.970
151949 98.545 -78.135 45.951
151950 98.546 -78.142 45.931
151951 98.567 -78.150 45.912
151952 98.620 -78.157 45.893
151953 98.673 -78.166 45.875
151954 98.725 -78.173 45.858
151955 98.778 -78.181 45.840
151956 98.822 -78.188 45.825
151957 98.866 -78.196 45.810
151958 98.909 -78.203 45.796
151959 98.952 -78.210 45.781
151960 98.995 -78.217 45.766
151961 99.030 -78.218 45.751
151962 99.059 -78.216 45.736
151963 99.086 -78.228 45.722
151964 99.113 -78.243 45.707
151965 99.127 -78.252 45.693
151966 99.120 -78.253 45.679
151967 99.118 -78.253 45.667
151968 99.140 -78.256 45.654
151969 99.162 -78.259 45.642
151970 99.185 -78.261 45.630
151971 99.207 -78.264 45.618
151972 99.229 -78.267 45.605
151973 99.252 -78.269 45.593
151974 99.279 -78.275 45.581
151975 99.311 -78.282 45.567
151976 99.343 -78.291 45.554
151977 99.376 -78.298 45.544
151978 99.408 -78.307 45.531
151979 99.440 -78.314 45.520
151980 99.473 -78.322 45.509
151981 99.505 -78.330 45.496
151982 99.538 -78.338 45.484
151983 99.564 -78.348 45.472
151984 99.551 -78.376 45.460
151985 99.537 -78.402 45.448
151986 99.524 -78.429 45.435
151987 99.523 -78.454 45.424
151988 99.571 -78.468 45.413
151989 99.655 -78.473 45.402
151990 99.739 -78.479 45.390
151991 99.817 -78.487 45.379
151992 99.849 -78.515 45.367
151993 99.882 -78.529 45.355
151994 99.909 -78.539 45.343
151995 99.937 -78.548 45.331
151996 99.964 -78.558 45.319
151997 99.991 -78.568 45.305
151998 100.018 -78.578 45.292
151999 100.046 -78.588 45.279
152000 100.073 -78.597 45.265
152001 100.101 -78.607 45.252
152002 100.128 -78.617 45.240
152003 100.147 -78.625 45.229
152004 100.156 -78.632 45.215
152005 100.186 -78.644 45.200
152006 100.210 -78.658 45.183
152007 100.228 -78.669 45.168
152008 100.246 -78.681 45.151
152009 100.264 -78.692 45.134
152010 100.282 -78.704 45.116
152011 100.300 -78.716 45.098
152012 100.319 -78.727 45.081
152013 100.338 -78.739 45.064
152014 100.357 -78.750 45.048
152015 100.376 -78.761 45.030
152016 100.395 -78.773 45.013
152017 100.414 -78.785 44.997
152018 100.433 -78.796 44.980
152019 100.452 -78.808 44.963
152020 100.471 -78.818 44.947
152021 100.492 -78.825 44.931
152022 100.512 -78.833 44.914
152023 100.528 -78.851 44.898
152024 100.545 -78.868 44.882
152025 100.561 -78.884 44.865
152026 100.574 -78.904 44.848
152027 100.579 -78.930 44.831
152028 100.584 -78.957 44.815
152029 100.588 -78.983 44.798
152030 100.594 -79.009 44.780
152031 100.610 -79.034 44.763
152032 100.628 -79.060 44.747
152033 100.650 -79.068 44.730
152034 100.680 -79.049 44.713
152035 100.740 -79.043 44.696
152036 100.798 -79.042 44.679
152037 100.850 -79.049 44.663
152038 100.902 -79.055 44.647
152039 100.954 -79.061 44.631
152040 101.006 -79.067 44.617
152041 101.058 -79.073 44.602
152042 101.110 -79.080 44.587
152043 101.162 -79.086 44.573
152044 101.214 -79.092 44.558
152045 101.266 -79.098 44.544
152046 101.293 -79.111 44.532
152047 101.318 -79.125 44.518
152048 101.342 -79.138 44.505
152049 101.367 -79.152 44.492
152050 101.391 -79.165 44.478
152281 92.615 -77.727 54.205
152282 92.635 -77.733 54.204
152283 92.655 -77.741 54.203
152284 92.675 -77.749 54.201
152285 92.695 -77.756 54.201
152286 92.715 -77.763 54.199
152287 92.735 -77.771 54.199
152288 92.755 -77.778 54.198
152289 92.775 -77.785 54.197
152290 92.795 -77.793 54.196
152291 92.815 -77.800 54.197
152292 92.835 -77.807 54.197
152293 92.855 -77.815 54.196
152294 92.875 -77.822 54.197
152295 92.895 -77.829 54.198
152296 92.915 -77.837 54.200
152297 92.939 -77.834 54.201
152298 92.959 -77.826 54.203
152299 92.977 -77.821 54.204
152300 92.993 -77.815 54.206
152301 93.010 -77.810 54.208
152302 93.027 -77.805 54.211
152303 93.044 -77.799 54.216
152304 93.061 -77.794 54.223
152305 93.078 -77.789 54.229
152306 93.092 -77.780 54.233
152307 93.105 -77.774 54.234
152308 93.116 -77.762 54.232
152309 93.128 -77.749 54.230
152310 93.139 -77.736 54.228
152311 93.151 -77.724 54.226
152312 93.162 -77.711 54.221
152313 93.174 -77.699 54.216
152314 93.186 -77.686 54.212
152315 93.198 -77.673 54.208
152316 93.209 -77.660 54.205
152317 93.221 -77.647 54.201
152318 93.232 -77.634 54.198
152319 93.245 -77.622 54.193
152320 93.258 -77.610 54.187
152321 93.271 -77.597 54.181
152322 93.284 -77.585 54.174
152323 93.297 -77.576 54.166
152324 93.311 -77.566 54.158
152325 93.326 -77.556 54.151
152326 93.341 -77.548 54.142
152327 93.357 -77.541 54.134
152328 93.373 -77.533 54.127
152329 93.389 -77.524 54.120
152330 93.405 -77.516 54.112
152331 93.421 -77.507 54.103
152332 93.437 -77.498 54.093
152333 93.452 -77.490 54.080
152334 93.469 -77.482 54.067
152335 93.484 -77.474 54.054
152336 93.497 -77.471 54.041
152337 93.510 -77.467 54.028
152338 93.523 -77.464 54.016
152339 93.539 -77.459 54.004
152340 93.557 -77.454 53.991
152341 93.574 -77.448 53.975
152342 93.584 -77.442 53.958
152343 93.590 -77.434 53.942
152344 93.597 -77.427 53.926
152345 93.604 -77.420 53.910
152346 93.615 -77.412 53.894
152347 93.630 -77.403 53.877
152348 93.645 -77.394 53.862
152349 93.663 -77.385 53.848
152350 93.685 -77.373 53.835
152351 93.706 -77.363 53.822
152352 93.728 -77.351 53.810
152353 93.753 -77.339 53.797
152354 93.779 -77.326 53.779
152355 93.804 -77.314 53.760
152356 93.826 -77.302 53.741
152357 93.842 -77.291 53.722
152358 93.850 -77.294 53.703
152359 93.852 -77.310 53.689
152360 93.861 -77.308 53.675
152361 93.877 -77.288 53.660
152362 93.885 -77.294 53.646
152363 93.893 -77.295 53.631
152364 93.902 -77.289 53.616
152365 93.911 -77.284 53.600
152366 93.920 -77.279 53.583
152367 93.934 -77.281 53.565
152368 93.952 -77.287 53.547
152369 93.969 -77.293 53.530
152370 93.986 -77.300 53.515
152371 94.000 -77.294 53.502
152372 94.011 -77.282 53.489
152373 94.020 -77.269 53.479
152374 94.036 -77.260 53.467
152375 94.066 -77.260 53.453
152376 94.096 -77.260 53.438
152377 94.125 -77.259 53.421
152378 94.154 -77.258 53.402
152379 94.184 -77.257 53.383
152380 94.214 -77.257 53.367
152381 94.243 -77.257 53.357
152382 94.268 -77.254 53.351
152383 94.287 -77.248 53.347
152384 94.302 -77.241 53.346
152385 94.303 -77.235 53.346
152386 94.305 -77.228 53.344
152387 94.306 -77.221 53.341
152388 94.307 -77.214 53.335
152389 94.307 -77.207 53.329
152390 94.308 -77.200 53.322
152391 94.309 -77.194 53.316
152392 94.309 -77.194 53.309
152393 94.312 -77.192 53.302
152394 94.313 -77.191 53.294
152395 94.314 -77.189 53.282
152396 94.315 -77.188 53.270
152397 94.316 -77.186 53.258
152398 94.317 -77.185 53.247
152399 94.328 -77.174 53.235
152400 94.340 -77.161 53.222
152401 94.354 -77.150 53.206
152402 94.377 -77.146 53.187
152403 94.399 -77.141 53.172
152404 94.422 -77.138 53.160
152405 94.443 -77.134 53.150
152406 94.464 -77.129 53.136
152407 94.483 -77.124 53.120
152408 94.503 -77.119 53.101
152409 94.524 -77.115 53.079
152410 94.544 -77.110 53.057
152411 94.564 -77.106 53.036
152412 94.584 -77.101 53.015
152413 94.604 -77.097 52.996
152414 94.623 -77.093 52.976
152415 94.633 -77.092 52.954
152416 94.643 -77.093 52.929
152417 94.657 -77.097 52.903
152418 94.670 -77.100 52.877
152419 94.683 -77.103 52.850
152420 94.697 -77.107 52.823
152421 94.710 -77.111 52.793
152422 94.722 -77.113 52.759
152423 94.735 -77.116 52.721
152424 94.757 -77.119 52.682
152425 94.773 -77.125 52.643
152426 94.792 -77.132 52.604
152427 94.811 -77.138 52.565
152428 94.830 -77.145 52.526
152429 94.849 -77.152 52.487
152430 94.867 -77.159 52.449
152431 94.887 -77.166 52.410
152432 94.906 -77.172 52.368
152433 94.922 -77.171 52.326
152434 94.938 -77.166 52.283
152435 94.955 -77.160 52.236
152436 94.971 -77.154 52.186
152437 94.987 -77.149 52.134
152438 95.007 -77.150 52.079
152439 95.028 -77.153 52.025
152440 95.049 -77.157 51.972
152441 95.069 -77.160 51.919
152442 95.091 -77.163 51.868
152443 95.114 -77.168 51.824
152444 95.138 -77.173 51.780
152445 95.162 -77.179 51.735
152446 95.177 -77.181 51.687
152447 95.190 -77.182 51.638
152448 95.204 -77.183 51.588
152449 95.217 -77.185 51.537
152450 95.233 -77.179 51.488
152451 95.249 -77.172 51.439
152452 95.265 -77.165 51.390
152453 95.281 -77.158 51.339
152454 95.297 -77.151 51.287
152455 95.313 -77.144 51.237
152456 95.329 -77.138 51.190
152457 95.346 -77.136 51.143
152458 95.369 -77.151 51.097
152459 95.391 -77.166 51.052
152460 95.413 -77.181 51.008
152461 95.435 -77.195 50.966
152462 95.453 -77.210 50.924
152463 95.471 -77.225 50.883
152464 95.489 -77.239 50.840
152465 95.507 -77.254 50.797
152466 95.524 -77.269 50.755
152467 95.542 -77.284 50.716
152468 95.560 -77.298 50.681
152469 95.582 -77.317 50.643
152470 95.605 -77.338 50.604
152471 95.628 -77.358 50.561
152472 95.652 -77.379 50.521
152473 95.671 -77.397 50.482
152474 95.687 -77.414 50.443
152475 95.703 -77.430 50.406
152476 95.720 -77.447 50.369
152477 95.736 -77.464 50.333
152478 95.752 -77.480 50.299
152479 95.769 -77.497 50.266
152480 95.785 -77.514 50.231
152481 95.801 -77.530 50.194
152482 95.818 -77.539 50.154
152483 95.835 -77.540 50.110
152484 95.851 -77.542 50.064
152485 95.868 -77.544 50.014
152486 95.885 -77.546 49.959
152487 95.901 -77.548 49.904
152488 95.918 -77.550 49.849
152489 95.935 -77.552 49.796
152490 95.952 -77.554 49.745
152491 95.968 -77.556 49.697
152492 95.983 -77.556 49.651
152493 95.993 -77.558 49.608
152494 96.002 -77.558 49.566
152495 96.012 -77.558 49.524
152496 96.021 -77.558 49.483
152497 96.030 -77.558 49.442
152498 96.039 -77.558 49.403
152499 96.049 -77.558 49.367
152500 96.058 -77.558 49.332
152501 96.067 -77.558 49.301
152502 96.078 -77.555 49.269
152503 96.094 -77.527 49.238
152504 96.105 -77.511 49.207
152505 96.098 -77.551 49.181
152506 96.093 -77.582 49.156
152507 96.097 -77.573 49.131
152508 96.102 -77.565 49.108
152509 96.107 -77.557 49.087
152510 96.105 -77.574 49.067
152511 96.101 -77.599 49.048
152512 96.109 -77.609 49.026
152513 96.124 -77.609 49.002
152514 96.138 -77.609 48.976
152515 96.154 -77.610 48.948
152516 96.168 -77.616 48.920
152517 96.182 -77.627 48.890
152518 96.195 -77.639 48.859
152519 96.209 -77.650 48.831
152520 96.223 -77.661 48.802
152521 96.236 -77.670 48.771
152522 96.249 -77.679 48.743
152523 96.261 -77.687 48.715
152524 96.274 -77.696 48.688
152525 96.281 -77.706 48.662
152526 96.289 -77.715 48.637
152527 96.309 -77.721 48.613
152528 96.329 -77.725 48.588
152529 96.349 -77.730 48.564
152530 96.369 -77.736 48.539
152531 96.389 -77.741 48.515
152532 96.409 -77.746 48.492
152533 96.432 -77.748 48.471
152534 96.459 -77.746 48.450
152535 96.474 -77.748 48.429
152536 96.483 -77.752 48.408
152537 96.492 -77.756 48.387
152538 96.502 -77.760 48.368
152539 96.514 -77.763 48.348
152540 96.525 -77.767 48.328
152541 96.536 -77.770 48.307
152542 96.546 -77.773 48.286
152543 96.553 -77.774 48.264
152544 96.561 -77.777 48.241
152545 96.571 -77.784 48.217
152546 96.580 -77.791 48.195
152547 96.590 -77.797 48.172
152548 96.600 -77.804 48.150
152549 96.609 -77.810 48.127
152550 96.620 -77.815 48.103
152551 96.635 -77.818 48.078
152552 96.650 -77.811 48.052
152553 96.664 -77.799 48.026
152554 96.679 -77.787 48.001
152555 96.694 -77.774 47.974
152556 96.710 -77.762 47.947
152557 96.724 -77.750 47.921
152558 96.739 -77.738 47.895
152559 96.759 -77.743 47.870
152560 96.782 -77.753 47.845
152561 96.804 -77.763 47.819
152562 96.826 -77.774 47.795
152563 96.848 -77.783 47.771
152564 96.870 -77.794 47.747
152565 96.892 -77.805 47.722
152566 96.914 -77.815 47.699
152567 96.936 -77.825 47.676
152568 96.958 -77.838 47.653
152569 96.968 -77.837 47.631
152570 96.979 -77.834 47.607
152571 96.990 -77.832 47.585
152572 97.001 -77.830 47.564
152573 97.011 -77.827 47.543
152574 97.022 -77.826 47.522
152575 97.033 -77.824 47.502
152576 97.044 -77.821 47.481
152577 97.055 -77.820 47.461
152578 97.069 -77.819 47.443
152579 97.082 -77.819 47.425
152580 97.095 -77.818 47.408
152581 97.113 -77.821 47.390
152582 97.133 -77.826 47.372
152583 97.164 -77.837 47.354
152584 97.195 -77.849 47.334
152585 97.227 -77.862 47.315
152586 97.258 -77.874 47.297
152587 97.288 -77.884 47.279
152588 97.311 -77.883 47.261
152589 97.334 -77.882 47.244
152590 97.357 -77.881 47.228
152591 97.380 -77.879 47.211
152592 97.403 -77.878 47.194
152593 97.415 -77.882 47.178
152594 97.425 -77.894 47.160
152595 97.437 -77.914 47.139
152596 97.450 -77.934 47.121
152597 97.463 -77.953 47.103
152598 97.478 -77.967 47.084
152599 97.494 -77.981 47.065
152600 97.509 -77.993 47.046
152601 97.524 -78.007 47.026
152602 97.542 -78.014 47.006
152603 97.563 -78.034 46.985
152604 97.577 -78.051 46.964
152605 97.591 -78.065 46.943
152606 97.604 -78.080 46.922
152607 97.605 -78.083 46.902
152608 97.605 -78.084 46.882
152609 97.604 -78.084 46.861
152610 97.607 -78.100 46.841
152611 97.629 -78.125 46.821
152612 97.650 -78.150 46.801
152613 97.671 -78.175 46.783
152614 97.691 -78.197 46.765
152615 97.720 -78.203 46.745
152616 97.748 -78.210 46.724
152617 97.776 -78.218 46.703
152618 97.804 -78.225 46.682
152619 97.832 -78.232 46.661
152620 97.860 -78.229 46.641
152621 97.923 -78.218 46.619
152622 97.986 -78.207 46.596
152623 98.050 -78.196 46.572
152624 98.112 -78.185 46.550
152625 98.135 -78.174 46.527
152626 98.156 -78.162 46.504
152627 98.177 -78.150 46.480
152628 98.188 -78.138 46.456
152629 98.218 -78.139 46.433
152630 98.248 -78.141 46.410
152631 98.278 -78.142 46.386
152632 98.307 -78.144 46.361
152633 98.338 -78.145 46.338
152634 98.367 -78.147 46.314
152635 98.397 -78.148 46.291
152636 98.427 -78.150 46.268
152637 98.461 -78.141 46.246
152638 98.498 -78.125 46.222
152639 98.535 -78.111 46.199
152640 98.572 -78.097 46.176
152641 98.608 -78.082 46.154
152642 98.624 -78.113 46.131
152643 98.625 -78.131 46.110
152644 98.612 -78.127 46.089
152645 98.599 -78.122 46.068
152646 98.586 -78.118 46.046
152647 98.573 -78.113 46.025
152648 98.561 -78.109 46.004
152649 98.562 -78.116 45.981
152650 98.562 -78.124 45.962
152651 98.563 -78.131 45.943
152652 98.564 -78.138 45.924
152653 98.608 -78.146 45.906
152654 98.661 -78.153 45.888
152655 98.714 -78.162 45.870
152656 98.767 -78.169 45.853
152657 98.819 -78.178 45.836
152658 98.872 -78.185 45.821
152659 98.915 -78.192 45.806
152660 98.958 -78.200 45.792
152661 99.002 -78.206 45.776
152662 99.045 -78.214 45.761
152663 99.076 -78.213 45.746
152664 99.104 -78.218 45.732
152665 99.131 -78.233 45.718
152666 99.141 -78.241 45.703
152667 99.134 -78.241 45.688
152668 99.128 -78.241 45.675
152669 99.148 -78.244 45.663
152670 99.170 -78.247 45.651
152671 99.192 -78.249 45.639
152672 99.215 -78.252 45.627
152673 99.237 -78.255 45.615
152674 99.259 -78.257 45.603
152675 99.282 -78.260 45.591
152676 99.306 -78.264 45.578
152677 99.339 -78.272 45.565
152678 99.371 -78.280 45.554
152679 99.403 -78.288 45.542
152680 99.436 -78.296 45.531
152681 99.468 -78.304 45.520
152682 99.500 -78.311 45.508
152683 99.533 -78.319 45.496
152684 99.536 -78.340 45.483
152685 99.522 -78.366 45.471
152686 99.509 -78.394 45.460
152687 99.517 -78.417 45.448
152688 99.547 -78.436 45.436
152689 99.575 -78.455 45.425
152690 99.642 -78.462 45.414
152691 99.726 -78.468 45.403
152692 99.802 -78.476 45.391
152693 99.835 -78.505 45.379
152694 99.867 -78.529 45.367
152695 99.897 -78.532 45.354
152696 99.924 -78.542 45.342
152697 99.951 -78.551 45.330
152698 99.979 -78.561 45.316
152699 100.006 -78.571 45.303
152700 100.033 -78.580 45.291
152701 100.060 -78.590 45.276
152702 100.088 -78.600 45.262
152703 100.115 -78.610 45.249
152704 100.142 -78.620 45.237
152705 100.170 -78.629 45.224
152706 100.192 -78.639 45.209
152707 100.210 -78.651 45.193
152708 100.229 -78.662 45.178
152709 100.249 -78.673 45.161
152710 100.267 -78.685 45.143
152711 100.286 -78.697 45.125
152712 100.305 -78.708 45.108
152713 100.324 -78.720 45.090
152714 100.343 -78.731 45.073
152715 100.362 -78.742 45.057
152716 100.381 -78.754 45.040
152717 100.400 -78.765 45.022
152718 100.419 -78.777 45.005
152719 100.438 -78.788 44.989
152720 100.457 -78.800 44.972
152721 100.477 -78.809 44.955
152722 100.498 -78.816 44.939
152723 100.519 -78.823 44.923
152724 100.538 -78.832 44.906
152725 100.555 -78.849 44.890
152726 100.571 -78.866 44.874
152727 100.579 -78.890 44.857
152728 100.584 -78.915 44.840
152729 100.590 -78.942 44.823
152730 100.594 -78.968 44.806
152731 100.599 -78.994 44.789
152732 100.604 -79.021 44.772
152733 100.623 -79.040 44.755
152734 100.648 -79.037 44.739
152735 100.673 -79.034 44.722
152736 100.701 -79.046 44.705
152737 100.793 -79.037 44.688
152738 100.855 -79.032 44.672
152739 100.907 -79.038 44.656
152740 100.960 -79.044 44.640
152741 101.011 -79.050 44.625
152742 101.063 -79.057 44.611
152743 101.116 -79.063 44.596
152744 101.168 -79.069 44.581
152745 101.219 -79.075 44.566
152746 101.268 -79.083 44.552
152747 101.292 -79.096 44.539
152748 101.317 -79.109 44.526
152749 101.341 -79.123 44.513
152750 101.365 -79.137 44.500
152751 101.390 -79.150 44.487
152752 101.412 -79.169 44.474
152981 92.584 -77.699 54.200
152982 92.604 -77.706 54.199
152983 92.624 -77.714 54.199
152984 92.644 -77.721 54.198
152985 92.664 -77.728 54.197
152986 92.684 -77.736 54.198
152987 92.704 -77.743 54.197
152988 92.724 -77.751 54.196
152989 92.744 -77.758 54.196
152990 92.764 -77.765 54.195
152991 92.784 -77.773 54.195
152992 92.804 -77.780 54.195
152993 92.824 -77.787 54.195
152994 92.844 -77.795 54.195
152995 92.864 -77.802 54.196
152996 92.883 -77.809 54.198
152997 92.904 -77.817 54.201
152998 92.928 -77.813 54.202
152999 92.950 -77.805 54.203
153000 92.967 -77.799 54.205
153001 92.984 -77.794 54.208
153002 93.000 -77.790 54.210
153003 93.017 -77.784 54.214
153004 93.034 -77.779 54.219
153005 93.051 -77.774 54.226
153006 93.066 -77.766 54.233
153007 93.079 -77.757 54.238
153008 93.092 -77.749 54.238
153009 93.105 -77.743 54.236
153010 93.117 -77.736 54.234
153011 93.129 -77.727 54.232
153012 93.141 -77.714 54.230
153013 93.152 -77.701 54.227
153014 93.164 -77.688 54.223
153015 93.176 -77.675 54.220
153016 93.187 -77.662 54.217
153017 93.199 -77.650 54.215
153018 93.210 -77.637 54.212
153019 93.222 -77.624 54.211
153020 93.235 -77.612 54.206
153021 93.248 -77.599 54.201
153022 93.261 -77.587 54.195
153023 93.274 -77.575 54.188
153024 93.287 -77.563 54.181
153025 93.300 -77.551 54.173
153026 93.314 -77.541 54.166
153027 93.330 -77.533 54.157
153028 93.346 -77.525 54.149
153029 93.362 -77.516 54.142
153030 93.377 -77.508 54.134
153031 93.394 -77.499 54.126
153032 93.409 -77.491 54.118
153033 93.425 -77.483 54.108
153034 93.441 -77.474 54.097
153035 93.457 -77.466 54.084
153036 93.473 -77.457 54.072
153037 93.487 -77.452 54.060
153038 93.500 -77.449 54.048
153039 93.513 -77.445 54.036
153040 93.530 -77.441 54.024
153041 93.547 -77.435 54.011
153042 93.565 -77.429 53.995
153043 93.581 -77.423 53.980
153044 93.587 -77.417 53.965
153045 93.594 -77.409 53.949
153046 93.604 -77.401 53.933
153047 93.619 -77.392 53.917
153048 93.635 -77.384 53.902
153049 93.650 -77.375 53.886
153050 93.665 -77.366 53.872
153051 93.685 -77.355 53.859
153052 93.707 -77.345 53.845
153053 93.729 -77.333 53.833
153054 93.751 -77.323 53.819
153055 93.773 -77.311 53.803
153056 93.796 -77.299 53.783
153057 93.822 -77.287 53.763
153058 93.841 -77.276 53.743
153059 93.847 -77.282 53.724
153060 93.849 -77.297 53.708
153061 93.856 -77.301 53.693
153062 93.868 -77.292 53.678
153063 93.876 -77.293 53.663
153064 93.886 -77.287 53.649
153065 93.895 -77.282 53.635
153066 93.904 -77.276 53.619
153067 93.913 -77.270 53.602
153068 93.927 -77.272 53.585
153069 93.944 -77.279 53.566
153070 93.961 -77.285 53.548
153071 93.977 -77.284 53.532
153072 93.988 -77.272 53.519
153073 94.000 -77.259 53.507
153074 94.010 -77.246 53.495
153075 94.035 -77.243 53.483
153076 94.065 -77.242 53.468
153077 94.095 -77.241 53.453
153078 94.124 -77.241 53.436
153079 94.153 -77.241 53.418
153080 94.183 -77.240 53.401
153081 94.213 -77.239 53.387
153082 94.242 -77.239 53.379
153083 94.272 -77.238 53.373
153084 94.292 -77.233 53.370
153085 94.308 -77.227 53.369
153086 94.309 -77.220 53.369
153087 94.310 -77.213 53.370
153088 94.310 -77.207 53.370
153089 94.311 -77.200 53.367
153090 94.312 -77.192 53.363
153091 94.313 -77.185 53.360
153092 94.314 -77.179 53.355
153093 94.315 -77.172 53.349
153094 94.316 -77.170 53.342
153095 94.317 -77.169 53.335
153096 94.317 -77.167 53.323
153097 94.318 -77.166 53.310
153098 94.319 -77.164 53.298
153099 94.320 -77.163 53.286
153100 94.325 -77.157 53.275
153101 94.338 -77.144 53.262
153102 94.353 -77.134 53.246
153103 94.375 -77.129 53.224
153104 94.398 -77.125 53.207
153105 94.420 -77.122 53.193
153106 94.439 -77.117 53.181
153107 94.460 -77.113 53.168
153108 94.480 -77.108 53.152
153109 94.500 -77.103 53.134
153110 94.520 -77.099 53.111
153111 94.540 -77.094 53.089
153112 94.560 -77.089 53.067
153113 94.580 -77.085 53.046
153114 94.601 -77.080 53.026
153115 94.621 -77.075 53.006
153116 94.632 -77.075 52.984
153117 94.645 -77.078 52.959
153118 94.658 -77.081 52.933
153119 94.671 -77.085 52.906
153120 94.684 -77.088 52.879
153121 94.696 -77.090 52.850
153122 94.709 -77.092 52.819
153123 94.721 -77.094 52.783
153124 94.734 -77.097 52.744
153125 94.762 -77.101 52.703
153126 94.780 -77.106 52.660
153127 94.796 -77.111 52.616
153128 94.815 -77.118 52.574
153129 94.834 -77.125 52.532
153130 94.853 -77.132 52.491
153131 94.872 -77.138 52.450
153132 94.891 -77.144 52.409
153133 94.909 -77.151 52.367
153134 94.927 -77.151 52.325
153135 94.943 -77.146 52.283
153136 94.959 -77.141 52.237
153137 94.975 -77.135 52.187
153138 94.992 -77.132 52.136
153139 95.013 -77.135 52.083
153140 95.034 -77.138 52.031
153141 95.055 -77.141 51.982
153142 95.075 -77.144 51.936
153143 95.097 -77.147 51.893
153144 95.117 -77.150 51.853
153145 95.139 -77.153 51.813
153146 95.163 -77.159 51.771
153147 95.178 -77.161 51.723
153148 95.191 -77.163 51.673
153149 95.205 -77.163 51.622
153150 95.218 -77.165 51.568
153151 95.233 -77.166 51.520
153152 95.248 -77.159 51.471
153153 95.264 -77.152 51.421
153154 95.280 -77.145 51.367
153155 95.297 -77.138 51.315
153156 95.312 -77.132 51.264
153157 95.328 -77.125 51.217
153158 95.349 -77.135 51.171
153159 95.372 -77.150 51.126
153160 95.394 -77.165 51.081
153161 95.416 -77.180 51.036
153162 95.438 -77.194 50.992
153163 95.457 -77.210 50.949
153164 95.474 -77.224 50.906
153165 95.492 -77.238 50.860
153166 95.510 -77.254 50.817
153167 95.528 -77.268 50.777
153168 95.545 -77.283 50.741
153169 95.563 -77.298 50.704
153170 95.581 -77.312 50.665
153171 95.602 -77.331 50.623
153172 95.626 -77.351 50.578
153173 95.649 -77.372 50.535
153174 95.669 -77.390 50.493
153175 95.685 -77.407 50.452
153176 95.701 -77.423 50.416
153177 95.718 -77.440 50.379
153178 95.734 -77.457 50.341
153179 95.750 -77.473 50.307
153180 95.766 -77.490 50.277
153181 95.783 -77.503 50.245
153182 95.801 -77.515 50.209
153183 95.822 -77.527 50.169
153184 95.843 -77.537 50.126
153185 95.860 -77.539 50.080
153186 95.876 -77.542 50.030
153187 95.893 -77.543 49.978
153188 95.910 -77.545 49.924
153189 95.926 -77.547 49.869
153190 95.943 -77.549 49.815
153191 95.960 -77.551 49.762
153192 95.975 -77.551 49.712
153193 95.986 -77.545 49.664
153194 95.996 -77.547 49.620
153195 96.004 -77.555 49.576
153196 96.013 -77.556 49.533
153197 96.022 -77.556 49.491
153198 96.032 -77.556 49.449
153199 96.041 -77.556 49.409
153200 96.050 -77.555 49.372
153201 96.059 -77.555 49.337
153202 96.069 -77.555 49.303
153203 96.079 -77.552 49.271
153204 96.096 -77.524 49.239
153205 96.106 -77.504 49.207
153206 96.114 -77.489 49.180
153207 96.110 -77.517 49.155
153208 96.109 -77.533 49.130
153209 96.113 -77.524 49.107
153210 96.116 -77.523 49.085
153211 96.113 -77.548 49.064
153212 96.109 -77.573 49.046
153213 96.111 -77.591 49.027
153214 96.125 -77.591 49.005
153215 96.141 -77.591 48.981
153216 96.155 -77.591 48.955
153217 96.169 -77.602 48.928
153218 96.183 -77.614 48.900
153219 96.197 -77.625 48.868
153220 96.210 -77.636 48.840
153221 96.225 -77.648 48.811
153222 96.238 -77.658 48.782
153223 96.250 -77.667 48.755
153224 96.263 -77.675 48.727
153225 96.276 -77.684 48.700
153226 96.289 -77.693 48.675
153227 96.310 -77.700 48.651
153228 96.330 -77.705 48.627
153229 96.350 -77.710 48.602
153230 96.371 -77.715 48.576
153231 96.391 -77.720 48.552
153232 96.411 -77.725 48.529
153233 96.431 -77.730 48.507
153234 96.451 -77.735 48.487
153235 96.471 -77.740 48.467
153236 96.482 -77.743 48.447
153237 96.493 -77.746 48.426
153238 96.505 -77.750 48.406
153239 96.516 -77.753 48.386
153240 96.527 -77.757 48.366
153241 96.539 -77.761 48.346
153242 96.549 -77.764 48.325
153243 96.556 -77.765 48.303
153244 96.563 -77.766 48.280
153245 96.570 -77.768 48.255
153246 96.579 -77.774 48.231
153247 96.589 -77.780 48.208
153248 96.599 -77.787 48.187
153249 96.608 -77.793 48.167
153250 96.618 -77.800 48.145
153251 96.629 -77.805 48.122
153252 96.643 -77.808 48.098
153253 96.657 -77.810 48.071
153254 96.672 -77.807 48.045
153255 96.687 -77.795 48.019
153256 96.702 -77.783 47.993
153257 96.717 -77.771 47.968
153258 96.732 -77.758 47.943
153259 96.747 -77.746 47.917
153260 96.763 -77.739 47.892
153261 96.785 -77.749 47.866
153262 96.807 -77.759 47.840
153263 96.829 -77.770 47.816
153264 96.852 -77.780 47.792
153265 96.874 -77.790 47.768
153266 96.896 -77.801 47.743
153267 96.918 -77.811 47.720
153268 96.940 -77.821 47.697
153269 96.962 -77.831 47.674
153270 96.977 -77.831 47.652
153271 96.990 -77.830 47.629
153272 97.001 -77.827 47.607
153273 97.011 -77.826 47.585
153274 97.022 -77.824 47.564
153275 97.033 -77.821 47.542
153276 97.044 -77.819 47.521
153277 97.054 -77.818 47.501
153278 97.066 -77.815 47.481
153279 97.079 -77.815 47.462
153280 97.093 -77.815 47.444
153281 97.110 -77.818 47.426
153282 97.131 -77.822 47.407
153283 97.151 -77.827 47.388
153284 97.181 -77.838 47.370
153285 97.213 -77.850 47.351
153286 97.244 -77.862 47.332
153287 97.272 -77.868 47.314
153288 97.295 -77.868 47.296
153289 97.317 -77.866 47.278
153290 97.341 -77.865 47.261
153291 97.363 -77.864 47.245
153292 97.386 -77.863 47.229
153293 97.409 -77.862 47.211
153294 97.424 -77.865 47.194
153295 97.434 -77.878 47.176
153296 97.447 -77.899 47.156
153297 97.460 -77.918 47.138
153298 97.475 -77.932 47.119
153299 97.490 -77.945 47.099
153300 97.506 -77.959 47.081
153301 97.521 -77.972 47.061
153302 97.536 -77.986 47.040
153303 97.551 -78.000 47.020
153304 97.567 -78.012 46.998
153305 97.586 -78.028 46.977
153306 97.600 -78.042 46.955
153307 97.606 -78.050 46.934
153308 97.606 -78.051 46.913
153309 97.606 -78.066 46.892
153310 97.607 -78.083 46.870
153311 97.610 -78.101 46.849
153312 97.631 -78.126 46.828
153313 97.652 -78.151 46.808
153314 97.673 -78.176 46.788
153315 97.703 -78.173 46.770
153316 97.737 -78.162 46.750
153317 97.720 -78.200 46.729
153318 97.748 -78.207 46.710
153319 97.777 -78.214 46.691
153320 97.805 -78.221 46.671
153321 97.850 -78.213 46.651
153322 97.913 -78.203 46.628
153323 97.976 -78.191 46.605
153324 98.039 -78.181 46.582
153325 98.102 -78.169 46.559
153326 98.128 -78.158 46.537
153327 98.149 -78.147 46.514
153328 98.162 -78.132 46.491
153329 98.189 -78.131 46.467
153330 98.219 -78.133 46.444
153331 98.248 -78.134 46.420
153332 98.278 -78.136 46.396
153333 98.308 -78.138 46.372
153334 98.338 -78.139 46.348
153335 98.368 -78.141 46.324
153336 98.405 -78.126 46.301
153337 98.441 -78.112 46.279
153338 98.478 -78.097 46.256
153339 98.514 -78.083 46.233
153340 98.551 -78.069 46.207
153341 98.588 -78.053 46.185
153342 98.618 -78.053 46.164
153343 98.631 -78.089 46.142
153344 98.637 -78.115 46.121
153345 98.624 -78.111 46.100
153346 98.611 -78.106 46.080
153347 98.598 -78.102 46.058
153348 98.585 -78.098 46.037
153349 98.578 -78.098 46.015
153350 98.579 -78.106 45.993
153351 98.580 -78.112 45.974
153352 98.581 -78.119 45.955
153353 98.581 -78.127 45.936
153354 98.597 -78.134 45.918
153355 98.650 -78.142 45.899
153356 98.703 -78.150 45.881
153357 98.756 -78.158 45.864
153358 98.808 -78.166 45.847
153359 98.861 -78.174 45.831
153360 98.913 -78.181 45.816
153361 98.964 -78.189 45.801
153362 99.008 -78.197 45.785
153363 99.051 -78.203 45.770
153364 99.093 -78.210 45.755
153365 99.122 -78.208 45.741
153366 99.149 -78.223 45.726
153367 99.155 -78.229 45.712
153368 99.148 -78.229 45.697
153369 99.142 -78.229 45.684
153370 99.155 -78.232 45.672
153371 99.178 -78.234 45.660
153372 99.200 -78.237 45.649
153373 99.222 -78.240 45.636
153374 99.244 -78.242 45.624
153375 99.267 -78.245 45.613
153376 99.289 -78.248 45.601
153377 99.311 -78.250 45.588
153378 99.334 -78.253 45.576
153379 99.366 -78.262 45.565
153380 99.399 -78.269 45.553
153381 99.431 -78.277 45.543
153382 99.463 -78.285 45.531
153383 99.496 -78.293 45.519
153384 99.521 -78.304 45.507
153385 99.508 -78.331 45.494
153386 99.494 -78.358 45.482
153387 99.512 -78.379 45.471
153388 99.541 -78.399 45.459
153389 99.570 -78.418 45.447
153390 99.599 -78.437 45.436
153391 99.629 -78.451 45.426
153392 99.713 -78.457 45.415
153393 99.788 -78.467 45.403
153394 99.820 -78.495 45.391
153395 99.852 -78.523 45.379
153396 99.884 -78.525 45.367
153397 99.911 -78.535 45.354
153398 99.938 -78.545 45.340
153399 99.966 -78.554 45.327
153400 99.993 -78.564 45.314
153401 100.020 -78.573 45.300
153402 100.048 -78.583 45.285
153403 100.075 -78.593 45.271
153404 100.102 -78.603 45.258
153405 100.130 -78.613 45.245
153406 100.162 -78.626 45.232
153407 100.196 -78.642 45.217
153408 100.216 -78.647 45.202
153409 100.234 -78.655 45.186
153410 100.254 -78.666 45.170
153411 100.273 -78.677 45.152
153412 100.291 -78.689 45.135
153413 100.311 -78.701 45.118
153414 100.330 -78.712 45.100
153415 100.348 -78.724 45.082
153416 100.367 -78.735 45.066
153417 100.387 -78.746 45.049
153418 100.405 -78.758 45.031
153419 100.424 -78.769 45.014
153420 100.443 -78.780 44.998
153421 100.462 -78.792 44.980
153422 100.483 -78.800 44.963
153423 100.503 -78.807 44.947
153424 100.524 -78.814 44.931
153425 100.545 -78.821 44.915
153426 100.565 -78.830 44.899
153427 100.580 -78.848 44.882
153428 100.585 -78.874 44.866
153429 100.590 -78.901 44.849
153430 100.595 -78.927 44.831
153431 100.600 -78.953 44.815
153432 100.606 -78.979 44.798
153433 100.612 -79.006 44.781
153434 100.635 -79.018 44.765
153435 100.659 -79.014 44.748
153436 100.683 -79.011 44.732
153437 100.707 -79.007 44.715
153438 100.746 -79.083 44.698
153439 100.846 -79.032 44.681
153440 100.913 -79.021 44.665
153441 100.965 -79.028 44.649
153442 101.016 -79.034 44.634
153443 101.069 -79.040 44.619
153444 101.121 -79.046 44.605
153445 101.173 -79.053 44.590
153446 101.225 -79.059 44.575
153447 101.267 -79.068 44.561
153448 101.291 -79.081 44.548
153449 101.316 -79.095 44.534
153450 101.340 -79.108 44.521
153451 101.364 -79.122 44.508
153452 101.389 -79.135 44.496
153453 101.411 -79.156 44.482
153454 101.427 -79.190 44.470
153681 92.553 -77.672 54.193
153682 92.573 -77.679 54.194
153683 92.593 -77.686 54.194
153684 92.613 -77.694 54.194
153685 92.633 -77.701 54.194
153686 92.653 -77.708 54.194
153687 92.673 -77.716 54.195
153688 92.693 -77.723 54.194
153689 92.713 -77.730 54.194
153690 92.733 -77.738 54.195
153691 92.753 -77.745 54.194
153692 92.773 -77.752 54.194
153693 92.793 -77.760 54.195
153694 92.813 -77.768 54.195
153695 92.833 -77.774 54.195
153696 92.853 -77.782 54.197
153697 92.873 -77.790 54.199
153698 92.893 -77.796 54.201
153699 92.917 -77.793 54.203
153700 92.940 -77.784 54.205
153701 92.957 -77.779 54.207
153702 92.974 -77.774 54.210
153703 92.991 -77.768 54.213
153704 93.007 -77.763 54.217
153705 93.024 -77.758 54.222
153706 93.040 -77.751 54.230
153707 93.053 -77.742 54.237
153708 93.066 -77.733 54.242
153709 93.080 -77.724 54.243
153710 93.092 -77.718 54.241
153711 93.105 -77.711 54.239
153712 93.117 -77.705 54.238
153713 93.130 -77.699 54.237
153714 93.142 -77.690 54.234
153715 93.153 -77.678 54.231
153716 93.165 -77.665 54.229
153717 93.177 -77.652 54.227
153718 93.188 -77.639 54.225
153719 93.200 -77.627 54.223
153720 93.212 -77.614 54.224
153721 93.225 -77.602 54.220
153722 93.238 -77.589 54.215
153723 93.251 -77.577 54.209
153724 93.264 -77.565 54.202
153725 93.277 -77.552 54.196
153726 93.290 -77.539 54.189
153727 93.303 -77.526 54.181
153728 93.319 -77.517 54.173
153729 93.334 -77.508 54.165
153730 93.350 -77.500 54.157
153731 93.366 -77.492 54.149
153732 93.382 -77.483 54.142
153733 93.398 -77.475 54.135
153734 93.414 -77.466 54.125
153735 93.430 -77.457 54.114
153736 93.445 -77.449 54.103
153737 93.462 -77.441 54.092
153738 93.477 -77.434 54.082
153739 93.490 -77.430 54.071
153740 93.503 -77.427 54.061
153741 93.520 -77.422 54.050
153742 93.538 -77.416 54.036
153743 93.556 -77.410 54.021
153744 93.573 -77.404 54.005
153745 93.585 -77.398 53.991
153746 93.593 -77.391 53.976
153747 93.608 -77.382 53.960
153748 93.624 -77.373 53.945
153749 93.639 -77.364 53.930
153750 93.654 -77.355 53.915
153751 93.670 -77.346 53.900
153752 93.686 -77.338 53.884
153753 93.707 -77.326 53.869
153754 93.729 -77.315 53.855
153755 93.751 -77.304 53.841
153756 93.773 -77.293 53.825
153757 93.795 -77.282 53.805
153758 93.817 -77.271 53.785
153759 93.840 -77.260 53.766
153760 93.845 -77.269 53.745
153761 93.847 -77.285 53.728
153762 93.851 -77.294 53.712
153763 93.860 -77.291 53.697
153764 93.869 -77.285 53.681
153765 93.878 -77.279 53.666
153766 93.888 -77.273 53.652
153767 93.897 -77.268 53.637
153768 93.906 -77.263 53.621
153769 93.919 -77.263 53.603
153770 93.936 -77.270 53.585
153771 93.953 -77.274 53.566
153772 93.965 -77.261 53.549
153773 93.977 -77.248 53.535
153774 93.989 -77.236 53.522
153775 94.004 -77.226 53.510
153776 94.035 -77.226 53.496
153777 94.064 -77.225 53.481
153778 94.094 -77.224 53.466
153779 94.123 -77.223 53.449
153780 94.152 -77.223 53.433
153781 94.182 -77.222 53.418
153782 94.212 -77.222 53.405
153783 94.242 -77.221 53.398
153784 94.271 -77.221 53.392
153785 94.298 -77.219 53.388
153786 94.313 -77.213 53.387
153787 94.313 -77.206 53.386
153788 94.314 -77.199 53.387
153789 94.315 -77.192 53.388
153790 94.316 -77.185 53.386
153791 94.316 -77.177 53.384
153792 94.314 -77.166 53.381
153793 94.312 -77.155 53.377
153794 94.318 -77.154 53.373
153795 94.319 -77.151 53.368
153796 94.320 -77.147 53.362
153797 94.321 -77.145 53.353
153798 94.321 -77.144 53.342
153799 94.322 -77.142 53.331
153800 94.323 -77.141 53.321
153801 94.324 -77.139 53.310
153802 94.336 -77.128 53.297
153803 94.351 -77.117 53.281
153804 94.373 -77.113 53.260
153805 94.396 -77.110 53.241
153806 94.416 -77.105 53.224
153807 94.436 -77.100 53.209
153808 94.456 -77.096 53.195
153809 94.476 -77.091 53.180
153810 94.496 -77.087 53.162
153811 94.517 -77.082 53.139
153812 94.537 -77.078 53.116
153813 94.557 -77.073 53.093
153814 94.577 -77.069 53.073
153815 94.597 -77.064 53.053
153816 94.617 -77.059 53.033
153817 94.632 -77.059 53.011
153818 94.645 -77.061 52.986
153819 94.657 -77.064 52.960
153820 94.670 -77.066 52.934
153821 94.683 -77.069 52.906
153822 94.695 -77.071 52.876
153823 94.708 -77.073 52.843
153824 94.720 -77.075 52.807
153825 94.733 -77.078 52.768
153826 94.763 -77.083 52.724
153827 94.795 -77.088 52.678
153828 94.804 -77.091 52.630
153829 94.819 -77.097 52.583
153830 94.838 -77.103 52.539
153831 94.856 -77.110 52.494
153832 94.876 -77.117 52.449
153833 94.895 -77.124 52.407
153834 94.913 -77.131 52.364
153835 94.931 -77.132 52.321
153836 94.947 -77.127 52.280
153837 94.963 -77.122 52.236
153838 94.979 -77.116 52.189
153839 94.998 -77.116 52.139
153840 95.019 -77.119 52.090
153841 95.040 -77.122 52.043
153842 95.061 -77.125 52.000
153843 95.082 -77.128 51.958
153844 95.103 -77.131 51.920
153845 95.124 -77.134 51.883
153846 95.144 -77.137 51.845
153847 95.165 -77.141 51.804
153848 95.179 -77.141 51.756
153849 95.193 -77.142 51.705
153850 95.206 -77.144 51.652
153851 95.220 -77.145 51.598
153852 95.234 -77.147 51.549
153853 95.248 -77.146 51.500
153854 95.264 -77.139 51.449
153855 95.280 -77.132 51.395
153856 95.296 -77.125 51.342
153857 95.312 -77.119 51.291
153858 95.330 -77.120 51.245
153859 95.352 -77.135 51.199
153860 95.375 -77.150 51.154
153861 95.397 -77.164 51.108
153862 95.419 -77.179 51.062
153863 95.441 -77.194 51.017
153864 95.460 -77.209 50.972
153865 95.478 -77.223 50.927
153866 95.495 -77.238 50.881
153867 95.513 -77.253 50.839
153868 95.531 -77.267 50.799
153869 95.549 -77.282 50.763
153870 95.566 -77.297 50.724
153871 95.584 -77.311 50.683
153872 95.602 -77.326 50.640
153873 95.623 -77.344 50.594
153874 95.646 -77.364 50.549
153875 95.666 -77.383 50.505
153876 95.682 -77.400 50.463
153877 95.699 -77.417 50.428
153878 95.715 -77.433 50.390
153879 95.732 -77.450 50.354
153880 95.748 -77.462 50.319
153881 95.765 -77.475 50.288
153882 95.782 -77.487 50.254
153883 95.803 -77.499 50.218
153884 95.824 -77.511 50.180
153885 95.845 -77.523 50.138
153886 95.867 -77.534 50.093
153887 95.885 -77.539 50.044
153888 95.901 -77.541 49.993
153889 95.918 -77.542 49.940
153890 95.935 -77.545 49.888
153891 95.951 -77.546 49.834
153892 95.967 -77.547 49.779
153893 95.978 -77.541 49.728
153894 95.989 -77.536 49.679
153895 96.000 -77.536 49.633
153896 96.007 -77.543 49.588
153897 96.015 -77.551 49.544
153898 96.024 -77.554 49.501
153899 96.033 -77.554 49.458
153900 96.042 -77.554 49.417
153901 96.052 -77.553 49.379
153902 96.061 -77.553 49.343
153903 96.071 -77.553 49.308
153904 96.081 -77.550 49.274
153905 96.097 -77.521 49.242
153906 96.106 -77.504 49.211
153907 96.113 -77.489 49.184
153908 96.121 -77.473 49.159
153909 96.122 -77.482 49.132
153910 96.125 -77.483 49.108
153911 96.124 -77.496 49.084
153912 96.120 -77.521 49.062
153913 96.117 -77.547 49.042
153914 96.113 -77.572 49.023
153915 96.127 -77.573 49.003
153916 96.142 -77.573 48.980
153917 96.157 -77.577 48.955
153918 96.170 -77.589 48.930
153919 96.184 -77.600 48.903
153920 96.198 -77.611 48.875
153921 96.212 -77.623 48.847
153922 96.226 -77.634 48.819
153923 96.240 -77.645 48.791
153924 96.252 -77.655 48.765
153925 96.265 -77.664 48.739
153926 96.278 -77.672 48.713
153927 96.298 -77.682 48.689
153928 96.322 -77.692 48.665
153929 96.344 -77.699 48.640
153930 96.365 -77.704 48.615
153931 96.385 -77.709 48.591
153932 96.405 -77.714 48.567
153933 96.425 -77.719 48.544
153934 96.445 -77.724 48.522
153935 96.465 -77.729 48.503
153936 96.484 -77.733 48.483
153937 96.496 -77.736 48.463
153938 96.507 -77.740 48.442
153939 96.519 -77.744 48.424
153940 96.530 -77.747 48.405
153941 96.541 -77.751 48.386
153942 96.552 -77.754 48.366
153943 96.559 -77.755 48.343
153944 96.566 -77.757 48.320
153945 96.573 -77.758 48.295
153946 96.580 -77.760 48.272
153947 96.588 -77.764 48.248
153948 96.598 -77.770 48.225
153949 96.607 -77.777 48.204
153950 96.617 -77.783 48.182
153951 96.626 -77.790 48.161
153952 96.638 -77.796 48.139
153953 96.652 -77.798 48.116
153954 96.666 -77.800 48.090
153955 96.680 -77.802 48.064
153956 96.694 -77.803 48.038
153957 96.709 -77.791 48.014
153958 96.724 -77.779 47.990
153959 96.739 -77.767 47.965
153960 96.754 -77.755 47.939
153961 96.769 -77.743 47.913
153962 96.789 -77.745 47.887
153963 96.811 -77.755 47.861
153964 96.833 -77.766 47.836
153965 96.855 -77.776 47.812
153966 96.877 -77.787 47.788
153967 96.899 -77.797 47.764
153968 96.922 -77.807 47.740
153969 96.944 -77.818 47.717
153970 96.966 -77.823 47.694
153971 96.983 -77.823 47.672
153972 96.996 -77.822 47.649
153973 97.009 -77.821 47.627
153974 97.021 -77.821 47.606
153975 97.033 -77.819 47.585
153976 97.044 -77.818 47.563
153977 97.054 -77.815 47.542
153978 97.066 -77.813 47.522
153979 97.077 -77.812 47.502
153980 97.090 -77.811 47.482
153981 97.108 -77.814 47.463
153982 97.128 -77.818 47.444
153983 97.149 -77.823 47.425
153984 97.169 -77.827 47.405
153985 97.198 -77.838 47.386
153986 97.230 -77.850 47.368
153987 97.255 -77.853 47.349
153988 97.278 -77.852 47.332
153989 97.301 -77.851 47.314
153990 97.324 -77.849 47.296
153991 97.347 -77.849 47.279
153992 97.370 -77.847 47.263
153993 97.393 -77.846 47.246
153994 97.416 -77.845 47.229
153995 97.433 -77.847 47.211
153996 97.444 -77.863 47.193
153997 97.456 -77.883 47.174
153998 97.471 -77.896 47.155
153999 97.487 -77.910 47.135
154000 97.502 -77.924 47.115
154001 97.517 -77.937 47.096
154002 97.533 -77.951 47.076
154003 97.548 -77.965 47.055
154004 97.563 -77.978 47.035
154005 97.579 -77.992 47.013
154006 97.594 -78.005 46.991
154007 97.607 -78.018 46.968
154008 97.608 -78.034 46.946
154009 97.609 -78.051 46.924
154010 97.610 -78.069 46.902
154011 97.611 -78.086 46.880
154012 97.613 -78.103 46.857
154013 97.634 -78.128 46.835
154014 97.655 -78.153 46.814
154015 97.676 -78.170 46.792
154016 97.711 -78.153 46.773
154017 97.746 -78.141 46.754
154018 97.781 -78.130 46.735
154019 97.764 -78.166 46.717
154020 97.749 -78.203 46.699
154021 97.776 -78.209 46.680
154022 97.840 -78.198 46.660
154023 97.902 -78.188 46.636
154024 97.965 -78.176 46.613
154025 98.029 -78.166 46.591
154026 98.092 -78.154 46.569
154027 98.120 -78.143 46.546
154028 98.141 -78.131 46.524
154029 98.159 -78.124 46.501
154030 98.189 -78.125 46.477
154031 98.219 -78.127 46.453
154032 98.249 -78.128 46.429
154033 98.279 -78.130 46.406
154034 98.310 -78.127 46.383
154035 98.347 -78.112 46.359
154036 98.384 -78.098 46.334
154037 98.421 -78.084 46.310
154038 98.457 -78.069 46.287
154039 98.494 -78.054 46.264
154040 98.531 -78.040 46.241
154041 98.568 -78.025 46.218
154042 98.604 -78.011 46.196
154043 98.625 -78.030 46.175
154044 98.638 -78.065 46.153
154045 98.649 -78.100 46.132
154046 98.636 -78.095 46.111
154047 98.624 -78.091 46.091
154048 98.610 -78.087 46.069
154049 98.597 -78.082 46.047
154050 98.595 -78.087 46.026
154051 98.596 -78.094 46.004
154052 98.597 -78.101 45.985
154053 98.598 -78.108 45.967
154054 98.599 -78.116 45.949
154055 98.599 -78.122 45.930
154056 98.639 -78.131 45.911
154057 98.691 -78.138 45.892
154058 98.744 -78.146 45.875
154059 98.797 -78.154 45.858
154060 98.849 -78.162 45.842
154061 98.902 -78.169 45.826
154062 98.955 -78.178 45.810
154063 99.007 -78.185 45.794
154064 99.057 -78.193 45.779
154065 99.100 -78.200 45.764
154066 99.139 -78.204 45.750
154067 99.167 -78.213 45.735
154068 99.169 -78.217 45.720
154069 99.162 -78.218 45.706
154070 99.156 -78.218 45.693
154071 99.163 -78.219 45.680
154072 99.185 -78.222 45.669
154073 99.208 -78.225 45.658
154074 99.230 -78.228 45.645
154075 99.252 -78.230 45.634
154076 99.274 -78.233 45.623
154077 99.297 -78.236 45.612
154078 99.319 -78.238 45.598
154079 99.341 -78.241 45.586
154080 99.364 -78.244 45.575
154081 99.394 -78.251 45.564
154082 99.426 -78.259 45.552
154083 99.458 -78.267 45.540
154084 99.491 -78.275 45.528
154085 99.493 -78.295 45.516
154086 99.480 -78.322 45.505
154087 99.507 -78.342 45.493
154088 99.536 -78.361 45.481
154089 99.565 -78.380 45.470
154090 99.593 -78.400 45.459
154091 99.623 -78.419 45.447
154092 99.651 -78.438 45.437
154093 99.700 -78.446 45.426
154094 99.774 -78.456 45.415
154095 99.805 -78.484 45.402
154096 99.838 -78.512 45.390
154097 99.871 -78.518 45.377
154098 99.898 -78.528 45.364
154099 99.925 -78.538 45.351
154100 99.953 -78.548 45.337
154101 99.980 -78.557 45.324
154102 100.007 -78.567 45.310
154103 100.034 -78.576 45.295
154104 100.062 -78.586 45.281
154105 100.095 -78.601 45.267
154106 100.129 -78.616 45.254
154107 100.162 -78.630 45.239
154108 100.195 -78.645 45.225
154109 100.228 -78.660 45.210
154110 100.241 -78.661 45.194
154111 100.259 -78.664 45.178
154112 100.278 -78.670 45.162
154113 100.297 -78.681 45.145
154114 100.316 -78.692 45.128
154115 100.335 -78.704 45.110
154116 100.354 -78.716 45.092
154117 100.372 -78.727 45.074
154118 100.392 -78.739 45.057
154119 100.411 -78.750 45.040
154120 100.429 -78.761 45.024
154121 100.449 -78.773 45.007
154122 100.468 -78.784 44.989
154123 100.488 -78.791 44.972
154124 100.509 -78.799 44.956
154125 100.530 -78.805 44.939
154126 100.551 -78.813 44.923
154127 100.571 -78.820 44.907
154128 100.587 -78.833 44.891
154129 100.593 -78.859 44.875
154130 100.599 -78.886 44.858
154131 100.606 -78.912 44.840
154132 100.612 -78.937 44.824
154133 100.618 -78.964 44.808
154134 100.629 -78.987 44.791
154135 100.657 -78.997 44.775
154136 100.683 -79.002 44.758
154137 100.707 -78.998 44.741
154138 100.731 -78.995 44.724
154139 100.754 -79.074 44.707
154140 100.799 -79.078 44.691
154141 100.899 -79.027 44.675
154142 100.970 -79.011 44.659
154143 101.022 -79.018 44.644
154144 101.074 -79.024 44.628
154145 101.126 -79.030 44.613
154146 101.178 -79.036 44.599
154147 101.230 -79.042 44.584
154148 101.266 -79.053 44.570
154149 101.290 -79.066 44.556
154150 101.315 -79.080 44.543
154151 101.339 -79.093 44.529
154152 101.363 -79.106 44.517
154153 101.388 -79.120 44.505
154154 101.409 -79.143 44.491
154155 101.425 -79.177 44.479
154382 92.542 -77.652 54.185
154383 92.562 -77.659 54.187
154384 92.582 -77.667 54.188
154385 92.602 -77.674 54.189
154386 92.622 -77.681 54.190
154387 92.642 -77.689 54.190
154388 92.662 -77.696 54.191
154389 92.682 -77.703 54.191
154390 92.702 -77.711 54.192
154391 92.722 -77.718 54.193
154392 92.742 -77.725 54.193
154393 92.762 -77.733 54.194
154394 92.782 -77.740 54.195
154395 92.802 -77.747 54.196
154396 92.821 -77.755 54.197
154397 92.842 -77.762 54.199
154398 92.862 -77.769 54.201
154399 92.881 -77.777 54.203
154400 92.906 -77.774 54.205
154401 92.930 -77.764 54.208
154402 92.947 -77.758 54.210
154403 92.964 -77.753 54.213
154404 92.981 -77.748 54.216
154405 92.998 -77.743 54.220
154406 93.013 -77.736 54.226
154407 93.026 -77.727 54.234
154408 93.040 -77.718 54.241
154409 93.054 -77.709 54.246
154410 93.067 -77.700 54.248
154411 93.080 -77.693 54.248
154412 93.092 -77.686 54.247
154413 93.105 -77.680 54.246
154414 93.117 -77.674 54.245
154415 93.130 -77.667 54.242
154416 93.142 -77.661 54.239
154417 93.154 -77.655 54.238
154418 93.166 -77.642 54.237
154419 93.178 -77.629 54.236
154420 93.190 -77.617 54.235
154421 93.203 -77.604 54.235
154422 93.216 -77.592 54.233
154423 93.228 -77.580 54.228
154424 93.240 -77.565 54.223
154425 93.252 -77.551 54.217
154426 93.263 -77.536 54.210
154427 93.274 -77.521 54.203
154428 93.295 -77.511 54.196
154429 93.314 -77.503 54.188
154430 93.328 -77.494 54.181
154431 93.342 -77.485 54.173
154432 93.355 -77.476 54.167
154433 93.370 -77.467 54.160
154434 93.387 -77.458 54.153
154435 93.403 -77.450 54.144
154436 93.418 -77.442 54.133
154437 93.434 -77.433 54.124
154438 93.450 -77.425 54.115
154439 93.466 -77.417 54.106
154440 93.480 -77.412 54.097
154441 93.493 -77.408 54.087
154442 93.511 -77.403 54.077
154443 93.529 -77.397 54.062
154444 93.546 -77.392 54.047
154445 93.564 -77.386 54.032
154446 93.582 -77.380 54.019
154447 93.597 -77.372 54.003
154448 93.613 -77.363 53.989
154449 93.628 -77.354 53.975
154450 93.643 -77.345 53.959
154451 93.659 -77.336 53.944
154452 93.674 -77.327 53.928
154453 93.690 -77.318 53.913
154454 93.708 -77.308 53.897
154455 93.730 -77.298 53.882
154456 93.752 -77.286 53.867
154457 93.774 -77.275 53.851
154458 93.795 -77.264 53.831
154459 93.817 -77.253 53.810
154460 93.834 -77.232 53.790
154461 93.839 -77.215 53.769
154462 93.842 -77.238 53.751
154463 93.847 -77.267 53.734
154464 93.854 -77.279 53.716
154465 93.862 -77.277 53.699
154466 93.871 -77.271 53.683
154467 93.880 -77.266 53.667
154468 93.890 -77.260 53.651
154469 93.899 -77.254 53.635
154470 93.912 -77.255 53.618
154471 93.929 -77.261 53.600
154472 93.941 -77.251 53.582
154473 93.954 -77.238 53.565
154474 93.967 -77.226 53.549
154475 93.983 -77.216 53.535
154476 94.000 -77.208 53.521
154477 94.025 -77.206 53.506
154478 94.056 -77.206 53.490
154479 94.087 -77.205 53.477
154480 94.118 -77.205 53.462
154481 94.149 -77.205 53.447
154482 94.180 -77.204 53.433
154483 94.211 -77.204 53.422
154484 94.241 -77.204 53.414
154485 94.270 -77.203 53.408
154486 94.299 -77.203 53.404
154487 94.317 -77.198 53.402
154488 94.318 -77.191 53.402
154489 94.318 -77.183 53.402
154490 94.316 -77.172 53.402
154491 94.314 -77.161 53.400
154492 94.311 -77.150 53.398
154493 94.309 -77.139 53.395
154494 94.314 -77.137 53.392
154495 94.321 -77.137 53.389
154496 94.323 -77.134 53.386
154497 94.324 -77.130 53.383
154498 94.324 -77.126 53.376
154499 94.325 -77.122 53.368
154500 94.325 -77.120 53.359
154501 94.326 -77.119 53.351
154502 94.327 -77.117 53.341
154503 94.333 -77.110 53.328
154504 94.350 -77.101 53.313
154505 94.372 -77.097 53.291
154506 94.392 -77.093 53.270
154507 94.412 -77.088 53.250
154508 94.432 -77.084 53.232
154509 94.453 -77.079 53.216
154510 94.473 -77.075 53.198
154511 94.493 -77.070 53.179
154512 94.513 -77.066 53.156
154513 94.534 -77.061 53.133
154514 94.555 -77.056 53.111
154515 94.576 -77.053 53.089
154516 94.597 -77.049 53.071
154517 94.620 -77.044 53.051
154518 94.632 -77.041 53.028
154519 94.644 -77.043 53.004
154520 94.657 -77.045 52.980
154521 94.669 -77.047 52.953
154522 94.682 -77.050 52.925
154523 94.694 -77.052 52.894
154524 94.707 -77.054 52.861
154525 94.719 -77.056 52.826
154526 94.732 -77.059 52.787
154527 94.764 -77.064 52.743
154528 94.804 -77.071 52.696
154529 94.818 -77.075 52.647
154530 94.826 -77.077 52.596
154531 94.842 -77.082 52.548
154532 94.860 -77.089 52.501
154533 94.880 -77.096 52.455
154534 94.898 -77.103 52.407
154535 94.917 -77.110 52.361
154536 94.935 -77.113 52.316
154537 94.952 -77.108 52.274
154538 94.968 -77.102 52.234
154539 94.984 -77.097 52.191
154540 95.004 -77.100 52.147
154541 95.025 -77.103 52.104
154542 95.046 -77.106 52.065
154543 95.067 -77.109 52.026
154544 95.088 -77.112 51.989
154545 95.109 -77.115 51.951
154546 95.130 -77.118 51.913
154547 95.150 -77.122 51.875
154548 95.170 -77.126 51.834
154549 95.186 -77.128 51.786
154550 95.197 -77.125 51.735
154551 95.208 -77.124 51.681
154552 95.222 -77.125 51.626
154553 95.235 -77.126 51.574
154554 95.249 -77.128 51.524
154555 95.263 -77.125 51.474
154556 95.279 -77.119 51.421
154557 95.295 -77.112 51.369
154558 95.311 -77.106 51.318
154559 95.333 -77.119 51.271
154560 95.355 -77.134 51.227
154561 95.378 -77.149 51.181
154562 95.400 -77.163 51.135
154563 95.422 -77.179 51.088
154564 95.444 -77.193 51.040
154565 95.463 -77.208 50.993
154566 95.481 -77.222 50.947
154567 95.498 -77.237 50.901
154568 95.517 -77.252 50.860
154569 95.534 -77.266 50.821
154570 95.552 -77.282 50.783
154571 95.570 -77.296 50.741
154572 95.588 -77.310 50.699
154573 95.605 -77.326 50.655
154574 95.623 -77.340 50.609
154575 95.643 -77.357 50.563
154576 95.664 -77.376 50.519
154577 95.680 -77.393 50.477
154578 95.696 -77.409 50.439
154579 95.713 -77.421 50.402
154580 95.730 -77.434 50.366
154581 95.747 -77.446 50.332
154582 95.764 -77.458 50.299
154583 95.782 -77.471 50.263
154584 95.804 -77.483 50.226
154585 95.825 -77.495 50.189
154586 95.847 -77.506 50.148
154587 95.868 -77.518 50.105
154588 95.890 -77.530 50.057
154589 95.909 -77.538 50.007
154590 95.926 -77.539 49.956
154591 95.943 -77.542 49.905
154592 95.958 -77.542 49.852
154593 95.970 -77.536 49.796
154594 95.981 -77.531 49.743
154595 95.992 -77.525 49.693
154596 96.003 -77.524 49.646
154597 96.011 -77.532 49.601
154598 96.018 -77.540 49.556
154599 96.026 -77.548 49.511
154600 96.034 -77.551 49.467
154601 96.044 -77.551 49.425
154602 96.053 -77.551 49.386
154603 96.062 -77.551 49.349
154604 96.072 -77.551 49.314
154605 96.082 -77.547 49.279
154606 96.098 -77.520 49.246
154607 96.105 -77.504 49.215
154608 96.113 -77.489 49.188
154609 96.121 -77.473 49.161
154610 96.128 -77.457 49.134
154611 96.135 -77.445 49.109
154612 96.131 -77.470 49.084
154613 96.128 -77.495 49.060
154614 96.124 -77.520 49.038
154615 96.121 -77.545 49.018
154616 96.129 -77.555 48.998
154617 96.144 -77.555 48.976
154618 96.158 -77.564 48.953
154619 96.172 -77.576 48.929
154620 96.185 -77.587 48.904
154621 96.199 -77.599 48.879
154622 96.213 -77.610 48.853
154623 96.227 -77.621 48.826
154624 96.241 -77.633 48.800
154625 96.254 -77.642 48.775
154626 96.267 -77.652 48.750
154627 96.282 -77.661 48.725
154628 96.307 -77.671 48.701
154629 96.332 -77.681 48.677
154630 96.358 -77.692 48.653
154631 96.379 -77.698 48.629
154632 96.399 -77.703 48.606
154633 96.419 -77.708 48.583
154634 96.440 -77.713 48.560
154635 96.460 -77.718 48.538
154636 96.480 -77.723 48.518
154637 96.499 -77.727 48.499
154638 96.510 -77.730 48.479
154639 96.521 -77.734 48.459
154640 96.532 -77.737 48.442
154641 96.544 -77.741 48.424
154642 96.555 -77.745 48.405
154643 96.562 -77.746 48.385
154644 96.569 -77.748 48.362
154645 96.576 -77.749 48.337
154646 96.583 -77.750 48.312
154647 96.590 -77.752 48.289
154648 96.597 -77.753 48.265
154649 96.606 -77.760 48.242
154650 96.616 -77.767 48.221
154651 96.626 -77.773 48.199
154652 96.636 -77.780 48.177
154653 96.646 -77.786 48.155
154654 96.660 -77.788 48.133
154655 96.674 -77.790 48.108
154656 96.688 -77.793 48.083
154657 96.702 -77.795 48.058
154658 96.717 -77.797 48.036
154659 96.732 -77.787 48.011
154660 96.747 -77.774 47.986
154661 96.761 -77.762 47.960
154662 96.777 -77.750 47.934
154663 96.793 -77.741 47.908
154664 96.815 -77.752 47.882
154665 96.837 -77.762 47.856
154666 96.859 -77.772 47.832
154667 96.881 -77.783 47.808
154668 96.903 -77.793 47.784
154669 96.925 -77.803 47.760
154670 96.947 -77.813 47.737
154671 96.970 -77.814 47.715
154672 96.989 -77.815 47.692
154673 97.001 -77.814 47.669
154674 97.014 -77.813 47.648
154675 97.027 -77.812 47.627
154676 97.039 -77.812 47.606
154677 97.052 -77.812 47.584
154678 97.065 -77.811 47.563
154679 97.076 -77.809 47.542
154680 97.087 -77.807 47.523
154681 97.105 -77.810 47.503
154682 97.126 -77.815 47.483
154683 97.146 -77.819 47.463
154684 97.166 -77.824 47.444
154685 97.186 -77.828 47.424
154686 97.216 -77.839 47.404
154687 97.239 -77.838 47.385
154688 97.262 -77.837 47.367
154689 97.285 -77.836 47.349
154690 97.308 -77.834 47.333
154691 97.330 -77.834 47.314
154692 97.354 -77.832 47.298
154693 97.377 -77.831 47.281
154694 97.399 -77.830 47.264
154695 97.423 -77.829 47.247
154696 97.442 -77.829 47.229
154697 97.453 -77.847 47.211
154698 97.468 -77.862 47.192
154699 97.484 -77.875 47.173
154700 97.499 -77.889 47.153
154701 97.514 -77.902 47.132
154702 97.529 -77.916 47.112
154703 97.545 -77.930 47.091
154704 97.560 -77.944 47.070
154705 97.576 -77.959 47.050
154706 97.591 -77.975 47.028
154707 97.607 -77.990 47.005
154708 97.614 -78.011 46.983
154709 97.612 -78.036 46.960
154710 97.613 -78.053 46.937
154711 97.614 -78.071 46.914
154712 97.615 -78.088 46.890
154713 97.616 -78.106 46.867
154714 97.637 -78.130 46.844
154715 97.658 -78.155 46.821
154716 97.684 -78.144 46.799
154717 97.720 -78.132 46.779
154718 97.754 -78.121 46.759
154719 97.790 -78.109 46.740
154720 97.824 -78.098 46.721
154721 97.813 -78.129 46.705
154722 97.798 -78.169 46.687
154723 97.829 -78.183 46.667
154724 97.892 -78.172 46.644
154725 97.955 -78.161 46.622
154726 98.018 -78.150 46.600
154727 98.081 -78.139 46.578
154728 98.112 -78.128 46.555
154729 98.130 -78.116 46.532
154730 98.160 -78.118 46.509
154731 98.190 -78.119 46.485
154732 98.220 -78.122 46.461
154733 98.253 -78.113 46.438
154734 98.290 -78.099 46.416
154735 98.327 -78.084 46.393
154736 98.364 -78.069 46.368
154737 98.401 -78.055 46.343
154738 98.437 -78.040 46.319
154739 98.474 -78.026 46.297
154740 98.510 -78.012 46.274
154741 98.547 -77.996 46.251
154742 98.584 -77.982 46.229
154743 98.620 -77.970 46.207
154744 98.632 -78.006 46.186
154745 98.645 -78.042 46.164
154746 98.658 -78.078 46.142
154747 98.648 -78.079 46.122
154748 98.636 -78.075 46.101
154749 98.623 -78.071 46.079
154750 98.613 -78.069 46.057
154751 98.613 -78.075 46.036
154752 98.614 -78.083 46.015
154753 98.615 -78.090 45.997
154754 98.616 -78.097 45.979
154755 98.616 -78.104 45.961
154756 98.617 -78.111 45.942
154757 98.627 -78.119 45.923
154758 98.680 -78.126 45.904
154759 98.733 -78.134 45.885
154760 98.785 -78.142 45.868
154761 98.838 -78.150 45.852
154762 98.890 -78.158 45.836
154763 98.943 -78.166 45.819
154764 98.996 -78.174 45.802
154765 99.049 -78.181 45.787
154766 99.101 -78.190 45.773
154767 99.150 -78.197 45.759
154768 99.185 -78.203 45.744
154769 99.183 -78.206 45.729
154770 99.177 -78.206 45.715
154771 99.169 -78.206 45.702
154772 99.170 -78.207 45.689
154773 99.193 -78.210 45.678
154774 99.215 -78.213 45.667
154775 99.237 -78.216 45.655
154776 99.260 -78.218 45.643
154777 99.282 -78.221 45.633
154778 99.304 -78.224 45.621
154779 99.327 -78.226 45.607
154780 99.349 -78.229 45.595
154781 99.371 -78.232 45.584
154782 99.394 -78.235 45.574
154783 99.422 -78.241 45.562
154784 99.454 -78.248 45.550
154785 99.478 -78.260 45.537
154786 99.473 -78.285 45.526
154787 99.502 -78.304 45.514
154788 99.531 -78.324 45.503
154789 99.560 -78.343 45.492
154790 99.588 -78.362 45.481
154791 99.617 -78.382 45.470
154792 99.646 -78.401 45.459
154793 99.675 -78.420 45.448
154794 99.691 -78.435 45.437
154795 99.759 -78.445 45.426
154796 99.791 -78.473 45.413
154797 99.823 -78.502 45.400
154798 99.857 -78.515 45.387
154799 99.886 -78.521 45.374
154800 99.913 -78.531 45.361
154801 99.940 -78.541 45.347
154802 99.967 -78.550 45.334
154803 99.995 -78.560 45.320
154804 100.028 -78.575 45.305
154805 100.061 -78.589 45.292
154806 100.094 -78.604 45.277
154807 100.128 -78.619 45.262
154808 100.161 -78.634 45.247
154809 100.195 -78.648 45.232
154810 100.228 -78.664 45.217
154811 100.254 -78.673 45.202
154812 100.265 -78.673 45.186
154813 100.284 -78.677 45.170
154814 100.302 -78.680 45.154
154815 100.321 -78.685 45.136
154816 100.340 -78.696 45.118
154817 100.359 -78.708 45.100
154818 100.378 -78.720 45.083
154819 100.397 -78.731 45.066
154820 100.416 -78.742 45.049
154821 100.435 -78.754 45.032
154822 100.454 -78.765 45.016
154823 100.473 -78.775 44.998
154824 100.494 -78.783 44.981
154825 100.514 -78.789 44.964
154826 100.536 -78.797 44.948
154827 100.556 -78.804 44.932
154828 100.577 -78.811 44.915
154829 100.599 -78.818 44.900
154830 100.605 -78.844 44.883
154831 100.611 -78.870 44.866
154832 100.617 -78.896 44.849
154833 100.623 -78.922 44.833
154834 100.629 -78.948 44.817
154835 100.650 -78.964 44.801
154836 100.678 -78.975 44.784
154837 100.706 -78.986 44.767
154838 100.732 -78.985 44.750
154839 100.755 -78.982 44.733
154840 100.785 -79.040 44.717
154841 100.829 -79.046 44.701
154842 100.873 -79.052 44.685
154843 100.953 -79.021 44.669
154844 101.027 -79.001 44.654
154845 101.079 -79.007 44.638
154846 101.131 -79.013 44.622
154847 101.183 -79.019 44.607
154848 101.235 -79.026 44.592
154849 101.265 -79.038 44.578
154850 101.289 -79.051 44.564
154851 101.314 -79.065 44.551
154852 101.338 -79.078 44.538
154853 101.362 -79.092 44.525
154854 101.387 -79.105 44.513
154855 101.407 -79.130 44.500
154856 101.424 -79.164 44.487
155083 92.531 -77.632 54.177
155084 92.551 -77.639 54.179
155085 92.571 -77.646 54.181
155086 92.591 -77.654 54.183
155087 92.611 -77.661 54.185
155088 92.631 -77.668 54.186
155089 92.651 -77.676 54.187
155090 92.671 -77.683 54.189
155091 92.691 -77.690 54.191
155092 92.711 -77.698 54.193
155093 92.731 -77.705 54.193
155094 92.751 -77.712 54.195
155095 92.771 -77.720 54.196
155096 92.791 -77.727 54.197
155097 92.810 -77.734 54.199
155098 92.831 -77.742 54.201
155099 92.851 -77.749 54.203
155100 92.870 -77.757 54.206
155101 92.895 -77.754 54.208
155102 92.920 -77.743 54.211
155103 92.937 -77.738 54.214
155104 92.954 -77.733 54.217
155105 92.971 -77.727 54.220
155106 92.987 -77.721 54.225
155107 93.000 -77.711 54.231
155108 93.014 -77.703 54.239
155109 93.027 -77.694 54.246
155110 93.041 -77.685 54.250
155111 93.054 -77.677 54.254
155112 93.068 -77.668 54.255
155113 93.080 -77.661 54.256
155114 93.092 -77.655 54.255
155115 93.105 -77.649 54.254
155116 93.117 -77.642 54.251
155117 93.130 -77.636 54.249
155118 93.142 -77.630 54.248
155119 93.155 -77.623 54.248
155120 93.167 -77.617 54.248
155121 93.179 -77.605 54.247
155122 93.190 -77.590 54.247
155123 93.202 -77.576 54.246
155124 93.213 -77.561 54.242
155125 93.224 -77.547 54.237
155126 93.236 -77.532 54.232
155127 93.247 -77.517 54.226
155128 93.263 -77.505 54.219
155129 93.287 -77.497 54.214
155130 93.310 -77.489 54.206
155131 93.324 -77.479 54.199
155132 93.338 -77.470 54.192
155133 93.351 -77.461 54.187
155134 93.364 -77.452 54.181
155135 93.378 -77.443 54.175
155136 93.391 -77.434 54.167
155137 93.407 -77.426 54.157
155138 93.423 -77.417 54.148
155139 93.439 -77.408 54.141
155140 93.454 -77.400 54.133
155141 93.469 -77.394 54.125
155142 93.484 -77.389 54.116
155143 93.501 -77.384 54.105
155144 93.519 -77.379 54.090
155145 93.537 -77.373 54.075
155146 93.556 -77.367 54.061
155147 93.574 -77.360 54.047
155148 93.595 -77.352 54.031
155149 93.616 -77.343 54.018
155150 93.632 -77.335 54.005
155151 93.648 -77.326 53.989
155152 93.663 -77.316 53.973
155153 93.679 -77.307 53.957
155154 93.694 -77.299 53.943
155155 93.709 -77.290 53.926
155156 93.730 -77.279 53.910
155157 93.752 -77.268 53.895
155158 93.774 -77.257 53.878
155159 93.796 -77.246 53.859
155160 93.817 -77.233 53.838
155161 93.823 -77.196 53.817
155162 93.834 -77.178 53.797
155163 93.841 -77.191 53.777
155164 93.847 -77.215 53.758
155165 93.854 -77.226 53.738
155166 93.860 -77.238 53.721
155167 93.867 -77.248 53.703
155168 93.873 -77.258 53.686
155169 93.883 -77.252 53.669
155170 93.892 -77.247 53.652
155171 93.904 -77.247 53.635
155172 93.918 -77.241 53.617
155173 93.931 -77.229 53.599
155174 93.947 -77.219 53.581
155175 93.964 -77.209 53.564
155176 93.979 -77.199 53.550
155177 93.997 -77.191 53.534
155178 94.021 -77.188 53.519
155179 94.046 -77.186 53.503
155180 94.077 -77.186 53.488
155181 94.108 -77.185 53.472
155182 94.139 -77.185 53.458
155183 94.170 -77.185 53.445
155184 94.201 -77.185 53.435
155185 94.232 -77.185 53.428
155186 94.263 -77.185 53.423
155187 94.294 -77.184 53.420
155188 94.315 -77.179 53.418
155189 94.315 -77.167 53.417
155190 94.313 -77.156 53.415
155191 94.311 -77.145 53.413
155192 94.309 -77.134 53.408
155193 94.306 -77.123 53.403
155194 94.309 -77.119 53.399
155195 94.317 -77.119 53.397
155196 94.324 -77.119 53.396
155197 94.327 -77.117 53.396
155198 94.328 -77.113 53.396
155199 94.328 -77.110 53.393
155200 94.329 -77.106 53.389
155201 94.329 -77.101 53.384
155202 94.330 -77.097 53.377
155203 94.330 -77.095 53.368
155204 94.331 -77.093 53.357
155205 94.348 -77.085 53.341
155206 94.368 -77.081 53.319
155207 94.389 -77.076 53.296
155208 94.409 -77.072 53.272
155209 94.429 -77.067 53.251
155210 94.451 -77.063 53.231
155211 94.472 -77.059 53.212
155212 94.493 -77.055 53.192
155213 94.514 -77.050 53.168
155214 94.535 -77.047 53.145
155215 94.556 -77.042 53.123
155216 94.577 -77.038 53.102
155217 94.599 -77.034 53.084
155218 94.623 -77.030 53.064
155219 94.631 -77.025 53.041
155220 94.643 -77.025 53.017
155221 94.655 -77.026 52.993
155222 94.668 -77.028 52.967
155223 94.680 -77.031 52.939
155224 94.693 -77.033 52.908
155225 94.705 -77.035 52.874
155226 94.718 -77.038 52.837
155227 94.731 -77.040 52.797
155228 94.765 -77.046 52.754
155229 94.805 -77.053 52.707
155230 94.833 -77.057 52.657
155231 94.841 -77.060 52.605
155232 94.849 -77.063 52.554
155233 94.864 -77.069 52.505
155234 94.883 -77.075 52.458
155235 94.902 -77.081 52.411
155236 94.921 -77.088 52.367
155237 94.939 -77.094 52.324
155238 94.956 -77.088 52.283
155239 94.972 -77.083 52.246
155240 94.990 -77.081 52.208
155241 95.011 -77.084 52.169
155242 95.032 -77.087 52.135
155243 95.052 -77.090 52.101
155244 95.073 -77.093 52.066
155245 95.094 -77.096 52.027
155246 95.115 -77.100 51.986
155247 95.135 -77.103 51.945
155248 95.155 -77.108 51.903
155249 95.175 -77.113 51.859
155250 95.194 -77.116 51.811
155251 95.205 -77.114 51.761
155252 95.216 -77.112 51.708
155253 95.227 -77.110 51.652
155254 95.237 -77.107 51.599
155255 95.250 -77.108 51.547
155256 95.264 -77.110 51.497
155257 95.279 -77.106 51.446
155258 95.295 -77.099 51.395
155259 95.314 -77.103 51.345
155260 95.336 -77.119 51.298
155261 95.358 -77.133 51.254
155262 95.381 -77.148 51.209
155263 95.403 -77.163 51.162
155264 95.425 -77.178 51.112
155265 95.447 -77.192 51.061
155266 95.466 -77.207 51.012
155267 95.484 -77.222 50.965
155268 95.502 -77.236 50.921
155269 95.520 -77.251 50.879
155270 95.538 -77.266 50.839
155271 95.555 -77.281 50.797
155272 95.573 -77.295 50.754
155273 95.591 -77.310 50.710
155274 95.609 -77.325 50.665
155275 95.626 -77.339 50.620
155276 95.644 -77.354 50.576
155277 95.662 -77.368 50.534
155278 95.679 -77.380 50.492
155279 95.695 -77.393 50.451
155280 95.712 -77.405 50.412
155281 95.729 -77.417 50.376
155282 95.746 -77.430 50.343
155283 95.763 -77.442 50.309
155284 95.784 -77.454 50.273
155285 95.806 -77.466 50.234
155286 95.827 -77.478 50.196
155287 95.848 -77.490 50.156
155288 95.870 -77.501 50.114
155289 95.891 -77.514 50.068
155290 95.913 -77.525 50.020
155291 95.934 -77.537 49.971
155292 95.950 -77.537 49.922
155293 95.961 -77.532 49.868
155294 95.973 -77.526 49.811
155295 95.984 -77.520 49.757
155296 95.995 -77.515 49.706
155297 96.006 -77.513 49.658
155298 96.014 -77.520 49.612
155299 96.022 -77.529 49.567
155300 96.029 -77.536 49.522
155301 96.037 -77.545 49.478
155302 96.045 -77.549 49.435
155303 96.054 -77.548 49.394
155304 96.064 -77.548 49.355
155305 96.073 -77.548 49.320
155306 96.084 -77.544 49.285
155307 96.097 -77.520 49.252
155308 96.105 -77.504 49.220
155309 96.112 -77.489 49.191
155310 96.120 -77.473 49.163
155311 96.128 -77.461 49.136
155312 96.136 -77.457 49.110
155313 96.139 -77.459 49.084
155314 96.135 -77.470 49.059
155315 96.132 -77.494 49.034
155316 96.128 -77.519 49.013
155317 96.130 -77.536 48.992
155318 96.145 -77.539 48.969
155319 96.159 -77.551 48.948
155320 96.173 -77.562 48.926
155321 96.187 -77.573 48.904
155322 96.200 -77.585 48.880
155323 96.215 -77.596 48.858
155324 96.228 -77.608 48.834
155325 96.242 -77.619 48.810
155326 96.256 -77.630 48.786
155327 96.269 -77.639 48.761
155328 96.292 -77.649 48.737
155329 96.317 -77.660 48.713
155330 96.342 -77.671 48.689
155331 96.368 -77.680 48.666
155332 96.393 -77.691 48.643
155333 96.413 -77.697 48.621
155334 96.434 -77.702 48.598
155335 96.454 -77.706 48.575
155336 96.474 -77.711 48.554
155337 96.495 -77.716 48.534
155338 96.512 -77.721 48.514
155339 96.524 -77.724 48.494
155340 96.535 -77.728 48.476
155341 96.546 -77.731 48.460
155342 96.557 -77.735 48.443
155343 96.566 -77.736 48.424
155344 96.573 -77.738 48.403
155345 96.579 -77.740 48.379
155346 96.586 -77.741 48.355
155347 96.593 -77.743 48.330
155348 96.600 -77.744 48.307
155349 96.607 -77.745 48.284
155350 96.615 -77.749 48.261
155351 96.625 -77.756 48.240
155352 96.635 -77.763 48.218
155353 96.644 -77.769 48.195
155354 96.654 -77.775 48.171
155355 96.668 -77.778 48.150
155356 96.682 -77.780 48.124
155357 96.697 -77.783 48.101
155358 96.711 -77.785 48.079
155359 96.725 -77.787 48.057
155360 96.739 -77.790 48.032
155361 96.754 -77.783 48.007
155362 96.769 -77.771 47.980
155363 96.784 -77.758 47.954
155364 96.799 -77.746 47.928
155365 96.818 -77.748 47.902
155366 96.840 -77.758 47.878
155367 96.863 -77.768 47.853
155368 96.885 -77.779 47.829
155369 96.907 -77.790 47.804
155370 96.929 -77.799 47.780
155371 96.952 -77.804 47.758
155372 96.975 -77.805 47.735
155373 96.994 -77.805 47.713
155374 97.007 -77.805 47.691
155375 97.019 -77.805 47.669
155376 97.032 -77.804 47.648
155377 97.045 -77.803 47.627
155378 97.058 -77.802 47.606
155379 97.070 -77.802 47.584
155380 97.083 -77.802 47.563
155381 97.095 -77.799 47.543
155382 97.113 -77.800 47.522
155383 97.134 -77.805 47.502
155384 97.155 -77.810 47.484
155385 97.176 -77.815 47.463
155386 97.196 -77.821 47.442
155387 97.220 -77.821 47.423
155388 97.243 -77.820 47.404
155389 97.267 -77.819 47.385
155390 97.290 -77.818 47.368
155391 97.314 -77.818 47.351
155392 97.337 -77.817 47.333
155393 97.360 -77.815 47.316
155394 97.383 -77.815 47.300
155395 97.406 -77.813 47.283
155396 97.429 -77.812 47.266
155397 97.451 -77.812 47.248
155398 97.465 -77.827 47.231
155399 97.480 -77.840 47.211
155400 97.496 -77.855 47.191
155401 97.512 -77.871 47.170
155402 97.527 -77.887 47.150
155403 97.542 -77.902 47.128
155404 97.558 -77.917 47.107
155405 97.574 -77.932 47.085
155406 97.589 -77.947 47.063
155407 97.605 -77.962 47.042
155408 97.620 -77.978 47.019
155409 97.623 -78.002 46.995
155410 97.619 -78.030 46.972
155411 97.617 -78.056 46.948
155412 97.618 -78.073 46.924
155413 97.619 -78.091 46.899
155414 97.620 -78.108 46.876
155415 97.640 -78.132 46.852
155416 97.660 -78.149 46.829
155417 97.693 -78.123 46.806
155418 97.728 -78.112 46.785
155419 97.763 -78.100 46.766
155420 97.798 -78.089 46.746
155421 97.833 -78.078 46.727
155422 97.868 -78.065 46.710
155423 97.868 -78.090 46.693
155424 97.892 -78.109 46.673
155425 97.917 -78.128 46.651
155426 97.945 -78.145 46.630
155427 98.008 -78.134 46.608
155428 98.071 -78.123 46.586
155429 98.105 -78.112 46.563
155430 98.131 -78.110 46.540
155431 98.160 -78.112 46.517
155432 98.197 -78.100 46.492
155433 98.233 -78.084 46.469
155434 98.270 -78.070 46.447
155435 98.306 -78.056 46.425
155436 98.343 -78.041 46.402
155437 98.380 -78.027 46.377
155438 98.417 -78.012 46.353
155439 98.453 -77.997 46.330
155440 98.490 -77.983 46.307
155441 98.527 -77.968 46.284
155442 98.563 -77.954 46.261
155443 98.600 -77.940 46.239
155444 98.627 -77.946 46.217
155445 98.639 -77.983 46.195
155446 98.652 -78.018 46.174
155447 98.665 -78.054 46.152
155448 98.660 -78.063 46.130
155449 98.648 -78.059 46.109
155450 98.635 -78.055 46.088
155451 98.630 -78.057 46.067
155452 98.631 -78.064 46.047
155453 98.631 -78.072 46.027
155454 98.632 -78.078 46.009
155455 98.633 -78.086 45.991
155456 98.634 -78.093 45.972
155457 98.635 -78.100 45.953
155458 98.635 -78.107 45.935
155459 98.668 -78.115 45.915
155460 98.721 -78.122 45.896
155461 98.774 -78.131 45.878
155462 98.827 -78.138 45.861
155463 98.879 -78.147 45.845
155464 98.932 -78.154 45.828
155465 98.985 -78.162 45.812
155466 99.038 -78.170 45.796
155467 99.090 -78.178 45.781
155468 99.143 -78.186 45.766
155469 99.193 -78.194 45.752
155470 99.197 -78.194 45.738
155471 99.191 -78.194 45.724
155472 99.184 -78.194 45.711
155473 99.178 -78.195 45.698
155474 99.200 -78.197 45.687
155475 99.223 -78.200 45.676
155476 99.245 -78.203 45.663
155477 99.267 -78.206 45.652
155478 99.290 -78.209 45.641
155479 99.312 -78.212 45.630
155480 99.334 -78.214 45.617
155481 99.356 -78.217 45.605
155482 99.379 -78.220 45.594
155483 99.401 -78.222 45.583
155484 99.423 -78.225 45.571
155485 99.449 -78.230 45.559
155486 99.468 -78.248 45.547
155487 99.497 -78.267 45.535
155488 99.525 -78.286 45.523
155489 99.554 -78.306 45.513
155490 99.583 -78.325 45.503
155491 99.612 -78.344 45.492
155492 99.641 -78.363 45.481
155493 99.669 -78.382 45.470
155494 99.698 -78.401 45.458
155495 99.727 -78.421 45.447
155496 99.745 -78.435 45.436
155497 99.777 -78.463 45.423
155498 99.808 -78.492 45.409
155499 99.841 -78.516 45.396
155500 99.872 -78.514 45.383
155501 99.900 -78.524 45.371
155502 99.927 -78.534 45.357
155503 99.961 -78.548 45.343
155504 99.994 -78.564 45.329
155505 100.028 -78.578 45.315
155506 100.061 -78.593 45.301
155507 100.094 -78.608 45.286
155508 100.128 -78.623 45.270
155509 100.161 -78.637 45.256
155510 100.194 -78.652 45.240
155511 100.227 -78.667 45.224
155512 100.261 -78.682 45.209
155513 100.272 -78.684 45.193
155514 100.290 -78.687 45.177
155515 100.309 -78.690 45.160
155516 100.327 -78.693 45.143
155517 100.346 -78.696 45.126
155518 100.364 -78.700 45.108
155519 100.383 -78.711 45.091
155520 100.402 -78.723 45.073
155521 100.421 -78.735 45.056
155522 100.440 -78.746 45.039
155523 100.459 -78.758 45.023
155524 100.479 -78.767 45.006
155525 100.499 -78.774 44.989
155526 100.521 -78.781 44.973
155527 100.541 -78.788 44.956
155528 100.562 -78.795 44.940
155529 100.583 -78.802 44.924
155530 100.605 -78.808 44.909
155531 100.616 -78.828 44.892
155532 100.622 -78.855 44.875
155533 100.628 -78.880 44.858
155534 100.634 -78.906 44.841
155535 100.643 -78.930 44.826
155536 100.671 -78.942 44.810
155537 100.700 -78.952 44.793
155538 100.728 -78.964 44.776
155539 100.756 -78.973 44.759
155540 100.777 -78.986 44.742
155541 100.817 -79.006 44.726
155542 100.860 -79.012 44.710
155543 100.904 -79.018 44.695
155544 100.948 -79.024 44.679
155545 101.006 -79.015 44.664
155546 101.085 -78.990 44.648
155547 101.136 -78.996 44.631
155548 101.188 -79.002 44.615
155549 101.240 -79.009 44.601
155550 101.264 -79.023 44.587
155551 101.288 -79.037 44.573
155552 101.313 -79.049 44.560
155553 101.337 -79.063 44.547
155554 101.361 -79.077 44.534
155555 101.386 -79.090 44.521
155556 101.406 -79.117 44.509
155557 101.422 -79.151 44.496
155558 101.450 -79.179 44.484
155784 92.520 -77.611 54.166
155785 92.540 -77.619 54.170
155786 92.560 -77.627 54.173
155787 92.580 -77.633 54.177
155788 92.600 -77.641 54.179
155789 92.620 -77.649 54.181
155790 92.640 -77.656 54.184
155791 92.660 -77.663 54.186
155792 92.680 -77.671 54.189
155793 92.700 -77.678 54.192
155794 92.720 -77.685 54.194
155795 92.740 -77.692 54.196
155796 92.760 -77.700 54.198
155797 92.780 -77.707 54.199
155798 92.800 -77.714 54.201
155799 92.819 -77.722 54.204
155800 92.840 -77.729 54.207
155801 92.859 -77.736 54.210
155802 92.884 -77.734 54.213
155803 92.911 -77.722 54.216
155804 92.927 -77.717 54.220
155805 92.944 -77.711 54.223
155806 92.960 -77.705 54.227
155807 92.974 -77.697 54.232
155808 92.987 -77.688 54.238
155809 93.001 -77.679 54.245
155810 93.014 -77.670 54.251
155811 93.028 -77.661 54.256
155812 93.042 -77.652 54.261
155813 93.055 -77.643 54.264
155814 93.068 -77.636 54.266
155815 93.080 -77.630 54.266
155816 93.092 -77.624 54.265
155817 93.105 -77.617 54.262
155818 93.117 -77.611 54.260
155819 93.129 -77.608 54.259
155820 93.141 -77.605 54.260
155821 93.152 -77.602 54.260
155822 93.163 -77.587 54.260
155823 93.174 -77.572 54.260
155824 93.186 -77.558 54.260
155825 93.197 -77.543 54.258
155826 93.209 -77.528 54.253
155827 93.220 -77.514 54.248
155828 93.231 -77.499 54.243
155829 93.255 -77.491 54.238
155830 93.279 -77.483 54.233
155831 93.303 -77.474 54.226
155832 93.320 -77.465 54.219
155833 93.334 -77.456 54.212
155834 93.347 -77.447 54.208
155835 93.360 -77.438 54.203
155836 93.374 -77.429 54.198
155837 93.387 -77.420 54.191
155838 93.401 -77.410 54.182
155839 93.414 -77.401 54.173
155840 93.427 -77.392 54.166
155841 93.443 -77.384 54.160
155842 93.459 -77.376 54.152
155843 93.475 -77.371 54.142
155844 93.492 -77.365 54.131
155845 93.511 -77.359 54.117
155846 93.530 -77.353 54.103
155847 93.548 -77.348 54.089
155848 93.567 -77.341 54.075
155849 93.588 -77.332 54.060
155850 93.609 -77.323 54.047
155851 93.629 -77.315 54.035
155852 93.650 -77.306 54.019
155853 93.667 -77.298 54.003
155854 93.683 -77.288 53.987
155855 93.698 -77.279 53.972
155856 93.714 -77.270 53.956
155857 93.731 -77.261 53.941
155858 93.753 -77.250 53.926
155859 93.775 -77.239 53.908
155860 93.796 -77.228 53.889
155861 93.806 -77.197 53.869
155862 93.818 -77.183 53.849
155863 93.830 -77.170 53.827
155864 93.841 -77.150 53.806
155865 93.847 -77.163 53.786
155866 93.854 -77.174 53.764
155867 93.860 -77.185 53.745
155868 93.867 -77.197 53.727
155869 93.873 -77.207 53.709
155870 93.880 -77.219 53.690
155871 93.886 -77.230 53.673
155872 93.896 -77.232 53.656
155873 93.912 -77.222 53.638
155874 93.928 -77.212 53.618
155875 93.944 -77.201 53.600
155876 93.960 -77.191 53.583
155877 93.976 -77.182 53.568
155878 93.994 -77.173 53.552
155879 94.018 -77.171 53.536
155880 94.042 -77.169 53.520
155881 94.067 -77.166 53.504
155882 94.098 -77.166 53.488
155883 94.129 -77.166 53.474
155884 94.160 -77.166 53.461
155885 94.191 -77.165 53.450
155886 94.222 -77.164 53.441
155887 94.254 -77.163 53.434
155888 94.285 -77.163 53.431
155889 94.297 -77.153 53.430
155890 94.302 -77.141 53.428
155891 94.307 -77.129 53.426
155892 94.306 -77.118 53.424
155893 94.303 -77.107 53.418
155894 94.305 -77.100 53.410
155895 94.312 -77.101 53.403
155896 94.319 -77.101 53.398
155897 94.327 -77.102 53.395
155898 94.331 -77.100 53.395
155899 94.331 -77.097 53.398
155900 94.332 -77.092 53.399
155901 94.332 -77.088 53.398
155902 94.333 -77.085 53.396
155903 94.333 -77.081 53.393
155904 94.334 -77.077 53.385
155905 94.335 -77.073 53.374
155906 94.346 -77.069 53.360
155907 94.367 -77.065 53.338
155908 94.389 -77.061 53.313
155909 94.410 -77.056 53.288
155910 94.431 -77.053 53.263
155911 94.452 -77.049 53.240
155912 94.473 -77.044 53.219
155913 94.494 -77.041 53.198
155914 94.516 -77.036 53.175
155915 94.537 -77.032 53.151
155916 94.558 -77.028 53.129
155917 94.579 -77.024 53.110
155918 94.605 -77.020 53.090
155919 94.624 -77.016 53.070
155920 94.632 -77.010 53.049
155921 94.643 -77.009 53.024
155922 94.654 -77.009 53.000
155923 94.667 -77.009 52.975
155924 94.679 -77.012 52.947
155925 94.692 -77.014 52.916
155926 94.704 -77.016 52.881
155927 94.717 -77.019 52.842
155928 94.730 -77.021 52.801
155929 94.766 -77.027 52.758
155930 94.806 -77.034 52.712
155931 94.846 -77.041 52.663
155932 94.856 -77.044 52.611
155933 94.864 -77.046 52.558
155934 94.873 -77.049 52.508
155935 94.887 -77.054 52.462
155936 94.906 -77.061 52.418
155937 94.925 -77.068 52.376
155938 94.944 -77.075 52.337
155939 94.960 -77.069 52.298
155940 94.976 -77.064 52.264
155941 94.996 -77.065 52.230
155942 95.017 -77.068 52.198
155943 95.038 -77.071 52.171
155944 95.058 -77.074 52.140
155945 95.079 -77.077 52.106
155946 95.100 -77.081 52.067
155947 95.120 -77.085 52.023
155948 95.140 -77.090 51.980
155949 95.160 -77.094 51.936
155950 95.180 -77.099 51.890
155951 95.200 -77.103 51.839
155952 95.213 -77.103 51.788
155953 95.224 -77.100 51.736
155954 95.235 -77.098 51.679
155955 95.245 -77.096 51.625
155956 95.256 -77.094 51.572
155957 95.267 -77.091 51.520
155958 95.279 -77.091 51.470
155959 95.295 -77.088 51.420
155960 95.317 -77.103 51.371
155961 95.339 -77.118 51.322
155962 95.361 -77.132 51.277
155963 95.384 -77.147 51.231
155964 95.406 -77.162 51.185
155965 95.428 -77.177 51.135
155966 95.450 -77.191 51.083
155967 95.470 -77.207 51.032
155968 95.487 -77.221 50.985
155969 95.506 -77.235 50.941
155970 95.523 -77.251 50.898
155971 95.541 -77.265 50.855
155972 95.558 -77.280 50.810
155973 95.577 -77.294 50.766
155974 95.594 -77.309 50.721
155975 95.612 -77.324 50.676
155976 95.627 -77.335 50.632
155977 95.642 -77.345 50.590
155978 95.658 -77.356 50.548
155979 95.676 -77.367 50.505
155980 95.693 -77.378 50.463
155981 95.711 -77.389 50.424
155982 95.728 -77.401 50.387
155983 95.745 -77.414 50.352
155984 95.764 -77.426 50.316
155985 95.785 -77.438 50.279
155986 95.807 -77.450 50.240
155987 95.828 -77.461 50.201
155988 95.849 -77.473 50.161
155989 95.870 -77.483 50.119
155990 95.890 -77.494 50.076
155991 95.911 -77.504 50.031
155992 95.931 -77.515 49.985
155993 95.952 -77.523 49.935
155994 95.964 -77.522 49.880
155995 95.976 -77.516 49.823
155996 95.987 -77.511 49.769
155997 95.999 -77.505 49.718
155998 96.010 -77.501 49.670
155999 96.017 -77.509 49.622
156000 96.025 -77.517 49.575
156001 96.032 -77.525 49.530
156002 96.040 -77.533 49.484
156003 96.048 -77.541 49.443
156004 96.056 -77.546 49.403
156005 96.065 -77.546 49.364
156006 96.075 -77.546 49.327
156007 96.085 -77.541 49.292
156008 96.097 -77.520 49.257
156009 96.104 -77.504 49.225
156010 96.112 -77.489 49.196
156011 96.120 -77.480 49.167
156012 96.128 -77.476 49.139
156013 96.135 -77.471 49.112
156014 96.144 -77.467 49.086
156015 96.143 -77.475 49.060
156016 96.139 -77.486 49.034
156017 96.136 -77.497 49.013
156018 96.132 -77.518 48.992
156019 96.147 -77.526 48.971
156020 96.160 -77.537 48.952
156021 96.174 -77.548 48.933
156022 96.188 -77.560 48.913
156023 96.202 -77.571 48.890
156024 96.216 -77.583 48.868
156025 96.230 -77.594 48.845
156026 96.244 -77.606 48.822
156027 96.258 -77.618 48.798
156028 96.276 -77.628 48.774
156029 96.302 -77.639 48.749
156030 96.327 -77.649 48.725
156031 96.352 -77.659 48.701
156032 96.377 -77.670 48.679
156033 96.402 -77.680 48.656
156034 96.428 -77.690 48.634
156035 96.448 -77.696 48.611
156036 96.468 -77.700 48.588
156037 96.488 -77.705 48.567
156038 96.509 -77.710 48.548
156039 96.526 -77.714 48.529
156040 96.537 -77.718 48.510
156041 96.549 -77.721 48.492
156042 96.560 -77.725 48.476
156043 96.569 -77.727 48.458
156044 96.576 -77.729 48.438
156045 96.583 -77.730 48.416
156046 96.589 -77.732 48.393
156047 96.596 -77.733 48.370
156048 96.603 -77.734 48.347
156049 96.610 -77.736 48.324
156050 96.617 -77.737 48.302
156051 96.624 -77.740 48.282
156052 96.633 -77.746 48.262
156053 96.643 -77.752 48.241
156054 96.653 -77.759 48.218
156055 96.663 -77.765 48.194
156056 96.677 -77.768 48.170
156057 96.691 -77.771 48.145
156058 96.705 -77.773 48.122
156059 96.720 -77.775 48.100
156060 96.734 -77.777 48.078
156061 96.748 -77.780 48.053
156062 96.762 -77.783 48.028
156063 96.777 -77.779 48.001
156064 96.792 -77.767 47.975
156065 96.806 -77.755 47.948
156066 96.822 -77.744 47.923
156067 96.844 -77.755 47.898
156068 96.866 -77.765 47.873
156069 96.888 -77.775 47.848
156070 96.910 -77.786 47.824
156071 96.933 -77.795 47.801
156072 96.956 -77.796 47.778
156073 96.979 -77.796 47.756
156074 97.000 -77.797 47.734
156075 97.012 -77.796 47.712
156076 97.025 -77.796 47.690
156077 97.037 -77.796 47.669
156078 97.050 -77.795 47.648
156079 97.063 -77.794 47.627
156080 97.075 -77.795 47.606
156081 97.086 -77.799 47.584
156082 97.099 -77.790 47.563
156083 97.113 -77.782 47.543
156084 97.131 -77.782 47.522
156085 97.152 -77.787 47.503
156086 97.172 -77.793 47.482
156087 97.193 -77.798 47.462
156088 97.217 -77.797 47.442
156089 97.240 -77.796 47.423
156090 97.264 -77.796 47.405
156091 97.288 -77.795 47.387
156092 97.311 -77.794 47.370
156093 97.334 -77.793 47.352
156094 97.358 -77.793 47.335
156095 97.382 -77.792 47.318
156096 97.405 -77.791 47.301
156097 97.435 -77.799 47.284
156098 97.465 -77.808 47.267
156099 97.478 -77.813 47.248
156100 97.494 -77.828 47.229
156101 97.509 -77.843 47.208
156102 97.525 -77.859 47.187
156103 97.540 -77.874 47.166
156104 97.556 -77.889 47.144
156105 97.572 -77.904 47.122
156106 97.587 -77.920 47.100
156107 97.603 -77.935 47.077
156108 97.618 -77.950 47.056
156109 97.634 -77.965 47.032
156110 97.631 -77.993 47.007
156111 97.628 -78.021 46.983
156112 97.625 -78.048 46.958
156113 97.623 -78.075 46.933
156114 97.624 -78.093 46.909
156115 97.624 -78.110 46.885
156116 97.642 -78.134 46.861
156117 97.662 -78.126 46.837
156118 97.702 -78.103 46.815
156119 97.737 -78.091 46.793
156120 97.772 -78.080 46.773
156121 97.806 -78.069 46.753
156122 97.842 -78.056 46.733
156123 97.876 -78.045 46.716
156124 97.908 -78.028 46.697
156125 97.940 -78.039 46.677
156126 97.972 -78.055 46.658
156127 98.005 -78.072 46.637
156128 98.036 -78.089 46.616
156129 98.061 -78.108 46.593
156130 98.102 -78.100 46.570
156131 98.139 -78.085 46.546
156132 98.176 -78.071 46.523
156133 98.213 -78.056 46.500
156134 98.249 -78.042 46.479
156135 98.286 -78.028 46.457
156136 98.323 -78.012 46.434
156137 98.359 -77.998 46.410
156138 98.396 -77.984 46.386
156139 98.433 -77.969 46.362
156140 98.469 -77.955 46.339
156141 98.506 -77.940 46.316
156142 98.543 -77.925 46.294
156143 98.580 -77.911 46.271
156144 98.617 -77.896 46.249
156145 98.634 -77.923 46.226
156146 98.646 -77.959 46.204
156147 98.659 -77.995 46.182
156148 98.672 -78.031 46.161
156149 98.673 -78.048 46.139
156150 98.660 -78.044 46.118
156151 98.647 -78.039 46.098
156152 98.647 -78.046 46.078
156153 98.648 -78.053 46.058
156154 98.649 -78.060 46.039
156155 98.650 -78.067 46.020
156156 98.650 -78.075 46.001
156157 98.651 -78.081 45.982
156158 98.652 -78.089 45.963
156159 98.653 -78.096 45.945
156160 98.657 -78.103 45.925
156161 98.710 -78.111 45.906
156162 98.763 -78.119 45.888
156163 98.815 -78.127 45.871
156164 98.868 -78.134 45.854
156165 98.921 -78.143 45.837
156166 98.973 -78.150 45.821
156167 99.026 -78.158 45.805
156168 99.079 -78.166 45.789
156169 99.131 -78.174 45.774
156170 99.156 -78.178 45.761
156171 99.161 -78.179 45.746
156172 99.167 -78.180 45.733
156173 99.172 -78.181 45.720
156174 99.183 -78.182 45.707
156175 99.207 -78.181 45.695
156176 99.229 -78.183 45.684
156177 99.252 -78.185 45.672
156178 99.274 -78.188 45.661
156179 99.296 -78.190 45.650
156180 99.318 -78.192 45.638
156181 99.340 -78.194 45.626
156182 99.362 -78.197 45.614
156183 99.384 -78.199 45.602
156184 99.407 -78.201 45.591
156185 99.429 -78.203 45.579
156186 99.452 -78.209 45.567
156187 99.482 -78.228 45.556
156188 99.511 -78.247 45.545
156189 99.540 -78.266 45.534
156190 99.569 -78.286 45.522
156191 99.597 -78.305 45.511
156192 99.626 -78.324 45.501
156193 99.655 -78.344 45.490
156194 99.684 -78.363 45.479
156195 99.713 -78.382 45.468
156196 99.739 -78.408 45.457
156197 99.769 -78.435 45.444
156198 99.792 -78.477 45.432
156199 99.811 -78.496 45.418
156200 99.832 -78.508 45.405
156201 99.860 -78.507 45.393
156202 99.894 -78.523 45.380
156203 99.927 -78.537 45.366
156204 99.960 -78.552 45.351
156205 99.994 -78.567 45.337
156206 100.027 -78.582 45.323
156207 100.060 -78.596 45.309
156208 100.093 -78.611 45.294
156209 100.127 -78.626 45.278
156210 100.160 -78.641 45.264
156211 100.194 -78.655 45.248
156212 100.227 -78.670 45.232
156213 100.261 -78.685 45.216
156214 100.294 -78.700 45.200
156215 100.296 -78.697 45.183
156216 100.315 -78.700 45.167
156217 100.333 -78.703 45.150
156218 100.352 -78.707 45.133
156219 100.370 -78.710 45.115
156220 100.388 -78.713 45.097
156221 100.407 -78.716 45.081
156222 100.426 -78.727 45.064
156223 100.445 -78.738 45.047
156224 100.464 -78.750 45.029
156225 100.485 -78.758 45.013
156226 100.505 -78.765 44.997
156227 100.527 -78.772 44.981
156228 100.548 -78.779 44.965
156229 100.569 -78.786 44.949
156230 100.590 -78.793 44.933
156231 100.611 -78.800 44.917
156232 100.627 -78.813 44.900
156233 100.633 -78.839 44.884
156234 100.639 -78.865 44.867
156235 100.645 -78.891 44.850
156236 100.665 -78.908 44.834
156237 100.693 -78.919 44.818
156238 100.721 -78.930 44.801
156239 100.749 -78.941 44.785
156240 100.777 -78.952 44.768
156241 100.805 -78.967 44.752
156242 100.848 -78.972 44.735
156243 100.892 -78.978 44.719
156244 100.936 -78.984 44.703
156245 100.979 -78.990 44.688
156246 101.023 -78.996 44.672
156247 101.067 -79.002 44.656
156248 101.141 -78.980 44.640
156249 101.194 -78.987 44.624
156250 101.239 -78.994 44.610
156251 101.263 -79.008 44.596
156252 101.287 -79.021 44.582
156253 101.312 -79.035 44.568
156254 101.336 -79.048 44.556
156255 101.360 -79.062 44.543
156256 101.385 -79.075 44.530
156257 101.404 -79.104 44.517
156258 101.420 -79.138 44.505
156259 101.450 -79.165 44.493
156485 92.509 -77.592 54.155
156486 92.529 -77.599 54.160
156487 92.549 -77.606 54.164
156488 92.569 -77.614 54.169
156489 92.589 -77.621 54.173
156490 92.609 -77.628 54.176
156491 92.629 -77.636 54.180
156492 92.649 -77.643 54.184
156493 92.669 -77.651 54.188
156494 92.689 -77.658 54.194
156495 92.709 -77.665 54.195
156496 92.729 -77.673 54.197
156497 92.748 -77.680 54.201
156498 92.769 -77.687 54.202
156499 92.789 -77.695 54.204
156500 92.808 -77.702 54.208
156501 92.829 -77.709 54.212
156502 92.849 -77.717 54.215
156503 92.873 -77.714 54.218
156504 92.901 -77.702 54.222
156505 92.918 -77.696 54.227
156506 92.934 -77.690 54.231
156507 92.947 -77.682 54.236
156508 92.961 -77.673 54.240
156509 92.975 -77.664 54.247
156510 92.988 -77.655 54.253
156511 93.002 -77.646 54.258
156512 93.015 -77.637 54.263
156513 93.028 -77.628 54.268
156514 93.042 -77.620 54.273
156515 93.055 -77.612 54.275
156516 93.067 -77.608 54.275
156517 93.079 -77.604 54.275
156518 93.091 -77.601 54.274
156519 93.102 -77.598 54.272
156520 93.114 -77.595 54.270
156521 93.126 -77.591 54.271
156522 93.138 -77.587 54.273
156523 93.148 -77.570 54.273
156524 93.159 -77.554 54.275
156525 93.170 -77.539 54.275
156526 93.182 -77.524 54.273
156527 93.193 -77.510 54.270
156528 93.205 -77.495 54.266
156529 93.223 -77.484 54.262
156530 93.247 -77.476 54.258
156531 93.271 -77.468 54.253
156532 93.295 -77.460 54.247
156533 93.316 -77.451 54.241
156534 93.330 -77.442 54.234
156535 93.343 -77.433 54.231
156536 93.356 -77.424 54.227
156537 93.370 -77.415 54.223
156538 93.383 -77.405 54.216
156539 93.397 -77.396 54.207
156540 93.410 -77.387 54.199
156541 93.423 -77.378 54.190
156542 93.437 -77.369 54.185
156543 93.450 -77.360 54.178
156544 93.466 -77.351 54.168
156545 93.485 -77.345 54.156
156546 93.503 -77.340 54.142
156547 93.522 -77.334 54.129
156548 93.541 -77.328 54.115
156549 93.560 -77.320 54.102
156550 93.581 -77.312 54.089
156551 93.602 -77.304 54.076
156552 93.622 -77.294 54.063
156553 93.643 -77.286 54.049
156554 93.663 -77.277 54.033
156555 93.684 -77.269 54.017
156556 93.703 -77.260 54.001
156557 93.718 -77.251 53.985
156558 93.733 -77.242 53.970
156559 93.753 -77.232 53.955
156560 93.775 -77.221 53.938
156561 93.790 -77.197 53.919
156562 93.801 -77.185 53.899
156563 93.815 -77.177 53.879
156564 93.826 -77.162 53.857
156565 93.835 -77.145 53.836
156566 93.837 -77.140 53.815
156567 93.842 -77.148 53.794
156568 93.848 -77.161 53.774
156569 93.854 -77.173 53.756
156570 93.860 -77.186 53.739
156571 93.866 -77.198 53.720
156572 93.872 -77.210 53.704
156573 93.896 -77.210 53.686
156574 93.909 -77.204 53.666
156575 93.925 -77.194 53.646
156576 93.941 -77.184 53.626
156577 93.957 -77.174 53.608
156578 93.973 -77.164 53.591
156579 93.991 -77.157 53.575
156580 94.015 -77.153 53.559
156581 94.039 -77.151 53.542
156582 94.063 -77.149 53.526
156583 94.088 -77.146 53.509
156584 94.119 -77.145 53.493
156585 94.150 -77.144 53.480
156586 94.182 -77.144 53.467
156587 94.213 -77.143 53.455
156588 94.245 -77.142 53.446
156589 94.274 -77.141 53.441
156590 94.279 -77.128 53.439
156591 94.284 -77.116 53.438
156592 94.289 -77.104 53.436
156593 94.294 -77.092 53.433
156594 94.300 -77.082 53.426
156595 94.308 -77.083 53.417
156596 94.315 -77.083 53.407
156597 94.322 -77.084 53.400
156598 94.330 -77.085 53.397
156599 94.335 -77.083 53.396
156600 94.335 -77.079 53.399
156601 94.336 -77.075 53.402
156602 94.336 -77.072 53.403
156603 94.337 -77.068 53.402
156604 94.337 -77.064 53.401
156605 94.337 -77.060 53.395
156606 94.336 -77.059 53.385
156607 94.350 -77.055 53.371
156608 94.369 -77.050 53.349
156609 94.390 -77.047 53.325
156610 94.411 -77.042 53.298
156611 94.432 -77.038 53.271
156612 94.454 -77.034 53.246
156613 94.475 -77.030 53.223
156614 94.496 -77.026 53.201
156615 94.517 -77.022 53.178
156616 94.538 -77.018 53.154
156617 94.559 -77.013 53.130
156618 94.582 -77.009 53.109
156619 94.611 -77.006 53.090
156620 94.625 -77.001 53.069
156621 94.633 -76.996 53.046
156622 94.643 -76.993 53.020
156623 94.654 -76.993 52.996
156624 94.666 -76.992 52.970
156625 94.678 -76.993 52.942
156626 94.691 -76.995 52.911
156627 94.703 -76.997 52.875
156628 94.716 -77.000 52.837
156629 94.728 -77.003 52.795
156630 94.767 -77.009 52.753
156631 94.807 -77.016 52.709
156632 94.847 -77.022 52.663
156633 94.871 -77.026 52.613
156634 94.879 -77.029 52.562
156635 94.887 -77.032 52.512
156636 94.895 -77.034 52.466
156637 94.910 -77.040 52.427
156638 94.929 -77.047 52.390
156639 94.948 -77.053 52.354
156640 94.965 -77.050 52.320
156641 94.981 -77.046 52.286
156642 95.002 -77.049 52.257
156643 95.023 -77.052 52.232
156644 95.044 -77.055 52.211
156645 95.064 -77.059 52.181
156646 95.085 -77.063 52.145
156647 95.105 -77.067 52.106
156648 95.125 -77.072 52.062
156649 95.145 -77.076 52.019
156650 95.165 -77.081 51.974
156651 95.185 -77.085 51.925
156652 95.205 -77.089 51.871
156653 95.222 -77.091 51.818
156654 95.233 -77.089 51.764
156655 95.243 -77.087 51.707
156656 95.254 -77.085 51.652
156657 95.264 -77.082 51.597
156658 95.275 -77.080 51.543
156659 95.285 -77.078 51.494
156660 95.303 -77.089 51.444
156661 95.321 -77.103 51.394
156662 95.342 -77.117 51.344
156663 95.365 -77.132 51.296
156664 95.387 -77.147 51.250
156665 95.409 -77.161 51.204
156666 95.431 -77.176 51.155
156667 95.453 -77.191 51.103
156668 95.473 -77.206 51.052
156669 95.491 -77.220 51.004
156670 95.509 -77.235 50.961
156671 95.526 -77.250 50.916
156672 95.544 -77.264 50.870
156673 95.562 -77.279 50.824
156674 95.578 -77.291 50.778
156675 95.593 -77.301 50.733
156676 95.608 -77.313 50.688
156677 95.622 -77.323 50.645
156678 95.637 -77.334 50.603
156679 95.655 -77.345 50.561
156680 95.672 -77.356 50.518
156681 95.690 -77.367 50.476
156682 95.707 -77.378 50.436
156683 95.725 -77.389 50.398
156684 95.745 -77.396 50.361
156685 95.762 -77.405 50.324
156686 95.782 -77.416 50.286
156687 95.803 -77.426 50.248
156688 95.823 -77.437 50.208
156689 95.843 -77.448 50.167
156690 95.864 -77.458 50.125
156691 95.884 -77.469 50.083
156692 95.904 -77.479 50.039
156693 95.925 -77.490 49.993
156694 95.952 -77.499 49.944
156695 95.966 -77.504 49.889
156696 95.979 -77.506 49.833
156697 95.990 -77.500 49.780
156698 96.002 -77.495 49.728
156699 96.013 -77.489 49.679
156700 96.021 -77.498 49.631
156701 96.028 -77.506 49.584
156702 96.036 -77.514 49.538
156703 96.043 -77.522 49.493
156704 96.051 -77.530 49.451
156705 96.058 -77.538 49.412
156706 96.066 -77.544 49.372
156707 96.076 -77.544 49.335
156708 96.087 -77.538 49.299
156709 96.096 -77.520 49.263
156710 96.104 -77.504 49.231
156711 96.112 -77.499 49.200
156712 96.119 -77.495 49.171
156713 96.127 -77.491 49.142
156714 96.135 -77.486 49.114
156715 96.143 -77.482 49.088
156716 96.149 -77.479 49.061
156717 96.146 -77.491 49.035
156718 96.143 -77.501 49.013
156719 96.144 -77.511 48.993
156720 96.152 -77.521 48.973
156721 96.167 -77.534 48.956
156722 96.181 -77.546 48.939
156723 96.195 -77.558 48.921
156724 96.209 -77.571 48.900
156725 96.224 -77.583 48.878
156726 96.238 -77.595 48.855
156727 96.251 -77.605 48.832
156728 96.262 -77.609 48.808
156729 96.279 -77.617 48.784
156730 96.308 -77.627 48.760
156731 96.336 -77.638 48.735
156732 96.362 -77.649 48.712
156733 96.387 -77.658 48.690
156734 96.412 -77.669 48.668
156735 96.438 -77.680 48.644
156736 96.463 -77.689 48.622
156737 96.483 -77.694 48.601
156738 96.503 -77.699 48.581
156739 96.523 -77.704 48.563
156740 96.540 -77.708 48.545
156741 96.551 -77.712 48.526
156742 96.562 -77.715 48.508
156743 96.572 -77.718 48.491
156744 96.579 -77.720 48.473
156745 96.586 -77.721 48.453
156746 96.592 -77.722 48.430
156747 96.599 -77.724 48.408
156748 96.606 -77.725 48.385
156749 96.613 -77.727 48.363
156750 96.620 -77.727 48.342
156751 96.626 -77.729 48.321
156752 96.633 -77.730 48.301
156753 96.643 -77.736 48.283
156754 96.652 -77.742 48.263
156755 96.662 -77.749 48.240
156756 96.671 -77.755 48.216
156757 96.685 -77.758 48.192
156758 96.699 -77.761 48.167
156759 96.714 -77.763 48.144
156760 96.728 -77.765 48.121
156761 96.742 -77.768 48.099
156762 96.756 -77.770 48.074
156763 96.770 -77.772 48.048
156764 96.785 -77.775 48.022
156765 96.799 -77.775 47.995
156766 96.814 -77.762 47.969
156767 96.829 -77.750 47.943
156768 96.848 -77.751 47.918
156769 96.870 -77.761 47.893
156770 96.892 -77.771 47.868
156771 96.914 -77.781 47.843
156772 96.937 -77.786 47.821
156773 96.960 -77.787 47.798
156774 96.983 -77.788 47.776
156775 97.005 -77.789 47.755
156776 97.018 -77.788 47.732
156777 97.030 -77.788 47.711
156778 97.043 -77.787 47.690
156779 97.055 -77.787 47.669
156780 97.067 -77.789 47.648
156781 97.078 -77.793 47.626
156782 97.090 -77.790 47.605
156783 97.104 -77.781 47.584
156784 97.117 -77.773 47.563
156785 97.131 -77.765 47.543
156786 97.149 -77.765 47.523
156787 97.169 -77.770 47.501
156788 97.190 -77.774 47.481
156789 97.214 -77.774 47.461
156790 97.237 -77.773 47.442
156791 97.261 -77.772 47.424
156792 97.285 -77.771 47.407
156793 97.308 -77.771 47.389
156794 97.332 -77.770 47.370
156795 97.361 -77.777 47.354
156796 97.391 -77.786 47.337
156797 97.422 -77.794 47.319
156798 97.452 -77.802 47.302
156799 97.482 -77.812 47.284
156800 97.493 -77.805 47.264
156801 97.507 -77.815 47.244
156802 97.523 -77.831 47.223
156803 97.538 -77.846 47.202
156804 97.554 -77.862 47.181
156805 97.570 -77.877 47.159
156806 97.585 -77.892 47.136
156807 97.601 -77.907 47.114
156808 97.616 -77.922 47.091
156809 97.632 -77.937 47.069
156810 97.642 -77.957 47.045
156811 97.639 -77.984 47.021
156812 97.636 -78.012 46.996
156813 97.633 -78.040 46.970
156814 97.631 -78.067 46.945
156815 97.628 -78.094 46.920
156816 97.628 -78.112 46.896
156817 97.645 -78.128 46.871
156818 97.665 -78.103 46.848
156819 97.710 -78.082 46.824
156820 97.745 -78.071 46.802
156821 97.780 -78.059 46.781
156822 97.814 -78.046 46.761
156823 97.846 -78.029 46.741
156824 97.877 -78.012 46.722
156825 97.908 -77.994 46.702
156826 97.940 -77.987 46.681
156827 97.972 -78.003 46.662
156828 98.003 -78.021 46.641
156829 98.025 -78.042 46.619
156830 98.046 -78.063 46.598
156831 98.064 -78.067 46.574
156832 98.146 -78.044 46.551
156833 98.192 -78.028 46.529
156834 98.229 -78.013 46.508
156835 98.266 -77.999 46.487
156836 98.302 -77.984 46.465
156837 98.339 -77.970 46.443
156838 98.376 -77.956 46.419
156839 98.413 -77.940 46.394
156840 98.449 -77.926 46.371
156841 98.486 -77.912 46.347
156842 98.522 -77.897 46.325
156843 98.559 -77.883 46.303
156844 98.596 -77.868 46.281
156845 98.628 -77.863 46.258
156846 98.641 -77.899 46.235
156847 98.653 -77.935 46.212
156848 98.666 -77.971 46.191
156849 98.679 -78.007 46.170
156850 98.685 -78.032 46.148
156851 98.672 -78.028 46.128
156852 98.664 -78.028 46.108
156853 98.665 -78.034 46.088
156854 98.665 -78.042 46.069
156855 98.666 -78.049 46.049
156856 98.667 -78.056 46.030
156857 98.668 -78.063 46.011
156858 98.668 -78.070 45.992
156859 98.669 -78.078 45.973
156860 98.670 -78.084 45.955
156861 98.671 -78.092 45.935
156862 98.699 -78.100 45.916
156863 98.751 -78.107 45.897
156864 98.804 -78.115 45.880
156865 98.857 -78.123 45.863
156866 98.909 -78.131 45.846
156867 98.962 -78.139 45.829
156868 99.015 -78.147 45.813
156869 99.067 -78.154 45.798
156870 99.114 -78.162 45.783
156871 99.119 -78.163 45.769
156872 99.124 -78.163 45.755
156873 99.130 -78.165 45.741
156874 99.144 -78.166 45.728
156875 99.199 -78.166 45.716
156876 99.218 -78.163 45.704
156877 99.237 -78.159 45.693
156878 99.256 -78.157 45.681
156879 99.278 -78.159 45.670
156880 99.300 -78.162 45.658
156881 99.322 -78.163 45.646
156882 99.344 -78.166 45.634
156883 99.366 -78.168 45.623
156884 99.389 -78.170 45.611
156885 99.411 -78.172 45.599
156886 99.433 -78.175 45.587
156887 99.460 -78.188 45.576
156888 99.490 -78.207 45.565
156889 99.520 -78.227 45.555
156890 99.549 -78.246 45.543
156891 99.577 -78.265 45.532
156892 99.607 -78.285 45.521
156893 99.635 -78.304 45.511
156894 99.664 -78.322 45.500
156895 99.693 -78.342 45.488
156896 99.719 -78.373 45.478
156897 99.753 -78.391 45.465
156898 99.786 -78.409 45.452
156899 99.820 -78.426 45.440
156900 99.859 -78.426 45.427
156901 99.862 -78.464 45.414
156902 99.854 -78.485 45.402
156903 99.883 -78.506 45.388
156904 99.919 -78.526 45.374
156905 99.956 -78.548 45.360
156906 99.992 -78.568 45.345
156907 100.027 -78.586 45.331
156908 100.060 -78.600 45.317
156909 100.093 -78.615 45.302
156910 100.127 -78.630 45.287
156911 100.160 -78.645 45.273
156912 100.193 -78.659 45.256
156913 100.226 -78.674 45.240
156914 100.260 -78.689 45.223
156915 100.293 -78.704 45.206
156916 100.327 -78.718 45.190
156917 100.321 -78.711 45.174
156918 100.340 -78.714 45.156
156919 100.358 -78.717 45.139
156920 100.376 -78.720 45.122
156921 100.395 -78.723 45.104
156922 100.414 -78.726 45.087
156923 100.432 -78.730 45.070
156924 100.451 -78.733 45.053
156925 100.470 -78.742 45.036
156926 100.490 -78.749 45.020
156927 100.511 -78.756 45.004
156928 100.533 -78.763 44.988
156929 100.554 -78.770 44.973
156930 100.575 -78.777 44.957
156931 100.596 -78.783 44.941
156932 100.617 -78.791 44.925
156933 100.638 -78.798 44.908
156934 100.645 -78.824 44.892
156935 100.651 -78.849 44.876
156936 100.658 -78.874 44.858
156937 100.686 -78.886 44.842
156938 100.714 -78.897 44.826
156939 100.743 -78.908 44.810
156940 100.771 -78.919 44.793
156941 100.796 -78.932 44.777
156942 100.822 -78.945 44.761
156943 100.868 -78.949 44.745
156944 100.916 -78.952 44.728
156945 100.963 -78.954 44.711
156946 101.010 -78.957 44.694
156947 101.055 -78.962 44.679
156948 101.098 -78.968 44.664
156949 101.142 -78.974 44.648
156950 101.199 -78.970 44.632
156951 101.238 -78.979 44.618
156952 101.262 -78.993 44.604
156953 101.286 -79.006 44.591
156954 101.311 -79.020 44.577
156955 101.335 -79.034 44.565
156956 101.359 -79.046 44.552
156957 101.384 -79.060 44.539
156958 101.402 -79.090 44.526
156959 101.419 -79.125 44.513
156960 101.449 -79.151 44.501
157186 92.498 -77.572 54.143
157187 92.518 -77.579 54.149
157188 92.538 -77.586 54.154
157189 92.558 -77.594 54.160
157190 92.578 -77.601 54.165
157191 92.598 -77.608 54.170
157192 92.618 -77.616 54.175
157193 92.638 -77.623 54.181
157194 92.658 -77.630 54.186
157195 92.678 -77.638 54.193
157196 92.698 -77.645 54.195
157197 92.718 -77.652 54.200
157198 92.738 -77.660 54.204
157199 92.758 -77.667 54.207
157200 92.778 -77.674 54.210
157201 92.797 -77.682 54.214
157202 92.817 -77.689 54.218
157203 92.838 -77.696 54.221
157204 92.862 -77.694 54.226
157205 92.890 -77.681 54.231
157206 92.908 -77.676 54.236
157207 92.921 -77.667 54.242
157208 92.935 -77.658 54.247
157209 92.948 -77.649 54.252
157210 92.961 -77.640 54.258
157211 92.976 -77.631 54.262
157212 92.990 -77.624 54.266
157213 93.004 -77.616 54.271
157214 93.018 -77.608 54.276
157215 93.032 -77.601 54.280
157216 93.042 -77.597 54.284
157217 93.053 -77.594 54.284
157218 93.064 -77.590 54.284
157219 93.076 -77.587 54.284
157220 93.087 -77.584 54.283
157221 93.099 -77.580 54.281
157222 93.111 -77.577 54.282
157223 93.123 -77.571 54.284
157224 93.133 -77.554 54.286
157225 93.144 -77.537 54.288
157226 93.155 -77.520 54.289
157227 93.166 -77.506 54.289
157228 93.178 -77.492 54.287
157229 93.191 -77.478 54.285
157230 93.215 -77.470 54.282
157231 93.239 -77.461 54.279
157232 93.264 -77.454 54.276
157233 93.288 -77.445 54.271
157234 93.312 -77.437 54.265
157235 93.326 -77.428 54.259
157236 93.339 -77.419 54.254
157237 93.352 -77.410 54.250
157238 93.365 -77.401 54.247
157239 93.379 -77.392 54.241
157240 93.393 -77.382 54.233
157241 93.406 -77.373 54.224
157242 93.419 -77.364 54.214
157243 93.431 -77.353 54.208
157244 93.441 -77.339 54.200
157245 93.460 -77.332 54.191
157246 93.478 -77.326 54.180
157247 93.496 -77.320 54.166
157248 93.515 -77.314 54.153
157249 93.534 -77.308 54.140
157250 93.553 -77.301 54.129
157251 93.574 -77.292 54.116
157252 93.594 -77.283 54.104
157253 93.615 -77.275 54.091
157254 93.636 -77.266 54.078
157255 93.656 -77.257 54.062
157256 93.677 -77.249 54.045
157257 93.698 -77.240 54.030
157258 93.718 -77.232 54.014
157259 93.738 -77.223 53.999
157260 93.754 -77.214 53.984
157261 93.773 -77.198 53.967
157262 93.785 -77.188 53.949
157263 93.798 -77.179 53.929
157264 93.811 -77.171 53.909
157265 93.822 -77.154 53.887
157266 93.825 -77.148 53.865
157267 93.827 -77.143 53.844
157268 93.829 -77.138 53.825
157269 93.833 -77.143 53.805
157270 93.839 -77.156 53.788
157271 93.845 -77.168 53.772
157272 93.851 -77.181 53.755
157273 93.868 -77.186 53.738
157274 93.897 -77.185 53.720
157275 93.908 -77.184 53.699
157276 93.921 -77.176 53.678
157277 93.937 -77.166 53.657
157278 93.954 -77.157 53.638
157279 93.970 -77.147 53.620
157280 93.988 -77.139 53.603
157281 94.012 -77.136 53.586
157282 94.036 -77.134 53.568
157283 94.060 -77.131 53.551
157284 94.083 -77.127 53.533
157285 94.110 -77.124 53.517
157286 94.141 -77.123 53.502
157287 94.173 -77.122 53.488
157288 94.204 -77.122 53.474
157289 94.236 -77.121 53.461
157290 94.257 -77.116 53.452
157291 94.262 -77.103 53.448
157292 94.267 -77.091 53.445
157293 94.272 -77.079 53.442
157294 94.280 -77.071 53.439
157295 94.293 -77.068 53.432
157296 94.308 -77.066 53.424
157297 94.318 -77.066 53.413
157298 94.325 -77.066 53.405
157299 94.333 -77.067 53.401
157300 94.339 -77.066 53.399
157301 94.339 -77.063 53.400
157302 94.340 -77.059 53.402
157303 94.340 -77.055 53.403
157304 94.341 -77.050 53.403
157305 94.340 -77.049 53.403
157306 94.338 -77.047 53.398
157307 94.338 -77.045 53.390
157308 94.356 -77.041 53.378
157309 94.374 -77.036 53.356
157310 94.393 -77.032 53.332
157311 94.413 -77.028 53.305
157312 94.434 -77.024 53.277
157313 94.455 -77.020 53.250
157314 94.476 -77.016 53.226
157315 94.497 -77.012 53.203
157316 94.519 -77.007 53.179
157317 94.540 -77.003 53.154
157318 94.561 -76.999 53.130
157319 94.589 -76.995 53.108
157320 94.617 -76.992 53.088
157321 94.626 -76.986 53.066
157322 94.634 -76.981 53.041
157323 94.642 -76.977 53.014
157324 94.654 -76.976 52.988
157325 94.666 -76.976 52.962
157326 94.678 -76.975 52.933
157327 94.689 -76.976 52.901
157328 94.702 -76.978 52.865
157329 94.714 -76.981 52.827
157330 94.728 -76.984 52.785
157331 94.768 -76.990 52.744
157332 94.808 -76.997 52.702
157333 94.848 -77.003 52.659
157334 94.886 -77.009 52.612
157335 94.894 -77.012 52.564
157336 94.902 -77.015 52.518
157337 94.910 -77.018 52.475
157338 94.918 -77.021 52.441
157339 94.933 -77.025 52.408
157340 94.952 -77.032 52.375
157341 94.969 -77.031 52.345
157342 94.987 -77.030 52.313
157343 95.008 -77.033 52.286
157344 95.029 -77.036 52.262
157345 95.049 -77.041 52.241
157346 95.069 -77.044 52.212
157347 95.090 -77.049 52.177
157348 95.110 -77.053 52.136
157349 95.130 -77.058 52.094
157350 95.150 -77.063 52.051
157351 95.170 -77.066 52.006
157352 95.190 -77.071 51.956
157353 95.210 -77.075 51.901
157354 95.230 -77.079 51.847
157355 95.241 -77.078 51.793
157356 95.251 -77.075 51.739
157357 95.262 -77.073 51.681
157358 95.273 -77.071 51.623
157359 95.283 -77.069 51.566
157360 95.300 -77.079 51.513
157361 95.318 -77.093 51.462
157362 95.336 -77.106 51.411
157363 95.354 -77.119 51.361
157364 95.373 -77.132 51.310
157365 95.390 -77.146 51.261
157366 95.412 -77.160 51.213
157367 95.434 -77.175 51.166
157368 95.456 -77.190 51.117
157369 95.476 -77.205 51.070
157370 95.494 -77.219 51.025
157371 95.512 -77.234 50.980
157372 95.529 -77.248 50.934
157373 95.544 -77.258 50.887
157374 95.558 -77.269 50.839
157375 95.573 -77.279 50.792
157376 95.589 -77.290 50.747
157377 95.603 -77.301 50.703
157378 95.618 -77.312 50.660
157379 95.633 -77.323 50.618
157380 95.651 -77.333 50.576
157381 95.669 -77.345 50.532
157382 95.687 -77.354 50.490
157383 95.707 -77.360 50.448
157384 95.727 -77.367 50.408
157385 95.747 -77.373 50.371
157386 95.755 -77.380 50.334
157387 95.776 -77.391 50.296
157388 95.796 -77.401 50.257
157389 95.817 -77.412 50.216
157390 95.837 -77.423 50.174
157391 95.858 -77.433 50.132
157392 95.878 -77.444 50.089
157393 95.898 -77.454 50.044
157394 95.920 -77.465 49.997
157395 95.952 -77.475 49.948
157396 95.965 -77.479 49.896
157397 95.979 -77.484 49.842
157398 95.993 -77.489 49.789
157399 96.005 -77.485 49.737
157400 96.016 -77.479 49.688
157401 96.024 -77.486 49.640
157402 96.031 -77.495 49.593
157403 96.039 -77.502 49.547
157404 96.047 -77.511 49.503
157405 96.054 -77.518 49.461
157406 96.062 -77.526 49.421
157407 96.070 -77.534 49.380
157408 96.077 -77.542 49.343
157409 96.088 -77.536 49.306
157410 96.096 -77.523 49.270
157411 96.103 -77.519 49.236
157412 96.111 -77.514 49.205
157413 96.119 -77.510 49.175
157414 96.127 -77.505 49.145
157415 96.135 -77.501 49.117
157416 96.143 -77.496 49.090
157417 96.151 -77.492 49.063
157418 96.153 -77.495 49.036
157419 96.155 -77.504 49.015
157420 96.159 -77.513 48.995
157421 96.164 -77.521 48.976
157422 96.174 -77.533 48.959
157423 96.189 -77.547 48.944
157424 96.203 -77.560 48.928
157425 96.218 -77.572 48.908
157426 96.231 -77.583 48.887
157427 96.243 -77.588 48.864
157428 96.254 -77.592 48.840
157429 96.265 -77.598 48.816
157430 96.282 -77.605 48.793
157431 96.310 -77.615 48.769
157432 96.339 -77.626 48.745
157433 96.368 -77.636 48.721
157434 96.396 -77.648 48.700
157435 96.422 -77.658 48.677
157436 96.448 -77.668 48.653
157437 96.473 -77.679 48.633
157438 96.497 -77.688 48.614
157439 96.517 -77.693 48.596
157440 96.537 -77.698 48.579
157441 96.553 -77.702 48.560
157442 96.565 -77.705 48.542
157443 96.575 -77.708 48.524
157444 96.582 -77.710 48.507
157445 96.589 -77.711 48.489
157446 96.596 -77.713 48.469
157447 96.602 -77.714 48.446
157448 96.609 -77.715 48.424
157449 96.616 -77.717 48.401
157450 96.623 -77.718 48.380
157451 96.630 -77.720 48.361
157452 96.637 -77.721 48.340
157453 96.644 -77.723 48.321
157454 96.651 -77.725 48.303
157455 96.661 -77.732 48.284
157456 96.670 -77.739 48.261
157457 96.680 -77.745 48.238
157458 96.693 -77.748 48.214
157459 96.708 -77.750 48.190
157460 96.722 -77.753 48.166
157461 96.736 -77.755 48.142
157462 96.750 -77.758 48.119
157463 96.764 -77.760 48.092
157464 96.779 -77.762 48.066
157465 96.793 -77.765 48.040
157466 96.807 -77.768 48.014
157467 96.821 -77.770 47.989
157468 96.836 -77.758 47.963
157469 96.852 -77.747 47.937
157470 96.873 -77.757 47.912
157471 96.895 -77.768 47.887
157472 96.918 -77.776 47.863
157473 96.941 -77.777 47.839
157474 96.964 -77.778 47.818
157475 96.987 -77.780 47.796
157476 97.011 -77.780 47.773
157477 97.023 -77.780 47.752
157478 97.036 -77.779 47.731
157479 97.048 -77.779 47.710
157480 97.060 -77.783 47.689
157481 97.071 -77.786 47.669
157482 97.081 -77.790 47.648
157483 97.095 -77.781 47.626
157484 97.108 -77.773 47.604
157485 97.121 -77.764 47.583
157486 97.135 -77.755 47.563
157487 97.148 -77.747 47.543
157488 97.166 -77.747 47.522
157489 97.187 -77.751 47.501
157490 97.211 -77.750 47.481
157491 97.235 -77.749 47.461
157492 97.258 -77.749 47.443
157493 97.287 -77.755 47.425
157494 97.317 -77.764 47.407
157495 97.348 -77.772 47.388
157496 97.378 -77.780 47.371
157497 97.408 -77.790 47.354
157498 97.438 -77.798 47.336
157499 97.468 -77.806 47.317
157500 97.499 -77.815 47.298
157501 97.512 -77.810 47.279
157502 97.521 -77.803 47.258
157503 97.536 -77.818 47.237
157504 97.552 -77.834 47.216
157505 97.568 -77.849 47.195
157506 97.583 -77.865 47.173
157507 97.599 -77.880 47.151
157508 97.614 -77.895 47.129
157509 97.630 -77.910 47.105
157510 97.645 -77.925 47.082
157511 97.650 -77.948 47.058
157512 97.647 -77.975 47.034
157513 97.644 -78.003 47.009
157514 97.641 -78.031 46.983
157515 97.639 -78.058 46.957
157516 97.636 -78.086 46.931
157517 97.633 -78.113 46.906
157518 97.647 -78.105 46.882
157519 97.684 -78.073 46.858
157520 97.719 -78.062 46.834
157521 97.752 -78.047 46.811
157522 97.783 -78.030 46.789
157523 97.814 -78.012 46.768
157524 97.846 -77.996 46.748
157525 97.881 -77.980 46.727
157526 97.916 -77.965 46.707
157527 97.945 -77.956 46.686
157528 97.947 -77.962 46.666
157529 97.969 -77.984 46.644
157530 97.991 -78.005 46.623
157531 98.013 -78.026 46.602
157532 98.034 -78.047 46.579
157533 98.060 -78.021 46.557
157534 98.142 -77.997 46.536
157535 98.225 -77.975 46.516
157536 98.282 -77.956 46.495
157537 98.318 -77.941 46.473
157538 98.355 -77.927 46.450
157539 98.392 -77.912 46.427
157540 98.429 -77.898 46.403
157541 98.465 -77.884 46.379
157542 98.502 -77.868 46.356
157543 98.539 -77.854 46.334
157544 98.576 -77.840 46.312
157545 98.612 -77.825 46.289
157546 98.635 -77.840 46.266
157547 98.648 -77.876 46.242
157548 98.661 -77.912 46.220
157549 98.673 -77.947 46.199
157550 98.686 -77.984 46.178
157551 98.697 -78.016 46.158
157552 98.684 -78.012 46.138
157553 98.681 -78.016 46.118
157554 98.682 -78.023 46.099
157555 98.683 -78.031 46.079
157556 98.684 -78.037 46.059
157557 98.685 -78.045 46.040
157558 98.685 -78.052 46.021
157559 98.686 -78.059 46.003
157560 98.687 -78.066 45.984
157561 98.688 -78.073 45.964
157562 98.689 -78.081 45.944
157563 98.689 -78.087 45.925
157564 98.740 -78.096 45.906
157565 98.792 -78.103 45.889
157566 98.845 -78.111 45.872
157567 98.898 -78.119 45.856
157568 98.950 -78.127 45.840
157569 99.003 -78.135 45.822
157570 99.056 -78.143 45.807
157571 99.076 -78.146 45.792
157572 99.082 -78.147 45.778
157573 99.087 -78.148 45.764
157574 99.104 -78.149 45.749
157575 99.160 -78.150 45.736
157576 99.213 -78.150 45.723
157577 99.231 -78.147 45.712
157578 99.250 -78.144 45.701
157579 99.268 -78.140 45.689
157580 99.286 -78.137 45.678
157581 99.305 -78.134 45.667
157582 99.327 -78.135 45.655
157583 99.349 -78.138 45.643
157584 99.371 -78.140 45.631
157585 99.393 -78.142 45.619
157586 99.415 -78.144 45.607
157587 99.438 -78.147 45.596
157588 99.468 -78.167 45.585
157589 99.498 -78.187 45.575
157590 99.528 -78.206 45.564
157591 99.558 -78.225 45.552
157592 99.586 -78.244 45.541
157593 99.615 -78.264 45.532
157594 99.644 -78.283 45.521
157595 99.673 -78.304 45.509
157596 99.703 -78.329 45.497
157597 99.736 -78.346 45.487
157598 99.770 -78.364 45.474
157599 99.803 -78.382 45.461
157600 99.837 -78.400 45.447
157601 99.870 -78.417 45.436
157602 99.865 -78.438 45.423
157603 99.857 -78.459 45.410
157604 99.868 -78.480 45.396
157605 99.904 -78.501 45.382
157606 99.940 -78.522 45.368
157607 99.977 -78.542 45.355
157608 100.013 -78.563 45.341
157609 100.050 -78.584 45.326
157610 100.086 -78.604 45.310
157611 100.122 -78.625 45.295
157612 100.158 -78.646 45.280
157613 100.193 -78.663 45.263
157614 100.226 -78.677 45.247
157615 100.260 -78.692 45.230
157616 100.293 -78.707 45.213
157617 100.326 -78.722 45.196
157618 100.353 -78.733 45.179
157619 100.346 -78.724 45.161
157620 100.364 -78.727 45.144
157621 100.382 -78.730 45.127
157622 100.401 -78.733 45.110
157623 100.420 -78.736 45.093
157624 100.437 -78.738 45.076
157625 100.442 -78.722 45.060
157626 100.460 -78.727 45.043
157627 100.498 -78.738 45.027
157628 100.518 -78.747 45.011
157629 100.539 -78.754 44.996
157630 100.560 -78.761 44.980
157631 100.581 -78.767 44.964
157632 100.602 -78.775 44.948
157633 100.623 -78.782 44.932
157634 100.644 -78.789 44.916
157635 100.656 -78.808 44.900
157636 100.662 -78.833 44.884
157637 100.680 -78.852 44.867
157638 100.708 -78.863 44.849
157639 100.736 -78.874 44.834
157640 100.764 -78.886 44.818
157641 100.781 -78.903 44.801
157642 100.799 -78.921 44.785
157643 100.833 -78.930 44.769
157644 100.875 -78.937 44.752
157645 100.922 -78.940 44.735
157646 100.969 -78.943 44.719
157647 101.016 -78.946 44.702
157648 101.064 -78.948 44.687
157649 101.111 -78.951 44.671
157650 101.159 -78.953 44.656
157651 101.206 -78.956 44.641
157652 101.237 -78.965 44.626
157653 101.261 -78.977 44.612
157654 101.285 -78.991 44.599
157655 101.310 -79.005 44.586
157656 101.334 -79.018 44.573
157657 101.359 -79.032 44.560
157658 101.383 -79.045 44.547
157659 101.400 -79.078 44.534
157660 101.417 -79.112 44.522
157661 101.449 -79.137 44.509
157887 92.487 -77.551 54.129
157888 92.507 -77.559 54.137
157889 92.527 -77.567 54.144
157890 92.547 -77.573 54.151
157891 92.567 -77.581 54.156
157892 92.587 -77.589 54.162
157893 92.607 -77.595 54.170
157894 92.627 -77.603 54.176
157895 92.647 -77.611 54.183
157896 92.667 -77.617 54.191
157897 92.687 -77.625 54.196
157898 92.707 -77.633 54.201
157899 92.727 -77.639 54.207
157900 92.746 -77.647 54.212
157901 92.767 -77.655 54.216
157902 92.787 -77.662 54.221
157903 92.806 -77.669 54.225
157904 92.827 -77.677 54.230
157905 92.851 -77.674 54.234
157906 92.879 -77.661 54.239
157907 92.896 -77.653 54.244
157908 92.910 -77.645 54.251
157909 92.924 -77.638 54.257
157910 92.938 -77.630 54.262
157911 92.952 -77.623 54.268
157912 92.967 -77.615 54.272
157913 92.981 -77.608 54.275
157914 92.995 -77.600 54.279
157915 93.009 -77.592 54.283
157916 93.020 -77.588 54.287
157917 93.028 -77.584 54.289
157918 93.038 -77.580 54.290
157919 93.049 -77.577 54.291
157920 93.061 -77.573 54.292
157921 93.073 -77.570 54.292
157922 93.084 -77.567 54.292
157923 93.096 -77.564 54.293
157924 93.108 -77.555 54.295
157925 93.118 -77.538 54.299
157926 93.129 -77.520 54.302
157927 93.139 -77.504 54.304
157928 93.150 -77.488 54.306
157929 93.162 -77.473 54.304
157930 93.183 -77.464 54.303
157931 93.207 -77.455 54.302
157932 93.231 -77.447 54.300
157933 93.256 -77.439 54.298
157934 93.280 -77.431 54.294
157935 93.304 -77.423 54.289
157936 93.321 -77.414 54.283
157937 93.335 -77.404 54.277
157938 93.348 -77.395 54.273
157939 93.361 -77.386 54.271
157940 93.375 -77.377 54.266
157941 93.389 -77.368 54.258
157942 93.402 -77.359 54.248
157943 93.412 -77.345 54.237
157944 93.420 -77.329 54.228
157945 93.435 -77.320 54.220
157946 93.453 -77.314 54.212
157947 93.471 -77.307 54.202
157948 93.489 -77.301 54.190
157949 93.507 -77.294 54.178
157950 93.526 -77.289 54.166
157951 93.546 -77.281 54.154
157952 93.567 -77.272 54.142
157953 93.587 -77.263 54.129
157954 93.608 -77.255 54.117
157955 93.629 -77.246 54.104
157956 93.649 -77.238 54.088
157957 93.670 -77.229 54.072
157958 93.690 -77.220 54.057
157959 93.711 -77.211 54.042
157960 93.732 -77.202 54.027
157961 93.754 -77.193 54.014
157962 93.768 -77.189 53.996
157963 93.782 -77.182 53.977
157964 93.794 -77.173 53.958
157965 93.807 -77.165 53.937
157966 93.812 -77.157 53.916
157967 93.814 -77.151 53.894
157968 93.816 -77.146 53.873
157969 93.818 -77.141 53.855
157970 93.820 -77.135 53.837
157971 93.824 -77.138 53.821
157972 93.830 -77.150 53.806
157973 93.837 -77.162 53.791
157974 93.866 -77.161 53.773
157975 93.895 -77.160 53.755
157976 93.911 -77.158 53.734
157977 93.920 -77.157 53.711
157978 93.934 -77.149 53.689
157979 93.950 -77.139 53.669
157980 93.966 -77.129 53.650
157981 93.985 -77.122 53.632
157982 94.009 -77.119 53.614
157983 94.033 -77.116 53.596
157984 94.056 -77.113 53.578
157985 94.080 -77.108 53.561
157986 94.104 -77.104 53.544
157987 94.132 -77.102 53.529
157988 94.164 -77.101 53.514
157989 94.195 -77.100 53.497
157990 94.226 -77.100 53.481
157991 94.239 -77.091 53.468
157992 94.244 -77.078 53.459
157993 94.249 -77.066 53.452
157994 94.258 -77.059 53.447
157995 94.271 -77.056 53.441
157996 94.284 -77.054 53.434
157997 94.299 -77.052 53.426
157998 94.315 -77.050 53.418
157999 94.328 -77.049 53.412
158000 94.335 -77.049 53.408
158001 94.342 -77.050 53.406
158002 94.343 -77.046 53.406
158003 94.343 -77.041 53.405
158004 94.343 -77.038 53.405
158005 94.342 -77.037 53.405
158006 94.340 -77.035 53.404
158007 94.339 -77.033 53.399
158008 94.343 -77.031 53.390
158009 94.361 -77.026 53.378
158010 94.380 -77.022 53.357
158011 94.398 -77.018 53.333
158012 94.416 -77.014 53.308
158013 94.435 -77.009 53.281
158014 94.457 -77.006 53.254
158015 94.478 -77.001 53.228
158016 94.499 -76.997 53.205
158017 94.521 -76.993 53.179
158018 94.542 -76.989 53.153
158019 94.567 -76.985 53.128
158020 94.595 -76.981 53.106
158021 94.619 -76.977 53.084
158022 94.627 -76.972 53.061
158023 94.634 -76.966 53.035
158024 94.642 -76.961 53.006
158025 94.654 -76.960 52.978
158026 94.666 -76.960 52.949
158027 94.677 -76.959 52.920
158028 94.689 -76.959 52.888
158029 94.701 -76.960 52.853
158030 94.713 -76.962 52.815
158031 94.729 -76.965 52.774
158032 94.769 -76.972 52.733
158033 94.809 -76.978 52.693
158034 94.849 -76.984 52.652
158035 94.889 -76.991 52.609
158036 94.908 -76.995 52.566
158037 94.917 -76.998 52.525
158038 94.925 -77.001 52.487
158039 94.933 -77.003 52.456
158040 94.942 -77.006 52.426
158041 94.956 -77.012 52.397
158042 94.973 -77.012 52.370
158043 94.993 -77.014 52.341
158044 95.014 -77.018 52.314
158045 95.034 -77.022 52.290
158046 95.054 -77.026 52.268
158047 95.074 -77.031 52.241
158048 95.095 -77.035 52.207
158049 95.115 -77.040 52.167
158050 95.135 -77.044 52.126
158051 95.155 -77.048 52.083
158052 95.175 -77.053 52.037
158053 95.195 -77.057 51.986
158054 95.215 -77.062 51.931
158055 95.235 -77.066 51.877
158056 95.249 -77.066 51.823
158057 95.260 -77.063 51.768
158058 95.270 -77.061 51.708
158059 95.281 -77.059 51.648
158060 95.298 -77.069 51.589
158061 95.316 -77.083 51.534
158062 95.334 -77.096 51.480
158063 95.352 -77.110 51.427
158064 95.370 -77.122 51.376
158065 95.388 -77.136 51.325
158066 95.406 -77.150 51.275
158067 95.424 -77.163 51.226
158068 95.442 -77.176 51.180
158069 95.460 -77.190 51.135
158070 95.480 -77.204 51.091
158071 95.494 -77.215 51.047
158072 95.510 -77.226 51.000
158073 95.525 -77.236 50.952
158074 95.539 -77.247 50.903
158075 95.554 -77.257 50.854
158076 95.569 -77.268 50.807
158077 95.584 -77.279 50.761
158078 95.599 -77.289 50.718
158079 95.614 -77.300 50.675
158080 95.630 -77.311 50.632
158081 95.650 -77.318 50.589
158082 95.670 -77.324 50.546
158083 95.689 -77.331 50.503
158084 95.709 -77.338 50.460
158085 95.729 -77.344 50.419
158086 95.749 -77.351 50.381
158087 95.756 -77.357 50.342
158088 95.770 -77.365 50.304
158089 95.790 -77.376 50.265
158090 95.811 -77.386 50.224
158091 95.831 -77.397 50.183
158092 95.851 -77.407 50.140
158093 95.872 -77.418 50.097
158094 95.892 -77.429 50.051
158095 95.916 -77.439 50.004
158096 95.949 -77.450 49.954
158097 95.965 -77.455 49.902
158098 95.979 -77.460 49.849
158099 95.993 -77.464 49.797
158100 96.007 -77.470 49.747
158101 96.019 -77.469 49.697
158102 96.027 -77.475 49.650
158103 96.035 -77.483 49.603
158104 96.042 -77.491 49.556
158105 96.050 -77.499 49.512
158106 96.057 -77.507 49.470
158107 96.065 -77.515 49.429
158108 96.070 -77.520 49.389
158109 96.075 -77.525 49.351
158110 96.080 -77.526 49.314
158111 96.087 -77.519 49.276
158112 96.094 -77.512 49.241
158113 96.101 -77.506 49.209
158114 96.108 -77.499 49.178
158115 96.115 -77.492 49.149
158116 96.122 -77.486 49.120
158117 96.129 -77.479 49.092
158118 96.139 -77.484 49.065
158119 96.158 -77.494 49.038
158120 96.169 -77.505 49.017
158121 96.174 -77.515 48.997
158122 96.179 -77.523 48.978
158123 96.183 -77.532 48.961
158124 96.195 -77.545 48.947
158125 96.211 -77.558 48.932
158126 96.223 -77.566 48.916
158127 96.235 -77.570 48.894
158128 96.246 -77.576 48.870
158129 96.257 -77.581 48.846
158130 96.268 -77.586 48.822
158131 96.284 -77.592 48.800
158132 96.312 -77.603 48.777
158133 96.341 -77.614 48.752
158134 96.370 -77.625 48.731
158135 96.398 -77.636 48.709
158136 96.427 -77.646 48.687
158137 96.456 -77.657 48.666
158138 96.482 -77.667 48.648
158139 96.508 -77.678 48.629
158140 96.532 -77.687 48.611
158141 96.552 -77.692 48.594
158142 96.568 -77.696 48.577
158143 96.578 -77.699 48.560
158144 96.585 -77.701 48.541
158145 96.592 -77.702 48.524
158146 96.599 -77.703 48.505
158147 96.606 -77.705 48.483
158148 96.612 -77.706 48.462
158149 96.619 -77.708 48.441
158150 96.626 -77.709 48.419
158151 96.633 -77.711 48.398
158152 96.640 -77.711 48.379
158153 96.647 -77.713 48.360
158154 96.654 -77.714 48.341
158155 96.660 -77.716 48.323
158156 96.670 -77.721 48.305
158157 96.679 -77.728 48.284
158158 96.689 -77.735 48.261
158159 96.702 -77.738 48.237
158160 96.716 -77.740 48.212
158161 96.730 -77.743 48.187
158162 96.745 -77.746 48.163
158163 96.759 -77.748 48.138
158164 96.773 -77.750 48.111
158165 96.787 -77.752 48.085
158166 96.801 -77.755 48.059
158167 96.816 -77.758 48.034
158168 96.830 -77.760 48.008
158169 96.844 -77.762 47.982
158170 96.859 -77.755 47.956
158171 96.877 -77.753 47.931
158172 96.899 -77.764 47.906
158173 96.922 -77.768 47.882
158174 96.945 -77.768 47.858
158175 96.968 -77.770 47.837
158176 96.992 -77.771 47.815
158177 97.015 -77.772 47.794
158178 97.029 -77.771 47.772
158179 97.040 -77.773 47.751
158180 97.051 -77.777 47.730
158181 97.063 -77.780 47.710
158182 97.074 -77.783 47.689
158183 97.086 -77.781 47.668
158184 97.099 -77.772 47.646
158185 97.112 -77.764 47.625
158186 97.126 -77.755 47.603
158187 97.139 -77.747 47.583
158188 97.153 -77.738 47.563
158189 97.166 -77.730 47.542
158190 97.184 -77.727 47.522
158191 97.213 -77.733 47.501
158192 97.243 -77.742 47.481
158193 97.273 -77.750 47.462
158194 97.304 -77.758 47.444
158195 97.334 -77.768 47.425
158196 97.364 -77.776 47.406
158197 97.394 -77.784 47.389
158198 97.425 -77.793 47.370
158199 97.455 -77.802 47.352
158200 97.485 -77.810 47.332
158201 97.515 -77.818 47.313
158202 97.530 -77.815 47.293
158203 97.539 -77.807 47.272
158204 97.550 -77.806 47.250
158205 97.566 -77.821 47.229
158206 97.581 -77.837 47.209
158207 97.597 -77.852 47.187
158208 97.612 -77.867 47.166
158209 97.628 -77.882 47.143
158210 97.643 -77.898 47.119
158211 97.659 -77.913 47.095
158212 97.658 -77.939 47.071
158213 97.655 -77.967 47.047
158214 97.652 -77.994 47.022
158215 97.650 -78.022 46.996
158216 97.647 -78.050 46.970
158217 97.644 -78.077 46.943
158218 97.656 -78.090 46.917
158219 97.659 -78.075 46.892
158220 97.695 -78.051 46.868
158221 97.730 -78.036 46.845
158222 97.764 -78.021 46.821
158223 97.800 -78.006 46.798
158224 97.835 -77.990 46.776
158225 97.870 -77.975 46.754
158226 97.906 -77.960 46.732
158227 97.941 -77.945 46.712
158228 97.955 -77.951 46.692
158229 97.960 -77.956 46.671
158230 97.955 -77.960 46.649
158231 97.956 -77.968 46.627
158232 97.979 -77.989 46.606
158233 98.000 -78.010 46.584
158234 97.998 -77.998 46.564
158235 98.055 -77.974 46.544
158236 98.138 -77.950 46.524
158237 98.220 -77.928 46.503
158238 98.303 -77.904 46.481
158239 98.372 -77.884 46.458
158240 98.408 -77.869 46.434
158241 98.445 -77.855 46.411
158242 98.482 -77.840 46.387
158243 98.518 -77.826 46.365
158244 98.555 -77.812 46.343
158245 98.592 -77.796 46.320
158246 98.629 -77.782 46.297
158247 98.642 -77.816 46.274
158248 98.655 -77.852 46.252
158249 98.668 -77.888 46.229
158250 98.680 -77.924 46.208
158251 98.694 -77.960 46.189
158252 98.706 -77.996 46.168
158253 98.698 -77.998 46.148
158254 98.699 -78.005 46.127
158255 98.700 -78.012 46.109
158256 98.700 -78.019 46.090
158257 98.701 -78.026 46.070
158258 98.702 -78.034 46.051
158259 98.703 -78.040 46.032
158260 98.704 -78.048 46.013
158261 98.704 -78.055 45.994
158262 98.705 -78.062 45.974
158263 98.706 -78.069 45.954
158264 98.707 -78.076 45.934
158265 98.728 -78.084 45.915
158266 98.781 -78.091 45.899
158267 98.834 -78.100 45.882
158268 98.886 -78.107 45.865
158269 98.939 -78.116 45.849
158270 98.992 -78.123 45.832
158271 99.034 -78.129 45.816
158272 99.040 -78.131 45.801
158273 99.045 -78.131 45.787
158274 99.065 -78.132 45.772
158275 99.121 -78.133 45.758
158276 99.175 -78.134 45.744
158277 99.225 -78.133 45.731
158278 99.243 -78.130 45.719
158279 99.262 -78.127 45.708
158280 99.281 -78.124 45.697
158281 99.299 -78.120 45.686
158282 99.317 -78.117 45.675
158283 99.336 -78.114 45.663
158284 99.354 -78.111 45.651
158285 99.375 -78.111 45.639
158286 99.397 -78.113 45.626
158287 99.419 -78.116 45.615
158288 99.445 -78.126 45.603
158289 99.476 -78.146 45.594
158290 99.506 -78.166 45.584
158291 99.536 -78.185 45.573
158292 99.566 -78.205 45.562
158293 99.595 -78.224 45.551
158294 99.624 -78.243 45.541
158295 99.653 -78.266 45.529
158296 99.687 -78.284 45.518
158297 99.720 -78.302 45.507
158298 99.752 -78.320 45.494
158299 99.780 -78.339 45.482
158300 99.808 -78.359 45.469
158301 99.836 -78.379 45.455
158302 99.858 -78.395 45.442
158303 99.868 -78.412 45.430
158304 99.860 -78.433 45.417
158305 99.853 -78.454 45.403
158306 99.889 -78.475 45.389
158307 99.925 -78.496 45.376
158308 99.962 -78.517 45.363
158309 99.998 -78.537 45.349
158310 100.034 -78.558 45.334
158311 100.071 -78.579 45.319
158312 100.107 -78.599 45.303
158313 100.143 -78.620 45.287
158314 100.180 -78.641 45.270
158315 100.216 -78.661 45.254
158316 100.252 -78.683 45.237
158317 100.288 -78.703 45.219
158318 100.325 -78.724 45.202
158319 100.359 -78.740 45.184
158320 100.371 -78.743 45.167
158321 100.370 -78.737 45.150
158322 100.389 -78.740 45.133
158323 100.395 -78.726 45.116
158324 100.399 -78.711 45.099
158325 100.404 -78.695 45.083
158326 100.409 -78.680 45.066
158327 100.438 -78.708 45.049
158328 100.486 -78.720 45.033
158329 100.529 -78.731 45.018
158330 100.548 -78.741 45.003
158331 100.567 -78.750 44.988
158332 100.587 -78.759 44.972
158333 100.609 -78.766 44.956
158334 100.630 -78.773 44.939
158335 100.650 -78.780 44.924
158336 100.668 -78.792 44.908
158337 100.674 -78.818 44.891
158338 100.701 -78.830 44.875
158339 100.729 -78.841 44.858
158340 100.749 -78.857 44.842
158341 100.767 -78.874 44.826
158342 100.785 -78.892 44.809
158343 100.804 -78.908 44.793
158344 100.843 -78.916 44.777
158345 100.882 -78.924 44.760
158346 100.928 -78.928 44.743
158347 100.975 -78.930 44.727
158348 101.022 -78.933 44.711
158349 101.070 -78.937 44.695
158350 101.117 -78.939 44.680
158351 101.165 -78.942 44.665
158352 101.203 -78.949 44.650
158353 101.236 -78.950 44.634
158354 101.260 -78.963 44.621
158355 101.284 -78.976 44.607
158356 101.309 -78.990 44.594
158357 101.333 -79.003 44.581
158358 101.358 -79.017 44.567
158359 101.382 -79.031 44.554
158360 101.399 -79.065 44.542
158361 101.415 -79.100 44.530
158362 101.448 -79.124 44.517
158363 101.490 -79.143 44.505
158588 92.476 -77.532 54.114
158589 92.496 -77.539 54.122
158590 92.516 -77.546 54.131
158591 92.536 -77.554 54.138
158592 92.556 -77.561 54.146
158593 92.576 -77.568 54.154
158594 92.596 -77.576 54.162
158595 92.616 -77.583 54.171
158596 92.636 -77.590 54.180
158597 92.656 -77.598 54.188
158598 92.675 -77.605 54.195
158599 92.696 -77.612 54.202
158600 92.716 -77.620 54.209
158601 92.735 -77.627 54.217
158602 92.756 -77.634 54.223
158603 92.776 -77.642 54.228
158604 92.795 -77.649 54.232
158605 92.815 -77.657 54.238
158606 92.840 -77.655 54.243
158607 92.867 -77.644 54.247
158608 92.887 -77.637 54.252
158609 92.901 -77.630 54.259
158610 92.915 -77.622 54.265
158611 92.929 -77.614 54.271
158612 92.943 -77.607 54.277
158613 92.958 -77.599 54.281
158614 92.972 -77.592 54.284
158615 92.986 -77.584 54.288
158616 92.998 -77.578 54.291
158617 93.007 -77.574 54.293
158618 93.016 -77.570 54.295
158619 93.025 -77.567 54.297
158620 93.035 -77.563 54.298
158621 93.046 -77.560 54.298
158622 93.058 -77.556 54.299
158623 93.070 -77.553 54.300
158624 93.081 -77.550 54.304
158625 93.092 -77.538 54.308
158626 93.103 -77.521 54.312
158627 93.114 -77.504 54.316
158628 93.124 -77.487 54.320
158629 93.135 -77.470 54.323
158630 93.150 -77.457 54.322
158631 93.174 -77.449 54.322
158632 93.199 -77.441 54.323
158633 93.223 -77.432 54.323
158634 93.248 -77.424 54.322
158635 93.272 -77.417 54.318
158636 93.296 -77.408 54.312
158637 93.317 -77.400 54.306
158638 93.331 -77.391 54.300
158639 93.344 -77.382 54.297
158640 93.357 -77.373 54.294
158641 93.371 -77.363 54.290
158642 93.384 -77.354 54.281
158643 93.393 -77.338 54.270
158644 93.401 -77.323 54.258
158645 93.411 -77.308 54.248
158646 93.429 -77.302 54.239
158647 93.446 -77.295 54.232
158648 93.465 -77.289 54.223
158649 93.483 -77.282 54.212
158650 93.501 -77.276 54.201
158651 93.519 -77.269 54.190
158652 93.541 -77.260 54.178
158653 93.562 -77.251 54.165
158654 93.583 -77.241 54.152
158655 93.604 -77.232 54.140
158656 93.625 -77.223 54.127
158657 93.646 -77.214 54.113
158658 93.667 -77.205 54.097
158659 93.688 -77.196 54.082
158660 93.709 -77.187 54.066
158661 93.730 -77.178 54.051
158662 93.748 -77.174 54.037
158663 93.763 -77.177 54.019
158664 93.776 -77.170 54.000
158665 93.789 -77.163 53.981
158666 93.802 -77.156 53.961
158667 93.803 -77.153 53.941
158668 93.805 -77.149 53.921
158669 93.806 -77.146 53.902
158670 93.808 -77.143 53.885
158671 93.810 -77.138 53.869
158672 93.812 -77.133 53.854
158673 93.815 -77.133 53.840
158674 93.835 -77.137 53.824
158675 93.864 -77.135 53.807
158676 93.893 -77.134 53.789
158677 93.913 -77.133 53.769
158678 93.922 -77.132 53.744
158679 93.932 -77.130 53.721
158680 93.947 -77.122 53.700
158681 93.963 -77.111 53.682
158682 93.982 -77.104 53.664
158683 94.006 -77.102 53.645
158684 94.030 -77.098 53.626
158685 94.053 -77.094 53.608
158686 94.077 -77.090 53.591
158687 94.101 -77.085 53.574
158688 94.124 -77.081 53.558
158689 94.154 -77.080 53.541
158690 94.186 -77.079 53.523
158691 94.216 -77.078 53.503
158692 94.221 -77.066 53.487
158693 94.226 -77.054 53.473
158694 94.237 -77.049 53.462
158695 94.249 -77.046 53.453
158696 94.262 -77.043 53.445
158697 94.275 -77.041 53.437
158698 94.290 -77.038 53.429
158699 94.306 -77.036 53.422
158700 94.321 -77.034 53.419
158701 94.336 -77.032 53.416
158702 94.346 -77.032 53.414
158703 94.347 -77.028 53.413
158704 94.346 -77.027 53.411
158705 94.344 -77.025 53.409
158706 94.342 -77.023 53.408
158707 94.341 -77.021 53.405
158708 94.339 -77.019 53.399
158709 94.348 -77.016 53.390
158710 94.367 -77.012 53.377
158711 94.385 -77.008 53.357
158712 94.403 -77.003 53.335
158713 94.422 -77.000 53.310
158714 94.440 -76.995 53.285
158715 94.459 -76.991 53.260
158716 94.480 -76.987 53.235
158717 94.501 -76.983 53.210
158718 94.522 -76.978 53.183
158719 94.545 -76.975 53.157
158720 94.573 -76.971 53.131
158721 94.601 -76.967 53.109
158722 94.620 -76.962 53.085
158723 94.627 -76.957 53.060
158724 94.635 -76.952 53.034
158725 94.643 -76.947 53.002
158726 94.654 -76.944 52.970
158727 94.666 -76.944 52.939
158728 94.677 -76.943 52.908
158729 94.689 -76.943 52.875
158730 94.700 -76.943 52.841
158731 94.712 -76.943 52.805
158732 94.730 -76.947 52.767
158733 94.770 -76.953 52.725
158734 94.810 -76.959 52.684
158735 94.850 -76.966 52.644
158736 94.890 -76.973 52.605
158737 94.923 -76.978 52.569
158738 94.931 -76.981 52.534
158739 94.939 -76.984 52.500
158740 94.948 -76.987 52.469
158741 94.956 -76.990 52.443
158742 94.964 -76.992 52.419
158743 94.979 -76.995 52.395
158744 94.999 -77.000 52.368
158745 95.019 -77.003 52.341
158746 95.039 -77.008 52.315
158747 95.059 -77.012 52.294
158748 95.079 -77.017 52.267
158749 95.100 -77.022 52.235
158750 95.120 -77.025 52.199
158751 95.139 -77.030 52.159
158752 95.160 -77.034 52.115
158753 95.180 -77.039 52.067
158754 95.200 -77.044 52.014
158755 95.220 -77.047 51.960
158756 95.240 -77.052 51.906
158757 95.257 -77.054 51.851
158758 95.268 -77.052 51.796
158759 95.278 -77.050 51.734
158760 95.295 -77.059 51.672
158761 95.313 -77.073 51.612
158762 95.331 -77.086 51.555
158763 95.349 -77.100 51.499
158764 95.367 -77.113 51.444
158765 95.385 -77.126 51.391
158766 95.403 -77.140 51.340
158767 95.419 -77.151 51.289
158768 95.433 -77.160 51.239
158769 95.447 -77.170 51.194
158770 95.461 -77.179 51.153
158771 95.477 -77.191 51.111
158772 95.493 -77.204 51.067
158773 95.509 -77.214 51.018
158774 95.523 -77.225 50.968
158775 95.538 -77.235 50.919
158776 95.553 -77.246 50.870
158777 95.568 -77.257 50.822
158778 95.583 -77.268 50.776
158779 95.598 -77.279 50.732
158780 95.613 -77.289 50.689
158781 95.632 -77.290 50.646
158782 95.652 -77.295 50.602
158783 95.672 -77.301 50.559
158784 95.691 -77.308 50.516
158785 95.711 -77.315 50.473
158786 95.731 -77.322 50.431
158787 95.750 -77.328 50.392
158788 95.757 -77.334 50.353
158789 95.764 -77.340 50.314
158790 95.784 -77.351 50.275
158791 95.804 -77.361 50.235
158792 95.825 -77.372 50.193
158793 95.845 -77.382 50.150
158794 95.866 -77.393 50.107
158795 95.886 -77.404 50.060
158796 95.912 -77.414 50.012
158797 95.946 -77.425 49.962
158798 95.965 -77.431 49.910
158799 95.979 -77.436 49.857
158800 95.992 -77.441 49.806
158801 96.006 -77.445 49.757
158802 96.019 -77.449 49.708
158803 96.013 -77.445 49.661
158804 96.016 -77.449 49.614
158805 96.021 -77.454 49.567
158806 96.026 -77.460 49.521
158807 96.031 -77.464 49.478
158808 96.035 -77.470 49.436
158809 96.040 -77.475 49.397
158810 96.045 -77.479 49.358
158811 96.052 -77.473 49.320
158812 96.058 -77.467 49.281
158813 96.065 -77.460 49.245
158814 96.073 -77.453 49.213
158815 96.080 -77.446 49.182
158816 96.089 -77.447 49.153
158817 96.101 -77.457 49.123
158818 96.113 -77.468 49.094
158819 96.132 -77.478 49.067
158820 96.153 -77.487 49.040
158821 96.170 -77.497 49.018
158822 96.180 -77.509 48.998
158823 96.189 -77.521 48.980
158824 96.198 -77.533 48.964
158825 96.203 -77.542 48.950
158826 96.217 -77.553 48.936
158827 96.227 -77.557 48.919
158828 96.238 -77.561 48.897
158829 96.249 -77.564 48.874
158830 96.260 -77.569 48.849
158831 96.271 -77.574 48.825
158832 96.286 -77.580 48.805
158833 96.315 -77.591 48.783
158834 96.343 -77.602 48.760
158835 96.372 -77.613 48.740
158836 96.401 -77.624 48.719
158837 96.429 -77.634 48.699
158838 96.458 -77.645 48.680
158839 96.486 -77.656 48.661
158840 96.515 -77.667 48.643
158841 96.543 -77.677 48.625
158842 96.566 -77.686 48.608
158843 96.581 -77.690 48.593
158844 96.588 -77.691 48.576
158845 96.595 -77.692 48.558
158846 96.602 -77.694 48.541
158847 96.609 -77.696 48.521
158848 96.615 -77.697 48.499
158849 96.622 -77.699 48.478
158850 96.629 -77.699 48.458
158851 96.636 -77.701 48.438
158852 96.643 -77.702 48.418
158853 96.650 -77.704 48.399
158854 96.657 -77.705 48.380
158855 96.664 -77.707 48.361
158856 96.670 -77.708 48.343
158857 96.678 -77.711 48.325
158858 96.688 -77.718 48.304
158859 96.697 -77.724 48.283
158860 96.711 -77.728 48.259
158861 96.725 -77.730 48.234
158862 96.739 -77.733 48.208
158863 96.753 -77.736 48.182
158864 96.767 -77.738 48.156
158865 96.782 -77.740 48.130
158866 96.796 -77.743 48.104
158867 96.810 -77.745 48.078
158868 96.824 -77.747 48.053
158869 96.838 -77.750 48.027
158870 96.853 -77.752 48.001
158871 96.867 -77.755 47.975
158872 96.881 -77.751 47.950
158873 96.903 -77.758 47.925
158874 96.927 -77.759 47.901
158875 96.950 -77.760 47.877
158876 96.973 -77.761 47.856
158877 96.996 -77.762 47.835
158878 97.020 -77.763 47.814
158879 97.033 -77.767 47.792
158880 97.044 -77.770 47.771
158881 97.055 -77.774 47.751
158882 97.066 -77.777 47.731
158883 97.077 -77.780 47.710
158884 97.090 -77.772 47.688
158885 97.103 -77.764 47.667
158886 97.117 -77.755 47.645
158887 97.130 -77.746 47.623
158888 97.144 -77.738 47.603
158889 97.157 -77.730 47.583
158890 97.179 -77.725 47.562
158891 97.204 -77.727 47.542
158892 97.229 -77.736 47.521
158893 97.260 -77.746 47.501
158894 97.290 -77.754 47.482
158895 97.320 -77.762 47.462
158896 97.351 -77.771 47.443
158897 97.381 -77.780 47.424
158898 97.411 -77.788 47.406
158899 97.441 -77.796 47.387
158900 97.471 -77.805 47.368
158901 97.502 -77.814 47.348
158902 97.532 -77.822 47.328
158903 97.548 -77.821 47.307
158904 97.558 -77.812 47.286
158905 97.567 -77.804 47.265
158906 97.579 -77.809 47.244
158907 97.595 -77.824 47.223
158908 97.610 -77.840 47.200
158909 97.626 -77.855 47.179
158910 97.641 -77.870 47.157
158911 97.657 -77.885 47.132
158912 97.669 -77.902 47.108
158913 97.666 -77.931 47.084
158914 97.663 -77.958 47.059
158915 97.661 -77.985 47.034
158916 97.658 -78.013 47.008
158917 97.655 -78.040 46.981
158918 97.671 -78.052 46.955
158919 97.682 -78.051 46.929
158920 97.702 -78.028 46.902
158921 97.721 -78.029 46.878
158922 97.754 -78.016 46.855
158923 97.790 -78.001 46.832
158924 97.825 -77.986 46.809
158925 97.860 -77.971 46.786
158926 97.895 -77.956 46.764
158927 97.931 -77.940 46.742
158928 97.951 -77.940 46.720
158929 97.964 -77.946 46.698
158930 97.977 -77.953 46.677
158931 97.972 -77.956 46.655
158932 97.967 -77.960 46.634
158933 97.962 -77.964 46.613
158934 97.966 -77.973 46.592
158935 97.977 -77.979 46.571
158936 97.968 -77.949 46.551
158937 98.051 -77.927 46.530
158938 98.134 -77.904 46.509
158939 98.216 -77.881 46.487
158940 98.299 -77.858 46.464
158941 98.381 -77.834 46.442
158942 98.461 -77.812 46.419
158943 98.498 -77.797 46.396
158944 98.534 -77.783 46.373
158945 98.571 -77.768 46.350
158946 98.608 -77.754 46.327
158947 98.637 -77.757 46.304
158948 98.649 -77.793 46.282
158949 98.662 -77.828 46.260
158950 98.675 -77.865 46.239
158951 98.688 -77.900 46.218
158952 98.701 -77.936 46.199
158953 98.713 -77.972 46.179
158954 98.715 -77.987 46.158
158955 98.716 -77.993 46.137
158956 98.717 -78.000 46.119
158957 98.718 -78.008 46.099
158958 98.719 -78.015 46.080
158959 98.719 -78.022 46.062
158960 98.720 -78.029 46.042
158961 98.721 -78.037 46.022
158962 98.722 -78.044 46.003
158963 98.722 -78.051 45.983
158964 98.723 -78.058 45.963
158965 98.724 -78.065 45.943
158966 98.725 -78.072 45.924
158967 98.770 -78.080 45.908
158968 98.822 -78.087 45.892
158969 98.875 -78.096 45.875
158970 98.928 -78.103 45.857
158971 98.981 -78.112 45.840
158972 98.997 -78.114 45.825
158973 99.002 -78.116 45.810
158974 99.026 -78.116 45.795
158975 99.081 -78.116 45.781
158976 99.136 -78.117 45.766
158977 99.191 -78.117 45.751
158978 99.238 -78.117 45.739
158979 99.256 -78.114 45.727
158980 99.275 -78.110 45.715
158981 99.293 -78.107 45.704
158982 99.311 -78.104 45.693
158983 99.330 -78.101 45.682
158984 99.349 -78.097 45.671
158985 99.367 -78.094 45.659
158986 99.385 -78.091 45.647
158987 99.404 -78.088 45.635
158988 99.424 -78.087 45.623
158989 99.453 -78.106 45.613
158990 99.484 -78.125 45.603
158991 99.514 -78.144 45.593
158992 99.544 -78.164 45.582
158993 99.574 -78.184 45.571
158994 99.604 -78.203 45.560
158995 99.634 -78.223 45.549
158996 99.662 -78.242 45.538
158997 99.690 -78.262 45.526
158998 99.718 -78.282 45.514
158999 99.746 -78.300 45.501
159000 99.774 -78.320 45.489
159001 99.801 -78.340 45.476
159002 99.828 -78.359 45.463
159003 99.845 -78.374 45.450
159004 99.862 -78.388 45.437
159005 99.863 -78.407 45.424
159006 99.855 -78.429 45.411
159007 99.874 -78.450 45.397
159008 99.910 -78.470 45.384
159009 99.946 -78.491 45.370
159010 99.983 -78.511 45.357
159011 100.019 -78.532 45.342
159012 100.056 -78.553 45.327
159013 100.092 -78.573 45.311
159014 100.128 -78.595 45.294
159015 100.164 -78.615 45.277
159016 100.201 -78.636 45.260
159017 100.237 -78.657 45.243
159018 100.273 -78.677 45.225
159019 100.310 -78.698 45.208
159020 100.346 -78.719 45.190
159021 100.343 -78.724 45.173
159022 100.341 -78.713 45.155
159023 100.352 -78.698 45.138
159024 100.361 -78.683 45.121
159025 100.365 -78.667 45.104
159026 100.370 -78.651 45.088
159027 100.388 -78.661 45.071
159028 100.418 -78.689 45.055
159029 100.465 -78.701 45.040
159030 100.512 -78.713 45.025
159031 100.559 -78.724 45.009
159032 100.579 -78.734 44.995
159033 100.598 -78.744 44.979
159034 100.617 -78.753 44.963
159035 100.636 -78.763 44.947
159036 100.657 -78.771 44.932
159037 100.678 -78.777 44.915
159038 100.694 -78.796 44.899
159039 100.716 -78.811 44.883
159040 100.735 -78.828 44.866
159041 100.752 -78.846 44.850
159042 100.770 -78.863 44.834
159043 100.788 -78.880 44.818
159044 100.814 -78.894 44.801
159045 100.853 -78.902 44.785
159046 100.893 -78.910 44.768
159047 100.934 -78.916 44.751
159048 100.981 -78.919 44.735
159049 101.029 -78.922 44.719
159050 101.076 -78.924 44.703
159051 101.123 -78.927 44.688
159052 101.171 -78.930 44.673
159053 101.200 -78.943 44.657
159054 101.237 -78.949 44.642
159055 101.259 -78.948 44.629
159056 101.283 -78.962 44.615
159057 101.308 -78.974 44.602
159058 101.332 -78.988 44.588
159059 101.357 -79.002 44.575
159060 101.380 -79.018 44.563
159061 101.397 -79.052 44.550
159062 101.413 -79.086 44.538
159063 101.448 -79.110 44.525
159064 101.490 -79.129 44.512
159289 92.465 -77.511 54.097
159290 92.485 -77.519 54.106
159291 92.505 -77.526 54.116
159292 92.525 -77.533 54.125
159293 92.545 -77.541 54.135
159294 92.565 -77.548 54.144
159295 92.585 -77.556 54.153
159296 92.605 -77.563 54.163
159297 92.625 -77.570 54.174
159298 92.645 -77.578 54.183
159299 92.665 -77.585 54.192
159300 92.685 -77.592 54.201
159301 92.705 -77.600 54.209
159302 92.724 -77.607 54.219
159303 92.744 -77.614 54.227
159304 92.765 -77.622 54.232
159305 92.784 -77.629 54.237
159306 92.804 -77.636 54.243
159307 92.829 -77.634 54.248
159308 92.855 -77.627 54.254
159309 92.878 -77.623 54.259
159310 92.892 -77.614 54.266
159311 92.907 -77.606 54.272
159312 92.921 -77.599 54.278
159313 92.935 -77.591 54.284
159314 92.949 -77.583 54.288
159315 92.963 -77.576 54.292
159316 92.977 -77.569 54.296
159317 92.986 -77.565 54.299
159318 92.994 -77.561 54.300
159319 93.003 -77.558 54.302
159320 93.012 -77.554 54.303
159321 93.021 -77.549 54.304
159322 93.031 -77.546 54.305
159323 93.043 -77.542 54.307
159324 93.055 -77.539 54.309
159325 93.067 -77.536 54.314
159326 93.077 -77.522 54.320
159327 93.088 -77.504 54.325
159328 93.098 -77.488 54.330
159329 93.110 -77.471 54.335
159330 93.123 -77.454 54.338
159331 93.140 -77.438 54.340
159332 93.163 -77.429 54.341
159333 93.191 -77.426 54.342
159334 93.215 -77.418 54.342
159335 93.239 -77.410 54.342
159336 93.264 -77.402 54.339
159337 93.288 -77.394 54.334
159338 93.312 -77.385 54.329
159339 93.327 -77.376 54.324
159340 93.340 -77.367 54.319
159341 93.353 -77.358 54.314
159342 93.365 -77.347 54.308
159343 93.373 -77.331 54.300
159344 93.381 -77.315 54.291
159345 93.390 -77.299 54.279
159346 93.404 -77.289 54.267
159347 93.422 -77.283 54.258
159348 93.440 -77.276 54.251
159349 93.458 -77.270 54.244
159350 93.476 -77.265 54.234
159351 93.494 -77.260 54.223
159352 93.514 -77.251 54.212
159353 93.537 -77.238 54.199
159354 93.560 -77.226 54.186
159355 93.581 -77.217 54.173
159356 93.602 -77.208 54.162
159357 93.623 -77.199 54.150
159358 93.644 -77.190 54.136
159359 93.665 -77.181 54.121
159360 93.686 -77.172 54.106
159361 93.707 -77.163 54.090
159362 93.727 -77.155 54.074
159363 93.742 -77.157 54.060
159364 93.756 -77.160 54.040
159365 93.770 -77.159 54.021
159366 93.783 -77.151 54.002
159367 93.795 -77.144 53.983
159368 93.796 -77.141 53.965
159369 93.798 -77.138 53.946
159370 93.799 -77.135 53.928
159371 93.801 -77.132 53.912
159372 93.802 -77.128 53.896
159373 93.804 -77.125 53.882
159374 93.809 -77.122 53.869
159375 93.832 -77.114 53.854
159376 93.861 -77.110 53.838
159377 93.891 -77.109 53.821
159378 93.916 -77.107 53.802
159379 93.925 -77.106 53.779
159380 93.934 -77.105 53.756
159381 93.944 -77.103 53.735
159382 93.960 -77.094 53.716
159383 93.979 -77.087 53.698
159384 94.002 -77.084 53.678
159385 94.026 -77.080 53.658
159386 94.050 -77.075 53.640
159387 94.073 -77.071 53.623
159388 94.097 -77.067 53.605
159389 94.121 -77.063 53.588
159390 94.145 -77.059 53.570
159391 94.177 -77.058 53.550
159392 94.199 -77.053 53.529
159393 94.204 -77.041 53.509
159394 94.216 -77.038 53.492
159395 94.228 -77.035 53.477
159396 94.240 -77.032 53.462
159397 94.253 -77.029 53.452
159398 94.267 -77.027 53.442
159399 94.282 -77.025 53.433
159400 94.297 -77.022 53.426
159401 94.312 -77.021 53.425
159402 94.327 -77.019 53.425
159403 94.343 -77.016 53.424
159404 94.348 -77.015 53.422
159405 94.346 -77.013 53.417
159406 94.345 -77.011 53.413
159407 94.343 -77.009 53.411
159408 94.342 -77.007 53.408
159409 94.340 -77.006 53.400
159410 94.354 -77.002 53.390
159411 94.372 -76.998 53.377
159412 94.391 -76.994 53.358
159413 94.409 -76.990 53.337
159414 94.427 -76.985 53.313
159415 94.445 -76.981 53.289
159416 94.464 -76.977 53.266
159417 94.482 -76.972 53.241
159418 94.502 -76.969 53.217
159419 94.524 -76.964 53.190
159420 94.551 -76.960 53.162
159421 94.579 -76.956 53.136
159422 94.608 -76.953 53.113
159423 94.620 -76.948 53.088
159424 94.628 -76.943 53.062
159425 94.636 -76.937 53.035
159426 94.644 -76.931 53.003
159427 94.653 -76.928 52.969
159428 94.665 -76.928 52.935
159429 94.677 -76.928 52.901
159430 94.688 -76.928 52.868
159431 94.700 -76.927 52.835
159432 94.712 -76.927 52.801
159433 94.731 -76.928 52.764
159434 94.771 -76.935 52.724
159435 94.812 -76.942 52.684
159436 94.852 -76.948 52.646
159437 94.892 -76.955 52.611
159438 94.932 -76.962 52.579
159439 94.943 -76.963 52.546
159440 94.948 -76.965 52.512
159441 94.954 -76.968 52.483
159442 94.959 -76.970 52.459
159443 94.966 -76.971 52.437
159444 94.977 -76.969 52.416
159445 94.998 -76.976 52.391
159446 95.019 -76.982 52.364
159447 95.040 -76.988 52.338
159448 95.061 -76.994 52.318
159449 95.083 -77.000 52.293
159450 95.104 -77.006 52.263
159451 95.124 -77.012 52.229
159452 95.144 -77.016 52.189
159453 95.165 -77.021 52.144
159454 95.185 -77.025 52.095
159455 95.205 -77.029 52.043
159456 95.225 -77.034 51.988
159457 95.245 -77.038 51.932
159458 95.265 -77.043 51.874
159459 95.276 -77.041 51.816
159460 95.293 -77.050 51.755
159461 95.310 -77.063 51.694
159462 95.328 -77.076 51.636
159463 95.345 -77.089 51.579
159464 95.359 -77.098 51.521
159465 95.374 -77.108 51.464
159466 95.387 -77.117 51.408
159467 95.401 -77.126 51.355
159468 95.415 -77.136 51.303
159469 95.429 -77.145 51.255
159470 95.443 -77.154 51.210
159471 95.458 -77.165 51.168
159472 95.475 -77.176 51.125
159473 95.492 -77.191 51.078
159474 95.508 -77.204 51.030
159475 95.523 -77.214 50.982
159476 95.538 -77.225 50.934
159477 95.552 -77.235 50.886
159478 95.567 -77.246 50.837
159479 95.582 -77.257 50.791
159480 95.597 -77.267 50.747
159481 95.613 -77.276 50.703
159482 95.633 -77.274 50.659
159483 95.654 -77.273 50.616
159484 95.674 -77.279 50.573
159485 95.693 -77.285 50.529
159486 95.713 -77.292 50.486
159487 95.733 -77.299 50.445
159488 95.751 -77.305 50.406
159489 95.758 -77.311 50.367
159490 95.765 -77.317 50.328
159491 95.777 -77.326 50.288
159492 95.798 -77.336 50.247
159493 95.818 -77.347 50.205
159494 95.839 -77.357 50.162
159495 95.859 -77.368 50.118
159496 95.880 -77.379 50.069
159497 95.909 -77.389 50.020
159498 95.942 -77.400 49.970
159499 95.965 -77.407 49.919
159500 95.979 -77.412 49.867
159501 95.992 -77.417 49.816
159502 96.006 -77.421 49.768
159503 96.011 -77.422 49.720
159504 96.005 -77.418 49.672
159505 95.999 -77.414 49.624
159506 95.992 -77.410 49.578
159507 95.996 -77.414 49.530
159508 96.001 -77.420 49.486
159509 96.005 -77.425 49.445
159510 96.010 -77.429 49.404
159511 96.016 -77.427 49.364
159512 96.023 -77.420 49.325
159513 96.030 -77.414 49.287
159514 96.038 -77.410 49.250
159515 96.050 -77.420 49.217
159516 96.062 -77.431 49.186
159517 96.075 -77.442 49.156
159518 96.087 -77.452 49.126
159519 96.107 -77.461 49.097
159520 96.127 -77.471 49.069
159521 96.148 -77.480 49.043
159522 96.168 -77.490 49.020
159523 96.181 -77.501 49.000
159524 96.191 -77.513 48.982
159525 96.200 -77.525 48.966
159526 96.209 -77.537 48.954
159527 96.220 -77.544 48.939
159528 96.231 -77.548 48.920
159529 96.242 -77.551 48.898
159530 96.253 -77.555 48.875
159531 96.263 -77.558 48.851
159532 96.274 -77.562 48.827
159533 96.288 -77.568 48.809
159534 96.317 -77.579 48.790
159535 96.345 -77.589 48.769
159536 96.374 -77.601 48.748
159537 96.403 -77.611 48.729
159538 96.432 -77.622 48.711
159539 96.460 -77.633 48.693
159540 96.489 -77.644 48.674
159541 96.518 -77.655 48.656
159542 96.547 -77.664 48.638
159543 96.576 -77.674 48.623
159544 96.591 -77.682 48.608
159545 96.598 -77.683 48.592
159546 96.605 -77.685 48.575
159547 96.612 -77.686 48.557
159548 96.619 -77.687 48.537
159549 96.625 -77.689 48.516
159550 96.632 -77.690 48.495
159551 96.639 -77.692 48.474
159552 96.646 -77.693 48.456
159553 96.653 -77.695 48.438
159554 96.660 -77.696 48.420
159555 96.667 -77.697 48.400
159556 96.673 -77.699 48.381
159557 96.680 -77.700 48.361
159558 96.687 -77.702 48.343
159559 96.697 -77.708 48.323
159560 96.707 -77.714 48.302
159561 96.719 -77.718 48.279
159562 96.733 -77.721 48.254
159563 96.747 -77.723 48.228
159564 96.761 -77.725 48.202
159565 96.776 -77.728 48.175
159566 96.790 -77.730 48.150
159567 96.804 -77.733 48.124
159568 96.818 -77.735 48.097
159569 96.832 -77.737 48.071
159570 96.847 -77.740 48.045
159571 96.861 -77.743 48.019
159572 96.875 -77.745 47.994
159573 96.890 -77.747 47.969
159574 96.913 -77.749 47.945
159575 96.939 -77.750 47.921
159576 96.964 -77.752 47.896
159577 96.990 -77.752 47.874
159578 97.015 -77.754 47.854
159579 97.028 -77.758 47.834
159580 97.036 -77.764 47.813
159581 97.047 -77.768 47.792
159582 97.058 -77.771 47.772
159583 97.069 -77.774 47.752
159584 97.081 -77.772 47.730
159585 97.094 -77.763 47.709
159586 97.108 -77.755 47.687
159587 97.121 -77.746 47.665
159588 97.135 -77.738 47.644
159589 97.154 -77.732 47.623
159590 97.179 -77.729 47.603
159591 97.205 -77.726 47.582
159592 97.229 -77.730 47.560
159593 97.252 -77.740 47.539
159594 97.277 -77.749 47.518
159595 97.307 -77.758 47.498
159596 97.337 -77.766 47.479
159597 97.367 -77.774 47.460
159598 97.397 -77.783 47.441
159599 97.428 -77.792 47.421
159600 97.458 -77.800 47.402
159601 97.488 -77.808 47.382
159602 97.518 -77.818 47.362
159603 97.548 -77.826 47.342
159604 97.567 -77.825 47.321
159605 97.576 -77.818 47.299
159606 97.585 -77.809 47.278
159607 97.594 -77.801 47.257
159608 97.608 -77.812 47.235
159609 97.624 -77.827 47.213
159610 97.639 -77.843 47.192
159611 97.655 -77.858 47.169
159612 97.670 -77.873 47.145
159613 97.677 -77.894 47.121
159614 97.674 -77.921 47.096
159615 97.672 -77.949 47.070
159616 97.669 -77.977 47.045
159617 97.667 -78.003 47.018
159618 97.685 -78.013 46.992
159619 97.703 -78.023 46.966
159620 97.709 -78.013 46.940
159621 97.735 -78.000 46.914
159622 97.755 -78.000 46.889
159623 97.780 -77.996 46.866
159624 97.815 -77.981 46.842
159625 97.850 -77.966 46.819
159626 97.885 -77.951 46.796
159627 97.920 -77.936 46.773
159628 97.948 -77.928 46.751
159629 97.961 -77.935 46.728
159630 97.974 -77.941 46.705
159631 97.988 -77.948 46.684
159632 97.989 -77.953 46.662
159633 97.984 -77.956 46.641
159634 97.979 -77.960 46.620
159635 97.974 -77.964 46.599
159636 97.982 -77.962 46.578
159637 98.005 -77.969 46.557
159638 98.034 -77.956 46.537
159639 98.100 -77.920 46.515
159640 98.166 -77.884 46.493
159641 98.232 -77.849 46.471
159642 98.298 -77.813 46.449
159643 98.377 -77.787 46.426
159644 98.460 -77.765 46.403
159645 98.543 -77.741 46.379
159646 98.588 -77.725 46.356
159647 98.625 -77.711 46.333
159648 98.644 -77.733 46.310
159649 98.656 -77.769 46.289
159650 98.669 -77.805 46.269
159651 98.684 -77.843 46.248
159652 98.695 -77.896 46.227
159653 98.698 -77.938 46.210
159654 98.702 -77.980 46.190
159655 98.717 -77.987 46.169
159656 98.731 -77.993 46.147
159657 98.734 -77.990 46.128
159658 98.735 -77.996 46.108
159659 98.736 -78.003 46.090
159660 98.737 -78.011 46.072
159661 98.737 -78.018 46.052
159662 98.738 -78.025 46.032
159663 98.739 -78.032 46.012
159664 98.740 -78.040 45.991
159665 98.741 -78.047 45.970
159666 98.741 -78.053 45.951
159667 98.742 -78.061 45.934
159668 98.759 -78.069 45.916
159669 98.811 -78.076 45.900
159670 98.864 -78.084 45.883
159671 98.917 -78.092 45.866
159672 98.954 -78.097 45.849
159673 98.960 -78.099 45.833
159674 98.987 -78.100 45.819
159675 99.042 -78.100 45.803
159676 99.097 -78.100 45.788
159677 99.152 -78.101 45.773
159678 99.207 -78.101 45.758
159679 99.251 -78.100 45.746
159680 99.269 -78.097 45.734
159681 99.287 -78.094 45.723
159682 99.306 -78.091 45.712
159683 99.324 -78.087 45.701
159684 99.343 -78.084 45.690
159685 99.361 -78.081 45.678
159686 99.379 -78.078 45.666
159687 99.398 -78.075 45.654
159688 99.416 -78.072 45.643
159689 99.426 -78.078 45.632
159690 99.439 -78.086 45.622
159691 99.470 -78.106 45.611
159692 99.500 -78.125 45.601
159693 99.530 -78.145 45.590
159694 99.560 -78.165 45.580
159695 99.595 -78.184 45.569
159696 99.628 -78.203 45.557
159697 99.655 -78.223 45.546
159698 99.684 -78.242 45.533
159699 99.711 -78.262 45.521
159700 99.739 -78.282 45.508
159701 99.767 -78.301 45.496
159702 99.795 -78.320 45.483
159703 99.815 -78.337 45.471
159704 99.832 -78.352 45.459
159705 99.849 -78.366 45.446
159706 99.866 -78.382 45.433
159707 99.859 -78.403 45.420
159708 99.859 -78.424 45.406
159709 99.895 -78.445 45.392
159710 99.931 -78.465 45.378
159711 99.968 -78.486 45.363
159712 100.004 -78.507 45.348
159713 100.041 -78.527 45.333
159714 100.077 -78.548 45.317
159715 100.113 -78.569 45.301
159716 100.149 -78.589 45.283
159717 100.186 -78.610 45.265
159718 100.222 -78.631 45.247
159719 100.258 -78.651 45.230
159720 100.291 -78.671 45.213
159721 100.286 -78.680 45.195
159722 100.277 -78.680 45.178
159723 100.283 -78.667 45.161
159724 100.294 -78.651 45.144
159725 100.305 -78.636 45.127
159726 100.317 -78.621 45.111
159727 100.337 -78.617 45.094
159728 100.366 -78.643 45.078
159729 100.397 -78.670 45.063
159730 100.444 -78.682 45.047
159731 100.491 -78.694 45.032
159732 100.538 -78.706 45.016
159733 100.585 -78.717 45.002
159734 100.610 -78.727 44.986
159735 100.629 -78.737 44.971
159736 100.648 -78.746 44.955
159737 100.668 -78.756 44.939
159738 100.687 -78.765 44.922
159739 100.705 -78.779 44.906
159740 100.720 -78.799 44.891
159741 100.738 -78.817 44.874
159742 100.756 -78.834 44.859
159743 100.774 -78.852 44.843
159744 100.791 -78.869 44.826
159745 100.824 -78.880 44.809
159746 100.863 -78.887 44.793
159747 100.903 -78.895 44.777
159748 100.943 -78.903 44.760
159749 100.987 -78.908 44.743
159750 101.035 -78.910 44.727
159751 101.082 -78.913 44.711
159752 101.129 -78.915 44.696
159753 101.162 -78.925 44.680
159754 101.196 -78.937 44.665
159755 101.235 -78.946 44.650
159756 101.258 -78.933 44.637
159757 101.282 -78.946 44.623
159758 101.307 -78.960 44.609
159759 101.331 -78.973 44.596
159760 101.356 -78.987 44.583
159761 101.379 -79.005 44.570
159762 101.395 -79.039 44.557
159763 101.412 -79.073 44.545
159764 101.448 -79.096 44.532
159765 101.490 -79.115 44.520
159766 101.532 -79.135 44.506
159990 92.454 -77.492 54.079
159991 92.474 -77.499 54.089
159992 92.494 -77.506 54.100
159993 92.514 -77.514 54.111
159994 92.534 -77.521 54.122
159995 92.554 -77.528 54.133
159996 92.574 -77.536 54.144
159997 92.594 -77.543 54.155
159998 92.614 -77.550 54.166
159999 92.634 -77.558 54.176
160000 92.654 -77.565 54.187
160001 92.673 -77.573 54.198
160002 92.694 -77.580 54.208
160003 92.714 -77.587 54.219
160004 92.733 -77.595 54.228
160005 92.754 -77.602 54.235
160006 92.774 -77.609 54.241
160007 92.793 -77.617 54.247
160008 92.817 -77.614 54.253
160009 92.842 -77.611 54.259
160010 92.867 -77.607 54.265
160011 92.883 -77.599 54.272
160012 92.898 -77.590 54.278
160013 92.912 -77.583 54.284
160014 92.926 -77.575 54.290
160015 92.940 -77.567 54.295
160016 92.954 -77.560 54.299
160017 92.964 -77.556 54.303
160018 92.973 -77.551 54.306
160019 92.982 -77.548 54.308
160020 92.991 -77.544 54.309
160021 93.000 -77.540 54.309
160022 93.009 -77.536 54.310
160023 93.018 -77.533 54.312
160024 93.028 -77.529 54.316
160025 93.040 -77.526 54.320
160026 93.052 -77.522 54.325
160027 93.062 -77.505 54.331
160028 93.073 -77.489 54.338
160029 93.083 -77.471 54.344
160030 93.100 -77.455 54.349
160031 93.117 -77.439 54.354
160032 93.134 -77.423 54.358
160033 93.151 -77.406 54.360
160034 93.169 -77.391 54.361
160035 93.199 -77.391 54.361
160036 93.228 -77.391 54.362
160037 93.256 -77.387 54.360
160038 93.280 -77.379 54.357
160039 93.304 -77.371 54.352
160040 93.323 -77.362 54.346
160041 93.336 -77.353 54.340
160042 93.346 -77.339 54.333
160043 93.354 -77.323 54.326
160044 93.362 -77.307 54.320
160045 93.370 -77.292 54.311
160046 93.379 -77.277 54.300
160047 93.398 -77.271 54.287
160048 93.415 -77.266 54.277
160049 93.433 -77.260 54.268
160050 93.451 -77.255 54.260
160051 93.469 -77.250 54.251
160052 93.486 -77.244 54.242
160053 93.509 -77.232 54.232
160054 93.532 -77.219 54.219
160055 93.556 -77.205 54.205
160056 93.579 -77.193 54.193
160057 93.600 -77.184 54.182
160058 93.621 -77.175 54.170
160059 93.642 -77.166 54.156
160060 93.663 -77.157 54.143
160061 93.684 -77.147 54.129
160062 93.705 -77.138 54.113
160063 93.721 -77.138 54.098
160064 93.735 -77.141 54.082
160065 93.750 -77.144 54.061
160066 93.764 -77.147 54.041
160067 93.777 -77.140 54.022
160068 93.788 -77.134 54.004
160069 93.789 -77.130 53.988
160070 93.791 -77.127 53.970
160071 93.792 -77.124 53.952
160072 93.794 -77.120 53.936
160073 93.798 -77.117 53.921
160074 93.804 -77.114 53.907
160075 93.809 -77.110 53.895
160076 93.827 -77.104 53.881
160077 93.857 -77.094 53.867
160078 93.888 -77.085 53.851
160079 93.917 -77.082 53.832
160080 93.927 -77.081 53.810
160081 93.937 -77.079 53.789
160082 93.946 -77.078 53.767
160083 93.956 -77.076 53.749
160084 93.976 -77.069 53.731
160085 93.999 -77.066 53.712
160086 94.023 -77.061 53.691
160087 94.047 -77.057 53.672
160088 94.070 -77.053 53.654
160089 94.094 -77.048 53.636
160090 94.118 -77.044 53.617
160091 94.141 -77.040 53.598
160092 94.168 -77.037 53.578
160093 94.182 -77.030 53.556
160094 94.194 -77.027 53.533
160095 94.207 -77.024 53.514
160096 94.219 -77.021 53.496
160097 94.231 -77.018 53.480
160098 94.244 -77.015 53.466
160099 94.258 -77.012 53.453
160100 94.273 -77.011 53.442
160101 94.288 -77.009 53.434
160102 94.303 -77.006 53.430
160103 94.319 -77.005 53.428
160104 94.334 -77.002 53.426
160105 94.344 -76.998 53.425
160106 94.346 -76.998 53.422
160107 94.345 -76.997 53.418
160108 94.344 -76.996 53.415
160109 94.342 -76.994 53.411
160110 94.341 -76.992 53.404
160111 94.359 -76.988 53.394
160112 94.378 -76.984 53.381
160113 94.396 -76.979 53.362
160114 94.414 -76.975 53.341
160115 94.432 -76.971 53.318
160116 94.451 -76.967 53.295
160117 94.469 -76.962 53.271
160118 94.488 -76.958 53.248
160119 94.505 -76.954 53.224
160120 94.526 -76.950 53.197
160121 94.553 -76.947 53.170
160122 94.581 -76.943 53.144
160123 94.608 -76.940 53.119
160124 94.618 -76.934 53.092
160125 94.627 -76.928 53.065
160126 94.636 -76.922 53.039
160127 94.645 -76.917 53.007
160128 94.653 -76.912 52.972
160129 94.666 -76.913 52.936
160130 94.679 -76.916 52.900
160131 94.692 -76.919 52.865
160132 94.705 -76.922 52.833
160133 94.718 -76.925 52.800
160134 94.754 -76.931 52.764
160135 94.794 -76.938 52.727
160136 94.835 -76.945 52.690
160137 94.875 -76.951 52.654
160138 94.909 -76.943 52.622
160139 94.931 -76.940 52.591
160140 94.937 -76.942 52.560
160141 94.942 -76.943 52.526
160142 94.948 -76.946 52.498
160143 94.955 -76.947 52.475
160144 94.965 -76.945 52.454
160145 94.976 -76.943 52.436
160146 94.994 -76.947 52.413
160147 95.015 -76.953 52.388
160148 95.036 -76.959 52.362
160149 95.057 -76.965 52.343
160150 95.078 -76.972 52.320
160151 95.100 -76.978 52.292
160152 95.121 -76.984 52.258
160153 95.142 -76.990 52.218
160154 95.164 -76.997 52.172
160155 95.185 -77.003 52.122
160156 95.206 -77.009 52.068
160157 95.227 -77.015 52.012
160158 95.248 -77.021 51.954
160159 95.269 -77.027 51.896
160160 95.286 -77.036 51.836
160161 95.300 -77.045 51.773
160162 95.314 -77.055 51.713
160163 95.327 -77.064 51.657
160164 95.341 -77.073 51.599
160165 95.355 -77.083 51.541
160166 95.369 -77.092 51.482
160167 95.383 -77.101 51.425
160168 95.397 -77.111 51.370
160169 95.410 -77.120 51.319
160170 95.424 -77.129 51.271
160171 95.438 -77.139 51.225
160172 95.455 -77.150 51.182
160173 95.472 -77.162 51.138
160174 95.489 -77.175 51.092
160175 95.507 -77.192 51.045
160176 95.522 -77.203 50.998
160177 95.537 -77.213 50.950
160178 95.552 -77.224 50.901
160179 95.567 -77.235 50.852
160180 95.582 -77.246 50.805
160181 95.597 -77.257 50.761
160182 95.615 -77.260 50.717
160183 95.635 -77.258 50.672
160184 95.656 -77.256 50.628
160185 95.676 -77.257 50.584
160186 95.695 -77.263 50.541
160187 95.716 -77.269 50.498
160188 95.735 -77.276 50.458
160189 95.753 -77.282 50.419
160190 95.760 -77.289 50.381
160191 95.767 -77.295 50.342
160192 95.774 -77.301 50.303
160193 95.792 -77.310 50.262
160194 95.812 -77.321 50.219
160195 95.833 -77.332 50.175
160196 95.853 -77.342 50.130
160197 95.874 -77.353 50.082
160198 95.905 -77.363 50.034
160199 95.938 -77.374 49.984
160200 95.965 -77.383 49.933
160201 95.978 -77.388 49.882
160202 95.992 -77.392 49.831
160203 96.006 -77.397 49.783
160204 96.002 -77.394 49.735
160205 95.996 -77.390 49.685
160206 95.989 -77.386 49.636
160207 95.983 -77.382 49.589
160208 95.977 -77.379 49.541
160209 95.971 -77.375 49.497
160210 95.975 -77.379 49.454
160211 95.981 -77.381 49.412
160212 95.988 -77.374 49.371
160213 96.000 -77.383 49.331
160214 96.012 -77.394 49.292
160215 96.024 -77.404 49.256
160216 96.036 -77.415 49.223
160217 96.048 -77.426 49.191
160218 96.061 -77.436 49.160
160219 96.082 -77.445 49.130
160220 96.102 -77.454 49.100
160221 96.122 -77.464 49.072
160222 96.143 -77.473 49.045
160223 96.163 -77.483 49.023
160224 96.183 -77.492 49.003
160225 96.192 -77.504 48.985
160226 96.201 -77.517 48.970
160227 96.211 -77.529 48.956
160228 96.222 -77.533 48.939
160229 96.233 -77.537 48.918
160230 96.244 -77.541 48.894
160231 96.255 -77.545 48.871
160232 96.266 -77.548 48.850
160233 96.277 -77.552 48.832
160234 96.292 -77.555 48.816
160235 96.321 -77.564 48.798
160236 96.350 -77.573 48.777
160237 96.380 -77.583 48.758
160238 96.409 -77.593 48.740
160239 96.439 -77.603 48.723
160240 96.468 -77.612 48.705
160241 96.498 -77.622 48.687
160242 96.527 -77.632 48.669
160243 96.555 -77.643 48.651
160244 96.572 -77.664 48.636
160245 96.588 -77.683 48.621
160246 96.600 -77.680 48.606
160247 96.612 -77.679 48.590
160248 96.622 -77.678 48.572
160249 96.629 -77.680 48.551
160250 96.636 -77.681 48.530
160251 96.643 -77.682 48.509
160252 96.649 -77.683 48.491
160253 96.656 -77.685 48.474
160254 96.663 -77.686 48.458
160255 96.670 -77.688 48.441
160256 96.677 -77.689 48.419
160257 96.683 -77.690 48.399
160258 96.690 -77.692 48.380
160259 96.697 -77.693 48.360
160260 96.705 -77.697 48.341
160261 96.715 -77.704 48.321
160262 96.727 -77.708 48.298
160263 96.741 -77.711 48.274
160264 96.755 -77.713 48.248
160265 96.770 -77.715 48.222
160266 96.784 -77.718 48.195
160267 96.798 -77.721 48.169
160268 96.812 -77.723 48.144
160269 96.826 -77.725 48.117
160270 96.841 -77.727 48.091
160271 96.855 -77.730 48.064
160272 96.869 -77.733 48.039
160273 96.891 -77.735 48.013
160274 96.912 -77.738 47.988
160275 96.934 -77.740 47.963
160276 96.960 -77.742 47.939
160277 96.985 -77.743 47.915
160278 97.011 -77.744 47.893
160279 97.028 -77.747 47.872
160280 97.034 -77.754 47.852
160281 97.040 -77.761 47.832
160282 97.050 -77.765 47.812
160283 97.061 -77.768 47.792
160284 97.072 -77.771 47.771
160285 97.086 -77.763 47.750
160286 97.099 -77.755 47.729
160287 97.112 -77.746 47.707
160288 97.129 -77.739 47.686
160289 97.154 -77.736 47.665
160290 97.179 -77.733 47.643
160291 97.205 -77.730 47.622
160292 97.230 -77.727 47.601
160293 97.254 -77.731 47.579
160294 97.278 -77.741 47.557
160295 97.301 -77.751 47.536
160296 97.325 -77.761 47.515
160297 97.354 -77.770 47.495
160298 97.384 -77.778 47.476
160299 97.414 -77.787 47.456
160300 97.444 -77.796 47.436
160301 97.474 -77.804 47.416
160302 97.505 -77.812 47.396
160303 97.535 -77.821 47.376
160304 97.565 -77.830 47.356
160305 97.586 -77.830 47.335
160306 97.595 -77.822 47.313
160307 97.603 -77.815 47.292
160308 97.612 -77.806 47.270
160309 97.621 -77.799 47.248
160310 97.637 -77.815 47.227
160311 97.653 -77.830 47.205
160312 97.668 -77.845 47.182
160313 97.684 -77.861 47.157
160314 97.685 -77.885 47.132
160315 97.682 -77.912 47.107
160316 97.680 -77.940 47.080
160317 97.681 -77.964 47.055
160318 97.699 -77.974 47.028
160319 97.717 -77.984 47.002
160320 97.732 -77.989 46.976
160321 97.749 -77.970 46.951
160322 97.770 -77.971 46.925
160323 97.789 -77.971 46.900
160324 97.809 -77.971 46.876
160325 97.840 -77.962 46.852
160326 97.875 -77.946 46.828
160327 97.910 -77.931 46.805
160328 97.945 -77.917 46.783
160329 97.958 -77.924 46.760
160330 97.971 -77.930 46.736
160331 97.984 -77.937 46.713
160332 97.997 -77.943 46.691
160333 98.006 -77.949 46.669
160334 98.001 -77.953 46.649
160335 97.996 -77.956 46.628
160336 97.991 -77.960 46.606
160337 98.026 -77.938 46.584
160338 98.054 -77.965 46.563
160339 98.080 -77.987 46.543
160340 98.112 -77.971 46.521
160341 98.178 -77.936 46.499
160342 98.244 -77.900 46.476
160343 98.310 -77.865 46.454
160344 98.377 -77.829 46.431
160345 98.443 -77.793 46.407
160346 98.509 -77.758 46.383
160347 98.575 -77.722 46.360
160348 98.624 -77.690 46.338
160349 98.641 -77.733 46.317
160350 98.644 -77.775 46.296
160351 98.647 -77.818 46.276
160352 98.650 -77.859 46.256
160353 98.653 -77.902 46.236
160354 98.656 -77.943 46.219
160355 98.670 -77.965 46.199
160356 98.684 -77.972 46.178
160357 98.698 -77.979 46.158
160358 98.712 -77.987 46.138
160359 98.727 -77.993 46.118
160360 98.741 -78.000 46.099
160361 98.754 -78.006 46.081
160362 98.755 -78.006 46.061
160363 98.756 -78.014 46.041
160364 98.757 -78.021 46.020
160365 98.758 -78.028 45.999
160366 98.758 -78.035 45.979
160367 98.759 -78.043 45.959
160368 98.760 -78.050 45.942
160369 98.761 -78.056 45.925
160370 98.800 -78.065 45.908
160371 98.852 -78.072 45.891
160372 98.905 -78.080 45.874
160373 98.917 -78.082 45.857
160374 98.947 -78.083 45.841
160375 99.002 -78.084 45.827
160376 99.058 -78.084 45.811
160377 99.113 -78.084 45.795
160378 99.168 -78.084 45.780
160379 99.223 -78.085 45.765
160380 99.263 -78.084 45.753
160381 99.281 -78.081 45.741
160382 99.300 -78.078 45.730
160383 99.319 -78.075 45.720
160384 99.337 -78.072 45.709
160385 99.355 -78.068 45.697
160386 99.374 -78.065 45.686
160387 99.392 -78.062 45.673
160388 99.408 -78.062 45.662
160389 99.418 -78.067 45.651
160390 99.428 -78.072 45.641
160391 99.438 -78.078 45.630
160392 99.452 -78.086 45.620
160393 99.483 -78.106 45.609
160394 99.512 -78.125 45.598
160395 99.544 -78.145 45.587
160396 99.583 -78.165 45.576
160397 99.621 -78.184 45.565
160398 99.649 -78.203 45.553
160399 99.677 -78.223 45.541
160400 99.705 -78.243 45.529
160401 99.733 -78.262 45.517
160402 99.761 -78.282 45.504
160403 99.785 -78.300 45.491
160404 99.802 -78.315 45.478
160405 99.819 -78.330 45.466
160406 99.836 -78.344 45.455
160407 99.853 -78.360 45.442
160408 99.862 -78.377 45.428
160409 99.854 -78.398 45.414
160410 99.880 -78.419 45.399
160411 99.917 -78.439 45.384
160412 99.953 -78.460 45.369
160413 99.989 -78.481 45.353
160414 100.025 -78.501 45.338
160415 100.062 -78.523 45.322
160416 100.098 -78.543 45.305
160417 100.134 -78.564 45.288
160418 100.171 -78.585 45.270
160419 100.207 -78.605 45.252
160420 100.232 -78.623 45.235
160421 100.226 -78.632 45.218
160422 100.221 -78.640 45.200
160423 100.215 -78.636 45.184
160424 100.226 -78.621 45.167
160425 100.237 -78.606 45.150
160426 100.249 -78.591 45.134
160427 100.275 -78.596 45.118
160428 100.310 -78.611 45.102
160429 100.344 -78.626 45.086
160430 100.375 -78.651 45.071
160431 100.423 -78.664 45.055
160432 100.470 -78.675 45.039
160433 100.517 -78.687 45.024
160434 100.565 -78.698 45.009
160435 100.612 -78.711 44.994
160436 100.641 -78.720 44.977
160437 100.660 -78.730 44.961
160438 100.679 -78.740 44.947
160439 100.700 -78.749 44.930
160440 100.721 -78.756 44.914
160441 100.739 -78.771 44.899
160442 100.753 -78.792 44.882
160443 100.768 -78.813 44.866
160444 100.782 -78.833 44.850
160445 100.797 -78.855 44.834
160446 100.834 -78.865 44.817
160447 100.874 -78.873 44.801
160448 100.913 -78.880 44.785
160449 100.953 -78.889 44.768
160450 100.993 -78.896 44.751
160451 101.040 -78.899 44.735
160452 101.088 -78.901 44.719
160453 101.135 -78.904 44.703
160454 101.155 -78.921 44.688
160455 101.193 -78.930 44.673
160456 101.232 -78.940 44.658
160457 101.259 -78.930 44.645
160458 101.281 -78.931 44.631
160459 101.306 -78.945 44.617
160460 101.330 -78.959 44.604
160461 101.355 -78.971 44.590
160462 101.377 -78.992 44.577
160463 101.394 -79.026 44.563
160464 101.410 -79.060 44.551
160465 101.447 -79.083 44.539
160466 101.489 -79.102 44.527
160467 101.532 -79.121 44.514
160468 101.574 -79.140 44.502
160691 92.443 -77.472 54.058
160692 92.463 -77.479 54.070
160693 92.483 -77.486 54.082
160694 92.503 -77.494 54.094
160695 92.523 -77.501 54.107
160696 92.543 -77.508 54.119
160697 92.563 -77.516 54.131
160698 92.583 -77.523 54.144
160699 92.602 -77.530 54.157
160700 92.623 -77.538 54.169
160701 92.643 -77.545 54.181
160702 92.662 -77.552 54.193
160703 92.683 -77.560 54.205
160704 92.703 -77.567 54.216
160705 92.722 -77.574 54.226
160706 92.742 -77.582 54.234
160707 92.763 -77.589 54.242
160708 92.782 -77.596 54.250
160709 92.805 -77.598 54.257
160710 92.830 -77.594 54.263
160711 92.854 -77.590 54.270
160712 92.874 -77.584 54.277
160713 92.889 -77.575 54.283
160714 92.903 -77.567 54.290
160715 92.917 -77.559 54.296
160716 92.931 -77.551 54.301
160717 92.942 -77.546 54.306
160718 92.951 -77.542 54.310
160719 92.960 -77.539 54.313
160720 92.969 -77.535 54.315
160721 92.978 -77.531 54.315
160722 92.987 -77.527 54.316
160723 92.996 -77.523 54.318
160724 93.005 -77.519 54.321
160725 93.014 -77.515 54.325
160726 93.025 -77.512 54.331
160727 93.037 -77.506 54.337
160728 93.047 -77.489 54.343
160729 93.060 -77.472 54.350
160730 93.077 -77.456 54.356
160731 93.094 -77.439 54.363
160732 93.111 -77.423 54.369
160733 93.128 -77.407 54.375
160734 93.145 -77.390 54.378
160735 93.162 -77.374 54.379
160736 93.179 -77.357 54.380
160737 93.205 -77.353 54.382
160738 93.234 -77.353 54.382
160739 93.264 -77.352 54.380
160740 93.294 -77.352 54.374
160741 93.318 -77.348 54.367
160742 93.326 -77.331 54.359
160743 93.331 -77.313 54.350
160744 93.337 -77.294 54.344
160745 93.342 -77.276 54.338
160746 93.355 -77.266 54.331
160747 93.373 -77.260 54.319
160748 93.391 -77.255 54.306
160749 93.408 -77.250 54.294
160750 93.426 -77.245 54.284
160751 93.443 -77.240 54.275
160752 93.461 -77.235 54.266
160753 93.481 -77.227 54.257
160754 93.504 -77.213 54.248
160755 93.528 -77.200 54.235
160756 93.551 -77.187 54.222
160757 93.574 -77.173 54.210
160758 93.597 -77.160 54.200
160759 93.619 -77.150 54.188
160760 93.640 -77.141 54.176
160761 93.661 -77.132 54.164
160762 93.682 -77.123 54.150
160763 93.701 -77.118 54.135
160764 93.715 -77.121 54.119
160765 93.729 -77.124 54.104
160766 93.743 -77.127 54.084
160767 93.757 -77.129 54.064
160768 93.771 -77.128 54.044
160769 93.781 -77.122 54.027
160770 93.783 -77.119 54.009
160771 93.784 -77.116 53.991
160772 93.787 -77.113 53.974
160773 93.793 -77.110 53.958
160774 93.798 -77.106 53.943
160775 93.804 -77.103 53.929
160776 93.810 -77.100 53.918
160777 93.821 -77.094 53.905
160778 93.852 -77.085 53.891
160779 93.883 -77.075 53.877
160780 93.913 -77.065 53.857
160781 93.928 -77.060 53.836
160782 93.939 -77.058 53.816
160783 93.948 -77.056 53.796
160784 93.958 -77.055 53.779
160785 93.971 -77.053 53.763
160786 93.994 -77.049 53.746
160787 94.018 -77.044 53.724
160788 94.042 -77.041 53.705
160789 94.065 -77.036 53.685
160790 94.089 -77.032 53.665
160791 94.113 -77.028 53.646
160792 94.136 -77.023 53.625
160793 94.160 -77.019 53.604
160794 94.173 -77.016 53.582
160795 94.185 -77.013 53.557
160796 94.198 -77.010 53.536
160797 94.210 -77.007 53.517
160798 94.222 -77.004 53.500
160799 94.235 -77.001 53.483
160800 94.249 -76.999 53.468
160801 94.264 -76.997 53.455
160802 94.279 -76.995 53.445
160803 94.294 -76.993 53.438
160804 94.310 -76.990 53.434
160805 94.325 -76.987 53.431
160806 94.337 -76.983 53.429
160807 94.342 -76.981 53.426
160808 94.343 -76.981 53.423
160809 94.344 -76.981 53.421
160810 94.343 -76.980 53.417
160811 94.346 -76.978 53.411
160812 94.364 -76.974 53.401
160813 94.383 -76.969 53.388
160814 94.401 -76.965 53.370
160815 94.419 -76.961 53.349
160816 94.436 -76.957 53.327
160817 94.453 -76.953 53.304
160818 94.470 -76.949 53.280
160819 94.487 -76.944 53.256
160820 94.504 -76.940 53.232
160821 94.519 -76.938 53.205
160822 94.544 -76.935 53.178
160823 94.572 -76.931 53.152
160824 94.598 -76.928 53.126
160825 94.607 -76.922 53.098
160826 94.616 -76.916 53.070
160827 94.625 -76.911 53.042
160828 94.635 -76.905 53.009
160829 94.649 -76.905 52.976
160830 94.671 -76.912 52.942
160831 94.687 -76.918 52.907
160832 94.700 -76.921 52.872
160833 94.713 -76.924 52.838
160834 94.737 -76.928 52.806
160835 94.777 -76.935 52.771
160836 94.817 -76.939 52.736
160837 94.851 -76.931 52.702
160838 94.886 -76.924 52.670
160839 94.920 -76.916 52.639
160840 94.925 -76.918 52.607
160841 94.931 -76.920 52.574
160842 94.936 -76.922 52.542
160843 94.945 -76.922 52.514
160844 94.955 -76.921 52.490
160845 94.965 -76.918 52.471
160846 94.976 -76.917 52.453
160847 94.990 -76.918 52.433
160848 95.011 -76.925 52.412
160849 95.032 -76.931 52.388
160850 95.053 -76.937 52.369
160851 95.074 -76.943 52.348
160852 95.096 -76.949 52.321
160853 95.117 -76.955 52.286
160854 95.138 -76.961 52.245
160855 95.160 -76.968 52.198
160856 95.181 -76.974 52.146
160857 95.202 -76.980 52.089
160858 95.223 -76.986 52.032
160859 95.244 -76.992 51.975
160860 95.271 -77.000 51.917
160861 95.281 -77.021 51.855
160862 95.295 -77.030 51.791
160863 95.309 -77.039 51.730
160864 95.323 -77.049 51.675
160865 95.337 -77.058 51.616
160866 95.350 -77.067 51.558
160867 95.365 -77.077 51.500
160868 95.378 -77.086 51.441
160869 95.392 -77.095 51.386
160870 95.406 -77.105 51.334
160871 95.420 -77.114 51.285
160872 95.435 -77.125 51.239
160873 95.452 -77.136 51.194
160874 95.469 -77.148 51.149
160875 95.486 -77.160 51.105
160876 95.504 -77.175 51.059
160877 95.522 -77.192 51.012
160878 95.537 -77.203 50.964
160879 95.551 -77.213 50.915
160880 95.566 -77.224 50.866
160881 95.581 -77.235 50.819
160882 95.596 -77.245 50.775
160883 95.616 -77.244 50.731
160884 95.636 -77.242 50.685
160885 95.657 -77.240 50.640
160886 95.678 -77.238 50.597
160887 95.698 -77.241 50.554
160888 95.718 -77.247 50.512
160889 95.737 -77.254 50.471
160890 95.754 -77.260 50.432
160891 95.761 -77.266 50.395
160892 95.768 -77.273 50.357
160893 95.775 -77.279 50.318
160894 95.785 -77.285 50.276
160895 95.806 -77.296 50.232
160896 95.826 -77.307 50.187
160897 95.847 -77.317 50.142
160898 95.869 -77.328 50.095
160899 95.902 -77.338 50.047
160900 95.935 -77.349 49.999
160901 95.965 -77.359 49.948
160902 95.978 -77.363 49.897
160903 95.992 -77.368 49.847
160904 96.000 -77.370 49.798
160905 95.993 -77.366 49.748
160906 95.987 -77.362 49.698
160907 95.981 -77.358 49.648
160908 95.975 -77.354 49.600
160909 95.969 -77.351 49.552
160910 95.962 -77.347 49.506
160911 95.956 -77.348 49.463
160912 95.961 -77.357 49.420
160913 95.973 -77.367 49.378
160914 95.985 -77.378 49.337
160915 95.997 -77.389 49.298
160916 96.010 -77.399 49.262
160917 96.022 -77.409 49.228
160918 96.036 -77.420 49.195
160919 96.056 -77.429 49.164
160920 96.077 -77.439 49.133
160921 96.097 -77.448 49.104
160922 96.117 -77.457 49.076
160923 96.137 -77.467 49.049
160924 96.158 -77.476 49.027
160925 96.178 -77.486 49.007
160926 96.194 -77.496 48.988
160927 96.203 -77.508 48.972
160928 96.213 -77.519 48.957
160929 96.224 -77.523 48.940
160930 96.235 -77.526 48.919
160931 96.246 -77.530 48.895
160932 96.258 -77.534 48.872
160933 96.273 -77.536 48.852
160934 96.291 -77.537 48.836
160935 96.309 -77.538 48.821
160936 96.331 -77.541 48.803
160937 96.361 -77.551 48.784
160938 96.390 -77.561 48.767
160939 96.419 -77.570 48.751
160940 96.449 -77.580 48.734
160941 96.478 -77.589 48.718
160942 96.508 -77.599 48.700
160943 96.533 -77.613 48.682
160944 96.550 -77.635 48.665
160945 96.567 -77.656 48.649
160946 96.584 -77.677 48.634
160947 96.597 -77.681 48.619
160948 96.609 -77.679 48.603
160949 96.621 -77.677 48.585
160950 96.632 -77.676 48.564
160951 96.645 -77.674 48.544
160952 96.653 -77.674 48.524
160953 96.659 -77.676 48.508
160954 96.666 -77.677 48.492
160955 96.673 -77.678 48.476
160956 96.680 -77.680 48.459
160957 96.687 -77.681 48.439
160958 96.693 -77.683 48.419
160959 96.700 -77.684 48.399
160960 96.707 -77.686 48.378
160961 96.714 -77.687 48.359
160962 96.724 -77.693 48.339
160963 96.736 -77.699 48.317
160964 96.750 -77.701 48.293
160965 96.764 -77.703 48.268
160966 96.778 -77.705 48.242
160967 96.793 -77.708 48.215
160968 96.807 -77.710 48.188
160969 96.821 -77.713 48.162
160970 96.835 -77.715 48.136
160971 96.849 -77.718 48.109
160972 96.870 -77.721 48.082
160973 96.891 -77.724 48.056
160974 96.912 -77.727 48.031
160975 96.934 -77.729 48.006
160976 96.955 -77.732 47.982
160977 96.980 -77.733 47.958
160978 97.006 -77.734 47.933
160979 97.028 -77.736 47.911
160980 97.033 -77.743 47.890
160981 97.039 -77.750 47.871
160982 97.045 -77.757 47.851
160983 97.053 -77.762 47.831
160984 97.065 -77.765 47.811
160985 97.077 -77.763 47.791
160986 97.090 -77.754 47.770
160987 97.103 -77.746 47.749
160988 97.129 -77.743 47.728
160989 97.154 -77.740 47.706
160990 97.179 -77.736 47.685
160991 97.205 -77.733 47.663
160992 97.230 -77.730 47.641
160993 97.255 -77.727 47.619
160994 97.280 -77.733 47.597
160995 97.303 -77.743 47.575
160996 97.327 -77.753 47.554
160997 97.350 -77.763 47.532
160998 97.374 -77.773 47.512
160999 97.400 -77.782 47.492
161000 97.431 -77.790 47.471
161001 97.461 -77.799 47.451
161002 97.491 -77.808 47.431
161003 97.521 -77.816 47.410
161004 97.551 -77.824 47.389
161005 97.582 -77.834 47.370
161006 97.604 -77.836 47.350
161007 97.613 -77.827 47.328
161008 97.622 -77.820 47.306
161009 97.631 -77.812 47.284
161010 97.640 -77.803 47.262
161011 97.651 -77.802 47.240
161012 97.666 -77.818 47.218
161013 97.682 -77.833 47.194
161014 97.697 -77.849 47.168
161015 97.693 -77.876 47.142
161016 97.691 -77.904 47.116
161017 97.696 -77.925 47.090
161018 97.713 -77.935 47.064
161019 97.731 -77.945 47.038
161020 97.749 -77.955 47.013
161021 97.758 -77.950 46.988
161022 97.784 -77.941 46.963
161023 97.803 -77.942 46.938
161024 97.823 -77.942 46.913
161025 97.843 -77.943 46.888
161026 97.865 -77.942 46.863
161027 97.899 -77.927 46.839
161028 97.935 -77.912 46.815
161029 97.954 -77.912 46.792
161030 97.967 -77.918 46.768
161031 97.981 -77.925 46.744
161032 97.994 -77.932 46.720
161033 98.007 -77.938 46.696
161034 98.020 -77.945 46.675
161035 98.017 -77.949 46.654
161036 98.013 -77.953 46.633
161037 98.008 -77.956 46.611
161038 98.070 -77.914 46.589
161039 98.098 -77.941 46.568
161040 98.126 -77.968 46.547
161041 98.154 -77.996 46.526
161042 98.191 -77.987 46.504
161043 98.257 -77.952 46.481
161044 98.323 -77.916 46.458
161045 98.389 -77.881 46.434
161046 98.435 -77.846 46.410
161047 98.474 -77.813 46.386
161048 98.512 -77.780 46.365
161049 98.551 -77.747 46.343
161050 98.598 -77.740 46.323
161051 98.602 -77.780 46.303
161052 98.605 -77.823 46.282
161053 98.608 -77.865 46.263
161054 98.611 -77.907 46.245
161055 98.619 -77.946 46.228
161056 98.637 -77.951 46.208
161057 98.651 -77.958 46.188
161058 98.665 -77.965 46.167
161059 98.680 -77.972 46.147
161060 98.694 -77.979 46.127
161061 98.708 -77.987 46.108
161062 98.722 -77.993 46.090
161063 98.737 -78.000 46.070
161064 98.751 -78.008 46.049
161065 98.765 -78.015 46.028
161066 98.775 -78.017 46.008
161067 98.776 -78.024 45.987
161068 98.776 -78.031 45.968
161069 98.777 -78.038 45.950
161070 98.778 -78.046 45.933
161071 98.788 -78.053 45.916
161072 98.841 -78.061 45.899
161073 98.875 -78.065 45.882
161074 98.908 -78.066 45.864
161075 98.963 -78.067 45.849
161076 99.018 -78.067 45.834
161077 99.073 -78.068 45.818
161078 99.129 -78.069 45.803
161079 99.184 -78.069 45.788
161080 99.239 -78.069 45.774
161081 99.276 -78.068 45.760
161082 99.294 -78.065 45.748
161083 99.312 -78.061 45.737
161084 99.331 -78.058 45.727
161085 99.349 -78.055 45.715
161086 99.368 -78.052 45.704
161087 99.386 -78.048 45.694
161088 99.400 -78.050 45.681
161089 99.410 -78.056 45.670
161090 99.420 -78.062 45.659
161091 99.430 -78.067 45.648
161092 99.440 -78.072 45.638
161093 99.450 -78.078 45.628
161094 99.465 -78.087 45.617
161095 99.495 -78.106 45.606
161096 99.532 -78.126 45.594
161097 99.571 -78.145 45.583
161098 99.610 -78.165 45.572
161099 99.643 -78.184 45.560
161100 99.671 -78.204 45.548
161101 99.699 -78.223 45.536
161102 99.727 -78.243 45.524
161103 99.755 -78.263 45.511
161104 99.772 -78.278 45.498
161105 99.789 -78.293 45.486
161106 99.806 -78.308 45.474
161107 99.823 -78.323 45.462
161108 99.840 -78.338 45.449
161109 99.857 -78.353 45.434
161110 99.857 -78.372 45.420
161111 99.865 -78.393 45.405
161112 99.902 -78.413 45.389
161113 99.938 -78.435 45.374
161114 99.974 -78.455 45.358
161115 100.010 -78.476 45.342
161116 100.047 -78.497 45.325
161117 100.083 -78.517 45.308
161118 100.119 -78.538 45.292
161119 100.155 -78.559 45.274
161120 100.173 -78.574 45.256
161121 100.168 -78.582 45.239
161122 100.162 -78.592 45.223
161123 100.157 -78.601 45.207
161124 100.157 -78.590 45.190
161125 100.168 -78.575 45.174
161126 100.180 -78.560 45.157
161127 100.214 -78.575 45.141
161128 100.249 -78.590 45.126
161129 100.283 -78.605 45.109
161130 100.317 -78.620 45.094
161131 100.357 -78.633 45.078
161132 100.402 -78.645 45.062
161133 100.449 -78.656 45.046
161134 100.496 -78.668 45.031
161135 100.544 -78.680 45.016
161136 100.591 -78.692 45.000
161137 100.638 -78.704 44.984
161138 100.671 -78.714 44.968
161139 100.691 -78.723 44.954
161140 100.713 -78.730 44.937
161141 100.734 -78.738 44.921
161142 100.755 -78.746 44.906
161143 100.772 -78.761 44.890
161144 100.787 -78.783 44.874
161145 100.802 -78.804 44.858
161146 100.826 -78.822 44.842
161147 100.857 -78.839 44.825
161148 100.888 -78.855 44.809
161149 100.923 -78.866 44.792
161150 100.963 -78.874 44.776
161151 101.002 -78.882 44.760
161152 101.046 -78.886 44.743
161153 101.094 -78.890 44.727
161154 101.124 -78.901 44.712
161155 101.152 -78.915 44.696
161156 101.190 -78.924 44.681
161157 101.229 -78.933 44.666
161158 101.259 -78.930 44.652
161159 101.281 -78.916 44.638
161160 101.305 -78.930 44.624
161161 101.329 -78.943 44.611
161162 101.354 -78.957 44.598
161163 101.375 -78.979 44.585
161164 101.392 -79.013 44.572
161165 101.408 -79.047 44.559
161166 101.447 -79.068 44.547
161167 101.489 -79.088 44.534
161168 101.531 -79.107 44.520
161169 101.573 -79.127 44.508
161170 101.616 -79.146 44.496
161392 92.432 -77.451 54.036
161393 92.452 -77.459 54.050
161394 92.472 -77.467 54.063
161395 92.492 -77.473 54.076
161396 92.512 -77.481 54.090
161397 92.532 -77.489 54.103
161398 92.552 -77.495 54.117
161399 92.572 -77.503 54.132
161400 92.592 -77.511 54.147
161401 92.612 -77.517 54.160
161402 92.632 -77.525 54.173
161403 92.652 -77.533 54.185
161404 92.671 -77.539 54.197
161405 92.690 -77.544 54.208
161406 92.709 -77.550 54.219
161407 92.728 -77.555 54.228
161408 92.747 -77.561 54.239
161409 92.769 -77.569 54.247
161410 92.788 -77.581 54.256
161411 92.813 -77.579 54.264
161412 92.842 -77.573 54.272
161413 92.866 -77.570 54.280
161414 92.880 -77.560 54.287
161415 92.894 -77.551 54.294
161416 92.909 -77.543 54.301
161417 92.921 -77.537 54.307
161418 92.929 -77.533 54.311
161419 92.938 -77.530 54.315
161420 92.947 -77.526 54.319
161421 92.956 -77.521 54.321
161422 92.966 -77.517 54.321
161423 92.975 -77.514 54.322
161424 92.984 -77.510 54.325
161425 92.993 -77.506 54.329
161426 93.001 -77.502 54.335
161427 93.010 -77.498 54.342
161428 93.021 -77.489 54.349
161429 93.037 -77.473 54.355
161430 93.054 -77.457 54.363
161431 93.071 -77.440 54.370
161432 93.088 -77.424 54.376
161433 93.105 -77.407 54.383
161434 93.122 -77.391 54.390
161435 93.139 -77.375 54.394
161436 93.156 -77.358 54.397
161437 93.173 -77.342 54.399
161438 93.190 -77.326 54.401
161439 93.211 -77.315 54.401
161440 93.241 -77.315 54.400
161441 93.271 -77.315 54.393
161442 93.280 -77.300 54.384
161443 93.286 -77.282 54.375
161444 93.295 -77.267 54.367
161445 93.313 -77.262 54.360
161446 93.330 -77.257 54.353
161447 93.348 -77.251 54.346
161448 93.365 -77.245 54.337
161449 93.383 -77.240 54.324
161450 93.401 -77.235 54.311
161451 93.418 -77.229 54.299
161452 93.436 -77.224 54.288
161453 93.454 -77.219 54.279
161454 93.476 -77.208 54.270
161455 93.499 -77.194 54.260
161456 93.522 -77.182 54.248
161457 93.546 -77.168 54.237
161458 93.569 -77.154 54.226
161459 93.592 -77.141 54.216
161460 93.616 -77.128 54.205
161461 93.638 -77.116 54.194
161462 93.659 -77.107 54.182
161463 93.680 -77.099 54.169
161464 93.694 -77.101 54.155
161465 93.708 -77.104 54.139
161466 93.722 -77.107 54.124
161467 93.737 -77.110 54.105
161468 93.751 -77.113 54.086
161469 93.765 -77.116 54.066
161470 93.774 -77.111 54.049
161471 93.776 -77.108 54.031
161472 93.782 -77.105 54.013
161473 93.787 -77.101 53.996
161474 93.793 -77.098 53.980
161475 93.799 -77.095 53.965
161476 93.804 -77.091 53.952
161477 93.811 -77.088 53.941
161478 93.816 -77.085 53.928
161479 93.847 -77.075 53.915
161480 93.877 -77.066 53.898
161481 93.901 -77.056 53.878
161482 93.917 -77.052 53.857
161483 93.931 -77.047 53.836
161484 93.946 -77.043 53.817
161485 93.960 -77.040 53.802
161486 93.969 -77.038 53.788
161487 93.987 -77.035 53.773
161488 94.010 -77.031 53.753
161489 94.034 -77.027 53.734
161490 94.058 -77.022 53.715
161491 94.082 -77.018 53.695
161492 94.106 -77.013 53.673
161493 94.129 -77.009 53.651
161494 94.150 -77.005 53.627
161495 94.164 -77.002 53.603
161496 94.176 -76.999 53.580
161497 94.189 -76.996 53.559
161498 94.201 -76.994 53.539
161499 94.213 -76.990 53.521
161500 94.225 -76.987 53.504
161501 94.240 -76.985 53.487
161502 94.255 -76.983 53.471
161503 94.270 -76.981 53.459
161504 94.286 -76.979 53.451
161505 94.301 -76.976 53.444
161506 94.316 -76.973 53.438
161507 94.328 -76.969 53.434
161508 94.338 -76.965 53.431
161509 94.339 -76.965 53.428
161510 94.340 -76.965 53.427
161511 94.341 -76.965 53.424
161512 94.350 -76.964 53.418
161513 94.367 -76.959 53.408
161514 94.384 -76.956 53.397
161515 94.401 -76.951 53.379
161516 94.418 -76.947 53.359
161517 94.435 -76.943 53.338
161518 94.452 -76.939 53.315
161519 94.469 -76.934 53.292
161520 94.486 -76.931 53.268
161521 94.502 -76.928 53.245
161522 94.517 -76.925 53.217
161523 94.535 -76.923 53.190
161524 94.563 -76.920 53.163
161525 94.587 -76.915 53.136
161526 94.596 -76.910 53.109
161527 94.606 -76.904 53.081
161528 94.615 -76.899 53.052
161529 94.625 -76.893 53.019
161530 94.643 -76.898 52.986
161531 94.665 -76.906 52.953
161532 94.686 -76.914 52.919
161533 94.707 -76.922 52.886
161534 94.721 -76.925 52.852
161535 94.758 -76.927 52.819
161536 94.793 -76.919 52.784
161537 94.827 -76.912 52.750
161538 94.862 -76.904 52.718
161539 94.897 -76.897 52.687
161540 94.914 -76.894 52.655
161541 94.919 -76.896 52.623
161542 94.925 -76.898 52.590
161543 94.933 -76.898 52.557
161544 94.944 -76.896 52.529
161545 94.955 -76.894 52.505
161546 94.965 -76.893 52.485
161547 94.976 -76.891 52.469
161548 94.986 -76.890 52.449
161549 95.007 -76.896 52.430
161550 95.028 -76.902 52.410
161551 95.049 -76.908 52.392
161552 95.070 -76.914 52.370
161553 95.092 -76.920 52.343
161554 95.113 -76.926 52.308
161555 95.134 -76.932 52.265
161556 95.156 -76.939 52.217
161557 95.177 -76.945 52.164
161558 95.198 -76.951 52.108
161559 95.219 -76.957 52.051
161560 95.242 -76.964 51.993
161561 95.279 -76.976 51.934
161562 95.289 -76.997 51.869
161563 95.290 -77.015 51.806
161564 95.305 -77.024 51.745
161565 95.318 -77.033 51.689
161566 95.332 -77.043 51.630
161567 95.346 -77.052 51.572
161568 95.360 -77.061 51.514
161569 95.374 -77.071 51.456
161570 95.388 -77.080 51.401
161571 95.402 -77.089 51.347
161572 95.415 -77.099 51.297
161573 95.432 -77.110 51.248
161574 95.449 -77.122 51.202
161575 95.466 -77.134 51.157
161576 95.483 -77.145 51.112
161577 95.500 -77.160 51.068
161578 95.519 -77.176 51.023
161579 95.536 -77.191 50.977
161580 95.551 -77.202 50.928
161581 95.566 -77.213 50.881
161582 95.581 -77.224 50.835
161583 95.597 -77.230 50.789
161584 95.618 -77.228 50.743
161585 95.638 -77.226 50.698
161586 95.659 -77.224 50.653
161587 95.679 -77.222 50.611
161588 95.699 -77.220 50.568
161589 95.720 -77.224 50.526
161590 95.739 -77.231 50.485
161591 95.756 -77.238 50.446
161592 95.763 -77.244 50.408
161593 95.770 -77.250 50.371
161594 95.777 -77.256 50.333
161595 95.783 -77.262 50.290
161596 95.800 -77.271 50.244
161597 95.820 -77.282 50.198
161598 95.840 -77.292 50.153
161599 95.865 -77.303 50.108
161600 95.898 -77.313 50.061
161601 95.931 -77.324 50.013
161602 95.964 -77.334 49.964
161603 95.978 -77.339 49.913
161604 95.992 -77.344 49.863
161605 95.991 -77.342 49.814
161606 95.985 -77.338 49.761
161607 95.979 -77.335 49.711
161608 95.972 -77.331 49.660
161609 95.966 -77.326 49.611
161610 95.960 -77.326 49.561
161611 95.952 -77.334 49.514
161612 95.944 -77.342 49.470
161613 95.947 -77.351 49.427
161614 95.959 -77.362 49.385
161615 95.971 -77.373 49.344
161616 95.983 -77.382 49.303
161617 95.995 -77.393 49.266
161618 96.011 -77.403 49.232
161619 96.031 -77.413 49.200
161620 96.051 -77.423 49.168
161621 96.072 -77.432 49.137
161622 96.092 -77.442 49.108
161623 96.112 -77.451 49.080
161624 96.132 -77.461 49.052
161625 96.153 -77.470 49.031
161626 96.173 -77.479 49.010
161627 96.193 -77.489 48.990
161628 96.205 -77.500 48.974
161629 96.216 -77.509 48.958
161630 96.227 -77.513 48.941
161631 96.238 -77.517 48.921
161632 96.253 -77.518 48.899
161633 96.271 -77.519 48.876
161634 96.290 -77.520 48.854
161635 96.308 -77.520 48.839
161636 96.326 -77.520 48.823
161637 96.345 -77.521 48.807
161638 96.370 -77.527 48.791
161639 96.399 -77.537 48.776
161640 96.429 -77.547 48.761
161641 96.458 -77.557 48.746
161642 96.488 -77.567 48.730
161643 96.510 -77.583 48.713
161644 96.527 -77.605 48.695
161645 96.544 -77.626 48.678
161646 96.561 -77.648 48.662
161647 96.579 -77.669 48.648
161648 96.594 -77.682 48.633
161649 96.605 -77.680 48.616
161650 96.617 -77.678 48.597
161651 96.629 -77.676 48.577
161652 96.641 -77.674 48.557
161653 96.653 -77.672 48.539
161654 96.665 -77.671 48.524
161655 96.676 -77.669 48.508
161656 96.683 -77.671 48.492
161657 96.690 -77.672 48.477
161658 96.696 -77.674 48.459
161659 96.703 -77.674 48.439
161660 96.711 -77.676 48.419
161661 96.717 -77.677 48.397
161662 96.724 -77.679 48.378
161663 96.733 -77.683 48.358
161664 96.744 -77.688 48.336
161665 96.758 -77.691 48.312
161666 96.772 -77.693 48.287
161667 96.787 -77.696 48.261
161668 96.801 -77.698 48.234
161669 96.815 -77.700 48.207
161670 96.829 -77.703 48.179
161671 96.849 -77.705 48.152
161672 96.870 -77.708 48.125
161673 96.891 -77.711 48.098
161674 96.912 -77.714 48.072
161675 96.934 -77.718 48.047
161676 96.955 -77.721 48.022
161677 96.976 -77.723 47.999
161678 97.001 -77.724 47.976
161679 97.027 -77.726 47.952
161680 97.033 -77.733 47.929
161681 97.039 -77.739 47.909
161682 97.045 -77.746 47.890
161683 97.051 -77.752 47.871
161684 97.056 -77.759 47.850
161685 97.068 -77.762 47.830
161686 97.081 -77.754 47.811
161687 97.103 -77.749 47.790
161688 97.129 -77.746 47.768
161689 97.154 -77.743 47.747
161690 97.179 -77.740 47.726
161691 97.205 -77.737 47.704
161692 97.230 -77.734 47.683
161693 97.256 -77.731 47.661
161694 97.281 -77.728 47.639
161695 97.305 -77.736 47.616
161696 97.328 -77.746 47.594
161697 97.352 -77.755 47.572
161698 97.376 -77.765 47.550
161699 97.399 -77.775 47.529
161700 97.423 -77.785 47.509
161701 97.447 -77.794 47.486
161702 97.477 -77.802 47.465
161703 97.508 -77.812 47.444
161704 97.538 -77.820 47.423
161705 97.568 -77.828 47.402
161706 97.599 -77.837 47.382
161707 97.623 -77.841 47.361
161708 97.632 -77.833 47.339
161709 97.641 -77.824 47.318
161710 97.650 -77.817 47.296
161711 97.658 -77.808 47.274
161712 97.667 -77.800 47.253
161713 97.680 -77.805 47.229
161714 97.696 -77.821 47.204
161715 97.705 -77.840 47.179
161716 97.702 -77.868 47.152
161717 97.710 -77.886 47.126
161718 97.728 -77.896 47.100
161719 97.745 -77.906 47.074
161720 97.764 -77.916 47.048
161721 97.782 -77.926 47.024
161722 97.797 -77.912 47.000
161723 97.817 -77.912 46.976
161724 97.837 -77.913 46.950
161725 97.857 -77.913 46.925
161726 97.877 -77.914 46.900
161727 97.897 -77.915 46.874
161728 97.925 -77.907 46.850
161729 97.951 -77.900 46.825
161730 97.964 -77.907 46.801
161731 97.978 -77.914 46.777
161732 97.991 -77.920 46.751
161733 98.004 -77.927 46.726
161734 98.017 -77.934 46.702
161735 98.030 -77.940 46.679
161736 98.034 -77.945 46.658
161737 98.029 -77.949 46.636
161738 98.028 -77.950 46.614
161739 98.114 -77.890 46.592
161740 98.142 -77.917 46.571
161741 98.170 -77.944 46.549
161742 98.198 -77.972 46.528
161743 98.226 -78.000 46.507
161744 98.246 -78.003 46.485
161745 98.286 -77.972 46.461
161746 98.324 -77.939 46.437
161747 98.363 -77.906 46.413
161748 98.402 -77.872 46.390
161749 98.440 -77.839 46.369
161750 98.479 -77.805 46.348
161751 98.523 -77.792 46.328
161752 98.560 -77.786 46.309
161753 98.563 -77.828 46.289
161754 98.566 -77.870 46.271
161755 98.569 -77.912 46.253
161756 98.590 -77.930 46.237
161757 98.604 -77.937 46.217
161758 98.619 -77.943 46.197
161759 98.633 -77.951 46.177
161760 98.647 -77.958 46.155
161761 98.661 -77.965 46.135
161762 98.675 -77.972 46.117
161763 98.690 -77.979 46.099
161764 98.704 -77.987 46.079
161765 98.718 -77.993 46.058
161766 98.732 -78.000 46.038
161767 98.747 -78.008 46.017
161768 98.761 -78.015 45.996
161769 98.775 -78.022 45.977
161770 98.789 -78.029 45.959
161771 98.795 -78.034 45.941
161772 98.796 -78.041 45.924
161773 98.830 -78.049 45.907
161774 98.869 -78.050 45.890
161775 98.924 -78.050 45.872
161776 98.979 -78.051 45.856
161777 99.034 -78.051 45.841
161778 99.089 -78.052 45.825
161779 99.144 -78.052 45.809
161780 99.199 -78.053 45.795
161781 99.255 -78.053 45.781
161782 99.288 -78.051 45.767
161783 99.306 -78.048 45.755
161784 99.325 -78.045 45.743
161785 99.344 -78.042 45.732
161786 99.362 -78.038 45.722
161787 99.380 -78.035 45.712
161788 99.392 -78.040 45.701
161789 99.402 -78.045 45.690
161790 99.412 -78.050 45.678
161791 99.422 -78.056 45.667
161792 99.432 -78.062 45.655
161793 99.443 -78.067 45.646
161794 99.453 -78.073 45.635
161795 99.463 -78.078 45.624
161796 99.481 -78.087 45.613
161797 99.519 -78.106 45.601
161798 99.559 -78.126 45.590
161799 99.597 -78.146 45.578
161800 99.636 -78.166 45.566
161801 99.664 -78.184 45.554
161802 99.693 -78.204 45.542
161803 99.720 -78.224 45.530
161804 99.743 -78.241 45.517
161805 99.760 -78.256 45.505
161806 99.776 -78.271 45.493
161807 99.793 -78.286 45.481
161808 99.810 -78.301 45.467
161809 99.827 -78.316 45.454
161810 99.844 -78.331 45.439
161811 99.860 -78.346 45.425
161812 99.852 -78.367 45.410
161813 99.887 -78.388 45.395
161814 99.923 -78.409 45.379
161815 99.959 -78.429 45.362
161816 99.995 -78.450 45.346
161817 100.031 -78.471 45.329
161818 100.068 -78.492 45.313
161819 100.104 -78.512 45.296
161820 100.113 -78.525 45.279
161821 100.108 -78.534 45.263
161822 100.103 -78.543 45.246
161823 100.098 -78.552 45.230
161824 100.091 -78.558 45.214
161825 100.100 -78.544 45.197
161826 100.119 -78.539 45.180
161827 100.153 -78.554 45.164
161828 100.187 -78.569 45.148
161829 100.231 -78.583 45.133
161830 100.281 -78.598 45.117
161831 100.328 -78.611 45.101
161832 100.365 -78.620 45.085
161833 100.406 -78.631 45.068
161834 100.451 -78.642 45.053
161835 100.495 -78.653 45.037
161836 100.540 -78.664 45.022
161837 100.585 -78.676 45.007
161838 100.630 -78.687 44.991
161839 100.675 -78.698 44.975
161840 100.702 -78.705 44.961
161841 100.726 -78.713 44.944
161842 100.747 -78.720 44.928
161843 100.768 -78.728 44.912
161844 100.789 -78.736 44.896
161845 100.806 -78.753 44.881
161846 100.824 -78.774 44.865
161847 100.855 -78.789 44.849
161848 100.886 -78.806 44.832
161849 100.916 -78.822 44.816
161850 100.948 -78.838 44.800
161851 100.978 -78.855 44.784
161852 101.012 -78.868 44.768
161853 101.052 -78.875 44.752
161854 101.097 -78.880 44.736
161855 101.113 -78.897 44.720
161856 101.148 -78.908 44.704
161857 101.187 -78.918 44.689
161858 101.226 -78.927 44.673
161859 101.260 -78.930 44.660
161860 101.281 -78.911 44.646
161861 101.304 -78.915 44.632
161862 101.329 -78.928 44.619
161863 101.353 -78.942 44.606
161864 101.374 -78.965 44.593
161865 101.390 -79.000 44.579
161866 101.407 -79.034 44.566
161867 101.447 -79.055 44.554
161868 101.489 -79.075 44.540
161869 101.531 -79.093 44.527
161870 101.573 -79.112 44.515
161871 101.616 -79.132 44.502
161872 101.656 -79.150 44.489
162094 92.440 -77.438 54.028
162095 92.459 -77.443 54.042
162096 92.479 -77.448 54.057
162097 92.498 -77.454 54.072
162098 92.517 -77.460 54.086
162099 92.536 -77.466 54.103
162100 92.555 -77.471 54.119
162101 92.574 -77.476 54.134
162102 92.593 -77.483 54.149
162103 92.613 -77.488 54.163
162104 92.632 -77.493 54.176
162105 92.651 -77.499 54.188
162106 92.670 -77.504 54.200
162107 92.689 -77.511 54.211
162108 92.710 -77.518 54.221
162109 92.733 -77.529 54.232
162110 92.756 -77.540 54.241
162111 92.779 -77.551 54.250
162112 92.788 -77.564 54.259
162113 92.810 -77.564 54.268
162114 92.840 -77.558 54.277
162115 92.865 -77.551 54.286
162116 92.882 -77.539 54.294
162117 92.900 -77.528 54.301
162118 92.910 -77.523 54.307
162119 92.920 -77.520 54.312
162120 92.930 -77.516 54.316
162121 92.940 -77.512 54.319
162122 92.950 -77.508 54.322
162123 92.959 -77.504 54.323
162124 92.970 -77.501 54.326
162125 92.979 -77.497 54.331
162126 92.989 -77.493 54.336
162127 92.999 -77.489 54.345
162128 93.009 -77.485 54.354
162129 93.022 -77.473 54.361
162130 93.039 -77.457 54.368
162131 93.055 -77.441 54.376
162132 93.070 -77.425 54.383
162133 93.086 -77.408 54.390
162134 93.102 -77.392 54.396
162135 93.118 -77.376 54.402
162136 93.134 -77.359 54.406
162137 93.150 -77.343 54.411
162138 93.167 -77.326 54.415
162139 93.184 -77.310 54.416
162140 93.201 -77.294 54.417
162141 93.218 -77.278 54.416
162142 93.235 -77.269 54.408
162143 93.253 -77.263 54.399
162144 93.270 -77.257 54.389
162145 93.288 -77.252 54.381
162146 93.305 -77.247 54.373
162147 93.323 -77.241 54.366
162148 93.341 -77.235 54.358
162149 93.358 -77.230 54.350
162150 93.376 -77.225 54.337
162151 93.394 -77.219 54.324
162152 93.411 -77.214 54.311
162153 93.429 -77.209 54.299
162154 93.447 -77.203 54.288
162155 93.471 -77.189 54.279
162156 93.494 -77.175 54.271
162157 93.517 -77.163 54.259
162158 93.541 -77.149 54.248
162159 93.564 -77.135 54.238
162160 93.587 -77.122 54.229
162161 93.611 -77.109 54.218
162162 93.634 -77.095 54.208
162163 93.657 -77.083 54.198
162164 93.674 -77.081 54.186
162165 93.688 -77.085 54.173
162166 93.702 -77.088 54.159
162167 93.716 -77.091 54.143
162168 93.730 -77.093 54.126
162169 93.745 -77.096 54.107
162170 93.759 -77.099 54.088
162171 93.770 -77.100 54.071
162172 93.776 -77.097 54.054
162173 93.782 -77.094 54.036
162174 93.788 -77.091 54.019
162175 93.793 -77.087 54.003
162176 93.799 -77.084 53.989
162177 93.805 -77.081 53.976
162178 93.811 -77.077 53.964
162179 93.817 -77.074 53.953
162180 93.841 -77.066 53.938
162181 93.872 -77.056 53.921
162182 93.890 -77.048 53.900
162183 93.905 -77.043 53.878
162184 93.920 -77.039 53.857
162185 93.935 -77.034 53.838
162186 93.949 -77.031 53.823
162187 93.965 -77.026 53.809
162188 93.980 -77.022 53.795
162189 94.003 -77.017 53.777
162190 94.027 -77.012 53.758
162191 94.051 -77.008 53.739
162192 94.075 -77.003 53.719
162193 94.099 -76.999 53.698
162194 94.123 -76.994 53.676
162195 94.141 -76.990 53.652
162196 94.154 -76.987 53.627
162197 94.167 -76.985 53.605
162198 94.180 -76.982 53.585
162199 94.192 -76.979 53.566
162200 94.204 -76.976 53.546
162201 94.216 -76.973 53.527
162202 94.231 -76.972 53.509
162203 94.247 -76.969 53.492
162204 94.262 -76.967 53.477
162205 94.277 -76.965 53.466
162206 94.293 -76.962 53.458
162207 94.307 -76.958 53.451
162208 94.319 -76.954 53.446
162209 94.331 -76.950 53.440
162210 94.335 -76.948 53.435
162211 94.336 -76.948 53.432
162212 94.338 -76.948 53.430
162213 94.348 -76.949 53.425
162214 94.366 -76.946 53.417
162215 94.383 -76.941 53.407
162216 94.400 -76.937 53.391
162217 94.417 -76.933 53.372
162218 94.434 -76.929 53.351
162219 94.451 -76.925 53.329
162220 94.468 -76.921 53.308
162221 94.485 -76.917 53.286
162222 94.500 -76.915 53.263
162223 94.515 -76.913 53.235
162224 94.530 -76.911 53.206
162225 94.554 -76.908 53.177
162226 94.576 -76.903 53.151
162227 94.585 -76.898 53.124
162228 94.595 -76.893 53.097
162229 94.605 -76.887 53.068
162230 94.617 -76.884 53.035
162231 94.638 -76.891 53.001
162232 94.659 -76.899 52.968
162233 94.680 -76.907 52.936
162234 94.702 -76.914 52.904
162235 94.735 -76.907 52.871
162236 94.769 -76.900 52.836
162237 94.804 -76.892 52.801
162238 94.838 -76.884 52.768
162239 94.873 -76.876 52.736
162240 94.903 -76.870 52.704
162241 94.908 -76.872 52.672
162242 94.914 -76.875 52.638
162243 94.923 -76.874 52.605
162244 94.933 -76.871 52.571
162245 94.944 -76.870 52.543
162246 94.955 -76.868 52.518
162247 94.965 -76.867 52.497
162248 94.976 -76.865 52.480
162249 94.986 -76.864 52.462
162250 95.003 -76.867 52.444
162251 95.024 -76.873 52.428
162252 95.045 -76.879 52.411
162253 95.066 -76.885 52.389
162254 95.088 -76.891 52.361
162255 95.109 -76.897 52.327
162256 95.130 -76.903 52.282
162257 95.151 -76.910 52.233
162258 95.173 -76.916 52.180
162259 95.194 -76.922 52.125
162260 95.215 -76.928 52.067
162261 95.249 -76.940 52.008
162262 95.286 -76.952 51.946
162263 95.300 -76.972 51.882
162264 95.286 -76.999 51.819
162265 95.300 -77.009 51.759
162266 95.314 -77.018 51.701
162267 95.328 -77.027 51.644
162268 95.342 -77.036 51.586
162269 95.355 -77.046 51.528
162270 95.370 -77.055 51.470
162271 95.383 -77.064 51.415
162272 95.397 -77.074 51.361
162273 95.412 -77.085 51.309
162274 95.429 -77.096 51.260
162275 95.446 -77.108 51.213
162276 95.463 -77.119 51.167
162277 95.480 -77.131 51.120
162278 95.497 -77.143 51.076
162279 95.515 -77.160 51.031
162280 95.533 -77.176 50.986
162281 95.550 -77.191 50.940
162282 95.565 -77.202 50.895
162283 95.580 -77.213 50.850
162284 95.599 -77.214 50.804
162285 95.619 -77.212 50.759
162286 95.640 -77.210 50.714
162287 95.660 -77.208 50.668
162288 95.681 -77.206 50.624
162289 95.701 -77.204 50.582
162290 95.722 -77.202 50.541
162291 95.742 -77.208 50.500
162292 95.757 -77.215 50.460
162293 95.764 -77.221 50.422
162294 95.771 -77.227 50.384
162295 95.778 -77.233 50.345
162296 95.785 -77.239 50.302
162297 95.794 -77.245 50.257
162298 95.814 -77.256 50.211
162299 95.834 -77.267 50.166
162300 95.862 -77.277 50.121
162301 95.895 -77.288 50.074
162302 95.927 -77.298 50.026
162303 95.960 -77.309 49.978
162304 95.978 -77.315 49.929
162305 95.988 -77.318 49.879
162306 95.982 -77.314 49.831
162307 95.976 -77.310 49.779
162308 95.970 -77.307 49.728
162309 95.963 -77.305 49.676
162310 95.956 -77.313 49.623
162311 95.948 -77.321 49.571
162312 95.940 -77.329 49.524
162313 95.932 -77.337 49.478
162314 95.932 -77.346 49.433
162315 95.944 -77.356 49.391
162316 95.957 -77.367 49.349
162317 95.969 -77.377 49.309
162318 95.985 -77.387 49.271
162319 96.005 -77.397 49.237
162320 96.026 -77.406 49.204
162321 96.046 -77.416 49.173
162322 96.066 -77.425 49.141
162323 96.087 -77.435 49.111
162324 96.107 -77.444 49.083
162325 96.127 -77.454 49.056
162326 96.148 -77.463 49.034
162327 96.168 -77.473 49.013
162328 96.188 -77.482 48.993
162329 96.207 -77.492 48.975
162330 96.218 -77.498 48.955
162331 96.234 -77.500 48.935
162332 96.252 -77.501 48.915
162333 96.270 -77.501 48.895
162334 96.289 -77.502 48.874
162335 96.307 -77.502 48.854
162336 96.325 -77.503 48.837
162337 96.343 -77.504 48.822
162338 96.362 -77.504 48.808
162339 96.380 -77.505 48.795
162340 96.409 -77.514 48.783
162341 96.439 -77.524 48.770
162342 96.468 -77.534 48.756
162343 96.487 -77.553 48.742
162344 96.505 -77.574 48.725
162345 96.522 -77.596 48.708
162346 96.539 -77.617 48.691
162347 96.556 -77.639 48.676
162348 96.573 -77.661 48.662
162349 96.590 -77.682 48.646
162350 96.602 -77.680 48.628
162351 96.614 -77.678 48.609
162352 96.626 -77.677 48.590
162353 96.638 -77.675 48.571
162354 96.650 -77.673 48.553
162355 96.662 -77.671 48.537
162356 96.673 -77.669 48.521
162357 96.685 -77.667 48.505
162358 96.697 -77.665 48.490
162359 96.707 -77.665 48.473
162360 96.714 -77.667 48.455
162361 96.721 -77.668 48.436
162362 96.727 -77.669 48.415
162363 96.734 -77.671 48.394
162364 96.741 -77.673 48.373
162365 96.752 -77.678 48.351
162366 96.766 -77.681 48.327
162367 96.781 -77.683 48.303
162368 96.795 -77.686 48.278
162369 96.809 -77.688 48.250
162370 96.828 -77.691 48.223
162371 96.849 -77.694 48.196
162372 96.871 -77.696 48.169
162373 96.892 -77.699 48.142
162374 96.913 -77.702 48.116
162375 96.935 -77.705 48.090
162376 96.956 -77.708 48.065
162377 96.977 -77.711 48.041
162378 96.999 -77.714 48.017
162379 97.022 -77.716 47.994
162380 97.033 -77.721 47.971
162381 97.038 -77.728 47.948
162382 97.044 -77.735 47.927
162383 97.050 -77.742 47.908
162384 97.056 -77.749 47.889
162385 97.062 -77.755 47.870
162386 97.078 -77.756 47.850
162387 97.103 -77.753 47.830
162388 97.129 -77.751 47.809
162389 97.154 -77.748 47.789
162390 97.180 -77.745 47.768
162391 97.205 -77.742 47.746
162392 97.231 -77.739 47.724
162393 97.256 -77.736 47.703
162394 97.282 -77.733 47.681
162395 97.307 -77.730 47.658
162396 97.330 -77.738 47.635
162397 97.354 -77.748 47.612
162398 97.377 -77.757 47.589
162399 97.401 -77.767 47.567
162400 97.425 -77.777 47.546
162401 97.448 -77.787 47.524
162402 97.471 -77.796 47.501
162403 97.495 -77.806 47.479
162404 97.524 -77.815 47.458
162405 97.555 -77.824 47.437
162406 97.585 -77.832 47.415
162407 97.615 -77.840 47.394
162408 97.642 -77.846 47.373
162409 97.650 -77.838 47.352
162410 97.659 -77.830 47.330
162411 97.668 -77.821 47.308
162412 97.677 -77.814 47.286
162413 97.686 -77.805 47.264
162414 97.695 -77.797 47.239
162415 97.709 -77.808 47.214
162416 97.713 -77.831 47.189
162417 97.724 -77.847 47.162
162418 97.742 -77.857 47.136
162419 97.759 -77.867 47.110
162420 97.778 -77.877 47.084
162421 97.796 -77.887 47.059
162422 97.808 -77.888 47.035
162423 97.831 -77.883 47.012
162424 97.851 -77.884 46.989
162425 97.871 -77.884 46.962
162426 97.891 -77.884 46.937
162427 97.911 -77.885 46.911
162428 97.931 -77.886 46.886
162429 97.948 -77.888 46.860
162430 97.961 -77.896 46.835
162431 97.974 -77.902 46.810
162432 97.987 -77.909 46.785
162433 98.000 -77.915 46.758
162434 98.013 -77.922 46.732
162435 98.027 -77.928 46.707
162436 98.040 -77.935 46.682
162437 98.051 -77.941 46.660
162438 98.046 -77.945 46.637
162439 98.070 -77.928 46.615
162440 98.157 -77.867 46.593
162441 98.185 -77.893 46.571
162442 98.213 -77.920 46.548
162443 98.218 -77.947 46.527
162444 98.224 -77.958 46.506
162445 98.231 -77.953 46.485
162446 98.238 -77.948 46.463
162447 98.245 -77.943 46.439
162448 98.285 -77.918 46.416
162449 98.324 -77.894 46.393
162450 98.364 -77.870 46.372
162451 98.404 -77.845 46.352
162452 98.448 -77.844 46.333
162453 98.491 -77.830 46.315
162454 98.521 -77.834 46.296
162455 98.524 -77.876 46.277
162456 98.535 -77.911 46.260
162457 98.558 -77.915 46.244
162458 98.572 -77.922 46.225
162459 98.586 -77.930 46.205
162460 98.600 -77.937 46.185
162461 98.615 -77.943 46.164
162462 98.629 -77.951 46.144
162463 98.643 -77.958 46.125
162464 98.657 -77.965 46.107
162465 98.671 -77.972 46.087
162466 98.686 -77.979 46.067
162467 98.700 -77.987 46.047
162468 98.714 -77.993 46.026
162469 98.728 -78.000 46.005
162470 98.742 -78.008 45.986
162471 98.758 -78.015 45.967
162472 98.777 -78.022 45.950
162473 98.795 -78.030 45.933
162474 98.834 -78.037 45.915
162475 98.884 -78.034 45.898
162476 98.939 -78.034 45.880
162477 98.995 -78.034 45.864
162478 99.050 -78.035 45.848
162479 99.105 -78.036 45.831
162480 99.160 -78.036 45.816
162481 99.215 -78.037 45.801
162482 99.270 -78.037 45.788
162483 99.301 -78.035 45.774
162484 99.319 -78.031 45.761
162485 99.338 -78.028 45.750
162486 99.356 -78.025 45.738
162487 99.374 -78.023 45.728
162488 99.384 -78.028 45.719
162489 99.394 -78.034 45.708
162490 99.404 -78.040 45.698
162491 99.414 -78.045 45.687
162492 99.425 -78.050 45.675
162493 99.435 -78.056 45.663
162494 99.444 -78.062 45.653
162495 99.453 -78.070 45.642
162496 99.462 -78.077 45.632
162497 99.482 -78.085 45.620
162498 99.513 -78.095 45.608
162499 99.546 -78.107 45.596
162500 99.585 -78.126 45.584
162501 99.624 -78.146 45.572
162502 99.658 -78.166 45.560
162503 99.686 -78.185 45.548
162504 99.713 -78.203 45.536
162505 99.729 -78.219 45.525
162506 99.747 -78.234 45.512
162507 99.763 -78.249 45.499
162508 99.780 -78.264 45.486
162509 99.797 -78.279 45.473
162510 99.814 -78.294 45.459
162511 99.831 -78.309 45.445
162512 99.848 -78.324 45.430
162513 99.855 -78.341 45.415
162514 99.871 -78.363 45.400
162515 99.908 -78.383 45.384
162516 99.944 -78.404 45.367
162517 99.980 -78.425 45.350
162518 100.016 -78.445 45.334
162519 100.053 -78.466 45.317
162520 100.054 -78.476 45.301
162521 100.049 -78.485 45.284
162522 100.044 -78.495 45.268
162523 100.039 -78.503 45.252
162524 100.034 -78.512 45.236
162525 100.031 -78.513 45.219
162526 100.058 -78.517 45.202
162527 100.101 -78.532 45.186
162528 100.151 -78.546 45.170
162529 100.202 -78.561 45.155
162530 100.253 -78.574 45.139
162531 100.300 -78.588 45.124
162532 100.337 -78.598 45.107
162533 100.373 -78.607 45.091
162534 100.410 -78.617 45.074
162535 100.455 -78.628 45.060
162536 100.500 -78.639 45.044
162537 100.544 -78.651 45.029
162538 100.589 -78.662 45.013
162539 100.634 -78.673 44.998
162540 100.678 -78.680 44.982
162541 100.701 -78.684 44.967
162542 100.725 -78.692 44.950
162543 100.750 -78.701 44.934
162544 100.775 -78.709 44.918
162545 100.799 -78.717 44.902
162546 100.824 -78.726 44.887
162547 100.852 -78.741 44.871
162548 100.883 -78.757 44.854
162549 100.913 -78.774 44.839
162550 100.944 -78.790 44.823
162551 100.974 -78.807 44.807
162552 101.004 -78.824 44.791
162553 101.035 -78.839 44.775
162554 101.065 -78.856 44.759
162555 101.086 -78.876 44.744
162556 101.107 -78.892 44.728
162557 101.145 -78.902 44.712
162558 101.184 -78.911 44.696
162559 101.222 -78.921 44.681
162560 101.261 -78.930 44.667
162561 101.281 -78.911 44.653
162562 101.303 -78.900 44.639
162563 101.328 -78.913 44.626
162564 101.352 -78.927 44.613
162565 101.372 -78.952 44.600
162566 101.388 -78.987 44.586
162567 101.405 -79.021 44.573
162568 101.446 -79.041 44.560
162569 101.488 -79.061 44.547
162570 101.531 -79.080 44.533
162571 101.573 -79.099 44.521
162572 101.615 -79.118 44.508
162573 101.657 -79.137 44.496
162574 101.674 -79.143 44.484
162796 92.439 -77.404 54.021
162797 92.458 -77.409 54.037
162798 92.478 -77.415 54.052
162799 92.497 -77.420 54.068
162800 92.516 -77.426 54.086
162801 92.535 -77.432 54.103
162802 92.555 -77.437 54.120
162803 92.574 -77.443 54.136
162804 92.593 -77.448 54.151
162805 92.612 -77.454 54.166
162806 92.631 -77.460 54.179
162807 92.651 -77.467 54.191
162808 92.674 -77.478 54.203
162809 92.698 -77.489 54.214
162810 92.720 -77.500 54.224
162811 92.743 -77.511 54.233
162812 92.767 -77.522 54.243
162813 92.782 -77.534 54.252
162814 92.789 -77.547 54.262
162815 92.807 -77.550 54.272
162816 92.837 -77.545 54.281
162817 92.860 -77.536 54.291
162818 92.877 -77.524 54.297
162819 92.900 -77.514 54.304
162820 92.929 -77.508 54.311
162821 92.941 -77.503 54.314
162822 92.950 -77.499 54.319
162823 92.960 -77.495 54.322
162824 92.971 -77.492 54.326
162825 92.980 -77.488 54.330
162826 92.990 -77.484 54.334
162827 93.000 -77.480 54.340
162828 93.010 -77.476 54.348
162829 93.020 -77.472 54.356
162830 93.035 -77.457 54.365
162831 93.051 -77.442 54.374
162832 93.067 -77.425 54.382
162833 93.082 -77.409 54.391
162834 93.098 -77.392 54.399
162835 93.115 -77.376 54.406
162836 93.130 -77.360 54.409
162837 93.146 -77.344 54.414
162838 93.162 -77.327 54.419
162839 93.178 -77.311 54.425
162840 93.194 -77.294 54.428
162841 93.204 -77.280 54.428
162842 93.214 -77.266 54.425
162843 93.229 -77.257 54.418
162844 93.246 -77.249 54.409
162845 93.263 -77.242 54.400
162846 93.281 -77.237 54.392
162847 93.298 -77.232 54.384
162848 93.316 -77.226 54.375
162849 93.333 -77.220 54.367
162850 93.351 -77.215 54.358
162851 93.368 -77.210 54.347
162852 93.386 -77.204 54.334
162853 93.404 -77.199 54.321
162854 93.421 -77.194 54.308
162855 93.442 -77.184 54.297
162856 93.466 -77.170 54.287
162857 93.489 -77.157 54.278
162858 93.512 -77.144 54.268
162859 93.536 -77.130 54.257
162860 93.559 -77.117 54.247
162861 93.582 -77.103 54.238
162862 93.606 -77.090 54.228
162863 93.630 -77.080 54.218
162864 93.651 -77.074 54.210
162865 93.664 -77.077 54.200
162866 93.683 -77.069 54.188
162867 93.698 -77.072 54.174
162868 93.712 -77.075 54.160
162869 93.727 -77.078 54.143
162870 93.741 -77.081 54.126
162871 93.754 -77.082 54.109
162872 93.762 -77.081 54.093
162873 93.770 -77.080 54.076
162874 93.778 -77.079 54.059
162875 93.787 -77.078 54.042
162876 93.794 -77.075 54.026
162877 93.799 -77.072 54.013
162878 93.805 -77.069 54.000
162879 93.811 -77.066 53.990
162880 93.817 -77.063 53.978
162881 93.836 -77.056 53.964
162882 93.860 -77.047 53.947
162883 93.877 -77.039 53.925
162884 93.893 -77.034 53.903
162885 93.908 -77.031 53.880
162886 93.923 -77.026 53.858
162887 93.938 -77.022 53.841
162888 93.956 -77.018 53.827
162889 93.980 -77.013 53.814
162890 94.002 -77.006 53.797
162891 94.024 -77.000 53.779
162892 94.045 -76.994 53.760
162893 94.068 -76.988 53.740
162894 94.092 -76.984 53.720
162895 94.116 -76.979 53.699
162896 94.131 -76.976 53.676
162897 94.144 -76.974 53.651
162898 94.157 -76.971 53.631
162899 94.170 -76.969 53.612
162900 94.183 -76.965 53.593
162901 94.195 -76.962 53.574
162902 94.207 -76.959 53.554
162903 94.222 -76.957 53.535
162904 94.238 -76.956 53.516
162905 94.253 -76.953 53.498
162906 94.269 -76.950 53.484
162907 94.284 -76.947 53.473
162908 94.297 -76.943 53.465
162909 94.310 -76.939 53.459
162910 94.322 -76.935 53.452
162911 94.332 -76.931 53.444
162912 94.333 -76.931 53.439
162913 94.336 -76.933 53.436
162914 94.347 -76.933 53.431
162915 94.364 -76.931 53.424
162916 94.382 -76.928 53.414
162917 94.399 -76.924 53.400
162918 94.416 -76.919 53.384
162919 94.433 -76.915 53.366
162920 94.450 -76.911 53.347
162921 94.467 -76.907 53.327
162922 94.483 -76.904 53.306
162923 94.498 -76.902 53.283
162924 94.513 -76.900 53.255
162925 94.528 -76.898 53.227
162926 94.544 -76.896 53.201
162927 94.565 -76.892 53.175
162928 94.574 -76.886 53.148
162929 94.584 -76.881 53.119
162930 94.595 -76.875 53.090
162931 94.611 -76.877 53.057
162932 94.632 -76.884 53.024
162933 94.653 -76.892 52.990
162934 94.680 -76.893 52.959
162935 94.713 -76.886 52.926
162936 94.747 -76.879 52.892
162937 94.780 -76.872 52.857
162938 94.815 -76.865 52.822
162939 94.849 -76.856 52.789
162940 94.884 -76.849 52.756
162941 94.897 -76.849 52.722
162942 94.902 -76.850 52.687
162943 94.912 -76.850 52.654
162944 94.923 -76.847 52.620
162945 94.933 -76.846 52.585
162946 94.944 -76.844 52.555
162947 94.955 -76.843 52.528
162948 94.965 -76.841 52.505
162949 94.976 -76.839 52.487
162950 94.986 -76.837 52.471
162951 94.999 -76.838 52.456
162952 95.020 -76.844 52.441
162953 95.041 -76.850 52.425
162954 95.062 -76.856 52.404
162955 95.084 -76.862 52.376
162956 95.105 -76.868 52.340
162957 95.126 -76.875 52.296
162958 95.147 -76.881 52.247
162959 95.169 -76.887 52.195
162960 95.190 -76.893 52.140
162961 95.220 -76.903 52.081
162962 95.257 -76.915 52.020
162963 95.293 -76.928 51.957
162964 95.311 -76.946 51.894
162965 95.294 -76.975 51.833
162966 95.296 -76.993 51.772
162967 95.310 -77.002 51.714
162968 95.323 -77.012 51.657
162969 95.337 -77.021 51.599
162970 95.351 -77.030 51.541
162971 95.365 -77.040 51.484
162972 95.379 -77.049 51.429
162973 95.393 -77.058 51.376
162974 95.410 -77.070 51.322
162975 95.426 -77.081 51.273
162976 95.444 -77.094 51.225
162977 95.461 -77.105 51.177
162978 95.477 -77.117 51.130
162979 95.494 -77.128 51.085
162980 95.512 -77.144 51.040
162981 95.530 -77.160 50.996
162982 95.548 -77.177 50.952
162983 95.565 -77.191 50.909
162984 95.580 -77.201 50.866
162985 95.601 -77.198 50.820
162986 95.621 -77.196 50.775
162987 95.641 -77.194 50.729
162988 95.662 -77.192 50.683
162989 95.682 -77.190 50.638
162990 95.703 -77.188 50.596
162991 95.723 -77.186 50.555
162992 95.744 -77.186 50.514
162993 95.758 -77.192 50.474
162994 95.765 -77.198 50.436
162995 95.772 -77.204 50.398
162996 95.779 -77.210 50.359
162997 95.787 -77.216 50.316
162998 95.794 -77.222 50.271
162999 95.808 -77.231 50.224
163000 95.828 -77.241 50.179
163001 95.858 -77.252 50.134
163002 95.891 -77.263 50.088
163003 95.924 -77.273 50.041
163004 95.957 -77.284 49.994
163005 95.978 -77.291 49.945
163006 95.980 -77.290 49.897
163007 95.974 -77.286 49.848
163008 95.967 -77.284 49.797
163009 95.959 -77.291 49.745
163010 95.952 -77.300 49.691
163011 95.944 -77.307 49.636
163012 95.936 -77.316 49.583
163013 95.927 -77.323 49.534
163014 95.919 -77.332 49.487
163015 95.918 -77.340 49.441
163016 95.930 -77.351 49.398
163017 95.942 -77.361 49.356
163018 95.959 -77.371 49.315
163019 95.980 -77.380 49.277
163020 96.000 -77.390 49.242
163021 96.020 -77.399 49.209
163022 96.041 -77.409 49.177
163023 96.061 -77.418 49.146
163024 96.081 -77.428 49.116
163025 96.102 -77.437 49.087
163026 96.122 -77.447 49.060
163027 96.142 -77.457 49.037
163028 96.163 -77.465 49.016
163029 96.184 -77.473 48.996
163030 96.204 -77.480 48.977
163031 96.225 -77.482 48.955
163032 96.245 -77.483 48.933
163033 96.265 -77.483 48.912
163034 96.286 -77.484 48.893
163035 96.306 -77.485 48.874
163036 96.324 -77.486 48.855
163037 96.342 -77.486 48.837
163038 96.361 -77.487 48.823
163039 96.379 -77.488 48.810
163040 96.397 -77.488 48.800
163041 96.419 -77.492 48.790
163042 96.448 -77.501 48.779
163043 96.465 -77.523 48.767
163044 96.482 -77.545 48.752
163045 96.500 -77.566 48.736
163046 96.517 -77.588 48.720
163047 96.534 -77.609 48.704
163048 96.551 -77.630 48.690
163049 96.568 -77.652 48.675
163050 96.585 -77.674 48.658
163051 96.599 -77.681 48.640
163052 96.611 -77.679 48.621
163053 96.622 -77.677 48.602
163054 96.635 -77.675 48.584
163055 96.647 -77.674 48.566
163056 96.658 -77.672 48.551
163057 96.670 -77.670 48.534
163058 96.682 -77.668 48.517
163059 96.694 -77.666 48.501
163060 96.705 -77.664 48.484
163061 96.718 -77.662 48.466
163062 96.729 -77.661 48.448
163063 96.737 -77.661 48.429
163064 96.744 -77.663 48.409
163065 96.751 -77.664 48.388
163066 96.761 -77.668 48.368
163067 96.775 -77.671 48.343
163068 96.789 -77.674 48.319
163069 96.807 -77.676 48.294
163070 96.828 -77.679 48.267
163071 96.850 -77.682 48.240
163072 96.871 -77.685 48.214
163073 96.892 -77.688 48.187
163074 96.913 -77.690 48.160
163075 96.935 -77.693 48.134
163076 96.956 -77.696 48.108
163077 96.977 -77.699 48.084
163078 96.999 -77.702 48.060
163079 97.019 -77.704 48.036
163080 97.033 -77.707 48.013
163081 97.039 -77.712 47.990
163082 97.045 -77.717 47.967
163083 97.052 -77.722 47.946
163084 97.058 -77.727 47.926
163085 97.065 -77.732 47.908
163086 97.074 -77.738 47.889
163087 97.102 -77.746 47.869
163088 97.129 -77.755 47.850
163089 97.155 -77.752 47.829
163090 97.180 -77.749 47.809
163091 97.206 -77.746 47.788
163092 97.231 -77.743 47.766
163093 97.256 -77.740 47.744
163094 97.282 -77.736 47.722
163095 97.307 -77.733 47.700
163096 97.332 -77.730 47.677
163097 97.356 -77.740 47.653
163098 97.379 -77.749 47.630
163099 97.403 -77.759 47.607
163100 97.427 -77.769 47.584
163101 97.450 -77.779 47.562
163102 97.473 -77.789 47.539
163103 97.497 -77.799 47.515
163104 97.520 -77.808 47.493
163105 97.544 -77.818 47.471
163106 97.572 -77.827 47.450
163107 97.602 -77.836 47.428
163108 97.632 -77.844 47.407
163109 97.660 -77.851 47.385
163110 97.669 -77.843 47.364
163111 97.678 -77.835 47.343
163112 97.686 -77.827 47.321
163113 97.696 -77.819 47.298
163114 97.704 -77.811 47.274
163115 97.713 -77.802 47.248
163116 97.722 -77.796 47.223
163117 97.738 -77.808 47.197
163118 97.756 -77.818 47.172
163119 97.774 -77.828 47.146
163120 97.792 -77.838 47.120
163121 97.810 -77.848 47.094
163122 97.828 -77.858 47.069
163123 97.846 -77.853 47.046
163124 97.865 -77.854 47.022
163125 97.885 -77.854 47.000
163126 97.905 -77.855 46.973
163127 97.925 -77.856 46.947
163128 97.934 -77.870 46.922
163129 97.947 -77.876 46.896
163130 97.964 -77.877 46.870
163131 97.978 -77.883 46.844
163132 97.988 -77.893 46.818
163133 97.999 -77.902 46.791
163134 98.010 -77.911 46.764
163135 98.023 -77.917 46.738
163136 98.036 -77.924 46.713
163137 98.050 -77.931 46.688
163138 98.063 -77.937 46.664
163139 98.063 -77.941 46.641
163140 98.113 -77.906 46.617
163141 98.189 -77.854 46.594
163142 98.200 -77.868 46.571
163143 98.205 -77.896 46.548
163144 98.211 -77.908 46.527
163145 98.218 -77.902 46.506
163146 98.225 -77.898 46.485
163147 98.232 -77.893 46.464
163148 98.239 -77.889 46.441
163149 98.245 -77.884 46.418
163150 98.252 -77.879 46.396
163151 98.284 -77.859 46.375
163152 98.323 -77.837 46.357
163153 98.369 -77.849 46.338
163154 98.414 -77.862 46.321
163155 98.458 -77.869 46.302
163156 98.483 -77.881 46.283
163157 98.511 -77.894 46.266
163158 98.525 -77.901 46.249
163159 98.539 -77.908 46.232
163160 98.554 -77.915 46.213
163161 98.568 -77.922 46.192
163162 98.582 -77.929 46.173
163163 98.596 -77.937 46.153
163164 98.611 -77.943 46.133
163165 98.630 -77.951 46.114
163166 98.648 -77.959 46.096
163167 98.667 -77.965 46.076
163168 98.686 -77.973 46.056
163169 98.704 -77.981 46.035
163170 98.723 -77.988 46.014
163171 98.741 -77.995 45.994
163172 98.760 -78.003 45.976
163173 98.779 -78.010 45.958
163174 98.797 -78.018 45.941
163175 98.848 -78.027 45.923
163176 98.914 -78.030 45.906
163177 98.959 -78.022 45.888
163178 99.010 -78.018 45.871
163179 99.066 -78.019 45.855
163180 99.121 -78.019 45.837
163181 99.176 -78.020 45.823
163182 99.231 -78.020 45.809
163183 99.286 -78.021 45.795
163184 99.313 -78.018 45.781
163185 99.332 -78.015 45.768
163186 99.350 -78.012 45.756
163187 99.366 -78.012 45.746
163188 99.376 -78.018 45.736
163189 99.386 -78.023 45.725
163190 99.397 -78.028 45.714
163191 99.407 -78.034 45.703
163192 99.416 -78.040 45.693
163193 99.425 -78.047 45.681
163194 99.433 -78.055 45.670
163195 99.442 -78.062 45.660
163196 99.450 -78.069 45.650
163197 99.459 -78.077 45.638
163198 99.484 -78.085 45.626
163199 99.515 -78.095 45.614
163200 99.547 -78.105 45.602
163201 99.578 -78.114 45.591
163202 99.612 -78.127 45.579
163203 99.650 -78.146 45.567
163204 99.680 -78.166 45.555
163205 99.700 -78.182 45.544
163206 99.717 -78.197 45.532
163207 99.733 -78.212 45.519
163208 99.751 -78.227 45.506
163209 99.767 -78.242 45.492
163210 99.784 -78.257 45.479
163211 99.801 -78.272 45.465
163212 99.818 -78.287 45.450
163213 99.835 -78.302 45.435
163214 99.852 -78.317 45.420
163215 99.856 -78.337 45.404
163216 99.893 -78.357 45.389
163217 99.929 -78.378 45.372
163218 99.965 -78.399 45.355
163219 99.999 -78.419 45.339
163220 99.994 -78.428 45.321
163221 99.989 -78.437 45.306
163222 99.984 -78.445 45.289
163223 99.979 -78.454 45.272
163224 99.974 -78.463 45.256
163225 99.985 -78.479 45.239
163226 100.021 -78.495 45.223
163227 100.072 -78.509 45.206
163228 100.123 -78.523 45.190
163229 100.173 -78.537 45.174
163230 100.224 -78.551 45.159
163231 100.272 -78.565 45.144
163232 100.309 -78.575 45.128
163233 100.346 -78.585 45.112
163234 100.382 -78.594 45.096
163235 100.419 -78.604 45.080
163236 100.459 -78.614 45.064
163237 100.504 -78.626 45.049
163238 100.549 -78.637 45.033
163239 100.593 -78.648 45.018
163240 100.638 -78.659 45.002
163241 100.676 -78.655 44.986
163242 100.701 -78.664 44.971
163243 100.725 -78.672 44.954
163244 100.750 -78.680 44.939
163245 100.774 -78.689 44.923
163246 100.798 -78.697 44.907
163247 100.826 -78.708 44.892
163248 100.856 -78.721 44.876
163249 100.886 -78.738 44.860
163250 100.916 -78.755 44.845
163251 100.947 -78.771 44.829
163252 100.977 -78.788 44.813
163253 101.007 -78.804 44.797
163254 101.038 -78.821 44.782
163255 101.067 -78.823 44.766
163256 101.085 -78.847 44.750
163257 101.096 -78.880 44.735
163258 101.140 -78.894 44.719
163259 101.181 -78.905 44.704
163260 101.219 -78.914 44.689
163261 101.258 -78.924 44.675
163262 101.282 -78.911 44.661
163263 101.303 -78.891 44.647
163264 101.327 -78.899 44.633
163265 101.351 -78.912 44.619
163266 101.370 -78.940 44.605
163267 101.387 -78.974 44.592
163268 101.404 -79.009 44.579
163269 101.446 -79.028 44.566
163270 101.488 -79.046 44.553
163271 101.530 -79.066 44.539
163272 101.572 -79.085 44.527
163273 101.615 -79.105 44.514
163274 101.657 -79.124 44.502
163275 101.675 -79.130 44.491
163276 101.691 -79.135 44.479
163497 92.420 -77.364 54.000
163498 92.439 -77.370 54.016
163499 92.458 -77.376 54.033
163500 92.477 -77.381 54.050
163501 92.496 -77.386 54.068
163502 92.515 -77.392 54.087
163503 92.534 -77.398 54.105
163504 92.554 -77.403 54.122
163505 92.573 -77.409 54.138
163506 92.592 -77.416 54.153
163507 92.616 -77.426 54.169
163508 92.639 -77.438 54.182
163509 92.662 -77.448 54.194
163510 92.685 -77.460 54.205
163511 92.708 -77.471 54.215
163512 92.731 -77.482 54.225
163513 92.754 -77.493 54.235
163514 92.775 -77.504 54.245
163515 92.782 -77.517 54.254
163516 92.789 -77.530 54.263
163517 92.804 -77.536 54.273
163518 92.834 -77.530 54.282
163519 92.855 -77.520 54.290
163520 92.872 -77.509 54.298
163521 92.900 -77.501 54.305
163522 92.929 -77.494 54.311
163523 92.958 -77.487 54.317
163524 92.972 -77.483 54.323
163525 92.981 -77.479 54.328
163526 92.991 -77.475 54.332
163527 93.001 -77.471 54.337
163528 93.011 -77.467 54.343
163529 93.021 -77.463 54.350
163530 93.031 -77.458 54.358
163531 93.047 -77.442 54.368
163532 93.063 -77.426 54.377
163533 93.079 -77.410 54.386
163534 93.095 -77.393 54.395
163535 93.111 -77.377 54.402
163536 93.127 -77.360 54.408
163537 93.142 -77.344 54.412
163538 93.158 -77.328 54.416
163539 93.168 -77.313 54.420
163540 93.179 -77.299 54.425
163541 93.188 -77.285 54.429
163542 93.198 -77.271 54.428
163543 93.208 -77.257 54.423
163544 93.223 -77.247 54.416
163545 93.240 -77.239 54.407
163546 93.258 -77.232 54.398
163547 93.274 -77.225 54.391
163548 93.291 -77.217 54.384
163549 93.308 -77.210 54.377
163550 93.326 -77.205 54.370
163551 93.344 -77.200 54.361
163552 93.361 -77.194 54.351
163553 93.379 -77.188 54.341
163554 93.397 -77.184 54.328
163555 93.414 -77.179 54.316
163556 93.437 -77.165 54.305
163557 93.461 -77.152 54.294
163558 93.484 -77.138 54.284
163559 93.507 -77.125 54.273
163560 93.531 -77.112 54.263
163561 93.555 -77.101 54.253
163562 93.578 -77.091 54.244
163563 93.603 -77.081 54.235
163564 93.626 -77.072 54.226
163565 93.639 -77.078 54.219
163566 93.647 -77.090 54.211
163567 93.666 -77.082 54.201
163568 93.686 -77.075 54.188
163569 93.704 -77.068 54.175
163570 93.723 -77.062 54.158
163571 93.737 -77.065 54.142
163572 93.746 -77.064 54.127
163573 93.755 -77.063 54.112
163574 93.763 -77.062 54.096
163575 93.771 -77.060 54.080
163576 93.779 -77.059 54.064
163577 93.787 -77.058 54.048
163578 93.795 -77.057 54.034
163579 93.803 -77.056 54.022
163580 93.812 -77.055 54.011
163581 93.817 -77.051 54.000
163582 93.830 -77.046 53.987
163583 93.849 -77.038 53.972
163584 93.865 -77.031 53.951
163585 93.881 -77.026 53.928
163586 93.896 -77.022 53.906
163587 93.911 -77.018 53.882
163588 93.932 -77.013 53.865
163589 93.957 -77.009 53.849
163590 93.980 -77.004 53.832
163591 94.002 -76.997 53.814
163592 94.024 -76.991 53.794
163593 94.046 -76.984 53.775
163594 94.067 -76.978 53.758
163595 94.089 -76.972 53.739
163596 94.109 -76.965 53.720
163597 94.121 -76.962 53.699
163598 94.134 -76.959 53.676
163599 94.147 -76.956 53.657
163600 94.160 -76.954 53.639
163601 94.174 -76.952 53.621
163602 94.186 -76.949 53.603
163603 94.198 -76.946 53.584
163604 94.214 -76.943 53.563
163605 94.229 -76.941 53.541
163606 94.245 -76.938 53.522
163607 94.260 -76.936 53.505
163608 94.276 -76.933 53.490
163609 94.288 -76.929 53.479
163610 94.300 -76.925 53.472
163611 94.313 -76.920 53.464
163612 94.325 -76.916 53.456
163613 94.329 -76.915 53.449
163614 94.334 -76.918 53.444
163615 94.346 -76.918 53.438
163616 94.363 -76.915 53.430
163617 94.381 -76.912 53.422
163618 94.398 -76.909 53.410
163619 94.415 -76.906 53.397
163620 94.432 -76.901 53.383
163621 94.449 -76.897 53.366
163622 94.466 -76.893 53.347
163623 94.481 -76.891 53.325
163624 94.496 -76.889 53.302
163625 94.511 -76.887 53.275
163626 94.526 -76.885 53.249
163627 94.541 -76.884 53.224
163628 94.554 -76.880 53.198
163629 94.564 -76.875 53.170
163630 94.575 -76.869 53.141
163631 94.585 -76.864 53.112
163632 94.605 -76.870 53.079
163633 94.626 -76.878 53.046
163634 94.658 -76.872 53.013
163635 94.691 -76.865 52.983
163636 94.724 -76.859 52.949
163637 94.758 -76.852 52.915
163638 94.791 -76.844 52.880
163639 94.826 -76.837 52.845
163640 94.860 -76.829 52.811
163641 94.886 -76.824 52.776
163642 94.891 -76.827 52.741
163643 94.902 -76.825 52.704
163644 94.912 -76.823 52.670
163645 94.923 -76.821 52.636
163646 94.933 -76.820 52.599
163647 94.944 -76.818 52.567
163648 94.955 -76.817 52.537
163649 94.965 -76.815 52.512
163650 94.975 -76.813 52.491
163651 94.986 -76.812 52.475
163652 94.996 -76.810 52.461
163653 95.016 -76.815 52.446
163654 95.037 -76.821 52.430
163655 95.058 -76.828 52.409
163656 95.080 -76.834 52.383
163657 95.101 -76.840 52.349
163658 95.122 -76.846 52.307
163659 95.143 -76.853 52.260
163660 95.165 -76.859 52.209
163661 95.190 -76.866 52.154
163662 95.227 -76.878 52.094
163663 95.264 -76.890 52.032
163664 95.301 -76.903 51.968
163665 95.322 -76.920 51.907
163666 95.305 -76.950 51.848
163667 95.291 -76.978 51.788
163668 95.305 -76.987 51.729
163669 95.319 -76.996 51.671
163670 95.333 -77.006 51.612
163671 95.346 -77.015 51.554
163672 95.360 -77.024 51.497
163673 95.375 -77.034 51.443
163674 95.390 -77.044 51.390
163675 95.407 -77.056 51.338
163676 95.424 -77.067 51.288
163677 95.441 -77.079 51.238
163678 95.458 -77.091 51.189
163679 95.475 -77.103 51.142
163680 95.491 -77.114 51.095
163681 95.509 -77.128 51.051
163682 95.527 -77.144 51.007
163683 95.545 -77.161 50.964
163684 95.563 -77.178 50.923
163685 95.582 -77.185 50.881
163686 95.602 -77.182 50.835
163687 95.623 -77.181 50.789
163688 95.643 -77.179 50.744
163689 95.664 -77.176 50.698
163690 95.684 -77.175 50.653
163691 95.704 -77.172 50.610
163692 95.725 -77.170 50.568
163693 95.746 -77.169 50.528
163694 95.760 -77.169 50.489
163695 95.767 -77.175 50.450
163696 95.774 -77.182 50.412
163697 95.781 -77.188 50.373
163698 95.788 -77.194 50.332
163699 95.795 -77.200 50.285
163700 95.802 -77.206 50.238
163701 95.822 -77.216 50.192
163702 95.854 -77.227 50.148
163703 95.887 -77.238 50.103
163704 95.920 -77.248 50.057
163705 95.953 -77.258 50.010
163706 95.977 -77.266 49.963
163707 95.971 -77.263 49.915
163708 95.963 -77.270 49.866
163709 95.955 -77.279 49.815
163710 95.948 -77.286 49.762
163711 95.940 -77.294 49.707
163712 95.932 -77.302 49.650
163713 95.923 -77.310 49.596
163714 95.915 -77.318 49.545
163715 95.907 -77.326 49.496
163716 95.904 -77.335 49.450
163717 95.916 -77.345 49.406
163718 95.934 -77.354 49.364
163719 95.954 -77.364 49.323
163720 95.974 -77.373 49.283
163721 95.995 -77.383 49.248
163722 96.015 -77.392 49.215
163723 96.036 -77.401 49.182
163724 96.056 -77.410 49.151
163725 96.078 -77.417 49.121
163726 96.098 -77.425 49.091
163727 96.119 -77.432 49.064
163728 96.140 -77.440 49.041
163729 96.161 -77.448 49.020
163730 96.182 -77.456 49.000
163731 96.202 -77.463 48.980
163732 96.223 -77.464 48.957
163733 96.243 -77.464 48.934
163734 96.264 -77.465 48.911
163735 96.284 -77.466 48.893
163736 96.304 -77.467 48.875
163737 96.325 -77.467 48.857
163738 96.345 -77.468 48.840
163739 96.365 -77.469 48.825
163740 96.386 -77.470 48.813
163741 96.406 -77.470 48.805
163742 96.428 -77.474 48.798
163743 96.443 -77.493 48.788
163744 96.460 -77.514 48.776
163745 96.477 -77.536 48.762
163746 96.494 -77.558 48.747
163747 96.511 -77.579 48.731
163748 96.528 -77.601 48.717
163749 96.545 -77.622 48.702
163750 96.562 -77.643 48.688
163751 96.580 -77.664 48.671
163752 96.596 -77.681 48.652
163753 96.607 -77.680 48.633
163754 96.619 -77.678 48.614
163755 96.631 -77.676 48.596
163756 96.643 -77.674 48.579
163757 96.655 -77.672 48.564
163758 96.667 -77.671 48.546
163759 96.679 -77.668 48.528
163760 96.690 -77.667 48.510
163761 96.702 -77.665 48.492
163762 96.714 -77.663 48.475
163763 96.726 -77.661 48.458
163764 96.738 -77.659 48.440
163765 96.750 -77.658 48.422
163766 96.761 -77.656 48.404
163767 96.769 -77.658 48.385
163768 96.786 -77.661 48.361
163769 96.807 -77.664 48.336
163770 96.829 -77.667 48.310
163771 96.849 -77.669 48.284
163772 96.869 -77.671 48.258
163773 96.889 -77.673 48.232
163774 96.909 -77.675 48.206
163775 96.930 -77.677 48.179
163776 96.950 -77.679 48.153
163777 96.970 -77.680 48.127
163778 96.990 -77.683 48.103
163779 97.010 -77.684 48.080
163780 97.030 -77.686 48.057
163781 97.042 -77.689 48.033
163782 97.048 -77.693 48.010
163783 97.054 -77.699 47.987
163784 97.061 -77.704 47.965
163785 97.067 -77.708 47.945
163786 97.074 -77.714 47.926
163787 97.096 -77.722 47.907
163788 97.124 -77.730 47.888
163789 97.152 -77.739 47.870
163790 97.179 -77.747 47.849
163791 97.206 -77.749 47.829
163792 97.231 -77.746 47.808
163793 97.256 -77.743 47.786
163794 97.282 -77.740 47.764
163795 97.307 -77.737 47.742
163796 97.332 -77.734 47.718
163797 97.358 -77.732 47.694
163798 97.381 -77.742 47.670
163799 97.405 -77.752 47.647
163800 97.428 -77.761 47.623
163801 97.452 -77.771 47.599
163802 97.475 -77.781 47.577
163803 97.499 -77.791 47.554
163804 97.522 -77.801 47.529
163805 97.546 -77.811 47.506
163806 97.569 -77.820 47.483
163807 97.593 -77.830 47.461
163808 97.618 -77.840 47.438
163809 97.649 -77.848 47.417
163810 97.678 -77.856 47.396
163811 97.687 -77.848 47.374
163812 97.696 -77.840 47.353
163813 97.705 -77.832 47.330
163814 97.714 -77.824 47.307
163815 97.722 -77.814 47.283
163816 97.723 -77.787 47.257
163817 97.726 -77.763 47.232
163818 97.759 -77.786 47.206
163819 97.784 -77.791 47.180
163820 97.806 -77.799 47.154
163821 97.824 -77.809 47.127
163822 97.842 -77.819 47.101
163823 97.858 -77.825 47.077
163824 97.879 -77.824 47.053
163825 97.899 -77.824 47.031
163826 97.911 -77.836 47.008
163827 97.920 -77.849 46.981
163828 97.930 -77.863 46.955
163829 97.946 -77.864 46.929
163830 97.963 -77.865 46.904
163831 97.980 -77.865 46.877
163832 97.996 -77.868 46.851
163833 98.007 -77.877 46.824
163834 98.017 -77.887 46.798
163835 98.027 -77.897 46.770
163836 98.038 -77.907 46.744
163837 98.048 -77.917 46.719
163838 98.060 -77.925 46.694
163839 98.073 -77.932 46.670
163840 98.079 -77.937 46.646
163841 98.135 -77.902 46.621
163842 98.152 -77.901 46.597
163843 98.176 -77.893 46.573
163844 98.208 -77.879 46.550
163845 98.216 -77.873 46.528
163846 98.223 -77.868 46.507
163847 98.230 -77.863 46.486
163848 98.236 -77.858 46.465
163849 98.243 -77.852 46.443
163850 98.249 -77.847 46.421
163851 98.256 -77.842 46.399
163852 98.263 -77.837 46.379
163853 98.282 -77.837 46.361
163854 98.326 -77.846 46.343
163855 98.371 -77.857 46.325
163856 98.415 -77.868 46.307
163857 98.460 -77.879 46.289
163858 98.483 -77.880 46.271
163859 98.502 -77.887 46.254
163860 98.520 -77.895 46.237
163861 98.538 -77.902 46.219
163862 98.557 -77.909 46.200
163863 98.576 -77.917 46.180
163864 98.594 -77.924 46.160
163865 98.613 -77.931 46.141
163866 98.632 -77.939 46.123
163867 98.650 -77.946 46.104
163868 98.669 -77.954 46.084
163869 98.688 -77.961 46.064
163870 98.706 -77.968 46.043
163871 98.725 -77.976 46.022
163872 98.743 -77.983 46.003
163873 98.762 -77.990 45.984
163874 98.781 -77.998 45.966
163875 98.806 -78.006 45.949
163876 98.861 -78.015 45.931
163877 98.942 -78.027 45.914
163878 98.995 -78.022 45.896
163879 99.040 -78.014 45.879
163880 99.084 -78.006 45.862
163881 99.137 -78.003 45.845
163882 99.192 -78.004 45.829
163883 99.247 -78.004 45.814
163884 99.302 -78.005 45.800
163885 99.326 -78.002 45.786
163886 99.345 -77.999 45.773
163887 99.358 -78.001 45.762
163888 99.368 -78.006 45.752
163889 99.378 -78.012 45.741
163890 99.388 -78.018 45.730
163891 99.397 -78.025 45.719
163892 99.405 -78.033 45.709
163893 99.414 -78.040 45.699
163894 99.422 -78.047 45.688
163895 99.431 -78.055 45.676
163896 99.439 -78.062 45.666
163897 99.448 -78.069 45.657
163898 99.457 -78.076 45.645
163899 99.486 -78.086 45.631
163900 99.517 -78.095 45.620
163901 99.549 -78.105 45.609
163902 99.580 -78.115 45.598
163903 99.612 -78.124 45.586
163904 99.643 -78.134 45.574
163905 99.670 -78.145 45.564
163906 99.687 -78.160 45.553
163907 99.704 -78.175 45.540
163908 99.721 -78.190 45.527
163909 99.737 -78.205 45.513
163910 99.755 -78.220 45.499
163911 99.771 -78.235 45.485
163912 99.788 -78.250 45.470
163913 99.805 -78.265 45.454
163914 99.822 -78.280 45.439
163915 99.839 -78.295 45.425
163916 99.853 -78.311 45.409
163917 99.877 -78.332 45.393
163918 99.914 -78.352 45.377
163919 99.940 -78.370 45.360
163920 99.935 -78.379 45.343
163921 99.930 -78.388 45.326
163922 99.925 -78.397 45.309
163923 99.944 -78.415 45.292
163924 99.964 -78.433 45.275
163925 99.985 -78.451 45.258
163926 100.005 -78.470 45.242
163927 100.043 -78.485 45.225
163928 100.093 -78.500 45.209
163929 100.144 -78.514 45.193
163930 100.195 -78.529 45.177
163931 100.244 -78.542 45.162
163932 100.281 -78.552 45.147
163933 100.318 -78.562 45.131
163934 100.354 -78.571 45.115
163935 100.391 -78.581 45.100
163936 100.428 -78.591 45.084
163937 100.465 -78.601 45.068
163938 100.508 -78.611 45.053
163939 100.553 -78.623 45.037
163940 100.598 -78.634 45.022
163941 100.642 -78.645 45.006
163942 100.676 -78.635 44.990
163943 100.700 -78.643 44.975
163944 100.725 -78.651 44.959
163945 100.749 -78.660 44.943
163946 100.774 -78.668 44.928
163947 100.801 -78.679 44.912
163948 100.830 -78.692 44.895
163949 100.859 -78.704 44.880
163950 100.889 -78.719 44.864
163951 100.920 -78.736 44.850
163952 100.950 -78.752 44.834
163953 100.980 -78.768 44.818
163954 101.010 -78.781 44.802
163955 101.039 -78.783 44.787
163956 101.068 -78.786 44.771
163957 101.090 -78.803 44.756
163958 101.092 -78.858 44.741
163959 101.121 -78.880 44.725
163960 101.168 -78.893 44.710
163961 101.215 -78.908 44.695
163962 101.255 -78.917 44.681
163963 101.283 -78.910 44.667
163964 101.304 -78.891 44.654
163965 101.326 -78.883 44.639
163966 101.350 -78.897 44.625
163967 101.369 -78.927 44.611
163968 101.385 -78.961 44.598
163969 101.403 -78.994 44.584
163970 101.446 -79.014 44.571
163971 101.488 -79.033 44.558
163972 101.530 -79.053 44.545
163973 101.572 -79.071 44.533
163974 101.614 -79.090 44.520
163975 101.656 -79.110 44.508
163976 101.676 -79.117 44.496
163977 101.692 -79.122 44.485
164199 92.419 -77.330 53.995
164200 92.438 -77.335 54.013
164201 92.457 -77.341 54.030
164202 92.476 -77.347 54.050
164203 92.495 -77.353 54.070
164204 92.514 -77.358 54.089
164205 92.534 -77.364 54.107
164206 92.557 -77.376 54.124
164207 92.580 -77.386 54.140
164208 92.603 -77.398 54.157
164209 92.626 -77.408 54.171
164210 92.649 -77.420 54.184
164211 92.672 -77.431 54.196
164212 92.695 -77.442 54.207
164213 92.718 -77.453 54.217
164214 92.741 -77.464 54.227
164215 92.765 -77.475 54.236
164216 92.775 -77.487 54.246
164217 92.782 -77.500 54.255
164218 92.789 -77.512 54.265
164219 92.801 -77.521 54.274
164220 92.831 -77.516 54.282
164221 92.851 -77.505 54.290
164222 92.872 -77.495 54.298
164223 92.901 -77.488 54.306
164224 92.929 -77.481 54.314
164225 92.958 -77.473 54.322
164226 92.987 -77.467 54.329
164227 93.002 -77.462 54.334
164228 93.012 -77.458 54.340
164229 93.022 -77.454 54.344
164230 93.031 -77.450 54.351
164231 93.043 -77.443 54.360
164232 93.059 -77.426 54.369
164233 93.075 -77.410 54.377
164234 93.091 -77.394 54.387
164235 93.107 -77.377 54.395
164236 93.123 -77.361 54.402
164237 93.133 -77.347 54.407
164238 93.143 -77.332 54.411
164239 93.153 -77.318 54.415
164240 93.162 -77.304 54.418
164241 93.172 -77.290 54.421
164242 93.183 -77.276 54.423
164243 93.192 -77.261 54.422
164244 93.202 -77.248 54.415
164245 93.218 -77.237 54.406
164246 93.235 -77.229 54.397
164247 93.252 -77.222 54.389
164248 93.269 -77.215 54.383
164249 93.286 -77.207 54.377
164250 93.303 -77.200 54.371
164251 93.320 -77.192 54.365
164252 93.337 -77.185 54.358
164253 93.354 -77.179 54.351
164254 93.371 -77.173 54.343
164255 93.389 -77.168 54.334
164256 93.409 -77.160 54.323
164257 93.432 -77.147 54.311
164258 93.455 -77.133 54.299
164259 93.480 -77.122 54.287
164260 93.503 -77.113 54.275
164261 93.528 -77.103 54.264
164262 93.551 -77.093 54.254
164263 93.575 -77.083 54.245
164264 93.599 -77.073 54.237
164265 93.623 -77.063 54.230
164266 93.627 -77.082 54.224
164267 93.631 -77.102 54.218
164268 93.650 -77.095 54.210
164269 93.671 -77.084 54.199
164270 93.692 -77.072 54.186
164271 93.712 -77.060 54.171
164272 93.729 -77.053 54.157
164273 93.736 -77.044 54.143
164274 93.744 -77.042 54.127
164275 93.752 -77.041 54.113
164276 93.760 -77.040 54.098
164277 93.769 -77.039 54.083
164278 93.777 -77.038 54.067
164279 93.785 -77.037 54.053
164280 93.793 -77.035 54.040
164281 93.801 -77.034 54.029
164282 93.809 -77.034 54.019
164283 93.820 -77.034 54.008
164284 93.837 -77.030 53.995
164285 93.854 -77.022 53.975
164286 93.869 -77.018 53.953
164287 93.884 -77.013 53.931
164288 93.908 -77.009 53.908
164289 93.933 -77.005 53.890
164290 93.958 -77.001 53.873
164291 93.980 -76.994 53.856
164292 94.002 -76.988 53.835
164293 94.024 -76.982 53.814
164294 94.046 -76.975 53.794
164295 94.067 -76.969 53.775
164296 94.086 -76.960 53.757
164297 94.103 -76.950 53.739
164298 94.115 -76.948 53.720
164299 94.127 -76.945 53.700
164300 94.138 -76.943 53.683
164301 94.151 -76.940 53.667
164302 94.164 -76.937 53.651
164303 94.177 -76.934 53.633
164304 94.190 -76.932 53.614
164305 94.205 -76.929 53.594
164306 94.220 -76.927 53.572
164307 94.236 -76.924 53.551
164308 94.252 -76.922 53.532
164309 94.267 -76.918 53.514
164310 94.279 -76.915 53.499
164311 94.291 -76.911 53.488
164312 94.304 -76.906 53.477
164313 94.316 -76.903 53.468
164314 94.328 -76.900 53.461
164315 94.332 -76.903 53.454
164316 94.344 -76.902 53.446
164317 94.362 -76.900 53.437
164318 94.379 -76.897 53.429
164319 94.397 -76.894 53.422
164320 94.414 -76.892 53.413
164321 94.431 -76.887 53.401
164322 94.448 -76.884 53.386
164323 94.464 -76.881 53.367
164324 94.479 -76.878 53.345
164325 94.494 -76.877 53.322
164326 94.509 -76.875 53.296
164327 94.524 -76.872 53.271
164328 94.539 -76.870 53.246
164329 94.553 -76.868 53.218
164330 94.555 -76.863 53.190
164331 94.565 -76.858 53.162
164332 94.577 -76.856 53.135
164333 94.603 -76.859 53.102
164334 94.636 -76.852 53.069
164335 94.670 -76.845 53.037
164336 94.703 -76.837 53.007
164337 94.736 -76.831 52.973
164338 94.770 -76.824 52.939
164339 94.803 -76.817 52.905
164340 94.837 -76.809 52.869
164341 94.872 -76.802 52.834
164342 94.881 -76.802 52.798
164343 94.891 -76.800 52.761
164344 94.902 -76.799 52.723
164345 94.912 -76.797 52.687
164346 94.923 -76.796 52.652
164347 94.933 -76.794 52.613
164348 94.944 -76.793 52.578
164349 94.954 -76.790 52.546
164350 94.965 -76.789 52.518
164351 94.975 -76.787 52.496
164352 94.986 -76.786 52.479
164353 94.996 -76.784 52.464
164354 95.012 -76.787 52.448
164355 95.033 -76.793 52.432
164356 95.054 -76.799 52.411
164357 95.076 -76.805 52.386
164358 95.097 -76.811 52.355
164359 95.118 -76.817 52.316
164360 95.139 -76.824 52.271
164361 95.161 -76.830 52.221
164362 95.198 -76.842 52.167
164363 95.234 -76.854 52.107
164364 95.271 -76.866 52.044
164365 95.308 -76.878 51.981
164366 95.333 -76.894 51.921
164367 95.316 -76.924 51.862
164368 95.299 -76.953 51.803
164369 95.301 -76.972 51.744
164370 95.315 -76.981 51.685
164371 95.328 -76.990 51.626
164372 95.342 -76.999 51.567
164373 95.356 -77.009 51.511
164374 95.370 -77.018 51.457
164375 95.387 -77.030 51.404
164376 95.404 -77.041 51.352
164377 95.421 -77.053 51.301
164378 95.438 -77.065 51.251
164379 95.455 -77.077 51.202
164380 95.472 -77.088 51.153
164381 95.489 -77.100 51.107
164382 95.506 -77.112 51.061
164383 95.524 -77.128 51.018
164384 95.542 -77.144 50.975
164385 95.559 -77.161 50.933
164386 95.589 -77.168 50.890
164387 95.609 -77.166 50.844
164388 95.629 -77.164 50.799
164389 95.650 -77.162 50.753
164390 95.671 -77.160 50.708
164391 95.691 -77.158 50.664
164392 95.711 -77.156 50.621
164393 95.732 -77.154 50.580
164394 95.742 -77.158 50.542
164395 95.751 -77.171 50.504
164396 95.758 -77.184 50.465
164397 95.767 -77.183 50.427
164398 95.777 -77.182 50.388
164399 95.786 -77.180 50.345
164400 95.796 -77.179 50.299
164401 95.803 -77.183 50.252
164402 95.818 -77.191 50.207
164403 95.851 -77.201 50.163
164404 95.884 -77.212 50.118
164405 95.916 -77.228 50.074
164406 95.948 -77.244 50.029
164407 95.971 -77.247 49.982
164408 95.962 -77.255 49.934
164409 95.954 -77.263 49.886
164410 95.945 -77.272 49.835
164411 95.937 -77.280 49.780
164412 95.928 -77.288 49.723
164413 95.919 -77.297 49.666
164414 95.911 -77.305 49.609
164415 95.903 -77.313 49.556
164416 95.896 -77.321 49.505
164417 95.889 -77.329 49.458
164418 95.908 -77.338 49.414
164419 95.930 -77.346 49.372
164420 95.950 -77.354 49.332
164421 95.971 -77.361 49.292
164422 95.992 -77.370 49.256
164423 96.013 -77.377 49.222
164424 96.034 -77.385 49.188
164425 96.055 -77.392 49.157
164426 96.076 -77.400 49.126
164427 96.097 -77.408 49.097
164428 96.117 -77.416 49.069
164429 96.138 -77.423 49.045
164430 96.159 -77.431 49.023
164431 96.180 -77.439 49.002
164432 96.201 -77.445 48.982
164433 96.221 -77.446 48.960
164434 96.241 -77.447 48.938
164435 96.262 -77.448 48.917
164436 96.282 -77.448 48.898
164437 96.302 -77.448 48.880
164438 96.323 -77.449 48.862
164439 96.343 -77.450 48.844
164440 96.364 -77.451 48.830
164441 96.384 -77.452 48.819
164442 96.408 -77.459 48.811
164443 96.430 -77.469 48.805
164444 96.447 -77.485 48.797
164445 96.458 -77.506 48.786
164446 96.472 -77.527 48.773
164447 96.488 -77.549 48.758
164448 96.506 -77.570 48.743
164449 96.523 -77.592 48.728
164450 96.540 -77.614 48.714
164451 96.557 -77.635 48.699
164452 96.574 -77.656 48.682
164453 96.591 -77.678 48.664
164454 96.604 -77.680 48.645
164455 96.616 -77.678 48.626
164456 96.628 -77.677 48.609
164457 96.640 -77.674 48.592
164458 96.652 -77.673 48.575
164459 96.664 -77.671 48.557
164460 96.675 -77.669 48.539
164461 96.687 -77.667 48.520
164462 96.699 -77.665 48.503
164463 96.711 -77.664 48.487
164464 96.723 -77.661 48.470
164465 96.735 -77.660 48.454
164466 96.746 -77.658 48.437
164467 96.761 -77.655 48.419
164468 96.785 -77.659 48.400
164469 96.800 -77.645 48.376
164470 96.820 -77.648 48.352
164471 96.840 -77.649 48.327
164472 96.861 -77.652 48.300
164473 96.881 -77.654 48.275
164474 96.901 -77.655 48.249
164475 96.921 -77.658 48.224
164476 96.941 -77.659 48.197
164477 96.961 -77.661 48.172
164478 96.981 -77.663 48.147
164479 97.002 -77.665 48.123
164480 97.022 -77.667 48.099
164481 97.042 -77.669 48.075
164482 97.051 -77.670 48.052
164483 97.057 -77.675 48.029
164484 97.064 -77.680 48.007
164485 97.070 -77.685 47.985
164486 97.076 -77.690 47.965
164487 97.091 -77.697 47.947
164488 97.119 -77.705 47.928
164489 97.146 -77.714 47.908
164490 97.174 -77.723 47.889
164491 97.202 -77.731 47.868
164492 97.229 -77.740 47.847
164493 97.256 -77.747 47.826
164494 97.282 -77.744 47.804
164495 97.307 -77.741 47.782
164496 97.333 -77.738 47.759
164497 97.358 -77.735 47.735
164498 97.383 -77.734 47.711
164499 97.407 -77.744 47.687
164500 97.430 -77.754 47.663
164501 97.454 -77.764 47.638
164502 97.477 -77.774 47.614
164503 97.501 -77.783 47.591
164504 97.524 -77.793 47.567
164505 97.547 -77.802 47.542
164506 97.571 -77.812 47.518
164507 97.595 -77.822 47.494
164508 97.618 -77.832 47.471
164509 97.642 -77.842 47.449
164510 97.665 -77.852 47.428
164511 97.696 -77.860 47.406
164512 97.706 -77.853 47.384
164513 97.715 -77.845 47.362
164514 97.719 -77.826 47.338
164515 97.721 -77.799 47.314
164516 97.727 -77.775 47.291
164517 97.737 -77.752 47.265
164518 97.748 -77.741 47.240
164519 97.772 -77.770 47.214
164520 97.797 -77.775 47.187
164521 97.822 -77.781 47.161
164522 97.847 -77.787 47.135
164523 97.872 -77.792 47.108
164524 97.888 -77.801 47.084
164525 97.897 -77.815 47.061
164526 97.907 -77.828 47.038
164527 97.916 -77.842 47.015
164528 97.928 -77.851 46.988
164529 97.945 -77.852 46.961
164530 97.962 -77.852 46.935
164531 97.979 -77.853 46.910
164532 97.996 -77.854 46.884
164533 98.013 -77.855 46.856
164534 98.025 -77.862 46.829
164535 98.035 -77.872 46.804
164536 98.046 -77.882 46.777
164537 98.056 -77.892 46.750
164538 98.067 -77.902 46.725
164539 98.077 -77.912 46.700
164540 98.087 -77.921 46.676
164541 98.102 -77.929 46.652
164542 98.131 -77.919 46.627
164543 98.162 -77.905 46.602
164544 98.194 -77.890 46.577
164545 98.225 -77.877 46.553
164546 98.239 -77.868 46.532
164547 98.246 -77.864 46.510
164548 98.252 -77.859 46.489
164549 98.260 -77.853 46.468
164550 98.266 -77.848 46.446
164551 98.273 -77.843 46.424
164552 98.279 -77.837 46.403
164553 98.291 -77.834 46.384
164554 98.317 -77.836 46.365
164555 98.362 -77.846 46.346
164556 98.406 -77.857 46.328
164557 98.450 -77.868 46.311
164558 98.483 -77.868 46.292
164559 98.486 -77.868 46.274
164560 98.503 -77.875 46.256
164561 98.522 -77.883 46.240
164562 98.541 -77.890 46.223
164563 98.559 -77.897 46.205
164564 98.578 -77.905 46.187
164565 98.596 -77.912 46.167
164566 98.615 -77.919 46.149
164567 98.634 -77.927 46.131
164568 98.652 -77.934 46.112
164569 98.671 -77.942 46.092
164570 98.690 -77.949 46.072
164571 98.708 -77.956 46.052
164572 98.727 -77.964 46.031
164573 98.745 -77.971 46.011
164574 98.764 -77.978 45.993
164575 98.783 -77.986 45.975
164576 98.818 -77.993 45.957
164577 98.874 -78.003 45.939
164578 98.955 -78.015 45.922
164579 99.030 -78.023 45.903
164580 99.075 -78.015 45.886
164581 99.120 -78.006 45.869
164582 99.165 -77.998 45.852
164583 99.210 -77.990 45.836
164584 99.263 -77.988 45.821
164585 99.317 -77.989 45.806
164586 99.339 -77.986 45.792
164587 99.350 -77.990 45.778
164588 99.360 -77.996 45.767
164589 99.368 -78.003 45.757
164590 99.377 -78.011 45.746
164591 99.385 -78.018 45.736
164592 99.394 -78.025 45.726
164593 99.403 -78.032 45.715
164594 99.411 -78.040 45.705
164595 99.420 -78.047 45.694
164596 99.428 -78.054 45.683
164597 99.437 -78.062 45.673
164598 99.446 -78.069 45.663
164599 99.456 -78.076 45.651
164600 99.488 -78.086 45.639
164601 99.519 -78.095 45.627
164602 99.551 -78.105 45.617
164603 99.582 -78.115 45.606
164604 99.614 -78.124 45.594
164605 99.644 -78.134 45.582
164606 99.674 -78.148 45.571
164607 99.693 -78.161 45.559
164608 99.707 -78.172 45.547
164609 99.721 -78.184 45.533
164610 99.735 -78.196 45.519
164611 99.749 -78.207 45.505
164612 99.763 -78.219 45.490
164613 99.777 -78.230 45.474
164614 99.792 -78.244 45.459
164615 99.809 -78.258 45.445
164616 99.826 -78.273 45.429
164617 99.843 -78.288 45.414
164618 99.862 -78.306 45.398
164619 99.881 -78.322 45.381
164620 99.883 -78.333 45.365
164621 99.903 -78.351 45.348
164622 99.923 -78.369 45.331
164623 99.943 -78.388 45.314
164624 99.964 -78.407 45.296
164625 99.984 -78.425 45.279
164626 100.004 -78.443 45.262
164627 100.025 -78.461 45.245
164628 100.065 -78.477 45.229
164629 100.116 -78.492 45.213
164630 100.166 -78.506 45.197
164631 100.216 -78.520 45.181
164632 100.253 -78.529 45.165
164633 100.289 -78.539 45.150
164634 100.326 -78.548 45.135
164635 100.363 -78.558 45.119
164636 100.400 -78.568 45.103
164637 100.436 -78.578 45.088
164638 100.473 -78.588 45.073
164639 100.512 -78.598 45.057
164640 100.557 -78.609 45.041
164641 100.602 -78.620 45.026
164642 100.646 -78.632 45.011
164643 100.675 -78.614 44.996
164644 100.700 -78.623 44.980
164645 100.724 -78.631 44.965
164646 100.749 -78.639 44.949
164647 100.776 -78.650 44.933
164648 100.805 -78.662 44.918
164649 100.834 -78.675 44.903
164650 100.863 -78.687 44.887
164651 100.893 -78.700 44.871
164652 100.923 -78.717 44.855
164653 100.953 -78.733 44.840
164654 100.983 -78.742 44.824
164655 101.011 -78.744 44.808
164656 101.041 -78.747 44.793
164657 101.069 -78.749 44.778
164658 101.095 -78.760 44.763
164659 101.097 -78.814 44.747
164660 101.101 -78.865 44.732
164661 101.149 -78.879 44.717
164662 101.197 -78.893 44.702
164663 101.247 -78.905 44.688
164664 101.283 -78.910 44.674
164665 101.305 -78.891 44.660
164666 101.325 -78.872 44.646
164667 101.349 -78.882 44.632
164668 101.367 -78.914 44.618
164669 101.383 -78.948 44.604
164670 101.403 -78.981 44.590
164671 101.445 -79.000 44.577
164672 101.487 -79.019 44.564
164673 101.530 -79.038 44.551
164674 101.572 -79.058 44.538
164675 101.614 -79.077 44.525
164676 101.656 -79.096 44.513
164677 101.676 -79.103 44.500
164678 101.693 -79.109 44.489
164679 101.710 -79.115 44.477
164900 92.399 -77.291 53.972
164901 92.418 -77.296 53.992
164902 92.437 -77.302 54.010
164903 92.456 -77.307 54.030
164904 92.475 -77.313 54.051
164905 92.498 -77.324 54.071
164906 92.521 -77.335 54.090
164907 92.545 -77.346 54.109
164908 92.567 -77.357 54.127
164909 92.590 -77.368 54.143
164910 92.614 -77.379 54.159
164911 92.637 -77.391 54.173
164912 92.660 -77.401 54.187
164913 92.683 -77.413 54.198
164914 92.706 -77.424 54.209
164915 92.729 -77.435 54.220
164916 92.752 -77.446 54.228
164917 92.769 -77.457 54.238
164918 92.776 -77.470 54.248
164919 92.783 -77.483 54.257
164920 92.790 -77.495 54.266
164921 92.798 -77.507 54.274
164922 92.828 -77.501 54.282
164923 92.846 -77.490 54.290
164924 92.872 -77.482 54.300
164925 92.901 -77.475 54.311
164926 92.930 -77.467 54.321
164927 92.958 -77.461 54.329
164928 92.987 -77.453 54.335
164929 93.016 -77.446 54.340
164930 93.032 -77.441 54.344
164931 93.043 -77.437 54.350
164932 93.055 -77.427 54.357
164933 93.071 -77.410 54.365
164934 93.087 -77.395 54.373
164935 93.097 -77.380 54.382
164936 93.108 -77.366 54.391
164937 93.117 -77.351 54.399
164938 93.127 -77.338 54.405
164939 93.137 -77.323 54.408
164940 93.147 -77.309 54.411
164941 93.157 -77.294 54.413
164942 93.166 -77.280 54.413
164943 93.177 -77.266 54.412
164944 93.186 -77.252 54.409
164945 93.196 -77.238 54.402
164946 93.212 -77.228 54.391
164947 93.229 -77.219 54.382
164948 93.247 -77.213 54.375
164949 93.264 -77.205 54.370
164950 93.280 -77.197 54.365
164951 93.297 -77.190 54.361
164952 93.314 -77.183 54.357
164953 93.331 -77.175 54.352
164954 93.348 -77.168 54.348
164955 93.365 -77.160 54.342
164956 93.382 -77.153 54.335
164957 93.405 -77.144 54.326
164958 93.428 -77.134 54.314
164959 93.452 -77.124 54.302
164960 93.476 -77.114 54.289
164961 93.500 -77.104 54.276
164962 93.524 -77.094 54.265
164963 93.548 -77.085 54.254
164964 93.572 -77.075 54.247
164965 93.596 -77.065 54.239
164966 93.612 -77.067 54.233
164967 93.616 -77.087 54.229
164968 93.623 -77.094 54.223
164969 93.644 -77.082 54.215
164970 93.665 -77.070 54.204
164971 93.686 -77.059 54.191
164972 93.707 -77.047 54.176
164973 93.719 -77.044 54.162
164974 93.731 -77.039 54.150
164975 93.735 -77.028 54.136
164976 93.740 -77.019 54.123
164977 93.748 -77.017 54.110
164978 93.756 -77.016 54.095
164979 93.764 -77.016 54.080
164980 93.773 -77.014 54.066
164981 93.781 -77.013 54.052
164982 93.789 -77.012 54.042
164983 93.797 -77.011 54.034
164984 93.808 -77.012 54.025
164985 93.825 -77.016 54.014
164986 93.842 -77.014 53.996
164987 93.858 -77.009 53.977
164988 93.884 -77.005 53.956
164989 93.909 -77.000 53.936
164990 93.934 -76.997 53.917
164991 93.959 -76.992 53.900
164992 93.981 -76.985 53.882
164993 94.002 -76.979 53.860
164994 94.024 -76.973 53.837
164995 94.046 -76.966 53.815
164996 94.064 -76.956 53.795
164997 94.081 -76.946 53.776
164998 94.098 -76.936 53.757
164999 94.109 -76.934 53.739
165000 94.121 -76.931 53.721
165001 94.133 -76.928 53.706
165002 94.144 -76.926 53.692
165003 94.156 -76.923 53.678
165004 94.168 -76.921 53.662
165005 94.181 -76.918 53.643
165006 94.197 -76.915 53.623
165007 94.214 -76.912 53.601
165008 94.230 -76.909 53.579
165009 94.246 -76.907 53.558
165010 94.260 -76.904 53.539
165011 94.272 -76.902 53.521
165012 94.284 -76.899 53.507
165013 94.296 -76.896 53.494
165014 94.309 -76.892 53.482
165015 94.322 -76.888 53.474
165016 94.330 -76.887 53.465
165017 94.343 -76.887 53.457
165018 94.360 -76.884 53.449
165019 94.378 -76.881 53.443
165020 94.395 -76.879 53.435
165021 94.413 -76.876 53.427
165022 94.430 -76.874 53.417
165023 94.447 -76.869 53.404
165024 94.462 -76.868 53.386
165025 94.477 -76.865 53.364
165026 94.492 -76.863 53.342
165027 94.507 -76.861 53.317
165028 94.522 -76.859 53.291
165029 94.537 -76.856 53.267
165030 94.551 -76.854 53.240
165031 94.561 -76.851 53.213
165032 94.561 -76.846 53.185
165033 94.578 -76.840 53.158
165034 94.613 -76.832 53.125
165035 94.647 -76.824 53.092
165036 94.681 -76.817 53.060
165037 94.714 -76.810 53.030
165038 94.748 -76.803 52.997
165039 94.781 -76.796 52.963
165040 94.815 -76.789 52.930
165041 94.848 -76.782 52.896
165042 94.870 -76.778 52.860
165043 94.881 -76.776 52.822
165044 94.891 -76.774 52.783
165045 94.902 -76.773 52.745
165046 94.912 -76.771 52.706
165047 94.923 -76.770 52.666
165048 94.933 -76.768 52.627
165049 94.944 -76.766 52.590
165050 94.954 -76.765 52.556
165051 94.965 -76.763 52.524
165052 94.975 -76.762 52.501
165053 94.986 -76.760 52.482
165054 94.996 -76.758 52.465
165055 95.008 -76.758 52.447
165056 95.029 -76.764 52.429
165057 95.051 -76.770 52.409
165058 95.072 -76.776 52.386
165059 95.093 -76.782 52.356
165060 95.114 -76.788 52.322
165061 95.135 -76.794 52.280
165062 95.168 -76.806 52.231
165063 95.205 -76.818 52.178
165064 95.242 -76.830 52.118
165065 95.278 -76.842 52.056
165066 95.315 -76.854 51.995
165067 95.344 -76.868 51.934
165068 95.327 -76.898 51.875
165069 95.310 -76.928 51.816
165070 95.296 -76.956 51.757
165071 95.310 -76.965 51.697
165072 95.324 -76.975 51.638
165073 95.338 -76.984 51.580
165074 95.351 -76.993 51.525
165075 95.367 -77.003 51.471
165076 95.384 -77.016 51.418
165077 95.401 -77.027 51.365
165078 95.418 -77.039 51.314
165079 95.435 -77.050 51.263
165080 95.452 -77.063 51.214
165081 95.469 -77.074 51.165
165082 95.486 -77.086 51.118
165083 95.503 -77.097 51.072
165084 95.520 -77.112 51.028
165085 95.538 -77.128 50.984
165086 95.558 -77.144 50.941
165087 95.595 -77.152 50.897
165088 95.616 -77.150 50.852
165089 95.636 -77.148 50.807
165090 95.657 -77.146 50.762
165091 95.677 -77.144 50.718
165092 95.698 -77.142 50.674
165093 95.712 -77.144 50.633
165094 95.724 -77.147 50.593
165095 95.734 -77.150 50.556
165096 95.743 -77.163 50.519
165097 95.748 -77.187 50.480
165098 95.755 -77.200 50.441
165099 95.765 -77.198 50.402
165100 95.774 -77.197 50.358
165101 95.784 -77.196 50.312
165102 95.793 -77.194 50.266
165103 95.810 -77.199 50.222
165104 95.842 -77.215 50.178
165105 95.875 -77.231 50.135
165106 95.907 -77.247 50.091
165107 95.936 -77.248 50.048
165108 95.961 -77.237 50.003
165109 95.961 -77.240 49.957
165110 95.953 -77.248 49.909
165111 95.945 -77.257 49.857
165112 95.936 -77.265 49.801
165113 95.927 -77.273 49.743
165114 95.919 -77.282 49.683
165115 95.910 -77.291 49.623
165116 95.902 -77.299 49.568
165117 95.893 -77.307 49.516
165118 95.894 -77.317 49.468
165119 95.910 -77.323 49.422
165120 95.927 -77.329 49.380
165121 95.949 -77.337 49.339
165122 95.969 -77.345 49.299
165123 95.990 -77.352 49.263
165124 96.011 -77.360 49.228
165125 96.032 -77.368 49.195
165126 96.053 -77.376 49.164
165127 96.074 -77.383 49.133
165128 96.095 -77.391 49.104
165129 96.115 -77.399 49.076
165130 96.136 -77.407 49.053
165131 96.157 -77.414 49.030
165132 96.178 -77.422 49.007
165133 96.199 -77.427 48.986
165134 96.219 -77.428 48.965
165135 96.239 -77.429 48.944
165136 96.260 -77.429 48.925
165137 96.280 -77.430 48.907
165138 96.301 -77.431 48.887
165139 96.321 -77.432 48.868
165140 96.341 -77.432 48.850
165141 96.365 -77.438 48.835
165142 96.389 -77.445 48.823
165143 96.413 -77.451 48.814
165144 96.432 -77.464 48.809
165145 96.449 -77.479 48.801
165146 96.464 -77.496 48.791
165147 96.474 -77.518 48.781
165148 96.485 -77.540 48.768
165149 96.501 -77.562 48.753
165150 96.518 -77.583 48.738
165151 96.535 -77.605 48.724
165152 96.552 -77.627 48.707
165153 96.569 -77.648 48.690
165154 96.586 -77.669 48.673
165155 96.601 -77.680 48.655
165156 96.613 -77.679 48.637
165157 96.624 -77.677 48.619
165158 96.637 -77.675 48.602
165159 96.648 -77.674 48.585
165160 96.660 -77.671 48.568
165161 96.672 -77.670 48.550
165162 96.684 -77.667 48.532
165163 96.696 -77.666 48.515
165164 96.708 -77.664 48.499
165165 96.720 -77.662 48.483
165166 96.734 -77.660 48.467
165167 96.755 -77.656 48.450
165168 96.777 -77.652 48.432
165169 96.804 -77.663 48.413
165170 96.822 -77.656 48.391
165171 96.834 -77.637 48.368
165172 96.852 -77.632 48.344
165173 96.872 -77.634 48.319
165174 96.892 -77.636 48.294
165175 96.912 -77.638 48.269
165176 96.933 -77.640 48.243
165177 96.953 -77.642 48.217
165178 96.973 -77.644 48.192
165179 96.993 -77.645 48.167
165180 97.013 -77.648 48.142
165181 97.033 -77.649 48.118
165182 97.052 -77.651 48.094
165183 97.060 -77.652 48.071
165184 97.066 -77.657 48.048
165185 97.072 -77.661 48.026
165186 97.079 -77.667 48.006
165187 97.086 -77.672 47.986
165188 97.113 -77.680 47.968
165189 97.141 -77.689 47.948
165190 97.169 -77.698 47.928
165191 97.196 -77.706 47.909
165192 97.224 -77.715 47.887
165193 97.251 -77.724 47.865
165194 97.279 -77.732 47.844
165195 97.307 -77.741 47.822
165196 97.333 -77.742 47.798
165197 97.358 -77.739 47.775
165198 97.384 -77.736 47.752
165199 97.408 -77.736 47.727
165200 97.432 -77.746 47.703
165201 97.456 -77.755 47.678
165202 97.479 -77.765 47.653
165203 97.503 -77.775 47.629
165204 97.526 -77.785 47.605
165205 97.549 -77.795 47.580
165206 97.573 -77.805 47.555
165207 97.597 -77.815 47.530
165208 97.620 -77.824 47.505
165209 97.644 -77.834 47.481
165210 97.667 -77.844 47.459
165211 97.690 -77.854 47.437
165212 97.713 -77.862 47.415
165213 97.728 -77.843 47.392
165214 97.739 -77.819 47.368
165215 97.749 -77.796 47.344
165216 97.760 -77.773 47.321
165217 97.771 -77.749 47.298
165218 97.782 -77.727 47.272
165219 97.792 -77.733 47.246
165220 97.800 -77.758 47.220
165221 97.869 -77.762 47.194
165222 97.900 -77.768 47.167
165223 97.884 -77.773 47.141
165224 97.872 -77.784 47.115
165225 97.891 -77.806 47.091
165226 97.902 -77.821 47.067
165227 97.912 -77.835 47.044
165228 97.927 -77.839 47.020
165229 97.944 -77.840 46.994
165230 97.961 -77.840 46.967
165231 97.978 -77.841 46.940
165232 97.995 -77.842 46.916
165233 98.012 -77.843 46.890
165234 98.029 -77.843 46.863
165235 98.043 -77.847 46.837
165236 98.054 -77.857 46.811
165237 98.064 -77.867 46.784
165238 98.075 -77.877 46.758
165239 98.085 -77.887 46.732
165240 98.095 -77.896 46.707
165241 98.112 -77.902 46.682
165242 98.131 -77.909 46.658
165243 98.150 -77.915 46.633
165244 98.179 -77.902 46.608
165245 98.211 -77.888 46.583
165246 98.242 -77.874 46.558
165247 98.262 -77.865 46.537
165248 98.269 -77.859 46.516
165249 98.276 -77.854 46.494
165250 98.282 -77.849 46.472
165251 98.289 -77.843 46.450
165252 98.296 -77.838 46.428
165253 98.307 -77.834 46.407
165254 98.323 -77.832 46.389
165255 98.353 -77.836 46.369
165256 98.397 -77.846 46.351
165257 98.441 -77.856 46.333
165258 98.485 -77.867 46.314
165259 98.498 -77.862 46.295
165260 98.501 -77.862 46.277
165261 98.505 -77.863 46.259
165262 98.524 -77.871 46.243
165263 98.543 -77.878 46.226
165264 98.561 -77.885 46.209
165265 98.580 -77.893 46.191
165266 98.598 -77.900 46.173
165267 98.617 -77.907 46.155
165268 98.636 -77.915 46.138
165269 98.654 -77.922 46.119
165270 98.673 -77.930 46.099
165271 98.692 -77.937 46.079
165272 98.710 -77.944 46.060
165273 98.729 -77.952 46.040
165274 98.747 -77.959 46.020
165275 98.766 -77.966 46.001
165276 98.788 -77.974 45.983
165277 98.830 -77.982 45.965
165278 98.886 -77.991 45.947
165279 98.968 -78.003 45.930
165280 99.049 -78.015 45.911
165281 99.111 -78.015 45.894
165282 99.156 -78.007 45.877
165283 99.201 -77.999 45.860
165284 99.245 -77.990 45.844
165285 99.290 -77.982 45.828
165286 99.330 -77.974 45.813
165287 99.340 -77.981 45.798
165288 99.349 -77.988 45.783
165289 99.357 -77.996 45.772
165290 99.366 -78.003 45.762
165291 99.374 -78.010 45.751
165292 99.383 -78.018 45.742
165293 99.392 -78.025 45.731
165294 99.400 -78.032 45.721
165295 99.409 -78.039 45.711
165296 99.418 -78.047 45.701
165297 99.426 -78.053 45.690
165298 99.435 -78.061 45.679
165299 99.443 -78.069 45.668
165300 99.459 -78.076 45.657
165301 99.490 -78.086 45.646
165302 99.522 -78.095 45.636
165303 99.553 -78.105 45.625
165304 99.585 -78.115 45.613
165305 99.615 -78.125 45.602
165306 99.642 -78.135 45.590
165307 99.673 -78.150 45.578
165308 99.703 -78.166 45.566
165309 99.717 -78.178 45.554
165310 99.731 -78.189 45.540
165311 99.745 -78.200 45.525
165312 99.759 -78.213 45.510
165313 99.773 -78.224 45.495
165314 99.787 -78.235 45.479
165315 99.800 -78.247 45.464
165316 99.815 -78.259 45.450
165317 99.829 -78.270 45.434
165318 99.843 -78.282 45.419
165319 99.858 -78.294 45.403
165320 99.881 -78.307 45.386
165321 99.902 -78.325 45.369
165322 99.923 -78.343 45.352
165323 99.943 -78.361 45.335
165324 99.963 -78.379 45.318
165325 99.984 -78.398 45.301
165326 100.004 -78.416 45.284
165327 100.024 -78.435 45.267
165328 100.045 -78.453 45.251
165329 100.086 -78.469 45.235
165330 100.137 -78.482 45.219
165331 100.188 -78.497 45.203
165332 100.224 -78.507 45.186
165333 100.261 -78.517 45.170
165334 100.298 -78.526 45.154
165335 100.335 -78.535 45.137
165336 100.371 -78.545 45.122
165337 100.408 -78.555 45.107
165338 100.445 -78.565 45.092
165339 100.482 -78.575 45.077
165340 100.519 -78.584 45.062
165341 100.561 -78.595 45.048
165342 100.606 -78.607 45.033
165343 100.650 -78.617 45.018
165344 100.675 -78.594 45.003
165345 100.699 -78.602 44.988
165346 100.724 -78.611 44.973
165347 100.751 -78.620 44.957
165348 100.780 -78.633 44.941
165349 100.809 -78.645 44.926
165350 100.838 -78.658 44.911
165351 100.867 -78.670 44.895
165352 100.896 -78.683 44.878
165353 100.926 -78.697 44.862
165354 100.955 -78.702 44.847
165355 100.984 -78.705 44.831
165356 101.013 -78.708 44.816
165357 101.042 -78.711 44.801
165358 101.070 -78.713 44.785
165359 101.100 -78.716 44.770
165360 101.102 -78.770 44.755
165361 101.127 -78.808 44.739
165362 101.167 -78.836 44.724
165363 101.213 -78.854 44.709
165364 101.263 -78.865 44.695
165365 101.299 -78.874 44.682
165366 101.310 -78.878 44.668
165367 101.326 -78.871 44.653
165368 101.348 -78.867 44.639
165369 101.365 -78.901 44.624
165370 101.382 -78.935 44.609
165371 101.403 -78.967 44.596
165372 101.445 -78.986 44.583
165373 101.487 -79.006 44.570
165374 101.529 -79.024 44.557
165375 101.571 -79.044 44.544
165376 101.614 -79.063 44.531
165377 101.656 -79.082 44.518
165378 101.677 -79.090 44.506
165379 101.694 -79.096 44.495
165380 101.710 -79.102 44.483
165381 101.728 -79.107 44.470
165602 92.398 -77.257 53.970
165603 92.417 -77.263 53.989
165604 92.439 -77.273 54.010
165605 92.462 -77.284 54.031
165606 92.486 -77.295 54.052
165607 92.509 -77.306 54.073
165608 92.532 -77.317 54.092
165609 92.555 -77.328 54.112
165610 92.578 -77.339 54.128
165611 92.601 -77.351 54.145
165612 92.624 -77.361 54.161
165613 92.647 -77.373 54.175
165614 92.670 -77.384 54.188
165615 92.693 -77.395 54.200
165616 92.716 -77.406 54.211
165617 92.739 -77.417 54.222
165618 92.762 -77.428 54.231
165619 92.769 -77.441 54.240
165620 92.776 -77.453 54.250
165621 92.783 -77.466 54.258
165622 92.790 -77.478 54.266
165623 92.797 -77.491 54.275
165624 92.824 -77.486 54.283
165625 92.843 -77.476 54.295
165626 92.872 -77.469 54.307
165627 92.901 -77.461 54.319
165628 92.930 -77.454 54.329
165629 92.958 -77.447 54.335
165630 92.988 -77.440 54.339
165631 93.016 -77.433 54.341
165632 93.045 -77.426 54.347
165633 93.062 -77.414 54.353
165634 93.072 -77.399 54.360
165635 93.082 -77.385 54.368
165636 93.091 -77.370 54.376
165637 93.101 -77.357 54.386
165638 93.112 -77.342 54.395
165639 93.121 -77.328 54.402
165640 93.131 -77.313 54.405
165641 93.141 -77.300 54.406
165642 93.151 -77.285 54.406
165643 93.161 -77.271 54.403
165644 93.170 -77.257 54.399
165645 93.181 -77.243 54.392
165646 93.190 -77.229 54.386
165647 93.206 -77.219 54.376
165648 93.224 -77.210 54.368
165649 93.241 -77.203 54.360
165650 93.258 -77.195 54.356
165651 93.275 -77.188 54.353
165652 93.292 -77.181 54.350
165653 93.309 -77.173 54.347
165654 93.326 -77.166 54.345
165655 93.343 -77.158 54.342
165656 93.360 -77.151 54.339
165657 93.370 -77.141 54.333
165658 93.398 -77.134 54.324
165659 93.425 -77.125 54.313
165660 93.449 -77.116 54.301
165661 93.473 -77.106 54.289
165662 93.497 -77.096 54.276
165663 93.521 -77.086 54.264
165664 93.545 -77.076 54.254
165665 93.569 -77.066 54.247
165666 93.592 -77.056 54.241
165667 93.600 -77.072 54.235
165668 93.606 -77.083 54.232
165669 93.618 -77.080 54.225
165670 93.639 -77.069 54.217
165671 93.659 -77.056 54.205
165672 93.681 -77.045 54.192
165673 93.697 -77.038 54.178
165674 93.709 -77.034 54.165
165675 93.722 -77.031 54.152
165676 93.730 -77.023 54.140
165677 93.734 -77.012 54.128
165678 93.739 -77.002 54.117
165679 93.744 -76.994 54.104
165680 93.752 -76.993 54.089
165681 93.760 -76.991 54.075
165682 93.769 -76.990 54.062
165683 93.777 -76.990 54.052
165684 93.785 -76.988 54.044
165685 93.797 -76.990 54.037
165686 93.814 -76.994 54.030
165687 93.833 -76.999 54.015
165688 93.860 -77.000 53.998
165689 93.885 -76.997 53.979
165690 93.911 -76.992 53.960
165691 93.936 -76.988 53.944
165692 93.959 -76.983 53.927
165693 93.981 -76.976 53.908
165694 94.003 -76.970 53.886
165695 94.024 -76.963 53.863
165696 94.041 -76.953 53.841
165697 94.058 -76.941 53.820
165698 94.076 -76.931 53.798
165699 94.091 -76.922 53.778
165700 94.104 -76.919 53.759
165701 94.115 -76.917 53.739
165702 94.127 -76.914 53.725
165703 94.139 -76.912 53.713
165704 94.150 -76.909 53.702
165705 94.163 -76.906 53.688
165706 94.177 -76.903 53.671
165707 94.193 -76.900 53.650
165708 94.209 -76.898 53.629
165709 94.225 -76.895 53.606
165710 94.242 -76.892 53.585
165711 94.254 -76.890 53.565
165712 94.266 -76.887 53.546
165713 94.278 -76.884 53.528
165714 94.290 -76.881 53.513
165715 94.302 -76.879 53.499
165716 94.315 -76.876 53.489
165717 94.327 -76.873 53.480
165718 94.341 -76.871 53.472
165719 94.359 -76.868 53.466
165720 94.377 -76.866 53.460
165721 94.394 -76.863 53.451
165722 94.411 -76.861 53.442
165723 94.428 -76.859 53.433
165724 94.445 -76.856 53.421
165725 94.460 -76.854 53.404
165726 94.475 -76.852 53.384
165727 94.490 -76.850 53.362
165728 94.504 -76.847 53.337
165729 94.520 -76.845 53.312
165730 94.534 -76.843 53.288
165731 94.549 -76.840 53.261
165732 94.564 -76.838 53.235
165733 94.566 -76.835 53.209
165734 94.558 -76.838 53.180
165735 94.594 -76.830 53.147
165736 94.628 -76.821 53.115
165737 94.663 -76.813 53.083
165738 94.698 -76.805 53.052
165739 94.733 -76.796 53.019
165740 94.768 -76.788 52.987
165741 94.803 -76.781 52.955
165742 94.846 -76.766 52.921
165743 94.868 -76.758 52.885
165744 94.880 -76.752 52.847
165745 94.891 -76.749 52.808
165746 94.902 -76.747 52.767
165747 94.912 -76.746 52.727
165748 94.922 -76.743 52.685
165749 94.933 -76.742 52.644
165750 94.944 -76.740 52.605
165751 94.954 -76.739 52.568
165752 94.965 -76.737 52.533
165753 94.975 -76.736 52.506
165754 94.986 -76.734 52.484
165755 94.996 -76.732 52.463
165756 95.006 -76.730 52.443
165757 95.025 -76.735 52.423
165758 95.047 -76.741 52.402
165759 95.068 -76.747 52.380
165760 95.089 -76.753 52.354
165761 95.110 -76.759 52.321
165762 95.139 -76.768 52.282
165763 95.176 -76.781 52.236
165764 95.212 -76.793 52.183
165765 95.249 -76.805 52.125
165766 95.285 -76.817 52.065
165767 95.322 -76.829 52.006
165768 95.355 -76.843 51.947
165769 95.338 -76.872 51.887
165770 95.321 -76.902 51.827
165771 95.304 -76.932 51.767
165772 95.306 -76.950 51.707
165773 95.319 -76.959 51.648
165774 95.333 -76.969 51.591
165775 95.347 -76.978 51.538
165776 95.365 -76.990 51.483
165777 95.381 -77.001 51.430
165778 95.398 -77.013 51.378
165779 95.415 -77.025 51.325
165780 95.432 -77.037 51.274
165781 95.449 -77.048 51.225
165782 95.466 -77.059 51.177
165783 95.483 -77.072 51.129
165784 95.500 -77.083 51.082
165785 95.517 -77.095 51.037
165786 95.535 -77.112 50.992
165787 95.561 -77.125 50.948
165788 95.600 -77.134 50.903
165789 95.623 -77.134 50.858
165790 95.643 -77.132 50.813
165791 95.664 -77.130 50.769
165792 95.683 -77.128 50.727
165793 95.694 -77.132 50.685
165794 95.704 -77.136 50.645
165795 95.716 -77.140 50.607
165796 95.726 -77.144 50.570
165797 95.735 -77.157 50.533
165798 95.740 -77.179 50.494
165799 95.746 -77.203 50.454
165800 95.753 -77.215 50.415
165801 95.762 -77.213 50.370
165802 95.772 -77.213 50.324
165803 95.781 -77.211 50.279
165804 95.801 -77.218 50.236
165805 95.833 -77.234 50.194
165806 95.866 -77.250 50.152
165807 95.894 -77.253 50.109
165808 95.920 -77.241 50.068
165809 95.945 -77.231 50.025
165810 95.961 -77.226 49.980
165811 95.952 -77.234 49.933
165812 95.944 -77.242 49.881
165813 95.935 -77.251 49.824
165814 95.926 -77.259 49.763
165815 95.918 -77.267 49.699
165816 95.909 -77.276 49.638
165817 95.901 -77.284 49.581
165818 95.893 -77.292 49.528
165819 95.902 -77.304 49.478
165820 95.919 -77.310 49.431
165821 95.935 -77.316 49.387
165822 95.951 -77.322 49.346
165823 95.968 -77.328 49.306
165824 95.988 -77.335 49.269
165825 96.009 -77.343 49.234
165826 96.030 -77.351 49.202
165827 96.051 -77.359 49.171
165828 96.072 -77.367 49.141
165829 96.093 -77.374 49.112
165830 96.114 -77.382 49.084
165831 96.134 -77.390 49.061
165832 96.156 -77.398 49.038
165833 96.176 -77.405 49.015
165834 96.197 -77.409 48.993
165835 96.217 -77.410 48.972
165836 96.238 -77.410 48.952
165837 96.258 -77.411 48.933
165838 96.278 -77.412 48.915
165839 96.299 -77.413 48.895
165840 96.320 -77.416 48.874
165841 96.345 -77.423 48.856
165842 96.369 -77.429 48.841
165843 96.394 -77.437 48.828
165844 96.418 -77.445 48.816
165845 96.434 -77.460 48.812
165846 96.451 -77.475 48.804
165847 96.467 -77.491 48.795
165848 96.481 -77.508 48.788
165849 96.491 -77.530 48.776
165850 96.502 -77.553 48.762
165851 96.512 -77.575 48.747
165852 96.529 -77.596 48.733
165853 96.546 -77.617 48.714
165854 96.564 -77.639 48.697
165855 96.581 -77.661 48.680
165856 96.598 -77.681 48.664
165857 96.609 -77.680 48.646
165858 96.621 -77.677 48.628
165859 96.633 -77.676 48.612
165860 96.645 -77.674 48.596
165861 96.657 -77.672 48.578
165862 96.669 -77.671 48.561
165863 96.681 -77.668 48.544
165864 96.692 -77.667 48.526
165865 96.707 -77.664 48.511
165866 96.728 -77.660 48.496
165867 96.749 -77.656 48.479
165868 96.770 -77.652 48.461
165869 96.792 -77.649 48.443
165870 96.822 -77.667 48.425
165871 96.844 -77.671 48.405
165872 96.857 -77.652 48.383
165873 96.870 -77.632 48.361
165874 96.884 -77.617 48.339
165875 96.904 -77.619 48.315
165876 96.924 -77.620 48.290
165877 96.944 -77.623 48.264
165878 96.964 -77.624 48.238
165879 96.984 -77.627 48.212
165880 97.005 -77.628 48.187
165881 97.025 -77.630 48.162
165882 97.045 -77.632 48.138
165883 97.062 -77.633 48.114
165884 97.069 -77.633 48.091
165885 97.075 -77.638 48.068
165886 97.081 -77.643 48.047
165887 97.087 -77.649 48.026
165888 97.108 -77.656 48.007
165889 97.136 -77.664 47.988
165890 97.163 -77.673 47.968
165891 97.191 -77.682 47.948
165892 97.219 -77.690 47.928
165893 97.246 -77.699 47.906
165894 97.274 -77.708 47.883
165895 97.301 -77.716 47.861
165896 97.329 -77.724 47.838
165897 97.357 -77.733 47.814
165898 97.384 -77.740 47.791
165899 97.409 -77.737 47.767
165900 97.434 -77.738 47.742
165901 97.457 -77.748 47.717
165902 97.481 -77.758 47.693
165903 97.505 -77.768 47.669
165904 97.528 -77.777 47.644
165905 97.551 -77.787 47.620
165906 97.575 -77.797 47.595
165907 97.597 -77.805 47.570
165908 97.616 -77.812 47.545
165909 97.636 -77.819 47.519
165910 97.656 -77.826 47.493
165911 97.675 -77.833 47.469
165912 97.695 -77.840 47.446
165913 97.728 -77.849 47.423
165914 97.755 -77.840 47.399
165915 97.772 -77.809 47.374
165916 97.788 -77.778 47.350
165917 97.804 -77.748 47.327
165918 97.815 -77.724 47.304
165919 97.825 -77.701 47.279
165920 97.837 -77.717 47.253
165921 97.830 -77.745 47.227
165922 97.897 -77.749 47.200
165923 97.872 -77.755 47.173
165924 97.851 -77.762 47.146
165925 97.870 -77.784 47.120
165926 97.889 -77.807 47.094
165927 97.909 -77.826 47.070
165928 97.926 -77.827 47.046
165929 97.943 -77.827 47.021
165930 97.960 -77.828 46.997
165931 97.977 -77.829 46.971
165932 97.994 -77.830 46.944
165933 98.011 -77.830 46.920
165934 98.027 -77.831 46.894
165935 98.044 -77.832 46.869
165936 98.062 -77.833 46.843
165937 98.072 -77.842 46.816
165938 98.083 -77.852 46.790
165939 98.093 -77.862 46.764
165940 98.105 -77.871 46.737
165941 98.123 -77.877 46.713
165942 98.142 -77.882 46.688
165943 98.160 -77.888 46.663
165944 98.178 -77.893 46.639
165945 98.197 -77.899 46.614
165946 98.228 -77.886 46.589
165947 98.260 -77.872 46.565
165948 98.285 -77.860 46.543
165949 98.292 -77.855 46.522
165950 98.298 -77.849 46.500
165951 98.305 -77.844 46.477
165952 98.312 -77.839 46.455
165953 98.323 -77.835 46.433
165954 98.339 -77.833 46.412
165955 98.356 -77.830 46.393
165956 98.387 -77.835 46.374
165957 98.432 -77.846 46.355
165958 98.476 -77.856 46.337
165959 98.503 -77.860 46.318
165960 98.513 -77.856 46.298
165961 98.516 -77.856 46.280
165962 98.520 -77.856 46.263
165963 98.526 -77.859 46.245
165964 98.545 -77.866 46.228
165965 98.563 -77.873 46.211
165966 98.582 -77.881 46.194
165967 98.600 -77.888 46.178
165968 98.619 -77.896 46.161
165969 98.638 -77.902 46.144
165970 98.656 -77.910 46.125
165971 98.675 -77.918 46.106
165972 98.694 -77.924 46.087
165973 98.712 -77.932 46.068
165974 98.731 -77.940 46.048
165975 98.749 -77.947 46.028
165976 98.768 -77.954 46.009
165977 98.800 -77.962 45.990
165978 98.842 -77.970 45.972
165979 98.899 -77.980 45.954
165980 98.980 -77.991 45.936
165981 99.062 -78.003 45.918
165982 99.143 -78.015 45.901
165983 99.186 -78.008 45.883
165984 99.217 -78.000 45.866
165985 99.249 -77.990 45.850
165986 99.280 -77.982 45.834
165987 99.311 -77.974 45.818
165988 99.333 -77.981 45.802
165989 99.346 -77.988 45.789
165990 99.355 -77.995 45.777
165991 99.363 -78.003 45.767
165992 99.372 -78.010 45.757
165993 99.380 -78.017 45.747
165994 99.390 -78.025 45.737
165995 99.398 -78.031 45.727
165996 99.407 -78.039 45.716
165997 99.415 -78.046 45.707
165998 99.424 -78.053 45.697
165999 99.432 -78.060 45.686
166000 99.441 -78.068 45.674
166001 99.461 -78.076 45.664
166002 99.492 -78.086 45.654
166003 99.524 -78.095 45.644
166004 99.555 -78.105 45.632
166005 99.586 -78.115 45.621
166006 99.613 -78.125 45.610
166007 99.641 -78.138 45.597
166008 99.671 -78.153 45.585
166009 99.702 -78.169 45.573
166010 99.726 -78.183 45.560
166011 99.740 -78.194 45.545
166012 99.755 -78.206 45.530
166013 99.768 -78.218 45.515
166014 99.782 -78.229 45.499
166015 99.796 -78.241 45.484
166016 99.810 -78.253 45.469
166017 99.825 -78.264 45.454
166018 99.838 -78.275 45.438
166019 99.852 -78.287 45.422
166020 99.871 -78.301 45.407
166021 99.892 -78.314 45.391
166022 99.916 -78.326 45.375
166023 99.940 -78.338 45.358
166024 99.963 -78.353 45.340
166025 99.983 -78.371 45.323
166026 100.003 -78.389 45.306
166027 100.023 -78.408 45.288
166028 100.044 -78.426 45.272
166029 100.064 -78.445 45.257
166030 100.109 -78.460 45.242
166031 100.159 -78.474 45.226
166032 100.196 -78.484 45.209
166033 100.233 -78.494 45.193
166034 100.270 -78.504 45.176
166035 100.307 -78.513 45.160
166036 100.343 -78.523 45.143
166037 100.380 -78.532 45.127
166038 100.417 -78.542 45.112
166039 100.454 -78.552 45.097
166040 100.490 -78.561 45.083
166041 100.527 -78.571 45.069
166042 100.565 -78.581 45.055
166043 100.610 -78.592 45.040
166044 100.654 -78.604 45.026
166045 100.674 -78.573 45.011
166046 100.699 -78.581 44.996
166047 100.725 -78.592 44.981
166048 100.755 -78.604 44.965
166049 100.783 -78.617 44.949
166050 100.813 -78.629 44.934
166051 100.842 -78.641 44.918
166052 100.871 -78.654 44.902
166053 100.899 -78.670 44.886
166054 100.930 -78.680 44.871
166055 100.958 -78.677 44.855
166056 100.986 -78.674 44.839
166057 101.014 -78.671 44.824
166058 101.043 -78.673 44.809
166059 101.072 -78.677 44.794
166060 101.101 -78.679 44.778
166061 101.120 -78.717 44.762
166062 101.159 -78.745 44.747
166063 101.199 -78.773 44.732
166064 101.239 -78.801 44.717
166065 101.279 -78.827 44.703
166066 101.314 -78.836 44.689
166067 101.326 -78.839 44.674
166068 101.337 -78.843 44.659
166069 101.347 -78.855 44.645
166070 101.363 -78.888 44.629
166071 101.379 -78.922 44.615
166072 101.402 -78.953 44.601
166073 101.445 -78.972 44.589
166074 101.487 -78.992 44.576
166075 101.529 -79.011 44.562
166076 101.571 -79.030 44.550
166077 101.613 -79.049 44.537
166078 101.656 -79.068 44.524
166079 101.678 -79.078 44.511
166080 101.695 -79.083 44.501
166081 101.711 -79.088 44.488
166082 101.728 -79.094 44.475
166083 101.745 -79.100 44.463
166305 92.427 -77.244 53.989
166306 92.450 -77.255 54.011
166307 92.473 -77.266 54.033
166308 92.496 -77.277 54.054
166309 92.519 -77.288 54.075
166310 92.542 -77.299 54.095
166311 92.565 -77.310 54.113
166312 92.588 -77.321 54.130
166313 92.612 -77.332 54.148
166314 92.635 -77.344 54.163
166315 92.657 -77.354 54.177
166316 92.681 -77.366 54.190
166317 92.704 -77.376 54.202
166318 92.727 -77.388 54.213
166319 92.750 -77.399 54.223
166320 92.762 -77.411 54.232
166321 92.769 -77.423 54.243
166322 92.776 -77.436 54.251
166323 92.783 -77.448 54.261
166324 92.790 -77.461 54.270
166325 92.797 -77.473 54.278
166326 92.819 -77.471 54.292
166327 92.844 -77.463 54.305
166328 92.872 -77.455 54.318
166329 92.901 -77.448 54.327
166330 92.930 -77.441 54.333
166331 92.959 -77.434 54.337
166332 92.985 -77.423 54.337
166333 93.010 -77.412 54.341
166334 93.036 -77.401 54.346
166335 93.061 -77.389 54.352
166336 93.076 -77.376 54.361
166337 93.086 -77.361 54.369
166338 93.095 -77.347 54.379
166339 93.106 -77.332 54.390
166340 93.115 -77.319 54.395
166341 93.125 -77.304 54.397
166342 93.135 -77.290 54.398
166343 93.145 -77.276 54.397
166344 93.155 -77.262 54.392
166345 93.165 -77.248 54.384
166346 93.174 -77.233 54.376
166347 93.185 -77.219 54.370
166348 93.201 -77.209 54.362
166349 93.219 -77.201 54.354
166350 93.235 -77.193 54.348
166351 93.253 -77.185 54.345
166352 93.270 -77.178 54.344
166353 93.286 -77.171 54.343
166354 93.303 -77.163 54.342
166355 93.320 -77.156 54.340
166356 93.336 -77.148 54.337
166357 93.346 -77.138 54.333
166358 93.356 -77.128 54.326
166359 93.386 -77.122 54.317
166360 93.417 -77.115 54.306
166361 93.446 -77.107 54.294
166362 93.470 -77.097 54.284
166363 93.494 -77.088 54.273
166364 93.517 -77.078 54.262
166365 93.542 -77.068 54.252
166366 93.565 -77.058 54.247
166367 93.584 -77.056 54.242
166368 93.588 -77.072 54.237
166369 93.599 -77.071 54.233
166370 93.612 -77.066 54.225
166371 93.633 -77.055 54.214
166372 93.654 -77.043 54.202
166373 93.674 -77.032 54.189
166374 93.687 -77.028 54.176
166375 93.699 -77.025 54.164
166376 93.712 -77.022 54.152
166377 93.725 -77.018 54.140
166378 93.729 -77.007 54.131
166379 93.733 -76.997 54.121
166380 93.737 -76.986 54.109
166381 93.742 -76.976 54.096
166382 93.748 -76.969 54.083
166383 93.756 -76.968 54.071
166384 93.765 -76.967 54.061
166385 93.773 -76.965 54.054
166386 93.786 -76.968 54.048
166387 93.804 -76.972 54.043
166388 93.832 -76.981 54.031
166389 93.860 -76.988 54.018
166390 93.887 -76.988 54.002
166391 93.912 -76.984 53.985
166392 93.937 -76.980 53.969
166393 93.960 -76.974 53.951
166394 93.981 -76.967 53.932
166395 94.001 -76.959 53.910
166396 94.018 -76.949 53.887
166397 94.036 -76.937 53.865
166398 94.053 -76.927 53.843
166399 94.070 -76.916 53.821
166400 94.086 -76.908 53.801
166401 94.098 -76.906 53.781
166402 94.110 -76.903 53.760
166403 94.121 -76.900 53.745
166404 94.133 -76.897 53.732
166405 94.145 -76.895 53.722
166406 94.159 -76.892 53.710
166407 94.173 -76.889 53.695
166408 94.189 -76.886 53.677
166409 94.205 -76.883 53.656
166410 94.221 -76.881 53.634
166411 94.235 -76.878 53.611
166412 94.248 -76.875 53.590
166413 94.260 -76.872 53.570
166414 94.272 -76.870 53.553
166415 94.284 -76.867 53.537
166416 94.296 -76.864 53.522
166417 94.308 -76.862 53.510
166418 94.314 -76.863 53.501
166419 94.338 -76.857 53.494
166420 94.355 -76.854 53.487
166421 94.372 -76.852 53.481
166422 94.390 -76.850 53.471
166423 94.407 -76.846 53.460
166424 94.424 -76.844 53.449
166425 94.442 -76.842 53.437
166426 94.459 -76.839 53.421
166427 94.474 -76.837 53.403
166428 94.488 -76.834 53.382
166429 94.503 -76.833 53.358
166430 94.518 -76.831 53.334
166431 94.533 -76.828 53.310
166432 94.547 -76.826 53.283
166433 94.562 -76.824 53.258
166434 94.568 -76.827 53.231
166435 94.559 -76.831 53.202
166436 94.573 -76.828 53.169
166437 94.609 -76.819 53.137
166438 94.643 -76.811 53.105
166439 94.678 -76.802 53.074
166440 94.713 -76.795 53.042
166441 94.748 -76.787 53.010
166442 94.796 -76.769 52.979
166443 94.844 -76.752 52.945
166444 94.864 -76.744 52.909
166445 94.877 -76.738 52.871
166446 94.888 -76.732 52.831
166447 94.900 -76.726 52.790
166448 94.912 -76.720 52.748
166449 94.922 -76.718 52.705
166450 94.933 -76.716 52.662
166451 94.944 -76.715 52.620
166452 94.954 -76.713 52.580
166453 94.965 -76.712 52.543
166454 94.975 -76.709 52.512
166455 94.985 -76.708 52.485
166456 94.996 -76.706 52.461
166457 95.006 -76.705 52.437
166458 95.022 -76.706 52.416
166459 95.043 -76.712 52.394
166460 95.064 -76.718 52.372
166461 95.085 -76.724 52.347
166462 95.109 -76.732 52.316
166463 95.146 -76.744 52.280
166464 95.183 -76.756 52.236
166465 95.219 -76.768 52.185
166466 95.256 -76.781 52.130
166467 95.293 -76.793 52.073
166468 95.329 -76.805 52.016
166469 95.366 -76.817 51.958
166470 95.349 -76.846 51.899
166471 95.332 -76.877 51.838
166472 95.315 -76.906 51.777
166473 95.301 -76.934 51.718
166474 95.315 -76.943 51.661
166475 95.329 -76.953 51.605
166476 95.344 -76.963 51.551
166477 95.361 -76.975 51.496
166478 95.379 -76.987 51.442
166479 95.395 -76.999 51.388
166480 95.412 -77.010 51.336
166481 95.429 -77.022 51.285
166482 95.446 -77.034 51.236
166483 95.463 -77.046 51.189
166484 95.480 -77.057 51.141
166485 95.497 -77.069 51.094
166486 95.514 -77.081 51.047
166487 95.532 -77.096 51.002
166488 95.564 -77.106 50.957
166489 95.602 -77.116 50.912
166490 95.630 -77.117 50.869
166491 95.651 -77.116 50.824
166492 95.664 -77.117 50.780
166493 95.675 -77.121 50.737
166494 95.686 -77.125 50.696
166495 95.696 -77.128 50.658
166496 95.707 -77.132 50.621
166497 95.718 -77.136 50.585
166498 95.726 -77.149 50.548
166499 95.732 -77.172 50.510
166500 95.738 -77.195 50.470
166501 95.743 -77.219 50.430
166502 95.750 -77.230 50.384
166503 95.760 -77.229 50.339
166504 95.769 -77.228 50.295
166505 95.792 -77.236 50.253
166506 95.824 -77.252 50.211
166507 95.853 -77.257 50.169
166508 95.879 -77.246 50.126
166509 95.904 -77.235 50.086
166510 95.930 -77.225 50.044
166511 95.955 -77.213 50.000
166512 95.951 -77.219 49.954
166513 95.943 -77.228 49.904
166514 95.934 -77.236 49.847
166515 95.926 -77.244 49.784
166516 95.917 -77.253 49.717
166517 95.909 -77.261 49.654
166518 95.900 -77.269 49.595
166519 95.899 -77.279 49.540
166520 95.911 -77.291 49.488
166521 95.927 -77.298 49.441
166522 95.944 -77.304 49.396
166523 95.960 -77.309 49.353
166524 95.976 -77.315 49.313
166525 95.992 -77.321 49.276
166526 96.009 -77.327 49.241
166527 96.028 -77.334 49.209
166528 96.049 -77.342 49.177
166529 96.070 -77.350 49.148
166530 96.091 -77.357 49.121
166531 96.112 -77.365 49.093
166532 96.132 -77.373 49.070
166533 96.154 -77.381 49.048
166534 96.174 -77.389 49.026
166535 96.195 -77.391 49.004
166536 96.215 -77.392 48.981
166537 96.236 -77.392 48.959
166538 96.256 -77.393 48.938
166539 96.276 -77.394 48.920
166540 96.301 -77.401 48.900
166541 96.325 -77.408 48.880
166542 96.350 -77.415 48.863
166543 96.374 -77.422 48.847
166544 96.399 -77.429 48.833
166545 96.420 -77.440 48.820
166546 96.436 -77.455 48.814
166547 96.453 -77.470 48.806
166548 96.469 -77.486 48.798
166549 96.485 -77.501 48.792
166550 96.498 -77.521 48.782
166551 96.508 -77.543 48.769
166552 96.519 -77.565 48.755
166553 96.529 -77.587 48.739
166554 96.541 -77.609 48.721
166555 96.558 -77.630 48.703
166556 96.575 -77.652 48.686
166557 96.592 -77.674 48.670
166558 96.606 -77.680 48.653
166559 96.618 -77.678 48.636
166560 96.630 -77.677 48.620
166561 96.642 -77.674 48.605
166562 96.654 -77.673 48.589
166563 96.665 -77.671 48.572
166564 96.680 -77.668 48.556
166565 96.701 -77.664 48.538
166566 96.722 -77.661 48.521
166567 96.743 -77.656 48.506
166568 96.764 -77.652 48.489
166569 96.786 -77.649 48.472
166570 96.810 -77.654 48.455
166571 96.840 -77.672 48.438
166572 96.867 -77.685 48.418
166573 96.880 -77.666 48.398
166574 96.893 -77.647 48.377
166575 96.905 -77.627 48.356
166576 96.918 -77.608 48.332
166577 96.936 -77.603 48.307
166578 96.956 -77.605 48.282
166579 96.976 -77.607 48.257
166580 96.996 -77.609 48.232
166581 97.016 -77.611 48.207
166582 97.036 -77.613 48.182
166583 97.056 -77.614 48.158
166584 97.070 -77.614 48.135
166585 97.078 -77.614 48.112
166586 97.084 -77.620 48.089
166587 97.090 -77.625 48.067
166588 97.103 -77.631 48.046
166589 97.131 -77.639 48.026
166590 97.158 -77.649 48.007
166591 97.186 -77.657 47.987
166592 97.214 -77.665 47.966
166593 97.241 -77.674 47.946
166594 97.268 -77.683 47.923
166595 97.296 -77.691 47.900
166596 97.324 -77.700 47.877
166597 97.352 -77.708 47.854
166598 97.379 -77.717 47.830
166599 97.406 -77.726 47.807
166600 97.434 -77.734 47.784
166601 97.459 -77.740 47.759
166602 97.481 -77.749 47.734
166603 97.501 -77.755 47.709
166604 97.520 -77.762 47.685
166605 97.540 -77.769 47.660
166606 97.560 -77.776 47.635
166607 97.579 -77.783 47.611
166608 97.599 -77.789 47.585
166609 97.618 -77.796 47.559
166610 97.638 -77.802 47.532
166611 97.658 -77.809 47.505
166612 97.677 -77.816 47.480
166613 97.706 -77.825 47.455
166614 97.744 -77.837 47.431
166615 97.779 -77.841 47.406
166616 97.796 -77.811 47.382
166617 97.812 -77.780 47.358
166618 97.828 -77.749 47.334
166619 97.845 -77.718 47.310
166620 97.839 -77.693 47.285
166621 97.839 -77.721 47.260
166622 97.829 -77.750 47.234
166623 97.852 -77.746 47.206
166624 97.835 -77.742 47.179
166625 97.854 -77.762 47.152
166626 97.888 -77.784 47.125
166627 97.923 -77.805 47.097
166628 97.918 -77.815 47.073
166629 97.934 -77.816 47.048
166630 97.953 -77.816 47.023
166631 97.972 -77.817 46.999
166632 97.991 -77.818 46.974
166633 98.009 -77.818 46.948
166634 98.026 -77.819 46.923
166635 98.043 -77.820 46.898
166636 98.060 -77.821 46.873
166637 98.077 -77.821 46.848
166638 98.091 -77.827 46.821
166639 98.101 -77.837 46.795
166640 98.115 -77.845 46.769
166641 98.134 -77.850 46.743
166642 98.152 -77.856 46.718
166643 98.171 -77.862 46.693
166644 98.190 -77.868 46.669
166645 98.208 -77.873 46.644
166646 98.226 -77.879 46.620
166647 98.245 -77.884 46.596
166648 98.277 -77.870 46.572
166649 98.308 -77.855 46.550
166650 98.315 -77.850 46.528
166651 98.321 -77.845 46.505
166652 98.328 -77.840 46.483
166653 98.339 -77.836 46.461
166654 98.355 -77.834 46.439
166655 98.372 -77.831 46.417
166656 98.388 -77.829 46.398
166657 98.423 -77.835 46.378
166658 98.467 -77.846 46.359
166659 98.506 -77.854 46.340
166660 98.521 -77.852 46.321
166661 98.528 -77.850 46.301
166662 98.531 -77.850 46.283
166663 98.534 -77.850 46.267
166664 98.538 -77.851 46.249
166665 98.547 -77.854 46.231
166666 98.565 -77.861 46.214
166667 98.584 -77.868 46.196
166668 98.602 -77.876 46.180
166669 98.621 -77.884 46.164
166670 98.640 -77.890 46.148
166671 98.658 -77.898 46.130
166672 98.677 -77.906 46.111
166673 98.696 -77.913 46.092
166674 98.714 -77.920 46.073
166675 98.733 -77.928 46.053
166676 98.751 -77.935 46.033
166677 98.770 -77.942 46.014
166678 98.812 -77.950 45.996
166679 98.854 -77.959 45.978
166680 98.912 -77.968 45.959
166681 98.993 -77.980 45.941
166682 99.074 -77.991 45.924
166683 99.144 -78.001 45.907
166684 99.169 -78.004 45.889
166685 99.198 -77.999 45.872
166686 99.229 -77.990 45.855
166687 99.261 -77.981 45.839
166688 99.292 -77.973 45.823
166689 99.316 -77.981 45.807
166690 99.336 -77.987 45.794
166691 99.349 -77.995 45.783
166692 99.361 -78.002 45.772
166693 99.370 -78.009 45.762
166694 99.378 -78.017 45.752
166695 99.387 -78.024 45.742
166696 99.396 -78.031 45.732
166697 99.404 -78.038 45.722
166698 99.413 -78.046 45.712
166699 99.421 -78.053 45.703
166700 99.430 -78.060 45.692
166701 99.439 -78.067 45.681
166702 99.463 -78.076 45.671
166703 99.494 -78.086 45.662
166704 99.526 -78.096 45.651
166705 99.557 -78.105 45.640
166706 99.584 -78.116 45.629
166707 99.611 -78.127 45.617
166708 99.640 -78.140 45.605
166709 99.671 -78.156 45.592
166710 99.701 -78.172 45.578
166711 99.732 -78.187 45.566
166712 99.750 -78.200 45.550
166713 99.764 -78.211 45.535
166714 99.778 -78.223 45.520
166715 99.792 -78.235 45.505
166716 99.806 -78.246 45.490
166717 99.820 -78.257 45.475
166718 99.834 -78.269 45.459
166719 99.848 -78.281 45.443
166720 99.865 -78.294 45.428
166721 99.885 -78.307 45.412
166722 99.904 -78.322 45.396
166723 99.927 -78.334 45.379
166724 99.950 -78.346 45.362
166725 99.975 -78.358 45.345
166726 99.998 -78.370 45.327
166727 100.022 -78.383 45.310
166728 100.044 -78.400 45.294
166729 100.064 -78.418 45.278
166730 100.084 -78.436 45.263
166731 100.130 -78.451 45.247
166732 100.168 -78.461 45.232
166733 100.205 -78.471 45.216
166734 100.242 -78.481 45.199
166735 100.279 -78.490 45.183
166736 100.316 -78.500 45.166
166737 100.352 -78.510 45.151
166738 100.389 -78.520 45.135
166739 100.426 -78.529 45.119
166740 100.463 -78.539 45.104
166741 100.499 -78.548 45.089
166742 100.536 -78.558 45.076
166743 100.573 -78.568 45.063
166744 100.614 -78.579 45.048
166745 100.657 -78.586 45.034
166746 100.674 -78.552 45.019
166747 100.700 -78.562 45.003
166748 100.729 -78.575 44.988
166749 100.758 -78.587 44.973
166750 100.787 -78.599 44.957
166751 100.816 -78.615 44.942
166752 100.844 -78.633 44.926
166753 100.873 -78.649 44.910
166754 100.901 -78.667 44.894
166755 100.933 -78.672 44.879
166756 100.963 -78.670 44.863
166757 100.991 -78.667 44.847
166758 101.019 -78.664 44.832
166759 101.047 -78.662 44.817
166760 101.075 -78.659 44.801
166761 101.104 -78.657 44.785
166762 101.142 -78.674 44.769
166763 101.181 -78.703 44.754
166764 101.220 -78.732 44.739
166765 101.260 -78.761 44.725
166766 101.299 -78.789 44.711
166767 101.322 -78.815 44.696
166768 101.331 -78.833 44.681
166769 101.340 -78.844 44.666
166770 101.349 -78.855 44.651
166771 101.361 -78.875 44.636
166772 101.378 -78.909 44.621
166773 101.402 -78.940 44.607
166774 101.444 -78.959 44.594
166775 101.486 -78.978 44.581
166776 101.529 -78.997 44.568
166777 101.571 -79.016 44.554
166778 101.613 -79.036 44.542
166779 101.655 -79.055 44.529
166780 101.679 -79.064 44.517
166781 101.695 -79.070 44.506
166782 101.712 -79.075 44.494
166783 101.729 -79.081 44.481
166784 101.746 -79.087 44.469
166785 101.762 -79.092 44.457
167009 92.484 -77.248 54.035
167010 92.507 -77.259 54.056
167011 92.529 -77.270 54.077
167012 92.553 -77.281 54.097
167013 92.576 -77.292 54.116
167014 92.599 -77.304 54.134
167015 92.622 -77.314 54.150
167016 92.645 -77.326 54.165
167017 92.668 -77.336 54.179
167018 92.691 -77.348 54.193
167019 92.714 -77.359 54.203
167020 92.737 -77.370 54.214
167021 92.755 -77.382 54.224
167022 92.762 -77.394 54.235
167023 92.770 -77.407 54.245
167024 92.777 -77.419 54.255
167025 92.784 -77.432 54.264
167026 92.791 -77.444 54.274
167027 92.798 -77.457 54.287
167028 92.815 -77.457 54.302
167029 92.844 -77.449 54.316
167030 92.872 -77.442 54.324
167031 92.899 -77.432 54.330
167032 92.924 -77.420 54.333
167033 92.950 -77.409 54.333
167034 92.975 -77.398 54.333
167035 93.001 -77.386 54.335
167036 93.026 -77.374 54.340
167037 93.052 -77.363 54.348
167038 93.077 -77.351 54.356
167039 93.090 -77.338 54.366
167040 93.099 -77.323 54.377
167041 93.110 -77.309 54.379
167042 93.119 -77.295 54.381
167043 93.129 -77.281 54.383
167044 93.139 -77.266 54.382
167045 93.149 -77.252 54.377
167046 93.159 -77.238 54.369
167047 93.168 -77.224 54.359
167048 93.179 -77.210 54.354
167049 93.195 -77.200 54.348
167050 93.213 -77.191 54.343
167051 93.230 -77.183 54.338
167052 93.247 -77.175 54.337
167053 93.264 -77.169 54.337
167054 93.281 -77.161 54.337
167055 93.298 -77.153 54.338
167056 93.313 -77.145 54.335
167057 93.323 -77.135 54.332
167058 93.333 -77.125 54.327
167059 93.343 -77.115 54.319
167060 93.373 -77.109 54.309
167061 93.406 -77.103 54.299
167062 93.435 -77.095 54.288
167063 93.465 -77.088 54.278
167064 93.490 -77.079 54.268
167065 93.514 -77.069 54.258
167066 93.538 -77.059 54.251
167067 93.562 -77.050 54.247
167068 93.572 -77.060 54.242
167069 93.582 -77.060 54.236
167070 93.592 -77.059 54.230
167071 93.607 -77.053 54.221
167072 93.628 -77.041 54.210
167073 93.649 -77.029 54.197
167074 93.664 -77.023 54.184
167075 93.677 -77.019 54.172
167076 93.689 -77.016 54.161
167077 93.702 -77.012 54.150
167078 93.715 -77.009 54.139
167079 93.724 -77.002 54.131
167080 93.728 -76.991 54.124
167081 93.732 -76.981 54.114
167082 93.737 -76.971 54.102
167083 93.741 -76.960 54.090
167084 93.746 -76.950 54.079
167085 93.752 -76.944 54.071
167086 93.760 -76.943 54.064
167087 93.774 -76.946 54.058
167088 93.803 -76.954 54.054
167089 93.831 -76.962 54.046
167090 93.859 -76.970 54.036
167091 93.888 -76.978 54.024
167092 93.913 -76.975 54.008
167093 93.938 -76.971 53.992
167094 93.960 -76.965 53.974
167095 93.979 -76.956 53.955
167096 93.996 -76.944 53.932
167097 94.013 -76.934 53.910
167098 94.031 -76.923 53.888
167099 94.048 -76.912 53.866
167100 94.065 -76.901 53.845
167101 94.080 -76.893 53.825
167102 94.092 -76.891 53.804
167103 94.104 -76.889 53.784
167104 94.115 -76.886 53.767
167105 94.127 -76.884 53.754
167106 94.141 -76.881 53.744
167107 94.155 -76.877 53.733
167108 94.169 -76.874 53.719
167109 94.184 -76.871 53.703
167110 94.200 -76.868 53.684
167111 94.216 -76.865 53.662
167112 94.229 -76.863 53.639
167113 94.241 -76.860 53.616
167114 94.253 -76.858 53.596
167115 94.265 -76.855 53.579
167116 94.277 -76.853 53.564
167117 94.284 -76.853 53.548
167118 94.290 -76.855 53.534
167119 94.303 -76.853 53.522
167120 94.333 -76.843 53.513
167121 94.351 -76.840 53.506
167122 94.368 -76.837 53.500
167123 94.386 -76.835 53.491
167124 94.403 -76.833 53.478
167125 94.420 -76.830 53.466
167126 94.438 -76.828 53.453
167127 94.456 -76.824 53.437
167128 94.473 -76.822 53.419
167129 94.487 -76.820 53.399
167130 94.502 -76.818 53.377
167131 94.517 -76.815 53.354
167132 94.532 -76.813 53.330
167133 94.546 -76.812 53.305
167134 94.561 -76.809 53.279
167135 94.561 -76.816 53.252
167136 94.561 -76.822 53.222
167137 94.554 -76.825 53.190
167138 94.589 -76.817 53.158
167139 94.624 -76.809 53.126
167140 94.659 -76.801 53.095
167141 94.694 -76.792 53.063
167142 94.746 -76.772 53.032
167143 94.797 -76.752 53.003
167144 94.841 -76.738 52.968
167145 94.860 -76.731 52.932
167146 94.873 -76.725 52.893
167147 94.884 -76.719 52.852
167148 94.896 -76.713 52.811
167149 94.908 -76.707 52.768
167150 94.920 -76.701 52.725
167151 94.931 -76.694 52.681
167152 94.944 -76.689 52.635
167153 94.954 -76.687 52.591
167154 94.964 -76.685 52.552
167155 94.975 -76.683 52.518
167156 94.985 -76.682 52.487
167157 94.996 -76.680 52.458
167158 95.006 -76.679 52.431
167159 95.018 -76.677 52.409
167160 95.039 -76.683 52.386
167161 95.060 -76.690 52.362
167162 95.081 -76.696 52.338
167163 95.117 -76.708 52.309
167164 95.153 -76.720 52.274
167165 95.190 -76.732 52.232
167166 95.227 -76.744 52.185
167167 95.263 -76.756 52.134
167168 95.299 -76.770 52.080
167169 95.333 -76.788 52.025
167170 95.348 -76.808 51.969
167171 95.347 -76.829 51.909
167172 95.343 -76.851 51.848
167173 95.326 -76.881 51.788
167174 95.309 -76.910 51.730
167175 95.311 -76.928 51.674
167176 95.325 -76.937 51.619
167177 95.342 -76.950 51.564
167178 95.358 -76.961 51.508
167179 95.376 -76.973 51.453
167180 95.393 -76.984 51.398
167181 95.409 -76.996 51.347
167182 95.426 -77.008 51.297
167183 95.444 -77.019 51.248
167184 95.460 -77.031 51.200
167185 95.477 -77.043 51.153
167186 95.494 -77.055 51.105
167187 95.511 -77.066 51.058
167188 95.529 -77.080 51.012
167189 95.567 -77.088 50.966
167190 95.605 -77.097 50.922
167191 95.635 -77.103 50.878
167192 95.645 -77.106 50.835
167193 95.656 -77.110 50.791
167194 95.667 -77.113 50.748
167195 95.678 -77.117 50.707
167196 95.688 -77.121 50.671
167197 95.699 -77.125 50.635
167198 95.710 -77.128 50.599
167199 95.718 -77.142 50.562
167200 95.724 -77.165 50.525
167201 95.730 -77.188 50.487
167202 95.735 -77.212 50.446
167203 95.741 -77.235 50.400
167204 95.748 -77.246 50.355
167205 95.758 -77.244 50.311
167206 95.782 -77.255 50.269
167207 95.813 -77.262 50.228
167208 95.838 -77.251 50.186
167209 95.863 -77.240 50.145
167210 95.889 -77.229 50.105
167211 95.914 -77.218 50.064
167212 95.939 -77.207 50.021
167213 95.951 -77.204 49.975
167214 95.942 -77.213 49.926
167215 95.934 -77.221 49.868
167216 95.925 -77.229 49.804
167217 95.916 -77.238 49.735
167218 95.908 -77.246 49.670
167219 95.899 -77.254 49.609
167220 95.908 -77.266 49.552
167221 95.919 -77.279 49.500
167222 95.936 -77.285 49.451
167223 95.952 -77.291 49.405
167224 95.968 -77.296 49.361
167225 95.984 -77.302 49.320
167226 96.001 -77.308 49.283
167227 96.017 -77.314 49.248
167228 96.033 -77.320 49.217
167229 96.050 -77.326 49.186
167230 96.068 -77.333 49.157
167231 96.089 -77.341 49.129
167232 96.110 -77.348 49.103
167233 96.131 -77.356 49.081
167234 96.152 -77.363 49.060
167235 96.173 -77.372 49.040
167236 96.193 -77.373 49.020
167237 96.213 -77.373 48.997
167238 96.234 -77.374 48.972
167239 96.257 -77.379 48.948
167240 96.282 -77.386 48.927
167241 96.306 -77.393 48.907
167242 96.330 -77.401 48.887
167243 96.355 -77.407 48.870
167244 96.379 -77.414 48.854
167245 96.404 -77.422 48.839
167246 96.422 -77.436 48.826
167247 96.438 -77.451 48.817
167248 96.455 -77.466 48.809
167249 96.471 -77.481 48.802
167250 96.487 -77.496 48.795
167251 96.504 -77.512 48.786
167252 96.515 -77.533 48.774
167253 96.525 -77.555 48.760
167254 96.535 -77.577 48.745
167255 96.546 -77.600 48.727
167256 96.556 -77.622 48.710
167257 96.570 -77.643 48.693
167258 96.587 -77.665 48.678
167259 96.603 -77.680 48.662
167260 96.615 -77.679 48.646
167261 96.626 -77.677 48.631
167262 96.639 -77.675 48.616
167263 96.653 -77.673 48.599
167264 96.674 -77.668 48.583
167265 96.695 -77.664 48.567
167266 96.716 -77.661 48.550
167267 96.737 -77.657 48.534
167268 96.758 -77.652 48.519
167269 96.780 -77.649 48.503
167270 96.801 -77.645 48.486
167271 96.828 -77.658 48.469
167272 96.858 -77.677 48.452
167273 96.888 -77.695 48.433
167274 96.903 -77.680 48.414
167275 96.915 -77.661 48.394
167276 96.928 -77.642 48.372
167277 96.941 -77.623 48.348
167278 96.954 -77.603 48.324
167279 96.967 -77.588 48.299
167280 96.987 -77.589 48.275
167281 97.008 -77.592 48.251
167282 97.028 -77.593 48.227
167283 97.048 -77.595 48.202
167284 97.068 -77.597 48.178
167285 97.079 -77.596 48.155
167286 97.086 -77.596 48.132
167287 97.093 -77.602 48.109
167288 97.099 -77.607 48.087
167289 97.125 -77.615 48.065
167290 97.153 -77.624 48.045
167291 97.180 -77.632 48.026
167292 97.208 -77.641 48.005
167293 97.236 -77.649 47.984
167294 97.263 -77.658 47.963
167295 97.291 -77.667 47.939
167296 97.318 -77.675 47.916
167297 97.346 -77.683 47.893
167298 97.374 -77.692 47.870
167299 97.400 -77.701 47.847
167300 97.423 -77.711 47.823
167301 97.447 -77.721 47.799
167302 97.468 -77.729 47.775
167303 97.484 -77.733 47.750
167304 97.503 -77.739 47.725
167305 97.522 -77.746 47.700
167306 97.542 -77.752 47.675
167307 97.562 -77.759 47.651
167308 97.581 -77.766 47.626
167309 97.601 -77.773 47.598
167310 97.620 -77.780 47.572
167311 97.640 -77.787 47.544
167312 97.660 -77.793 47.517
167313 97.684 -77.801 47.490
167314 97.723 -77.812 47.464
167315 97.761 -77.824 47.439
167316 97.800 -77.836 47.413
167317 97.819 -77.812 47.390
167318 97.836 -77.781 47.366
167319 97.852 -77.750 47.342
167320 97.854 -77.724 47.317
167321 97.816 -77.707 47.293
167322 97.814 -77.727 47.267
167323 97.822 -77.754 47.241
167324 97.840 -77.758 47.213
167325 97.846 -77.757 47.185
167326 97.886 -77.765 47.157
167327 97.933 -77.782 47.129
167328 97.956 -77.800 47.102
167329 97.933 -77.804 47.077
167330 97.941 -77.805 47.052
167331 97.960 -77.805 47.026
167332 97.979 -77.806 47.001
167333 97.998 -77.807 46.976
167334 98.017 -77.808 46.951
167335 98.036 -77.808 46.926
167336 98.055 -77.809 46.901
167337 98.074 -77.809 46.877
167338 98.093 -77.810 46.852
167339 98.109 -77.812 46.826
167340 98.126 -77.818 46.801
167341 98.145 -77.824 46.776
167342 98.163 -77.830 46.750
167343 98.181 -77.836 46.726
167344 98.200 -77.841 46.701
167345 98.219 -77.847 46.676
167346 98.237 -77.852 46.652
167347 98.255 -77.859 46.627
167348 98.274 -77.865 46.603
167349 98.294 -77.867 46.579
167350 98.326 -77.853 46.555
167351 98.338 -77.846 46.533
167352 98.345 -77.841 46.511
167353 98.355 -77.837 46.488
167354 98.371 -77.834 46.466
167355 98.387 -77.832 46.444
167356 98.404 -77.830 46.423
167357 98.420 -77.827 46.402
167358 98.458 -77.834 46.383
167359 98.502 -77.845 46.363
167360 98.524 -77.847 46.344
167361 98.540 -77.843 46.324
167362 98.543 -77.843 46.305
167363 98.547 -77.844 46.287
167364 98.550 -77.844 46.270
167365 98.553 -77.845 46.253
167366 98.556 -77.845 46.235
167367 98.567 -77.849 46.217
167368 98.586 -77.856 46.199
167369 98.604 -77.864 46.182
167370 98.623 -77.871 46.166
167371 98.642 -77.878 46.150
167372 98.660 -77.886 46.134
167373 98.679 -77.893 46.115
167374 98.698 -77.901 46.096
167375 98.716 -77.908 46.076
167376 98.735 -77.915 46.056
167377 98.754 -77.923 46.036
167378 98.782 -77.931 46.017
167379 98.824 -77.938 46.000
167380 98.866 -77.946 45.982
167381 98.924 -77.956 45.964
167382 99.006 -77.968 45.947
167383 99.087 -77.979 45.929
167384 99.133 -77.986 45.912
167385 99.158 -77.989 45.894
167386 99.183 -77.992 45.877
167387 99.210 -77.990 45.860
167388 99.242 -77.981 45.843
167389 99.273 -77.973 45.828
167390 99.297 -77.980 45.813
167391 99.321 -77.987 45.799
167392 99.339 -77.994 45.787
167393 99.351 -78.002 45.776
167394 99.364 -78.009 45.767
167395 99.376 -78.016 45.756
167396 99.384 -78.024 45.746
167397 99.393 -78.031 45.737
167398 99.402 -78.038 45.727
167399 99.411 -78.045 45.718
167400 99.419 -78.053 45.708
167401 99.428 -78.059 45.699
167402 99.436 -78.067 45.688
167403 99.465 -78.076 45.678
167404 99.497 -78.086 45.669
167405 99.527 -78.096 45.659
167406 99.555 -78.106 45.648
167407 99.582 -78.117 45.636
167408 99.609 -78.128 45.624
167409 99.639 -78.143 45.612
167410 99.669 -78.159 45.598
167411 99.700 -78.174 45.585
167412 99.731 -78.190 45.570
167413 99.760 -78.205 45.554
167414 99.774 -78.216 45.539
167415 99.788 -78.228 45.524
167416 99.802 -78.240 45.509
167417 99.816 -78.251 45.494
167418 99.830 -78.263 45.479
167419 99.844 -78.275 45.464
167420 99.859 -78.287 45.448
167421 99.878 -78.300 45.432
167422 99.898 -78.314 45.416
167423 99.917 -78.328 45.400
167424 99.937 -78.341 45.383
167425 99.961 -78.354 45.366
167426 99.985 -78.366 45.349
167427 100.009 -78.378 45.331
167428 100.033 -78.390 45.315
167429 100.057 -78.403 45.299
167430 100.080 -78.415 45.285
167431 100.104 -78.428 45.269
167432 100.140 -78.438 45.253
167433 100.177 -78.448 45.237
167434 100.214 -78.458 45.221
167435 100.251 -78.468 45.205
167436 100.287 -78.477 45.189
167437 100.324 -78.487 45.173
167438 100.361 -78.497 45.157
167439 100.398 -78.507 45.142
167440 100.434 -78.517 45.126
167441 100.471 -78.526 45.111
167442 100.508 -78.535 45.096
167443 100.545 -78.545 45.082
167444 100.581 -78.555 45.069
167445 100.618 -78.565 45.055
167446 100.649 -78.523 45.041
167447 100.675 -78.533 45.025
167448 100.704 -78.545 45.010
167449 100.733 -78.560 44.995
167450 100.761 -78.577 44.980
167451 100.789 -78.594 44.965
167452 100.818 -78.611 44.949
167453 100.846 -78.628 44.933
167454 100.875 -78.645 44.917
167455 100.903 -78.662 44.902
167456 100.935 -78.664 44.886
167457 100.968 -78.663 44.870
167458 100.996 -78.661 44.855
167459 101.025 -78.658 44.839
167460 101.053 -78.655 44.823
167461 101.080 -78.652 44.808
167462 101.109 -78.647 44.792
167463 101.151 -78.657 44.776
167464 101.191 -78.686 44.760
167465 101.230 -78.714 44.746
167466 101.269 -78.743 44.732
167467 101.308 -78.772 44.718
167468 101.323 -78.797 44.702
167469 101.333 -78.821 44.687
167470 101.343 -78.844 44.673
167471 101.352 -78.856 44.658
167472 101.361 -78.868 44.642
167473 101.376 -78.896 44.628
167474 101.401 -78.926 44.614
167475 101.444 -78.945 44.601
167476 101.486 -78.964 44.587
167477 101.528 -78.984 44.573
167478 101.570 -79.002 44.560
167479 101.613 -79.022 44.547
167480 101.655 -79.041 44.534
167481 101.680 -79.051 44.522
167482 101.696 -79.056 44.511
167483 101.713 -79.062 44.499
167484 101.730 -79.068 44.486
167485 101.747 -79.074 44.474
167486 101.763 -79.079 44.462
167487 101.780 -79.084 44.451
167712 92.517 -77.241 54.060
167713 92.540 -77.252 54.080
167714 92.563 -77.263 54.100
167715 92.586 -77.274 54.119
167716 92.609 -77.285 54.137
167717 92.632 -77.296 54.153
167718 92.655 -77.307 54.167
167719 92.678 -77.319 54.181
167720 92.702 -77.329 54.194
167721 92.725 -77.341 54.205
167722 92.747 -77.352 54.216
167723 92.756 -77.364 54.227
167724 92.763 -77.377 54.239
167725 92.770 -77.389 54.249
167726 92.777 -77.402 54.258
167727 92.784 -77.414 54.268
167728 92.791 -77.427 54.282
167729 92.798 -77.439 54.297
167730 92.813 -77.441 54.312
167731 92.839 -77.429 54.319
167732 92.864 -77.417 54.324
167733 92.889 -77.406 54.326
167734 92.915 -77.395 54.326
167735 92.940 -77.383 54.324
167736 92.966 -77.371 54.324
167737 92.991 -77.360 54.327
167738 93.016 -77.348 54.333
167739 93.042 -77.337 54.340
167740 93.067 -77.326 54.348
167741 93.093 -77.313 54.358
167742 93.103 -77.300 54.359
167743 93.114 -77.285 54.360
167744 93.123 -77.271 54.360
167745 93.133 -77.257 54.359
167746 93.143 -77.243 54.355
167747 93.153 -77.229 54.350
167748 93.163 -77.214 54.344
167749 93.172 -77.201 54.340
167750 93.189 -77.191 54.337
167751 93.207 -77.182 54.335
167752 93.225 -77.173 54.333
167753 93.241 -77.166 54.333
167754 93.259 -77.159 54.334
167755 93.276 -77.151 54.334
167756 93.289 -77.143 54.334
167757 93.299 -77.132 54.332
167758 93.309 -77.122 54.328
167759 93.320 -77.112 54.322
167760 93.330 -77.101 54.312
167761 93.359 -77.096 54.303
167762 93.395 -77.091 54.293
167763 93.424 -77.084 54.283
167764 93.453 -77.076 54.272
167765 93.483 -77.069 54.262
167766 93.511 -77.061 54.254
167767 93.535 -77.051 54.250
167768 93.556 -77.045 54.247
167769 93.564 -77.050 54.241
167770 93.575 -77.048 54.234
167771 93.585 -77.047 54.226
167772 93.602 -77.039 54.216
167773 93.622 -77.028 54.205
167774 93.641 -77.018 54.192
167775 93.654 -77.014 54.181
167776 93.666 -77.010 54.170
167777 93.679 -77.007 54.160
167778 93.692 -77.003 54.151
167779 93.705 -77.000 54.141
167780 93.717 -76.996 54.133
167781 93.723 -76.987 54.126
167782 93.727 -76.976 54.118
167783 93.732 -76.965 54.107
167784 93.736 -76.955 54.097
167785 93.741 -76.944 54.086
167786 93.745 -76.934 54.079
167787 93.749 -76.923 54.072
167788 93.773 -76.928 54.066
167789 93.801 -76.936 54.062
167790 93.830 -76.943 54.055
167791 93.858 -76.952 54.046
167792 93.887 -76.960 54.037
167793 93.915 -76.967 54.025
167794 93.938 -76.962 54.011
167795 93.956 -76.952 53.995
167796 93.973 -76.940 53.976
167797 93.991 -76.930 53.953
167798 94.008 -76.919 53.931
167799 94.025 -76.909 53.909
167800 94.042 -76.897 53.888
167801 94.059 -76.887 53.869
167802 94.074 -76.880 53.850
167803 94.086 -76.877 53.830
167804 94.098 -76.875 53.809
167805 94.110 -76.871 53.791
167806 94.123 -76.868 53.777
167807 94.137 -76.865 53.767
167808 94.151 -76.862 53.757
167809 94.165 -76.859 53.745
167810 94.180 -76.856 53.731
167811 94.196 -76.854 53.712
167812 94.210 -76.851 53.691
167813 94.222 -76.848 53.668
167814 94.235 -76.846 53.644
167815 94.247 -76.843 53.622
167816 94.255 -76.843 53.606
167817 94.261 -76.844 53.591
167818 94.267 -76.846 53.575
167819 94.273 -76.847 53.559
167820 94.295 -76.841 53.545
167821 94.324 -76.832 53.534
167822 94.347 -76.826 53.526
167823 94.364 -76.823 53.520
167824 94.382 -76.821 53.510
167825 94.399 -76.818 53.497
167826 94.416 -76.816 53.484
167827 94.434 -76.813 53.469
167828 94.453 -76.810 53.453
167829 94.470 -76.807 53.436
167830 94.486 -76.805 53.418
167831 94.501 -76.803 53.397
167832 94.516 -76.801 53.375
167833 94.530 -76.799 53.351
167834 94.545 -76.796 53.325
167835 94.553 -76.798 53.299
167836 94.554 -76.805 53.272
167837 94.554 -76.812 53.242
167838 94.554 -76.818 53.211
167839 94.569 -76.815 53.179
167840 94.604 -76.807 53.147
167841 94.644 -76.795 53.115
167842 94.696 -76.775 53.085
167843 94.747 -76.755 53.055
167844 94.799 -76.736 53.025
167845 94.837 -76.725 52.989
167846 94.857 -76.718 52.951
167847 94.868 -76.712 52.912
167848 94.881 -76.705 52.871
167849 94.892 -76.699 52.830
167850 94.904 -76.693 52.787
167851 94.916 -76.687 52.743
167852 94.928 -76.681 52.696
167853 94.939 -76.675 52.650
167854 94.952 -76.669 52.606
167855 94.963 -76.663 52.565
167856 94.975 -76.658 52.529
167857 94.985 -76.656 52.494
167858 94.996 -76.655 52.462
167859 95.006 -76.652 52.432
167860 95.017 -76.651 52.405
167861 95.035 -76.655 52.379
167862 95.056 -76.661 52.354
167863 95.087 -76.671 52.329
167864 95.124 -76.683 52.300
167865 95.161 -76.695 52.267
167866 95.197 -76.708 52.228
167867 95.231 -76.726 52.184
167868 95.264 -76.745 52.136
167869 95.298 -76.763 52.086
167870 95.325 -76.783 52.033
167871 95.323 -76.803 51.976
167872 95.322 -76.824 51.916
167873 95.321 -76.845 51.856
167874 95.319 -76.865 51.798
167875 95.318 -76.886 51.742
167876 95.306 -76.912 51.687
167877 95.322 -76.923 51.633
167878 95.339 -76.935 51.576
167879 95.356 -76.947 51.519
167880 95.373 -76.959 51.464
167881 95.390 -76.970 51.410
167882 95.407 -76.981 51.359
167883 95.423 -76.991 51.309
167884 95.441 -77.002 51.261
167885 95.458 -77.012 51.212
167886 95.474 -77.023 51.164
167887 95.491 -77.033 51.117
167888 95.508 -77.044 51.069
167889 95.530 -77.053 51.022
167890 95.569 -77.064 50.976
167891 95.603 -77.060 50.931
167892 95.624 -77.091 50.887
167893 95.637 -77.099 50.845
167894 95.648 -77.103 50.802
167895 95.659 -77.106 50.760
167896 95.670 -77.110 50.720
167897 95.680 -77.114 50.684
167898 95.691 -77.118 50.648
167899 95.702 -77.121 50.612
167900 95.710 -77.135 50.576
167901 95.716 -77.158 50.540
167902 95.722 -77.182 50.503
167903 95.727 -77.204 50.461
167904 95.733 -77.228 50.416
167905 95.738 -77.251 50.370
167906 95.746 -77.261 50.326
167907 95.771 -77.266 50.285
167908 95.797 -77.256 50.245
167909 95.822 -77.244 50.204
167910 95.847 -77.234 50.164
167911 95.873 -77.222 50.125
167912 95.898 -77.212 50.084
167913 95.923 -77.201 50.041
167914 95.949 -77.190 49.996
167915 95.941 -77.198 49.946
167916 95.933 -77.207 49.888
167917 95.924 -77.215 49.823
167918 95.915 -77.223 49.753
167919 95.907 -77.232 49.686
167920 95.904 -77.241 49.623
167921 95.916 -77.254 49.565
167922 95.928 -77.266 49.511
167923 95.945 -77.272 49.460
167924 95.961 -77.278 49.413
167925 95.977 -77.284 49.370
167926 95.993 -77.289 49.328
167927 96.010 -77.295 49.290
167928 96.026 -77.301 49.255
167929 96.042 -77.307 49.224
167930 96.058 -77.313 49.194
167931 96.075 -77.320 49.165
167932 96.091 -77.325 49.138
167933 96.108 -77.332 49.114
167934 96.129 -77.339 49.094
167935 96.150 -77.347 49.075
167936 96.171 -77.354 49.056
167937 96.191 -77.354 49.037
167938 96.213 -77.357 49.013
167939 96.237 -77.364 48.987
167940 96.262 -77.371 48.958
167941 96.286 -77.379 48.936
167942 96.311 -77.385 48.915
167943 96.335 -77.392 48.895
167944 96.360 -77.400 48.877
167945 96.384 -77.407 48.861
167946 96.407 -77.415 48.846
167947 96.424 -77.431 48.833
167948 96.440 -77.446 48.822
167949 96.457 -77.461 48.813
167950 96.473 -77.476 48.805
167951 96.489 -77.492 48.797
167952 96.506 -77.507 48.790
167953 96.521 -77.523 48.778
167954 96.531 -77.546 48.763
167955 96.542 -77.568 48.750
167956 96.552 -77.590 48.733
167957 96.562 -77.612 48.717
167958 96.573 -77.634 48.702
167959 96.583 -77.656 48.689
167960 96.599 -77.678 48.673
167961 96.611 -77.680 48.657
167962 96.626 -77.677 48.643
167963 96.647 -77.673 48.627
167964 96.668 -77.669 48.611
167965 96.689 -77.665 48.594
167966 96.710 -77.661 48.577
167967 96.731 -77.657 48.562
167968 96.752 -77.653 48.548
167969 96.773 -77.649 48.534
167970 96.795 -77.645 48.518
167971 96.817 -77.645 48.501
167972 96.847 -77.663 48.484
167973 96.876 -77.681 48.467
167974 96.906 -77.699 48.449
167975 96.926 -77.695 48.430
167976 96.938 -77.676 48.410
167977 96.951 -77.656 48.387
167978 96.964 -77.637 48.364
167979 96.976 -77.617 48.341
167980 96.989 -77.599 48.316
167981 97.002 -77.580 48.293
167982 97.019 -77.574 48.270
167983 97.039 -77.576 48.246
167984 97.060 -77.578 48.222
167985 97.080 -77.580 48.199
167986 97.088 -77.578 48.175
167987 97.095 -77.578 48.151
167988 97.101 -77.583 48.127
167989 97.120 -77.590 48.106
167990 97.148 -77.599 48.085
167991 97.175 -77.608 48.064
167992 97.203 -77.616 48.044
167993 97.230 -77.624 48.023
167994 97.258 -77.633 48.002
167995 97.286 -77.642 47.979
167996 97.313 -77.650 47.955
167997 97.337 -77.660 47.932
167998 97.361 -77.670 47.909
167999 97.384 -77.680 47.887
168000 97.407 -77.689 47.862
168001 97.431 -77.699 47.838
168002 97.454 -77.708 47.814
168003 97.478 -77.718 47.790
168004 97.488 -77.718 47.765
168005 97.505 -77.723 47.741
168006 97.524 -77.730 47.716
168007 97.544 -77.736 47.690
168008 97.564 -77.743 47.666
168009 97.583 -77.749 47.640
168010 97.603 -77.756 47.612
168011 97.623 -77.763 47.584
168012 97.642 -77.770 47.556
168013 97.662 -77.777 47.529
168014 97.701 -77.789 47.503
168015 97.739 -77.800 47.476
168016 97.778 -77.812 47.449
168017 97.816 -77.824 47.423
168018 97.844 -77.813 47.398
168019 97.860 -77.782 47.373
168020 97.868 -77.753 47.349
168021 97.830 -77.737 47.325
168022 97.793 -77.721 47.300
168023 97.784 -77.733 47.274
168024 97.803 -77.754 47.247
168025 97.835 -77.767 47.220
168026 97.867 -77.779 47.192
168027 97.877 -77.780 47.163
168028 97.938 -77.793 47.135
168029 97.925 -77.777 47.107
168030 97.900 -77.774 47.081
168031 97.967 -77.788 47.055
168032 97.986 -77.790 47.029
168033 98.005 -77.790 47.004
168034 98.024 -77.792 46.980
168035 98.043 -77.793 46.955
168036 98.063 -77.793 46.930
168037 98.082 -77.795 46.905
168038 98.101 -77.796 46.881
168039 98.120 -77.797 46.856
168040 98.132 -77.799 46.830
168041 98.143 -77.805 46.805
168042 98.164 -77.808 46.781
168043 98.185 -77.813 46.756
168044 98.207 -77.817 46.732
168045 98.228 -77.821 46.708
168046 98.247 -77.827 46.684
168047 98.266 -77.832 46.660
168048 98.285 -77.838 46.635
168049 98.303 -77.844 46.610
168050 98.321 -77.849 46.585
168051 98.343 -77.851 46.561
168052 98.361 -77.842 46.539
168053 98.371 -77.837 46.516
168054 98.387 -77.835 46.493
168055 98.404 -77.833 46.471
168056 98.420 -77.830 46.450
168057 98.436 -77.828 46.428
168058 98.452 -77.826 46.407
168059 98.493 -77.834 46.388
168060 98.528 -77.841 46.368
168061 98.543 -77.840 46.348
168062 98.555 -77.837 46.328
168063 98.558 -77.837 46.309
168064 98.562 -77.838 46.291
168065 98.565 -77.838 46.273
168066 98.568 -77.839 46.257
168067 98.571 -77.839 46.238
168068 98.574 -77.840 46.220
168069 98.588 -77.844 46.202
168070 98.606 -77.852 46.184
168071 98.625 -77.859 46.168
168072 98.644 -77.866 46.152
168073 98.662 -77.874 46.135
168074 98.681 -77.881 46.117
168075 98.700 -77.889 46.098
168076 98.718 -77.896 46.077
168077 98.737 -77.903 46.056
168078 98.756 -77.911 46.037
168079 98.794 -77.918 46.019
168080 98.836 -77.927 46.001
168081 98.878 -77.934 45.984
168082 98.937 -77.944 45.967
168083 99.018 -77.956 45.950
168084 99.097 -77.968 45.933
168085 99.123 -77.971 45.916
168086 99.147 -77.974 45.899
168087 99.172 -77.977 45.883
168088 99.197 -77.980 45.866
168089 99.223 -77.981 45.850
168090 99.254 -77.972 45.834
168091 99.278 -77.979 45.819
168092 99.302 -77.987 45.804
168093 99.326 -77.993 45.792
168094 99.341 -78.001 45.781
168095 99.354 -78.009 45.771
168096 99.367 -78.015 45.762
168097 99.380 -78.023 45.752
168098 99.391 -78.031 45.742
168099 99.400 -78.037 45.733
168100 99.408 -78.045 45.724
168101 99.417 -78.052 45.714
168102 99.425 -78.059 45.705
168103 99.436 -78.067 45.695
168104 99.467 -78.076 45.685
168105 99.498 -78.086 45.675
168106 99.525 -78.097 45.666
168107 99.553 -78.108 45.655
168108 99.580 -78.119 45.644
168109 99.607 -78.130 45.631
168110 99.638 -78.145 45.618
168111 99.668 -78.161 45.605
168112 99.699 -78.177 45.591
168113 99.729 -78.192 45.575
168114 99.760 -78.208 45.559
168115 99.784 -78.222 45.544
168116 99.797 -78.234 45.528
168117 99.811 -78.245 45.514
168118 99.826 -78.257 45.498
168119 99.840 -78.268 45.483
168120 99.854 -78.280 45.467
168121 99.872 -78.293 45.451
168122 99.891 -78.307 45.435
168123 99.910 -78.321 45.419
168124 99.930 -78.335 45.404
168125 99.949 -78.348 45.387
168126 99.972 -78.361 45.370
168127 99.996 -78.373 45.352
168128 100.019 -78.385 45.335
168129 100.043 -78.398 45.319
168130 100.067 -78.410 45.303
168131 100.091 -78.423 45.288
168132 100.115 -78.427 45.273
168133 100.141 -78.431 45.257
168134 100.179 -78.440 45.242
168135 100.216 -78.449 45.226
168136 100.254 -78.459 45.210
168137 100.291 -78.468 45.194
168138 100.328 -78.477 45.179
168139 100.365 -78.486 45.163
168140 100.403 -78.496 45.148
168141 100.440 -78.505 45.133
168142 100.478 -78.514 45.118
168143 100.515 -78.523 45.102
168144 100.552 -78.533 45.088
168145 100.590 -78.542 45.075
168146 100.627 -78.551 45.061
168147 100.649 -78.504 45.047
168148 100.678 -78.522 45.031
168149 100.706 -78.539 45.016
168150 100.735 -78.556 45.001
168151 100.763 -78.573 44.986
168152 100.791 -78.590 44.971
168153 100.819 -78.607 44.955
168154 100.848 -78.624 44.939
168155 100.876 -78.641 44.924
168156 100.905 -78.657 44.908
168157 100.938 -78.656 44.892
168158 100.972 -78.655 44.877
168159 101.001 -78.654 44.861
168160 101.030 -78.651 44.844
168161 101.058 -78.648 44.829
168162 101.086 -78.645 44.813
168163 101.114 -78.636 44.798
168164 101.158 -78.648 44.783
168165 101.200 -78.668 44.768
168166 101.239 -78.697 44.753
168167 101.278 -78.727 44.739
168168 101.315 -78.755 44.724
168169 101.325 -78.778 44.709
168170 101.335 -78.802 44.693
168171 101.344 -78.825 44.678
168172 101.354 -78.849 44.664
168173 101.364 -78.868 44.649
168174 101.375 -78.883 44.634
168175 101.401 -78.912 44.620
168176 101.444 -78.931 44.606
168177 101.485 -78.950 44.593
168178 101.528 -78.970 44.579
168179 101.570 -78.989 44.566
168180 101.612 -79.008 44.552
168181 101.655 -79.028 44.540
168182 101.680 -79.037 44.528
168183 101.697 -79.043 44.516
168184 101.714 -79.049 44.504
168185 101.731 -79.055 44.491
168186 101.747 -79.060 44.479
168187 101.764 -79.066 44.468
168188 101.781 -79.071 44.456
168189 101.798 -79.077 44.444
168415 92.551 -77.234 54.083
168416 92.574 -77.245 54.103
168417 92.597 -77.256 54.122
168418 92.620 -77.267 54.139
168419 92.643 -77.279 54.155
168420 92.666 -77.289 54.169
168421 92.689 -77.301 54.181
168422 92.712 -77.311 54.194
168423 92.735 -77.323 54.206
168424 92.749 -77.335 54.218
168425 92.756 -77.347 54.229
168426 92.763 -77.360 54.240
168427 92.770 -77.373 54.250
168428 92.777 -77.385 54.260
168429 92.784 -77.398 54.273
168430 92.791 -77.410 54.288
168431 92.801 -77.418 54.302
168432 92.821 -77.409 54.309
168433 92.842 -77.401 54.313
168434 92.861 -77.390 54.315
168435 92.881 -77.379 54.315
168436 92.905 -77.369 54.314
168437 92.930 -77.357 54.313
168438 92.956 -77.345 54.314
168439 92.981 -77.334 54.318
168440 93.007 -77.323 54.324
168441 93.032 -77.311 54.330
168442 93.058 -77.299 54.337
168443 93.083 -77.288 54.337
168444 93.108 -77.276 54.338
168445 93.118 -77.262 54.338
168446 93.127 -77.248 54.338
168447 93.137 -77.233 54.336
168448 93.147 -77.219 54.334
168449 93.157 -77.205 54.333
168450 93.167 -77.191 54.332
168451 93.183 -77.181 54.331
168452 93.201 -77.172 54.331
168453 93.219 -77.164 54.331
168454 93.236 -77.157 54.331
168455 93.253 -77.149 54.332
168456 93.266 -77.140 54.331
168457 93.276 -77.129 54.328
168458 93.286 -77.119 54.326
168459 93.296 -77.109 54.321
168460 93.306 -77.099 54.314
168461 93.317 -77.088 54.305
168462 93.346 -77.082 54.297
168463 93.381 -77.078 54.290
168464 93.413 -77.072 54.281
168465 93.442 -77.064 54.269
168466 93.472 -77.057 54.259
168467 93.502 -77.050 54.252
168468 93.532 -77.043 54.248
168469 93.547 -77.039 54.244
168470 93.558 -77.038 54.238
168471 93.568 -77.036 54.229
168472 93.578 -77.034 54.220
168473 93.596 -77.025 54.211
168474 93.617 -77.014 54.201
168475 93.631 -77.009 54.190
168476 93.644 -77.005 54.180
168477 93.656 -77.001 54.170
168478 93.669 -76.998 54.161
168479 93.682 -76.994 54.153
168480 93.695 -76.990 54.145
168481 93.708 -76.987 54.137
168482 93.718 -76.981 54.130
168483 93.722 -76.971 54.123
168484 93.726 -76.960 54.114
168485 93.731 -76.950 54.103
168486 93.735 -76.939 54.094
168487 93.740 -76.928 54.086
168488 93.747 -76.912 54.079
168489 93.767 -76.910 54.073
168490 93.796 -76.918 54.068
168491 93.825 -76.926 54.061
168492 93.854 -76.934 54.054
168493 93.883 -76.942 54.046
168494 93.911 -76.950 54.037
168495 93.933 -76.947 54.025
168496 93.951 -76.937 54.011
168497 93.968 -76.926 53.995
168498 93.985 -76.915 53.974
168499 94.002 -76.904 53.952
168500 94.019 -76.893 53.930
168501 94.037 -76.883 53.910
168502 94.054 -76.872 53.892
168503 94.068 -76.865 53.874
168504 94.080 -76.863 53.856
168505 94.092 -76.860 53.834
168506 94.106 -76.857 53.816
168507 94.120 -76.854 53.801
168508 94.133 -76.851 53.791
168509 94.147 -76.848 53.782
168510 94.161 -76.845 53.773
168511 94.175 -76.842 53.761
168512 94.191 -76.839 53.744
168513 94.204 -76.837 53.722
168514 94.216 -76.834 53.698
168515 94.226 -76.833 53.673
168516 94.231 -76.834 53.650
168517 94.238 -76.836 53.632
168518 94.244 -76.837 53.616
168519 94.250 -76.838 53.602
168520 94.257 -76.839 53.586
168521 94.286 -76.830 53.571
168522 94.315 -76.821 53.558
168523 94.343 -76.812 53.547
168524 94.360 -76.809 53.540
168525 94.378 -76.806 53.530
168526 94.395 -76.804 53.517
168527 94.413 -76.802 53.502
168528 94.430 -76.799 53.486
168529 94.449 -76.796 53.469
168530 94.467 -76.793 53.453
168531 94.485 -76.790 53.437
168532 94.499 -76.788 53.417
168533 94.514 -76.786 53.395
168534 94.529 -76.784 53.371
168535 94.544 -76.781 53.345
168536 94.546 -76.787 53.319
168537 94.547 -76.794 53.292
168538 94.547 -76.801 53.262
168539 94.548 -76.808 53.231
168540 94.549 -76.813 53.200
168541 94.594 -76.798 53.168
168542 94.645 -76.778 53.136
168543 94.697 -76.759 53.107
168544 94.749 -76.739 53.077
168545 94.797 -76.721 53.046
168546 94.834 -76.712 53.008
168547 94.853 -76.705 52.969
168548 94.864 -76.699 52.929
168549 94.877 -76.693 52.888
168550 94.888 -76.687 52.846
168551 94.900 -76.680 52.803
168552 94.912 -76.674 52.759
168553 94.924 -76.668 52.711
168554 94.935 -76.661 52.666
168555 94.948 -76.655 52.623
168556 94.959 -76.649 52.583
168557 94.971 -76.643 52.544
168558 94.983 -76.637 52.507
168559 94.995 -76.631 52.470
168560 95.006 -76.627 52.436
168561 95.017 -76.625 52.404
168562 95.031 -76.626 52.375
168563 95.058 -76.634 52.348
168564 95.095 -76.646 52.321
168565 95.129 -76.663 52.291
168566 95.162 -76.682 52.259
168567 95.196 -76.701 52.223
168568 95.229 -76.720 52.182
168569 95.262 -76.739 52.136
168570 95.296 -76.757 52.088
168571 95.300 -76.777 52.035
168572 95.299 -76.799 51.978
168573 95.297 -76.819 51.920
168574 95.296 -76.840 51.862
168575 95.294 -76.860 51.806
168576 95.293 -76.881 51.752
168577 95.302 -76.895 51.697
168578 95.318 -76.903 51.640
168579 95.336 -76.909 51.583
168580 95.353 -76.919 51.527
168581 95.370 -76.930 51.474
168582 95.387 -76.940 51.422
168583 95.403 -76.950 51.371
168584 95.420 -76.961 51.322
168585 95.438 -76.971 51.274
168586 95.454 -76.981 51.225
168587 95.471 -76.992 51.177
168588 95.488 -77.003 51.129
168589 95.505 -77.012 51.083
168590 95.527 -77.022 51.036
168591 95.561 -77.002 50.990
168592 95.593 -76.990 50.944
168593 95.603 -77.046 50.900
168594 95.622 -77.072 50.858
168595 95.641 -77.100 50.816
168596 95.651 -77.099 50.774
168597 95.662 -77.103 50.734
168598 95.672 -77.106 50.696
168599 95.683 -77.110 50.660
168600 95.694 -77.114 50.625
168601 95.702 -77.128 50.590
168602 95.707 -77.150 50.553
168603 95.713 -77.174 50.514
168604 95.719 -77.197 50.471
168605 95.725 -77.220 50.428
168606 95.730 -77.244 50.384
168607 95.739 -77.260 50.341
168608 95.756 -77.260 50.300
168609 95.781 -77.249 50.260
168610 95.807 -77.238 50.220
168611 95.832 -77.228 50.180
168612 95.857 -77.216 50.142
168613 95.883 -77.206 50.101
168614 95.908 -77.194 50.058
168615 95.933 -77.184 50.012
168616 95.940 -77.183 49.963
168617 95.932 -77.191 49.905
168618 95.923 -77.200 49.841
168619 95.915 -77.208 49.770
168620 95.906 -77.217 49.701
168621 95.913 -77.229 49.637
168622 95.924 -77.241 49.578
168623 95.937 -77.253 49.523
168624 95.953 -77.259 49.471
168625 95.969 -77.265 49.423
168626 95.985 -77.271 49.378
168627 96.002 -77.277 49.336
168628 96.018 -77.282 49.299
168629 96.034 -77.288 49.264
168630 96.050 -77.294 49.231
168631 96.066 -77.301 49.202
168632 96.083 -77.307 49.173
168633 96.099 -77.312 49.146
168634 96.115 -77.318 49.123
168635 96.132 -77.324 49.108
168636 96.148 -77.330 49.092
168637 96.169 -77.335 49.073
168638 96.193 -77.342 49.053
168639 96.218 -77.349 49.027
168640 96.242 -77.357 48.999
168641 96.267 -77.363 48.969
168642 96.291 -77.371 48.946
168643 96.316 -77.378 48.924
168644 96.340 -77.385 48.903
168645 96.365 -77.392 48.887
168646 96.389 -77.399 48.869
168647 96.409 -77.410 48.854
168648 96.426 -77.426 48.840
168649 96.442 -77.442 48.828
168650 96.459 -77.457 48.818
168651 96.475 -77.472 48.808
168652 96.491 -77.487 48.799
168653 96.508 -77.502 48.791
168654 96.524 -77.517 48.781
168655 96.538 -77.536 48.768
168656 96.548 -77.558 48.753
168657 96.558 -77.580 48.739
168658 96.569 -77.602 48.725
168659 96.580 -77.624 48.711
168660 96.590 -77.647 48.698
168661 96.600 -77.669 48.684
168662 96.620 -77.677 48.669
168663 96.641 -77.673 48.654
168664 96.662 -77.669 48.639
168665 96.683 -77.665 48.623
168666 96.704 -77.661 48.606
168667 96.725 -77.657 48.589
168668 96.746 -77.653 48.574
168669 96.767 -77.649 48.560
168670 96.789 -77.645 48.547
168671 96.810 -77.641 48.532
168672 96.835 -77.649 48.514
168673 96.865 -77.667 48.497
168674 96.894 -77.685 48.481
168675 96.924 -77.703 48.463
168676 96.948 -77.709 48.445
168677 96.961 -77.690 48.425
168678 96.974 -77.671 48.403
168679 96.986 -77.652 48.379
168680 96.999 -77.633 48.357
168681 97.012 -77.613 48.334
168682 97.025 -77.594 48.312
168683 97.037 -77.574 48.288
168684 97.051 -77.558 48.265
168685 97.071 -77.561 48.241
168686 97.089 -77.561 48.217
168687 97.097 -77.559 48.192
168688 97.104 -77.560 48.168
168689 97.115 -77.565 48.145
168690 97.143 -77.574 48.122
168691 97.170 -77.583 48.101
168692 97.197 -77.591 48.080
168693 97.225 -77.600 48.059
168694 97.251 -77.608 48.038
168695 97.275 -77.618 48.017
168696 97.298 -77.628 47.994
168697 97.322 -77.638 47.971
168698 97.345 -77.648 47.948
168699 97.369 -77.658 47.925
168700 97.392 -77.667 47.901
168701 97.415 -77.677 47.877
168702 97.439 -77.687 47.853
168703 97.462 -77.697 47.828
168704 97.486 -77.707 47.803
168705 97.504 -77.712 47.779
168706 97.508 -77.707 47.755
168707 97.526 -77.713 47.731
168708 97.546 -77.720 47.705
168709 97.566 -77.727 47.679
168710 97.585 -77.733 47.653
168711 97.605 -77.740 47.624
168712 97.625 -77.747 47.596
168713 97.644 -77.754 47.569
168714 97.679 -77.765 47.541
168715 97.717 -77.776 47.514
168716 97.756 -77.787 47.487
168717 97.794 -77.799 47.459
168718 97.833 -77.811 47.432
168719 97.867 -77.815 47.406
168720 97.882 -77.783 47.381
168721 97.845 -77.768 47.356
168722 97.808 -77.751 47.332
168723 97.771 -77.734 47.307
168724 97.754 -77.739 47.281
168725 97.783 -77.754 47.254
168726 97.815 -77.767 47.227
168727 97.847 -77.779 47.199
168728 97.881 -77.787 47.170
168729 97.910 -77.803 47.141
168730 97.917 -77.793 47.112
168731 97.924 -77.760 47.085
168732 97.963 -77.765 47.059
168733 98.018 -77.776 47.034
168734 98.037 -77.777 47.008
168735 98.056 -77.778 46.984
168736 98.075 -77.780 46.960
168737 98.094 -77.780 46.934
168738 98.113 -77.782 46.909
168739 98.133 -77.783 46.884
168740 98.152 -77.783 46.858
168741 98.155 -77.788 46.833
168742 98.160 -77.793 46.809
168743 98.174 -77.798 46.785
168744 98.195 -77.802 46.761
168745 98.217 -77.806 46.738
168746 98.238 -77.811 46.715
168747 98.260 -77.815 46.692
168748 98.281 -77.819 46.668
168749 98.302 -77.824 46.643
168750 98.324 -77.827 46.617
168751 98.346 -77.832 46.592
168752 98.367 -77.836 46.568
168753 98.387 -77.838 46.545
168754 98.404 -77.836 46.522
168755 98.420 -77.834 46.498
168756 98.436 -77.831 46.477
168757 98.452 -77.829 46.455
168758 98.468 -77.826 46.433
168759 98.485 -77.824 46.412
168760 98.528 -77.834 46.392
168761 98.547 -77.834 46.373
168762 98.561 -77.834 46.354
168763 98.570 -77.831 46.333
168764 98.573 -77.831 46.313
168765 98.577 -77.832 46.295
168766 98.580 -77.832 46.276
168767 98.583 -77.833 46.258
168768 98.586 -77.833 46.240
168769 98.589 -77.834 46.221
168770 98.593 -77.834 46.203
168771 98.608 -77.840 46.185
168772 98.627 -77.847 46.168
168773 98.646 -77.854 46.151
168774 98.664 -77.862 46.134
168775 98.683 -77.869 46.117
168776 98.702 -77.877 46.098
168777 98.720 -77.884 46.079
168778 98.739 -77.891 46.058
168779 98.764 -77.899 46.039
168780 98.806 -77.907 46.020
168781 98.848 -77.915 46.002
168782 98.890 -77.923 45.985
168783 98.950 -77.933 45.969
168784 99.031 -77.944 45.953
168785 99.087 -77.952 45.937
168786 99.112 -77.955 45.920
168787 99.136 -77.958 45.903
168788 99.161 -77.961 45.887
168789 99.186 -77.964 45.871
168790 99.211 -77.967 45.855
168791 99.235 -77.971 45.839
168792 99.259 -77.979 45.824
168793 99.283 -77.986 45.809
168794 99.307 -77.993 45.797
168795 99.331 -78.000 45.786
168796 99.344 -78.008 45.776
168797 99.357 -78.015 45.767
168798 99.370 -78.022 45.758
168799 99.383 -78.030 45.748
168800 99.396 -78.037 45.739
168801 99.406 -78.044 45.730
168802 99.414 -78.052 45.721
168803 99.423 -78.059 45.712
168804 99.438 -78.067 45.703
168805 99.469 -78.076 45.693
168806 99.496 -78.087 45.682
168807 99.523 -78.098 45.671
168808 99.551 -78.109 45.661
168809 99.577 -78.120 45.650
168810 99.606 -78.132 45.637
168811 99.637 -78.148 45.625
168812 99.667 -78.163 45.611
168813 99.698 -78.179 45.596
168814 99.728 -78.195 45.580
168815 99.759 -78.210 45.565
168816 99.790 -78.226 45.549
168817 99.807 -78.238 45.533
168818 99.822 -78.250 45.517
168819 99.835 -78.262 45.503
168820 99.849 -78.273 45.487
168821 99.865 -78.286 45.470
168822 99.885 -78.300 45.454
168823 99.904 -78.313 45.438
168824 99.923 -78.327 45.423
168825 99.943 -78.341 45.407
168826 99.962 -78.355 45.392
168827 99.982 -78.369 45.374
168828 100.006 -78.381 45.356
168829 100.030 -78.393 45.339
168830 100.054 -78.405 45.323
168831 100.078 -78.417 45.308
168832 100.102 -78.422 45.293
168833 100.126 -78.425 45.279
168834 100.150 -78.428 45.263
168835 100.179 -78.433 45.247
168836 100.216 -78.442 45.232
168837 100.254 -78.451 45.215
168838 100.291 -78.461 45.200
168839 100.328 -78.470 45.184
168840 100.366 -78.479 45.169
168841 100.403 -78.489 45.154
168842 100.440 -78.498 45.139
168843 100.478 -78.507 45.124
168844 100.515 -78.517 45.108
168845 100.552 -78.526 45.094
168846 100.590 -78.535 45.080
168847 100.627 -78.545 45.066
168848 100.681 -78.552 45.052
168849 100.682 -78.517 45.037
168850 100.708 -78.535 45.022
168851 100.736 -78.551 45.006
168852 100.764 -78.569 44.991
168853 100.792 -78.586 44.976
168854 100.821 -78.603 44.960
168855 100.849 -78.620 44.945
168856 100.878 -78.637 44.929
168857 100.907 -78.649 44.913
168858 100.941 -78.648 44.897
168859 100.974 -78.648 44.881
168860 101.006 -78.647 44.865
168861 101.035 -78.644 44.850
168862 101.063 -78.642 44.835
168863 101.091 -78.639 44.820
168864 101.119 -78.625 44.804
168865 101.164 -78.639 44.789
168866 101.209 -78.652 44.775
168867 101.249 -78.680 44.760
168868 101.287 -78.709 44.745
168869 101.316 -78.736 44.730
168870 101.326 -78.759 44.715
168871 101.336 -78.783 44.699
168872 101.346 -78.807 44.683
168873 101.356 -78.830 44.669
168874 101.366 -78.854 44.654
168875 101.376 -78.877 44.640
168876 101.401 -78.898 44.626
168877 101.443 -78.918 44.612
168878 101.485 -78.937 44.598
168879 101.528 -78.955 44.584
168880 101.570 -78.975 44.570
168881 101.612 -78.994 44.558
168882 101.654 -79.014 44.545
168883 101.681 -79.024 44.533
168884 101.698 -79.031 44.521
168885 101.714 -79.036 44.509
168886 101.732 -79.041 44.496
168887 101.748 -79.047 44.484
168888 101.765 -79.053 44.472
168889 101.781 -79.059 44.460
168890 101.799 -79.064 44.449
168891 101.815 -79.069 44.437
169118 92.584 -77.227 54.106
169119 92.607 -77.238 54.124
169120 92.630 -77.249 54.141
169121 92.653 -77.260 54.156
169122 92.676 -77.272 54.168
169123 92.700 -77.282 54.182
169124 92.722 -77.294 54.195
169125 92.742 -77.305 54.208
169126 92.749 -77.317 54.219
169127 92.756 -77.330 54.229
169128 92.763 -77.342 54.240
169129 92.771 -77.355 54.250
169130 92.778 -77.368 54.262
169131 92.785 -77.380 54.275
169132 92.792 -77.393 54.286
169133 92.811 -77.385 54.292
169134 92.832 -77.376 54.296
169135 92.852 -77.368 54.298
169136 92.872 -77.359 54.299
169137 92.892 -77.350 54.299
169138 92.912 -77.339 54.299
169139 92.931 -77.329 54.301
169140 92.950 -77.319 54.305
169141 92.972 -77.308 54.309
169142 92.997 -77.296 54.313
169143 93.022 -77.285 54.318
169144 93.048 -77.273 54.319
169145 93.073 -77.262 54.319
169146 93.099 -77.251 54.319
169147 93.121 -77.238 54.321
169148 93.131 -77.224 54.323
169149 93.141 -77.210 54.324
169150 93.151 -77.195 54.326
169151 93.161 -77.182 54.327
169152 93.178 -77.172 54.328
169153 93.195 -77.163 54.330
169154 93.213 -77.154 54.330
169155 93.231 -77.147 54.330
169156 93.242 -77.137 54.330
169157 93.253 -77.127 54.328
169158 93.263 -77.116 54.322
169159 93.273 -77.106 54.317
169160 93.283 -77.096 54.312
169161 93.293 -77.086 54.305
169162 93.303 -77.075 54.297
169163 93.332 -77.069 54.290
169164 93.367 -77.065 54.284
169165 93.402 -77.059 54.276
169166 93.431 -77.052 54.265
169167 93.461 -77.045 54.256
169168 93.491 -77.038 54.249
169169 93.520 -77.031 54.244
169170 93.540 -77.027 54.240
169171 93.551 -77.025 54.234
169172 93.561 -77.024 54.226
169173 93.571 -77.022 54.217
169174 93.590 -77.012 54.209
169175 93.608 -77.003 54.201
169176 93.621 -77.000 54.193
169177 93.634 -76.996 54.185
169178 93.646 -76.992 54.177
169179 93.659 -76.989 54.169
169180 93.672 -76.985 54.162
169181 93.685 -76.981 54.155
169182 93.698 -76.978 54.147
169183 93.710 -76.975 54.139
169184 93.717 -76.965 54.130
169185 93.721 -76.955 54.121
169186 93.726 -76.944 54.112
169187 93.731 -76.932 54.103
169188 93.739 -76.915 54.094
169189 93.748 -76.901 54.085
169190 93.766 -76.898 54.077
169191 93.790 -76.900 54.072
169192 93.819 -76.909 54.065
169193 93.847 -76.916 54.058
169194 93.876 -76.925 54.052
169195 93.905 -76.933 54.046
169196 93.928 -76.932 54.036
169197 93.945 -76.922 54.024
169198 93.963 -76.911 54.012
169199 93.980 -76.900 53.993
169200 93.997 -76.890 53.973
169201 94.014 -76.879 53.951
169202 94.032 -76.868 53.932
169203 94.049 -76.857 53.915
169204 94.063 -76.852 53.897
169205 94.074 -76.849 53.880
169206 94.088 -76.846 53.858
169207 94.102 -76.843 53.841
169208 94.116 -76.840 53.827
169209 94.129 -76.837 53.816
169210 94.143 -76.834 53.809
169211 94.157 -76.831 53.801
169212 94.171 -76.828 53.791
169213 94.186 -76.824 53.775
169214 94.197 -76.822 53.755
169215 94.203 -76.824 53.732
169216 94.208 -76.825 53.707
169217 94.214 -76.827 53.682
169218 94.220 -76.828 53.662
169219 94.226 -76.830 53.644
169220 94.232 -76.831 53.629
169221 94.248 -76.828 53.614
169222 94.277 -76.818 53.599
169223 94.307 -76.809 53.584
169224 94.335 -76.799 53.573
169225 94.357 -76.795 53.564
169226 94.374 -76.793 53.552
169227 94.391 -76.790 53.538
169228 94.409 -76.787 53.521
169229 94.427 -76.784 53.503
169230 94.445 -76.781 53.485
169231 94.463 -76.779 53.469
169232 94.481 -76.776 53.454
169233 94.498 -76.773 53.435
169234 94.513 -76.771 53.413
169235 94.528 -76.769 53.390
169236 94.538 -76.770 53.365
169237 94.539 -76.776 53.339
169238 94.539 -76.783 53.311
169239 94.540 -76.790 53.283
169240 94.540 -76.795 53.252
169241 94.546 -76.797 53.221
169242 94.596 -76.781 53.188
169243 94.647 -76.762 53.156
169244 94.698 -76.742 53.127
169245 94.750 -76.722 53.097
169246 94.793 -76.708 53.065
169247 94.830 -76.699 53.026
169248 94.849 -76.691 52.986
169249 94.861 -76.685 52.945
169250 94.873 -76.679 52.902
169251 94.885 -76.673 52.860
169252 94.896 -76.667 52.817
169253 94.908 -76.661 52.772
169254 94.920 -76.655 52.725
169255 94.932 -76.649 52.683
169256 94.944 -76.643 52.643
169257 94.956 -76.636 52.606
169258 94.967 -76.630 52.565
169259 94.979 -76.624 52.524
169260 94.991 -76.618 52.483
169261 95.003 -76.612 52.443
169262 95.015 -76.606 52.407
169263 95.028 -76.602 52.374
169264 95.060 -76.620 52.344
169265 95.094 -76.639 52.314
169266 95.127 -76.657 52.283
169267 95.161 -76.676 52.252
169268 95.194 -76.695 52.217
169269 95.228 -76.714 52.178
169270 95.261 -76.732 52.134
169271 95.276 -76.752 52.087
169272 95.275 -76.773 52.034
169273 95.273 -76.793 51.979
169274 95.272 -76.814 51.923
169275 95.271 -76.835 51.868
169276 95.277 -76.850 51.815
169277 95.287 -76.864 51.760
169278 95.298 -76.877 51.704
169279 95.312 -76.890 51.646
169280 95.329 -76.895 51.590
169281 95.348 -76.897 51.536
169282 95.367 -76.899 51.484
169283 95.384 -76.909 51.435
169284 95.400 -76.919 51.384
169285 95.417 -76.930 51.335
169286 95.434 -76.940 51.287
169287 95.451 -76.951 51.239
169288 95.468 -76.961 51.191
169289 95.485 -76.972 51.143
169290 95.502 -76.982 51.097
169291 95.524 -76.991 51.050
169292 95.555 -76.956 51.004
169293 95.574 -76.942 50.959
169294 95.582 -77.000 50.915
169295 95.601 -77.028 50.873
169296 95.620 -77.055 50.831
169297 95.639 -77.082 50.790
169298 95.654 -77.095 50.750
169299 95.664 -77.099 50.711
169300 95.675 -77.103 50.673
169301 95.686 -77.106 50.637
169302 95.694 -77.120 50.600
169303 95.699 -77.144 50.563
169304 95.705 -77.167 50.524
169305 95.710 -77.190 50.481
169306 95.718 -77.210 50.439
169307 95.731 -77.219 50.397
169308 95.744 -77.229 50.356
169309 95.757 -77.238 50.315
169310 95.772 -77.237 50.275
169311 95.791 -77.232 50.235
169312 95.816 -77.221 50.195
169313 95.841 -77.210 50.158
169314 95.867 -77.199 50.117
169315 95.892 -77.188 50.074
169316 95.917 -77.178 50.028
169317 95.940 -77.169 49.978
169318 95.931 -77.177 49.921
169319 95.922 -77.185 49.858
169320 95.914 -77.194 49.787
169321 95.910 -77.203 49.716
169322 95.921 -77.216 49.651
169323 95.933 -77.229 49.590
169324 95.945 -77.240 49.534
169325 95.961 -77.246 49.483
169326 95.978 -77.252 49.434
169327 95.994 -77.258 49.387
169328 96.010 -77.264 49.345
169329 96.026 -77.269 49.307
169330 96.043 -77.276 49.272
169331 96.059 -77.282 49.239
169332 96.075 -77.288 49.210
169333 96.092 -77.294 49.181
169334 96.108 -77.300 49.154
169335 96.124 -77.307 49.130
169336 96.142 -77.317 49.114
169337 96.159 -77.328 49.098
169338 96.178 -77.333 49.081
169339 96.198 -77.335 49.059
169340 96.223 -77.341 49.033
169341 96.247 -77.349 49.005
169342 96.271 -77.356 48.976
169343 96.296 -77.363 48.953
169344 96.321 -77.370 48.932
169345 96.345 -77.377 48.912
169346 96.370 -77.385 48.895
169347 96.394 -77.392 48.878
169348 96.411 -77.406 48.863
169349 96.428 -77.421 48.848
169350 96.444 -77.436 48.836
169351 96.461 -77.452 48.825
169352 96.477 -77.467 48.814
169353 96.493 -77.483 48.804
169354 96.510 -77.498 48.793
169355 96.526 -77.513 48.783
169356 96.543 -77.528 48.772
169357 96.555 -77.548 48.759
169358 96.565 -77.570 48.747
169359 96.576 -77.593 48.735
169360 96.586 -77.615 48.723
169361 96.596 -77.637 48.710
169362 96.610 -77.655 48.696
169363 96.633 -77.665 48.681
169364 96.656 -77.670 48.666
169365 96.677 -77.665 48.651
169366 96.698 -77.661 48.635
169367 96.719 -77.658 48.618
169368 96.740 -77.653 48.601
169369 96.761 -77.649 48.585
169370 96.783 -77.645 48.571
169371 96.804 -77.642 48.557
169372 96.825 -77.637 48.542
169373 96.854 -77.653 48.526
169374 96.883 -77.671 48.510
169375 96.912 -77.689 48.494
169376 96.942 -77.708 48.477
169377 96.971 -77.724 48.459
169378 96.984 -77.705 48.439
169379 96.997 -77.686 48.418
169380 97.009 -77.666 48.394
169381 97.022 -77.647 48.373
169382 97.035 -77.627 48.351
169383 97.047 -77.608 48.330
169384 97.060 -77.589 48.306
169385 97.073 -77.570 48.282
169386 97.085 -77.551 48.258
169387 97.098 -77.543 48.234
169388 97.106 -77.541 48.209
169389 97.113 -77.541 48.184
169390 97.137 -77.549 48.161
169391 97.165 -77.558 48.139
169392 97.189 -77.567 48.117
169393 97.213 -77.577 48.095
169394 97.236 -77.587 48.074
169395 97.259 -77.597 48.052
169396 97.283 -77.606 48.031
169397 97.306 -77.616 48.009
169398 97.329 -77.626 47.986
169399 97.353 -77.636 47.963
169400 97.376 -77.645 47.939
169401 97.399 -77.655 47.916
169402 97.423 -77.665 47.891
169403 97.446 -77.675 47.867
169404 97.470 -77.685 47.842
169405 97.493 -77.695 47.817
169406 97.516 -77.705 47.793
169407 97.523 -77.702 47.769
169408 97.528 -77.697 47.744
169409 97.548 -77.704 47.719
169410 97.568 -77.710 47.692
169411 97.587 -77.717 47.665
169412 97.607 -77.724 47.636
169413 97.626 -77.730 47.608
169414 97.657 -77.740 47.581
169415 97.696 -77.752 47.553
169416 97.734 -77.763 47.525
169417 97.773 -77.775 47.497
169418 97.811 -77.787 47.469
169419 97.850 -77.799 47.441
169420 97.888 -77.810 47.414
169421 97.860 -77.797 47.388
169422 97.822 -77.781 47.363
169423 97.785 -77.765 47.339
169424 97.748 -77.748 47.314
169425 97.731 -77.741 47.288
169426 97.772 -77.740 47.262
169427 97.812 -77.736 47.235
169428 97.853 -77.733 47.206
169429 97.894 -77.731 47.177
169430 97.926 -77.743 47.147
169431 97.952 -77.772 47.118
169432 97.983 -77.793 47.091
169433 97.984 -77.750 47.065
169434 98.025 -77.756 47.039
169435 98.069 -77.764 47.015
169436 98.088 -77.765 46.990
169437 98.107 -77.766 46.965
169438 98.126 -77.768 46.938
169439 98.145 -77.768 46.912
169440 98.164 -77.770 46.886
169441 98.176 -77.772 46.860
169442 98.177 -77.777 46.835
169443 98.180 -77.782 46.811
169444 98.190 -77.787 46.787
169445 98.205 -77.792 46.763
169446 98.227 -77.796 46.740
169447 98.248 -77.800 46.718
169448 98.270 -77.804 46.695
169449 98.291 -77.808 46.672
169450 98.312 -77.812 46.648
169451 98.334 -77.817 46.623
169452 98.355 -77.821 46.598
169453 98.378 -77.824 46.574
169454 98.405 -77.825 46.552
169455 98.427 -77.827 46.529
169456 98.450 -77.830 46.506
169457 98.468 -77.830 46.483
169458 98.485 -77.827 46.461
169459 98.501 -77.824 46.439
169460 98.519 -77.823 46.418
169461 98.550 -77.828 46.398
169462 98.565 -77.827 46.378
169463 98.579 -77.826 46.358
169464 98.585 -77.825 46.338
169465 98.588 -77.825 46.317
169466 98.591 -77.826 46.298
169467 98.595 -77.826 46.279
169468 98.598 -77.827 46.260
169469 98.601 -77.827 46.241
169470 98.604 -77.827 46.223
169471 98.608 -77.827 46.203
169472 98.611 -77.827 46.185
169473 98.629 -77.835 46.167
169474 98.648 -77.843 46.150
169475 98.666 -77.849 46.133
169476 98.685 -77.857 46.116
169477 98.704 -77.865 46.098
169478 98.722 -77.871 46.080
169479 98.741 -77.879 46.060
169480 98.776 -77.887 46.041
169481 98.818 -77.895 46.022
169482 98.860 -77.903 46.004
169483 98.902 -77.911 45.987
169484 98.962 -77.921 45.971
169485 99.044 -77.933 45.955
169486 99.076 -77.937 45.939
169487 99.101 -77.940 45.923
169488 99.126 -77.943 45.907
169489 99.150 -77.946 45.891
169490 99.175 -77.949 45.874
169491 99.199 -77.957 45.858
169492 99.223 -77.966 45.843
169493 99.244 -77.975 45.828
169494 99.265 -77.985 45.814
169495 99.288 -77.993 45.803
169496 99.311 -78.000 45.792
169497 99.333 -78.007 45.782
169498 99.346 -78.015 45.773
169499 99.359 -78.022 45.763
169500 99.372 -78.029 45.754
169501 99.385 -78.037 45.744
169502 99.399 -78.044 45.736
169503 99.412 -78.051 45.728
169504 99.421 -78.059 45.719
169505 99.440 -78.067 45.710
169506 99.467 -78.078 45.699
169507 99.494 -78.088 45.688
169508 99.521 -78.100 45.676
169509 99.549 -78.110 45.666
169510 99.575 -78.121 45.655
169511 99.605 -78.135 45.643
169512 99.636 -78.150 45.630
169513 99.666 -78.166 45.616
169514 99.697 -78.182 45.601
169515 99.727 -78.197 45.585
169516 99.758 -78.213 45.571
169517 99.789 -78.229 45.555
169518 99.817 -78.244 45.539
169519 99.831 -78.256 45.523
169520 99.845 -78.267 45.507
169521 99.859 -78.279 45.491
169522 99.878 -78.293 45.475
169523 99.898 -78.307 45.458
169524 99.917 -78.320 45.442
169525 99.936 -78.334 45.426
169526 99.956 -78.347 45.411
169527 99.975 -78.362 45.396
169528 99.994 -78.376 45.378
169529 100.017 -78.388 45.361
169530 100.041 -78.401 45.343
169531 100.065 -78.413 45.329
169532 100.089 -78.417 45.314
169533 100.113 -78.420 45.300
169534 100.137 -78.423 45.285
169535 100.161 -78.426 45.269
169536 100.186 -78.429 45.253
169537 100.216 -78.435 45.237
169538 100.254 -78.445 45.221
169539 100.291 -78.454 45.205
169540 100.329 -78.463 45.190
169541 100.366 -78.473 45.175
169542 100.403 -78.482 45.161
169543 100.440 -78.491 45.146
169544 100.478 -78.501 45.130
169545 100.515 -78.510 45.113
169546 100.553 -78.519 45.099
169547 100.590 -78.528 45.085
169548 100.632 -78.537 45.072
169549 100.690 -78.545 45.057
169550 100.748 -78.552 45.042
169551 100.732 -78.530 45.027
169552 100.738 -78.548 45.012
169553 100.766 -78.565 44.997
169554 100.794 -78.582 44.981
169555 100.823 -78.599 44.966
169556 100.851 -78.616 44.950
169557 100.879 -78.633 44.934
169558 100.910 -78.642 44.918
169559 100.944 -78.641 44.901
169560 100.977 -78.640 44.886
169561 101.010 -78.639 44.870
169562 101.040 -78.637 44.856
169563 101.068 -78.635 44.842
169564 101.096 -78.632 44.826
169565 101.124 -78.615 44.811
169566 101.170 -78.630 44.796
169567 101.215 -78.643 44.781
169568 101.258 -78.663 44.766
169569 101.297 -78.692 44.751
169570 101.318 -78.717 44.736
169571 101.327 -78.741 44.719
169572 101.337 -78.764 44.704
169573 101.347 -78.788 44.689
169574 101.357 -78.811 44.675
169575 101.368 -78.836 44.660
169576 101.378 -78.862 44.645
169577 101.403 -78.884 44.631
169578 101.443 -78.904 44.616
169579 101.485 -78.923 44.603
169580 101.527 -78.942 44.589
169581 101.569 -78.962 44.575
169582 101.612 -78.981 44.562
169583 101.654 -78.999 44.550
169584 101.682 -79.012 44.537
169585 101.699 -79.017 44.525
169586 101.715 -79.023 44.513
169587 101.732 -79.028 44.500
169588 101.749 -79.034 44.488
169589 101.766 -79.040 44.476
169590 101.782 -79.045 44.464
169591 101.799 -79.051 44.452
169592 101.816 -79.056 44.441
169593 101.833 -79.062 44.430
169820 92.594 -77.209 54.108
169821 92.618 -77.220 54.126
169822 92.641 -77.231 54.141
169823 92.664 -77.242 54.155
169824 92.687 -77.254 54.169
169825 92.710 -77.264 54.183
169826 92.733 -77.276 54.196
169827 92.742 -77.288 54.208
169828 92.749 -77.301 54.219
169829 92.757 -77.313 54.228
169830 92.764 -77.326 54.239
169831 92.771 -77.338 54.250
169832 92.778 -77.351 54.260
169833 92.785 -77.363 54.269
169834 92.801 -77.362 54.275
169835 92.821 -77.353 54.279
169836 92.841 -77.345 54.281
169837 92.862 -77.335 54.282
169838 92.882 -77.327 54.284
169839 92.902 -77.318 54.286
169840 92.923 -77.309 54.289
169841 92.942 -77.299 54.291
169842 92.961 -77.288 54.295
169843 92.981 -77.278 54.298
169844 93.001 -77.268 54.302
169845 93.020 -77.257 54.303
169846 93.040 -77.247 54.304
169847 93.064 -77.236 54.306
169848 93.089 -77.224 54.310
169849 93.115 -77.213 54.314
169850 93.135 -77.201 54.319
169851 93.145 -77.186 54.323
169852 93.155 -77.172 54.325
169853 93.171 -77.162 54.328
169854 93.190 -77.153 54.329
169855 93.208 -77.144 54.328
169856 93.219 -77.135 54.327
169857 93.229 -77.124 54.325
169858 93.239 -77.114 54.322
169859 93.250 -77.103 54.316
169860 93.260 -77.093 54.308
169861 93.270 -77.083 54.301
169862 93.280 -77.072 54.294
169863 93.290 -77.063 54.288
169864 93.319 -77.056 54.282
169865 93.354 -77.051 54.276
169866 93.389 -77.047 54.269
169867 93.420 -77.041 54.261
169868 93.450 -77.033 54.253
169869 93.480 -77.026 54.246
169870 93.505 -77.020 54.242
169871 93.525 -77.016 54.238
169872 93.544 -77.013 54.233
169873 93.554 -77.011 54.225
169874 93.564 -77.009 54.218
169875 93.585 -76.998 54.212
169876 93.598 -76.994 54.206
169877 93.611 -76.990 54.200
169878 93.624 -76.987 54.194
169879 93.636 -76.983 54.188
169880 93.649 -76.980 54.180
169881 93.662 -76.976 54.175
169882 93.675 -76.972 54.168
169883 93.688 -76.969 54.160
169884 93.700 -76.965 54.152
169885 93.712 -76.960 54.143
169886 93.716 -76.950 54.135
169887 93.722 -76.936 54.125
169888 93.729 -76.919 54.116
169889 93.737 -76.903 54.106
169890 93.748 -76.891 54.096
169891 93.767 -76.888 54.087
169892 93.785 -76.884 54.079
169893 93.813 -76.891 54.070
169894 93.841 -76.899 54.063
169895 93.870 -76.907 54.056
169896 93.899 -76.915 54.051
169897 93.924 -76.917 54.043
169898 93.940 -76.906 54.034
169899 93.958 -76.896 54.026
169900 93.975 -76.885 54.011
169901 93.992 -76.875 53.994
169902 94.009 -76.865 53.975
169903 94.026 -76.853 53.956
169904 94.043 -76.843 53.938
169905 94.057 -76.837 53.920
169906 94.070 -76.834 53.902
169907 94.084 -76.831 53.881
169908 94.098 -76.828 53.865
169909 94.112 -76.825 53.853
169910 94.125 -76.822 53.842
169911 94.139 -76.819 53.835
169912 94.153 -76.816 53.829
169913 94.166 -76.814 53.821
169914 94.174 -76.814 53.807
169915 94.179 -76.815 53.788
169916 94.185 -76.816 53.767
169917 94.191 -76.818 53.743
169918 94.197 -76.819 53.717
169919 94.203 -76.821 53.694
169920 94.209 -76.822 53.674
169921 94.215 -76.823 53.658
169922 94.240 -76.816 53.644
169923 94.269 -76.807 53.628
169924 94.297 -76.797 53.613
169925 94.327 -76.788 53.602
169926 94.353 -76.781 53.590
169927 94.370 -76.778 53.576
169928 94.387 -76.776 53.559
169929 94.405 -76.773 53.540
169930 94.423 -76.770 53.520
169931 94.441 -76.768 53.501
169932 94.459 -76.765 53.484
169933 94.477 -76.762 53.469
169934 94.495 -76.759 53.451
169935 94.511 -76.756 53.430
169936 94.527 -76.754 53.407
169937 94.531 -76.759 53.382
169938 94.531 -76.765 53.356
169939 94.532 -76.771 53.329
169940 94.533 -76.774 53.302
169941 94.533 -76.777 53.271
169942 94.548 -76.777 53.240
169943 94.597 -76.765 53.207
169944 94.648 -76.745 53.175
169945 94.700 -76.725 53.145
169946 94.751 -76.706 53.113
169947 94.790 -76.695 53.079
169948 94.826 -76.685 53.039
169949 94.845 -76.678 52.999
169950 94.857 -76.672 52.958
169951 94.869 -76.666 52.915
169952 94.881 -76.660 52.873
169953 94.893 -76.654 52.830
169954 94.904 -76.648 52.788
169955 94.916 -76.642 52.746
169956 94.928 -76.636 52.707
169957 94.940 -76.629 52.669
169958 94.952 -76.623 52.630
169959 94.964 -76.617 52.589
169960 94.975 -76.611 52.547
169961 94.987 -76.605 52.504
169962 94.999 -76.599 52.460
169963 95.011 -76.596 52.418
169964 95.036 -76.606 52.379
169965 95.061 -76.616 52.344
169966 95.092 -76.632 52.310
169967 95.125 -76.651 52.277
169968 95.159 -76.670 52.245
169969 95.192 -76.689 52.210
169970 95.226 -76.708 52.172
169971 95.253 -76.727 52.129
169972 95.251 -76.747 52.082
169973 95.250 -76.768 52.032
169974 95.248 -76.789 51.979
169975 95.253 -76.806 51.926
169976 95.263 -76.818 51.874
169977 95.272 -76.832 51.822
169978 95.282 -76.846 51.766
169979 95.293 -76.859 51.709
169980 95.308 -76.871 51.653
169981 95.322 -76.884 51.600
169982 95.341 -76.887 51.548
169983 95.359 -76.888 51.497
169984 95.378 -76.890 51.446
169985 95.397 -76.891 51.397
169986 95.414 -76.899 51.349
169987 95.431 -76.909 51.301
169988 95.448 -76.920 51.254
169989 95.465 -76.931 51.206
169990 95.482 -76.940 51.158
169991 95.498 -76.951 51.112
169992 95.522 -76.960 51.066
169993 95.549 -76.917 51.020
169994 95.552 -76.933 50.975
169995 95.561 -76.957 50.931
169996 95.580 -76.983 50.888
169997 95.599 -77.009 50.847
169998 95.618 -77.037 50.807
169999 95.637 -77.064 50.768
170000 95.656 -77.091 50.726
170001 95.667 -77.095 50.686
170002 95.678 -77.099 50.647
170003 95.686 -77.113 50.609
170004 95.691 -77.137 50.571
170005 95.697 -77.160 50.532
170006 95.710 -77.170 50.492
170007 95.723 -77.180 50.452
170008 95.736 -77.189 50.411
170009 95.749 -77.199 50.370
170010 95.762 -77.209 50.331
170011 95.776 -77.213 50.290
170012 95.792 -77.212 50.250
170013 95.807 -77.210 50.211
170014 95.826 -77.204 50.173
170015 95.851 -77.193 50.132
170016 95.877 -77.182 50.088
170017 95.902 -77.171 50.042
170018 95.927 -77.160 49.992
170019 95.930 -77.163 49.936
170020 95.921 -77.171 49.874
170021 95.913 -77.179 49.802
170022 95.918 -77.190 49.731
170023 95.930 -77.203 49.664
170024 95.942 -77.216 49.602
170025 95.954 -77.228 49.546
170026 95.970 -77.233 49.494
170027 95.986 -77.239 49.444
170028 96.003 -77.245 49.397
170029 96.019 -77.251 49.354
170030 96.035 -77.257 49.316
170031 96.051 -77.263 49.280
170032 96.068 -77.271 49.247
170033 96.085 -77.281 49.218
170034 96.102 -77.291 49.189
170035 96.119 -77.301 49.161
170036 96.136 -77.312 49.137
170037 96.154 -77.322 49.117
170038 96.171 -77.332 49.100
170039 96.191 -77.335 49.084
170040 96.212 -77.337 49.061
170041 96.232 -77.339 49.035
170042 96.252 -77.341 49.009
170043 96.276 -77.348 48.983
170044 96.301 -77.355 48.961
170045 96.325 -77.363 48.940
170046 96.350 -77.370 48.921
170047 96.375 -77.377 48.903
170048 96.397 -77.386 48.888
170049 96.413 -77.401 48.873
170050 96.430 -77.417 48.858
170051 96.446 -77.432 48.845
170052 96.463 -77.447 48.833
170053 96.479 -77.463 48.822
170054 96.496 -77.478 48.810
170055 96.512 -77.493 48.797
170056 96.528 -77.508 48.786
170057 96.545 -77.523 48.775
170058 96.561 -77.539 48.766
170059 96.572 -77.561 48.756
170060 96.582 -77.583 48.746
170061 96.592 -77.605 48.735
170062 96.603 -77.627 48.723
170063 96.623 -77.639 48.709
170064 96.647 -77.649 48.693
170065 96.670 -77.658 48.677
170066 96.692 -77.661 48.662
170067 96.714 -77.658 48.646
170068 96.734 -77.654 48.630
170069 96.755 -77.650 48.612
170070 96.777 -77.645 48.595
170071 96.798 -77.642 48.580
170072 96.819 -77.638 48.566
170073 96.842 -77.639 48.551
170074 96.872 -77.658 48.537
170075 96.901 -77.676 48.522
170076 96.931 -77.694 48.506
170077 96.960 -77.712 48.490
170078 96.990 -77.730 48.472
170079 97.007 -77.719 48.452
170080 97.019 -77.700 48.432
170081 97.032 -77.680 48.409
170082 97.044 -77.661 48.388
170083 97.056 -77.642 48.367
170084 97.061 -77.620 48.346
170085 97.065 -77.599 48.322
170086 97.070 -77.577 48.298
170087 97.074 -77.556 48.275
170088 97.079 -77.535 48.249
170089 97.095 -77.527 48.224
170090 97.127 -77.527 48.200
170091 97.150 -77.536 48.178
170092 97.173 -77.545 48.157
170093 97.197 -77.555 48.135
170094 97.220 -77.565 48.112
170095 97.243 -77.575 48.088
170096 97.267 -77.585 48.066
170097 97.290 -77.595 48.045
170098 97.314 -77.605 48.024
170099 97.337 -77.614 48.001
170100 97.361 -77.624 47.978
170101 97.384 -77.633 47.954
170102 97.407 -77.643 47.930
170103 97.431 -77.653 47.905
170104 97.454 -77.663 47.881
170105 97.477 -77.673 47.856
170106 97.501 -77.683 47.831
170107 97.524 -77.692 47.806
170108 97.539 -77.696 47.781
170109 97.543 -77.690 47.756
170110 97.550 -77.687 47.731
170111 97.570 -77.694 47.705
170112 97.589 -77.701 47.678
170113 97.608 -77.708 47.650
170114 97.635 -77.716 47.621
170115 97.673 -77.727 47.593
170116 97.712 -77.740 47.564
170117 97.751 -77.751 47.536
170118 97.789 -77.762 47.508
170119 97.827 -77.774 47.479
170120 97.864 -77.787 47.451
170121 97.900 -77.797 47.424
170122 97.889 -77.784 47.397
170123 97.850 -77.768 47.371
170124 97.812 -77.752 47.346
170125 97.773 -77.736 47.320
170126 97.751 -77.711 47.294
170127 97.784 -77.683 47.268
170128 97.825 -77.680 47.241
170129 97.866 -77.677 47.211
170130 97.907 -77.675 47.182
170131 97.941 -77.682 47.153
170132 97.968 -77.711 47.125
170133 98.003 -77.741 47.098
170134 98.037 -77.770 47.072
170135 98.045 -77.744 47.047
170136 98.086 -77.747 47.021
170137 98.120 -77.752 46.996
170138 98.139 -77.753 46.970
170139 98.158 -77.754 46.944
170140 98.177 -77.755 46.917
170141 98.196 -77.756 46.890
170142 98.199 -77.761 46.863
170143 98.201 -77.765 46.838
170144 98.203 -77.771 46.813
170145 98.211 -77.775 46.788
170146 98.220 -77.780 46.764
170147 98.236 -77.785 46.742
170148 98.258 -77.790 46.720
170149 98.280 -77.793 46.698
170150 98.301 -77.798 46.675
170151 98.322 -77.802 46.652
170152 98.344 -77.806 46.628
170153 98.365 -77.811 46.604
170154 98.398 -77.808 46.581
170155 98.421 -77.810 46.558
170156 98.443 -77.813 46.536
170157 98.465 -77.815 46.513
170158 98.488 -77.818 46.489
170159 98.511 -77.821 46.467
170160 98.533 -77.823 46.445
170161 98.554 -77.823 46.424
170162 98.568 -77.821 46.404
170163 98.583 -77.821 46.383
170164 98.597 -77.818 46.363
170165 98.600 -77.819 46.342
170166 98.603 -77.819 46.321
170167 98.606 -77.820 46.301
170168 98.610 -77.820 46.281
170169 98.613 -77.821 46.262
170170 98.617 -77.821 46.244
170171 98.620 -77.821 46.225
170172 98.623 -77.821 46.205
170173 98.626 -77.821 46.185
170174 98.631 -77.823 46.166
170175 98.650 -77.830 46.149
170176 98.668 -77.837 46.132
170177 98.687 -77.845 46.115
170178 98.706 -77.852 46.096
170179 98.724 -77.860 46.078
170180 98.746 -77.867 46.061
170181 98.788 -77.875 46.043
170182 98.830 -77.884 46.024
170183 98.872 -77.891 46.005
170184 98.914 -77.899 45.988
170185 98.975 -77.909 45.971
170186 99.041 -77.918 45.955
170187 99.065 -77.921 45.940
170188 99.090 -77.924 45.925
170189 99.115 -77.928 45.908
170190 99.139 -77.934 45.892
170191 99.163 -77.943 45.876
170192 99.187 -77.951 45.861
170193 99.211 -77.960 45.846
170194 99.233 -77.970 45.831
170195 99.253 -77.979 45.819
170196 99.274 -77.989 45.809
170197 99.294 -77.998 45.799
170198 99.316 -78.006 45.789
170199 99.336 -78.014 45.779
170200 99.349 -78.022 45.769
170201 99.362 -78.028 45.759
170202 99.375 -78.036 45.750
170203 99.388 -78.044 45.742
170204 99.401 -78.050 45.734
170205 99.415 -78.057 45.725
170206 99.439 -78.067 45.716
170207 99.465 -78.079 45.705
170208 99.492 -78.090 45.694
170209 99.519 -78.100 45.683
170210 99.546 -78.112 45.672
170211 99.573 -78.122 45.660
170212 99.604 -78.138 45.648
170213 99.635 -78.153 45.635
170214 99.665 -78.169 45.621
170215 99.696 -78.184 45.606
170216 99.726 -78.200 45.591
170217 99.757 -78.216 45.577
170218 99.788 -78.232 45.562
170219 99.819 -78.247 45.546
170220 99.841 -78.261 45.529
170221 99.855 -78.272 45.512
170222 99.872 -78.285 45.496
170223 99.892 -78.299 45.481
170224 99.911 -78.313 45.464
170225 99.930 -78.327 45.447
170226 99.949 -78.341 45.430
170227 99.969 -78.354 45.414
170228 99.988 -78.368 45.399
170229 100.007 -78.382 45.382
170230 100.028 -78.395 45.365
170231 100.051 -78.408 45.350
170232 100.075 -78.413 45.335
170233 100.100 -78.416 45.320
170234 100.124 -78.419 45.305
170235 100.148 -78.422 45.291
170236 100.173 -78.425 45.274
170237 100.197 -78.428 45.258
170238 100.221 -78.431 45.243
170239 100.254 -78.438 45.227
170240 100.291 -78.447 45.212
170241 100.329 -78.456 45.197
170242 100.366 -78.466 45.182
170243 100.403 -78.475 45.168
170244 100.440 -78.484 45.152
170245 100.478 -78.493 45.136
170246 100.515 -78.503 45.120
170247 100.553 -78.512 45.106
170248 100.590 -78.521 45.092
170249 100.641 -78.529 45.078
170250 100.699 -78.537 45.063
170251 100.758 -78.545 45.048
170252 100.795 -78.524 45.033
170253 100.781 -78.542 45.018
170254 100.768 -78.561 45.002
170255 100.796 -78.578 44.987
170256 100.824 -78.595 44.971
170257 100.853 -78.612 44.955
170258 100.881 -78.629 44.939
170259 100.913 -78.634 44.922
170260 100.946 -78.633 44.906
170261 100.980 -78.633 44.892
170262 101.013 -78.632 44.877
170263 101.045 -78.630 44.863
170264 101.073 -78.628 44.849
170265 101.101 -78.621 44.832
170266 101.130 -78.606 44.818
170267 101.176 -78.620 44.803
170268 101.221 -78.634 44.787
170269 101.266 -78.648 44.771
170270 101.306 -78.675 44.756
170271 101.319 -78.698 44.740
170272 101.329 -78.722 44.725
170273 101.339 -78.746 44.709
170274 101.349 -78.772 44.695
170275 101.360 -78.798 44.680
170276 101.370 -78.824 44.665
170277 101.389 -78.846 44.651
170278 101.415 -78.868 44.636
170279 101.442 -78.890 44.622
170280 101.484 -78.909 44.608
170281 101.527 -78.928 44.594
170282 101.569 -78.948 44.580
170283 101.611 -78.967 44.566
170284 101.653 -78.986 44.554
170285 101.683 -78.998 44.541
170286 101.699 -79.004 44.529
170287 101.716 -79.009 44.517
170288 101.733 -79.015 44.504
170289 101.750 -79.021 44.491
170290 101.766 -79.027 44.479
170291 101.783 -79.032 44.467
170292 101.800 -79.037 44.456
170293 101.817 -79.043 44.444
170294 101.833 -79.049 44.434
170295 101.850 -79.055 44.421
170523 92.628 -77.202 54.125
170524 92.651 -77.213 54.139
170525 92.674 -77.224 54.153
170526 92.697 -77.235 54.168
170527 92.720 -77.247 54.181
170528 92.736 -77.258 54.193
170529 92.743 -77.271 54.204
170530 92.750 -77.283 54.214
170531 92.757 -77.296 54.224
170532 92.764 -77.308 54.234
170533 92.771 -77.321 54.243
170534 92.778 -77.334 54.252
170535 92.790 -77.338 54.257
170536 92.810 -77.329 54.261
170537 92.831 -77.320 54.265
170538 92.851 -77.312 54.267
170539 92.871 -77.303 54.269
170540 92.892 -77.294 54.272
170541 92.912 -77.285 54.275
170542 92.932 -77.277 54.278
170543 92.953 -77.268 54.281
170544 92.973 -77.258 54.284
170545 92.992 -77.248 54.287
170546 93.012 -77.238 54.289
170547 93.031 -77.227 54.291
170548 93.051 -77.217 54.293
170549 93.070 -77.207 54.298
170550 93.090 -77.196 54.304
170551 93.109 -77.186 54.310
170552 93.130 -77.175 54.315
170553 93.149 -77.163 54.319
170554 93.165 -77.152 54.323
170555 93.183 -77.139 54.325
170556 93.201 -77.126 54.324
170557 93.213 -77.116 54.322
170558 93.217 -77.110 54.321
170559 93.226 -77.100 54.317
170560 93.236 -77.091 54.310
170561 93.246 -77.080 54.301
170562 93.256 -77.070 54.293
170563 93.266 -77.059 54.286
170564 93.276 -77.050 54.281
170565 93.305 -77.043 54.275
170566 93.340 -77.038 54.269
170567 93.375 -77.034 54.264
170568 93.409 -77.028 54.258
170569 93.439 -77.021 54.253
170570 93.469 -77.014 54.248
170571 93.489 -77.010 54.244
170572 93.510 -77.006 54.241
170573 93.531 -77.002 54.237
170574 93.547 -76.999 54.231
170575 93.558 -76.997 54.225
170576 93.575 -76.988 54.221
170577 93.588 -76.985 54.216
170578 93.601 -76.981 54.212
170579 93.614 -76.978 54.208
170580 93.626 -76.974 54.202
170581 93.639 -76.971 54.195
170582 93.652 -76.967 54.190
170583 93.664 -76.963 54.183
170584 93.678 -76.959 54.175
170585 93.690 -76.956 54.167
170586 93.703 -76.951 54.159
170587 93.713 -76.940 54.151
170588 93.721 -76.924 54.141
170589 93.728 -76.907 54.131
170590 93.735 -76.890 54.121
170591 93.749 -76.881 54.110
170592 93.767 -76.878 54.099
170593 93.785 -76.875 54.088
170594 93.806 -76.874 54.078
170595 93.835 -76.881 54.068
170596 93.864 -76.890 54.061
170597 93.893 -76.898 54.056
170598 93.919 -76.902 54.049
170599 93.936 -76.891 54.043
170600 93.953 -76.881 54.036
170601 93.970 -76.870 54.024
170602 93.987 -76.859 54.010
170603 94.004 -76.849 53.994
170604 94.021 -76.839 53.976
170605 94.038 -76.828 53.959
170606 94.052 -76.823 53.942
170607 94.065 -76.821 53.924
170608 94.077 -76.819 53.904
170609 94.090 -76.817 53.888
170610 94.103 -76.815 53.876
170611 94.115 -76.814 53.867
170612 94.127 -76.812 53.860
170613 94.140 -76.810 53.854
170614 94.153 -76.808 53.847
170615 94.171 -76.805 53.835
170616 94.162 -76.808 53.819
170617 94.168 -76.809 53.799
170618 94.174 -76.810 53.777
170619 94.180 -76.812 53.753
170620 94.186 -76.813 53.729
170621 94.192 -76.815 53.707
170622 94.202 -76.814 53.688
170623 94.230 -76.805 53.676
170624 94.260 -76.796 53.661
170625 94.289 -76.786 53.646
170626 94.318 -76.777 53.633
170627 94.347 -76.768 53.619
170628 94.366 -76.764 53.602
170629 94.384 -76.762 53.582
170630 94.401 -76.759 53.560
170631 94.419 -76.755 53.538
170632 94.437 -76.753 53.517
170633 94.456 -76.750 53.498
170634 94.473 -76.747 53.482
170635 94.491 -76.745 53.464
170636 94.509 -76.742 53.444
170637 94.523 -76.741 53.422
170638 94.524 -76.746 53.398
170639 94.524 -76.749 53.373
170640 94.525 -76.752 53.346
170641 94.525 -76.756 53.320
170642 94.526 -76.759 53.289
170643 94.551 -76.758 53.257
170644 94.598 -76.748 53.225
170645 94.649 -76.728 53.194
170646 94.701 -76.709 53.162
170647 94.750 -76.691 53.130
170648 94.786 -76.681 53.093
170649 94.823 -76.672 53.053
170650 94.841 -76.665 53.012
170651 94.853 -76.658 52.970
170652 94.865 -76.652 52.927
170653 94.877 -76.646 52.886
170654 94.889 -76.640 52.845
170655 94.900 -76.634 52.806
170656 94.912 -76.628 52.768
170657 94.924 -76.622 52.732
170658 94.936 -76.616 52.695
170659 94.948 -76.610 52.656
170660 94.960 -76.604 52.615
170661 94.971 -76.598 52.573
170662 94.980 -76.592 52.528
170663 94.992 -76.590 52.481
170664 95.018 -76.600 52.436
170665 95.043 -76.610 52.391
170666 95.068 -76.620 52.349
170667 95.093 -76.629 52.312
170668 95.124 -76.645 52.277
170669 95.158 -76.664 52.242
170670 95.191 -76.683 52.205
170671 95.224 -76.701 52.166
170672 95.228 -76.721 52.124
170673 95.226 -76.743 52.079
170674 95.229 -76.760 52.031
170675 95.239 -76.774 51.982
170676 95.248 -76.787 51.931
170677 95.258 -76.801 51.880
170678 95.267 -76.814 51.826
170679 95.276 -76.828 51.770
170680 95.288 -76.840 51.715
170681 95.304 -76.853 51.662
170682 95.318 -76.866 51.611
170683 95.334 -76.877 51.561
170684 95.352 -76.878 51.511
170685 95.371 -76.880 51.462
170686 95.390 -76.881 51.414
170687 95.408 -76.883 51.366
170688 95.427 -76.884 51.318
170689 95.445 -76.889 51.271
170690 95.462 -76.900 51.223
170691 95.479 -76.910 51.176
170692 95.495 -76.920 51.130
170693 95.519 -76.929 51.085
170694 95.534 -76.906 51.039
170695 95.543 -76.931 50.993
170696 95.558 -76.950 50.948
170697 95.573 -76.969 50.905
170698 95.589 -76.988 50.863
170699 95.604 -77.008 50.823
170700 95.619 -77.027 50.783
170701 95.634 -77.047 50.743
170702 95.654 -77.073 50.703
170703 95.670 -77.091 50.664
170704 95.678 -77.106 50.625
170705 95.688 -77.120 50.586
170706 95.701 -77.130 50.546
170707 95.714 -77.140 50.507
170708 95.728 -77.150 50.467
170709 95.741 -77.159 50.427
170710 95.754 -77.169 50.387
170711 95.767 -77.179 50.347
170712 95.780 -77.188 50.307
170713 95.795 -77.188 50.267
170714 95.810 -77.186 50.226
170715 95.825 -77.185 50.186
170716 95.840 -77.183 50.144
170717 95.861 -77.176 50.101
170718 95.886 -77.165 50.056
170719 95.911 -77.154 50.006
170720 95.930 -77.147 49.952
170721 95.921 -77.156 49.891
170722 95.915 -77.165 49.818
170723 95.927 -77.178 49.746
170724 95.939 -77.190 49.678
170725 95.950 -77.203 49.614
170726 95.962 -77.215 49.558
170727 95.978 -77.220 49.504
170728 95.994 -77.226 49.454
170729 96.011 -77.235 49.408
170730 96.029 -77.244 49.365
170731 96.046 -77.255 49.326
170732 96.063 -77.265 49.290
170733 96.080 -77.275 49.256
170734 96.097 -77.285 49.226
170735 96.114 -77.295 49.197
170736 96.131 -77.306 49.168
170737 96.149 -77.316 49.141
170738 96.166 -77.326 49.120
170739 96.183 -77.335 49.100
170740 96.204 -77.338 49.080
170741 96.224 -77.339 49.060
170742 96.245 -77.341 49.036
170743 96.265 -77.344 49.013
170744 96.286 -77.346 48.990
170745 96.306 -77.348 48.969
170746 96.330 -77.355 48.950
170747 96.355 -77.362 48.931
170748 96.379 -77.370 48.914
170749 96.399 -77.382 48.899
170750 96.415 -77.397 48.883
170751 96.432 -77.412 48.868
170752 96.448 -77.427 48.855
170753 96.465 -77.442 48.842
170754 96.481 -77.457 48.830
170755 96.498 -77.473 48.818
170756 96.514 -77.489 48.804
170757 96.530 -77.504 48.791
170758 96.547 -77.519 48.780
170759 96.564 -77.534 48.771
170760 96.578 -77.551 48.763
170761 96.589 -77.573 48.755
170762 96.599 -77.595 48.747
170763 96.613 -77.614 48.737
170764 96.637 -77.624 48.723
170765 96.660 -77.633 48.706
170766 96.683 -77.642 48.689
170767 96.706 -77.652 48.672
170768 96.729 -77.654 48.657
170769 96.749 -77.650 48.641
170770 96.770 -77.646 48.623
170771 96.792 -77.642 48.607
170772 96.813 -77.638 48.592
170773 96.834 -77.634 48.576
170774 96.861 -77.644 48.561
170775 96.890 -77.662 48.546
170776 96.920 -77.680 48.531
170777 96.949 -77.699 48.516
170778 96.979 -77.717 48.499
170779 97.008 -77.735 48.482
170780 97.023 -77.724 48.464
170781 97.029 -77.702 48.445
170782 97.035 -77.680 48.424
170783 97.040 -77.657 48.403
170784 97.046 -77.635 48.382
170785 97.052 -77.612 48.361
170786 97.058 -77.590 48.336
170787 97.064 -77.567 48.313
170788 97.070 -77.545 48.289
170789 97.079 -77.533 48.264
170790 97.095 -77.523 48.239
170791 97.136 -77.521 48.216
170792 97.160 -77.529 48.194
170793 97.182 -77.537 48.172
170794 97.205 -77.545 48.149
170795 97.228 -77.553 48.126
170796 97.251 -77.563 48.103
170797 97.275 -77.573 48.081
170798 97.298 -77.583 48.059
170799 97.321 -77.592 48.037
170800 97.345 -77.602 48.015
170801 97.368 -77.612 47.991
170802 97.391 -77.621 47.968
170803 97.415 -77.631 47.943
170804 97.438 -77.641 47.919
170805 97.461 -77.651 47.894
170806 97.485 -77.661 47.868
170807 97.508 -77.671 47.843
170808 97.532 -77.680 47.817
170809 97.555 -77.690 47.792
170810 97.559 -77.685 47.767
170811 97.563 -77.680 47.742
170812 97.572 -77.677 47.716
170813 97.591 -77.684 47.688
170814 97.613 -77.692 47.661
170815 97.652 -77.703 47.632
170816 97.690 -77.715 47.604
170817 97.728 -77.727 47.575
170818 97.766 -77.739 47.546
170819 97.803 -77.751 47.518
170820 97.840 -77.763 47.490
170821 97.877 -77.775 47.461
170822 97.910 -77.782 47.433
170823 97.918 -77.771 47.405
170824 97.879 -77.755 47.379
170825 97.841 -77.740 47.353
170826 97.802 -77.724 47.327
170827 97.778 -77.699 47.300
170828 97.803 -77.645 47.274
170829 97.838 -77.624 47.247
170830 97.879 -77.621 47.217
170831 97.920 -77.618 47.188
170832 97.957 -77.621 47.159
170833 97.984 -77.651 47.131
170834 98.019 -77.680 47.104
170835 98.053 -77.709 47.078
170836 98.088 -77.739 47.053
170837 98.110 -77.743 47.028
170838 98.148 -77.738 47.002
170839 98.170 -77.740 46.977
170840 98.190 -77.741 46.951
170841 98.209 -77.743 46.922
170842 98.220 -77.745 46.894
170843 98.222 -77.749 46.867
170844 98.224 -77.754 46.841
170845 98.226 -77.759 46.815
170846 98.231 -77.764 46.790
170847 98.241 -77.769 46.766
170848 98.250 -77.774 46.744
170849 98.268 -77.779 46.721
170850 98.289 -77.783 46.699
170851 98.311 -77.787 46.677
170852 98.332 -77.791 46.655
170853 98.354 -77.796 46.632
170854 98.383 -77.794 46.609
170855 98.414 -77.793 46.587
170856 98.436 -77.796 46.563
170857 98.459 -77.798 46.541
170858 98.482 -77.801 46.518
170859 98.504 -77.803 46.495
170860 98.527 -77.806 46.472
170861 98.550 -77.808 46.450
170862 98.568 -77.813 46.429
170863 98.587 -77.815 46.408
170864 98.601 -77.814 46.387
170865 98.612 -77.812 46.367
170866 98.615 -77.813 46.346
170867 98.619 -77.813 46.324
170868 98.622 -77.814 46.304
170869 98.625 -77.814 46.284
170870 98.628 -77.815 46.265
170871 98.631 -77.815 46.246
170872 98.635 -77.815 46.226
170873 98.638 -77.815 46.206
170874 98.641 -77.815 46.185
170875 98.644 -77.816 46.167
170876 98.652 -77.818 46.148
170877 98.670 -77.825 46.130
170878 98.689 -77.833 46.113
170879 98.708 -77.840 46.095
170880 98.726 -77.848 46.077
170881 98.758 -77.855 46.059
170882 98.800 -77.863 46.042
170883 98.842 -77.871 46.024
170884 98.884 -77.879 46.006
170885 98.926 -77.887 45.988
170886 98.988 -77.897 45.971
170887 99.029 -77.903 45.955
170888 99.055 -77.906 45.940
170889 99.079 -77.910 45.925
170890 99.103 -77.919 45.909
170891 99.127 -77.928 45.894
170892 99.150 -77.937 45.877
170893 99.174 -77.946 45.863
170894 99.198 -77.955 45.849
170895 99.221 -77.964 45.835
170896 99.242 -77.974 45.824
170897 99.262 -77.983 45.815
170898 99.283 -77.993 45.806
170899 99.303 -78.003 45.795
170900 99.324 -78.012 45.785
170901 99.340 -78.020 45.774
170902 99.354 -78.027 45.764
170903 99.368 -78.034 45.755
170904 99.382 -78.040 45.747
170905 99.397 -78.048 45.739
170906 99.411 -78.055 45.731
170907 99.439 -78.060 45.722
170908 99.471 -78.070 45.712
170909 99.494 -78.086 45.701
170910 99.517 -78.102 45.690
170911 99.544 -78.112 45.677
170912 99.572 -78.125 45.665
170913 99.603 -78.141 45.653
170914 99.633 -78.156 45.639
170915 99.664 -78.172 45.625
170916 99.695 -78.188 45.610
170917 99.725 -78.203 45.596
170918 99.756 -78.219 45.581
170919 99.787 -78.235 45.567
170920 99.818 -78.250 45.551
170921 99.848 -78.266 45.535
170922 99.866 -78.279 45.519
170923 99.885 -78.292 45.503
170924 99.904 -78.306 45.486
170925 99.924 -78.319 45.470
170926 99.943 -78.333 45.453
170927 99.962 -78.347 45.436
170928 99.982 -78.361 45.420
170929 100.001 -78.375 45.404
170930 100.020 -78.388 45.388
170931 100.039 -78.402 45.372
170932 100.062 -78.407 45.357
170933 100.086 -78.410 45.342
170934 100.111 -78.413 45.327
170935 100.135 -78.416 45.312
170936 100.159 -78.420 45.297
170937 100.184 -78.423 45.281
170938 100.208 -78.426 45.265
170939 100.232 -78.429 45.249
170940 100.257 -78.432 45.234
170941 100.291 -78.440 45.219
170942 100.329 -78.449 45.205
170943 100.366 -78.459 45.190
170944 100.403 -78.468 45.175
170945 100.440 -78.477 45.159
170946 100.478 -78.486 45.143
170947 100.515 -78.496 45.127
170948 100.553 -78.505 45.115
170949 100.592 -78.514 45.100
170950 100.650 -78.522 45.085
170951 100.709 -78.529 45.070
170952 100.767 -78.536 45.054
170953 100.825 -78.544 45.039
170954 100.844 -78.537 45.024
170955 100.830 -78.555 45.008
170956 100.816 -78.573 44.992
170957 100.826 -78.591 44.976
170958 100.854 -78.608 44.961
170959 100.882 -78.625 44.945
170960 100.915 -78.626 44.928
170961 100.949 -78.626 44.913
170962 100.982 -78.625 44.899
170963 101.016 -78.624 44.884
170964 101.049 -78.623 44.870
170965 101.078 -78.621 44.855
170966 101.106 -78.610 44.839
170967 101.136 -78.597 44.824
170968 101.182 -78.611 44.808
170969 101.227 -78.625 44.793
170970 101.272 -78.639 44.777
170971 101.310 -78.657 44.762
170972 101.321 -78.682 44.746
170973 101.331 -78.708 44.730
170974 101.341 -78.733 44.715
170975 101.352 -78.758 44.701
170976 101.362 -78.784 44.686
170977 101.376 -78.809 44.671
170978 101.401 -78.831 44.656
170979 101.427 -78.852 44.642
170980 101.453 -78.874 44.627
170981 101.484 -78.896 44.612
170982 101.527 -78.915 44.598
170983 101.568 -78.933 44.584
170984 101.611 -78.953 44.571
170985 101.653 -78.972 44.559
170986 101.683 -78.985 44.546
170987 101.700 -78.990 44.533
170988 101.717 -78.996 44.521
170989 101.734 -79.002 44.508
170990 101.751 -79.008 44.496
170991 101.767 -79.013 44.483
170992 101.784 -79.019 44.471
170993 101.801 -79.024 44.460
170994 101.818 -79.030 44.449
170995 101.834 -79.036 44.437
170996 101.851 -79.041 44.425
170997 101.868 -79.047 44.413
171225 92.639 -77.184 54.122
171226 92.661 -77.195 54.136
171227 92.685 -77.207 54.151
171228 92.708 -77.217 54.164
171229 92.729 -77.229 54.176
171230 92.736 -77.241 54.187
171231 92.743 -77.254 54.197
171232 92.750 -77.266 54.207
171233 92.757 -77.279 54.216
171234 92.764 -77.291 54.226
171235 92.772 -77.304 54.233
171236 92.780 -77.314 54.239
171237 92.800 -77.306 54.244
171238 92.820 -77.297 54.247
171239 92.841 -77.288 54.250
171240 92.861 -77.279 54.253
171241 92.881 -77.271 54.256
171242 92.902 -77.262 54.260
171243 92.922 -77.253 54.263
171244 92.942 -77.244 54.266
171245 92.963 -77.235 54.269
171246 92.983 -77.227 54.272
171247 93.003 -77.218 54.275
171248 93.023 -77.207 54.278
171249 93.042 -77.197 54.281
171250 93.062 -77.187 54.286
171251 93.081 -77.176 54.292
171252 93.101 -77.166 54.299
171253 93.120 -77.156 54.304
171254 93.140 -77.145 54.310
171255 93.159 -77.135 54.315
171256 93.176 -77.122 54.318
171257 93.194 -77.109 54.317
171258 93.207 -77.097 54.317
171259 93.211 -77.091 54.316
171260 93.216 -77.085 54.313
171261 93.222 -77.078 54.304
171262 93.232 -77.067 54.295
171263 93.243 -77.056 54.286
171264 93.253 -77.047 54.279
171265 93.263 -77.036 54.275
171266 93.291 -77.030 54.271
171267 93.327 -77.025 54.267
171268 93.362 -77.021 54.263
171269 93.398 -77.016 54.260
171270 93.428 -77.009 54.257
171271 93.453 -77.003 54.254
171272 93.474 -77.000 54.251
171273 93.495 -76.996 54.249
171274 93.516 -76.991 54.246
171275 93.537 -76.987 54.241
171276 93.553 -76.983 54.235
171277 93.565 -76.979 54.231
171278 93.578 -76.976 54.228
171279 93.591 -76.972 54.227
171280 93.604 -76.967 54.224
171281 93.617 -76.962 54.218
171282 93.629 -76.957 54.212
171283 93.642 -76.952 54.207
171284 93.655 -76.947 54.200
171285 93.667 -76.941 54.192
171286 93.680 -76.937 54.183
171287 93.694 -76.932 54.177
171288 93.712 -76.928 54.169
171289 93.719 -76.911 54.159
171290 93.726 -76.894 54.148
171291 93.734 -76.878 54.137
171292 93.749 -76.871 54.126
171293 93.767 -76.868 54.114
171294 93.786 -76.864 54.101
171295 93.804 -76.861 54.087
171296 93.829 -76.864 54.076
171297 93.857 -76.872 54.067
171298 93.886 -76.881 54.061
171299 93.914 -76.886 54.055
171300 93.931 -76.876 54.050
171301 93.948 -76.865 54.044
171302 93.965 -76.855 54.035
171303 93.982 -76.844 54.023
171304 93.999 -76.834 54.010
171305 94.016 -76.823 53.995
171306 94.033 -76.816 53.980
171307 94.046 -76.814 53.964
171308 94.052 -76.815 53.947
171309 94.064 -76.813 53.927
171310 94.077 -76.812 53.912
171311 94.089 -76.810 53.900
171312 94.102 -76.808 53.892
171313 94.114 -76.806 53.884
171314 94.127 -76.805 53.878
171315 94.143 -76.802 53.872
171316 94.163 -76.798 53.861
171317 94.163 -76.798 53.847
171318 94.151 -76.802 53.830
171319 94.156 -76.802 53.810
171320 94.163 -76.804 53.788
171321 94.169 -76.806 53.764
171322 94.175 -76.807 53.741
171323 94.193 -76.802 53.722
171324 94.222 -76.793 53.710
171325 94.251 -76.784 53.696
171326 94.280 -76.774 53.681
171327 94.309 -76.765 53.666
171328 94.338 -76.756 53.651
171329 94.362 -76.749 53.631
171330 94.380 -76.747 53.607
171331 94.398 -76.744 53.580
171332 94.415 -76.742 53.555
171333 94.433 -76.739 53.531
171334 94.452 -76.736 53.509
171335 94.470 -76.733 53.491
171336 94.487 -76.730 53.473
171337 94.505 -76.727 53.453
171338 94.515 -76.726 53.433
171339 94.516 -76.728 53.411
171340 94.517 -76.732 53.387
171341 94.518 -76.735 53.362
171342 94.518 -76.739 53.335
171343 94.519 -76.742 53.305
171344 94.554 -76.738 53.273
171345 94.599 -76.731 53.242
171346 94.651 -76.712 53.211
171347 94.702 -76.693 53.179
171348 94.747 -76.677 53.146
171349 94.782 -76.668 53.110
171350 94.819 -76.658 53.069
171351 94.837 -76.652 53.026
171352 94.849 -76.646 52.983
171353 94.861 -76.640 52.940
171354 94.873 -76.633 52.900
171355 94.885 -76.627 52.862
171356 94.897 -76.621 52.826
171357 94.908 -76.615 52.792
171358 94.920 -76.609 52.757
171359 94.932 -76.602 52.721
171360 94.944 -76.596 52.683
171361 94.955 -76.590 52.643
171362 94.961 -76.584 52.601
171363 94.974 -76.584 52.556
171364 94.999 -76.594 52.506
171365 95.025 -76.604 52.458
171366 95.050 -76.614 52.409
171367 95.075 -76.623 52.360
171368 95.101 -76.633 52.319
171369 95.126 -76.643 52.280
171370 95.156 -76.658 52.242
171371 95.189 -76.677 52.203
171372 95.204 -76.696 52.162
171373 95.205 -76.715 52.120
171374 95.214 -76.729 52.076
171375 95.224 -76.743 52.031
171376 95.233 -76.755 51.985
171377 95.243 -76.769 51.936
171378 95.252 -76.783 51.885
171379 95.262 -76.796 51.830
171380 95.271 -76.809 51.776
171381 95.284 -76.823 51.723
171382 95.299 -76.836 51.673
171383 95.314 -76.849 51.625
171384 95.328 -76.862 51.577
171385 95.345 -76.869 51.529
171386 95.364 -76.871 51.481
171387 95.382 -76.872 51.433
171388 95.401 -76.874 51.385
171389 95.419 -76.875 51.337
171390 95.438 -76.876 51.290
171391 95.457 -76.878 51.242
171392 95.475 -76.879 51.195
171393 95.492 -76.890 51.149
171394 95.513 -76.890 51.103
171395 95.524 -76.904 51.058
171396 95.539 -76.924 51.012
171397 95.555 -76.943 50.966
171398 95.570 -76.962 50.922
171399 95.586 -76.981 50.880
171400 95.601 -77.000 50.839
171401 95.616 -77.019 50.799
171402 95.632 -77.038 50.760
171403 95.648 -77.056 50.721
171404 95.663 -77.075 50.682
171405 95.678 -77.086 50.642
171406 95.693 -77.091 50.602
171407 95.706 -77.100 50.562
171408 95.720 -77.110 50.523
171409 95.733 -77.119 50.484
171410 95.746 -77.128 50.444
171411 95.759 -77.138 50.404
171412 95.772 -77.148 50.365
171413 95.785 -77.158 50.325
171414 95.799 -77.165 50.284
171415 95.814 -77.163 50.243
171416 95.829 -77.161 50.202
171417 95.844 -77.160 50.159
171418 95.859 -77.157 50.115
171419 95.874 -77.156 50.070
171420 95.896 -77.148 50.021
171421 95.921 -77.137 49.967
171422 95.920 -77.141 49.907
171423 95.924 -77.152 49.834
171424 95.936 -77.165 49.761
171425 95.947 -77.177 49.692
171426 95.958 -77.189 49.627
171427 95.976 -77.201 49.569
171428 95.990 -77.216 49.515
171429 96.006 -77.229 49.465
171430 96.023 -77.238 49.419
171431 96.040 -77.249 49.376
171432 96.057 -77.259 49.337
171433 96.075 -77.269 49.301
171434 96.092 -77.279 49.267
171435 96.109 -77.290 49.236
171436 96.126 -77.300 49.206
171437 96.144 -77.310 49.177
171438 96.161 -77.320 49.150
171439 96.178 -77.331 49.127
171440 96.196 -77.338 49.106
171441 96.217 -77.340 49.085
171442 96.237 -77.341 49.064
171443 96.258 -77.344 49.042
171444 96.278 -77.346 49.020
171445 96.299 -77.348 48.997
171446 96.319 -77.351 48.979
171447 96.340 -77.353 48.961
171448 96.360 -77.355 48.943
171449 96.384 -77.361 48.926
171450 96.401 -77.376 48.911
171451 96.417 -77.392 48.896
171452 96.434 -77.407 48.880
171453 96.450 -77.423 48.867
171454 96.467 -77.438 48.853
171455 96.483 -77.453 48.841
171456 96.500 -77.468 48.828
171457 96.516 -77.484 48.814
171458 96.532 -77.499 48.801
171459 96.549 -77.514 48.788
171460 96.566 -77.530 48.777
171461 96.582 -77.545 48.769
171462 96.595 -77.564 48.761
171463 96.605 -77.586 48.753
171464 96.626 -77.598 48.744
171465 96.649 -77.607 48.730
171466 96.673 -77.617 48.715
171467 96.696 -77.626 48.698
171468 96.719 -77.635 48.683
171469 96.743 -77.645 48.667
171470 96.764 -77.646 48.651
171471 96.786 -77.642 48.633
171472 96.807 -77.638 48.618
171473 96.828 -77.634 48.603
171474 96.849 -77.630 48.587
171475 96.879 -77.649 48.571
171476 96.908 -77.667 48.555
171477 96.938 -77.685 48.539
171478 96.967 -77.702 48.523
171479 97.001 -77.719 48.507
171480 97.028 -77.723 48.492
171481 97.034 -77.700 48.475
171482 97.039 -77.678 48.456
171483 97.045 -77.655 48.436
171484 97.051 -77.633 48.416
171485 97.057 -77.611 48.395
171486 97.063 -77.588 48.374
171487 97.070 -77.571 48.349
171488 97.079 -77.560 48.326
171489 97.088 -77.548 48.302
171490 97.098 -77.537 48.277
171491 97.105 -77.522 48.253
171492 97.140 -77.518 48.230
171493 97.170 -77.523 48.208
171494 97.193 -77.530 48.185
171495 97.216 -77.539 48.162
171496 97.238 -77.547 48.139
171497 97.260 -77.555 48.117
171498 97.283 -77.563 48.095
171499 97.306 -77.571 48.073
171500 97.329 -77.580 48.051
171501 97.353 -77.590 48.028
171502 97.376 -77.600 48.004
171503 97.399 -77.610 47.980
171504 97.423 -77.620 47.955
171505 97.446 -77.629 47.930
171506 97.469 -77.639 47.905
171507 97.493 -77.649 47.880
171508 97.516 -77.658 47.854
171509 97.539 -77.668 47.828
171510 97.563 -77.678 47.802
171511 97.575 -77.679 47.777
171512 97.579 -77.674 47.751
171513 97.583 -77.668 47.725
171514 97.593 -77.668 47.698
171515 97.630 -77.680 47.671
171516 97.667 -77.691 47.643
171517 97.704 -77.703 47.614
171518 97.741 -77.715 47.585
171519 97.779 -77.727 47.555
171520 97.816 -77.740 47.528
171521 97.853 -77.752 47.499
171522 97.890 -77.764 47.470
171523 97.920 -77.767 47.442
171524 97.937 -77.757 47.414
171525 97.909 -77.742 47.387
171526 97.870 -77.726 47.361
171527 97.831 -77.711 47.335
171528 97.804 -77.687 47.307
171529 97.829 -77.633 47.279
171530 97.855 -77.578 47.251
171531 97.892 -77.565 47.223
171532 97.933 -77.562 47.194
171533 97.973 -77.561 47.165
171534 98.000 -77.590 47.137
171535 98.035 -77.620 47.110
171536 98.069 -77.649 47.084
171537 98.104 -77.678 47.059
171538 98.138 -77.707 47.035
171539 98.173 -77.736 47.009
171540 98.202 -77.727 46.983
171541 98.221 -77.728 46.957
171542 98.240 -77.729 46.929
171543 98.243 -77.733 46.901
171544 98.245 -77.738 46.873
171545 98.247 -77.743 46.845
171546 98.249 -77.748 46.819
171547 98.252 -77.752 46.794
171548 98.262 -77.758 46.769
171549 98.271 -77.763 46.746
171550 98.281 -77.768 46.723
171551 98.299 -77.772 46.700
171552 98.320 -77.777 46.678
171553 98.342 -77.780 46.656
171554 98.367 -77.783 46.634
171555 98.406 -77.776 46.612
171556 98.430 -77.778 46.591
171557 98.452 -77.780 46.568
171558 98.475 -77.783 46.545
171559 98.497 -77.786 46.523
171560 98.520 -77.789 46.499
171561 98.544 -77.788 46.476
171562 98.568 -77.787 46.454
171563 98.582 -77.798 46.433
171564 98.591 -77.815 46.412
171565 98.608 -77.819 46.392
171566 98.627 -77.806 46.371
171567 98.630 -77.807 46.350
171568 98.633 -77.807 46.328
171569 98.637 -77.808 46.307
171570 98.640 -77.808 46.287
171571 98.643 -77.808 46.267
171572 98.646 -77.808 46.247
171573 98.650 -77.808 46.227
171574 98.653 -77.809 46.207
171575 98.656 -77.809 46.186
171576 98.659 -77.810 46.167
171577 98.662 -77.810 46.148
171578 98.672 -77.813 46.129
171579 98.691 -77.821 46.112
171580 98.710 -77.828 46.094
171581 98.728 -77.836 46.075
171582 98.770 -77.843 46.058
171583 98.812 -77.852 46.041
171584 98.854 -77.859 46.023
171585 98.896 -77.868 46.005
171586 98.938 -77.875 45.988
171587 98.994 -77.884 45.971
171588 99.019 -77.887 45.955
171589 99.043 -77.896 45.939
171590 99.067 -77.905 45.924
171591 99.090 -77.913 45.910
171592 99.114 -77.922 45.896
171593 99.138 -77.931 45.880
171594 99.162 -77.940 45.866
171595 99.186 -77.949 45.852
171596 99.209 -77.959 45.839
171597 99.230 -77.968 45.829
171598 99.251 -77.978 45.820
171599 99.271 -77.987 45.811
171600 99.294 -77.996 45.801
171601 99.333 -78.002 45.790
171602 99.338 -78.015 45.779
171603 99.350 -78.025 45.768
171604 99.364 -78.031 45.759
171605 99.378 -78.038 45.751
171606 99.392 -78.045 45.743
171607 99.406 -78.052 45.735
171608 99.432 -78.057 45.728
171609 99.476 -78.060 45.717
171610 99.503 -78.074 45.706
171611 99.526 -78.090 45.695
171612 99.549 -78.105 45.683
171613 99.567 -78.119 45.670
171614 99.588 -78.134 45.657
171615 99.619 -78.149 45.643
171616 99.649 -78.165 45.629
171617 99.680 -78.180 45.615
171618 99.710 -78.196 45.600
171619 99.740 -78.211 45.586
171620 99.771 -78.227 45.571
171621 99.801 -78.242 45.556
171622 99.832 -78.257 45.540
171623 99.874 -78.280 45.525
171624 99.893 -78.294 45.509
171625 99.911 -78.307 45.492
171626 99.930 -78.320 45.475
171627 99.949 -78.334 45.458
171628 99.968 -78.347 45.442
171629 99.986 -78.360 45.426
171630 100.005 -78.373 45.409
171631 100.028 -78.385 45.394
171632 100.056 -78.392 45.379
171633 100.064 -78.403 45.364
171634 100.094 -78.407 45.349
171635 100.122 -78.412 45.334
171636 100.146 -78.415 45.319
171637 100.171 -78.418 45.303
171638 100.195 -78.421 45.287
171639 100.219 -78.424 45.272
171640 100.243 -78.427 45.257
171641 100.268 -78.430 45.241
171642 100.292 -78.433 45.227
171643 100.329 -78.442 45.212
171644 100.366 -78.451 45.196
171645 100.404 -78.461 45.181
171646 100.441 -78.470 45.165
171647 100.478 -78.479 45.149
171648 100.515 -78.489 45.134
171649 100.553 -78.498 45.121
171650 100.602 -78.507 45.107
171651 100.660 -78.514 45.093
171652 100.718 -78.521 45.077
171653 100.776 -78.529 45.062
171654 100.835 -78.536 45.046
171655 100.893 -78.544 45.030
171656 100.893 -78.549 45.013
171657 100.879 -78.567 44.998
171658 100.865 -78.586 44.982
171659 100.856 -78.604 44.967
171660 100.885 -78.620 44.950
171661 100.918 -78.619 44.934
171662 100.952 -78.618 44.919
171663 100.985 -78.617 44.906
171664 101.019 -78.617 44.891
171665 101.052 -78.616 44.876
171666 101.083 -78.614 44.861
171667 101.111 -78.598 44.846
171668 101.142 -78.588 44.830
171669 101.187 -78.601 44.813
171670 101.233 -78.616 44.798
171671 101.286 -78.630 44.784
171672 101.312 -78.645 44.768
171673 101.323 -78.669 44.753
171674 101.334 -78.694 44.737
171675 101.344 -78.720 44.721
171676 101.354 -78.745 44.706
171677 101.365 -78.771 44.692
171678 101.388 -78.794 44.677
171679 101.413 -78.815 44.662
171680 101.439 -78.837 44.647
171681 101.465 -78.858 44.632
171682 101.490 -78.880 44.617
171683 101.526 -78.901 44.603
171684 101.568 -78.920 44.589
171685 101.611 -78.940 44.576
171686 101.653 -78.959 44.563
171687 101.684 -78.972 44.550
171688 101.701 -78.977 44.537
171689 101.718 -78.984 44.525
171690 101.735 -78.989 44.513
171691 101.751 -78.994 44.500
171692 101.768 -79.000 44.487
171693 101.785 -79.006 44.475
171694 101.802 -79.012 44.464
171695 101.818 -79.017 44.453
171696 101.835 -79.022 44.441
171697 101.852 -79.028 44.429
171698 101.869 -79.034 44.417
171699 101.885 -79.040 44.404
171927 92.649 -77.166 54.117
171928 92.672 -77.177 54.133
171929 92.695 -77.188 54.146
171930 92.718 -77.199 54.158
171931 92.729 -77.212 54.170
171932 92.736 -77.224 54.180
171933 92.743 -77.237 54.189
171934 92.751 -77.249 54.198
171935 92.758 -77.262 54.207
171936 92.765 -77.274 54.214
171937 92.772 -77.287 54.221
171938 92.790 -77.282 54.225
171939 92.810 -77.273 54.229
171940 92.830 -77.264 54.232
171941 92.851 -77.256 54.236
171942 92.871 -77.247 54.240
171943 92.892 -77.238 54.244
171944 92.912 -77.229 54.247
171945 92.932 -77.221 54.251
171946 92.952 -77.212 54.254
171947 92.973 -77.203 54.257
171948 92.993 -77.194 54.260
171949 93.013 -77.185 54.264
171950 93.034 -77.177 54.267
171951 93.054 -77.167 54.271
171952 93.073 -77.157 54.277
171953 93.092 -77.147 54.282
171954 93.112 -77.136 54.287
171955 93.131 -77.125 54.293
171956 93.151 -77.116 54.299
171957 93.169 -77.103 54.305
171958 93.187 -77.091 54.309
171959 93.201 -77.079 54.311
171960 93.206 -77.073 54.311
171961 93.210 -77.067 54.309
171962 93.214 -77.061 54.299
171963 93.219 -77.054 54.290
171964 93.229 -77.044 54.281
171965 93.239 -77.034 54.274
171966 93.250 -77.023 54.271
171967 93.278 -77.017 54.269
171968 93.313 -77.012 54.267
171969 93.348 -77.007 54.265
171970 93.384 -77.003 54.264
171971 93.417 -76.997 54.263
171972 93.438 -76.993 54.260
171973 93.459 -76.989 54.258
171974 93.479 -76.984 54.257
171975 93.499 -76.978 54.254
171976 93.519 -76.973 54.250
171977 93.541 -76.970 54.244
171978 93.556 -76.963 54.240
171979 93.568 -76.957 54.238
171980 93.581 -76.953 54.239
171981 93.593 -76.947 54.237
171982 93.606 -76.942 54.233
171983 93.619 -76.937 54.228
171984 93.632 -76.931 54.223
171985 93.644 -76.927 54.216
171986 93.657 -76.922 54.208
171987 93.670 -76.916 54.200
171988 93.688 -76.915 54.193
171989 93.710 -76.915 54.185
171990 93.718 -76.899 54.175
171991 93.725 -76.882 54.166
171992 93.732 -76.865 54.155
171993 93.749 -76.860 54.144
171994 93.768 -76.857 54.131
171995 93.786 -76.854 54.117
171996 93.804 -76.851 54.102
171997 93.823 -76.847 54.088
171998 93.851 -76.855 54.076
171999 93.880 -76.862 54.068
172000 93.909 -76.871 54.062
172001 93.926 -76.860 54.056
172002 93.943 -76.850 54.050
172003 93.960 -76.840 54.040
172004 93.977 -76.829 54.030
172005 93.994 -76.818 54.021
172006 94.011 -76.811 54.011
172007 94.028 -76.807 53.998
172008 94.045 -76.803 53.984
172009 94.052 -76.804 53.969
172010 94.056 -76.806 53.951
172011 94.063 -76.806 53.936
172012 94.076 -76.804 53.925
172013 94.088 -76.802 53.917
172014 94.101 -76.801 53.909
172015 94.113 -76.799 53.902
172016 94.132 -76.795 53.895
172017 94.152 -76.791 53.886
172018 94.167 -76.788 53.874
172019 94.155 -76.792 53.860
172020 94.143 -76.795 53.841
172021 94.145 -76.796 53.819
172022 94.151 -76.798 53.797
172023 94.157 -76.799 53.775
172024 94.184 -76.791 53.757
172025 94.213 -76.782 53.744
172026 94.242 -76.773 53.730
172027 94.271 -76.763 53.716
172028 94.300 -76.754 53.700
172029 94.329 -76.745 53.682
172030 94.358 -76.736 53.660
172031 94.376 -76.733 53.633
172032 94.394 -76.730 53.602
172033 94.412 -76.727 53.573
172034 94.429 -76.724 53.546
172035 94.447 -76.721 53.521
172036 94.464 -76.718 53.501
172037 94.481 -76.715 53.480
172038 94.498 -76.712 53.460
172039 94.505 -76.712 53.441
172040 94.508 -76.712 53.420
172041 94.509 -76.714 53.397
172042 94.510 -76.718 53.374
172043 94.510 -76.721 53.348
172044 94.511 -76.724 53.318
172045 94.557 -76.718 53.287
172046 94.602 -76.712 53.256
172047 94.652 -76.696 53.224
172048 94.703 -76.676 53.192
172049 94.743 -76.665 53.159
172050 94.779 -76.655 53.125
172051 94.815 -76.646 53.085
172052 94.833 -76.638 53.042
172053 94.845 -76.632 52.999
172054 94.857 -76.626 52.958
172055 94.869 -76.620 52.922
172056 94.881 -76.614 52.888
172057 94.893 -76.608 52.852
172058 94.904 -76.602 52.816
172059 94.916 -76.596 52.780
172060 94.928 -76.589 52.744
172061 94.936 -76.583 52.707
172062 94.942 -76.577 52.667
172063 94.956 -76.578 52.626
172064 94.981 -76.588 52.582
172065 95.006 -76.598 52.533
172066 95.032 -76.608 52.484
172067 95.057 -76.617 52.433
172068 95.083 -76.627 52.381
172069 95.108 -76.636 52.334
172070 95.133 -76.646 52.290
172071 95.158 -76.655 52.248
172072 95.180 -76.671 52.205
172073 95.190 -76.684 52.162
172074 95.199 -76.697 52.117
172075 95.209 -76.711 52.074
172076 95.218 -76.724 52.032
172077 95.228 -76.738 51.988
172078 95.238 -76.751 51.940
172079 95.247 -76.765 51.887
172080 95.257 -76.778 51.833
172081 95.266 -76.792 51.783
172082 95.280 -76.805 51.733
172083 95.295 -76.818 51.685
172084 95.310 -76.831 51.640
172085 95.324 -76.843 51.595
172086 95.339 -76.856 51.549
172087 95.356 -76.861 51.502
172088 95.375 -76.862 51.454
172089 95.394 -76.864 51.406
172090 95.412 -76.865 51.358
172091 95.431 -76.867 51.311
172092 95.450 -76.868 51.263
172093 95.468 -76.870 51.215
172094 95.487 -76.871 51.169
172095 95.506 -76.877 51.122
172096 95.522 -76.895 51.076
172097 95.537 -76.914 51.030
172098 95.553 -76.932 50.984
172099 95.568 -76.951 50.940
172100 95.584 -76.969 50.898
172101 95.599 -76.988 50.857
172102 95.615 -77.007 50.817
172103 95.630 -77.025 50.777
172104 95.646 -77.044 50.738
172105 95.662 -77.063 50.699
172106 95.677 -77.072 50.660
172107 95.691 -77.078 50.620
172108 95.706 -77.083 50.580
172109 95.721 -77.088 50.541
172110 95.736 -77.094 50.501
172111 95.751 -77.098 50.461
172112 95.764 -77.108 50.422
172113 95.777 -77.118 50.384
172114 95.790 -77.128 50.344
172115 95.803 -77.138 50.303
172116 95.818 -77.140 50.261
172117 95.833 -77.138 50.219
172118 95.848 -77.136 50.174
172119 95.863 -77.134 50.129
172120 95.878 -77.132 50.084
172121 95.893 -77.131 50.036
172122 95.908 -77.128 49.983
172123 95.921 -77.127 49.922
172124 95.932 -77.138 49.848
172125 95.940 -77.149 49.774
172126 95.948 -77.160 49.705
172127 95.980 -77.174 49.640
172128 96.002 -77.183 49.581
172129 96.015 -77.195 49.527
172130 96.029 -77.212 49.476
172131 96.043 -77.229 49.429
172132 96.056 -77.245 49.387
172133 96.071 -77.262 49.347
172134 96.087 -77.274 49.311
172135 96.104 -77.284 49.279
172136 96.121 -77.294 49.247
172137 96.138 -77.304 49.216
172138 96.155 -77.315 49.187
172139 96.172 -77.325 49.159
172140 96.190 -77.335 49.136
172141 96.209 -77.340 49.114
172142 96.230 -77.342 49.092
172143 96.250 -77.344 49.071
172144 96.270 -77.346 49.050
172145 96.291 -77.348 49.029
172146 96.312 -77.351 49.006
172147 96.332 -77.353 48.989
172148 96.353 -77.352 48.972
172149 96.374 -77.350 48.955
172150 96.394 -77.349 48.938
172151 96.409 -77.366 48.924
172152 96.425 -77.382 48.909
172153 96.441 -77.398 48.894
172154 96.456 -77.414 48.879
172155 96.471 -77.430 48.866
172156 96.487 -77.447 48.853
172157 96.503 -77.463 48.840
172158 96.518 -77.479 48.825
172159 96.534 -77.495 48.811
172160 96.551 -77.510 48.798
172161 96.568 -77.525 48.785
172162 96.584 -77.540 48.775
172163 96.600 -77.555 48.767
172164 96.616 -77.573 48.759
172165 96.639 -77.582 48.749
172166 96.662 -77.591 48.735
172167 96.686 -77.600 48.721
172168 96.709 -77.610 48.707
172169 96.732 -77.619 48.692
172170 96.756 -77.628 48.677
172171 96.779 -77.638 48.661
172172 96.801 -77.639 48.644
172173 96.822 -77.634 48.628
172174 96.843 -77.630 48.613
172175 96.867 -77.634 48.598
172176 96.897 -77.652 48.580
172177 96.927 -77.671 48.564
172178 96.957 -77.689 48.547
172179 96.994 -77.703 48.530
172180 97.030 -77.718 48.515
172181 97.038 -77.699 48.500
172182 97.044 -77.676 48.484
172183 97.050 -77.654 48.466
172184 97.056 -77.631 48.447
172185 97.062 -77.609 48.427
172186 97.071 -77.598 48.407
172187 97.080 -77.586 48.385
172188 97.089 -77.575 48.361
172189 97.098 -77.564 48.337
172190 97.107 -77.552 48.313
172191 97.116 -77.539 48.289
172192 97.123 -77.524 48.265
172193 97.145 -77.515 48.242
172194 97.180 -77.516 48.220
172195 97.203 -77.524 48.196
172196 97.226 -77.533 48.174
172197 97.248 -77.540 48.152
172198 97.271 -77.548 48.131
172199 97.294 -77.557 48.109
172200 97.316 -77.564 48.087
172201 97.338 -77.573 48.065
172202 97.361 -77.581 48.041
172203 97.384 -77.589 48.016
172204 97.407 -77.598 47.991
172205 97.430 -77.608 47.965
172206 97.453 -77.617 47.940
172207 97.477 -77.627 47.915
172208 97.500 -77.636 47.890
172209 97.524 -77.646 47.864
172210 97.547 -77.656 47.838
172211 97.571 -77.666 47.812
172212 97.590 -77.674 47.786
172213 97.594 -77.668 47.759
172214 97.598 -77.663 47.733
172215 97.610 -77.659 47.706
172216 97.643 -77.668 47.680
172217 97.680 -77.680 47.652
172218 97.717 -77.692 47.624
172219 97.754 -77.704 47.595
172220 97.792 -77.716 47.566
172221 97.829 -77.728 47.537
172222 97.866 -77.740 47.508
172223 97.903 -77.752 47.478
172224 97.930 -77.752 47.450
172225 97.947 -77.742 47.422
172226 97.938 -77.729 47.395
172227 97.899 -77.713 47.369
172228 97.861 -77.697 47.343
172229 97.831 -77.676 47.315
172230 97.856 -77.621 47.286
172231 97.881 -77.567 47.258
172232 97.906 -77.512 47.229
172233 97.946 -77.506 47.200
172234 97.987 -77.503 47.172
172235 98.016 -77.530 47.143
172236 98.051 -77.559 47.117
172237 98.085 -77.588 47.091
172238 98.120 -77.617 47.065
172239 98.154 -77.646 47.041
172240 98.189 -77.676 47.015
172241 98.223 -77.705 46.988
172242 98.253 -77.714 46.963
172243 98.265 -77.718 46.935
172244 98.267 -77.722 46.907
172245 98.268 -77.727 46.878
172246 98.270 -77.731 46.850
172247 98.272 -77.736 46.824
172248 98.274 -77.741 46.798
172249 98.282 -77.746 46.772
172250 98.292 -77.751 46.750
172251 98.301 -77.756 46.726
172252 98.310 -77.761 46.703
172253 98.331 -77.766 46.680
172254 98.352 -77.770 46.658
172255 98.389 -77.765 46.636
172256 98.423 -77.761 46.615
172257 98.445 -77.763 46.594
172258 98.468 -77.766 46.572
172259 98.491 -77.768 46.549
172260 98.514 -77.769 46.527
172261 98.538 -77.768 46.504
172262 98.563 -77.768 46.481
172263 98.587 -77.766 46.459
172264 98.596 -77.783 46.438
172265 98.604 -77.799 46.416
172266 98.619 -77.811 46.396
172267 98.647 -77.795 46.376
172268 98.650 -77.797 46.354
172269 98.652 -77.799 46.332
172270 98.653 -77.800 46.312
172271 98.655 -77.802 46.291
172272 98.658 -77.802 46.269
172273 98.661 -77.802 46.249
172274 98.665 -77.802 46.229
172275 98.668 -77.803 46.208
172276 98.671 -77.803 46.188
172277 98.674 -77.804 46.168
172278 98.677 -77.804 46.149
172279 98.681 -77.804 46.131
172280 98.693 -77.808 46.112
172281 98.711 -77.816 46.093
172282 98.740 -77.824 46.074
172283 98.782 -77.832 46.057
172284 98.824 -77.840 46.040
172285 98.866 -77.848 46.022
172286 98.908 -77.855 46.005
172287 98.950 -77.864 45.989
172288 98.983 -77.872 45.972
172289 99.007 -77.881 45.955
172290 99.030 -77.890 45.939
172291 99.054 -77.899 45.925
172292 99.078 -77.908 45.911
172293 99.102 -77.917 45.897
172294 99.126 -77.926 45.882
172295 99.149 -77.935 45.868
172296 99.173 -77.944 45.855
172297 99.197 -77.953 45.843
172298 99.219 -77.962 45.833
172299 99.239 -77.972 45.824
172300 99.263 -77.981 45.815
172301 99.302 -77.986 45.804
172302 99.337 -77.992 45.793
172303 99.342 -78.006 45.782
172304 99.348 -78.019 45.771
172305 99.359 -78.028 45.762
172306 99.373 -78.035 45.754
172307 99.387 -78.043 45.747
172308 99.402 -78.050 45.740
172309 99.425 -78.055 45.732
172310 99.469 -78.058 45.722
172311 99.513 -78.061 45.711
172312 99.536 -78.077 45.700
172313 99.556 -78.091 45.688
172314 99.573 -78.103 45.675
172315 99.588 -78.118 45.661
172316 99.603 -78.131 45.648
172317 99.617 -78.146 45.634
172318 99.646 -78.161 45.620
172319 99.677 -78.177 45.605
172320 99.707 -78.192 45.591
172321 99.737 -78.208 45.575
172322 99.768 -78.223 45.560
172323 99.826 -78.252 45.546
172324 99.877 -78.279 45.530
172325 99.897 -78.292 45.514
172326 99.915 -78.305 45.498
172327 99.934 -78.319 45.481
172328 99.952 -78.332 45.464
172329 99.971 -78.345 45.448
172330 99.990 -78.359 45.432
172331 100.018 -78.367 45.415
172332 100.050 -78.373 45.400
172333 100.069 -78.382 45.385
172334 100.074 -78.395 45.371
172335 100.093 -78.402 45.356
172336 100.123 -78.407 45.341
172337 100.153 -78.412 45.325
172338 100.182 -78.416 45.310
172339 100.206 -78.420 45.295
172340 100.230 -78.423 45.279
172341 100.254 -78.426 45.264
172342 100.278 -78.429 45.249
172343 100.303 -78.432 45.233
172344 100.329 -78.435 45.218
172345 100.366 -78.445 45.202
172346 100.404 -78.454 45.187
172347 100.441 -78.463 45.171
172348 100.478 -78.473 45.156
172349 100.515 -78.482 45.141
172350 100.553 -78.491 45.128
172351 100.611 -78.498 45.114
172352 100.669 -78.506 45.099
172353 100.727 -78.514 45.084
172354 100.786 -78.521 45.068
172355 100.844 -78.529 45.052
172356 100.902 -78.536 45.035
172357 100.957 -78.543 45.019
172358 100.939 -78.561 45.004
172359 100.918 -78.578 44.988
172360 100.898 -78.595 44.972
172361 100.879 -78.588 44.956
172362 100.924 -78.593 44.941
172363 100.957 -78.594 44.926
172364 100.990 -78.594 44.913
172365 101.024 -78.595 44.898
172366 101.057 -78.595 44.883
172367 101.088 -78.604 44.869
172368 101.116 -78.588 44.852
172369 101.148 -78.579 44.837
172370 101.201 -78.593 44.821
172371 101.256 -78.608 44.805
172372 101.289 -78.622 44.790
172373 101.310 -78.636 44.774
172374 101.326 -78.655 44.758
172375 101.336 -78.681 44.742
172376 101.346 -78.707 44.727
172377 101.357 -78.733 44.712
172378 101.374 -78.756 44.698
172379 101.400 -78.778 44.683
172380 101.425 -78.800 44.668
172381 101.451 -78.821 44.653
172382 101.477 -78.843 44.639
172383 101.502 -78.865 44.623
172384 101.528 -78.886 44.609
172385 101.568 -78.906 44.595
172386 101.610 -78.925 44.581
172387 101.653 -78.945 44.568
172388 101.685 -78.959 44.555
172389 101.702 -78.965 44.542
172390 101.718 -78.970 44.529
172391 101.735 -78.976 44.517
172392 101.752 -78.981 44.504
172393 101.769 -78.987 44.491
172394 101.785 -78.993 44.479
172395 101.802 -78.998 44.467
172396 101.819 -79.004 44.456
172397 101.836 -79.009 44.444
172398 101.852 -79.015 44.432
172399 101.870 -79.021 44.420
172400 101.886 -79.026 44.408
172401 101.903 -79.032 44.396
172629 92.659 -77.148 54.112
172630 92.683 -77.159 54.126
172631 92.706 -77.170 54.138
172632 92.722 -77.182 54.150
172633 92.730 -77.194 54.160
172634 92.737 -77.207 54.168
172635 92.744 -77.219 54.177
172636 92.751 -77.232 54.187
172637 92.758 -77.244 54.194
172638 92.765 -77.257 54.200
172639 92.779 -77.258 54.205
172640 92.799 -77.250 54.210
172641 92.820 -77.241 54.213
172642 92.840 -77.232 54.217
172643 92.860 -77.223 54.222
172644 92.881 -77.215 54.226
172645 92.901 -77.206 54.230
172646 92.922 -77.197 54.234
172647 92.942 -77.188 54.238
172648 92.962 -77.179 54.242
172649 92.983 -77.171 54.245
172650 93.003 -77.162 54.249
172651 93.023 -77.153 54.253
172652 93.044 -77.144 54.257
172653 93.064 -77.135 54.262
172654 93.084 -77.126 54.266
172655 93.103 -77.116 54.271
172656 93.123 -77.106 54.277
172657 93.142 -77.095 54.283
172658 93.162 -77.085 54.290
172659 93.180 -77.072 54.299
172660 93.196 -77.061 54.303
172661 93.200 -77.054 54.304
172662 93.204 -77.048 54.301
172663 93.209 -77.042 54.291
172664 93.213 -77.036 54.283
172665 93.217 -77.030 54.276
172666 93.226 -77.020 54.272
172667 93.236 -77.010 54.271
172668 93.264 -77.003 54.271
172669 93.299 -76.999 54.271
172670 93.335 -76.994 54.270
172671 93.372 -76.989 54.270
172672 93.404 -76.982 54.268
172673 93.419 -76.975 54.266
172674 93.439 -76.969 54.264
172675 93.459 -76.963 54.262
172676 93.479 -76.957 54.261
172677 93.499 -76.952 54.257
172678 93.521 -76.951 54.251
172679 93.544 -76.951 54.247
172680 93.562 -76.947 54.246
172681 93.573 -76.937 54.249
172682 93.583 -76.927 54.249
172683 93.596 -76.922 54.247
172684 93.609 -76.917 54.243
172685 93.622 -76.912 54.239
172686 93.634 -76.906 54.232
172687 93.647 -76.901 54.224
172688 93.660 -76.897 54.216
172689 93.682 -76.897 54.209
172690 93.705 -76.898 54.201
172691 93.716 -76.886 54.192
172692 93.724 -76.870 54.183
172693 93.731 -76.853 54.173
172694 93.750 -76.850 54.161
172695 93.768 -76.847 54.148
172696 93.786 -76.843 54.133
172697 93.805 -76.840 54.117
172698 93.823 -76.837 54.102
172699 93.845 -76.837 54.088
172700 93.873 -76.845 54.078
172701 93.903 -76.853 54.069
172702 93.921 -76.845 54.062
172703 93.938 -76.834 54.055
172704 93.955 -76.824 54.045
172705 93.972 -76.814 54.036
172706 93.989 -76.805 54.029
172707 94.006 -76.801 54.021
172708 94.023 -76.797 54.012
172709 94.040 -76.794 54.002
172710 94.054 -76.791 53.991
172711 94.058 -76.793 53.975
172712 94.061 -76.796 53.962
172713 94.065 -76.797 53.952
172714 94.075 -76.796 53.943
172715 94.087 -76.795 53.936
172716 94.103 -76.793 53.929
172717 94.122 -76.789 53.921
172718 94.142 -76.784 53.910
172719 94.166 -76.779 53.898
172720 94.159 -76.781 53.884
172721 94.147 -76.785 53.867
172722 94.135 -76.789 53.846
172723 94.134 -76.790 53.825
172724 94.146 -76.789 53.806
172725 94.175 -76.780 53.790
172726 94.204 -76.771 53.776
172727 94.234 -76.761 53.763
172728 94.262 -76.752 53.751
172729 94.291 -76.743 53.733
172730 94.321 -76.734 53.714
172731 94.350 -76.724 53.689
172732 94.371 -76.718 53.659
172733 94.389 -76.715 53.626
172734 94.406 -76.712 53.593
172735 94.423 -76.710 53.563
172736 94.440 -76.707 53.535
172737 94.458 -76.704 53.511
172738 94.475 -76.701 53.488
172739 94.492 -76.698 53.466
172740 94.495 -76.698 53.446
172741 94.498 -76.699 53.426
172742 94.501 -76.699 53.405
172743 94.503 -76.699 53.384
172744 94.503 -76.703 53.358
172745 94.514 -76.705 53.330
172746 94.559 -76.698 53.299
172747 94.605 -76.692 53.267
172748 94.653 -76.679 53.234
172749 94.703 -76.661 53.202
172750 94.739 -76.651 53.170
172751 94.775 -76.642 53.138
172752 94.812 -76.632 53.099
172753 94.830 -76.625 53.058
172754 94.841 -76.619 53.015
172755 94.853 -76.613 52.977
172756 94.865 -76.607 52.943
172757 94.877 -76.601 52.911
172758 94.889 -76.595 52.878
172759 94.901 -76.589 52.841
172760 94.911 -76.583 52.804
172761 94.916 -76.576 52.768
172762 94.922 -76.570 52.729
172763 94.938 -76.572 52.690
172764 94.964 -76.582 52.650
172765 94.989 -76.592 52.607
172766 95.014 -76.602 52.559
172767 95.039 -76.611 52.511
172768 95.064 -76.621 52.460
172769 95.090 -76.630 52.406
172770 95.115 -76.640 52.355
172771 95.140 -76.649 52.306
172772 95.147 -76.651 52.260
172773 95.165 -76.655 52.214
172774 95.185 -76.666 52.168
172775 95.194 -76.680 52.122
172776 95.204 -76.693 52.076
172777 95.213 -76.706 52.033
172778 95.223 -76.720 51.988
172779 95.233 -76.734 51.941
172780 95.242 -76.746 51.891
172781 95.251 -76.760 51.841
172782 95.261 -76.774 51.793
172783 95.276 -76.787 51.746
172784 95.290 -76.799 51.700
172785 95.305 -76.812 51.656
172786 95.320 -76.825 51.613
172787 95.335 -76.838 51.568
172788 95.349 -76.851 51.522
172789 95.368 -76.853 51.475
172790 95.386 -76.855 51.428
172791 95.405 -76.856 51.381
172792 95.423 -76.857 51.333
172793 95.442 -76.859 51.285
172794 95.461 -76.860 51.236
172795 95.479 -76.862 51.189
172796 95.494 -76.865 51.141
172797 95.520 -76.883 51.095
172798 95.536 -76.901 51.049
172799 95.551 -76.920 51.003
172800 95.567 -76.938 50.960
172801 95.583 -76.957 50.918
172802 95.598 -76.976 50.877
172803 95.614 -76.994 50.837
172804 95.629 -77.013 50.796
172805 95.644 -77.031 50.756
172806 95.660 -77.050 50.717
172807 95.675 -77.059 50.679
172808 95.690 -77.064 50.639
172809 95.705 -77.069 50.599
172810 95.720 -77.074 50.561
172811 95.735 -77.079 50.520
172812 95.749 -77.085 50.481
172813 95.764 -77.090 50.442
172814 95.779 -77.095 50.405
172815 95.794 -77.100 50.365
172816 95.808 -77.107 50.324
172817 95.821 -77.116 50.281
172818 95.836 -77.114 50.237
172819 95.851 -77.113 50.191
172820 95.869 -77.113 50.144
172821 95.887 -77.114 50.098
172822 95.906 -77.116 50.051
172823 95.924 -77.117 49.997
172824 95.931 -77.117 49.936
172825 95.930 -77.119 49.862
172826 95.938 -77.129 49.787
172827 95.969 -77.144 49.717
172828 96.016 -77.160 49.652
172829 96.027 -77.166 49.592
172830 96.040 -77.173 49.538
172831 96.053 -77.190 49.487
172832 96.067 -77.207 49.441
172833 96.082 -77.223 49.397
172834 96.095 -77.240 49.358
172835 96.109 -77.257 49.322
172836 96.123 -77.274 49.289
172837 96.137 -77.291 49.258
172838 96.151 -77.307 49.227
172839 96.167 -77.319 49.198
172840 96.184 -77.329 49.170
172841 96.201 -77.339 49.146
172842 96.222 -77.342 49.123
172843 96.242 -77.345 49.101
172844 96.263 -77.345 49.080
172845 96.284 -77.344 49.059
172846 96.305 -77.342 49.039
172847 96.325 -77.340 49.018
172848 96.346 -77.338 49.000
172849 96.367 -77.337 48.984
172850 96.388 -77.335 48.968
172851 96.408 -77.333 48.953
172852 96.424 -77.348 48.938
172853 96.440 -77.364 48.923
172854 96.455 -77.381 48.908
172855 96.471 -77.397 48.893
172856 96.486 -77.413 48.879
172857 96.502 -77.429 48.866
172858 96.517 -77.445 48.852
172859 96.533 -77.462 48.838
172860 96.548 -77.478 48.823
172861 96.564 -77.494 48.808
172862 96.579 -77.511 48.796
172863 96.595 -77.526 48.784
172864 96.611 -77.542 48.775
172865 96.631 -77.552 48.766
172866 96.653 -77.564 48.753
172867 96.676 -77.574 48.741
172868 96.699 -77.584 48.728
172869 96.722 -77.593 48.716
172870 96.745 -77.603 48.700
172871 96.768 -77.612 48.685
172872 96.792 -77.621 48.670
172873 96.815 -77.631 48.654
172874 96.837 -77.631 48.638
172875 96.858 -77.627 48.622
172876 96.885 -77.639 48.605
172877 96.915 -77.657 48.589
172878 96.950 -77.672 48.572
172879 96.986 -77.687 48.555
172880 97.023 -77.702 48.538
172881 97.043 -77.696 48.522
172882 97.049 -77.674 48.507
172883 97.055 -77.652 48.492
172884 97.063 -77.636 48.476
172885 97.072 -77.624 48.457
172886 97.081 -77.613 48.437
172887 97.090 -77.602 48.416
172888 97.099 -77.590 48.393
172889 97.108 -77.579 48.371
172890 97.117 -77.567 48.348
172891 97.126 -77.555 48.324
172892 97.135 -77.542 48.302
172893 97.142 -77.526 48.277
172894 97.150 -77.511 48.252
172895 97.190 -77.510 48.229
172896 97.213 -77.517 48.208
172897 97.236 -77.526 48.185
172898 97.258 -77.534 48.163
172899 97.281 -77.542 48.141
172900 97.304 -77.550 48.119
172901 97.326 -77.558 48.098
172902 97.349 -77.567 48.076
172903 97.372 -77.574 48.051
172904 97.394 -77.583 48.026
172905 97.417 -77.591 48.001
172906 97.439 -77.599 47.976
172907 97.462 -77.607 47.950
172908 97.485 -77.615 47.925
172909 97.508 -77.625 47.899
172910 97.531 -77.635 47.872
172911 97.555 -77.644 47.847
172912 97.578 -77.654 47.820
172913 97.601 -77.664 47.794
172914 97.610 -77.662 47.767
172915 97.614 -77.657 47.741
172916 97.634 -77.656 47.715
172917 97.656 -77.657 47.688
172918 97.693 -77.669 47.660
172919 97.730 -77.680 47.631
172920 97.768 -77.692 47.603
172921 97.805 -77.705 47.574
172922 97.843 -77.717 47.545
172923 97.880 -77.729 47.516
172924 97.917 -77.741 47.486
172925 97.939 -77.736 47.458
172926 97.957 -77.727 47.431
172927 97.967 -77.716 47.404
172928 97.928 -77.700 47.377
172929 97.890 -77.684 47.350
172930 97.858 -77.664 47.322
172931 97.883 -77.609 47.293
172932 97.908 -77.555 47.264
172933 97.933 -77.501 47.235
172934 97.959 -77.450 47.206
172935 98.000 -77.447 47.178
172936 98.032 -77.469 47.150
172937 98.067 -77.498 47.124
172938 98.101 -77.527 47.098
172939 98.136 -77.557 47.071
172940 98.170 -77.586 47.047
172941 98.205 -77.615 47.020
172942 98.238 -77.645 46.994
172943 98.270 -77.682 46.968
172944 98.286 -77.705 46.940
172945 98.288 -77.711 46.912
172946 98.291 -77.715 46.884
172947 98.293 -77.720 46.856
172948 98.295 -77.725 46.829
172949 98.297 -77.730 46.802
172950 98.303 -77.734 46.776
172951 98.312 -77.740 46.753
172952 98.322 -77.745 46.730
172953 98.332 -77.750 46.707
172954 98.341 -77.755 46.684
172955 98.373 -77.752 46.660
172956 98.411 -77.746 46.639
172957 98.438 -77.746 46.618
172958 98.461 -77.749 46.597
172959 98.484 -77.750 46.576
172960 98.508 -77.749 46.554
172961 98.532 -77.749 46.531
172962 98.557 -77.748 46.509
172963 98.581 -77.746 46.486
172964 98.601 -77.751 46.464
172965 98.610 -77.767 46.443
172966 98.621 -77.782 46.421
172967 98.641 -77.787 46.401
172968 98.667 -77.775 46.380
172969 98.673 -77.783 46.359
172970 98.676 -77.786 46.337
172971 98.678 -77.787 46.317
172972 98.679 -77.789 46.296
172973 98.681 -77.790 46.274
172974 98.683 -77.792 46.253
172975 98.685 -77.793 46.233
172976 98.686 -77.795 46.212
172977 98.688 -77.796 46.191
172978 98.690 -77.798 46.171
172979 98.693 -77.798 46.152
172980 98.696 -77.798 46.133
172981 98.699 -77.799 46.114
172982 98.713 -77.804 46.095
172983 98.752 -77.812 46.076
172984 98.794 -77.820 46.057
172985 98.835 -77.827 46.040
172986 98.875 -77.835 46.023
172987 98.916 -77.843 46.006
172988 98.953 -77.850 45.990
172989 98.984 -77.855 45.973
172990 99.008 -77.864 45.956
172991 99.028 -77.876 45.940
172992 99.049 -77.888 45.926
172993 99.069 -77.899 45.912
172994 99.089 -77.912 45.898
172995 99.113 -77.921 45.884
172996 99.137 -77.930 45.870
172997 99.161 -77.938 45.858
172998 99.185 -77.947 45.846
172999 99.207 -77.956 45.837
173000 99.232 -77.965 45.828
173001 99.271 -77.971 45.818
173002 99.310 -77.975 45.807
173003 99.341 -77.983 45.795
173004 99.346 -77.996 45.784
173005 99.351 -78.010 45.774
173006 99.357 -78.025 45.766
173007 99.369 -78.033 45.758
173008 99.383 -78.040 45.751
173009 99.397 -78.047 45.744
173010 99.418 -78.053 45.736
173011 99.462 -78.056 45.726
173012 99.506 -78.059 45.716
173013 99.545 -78.065 45.704
173014 99.561 -78.075 45.691
173015 99.580 -78.089 45.679
173016 99.603 -78.106 45.666
173017 99.614 -78.119 45.653
173018 99.624 -78.131 45.640
173019 99.638 -78.144 45.626
173020 99.652 -78.159 45.612
173021 99.674 -78.173 45.597
173022 99.705 -78.189 45.582
173023 99.761 -78.217 45.566
173024 99.820 -78.247 45.550
173025 99.878 -78.275 45.535
173026 99.900 -78.291 45.519
173027 99.919 -78.304 45.503
173028 99.937 -78.317 45.487
173029 99.957 -78.330 45.470
173030 99.976 -78.343 45.453
173031 100.008 -78.349 45.437
173032 100.041 -78.356 45.421
173033 100.073 -78.362 45.405
173034 100.082 -78.373 45.391
173035 100.087 -78.385 45.377
173036 100.092 -78.396 45.362
173037 100.123 -78.401 45.347
173038 100.152 -78.406 45.332
173039 100.182 -78.411 45.317
173040 100.212 -78.416 45.302
173041 100.241 -78.420 45.287
173042 100.265 -78.423 45.271
173043 100.289 -78.426 45.256
173044 100.314 -78.429 45.239
173045 100.338 -78.432 45.224
173046 100.366 -78.438 45.209
173047 100.404 -78.447 45.192
173048 100.441 -78.456 45.177
173049 100.478 -78.466 45.162
173050 100.515 -78.475 45.149
173051 100.562 -78.483 45.135
173052 100.620 -78.491 45.120
173053 100.679 -78.498 45.105
173054 100.737 -78.506 45.090
173055 100.795 -78.514 45.074
173056 100.853 -78.520 45.058
173057 100.891 -78.528 45.042
173058 100.923 -78.535 45.026
173059 100.945 -78.545 45.009
173060 100.924 -78.563 44.993
173061 100.904 -78.579 44.977
173062 100.888 -78.542 44.962
173063 100.915 -78.524 44.948
173064 100.943 -78.523 44.933
173065 100.961 -78.523 44.919
173066 101.005 -78.549 44.905
173067 101.050 -78.576 44.889
173068 101.088 -78.586 44.874
173069 101.117 -78.570 44.858
173070 101.164 -78.568 44.842
173071 101.223 -78.584 44.827
173072 101.266 -78.599 44.811
173073 101.287 -78.613 44.795
173074 101.308 -78.627 44.779
173075 101.328 -78.642 44.763
173076 101.339 -78.668 44.747
173077 101.349 -78.694 44.733
173078 101.360 -78.719 44.718
173079 101.386 -78.741 44.703
173080 101.412 -78.762 44.688
173081 101.438 -78.784 44.673
173082 101.463 -78.806 44.659
173083 101.489 -78.827 44.645
173084 101.515 -78.849 44.629
173085 101.540 -78.871 44.614
173086 101.567 -78.893 44.600
173087 101.610 -78.912 44.586
173088 101.652 -78.931 44.572
173089 101.686 -78.946 44.559
173090 101.702 -78.951 44.546
173091 101.720 -78.957 44.533
173092 101.736 -78.962 44.521
173093 101.753 -78.968 44.508
173094 101.770 -78.974 44.496
173095 101.786 -78.980 44.484
173096 101.803 -78.985 44.472
173097 101.820 -78.990 44.460
173098 101.837 -78.996 44.448
173099 101.853 -79.002 44.436
173100 101.870 -79.008 44.424
173101 101.887 -79.013 44.412
173102 101.904 -79.018 44.400
173331 92.670 -77.130 54.104
173332 92.693 -77.141 54.117
173333 92.716 -77.152 54.128
173334 92.723 -77.165 54.137
173335 92.730 -77.177 54.146
173336 92.737 -77.190 54.155
173337 92.744 -77.203 54.164
173338 92.751 -77.215 54.172
173339 92.758 -77.228 54.178
173340 92.769 -77.235 54.183
173341 92.789 -77.226 54.188
173342 92.809 -77.217 54.192
173343 92.830 -77.208 54.197
173344 92.850 -77.200 54.203
173345 92.870 -77.191 54.206
173346 92.890 -77.182 54.211
173347 92.911 -77.173 54.216
173348 92.931 -77.165 54.221
173349 92.952 -77.156 54.224
173350 92.972 -77.147 54.229
173351 92.992 -77.138 54.234
173352 93.013 -77.129 54.238
173353 93.033 -77.121 54.243
173354 93.053 -77.112 54.247
173355 93.074 -77.103 54.251
173356 93.094 -77.094 54.256
173357 93.115 -77.086 54.262
173358 93.134 -77.075 54.268
173359 93.153 -77.066 54.275
173360 93.173 -77.055 54.286
173361 93.190 -77.042 54.293
173362 93.194 -77.036 54.295
173363 93.199 -77.030 54.290
173364 93.203 -77.024 54.282
173365 93.207 -77.018 54.276
173366 93.212 -77.011 54.272
173367 93.216 -77.005 54.271
173368 93.223 -76.997 54.271
173369 93.250 -76.990 54.273
173370 93.288 -76.985 54.274
173371 93.327 -76.979 54.274
173372 93.367 -76.973 54.274
173373 93.398 -76.966 54.273
173374 93.404 -76.956 54.271
173375 93.418 -76.948 54.269
173376 93.438 -76.942 54.268
173377 93.458 -76.937 54.267
173378 93.479 -76.931 54.263
173379 93.501 -76.931 54.258
173380 93.524 -76.932 54.255
173381 93.547 -76.932 54.254
173382 93.569 -76.932 54.257
173383 93.580 -76.922 54.258
173384 93.590 -76.911 54.258
173385 93.601 -76.900 54.256
173386 93.612 -76.892 54.251
173387 93.624 -76.887 54.246
173388 93.637 -76.881 54.240
173389 93.654 -76.878 54.233
173390 93.676 -76.879 54.225
173391 93.699 -76.880 54.217
173392 93.715 -76.874 54.207
173393 93.722 -76.857 54.200
173394 93.731 -76.843 54.190
173395 93.750 -76.840 54.178
173396 93.768 -76.837 54.165
173397 93.787 -76.834 54.150
173398 93.805 -76.831 54.134
173399 93.824 -76.827 54.117
173400 93.842 -76.824 54.101
173401 93.867 -76.828 54.089
173402 93.896 -76.836 54.078
173403 93.916 -76.830 54.069
173404 93.933 -76.819 54.061
173405 93.950 -76.809 54.051
173406 93.967 -76.799 54.042
173407 93.984 -76.796 54.036
173408 94.001 -76.792 54.031
173409 94.018 -76.788 54.025
173410 94.035 -76.784 54.018
173411 94.052 -76.781 54.011
173412 94.060 -76.781 53.998
173413 94.064 -76.783 53.987
173414 94.067 -76.785 53.978
173415 94.071 -76.787 53.969
173416 94.074 -76.789 53.962
173417 94.096 -76.784 53.956
173418 94.122 -76.777 53.948
173419 94.149 -76.771 53.935
173420 94.161 -76.767 53.922
173421 94.161 -76.768 53.908
173422 94.150 -76.772 53.892
173423 94.138 -76.777 53.873
173424 94.127 -76.781 53.854
173425 94.139 -76.777 53.836
173426 94.169 -76.768 53.819
173427 94.198 -76.759 53.805
173428 94.228 -76.749 53.792
173429 94.258 -76.740 53.780
173430 94.287 -76.730 53.763
173431 94.317 -76.721 53.742
173432 94.347 -76.712 53.716
173433 94.365 -76.705 53.685
173434 94.383 -76.701 53.650
173435 94.400 -76.698 53.616
173436 94.417 -76.695 53.583
173437 94.434 -76.692 53.552
173438 94.452 -76.689 53.524
173439 94.469 -76.686 53.497
173440 94.483 -76.683 53.472
173441 94.485 -76.684 53.450
173442 94.488 -76.684 53.429
173443 94.491 -76.685 53.409
173444 94.493 -76.685 53.389
173445 94.496 -76.686 53.365
173446 94.517 -76.684 53.337
173447 94.562 -76.678 53.306
173448 94.608 -76.672 53.274
173449 94.654 -76.662 53.240
173450 94.699 -76.647 53.209
173451 94.736 -76.638 53.179
173452 94.771 -76.628 53.148
173453 94.808 -76.619 53.112
173454 94.826 -76.611 53.073
173455 94.838 -76.605 53.033
173456 94.849 -76.599 52.996
173457 94.861 -76.593 52.964
173458 94.873 -76.587 52.933
173459 94.885 -76.581 52.904
173460 94.891 -76.575 52.866
173461 94.897 -76.569 52.829
173462 94.902 -76.563 52.792
173463 94.920 -76.566 52.752
173464 94.946 -76.576 52.713
173465 94.971 -76.586 52.673
173466 94.996 -76.596 52.631
173467 95.021 -76.605 52.585
173468 95.046 -76.614 52.538
173469 95.071 -76.624 52.487
173470 95.096 -76.633 52.433
173471 95.103 -76.635 52.379
173472 95.111 -76.636 52.327
173473 95.118 -76.638 52.276
173474 95.135 -76.642 52.227
173475 95.173 -76.649 52.179
173476 95.189 -76.661 52.130
173477 95.199 -76.675 52.083
173478 95.208 -76.689 52.037
173479 95.217 -76.702 51.991
173480 95.227 -76.715 51.945
173481 95.237 -76.729 51.897
173482 95.246 -76.743 51.851
173483 95.257 -76.755 51.804
173484 95.271 -76.768 51.759
173485 95.286 -76.781 51.715
173486 95.301 -76.794 51.673
173487 95.316 -76.807 51.631
173488 95.330 -76.820 51.588
173489 95.345 -76.833 51.544
173490 95.360 -76.843 51.498
173491 95.379 -76.845 51.451
173492 95.398 -76.846 51.404
173493 95.416 -76.848 51.355
173494 95.435 -76.850 51.306
173495 95.454 -76.851 51.258
173496 95.471 -76.853 51.209
173497 95.482 -76.855 51.161
173498 95.508 -76.872 51.114
173499 95.533 -76.889 51.068
173500 95.550 -76.908 51.024
173501 95.565 -76.926 50.981
173502 95.581 -76.945 50.939
173503 95.597 -76.963 50.898
173504 95.612 -76.982 50.857
173505 95.628 -77.001 50.817
173506 95.643 -77.019 50.776
173507 95.659 -77.038 50.737
173508 95.674 -77.045 50.699
173509 95.688 -77.050 50.660
173510 95.703 -77.055 50.620
173511 95.718 -77.060 50.581
173512 95.733 -77.066 50.541
173513 95.748 -77.071 50.501
173514 95.763 -77.076 50.463
173515 95.777 -77.081 50.426
173516 95.792 -77.087 50.386
173517 95.807 -77.092 50.344
173518 95.825 -77.100 50.301
173519 95.857 -77.107 50.255
173520 95.876 -77.109 50.208
173521 95.894 -77.110 50.160
173522 95.913 -77.112 50.113
173523 95.932 -77.113 50.063
173524 95.950 -77.115 50.008
173525 95.948 -77.113 49.946
173526 95.941 -77.110 49.871
173527 95.958 -77.113 49.797
173528 96.026 -77.135 49.727
173529 96.041 -77.142 49.663
173530 96.053 -77.148 49.603
173531 96.065 -77.154 49.549
173532 96.079 -77.169 49.499
173533 96.092 -77.185 49.451
173534 96.106 -77.202 49.408
173535 96.120 -77.219 49.370
173536 96.134 -77.235 49.334
173537 96.148 -77.252 49.301
173538 96.162 -77.269 49.270
173539 96.176 -77.285 49.239
173540 96.190 -77.302 49.209
173541 96.204 -77.319 49.182
173542 96.215 -77.334 49.158
173543 96.236 -77.334 49.135
173544 96.256 -77.332 49.113
173545 96.277 -77.330 49.092
173546 96.298 -77.329 49.071
173547 96.319 -77.327 49.051
173548 96.339 -77.326 49.031
173549 96.360 -77.323 49.013
173550 96.381 -77.322 48.998
173551 96.401 -77.320 48.983
173552 96.422 -77.318 48.969
173553 96.439 -77.331 48.953
173554 96.454 -77.347 48.938
173555 96.470 -77.363 48.923
173556 96.485 -77.379 48.908
173557 96.501 -77.395 48.894
173558 96.516 -77.412 48.880
173559 96.532 -77.428 48.866
173560 96.547 -77.445 48.851
173561 96.563 -77.461 48.835
173562 96.579 -77.476 48.820
173563 96.594 -77.493 48.806
173564 96.610 -77.508 48.794
173565 96.629 -77.518 48.783
173566 96.648 -77.528 48.773
173567 96.670 -77.539 48.760
173568 96.693 -77.550 48.748
173569 96.715 -77.561 48.736
173570 96.738 -77.572 48.724
173571 96.760 -77.583 48.709
173572 96.783 -77.594 48.695
173573 96.805 -77.605 48.680
173574 96.828 -77.614 48.664
173575 96.852 -77.624 48.649
173576 96.874 -77.625 48.632
173577 96.906 -77.642 48.614
173578 96.943 -77.656 48.598
173579 96.979 -77.671 48.580
173580 97.016 -77.686 48.563
173581 97.048 -77.695 48.546
173582 97.054 -77.674 48.529
173583 97.064 -77.662 48.512
173584 97.073 -77.651 48.497
173585 97.082 -77.639 48.483
173586 97.091 -77.628 48.465
173587 97.100 -77.617 48.445
173588 97.109 -77.605 48.423
173589 97.118 -77.593 48.401
173590 97.127 -77.582 48.378
173591 97.137 -77.570 48.356
173592 97.146 -77.559 48.333
173593 97.153 -77.544 48.310
173594 97.160 -77.529 48.286
173595 97.168 -77.514 48.261
173596 97.195 -77.507 48.238
173597 97.224 -77.511 48.217
173598 97.246 -77.520 48.195
173599 97.269 -77.527 48.172
173600 97.291 -77.536 48.150
173601 97.314 -77.544 48.128
173602 97.336 -77.551 48.107
173603 97.359 -77.560 48.086
173604 97.382 -77.568 48.062
173605 97.404 -77.576 48.037
173606 97.427 -77.584 48.012
173607 97.450 -77.592 47.987
173608 97.472 -77.600 47.961
173609 97.495 -77.608 47.936
173610 97.517 -77.617 47.909
173611 97.540 -77.624 47.881
173612 97.563 -77.633 47.855
173613 97.586 -77.642 47.829
173614 97.609 -77.652 47.802
173615 97.626 -77.657 47.775
173616 97.639 -77.654 47.749
173617 97.658 -77.653 47.722
173618 97.678 -77.652 47.695
173619 97.706 -77.658 47.666
173620 97.743 -77.669 47.638
173621 97.781 -77.681 47.610
173622 97.818 -77.693 47.581
173623 97.856 -77.705 47.552
173624 97.893 -77.718 47.523
173625 97.930 -77.730 47.494
173626 97.949 -77.721 47.466
173627 97.966 -77.711 47.439
173628 97.984 -77.702 47.413
173629 97.957 -77.686 47.386
173630 97.919 -77.671 47.357
173631 97.884 -77.652 47.329
173632 97.910 -77.598 47.300
173633 97.934 -77.543 47.270
173634 97.959 -77.489 47.241
173635 97.985 -77.434 47.212
173636 98.013 -77.391 47.184
173637 98.048 -77.408 47.157
173638 98.082 -77.438 47.131
173639 98.117 -77.467 47.105
173640 98.149 -77.501 47.079
173641 98.181 -77.538 47.053
173642 98.212 -77.574 47.026
173643 98.244 -77.611 46.999
173644 98.275 -77.646 46.972
173645 98.307 -77.683 46.945
173646 98.307 -77.697 46.917
173647 98.308 -77.702 46.889
173648 98.311 -77.708 46.861
173649 98.313 -77.712 46.834
173650 98.316 -77.718 46.807
173651 98.323 -77.722 46.781
173652 98.364 -77.733 46.757
173653 98.375 -77.739 46.734
173654 98.370 -77.743 46.711
173655 98.360 -77.742 46.689
173656 98.395 -77.734 46.665
173657 98.432 -77.728 46.643
173658 98.454 -77.731 46.622
173659 98.479 -77.730 46.601
173660 98.503 -77.730 46.580
173661 98.527 -77.729 46.559
173662 98.551 -77.728 46.537
173663 98.576 -77.727 46.514
173664 98.600 -77.726 46.491
173665 98.615 -77.735 46.469
173666 98.624 -77.752 46.448
173667 98.643 -77.758 46.427
173668 98.664 -77.765 46.407
173669 98.687 -77.756 46.386
173670 98.693 -77.763 46.365
173671 98.699 -77.771 46.344
173672 98.703 -77.774 46.323
173673 98.704 -77.776 46.302
173674 98.706 -77.777 46.281
173675 98.708 -77.779 46.260
173676 98.709 -77.780 46.238
173677 98.711 -77.782 46.217
173678 98.712 -77.783 46.195
173679 98.714 -77.785 46.175
173680 98.716 -77.787 46.156
173681 98.717 -77.788 46.136
173682 98.719 -77.790 46.117
173683 98.720 -77.792 46.098
173684 98.760 -77.799 46.079
173685 98.800 -77.807 46.058
173686 98.840 -77.815 46.041
173687 98.880 -77.822 46.024
173688 98.920 -77.830 46.007
173689 98.951 -77.834 45.991
173690 98.982 -77.839 45.974
173691 99.012 -77.843 45.958
173692 99.035 -77.854 45.941
173693 99.056 -77.865 45.927
173694 99.076 -77.877 45.913
173695 99.096 -77.889 45.899
173696 99.117 -77.901 45.886
173697 99.138 -77.913 45.872
173698 99.158 -77.924 45.861
173699 99.179 -77.937 45.850
173700 99.202 -77.949 45.841
173701 99.240 -77.955 45.831
173702 99.279 -77.960 45.821
173703 99.318 -77.965 45.809
173704 99.345 -77.974 45.798
173705 99.350 -77.987 45.787
173706 99.355 -78.001 45.778
173707 99.361 -78.015 45.770
173708 99.366 -78.029 45.763
173709 99.378 -78.037 45.757
173710 99.393 -78.044 45.749
173711 99.411 -78.050 45.740
173712 99.455 -78.053 45.730
173713 99.498 -78.056 45.720
173714 99.534 -78.059 45.708
173715 99.563 -78.062 45.696
173716 99.591 -78.084 45.684
173717 99.617 -78.106 45.672
173718 99.628 -78.118 45.658
173719 99.639 -78.130 45.646
173720 99.650 -78.142 45.632
173721 99.661 -78.154 45.618
173722 99.673 -78.166 45.603
173723 99.705 -78.178 45.588
173724 99.755 -78.212 45.572
173725 99.814 -78.241 45.556
173726 99.872 -78.270 45.541
173727 99.904 -78.288 45.525
173728 99.923 -78.302 45.509
173729 99.941 -78.315 45.493
173730 99.967 -78.326 45.476
173731 99.999 -78.332 45.459
173732 100.031 -78.338 45.443
173733 100.063 -78.344 45.428
173734 100.089 -78.351 45.412
173735 100.094 -78.363 45.397
173736 100.100 -78.376 45.383
173737 100.105 -78.387 45.368
173738 100.122 -78.395 45.354
173739 100.151 -78.401 45.339
173740 100.182 -78.405 45.324
173741 100.211 -78.410 45.309
173742 100.241 -78.415 45.293
173743 100.271 -78.420 45.278
173744 100.300 -78.425 45.262
173745 100.325 -78.428 45.245
173746 100.349 -78.431 45.230
173747 100.373 -78.434 45.215
173748 100.404 -78.440 45.198
173749 100.441 -78.449 45.183
173750 100.478 -78.459 45.169
173751 100.516 -78.468 45.155
173752 100.571 -78.476 45.141
173753 100.630 -78.483 45.126
173754 100.688 -78.491 45.111
173755 100.747 -78.498 45.095
173756 100.785 -78.505 45.079
173757 100.819 -78.513 45.064
173758 100.851 -78.520 45.048
173759 100.884 -78.528 45.031
173760 100.917 -78.535 45.014
173761 100.930 -78.547 44.998
173762 100.912 -78.542 44.983
173763 100.896 -78.496 44.968
173764 100.885 -78.470 44.954
173765 100.922 -78.490 44.940
173766 100.966 -78.517 44.925
173767 101.010 -78.544 44.910
173768 101.055 -78.570 44.894
173769 101.088 -78.567 44.879
173770 101.121 -78.551 44.864
173771 101.179 -78.557 44.848
173772 101.238 -78.573 44.832
173773 101.263 -78.589 44.816
173774 101.285 -78.604 44.799
173775 101.307 -78.619 44.783
173776 101.327 -78.633 44.768
173777 101.341 -78.655 44.753
173778 101.351 -78.680 44.739
173779 101.373 -78.703 44.723
173780 101.398 -78.725 44.708
173781 101.424 -78.747 44.694
173782 101.450 -78.768 44.678
173783 101.475 -78.790 44.664
173784 101.501 -78.812 44.650
173785 101.527 -78.833 44.635
173786 101.552 -78.855 44.620
173787 101.578 -78.877 44.605
173788 101.610 -78.898 44.590
173789 101.652 -78.917 44.577
173790 101.687 -78.933 44.563
173791 101.703 -78.938 44.550
173792 101.720 -78.943 44.537
173793 101.737 -78.949 44.524
173794 101.754 -78.955 44.512
173795 101.770 -78.961 44.500
173796 101.787 -78.966 44.488
173797 101.804 -78.972 44.476
173798 101.821 -78.977 44.464
173799 101.837 -78.983 44.451
173800 101.854 -78.989 44.439
173801 101.871 -78.994 44.427
173802 101.888 -79.000 44.415
173803 101.904 -79.006 44.404
174033 92.681 -77.112 54.094
174034 92.704 -77.123 54.105
174035 92.716 -77.135 54.114
174036 92.723 -77.147 54.122
174037 92.730 -77.160 54.131
174038 92.737 -77.172 54.140
174039 92.744 -77.185 54.148
174040 92.752 -77.198 54.154
174041 92.759 -77.210 54.159
174042 92.779 -77.202 54.165
174043 92.799 -77.194 54.170
174044 92.819 -77.185 54.175
174045 92.840 -77.176 54.181
174046 92.860 -77.167 54.186
174047 92.880 -77.159 54.191
174048 92.901 -77.150 54.196
174049 92.921 -77.141 54.202
174050 92.941 -77.132 54.206
174051 92.961 -77.123 54.211
174052 92.982 -77.115 54.216
174053 93.002 -77.106 54.221
174054 93.022 -77.097 54.226
174055 93.043 -77.088 54.231
174056 93.063 -77.079 54.237
174057 93.083 -77.071 54.242
174058 93.104 -77.062 54.247
174059 93.124 -77.053 54.254
174060 93.145 -77.044 54.262
174061 93.165 -77.035 54.272
174062 93.184 -77.024 54.279
174063 93.189 -77.018 54.282
174064 93.193 -77.011 54.277
174065 93.197 -77.005 54.272
174066 93.202 -76.999 54.269
174067 93.209 -76.993 54.269
174068 93.217 -76.987 54.271
174069 93.225 -76.981 54.272
174070 93.243 -76.975 54.274
174071 93.283 -76.969 54.277
174072 93.323 -76.963 54.277
174073 93.362 -76.958 54.276
174074 93.391 -76.950 54.275
174075 93.398 -76.940 54.274
174076 93.404 -76.929 54.273
174077 93.417 -76.922 54.273
174078 93.437 -76.915 54.272
174079 93.459 -76.912 54.268
174080 93.482 -76.912 54.264
174081 93.504 -76.912 54.262
174082 93.527 -76.913 54.263
174083 93.550 -76.913 54.264
174084 93.573 -76.913 54.267
174085 93.586 -76.906 54.268
174086 93.597 -76.895 54.268
174087 93.607 -76.884 54.264
174088 93.618 -76.874 54.260
174089 93.628 -76.863 54.254
174090 93.648 -76.861 54.247
174091 93.670 -76.862 54.239
174092 93.693 -76.862 54.232
174093 93.713 -76.861 54.223
174094 93.721 -76.845 54.215
174095 93.732 -76.833 54.204
174096 93.750 -76.830 54.192
174097 93.769 -76.826 54.179
174098 93.787 -76.823 54.164
174099 93.805 -76.820 54.149
174100 93.824 -76.817 54.132
174101 93.843 -76.814 54.116
174102 93.861 -76.810 54.104
174103 93.890 -76.818 54.093
174104 93.911 -76.815 54.082
174105 93.928 -76.804 54.072
174106 93.945 -76.794 54.060
174107 93.962 -76.790 54.051
174108 93.979 -76.787 54.044
174109 93.996 -76.783 54.040
174110 94.013 -76.779 54.036
174111 94.030 -76.775 54.032
174112 94.047 -76.771 54.029
174113 94.063 -76.768 54.019
174114 94.066 -76.771 54.011
174115 94.070 -76.773 54.003
174116 94.073 -76.774 53.995
174117 94.077 -76.777 53.989
174118 94.104 -76.770 53.983
174119 94.130 -76.763 53.975
174120 94.147 -76.759 53.961
174121 94.156 -76.755 53.947
174122 94.164 -76.753 53.933
174123 94.152 -76.758 53.918
174124 94.141 -76.762 53.900
174125 94.129 -76.767 53.883
174126 94.140 -76.764 53.866
174127 94.167 -76.755 53.848
174128 94.196 -76.746 53.833
174129 94.226 -76.737 53.820
174130 94.256 -76.727 53.806
174131 94.285 -76.718 53.790
174132 94.315 -76.708 53.769
174133 94.344 -76.699 53.742
174134 94.360 -76.693 53.710
174135 94.377 -76.687 53.674
174136 94.394 -76.683 53.639
174137 94.411 -76.680 53.606
174138 94.428 -76.677 53.573
174139 94.445 -76.674 53.542
174140 94.463 -76.671 53.512
174141 94.473 -76.670 53.483
174142 94.476 -76.670 53.457
174143 94.478 -76.671 53.434
174144 94.481 -76.671 53.412
174145 94.484 -76.671 53.391
174146 94.488 -76.670 53.367
174147 94.521 -76.665 53.340
174148 94.565 -76.658 53.310
174149 94.610 -76.652 53.280
174150 94.656 -76.646 53.246
174151 94.695 -76.634 53.214
174152 94.732 -76.624 53.183
174153 94.768 -76.615 53.153
174154 94.804 -76.605 53.118
174155 94.822 -76.599 53.081
174156 94.834 -76.592 53.043
174157 94.845 -76.586 53.010
174158 94.857 -76.580 52.979
174159 94.866 -76.574 52.950
174160 94.872 -76.567 52.922
174161 94.878 -76.561 52.888
174162 94.883 -76.555 52.854
174163 94.902 -76.560 52.819
174164 94.927 -76.570 52.781
174165 94.953 -76.580 52.741
174166 94.978 -76.589 52.700
174167 95.003 -76.599 52.656
174168 95.028 -76.608 52.610
174169 95.051 -76.617 52.563
174170 95.058 -76.619 52.514
174171 95.066 -76.621 52.462
174172 95.073 -76.622 52.406
174173 95.081 -76.624 52.352
174174 95.089 -76.625 52.298
174175 95.106 -76.629 52.246
174176 95.143 -76.636 52.195
174177 95.181 -76.644 52.146
174178 95.193 -76.657 52.097
174179 95.203 -76.671 52.050
174180 95.212 -76.684 52.002
174181 95.222 -76.697 51.955
174182 95.232 -76.711 51.908
174183 95.241 -76.724 51.862
174184 95.252 -76.737 51.816
174185 95.267 -76.750 51.772
174186 95.282 -76.763 51.730
174187 95.297 -76.776 51.691
174188 95.311 -76.789 51.651
174189 95.326 -76.802 51.610
174190 95.341 -76.815 51.566
174191 95.355 -76.828 51.522
174192 95.372 -76.836 51.475
174193 95.390 -76.837 51.427
174194 95.409 -76.838 51.378
174195 95.427 -76.840 51.328
174196 95.446 -76.841 51.279
174197 95.460 -76.843 51.230
174198 95.471 -76.846 51.181
174199 95.495 -76.862 51.134
174200 95.521 -76.878 51.089
174201 95.547 -76.896 51.045
174202 95.564 -76.914 51.003
174203 95.580 -76.932 50.962
174204 95.595 -76.951 50.920
174205 95.611 -76.970 50.879
174206 95.626 -76.988 50.839
174207 95.642 -77.007 50.799
174208 95.658 -77.025 50.759
174209 95.672 -77.031 50.721
174210 95.687 -77.036 50.682
174211 95.702 -77.041 50.642
174212 95.716 -77.047 50.603
174213 95.732 -77.052 50.563
174214 95.746 -77.057 50.523
174215 95.761 -77.063 50.483
174216 95.776 -77.068 50.446
174217 95.792 -77.075 50.406
174218 95.811 -77.083 50.364
174219 95.831 -77.091 50.320
174220 95.867 -77.099 50.274
174221 95.902 -77.106 50.224
174222 95.920 -77.108 50.174
174223 95.939 -77.110 50.126
174224 95.958 -77.111 50.074
174225 95.971 -77.112 50.016
174226 95.964 -77.109 49.952
174227 95.979 -77.109 49.879
174228 96.024 -77.113 49.805
174229 96.055 -77.119 49.737
174230 96.066 -77.125 49.674
174231 96.079 -77.131 49.615
174232 96.091 -77.137 49.561
174233 96.103 -77.147 49.511
174234 96.117 -77.163 49.463
174235 96.131 -77.180 49.420
174236 96.145 -77.197 49.382
174237 96.159 -77.213 49.347
174238 96.173 -77.230 49.313
174239 96.187 -77.247 49.282
174240 96.201 -77.264 49.252
174241 96.210 -77.282 49.222
174242 96.214 -77.303 49.194
174243 96.231 -77.311 49.170
174244 96.250 -77.316 49.147
174245 96.270 -77.317 49.125
174246 96.291 -77.316 49.105
174247 96.312 -77.313 49.085
174248 96.332 -77.312 49.065
174249 96.353 -77.310 49.046
174250 96.374 -77.308 49.029
174251 96.395 -77.307 49.014
174252 96.415 -77.305 49.000
174253 96.436 -77.303 48.987
174254 96.454 -77.313 48.970
174255 96.469 -77.329 48.954
174256 96.484 -77.346 48.938
174257 96.500 -77.362 48.924
174258 96.516 -77.378 48.910
174259 96.531 -77.395 48.895
174260 96.546 -77.410 48.880
174261 96.562 -77.427 48.865
174262 96.578 -77.443 48.848
174263 96.593 -77.460 48.832
174264 96.610 -77.473 48.818
174265 96.628 -77.484 48.805
174266 96.647 -77.494 48.793
174267 96.666 -77.504 48.781
174268 96.687 -77.515 48.769
174269 96.710 -77.526 48.757
174270 96.732 -77.537 48.745
174271 96.755 -77.548 48.733
174272 96.778 -77.559 48.719
174273 96.800 -77.570 48.705
174274 96.822 -77.581 48.691
174275 96.845 -77.592 48.675
174276 96.866 -77.607 48.659
174277 96.888 -77.614 48.642
174278 96.924 -77.629 48.624
174279 96.960 -77.643 48.606
174280 96.997 -77.658 48.588
174281 97.033 -77.673 48.570
174282 97.064 -77.689 48.553
174283 97.074 -77.677 48.535
174284 97.083 -77.666 48.517
174285 97.092 -77.655 48.501
174286 97.101 -77.643 48.488
174287 97.110 -77.632 48.470
174288 97.119 -77.620 48.451
174289 97.129 -77.608 48.429
174290 97.138 -77.597 48.408
174291 97.147 -77.586 48.385
174292 97.156 -77.574 48.362
174293 97.164 -77.561 48.340
174294 97.172 -77.546 48.317
174295 97.179 -77.531 48.293
174296 97.186 -77.516 48.268
174297 97.199 -77.504 48.245
174298 97.234 -77.505 48.224
174299 97.256 -77.513 48.202
174300 97.279 -77.521 48.180
174301 97.302 -77.530 48.157
174302 97.324 -77.537 48.137
174303 97.347 -77.545 48.116
174304 97.369 -77.554 48.096
174305 97.392 -77.561 48.073
174306 97.415 -77.570 48.050
174307 97.437 -77.578 48.025
174308 97.460 -77.586 47.999
174309 97.483 -77.594 47.973
174310 97.505 -77.602 47.947
174311 97.528 -77.610 47.919
174312 97.550 -77.618 47.891
174313 97.573 -77.627 47.864
174314 97.595 -77.634 47.838
174315 97.618 -77.642 47.811
174316 97.641 -77.651 47.783
174317 97.663 -77.650 47.756
174318 97.682 -77.649 47.728
174319 97.702 -77.649 47.701
174320 97.722 -77.648 47.672
174321 97.757 -77.658 47.643
174322 97.794 -77.670 47.615
174323 97.831 -77.682 47.588
174324 97.869 -77.694 47.558
174325 97.906 -77.706 47.530
174326 97.941 -77.716 47.502
174327 97.958 -77.706 47.474
174328 97.976 -77.696 47.447
174329 97.994 -77.686 47.421
174330 97.987 -77.674 47.393
174331 97.948 -77.658 47.365
174332 97.911 -77.641 47.335
174333 97.936 -77.586 47.305
174334 97.961 -77.532 47.276
174335 97.986 -77.477 47.247
174336 98.011 -77.423 47.219
174337 98.036 -77.368 47.192
174338 98.060 -77.357 47.164
174339 98.091 -77.393 47.139
174340 98.123 -77.429 47.112
174341 98.155 -77.466 47.086
174342 98.186 -77.502 47.059
174343 98.218 -77.539 47.033
174344 98.249 -77.575 47.005
174345 98.281 -77.611 46.978
174346 98.312 -77.648 46.950
174347 98.334 -77.677 46.922
174348 98.329 -77.687 46.894
174349 98.329 -77.695 46.866
174350 98.331 -77.699 46.840
174351 98.334 -77.705 46.813
174352 98.336 -77.709 46.787
174353 98.368 -77.717 46.763
174354 98.386 -77.728 46.739
174355 98.361 -77.732 46.716
174356 98.384 -77.725 46.693
174357 98.419 -77.717 46.671
174358 98.448 -77.711 46.649
174359 98.473 -77.711 46.627
174360 98.497 -77.710 46.605
174361 98.521 -77.709 46.585
174362 98.546 -77.708 46.564
174363 98.570 -77.708 46.542
174364 98.594 -77.706 46.519
174365 98.619 -77.705 46.496
174366 98.629 -77.720 46.475
174367 98.645 -77.730 46.455
174368 98.665 -77.736 46.434
174369 98.686 -77.741 46.414
174370 98.706 -77.736 46.393
174371 98.712 -77.744 46.372
174372 98.719 -77.752 46.351
174373 98.725 -77.758 46.330
174374 98.729 -77.763 46.309
174375 98.731 -77.765 46.289
174376 98.732 -77.766 46.267
174377 98.734 -77.768 46.245
174378 98.735 -77.769 46.223
174379 98.737 -77.771 46.201
174380 98.739 -77.772 46.181
174381 98.740 -77.773 46.161
174382 98.742 -77.774 46.140
174383 98.743 -77.777 46.120
174384 98.744 -77.780 46.101
174385 98.765 -77.786 46.081
174386 98.805 -77.793 46.060
174387 98.845 -77.801 46.042
174388 98.885 -77.808 46.025
174389 98.919 -77.815 46.009
174390 98.950 -77.819 45.992
174391 98.980 -77.824 45.976
174392 99.010 -77.828 45.960
174393 99.041 -77.833 45.944
174394 99.063 -77.843 45.930
174395 99.083 -77.855 45.916
174396 99.104 -77.867 45.902
174397 99.124 -77.879 45.888
174398 99.145 -77.890 45.875
174399 99.165 -77.902 45.863
174400 99.186 -77.915 45.853
174401 99.214 -77.930 45.843
174402 99.249 -77.944 45.833
174403 99.287 -77.949 45.823
174404 99.326 -77.955 45.812
174405 99.349 -77.965 45.802
174406 99.354 -77.978 45.792
174407 99.359 -77.992 45.784
174408 99.364 -78.006 45.776
174409 99.370 -78.019 45.769
174410 99.375 -78.034 45.762
174411 99.387 -78.041 45.754
174412 99.404 -78.048 45.745
174413 99.448 -78.051 45.735
174414 99.487 -78.054 45.725
174415 99.520 -78.056 45.713
174416 99.563 -78.062 45.701
174417 99.593 -78.084 45.690
174418 99.618 -78.105 45.677
174419 99.629 -78.117 45.664
174420 99.641 -78.129 45.652
174421 99.652 -78.141 45.638
174422 99.663 -78.153 45.624
174423 99.687 -78.150 45.610
174424 99.724 -78.162 45.595
174425 99.768 -78.198 45.579
174426 99.811 -78.234 45.564
174427 99.866 -78.265 45.548
174428 99.908 -78.287 45.531
174429 99.926 -78.300 45.515
174430 99.957 -78.307 45.498
174431 99.989 -78.313 45.483
174432 100.021 -78.320 45.467
174433 100.054 -78.326 45.451
174434 100.086 -78.332 45.436
174435 100.102 -78.342 45.421
174436 100.108 -78.354 45.405
174437 100.113 -78.366 45.390
174438 100.118 -78.378 45.376
174439 100.123 -78.389 45.361
174440 100.151 -78.395 45.346
174441 100.181 -78.400 45.331
174442 100.210 -78.404 45.315
174443 100.241 -78.410 45.299
174444 100.270 -78.414 45.283
174445 100.300 -78.420 45.267
174446 100.330 -78.424 45.251
174447 100.360 -78.429 45.235
174448 100.385 -78.432 45.220
174449 100.409 -78.435 45.205
174450 100.441 -78.442 45.190
174451 100.479 -78.451 45.176
174452 100.522 -78.460 45.162
174453 100.581 -78.468 45.148
174454 100.639 -78.475 45.132
174455 100.680 -78.482 45.116
174456 100.713 -78.490 45.100
174457 100.746 -78.498 45.084
174458 100.779 -78.505 45.069
174459 100.812 -78.513 45.052
174460 100.845 -78.520 45.036
174461 100.879 -78.528 45.019
174462 100.918 -78.535 45.004
174463 100.927 -78.525 44.989
174464 100.916 -78.502 44.974
174465 100.906 -78.479 44.959
174466 100.927 -78.484 44.944
174467 100.971 -78.511 44.929
174468 101.016 -78.539 44.914
174469 101.059 -78.564 44.899
174470 101.089 -78.548 44.883
174471 101.134 -78.531 44.867
174472 101.193 -78.545 44.852
174473 101.238 -78.561 44.836
174474 101.260 -78.577 44.820
174475 101.282 -78.593 44.804
174476 101.304 -78.609 44.788
174477 101.326 -78.624 44.772
174478 101.344 -78.642 44.758
174479 101.359 -78.666 44.743
174480 101.385 -78.688 44.728
174481 101.410 -78.709 44.713
174482 101.436 -78.731 44.698
174483 101.462 -78.753 44.683
174484 101.487 -78.774 44.668
174485 101.513 -78.796 44.654
174486 101.538 -78.818 44.639
174487 101.564 -78.839 44.624
174488 101.590 -78.861 44.609
174489 101.615 -78.883 44.594
174490 101.652 -78.903 44.581
174491 101.687 -78.919 44.567
174492 101.704 -78.925 44.554
174493 101.721 -78.930 44.540
174494 101.738 -78.937 44.527
174495 101.754 -78.942 44.515
174496 101.771 -78.947 44.503
174497 101.788 -78.953 44.491
174498 101.805 -78.959 44.479
174499 101.822 -78.965 44.467
174500 101.838 -78.970 44.455
174501 101.855 -78.975 44.443
174502 101.872 -78.981 44.430
174503 101.889 -78.987 44.418
174504 101.905 -78.993 44.407
174734 92.668 -77.083 54.069
174735 92.691 -77.094 54.080
174736 92.709 -77.106 54.089
174737 92.716 -77.118 54.097
174738 92.723 -77.131 54.105
174739 92.730 -77.143 54.115
174740 92.738 -77.156 54.123
174741 92.745 -77.168 54.129
174742 92.752 -77.181 54.135
174743 92.768 -77.179 54.140
174744 92.788 -77.170 54.146
174745 92.809 -77.161 54.152
174746 92.829 -77.152 54.157
174747 92.849 -77.144 54.164
174748 92.870 -77.135 54.170
174749 92.890 -77.126 54.175
174750 92.911 -77.117 54.181
174751 92.931 -77.109 54.186
174752 92.951 -77.100 54.191
174753 92.972 -77.091 54.197
174754 92.992 -77.082 54.203
174755 93.012 -77.073 54.209
174756 93.032 -77.065 54.215
174757 93.053 -77.056 54.221
174758 93.073 -77.047 54.226
174759 93.093 -77.038 54.232
174760 93.114 -77.030 54.239
174761 93.134 -77.021 54.247
174762 93.154 -77.012 54.256
174763 93.175 -77.003 54.263
174764 93.194 -77.000 54.267
174765 93.207 -76.994 54.264
174766 93.215 -76.989 54.263
174767 93.223 -76.983 54.264
174768 93.231 -76.977 54.267
174769 93.240 -76.972 54.270
174770 93.248 -76.965 54.272
174771 93.256 -76.959 54.274
174772 93.278 -76.953 54.278
174773 93.318 -76.948 54.276
174774 93.357 -76.942 54.275
174775 93.385 -76.934 54.274
174776 93.392 -76.924 54.272
174777 93.398 -76.913 54.273
174778 93.405 -76.903 54.274
174779 93.417 -76.894 54.275
174780 93.439 -76.893 54.273
174781 93.462 -76.893 54.271
174782 93.485 -76.893 54.270
174783 93.507 -76.893 54.271
174784 93.530 -76.894 54.271
174785 93.553 -76.894 54.275
174786 93.576 -76.894 54.278
174787 93.593 -76.890 54.278
174788 93.604 -76.879 54.276
174789 93.616 -76.867 54.272
174790 93.628 -76.854 54.266
174791 93.643 -76.845 54.259
174792 93.664 -76.844 54.252
174793 93.687 -76.845 54.245
174794 93.709 -76.846 54.238
174795 93.719 -76.832 54.229
174796 93.732 -76.822 54.218
174797 93.751 -76.819 54.206
174798 93.769 -76.816 54.193
174799 93.787 -76.813 54.179
174800 93.806 -76.809 54.163
174801 93.825 -76.806 54.147
174802 93.843 -76.803 54.131
174803 93.861 -76.800 54.119
174804 93.884 -76.801 54.108
174805 93.906 -76.799 54.098
174806 93.923 -76.788 54.087
174807 93.940 -76.784 54.075
174808 93.957 -76.781 54.064
174809 93.974 -76.777 54.056
174810 93.991 -76.774 54.050
174811 94.008 -76.770 54.046
174812 94.025 -76.766 54.044
174813 94.042 -76.762 54.044
174814 94.059 -76.759 54.037
174815 94.069 -76.758 54.032
174816 94.072 -76.760 54.027
174817 94.076 -76.762 54.021
174818 94.085 -76.762 54.014
174819 94.112 -76.755 54.007
174820 94.133 -76.749 54.000
174821 94.143 -76.746 53.987
174822 94.152 -76.743 53.973
174823 94.161 -76.741 53.959
174824 94.154 -76.743 53.944
174825 94.143 -76.748 53.928
174826 94.131 -76.752 53.911
174827 94.145 -76.749 53.894
174828 94.166 -76.743 53.877
174829 94.194 -76.734 53.861
174830 94.223 -76.724 53.846
174831 94.253 -76.715 53.831
174832 94.283 -76.706 53.814
174833 94.313 -76.696 53.793
174834 94.339 -76.688 53.766
174835 94.355 -76.681 53.734
174836 94.371 -76.674 53.699
174837 94.388 -76.668 53.664
174838 94.405 -76.665 53.630
174839 94.422 -76.662 53.596
174840 94.439 -76.659 53.563
174841 94.457 -76.657 53.530
174842 94.463 -76.656 53.497
174843 94.466 -76.656 53.468
174844 94.469 -76.656 53.441
174845 94.471 -76.657 53.416
174846 94.477 -76.655 53.393
174847 94.487 -76.650 53.368
174848 94.528 -76.644 53.341
174849 94.572 -76.638 53.312
174850 94.616 -76.632 53.283
174851 94.658 -76.627 53.250
174852 94.691 -76.621 53.218
174853 94.728 -76.611 53.185
174854 94.764 -76.602 53.155
174855 94.801 -76.592 53.122
174856 94.818 -76.586 53.088
174857 94.830 -76.579 53.053
174858 94.841 -76.573 53.024
174859 94.847 -76.567 52.995
174860 94.852 -76.561 52.966
174861 94.858 -76.555 52.938
174862 94.863 -76.549 52.907
174863 94.884 -76.554 52.875
174864 94.909 -76.564 52.842
174865 94.934 -76.574 52.806
174866 94.960 -76.583 52.767
174867 94.985 -76.592 52.726
174868 95.006 -76.601 52.683
174869 95.014 -76.602 52.635
174870 95.022 -76.604 52.588
174871 95.029 -76.605 52.539
174872 95.037 -76.608 52.489
174873 95.044 -76.609 52.435
174874 95.052 -76.611 52.380
174875 95.059 -76.612 52.326
174876 95.077 -76.616 52.273
174877 95.114 -76.624 52.220
174878 95.151 -76.631 52.168
174879 95.188 -76.639 52.117
174880 95.198 -76.652 52.069
174881 95.207 -76.666 52.021
174882 95.217 -76.680 51.972
174883 95.227 -76.693 51.924
174884 95.236 -76.706 51.878
174885 95.248 -76.719 51.833
174886 95.263 -76.732 51.789
174887 95.277 -76.745 51.747
174888 95.292 -76.758 51.708
174889 95.307 -76.771 51.669
174890 95.322 -76.784 51.629
174891 95.336 -76.796 51.586
174892 95.351 -76.809 51.540
174893 95.366 -76.822 51.492
174894 95.383 -76.828 51.444
174895 95.402 -76.829 51.395
174896 95.420 -76.831 51.346
174897 95.439 -76.832 51.297
174898 95.450 -76.834 51.251
174899 95.460 -76.837 51.204
174900 95.483 -76.851 51.157
174901 95.509 -76.868 51.111
174902 95.535 -76.884 51.068
174903 95.561 -76.902 51.026
174904 95.579 -76.920 50.985
174905 95.594 -76.939 50.944
174906 95.609 -76.957 50.903
174907 95.625 -76.976 50.863
174908 95.640 -76.994 50.823
174909 95.656 -77.012 50.782
174910 95.671 -77.017 50.743
174911 95.686 -77.022 50.704
174912 95.700 -77.028 50.665
174913 95.716 -77.033 50.625
174914 95.730 -77.038 50.585
174915 95.745 -77.044 50.545
174916 95.760 -77.049 50.504
174917 95.778 -77.057 50.465
174918 95.797 -77.066 50.425
174919 95.815 -77.074 50.383
174920 95.839 -77.082 50.339
174921 95.874 -77.090 50.291
174922 95.909 -77.097 50.240
174923 95.945 -77.105 50.188
174924 95.965 -77.107 50.137
174925 95.983 -77.109 50.082
174926 95.988 -77.108 50.021
174927 96.005 -77.108 49.955
174928 96.044 -77.110 49.884
174929 96.060 -77.110 49.813
174930 96.076 -77.108 49.747
174931 96.092 -77.108 49.686
174932 96.104 -77.114 49.627
174933 96.116 -77.119 49.574
174934 96.128 -77.125 49.523
174935 96.142 -77.142 49.475
174936 96.156 -77.159 49.433
174937 96.170 -77.175 49.395
174938 96.184 -77.192 49.360
174939 96.198 -77.209 49.327
174940 96.210 -77.226 49.296
174941 96.214 -77.247 49.265
174942 96.218 -77.267 49.236
174943 96.226 -77.283 49.207
174944 96.246 -77.288 49.183
174945 96.265 -77.293 49.160
174946 96.285 -77.298 49.139
174947 96.305 -77.301 49.119
174948 96.326 -77.298 49.100
174949 96.346 -77.297 49.082
174950 96.367 -77.295 49.065
174951 96.388 -77.294 49.050
174952 96.409 -77.291 49.035
174953 96.430 -77.290 49.019
174954 96.450 -77.288 49.005
174955 96.468 -77.296 48.987
174956 96.484 -77.312 48.970
174957 96.499 -77.329 48.954
174958 96.515 -77.345 48.941
174959 96.530 -77.361 48.926
174960 96.546 -77.377 48.911
174961 96.561 -77.393 48.896
174962 96.577 -77.410 48.880
174963 96.592 -77.426 48.864
174964 96.609 -77.439 48.847
174965 96.628 -77.450 48.832
174966 96.647 -77.460 48.818
174967 96.665 -77.470 48.805
174968 96.684 -77.480 48.792
174969 96.704 -77.491 48.779
174970 96.727 -77.502 48.767
174971 96.749 -77.513 48.754
174972 96.772 -77.524 48.742
174973 96.795 -77.535 48.729
174974 96.817 -77.546 48.715
174975 96.836 -77.563 48.701
174976 96.854 -77.583 48.686
174977 96.869 -77.588 48.669
174978 96.890 -77.590 48.652
174979 96.928 -77.604 48.634
174980 96.965 -77.618 48.615
174981 97.001 -77.633 48.596
174982 97.037 -77.647 48.577
174983 97.070 -77.667 48.559
174984 97.093 -77.681 48.539
174985 97.102 -77.670 48.522
174986 97.111 -77.658 48.505
174987 97.120 -77.646 48.488
174988 97.129 -77.635 48.472
174989 97.138 -77.624 48.454
174990 97.148 -77.612 48.433
174991 97.157 -77.601 48.412
174992 97.166 -77.589 48.390
174993 97.175 -77.577 48.369
174994 97.183 -77.564 48.348
174995 97.190 -77.548 48.324
174996 97.197 -77.533 48.300
174997 97.205 -77.518 48.277
174998 97.212 -77.503 48.254
174999 97.244 -77.498 48.232
175000 97.267 -77.507 48.210
175001 97.289 -77.515 48.188
175002 97.312 -77.523 48.166
175003 97.334 -77.531 48.145
175004 97.357 -77.539 48.125
175005 97.380 -77.547 48.105
175006 97.402 -77.555 48.082
175007 97.425 -77.564 48.059
175008 97.447 -77.571 48.035
175009 97.470 -77.580 48.010
175010 97.493 -77.588 47.984
175011 97.515 -77.595 47.958
175012 97.538 -77.604 47.930
175013 97.561 -77.612 47.902
175014 97.583 -77.620 47.875
175015 97.605 -77.629 47.847
175016 97.624 -77.641 47.818
175017 97.654 -77.645 47.790
175018 97.688 -77.648 47.763
175019 97.706 -77.646 47.735
175020 97.726 -77.645 47.707
175021 97.745 -77.644 47.679
175022 97.770 -77.647 47.650
175023 97.807 -77.658 47.621
175024 97.845 -77.671 47.594
175025 97.882 -77.683 47.565
175026 97.919 -77.695 47.537
175027 97.951 -77.701 47.509
175028 97.968 -77.691 47.482
175029 97.986 -77.680 47.456
175030 98.003 -77.671 47.428
175031 98.016 -77.661 47.400
175032 97.977 -77.645 47.372
175033 97.939 -77.629 47.342
175034 97.963 -77.574 47.311
175035 97.988 -77.520 47.281
175036 98.013 -77.465 47.252
175037 98.038 -77.411 47.226
175038 98.061 -77.373 47.199
175039 98.081 -77.373 47.170
175040 98.101 -77.373 47.145
175041 98.129 -77.394 47.119
175042 98.160 -77.430 47.092
175043 98.192 -77.467 47.065
175044 98.223 -77.503 47.039
175045 98.254 -77.539 47.012
175046 98.286 -77.576 46.984
175047 98.317 -77.612 46.955
175048 98.349 -77.649 46.927
175049 98.356 -77.667 46.899
175050 98.352 -77.677 46.872
175051 98.349 -77.687 46.845
175052 98.351 -77.692 46.819
175053 98.354 -77.696 46.794
175054 98.364 -77.700 46.769
175055 98.377 -77.708 46.745
175056 98.381 -77.712 46.722
175057 98.404 -77.710 46.699
175058 98.438 -77.702 46.677
175059 98.467 -77.696 46.654
175060 98.491 -77.692 46.632
175061 98.515 -77.689 46.611
175062 98.540 -77.689 46.591
175063 98.564 -77.688 46.570
175064 98.588 -77.687 46.548
175065 98.613 -77.686 46.524
175066 98.635 -77.688 46.501
175067 98.646 -77.702 46.481
175068 98.667 -77.707 46.463
175069 98.688 -77.713 46.443
175070 98.709 -77.718 46.422
175071 98.725 -77.717 46.400
175072 98.732 -77.724 46.380
175073 98.738 -77.732 46.358
175074 98.745 -77.740 46.339
175075 98.751 -77.746 46.318
175076 98.756 -77.752 46.296
175077 98.757 -77.753 46.275
175078 98.759 -77.755 46.254
175079 98.760 -77.756 46.231
175080 98.762 -77.757 46.209
175081 98.764 -77.758 46.188
175082 98.765 -77.760 46.168
175083 98.767 -77.762 46.147
175084 98.768 -77.765 46.127
175085 98.769 -77.769 46.106
175086 98.769 -77.772 46.085
175087 98.809 -77.780 46.065
175088 98.850 -77.787 46.047
175089 98.887 -77.794 46.029
175090 98.917 -77.799 46.012
175091 98.948 -77.803 45.995
175092 98.978 -77.808 45.979
175093 99.008 -77.812 45.963
175094 99.039 -77.817 45.947
175095 99.069 -77.822 45.932
175096 99.090 -77.833 45.918
175097 99.111 -77.845 45.904
175098 99.132 -77.856 45.891
175099 99.152 -77.868 45.878
175100 99.172 -77.881 45.866
175101 99.193 -77.892 45.855
175102 99.225 -77.910 45.846
175103 99.259 -77.928 45.836
175104 99.296 -77.940 45.826
175105 99.335 -77.945 45.816
175106 99.352 -77.956 45.807
175107 99.358 -77.969 45.798
175108 99.363 -77.983 45.790
175109 99.368 -77.996 45.782
175110 99.373 -78.010 45.774
175111 99.379 -78.025 45.767
175112 99.384 -78.038 45.759
175113 99.397 -78.046 45.751
175114 99.440 -78.050 45.741
175115 99.477 -78.053 45.731
175116 99.519 -78.059 45.720
175117 99.562 -78.066 45.707
175118 99.595 -78.084 45.695
175119 99.620 -78.103 45.683
175120 99.631 -78.116 45.670
175121 99.642 -78.128 45.658
175122 99.653 -78.140 45.645
175123 99.673 -78.141 45.631
175124 99.704 -78.128 45.616
175125 99.743 -78.146 45.603
175126 99.786 -78.181 45.587
175127 99.830 -78.217 45.571
175128 99.873 -78.253 45.556
175129 99.915 -78.283 45.540
175130 99.947 -78.290 45.524
175131 99.980 -78.296 45.507
175132 100.012 -78.302 45.491
175133 100.044 -78.309 45.474
175134 100.076 -78.315 45.459
175135 100.109 -78.321 45.443
175136 100.115 -78.333 45.429
175137 100.120 -78.344 45.413
175138 100.125 -78.357 45.398
175139 100.130 -78.368 45.383
175140 100.135 -78.380 45.368
175141 100.150 -78.389 45.353
175142 100.180 -78.394 45.338
175143 100.210 -78.399 45.322
175144 100.240 -78.404 45.305
175145 100.269 -78.409 45.289
175146 100.299 -78.413 45.273
175147 100.329 -78.419 45.257
175148 100.359 -78.423 45.241
175149 100.388 -78.429 45.226
175150 100.419 -78.433 45.212
175151 100.444 -78.437 45.197
175152 100.479 -78.445 45.183
175153 100.532 -78.453 45.168
175154 100.575 -78.460 45.154
175155 100.608 -78.467 45.138
175156 100.641 -78.475 45.122
175157 100.679 -78.482 45.105
175158 100.719 -78.489 45.089
175159 100.759 -78.495 45.073
175160 100.799 -78.502 45.057
175161 100.839 -78.508 45.040
175162 100.879 -78.515 45.024
175163 100.912 -78.521 45.009
175164 100.938 -78.526 44.994
175165 100.937 -78.511 44.980
175166 100.927 -78.488 44.965
175167 100.932 -78.479 44.950
175168 100.976 -78.506 44.935
175169 101.021 -78.532 44.920
175170 101.060 -78.545 44.904
175171 101.095 -78.529 44.888
175172 101.148 -78.518 44.872
175173 101.207 -78.534 44.856
175174 101.235 -78.550 44.841
175175 101.257 -78.566 44.825
175176 101.279 -78.582 44.809
175177 101.301 -78.598 44.792
175178 101.323 -78.614 44.777
175179 101.345 -78.630 44.763
175180 101.371 -78.651 44.748
175181 101.397 -78.672 44.733
175182 101.422 -78.694 44.718
175183 101.448 -78.715 44.702
175184 101.474 -78.737 44.687
175185 101.499 -78.759 44.673
175186 101.525 -78.780 44.658
175187 101.550 -78.802 44.642
175188 101.576 -78.824 44.627
175189 101.602 -78.846 44.613
175190 101.627 -78.868 44.599
175191 101.653 -78.890 44.585
175192 101.688 -78.906 44.571
175193 101.705 -78.912 44.558
175194 101.722 -78.918 44.544
175195 101.739 -78.923 44.531
175196 101.755 -78.929 44.517
175197 101.772 -78.934 44.505
175198 101.789 -78.940 44.493
175199 101.806 -78.946 44.481
175200 101.822 -78.951 44.469
175201 101.839 -78.957 44.457
175202 101.856 -78.962 44.445
175203 101.873 -78.968 44.432
175204 101.889 -78.974 44.420
175205 101.906 -78.979 44.409
175436 92.678 -77.065 54.054
175437 92.701 -77.076 54.062
175438 92.710 -77.088 54.070
175439 92.717 -77.101 54.078
175440 92.724 -77.113 54.087
175441 92.731 -77.126 54.095
175442 92.738 -77.138 54.102
175443 92.745 -77.151 54.108
175444 92.758 -77.155 54.114
175445 92.778 -77.146 54.120
175446 92.798 -77.138 54.126
175447 92.818 -77.128 54.133
175448 92.839 -77.120 54.140
175449 92.859 -77.111 54.147
175450 92.879 -77.103 54.153
175451 92.900 -77.094 54.158
175452 92.920 -77.085 54.164
175453 92.941 -77.076 54.170
175454 92.961 -77.067 54.176
175455 92.981 -77.059 54.183
175456 93.002 -77.050 54.191
175457 93.022 -77.041 54.198
175458 93.043 -77.032 54.204
175459 93.063 -77.023 54.210
175460 93.083 -77.015 54.217
175461 93.103 -77.006 54.224
175462 93.124 -76.997 54.231
175463 93.144 -76.988 54.239
175464 93.164 -76.980 54.245
175465 93.184 -76.975 54.250
175466 93.203 -76.972 54.250
175467 93.222 -76.969 54.252
175468 93.241 -76.965 54.256
175469 93.255 -76.961 54.261
175470 93.263 -76.956 54.265
175471 93.271 -76.950 54.268
175472 93.279 -76.943 54.271
175473 93.287 -76.938 54.274
175474 93.313 -76.932 54.272
175475 93.352 -76.926 54.271
175476 93.379 -76.918 54.270
175477 93.386 -76.908 54.267
175478 93.392 -76.897 54.268
175479 93.399 -76.887 54.271
175480 93.411 -76.879 54.274
175481 93.429 -76.875 54.274
175482 93.452 -76.875 54.274
175483 93.475 -76.875 54.276
175484 93.497 -76.875 54.278
175485 93.520 -76.875 54.279
175486 93.543 -76.876 54.283
175487 93.565 -76.876 54.286
175488 93.588 -76.876 54.288
175489 93.604 -76.870 54.285
175490 93.617 -76.857 54.281
175491 93.629 -76.844 54.276
175492 93.641 -76.831 54.269
175493 93.660 -76.829 54.263
175494 93.681 -76.828 54.257
175495 93.703 -76.828 54.250
175496 93.718 -76.820 54.242
175497 93.732 -76.812 54.232
175498 93.751 -76.809 54.220
175499 93.769 -76.806 54.207
175500 93.788 -76.802 54.192
175501 93.806 -76.799 54.177
175502 93.825 -76.796 54.161
175503 93.843 -76.793 54.146
175504 93.862 -76.790 54.134
175505 93.880 -76.787 54.123
175506 93.901 -76.784 54.114
175507 93.918 -76.779 54.103
175508 93.935 -76.775 54.090
175509 93.952 -76.771 54.078
175510 93.969 -76.768 54.068
175511 93.986 -76.764 54.061
175512 94.003 -76.761 54.057
175513 94.020 -76.757 54.056
175514 94.037 -76.753 54.056
175515 94.054 -76.749 54.053
175516 94.071 -76.746 54.050
175517 94.074 -76.748 54.048
175518 94.078 -76.750 54.044
175519 94.093 -76.748 54.038
175520 94.119 -76.740 54.032
175521 94.129 -76.738 54.024
175522 94.138 -76.735 54.012
175523 94.148 -76.732 53.998
175524 94.157 -76.729 53.984
175525 94.157 -76.730 53.969
175526 94.145 -76.734 53.954
175527 94.134 -76.739 53.939
175528 94.151 -76.734 53.922
175529 94.172 -76.727 53.905
175530 94.192 -76.721 53.888
175531 94.221 -76.712 53.870
175532 94.251 -76.703 53.853
175533 94.281 -76.693 53.836
175534 94.311 -76.684 53.814
175535 94.333 -76.676 53.788
175536 94.349 -76.670 53.756
175537 94.365 -76.663 53.724
175538 94.382 -76.656 53.691
175539 94.398 -76.651 53.657
175540 94.416 -76.648 53.621
175541 94.433 -76.645 53.586
175542 94.450 -76.642 53.550
175543 94.454 -76.642 53.515
175544 94.456 -76.643 53.483
175545 94.459 -76.643 53.452
175546 94.467 -76.640 53.424
175547 94.477 -76.634 53.397
175548 94.490 -76.630 53.369
175549 94.534 -76.624 53.341
175550 94.578 -76.618 53.311
175551 94.623 -76.611 53.283
175552 94.659 -76.611 53.251
175553 94.677 -76.606 53.219
175554 94.714 -76.599 53.187
175555 94.760 -76.589 53.154
175556 94.796 -76.579 53.123
175557 94.814 -76.572 53.093
175558 94.822 -76.566 53.065
175559 94.827 -76.560 53.040
175560 94.833 -76.553 53.012
175561 94.838 -76.547 52.983
175562 94.844 -76.541 52.953
175563 94.866 -76.548 52.922
175564 94.891 -76.558 52.892
175565 94.916 -76.567 52.861
175566 94.942 -76.577 52.829
175567 94.962 -76.585 52.792
175568 94.970 -76.586 52.752
175569 94.977 -76.588 52.709
175570 94.985 -76.589 52.662
175571 94.992 -76.591 52.614
175572 95.000 -76.592 52.564
175573 95.007 -76.594 52.514
175574 95.015 -76.596 52.462
175575 95.023 -76.598 52.409
175576 95.030 -76.599 52.355
175577 95.048 -76.603 52.300
175578 95.085 -76.611 52.246
175579 95.122 -76.618 52.193
175580 95.160 -76.626 52.141
175581 95.193 -76.634 52.090
175582 95.202 -76.648 52.041
175583 95.211 -76.661 51.992
175584 95.221 -76.675 51.943
175585 95.231 -76.688 51.896
175586 95.244 -76.702 51.851
175587 95.259 -76.715 51.807
175588 95.273 -76.727 51.764
175589 95.288 -76.740 51.724
175590 95.303 -76.753 51.684
175591 95.317 -76.766 51.645
175592 95.332 -76.779 51.601
175593 95.347 -76.792 51.555
175594 95.361 -76.805 51.507
175595 95.376 -76.818 51.459
175596 95.395 -76.819 51.412
175597 95.413 -76.821 51.365
175598 95.428 -76.823 51.319
175599 95.439 -76.825 51.274
175600 95.450 -76.828 51.228
175601 95.471 -76.840 51.182
175602 95.497 -76.857 51.135
175603 95.523 -76.874 51.092
175604 95.549 -76.891 51.050
175605 95.574 -76.908 51.009
175606 95.593 -76.926 50.968
175607 95.608 -76.945 50.928
175608 95.624 -76.963 50.888
175609 95.639 -76.982 50.848
175610 95.655 -76.998 50.808
175611 95.669 -77.003 50.769
175612 95.684 -77.009 50.729
175613 95.699 -77.014 50.689
175614 95.714 -77.019 50.650
175615 95.729 -77.025 50.610
175616 95.745 -77.031 50.569
175617 95.763 -77.040 50.527
175618 95.782 -77.048 50.486
175619 95.801 -77.057 50.444
175620 95.819 -77.066 50.401
175621 95.846 -77.073 50.356
175622 95.882 -77.081 50.306
175623 95.917 -77.088 50.253
175624 95.953 -77.096 50.200
175625 95.988 -77.103 50.147
175626 96.010 -77.106 50.090
175627 96.030 -77.106 50.026
175628 96.046 -77.106 49.959
175629 96.062 -77.104 49.890
175630 96.079 -77.103 49.821
175631 96.095 -77.102 49.757
175632 96.111 -77.101 49.696
175633 96.128 -77.100 49.639
175634 96.142 -77.103 49.586
175635 96.154 -77.109 49.536
175636 96.167 -77.120 49.488
175637 96.181 -77.137 49.446
175638 96.195 -77.153 49.409
175639 96.208 -77.170 49.374
175640 96.214 -77.191 49.341
175641 96.217 -77.211 49.310
175642 96.221 -77.232 49.279
175643 96.225 -77.252 49.250
175644 96.242 -77.260 49.221
175645 96.261 -77.265 49.198
175646 96.281 -77.269 49.175
175647 96.300 -77.275 49.153
175648 96.320 -77.279 49.134
175649 96.340 -77.284 49.115
175650 96.361 -77.282 49.098
175651 96.381 -77.280 49.081
175652 96.402 -77.279 49.068
175653 96.422 -77.276 49.053
175654 96.444 -77.275 49.038
175655 96.464 -77.273 49.023
175656 96.483 -77.279 49.006
175657 96.499 -77.294 48.989
175658 96.514 -77.311 48.973
175659 96.529 -77.327 48.959
175660 96.545 -77.344 48.944
175661 96.560 -77.360 48.929
175662 96.576 -77.376 48.913
175663 96.591 -77.392 48.896
175664 96.609 -77.405 48.879
175665 96.627 -77.415 48.863
175666 96.646 -77.426 48.847
175667 96.665 -77.436 48.832
175668 96.683 -77.446 48.818
175669 96.702 -77.457 48.805
175670 96.722 -77.467 48.792
175671 96.744 -77.478 48.779
175672 96.766 -77.489 48.766
175673 96.789 -77.500 48.753
175674 96.807 -77.520 48.740
175675 96.825 -77.541 48.727
175676 96.841 -77.557 48.712
175677 96.857 -77.562 48.696
175678 96.872 -77.567 48.678
175679 96.888 -77.572 48.660
175680 96.933 -77.578 48.642
175681 96.969 -77.593 48.622
175682 97.006 -77.608 48.602
175683 97.042 -77.622 48.583
175684 97.074 -77.643 48.564
175685 97.102 -77.668 48.545
175686 97.121 -77.673 48.527
175687 97.130 -77.661 48.509
175688 97.139 -77.650 48.491
175689 97.148 -77.639 48.475
175690 97.157 -77.627 48.457
175691 97.167 -77.616 48.438
175692 97.176 -77.604 48.418
175693 97.185 -77.592 48.397
175694 97.194 -77.581 48.375
175695 97.202 -77.567 48.354
175696 97.209 -77.551 48.330
175697 97.216 -77.536 48.307
175698 97.223 -77.521 48.284
175699 97.230 -77.506 48.262
175700 97.249 -77.495 48.241
175701 97.277 -77.501 48.219
175702 97.300 -77.508 48.197
175703 97.322 -77.517 48.175
175704 97.345 -77.525 48.155
175705 97.367 -77.533 48.134
175706 97.390 -77.541 48.113
175707 97.413 -77.549 48.091
175708 97.435 -77.557 48.069
175709 97.458 -77.565 48.045
175710 97.480 -77.573 48.020
175711 97.503 -77.581 47.994
175712 97.525 -77.589 47.968
175713 97.548 -77.598 47.940
175714 97.571 -77.606 47.912
175715 97.590 -77.618 47.884
175716 97.609 -77.630 47.855
175717 97.631 -77.638 47.826
175718 97.667 -77.640 47.798
175719 97.713 -77.647 47.770
175720 97.733 -77.644 47.742
175721 97.750 -77.642 47.714
175722 97.770 -77.641 47.686
175723 97.789 -77.640 47.658
175724 97.820 -77.647 47.629
175725 97.858 -77.659 47.600
175726 97.895 -77.671 47.572
175727 97.932 -77.683 47.544
175728 97.960 -77.686 47.516
175729 97.978 -77.675 47.489
175730 97.996 -77.665 47.463
175731 98.013 -77.655 47.436
175732 98.030 -77.645 47.407
175733 98.006 -77.631 47.378
175734 97.968 -77.615 47.348
175735 97.990 -77.563 47.317
175736 98.014 -77.508 47.286
175737 98.039 -77.454 47.258
175738 98.065 -77.399 47.231
175739 98.085 -77.397 47.204
175740 98.105 -77.396 47.175
175741 98.125 -77.396 47.149
175742 98.145 -77.395 47.123
175743 98.165 -77.395 47.097
175744 98.197 -77.432 47.071
175745 98.228 -77.467 47.044
175746 98.260 -77.504 47.017
175747 98.291 -77.540 46.989
175748 98.323 -77.577 46.961
175749 98.354 -77.613 46.934
175750 98.382 -77.646 46.907
175751 98.378 -77.657 46.880
175752 98.374 -77.667 46.852
175753 98.370 -77.678 46.826
175754 98.371 -77.684 46.801
175755 98.377 -77.688 46.777
175756 98.397 -77.688 46.752
175757 98.416 -77.688 46.729
175758 98.430 -77.690 46.706
175759 98.456 -77.686 46.684
175760 98.486 -77.680 46.660
175761 98.510 -77.677 46.638
175762 98.534 -77.673 46.617
175763 98.559 -77.670 46.597
175764 98.583 -77.667 46.576
175765 98.607 -77.666 46.554
175766 98.631 -77.665 46.531
175767 98.649 -77.672 46.509
175768 98.669 -77.678 46.489
175769 98.690 -77.684 46.470
175770 98.711 -77.689 46.450
175771 98.731 -77.696 46.430
175772 98.745 -77.697 46.409
175773 98.751 -77.705 46.388
175774 98.758 -77.712 46.366
175775 98.764 -77.720 46.347
175776 98.771 -77.727 46.328
175777 98.777 -77.735 46.306
175778 98.782 -77.740 46.285
175779 98.783 -77.741 46.263
175780 98.785 -77.743 46.241
175781 98.787 -77.744 46.219
175782 98.788 -77.746 46.197
175783 98.790 -77.748 46.176
175784 98.791 -77.751 46.154
175785 98.792 -77.754 46.134
175786 98.792 -77.758 46.112
175787 98.793 -77.761 46.091
175788 98.814 -77.766 46.071
175789 98.854 -77.774 46.052
175790 98.885 -77.779 46.034
175791 98.916 -77.783 46.016
175792 98.946 -77.788 45.999
175793 98.976 -77.793 45.982
175794 99.006 -77.797 45.966
175795 99.037 -77.802 45.950
175796 99.067 -77.806 45.936
175797 99.097 -77.811 45.921
175798 99.118 -77.823 45.908
175799 99.139 -77.834 45.894
175800 99.159 -77.846 45.882
175801 99.180 -77.858 45.870
175802 99.204 -77.871 45.859
175803 99.237 -77.890 45.849
175804 99.271 -77.909 45.839
175805 99.305 -77.927 45.829
175806 99.343 -77.934 45.820
175807 99.356 -77.946 45.812
175808 99.362 -77.960 45.803
175809 99.367 -77.974 45.795
175810 99.372 -77.987 45.787
175811 99.377 -78.001 45.780
175812 99.383 -78.015 45.772
175813 99.388 -78.029 45.764
175814 99.394 -78.043 45.756
175815 99.433 -78.050 45.747
175816 99.476 -78.056 45.736
175817 99.518 -78.062 45.725
175818 99.561 -78.069 45.713
175819 99.597 -78.083 45.701
175820 99.621 -78.103 45.689
175821 99.632 -78.115 45.677
175822 99.643 -78.127 45.665
175823 99.659 -78.132 45.652
175824 99.690 -78.119 45.638
175825 99.720 -78.106 45.624
175826 99.761 -78.129 45.610
175827 99.805 -78.165 45.595
175828 99.848 -78.201 45.580
175829 99.892 -78.237 45.564
175830 99.929 -78.252 45.549
175831 99.965 -78.267 45.532
175832 100.001 -78.282 45.515
175833 100.034 -78.291 45.499
175834 100.067 -78.297 45.482
175835 100.099 -78.304 45.466
175836 100.123 -78.311 45.451
175837 100.128 -78.323 45.436
175838 100.133 -78.335 45.421
175839 100.138 -78.347 45.406
175840 100.143 -78.359 45.391
175841 100.148 -78.370 45.376
175842 100.153 -78.382 45.361
175843 100.179 -78.388 45.345
175844 100.209 -78.393 45.329
175845 100.239 -78.398 45.313
175846 100.269 -78.403 45.297
175847 100.298 -78.408 45.281
175848 100.328 -78.413 45.265
175849 100.358 -78.418 45.250
175850 100.388 -78.423 45.234
175851 100.418 -78.428 45.219
175852 100.447 -78.432 45.204
175853 100.479 -78.436 45.190
175854 100.519 -78.442 45.175
175855 100.560 -78.449 45.161
175856 100.600 -78.456 45.144
175857 100.640 -78.462 45.128
175858 100.680 -78.469 45.112
175859 100.720 -78.476 45.096
175860 100.760 -78.482 45.079
175861 100.800 -78.489 45.063
175862 100.840 -78.495 45.047
175863 100.878 -78.501 45.030
175864 100.904 -78.507 45.015
175865 100.930 -78.513 45.000
175866 100.957 -78.518 44.986
175867 100.948 -78.497 44.971
175868 100.938 -78.473 44.956
175869 100.982 -78.500 44.941
175870 101.026 -78.527 44.925
175871 101.060 -78.526 44.910
175872 101.107 -78.509 44.894
175873 101.162 -78.507 44.877
175874 101.209 -78.523 44.861
175875 101.232 -78.539 44.846
175876 101.253 -78.554 44.830
175877 101.275 -78.570 44.814
175878 101.298 -78.586 44.798
175879 101.321 -78.602 44.782
175880 101.347 -78.620 44.767
175881 101.374 -78.641 44.753
175882 101.400 -78.662 44.738
175883 101.426 -78.683 44.722
175884 101.453 -78.705 44.705
175885 101.479 -78.726 44.691
175886 101.505 -78.747 44.677
175887 101.532 -78.768 44.662
175888 101.558 -78.789 44.647
175889 101.585 -78.811 44.632
175890 101.611 -78.832 44.616
175891 101.637 -78.853 44.602
175892 101.664 -78.874 44.588
175893 101.689 -78.893 44.574
175894 101.705 -78.899 44.561
175895 101.722 -78.905 44.547
175896 101.739 -78.910 44.534
175897 101.756 -78.916 44.520
175898 101.772 -78.921 44.507
175899 101.790 -78.927 44.496
175900 101.806 -78.933 44.483
175901 101.823 -78.939 44.471
175902 101.840 -78.944 44.459
175903 101.856 -78.949 44.447
175904 101.873 -78.955 44.434
175905 101.890 -78.961 44.422
175906 101.907 -78.967 44.411
175907 101.923 -78.972 44.400
176141 92.717 -77.084 54.059
176142 92.724 -77.097 54.067
176143 92.731 -77.109 54.074
176144 92.738 -77.122 54.081
176145 92.747 -77.132 54.087
176146 92.768 -77.122 54.093
176147 92.788 -77.114 54.100
176148 92.808 -77.105 54.108
176149 92.829 -77.096 54.115
176150 92.849 -77.088 54.122
176151 92.869 -77.078 54.129
176152 92.889 -77.070 54.135
176153 92.910 -77.061 54.141
176154 92.930 -77.053 54.148
176155 92.950 -77.044 54.156
176156 92.971 -77.035 54.163
176157 92.991 -77.026 54.171
176158 93.011 -77.017 54.179
176159 93.032 -77.009 54.186
176160 93.052 -77.000 54.193
176161 93.073 -76.991 54.200
176162 93.093 -76.982 54.208
176163 93.113 -76.974 54.215
176164 93.134 -76.965 54.222
176165 93.154 -76.956 54.227
176166 93.174 -76.952 54.231
176167 93.193 -76.949 54.235
176168 93.212 -76.945 54.241
176169 93.231 -76.942 54.247
176170 93.250 -76.939 54.252
176171 93.269 -76.936 54.258
176172 93.289 -76.932 54.262
176173 93.302 -76.928 54.265
176174 93.310 -76.922 54.266
176175 93.319 -76.916 54.265
176176 93.348 -76.910 54.264
176177 93.373 -76.902 54.262
176178 93.379 -76.892 54.260
176179 93.389 -76.883 54.261
176180 93.403 -76.875 54.264
176181 93.416 -76.867 54.268
176182 93.430 -76.859 54.271
176183 93.450 -76.858 54.274
176184 93.473 -76.858 54.278
176185 93.495 -76.858 54.282
176186 93.518 -76.858 54.284
176187 93.541 -76.858 54.287
176188 93.563 -76.859 54.291
176189 93.583 -76.856 54.292
176190 93.601 -76.852 54.290
176191 93.617 -76.847 54.287
176192 93.630 -76.834 54.283
176193 93.642 -76.821 54.278
176194 93.659 -76.815 54.273
176195 93.678 -76.812 54.267
176196 93.697 -76.811 54.259
176197 93.716 -76.807 54.251
176198 93.733 -76.802 54.242
176199 93.751 -76.799 54.230
176200 93.770 -76.796 54.217
176201 93.788 -76.793 54.203
176202 93.806 -76.789 54.188
176203 93.825 -76.786 54.173
176204 93.843 -76.782 54.159
176205 93.861 -76.777 54.148
176206 93.878 -76.772 54.137
176207 93.895 -76.770 54.127
176208 93.913 -76.768 54.117
176209 93.930 -76.766 54.105
176210 93.947 -76.762 54.093
176211 93.964 -76.759 54.081
176212 93.981 -76.755 54.074
176213 93.998 -76.751 54.069
176214 94.015 -76.748 54.066
176215 94.032 -76.744 54.067
176216 94.049 -76.740 54.067
176217 94.066 -76.737 54.067
176218 94.077 -76.735 54.066
176219 94.080 -76.737 54.064
176220 94.101 -76.733 54.061
176221 94.115 -76.729 54.055
176222 94.125 -76.726 54.048
176223 94.134 -76.723 54.036
176224 94.143 -76.721 54.022
176225 94.153 -76.718 54.008
176226 94.159 -76.715 53.994
176227 94.148 -76.720 53.980
176228 94.136 -76.724 53.965
176229 94.156 -76.719 53.949
176230 94.177 -76.713 53.932
176231 94.198 -76.707 53.913
176232 94.219 -76.700 53.894
176233 94.249 -76.691 53.875
176234 94.279 -76.681 53.857
176235 94.309 -76.672 53.836
176236 94.328 -76.665 53.810
176237 94.344 -76.658 53.779
176238 94.360 -76.651 53.748
176239 94.377 -76.644 53.716
176240 94.393 -76.637 53.682
176241 94.409 -76.633 53.646
176242 94.427 -76.630 53.610
176243 94.441 -76.628 53.573
176244 94.443 -76.628 53.535
176245 94.447 -76.628 53.502
176246 94.456 -76.624 53.469
176247 94.466 -76.619 53.438
176248 94.476 -76.614 53.408
176249 94.496 -76.609 53.378
176250 94.540 -76.603 53.347
176251 94.584 -76.597 53.316
176252 94.626 -76.592 53.284
176253 94.659 -76.594 53.252
176254 94.678 -76.589 53.220
176255 94.695 -76.585 53.189
176256 94.737 -76.576 53.159
176257 94.784 -76.566 53.130
176258 94.803 -76.558 53.102
176259 94.808 -76.552 53.076
176260 94.814 -76.546 53.055
176261 94.819 -76.540 53.029
176262 94.825 -76.533 53.001
176263 94.848 -76.542 52.971
176264 94.873 -76.552 52.938
176265 94.899 -76.561 52.908
176266 94.918 -76.568 52.879
176267 94.925 -76.570 52.850
176268 94.933 -76.571 52.815
176269 94.941 -76.574 52.777
176270 94.948 -76.575 52.735
176271 94.956 -76.577 52.688
176272 94.963 -76.578 52.639
176273 94.971 -76.580 52.590
176274 94.978 -76.581 52.539
176275 94.986 -76.583 52.488
176276 94.993 -76.585 52.436
176277 95.001 -76.586 52.383
176278 95.019 -76.590 52.327
176279 95.056 -76.598 52.274
176280 95.093 -76.605 52.220
176281 95.130 -76.613 52.167
176282 95.168 -76.621 52.115
176283 95.197 -76.630 52.064
176284 95.206 -76.643 52.014
176285 95.216 -76.657 51.964
176286 95.226 -76.671 51.916
176287 95.240 -76.683 51.871
176288 95.254 -76.696 51.826
176289 95.269 -76.709 51.781
176290 95.283 -76.722 51.739
176291 95.299 -76.735 51.697
176292 95.313 -76.748 51.655
176293 95.328 -76.761 51.611
176294 95.342 -76.774 51.566
176295 95.357 -76.787 51.521
176296 95.372 -76.799 51.474
176297 95.387 -76.810 51.429
176298 95.406 -76.812 51.385
176299 95.418 -76.814 51.342
176300 95.429 -76.816 51.298
176301 95.440 -76.818 51.254
176302 95.459 -76.829 51.208
176303 95.485 -76.846 51.161
176304 95.511 -76.863 51.116
176305 95.536 -76.881 51.074
176306 95.562 -76.897 51.033
176307 95.588 -76.915 50.992
176308 95.607 -76.933 50.953
176309 95.622 -76.951 50.915
176310 95.638 -76.970 50.876
176311 95.653 -76.984 50.835
176312 95.668 -76.990 50.796
176313 95.683 -76.995 50.755
176314 95.697 -77.000 50.714
176315 95.712 -77.006 50.675
176316 95.731 -77.014 50.635
176317 95.749 -77.022 50.593
176318 95.768 -77.031 50.550
176319 95.787 -77.040 50.506
176320 95.805 -77.048 50.463
176321 95.824 -77.056 50.419
176322 95.853 -77.065 50.372
176323 95.889 -77.072 50.319
176324 95.924 -77.079 50.265
176325 95.960 -77.087 50.211
176326 95.995 -77.094 50.156
176327 96.026 -77.100 50.096
176328 96.043 -77.104 50.031
176329 96.061 -77.101 49.964
176330 96.080 -77.099 49.896
176331 96.098 -77.097 49.829
176332 96.114 -77.096 49.766
176333 96.130 -77.095 49.707
176334 96.147 -77.094 49.650
176335 96.163 -77.093 49.598
176336 96.179 -77.091 49.548
176337 96.192 -77.098 49.501
176338 96.205 -77.115 49.459
176339 96.214 -77.135 49.422
176340 96.217 -77.155 49.389
176341 96.221 -77.175 49.355
176342 96.225 -77.196 49.324
176343 96.228 -77.217 49.294
176344 96.237 -77.232 49.264
176345 96.257 -77.237 49.236
176346 96.276 -77.241 49.213
176347 96.296 -77.247 49.191
176348 96.316 -77.252 49.169
176349 96.335 -77.257 49.150
176350 96.355 -77.262 49.131
176351 96.375 -77.266 49.113
176352 96.395 -77.265 49.097
176353 96.416 -77.263 49.084
176354 96.437 -77.261 49.071
176355 96.457 -77.260 49.057
176356 96.478 -77.258 49.042
176357 96.498 -77.261 49.025
176358 96.513 -77.278 49.009
176359 96.529 -77.294 48.992
176360 96.544 -77.310 48.978
176361 96.559 -77.326 48.963
176362 96.575 -77.342 48.947
176363 96.591 -77.359 48.930
176364 96.608 -77.371 48.913
176365 96.627 -77.381 48.896
176366 96.646 -77.392 48.879
176367 96.664 -77.401 48.863
176368 96.683 -77.412 48.848
176369 96.701 -77.422 48.833
176370 96.720 -77.432 48.818
176371 96.739 -77.443 48.805
176372 96.760 -77.455 48.791
176373 96.778 -77.476 48.778
176374 96.796 -77.498 48.765
176375 96.813 -77.519 48.751
176376 96.829 -77.531 48.737
176377 96.845 -77.536 48.722
176378 96.860 -77.541 48.706
176379 96.876 -77.546 48.687
176380 96.891 -77.551 48.668
176381 96.937 -77.553 48.649
176382 96.973 -77.567 48.629
176383 97.010 -77.582 48.609
176384 97.046 -77.597 48.590
176385 97.077 -77.618 48.571
176386 97.106 -77.643 48.551
176387 97.135 -77.668 48.532
176388 97.149 -77.665 48.514
176389 97.158 -77.654 48.495
176390 97.167 -77.642 48.478
176391 97.176 -77.630 48.461
176392 97.186 -77.619 48.444
176393 97.195 -77.608 48.424
176394 97.204 -77.596 48.403
176395 97.213 -77.584 48.382
176396 97.220 -77.569 48.359
176397 97.227 -77.554 48.336
176398 97.234 -77.539 48.313
176399 97.242 -77.523 48.291
176400 97.249 -77.508 48.270
176401 97.256 -77.493 48.249
176402 97.287 -77.494 48.228
176403 97.310 -77.502 48.206
176404 97.332 -77.511 48.186
176405 97.355 -77.518 48.165
176406 97.378 -77.526 48.143
176407 97.400 -77.535 48.122
176408 97.423 -77.542 48.100
176409 97.445 -77.551 48.077
176410 97.468 -77.559 48.054
176411 97.491 -77.567 48.029
176412 97.513 -77.575 48.003
176413 97.536 -77.583 47.976
176414 97.555 -77.595 47.949
176415 97.574 -77.608 47.920
176416 97.593 -77.620 47.892
176417 97.612 -77.633 47.863
176418 97.640 -77.631 47.835
176419 97.680 -77.635 47.807
176420 97.726 -77.642 47.778
176421 97.762 -77.645 47.750
176422 97.777 -77.640 47.722
176423 97.794 -77.637 47.694
176424 97.813 -77.636 47.666
176425 97.833 -77.636 47.637
176426 97.871 -77.648 47.608
176427 97.908 -77.660 47.579
176428 97.945 -77.672 47.550
176429 97.970 -77.670 47.523
176430 97.988 -77.660 47.497
176431 98.005 -77.650 47.470
176432 98.023 -77.640 47.442
176433 98.040 -77.630 47.413
176434 98.035 -77.618 47.383
176435 97.997 -77.602 47.353
176436 98.016 -77.551 47.323
176437 98.041 -77.497 47.294
176438 98.066 -77.442 47.265
176439 98.088 -77.420 47.236
176440 98.108 -77.420 47.208
176441 98.129 -77.420 47.180
176442 98.149 -77.419 47.154
176443 98.169 -77.419 47.128
176444 98.189 -77.418 47.102
176445 98.209 -77.418 47.076
176446 98.234 -77.432 47.049
176447 98.265 -77.469 47.023
176448 98.297 -77.505 46.996
176449 98.329 -77.541 46.968
176450 98.360 -77.577 46.941
176451 98.391 -77.614 46.914
176452 98.404 -77.636 46.887
176453 98.400 -77.647 46.860
176454 98.396 -77.658 46.833
176455 98.392 -77.668 46.808
176456 98.391 -77.676 46.785
176457 98.411 -77.675 46.761
176458 98.435 -77.672 46.737
176459 98.459 -77.669 46.714
176460 98.479 -77.668 46.691
176461 98.505 -77.664 46.668
176462 98.529 -77.661 46.646
176463 98.553 -77.658 46.625
176464 98.577 -77.655 46.603
176465 98.601 -77.651 46.582
176466 98.626 -77.648 46.561
176467 98.650 -77.645 46.538
176468 98.670 -77.650 46.517
176469 98.692 -77.655 46.496
176470 98.712 -77.661 46.477
176471 98.733 -77.667 46.457
176472 98.754 -77.672 46.438
176473 98.765 -77.678 46.418
176474 98.771 -77.686 46.397
176475 98.777 -77.692 46.375
176476 98.784 -77.700 46.355
176477 98.790 -77.708 46.336
176478 98.797 -77.715 46.316
176479 98.803 -77.723 46.295
176480 98.808 -77.728 46.273
176481 98.810 -77.730 46.251
176482 98.811 -77.731 46.230
176483 98.813 -77.733 46.208
176484 98.814 -77.736 46.185
176485 98.815 -77.740 46.163
176486 98.816 -77.743 46.141
176487 98.816 -77.746 46.120
176488 98.817 -77.749 46.098
176489 98.819 -77.753 46.077
176490 98.853 -77.758 46.057
176491 98.883 -77.763 46.039
176492 98.914 -77.768 46.020
176493 98.944 -77.772 46.003
176494 98.974 -77.777 45.985
176495 99.004 -77.781 45.969
176496 99.035 -77.787 45.954
176497 99.065 -77.791 45.939
176498 99.095 -77.796 45.925
176499 99.126 -77.800 45.911
176500 99.146 -77.812 45.897
176501 99.166 -77.824 45.885
176502 99.187 -77.836 45.874
176503 99.215 -77.852 45.863
176504 99.249 -77.870 45.853
176505 99.283 -77.889 45.844
176506 99.316 -77.907 45.835
176507 99.351 -77.924 45.826
176508 99.360 -77.937 45.817
176509 99.366 -77.952 45.808
176510 99.373 -77.968 45.800
176511 99.380 -77.984 45.792
176512 99.387 -78.000 45.785
176513 99.394 -78.015 45.777
176514 99.401 -78.031 45.769
176515 99.420 -78.042 45.761
176516 99.441 -78.052 45.751
176517 99.474 -78.059 45.741
176518 99.517 -78.066 45.730
176519 99.560 -78.073 45.718
176520 99.600 -78.082 45.706
176521 99.622 -78.102 45.695
176522 99.634 -78.114 45.683
176523 99.646 -78.124 45.671
176524 99.677 -78.110 45.658
176525 99.707 -78.097 45.645
176526 99.737 -78.084 45.631
176527 99.780 -78.112 45.617
176528 99.823 -78.149 45.602
176529 99.867 -78.184 45.588
176530 99.906 -78.206 45.572
176531 99.942 -78.221 45.556
176532 99.978 -78.235 45.539
176533 100.014 -78.250 45.522
176534 100.050 -78.266 45.506
176535 100.086 -78.280 45.490
176536 100.121 -78.292 45.475
176537 100.136 -78.302 45.460
176538 100.141 -78.313 45.445
176539 100.146 -78.326 45.429
176540 100.151 -78.338 45.414
176541 100.156 -78.349 45.400
176542 100.161 -78.361 45.384
176543 100.166 -78.373 45.369
176544 100.178 -78.382 45.353
176545 100.208 -78.388 45.336
176546 100.238 -78.392 45.320
176547 100.268 -78.398 45.304
176548 100.297 -78.402 45.287
176549 100.328 -78.407 45.272
176550 100.357 -78.412 45.256
176551 100.387 -78.417 45.241
176552 100.417 -78.422 45.226
176553 100.448 -78.423 45.211
176554 100.482 -78.424 45.196
176555 100.521 -78.429 45.182
176556 100.561 -78.436 45.167
176557 100.601 -78.442 45.151
176558 100.641 -78.449 45.135
176559 100.681 -78.456 45.119
176560 100.721 -78.462 45.102
176561 100.761 -78.469 45.086
176562 100.802 -78.476 45.069
176563 100.842 -78.482 45.053
176564 100.869 -78.488 45.037
176565 100.896 -78.493 45.021
176566 100.922 -78.499 45.006
176567 100.949 -78.504 44.992
176568 100.969 -78.506 44.977
176569 100.959 -78.482 44.962
176570 100.987 -78.495 44.947
176571 101.031 -78.522 44.930
176572 101.068 -78.507 44.915
176573 101.120 -78.489 44.900
176574 101.177 -78.495 44.883
176575 101.206 -78.511 44.867
176576 101.228 -78.527 44.851
176577 101.250 -78.543 44.835
176578 101.272 -78.559 44.819
176579 101.298 -78.576 44.804
176580 101.324 -78.593 44.788
176581 101.350 -78.610 44.772
176582 101.376 -78.630 44.757
176583 101.402 -78.652 44.742
176584 101.428 -78.674 44.726
176585 101.455 -78.695 44.711
176586 101.481 -78.717 44.696
176587 101.507 -78.738 44.681
176588 101.534 -78.759 44.666
176589 101.560 -78.780 44.651
176590 101.587 -78.802 44.636
176591 101.613 -78.823 44.620
176592 101.640 -78.844 44.605
176593 101.665 -78.868 44.591
176594 101.688 -78.902 44.578
176595 101.700 -78.898 44.564
176596 101.717 -78.904 44.550
176597 101.734 -78.909 44.537
176598 101.751 -78.915 44.523
176599 101.767 -78.921 44.510
176600 101.784 -78.926 44.498
176601 101.801 -78.932 44.486
176602 101.818 -78.937 44.473
176603 101.835 -78.943 44.461
176604 101.851 -78.949 44.449
176605 101.868 -78.954 44.437
176606 101.885 -78.960 44.425
176607 101.902 -78.965 44.413
176608 101.918 -78.971 44.402
176852 92.859 -77.055 54.103
176853 92.879 -77.047 54.111
176854 92.900 -77.038 54.118
176855 92.920 -77.028 54.126
176856 92.940 -77.020 54.134
176857 92.960 -77.011 54.143
176858 92.981 -77.003 54.152
176859 93.001 -76.994 54.160
176860 93.021 -76.985 54.168
176861 93.042 -76.976 54.176
176862 93.062 -76.967 54.184
176863 93.082 -76.959 54.192
176864 93.103 -76.950 54.199
176865 93.123 -76.941 54.206
176866 93.143 -76.932 54.211
176867 93.163 -76.928 54.217
176868 93.182 -76.925 54.221
176869 93.201 -76.922 54.226
176870 93.220 -76.918 54.232
176871 93.239 -76.915 54.237
176872 93.259 -76.912 54.242
176873 93.278 -76.909 54.246
176874 93.297 -76.906 54.250
176875 93.317 -76.902 54.252
176876 93.336 -76.899 54.252
176877 93.350 -76.894 54.251
176878 93.366 -76.886 54.250
176879 93.380 -76.878 54.248
176880 93.394 -76.871 54.249
176881 93.408 -76.863 54.252
176882 93.421 -76.856 54.258
176883 93.435 -76.848 54.264
176884 93.448 -76.840 54.270
176885 93.471 -76.840 54.276
176886 93.493 -76.840 54.282
176887 93.516 -76.840 54.286
176888 93.539 -76.840 54.289
176889 93.561 -76.840 54.292
176890 93.579 -76.837 54.292
176891 93.596 -76.832 54.291
176892 93.613 -76.828 54.289
176893 93.630 -76.824 54.287
176894 93.643 -76.812 54.284
176895 93.657 -76.801 54.280
176896 93.676 -76.799 54.274
176897 93.695 -76.796 54.267
176898 93.714 -76.793 54.258
176899 93.731 -76.788 54.249
176900 93.749 -76.784 54.238
176901 93.767 -76.779 54.226
176902 93.784 -76.774 54.212
176903 93.802 -76.769 54.198
176904 93.820 -76.765 54.183
176905 93.837 -76.760 54.170
176906 93.855 -76.755 54.159
176907 93.872 -76.750 54.148
176908 93.889 -76.748 54.139
176909 93.904 -76.749 54.130
176910 93.922 -76.748 54.118
176911 93.939 -76.746 54.105
176912 93.958 -76.744 54.095
176913 93.975 -76.743 54.086
176914 93.993 -76.741 54.081
176915 94.010 -76.738 54.078
176916 94.027 -76.734 54.078
176917 94.044 -76.731 54.078
176918 94.061 -76.727 54.079
176919 94.078 -76.724 54.082
176920 94.083 -76.725 54.082
176921 94.102 -76.720 54.080
176922 94.111 -76.717 54.076
176923 94.121 -76.715 54.071
176924 94.130 -76.712 54.058
176925 94.139 -76.708 54.044
176926 94.148 -76.705 54.031
176927 94.158 -76.703 54.017
176928 94.150 -76.706 54.003
176929 94.141 -76.710 53.989
176930 94.161 -76.704 53.975
176931 94.182 -76.698 53.956
176932 94.203 -76.692 53.938
176933 94.223 -76.686 53.920
176934 94.247 -76.678 53.900
176935 94.276 -76.669 53.880
176936 94.306 -76.660 53.858
176937 94.322 -76.653 53.833
176938 94.339 -76.646 53.801
176939 94.355 -76.640 53.771
176940 94.371 -76.633 53.740
176941 94.387 -76.626 53.707
176942 94.403 -76.619 53.672
176943 94.420 -76.615 53.635
176944 94.431 -76.614 53.597
176945 94.435 -76.613 53.558
176946 94.445 -76.608 53.523
176947 94.456 -76.603 53.489
176948 94.466 -76.599 53.456
176949 94.476 -76.594 53.423
176950 94.502 -76.588 53.392
176951 94.547 -76.583 53.358
176952 94.591 -76.577 53.324
176953 94.627 -76.576 53.290
176954 94.659 -76.577 53.257
176955 94.679 -76.574 53.225
176956 94.696 -76.568 53.195
176957 94.713 -76.563 53.168
176958 94.760 -76.553 53.140
176959 94.788 -76.546 53.114
176960 94.794 -76.539 53.089
176961 94.800 -76.533 53.069
176962 94.805 -76.527 53.045
176963 94.830 -76.536 53.018
176964 94.855 -76.546 52.988
176965 94.874 -76.552 52.954
176966 94.881 -76.554 52.924
176967 94.889 -76.555 52.896
176968 94.896 -76.557 52.868
176969 94.904 -76.558 52.835
176970 94.911 -76.560 52.799
176971 94.919 -76.562 52.758
176972 94.926 -76.564 52.711
176973 94.934 -76.565 52.663
176974 94.942 -76.567 52.614
176975 94.949 -76.568 52.564
176976 94.957 -76.570 52.513
176977 94.964 -76.572 52.461
176978 94.972 -76.574 52.408
176979 94.990 -76.577 52.353
176980 95.027 -76.585 52.299
176981 95.064 -76.592 52.246
176982 95.101 -76.600 52.193
176983 95.138 -76.608 52.141
176984 95.175 -76.615 52.089
176985 95.201 -76.625 52.039
176986 95.211 -76.639 51.988
176987 95.220 -76.652 51.939
176988 95.235 -76.665 51.892
176989 95.250 -76.678 51.846
176990 95.265 -76.691 51.800
176991 95.279 -76.704 51.756
176992 95.294 -76.717 51.712
176993 95.309 -76.730 51.668
176994 95.323 -76.743 51.623
176995 95.338 -76.755 51.578
176996 95.353 -76.768 51.534
176997 95.368 -76.781 51.490
176998 95.382 -76.794 51.447
176999 95.397 -76.802 51.405
177000 95.407 -76.805 51.364
177001 95.418 -76.807 51.321
177002 95.429 -76.809 51.277
177003 95.447 -76.818 51.231
177004 95.473 -76.836 51.185
177005 95.498 -76.853 51.140
177006 95.524 -76.870 51.098
177007 95.550 -76.887 51.058
177008 95.576 -76.903 51.019
177009 95.602 -76.921 50.980
177010 95.621 -76.939 50.943
177011 95.636 -76.957 50.905
177012 95.652 -76.971 50.866
177013 95.667 -76.976 50.824
177014 95.681 -76.981 50.782
177015 95.698 -76.988 50.740
177016 95.716 -76.997 50.700
177017 95.735 -77.006 50.658
177018 95.754 -77.014 50.615
177019 95.772 -77.022 50.570
177020 95.791 -77.031 50.524
177021 95.809 -77.040 50.479
177022 95.828 -77.048 50.434
177023 95.861 -77.056 50.386
177024 95.896 -77.063 50.332
177025 95.932 -77.071 50.277
177026 95.967 -77.078 50.221
177027 96.002 -77.081 50.164
177028 96.029 -77.081 50.102
177029 96.043 -77.096 50.037
177030 96.058 -77.108 49.969
177031 96.077 -77.105 49.902
177032 96.095 -77.103 49.837
177033 96.114 -77.100 49.776
177034 96.133 -77.098 49.718
177035 96.152 -77.095 49.662
177036 96.170 -77.093 49.610
177037 96.189 -77.091 49.561
177038 96.207 -77.088 49.514
177039 96.217 -77.099 49.473
177040 96.221 -77.119 49.436
177041 96.224 -77.140 49.403
177042 96.228 -77.160 49.369
177043 96.232 -77.182 49.339
177044 96.235 -77.202 49.309
177045 96.253 -77.209 49.279
177046 96.272 -77.213 49.252
177047 96.292 -77.219 49.229
177048 96.312 -77.224 49.207
177049 96.331 -77.229 49.186
177050 96.350 -77.234 49.166
177051 96.370 -77.238 49.147
177052 96.390 -77.244 49.130
177053 96.409 -77.248 49.113
177054 96.430 -77.248 49.101
177055 96.451 -77.247 49.088
177056 96.471 -77.244 49.075
177057 96.492 -77.243 49.062
177058 96.512 -77.244 49.046
177059 96.528 -77.260 49.028
177060 96.543 -77.276 49.011
177061 96.558 -77.292 48.997
177062 96.574 -77.309 48.982
177063 96.590 -77.325 48.965
177064 96.608 -77.337 48.948
177065 96.626 -77.347 48.930
177066 96.645 -77.357 48.913
177067 96.664 -77.367 48.896
177068 96.682 -77.378 48.880
177069 96.701 -77.388 48.863
177070 96.720 -77.398 48.847
177071 96.736 -77.410 48.833
177072 96.751 -77.429 48.819
177073 96.766 -77.454 48.804
177074 96.784 -77.476 48.790
177075 96.802 -77.497 48.776
177076 96.817 -77.505 48.760
177077 96.833 -77.510 48.746
177078 96.848 -77.515 48.731
177079 96.864 -77.520 48.714
177080 96.879 -77.526 48.694
177081 96.894 -77.530 48.674
177082 96.941 -77.528 48.655
177083 96.977 -77.542 48.636
177084 97.014 -77.557 48.618
177085 97.050 -77.571 48.599
177086 97.081 -77.594 48.579
177087 97.110 -77.619 48.558
177088 97.139 -77.644 48.539
177089 97.168 -77.668 48.519
177090 97.177 -77.657 48.500
177091 97.186 -77.645 48.482
177092 97.195 -77.634 48.466
177093 97.205 -77.623 48.448
177094 97.214 -77.611 48.429
177095 97.223 -77.600 48.408
177096 97.231 -77.586 48.386
177097 97.238 -77.571 48.364
177098 97.246 -77.556 48.341
177099 97.253 -77.541 48.319
177100 97.260 -77.526 48.297
177101 97.267 -77.511 48.276
177102 97.275 -77.495 48.256
177103 97.298 -77.488 48.236
177104 97.320 -77.496 48.216
177105 97.343 -77.504 48.196
177106 97.365 -77.512 48.174
177107 97.388 -77.520 48.154
177108 97.410 -77.528 48.133
177109 97.433 -77.536 48.111
177110 97.456 -77.545 48.088
177111 97.478 -77.552 48.064
177112 97.501 -77.561 48.039
177113 97.521 -77.572 48.013
177114 97.540 -77.585 47.987
177115 97.559 -77.597 47.959
177116 97.578 -77.609 47.930
177117 97.597 -77.622 47.902
177118 97.621 -77.625 47.873
177119 97.648 -77.624 47.845
177120 97.693 -77.630 47.816
177121 97.739 -77.636 47.787
177122 97.785 -77.642 47.759
177123 97.806 -77.641 47.730
177124 97.821 -77.636 47.702
177125 97.837 -77.633 47.673
177126 97.857 -77.632 47.644
177127 97.884 -77.636 47.615
177128 97.921 -77.649 47.586
177129 97.959 -77.661 47.558
177130 97.980 -77.655 47.530
177131 97.997 -77.645 47.503
177132 98.015 -77.635 47.475
177133 98.032 -77.625 47.447
177134 98.050 -77.615 47.418
177135 98.065 -77.605 47.389
177136 98.026 -77.589 47.359
177137 98.043 -77.539 47.330
177138 98.068 -77.485 47.301
177139 98.091 -77.444 47.272
177140 98.111 -77.444 47.242
177141 98.132 -77.443 47.213
177142 98.152 -77.443 47.185
177143 98.172 -77.442 47.159
177144 98.192 -77.442 47.133
177145 98.212 -77.442 47.106
177146 98.232 -77.441 47.080
177147 98.252 -77.441 47.053
177148 98.273 -77.440 47.027
177149 98.302 -77.470 47.001
177150 98.334 -77.506 46.974
177151 98.365 -77.542 46.948
177152 98.396 -77.579 46.921
177153 98.428 -77.614 46.894
177154 98.426 -77.627 46.867
177155 98.422 -77.638 46.841
177156 98.418 -77.649 46.816
177157 98.414 -77.659 46.793
177158 98.423 -77.661 46.771
177159 98.444 -77.661 46.747
177160 98.466 -77.659 46.723
177161 98.488 -77.658 46.700
177162 98.511 -77.656 46.677
177163 98.537 -77.650 46.655
177164 98.563 -77.645 46.634
177165 98.588 -77.639 46.611
177166 98.614 -77.633 46.589
177167 98.639 -77.630 46.567
177168 98.661 -77.636 46.546
177169 98.683 -77.640 46.524
177170 98.704 -77.645 46.503
177171 98.725 -77.651 46.483
177172 98.746 -77.655 46.463
177173 98.769 -77.658 46.444
177174 98.780 -77.661 46.425
177175 98.787 -77.667 46.404
177176 98.795 -77.674 46.383
177177 98.802 -77.681 46.363
177178 98.810 -77.688 46.343
177179 98.816 -77.696 46.324
177180 98.822 -77.703 46.303
177181 98.829 -77.711 46.282
177182 98.835 -77.717 46.260
177183 98.836 -77.718 46.238
177184 98.837 -77.721 46.216
177185 98.838 -77.725 46.194
177186 98.839 -77.728 46.171
177187 98.839 -77.732 46.149
177188 98.838 -77.736 46.127
177189 98.837 -77.740 46.106
177190 98.838 -77.743 46.085
177191 98.857 -77.749 46.066
177192 98.889 -77.756 46.046
177193 98.915 -77.756 46.027
177194 98.942 -77.757 46.008
177195 98.972 -77.761 45.990
177196 99.003 -77.766 45.973
177197 99.033 -77.771 45.957
177198 99.063 -77.775 45.942
177199 99.093 -77.780 45.928
177200 99.124 -77.784 45.915
177201 99.153 -77.790 45.901
177202 99.173 -77.802 45.888
177203 99.194 -77.814 45.877
177204 99.227 -77.832 45.867
177205 99.261 -77.850 45.857
177206 99.295 -77.868 45.848
177207 99.328 -77.893 45.839
177208 99.354 -77.918 45.831
177209 99.374 -77.941 45.820
177210 99.381 -77.957 45.812
177211 99.388 -77.973 45.804
177212 99.395 -77.989 45.796
177213 99.402 -78.005 45.789
177214 99.409 -78.021 45.781
177215 99.426 -78.032 45.772
177216 99.447 -78.042 45.764
177217 99.469 -78.051 45.755
177218 99.491 -78.061 45.745
177219 99.515 -78.069 45.734
177220 99.558 -78.076 45.722
177221 99.601 -78.083 45.711
177222 99.624 -78.100 45.700
177223 99.635 -78.112 45.689
177224 99.663 -78.102 45.677
177225 99.694 -78.088 45.665
177226 99.724 -78.075 45.651
177227 99.755 -78.062 45.636
177228 99.798 -78.097 45.622
177229 99.842 -78.132 45.607
177230 99.883 -78.159 45.592
177231 99.919 -78.174 45.576
177232 99.955 -78.189 45.561
177233 99.991 -78.204 45.544
177234 100.028 -78.219 45.528
177235 100.063 -78.234 45.512
177236 100.098 -78.254 45.498
177237 100.130 -78.279 45.483
177238 100.149 -78.287 45.468
177239 100.153 -78.304 45.452
177240 100.158 -78.316 45.437
177241 100.163 -78.328 45.421
177242 100.168 -78.340 45.406
177243 100.174 -78.351 45.391
177244 100.179 -78.363 45.375
177245 100.184 -78.375 45.359
177246 100.207 -78.382 45.342
177247 100.237 -78.387 45.326
177248 100.267 -78.391 45.310
177249 100.297 -78.397 45.293
177250 100.327 -78.401 45.278
177251 100.356 -78.407 45.263
177252 100.387 -78.409 45.248
177253 100.418 -78.410 45.233
177254 100.449 -78.410 45.218
177255 100.484 -78.413 45.203
177256 100.522 -78.416 45.188
177257 100.562 -78.423 45.173
177258 100.602 -78.429 45.158
177259 100.642 -78.435 45.142
177260 100.682 -78.442 45.125
177261 100.722 -78.449 45.108
177262 100.763 -78.455 45.092
177263 100.802 -78.462 45.075
177264 100.835 -78.468 45.058
177265 100.861 -78.474 45.042
177266 100.888 -78.479 45.027
177267 100.914 -78.485 45.012
177268 100.940 -78.491 44.998
177269 100.967 -78.496 44.983
177270 100.980 -78.492 44.968
177271 100.992 -78.489 44.952
177272 101.032 -78.504 44.937
177273 101.080 -78.487 44.922
177274 101.132 -78.469 44.906
177275 101.181 -78.483 44.889
177276 101.203 -78.499 44.874
177277 101.225 -78.515 44.858
177278 101.248 -78.532 44.842
177279 101.274 -78.549 44.826
177280 101.301 -78.566 44.810
177281 101.327 -78.583 44.794
177282 101.353 -78.600 44.777
177283 101.379 -78.619 44.761
177284 101.405 -78.642 44.746
177285 101.431 -78.664 44.731
177286 101.457 -78.686 44.716
177287 101.483 -78.708 44.701
177288 101.510 -78.729 44.685
177289 101.536 -78.750 44.670
177290 101.563 -78.771 44.655
177291 101.589 -78.793 44.640
177292 101.616 -78.814 44.624
177293 101.641 -78.839 44.609
177294 101.664 -78.872 44.596
177295 101.687 -78.905 44.582
177296 101.696 -78.899 44.568
177297 101.712 -78.902 44.554
177298 101.729 -78.908 44.541
177299 101.746 -78.914 44.527
177300 101.762 -78.919 44.514
177301 101.779 -78.925 44.501
177302 101.796 -78.930 44.489
177303 101.813 -78.937 44.477
177304 101.830 -78.942 44.464
177305 101.846 -78.947 44.452
177306 101.863 -78.953 44.440
177307 101.880 -78.959 44.428
177308 101.897 -78.965 44.416
177309 101.913 -78.970 44.405
177555 92.889 -77.014 54.094
177556 92.909 -77.005 54.103
177557 92.930 -76.997 54.112
177558 92.950 -76.987 54.122
177559 92.970 -76.979 54.132
177560 92.991 -76.970 54.141
177561 93.011 -76.961 54.151
177562 93.031 -76.953 54.158
177563 93.052 -76.943 54.167
177564 93.072 -76.935 54.175
177565 93.092 -76.926 54.183
177566 93.113 -76.918 54.190
177567 93.133 -76.909 54.196
177568 93.152 -76.904 54.202
177569 93.171 -76.901 54.208
177570 93.191 -76.898 54.213
177571 93.210 -76.895 54.218
177572 93.229 -76.891 54.223
177573 93.249 -76.888 54.227
177574 93.268 -76.885 54.230
177575 93.287 -76.881 54.233
177576 93.306 -76.878 54.235
177577 93.326 -76.875 54.236
177578 93.345 -76.871 54.236
177579 93.364 -76.868 54.235
177580 93.383 -76.865 54.234
177581 93.399 -76.859 54.235
177582 93.413 -76.852 54.238
177583 93.426 -76.844 54.244
177584 93.440 -76.837 54.252
177585 93.453 -76.829 54.260
177586 93.469 -76.823 54.267
177587 93.491 -76.823 54.275
177588 93.514 -76.823 54.280
177589 93.537 -76.823 54.284
177590 93.557 -76.821 54.288
177591 93.574 -76.817 54.288
177592 93.591 -76.813 54.288
177593 93.608 -76.809 54.287
177594 93.625 -76.804 54.287
177595 93.642 -76.800 54.285
177596 93.656 -76.789 54.282
177597 93.674 -76.785 54.277
177598 93.693 -76.782 54.271
177599 93.711 -76.774 54.262
177600 93.726 -76.767 54.253
177601 93.743 -76.762 54.243
177602 93.761 -76.757 54.232
177603 93.778 -76.752 54.220
177604 93.796 -76.747 54.207
177605 93.814 -76.743 54.192
177606 93.831 -76.738 54.178
177607 93.849 -76.734 54.167
177608 93.866 -76.728 54.158
177609 93.883 -76.727 54.149
177610 93.898 -76.727 54.140
177611 93.913 -76.728 54.129
177612 93.931 -76.727 54.118
177613 93.949 -76.725 54.107
177614 93.967 -76.724 54.099
177615 93.984 -76.722 54.092
177616 94.002 -76.721 54.089
177617 94.020 -76.719 54.089
177618 94.038 -76.718 54.088
177619 94.055 -76.716 54.090
177620 94.073 -76.715 54.093
177621 94.088 -76.712 54.095
177622 94.098 -76.708 54.094
177623 94.107 -76.705 54.092
177624 94.116 -76.702 54.089
177625 94.126 -76.700 54.077
177626 94.135 -76.697 54.065
177627 94.144 -76.694 54.053
177628 94.153 -76.691 54.039
177629 94.153 -76.692 54.026
177630 94.146 -76.695 54.013
177631 94.167 -76.689 54.000
177632 94.188 -76.683 53.981
177633 94.208 -76.677 53.963
177634 94.229 -76.671 53.944
177635 94.250 -76.665 53.925
177636 94.274 -76.657 53.905
177637 94.300 -76.648 53.883
177638 94.317 -76.641 53.857
177639 94.333 -76.634 53.827
177640 94.349 -76.627 53.796
177641 94.365 -76.621 53.765
177642 94.382 -76.614 53.732
177643 94.398 -76.608 53.697
177644 94.414 -76.601 53.661
177645 94.424 -76.598 53.624
177646 94.434 -76.593 53.586
177647 94.445 -76.588 53.548
177648 94.455 -76.583 53.512
177649 94.465 -76.578 53.476
177650 94.475 -76.574 53.443
177651 94.509 -76.567 53.410
177652 94.553 -76.561 53.374
177653 94.595 -76.558 53.337
177654 94.627 -76.559 53.302
177655 94.660 -76.561 53.269
177656 94.680 -76.557 53.237
177657 94.697 -76.552 53.206
177658 94.714 -76.546 53.180
177659 94.736 -76.541 53.154
177660 94.774 -76.532 53.129
177661 94.780 -76.526 53.104
177662 94.787 -76.520 53.083
177663 94.812 -76.530 53.058
177664 94.829 -76.536 53.032
177665 94.837 -76.537 53.003
177666 94.844 -76.539 52.970
177667 94.852 -76.541 52.940
177668 94.859 -76.542 52.913
177669 94.867 -76.544 52.885
177670 94.875 -76.546 52.854
177671 94.882 -76.547 52.819
177672 94.890 -76.549 52.778
177673 94.897 -76.551 52.734
177674 94.905 -76.552 52.687
177675 94.912 -76.554 52.639
177676 94.920 -76.555 52.590
177677 94.927 -76.557 52.538
177678 94.935 -76.558 52.486
177679 94.943 -76.561 52.432
177680 94.961 -76.564 52.377
177681 94.998 -76.572 52.324
177682 95.035 -76.580 52.271
177683 95.072 -76.587 52.218
177684 95.109 -76.595 52.166
177685 95.146 -76.602 52.115
177686 95.183 -76.610 52.063
177687 95.205 -76.621 52.012
177688 95.216 -76.634 51.962
177689 95.231 -76.647 51.913
177690 95.246 -76.660 51.866
177691 95.260 -76.673 51.819
177692 95.275 -76.686 51.773
177693 95.289 -76.699 51.727
177694 95.305 -76.712 51.681
177695 95.319 -76.724 51.635
177696 95.334 -76.739 51.591
177697 95.348 -76.753 51.548
177698 95.362 -76.767 51.504
177699 95.377 -76.781 51.462
177700 95.405 -76.793 51.421
177701 95.397 -76.796 51.381
177702 95.408 -76.798 51.339
177703 95.418 -76.800 51.296
177704 95.435 -76.808 51.251
177705 95.461 -76.825 51.207
177706 95.486 -76.842 51.163
177707 95.512 -76.859 51.123
177708 95.538 -76.876 51.084
177709 95.564 -76.893 51.046
177710 95.590 -76.910 51.009
177711 95.616 -76.927 50.972
177712 95.635 -76.945 50.934
177713 95.650 -76.957 50.893
177714 95.665 -76.962 50.850
177715 95.684 -76.972 50.807
177716 95.702 -76.980 50.765
177717 95.721 -76.988 50.723
177718 95.739 -76.997 50.680
177719 95.758 -77.005 50.635
177720 95.776 -77.014 50.589
177721 95.795 -77.022 50.543
177722 95.814 -77.031 50.498
177723 95.833 -77.039 50.450
177724 95.868 -77.047 50.400
177725 95.904 -77.054 50.346
177726 95.939 -77.062 50.290
177727 95.974 -77.066 50.233
177728 96.006 -77.061 50.174
177729 96.032 -77.063 50.110
177730 96.046 -77.078 50.043
177731 96.060 -77.093 49.976
177732 96.074 -77.107 49.909
177733 96.092 -77.110 49.845
177734 96.110 -77.106 49.785
177735 96.129 -77.104 49.728
177736 96.148 -77.102 49.673
177737 96.167 -77.099 49.622
177738 96.185 -77.097 49.574
177739 96.204 -77.094 49.530
177740 96.224 -77.084 49.488
177741 96.228 -77.104 49.452
177742 96.231 -77.125 49.418
177743 96.235 -77.146 49.385
177744 96.239 -77.166 49.355
177745 96.249 -77.181 49.325
177746 96.268 -77.185 49.296
177747 96.288 -77.191 49.269
177748 96.307 -77.196 49.245
177749 96.327 -77.201 49.224
177750 96.346 -77.206 49.203
177751 96.366 -77.210 49.183
177752 96.386 -77.216 49.165
177753 96.405 -77.221 49.148
177754 96.425 -77.226 49.131
177755 96.444 -77.231 49.119
177756 96.465 -77.232 49.107
177757 96.485 -77.229 49.095
177758 96.506 -77.228 49.081
177759 96.527 -77.226 49.065
177760 96.542 -77.243 49.049
177761 96.558 -77.259 49.032
177762 96.574 -77.276 49.017
177763 96.589 -77.291 49.002
177764 96.607 -77.303 48.985
177765 96.626 -77.313 48.968
177766 96.645 -77.323 48.950
177767 96.663 -77.333 48.931
177768 96.682 -77.344 48.913
177769 96.699 -77.355 48.896
177770 96.715 -77.369 48.879
177771 96.731 -77.382 48.863
177772 96.747 -77.395 48.847
177773 96.758 -77.424 48.832
177774 96.772 -77.454 48.816
177775 96.790 -77.474 48.801
177776 96.805 -77.479 48.786
177777 96.821 -77.484 48.770
177778 96.836 -77.489 48.754
177779 96.852 -77.495 48.738
177780 96.867 -77.500 48.721
177781 96.882 -77.504 48.701
177782 96.898 -77.510 48.681
177783 96.946 -77.502 48.662
177784 96.982 -77.517 48.644
177785 97.018 -77.532 48.626
177786 97.055 -77.546 48.607
177787 97.085 -77.570 48.587
177788 97.114 -77.594 48.566
177789 97.143 -77.619 48.546
177790 97.172 -77.644 48.526
177791 97.196 -77.661 48.507
177792 97.206 -77.649 48.488
177793 97.215 -77.638 48.470
177794 97.224 -77.627 48.452
177795 97.233 -77.614 48.433
177796 97.242 -77.603 48.411
177797 97.250 -77.589 48.390
177798 97.257 -77.573 48.369
177799 97.264 -77.558 48.348
177800 97.272 -77.543 48.327
177801 97.279 -77.528 48.306
177802 97.286 -77.513 48.285
177803 97.293 -77.498 48.265
177804 97.304 -77.484 48.246
177805 97.330 -77.489 48.226
177806 97.353 -77.498 48.205
177807 97.376 -77.506 48.184
177808 97.398 -77.514 48.163
177809 97.421 -77.522 48.142
177810 97.443 -77.530 48.121
177811 97.466 -77.538 48.097
177812 97.486 -77.549 48.073
177813 97.505 -77.561 48.048
177814 97.524 -77.574 48.022
177815 97.543 -77.586 47.996
177816 97.563 -77.599 47.969
177817 97.582 -77.611 47.941
177818 97.603 -77.620 47.912
177819 97.630 -77.618 47.883
177820 97.661 -77.618 47.854
177821 97.707 -77.624 47.824
177822 97.752 -77.631 47.795
177823 97.798 -77.637 47.766
177824 97.836 -77.641 47.737
177825 97.851 -77.637 47.708
177826 97.866 -77.633 47.678
177827 97.881 -77.629 47.649
177828 97.900 -77.628 47.621
177829 97.935 -77.637 47.592
177830 97.972 -77.649 47.564
177831 97.989 -77.639 47.536
177832 98.007 -77.630 47.507
177833 98.024 -77.620 47.480
177834 98.042 -77.610 47.451
177835 98.060 -77.600 47.422
177836 98.077 -77.590 47.394
177837 98.055 -77.576 47.365
177838 98.070 -77.528 47.336
177839 98.094 -77.473 47.308
177840 98.115 -77.467 47.279
177841 98.135 -77.467 47.249
177842 98.155 -77.467 47.219
177843 98.175 -77.466 47.189
177844 98.196 -77.466 47.163
177845 98.216 -77.465 47.136
177846 98.236 -77.465 47.109
177847 98.256 -77.464 47.083
177848 98.276 -77.464 47.057
177849 98.296 -77.464 47.032
177850 98.316 -77.463 47.006
177851 98.339 -77.470 46.980
177852 98.370 -77.507 46.954
177853 98.402 -77.543 46.927
177854 98.434 -77.580 46.900
177855 98.452 -77.607 46.875
177856 98.448 -77.617 46.849
177857 98.444 -77.628 46.824
177858 98.440 -77.638 46.801
177859 98.443 -77.639 46.780
177860 98.459 -77.638 46.757
177861 98.481 -77.637 46.733
177862 98.502 -77.636 46.710
177863 98.527 -77.632 46.687
177864 98.553 -77.627 46.665
177865 98.578 -77.620 46.643
177866 98.604 -77.615 46.621
177867 98.627 -77.619 46.598
177868 98.648 -77.624 46.576
177869 98.670 -77.630 46.553
177870 98.692 -77.635 46.531
177871 98.713 -77.640 46.510
177872 98.734 -77.645 46.490
177873 98.756 -77.650 46.470
177874 98.778 -77.644 46.451
177875 98.791 -77.645 46.431
177876 98.799 -77.652 46.411
177877 98.806 -77.658 46.390
177878 98.814 -77.665 46.370
177879 98.821 -77.672 46.351
177880 98.829 -77.680 46.331
177881 98.837 -77.686 46.311
177882 98.844 -77.693 46.290
177883 98.850 -77.699 46.268
177884 98.856 -77.708 46.246
177885 98.855 -77.712 46.225
177886 98.854 -77.716 46.203
177887 98.853 -77.720 46.180
177888 98.852 -77.724 46.158
177889 98.851 -77.727 46.136
177890 98.850 -77.731 46.114
177891 98.857 -77.735 46.092
177892 98.868 -77.738 46.073
177893 98.893 -77.746 46.053
177894 98.930 -77.758 46.033
177895 98.964 -77.767 46.012
177896 98.989 -77.766 45.994
177897 99.015 -77.765 45.977
177898 99.040 -77.765 45.960
177899 99.066 -77.765 45.945
177900 99.091 -77.765 45.931
177901 99.122 -77.769 45.918
177902 99.152 -77.774 45.905
177903 99.181 -77.780 45.892
177904 99.206 -77.793 45.881
177905 99.239 -77.812 45.871
177906 99.273 -77.840 45.860
177907 99.306 -77.868 45.851
177908 99.330 -77.893 45.842
177909 99.354 -77.917 45.833
177910 99.378 -77.940 45.823
177911 99.396 -77.962 45.815
177912 99.403 -77.978 45.807
177913 99.410 -77.994 45.799
177914 99.417 -78.010 45.790
177915 99.432 -78.022 45.783
177916 99.454 -78.032 45.775
177917 99.476 -78.041 45.766
177918 99.497 -78.050 45.756
177919 99.519 -78.060 45.746
177920 99.541 -78.069 45.735
177921 99.562 -78.079 45.724
177922 99.600 -78.086 45.713
177923 99.625 -78.100 45.702
177924 99.650 -78.094 45.691
177925 99.681 -78.080 45.679
177926 99.711 -78.066 45.667
177927 99.742 -78.053 45.654
177928 99.773 -78.044 45.640
177929 99.817 -78.080 45.625
177930 99.859 -78.112 45.610
177931 99.896 -78.128 45.595
177932 99.932 -78.142 45.580
177933 99.968 -78.157 45.565
177934 100.004 -78.172 45.550
177935 100.037 -78.195 45.534
177936 100.069 -78.225 45.519
177937 100.101 -78.255 45.504
177938 100.134 -78.265 45.490
177939 100.163 -78.265 45.475
177940 100.168 -78.282 45.459
177941 100.172 -78.298 45.444
177942 100.177 -78.314 45.428
177943 100.182 -78.330 45.412
177944 100.187 -78.342 45.397
177945 100.192 -78.354 45.380
177946 100.197 -78.366 45.364
177947 100.206 -78.376 45.347
177948 100.236 -78.381 45.331
177949 100.266 -78.386 45.315
177950 100.296 -78.391 45.300
177951 100.326 -78.395 45.285
177952 100.357 -78.396 45.269
177953 100.388 -78.397 45.254
177954 100.419 -78.398 45.239
177955 100.451 -78.399 45.225
177956 100.486 -78.401 45.210
177957 100.524 -78.404 45.194
177958 100.563 -78.410 45.178
177959 100.603 -78.416 45.162
177960 100.643 -78.423 45.146
177961 100.684 -78.429 45.130
177962 100.723 -78.435 45.113
177963 100.764 -78.442 45.096
177964 100.801 -78.448 45.079
177965 100.827 -78.454 45.063
177966 100.853 -78.460 45.047
177967 100.880 -78.466 45.032
177968 100.906 -78.471 45.018
177969 100.932 -78.477 45.004
177970 100.959 -78.482 44.989
177971 100.985 -78.489 44.974
177972 100.997 -78.483 44.958
177973 101.041 -78.485 44.944
177974 101.093 -78.467 44.928
177975 101.146 -78.456 44.912
177976 101.178 -78.472 44.896
177977 101.199 -78.488 44.880
177978 101.225 -78.505 44.864
177979 101.251 -78.522 44.848
177980 101.277 -78.539 44.832
177981 101.304 -78.556 44.816
177982 101.330 -78.573 44.800
177983 101.356 -78.591 44.783
177984 101.382 -78.608 44.766
177985 101.408 -78.630 44.751
177986 101.433 -78.654 44.736
177987 101.460 -78.677 44.720
177988 101.485 -78.698 44.705
177989 101.512 -78.720 44.690
177990 101.538 -78.741 44.675
177991 101.565 -78.762 44.660
177992 101.591 -78.783 44.644
177993 101.616 -78.810 44.629
177994 101.639 -78.843 44.614
177995 101.662 -78.876 44.600
177996 101.685 -78.908 44.586
177997 101.697 -78.909 44.572
177998 101.707 -78.902 44.558
177999 101.724 -78.907 44.544
178000 101.741 -78.913 44.530
178001 101.757 -78.918 44.517
178002 101.774 -78.924 44.505
178003 101.791 -78.930 44.493
178004 101.808 -78.935 44.480
178005 101.825 -78.941 44.467
178006 101.841 -78.946 44.455
178007 101.858 -78.952 44.443
178008 101.875 -78.958 44.431
178009 101.892 -78.963 44.419
178010 101.908 -78.969 44.408
178258 92.919 -76.972 54.090
178259 92.939 -76.964 54.102
178260 92.960 -76.955 54.112
178261 92.980 -76.947 54.123
178262 93.000 -76.937 54.131
178263 93.021 -76.929 54.140
178264 93.041 -76.920 54.150
178265 93.062 -76.911 54.159
178266 93.082 -76.903 54.168
178267 93.102 -76.893 54.175
178268 93.123 -76.885 54.181
178269 93.142 -76.881 54.188
178270 93.161 -76.878 54.195
178271 93.181 -76.875 54.200
178272 93.200 -76.871 54.204
178273 93.219 -76.868 54.208
178274 93.238 -76.865 54.212
178275 93.257 -76.861 54.214
178276 93.276 -76.858 54.217
178277 93.296 -76.855 54.218
178278 93.315 -76.851 54.220
178279 93.334 -76.848 54.221
178280 93.354 -76.845 54.220
178281 93.373 -76.842 54.221
178282 93.393 -76.839 54.222
178283 93.412 -76.836 54.225
178284 93.431 -76.833 54.231
178285 93.445 -76.825 54.238
178286 93.459 -76.818 54.247
178287 93.472 -76.810 54.255
178288 93.489 -76.806 54.263
178289 93.512 -76.806 54.270
178290 93.535 -76.806 54.276
178291 93.552 -76.802 54.279
178292 93.569 -76.797 54.280
178293 93.586 -76.793 54.280
178294 93.604 -76.789 54.281
178295 93.621 -76.785 54.282
178296 93.638 -76.781 54.281
178297 93.655 -76.776 54.279
178298 93.672 -76.771 54.275
178299 93.690 -76.763 54.268
178300 93.708 -76.755 54.260
178301 93.725 -76.747 54.252
178302 93.739 -76.740 54.243
178303 93.755 -76.735 54.235
178304 93.772 -76.730 54.225
178305 93.790 -76.725 54.212
178306 93.807 -76.721 54.197
178307 93.825 -76.716 54.183
178308 93.842 -76.712 54.172
178309 93.860 -76.706 54.163
178310 93.876 -76.705 54.156
178311 93.892 -76.706 54.148
178312 93.907 -76.707 54.139
178313 93.922 -76.708 54.129
178314 93.940 -76.706 54.120
178315 93.958 -76.705 54.111
178316 93.976 -76.703 54.104
178317 93.993 -76.702 54.099
178318 94.011 -76.700 54.098
178319 94.029 -76.699 54.098
178320 94.047 -76.697 54.099
178321 94.064 -76.696 54.102
178322 94.080 -76.695 54.105
178323 94.092 -76.695 54.106
178324 94.103 -76.693 54.105
178325 94.112 -76.691 54.104
178326 94.121 -76.688 54.095
178327 94.131 -76.685 54.085
178328 94.140 -76.683 54.074
178329 94.149 -76.680 54.061
178330 94.155 -76.678 54.049
178331 94.151 -76.680 54.037
178332 94.173 -76.674 54.024
178333 94.193 -76.668 54.006
178334 94.214 -76.661 53.988
178335 94.235 -76.655 53.969
178336 94.255 -76.649 53.950
178337 94.276 -76.643 53.931
178338 94.295 -76.636 53.909
178339 94.312 -76.630 53.884
178340 94.327 -76.622 53.855
178341 94.343 -76.615 53.824
178342 94.359 -76.608 53.792
178343 94.374 -76.601 53.759
178344 94.391 -76.594 53.725
178345 94.408 -76.586 53.690
178346 94.424 -76.577 53.654
178347 94.434 -76.573 53.615
178348 94.444 -76.567 53.576
178349 94.454 -76.563 53.538
178350 94.464 -76.558 53.501
178351 94.474 -76.553 53.466
178352 94.516 -76.547 53.431
178353 94.560 -76.541 53.393
178354 94.596 -76.541 53.354
178355 94.628 -76.542 53.317
178356 94.660 -76.544 53.283
178357 94.681 -76.540 53.251
178358 94.698 -76.535 53.221
178359 94.715 -76.530 53.197
178360 94.732 -76.525 53.172
178361 94.758 -76.518 53.148
178362 94.769 -76.514 53.121
178363 94.784 -76.520 53.095
178364 94.792 -76.521 53.068
178365 94.800 -76.523 53.041
178366 94.807 -76.524 53.014
178367 94.815 -76.526 52.984
178368 94.822 -76.528 52.956
178369 94.830 -76.530 52.931
178370 94.837 -76.531 52.904
178371 94.845 -76.533 52.873
178372 94.853 -76.534 52.838
178373 94.860 -76.536 52.799
178374 94.868 -76.538 52.756
178375 94.876 -76.539 52.710
178376 94.883 -76.541 52.662
178377 94.891 -76.542 52.611
178378 94.898 -76.544 52.560
178379 94.906 -76.546 52.508
178380 94.913 -76.548 52.455
178381 94.931 -76.552 52.400
178382 94.969 -76.559 52.346
178383 95.005 -76.567 52.294
178384 95.043 -76.574 52.243
178385 95.080 -76.582 52.191
178386 95.117 -76.589 52.140
178387 95.154 -76.597 52.088
178388 95.191 -76.605 52.036
178389 95.212 -76.616 51.985
178390 95.227 -76.629 51.936
178391 95.241 -76.643 51.887
178392 95.256 -76.658 51.839
178393 95.270 -76.671 51.791
178394 95.284 -76.686 51.742
178395 95.299 -76.700 51.694
178396 95.314 -76.715 51.649
178397 95.328 -76.728 51.605
178398 95.343 -76.743 51.561
178399 95.357 -76.757 51.517
178400 95.372 -76.771 51.476
178401 95.411 -76.781 51.435
178402 95.410 -76.787 51.394
178403 95.397 -76.789 51.354
178404 95.408 -76.791 51.312
178405 95.422 -76.797 51.269
178406 95.448 -76.815 51.226
178407 95.474 -76.831 51.184
178408 95.500 -76.848 51.147
178409 95.526 -76.865 51.110
178410 95.552 -76.882 51.072
178411 95.578 -76.900 51.037
178412 95.603 -76.916 50.999
178413 95.629 -76.934 50.960
178414 95.651 -76.946 50.919
178415 95.670 -76.954 50.875
178416 95.688 -76.962 50.832
178417 95.706 -76.971 50.789
178418 95.725 -76.980 50.745
178419 95.744 -76.988 50.701
178420 95.762 -76.997 50.655
178421 95.781 -77.005 50.607
178422 95.800 -77.013 50.563
178423 95.818 -77.022 50.516
178424 95.840 -77.031 50.467
178425 95.876 -77.038 50.415
178426 95.911 -77.045 50.360
178427 95.946 -77.052 50.303
178428 95.978 -77.047 50.244
178429 96.010 -77.041 50.183
178430 96.035 -77.044 50.117
178431 96.049 -77.059 50.050
178432 96.063 -77.074 49.983
178433 96.077 -77.089 49.917
178434 96.091 -77.104 49.854
178435 96.107 -77.113 49.795
178436 96.125 -77.110 49.739
178437 96.144 -77.108 49.686
178438 96.163 -77.106 49.636
178439 96.182 -77.103 49.589
178440 96.200 -77.100 49.545
178441 96.221 -77.088 49.504
178442 96.235 -77.089 49.468
178443 96.239 -77.110 49.434
178444 96.242 -77.131 49.401
178445 96.246 -77.151 49.371
178446 96.264 -77.158 49.342
178447 96.284 -77.163 49.313
178448 96.303 -77.168 49.287
178449 96.323 -77.172 49.263
178450 96.342 -77.178 49.241
178451 96.362 -77.182 49.221
178452 96.381 -77.188 49.202
178453 96.401 -77.193 49.184
178454 96.420 -77.197 49.167
178455 96.440 -77.203 49.151
178456 96.460 -77.207 49.138
178457 96.479 -77.213 49.126
178458 96.500 -77.215 49.114
178459 96.520 -77.213 49.100
178460 96.541 -77.211 49.085
178461 96.557 -77.226 49.069
178462 96.573 -77.241 49.053
178463 96.588 -77.258 49.038
178464 96.606 -77.269 49.022
178465 96.625 -77.279 49.005
178466 96.644 -77.289 48.988
178467 96.662 -77.300 48.970
178468 96.678 -77.313 48.951
178469 96.694 -77.326 48.932
178470 96.710 -77.340 48.913
178471 96.726 -77.353 48.896
178472 96.742 -77.367 48.880
178473 96.757 -77.379 48.863
178474 96.766 -77.420 48.846
178475 96.778 -77.448 48.829
178476 96.793 -77.453 48.813
178477 96.809 -77.458 48.797
178478 96.824 -77.464 48.780
178479 96.839 -77.469 48.763
178480 96.855 -77.473 48.745
178481 96.870 -77.479 48.727
178482 96.886 -77.484 48.708
178483 96.901 -77.489 48.688
178484 96.944 -77.483 48.669
178485 96.986 -77.492 48.652
178486 97.023 -77.506 48.633
178487 97.059 -77.520 48.615
178488 97.089 -77.545 48.596
178489 97.117 -77.570 48.574
178490 97.147 -77.595 48.554
178491 97.175 -77.620 48.534
178492 97.205 -77.644 48.514
178493 97.225 -77.653 48.493
178494 97.234 -77.642 48.474
178495 97.243 -77.630 48.455
178496 97.252 -77.618 48.435
178497 97.261 -77.606 48.415
178498 97.268 -77.591 48.395
178499 97.276 -77.576 48.375
178500 97.283 -77.561 48.356
178501 97.290 -77.545 48.336
178502 97.297 -77.530 48.315
178503 97.305 -77.515 48.295
178504 97.312 -77.500 48.274
178505 97.319 -77.485 48.255
178506 97.341 -77.483 48.235
178507 97.363 -77.492 48.215
178508 97.386 -77.499 48.193
178509 97.408 -77.508 48.172
178510 97.431 -77.516 48.151
178511 97.452 -77.526 48.129
178512 97.470 -77.539 48.106
178513 97.490 -77.551 48.082
178514 97.509 -77.564 48.057
178515 97.528 -77.576 48.031
178516 97.547 -77.588 48.005
178517 97.566 -77.601 47.978
178518 97.585 -77.613 47.950
178519 97.611 -77.613 47.921
178520 97.639 -77.611 47.892
178521 97.674 -77.613 47.862
178522 97.720 -77.619 47.832
178523 97.765 -77.626 47.802
178524 97.811 -77.632 47.772
178525 97.856 -77.639 47.743
178526 97.880 -77.637 47.712
178527 97.895 -77.633 47.682
178528 97.910 -77.629 47.653
178529 97.925 -77.625 47.624
178530 97.948 -77.626 47.595
178531 97.982 -77.634 47.567
178532 97.999 -77.624 47.539
178533 98.017 -77.614 47.511
178534 98.034 -77.605 47.482
178535 98.052 -77.595 47.454
178536 98.069 -77.585 47.426
178537 98.087 -77.575 47.398
178538 98.084 -77.563 47.370
178539 98.096 -77.516 47.342
178540 98.119 -77.491 47.314
178541 98.139 -77.491 47.285
178542 98.159 -77.490 47.256
178543 98.179 -77.489 47.227
178544 98.199 -77.489 47.198
178545 98.219 -77.489 47.170
178546 98.239 -77.489 47.143
178547 98.260 -77.488 47.116
178548 98.280 -77.488 47.089
178549 98.300 -77.487 47.063
178550 98.320 -77.487 47.038
178551 98.340 -77.486 47.013
178552 98.360 -77.486 46.987
178553 98.380 -77.486 46.960
178554 98.408 -77.508 46.934
178555 98.439 -77.544 46.908
178556 98.471 -77.580 46.883
178557 98.475 -77.597 46.859
178558 98.471 -77.608 46.835
178559 98.468 -77.614 46.812
178560 98.471 -77.615 46.790
178561 98.473 -77.615 46.767
178562 98.495 -77.614 46.744
178563 98.517 -77.614 46.721
178564 98.543 -77.608 46.698
178565 98.568 -77.602 46.676
178566 98.592 -77.603 46.653
178567 98.614 -77.608 46.631
178568 98.636 -77.614 46.608
178569 98.657 -77.619 46.585
178570 98.679 -77.624 46.561
178571 98.701 -77.630 46.539
178572 98.722 -77.635 46.517
178573 98.744 -77.639 46.497
178574 98.766 -77.639 46.477
178575 98.786 -77.632 46.457
178576 98.800 -77.632 46.438
178577 98.810 -77.636 46.418
178578 98.818 -77.642 46.398
178579 98.826 -77.650 46.378
178580 98.833 -77.657 46.358
178581 98.841 -77.664 46.339
178582 98.848 -77.671 46.319
178583 98.844 -77.673 46.298
178584 98.841 -77.677 46.276
178585 98.864 -77.697 46.254
178586 98.868 -77.704 46.233
178587 98.867 -77.708 46.211
178588 98.866 -77.711 46.189
178589 98.865 -77.715 46.167
178590 98.864 -77.719 46.144
178591 98.866 -77.723 46.121
178592 98.876 -77.726 46.098
178593 98.887 -77.729 46.077
178594 98.898 -77.732 46.057
178595 98.929 -77.743 46.037
178596 98.966 -77.755 46.017
178597 99.004 -77.768 45.998
178598 99.038 -77.777 45.981
178599 99.063 -77.776 45.964
178600 99.089 -77.775 45.949
178601 99.115 -77.775 45.934
178602 99.140 -77.774 45.921
178603 99.165 -77.774 45.909
178604 99.191 -77.773 45.898
178605 99.217 -77.787 45.886
178606 99.251 -77.815 45.874
178607 99.283 -77.843 45.863
178608 99.306 -77.867 45.853
178609 99.330 -77.891 45.844
178610 99.354 -77.915 45.834
178611 99.377 -77.939 45.825
178612 99.401 -77.963 45.817
178613 99.418 -77.984 45.809
178614 99.425 -78.000 45.801
178615 99.439 -78.012 45.792
178616 99.461 -78.022 45.784
178617 99.482 -78.031 45.776
178618 99.504 -78.040 45.767
178619 99.525 -78.050 45.758
178620 99.547 -78.059 45.747
178621 99.568 -78.069 45.736
178622 99.590 -78.078 45.725
178623 99.612 -78.087 45.714
178624 99.637 -78.085 45.703
178625 99.667 -78.072 45.693
178626 99.698 -78.058 45.681
178627 99.728 -78.045 45.669
178628 99.759 -78.031 45.656
178629 99.791 -78.028 45.642
178630 99.835 -78.063 45.628
178631 99.872 -78.081 45.613
178632 99.909 -78.096 45.599
178633 99.945 -78.111 45.585
178634 99.978 -78.137 45.569
178635 100.009 -78.166 45.555
178636 100.041 -78.197 45.541
178637 100.073 -78.226 45.526
178638 100.105 -78.243 45.510
178639 100.138 -78.251 45.495
178640 100.177 -78.243 45.480
178641 100.182 -78.260 45.465
178642 100.186 -78.276 45.450
178643 100.191 -78.293 45.434
178644 100.196 -78.309 45.417
178645 100.200 -78.326 45.401
178646 100.205 -78.342 45.385
178647 100.210 -78.356 45.368
178648 100.215 -78.368 45.352
178649 100.235 -78.376 45.336
178650 100.265 -78.380 45.321
178651 100.296 -78.382 45.305
178652 100.327 -78.383 45.290
178653 100.358 -78.384 45.275
178654 100.389 -78.385 45.260
178655 100.420 -78.386 45.244
178656 100.452 -78.387 45.229
178657 100.489 -78.390 45.214
178658 100.527 -78.393 45.198
178659 100.564 -78.396 45.182
178660 100.604 -78.403 45.166
178661 100.644 -78.410 45.150
178662 100.685 -78.416 45.133
178663 100.725 -78.423 45.116
178664 100.765 -78.429 45.099
178665 100.793 -78.435 45.083
178666 100.819 -78.441 45.067
178667 100.845 -78.446 45.051
178668 100.872 -78.451 45.037
178669 100.898 -78.457 45.023
178670 100.924 -78.463 45.009
178671 100.951 -78.469 44.994
178672 100.977 -78.474 44.979
178673 101.005 -78.480 44.964
178674 101.054 -78.465 44.950
178675 101.106 -78.447 44.934
178676 101.152 -78.445 44.917
178677 101.175 -78.461 44.901
178678 101.201 -78.478 44.886
178679 101.228 -78.495 44.870
178680 101.254 -78.512 44.854
178681 101.280 -78.529 44.837
178682 101.306 -78.547 44.820
178683 101.332 -78.564 44.804
178684 101.358 -78.581 44.788
178685 101.385 -78.598 44.770
178686 101.411 -78.620 44.755
178687 101.437 -78.642 44.739
178688 101.462 -78.665 44.724
178689 101.488 -78.688 44.709
178690 101.514 -78.711 44.694
178691 101.541 -78.732 44.679
178692 101.567 -78.753 44.664
178693 101.592 -78.780 44.648
178694 101.615 -78.813 44.633
178695 101.637 -78.846 44.618
178696 101.661 -78.879 44.604
178697 101.683 -78.912 44.590
178698 101.698 -78.921 44.576
178699 101.705 -78.906 44.561
178700 101.719 -78.906 44.547
178701 101.736 -78.912 44.533
178702 101.752 -78.917 44.521
178703 101.769 -78.923 44.509
178704 101.786 -78.928 44.496
178705 101.802 -78.934 44.482
178706 101.820 -78.940 44.469
178707 101.836 -78.946 44.457
178708 101.853 -78.951 44.445
178709 101.870 -78.956 44.433
178710 101.887 -78.962 44.422
178711 101.903 -78.968 44.410
178961 92.949 -76.931 54.092
178962 92.970 -76.923 54.103
178963 92.990 -76.914 54.112
178964 93.010 -76.905 54.122
178965 93.030 -76.897 54.132
178966 93.051 -76.887 54.143
178967 93.071 -76.879 54.151
178968 93.092 -76.870 54.160
178969 93.112 -76.862 54.168
178970 93.132 -76.857 54.174
178971 93.151 -76.854 54.181
178972 93.170 -76.850 54.187
178973 93.189 -76.847 54.192
178974 93.208 -76.844 54.196
178975 93.227 -76.840 54.200
178976 93.247 -76.837 54.202
178977 93.266 -76.834 54.204
178978 93.285 -76.831 54.206
178979 93.304 -76.828 54.208
178980 93.324 -76.824 54.209
178981 93.343 -76.821 54.210
178982 93.363 -76.818 54.211
178983 93.382 -76.815 54.212
178984 93.402 -76.812 54.215
178985 93.421 -76.809 54.219
178986 93.440 -76.806 54.224
178987 93.460 -76.803 54.232
178988 93.477 -76.799 54.240
178989 93.491 -76.791 54.248
178990 93.509 -76.787 54.257
178991 93.531 -76.787 54.264
178992 93.548 -76.782 54.268
178993 93.565 -76.778 54.269
178994 93.582 -76.774 54.271
178995 93.599 -76.769 54.273
178996 93.617 -76.765 54.275
178997 93.634 -76.760 54.275
178998 93.651 -76.755 54.273
178999 93.669 -76.751 54.270
179000 93.687 -76.743 54.263
179001 93.705 -76.736 54.256
179002 93.723 -76.727 54.248
179003 93.736 -76.721 54.241
179004 93.751 -76.714 54.234
179005 93.766 -76.708 54.226
179006 93.783 -76.703 54.215
179007 93.801 -76.699 54.200
179008 93.819 -76.694 54.185
179009 93.836 -76.690 54.174
179010 93.854 -76.684 54.165
179011 93.870 -76.683 54.159
179012 93.886 -76.684 54.153
179013 93.901 -76.686 54.146
179014 93.916 -76.687 54.137
179015 93.932 -76.687 54.129
179016 93.949 -76.685 54.122
179017 93.967 -76.683 54.116
179018 93.985 -76.682 54.110
179019 94.003 -76.680 54.108
179020 94.020 -76.679 54.106
179021 94.038 -76.677 54.106
179022 94.056 -76.676 54.108
179023 94.070 -76.677 54.112
179024 94.085 -76.677 54.115
179025 94.097 -76.677 54.117
179026 94.107 -76.676 54.117
179027 94.116 -76.675 54.111
179028 94.126 -76.674 54.103
179029 94.136 -76.671 54.094
179030 94.145 -76.668 54.082
179031 94.154 -76.665 54.071
179032 94.157 -76.665 54.060
179033 94.178 -76.658 54.048
179034 94.198 -76.652 54.031
179035 94.219 -76.646 54.012
179036 94.238 -76.640 53.994
179037 94.256 -76.633 53.977
179038 94.272 -76.627 53.958
179039 94.287 -76.621 53.937
179040 94.302 -76.615 53.912
179041 94.318 -76.608 53.883
179042 94.334 -76.601 53.853
179043 94.350 -76.594 53.821
179044 94.365 -76.586 53.788
179045 94.382 -76.579 53.754
179046 94.402 -76.568 53.721
179047 94.421 -76.558 53.686
179048 94.433 -76.552 53.646
179049 94.443 -76.547 53.606
179050 94.454 -76.542 53.567
179051 94.464 -76.538 53.528
179052 94.478 -76.533 53.490
179053 94.522 -76.527 53.452
179054 94.563 -76.523 53.413
179055 94.596 -76.524 53.372
179056 94.628 -76.526 53.333
179057 94.660 -76.527 53.298
179058 94.681 -76.524 53.266
179059 94.698 -76.518 53.238
179060 94.715 -76.514 53.213
179061 94.733 -76.508 53.189
179062 94.750 -76.504 53.164
179063 94.752 -76.505 53.138
179064 94.755 -76.507 53.110
179065 94.763 -76.508 53.082
179066 94.770 -76.510 53.054
179067 94.778 -76.511 53.026
179068 94.785 -76.513 52.999
179069 94.793 -76.515 52.972
179070 94.801 -76.517 52.948
179071 94.808 -76.518 52.921
179072 94.816 -76.520 52.891
179073 94.823 -76.521 52.857
179074 94.831 -76.523 52.820
179075 94.838 -76.524 52.777
179076 94.846 -76.527 52.731
179077 94.853 -76.528 52.682
179078 94.861 -76.530 52.631
179079 94.868 -76.531 52.579
179080 94.876 -76.533 52.528
179081 94.884 -76.534 52.476
179082 94.902 -76.538 52.423
179083 94.939 -76.546 52.369
179084 94.977 -76.554 52.315
179085 95.014 -76.561 52.264
179086 95.051 -76.569 52.211
179087 95.088 -76.577 52.160
179088 95.125 -76.584 52.109
179089 95.166 -76.593 52.056
179090 95.207 -76.605 52.006
179091 95.221 -76.618 51.955
179092 95.236 -76.633 51.905
179093 95.250 -76.647 51.856
179094 95.265 -76.661 51.807
179095 95.279 -76.675 51.759
179096 95.294 -76.690 51.711
179097 95.308 -76.704 51.665
179098 95.323 -76.718 51.619
179099 95.337 -76.733 51.574
179100 95.352 -76.746 51.531
179101 95.373 -76.760 51.489
179102 95.414 -76.770 51.447
179103 95.433 -76.777 51.406
179104 95.416 -76.780 51.367
179105 95.398 -76.782 51.327
179106 95.410 -76.787 51.285
179107 95.436 -76.804 51.244
179108 95.462 -76.821 51.205
179109 95.488 -76.837 51.169
179110 95.514 -76.855 51.133
179111 95.540 -76.872 51.097
179112 95.566 -76.890 51.061
179113 95.593 -76.907 51.023
179114 95.619 -76.925 50.984
179115 95.645 -76.940 50.944
179116 95.670 -76.947 50.901
179117 95.692 -76.954 50.857
179118 95.711 -76.962 50.812
179119 95.730 -76.971 50.767
179120 95.748 -76.979 50.722
179121 95.766 -76.987 50.674
179122 95.785 -76.997 50.627
179123 95.804 -77.005 50.583
179124 95.822 -77.013 50.535
179125 95.847 -77.022 50.484
179126 95.883 -77.029 50.429
179127 95.918 -77.037 50.373
179128 95.950 -77.031 50.314
179129 95.982 -77.027 50.254
179130 96.014 -77.022 50.193
179131 96.038 -77.025 50.126
179132 96.052 -77.041 50.059
179133 96.066 -77.056 49.991
179134 96.080 -77.071 49.927
179135 96.094 -77.085 49.866
179136 96.108 -77.100 49.807
179137 96.122 -77.116 49.752
179138 96.141 -77.114 49.701
179139 96.159 -77.112 49.652
179140 96.178 -77.110 49.605
179141 96.197 -77.107 49.560
179142 96.218 -77.094 49.520
179143 96.239 -77.081 49.485
179144 96.246 -77.095 49.451
179145 96.249 -77.116 49.418
179146 96.260 -77.130 49.389
179147 96.279 -77.135 49.359
179148 96.299 -77.140 49.330
179149 96.318 -77.144 49.305
179150 96.338 -77.150 49.282
179151 96.358 -77.154 49.261
179152 96.377 -77.160 49.240
179153 96.397 -77.165 49.221
179154 96.416 -77.169 49.204
179155 96.436 -77.175 49.188
179156 96.456 -77.179 49.171
179157 96.475 -77.185 49.158
179158 96.495 -77.189 49.146
179159 96.514 -77.194 49.133
179160 96.534 -77.197 49.120
179161 96.555 -77.196 49.106
179162 96.572 -77.208 49.091
179163 96.587 -77.224 49.075
179164 96.606 -77.235 49.059
179165 96.624 -77.244 49.042
179166 96.641 -77.257 49.026
179167 96.657 -77.270 49.009
179168 96.673 -77.284 48.991
179169 96.689 -77.298 48.972
179170 96.704 -77.310 48.952
179171 96.721 -77.324 48.932
179172 96.736 -77.338 48.914
179173 96.752 -77.351 48.897
179174 96.766 -77.373 48.879
179175 96.836 -77.404 48.860
179176 96.802 -77.422 48.842
179177 96.797 -77.432 48.825
179178 96.812 -77.438 48.808
179179 96.827 -77.442 48.789
179180 96.843 -77.448 48.770
179181 96.858 -77.453 48.752
179182 96.874 -77.458 48.733
179183 96.889 -77.463 48.714
179184 96.904 -77.468 48.694
179185 96.939 -77.465 48.675
179186 96.991 -77.467 48.658
179187 97.027 -77.481 48.640
179188 97.063 -77.495 48.622
179189 97.092 -77.520 48.604
179190 97.121 -77.545 48.583
179191 97.150 -77.570 48.563
179192 97.179 -77.595 48.542
179193 97.209 -77.620 48.521
179194 97.237 -77.645 48.500
179195 97.253 -77.645 48.479
179196 97.262 -77.633 48.460
179197 97.272 -77.622 48.439
179198 97.280 -77.608 48.419
179199 97.287 -77.593 48.399
179200 97.294 -77.578 48.380
179201 97.301 -77.563 48.362
179202 97.309 -77.548 48.344
179203 97.316 -77.533 48.324
179204 97.323 -77.517 48.304
179205 97.330 -77.502 48.283
179206 97.337 -77.487 48.264
179207 97.351 -77.477 48.245
179208 97.374 -77.485 48.225
179209 97.396 -77.493 48.203
179210 97.417 -77.503 48.181
179211 97.436 -77.516 48.160
179212 97.455 -77.528 48.138
179213 97.474 -77.540 48.115
179214 97.494 -77.553 48.090
179215 97.513 -77.565 48.064
179216 97.531 -77.577 48.038
179217 97.550 -77.590 48.012
179218 97.570 -77.602 47.984
179219 97.593 -77.607 47.956
179220 97.620 -77.605 47.926
179221 97.648 -77.604 47.896
179222 97.687 -77.608 47.866
179223 97.733 -77.614 47.836
179224 97.779 -77.620 47.806
179225 97.824 -77.627 47.775
179226 97.870 -77.633 47.745
179227 97.910 -77.637 47.714
179228 97.924 -77.633 47.684
179229 97.939 -77.630 47.655
179230 97.954 -77.625 47.626
179231 97.969 -77.621 47.597
179232 97.991 -77.619 47.569
179233 98.009 -77.609 47.540
179234 98.026 -77.599 47.512
179235 98.044 -77.589 47.484
179236 98.061 -77.580 47.456
179237 98.079 -77.570 47.429
179238 98.096 -77.559 47.401
179239 98.113 -77.549 47.374
179240 98.122 -77.514 47.346
179241 98.142 -77.514 47.319
179242 98.162 -77.514 47.290
179243 98.182 -77.514 47.262
179244 98.203 -77.513 47.234
179245 98.223 -77.513 47.205
179246 98.243 -77.512 47.177
179247 98.263 -77.511 47.150
179248 98.283 -77.511 47.123
179249 98.303 -77.511 47.095
179250 98.323 -77.511 47.069
179251 98.344 -77.510 47.044
179252 98.364 -77.510 47.019
179253 98.384 -77.509 46.993
179254 98.404 -77.509 46.967
179255 98.424 -77.508 46.942
179256 98.444 -77.508 46.916
179257 98.476 -77.545 46.891
179258 98.501 -77.577 46.868
179259 98.497 -77.587 46.846
179260 98.497 -77.592 46.824
179261 98.499 -77.592 46.800
179262 98.501 -77.592 46.777
179263 98.509 -77.592 46.754
179264 98.532 -77.589 46.732
179265 98.557 -77.586 46.709
179266 98.579 -77.592 46.686
179267 98.601 -77.598 46.663
179268 98.623 -77.603 46.640
179269 98.645 -77.608 46.617
179270 98.667 -77.614 46.594
179271 98.689 -77.619 46.569
179272 98.710 -77.624 46.546
179273 98.732 -77.630 46.525
179274 98.753 -77.634 46.505
179275 98.777 -77.624 46.484
179276 98.795 -77.619 46.464
179277 98.808 -77.620 46.444
179278 98.822 -77.620 46.425
179279 98.830 -77.627 46.405
179280 98.837 -77.634 46.385
179281 98.845 -77.641 46.365
179282 98.844 -77.645 46.345
179283 98.838 -77.647 46.325
179284 98.831 -77.649 46.304
179285 98.842 -77.661 46.283
179286 98.865 -77.680 46.260
179287 98.881 -77.696 46.239
179288 98.880 -77.699 46.217
179289 98.879 -77.703 46.195
179290 98.878 -77.707 46.173
179291 98.878 -77.711 46.150
179292 98.885 -77.714 46.127
179293 98.896 -77.718 46.104
179294 98.906 -77.721 46.083
179295 98.917 -77.724 46.063
179296 98.928 -77.727 46.043
179297 98.964 -77.739 46.023
179298 99.002 -77.752 46.004
179299 99.040 -77.764 45.986
179300 99.077 -77.777 45.969
179301 99.112 -77.787 45.954
179302 99.138 -77.786 45.939
179303 99.163 -77.785 45.925
179304 99.189 -77.785 45.913
179305 99.206 -77.778 45.901
179306 99.228 -77.790 45.889
179307 99.259 -77.817 45.877
179308 99.283 -77.841 45.866
179309 99.306 -77.865 45.855
179310 99.330 -77.889 45.845
179311 99.354 -77.913 45.836
179312 99.377 -77.937 45.827
179313 99.401 -77.961 45.819
179314 99.425 -77.985 45.811
179315 99.445 -78.003 45.802
179316 99.467 -78.012 45.793
179317 99.487 -78.020 45.785
179318 99.508 -78.028 45.777
179319 99.529 -78.037 45.769
179320 99.550 -78.045 45.759
179321 99.571 -78.053 45.748
179322 99.592 -78.062 45.737
179323 99.613 -78.071 45.726
179324 99.635 -78.077 45.715
179325 99.661 -78.063 45.705
179326 99.688 -78.050 45.694
179327 99.715 -78.036 45.682
179328 99.745 -78.023 45.671
179329 99.776 -78.009 45.658
179330 99.810 -78.011 45.645
179331 99.849 -78.034 45.631
179332 99.886 -78.050 45.618
179333 99.918 -78.078 45.604
179334 99.949 -78.108 45.590
179335 99.981 -78.138 45.575
179336 100.012 -78.168 45.561
179337 100.044 -78.197 45.546
179338 100.076 -78.220 45.531
179339 100.109 -78.229 45.516
179340 100.142 -78.238 45.501
179341 100.179 -78.237 45.486
179342 100.196 -78.238 45.470
179343 100.200 -78.254 45.455
179344 100.205 -78.271 45.438
179345 100.209 -78.287 45.421
179346 100.214 -78.304 45.404
179347 100.219 -78.320 45.389
179348 100.223 -78.337 45.373
179349 100.228 -78.353 45.357
179350 100.235 -78.368 45.342
179351 100.266 -78.369 45.327
179352 100.297 -78.370 45.311
179353 100.328 -78.371 45.295
179354 100.359 -78.372 45.280
179355 100.390 -78.373 45.265
179356 100.421 -78.374 45.249
179357 100.453 -78.375 45.233
179358 100.491 -78.379 45.217
179359 100.528 -78.382 45.201
179360 100.566 -78.385 45.186
179361 100.606 -78.389 45.169
179362 100.645 -78.396 45.152
179363 100.686 -78.403 45.135
179364 100.726 -78.409 45.117
179365 100.758 -78.415 45.101
179366 100.785 -78.421 45.086
179367 100.811 -78.426 45.071
179368 100.837 -78.432 45.056
179369 100.864 -78.438 45.041
179370 100.890 -78.444 45.027
179371 100.916 -78.449 45.013
179372 100.954 -78.456 44.998
179373 100.993 -78.462 44.984
179374 101.025 -78.469 44.969
179375 101.071 -78.450 44.953
179376 101.119 -78.428 44.938
179377 101.151 -78.434 44.922
179378 101.178 -78.451 44.906
179379 101.204 -78.468 44.891
179380 101.230 -78.485 44.874
179381 101.256 -78.503 44.857
179382 101.282 -78.520 44.841
179383 101.309 -78.537 44.825
179384 101.335 -78.554 44.808
179385 101.361 -78.571 44.791
179386 101.388 -78.589 44.775
179387 101.413 -78.609 44.758
179388 101.440 -78.632 44.743
179389 101.465 -78.655 44.727
179390 101.491 -78.677 44.712
179391 101.517 -78.700 44.697
179392 101.543 -78.723 44.682
179393 101.567 -78.751 44.667
179394 101.590 -78.783 44.652
179395 101.613 -78.817 44.637
179396 101.636 -78.849 44.622
179397 101.659 -78.883 44.608
179398 101.682 -78.915 44.593
179399 101.699 -78.931 44.579
179400 101.706 -78.918 44.565
179401 101.714 -78.905 44.552
179402 101.731 -78.911 44.536
179403 101.747 -78.916 44.524
179404 101.764 -78.921 44.512
179405 101.781 -78.927 44.499
179406 101.798 -78.933 44.486
179407 101.815 -78.939 44.473
179408 101.831 -78.944 44.461
179409 101.848 -78.949 44.448
179410 101.865 -78.955 44.436
179411 101.882 -78.961 44.425
179412 101.898 -78.967 44.412
179663 92.959 -76.899 54.082
179664 92.980 -76.890 54.093
179665 93.000 -76.881 54.104
179666 93.020 -76.873 54.115
179667 93.041 -76.864 54.126
179668 93.061 -76.855 54.135
179669 93.081 -76.846 54.145
179670 93.101 -76.837 54.154
179671 93.121 -76.834 54.161
179672 93.140 -76.830 54.168
179673 93.159 -76.827 54.175
179674 93.179 -76.824 54.181
179675 93.198 -76.820 54.184
179676 93.217 -76.817 54.188
179677 93.236 -76.814 54.191
179678 93.256 -76.810 54.193
179679 93.275 -76.807 54.196
179680 93.294 -76.804 54.198
179681 93.313 -76.801 54.199
179682 93.333 -76.798 54.201
179683 93.352 -76.795 54.202
179684 93.372 -76.792 54.203
179685 93.391 -76.789 54.205
179686 93.411 -76.786 54.209
179687 93.430 -76.783 54.212
179688 93.449 -76.780 54.218
179689 93.469 -76.777 54.225
179690 93.489 -76.774 54.232
179691 93.508 -76.771 54.241
179692 93.526 -76.766 54.249
179693 93.544 -76.762 54.254
179694 93.561 -76.757 54.257
179695 93.579 -76.752 54.260
179696 93.596 -76.748 54.262
179697 93.614 -76.743 54.265
179698 93.631 -76.739 54.264
179699 93.649 -76.734 54.264
179700 93.666 -76.729 54.261
179701 93.684 -76.724 54.256
179702 93.702 -76.716 54.249
179703 93.721 -76.708 54.242
179704 93.735 -76.701 54.236
179705 93.749 -76.694 54.232
179706 93.763 -76.688 54.225
179707 93.777 -76.682 54.216
179708 93.795 -76.677 54.202
179709 93.813 -76.672 54.187
179710 93.830 -76.668 54.176
179711 93.848 -76.662 54.167
179712 93.864 -76.661 54.161
179713 93.879 -76.663 54.158
179714 93.895 -76.664 54.152
179715 93.910 -76.665 54.144
179716 93.925 -76.666 54.138
179717 93.941 -76.666 54.132
179718 93.959 -76.665 54.127
179719 93.977 -76.663 54.122
179720 93.994 -76.661 54.118
179721 94.012 -76.660 54.115
179722 94.030 -76.658 54.113
179723 94.046 -76.658 54.114
179724 94.060 -76.658 54.119
179725 94.075 -76.659 54.124
179726 94.089 -76.660 54.128
179727 94.101 -76.659 54.130
179728 94.111 -76.658 54.126
179729 94.120 -76.657 54.119
179730 94.130 -76.656 54.112
179731 94.140 -76.655 54.102
179732 94.150 -76.653 54.092
179733 94.163 -76.649 54.081
179734 94.181 -76.643 54.071
179735 94.198 -76.636 54.053
179736 94.215 -76.630 54.035
179737 94.232 -76.624 54.017
179738 94.250 -76.618 54.000
179739 94.264 -76.611 53.982
179740 94.279 -76.606 53.963
179741 94.293 -76.600 53.940
179742 94.309 -76.593 53.913
179743 94.325 -76.586 53.883
179744 94.341 -76.579 53.853
179745 94.356 -76.572 53.821
179746 94.376 -76.562 53.787
179747 94.396 -76.552 53.754
179748 94.416 -76.541 53.719
179749 94.433 -76.532 53.679
179750 94.443 -76.527 53.637
179751 94.453 -76.522 53.596
179752 94.463 -76.517 53.556
179753 94.484 -76.512 53.515
179754 94.529 -76.506 53.475
179755 94.564 -76.505 53.434
179756 94.596 -76.507 53.391
179757 94.628 -76.508 53.352
179758 94.661 -76.510 53.316
179759 94.682 -76.507 53.284
179760 94.699 -76.502 53.255
179761 94.716 -76.497 53.230
179762 94.734 -76.492 53.205
179763 94.752 -76.489 53.180
179764 94.758 -76.489 53.154
179765 94.754 -76.491 53.126
179766 94.749 -76.494 53.098
179767 94.745 -76.496 53.069
179768 94.749 -76.499 53.040
179769 94.756 -76.500 53.013
179770 94.764 -76.502 52.988
179771 94.771 -76.504 52.963
179772 94.779 -76.505 52.938
179773 94.786 -76.507 52.908
179774 94.794 -76.508 52.875
179775 94.802 -76.510 52.838
179776 94.809 -76.511 52.795
179777 94.817 -76.514 52.749
179778 94.824 -76.515 52.700
179779 94.832 -76.517 52.648
179780 94.839 -76.518 52.597
179781 94.847 -76.520 52.547
179782 94.854 -76.521 52.497
179783 94.873 -76.525 52.445
179784 94.910 -76.533 52.391
179785 94.948 -76.541 52.337
179786 94.984 -76.549 52.283
179787 95.022 -76.556 52.231
179788 95.059 -76.564 52.180
179789 95.102 -76.574 52.128
179790 95.153 -76.586 52.075
179791 95.195 -76.596 52.024
179792 95.216 -76.608 51.973
179793 95.230 -76.623 51.921
179794 95.245 -76.636 51.872
179795 95.259 -76.651 51.823
179796 95.274 -76.665 51.775
179797 95.288 -76.680 51.727
179798 95.303 -76.693 51.680
179799 95.317 -76.708 51.634
179800 95.332 -76.722 51.589
179801 95.346 -76.737 51.545
179802 95.377 -76.748 51.502
179803 95.418 -76.758 51.460
179804 95.457 -76.768 51.418
179805 95.439 -76.770 51.379
179806 95.421 -76.773 51.339
179807 95.404 -76.777 51.300
179808 95.425 -76.794 51.260
179809 95.451 -76.812 51.224
179810 95.478 -76.829 51.188
179811 95.504 -76.846 51.153
179812 95.530 -76.864 51.118
179813 95.556 -76.882 51.082
179814 95.583 -76.900 51.045
179815 95.609 -76.917 51.006
179816 95.635 -76.934 50.967
179817 95.661 -76.941 50.925
179818 95.686 -76.948 50.880
179819 95.711 -76.955 50.835
179820 95.734 -76.962 50.789
179821 95.752 -76.971 50.742
179822 95.771 -76.979 50.694
179823 95.790 -76.987 50.647
179824 95.808 -76.996 50.602
179825 95.827 -77.005 50.551
179826 95.855 -77.012 50.498
179827 95.890 -77.020 50.443
179828 95.923 -77.017 50.385
179829 95.955 -77.012 50.325
179830 95.986 -77.006 50.264
179831 96.018 -77.001 50.203
179832 96.040 -77.007 50.136
179833 96.054 -77.022 50.070
179834 96.069 -77.037 50.004
179835 96.083 -77.052 49.941
179836 96.097 -77.067 49.880
179837 96.111 -77.082 49.822
179838 96.125 -77.097 49.768
179839 96.139 -77.112 49.716
179840 96.156 -77.118 49.668
179841 96.174 -77.116 49.622
179842 96.193 -77.113 49.577
179843 96.214 -77.100 49.538
179844 96.235 -77.087 49.503
179845 96.253 -77.080 49.470
179846 96.256 -77.101 49.436
179847 96.275 -77.106 49.406
179848 96.295 -77.112 49.377
179849 96.314 -77.116 49.350
179850 96.334 -77.122 49.324
179851 96.354 -77.127 49.302
179852 96.373 -77.132 49.281
179853 96.393 -77.137 49.260
179854 96.412 -77.141 49.243
179855 96.432 -77.147 49.225
179856 96.451 -77.151 49.208
179857 96.471 -77.157 49.192
179858 96.490 -77.162 49.179
179859 96.510 -77.166 49.166
179860 96.530 -77.172 49.153
179861 96.549 -77.176 49.141
179862 96.569 -77.181 49.128
179863 96.587 -77.190 49.113
179864 96.604 -77.201 49.096
179865 96.620 -77.215 49.080
179866 96.636 -77.229 49.064
179867 96.652 -77.241 49.048
179868 96.668 -77.255 49.030
179869 96.683 -77.269 49.012
179870 96.699 -77.282 48.993
179871 96.715 -77.295 48.972
179872 96.731 -77.309 48.951
179873 96.746 -77.322 48.933
179874 96.762 -77.335 48.914
179875 96.879 -77.351 48.895
179876 96.892 -77.370 48.876
179877 96.858 -77.388 48.857
179878 96.824 -77.405 48.838
179879 96.815 -77.417 48.820
179880 96.831 -77.422 48.800
179881 96.847 -77.427 48.781
179882 96.862 -77.432 48.761
179883 96.877 -77.437 48.740
179884 96.892 -77.442 48.721
179885 96.908 -77.448 48.702
179886 96.935 -77.448 48.683
179887 96.995 -77.441 48.664
179888 97.031 -77.455 48.646
179889 97.067 -77.471 48.628
179890 97.096 -77.496 48.609
179891 97.125 -77.520 48.590
179892 97.154 -77.545 48.570
179893 97.183 -77.570 48.550
179894 97.212 -77.595 48.529
179895 97.241 -77.620 48.507
179896 97.270 -77.645 48.486
179897 97.282 -77.637 48.466
179898 97.291 -77.626 48.446
179899 97.298 -77.611 48.425
179900 97.305 -77.596 48.405
179901 97.313 -77.581 48.386
179902 97.320 -77.566 48.368
179903 97.327 -77.551 48.351
179904 97.334 -77.536 48.332
179905 97.342 -77.520 48.313
179906 97.349 -77.505 48.292
179907 97.356 -77.490 48.274
179908 97.363 -77.475 48.255
179909 97.382 -77.480 48.235
179910 97.401 -77.492 48.213
179911 97.421 -77.505 48.191
179912 97.440 -77.517 48.169
179913 97.459 -77.530 48.146
179914 97.478 -77.542 48.122
179915 97.497 -77.555 48.097
179916 97.516 -77.567 48.071
179917 97.535 -77.580 48.046
179918 97.555 -77.592 48.019
179919 97.575 -77.602 47.991
179920 97.602 -77.600 47.961
179921 97.629 -77.598 47.931
179922 97.657 -77.596 47.901
179923 97.701 -77.602 47.870
179924 97.746 -77.608 47.839
179925 97.792 -77.615 47.809
179926 97.837 -77.621 47.778
179927 97.883 -77.628 47.747
179928 97.928 -77.634 47.716
179929 97.954 -77.633 47.686
179930 97.969 -77.630 47.657
179931 97.984 -77.625 47.628
179932 98.000 -77.621 47.599
179933 98.020 -77.614 47.570
179934 98.035 -77.602 47.542
179935 98.051 -77.592 47.514
179936 98.067 -77.580 47.486
179937 98.082 -77.570 47.459
179938 98.098 -77.559 47.432
179939 98.113 -77.548 47.405
179940 98.126 -77.536 47.377
179941 98.146 -77.533 47.350
179942 98.166 -77.534 47.323
179943 98.186 -77.536 47.295
179944 98.206 -77.536 47.267
179945 98.226 -77.536 47.240
179946 98.246 -77.536 47.212
179947 98.266 -77.536 47.185
179948 98.286 -77.535 47.158
179949 98.306 -77.535 47.131
179950 98.326 -77.534 47.102
179951 98.347 -77.534 47.076
179952 98.367 -77.533 47.050
179953 98.387 -77.533 47.025
179954 98.407 -77.533 47.000
179955 98.427 -77.532 46.975
179956 98.447 -77.532 46.950
179957 98.467 -77.531 46.925
179958 98.488 -77.531 46.901
179959 98.513 -77.545 46.877
179960 98.523 -77.567 46.855
179961 98.524 -77.568 46.835
179962 98.527 -77.569 46.812
179963 98.529 -77.569 46.788
179964 98.530 -77.572 46.764
179965 98.545 -77.576 46.742
179966 98.566 -77.581 46.719
179967 98.588 -77.586 46.696
179968 98.610 -77.592 46.673
179969 98.632 -77.597 46.649
179970 98.654 -77.602 46.626
179971 98.676 -77.608 46.602
179972 98.698 -77.614 46.578
179973 98.720 -77.619 46.554
179974 98.741 -77.624 46.533
179975 98.764 -77.620 46.512
179976 98.788 -77.608 46.493
179977 98.803 -77.607 46.473
179978 98.817 -77.607 46.453
179979 98.831 -77.607 46.432
179980 98.841 -77.611 46.411
179981 98.845 -77.617 46.391
179982 98.838 -77.619 46.371
179983 98.831 -77.620 46.351
179984 98.825 -77.622 46.331
179985 98.819 -77.625 46.310
179986 98.843 -77.645 46.289
179987 98.867 -77.664 46.267
179988 98.890 -77.684 46.246
179989 98.894 -77.691 46.224
179990 98.893 -77.695 46.202
179991 98.892 -77.699 46.180
179992 98.894 -77.702 46.157
179993 98.904 -77.705 46.134
179994 98.915 -77.708 46.111
179995 98.925 -77.711 46.090
179996 98.936 -77.714 46.069
179997 98.947 -77.718 46.049
179998 98.962 -77.723 46.029
179999 99.000 -77.735 46.010
180000 99.038 -77.748 45.991
180001 99.075 -77.761 45.974
180002 99.113 -77.773 45.958
180003 99.150 -77.786 45.943
180004 99.187 -77.796 45.929
180005 99.211 -77.796 45.916
180006 99.216 -77.780 45.903
180007 99.235 -77.791 45.890
180008 99.259 -77.815 45.879
180009 99.282 -77.838 45.868
180010 99.304 -77.861 45.857
180011 99.327 -77.884 45.847
180012 99.350 -77.906 45.837
180013 99.373 -77.930 45.828
180014 99.396 -77.953 45.819
180015 99.427 -77.973 45.811
180016 99.466 -77.993 45.804
180017 99.489 -78.003 45.795
180018 99.510 -78.011 45.787
180019 99.531 -78.019 45.779
180020 99.552 -78.028 45.770
180021 99.573 -78.037 45.760
180022 99.593 -78.045 45.749
180023 99.615 -78.053 45.738
180024 99.636 -78.062 45.727
180025 99.660 -78.056 45.716
180026 99.687 -78.042 45.705
180027 99.714 -78.028 45.695
180028 99.740 -78.015 45.684
180029 99.767 -78.001 45.672
180030 99.793 -77.987 45.661
180031 99.826 -77.989 45.648
180032 99.857 -78.019 45.635
180033 99.889 -78.049 45.622
180034 99.921 -78.079 45.609
180035 99.952 -78.109 45.595
180036 99.984 -78.139 45.580
180037 100.015 -78.169 45.566
180038 100.047 -78.198 45.551
180039 100.080 -78.206 45.536
180040 100.113 -78.215 45.521
180041 100.146 -78.224 45.506
180042 100.179 -78.232 45.490
180043 100.210 -78.216 45.474
180044 100.214 -78.232 45.459
180045 100.219 -78.249 45.443
180046 100.223 -78.266 45.426
180047 100.228 -78.282 45.409
180048 100.233 -78.298 45.394
180049 100.238 -78.315 45.379
180050 100.242 -78.331 45.363
180051 100.249 -78.348 45.348
180052 100.267 -78.357 45.332
180053 100.298 -78.358 45.317
180054 100.329 -78.359 45.300
180055 100.360 -78.360 45.285
180056 100.392 -78.361 45.269
180057 100.422 -78.362 45.253
180058 100.456 -78.363 45.236
180059 100.493 -78.366 45.221
180060 100.531 -78.370 45.205
180061 100.569 -78.373 45.188
180062 100.607 -78.376 45.172
180063 100.647 -78.383 45.155
180064 100.687 -78.389 45.138
180065 100.724 -78.396 45.121
180066 100.750 -78.401 45.106
180067 100.777 -78.407 45.091
180068 100.803 -78.413 45.076
180069 100.829 -78.418 45.062
180070 100.862 -78.425 45.047
180071 100.902 -78.431 45.033
180072 100.941 -78.438 45.018
180073 100.980 -78.445 45.003
180074 101.018 -78.451 44.988
180075 101.045 -78.458 44.973
180076 101.092 -78.438 44.957
180077 101.137 -78.418 44.942
180078 101.152 -78.426 44.926
180079 101.179 -78.442 44.910
180080 101.207 -78.459 44.894
180081 101.233 -78.476 44.877
180082 101.259 -78.493 44.860
180083 101.285 -78.510 44.844
180084 101.312 -78.527 44.828
180085 101.338 -78.545 44.811
180086 101.364 -78.561 44.794
180087 101.390 -78.579 44.778
180088 101.416 -78.598 44.761
180089 101.442 -78.620 44.745
180090 101.468 -78.643 44.731
180091 101.494 -78.666 44.716
180092 101.519 -78.689 44.700
180093 101.544 -78.714 44.685
180094 101.567 -78.745 44.670
180095 101.590 -78.779 44.655
180096 101.613 -78.811 44.639
180097 101.636 -78.844 44.624
180098 101.659 -78.877 44.610
180099 101.682 -78.910 44.596
180100 101.694 -78.915 44.581
180101 101.705 -78.918 44.568
180102 101.714 -78.915 44.555
180103 101.726 -78.909 44.539
180104 101.742 -78.915 44.526
180105 101.759 -78.921 44.514
180106 101.776 -78.926 44.501
180107 101.793 -78.932 44.488
180108 101.810 -78.937 44.476
180109 101.826 -78.943 44.463
180110 101.843 -78.949 44.451
180111 101.859 -78.954 44.439
180112 101.877 -78.960 44.427
180113 101.893 -78.965 44.414
180114 101.911 -78.971 44.403
180365 92.969 -76.867 54.072
180366 92.990 -76.858 54.085
180367 93.010 -76.849 54.097
180368 93.030 -76.840 54.109
180369 93.051 -76.831 54.120
180370 93.071 -76.823 54.130
180371 93.091 -76.814 54.139
180372 93.111 -76.809 54.147
180373 93.130 -76.806 54.154
180374 93.149 -76.803 54.162
180375 93.168 -76.799 54.169
180376 93.188 -76.796 54.173
180377 93.207 -76.793 54.177
180378 93.226 -76.790 54.181
180379 93.245 -76.787 54.184
180380 93.264 -76.784 54.187
180381 93.284 -76.781 54.189
180382 93.303 -76.777 54.190
180383 93.322 -76.774 54.191
180384 93.342 -76.771 54.194
180385 93.361 -76.768 54.195
180386 93.381 -76.765 54.197
180387 93.400 -76.762 54.200
180388 93.420 -76.759 54.202
180389 93.439 -76.756 54.206
180390 93.459 -76.753 54.211
180391 93.478 -76.750 54.217
180392 93.499 -76.747 54.225
180393 93.518 -76.749 54.232
180394 93.539 -76.743 54.239
180395 93.559 -76.736 54.244
180396 93.576 -76.731 54.247
180397 93.594 -76.726 54.250
180398 93.612 -76.721 54.251
180399 93.629 -76.717 54.252
180400 93.646 -76.712 54.251
180401 93.664 -76.708 54.250
180402 93.682 -76.703 54.245
180403 93.700 -76.696 54.240
180404 93.718 -76.689 54.235
180405 93.733 -76.681 54.230
180406 93.747 -76.675 54.226
180407 93.761 -76.668 54.221
180408 93.774 -76.661 54.213
180409 93.789 -76.655 54.200
180410 93.806 -76.650 54.189
180411 93.824 -76.646 54.180
180412 93.842 -76.640 54.172
180413 93.858 -76.640 54.166
180414 93.873 -76.641 54.163
180415 93.888 -76.643 54.159
180416 93.903 -76.643 54.152
180417 93.919 -76.645 54.146
180418 93.934 -76.646 54.141
180419 93.950 -76.646 54.137
180420 93.968 -76.644 54.133
180421 93.986 -76.643 54.129
180422 94.003 -76.641 54.124
180423 94.021 -76.640 54.122
180424 94.036 -76.640 54.121
180425 94.050 -76.640 54.126
180426 94.065 -76.641 54.132
180427 94.079 -76.642 54.138
180428 94.094 -76.643 54.142
180429 94.105 -76.642 54.139
180430 94.115 -76.640 54.134
180431 94.124 -76.640 54.129
180432 94.134 -76.638 54.120
180433 94.144 -76.637 54.111
180434 94.157 -76.633 54.101
180435 94.175 -76.627 54.092
180436 94.192 -76.621 54.074
180437 94.209 -76.614 54.057
180438 94.226 -76.608 54.040
180439 94.242 -76.602 54.022
180440 94.256 -76.596 54.005
180441 94.271 -76.590 53.987
180442 94.285 -76.584 53.967
180443 94.300 -76.579 53.941
180444 94.316 -76.572 53.912
180445 94.332 -76.564 53.883
180446 94.350 -76.555 53.851
180447 94.370 -76.545 53.818
180448 94.390 -76.534 53.785
180449 94.410 -76.524 53.749
180450 94.429 -76.513 53.709
180451 94.442 -76.507 53.667
180452 94.452 -76.502 53.626
180453 94.463 -76.497 53.584
180454 94.491 -76.492 53.542
180455 94.532 -76.487 53.500
180456 94.564 -76.489 53.457
180457 94.597 -76.490 53.414
180458 94.629 -76.492 53.372
180459 94.661 -76.493 53.335
180460 94.683 -76.490 53.301
180461 94.700 -76.486 53.271
180462 94.717 -76.480 53.245
180463 94.735 -76.477 53.219
180464 94.754 -76.473 53.193
180465 94.765 -76.472 53.168
180466 94.760 -76.475 53.142
180467 94.756 -76.478 53.114
180468 94.751 -76.480 53.086
180469 94.747 -76.483 53.056
180470 94.742 -76.486 53.028
180471 94.738 -76.489 53.001
180472 94.742 -76.490 52.975
180473 94.750 -76.492 52.949
180474 94.757 -76.494 52.921
180475 94.765 -76.495 52.888
180476 94.772 -76.497 52.852
180477 94.780 -76.499 52.809
180478 94.787 -76.500 52.763
180479 94.795 -76.502 52.714
180480 94.803 -76.504 52.663
180481 94.810 -76.505 52.613
180482 94.818 -76.507 52.565
180483 94.825 -76.508 52.517
180484 94.844 -76.512 52.466
180485 94.881 -76.520 52.413
180486 94.918 -76.528 52.358
180487 94.955 -76.536 52.303
180488 94.992 -76.543 52.249
180489 95.038 -76.554 52.198
180490 95.089 -76.567 52.147
180491 95.139 -76.580 52.094
180492 95.179 -76.588 52.041
180493 95.210 -76.598 51.989
180494 95.225 -76.612 51.937
180495 95.239 -76.627 51.887
180496 95.254 -76.641 51.838
180497 95.268 -76.655 51.790
180498 95.283 -76.669 51.742
180499 95.298 -76.683 51.695
180500 95.312 -76.698 51.649
180501 95.326 -76.712 51.604
180502 95.341 -76.726 51.559
180503 95.380 -76.737 51.516
180504 95.422 -76.746 51.473
180505 95.464 -76.756 51.431
180506 95.462 -76.761 51.391
180507 95.445 -76.764 51.352
180508 95.427 -76.767 51.313
180509 95.435 -76.785 51.276
180510 95.442 -76.803 51.241
180511 95.468 -76.821 51.206
180512 95.494 -76.839 51.172
180513 95.520 -76.856 51.137
180514 95.547 -76.874 51.102
180515 95.573 -76.891 51.065
180516 95.599 -76.909 51.028
180517 95.626 -76.926 50.989
180518 95.651 -76.937 50.947
180519 95.676 -76.943 50.903
180520 95.701 -76.950 50.857
180521 95.727 -76.956 50.810
180522 95.752 -76.963 50.762
180523 95.775 -76.970 50.714
180524 95.794 -76.979 50.666
180525 95.812 -76.987 50.619
180526 95.831 -76.996 50.568
180527 95.862 -77.003 50.513
180528 95.895 -77.002 50.457
180529 95.927 -76.997 50.399
180530 95.959 -76.992 50.338
180531 95.990 -76.987 50.277
180532 96.023 -76.981 50.216
180533 96.043 -76.989 50.150
180534 96.057 -77.003 50.084
180535 96.072 -77.019 50.019
180536 96.086 -77.034 49.956
180537 96.100 -77.049 49.895
180538 96.114 -77.063 49.838
180539 96.128 -77.078 49.784
180540 96.142 -77.094 49.733
180541 96.156 -77.109 49.684
180542 96.171 -77.122 49.639
180543 96.189 -77.119 49.595
180544 96.210 -77.106 49.556
180545 96.232 -77.093 49.520
180546 96.253 -77.079 49.487
180547 96.271 -77.078 49.455
180548 96.291 -77.084 49.425
180549 96.310 -77.088 49.397
180550 96.328 -77.095 49.370
180551 96.345 -77.102 49.344
180552 96.362 -77.109 49.322
180553 96.379 -77.116 49.302
180554 96.396 -77.122 49.282
180555 96.413 -77.129 49.265
180556 96.430 -77.136 49.248
180557 96.447 -77.143 49.231
180558 96.464 -77.150 49.214
180559 96.481 -77.157 49.201
180560 96.498 -77.163 49.188
180561 96.515 -77.170 49.175
180562 96.539 -77.172 49.163
180563 96.570 -77.172 49.151
180564 96.594 -77.175 49.136
180565 96.609 -77.188 49.118
180566 96.625 -77.201 49.102
180567 96.641 -77.214 49.085
180568 96.656 -77.228 49.069
180569 96.672 -77.241 49.051
180570 96.687 -77.255 49.032
180571 96.702 -77.268 49.012
180572 96.718 -77.282 48.992
180573 96.734 -77.295 48.971
180574 96.749 -77.309 48.951
180575 96.778 -77.322 48.932
180576 96.912 -77.333 48.912
180577 96.927 -77.338 48.892
180578 96.914 -77.353 48.873
180579 96.880 -77.371 48.853
180580 96.846 -77.389 48.832
180581 96.834 -77.401 48.812
180582 96.850 -77.406 48.792
180583 96.865 -77.411 48.771
180584 96.881 -77.417 48.750
180585 96.896 -77.421 48.730
180586 96.911 -77.426 48.710
180587 96.930 -77.430 48.691
180588 96.999 -77.416 48.671
180589 97.035 -77.430 48.652
180590 97.071 -77.447 48.634
180591 97.100 -77.471 48.615
180592 97.129 -77.496 48.596
180593 97.158 -77.521 48.578
180594 97.187 -77.545 48.558
180595 97.216 -77.570 48.536
180596 97.245 -77.595 48.515
180597 97.274 -77.620 48.494
180598 97.300 -77.640 48.473
180599 97.309 -77.629 48.453
180600 97.317 -77.614 48.433
180601 97.324 -77.599 48.413
180602 97.331 -77.583 48.393
180603 97.338 -77.568 48.375
180604 97.346 -77.553 48.358
180605 97.352 -77.536 48.340
180606 97.357 -77.518 48.322
180607 97.362 -77.500 48.302
180608 97.367 -77.482 48.283
180609 97.373 -77.467 48.264
180610 97.386 -77.483 48.245
180611 97.405 -77.495 48.223
180612 97.425 -77.507 48.202
180613 97.443 -77.519 48.179
180614 97.462 -77.532 48.155
180615 97.481 -77.544 48.129
180616 97.501 -77.556 48.104
180617 97.520 -77.569 48.078
180618 97.539 -77.581 48.052
180619 97.558 -77.593 48.025
180620 97.584 -77.595 47.997
180621 97.611 -77.592 47.966
180622 97.638 -77.591 47.936
180623 97.668 -77.590 47.905
180624 97.714 -77.597 47.873
180625 97.759 -77.603 47.842
180626 97.805 -77.610 47.811
180627 97.850 -77.616 47.780
180628 97.896 -77.623 47.749
180629 97.941 -77.629 47.718
180630 97.983 -77.634 47.687
180631 97.998 -77.630 47.657
180632 98.014 -77.625 47.627
180633 98.035 -77.618 47.600
180634 98.056 -77.611 47.572
180635 98.073 -77.602 47.545
180636 98.087 -77.590 47.517
180637 98.096 -77.577 47.489
180638 98.105 -77.564 47.462
180639 98.113 -77.551 47.435
180640 98.123 -77.537 47.409
180641 98.132 -77.522 47.381
180642 98.152 -77.514 47.353
180643 98.189 -77.522 47.325
180644 98.209 -77.523 47.297
180645 98.229 -77.524 47.271
180646 98.249 -77.526 47.244
180647 98.270 -77.527 47.218
180648 98.290 -77.529 47.192
180649 98.310 -77.530 47.165
180650 98.330 -77.531 47.138
180651 98.350 -77.533 47.111
180652 98.370 -77.534 47.083
180653 98.390 -77.536 47.057
180654 98.411 -77.537 47.032
180655 98.431 -77.539 47.009
180656 98.451 -77.540 46.983
180657 98.471 -77.541 46.959
180658 98.491 -77.542 46.934
180659 98.511 -77.544 46.910
180660 98.531 -77.545 46.887
180661 98.547 -77.550 46.865
180662 98.547 -77.558 46.844
180663 98.546 -77.564 46.821
180664 98.546 -77.571 46.798
180665 98.548 -77.578 46.775
180666 98.567 -77.581 46.752
180667 98.586 -77.584 46.728
180668 98.605 -77.587 46.705
180669 98.624 -77.590 46.682
180670 98.643 -77.594 46.659
180671 98.663 -77.597 46.634
180672 98.685 -77.602 46.610
180673 98.707 -77.608 46.586
180674 98.729 -77.613 46.562
180675 98.750 -77.618 46.540
180676 98.775 -77.605 46.519
180677 98.797 -77.594 46.500
180678 98.811 -77.594 46.480
180679 98.826 -77.594 46.460
180680 98.840 -77.595 46.438
180681 98.852 -77.598 46.417
180682 98.848 -77.601 46.396
180683 98.841 -77.602 46.376
180684 98.834 -77.604 46.357
180685 98.827 -77.605 46.337
180686 98.831 -77.611 46.316
180687 98.844 -77.629 46.295
180688 98.868 -77.649 46.273
180689 98.892 -77.668 46.252
180690 98.907 -77.683 46.231
180691 98.907 -77.686 46.209
180692 98.906 -77.690 46.187
180693 98.913 -77.693 46.164
180694 98.923 -77.696 46.141
180695 98.934 -77.699 46.119
180696 98.944 -77.703 46.097
180697 98.955 -77.706 46.077
180698 98.966 -77.709 46.057
180699 98.977 -77.712 46.037
180700 98.998 -77.719 46.017
180701 99.036 -77.732 45.997
180702 99.074 -77.744 45.979
180703 99.111 -77.757 45.963
180704 99.149 -77.769 45.947
180705 99.185 -77.787 45.932
180706 99.197 -77.787 45.918
180707 99.211 -77.766 45.905
180708 99.228 -77.780 45.892
180709 99.251 -77.803 45.880
180710 99.274 -77.827 45.869
180711 99.297 -77.849 45.858
180712 99.320 -77.872 45.849
180713 99.343 -77.895 45.839
180714 99.365 -77.918 45.830
180715 99.401 -77.938 45.821
180716 99.439 -77.958 45.813
180717 99.478 -77.978 45.805
180718 99.511 -77.994 45.797
180719 99.532 -78.003 45.788
180720 99.553 -78.011 45.779
180721 99.574 -78.019 45.770
180722 99.595 -78.028 45.760
180723 99.616 -78.036 45.750
180724 99.637 -78.044 45.739
180725 99.659 -78.048 45.728
180726 99.686 -78.034 45.717
180727 99.713 -78.021 45.706
180728 99.739 -78.007 45.696
180729 99.766 -77.993 45.686
180730 99.793 -77.980 45.674
180731 99.813 -77.983 45.663
180732 99.831 -77.993 45.651
180733 99.861 -78.020 45.638
180734 99.892 -78.050 45.626
180735 99.924 -78.080 45.613
180736 99.956 -78.110 45.599
180737 99.987 -78.140 45.585
180738 100.019 -78.170 45.570
180739 100.051 -78.184 45.556
180740 100.084 -78.193 45.541
180741 100.117 -78.201 45.525
180742 100.150 -78.210 45.509
180743 100.183 -78.219 45.493
180744 100.223 -78.194 45.477
180745 100.228 -78.211 45.461
180746 100.233 -78.227 45.445
180747 100.238 -78.244 45.430
180748 100.242 -78.260 45.415
180749 100.247 -78.276 45.399
180750 100.252 -78.293 45.383
180751 100.257 -78.310 45.367
180752 100.264 -78.326 45.353
180753 100.272 -78.342 45.337
180754 100.299 -78.346 45.321
180755 100.330 -78.347 45.306
180756 100.361 -78.348 45.289
180757 100.393 -78.349 45.273
180758 100.423 -78.350 45.256
180759 100.458 -78.352 45.240
180760 100.496 -78.355 45.224
180761 100.534 -78.358 45.208
180762 100.571 -78.362 45.191
180763 100.609 -78.365 45.175
180764 100.648 -78.369 45.158
180765 100.688 -78.376 45.142
180766 100.716 -78.382 45.126
180767 100.742 -78.388 45.110
180768 100.771 -78.394 45.096
180769 100.811 -78.400 45.081
180770 100.850 -78.407 45.067
180771 100.889 -78.413 45.053
180772 100.928 -78.420 45.038
180773 100.968 -78.427 45.023
180774 101.007 -78.433 45.008
180775 101.038 -78.440 44.993
180776 101.065 -78.447 44.978
180777 101.112 -78.426 44.962
180778 101.146 -78.417 44.947
180779 101.162 -78.425 44.930
180780 101.177 -78.433 44.914
180781 101.204 -78.449 44.898
180782 101.234 -78.466 44.881
180783 101.262 -78.483 44.864
180784 101.288 -78.501 44.848
180785 101.314 -78.517 44.831
180786 101.340 -78.535 44.814
180787 101.367 -78.552 44.797
180788 101.393 -78.569 44.781
180789 101.419 -78.587 44.764
180790 101.445 -78.610 44.749
180791 101.471 -78.633 44.733
180792 101.496 -78.655 44.718
180793 101.521 -78.680 44.703
180794 101.544 -78.707 44.688
180795 101.567 -78.739 44.673
180796 101.590 -78.772 44.658
180797 101.613 -78.805 44.642
180798 101.636 -78.837 44.627
180799 101.659 -78.870 44.612
180800 101.676 -78.887 44.598
180801 101.687 -78.890 44.584
180802 101.697 -78.893 44.570
180803 101.708 -78.896 44.557
180804 101.729 -78.912 44.541
180805 101.747 -78.918 44.528
180806 101.766 -78.924 44.516
180807 101.784 -78.930 44.502
180808 101.802 -78.937 44.490
180809 101.821 -78.943 44.478
180810 101.839 -78.949 44.465
180811 101.857 -78.955 44.452
180812 101.875 -78.962 44.440
180813 101.894 -78.968 44.428
180814 101.912 -78.974 44.416
180815 101.930 -78.980 44.404
181067 92.979 -76.834 54.065
181068 92.999 -76.825 54.078
181069 93.020 -76.817 54.091
181070 93.040 -76.808 54.103
181071 93.060 -76.799 54.114
181072 93.081 -76.790 54.124
181073 93.100 -76.786 54.134
181074 93.119 -76.783 54.141
181075 93.139 -76.779 54.149
181076 93.158 -76.776 54.155
181077 93.177 -76.773 54.162
181078 93.196 -76.770 54.167
181079 93.215 -76.766 54.171
181080 93.234 -76.763 54.174
181081 93.254 -76.760 54.178
181082 93.273 -76.756 54.180
181083 93.292 -76.753 54.182
181084 93.312 -76.750 54.183
181085 93.332 -76.747 54.185
181086 93.351 -76.744 54.187
181087 93.370 -76.741 54.188
181088 93.390 -76.738 54.191
181089 93.409 -76.735 54.193
181090 93.429 -76.732 54.195
181091 93.448 -76.730 54.198
181092 93.468 -76.727 54.203
181093 93.492 -76.723 54.210
181094 93.510 -76.725 54.216
181095 93.529 -76.728 54.223
181096 93.549 -76.724 54.229
181097 93.570 -76.715 54.233
181098 93.591 -76.705 54.236
181099 93.609 -76.700 54.236
181100 93.627 -76.696 54.237
181101 93.644 -76.690 54.238
181102 93.661 -76.686 54.237
181103 93.679 -76.681 54.233
181104 93.697 -76.677 54.230
181105 93.715 -76.669 54.226
181106 93.731 -76.661 54.223
181107 93.745 -76.655 54.219
181108 93.759 -76.649 54.215
181109 93.773 -76.642 54.209
181110 93.786 -76.635 54.201
181111 93.800 -76.628 54.193
181112 93.818 -76.624 54.187
181113 93.836 -76.619 54.183
181114 93.852 -76.618 54.178
181115 93.867 -76.620 54.175
181116 93.882 -76.621 54.170
181117 93.897 -76.622 54.163
181118 93.912 -76.623 54.155
181119 93.927 -76.624 54.149
181120 93.943 -76.625 54.144
181121 93.960 -76.624 54.140
181122 93.977 -76.623 54.138
181123 93.995 -76.621 54.134
181124 94.011 -76.621 54.131
181125 94.026 -76.621 54.129
181126 94.040 -76.622 54.131
181127 94.055 -76.623 54.135
181128 94.069 -76.624 54.141
181129 94.084 -76.624 54.147
181130 94.098 -76.625 54.147
181131 94.109 -76.624 54.145
181132 94.119 -76.623 54.142
181133 94.128 -76.621 54.135
181134 94.138 -76.621 54.128
181135 94.152 -76.617 54.119
181136 94.169 -76.611 54.110
181137 94.186 -76.604 54.094
181138 94.203 -76.598 54.078
181139 94.219 -76.592 54.061
181140 94.233 -76.586 54.045
181141 94.248 -76.580 54.028
181142 94.262 -76.574 54.010
181143 94.277 -76.569 53.992
181144 94.291 -76.563 53.967
181145 94.307 -76.557 53.940
181146 94.325 -76.549 53.910
181147 94.344 -76.538 53.879
181148 94.364 -76.527 53.846
181149 94.384 -76.517 53.813
181150 94.404 -76.506 53.777
181151 94.424 -76.495 53.738
181152 94.441 -76.486 53.696
181153 94.452 -76.482 53.653
181154 94.462 -76.477 53.611
181155 94.497 -76.471 53.568
181156 94.533 -76.470 53.525
181157 94.565 -76.472 53.482
181158 94.597 -76.473 53.437
181159 94.629 -76.475 53.393
181160 94.662 -76.477 53.354
181161 94.684 -76.474 53.318
181162 94.701 -76.469 53.286
181163 94.718 -76.464 53.259
181164 94.737 -76.461 53.232
181165 94.756 -76.458 53.206
181166 94.771 -76.456 53.181
181167 94.767 -76.459 53.156
181168 94.762 -76.461 53.129
181169 94.758 -76.464 53.102
181170 94.753 -76.467 53.072
181171 94.749 -76.470 53.042
181172 94.744 -76.473 53.013
181173 94.739 -76.475 52.985
181174 94.735 -76.478 52.959
181175 94.730 -76.480 52.931
181176 94.736 -76.483 52.899
181177 94.743 -76.484 52.864
181178 94.751 -76.486 52.822
181179 94.758 -76.487 52.777
181180 94.766 -76.489 52.728
181181 94.773 -76.491 52.677
181182 94.781 -76.492 52.627
181183 94.788 -76.494 52.578
181184 94.796 -76.495 52.530
181185 94.815 -76.499 52.480
181186 94.852 -76.507 52.428
181187 94.889 -76.514 52.377
181188 94.926 -76.523 52.325
181189 94.973 -76.534 52.270
181190 95.024 -76.547 52.216
181191 95.075 -76.560 52.164
181192 95.124 -76.572 52.110
181193 95.164 -76.580 52.057
181194 95.203 -76.588 52.004
181195 95.219 -76.602 51.952
181196 95.234 -76.616 51.901
181197 95.248 -76.630 51.853
181198 95.263 -76.645 51.805
181199 95.277 -76.659 51.758
181200 95.292 -76.673 51.710
181201 95.306 -76.687 51.663
181202 95.321 -76.702 51.618
181203 95.342 -76.715 51.573
181204 95.384 -76.724 51.529
181205 95.425 -76.734 51.487
181206 95.467 -76.744 51.446
181207 95.486 -76.752 51.406
181208 95.468 -76.754 51.367
181209 95.451 -76.758 51.329
181210 95.458 -76.776 51.292
181211 95.465 -76.794 51.258
181212 95.472 -76.812 51.224
181213 95.484 -76.831 51.190
181214 95.511 -76.848 51.156
181215 95.537 -76.865 51.122
181216 95.563 -76.883 51.086
181217 95.590 -76.900 51.049
181218 95.616 -76.918 51.010
181219 95.641 -76.932 50.968
181220 95.667 -76.938 50.924
181221 95.692 -76.945 50.877
181222 95.717 -76.952 50.830
181223 95.742 -76.959 50.782
181224 95.767 -76.965 50.734
181225 95.793 -76.972 50.685
181226 95.817 -76.978 50.637
181227 95.835 -76.987 50.585
181228 95.868 -76.987 50.531
181229 95.899 -76.982 50.474
181230 95.931 -76.977 50.415
181231 95.963 -76.972 50.355
181232 95.995 -76.967 50.294
181233 96.027 -76.962 50.233
181234 96.046 -76.970 50.169
181235 96.060 -76.985 50.103
181236 96.075 -77.000 50.037
181237 96.089 -77.015 49.972
181238 96.103 -77.030 49.911
181239 96.117 -77.045 49.854
181240 96.131 -77.060 49.801
181241 96.145 -77.075 49.751
181242 96.159 -77.091 49.702
181243 96.172 -77.106 49.656
181244 96.187 -77.119 49.613
181245 96.190 -77.122 49.573
181246 96.203 -77.113 49.537
181247 96.217 -77.104 49.504
181248 96.230 -77.096 49.473
181249 96.247 -77.103 49.445
181250 96.264 -77.110 49.418
181251 96.281 -77.116 49.391
181252 96.298 -77.122 49.366
181253 96.315 -77.129 49.344
181254 96.332 -77.136 49.324
181255 96.349 -77.143 49.305
181256 96.366 -77.150 49.288
181257 96.383 -77.157 49.271
181258 96.400 -77.163 49.254
181259 96.417 -77.170 49.238
181260 96.447 -77.170 49.225
181261 96.477 -77.169 49.212
181262 96.508 -77.169 49.200
181263 96.538 -77.168 49.187
181264 96.569 -77.168 49.173
181265 96.590 -77.172 49.157
181266 96.611 -77.176 49.139
181267 96.628 -77.188 49.123
181268 96.643 -77.201 49.105
181269 96.659 -77.214 49.087
181270 96.674 -77.228 49.070
181271 96.690 -77.241 49.051
181272 96.705 -77.254 49.031
181273 96.721 -77.268 49.010
181274 96.736 -77.282 48.989
181275 96.752 -77.295 48.970
181276 96.854 -77.309 48.950
181277 96.895 -77.314 48.930
181278 96.909 -77.319 48.909
181279 96.924 -77.325 48.890
181280 96.937 -77.336 48.869
181281 96.902 -77.354 48.847
181282 96.868 -77.373 48.825
181283 96.853 -77.385 48.804
181284 96.869 -77.391 48.783
181285 96.884 -77.395 48.762
181286 96.899 -77.401 48.741
181287 96.914 -77.406 48.720
181288 96.930 -77.410 48.700
181289 97.002 -77.392 48.680
181290 97.040 -77.405 48.660
181291 97.075 -77.422 48.640
181292 97.103 -77.447 48.622
181293 97.133 -77.472 48.603
181294 97.162 -77.496 48.584
181295 97.190 -77.521 48.564
181296 97.220 -77.546 48.544
181297 97.248 -77.570 48.523
181298 97.278 -77.595 48.502
181299 97.307 -77.620 48.481
181300 97.326 -77.629 48.462
181301 97.331 -77.611 48.442
181302 97.336 -77.592 48.422
181303 97.342 -77.575 48.402
181304 97.347 -77.557 48.383
181305 97.352 -77.539 48.365
181306 97.358 -77.521 48.348
181307 97.363 -77.503 48.331
181308 97.368 -77.485 48.312
181309 97.373 -77.467 48.293
181310 97.382 -77.465 48.274
181311 97.395 -77.483 48.254
181312 97.409 -77.496 48.233
181313 97.428 -77.508 48.211
181314 97.447 -77.521 48.188
181315 97.466 -77.533 48.164
181316 97.485 -77.545 48.138
181317 97.504 -77.558 48.112
181318 97.523 -77.570 48.085
181319 97.542 -77.583 48.057
181320 97.565 -77.589 48.030
181321 97.592 -77.587 48.001
181322 97.619 -77.586 47.970
181323 97.647 -77.583 47.940
181324 97.681 -77.585 47.909
181325 97.727 -77.592 47.877
181326 97.773 -77.598 47.844
181327 97.818 -77.605 47.814
181328 97.864 -77.611 47.782
181329 97.909 -77.617 47.750
181330 97.955 -77.624 47.720
181331 98.000 -77.630 47.689
181332 98.028 -77.630 47.659
181333 98.047 -77.624 47.630
181334 98.061 -77.620 47.602
181335 98.075 -77.605 47.575
181336 98.084 -77.592 47.547
181337 98.093 -77.579 47.519
181338 98.102 -77.566 47.491
181339 98.111 -77.552 47.464
181340 98.120 -77.539 47.437
181341 98.130 -77.523 47.411
181342 98.139 -77.508 47.383
181343 98.148 -77.493 47.355
181344 98.192 -77.501 47.328
181345 98.233 -77.511 47.300
181346 98.253 -77.512 47.273
181347 98.273 -77.513 47.247
181348 98.293 -77.514 47.221
181349 98.313 -77.516 47.197
181350 98.334 -77.517 47.171
181351 98.354 -77.519 47.145
181352 98.374 -77.520 47.118
181353 98.394 -77.521 47.091
181354 98.414 -77.523 47.065
181355 98.434 -77.524 47.040
181356 98.454 -77.526 47.016
181357 98.475 -77.527 46.991
181358 98.495 -77.529 46.967
181359 98.515 -77.530 46.943
181360 98.535 -77.531 46.919
181361 98.549 -77.526 46.896
181362 98.555 -77.545 46.874
181363 98.562 -77.564 46.852
181364 98.561 -77.571 46.830
181365 98.560 -77.578 46.807
181366 98.570 -77.583 46.784
181367 98.589 -77.586 46.761
181368 98.608 -77.589 46.737
181369 98.627 -77.592 46.714
181370 98.646 -77.595 46.691
181371 98.665 -77.599 46.668
181372 98.685 -77.602 46.642
181373 98.704 -77.605 46.618
181374 98.723 -77.608 46.594
181375 98.744 -77.605 46.570
181376 98.762 -77.602 46.548
181377 98.786 -77.589 46.527
181378 98.806 -77.581 46.506
181379 98.820 -77.582 46.485
181380 98.836 -77.583 46.464
181381 98.851 -77.584 46.442
181382 98.862 -77.588 46.422
181383 98.870 -77.595 46.402
181384 98.870 -77.599 46.382
181385 98.863 -77.600 46.362
181386 98.856 -77.602 46.343
181387 98.868 -77.610 46.323
181388 98.885 -77.619 46.301
181389 98.871 -77.633 46.278
181390 98.893 -77.652 46.257
181391 98.917 -77.671 46.237
181392 98.920 -77.678 46.216
181393 98.921 -77.682 46.195
181394 98.932 -77.685 46.173
181395 98.942 -77.688 46.150
181396 98.953 -77.691 46.127
181397 98.964 -77.694 46.106
181398 98.975 -77.697 46.085
181399 98.985 -77.700 46.064
181400 98.996 -77.703 46.043
181401 99.006 -77.706 46.022
181402 99.034 -77.716 46.002
181403 99.072 -77.728 45.983
181404 99.109 -77.743 45.966
181405 99.142 -77.772 45.950
181406 99.167 -77.790 45.934
181407 99.187 -77.768 45.920
181408 99.199 -77.748 45.907
181409 99.221 -77.769 45.893
181410 99.243 -77.793 45.881
181411 99.267 -77.815 45.870
181412 99.289 -77.838 45.860
181413 99.312 -77.861 45.850
181414 99.336 -77.884 45.841
181415 99.374 -77.903 45.831
181416 99.413 -77.923 45.823
181417 99.451 -77.943 45.814
181418 99.490 -77.962 45.806
181419 99.528 -77.981 45.798
181420 99.555 -77.993 45.789
181421 99.576 -78.002 45.780
181422 99.596 -78.010 45.770
181423 99.618 -78.018 45.760
181424 99.639 -78.028 45.750
181425 99.659 -78.036 45.740
181426 99.685 -78.027 45.729
181427 99.712 -78.013 45.718
181428 99.738 -78.000 45.707
181429 99.765 -77.986 45.697
181430 99.790 -77.977 45.687
181431 99.807 -77.987 45.676
181432 99.825 -77.996 45.665
181433 99.843 -78.006 45.654
181434 99.864 -78.022 45.641
181435 99.896 -78.051 45.629
181436 99.927 -78.081 45.617
181437 99.959 -78.111 45.603
181438 99.990 -78.141 45.589
181439 100.022 -78.163 45.574
181440 100.055 -78.171 45.559
181441 100.088 -78.179 45.542
181442 100.121 -78.188 45.527
181443 100.154 -78.197 45.511
181444 100.187 -78.205 45.495
181445 100.238 -78.172 45.479
181446 100.242 -78.189 45.463
181447 100.247 -78.206 45.448
181448 100.252 -78.222 45.433
181449 100.256 -78.238 45.418
181450 100.261 -78.255 45.403
181451 100.266 -78.271 45.388
181452 100.272 -78.288 45.373
181453 100.279 -78.305 45.357
181454 100.285 -78.322 45.342
181455 100.302 -78.333 45.326
181456 100.331 -78.335 45.310
181457 100.362 -78.336 45.293
181458 100.394 -78.337 45.276
181459 100.425 -78.338 45.260
181460 100.460 -78.341 45.244
181461 100.498 -78.344 45.228
181462 100.536 -78.347 45.211
181463 100.573 -78.350 45.194
181464 100.612 -78.354 45.177
181465 100.649 -78.357 45.162
181466 100.682 -78.363 45.146
181467 100.719 -78.369 45.131
181468 100.759 -78.376 45.115
181469 100.798 -78.382 45.101
181470 100.837 -78.389 45.087
181471 100.876 -78.395 45.072
181472 100.916 -78.402 45.058
181473 100.956 -78.409 45.042
181474 100.995 -78.416 45.027
181475 101.030 -78.423 45.013
181476 101.057 -78.429 44.997
181477 101.085 -78.436 44.981
181478 101.132 -78.414 44.965
181479 101.155 -78.416 44.950
181480 101.170 -78.424 44.934
181481 101.186 -78.432 44.918
181482 101.201 -78.440 44.901
181483 101.229 -78.457 44.884
181484 101.259 -78.473 44.867
181485 101.288 -78.491 44.851
181486 101.317 -78.508 44.834
181487 101.343 -78.525 44.817
181488 101.370 -78.542 44.801
181489 101.396 -78.559 44.784
181490 101.422 -78.576 44.769
181491 101.448 -78.598 44.752
181492 101.474 -78.621 44.736
181493 101.498 -78.645 44.721
181494 101.522 -78.673 44.706
181495 101.544 -78.700 44.691
181496 101.567 -78.732 44.676
181497 101.591 -78.765 44.660
181498 101.613 -78.798 44.645
181499 101.636 -78.830 44.630
181500 101.658 -78.859 44.615
181501 101.669 -78.862 44.600
181502 101.680 -78.865 44.586
181503 101.690 -78.868 44.572
181504 101.714 -78.882 44.559
181505 101.753 -78.908 44.543
181506 101.766 -78.926 44.531
181507 101.784 -78.932 44.518
181508 101.802 -78.938 44.504
181509 101.821 -78.945 44.492
181510 101.839 -78.951 44.479
181511 101.857 -78.957 44.466
181512 101.876 -78.963 44.453
181513 101.894 -78.969 44.441
181514 101.912 -78.976 44.429
181515 101.930 -78.982 44.418
181516 101.949 -78.988 44.406
181769 92.989 -76.802 54.059
181770 93.009 -76.793 54.073
181771 93.029 -76.784 54.086
181772 93.050 -76.775 54.097
181773 93.070 -76.767 54.108
181774 93.090 -76.762 54.120
181775 93.109 -76.759 54.128
181776 93.128 -76.755 54.136
181777 93.147 -76.752 54.144
181778 93.166 -76.749 54.151
181779 93.186 -76.746 54.156
181780 93.205 -76.743 54.161
181781 93.224 -76.740 54.165
181782 93.243 -76.736 54.169
181783 93.263 -76.733 54.172
181784 93.282 -76.730 54.174
181785 93.301 -76.727 54.176
181786 93.321 -76.724 54.178
181787 93.341 -76.721 54.179
181788 93.360 -76.718 54.181
181789 93.379 -76.715 54.183
181790 93.399 -76.712 54.185
181791 93.418 -76.708 54.186
181792 93.438 -76.705 54.188
181793 93.458 -76.702 54.192
181794 93.484 -76.699 54.196
181795 93.502 -76.702 54.201
181796 93.521 -76.705 54.206
181797 93.539 -76.708 54.212
181798 93.559 -76.705 54.216
181799 93.580 -76.696 54.218
181800 93.601 -76.686 54.218
181801 93.622 -76.676 54.220
181802 93.641 -76.669 54.221
181803 93.659 -76.665 54.222
181804 93.677 -76.660 54.221
181805 93.694 -76.655 54.219
181806 93.712 -76.650 54.217
181807 93.729 -76.642 54.214
181808 93.744 -76.636 54.210
181809 93.757 -76.629 54.208
181810 93.771 -76.622 54.205
181811 93.785 -76.615 54.202
181812 93.798 -76.608 54.198
181813 93.812 -76.602 54.196
181814 93.830 -76.597 54.195
181815 93.845 -76.597 54.192
181816 93.860 -76.598 54.189
181817 93.876 -76.599 54.184
181818 93.891 -76.600 54.176
181819 93.906 -76.601 54.167
181820 93.921 -76.602 54.159
181821 93.936 -76.603 54.153
181822 93.952 -76.605 54.149
181823 93.969 -76.604 54.148
181824 93.986 -76.602 54.144
181825 94.001 -76.603 54.140
181826 94.016 -76.603 54.137
181827 94.030 -76.604 54.137
181828 94.045 -76.605 54.139
181829 94.059 -76.605 54.144
181830 94.074 -76.606 54.150
181831 94.088 -76.607 54.152
181832 94.103 -76.608 54.152
181833 94.113 -76.606 54.151
181834 94.123 -76.605 54.148
181835 94.134 -76.603 54.143
181836 94.147 -76.599 54.136
181837 94.163 -76.594 54.126
181838 94.180 -76.588 54.112
181839 94.196 -76.582 54.095
181840 94.210 -76.576 54.079
181841 94.225 -76.571 54.062
181842 94.240 -76.564 54.046
181843 94.254 -76.559 54.030
181844 94.269 -76.553 54.013
181845 94.285 -76.547 53.988
181846 94.302 -76.541 53.960
181847 94.321 -76.533 53.931
181848 94.338 -76.520 53.900
181849 94.358 -76.510 53.868
181850 94.378 -76.499 53.835
181851 94.398 -76.489 53.800
181852 94.418 -76.479 53.762
181853 94.438 -76.468 53.721
181854 94.451 -76.461 53.679
181855 94.461 -76.456 53.636
181856 94.500 -76.452 53.594
181857 94.533 -76.454 53.552
181858 94.565 -76.455 53.507
181859 94.597 -76.457 53.461
181860 94.629 -76.458 53.415
181861 94.662 -76.460 53.373
181862 94.685 -76.458 53.334
181863 94.702 -76.452 53.301
181864 94.720 -76.449 53.273
181865 94.739 -76.446 53.245
181866 94.758 -76.443 53.218
181867 94.776 -76.440 53.193
181868 94.773 -76.442 53.169
181869 94.768 -76.445 53.143
181870 94.764 -76.448 53.116
181871 94.759 -76.451 53.086
181872 94.755 -76.454 53.055
181873 94.750 -76.456 53.025
181874 94.746 -76.459 52.996
181875 94.741 -76.462 52.970
181876 94.737 -76.464 52.942
181877 94.732 -76.467 52.911
181878 94.728 -76.470 52.877
181879 94.723 -76.473 52.835
181880 94.729 -76.474 52.790
181881 94.737 -76.476 52.742
181882 94.744 -76.478 52.691
181883 94.752 -76.480 52.641
181884 94.759 -76.481 52.592
181885 94.767 -76.483 52.543
181886 94.785 -76.486 52.493
181887 94.823 -76.494 52.444
181888 94.860 -76.502 52.395
181889 94.909 -76.514 52.345
181890 94.960 -76.527 52.291
181891 95.011 -76.540 52.236
181892 95.062 -76.553 52.182
181893 95.109 -76.564 52.127
181894 95.148 -76.572 52.073
181895 95.187 -76.580 52.021
181896 95.214 -76.592 51.969
181897 95.229 -76.606 51.917
181898 95.243 -76.620 51.868
181899 95.257 -76.634 51.820
181900 95.272 -76.649 51.773
181901 95.286 -76.663 51.725
181902 95.301 -76.677 51.679
181903 95.315 -76.691 51.633
181904 95.345 -76.703 51.588
181905 95.387 -76.713 51.545
181906 95.429 -76.723 51.502
181907 95.471 -76.733 51.461
181908 95.509 -76.742 51.421
181909 95.491 -76.745 51.382
181910 95.474 -76.749 51.345
181911 95.482 -76.767 51.309
181912 95.489 -76.785 51.275
181913 95.496 -76.803 51.241
181914 95.503 -76.821 51.208
181915 95.510 -76.840 51.174
181916 95.527 -76.857 51.140
181917 95.553 -76.875 51.104
181918 95.580 -76.893 51.067
181919 95.606 -76.910 51.027
181920 95.632 -76.927 50.987
181921 95.657 -76.934 50.943
181922 95.682 -76.940 50.896
181923 95.707 -76.947 50.849
181924 95.733 -76.953 50.801
181925 95.758 -76.960 50.753
181926 95.783 -76.967 50.704
181927 95.808 -76.974 50.657
181928 95.838 -76.972 50.606
181929 95.870 -76.966 50.552
181930 95.902 -76.962 50.495
181931 95.934 -76.956 50.436
181932 95.965 -76.951 50.375
181933 95.997 -76.946 50.314
181934 96.027 -76.943 50.253
181935 96.044 -76.953 50.188
181936 96.058 -76.969 50.123
181937 96.072 -76.984 50.057
181938 96.085 -76.999 49.992
181939 96.099 -77.014 49.930
181940 96.112 -77.029 49.872
181941 96.126 -77.044 49.818
181942 96.139 -77.059 49.769
181943 96.153 -77.075 49.721
181944 96.170 -77.088 49.675
181945 96.184 -77.100 49.632
181946 96.174 -77.118 49.593
181947 96.164 -77.137 49.558
181948 96.172 -77.134 49.525
181949 96.185 -77.125 49.494
181950 96.200 -77.123 49.466
181951 96.218 -77.130 49.439
181952 96.235 -77.137 49.413
181953 96.251 -77.144 49.388
181954 96.268 -77.150 49.366
181955 96.286 -77.157 49.346
181956 96.303 -77.164 49.328
181957 96.324 -77.169 49.310
181958 96.355 -77.168 49.293
181959 96.385 -77.167 49.277
181960 96.415 -77.166 49.262
181961 96.446 -77.166 49.250
181962 96.476 -77.166 49.237
181963 96.507 -77.165 49.224
181964 96.537 -77.164 49.211
181965 96.567 -77.163 49.197
181966 96.587 -77.169 49.181
181967 96.608 -77.173 49.163
181968 96.628 -77.178 49.145
181969 96.646 -77.187 49.127
181970 96.661 -77.201 49.108
181971 96.677 -77.214 49.088
181972 96.692 -77.228 49.068
181973 96.708 -77.241 49.048
181974 96.724 -77.254 49.027
181975 96.739 -77.268 49.007
181976 96.754 -77.282 48.988
181977 96.863 -77.291 48.968
181978 96.877 -77.296 48.947
181979 96.891 -77.301 48.928
181980 96.906 -77.305 48.907
181981 96.920 -77.312 48.885
181982 96.935 -77.323 48.863
181983 96.925 -77.338 48.840
181984 96.890 -77.356 48.818
181985 96.872 -77.370 48.796
181986 96.887 -77.375 48.774
181987 96.903 -77.380 48.752
181988 96.918 -77.385 48.731
181989 96.934 -77.390 48.710
181990 96.997 -77.375 48.690
181991 97.044 -77.379 48.670
181992 97.078 -77.398 48.649
181993 97.107 -77.423 48.629
181994 97.136 -77.447 48.610
181995 97.165 -77.472 48.590
181996 97.194 -77.497 48.571
181997 97.223 -77.521 48.552
181998 97.252 -77.546 48.530
181999 97.281 -77.570 48.509
182000 97.305 -77.589 48.488
182001 97.330 -77.609 48.469
182002 97.337 -77.595 48.449
182003 97.343 -77.578 48.429
182004 97.348 -77.560 48.410
182005 97.353 -77.542 48.391
182006 97.358 -77.524 48.373
182007 97.363 -77.506 48.356
182008 97.369 -77.488 48.339
182009 97.374 -77.470 48.320
182010 97.379 -77.452 48.301
182011 97.391 -77.464 48.281
182012 97.404 -77.483 48.262
182013 97.417 -77.500 48.242
182014 97.432 -77.511 48.219
182015 97.451 -77.523 48.196
182016 97.470 -77.535 48.171
182017 97.489 -77.547 48.145
182018 97.508 -77.560 48.119
182019 97.527 -77.572 48.091
182020 97.546 -77.583 48.062
182021 97.574 -77.582 48.034
182022 97.601 -77.580 48.005
182023 97.629 -77.578 47.974
182024 97.656 -77.577 47.943
182025 97.695 -77.580 47.911
182026 97.740 -77.586 47.879
182027 97.786 -77.592 47.847
182028 97.831 -77.599 47.816
182029 97.877 -77.605 47.784
182030 97.922 -77.612 47.752
182031 97.968 -77.618 47.722
182032 98.009 -77.622 47.692
182033 98.041 -77.625 47.662
182034 98.055 -77.636 47.633
182035 98.069 -77.619 47.604
182036 98.081 -77.594 47.576
182037 98.090 -77.580 47.548
182038 98.099 -77.567 47.520
182039 98.108 -77.555 47.493
182040 98.118 -77.539 47.466
182041 98.127 -77.524 47.439
182042 98.136 -77.509 47.412
182043 98.146 -77.494 47.384
182044 98.155 -77.479 47.357
182045 98.183 -77.476 47.330
182046 98.233 -77.489 47.302
182047 98.277 -77.499 47.275
182048 98.297 -77.501 47.249
182049 98.317 -77.501 47.224
182050 98.337 -77.503 47.200
182051 98.357 -77.504 47.176
182052 98.377 -77.506 47.151
182053 98.397 -77.508 47.125
182054 98.418 -77.509 47.099
182055 98.438 -77.510 47.074
182056 98.458 -77.511 47.049
182057 98.478 -77.513 47.022
182058 98.498 -77.514 46.999
182059 98.518 -77.516 46.975
182060 98.538 -77.517 46.950
182061 98.558 -77.518 46.927
182062 98.556 -77.519 46.903
182063 98.563 -77.539 46.882
182064 98.569 -77.558 46.861
182065 98.575 -77.578 46.838
182066 98.575 -77.585 46.816
182067 98.592 -77.589 46.793
182068 98.611 -77.592 46.770
182069 98.630 -77.595 46.745
182070 98.649 -77.598 46.722
182071 98.670 -77.593 46.698
182072 98.692 -77.587 46.674
182073 98.713 -77.580 46.649
182074 98.734 -77.574 46.625
182075 98.756 -77.568 46.601
182076 98.777 -77.561 46.576
182077 98.797 -77.558 46.554
182078 98.805 -77.564 46.532
182079 98.816 -77.570 46.511
182080 98.832 -77.571 46.490
182081 98.847 -77.572 46.469
182082 98.862 -77.573 46.448
182083 98.873 -77.578 46.428
182084 98.880 -77.585 46.407
182085 98.888 -77.591 46.388
182086 98.892 -77.597 46.368
182087 98.887 -77.599 46.348
182088 98.905 -77.608 46.327
182089 98.922 -77.618 46.306
182090 98.939 -77.627 46.284
182091 98.931 -77.640 46.263
182092 98.918 -77.655 46.243
182093 98.933 -77.670 46.223
182094 98.940 -77.673 46.201
182095 98.951 -77.676 46.180
182096 98.961 -77.679 46.157
182097 98.973 -77.682 46.134
182098 98.983 -77.685 46.113
182099 98.994 -77.689 46.091
182100 99.004 -77.692 46.069
182101 99.015 -77.695 46.048
182102 99.025 -77.698 46.026
182103 99.037 -77.701 46.006
182104 99.066 -77.727 45.987
182105 99.100 -77.757 45.969
182106 99.133 -77.787 45.952
182107 99.161 -77.769 45.937
182108 99.188 -77.745 45.922
182109 99.191 -77.735 45.909
182110 99.213 -77.758 45.894
182111 99.236 -77.781 45.882
182112 99.259 -77.804 45.871
182113 99.282 -77.827 45.862
182114 99.309 -77.849 45.852
182115 99.347 -77.868 45.843
182116 99.386 -77.888 45.833
182117 99.424 -77.907 45.825
182118 99.463 -77.927 45.816
182119 99.501 -77.946 45.807
182120 99.540 -77.966 45.798
182121 99.577 -77.984 45.788
182122 99.598 -77.993 45.778
182123 99.619 -78.002 45.769
182124 99.640 -78.010 45.759
182125 99.661 -78.018 45.748
182126 99.684 -78.018 45.738
182127 99.711 -78.005 45.728
182128 99.737 -77.992 45.717
182129 99.764 -77.978 45.707
182130 99.784 -77.981 45.697
182131 99.802 -77.990 45.687
182132 99.820 -78.000 45.675
182133 99.837 -78.010 45.665
182134 99.855 -78.020 45.654
182135 99.872 -78.030 45.642
182136 99.899 -78.053 45.630
182137 99.930 -78.082 45.617
182138 99.962 -78.112 45.604
182139 99.993 -78.140 45.590
182140 100.027 -78.149 45.576
182141 100.059 -78.157 45.560
182142 100.092 -78.166 45.543
182143 100.125 -78.174 45.528
182144 100.158 -78.183 45.512
182145 100.191 -78.191 45.496
182146 100.241 -78.166 45.480
182147 100.256 -78.167 45.464
182148 100.261 -78.184 45.450
182149 100.265 -78.200 45.436
182150 100.270 -78.216 45.421
182151 100.275 -78.233 45.407
182152 100.281 -78.250 45.392
182153 100.288 -78.267 45.377
182154 100.294 -78.284 45.362
182155 100.301 -78.300 45.346
182156 100.310 -78.316 45.329
182157 100.333 -78.323 45.313
182158 100.363 -78.324 45.298
182159 100.395 -78.326 45.281
182160 100.426 -78.326 45.263
182161 100.463 -78.329 45.247
182162 100.500 -78.332 45.230
182163 100.538 -78.335 45.215
182164 100.576 -78.338 45.199
182165 100.614 -78.341 45.182
182166 100.655 -78.343 45.167
182167 100.702 -78.351 45.151
182168 100.746 -78.358 45.136
182169 100.786 -78.364 45.121
182170 100.825 -78.371 45.106
182171 100.864 -78.378 45.091
182172 100.904 -78.385 45.077
182173 100.943 -78.391 45.063
182174 100.982 -78.398 45.048
182175 101.022 -78.404 45.033
182176 101.050 -78.411 45.018
182177 101.077 -78.418 45.001
182178 101.104 -78.425 44.985
182179 101.148 -78.407 44.969
182180 101.164 -78.415 44.954
182181 101.179 -78.423 44.938
182182 101.194 -78.431 44.921
182183 101.210 -78.439 44.904
182184 101.226 -78.448 44.888
182185 101.254 -78.464 44.871
182186 101.284 -78.481 44.854
182187 101.314 -78.498 44.837
182188 101.343 -78.515 44.821
182189 101.372 -78.532 44.805
182190 101.398 -78.550 44.789
182191 101.425 -78.567 44.773
182192 101.451 -78.588 44.756
182193 101.476 -78.611 44.740
182194 101.499 -78.639 44.725
182195 101.522 -78.666 44.709
182196 101.545 -78.692 44.694
182197 101.567 -78.725 44.679
182198 101.591 -78.758 44.664
182199 101.614 -78.791 44.647
182200 101.636 -78.824 44.633
182201 101.650 -78.834 44.618
182202 101.662 -78.837 44.604
182203 101.672 -78.840 44.590
182204 101.698 -78.855 44.576
182205 101.737 -78.880 44.563
182206 101.775 -78.905 44.547
182207 101.785 -78.934 44.534
182208 101.803 -78.940 44.520
182209 101.821 -78.946 44.507
182210 101.840 -78.952 44.494
182211 101.858 -78.959 44.481
182212 101.876 -78.965 44.468
182213 101.895 -78.971 44.455
182214 101.913 -78.977 44.443
182215 101.931 -78.984 44.431
182216 101.949 -78.990 44.420
182217 101.968 -78.996 44.408
182471 92.999 -76.769 54.056
182472 93.019 -76.761 54.069
182473 93.040 -76.752 54.082
182474 93.060 -76.743 54.094
182475 93.079 -76.739 54.105
182476 93.098 -76.735 54.114
182477 93.118 -76.732 54.124
182478 93.137 -76.729 54.133
182479 93.156 -76.725 54.140
182480 93.176 -76.722 54.146
182481 93.195 -76.719 54.151
182482 93.214 -76.715 54.156
182483 93.233 -76.712 54.160
182484 93.252 -76.709 54.164
182485 93.272 -76.706 54.166
182486 93.291 -76.703 54.169
182487 93.310 -76.700 54.171
182488 93.330 -76.697 54.172
182489 93.350 -76.694 54.174
182490 93.369 -76.691 54.176
182491 93.389 -76.688 54.177
182492 93.408 -76.685 54.178
182493 93.427 -76.682 54.179
182494 93.447 -76.679 54.181
182495 93.476 -76.676 54.184
182496 93.494 -76.679 54.187
182497 93.512 -76.681 54.192
182498 93.531 -76.684 54.196
182499 93.549 -76.687 54.198
182500 93.569 -76.687 54.200
182501 93.590 -76.677 54.200
182502 93.611 -76.667 54.202
182503 93.632 -76.657 54.204
182504 93.653 -76.647 54.207
182505 93.674 -76.638 54.208
182506 93.692 -76.633 54.208
182507 93.709 -76.629 54.208
182508 93.727 -76.623 54.206
182509 93.742 -76.616 54.205
182510 93.755 -76.609 54.204
182511 93.769 -76.602 54.203
182512 93.783 -76.596 54.203
182513 93.797 -76.589 54.204
182514 93.811 -76.582 54.206
182515 93.824 -76.575 54.207
182516 93.839 -76.575 54.206
182517 93.854 -76.577 54.204
182518 93.869 -76.577 54.199
182519 93.885 -76.578 54.191
182520 93.900 -76.580 54.180
182521 93.915 -76.580 54.172
182522 93.930 -76.582 54.165
182523 93.945 -76.583 54.160
182524 93.961 -76.584 54.158
182525 93.977 -76.584 54.154
182526 93.991 -76.585 54.149
182527 94.006 -76.585 54.145
182528 94.020 -76.586 54.143
182529 94.035 -76.586 54.144
182530 94.049 -76.587 54.148
182531 94.064 -76.588 54.153
182532 94.078 -76.589 54.157
182533 94.092 -76.589 54.159
182534 94.107 -76.589 54.159
182535 94.117 -76.589 54.159
182536 94.129 -76.586 54.156
182537 94.143 -76.582 54.150
182538 94.156 -76.578 54.141
182539 94.173 -76.571 54.127
182540 94.188 -76.566 54.111
182541 94.203 -76.560 54.094
182542 94.219 -76.554 54.078
182543 94.235 -76.547 54.062
182544 94.252 -76.541 54.046
182545 94.268 -76.534 54.030
182546 94.284 -76.528 54.004
182547 94.303 -76.526 53.976
182548 94.323 -76.524 53.946
182549 94.340 -76.511 53.914
182550 94.357 -76.498 53.882
182551 94.374 -76.485 53.850
182552 94.392 -76.472 53.816
182553 94.412 -76.461 53.779
182554 94.432 -76.451 53.740
182555 94.451 -76.441 53.700
182556 94.466 -76.436 53.657
182557 94.501 -76.436 53.615
182558 94.533 -76.437 53.574
182559 94.565 -76.439 53.530
182560 94.598 -76.440 53.483
182561 94.630 -76.442 53.436
182562 94.662 -76.443 53.393
182563 94.685 -76.441 53.353
182564 94.703 -76.437 53.319
182565 94.722 -76.434 53.288
182566 94.741 -76.431 53.259
182567 94.759 -76.428 53.231
182568 94.778 -76.426 53.206
182569 94.779 -76.426 53.180
182570 94.775 -76.430 53.154
182571 94.770 -76.432 53.127
182572 94.766 -76.435 53.099
182573 94.761 -76.437 53.068
182574 94.757 -76.440 53.038
182575 94.752 -76.443 53.010
182576 94.748 -76.445 52.984
182577 94.743 -76.448 52.956
182578 94.739 -76.451 52.926
182579 94.734 -76.454 52.893
182580 94.730 -76.457 52.851
182581 94.724 -76.459 52.806
182582 94.720 -76.462 52.757
182583 94.715 -76.464 52.706
182584 94.722 -76.467 52.655
182585 94.730 -76.468 52.606
182586 94.738 -76.470 52.558
182587 94.757 -76.473 52.509
182588 94.794 -76.481 52.460
182589 94.845 -76.494 52.411
182590 94.896 -76.507 52.363
182591 94.947 -76.520 52.311
182592 94.997 -76.533 52.258
182593 95.048 -76.546 52.203
182594 95.093 -76.556 52.147
182595 95.132 -76.564 52.091
182596 95.172 -76.573 52.038
182597 95.208 -76.581 51.987
182598 95.223 -76.596 51.934
182599 95.238 -76.610 51.885
182600 95.252 -76.624 51.836
182601 95.266 -76.638 51.787
182602 95.281 -76.652 51.741
182603 95.296 -76.667 51.695
182604 95.310 -76.681 51.649
182605 95.349 -76.691 51.604
182606 95.391 -76.701 51.561
182607 95.432 -76.711 51.518
182608 95.474 -76.721 51.477
182609 95.516 -76.731 51.437
182610 95.515 -76.736 51.398
182611 95.498 -76.739 51.361
182612 95.505 -76.757 51.326
182613 95.512 -76.775 51.292
182614 95.520 -76.793 51.259
182615 95.527 -76.812 51.226
182616 95.534 -76.830 51.192
182617 95.541 -76.849 51.158
182618 95.548 -76.867 51.122
182619 95.569 -76.884 51.083
182620 95.596 -76.902 51.043
182621 95.622 -76.919 51.004
182622 95.648 -76.929 50.960
182623 95.673 -76.935 50.914
182624 95.698 -76.942 50.868
182625 95.723 -76.949 50.819
182626 95.748 -76.956 50.771
182627 95.778 -76.954 50.723
182628 95.808 -76.953 50.676
182629 95.838 -76.950 50.626
182630 95.871 -76.945 50.572
182631 95.902 -76.940 50.516
182632 95.934 -76.934 50.458
182633 95.966 -76.929 50.397
182634 95.995 -76.927 50.335
182635 96.020 -76.928 50.273
182636 96.039 -76.937 50.209
182637 96.052 -76.953 50.143
182638 96.066 -76.968 50.077
182639 96.080 -76.983 50.011
182640 96.093 -76.998 49.949
182641 96.107 -77.013 49.891
182642 96.120 -77.028 49.837
182643 96.134 -77.044 49.788
182644 96.151 -77.056 49.740
182645 96.176 -77.066 49.694
182646 96.182 -77.079 49.652
182647 96.172 -77.098 49.614
182648 96.162 -77.116 49.579
182649 96.152 -77.135 49.547
182650 96.142 -77.153 49.516
182651 96.154 -77.145 49.488
182652 96.171 -77.151 49.462
182653 96.188 -77.158 49.436
182654 96.205 -77.165 49.410
182655 96.231 -77.166 49.389
182656 96.262 -77.166 49.370
182657 96.292 -77.165 49.351
182658 96.323 -77.164 49.334
182659 96.353 -77.164 49.318
182660 96.384 -77.163 49.302
182661 96.414 -77.163 49.287
182662 96.444 -77.162 49.275
182663 96.475 -77.161 49.262
182664 96.505 -77.160 49.249
182665 96.536 -77.160 49.235
182666 96.564 -77.160 49.221
182667 96.584 -77.165 49.204
182668 96.605 -77.170 49.186
182669 96.625 -77.175 49.168
182670 96.646 -77.179 49.149
182671 96.664 -77.187 49.129
182672 96.680 -77.201 49.109
182673 96.695 -77.214 49.088
182674 96.711 -77.228 49.066
182675 96.726 -77.241 49.045
182676 96.742 -77.254 49.024
182677 96.819 -77.268 49.005
182678 96.846 -77.273 48.985
182679 96.860 -77.278 48.964
182680 96.874 -77.282 48.945
182681 96.888 -77.287 48.924
182682 96.902 -77.291 48.901
182683 96.917 -77.299 48.879
182684 96.932 -77.310 48.855
182685 96.946 -77.321 48.832
182686 96.912 -77.339 48.809
182687 96.891 -77.354 48.786
182688 96.906 -77.359 48.764
182689 96.922 -77.364 48.742
182690 96.937 -77.370 48.721
182691 96.993 -77.357 48.701
182692 97.048 -77.354 48.680
182693 97.082 -77.373 48.659
182694 97.111 -77.398 48.637
182695 97.140 -77.423 48.617
182696 97.169 -77.448 48.597
182697 97.198 -77.472 48.578
182698 97.227 -77.497 48.559
182699 97.253 -77.518 48.538
182700 97.277 -77.537 48.516
182701 97.302 -77.557 48.496
182702 97.326 -77.576 48.476
182703 97.343 -77.581 48.456
182704 97.349 -77.563 48.437
182705 97.354 -77.545 48.417
182706 97.359 -77.527 48.398
182707 97.364 -77.509 48.381
182708 97.369 -77.491 48.364
182709 97.374 -77.473 48.347
182710 97.380 -77.455 48.328
182711 97.387 -77.445 48.307
182712 97.400 -77.464 48.287
182713 97.413 -77.482 48.268
182714 97.426 -77.499 48.248
182715 97.440 -77.514 48.227
182716 97.455 -77.524 48.203
182717 97.473 -77.536 48.177
182718 97.493 -77.549 48.152
182719 97.512 -77.561 48.125
182720 97.531 -77.573 48.097
182721 97.556 -77.577 48.068
182722 97.583 -77.574 48.039
182723 97.610 -77.573 48.009
182724 97.638 -77.571 47.977
182725 97.665 -77.569 47.945
182726 97.708 -77.574 47.913
182727 97.753 -77.581 47.881
182728 97.799 -77.587 47.850
182729 97.844 -77.594 47.817
182730 97.890 -77.600 47.786
182731 97.935 -77.606 47.755
182732 97.972 -77.608 47.725
182733 98.009 -77.609 47.694
182734 98.046 -77.611 47.664
182735 98.063 -77.620 47.635
182736 98.077 -77.599 47.605
182737 98.087 -77.583 47.576
182738 98.096 -77.570 47.548
182739 98.105 -77.556 47.520
182740 98.114 -77.541 47.494
182741 98.124 -77.526 47.467
182742 98.134 -77.511 47.440
182743 98.143 -77.495 47.413
182744 98.152 -77.480 47.387
182745 98.161 -77.465 47.359
182746 98.174 -77.451 47.331
182747 98.223 -77.464 47.304
182748 98.273 -77.476 47.278
182749 98.320 -77.488 47.252
182750 98.341 -77.489 47.227
182751 98.361 -77.490 47.203
182752 98.381 -77.492 47.179
182753 98.401 -77.493 47.155
182754 98.421 -77.495 47.132
182755 98.441 -77.496 47.107
182756 98.461 -77.498 47.082
182757 98.482 -77.499 47.057
182758 98.502 -77.500 47.032
182759 98.522 -77.501 47.007
182760 98.542 -77.503 46.983
182761 98.562 -77.504 46.959
182762 98.557 -77.494 46.934
182763 98.564 -77.514 46.912
182764 98.570 -77.533 46.890
182765 98.577 -77.553 46.869
182766 98.583 -77.573 46.846
182767 98.593 -77.586 46.824
182768 98.613 -77.581 46.801
182769 98.632 -77.577 46.778
182770 98.651 -77.573 46.753
182771 98.670 -77.569 46.729
182772 98.690 -77.564 46.706
182773 98.709 -77.561 46.681
182774 98.728 -77.556 46.657
182775 98.748 -77.551 46.632
182776 98.767 -77.548 46.607
182777 98.787 -77.543 46.583
182778 98.805 -77.541 46.559
182779 98.820 -77.548 46.537
182780 98.834 -77.554 46.516
182781 98.843 -77.561 46.495
182782 98.858 -77.561 46.474
182783 98.874 -77.563 46.454
182784 98.883 -77.568 46.433
182785 98.890 -77.575 46.413
182786 98.899 -77.581 46.394
182787 98.906 -77.588 46.374
182788 98.920 -77.595 46.354
182789 98.939 -77.606 46.332
182790 98.958 -77.617 46.311
182791 98.976 -77.626 46.290
182792 98.994 -77.636 46.269
182793 98.990 -77.648 46.249
182794 98.977 -77.645 46.228
182795 98.982 -77.645 46.206
182796 98.991 -77.647 46.185
182797 98.998 -77.651 46.163
182798 99.006 -77.655 46.141
182799 99.014 -77.658 46.120
182800 99.022 -77.663 46.096
182801 99.029 -77.667 46.074
182802 99.038 -77.671 46.052
182803 99.045 -77.674 46.030
182804 99.053 -77.678 46.010
182805 99.072 -77.696 45.991
182806 99.104 -77.729 45.973
182807 99.135 -77.758 45.955
182808 99.166 -77.746 45.939
182809 99.175 -77.727 45.925
182810 99.183 -77.724 45.910
182811 99.206 -77.747 45.896
182812 99.228 -77.770 45.884
182813 99.252 -77.793 45.873
182814 99.282 -77.814 45.863
182815 99.321 -77.834 45.854
182816 99.359 -77.852 45.844
182817 99.398 -77.872 45.834
182818 99.436 -77.892 45.826
182819 99.475 -77.912 45.816
182820 99.513 -77.931 45.807
182821 99.552 -77.950 45.797
182822 99.590 -77.970 45.787
182823 99.621 -77.984 45.777
182824 99.642 -77.993 45.767
182825 99.662 -78.001 45.757
182826 99.684 -78.009 45.746
182827 99.710 -77.997 45.736
182828 99.736 -77.984 45.726
182829 99.762 -77.975 45.716
182830 99.779 -77.984 45.706
182831 99.796 -77.994 45.696
182832 99.814 -78.004 45.685
182833 99.832 -78.014 45.674
182834 99.849 -78.024 45.664
182835 99.867 -78.034 45.653
182836 99.885 -78.044 45.642
182837 99.902 -78.054 45.630
182838 99.933 -78.084 45.617
182839 99.965 -78.113 45.603
182840 99.997 -78.126 45.589
182841 100.031 -78.135 45.575
182842 100.063 -78.144 45.559
182843 100.096 -78.152 45.543
182844 100.129 -78.160 45.528
182845 100.162 -78.169 45.512
182846 100.195 -78.178 45.496
182847 100.238 -78.166 45.480
182848 100.270 -78.146 45.464
182849 100.275 -78.162 45.450
182850 100.279 -78.178 45.436
182851 100.284 -78.195 45.423
182852 100.290 -78.212 45.409
182853 100.296 -78.228 45.395
182854 100.303 -78.245 45.380
182855 100.310 -78.263 45.364
182856 100.316 -78.279 45.349
182857 100.323 -78.296 45.333
182858 100.340 -78.307 45.317
182859 100.364 -78.313 45.301
182860 100.396 -78.313 45.284
182861 100.427 -78.314 45.267
182862 100.465 -78.317 45.251
182863 100.502 -78.321 45.234
182864 100.541 -78.324 45.218
182865 100.580 -78.326 45.202
182866 100.622 -78.326 45.186
182867 100.667 -78.329 45.171
182868 100.714 -78.337 45.156
182869 100.761 -78.345 45.141
182870 100.808 -78.353 45.126
182871 100.852 -78.360 45.111
182872 100.891 -78.366 45.096
182873 100.930 -78.373 45.082
182874 100.970 -78.380 45.068
182875 101.009 -78.387 45.053
182876 101.042 -78.394 45.038
182877 101.069 -78.401 45.022
182878 101.097 -78.407 45.005
182879 101.124 -78.413 44.989
182880 101.156 -78.405 44.973
182881 101.172 -78.413 44.956
182882 101.188 -78.422 44.941
182883 101.203 -78.430 44.924
182884 101.218 -78.438 44.908
182885 101.234 -78.446 44.891
182886 101.250 -78.454 44.874
182887 101.279 -78.472 44.857
182888 101.309 -78.489 44.841
182889 101.338 -78.506 44.826
182890 101.368 -78.523 44.810
182891 101.398 -78.540 44.794
182892 101.427 -78.557 44.777
182893 101.453 -78.577 44.761
182894 101.476 -78.604 44.745
182895 101.499 -78.632 44.729
182896 101.522 -78.658 44.713
182897 101.545 -78.686 44.698
182898 101.568 -78.718 44.683
182899 101.591 -78.752 44.667
182900 101.614 -78.784 44.651
182901 101.632 -78.806 44.637
182902 101.643 -78.809 44.623
182903 101.654 -78.812 44.608
182904 101.681 -78.828 44.594
182905 101.720 -78.853 44.580
182906 101.759 -78.879 44.566
182907 101.798 -78.904 44.551
182908 101.804 -78.943 44.537
182909 101.822 -78.949 44.523
182910 101.840 -78.955 44.510
182911 101.858 -78.961 44.496
182912 101.877 -78.967 44.483
182913 101.895 -78.974 44.471
182914 101.913 -78.980 44.458
182915 101.932 -78.986 44.446
182916 101.950 -78.992 44.434
182917 101.968 -78.998 44.421
182918 101.986 -79.005 44.408
183173 93.009 -76.737 54.052
183174 93.029 -76.728 54.066
183175 93.049 -76.719 54.079
183176 93.069 -76.715 54.091
183177 93.088 -76.712 54.101
183178 93.107 -76.708 54.111
183179 93.127 -76.705 54.121
183180 93.146 -76.702 54.129
183181 93.165 -76.699 54.136
183182 93.184 -76.695 54.142
183183 93.203 -76.692 54.148
183184 93.222 -76.689 54.152
183185 93.242 -76.685 54.155
183186 93.261 -76.682 54.158
183187 93.281 -76.679 54.161
183188 93.300 -76.677 54.163
183189 93.320 -76.674 54.165
183190 93.339 -76.671 54.167
183191 93.359 -76.668 54.168
183192 93.378 -76.665 54.169
183193 93.398 -76.661 54.170
183194 93.417 -76.658 54.171
183195 93.439 -76.655 54.171
183196 93.468 -76.652 54.173
183197 93.486 -76.655 54.175
183198 93.504 -76.658 54.179
183199 93.523 -76.661 54.180
183200 93.541 -76.664 54.182
183201 93.560 -76.667 54.182
183202 93.578 -76.668 54.183
183203 93.600 -76.658 54.185
183204 93.621 -76.648 54.188
183205 93.642 -76.638 54.192
183206 93.663 -76.628 54.195
183207 93.685 -76.618 54.196
183208 93.706 -76.609 54.198
183209 93.724 -76.602 54.199
183210 93.740 -76.596 54.199
183211 93.754 -76.589 54.200
183212 93.766 -76.582 54.202
183213 93.776 -76.575 54.204
183214 93.786 -76.567 54.207
183215 93.796 -76.560 54.210
183216 93.813 -76.561 54.212
183217 93.832 -76.567 54.214
183218 93.847 -76.567 54.214
183219 93.863 -76.568 54.211
183220 93.878 -76.568 54.205
183221 93.893 -76.569 54.196
183222 93.908 -76.570 54.188
183223 93.924 -76.571 54.180
183224 93.939 -76.571 54.174
183225 93.955 -76.569 54.170
183226 93.964 -76.566 54.165
183227 93.978 -76.567 54.160
183228 93.993 -76.567 54.154
183229 94.008 -76.568 54.151
183230 94.023 -76.568 54.150
183231 94.038 -76.569 54.152
183232 94.053 -76.570 54.158
183233 94.068 -76.571 54.161
183234 94.082 -76.571 54.164
183235 94.097 -76.572 54.167
183236 94.111 -76.572 54.166
183237 94.125 -76.568 54.165
183238 94.139 -76.564 54.162
183239 94.153 -76.560 54.154
183240 94.169 -76.554 54.139
183241 94.185 -76.548 54.123
183242 94.201 -76.542 54.107
183243 94.218 -76.535 54.093
183244 94.234 -76.529 54.077
183245 94.250 -76.522 54.061
183246 94.267 -76.516 54.043
183247 94.284 -76.511 54.016
183248 94.305 -76.510 53.988
183249 94.324 -76.508 53.958
183250 94.343 -76.504 53.926
183251 94.360 -76.490 53.892
183252 94.377 -76.477 53.861
183253 94.394 -76.463 53.828
183254 94.411 -76.449 53.792
183255 94.428 -76.436 53.756
183256 94.446 -76.423 53.716
183257 94.469 -76.417 53.675
183258 94.501 -76.419 53.633
183259 94.534 -76.420 53.593
183260 94.566 -76.422 53.549
183261 94.598 -76.423 53.501
183262 94.630 -76.425 53.455
183263 94.663 -76.426 53.412
183264 94.687 -76.425 53.373
183265 94.705 -76.422 53.337
183266 94.723 -76.420 53.304
183267 94.740 -76.419 53.273
183268 94.757 -76.417 53.243
183269 94.774 -76.416 53.216
183270 94.778 -76.417 53.189
183271 94.774 -76.419 53.162
183272 94.770 -76.421 53.134
183273 94.766 -76.423 53.107
183274 94.762 -76.425 53.079
183275 94.759 -76.427 53.050
183276 94.755 -76.430 53.023
183277 94.751 -76.432 52.998
183278 94.747 -76.434 52.971
183279 94.743 -76.436 52.943
183280 94.739 -76.439 52.911
183281 94.735 -76.441 52.869
183282 94.731 -76.443 52.823
183283 94.726 -76.445 52.774
183284 94.722 -76.448 52.723
183285 94.717 -76.451 52.671
183286 94.713 -76.454 52.621
183287 94.708 -76.457 52.572
183288 94.730 -76.461 52.522
183289 94.780 -76.474 52.473
183290 94.831 -76.487 52.423
183291 94.882 -76.500 52.374
183292 94.933 -76.513 52.324
183293 94.984 -76.526 52.272
183294 95.035 -76.539 52.219
183295 95.078 -76.549 52.165
183296 95.117 -76.557 52.110
183297 95.156 -76.565 52.057
183298 95.195 -76.573 52.004
183299 95.217 -76.586 51.952
183300 95.232 -76.599 51.902
183301 95.247 -76.614 51.853
183302 95.261 -76.628 51.805
183303 95.275 -76.643 51.758
183304 95.290 -76.656 51.712
183305 95.311 -76.670 51.666
183306 95.353 -76.680 51.620
183307 95.394 -76.690 51.577
183308 95.436 -76.699 51.534
183309 95.478 -76.709 51.492
183310 95.520 -76.719 51.452
183311 95.538 -76.726 51.413
183312 95.522 -76.730 51.376
183313 95.529 -76.748 51.341
183314 95.536 -76.766 51.308
183315 95.543 -76.784 51.276
183316 95.550 -76.802 51.243
183317 95.557 -76.821 51.210
183318 95.564 -76.839 51.175
183319 95.571 -76.857 51.139
183320 95.579 -76.875 51.101
183321 95.586 -76.893 51.061
183322 95.612 -76.911 51.021
183323 95.638 -76.924 50.977
183324 95.663 -76.931 50.931
183325 95.688 -76.937 50.885
183326 95.718 -76.937 50.835
183327 95.748 -76.935 50.786
183328 95.778 -76.934 50.740
183329 95.808 -76.932 50.693
183330 95.839 -76.928 50.644
183331 95.871 -76.923 50.591
183332 95.903 -76.918 50.536
183333 95.935 -76.913 50.480
183334 95.963 -76.911 50.420
183335 95.988 -76.912 50.357
183336 96.014 -76.914 50.293
183337 96.033 -76.922 50.229
183338 96.047 -76.937 50.164
183339 96.060 -76.952 50.099
183340 96.074 -76.967 50.032
183341 96.088 -76.982 49.970
183342 96.101 -76.997 49.912
183343 96.115 -77.012 49.859
183344 96.132 -77.025 49.808
183345 96.158 -77.034 49.760
183346 96.183 -77.043 49.714
183347 96.179 -77.059 49.673
183348 96.169 -77.078 49.635
183349 96.159 -77.097 49.600
183350 96.149 -77.115 49.568
183351 96.139 -77.134 49.538
183352 96.129 -77.152 49.511
183353 96.139 -77.162 49.485
183354 96.170 -77.161 49.459
183355 96.200 -77.161 49.434
183356 96.231 -77.160 49.413
183357 96.261 -77.160 49.394
183358 96.292 -77.160 49.376
183359 96.322 -77.159 49.359
183360 96.352 -77.159 49.344
183361 96.382 -77.158 49.328
183362 96.413 -77.158 49.312
183363 96.443 -77.157 49.299
183364 96.473 -77.157 49.287
183365 96.504 -77.157 49.274
183366 96.534 -77.156 49.261
183367 96.560 -77.157 49.246
183368 96.581 -77.162 49.228
183369 96.601 -77.166 49.209
183370 96.622 -77.171 49.191
183371 96.643 -77.176 49.171
183372 96.663 -77.181 49.151
183373 96.682 -77.187 49.131
183374 96.698 -77.201 49.109
183375 96.714 -77.214 49.086
183376 96.729 -77.228 49.063
183377 96.744 -77.241 49.042
183378 96.814 -77.251 49.021
183379 96.828 -77.255 49.001
183380 96.842 -77.260 48.981
183381 96.857 -77.264 48.961
183382 96.871 -77.269 48.940
183383 96.885 -77.273 48.917
183384 96.899 -77.278 48.893
183385 96.913 -77.286 48.869
183386 96.928 -77.298 48.846
183387 96.943 -77.308 48.822
183388 96.935 -77.323 48.799
183389 96.909 -77.338 48.776
183390 96.925 -77.343 48.754
183391 96.940 -77.348 48.733
183392 96.988 -77.340 48.711
183393 97.053 -77.329 48.690
183394 97.086 -77.348 48.668
183395 97.115 -77.373 48.646
183396 97.144 -77.398 48.625
183397 97.173 -77.423 48.605
183398 97.200 -77.446 48.586
183399 97.225 -77.465 48.566
183400 97.249 -77.485 48.546
183401 97.274 -77.504 48.524
183402 97.298 -77.523 48.503
183403 97.322 -77.542 48.482
183404 97.347 -77.562 48.464
183405 97.354 -77.548 48.444
183406 97.360 -77.530 48.425
183407 97.365 -77.512 48.406
183408 97.370 -77.494 48.388
183409 97.375 -77.476 48.371
183410 97.380 -77.458 48.354
183411 97.385 -77.440 48.335
183412 97.396 -77.445 48.314
183413 97.409 -77.463 48.294
183414 97.422 -77.480 48.274
183415 97.435 -77.498 48.254
183416 97.448 -77.517 48.233
183417 97.463 -77.529 48.211
183418 97.477 -77.539 48.187
183419 97.496 -77.551 48.160
183420 97.515 -77.563 48.133
183421 97.537 -77.570 48.105
183422 97.565 -77.569 48.075
183423 97.592 -77.567 48.045
183424 97.619 -77.565 48.013
183425 97.646 -77.564 47.981
183426 97.675 -77.563 47.950
183427 97.721 -77.569 47.918
183428 97.766 -77.576 47.886
183429 97.812 -77.582 47.855
183430 97.857 -77.589 47.823
183431 97.898 -77.592 47.791
183432 97.935 -77.593 47.759
183433 97.972 -77.595 47.728
183434 98.009 -77.596 47.697
183435 98.046 -77.598 47.666
183436 98.072 -77.603 47.636
183437 98.084 -77.584 47.608
183438 98.093 -77.571 47.579
183439 98.103 -77.557 47.550
183440 98.112 -77.542 47.522
183441 98.122 -77.526 47.495
183442 98.131 -77.511 47.467
183443 98.140 -77.496 47.440
183444 98.150 -77.481 47.413
183445 98.159 -77.466 47.387
183446 98.168 -77.451 47.360
183447 98.177 -77.436 47.332
183448 98.214 -77.439 47.305
183449 98.264 -77.451 47.279
183450 98.313 -77.464 47.254
183451 98.363 -77.476 47.229
183452 98.384 -77.478 47.205
183453 98.405 -77.479 47.182
183454 98.425 -77.480 47.160
183455 98.445 -77.482 47.137
183456 98.465 -77.483 47.113
183457 98.485 -77.485 47.089
183458 98.505 -77.486 47.065
183459 98.525 -77.488 47.040
183460 98.546 -77.489 47.015
183461 98.566 -77.490 46.991
183462 98.558 -77.469 46.967
183463 98.565 -77.489 46.943
183464 98.571 -77.508 46.920
183465 98.578 -77.528 46.899
183466 98.584 -77.548 46.877
183467 98.598 -77.564 46.853
183468 98.610 -77.567 46.831
183469 98.622 -77.564 46.808
183470 98.641 -77.559 46.784
183471 98.660 -77.555 46.760
183472 98.679 -77.551 46.736
183473 98.699 -77.547 46.712
183474 98.718 -77.542 46.688
183475 98.737 -77.539 46.664
183476 98.757 -77.534 46.639
183477 98.776 -77.530 46.614
183478 98.795 -77.526 46.590
183479 98.813 -77.524 46.566
183480 98.828 -77.531 46.543
183481 98.843 -77.537 46.521
183482 98.857 -77.544 46.501
183483 98.870 -77.550 46.480
183484 98.885 -77.552 46.459
183485 98.893 -77.558 46.438
183486 98.897 -77.565 46.419
183487 98.896 -77.572 46.400
183488 98.895 -77.578 46.380
183489 98.940 -77.586 46.358
183490 98.959 -77.597 46.338
183491 98.979 -77.608 46.316
183492 98.998 -77.618 46.294
183493 99.018 -77.628 46.274
183494 99.026 -77.639 46.254
183495 98.999 -77.642 46.233
183496 98.987 -77.643 46.212
183497 98.997 -77.643 46.189
183498 99.007 -77.643 46.166
183499 99.016 -77.643 46.144
183500 99.026 -77.643 46.123
183501 99.037 -77.643 46.100
183502 99.046 -77.644 46.077
183503 99.054 -77.648 46.055
183504 99.062 -77.652 46.034
183505 99.069 -77.656 46.014
183506 99.077 -77.661 45.995
183507 99.110 -77.692 45.977
183508 99.141 -77.718 45.958
183509 99.164 -77.724 45.942
183510 99.164 -77.708 45.927
183511 99.175 -77.712 45.912
183512 99.198 -77.736 45.898
183513 99.221 -77.758 45.886
183514 99.256 -77.779 45.875
183515 99.294 -77.798 45.864
183516 99.333 -77.818 45.855
183517 99.371 -77.837 45.845
183518 99.410 -77.857 45.836
183519 99.448 -77.876 45.826
183520 99.487 -77.896 45.817
183521 99.525 -77.915 45.807
183522 99.563 -77.935 45.796
183523 99.602 -77.954 45.786
183524 99.640 -77.974 45.776
183525 99.664 -77.984 45.765
183526 99.685 -77.993 45.754
183527 99.709 -77.990 45.744
183528 99.735 -77.976 45.733
183529 99.756 -77.978 45.724
183530 99.774 -77.988 45.714
183531 99.791 -77.998 45.705
183532 99.808 -78.008 45.694
183533 99.826 -78.018 45.682
183534 99.844 -78.028 45.672
183535 99.861 -78.038 45.662
183536 99.879 -78.048 45.652
183537 99.897 -78.058 45.641
183538 99.914 -78.068 45.628
183539 99.936 -78.084 45.615
183540 99.969 -78.104 45.601
183541 100.001 -78.112 45.586
183542 100.034 -78.122 45.572
183543 100.067 -78.130 45.557
183544 100.100 -78.138 45.543
183545 100.133 -78.147 45.527
183546 100.166 -78.156 45.512
183547 100.199 -78.164 45.496
183548 100.234 -78.166 45.480
183549 100.284 -78.124 45.465
183550 100.289 -78.141 45.451
183551 100.293 -78.156 45.437
183552 100.298 -78.173 45.424
183553 100.305 -78.190 45.410
183554 100.312 -78.207 45.396
183555 100.318 -78.224 45.381
183556 100.325 -78.241 45.365
183557 100.332 -78.258 45.350
183558 100.338 -78.275 45.335
183559 100.348 -78.290 45.319
183560 100.370 -78.297 45.304
183561 100.397 -78.301 45.287
183562 100.429 -78.303 45.270
183563 100.467 -78.306 45.254
183564 100.506 -78.309 45.237
183565 100.548 -78.309 45.221
183566 100.590 -78.309 45.205
183567 100.632 -78.309 45.188
183568 100.678 -78.316 45.174
183569 100.725 -78.324 45.160
183570 100.772 -78.332 45.145
183571 100.819 -78.339 45.130
183572 100.865 -78.347 45.116
183573 100.913 -78.355 45.102
183574 100.957 -78.363 45.087
183575 100.997 -78.369 45.071
183576 101.034 -78.376 45.056
183577 101.061 -78.382 45.041
183578 101.089 -78.389 45.026
183579 101.116 -78.396 45.009
183580 101.145 -78.401 44.992
183581 101.165 -78.404 44.976
183582 101.181 -78.413 44.959
183583 101.196 -78.420 44.943
183584 101.212 -78.429 44.927
183585 101.228 -78.437 44.910
183586 101.243 -78.445 44.895
183587 101.258 -78.453 44.878
183588 101.274 -78.462 44.862
183589 101.304 -78.479 44.845
183590 101.334 -78.496 44.830
183591 101.363 -78.514 44.814
183592 101.393 -78.530 44.798
183593 101.423 -78.548 44.782
183594 101.453 -78.571 44.766
183595 101.476 -78.598 44.750
183596 101.499 -78.624 44.734
183597 101.522 -78.651 44.717
183598 101.545 -78.679 44.702
183599 101.568 -78.711 44.686
183600 101.591 -78.744 44.671
183601 101.614 -78.777 44.656
183602 101.625 -78.781 44.641
183603 101.636 -78.784 44.627
183604 101.665 -78.801 44.613
183605 101.703 -78.827 44.598
183606 101.742 -78.852 44.583
183607 101.781 -78.877 44.568
183608 101.820 -78.902 44.554
183609 101.822 -78.950 44.540
183610 101.840 -78.957 44.526
183611 101.859 -78.963 44.512
183612 101.877 -78.969 44.499
183613 101.895 -78.975 44.486
183614 101.914 -78.981 44.473
183615 101.932 -78.988 44.460
183616 101.950 -78.994 44.448
183617 101.968 -79.000 44.436
183618 101.987 -79.006 44.423
183619 102.005 -79.012 44.410
183875 93.018 -76.705 54.049
183876 93.039 -76.696 54.063
183877 93.058 -76.691 54.076
183878 93.077 -76.688 54.087
183879 93.097 -76.684 54.098
183880 93.116 -76.681 54.108
183881 93.135 -76.678 54.117
183882 93.154 -76.674 54.125
183883 93.174 -76.671 54.132
183884 93.193 -76.668 54.138
183885 93.212 -76.665 54.143
183886 93.231 -76.661 54.147
183887 93.251 -76.658 54.150
183888 93.270 -76.655 54.153
183889 93.290 -76.652 54.155
183890 93.309 -76.649 54.157
183891 93.329 -76.646 54.159
183892 93.348 -76.643 54.161
183893 93.368 -76.640 54.161
183894 93.387 -76.638 54.161
183895 93.407 -76.635 54.162
183896 93.432 -76.631 54.162
183897 93.460 -76.629 54.163
183898 93.478 -76.632 54.165
183899 93.496 -76.635 54.167
183900 93.515 -76.637 54.167
183901 93.533 -76.640 54.168
183902 93.552 -76.643 54.168
183903 93.570 -76.646 54.168
183904 93.588 -76.649 54.170
183905 93.610 -76.639 54.174
183906 93.631 -76.629 54.177
183907 93.652 -76.619 54.181
183908 93.673 -76.609 54.184
183909 93.694 -76.599 54.187
183910 93.715 -76.590 54.191
183911 93.734 -76.580 54.193
183912 93.737 -76.567 54.196
183913 93.747 -76.560 54.200
183914 93.757 -76.552 54.203
183915 93.767 -76.545 54.208
183916 93.777 -76.537 54.212
183917 93.803 -76.550 54.216
183918 93.825 -76.558 54.219
183919 93.840 -76.559 54.221
183920 93.856 -76.560 54.221
183921 93.871 -76.561 54.217
183922 93.886 -76.561 54.211
183923 93.901 -76.561 54.203
183924 93.917 -76.562 54.195
183925 93.933 -76.557 54.188
183926 93.948 -76.552 54.182
183927 93.952 -76.549 54.176
183928 93.963 -76.549 54.170
183929 93.978 -76.549 54.163
183930 93.993 -76.550 54.159
183931 94.008 -76.551 54.157
183932 94.023 -76.552 54.157
183933 94.038 -76.552 54.162
183934 94.053 -76.552 54.165
183935 94.066 -76.555 54.168
183936 94.078 -76.558 54.171
183937 94.105 -76.553 54.171
183938 94.124 -76.549 54.170
183939 94.138 -76.545 54.167
183940 94.152 -76.541 54.160
183941 94.168 -76.536 54.148
183942 94.184 -76.529 54.133
183943 94.200 -76.523 54.118
183944 94.216 -76.516 54.103
183945 94.232 -76.510 54.088
183946 94.249 -76.503 54.071
183947 94.266 -76.498 54.052
183948 94.285 -76.496 54.025
183949 94.305 -76.495 53.996
183950 94.325 -76.493 53.965
183951 94.345 -76.491 53.932
183952 94.362 -76.482 53.900
183953 94.380 -76.468 53.868
183954 94.397 -76.455 53.836
183955 94.414 -76.441 53.801
183956 94.430 -76.427 53.765
183957 94.448 -76.414 53.727
183958 94.470 -76.410 53.687
183959 94.496 -76.411 53.647
183960 94.524 -76.413 53.605
183961 94.556 -76.414 53.562
183962 94.588 -76.417 53.516
183963 94.620 -76.418 53.472
183964 94.652 -76.420 53.431
183965 94.680 -76.420 53.392
183966 94.697 -76.419 53.357
183967 94.714 -76.417 53.321
183968 94.731 -76.416 53.287
183969 94.748 -76.414 53.254
183970 94.764 -76.412 53.225
183971 94.769 -76.412 53.196
183972 94.766 -76.414 53.167
183973 94.762 -76.417 53.139
183974 94.758 -76.419 53.114
183975 94.754 -76.421 53.087
183976 94.750 -76.423 53.061
183977 94.746 -76.426 53.035
183978 94.742 -76.428 53.011
183979 94.738 -76.430 52.986
183980 94.734 -76.433 52.960
183981 94.730 -76.435 52.930
183982 94.726 -76.437 52.889
183983 94.722 -76.439 52.843
183984 94.718 -76.442 52.793
183985 94.714 -76.443 52.742
183986 94.710 -76.445 52.689
183987 94.706 -76.448 52.638
183988 94.709 -76.451 52.588
183989 94.718 -76.455 52.537
183990 94.767 -76.467 52.487
183991 94.818 -76.480 52.438
183992 94.869 -76.493 52.388
183993 94.920 -76.506 52.337
183994 94.971 -76.519 52.286
183995 95.022 -76.532 52.234
183996 95.062 -76.541 52.181
183997 95.102 -76.549 52.128
183998 95.140 -76.557 52.076
183999 95.180 -76.565 52.024
184000 95.212 -76.575 51.972
184001 95.227 -76.589 51.922
184002 95.241 -76.603 51.874
184003 95.256 -76.618 51.825
184004 95.270 -76.632 51.777
184005 95.284 -76.646 51.730
184006 95.315 -76.658 51.684
184007 95.356 -76.668 51.638
184008 95.398 -76.677 51.594
184009 95.440 -76.687 51.551
184010 95.481 -76.698 51.509
184011 95.523 -76.708 51.469
184012 95.561 -76.717 51.430
184013 95.545 -76.721 51.393
184014 95.552 -76.739 51.357
184015 95.559 -76.757 51.324
184016 95.566 -76.775 51.291
184017 95.573 -76.793 51.259
184018 95.581 -76.812 51.226
184019 95.588 -76.830 51.191
184020 95.595 -76.848 51.154
184021 95.602 -76.866 51.116
184022 95.609 -76.884 51.076
184023 95.616 -76.903 51.035
184024 95.629 -76.919 50.992
184025 95.658 -76.918 50.946
184026 95.688 -76.917 50.897
184027 95.718 -76.915 50.849
184028 95.748 -76.915 50.802
184029 95.778 -76.913 50.757
184030 95.808 -76.912 50.711
184031 95.839 -76.907 50.662
184032 95.872 -76.902 50.610
184033 95.903 -76.897 50.556
184034 95.931 -76.896 50.500
184035 95.956 -76.897 50.440
184036 95.982 -76.898 50.377
184037 96.007 -76.899 50.315
184038 96.028 -76.906 50.252
184039 96.042 -76.921 50.187
184040 96.055 -76.936 50.121
184041 96.068 -76.951 50.056
184042 96.082 -76.966 49.994
184043 96.096 -76.981 49.935
184044 96.114 -76.994 49.881
184045 96.139 -77.003 49.830
184046 96.164 -77.012 49.782
184047 96.187 -77.021 49.737
184048 96.177 -77.040 49.695
184049 96.167 -77.058 49.657
184050 96.157 -77.077 49.623
184051 96.147 -77.095 49.591
184052 96.136 -77.113 49.561
184053 96.136 -77.129 49.535
184054 96.165 -77.125 49.509
184055 96.182 -77.136 49.483
184056 96.203 -77.144 49.459
184057 96.233 -77.144 49.438
184058 96.264 -77.143 49.420
184059 96.294 -77.142 49.402
184060 96.324 -77.142 49.386
184061 96.355 -77.141 49.371
184062 96.385 -77.141 49.355
184063 96.415 -77.141 49.339
184064 96.446 -77.140 49.326
184065 96.476 -77.140 49.313
184066 96.507 -77.139 49.300
184067 96.537 -77.139 49.286
184068 96.558 -77.148 49.270
184069 96.579 -77.155 49.251
184070 96.599 -77.162 49.231
184071 96.619 -77.168 49.213
184072 96.640 -77.172 49.193
184073 96.660 -77.177 49.174
184074 96.680 -77.182 49.153
184075 96.700 -77.187 49.130
184076 96.716 -77.201 49.106
184077 96.732 -77.214 49.082
184078 96.782 -77.227 49.060
184079 96.796 -77.232 49.039
184080 96.810 -77.237 49.018
184081 96.824 -77.241 48.996
184082 96.839 -77.246 48.976
184083 96.853 -77.251 48.954
184084 96.867 -77.255 48.931
184085 96.882 -77.260 48.907
184086 96.896 -77.264 48.883
184087 96.910 -77.273 48.859
184088 96.925 -77.285 48.835
184089 96.939 -77.295 48.811
184090 96.954 -77.307 48.788
184091 96.928 -77.323 48.766
184092 96.944 -77.328 48.744
184093 96.983 -77.323 48.721
184094 97.057 -77.304 48.699
184095 97.089 -77.324 48.677
184096 97.118 -77.349 48.655
184097 97.148 -77.373 48.633
184098 97.172 -77.393 48.613
184099 97.196 -77.413 48.593
184100 97.221 -77.432 48.572
184101 97.246 -77.451 48.553
184102 97.270 -77.470 48.532
184103 97.295 -77.490 48.511
184104 97.319 -77.510 48.490
184105 97.344 -77.529 48.471
184106 97.360 -77.533 48.451
184107 97.365 -77.515 48.432
184108 97.371 -77.498 48.413
184109 97.376 -77.479 48.395
184110 97.381 -77.461 48.377
184111 97.386 -77.444 48.360
184112 97.391 -77.426 48.342
184113 97.404 -77.444 48.321
184114 97.418 -77.462 48.301
184115 97.431 -77.479 48.280
184116 97.444 -77.498 48.260
184117 97.457 -77.515 48.239
184118 97.470 -77.533 48.217
184119 97.486 -77.542 48.193
184120 97.500 -77.552 48.167
184121 97.519 -77.565 48.140
184122 97.546 -77.564 48.112
184123 97.573 -77.561 48.082
184124 97.600 -77.560 48.051
184125 97.628 -77.558 48.019
184126 97.655 -77.556 47.986
184127 97.688 -77.558 47.955
184128 97.734 -77.564 47.923
184129 97.780 -77.570 47.891
184130 97.824 -77.577 47.860
184131 97.861 -77.577 47.829
184132 97.898 -77.579 47.797
184133 97.935 -77.580 47.764
184134 97.972 -77.582 47.732
184135 98.009 -77.583 47.701
184136 98.046 -77.584 47.670
184137 98.080 -77.587 47.639
184138 98.091 -77.573 47.610
184139 98.100 -77.558 47.582
184140 98.109 -77.543 47.554
184141 98.119 -77.528 47.526
184142 98.128 -77.513 47.497
184143 98.138 -77.498 47.468
184144 98.147 -77.483 47.441
184145 98.156 -77.467 47.414
184146 98.165 -77.452 47.388
184147 98.175 -77.437 47.360
184148 98.184 -77.422 47.333
184149 98.205 -77.414 47.307
184150 98.254 -77.426 47.281
184151 98.304 -77.439 47.255
184152 98.354 -77.451 47.231
184153 98.404 -77.464 47.207
184154 98.428 -77.467 47.186
184155 98.448 -77.468 47.164
184156 98.468 -77.469 47.141
184157 98.488 -77.470 47.118
184158 98.508 -77.472 47.094
184159 98.528 -77.473 47.070
184160 98.549 -77.475 47.047
184161 98.569 -77.476 47.023
184162 98.563 -77.452 46.999
184163 98.566 -77.463 46.975
184164 98.572 -77.483 46.952
184165 98.579 -77.502 46.929
184166 98.589 -77.520 46.906
184167 98.602 -77.536 46.883
184168 98.616 -77.553 46.859
184169 98.627 -77.549 46.837
184170 98.639 -77.545 46.813
184171 98.650 -77.542 46.790
184172 98.669 -77.537 46.765
184173 98.688 -77.533 46.741
184174 98.708 -77.529 46.718
184175 98.727 -77.525 46.694
184176 98.746 -77.520 46.669
184177 98.766 -77.517 46.645
184178 98.785 -77.512 46.620
184179 98.804 -77.508 46.595
184180 98.822 -77.508 46.572
184181 98.837 -77.514 46.549
184182 98.851 -77.520 46.527
184183 98.866 -77.527 46.507
184184 98.880 -77.533 46.486
184185 98.889 -77.542 46.465
184186 98.886 -77.548 46.444
184187 98.885 -77.555 46.424
184188 98.883 -77.562 46.404
184189 98.940 -77.567 46.384
184190 98.959 -77.577 46.363
184191 98.979 -77.588 46.341
184192 98.999 -77.599 46.320
184193 99.018 -77.608 46.298
184194 99.032 -77.617 46.278
184195 99.030 -77.630 46.257
184196 99.021 -77.641 46.236
184197 98.992 -77.641 46.215
184198 99.001 -77.642 46.192
184199 99.011 -77.642 46.169
184200 99.021 -77.642 46.147
184201 99.031 -77.642 46.125
184202 99.041 -77.642 46.102
184203 99.051 -77.642 46.080
184204 99.061 -77.642 46.058
184205 99.071 -77.642 46.037
184206 99.081 -77.642 46.017
184207 99.091 -77.642 45.998
184208 99.104 -77.652 45.980
184209 99.127 -77.677 45.962
184210 99.150 -77.702 45.945
184211 99.152 -77.690 45.929
184212 99.167 -77.702 45.914
184213 99.191 -77.724 45.900
184214 99.229 -77.743 45.888
184215 99.268 -77.763 45.877
184216 99.306 -77.783 45.865
184217 99.345 -77.802 45.855
184218 99.383 -77.822 45.846
184219 99.421 -77.841 45.837
184220 99.460 -77.861 45.827
184221 99.498 -77.881 45.817
184222 99.537 -77.900 45.806
184223 99.575 -77.919 45.795
184224 99.614 -77.939 45.785
184225 99.652 -77.959 45.774
184226 99.687 -77.975 45.762
184227 99.708 -77.982 45.751
184228 99.733 -77.972 45.741
184229 99.751 -77.982 45.731
184230 99.768 -77.992 45.721
184231 99.786 -78.002 45.712
184232 99.803 -78.012 45.703
184233 99.821 -78.022 45.692
184234 99.838 -78.032 45.681
184235 99.856 -78.042 45.669
184236 99.873 -78.052 45.660
184237 99.891 -78.059 45.650
184238 99.907 -78.062 45.639
184239 99.925 -78.070 45.625
184240 99.946 -78.081 45.612
184241 99.977 -78.086 45.599
184242 100.005 -78.099 45.585
184243 100.039 -78.108 45.570
184244 100.071 -78.116 45.556
184245 100.104 -78.125 45.541
184246 100.137 -78.133 45.527
184247 100.170 -78.142 45.511
184248 100.203 -78.150 45.496
184249 100.235 -78.159 45.480
184250 100.298 -78.102 45.466
184251 100.302 -78.119 45.452
184252 100.308 -78.134 45.438
184253 100.314 -78.152 45.425
184254 100.320 -78.169 45.412
184255 100.327 -78.185 45.397
184256 100.334 -78.203 45.382
184257 100.340 -78.219 45.366
184258 100.347 -78.237 45.351
184259 100.353 -78.253 45.335
184260 100.360 -78.270 45.320
184261 100.378 -78.281 45.304
184262 100.400 -78.288 45.288
184263 100.432 -78.291 45.272
184264 100.473 -78.291 45.256
184265 100.515 -78.292 45.240
184266 100.557 -78.292 45.223
184267 100.599 -78.292 45.208
184268 100.642 -78.295 45.192
184269 100.690 -78.303 45.177
184270 100.737 -78.310 45.162
184271 100.783 -78.319 45.148
184272 100.830 -78.326 45.134
184273 100.877 -78.334 45.119
184274 100.924 -78.341 45.105
184275 100.971 -78.349 45.090
184276 101.018 -78.357 45.075
184277 101.051 -78.363 45.059
184278 101.078 -78.370 45.044
184279 101.105 -78.377 45.028
184280 101.132 -78.384 45.011
184281 101.158 -78.391 44.995
184282 101.174 -78.402 44.978
184283 101.190 -78.411 44.962
184284 101.205 -78.420 44.946
184285 101.220 -78.428 44.930
184286 101.236 -78.435 44.914
184287 101.252 -78.444 44.898
184288 101.267 -78.452 44.882
184289 101.282 -78.460 44.866
184290 101.299 -78.470 44.850
184291 101.329 -78.486 44.834
184292 101.358 -78.504 44.818
184293 101.388 -78.521 44.802
184294 101.418 -78.544 44.786
184295 101.448 -78.567 44.771
184296 101.476 -78.590 44.755
184297 101.499 -78.617 44.738
184298 101.522 -78.645 44.722
184299 101.545 -78.672 44.706
184300 101.568 -78.705 44.690
184301 101.591 -78.737 44.674
184302 101.607 -78.753 44.658
184303 101.618 -78.756 44.644
184304 101.648 -78.774 44.630
184305 101.687 -78.799 44.615
184306 101.726 -78.825 44.601
184307 101.764 -78.850 44.586
184308 101.803 -78.875 44.571
184309 101.842 -78.901 44.557
184310 101.841 -78.959 44.543
184311 101.859 -78.965 44.529
184312 101.878 -78.971 44.515
184313 101.896 -78.977 44.502
184314 101.914 -78.984 44.488
184315 101.933 -78.990 44.475
184316 101.951 -78.996 44.462
184317 101.969 -79.002 44.449
184318 101.987 -79.009 44.437
184319 102.006 -79.015 44.425
184320 102.024 -79.021 44.412
184582 93.125 -76.654 54.104
184583 93.144 -76.651 54.113
184584 93.163 -76.648 54.120
184585 93.183 -76.644 54.127
184586 93.202 -76.641 54.133
184587 93.221 -76.638 54.138
184588 93.240 -76.635 54.141
184589 93.260 -76.632 54.144
184590 93.279 -76.629 54.147
184591 93.299 -76.626 54.149
184592 93.319 -76.623 54.151
184593 93.338 -76.620 54.153
184594 93.357 -76.617 54.153
184595 93.377 -76.614 54.153
184596 93.396 -76.611 54.153
184597 93.425 -76.607 54.154
184598 93.451 -76.605 54.155
184599 93.470 -76.608 54.156
184600 93.488 -76.611 54.157
184601 93.507 -76.614 54.157
184602 93.525 -76.617 54.157
184603 93.543 -76.620 54.157
184604 93.562 -76.623 54.158
184605 93.580 -76.626 54.159
184606 93.599 -76.629 54.162
184607 93.620 -76.620 54.165
184608 93.641 -76.610 54.167
184609 93.662 -76.600 54.171
184610 93.683 -76.590 54.175
184611 93.704 -76.580 54.180
184612 93.725 -76.571 54.185
184613 93.735 -76.560 54.192
184614 93.728 -76.537 54.199
184615 93.738 -76.530 54.203
184616 93.748 -76.522 54.207
184617 93.767 -76.526 54.212
184618 93.794 -76.539 54.217
184619 93.818 -76.550 54.222
184620 93.834 -76.551 54.226
184621 93.849 -76.552 54.228
184622 93.864 -76.552 54.227
184623 93.879 -76.553 54.223
184624 93.895 -76.550 54.217
184625 93.911 -76.546 54.209
184626 93.927 -76.540 54.202
184627 93.938 -76.536 54.195
184628 93.942 -76.533 54.188
184629 93.948 -76.531 54.180
184630 93.963 -76.532 54.173
184631 93.978 -76.533 54.168
184632 93.993 -76.533 54.164
184633 94.007 -76.534 54.162
184634 94.019 -76.537 54.166
184635 94.031 -76.540 54.168
184636 94.043 -76.543 54.169
184637 94.068 -76.540 54.172
184638 94.096 -76.535 54.173
184639 94.123 -76.530 54.172
184640 94.137 -76.526 54.170
184641 94.151 -76.522 54.165
184642 94.167 -76.517 54.155
184643 94.183 -76.511 54.142
184644 94.199 -76.504 54.127
184645 94.215 -76.498 54.113
184646 94.231 -76.491 54.097
184647 94.248 -76.485 54.079
184648 94.267 -76.482 54.060
184649 94.286 -76.480 54.033
184650 94.306 -76.479 54.003
184651 94.326 -76.477 53.972
184652 94.346 -76.476 53.941
184653 94.365 -76.473 53.908
184654 94.382 -76.460 53.876
184655 94.399 -76.446 53.843
184656 94.416 -76.433 53.807
184657 94.433 -76.419 53.771
184658 94.450 -76.406 53.734
184659 94.475 -76.405 53.694
184660 94.501 -76.408 53.653
184661 94.528 -76.409 53.611
184662 94.554 -76.411 53.568
184663 94.580 -76.412 53.525
184664 94.606 -76.414 53.485
184665 94.638 -76.416 53.447
184666 94.670 -76.417 53.411
184667 94.688 -76.416 53.374
184668 94.704 -76.414 53.337
184669 94.721 -76.412 53.301
184670 94.738 -76.410 53.266
184671 94.754 -76.408 53.235
184672 94.761 -76.408 53.205
184673 94.757 -76.410 53.176
184674 94.753 -76.412 53.147
184675 94.749 -76.414 53.121
184676 94.745 -76.417 53.095
184677 94.741 -76.419 53.069
184678 94.737 -76.421 53.045
184679 94.733 -76.423 53.024
184680 94.729 -76.426 53.002
184681 94.725 -76.428 52.978
184682 94.721 -76.430 52.950
184683 94.717 -76.433 52.911
184684 94.713 -76.435 52.866
184685 94.709 -76.437 52.817
184686 94.705 -76.439 52.765
184687 94.701 -76.442 52.711
184688 94.709 -76.445 52.658
184689 94.718 -76.449 52.606
184690 94.728 -76.453 52.554
184691 94.754 -76.461 52.505
184692 94.805 -76.473 52.456
184693 94.856 -76.486 52.405
184694 94.907 -76.499 52.352
184695 94.958 -76.512 52.299
184696 95.007 -76.525 52.247
184697 95.047 -76.533 52.196
184698 95.086 -76.541 52.145
184699 95.125 -76.549 52.095
184700 95.164 -76.558 52.044
184701 95.203 -76.565 51.992
184702 95.221 -76.579 51.944
184703 95.236 -76.593 51.896
184704 95.250 -76.608 51.845
184705 95.264 -76.621 51.797
184706 95.279 -76.636 51.750
184707 95.318 -76.646 51.703
184708 95.360 -76.656 51.658
184709 95.402 -76.666 51.612
184710 95.444 -76.676 51.569
184711 95.485 -76.686 51.527
184712 95.527 -76.696 51.486
184713 95.569 -76.705 51.446
184714 95.568 -76.712 51.409
184715 95.576 -76.730 51.373
184716 95.583 -76.748 51.340
184717 95.590 -76.766 51.307
184718 95.597 -76.784 51.274
184719 95.604 -76.802 51.241
184720 95.611 -76.821 51.205
184721 95.618 -76.839 51.167
184722 95.620 -76.856 51.129
184723 95.617 -76.871 51.088
184724 95.613 -76.887 51.047
184725 95.631 -76.894 51.004
184726 95.659 -76.897 50.958
184727 95.688 -76.897 50.909
184728 95.718 -76.895 50.862
184729 95.748 -76.893 50.817
184730 95.778 -76.892 50.774
184731 95.808 -76.890 50.728
184732 95.840 -76.885 50.680
184733 95.872 -76.880 50.628
184734 95.899 -76.880 50.575
184735 95.924 -76.881 50.518
184736 95.950 -76.882 50.458
184737 95.975 -76.884 50.396
184738 96.001 -76.884 50.336
184739 96.023 -76.890 50.275
184740 96.036 -76.905 50.211
184741 96.050 -76.920 50.145
184742 96.063 -76.935 50.081
184743 96.077 -76.950 50.018
184744 96.095 -76.962 49.959
184745 96.120 -76.972 49.904
184746 96.146 -76.980 49.853
184747 96.171 -76.989 49.805
184748 96.184 -77.001 49.760
184749 96.174 -77.019 49.719
184750 96.164 -77.038 49.681
184751 96.154 -77.056 49.647
184752 96.144 -77.075 49.615
184753 96.142 -77.091 49.585
184754 96.151 -77.105 49.559
184755 96.161 -77.119 49.533
184756 96.214 -77.094 49.508
184757 96.228 -77.108 49.484
184758 96.242 -77.122 49.464
184759 96.266 -77.125 49.446
184760 96.297 -77.125 49.429
184761 96.327 -77.125 49.413
184762 96.358 -77.125 49.398
184763 96.388 -77.124 49.382
184764 96.418 -77.123 49.366
184765 96.449 -77.123 49.353
184766 96.479 -77.122 49.340
184767 96.510 -77.122 49.327
184768 96.540 -77.115 49.312
184769 96.558 -77.131 49.295
184770 96.577 -77.144 49.275
184771 96.597 -77.151 49.255
184772 96.617 -77.158 49.235
184773 96.637 -77.165 49.216
184774 96.657 -77.172 49.196
184775 96.677 -77.178 49.175
184776 96.697 -77.184 49.151
184777 96.718 -77.188 49.126
184778 96.734 -77.201 49.102
184779 96.764 -77.209 49.079
184780 96.779 -77.213 49.057
184781 96.793 -77.218 49.035
184782 96.807 -77.222 49.013
184783 96.821 -77.227 48.991
184784 96.835 -77.232 48.969
184785 96.850 -77.237 48.945
184786 96.864 -77.241 48.921
184787 96.878 -77.246 48.897
184788 96.892 -77.251 48.872
184789 96.907 -77.260 48.848
184790 96.922 -77.272 48.823
184791 96.936 -77.282 48.800
184792 96.951 -77.294 48.777
184793 96.947 -77.307 48.756
184794 96.979 -77.305 48.733
184795 97.054 -77.285 48.710
184796 97.093 -77.299 48.687
184797 97.120 -77.321 48.665
184798 97.144 -77.341 48.643
184799 97.169 -77.360 48.621
184800 97.193 -77.379 48.600
184801 97.218 -77.398 48.579
184802 97.242 -77.418 48.560
184803 97.266 -77.438 48.539
184804 97.291 -77.457 48.519
184805 97.315 -77.476 48.498
184806 97.340 -77.495 48.478
184807 97.364 -77.515 48.458
184808 97.371 -77.501 48.439
184809 97.377 -77.483 48.420
184810 97.382 -77.464 48.402
184811 97.387 -77.447 48.384
184812 97.392 -77.429 48.366
184813 97.400 -77.425 48.348
184814 97.414 -77.443 48.328
184815 97.427 -77.461 48.307
184816 97.440 -77.479 48.287
184817 97.453 -77.497 48.266
184818 97.466 -77.514 48.245
184819 97.479 -77.533 48.223
184820 97.494 -77.547 48.199
184821 97.508 -77.557 48.173
184822 97.525 -77.552 48.146
184823 97.555 -77.556 48.116
184824 97.582 -77.555 48.086
184825 97.609 -77.553 48.056
184826 97.637 -77.551 48.024
184827 97.664 -77.549 47.991
184828 97.702 -77.552 47.959
184829 97.747 -77.558 47.928
184830 97.787 -77.562 47.897
184831 97.824 -77.563 47.866
184832 97.861 -77.564 47.835
184833 97.898 -77.566 47.802
184834 97.935 -77.567 47.770
184835 97.972 -77.569 47.737
184836 98.009 -77.570 47.706
184837 98.046 -77.567 47.674
184838 98.083 -77.564 47.643
184839 98.097 -77.559 47.614
184840 98.107 -77.544 47.585
184841 98.116 -77.529 47.558
184842 98.126 -77.514 47.530
184843 98.135 -77.498 47.501
184844 98.144 -77.483 47.473
184845 98.154 -77.468 47.445
184846 98.163 -77.453 47.417
184847 98.172 -77.438 47.389
184848 98.181 -77.423 47.362
184849 98.191 -77.407 47.335
184850 98.201 -77.392 47.308
184851 98.245 -77.401 47.282
184852 98.295 -77.414 47.257
184853 98.345 -77.426 47.232
184854 98.394 -77.439 47.209
184855 98.444 -77.451 47.187
184856 98.472 -77.455 47.165
184857 98.492 -77.457 47.143
184858 98.512 -77.457 47.120
184859 98.532 -77.459 47.098
184860 98.552 -77.461 47.075
184861 98.572 -77.462 47.053
184862 98.572 -77.443 47.029
184863 98.567 -77.438 47.005
184864 98.573 -77.457 46.982
184865 98.581 -77.476 46.958
184866 98.594 -77.493 46.934
184867 98.607 -77.510 46.911
184868 98.621 -77.526 46.888
184869 98.633 -77.536 46.865
184870 98.645 -77.532 46.841
184871 98.656 -77.527 46.818
184872 98.668 -77.523 46.794
184873 98.679 -77.520 46.769
184874 98.697 -77.516 46.746
184875 98.717 -77.511 46.723
184876 98.736 -77.507 46.700
184877 98.755 -77.503 46.675
184878 98.775 -77.498 46.651
184879 98.794 -77.495 46.626
184880 98.813 -77.490 46.601
184881 98.831 -77.491 46.578
184882 98.845 -77.498 46.556
184883 98.860 -77.504 46.534
184884 98.874 -77.511 46.513
184885 98.887 -77.517 46.491
184886 98.880 -77.529 46.471
184887 98.874 -77.539 46.450
184888 98.873 -77.545 46.429
184889 98.896 -77.551 46.409
184890 98.960 -77.558 46.388
184891 98.980 -77.568 46.367
184892 98.999 -77.578 46.345
184893 99.019 -77.587 46.323
184894 99.038 -77.595 46.301
184895 99.037 -77.608 46.281
184896 99.035 -77.622 46.260
184897 99.034 -77.635 46.238
184898 99.010 -77.639 46.216
184899 99.006 -77.639 46.195
184900 99.016 -77.639 46.172
184901 99.026 -77.640 46.149
184902 99.036 -77.640 46.127
184903 99.046 -77.640 46.105
184904 99.056 -77.640 46.083
184905 99.066 -77.640 46.061
184906 99.076 -77.640 46.041
184907 99.086 -77.640 46.020
184908 99.095 -77.643 46.001
184909 99.104 -77.647 45.983
184910 99.113 -77.652 45.965
184911 99.125 -77.660 45.948
184912 99.140 -77.672 45.932
184913 99.164 -77.689 45.916
184914 99.202 -77.708 45.902
184915 99.241 -77.728 45.889
184916 99.279 -77.748 45.878
184917 99.318 -77.768 45.867
184918 99.356 -77.787 45.857
184919 99.395 -77.806 45.847
184920 99.433 -77.826 45.837
184921 99.472 -77.846 45.828
184922 99.510 -77.865 45.817
184923 99.549 -77.884 45.806
184924 99.587 -77.904 45.794
184925 99.626 -77.924 45.783
184926 99.664 -77.943 45.772
184927 99.703 -77.962 45.760
184928 99.726 -77.971 45.749
184929 99.743 -77.975 45.738
184930 99.759 -77.979 45.729
184931 99.775 -77.984 45.719
184932 99.792 -77.987 45.709
184933 99.808 -77.991 45.700
184934 99.824 -77.996 45.690
184935 99.841 -78.000 45.679
184936 99.860 -78.009 45.667
184937 99.881 -78.021 45.658
184938 99.902 -78.033 45.648
184939 99.923 -78.044 45.636
184940 99.943 -78.056 45.623
184941 99.970 -78.064 45.610
184942 100.000 -78.069 45.597
184943 100.030 -78.075 45.583
184944 100.060 -78.080 45.568
184945 100.075 -78.103 45.554
184946 100.108 -78.111 45.540
184947 100.141 -78.119 45.526
184948 100.174 -78.128 45.511
184949 100.207 -78.137 45.496
184950 100.240 -78.145 45.481
184951 100.301 -78.094 45.467
184952 100.317 -78.097 45.453
184953 100.322 -78.113 45.439
184954 100.329 -78.130 45.426
184955 100.336 -78.147 45.413
184956 100.342 -78.164 45.398
184957 100.349 -78.181 45.382
184958 100.355 -78.198 45.367
184959 100.362 -78.215 45.352
184960 100.368 -78.232 45.337
184961 100.375 -78.249 45.322
184962 100.385 -78.263 45.306
184963 100.409 -78.271 45.290
184964 100.440 -78.275 45.274
184965 100.482 -78.275 45.258
184966 100.524 -78.275 45.242
184967 100.566 -78.275 45.226
184968 100.608 -78.275 45.211
184969 100.654 -78.282 45.196
184970 100.701 -78.290 45.180
184971 100.748 -78.297 45.165
184972 100.794 -78.305 45.151
184973 100.841 -78.313 45.136
184974 100.888 -78.320 45.122
184975 100.935 -78.329 45.107
184976 100.979 -78.335 45.093
184977 101.016 -78.341 45.078
184978 101.055 -78.348 45.062
184979 101.085 -78.355 45.046
184980 101.112 -78.361 45.030
184981 101.139 -78.368 45.013
184982 101.165 -78.376 44.997
184983 101.181 -78.386 44.981
184984 101.196 -78.397 44.965
184985 101.212 -78.407 44.950
184986 101.228 -78.418 44.934
184987 101.244 -78.429 44.918
184988 101.260 -78.440 44.902
184989 101.276 -78.451 44.885
184990 101.291 -78.459 44.869
184991 101.307 -78.467 44.854
184992 101.324 -78.477 44.838
184993 101.354 -78.494 44.822
184994 101.384 -78.517 44.806
184995 101.414 -78.539 44.789
184996 101.444 -78.563 44.773
184997 101.473 -78.586 44.758
184998 101.499 -78.611 44.741
184999 101.523 -78.637 44.725
185000 101.545 -78.665 44.710
185001 101.568 -78.698 44.694
185002 101.589 -78.725 44.678
185003 101.600 -78.728 44.661
185004 101.631 -78.747 44.647
185005 101.670 -78.773 44.632
185006 101.709 -78.798 44.618
185007 101.748 -78.823 44.604
185008 101.786 -78.849 44.589
185009 101.825 -78.874 44.574
185010 101.864 -78.899 44.560
185011 101.859 -78.967 44.545
185012 101.878 -78.973 44.531
185013 101.896 -78.979 44.517
185014 101.914 -78.986 44.503
185015 101.933 -78.992 44.490
185016 101.951 -78.998 44.476
185017 101.969 -79.004 44.463
185018 101.988 -79.010 44.451
185019 102.006 -79.016 44.438
185020 102.024 -79.023 44.426
185021 102.042 -79.029 44.414
185291 93.269 -76.605 54.138
185292 93.288 -76.602 54.141
185293 93.308 -76.599 54.142
185294 93.328 -76.596 54.144
185295 93.347 -76.593 54.145
185296 93.366 -76.590 54.145
185297 93.387 -76.587 54.146
185298 93.418 -76.583 54.146
185299 93.443 -76.582 54.147
185300 93.462 -76.585 54.147
185301 93.480 -76.588 54.148
185302 93.499 -76.590 54.148
185303 93.517 -76.593 54.147
185304 93.535 -76.596 54.148
185305 93.554 -76.599 54.148
185306 93.572 -76.602 54.149
185307 93.590 -76.605 54.151
185308 93.609 -76.608 54.154
185309 93.629 -76.600 54.156
185310 93.650 -76.591 54.159
185311 93.672 -76.581 54.163
185312 93.693 -76.571 54.168
185313 93.714 -76.561 54.175
185314 93.735 -76.552 54.184
185315 93.737 -76.539 54.194
185316 93.719 -76.507 54.199
185317 93.730 -76.501 54.205
185318 93.757 -76.514 54.211
185319 93.784 -76.528 54.217
185320 93.811 -76.542 54.222
185321 93.827 -76.542 54.229
185322 93.842 -76.543 54.233
185323 93.857 -76.543 54.233
185324 93.873 -76.539 54.232
185325 93.890 -76.533 54.228
185326 93.906 -76.529 54.222
185327 93.922 -76.524 54.213
185328 93.929 -76.520 54.207
185329 93.932 -76.517 54.200
185330 93.936 -76.514 54.192
185331 93.948 -76.514 54.184
185332 93.960 -76.517 54.176
185333 93.972 -76.520 54.171
185334 93.984 -76.523 54.167
185335 93.996 -76.527 54.168
185336 94.008 -76.530 54.168
185337 94.031 -76.527 54.167
185338 94.059 -76.522 54.168
185339 94.087 -76.517 54.170
185340 94.116 -76.511 54.171
185341 94.136 -76.507 54.170
185342 94.150 -76.503 54.167
185343 94.165 -76.498 54.159
185344 94.181 -76.492 54.148
185345 94.198 -76.485 54.135
185346 94.214 -76.479 54.121
185347 94.230 -76.472 54.104
185348 94.248 -76.468 54.085
185349 94.268 -76.467 54.066
185350 94.287 -76.465 54.039
185351 94.307 -76.463 54.010
185352 94.327 -76.461 53.979
185353 94.347 -76.460 53.950
185354 94.366 -76.458 53.917
185355 94.385 -76.451 53.883
185356 94.402 -76.438 53.849
185357 94.419 -76.425 53.813
185358 94.436 -76.411 53.776
185359 94.454 -76.400 53.738
185360 94.480 -76.401 53.700
185361 94.506 -76.403 53.659
185362 94.533 -76.404 53.618
185363 94.559 -76.407 53.577
185364 94.585 -76.408 53.534
185365 94.611 -76.410 53.497
185366 94.638 -76.412 53.462
185367 94.662 -76.412 53.426
185368 94.678 -76.412 53.390
185369 94.694 -76.411 53.352
185370 94.711 -76.408 53.315
185371 94.727 -76.406 53.279
185372 94.744 -76.404 53.246
185373 94.752 -76.404 53.215
185374 94.748 -76.406 53.185
185375 94.744 -76.408 53.156
185376 94.740 -76.411 53.130
185377 94.736 -76.413 53.104
185378 94.732 -76.415 53.079
185379 94.728 -76.417 53.056
185380 94.724 -76.420 53.036
185381 94.720 -76.422 53.015
185382 94.716 -76.424 52.991
185383 94.712 -76.426 52.964
185384 94.708 -76.428 52.928
185385 94.704 -76.430 52.885
185386 94.700 -76.433 52.837
185387 94.700 -76.436 52.785
185388 94.709 -76.439 52.731
185389 94.718 -76.443 52.678
185390 94.728 -76.447 52.626
185391 94.737 -76.451 52.574
185392 94.746 -76.455 52.523
185393 94.791 -76.467 52.472
185394 94.842 -76.480 52.421
185395 94.893 -76.492 52.368
185396 94.944 -76.505 52.315
185397 94.992 -76.517 52.261
185398 95.031 -76.525 52.209
185399 95.070 -76.533 52.159
185400 95.110 -76.542 52.109
185401 95.148 -76.549 52.060
185402 95.188 -76.558 52.011
185403 95.215 -76.569 51.965
185404 95.230 -76.583 51.918
185405 95.245 -76.597 51.869
185406 95.259 -76.611 51.821
185407 95.280 -76.624 51.772
185408 95.322 -76.634 51.724
185409 95.364 -76.644 51.679
185410 95.405 -76.654 51.633
185411 95.447 -76.664 51.589
185412 95.489 -76.674 51.546
185413 95.531 -76.684 51.504
185414 95.572 -76.694 51.463
185415 95.592 -76.702 51.425
185416 95.599 -76.720 51.389
185417 95.606 -76.738 51.355
185418 95.613 -76.756 51.321
185419 95.618 -76.774 51.288
185420 95.614 -76.790 51.254
185421 95.610 -76.806 51.218
185422 95.607 -76.822 51.179
185423 95.603 -76.838 51.139
185424 95.599 -76.853 51.098
185425 95.607 -76.865 51.057
185426 95.635 -76.867 51.013
185427 95.663 -76.870 50.967
185428 95.690 -76.872 50.920
185429 95.718 -76.875 50.876
185430 95.748 -76.873 50.833
185431 95.778 -76.871 50.789
185432 95.809 -76.868 50.745
185433 95.840 -76.864 50.697
185434 95.867 -76.864 50.646
185435 95.892 -76.865 50.593
185436 95.918 -76.866 50.534
185437 95.943 -76.868 50.474
185438 95.969 -76.868 50.415
185439 95.994 -76.870 50.358
185440 96.017 -76.874 50.298
185441 96.031 -76.889 50.236
185442 96.044 -76.904 50.171
185443 96.058 -76.919 50.105
185444 96.077 -76.931 50.041
185445 96.102 -76.940 49.982
185446 96.127 -76.949 49.927
185447 96.152 -76.958 49.876
185448 96.177 -76.966 49.829
185449 96.182 -76.981 49.785
185450 96.172 -77.000 49.745
185451 96.162 -77.018 49.707
185452 96.152 -77.037 49.672
185453 96.147 -77.053 49.640
185454 96.157 -77.067 49.610
185455 96.166 -77.081 49.583
185456 96.176 -77.094 49.557
185457 96.245 -77.052 49.533
185458 96.259 -77.066 49.511
185459 96.273 -77.079 49.491
185460 96.288 -77.093 49.473
185461 96.302 -77.106 49.456
185462 96.330 -77.108 49.441
185463 96.361 -77.107 49.426
185464 96.391 -77.106 49.410
185465 96.421 -77.106 49.394
185466 96.452 -77.106 49.382
185467 96.482 -77.106 49.368
185468 96.512 -77.105 49.354
185469 96.542 -77.089 49.340
185470 96.559 -77.113 49.321
185471 96.575 -77.135 49.300
185472 96.595 -77.141 49.279
185473 96.615 -77.147 49.258
185474 96.636 -77.154 49.238
185475 96.656 -77.161 49.217
185476 96.676 -77.168 49.196
185477 96.695 -77.175 49.173
185478 96.716 -77.182 49.147
185479 96.736 -77.188 49.122
185480 96.747 -77.191 49.098
185481 96.761 -77.195 49.076
185482 96.775 -77.200 49.053
185483 96.789 -77.204 49.031
185484 96.804 -77.209 49.007
185485 96.818 -77.213 48.984
185486 96.832 -77.218 48.960
185487 96.847 -77.222 48.937
185488 96.861 -77.227 48.912
185489 96.875 -77.232 48.886
185490 96.889 -77.237 48.861
185491 96.904 -77.248 48.837
185492 96.918 -77.258 48.812
185493 96.933 -77.269 48.790
185494 96.947 -77.281 48.768
185495 96.984 -77.279 48.745
185496 97.061 -77.260 48.721
185497 97.093 -77.270 48.697
185498 97.116 -77.288 48.675
185499 97.141 -77.307 48.653
185500 97.165 -77.327 48.630
185501 97.189 -77.346 48.608
185502 97.214 -77.366 48.587
185503 97.238 -77.385 48.566
185504 97.263 -77.404 48.546
185505 97.288 -77.423 48.526
185506 97.312 -77.443 48.506
185507 97.336 -77.462 48.486
185508 97.361 -77.482 48.467
185509 97.377 -77.486 48.447
185510 97.382 -77.467 48.428
185511 97.388 -77.450 48.409
185512 97.393 -77.432 48.391
185513 97.398 -77.414 48.373
185514 97.409 -77.424 48.353
185515 97.423 -77.442 48.333
185516 97.436 -77.460 48.313
185517 97.449 -77.478 48.293
185518 97.462 -77.496 48.271
185519 97.475 -77.514 48.249
185520 97.489 -77.532 48.227
185521 97.502 -77.549 48.203
185522 97.493 -77.517 48.177
185523 97.523 -77.523 48.149
185524 97.555 -77.529 48.120
185525 97.585 -77.535 48.090
185526 97.616 -77.541 48.059
185527 97.646 -77.544 48.028
185528 97.673 -77.542 47.995
185529 97.714 -77.546 47.964
185530 97.750 -77.548 47.933
185531 97.787 -77.548 47.902
185532 97.824 -77.546 47.871
185533 97.862 -77.542 47.839
185534 97.898 -77.539 47.806
185535 97.936 -77.536 47.773
185536 97.973 -77.532 47.740
185537 98.010 -77.528 47.708
185538 98.047 -77.525 47.676
185539 98.084 -77.521 47.646
185540 98.101 -77.517 47.618
185541 98.111 -77.513 47.590
185542 98.122 -77.508 47.562
185543 98.132 -77.500 47.535
185544 98.142 -77.485 47.506
185545 98.151 -77.470 47.478
185546 98.160 -77.454 47.449
185547 98.169 -77.439 47.421
185548 98.179 -77.424 47.394
185549 98.189 -77.409 47.367
185550 98.198 -77.394 47.340
185551 98.207 -77.379 47.313
185552 98.236 -77.377 47.286
185553 98.286 -77.389 47.260
185554 98.336 -77.401 47.235
185555 98.385 -77.414 47.212
185556 98.435 -77.426 47.190
185557 98.484 -77.439 47.167
185558 98.515 -77.444 47.145
185559 98.535 -77.445 47.123
185560 98.556 -77.447 47.100
185561 98.576 -77.448 47.079
185562 98.581 -77.435 47.057
185563 98.568 -77.413 47.034
185564 98.574 -77.432 47.011
185565 98.585 -77.450 46.988
185566 98.599 -77.466 46.964
185567 98.612 -77.483 46.940
185568 98.626 -77.498 46.916
185569 98.639 -77.515 46.893
185570 98.651 -77.517 46.870
185571 98.662 -77.514 46.846
185572 98.674 -77.510 46.822
185573 98.686 -77.506 46.798
185574 98.697 -77.501 46.773
185575 98.708 -77.498 46.750
185576 98.725 -77.494 46.727
185577 98.745 -77.489 46.704
185578 98.764 -77.486 46.680
185579 98.784 -77.481 46.656
185580 98.803 -77.476 46.632
185581 98.822 -77.473 46.607
185582 98.839 -77.474 46.584
185583 98.854 -77.481 46.562
185584 98.868 -77.487 46.540
185585 98.883 -77.494 46.519
185586 98.878 -77.504 46.498
185587 98.871 -77.516 46.477
185588 98.864 -77.527 46.456
185589 98.861 -77.536 46.434
185590 98.960 -77.539 46.413
185591 98.980 -77.549 46.391
185592 98.999 -77.558 46.369
185593 99.019 -77.566 46.348
185594 99.039 -77.574 46.326
185595 99.043 -77.586 46.304
185596 99.042 -77.600 46.283
185597 99.040 -77.613 46.262
185598 99.039 -77.627 46.241
185599 99.032 -77.638 46.219
185600 99.011 -77.638 46.197
185601 99.021 -77.638 46.174
185602 99.031 -77.638 46.152
185603 99.041 -77.638 46.130
185604 99.051 -77.638 46.107
185605 99.061 -77.639 46.086
185606 99.071 -77.639 46.065
185607 99.080 -77.639 46.043
185608 99.090 -77.641 46.023
185609 99.099 -77.645 46.005
185610 99.108 -77.649 45.986
185611 99.117 -77.653 45.969
185612 99.126 -77.658 45.951
185613 99.148 -77.658 45.935
185614 99.182 -77.670 45.919
185615 99.215 -77.686 45.905
185616 99.249 -77.703 45.892
185617 99.283 -77.720 45.880
185618 99.317 -77.736 45.869
185619 99.351 -77.752 45.858
185620 99.385 -77.769 45.848
185621 99.421 -77.787 45.838
185622 99.458 -77.805 45.828
185623 99.496 -77.823 45.818
185624 99.533 -77.841 45.806
185625 99.570 -77.859 45.794
185626 99.607 -77.877 45.782
185627 99.644 -77.896 45.770
185628 99.682 -77.914 45.758
185629 99.726 -77.924 45.746
185630 99.742 -77.928 45.736
185631 99.758 -77.932 45.726
185632 99.775 -77.937 45.716
185633 99.796 -77.949 45.706
185634 99.817 -77.960 45.697
185635 99.837 -77.972 45.687
185636 99.858 -77.984 45.676
185637 99.879 -77.996 45.666
185638 99.900 -78.007 45.655
185639 99.921 -78.019 45.645
185640 99.941 -78.031 45.633
185641 99.963 -78.042 45.621
185642 99.993 -78.047 45.608
185643 100.023 -78.053 45.595
185644 100.053 -78.058 45.581
185645 100.083 -78.062 45.567
185646 100.114 -78.068 45.554
185647 100.144 -78.073 45.540
185648 100.174 -78.078 45.525
185649 100.178 -78.115 45.511
185650 100.211 -78.123 45.497
185651 100.244 -78.131 45.482
185652 100.299 -78.094 45.468
185653 100.331 -78.075 45.454
185654 100.338 -78.092 45.440
185655 100.344 -78.109 45.426
185656 100.351 -78.126 45.413
185657 100.357 -78.143 45.398
185658 100.364 -78.159 45.383
185659 100.370 -78.177 45.368
185660 100.374 -78.199 45.353
185661 100.378 -78.220 45.338
185662 100.387 -78.235 45.323
185663 100.396 -78.248 45.308
185664 100.420 -78.253 45.292
185665 100.450 -78.258 45.276
185666 100.491 -78.259 45.260
185667 100.533 -78.259 45.244
185668 100.575 -78.259 45.229
185669 100.618 -78.261 45.214
185670 100.665 -78.269 45.199
185671 100.712 -78.276 45.184
185672 100.759 -78.284 45.168
185673 100.806 -78.292 45.153
185674 100.853 -78.300 45.139
185675 100.900 -78.307 45.124
185676 100.939 -78.314 45.109
185677 100.977 -78.320 45.094
185678 101.015 -78.326 45.080
185679 101.053 -78.333 45.064
185680 101.091 -78.339 45.048
185681 101.119 -78.346 45.032
185682 101.145 -78.353 45.016
185683 101.172 -78.360 45.000
185684 101.187 -78.370 44.984
185685 101.203 -78.381 44.969
185686 101.219 -78.392 44.953
185687 101.235 -78.403 44.936
185688 101.251 -78.413 44.921
185689 101.267 -78.424 44.906
185690 101.282 -78.435 44.889
185691 101.298 -78.446 44.873
185692 101.314 -78.457 44.857
185693 101.328 -78.471 44.841
185694 101.351 -78.491 44.825
185695 101.379 -78.513 44.809
185696 101.409 -78.535 44.793
185697 101.439 -78.558 44.776
185698 101.469 -78.581 44.761
185699 101.498 -78.604 44.744
185700 101.523 -78.630 44.728
185701 101.546 -78.658 44.713
185702 101.568 -78.691 44.697
185703 101.582 -78.700 44.681
185704 101.615 -78.720 44.665
185705 101.653 -78.745 44.650
185706 101.692 -78.771 44.635
185707 101.731 -78.796 44.620
185708 101.770 -78.821 44.604
185709 101.809 -78.847 44.590
185710 101.847 -78.872 44.576
185711 101.886 -78.898 44.560
185712 101.878 -78.975 44.546
185713 101.897 -78.981 44.532
185714 101.915 -78.987 44.518
185715 101.933 -78.993 44.504
185716 101.952 -78.999 44.491
185717 101.970 -79.006 44.477
185718 101.988 -79.012 44.465
185719 102.007 -79.018 44.452
185720 102.025 -79.024 44.439
185721 102.043 -79.031 44.427
185722 102.061 -79.037 44.415
185723 102.080 -79.043 44.402
185994 93.297 -76.576 54.133
185995 93.317 -76.573 54.135
185996 93.337 -76.570 54.137
185997 93.356 -76.567 54.137
185998 93.380 -76.563 54.137
185999 93.411 -76.559 54.137
186000 93.435 -76.558 54.138
186001 93.454 -76.561 54.138
186002 93.472 -76.564 54.139
186003 93.490 -76.567 54.139
186004 93.509 -76.570 54.138
186005 93.527 -76.573 54.138
186006 93.546 -76.576 54.139
186007 93.564 -76.579 54.140
186008 93.582 -76.582 54.142
186009 93.601 -76.585 54.145
186010 93.619 -76.587 54.146
186011 93.639 -76.581 54.148
186012 93.660 -76.572 54.151
186013 93.682 -76.562 54.156
186014 93.703 -76.552 54.165
186015 93.724 -76.542 54.175
186016 93.744 -76.536 54.185
186017 93.738 -76.517 54.193
186018 93.721 -76.490 54.201
186019 93.748 -76.504 54.208
186020 93.775 -76.517 54.215
186021 93.802 -76.531 54.222
186022 93.820 -76.534 54.229
186023 93.835 -76.532 54.235
186024 93.851 -76.527 54.237
186025 93.867 -76.521 54.237
186026 93.884 -76.517 54.236
186027 93.900 -76.511 54.232
186028 93.915 -76.507 54.225
186029 93.919 -76.504 54.218
186030 93.923 -76.501 54.211
186031 93.931 -76.497 54.202
186032 93.937 -76.499 54.193
186033 93.938 -76.505 54.185
186034 93.949 -76.509 54.178
186035 93.961 -76.512 54.172
186036 93.973 -76.515 54.171
186037 93.995 -76.514 54.168
186038 94.023 -76.508 54.166
186039 94.051 -76.503 54.166
186040 94.079 -76.498 54.166
186041 94.108 -76.493 54.167
186042 94.135 -76.488 54.167
186043 94.149 -76.483 54.168
186044 94.164 -76.480 54.161
186045 94.180 -76.473 54.152
186046 94.196 -76.467 54.140
186047 94.213 -76.460 54.127
186048 94.229 -76.454 54.110
186049 94.249 -76.452 54.091
186050 94.269 -76.451 54.071
186051 94.288 -76.449 54.046
186052 94.308 -76.448 54.018
186053 94.328 -76.446 53.989
186054 94.347 -76.445 53.960
186055 94.367 -76.443 53.927
186056 94.387 -76.441 53.892
186057 94.405 -76.430 53.856
186058 94.421 -76.417 53.819
186059 94.438 -76.403 53.780
186060 94.459 -76.395 53.742
186061 94.485 -76.397 53.703
186062 94.511 -76.399 53.665
186063 94.538 -76.401 53.626
186064 94.564 -76.402 53.586
186065 94.590 -76.404 53.545
186066 94.615 -76.404 53.509
186067 94.635 -76.402 53.474
186068 94.655 -76.401 53.439
186069 94.668 -76.408 53.405
186070 94.684 -76.407 53.368
186071 94.700 -76.404 53.330
186072 94.717 -76.402 53.292
186073 94.734 -76.400 53.257
186074 94.743 -76.399 53.225
186075 94.739 -76.401 53.195
186076 94.735 -76.404 53.167
186077 94.731 -76.406 53.141
186078 94.728 -76.408 53.114
186079 94.723 -76.411 53.089
186080 94.719 -76.413 53.066
186081 94.715 -76.415 53.046
186082 94.711 -76.417 53.025
186083 94.707 -76.420 53.002
186084 94.703 -76.422 52.976
186085 94.700 -76.424 52.940
186086 94.696 -76.426 52.898
186087 94.701 -76.430 52.851
186088 94.710 -76.433 52.800
186089 94.719 -76.437 52.747
186090 94.728 -76.442 52.695
186091 94.737 -76.445 52.643
186092 94.746 -76.449 52.590
186093 94.755 -76.453 52.539
186094 94.778 -76.460 52.487
186095 94.829 -76.473 52.436
186096 94.880 -76.486 52.384
186097 94.931 -76.499 52.331
186098 94.976 -76.509 52.278
186099 95.016 -76.517 52.226
186100 95.055 -76.526 52.176
186101 95.094 -76.533 52.127
186102 95.133 -76.542 52.077
186103 95.172 -76.550 52.029
186104 95.210 -76.558 51.983
186105 95.225 -76.573 51.936
186106 95.239 -76.586 51.888
186107 95.254 -76.601 51.841
186108 95.283 -76.613 51.793
186109 95.325 -76.623 51.745
186110 95.367 -76.633 51.701
186111 95.409 -76.643 51.655
186112 95.451 -76.652 51.609
186113 95.492 -76.662 51.565
186114 95.534 -76.672 51.521
186115 95.576 -76.682 51.480
186116 95.615 -76.693 51.441
186117 95.611 -76.708 51.404
186118 95.608 -76.724 51.369
186119 95.604 -76.740 51.334
186120 95.600 -76.756 51.299
186121 95.597 -76.772 51.264
186122 95.593 -76.788 51.226
186123 95.590 -76.804 51.186
186124 95.586 -76.820 51.145
186125 95.584 -76.835 51.104
186126 95.611 -76.837 51.063
186127 95.639 -76.840 51.021
186128 95.667 -76.843 50.976
186129 95.694 -76.846 50.931
186130 95.722 -76.848 50.889
186131 95.749 -76.850 50.848
186132 95.778 -76.850 50.806
186133 95.809 -76.847 50.762
186134 95.835 -76.848 50.714
186135 95.861 -76.850 50.662
186136 95.886 -76.850 50.607
186137 95.911 -76.852 50.549
186138 95.937 -76.853 50.491
186139 95.962 -76.854 50.436
186140 95.987 -76.856 50.384
186141 96.012 -76.858 50.325
186142 96.025 -76.873 50.263
186143 96.039 -76.888 50.197
186144 96.058 -76.900 50.132
186145 96.084 -76.909 50.068
186146 96.109 -76.918 50.008
186147 96.134 -76.926 49.954
186148 96.159 -76.935 49.903
186149 96.184 -76.943 49.857
186150 96.179 -76.961 49.814
186151 96.169 -76.979 49.772
186152 96.159 -76.998 49.733
186153 96.152 -77.016 49.698
186154 96.162 -77.029 49.666
186155 96.171 -77.043 49.637
186156 96.181 -77.056 49.610
186157 96.190 -77.069 49.584
186158 96.242 -77.044 49.560
186159 96.291 -77.024 49.538
186160 96.305 -77.038 49.519
186161 96.319 -77.051 49.503
186162 96.333 -77.065 49.487
186163 96.347 -77.078 49.471
186164 96.363 -77.090 49.455
186165 96.393 -77.090 49.440
186166 96.424 -77.089 49.426
186167 96.454 -77.089 49.412
186168 96.484 -77.088 49.398
186169 96.515 -77.081 49.383
186170 96.543 -77.068 49.368
186171 96.559 -77.094 49.347
186172 96.575 -77.119 49.325
186173 96.594 -77.131 49.304
186174 96.614 -77.138 49.282
186175 96.634 -77.144 49.261
186176 96.654 -77.150 49.240
186177 96.674 -77.157 49.218
186178 96.694 -77.164 49.194
186179 96.714 -77.169 49.168
186180 96.734 -77.174 49.142
186181 96.750 -77.179 49.118
186182 96.760 -77.182 49.094
186183 96.770 -77.185 49.071
186184 96.780 -77.188 49.048
186185 96.790 -77.192 49.024
186186 96.800 -77.195 49.000
186187 96.814 -77.200 48.976
186188 96.829 -77.204 48.953
186189 96.843 -77.209 48.928
186190 96.857 -77.213 48.902
186191 96.872 -77.218 48.876
186192 96.886 -77.223 48.851
186193 96.900 -77.235 48.826
186194 96.915 -77.245 48.802
186195 96.950 -77.245 48.779
186196 97.013 -77.241 48.756
186197 97.064 -77.250 48.732
186198 97.110 -77.266 48.709
186199 97.127 -77.274 48.685
186200 97.144 -77.282 48.662
186201 97.162 -77.294 48.640
186202 97.186 -77.313 48.618
186203 97.211 -77.332 48.597
186204 97.235 -77.351 48.576
186205 97.259 -77.371 48.556
186206 97.284 -77.390 48.536
186207 97.308 -77.410 48.515
186208 97.333 -77.429 48.495
186209 97.357 -77.448 48.476
186210 97.382 -77.468 48.456
186211 97.388 -77.453 48.436
186212 97.393 -77.435 48.417
186213 97.399 -77.417 48.398
186214 97.405 -77.405 48.379
186215 97.419 -77.423 48.359
186216 97.432 -77.442 48.339
186217 97.445 -77.459 48.319
186218 97.458 -77.477 48.299
186219 97.471 -77.495 48.276
186220 97.485 -77.513 48.255
186221 97.498 -77.531 48.232
186222 97.502 -77.511 48.208
186223 97.519 -77.512 48.180
186224 97.546 -77.514 48.152
186225 97.574 -77.517 48.123
186226 97.602 -77.518 48.093
186227 97.629 -77.520 48.062
186228 97.657 -77.523 48.031
186229 97.684 -77.525 47.999
186230 97.715 -77.514 47.968
186231 97.751 -77.511 47.937
186232 97.788 -77.507 47.906
186233 97.825 -77.504 47.874
186234 97.863 -77.500 47.841
186235 97.900 -77.496 47.809
186236 97.937 -77.493 47.776
186237 97.974 -77.489 47.743
186238 98.011 -77.486 47.711
186239 98.048 -77.483 47.679
186240 98.085 -77.479 47.649
186241 98.104 -77.474 47.622
186242 98.114 -77.470 47.594
186243 98.126 -77.466 47.567
186244 98.136 -77.462 47.539
186245 98.147 -77.457 47.511
186246 98.157 -77.453 47.482
186247 98.167 -77.440 47.454
186248 98.176 -77.425 47.427
186249 98.185 -77.410 47.400
186250 98.195 -77.395 47.372
186251 98.205 -77.379 47.345
186252 98.214 -77.364 47.319
186253 98.227 -77.352 47.291
186254 98.277 -77.364 47.265
186255 98.326 -77.376 47.239
186256 98.376 -77.389 47.214
186257 98.426 -77.401 47.192
186258 98.475 -77.414 47.169
186259 98.524 -77.426 47.147
186260 98.559 -77.432 47.125
186261 98.579 -77.434 47.103
186262 98.589 -77.426 47.081
186263 98.569 -77.387 47.061
186264 98.577 -77.406 47.038
186265 98.590 -77.423 47.015
186266 98.603 -77.439 46.992
186267 98.617 -77.455 46.970
186268 98.630 -77.472 46.946
186269 98.643 -77.488 46.922
186270 98.657 -77.504 46.899
186271 98.668 -77.500 46.875
186272 98.679 -77.495 46.852
186273 98.691 -77.492 46.828
186274 98.703 -77.488 46.803
186275 98.714 -77.484 46.778
186276 98.726 -77.479 46.755
186277 98.737 -77.476 46.732
186278 98.754 -77.472 46.709
186279 98.773 -77.467 46.686
186280 98.792 -77.464 46.662
186281 98.812 -77.459 46.639
186282 98.831 -77.455 46.616
186283 98.847 -77.457 46.593
186284 98.862 -77.464 46.571
186285 98.877 -77.470 46.548
186286 98.876 -77.481 46.525
186287 98.869 -77.492 46.503
186288 98.862 -77.503 46.482
186289 98.855 -77.514 46.461
186290 98.901 -77.523 46.440
186291 98.980 -77.529 46.418
186292 99.000 -77.537 46.396
186293 99.020 -77.545 46.373
186294 99.040 -77.554 46.351
186295 99.049 -77.564 46.329
186296 99.048 -77.577 46.308
186297 99.047 -77.591 46.286
186298 99.045 -77.604 46.265
186299 99.044 -77.617 46.243
186300 99.042 -77.631 46.221
186301 99.021 -77.636 46.198
186302 99.026 -77.636 46.176
186303 99.035 -77.636 46.154
186304 99.046 -77.636 46.132
186305 99.056 -77.636 46.110
186306 99.066 -77.636 46.089
186307 99.075 -77.636 46.067
186308 99.085 -77.639 46.046
186309 99.094 -77.642 46.026
186310 99.103 -77.647 46.008
186311 99.112 -77.651 45.990
186312 99.127 -77.654 45.972
186313 99.157 -77.653 45.954
186314 99.187 -77.652 45.938
186315 99.217 -77.653 45.922
186316 99.252 -77.670 45.907
186317 99.285 -77.686 45.894
186318 99.319 -77.702 45.882
186319 99.353 -77.719 45.870
186320 99.387 -77.736 45.860
186321 99.421 -77.752 45.849
186322 99.455 -77.768 45.838
186323 99.489 -77.785 45.829
186324 99.523 -77.801 45.818
186325 99.557 -77.818 45.807
186326 99.591 -77.834 45.794
186327 99.625 -77.851 45.783
186328 99.659 -77.867 45.770
186329 99.693 -77.884 45.757
186330 99.731 -77.888 45.745
186331 99.752 -77.899 45.735
186332 99.773 -77.912 45.725
186333 99.793 -77.923 45.714
186334 99.815 -77.935 45.704
186335 99.835 -77.946 45.694
186336 99.856 -77.959 45.684
186337 99.877 -77.970 45.674
186338 99.898 -77.982 45.664
186339 99.918 -77.993 45.653
186340 99.939 -78.006 45.642
186341 99.960 -78.017 45.630
186342 99.986 -78.025 45.618
186343 100.016 -78.031 45.607
186344 100.046 -78.036 45.594
186345 100.076 -78.040 45.581
186346 100.107 -78.046 45.567
186347 100.137 -78.051 45.554
186348 100.168 -78.056 45.540
186349 100.198 -78.062 45.525
186350 100.228 -78.066 45.511
186351 100.258 -78.072 45.497
186352 100.288 -78.077 45.483
186353 100.299 -78.093 45.469
186354 100.323 -78.087 45.455
186355 100.326 -78.109 45.441
186356 100.334 -78.125 45.427
186357 100.344 -78.139 45.413
186358 100.353 -78.153 45.398
186359 100.362 -78.167 45.384
186360 100.372 -78.181 45.369
186361 100.382 -78.195 45.354
186362 100.391 -78.210 45.339
186363 100.400 -78.223 45.324
186364 100.410 -78.238 45.310
186365 100.432 -78.235 45.294
186366 100.459 -78.241 45.278
186367 100.500 -78.242 45.261
186368 100.542 -78.242 45.247
186369 100.584 -78.242 45.232
186370 100.630 -78.247 45.217
186371 100.677 -78.255 45.201
186372 100.723 -78.263 45.187
186373 100.770 -78.271 45.171
186374 100.817 -78.279 45.156
186375 100.861 -78.286 45.142
186376 100.899 -78.292 45.126
186377 100.937 -78.298 45.112
186378 100.975 -78.305 45.097
186379 101.013 -78.311 45.082
186380 101.051 -78.317 45.067
186381 101.090 -78.324 45.051
186382 101.125 -78.330 45.035
186383 101.152 -78.337 45.019
186384 101.178 -78.344 45.003
186385 101.194 -78.354 44.988
186386 101.210 -78.366 44.972
186387 101.226 -78.376 44.956
186388 101.242 -78.387 44.940
186389 101.257 -78.398 44.924
186390 101.273 -78.409 44.909
186391 101.289 -78.420 44.893
186392 101.302 -78.437 44.877
186393 101.315 -78.454 44.861
186394 101.331 -78.472 44.844
186395 101.357 -78.492 44.828
186396 101.383 -78.514 44.812
186397 101.409 -78.534 44.796
186398 101.435 -78.554 44.780
186399 101.464 -78.577 44.763
186400 101.494 -78.600 44.747
186401 101.523 -78.623 44.731
186402 101.546 -78.651 44.716
186403 101.564 -78.671 44.700
186404 101.598 -78.693 44.684
186405 101.637 -78.718 44.669
186406 101.676 -78.744 44.653
186407 101.714 -78.769 44.638
186408 101.753 -78.795 44.622
186409 101.792 -78.820 44.607
186410 101.830 -78.845 44.592
186411 101.869 -78.871 44.577
186412 101.908 -78.896 44.562
186413 101.947 -78.921 44.547
186414 101.915 -78.990 44.533
186415 101.934 -78.996 44.519
186416 101.952 -79.002 44.505
186417 101.970 -79.008 44.492
186418 101.988 -79.014 44.479
186419 102.007 -79.021 44.466
186420 102.025 -79.027 44.453
186421 102.043 -79.033 44.440
186422 102.061 -79.039 44.428
186423 102.080 -79.045 44.416
186424 102.098 -79.052 44.403
186696 93.306 -76.549 54.126
186697 93.326 -76.546 54.128
186698 93.346 -76.543 54.128
186699 93.373 -76.539 54.129
186700 93.404 -76.535 54.129
186701 93.427 -76.535 54.130
186702 93.446 -76.538 54.130
186703 93.464 -76.541 54.131
186704 93.482 -76.544 54.131
186705 93.501 -76.546 54.131
186706 93.519 -76.549 54.130
186707 93.538 -76.552 54.132
186708 93.556 -76.555 54.133
186709 93.574 -76.558 54.135
186710 93.592 -76.561 54.137
186711 93.611 -76.564 54.138
186712 93.629 -76.567 54.140
186713 93.649 -76.562 54.143
186714 93.671 -76.552 54.147
186715 93.692 -76.543 54.154
186716 93.713 -76.533 54.163
186717 93.734 -76.524 54.173
186718 93.751 -76.520 54.183
186719 93.740 -76.496 54.192
186720 93.752 -76.498 54.202
186721 93.775 -76.510 54.212
186722 93.798 -76.523 54.220
186723 93.814 -76.520 54.228
186724 93.830 -76.514 54.234
186725 93.846 -76.510 54.238
186726 93.862 -76.505 54.240
186727 93.878 -76.500 54.241
186728 93.895 -76.495 54.239
186729 93.906 -76.491 54.235
186730 93.909 -76.488 54.228
186731 93.919 -76.483 54.220
186732 93.931 -76.478 54.210
186733 93.938 -76.479 54.202
186734 93.939 -76.486 54.193
186735 93.940 -76.493 54.185
186736 93.941 -76.500 54.178
186737 93.958 -76.501 54.174
186738 93.986 -76.495 54.170
186739 94.014 -76.490 54.167
186740 94.043 -76.485 54.165
186741 94.071 -76.480 54.165
186742 94.099 -76.474 54.166
186743 94.127 -76.469 54.167
186744 94.148 -76.464 54.168
186745 94.163 -76.461 54.164
186746 94.179 -76.455 54.156
186747 94.195 -76.448 54.145
186748 94.211 -76.442 54.132
186749 94.230 -76.439 54.115
186750 94.250 -76.437 54.097
186751 94.270 -76.435 54.079
186752 94.289 -76.433 54.054
186753 94.309 -76.432 54.027
186754 94.329 -76.430 54.000
186755 94.348 -76.429 53.969
186756 94.368 -76.427 53.937
186757 94.388 -76.426 53.903
186758 94.407 -76.422 53.867
186759 94.424 -76.408 53.828
186760 94.441 -76.395 53.788
186761 94.464 -76.392 53.749
186762 94.490 -76.393 53.709
186763 94.517 -76.395 53.671
186764 94.543 -76.397 53.635
186765 94.568 -76.398 53.598
186766 94.589 -76.395 53.560
186767 94.609 -76.393 53.523
186768 94.630 -76.391 53.487
186769 94.647 -76.392 53.452
186770 94.660 -76.398 53.418
186771 94.674 -76.402 53.382
186772 94.690 -76.401 53.345
186773 94.707 -76.398 53.307
186774 94.723 -76.396 53.272
186775 94.735 -76.395 53.239
186776 94.731 -76.398 53.208
186777 94.726 -76.400 53.180
186778 94.722 -76.402 53.153
186779 94.718 -76.404 53.126
186780 94.714 -76.407 53.099
186781 94.710 -76.409 53.075
186782 94.707 -76.411 53.054
186783 94.703 -76.413 53.032
186784 94.699 -76.415 53.009
186785 94.695 -76.417 52.983
186786 94.692 -76.420 52.947
186787 94.701 -76.424 52.905
186788 94.710 -76.428 52.859
186789 94.719 -76.432 52.810
186790 94.729 -76.436 52.759
186791 94.738 -76.439 52.707
186792 94.747 -76.443 52.657
186793 94.756 -76.447 52.604
186794 94.765 -76.451 52.552
186795 94.774 -76.455 52.501
186796 94.816 -76.466 52.450
186797 94.866 -76.479 52.399
186798 94.917 -76.492 52.348
186799 94.961 -76.502 52.297
186800 95.000 -76.510 52.246
186801 95.039 -76.517 52.196
186802 95.078 -76.526 52.146
186803 95.118 -76.534 52.096
186804 95.157 -76.542 52.048
186805 95.196 -76.550 52.000
186806 95.219 -76.562 51.953
186807 95.234 -76.577 51.906
186808 95.248 -76.591 51.859
186809 95.287 -76.601 51.812
186810 95.329 -76.611 51.767
186811 95.371 -76.621 51.723
186812 95.412 -76.630 51.677
186813 95.454 -76.641 51.630
186814 95.496 -76.651 51.584
186815 95.533 -76.661 51.539
186816 95.552 -76.671 51.497
186817 95.570 -76.681 51.456
186818 95.589 -76.692 51.418
186819 95.590 -76.706 51.382
186820 95.587 -76.722 51.346
186821 95.583 -76.738 51.310
186822 95.580 -76.754 51.272
186823 95.576 -76.770 51.233
186824 95.572 -76.786 51.192
186825 95.568 -76.802 51.151
186826 95.588 -76.808 51.110
186827 95.615 -76.811 51.070
186828 95.642 -76.813 51.028
186829 95.670 -76.816 50.985
186830 95.698 -76.818 50.943
186831 95.726 -76.821 50.902
186832 95.753 -76.824 50.862
186833 95.782 -76.827 50.821
186834 95.801 -76.826 50.777
186835 95.828 -76.832 50.728
186836 95.854 -76.835 50.677
186837 95.879 -76.836 50.622
186838 95.905 -76.837 50.565
186839 95.930 -76.838 50.511
186840 95.956 -76.840 50.459
186841 95.981 -76.840 50.409
186842 96.006 -76.842 50.351
186843 96.020 -76.857 50.289
186844 96.040 -76.869 50.224
186845 96.065 -76.878 50.158
186846 96.090 -76.887 50.095
186847 96.115 -76.895 50.036
186848 96.141 -76.904 49.981
186849 96.166 -76.912 49.931
186850 96.187 -76.922 49.885
186851 96.177 -76.941 49.842
186852 96.167 -76.959 49.801
186853 96.157 -76.978 49.761
186854 96.167 -76.991 49.726
186855 96.176 -77.005 49.695
186856 96.186 -77.019 49.666
186857 96.195 -77.031 49.638
186858 96.205 -77.045 49.612
186859 96.233 -77.041 49.589
186860 96.322 -76.982 49.567
186861 96.336 -76.996 49.549
186862 96.350 -77.009 49.533
186863 96.365 -77.022 49.518
186864 96.379 -77.036 49.501
186865 96.393 -77.050 49.486
186866 96.407 -77.063 49.471
186867 96.427 -77.072 49.457
186868 96.457 -77.072 49.443
186869 96.487 -77.071 49.428
186870 96.518 -77.056 49.413
186871 96.544 -77.050 49.395
186872 96.559 -77.075 49.373
186873 96.575 -77.101 49.351
186874 96.592 -77.120 49.328
186875 96.612 -77.127 49.307
186876 96.632 -77.134 49.286
186877 96.652 -77.141 49.264
186878 96.672 -77.147 49.240
186879 96.692 -77.152 49.215
186880 96.713 -77.157 49.189
186881 96.733 -77.161 49.163
186882 96.753 -77.166 49.137
186883 96.764 -77.169 49.113
186884 96.774 -77.172 49.089
186885 96.784 -77.176 49.065
186886 96.794 -77.179 49.041
186887 96.804 -77.183 49.017
186888 96.813 -77.186 48.993
186889 96.823 -77.190 48.969
186890 96.833 -77.193 48.944
186891 96.843 -77.196 48.918
186892 96.854 -77.200 48.892
186893 96.868 -77.204 48.867
186894 96.883 -77.210 48.841
186895 96.915 -77.213 48.816
186896 96.966 -77.222 48.792
186897 97.017 -77.231 48.768
186898 97.067 -77.241 48.744
186899 97.114 -77.257 48.720
186900 97.145 -77.269 48.696
186901 97.161 -77.279 48.673
186902 97.178 -77.287 48.650
186903 97.194 -77.295 48.628
186904 97.211 -77.304 48.607
186905 97.231 -77.319 48.587
186906 97.256 -77.338 48.566
186907 97.280 -77.357 48.546
186908 97.305 -77.376 48.525
186909 97.329 -77.396 48.505
186910 97.354 -77.415 48.485
186911 97.378 -77.435 48.465
186912 97.394 -77.438 48.445
186913 97.399 -77.420 48.425
186914 97.404 -77.402 48.406
186915 97.415 -77.404 48.386
186916 97.428 -77.423 48.366
186917 97.441 -77.440 48.346
186918 97.454 -77.458 48.325
186919 97.467 -77.476 48.304
186920 97.479 -77.491 48.282
186921 97.490 -77.480 48.260
186922 97.511 -77.489 48.238
186923 97.507 -77.489 48.213
186924 97.534 -77.487 48.184
186925 97.561 -77.487 48.155
186926 97.589 -77.489 48.126
186927 97.616 -77.491 48.095
186928 97.644 -77.493 48.065
186929 97.671 -77.495 48.034
186930 97.699 -77.497 48.003
186931 97.729 -77.489 47.972
186932 97.759 -77.477 47.941
186933 97.790 -77.465 47.909
186934 97.826 -77.461 47.876
186935 97.864 -77.457 47.843
186936 97.901 -77.454 47.811
186937 97.938 -77.451 47.778
186938 97.975 -77.447 47.745
186939 98.012 -77.443 47.713
186940 98.050 -77.440 47.682
186941 98.086 -77.436 47.653
186942 98.107 -77.432 47.627
186943 98.118 -77.428 47.600
186944 98.129 -77.423 47.571
186945 98.139 -77.420 47.543
186946 98.150 -77.415 47.515
186947 98.161 -77.410 47.487
186948 98.171 -77.407 47.459
186949 98.182 -77.402 47.432
186950 98.193 -77.396 47.404
186951 98.202 -77.381 47.377
186952 98.211 -77.366 47.350
186953 98.221 -77.351 47.323
186954 98.230 -77.335 47.296
186955 98.268 -77.339 47.269
186956 98.317 -77.352 47.243
186957 98.367 -77.364 47.217
186958 98.416 -77.376 47.194
186959 98.466 -77.389 47.172
186960 98.515 -77.401 47.149
186961 98.565 -77.413 47.126
186962 98.598 -77.417 47.104
186963 98.577 -77.377 47.083
186964 98.582 -77.379 47.062
186965 98.595 -77.395 47.040
186966 98.608 -77.412 47.018
186967 98.622 -77.428 46.996
186968 98.635 -77.445 46.973
186969 98.648 -77.461 46.950
186970 98.661 -77.477 46.927
186971 98.674 -77.486 46.904
186972 98.686 -77.482 46.881
186973 98.697 -77.478 46.857
186974 98.709 -77.473 46.833
186975 98.720 -77.470 46.810
186976 98.732 -77.466 46.785
186977 98.743 -77.462 46.762
186978 98.755 -77.457 46.739
186979 98.767 -77.454 46.717
186980 98.782 -77.450 46.694
186981 98.801 -77.445 46.671
186982 98.821 -77.442 46.648
186983 98.840 -77.437 46.625
186984 98.856 -77.441 46.602
186985 98.870 -77.448 46.579
186986 98.874 -77.457 46.556
186987 98.867 -77.468 46.531
186988 98.860 -77.479 46.509
186989 98.878 -77.489 46.488
186990 98.958 -77.495 46.467
186991 98.979 -77.504 46.446
186992 99.000 -77.516 46.424
186993 99.020 -77.524 46.401
186994 99.040 -77.533 46.377
186995 99.056 -77.542 46.355
186996 99.054 -77.555 46.333
186997 99.053 -77.569 46.311
186998 99.051 -77.582 46.289
186999 99.050 -77.595 46.267
187000 99.049 -77.608 46.245
187001 99.047 -77.622 46.223
187002 99.043 -77.634 46.200
187003 99.030 -77.635 46.177
187004 99.041 -77.635 46.155
187005 99.051 -77.635 46.134
187006 99.060 -77.635 46.113
187007 99.070 -77.635 46.091
187008 99.080 -77.636 46.070
187009 99.089 -77.640 46.049
187010 99.098 -77.644 46.029
187011 99.107 -77.649 46.011
187012 99.135 -77.649 45.992
187013 99.165 -77.648 45.974
187014 99.196 -77.647 45.957
187015 99.225 -77.646 45.941
187016 99.256 -77.645 45.924
187017 99.287 -77.652 45.909
187018 99.322 -77.669 45.896
187019 99.355 -77.686 45.884
187020 99.389 -77.702 45.872
187021 99.423 -77.718 45.861
187022 99.457 -77.735 45.850
187023 99.491 -77.752 45.839
187024 99.525 -77.768 45.830
187025 99.559 -77.784 45.818
187026 99.593 -77.801 45.806
187027 99.627 -77.817 45.794
187028 99.661 -77.834 45.781
187029 99.695 -77.850 45.769
187030 99.736 -77.855 45.756
187031 99.749 -77.868 45.743
187032 99.771 -77.886 45.732
187033 99.791 -77.897 45.722
187034 99.812 -77.909 45.711
187035 99.833 -77.921 45.700
187036 99.854 -77.933 45.690
187037 99.874 -77.944 45.680
187038 99.896 -77.956 45.670
187039 99.916 -77.968 45.661
187040 99.937 -77.980 45.649
187041 99.958 -77.992 45.639
187042 99.979 -78.003 45.628
187043 100.009 -78.009 45.616
187044 100.039 -78.014 45.605
187045 100.069 -78.018 45.593
187046 100.100 -78.024 45.580
187047 100.130 -78.029 45.567
187048 100.160 -78.034 45.554
187049 100.190 -78.040 45.540
187050 100.221 -78.044 45.527
187051 100.251 -78.050 45.512
187052 100.281 -78.055 45.498
187053 100.302 -78.069 45.484
187054 100.317 -78.079 45.471
187055 100.329 -78.087 45.457
187056 100.338 -78.100 45.442
187057 100.348 -78.115 45.428
187058 100.357 -78.128 45.414
187059 100.366 -78.142 45.398
187060 100.376 -78.156 45.384
187061 100.385 -78.170 45.370
187062 100.395 -78.184 45.355
187063 100.404 -78.198 45.339
187064 100.414 -78.213 45.325
187065 100.423 -78.226 45.310
187066 100.445 -78.219 45.295
187067 100.471 -78.224 45.279
187068 100.509 -78.225 45.264
187069 100.551 -78.225 45.249
187070 100.594 -78.226 45.234
187071 100.641 -78.235 45.219
187072 100.688 -78.242 45.204
187073 100.735 -78.250 45.188
187074 100.782 -78.257 45.173
187075 100.822 -78.264 45.159
187076 100.860 -78.270 45.144
187077 100.898 -78.277 45.128
187078 100.936 -78.283 45.114
187079 100.974 -78.289 45.099
187080 101.012 -78.296 45.084
187081 101.050 -78.302 45.069
187082 101.088 -78.308 45.054
187083 101.126 -78.315 45.038
187084 101.159 -78.321 45.022
187085 101.185 -78.329 45.006
187086 101.201 -78.339 44.991
187087 101.217 -78.350 44.976
187088 101.233 -78.360 44.960
187089 101.248 -78.371 44.944
187090 101.262 -78.386 44.928
187091 101.275 -78.404 44.912
187092 101.289 -78.420 44.897
187093 101.302 -78.437 44.881
187094 101.315 -78.454 44.864
187095 101.337 -78.473 44.848
187096 101.363 -78.495 44.832
187097 101.389 -78.515 44.815
187098 101.415 -78.535 44.799
187099 101.441 -78.556 44.782
187100 101.467 -78.576 44.766
187101 101.493 -78.598 44.750
187102 101.519 -78.618 44.734
187103 101.546 -78.643 44.719
187104 101.582 -78.666 44.703
187105 101.620 -78.692 44.687
187106 101.659 -78.717 44.671
187107 101.697 -78.742 44.656
187108 101.736 -78.767 44.640
187109 101.775 -78.793 44.624
187110 101.814 -78.818 44.609
187111 101.852 -78.843 44.593
187112 101.891 -78.869 44.578
187113 101.930 -78.894 44.563
187114 101.969 -78.920 44.548
187115 101.934 -78.997 44.534
187116 101.952 -79.004 44.521
187117 101.971 -79.010 44.506
187118 101.989 -79.016 44.494
187119 102.007 -79.022 44.481
187120 102.026 -79.028 44.468
187121 102.044 -79.035 44.455
187122 102.062 -79.041 44.442
187123 102.081 -79.047 44.430
187124 102.099 -79.053 44.416
187125 102.117 -79.059 44.403
187398 93.315 -76.522 54.117
187399 93.335 -76.519 54.118
187400 93.366 -76.515 54.120
187401 93.397 -76.511 54.121
187402 93.419 -76.511 54.122
187403 93.437 -76.514 54.123
187404 93.456 -76.517 54.124
187405 93.474 -76.520 54.123
187406 93.493 -76.524 54.123
187407 93.511 -76.526 54.123
187408 93.530 -76.529 54.124
187409 93.548 -76.532 54.126
187410 93.566 -76.535 54.127
187411 93.584 -76.538 54.130
187412 93.603 -76.541 54.131
187413 93.621 -76.543 54.133
187414 93.639 -76.546 54.135
187415 93.659 -76.543 54.139
187416 93.680 -76.533 54.146
187417 93.701 -76.524 54.153
187418 93.722 -76.514 54.162
187419 93.744 -76.505 54.171
187420 93.758 -76.505 54.181
187421 93.759 -76.493 54.191
187422 93.782 -76.506 54.203
187423 93.797 -76.503 54.212
187424 93.812 -76.499 54.221
187425 93.827 -76.495 54.229
187426 93.841 -76.492 54.235
187427 93.856 -76.488 54.239
187428 93.872 -76.483 54.242
187429 93.889 -76.478 54.243
187430 93.896 -76.475 54.242
187431 93.908 -76.469 54.237
187432 93.920 -76.464 54.228
187433 93.932 -76.458 54.217
187434 93.939 -76.459 54.209
187435 93.940 -76.467 54.201
187436 93.941 -76.473 54.194
187437 93.950 -76.479 54.187
187438 93.966 -76.480 54.181
187439 93.984 -76.477 54.176
187440 94.006 -76.472 54.171
187441 94.034 -76.467 54.167
187442 94.062 -76.461 54.167
187443 94.090 -76.456 54.168
187444 94.119 -76.451 54.170
187445 94.147 -76.445 54.170
187446 94.161 -76.442 54.168
187447 94.178 -76.436 54.161
187448 94.194 -76.429 54.150
187449 94.211 -76.424 54.137
187450 94.231 -76.423 54.122
187451 94.251 -76.421 54.105
187452 94.271 -76.420 54.088
187453 94.290 -76.418 54.063
187454 94.310 -76.417 54.037
187455 94.330 -76.415 54.009
187456 94.349 -76.413 53.979
187457 94.369 -76.411 53.946
187458 94.389 -76.410 53.913
187459 94.409 -76.408 53.879
187460 94.427 -76.400 53.839
187461 94.443 -76.386 53.799
187462 94.469 -76.387 53.759
187463 94.495 -76.389 53.718
187464 94.522 -76.391 53.680
187465 94.542 -76.389 53.645
187466 94.563 -76.386 53.611
187467 94.583 -76.383 53.575
187468 94.604 -76.381 53.538
187469 94.624 -76.379 53.502
187470 94.639 -76.383 53.467
187471 94.652 -76.389 53.434
187472 94.665 -76.396 53.400
187473 94.680 -76.396 53.364
187474 94.696 -76.395 53.326
187475 94.713 -76.392 53.289
187476 94.725 -76.391 53.255
187477 94.721 -76.393 53.224
187478 94.717 -76.395 53.195
187479 94.713 -76.398 53.167
187480 94.710 -76.400 53.138
187481 94.706 -76.402 53.110
187482 94.702 -76.404 53.083
187483 94.698 -76.407 53.059
187484 94.694 -76.409 53.036
187485 94.690 -76.411 53.012
187486 94.692 -76.414 52.983
187487 94.701 -76.418 52.948
187488 94.710 -76.422 52.908
187489 94.720 -76.426 52.863
187490 94.729 -76.430 52.816
187491 94.738 -76.433 52.767
187492 94.747 -76.437 52.718
187493 94.756 -76.442 52.669
187494 94.765 -76.445 52.618
187495 94.774 -76.449 52.568
187496 94.783 -76.453 52.518
187497 94.803 -76.459 52.470
187498 94.853 -76.472 52.419
187499 94.904 -76.485 52.369
187500 94.946 -76.494 52.321
187501 94.984 -76.502 52.270
187502 95.024 -76.510 52.219
187503 95.063 -76.518 52.168
187504 95.102 -76.526 52.116
187505 95.141 -76.534 52.066
187506 95.180 -76.542 52.018
187507 95.214 -76.552 51.970
187508 95.228 -76.566 51.921
187509 95.249 -76.580 51.874
187510 95.291 -76.589 51.828
187511 95.333 -76.599 51.784
187512 95.375 -76.609 51.743
187513 95.416 -76.619 51.698
187514 95.450 -76.629 51.652
187515 95.469 -76.640 51.605
187516 95.487 -76.650 51.558
187517 95.506 -76.660 51.514
187518 95.525 -76.671 51.471
187519 95.543 -76.680 51.431
187520 95.562 -76.691 51.393
187521 95.569 -76.704 51.356
187522 95.566 -76.720 51.319
187523 95.562 -76.736 51.281
187524 95.558 -76.752 51.240
187525 95.555 -76.768 51.200
187526 95.564 -76.778 51.158
187527 95.591 -76.781 51.117
187528 95.619 -76.784 51.076
187529 95.647 -76.786 51.036
187530 95.674 -76.789 50.995
187531 95.702 -76.792 50.953
187532 95.732 -76.795 50.913
187533 95.762 -76.798 50.873
187534 95.788 -76.802 50.832
187535 95.811 -76.806 50.788
187536 95.817 -76.794 50.741
187537 95.840 -76.797 50.690
187538 95.867 -76.803 50.635
187539 95.894 -76.809 50.583
187540 95.921 -76.816 50.532
187541 95.948 -76.822 50.483
187542 95.974 -76.826 50.434
187543 96.000 -76.828 50.377
187544 96.021 -76.837 50.316
187545 96.046 -76.846 50.251
187546 96.072 -76.855 50.185
187547 96.097 -76.864 50.123
187548 96.122 -76.872 50.064
187549 96.147 -76.881 50.009
187550 96.172 -76.890 49.960
187551 96.184 -76.903 49.914
187552 96.174 -76.921 49.871
187553 96.164 -76.940 49.829
187554 96.172 -76.953 49.790
187555 96.181 -76.967 49.755
187556 96.191 -76.981 49.724
187557 96.200 -76.994 49.695
187558 96.210 -77.007 49.668
187559 96.220 -77.021 49.642
187560 96.230 -77.034 49.618
187561 96.329 -76.964 49.597
187562 96.368 -76.953 49.579
187563 96.382 -76.967 49.563
187564 96.396 -76.981 49.547
187565 96.410 -76.994 49.531
187566 96.424 -77.008 49.517
187567 96.438 -77.022 49.503
187568 96.452 -77.035 49.488
187569 96.466 -77.049 49.475
187570 96.489 -77.049 49.460
187571 96.516 -77.032 49.443
187572 96.513 -77.059 49.422
187573 96.550 -77.065 49.400
187574 96.575 -77.083 49.376
187575 96.591 -77.109 49.353
187576 96.610 -77.117 49.332
187577 96.630 -77.124 49.311
187578 96.651 -77.129 49.288
187579 96.671 -77.134 49.264
187580 96.691 -77.138 49.237
187581 96.711 -77.144 49.211
187582 96.731 -77.148 49.184
187583 96.751 -77.153 49.157
187584 96.767 -77.157 49.133
187585 96.777 -77.160 49.107
187586 96.787 -77.164 49.082
187587 96.797 -77.167 49.058
187588 96.807 -77.170 49.033
187589 96.817 -77.174 49.009
187590 96.827 -77.177 48.985
187591 96.836 -77.181 48.960
187592 96.847 -77.184 48.934
187593 96.858 -77.188 48.908
187594 96.870 -77.192 48.883
187595 96.882 -77.196 48.857
187596 96.919 -77.203 48.832
187597 96.969 -77.213 48.807
187598 97.020 -77.222 48.782
187599 97.070 -77.232 48.757
187600 97.116 -77.249 48.732
187601 97.163 -77.265 48.708
187602 97.179 -77.274 48.685
187603 97.195 -77.282 48.662
187604 97.212 -77.291 48.640
187605 97.229 -77.300 48.618
187606 97.245 -77.308 48.598
187607 97.261 -77.317 48.576
187608 97.278 -77.326 48.555
187609 97.301 -77.343 48.535
187610 97.325 -77.363 48.515
187611 97.350 -77.382 48.495
187612 97.375 -77.401 48.476
187613 97.399 -77.421 48.456
187614 97.405 -77.405 48.435
187615 97.410 -77.387 48.415
187616 97.424 -77.404 48.395
187617 97.437 -77.422 48.374
187618 97.450 -77.439 48.353
187619 97.460 -77.450 48.332
187620 97.477 -77.449 48.310
187621 97.499 -77.457 48.288
187622 97.520 -77.466 48.266
187623 97.526 -77.470 48.242
187624 97.522 -77.464 48.217
187625 97.548 -77.462 48.189
187626 97.576 -77.461 48.160
187627 97.603 -77.462 48.130
187628 97.631 -77.464 48.099
187629 97.658 -77.466 48.069
187630 97.685 -77.468 48.038
187631 97.713 -77.470 48.007
187632 97.743 -77.463 47.976
187633 97.774 -77.451 47.944
187634 97.804 -77.439 47.912
187635 97.835 -77.427 47.880
187636 97.865 -77.416 47.847
187637 97.902 -77.411 47.814
187638 97.939 -77.408 47.782
187639 97.977 -77.404 47.750
187640 98.013 -77.401 47.718
187641 98.051 -77.398 47.687
187642 98.088 -77.394 47.658
187643 98.111 -77.390 47.630
187644 98.122 -77.385 47.602
187645 98.132 -77.381 47.574
187646 98.143 -77.377 47.546
187647 98.153 -77.373 47.519
187648 98.164 -77.369 47.491
187649 98.175 -77.364 47.464
187650 98.185 -77.360 47.437
187651 98.196 -77.356 47.409
187652 98.207 -77.351 47.382
187653 98.217 -77.347 47.355
187654 98.227 -77.336 47.327
187655 98.236 -77.321 47.300
187656 98.258 -77.315 47.273
187657 98.308 -77.327 47.246
187658 98.358 -77.339 47.220
187659 98.407 -77.351 47.197
187660 98.457 -77.363 47.173
187661 98.511 -77.375 47.150
187662 98.564 -77.387 47.128
187663 98.572 -77.369 47.106
187664 98.569 -77.351 47.085
187665 98.598 -77.368 47.064
187666 98.613 -77.385 47.043
187667 98.627 -77.401 47.021
187668 98.640 -77.417 47.000
187669 98.653 -77.434 46.977
187670 98.666 -77.450 46.954
187671 98.679 -77.467 46.932
187672 98.692 -77.468 46.910
187673 98.703 -77.464 46.887
187674 98.715 -77.460 46.863
187675 98.726 -77.456 46.840
187676 98.738 -77.452 46.816
187677 98.749 -77.448 46.793
187678 98.761 -77.444 46.770
187679 98.772 -77.440 46.748
187680 98.784 -77.436 46.726
187681 98.795 -77.432 46.703
187682 98.810 -77.428 46.680
187683 98.830 -77.423 46.657
187684 98.849 -77.420 46.633
187685 98.864 -77.425 46.610
187686 98.871 -77.432 46.587
187687 98.865 -77.444 46.563
187688 98.866 -77.455 46.539
187689 98.934 -77.461 46.516
187690 98.954 -77.467 46.494
187691 98.974 -77.472 46.472
187692 98.997 -77.487 46.451
187693 99.020 -77.504 46.429
187694 99.041 -77.512 46.406
187695 99.060 -77.520 46.382
187696 99.060 -77.533 46.359
187697 99.059 -77.546 46.336
187698 99.058 -77.560 46.314
187699 99.056 -77.573 46.292
187700 99.055 -77.586 46.270
187701 99.053 -77.600 46.248
187702 99.052 -77.613 46.226
187703 99.051 -77.627 46.203
187704 99.035 -77.633 46.180
187705 99.046 -77.633 46.157
187706 99.055 -77.633 46.136
187707 99.065 -77.633 46.114
187708 99.075 -77.633 46.093
187709 99.084 -77.638 46.073
187710 99.093 -77.642 46.052
187711 99.114 -77.643 46.032
187712 99.144 -77.642 46.013
187713 99.174 -77.642 45.995
187714 99.204 -77.642 45.977
187715 99.234 -77.641 45.960
187716 99.264 -77.640 45.943
187717 99.294 -77.639 45.927
187718 99.324 -77.639 45.911
187719 99.357 -77.652 45.898
187720 99.392 -77.668 45.885
187721 99.425 -77.685 45.873
187722 99.460 -77.702 45.862
187723 99.493 -77.718 45.850
187724 99.527 -77.734 45.840
187725 99.561 -77.751 45.829
187726 99.595 -77.768 45.817
187727 99.629 -77.783 45.805
187728 99.663 -77.800 45.793
187729 99.697 -77.817 45.779
187730 99.739 -77.825 45.767
187731 99.756 -77.836 45.754
187732 99.769 -77.848 45.741
187733 99.788 -77.867 45.730
187734 99.810 -77.884 45.718
187735 99.831 -77.896 45.707
187736 99.852 -77.907 45.696
187737 99.872 -77.919 45.686
187738 99.893 -77.931 45.676
187739 99.914 -77.943 45.666
187740 99.935 -77.954 45.656
187741 99.956 -77.966 45.646
187742 99.977 -77.978 45.636
187743 100.002 -77.987 45.625
187744 100.032 -77.992 45.614
187745 100.062 -77.996 45.602
187746 100.092 -78.002 45.591
187747 100.123 -78.007 45.579
187748 100.153 -78.012 45.567
187749 100.183 -78.018 45.554
187750 100.213 -78.022 45.541
187751 100.244 -78.028 45.527
187752 100.268 -78.038 45.512
187753 100.288 -78.052 45.498
187754 100.309 -78.065 45.484
187755 100.325 -78.078 45.471
187756 100.337 -78.084 45.457
187757 100.350 -78.092 45.442
187758 100.361 -78.103 45.428
187759 100.370 -78.118 45.414
187760 100.380 -78.131 45.399
187761 100.390 -78.145 45.384
187762 100.399 -78.159 45.369
187763 100.408 -78.173 45.354
187764 100.418 -78.188 45.339
187765 100.427 -78.201 45.325
187766 100.436 -78.216 45.310
187767 100.458 -78.208 45.295
187768 100.483 -78.206 45.281
187769 100.519 -78.209 45.266
187770 100.561 -78.209 45.251
187771 100.606 -78.213 45.235
187772 100.652 -78.221 45.220
187773 100.699 -78.228 45.206
187774 100.744 -78.236 45.190
187775 100.782 -78.242 45.175
187776 100.820 -78.249 45.160
187777 100.858 -78.255 45.146
187778 100.896 -78.261 45.131
187779 100.934 -78.268 45.116
187780 100.972 -78.274 45.100
187781 101.010 -78.280 45.086
187782 101.048 -78.287 45.072
187783 101.086 -78.293 45.057
187784 101.124 -78.299 45.042
187785 101.162 -78.306 45.025
187786 101.192 -78.313 45.010
187787 101.208 -78.323 44.995
187788 101.223 -78.335 44.980
187789 101.236 -78.353 44.963
187790 101.249 -78.369 44.947
187791 101.262 -78.386 44.932
187792 101.276 -78.404 44.916
187793 101.289 -78.420 44.900
187794 101.302 -78.437 44.884
187795 101.317 -78.455 44.867
187796 101.343 -78.476 44.851
187797 101.369 -78.496 44.834
187798 101.395 -78.517 44.818
187799 101.421 -78.537 44.801
187800 101.447 -78.558 44.785
187801 101.473 -78.579 44.769
187802 101.490 -78.589 44.753
187803 101.525 -78.610 44.738
187804 101.555 -78.633 44.722
187805 101.593 -78.660 44.706
187806 101.636 -78.687 44.690
187807 101.679 -78.714 44.673
187808 101.720 -78.740 44.658
187809 101.758 -78.766 44.642
187810 101.797 -78.791 44.627
187811 101.836 -78.817 44.611
187812 101.875 -78.842 44.594
187813 101.913 -78.868 44.578
187814 101.952 -78.893 44.563
187815 101.991 -78.918 44.549
187816 101.953 -79.006 44.535
187817 101.971 -79.012 44.522
187818 101.989 -79.018 44.508
187819 102.007 -79.024 44.495
187820 102.026 -79.031 44.482
187821 102.044 -79.037 44.469
187822 102.062 -79.043 44.456
187823 102.081 -79.049 44.444
187824 102.100 -79.054 44.431
187825 102.119 -79.056 44.417
187826 102.138 -79.057 44.405
188100 93.328 -76.495 54.109
188101 93.359 -76.491 54.111
188102 93.390 -76.487 54.112
188103 93.411 -76.488 54.114
188104 93.429 -76.491 54.115
188105 93.448 -76.494 54.116
188106 93.466 -76.497 54.116
188107 93.485 -76.500 54.115
188108 93.503 -76.502 54.116
188109 93.521 -76.505 54.117
188110 93.540 -76.508 54.118
188111 93.558 -76.511 54.121
188112 93.576 -76.514 54.123
188113 93.594 -76.517 54.125
188114 93.613 -76.520 54.126
188115 93.631 -76.523 54.129
188116 93.649 -76.526 54.132
188117 93.669 -76.524 54.138
188118 93.690 -76.514 54.145
188119 93.711 -76.505 54.153
188120 93.732 -76.495 54.160
188121 93.751 -76.489 54.170
188122 93.766 -76.490 54.181
188123 93.780 -76.487 54.193
188124 93.795 -76.483 54.203
188125 93.809 -76.480 54.214
188126 93.824 -76.477 54.224
188127 93.839 -76.473 54.231
188128 93.854 -76.469 54.236
188129 93.868 -76.465 54.241
188130 93.883 -76.461 54.244
188131 93.896 -76.455 54.244
188132 93.909 -76.450 54.241
188133 93.921 -76.445 54.234
188134 93.933 -76.439 54.223
188135 93.940 -76.439 54.215
188136 93.941 -76.447 54.208
188137 93.952 -76.451 54.201
188138 93.966 -76.456 54.195
188139 93.980 -76.460 54.190
188140 93.998 -76.455 54.185
188141 94.016 -76.451 54.178
188142 94.035 -76.447 54.175
188143 94.054 -76.442 54.175
188144 94.082 -76.438 54.177
188145 94.111 -76.433 54.179
188146 94.139 -76.427 54.181
188147 94.161 -76.423 54.177
188148 94.176 -76.417 54.170
188149 94.193 -76.411 54.159
188150 94.212 -76.409 54.146
188151 94.232 -76.408 54.130
188152 94.252 -76.405 54.114
188153 94.271 -76.404 54.098
188154 94.291 -76.402 54.074
188155 94.311 -76.401 54.047
188156 94.331 -76.399 54.019
188157 94.350 -76.398 53.988
188158 94.370 -76.396 53.956
188159 94.390 -76.395 53.923
188160 94.409 -76.393 53.890
188161 94.429 -76.391 53.852
188162 94.448 -76.382 53.813
188163 94.474 -76.383 53.772
188164 94.496 -76.381 53.730
188165 94.516 -76.379 53.692
188166 94.537 -76.376 53.657
188167 94.557 -76.374 53.624
188168 94.577 -76.372 53.590
188169 94.598 -76.370 53.555
188170 94.618 -76.367 53.519
188171 94.631 -76.373 53.484
188172 94.644 -76.379 53.452
188173 94.657 -76.387 53.420
188174 94.670 -76.392 53.385
188175 94.686 -76.390 53.347
188176 94.703 -76.389 53.308
188177 94.717 -76.387 53.273
188178 94.713 -76.389 53.241
188179 94.709 -76.392 53.211
188180 94.705 -76.394 53.182
188181 94.701 -76.395 53.152
188182 94.697 -76.398 53.122
188183 94.693 -76.400 53.092
188184 94.689 -76.402 53.066
188185 94.685 -76.404 53.040
188186 94.693 -76.408 53.015
188187 94.702 -76.412 52.985
188188 94.711 -76.416 52.949
188189 94.720 -76.420 52.910
188190 94.729 -76.424 52.867
188191 94.738 -76.428 52.821
188192 94.747 -76.432 52.775
188193 94.757 -76.436 52.729
188194 94.766 -76.439 52.682
188195 94.775 -76.443 52.633
188196 94.784 -76.447 52.584
188197 94.793 -76.451 52.537
188198 94.802 -76.455 52.490
188199 94.840 -76.465 52.441
188200 94.891 -76.478 52.392
188201 94.930 -76.486 52.344
188202 94.969 -76.494 52.292
188203 95.008 -76.502 52.240
188204 95.047 -76.510 52.187
188205 95.087 -76.518 52.133
188206 95.126 -76.527 52.082
188207 95.165 -76.535 52.033
188208 95.204 -76.542 51.985
188209 95.223 -76.556 51.936
188210 95.253 -76.567 51.890
188211 95.295 -76.577 51.843
188212 95.336 -76.587 51.799
188213 95.367 -76.598 51.758
188214 95.386 -76.608 51.715
188215 95.404 -76.618 51.670
188216 95.422 -76.629 51.624
188217 95.441 -76.639 51.576
188218 95.460 -76.649 51.530
188219 95.478 -76.660 51.486
188220 95.497 -76.670 51.444
188221 95.516 -76.680 51.405
188222 95.534 -76.690 51.367
188223 95.548 -76.702 51.328
188224 95.545 -76.718 51.290
188225 95.541 -76.734 51.249
188226 95.540 -76.749 51.209
188227 95.567 -76.752 51.168
188228 95.595 -76.754 51.127
188229 95.623 -76.756 51.087
188230 95.652 -76.759 51.047
188231 95.682 -76.763 51.006
188232 95.711 -76.766 50.965
188233 95.741 -76.769 50.923
188234 95.770 -76.772 50.881
188235 95.794 -76.777 50.840
188236 95.816 -76.781 50.797
188237 95.834 -76.782 50.750
188238 95.837 -76.767 50.702
188239 95.852 -76.762 50.652
188240 95.879 -76.768 50.603
188241 95.906 -76.774 50.555
188242 95.933 -76.781 50.506
188243 95.967 -76.793 50.457
188244 96.001 -76.807 50.403
188245 96.025 -76.815 50.343
188246 96.053 -76.824 50.280
188247 96.078 -76.833 50.215
188248 96.103 -76.841 50.153
188249 96.128 -76.850 50.095
188250 96.154 -76.859 50.040
188251 96.179 -76.868 49.990
188252 96.182 -76.882 49.945
188253 96.172 -76.901 49.902
188254 96.177 -76.915 49.860
188255 96.186 -76.929 49.821
188256 96.196 -76.942 49.786
188257 96.206 -76.956 49.755
188258 96.216 -76.969 49.726
188259 96.225 -76.983 49.700
188260 96.235 -76.996 49.674
188261 96.244 -77.009 49.650
188262 96.320 -76.962 49.628
188263 96.399 -76.912 49.610
188264 96.413 -76.925 49.593
188265 96.428 -76.939 49.578
188266 96.442 -76.953 49.563
188267 96.456 -76.966 49.549
188268 96.470 -76.980 49.535
188269 96.464 -76.979 49.520
188270 96.474 -77.021 49.507
188271 96.515 -77.029 49.490
188272 96.494 -77.035 49.471
188273 96.477 -77.074 49.450
188274 96.511 -77.082 49.427
188275 96.549 -77.088 49.403
188276 96.587 -77.094 49.380
188277 96.609 -77.106 49.359
188278 96.629 -77.111 49.337
188279 96.649 -77.116 49.314
188280 96.669 -77.121 49.289
188281 96.689 -77.125 49.262
188282 96.710 -77.130 49.234
188283 96.730 -77.135 49.206
188284 96.750 -77.140 49.178
188285 96.770 -77.144 49.153
188286 96.781 -77.148 49.127
188287 96.791 -77.151 49.101
188288 96.803 -77.156 49.076
188289 96.815 -77.160 49.051
188290 96.826 -77.163 49.026
188291 96.838 -77.168 49.001
188292 96.851 -77.172 48.977
188293 96.863 -77.175 48.951
188294 96.875 -77.180 48.925
188295 96.886 -77.184 48.900
188296 96.898 -77.188 48.874
188297 96.922 -77.194 48.849
188298 96.973 -77.203 48.823
188299 97.023 -77.212 48.797
188300 97.073 -77.224 48.771
188301 97.119 -77.241 48.747
188302 97.166 -77.257 48.723
188303 97.197 -77.269 48.700
188304 97.214 -77.278 48.676
188305 97.230 -77.287 48.653
188306 97.246 -77.295 48.631
188307 97.263 -77.304 48.609
188308 97.279 -77.313 48.587
188309 97.296 -77.321 48.566
188310 97.312 -77.330 48.546
188311 97.329 -77.338 48.526
188312 97.347 -77.349 48.506
188313 97.371 -77.368 48.486
188314 97.395 -77.388 48.466
188315 97.412 -77.390 48.445
188316 97.419 -77.385 48.425
188317 97.433 -77.403 48.404
188318 97.443 -77.410 48.383
188319 97.464 -77.419 48.362
188320 97.486 -77.426 48.340
188321 97.508 -77.435 48.317
188322 97.529 -77.443 48.295
188323 97.548 -77.451 48.273
188324 97.511 -77.442 48.248
188325 97.536 -77.439 48.222
188326 97.564 -77.438 48.195
188327 97.590 -77.436 48.165
188328 97.617 -77.435 48.135
188329 97.645 -77.436 48.104
188330 97.672 -77.439 48.073
188331 97.700 -77.441 48.043
188332 97.728 -77.442 48.012
188333 97.757 -77.437 47.980
188334 97.788 -77.425 47.948
188335 97.818 -77.414 47.917
188336 97.849 -77.401 47.885
188337 97.879 -77.390 47.853
188338 97.910 -77.378 47.820
188339 97.941 -77.367 47.787
188340 97.978 -77.362 47.754
188341 98.015 -77.358 47.723
188342 98.052 -77.355 47.692
188343 98.089 -77.351 47.663
188344 98.114 -77.348 47.635
188345 98.125 -77.343 47.606
188346 98.135 -77.339 47.578
188347 98.146 -77.335 47.551
188348 98.157 -77.330 47.524
188349 98.167 -77.326 47.497
188350 98.178 -77.322 47.469
188351 98.189 -77.317 47.442
188352 98.199 -77.313 47.414
188353 98.210 -77.309 47.387
188354 98.221 -77.304 47.360
188355 98.231 -77.301 47.332
188356 98.242 -77.296 47.305
188357 98.252 -77.292 47.277
188358 98.302 -77.301 47.249
188359 98.355 -77.313 47.223
188360 98.408 -77.325 47.199
188361 98.461 -77.336 47.175
188362 98.514 -77.348 47.152
188363 98.567 -77.360 47.130
188364 98.546 -77.323 47.108
188365 98.575 -77.340 47.087
188366 98.604 -77.357 47.067
188367 98.631 -77.374 47.046
188368 98.645 -77.390 47.025
188369 98.658 -77.407 47.003
188370 98.671 -77.423 46.981
188371 98.685 -77.439 46.958
188372 98.698 -77.454 46.937
188373 98.709 -77.450 46.915
188374 98.721 -77.446 46.893
188375 98.732 -77.442 46.871
188376 98.743 -77.438 46.848
188377 98.755 -77.434 46.824
188378 98.767 -77.430 46.801
188379 98.778 -77.426 46.778
188380 98.790 -77.422 46.756
188381 98.801 -77.418 46.735
188382 98.813 -77.414 46.712
188383 98.825 -77.410 46.689
188384 98.839 -77.406 46.665
188385 98.858 -77.401 46.642
188386 98.869 -77.409 46.618
188387 98.863 -77.420 46.595
188388 98.910 -77.426 46.571
188389 98.930 -77.432 46.547
188390 98.950 -77.438 46.524
188391 98.970 -77.443 46.501
188392 98.990 -77.449 46.479
188393 99.016 -77.470 46.458
188394 99.041 -77.491 46.436
188395 99.061 -77.499 46.413
188396 99.067 -77.511 46.389
188397 99.065 -77.524 46.364
188398 99.064 -77.538 46.341
188399 99.062 -77.551 46.318
188400 99.061 -77.564 46.296
188401 99.060 -77.577 46.273
188402 99.058 -77.591 46.251
188403 99.057 -77.605 46.229
188404 99.055 -77.617 46.205
188405 99.054 -77.631 46.182
188406 99.050 -77.631 46.160
188407 99.060 -77.631 46.139
188408 99.070 -77.631 46.117
188409 99.079 -77.635 46.096
188410 99.092 -77.639 46.075
188411 99.122 -77.638 46.055
188412 99.152 -77.637 46.034
188413 99.183 -77.636 46.015
188414 99.212 -77.636 45.997
188415 99.242 -77.635 45.979
188416 99.273 -77.635 45.962
188417 99.302 -77.634 45.945
188418 99.333 -77.633 45.929
188419 99.363 -77.633 45.913
188420 99.394 -77.635 45.899
188421 99.427 -77.652 45.887
188422 99.462 -77.668 45.875
188423 99.495 -77.684 45.863
188424 99.529 -77.701 45.851
188425 99.563 -77.718 45.839
188426 99.597 -77.734 45.828
188427 99.631 -77.750 45.816
188428 99.665 -77.767 45.804
188429 99.699 -77.783 45.792
188430 99.734 -77.799 45.778
188431 99.762 -77.805 45.765
188432 99.775 -77.817 45.752
188433 99.789 -77.829 45.740
188434 99.803 -77.843 45.727
188435 99.828 -77.868 45.715
188436 99.850 -77.882 45.703
188437 99.870 -77.893 45.692
188438 99.891 -77.906 45.681
188439 99.912 -77.917 45.671
188440 99.933 -77.929 45.661
188441 99.954 -77.940 45.652
188442 99.974 -77.953 45.643
188443 99.995 -77.964 45.633
188444 100.025 -77.970 45.622
188445 100.055 -77.975 45.612
188446 100.085 -77.980 45.600
188447 100.116 -77.985 45.589
188448 100.146 -77.990 45.578
188449 100.176 -77.996 45.566
188450 100.206 -78.000 45.553
188451 100.234 -78.008 45.539
188452 100.255 -78.022 45.525
188453 100.275 -78.035 45.512
188454 100.296 -78.049 45.497
188455 100.316 -78.063 45.484
188456 100.333 -78.075 45.471
188457 100.345 -78.082 45.456
188458 100.358 -78.090 45.442
188459 100.371 -78.097 45.428
188460 100.384 -78.106 45.413
188461 100.394 -78.121 45.398
188462 100.403 -78.134 45.383
188463 100.412 -78.148 45.368
188464 100.422 -78.163 45.353
188465 100.431 -78.176 45.339
188466 100.440 -78.191 45.324
188467 100.450 -78.204 45.310
188468 100.472 -78.197 45.296
188469 100.495 -78.188 45.282
188470 100.528 -78.192 45.267
188471 100.570 -78.192 45.252
188472 100.617 -78.200 45.236
188473 100.664 -78.208 45.221
188474 100.704 -78.214 45.206
188475 100.743 -78.221 45.190
188476 100.780 -78.227 45.175
188477 100.819 -78.233 45.161
188478 100.856 -78.240 45.146
188479 100.894 -78.246 45.131
188480 100.932 -78.252 45.116
188481 100.970 -78.259 45.101
188482 101.008 -78.265 45.087
188483 101.046 -78.271 45.073
188484 101.085 -78.278 45.059
188485 101.122 -78.284 45.044
188486 101.163 -78.291 45.028
188487 101.199 -78.300 45.013
188488 101.210 -78.319 44.998
188489 101.223 -78.335 44.982
188490 101.236 -78.353 44.966
188491 101.249 -78.369 44.951
188492 101.262 -78.386 44.935
188493 101.276 -78.404 44.918
188494 101.289 -78.420 44.903
188495 101.302 -78.438 44.887
188496 101.323 -78.457 44.870
188497 101.349 -78.477 44.854
188498 101.375 -78.498 44.837
188499 101.401 -78.518 44.820
188500 101.427 -78.539 44.804
188501 101.435 -78.534 44.788
188502 101.471 -78.555 44.772
188503 101.506 -78.576 44.757
188504 101.541 -78.598 44.742
188505 101.568 -78.623 44.725
188506 101.595 -78.648 44.709
188507 101.634 -78.676 44.693
188508 101.677 -78.702 44.676
188509 101.720 -78.730 44.660
188510 101.763 -78.757 44.644
188511 101.806 -78.784 44.627
188512 101.849 -78.811 44.611
188513 101.892 -78.838 44.596
188514 101.935 -78.865 44.580
188515 101.974 -78.891 44.566
188516 102.013 -78.916 44.552
188517 101.971 -79.014 44.537
188518 101.990 -79.020 44.523
188519 102.008 -79.026 44.510
188520 102.028 -79.028 44.496
188521 102.047 -79.030 44.483
188522 102.066 -79.031 44.470
188523 102.086 -79.033 44.457
188524 102.105 -79.035 44.445
188525 102.125 -79.037 44.432
188526 102.144 -79.038 44.419
188527 102.164 -79.040 44.407
188802 93.352 -76.467 54.100
188803 93.382 -76.463 54.102
188804 93.403 -76.464 54.104
188805 93.421 -76.467 54.106
188806 93.440 -76.470 54.107
188807 93.458 -76.473 54.108
188808 93.477 -76.477 54.108
188809 93.495 -76.479 54.109
188810 93.513 -76.482 54.110
188811 93.532 -76.485 54.112
188812 93.550 -76.488 54.114
188813 93.568 -76.491 54.117
188814 93.586 -76.494 54.119
188815 93.605 -76.497 54.120
188816 93.623 -76.499 54.123
188817 93.641 -76.502 54.127
188818 93.660 -76.505 54.132
188819 93.679 -76.505 54.138
188820 93.700 -76.495 54.146
188821 93.721 -76.486 54.152
188822 93.743 -76.476 54.161
188823 93.759 -76.474 54.171
188824 93.773 -76.473 54.182
188825 93.788 -76.470 54.195
188826 93.802 -76.466 54.207
188827 93.817 -76.462 54.218
188828 93.832 -76.459 54.226
188829 93.846 -76.455 54.232
188830 93.861 -76.451 54.238
188831 93.877 -76.440 54.242
188832 93.892 -76.434 54.244
188833 93.907 -76.430 54.242
188834 93.922 -76.425 54.237
188835 93.934 -76.420 54.228
188836 93.942 -76.420 54.220
188837 93.953 -76.425 54.213
188838 93.967 -76.429 54.207
188839 93.981 -76.433 54.201
188840 93.995 -76.437 54.198
188841 94.012 -76.435 54.194
188842 94.031 -76.430 54.187
188843 94.049 -76.426 54.183
188844 94.068 -76.422 54.183
188845 94.086 -76.418 54.185
188846 94.104 -76.414 54.189
188847 94.130 -76.408 54.191
188848 94.158 -76.404 54.187
188849 94.175 -76.398 54.179
188850 94.194 -76.395 54.170
188851 94.213 -76.393 54.156
188852 94.233 -76.392 54.141
188853 94.253 -76.390 54.125
188854 94.272 -76.389 54.108
188855 94.292 -76.387 54.083
188856 94.312 -76.385 54.056
188857 94.332 -76.383 54.027
188858 94.351 -76.382 53.996
188859 94.371 -76.380 53.964
188860 94.391 -76.379 53.933
188861 94.410 -76.377 53.902
188862 94.430 -76.376 53.866
188863 94.449 -76.374 53.827
188864 94.470 -76.372 53.786
188865 94.490 -76.370 53.745
188866 94.510 -76.367 53.708
188867 94.531 -76.364 53.673
188868 94.551 -76.362 53.639
188869 94.571 -76.360 53.605
188870 94.592 -76.357 53.570
188871 94.611 -76.357 53.536
188872 94.624 -76.364 53.503
188873 94.636 -76.370 53.472
188874 94.649 -76.377 53.441
188875 94.662 -76.384 53.407
188876 94.676 -76.386 53.368
188877 94.693 -76.385 53.329
188878 94.708 -76.383 53.292
188879 94.704 -76.385 53.258
188880 94.700 -76.387 53.227
188881 94.696 -76.389 53.198
188882 94.692 -76.392 53.167
188883 94.688 -76.394 53.136
188884 94.684 -76.396 53.105
188885 94.684 -76.398 53.075
188886 94.693 -76.402 53.047
188887 94.702 -76.407 53.019
188888 94.711 -76.411 52.989
188889 94.720 -76.414 52.953
188890 94.730 -76.418 52.912
188891 94.739 -76.422 52.870
188892 94.748 -76.426 52.826
188893 94.757 -76.430 52.782
188894 94.766 -76.433 52.738
188895 94.775 -76.437 52.694
188896 94.784 -76.441 52.648
188897 94.793 -76.445 52.602
188898 94.803 -76.449 52.556
188899 94.812 -76.453 52.510
188900 94.827 -76.458 52.463
188901 94.875 -76.470 52.415
188902 94.914 -76.478 52.366
188903 94.954 -76.486 52.313
188904 94.993 -76.494 52.260
188905 95.032 -76.502 52.205
188906 95.071 -76.511 52.149
188907 95.110 -76.519 52.097
188908 95.149 -76.527 52.048
188909 95.188 -76.535 51.999
188910 95.217 -76.546 51.951
188911 95.256 -76.556 51.904
188912 95.283 -76.566 51.857
188913 95.302 -76.577 51.812
188914 95.321 -76.587 51.772
188915 95.339 -76.597 51.730
188916 95.358 -76.608 51.687
188917 95.377 -76.618 51.642
188918 95.395 -76.628 51.594
188919 95.414 -76.639 51.547
188920 95.432 -76.649 51.502
188921 95.451 -76.659 51.459
188922 95.470 -76.670 51.418
188923 95.488 -76.680 51.378
188924 95.507 -76.690 51.339
188925 95.525 -76.700 51.300
188926 95.524 -76.716 51.260
188927 95.544 -76.722 51.220
188928 95.572 -76.724 51.180
188929 95.602 -76.727 51.139
188930 95.632 -76.730 51.100
188931 95.662 -76.734 51.060
188932 95.691 -76.737 51.019
188933 95.721 -76.740 50.977
188934 95.751 -76.743 50.934
188935 95.776 -76.747 50.892
188936 95.799 -76.752 50.850
188937 95.821 -76.756 50.808
188938 95.844 -76.762 50.764
188939 95.855 -76.754 50.718
188940 95.858 -76.739 50.673
188941 95.864 -76.727 50.627
188942 95.899 -76.740 50.579
188943 95.943 -76.761 50.531
188944 95.983 -76.780 50.481
188945 96.003 -76.785 50.428
188946 96.023 -76.790 50.370
188947 96.051 -76.799 50.308
188948 96.084 -76.810 50.245
188949 96.110 -76.818 50.183
188950 96.135 -76.828 50.125
188951 96.160 -76.836 50.070
188952 96.185 -76.845 50.020
188953 96.179 -76.862 49.976
188954 96.182 -76.878 49.933
188955 96.192 -76.891 49.892
188956 96.201 -76.904 49.853
188957 96.211 -76.918 49.819
188958 96.221 -76.931 49.787
188959 96.230 -76.945 49.758
188960 96.240 -76.958 49.732
188961 96.249 -76.972 49.707
188962 96.259 -76.985 49.683
188963 96.310 -76.959 49.662
188964 96.416 -76.884 49.643
188965 96.445 -76.884 49.626
188966 96.459 -76.897 49.611
188967 96.473 -76.910 49.596
188968 96.453 -76.901 49.582
188969 96.438 -76.914 49.568
188970 96.461 -76.970 49.554
188971 96.484 -77.026 49.539
188972 96.519 -77.028 49.521
188973 96.491 -77.060 49.500
188974 96.495 -77.049 49.478
188975 96.514 -77.034 49.454
188976 96.548 -77.045 49.430
188977 96.583 -77.056 49.407
188978 96.617 -77.076 49.386
188979 96.641 -77.088 49.364
188980 96.661 -77.092 49.340
188981 96.681 -77.097 49.314
188982 96.701 -77.102 49.286
188983 96.721 -77.106 49.256
188984 96.741 -77.111 49.227
188985 96.761 -77.116 49.200
188986 96.782 -77.121 49.174
188987 96.795 -77.132 49.147
188988 96.807 -77.143 49.120
188989 96.819 -77.147 49.094
188990 96.831 -77.151 49.068
188991 96.843 -77.156 49.043
188992 96.855 -77.160 49.018
188993 96.867 -77.164 48.994
188994 96.879 -77.168 48.968
188995 96.891 -77.172 48.943
188996 96.903 -77.176 48.917
188997 96.915 -77.180 48.892
188998 96.927 -77.184 48.866
188999 96.976 -77.193 48.839
189000 97.027 -77.202 48.813
189001 97.075 -77.216 48.787
189002 97.122 -77.232 48.762
189003 97.168 -77.248 48.738
189004 97.215 -77.265 48.714
189005 97.231 -77.273 48.690
189006 97.248 -77.282 48.667
189007 97.264 -77.291 48.644
189008 97.281 -77.300 48.622
189009 97.297 -77.308 48.599
189010 97.314 -77.316 48.578
189011 97.330 -77.326 48.557
189012 97.347 -77.334 48.538
189013 97.363 -77.342 48.518
189014 97.379 -77.351 48.498
189015 97.396 -77.360 48.477
189016 97.417 -77.374 48.456
189017 97.430 -77.379 48.435
189018 97.452 -77.388 48.415
189019 97.473 -77.396 48.392
189020 97.495 -77.404 48.371
189021 97.517 -77.412 48.348
189022 97.538 -77.420 48.325
189023 97.560 -77.429 48.303
189024 97.532 -77.423 48.280
189025 97.524 -77.417 48.255
189026 97.551 -77.415 48.229
189027 97.578 -77.413 48.200
189028 97.605 -77.410 48.170
189029 97.632 -77.409 48.140
189030 97.659 -77.409 48.109
189031 97.687 -77.411 48.078
189032 97.715 -77.414 48.047
189033 97.742 -77.415 48.016
189034 97.771 -77.411 47.985
189035 97.802 -77.399 47.954
189036 97.832 -77.388 47.922
189037 97.863 -77.376 47.891
189038 97.893 -77.364 47.860
189039 97.924 -77.352 47.826
189040 97.955 -77.340 47.793
189041 97.986 -77.329 47.759
189042 98.016 -77.316 47.728
189043 98.053 -77.313 47.697
189044 98.090 -77.309 47.668
189045 98.118 -77.305 47.640
189046 98.128 -77.301 47.612
189047 98.139 -77.297 47.585
189048 98.149 -77.292 47.557
189049 98.160 -77.288 47.531
189050 98.171 -77.284 47.503
189051 98.181 -77.279 47.475
189052 98.192 -77.275 47.448
189053 98.204 -77.271 47.421
189054 98.217 -77.266 47.393
189055 98.230 -77.263 47.366
189056 98.244 -77.258 47.339
189057 98.257 -77.254 47.311
189058 98.273 -77.273 47.283
189059 98.308 -77.276 47.255
189060 98.358 -77.286 47.228
189061 98.411 -77.298 47.204
189062 98.464 -77.309 47.180
189063 98.517 -77.321 47.157
189064 98.538 -77.311 47.134
189065 98.552 -77.312 47.112
189066 98.581 -77.329 47.092
189067 98.610 -77.346 47.072
189068 98.639 -77.363 47.050
189069 98.663 -77.379 47.029
189070 98.676 -77.396 47.008
189071 98.689 -77.412 46.987
189072 98.703 -77.429 46.964
189073 98.715 -77.436 46.942
189074 98.726 -77.432 46.920
189075 98.738 -77.428 46.899
189076 98.749 -77.424 46.876
189077 98.761 -77.420 46.854
189078 98.773 -77.416 46.832
189079 98.784 -77.412 46.808
189080 98.796 -77.408 46.786
189081 98.807 -77.404 46.765
189082 98.819 -77.400 46.744
189083 98.830 -77.396 46.720
189084 98.842 -77.392 46.697
189085 98.853 -77.388 46.673
189086 98.867 -77.385 46.649
189087 98.886 -77.392 46.626
189088 98.906 -77.398 46.603
189089 98.926 -77.403 46.579
189090 98.946 -77.409 46.554
189091 98.966 -77.414 46.531
189092 98.986 -77.420 46.507
189093 99.007 -77.428 46.484
189094 99.034 -77.454 46.464
189095 99.061 -77.479 46.442
189096 99.073 -77.489 46.419
189097 99.071 -77.502 46.395
189098 99.070 -77.516 46.370
189099 99.069 -77.529 46.345
189100 99.067 -77.542 46.322
189101 99.066 -77.555 46.301
189102 99.064 -77.569 46.278
189103 99.063 -77.582 46.255
189104 99.062 -77.595 46.232
189105 99.060 -77.609 46.208
189106 99.059 -77.622 46.185
189107 99.055 -77.630 46.164
189108 99.065 -77.630 46.143
189109 99.074 -77.633 46.120
189110 99.100 -77.633 46.099
189111 99.131 -77.633 46.078
189112 99.161 -77.632 46.058
189113 99.191 -77.631 46.037
189114 99.221 -77.630 46.018
189115 99.251 -77.630 45.999
189116 99.281 -77.629 45.982
189117 99.311 -77.629 45.965
189118 99.341 -77.628 45.948
189119 99.371 -77.627 45.931
189120 99.401 -77.627 45.915
189121 99.431 -77.626 45.901
189122 99.464 -77.635 45.889
189123 99.497 -77.651 45.876
189124 99.532 -77.667 45.864
189125 99.565 -77.684 45.852
189126 99.600 -77.700 45.840
189127 99.633 -77.717 45.828
189128 99.667 -77.733 45.816
189129 99.701 -77.750 45.803
189130 99.735 -77.766 45.790
189131 99.768 -77.774 45.777
189132 99.781 -77.786 45.764
189133 99.795 -77.798 45.751
189134 99.808 -77.810 45.737
189135 99.822 -77.822 45.724
189136 99.844 -77.845 45.712
189137 99.868 -77.868 45.700
189138 99.889 -77.880 45.689
189139 99.910 -77.892 45.677
189140 99.931 -77.903 45.667
189141 99.951 -77.915 45.658
189142 99.972 -77.927 45.648
189143 99.993 -77.939 45.639
189144 100.017 -77.948 45.630
189145 100.048 -77.953 45.620
189146 100.078 -77.958 45.609
189147 100.109 -77.963 45.599
189148 100.139 -77.968 45.587
189149 100.169 -77.974 45.576
189150 100.199 -77.978 45.563
189151 100.221 -77.991 45.550
189152 100.242 -78.005 45.537
189153 100.262 -78.018 45.523
189154 100.282 -78.033 45.511
189155 100.303 -78.047 45.496
189156 100.323 -78.060 45.483
189157 100.342 -78.073 45.469
189158 100.353 -78.081 45.456
189159 100.366 -78.088 45.442
189160 100.379 -78.096 45.427
189161 100.393 -78.103 45.412
189162 100.406 -78.110 45.397
189163 100.416 -78.124 45.382
189164 100.426 -78.138 45.367
189165 100.435 -78.151 45.352
189166 100.444 -78.166 45.337
189167 100.454 -78.179 45.323
189168 100.464 -78.194 45.309
189169 100.485 -78.187 45.294
189170 100.508 -78.177 45.280
189171 100.537 -78.175 45.266
189172 100.581 -78.179 45.250
189173 100.627 -78.187 45.235
189174 100.665 -78.193 45.221
189175 100.703 -78.199 45.206
189176 100.741 -78.205 45.190
189177 100.779 -78.212 45.175
189178 100.817 -78.218 45.160
189179 100.855 -78.224 45.145
189180 100.894 -78.232 45.131
189181 100.935 -78.239 45.116
189182 100.976 -78.247 45.102
189183 101.016 -78.254 45.088
189184 101.057 -78.262 45.074
189185 101.098 -78.269 45.060
189186 101.139 -78.277 45.045
189187 101.179 -78.285 45.030
189188 101.208 -78.297 45.015
189189 101.220 -78.315 45.000
189190 101.230 -78.333 44.985
189191 101.239 -78.351 44.969
189192 101.249 -78.369 44.953
189193 101.263 -78.387 44.937
189194 101.276 -78.404 44.921
189195 101.289 -78.420 44.906
189196 101.303 -78.438 44.890
189197 101.329 -78.458 44.873
189198 101.355 -78.479 44.857
189199 101.346 -78.458 44.840
189200 101.381 -78.479 44.823
189201 101.416 -78.501 44.807
189202 101.452 -78.522 44.791
189203 101.487 -78.543 44.775
189204 101.523 -78.564 44.759
189205 101.554 -78.587 44.743
189206 101.581 -78.613 44.726
189207 101.608 -78.639 44.710
189208 101.634 -78.664 44.694
189209 101.675 -78.691 44.677
189210 101.718 -78.718 44.661
189211 101.761 -78.745 44.645
189212 101.804 -78.772 44.628
189213 101.847 -78.799 44.612
189214 101.890 -78.827 44.597
189215 101.933 -78.854 44.582
189216 101.976 -78.880 44.568
189217 101.956 -78.981 44.554
189218 101.994 -79.006 44.539
189219 102.013 -79.008 44.525
189220 102.033 -79.009 44.511
189221 102.052 -79.011 44.498
189222 102.072 -79.012 44.485
189223 102.092 -79.014 44.472
189224 102.111 -79.016 44.459
189225 102.130 -79.018 44.447
189226 102.150 -79.019 44.433
189227 102.169 -79.021 44.421
189228 102.189 -79.022 44.408
189229 102.208 -79.024 44.394
189505 93.395 -76.441 54.094
189506 93.413 -76.444 54.096
189507 93.432 -76.447 54.098
189508 93.450 -76.450 54.099
189509 93.468 -76.453 54.100
189510 93.487 -76.456 54.101
189511 93.505 -76.458 54.103
189512 93.523 -76.461 54.106
189513 93.542 -76.464 54.108
189514 93.560 -76.467 54.111
189515 93.578 -76.470 54.113
189516 93.596 -76.473 54.115
189517 93.615 -76.476 54.118
189518 93.633 -76.479 54.121
189519 93.652 -76.482 54.126
189520 93.670 -76.485 54.132
189521 93.689 -76.486 54.140
189522 93.710 -76.477 54.145
189523 93.731 -76.467 54.153
189524 93.751 -76.458 54.163
189525 93.766 -76.459 54.175
189526 93.780 -76.456 54.188
189527 93.795 -76.452 54.201
189528 93.810 -76.448 54.212
189529 93.825 -76.445 54.219
189530 93.839 -76.441 54.226
189531 93.854 -76.436 54.231
189532 93.870 -76.420 54.236
189533 93.886 -76.413 54.238
189534 93.901 -76.408 54.239
189535 93.916 -76.404 54.236
189536 93.931 -76.399 54.229
189537 93.948 -76.400 54.222
189538 93.965 -76.403 54.216
189539 93.982 -76.406 54.210
189540 93.996 -76.410 54.205
189541 94.011 -76.414 54.204
189542 94.027 -76.414 54.201
189543 94.045 -76.410 54.197
189544 94.064 -76.405 54.192
189545 94.082 -76.401 54.191
189546 94.100 -76.397 54.193
189547 94.119 -76.393 54.197
189548 94.137 -76.389 54.200
189549 94.155 -76.384 54.196
189550 94.175 -76.381 54.189
189551 94.195 -76.379 54.180
189552 94.214 -76.377 54.167
189553 94.234 -76.376 54.152
189554 94.254 -76.374 54.136
189555 94.273 -76.373 54.118
189556 94.293 -76.371 54.092
189557 94.313 -76.370 54.064
189558 94.333 -76.368 54.035
189559 94.352 -76.366 54.004
189560 94.372 -76.364 53.973
189561 94.392 -76.363 53.943
189562 94.412 -76.361 53.913
189563 94.431 -76.360 53.878
189564 94.449 -76.358 53.840
189565 94.468 -76.357 53.800
189566 94.486 -76.356 53.761
189567 94.505 -76.355 53.724
189568 94.525 -76.353 53.689
189569 94.545 -76.351 53.656
189570 94.566 -76.348 53.622
189571 94.586 -76.345 53.589
189572 94.603 -76.348 53.556
189573 94.616 -76.354 53.525
189574 94.628 -76.361 53.494
189575 94.641 -76.368 53.462
189576 94.653 -76.375 53.427
189577 94.666 -76.382 53.389
189578 94.682 -76.380 53.349
189579 94.699 -76.379 53.313
189580 94.695 -76.380 53.278
189581 94.691 -76.383 53.246
189582 94.687 -76.385 53.216
189583 94.683 -76.387 53.184
189584 94.679 -76.389 53.151
189585 94.684 -76.393 53.120
189586 94.693 -76.397 53.088
189587 94.702 -76.401 53.057
189588 94.712 -76.404 53.027
189589 94.721 -76.408 52.997
189590 94.730 -76.412 52.960
189591 94.739 -76.416 52.921
189592 94.748 -76.420 52.880
189593 94.757 -76.423 52.837
189594 94.766 -76.428 52.794
189595 94.775 -76.432 52.752
189596 94.784 -76.436 52.711
189597 94.793 -76.439 52.665
189598 94.803 -76.443 52.619
189599 94.812 -76.447 52.572
189600 94.821 -76.451 52.525
189601 94.830 -76.455 52.479
189602 94.859 -76.462 52.432
189603 94.899 -76.470 52.383
189604 94.938 -76.478 52.330
189605 94.977 -76.486 52.276
189606 95.016 -76.495 52.221
189607 95.057 -76.503 52.164
189608 95.099 -76.511 52.111
189609 95.140 -76.519 52.060
189610 95.181 -76.527 52.011
189611 95.216 -76.533 51.964
189612 95.237 -76.539 51.915
189613 95.256 -76.550 51.868
189614 95.275 -76.560 51.825
189615 95.294 -76.571 51.786
189616 95.313 -76.580 51.746
189617 95.332 -76.591 51.703
189618 95.351 -76.601 51.659
189619 95.371 -76.611 51.612
189620 95.390 -76.621 51.565
189621 95.408 -76.631 51.519
189622 95.427 -76.642 51.474
189623 95.447 -76.652 51.432
189624 95.466 -76.662 51.392
189625 95.485 -76.672 51.353
189626 95.503 -76.683 51.313
189627 95.523 -76.693 51.274
189628 95.552 -76.696 51.235
189629 95.582 -76.699 51.195
189630 95.612 -76.702 51.155
189631 95.641 -76.705 51.115
189632 95.671 -76.708 51.075
189633 95.701 -76.712 51.034
189634 95.731 -76.715 50.992
189635 95.758 -76.718 50.949
189636 95.781 -76.722 50.906
189637 95.804 -76.727 50.864
189638 95.827 -76.732 50.822
189639 95.849 -76.737 50.780
189640 95.872 -76.741 50.738
189641 95.875 -76.726 50.697
189642 95.878 -76.712 50.653
189643 95.918 -76.729 50.606
189644 95.962 -76.749 50.556
189645 95.986 -76.758 50.506
189646 96.006 -76.763 50.453
189647 96.026 -76.768 50.396
189648 96.045 -76.774 50.337
189649 96.077 -76.784 50.275
189650 96.110 -76.795 50.214
189651 96.142 -76.805 50.155
189652 96.167 -76.814 50.101
189653 96.187 -76.824 50.051
189654 96.187 -76.840 50.006
189655 96.197 -76.853 49.964
189656 96.206 -76.866 49.924
189657 96.216 -76.880 49.886
189658 96.226 -76.893 49.852
189659 96.235 -76.907 49.821
189660 96.245 -76.920 49.792
189661 96.254 -76.934 49.766
189662 96.264 -76.947 49.741
189663 96.273 -76.961 49.718
189664 96.301 -76.958 49.696
189665 96.406 -76.882 49.679
189666 96.476 -76.841 49.662
189667 96.443 -76.823 49.646
189668 96.401 -76.807 49.631
189669 96.425 -76.863 49.616
189670 96.448 -76.919 49.603
189671 96.472 -76.975 49.589
189672 96.495 -77.032 49.572
189673 96.506 -77.046 49.552
189674 96.505 -77.046 49.530
189675 96.515 -77.019 49.506
189676 96.525 -76.991 49.482
189677 96.553 -76.992 49.458
189678 96.588 -77.003 49.435
189679 96.622 -77.027 49.414
189680 96.655 -77.056 49.391
189681 96.678 -77.064 49.366
189682 96.698 -77.069 49.338
189683 96.718 -77.074 49.308
189684 96.738 -77.078 49.278
189685 96.758 -77.084 49.250
189686 96.779 -77.088 49.222
189687 96.797 -77.094 49.194
189688 96.810 -77.106 49.167
189689 96.822 -77.117 49.139
189690 96.834 -77.128 49.111
189691 96.847 -77.139 49.085
189692 96.859 -77.147 49.060
189693 96.871 -77.152 49.036
189694 96.883 -77.156 49.011
189695 96.895 -77.160 48.985
189696 96.907 -77.164 48.960
189697 96.919 -77.168 48.935
189698 96.931 -77.172 48.910
189699 96.943 -77.176 48.883
189700 96.979 -77.183 48.855
189701 97.030 -77.192 48.829
189702 97.078 -77.207 48.802
189703 97.124 -77.224 48.777
189704 97.171 -77.240 48.753
189705 97.218 -77.257 48.728
189706 97.249 -77.269 48.705
189707 97.265 -77.278 48.681
189708 97.282 -77.286 48.657
189709 97.299 -77.295 48.635
189710 97.315 -77.304 48.613
189711 97.331 -77.312 48.591
189712 97.348 -77.321 48.571
189713 97.365 -77.329 48.551
189714 97.381 -77.338 48.531
189715 97.397 -77.347 48.511
189716 97.414 -77.355 48.490
189717 97.425 -77.351 48.469
189718 97.461 -77.365 48.447
189719 97.483 -77.373 48.426
189720 97.504 -77.382 48.403
189721 97.526 -77.390 48.380
189722 97.547 -77.398 48.357
189723 97.569 -77.406 48.334
189724 97.554 -77.404 48.311
189725 97.516 -77.395 48.288
189726 97.539 -77.392 48.262
189727 97.566 -77.390 48.234
189728 97.593 -77.388 48.205
189729 97.620 -77.386 48.175
189730 97.647 -77.384 48.145
189731 97.674 -77.382 48.115
189732 97.701 -77.384 48.084
189733 97.729 -77.386 48.052
189734 97.756 -77.388 48.021
189735 97.785 -77.385 47.991
189736 97.816 -77.373 47.959
189737 97.846 -77.361 47.927
189738 97.877 -77.350 47.895
189739 97.908 -77.338 47.863
189740 97.938 -77.326 47.830
189741 97.969 -77.314 47.798
189742 98.000 -77.303 47.765
189743 98.030 -77.291 47.734
189744 98.061 -77.279 47.704
189745 98.092 -77.267 47.674
189746 98.121 -77.263 47.646
189747 98.132 -77.258 47.617
189748 98.143 -77.254 47.589
189749 98.156 -77.250 47.562
189750 98.169 -77.246 47.534
189751 98.182 -77.241 47.507
189752 98.196 -77.237 47.481
189753 98.209 -77.233 47.454
189754 98.222 -77.229 47.426
189755 98.235 -77.225 47.399
189756 98.248 -77.220 47.371
189757 98.262 -77.216 47.344
189758 98.277 -77.235 47.316
189759 98.295 -77.254 47.288
189760 98.331 -77.257 47.260
189761 98.366 -77.261 47.234
189762 98.414 -77.270 47.210
189763 98.467 -77.282 47.187
189764 98.520 -77.294 47.163
189765 98.529 -77.284 47.139
189766 98.558 -77.301 47.117
189767 98.587 -77.318 47.096
189768 98.616 -77.335 47.076
189769 98.645 -77.351 47.055
189770 98.674 -77.368 47.034
189771 98.694 -77.385 47.014
189772 98.707 -77.401 46.992
189773 98.721 -77.418 46.970
189774 98.732 -77.418 46.948
189775 98.744 -77.414 46.927
189776 98.756 -77.410 46.905
189777 98.767 -77.406 46.883
189778 98.779 -77.402 46.861
189779 98.790 -77.398 46.839
189780 98.802 -77.394 46.816
189781 98.813 -77.390 46.795
189782 98.825 -77.386 46.774
189783 98.836 -77.382 46.752
189784 98.847 -77.382 46.728
189785 98.858 -77.382 46.704
189786 98.869 -77.382 46.680
189787 98.880 -77.382 46.656
189788 98.902 -77.373 46.634
189789 98.922 -77.374 46.611
189790 98.942 -77.380 46.586
189791 98.962 -77.385 46.561
189792 98.982 -77.392 46.536
189793 99.002 -77.398 46.512
189794 99.025 -77.411 46.490
189795 99.053 -77.437 46.468
189796 99.077 -77.463 46.446
189797 99.080 -77.480 46.424
189798 99.078 -77.493 46.401
189799 99.077 -77.507 46.376
189800 99.075 -77.520 46.351
189801 99.073 -77.533 46.327
189802 99.071 -77.547 46.305
189803 99.069 -77.560 46.282
189804 99.068 -77.573 46.259
189805 99.066 -77.586 46.237
189806 99.065 -77.600 46.214
189807 99.063 -77.613 46.190
189808 99.062 -77.627 46.168
189809 99.079 -77.628 46.147
189810 99.109 -77.627 46.125
189811 99.139 -77.627 46.103
189812 99.169 -77.627 46.082
189813 99.199 -77.626 46.061
189814 99.229 -77.625 46.040
189815 99.260 -77.624 46.020
189816 99.289 -77.624 46.002
189817 99.320 -77.623 45.984
189818 99.350 -77.622 45.966
189819 99.379 -77.622 45.949
189820 99.410 -77.621 45.933
189821 99.440 -77.620 45.918
189822 99.470 -77.620 45.904
189823 99.500 -77.619 45.890
189824 99.534 -77.634 45.877
189825 99.567 -77.651 45.865
189826 99.602 -77.667 45.853
189827 99.635 -77.683 45.841
189828 99.669 -77.700 45.828
189829 99.703 -77.716 45.815
189830 99.737 -77.733 45.803
189831 99.774 -77.743 45.790
189832 99.788 -77.755 45.776
189833 99.801 -77.767 45.762
189834 99.815 -77.779 45.749
189835 99.828 -77.791 45.736
189836 99.842 -77.803 45.722
189837 99.859 -77.821 45.709
189838 99.885 -77.846 45.698
189839 99.908 -77.866 45.686
189840 99.928 -77.878 45.674
189841 99.949 -77.890 45.664
189842 99.970 -77.902 45.655
189843 99.991 -77.913 45.645
189844 100.012 -77.925 45.636
189845 100.041 -77.931 45.627
189846 100.071 -77.936 45.618
189847 100.102 -77.941 45.608
189848 100.132 -77.946 45.597
189849 100.162 -77.952 45.585
189850 100.188 -77.961 45.574
189851 100.208 -77.975 45.561
189852 100.228 -77.988 45.548
189853 100.249 -78.003 45.535
189854 100.269 -78.016 45.522
189855 100.289 -78.030 45.508
189856 100.310 -78.044 45.495
189857 100.331 -78.058 45.481
189858 100.350 -78.072 45.468
189859 100.361 -78.078 45.454
189860 100.374 -78.086 45.440
189861 100.388 -78.094 45.426
189862 100.401 -78.101 45.411
189863 100.414 -78.109 45.395
189864 100.428 -78.116 45.380
189865 100.439 -78.127 45.365
189866 100.449 -78.141 45.351
189867 100.458 -78.154 45.336
189868 100.468 -78.169 45.322
189869 100.477 -78.182 45.308
189870 100.498 -78.175 45.293
189871 100.522 -78.166 45.279
189872 100.546 -78.159 45.264
189873 100.587 -78.165 45.249
189874 100.626 -78.172 45.234
189875 100.666 -78.179 45.220
189876 100.707 -78.187 45.206
189877 100.748 -78.194 45.190
189878 100.788 -78.202 45.175
189879 100.829 -78.210 45.159
189880 100.870 -78.217 45.144
189881 100.911 -78.225 45.130
189882 100.952 -78.232 45.115
189883 100.992 -78.240 45.101
189884 101.033 -78.247 45.087
189885 101.073 -78.255 45.074
189886 101.114 -78.263 45.060
189887 101.155 -78.270 45.045
189888 101.196 -78.278 45.031
189889 101.218 -78.292 45.016
189890 101.230 -78.310 45.001
189891 101.241 -78.329 44.986
189892 101.250 -78.347 44.970
189893 101.260 -78.366 44.955
189894 101.269 -78.384 44.939
189895 101.279 -78.402 44.923
189896 101.289 -78.420 44.908
189897 101.281 -78.409 44.892
189898 101.292 -78.404 44.875
189899 101.327 -78.425 44.859
189900 101.362 -78.446 44.843
189901 101.398 -78.467 44.826
189902 101.433 -78.488 44.810
189903 101.468 -78.509 44.793
189904 101.503 -78.530 44.777
189905 101.539 -78.551 44.761
189906 101.566 -78.576 44.744
189907 101.594 -78.602 44.727
189908 101.620 -78.628 44.711
189909 101.647 -78.654 44.694
189910 101.674 -78.680 44.678
189911 101.716 -78.706 44.662
189912 101.759 -78.733 44.645
189913 101.802 -78.761 44.629
189914 101.845 -78.788 44.614
189915 101.888 -78.814 44.599
189916 101.889 -78.891 44.584
189917 101.932 -78.923 44.569
189918 101.975 -78.955 44.556
189919 102.018 -78.987 44.541
189920 102.039 -78.990 44.527
189921 102.058 -78.992 44.513
189922 102.078 -78.993 44.500
189923 102.097 -78.995 44.486
189924 102.117 -78.996 44.473
189925 102.136 -78.999 44.460
189926 102.156 -79.000 44.447
189927 102.175 -79.002 44.435
189928 102.194 -79.003 44.422
189929 102.214 -79.005 44.408
189930 102.233 -79.006 44.395
190209 93.442 -76.426 54.091
190210 93.460 -76.430 54.093
190211 93.479 -76.433 54.093
190212 93.497 -76.435 54.096
190213 93.515 -76.438 54.099
190214 93.534 -76.441 54.102
190215 93.552 -76.444 54.106
190216 93.570 -76.447 54.108
190217 93.588 -76.450 54.110
190218 93.607 -76.453 54.113
190219 93.625 -76.455 54.116
190220 93.644 -76.458 54.120
190221 93.662 -76.461 54.126
190222 93.680 -76.464 54.134
190223 93.699 -76.467 54.139
190224 93.720 -76.458 54.146
190225 93.741 -76.448 54.155
190226 93.759 -76.443 54.168
190227 93.773 -76.442 54.182
190228 93.788 -76.438 54.195
190229 93.802 -76.435 54.207
190230 93.818 -76.431 54.213
190231 93.832 -76.427 54.220
190232 93.848 -76.415 54.225
190233 93.864 -76.399 54.229
190234 93.880 -76.391 54.231
190235 93.895 -76.386 54.233
190236 93.910 -76.382 54.233
190237 93.925 -76.378 54.227
190238 93.944 -76.384 54.219
190239 93.962 -76.389 54.213
190240 93.979 -76.391 54.208
190241 93.995 -76.394 54.206
190242 94.012 -76.397 54.207
190243 94.030 -76.397 54.205
190244 94.047 -76.393 54.200
190245 94.065 -76.389 54.196
190246 94.082 -76.385 54.195
190247 94.101 -76.381 54.197
190248 94.118 -76.377 54.201
190249 94.136 -76.373 54.204
190250 94.154 -76.370 54.201
190251 94.175 -76.367 54.195
190252 94.196 -76.364 54.188
190253 94.215 -76.361 54.175
190254 94.235 -76.360 54.161
190255 94.255 -76.358 54.145
190256 94.275 -76.357 54.127
190257 94.294 -76.355 54.100
190258 94.314 -76.354 54.072
190259 94.334 -76.352 54.042
190260 94.353 -76.350 54.012
190261 94.373 -76.348 53.982
190262 94.393 -76.347 53.953
190263 94.413 -76.345 53.924
190264 94.431 -76.344 53.890
190265 94.449 -76.342 53.853
190266 94.468 -76.341 53.814
190267 94.486 -76.340 53.776
190268 94.505 -76.339 53.741
190269 94.523 -76.338 53.707
190270 94.542 -76.337 53.675
190271 94.560 -76.336 53.642
190272 94.580 -76.334 53.609
190273 94.595 -76.338 53.578
190274 94.608 -76.345 53.547
190275 94.620 -76.351 53.515
190276 94.633 -76.359 53.482
190277 94.645 -76.366 53.447
190278 94.659 -76.373 53.409
190279 94.672 -76.376 53.370
190280 94.689 -76.374 53.333
190281 94.686 -76.376 53.299
190282 94.682 -76.379 53.265
190283 94.679 -76.381 53.234
190284 94.676 -76.383 53.201
190285 94.685 -76.387 53.167
190286 94.694 -76.391 53.135
190287 94.703 -76.395 53.102
190288 94.712 -76.398 53.069
190289 94.721 -76.402 53.037
190290 94.730 -76.407 53.005
190291 94.740 -76.411 52.968
190292 94.749 -76.414 52.930
190293 94.758 -76.418 52.890
190294 94.767 -76.422 52.848
190295 94.776 -76.426 52.807
190296 94.785 -76.430 52.766
190297 94.794 -76.433 52.725
190298 94.804 -76.437 52.679
190299 94.815 -76.440 52.632
190300 94.827 -76.444 52.585
190301 94.838 -76.447 52.537
190302 94.849 -76.450 52.492
190303 94.868 -76.455 52.444
190304 94.910 -76.463 52.395
190305 94.951 -76.471 52.343
190306 94.992 -76.480 52.289
190307 95.034 -76.488 52.233
190308 95.074 -76.495 52.177
190309 95.116 -76.504 52.122
190310 95.152 -76.511 52.071
190311 95.177 -76.518 52.021
190312 95.214 -76.519 51.973
190313 95.243 -76.521 51.925
190314 95.266 -76.526 51.879
190315 95.288 -76.532 51.836
190316 95.308 -76.542 51.796
190317 95.327 -76.552 51.756
190318 95.346 -76.562 51.716
190319 95.365 -76.572 51.675
190320 95.384 -76.583 51.631
190321 95.403 -76.592 51.584
190322 95.422 -76.603 51.537
190323 95.441 -76.613 51.491
190324 95.460 -76.624 51.448
190325 95.479 -76.633 51.408
190326 95.498 -76.644 51.369
190327 95.517 -76.654 51.330
190328 95.536 -76.665 51.291
190329 95.561 -76.670 51.252
190330 95.591 -76.673 51.213
190331 95.621 -76.676 51.174
190332 95.651 -76.680 51.135
190333 95.680 -76.683 51.095
190334 95.710 -76.686 51.054
190335 95.740 -76.689 51.011
190336 95.764 -76.693 50.969
190337 95.787 -76.698 50.926
190338 95.809 -76.702 50.884
190339 95.832 -76.707 50.844
190340 95.855 -76.712 50.803
190341 95.878 -76.716 50.763
190342 95.893 -76.713 50.723
190343 95.896 -76.699 50.678
190344 95.938 -76.718 50.631
190345 95.969 -76.730 50.582
190346 95.988 -76.736 50.532
190347 96.008 -76.741 50.479
190348 96.028 -76.747 50.424
190349 96.048 -76.752 50.366
190350 96.070 -76.759 50.307
190351 96.103 -76.769 50.245
190352 96.136 -76.780 50.185
190353 96.169 -76.790 50.130
190354 96.192 -76.802 50.081
190355 96.202 -76.815 50.036
190356 96.212 -76.828 49.995
190357 96.221 -76.842 49.955
190358 96.231 -76.856 49.918
190359 96.240 -76.869 49.885
190360 96.250 -76.882 49.854
190361 96.259 -76.896 49.825
190362 96.269 -76.909 49.800
190363 96.278 -76.923 49.776
190364 96.289 -76.936 49.754
190365 96.298 -76.950 49.733
190366 96.345 -76.856 49.715
190367 96.369 -76.746 49.698
190368 96.392 -76.758 49.682
190369 96.412 -76.813 49.668
190370 96.436 -76.869 49.654
190371 96.459 -76.925 49.640
190372 96.482 -76.981 49.624
190373 96.506 -77.038 49.605
190374 96.514 -77.043 49.584
190375 96.525 -77.016 49.561
190376 96.535 -76.988 49.537
190377 96.545 -76.961 49.511
190378 96.558 -76.940 49.487
190379 96.593 -76.950 49.463
190380 96.627 -76.978 49.441
190381 96.660 -77.006 49.416
190382 96.693 -77.034 49.390
190383 96.715 -77.041 49.362
190384 96.735 -77.046 49.331
190385 96.755 -77.051 49.301
190386 96.775 -77.056 49.272
190387 96.795 -77.060 49.242
190388 96.812 -77.068 49.214
190389 96.825 -77.079 49.187
190390 96.837 -77.091 49.159
190391 96.850 -77.102 49.131
190392 96.862 -77.113 49.104
190393 96.874 -77.124 49.079
190394 96.887 -77.135 49.054
190395 96.899 -77.147 49.029
190396 96.911 -77.152 49.003
190397 96.924 -77.156 48.978
190398 96.935 -77.160 48.952
190399 96.947 -77.164 48.926
190400 96.959 -77.168 48.899
190401 96.983 -77.173 48.872
190402 97.034 -77.183 48.845
190403 97.080 -77.199 48.818
190404 97.127 -77.216 48.793
190405 97.174 -77.232 48.768
190406 97.221 -77.248 48.744
190407 97.267 -77.265 48.720
190408 97.284 -77.273 48.696
190409 97.300 -77.282 48.673
190410 97.316 -77.291 48.650
190411 97.333 -77.299 48.628
190412 97.350 -77.308 48.607
190413 97.366 -77.316 48.587
190414 97.382 -77.325 48.567
190415 97.399 -77.334 48.547
190416 97.413 -77.335 48.525
190417 97.423 -77.331 48.504
190418 97.434 -77.327 48.483
190419 97.492 -77.351 48.461
190420 97.513 -77.359 48.438
190421 97.535 -77.367 48.414
190422 97.557 -77.376 48.390
190423 97.578 -77.383 48.366
190424 97.575 -77.385 48.341
190425 97.538 -77.376 48.318
190426 97.527 -77.369 48.293
190427 97.554 -77.367 48.266
190428 97.581 -77.365 48.238
190429 97.607 -77.363 48.210
190430 97.635 -77.361 48.180
190431 97.662 -77.360 48.150
190432 97.688 -77.357 48.120
190433 97.716 -77.357 48.089
190434 97.743 -77.359 48.058
190435 97.771 -77.360 48.028
190436 97.799 -77.359 47.997
190437 97.830 -77.348 47.965
190438 97.860 -77.335 47.932
190439 97.891 -77.324 47.900
190440 97.922 -77.312 47.867
190441 97.952 -77.301 47.835
190442 97.983 -77.288 47.803
190443 98.014 -77.277 47.771
190444 98.044 -77.265 47.740
190445 98.075 -77.253 47.710
190446 98.104 -77.240 47.681
190447 98.129 -77.226 47.652
190448 98.148 -77.216 47.623
190449 98.161 -77.212 47.595
190450 98.174 -77.208 47.567
190451 98.187 -77.204 47.539
190452 98.201 -77.200 47.512
190453 98.214 -77.195 47.485
190454 98.227 -77.191 47.459
190455 98.240 -77.187 47.432
190456 98.253 -77.182 47.404
190457 98.267 -77.179 47.377
190458 98.282 -77.197 47.349
190459 98.297 -77.219 47.321
190460 98.317 -77.235 47.293
190461 98.353 -77.239 47.265
190462 98.389 -77.243 47.240
190463 98.425 -77.247 47.216
190464 98.471 -77.255 47.194
190465 98.507 -77.256 47.171
190466 98.535 -77.273 47.146
190467 98.564 -77.289 47.123
190468 98.593 -77.307 47.101
190469 98.622 -77.323 47.080
190470 98.651 -77.340 47.060
190471 98.680 -77.357 47.040
190472 98.709 -77.374 47.019
190473 98.725 -77.391 46.997
190474 98.738 -77.404 46.976
190475 98.750 -77.400 46.956
190476 98.762 -77.396 46.935
190477 98.773 -77.392 46.913
190478 98.784 -77.388 46.890
190479 98.796 -77.384 46.868
190480 98.807 -77.381 46.846
190481 98.818 -77.381 46.824
190482 98.829 -77.382 46.803
190483 98.840 -77.382 46.782
190484 98.851 -77.382 46.760
190485 98.862 -77.382 46.735
190486 98.872 -77.382 46.710
190487 98.883 -77.383 46.686
190488 98.895 -77.382 46.661
190489 98.916 -77.374 46.639
190490 98.937 -77.366 46.615
190491 98.958 -77.357 46.591
190492 98.979 -77.363 46.566
190493 98.999 -77.369 46.542
190494 99.018 -77.374 46.519
190495 99.044 -77.394 46.495
190496 99.066 -77.420 46.473
190497 99.080 -77.446 46.451
190498 99.094 -77.472 46.428
190499 99.092 -77.485 46.405
190500 99.090 -77.498 46.380
190501 99.088 -77.511 46.356
190502 99.087 -77.525 46.332
190503 99.085 -77.538 46.310
190504 99.083 -77.551 46.287
190505 99.081 -77.565 46.264
190506 99.079 -77.578 46.241
190507 99.078 -77.592 46.218
190508 99.076 -77.605 46.195
190509 99.076 -77.617 46.173
190510 99.099 -77.614 46.151
190511 99.139 -77.617 46.128
190512 99.178 -77.620 46.107
190513 99.208 -77.620 46.085
190514 99.238 -77.620 46.064
190515 99.268 -77.619 46.043
190516 99.298 -77.618 46.023
190517 99.328 -77.617 46.005
190518 99.358 -77.617 45.986
190519 99.388 -77.616 45.969
190520 99.418 -77.616 45.951
190521 99.448 -77.615 45.935
190522 99.478 -77.614 45.921
190523 99.508 -77.614 45.906
190524 99.539 -77.613 45.892
190525 99.569 -77.617 45.879
190526 99.604 -77.633 45.866
190527 99.637 -77.650 45.854
190528 99.672 -77.667 45.842
190529 99.705 -77.683 45.829
190530 99.739 -77.699 45.815
190531 99.776 -77.713 45.802
190532 99.794 -77.724 45.789
190533 99.807 -77.736 45.775
190534 99.821 -77.748 45.761
190535 99.834 -77.760 45.748
190536 99.848 -77.772 45.735
190537 99.861 -77.784 45.722
190538 99.875 -77.796 45.708
190539 99.900 -77.823 45.695
190540 99.925 -77.849 45.684
190541 99.947 -77.865 45.673
190542 99.968 -77.876 45.663
190543 99.989 -77.888 45.653
190544 100.009 -77.899 45.644
190545 100.034 -77.909 45.635
190546 100.064 -77.914 45.625
190547 100.094 -77.919 45.616
190548 100.125 -77.924 45.607
190549 100.154 -77.931 45.596
190550 100.174 -77.944 45.584
190551 100.195 -77.958 45.572
190552 100.215 -77.972 45.559
190553 100.235 -77.986 45.547
190554 100.256 -78.000 45.534
190555 100.276 -78.013 45.521
190556 100.297 -78.028 45.507
190557 100.317 -78.041 45.494
190558 100.338 -78.055 45.480
190559 100.358 -78.069 45.466
190560 100.369 -78.076 45.453
190561 100.383 -78.084 45.438
190562 100.396 -78.091 45.424
190563 100.409 -78.099 45.410
190564 100.423 -78.106 45.394
190565 100.436 -78.114 45.378
190566 100.449 -78.122 45.363
190567 100.462 -78.130 45.350
190568 100.472 -78.144 45.335
190569 100.481 -78.157 45.321
190570 100.490 -78.172 45.307
190571 100.512 -78.165 45.291
190572 100.535 -78.155 45.277
190573 100.567 -78.147 45.263
190574 100.602 -78.155 45.248
190575 100.642 -78.165 45.234
190576 100.683 -78.172 45.220
190577 100.723 -78.180 45.206
190578 100.764 -78.188 45.191
190579 100.805 -78.195 45.175
190580 100.845 -78.203 45.160
190581 100.886 -78.210 45.145
190582 100.927 -78.218 45.130
190583 100.968 -78.225 45.115
190584 101.008 -78.233 45.101
190585 101.049 -78.241 45.087
190586 101.090 -78.249 45.074
190587 101.131 -78.257 45.060
190588 101.172 -78.264 45.046
190589 101.212 -78.272 45.032
190590 101.228 -78.288 45.016
190591 101.239 -78.307 45.002
190592 101.251 -78.325 44.987
190593 101.261 -78.343 44.971
190594 101.271 -78.361 44.956
190595 101.280 -78.379 44.940
190596 101.290 -78.398 44.925
190597 101.255 -78.366 44.910
190598 101.273 -78.369 44.894
190599 101.308 -78.391 44.878
190600 101.343 -78.412 44.861
190601 101.379 -78.433 44.845
190602 101.414 -78.454 44.828
190603 101.450 -78.476 44.811
190604 101.485 -78.497 44.795
190605 101.520 -78.518 44.777
190606 101.552 -78.541 44.761
190607 101.579 -78.567 44.745
190608 101.606 -78.592 44.728
190609 101.633 -78.617 44.712
190610 101.660 -78.643 44.696
190611 101.687 -78.669 44.679
190612 101.714 -78.695 44.663
190613 101.757 -78.722 44.647
190614 101.800 -78.749 44.631
190615 101.821 -78.802 44.616
190616 101.864 -78.833 44.601
190617 101.907 -78.865 44.586
190618 101.950 -78.897 44.571
190619 101.993 -78.929 44.557
190620 102.036 -78.961 44.543
190621 102.064 -78.973 44.529
190622 102.084 -78.974 44.515
190623 102.103 -78.976 44.501
190624 102.122 -78.977 44.487
190625 102.142 -78.979 44.474
190626 102.161 -78.981 44.461
190627 102.181 -78.983 44.448
190628 102.200 -78.984 44.436
190629 102.220 -78.986 44.422
190630 102.239 -78.987 44.408
190631 102.258 -78.989 44.395
190632 102.278 -78.990 44.382
190913 93.489 -76.412 54.088
190914 93.507 -76.414 54.091
190915 93.525 -76.417 54.095
190916 93.544 -76.420 54.099
190917 93.562 -76.423 54.102
190918 93.580 -76.426 54.104
190919 93.599 -76.430 54.108
190920 93.617 -76.432 54.111
190921 93.635 -76.435 54.116
190922 93.654 -76.438 54.122
190923 93.672 -76.441 54.129
190924 93.691 -76.444 54.134
190925 93.709 -76.447 54.140
190926 93.730 -76.439 54.149
190927 93.751 -76.429 54.161
190928 93.766 -76.428 54.176
190929 93.781 -76.424 54.190
190930 93.795 -76.420 54.202
190931 93.810 -76.417 54.209
190932 93.825 -76.410 54.214
190933 93.841 -76.395 54.219
190934 93.858 -76.379 54.222
190935 93.873 -76.370 54.224
190936 93.888 -76.365 54.227
190937 93.903 -76.361 54.228
190938 93.921 -76.362 54.222
190939 93.939 -76.367 54.215
190940 93.959 -76.373 54.209
190941 93.976 -76.376 54.205
190942 93.992 -76.379 54.206
190943 94.009 -76.382 54.209
190944 94.027 -76.382 54.208
190945 94.044 -76.378 54.203
190946 94.062 -76.373 54.199
190947 94.079 -76.370 54.198
190948 94.098 -76.366 54.199
190949 94.115 -76.362 54.202
190950 94.134 -76.359 54.204
190951 94.153 -76.358 54.202
190952 94.174 -76.357 54.197
190953 94.196 -76.348 54.191
190954 94.216 -76.346 54.179
190955 94.237 -76.344 54.166
190956 94.256 -76.342 54.151
190957 94.276 -76.341 54.133
190958 94.295 -76.339 54.106
190959 94.315 -76.338 54.078
190960 94.335 -76.336 54.048
190961 94.355 -76.335 54.019
190962 94.374 -76.332 53.990
190963 94.394 -76.331 53.961
190964 94.413 -76.329 53.932
190965 94.431 -76.328 53.898
190966 94.449 -76.326 53.861
190967 94.468 -76.325 53.825
190968 94.486 -76.324 53.789
190969 94.505 -76.323 53.755
190970 94.523 -76.322 53.723
190971 94.542 -76.321 53.691
190972 94.560 -76.320 53.658
190973 94.576 -76.320 53.627
190974 94.589 -76.326 53.596
190975 94.600 -76.334 53.565
190976 94.613 -76.342 53.532
190977 94.625 -76.349 53.499
190978 94.638 -76.356 53.464
190979 94.650 -76.363 53.426
190980 94.663 -76.370 53.388
190981 94.678 -76.370 53.351
190982 94.678 -76.372 53.316
190983 94.674 -76.374 53.283
190984 94.676 -76.377 53.250
190985 94.685 -76.381 53.217
190986 94.694 -76.385 53.184
190987 94.703 -76.389 53.150
190988 94.712 -76.393 53.117
190989 94.721 -76.397 53.083
190990 94.731 -76.401 53.049
190991 94.740 -76.404 53.016
190992 94.749 -76.408 52.978
190993 94.760 -76.411 52.939
190994 94.772 -76.415 52.900
190995 94.783 -76.418 52.860
190996 94.794 -76.421 52.820
190997 94.806 -76.425 52.780
190998 94.817 -76.428 52.738
190999 94.829 -76.431 52.691
191000 94.840 -76.435 52.643
191001 94.851 -76.438 52.594
191002 94.863 -76.441 52.547
191003 94.875 -76.445 52.501
191004 94.886 -76.448 52.453
191005 94.927 -76.456 52.404
191006 94.968 -76.464 52.352
191007 95.010 -76.472 52.298
191008 95.051 -76.480 52.243
191009 95.092 -76.489 52.187
191010 95.117 -76.495 52.133
191011 95.142 -76.502 52.081
191012 95.167 -76.508 52.031
191013 95.210 -76.506 51.982
191014 95.247 -76.505 51.935
191015 95.270 -76.511 51.890
191016 95.293 -76.515 51.847
191017 95.316 -76.520 51.806
191018 95.339 -76.524 51.767
191019 95.359 -76.534 51.729
191020 95.378 -76.544 51.690
191021 95.397 -76.554 51.648
191022 95.416 -76.564 51.602
191023 95.435 -76.574 51.555
191024 95.454 -76.585 51.508
191025 95.473 -76.595 51.465
191026 95.492 -76.605 51.425
191027 95.512 -76.615 51.386
191028 95.531 -76.626 51.347
191029 95.549 -76.636 51.309
191030 95.571 -76.644 51.271
191031 95.601 -76.647 51.233
191032 95.630 -76.650 51.195
191033 95.660 -76.654 51.156
191034 95.690 -76.657 51.116
191035 95.720 -76.660 51.075
191036 95.746 -76.664 51.033
191037 95.769 -76.668 50.991
191038 95.792 -76.673 50.948
191039 95.815 -76.677 50.905
191040 95.837 -76.682 50.866
191041 95.860 -76.687 50.826
191042 95.883 -76.692 50.787
191043 95.906 -76.696 50.747
191044 95.913 -76.686 50.703
191045 95.952 -76.703 50.657
191046 95.971 -76.708 50.608
191047 95.991 -76.714 50.559
191048 96.011 -76.719 50.506
191049 96.031 -76.725 50.451
191050 96.050 -76.730 50.395
191051 96.071 -76.736 50.336
191052 96.095 -76.743 50.275
191053 96.131 -76.753 50.215
191054 96.168 -76.762 50.160
191055 96.196 -76.771 50.111
191056 96.216 -76.777 50.067
191057 96.234 -76.786 50.026
191058 96.243 -76.799 49.987
191059 96.253 -76.813 49.951
191060 96.262 -76.827 49.918
191061 96.272 -76.840 49.888
191062 96.281 -76.855 49.860
191063 96.291 -76.868 49.835
191064 96.300 -76.882 49.812
191065 96.310 -76.896 49.790
191066 96.327 -76.903 49.770
191067 96.390 -76.875 49.753
191068 96.453 -76.843 49.736
191069 96.489 -76.790 49.721
191070 96.505 -76.844 49.706
191071 96.520 -76.897 49.692
191072 96.510 -76.934 49.677
191073 96.499 -76.970 49.660
191074 96.493 -77.008 49.639
191075 96.522 -77.008 49.616
191076 96.538 -76.983 49.591
191077 96.553 -76.958 49.566
191078 96.565 -76.931 49.540
191079 96.575 -76.904 49.516
191080 96.598 -76.901 49.492
191081 96.631 -76.929 49.468
191082 96.665 -76.958 49.442
191083 96.698 -76.986 49.414
191084 96.732 -77.013 49.385
191085 96.752 -77.019 49.354
191086 96.772 -77.023 49.323
191087 96.792 -77.028 49.293
191088 96.812 -77.033 49.263
191089 96.827 -77.041 49.234
191090 96.840 -77.053 49.206
191091 96.852 -77.064 49.178
191092 96.865 -77.075 49.150
191093 96.877 -77.087 49.123
191094 96.889 -77.097 49.097
191095 96.902 -77.109 49.072
191096 96.914 -77.120 49.048
191097 96.927 -77.132 49.022
191098 96.939 -77.143 48.996
191099 96.952 -77.152 48.970
191100 96.964 -77.156 48.944
191101 96.975 -77.160 48.917
191102 96.987 -77.164 48.890
191103 97.036 -77.175 48.862
191104 97.083 -77.191 48.835
191105 97.130 -77.207 48.809
191106 97.176 -77.223 48.784
191107 97.223 -77.240 48.759
191108 97.270 -77.257 48.735
191109 97.301 -77.269 48.712
191110 97.318 -77.278 48.689
191111 97.334 -77.286 48.666
191112 97.351 -77.294 48.645
191113 97.367 -77.304 48.624
191114 97.384 -77.312 48.603
191115 97.400 -77.320 48.583
191116 97.411 -77.316 48.561
191117 97.422 -77.311 48.539
191118 97.433 -77.307 48.518
191119 97.444 -77.302 48.496
191120 97.491 -77.320 48.473
191121 97.544 -77.345 48.449
191122 97.566 -77.353 48.425
191123 97.588 -77.361 48.400
191124 97.596 -77.365 48.374
191125 97.559 -77.356 48.349
191126 97.522 -77.347 48.325
191127 97.541 -77.345 48.298
191128 97.568 -77.342 48.270
191129 97.595 -77.341 48.242
191130 97.623 -77.338 48.214
191131 97.649 -77.336 48.185
191132 97.676 -77.335 48.155
191133 97.703 -77.332 48.125
191134 97.730 -77.331 48.095
191135 97.757 -77.332 48.065
191136 97.785 -77.333 48.034
191137 97.813 -77.333 48.003
191138 97.844 -77.322 47.971
191139 97.874 -77.310 47.938
191140 97.905 -77.298 47.906
191141 97.936 -77.286 47.873
191142 97.966 -77.274 47.841
191143 97.997 -77.263 47.809
191144 98.028 -77.251 47.777
191145 98.059 -77.239 47.746
191146 98.084 -77.224 47.716
191147 98.108 -77.210 47.687
191148 98.134 -77.195 47.659
191149 98.159 -77.181 47.630
191150 98.179 -77.170 47.601
191151 98.193 -77.166 47.572
191152 98.206 -77.162 47.544
191153 98.219 -77.157 47.516
191154 98.232 -77.153 47.490
191155 98.245 -77.149 47.463
191156 98.258 -77.144 47.438
191157 98.272 -77.141 47.409
191158 98.287 -77.160 47.380
191159 98.302 -77.182 47.353
191160 98.318 -77.203 47.326
191161 98.340 -77.216 47.298
191162 98.375 -77.220 47.272
191163 98.412 -77.224 47.246
191164 98.447 -77.228 47.224
191165 98.483 -77.232 47.201
191166 98.513 -77.244 47.178
191167 98.541 -77.261 47.155
191168 98.570 -77.279 47.130
191169 98.599 -77.295 47.108
191170 98.628 -77.312 47.087
191171 98.657 -77.329 47.067
191172 98.686 -77.346 47.046
191173 98.715 -77.363 47.026
191174 98.743 -77.380 47.004
191175 98.756 -77.386 46.983
191176 98.767 -77.382 46.962
191177 98.778 -77.381 46.941
191178 98.789 -77.381 46.920
191179 98.800 -77.382 46.897
191180 98.811 -77.382 46.876
191181 98.822 -77.382 46.854
191182 98.833 -77.382 46.832
191183 98.844 -77.382 46.810
191184 98.854 -77.382 46.787
191185 98.865 -77.383 46.764
191186 98.876 -77.383 46.739
191187 98.887 -77.383 46.715
191188 98.898 -77.384 46.690
191189 98.909 -77.383 46.665
191190 98.930 -77.375 46.643
191191 98.951 -77.367 46.620
191192 98.973 -77.358 46.596
191193 98.994 -77.350 46.571
191194 99.015 -77.345 46.548
191195 99.035 -77.351 46.524
191196 99.056 -77.377 46.500
191197 99.070 -77.403 46.478
191198 99.083 -77.429 46.455
191199 99.097 -77.455 46.432
191200 99.106 -77.476 46.409
191201 99.104 -77.490 46.385
191202 99.102 -77.503 46.361
191203 99.100 -77.517 46.338
191204 99.098 -77.530 46.315
191205 99.097 -77.543 46.292
191206 99.095 -77.556 46.269
191207 99.093 -77.570 46.246
191208 99.091 -77.583 46.223
191209 99.089 -77.596 46.200
191210 99.099 -77.602 46.177
191211 99.121 -77.599 46.154
191212 99.144 -77.596 46.132
191213 99.184 -77.600 46.110
191214 99.223 -77.604 46.089
191215 99.264 -77.608 46.067
191216 99.303 -77.611 46.045
191217 99.337 -77.612 46.027
191218 99.366 -77.611 46.008
191219 99.397 -77.611 45.989
191220 99.427 -77.610 45.972
191221 99.456 -77.609 45.954
191222 99.487 -77.609 45.938
191223 99.517 -77.608 45.923
191224 99.547 -77.608 45.908
191225 99.577 -77.607 45.894
191226 99.607 -77.606 45.881
191227 99.639 -77.617 45.868
191228 99.674 -77.633 45.856
191229 99.707 -77.649 45.843
191230 99.742 -77.666 45.829
191231 99.775 -77.683 45.816
191232 99.800 -77.692 45.803
191233 99.814 -77.704 45.789
191234 99.827 -77.716 45.775
191235 99.841 -77.728 45.761
191236 99.854 -77.740 45.747
191237 99.867 -77.752 45.734
191238 99.881 -77.765 45.720
191239 99.894 -77.777 45.707
191240 99.916 -77.799 45.694
191241 99.941 -77.824 45.682
191242 99.966 -77.850 45.672
191243 99.987 -77.862 45.662
191244 100.007 -77.874 45.653
191245 100.028 -77.886 45.643
191246 100.057 -77.892 45.634
191247 100.087 -77.897 45.625
191248 100.117 -77.902 45.615
191249 100.140 -77.914 45.605
191250 100.161 -77.928 45.595
191251 100.181 -77.942 45.583
191252 100.202 -77.956 45.571
191253 100.222 -77.969 45.559
191254 100.243 -77.983 45.547
191255 100.263 -77.997 45.534
191256 100.283 -78.011 45.521
191257 100.304 -78.025 45.508
191258 100.324 -78.038 45.494
191259 100.345 -78.053 45.480
191260 100.365 -78.066 45.466
191261 100.378 -78.074 45.451
191262 100.391 -78.081 45.437
191263 100.404 -78.089 45.423
191264 100.418 -78.097 45.409
191265 100.431 -78.104 45.394
191266 100.444 -78.112 45.378
191267 100.458 -78.119 45.362
191268 100.471 -78.127 45.349
191269 100.484 -78.134 45.334
191270 100.494 -78.147 45.320
191271 100.504 -78.161 45.305
191272 100.525 -78.154 45.290
191273 100.556 -78.146 45.276
191274 100.593 -78.139 45.262
191275 100.622 -78.138 45.247
191276 100.661 -78.151 45.233
191277 100.700 -78.163 45.219
191278 100.740 -78.173 45.206
191279 100.780 -78.181 45.192
191280 100.821 -78.188 45.176
191281 100.862 -78.196 45.161
191282 100.903 -78.204 45.146
191283 100.944 -78.212 45.131
191284 100.984 -78.219 45.116
191285 101.025 -78.227 45.102
191286 101.065 -78.235 45.088
191287 101.106 -78.242 45.074
191288 101.147 -78.250 45.060
191289 101.188 -78.257 45.047
191290 101.226 -78.266 45.033
191291 101.237 -78.284 45.017
191292 101.249 -78.302 45.003
191293 101.260 -78.320 44.988
191294 101.272 -78.338 44.972
191295 101.282 -78.357 44.957
191296 101.291 -78.375 44.942
191297 101.301 -78.394 44.926
191298 101.283 -78.366 44.911
191299 101.296 -78.365 44.896
191300 101.325 -78.379 44.880
191301 101.360 -78.400 44.863
191302 101.395 -78.421 44.847
191303 101.430 -78.442 44.830
191304 101.466 -78.463 44.813
191305 101.501 -78.484 44.796
191306 101.537 -78.505 44.779
191307 101.565 -78.530 44.762
191308 101.592 -78.556 44.746
191309 101.619 -78.582 44.729
191310 101.645 -78.608 44.713
191311 101.673 -78.633 44.697
191312 101.699 -78.658 44.680
191313 101.727 -78.684 44.664
191314 101.755 -78.710 44.648
191315 101.796 -78.744 44.632
191316 101.839 -78.776 44.617
191317 101.883 -78.808 44.602
191318 101.925 -78.839 44.588
191319 101.969 -78.871 44.573
191320 102.012 -78.903 44.558
191321 102.055 -78.935 44.544
191322 102.089 -78.955 44.530
191323 102.109 -78.957 44.516
191324 102.128 -78.959 44.502
191325 102.147 -78.960 44.488
191326 102.167 -78.962 44.475
191327 102.186 -78.964 44.463
191328 102.206 -78.965 44.449
191329 102.226 -78.967 44.436
191330 102.245 -78.968 44.423
191331 102.264 -78.970 44.408
191332 102.283 -78.971 44.396
191333 102.303 -78.973 44.382
191615 93.499 -76.391 54.084
191616 93.517 -76.394 54.089
191617 93.536 -76.397 54.093
191618 93.554 -76.400 54.096
191619 93.572 -76.403 54.098
191620 93.591 -76.406 54.102
191621 93.609 -76.408 54.106
191622 93.627 -76.411 54.111
191623 93.646 -76.414 54.117
191624 93.664 -76.417 54.123
191625 93.683 -76.420 54.129
191626 93.701 -76.423 54.135
191627 93.719 -76.426 54.143
191628 93.740 -76.420 54.155
191629 93.759 -76.412 54.170
191630 93.774 -76.411 54.185
191631 93.788 -76.407 54.197
191632 93.803 -76.403 54.204
191633 93.819 -76.389 54.209
191634 93.835 -76.373 54.214
191635 93.851 -76.358 54.218
191636 93.867 -76.348 54.219
191637 93.882 -76.343 54.220
191638 93.897 -76.339 54.219
191639 93.916 -76.345 54.213
191640 93.935 -76.351 54.208
191641 93.955 -76.357 54.203
191642 93.972 -76.361 54.202
191643 93.989 -76.364 54.207
191644 94.006 -76.367 54.209
191645 94.024 -76.367 54.209
191646 94.041 -76.363 54.206
191647 94.059 -76.358 54.201
191648 94.076 -76.354 54.199
191649 94.095 -76.351 54.199
191650 94.113 -76.348 54.201
191651 94.133 -76.348 54.203
191652 94.153 -76.346 54.200
191653 94.173 -76.345 54.196
191654 94.195 -76.338 54.191
191655 94.218 -76.330 54.182
191656 94.238 -76.328 54.170
191657 94.257 -76.326 54.155
191658 94.277 -76.325 54.137
191659 94.296 -76.323 54.111
191660 94.316 -76.322 54.083
191661 94.336 -76.320 54.054
191662 94.356 -76.319 54.025
191663 94.376 -76.317 53.997
191664 94.395 -76.315 53.968
191665 94.413 -76.314 53.939
191666 94.431 -76.311 53.904
191667 94.449 -76.310 53.869
191668 94.468 -76.309 53.835
191669 94.486 -76.308 53.801
191670 94.505 -76.307 53.768
191671 94.523 -76.306 53.736
191672 94.542 -76.305 53.705
191673 94.557 -76.304 53.672
191674 94.570 -76.304 53.642
191675 94.582 -76.308 53.611
191676 94.595 -76.317 53.581
191677 94.607 -76.326 53.549
191678 94.619 -76.336 53.516
191679 94.631 -76.345 53.481
191680 94.642 -76.354 53.444
191681 94.655 -76.361 53.406
191682 94.668 -76.367 53.369
191683 94.669 -76.368 53.334
191684 94.676 -76.372 53.300
191685 94.685 -76.376 53.267
191686 94.694 -76.379 53.234
191687 94.705 -76.383 53.200
191688 94.716 -76.386 53.166
191689 94.728 -76.389 53.133
191690 94.740 -76.392 53.099
191691 94.751 -76.396 53.065
191692 94.762 -76.399 53.030
191693 94.774 -76.402 52.991
191694 94.785 -76.406 52.952
191695 94.796 -76.409 52.912
191696 94.808 -76.412 52.872
191697 94.819 -76.416 52.831
191698 94.831 -76.419 52.790
191699 94.842 -76.422 52.746
191700 94.853 -76.426 52.699
191701 94.865 -76.429 52.652
191702 94.877 -76.432 52.603
191703 94.888 -76.436 52.556
191704 94.899 -76.439 52.509
191705 94.911 -76.442 52.461
191706 94.945 -76.449 52.412
191707 94.985 -76.457 52.360
191708 95.027 -76.465 52.306
191709 95.057 -76.472 52.252
191710 95.082 -76.479 52.196
191711 95.107 -76.485 52.143
191712 95.131 -76.492 52.091
191713 95.158 -76.498 52.041
191714 95.207 -76.492 51.993
191715 95.251 -76.490 51.948
191716 95.274 -76.495 51.903
191717 95.297 -76.499 51.861
191718 95.320 -76.504 51.820
191719 95.343 -76.508 51.781
191720 95.367 -76.514 51.743
191721 95.390 -76.518 51.704
191722 95.410 -76.526 51.664
191723 95.429 -76.536 51.618
191724 95.449 -76.546 51.571
191725 95.468 -76.556 51.524
191726 95.487 -76.567 51.482
191727 95.506 -76.577 51.442
191728 95.525 -76.587 51.403
191729 95.544 -76.597 51.365
191730 95.563 -76.608 51.329
191731 95.582 -76.618 51.292
191732 95.610 -76.621 51.255
191733 95.640 -76.625 51.217
191734 95.670 -76.628 51.179
191735 95.699 -76.631 51.140
191736 95.729 -76.634 51.099
191737 95.752 -76.639 51.056
191738 95.774 -76.643 51.014
191739 95.797 -76.648 50.971
191740 95.820 -76.652 50.927
191741 95.843 -76.658 50.888
191742 95.866 -76.662 50.849
191743 95.888 -76.667 50.810
191744 95.911 -76.671 50.771
191745 95.933 -76.674 50.728
191746 95.954 -76.681 50.682
191747 95.974 -76.687 50.634
191748 95.995 -76.692 50.585
191749 96.016 -76.696 50.532
191750 96.036 -76.702 50.477
191751 96.057 -76.707 50.422
191752 96.078 -76.712 50.363
191753 96.098 -76.716 50.303
191754 96.132 -76.724 50.245
191755 96.169 -76.734 50.191
191756 96.195 -76.743 50.143
191757 96.215 -76.749 50.099
191758 96.235 -76.756 50.058
191759 96.254 -76.763 50.020
191760 96.274 -76.771 49.984
191761 96.285 -76.783 49.953
191762 96.294 -76.797 49.924
191763 96.304 -76.811 49.896
191764 96.313 -76.824 49.872
191765 96.323 -76.838 49.849
191766 96.332 -76.852 49.828
191767 96.368 -76.846 49.808
191768 96.431 -76.816 49.791
191769 96.495 -76.787 49.775
191770 96.488 -76.775 49.761
191771 96.477 -76.794 49.745
191772 96.465 -76.830 49.730
191773 96.453 -76.865 49.715
191774 96.441 -76.900 49.697
191775 96.475 -76.953 49.673
191776 96.513 -76.965 49.648
191777 96.528 -76.940 49.621
191778 96.544 -76.915 49.595
191779 96.559 -76.890 49.569
191780 96.575 -76.865 49.544
191781 96.609 -76.859 49.519
191782 96.637 -76.881 49.494
191783 96.670 -76.909 49.467
191784 96.703 -76.937 49.438
191785 96.737 -76.965 49.407
191786 96.768 -76.990 49.376
191787 96.789 -76.995 49.345
191788 96.809 -77.000 49.314
191789 96.829 -77.005 49.284
191790 96.842 -77.016 49.254
191791 96.855 -77.026 49.225
191792 96.867 -77.038 49.197
191793 96.880 -77.049 49.169
191794 96.892 -77.060 49.142
191795 96.904 -77.072 49.116
191796 96.917 -77.083 49.090
191797 96.930 -77.094 49.066
191798 96.942 -77.105 49.041
191799 96.954 -77.116 49.015
191800 96.967 -77.128 48.988
191801 96.979 -77.139 48.962
191802 96.992 -77.150 48.936
191803 97.004 -77.156 48.908
191804 97.039 -77.166 48.880
191805 97.086 -77.182 48.853
191806 97.133 -77.199 48.826
191807 97.179 -77.215 48.800
191808 97.226 -77.232 48.775
191809 97.273 -77.248 48.751
191810 97.319 -77.264 48.728
191811 97.335 -77.273 48.706
191812 97.352 -77.282 48.684
191813 97.369 -77.290 48.662
191814 97.385 -77.299 48.640
191815 97.398 -77.300 48.618
191816 97.409 -77.296 48.596
191817 97.420 -77.291 48.575
191818 97.431 -77.287 48.553
191819 97.442 -77.282 48.531
191820 97.453 -77.278 48.507
191821 97.487 -77.287 48.484
191822 97.575 -77.330 48.460
191823 97.597 -77.338 48.436
191824 97.617 -77.346 48.409
191825 97.580 -77.337 48.383
191826 97.543 -77.328 48.356
191827 97.529 -77.322 48.330
191828 97.556 -77.320 48.302
191829 97.583 -77.317 48.274
191830 97.610 -77.316 48.246
191831 97.637 -77.313 48.217
191832 97.664 -77.312 48.189
191833 97.691 -77.310 48.159
191834 97.718 -77.308 48.129
191835 97.745 -77.306 48.101
191836 97.772 -77.304 48.071
191837 97.800 -77.306 48.041
191838 97.827 -77.307 48.009
191839 97.858 -77.295 47.977
191840 97.888 -77.284 47.945
191841 97.919 -77.272 47.913
191842 97.950 -77.260 47.881
191843 97.981 -77.248 47.849
191844 98.011 -77.237 47.817
191845 98.038 -77.223 47.785
191846 98.063 -77.208 47.753
191847 98.088 -77.194 47.722
191848 98.113 -77.179 47.693
191849 98.139 -77.165 47.666
191850 98.164 -77.150 47.636
191851 98.189 -77.135 47.607
191852 98.211 -77.124 47.578
191853 98.224 -77.119 47.550
191854 98.237 -77.116 47.522
191855 98.250 -77.111 47.495
191856 98.263 -77.106 47.469
191857 98.277 -77.103 47.441
191858 98.292 -77.123 47.413
191859 98.307 -77.144 47.385
191860 98.323 -77.165 47.357
191861 98.339 -77.186 47.331
191862 98.362 -77.198 47.304
191863 98.398 -77.202 47.278
191864 98.434 -77.206 47.252
191865 98.469 -77.210 47.231
191866 98.489 -77.219 47.208
191867 98.513 -77.230 47.186
191868 98.546 -77.243 47.163
191869 98.577 -77.257 47.138
191870 98.605 -77.275 47.115
191871 98.635 -77.292 47.094
191872 98.663 -77.310 47.073
191873 98.692 -77.328 47.052
191874 98.721 -77.345 47.032
191875 98.750 -77.363 47.011
191876 98.769 -77.377 46.989
191877 98.782 -77.382 46.968
191878 98.793 -77.382 46.947
191879 98.804 -77.382 46.926
191880 98.814 -77.382 46.904
191881 98.826 -77.382 46.883
191882 98.836 -77.382 46.861
191883 98.847 -77.383 46.838
191884 98.858 -77.383 46.815
191885 98.869 -77.383 46.791
191886 98.879 -77.384 46.767
191887 98.890 -77.384 46.744
191888 98.901 -77.384 46.719
191889 98.912 -77.385 46.694
191890 98.924 -77.385 46.669
191891 98.945 -77.376 46.647
191892 98.966 -77.367 46.624
191893 98.987 -77.359 46.600
191894 99.008 -77.351 46.576
191895 99.029 -77.342 46.553
191896 99.046 -77.339 46.529
191897 99.059 -77.360 46.505
191898 99.073 -77.386 46.483
191899 99.087 -77.413 46.459
191900 99.101 -77.439 46.436
191901 99.115 -77.464 46.412
191902 99.118 -77.481 46.390
191903 99.116 -77.495 46.367
191904 99.114 -77.508 46.344
191905 99.113 -77.521 46.322
191906 99.111 -77.535 46.298
191907 99.109 -77.548 46.275
191908 99.107 -77.561 46.252
191909 99.105 -77.574 46.228
191910 99.103 -77.588 46.204
191911 99.122 -77.587 46.181
191912 99.144 -77.584 46.158
191913 99.166 -77.582 46.135
191914 99.189 -77.579 46.113
191915 99.229 -77.583 46.092
191916 99.269 -77.586 46.070
191917 99.308 -77.590 46.049
191918 99.348 -77.594 46.029
191919 99.388 -77.598 46.010
191920 99.428 -77.602 45.991
191921 99.465 -77.604 45.974
191922 99.495 -77.603 45.957
191923 99.525 -77.602 45.940
191924 99.555 -77.602 45.925
191925 99.585 -77.602 45.910
191926 99.615 -77.601 45.896
191927 99.645 -77.600 45.883
191928 99.676 -77.599 45.869
191929 99.709 -77.616 45.856
191930 99.744 -77.633 45.843
191931 99.777 -77.649 45.830
191932 99.806 -77.661 45.816
191933 99.820 -77.673 45.803
191934 99.833 -77.685 45.789
191935 99.847 -77.697 45.775
191936 99.860 -77.709 45.760
191937 99.873 -77.721 45.747
191938 99.887 -77.733 45.733
191939 99.901 -77.746 45.719
191940 99.914 -77.758 45.706
191941 99.931 -77.774 45.694
191942 99.957 -77.801 45.682
191943 99.981 -77.827 45.671
191944 100.005 -77.849 45.662
191945 100.026 -77.860 45.653
191946 100.050 -77.870 45.643
191947 100.080 -77.875 45.634
191948 100.107 -77.884 45.624
191949 100.127 -77.897 45.615
191950 100.147 -77.911 45.605
191951 100.168 -77.925 45.595
191952 100.189 -77.939 45.583
191953 100.209 -77.953 45.572
191954 100.229 -77.967 45.560
191955 100.250 -77.981 45.547
191956 100.270 -77.994 45.534
191957 100.290 -78.008 45.521
191958 100.311 -78.022 45.509
191959 100.332 -78.036 45.495
191960 100.352 -78.050 45.480
191961 100.372 -78.063 45.466
191962 100.386 -78.072 45.450
191963 100.399 -78.080 45.436
191964 100.412 -78.087 45.422
191965 100.426 -78.094 45.407
191966 100.439 -78.102 45.392
191967 100.453 -78.109 45.376
191968 100.466 -78.117 45.362
191969 100.479 -78.125 45.347
191970 100.492 -78.132 45.333
191971 100.505 -78.140 45.319
191972 100.517 -78.150 45.304
191973 100.546 -78.144 45.289
191974 100.582 -78.138 45.275
191975 100.619 -78.131 45.261
191976 100.651 -78.128 45.247
191977 100.681 -78.134 45.233
191978 100.719 -78.147 45.219
191979 100.759 -78.159 45.205
191980 100.797 -78.172 45.192
191981 100.838 -78.182 45.177
191982 100.879 -78.190 45.162
191983 100.919 -78.197 45.146
191984 100.960 -78.205 45.132
191985 101.000 -78.213 45.117
191986 101.041 -78.220 45.104
191987 101.082 -78.228 45.090
191988 101.123 -78.235 45.075
191989 101.164 -78.243 45.061
191990 101.204 -78.250 45.047
191991 101.235 -78.262 45.033
191992 101.246 -78.280 45.019
191993 101.258 -78.298 45.004
191994 101.270 -78.316 44.989
191995 101.281 -78.334 44.973
191996 101.293 -78.352 44.958
191997 101.303 -78.370 44.944
191998 101.298 -78.366 44.928
191999 101.312 -78.366 44.913
192000 101.325 -78.364 44.898
192001 101.341 -78.366 44.882
192002 101.376 -78.387 44.865
192003 101.412 -78.408 44.849
192004 101.447 -78.429 44.831
192005 101.482 -78.451 44.814
192006 101.518 -78.472 44.797
192007 101.550 -78.495 44.779
192008 101.578 -78.520 44.763
192009 101.604 -78.545 44.747
192010 101.631 -78.571 44.730
192011 101.658 -78.597 44.713
192012 101.685 -78.623 44.697
192013 101.712 -78.648 44.681
192014 101.740 -78.671 44.665
192015 101.770 -78.692 44.649
192016 101.815 -78.717 44.633
192017 101.858 -78.749 44.618
192018 101.901 -78.781 44.604
192019 101.944 -78.813 44.590
192020 101.987 -78.845 44.575
192021 102.030 -78.877 44.560
192022 102.074 -78.908 44.546
192023 102.114 -78.938 44.532
192024 102.134 -78.940 44.518
192025 102.153 -78.941 44.504
192026 102.173 -78.943 44.490
192027 102.192 -78.945 44.477
192028 102.211 -78.946 44.464
192029 102.231 -78.948 44.450
192030 102.250 -78.949 44.436
192031 102.270 -78.951 44.422
192032 102.290 -78.952 44.409
192033 102.309 -78.954 44.396
192034 102.328 -78.955 44.382
192035 102.348 -78.958 44.368
192317 93.509 -76.370 54.082
192318 93.528 -76.373 54.086
192319 93.546 -76.376 54.089
192320 93.564 -76.379 54.092
192321 93.582 -76.383 54.096
192322 93.601 -76.386 54.100
192323 93.619 -76.388 54.105
192324 93.638 -76.391 54.112
192325 93.656 -76.394 54.118
192326 93.675 -76.397 54.124
192327 93.693 -76.400 54.130
192328 93.711 -76.403 54.138
192329 93.729 -76.405 54.149
192330 93.750 -76.400 54.163
192331 93.766 -76.396 54.178
192332 93.781 -76.392 54.193
192333 93.796 -76.385 54.198
192334 93.813 -76.369 54.203
192335 93.829 -76.353 54.206
192336 93.845 -76.337 54.210
192337 93.861 -76.326 54.211
192338 93.875 -76.321 54.212
192339 93.893 -76.323 54.211
192340 93.912 -76.329 54.206
192341 93.931 -76.335 54.203
192342 93.950 -76.340 54.202
192343 93.969 -76.346 54.202
192344 93.986 -76.350 54.207
192345 94.003 -76.352 54.212
192346 94.020 -76.351 54.214
192347 94.038 -76.348 54.211
192348 94.056 -76.344 54.207
192349 94.073 -76.339 54.204
192350 94.092 -76.338 54.203
192351 94.113 -76.336 54.203
192352 94.132 -76.336 54.204
192353 94.152 -76.335 54.201
192354 94.172 -76.333 54.196
192355 94.194 -76.327 54.192
192356 94.217 -76.319 54.183
192357 94.239 -76.312 54.171
192358 94.258 -76.310 54.156
192359 94.278 -76.309 54.139
192360 94.298 -76.307 54.115
192361 94.318 -76.306 54.088
192362 94.337 -76.304 54.060
192363 94.357 -76.303 54.032
192364 94.377 -76.301 54.004
192365 94.395 -76.299 53.974
192366 94.413 -76.298 53.944
192367 94.431 -76.295 53.909
192368 94.449 -76.295 53.876
192369 94.468 -76.293 53.844
192370 94.486 -76.292 53.812
192371 94.505 -76.292 53.780
192372 94.523 -76.290 53.748
192373 94.537 -76.290 53.717
192374 94.551 -76.289 53.684
192375 94.564 -76.289 53.653
192376 94.577 -76.292 53.624
192377 94.589 -76.301 53.594
192378 94.601 -76.309 53.564
192379 94.613 -76.318 53.532
192380 94.625 -76.327 53.497
192381 94.637 -76.336 53.461
192382 94.649 -76.345 53.423
192383 94.661 -76.354 53.387
192384 94.675 -76.361 53.351
192385 94.687 -76.365 53.317
192386 94.700 -76.368 53.284
192387 94.712 -76.372 53.251
192388 94.725 -76.376 53.217
192389 94.738 -76.379 53.183
192390 94.750 -76.383 53.150
192391 94.763 -76.386 53.116
192392 94.775 -76.390 53.081
192393 94.787 -76.393 53.046
192394 94.799 -76.397 53.006
192395 94.810 -76.400 52.966
192396 94.821 -76.403 52.924
192397 94.833 -76.407 52.883
192398 94.844 -76.410 52.841
192399 94.855 -76.413 52.798
192400 94.867 -76.417 52.753
192401 94.879 -76.420 52.707
192402 94.890 -76.423 52.662
192403 94.901 -76.426 52.616
192404 94.913 -76.430 52.570
192405 94.924 -76.433 52.524
192406 94.935 -76.436 52.476
192407 94.962 -76.442 52.427
192408 94.997 -76.449 52.374
192409 95.022 -76.456 52.320
192410 95.047 -76.462 52.265
192411 95.072 -76.469 52.209
192412 95.097 -76.475 52.156
192413 95.121 -76.482 52.105
192414 95.154 -76.484 52.055
192415 95.203 -76.480 52.008
192416 95.253 -76.475 51.963
192417 95.278 -76.479 51.920
192418 95.301 -76.483 51.877
192419 95.324 -76.489 51.835
192420 95.347 -76.493 51.795
192421 95.371 -76.498 51.756
192422 95.394 -76.502 51.717
192423 95.417 -76.507 51.676
192424 95.440 -76.511 51.632
192425 95.462 -76.517 51.586
192426 95.481 -76.528 51.539
192427 95.500 -76.538 51.497
192428 95.519 -76.549 51.459
192429 95.538 -76.558 51.422
192430 95.557 -76.569 51.386
192431 95.577 -76.579 51.351
192432 95.595 -76.589 51.317
192433 95.619 -76.596 51.282
192434 95.649 -76.599 51.246
192435 95.679 -76.602 51.208
192436 95.708 -76.605 51.168
192437 95.734 -76.609 51.126
192438 95.757 -76.614 51.083
192439 95.779 -76.618 51.039
192440 95.803 -76.624 50.996
192441 95.825 -76.628 50.952
192442 95.848 -76.633 50.911
192443 95.871 -76.637 50.871
192444 95.894 -76.642 50.831
192445 95.918 -76.649 50.792
192446 95.942 -76.653 50.748
192447 95.963 -76.657 50.703
192448 95.983 -76.662 50.656
192449 96.004 -76.667 50.608
192450 96.025 -76.672 50.556
192451 96.045 -76.677 50.503
192452 96.066 -76.683 50.448
192453 96.087 -76.687 50.390
192454 96.107 -76.691 50.331
192455 96.133 -76.696 50.274
192456 96.170 -76.706 50.222
192457 96.195 -76.714 50.175
192458 96.215 -76.721 50.132
192459 96.234 -76.728 50.092
192460 96.254 -76.735 50.054
192461 96.273 -76.742 50.020
192462 96.293 -76.749 49.989
192463 96.313 -76.756 49.960
192464 96.326 -76.767 49.932
192465 96.336 -76.781 49.909
192466 96.345 -76.794 49.887
192467 96.355 -76.809 49.867
192468 96.408 -76.787 49.847
192469 96.448 -76.764 49.831
192470 96.439 -76.753 49.816
192471 96.429 -76.742 49.800
192472 96.419 -76.731 49.785
192473 96.408 -76.760 49.770
192474 96.396 -76.796 49.753
192475 96.392 -76.834 49.732
192476 96.458 -76.897 49.707
192477 96.504 -76.923 49.679
192478 96.519 -76.899 49.650
192479 96.534 -76.874 49.624
192480 96.550 -76.849 49.598
192481 96.587 -76.842 49.572
192482 96.625 -76.837 49.545
192483 96.656 -76.847 49.519
192484 96.683 -76.868 49.490
192485 96.710 -76.890 49.460
192486 96.742 -76.916 49.429
192487 96.775 -76.944 49.397
192488 96.806 -76.968 49.365
192489 96.826 -76.972 49.334
192490 96.845 -76.978 49.303
192491 96.858 -76.989 49.272
192492 96.870 -77.000 49.243
192493 96.882 -77.011 49.215
192494 96.895 -77.022 49.186
192495 96.907 -77.034 49.160
192496 96.920 -77.045 49.134
192497 96.932 -77.056 49.108
192498 96.945 -77.068 49.084
192499 96.957 -77.078 49.059
192500 96.969 -77.090 49.034
192501 96.982 -77.101 49.007
192502 96.995 -77.111 48.981
192503 97.008 -77.120 48.954
192504 97.022 -77.129 48.926
192505 97.053 -77.147 48.898
192506 97.093 -77.169 48.871
192507 97.135 -77.191 48.844
192508 97.182 -77.207 48.818
192509 97.228 -77.223 48.793
192510 97.275 -77.240 48.768
192511 97.322 -77.256 48.744
192512 97.354 -77.269 48.721
192513 97.370 -77.277 48.699
192514 97.386 -77.285 48.677
192515 97.397 -77.280 48.655
192516 97.408 -77.276 48.634
192517 97.419 -77.271 48.610
192518 97.430 -77.267 48.588
192519 97.441 -77.263 48.566
192520 97.451 -77.258 48.542
192521 97.462 -77.254 48.519
192522 97.483 -77.255 48.495
192523 97.581 -77.303 48.470
192524 97.628 -77.324 48.445
192525 97.602 -77.318 48.419
192526 97.565 -77.309 48.392
192527 97.528 -77.300 48.364
192528 97.544 -77.297 48.336
192529 97.571 -77.294 48.308
192530 97.598 -77.293 48.280
192531 97.625 -77.291 48.251
192532 97.652 -77.289 48.221
192533 97.679 -77.287 48.192
192534 97.706 -77.285 48.164
192535 97.733 -77.283 48.135
192536 97.760 -77.281 48.107
192537 97.787 -77.279 48.077
192538 97.814 -77.279 48.046
192539 97.842 -77.281 48.016
192540 97.872 -77.269 47.985
192541 97.902 -77.258 47.954
192542 97.933 -77.246 47.922
192543 97.964 -77.235 47.889
192544 97.993 -77.222 47.857
192545 98.018 -77.207 47.825
192546 98.043 -77.192 47.793
192547 98.068 -77.178 47.760
192548 98.093 -77.163 47.730
192549 98.119 -77.149 47.701
192550 98.143 -77.134 47.673
192551 98.168 -77.119 47.644
192552 98.194 -77.105 47.614
192553 98.219 -77.091 47.585
192554 98.242 -77.078 47.557
192555 98.255 -77.073 47.529
192556 98.268 -77.069 47.502
192557 98.281 -77.065 47.474
192558 98.297 -77.085 47.446
192559 98.312 -77.106 47.418
192560 98.328 -77.128 47.390
192561 98.343 -77.149 47.362
192562 98.359 -77.169 47.335
192563 98.384 -77.180 47.309
192564 98.421 -77.184 47.283
192565 98.456 -77.188 47.259
192566 98.473 -77.197 47.238
192567 98.489 -77.207 47.216
192568 98.511 -77.219 47.193
192569 98.544 -77.232 47.170
192570 98.577 -77.244 47.146
192571 98.609 -77.257 47.123
192572 98.641 -77.271 47.101
192573 98.670 -77.288 47.080
192574 98.699 -77.306 47.058
192575 98.727 -77.323 47.038
192576 98.756 -77.341 47.017
192577 98.773 -77.354 46.995
192578 98.790 -77.368 46.973
192579 98.807 -77.382 46.952
192580 98.818 -77.382 46.932
192581 98.829 -77.383 46.910
192582 98.840 -77.383 46.888
192583 98.851 -77.383 46.865
192584 98.861 -77.384 46.843
192585 98.872 -77.384 46.819
192586 98.883 -77.384 46.795
192587 98.894 -77.385 46.771
192588 98.905 -77.385 46.748
192589 98.916 -77.385 46.724
192590 98.927 -77.385 46.699
192591 98.938 -77.385 46.675
192592 98.959 -77.377 46.652
192593 98.980 -77.368 46.629
192594 99.001 -77.360 46.605
192595 99.022 -77.351 46.580
192596 99.042 -77.345 46.557
192597 99.053 -77.350 46.534
192598 99.064 -77.354 46.511
192599 99.076 -77.370 46.489
192600 99.090 -77.396 46.465
192601 99.104 -77.422 46.442
192602 99.118 -77.448 46.418
192603 99.132 -77.473 46.395
192604 99.130 -77.486 46.372
192605 99.128 -77.499 46.349
192606 99.126 -77.513 46.326
192607 99.124 -77.526 46.303
192608 99.123 -77.539 46.280
192609 99.121 -77.553 46.256
192610 99.119 -77.566 46.233
192611 99.122 -77.576 46.208
192612 99.144 -77.573 46.185
192613 99.166 -77.570 46.162
192614 99.189 -77.567 46.139
192615 99.211 -77.564 46.117
192616 99.234 -77.561 46.095
192617 99.274 -77.565 46.073
192618 99.313 -77.569 46.051
192619 99.353 -77.573 46.032
192620 99.394 -77.577 46.012
192621 99.433 -77.580 45.994
192622 99.473 -77.584 45.977
192623 99.513 -77.588 45.960
192624 99.553 -77.592 45.943
192625 99.592 -77.595 45.927
192626 99.624 -77.595 45.912
192627 99.654 -77.595 45.898
192628 99.684 -77.594 45.883
192629 99.714 -77.593 45.871
192630 99.746 -77.599 45.857
192631 99.779 -77.615 45.844
192632 99.812 -77.630 45.831
192633 99.826 -77.642 45.817
192634 99.840 -77.654 45.803
192635 99.853 -77.666 45.789
192636 99.866 -77.678 45.775
192637 99.880 -77.690 45.760
192638 99.894 -77.702 45.747
192639 99.907 -77.714 45.733
192640 99.920 -77.727 45.718
192641 99.934 -77.739 45.707
192642 99.947 -77.751 45.695
192643 99.972 -77.777 45.683
192644 99.997 -77.802 45.671
192645 100.022 -77.829 45.663
192646 100.045 -77.846 45.653
192647 100.073 -77.853 45.643
192648 100.093 -77.867 45.633
192649 100.114 -77.881 45.623
192650 100.134 -77.895 45.614
192651 100.155 -77.909 45.605
192652 100.175 -77.922 45.594
192653 100.196 -77.936 45.584
192654 100.216 -77.950 45.572
192655 100.237 -77.964 45.560
192656 100.257 -77.978 45.547
192657 100.277 -77.991 45.533
192658 100.298 -78.006 45.520
192659 100.318 -78.019 45.508
192660 100.339 -78.033 45.494
192661 100.359 -78.047 45.480
192662 100.379 -78.061 45.465
192663 100.394 -78.070 45.450
192664 100.407 -78.078 45.435
192665 100.421 -78.085 45.420
192666 100.434 -78.093 45.405
192667 100.447 -78.100 45.390
192668 100.461 -78.107 45.375
192669 100.474 -78.115 45.360
192670 100.487 -78.122 45.346
192671 100.504 -78.125 45.331
192672 100.522 -78.128 45.317
192673 100.542 -78.134 45.303
192674 100.571 -78.136 45.288
192675 100.608 -78.129 45.274
192676 100.645 -78.122 45.260
192677 100.680 -78.117 45.247
192678 100.703 -78.119 45.233
192679 100.739 -78.130 45.219
192680 100.778 -78.143 45.205
192681 100.818 -78.155 45.191
192682 100.856 -78.168 45.177
192683 100.896 -78.180 45.163
192684 100.935 -78.191 45.148
192685 100.976 -78.198 45.134
192686 101.017 -78.206 45.118
192687 101.058 -78.213 45.104
192688 101.098 -78.221 45.090
192689 101.139 -78.228 45.075
192690 101.180 -78.237 45.062
192691 101.220 -78.244 45.048
192692 101.244 -78.258 45.034
192693 101.256 -78.275 45.020
192694 101.267 -78.294 45.005
192695 101.279 -78.312 44.990
192696 101.291 -78.330 44.974
192697 101.302 -78.348 44.959
192698 101.314 -78.366 44.945
192699 101.326 -78.366 44.930
192700 101.339 -78.365 44.914
192701 101.353 -78.363 44.899
192702 101.366 -78.362 44.883
192703 101.393 -78.375 44.867
192704 101.428 -78.396 44.851
192705 101.464 -78.417 44.832
192706 101.499 -78.438 44.815
192707 101.534 -78.460 44.798
192708 101.563 -78.484 44.780
192709 101.590 -78.510 44.764
192710 101.617 -78.535 44.747
192711 101.644 -78.561 44.730
192712 101.671 -78.586 44.713
192713 101.698 -78.612 44.698
192714 101.728 -78.633 44.682
192715 101.757 -78.653 44.666
192716 101.787 -78.673 44.650
192717 101.832 -78.696 44.635
192718 101.877 -78.724 44.620
192719 101.919 -78.755 44.605
192720 101.963 -78.787 44.590
192721 102.006 -78.819 44.576
192722 102.049 -78.851 44.562
192723 102.092 -78.883 44.548
192724 102.135 -78.915 44.533
192725 102.159 -78.922 44.519
192726 102.178 -78.924 44.505
192727 102.198 -78.926 44.491
192728 102.218 -78.927 44.478
192729 102.237 -78.929 44.464
192730 102.256 -78.930 44.451
192731 102.275 -78.932 44.436
192732 102.295 -78.933 44.422
192733 102.315 -78.935 44.409
192734 102.334 -78.937 44.395
192735 102.353 -78.938 44.382
192736 102.373 -78.940 44.368
193019 93.519 -76.350 54.077
193020 93.538 -76.353 54.081
193021 93.556 -76.356 54.085
193022 93.574 -76.359 54.089
193023 93.593 -76.362 54.094
193024 93.611 -76.364 54.099
193025 93.630 -76.367 54.106
193026 93.648 -76.370 54.113
193027 93.666 -76.373 54.119
193028 93.685 -76.376 54.125
193029 93.703 -76.379 54.133
193030 93.721 -76.383 54.143
193031 93.740 -76.385 54.156
193032 93.759 -76.382 54.171
193033 93.774 -76.379 54.188
193034 93.790 -76.364 54.193
193035 93.806 -76.348 54.196
193036 93.822 -76.332 54.198
193037 93.839 -76.317 54.201
193038 93.854 -76.304 54.203
193039 93.869 -76.300 54.204
193040 93.889 -76.306 54.204
193041 93.908 -76.312 54.202
193042 93.927 -76.318 54.202
193043 93.945 -76.323 54.202
193044 93.965 -76.329 54.205
193045 93.983 -76.335 54.210
193046 94.000 -76.338 54.216
193047 94.017 -76.336 54.219
193048 94.035 -76.332 54.216
193049 94.053 -76.329 54.212
193050 94.072 -76.326 54.210
193051 94.092 -76.326 54.209
193052 94.112 -76.325 54.208
193053 94.131 -76.323 54.209
193054 94.151 -76.323 54.206
193055 94.171 -76.322 54.201
193056 94.193 -76.317 54.195
193057 94.215 -76.308 54.187
193058 94.239 -76.299 54.175
193059 94.259 -76.295 54.159
193060 94.279 -76.293 54.142
193061 94.299 -76.292 54.119
193062 94.319 -76.290 54.093
193063 94.338 -76.289 54.066
193064 94.358 -76.287 54.038
193065 94.377 -76.285 54.009
193066 94.395 -76.283 53.980
193067 94.413 -76.282 53.949
193068 94.431 -76.279 53.916
193069 94.449 -76.279 53.883
193070 94.468 -76.277 53.852
193071 94.486 -76.276 53.822
193072 94.504 -76.276 53.790
193073 94.518 -76.275 53.758
193074 94.531 -76.275 53.727
193075 94.545 -76.274 53.694
193076 94.558 -76.274 53.663
193077 94.571 -76.274 53.634
193078 94.583 -76.283 53.606
193079 94.596 -76.292 53.576
193080 94.607 -76.301 53.546
193081 94.619 -76.310 53.513
193082 94.631 -76.320 53.477
193083 94.643 -76.329 53.440
193084 94.663 -76.340 53.403
193085 94.675 -76.345 53.368
193086 94.686 -76.351 53.334
193087 94.698 -76.355 53.301
193088 94.711 -76.359 53.268
193089 94.723 -76.362 53.235
193090 94.736 -76.366 53.201
193091 94.748 -76.370 53.168
193092 94.761 -76.373 53.134
193093 94.773 -76.376 53.100
193094 94.786 -76.380 53.063
193095 94.799 -76.384 53.023
193096 94.811 -76.388 52.981
193097 94.824 -76.391 52.938
193098 94.836 -76.395 52.893
193099 94.849 -76.398 52.849
193100 94.861 -76.402 52.804
193101 94.874 -76.405 52.760
193102 94.887 -76.409 52.716
193103 94.899 -76.413 52.673
193104 94.912 -76.417 52.631
193105 94.924 -76.420 52.586
193106 94.936 -76.423 52.541
193107 94.949 -76.427 52.494
193108 94.963 -76.433 52.445
193109 94.987 -76.439 52.391
193110 95.012 -76.446 52.337
193111 95.037 -76.452 52.282
193112 95.061 -76.459 52.226
193113 95.086 -76.465 52.172
193114 95.111 -76.472 52.120
193115 95.151 -76.471 52.070
193116 95.200 -76.467 52.024
193117 95.250 -76.462 51.980
193118 95.282 -76.463 51.937
193119 95.306 -76.468 51.895
193120 95.328 -76.473 51.851
193121 95.351 -76.477 51.808
193122 95.375 -76.482 51.766
193123 95.398 -76.486 51.726
193124 95.421 -76.491 51.684
193125 95.444 -76.496 51.642
193126 95.467 -76.501 51.598
193127 95.490 -76.505 51.554
193128 95.513 -76.510 51.513
193129 95.533 -76.520 51.475
193130 95.551 -76.530 51.440
193131 95.570 -76.540 51.406
193132 95.590 -76.551 51.374
193133 95.609 -76.561 51.342
193134 95.629 -76.571 51.310
193135 95.659 -76.574 51.275
193136 95.688 -76.577 51.238
193137 95.716 -76.580 51.197
193138 95.739 -76.584 51.154
193139 95.762 -76.589 51.109
193140 95.785 -76.594 51.065
193141 95.808 -76.599 51.021
193142 95.831 -76.603 50.977
193143 95.854 -76.608 50.935
193144 95.878 -76.614 50.893
193145 95.902 -76.621 50.852
193146 95.926 -76.627 50.811
193147 95.950 -76.630 50.767
193148 95.972 -76.632 50.721
193149 95.992 -76.637 50.674
193150 96.013 -76.643 50.627
193151 96.034 -76.648 50.577
193152 96.054 -76.652 50.525
193153 96.075 -76.658 50.471
193154 96.096 -76.661 50.415
193155 96.116 -76.665 50.358
193156 96.136 -76.669 50.303
193157 96.171 -76.678 50.254
193158 96.194 -76.686 50.207
193159 96.214 -76.693 50.164
193160 96.234 -76.699 50.124
193161 96.253 -76.706 50.087
193162 96.273 -76.714 50.054
193163 96.293 -76.721 50.024
193164 96.312 -76.727 49.996
193165 96.332 -76.734 49.970
193166 96.352 -76.742 49.946
193167 96.368 -76.751 49.925
193168 96.385 -76.759 49.905
193169 96.399 -76.742 49.886
193170 96.389 -76.730 49.871
193171 96.380 -76.720 49.856
193172 96.370 -76.709 49.840
193173 96.361 -76.698 49.825
193174 96.351 -76.690 49.809
193175 96.339 -76.726 49.790
193176 96.375 -76.778 49.767
193177 96.441 -76.841 49.739
193178 96.494 -76.881 49.709
193179 96.509 -76.856 49.680
193180 96.526 -76.833 49.652
193181 96.565 -76.827 49.625
193182 96.602 -76.821 49.598
193183 96.641 -76.815 49.571
193184 96.677 -76.814 49.543
193185 96.703 -76.835 49.513
193186 96.730 -76.856 49.482
193187 96.756 -76.878 49.450
193188 96.783 -76.899 49.418
193189 96.813 -76.924 49.385
193190 96.842 -76.944 49.352
193191 96.860 -76.951 49.321
193192 96.873 -76.962 49.291
193193 96.885 -76.974 49.261
193194 96.897 -76.985 49.232
193195 96.910 -76.997 49.203
193196 96.923 -77.007 49.177
193197 96.936 -77.016 49.151
193198 96.949 -77.025 49.125
193199 96.962 -77.035 49.102
193200 96.975 -77.044 49.077
193201 96.989 -77.053 49.052
193202 97.002 -77.063 49.027
193203 97.015 -77.072 48.999
193204 97.029 -77.081 48.972
193205 97.062 -77.106 48.944
193206 97.085 -77.117 48.916
193207 97.123 -77.137 48.889
193208 97.164 -77.159 48.862
193209 97.205 -77.181 48.835
193210 97.245 -77.202 48.811
193211 97.286 -77.224 48.786
193212 97.326 -77.246 48.761
193213 97.371 -77.264 48.737
193214 97.384 -77.265 48.715
193215 97.395 -77.260 48.692
193216 97.406 -77.256 48.670
193217 97.417 -77.251 48.648
193218 97.428 -77.247 48.624
193219 97.439 -77.243 48.601
193220 97.450 -77.238 48.578
193221 97.461 -77.234 48.553
193222 97.472 -77.229 48.529
193223 97.483 -77.225 48.505
193224 97.577 -77.270 48.480
193225 97.623 -77.299 48.454
193226 97.586 -77.290 48.427
193227 97.549 -77.281 48.400
193228 97.532 -77.274 48.372
193229 97.559 -77.272 48.343
193230 97.586 -77.270 48.314
193231 97.612 -77.268 48.285
193232 97.640 -77.266 48.256
193233 97.667 -77.264 48.226
193234 97.693 -77.262 48.197
193235 97.721 -77.260 48.168
193236 97.747 -77.258 48.140
193237 97.775 -77.257 48.112
193238 97.801 -77.254 48.082
193239 97.828 -77.253 48.053
193240 97.856 -77.254 48.024
193241 97.886 -77.244 47.994
193242 97.917 -77.232 47.963
193243 97.947 -77.220 47.931
193244 97.973 -77.206 47.897
193245 97.998 -77.191 47.866
193246 98.023 -77.176 47.834
193247 98.048 -77.162 47.801
193248 98.073 -77.147 47.769
193249 98.098 -77.133 47.738
193250 98.123 -77.118 47.709
193251 98.148 -77.103 47.681
193252 98.173 -77.089 47.652
193253 98.199 -77.075 47.622
193254 98.224 -77.059 47.593
193255 98.248 -77.045 47.565
193256 98.273 -77.031 47.537
193257 98.286 -77.027 47.509
193258 98.302 -77.048 47.480
193259 98.317 -77.069 47.452
193260 98.333 -77.091 47.423
193261 98.348 -77.111 47.395
193262 98.364 -77.132 47.368
193263 98.379 -77.153 47.340
193264 98.407 -77.161 47.314
193265 98.442 -77.166 47.289
193266 98.458 -77.175 47.267
193267 98.473 -77.186 47.244
193268 98.489 -77.196 47.222
193269 98.510 -77.207 47.200
193270 98.543 -77.219 47.176
193271 98.575 -77.232 47.153
193272 98.608 -77.245 47.130
193273 98.641 -77.257 47.107
193274 98.673 -77.270 47.087
193275 98.705 -77.285 47.066
193276 98.734 -77.302 47.045
193277 98.760 -77.319 47.024
193278 98.776 -77.332 47.002
193279 98.793 -77.345 46.980
193280 98.810 -77.359 46.959
193281 98.828 -77.373 46.937
193282 98.843 -77.384 46.915
193283 98.854 -77.384 46.892
193284 98.865 -77.385 46.869
193285 98.876 -77.385 46.846
193286 98.887 -77.385 46.823
193287 98.898 -77.385 46.800
193288 98.909 -77.385 46.777
193289 98.919 -77.385 46.753
193290 98.930 -77.386 46.729
193291 98.941 -77.386 46.704
193292 98.952 -77.386 46.679
193293 98.973 -77.378 46.656
193294 98.994 -77.370 46.633
193295 99.015 -77.361 46.609
193296 99.037 -77.352 46.586
193297 99.049 -77.355 46.562
193298 99.060 -77.360 46.539
193299 99.071 -77.364 46.517
193300 99.082 -77.368 46.494
193301 99.094 -77.379 46.471
193302 99.108 -77.405 46.447
193303 99.122 -77.431 46.424
193304 99.136 -77.457 46.400
193305 99.143 -77.478 46.376
193306 99.142 -77.491 46.353
193307 99.140 -77.504 46.330
193308 99.138 -77.517 46.308
193309 99.136 -77.531 46.284
193310 99.134 -77.544 46.261
193311 99.133 -77.558 46.237
193312 99.145 -77.561 46.212
193313 99.167 -77.558 46.189
193314 99.189 -77.555 46.166
193315 99.211 -77.552 46.143
193316 99.233 -77.549 46.121
193317 99.255 -77.546 46.098
193318 99.279 -77.544 46.076
193319 99.319 -77.548 46.054
193320 99.359 -77.551 46.034
193321 99.399 -77.555 46.016
193322 99.438 -77.559 45.998
193323 99.478 -77.563 45.980
193324 99.518 -77.567 45.963
193325 99.558 -77.570 45.945
193326 99.598 -77.574 45.928
193327 99.638 -77.578 45.913
193328 99.678 -77.582 45.899
193329 99.717 -77.586 45.885
193330 99.753 -77.587 45.872
193331 99.783 -77.586 45.858
193332 99.816 -77.599 45.843
193333 99.832 -77.611 45.830
193334 99.846 -77.623 45.816
193335 99.859 -77.635 45.802
193336 99.873 -77.647 45.788
193337 99.886 -77.659 45.774
193338 99.900 -77.671 45.759
193339 99.913 -77.683 45.745
193340 99.926 -77.696 45.732
193341 99.940 -77.708 45.719
193342 99.954 -77.720 45.707
193343 99.967 -77.732 45.695
193344 99.988 -77.752 45.683
193345 100.012 -77.779 45.672
193346 100.037 -77.805 45.663
193347 100.063 -77.830 45.653
193348 100.085 -77.845 45.643
193349 100.104 -77.860 45.633
193350 100.124 -77.875 45.623
193351 100.143 -77.890 45.614
193352 100.162 -77.906 45.604
193353 100.182 -77.920 45.594
193354 100.203 -77.934 45.584
193355 100.223 -77.947 45.572
193356 100.244 -77.961 45.559
193357 100.264 -77.975 45.545
193358 100.284 -77.989 45.532
193359 100.305 -78.003 45.519
193360 100.325 -78.016 45.506
193361 100.346 -78.031 45.492
193362 100.366 -78.044 45.478
193363 100.387 -78.058 45.464
193364 100.402 -78.068 45.448
193365 100.416 -78.075 45.434
193366 100.429 -78.083 45.419
193367 100.444 -78.087 45.403
193368 100.462 -78.090 45.388
193369 100.480 -78.091 45.373
193370 100.498 -78.094 45.357
193371 100.516 -78.095 45.342
193372 100.534 -78.097 45.328
193373 100.554 -78.103 45.315
193374 100.577 -78.113 45.301
193375 100.601 -78.124 45.287
193376 100.634 -78.121 45.272
193377 100.671 -78.114 45.258
193378 100.707 -78.107 45.245
193379 100.733 -78.109 45.231
193380 100.759 -78.113 45.218
193381 100.798 -78.126 45.204
193382 100.837 -78.138 45.191
193383 100.876 -78.151 45.177
193384 100.915 -78.163 45.163
193385 100.955 -78.176 45.149
193386 100.993 -78.188 45.134
193387 101.033 -78.200 45.119
193388 101.074 -78.207 45.104
193389 101.115 -78.215 45.089
193390 101.156 -78.222 45.076
193391 101.196 -78.230 45.061
193392 101.237 -78.238 45.047
193393 101.254 -78.253 45.033
193394 101.265 -78.272 45.019
193395 101.277 -78.290 45.005
193396 101.288 -78.308 44.990
193397 101.301 -78.326 44.975
193398 101.317 -78.338 44.960
193399 101.334 -78.350 44.945
193400 101.351 -78.361 44.930
193401 101.368 -78.364 44.914
193402 101.381 -78.363 44.898
193403 101.394 -78.361 44.884
193404 101.409 -78.363 44.868
193405 101.445 -78.384 44.852
193406 101.480 -78.404 44.833
193407 101.515 -78.426 44.816
193408 101.549 -78.448 44.799
193409 101.575 -78.473 44.782
193410 101.603 -78.499 44.765
193411 101.629 -78.525 44.748
193412 101.656 -78.551 44.731
193413 101.685 -78.573 44.715
193414 101.714 -78.593 44.699
193415 101.744 -78.614 44.683
193416 101.773 -78.634 44.668
193417 101.804 -78.655 44.651
193418 101.849 -78.678 44.636
193419 101.894 -78.701 44.621
193420 101.938 -78.730 44.605
193421 101.981 -78.761 44.591
193422 102.024 -78.793 44.576
193423 102.067 -78.825 44.562
193424 102.110 -78.857 44.549
193425 102.153 -78.889 44.534
193426 102.184 -78.905 44.520
193427 102.203 -78.906 44.507
193428 102.223 -78.908 44.493
193429 102.243 -78.910 44.478
193430 102.262 -78.912 44.464
193431 102.281 -78.913 44.451
193432 102.301 -78.915 44.436
193433 102.320 -78.916 44.422
193434 102.340 -78.918 44.409
193435 102.359 -78.919 44.395
193436 102.379 -78.921 44.381
193437 102.398 -78.923 44.367
193720 93.511 -76.326 54.069
193721 93.530 -76.329 54.073
193722 93.548 -76.332 54.078
193723 93.566 -76.336 54.083
193724 93.585 -76.339 54.088
193725 93.603 -76.342 54.093
193726 93.621 -76.344 54.100
193727 93.640 -76.347 54.108
193728 93.658 -76.350 54.114
193729 93.677 -76.353 54.120
193730 93.695 -76.356 54.128
193731 93.713 -76.359 54.137
193732 93.731 -76.361 54.149
193733 93.750 -76.364 54.164
193734 93.767 -76.360 54.180
193735 93.783 -76.344 54.185
193736 93.800 -76.328 54.189
193737 93.816 -76.312 54.192
193738 93.832 -76.296 54.196
193739 93.848 -76.282 54.198
193740 93.865 -76.283 54.201
193741 93.884 -76.289 54.203
193742 93.903 -76.295 54.203
193743 93.922 -76.301 54.204
193744 93.941 -76.307 54.207
193745 93.960 -76.313 54.210
193746 93.979 -76.319 54.217
193747 93.997 -76.323 54.222
193748 94.014 -76.321 54.225
193749 94.032 -76.317 54.222
193750 94.052 -76.316 54.218
193751 94.071 -76.315 54.217
193752 94.091 -76.314 54.216
193753 94.111 -76.313 54.216
193754 94.131 -76.312 54.217
193755 94.150 -76.310 54.215
193756 94.170 -76.310 54.209
193757 94.191 -76.306 54.202
193758 94.214 -76.297 54.193
193759 94.237 -76.289 54.180
193760 94.260 -76.280 54.164
193761 94.280 -76.277 54.147
193762 94.300 -76.276 54.124
193763 94.320 -76.274 54.098
193764 94.339 -76.273 54.072
193765 94.358 -76.271 54.043
193766 94.376 -76.269 54.014
193767 94.394 -76.267 53.984
193768 94.413 -76.265 53.954
193769 94.431 -76.263 53.922
193770 94.449 -76.263 53.890
193771 94.468 -76.261 53.859
193772 94.485 -76.260 53.829
193773 94.498 -76.260 53.797
193774 94.512 -76.260 53.766
193775 94.526 -76.260 53.736
193776 94.539 -76.259 53.703
193777 94.552 -76.259 53.672
193778 94.566 -76.259 53.644
193779 94.578 -76.267 53.617
193780 94.590 -76.276 53.589
193781 94.602 -76.284 53.561
193782 94.614 -76.293 53.530
193783 94.626 -76.302 53.494
193784 94.643 -76.313 53.457
193785 94.665 -76.325 53.420
193786 94.676 -76.330 53.384
193787 94.687 -76.335 53.350
193788 94.698 -76.340 53.318
193789 94.709 -76.345 53.286
193790 94.721 -76.349 53.252
193791 94.734 -76.353 53.219
193792 94.747 -76.356 53.185
193793 94.759 -76.360 53.150
193794 94.772 -76.364 53.114
193795 94.784 -76.367 53.077
193796 94.797 -76.370 53.036
193797 94.809 -76.374 52.992
193798 94.822 -76.378 52.947
193799 94.834 -76.382 52.903
193800 94.847 -76.385 52.858
193801 94.859 -76.389 52.814
193802 94.872 -76.392 52.772
193803 94.885 -76.396 52.729
193804 94.897 -76.399 52.687
193805 94.910 -76.403 52.646
193806 94.922 -76.407 52.601
193807 94.935 -76.411 52.556
193808 94.948 -76.414 52.509
193809 94.953 -76.423 52.461
193810 94.977 -76.430 52.408
193811 95.001 -76.436 52.356
193812 95.027 -76.442 52.302
193813 95.051 -76.448 52.246
193814 95.076 -76.455 52.189
193815 95.101 -76.461 52.136
193816 95.147 -76.458 52.087
193817 95.197 -76.454 52.041
193818 95.246 -76.448 51.998
193819 95.286 -76.448 51.955
193820 95.310 -76.452 51.912
193821 95.332 -76.457 51.864
193822 95.355 -76.461 51.818
193823 95.379 -76.466 51.773
193824 95.402 -76.471 51.731
193825 95.425 -76.476 51.689
193826 95.448 -76.480 51.647
193827 95.471 -76.485 51.606
193828 95.494 -76.489 51.566
193829 95.517 -76.494 51.529
193830 95.540 -76.499 51.493
193831 95.563 -76.504 51.458
193832 95.584 -76.512 51.426
193833 95.603 -76.522 51.396
193834 95.622 -76.532 51.367
193835 95.641 -76.542 51.338
193836 95.668 -76.548 51.305
193837 95.697 -76.551 51.268
193838 95.722 -76.555 51.227
193839 95.745 -76.560 51.183
193840 95.767 -76.564 51.136
193841 95.791 -76.569 51.091
193842 95.814 -76.575 51.046
193843 95.838 -76.581 51.002
193844 95.863 -76.587 50.958
193845 95.887 -76.593 50.915
193846 95.911 -76.599 50.872
193847 95.935 -76.605 50.829
193848 95.959 -76.608 50.785
193849 95.981 -76.608 50.738
193850 96.002 -76.613 50.690
193851 96.023 -76.618 50.645
193852 96.043 -76.623 50.596
193853 96.063 -76.628 50.545
193854 96.084 -76.633 50.492
193855 96.104 -76.636 50.438
193856 96.125 -76.640 50.384
193857 96.145 -76.643 50.332
193858 96.172 -76.649 50.284
193859 96.194 -76.657 50.239
193860 96.214 -76.664 50.197
193861 96.233 -76.671 50.157
193862 96.253 -76.678 50.120
193863 96.272 -76.685 50.087
193864 96.292 -76.692 50.058
193865 96.312 -76.699 50.032
193866 96.332 -76.707 50.007
193867 96.352 -76.715 49.984
193868 96.372 -76.723 49.963
193869 96.391 -76.729 49.943
193870 96.374 -76.716 49.924
193871 96.357 -76.704 49.910
193872 96.340 -76.691 49.895
193873 96.323 -76.678 49.879
193874 96.306 -76.666 49.864
193875 96.292 -76.654 49.847
193876 96.292 -76.660 49.826
193877 96.358 -76.723 49.800
193878 96.424 -76.786 49.771
193879 96.484 -76.839 49.740
193880 96.504 -76.817 49.709
193881 96.541 -76.812 49.680
193882 96.580 -76.806 49.652
193883 96.618 -76.799 49.624
193884 96.656 -76.794 49.595
193885 96.694 -76.788 49.566
193886 96.724 -76.802 49.535
193887 96.750 -76.824 49.503
193888 96.777 -76.845 49.471
193889 96.803 -76.866 49.437
193890 96.830 -76.887 49.403
193891 96.857 -76.909 49.371
193892 96.876 -76.922 49.340
193893 96.889 -76.931 49.309
193894 96.903 -76.940 49.279
193895 96.916 -76.950 49.249
193896 96.930 -76.959 49.221
193897 96.943 -76.968 49.193
193898 96.956 -76.977 49.167
193899 96.969 -76.986 49.143
193900 96.982 -76.996 49.120
193901 96.996 -77.005 49.095
193902 97.009 -77.014 49.071
193903 97.022 -77.023 49.046
193904 97.037 -77.033 49.019
193905 97.069 -77.058 48.991
193906 97.101 -77.084 48.963
193907 97.124 -77.094 48.934
193908 97.153 -77.103 48.907
193909 97.193 -77.125 48.879
193910 97.234 -77.147 48.853
193911 97.275 -77.169 48.829
193912 97.315 -77.191 48.804
193913 97.356 -77.213 48.778
193914 97.389 -77.248 48.754
193915 97.408 -77.259 48.731
193916 97.417 -77.252 48.707
193917 97.426 -77.245 48.685
193918 97.435 -77.238 48.662
193919 97.444 -77.232 48.638
193920 97.454 -77.226 48.613
193921 97.463 -77.219 48.589
193922 97.472 -77.212 48.565
193923 97.481 -77.205 48.540
193924 97.492 -77.201 48.515
193925 97.532 -77.224 48.489
193926 97.568 -77.251 48.462
193927 97.570 -77.262 48.435
193928 97.533 -77.253 48.407
193929 97.546 -77.249 48.379
193930 97.573 -77.248 48.350
193931 97.600 -77.245 48.321
193932 97.628 -77.243 48.291
193933 97.654 -77.241 48.262
193934 97.681 -77.239 48.232
193935 97.708 -77.238 48.203
193936 97.735 -77.235 48.175
193937 97.762 -77.234 48.146
193938 97.789 -77.232 48.118
193939 97.816 -77.229 48.089
193940 97.843 -77.228 48.061
193941 97.870 -77.226 48.032
193942 97.900 -77.218 48.002
193943 97.928 -77.204 47.972
193944 97.953 -77.190 47.940
193945 97.978 -77.175 47.906
193946 98.003 -77.160 47.875
193947 98.028 -77.146 47.843
193948 98.053 -77.132 47.811
193949 98.078 -77.117 47.779
193950 98.103 -77.102 47.748
193951 98.129 -77.089 47.718
193952 98.154 -77.076 47.688
193953 98.179 -77.064 47.660
193954 98.205 -77.051 47.630
193955 98.230 -77.039 47.601
193956 98.256 -77.026 47.572
193957 98.282 -77.014 47.544
193958 98.306 -77.010 47.515
193959 98.314 -77.041 47.486
193960 98.334 -77.058 47.458
193961 98.353 -77.074 47.429
193962 98.369 -77.095 47.401
193963 98.384 -77.116 47.374
193964 98.400 -77.137 47.347
193965 98.427 -77.144 47.321
193966 98.442 -77.154 47.297
193967 98.458 -77.164 47.274
193968 98.474 -77.174 47.252
193969 98.489 -77.185 47.229
193970 98.508 -77.195 47.207
193971 98.541 -77.208 47.184
193972 98.574 -77.221 47.161
193973 98.606 -77.233 47.138
193974 98.639 -77.246 47.115
193975 98.672 -77.259 47.094
193976 98.705 -77.271 47.074
193977 98.737 -77.284 47.053
193978 98.763 -77.296 47.032
193979 98.780 -77.310 47.010
193980 98.797 -77.323 46.988
193981 98.814 -77.337 46.967
193982 98.831 -77.351 46.944
193983 98.848 -77.364 46.922
193984 98.865 -77.378 46.898
193985 98.879 -77.385 46.875
193986 98.890 -77.385 46.851
193987 98.901 -77.385 46.829
193988 98.912 -77.386 46.806
193989 98.923 -77.386 46.782
193990 98.934 -77.386 46.758
193991 98.944 -77.387 46.734
193992 98.955 -77.387 46.708
193993 98.967 -77.387 46.683
193994 98.988 -77.379 46.660
193995 99.009 -77.370 46.636
193996 99.030 -77.362 46.613
193997 99.045 -77.360 46.591
193998 99.056 -77.365 46.568
193999 99.067 -77.370 46.544
194000 99.078 -77.374 46.521
194001 99.089 -77.378 46.498
194002 99.100 -77.382 46.475
194003 99.111 -77.389 46.452
194004 99.125 -77.414 46.428
194005 99.139 -77.440 46.404
194006 99.153 -77.466 46.381
194007 99.155 -77.483 46.358
194008 99.153 -77.496 46.335
194009 99.152 -77.509 46.312
194010 99.150 -77.523 46.289
194011 99.148 -77.536 46.265
194012 99.146 -77.549 46.241
194013 99.167 -77.547 46.216
194014 99.189 -77.544 46.193
194015 99.211 -77.541 46.169
194016 99.233 -77.538 46.146
194017 99.256 -77.535 46.123
194018 99.278 -77.532 46.101
194019 99.300 -77.529 46.080
194020 99.324 -77.526 46.057
194021 99.364 -77.530 46.037
194022 99.404 -77.534 46.019
194023 99.443 -77.538 46.001
194024 99.484 -77.542 45.983
194025 99.523 -77.545 45.965
194026 99.563 -77.549 45.947
194027 99.603 -77.553 45.929
194028 99.643 -77.557 45.914
194029 99.683 -77.561 45.900
194030 99.723 -77.564 45.886
194031 99.763 -77.568 45.872
194032 99.802 -77.572 45.858
194033 99.839 -77.580 45.844
194034 99.852 -77.592 45.829
194035 99.865 -77.604 45.816
194036 99.879 -77.616 45.802
194037 99.893 -77.628 45.787
194038 99.906 -77.640 45.773
194039 99.919 -77.652 45.758
194040 99.933 -77.664 45.745
194041 99.946 -77.677 45.732
194042 99.960 -77.689 45.719
194043 99.973 -77.701 45.707
194044 99.987 -77.713 45.696
194045 100.003 -77.729 45.684
194046 100.028 -77.755 45.673
194047 100.053 -77.780 45.663
194048 100.077 -77.805 45.652
194049 100.107 -77.820 45.642
194050 100.129 -77.834 45.633
194051 100.149 -77.849 45.624
194052 100.168 -77.865 45.614
194053 100.188 -77.880 45.604
194054 100.207 -77.895 45.594
194055 100.226 -77.910 45.583
194056 100.246 -77.926 45.571
194057 100.265 -77.941 45.557
194058 100.284 -77.956 45.544
194059 100.303 -77.971 45.531
194060 100.323 -77.987 45.517
194061 100.342 -78.002 45.504
194062 100.362 -78.017 45.490
194063 100.381 -78.032 45.475
194064 100.401 -78.047 45.461
194065 100.421 -78.053 45.446
194066 100.438 -78.056 45.431
194067 100.457 -78.057 45.416
194068 100.474 -78.059 45.400
194069 100.492 -78.061 45.385
194070 100.510 -78.062 45.370
194071 100.528 -78.065 45.355
194072 100.546 -78.066 45.340
194073 100.566 -78.072 45.326
194074 100.590 -78.083 45.312
194075 100.613 -78.094 45.298
194076 100.635 -78.104 45.284
194077 100.660 -78.112 45.270
194078 100.696 -78.106 45.256
194079 100.733 -78.099 45.243
194080 100.762 -78.098 45.229
194081 100.785 -78.101 45.216
194082 100.818 -78.109 45.203
194083 100.857 -78.122 45.190
194084 100.896 -78.134 45.176
194085 100.935 -78.147 45.162
194086 100.974 -78.159 45.148
194087 101.013 -78.172 45.134
194088 101.052 -78.184 45.119
194089 101.092 -78.197 45.104
194090 101.131 -78.208 45.090
194091 101.172 -78.216 45.076
194092 101.212 -78.223 45.061
194093 101.252 -78.232 45.047
194094 101.263 -78.250 45.033
194095 101.275 -78.268 45.018
194096 101.287 -78.285 45.005
194097 101.305 -78.296 44.990
194098 101.321 -78.308 44.975
194099 101.338 -78.320 44.960
194100 101.355 -78.332 44.945
194101 101.372 -78.344 44.929
194102 101.389 -78.356 44.914
194103 101.409 -78.362 44.899
194104 101.422 -78.361 44.884
194105 101.435 -78.360 44.868
194106 101.461 -78.371 44.852
194107 101.496 -78.392 44.834
194108 101.532 -78.413 44.817
194109 101.561 -78.438 44.800
194110 101.588 -78.463 44.783
194111 101.615 -78.489 44.766
194112 101.643 -78.513 44.749
194113 101.672 -78.533 44.732
194114 101.702 -78.554 44.716
194115 101.731 -78.574 44.699
194116 101.761 -78.595 44.684
194117 101.790 -78.615 44.669
194118 101.820 -78.636 44.652
194119 101.866 -78.659 44.637
194120 101.911 -78.682 44.621
194121 101.956 -78.705 44.606
194122 101.999 -78.736 44.591
194123 102.043 -78.767 44.577
194124 102.086 -78.799 44.563
194125 102.129 -78.831 44.549
194126 102.172 -78.863 44.535
194127 102.209 -78.887 44.521
194128 102.229 -78.889 44.506
194129 102.248 -78.891 44.492
194130 102.268 -78.893 44.478
194131 102.287 -78.894 44.464
194132 102.307 -78.896 44.451
194133 102.326 -78.897 44.436
194134 102.345 -78.899 44.422
194135 102.365 -78.900 44.408
194136 102.384 -78.902 44.394
194137 102.404 -78.904 44.380
194138 102.423 -78.905 44.366
194139 102.443 -78.907 44.352
194422 93.521 -76.306 54.065
194423 93.540 -76.309 54.070
194424 93.558 -76.312 54.075
194425 93.577 -76.315 54.081
194426 93.595 -76.318 54.087
194427 93.613 -76.320 54.094
194428 93.632 -76.323 54.102
194429 93.650 -76.326 54.109
194430 93.669 -76.329 54.115
194431 93.687 -76.332 54.123
194432 93.705 -76.336 54.131
194433 93.723 -76.338 54.143
194434 93.742 -76.341 54.156
194435 93.755 -76.353 54.172
194436 93.766 -76.336 54.178
194437 93.785 -76.316 54.183
194438 93.805 -76.296 54.188
194439 93.825 -76.276 54.194
194440 93.842 -76.261 54.198
194441 93.861 -76.267 54.202
194442 93.880 -76.273 54.206
194443 93.899 -76.279 54.206
194444 93.918 -76.285 54.207
194445 93.937 -76.291 54.210
194446 93.956 -76.296 54.214
194447 93.975 -76.302 54.221
194448 93.994 -76.308 54.226
194449 94.011 -76.306 54.229
194450 94.031 -76.305 54.225
194451 94.051 -76.304 54.223
194452 94.070 -76.303 54.222
194453 94.090 -76.302 54.222
194454 94.110 -76.301 54.224
194455 94.130 -76.300 54.225
194456 94.150 -76.299 54.224
194457 94.170 -76.298 54.220
194458 94.190 -76.295 54.212
194459 94.213 -76.286 54.200
194460 94.236 -76.278 54.187
194461 94.259 -76.270 54.173
194462 94.281 -76.261 54.154
194463 94.301 -76.260 54.131
194464 94.321 -76.258 54.105
194465 94.340 -76.257 54.078
194466 94.358 -76.254 54.050
194467 94.376 -76.253 54.019
194468 94.394 -76.251 53.988
194469 94.412 -76.249 53.959
194470 94.431 -76.248 53.927
194471 94.449 -76.247 53.896
194472 94.466 -76.246 53.865
194473 94.479 -76.245 53.834
194474 94.493 -76.245 53.803
194475 94.506 -76.245 53.773
194476 94.520 -76.245 53.742
194477 94.533 -76.245 53.710
194478 94.546 -76.244 53.681
194479 94.560 -76.244 53.654
194480 94.572 -76.249 53.628
194481 94.584 -76.258 53.602
194482 94.596 -76.267 53.575
194483 94.608 -76.276 53.545
194484 94.624 -76.286 53.509
194485 94.645 -76.298 53.473
194486 94.665 -76.309 53.436
194487 94.676 -76.314 53.399
194488 94.687 -76.320 53.365
194489 94.698 -76.325 53.334
194490 94.709 -76.329 53.302
194491 94.720 -76.335 53.269
194492 94.732 -76.339 53.235
194493 94.745 -76.343 53.200
194494 94.757 -76.347 53.165
194495 94.770 -76.350 53.128
194496 94.782 -76.354 53.089
194497 94.795 -76.357 53.047
194498 94.808 -76.361 53.002
194499 94.820 -76.364 52.957
194500 94.833 -76.368 52.912
194501 94.845 -76.372 52.868
194502 94.858 -76.376 52.825
194503 94.871 -76.379 52.785
194504 94.883 -76.383 52.743
194505 94.895 -76.386 52.701
194506 94.908 -76.390 52.659
194507 94.920 -76.393 52.615
194508 94.933 -76.397 52.570
194509 94.946 -76.401 52.523
194510 94.952 -76.409 52.476
194511 94.967 -76.420 52.426
194512 94.991 -76.426 52.374
194513 95.016 -76.433 52.321
194514 95.041 -76.439 52.265
194515 95.065 -76.445 52.208
194516 95.095 -76.450 52.154
194517 95.144 -76.445 52.104
194518 95.193 -76.440 52.058
194519 95.243 -76.436 52.016
194520 95.290 -76.432 51.973
194521 95.314 -76.436 51.928
194522 95.336 -76.441 51.879
194523 95.359 -76.445 51.831
194524 95.383 -76.451 51.785
194525 95.406 -76.455 51.742
194526 95.429 -76.460 51.699
194527 95.452 -76.464 51.659
194528 95.475 -76.469 51.619
194529 95.498 -76.474 51.581
194530 95.521 -76.479 51.546
194531 95.544 -76.483 51.511
194532 95.567 -76.488 51.477
194533 95.591 -76.492 51.445
194534 95.614 -76.497 51.416
194535 95.635 -76.504 51.388
194536 95.655 -76.514 51.360
194537 95.677 -76.522 51.330
194538 95.704 -76.525 51.294
194539 95.727 -76.530 51.253
194540 95.751 -76.536 51.209
194541 95.775 -76.542 51.161
194542 95.799 -76.548 51.114
194543 95.823 -76.554 51.068
194544 95.847 -76.560 51.023
194545 95.872 -76.566 50.978
194546 95.896 -76.572 50.935
194547 95.919 -76.578 50.892
194548 95.944 -76.584 50.848
194549 95.967 -76.585 50.803
194550 95.990 -76.584 50.756
194551 96.011 -76.588 50.709
194552 96.032 -76.593 50.663
194553 96.052 -76.599 50.615
194554 96.073 -76.603 50.565
194555 96.093 -76.607 50.514
194556 96.113 -76.611 50.462
194557 96.133 -76.614 50.411
194558 96.154 -76.618 50.362
194559 96.174 -76.622 50.315
194560 96.194 -76.630 50.270
194561 96.214 -76.638 50.228
194562 96.234 -76.646 50.189
194563 96.254 -76.654 50.153
194564 96.274 -76.661 50.121
194565 96.294 -76.670 50.092
194566 96.314 -76.677 50.065
194567 96.334 -76.686 50.041
194568 96.354 -76.694 50.019
194569 96.373 -76.701 49.998
194570 96.392 -76.705 49.979
194571 96.395 -76.702 49.962
194572 96.378 -76.689 49.947
194573 96.361 -76.677 49.933
194574 96.343 -76.664 49.917
194575 96.326 -76.651 49.901
194576 96.316 -76.641 49.881
194577 96.330 -76.654 49.858
194578 96.383 -76.705 49.831
194579 96.436 -76.756 49.801
194580 96.487 -76.803 49.770
194581 96.524 -76.797 49.739
194582 96.561 -76.791 49.708
194583 96.599 -76.785 49.679
194584 96.637 -76.779 49.649
194585 96.674 -76.773 49.618
194586 96.712 -76.768 49.588
194587 96.744 -76.771 49.557
194588 96.770 -76.793 49.524
194589 96.797 -76.815 49.491
194590 96.824 -76.837 49.457
194591 96.851 -76.859 49.423
194592 96.878 -76.882 49.389
194593 96.895 -76.884 49.357
194594 96.910 -76.892 49.326
194595 96.923 -76.901 49.296
194596 96.936 -76.910 49.265
194597 96.950 -76.919 49.237
194598 96.963 -76.928 49.209
194599 96.976 -76.937 49.184
194600 96.990 -76.947 49.160
194601 97.003 -76.956 49.137
194602 97.016 -76.965 49.113
194603 97.029 -76.975 49.089
194604 97.044 -76.985 49.065
194605 97.077 -77.010 49.038
194606 97.109 -77.035 49.010
194607 97.141 -77.061 48.982
194608 97.164 -77.071 48.953
194609 97.183 -77.071 48.926
194610 97.223 -77.093 48.899
194611 97.264 -77.115 48.872
194612 97.305 -77.137 48.846
194613 97.345 -77.159 48.821
194614 97.377 -77.196 48.796
194615 97.409 -77.232 48.772
194616 97.436 -77.260 48.748
194617 97.445 -77.254 48.725
194618 97.454 -77.247 48.701
194619 97.464 -77.240 48.677
194620 97.473 -77.233 48.651
194621 97.482 -77.226 48.626
194622 97.491 -77.219 48.601
194623 97.500 -77.213 48.576
194624 97.507 -77.203 48.550
194625 97.507 -77.185 48.524
194626 97.511 -77.192 48.498
194627 97.515 -77.204 48.471
194628 97.537 -77.225 48.443
194629 97.534 -77.226 48.415
194630 97.561 -77.224 48.387
194631 97.588 -77.222 48.358
194632 97.615 -77.220 48.328
194633 97.642 -77.219 48.299
194634 97.669 -77.216 48.269
194635 97.696 -77.215 48.240
194636 97.723 -77.213 48.211
194637 97.750 -77.211 48.183
194638 97.777 -77.210 48.154
194639 97.804 -77.209 48.126
194640 97.830 -77.208 48.098
194641 97.858 -77.207 48.070
194642 97.885 -77.204 48.043
194643 97.910 -77.200 48.012
194644 97.935 -77.187 47.982
194645 97.961 -77.175 47.951
194646 97.987 -77.162 47.918
194647 98.012 -77.150 47.887
194648 98.037 -77.137 47.855
194649 98.063 -77.125 47.823
194650 98.088 -77.112 47.791
194651 98.114 -77.100 47.760
194652 98.140 -77.087 47.728
194653 98.165 -77.075 47.698
194654 98.191 -77.062 47.669
194655 98.216 -77.050 47.640
194656 98.242 -77.037 47.610
194657 98.267 -77.025 47.581
194658 98.293 -77.015 47.552
194659 98.317 -77.015 47.522
194660 98.342 -77.015 47.492
194661 98.353 -77.038 47.463
194662 98.357 -77.076 47.435
194663 98.366 -77.106 47.408
194664 98.385 -77.122 47.381
194665 98.416 -77.124 47.354
194666 98.432 -77.130 47.329
194667 98.442 -77.143 47.305
194668 98.458 -77.153 47.282
194669 98.474 -77.163 47.259
194670 98.489 -77.173 47.237
194671 98.507 -77.184 47.215
194672 98.539 -77.196 47.192
194673 98.572 -77.209 47.169
194674 98.605 -77.222 47.146
194675 98.638 -77.235 47.124
194676 98.670 -77.247 47.102
194677 98.703 -77.260 47.081
194678 98.744 -77.269 47.060
194679 98.774 -77.281 47.039
194680 98.788 -77.292 47.018
194681 98.803 -77.303 46.997
194682 98.817 -77.314 46.975
194683 98.835 -77.328 46.952
194684 98.852 -77.341 46.929
194685 98.869 -77.355 46.906
194686 98.886 -77.369 46.882
194687 98.903 -77.382 46.859
194688 98.916 -77.386 46.836
194689 98.926 -77.387 46.812
194690 98.937 -77.387 46.788
194691 98.948 -77.388 46.763
194692 98.959 -77.388 46.738
194693 98.970 -77.388 46.713
194694 98.981 -77.388 46.687
194695 99.002 -77.379 46.664
194696 99.023 -77.371 46.641
194697 99.042 -77.367 46.618
194698 99.053 -77.370 46.596
194699 99.063 -77.375 46.572
194700 99.074 -77.379 46.547
194701 99.085 -77.384 46.523
194702 99.096 -77.389 46.501
194703 99.107 -77.392 46.478
194704 99.118 -77.397 46.455
194705 99.129 -77.401 46.431
194706 99.142 -77.423 46.407
194707 99.156 -77.449 46.385
194708 99.169 -77.474 46.361
194709 99.167 -77.487 46.338
194710 99.165 -77.501 46.315
194711 99.163 -77.514 46.291
194712 99.162 -77.527 46.267
194713 99.167 -77.536 46.243
194714 99.190 -77.533 46.218
194715 99.212 -77.530 46.195
194716 99.234 -77.526 46.171
194717 99.256 -77.523 46.148
194718 99.278 -77.520 46.125
194719 99.300 -77.517 46.103
194720 99.323 -77.514 46.082
194721 99.345 -77.511 46.060
194722 99.369 -77.509 46.040
194723 99.409 -77.513 46.021
194724 99.449 -77.517 46.003
194725 99.489 -77.520 45.985
194726 99.528 -77.524 45.966
194727 99.568 -77.528 45.948
194728 99.608 -77.532 45.931
194729 99.648 -77.536 45.916
194730 99.688 -77.539 45.901
194731 99.728 -77.543 45.886
194732 99.768 -77.547 45.872
194733 99.809 -77.553 45.858
194734 99.854 -77.566 45.844
194735 99.874 -77.577 45.829
194736 99.887 -77.589 45.815
194737 99.901 -77.600 45.801
194738 99.914 -77.611 45.787
194739 99.927 -77.624 45.773
194740 99.940 -77.635 45.759
194741 99.953 -77.646 45.746
194742 99.966 -77.658 45.733
194743 99.979 -77.670 45.719
194744 99.993 -77.682 45.708
194745 100.006 -77.694 45.697
194746 100.019 -77.706 45.685
194747 100.044 -77.730 45.674
194748 100.068 -77.755 45.664
194749 100.092 -77.779 45.653
194750 100.121 -77.796 45.643
194751 100.151 -77.810 45.633
194752 100.174 -77.824 45.624
194753 100.193 -77.840 45.614
194754 100.213 -77.855 45.603
194755 100.232 -77.870 45.593
194756 100.252 -77.885 45.582
194757 100.271 -77.900 45.569
194758 100.290 -77.915 45.556
194759 100.310 -77.931 45.543
194760 100.329 -77.946 45.528
194761 100.348 -77.961 45.515
194762 100.367 -77.976 45.501
194763 100.387 -77.991 45.486
194764 100.409 -78.010 45.472
194765 100.434 -78.032 45.457
194766 100.451 -78.025 45.443
194767 100.469 -78.027 45.428
194768 100.486 -78.028 45.412
194769 100.504 -78.031 45.397
194770 100.522 -78.032 45.382
194771 100.540 -78.034 45.367
194772 100.558 -78.036 45.353
194773 100.578 -78.041 45.338
194774 100.601 -78.052 45.323
194775 100.624 -78.062 45.309
194776 100.647 -78.073 45.295
194777 100.671 -78.084 45.282
194778 100.694 -78.094 45.268
194779 100.722 -78.097 45.255
194780 100.759 -78.091 45.241
194781 100.791 -78.087 45.227
194782 100.815 -78.090 45.213
194783 100.838 -78.093 45.201
194784 100.876 -78.105 45.188
194785 100.916 -78.118 45.175
194786 100.955 -78.130 45.161
194787 100.994 -78.143 45.148
194788 101.033 -78.155 45.133
194789 101.072 -78.168 45.118
194790 101.111 -78.181 45.103
194791 101.150 -78.193 45.089
194792 101.189 -78.206 45.075
194793 101.229 -78.216 45.060
194794 101.261 -78.227 45.045
194795 101.275 -78.243 45.031
194796 101.292 -78.255 45.017
194797 101.309 -78.266 45.004
194798 101.326 -78.279 44.989
194799 101.342 -78.291 44.974
194800 101.359 -78.302 44.960
194801 101.376 -78.314 44.944
194802 101.393 -78.326 44.929
194803 101.410 -78.338 44.913
194804 101.426 -78.350 44.898
194805 101.444 -78.362 44.883
194806 101.463 -78.359 44.867
194807 101.478 -78.359 44.851
194808 101.513 -78.380 44.835
194809 101.547 -78.401 44.817
194810 101.574 -78.427 44.800
194811 101.601 -78.453 44.784
194812 101.630 -78.474 44.767
194813 101.659 -78.495 44.751
194814 101.689 -78.515 44.734
194815 101.718 -78.535 44.717
194816 101.748 -78.556 44.700
194817 101.777 -78.576 44.684
194818 101.807 -78.597 44.668
194819 101.837 -78.617 44.652
194820 101.882 -78.640 44.637
194821 101.927 -78.664 44.622
194822 101.973 -78.686 44.607
194823 102.018 -78.710 44.592
194824 102.061 -78.742 44.577
194825 102.104 -78.774 44.562
194826 102.147 -78.805 44.549
194827 102.190 -78.837 44.535
194828 102.234 -78.869 44.521
194829 102.254 -78.872 44.507
194830 102.273 -78.874 44.492
194831 102.293 -78.875 44.478
194832 102.312 -78.877 44.464
194833 102.332 -78.878 44.450
194834 102.351 -78.880 44.436
194835 102.371 -78.881 44.422
194836 102.390 -78.883 44.407
194837 102.409 -78.884 44.393
194838 102.429 -78.886 44.379
194839 102.449 -78.888 44.366
194840 102.468 -78.890 44.352
195123 93.513 -76.282 54.056
195124 93.532 -76.285 54.061
195125 93.550 -76.289 54.067
195126 93.568 -76.292 54.073
195127 93.587 -76.295 54.080
195128 93.605 -76.297 54.088
195129 93.624 -76.300 54.096
195130 93.642 -76.303 54.103
195131 93.660 -76.306 54.110
195132 93.679 -76.309 54.118
195133 93.697 -76.312 54.125
195134 93.715 -76.315 54.136
195135 93.733 -76.318 54.149
195136 93.747 -76.332 54.163
195137 93.755 -76.329 54.171
195138 93.763 -76.312 54.178
195139 93.782 -76.292 54.185
195140 93.802 -76.273 54.192
195141 93.830 -76.259 54.198
195142 93.865 -76.254 54.204
195143 93.884 -76.260 54.209
195144 93.902 -76.266 54.209
195145 93.920 -76.271 54.210
195146 93.938 -76.277 54.212
195147 93.957 -76.282 54.217
195148 93.975 -76.288 54.224
195149 93.993 -76.293 54.230
195150 94.004 -76.289 54.233
195151 94.023 -76.287 54.230
195152 94.042 -76.285 54.229
195153 94.061 -76.284 54.229
195154 94.079 -76.282 54.229
195155 94.099 -76.280 54.231
195156 94.117 -76.279 54.233
195157 94.137 -76.277 54.232
195158 94.166 -76.277 54.228
195159 94.193 -76.276 54.220
195160 94.214 -76.270 54.210
195161 94.236 -76.265 54.196
195162 94.258 -76.259 54.181
195163 94.280 -76.251 54.163
195164 94.302 -76.244 54.139
195165 94.322 -76.242 54.112
195166 94.340 -76.241 54.083
195167 94.358 -76.238 54.054
195168 94.376 -76.237 54.023
195169 94.394 -76.235 53.992
195170 94.412 -76.233 53.961
195171 94.431 -76.232 53.930
195172 94.447 -76.231 53.899
195173 94.460 -76.231 53.869
195174 94.473 -76.230 53.838
195175 94.487 -76.230 53.807
195176 94.500 -76.229 53.777
195177 94.513 -76.229 53.748
195178 94.527 -76.229 53.718
195179 94.540 -76.229 53.691
195180 94.554 -76.229 53.666
195181 94.566 -76.232 53.641
195182 94.578 -76.242 53.614
195183 94.591 -76.250 53.586
195184 94.604 -76.260 53.557
195185 94.625 -76.271 53.523
195186 94.646 -76.283 53.486
195187 94.666 -76.294 53.450
195188 94.677 -76.298 53.414
195189 94.688 -76.304 53.379
195190 94.699 -76.309 53.348
195191 94.710 -76.314 53.317
195192 94.721 -76.319 53.284
195193 94.732 -76.324 53.249
195194 94.744 -76.329 53.214
195195 94.756 -76.333 53.177
195196 94.768 -76.337 53.139
195197 94.780 -76.341 53.098
195198 94.793 -76.344 53.056
195199 94.806 -76.348 53.012
195200 94.818 -76.351 52.965
195201 94.831 -76.355 52.922
195202 94.843 -76.358 52.879
195203 94.856 -76.362 52.838
195204 94.868 -76.366 52.799
195205 94.881 -76.370 52.758
195206 94.894 -76.373 52.714
195207 94.906 -76.376 52.671
195208 94.919 -76.380 52.627
195209 94.931 -76.384 52.582
195210 94.944 -76.387 52.537
195211 94.951 -76.395 52.492
195212 94.956 -76.410 52.443
195213 94.981 -76.416 52.391
195214 95.006 -76.423 52.338
195215 95.031 -76.429 52.283
195216 95.055 -76.436 52.226
195217 95.091 -76.436 52.172
195218 95.140 -76.432 52.121
195219 95.190 -76.427 52.076
195220 95.240 -76.423 52.033
195221 95.289 -76.418 51.990
195222 95.318 -76.420 51.943
195223 95.341 -76.425 51.894
195224 95.364 -76.430 51.845
195225 95.387 -76.435 51.798
195226 95.410 -76.439 51.755
195227 95.433 -76.444 51.713
195228 95.456 -76.448 51.673
195229 95.479 -76.454 51.636
195230 95.502 -76.458 51.599
195231 95.525 -76.463 51.565
195232 95.548 -76.467 51.530
195233 95.571 -76.472 51.497
195234 95.595 -76.477 51.464
195235 95.618 -76.482 51.435
195236 95.641 -76.486 51.407
195237 95.664 -76.491 51.380
195238 95.687 -76.495 51.351
195239 95.711 -76.502 51.316
195240 95.735 -76.508 51.276
195241 95.759 -76.514 51.231
195242 95.783 -76.520 51.182
195243 95.808 -76.527 51.134
195244 95.832 -76.533 51.087
195245 95.856 -76.539 51.041
195246 95.880 -76.545 50.997
195247 95.904 -76.551 50.953
195248 95.928 -76.557 50.910
195249 95.953 -76.563 50.866
195250 95.975 -76.562 50.821
195251 95.999 -76.561 50.775
195252 96.020 -76.564 50.729
195253 96.041 -76.569 50.682
195254 96.061 -76.574 50.635
195255 96.082 -76.578 50.587
195256 96.102 -76.582 50.538
195257 96.122 -76.585 50.489
195258 96.142 -76.589 50.441
195259 96.155 -76.597 50.392
195260 96.172 -76.603 50.345
195261 96.194 -76.609 50.301
195262 96.216 -76.616 50.260
195263 96.236 -76.624 50.222
195264 96.256 -76.632 50.185
195265 96.276 -76.640 50.154
195266 96.296 -76.649 50.125
195267 96.316 -76.656 50.099
195268 96.336 -76.665 50.075
195269 96.356 -76.672 50.054
195270 96.374 -76.677 50.034
195271 96.393 -76.680 50.016
195272 96.411 -76.685 49.999
195273 96.398 -76.674 49.985
195274 96.381 -76.661 49.970
195275 96.364 -76.649 49.954
195276 96.361 -76.641 49.935
195277 96.369 -76.637 49.914
195278 96.379 -76.658 49.889
195279 96.432 -76.708 49.860
195280 96.483 -76.757 49.829
195281 96.528 -76.787 49.797
195282 96.566 -76.781 49.765
195283 96.603 -76.775 49.732
195284 96.641 -76.769 49.702
195285 96.678 -76.763 49.671
195286 96.710 -76.744 49.640
195287 96.737 -76.740 49.609
195288 96.764 -76.744 49.577
195289 96.792 -76.766 49.544
195290 96.818 -76.789 49.510
195291 96.845 -76.811 49.476
195292 96.872 -76.834 49.441
195293 96.895 -76.847 49.407
195294 96.912 -76.850 49.374
195295 96.930 -76.853 49.343
195296 96.943 -76.862 49.312
195297 96.957 -76.871 49.282
195298 96.970 -76.880 49.253
195299 96.983 -76.889 49.226
195300 96.997 -76.898 49.200
195301 97.010 -76.907 49.176
195302 97.023 -76.916 49.154
195303 97.036 -76.926 49.131
195304 97.052 -76.937 49.108
195305 97.084 -76.962 49.083
195306 97.116 -76.987 49.056
195307 97.149 -77.012 49.029
195308 97.181 -77.038 49.001
195309 97.204 -77.048 48.972
195310 97.218 -77.044 48.945
195311 97.253 -77.060 48.918
195312 97.294 -77.081 48.891
195313 97.333 -77.106 48.865
195314 97.365 -77.143 48.839
195315 97.397 -77.180 48.814
195316 97.429 -77.217 48.789
195317 97.461 -77.254 48.765
195318 97.474 -77.255 48.741
195319 97.483 -77.248 48.717
195320 97.492 -77.241 48.692
195321 97.501 -77.235 48.665
195322 97.510 -77.228 48.639
195323 97.519 -77.220 48.613
195324 97.519 -77.202 48.588
195325 97.519 -77.184 48.562
195326 97.520 -77.166 48.535
195327 97.524 -77.175 48.508
195328 97.528 -77.188 48.481
195329 97.533 -77.201 48.452
195330 97.550 -77.207 48.424
195331 97.575 -77.209 48.396
195332 97.602 -77.208 48.366
195333 97.629 -77.207 48.337
195334 97.656 -77.206 48.308
195335 97.682 -77.204 48.278
195336 97.710 -77.204 48.248
195337 97.736 -77.202 48.220
195338 97.764 -77.201 48.191
195339 97.790 -77.200 48.163
195340 97.817 -77.199 48.136
195341 97.844 -77.197 48.108
195342 97.873 -77.192 48.081
195343 97.903 -77.182 48.053
195344 97.925 -77.185 48.023
195345 97.947 -77.185 47.992
195346 97.972 -77.173 47.961
195347 97.998 -77.160 47.931
195348 98.023 -77.148 47.899
195349 98.049 -77.135 47.868
195350 98.074 -77.123 47.836
195351 98.100 -77.110 47.805
195352 98.125 -77.098 47.772
195353 98.151 -77.085 47.740
195354 98.176 -77.073 47.709
195355 98.202 -77.060 47.679
195356 98.227 -77.048 47.650
195357 98.253 -77.035 47.620
195358 98.279 -77.023 47.589
195359 98.303 -77.019 47.559
195360 98.327 -77.019 47.529
195361 98.352 -77.019 47.499
195362 98.377 -77.019 47.468
195363 98.393 -77.034 47.442
195364 98.397 -77.071 47.415
195365 98.401 -77.109 47.388
195366 98.438 -77.101 47.361
195367 98.450 -77.109 47.337
195368 98.442 -77.131 47.314
195369 98.458 -77.141 47.291
195370 98.474 -77.151 47.267
195371 98.489 -77.162 47.245
195372 98.505 -77.172 47.223
195373 98.538 -77.185 47.200
195374 98.571 -77.197 47.176
195375 98.603 -77.210 47.154
195376 98.637 -77.222 47.132
195377 98.669 -77.235 47.109
195378 98.708 -77.246 47.088
195379 98.758 -77.253 47.067
195380 98.786 -77.264 47.046
195381 98.799 -77.278 47.025
195382 98.814 -77.289 47.004
195383 98.829 -77.300 46.982
195384 98.843 -77.311 46.959
195385 98.858 -77.322 46.937
195386 98.872 -77.333 46.913
195387 98.889 -77.346 46.889
195388 98.907 -77.360 46.865
195389 98.923 -77.373 46.842
195390 98.940 -77.387 46.818
195391 98.952 -77.388 46.793
195392 98.962 -77.389 46.767
195393 98.974 -77.389 46.742
195394 98.984 -77.389 46.716
195395 98.995 -77.389 46.692
195396 99.016 -77.381 46.668
195397 99.038 -77.372 46.645
195398 99.049 -77.376 46.622
195399 99.060 -77.381 46.599
195400 99.070 -77.385 46.575
195401 99.081 -77.389 46.550
195402 99.092 -77.394 46.526
195403 99.103 -77.398 46.503
195404 99.114 -77.403 46.480
195405 99.125 -77.407 46.457
195406 99.136 -77.411 46.433
195407 99.147 -77.416 46.410
195408 99.160 -77.432 46.387
195409 99.174 -77.458 46.364
195410 99.181 -77.479 46.341
195411 99.179 -77.492 46.317
195412 99.178 -77.505 46.293
195413 99.175 -77.519 46.269
195414 99.190 -77.521 46.245
195415 99.212 -77.518 46.219
195416 99.234 -77.515 46.196
195417 99.257 -77.512 46.173
195418 99.279 -77.509 46.150
195419 99.301 -77.506 46.126
195420 99.323 -77.503 46.104
195421 99.345 -77.500 46.083
195422 99.367 -77.497 46.063
195423 99.390 -77.494 46.043
195424 99.414 -77.492 46.024
195425 99.454 -77.495 46.004
195426 99.494 -77.499 45.985
195427 99.534 -77.503 45.967
195428 99.574 -77.507 45.949
195429 99.614 -77.511 45.933
195430 99.653 -77.514 45.918
195431 99.693 -77.518 45.903
195432 99.733 -77.522 45.888
195433 99.773 -77.526 45.873
195434 99.817 -77.538 45.859
195435 99.862 -77.550 45.844
195436 99.896 -77.563 45.830
195437 99.909 -77.574 45.816
195438 99.922 -77.586 45.801
195439 99.935 -77.597 45.787
195440 99.948 -77.608 45.774
195441 99.962 -77.620 45.759
195442 99.975 -77.631 45.747
195443 99.988 -77.642 45.734
195444 100.001 -77.655 45.721
195445 100.014 -77.666 45.710
195446 100.027 -77.677 45.698
195447 100.040 -77.689 45.687
195448 100.059 -77.705 45.676
195449 100.083 -77.729 45.665
195450 100.108 -77.753 45.654
195451 100.135 -77.773 45.643
195452 100.165 -77.787 45.633
195453 100.196 -77.799 45.623
195454 100.218 -77.815 45.613
195455 100.238 -77.830 45.602
195456 100.257 -77.845 45.591
195457 100.277 -77.860 45.580
195458 100.296 -77.875 45.568
195459 100.316 -77.890 45.554
195460 100.335 -77.906 45.540
195461 100.354 -77.921 45.526
195462 100.373 -77.936 45.512
195463 100.393 -77.951 45.497
195464 100.417 -77.973 45.482
195465 100.442 -77.995 45.467
195466 100.465 -78.007 45.453
195467 100.481 -77.996 45.439
195468 100.498 -77.998 45.424
195469 100.516 -78.000 45.409
195470 100.534 -78.002 45.394
195471 100.552 -78.003 45.380
195472 100.570 -78.006 45.365
195473 100.590 -78.011 45.351
195474 100.613 -78.022 45.336
195475 100.636 -78.032 45.321
195476 100.659 -78.043 45.306
195477 100.683 -78.053 45.293
195478 100.705 -78.064 45.280
195479 100.728 -78.074 45.265
195480 100.752 -78.084 45.252
195481 100.785 -78.082 45.238
195482 100.820 -78.077 45.225
195483 100.844 -78.079 45.212
195484 100.867 -78.082 45.200
195485 100.897 -78.088 45.187
195486 100.935 -78.101 45.174
195487 100.975 -78.113 45.161
195488 101.013 -78.126 45.147
195489 101.053 -78.138 45.133
195490 101.092 -78.151 45.117
195491 101.131 -78.164 45.103
195492 101.170 -78.176 45.088
195493 101.209 -78.189 45.074
195494 101.247 -78.203 45.059
195495 101.279 -78.218 45.044
195496 101.296 -78.228 45.030
195497 101.313 -78.240 45.016
195498 101.330 -78.250 45.002
195499 101.346 -78.261 44.988
195500 101.363 -78.272 44.973
195501 101.380 -78.285 44.959
195502 101.397 -78.297 44.943
195503 101.414 -78.308 44.927
195504 101.431 -78.320 44.911
195505 101.448 -78.332 44.896
195506 101.465 -78.344 44.881
195507 101.481 -78.356 44.865
195508 101.504 -78.357 44.849
195509 101.529 -78.367 44.833
195510 101.559 -78.391 44.816
195511 101.588 -78.414 44.800
195512 101.617 -78.435 44.783
195513 101.647 -78.455 44.766
195514 101.676 -78.476 44.750
195515 101.706 -78.496 44.733
195516 101.735 -78.517 44.717
195517 101.765 -78.537 44.700
195518 101.794 -78.557 44.684
195519 101.824 -78.578 44.669
195520 101.853 -78.598 44.654
195521 101.899 -78.622 44.638
195522 101.944 -78.645 44.623
195523 101.989 -78.668 44.608
195524 102.035 -78.691 44.592
195525 102.080 -78.716 44.577
195526 102.123 -78.748 44.563
195527 102.166 -78.780 44.549
195528 102.209 -78.811 44.535
195529 102.252 -78.843 44.521
195530 102.279 -78.855 44.507
195531 102.299 -78.856 44.492
195532 102.318 -78.858 44.478
195533 102.337 -78.859 44.463
195534 102.357 -78.861 44.448
195535 102.376 -78.862 44.435
195536 102.396 -78.864 44.421
195537 102.415 -78.865 44.407
195538 102.435 -78.867 44.392
195539 102.454 -78.869 44.378
195540 102.473 -78.871 44.364
195541 102.493 -78.872 44.349
195542 102.513 -78.874 44.335
195825 93.524 -76.262 54.053
195826 93.542 -76.265 54.059
195827 93.560 -76.268 54.066
195828 93.579 -76.271 54.073
195829 93.597 -76.274 54.082
195830 93.616 -76.276 54.091
195831 93.634 -76.279 54.098
195832 93.652 -76.282 54.105
195833 93.671 -76.285 54.113
195834 93.689 -76.289 54.121
195835 93.707 -76.292 54.132
195836 93.725 -76.295 54.144
195837 93.739 -76.307 54.157
195838 93.745 -76.323 54.165
195839 93.754 -76.305 54.173
195840 93.763 -76.289 54.181
195841 93.779 -76.270 54.189
195842 93.815 -76.260 54.196
195843 93.854 -76.254 54.203
195844 93.890 -76.249 54.207
195845 93.908 -76.254 54.209
195846 93.926 -76.260 54.210
195847 93.945 -76.266 54.213
195848 93.963 -76.271 54.219
195849 93.981 -76.276 54.226
195850 93.985 -76.270 54.232
195851 93.989 -76.264 54.237
195852 94.005 -76.262 54.237
195853 94.024 -76.260 54.237
195854 94.043 -76.258 54.236
195855 94.062 -76.257 54.236
195856 94.083 -76.255 54.239
195857 94.112 -76.256 54.242
195858 94.141 -76.256 54.241
195859 94.170 -76.256 54.236
195860 94.196 -76.254 54.229
195861 94.217 -76.249 54.219
195862 94.239 -76.244 54.206
195863 94.261 -76.238 54.189
195864 94.282 -76.233 54.171
195865 94.303 -76.228 54.146
195866 94.322 -76.226 54.117
195867 94.340 -76.223 54.087
195868 94.358 -76.222 54.056
195869 94.377 -76.220 54.025
195870 94.395 -76.218 53.994
195871 94.413 -76.216 53.963
195872 94.429 -76.215 53.932
195873 94.441 -76.216 53.900
195874 94.454 -76.216 53.870
195875 94.468 -76.215 53.839
195876 94.481 -76.215 53.809
195877 94.494 -76.215 53.780
195878 94.507 -76.214 53.753
195879 94.521 -76.214 53.724
195880 94.535 -76.213 53.699
195881 94.548 -76.213 53.675
195882 94.561 -76.215 53.651
195883 94.573 -76.224 53.625
195884 94.585 -76.233 53.597
195885 94.606 -76.245 53.566
195886 94.627 -76.257 53.533
195887 94.648 -76.268 53.496
195888 94.666 -76.278 53.460
195889 94.677 -76.283 53.424
195890 94.688 -76.288 53.390
195891 94.699 -76.293 53.359
195892 94.710 -76.298 53.327
195893 94.721 -76.304 53.294
195894 94.733 -76.308 53.257
195895 94.744 -76.314 53.220
195896 94.755 -76.319 53.183
195897 94.766 -76.323 53.145
195898 94.779 -76.327 53.104
195899 94.791 -76.331 53.061
195900 94.804 -76.335 53.017
195901 94.817 -76.338 52.974
195902 94.829 -76.342 52.933
195903 94.841 -76.345 52.893
195904 94.854 -76.349 52.853
195905 94.866 -76.352 52.813
195906 94.879 -76.356 52.772
195907 94.892 -76.360 52.729
195908 94.904 -76.364 52.685
195909 94.917 -76.367 52.640
195910 94.929 -76.370 52.596
195911 94.942 -76.374 52.551
195912 94.950 -76.381 52.506
195913 94.946 -76.399 52.457
195914 94.971 -76.406 52.405
195915 94.995 -76.412 52.352
195916 95.020 -76.419 52.297
195917 95.045 -76.425 52.242
195918 95.088 -76.423 52.189
195919 95.137 -76.419 52.138
195920 95.187 -76.414 52.092
195921 95.236 -76.410 52.050
195922 95.285 -76.404 52.005
195923 95.322 -76.405 51.959
195924 95.345 -76.410 51.909
195925 95.368 -76.414 51.860
195926 95.391 -76.419 51.813
195927 95.414 -76.423 51.769
195928 95.437 -76.428 51.729
195929 95.460 -76.433 51.691
195930 95.483 -76.438 51.654
195931 95.507 -76.442 51.619
195932 95.529 -76.447 51.586
195933 95.552 -76.451 51.552
195934 95.576 -76.457 51.517
195935 95.599 -76.461 51.482
195936 95.622 -76.466 51.452
195937 95.645 -76.470 51.424
195938 95.668 -76.475 51.397
195939 95.693 -76.477 51.369
195940 95.720 -76.493 51.333
195941 95.744 -76.492 51.293
195942 95.768 -76.493 51.249
195943 95.792 -76.499 51.201
195944 95.816 -76.505 51.152
195945 95.840 -76.511 51.104
195946 95.865 -76.517 51.057
195947 95.889 -76.524 51.013
195948 95.913 -76.530 50.970
195949 95.937 -76.536 50.927
195950 95.961 -76.541 50.884
195951 95.984 -76.540 50.840
195952 96.007 -76.539 50.795
195953 96.029 -76.539 50.749
195954 96.050 -76.544 50.702
195955 96.071 -76.549 50.656
195956 96.091 -76.552 50.609
195957 96.110 -76.557 50.562
195958 96.122 -76.564 50.517
195959 96.135 -76.571 50.470
195960 96.148 -76.579 50.423
195961 96.171 -76.585 50.376
195962 96.193 -76.591 50.332
195963 96.215 -76.597 50.291
195964 96.238 -76.603 50.253
195965 96.258 -76.611 50.218
195966 96.278 -76.619 50.186
195967 96.298 -76.627 50.158
195968 96.318 -76.635 50.133
195969 96.337 -76.643 50.110
195970 96.357 -76.649 50.088
195971 96.375 -76.653 50.069
195972 96.394 -76.657 50.052
195973 96.412 -76.661 50.036
195974 96.419 -76.659 50.022
195975 96.402 -76.647 50.006
195976 96.404 -76.641 49.988
195977 96.413 -76.636 49.967
195978 96.422 -76.633 49.944
195979 96.432 -76.660 49.917
195980 96.480 -76.711 49.887
195981 96.525 -76.742 49.855
195982 96.570 -76.771 49.823
195983 96.607 -76.765 49.789
195984 96.645 -76.759 49.755
195985 96.676 -76.730 49.723
195986 96.703 -76.723 49.691
195987 96.731 -76.719 49.660
195988 96.758 -76.715 49.629
195989 96.786 -76.718 49.597
195990 96.812 -76.740 49.563
195991 96.839 -76.762 49.528
195992 96.866 -76.784 49.494
195993 96.893 -76.807 49.458
195994 96.912 -76.813 49.424
195995 96.930 -76.815 49.391
195996 96.947 -76.818 49.359
195997 96.964 -76.822 49.328
195998 96.977 -76.831 49.297
195999 96.990 -76.840 49.269
196000 97.004 -76.850 49.241
196001 97.017 -76.859 49.216
196002 97.030 -76.868 49.192
196003 97.043 -76.877 49.170
196004 97.060 -76.889 49.149
196005 97.092 -76.914 49.126
196006 97.124 -76.939 49.101
196007 97.156 -76.965 49.075
196008 97.188 -76.990 49.048
196009 97.203 -77.011 49.020
196010 97.223 -77.024 48.991
196011 97.258 -77.021 48.964
196012 97.283 -77.027 48.936
196013 97.321 -77.053 48.910
196014 97.353 -77.091 48.885
196015 97.385 -77.128 48.858
196016 97.417 -77.165 48.832
196017 97.449 -77.202 48.806
196018 97.480 -77.239 48.782
196019 97.502 -77.256 48.758
196020 97.511 -77.250 48.733
196021 97.520 -77.243 48.707
196022 97.530 -77.236 48.680
196023 97.531 -77.219 48.653
196024 97.532 -77.201 48.627
196025 97.532 -77.183 48.601
196026 97.532 -77.165 48.574
196027 97.533 -77.147 48.547
196028 97.537 -77.159 48.520
196029 97.542 -77.172 48.492
196030 97.548 -77.183 48.462
196031 97.571 -77.188 48.434
196032 97.593 -77.193 48.406
196033 97.615 -77.197 48.377
196034 97.642 -77.197 48.347
196035 97.669 -77.195 48.318
196036 97.696 -77.194 48.288
196037 97.723 -77.193 48.259
196038 97.749 -77.192 48.229
196039 97.777 -77.191 48.201
196040 97.803 -77.190 48.174
196041 97.830 -77.188 48.147
196042 97.860 -77.180 48.120
196043 97.891 -77.170 48.092
196044 97.921 -77.160 48.063
196045 97.942 -77.164 48.032
196046 97.962 -77.169 48.001
196047 97.984 -77.171 47.971
196048 98.009 -77.159 47.942
196049 98.034 -77.146 47.911
196050 98.060 -77.134 47.879
196051 98.085 -77.121 47.847
196052 98.111 -77.109 47.814
196053 98.137 -77.096 47.782
196054 98.162 -77.084 47.750
196055 98.187 -77.071 47.718
196056 98.213 -77.059 47.688
196057 98.239 -77.046 47.659
196058 98.264 -77.034 47.628
196059 98.290 -77.024 47.598
196060 98.314 -77.023 47.568
196061 98.339 -77.023 47.538
196062 98.363 -77.023 47.508
196063 98.387 -77.023 47.478
196064 98.412 -77.023 47.450
196065 98.433 -77.028 47.423
196066 98.437 -77.066 47.396
196067 98.459 -77.078 47.370
196068 98.468 -77.088 47.346
196069 98.460 -77.111 47.323
196070 98.458 -77.130 47.300
196071 98.474 -77.140 47.277
196072 98.489 -77.150 47.255
196073 98.505 -77.160 47.233
196074 98.537 -77.173 47.210
196075 98.570 -77.185 47.186
196076 98.602 -77.198 47.164
196077 98.635 -77.211 47.141
196078 98.671 -77.222 47.118
196079 98.721 -77.229 47.096
196080 98.771 -77.237 47.074
196081 98.797 -77.248 47.053
196082 98.811 -77.262 47.031
196083 98.826 -77.275 47.010
196084 98.840 -77.285 46.988
196085 98.855 -77.297 46.966
196086 98.869 -77.307 46.944
196087 98.884 -77.319 46.920
196088 98.899 -77.329 46.896
196089 98.913 -77.341 46.872
196090 98.928 -77.351 46.848
196091 98.944 -77.364 46.823
196092 98.961 -77.378 46.797
196093 98.977 -77.389 46.771
196094 98.988 -77.389 46.745
196095 98.999 -77.389 46.720
196096 99.009 -77.390 46.695
196097 99.030 -77.382 46.672
196098 99.045 -77.382 46.649
196099 99.056 -77.386 46.626
196100 99.067 -77.391 46.602
196101 99.078 -77.395 46.578
196102 99.088 -77.399 46.553
196103 99.099 -77.404 46.529
196104 99.111 -77.408 46.505
196105 99.122 -77.413 46.483
196106 99.132 -77.417 46.459
196107 99.143 -77.421 46.435
196108 99.154 -77.426 46.411
196109 99.165 -77.430 46.388
196110 99.177 -77.442 46.365
196111 99.191 -77.467 46.343
196112 99.193 -77.484 46.319
196113 99.191 -77.497 46.294
196114 99.191 -77.510 46.270
196115 99.213 -77.507 46.246
196116 99.235 -77.504 46.219
196117 99.257 -77.501 46.197
196118 99.279 -77.498 46.173
196119 99.301 -77.495 46.150
196120 99.324 -77.492 46.127
196121 99.346 -77.489 46.106
196122 99.368 -77.486 46.085
196123 99.390 -77.483 46.064
196124 99.412 -77.479 46.046
196125 99.434 -77.476 46.026
196126 99.459 -77.474 46.007
196127 99.499 -77.478 45.988
196128 99.539 -77.482 45.969
196129 99.579 -77.486 45.951
196130 99.619 -77.489 45.935
196131 99.658 -77.493 45.919
196132 99.698 -77.497 45.904
196133 99.738 -77.501 45.889
196134 99.781 -77.510 45.874
196135 99.826 -77.522 45.859
196136 99.870 -77.535 45.844
196137 99.915 -77.548 45.830
196138 99.931 -77.559 45.816
196139 99.944 -77.570 45.801
196140 99.957 -77.582 45.788
196141 99.970 -77.594 45.775
196142 99.983 -77.605 45.761
196143 99.996 -77.617 45.749
196144 100.009 -77.628 45.736
196145 100.022 -77.639 45.723
196146 100.036 -77.651 45.711
196147 100.049 -77.662 45.700
196148 100.063 -77.674 45.688
196149 100.078 -77.685 45.677
196150 100.098 -77.703 45.666
196151 100.123 -77.727 45.654
196152 100.148 -77.749 45.643
196153 100.179 -77.763 45.632
196154 100.210 -77.776 45.621
196155 100.240 -77.790 45.611
196156 100.263 -77.804 45.601
196157 100.283 -77.819 45.590
196158 100.302 -77.834 45.578
196159 100.321 -77.849 45.565
196160 100.341 -77.865 45.551
196161 100.360 -77.880 45.537
196162 100.379 -77.895 45.523
196163 100.401 -77.913 45.509
196164 100.426 -77.935 45.493
196165 100.451 -77.957 45.477
196166 100.475 -77.979 45.463
196167 100.495 -77.980 45.449
196168 100.510 -77.968 45.435
196169 100.528 -77.969 45.420
196170 100.546 -77.971 45.405
196171 100.564 -77.973 45.391
196172 100.582 -77.975 45.377
196173 100.602 -77.981 45.362
196174 100.625 -77.990 45.347
196175 100.648 -78.001 45.333
196176 100.671 -78.012 45.318
196177 100.694 -78.022 45.303
196178 100.717 -78.033 45.290
196179 100.741 -78.044 45.277
196180 100.764 -78.054 45.263
196181 100.787 -78.065 45.249
196182 100.811 -78.074 45.236
196183 100.847 -78.067 45.224
196184 100.873 -78.069 45.211
196185 100.896 -78.072 45.198
196186 100.920 -78.074 45.186
196187 100.955 -78.084 45.174
196188 100.994 -78.097 45.160
196189 101.034 -78.109 45.146
196190 101.072 -78.122 45.132
196191 101.112 -78.134 45.116
196192 101.150 -78.147 45.102
196193 101.190 -78.159 45.087
196194 101.227 -78.176 45.072
196195 101.263 -78.193 45.057
196196 101.299 -78.210 45.043
196197 101.316 -78.220 45.029
196198 101.333 -78.231 45.014
196199 101.350 -78.242 45.000
196200 101.367 -78.253 44.985
196201 101.384 -78.263 44.971
196202 101.401 -78.275 44.956
196203 101.418 -78.285 44.941
196204 101.434 -78.297 44.925
196205 101.452 -78.307 44.909
196206 101.468 -78.318 44.894
196207 101.485 -78.329 44.879
196208 101.502 -78.340 44.863
196209 101.519 -78.351 44.847
196210 101.543 -78.357 44.832
196211 101.575 -78.375 44.815
196212 101.605 -78.395 44.798
196213 101.634 -78.416 44.781
196214 101.664 -78.436 44.765
196215 101.693 -78.457 44.749
196216 101.723 -78.477 44.733
196217 101.752 -78.498 44.717
196218 101.781 -78.519 44.700
196219 101.811 -78.539 44.684
196220 101.840 -78.560 44.669
196221 101.870 -78.580 44.655
196222 101.915 -78.603 44.640
196223 101.961 -78.626 44.624
196224 102.006 -78.649 44.609
196225 102.051 -78.672 44.592
196226 102.097 -78.695 44.578
196227 102.141 -78.721 44.563
196228 102.184 -78.753 44.549
196229 102.228 -78.786 44.535
196230 102.270 -78.818 44.520
196231 102.304 -78.837 44.505
196232 102.324 -78.839 44.490
196233 102.343 -78.840 44.475
196234 102.363 -78.842 44.461
196235 102.382 -78.843 44.447
196236 102.401 -78.845 44.433
196237 102.421 -78.846 44.419
196238 102.441 -78.848 44.405
196239 102.460 -78.850 44.390
196240 102.479 -78.852 44.377
196241 102.499 -78.853 44.362
196242 102.518 -78.855 44.348
196243 102.538 -78.856 44.333
196527 93.534 -76.242 54.051
196528 93.552 -76.245 54.058
196529 93.571 -76.248 54.066
196530 93.589 -76.251 54.075
196531 93.608 -76.253 54.085
196532 93.626 -76.256 54.093
196533 93.644 -76.259 54.100
196534 93.662 -76.262 54.108
196535 93.681 -76.265 54.117
196536 93.699 -76.268 54.128
196537 93.717 -76.271 54.139
196538 93.731 -76.283 54.152
196539 93.735 -76.316 54.160
196540 93.745 -76.299 54.168
196541 93.754 -76.282 54.177
196542 93.768 -76.267 54.185
196543 93.801 -76.261 54.193
196544 93.839 -76.254 54.200
196545 93.879 -76.248 54.206
196546 93.914 -76.244 54.209
196547 93.933 -76.249 54.212
196548 93.951 -76.254 54.217
196549 93.969 -76.260 54.223
196550 93.973 -76.254 54.230
196551 93.978 -76.248 54.238
196552 93.982 -76.242 54.245
196553 93.988 -76.236 54.247
196554 94.006 -76.235 54.247
196555 94.029 -76.234 54.246
196556 94.058 -76.234 54.246
196557 94.087 -76.235 54.248
196558 94.116 -76.235 54.249
196559 94.145 -76.235 54.249
196560 94.174 -76.235 54.246
196561 94.199 -76.232 54.238
196562 94.221 -76.227 54.226
196563 94.243 -76.222 54.212
196564 94.264 -76.216 54.196
196565 94.285 -76.211 54.176
196566 94.305 -76.208 54.150
196567 94.323 -76.205 54.120
196568 94.341 -76.204 54.088
196569 94.359 -76.201 54.057
196570 94.377 -76.200 54.025
196571 94.395 -76.198 53.993
196572 94.413 -76.196 53.964
196573 94.430 -76.195 53.932
196574 94.442 -76.195 53.900
196575 94.454 -76.196 53.869
196576 94.465 -76.198 53.839
196577 94.477 -76.198 53.810
196578 94.488 -76.199 53.782
196579 94.502 -76.199 53.756
196580 94.515 -76.199 53.729
196581 94.529 -76.198 53.705
196582 94.542 -76.198 53.682
196583 94.555 -76.198 53.660
196584 94.567 -76.207 53.635
196585 94.586 -76.218 53.606
196586 94.607 -76.229 53.574
196587 94.628 -76.242 53.540
196588 94.649 -76.253 53.503
196589 94.667 -76.262 53.468
196590 94.678 -76.267 53.433
196591 94.689 -76.273 53.400
196592 94.700 -76.278 53.369
196593 94.711 -76.282 53.336
196594 94.722 -76.288 53.302
196595 94.733 -76.293 53.264
196596 94.745 -76.298 53.226
196597 94.756 -76.303 53.188
196598 94.767 -76.308 53.149
196599 94.778 -76.314 53.109
196600 94.789 -76.317 53.067
196601 94.802 -76.321 53.024
196602 94.815 -76.325 52.983
196603 94.827 -76.329 52.944
196604 94.840 -76.332 52.904
196605 94.852 -76.336 52.865
196606 94.865 -76.339 52.825
196607 94.878 -76.343 52.783
196608 94.890 -76.346 52.739
196609 94.903 -76.350 52.695
196610 94.915 -76.354 52.649
196611 94.927 -76.357 52.604
196612 94.940 -76.361 52.560
196613 94.950 -76.367 52.516
196614 94.936 -76.389 52.467
196615 94.961 -76.396 52.416
196616 94.985 -76.402 52.363
196617 95.010 -76.409 52.309
196618 95.035 -76.415 52.257
196619 95.085 -76.411 52.206
196620 95.134 -76.406 52.155
196621 95.183 -76.401 52.108
196622 95.233 -76.396 52.065
196623 95.282 -76.392 52.021
196624 95.326 -76.389 51.975
196625 95.349 -76.394 51.926
196626 95.372 -76.398 51.877
196627 95.395 -76.403 51.831
196628 95.418 -76.408 51.788
196629 95.441 -76.413 51.747
196630 95.464 -76.417 51.708
196631 95.487 -76.422 51.672
196632 95.511 -76.426 51.640
196633 95.533 -76.431 51.608
196634 95.556 -76.436 51.573
196635 95.580 -76.441 51.538
196636 95.603 -76.445 51.502
196637 95.626 -76.450 51.471
196638 95.649 -76.455 51.440
196639 95.674 -76.458 51.410
196640 95.699 -76.459 51.380
196641 95.730 -76.492 51.344
196642 95.754 -76.490 51.305
196643 95.777 -76.489 51.260
196644 95.801 -76.488 51.214
196645 95.825 -76.487 51.167
196646 95.849 -76.490 51.120
196647 95.873 -76.496 51.072
196648 95.897 -76.502 51.030
196649 95.921 -76.508 50.988
196650 95.946 -76.514 50.945
196651 95.969 -76.518 50.901
196652 95.992 -76.517 50.856
196653 96.015 -76.517 50.811
196654 96.038 -76.516 50.766
196655 96.059 -76.520 50.721
196656 96.078 -76.524 50.674
196657 96.090 -76.532 50.629
196658 96.103 -76.539 50.584
196659 96.116 -76.546 50.542
196660 96.128 -76.554 50.497
196661 96.147 -76.561 50.452
196662 96.169 -76.567 50.405
196663 96.191 -76.573 50.361
196664 96.214 -76.579 50.321
196665 96.236 -76.585 50.284
196666 96.258 -76.591 50.249
196667 96.279 -76.597 50.218
196668 96.300 -76.605 50.191
196669 96.320 -76.613 50.166
196670 96.339 -76.621 50.143
196671 96.358 -76.625 50.122
196672 96.376 -76.629 50.103
196673 96.395 -76.633 50.087
196674 96.413 -76.636 50.072
196675 96.431 -76.640 50.057
196676 96.443 -76.647 50.040
196677 96.451 -76.653 50.018
196678 96.459 -76.658 49.995
196679 96.467 -76.664 49.970
196680 96.476 -76.669 49.944
196681 96.520 -76.691 49.913
196682 96.566 -76.726 49.881
196683 96.611 -76.755 49.847
196684 96.641 -76.723 49.813
196685 96.669 -76.705 49.778
196686 96.697 -76.702 49.744
196687 96.724 -76.698 49.711
196688 96.752 -76.694 49.679
196689 96.780 -76.690 49.647
196690 96.807 -76.691 49.615
196691 96.833 -76.714 49.581
196692 96.860 -76.736 49.546
196693 96.887 -76.758 49.511
196694 96.911 -76.776 49.475
196695 96.929 -76.778 49.440
196696 96.947 -76.781 49.407
196697 96.965 -76.783 49.374
196698 96.982 -76.785 49.342
196699 96.997 -76.792 49.312
196700 97.011 -76.801 49.284
196701 97.024 -76.810 49.257
196702 97.037 -76.819 49.231
196703 97.050 -76.828 49.208
196704 97.067 -76.841 49.185
196705 97.099 -76.866 49.164
196706 97.132 -76.891 49.143
196707 97.164 -76.916 49.119
196708 97.174 -76.937 49.093
196709 97.176 -76.956 49.066
196710 97.179 -76.975 49.039
196711 97.181 -76.994 49.010
196712 97.235 -76.994 48.982
196713 97.310 -77.001 48.954
196714 97.344 -77.038 48.928
196715 97.373 -77.075 48.901
196716 97.404 -77.113 48.875
196717 97.437 -77.150 48.848
196718 97.468 -77.187 48.823
196719 97.501 -77.224 48.797
196720 97.530 -77.257 48.773
196721 97.540 -77.251 48.748
196722 97.543 -77.237 48.722
196723 97.544 -77.219 48.695
196724 97.544 -77.201 48.668
196725 97.544 -77.182 48.641
196726 97.545 -77.164 48.614
196727 97.545 -77.147 48.586
196728 97.546 -77.130 48.559
196729 97.550 -77.142 48.532
196730 97.555 -77.155 48.503
196731 97.570 -77.163 48.473
196732 97.591 -77.168 48.445
196733 97.613 -77.173 48.417
196734 97.636 -77.179 48.387
196735 97.658 -77.184 48.358
196736 97.682 -77.185 48.329
196737 97.709 -77.184 48.300
196738 97.736 -77.183 48.270
196739 97.763 -77.182 48.241
196740 97.790 -77.181 48.213
196741 97.817 -77.178 48.185
196742 97.848 -77.168 48.158
196743 97.878 -77.158 48.130
196744 97.909 -77.148 48.101
196745 97.939 -77.138 48.072
196746 97.959 -77.144 48.041
196747 97.980 -77.149 48.011
196748 97.999 -77.154 47.981
196749 98.020 -77.157 47.951
196750 98.045 -77.144 47.920
196751 98.071 -77.132 47.887
196752 98.097 -77.119 47.855
196753 98.122 -77.106 47.822
196754 98.148 -77.094 47.790
196755 98.173 -77.081 47.759
196756 98.199 -77.069 47.727
196757 98.224 -77.056 47.697
196758 98.250 -77.044 47.667
196759 98.276 -77.031 47.637
196760 98.300 -77.028 47.606
196761 98.324 -77.028 47.576
196762 98.349 -77.028 47.546
196763 98.373 -77.028 47.517
196764 98.398 -77.028 47.488
196765 98.423 -77.028 47.459
196766 98.447 -77.027 47.432
196767 98.472 -77.027 47.406
196768 98.482 -77.055 47.381
196769 98.487 -77.067 47.356
196770 98.478 -77.090 47.333
196771 98.469 -77.113 47.310
196772 98.474 -77.128 47.288
196773 98.489 -77.138 47.265
196774 98.505 -77.149 47.242
196775 98.535 -77.161 47.220
196776 98.568 -77.174 47.196
196777 98.601 -77.187 47.174
196778 98.635 -77.199 47.151
196779 98.685 -77.206 47.128
196780 98.735 -77.213 47.103
196781 98.785 -77.220 47.082
196782 98.809 -77.232 47.061
196783 98.823 -77.246 47.038
196784 98.837 -77.260 47.015
196785 98.851 -77.272 46.993
196786 98.866 -77.282 46.971
196787 98.880 -77.294 46.950
196788 98.895 -77.304 46.927
196789 98.910 -77.316 46.903
196790 98.924 -77.326 46.878
196791 98.939 -77.337 46.853
196792 98.953 -77.348 46.828
196793 98.968 -77.359 46.802
196794 98.983 -77.370 46.775
196795 98.999 -77.382 46.749
196796 99.013 -77.391 46.723
196797 99.024 -77.391 46.698
196798 99.041 -77.387 46.675
196799 99.052 -77.392 46.652
196800 99.063 -77.396 46.628
196801 99.074 -77.401 46.603
196802 99.085 -77.405 46.579
196803 99.096 -77.410 46.554
196804 99.106 -77.414 46.530
196805 99.118 -77.418 46.506
196806 99.129 -77.423 46.483
196807 99.140 -77.427 46.459
196808 99.150 -77.432 46.435
196809 99.161 -77.436 46.412
196810 99.172 -77.440 46.388
196811 99.184 -77.445 46.365
196812 99.195 -77.451 46.342
196813 99.207 -77.476 46.318
196814 99.205 -77.489 46.293
196815 99.213 -77.495 46.268
196816 99.235 -77.492 46.244
196817 99.258 -77.489 46.219
196818 99.280 -77.486 46.196
196819 99.302 -77.483 46.173
196820 99.324 -77.480 46.150
196821 99.346 -77.477 46.128
196822 99.368 -77.474 46.106
196823 99.391 -77.471 46.085
196824 99.413 -77.468 46.065
196825 99.435 -77.465 46.046
196826 99.457 -77.462 46.027
196827 99.479 -77.459 46.008
196828 99.504 -77.457 45.990
196829 99.544 -77.461 45.970
196830 99.584 -77.464 45.953
196831 99.624 -77.468 45.936
196832 99.663 -77.472 45.920
196833 99.704 -77.476 45.904
196834 99.745 -77.482 45.889
196835 99.789 -77.494 45.875
196836 99.834 -77.507 45.859
196837 99.879 -77.520 45.844
196838 99.924 -77.533 45.830
196839 99.952 -77.545 45.817
196840 99.966 -77.556 45.802
196841 99.979 -77.567 45.789
196842 99.992 -77.579 45.776
196843 100.005 -77.590 45.763
196844 100.018 -77.602 45.751
196845 100.031 -77.613 45.738
196846 100.044 -77.625 45.725
196847 100.057 -77.636 45.713
196848 100.071 -77.648 45.700
196849 100.086 -77.659 45.688
196850 100.101 -77.671 45.677
196851 100.116 -77.682 45.665
196852 100.138 -77.702 45.654
196853 100.162 -77.726 45.642
196854 100.193 -77.740 45.631
196855 100.223 -77.752 45.620
196856 100.254 -77.766 45.609
196857 100.285 -77.780 45.597
196858 100.308 -77.794 45.586
196859 100.327 -77.809 45.574
196860 100.347 -77.824 45.561
196861 100.366 -77.840 45.548
196862 100.385 -77.855 45.534
196863 100.410 -77.876 45.520
196864 100.434 -77.898 45.505
196865 100.459 -77.920 45.490
196866 100.484 -77.942 45.474
196867 100.508 -77.964 45.459
196868 100.525 -77.953 45.445
196869 100.540 -77.939 45.431
196870 100.558 -77.940 45.415
196871 100.576 -77.943 45.401
196872 100.594 -77.944 45.387
196873 100.614 -77.949 45.373
196874 100.637 -77.960 45.358
196875 100.660 -77.971 45.344
196876 100.683 -77.981 45.329
196877 100.706 -77.992 45.314
196878 100.729 -78.003 45.301
196879 100.753 -78.013 45.287
196880 100.776 -78.024 45.273
196881 100.798 -78.034 45.260
196882 100.822 -78.044 45.248
196883 100.845 -78.055 45.235
196884 100.873 -78.059 45.222
196885 100.902 -78.058 45.209
196886 100.925 -78.061 45.197
196887 100.949 -78.063 45.184
196888 100.975 -78.068 45.172
196889 101.014 -78.080 45.158
196890 101.053 -78.093 45.144
196891 101.092 -78.105 45.129
196892 101.131 -78.118 45.114
196893 101.169 -78.133 45.100
196894 101.205 -78.149 45.085
196895 101.242 -78.166 45.070
196896 101.278 -78.182 45.056
196897 101.315 -78.199 45.042
196898 101.336 -78.211 45.028
196899 101.353 -78.222 45.013
196900 101.371 -78.233 44.998
196901 101.387 -78.244 44.984
196902 101.404 -78.255 44.969
196903 101.421 -78.266 44.954
196904 101.438 -78.277 44.939
196905 101.455 -78.288 44.923
196906 101.472 -78.298 44.907
196907 101.489 -78.310 44.891
196908 101.505 -78.320 44.876
196909 101.523 -78.332 44.861
196910 101.540 -78.342 44.845
196911 101.564 -78.356 44.830
196912 101.589 -78.369 44.813
196913 101.621 -78.377 44.797
196914 101.650 -78.398 44.780
196915 101.680 -78.418 44.766
196916 101.709 -78.438 44.749
196917 101.739 -78.459 44.732
196918 101.768 -78.479 44.717
196919 101.798 -78.500 44.700
196920 101.827 -78.520 44.685
196921 101.857 -78.541 44.670
196922 101.887 -78.561 44.656
196923 101.932 -78.585 44.639
196924 101.977 -78.608 44.623
196925 102.023 -78.630 44.608
196926 102.068 -78.654 44.592
196927 102.114 -78.677 44.577
196928 102.159 -78.700 44.562
196929 102.203 -78.727 44.548
196930 102.246 -78.759 44.533
196931 102.289 -78.791 44.518
196932 102.329 -78.820 44.504
196933 102.349 -78.821 44.489
196934 102.368 -78.823 44.474
196935 102.388 -78.824 44.460
196936 102.407 -78.826 44.446
196937 102.427 -78.827 44.431
196938 102.446 -78.829 44.417
196939 102.466 -78.830 44.403
196940 102.485 -78.833 44.388
196941 102.504 -78.834 44.375
196942 102.524 -78.836 44.360
196943 102.543 -78.837 44.346
196944 102.563 -78.839 44.331
197228 93.526 -76.218 54.042
197229 93.544 -76.221 54.050
197230 93.563 -76.224 54.058
197231 93.581 -76.227 54.067
197232 93.599 -76.229 54.079
197233 93.618 -76.232 54.087
197234 93.636 -76.235 54.095
197235 93.654 -76.238 54.104
197236 93.673 -76.242 54.114
197237 93.691 -76.245 54.124
197238 93.709 -76.248 54.134
197239 93.724 -76.260 54.147
197240 93.727 -76.293 54.156
197241 93.735 -76.292 54.165
197242 93.744 -76.275 54.174
197243 93.770 -76.266 54.183
197244 93.803 -76.260 54.191
197245 93.836 -76.254 54.198
197246 93.869 -76.248 54.206
197247 93.904 -76.242 54.211
197248 93.939 -76.238 54.216
197249 93.958 -76.244 54.222
197250 93.962 -76.238 54.229
197251 93.966 -76.232 54.237
197252 93.970 -76.226 54.246
197253 93.973 -76.220 54.254
197254 93.977 -76.213 54.258
197255 94.004 -76.213 54.257
197256 94.033 -76.213 54.257
197257 94.062 -76.213 54.256
197258 94.091 -76.213 54.257
197259 94.120 -76.213 54.259
197260 94.149 -76.213 54.258
197261 94.178 -76.214 54.252
197262 94.203 -76.211 54.244
197263 94.224 -76.206 54.232
197264 94.246 -76.201 54.217
197265 94.267 -76.195 54.200
197266 94.287 -76.191 54.178
197267 94.305 -76.188 54.152
197268 94.323 -76.186 54.121
197269 94.341 -76.183 54.090
197270 94.359 -76.182 54.057
197271 94.378 -76.179 54.025
197272 94.396 -76.178 53.995
197273 94.414 -76.176 53.965
197274 94.432 -76.174 53.932
197275 94.443 -76.175 53.898
197276 94.455 -76.176 53.866
197277 94.467 -76.177 53.836
197278 94.478 -76.178 53.808
197279 94.490 -76.179 53.782
197280 94.501 -76.180 53.757
197281 94.513 -76.181 53.732
197282 94.525 -76.182 53.708
197283 94.536 -76.183 53.685
197284 94.549 -76.183 53.664
197285 94.566 -76.191 53.639
197286 94.588 -76.203 53.611
197287 94.609 -76.215 53.580
197288 94.630 -76.226 53.545
197289 94.651 -76.238 53.509
197290 94.667 -76.247 53.474
197291 94.678 -76.251 53.440
197292 94.689 -76.257 53.408
197293 94.700 -76.262 53.376
197294 94.711 -76.267 53.343
197295 94.722 -76.272 53.307
197296 94.734 -76.277 53.270
197297 94.745 -76.282 53.232
197298 94.756 -76.287 53.192
197299 94.767 -76.292 53.151
197300 94.778 -76.298 53.113
197301 94.789 -76.303 53.073
197302 94.801 -76.307 53.033
197303 94.813 -76.311 52.993
197304 94.825 -76.315 52.954
197305 94.838 -76.319 52.914
197306 94.850 -76.323 52.874
197307 94.863 -76.326 52.834
197308 94.876 -76.329 52.791
197309 94.888 -76.333 52.746
197310 94.901 -76.337 52.701
197311 94.913 -76.340 52.655
197312 94.926 -76.344 52.610
197313 94.938 -76.348 52.566
197314 94.949 -76.353 52.523
197315 94.935 -76.376 52.475
197316 94.950 -76.386 52.425
197317 94.975 -76.392 52.374
197318 94.999 -76.399 52.321
197319 95.032 -76.402 52.272
197320 95.081 -76.398 52.222
197321 95.130 -76.392 52.172
197322 95.180 -76.388 52.124
197323 95.230 -76.383 52.080
197324 95.279 -76.379 52.036
197325 95.328 -76.374 51.991
197326 95.353 -76.378 51.943
197327 95.376 -76.383 51.896
197328 95.399 -76.388 51.851
197329 95.422 -76.392 51.807
197330 95.445 -76.397 51.766
197331 95.468 -76.401 51.726
197332 95.491 -76.406 51.690
197333 95.515 -76.411 51.659
197334 95.538 -76.416 51.627
197335 95.560 -76.420 51.593
197336 95.584 -76.425 51.558
197337 95.607 -76.430 51.523
197338 95.630 -76.434 51.490
197339 95.654 -76.438 51.458
197340 95.680 -76.439 51.426
197341 95.705 -76.441 51.394
197342 95.739 -76.490 51.358
197343 95.763 -76.489 51.318
197344 95.787 -76.488 51.273
197345 95.811 -76.486 51.227
197346 95.835 -76.486 51.181
197347 95.859 -76.484 51.135
197348 95.882 -76.483 51.087
197349 95.906 -76.482 51.046
197350 95.930 -76.487 51.004
197351 95.954 -76.493 50.961
197352 95.978 -76.495 50.917
197353 96.001 -76.495 50.872
197354 96.024 -76.494 50.826
197355 96.045 -76.495 50.781
197356 96.061 -76.499 50.737
197357 96.075 -76.507 50.692
197358 96.088 -76.514 50.648
197359 96.101 -76.522 50.606
197360 96.114 -76.530 50.566
197361 96.129 -76.534 50.523
197362 96.147 -76.538 50.479
197363 96.168 -76.545 50.434
197364 96.190 -76.552 50.391
197365 96.212 -76.559 50.351
197366 96.234 -76.566 50.314
197367 96.256 -76.572 50.280
197368 96.278 -76.578 50.249
197369 96.301 -76.584 50.222
197370 96.322 -76.592 50.198
197371 96.340 -76.597 50.175
197372 96.359 -76.601 50.154
197373 96.374 -76.603 50.136
197374 96.389 -76.605 50.119
197375 96.404 -76.608 50.103
197376 96.432 -76.623 50.086
197377 96.451 -76.636 50.066
197378 96.459 -76.641 50.043
197379 96.467 -76.647 50.019
197380 96.476 -76.652 49.994
197381 96.501 -76.660 49.967
197382 96.549 -76.671 49.937
197383 96.597 -76.683 49.904
197384 96.636 -76.689 49.870
197385 96.663 -76.684 49.835
197386 96.690 -76.680 49.800
197387 96.718 -76.677 49.765
197388 96.745 -76.673 49.731
197389 96.773 -76.668 49.698
197390 96.800 -76.665 49.665
197391 96.827 -76.665 49.632
197392 96.855 -76.687 49.599
197393 96.881 -76.709 49.564
197394 96.908 -76.732 49.528
197395 96.929 -76.741 49.492
197396 96.947 -76.743 49.456
197397 96.964 -76.746 49.422
197398 96.982 -76.749 49.389
197399 97.000 -76.751 49.358
197400 97.017 -76.753 49.327
197401 97.031 -76.762 49.299
197402 97.044 -76.771 49.272
197403 97.058 -76.780 49.247
197404 97.075 -76.793 49.224
197405 97.107 -76.818 49.202
197406 97.139 -76.843 49.181
197407 97.145 -76.863 49.160
197408 97.148 -76.882 49.138
197409 97.150 -76.901 49.112
197410 97.153 -76.920 49.084
197411 97.155 -76.940 49.057
197412 97.165 -76.960 49.029
197413 97.219 -76.982 49.001
197414 97.308 -76.995 48.973
197415 97.386 -77.013 48.946
197416 97.415 -77.051 48.919
197417 97.444 -77.090 48.892
197418 97.473 -77.128 48.866
197419 97.502 -77.166 48.839
197420 97.531 -77.204 48.814
197421 97.549 -77.240 48.788
197422 97.556 -77.236 48.763
197423 97.557 -77.218 48.737
197424 97.557 -77.201 48.710
197425 97.557 -77.182 48.682
197426 97.558 -77.164 48.654
197427 97.558 -77.146 48.627
197428 97.558 -77.128 48.599
197429 97.559 -77.113 48.572
197430 97.564 -77.126 48.544
197431 97.568 -77.138 48.514
197432 97.590 -77.144 48.485
197433 97.612 -77.149 48.456
197434 97.634 -77.153 48.427
197435 97.656 -77.159 48.398
197436 97.678 -77.164 48.369
197437 97.700 -77.169 48.341
197438 97.722 -77.174 48.312
197439 97.749 -77.172 48.283
197440 97.776 -77.172 48.254
197441 97.805 -77.166 48.225
197442 97.836 -77.156 48.196
197443 97.866 -77.146 48.169
197444 97.896 -77.136 48.140
197445 97.927 -77.126 48.111
197446 97.956 -77.117 48.081
197447 97.977 -77.122 48.051
197448 97.997 -77.128 48.021
197449 98.016 -77.133 47.990
197450 98.036 -77.138 47.959
197451 98.057 -77.143 47.928
197452 98.082 -77.130 47.895
197453 98.108 -77.118 47.863
197454 98.134 -77.105 47.829
197455 98.159 -77.093 47.798
197456 98.184 -77.080 47.766
197457 98.210 -77.068 47.735
197458 98.235 -77.055 47.705
197459 98.261 -77.043 47.675
197460 98.287 -77.032 47.644
197461 98.311 -77.032 47.614
197462 98.336 -77.032 47.584
197463 98.360 -77.032 47.554
197464 98.384 -77.031 47.525
197465 98.409 -77.031 47.497
197466 98.433 -77.031 47.469
197467 98.457 -77.031 47.443
197468 98.482 -77.031 47.417
197469 98.507 -77.031 47.392
197470 98.505 -77.047 47.367
197471 98.497 -77.069 47.343
197472 98.488 -77.091 47.321
197473 98.480 -77.114 47.298
197474 98.489 -77.127 47.275
197475 98.505 -77.138 47.253
197476 98.534 -77.150 47.230
197477 98.567 -77.162 47.207
197478 98.599 -77.175 47.183
197479 98.648 -77.182 47.159
197480 98.698 -77.189 47.135
197481 98.748 -77.197 47.112
197482 98.798 -77.204 47.089
197483 98.820 -77.216 47.066
197484 98.835 -77.229 47.042
197485 98.849 -77.243 47.020
197486 98.862 -77.257 46.998
197487 98.877 -77.268 46.977
197488 98.891 -77.279 46.956
197489 98.906 -77.290 46.932
197490 98.921 -77.301 46.907
197491 98.935 -77.312 46.882
197492 98.950 -77.323 46.857
197493 98.964 -77.334 46.831
197494 98.979 -77.345 46.805
197495 98.994 -77.356 46.779
197496 99.008 -77.367 46.753
197497 99.023 -77.378 46.727
197498 99.038 -77.389 46.701
197499 99.049 -77.395 46.678
197500 99.060 -77.400 46.653
197501 99.070 -77.405 46.628
197502 99.081 -77.410 46.603
197503 99.092 -77.415 46.579
197504 99.103 -77.420 46.553
197505 99.114 -77.423 46.529
197506 99.125 -77.428 46.506
197507 99.136 -77.432 46.483
197508 99.147 -77.437 46.459
197509 99.157 -77.442 46.435
197510 99.168 -77.446 46.412
197511 99.180 -77.450 46.387
197512 99.191 -77.454 46.365
197513 99.201 -77.459 46.341
197514 99.212 -77.464 46.316
197515 99.218 -77.480 46.291
197516 99.236 -77.480 46.266
197517 99.258 -77.477 46.242
197518 99.280 -77.474 46.218
197519 99.302 -77.471 46.195
197520 99.325 -77.468 46.172
197521 99.347 -77.465 46.150
197522 99.369 -77.462 46.128
197523 99.391 -77.460 46.106
197524 99.413 -77.457 46.085
197525 99.435 -77.454 46.065
197526 99.457 -77.451 46.046
197527 99.479 -77.448 46.027
197528 99.501 -77.445 46.008
197529 99.523 -77.442 45.990
197530 99.549 -77.440 45.971
197531 99.589 -77.444 45.953
197532 99.629 -77.447 45.936
197533 99.669 -77.451 45.919
197534 99.709 -77.454 45.904
197535 99.753 -77.466 45.890
197536 99.798 -77.479 45.875
197537 99.843 -77.492 45.859
197538 99.888 -77.504 45.845
197539 99.932 -77.517 45.831
197540 99.974 -77.530 45.817
197541 99.987 -77.541 45.803
197542 100.000 -77.552 45.790
197543 100.013 -77.564 45.777
197544 100.027 -77.576 45.764
197545 100.040 -77.587 45.751
197546 100.053 -77.599 45.738
197547 100.066 -77.610 45.726
197548 100.079 -77.621 45.714
197549 100.093 -77.633 45.700
197550 100.109 -77.644 45.688
197551 100.124 -77.655 45.676
197552 100.139 -77.667 45.664
197553 100.154 -77.678 45.652
197554 100.177 -77.700 45.640
197555 100.207 -77.716 45.628
197556 100.237 -77.730 45.617
197557 100.268 -77.743 45.605
197558 100.298 -77.756 45.594
197559 100.329 -77.769 45.582
197560 100.352 -77.783 45.570
197561 100.372 -77.799 45.557
197562 100.393 -77.817 45.544
197563 100.418 -77.839 45.530
197564 100.442 -77.861 45.516
197565 100.467 -77.883 45.501
197566 100.492 -77.905 45.486
197567 100.516 -77.926 45.471
197568 100.541 -77.944 45.455
197569 100.555 -77.925 45.440
197570 100.570 -77.910 45.426
197571 100.588 -77.912 45.411
197572 100.606 -77.914 45.397
197573 100.626 -77.919 45.382
197574 100.649 -77.930 45.368
197575 100.672 -77.940 45.354
197576 100.695 -77.950 45.339
197577 100.718 -77.961 45.325
197578 100.742 -77.971 45.311
197579 100.764 -77.982 45.298
197580 100.787 -77.993 45.285
197581 100.811 -78.003 45.271
197582 100.834 -78.014 45.258
197583 100.857 -78.025 45.246
197584 100.880 -78.035 45.234
197585 100.903 -78.046 45.222
197586 100.931 -78.047 45.209
197587 100.955 -78.050 45.196
197588 100.978 -78.053 45.184
197589 101.001 -78.055 45.171
197590 101.034 -78.063 45.157
197591 101.073 -78.076 45.143
197592 101.112 -78.089 45.128
197593 101.148 -78.106 45.113
197594 101.185 -78.122 45.099
197595 101.221 -78.139 45.084
197596 101.257 -78.156 45.069
197597 101.294 -78.172 45.055
197598 101.330 -78.189 45.041
197599 101.357 -78.203 45.027
197600 101.374 -78.213 45.012
197601 101.391 -78.225 44.997
197602 101.408 -78.235 44.982
197603 101.424 -78.247 44.967
197604 101.442 -78.257 44.952
197605 101.458 -78.268 44.936
197606 101.475 -78.279 44.920
197607 101.492 -78.290 44.905
197608 101.509 -78.300 44.889
197609 101.526 -78.312 44.873
197610 101.543 -78.322 44.858
197611 101.567 -78.336 44.842
197612 101.592 -78.349 44.825
197613 101.616 -78.362 44.811
197614 101.640 -78.376 44.795
197615 101.665 -78.388 44.779
197616 101.697 -78.399 44.764
197617 101.726 -78.420 44.747
197618 101.756 -78.440 44.731
197619 101.785 -78.460 44.716
197620 101.815 -78.481 44.699
197621 101.844 -78.501 44.685
197622 101.874 -78.522 44.670
197623 101.903 -78.543 44.655
197624 101.949 -78.566 44.639
197625 101.994 -78.589 44.623
197626 102.040 -78.612 44.607
197627 102.085 -78.635 44.592
197628 102.130 -78.658 44.576
197629 102.176 -78.681 44.561
197630 102.221 -78.704 44.546
197631 102.265 -78.733 44.531
197632 102.308 -78.765 44.517
197633 102.351 -78.797 44.503
197634 102.374 -78.804 44.488
197635 102.393 -78.805 44.473
197636 102.413 -78.807 44.459
197637 102.433 -78.808 44.444
197638 102.452 -78.810 44.429
197639 102.471 -78.811 44.415
197640 102.491 -78.813 44.401
197641 102.510 -78.815 44.387
197642 102.530 -78.817 44.372
197643 102.549 -78.818 44.357
197644 102.569 -78.820 44.342
197645 102.588 -78.821 44.328
197930 93.536 -76.198 54.041
197931 93.555 -76.201 54.051
197932 93.573 -76.204 54.061
197933 93.591 -76.207 54.072
197934 93.610 -76.209 54.082
197935 93.628 -76.212 54.090
197936 93.646 -76.215 54.099
197937 93.664 -76.218 54.109
197938 93.683 -76.222 54.120
197939 93.701 -76.225 54.132
197940 93.716 -76.235 54.143
197941 93.720 -76.269 54.153
197942 93.725 -76.286 54.164
197943 93.739 -76.271 54.175
197944 93.772 -76.265 54.184
197945 93.805 -76.259 54.191
197946 93.838 -76.253 54.199
197947 93.871 -76.248 54.210
197948 93.904 -76.240 54.217
197949 93.936 -76.227 54.223
197950 93.946 -76.220 54.230
197951 93.949 -76.213 54.238
197952 93.952 -76.207 54.246
197953 93.955 -76.201 54.256
197954 93.957 -76.195 54.265
197955 93.978 -76.189 54.269
197956 94.007 -76.190 54.269
197957 94.037 -76.191 54.268
197958 94.066 -76.191 54.267
197959 94.096 -76.192 54.268
197960 94.124 -76.192 54.269
197961 94.153 -76.193 54.267
197962 94.183 -76.193 54.259
197963 94.206 -76.189 54.250
197964 94.228 -76.184 54.237
197965 94.250 -76.179 54.222
197966 94.270 -76.174 54.203
197967 94.288 -76.172 54.180
197968 94.306 -76.169 54.153
197969 94.324 -76.166 54.123
197970 94.342 -76.163 54.091
197971 94.360 -76.161 54.059
197972 94.378 -76.160 54.028
197973 94.396 -76.157 53.997
197974 94.414 -76.156 53.967
197975 94.432 -76.154 53.932
197976 94.445 -76.154 53.897
197977 94.457 -76.155 53.863
197978 94.468 -76.157 53.833
197979 94.480 -76.157 53.805
197980 94.491 -76.158 53.780
197981 94.503 -76.160 53.756
197982 94.514 -76.160 53.731
197983 94.526 -76.161 53.708
197984 94.538 -76.163 53.685
197985 94.551 -76.167 53.664
197986 94.568 -76.176 53.640
197987 94.589 -76.188 53.614
197988 94.610 -76.200 53.583
197989 94.631 -76.211 53.549
197990 94.652 -76.223 53.513
197991 94.668 -76.231 53.479
197992 94.679 -76.236 53.445
197993 94.690 -76.241 53.414
197994 94.701 -76.246 53.383
197995 94.712 -76.251 53.349
197996 94.723 -76.257 53.313
197997 94.735 -76.261 53.275
197998 94.746 -76.267 53.236
197999 94.757 -76.272 53.197
198000 94.768 -76.277 53.158
198001 94.779 -76.282 53.121
198002 94.790 -76.287 53.083
198003 94.801 -76.292 53.043
198004 94.812 -76.298 53.003
198005 94.824 -76.302 52.962
198006 94.836 -76.305 52.921
198007 94.849 -76.309 52.881
198008 94.861 -76.313 52.840
198009 94.874 -76.317 52.796
198010 94.886 -76.320 52.751
198011 94.899 -76.323 52.705
198012 94.911 -76.327 52.658
198013 94.924 -76.331 52.614
198014 94.936 -76.335 52.571
198015 94.948 -76.339 52.528
198016 94.934 -76.361 52.480
198017 94.940 -76.376 52.433
198018 94.965 -76.383 52.384
198019 94.989 -76.389 52.333
198020 95.028 -76.389 52.285
198021 95.077 -76.384 52.237
198022 95.127 -76.379 52.188
198023 95.177 -76.375 52.142
198024 95.226 -76.370 52.096
198025 95.275 -76.365 52.051
198026 95.325 -76.361 52.007
198027 95.357 -76.362 51.961
198028 95.380 -76.367 51.916
198029 95.403 -76.372 51.871
198030 95.426 -76.376 51.827
198031 95.449 -76.381 51.785
198032 95.472 -76.386 51.745
198033 95.495 -76.391 51.707
198034 95.519 -76.395 51.674
198035 95.542 -76.400 51.644
198036 95.564 -76.404 51.612
198037 95.588 -76.409 51.578
198038 95.611 -76.414 51.545
198039 95.635 -76.417 51.511
198040 95.660 -76.419 51.477
198041 95.686 -76.421 51.445
198042 95.711 -76.423 51.411
198043 95.746 -76.474 51.374
198044 95.773 -76.488 51.333
198045 95.797 -76.486 51.287
198046 95.820 -76.486 51.241
198047 95.844 -76.484 51.194
198048 95.868 -76.482 51.148
198049 95.892 -76.480 51.102
198050 95.915 -76.477 51.061
198051 95.939 -76.475 51.020
198052 95.963 -76.473 50.977
198053 95.984 -76.475 50.932
198054 96.006 -76.475 50.886
198055 96.028 -76.476 50.841
198056 96.049 -76.477 50.796
198057 96.063 -76.483 50.753
198058 96.076 -76.491 50.710
198059 96.089 -76.499 50.668
198060 96.102 -76.505 50.628
198061 96.118 -76.510 50.590
198062 96.137 -76.510 50.548
198063 96.151 -76.513 50.505
198064 96.170 -76.518 50.461
198065 96.191 -76.526 50.419
198066 96.214 -76.533 50.381
198067 96.235 -76.539 50.344
198068 96.257 -76.547 50.309
198069 96.279 -76.554 50.278
198070 96.301 -76.561 50.252
198071 96.320 -76.567 50.229
198072 96.333 -76.568 50.206
198073 96.348 -76.571 50.185
198074 96.363 -76.573 50.166
198075 96.379 -76.575 50.148
198076 96.395 -76.578 50.131
198077 96.436 -76.608 50.111
198078 96.459 -76.624 50.089
198079 96.467 -76.630 50.066
198080 96.476 -76.635 50.041
198081 96.484 -76.640 50.016
198082 96.528 -76.652 49.989
198083 96.576 -76.663 49.960
198084 96.615 -76.667 49.927
198085 96.649 -76.665 49.892
198086 96.684 -76.659 49.857
198087 96.712 -76.655 49.821
198088 96.739 -76.652 49.785
198089 96.766 -76.648 49.751
198090 96.794 -76.643 49.717
198091 96.821 -76.640 49.684
198092 96.849 -76.638 49.650
198093 96.875 -76.661 49.616
198094 96.902 -76.683 49.580
198095 96.929 -76.704 49.544
198096 96.946 -76.706 49.507
198097 96.964 -76.708 49.471
198098 96.981 -76.712 49.437
198099 96.999 -76.714 49.404
198100 97.017 -76.716 49.373
198101 97.034 -76.718 49.342
198102 97.051 -76.722 49.313
198103 97.064 -76.731 49.286
198104 97.082 -76.745 49.262
198105 97.114 -76.770 49.240
198106 97.116 -76.789 49.218
198107 97.119 -76.808 49.199
198108 97.125 -76.828 49.178
198109 97.137 -76.849 49.156
198110 97.149 -76.870 49.130
198111 97.161 -76.891 49.102
198112 97.173 -76.912 49.075
198113 97.185 -76.934 49.047
198114 97.219 -76.956 49.019
198115 97.288 -76.964 48.991
198116 97.394 -77.003 48.964
198117 97.457 -77.027 48.938
198118 97.486 -77.066 48.910
198119 97.508 -77.102 48.883
198120 97.521 -77.135 48.856
198121 97.534 -77.169 48.830
198122 97.547 -77.193 48.805
198123 97.560 -77.200 48.779
198124 97.569 -77.200 48.752
198125 97.570 -77.182 48.723
198126 97.570 -77.163 48.696
198127 97.570 -77.145 48.668
198128 97.571 -77.127 48.640
198129 97.571 -77.109 48.612
198130 97.572 -77.097 48.586
198131 97.577 -77.110 48.557
198132 97.589 -77.119 48.526
198133 97.611 -77.124 48.497
198134 97.633 -77.129 48.467
198135 97.655 -77.134 48.438
198136 97.677 -77.139 48.409
198137 97.699 -77.144 48.381
198138 97.721 -77.149 48.353
198139 97.743 -77.154 48.325
198140 97.765 -77.160 48.296
198141 97.793 -77.154 48.267
198142 97.823 -77.144 48.238
198143 97.854 -77.134 48.209
198144 97.884 -77.124 48.179
198145 97.915 -77.114 48.150
198146 97.945 -77.104 48.120
198147 97.973 -77.097 48.090
198148 97.994 -77.102 48.060
198149 98.014 -77.107 48.029
198150 98.033 -77.113 47.999
198151 98.054 -77.118 47.968
198152 98.074 -77.123 47.936
198153 98.094 -77.128 47.904
198154 98.119 -77.116 47.871
198155 98.145 -77.103 47.838
198156 98.170 -77.091 47.806
198157 98.196 -77.078 47.774
198158 98.221 -77.066 47.744
198159 98.247 -77.053 47.713
198160 98.272 -77.041 47.682
198161 98.297 -77.036 47.652
198162 98.321 -77.036 47.621
198163 98.346 -77.036 47.593
198164 98.370 -77.036 47.564
198165 98.395 -77.036 47.535
198166 98.420 -77.036 47.508
198167 98.444 -77.036 47.481
198168 98.468 -77.035 47.454
198169 98.492 -77.029 47.429
198170 98.516 -77.024 47.403
198171 98.520 -77.033 47.378
198172 98.514 -77.050 47.354
198173 98.506 -77.071 47.331
198174 98.498 -77.094 47.308
198175 98.490 -77.116 47.286
198176 98.505 -77.126 47.263
198177 98.532 -77.138 47.240
198178 98.565 -77.150 47.217
198179 98.612 -77.159 47.193
198180 98.662 -77.166 47.169
198181 98.712 -77.173 47.145
198182 98.762 -77.180 47.121
198183 98.812 -77.187 47.097
198184 98.833 -77.200 47.073
198185 98.846 -77.213 47.048
198186 98.860 -77.227 47.026
198187 98.874 -77.241 47.004
198188 98.888 -77.254 46.982
198189 98.903 -77.265 46.961
198190 98.917 -77.276 46.936
198191 98.932 -77.287 46.911
198192 98.946 -77.298 46.886
198193 98.961 -77.309 46.861
198194 98.976 -77.320 46.835
198195 98.990 -77.331 46.808
198196 99.005 -77.341 46.783
198197 99.019 -77.353 46.757
198198 99.034 -77.363 46.730
198199 99.049 -77.373 46.704
198200 99.061 -77.378 46.679
198201 99.073 -77.383 46.653
198202 99.084 -77.389 46.627
198203 99.094 -77.395 46.602
198204 99.105 -77.401 46.578
198205 99.116 -77.406 46.552
198206 99.126 -77.412 46.528
198207 99.137 -77.417 46.504
198208 99.147 -77.423 46.482
198209 99.158 -77.429 46.458
198210 99.169 -77.435 46.434
198211 99.180 -77.440 46.410
198212 99.190 -77.446 46.386
198213 99.201 -77.451 46.363
198214 99.211 -77.457 46.339
198215 99.222 -77.463 46.314
198216 99.238 -77.468 46.289
198217 99.264 -77.468 46.264
198218 99.283 -77.464 46.240
198219 99.303 -77.460 46.216
198220 99.325 -77.457 46.193
198221 99.347 -77.454 46.171
198222 99.369 -77.451 46.149
198223 99.391 -77.448 46.126
198224 99.413 -77.445 46.105
198225 99.435 -77.442 46.084
198226 99.457 -77.439 46.064
198227 99.480 -77.436 46.045
198228 99.502 -77.433 46.025
198229 99.524 -77.430 46.006
198230 99.546 -77.427 45.988
198231 99.568 -77.424 45.970
198232 99.594 -77.423 45.952
198233 99.634 -77.426 45.935
198234 99.674 -77.430 45.918
198235 99.716 -77.438 45.903
198236 99.761 -77.451 45.889
198237 99.806 -77.464 45.874
198238 99.851 -77.476 45.859
198239 99.896 -77.489 45.845
198240 99.941 -77.501 45.831
198241 99.986 -77.514 45.817
198242 100.009 -77.526 45.803
198243 100.022 -77.538 45.790
198244 100.035 -77.549 45.777
198245 100.049 -77.561 45.763
198246 100.062 -77.572 45.751
198247 100.075 -77.583 45.738
198248 100.088 -77.595 45.726
198249 100.101 -77.607 45.713
198250 100.116 -77.618 45.699
198251 100.132 -77.630 45.687
198252 100.147 -77.641 45.675
198253 100.162 -77.652 45.663
198254 100.177 -77.664 45.651
198255 100.193 -77.675 45.638
198256 100.220 -77.692 45.626
198257 100.251 -77.706 45.613
198258 100.282 -77.719 45.601
198259 100.312 -77.733 45.590
198260 100.343 -77.746 45.578
198261 100.374 -77.759 45.566
198262 100.402 -77.780 45.553
198263 100.426 -77.801 45.540
198264 100.451 -77.823 45.527
198265 100.476 -77.845 45.513
198266 100.500 -77.867 45.498
198267 100.525 -77.889 45.483
198268 100.550 -77.911 45.468
198269 100.571 -77.918 45.452
198270 100.585 -77.898 45.436
198271 100.600 -77.881 45.422
198272 100.618 -77.884 45.407
198273 100.638 -77.888 45.393
198274 100.661 -77.899 45.378
198275 100.684 -77.909 45.363
198276 100.707 -77.920 45.349
198277 100.731 -77.931 45.335
198278 100.753 -77.941 45.321
198279 100.776 -77.952 45.308
198280 100.799 -77.962 45.295
198281 100.823 -77.973 45.282
198282 100.846 -77.984 45.269
198283 100.868 -77.993 45.257
198284 100.892 -78.004 45.245
198285 100.915 -78.015 45.233
198286 100.937 -78.026 45.221
198287 100.960 -78.037 45.208
198288 100.984 -78.040 45.196
198289 101.007 -78.042 45.183
198290 101.031 -78.044 45.170
198291 101.054 -78.047 45.156
198292 101.091 -78.062 45.142
198293 101.127 -78.079 45.127
198294 101.164 -78.096 45.113
198295 101.200 -78.112 45.097
198296 101.237 -78.129 45.082
198297 101.273 -78.146 45.068
198298 101.310 -78.163 45.054
198299 101.346 -78.179 45.040
198300 101.377 -78.194 45.025
198301 101.394 -78.205 45.010
198302 101.411 -78.216 44.995
198303 101.428 -78.227 44.980
198304 101.445 -78.238 44.965
198305 101.462 -78.249 44.949
198306 101.479 -78.260 44.933
198307 101.495 -78.270 44.917
198308 101.513 -78.282 44.900
198309 101.530 -78.292 44.885
198310 101.546 -78.303 44.869
198311 101.570 -78.316 44.854
198312 101.595 -78.329 44.838
198313 101.619 -78.342 44.822
198314 101.643 -78.356 44.807
198315 101.668 -78.369 44.792
198316 101.692 -78.382 44.777
198317 101.717 -78.395 44.761
198318 101.741 -78.408 44.745
198319 101.772 -78.422 44.730
198320 101.802 -78.442 44.714
198321 101.831 -78.463 44.699
198322 101.861 -78.483 44.684
198323 101.890 -78.504 44.669
198324 101.920 -78.524 44.654
198325 101.965 -78.547 44.638
198326 102.011 -78.570 44.622
198327 102.056 -78.593 44.606
198328 102.102 -78.617 44.591
198329 102.147 -78.639 44.576
198330 102.192 -78.662 44.560
198331 102.238 -78.686 44.545
198332 102.283 -78.708 44.530
198333 102.326 -78.739 44.515
198334 102.369 -78.771 44.501
198335 102.399 -78.786 44.486
198336 102.418 -78.788 44.471
198337 102.438 -78.789 44.456
198338 102.458 -78.791 44.441
198339 102.477 -78.793 44.426
198340 102.497 -78.794 44.412
198341 102.516 -78.796 44.398
198342 102.535 -78.798 44.383
198343 102.555 -78.799 44.369
198344 102.574 -78.801 44.354
198345 102.594 -78.802 44.339
198346 102.613 -78.804 44.324
198631 93.528 -76.174 54.033
198632 93.546 -76.177 54.043
198633 93.565 -76.180 54.054
198634 93.583 -76.183 54.066
198635 93.602 -76.185 54.076
198636 93.620 -76.188 54.085
198637 93.638 -76.191 54.095
198638 93.656 -76.195 54.106
198639 93.675 -76.198 54.117
198640 93.693 -76.201 54.130
198641 93.708 -76.211 54.142
198642 93.712 -76.245 54.153
198643 93.716 -76.279 54.164
198644 93.741 -76.270 54.176
198645 93.773 -76.263 54.185
198646 93.805 -76.250 54.193
198647 93.838 -76.236 54.202
198648 93.870 -76.223 54.213
198649 93.902 -76.210 54.222
198650 93.921 -76.203 54.231
198651 93.930 -76.195 54.239
198652 93.933 -76.189 54.249
198653 93.936 -76.183 54.259
198654 93.939 -76.176 54.269
198655 93.959 -76.171 54.279
198656 93.980 -76.166 54.281
198657 94.009 -76.167 54.282
198658 94.040 -76.169 54.281
198659 94.070 -76.171 54.281
198660 94.100 -76.171 54.282
198661 94.129 -76.171 54.283
198662 94.157 -76.172 54.280
198663 94.187 -76.172 54.272
198664 94.210 -76.168 54.258
198665 94.231 -76.163 54.243
198666 94.253 -76.157 54.227
198667 94.271 -76.154 54.206
198668 94.289 -76.152 54.184
198669 94.307 -76.150 54.157
198670 94.325 -76.147 54.125
198671 94.342 -76.144 54.094
198672 94.360 -76.141 54.063
198673 94.379 -76.139 54.032
198674 94.397 -76.138 54.001
198675 94.415 -76.135 53.968
198676 94.433 -76.134 53.932
198677 94.447 -76.134 53.896
198678 94.458 -76.135 53.861
198679 94.470 -76.136 53.829
198680 94.481 -76.137 53.802
198681 94.493 -76.138 53.777
198682 94.504 -76.139 53.755
198683 94.516 -76.140 53.729
198684 94.528 -76.141 53.705
198685 94.541 -76.148 53.683
198686 94.555 -76.154 53.662
198687 94.569 -76.161 53.639
198688 94.591 -76.173 53.614
198689 94.612 -76.185 53.585
198690 94.633 -76.197 53.551
198691 94.654 -76.208 53.516
198692 94.668 -76.215 53.482
198693 94.679 -76.220 53.449
198694 94.690 -76.226 53.418
198695 94.701 -76.230 53.388
198696 94.712 -76.235 53.355
198697 94.723 -76.241 53.318
198698 94.735 -76.246 53.280
198699 94.746 -76.251 53.242
198700 94.757 -76.256 53.204
198701 94.768 -76.261 53.166
198702 94.779 -76.267 53.130
198703 94.790 -76.271 53.092
198704 94.802 -76.276 53.053
198705 94.813 -76.282 53.012
198706 94.824 -76.287 52.971
198707 94.835 -76.292 52.929
198708 94.847 -76.296 52.888
198709 94.859 -76.300 52.846
198710 94.872 -76.303 52.802
198711 94.885 -76.307 52.757
198712 94.897 -76.310 52.712
198713 94.910 -76.314 52.666
198714 94.922 -76.317 52.623
198715 94.935 -76.321 52.580
198716 94.947 -76.325 52.537
198717 94.933 -76.348 52.492
198718 94.929 -76.366 52.445
198719 94.954 -76.373 52.398
198720 94.979 -76.379 52.349
198721 95.025 -76.376 52.303
198722 95.074 -76.371 52.256
198723 95.124 -76.367 52.209
198724 95.173 -76.362 52.162
198725 95.222 -76.357 52.115
198726 95.272 -76.352 52.068
198727 95.322 -76.348 52.023
198728 95.361 -76.347 51.979
198729 95.384 -76.351 51.935
198730 95.407 -76.356 51.892
198731 95.430 -76.361 51.849
198732 95.453 -76.365 51.806
198733 95.476 -76.370 51.765
198734 95.499 -76.375 51.726
198735 95.523 -76.379 51.691
198736 95.546 -76.384 51.659
198737 95.568 -76.389 51.628
198738 95.592 -76.394 51.597
198739 95.615 -76.398 51.565
198740 95.641 -76.399 51.532
198741 95.666 -76.401 51.500
198742 95.692 -76.402 51.467
198743 95.718 -76.404 51.432
198744 95.752 -76.455 51.394
198745 95.779 -76.462 51.351
198746 95.803 -76.461 51.303
198747 95.827 -76.459 51.256
198748 95.851 -76.457 51.209
198749 95.875 -76.455 51.162
198750 95.898 -76.452 51.116
198751 95.922 -76.451 51.075
198752 95.945 -76.451 51.032
198753 95.966 -76.455 50.989
198754 95.988 -76.457 50.944
198755 96.010 -76.457 50.899
198756 96.031 -76.458 50.853
198757 96.051 -76.460 50.808
198758 96.064 -76.467 50.767
198759 96.078 -76.474 50.726
198760 96.091 -76.482 50.687
198761 96.106 -76.486 50.650
198762 96.126 -76.486 50.613
198763 96.144 -76.486 50.572
198764 96.157 -76.489 50.530
198765 96.171 -76.492 50.489
198766 96.193 -76.499 50.448
198767 96.215 -76.506 50.410
198768 96.237 -76.514 50.374
198769 96.258 -76.520 50.338
198770 96.281 -76.527 50.307
198771 96.302 -76.535 50.280
198772 96.315 -76.538 50.257
198773 96.322 -76.538 50.234
198774 96.337 -76.540 50.212
198775 96.353 -76.542 50.192
198776 96.368 -76.544 50.174
198777 96.398 -76.563 50.155
198778 96.439 -76.592 50.133
198779 96.467 -76.613 50.111
198780 96.476 -76.618 50.087
198781 96.484 -76.624 50.062
198782 96.508 -76.631 50.037
198783 96.556 -76.643 50.010
198784 96.596 -76.648 49.981
198785 96.619 -76.646 49.948
198786 96.651 -76.645 49.914
198787 96.696 -76.636 49.878
198788 96.732 -76.630 49.842
198789 96.760 -76.627 49.806
198790 96.787 -76.622 49.770
198791 96.815 -76.618 49.736
198792 96.842 -76.614 49.702
198793 96.870 -76.612 49.668
198794 96.896 -76.634 49.633
198795 96.923 -76.656 49.598
198796 96.946 -76.669 49.560
198797 96.963 -76.672 49.523
198798 96.981 -76.674 49.486
198799 96.999 -76.677 49.451
198800 97.016 -76.679 49.418
198801 97.034 -76.681 49.386
198802 97.051 -76.683 49.355
198803 97.069 -76.687 49.328
198804 97.086 -76.696 49.302
198805 97.096 -76.717 49.277
198806 97.108 -76.738 49.255
198807 97.120 -76.759 49.235
198808 97.132 -76.781 49.216
198809 97.144 -76.802 49.196
198810 97.156 -76.823 49.173
198811 97.168 -76.844 49.148
198812 97.180 -76.865 49.121
198813 97.192 -76.887 49.093
198814 97.204 -76.908 49.066
198815 97.231 -76.929 49.038
198816 97.296 -76.949 49.010
198817 97.366 -76.953 48.983
198818 97.441 -76.963 48.955
198819 97.468 -76.997 48.927
198820 97.492 -77.047 48.900
198821 97.515 -77.095 48.873
198822 97.533 -77.134 48.846
198823 97.541 -77.150 48.820
198824 97.557 -77.155 48.793
198825 97.572 -77.160 48.766
198826 97.583 -77.163 48.738
198827 97.583 -77.144 48.710
198828 97.583 -77.126 48.682
198829 97.584 -77.108 48.654
198830 97.584 -77.090 48.626
198831 97.585 -77.081 48.597
198832 97.590 -77.093 48.568
198833 97.609 -77.099 48.538
198834 97.632 -77.104 48.509
198835 97.654 -77.110 48.478
198836 97.675 -77.114 48.449
198837 97.698 -77.119 48.420
198838 97.720 -77.125 48.392
198839 97.742 -77.129 48.364
198840 97.765 -77.135 48.335
198841 97.788 -77.141 48.307
198842 97.812 -77.134 48.279
198843 97.841 -77.122 48.250
198844 97.872 -77.112 48.220
198845 97.902 -77.102 48.191
198846 97.933 -77.092 48.161
198847 97.963 -77.081 48.130
198848 97.991 -77.076 48.100
198849 98.011 -77.081 48.069
198850 98.031 -77.087 48.039
198851 98.051 -77.092 48.008
198852 98.071 -77.097 47.977
198853 98.091 -77.103 47.945
198854 98.110 -77.108 47.913
198855 98.131 -77.113 47.880
198856 98.156 -77.102 47.847
198857 98.181 -77.089 47.815
198858 98.207 -77.077 47.784
198859 98.232 -77.064 47.753
198860 98.258 -77.052 47.722
198861 98.283 -77.041 47.692
198862 98.308 -77.041 47.661
198863 98.333 -77.041 47.631
198864 98.357 -77.041 47.602
198865 98.381 -77.041 47.573
198866 98.406 -77.041 47.545
198867 98.430 -77.039 47.518
198868 98.454 -77.034 47.491
198869 98.478 -77.028 47.465
198870 98.502 -77.022 47.440
198871 98.526 -77.016 47.414
198872 98.535 -77.022 47.389
198873 98.528 -77.040 47.365
198874 98.522 -77.057 47.343
198875 98.515 -77.075 47.320
198876 98.507 -77.095 47.297
198877 98.505 -77.115 47.274
198878 98.531 -77.126 47.251
198879 98.575 -77.135 47.226
198880 98.625 -77.142 47.202
198881 98.675 -77.149 47.179
198882 98.725 -77.157 47.155
198883 98.775 -77.163 47.130
198884 98.826 -77.171 47.105
198885 98.844 -77.184 47.080
198886 98.858 -77.197 47.055
198887 98.872 -77.211 47.032
198888 98.886 -77.225 47.009
198889 98.900 -77.238 46.987
198890 98.914 -77.251 46.965
198891 98.928 -77.262 46.940
198892 98.943 -77.273 46.915
198893 98.957 -77.284 46.889
198894 98.973 -77.294 46.864
198895 98.987 -77.306 46.838
198896 99.002 -77.316 46.812
198897 99.016 -77.328 46.787
198898 99.031 -77.338 46.760
198899 99.046 -77.349 46.733
198900 99.059 -77.357 46.706
198901 99.072 -77.363 46.680
198902 99.085 -77.367 46.653
198903 99.097 -77.373 46.628
198904 99.108 -77.379 46.603
198905 99.119 -77.384 46.577
198906 99.129 -77.390 46.551
198907 99.140 -77.395 46.527
198908 99.150 -77.401 46.503
198909 99.161 -77.407 46.480
198910 99.172 -77.413 46.456
198911 99.183 -77.418 46.432
198912 99.193 -77.423 46.407
198913 99.204 -77.429 46.383
198914 99.214 -77.435 46.359
198915 99.225 -77.441 46.335
198916 99.235 -77.446 46.310
198917 99.266 -77.451 46.287
198918 99.297 -77.454 46.262
198919 99.328 -77.458 46.237
198920 99.348 -77.454 46.213
198921 99.367 -77.450 46.191
198922 99.387 -77.446 46.169
198923 99.406 -77.442 46.147
198924 99.426 -77.439 46.125
198925 99.445 -77.434 46.103
198926 99.465 -77.430 46.082
198927 99.485 -77.426 46.062
198928 99.504 -77.422 46.043
198929 99.524 -77.418 46.024
198930 99.547 -77.415 46.005
198931 99.569 -77.412 45.986
198932 99.591 -77.409 45.968
198933 99.613 -77.407 45.951
198934 99.639 -77.405 45.934
198935 99.680 -77.410 45.917
198936 99.725 -77.423 45.902
198937 99.770 -77.436 45.888
198938 99.815 -77.448 45.873
198939 99.859 -77.461 45.859
198940 99.904 -77.473 45.845
198941 99.949 -77.486 45.831
198942 99.994 -77.499 45.817
198943 100.031 -77.511 45.803
198944 100.044 -77.523 45.790
198945 100.057 -77.535 45.777
198946 100.070 -77.546 45.763
198947 100.083 -77.558 45.751
198948 100.097 -77.569 45.738
198949 100.110 -77.580 45.725
198950 100.124 -77.592 45.712
198951 100.139 -77.603 45.698
198952 100.154 -77.614 45.685
198953 100.170 -77.626 45.673
198954 100.185 -77.637 45.661
198955 100.200 -77.649 45.648
198956 100.215 -77.660 45.635
198957 100.239 -77.671 45.622
198958 100.271 -77.685 45.609
198959 100.302 -77.699 45.597
198960 100.334 -77.713 45.586
198961 100.366 -77.727 45.574
198962 100.398 -77.741 45.563
198963 100.420 -77.746 45.550
198964 100.443 -77.751 45.537
198965 100.467 -77.761 45.523
198966 100.497 -77.799 45.508
198967 100.528 -77.838 45.494
198968 100.558 -77.874 45.479
198969 100.583 -77.896 45.464
198970 100.601 -77.890 45.448
198971 100.616 -77.871 45.432
198972 100.630 -77.853 45.417
198973 100.650 -77.858 45.403
198974 100.673 -77.868 45.389
198975 100.696 -77.879 45.374
198976 100.719 -77.890 45.359
198977 100.742 -77.900 45.345
198978 100.765 -77.911 45.331
198979 100.787 -77.923 45.318
198980 100.810 -77.934 45.305
198981 100.832 -77.946 45.292
198982 100.854 -77.959 45.280
198983 100.877 -77.971 45.267
198984 100.899 -77.983 45.255
198985 100.921 -77.994 45.244
198986 100.944 -78.006 45.232
198987 100.966 -78.018 45.220
198988 100.989 -78.031 45.208
198989 101.012 -78.034 45.195
198990 101.036 -78.036 45.182
198991 101.059 -78.039 45.169
198992 101.082 -78.041 45.155
198993 101.106 -78.052 45.141
198994 101.142 -78.069 45.126
198995 101.179 -78.086 45.111
198996 101.215 -78.103 45.095
198997 101.252 -78.119 45.081
198998 101.288 -78.136 45.066
198999 101.325 -78.152 45.051
199000 101.361 -78.169 45.037
199001 101.397 -78.185 45.022
199002 101.414 -78.197 45.007
199003 101.431 -78.207 44.992
199004 101.448 -78.218 44.977
199005 101.465 -78.229 44.962
199006 101.482 -78.240 44.946
199007 101.499 -78.250 44.930
199008 101.516 -78.262 44.913
199009 101.533 -78.272 44.896
199010 101.550 -78.284 44.880
199011 101.573 -78.296 44.865
199012 101.597 -78.310 44.849
199013 101.622 -78.322 44.834
199014 101.646 -78.336 44.819
199015 101.671 -78.349 44.803
199016 101.695 -78.363 44.788
199017 101.720 -78.376 44.773
199018 101.744 -78.388 44.757
199019 101.768 -78.402 44.743
199020 101.793 -78.415 44.728
199021 101.818 -78.429 44.712
199022 101.848 -78.444 44.697
199023 101.878 -78.464 44.682
199024 101.907 -78.485 44.668
199025 101.937 -78.505 44.653
199026 101.982 -78.529 44.637
199027 102.028 -78.551 44.621
199028 102.073 -78.575 44.605
199029 102.118 -78.598 44.590
199030 102.164 -78.620 44.574
199031 102.209 -78.644 44.559
199032 102.254 -78.667 44.543
199033 102.300 -78.689 44.528
199034 102.345 -78.714 44.513
199035 102.388 -78.745 44.498
199036 102.424 -78.769 44.483
199037 102.444 -78.771 44.468
199038 102.463 -78.772 44.453
199039 102.483 -78.774 44.439
199040 102.502 -78.775 44.424
199041 102.522 -78.777 44.409
199042 102.541 -78.779 44.394
199043 102.560 -78.780 44.380
199044 102.580 -78.782 44.366
199045 102.599 -78.783 44.351
199046 102.617 -78.785 44.336
199047 102.633 -78.787 44.321
199048 102.649 -78.789 44.306
199333 93.538 -76.154 54.034
199334 93.557 -76.157 54.047
199335 93.575 -76.160 54.060
199336 93.593 -76.163 54.071
199337 93.612 -76.165 54.080
199338 93.630 -76.169 54.091
199339 93.648 -76.172 54.103
199340 93.666 -76.175 54.115
199341 93.685 -76.179 54.127
199342 93.700 -76.188 54.141
199343 93.704 -76.221 54.154
199344 93.708 -76.254 54.166
199345 93.740 -76.246 54.178
199346 93.771 -76.233 54.187
199347 93.803 -76.220 54.196
199348 93.836 -76.207 54.206
199349 93.868 -76.194 54.218
199350 93.888 -76.185 54.227
199351 93.905 -76.178 54.237
199352 93.915 -76.171 54.248
199353 93.918 -76.165 54.259
199354 93.920 -76.159 54.270
199355 93.939 -76.153 54.281
199356 93.960 -76.148 54.291
199357 93.982 -76.143 54.293
199358 94.012 -76.144 54.294
199359 94.043 -76.147 54.295
199360 94.073 -76.148 54.296
199361 94.104 -76.150 54.296
199362 94.133 -76.151 54.297
199363 94.161 -76.151 54.294
199364 94.191 -76.151 54.286
199365 94.213 -76.146 54.270
199366 94.235 -76.141 54.253
199367 94.254 -76.138 54.235
199368 94.271 -76.135 54.213
199369 94.289 -76.132 54.190
199370 94.308 -76.130 54.162
199371 94.325 -76.127 54.131
199372 94.343 -76.125 54.098
199373 94.361 -76.122 54.066
199374 94.379 -76.119 54.034
199375 94.397 -76.117 54.001
199376 94.415 -76.115 53.968
199377 94.433 -76.113 53.932
199378 94.448 -76.113 53.895
199379 94.459 -76.114 53.858
199380 94.471 -76.115 53.826
199381 94.482 -76.116 53.797
199382 94.494 -76.117 53.771
199383 94.505 -76.119 53.749
199384 94.518 -76.120 53.722
199385 94.532 -76.127 53.697
199386 94.546 -76.135 53.675
199387 94.560 -76.141 53.656
199388 94.574 -76.148 53.635
199389 94.592 -76.158 53.611
199390 94.613 -76.169 53.584
199391 94.634 -76.182 53.549
199392 94.655 -76.193 53.516
199393 94.669 -76.200 53.483
199394 94.680 -76.204 53.451
199395 94.691 -76.210 53.420
199396 94.702 -76.215 53.391
199397 94.713 -76.220 53.360
199398 94.724 -76.225 53.323
199399 94.736 -76.230 53.285
199400 94.747 -76.235 53.247
199401 94.758 -76.240 53.210
199402 94.769 -76.245 53.175
199403 94.780 -76.251 53.139
199404 94.791 -76.256 53.101
199405 94.802 -76.260 53.062
199406 94.813 -76.266 53.021
199407 94.824 -76.271 52.979
199408 94.835 -76.276 52.938
199409 94.846 -76.281 52.897
199410 94.857 -76.286 52.854
199411 94.870 -76.290 52.810
199412 94.883 -76.294 52.765
199413 94.895 -76.297 52.721
199414 94.908 -76.301 52.677
199415 94.920 -76.304 52.633
199416 94.933 -76.308 52.590
199417 94.946 -76.311 52.548
199418 94.933 -76.333 52.503
199419 94.919 -76.356 52.458
199420 94.944 -76.363 52.412
199421 94.972 -76.367 52.365
199422 95.022 -76.363 52.321
199423 95.071 -76.358 52.276
199424 95.120 -76.353 52.231
199425 95.170 -76.348 52.183
199426 95.219 -76.344 52.135
199427 95.269 -76.339 52.087
199428 95.318 -76.335 52.041
199429 95.365 -76.331 51.998
199430 95.388 -76.336 51.956
199431 95.411 -76.340 51.914
199432 95.434 -76.345 51.872
199433 95.457 -76.350 51.829
199434 95.480 -76.354 51.787
199435 95.503 -76.359 51.746
199436 95.527 -76.364 51.708
199437 95.550 -76.368 51.675
199438 95.573 -76.373 51.644
199439 95.596 -76.378 51.614
199440 95.621 -76.379 51.583
199441 95.647 -76.381 51.552
199442 95.672 -76.383 51.520
199443 95.698 -76.384 51.488
199444 95.724 -76.386 51.454
199445 95.756 -76.426 51.414
199446 95.782 -76.429 51.369
199447 95.809 -76.432 51.321
199448 95.834 -76.432 51.271
199449 95.858 -76.430 51.224
199450 95.881 -76.428 51.177
199451 95.904 -76.426 51.131
199452 95.926 -76.430 51.089
199453 95.948 -76.433 51.045
199454 95.970 -76.436 51.001
199455 95.992 -76.439 50.956
199456 96.014 -76.439 50.911
199457 96.035 -76.439 50.867
199458 96.053 -76.443 50.823
199459 96.066 -76.451 50.783
199460 96.080 -76.458 50.744
199461 96.095 -76.463 50.706
199462 96.114 -76.462 50.671
199463 96.133 -76.462 50.634
199464 96.151 -76.463 50.594
199465 96.164 -76.466 50.554
199466 96.177 -76.469 50.515
199467 96.194 -76.473 50.475
199468 96.216 -76.480 50.437
199469 96.238 -76.487 50.400
199470 96.260 -76.495 50.364
199471 96.282 -76.502 50.333
199472 96.304 -76.508 50.307
199473 96.310 -76.508 50.283
199474 96.315 -76.508 50.259
199475 96.326 -76.509 50.237
199476 96.341 -76.511 50.217
199477 96.361 -76.518 50.198
199478 96.402 -76.548 50.178
199479 96.443 -76.578 50.155
199480 96.476 -76.602 50.132
199481 96.484 -76.607 50.109
199482 96.493 -76.612 50.084
199483 96.536 -76.623 50.059
199484 96.578 -76.630 50.032
199485 96.597 -76.621 50.003
199486 96.622 -76.624 49.971
199487 96.653 -76.624 49.937
199488 96.698 -76.616 49.901
199489 96.744 -76.608 49.864
199490 96.781 -76.601 49.827
199491 96.808 -76.597 49.791
199492 96.835 -76.593 49.755
199493 96.863 -76.589 49.721
199494 96.890 -76.586 49.685
199495 96.918 -76.608 49.650
199496 96.944 -76.630 49.614
199497 96.963 -76.635 49.576
199498 96.980 -76.637 49.538
199499 96.998 -76.640 49.501
199500 97.016 -76.642 49.465
199501 97.033 -76.644 49.432
199502 97.051 -76.647 49.400
199503 97.067 -76.649 49.369
199504 97.061 -76.657 49.341
199505 97.087 -76.684 49.316
199506 97.115 -76.690 49.293
199507 97.127 -76.712 49.271
199508 97.140 -76.733 49.251
199509 97.152 -76.754 49.233
199510 97.164 -76.775 49.213
199511 97.175 -76.796 49.191
199512 97.187 -76.818 49.167
199513 97.199 -76.839 49.140
199514 97.212 -76.860 49.112
199515 97.224 -76.881 49.085
199516 97.242 -76.903 49.057
199517 97.307 -76.925 49.029
199518 97.374 -76.939 49.001
199519 97.444 -76.943 48.972
199520 97.478 -76.981 48.944
199521 97.502 -77.030 48.917
199522 97.521 -77.071 48.890
199523 97.529 -77.087 48.863
199524 97.537 -77.103 48.836
199525 97.545 -77.119 48.809
199526 97.563 -77.119 48.781
199527 97.583 -77.119 48.753
199528 97.595 -77.125 48.725
199529 97.596 -77.108 48.696
199530 97.596 -77.090 48.668
199531 97.597 -77.072 48.639
199532 97.598 -77.064 48.610
199533 97.608 -77.075 48.580
199534 97.630 -77.079 48.550
199535 97.652 -77.085 48.521
199536 97.674 -77.090 48.490
199537 97.697 -77.094 48.460
199538 97.719 -77.100 48.431
199539 97.742 -77.106 48.404
199540 97.765 -77.111 48.375
199541 97.788 -77.117 48.346
199542 97.811 -77.122 48.318
199543 97.836 -77.117 48.290
199544 97.860 -77.101 48.261
199545 97.890 -77.090 48.231
199546 97.920 -77.080 48.202
199547 97.951 -77.069 48.172
199548 97.981 -77.059 48.141
199549 98.008 -77.055 48.110
199550 98.028 -77.060 48.078
199551 98.048 -77.066 48.048
199552 98.068 -77.072 48.017
199553 98.088 -77.077 47.986
199554 98.108 -77.082 47.954
199555 98.128 -77.088 47.921
199556 98.148 -77.093 47.889
199557 98.168 -77.098 47.856
199558 98.193 -77.088 47.824
199559 98.218 -77.075 47.793
199560 98.244 -77.063 47.762
199561 98.269 -77.050 47.733
199562 98.294 -77.045 47.702
199563 98.318 -77.045 47.671
199564 98.343 -77.045 47.641
199565 98.367 -77.044 47.612
199566 98.392 -77.043 47.583
199567 98.416 -77.038 47.555
199568 98.440 -77.032 47.527
199569 98.464 -77.026 47.501
199570 98.488 -77.021 47.476
199571 98.512 -77.015 47.451
199572 98.536 -77.009 47.425
199573 98.550 -77.012 47.400
199574 98.543 -77.029 47.377
199575 98.536 -77.047 47.354
199576 98.530 -77.064 47.332
199577 98.523 -77.081 47.309
199578 98.516 -77.100 47.285
199579 98.538 -77.112 47.260
199580 98.589 -77.119 47.234
199581 98.639 -77.125 47.212
199582 98.689 -77.133 47.189
199583 98.739 -77.140 47.164
199584 98.789 -77.147 47.139
199585 98.839 -77.154 47.114
199586 98.856 -77.167 47.088
199587 98.870 -77.181 47.062
199588 98.883 -77.194 47.037
199589 98.898 -77.209 47.014
199590 98.911 -77.222 46.992
199591 98.925 -77.236 46.968
199592 98.940 -77.248 46.944
199593 98.954 -77.259 46.919
199594 98.969 -77.269 46.893
199595 98.984 -77.281 46.868
199596 98.998 -77.291 46.842
199597 99.013 -77.302 46.816
199598 99.027 -77.313 46.790
199599 99.042 -77.324 46.764
199600 99.057 -77.335 46.736
199601 99.070 -77.342 46.707
199602 99.083 -77.347 46.680
199603 99.095 -77.352 46.654
199604 99.109 -77.357 46.628
199605 99.121 -77.362 46.602
199606 99.132 -77.367 46.575
199607 99.143 -77.373 46.550
199608 99.153 -77.379 46.525
199609 99.164 -77.385 46.501
199610 99.174 -77.390 46.477
199611 99.186 -77.396 46.454
199612 99.196 -77.401 46.429
199613 99.207 -77.407 46.405
199614 99.217 -77.413 46.379
199615 99.228 -77.419 46.355
199616 99.238 -77.424 46.331
199617 99.262 -77.429 46.307
199618 99.293 -77.433 46.284
199619 99.325 -77.437 46.260
199620 99.356 -77.442 46.235
199621 99.388 -77.445 46.211
199622 99.412 -77.445 46.189
199623 99.432 -77.441 46.166
199624 99.451 -77.436 46.145
199625 99.471 -77.432 46.123
199626 99.490 -77.429 46.101
199627 99.510 -77.424 46.080
199628 99.529 -77.420 46.059
199629 99.549 -77.417 46.041
199630 99.569 -77.413 46.022
199631 99.588 -77.408 46.003
199632 99.608 -77.404 45.985
199633 99.627 -77.401 45.967
199634 99.647 -77.396 45.950
199635 99.666 -77.392 45.933
199636 99.689 -77.395 45.917
199637 99.733 -77.407 45.901
199638 99.778 -77.420 45.887
199639 99.823 -77.433 45.872
199640 99.868 -77.445 45.857
199641 99.913 -77.458 45.844
199642 99.958 -77.471 45.830
199643 100.002 -77.484 45.816
199644 100.047 -77.497 45.802
199645 100.066 -77.508 45.790
199646 100.079 -77.520 45.776
199647 100.092 -77.531 45.762
199648 100.105 -77.542 45.749
199649 100.118 -77.554 45.736
199650 100.132 -77.566 45.723
199651 100.147 -77.577 45.709
199652 100.162 -77.589 45.696
199653 100.177 -77.600 45.683
199654 100.193 -77.611 45.670
199655 100.208 -77.623 45.658
199656 100.225 -77.632 45.644
199657 100.244 -77.639 45.632
199658 100.263 -77.653 45.619
199659 100.294 -77.667 45.606
199660 100.327 -77.681 45.593
199661 100.358 -77.695 45.581
199662 100.390 -77.709 45.570
199663 100.419 -77.720 45.558
199664 100.441 -77.725 45.545
199665 100.464 -77.730 45.532
199666 100.487 -77.735 45.518
199667 100.509 -77.740 45.504
199668 100.531 -77.746 45.489
199669 100.561 -77.777 45.474
199670 100.592 -77.815 45.459
199671 100.618 -77.835 45.443
199672 100.635 -77.822 45.427
199673 100.656 -77.823 45.412
199674 100.682 -77.843 45.398
199675 100.704 -77.855 45.384
199676 100.727 -77.867 45.370
199677 100.749 -77.879 45.356
199678 100.772 -77.891 45.341
199679 100.794 -77.902 45.328
199680 100.816 -77.915 45.315
199681 100.839 -77.927 45.301
199682 100.861 -77.939 45.289
199683 100.884 -77.951 45.276
199684 100.906 -77.962 45.265
199685 100.928 -77.975 45.253
199686 100.951 -77.987 45.241
199687 100.973 -77.999 45.230
199688 100.995 -78.010 45.218
199689 101.018 -78.022 45.206
199690 101.041 -78.028 45.192
199691 101.064 -78.031 45.179
199692 101.088 -78.033 45.166
199693 101.109 -78.035 45.153
199694 101.118 -78.037 45.138
199695 101.157 -78.057 45.124
199696 101.195 -78.075 45.109
199697 101.231 -78.092 45.093
199698 101.267 -78.109 45.079
199699 101.304 -78.125 45.064
199700 101.340 -78.142 45.049
199701 101.377 -78.159 45.033
199702 101.413 -78.175 45.018
199703 101.434 -78.188 45.003
199704 101.452 -78.199 44.988
199705 101.469 -78.210 44.974
199706 101.485 -78.220 44.958
199707 101.502 -78.232 44.943
199708 101.519 -78.242 44.926
199709 101.536 -78.253 44.909
199710 101.553 -78.264 44.893
199711 101.575 -78.276 44.876
199712 101.600 -78.290 44.859
199713 101.624 -78.303 44.844
199714 101.649 -78.316 44.829
199715 101.673 -78.329 44.814
199716 101.698 -78.343 44.798
199717 101.723 -78.356 44.784
199718 101.747 -78.369 44.770
199719 101.771 -78.382 44.755
199720 101.796 -78.395 44.740
199721 101.820 -78.409 44.724
199722 101.845 -78.422 44.709
199723 101.869 -78.435 44.694
199724 101.894 -78.448 44.680
199725 101.923 -78.467 44.665
199726 101.954 -78.487 44.649
199727 101.999 -78.510 44.634
199728 102.044 -78.533 44.618
199729 102.090 -78.556 44.602
199730 102.135 -78.579 44.587
199731 102.180 -78.602 44.572
199732 102.226 -78.625 44.557
199733 102.271 -78.648 44.541
199734 102.316 -78.671 44.526
199735 102.362 -78.694 44.511
199736 102.406 -78.720 44.495
199737 102.450 -78.752 44.480
199738 102.469 -78.753 44.466
199739 102.488 -78.755 44.450
199740 102.508 -78.756 44.436
199741 102.527 -78.758 44.421
199742 102.547 -78.760 44.407
199743 102.566 -78.761 44.391
199744 102.586 -78.763 44.377
199745 102.601 -78.765 44.362
199746 102.616 -78.767 44.348
199747 102.632 -78.770 44.333
199748 102.647 -78.772 44.317
199749 102.663 -78.774 44.302
200034 93.530 -76.130 54.027
200035 93.549 -76.133 54.039
200036 93.567 -76.136 54.052
200037 93.585 -76.139 54.064
200038 93.604 -76.142 54.075
200039 93.622 -76.145 54.087
200040 93.640 -76.148 54.100
200041 93.658 -76.152 54.112
200042 93.677 -76.155 54.125
200043 93.693 -76.163 54.140
200044 93.697 -76.197 54.152
200045 93.701 -76.231 54.165
200046 93.737 -76.216 54.180
200047 93.769 -76.204 54.190
200048 93.801 -76.190 54.200
200049 93.834 -76.177 54.211
200050 93.854 -76.169 54.223
200051 93.872 -76.161 54.233
200052 93.890 -76.154 54.243
200053 93.899 -76.147 54.254
200054 93.902 -76.141 54.268
200055 93.920 -76.135 54.280
200056 93.941 -76.130 54.291
200057 93.962 -76.125 54.301
200058 93.984 -76.120 54.305
200059 94.014 -76.122 54.307
200060 94.045 -76.124 54.308
200061 94.076 -76.126 54.310
200062 94.107 -76.128 54.310
200063 94.137 -76.129 54.311
200064 94.166 -76.129 54.308
200065 94.195 -76.130 54.299
200066 94.217 -76.125 54.283
200067 94.237 -76.121 54.265
200068 94.254 -76.118 54.245
200069 94.272 -76.116 54.222
200070 94.290 -76.113 54.197
200071 94.308 -76.110 54.169
200072 94.326 -76.108 54.137
200073 94.344 -76.105 54.103
200074 94.362 -76.103 54.069
200075 94.380 -76.100 54.036
200076 94.398 -76.097 54.003
200077 94.416 -76.095 53.969
200078 94.434 -76.093 53.932
200079 94.449 -76.093 53.895
200080 94.461 -76.094 53.858
200081 94.472 -76.094 53.825
200082 94.484 -76.096 53.795
200083 94.495 -76.097 53.767
200084 94.508 -76.100 53.743
200085 94.522 -76.107 53.714
200086 94.536 -76.114 53.689
200087 94.550 -76.121 53.667
200088 94.564 -76.129 53.648
200089 94.578 -76.135 53.628
200090 94.594 -76.143 53.606
200091 94.615 -76.155 53.580
200092 94.636 -76.166 53.546
200093 94.657 -76.178 53.513
200094 94.669 -76.184 53.480
200095 94.680 -76.189 53.449
200096 94.691 -76.194 53.420
200097 94.702 -76.199 53.390
200098 94.713 -76.204 53.359
200099 94.725 -76.210 53.324
200100 94.736 -76.214 53.288
200101 94.747 -76.220 53.252
200102 94.758 -76.225 53.216
200103 94.769 -76.230 53.181
200104 94.780 -76.235 53.145
200105 94.791 -76.240 53.108
200106 94.803 -76.245 53.069
200107 94.814 -76.250 53.029
200108 94.825 -76.255 52.989
200109 94.836 -76.260 52.948
200110 94.847 -76.266 52.907
200111 94.858 -76.270 52.866
200112 94.869 -76.276 52.822
200113 94.881 -76.280 52.778
200114 94.894 -76.284 52.733
200115 94.906 -76.288 52.688
200116 94.919 -76.291 52.644
200117 94.931 -76.295 52.600
200118 94.944 -76.298 52.556
200119 94.932 -76.320 52.513
200120 94.919 -76.342 52.469
200121 94.933 -76.352 52.425
200122 94.969 -76.354 52.380
200123 95.018 -76.350 52.338
200124 95.067 -76.345 52.295
200125 95.117 -76.340 52.252
200126 95.167 -76.336 52.205
200127 95.216 -76.331 52.157
200128 95.265 -76.326 52.109
200129 95.315 -76.321 52.061
200130 95.365 -76.317 52.019
200131 95.392 -76.320 51.978
200132 95.415 -76.325 51.938
200133 95.438 -76.329 51.898
200134 95.461 -76.334 51.853
200135 95.484 -76.339 51.809
200136 95.508 -76.343 51.767
200137 95.531 -76.348 51.727
200138 95.554 -76.353 51.691
200139 95.577 -76.357 51.658
200140 95.602 -76.360 51.628
200141 95.627 -76.361 51.598
200142 95.653 -76.363 51.568
200143 95.678 -76.364 51.538
200144 95.704 -76.366 51.508
200145 95.730 -76.367 51.475
200146 95.760 -76.393 51.434
200147 95.786 -76.396 51.388
200148 95.812 -76.399 51.338
200149 95.839 -76.402 51.288
200150 95.864 -76.403 51.239
200151 95.886 -76.404 51.192
200152 95.908 -76.408 51.146
200153 95.930 -76.411 51.104
200154 95.952 -76.414 51.059
200155 95.974 -76.417 51.013
200156 95.996 -76.420 50.968
200157 96.018 -76.421 50.925
200158 96.039 -76.421 50.881
200159 96.055 -76.426 50.839
200160 96.068 -76.434 50.800
200161 96.083 -76.439 50.762
200162 96.103 -76.439 50.725
200163 96.122 -76.439 50.690
200164 96.142 -76.438 50.654
200165 96.157 -76.439 50.616
200166 96.170 -76.442 50.578
200167 96.184 -76.445 50.539
200168 96.197 -76.448 50.501
200169 96.217 -76.454 50.462
200170 96.239 -76.461 50.425
200171 96.261 -76.468 50.390
200172 96.283 -76.475 50.359
200173 96.300 -76.480 50.332
200174 96.304 -76.480 50.308
200175 96.309 -76.479 50.284
200176 96.316 -76.479 50.262
200177 96.331 -76.481 50.242
200178 96.365 -76.503 50.222
200179 96.406 -76.533 50.200
200180 96.447 -76.563 50.177
200181 96.484 -76.590 50.154
200182 96.492 -76.596 50.131
200183 96.516 -76.603 50.107
200184 96.559 -76.611 50.082
200185 96.576 -76.599 50.055
200186 96.601 -76.599 50.026
200187 96.626 -76.602 49.993
200188 96.655 -76.603 49.959
200189 96.701 -76.595 49.924
200190 96.746 -76.586 49.886
200191 96.792 -76.578 49.849
200192 96.829 -76.572 49.811
200193 96.857 -76.568 49.774
200194 96.884 -76.564 49.739
200195 96.911 -76.560 49.703
200196 96.938 -76.581 49.667
200197 96.962 -76.598 49.629
200198 96.980 -76.600 49.592
200199 96.998 -76.602 49.553
200200 97.015 -76.605 49.516
200201 97.033 -76.607 49.480
200202 97.050 -76.610 49.447
200203 97.045 -76.617 49.414
200204 97.038 -76.624 49.384
200205 97.070 -76.644 49.355
200206 97.085 -76.673 49.331
200207 97.103 -76.698 49.308
200208 97.141 -76.691 49.287
200209 97.159 -76.706 49.267
200210 97.171 -76.727 49.249
200211 97.183 -76.749 49.230
200212 97.195 -76.770 49.209
200213 97.207 -76.791 49.185
200214 97.219 -76.812 49.159
200215 97.231 -76.834 49.132
200216 97.243 -76.855 49.104
200217 97.255 -76.876 49.076
200218 97.318 -76.898 49.048
200219 97.383 -76.919 49.019
200220 97.452 -76.928 48.990
200221 97.488 -76.965 48.961
200222 97.509 -77.008 48.933
200223 97.517 -77.024 48.906
200224 97.525 -77.040 48.879
200225 97.532 -77.056 48.852
200226 97.540 -77.072 48.824
200227 97.549 -77.087 48.797
200228 97.570 -77.085 48.768
200229 97.591 -77.081 48.739
200230 97.607 -77.085 48.710
200231 97.609 -77.071 48.680
200232 97.609 -77.053 48.651
200233 97.612 -77.047 48.623
200234 97.629 -77.055 48.594
200235 97.651 -77.059 48.563
200236 97.673 -77.065 48.533
200237 97.696 -77.070 48.503
200238 97.719 -77.075 48.474
200239 97.742 -77.081 48.444
200240 97.765 -77.087 48.416
200241 97.788 -77.093 48.387
200242 97.811 -77.098 48.357
200243 97.835 -77.104 48.328
200244 97.859 -77.100 48.300
200245 97.883 -77.085 48.271
200246 97.909 -77.068 48.242
200247 97.938 -77.057 48.213
200248 97.969 -77.047 48.182
200249 97.999 -77.038 48.151
200250 98.025 -77.034 48.119
200251 98.045 -77.040 48.088
200252 98.065 -77.045 48.056
200253 98.085 -77.050 48.024
200254 98.105 -77.056 47.993
200255 98.125 -77.061 47.960
200256 98.145 -77.066 47.928
200257 98.165 -77.072 47.897
200258 98.185 -77.078 47.865
200259 98.205 -77.083 47.833
200260 98.229 -77.073 47.801
200261 98.255 -77.060 47.770
200262 98.280 -77.049 47.740
200263 98.305 -77.049 47.710
200264 98.329 -77.049 47.681
200265 98.354 -77.047 47.652
200266 98.378 -77.041 47.623
200267 98.402 -77.036 47.595
200268 98.426 -77.031 47.567
200269 98.450 -77.025 47.540
200270 98.474 -77.019 47.514
200271 98.498 -77.013 47.488
200272 98.522 -77.008 47.462
200273 98.546 -77.002 47.437
200274 98.565 -77.000 47.412
200275 98.558 -77.019 47.389
200276 98.551 -77.036 47.366
200277 98.545 -77.053 47.344
200278 98.537 -77.071 47.320
200279 98.531 -77.088 47.295
200280 98.553 -77.095 47.270
200281 98.602 -77.102 47.245
200282 98.652 -77.110 47.223
200283 98.702 -77.116 47.199
200284 98.752 -77.123 47.174
200285 98.802 -77.131 47.148
200286 98.852 -77.138 47.122
200287 98.867 -77.151 47.095
200288 98.881 -77.165 47.068
200289 98.895 -77.179 47.042
200290 98.909 -77.192 47.018
200291 98.923 -77.206 46.995
200292 98.937 -77.219 46.971
200293 98.951 -77.234 46.947
200294 98.966 -77.244 46.922
200295 98.980 -77.255 46.897
200296 98.995 -77.266 46.872
200297 99.009 -77.277 46.846
200298 99.024 -77.288 46.820
200299 99.039 -77.299 46.793
200300 99.053 -77.310 46.766
200301 99.068 -77.321 46.737
200302 99.081 -77.326 46.708
200303 99.093 -77.332 46.681
200304 99.106 -77.336 46.654
200305 99.119 -77.341 46.628
200306 99.132 -77.346 46.601
200307 99.145 -77.351 46.574
200308 99.156 -77.357 46.549
200309 99.167 -77.363 46.524
200310 99.178 -77.368 46.499
200311 99.188 -77.374 46.475
200312 99.199 -77.379 46.451
200313 99.210 -77.385 46.427
200314 99.220 -77.391 46.402
200315 99.231 -77.397 46.376
200316 99.241 -77.402 46.352
200317 99.257 -77.407 46.328
200318 99.289 -77.411 46.304
200319 99.321 -77.416 46.281
200320 99.352 -77.420 46.257
200321 99.384 -77.424 46.233
200322 99.415 -77.428 46.209
200323 99.447 -77.432 46.186
200324 99.477 -77.435 46.164
200325 99.496 -77.431 46.142
200326 99.516 -77.427 46.120
200327 99.536 -77.423 46.099
200328 99.555 -77.419 46.078
200329 99.574 -77.415 46.057
200330 99.594 -77.410 46.039
200331 99.614 -77.407 46.020
200332 99.633 -77.403 46.001
200333 99.653 -77.398 45.983
200334 99.673 -77.395 45.965
200335 99.688 -77.389 45.948
200336 99.699 -77.380 45.932
200337 99.717 -77.378 45.916
200338 99.751 -77.392 45.900
200339 99.787 -77.405 45.886
200340 99.832 -77.417 45.871
200341 99.876 -77.430 45.856
200342 99.921 -77.443 45.843
200343 99.966 -77.456 45.828
200344 100.011 -77.469 45.814
200345 100.056 -77.481 45.800
200346 100.087 -77.494 45.788
200347 100.101 -77.505 45.774
200348 100.114 -77.517 45.760
200349 100.127 -77.528 45.746
200350 100.140 -77.539 45.734
200351 100.154 -77.551 45.721
200352 100.170 -77.562 45.707
200353 100.185 -77.573 45.693
200354 100.200 -77.585 45.681
200355 100.216 -77.596 45.668
200356 100.234 -77.603 45.654
200357 100.253 -77.611 45.641
200358 100.271 -77.618 45.628
200359 100.290 -77.634 45.615
200360 100.319 -77.649 45.603
200361 100.350 -77.663 45.590
200362 100.382 -77.677 45.577
200363 100.414 -77.690 45.565
200364 100.440 -77.699 45.552
200365 100.463 -77.704 45.540
200366 100.485 -77.709 45.527
200367 100.508 -77.714 45.513
200368 100.531 -77.720 45.499
200369 100.553 -77.725 45.485
200370 100.575 -77.730 45.469
200371 100.598 -77.735 45.453
200372 100.624 -77.754 45.438
200373 100.648 -77.764 45.424
200374 100.676 -77.790 45.410
200375 100.703 -77.816 45.395
200376 100.732 -77.842 45.380
200377 100.756 -77.859 45.366
200378 100.778 -77.871 45.352
200379 100.800 -77.883 45.338
200380 100.823 -77.895 45.324
200381 100.845 -77.907 45.311
200382 100.867 -77.919 45.298
200383 100.890 -77.931 45.285
200384 100.912 -77.943 45.273
200385 100.934 -77.955 45.261
200386 100.957 -77.967 45.250
200387 100.979 -77.978 45.239
200388 101.001 -77.990 45.227
200389 101.024 -78.003 45.215
200390 101.046 -78.015 45.203
200391 101.069 -78.022 45.189
200392 101.093 -78.025 45.176
200393 101.116 -78.027 45.163
200394 101.135 -78.028 45.150
200395 101.140 -78.028 45.135
200396 101.170 -78.043 45.121
200397 101.208 -78.063 45.106
200398 101.247 -78.082 45.090
200399 101.283 -78.099 45.076
200400 101.320 -78.116 45.061
200401 101.356 -78.132 45.046
200402 101.392 -78.149 45.031
200403 101.429 -78.166 45.015
200404 101.455 -78.179 45.000
200405 101.472 -78.190 44.986
200406 101.489 -78.200 44.970
200407 101.506 -78.212 44.954
200408 101.523 -78.222 44.937
200409 101.540 -78.234 44.922
200410 101.557 -78.244 44.905
200411 101.579 -78.257 44.888
200412 101.603 -78.270 44.872
200413 101.627 -78.283 44.855
200414 101.652 -78.297 44.839
200415 101.676 -78.310 44.823
200416 101.701 -78.322 44.808
200417 101.725 -78.336 44.793
200418 101.750 -78.349 44.779
200419 101.774 -78.363 44.765
200420 101.799 -78.376 44.750
200421 101.823 -78.389 44.735
200422 101.847 -78.402 44.720
200423 101.872 -78.415 44.705
200424 101.896 -78.429 44.690
200425 101.921 -78.442 44.677
200426 101.945 -78.455 44.662
200427 101.970 -78.468 44.646
200428 102.016 -78.492 44.630
200429 102.061 -78.514 44.614
200430 102.106 -78.537 44.599
200431 102.152 -78.561 44.583
200432 102.197 -78.583 44.568
200433 102.242 -78.606 44.553
200434 102.287 -78.630 44.539
200435 102.333 -78.652 44.523
200436 102.378 -78.676 44.507
200437 102.423 -78.698 44.492
200438 102.468 -78.726 44.477
200439 102.494 -78.736 44.462
200440 102.514 -78.737 44.446
200441 102.533 -78.739 44.432
200442 102.552 -78.740 44.417
200443 102.569 -78.742 44.402
200444 102.585 -78.745 44.388
200445 102.600 -78.747 44.373
200446 102.615 -78.749 44.358
200447 102.631 -78.752 44.343
200448 102.647 -78.754 44.328
200449 102.662 -78.756 44.313
200450 102.677 -78.758 44.298
200735 93.522 -76.107 54.019
200736 93.541 -76.110 54.032
200737 93.559 -76.113 54.046
200738 93.577 -76.116 54.058
200739 93.595 -76.119 54.070
200740 93.614 -76.122 54.083
200741 93.632 -76.126 54.096
200742 93.650 -76.129 54.110
200743 93.669 -76.132 54.123
200744 93.685 -76.139 54.139
200745 93.689 -76.173 54.152
200746 93.695 -76.204 54.166
200747 93.732 -76.190 54.182
200748 93.767 -76.174 54.194
200749 93.799 -76.160 54.205
200750 93.821 -76.152 54.215
200751 93.839 -76.144 54.228
200752 93.856 -76.137 54.240
200753 93.873 -76.129 54.252
200754 93.884 -76.122 54.264
200755 93.901 -76.117 54.275
200756 93.921 -76.112 54.287
200757 93.942 -76.107 54.300
200758 93.963 -76.101 54.310
200759 93.987 -76.097 54.315
200760 94.017 -76.099 54.317
200761 94.048 -76.101 54.319
200762 94.078 -76.103 54.321
200763 94.109 -76.105 54.322
200764 94.140 -76.107 54.323
200765 94.170 -76.108 54.321
200766 94.199 -76.108 54.312
200767 94.219 -76.104 54.298
200768 94.237 -76.101 54.280
200769 94.255 -76.099 54.258
200770 94.273 -76.096 54.232
200771 94.291 -76.094 54.206
200772 94.309 -76.091 54.176
200773 94.327 -76.088 54.143
200774 94.345 -76.085 54.108
200775 94.362 -76.083 54.074
200776 94.381 -76.080 54.039
200777 94.398 -76.078 54.005
200778 94.416 -76.075 53.970
200779 94.434 -76.073 53.932
200780 94.451 -76.072 53.895
200781 94.462 -76.073 53.859
200782 94.474 -76.074 53.825
200783 94.485 -76.075 53.794
200784 94.498 -76.081 53.764
200785 94.512 -76.088 53.737
200786 94.527 -76.094 53.707
200787 94.541 -76.101 53.681
200788 94.555 -76.108 53.658
200789 94.569 -76.115 53.639
200790 94.583 -76.122 53.620
200791 94.597 -76.129 53.599
200792 94.616 -76.140 53.573
200793 94.637 -76.151 53.540
200794 94.659 -76.163 53.506
200795 94.670 -76.168 53.474
200796 94.681 -76.173 53.443
200797 94.692 -76.179 53.415
200798 94.703 -76.183 53.386
200799 94.714 -76.188 53.355
200800 94.725 -76.194 53.322
200801 94.737 -76.199 53.288
200802 94.748 -76.204 53.253
200803 94.759 -76.209 53.218
200804 94.770 -76.214 53.183
200805 94.781 -76.220 53.149
200806 94.792 -76.224 53.113
200807 94.803 -76.229 53.075
200808 94.814 -76.235 53.036
200809 94.825 -76.240 52.997
200810 94.836 -76.245 52.958
200811 94.847 -76.250 52.918
200812 94.858 -76.255 52.877
200813 94.870 -76.260 52.833
200814 94.881 -76.265 52.788
200815 94.892 -76.270 52.743
200816 94.904 -76.274 52.699
200817 94.917 -76.278 52.655
200818 94.929 -76.282 52.610
200819 94.942 -76.285 52.564
200820 94.931 -76.305 52.521
200821 94.918 -76.328 52.479
200822 94.923 -76.342 52.436
200823 94.965 -76.341 52.393
200824 95.015 -76.336 52.353
200825 95.064 -76.332 52.312
200826 95.114 -76.327 52.269
200827 95.163 -76.323 52.224
200828 95.212 -76.318 52.177
200829 95.262 -76.313 52.130
200830 95.312 -76.308 52.083
200831 95.361 -76.304 52.040
200832 95.396 -76.304 51.999
200833 95.419 -76.309 51.958
200834 95.442 -76.314 51.918
200835 95.465 -76.318 51.875
200836 95.488 -76.323 51.832
200837 95.512 -76.327 51.789
200838 95.535 -76.332 51.747
200839 95.558 -76.337 51.709
200840 95.583 -76.339 51.674
200841 95.608 -76.341 51.642
200842 95.633 -76.342 51.612
200843 95.659 -76.344 51.582
200844 95.685 -76.346 51.553
200845 95.710 -76.348 51.522
200846 95.736 -76.349 51.489
200847 95.763 -76.361 51.449
200848 95.790 -76.364 51.403
200849 95.816 -76.368 51.352
200850 95.842 -76.373 51.302
200851 95.869 -76.379 51.254
200852 95.890 -76.384 51.207
200853 95.912 -76.389 51.162
200854 95.934 -76.392 51.118
200855 95.956 -76.395 51.073
200856 95.977 -76.398 51.027
200857 96.000 -76.401 50.983
200858 96.021 -76.403 50.939
200859 96.043 -76.403 50.896
200860 96.056 -76.411 50.855
200861 96.072 -76.416 50.815
200862 96.091 -76.415 50.778
200863 96.110 -76.414 50.742
200864 96.130 -76.414 50.708
200865 96.149 -76.414 50.673
200866 96.164 -76.416 50.637
200867 96.177 -76.419 50.600
200868 96.190 -76.422 50.563
200869 96.204 -76.424 50.524
200870 96.218 -76.428 50.486
200871 96.240 -76.435 50.449
200872 96.262 -76.442 50.416
200873 96.284 -76.449 50.384
200874 96.294 -76.451 50.357
200875 96.299 -76.451 50.332
200876 96.304 -76.450 50.310
200877 96.309 -76.449 50.288
200878 96.327 -76.458 50.267
200879 96.368 -76.489 50.246
200880 96.409 -76.518 50.222
200881 96.450 -76.548 50.199
200882 96.491 -76.578 50.176
200883 96.501 -76.584 50.153
200884 96.541 -76.592 50.130
200885 96.557 -76.580 50.105
200886 96.579 -76.575 50.078
200887 96.605 -76.578 50.048
200888 96.630 -76.580 50.015
200889 96.658 -76.583 49.981
200890 96.703 -76.574 49.945
200891 96.749 -76.566 49.907
200892 96.794 -76.558 49.870
200893 96.840 -76.549 49.831
200894 96.878 -76.543 49.793
200895 96.905 -76.539 49.757
200896 96.933 -76.535 49.721
200897 96.959 -76.555 49.684
200898 96.979 -76.563 49.646
200899 96.997 -76.565 49.608
200900 97.015 -76.567 49.569
200901 97.032 -76.571 49.531
200902 97.029 -76.577 49.495
200903 97.023 -76.584 49.461
200904 97.017 -76.591 49.429
200905 97.016 -76.600 49.399
200906 97.069 -76.633 49.371
200907 97.084 -76.661 49.346
200908 97.099 -76.690 49.323
200909 97.124 -76.704 49.303
200910 97.162 -76.697 49.283
200911 97.190 -76.702 49.266
200912 97.203 -76.723 49.247
200913 97.214 -76.744 49.226
200914 97.226 -76.765 49.202
200915 97.238 -76.787 49.176
200916 97.250 -76.808 49.149
200917 97.262 -76.829 49.122
200918 97.275 -76.850 49.093
200919 97.329 -76.871 49.064
200920 97.394 -76.893 49.035
200921 97.460 -76.914 49.007
200922 97.497 -76.945 48.977
200923 97.505 -76.961 48.949
200924 97.512 -76.977 48.921
200925 97.520 -76.994 48.893
200926 97.528 -77.009 48.866
200927 97.536 -77.026 48.838
200928 97.544 -77.042 48.810
200929 97.556 -77.052 48.780
200930 97.577 -77.049 48.751
200931 97.597 -77.046 48.722
200932 97.618 -77.044 48.693
200933 97.623 -77.034 48.664
200934 97.628 -77.030 48.636
200935 97.650 -77.035 48.606
200936 97.672 -77.041 48.575
200937 97.696 -77.046 48.545
200938 97.719 -77.052 48.516
200939 97.742 -77.057 48.486
200940 97.765 -77.063 48.456
200941 97.788 -77.069 48.428
200942 97.812 -77.075 48.398
200943 97.835 -77.080 48.369
200944 97.858 -77.086 48.340
200945 97.882 -77.084 48.310
200946 97.907 -77.068 48.281
200947 97.932 -77.052 48.252
200948 97.956 -77.035 48.224
200949 97.987 -77.025 48.193
200950 98.017 -77.016 48.162
200951 98.042 -77.014 48.130
200952 98.062 -77.019 48.097
200953 98.082 -77.025 48.065
200954 98.102 -77.030 48.033
200955 98.122 -77.035 48.000
200956 98.142 -77.041 47.967
200957 98.162 -77.046 47.935
200958 98.182 -77.051 47.904
200959 98.202 -77.056 47.872
200960 98.222 -77.062 47.841
200961 98.242 -77.067 47.809
200962 98.266 -77.059 47.778
200963 98.291 -77.053 47.748
200964 98.315 -77.051 47.719
200965 98.340 -77.046 47.690
200966 98.364 -77.040 47.662
200967 98.387 -77.034 47.633
200968 98.412 -77.029 47.606
200969 98.436 -77.023 47.578
200970 98.460 -77.018 47.551
200971 98.484 -77.012 47.525
200972 98.508 -77.006 47.499
200973 98.532 -77.000 47.474
200974 98.556 -76.995 47.449
200975 98.579 -76.990 47.425
200976 98.573 -77.007 47.401
200977 98.566 -77.025 47.379
200978 98.559 -77.043 47.355
200979 98.553 -77.060 47.331
200980 98.546 -77.078 47.306
200981 98.575 -77.081 47.281
200982 98.616 -77.085 47.256
200983 98.666 -77.093 47.233
200984 98.716 -77.100 47.209
200985 98.766 -77.107 47.183
200986 98.816 -77.114 47.155
200987 98.865 -77.121 47.129
200988 98.879 -77.135 47.102
200989 98.893 -77.149 47.074
200990 98.907 -77.163 47.046
200991 98.921 -77.176 47.021
200992 98.935 -77.190 46.997
200993 98.948 -77.204 46.973
200994 98.962 -77.217 46.949
200995 98.977 -77.230 46.924
200996 98.991 -77.241 46.900
200997 99.006 -77.252 46.875
200998 99.020 -77.263 46.850
200999 99.035 -77.274 46.823
201000 99.050 -77.285 46.795
201001 99.064 -77.296 46.767
201002 99.079 -77.306 46.738
201003 99.091 -77.311 46.709
201004 99.104 -77.316 46.682
201005 99.117 -77.321 46.654
201006 99.130 -77.326 46.627
201007 99.143 -77.331 46.600
201008 99.155 -77.336 46.574
201009 99.168 -77.341 46.547
201010 99.181 -77.346 46.522
201011 99.191 -77.351 46.498
201012 99.202 -77.357 46.473
201013 99.213 -77.363 46.448
201014 99.223 -77.369 46.424
201015 99.234 -77.374 46.399
201016 99.244 -77.380 46.373
201017 99.255 -77.385 46.349
201018 99.285 -77.390 46.326
201019 99.316 -77.394 46.301
201020 99.348 -77.398 46.277
201021 99.379 -77.402 46.253
201022 99.411 -77.407 46.229
201023 99.443 -77.410 46.205
201024 99.475 -77.415 46.183
201025 99.506 -77.419 46.162
201026 99.538 -77.423 46.140
201027 99.561 -77.421 46.118
201028 99.580 -77.417 46.097
201029 99.600 -77.414 46.076
201030 99.620 -77.409 46.056
201031 99.639 -77.405 46.037
201032 99.658 -77.401 46.018
201033 99.678 -77.397 45.999
201034 99.697 -77.393 45.981
201035 99.709 -77.385 45.964
201036 99.721 -77.376 45.947
201037 99.733 -77.369 45.931
201038 99.745 -77.360 45.915
201039 99.779 -77.374 45.899
201040 99.813 -77.388 45.885
201041 99.847 -77.401 45.870
201042 99.885 -77.415 45.855
201043 99.930 -77.428 45.841
201044 99.975 -77.441 45.827
201045 100.019 -77.453 45.813
201046 100.064 -77.466 45.798
201047 100.109 -77.479 45.786
201048 100.122 -77.490 45.772
201049 100.135 -77.501 45.757
201050 100.148 -77.513 45.744
201051 100.162 -77.524 45.731
201052 100.177 -77.536 45.718
201053 100.193 -77.548 45.705
201054 100.208 -77.559 45.691
201055 100.225 -77.567 45.678
201056 100.244 -77.575 45.665
201057 100.262 -77.582 45.651
201058 100.280 -77.589 45.638
201059 100.299 -77.599 45.625
201060 100.318 -77.615 45.612
201061 100.343 -77.630 45.598
201062 100.374 -77.644 45.586
201063 100.406 -77.658 45.573
201064 100.438 -77.672 45.560
201065 100.461 -77.677 45.547
201066 100.484 -77.683 45.534
201067 100.506 -77.688 45.521
201068 100.529 -77.693 45.508
201069 100.552 -77.699 45.494
201070 100.574 -77.704 45.479
201071 100.597 -77.709 45.464
201072 100.619 -77.714 45.449
201073 100.642 -77.724 45.435
201074 100.668 -77.739 45.421
201075 100.695 -77.757 45.407
201076 100.722 -77.783 45.391
201077 100.750 -77.809 45.377
201078 100.778 -77.835 45.363
201079 100.806 -77.862 45.349
201080 100.829 -77.875 45.335
201081 100.851 -77.887 45.322
201082 100.874 -77.899 45.308
201083 100.896 -77.911 45.295
201084 100.918 -77.923 45.282
201085 100.941 -77.935 45.270
201086 100.963 -77.947 45.258
201087 100.986 -77.959 45.247
201088 101.008 -77.971 45.236
201089 101.030 -77.983 45.224
201090 101.053 -77.995 45.212
201091 101.075 -78.006 45.199
201092 101.098 -78.016 45.186
201093 101.121 -78.018 45.173
201094 101.145 -78.022 45.160
201095 101.161 -78.022 45.147
201096 101.166 -78.022 45.132
201097 101.182 -78.028 45.118
201098 101.221 -78.049 45.103
201099 101.260 -78.069 45.089
201100 101.299 -78.089 45.074
201101 101.335 -78.106 45.058
201102 101.372 -78.122 45.043
201103 101.408 -78.139 45.028
201104 101.445 -78.155 45.012
201105 101.475 -78.170 44.996
201106 101.492 -78.181 44.981
201107 101.509 -78.192 44.966
201108 101.526 -78.203 44.949
201109 101.543 -78.214 44.933
201110 101.560 -78.225 44.918
201111 101.581 -78.237 44.901
201112 101.606 -78.250 44.884
201113 101.630 -78.263 44.868
201114 101.655 -78.277 44.851
201115 101.679 -78.290 44.834
201116 101.703 -78.303 44.818
201117 101.728 -78.316 44.803
201118 101.752 -78.329 44.788
201119 101.777 -78.343 44.774
201120 101.801 -78.356 44.759
201121 101.826 -78.369 44.745
201122 101.850 -78.382 44.730
201123 101.875 -78.395 44.715
201124 101.899 -78.409 44.701
201125 101.920 -78.420 44.687
201126 101.941 -78.432 44.672
201127 101.961 -78.444 44.657
201128 101.995 -78.456 44.642
201129 102.036 -78.477 44.627
201130 102.069 -78.499 44.611
201131 102.112 -78.522 44.595
201132 102.160 -78.545 44.579
201133 102.207 -78.567 44.564
201134 102.254 -78.589 44.549
201135 102.301 -78.612 44.534
201136 102.348 -78.634 44.519
201137 102.395 -78.657 44.503
201138 102.440 -78.680 44.488
201139 102.485 -78.703 44.473
201140 102.519 -78.718 44.458
201141 102.537 -78.720 44.442
201142 102.552 -78.723 44.427
201143 102.568 -78.725 44.413
201144 102.584 -78.727 44.397
201145 102.599 -78.730 44.383
201146 102.614 -78.731 44.369
201147 102.630 -78.733 44.354
201148 102.646 -78.736 44.338
201149 102.661 -78.738 44.323
201150 102.677 -78.740 44.308
201151 102.692 -78.742 44.293
201437 93.533 -76.086 54.024
201438 93.551 -76.089 54.039
201439 93.569 -76.092 54.053
201440 93.587 -76.095 54.066
201441 93.606 -76.099 54.079
201442 93.624 -76.102 54.092
201443 93.642 -76.105 54.106
201444 93.660 -76.109 54.121
201445 93.677 -76.116 54.137
201446 93.681 -76.149 54.152
201447 93.690 -76.176 54.167
201448 93.723 -76.168 54.183
201449 93.765 -76.144 54.196
201450 93.788 -76.135 54.208
201451 93.805 -76.127 54.218
201452 93.823 -76.120 54.232
201453 93.841 -76.113 54.246
201454 93.859 -76.105 54.259
201455 93.881 -76.099 54.271
201456 93.902 -76.094 54.282
201457 93.923 -76.088 54.294
201458 93.944 -76.083 54.306
201459 93.965 -76.078 54.317
201460 93.989 -76.075 54.323
201461 94.019 -76.076 54.326
201462 94.050 -76.079 54.328
201463 94.081 -76.080 54.330
201464 94.112 -76.082 54.332
201465 94.142 -76.084 54.333
201466 94.173 -76.086 54.330
201467 94.202 -76.087 54.321
201468 94.220 -76.085 54.308
201469 94.238 -76.082 54.290
201470 94.256 -76.079 54.269
201471 94.274 -76.076 54.242
201472 94.291 -76.074 54.215
201473 94.310 -76.071 54.184
201474 94.327 -76.069 54.150
201475 94.345 -76.066 54.115
201476 94.363 -76.063 54.081
201477 94.381 -76.061 54.046
201478 94.399 -76.058 54.011
201479 94.417 -76.056 53.976
201480 94.435 -76.053 53.938
201481 94.452 -76.052 53.900
201482 94.464 -76.053 53.862
201483 94.475 -76.053 53.827
201484 94.489 -76.060 53.794
201485 94.503 -76.068 53.763
201486 94.517 -76.075 53.734
201487 94.531 -76.082 53.704
201488 94.545 -76.088 53.675
201489 94.559 -76.095 53.649
201490 94.573 -76.102 53.629
201491 94.588 -76.109 53.609
201492 94.602 -76.116 53.587
201493 94.618 -76.125 53.563
201494 94.639 -76.136 53.530
201495 94.659 -76.148 53.496
201496 94.670 -76.153 53.463
201497 94.681 -76.157 53.433
201498 94.692 -76.163 53.406
201499 94.703 -76.168 53.378
201500 94.715 -76.173 53.349
201501 94.726 -76.178 53.317
201502 94.737 -76.183 53.283
201503 94.748 -76.188 53.250
201504 94.759 -76.193 53.216
201505 94.770 -76.198 53.184
201506 94.781 -76.204 53.150
201507 94.792 -76.209 53.115
201508 94.804 -76.213 53.078
201509 94.815 -76.219 53.042
201510 94.826 -76.224 53.004
201511 94.837 -76.229 52.966
201512 94.848 -76.234 52.928
201513 94.859 -76.239 52.886
201514 94.870 -76.245 52.842
201515 94.882 -76.250 52.796
201516 94.893 -76.254 52.752
201517 94.904 -76.260 52.709
201518 94.915 -76.264 52.664
201519 94.927 -76.268 52.619
201520 94.940 -76.272 52.573
201521 94.930 -76.292 52.530
201522 94.917 -76.314 52.489
201523 94.913 -76.332 52.448
201524 94.962 -76.328 52.407
201525 95.012 -76.323 52.367
201526 95.061 -76.319 52.327
201527 95.110 -76.314 52.286
201528 95.160 -76.309 52.242
201529 95.209 -76.304 52.197
201530 95.259 -76.300 52.152
201531 95.308 -76.295 52.106
201532 95.357 -76.291 52.061
201533 95.400 -76.289 52.018
201534 95.423 -76.293 51.977
201535 95.447 -76.298 51.936
201536 95.469 -76.302 51.895
201537 95.492 -76.307 51.853
201538 95.516 -76.312 51.810
201539 95.539 -76.317 51.768
201540 95.563 -76.320 51.728
201541 95.589 -76.321 51.691
201542 95.614 -76.323 51.656
201543 95.639 -76.324 51.625
201544 95.665 -76.326 51.594
201545 95.691 -76.327 51.564
201546 95.716 -76.329 51.534
201547 95.742 -76.329 51.500
201548 95.767 -76.332 51.460
201549 95.793 -76.338 51.415
201550 95.820 -76.343 51.365
201551 95.846 -76.348 51.316
201552 95.873 -76.354 51.268
201553 95.895 -76.359 51.222
201554 95.917 -76.364 51.177
201555 95.938 -76.370 51.133
201556 95.960 -76.375 51.088
201557 95.981 -76.379 51.042
201558 96.003 -76.382 50.997
201559 96.025 -76.385 50.953
201560 96.045 -76.386 50.910
201561 96.060 -76.392 50.869
201562 96.079 -76.392 50.830
201563 96.099 -76.391 50.793
201564 96.118 -76.391 50.758
201565 96.137 -76.390 50.725
201566 96.157 -76.389 50.690
201567 96.170 -76.392 50.655
201568 96.184 -76.395 50.619
201569 96.197 -76.398 50.582
201570 96.210 -76.401 50.545
201571 96.224 -76.404 50.508
201572 96.241 -76.409 50.471
201573 96.263 -76.416 50.439
201574 96.284 -76.423 50.408
201575 96.289 -76.422 50.381
201576 96.294 -76.421 50.357
201577 96.299 -76.421 50.335
201578 96.312 -76.427 50.314
201579 96.338 -76.439 50.292
201580 96.379 -76.469 50.270
201581 96.413 -76.503 50.246
201582 96.454 -76.533 50.222
201583 96.495 -76.563 50.199
201584 96.523 -76.574 50.177
201585 96.539 -76.561 50.153
201586 96.556 -76.551 50.127
201587 96.583 -76.554 50.099
201588 96.609 -76.556 50.069
201589 96.635 -76.559 50.036
201590 96.661 -76.561 50.001
201591 96.705 -76.554 49.965
201592 96.751 -76.546 49.928
201593 96.797 -76.537 49.889
201594 96.842 -76.529 49.851
201595 96.888 -76.520 49.812
201596 96.926 -76.514 49.775
201597 96.953 -76.510 49.739
201598 96.979 -76.526 49.701
201599 96.997 -76.528 49.662
201600 97.014 -76.530 49.624
201601 97.013 -76.537 49.585
201602 97.007 -76.544 49.547
201603 97.001 -76.551 49.511
201604 96.995 -76.558 49.476
201605 96.989 -76.565 49.444
201606 97.046 -76.592 49.414
201607 97.068 -76.621 49.386
201608 97.083 -76.649 49.361
201609 97.098 -76.677 49.339
201610 97.113 -76.705 49.318
201611 97.145 -76.710 49.299
201612 97.183 -76.703 49.281
201613 97.221 -76.697 49.262
201614 97.234 -76.718 49.241
201615 97.246 -76.739 49.217
201616 97.258 -76.760 49.191
201617 97.270 -76.781 49.165
201618 97.282 -76.802 49.138
201619 97.294 -76.824 49.109
201620 97.341 -76.845 49.080
201621 97.405 -76.867 49.051
201622 97.471 -76.888 49.023
201623 97.492 -76.903 48.993
201624 97.500 -76.915 48.963
201625 97.508 -76.931 48.935
201626 97.516 -76.947 48.906
201627 97.524 -76.963 48.878
201628 97.532 -76.979 48.850
201629 97.539 -76.995 48.822
201630 97.547 -77.012 48.792
201631 97.563 -77.016 48.762
201632 97.584 -77.013 48.734
201633 97.604 -77.010 48.705
201634 97.625 -77.008 48.676
201635 97.649 -77.010 48.647
201636 97.672 -77.016 48.618
201637 97.696 -77.022 48.587
201638 97.719 -77.028 48.558
201639 97.742 -77.033 48.528
201640 97.765 -77.039 48.499
201641 97.789 -77.044 48.469
201642 97.812 -77.050 48.440
201643 97.835 -77.056 48.411
201644 97.858 -77.062 48.381
201645 97.881 -77.068 48.351
201646 97.905 -77.067 48.321
201647 97.930 -77.051 48.291
201648 97.955 -77.035 48.262
201649 97.980 -77.019 48.234
201650 98.005 -77.003 48.203
201651 98.035 -76.994 48.172
201652 98.059 -76.994 48.141
201653 98.079 -76.999 48.108
201654 98.099 -77.004 48.075
201655 98.119 -77.009 48.042
201656 98.139 -77.015 48.009
201657 98.159 -77.020 47.975
201658 98.179 -77.025 47.943
201659 98.199 -77.031 47.911
201660 98.219 -77.036 47.879
201661 98.239 -77.041 47.848
201662 98.259 -77.046 47.817
201663 98.279 -77.050 47.786
201664 98.301 -77.050 47.757
201665 98.325 -77.044 47.729
201666 98.350 -77.038 47.700
201667 98.373 -77.033 47.671
201668 98.397 -77.028 47.643
201669 98.422 -77.022 47.616
201670 98.446 -77.016 47.588
201671 98.469 -77.010 47.561
201672 98.494 -77.005 47.535
201673 98.518 -76.999 47.510
201674 98.542 -76.994 47.485
201675 98.566 -76.987 47.461
201676 98.590 -76.982 47.437
201677 98.587 -76.997 47.414
201678 98.581 -77.014 47.391
201679 98.574 -77.032 47.366
201680 98.567 -77.050 47.342
201681 98.565 -77.066 47.317
201682 98.598 -77.067 47.292
201683 98.631 -77.069 47.266
201684 98.679 -77.076 47.243
201685 98.729 -77.083 47.217
201686 98.779 -77.091 47.190
201687 98.829 -77.097 47.163
201688 98.877 -77.105 47.135
201689 98.891 -77.119 47.107
201690 98.905 -77.132 47.078
201691 98.919 -77.146 47.050
201692 98.932 -77.160 47.024
201693 98.946 -77.174 46.999
201694 98.960 -77.188 46.974
201695 98.974 -77.201 46.951
201696 98.988 -77.215 46.926
201697 99.002 -77.227 46.901
201698 99.017 -77.238 46.876
201699 99.031 -77.249 46.851
201700 99.046 -77.260 46.825
201701 99.061 -77.271 46.797
201702 99.075 -77.282 46.768
201703 99.089 -77.291 46.738
201704 99.102 -77.295 46.709
201705 99.115 -77.301 46.682
201706 99.128 -77.305 46.654
201707 99.141 -77.310 46.626
201708 99.153 -77.315 46.599
201709 99.166 -77.320 46.572
201710 99.179 -77.326 46.545
201711 99.192 -77.330 46.519
201712 99.205 -77.335 46.495
201713 99.215 -77.341 46.470
201714 99.226 -77.347 46.445
201715 99.237 -77.352 46.420
201716 99.248 -77.358 46.395
201717 99.258 -77.363 46.369
201718 99.280 -77.368 46.346
201719 99.312 -77.373 46.322
201720 99.344 -77.376 46.298
201721 99.375 -77.381 46.274
201722 99.407 -77.385 46.250
201723 99.438 -77.389 46.226
201724 99.470 -77.393 46.202
201725 99.502 -77.398 46.181
201726 99.534 -77.401 46.159
201727 99.565 -77.405 46.137
201728 99.596 -77.410 46.115
201729 99.625 -77.412 46.095
201730 99.645 -77.407 46.074
201731 99.664 -77.404 46.054
201732 99.684 -77.400 46.034
201733 99.704 -77.395 46.016
201734 99.719 -77.389 45.998
201735 99.730 -77.381 45.980
201736 99.743 -77.373 45.963
201737 99.755 -77.365 45.946
201738 99.767 -77.357 45.929
201739 99.779 -77.349 45.913
201740 99.807 -77.357 45.898
201741 99.841 -77.371 45.883
201742 99.875 -77.385 45.869
201743 99.909 -77.398 45.854
201744 99.943 -77.412 45.839
201745 99.983 -77.425 45.825
201746 100.028 -77.438 45.810
201747 100.073 -77.451 45.796
201748 100.118 -77.464 45.783
201749 100.144 -77.476 45.769
201750 100.157 -77.487 45.755
201751 100.170 -77.498 45.741
201752 100.185 -77.510 45.728
201753 100.200 -77.521 45.715
201754 100.216 -77.533 45.701
201755 100.234 -77.539 45.688
201756 100.253 -77.546 45.675
201757 100.271 -77.554 45.661
201758 100.290 -77.561 45.648
201759 100.309 -77.568 45.635
201760 100.327 -77.580 45.621
201761 100.346 -77.596 45.608
201762 100.367 -77.612 45.595
201763 100.399 -77.626 45.582
201764 100.430 -77.640 45.569
201765 100.460 -77.652 45.556
201766 100.482 -77.657 45.542
201767 100.505 -77.662 45.529
201768 100.528 -77.667 45.516
201769 100.550 -77.672 45.502
201770 100.573 -77.677 45.488
201771 100.596 -77.683 45.474
201772 100.618 -77.690 45.460
201773 100.641 -77.701 45.445
201774 100.664 -77.711 45.432
201775 100.690 -77.727 45.417
201776 100.716 -77.744 45.403
201777 100.743 -77.761 45.388
201778 100.770 -77.777 45.374
201779 100.797 -77.802 45.361
201780 100.825 -77.828 45.347
201781 100.853 -77.855 45.332
201782 100.880 -77.879 45.319
201783 100.903 -77.891 45.306
201784 100.925 -77.903 45.293
201785 100.948 -77.915 45.280
201786 100.970 -77.927 45.267
201787 100.992 -77.939 45.254
201788 101.015 -77.951 45.243
201789 101.037 -77.963 45.231
201790 101.059 -77.975 45.220
201791 101.081 -77.987 45.207
201792 101.104 -77.999 45.194
201793 101.126 -78.011 45.181
201794 101.150 -78.013 45.168
201795 101.174 -78.015 45.155
201796 101.187 -78.016 45.142
201797 101.192 -78.016 45.128
201798 101.198 -78.015 45.114
201799 101.234 -78.034 45.099
201800 101.273 -78.054 45.086
201801 101.312 -78.075 45.070
201802 101.350 -78.095 45.055
201803 101.387 -78.112 45.039
201804 101.423 -78.128 45.024
201805 101.460 -78.145 45.008
201806 101.496 -78.162 44.993
201807 101.513 -78.172 44.977
201808 101.530 -78.184 44.962
201809 101.546 -78.194 44.945
201810 101.563 -78.205 44.929
201811 101.584 -78.217 44.914
201812 101.608 -78.231 44.898
201813 101.633 -78.244 44.881
201814 101.657 -78.257 44.864
201815 101.682 -78.270 44.847
201816 101.706 -78.283 44.830
201817 101.731 -78.297 44.814
201818 101.755 -78.310 44.798
201819 101.779 -78.323 44.782
201820 101.804 -78.336 44.768
201821 101.829 -78.349 44.754
201822 101.851 -78.362 44.739
201823 101.871 -78.373 44.725
201824 101.892 -78.385 44.709
201825 101.912 -78.396 44.695
201826 101.933 -78.407 44.681
201827 101.953 -78.419 44.667
201828 101.983 -78.430 44.652
201829 102.020 -78.444 44.637
201830 102.061 -78.465 44.622
201831 102.102 -78.486 44.606
201832 102.137 -78.508 44.590
201833 102.168 -78.531 44.575
201834 102.208 -78.553 44.560
201835 102.255 -78.576 44.545
201836 102.303 -78.598 44.530
201837 102.349 -78.620 44.514
201838 102.397 -78.643 44.499
201839 102.444 -78.665 44.484
201840 102.491 -78.688 44.470
201841 102.536 -78.702 44.455
201842 102.552 -78.705 44.438
201843 102.567 -78.707 44.423
201844 102.583 -78.709 44.409
201845 102.598 -78.711 44.393
201846 102.614 -78.714 44.379
201847 102.629 -78.716 44.364
201848 102.644 -78.718 44.349
201849 102.660 -78.720 44.334
201850 102.676 -78.723 44.318
201851 102.691 -78.725 44.303
201852 102.706 -78.727 44.288
202138 93.524 -76.063 54.017
202139 93.543 -76.066 54.033
202140 93.561 -76.069 54.047
202141 93.579 -76.072 54.060
202142 93.597 -76.075 54.074
202143 93.616 -76.079 54.088
202144 93.634 -76.082 54.103
202145 93.652 -76.085 54.119
202146 93.670 -76.091 54.135
202147 93.674 -76.125 54.150
202148 93.685 -76.148 54.164
202149 93.717 -76.142 54.180
202150 93.759 -76.118 54.194
202151 93.777 -76.111 54.207
202152 93.795 -76.104 54.219
202153 93.814 -76.097 54.234
202154 93.833 -76.089 54.248
202155 93.851 -76.082 54.262
202156 93.874 -76.075 54.275
202157 93.904 -76.070 54.287
202158 93.924 -76.065 54.299
202159 93.945 -76.060 54.310
202160 93.966 -76.054 54.323
202161 93.992 -76.052 54.330
202162 94.022 -76.053 54.334
202163 94.053 -76.056 54.336
202164 94.083 -76.057 54.337
202165 94.114 -76.060 54.339
202166 94.145 -76.061 54.340
202167 94.173 -76.060 54.337
202168 94.201 -76.056 54.327
202169 94.222 -76.054 54.315
202170 94.240 -76.053 54.298
202171 94.258 -76.052 54.277
202172 94.275 -76.050 54.251
202173 94.293 -76.049 54.222
202174 94.311 -76.048 54.190
202175 94.328 -76.047 54.156
202176 94.346 -76.045 54.121
202177 94.364 -76.044 54.087
202178 94.382 -76.041 54.052
202179 94.400 -76.038 54.018
202180 94.418 -76.036 53.982
202181 94.435 -76.033 53.945
202182 94.453 -76.032 53.906
202183 94.466 -76.034 53.868
202184 94.480 -76.041 53.830
202185 94.493 -76.047 53.795
202186 94.507 -76.054 53.762
202187 94.522 -76.061 53.732
202188 94.536 -76.069 53.701
202189 94.550 -76.075 53.671
202190 94.564 -76.082 53.644
202191 94.578 -76.089 53.618
202192 94.592 -76.096 53.597
202193 94.606 -76.103 53.575
202194 94.620 -76.110 53.551
202195 94.640 -76.122 53.518
202196 94.660 -76.132 53.484
202197 94.671 -76.137 53.450
202198 94.682 -76.142 53.420
202199 94.693 -76.147 53.392
202200 94.704 -76.152 53.365
202201 94.715 -76.157 53.337
202202 94.726 -76.163 53.305
202203 94.738 -76.167 53.272
202204 94.749 -76.173 53.238
202205 94.760 -76.178 53.205
202206 94.771 -76.183 53.174
202207 94.782 -76.188 53.143
202208 94.793 -76.193 53.111
202209 94.804 -76.198 53.076
202210 94.815 -76.203 53.041
202211 94.826 -76.208 53.005
202212 94.837 -76.213 52.969
202213 94.848 -76.219 52.931
202214 94.859 -76.223 52.890
202215 94.871 -76.229 52.847
202216 94.882 -76.234 52.803
202217 94.893 -76.239 52.759
202218 94.904 -76.244 52.716
202219 94.915 -76.249 52.672
202220 94.926 -76.254 52.627
202221 94.938 -76.258 52.582
202222 94.930 -76.277 52.541
202223 94.925 -76.295 52.501
202224 94.933 -76.307 52.462
202225 94.959 -76.315 52.421
202226 95.008 -76.310 52.381
202227 95.057 -76.306 52.341
202228 95.107 -76.301 52.302
202229 95.157 -76.296 52.262
202230 95.206 -76.292 52.218
202231 95.255 -76.287 52.174
202232 95.305 -76.282 52.128
202233 95.354 -76.277 52.082
202234 95.404 -76.273 52.037
202235 95.427 -76.277 51.994
202236 95.451 -76.282 51.954
202237 95.473 -76.287 51.914
202238 95.496 -76.292 51.873
202239 95.520 -76.296 51.831
202240 95.544 -76.300 51.790
202241 95.569 -76.301 51.748
202242 95.595 -76.303 51.708
202243 95.620 -76.304 51.670
202244 95.645 -76.306 51.637
202245 95.671 -76.307 51.604
202246 95.697 -76.307 51.573
202247 95.725 -76.304 51.543
202248 95.750 -76.304 51.509
202249 95.773 -76.308 51.469
202250 95.797 -76.314 51.425
202251 95.823 -76.319 51.377
202252 95.850 -76.324 51.328
202253 95.876 -76.329 51.282
202254 95.900 -76.335 51.238
202255 95.922 -76.340 51.192
202256 95.943 -76.345 51.149
202257 95.965 -76.351 51.103
202258 95.986 -76.356 51.057
202259 96.008 -76.361 51.011
202260 96.029 -76.366 50.967
202261 96.048 -76.368 50.924
202262 96.067 -76.368 50.882
202263 96.087 -76.367 50.843
202264 96.107 -76.367 50.808
202265 96.126 -76.367 50.773
202266 96.146 -76.366 50.740
202267 96.164 -76.366 50.707
202268 96.177 -76.369 50.672
202269 96.190 -76.372 50.636
202270 96.204 -76.375 50.600
202271 96.217 -76.378 50.564
202272 96.231 -76.380 50.528
202273 96.244 -76.383 50.493
202274 96.264 -76.389 50.461
202275 96.279 -76.393 50.430
202276 96.284 -76.393 50.404
202277 96.295 -76.398 50.381
202278 96.312 -76.407 50.359
202279 96.329 -76.417 50.338
202280 96.346 -76.426 50.317
202281 96.398 -76.442 50.294
202282 96.442 -76.470 50.270
202283 96.474 -76.506 50.246
202284 96.507 -76.542 50.223
202285 96.520 -76.543 50.201
202286 96.536 -76.530 50.176
202287 96.560 -76.530 50.148
202288 96.587 -76.532 50.119
202289 96.613 -76.535 50.088
202290 96.639 -76.537 50.055
202291 96.665 -76.540 50.021
202292 96.708 -76.533 49.985
202293 96.753 -76.525 49.947
202294 96.799 -76.517 49.908
202295 96.844 -76.508 49.869
202296 96.890 -76.500 49.830
202297 96.936 -76.492 49.793
202298 96.974 -76.485 49.756
202299 96.996 -76.491 49.717
202300 96.997 -76.497 49.678
202301 96.991 -76.504 49.640
202302 96.985 -76.511 49.602
202303 96.979 -76.518 49.563
202304 96.972 -76.526 49.526
202305 96.966 -76.533 49.492
202306 96.980 -76.546 49.459
202307 97.051 -76.580 49.429
202308 97.066 -76.608 49.401
202309 97.081 -76.637 49.377
202310 97.097 -76.665 49.354
202311 97.112 -76.694 49.333
202312 97.127 -76.722 49.313
202313 97.166 -76.716 49.295
202314 97.204 -76.709 49.276
202315 97.242 -76.703 49.255
202316 97.265 -76.712 49.231
202317 97.277 -76.734 49.206
202318 97.289 -76.755 49.181
202319 97.301 -76.776 49.154
202320 97.313 -76.797 49.125
202321 97.352 -76.818 49.096
202322 97.417 -76.840 49.067
202323 97.471 -76.862 49.037
202324 97.485 -76.876 49.007
202325 97.494 -76.886 48.978
202326 97.502 -76.896 48.948
202327 97.511 -76.906 48.918
202328 97.519 -76.916 48.890
202329 97.527 -76.932 48.861
202330 97.535 -76.949 48.832
202331 97.543 -76.965 48.803
202332 97.551 -76.981 48.774
202333 97.570 -76.981 48.745
202334 97.590 -76.978 48.717
202335 97.618 -76.980 48.688
202336 97.657 -76.990 48.658
202337 97.696 -76.998 48.629
202338 97.719 -77.003 48.598
202339 97.742 -77.009 48.569
202340 97.765 -77.015 48.540
202341 97.789 -77.021 48.512
202342 97.812 -77.026 48.482
202343 97.835 -77.032 48.453
202344 97.858 -77.038 48.423
202345 97.881 -77.044 48.393
202346 97.904 -77.049 48.362
202347 97.928 -77.051 48.331
202348 97.953 -77.034 48.302
202349 97.978 -77.019 48.273
202350 98.003 -77.002 48.244
202351 98.027 -76.986 48.213
202352 98.054 -76.972 48.183
202353 98.076 -76.972 48.151
202354 98.096 -76.978 48.119
202355 98.117 -76.983 48.085
202356 98.136 -76.988 48.052
202357 98.156 -76.992 48.019
202358 98.176 -76.995 47.986
202359 98.196 -76.999 47.953
202360 98.216 -77.003 47.921
202361 98.235 -77.006 47.889
202362 98.255 -77.010 47.857
202363 98.275 -77.013 47.826
202364 98.295 -77.018 47.795
202365 98.313 -77.034 47.766
202366 98.335 -77.037 47.738
202367 98.359 -77.031 47.709
202368 98.383 -77.026 47.679
202369 98.408 -77.020 47.650
202370 98.432 -77.015 47.623
202371 98.455 -77.009 47.597
202372 98.480 -77.003 47.571
202373 98.504 -76.997 47.546
202374 98.528 -76.992 47.521
202375 98.552 -76.986 47.496
202376 98.576 -76.981 47.472
202377 98.600 -76.975 47.449
202378 98.602 -76.986 47.426
202379 98.595 -77.003 47.403
202380 98.588 -77.021 47.378
202381 98.582 -77.039 47.353
202382 98.587 -77.051 47.328
202383 98.621 -77.053 47.303
202384 98.653 -77.056 47.276
202385 98.693 -77.059 47.251
202386 98.743 -77.067 47.225
202387 98.793 -77.074 47.197
202388 98.843 -77.081 47.169
202389 98.888 -77.089 47.140
202390 98.903 -77.103 47.111
202391 98.917 -77.116 47.081
202392 98.930 -77.130 47.054
202393 98.944 -77.144 47.027
202394 98.958 -77.157 47.001
202395 98.972 -77.171 46.976
202396 98.986 -77.185 46.951
202397 99.000 -77.199 46.927
202398 99.014 -77.213 46.902
202399 99.028 -77.224 46.877
202400 99.043 -77.235 46.851
202401 99.058 -77.246 46.825
202402 99.072 -77.257 46.797
202403 99.087 -77.267 46.769
202404 99.100 -77.275 46.738
202405 99.113 -77.280 46.709
202406 99.126 -77.285 46.681
202407 99.139 -77.290 46.653
202408 99.151 -77.294 46.625
202409 99.164 -77.300 46.596
202410 99.177 -77.305 46.569
202411 99.190 -77.310 46.542
202412 99.203 -77.315 46.517
202413 99.215 -77.320 46.491
202414 99.228 -77.325 46.466
202415 99.239 -77.330 46.442
202416 99.251 -77.335 46.416
202417 99.261 -77.341 46.391
202418 99.276 -77.347 46.366
202419 99.307 -77.351 46.342
202420 99.339 -77.355 46.318
202421 99.371 -77.359 46.294
202422 99.403 -77.363 46.271
202423 99.434 -77.367 46.247
202424 99.466 -77.372 46.223
202425 99.497 -77.376 46.200
202426 99.529 -77.379 46.179
202427 99.561 -77.384 46.156
202428 99.592 -77.388 46.134
202429 99.624 -77.392 46.113
202430 99.656 -77.396 46.092
202431 99.687 -77.401 46.071
202432 99.709 -77.398 46.052
202433 99.728 -77.394 46.032
202434 99.740 -77.385 46.015
202435 99.752 -77.378 45.997
202436 99.764 -77.370 45.978
202437 99.776 -77.361 45.961
202438 99.788 -77.354 45.944
202439 99.800 -77.345 45.927
202440 99.812 -77.338 45.912
202441 99.835 -77.340 45.897
202442 99.869 -77.354 45.882
202443 99.903 -77.367 45.866
202444 99.937 -77.381 45.851
202445 99.971 -77.395 45.836
202446 100.005 -77.409 45.822
202447 100.039 -77.423 45.807
202448 100.081 -77.436 45.794
202449 100.126 -77.448 45.780
202450 100.165 -77.461 45.766
202451 100.179 -77.472 45.752
202452 100.193 -77.483 45.738
202453 100.208 -77.495 45.725
202454 100.225 -77.504 45.711
202455 100.244 -77.511 45.697
202456 100.262 -77.518 45.685
202457 100.280 -77.525 45.671
202458 100.299 -77.533 45.658
202459 100.318 -77.539 45.645
202460 100.336 -77.547 45.631
202461 100.355 -77.561 45.618
202462 100.373 -77.577 45.605
202463 100.392 -77.593 45.592
202464 100.423 -77.608 45.578
202465 100.455 -77.621 45.565
202466 100.481 -77.630 45.551
202467 100.503 -77.636 45.537
202468 100.526 -77.641 45.524
202469 100.549 -77.646 45.510
202470 100.571 -77.652 45.496
202471 100.594 -77.658 45.483
202472 100.617 -77.667 45.469
202473 100.639 -77.678 45.455
202474 100.662 -77.688 45.441
202475 100.685 -77.699 45.427
202476 100.711 -77.715 45.412
202477 100.738 -77.732 45.398
202478 100.765 -77.749 45.384
202479 100.791 -77.766 45.371
202480 100.818 -77.783 45.357
202481 100.845 -77.799 45.343
202482 100.872 -77.821 45.329
202483 100.900 -77.848 45.316
202484 100.928 -77.874 45.302
202485 100.954 -77.895 45.289
202486 100.976 -77.907 45.276
202487 100.998 -77.919 45.263
202488 101.021 -77.931 45.251
202489 101.043 -77.943 45.239
202490 101.065 -77.955 45.227
202491 101.088 -77.967 45.215
202492 101.110 -77.979 45.202
202493 101.133 -77.991 45.189
202494 101.155 -78.003 45.177
202495 101.178 -78.007 45.164
202496 101.202 -78.009 45.151
202497 101.213 -78.010 45.137
202498 101.218 -78.009 45.123
202499 101.224 -78.009 45.109
202500 101.247 -78.019 45.095
202501 101.285 -78.040 45.081
202502 101.324 -78.060 45.066
202503 101.363 -78.081 45.051
202504 101.402 -78.100 45.035
202505 101.439 -78.119 45.020
202506 101.475 -78.135 45.004
202507 101.512 -78.152 44.988
202508 101.533 -78.164 44.973
202509 101.550 -78.175 44.957
202510 101.567 -78.186 44.942
202511 101.587 -78.197 44.926
202512 101.611 -78.211 44.910
202513 101.635 -78.224 44.893
202514 101.660 -78.237 44.876
202515 101.684 -78.250 44.860
202516 101.709 -78.263 44.843
202517 101.734 -78.277 44.825
202518 101.758 -78.290 44.809
202519 101.781 -78.303 44.793
202520 101.802 -78.314 44.777
202521 101.822 -78.326 44.762
202522 101.842 -78.337 44.748
202523 101.863 -78.349 44.734
202524 101.883 -78.360 44.719
202525 101.904 -78.372 44.703
202526 101.924 -78.383 44.689
202527 101.944 -78.395 44.675
202528 101.973 -78.406 44.662
202529 102.007 -78.416 44.647
202530 102.045 -78.432 44.632
202531 102.086 -78.453 44.616
202532 102.127 -78.474 44.601
202533 102.168 -78.496 44.585
202534 102.205 -78.517 44.570
202535 102.236 -78.540 44.555
202536 102.267 -78.562 44.539
202537 102.304 -78.584 44.524
202538 102.351 -78.607 44.509
202539 102.398 -78.630 44.494
202540 102.445 -78.651 44.479
202541 102.492 -78.674 44.465
202542 102.551 -78.686 44.449
202543 102.567 -78.689 44.434
202544 102.582 -78.691 44.419
202545 102.597 -78.693 44.404
202546 102.613 -78.695 44.388
202547 102.628 -78.698 44.374
202548 102.644 -78.700 44.359
202549 102.659 -78.702 44.344
202550 102.675 -78.705 44.329
202551 102.690 -78.707 44.313
202552 102.705 -78.709 44.298
202553 102.721 -78.711 44.282
202554 102.737 -78.714 44.266
202839 93.516 -76.039 54.011
202840 93.535 -76.042 54.027
202841 93.553 -76.046 54.042
202842 93.571 -76.049 54.055
202843 93.589 -76.052 54.069
202844 93.608 -76.056 54.085
202845 93.626 -76.059 54.100
202846 93.644 -76.062 54.116
202847 93.662 -76.067 54.133
202848 93.665 -76.101 54.147
202849 93.680 -76.121 54.161
202850 93.712 -76.114 54.175
202851 93.752 -76.094 54.189
202852 93.770 -76.088 54.202
202853 93.788 -76.081 54.215
202854 93.806 -76.074 54.232
202855 93.825 -76.066 54.246
202856 93.844 -76.060 54.261
202857 93.864 -76.052 54.275
202858 93.898 -76.047 54.289
202859 93.926 -76.041 54.301
202860 93.947 -76.036 54.315
202861 93.968 -76.031 54.329
202862 93.994 -76.029 54.337
202863 94.025 -76.032 54.341
202864 94.055 -76.033 54.343
202865 94.086 -76.035 54.343
202866 94.116 -76.035 54.344
202867 94.143 -76.032 54.343
202868 94.171 -76.028 54.340
202869 94.199 -76.025 54.331
202870 94.224 -76.022 54.319
202871 94.242 -76.021 54.303
202872 94.260 -76.019 54.283
202873 94.278 -76.018 54.258
202874 94.295 -76.017 54.229
202875 94.313 -76.016 54.196
202876 94.331 -76.015 54.161
202877 94.349 -76.013 54.126
202878 94.366 -76.012 54.091
202879 94.384 -76.011 54.057
202880 94.402 -76.010 54.024
202881 94.419 -76.008 53.989
202882 94.437 -76.007 53.952
202883 94.456 -76.008 53.913
202884 94.470 -76.018 53.875
202885 94.484 -76.028 53.835
202886 94.498 -76.035 53.797
202887 94.512 -76.041 53.762
202888 94.526 -76.048 53.732
202889 94.540 -76.055 53.700
202890 94.554 -76.063 53.669
202891 94.568 -76.069 53.638
202892 94.582 -76.076 53.609
202893 94.597 -76.083 53.585
202894 94.610 -76.090 53.562
202895 94.624 -76.097 53.538
202896 94.642 -76.107 53.506
202897 94.660 -76.116 53.472
202898 94.671 -76.121 53.438
202899 94.682 -76.126 53.407
202900 94.693 -76.132 53.379
202901 94.704 -76.136 53.352
202902 94.716 -76.141 53.324
202903 94.727 -76.147 53.292
202904 94.738 -76.152 53.259
202905 94.749 -76.157 53.225
202906 94.760 -76.162 53.192
202907 94.771 -76.167 53.161
202908 94.782 -76.173 53.132
202909 94.793 -76.177 53.101
202910 94.805 -76.182 53.068
202911 94.816 -76.188 53.034
202912 94.827 -76.193 53.000
202913 94.838 -76.198 52.965
202914 94.849 -76.203 52.929
202915 94.860 -76.208 52.889
202916 94.872 -76.213 52.848
202917 94.883 -76.218 52.805
202918 94.894 -76.223 52.762
202919 94.905 -76.226 52.721
202920 94.916 -76.230 52.678
202921 94.927 -76.233 52.634
202922 94.938 -76.237 52.590
202923 94.973 -76.241 52.549
202924 94.981 -76.253 52.510
202925 94.988 -76.265 52.472
202926 94.997 -76.276 52.433
202927 95.026 -76.284 52.395
202928 95.055 -76.292 52.358
202929 95.104 -76.288 52.322
202930 95.153 -76.283 52.285
202931 95.202 -76.279 52.243
202932 95.252 -76.274 52.198
202933 95.302 -76.269 52.152
202934 95.351 -76.264 52.105
202935 95.400 -76.260 52.059
202936 95.431 -76.261 52.014
202937 95.455 -76.267 51.971
202938 95.477 -76.271 51.930
202939 95.500 -76.276 51.890
202940 95.524 -76.279 51.849
202941 95.550 -76.281 51.807
202942 95.576 -76.282 51.765
202943 95.601 -76.285 51.724
202944 95.626 -76.286 51.684
202945 95.653 -76.285 51.647
202946 95.680 -76.282 51.613
202947 95.707 -76.279 51.581
202948 95.734 -76.276 51.548
202949 95.757 -76.281 51.513
202950 95.780 -76.285 51.474
202951 95.803 -76.289 51.432
202952 95.827 -76.294 51.387
202953 95.853 -76.299 51.341
202954 95.880 -76.304 51.297
202955 95.905 -76.310 51.253
202956 95.926 -76.315 51.208
202957 95.948 -76.320 51.164
202958 95.969 -76.326 51.118
202959 95.991 -76.331 51.071
202960 96.012 -76.336 51.025
202961 96.034 -76.341 50.981
202962 96.055 -76.342 50.937
202963 96.076 -76.344 50.894
202964 96.095 -76.343 50.856
202965 96.114 -76.343 50.820
202966 96.134 -76.342 50.787
202967 96.153 -76.342 50.754
202968 96.170 -76.342 50.722
202969 96.184 -76.345 50.687
202970 96.197 -76.348 50.651
202971 96.210 -76.351 50.616
202972 96.224 -76.354 50.581
202973 96.237 -76.357 50.546
202974 96.251 -76.360 50.513
202975 96.265 -76.364 50.482
202976 96.278 -76.367 50.451
202977 96.295 -76.377 50.426
202978 96.312 -76.387 50.404
202979 96.329 -76.396 50.382
202980 96.346 -76.406 50.362
202981 96.363 -76.415 50.341
202982 96.400 -76.428 50.319
202983 96.459 -76.445 50.295
202984 96.505 -76.473 50.272
202985 96.516 -76.483 50.249
202986 96.526 -76.493 50.226
202987 96.539 -76.505 50.199
202988 96.565 -76.508 50.169
202989 96.591 -76.511 50.137
202990 96.616 -76.513 50.106
202991 96.643 -76.516 50.073
202992 96.669 -76.518 50.038
202993 96.710 -76.512 50.002
202994 96.755 -76.504 49.964
202995 96.801 -76.495 49.927
202996 96.847 -76.487 49.888
202997 96.892 -76.479 49.849
202998 96.939 -76.471 49.811
202999 96.980 -76.467 49.773
203000 96.975 -76.464 49.734
203001 96.969 -76.471 49.695
203002 96.963 -76.478 49.656
203003 96.957 -76.486 49.617
203004 96.951 -76.492 49.578
203005 96.944 -76.500 49.542
203006 96.938 -76.507 49.508
203007 97.011 -76.537 49.475
203008 97.049 -76.568 49.445
203009 97.065 -76.597 49.417
203010 97.080 -76.625 49.392
203011 97.095 -76.654 49.369
203012 97.111 -76.682 49.347
203013 97.126 -76.710 49.329
203014 97.148 -76.728 49.310
203015 97.186 -76.722 49.290
203016 97.225 -76.715 49.269
203017 97.263 -76.709 49.246
203018 97.297 -76.707 49.221
203019 97.309 -76.728 49.196
203020 97.320 -76.749 49.169
203021 97.332 -76.771 49.141
203022 97.363 -76.793 49.112
203023 97.428 -76.814 49.082
203024 97.463 -76.834 49.052
203025 97.478 -76.850 49.020
203026 97.487 -76.859 48.990
203027 97.496 -76.869 48.960
203028 97.504 -76.880 48.930
203029 97.513 -76.890 48.900
203030 97.521 -76.900 48.872
203031 97.530 -76.909 48.843
203032 97.538 -76.920 48.814
203033 97.546 -76.934 48.785
203034 97.556 -76.948 48.756
203035 97.580 -76.947 48.727
203036 97.618 -76.957 48.698
203037 97.657 -76.968 48.669
203038 97.697 -76.977 48.639
203039 97.737 -76.984 48.609
203040 97.766 -76.991 48.579
203041 97.789 -76.997 48.550
203042 97.812 -77.003 48.522
203043 97.835 -77.008 48.493
203044 97.858 -77.014 48.464
203045 97.881 -77.019 48.434
203046 97.904 -77.025 48.404
203047 97.928 -77.031 48.374
203048 97.951 -77.034 48.343
203049 97.976 -77.018 48.312
203050 98.001 -77.002 48.282
203051 98.026 -76.986 48.253
203052 98.051 -76.969 48.223
203053 98.075 -76.954 48.193
203054 98.093 -76.947 48.162
203055 98.113 -76.949 48.130
203056 98.133 -76.952 48.097
203057 98.152 -76.956 48.064
203058 98.172 -76.959 48.031
203059 98.192 -76.963 47.998
203060 98.212 -76.967 47.965
203061 98.232 -76.970 47.932
203062 98.251 -76.974 47.900
203063 98.272 -76.978 47.868
203064 98.291 -76.981 47.836
203065 98.311 -76.987 47.805
203066 98.329 -77.003 47.774
203067 98.348 -77.019 47.746
203068 98.369 -77.025 47.717
203069 98.393 -77.019 47.688
203070 98.417 -77.013 47.658
203071 98.441 -77.007 47.630
203072 98.465 -77.002 47.604
203073 98.490 -76.996 47.580
203074 98.514 -76.990 47.556
203075 98.537 -76.984 47.532
203076 98.562 -76.979 47.508
203077 98.586 -76.973 47.484
203078 98.610 -76.968 47.461
203079 98.617 -76.975 47.438
203080 98.610 -76.993 47.414
203081 98.603 -77.010 47.389
203082 98.596 -77.028 47.364
203083 98.610 -77.038 47.338
203084 98.643 -77.040 47.312
203085 98.676 -77.042 47.285
203086 98.709 -77.044 47.259
203087 98.756 -77.050 47.231
203088 98.806 -77.057 47.203
203089 98.856 -77.065 47.175
203090 98.901 -77.072 47.145
203091 98.914 -77.086 47.115
203092 98.928 -77.100 47.085
203093 98.942 -77.114 47.056
203094 98.956 -77.128 47.028
203095 98.970 -77.141 47.001
203096 98.984 -77.155 46.977
203097 98.998 -77.169 46.951
203098 99.011 -77.182 46.926
203099 99.025 -77.196 46.901
203100 99.040 -77.210 46.877
203101 99.054 -77.220 46.850
203102 99.069 -77.232 46.823
203103 99.083 -77.242 46.796
203104 99.098 -77.254 46.767
203105 99.111 -77.260 46.736
203106 99.124 -77.264 46.707
203107 99.137 -77.269 46.679
203108 99.149 -77.274 46.651
203109 99.162 -77.279 46.622
203110 99.175 -77.284 46.594
203111 99.188 -77.289 46.567
203112 99.201 -77.294 46.540
203113 99.213 -77.299 46.514
203114 99.226 -77.304 46.488
203115 99.239 -77.309 46.463
203116 99.252 -77.314 46.438
203117 99.264 -77.320 46.413
203118 99.275 -77.325 46.388
203119 99.303 -77.329 46.362
203120 99.335 -77.333 46.338
203121 99.367 -77.338 46.314
203122 99.398 -77.341 46.291
203123 99.430 -77.346 46.267
203124 99.462 -77.350 46.244
203125 99.493 -77.354 46.220
203126 99.525 -77.358 46.197
203127 99.557 -77.362 46.175
203128 99.586 -77.366 46.154
203129 99.616 -77.369 46.132
203130 99.645 -77.373 46.111
203131 99.675 -77.376 46.089
203132 99.704 -77.379 46.069
203133 99.742 -77.382 46.050
203134 99.762 -77.382 46.031
203135 99.774 -77.374 46.013
203136 99.786 -77.366 45.995
203137 99.798 -77.358 45.977
203138 99.810 -77.350 45.959
203139 99.822 -77.341 45.942
203140 99.834 -77.334 45.925
203141 99.846 -77.326 45.910
203142 99.863 -77.323 45.895
203143 99.897 -77.337 45.880
203144 99.931 -77.351 45.864
203145 99.965 -77.364 45.848
203146 99.999 -77.378 45.833
203147 100.033 -77.392 45.818
203148 100.067 -77.405 45.804
203149 100.101 -77.419 45.790
203150 100.135 -77.432 45.776
203151 100.179 -77.445 45.762
203152 100.201 -77.457 45.748
203153 100.216 -77.468 45.734
203154 100.234 -77.476 45.721
203155 100.253 -77.483 45.708
203156 100.271 -77.490 45.693
203157 100.290 -77.497 45.680
203158 100.309 -77.504 45.668
203159 100.327 -77.511 45.655
203160 100.345 -77.518 45.642
203161 100.364 -77.526 45.627
203162 100.382 -77.542 45.614
203163 100.401 -77.558 45.602
203164 100.420 -77.574 45.589
203165 100.446 -77.589 45.575
203166 100.479 -77.603 45.562
203167 100.502 -77.609 45.548
203168 100.525 -77.614 45.533
203169 100.547 -77.620 45.520
203170 100.570 -77.625 45.506
203171 100.593 -77.635 45.492
203172 100.615 -77.645 45.478
203173 100.638 -77.655 45.465
203174 100.661 -77.665 45.451
203175 100.684 -77.676 45.436
203176 100.707 -77.686 45.422
203177 100.734 -77.704 45.408
203178 100.760 -77.721 45.394
203179 100.786 -77.737 45.381
203180 100.813 -77.754 45.367
203181 100.840 -77.771 45.353
203182 100.866 -77.788 45.339
203183 100.893 -77.805 45.326
203184 100.920 -77.821 45.313
203185 100.947 -77.841 45.299
203186 100.975 -77.868 45.285
203187 101.003 -77.893 45.272
203188 101.027 -77.912 45.260
203189 101.050 -77.923 45.247
203190 101.072 -77.935 45.235
203191 101.095 -77.947 45.223
203192 101.117 -77.959 45.210
203193 101.139 -77.971 45.197
203194 101.162 -77.983 45.185
203195 101.184 -77.995 45.172
203196 101.207 -78.001 45.159
203197 101.231 -78.004 45.145
203198 101.240 -78.004 45.132
203199 101.245 -78.003 45.118
203200 101.250 -78.003 45.104
203201 101.259 -78.005 45.090
203202 101.298 -78.025 45.076
203203 101.337 -78.046 45.061
203204 101.376 -78.065 45.046
203205 101.414 -78.086 45.030
203206 101.453 -78.106 45.015
203207 101.491 -78.125 44.999
203208 101.527 -78.142 44.984
203209 101.553 -78.155 44.968
203210 101.570 -78.166 44.953
203211 101.590 -78.178 44.937
203212 101.614 -78.191 44.921
203213 101.638 -78.204 44.905
203214 101.663 -78.217 44.888
203215 101.687 -78.231 44.872
203216 101.712 -78.244 44.855
203217 101.733 -78.256 44.838
203218 101.753 -78.267 44.821
203219 101.773 -78.279 44.805
203220 101.793 -78.290 44.788
203221 101.814 -78.301 44.772
203222 101.834 -78.313 44.756
203223 101.854 -78.324 44.741
203224 101.875 -78.335 44.727
203225 101.895 -78.347 44.712
203226 101.915 -78.359 44.697
203227 101.936 -78.370 44.683
203228 101.962 -78.382 44.669
203229 101.996 -78.392 44.656
203230 102.032 -78.402 44.642
203231 102.070 -78.420 44.627
203232 102.111 -78.441 44.611
203233 102.153 -78.463 44.596
203234 102.193 -78.484 44.580
203235 102.235 -78.505 44.565
203236 102.272 -78.526 44.550
203237 102.304 -78.548 44.534
203238 102.335 -78.571 44.519
203239 102.366 -78.593 44.504
203240 102.399 -78.615 44.489
203241 102.447 -78.638 44.474
203242 102.494 -78.661 44.459
203243 102.565 -78.671 44.444
203244 102.581 -78.673 44.429
203245 102.596 -78.676 44.414
203246 102.612 -78.678 44.399
203247 102.627 -78.680 44.383
203248 102.643 -78.683 44.369
203249 102.658 -78.685 44.354
203250 102.674 -78.687 44.339
203251 102.689 -78.689 44.323
203252 102.705 -78.691 44.308
203253 102.720 -78.693 44.292
203254 102.736 -78.695 44.276
203255 102.751 -78.698 44.261
203541 93.526 -76.019 54.020
203542 93.545 -76.022 54.036
203543 93.563 -76.025 54.051
203544 93.581 -76.029 54.066
203545 93.600 -76.032 54.081
203546 93.618 -76.035 54.096
203547 93.636 -76.039 54.113
203548 93.654 -76.043 54.130
203549 93.658 -76.077 54.145
203550 93.675 -76.093 54.157
203551 93.707 -76.087 54.170
203552 93.745 -76.070 54.184
203553 93.763 -76.063 54.197
203554 93.781 -76.057 54.211
203555 93.799 -76.050 54.228
203556 93.818 -76.044 54.244
203557 93.836 -76.036 54.260
203558 93.855 -76.029 54.276
203559 93.887 -76.023 54.291
203560 93.921 -76.018 54.306
203561 93.948 -76.013 54.321
203562 93.969 -76.007 54.337
203563 93.996 -76.007 54.344
203564 94.027 -76.009 54.349
203565 94.058 -76.010 54.350
203566 94.085 -76.007 54.349
203567 94.114 -76.003 54.348
203568 94.141 -76.000 54.346
203569 94.169 -75.997 54.342
203570 94.197 -75.994 54.332
203571 94.224 -75.991 54.320
203572 94.245 -75.988 54.304
203573 94.262 -75.988 54.284
203574 94.280 -75.986 54.260
203575 94.298 -75.985 54.233
203576 94.316 -75.984 54.202
203577 94.333 -75.982 54.165
203578 94.351 -75.981 54.130
203579 94.369 -75.980 54.096
203580 94.387 -75.978 54.063
203581 94.404 -75.978 54.030
203582 94.422 -75.976 53.996
203583 94.441 -75.979 53.959
203584 94.459 -75.983 53.920
203585 94.473 -75.992 53.881
203586 94.487 -76.002 53.841
203587 94.502 -76.011 53.802
203588 94.516 -76.021 53.765
203589 94.530 -76.030 53.733
203590 94.544 -76.040 53.699
203591 94.559 -76.049 53.666
203592 94.573 -76.057 53.633
203593 94.587 -76.063 53.602
203594 94.601 -76.070 53.576
203595 94.615 -76.077 53.551
203596 94.629 -76.084 53.526
203597 94.644 -76.091 53.494
203598 94.661 -76.100 53.460
203599 94.672 -76.106 53.427
203600 94.683 -76.110 53.396
203601 94.694 -76.116 53.367
203602 94.705 -76.121 53.339
203603 94.716 -76.126 53.311
203604 94.728 -76.131 53.280
203605 94.739 -76.136 53.246
203606 94.750 -76.141 53.212
203607 94.761 -76.146 53.179
203608 94.772 -76.151 53.147
203609 94.783 -76.157 53.117
203610 94.794 -76.161 53.087
203611 94.805 -76.164 53.055
203612 94.816 -76.168 53.023
203613 94.827 -76.172 52.990
203614 94.838 -76.176 52.956
203615 94.849 -76.179 52.922
203616 94.860 -76.182 52.884
203617 94.872 -76.186 52.844
203618 94.883 -76.189 52.804
203619 94.894 -76.193 52.763
203620 94.905 -76.197 52.723
203621 94.916 -76.201 52.681
203622 94.927 -76.204 52.639
203623 94.938 -76.207 52.597
203624 94.988 -76.207 52.557
203625 95.036 -76.210 52.518
203626 95.043 -76.223 52.481
203627 95.050 -76.234 52.444
203628 95.067 -76.245 52.408
203629 95.096 -76.252 52.373
203630 95.124 -76.260 52.341
203631 95.153 -76.268 52.307
203632 95.199 -76.265 52.267
203633 95.249 -76.260 52.223
203634 95.298 -76.256 52.176
203635 95.347 -76.251 52.127
203636 95.397 -76.247 52.080
203637 95.436 -76.246 52.034
203638 95.459 -76.251 51.989
203639 95.482 -76.255 51.947
203640 95.505 -76.260 51.905
203641 95.530 -76.261 51.864
203642 95.556 -76.263 51.823
203643 95.582 -76.264 51.781
203644 95.608 -76.263 51.739
203645 95.635 -76.260 51.698
203646 95.662 -76.257 51.659
203647 95.689 -76.255 51.622
203648 95.716 -76.252 51.587
203649 95.741 -76.253 51.553
203650 95.764 -76.257 51.517
203651 95.788 -76.261 51.479
203652 95.810 -76.266 51.440
203653 95.833 -76.270 51.398
203654 95.858 -76.275 51.355
203655 95.884 -76.280 51.312
203656 95.910 -76.285 51.268
203657 95.932 -76.291 51.223
203658 95.953 -76.296 51.177
203659 95.974 -76.301 51.131
203660 95.996 -76.307 51.084
203661 96.017 -76.310 51.038
203662 96.038 -76.312 50.993
203663 96.060 -76.314 50.949
203664 96.081 -76.317 50.906
203665 96.102 -76.319 50.868
203666 96.122 -76.319 50.833
203667 96.142 -76.318 50.800
203668 96.161 -76.317 50.767
203669 96.177 -76.319 50.735
203670 96.190 -76.322 50.700
203671 96.204 -76.325 50.666
203672 96.217 -76.328 50.631
203673 96.231 -76.331 50.598
203674 96.244 -76.333 50.565
203675 96.257 -76.337 50.533
203676 96.269 -76.342 50.502
203677 96.287 -76.351 50.472
203678 96.311 -76.366 50.446
203679 96.329 -76.376 50.424
203680 96.346 -76.386 50.403
203681 96.363 -76.395 50.383
203682 96.380 -76.404 50.363
203683 96.402 -76.414 50.342
203684 96.461 -76.432 50.320
203685 96.504 -76.443 50.297
203686 96.530 -76.451 50.273
203687 96.541 -76.464 50.248
203688 96.551 -76.477 50.219
203689 96.572 -76.483 50.187
203690 96.596 -76.487 50.155
203691 96.621 -76.492 50.123
203692 96.647 -76.494 50.090
203693 96.673 -76.497 50.056
203694 96.713 -76.492 50.020
203695 96.758 -76.483 49.983
203696 96.805 -76.477 49.944
203697 96.856 -76.472 49.905
203698 96.906 -76.468 49.865
203699 96.949 -76.464 49.828
203700 96.956 -76.455 49.790
203701 96.954 -76.445 49.751
203702 96.941 -76.445 49.712
203703 96.935 -76.453 49.672
203704 96.929 -76.460 49.633
203705 96.923 -76.467 49.595
203706 96.916 -76.474 49.559
203707 96.945 -76.492 49.525
203708 97.033 -76.528 49.492
203709 97.048 -76.557 49.461
203710 97.064 -76.585 49.432
203711 97.079 -76.613 49.407
203712 97.094 -76.642 49.384
203713 97.109 -76.670 49.363
203714 97.124 -76.699 49.344
203715 97.140 -76.727 49.325
203716 97.169 -76.734 49.304
203717 97.207 -76.728 49.283
203718 97.245 -76.721 49.260
203719 97.284 -76.715 49.237
203720 97.322 -76.708 49.211
203721 97.340 -76.724 49.185
203722 97.352 -76.745 49.156
203723 97.375 -76.766 49.127
203724 97.438 -76.787 49.096
203725 97.455 -76.806 49.065
203726 97.472 -76.823 49.033
203727 97.480 -76.833 49.001
203728 97.489 -76.843 48.970
203729 97.498 -76.853 48.940
203730 97.506 -76.863 48.911
203731 97.515 -76.873 48.882
203732 97.523 -76.883 48.853
203733 97.532 -76.893 48.824
203734 97.540 -76.903 48.796
203735 97.551 -76.913 48.767
203736 97.580 -76.925 48.738
203737 97.618 -76.935 48.709
203738 97.657 -76.945 48.679
203739 97.696 -76.955 48.650
203740 97.736 -76.963 48.620
203741 97.777 -76.971 48.590
203742 97.812 -76.978 48.561
203743 97.835 -76.984 48.532
203744 97.858 -76.990 48.502
203745 97.881 -76.995 48.474
203746 97.904 -77.001 48.444
203747 97.928 -77.006 48.413
203748 97.951 -77.012 48.383
203749 97.974 -77.018 48.352
203750 97.999 -77.001 48.321
203751 98.024 -76.985 48.291
203752 98.049 -76.969 48.262
203753 98.074 -76.953 48.232
203754 98.093 -76.944 48.202
203755 98.112 -76.937 48.171
203756 98.131 -76.929 48.139
203757 98.149 -76.921 48.107
203758 98.169 -76.923 48.075
203759 98.189 -76.927 48.043
203760 98.208 -76.931 48.009
203761 98.228 -76.934 47.976
203762 98.248 -76.938 47.943
203763 98.268 -76.941 47.910
203764 98.288 -76.945 47.878
203765 98.307 -76.949 47.845
203766 98.327 -76.955 47.814
203767 98.346 -76.972 47.783
203768 98.364 -76.987 47.755
203769 98.382 -77.003 47.726
203770 98.403 -77.012 47.697
203771 98.427 -77.006 47.667
203772 98.451 -77.000 47.638
203773 98.476 -76.994 47.612
203774 98.500 -76.989 47.588
203775 98.523 -76.983 47.565
203776 98.548 -76.978 47.542
203777 98.572 -76.972 47.519
203778 98.596 -76.966 47.495
203779 98.620 -76.961 47.473
203780 98.632 -76.964 47.449
203781 98.625 -76.982 47.425
203782 98.618 -77.000 47.400
203783 98.611 -77.017 47.373
203784 98.633 -77.024 47.347
203785 98.665 -77.026 47.320
203786 98.699 -77.028 47.292
203787 98.731 -77.030 47.266
203788 98.770 -77.034 47.238
203789 98.819 -77.041 47.208
203790 98.870 -77.048 47.179
203791 98.912 -77.056 47.149
203792 98.926 -77.070 47.118
203793 98.940 -77.084 47.087
203794 98.954 -77.097 47.059
203795 98.968 -77.112 47.029
203796 98.982 -77.125 47.001
203797 98.995 -77.139 46.975
203798 99.009 -77.153 46.949
203799 99.023 -77.166 46.924
203800 99.037 -77.180 46.900
203801 99.051 -77.194 46.874
203802 99.065 -77.207 46.848
203803 99.080 -77.217 46.821
203804 99.094 -77.229 46.794
203805 99.109 -77.238 46.765
203806 99.122 -77.244 46.735
203807 99.135 -77.249 46.705
203808 99.147 -77.254 46.677
203809 99.160 -77.259 46.648
203810 99.173 -77.263 46.620
203811 99.186 -77.269 46.592
203812 99.199 -77.274 46.564
203813 99.211 -77.279 46.537
203814 99.224 -77.284 46.510
203815 99.237 -77.288 46.485
203816 99.250 -77.294 46.459
203817 99.262 -77.298 46.434
203818 99.281 -77.298 46.409
203819 99.289 -77.304 46.384
203820 99.319 -77.307 46.358
203821 99.348 -77.311 46.335
203822 99.377 -77.314 46.311
203823 99.407 -77.318 46.287
203824 99.436 -77.321 46.264
203825 99.466 -77.325 46.241
203826 99.495 -77.328 46.218
203827 99.525 -77.332 46.195
203828 99.555 -77.335 46.172
203829 99.584 -77.338 46.151
203830 99.614 -77.341 46.130
203831 99.643 -77.345 46.109
203832 99.683 -77.350 46.087
203833 99.733 -77.357 46.068
203834 99.761 -77.352 46.048
203835 99.792 -77.364 46.029
203836 99.807 -77.362 46.011
203837 99.820 -77.354 45.993
203838 99.832 -77.346 45.975
203839 99.844 -77.338 45.958
203840 99.856 -77.330 45.941
203841 99.868 -77.323 45.924
203842 99.879 -77.314 45.907
203843 99.892 -77.306 45.893
203844 99.925 -77.320 45.877
203845 99.960 -77.333 45.861
203846 99.993 -77.347 45.845
203847 100.028 -77.360 45.830
203848 100.061 -77.375 45.815
203849 100.096 -77.389 45.800
203850 100.129 -77.402 45.786
203851 100.163 -77.416 45.771
203852 100.197 -77.429 45.757
203853 100.223 -77.440 45.743
203854 100.243 -77.447 45.729
203855 100.262 -77.454 45.716
203856 100.281 -77.461 45.702
203857 100.299 -77.469 45.688
203858 100.318 -77.476 45.675
203859 100.336 -77.483 45.663
203860 100.355 -77.490 45.650
203861 100.373 -77.497 45.637
203862 100.392 -77.507 45.623
203863 100.410 -77.523 45.610
203864 100.429 -77.539 45.598
203865 100.448 -77.555 45.585
203866 100.471 -77.571 45.571
203867 100.501 -77.583 45.557
203868 100.524 -77.589 45.543
203869 100.546 -77.594 45.529
203870 100.569 -77.602 45.516
203871 100.592 -77.612 45.502
203872 100.614 -77.622 45.487
203873 100.637 -77.633 45.473
203874 100.659 -77.642 45.460
203875 100.682 -77.653 45.446
203876 100.705 -77.663 45.432
203877 100.728 -77.675 45.417
203878 100.755 -77.692 45.404
203879 100.782 -77.708 45.391
203880 100.809 -77.726 45.377
203881 100.835 -77.743 45.363
203882 100.862 -77.759 45.349
203883 100.889 -77.776 45.336
203884 100.915 -77.793 45.323
203885 100.942 -77.810 45.309
203886 100.968 -77.827 45.296
203887 100.995 -77.843 45.282
203888 101.022 -77.861 45.269
203889 101.050 -77.887 45.256
203890 101.077 -77.913 45.243
203891 101.101 -77.928 45.231
203892 101.123 -77.939 45.218
203893 101.145 -77.951 45.206
203894 101.168 -77.963 45.193
203895 101.190 -77.975 45.179
203896 101.212 -77.987 45.166
203897 101.236 -77.996 45.153
203898 101.259 -77.998 45.140
203899 101.266 -77.998 45.127
203900 101.271 -77.997 45.113
203901 101.276 -77.997 45.099
203902 101.282 -77.996 45.084
203903 101.311 -78.011 45.070
203904 101.349 -78.031 45.055
203905 101.388 -78.051 45.040
203906 101.427 -78.072 45.025
203907 101.466 -78.092 45.009
203908 101.504 -78.112 44.994
203909 101.543 -78.131 44.978
203910 101.573 -78.147 44.963
203911 101.592 -78.158 44.948
203912 101.617 -78.171 44.932
203913 101.641 -78.184 44.916
203914 101.663 -78.197 44.900
203915 101.683 -78.208 44.883
203916 101.704 -78.219 44.867
203917 101.724 -78.231 44.850
203918 101.744 -78.242 44.834
203919 101.765 -78.254 44.818
203920 101.785 -78.266 44.802
203921 101.806 -78.277 44.785
203922 101.826 -78.288 44.767
203923 101.846 -78.300 44.750
203924 101.867 -78.311 44.735
203925 101.887 -78.322 44.721
203926 101.907 -78.335 44.706
203927 101.927 -78.346 44.691
203928 101.951 -78.357 44.677
203929 101.986 -78.367 44.663
203930 102.021 -78.378 44.649
203931 102.055 -78.388 44.635
203932 102.095 -78.407 44.620
203933 102.136 -78.429 44.605
203934 102.177 -78.451 44.590
203935 102.219 -78.472 44.575
203936 102.259 -78.493 44.559
203937 102.301 -78.514 44.544
203938 102.340 -78.535 44.528
203939 102.372 -78.557 44.513
203940 102.403 -78.580 44.499
203941 102.434 -78.602 44.483
203942 102.465 -78.624 44.468
203943 102.500 -78.645 44.453
203944 102.580 -78.655 44.438
203945 102.595 -78.658 44.423
203946 102.611 -78.660 44.409
203947 102.626 -78.662 44.394
203948 102.642 -78.664 44.378
203949 102.657 -78.667 44.363
203950 102.673 -78.669 44.347
203951 102.688 -78.671 44.332
203952 102.704 -78.673 44.317
203953 102.719 -78.676 44.301
203954 102.735 -78.678 44.286
203955 102.755 -78.680 44.270
203956 102.781 -78.683 44.255
204242 93.518 -75.996 54.015
204243 93.537 -75.999 54.031
204244 93.555 -76.003 54.047
204245 93.573 -76.006 54.062
204246 93.591 -76.009 54.077
204247 93.610 -76.013 54.093
204248 93.628 -76.016 54.111
204249 93.646 -76.019 54.128
204250 93.650 -76.053 54.143
204251 93.669 -76.066 54.156
204252 93.702 -76.059 54.168
204253 93.738 -76.047 54.180
204254 93.757 -76.040 54.193
204255 93.775 -76.033 54.207
204256 93.793 -76.026 54.224
204257 93.811 -76.019 54.240
204258 93.829 -76.013 54.257
204259 93.847 -76.006 54.278
204260 93.877 -76.000 54.295
204261 93.910 -75.994 54.313
204262 93.944 -75.989 54.330
204263 93.971 -75.984 54.346
204264 93.999 -75.984 54.351
204265 94.028 -75.983 54.355
204266 94.056 -75.979 54.356
204267 94.083 -75.976 54.355
204268 94.111 -75.972 54.352
204269 94.139 -75.969 54.348
204270 94.167 -75.966 54.342
204271 94.195 -75.963 54.332
204272 94.222 -75.960 54.319
204273 94.247 -75.956 54.303
204274 94.265 -75.955 54.284
204275 94.282 -75.954 54.261
204276 94.300 -75.953 54.236
204277 94.318 -75.951 54.206
204278 94.336 -75.950 54.170
204279 94.353 -75.949 54.134
204280 94.371 -75.947 54.101
204281 94.389 -75.947 54.069
204282 94.407 -75.947 54.037
204283 94.426 -75.950 54.004
204284 94.445 -75.953 53.968
204285 94.462 -75.957 53.929
204286 94.477 -75.966 53.890
204287 94.491 -75.976 53.851
204288 94.505 -75.985 53.811
204289 94.520 -75.995 53.772
204290 94.534 -76.005 53.738
204291 94.548 -76.014 53.701
204292 94.562 -76.024 53.666
204293 94.576 -76.033 53.632
204294 94.591 -76.043 53.600
204295 94.605 -76.052 53.571
204296 94.619 -76.062 53.544
204297 94.633 -76.071 53.518
204298 94.647 -76.078 53.485
204299 94.661 -76.085 53.452
204300 94.672 -76.090 53.418
204301 94.683 -76.095 53.386
204302 94.694 -76.099 53.356
204303 94.705 -76.103 53.328
204304 94.716 -76.107 53.299
204305 94.728 -76.110 53.268
204306 94.739 -76.113 53.234
204307 94.750 -76.117 53.201
204308 94.761 -76.120 53.167
204309 94.772 -76.124 53.135
204310 94.783 -76.128 53.104
204311 94.794 -76.132 53.073
204312 94.805 -76.135 53.040
204313 94.816 -76.138 53.008
204314 94.827 -76.142 52.975
204315 94.838 -76.146 52.942
204316 94.849 -76.149 52.908
204317 94.860 -76.153 52.872
204318 94.872 -76.157 52.834
204319 94.883 -76.160 52.796
204320 94.894 -76.163 52.759
204321 94.905 -76.167 52.720
204322 94.916 -76.171 52.681
204323 94.927 -76.174 52.642
204324 94.938 -76.178 52.603
204325 94.986 -76.179 52.565
204326 95.049 -76.177 52.527
204327 95.098 -76.180 52.490
204328 95.106 -76.191 52.455
204329 95.113 -76.204 52.421
204330 95.136 -76.213 52.388
204331 95.165 -76.220 52.359
204332 95.193 -76.229 52.327
204333 95.222 -76.236 52.290
204334 95.251 -76.245 52.247
204335 95.295 -76.243 52.200
204336 95.344 -76.238 52.149
204337 95.394 -76.233 52.102
204338 95.440 -76.230 52.056
204339 95.463 -76.235 52.008
204340 95.486 -76.239 51.963
204341 95.511 -76.242 51.919
204342 95.536 -76.243 51.877
204343 95.563 -76.242 51.837
204344 95.590 -76.239 51.795
204345 95.617 -76.236 51.754
204346 95.644 -76.233 51.713
204347 95.672 -76.230 51.671
204348 95.698 -76.227 51.632
204349 95.726 -76.225 51.594
204350 95.749 -76.229 51.558
204351 95.771 -76.233 51.522
204352 95.795 -76.238 51.486
204353 95.817 -76.242 51.448
204354 95.840 -76.246 51.409
204355 95.863 -76.251 51.368
204356 95.887 -76.255 51.326
204357 95.914 -76.261 51.282
204358 95.937 -76.267 51.237
204359 95.958 -76.271 51.190
204360 95.979 -76.276 51.143
204361 96.001 -76.280 51.096
204362 96.022 -76.282 51.050
204363 96.043 -76.284 51.005
204364 96.064 -76.286 50.961
204365 96.086 -76.289 50.918
204366 96.107 -76.290 50.881
204367 96.128 -76.292 50.846
204368 96.150 -76.294 50.813
204369 96.169 -76.294 50.780
204370 96.184 -76.295 50.748
204371 96.197 -76.298 50.714
204372 96.210 -76.301 50.679
204373 96.223 -76.304 50.646
204374 96.236 -76.308 50.614
204375 96.249 -76.311 50.583
204376 96.261 -76.314 50.552
204377 96.273 -76.322 50.522
204378 96.285 -76.329 50.492
204379 96.307 -76.341 50.466
204380 96.331 -76.355 50.443
204381 96.355 -76.369 50.423
204382 96.379 -76.383 50.404
204383 96.397 -76.393 50.385
204384 96.414 -76.403 50.365
204385 96.461 -76.417 50.344
204386 96.491 -76.423 50.319
204387 96.521 -76.430 50.294
204388 96.550 -76.436 50.267
204389 96.567 -76.447 50.237
204390 96.580 -76.458 50.205
204391 96.604 -76.462 50.173
204392 96.629 -76.466 50.140
204393 96.654 -76.470 50.108
204394 96.680 -76.474 50.074
204395 96.722 -76.477 50.038
204396 96.772 -76.473 50.001
204397 96.823 -76.469 49.962
204398 96.874 -76.465 49.922
204399 96.919 -76.461 49.882
204400 96.925 -76.452 49.845
204401 96.931 -76.444 49.808
204402 96.937 -76.436 49.769
204403 96.915 -76.422 49.730
204404 96.906 -76.427 49.689
204405 96.900 -76.434 49.650
204406 96.894 -76.442 49.613
204407 96.888 -76.448 49.576
204408 96.975 -76.483 49.542
204409 97.031 -76.517 49.509
204410 97.047 -76.545 49.477
204411 97.062 -76.573 49.447
204412 97.078 -76.602 49.422
204413 97.093 -76.630 49.400
204414 97.108 -76.658 49.379
204415 97.123 -76.687 49.359
204416 97.139 -76.715 49.339
204417 97.154 -76.743 49.319
204418 97.190 -76.740 49.298
204419 97.228 -76.734 49.275
204420 97.266 -76.727 49.252
204421 97.305 -76.721 49.227
204422 97.343 -76.715 49.200
204423 97.371 -76.718 49.171
204424 97.386 -76.740 49.141
204425 97.430 -76.760 49.109
204426 97.448 -76.778 49.077
204427 97.465 -76.796 49.045
204428 97.474 -76.806 49.012
204429 97.483 -76.817 48.980
204430 97.491 -76.827 48.950
204431 97.500 -76.837 48.921
204432 97.508 -76.846 48.892
204433 97.517 -76.857 48.863
204434 97.525 -76.867 48.835
204435 97.542 -76.878 48.806
204436 97.566 -76.890 48.778
204437 97.589 -76.901 48.749
204438 97.618 -76.912 48.720
204439 97.657 -76.922 48.690
204440 97.696 -76.933 48.660
204441 97.735 -76.941 48.631
204442 97.776 -76.950 48.600
204443 97.816 -76.957 48.571
204444 97.856 -76.965 48.542
204445 97.881 -76.972 48.511
204446 97.904 -76.977 48.483
204447 97.928 -76.983 48.453
204448 97.951 -76.988 48.422
204449 97.974 -76.994 48.391
204450 97.998 -77.000 48.360
204451 98.022 -76.985 48.330
204452 98.047 -76.969 48.300
204453 98.072 -76.953 48.271
204454 98.094 -76.941 48.240
204455 98.112 -76.934 48.209
204456 98.131 -76.927 48.179
204457 98.149 -76.919 48.147
204458 98.168 -76.911 48.116
204459 98.186 -76.903 48.085
204460 98.205 -76.895 48.053
204461 98.225 -76.898 48.020
204462 98.244 -76.902 47.986
204463 98.264 -76.906 47.953
204464 98.284 -76.909 47.921
204465 98.304 -76.913 47.886
204466 98.324 -76.916 47.854
204467 98.343 -76.924 47.823
204468 98.362 -76.940 47.793
204469 98.380 -76.956 47.765
204470 98.398 -76.972 47.736
204471 98.417 -76.988 47.707
204472 98.437 -76.999 47.677
204473 98.461 -76.993 47.649
204474 98.485 -76.987 47.622
204475 98.509 -76.981 47.597
204476 98.533 -76.976 47.573
204477 98.558 -76.970 47.550
204478 98.582 -76.965 47.528
204479 98.605 -76.959 47.506
204480 98.630 -76.953 47.483
204481 98.646 -76.953 47.459
204482 98.639 -76.971 47.435
204483 98.633 -76.989 47.409
204484 98.626 -77.006 47.382
204485 98.655 -77.010 47.355
204486 98.688 -77.012 47.328
204487 98.721 -77.014 47.299
204488 98.754 -77.016 47.272
204489 98.787 -77.019 47.243
204490 98.833 -77.025 47.213
204491 98.883 -77.031 47.183
204492 98.924 -77.041 47.152
204493 98.938 -77.054 47.120
204494 98.951 -77.068 47.088
204495 98.966 -77.081 47.059
204496 98.979 -77.095 47.028
204497 98.993 -77.109 47.000
204498 99.007 -77.122 46.973
204499 99.021 -77.137 46.946
204500 99.035 -77.150 46.921
204501 99.049 -77.164 46.896
204502 99.063 -77.178 46.871
204503 99.076 -77.191 46.845
204504 99.091 -77.203 46.818
204505 99.106 -77.214 46.791
204506 99.120 -77.223 46.762
204507 99.132 -77.229 46.732
204508 99.145 -77.233 46.703
204509 99.158 -77.238 46.674
204510 99.171 -77.243 46.645
204511 99.184 -77.248 46.618
204512 99.196 -77.253 46.590
204513 99.209 -77.258 46.561
204514 99.222 -77.263 46.533
204515 99.237 -77.266 46.507
204516 99.263 -77.260 46.481
204517 99.289 -77.259 46.455
204518 99.311 -77.274 46.430
204519 99.316 -77.286 46.404
204520 99.302 -77.280 46.379
204521 99.316 -77.280 46.354
204522 99.346 -77.284 46.331
204523 99.375 -77.287 46.307
204524 99.405 -77.291 46.284
204525 99.434 -77.294 46.261
204526 99.464 -77.298 46.238
204527 99.493 -77.301 46.214
204528 99.522 -77.304 46.192
204529 99.552 -77.307 46.170
204530 99.581 -77.311 46.148
204531 99.616 -77.315 46.127
204532 99.667 -77.322 46.107
204533 99.718 -77.329 46.085
204534 99.749 -77.326 46.066
204535 99.779 -77.322 46.047
204536 99.813 -77.332 46.028
204537 99.838 -77.346 46.009
204538 99.853 -77.343 45.991
204539 99.865 -77.335 45.974
204540 99.877 -77.326 45.956
204541 99.889 -77.319 45.939
204542 99.901 -77.310 45.922
204543 99.913 -77.303 45.906
204544 99.925 -77.294 45.890
204545 99.954 -77.303 45.873
204546 99.988 -77.316 45.857
204547 100.021 -77.330 45.841
204548 100.055 -77.344 45.826
204549 100.089 -77.357 45.810
204550 100.123 -77.371 45.795
204551 100.157 -77.385 45.780
204552 100.191 -77.398 45.765
204553 100.224 -77.412 45.751
204554 100.235 -77.418 45.737
204555 100.253 -77.426 45.724
204556 100.272 -77.433 45.711
204557 100.291 -77.440 45.697
204558 100.311 -77.448 45.683
204559 100.330 -77.454 45.670
204560 100.349 -77.462 45.657
204561 100.368 -77.469 45.645
204562 100.387 -77.476 45.632
204563 100.410 -77.490 45.619
204564 100.434 -77.505 45.606
204565 100.457 -77.520 45.593
204566 100.476 -77.536 45.581
204567 100.495 -77.552 45.567
204568 100.522 -77.562 45.553
204569 100.545 -77.569 45.539
204570 100.567 -77.579 45.524
204571 100.590 -77.589 45.511
204572 100.613 -77.599 45.496
204573 100.635 -77.610 45.482
204574 100.658 -77.620 45.468
204575 100.681 -77.630 45.454
204576 100.703 -77.640 45.441
204577 100.726 -77.651 45.427
204578 100.750 -77.663 45.413
204579 100.777 -77.680 45.400
204580 100.804 -77.697 45.386
204581 100.830 -77.714 45.372
204582 100.857 -77.730 45.359
204583 100.884 -77.748 45.344
204584 100.910 -77.765 45.331
204585 100.937 -77.781 45.319
204586 100.964 -77.798 45.304
204587 100.990 -77.815 45.291
204588 101.017 -77.832 45.278
204589 101.044 -77.849 45.264
204590 101.070 -77.865 45.252
204591 101.097 -77.883 45.239
204592 101.124 -77.906 45.226
204593 101.152 -77.931 45.213
204594 101.174 -77.943 45.200
204595 101.197 -77.956 45.187
204596 101.219 -77.967 45.173
204597 101.241 -77.979 45.160
204598 101.264 -77.990 45.148
204599 101.286 -77.992 45.135
204600 101.292 -77.992 45.121
204601 101.297 -77.991 45.107
204602 101.302 -77.990 45.092
204603 101.308 -77.990 45.078
204604 101.323 -77.996 45.063
204605 101.362 -78.016 45.049
204606 101.401 -78.037 45.034
204607 101.440 -78.057 45.018
204608 101.478 -78.078 45.002
204609 101.517 -78.097 44.987
204610 101.556 -78.118 44.972
204611 101.594 -78.138 44.957
204612 101.614 -78.149 44.942
204613 101.634 -78.160 44.926
204614 101.655 -78.172 44.910
204615 101.675 -78.184 44.894
204616 101.695 -78.195 44.878
204617 101.715 -78.206 44.862
204618 101.736 -78.218 44.846
204619 101.756 -78.229 44.830
204620 101.777 -78.241 44.813
204621 101.797 -78.253 44.796
204622 101.818 -78.264 44.779
204623 101.838 -78.275 44.762
204624 101.858 -78.287 44.745
204625 101.879 -78.298 44.729
204626 101.899 -78.310 44.714
204627 101.919 -78.321 44.699
204628 101.941 -78.332 44.684
204629 101.975 -78.343 44.670
204630 102.010 -78.354 44.656
204631 102.045 -78.364 44.642
204632 102.080 -78.375 44.627
204633 102.120 -78.395 44.613
204634 102.161 -78.417 44.598
204635 102.202 -78.438 44.583
204636 102.243 -78.460 44.568
204637 102.284 -78.481 44.553
204638 102.325 -78.502 44.537
204639 102.367 -78.523 44.521
204640 102.407 -78.545 44.507
204641 102.440 -78.567 44.492
204642 102.470 -78.589 44.477
204643 102.501 -78.611 44.462
204644 102.533 -78.633 44.447
204645 102.591 -78.639 44.432
204646 102.610 -78.642 44.417
204647 102.625 -78.644 44.403
204648 102.641 -78.646 44.387
204649 102.657 -78.648 44.372
204650 102.672 -78.651 44.356
204651 102.687 -78.653 44.342
204652 102.703 -78.655 44.327
204653 102.719 -78.658 44.311
204654 102.740 -78.660 44.295
204655 102.766 -78.663 44.279
204656 102.792 -78.665 44.263
204657 102.817 -78.668 44.248
204943 93.510 -75.972 54.010
204944 93.529 -75.976 54.027
204945 93.547 -75.979 54.043
204946 93.565 -75.982 54.058
204947 93.583 -75.986 54.075
204948 93.602 -75.989 54.091
204949 93.620 -75.992 54.109
204950 93.638 -75.996 54.127
204951 93.642 -76.028 54.142
204952 93.664 -76.038 54.155
204953 93.696 -76.032 54.167
204954 93.729 -76.025 54.179
204955 93.750 -76.016 54.192
204956 93.768 -76.009 54.205
204957 93.787 -76.002 54.222
204958 93.807 -75.994 54.238
204959 93.829 -75.987 54.255
204960 93.850 -75.979 54.275
204961 93.876 -75.973 54.295
204962 93.907 -75.969 54.315
204963 93.939 -75.964 54.333
204964 93.971 -75.960 54.347
204965 93.998 -75.955 54.352
204966 94.026 -75.951 54.357
204967 94.054 -75.948 54.360
204968 94.081 -75.945 54.358
204969 94.109 -75.941 54.354
204970 94.137 -75.938 54.349
204971 94.165 -75.934 54.342
204972 94.192 -75.931 54.331
204973 94.220 -75.928 54.317
204974 94.248 -75.925 54.302
204975 94.267 -75.923 54.283
204976 94.285 -75.922 54.261
204977 94.303 -75.921 54.237
204978 94.321 -75.919 54.210
204979 94.338 -75.918 54.176
204980 94.356 -75.917 54.140
204981 94.374 -75.916 54.107
204982 94.393 -75.919 54.076
204983 94.411 -75.921 54.045
204984 94.430 -75.924 54.013
204985 94.449 -75.926 53.977
204986 94.466 -75.931 53.939
204987 94.480 -75.941 53.902
204988 94.494 -75.950 53.862
204989 94.508 -75.960 53.822
204990 94.523 -75.969 53.783
204991 94.537 -75.979 53.746
204992 94.551 -75.988 53.708
204993 94.565 -75.998 53.670
204994 94.580 -76.007 53.634
204995 94.594 -76.017 53.601
204996 94.608 -76.027 53.570
204997 94.623 -76.036 53.540
204998 94.637 -76.046 53.513
204999 94.652 -76.054 53.480
205000 94.664 -76.060 53.447
205001 94.673 -76.063 53.415
205002 94.684 -76.066 53.383
205003 94.695 -76.069 53.351
205004 94.706 -76.073 53.322
205005 94.717 -76.077 53.291
205006 94.728 -76.080 53.258
205007 94.739 -76.084 53.225
205008 94.750 -76.088 53.192
205009 94.761 -76.091 53.159
205010 94.772 -76.094 53.128
205011 94.783 -76.098 53.097
205012 94.794 -76.102 53.065
205013 94.805 -76.105 53.032
205014 94.816 -76.109 52.998
205015 94.827 -76.113 52.964
205016 94.838 -76.116 52.930
205017 94.849 -76.119 52.895
205018 94.860 -76.123 52.860
205019 94.872 -76.127 52.823
205020 94.883 -76.130 52.787
205021 94.894 -76.134 52.751
205022 94.905 -76.138 52.715
205023 94.916 -76.141 52.679
205024 94.927 -76.144 52.643
205025 94.938 -76.148 52.608
205026 94.985 -76.149 52.573
205027 95.048 -76.148 52.538
205028 95.110 -76.148 52.503
205029 95.161 -76.149 52.469
205030 95.168 -76.161 52.436
205031 95.177 -76.173 52.404
205032 95.205 -76.181 52.374
205033 95.234 -76.188 52.344
205034 95.263 -76.197 52.310
205035 95.291 -76.204 52.270
205036 95.320 -76.213 52.223
205037 95.348 -76.220 52.172
205038 95.390 -76.220 52.124
205039 95.440 -76.216 52.076
205040 95.467 -76.219 52.028
205041 95.491 -76.221 51.981
205042 95.518 -76.220 51.935
205043 95.545 -76.217 51.890
205044 95.572 -76.214 51.847
205045 95.600 -76.211 51.806
205046 95.626 -76.209 51.765
205047 95.654 -76.206 51.724
205048 95.681 -76.203 51.683
205049 95.708 -76.200 51.643
205050 95.733 -76.201 51.603
205051 95.756 -76.205 51.565
205052 95.779 -76.210 51.529
205053 95.802 -76.214 51.494
205054 95.825 -76.219 51.458
205055 95.847 -76.223 51.418
205056 95.871 -76.227 51.380
205057 95.893 -76.232 51.339
205058 95.917 -76.236 51.295
205059 95.942 -76.242 51.249
205060 95.963 -76.247 51.203
205061 95.984 -76.250 51.155
205062 96.006 -76.252 51.107
205063 96.027 -76.254 51.062
205064 96.048 -76.256 51.017
205065 96.070 -76.258 50.973
205066 96.091 -76.260 50.931
205067 96.112 -76.262 50.895
205068 96.133 -76.264 50.860
205069 96.155 -76.266 50.827
205070 96.177 -76.267 50.793
205071 96.189 -76.273 50.759
205072 96.202 -76.276 50.725
205073 96.215 -76.279 50.692
205074 96.228 -76.282 50.660
205075 96.240 -76.285 50.630
205076 96.253 -76.289 50.600
205077 96.265 -76.295 50.569
205078 96.277 -76.301 50.539
205079 96.289 -76.308 50.510
205080 96.302 -76.316 50.485
205081 96.326 -76.330 50.464
205082 96.350 -76.344 50.444
205083 96.375 -76.358 50.426
205084 96.399 -76.373 50.407
205085 96.423 -76.386 50.388
205086 96.455 -76.395 50.365
205087 96.478 -76.404 50.339
205088 96.508 -76.411 50.311
205089 96.538 -76.417 50.282
205090 96.568 -76.423 50.251
205091 96.592 -76.430 50.221
205092 96.612 -76.437 50.190
205093 96.640 -76.442 50.158
205094 96.676 -76.449 50.126
205095 96.712 -76.456 50.091
205096 96.749 -76.464 50.055
205097 96.791 -76.466 50.018
205098 96.841 -76.461 49.979
205099 96.888 -76.458 49.939
205100 96.894 -76.449 49.900
205101 96.900 -76.441 49.864
205102 96.906 -76.433 49.827
205103 96.912 -76.424 49.789
205104 96.904 -76.414 49.750
205105 96.878 -76.401 49.709
205106 96.872 -76.408 49.669
205107 96.866 -76.416 49.630
205108 96.909 -76.438 49.593
205109 97.005 -76.475 49.558
205110 97.030 -76.505 49.525
205111 97.045 -76.533 49.494
205112 97.061 -76.561 49.464
205113 97.076 -76.589 49.439
205114 97.091 -76.618 49.415
205115 97.107 -76.646 49.393
205116 97.122 -76.674 49.373
205117 97.138 -76.703 49.353
205118 97.153 -76.731 49.333
205119 97.172 -76.752 49.313
205120 97.211 -76.746 49.290
205121 97.249 -76.740 49.266
205122 97.287 -76.734 49.241
205123 97.325 -76.727 49.215
205124 97.364 -76.721 49.185
205125 97.402 -76.714 49.153
205126 97.422 -76.732 49.121
205127 97.440 -76.751 49.088
205128 97.458 -76.769 49.055
205129 97.467 -76.781 49.022
205130 97.476 -76.790 48.990
205131 97.485 -76.800 48.959
205132 97.493 -76.810 48.930
205133 97.502 -76.820 48.901
205134 97.511 -76.831 48.872
205135 97.534 -76.842 48.844
205136 97.557 -76.854 48.817
205137 97.580 -76.865 48.788
205138 97.603 -76.878 48.760
205139 97.627 -76.889 48.730
205140 97.657 -76.900 48.701
205141 97.696 -76.910 48.671
205142 97.735 -76.920 48.640
205143 97.775 -76.928 48.610
205144 97.815 -76.936 48.580
205145 97.856 -76.943 48.550
205146 97.896 -76.952 48.521
205147 97.928 -76.959 48.491
205148 97.951 -76.964 48.460
205149 97.974 -76.970 48.430
205150 97.998 -76.975 48.399
205151 98.021 -76.981 48.367
205152 98.045 -76.969 48.337
205153 98.070 -76.952 48.308
205154 98.094 -76.937 48.279
205155 98.112 -76.931 48.248
205156 98.131 -76.925 48.217
205157 98.149 -76.917 48.187
205158 98.168 -76.909 48.156
205159 98.186 -76.901 48.125
205160 98.205 -76.893 48.094
205161 98.224 -76.885 48.062
205162 98.242 -76.878 48.028
205163 98.261 -76.869 47.995
205164 98.281 -76.873 47.962
205165 98.300 -76.877 47.930
205166 98.320 -76.881 47.897
205167 98.340 -76.884 47.865
205168 98.359 -76.893 47.834
205169 98.378 -76.909 47.804
205170 98.396 -76.925 47.775
205171 98.415 -76.941 47.746
205172 98.433 -76.957 47.717
205173 98.452 -76.973 47.689
205174 98.471 -76.986 47.661
205175 98.495 -76.980 47.634
205176 98.519 -76.975 47.609
205177 98.544 -76.969 47.583
205178 98.568 -76.963 47.560
205179 98.591 -76.958 47.537
205180 98.616 -76.952 47.515
205181 98.640 -76.947 47.492
205182 98.661 -76.943 47.468
205183 98.654 -76.960 47.443
205184 98.647 -76.978 47.418
205185 98.645 -76.994 47.390
205186 98.677 -76.996 47.363
205187 98.711 -76.998 47.335
205188 98.743 -77.000 47.305
205189 98.777 -77.003 47.278
205190 98.809 -77.004 47.248
205191 98.847 -77.008 47.218
205192 98.897 -77.015 47.186
205193 98.935 -77.024 47.154
205194 98.949 -77.038 47.122
205195 98.963 -77.052 47.089
205196 98.977 -77.066 47.059
205197 98.991 -77.079 47.028
205198 99.005 -77.093 46.998
205199 99.019 -77.106 46.971
205200 99.032 -77.120 46.943
205201 99.047 -77.134 46.917
205202 99.060 -77.147 46.892
205203 99.074 -77.162 46.867
205204 99.088 -77.175 46.841
205205 99.102 -77.189 46.814
205206 99.117 -77.200 46.787
205207 99.130 -77.207 46.758
205208 99.143 -77.213 46.729
205209 99.156 -77.218 46.700
205210 99.169 -77.222 46.672
205211 99.182 -77.228 46.643
205212 99.194 -77.232 46.614
205213 99.219 -77.228 46.586
205214 99.245 -77.222 46.558
205215 99.271 -77.216 46.530
205216 99.296 -77.219 46.503
205217 99.321 -77.235 46.477
205218 99.337 -77.249 46.452
205219 99.343 -77.263 46.426
205220 99.336 -77.264 46.401
205221 99.323 -77.258 46.376
205222 99.314 -77.253 46.351
205223 99.343 -77.257 46.327
205224 99.373 -77.260 46.303
205225 99.402 -77.263 46.280
205226 99.432 -77.266 46.257
205227 99.462 -77.270 46.234
205228 99.491 -77.273 46.211
205229 99.520 -77.276 46.188
205230 99.550 -77.280 46.167
205231 99.600 -77.287 46.145
205232 99.651 -77.294 46.124
205233 99.702 -77.301 46.104
205234 99.739 -77.301 46.083
205235 99.764 -77.294 46.064
205236 99.797 -77.291 46.044
205237 99.833 -77.297 46.025
205238 99.862 -77.318 46.007
205239 99.885 -77.328 45.989
205240 99.899 -77.323 45.972
205241 99.911 -77.315 45.954
205242 99.923 -77.307 45.936
205243 99.935 -77.299 45.919
205244 99.947 -77.291 45.903
205245 99.959 -77.283 45.886
205246 99.982 -77.285 45.870
205247 100.015 -77.299 45.853
205248 100.050 -77.313 45.837
205249 100.083 -77.326 45.821
205250 100.118 -77.340 45.805
205251 100.151 -77.354 45.789
205252 100.186 -77.368 45.774
205253 100.219 -77.382 45.759
205254 100.240 -77.389 45.745
205255 100.251 -77.395 45.731
205256 100.263 -77.405 45.717
205257 100.282 -77.412 45.704
205258 100.301 -77.420 45.691
205259 100.321 -77.426 45.678
205260 100.340 -77.433 45.665
205261 100.359 -77.441 45.652
205262 100.379 -77.448 45.639
205263 100.400 -77.458 45.627
205264 100.424 -77.473 45.614
205265 100.447 -77.489 45.601
205266 100.471 -77.504 45.589
205267 100.495 -77.519 45.576
205268 100.519 -77.534 45.562
205269 100.543 -77.546 45.548
205270 100.566 -77.556 45.534
205271 100.588 -77.567 45.520
205272 100.611 -77.577 45.505
205273 100.634 -77.587 45.491
205274 100.656 -77.597 45.477
205275 100.680 -77.608 45.463
205276 100.702 -77.617 45.449
205277 100.725 -77.628 45.435
205278 100.748 -77.638 45.422
205279 100.772 -77.652 45.408
205280 100.799 -77.668 45.395
205281 100.826 -77.685 45.381
205282 100.852 -77.702 45.368
205283 100.879 -77.719 45.355
205284 100.905 -77.736 45.340
205285 100.932 -77.752 45.326
205286 100.959 -77.770 45.314
205287 100.985 -77.787 45.300
205288 101.012 -77.803 45.286
205289 101.039 -77.820 45.273
205290 101.065 -77.837 45.259
205291 101.092 -77.854 45.247
205292 101.119 -77.871 45.234
205293 101.145 -77.888 45.221
205294 101.172 -77.905 45.208
205295 101.199 -77.925 45.194
205296 101.226 -77.947 45.180
205297 101.248 -77.959 45.167
205298 101.270 -77.971 45.155
205299 101.292 -77.984 45.142
205300 101.313 -77.986 45.128
205301 101.318 -77.985 45.115
205302 101.323 -77.985 45.101
205303 101.329 -77.984 45.087
205304 101.334 -77.984 45.072
205305 101.339 -77.984 45.058
205306 101.375 -78.002 45.042
205307 101.414 -78.022 45.027
205308 101.453 -78.043 45.011
205309 101.494 -78.065 44.995
205310 101.538 -78.088 44.980
205311 101.582 -78.112 44.965
205312 101.611 -78.128 44.950
205313 101.626 -78.136 44.934
205314 101.646 -78.147 44.919
205315 101.667 -78.159 44.903
205316 101.687 -78.171 44.887
205317 101.707 -78.182 44.871
205318 101.728 -78.194 44.855
205319 101.748 -78.205 44.840
205320 101.768 -78.216 44.824
205321 101.789 -78.228 44.807
205322 101.809 -78.239 44.790
205323 101.829 -78.251 44.773
205324 101.850 -78.263 44.756
205325 101.870 -78.274 44.740
205326 101.891 -78.285 44.724
205327 101.911 -78.297 44.708
205328 101.931 -78.308 44.691
205329 101.965 -78.319 44.677
205330 101.999 -78.329 44.662
205331 102.034 -78.340 44.648
205332 102.069 -78.351 44.634
205333 102.104 -78.362 44.620
205334 102.146 -78.383 44.605
205335 102.186 -78.405 44.591
205336 102.228 -78.426 44.575
205337 102.268 -78.448 44.560
205338 102.310 -78.469 44.545
205339 102.350 -78.490 44.530
205340 102.392 -78.511 44.515
205341 102.433 -78.532 44.500
205342 102.474 -78.554 44.485
205343 102.507 -78.576 44.470
205344 102.538 -78.598 44.456
205345 102.570 -78.620 44.441
205346 102.599 -78.623 44.426
205347 102.624 -78.626 44.411
205348 102.640 -78.629 44.396
205349 102.656 -78.631 44.381
205350 102.671 -78.633 44.365
205351 102.686 -78.636 44.351
205352 102.702 -78.638 44.336
205353 102.726 -78.640 44.320
205354 102.751 -78.642 44.304
205355 102.777 -78.645 44.289
205356 102.803 -78.648 44.272
205357 102.828 -78.650 44.256
205358 102.854 -78.653 44.240
205644 93.502 -75.950 54.006
205645 93.520 -75.953 54.023
205646 93.539 -75.956 54.040
205647 93.557 -75.960 54.056
205648 93.575 -75.963 54.073
205649 93.593 -75.966 54.091
205650 93.612 -75.969 54.109
205651 93.630 -75.972 54.127
205652 93.635 -76.005 54.142
205653 93.659 -76.010 54.155
205654 93.691 -76.003 54.168
205655 93.723 -75.997 54.180
205656 93.748 -75.991 54.193
205657 93.769 -75.983 54.206
205658 93.790 -75.975 54.222
205659 93.811 -75.968 54.236
205660 93.832 -75.960 54.252
205661 93.853 -75.953 54.271
205662 93.876 -75.946 54.293
205663 93.908 -75.941 54.313
205664 93.939 -75.937 54.331
205665 93.971 -75.932 54.345
205666 93.998 -75.927 54.351
205667 94.025 -75.922 54.356
205668 94.052 -75.918 54.358
205669 94.079 -75.914 54.356
205670 94.107 -75.910 54.351
205671 94.135 -75.907 54.344
205672 94.163 -75.903 54.336
205673 94.190 -75.900 54.323
205674 94.218 -75.897 54.309
205675 94.246 -75.894 54.294
205676 94.270 -75.891 54.278
205677 94.287 -75.890 54.259
205678 94.305 -75.888 54.237
205679 94.323 -75.887 54.211
205680 94.341 -75.886 54.179
205681 94.359 -75.887 54.146
205682 94.378 -75.890 54.115
205683 94.397 -75.892 54.085
205684 94.415 -75.895 54.054
205685 94.434 -75.897 54.025
205686 94.453 -75.900 53.991
205687 94.469 -75.906 53.953
205688 94.483 -75.916 53.915
205689 94.498 -75.925 53.878
205690 94.512 -75.934 53.839
205691 94.526 -75.944 53.799
205692 94.541 -75.953 53.761
205693 94.555 -75.963 53.722
205694 94.569 -75.972 53.682
205695 94.583 -75.982 53.643
205696 94.598 -75.991 53.608
205697 94.612 -76.000 53.575
205698 94.627 -76.010 53.543
205699 94.642 -76.018 53.512
205700 94.657 -76.026 53.479
205701 94.669 -76.032 53.447
205702 94.678 -76.035 53.417
205703 94.687 -76.038 53.386
205704 94.696 -76.041 53.353
205705 94.706 -76.044 53.320
205706 94.717 -76.047 53.287
205707 94.728 -76.050 53.251
205708 94.739 -76.054 53.218
205709 94.750 -76.058 53.186
205710 94.761 -76.061 53.153
205711 94.772 -76.065 53.123
205712 94.783 -76.069 53.092
205713 94.794 -76.072 53.061
205714 94.806 -76.075 53.027
205715 94.817 -76.079 52.992
205716 94.828 -76.083 52.956
205717 94.839 -76.086 52.919
205718 94.850 -76.090 52.883
205719 94.861 -76.094 52.848
205720 94.872 -76.097 52.813
205721 94.883 -76.100 52.778
205722 94.894 -76.104 52.744
205723 94.905 -76.108 52.711
205724 94.916 -76.111 52.679
205725 94.927 -76.115 52.647
205726 94.938 -76.119 52.616
205727 94.983 -76.119 52.584
205728 95.046 -76.119 52.551
205729 95.109 -76.118 52.518
205730 95.172 -76.117 52.485
205731 95.224 -76.119 52.452
205732 95.231 -76.131 52.420
205733 95.246 -76.141 52.390
205734 95.275 -76.149 52.361
205735 95.304 -76.157 52.328
205736 95.332 -76.165 52.290
205737 95.360 -76.173 52.245
205738 95.389 -76.181 52.193
205739 95.418 -76.188 52.143
205740 95.447 -76.197 52.094
205741 95.473 -76.198 52.046
205742 95.500 -76.196 51.998
205743 95.528 -76.193 51.950
205744 95.554 -76.190 51.902
205745 95.582 -76.187 51.858
205746 95.609 -76.184 51.815
205747 95.636 -76.182 51.773
205748 95.663 -76.179 51.734
205749 95.690 -76.176 51.694
205750 95.717 -76.173 51.653
205751 95.740 -76.177 51.613
205752 95.763 -76.182 51.573
205753 95.786 -76.186 51.535
205754 95.809 -76.191 51.500
205755 95.832 -76.195 51.465
205756 95.854 -76.199 51.426
205757 95.878 -76.204 51.387
205758 95.901 -76.208 51.347
205759 95.923 -76.212 51.305
205760 95.947 -76.217 51.259
205761 95.968 -76.220 51.214
205762 95.989 -76.221 51.167
205763 96.012 -76.223 51.120
205764 96.033 -76.225 51.075
205765 96.055 -76.226 51.029
205766 96.078 -76.229 50.986
205767 96.099 -76.230 50.945
205768 96.121 -76.232 50.909
205769 96.143 -76.233 50.874
205770 96.165 -76.235 50.839
205771 96.187 -76.237 50.806
205772 96.196 -76.246 50.772
205773 96.206 -76.254 50.738
205774 96.220 -76.257 50.705
205775 96.232 -76.260 50.675
205776 96.245 -76.263 50.645
205777 96.257 -76.268 50.615
205778 96.269 -76.275 50.585
205779 96.281 -76.282 50.555
205780 96.293 -76.289 50.527
205781 96.305 -76.295 50.504
205782 96.321 -76.305 50.484
205783 96.345 -76.319 50.465
205784 96.370 -76.333 50.447
205785 96.394 -76.348 50.428
205786 96.418 -76.361 50.408
205787 96.455 -76.374 50.383
205788 96.500 -76.385 50.356
205789 96.537 -76.377 50.327
205790 96.562 -76.383 50.297
205791 96.586 -76.392 50.267
205792 96.609 -76.400 50.237
205793 96.639 -76.409 50.206
205794 96.674 -76.418 50.175
205795 96.710 -76.427 50.142
205796 96.745 -76.436 50.107
205797 96.781 -76.445 50.071
205798 96.817 -76.452 50.035
205799 96.857 -76.454 49.997
205800 96.863 -76.446 49.958
205801 96.869 -76.438 49.920
205802 96.875 -76.430 49.883
205803 96.881 -76.421 49.846
205804 96.887 -76.413 49.808
205805 96.893 -76.404 49.769
205806 96.865 -76.390 49.729
205807 96.844 -76.383 49.689
205808 96.843 -76.392 49.649
205809 96.939 -76.430 49.612
205810 97.013 -76.464 49.576
205811 97.029 -76.492 49.543
205812 97.044 -76.521 49.511
205813 97.060 -76.549 49.481
205814 97.075 -76.577 49.455
205815 97.090 -76.606 49.431
205816 97.105 -76.634 49.409
205817 97.121 -76.663 49.389
205818 97.136 -76.691 49.369
205819 97.151 -76.719 49.348
205820 97.167 -76.748 49.327
205821 97.193 -76.759 49.305
205822 97.232 -76.752 49.280
205823 97.270 -76.746 49.255
205824 97.308 -76.740 49.228
205825 97.347 -76.733 49.197
205826 97.403 -76.683 49.164
205827 97.414 -76.704 49.131
205828 97.432 -76.723 49.099
205829 97.450 -76.741 49.065
205830 97.461 -76.754 49.031
205831 97.469 -76.764 48.999
205832 97.478 -76.774 48.968
205833 97.487 -76.784 48.939
205834 97.502 -76.795 48.910
205835 97.525 -76.806 48.882
205836 97.548 -76.818 48.854
205837 97.571 -76.830 48.827
205838 97.594 -76.841 48.799
205839 97.618 -76.853 48.771
205840 97.641 -76.865 48.741
205841 97.664 -76.877 48.712
205842 97.696 -76.887 48.682
205843 97.734 -76.898 48.651
205844 97.775 -76.906 48.620
205845 97.815 -76.914 48.589
205846 97.855 -76.922 48.559
205847 97.895 -76.930 48.528
205848 97.936 -76.938 48.498
205849 97.974 -76.946 48.468
205850 97.998 -76.952 48.437
205851 98.021 -76.957 48.406
205852 98.044 -76.963 48.375
205853 98.069 -76.952 48.344
205854 98.093 -76.935 48.315
205855 98.113 -76.928 48.286
205856 98.131 -76.921 48.255
205857 98.150 -76.915 48.225
205858 98.168 -76.907 48.195
205859 98.186 -76.900 48.165
205860 98.205 -76.891 48.134
205861 98.224 -76.884 48.103
205862 98.242 -76.875 48.071
205863 98.261 -76.868 48.037
205864 98.279 -76.859 48.004
205865 98.298 -76.852 47.972
205866 98.316 -76.844 47.940
205867 98.336 -76.848 47.908
205868 98.356 -76.852 47.876
205869 98.375 -76.861 47.845
205870 98.394 -76.878 47.815
205871 98.413 -76.893 47.786
205872 98.431 -76.909 47.756
205873 98.449 -76.926 47.728
205874 98.468 -76.942 47.701
205875 98.487 -76.958 47.675
205876 98.505 -76.973 47.648
205877 98.529 -76.967 47.622
205878 98.553 -76.962 47.596
205879 98.577 -76.956 47.570
205880 98.601 -76.950 47.546
205881 98.626 -76.945 47.523
205882 98.650 -76.939 47.500
205883 98.673 -76.934 47.476
205884 98.669 -76.950 47.452
205885 98.662 -76.967 47.425
205886 98.667 -76.980 47.397
205887 98.700 -76.982 47.370
205888 98.733 -76.984 47.341
205889 98.766 -76.987 47.310
205890 98.799 -76.988 47.283
205891 98.832 -76.990 47.253
205892 98.865 -76.993 47.221
205893 98.910 -76.998 47.188
205894 98.947 -77.008 47.156
205895 98.961 -77.022 47.123
205896 98.975 -77.035 47.090
205897 98.989 -77.049 47.059
205898 99.003 -77.063 47.028
205899 99.016 -77.077 46.997
205900 99.030 -77.091 46.968
205901 99.045 -77.104 46.940
205902 99.058 -77.118 46.913
205903 99.072 -77.132 46.887
205904 99.086 -77.145 46.862
205905 99.100 -77.159 46.835
205906 99.114 -77.172 46.809
205907 99.128 -77.185 46.782
205908 99.141 -77.192 46.753
205909 99.154 -77.197 46.723
205910 99.174 -77.196 46.696
205911 99.201 -77.190 46.668
205912 99.227 -77.184 46.640
205913 99.253 -77.179 46.612
205914 99.279 -77.172 46.583
205915 99.304 -77.180 46.555
205916 99.329 -77.195 46.526
205917 99.353 -77.210 46.500
205918 99.362 -77.224 46.474
205919 99.368 -77.238 46.448
205920 99.370 -77.248 46.422
205921 99.356 -77.241 46.397
205922 99.342 -77.235 46.372
205923 99.329 -77.229 46.346
205924 99.341 -77.229 46.322
205925 99.370 -77.232 46.299
205926 99.400 -77.235 46.276
205927 99.429 -77.239 46.253
205928 99.459 -77.242 46.230
205929 99.488 -77.246 46.207
205930 99.534 -77.252 46.185
205931 99.584 -77.259 46.164
205932 99.636 -77.266 46.142
205933 99.687 -77.273 46.121
205934 99.730 -77.276 46.101
205935 99.755 -77.269 46.081
205936 99.780 -77.263 46.061
205937 99.816 -77.262 46.042
205938 99.852 -77.263 46.023
205939 99.883 -77.285 46.005
205940 99.909 -77.300 45.988
205941 99.932 -77.310 45.969
205942 99.944 -77.304 45.950
205943 99.957 -77.295 45.933
205944 99.969 -77.288 45.916
205945 99.981 -77.279 45.899
205946 99.993 -77.271 45.882
205947 100.010 -77.269 45.865
205948 100.044 -77.282 45.849
205949 100.078 -77.296 45.832
205950 100.112 -77.310 45.815
205951 100.146 -77.323 45.800
205952 100.180 -77.337 45.784
205953 100.214 -77.351 45.768
205954 100.244 -77.363 45.753
205955 100.255 -77.366 45.739
205956 100.266 -77.372 45.724
205957 100.277 -77.382 45.710
205958 100.293 -77.391 45.697
205959 100.312 -77.398 45.685
205960 100.331 -77.405 45.671
205961 100.350 -77.413 45.658
205962 100.369 -77.420 45.646
205963 100.389 -77.427 45.633
205964 100.413 -77.442 45.621
205965 100.437 -77.457 45.609
205966 100.461 -77.473 45.595
205967 100.485 -77.488 45.583
205968 100.508 -77.506 45.570
205969 100.531 -77.526 45.556
205970 100.559 -77.537 45.542
205971 100.587 -77.544 45.528
205972 100.610 -77.554 45.514
205973 100.633 -77.564 45.500
205974 100.655 -77.574 45.486
205975 100.678 -77.585 45.472
205976 100.701 -77.595 45.458
205977 100.723 -77.605 45.444
205978 100.746 -77.615 45.430
205979 100.769 -77.625 45.416
205980 100.794 -77.639 45.402
205981 100.821 -77.657 45.389
205982 100.847 -77.674 45.376
205983 100.874 -77.690 45.363
205984 100.901 -77.707 45.350
205985 100.927 -77.724 45.336
205986 100.954 -77.741 45.322
205987 100.981 -77.758 45.309
205988 101.007 -77.774 45.297
205989 101.034 -77.792 45.282
205990 101.060 -77.808 45.269
205991 101.087 -77.825 45.255
205992 101.114 -77.842 45.242
205993 101.140 -77.859 45.229
205994 101.167 -77.876 45.215
205995 101.194 -77.893 45.202
205996 101.220 -77.910 45.188
205997 101.247 -77.927 45.175
205998 101.274 -77.945 45.162
205999 101.299 -77.963 45.149
206000 101.321 -77.975 45.136
206001 101.339 -77.980 45.122
206002 101.344 -77.979 45.109
206003 101.350 -77.979 45.096
206004 101.355 -77.978 45.081
206005 101.360 -77.978 45.067
206006 101.367 -77.978 45.052
206007 101.393 -77.990 45.036
206008 101.437 -78.014 45.020
206009 101.480 -78.037 45.004
206010 101.524 -78.061 44.988
206011 101.568 -78.084 44.973
206012 101.611 -78.107 44.958
206013 101.629 -78.118 44.942
206014 101.640 -78.124 44.927
206015 101.658 -78.134 44.912
206016 101.679 -78.146 44.896
206017 101.699 -78.158 44.880
206018 101.720 -78.169 44.864
206019 101.740 -78.181 44.848
206020 101.760 -78.192 44.833
206021 101.780 -78.203 44.817
206022 101.801 -78.215 44.799
206023 101.821 -78.226 44.784
206024 101.841 -78.238 44.768
206025 101.862 -78.250 44.751
206026 101.882 -78.261 44.734
206027 101.902 -78.272 44.718
206028 101.923 -78.284 44.702
206029 101.954 -78.294 44.686
206030 101.988 -78.305 44.671
206031 102.024 -78.316 44.657
206032 102.058 -78.326 44.643
206033 102.093 -78.337 44.629
206034 102.129 -78.350 44.614
206035 102.170 -78.371 44.599
206036 102.211 -78.393 44.584
206037 102.252 -78.414 44.569
206038 102.293 -78.435 44.553
206039 102.334 -78.457 44.538
206040 102.376 -78.478 44.523
206041 102.416 -78.499 44.509
206042 102.458 -78.520 44.493
206043 102.499 -78.542 44.478
206044 102.540 -78.563 44.464
206045 102.575 -78.585 44.449
206046 102.606 -78.607 44.435
206047 102.608 -78.607 44.420
206048 102.639 -78.611 44.405
206049 102.655 -78.613 44.390
206050 102.670 -78.615 44.375
206051 102.685 -78.617 44.359
206052 102.711 -78.620 44.345
206053 102.736 -78.623 44.329
206054 102.762 -78.625 44.313
206055 102.787 -78.627 44.298
206056 102.813 -78.630 44.282
206057 102.839 -78.633 44.265
206058 102.865 -78.636 44.249
206059 102.890 -78.638 44.233
206344 93.476 -75.923 53.982
206345 93.494 -75.926 54.001
206346 93.512 -75.929 54.020
206347 93.531 -75.933 54.037
206348 93.549 -75.936 54.055
206349 93.567 -75.939 54.073
206350 93.585 -75.943 54.092
206351 93.604 -75.946 54.110
206352 93.622 -75.949 54.128
206353 93.627 -75.981 54.144
206354 93.654 -75.982 54.158
206355 93.686 -75.976 54.171
206356 93.718 -75.970 54.183
206357 93.751 -75.964 54.196
206358 93.772 -75.956 54.209
206359 93.793 -75.949 54.223
206360 93.815 -75.941 54.235
206361 93.835 -75.934 54.249
206362 93.856 -75.926 54.266
206363 93.877 -75.919 54.288
206364 93.909 -75.914 54.307
206365 93.940 -75.909 54.325
206366 93.972 -75.905 54.341
206367 93.999 -75.900 54.348
206368 94.025 -75.894 54.351
206369 94.052 -75.890 54.352
206370 94.079 -75.886 54.350
206371 94.106 -75.881 54.344
206372 94.133 -75.877 54.336
206373 94.160 -75.872 54.326
206374 94.188 -75.869 54.312
206375 94.216 -75.865 54.298
206376 94.244 -75.862 54.284
206377 94.271 -75.859 54.270
206378 94.290 -75.857 54.254
206379 94.308 -75.856 54.235
206380 94.326 -75.856 54.211
206381 94.344 -75.858 54.181
206382 94.363 -75.861 54.149
206383 94.382 -75.863 54.120
206384 94.400 -75.865 54.091
206385 94.419 -75.869 54.064
206386 94.438 -75.871 54.036
206387 94.457 -75.874 54.003
206388 94.473 -75.880 53.966
206389 94.487 -75.890 53.929
206390 94.501 -75.899 53.891
206391 94.516 -75.909 53.853
206392 94.530 -75.918 53.815
206393 94.544 -75.928 53.777
206394 94.558 -75.938 53.738
206395 94.572 -75.947 53.698
206396 94.587 -75.956 53.659
206397 94.602 -75.964 53.624
206398 94.617 -75.973 53.589
206399 94.632 -75.981 53.556
206400 94.646 -75.990 53.523
206401 94.661 -75.999 53.490
206402 94.673 -76.004 53.458
206403 94.683 -76.007 53.426
206404 94.692 -76.010 53.393
206405 94.701 -76.013 53.359
206406 94.710 -76.016 53.325
206407 94.720 -76.019 53.290
206408 94.729 -76.022 53.253
206409 94.739 -76.025 53.218
206410 94.750 -76.028 53.184
206411 94.761 -76.032 53.150
206412 94.772 -76.035 53.120
206413 94.783 -76.039 53.090
206414 94.794 -76.042 53.059
206415 94.806 -76.046 53.025
206416 94.817 -76.050 52.991
206417 94.828 -76.053 52.953
206418 94.839 -76.057 52.915
206419 94.850 -76.060 52.877
206420 94.861 -76.064 52.841
206421 94.872 -76.067 52.807
206422 94.883 -76.071 52.774
206423 94.894 -76.075 52.743
206424 94.905 -76.079 52.713
206425 94.916 -76.082 52.685
206426 94.927 -76.085 52.657
206427 94.938 -76.089 52.629
206428 94.982 -76.090 52.598
206429 95.045 -76.089 52.567
206430 95.108 -76.088 52.535
206431 95.170 -76.088 52.503
206432 95.233 -76.087 52.470
206433 95.286 -76.088 52.437
206434 95.293 -76.100 52.406
206435 95.316 -76.109 52.378
206436 95.344 -76.117 52.345
206437 95.373 -76.125 52.307
206438 95.401 -76.133 52.264
206439 95.430 -76.141 52.211
206440 95.459 -76.149 52.159
206441 95.486 -76.153 52.109
206442 95.511 -76.151 52.061
206443 95.536 -76.150 52.014
206444 95.561 -76.148 51.964
206445 95.587 -76.147 51.914
206446 95.612 -76.145 51.868
206447 95.637 -76.144 51.823
206448 95.662 -76.142 51.781
206449 95.687 -76.141 51.741
206450 95.705 -76.143 51.703
206451 95.718 -76.150 51.663
206452 95.740 -76.154 51.623
206453 95.762 -76.158 51.582
206454 95.784 -76.163 51.543
206455 95.807 -76.167 51.507
206456 95.829 -76.172 51.472
206457 95.851 -76.176 51.433
206458 95.874 -76.180 51.394
206459 95.896 -76.185 51.354
206460 95.918 -76.189 51.312
206461 95.951 -76.191 51.268
206462 95.976 -76.191 51.223
206463 96.000 -76.191 51.178
206464 96.021 -76.192 51.132
206465 96.043 -76.195 51.088
206466 96.065 -76.196 51.042
206467 96.087 -76.198 50.999
206468 96.109 -76.200 50.959
206469 96.131 -76.201 50.922
206470 96.153 -76.203 50.886
206471 96.175 -76.205 50.851
206472 96.197 -76.207 50.817
206473 96.206 -76.216 50.784
206474 96.215 -76.226 50.751
206475 96.224 -76.234 50.718
206476 96.237 -76.237 50.689
206477 96.249 -76.242 50.660
206478 96.261 -76.248 50.630
206479 96.273 -76.255 50.601
206480 96.285 -76.262 50.571
206481 96.297 -76.269 50.544
206482 96.309 -76.276 50.522
206483 96.321 -76.282 50.503
206484 96.341 -76.294 50.484
206485 96.365 -76.308 50.466
206486 96.389 -76.322 50.447
206487 96.413 -76.336 50.426
206488 96.452 -76.348 50.400
206489 96.497 -76.360 50.372
206490 96.542 -76.371 50.343
206491 96.583 -76.371 50.313
206492 96.619 -76.360 50.283
206493 96.647 -76.361 50.253
206494 96.681 -76.370 50.223
206495 96.715 -76.380 50.191
206496 96.740 -76.389 50.159
206497 96.765 -76.399 50.124
206498 96.790 -76.408 50.088
206499 96.815 -76.418 50.052
206500 96.840 -76.427 50.015
206501 96.846 -76.420 49.977
206502 96.852 -76.411 49.939
206503 96.858 -76.404 49.902
206504 96.864 -76.395 49.864
206505 96.870 -76.388 49.826
206506 96.858 -76.379 49.788
206507 96.839 -76.371 49.749
206508 96.820 -76.364 49.709
206509 96.873 -76.383 49.669
206510 96.970 -76.421 49.631
206511 97.012 -76.452 49.595
206512 97.027 -76.481 49.561
206513 97.043 -76.509 49.529
206514 97.058 -76.537 49.498
206515 97.074 -76.566 49.472
206516 97.089 -76.594 49.447
206517 97.104 -76.623 49.425
206518 97.119 -76.651 49.404
206519 97.135 -76.679 49.383
206520 97.150 -76.708 49.363
206521 97.165 -76.736 49.341
206522 97.181 -76.765 49.319
206523 97.214 -76.765 49.294
206524 97.252 -76.759 49.267
206525 97.291 -76.752 49.240
206526 97.350 -76.697 49.208
206527 97.403 -76.652 49.175
206528 97.418 -76.668 49.141
206529 97.424 -76.696 49.107
206530 97.442 -76.714 49.073
206531 97.454 -76.727 49.039
206532 97.463 -76.737 49.007
206533 97.471 -76.747 48.976
206534 97.493 -76.759 48.948
206535 97.516 -76.771 48.920
206536 97.539 -76.782 48.893
206537 97.563 -76.794 48.865
206538 97.586 -76.806 48.838
206539 97.609 -76.818 48.810
206540 97.632 -76.829 48.782
206541 97.655 -76.841 48.752
206542 97.678 -76.853 48.723
206543 97.702 -76.865 48.693
206544 97.735 -76.875 48.662
206545 97.774 -76.884 48.630
206546 97.814 -76.893 48.599
206547 97.855 -76.900 48.567
206548 97.895 -76.909 48.536
206549 97.935 -76.916 48.506
206550 97.975 -76.925 48.475
206551 98.016 -76.933 48.444
206552 98.044 -76.939 48.412
206553 98.068 -76.944 48.383
206554 98.092 -76.935 48.352
206555 98.113 -76.924 48.322
206556 98.132 -76.918 48.293
206557 98.150 -76.911 48.263
206558 98.168 -76.904 48.233
206559 98.187 -76.897 48.203
206560 98.206 -76.890 48.173
206561 98.224 -76.881 48.143
206562 98.242 -76.874 48.112
206563 98.261 -76.865 48.080
206564 98.280 -76.858 48.048
206565 98.298 -76.850 48.015
206566 98.316 -76.842 47.984
206567 98.335 -76.834 47.952
206568 98.354 -76.826 47.921
206569 98.372 -76.819 47.889
206570 98.391 -76.830 47.858
206571 98.410 -76.846 47.827
206572 98.429 -76.862 47.797
206573 98.447 -76.878 47.769
206574 98.465 -76.894 47.742
206575 98.484 -76.910 47.715
206576 98.502 -76.927 47.690
206577 98.521 -76.943 47.664
206578 98.539 -76.959 47.638
206579 98.563 -76.955 47.610
206580 98.587 -76.949 47.583
206581 98.612 -76.943 47.557
206582 98.635 -76.937 47.532
206583 98.659 -76.932 47.507
206584 98.684 -76.926 47.483
206585 98.684 -76.939 47.459
206586 98.677 -76.956 47.433
206587 98.690 -76.966 47.405
206588 98.723 -76.969 47.375
206589 98.756 -76.971 47.345
206590 98.789 -76.972 47.315
206591 98.821 -76.975 47.286
206592 98.855 -76.977 47.256
206593 98.887 -76.979 47.224
206594 98.924 -76.982 47.191
206595 98.959 -76.992 47.157
206596 98.973 -77.006 47.124
206597 98.987 -77.019 47.090
206598 99.000 -77.033 47.058
206599 99.014 -77.047 47.027
206600 99.028 -77.060 46.996
206601 99.042 -77.074 46.966
206602 99.056 -77.088 46.937
206603 99.070 -77.102 46.908
206604 99.084 -77.116 46.881
206605 99.097 -77.129 46.857
206606 99.114 -77.143 46.829
206607 99.131 -77.157 46.803
206608 99.154 -77.160 46.776
206609 99.183 -77.152 46.747
206610 99.208 -77.146 46.718
206611 99.234 -77.140 46.691
206612 99.261 -77.135 46.664
206613 99.287 -77.128 46.636
206614 99.311 -77.141 46.608
206615 99.336 -77.156 46.580
206616 99.360 -77.171 46.551
206617 99.381 -77.185 46.523
206618 99.387 -77.199 46.496
206619 99.394 -77.213 46.470
206620 99.400 -77.227 46.443
206621 99.390 -77.225 46.418
206622 99.376 -77.219 46.392
206623 99.362 -77.213 46.367
206624 99.348 -77.207 46.341
206625 99.339 -77.201 46.318
206626 99.368 -77.205 46.295
206627 99.398 -77.208 46.272
206628 99.427 -77.212 46.249
206629 99.467 -77.216 46.225
206630 99.518 -77.224 46.203
206631 99.569 -77.231 46.181
206632 99.620 -77.238 46.159
206633 99.671 -77.244 46.138
206634 99.721 -77.251 46.118
206635 99.746 -77.244 46.097
206636 99.771 -77.238 46.077
206637 99.797 -77.232 46.058
206638 99.834 -77.232 46.038
206639 99.870 -77.232 46.020
206640 99.903 -77.251 46.001
206641 99.933 -77.272 45.984
206642 99.956 -77.282 45.966
206643 99.978 -77.291 45.947
206644 99.990 -77.284 45.930
206645 100.002 -77.276 45.912
206646 100.014 -77.268 45.895
206647 100.027 -77.260 45.878
206648 100.038 -77.251 45.860
206649 100.072 -77.265 45.843
206650 100.106 -77.279 45.826
206651 100.140 -77.292 45.809
206652 100.174 -77.306 45.793
206653 100.208 -77.320 45.778
206654 100.242 -77.334 45.762
206655 100.259 -77.340 45.747
206656 100.271 -77.343 45.732
206657 100.282 -77.348 45.718
206658 100.292 -77.360 45.704
206659 100.303 -77.370 45.691
206660 100.323 -77.377 45.678
206661 100.342 -77.384 45.665
206662 100.361 -77.392 45.652
206663 100.380 -77.398 45.640
206664 100.402 -77.411 45.627
206665 100.426 -77.426 45.615
206666 100.450 -77.442 45.602
206667 100.474 -77.457 45.589
206668 100.497 -77.477 45.576
206669 100.521 -77.497 45.563
206670 100.544 -77.517 45.550
206671 100.570 -77.532 45.536
206672 100.601 -77.536 45.522
206673 100.631 -77.542 45.507
206674 100.654 -77.551 45.494
206675 100.677 -77.562 45.479
206676 100.699 -77.572 45.465
206677 100.722 -77.582 45.451
206678 100.745 -77.592 45.437
206679 100.767 -77.602 45.424
206680 100.790 -77.613 45.410
206681 100.816 -77.628 45.397
206682 100.842 -77.645 45.385
206683 100.869 -77.661 45.372
206684 100.896 -77.679 45.359
206685 100.922 -77.696 45.346
206686 100.949 -77.712 45.332
206687 100.976 -77.729 45.318
206688 101.002 -77.746 45.305
206689 101.029 -77.763 45.291
206690 101.056 -77.780 45.278
206691 101.082 -77.796 45.264
206692 101.109 -77.814 45.250
206693 101.136 -77.830 45.236
206694 101.162 -77.847 45.223
206695 101.189 -77.865 45.210
206696 101.216 -77.881 45.196
206697 101.243 -77.898 45.182
206698 101.269 -77.915 45.169
206699 101.296 -77.932 45.156
206700 101.322 -77.949 45.143
206701 101.349 -77.965 45.130
206702 101.365 -77.974 45.116
206703 101.373 -77.975 45.103
206704 101.383 -77.977 45.090
206705 101.392 -77.979 45.076
206706 101.402 -77.981 45.061
206707 101.411 -77.984 45.045
206708 101.422 -77.987 45.030
206709 101.466 -78.010 45.014
206710 101.510 -78.034 44.998
206711 101.553 -78.056 44.982
206712 101.597 -78.080 44.966
206713 101.636 -78.101 44.950
206714 101.647 -78.108 44.934
206715 101.658 -78.114 44.920
206716 101.670 -78.122 44.904
206717 101.691 -78.133 44.888
206718 101.711 -78.144 44.872
206719 101.732 -78.156 44.857
206720 101.752 -78.168 44.841
206721 101.772 -78.179 44.825
206722 101.793 -78.191 44.809
206723 101.813 -78.202 44.793
206724 101.833 -78.213 44.777
206725 101.853 -78.225 44.761
206726 101.874 -78.236 44.745
206727 101.894 -78.248 44.729
206728 101.914 -78.260 44.712
206729 101.943 -78.270 44.697
206730 101.978 -78.281 44.681
206731 102.013 -78.291 44.666
206732 102.047 -78.302 44.651
206733 102.082 -78.313 44.637
206734 102.117 -78.323 44.624
206735 102.155 -78.338 44.609
206736 102.195 -78.360 44.594
206737 102.237 -78.381 44.578
206738 102.277 -78.402 44.563
206739 102.319 -78.423 44.547
206740 102.360 -78.445 44.532
206741 102.401 -78.466 44.518
206742 102.442 -78.487 44.502
206743 102.483 -78.508 44.487
206744 102.524 -78.529 44.472
206745 102.564 -78.551 44.458
206746 102.606 -78.572 44.443
206747 102.643 -78.594 44.429
206748 102.637 -78.595 44.414
206749 102.654 -78.595 44.399
206750 102.670 -78.597 44.384
206751 102.696 -78.600 44.368
206752 102.722 -78.602 44.352
206753 102.747 -78.605 44.337
206754 102.773 -78.608 44.322
206755 102.799 -78.610 44.306
206756 102.824 -78.613 44.290
206757 102.850 -78.615 44.274
206758 102.876 -78.617 44.258
206759 102.901 -78.620 44.242
206760 102.927 -78.623 44.225
207045 93.468 -75.900 53.978
207046 93.486 -75.903 53.997
207047 93.504 -75.906 54.017
207048 93.522 -75.909 54.035
207049 93.541 -75.912 54.054
207050 93.559 -75.916 54.073
207051 93.577 -75.919 54.093
207052 93.595 -75.922 54.111
207053 93.614 -75.926 54.130
207054 93.619 -75.956 54.147
207055 93.648 -75.955 54.161
207056 93.681 -75.949 54.176
207057 93.713 -75.942 54.189
207058 93.746 -75.936 54.202
207059 93.776 -75.930 54.214
207060 93.797 -75.922 54.227
207061 93.818 -75.915 54.237
207062 93.839 -75.907 54.248
207063 93.860 -75.900 54.261
207064 93.880 -75.892 54.278
207065 93.910 -75.887 54.296
207066 93.941 -75.882 54.314
207067 93.973 -75.878 54.331
207068 93.999 -75.872 54.339
207069 94.026 -75.867 54.341
207070 94.052 -75.862 54.340
207071 94.079 -75.858 54.336
207072 94.106 -75.853 54.329
207073 94.133 -75.849 54.321
207074 94.160 -75.844 54.312
207075 94.188 -75.840 54.296
207076 94.215 -75.836 54.282
207077 94.242 -75.831 54.269
207078 94.269 -75.828 54.259
207079 94.292 -75.825 54.246
207080 94.311 -75.826 54.231
207081 94.330 -75.829 54.210
207082 94.348 -75.831 54.182
207083 94.367 -75.834 54.151
207084 94.386 -75.837 54.122
207085 94.404 -75.839 54.096
207086 94.423 -75.842 54.072
207087 94.442 -75.844 54.045
207088 94.461 -75.847 54.014
207089 94.476 -75.854 53.978
207090 94.490 -75.864 53.941
207091 94.504 -75.873 53.902
207092 94.519 -75.883 53.864
207093 94.533 -75.893 53.827
207094 94.548 -75.902 53.792
207095 94.562 -75.910 53.754
207096 94.577 -75.919 53.715
207097 94.592 -75.928 53.677
207098 94.607 -75.937 53.642
207099 94.622 -75.945 53.607
207100 94.636 -75.953 53.572
207101 94.651 -75.963 53.538
207102 94.666 -75.971 53.504
207103 94.678 -75.977 53.469
207104 94.688 -75.980 53.435
207105 94.697 -75.982 53.401
207106 94.706 -75.985 53.365
207107 94.715 -75.988 53.330
207108 94.725 -75.991 53.294
207109 94.734 -75.994 53.257
207110 94.744 -75.997 53.221
207111 94.753 -76.000 53.185
207112 94.762 -76.003 53.150
207113 94.772 -76.006 53.118
207114 94.783 -76.010 53.086
207115 94.794 -76.013 53.055
207116 94.806 -76.016 53.021
207117 94.817 -76.020 52.986
207118 94.828 -76.023 52.949
207119 94.839 -76.027 52.912
207120 94.850 -76.031 52.874
207121 94.861 -76.035 52.839
207122 94.872 -76.038 52.806
207123 94.883 -76.041 52.774
207124 94.894 -76.045 52.747
207125 94.905 -76.048 52.721
207126 94.916 -76.052 52.696
207127 94.927 -76.056 52.670
207128 94.938 -76.060 52.641
207129 94.980 -76.060 52.612
207130 95.043 -76.060 52.581
207131 95.106 -76.059 52.550
207132 95.169 -76.058 52.518
207133 95.232 -76.057 52.485
207134 95.295 -76.057 52.451
207135 95.348 -76.057 52.418
207136 95.356 -76.069 52.389
207137 95.385 -76.077 52.356
207138 95.413 -76.085 52.319
207139 95.442 -76.093 52.276
207140 95.471 -76.101 52.225
207141 95.498 -76.107 52.172
207142 95.524 -76.106 52.119
207143 95.549 -76.104 52.069
207144 95.574 -76.103 52.021
207145 95.600 -76.101 51.972
207146 95.625 -76.100 51.923
207147 95.650 -76.098 51.876
207148 95.669 -76.100 51.830
207149 95.679 -76.107 51.786
207150 95.689 -76.113 51.746
207151 95.698 -76.119 51.707
207152 95.708 -76.126 51.669
207153 95.731 -76.131 51.631
207154 95.753 -76.135 51.592
207155 95.775 -76.140 51.552
207156 95.798 -76.144 51.516
207157 95.820 -76.148 51.480
207158 95.842 -76.153 51.441
207159 95.865 -76.157 51.401
207160 95.887 -76.162 51.361
207161 95.910 -76.166 51.318
207162 95.951 -76.166 51.275
207163 95.983 -76.166 51.232
207164 96.007 -76.165 51.188
207165 96.030 -76.164 51.144
207166 96.053 -76.163 51.101
207167 96.075 -76.166 51.056
207168 96.097 -76.167 51.012
207169 96.119 -76.169 50.970
207170 96.141 -76.171 50.933
207171 96.163 -76.173 50.896
207172 96.185 -76.174 50.861
207173 96.206 -76.176 50.828
207174 96.217 -76.185 50.795
207175 96.225 -76.195 50.763
207176 96.233 -76.205 50.732
207177 96.242 -76.214 50.703
207178 96.253 -76.221 50.674
207179 96.265 -76.228 50.645
207180 96.277 -76.235 50.615
207181 96.289 -76.242 50.587
207182 96.301 -76.249 50.560
207183 96.313 -76.256 50.538
207184 96.325 -76.263 50.519
207185 96.337 -76.270 50.501
207186 96.360 -76.283 50.482
207187 96.384 -76.297 50.463
207188 96.409 -76.311 50.442
207189 96.449 -76.323 50.415
207190 96.495 -76.334 50.387
207191 96.540 -76.345 50.358
207192 96.585 -76.357 50.329
207193 96.629 -76.364 50.300
207194 96.657 -76.355 50.270
207195 96.674 -76.348 50.240
207196 96.694 -76.345 50.208
207197 96.719 -76.354 50.175
207198 96.744 -76.364 50.141
207199 96.769 -76.373 50.107
207200 96.795 -76.383 50.071
207201 96.820 -76.392 50.034
207202 96.831 -76.389 49.996
207203 96.837 -76.381 49.959
207204 96.842 -76.373 49.922
207205 96.849 -76.365 49.883
207206 96.830 -76.357 49.845
207207 96.812 -76.348 49.806
207208 96.843 -76.359 49.767
207209 96.889 -76.374 49.728
207210 96.935 -76.389 49.688
207211 96.996 -76.412 49.650
207212 97.011 -76.441 49.615
207213 97.026 -76.469 49.580
207214 97.041 -76.497 49.547
207215 97.056 -76.526 49.515
207216 97.072 -76.554 49.488
207217 97.087 -76.583 49.463
207218 97.103 -76.611 49.440
207219 97.118 -76.639 49.418
207220 97.134 -76.668 49.397
207221 97.149 -76.696 49.376
207222 97.164 -76.724 49.355
207223 97.179 -76.752 49.331
207224 97.197 -76.777 49.306
207225 97.237 -76.766 49.279
207226 97.285 -76.730 49.249
207227 97.318 -76.715 49.217
207228 97.353 -76.699 49.184
207229 97.386 -76.684 49.149
207230 97.420 -76.669 49.113
207231 97.434 -76.686 49.078
207232 97.447 -76.701 49.046
207233 97.461 -76.712 49.015
207234 97.484 -76.723 48.985
207235 97.507 -76.735 48.957
207236 97.530 -76.746 48.930
207237 97.554 -76.759 48.903
207238 97.577 -76.770 48.876
207239 97.600 -76.782 48.849
207240 97.623 -76.793 48.821
207241 97.646 -76.806 48.792
207242 97.670 -76.817 48.764
207243 97.693 -76.829 48.734
207244 97.716 -76.840 48.703
207245 97.739 -76.852 48.672
207246 97.774 -76.862 48.639
207247 97.813 -76.871 48.608
207248 97.854 -76.879 48.576
207249 97.894 -76.887 48.544
207250 97.934 -76.895 48.514
207251 97.975 -76.903 48.483
207252 98.015 -76.911 48.452
207253 98.056 -76.919 48.420
207254 98.091 -76.926 48.391
207255 98.113 -76.921 48.360
207256 98.132 -76.914 48.329
207257 98.150 -76.908 48.299
207258 98.169 -76.901 48.270
207259 98.187 -76.894 48.240
207260 98.206 -76.887 48.210
207261 98.224 -76.880 48.180
207262 98.243 -76.871 48.150
207263 98.261 -76.864 48.121
207264 98.280 -76.856 48.091
207265 98.298 -76.848 48.059
207266 98.317 -76.840 48.027
207267 98.336 -76.832 47.995
207268 98.354 -76.824 47.964
207269 98.372 -76.816 47.932
207270 98.394 -76.816 47.902
207271 98.415 -76.812 47.871
207272 98.428 -76.818 47.841
207273 98.444 -76.831 47.811
207274 98.463 -76.847 47.783
207275 98.482 -76.863 47.756
207276 98.500 -76.879 47.729
207277 98.518 -76.895 47.704
207278 98.537 -76.912 47.679
207279 98.556 -76.928 47.652
207280 98.574 -76.943 47.624
207281 98.597 -76.942 47.597
207282 98.621 -76.936 47.569
207283 98.645 -76.931 47.542
207284 98.669 -76.925 47.515
207285 98.694 -76.919 47.490
207286 98.698 -76.928 47.465
207287 98.692 -76.946 47.439
207288 98.713 -76.953 47.411
207289 98.745 -76.954 47.381
207290 98.778 -76.956 47.350
207291 98.811 -76.959 47.320
207292 98.844 -76.961 47.289
207293 98.877 -76.963 47.258
207294 98.910 -76.965 47.226
207295 98.943 -76.967 47.194
207296 98.971 -76.975 47.158
207297 98.984 -76.989 47.124
207298 98.998 -77.003 47.090
207299 99.014 -77.016 47.058
207300 99.031 -77.030 47.026
207301 99.048 -77.044 46.995
207302 99.065 -77.057 46.964
207303 99.082 -77.071 46.933
207304 99.098 -77.085 46.903
207305 99.115 -77.098 46.876
207306 99.132 -77.112 46.850
207307 99.149 -77.125 46.823
207308 99.179 -77.116 46.796
207309 99.210 -77.107 46.768
207310 99.241 -77.097 46.740
207311 99.268 -77.091 46.712
207312 99.294 -77.087 46.685
207313 99.319 -77.101 46.658
207314 99.343 -77.116 46.630
207315 99.368 -77.131 46.603
207316 99.393 -77.146 46.576
207317 99.406 -77.160 46.547
207318 99.413 -77.175 46.518
207319 99.419 -77.188 46.491
207320 99.425 -77.203 46.465
207321 99.424 -77.209 46.439
207322 99.410 -77.203 46.413
207323 99.396 -77.197 46.387
207324 99.382 -77.190 46.361
207325 99.368 -77.184 46.335
207326 99.355 -77.178 46.312
207327 99.366 -77.178 46.290
207328 99.400 -77.182 46.267
207329 99.451 -77.188 46.243
207330 99.502 -77.195 46.221
207331 99.554 -77.203 46.198
207332 99.605 -77.210 46.176
207333 99.655 -77.216 46.154
207334 99.707 -77.223 46.133
207335 99.737 -77.219 46.113
207336 99.762 -77.213 46.093
207337 99.787 -77.207 46.073
207338 99.815 -77.202 46.053
207339 99.852 -77.202 46.034
207340 99.889 -77.201 46.015
207341 99.922 -77.216 45.997
207342 99.954 -77.238 45.979
207343 99.979 -77.254 45.961
207344 100.002 -77.263 45.943
207345 100.024 -77.272 45.925
207346 100.036 -77.264 45.908
207347 100.048 -77.256 45.890
207348 100.060 -77.248 45.873
207349 100.072 -77.240 45.855
207350 100.100 -77.248 45.837
207351 100.134 -77.262 45.820
207352 100.168 -77.276 45.803
207353 100.202 -77.289 45.787
207354 100.236 -77.303 45.771
207355 100.263 -77.313 45.755
207356 100.274 -77.316 45.740
207357 100.286 -77.320 45.726
207358 100.297 -77.326 45.712
207359 100.308 -77.336 45.697
207360 100.318 -77.347 45.684
207361 100.333 -77.356 45.670
207362 100.352 -77.363 45.658
207363 100.371 -77.370 45.645
207364 100.392 -77.379 45.633
207365 100.416 -77.395 45.620
207366 100.439 -77.410 45.608
207367 100.463 -77.429 45.595
207368 100.486 -77.448 45.583
207369 100.510 -77.468 45.569
207370 100.533 -77.489 45.556
207371 100.557 -77.508 45.543
207372 100.581 -77.526 45.529
207373 100.612 -77.531 45.515
207374 100.642 -77.536 45.501
207375 100.673 -77.541 45.487
207376 100.698 -77.549 45.473
207377 100.721 -77.559 45.459
207378 100.744 -77.570 45.445
207379 100.766 -77.580 45.432
207380 100.789 -77.590 45.418
207381 100.812 -77.600 45.405
207382 100.838 -77.616 45.391
207383 100.864 -77.633 45.379
207384 100.891 -77.650 45.366
207385 100.918 -77.667 45.353
207386 100.945 -77.683 45.340
207387 100.971 -77.701 45.327
207388 100.997 -77.718 45.314
207389 101.024 -77.734 45.300
207390 101.051 -77.751 45.286
207391 101.077 -77.768 45.273
207392 101.104 -77.785 45.259
207393 101.131 -77.802 45.245
207394 101.158 -77.818 45.232
207395 101.184 -77.836 45.218
207396 101.211 -77.852 45.205
207397 101.238 -77.869 45.191
207398 101.264 -77.887 45.176
207399 101.290 -77.903 45.163
207400 101.315 -77.922 45.150
207401 101.339 -77.941 45.137
207402 101.364 -77.951 45.124
207403 101.393 -77.967 45.110
207404 101.418 -77.980 45.097
207405 101.427 -77.982 45.084
207406 101.437 -77.984 45.069
207407 101.446 -77.987 45.054
207408 101.456 -77.989 45.039
207409 101.465 -77.991 45.024
207410 101.495 -78.006 45.007
207411 101.539 -78.029 44.991
207412 101.583 -78.053 44.974
207413 101.627 -78.076 44.958
207414 101.654 -78.091 44.942
207415 101.665 -78.097 44.926
207416 101.676 -78.104 44.910
207417 101.686 -78.111 44.896
207418 101.703 -78.120 44.880
207419 101.723 -78.131 44.864
207420 101.743 -78.143 44.849
207421 101.764 -78.155 44.833
207422 101.784 -78.166 44.818
207423 101.805 -78.178 44.802
207424 101.825 -78.189 44.787
207425 101.845 -78.200 44.770
207426 101.866 -78.212 44.754
207427 101.886 -78.223 44.738
207428 101.906 -78.235 44.723
207429 101.933 -78.246 44.707
207430 101.967 -78.257 44.691
207431 102.002 -78.267 44.676
207432 102.037 -78.278 44.661
207433 102.071 -78.288 44.645
207434 102.106 -78.299 44.631
207435 102.141 -78.310 44.617
207436 102.179 -78.326 44.603
207437 102.221 -78.347 44.588
207438 102.261 -78.369 44.572
207439 102.303 -78.390 44.555
207440 102.343 -78.411 44.540
207441 102.385 -78.432 44.525
207442 102.426 -78.454 44.510
207443 102.467 -78.475 44.495
207444 102.507 -78.496 44.480
207445 102.549 -78.517 44.465
207446 102.590 -78.539 44.451
207447 102.631 -78.560 44.436
207448 102.672 -78.581 44.422
207449 102.682 -78.582 44.407
207450 102.681 -78.579 44.392
207451 102.707 -78.582 44.377
207452 102.733 -78.585 44.361
207453 102.758 -78.587 44.345
207454 102.784 -78.590 44.330
207455 102.810 -78.592 44.314
207456 102.835 -78.595 44.298
207457 102.861 -78.598 44.281
207458 102.887 -78.600 44.266
207459 102.912 -78.603 44.250
207460 102.938 -78.605 44.233
207461 102.964 -78.608 44.217
207746 93.460 -75.876 53.974
207747 93.478 -75.880 53.993
207748 93.496 -75.883 54.014
207749 93.514 -75.886 54.033
207750 93.533 -75.890 54.053
207751 93.551 -75.893 54.073
207752 93.569 -75.896 54.093
207753 93.587 -75.900 54.113
207754 93.606 -75.903 54.133
207755 93.612 -75.933 54.151
207756 93.643 -75.927 54.166
207757 93.676 -75.921 54.183
207758 93.708 -75.915 54.198
207759 93.741 -75.909 54.211
207760 93.773 -75.903 54.224
207761 93.800 -75.896 54.235
207762 93.821 -75.888 54.244
207763 93.842 -75.881 54.251
207764 93.863 -75.873 54.259
207765 93.884 -75.865 54.272
207766 93.910 -75.859 54.287
207767 93.942 -75.855 54.304
207768 93.974 -75.850 54.319
207769 94.000 -75.845 54.326
207770 94.026 -75.840 54.328
207771 94.052 -75.834 54.326
207772 94.079 -75.830 54.321
207773 94.106 -75.825 54.313
207774 94.133 -75.821 54.304
207775 94.160 -75.816 54.295
207776 94.187 -75.812 54.278
207777 94.214 -75.808 54.264
207778 94.242 -75.803 54.252
207779 94.269 -75.799 54.244
207780 94.293 -75.800 54.233
207781 94.313 -75.801 54.220
207782 94.333 -75.803 54.203
207783 94.352 -75.805 54.177
207784 94.371 -75.808 54.149
207785 94.390 -75.810 54.122
207786 94.408 -75.813 54.097
207787 94.427 -75.815 54.075
207788 94.445 -75.818 54.050
207789 94.464 -75.821 54.020
207790 94.480 -75.828 53.984
207791 94.494 -75.838 53.946
207792 94.508 -75.848 53.908
207793 94.523 -75.856 53.871
207794 94.538 -75.865 53.835
207795 94.552 -75.874 53.801
207796 94.567 -75.883 53.767
207797 94.582 -75.891 53.732
207798 94.597 -75.900 53.698
207799 94.612 -75.909 53.664
207800 94.626 -75.917 53.629
207801 94.641 -75.926 53.593
207802 94.656 -75.934 53.556
207803 94.671 -75.944 53.519
207804 94.683 -75.950 53.482
207805 94.692 -75.952 53.446
207806 94.702 -75.955 53.410
207807 94.711 -75.958 53.372
207808 94.720 -75.960 53.335
207809 94.730 -75.963 53.300
207810 94.739 -75.966 53.262
207811 94.748 -75.969 53.226
207812 94.758 -75.972 53.189
207813 94.767 -75.975 53.152
207814 94.776 -75.978 53.117
207815 94.785 -75.981 53.083
207816 94.795 -75.983 53.050
207817 94.806 -75.987 53.015
207818 94.817 -75.991 52.981
207819 94.828 -75.994 52.945
207820 94.839 -75.997 52.908
207821 94.850 -76.001 52.872
207822 94.861 -76.005 52.838
207823 94.872 -76.008 52.806
207824 94.883 -76.012 52.776
207825 94.894 -76.016 52.751
207826 94.905 -76.019 52.729
207827 94.916 -76.022 52.706
207828 94.927 -76.026 52.681
207829 94.938 -76.030 52.652
207830 94.979 -76.031 52.623
207831 95.042 -76.030 52.594
207832 95.105 -76.029 52.563
207833 95.167 -76.028 52.531
207834 95.230 -76.028 52.497
207835 95.293 -76.027 52.463
207836 95.356 -76.026 52.429
207837 95.411 -76.027 52.397
207838 95.426 -76.038 52.364
207839 95.454 -76.045 52.327
207840 95.483 -76.053 52.285
207841 95.512 -76.061 52.234
207842 95.537 -76.060 52.180
207843 95.562 -76.058 52.125
207844 95.587 -76.057 52.074
207845 95.612 -76.055 52.025
207846 95.634 -76.056 51.977
207847 95.643 -76.063 51.929
207848 95.653 -76.069 51.882
207849 95.663 -76.076 51.836
207850 95.673 -76.083 51.791
207851 95.682 -76.090 51.750
207852 95.692 -76.096 51.711
207853 95.702 -76.103 51.673
207854 95.722 -76.108 51.637
207855 95.744 -76.113 51.600
207856 95.766 -76.116 51.561
207857 95.789 -76.121 51.525
207858 95.811 -76.126 51.489
207859 95.833 -76.130 51.450
207860 95.856 -76.135 51.411
207861 95.878 -76.138 51.370
207862 95.910 -76.141 51.328
207863 95.950 -76.141 51.286
207864 95.990 -76.140 51.244
207865 96.013 -76.140 51.201
207866 96.036 -76.139 51.158
207867 96.060 -76.138 51.115
207868 96.083 -76.138 51.070
207869 96.107 -76.137 51.025
207870 96.129 -76.138 50.982
207871 96.151 -76.140 50.945
207872 96.173 -76.142 50.908
207873 96.194 -76.144 50.873
207874 96.217 -76.145 50.839
207875 96.227 -76.154 50.808
207876 96.235 -76.164 50.776
207877 96.242 -76.175 50.745
207878 96.245 -76.190 50.717
207879 96.267 -76.201 50.689
207880 96.273 -76.208 50.659
207881 96.282 -76.215 50.630
207882 96.293 -76.222 50.602
207883 96.305 -76.229 50.576
207884 96.317 -76.236 50.553
207885 96.329 -76.243 50.534
207886 96.341 -76.250 50.516
207887 96.356 -76.258 50.496
207888 96.380 -76.272 50.476
207889 96.404 -76.286 50.453
207890 96.447 -76.298 50.427
207891 96.492 -76.309 50.400
207892 96.537 -76.320 50.373
207893 96.583 -76.332 50.345
207894 96.626 -76.345 50.316
207895 96.660 -76.348 50.287
207896 96.675 -76.348 50.256
207897 96.690 -76.348 50.224
207898 96.708 -76.340 50.192
207899 96.726 -76.332 50.159
207900 96.749 -76.337 50.125
207901 96.774 -76.347 50.089
207902 96.799 -76.357 50.053
207903 96.816 -76.358 50.015
207904 96.822 -76.351 49.978
207905 96.822 -76.342 49.940
207906 96.804 -76.334 49.901
207907 96.844 -76.348 49.862
207908 96.890 -76.363 49.823
207909 96.929 -76.376 49.785
207910 96.952 -76.383 49.746
207911 96.975 -76.391 49.707
207912 96.998 -76.398 49.669
207913 97.036 -76.423 49.633
207914 97.050 -76.452 49.598
207915 97.065 -76.481 49.564
207916 97.079 -76.509 49.532
207917 97.094 -76.538 49.505
207918 97.108 -76.566 49.480
207919 97.122 -76.595 49.456
207920 97.137 -76.624 49.433
207921 97.151 -76.652 49.412
207922 97.166 -76.680 49.390
207923 97.180 -76.708 49.368
207924 97.194 -76.737 49.343
207925 97.197 -76.759 49.316
207926 97.232 -76.756 49.287
207927 97.264 -76.743 49.256
207928 97.296 -76.730 49.223
207929 97.328 -76.716 49.189
207930 97.361 -76.703 49.153
207931 97.393 -76.690 49.119
207932 97.426 -76.676 49.085
207933 97.450 -76.675 49.053
207934 97.473 -76.687 49.023
207935 97.497 -76.699 48.993
207936 97.521 -76.711 48.966
207937 97.545 -76.723 48.940
207938 97.568 -76.734 48.913
207939 97.591 -76.746 48.886
207940 97.614 -76.758 48.860
207941 97.638 -76.769 48.832
207942 97.661 -76.781 48.804
207943 97.684 -76.793 48.775
207944 97.707 -76.805 48.745
207945 97.730 -76.816 48.714
207946 97.753 -76.828 48.682
207947 97.777 -76.840 48.649
207948 97.813 -76.850 48.617
207949 97.853 -76.857 48.585
207950 97.893 -76.865 48.552
207951 97.934 -76.874 48.522
207952 97.974 -76.881 48.490
207953 98.017 -76.888 48.459
207954 98.062 -76.893 48.428
207955 98.107 -76.897 48.398
207956 98.132 -76.909 48.367
207957 98.151 -76.904 48.337
207958 98.169 -76.897 48.306
207959 98.187 -76.891 48.277
207960 98.206 -76.884 48.247
207961 98.224 -76.878 48.218
207962 98.243 -76.869 48.188
207963 98.262 -76.862 48.158
207964 98.280 -76.854 48.129
207965 98.299 -76.846 48.101
207966 98.317 -76.838 48.070
207967 98.336 -76.830 48.039
207968 98.354 -76.822 48.007
207969 98.375 -76.821 47.976
207970 98.397 -76.822 47.944
207971 98.419 -76.822 47.915
207972 98.443 -76.811 47.885
207973 98.458 -76.813 47.856
207974 98.471 -76.818 47.826
207975 98.484 -76.824 47.798
207976 98.498 -76.831 47.771
207977 98.516 -76.848 47.743
207978 98.534 -76.864 47.719
207979 98.553 -76.880 47.693
207980 98.572 -76.897 47.667
207981 98.590 -76.912 47.639
207982 98.608 -76.928 47.610
207983 98.631 -76.929 47.581
207984 98.655 -76.923 47.552
207985 98.679 -76.918 47.525
207986 98.703 -76.912 47.497
207987 98.713 -76.917 47.471
207988 98.706 -76.935 47.445
207989 98.735 -76.938 47.417
207990 98.768 -76.940 47.387
207991 98.801 -76.943 47.356
207992 98.834 -76.945 47.324
207993 98.867 -76.947 47.293
207994 98.900 -76.949 47.260
207995 98.933 -76.951 47.228
207996 98.966 -76.953 47.194
207997 98.999 -76.959 47.159
207998 99.015 -76.972 47.125
207999 99.032 -76.985 47.091
208000 99.049 -76.999 47.058
208001 99.066 -77.012 47.027
208002 99.083 -77.026 46.995
208003 99.099 -77.040 46.963
208004 99.116 -77.053 46.930
208005 99.133 -77.067 46.900
208006 99.150 -77.080 46.871
208007 99.173 -77.084 46.844
208008 99.204 -77.074 46.816
208009 99.235 -77.065 46.788
208010 99.266 -77.055 46.761
208011 99.289 -77.049 46.733
208012 99.307 -77.064 46.705
208013 99.351 -77.077 46.679
208014 99.375 -77.091 46.652
208015 99.400 -77.106 46.624
208016 99.424 -77.122 46.597
208017 99.431 -77.135 46.570
208018 99.438 -77.150 46.541
208019 99.444 -77.163 46.512
208020 99.451 -77.178 46.485
208021 99.457 -77.191 46.459
208022 99.444 -77.186 46.432
208023 99.430 -77.180 46.406
208024 99.416 -77.174 46.380
208025 99.402 -77.168 46.354
208026 99.389 -77.162 46.328
208027 99.375 -77.156 46.306
208028 99.385 -77.153 46.283
208029 99.436 -77.160 46.260
208030 99.487 -77.167 46.237
208031 99.538 -77.174 46.215
208032 99.589 -77.182 46.192
208033 99.640 -77.188 46.170
208034 99.691 -77.195 46.149
208035 99.728 -77.195 46.128
208036 99.753 -77.188 46.108
208037 99.778 -77.182 46.088
208038 99.802 -77.175 46.068
208039 99.833 -77.172 46.047
208040 99.870 -77.172 46.028
208041 99.907 -77.172 46.010
208042 99.941 -77.181 45.992
208043 99.973 -77.204 45.974
208044 100.003 -77.225 45.956
208045 100.026 -77.235 45.938
208046 100.049 -77.245 45.920
208047 100.069 -77.253 45.902
208048 100.081 -77.244 45.885
208049 100.093 -77.236 45.867
208050 100.106 -77.229 45.850
208051 100.128 -77.231 45.832
208052 100.162 -77.244 45.814
208053 100.196 -77.258 45.796
208054 100.230 -77.272 45.780
208055 100.264 -77.285 45.764
208056 100.278 -77.290 45.748
208057 100.289 -77.294 45.733
208058 100.301 -77.298 45.719
208059 100.313 -77.302 45.705
208060 100.323 -77.313 45.691
208061 100.333 -77.324 45.677
208062 100.343 -77.335 45.664
208063 100.362 -77.342 45.651
208064 100.382 -77.349 45.638
208065 100.405 -77.363 45.625
208066 100.429 -77.379 45.613
208067 100.452 -77.400 45.600
208068 100.476 -77.420 45.588
208069 100.499 -77.440 45.576
208070 100.522 -77.460 45.562
208071 100.546 -77.479 45.549
208072 100.569 -77.500 45.536
208073 100.593 -77.520 45.522
208074 100.623 -77.525 45.508
208075 100.653 -77.530 45.495
208076 100.684 -77.535 45.481
208077 100.714 -77.540 45.468
208078 100.742 -77.547 45.454
208079 100.765 -77.557 45.441
208080 100.787 -77.567 45.426
208081 100.810 -77.577 45.413
208082 100.833 -77.588 45.400
208083 100.859 -77.605 45.386
208084 100.886 -77.621 45.374
208085 100.913 -77.638 45.361
208086 100.939 -77.655 45.348
208087 100.966 -77.672 45.335
208088 100.993 -77.689 45.322
208089 101.019 -77.705 45.309
208090 101.046 -77.723 45.296
208091 101.073 -77.740 45.282
208092 101.100 -77.756 45.268
208093 101.126 -77.773 45.254
208094 101.153 -77.790 45.240
208095 101.179 -77.807 45.228
208096 101.204 -77.825 45.214
208097 101.229 -77.844 45.200
208098 101.252 -77.863 45.186
208099 101.276 -77.882 45.171
208100 101.301 -77.901 45.158
208101 101.326 -77.912 45.144
208102 101.352 -77.921 45.131
208103 101.379 -77.934 45.117
208104 101.408 -77.950 45.104
208105 101.437 -77.967 45.090
208106 101.466 -77.983 45.076
208107 101.481 -77.990 45.061
208108 101.491 -77.992 45.046
208109 101.500 -77.994 45.031
208110 101.510 -77.996 45.016
208111 101.525 -78.002 45.000
208112 101.568 -78.025 44.983
208113 101.612 -78.049 44.966
208114 101.656 -78.072 44.950
208115 101.672 -78.081 44.934
208116 101.682 -78.087 44.918
208117 101.693 -78.094 44.902
208118 101.704 -78.101 44.887
208119 101.715 -78.107 44.872
208120 101.735 -78.119 44.856
208121 101.755 -78.130 44.841
208122 101.776 -78.141 44.825
208123 101.796 -78.153 44.810
208124 101.816 -78.165 44.796
208125 101.837 -78.176 44.781
208126 101.857 -78.188 44.764
208127 101.878 -78.199 44.747
208128 101.898 -78.210 44.731
208129 101.921 -78.222 44.717
208130 101.957 -78.232 44.701
208131 101.991 -78.243 44.685
208132 102.026 -78.253 44.670
208133 102.061 -78.264 44.654
208134 102.096 -78.275 44.639
208135 102.130 -78.285 44.624
208136 102.165 -78.296 44.609
208137 102.204 -78.313 44.595
208138 102.245 -78.335 44.580
208139 102.287 -78.357 44.565
208140 102.327 -78.378 44.548
208141 102.369 -78.399 44.532
208142 102.409 -78.420 44.517
208143 102.451 -78.442 44.501
208144 102.491 -78.463 44.487
208145 102.533 -78.484 44.472
208146 102.574 -78.505 44.458
208147 102.615 -78.526 44.443
208148 102.655 -78.547 44.429
208149 102.686 -78.553 44.415
208150 102.701 -78.543 44.400
208151 102.699 -78.546 44.385
208152 102.736 -78.560 44.370
208153 102.769 -78.570 44.353
208154 102.795 -78.573 44.337
208155 102.820 -78.575 44.322
208156 102.846 -78.577 44.306
208157 102.872 -78.580 44.289
208158 102.897 -78.582 44.273
208159 102.923 -78.585 44.258
208160 102.949 -78.588 44.241
208161 102.974 -78.590 44.225
208162 103.000 -78.593 44.208
208446 93.433 -75.850 53.951
208447 93.451 -75.853 53.970
208448 93.470 -75.856 53.990
208449 93.488 -75.859 54.011
208450 93.506 -75.863 54.030
208451 93.524 -75.866 54.051
208452 93.543 -75.869 54.072
208453 93.561 -75.873 54.093
208454 93.579 -75.876 54.115
208455 93.597 -75.879 54.136
208456 93.606 -75.906 54.156
208457 93.638 -75.900 54.173
208458 93.671 -75.894 54.193
208459 93.703 -75.887 54.211
208460 93.735 -75.881 54.226
208461 93.768 -75.875 54.239
208462 93.800 -75.869 54.249
208463 93.825 -75.862 54.255
208464 93.845 -75.854 54.258
208465 93.866 -75.847 54.261
208466 93.887 -75.839 54.269
208467 93.911 -75.832 54.280
208468 93.943 -75.828 54.293
208469 93.974 -75.823 54.307
208470 94.001 -75.818 54.312
208471 94.027 -75.812 54.313
208472 94.053 -75.807 54.311
208473 94.079 -75.802 54.306
208474 94.106 -75.797 54.297
208475 94.133 -75.793 54.288
208476 94.160 -75.789 54.278
208477 94.187 -75.784 54.261
208478 94.214 -75.780 54.248
208479 94.241 -75.777 54.237
208480 94.266 -75.777 54.229
208481 94.291 -75.777 54.220
208482 94.313 -75.778 54.209
208483 94.332 -75.780 54.194
208484 94.352 -75.782 54.171
208485 94.371 -75.784 54.145
208486 94.391 -75.785 54.120
208487 94.411 -75.787 54.097
208488 94.431 -75.789 54.077
208489 94.450 -75.792 54.053
208490 94.468 -75.794 54.024
208491 94.483 -75.803 53.988
208492 94.498 -75.812 53.950
208493 94.512 -75.820 53.913
208494 94.528 -75.829 53.878
208495 94.542 -75.837 53.845
208496 94.557 -75.847 53.815
208497 94.572 -75.855 53.784
208498 94.586 -75.864 53.752
208499 94.602 -75.872 53.719
208500 94.616 -75.881 53.685
208501 94.631 -75.890 53.650
208502 94.646 -75.898 53.613
208503 94.661 -75.907 53.575
208504 94.676 -75.916 53.536
208505 94.688 -75.922 53.497
208506 94.697 -75.925 53.458
208507 94.706 -75.928 53.420
208508 94.716 -75.930 53.382
208509 94.725 -75.933 53.345
208510 94.735 -75.936 53.308
208511 94.744 -75.939 53.271
208512 94.753 -75.941 53.234
208513 94.762 -75.944 53.197
208514 94.772 -75.947 53.157
208515 94.781 -75.950 53.119
208516 94.790 -75.953 53.082
208517 94.800 -75.956 53.046
208518 94.809 -75.959 53.011
208519 94.819 -75.961 52.976
208520 94.828 -75.964 52.940
208521 94.839 -75.968 52.904
208522 94.850 -75.972 52.869
208523 94.861 -75.975 52.836
208524 94.873 -75.978 52.806
208525 94.884 -75.982 52.778
208526 94.895 -75.986 52.755
208527 94.905 -75.989 52.735
208528 94.916 -75.993 52.714
208529 94.927 -75.997 52.690
208530 94.938 -76.000 52.662
208531 94.977 -76.001 52.633
208532 95.040 -76.000 52.604
208533 95.103 -76.000 52.573
208534 95.166 -75.999 52.542
208535 95.229 -75.998 52.507
208536 95.291 -75.997 52.472
208537 95.354 -75.997 52.437
208538 95.417 -75.996 52.403
208539 95.473 -75.997 52.368
208540 95.495 -76.006 52.331
208541 95.524 -76.013 52.288
208542 95.550 -76.014 52.238
208543 95.575 -76.013 52.184
208544 95.598 -76.012 52.129
208545 95.608 -76.019 52.076
208546 95.618 -76.025 52.027
208547 95.627 -76.032 51.980
208548 95.637 -76.039 51.933
208549 95.647 -76.046 51.887
208550 95.657 -76.053 51.841
208551 95.666 -76.060 51.797
208552 95.676 -76.066 51.754
208553 95.686 -76.073 51.713
208554 95.695 -76.079 51.676
208555 95.712 -76.085 51.641
208556 95.735 -76.089 51.606
208557 95.757 -76.094 51.570
208558 95.780 -76.098 51.534
208559 95.802 -76.103 51.498
208560 95.824 -76.107 51.459
208561 95.847 -76.111 51.420
208562 95.870 -76.116 51.381
208563 95.910 -76.116 51.340
208564 95.950 -76.115 51.299
208565 95.990 -76.115 51.257
208566 96.020 -76.115 51.215
208567 96.043 -76.114 51.171
208568 96.066 -76.113 51.128
208569 96.090 -76.113 51.083
208570 96.113 -76.112 51.039
208571 96.136 -76.111 50.996
208572 96.160 -76.110 50.958
208573 96.182 -76.111 50.921
208574 96.204 -76.113 50.885
208575 96.226 -76.115 50.852
208576 96.237 -76.123 50.820
208577 96.242 -76.135 50.788
208578 96.241 -76.151 50.758
208579 96.248 -76.166 50.731
208580 96.270 -76.176 50.702
208581 96.292 -76.188 50.673
208582 96.300 -76.195 50.645
208583 96.306 -76.202 50.617
208584 96.312 -76.209 50.591
208585 96.321 -76.216 50.568
208586 96.333 -76.223 50.548
208587 96.345 -76.230 50.529
208588 96.357 -76.237 50.509
208589 96.375 -76.247 50.488
208590 96.399 -76.261 50.464
208591 96.444 -76.273 50.439
208592 96.489 -76.284 50.413
208593 96.534 -76.295 50.387
208594 96.577 -76.310 50.360
208595 96.620 -76.323 50.332
208596 96.655 -76.328 50.303
208597 96.683 -76.326 50.273
208598 96.698 -76.326 50.241
208599 96.714 -76.326 50.209
208600 96.729 -76.326 50.177
208601 96.744 -76.325 50.142
208602 96.759 -76.325 50.107
208603 96.779 -76.321 50.071
208604 96.801 -76.328 50.033
208605 96.799 -76.321 49.995
208606 96.839 -76.334 49.957
208607 96.862 -76.342 49.918
208608 96.885 -76.349 49.879
208609 96.908 -76.357 49.841
208610 96.931 -76.365 49.802
208611 96.954 -76.373 49.764
208612 96.977 -76.380 49.725
208613 97.000 -76.388 49.687
208614 97.043 -76.409 49.650
208615 97.084 -76.434 49.615
208616 97.098 -76.462 49.582
208617 97.112 -76.491 49.549
208618 97.126 -76.520 49.522
208619 97.141 -76.548 49.496
208620 97.156 -76.577 49.473
208621 97.170 -76.605 49.450
208622 97.184 -76.633 49.428
208623 97.199 -76.662 49.404
208624 97.213 -76.690 49.380
208625 97.180 -76.701 49.354
208626 97.220 -76.719 49.325
208627 97.255 -76.718 49.294
208628 97.288 -76.705 49.262
208629 97.320 -76.692 49.228
208630 97.352 -76.678 49.193
208631 97.384 -76.665 49.158
208632 97.417 -76.652 49.124
208633 97.449 -76.638 49.092
208634 97.460 -76.649 49.061
208635 97.479 -76.661 49.031
208636 97.503 -76.673 49.002
208637 97.527 -76.685 48.976
208638 97.550 -76.696 48.950
208639 97.574 -76.708 48.924
208640 97.598 -76.721 48.897
208641 97.622 -76.732 48.870
208642 97.646 -76.744 48.843
208643 97.669 -76.756 48.815
208644 97.693 -76.768 48.786
208645 97.717 -76.780 48.756
208646 97.741 -76.792 48.724
208647 97.764 -76.803 48.692
208648 97.788 -76.815 48.659
208649 97.820 -76.825 48.626
208650 97.865 -76.828 48.594
208651 97.911 -76.833 48.561
208652 97.956 -76.838 48.530
208653 98.001 -76.843 48.498
208654 98.046 -76.847 48.466
208655 98.092 -76.853 48.436
208656 98.125 -76.862 48.405
208657 98.147 -76.875 48.374
208658 98.168 -76.887 48.344
208659 98.188 -76.887 48.314
208660 98.206 -76.881 48.284
208661 98.225 -76.875 48.254
208662 98.243 -76.868 48.225
208663 98.262 -76.859 48.196
208664 98.280 -76.852 48.167
208665 98.299 -76.844 48.139
208666 98.317 -76.836 48.111
208667 98.336 -76.828 48.082
208668 98.357 -76.826 48.051
208669 98.378 -76.827 48.020
208670 98.400 -76.828 47.989
208671 98.422 -76.829 47.957
208672 98.444 -76.825 47.928
208673 98.469 -76.815 47.900
208674 98.490 -76.809 47.871
208675 98.502 -76.815 47.841
208676 98.515 -76.820 47.813
208677 98.527 -76.825 47.785
208678 98.541 -76.831 47.758
208679 98.553 -76.837 47.733
208680 98.569 -76.849 47.707
208681 98.588 -76.865 47.680
208682 98.606 -76.881 47.652
208683 98.625 -76.897 47.623
208684 98.643 -76.913 47.594
208685 98.665 -76.916 47.564
208686 98.689 -76.910 47.535
208687 98.713 -76.905 47.505
208688 98.727 -76.906 47.477
208689 98.725 -76.922 47.449
208690 98.758 -76.925 47.421
208691 98.791 -76.927 47.391
208692 98.823 -76.929 47.360
208693 98.856 -76.931 47.328
208694 98.889 -76.933 47.295
208695 98.922 -76.935 47.262
208696 98.955 -76.937 47.228
208697 98.995 -76.937 47.195
208698 99.033 -76.940 47.160
208699 99.050 -76.954 47.125
208700 99.067 -76.968 47.092
208701 99.084 -76.981 47.059
208702 99.100 -76.995 47.027
208703 99.118 -77.009 46.994
208704 99.134 -77.022 46.961
208705 99.151 -77.035 46.927
208706 99.168 -77.049 46.896
208707 99.198 -77.041 46.866
208708 99.228 -77.031 46.838
208709 99.260 -77.022 46.809
208710 99.274 -77.014 46.781
208711 99.292 -77.028 46.753
208712 99.309 -77.044 46.725
208713 99.327 -77.058 46.698
208714 99.368 -77.071 46.671
208715 99.432 -77.082 46.644
208716 99.450 -77.097 46.617
208717 99.457 -77.110 46.590
208718 99.463 -77.125 46.563
208719 99.470 -77.138 46.534
208720 99.476 -77.153 46.505
208721 99.483 -77.167 46.478
208722 99.478 -77.169 46.452
208723 99.464 -77.163 46.425
208724 99.450 -77.157 46.399
208725 99.436 -77.151 46.373
208726 99.422 -77.145 46.347
208727 99.430 -77.143 46.321
208728 99.440 -77.141 46.299
208729 99.450 -77.138 46.275
208730 99.471 -77.139 46.252
208731 99.522 -77.146 46.230
208732 99.573 -77.153 46.208
208733 99.625 -77.160 46.185
208734 99.676 -77.167 46.162
208735 99.719 -77.170 46.141
208736 99.744 -77.163 46.122
208737 99.769 -77.157 46.101
208738 99.793 -77.150 46.081
208739 99.818 -77.144 46.060
208740 99.851 -77.142 46.041
208741 99.889 -77.142 46.022
208742 99.925 -77.141 46.004
208743 99.961 -77.146 45.985
208744 99.992 -77.169 45.968
208745 100.023 -77.191 45.950
208746 100.050 -77.207 45.932
208747 100.072 -77.217 45.914
208748 100.095 -77.227 45.896
208749 100.115 -77.233 45.879
208750 100.127 -77.225 45.861
208751 100.139 -77.216 45.843
208752 100.156 -77.213 45.825
208753 100.190 -77.228 45.807
208754 100.224 -77.241 45.790
208755 100.258 -77.255 45.774
208756 100.282 -77.264 45.758
208757 100.293 -77.267 45.742
208758 100.305 -77.271 45.726
208759 100.317 -77.274 45.712
208760 100.328 -77.279 45.698
208761 100.338 -77.290 45.684
208762 100.348 -77.301 45.670
208763 100.359 -77.311 45.657
208764 100.373 -77.321 45.644
208765 100.395 -77.332 45.631
208766 100.418 -77.351 45.618
208767 100.441 -77.371 45.606
208768 100.465 -77.391 45.593
208769 100.488 -77.411 45.580
208770 100.511 -77.431 45.568
208771 100.535 -77.451 45.555
208772 100.558 -77.471 45.542
208773 100.581 -77.491 45.529
208774 100.605 -77.511 45.515
208775 100.634 -77.519 45.502
208776 100.665 -77.524 45.489
208777 100.695 -77.530 45.475
208778 100.725 -77.534 45.463
208779 100.756 -77.539 45.450
208780 100.786 -77.545 45.437
208781 100.809 -77.555 45.422
208782 100.832 -77.565 45.409
208783 100.855 -77.576 45.396
208784 100.881 -77.592 45.383
208785 100.908 -77.609 45.370
208786 100.934 -77.627 45.357
208787 100.961 -77.643 45.344
208788 100.988 -77.660 45.331
208789 101.015 -77.677 45.318
208790 101.041 -77.694 45.305
208791 101.068 -77.711 45.291
208792 101.094 -77.728 45.277
208793 101.118 -77.747 45.263
208794 101.142 -77.766 45.249
208795 101.166 -77.785 45.235
208796 101.190 -77.804 45.223
208797 101.214 -77.823 45.209
208798 101.238 -77.842 45.196
208799 101.262 -77.861 45.181
208800 101.287 -77.874 45.166
208801 101.314 -77.883 45.152
208802 101.339 -77.893 45.138
208803 101.365 -77.902 45.125
208804 101.394 -77.918 45.111
208805 101.423 -77.934 45.097
208806 101.452 -77.950 45.083
208807 101.481 -77.967 45.067
208808 101.510 -77.983 45.052
208809 101.536 -77.997 45.038
208810 101.545 -78.000 45.023
208811 101.554 -78.002 45.007
208812 101.564 -78.004 44.991
208813 101.598 -78.022 44.975
208814 101.642 -78.044 44.958
208815 101.679 -78.065 44.941
208816 101.689 -78.071 44.925
208817 101.700 -78.078 44.909
208818 101.711 -78.084 44.894
208819 101.722 -78.091 44.879
208820 101.733 -78.097 44.863
208821 101.747 -78.106 44.848
208822 101.767 -78.117 44.834
208823 101.787 -78.128 44.818
208824 101.808 -78.140 44.803
208825 101.828 -78.152 44.788
208826 101.849 -78.163 44.772
208827 101.869 -78.175 44.756
208828 101.890 -78.186 44.740
208829 101.911 -78.197 44.724
208830 101.946 -78.208 44.708
208831 101.980 -78.219 44.693
208832 102.015 -78.229 44.677
208833 102.050 -78.240 44.662
208834 102.085 -78.250 44.646
208835 102.120 -78.261 44.631
208836 102.155 -78.272 44.616
208837 102.189 -78.282 44.601
208838 102.230 -78.302 44.587
208839 102.270 -78.323 44.572
208840 102.312 -78.344 44.556
208841 102.352 -78.366 44.540
208842 102.394 -78.387 44.524
208843 102.435 -78.408 44.508
208844 102.476 -78.429 44.493
208845 102.517 -78.451 44.479
208846 102.558 -78.472 44.465
208847 102.599 -78.493 44.450
208848 102.633 -78.505 44.435
208849 102.664 -78.511 44.421
208850 102.694 -78.517 44.407
208851 102.709 -78.504 44.392
208852 102.716 -78.512 44.377
208853 102.753 -78.526 44.361
208854 102.790 -78.539 44.345
208855 102.827 -78.553 44.329
208856 102.857 -78.560 44.313
208857 102.883 -78.563 44.297
208858 102.908 -78.565 44.280
208859 102.934 -78.567 44.264
208860 102.960 -78.570 44.249
208861 102.985 -78.573 44.232
208862 103.011 -78.575 44.216
208863 103.037 -78.578 44.199
209147 93.425 -75.827 53.947
209148 93.443 -75.830 53.966
209149 93.462 -75.833 53.987
209150 93.480 -75.837 54.008
209151 93.498 -75.840 54.028
209152 93.516 -75.843 54.050
209153 93.535 -75.847 54.072
209154 93.553 -75.850 54.094
209155 93.571 -75.853 54.118
209156 93.589 -75.856 54.140
209157 93.601 -75.878 54.161
209158 93.633 -75.872 54.181
209159 93.665 -75.865 54.202
209160 93.698 -75.859 54.223
209161 93.730 -75.853 54.242
209162 93.763 -75.847 54.257
209163 93.795 -75.841 54.267
209164 93.828 -75.835 54.273
209165 93.849 -75.828 54.274
209166 93.870 -75.820 54.273
209167 93.891 -75.812 54.275
209168 93.912 -75.805 54.279
209169 93.943 -75.800 54.286
209170 93.975 -75.796 54.296
209171 94.001 -75.790 54.299
209172 94.028 -75.785 54.299
209173 94.053 -75.780 54.296
209174 94.079 -75.775 54.292
209175 94.106 -75.769 54.284
209176 94.133 -75.765 54.274
209177 94.160 -75.761 54.264
209178 94.187 -75.756 54.249
209179 94.213 -75.754 54.235
209180 94.238 -75.755 54.224
209181 94.263 -75.755 54.215
209182 94.288 -75.755 54.207
209183 94.312 -75.756 54.198
209184 94.331 -75.757 54.186
209185 94.351 -75.759 54.164
209186 94.371 -75.761 54.140
209187 94.391 -75.762 54.116
209188 94.410 -75.765 54.095
209189 94.430 -75.766 54.076
209190 94.450 -75.768 54.054
209191 94.467 -75.771 54.026
209192 94.483 -75.778 53.990
209193 94.502 -75.784 53.955
209194 94.518 -75.793 53.919
209195 94.532 -75.801 53.885
209196 94.547 -75.810 53.856
209197 94.562 -75.818 53.830
209198 94.576 -75.828 53.802
209199 94.592 -75.836 53.771
209200 94.606 -75.844 53.739
209201 94.621 -75.853 53.705
209202 94.636 -75.862 53.669
209203 94.650 -75.871 53.632
209204 94.666 -75.879 53.593
209205 94.680 -75.887 53.552
209206 94.693 -75.894 53.510
209207 94.702 -75.897 53.469
209208 94.711 -75.900 53.431
209209 94.720 -75.903 53.392
209210 94.730 -75.906 53.355
209211 94.740 -75.908 53.319
209212 94.749 -75.911 53.281
209213 94.758 -75.914 53.243
209214 94.767 -75.917 53.204
209215 94.777 -75.919 53.164
209216 94.786 -75.922 53.124
209217 94.795 -75.925 53.084
209218 94.805 -75.928 53.046
209219 94.814 -75.931 53.008
209220 94.823 -75.934 52.972
209221 94.833 -75.937 52.936
209222 94.842 -75.939 52.901
209223 94.851 -75.942 52.867
209224 94.861 -75.945 52.836
209225 94.873 -75.949 52.806
209226 94.884 -75.953 52.779
209227 94.895 -75.956 52.757
209228 94.906 -75.960 52.737
209229 94.917 -75.963 52.716
209230 94.928 -75.967 52.694
209231 94.939 -75.970 52.667
209232 94.976 -75.972 52.639
209233 95.039 -75.971 52.609
209234 95.102 -75.970 52.579
209235 95.165 -75.969 52.548
209236 95.227 -75.969 52.514
209237 95.290 -75.968 52.478
209238 95.353 -75.967 52.442
209239 95.416 -75.966 52.405
209240 95.478 -75.966 52.367
209241 95.536 -75.966 52.328
209242 95.562 -75.969 52.285
209243 95.572 -75.975 52.235
209244 95.582 -75.981 52.182
209245 95.592 -75.988 52.127
209246 95.602 -75.995 52.073
209247 95.611 -76.002 52.024
209248 95.621 -76.009 51.978
209249 95.631 -76.016 51.934
209250 95.640 -76.022 51.889
209251 95.651 -76.029 51.844
209252 95.660 -76.036 51.800
209253 95.670 -76.043 51.758
209254 95.680 -76.050 51.717
209255 95.689 -76.056 51.677
209256 95.704 -76.062 51.641
209257 95.726 -76.066 51.608
209258 95.748 -76.071 51.575
209259 95.771 -76.075 51.542
209260 95.793 -76.079 51.507
209261 95.815 -76.084 51.469
209262 95.838 -76.088 51.432
209263 95.869 -76.091 51.393
209264 95.910 -76.090 51.354
209265 95.950 -76.090 51.314
209266 95.990 -76.090 51.272
209267 96.026 -76.089 51.229
209268 96.049 -76.089 51.184
209269 96.073 -76.088 51.141
209270 96.096 -76.088 51.096
209271 96.120 -76.087 51.052
209272 96.143 -76.086 51.011
209273 96.167 -76.085 50.972
209274 96.190 -76.085 50.934
209275 96.213 -76.085 50.898
209276 96.236 -76.085 50.864
209277 96.243 -76.095 50.832
209278 96.241 -76.112 50.801
209279 96.239 -76.128 50.771
209280 96.251 -76.141 50.743
209281 96.273 -76.152 50.715
209282 96.295 -76.163 50.686
209283 96.317 -76.174 50.658
209284 96.327 -76.182 50.631
209285 96.333 -76.189 50.605
209286 96.339 -76.196 50.581
209287 96.345 -76.203 50.560
209288 96.352 -76.210 50.541
209289 96.361 -76.216 50.521
209290 96.373 -76.223 50.499
209291 96.396 -76.236 50.477
209292 96.442 -76.247 50.451
209293 96.485 -76.260 50.426
209294 96.528 -76.274 50.401
209295 96.571 -76.289 50.375
209296 96.613 -76.302 50.348
209297 96.650 -76.307 50.319
209298 96.679 -76.306 50.289
209299 96.707 -76.304 50.259
209300 96.722 -76.304 50.227
209301 96.738 -76.304 50.194
209302 96.753 -76.304 50.159
209303 96.768 -76.303 50.124
209304 96.784 -76.303 50.088
209305 96.799 -76.303 50.050
209306 96.818 -76.316 50.012
209307 96.841 -76.323 49.973
209308 96.864 -76.331 49.934
209309 96.887 -76.339 49.895
209310 96.910 -76.346 49.857
209311 96.933 -76.354 49.820
209312 96.956 -76.361 49.781
209313 96.979 -76.369 49.743
209314 97.002 -76.376 49.705
209315 97.042 -76.395 49.668
209316 97.089 -76.420 49.631
209317 97.131 -76.444 49.597
209318 97.146 -76.473 49.566
209319 97.160 -76.501 49.538
209320 97.174 -76.530 49.513
209321 97.188 -76.558 49.489
209322 97.203 -76.586 49.466
209323 97.218 -76.615 49.442
209324 97.232 -76.643 49.417
209325 97.163 -76.643 49.390
209326 97.203 -76.661 49.363
209327 97.243 -76.679 49.333
209328 97.279 -76.680 49.301
209329 97.311 -76.667 49.267
209330 97.344 -76.654 49.232
209331 97.376 -76.640 49.197
209332 97.408 -76.627 49.164
209333 97.440 -76.614 49.132
209334 97.465 -76.611 49.101
209335 97.475 -76.622 49.070
209336 97.485 -76.635 49.040
209337 97.509 -76.647 49.012
209338 97.532 -76.659 48.985
209339 97.556 -76.671 48.960
209340 97.580 -76.683 48.934
209341 97.604 -76.695 48.907
209342 97.628 -76.706 48.882
209343 97.651 -76.718 48.855
209344 97.675 -76.730 48.827
209345 97.699 -76.742 48.798
209346 97.723 -76.754 48.767
209347 97.746 -76.765 48.736
209348 97.770 -76.777 48.703
209349 97.794 -76.790 48.671
209350 97.845 -76.796 48.638
209351 97.897 -76.792 48.605
209352 97.940 -76.793 48.572
209353 97.986 -76.797 48.540
209354 98.031 -76.802 48.507
209355 98.076 -76.807 48.475
209356 98.118 -76.814 48.444
209357 98.139 -76.827 48.413
209358 98.161 -76.840 48.382
209359 98.183 -76.853 48.352
209360 98.205 -76.865 48.323
209361 98.225 -76.871 48.293
209362 98.243 -76.865 48.264
209363 98.262 -76.858 48.235
209364 98.281 -76.850 48.205
209365 98.299 -76.842 48.177
209366 98.317 -76.834 48.149
209367 98.338 -76.831 48.121
209368 98.360 -76.832 48.092
209369 98.381 -76.833 48.062
209370 98.403 -76.834 48.032
209371 98.425 -76.834 48.001
209372 98.447 -76.835 47.971
209373 98.470 -76.828 47.941
209374 98.495 -76.818 47.912
209375 98.519 -76.807 47.885
209376 98.533 -76.810 47.856
209377 98.546 -76.815 47.828
209378 98.559 -76.821 47.800
209379 98.571 -76.827 47.772
209380 98.584 -76.832 47.745
209381 98.596 -76.838 47.719
209382 98.609 -76.843 47.692
209383 98.623 -76.850 47.663
209384 98.641 -76.865 47.636
209385 98.659 -76.882 47.607
209386 98.678 -76.898 47.576
209387 98.699 -76.903 47.545
209388 98.723 -76.897 47.515
209389 98.742 -76.896 47.484
209390 98.747 -76.909 47.454
209391 98.780 -76.911 47.425
209392 98.813 -76.913 47.394
209393 98.846 -76.915 47.363
209394 98.879 -76.917 47.331
209395 98.912 -76.919 47.297
209396 98.944 -76.921 47.263
209397 98.977 -76.923 47.230
209398 99.027 -76.921 47.197
209399 99.068 -76.923 47.161
209400 99.085 -76.937 47.127
209401 99.101 -76.950 47.093
209402 99.119 -76.964 47.061
209403 99.135 -76.978 47.027
209404 99.152 -76.990 46.994
209405 99.169 -77.004 46.960
209406 99.192 -77.008 46.924
209407 99.222 -76.999 46.892
209408 99.254 -76.989 46.862
209409 99.260 -76.979 46.832
209410 99.277 -76.994 46.803
209411 99.294 -77.008 46.774
209412 99.312 -77.023 46.746
209413 99.330 -77.038 46.718
209414 99.347 -77.052 46.689
209415 99.365 -77.067 46.663
209416 99.428 -77.078 46.637
209417 99.477 -77.087 46.610
209418 99.489 -77.100 46.583
209419 99.495 -77.114 46.555
209420 99.501 -77.128 46.526
209421 99.508 -77.142 46.498
209422 99.511 -77.153 46.471
209423 99.498 -77.147 46.444
209424 99.484 -77.141 46.418
209425 99.470 -77.135 46.391
209426 99.474 -77.132 46.364
209427 99.484 -77.130 46.338
209428 99.494 -77.128 46.313
209429 99.504 -77.125 46.290
209430 99.514 -77.124 46.267
209431 99.524 -77.122 46.244
209432 99.558 -77.125 46.221
209433 99.609 -77.132 46.199
209434 99.660 -77.138 46.177
209435 99.710 -77.145 46.155
209436 99.735 -77.139 46.134
209437 99.760 -77.132 46.114
209438 99.784 -77.126 46.093
209439 99.809 -77.119 46.073
209440 99.834 -77.113 46.054
209441 99.870 -77.112 46.034
209442 99.907 -77.112 46.015
209443 99.943 -77.111 45.997
209444 99.980 -77.111 45.979
209445 100.011 -77.134 45.961
209446 100.043 -77.157 45.943
209447 100.073 -77.179 45.925
209448 100.097 -77.189 45.908
209449 100.119 -77.199 45.890
209450 100.142 -77.210 45.872
209451 100.161 -77.213 45.854
209452 100.173 -77.205 45.837
209453 100.185 -77.197 45.819
209454 100.218 -77.210 45.801
209455 100.253 -77.224 45.784
209456 100.285 -77.238 45.767
209457 100.297 -77.241 45.751
209458 100.309 -77.244 45.735
209459 100.320 -77.248 45.720
209460 100.332 -77.251 45.705
209461 100.343 -77.257 45.691
209462 100.353 -77.268 45.678
209463 100.364 -77.279 45.663
209464 100.374 -77.291 45.650
209465 100.385 -77.304 45.637
209466 100.407 -77.323 45.624
209467 100.430 -77.342 45.611
209468 100.454 -77.363 45.598
209469 100.477 -77.382 45.585
209470 100.500 -77.402 45.572
209471 100.524 -77.423 45.560
209472 100.547 -77.442 45.548
209473 100.570 -77.463 45.535
209474 100.594 -77.483 45.522
209475 100.617 -77.502 45.509
209476 100.645 -77.513 45.496
209477 100.676 -77.518 45.483
209478 100.706 -77.523 45.471
209479 100.737 -77.529 45.458
209480 100.767 -77.533 45.446
209481 100.797 -77.539 45.433
209482 100.828 -77.544 45.418
209483 100.853 -77.552 45.405
209484 100.876 -77.564 45.392
209485 100.903 -77.581 45.379
209486 100.930 -77.598 45.366
209487 100.957 -77.614 45.353
209488 100.983 -77.631 45.340
209489 101.008 -77.650 45.327
209490 101.032 -77.669 45.314
209491 101.056 -77.688 45.301
209492 101.080 -77.707 45.287
209493 101.104 -77.726 45.272
209494 101.128 -77.745 45.258
209495 101.152 -77.763 45.244
209496 101.176 -77.782 45.231
209497 101.200 -77.801 45.218
209498 101.224 -77.820 45.204
209499 101.249 -77.836 45.191
209500 101.275 -77.845 45.177
209501 101.301 -77.854 45.162
209502 101.327 -77.863 45.147
209503 101.353 -77.872 45.132
209504 101.381 -77.885 45.119
209505 101.409 -77.902 45.104
209506 101.439 -77.918 45.090
209507 101.467 -77.934 45.074
209508 101.496 -77.950 45.059
209509 101.525 -77.967 45.043
209510 101.553 -77.983 45.029
209511 101.582 -78.000 45.014
209512 101.599 -78.007 44.998
209513 101.609 -78.009 44.983
209514 101.627 -78.017 44.966
209515 101.671 -78.040 44.949
209516 101.696 -78.054 44.932
209517 101.707 -78.061 44.917
209518 101.718 -78.068 44.901
209519 101.729 -78.074 44.885
209520 101.740 -78.081 44.870
209521 101.751 -78.087 44.855
209522 101.761 -78.094 44.840
209523 101.779 -78.104 44.825
209524 101.799 -78.116 44.810
209525 101.820 -78.127 44.795
209526 101.840 -78.138 44.780
209527 101.861 -78.150 44.764
209528 101.881 -78.162 44.748
209529 101.901 -78.173 44.731
209530 101.935 -78.184 44.715
209531 101.970 -78.194 44.700
209532 102.004 -78.205 44.684
209533 102.039 -78.216 44.669
209534 102.074 -78.226 44.653
209535 102.109 -78.237 44.637
209536 102.144 -78.247 44.622
209537 102.179 -78.257 44.607
209538 102.214 -78.268 44.592
209539 102.254 -78.290 44.577
209540 102.296 -78.311 44.562
209541 102.336 -78.332 44.547
209542 102.378 -78.354 44.531
209543 102.418 -78.375 44.514
209544 102.460 -78.396 44.499
209545 102.501 -78.417 44.485
209546 102.542 -78.438 44.471
209547 102.581 -78.457 44.457
209548 102.611 -78.463 44.443
209549 102.642 -78.469 44.428
209550 102.673 -78.475 44.414
209551 102.703 -78.480 44.399
209552 102.722 -78.470 44.385
209553 102.733 -78.478 44.369
209554 102.770 -78.492 44.353
209555 102.807 -78.505 44.337
209556 102.844 -78.519 44.320
209557 102.881 -78.532 44.304
209558 102.918 -78.546 44.288
209559 102.945 -78.550 44.272
209560 102.971 -78.553 44.255
209561 102.997 -78.555 44.239
209562 103.022 -78.557 44.224
209563 103.048 -78.561 44.207
209564 103.074 -78.563 44.189
209848 93.417 -75.803 53.944
209849 93.435 -75.806 53.962
209850 93.453 -75.810 53.983
209851 93.472 -75.813 54.005
209852 93.490 -75.816 54.026
209853 93.508 -75.820 54.048
209854 93.526 -75.823 54.071
209855 93.545 -75.826 54.095
209856 93.563 -75.830 54.120
209857 93.581 -75.833 54.144
209858 93.595 -75.850 54.167
209859 93.628 -75.844 54.188
209860 93.660 -75.838 54.211
209861 93.693 -75.832 54.235
209862 93.725 -75.825 54.257
209863 93.758 -75.819 54.275
209864 93.790 -75.813 54.287
209865 93.823 -75.807 54.293
209866 93.852 -75.801 54.293
209867 93.873 -75.793 54.290
209868 93.894 -75.786 54.286
209869 93.915 -75.778 54.284
209870 93.944 -75.773 54.285
209871 93.976 -75.768 54.292
209872 94.002 -75.763 54.292
209873 94.028 -75.758 54.289
209874 94.054 -75.753 54.286
209875 94.080 -75.747 54.281
209876 94.106 -75.742 54.273
209877 94.132 -75.737 54.263
209878 94.159 -75.733 54.254
209879 94.185 -75.732 54.240
209880 94.210 -75.732 54.229
209881 94.235 -75.732 54.218
209882 94.260 -75.733 54.208
209883 94.285 -75.733 54.199
209884 94.310 -75.733 54.190
209885 94.330 -75.734 54.178
209886 94.350 -75.736 54.158
209887 94.370 -75.738 54.136
209888 94.390 -75.740 54.113
209889 94.409 -75.742 54.092
209890 94.429 -75.743 54.074
209891 94.448 -75.746 54.053
209892 94.461 -75.752 54.025
209893 94.473 -75.757 53.992
209894 94.492 -75.764 53.959
209895 94.511 -75.771 53.926
209896 94.530 -75.777 53.896
209897 94.549 -75.784 53.870
209898 94.566 -75.791 53.847
209899 94.581 -75.800 53.819
209900 94.596 -75.808 53.789
209901 94.611 -75.817 53.757
209902 94.626 -75.825 53.723
209903 94.640 -75.834 53.686
209904 94.655 -75.843 53.648
209905 94.670 -75.851 53.609
209906 94.685 -75.860 53.566
209907 94.698 -75.866 53.523
209908 94.707 -75.869 53.481
209909 94.716 -75.872 53.441
209910 94.725 -75.875 53.402
209911 94.735 -75.878 53.365
209912 94.744 -75.881 53.331
209913 94.754 -75.884 53.292
209914 94.763 -75.886 53.252
209915 94.772 -75.889 53.212
209916 94.781 -75.892 53.172
209917 94.791 -75.895 53.129
209918 94.800 -75.897 53.088
209919 94.810 -75.900 53.048
209920 94.819 -75.903 53.007
209921 94.828 -75.906 52.970
209922 94.837 -75.909 52.934
209923 94.847 -75.912 52.899
209924 94.856 -75.915 52.866
209925 94.865 -75.918 52.835
209926 94.875 -75.920 52.806
209927 94.884 -75.923 52.779
209928 94.895 -75.926 52.757
209929 94.906 -75.930 52.737
209930 94.917 -75.934 52.716
209931 94.928 -75.938 52.694
209932 94.939 -75.941 52.670
209933 94.975 -75.942 52.642
209934 95.037 -75.941 52.612
209935 95.100 -75.941 52.582
209936 95.163 -75.940 52.551
209937 95.226 -75.939 52.517
209938 95.288 -75.938 52.482
209939 95.351 -75.938 52.443
209940 95.412 -75.938 52.405
209941 95.469 -75.939 52.365
209942 95.527 -75.941 52.323
209943 95.556 -75.941 52.279
209944 95.566 -75.952 52.229
209945 95.576 -75.958 52.176
209946 95.586 -75.965 52.123
209947 95.595 -75.972 52.070
209948 95.605 -75.978 52.020
209949 95.615 -75.985 51.975
209950 95.624 -75.992 51.931
209951 95.634 -75.999 51.887
209952 95.644 -76.006 51.844
209953 95.654 -76.013 51.802
209954 95.664 -76.019 51.760
209955 95.673 -76.026 51.719
209956 95.683 -76.032 51.679
209957 95.695 -76.039 51.641
209958 95.717 -76.044 51.608
209959 95.739 -76.047 51.576
209960 95.762 -76.052 51.545
209961 95.784 -76.057 51.513
209962 95.806 -76.061 51.477
209963 95.829 -76.066 51.441
209964 95.869 -76.065 51.404
209965 95.909 -76.065 51.365
209966 95.950 -76.065 51.326
209967 95.990 -76.065 51.284
209968 96.031 -76.064 51.241
209969 96.056 -76.064 51.197
209970 96.080 -76.063 51.152
209971 96.103 -76.063 51.108
209972 96.126 -76.062 51.065
209973 96.150 -76.061 51.024
209974 96.173 -76.060 50.984
209975 96.196 -76.060 50.947
209976 96.215 -76.062 50.912
209977 96.232 -76.064 50.878
209978 96.241 -76.072 50.845
209979 96.239 -76.088 50.814
209980 96.238 -76.104 50.784
209981 96.254 -76.116 50.756
209982 96.275 -76.127 50.729
209983 96.298 -76.138 50.701
209984 96.320 -76.150 50.673
209985 96.341 -76.160 50.646
209986 96.354 -76.169 50.620
209987 96.360 -76.176 50.596
209988 96.366 -76.183 50.574
209989 96.372 -76.190 50.553
209990 96.378 -76.197 50.532
209991 96.384 -76.204 50.510
209992 96.393 -76.212 50.488
209993 96.436 -76.226 50.464
209994 96.478 -76.239 50.440
209995 96.522 -76.253 50.415
209996 96.565 -76.267 50.390
209997 96.607 -76.281 50.363
209998 96.645 -76.286 50.335
209999 96.674 -76.285 50.306
210000 96.703 -76.284 50.275
210001 96.731 -76.282 50.243
210002 96.746 -76.282 50.210
210003 96.761 -76.282 50.176
210004 96.777 -76.282 50.140
210005 96.792 -76.281 50.102
210006 96.808 -76.282 50.064
210007 96.826 -76.297 50.025
210008 96.845 -76.310 49.986
210009 96.867 -76.320 49.947
210010 96.889 -76.328 49.910
210011 96.912 -76.336 49.873
210012 96.936 -76.343 49.836
210013 96.958 -76.351 49.798
210014 96.981 -76.358 49.760
210015 97.005 -76.366 49.722
210016 97.040 -76.383 49.685
210017 97.088 -76.406 49.649
210018 97.135 -76.430 49.615
210019 97.178 -76.455 49.583
210020 97.193 -76.483 49.554
210021 97.207 -76.511 49.528
210022 97.222 -76.539 49.504
210023 97.236 -76.568 49.481
210024 97.250 -76.597 49.455
210025 97.146 -76.585 49.429
210026 97.186 -76.603 49.401
210027 97.226 -76.621 49.371
210028 97.266 -76.639 49.339
210029 97.303 -76.643 49.306
210030 97.335 -76.629 49.272
210031 97.367 -76.616 49.238
210032 97.399 -76.602 49.204
210033 97.432 -76.589 49.171
210034 97.464 -76.576 49.140
210035 97.480 -76.584 49.110
210036 97.490 -76.596 49.080
210037 97.494 -76.609 49.050
210038 97.514 -76.621 49.022
210039 97.538 -76.633 48.996
210040 97.562 -76.645 48.971
210041 97.585 -76.657 48.946
210042 97.609 -76.669 48.920
210043 97.633 -76.680 48.894
210044 97.657 -76.693 48.868
210045 97.680 -76.704 48.840
210046 97.705 -76.716 48.810
210047 97.728 -76.728 48.780
210048 97.752 -76.740 48.748
210049 97.776 -76.752 48.716
210050 97.800 -76.764 48.682
210051 97.868 -76.768 48.649
210052 97.934 -76.769 48.616
210053 97.972 -76.758 48.584
210054 98.015 -76.757 48.551
210055 98.061 -76.762 48.517
210056 98.106 -76.767 48.485
210057 98.132 -76.779 48.453
210058 98.154 -76.792 48.422
210059 98.175 -76.805 48.391
210060 98.197 -76.818 48.361
210061 98.219 -76.831 48.332
210062 98.241 -76.843 48.303
210063 98.263 -76.855 48.274
210064 98.281 -76.848 48.245
210065 98.299 -76.840 48.215
210066 98.319 -76.836 48.187
210067 98.341 -76.837 48.159
210068 98.363 -76.837 48.131
210069 98.384 -76.838 48.102
210070 98.406 -76.840 48.073
210071 98.428 -76.840 48.043
210072 98.450 -76.841 48.013
210073 98.472 -76.842 47.984
210074 98.496 -76.831 47.954
210075 98.520 -76.821 47.925
210076 98.544 -76.806 47.897
210077 98.562 -76.803 47.869
210078 98.576 -76.811 47.841
210079 98.589 -76.817 47.814
210080 98.602 -76.822 47.786
210081 98.615 -76.828 47.758
210082 98.628 -76.834 47.730
210083 98.640 -76.839 47.703
210084 98.653 -76.845 47.674
210085 98.666 -76.850 47.648
210086 98.678 -76.856 47.619
210087 98.694 -76.867 47.588
210088 98.712 -76.883 47.557
210089 98.733 -76.890 47.524
210090 98.757 -76.885 47.492
210091 98.770 -76.895 47.460
210092 98.803 -76.897 47.428
210093 98.836 -76.899 47.397
210094 98.868 -76.901 47.365
210095 98.901 -76.903 47.333
210096 98.934 -76.905 47.300
210097 98.966 -76.906 47.266
210098 99.007 -76.906 47.232
210099 99.060 -76.903 47.199
210100 99.102 -76.906 47.164
210101 99.120 -76.919 47.129
210102 99.136 -76.932 47.096
210103 99.153 -76.946 47.063
210104 99.170 -76.959 47.029
210105 99.187 -76.973 46.994
210106 99.216 -76.965 46.959
210107 99.247 -76.956 46.924
210108 99.248 -76.947 46.889
210109 99.261 -76.958 46.858
210110 99.279 -76.973 46.828
210111 99.296 -76.987 46.798
210112 99.314 -77.003 46.769
210113 99.332 -77.017 46.740
210114 99.350 -77.031 46.712
210115 99.367 -77.047 46.683
210116 99.385 -77.061 46.656
210117 99.422 -77.075 46.629
210118 99.474 -77.084 46.603
210119 99.509 -77.091 46.575
210120 99.526 -77.103 46.547
210121 99.533 -77.117 46.518
210122 99.540 -77.131 46.489
210123 99.532 -77.131 46.462
210124 99.517 -77.125 46.435
210125 99.519 -77.121 46.408
210126 99.529 -77.119 46.381
210127 99.539 -77.117 46.355
210128 99.549 -77.115 46.329
210129 99.559 -77.113 46.304
210130 99.569 -77.111 46.281
210131 99.579 -77.109 46.258
210132 99.589 -77.106 46.235
210133 99.598 -77.105 46.213
210134 99.644 -77.110 46.191
210135 99.696 -77.117 46.169
210136 99.726 -77.114 46.147
210137 99.751 -77.107 46.126
210138 99.776 -77.101 46.106
210139 99.800 -77.094 46.085
210140 99.825 -77.088 46.066
210141 99.851 -77.082 46.046
210142 99.888 -77.082 46.027
210143 99.925 -77.081 46.008
210144 99.962 -77.081 45.990
210145 99.999 -77.081 45.972
210146 100.031 -77.099 45.953
210147 100.062 -77.122 45.935
210148 100.093 -77.144 45.918
210149 100.120 -77.160 45.901
210150 100.143 -77.171 45.883
210151 100.165 -77.181 45.866
210152 100.189 -77.191 45.848
210153 100.207 -77.194 45.830
210154 100.218 -77.185 45.812
210155 100.247 -77.193 45.794
210156 100.280 -77.207 45.778
210157 100.301 -77.214 45.761
210158 100.313 -77.218 45.744
210159 100.324 -77.222 45.728
210160 100.336 -77.226 45.713
210161 100.346 -77.238 45.698
210162 100.357 -77.250 45.684
210163 100.367 -77.262 45.670
210164 100.378 -77.274 45.657
210165 100.388 -77.285 45.643
210166 100.400 -77.298 45.629
210167 100.421 -77.316 45.616
210168 100.443 -77.334 45.604
210169 100.466 -77.354 45.591
210170 100.490 -77.374 45.579
210171 100.513 -77.394 45.566
210172 100.536 -77.414 45.554
210173 100.560 -77.434 45.541
210174 100.583 -77.454 45.529
210175 100.607 -77.474 45.516
210176 100.630 -77.494 45.503
210177 100.656 -77.508 45.491
210178 100.687 -77.512 45.478
210179 100.717 -77.517 45.467
210180 100.748 -77.523 45.454
210181 100.778 -77.528 45.441
210182 100.809 -77.533 45.428
210183 100.839 -77.538 45.415
210184 100.869 -77.543 45.401
210185 100.897 -77.553 45.388
210186 100.921 -77.572 45.375
210187 100.946 -77.591 45.361
210188 100.970 -77.610 45.349
210189 100.994 -77.629 45.336
210190 101.018 -77.647 45.323
210191 101.042 -77.666 45.310
210192 101.066 -77.685 45.297
210193 101.090 -77.704 45.283
210194 101.114 -77.723 45.268
210195 101.138 -77.742 45.254
210196 101.162 -77.761 45.240
210197 101.186 -77.780 45.225
210198 101.210 -77.797 45.212
210199 101.237 -77.806 45.198
210200 101.262 -77.815 45.185
210201 101.288 -77.824 45.170
210202 101.315 -77.834 45.155
210203 101.341 -77.843 45.141
210204 101.367 -77.853 45.126
210205 101.396 -77.869 45.112
210206 101.424 -77.885 45.097
210207 101.454 -77.902 45.082
210208 101.482 -77.918 45.066
210209 101.511 -77.934 45.049
210210 101.540 -77.950 45.034
210211 101.568 -77.967 45.019
210212 101.597 -77.983 45.004
210213 101.626 -78.000 44.989
210214 101.653 -78.015 44.973
210215 101.663 -78.017 44.957
210216 101.701 -78.037 44.940
210217 101.714 -78.044 44.923
210218 101.725 -78.051 44.908
210219 101.736 -78.058 44.893
210220 101.747 -78.064 44.877
210221 101.758 -78.071 44.861
210222 101.768 -78.078 44.845
210223 101.779 -78.084 44.830
210224 101.792 -78.091 44.816
210225 101.812 -78.103 44.801
210226 101.832 -78.114 44.786
210227 101.852 -78.125 44.771
210228 101.873 -78.137 44.756
210229 101.893 -78.149 44.739
210230 101.924 -78.159 44.723
210231 101.959 -78.170 44.707
210232 101.994 -78.181 44.691
210233 102.029 -78.191 44.675
210234 102.063 -78.202 44.659
210235 102.098 -78.213 44.644
210236 102.133 -78.222 44.628
210237 102.168 -78.233 44.612
210238 102.203 -78.244 44.597
210239 102.239 -78.256 44.583
210240 102.279 -78.278 44.569
210241 102.321 -78.299 44.553
210242 102.362 -78.320 44.538
210243 102.403 -78.341 44.522
210244 102.444 -78.363 44.506
210245 102.485 -78.384 44.491
210246 102.526 -78.405 44.477
210247 102.559 -78.415 44.463
210248 102.590 -78.421 44.449
210249 102.620 -78.426 44.435
210250 102.650 -78.432 44.421
210251 102.681 -78.438 44.406
210252 102.712 -78.445 44.392
210253 102.734 -78.435 44.376
210254 102.750 -78.443 44.360
210255 102.787 -78.457 44.344
210256 102.824 -78.470 44.328
210257 102.861 -78.484 44.312
210258 102.898 -78.498 44.295
210259 102.935 -78.511 44.279
210260 102.972 -78.525 44.263
210261 103.007 -78.538 44.247
210262 103.033 -78.540 44.230
210263 103.058 -78.542 44.214
210264 103.084 -78.545 44.197
210265 103.110 -78.548 44.180
210548 93.390 -75.777 53.925
210549 93.409 -75.780 53.941
210550 93.427 -75.784 53.959
210551 93.445 -75.787 53.980
210552 93.464 -75.790 54.003
210553 93.482 -75.793 54.024
210554 93.500 -75.796 54.047
210555 93.518 -75.800 54.071
210556 93.537 -75.803 54.095
210557 93.555 -75.806 54.121
210558 93.573 -75.809 54.147
210559 93.590 -75.822 54.172
210560 93.623 -75.816 54.194
210561 93.655 -75.810 54.221
210562 93.688 -75.804 54.247
210563 93.720 -75.798 54.271
210564 93.753 -75.792 54.291
210565 93.785 -75.786 54.308
210566 93.818 -75.780 54.316
210567 93.850 -75.774 54.315
210568 93.876 -75.767 54.309
210569 93.897 -75.759 54.302
210570 93.918 -75.752 54.295
210571 93.945 -75.746 54.291
210572 93.977 -75.741 54.293
210573 94.003 -75.736 54.290
210574 94.029 -75.731 54.285
210575 94.055 -75.725 54.280
210576 94.081 -75.720 54.273
210577 94.107 -75.715 54.265
210578 94.133 -75.709 54.256
210579 94.157 -75.710 54.247
210580 94.182 -75.710 54.235
210581 94.207 -75.710 54.224
210582 94.232 -75.710 54.214
210583 94.257 -75.710 54.203
210584 94.282 -75.710 54.194
210585 94.308 -75.711 54.184
210586 94.330 -75.712 54.173
210587 94.349 -75.713 54.154
210588 94.369 -75.715 54.133
210589 94.389 -75.717 54.111
210590 94.408 -75.718 54.088
210591 94.428 -75.721 54.071
210592 94.442 -75.726 54.051
210593 94.455 -75.731 54.026
210594 94.467 -75.737 53.996
210595 94.481 -75.743 53.966
210596 94.500 -75.750 53.937
210597 94.520 -75.756 53.909
210598 94.539 -75.763 53.884
210599 94.558 -75.770 53.861
210600 94.576 -75.776 53.834
210601 94.596 -75.783 53.804
210602 94.615 -75.790 53.770
210603 94.630 -75.798 53.735
210604 94.645 -75.806 53.698
210605 94.660 -75.815 53.658
210606 94.675 -75.824 53.620
210607 94.690 -75.832 53.578
210608 94.702 -75.839 53.535
210609 94.712 -75.842 53.492
210610 94.721 -75.844 53.452
210611 94.731 -75.847 53.414
210612 94.740 -75.850 53.377
210613 94.749 -75.853 53.342
210614 94.758 -75.856 53.303
210615 94.768 -75.859 53.263
210616 94.777 -75.862 53.222
210617 94.786 -75.864 53.179
210618 94.795 -75.867 53.136
210619 94.805 -75.870 53.094
210620 94.814 -75.873 53.052
210621 94.824 -75.875 53.009
210622 94.833 -75.878 52.970
210623 94.842 -75.881 52.934
210624 94.851 -75.884 52.900
210625 94.861 -75.887 52.867
210626 94.870 -75.890 52.837
210627 94.880 -75.893 52.807
210628 94.889 -75.896 52.780
210629 94.898 -75.898 52.756
210630 94.908 -75.901 52.735
210631 94.917 -75.904 52.714
210632 94.928 -75.908 52.692
210633 94.939 -75.911 52.670
210634 94.973 -75.912 52.642
210635 95.036 -75.912 52.612
210636 95.099 -75.911 52.581
210637 95.162 -75.910 52.550
210638 95.219 -75.912 52.517
210639 95.276 -75.913 52.482
210640 95.334 -75.915 52.441
210641 95.391 -75.916 52.402
210642 95.448 -75.918 52.361
210643 95.506 -75.919 52.317
210644 95.547 -75.910 52.271
210645 95.558 -75.916 52.220
210646 95.569 -75.932 52.169
210647 95.580 -75.941 52.117
210648 95.589 -75.948 52.065
210649 95.599 -75.955 52.016
210650 95.609 -75.962 51.970
210651 95.618 -75.969 51.927
210652 95.628 -75.975 51.884
210653 95.638 -75.982 51.842
210654 95.648 -75.989 51.801
210655 95.657 -75.996 51.760
210656 95.667 -76.003 51.719
210657 95.677 -76.009 51.680
210658 95.686 -76.016 51.642
210659 95.708 -76.020 51.607
210660 95.730 -76.025 51.575
210661 95.753 -76.029 51.546
210662 95.775 -76.034 51.517
210663 95.797 -76.038 51.483
210664 95.829 -76.040 51.448
210665 95.869 -76.040 51.412
210666 95.909 -76.040 51.375
210667 95.950 -76.040 51.335
210668 95.990 -76.039 51.294
210669 96.030 -76.039 51.252
210670 96.063 -76.038 51.209
210671 96.086 -76.038 51.164
210672 96.109 -76.038 51.120
210673 96.133 -76.037 51.077
210674 96.152 -76.038 51.036
210675 96.170 -76.040 50.997
210676 96.187 -76.042 50.960
210677 96.204 -76.044 50.924
210678 96.222 -76.047 50.891
210679 96.239 -76.048 50.858
210680 96.238 -76.064 50.827
210681 96.236 -76.080 50.797
210682 96.257 -76.091 50.769
210683 96.278 -76.103 50.743
210684 96.301 -76.113 50.715
210685 96.323 -76.125 50.687
210686 96.344 -76.136 50.660
210687 96.367 -76.147 50.634
210688 96.381 -76.157 50.609
210689 96.387 -76.163 50.587
210690 96.393 -76.170 50.565
210691 96.399 -76.177 50.543
210692 96.405 -76.184 50.522
210693 96.407 -76.192 50.499
210694 96.430 -76.204 50.477
210695 96.472 -76.218 50.454
210696 96.515 -76.232 50.430
210697 96.558 -76.246 50.405
210698 96.601 -76.260 50.378
210699 96.640 -76.266 50.351
210700 96.669 -76.264 50.322
210701 96.698 -76.263 50.291
210702 96.727 -76.262 50.258
210703 96.755 -76.261 50.225
210704 96.770 -76.260 50.191
210705 96.785 -76.260 50.155
210706 96.800 -76.260 50.116
210707 96.816 -76.262 50.077
210708 96.835 -76.276 50.038
210709 96.854 -76.290 49.998
210710 96.872 -76.304 49.959
210711 96.892 -76.317 49.923
210712 96.914 -76.325 49.887
210713 96.938 -76.332 49.850
210714 96.961 -76.340 49.813
210715 96.983 -76.348 49.776
210716 97.007 -76.355 49.738
210717 97.039 -76.369 49.700
210718 97.086 -76.393 49.665
210719 97.133 -76.417 49.631
210720 97.180 -76.440 49.599
210721 97.226 -76.464 49.570
210722 97.240 -76.493 49.543
210723 97.255 -76.521 49.518
210724 97.239 -76.540 49.494
210725 97.129 -76.527 49.467
210726 97.169 -76.545 49.439
210727 97.209 -76.563 49.410
210728 97.249 -76.581 49.379
210729 97.289 -76.599 49.346
210730 97.326 -76.605 49.312
210731 97.359 -76.591 49.278
210732 97.391 -76.578 49.244
210733 97.423 -76.564 49.212
210734 97.455 -76.552 49.180
210735 97.480 -76.549 49.149
210736 97.497 -76.558 49.120
210737 97.505 -76.569 49.090
210738 97.509 -76.583 49.061
210739 97.520 -76.596 49.033
210740 97.543 -76.607 49.008
210741 97.567 -76.619 48.983
210742 97.591 -76.631 48.958
210743 97.615 -76.643 48.933
210744 97.639 -76.655 48.907
210745 97.662 -76.667 48.880
210746 97.686 -76.678 48.852
210747 97.710 -76.690 48.823
210748 97.734 -76.702 48.792
210749 97.757 -76.714 48.761
210750 97.782 -76.726 48.729
210751 97.823 -76.734 48.695
210752 97.892 -76.738 48.661
210753 97.961 -76.742 48.628
210754 98.010 -76.736 48.595
210755 98.048 -76.724 48.562
210756 98.090 -76.722 48.528
210757 98.125 -76.730 48.495
210758 98.146 -76.743 48.463
210759 98.168 -76.756 48.431
210760 98.190 -76.769 48.400
210761 98.212 -76.783 48.370
210762 98.233 -76.796 48.341
210763 98.255 -76.809 48.312
210764 98.277 -76.821 48.283
210765 98.298 -76.833 48.254
210766 98.322 -76.827 48.225
210767 98.344 -76.824 48.197
210768 98.365 -76.821 48.169
210769 98.387 -76.819 48.141
210770 98.409 -76.817 48.113
210771 98.431 -76.814 48.083
210772 98.452 -76.815 48.054
210773 98.474 -76.821 48.025
210774 98.495 -76.826 47.996
210775 98.518 -76.815 47.966
210776 98.541 -76.800 47.937
210777 98.565 -76.784 47.909
210778 98.583 -76.783 47.880
210779 98.595 -76.795 47.853
210780 98.609 -76.802 47.826
210781 98.624 -76.809 47.798
210782 98.638 -76.816 47.770
210783 98.652 -76.823 47.742
210784 98.666 -76.830 47.713
210785 98.680 -76.837 47.685
210786 98.694 -76.843 47.658
210787 98.708 -76.851 47.629
210788 98.722 -76.857 47.598
210789 98.735 -76.862 47.565
210790 98.747 -76.868 47.532
210791 98.767 -76.878 47.500
210792 98.792 -76.881 47.467
210793 98.825 -76.883 47.435
210794 98.858 -76.885 47.402
210795 98.890 -76.887 47.368
210796 98.923 -76.889 47.335
210797 98.955 -76.890 47.302
210798 98.988 -76.893 47.267
210799 99.040 -76.890 47.233
210800 99.093 -76.887 47.201
210801 99.137 -76.887 47.166
210802 99.154 -76.901 47.132
210803 99.171 -76.915 47.098
210804 99.188 -76.928 47.065
210805 99.210 -76.933 47.030
210806 99.241 -76.923 46.995
210807 99.235 -76.913 46.959
210808 99.246 -76.923 46.923
210809 99.264 -76.937 46.888
210810 99.281 -76.953 46.856
210811 99.299 -76.967 46.826
210812 99.316 -76.982 46.796
210813 99.334 -76.997 46.766
210814 99.352 -77.011 46.737
210815 99.370 -77.026 46.708
210816 99.387 -77.041 46.678
210817 99.405 -77.056 46.650
210818 99.423 -77.070 46.623
210819 99.472 -77.081 46.596
210820 99.507 -77.088 46.567
210821 99.542 -77.096 46.539
210822 99.565 -77.106 46.510
210823 99.565 -77.114 46.481
210824 99.563 -77.110 46.453
210825 99.573 -77.108 46.426
210826 99.583 -77.106 46.398
210827 99.593 -77.104 46.372
210828 99.603 -77.102 46.345
210829 99.613 -77.100 46.319
210830 99.623 -77.098 46.294
210831 99.633 -77.096 46.272
210832 99.643 -77.094 46.249
210833 99.653 -77.092 46.226
210834 99.663 -77.090 46.205
210835 99.680 -77.089 46.182
210836 99.717 -77.089 46.160
210837 99.742 -77.083 46.138
210838 99.767 -77.076 46.117
210839 99.791 -77.070 46.097
210840 99.816 -77.063 46.078
210841 99.841 -77.057 46.058
210842 99.869 -77.053 46.038
210843 99.906 -77.052 46.019
210844 99.943 -77.052 46.000
210845 99.980 -77.051 45.982
210846 100.017 -77.051 45.965
210847 100.051 -77.064 45.946
210848 100.082 -77.087 45.928
210849 100.113 -77.110 45.910
210850 100.144 -77.132 45.892
210851 100.167 -77.143 45.875
210852 100.190 -77.153 45.858
210853 100.212 -77.163 45.841
210854 100.235 -77.173 45.823
210855 100.252 -77.174 45.806
210856 100.275 -77.176 45.788
210857 100.304 -77.188 45.771
210858 100.316 -77.191 45.754
210859 100.328 -77.195 45.737
210860 100.338 -77.207 45.721
210861 100.349 -77.219 45.706
210862 100.360 -77.232 45.691
210863 100.370 -77.244 45.677
210864 100.381 -77.256 45.664
210865 100.391 -77.268 45.650
210866 100.402 -77.279 45.636
210867 100.416 -77.294 45.622
210868 100.437 -77.311 45.609
210869 100.458 -77.328 45.597
210870 100.479 -77.345 45.584
210871 100.502 -77.365 45.572
210872 100.525 -77.385 45.560
210873 100.549 -77.405 45.547
210874 100.572 -77.425 45.534
210875 100.596 -77.445 45.522
210876 100.619 -77.465 45.510
210877 100.642 -77.486 45.497
210878 100.668 -77.501 45.485
210879 100.698 -77.507 45.473
210880 100.728 -77.511 45.462
210881 100.759 -77.517 45.450
210882 100.789 -77.522 45.437
210883 100.820 -77.527 45.423
210884 100.850 -77.532 45.410
210885 100.878 -77.536 45.397
210886 100.904 -77.546 45.384
210887 100.931 -77.569 45.371
210888 100.956 -77.588 45.357
210889 100.980 -77.607 45.345
210890 101.003 -77.626 45.332
210891 101.028 -77.645 45.318
210892 101.052 -77.664 45.305
210893 101.076 -77.683 45.292
210894 101.100 -77.702 45.278
210895 101.124 -77.721 45.264
210896 101.148 -77.740 45.250
210897 101.172 -77.758 45.235
210898 101.198 -77.768 45.221
210899 101.224 -77.777 45.207
210900 101.250 -77.786 45.192
210901 101.276 -77.795 45.178
210902 101.302 -77.804 45.164
210903 101.328 -77.813 45.149
210904 101.354 -77.822 45.134
210905 101.382 -77.837 45.119
210906 101.411 -77.853 45.104
210907 101.440 -77.869 45.089
210908 101.468 -77.885 45.073
210909 101.497 -77.902 45.057
210910 101.526 -77.918 45.040
210911 101.555 -77.934 45.025
210912 101.584 -77.950 45.010
210913 101.612 -77.967 44.995
210914 101.641 -77.983 44.979
210915 101.670 -78.000 44.963
210916 101.699 -78.015 44.946
210917 101.721 -78.028 44.930
210918 101.732 -78.034 44.914
210919 101.743 -78.041 44.900
210920 101.754 -78.047 44.885
210921 101.764 -78.054 44.869
210922 101.775 -78.061 44.853
210923 101.786 -78.067 44.836
210924 101.797 -78.074 44.820
210925 101.808 -78.081 44.806
210926 101.824 -78.090 44.792
210927 101.844 -78.101 44.777
210928 101.865 -78.112 44.761
210929 101.885 -78.124 44.746
210930 101.913 -78.135 44.730
210931 101.949 -78.146 44.714
210932 101.983 -78.156 44.697
210933 102.018 -78.167 44.681
210934 102.053 -78.178 44.666
210935 102.088 -78.188 44.650
210936 102.122 -78.198 44.635
210937 102.157 -78.209 44.619
210938 102.192 -78.219 44.603
210939 102.227 -78.230 44.588
210940 102.263 -78.244 44.574
210941 102.305 -78.266 44.560
210942 102.345 -78.287 44.544
210943 102.387 -78.308 44.528
210944 102.427 -78.329 44.513
210945 102.469 -78.351 44.498
210946 102.507 -78.366 44.483
210947 102.537 -78.373 44.469
210948 102.568 -78.379 44.455
210949 102.598 -78.385 44.441
210950 102.629 -78.391 44.427
210951 102.660 -78.396 44.412
210952 102.690 -78.402 44.398
210953 102.721 -78.408 44.383
210954 102.747 -78.401 44.367
210955 102.767 -78.409 44.351
210956 102.804 -78.423 44.335
210957 102.841 -78.436 44.319
210958 102.878 -78.450 44.303
210959 102.915 -78.463 44.287
210960 102.952 -78.477 44.271
210961 102.988 -78.491 44.254
210962 103.026 -78.504 44.237
210963 103.063 -78.518 44.220
210964 103.095 -78.528 44.204
210965 103.121 -78.530 44.187
210966 103.147 -78.532 44.170
211249 93.382 -75.753 53.922
211250 93.401 -75.757 53.938
211251 93.419 -75.760 53.957
211252 93.437 -75.763 53.978
211253 93.455 -75.767 54.001
211254 93.474 -75.770 54.022
211255 93.492 -75.773 54.046
211256 93.510 -75.777 54.070
211257 93.529 -75.780 54.095
211258 93.547 -75.783 54.121
211259 93.565 -75.787 54.149
211260 93.585 -75.795 54.175
211261 93.618 -75.789 54.199
211262 93.650 -75.783 54.226
211263 93.683 -75.777 54.253
211264 93.715 -75.771 54.280
211265 93.748 -75.764 54.305
211266 93.780 -75.758 54.324
211267 93.812 -75.752 54.336
211268 93.845 -75.746 54.340
211269 93.877 -75.740 54.336
211270 93.901 -75.732 54.327
211271 93.922 -75.725 54.317
211272 93.946 -75.718 54.308
211273 93.977 -75.714 54.306
211274 94.003 -75.709 54.299
211275 94.030 -75.703 54.290
211276 94.055 -75.698 54.279
211277 94.081 -75.693 54.270
211278 94.107 -75.687 54.260
211279 94.131 -75.686 54.251
211280 94.155 -75.685 54.244
211281 94.180 -75.687 54.234
211282 94.204 -75.688 54.222
211283 94.229 -75.688 54.211
211284 94.254 -75.688 54.200
211285 94.279 -75.688 54.191
211286 94.305 -75.688 54.181
211287 94.329 -75.689 54.169
211288 94.348 -75.690 54.151
211289 94.368 -75.693 54.131
211290 94.388 -75.694 54.108
211291 94.408 -75.696 54.086
211292 94.423 -75.700 54.068
211293 94.436 -75.706 54.049
211294 94.449 -75.712 54.027
211295 94.461 -75.718 54.002
211296 94.473 -75.723 53.975
211297 94.490 -75.730 53.948
211298 94.509 -75.736 53.923
211299 94.528 -75.743 53.898
211300 94.547 -75.749 53.874
211301 94.566 -75.756 53.847
211302 94.585 -75.762 53.816
211303 94.604 -75.769 53.781
211304 94.623 -75.776 53.745
211305 94.642 -75.782 53.707
211306 94.662 -75.789 53.667
211307 94.680 -75.796 53.628
211308 94.694 -75.805 53.587
211309 94.707 -75.811 53.544
211310 94.716 -75.814 53.502
211311 94.726 -75.817 53.462
211312 94.735 -75.820 53.425
211313 94.745 -75.822 53.388
211314 94.754 -75.825 53.352
211315 94.763 -75.828 53.313
211316 94.772 -75.831 53.272
211317 94.782 -75.834 53.230
211318 94.791 -75.837 53.187
211319 94.801 -75.840 53.144
211320 94.810 -75.843 53.101
211321 94.819 -75.845 53.058
211322 94.829 -75.848 53.015
211323 94.838 -75.851 52.976
211324 94.847 -75.853 52.939
211325 94.856 -75.856 52.905
211326 94.866 -75.859 52.872
211327 94.875 -75.862 52.841
211328 94.885 -75.865 52.810
211329 94.894 -75.868 52.781
211330 94.903 -75.871 52.754
211331 94.912 -75.874 52.730
211332 94.922 -75.876 52.707
211333 94.931 -75.879 52.684
211334 94.940 -75.882 52.661
211335 94.969 -75.884 52.633
211336 95.026 -75.886 52.603
211337 95.083 -75.887 52.571
211338 95.140 -75.889 52.539
211339 95.198 -75.891 52.507
211340 95.255 -75.892 52.472
211341 95.312 -75.894 52.431
211342 95.370 -75.895 52.391
211343 95.426 -75.897 52.350
211344 95.484 -75.898 52.306
211345 95.537 -75.897 52.258
211346 95.549 -75.876 52.209
211347 95.560 -75.892 52.159
211348 95.571 -75.909 52.108
211349 95.583 -75.925 52.058
211350 95.593 -75.931 52.011
211351 95.602 -75.938 51.965
211352 95.612 -75.945 51.922
211353 95.622 -75.952 51.880
211354 95.631 -75.959 51.837
211355 95.641 -75.966 51.796
211356 95.651 -75.972 51.757
211357 95.661 -75.979 51.717
211358 95.671 -75.985 51.680
211359 95.680 -75.992 51.643
211360 95.699 -75.997 51.607
211361 95.721 -76.002 51.573
211362 95.744 -76.007 51.543
211363 95.766 -76.010 51.514
211364 95.788 -76.015 51.484
211365 95.828 -76.015 51.451
211366 95.869 -76.015 51.416
211367 95.909 -76.014 51.381
211368 95.950 -76.014 51.341
211369 95.990 -76.014 51.302
211370 96.030 -76.014 51.260
211371 96.070 -76.013 51.218
211372 96.090 -76.014 51.174
211373 96.108 -76.016 51.131
211374 96.125 -76.019 51.088
211375 96.142 -76.020 51.047
211376 96.160 -76.022 51.010
211377 96.177 -76.025 50.972
211378 96.194 -76.027 50.936
211379 96.212 -76.029 50.903
211380 96.229 -76.031 50.870
211381 96.236 -76.040 50.839
211382 96.238 -76.056 50.808
211383 96.260 -76.066 50.781
211384 96.282 -76.078 50.756
211385 96.304 -76.089 50.730
211386 96.326 -76.100 50.703
211387 96.347 -76.111 50.675
211388 96.370 -76.122 50.647
211389 96.392 -76.134 50.622
211390 96.408 -76.143 50.599
211391 96.414 -76.150 50.577
211392 96.420 -76.157 50.554
211393 96.426 -76.164 50.532
211394 96.426 -76.174 50.512
211395 96.427 -76.183 50.490
211396 96.466 -76.197 50.468
211397 96.509 -76.210 50.444
211398 96.552 -76.225 50.420
211399 96.595 -76.238 50.394
211400 96.635 -76.245 50.366
211401 96.664 -76.244 50.337
211402 96.693 -76.242 50.306
211403 96.722 -76.242 50.274
211404 96.751 -76.240 50.240
211405 96.779 -76.239 50.204
211406 96.794 -76.238 50.167
211407 96.809 -76.238 50.128
211408 96.825 -76.242 50.089
211409 96.843 -76.255 50.048
211410 96.863 -76.270 50.008
211411 96.881 -76.283 49.971
211412 96.900 -76.298 49.934
211413 96.919 -76.311 49.898
211414 96.940 -76.321 49.862
211415 96.963 -76.329 49.826
211416 96.986 -76.337 49.790
211417 97.009 -76.345 49.752
211418 97.038 -76.356 49.714
211419 97.085 -76.379 49.679
211420 97.132 -76.404 49.645
211421 97.178 -76.427 49.614
211422 97.225 -76.451 49.584
211423 97.273 -76.475 49.557
211424 97.225 -76.483 49.531
211425 97.111 -76.468 49.505
211426 97.152 -76.486 49.477
211427 97.191 -76.505 49.449
211428 97.232 -76.523 49.419
211429 97.272 -76.541 49.387
211430 97.312 -76.559 49.353
211431 97.350 -76.567 49.319
211432 97.382 -76.553 49.286
211433 97.415 -76.540 49.253
211434 97.447 -76.527 49.221
211435 97.479 -76.514 49.190
211436 97.496 -76.521 49.160
211437 97.512 -76.530 49.130
211438 97.519 -76.542 49.101
211439 97.524 -76.556 49.073
211440 97.528 -76.569 49.046
211441 97.549 -76.581 49.020
211442 97.573 -76.593 48.996
211443 97.597 -76.605 48.972
211444 97.620 -76.617 48.946
211445 97.644 -76.629 48.920
211446 97.668 -76.641 48.893
211447 97.692 -76.652 48.865
211448 97.716 -76.665 48.835
211449 97.739 -76.677 48.805
211450 97.763 -76.688 48.774
211451 97.787 -76.700 48.741
211452 97.847 -76.705 48.708
211453 97.916 -76.709 48.673
211454 97.985 -76.712 48.639
211455 98.047 -76.713 48.606
211456 98.085 -76.702 48.572
211457 98.124 -76.690 48.537
211458 98.136 -76.696 48.503
211459 98.157 -76.710 48.471
211460 98.177 -76.723 48.439
211461 98.199 -76.736 48.408
211462 98.220 -76.749 48.377
211463 98.241 -76.762 48.348
211464 98.262 -76.775 48.319
211465 98.283 -76.788 48.290
211466 98.308 -76.792 48.262
211467 98.346 -76.761 48.233
211468 98.367 -76.758 48.206
211469 98.389 -76.761 48.178
211470 98.410 -76.766 48.150
211471 98.432 -76.771 48.122
211472 98.453 -76.777 48.092
211473 98.475 -76.783 48.064
211474 98.496 -76.788 48.036
211475 98.517 -76.793 48.006
211476 98.539 -76.794 47.977
211477 98.563 -76.779 47.948
211478 98.586 -76.763 47.920
211479 98.604 -76.761 47.892
211480 98.617 -76.773 47.863
211481 98.629 -76.785 47.835
211482 98.642 -76.793 47.807
211483 98.657 -76.800 47.779
211484 98.671 -76.807 47.751
211485 98.685 -76.815 47.723
211486 98.699 -76.821 47.695
211487 98.713 -76.828 47.666
211488 98.727 -76.835 47.636
211489 98.741 -76.842 47.605
211490 98.756 -76.849 47.571
211491 98.770 -76.856 47.539
211492 98.784 -76.860 47.507
211493 98.806 -76.855 47.474
211494 98.844 -76.863 47.441
211495 98.880 -76.871 47.407
211496 98.913 -76.872 47.372
211497 98.945 -76.875 47.338
211498 98.978 -76.876 47.304
211499 99.020 -76.876 47.269
211500 99.073 -76.873 47.235
211501 99.126 -76.870 47.202
211502 99.172 -76.870 47.169
211503 99.189 -76.884 47.135
211504 99.206 -76.897 47.100
211505 99.235 -76.890 47.066
211506 99.223 -76.881 47.031
211507 99.230 -76.887 46.995
211508 99.248 -76.903 46.959
211509 99.266 -76.917 46.923
211510 99.284 -76.932 46.888
211511 99.301 -76.947 46.855
211512 99.319 -76.961 46.825
211513 99.337 -76.976 46.794
211514 99.354 -76.990 46.764
211515 99.372 -77.006 46.734
211516 99.390 -77.020 46.704
211517 99.408 -77.035 46.674
211518 99.425 -77.050 46.646
211519 99.443 -77.064 46.618
211520 99.470 -77.077 46.589
211521 99.505 -77.085 46.560
211522 99.543 -77.089 46.530
211523 99.582 -77.091 46.501
211524 99.594 -77.093 46.472
211525 99.604 -77.091 46.445
211526 99.613 -77.088 46.417
211527 99.622 -77.086 46.390
211528 99.631 -77.084 46.363
211529 99.640 -77.082 46.337
211530 99.649 -77.080 46.311
211531 99.658 -77.078 46.285
211532 99.667 -77.075 46.263
211533 99.677 -77.073 46.240
211534 99.686 -77.071 46.217
211535 99.695 -77.069 46.195
211536 99.704 -77.066 46.174
211537 99.732 -77.058 46.151
211538 99.758 -77.051 46.130
211539 99.782 -77.045 46.109
211540 99.807 -77.039 46.089
211541 99.832 -77.032 46.069
211542 99.856 -77.026 46.050
211543 99.888 -77.022 46.030
211544 99.925 -77.022 46.011
211545 99.962 -77.022 45.992
211546 99.998 -77.022 45.974
211547 100.035 -77.021 45.956
211548 100.070 -77.029 45.938
211549 100.101 -77.052 45.920
211550 100.132 -77.075 45.902
211551 100.163 -77.098 45.884
211552 100.191 -77.114 45.867
211553 100.213 -77.125 45.850
211554 100.236 -77.135 45.833
211555 100.259 -77.145 45.816
211556 100.282 -77.155 45.799
211557 100.303 -77.159 45.781
211558 100.320 -77.165 45.764
211559 100.331 -77.177 45.747
211560 100.341 -77.189 45.731
211561 100.352 -77.201 45.715
211562 100.363 -77.213 45.700
211563 100.373 -77.226 45.685
211564 100.384 -77.238 45.672
211565 100.395 -77.250 45.658
211566 100.405 -77.262 45.644
211567 100.415 -77.273 45.629
211568 100.432 -77.288 45.616
211569 100.453 -77.306 45.603
211570 100.474 -77.323 45.590
211571 100.495 -77.341 45.578
211572 100.516 -77.358 45.566
211573 100.538 -77.376 45.553
211574 100.561 -77.397 45.539
211575 100.584 -77.417 45.527
211576 100.608 -77.436 45.516
211577 100.631 -77.457 45.504
211578 100.655 -77.476 45.491
211579 100.679 -77.495 45.480
211580 100.709 -77.501 45.468
211581 100.740 -77.506 45.458
211582 100.770 -77.511 45.445
211583 100.800 -77.516 45.432
211584 100.831 -77.521 45.419
211585 100.860 -77.526 45.405
211586 100.882 -77.528 45.393
211587 100.907 -77.536 45.380
211588 100.941 -77.566 45.367
211589 100.966 -77.586 45.354
211590 100.989 -77.605 45.340
211591 101.013 -77.624 45.327
211592 101.038 -77.642 45.313
211593 101.062 -77.661 45.300
211594 101.086 -77.680 45.286
211595 101.110 -77.699 45.272
211596 101.134 -77.718 45.258
211597 101.159 -77.729 45.244
211598 101.186 -77.738 45.230
211599 101.211 -77.747 45.215
211600 101.238 -77.756 45.201
211601 101.264 -77.765 45.186
211602 101.289 -77.774 45.171
211603 101.316 -77.783 45.157
211604 101.342 -77.793 45.142
211605 101.369 -77.804 45.127
211606 101.397 -77.821 45.111
211607 101.426 -77.837 45.096
211608 101.455 -77.853 45.080
211609 101.483 -77.869 45.064
211610 101.513 -77.885 45.048
211611 101.541 -77.902 45.031
211612 101.570 -77.918 45.015
211613 101.599 -77.934 45.000
211614 101.627 -77.950 44.984
211615 101.656 -77.967 44.969
211616 101.683 -77.982 44.953
211617 101.707 -77.997 44.938
211618 101.754 -78.007 44.922
211619 101.760 -78.017 44.906
211620 101.766 -78.027 44.891
211621 101.772 -78.037 44.875
211622 101.782 -78.044 44.859
211623 101.793 -78.051 44.843
211624 101.804 -78.057 44.826
211625 101.815 -78.064 44.811
211626 101.826 -78.071 44.796
211627 101.836 -78.077 44.782
211628 101.856 -78.088 44.768
211629 101.877 -78.100 44.752
211630 101.903 -78.111 44.736
211631 101.938 -78.122 44.721
211632 101.972 -78.132 44.704
211633 102.007 -78.143 44.688
211634 102.042 -78.153 44.672
211635 102.077 -78.163 44.656
211636 102.112 -78.174 44.641
211637 102.147 -78.184 44.626
211638 102.181 -78.195 44.609
211639 102.216 -78.206 44.594
211640 102.251 -78.216 44.579
211641 102.289 -78.232 44.565
211642 102.329 -78.253 44.550
211643 102.371 -78.275 44.534
211644 102.411 -78.296 44.519
211645 102.453 -78.317 44.505
211646 102.484 -78.325 44.489
211647 102.515 -78.331 44.475
211648 102.546 -78.336 44.461
211649 102.576 -78.342 44.447
211650 102.607 -78.348 44.433
211651 102.638 -78.354 44.419
211652 102.668 -78.360 44.404
211653 102.699 -78.366 44.390
211654 102.726 -78.368 44.374
211655 102.760 -78.366 44.358
211656 102.784 -78.374 44.342
211657 102.821 -78.388 44.326
211658 102.858 -78.401 44.310
211659 102.895 -78.415 44.294
211660 102.932 -78.429 44.278
211661 102.969 -78.442 44.262
211662 103.006 -78.457 44.244
211663 103.042 -78.470 44.227
211664 103.080 -78.484 44.210
211665 103.116 -78.498 44.194
211666 103.153 -78.511 44.177
211667 103.183 -78.518 44.159
211950 93.374 -75.731 53.919
211951 93.392 -75.734 53.935
211952 93.411 -75.737 53.954
211953 93.429 -75.740 53.975
211954 93.447 -75.743 53.999
211955 93.466 -75.746 54.020
211956 93.484 -75.750 54.044
211957 93.502 -75.753 54.069
211958 93.520 -75.756 54.095
211959 93.539 -75.760 54.122
211960 93.557 -75.763 54.150
211961 93.580 -75.767 54.178
211962 93.613 -75.761 54.203
211963 93.645 -75.755 54.229
211964 93.678 -75.749 54.257
211965 93.710 -75.743 54.285
211966 93.742 -75.737 54.313
211967 93.775 -75.731 54.336
211968 93.807 -75.724 54.352
211969 93.840 -75.718 54.359
211970 93.872 -75.712 54.359
211971 93.904 -75.706 54.352
211972 93.925 -75.698 54.343
211973 93.946 -75.691 54.333
211974 93.978 -75.686 54.326
211975 94.004 -75.681 54.317
211976 94.030 -75.676 54.305
211977 94.056 -75.671 54.290
211978 94.082 -75.665 54.278
211979 94.106 -75.664 54.266
211980 94.130 -75.662 54.256
211981 94.154 -75.660 54.246
211982 94.178 -75.662 54.234
211983 94.202 -75.664 54.223
211984 94.226 -75.665 54.212
211985 94.252 -75.666 54.202
211986 94.276 -75.666 54.191
211987 94.301 -75.666 54.179
211988 94.327 -75.666 54.166
211989 94.348 -75.668 54.149
211990 94.367 -75.669 54.129
211991 94.387 -75.671 54.108
211992 94.405 -75.674 54.086
211993 94.417 -75.680 54.068
211994 94.430 -75.686 54.051
211995 94.442 -75.692 54.032
211996 94.455 -75.697 54.010
211997 94.467 -75.703 53.986
211998 94.480 -75.709 53.962
211999 94.499 -75.716 53.938
212000 94.518 -75.722 53.913
212001 94.537 -75.729 53.888
212002 94.556 -75.736 53.859
212003 94.575 -75.742 53.827
212004 94.594 -75.749 53.792
212005 94.613 -75.756 53.754
212006 94.632 -75.762 53.715
212007 94.651 -75.768 53.675
212008 94.670 -75.775 53.635
212009 94.690 -75.781 53.594
212010 94.705 -75.781 53.552
212011 94.710 -75.784 53.511
212012 94.719 -75.786 53.472
212013 94.728 -75.789 53.434
212014 94.737 -75.791 53.398
212015 94.746 -75.794 53.362
212016 94.755 -75.796 53.322
212017 94.764 -75.800 53.281
212018 94.772 -75.802 53.239
212019 94.781 -75.805 53.196
212020 94.790 -75.808 53.153
212021 94.799 -75.810 53.109
212022 94.808 -75.813 53.066
212023 94.817 -75.815 53.022
212024 94.826 -75.818 52.983
212025 94.834 -75.821 52.946
212026 94.843 -75.824 52.911
212027 94.852 -75.826 52.878
212028 94.861 -75.829 52.847
212029 94.870 -75.832 52.816
212030 94.879 -75.834 52.784
212031 94.888 -75.837 52.754
212032 94.896 -75.840 52.726
212033 94.905 -75.843 52.700
212034 94.914 -75.845 52.675
212035 94.936 -75.850 52.650
212036 94.965 -75.856 52.622
212037 95.004 -75.865 52.591
212038 95.062 -75.866 52.559
212039 95.119 -75.868 52.525
212040 95.176 -75.869 52.492
212041 95.234 -75.871 52.458
212042 95.291 -75.872 52.419
212043 95.348 -75.874 52.378
212044 95.405 -75.875 52.336
212045 95.463 -75.877 52.293
212046 95.520 -75.878 52.246
212047 95.540 -75.855 52.197
212048 95.551 -75.852 52.148
212049 95.563 -75.869 52.098
212050 95.574 -75.884 52.050
212051 95.585 -75.900 52.004
212052 95.596 -75.915 51.960
212053 95.606 -75.922 51.916
212054 95.615 -75.928 51.875
212055 95.625 -75.935 51.833
212056 95.635 -75.942 51.792
212057 95.644 -75.949 51.752
212058 95.655 -75.956 51.714
212059 95.664 -75.962 51.678
212060 95.674 -75.969 51.641
212061 95.690 -75.975 51.605
212062 95.712 -75.979 51.572
212063 95.735 -75.983 51.540
212064 95.757 -75.988 51.510
212065 95.788 -75.990 51.481
212066 95.828 -75.990 51.450
212067 95.869 -75.989 51.417
212068 95.909 -75.989 51.382
212069 95.949 -75.989 51.345
212070 95.990 -75.989 51.306
212071 96.033 -75.989 51.266
212072 96.071 -75.989 51.225
212073 96.097 -75.991 51.183
212074 96.109 -75.994 51.141
212075 96.121 -75.999 51.099
212076 96.133 -76.003 51.058
212077 96.149 -76.005 51.020
212078 96.167 -76.007 50.983
212079 96.184 -76.010 50.947
212080 96.201 -76.012 50.913
212081 96.219 -76.013 50.881
212082 96.235 -76.016 50.849
212083 96.241 -76.031 50.819
212084 96.262 -76.042 50.792
212085 96.285 -76.053 50.767
212086 96.307 -76.064 50.741
212087 96.328 -76.075 50.715
212088 96.350 -76.087 50.687
212089 96.373 -76.097 50.659
212090 96.395 -76.109 50.634
212091 96.416 -76.120 50.611
212092 96.435 -76.130 50.588
212093 96.441 -76.137 50.565
212094 96.444 -76.145 50.543
212095 96.445 -76.155 50.522
212096 96.445 -76.165 50.502
212097 96.460 -76.176 50.480
212098 96.503 -76.189 50.457
212099 96.546 -76.204 50.433
212100 96.589 -76.217 50.408
212101 96.629 -76.224 50.380
212102 96.659 -76.223 50.351
212103 96.688 -76.222 50.320
212104 96.717 -76.220 50.287
212105 96.746 -76.220 50.253
212106 96.776 -76.218 50.217
212107 96.802 -76.217 50.178
212108 96.817 -76.216 50.140
212109 96.834 -76.221 50.100
212110 96.853 -76.235 50.059
212111 96.871 -76.249 50.019
212112 96.890 -76.263 49.980
212113 96.908 -76.277 49.943
212114 96.927 -76.291 49.907
212115 96.946 -76.305 49.871
212116 96.965 -76.318 49.836
212117 96.988 -76.326 49.800
212118 97.011 -76.333 49.763
212119 97.036 -76.342 49.726
212120 97.083 -76.367 49.691
212121 97.130 -76.390 49.657
212122 97.177 -76.414 49.626
212123 97.224 -76.437 49.596
212124 97.271 -76.450 49.568
212125 97.177 -76.438 49.541
212126 97.134 -76.432 49.516
212127 97.174 -76.446 49.487
212128 97.215 -76.464 49.457
212129 97.255 -76.483 49.426
212130 97.295 -76.501 49.394
212131 97.335 -76.519 49.361
212132 97.374 -76.529 49.328
212133 97.406 -76.515 49.295
212134 97.438 -76.502 49.263
212135 97.470 -76.489 49.232
212136 97.495 -76.486 49.202
212137 97.512 -76.495 49.172
212138 97.528 -76.504 49.142
212139 97.534 -76.516 49.114
212140 97.538 -76.530 49.086
212141 97.543 -76.542 49.059
212142 97.555 -76.555 49.034
212143 97.578 -76.567 49.010
212144 97.602 -76.580 48.986
212145 97.626 -76.591 48.960
212146 97.650 -76.603 48.933
212147 97.673 -76.614 48.906
212148 97.698 -76.627 48.877
212149 97.721 -76.639 48.848
212150 97.745 -76.650 48.818
212151 97.769 -76.662 48.786
212152 97.801 -76.673 48.753
212153 97.870 -76.676 48.719
212154 97.939 -76.680 48.685
212155 98.008 -76.683 48.651
212156 98.077 -76.687 48.617
212157 98.123 -76.679 48.581
212158 98.155 -76.670 48.546
212159 98.158 -76.673 48.511
212160 98.163 -76.677 48.478
212161 98.184 -76.690 48.446
212162 98.206 -76.703 48.415
212163 98.227 -76.716 48.384
212164 98.248 -76.729 48.355
212165 98.269 -76.743 48.326
212166 98.290 -76.755 48.297
212167 98.318 -76.752 48.268
212168 98.355 -76.724 48.241
212169 98.383 -76.726 48.213
212170 98.411 -76.728 48.185
212171 98.433 -76.734 48.157
212172 98.454 -76.739 48.129
212173 98.476 -76.744 48.101
212174 98.497 -76.749 48.073
212175 98.518 -76.755 48.045
212176 98.539 -76.760 48.016
212177 98.561 -76.765 47.987
212178 98.584 -76.758 47.958
212179 98.607 -76.742 47.930
212180 98.626 -76.739 47.902
212181 98.638 -76.751 47.873
212182 98.650 -76.763 47.844
212183 98.662 -76.775 47.816
212184 98.675 -76.785 47.788
212185 98.690 -76.792 47.760
212186 98.704 -76.799 47.732
212187 98.718 -76.806 47.704
212188 98.732 -76.812 47.673
212189 98.746 -76.820 47.642
212190 98.761 -76.827 47.610
212191 98.775 -76.834 47.577
212192 98.789 -76.840 47.546
212193 98.802 -76.831 47.513
212194 98.814 -76.816 47.480
212195 98.844 -76.818 47.448
212196 98.881 -76.827 47.412
212197 98.919 -76.835 47.378
212198 98.956 -76.843 47.344
212199 98.994 -76.852 47.309
212200 99.044 -76.855 47.273
212201 99.106 -76.856 47.238
212202 99.158 -76.853 47.205
212203 99.207 -76.853 47.172
212204 99.226 -76.857 47.137
212205 99.211 -76.848 47.102
212206 99.215 -76.853 47.067
212207 99.233 -76.867 47.032
212208 99.251 -76.882 46.996
212209 99.268 -76.897 46.960
212210 99.286 -76.912 46.924
212211 99.304 -76.926 46.888
212212 99.322 -76.940 46.856
212213 99.339 -76.956 46.825
212214 99.357 -76.970 46.793
212215 99.374 -76.984 46.762
212216 99.391 -76.997 46.731
212217 99.408 -77.009 46.700
212218 99.425 -77.021 46.669
212219 99.442 -77.033 46.640
212220 99.460 -77.046 46.612
212221 99.488 -77.053 46.583
212222 99.527 -77.055 46.552
212223 99.567 -77.056 46.522
212224 99.589 -77.071 46.493
212225 99.598 -77.069 46.464
212226 99.608 -77.066 46.436
212227 99.617 -77.063 46.409
212228 99.626 -77.059 46.381
212229 99.635 -77.058 46.354
212230 99.644 -77.056 46.328
212231 99.653 -77.053 46.302
212232 99.662 -77.051 46.278
212233 99.671 -77.049 46.255
212234 99.681 -77.047 46.232
212235 99.690 -77.044 46.209
212236 99.699 -77.042 46.187
212237 99.708 -77.040 46.165
212238 99.736 -77.032 46.143
212239 99.767 -77.023 46.122
212240 99.798 -77.014 46.101
212241 99.823 -77.007 46.081
212242 99.848 -77.001 46.061
212243 99.872 -76.994 46.042
212244 99.906 -76.993 46.022
212245 99.943 -76.992 46.003
212246 99.980 -76.992 45.984
212247 100.016 -76.991 45.966
212248 100.054 -76.991 45.948
212249 100.089 -76.995 45.930
212250 100.121 -77.018 45.911
212251 100.152 -77.041 45.894
212252 100.183 -77.063 45.876
212253 100.214 -77.086 45.859
212254 100.237 -77.097 45.843
212255 100.260 -77.106 45.826
212256 100.283 -77.116 45.808
212257 100.304 -77.129 45.791
212258 100.324 -77.144 45.774
212259 100.334 -77.158 45.756
212260 100.344 -77.171 45.740
212261 100.355 -77.183 45.724
212262 100.366 -77.195 45.709
212263 100.376 -77.207 45.693
212264 100.387 -77.219 45.681
212265 100.398 -77.232 45.667
212266 100.408 -77.244 45.653
212267 100.419 -77.256 45.638
212268 100.429 -77.267 45.624
212269 100.448 -77.284 45.610
212270 100.469 -77.301 45.598
212271 100.490 -77.318 45.585
212272 100.511 -77.335 45.572
212273 100.531 -77.353 45.560
212274 100.553 -77.370 45.547
212275 100.574 -77.388 45.534
212276 100.597 -77.408 45.522
212277 100.621 -77.428 45.510
212278 100.644 -77.448 45.498
212279 100.667 -77.468 45.486
212280 100.691 -77.488 45.474
212281 100.720 -77.495 45.464
212282 100.751 -77.500 45.453
212283 100.781 -77.505 45.441
212284 100.812 -77.511 45.428
212285 100.841 -77.515 45.415
212286 100.863 -77.517 45.401
212287 100.886 -77.520 45.389
212288 100.909 -77.526 45.375
212289 100.944 -77.556 45.362
212290 100.975 -77.583 45.349
212291 100.999 -77.602 45.336
212292 101.024 -77.620 45.322
212293 101.048 -77.639 45.308
212294 101.071 -77.658 45.295
212295 101.096 -77.677 45.280
212296 101.121 -77.690 45.266
212297 101.147 -77.699 45.251
212298 101.173 -77.708 45.238
212299 101.199 -77.718 45.223
212300 101.225 -77.727 45.208
212301 101.251 -77.736 45.193
212302 101.277 -77.746 45.180
212303 101.303 -77.755 45.165
212304 101.329 -77.764 45.149
212305 101.355 -77.773 45.134
212306 101.384 -77.788 45.118
212307 101.412 -77.804 45.102
212308 101.441 -77.821 45.086
212309 101.470 -77.837 45.070
212310 101.498 -77.853 45.055
212311 101.528 -77.869 45.038
212312 101.556 -77.885 45.022
212313 101.585 -77.902 45.005
212314 101.613 -77.918 44.990
212315 101.638 -77.933 44.975
212316 101.663 -77.948 44.959
212317 101.687 -77.963 44.944
212318 101.712 -77.978 44.928
212319 101.780 -77.985 44.912
212320 101.797 -77.994 44.896
212321 101.803 -78.004 44.881
212322 101.809 -78.014 44.865
212323 101.815 -78.024 44.849
212324 101.821 -78.034 44.833
212325 101.827 -78.044 44.816
212326 101.834 -78.053 44.801
212327 101.843 -78.060 44.786
212328 101.854 -78.067 44.773
212329 101.868 -78.075 44.759
212330 101.892 -78.087 44.743
212331 101.927 -78.097 44.727
212332 101.962 -78.108 44.711
212333 101.996 -78.118 44.694
212334 102.031 -78.128 44.678
212335 102.066 -78.139 44.661
212336 102.101 -78.150 44.646
212337 102.136 -78.160 44.631
212338 102.171 -78.171 44.616
212339 102.205 -78.181 44.601
212340 102.240 -78.192 44.585
212341 102.275 -78.203 44.570
212342 102.314 -78.220 44.555
212343 102.355 -78.241 44.540
212344 102.396 -78.263 44.525
212345 102.432 -78.276 44.510
212346 102.462 -78.282 44.495
212347 102.493 -78.288 44.480
212348 102.524 -78.294 44.466
212349 102.554 -78.300 44.452
212350 102.585 -78.306 44.438
212351 102.615 -78.312 44.424
212352 102.646 -78.318 44.410
212353 102.677 -78.324 44.395
212354 102.708 -78.329 44.381
212355 102.739 -78.333 44.365
212356 102.773 -78.331 44.349
212357 102.801 -78.340 44.333
212358 102.838 -78.354 44.318
212359 102.875 -78.367 44.301
212360 102.912 -78.381 44.284
212361 102.949 -78.395 44.268
212362 102.986 -78.408 44.252
212363 103.023 -78.422 44.235
212364 103.059 -78.435 44.218
212365 103.096 -78.449 44.201
212366 103.134 -78.463 44.183
212367 103.170 -78.476 44.167
212368 103.207 -78.490 44.149
212651 93.366 -75.707 53.918
212652 93.384 -75.710 53.934
212653 93.403 -75.714 53.952
212654 93.421 -75.717 53.973
212655 93.439 -75.720 53.995
212656 93.458 -75.724 54.018
212657 93.476 -75.727 54.042
212658 93.494 -75.730 54.067
212659 93.512 -75.734 54.094
212660 93.531 -75.737 54.122
212661 93.549 -75.740 54.151
212662 93.575 -75.740 54.181
212663 93.608 -75.734 54.205
212664 93.640 -75.728 54.230
212665 93.672 -75.721 54.257
212666 93.705 -75.715 54.287
212667 93.737 -75.709 54.315
212668 93.770 -75.702 54.342
212669 93.802 -75.696 54.362
212670 93.834 -75.690 54.373
212671 93.867 -75.684 54.378
212672 93.899 -75.678 54.377
212673 93.928 -75.671 54.370
212674 93.949 -75.664 54.360
212675 93.979 -75.659 54.350
212676 94.005 -75.654 54.338
212677 94.031 -75.649 54.323
212678 94.057 -75.643 54.305
212679 94.080 -75.642 54.290
212680 94.105 -75.640 54.277
212681 94.129 -75.638 54.265
212682 94.153 -75.637 54.253
212683 94.177 -75.637 54.239
212684 94.201 -75.639 54.226
212685 94.225 -75.640 54.215
212686 94.249 -75.643 54.204
212687 94.274 -75.644 54.192
212688 94.298 -75.644 54.179
212689 94.324 -75.644 54.166
212690 94.347 -75.645 54.149
212691 94.366 -75.646 54.130
212692 94.386 -75.649 54.111
212693 94.399 -75.654 54.090
212694 94.411 -75.660 54.072
212695 94.424 -75.666 54.057
212696 94.436 -75.671 54.041
212697 94.449 -75.677 54.021
212698 94.461 -75.684 53.999
212699 94.474 -75.690 53.975
212700 94.489 -75.696 53.950
212701 94.507 -75.702 53.925
212702 94.527 -75.709 53.898
212703 94.546 -75.715 53.869
212704 94.565 -75.721 53.837
212705 94.583 -75.728 53.801
212706 94.603 -75.735 53.763
212707 94.622 -75.741 53.724
212708 94.641 -75.748 53.685
212709 94.666 -75.748 53.645
212710 94.691 -75.746 53.604
212711 94.707 -75.746 53.562
212712 94.698 -75.749 53.521
212713 94.702 -75.752 53.482
212714 94.711 -75.755 53.444
212715 94.720 -75.757 53.406
212716 94.729 -75.760 53.369
212717 94.738 -75.762 53.330
212718 94.747 -75.765 53.290
212719 94.755 -75.768 53.249
212720 94.764 -75.771 53.207
212721 94.773 -75.774 53.164
212722 94.782 -75.776 53.121
212723 94.790 -75.779 53.077
212724 94.800 -75.781 53.032
212725 94.809 -75.784 52.990
212726 94.817 -75.787 52.952
212727 94.826 -75.790 52.918
212728 94.835 -75.793 52.887
212729 94.844 -75.795 52.857
212730 94.852 -75.798 52.826
212731 94.861 -75.800 52.792
212732 94.871 -75.803 52.757
212733 94.879 -75.806 52.724
212734 94.888 -75.809 52.694
212735 94.915 -75.814 52.665
212736 94.945 -75.820 52.637
212737 94.971 -75.824 52.607
212738 95.005 -75.834 52.576
212739 95.041 -75.845 52.544
212740 95.098 -75.847 52.510
212741 95.155 -75.848 52.475
212742 95.212 -75.850 52.440
212743 95.269 -75.851 52.403
212744 95.327 -75.853 52.362
212745 95.384 -75.854 52.321
212746 95.441 -75.856 52.278
212747 95.498 -75.857 52.232
212748 95.530 -75.841 52.185
212749 95.543 -75.813 52.137
212750 95.554 -75.828 52.088
212751 95.565 -75.844 52.042
212752 95.576 -75.860 51.998
212753 95.588 -75.877 51.954
212754 95.599 -75.893 51.911
212755 95.609 -75.905 51.870
212756 95.619 -75.912 51.828
212757 95.629 -75.919 51.787
212758 95.638 -75.925 51.748
212759 95.648 -75.932 51.711
212760 95.658 -75.938 51.674
212761 95.668 -75.945 51.638
212762 95.681 -75.951 51.603
212763 95.703 -75.956 51.570
212764 95.726 -75.960 51.537
212765 95.748 -75.965 51.506
212766 95.788 -75.965 51.477
212767 95.828 -75.964 51.448
212768 95.869 -75.964 51.416
212769 95.909 -75.964 51.382
212770 95.949 -75.964 51.346
212771 95.998 -75.964 51.309
212772 96.037 -75.966 51.270
212773 96.071 -75.966 51.230
212774 96.115 -75.964 51.191
212775 96.127 -75.968 51.150
212776 96.139 -75.972 51.109
212777 96.152 -75.975 51.069
212778 96.164 -75.979 51.031
212779 96.176 -75.983 50.994
212780 96.188 -75.987 50.958
212781 96.200 -75.991 50.923
212782 96.212 -75.995 50.891
212783 96.226 -75.998 50.860
212784 96.244 -76.006 50.830
212785 96.265 -76.017 50.803
212786 96.288 -76.028 50.777
212787 96.310 -76.040 50.751
212788 96.331 -76.050 50.726
212789 96.354 -76.062 50.698
212790 96.376 -76.073 50.670
212791 96.397 -76.084 50.645
212792 96.419 -76.095 50.622
212793 96.442 -76.107 50.599
212794 96.462 -76.117 50.576
212795 96.462 -76.127 50.554
212796 96.463 -76.136 50.532
212797 96.463 -76.146 50.512
212798 96.464 -76.155 50.491
212799 96.497 -76.168 50.470
212800 96.540 -76.182 50.447
212801 96.583 -76.196 50.421
212802 96.623 -76.204 50.394
212803 96.654 -76.202 50.364
212804 96.683 -76.201 50.333
212805 96.712 -76.200 50.299
212806 96.741 -76.199 50.265
212807 96.770 -76.198 50.229
212808 96.800 -76.196 50.190
212809 96.826 -76.195 50.151
212810 96.842 -76.201 50.111
212811 96.861 -76.214 50.070
212812 96.880 -76.229 50.029
212813 96.898 -76.242 49.989
212814 96.917 -76.257 49.951
212815 96.936 -76.270 49.915
212816 96.955 -76.285 49.880
212817 96.973 -76.298 49.844
212818 96.992 -76.313 49.809
212819 97.013 -76.323 49.773
212820 97.036 -76.330 49.738
212821 97.082 -76.353 49.702
212822 97.129 -76.376 49.669
212823 97.175 -76.392 49.637
212824 97.222 -76.405 49.607
212825 97.268 -76.417 49.578
212826 97.160 -76.403 49.551
212827 97.155 -76.402 49.524
212828 97.197 -76.408 49.495
212829 97.238 -76.424 49.465
212830 97.278 -76.442 49.434
212831 97.318 -76.461 49.402
212832 97.359 -76.479 49.370
212833 97.397 -76.491 49.337
212834 97.430 -76.477 49.305
212835 97.462 -76.464 49.274
212836 97.494 -76.451 49.244
212837 97.511 -76.459 49.214
212838 97.527 -76.467 49.185
212839 97.543 -76.477 49.156
212840 97.549 -76.489 49.127
212841 97.554 -76.503 49.100
212842 97.558 -76.516 49.073
212843 97.562 -76.530 49.049
212844 97.584 -76.542 49.025
212845 97.608 -76.553 49.000
212846 97.632 -76.565 48.974
212847 97.655 -76.577 48.947
212848 97.679 -76.589 48.918
212849 97.703 -76.601 48.890
212850 97.727 -76.613 48.861
212851 97.750 -76.624 48.830
212852 97.775 -76.636 48.798
212853 97.825 -76.643 48.764
212854 97.894 -76.647 48.730
212855 97.963 -76.651 48.695
212856 98.032 -76.654 48.661
212857 98.101 -76.658 48.626
212858 98.160 -76.657 48.591
212859 98.185 -76.650 48.555
212860 98.189 -76.653 48.519
212861 98.193 -76.657 48.485
212862 98.196 -76.660 48.453
212863 98.213 -76.670 48.421
212864 98.233 -76.683 48.391
212865 98.254 -76.696 48.360
212866 98.276 -76.709 48.331
212867 98.297 -76.723 48.302
212868 98.327 -76.711 48.274
212869 98.360 -76.699 48.247
212870 98.388 -76.701 48.219
212871 98.417 -76.702 48.192
212872 98.445 -76.705 48.164
212873 98.474 -76.707 48.135
212874 98.498 -76.711 48.109
212875 98.519 -76.716 48.081
212876 98.541 -76.721 48.053
212877 98.562 -76.727 48.025
212878 98.583 -76.732 47.997
212879 98.605 -76.737 47.968
212880 98.629 -76.721 47.940
212881 98.647 -76.717 47.911
212882 98.659 -76.729 47.882
212883 98.672 -76.741 47.854
212884 98.684 -76.753 47.827
212885 98.696 -76.765 47.799
212886 98.709 -76.776 47.769
212887 98.723 -76.783 47.740
212888 98.737 -76.790 47.711
212889 98.751 -76.797 47.680
212890 98.766 -76.804 47.648
212891 98.780 -76.811 47.615
212892 98.794 -76.818 47.582
212893 98.807 -76.817 47.551
212894 98.819 -76.802 47.519
212895 98.832 -76.787 47.485
212896 98.845 -76.772 47.452
212897 98.882 -76.781 47.417
212898 98.919 -76.790 47.382
212899 98.956 -76.798 47.347
212900 99.002 -76.802 47.312
212901 99.048 -76.805 47.276
212902 99.094 -76.808 47.241
212903 99.140 -76.811 47.207
212904 99.199 -76.815 47.174
212905 99.200 -76.817 47.139
212906 99.218 -76.832 47.103
212907 99.235 -76.846 47.069
212908 99.252 -76.858 47.033
212909 99.269 -76.870 46.997
212910 99.286 -76.882 46.961
212911 99.303 -76.895 46.924
212912 99.320 -76.907 46.888
212913 99.337 -76.919 46.855
212914 99.354 -76.931 46.823
212915 99.371 -76.943 46.791
212916 99.389 -76.956 46.760
212917 99.406 -76.969 46.728
212918 99.422 -76.981 46.695
212919 99.439 -76.993 46.664
212920 99.456 -77.006 46.635
212921 99.474 -77.018 46.606
212922 99.512 -77.020 46.576
212923 99.553 -77.022 46.545
212924 99.580 -77.035 46.515
212925 99.593 -77.047 46.485
212926 99.603 -77.044 46.456
212927 99.612 -77.041 46.428
212928 99.621 -77.038 46.401
212929 99.630 -77.035 46.373
212930 99.639 -77.032 46.346
212931 99.648 -77.029 46.320
212932 99.657 -77.027 46.295
212933 99.666 -77.025 46.270
212934 99.675 -77.022 46.247
212935 99.684 -77.020 46.224
212936 99.694 -77.018 46.201
212937 99.703 -77.016 46.178
212938 99.712 -77.014 46.157
212939 99.739 -77.006 46.134
212940 99.771 -76.997 46.113
212941 99.802 -76.987 46.093
212942 99.834 -76.978 46.073
212943 99.863 -76.970 46.053
212944 99.888 -76.963 46.033
212945 99.924 -76.962 46.014
212946 99.961 -76.962 45.995
212947 99.998 -76.962 45.976
212948 100.035 -76.962 45.957
212949 100.072 -76.961 45.939
212950 100.109 -76.960 45.921
212951 100.140 -76.983 45.903
212952 100.171 -77.006 45.886
212953 100.202 -77.028 45.868
212954 100.234 -77.051 45.851
212955 100.259 -77.073 45.835
212956 100.279 -77.088 45.818
212957 100.299 -77.102 45.801
212958 100.320 -77.116 45.784
212959 100.339 -77.133 45.767
212960 100.354 -77.149 45.750
212961 100.363 -77.162 45.733
212962 100.372 -77.175 45.717
212963 100.381 -77.188 45.702
212964 100.390 -77.201 45.687
212965 100.401 -77.213 45.675
212966 100.411 -77.226 45.661
212967 100.422 -77.238 45.647
212968 100.433 -77.250 45.632
212969 100.443 -77.262 45.618
212970 100.464 -77.279 45.605
212971 100.485 -77.296 45.592
212972 100.506 -77.313 45.580
212973 100.527 -77.331 45.567
212974 100.548 -77.348 45.554
212975 100.568 -77.366 45.541
212976 100.590 -77.382 45.528
212977 100.610 -77.400 45.517
212978 100.633 -77.420 45.505
212979 100.656 -77.439 45.493
212980 100.680 -77.459 45.481
212981 100.703 -77.479 45.469
212982 100.732 -77.489 45.458
212983 100.762 -77.494 45.448
212984 100.792 -77.499 45.436
212985 100.823 -77.504 45.424
212986 100.845 -77.507 45.411
212987 100.867 -77.509 45.397
212988 100.889 -77.511 45.384
212989 100.912 -77.516 45.370
212990 100.947 -77.545 45.356
212991 100.981 -77.576 45.344
212992 101.009 -77.599 45.331
212993 101.034 -77.618 45.317
212994 101.058 -77.637 45.303
212995 101.082 -77.652 45.289
212996 101.109 -77.661 45.274
212997 101.134 -77.671 45.260
212998 101.161 -77.680 45.245
212999 101.187 -77.689 45.230
213000 101.212 -77.698 45.216
213001 101.239 -77.707 45.201
213002 101.265 -77.716 45.186
213003 101.290 -77.725 45.171
213004 101.317 -77.734 45.156
213005 101.343 -77.743 45.140
213006 101.370 -77.755 45.124
213007 101.399 -77.771 45.109
213008 101.427 -77.788 45.092
213009 101.456 -77.804 45.076
213010 101.485 -77.821 45.061
213011 101.513 -77.837 45.045
213012 101.543 -77.853 45.028
213013 101.568 -77.868 45.011
213014 101.593 -77.884 44.994
213015 101.618 -77.899 44.980
213016 101.642 -77.914 44.965
213017 101.668 -77.929 44.949
213018 101.692 -77.944 44.934
213019 101.717 -77.959 44.918
213020 101.791 -77.965 44.902
213021 101.834 -77.971 44.886
213022 101.840 -77.981 44.870
213023 101.846 -77.990 44.854
213024 101.852 -78.000 44.838
213025 101.858 -78.010 44.823
213026 101.865 -78.020 44.807
213027 101.871 -78.030 44.791
213028 101.877 -78.040 44.777
213029 101.883 -78.050 44.763
213030 101.882 -78.061 44.749
213031 101.916 -78.072 44.733
213032 101.951 -78.083 44.717
213033 101.986 -78.094 44.701
213034 102.021 -78.104 44.684
213035 102.055 -78.115 44.667
213036 102.090 -78.125 44.651
213037 102.125 -78.136 44.636
213038 102.160 -78.147 44.621
213039 102.195 -78.157 44.606
213040 102.230 -78.168 44.590
213041 102.264 -78.178 44.575
213042 102.299 -78.189 44.560
213043 102.338 -78.208 44.545
213044 102.379 -78.228 44.531
213045 102.410 -78.235 44.516
213046 102.441 -78.241 44.501
213047 102.471 -78.246 44.486
213048 102.501 -78.252 44.471
213049 102.532 -78.258 44.456
213050 102.563 -78.264 44.442
213051 102.594 -78.265 44.429
213052 102.624 -78.266 44.414
213053 102.655 -78.266 44.400
213054 102.685 -78.267 44.386
213055 102.716 -78.268 44.371
213056 102.750 -78.285 44.356
213057 102.786 -78.296 44.339
213058 102.818 -78.305 44.323
213059 102.855 -78.319 44.307
213060 102.892 -78.333 44.291
213061 102.929 -78.347 44.275
213062 102.966 -78.360 44.258
213063 103.003 -78.374 44.242
213064 103.040 -78.388 44.224
213065 103.077 -78.401 44.207
213066 103.113 -78.415 44.190
213067 103.151 -78.429 44.173
213068 103.187 -78.442 44.156
213069 103.224 -78.456 44.138
213352 93.358 -75.684 53.917
213353 93.376 -75.687 53.932
213354 93.395 -75.690 53.950
213355 93.413 -75.693 53.970
213356 93.431 -75.697 53.992
213357 93.449 -75.700 54.015
213358 93.468 -75.703 54.038
213359 93.486 -75.707 54.064
213360 93.504 -75.710 54.092
213361 93.522 -75.713 54.122
213362 93.542 -75.716 54.153
213363 93.574 -75.711 54.184
213364 93.603 -75.706 54.208
213365 93.635 -75.699 54.232
213366 93.667 -75.693 54.257
213367 93.700 -75.687 54.286
213368 93.732 -75.681 54.315
213369 93.764 -75.675 54.340
213370 93.797 -75.669 54.362
213371 93.829 -75.663 54.379
213372 93.862 -75.657 54.389
213373 93.894 -75.650 54.391
213374 93.926 -75.644 54.389
213375 93.953 -75.637 54.385
213376 93.979 -75.631 54.377
213377 94.005 -75.627 54.363
213378 94.031 -75.621 54.344
213379 94.055 -75.620 54.325
213380 94.079 -75.618 54.307
213381 94.104 -75.616 54.293
213382 94.128 -75.615 54.281
213383 94.152 -75.612 54.266
213384 94.176 -75.612 54.249
213385 94.200 -75.614 54.232
213386 94.224 -75.616 54.217
213387 94.248 -75.618 54.206
213388 94.272 -75.619 54.193
213389 94.296 -75.621 54.181
213390 94.321 -75.622 54.170
213391 94.346 -75.622 54.153
213392 94.365 -75.624 54.136
213393 94.380 -75.628 54.118
213394 94.393 -75.634 54.098
213395 94.405 -75.640 54.081
213396 94.418 -75.646 54.068
213397 94.430 -75.652 54.053
213398 94.442 -75.658 54.034
213399 94.455 -75.664 54.012
213400 94.468 -75.669 53.987
213401 94.480 -75.675 53.961
213402 94.497 -75.682 53.934
213403 94.517 -75.688 53.907
213404 94.535 -75.695 53.877
213405 94.554 -75.702 53.845
213406 94.573 -75.708 53.810
213407 94.593 -75.715 53.773
213408 94.616 -75.716 53.734
213409 94.641 -75.715 53.696
213410 94.667 -75.714 53.656
213411 94.693 -75.713 53.614
213412 94.709 -75.712 53.572
213413 94.700 -75.715 53.531
213414 94.691 -75.718 53.492
213415 94.694 -75.721 53.454
213416 94.703 -75.724 53.417
213417 94.711 -75.726 53.379
213418 94.720 -75.729 53.340
213419 94.730 -75.731 53.301
213420 94.738 -75.734 53.261
213421 94.747 -75.737 53.218
213422 94.756 -75.740 53.176
213423 94.765 -75.742 53.133
213424 94.773 -75.745 53.090
213425 94.782 -75.747 53.044
213426 94.791 -75.750 53.000
213427 94.800 -75.753 52.960
213428 94.809 -75.756 52.924
213429 94.818 -75.759 52.894
213430 94.827 -75.761 52.866
213431 94.835 -75.764 52.836
213432 94.844 -75.766 52.803
213433 94.853 -75.769 52.765
213434 94.862 -75.772 52.728
213435 94.893 -75.778 52.693
213436 94.923 -75.784 52.660
213437 94.952 -75.790 52.627
213438 94.977 -75.793 52.594
213439 95.010 -75.803 52.560
213440 95.044 -75.813 52.527
213441 95.078 -75.825 52.492
213442 95.133 -75.827 52.458
213443 95.191 -75.828 52.423
213444 95.248 -75.830 52.387
213445 95.305 -75.831 52.347
213446 95.362 -75.833 52.307
213447 95.420 -75.834 52.266
213448 95.477 -75.836 52.220
213449 95.521 -75.828 52.174
213450 95.533 -75.800 52.127
213451 95.545 -75.788 52.080
213452 95.556 -75.804 52.035
213453 95.567 -75.820 51.992
213454 95.579 -75.837 51.950
213455 95.590 -75.853 51.908
213456 95.601 -75.869 51.867
213457 95.612 -75.885 51.825
213458 95.622 -75.895 51.784
213459 95.632 -75.902 51.745
213460 95.642 -75.909 51.707
213461 95.652 -75.915 51.671
213462 95.661 -75.922 51.635
213463 95.672 -75.928 51.600
213464 95.694 -75.933 51.567
213465 95.716 -75.938 51.535
213466 95.747 -75.940 51.504
213467 95.788 -75.939 51.476
213468 95.828 -75.939 51.446
213469 95.868 -75.939 51.414
213470 95.913 -75.939 51.381
213471 95.963 -75.939 51.347
213472 96.005 -75.941 51.310
213473 96.037 -75.945 51.273
213474 96.072 -75.941 51.235
213475 96.133 -75.937 51.198
213476 96.145 -75.941 51.159
213477 96.157 -75.944 51.118
213478 96.169 -75.948 51.078
213479 96.181 -75.952 51.041
213480 96.193 -75.956 51.004
213481 96.205 -75.960 50.969
213482 96.218 -75.964 50.934
213483 96.227 -75.966 50.901
213484 96.235 -75.969 50.870
213485 96.256 -75.977 50.841
213486 96.277 -75.988 50.814
213487 96.300 -75.999 50.787
213488 96.322 -76.010 50.762
213489 96.344 -76.021 50.737
213490 96.366 -76.032 50.709
213491 96.388 -76.044 50.680
213492 96.410 -76.054 50.655
213493 96.432 -76.066 50.632
213494 96.454 -76.076 50.610
213495 96.473 -76.089 50.587
213496 96.477 -76.108 50.564
213497 96.481 -76.117 50.542
213498 96.482 -76.127 50.522
213499 96.482 -76.137 50.501
213500 96.490 -76.147 50.481
213501 96.533 -76.160 50.460
213502 96.577 -76.175 50.434
213503 96.617 -76.182 50.406
213504 96.649 -76.182 50.377
213505 96.678 -76.181 50.345
213506 96.707 -76.179 50.312
213507 96.736 -76.178 50.277
213508 96.765 -76.177 50.241
213509 96.795 -76.176 50.202
213510 96.824 -76.174 50.163
213511 96.848 -76.182 50.123
213512 96.867 -76.195 50.081
213513 96.887 -76.209 50.039
213514 96.906 -76.223 49.999
213515 96.926 -76.236 49.960
213516 96.945 -76.250 49.924
213517 96.963 -76.263 49.888
213518 96.982 -76.278 49.853
213519 97.001 -76.292 49.817
213520 97.019 -76.306 49.782
213521 97.038 -76.320 49.749
213522 97.080 -76.336 49.714
213523 97.126 -76.348 49.680
213524 97.173 -76.361 49.648
213525 97.219 -76.373 49.617
213526 97.265 -76.386 49.587
213527 97.143 -76.369 49.559
213528 97.176 -76.373 49.532
213529 97.218 -76.379 49.503
213530 97.261 -76.385 49.473
213531 97.301 -76.402 49.442
213532 97.342 -76.420 49.411
213533 97.381 -76.439 49.380
213534 97.421 -76.453 49.348
213535 97.453 -76.439 49.316
213536 97.486 -76.426 49.286
213537 97.511 -76.423 49.256
213538 97.527 -76.432 49.227
213539 97.543 -76.441 49.198
213540 97.559 -76.450 49.169
213541 97.564 -76.463 49.140
213542 97.568 -76.477 49.113
213543 97.573 -76.489 49.086
213544 97.577 -76.503 49.063
213545 97.590 -76.516 49.039
213546 97.613 -76.527 49.014
213547 97.637 -76.539 48.988
213548 97.661 -76.552 48.959
213549 97.685 -76.563 48.930
213550 97.709 -76.575 48.901
213551 97.732 -76.587 48.872
213552 97.756 -76.599 48.840
213553 97.780 -76.611 48.808
213554 97.849 -76.614 48.774
213555 97.918 -76.618 48.740
213556 97.987 -76.621 48.705
213557 98.056 -76.625 48.669
213558 98.125 -76.629 48.634
213559 98.194 -76.632 48.599
213560 98.216 -76.631 48.563
213561 98.220 -76.634 48.527
213562 98.223 -76.637 48.492
213563 98.226 -76.640 48.459
213564 98.230 -76.643 48.427
213565 98.240 -76.650 48.395
213566 98.262 -76.664 48.365
213567 98.283 -76.677 48.336
213568 98.304 -76.690 48.307
213569 98.337 -76.671 48.278
213570 98.365 -76.674 48.251
213571 98.393 -76.676 48.224
213572 98.422 -76.677 48.197
213573 98.450 -76.680 48.170
213574 98.479 -76.682 48.141
213575 98.507 -76.683 48.114
213576 98.536 -76.686 48.087
213577 98.563 -76.688 48.059
213578 98.584 -76.693 48.032
213579 98.606 -76.699 48.004
213580 98.627 -76.705 47.976
213581 98.650 -76.699 47.947
213582 98.668 -76.695 47.920
213583 98.681 -76.707 47.891
213584 98.693 -76.719 47.863
213585 98.706 -76.731 47.835
213586 98.718 -76.743 47.805
213587 98.730 -76.755 47.775
213588 98.742 -76.768 47.745
213589 98.757 -76.774 47.715
213590 98.771 -76.781 47.683
213591 98.785 -76.789 47.651
213592 98.799 -76.796 47.619
213593 98.813 -76.802 47.587
213594 98.826 -76.788 47.556
213595 98.838 -76.773 47.524
213596 98.850 -76.758 47.491
213597 98.862 -76.743 47.458
213598 98.882 -76.736 47.423
213599 98.920 -76.744 47.387
213600 98.964 -76.748 47.352
213601 99.010 -76.752 47.317
213602 99.056 -76.755 47.280
213603 99.102 -76.768 47.244
213604 99.149 -76.782 47.208
213605 99.195 -76.797 47.175
213606 99.215 -76.797 47.140
213607 99.232 -76.806 47.104
213608 99.249 -76.818 47.069
213609 99.266 -76.830 47.034
213610 99.283 -76.842 46.998
213611 99.300 -76.855 46.961
213612 99.317 -76.867 46.924
213613 99.335 -76.879 46.888
213614 99.351 -76.891 46.854
213615 99.368 -76.903 46.821
213616 99.385 -76.916 46.788
213617 99.403 -76.928 46.756
213618 99.420 -76.940 46.724
213619 99.437 -76.953 46.691
213620 99.453 -76.965 46.658
213621 99.471 -76.978 46.629
213622 99.497 -76.985 46.600
213623 99.538 -76.987 46.569
213624 99.570 -76.996 46.538
213625 99.587 -77.019 46.508
213626 99.597 -77.022 46.477
213627 99.607 -77.019 46.448
213628 99.616 -77.016 46.420
213629 99.625 -77.013 46.393
213630 99.634 -77.010 46.366
213631 99.643 -77.007 46.339
213632 99.652 -77.004 46.314
213633 99.661 -77.002 46.289
213634 99.670 -76.999 46.264
213635 99.679 -76.996 46.240
213636 99.688 -76.994 46.217
213637 99.697 -76.992 46.195
213638 99.706 -76.990 46.172
213639 99.716 -76.987 46.150
213640 99.744 -76.979 46.128
213641 99.775 -76.970 46.107
213642 99.806 -76.961 46.087
213643 99.838 -76.952 46.066
213644 99.869 -76.943 46.046
213645 99.907 -76.942 46.025
213646 99.943 -76.937 46.005
213647 99.979 -76.932 45.986
213648 100.016 -76.931 45.968
213649 100.053 -76.931 45.949
213650 100.090 -76.931 45.931
213651 100.127 -76.931 45.913
213652 100.159 -76.948 45.895
213653 100.191 -76.971 45.878
213654 100.225 -77.022 45.860
213655 100.253 -77.049 45.844
213656 100.275 -77.060 45.827
213657 100.295 -77.075 45.810
213658 100.316 -77.090 45.794
213659 100.336 -77.104 45.776
213660 100.354 -77.122 45.759
213661 100.372 -77.139 45.743
213662 100.382 -77.153 45.726
213663 100.391 -77.166 45.710
213664 100.400 -77.179 45.694
213665 100.409 -77.191 45.680
213666 100.418 -77.205 45.668
213667 100.427 -77.218 45.654
213668 100.436 -77.231 45.641
213669 100.446 -77.244 45.627
213670 100.459 -77.257 45.613
213671 100.480 -77.274 45.600
213672 100.501 -77.291 45.587
213673 100.522 -77.309 45.574
213674 100.543 -77.326 45.561
213675 100.563 -77.343 45.548
213676 100.584 -77.360 45.535
213677 100.605 -77.378 45.522
213678 100.626 -77.395 45.511
213679 100.647 -77.413 45.500
213680 100.669 -77.431 45.488
213681 100.692 -77.451 45.477
213682 100.715 -77.470 45.465
213683 100.743 -77.483 45.453
213684 100.773 -77.489 45.442
213685 100.804 -77.493 45.431
213686 100.826 -77.496 45.419
213687 100.848 -77.498 45.405
213688 100.870 -77.501 45.392
213689 100.893 -77.503 45.379
213690 100.915 -77.505 45.365
213691 100.949 -77.536 45.352
213692 100.984 -77.565 45.339
213693 101.019 -77.595 45.325
213694 101.044 -77.614 45.312
213695 101.070 -77.623 45.297
213696 101.096 -77.632 45.283
213697 101.122 -77.641 45.268
213698 101.148 -77.650 45.253
213699 101.174 -77.659 45.238
213700 101.200 -77.668 45.223
213701 101.226 -77.677 45.208
213702 101.252 -77.686 45.193
213703 101.278 -77.696 45.178
213704 101.304 -77.705 45.162
213705 101.330 -77.714 45.147
213706 101.356 -77.723 45.131
213707 101.385 -77.739 45.115
213708 101.414 -77.755 45.099
213709 101.443 -77.771 45.083
213710 101.471 -77.788 45.066
213711 101.498 -77.803 45.050
213712 101.523 -77.819 45.034
213713 101.548 -77.834 45.017
213714 101.573 -77.849 45.001
213715 101.598 -77.865 44.984
213716 101.622 -77.880 44.969
213717 101.647 -77.895 44.954
213718 101.672 -77.910 44.939
213719 101.697 -77.925 44.925
213720 101.722 -77.940 44.908
213721 101.802 -77.945 44.891
213722 101.872 -77.947 44.875
213723 101.878 -77.957 44.859
213724 101.884 -77.967 44.842
213725 101.890 -77.977 44.826
213726 101.896 -77.987 44.811
213727 101.902 -77.996 44.795
213728 101.908 -78.006 44.780
213729 101.914 -78.016 44.765
213730 101.906 -78.034 44.752
213731 101.893 -78.040 44.738
213732 101.939 -78.053 44.722
213733 101.974 -78.065 44.706
213734 102.009 -78.076 44.690
213735 102.044 -78.087 44.672
213736 102.079 -78.098 44.656
213737 102.114 -78.109 44.641
213738 102.149 -78.120 44.626
213739 102.184 -78.131 44.610
213740 102.219 -78.143 44.595
213741 102.253 -78.154 44.579
213742 102.289 -78.165 44.564
213743 102.323 -78.175 44.550
213744 102.358 -78.186 44.536
213745 102.388 -78.190 44.521
213746 102.418 -78.191 44.507
213747 102.449 -78.191 44.492
213748 102.480 -78.192 44.477
213749 102.510 -78.193 44.461
213750 102.541 -78.194 44.446
213751 102.572 -78.194 44.432
213752 102.602 -78.195 44.418
213753 102.632 -78.196 44.404
213754 102.663 -78.197 44.390
213755 102.694 -78.197 44.375
213756 102.725 -78.200 44.361
213757 102.761 -78.225 44.346
213758 102.798 -78.251 44.330
213759 102.835 -78.271 44.313
213760 102.872 -78.285 44.297
213761 102.909 -78.298 44.281
213762 102.946 -78.312 44.265
213763 102.983 -78.326 44.248
213764 103.020 -78.339 44.231
213765 103.056 -78.353 44.214
213766 103.094 -78.366 44.197
213767 103.130 -78.380 44.180
213768 103.167 -78.394 44.162
213769 103.205 -78.407 44.145
213770 103.241 -78.422 44.127
214053 93.350 -75.661 53.915
214054 93.368 -75.664 53.929
214055 93.386 -75.667 53.946
214056 93.405 -75.671 53.966
214057 93.423 -75.674 53.989
214058 93.441 -75.677 54.011
214059 93.460 -75.681 54.034
214060 93.478 -75.684 54.060
214061 93.496 -75.687 54.090
214062 93.517 -75.687 54.121
214063 93.541 -75.687 54.154
214064 93.578 -75.681 54.187
214065 93.615 -75.672 54.212
214066 93.644 -75.667 54.236
214067 93.674 -75.662 54.260
214068 93.703 -75.657 54.287
214069 93.733 -75.652 54.314
214070 93.763 -75.646 54.339
214071 93.792 -75.641 54.361
214072 93.824 -75.635 54.380
214073 93.856 -75.629 54.391
214074 93.889 -75.623 54.398
214075 93.921 -75.617 54.403
214076 93.954 -75.611 54.404
214077 93.980 -75.604 54.399
214078 94.006 -75.599 54.386
214079 94.030 -75.598 54.368
214080 94.054 -75.596 54.347
214081 94.078 -75.594 54.327
214082 94.102 -75.592 54.311
214083 94.126 -75.590 54.298
214084 94.150 -75.589 54.284
214085 94.175 -75.587 54.266
214086 94.199 -75.589 54.247
214087 94.223 -75.591 54.227
214088 94.247 -75.593 54.212
214089 94.271 -75.594 54.199
214090 94.295 -75.596 54.189
214091 94.319 -75.598 54.180
214092 94.343 -75.599 54.167
214093 94.362 -75.602 54.152
214094 94.374 -75.609 54.134
214095 94.387 -75.615 54.114
214096 94.399 -75.620 54.098
214097 94.411 -75.626 54.084
214098 94.424 -75.632 54.069
214099 94.436 -75.638 54.049
214100 94.449 -75.643 54.025
214101 94.461 -75.649 53.998
214102 94.474 -75.656 53.970
214103 94.487 -75.662 53.942
214104 94.506 -75.668 53.915
214105 94.525 -75.674 53.884
214106 94.544 -75.681 53.851
214107 94.566 -75.684 53.818
214108 94.592 -75.683 53.782
214109 94.617 -75.682 53.745
214110 94.643 -75.681 53.706
214111 94.669 -75.680 53.667
214112 94.694 -75.679 53.625
214113 94.712 -75.679 53.582
214114 94.702 -75.681 53.541
214115 94.693 -75.684 53.502
214116 94.684 -75.687 53.464
214117 94.686 -75.690 53.426
214118 94.694 -75.692 53.389
214119 94.703 -75.695 53.351
214120 94.712 -75.697 53.312
214121 94.721 -75.700 53.272
214122 94.730 -75.702 53.229
214123 94.739 -75.706 53.187
214124 94.748 -75.709 53.145
214125 94.756 -75.711 53.103
214126 94.765 -75.714 53.056
214127 94.774 -75.716 53.011
214128 94.783 -75.719 52.968
214129 94.792 -75.721 52.931
214130 94.801 -75.724 52.900
214131 94.810 -75.727 52.873
214132 94.818 -75.730 52.845
214133 94.827 -75.733 52.813
214134 94.841 -75.736 52.775
214135 94.871 -75.742 52.736
214136 94.901 -75.748 52.697
214137 94.931 -75.754 52.659
214138 94.959 -75.759 52.622
214139 94.984 -75.762 52.586
214140 95.016 -75.771 52.549
214141 95.050 -75.781 52.514
214142 95.084 -75.792 52.478
214143 95.118 -75.803 52.442
214144 95.170 -75.807 52.406
214145 95.227 -75.809 52.370
214146 95.284 -75.810 52.331
214147 95.341 -75.812 52.293
214148 95.398 -75.813 52.253
214149 95.456 -75.815 52.209
214150 95.512 -75.815 52.164
214151 95.524 -75.786 52.118
214152 95.536 -75.758 52.073
214153 95.547 -75.764 52.029
214154 95.558 -75.780 51.987
214155 95.569 -75.796 51.945
214156 95.581 -75.812 51.905
214157 95.592 -75.828 51.863
214158 95.603 -75.845 51.822
214159 95.614 -75.861 51.781
214160 95.626 -75.877 51.742
214161 95.635 -75.885 51.704
214162 95.645 -75.891 51.667
214163 95.655 -75.898 51.632
214164 95.665 -75.905 51.599
214165 95.685 -75.910 51.566
214166 95.707 -75.914 51.534
214167 95.747 -75.914 51.505
214168 95.788 -75.914 51.476
214169 95.828 -75.914 51.446
214170 95.879 -75.914 51.416
214171 95.928 -75.915 51.384
214172 95.971 -75.916 51.349
214173 96.004 -75.920 51.315
214174 96.037 -75.924 51.280
214175 96.072 -75.917 51.242
214176 96.151 -75.909 51.205
214177 96.163 -75.913 51.166
214178 96.175 -75.917 51.127
214179 96.187 -75.922 51.088
214180 96.199 -75.925 51.050
214181 96.212 -75.929 51.014
214182 96.220 -75.931 50.978
214183 96.223 -75.931 50.944
214184 96.228 -75.933 50.911
214185 96.249 -75.939 50.879
214186 96.269 -75.947 50.850
214187 96.291 -75.957 50.823
214188 96.313 -75.969 50.797
214189 96.335 -75.980 50.772
214190 96.357 -75.991 50.746
214191 96.379 -76.002 50.719
214192 96.401 -76.013 50.691
214193 96.424 -76.024 50.666
214194 96.446 -76.035 50.643
214195 96.467 -76.047 50.619
214196 96.483 -76.061 50.596
214197 96.493 -76.077 50.574
214198 96.493 -76.097 50.553
214199 96.501 -76.108 50.532
214200 96.501 -76.118 50.512
214201 96.502 -76.127 50.491
214202 96.527 -76.139 50.469
214203 96.570 -76.154 50.443
214204 96.611 -76.162 50.416
214205 96.644 -76.161 50.387
214206 96.673 -76.160 50.356
214207 96.702 -76.159 50.323
214208 96.731 -76.157 50.289
214209 96.760 -76.156 50.253
214210 96.790 -76.155 50.214
214211 96.816 -76.157 50.174
214212 96.837 -76.169 50.133
214213 96.856 -76.183 50.091
214214 96.876 -76.197 50.050
214215 96.895 -76.210 50.010
214216 96.914 -76.224 49.971
214217 96.934 -76.238 49.933
214218 96.953 -76.251 49.897
214219 96.973 -76.265 49.862
214220 96.993 -76.279 49.827
214221 97.016 -76.289 49.792
214222 97.044 -76.295 49.758
214223 97.078 -76.303 49.724
214224 97.124 -76.316 49.689
214225 97.170 -76.329 49.656
214226 97.217 -76.341 49.625
214227 97.253 -76.351 49.595
214228 97.154 -76.338 49.567
214229 97.196 -76.344 49.538
214230 97.239 -76.349 49.509
214231 97.282 -76.355 49.480
214232 97.324 -76.362 49.450
214233 97.364 -76.380 49.419
214234 97.404 -76.398 49.389
214235 97.444 -76.415 49.358
214236 97.476 -76.401 49.328
214237 97.509 -76.389 49.299
214238 97.526 -76.396 49.270
214239 97.542 -76.405 49.241
214240 97.559 -76.414 49.212
214241 97.574 -76.423 49.183
214242 97.579 -76.436 49.154
214243 97.583 -76.450 49.127
214244 97.587 -76.463 49.102
214245 97.592 -76.477 49.077
214246 97.596 -76.489 49.052
214247 97.619 -76.502 49.027
214248 97.643 -76.514 49.000
214249 97.667 -76.525 48.971
214250 97.690 -76.537 48.942
214251 97.714 -76.549 48.912
214252 97.738 -76.561 48.882
214253 97.762 -76.573 48.850
214254 97.803 -76.581 48.817
214255 97.872 -76.585 48.783
214256 97.941 -76.589 48.748
214257 98.010 -76.592 48.713
214258 98.079 -76.596 48.677
214259 98.148 -76.599 48.641
214260 98.211 -76.605 48.607
214261 98.236 -76.618 48.570
214262 98.246 -76.617 48.534
214263 98.253 -76.618 48.498
214264 98.257 -76.621 48.464
214265 98.261 -76.624 48.432
214266 98.264 -76.627 48.400
214267 98.268 -76.630 48.369
214268 98.289 -76.644 48.340
214269 98.310 -76.657 48.311
214270 98.342 -76.646 48.282
214271 98.371 -76.649 48.255
214272 98.399 -76.650 48.228
214273 98.427 -76.652 48.201
214274 98.456 -76.655 48.174
214275 98.484 -76.656 48.146
214276 98.513 -76.658 48.118
214277 98.541 -76.661 48.092
214278 98.570 -76.663 48.065
214279 98.598 -76.665 48.039
214280 98.627 -76.667 48.011
214281 98.650 -76.671 47.983
214282 98.671 -76.677 47.954
214283 98.690 -76.674 47.927
214284 98.702 -76.685 47.899
214285 98.715 -76.697 47.871
214286 98.727 -76.709 47.841
214287 98.739 -76.721 47.811
214288 98.751 -76.734 47.781
214289 98.764 -76.746 47.750
214290 98.776 -76.758 47.719
214291 98.790 -76.766 47.687
214292 98.804 -76.773 47.655
214293 98.818 -76.780 47.623
214294 98.831 -76.774 47.591
214295 98.843 -76.759 47.560
214296 98.855 -76.743 47.528
214297 98.868 -76.729 47.496
214298 98.880 -76.714 47.464
214299 98.892 -76.699 47.429
214300 98.926 -76.703 47.394
214301 98.973 -76.718 47.359
214302 99.019 -76.733 47.323
214303 99.065 -76.747 47.285
214304 99.112 -76.762 47.247
214305 99.158 -76.777 47.210
214306 99.205 -76.791 47.176
214307 99.227 -76.793 47.141
214308 99.245 -76.793 47.105
214309 99.263 -76.793 47.069
214310 99.280 -76.802 47.034
214311 99.297 -76.815 46.997
214312 99.314 -76.827 46.961
214313 99.332 -76.839 46.924
214314 99.349 -76.851 46.887
214315 99.366 -76.864 46.852
214316 99.382 -76.876 46.818
214317 99.400 -76.888 46.785
214318 99.417 -76.900 46.753
214319 99.434 -76.912 46.720
214320 99.451 -76.925 46.687
214321 99.468 -76.937 46.655
214322 99.485 -76.950 46.625
214323 99.522 -76.953 46.595
214324 99.559 -76.957 46.564
214325 99.576 -76.980 46.532
214326 99.593 -77.000 46.502
214327 99.602 -76.997 46.472
214328 99.611 -76.994 46.442
214329 99.620 -76.991 46.415
214330 99.629 -76.988 46.387
214331 99.638 -76.985 46.361
214332 99.647 -76.982 46.335
214333 99.656 -76.980 46.310
214334 99.665 -76.977 46.285
214335 99.674 -76.974 46.260
214336 99.683 -76.972 46.236
214337 99.692 -76.969 46.214
214338 99.701 -76.965 46.191
214339 99.710 -76.963 46.168
214340 99.719 -76.961 46.145
214341 99.747 -76.953 46.122
214342 99.779 -76.943 46.101
214343 99.816 -76.942 46.080
214344 99.854 -76.942 46.060
214345 99.893 -76.942 46.038
214346 99.931 -76.942 46.017
214347 99.970 -76.942 45.997
214348 100.006 -76.938 45.978
214349 100.042 -76.931 45.959
214350 100.077 -76.925 45.941
214351 100.113 -76.918 45.923
214352 100.148 -76.911 45.905
214353 100.183 -76.944 45.887
214354 100.218 -76.997 45.870
214355 100.245 -77.025 45.852
214356 100.268 -77.038 45.836
214357 100.291 -77.047 45.820
214358 100.311 -77.063 45.803
214359 100.332 -77.077 45.786
214360 100.352 -77.092 45.769
214361 100.369 -77.110 45.752
214362 100.388 -77.128 45.735
214363 100.402 -77.143 45.718
214364 100.411 -77.156 45.703
214365 100.420 -77.169 45.688
214366 100.429 -77.182 45.673
214367 100.438 -77.195 45.660
214368 100.447 -77.208 45.647
214369 100.456 -77.222 45.634
214370 100.465 -77.235 45.621
214371 100.478 -77.249 45.607
214372 100.498 -77.267 45.594
214373 100.518 -77.285 45.581
214374 100.538 -77.303 45.568
214375 100.558 -77.321 45.555
214376 100.579 -77.338 45.542
214377 100.600 -77.356 45.530
214378 100.621 -77.373 45.517
214379 100.642 -77.390 45.506
214380 100.663 -77.407 45.494
214381 100.684 -77.425 45.482
214382 100.705 -77.442 45.471
214383 100.728 -77.462 45.460
214384 100.754 -77.477 45.448
214385 100.784 -77.483 45.437
214386 100.808 -77.486 45.425
214387 100.830 -77.488 45.412
214388 100.852 -77.490 45.399
214389 100.874 -77.492 45.387
214390 100.896 -77.495 45.373
214391 100.918 -77.497 45.360
214392 100.952 -77.525 45.346
214393 100.987 -77.555 45.333
214394 101.024 -77.576 45.319
214395 101.059 -77.592 45.305
214396 101.085 -77.602 45.291
214397 101.111 -77.611 45.277
214398 101.136 -77.620 45.262
214399 101.162 -77.630 45.247
214400 101.188 -77.639 45.231
214401 101.213 -77.648 45.217
214402 101.240 -77.657 45.200
214403 101.266 -77.666 45.184
214404 101.292 -77.675 45.168
214405 101.318 -77.684 45.153
214406 101.344 -77.693 45.137
214407 101.372 -77.706 45.120
214408 101.400 -77.723 45.104
214409 101.428 -77.739 45.087
214410 101.453 -77.754 45.072
214411 101.478 -77.769 45.056
214412 101.503 -77.784 45.040
214413 101.528 -77.799 45.023
214414 101.553 -77.815 45.007
214415 101.578 -77.830 44.990
214416 101.603 -77.846 44.973
214417 101.627 -77.861 44.958
214418 101.652 -77.876 44.943
214419 101.677 -77.891 44.928
214420 101.702 -77.906 44.913
214421 101.727 -77.921 44.896
214422 101.812 -77.925 44.880
214423 101.908 -77.924 44.863
214424 101.914 -77.934 44.847
214425 101.921 -77.944 44.831
214426 101.927 -77.954 44.814
214427 101.933 -77.964 44.799
214428 101.939 -77.974 44.783
214429 101.942 -77.985 44.768
214430 101.929 -78.006 44.753
214431 101.901 -78.012 44.739
214432 101.909 -78.021 44.725
214433 101.962 -78.035 44.710
214434 101.997 -78.046 44.693
214435 102.032 -78.057 44.677
214436 102.067 -78.069 44.660
214437 102.102 -78.080 44.645
214438 102.137 -78.091 44.630
214439 102.172 -78.102 44.613
214440 102.207 -78.113 44.599
214441 102.242 -78.124 44.584
214442 102.276 -78.135 44.569
214443 102.309 -78.148 44.556
214444 102.338 -78.161 44.541
214445 102.364 -78.175 44.527
214446 102.395 -78.174 44.513
214447 102.427 -78.168 44.498
214448 102.459 -78.162 44.482
214449 102.490 -78.156 44.467
214450 102.522 -78.150 44.451
214451 102.553 -78.142 44.436
214452 102.581 -78.131 44.422
214453 102.611 -78.126 44.407
214454 102.641 -78.127 44.393
214455 102.672 -78.128 44.379
214456 102.702 -78.128 44.364
214457 102.736 -78.141 44.350
214458 102.766 -78.183 44.335
214459 102.802 -78.257 44.319
214460 102.844 -78.261 44.303
214461 102.886 -78.260 44.287
214462 102.926 -78.264 44.270
214463 102.963 -78.278 44.254
214464 103.000 -78.291 44.237
214465 103.037 -78.305 44.220
214466 103.074 -78.319 44.203
214467 103.110 -78.332 44.186
214468 103.148 -78.346 44.169
214469 103.184 -78.360 44.151
214470 103.221 -78.373 44.133
214471 103.258 -78.387 44.115
214753 93.323 -75.634 53.906
214754 93.342 -75.637 53.914
214755 93.360 -75.640 53.928
214756 93.378 -75.644 53.944
214757 93.397 -75.647 53.962
214758 93.415 -75.650 53.984
214759 93.433 -75.654 54.006
214760 93.451 -75.657 54.030
214761 93.470 -75.660 54.057
214762 93.493 -75.659 54.086
214763 93.516 -75.659 54.118
214764 93.539 -75.657 54.154
214765 93.574 -75.656 54.190
214766 93.625 -75.640 54.216
214767 93.655 -75.634 54.242
214768 93.685 -75.629 54.266
214769 93.714 -75.624 54.290
214770 93.744 -75.618 54.313
214771 93.774 -75.613 54.337
214772 93.803 -75.608 54.360
214773 93.833 -75.602 54.376
214774 93.863 -75.598 54.389
214775 93.893 -75.593 54.401
214776 93.922 -75.587 54.412
214777 93.952 -75.582 54.418
214778 93.980 -75.577 54.417
214779 94.004 -75.576 54.408
214780 94.029 -75.574 54.391
214781 94.053 -75.572 54.369
214782 94.077 -75.570 54.348
214783 94.101 -75.568 54.331
214784 94.125 -75.567 54.317
214785 94.149 -75.565 54.304
214786 94.173 -75.563 54.285
214787 94.197 -75.565 54.265
214788 94.221 -75.566 54.243
214789 94.246 -75.568 54.225
214790 94.270 -75.569 54.211
214791 94.294 -75.571 54.199
214792 94.318 -75.573 54.191
214793 94.343 -75.574 54.181
214794 94.362 -75.577 54.168
214795 94.373 -75.584 54.152
214796 94.385 -75.591 54.134
214797 94.396 -75.597 54.118
214798 94.408 -75.604 54.103
214799 94.419 -75.611 54.086
214800 94.431 -75.618 54.065
214801 94.443 -75.624 54.039
214802 94.455 -75.630 54.011
214803 94.468 -75.636 53.981
214804 94.480 -75.641 53.950
214805 94.496 -75.647 53.921
214806 94.517 -75.652 53.890
214807 94.542 -75.651 53.856
214808 94.567 -75.650 53.823
214809 94.593 -75.649 53.789
214810 94.619 -75.648 53.752
214811 94.644 -75.647 53.714
214812 94.670 -75.646 53.674
214813 94.695 -75.645 53.632
214814 94.714 -75.645 53.589
214815 94.705 -75.647 53.548
214816 94.695 -75.650 53.508
214817 94.686 -75.652 53.469
214818 94.677 -75.656 53.432
214819 94.677 -75.658 53.396
214820 94.686 -75.661 53.358
214821 94.695 -75.664 53.318
214822 94.704 -75.666 53.278
214823 94.713 -75.669 53.237
214824 94.721 -75.671 53.196
214825 94.731 -75.674 53.155
214826 94.739 -75.677 53.113
214827 94.748 -75.680 53.067
214828 94.757 -75.682 53.020
214829 94.766 -75.685 52.976
214830 94.775 -75.687 52.936
214831 94.783 -75.690 52.906
214832 94.792 -75.693 52.879
214833 94.801 -75.696 52.852
214834 94.819 -75.700 52.821
214835 94.849 -75.706 52.785
214836 94.879 -75.712 52.745
214837 94.909 -75.718 52.704
214838 94.939 -75.724 52.664
214839 94.966 -75.728 52.623
214840 94.991 -75.731 52.582
214841 95.021 -75.738 52.543
214842 95.055 -75.749 52.505
214843 95.089 -75.760 52.466
214844 95.123 -75.771 52.428
214845 95.158 -75.782 52.391
214846 95.205 -75.787 52.355
214847 95.263 -75.789 52.317
214848 95.320 -75.790 52.279
214849 95.377 -75.792 52.240
214850 95.435 -75.793 52.199
214851 95.491 -75.795 52.155
214852 95.514 -75.773 52.110
214853 95.526 -75.744 52.067
214854 95.538 -75.724 52.024
214855 95.549 -75.740 51.982
214856 95.561 -75.756 51.941
214857 95.572 -75.772 51.900
214858 95.583 -75.788 51.860
214859 95.594 -75.805 51.819
214860 95.606 -75.821 51.778
214861 95.617 -75.837 51.740
214862 95.628 -75.853 51.701
214863 95.639 -75.868 51.665
214864 95.649 -75.875 51.631
214865 95.659 -75.881 51.598
214866 95.676 -75.887 51.567
214867 95.706 -75.889 51.536
214868 95.747 -75.889 51.507
214869 95.794 -75.889 51.479
214870 95.844 -75.890 51.450
214871 95.894 -75.890 51.420
214872 95.939 -75.891 51.389
214873 95.971 -75.895 51.355
214874 96.001 -75.897 51.321
214875 96.025 -75.894 51.287
214876 96.075 -75.892 51.250
214877 96.169 -75.883 51.212
214878 96.181 -75.887 51.173
214879 96.193 -75.891 51.135
214880 96.205 -75.894 51.096
214881 96.213 -75.897 51.060
214882 96.216 -75.897 51.023
214883 96.219 -75.897 50.988
214884 96.222 -75.897 50.953
214885 96.242 -75.903 50.920
214886 96.262 -75.910 50.888
214887 96.283 -75.917 50.859
214888 96.304 -75.927 50.832
214889 96.326 -75.938 50.806
214890 96.348 -75.950 50.781
214891 96.371 -75.960 50.756
214892 96.393 -75.972 50.729
214893 96.415 -75.983 50.702
214894 96.437 -75.994 50.677
214895 96.459 -76.005 50.653
214896 96.477 -76.018 50.628
214897 96.493 -76.032 50.606
214898 96.509 -76.047 50.584
214899 96.509 -76.067 50.562
214900 96.508 -76.088 50.541
214901 96.519 -76.099 50.522
214902 96.520 -76.109 50.501
214903 96.521 -76.118 50.478
214904 96.564 -76.132 50.452
214905 96.605 -76.141 50.426
214906 96.639 -76.141 50.397
214907 96.668 -76.139 50.367
214908 96.697 -76.138 50.334
214909 96.726 -76.137 50.300
214910 96.754 -76.137 50.264
214911 96.781 -76.139 50.225
214912 96.806 -76.144 50.185
214913 96.826 -76.157 50.143
214914 96.845 -76.171 50.101
214915 96.865 -76.185 50.061
214916 96.884 -76.199 50.020
214917 96.903 -76.213 49.981
214918 96.923 -76.226 49.944
214919 96.943 -76.240 49.907
214920 96.969 -76.248 49.871
214921 96.997 -76.254 49.836
214922 97.024 -76.260 49.802
214923 97.052 -76.267 49.767
214924 97.080 -76.273 49.732
214925 97.121 -76.284 49.698
214926 97.168 -76.296 49.664
214927 97.215 -76.309 49.633
214928 97.235 -76.317 49.603
214929 97.175 -76.308 49.574
214930 97.218 -76.314 49.544
214931 97.260 -76.320 49.515
214932 97.303 -76.326 49.487
214933 97.346 -76.332 49.458
214934 97.387 -76.340 49.428
214935 97.428 -76.358 49.398
214936 97.468 -76.376 49.369
214937 97.500 -76.364 49.340
214938 97.526 -76.360 49.312
214939 97.542 -76.369 49.283
214940 97.558 -76.378 49.254
214941 97.574 -76.387 49.226
214942 97.589 -76.397 49.196
214943 97.593 -76.410 49.168
214944 97.598 -76.423 49.141
214945 97.602 -76.436 49.117
214946 97.607 -76.450 49.092
214947 97.611 -76.463 49.066
214948 97.625 -76.476 49.039
214949 97.648 -76.488 49.012
214950 97.672 -76.499 48.983
214951 97.696 -76.511 48.954
214952 97.720 -76.524 48.924
214953 97.743 -76.535 48.892
214954 97.768 -76.547 48.860
214955 97.827 -76.552 48.827
214956 97.896 -76.556 48.792
214957 97.965 -76.559 48.757
214958 98.034 -76.563 48.720
214959 98.103 -76.567 48.684
214960 98.163 -76.573 48.648
214961 98.206 -76.586 48.613
214962 98.249 -76.611 48.577
214963 98.249 -76.614 48.540
214964 98.259 -76.613 48.504
214965 98.269 -76.612 48.469
214966 98.279 -76.612 48.436
214967 98.288 -76.611 48.404
214968 98.298 -76.611 48.372
214969 98.301 -76.614 48.343
214970 98.318 -76.621 48.313
214971 98.348 -76.621 48.285
214972 98.376 -76.624 48.258
214973 98.404 -76.625 48.231
214974 98.433 -76.627 48.204
214975 98.461 -76.630 48.177
214976 98.490 -76.631 48.149
214977 98.518 -76.633 48.122
214978 98.547 -76.636 48.096
214979 98.575 -76.637 48.070
214980 98.603 -76.640 48.044
214981 98.632 -76.642 48.017
214982 98.660 -76.643 47.989
214983 98.689 -76.646 47.961
214984 98.712 -76.652 47.934
214985 98.724 -76.664 47.905
214986 98.736 -76.675 47.876
214987 98.748 -76.687 47.847
214988 98.761 -76.699 47.817
214989 98.773 -76.712 47.788
214990 98.785 -76.724 47.757
214991 98.797 -76.736 47.725
214992 98.810 -76.748 47.693
214993 98.823 -76.757 47.661
214994 98.837 -76.759 47.629
214995 98.849 -76.744 47.596
214996 98.861 -76.730 47.566
214997 98.873 -76.718 47.535
214998 98.885 -76.708 47.503
214999 98.898 -76.697 47.470
215000 98.910 -76.687 47.437
215001 98.935 -76.696 47.401
215002 98.982 -76.712 47.364
215003 99.028 -76.727 47.328
215004 99.075 -76.742 47.290
215005 99.122 -76.756 47.252
215006 99.168 -76.771 47.213
215007 99.214 -76.786 47.177
215008 99.240 -76.790 47.140
215009 99.258 -76.790 47.104
215010 99.276 -76.790 47.069
215011 99.294 -76.789 47.033
215012 99.311 -76.789 46.997
215013 99.329 -76.799 46.960
215014 99.346 -76.811 46.923
215015 99.363 -76.824 46.886
215016 99.380 -76.836 46.851
215017 99.397 -76.848 46.816
215018 99.414 -76.860 46.783
215019 99.431 -76.872 46.750
215020 99.448 -76.885 46.718
215021 99.465 -76.897 46.685
215022 99.482 -76.909 46.653
215023 99.507 -76.918 46.622
215024 99.548 -76.920 46.591
215025 99.566 -76.940 46.559
215026 99.583 -76.963 46.527
215027 99.597 -76.975 46.497
215028 99.606 -76.972 46.467
215029 99.615 -76.969 46.437
215030 99.624 -76.966 46.410
215031 99.633 -76.963 46.383
215032 99.642 -76.960 46.357
215033 99.651 -76.958 46.332
215034 99.660 -76.955 46.307
215035 99.669 -76.952 46.282
215036 99.679 -76.950 46.258
215037 99.688 -76.947 46.235
215038 99.697 -76.943 46.212
215039 99.706 -76.940 46.188
215040 99.714 -76.938 46.165
215041 99.728 -76.937 46.141
215042 99.761 -76.942 46.118
215043 99.801 -76.942 46.095
215044 99.840 -76.942 46.074
215045 99.878 -76.942 46.053
215046 99.917 -76.942 46.031
215047 99.955 -76.942 46.010
215048 99.994 -76.942 45.990
215049 100.032 -76.942 45.970
215050 100.069 -76.940 45.952
215051 100.105 -76.933 45.933
215052 100.140 -76.926 45.915
215053 100.176 -76.919 45.897
215054 100.210 -76.972 45.879
215055 100.238 -77.001 45.862
215056 100.261 -77.016 45.845
215057 100.284 -77.025 45.829
215058 100.307 -77.035 45.813
215059 100.327 -77.050 45.796
215060 100.348 -77.064 45.778
215061 100.367 -77.080 45.761
215062 100.385 -77.098 45.744
215063 100.403 -77.116 45.728
215064 100.421 -77.134 45.712
215065 100.430 -77.147 45.695
215066 100.439 -77.160 45.680
215067 100.448 -77.172 45.665
215068 100.457 -77.186 45.651
215069 100.466 -77.199 45.639
215070 100.475 -77.212 45.626
215071 100.484 -77.226 45.613
215072 100.499 -77.241 45.600
215073 100.519 -77.258 45.587
215074 100.539 -77.276 45.574
215075 100.559 -77.294 45.561
215076 100.579 -77.312 45.548
215077 100.600 -77.330 45.536
215078 100.620 -77.348 45.524
215079 100.640 -77.366 45.511
215080 100.660 -77.384 45.500
215081 100.680 -77.402 45.488
215082 100.700 -77.420 45.476
215083 100.721 -77.437 45.465
215084 100.742 -77.454 45.454
215085 100.765 -77.471 45.442
215086 100.789 -77.475 45.431
215087 100.811 -77.477 45.419
215088 100.833 -77.479 45.406
215089 100.855 -77.482 45.393
215090 100.878 -77.484 45.381
215091 100.900 -77.486 45.367
215092 100.922 -77.489 45.353
215093 100.956 -77.513 45.339
215094 100.994 -77.530 45.326
215095 101.032 -77.548 45.312
215096 101.071 -77.566 45.298
215097 101.102 -77.580 45.284
215098 101.127 -77.589 45.270
215099 101.153 -77.598 45.254
215100 101.178 -77.608 45.239
215101 101.204 -77.617 45.223
215102 101.230 -77.627 45.208
215103 101.255 -77.636 45.192
215104 101.280 -77.645 45.175
215105 101.305 -77.655 45.159
215106 101.331 -77.664 45.144
215107 101.357 -77.674 45.127
215108 101.383 -77.689 45.111
215109 101.408 -77.705 45.094
215110 101.433 -77.720 45.076
215111 101.458 -77.735 45.061
215112 101.483 -77.750 45.045
215113 101.507 -77.765 45.029
215114 101.533 -77.780 45.012
215115 101.557 -77.796 44.996
215116 101.583 -77.812 44.979
215117 101.607 -77.827 44.962
215118 101.632 -77.842 44.947
215119 101.657 -77.857 44.931
215120 101.682 -77.872 44.916
215121 101.707 -77.887 44.900
215122 101.732 -77.902 44.884
215123 101.823 -77.905 44.868
215124 101.946 -77.901 44.851
215125 101.952 -77.911 44.835
215126 101.958 -77.921 44.819
215127 101.964 -77.931 44.803
215128 101.970 -77.940 44.786
215129 101.964 -77.957 44.771
215130 101.952 -77.978 44.755
215131 101.925 -77.986 44.741
215132 101.897 -77.990 44.727
215133 101.925 -78.002 44.712
215134 101.981 -78.015 44.696
215135 102.020 -78.028 44.680
215136 102.055 -78.039 44.664
215137 102.090 -78.050 44.648
215138 102.125 -78.061 44.633
215139 102.160 -78.072 44.618
215140 102.195 -78.084 44.603
215141 102.230 -78.094 44.588
215142 102.261 -78.108 44.574
215143 102.291 -78.122 44.559
215144 102.321 -78.135 44.545
215145 102.349 -78.145 44.531
215146 102.371 -78.150 44.517
215147 102.400 -78.172 44.503
215148 102.430 -78.185 44.487
215149 102.463 -78.179 44.472
215150 102.494 -78.173 44.456
215151 102.526 -78.166 44.441
215152 102.558 -78.160 44.426
215153 102.590 -78.154 44.411
215154 102.621 -78.148 44.396
215155 102.653 -78.142 44.382
215156 102.685 -78.135 44.367
215157 102.717 -78.129 44.353
215158 102.732 -78.112 44.338
215159 102.751 -78.144 44.323
215160 102.809 -78.260 44.307
215161 102.851 -78.259 44.292
215162 102.893 -78.257 44.276
215163 102.935 -78.256 44.258
215164 102.976 -78.253 44.243
215165 103.017 -78.257 44.226
215166 103.054 -78.270 44.209
215167 103.091 -78.284 44.191
215168 103.127 -78.298 44.174
215169 103.164 -78.312 44.157
215170 103.201 -78.326 44.140
215171 103.238 -78.339 44.122
215172 103.275 -78.353 44.104
215455 93.333 -75.614 53.914
215456 93.352 -75.618 53.926
215457 93.370 -75.621 53.941
215458 93.389 -75.624 53.959
215459 93.407 -75.627 53.980
215460 93.425 -75.630 54.002
215461 93.445 -75.633 54.025
215462 93.468 -75.631 54.052
215463 93.491 -75.630 54.081
215464 93.515 -75.629 54.115
215465 93.538 -75.627 54.151
215466 93.572 -75.626 54.190
215467 93.620 -75.615 54.221
215468 93.666 -75.601 54.250
215469 93.696 -75.596 54.277
215470 93.726 -75.590 54.300
215471 93.756 -75.585 54.319
215472 93.785 -75.580 54.339
215473 93.815 -75.575 54.356
215474 93.844 -75.570 54.371
215475 93.874 -75.565 54.388
215476 93.904 -75.559 54.401
215477 93.933 -75.554 54.412
215478 93.957 -75.552 54.422
215479 93.979 -75.549 54.427
215480 94.003 -75.549 54.424
215481 94.028 -75.549 54.413
215482 94.052 -75.548 54.392
215483 94.075 -75.546 54.370
215484 94.100 -75.545 54.352
215485 94.124 -75.543 54.340
215486 94.148 -75.541 54.326
215487 94.172 -75.539 54.307
215488 94.196 -75.540 54.286
215489 94.220 -75.541 54.265
215490 94.244 -75.543 54.245
215491 94.268 -75.545 54.227
215492 94.292 -75.546 54.212
215493 94.318 -75.547 54.204
215494 94.343 -75.547 54.196
215495 94.364 -75.550 54.185
215496 94.376 -75.557 54.172
215497 94.387 -75.564 54.155
215498 94.398 -75.571 54.139
215499 94.410 -75.577 54.123
215500 94.421 -75.583 54.104
215501 94.433 -75.590 54.081
215502 94.445 -75.597 54.054
215503 94.456 -75.604 54.024
215504 94.467 -75.610 53.992
215505 94.480 -75.615 53.960
215506 94.494 -75.618 53.929
215507 94.518 -75.618 53.896
215508 94.543 -75.618 53.861
215509 94.569 -75.616 53.828
215510 94.594 -75.615 53.794
215511 94.620 -75.615 53.758
215512 94.645 -75.613 53.720
215513 94.671 -75.612 53.680
215514 94.696 -75.612 53.637
215515 94.716 -75.611 53.594
215516 94.707 -75.614 53.552
215517 94.698 -75.616 53.511
215518 94.688 -75.619 53.473
215519 94.679 -75.621 53.435
215520 94.670 -75.624 53.399
215521 94.669 -75.627 53.361
215522 94.678 -75.630 53.322
215523 94.687 -75.632 53.282
215524 94.696 -75.635 53.243
215525 94.704 -75.637 53.204
215526 94.713 -75.640 53.164
215527 94.722 -75.643 53.122
215528 94.731 -75.646 53.075
215529 94.740 -75.649 53.029
215530 94.749 -75.651 52.984
215531 94.758 -75.654 52.943
215532 94.766 -75.656 52.908
215533 94.775 -75.659 52.881
215534 94.797 -75.664 52.854
215535 94.827 -75.670 52.824
215536 94.857 -75.676 52.790
215537 94.888 -75.682 52.752
215538 94.918 -75.688 52.712
215539 94.947 -75.693 52.671
215540 94.972 -75.697 52.628
215541 94.998 -75.700 52.585
215542 95.027 -75.706 52.543
215543 95.060 -75.717 52.501
215544 95.095 -75.728 52.460
215545 95.129 -75.739 52.420
215546 95.163 -75.749 52.380
215547 95.197 -75.761 52.342
215548 95.241 -75.768 52.304
215549 95.299 -75.769 52.266
215550 95.356 -75.771 52.227
215551 95.413 -75.772 52.187
215552 95.470 -75.774 52.145
215553 95.505 -75.759 52.103
215554 95.517 -75.731 52.061
215555 95.529 -75.702 52.021
215556 95.541 -75.699 51.979
215557 95.552 -75.716 51.937
215558 95.563 -75.732 51.897
215559 95.574 -75.748 51.855
215560 95.586 -75.765 51.815
215561 95.597 -75.781 51.776
215562 95.608 -75.796 51.738
215563 95.619 -75.813 51.700
215564 95.630 -75.829 51.665
215565 95.641 -75.845 51.632
215566 95.652 -75.858 51.600
215567 95.667 -75.864 51.570
215568 95.709 -75.864 51.542
215569 95.759 -75.865 51.514
215570 95.809 -75.865 51.487
215571 95.860 -75.865 51.460
215572 95.906 -75.866 51.431
215573 95.937 -75.869 51.399
215574 95.967 -75.865 51.364
215575 96.005 -75.862 51.329
215576 96.042 -75.857 51.294
215577 96.095 -75.856 51.258
215578 96.172 -75.858 51.219
215579 96.199 -75.859 51.181
215580 96.206 -75.862 51.143
215581 96.209 -75.862 51.105
215582 96.212 -75.862 51.068
215583 96.215 -75.862 51.033
215584 96.218 -75.862 50.997
215585 96.236 -75.868 50.962
215586 96.256 -75.875 50.929
215587 96.276 -75.881 50.897
215588 96.297 -75.888 50.868
215589 96.318 -75.897 50.841
215590 96.340 -75.908 50.815
215591 96.362 -75.919 50.789
215592 96.384 -75.930 50.764
215593 96.406 -75.941 50.739
215594 96.428 -75.953 50.713
215595 96.450 -75.963 50.688
215596 96.471 -75.975 50.664
215597 96.487 -75.990 50.639
215598 96.504 -76.003 50.616
215599 96.520 -76.018 50.593
215600 96.524 -76.036 50.571
215601 96.524 -76.057 50.550
215602 96.523 -76.078 50.530
215603 96.536 -76.090 50.509
215604 96.539 -76.100 50.486
215605 96.558 -76.111 50.461
215606 96.599 -76.120 50.434
215607 96.633 -76.119 50.406
215608 96.663 -76.119 50.376
215609 96.692 -76.117 50.344
215610 96.720 -76.118 50.310
215611 96.746 -76.121 50.274
215612 96.773 -76.123 50.235
215613 96.796 -76.132 50.196
215614 96.815 -76.146 50.155
215615 96.834 -76.160 50.112
215616 96.854 -76.173 50.071
215617 96.873 -76.187 50.031
215618 96.893 -76.201 49.992
215619 96.921 -76.207 49.954
215620 96.949 -76.213 49.917
215621 96.977 -76.220 49.880
215622 97.005 -76.226 49.845
215623 97.033 -76.232 49.811
215624 97.061 -76.238 49.776
215625 97.089 -76.245 49.741
215626 97.119 -76.251 49.706
215627 97.166 -76.264 49.672
215628 97.212 -76.276 49.641
215629 97.218 -76.282 49.611
215630 97.196 -76.279 49.581
215631 97.239 -76.285 49.552
215632 97.281 -76.291 49.523
215633 97.324 -76.297 49.494
215634 97.366 -76.303 49.466
215635 97.409 -76.308 49.437
215636 97.451 -76.318 49.408
215637 97.491 -76.336 49.380
215638 97.524 -76.326 49.352
215639 97.541 -76.333 49.324
215640 97.558 -76.342 49.296
215641 97.574 -76.351 49.267
215642 97.590 -76.361 49.238
215643 97.604 -76.370 49.209
215644 97.608 -76.383 49.182
215645 97.612 -76.397 49.156
215646 97.617 -76.410 49.130
215647 97.621 -76.423 49.105
215648 97.626 -76.436 49.079
215649 97.631 -76.450 49.052
215650 97.654 -76.462 49.024
215651 97.678 -76.473 48.995
215652 97.702 -76.486 48.965
215653 97.725 -76.498 48.935
215654 97.749 -76.509 48.904
215655 97.782 -76.520 48.871
215656 97.851 -76.523 48.837
215657 97.920 -76.527 48.803
215658 97.989 -76.530 48.767
215659 98.058 -76.534 48.730
215660 98.114 -76.541 48.692
215661 98.158 -76.552 48.654
215662 98.202 -76.571 48.618
215663 98.244 -76.596 48.582
215664 98.271 -76.613 48.546
215665 98.262 -76.610 48.510
215666 98.272 -76.609 48.474
215667 98.282 -76.608 48.440
215668 98.291 -76.608 48.407
215669 98.301 -76.608 48.376
215670 98.314 -76.605 48.346
215671 98.337 -76.605 48.317
215672 98.357 -76.609 48.288
215673 98.383 -76.611 48.259
215674 98.412 -76.614 48.232
215675 98.440 -76.618 48.205
215676 98.468 -76.620 48.177
215677 98.497 -76.623 48.150
215678 98.525 -76.625 48.124
215679 98.554 -76.628 48.098
215680 98.583 -76.631 48.071
215681 98.611 -76.633 48.045
215682 98.640 -76.636 48.019
215683 98.668 -76.639 47.992
215684 98.697 -76.642 47.965
215685 98.715 -76.650 47.938
215686 98.722 -76.666 47.910
215687 98.730 -76.681 47.882
215688 98.744 -76.691 47.853
215689 98.759 -76.701 47.822
215690 98.774 -76.711 47.792
215691 98.788 -76.721 47.762
215692 98.802 -76.730 47.730
215693 98.817 -76.740 47.698
215694 98.835 -76.746 47.666
215695 98.854 -76.740 47.634
215696 98.867 -76.729 47.601
215697 98.879 -76.718 47.571
215698 98.891 -76.708 47.540
215699 98.903 -76.697 47.509
215700 98.915 -76.687 47.476
215701 98.928 -76.678 47.442
215702 98.951 -76.688 47.406
215703 98.990 -76.705 47.370
215704 99.038 -76.721 47.334
215705 99.084 -76.737 47.296
215706 99.131 -76.751 47.257
215707 99.178 -76.765 47.218
215708 99.224 -76.781 47.180
215709 99.253 -76.786 47.142
215710 99.271 -76.786 47.105
215711 99.288 -76.786 47.069
215712 99.306 -76.785 47.033
215713 99.324 -76.785 46.996
215714 99.342 -76.785 46.959
215715 99.360 -76.784 46.922
215716 99.377 -76.796 46.885
215717 99.394 -76.808 46.849
215718 99.411 -76.820 46.815
215719 99.428 -76.832 46.782
215720 99.445 -76.845 46.749
215721 99.462 -76.857 46.716
215722 99.479 -76.869 46.684
215723 99.496 -76.881 46.651
215724 99.533 -76.885 46.620
215725 99.556 -76.902 46.588
215726 99.573 -76.925 46.556
215727 99.590 -76.947 46.523
215728 99.601 -76.950 46.493
215729 99.610 -76.947 46.463
215730 99.619 -76.944 46.433
215731 99.628 -76.941 46.407
215732 99.637 -76.938 46.381
215733 99.646 -76.936 46.356
215734 99.655 -76.933 46.331
215735 99.664 -76.930 46.307
215736 99.674 -76.928 46.282
215737 99.683 -76.925 46.258
215738 99.692 -76.922 46.235
215739 99.701 -76.918 46.211
215740 99.710 -76.916 46.187
215741 99.732 -76.919 46.163
215742 99.753 -76.922 46.139
215743 99.784 -76.928 46.114
215744 99.816 -76.935 46.090
215745 99.854 -76.939 46.069
215746 99.896 -76.940 46.047
215747 99.937 -76.940 46.025
215748 99.979 -76.942 46.004
215749 100.017 -76.942 45.984
215750 100.056 -76.941 45.963
215751 100.095 -76.941 45.944
215752 100.132 -76.940 45.926
215753 100.168 -76.934 45.907
215754 100.203 -76.947 45.889
215755 100.231 -76.977 45.872
215756 100.253 -76.992 45.855
215757 100.276 -77.006 45.838
215758 100.300 -77.012 45.822
215759 100.323 -77.022 45.805
215760 100.343 -77.037 45.789
215761 100.363 -77.052 45.771
215762 100.382 -77.069 45.755
215763 100.400 -77.086 45.738
215764 100.418 -77.104 45.721
215765 100.436 -77.122 45.705
215766 100.450 -77.137 45.688
215767 100.459 -77.150 45.673
215768 100.468 -77.163 45.658
215769 100.477 -77.176 45.643
215770 100.486 -77.189 45.631
215771 100.495 -77.203 45.618
215772 100.504 -77.216 45.604
215773 100.519 -77.232 45.592
215774 100.540 -77.250 45.579
215775 100.560 -77.267 45.566
215776 100.580 -77.285 45.553
215777 100.600 -77.304 45.541
215778 100.620 -77.321 45.529
215779 100.640 -77.339 45.517
215780 100.661 -77.357 45.505
215781 100.681 -77.375 45.493
215782 100.701 -77.393 45.482
215783 100.721 -77.410 45.471
215784 100.741 -77.429 45.459
215785 100.761 -77.447 45.448
215786 100.787 -77.458 45.436
215787 100.809 -77.466 45.425
215788 100.830 -77.468 45.413
215789 100.852 -77.470 45.401
215790 100.874 -77.473 45.388
215791 100.896 -77.476 45.375
215792 100.917 -77.478 45.361
215793 100.939 -77.480 45.347
215794 100.966 -77.486 45.332
215795 101.002 -77.502 45.318
215796 101.041 -77.520 45.304
215797 101.079 -77.539 45.290
215798 101.118 -77.556 45.276
215799 101.144 -77.567 45.261
215800 101.170 -77.576 45.246
215801 101.195 -77.586 45.230
215802 101.220 -77.595 45.215
215803 101.246 -77.604 45.199
215804 101.272 -77.614 45.183
215805 101.297 -77.623 45.166
215806 101.323 -77.633 45.150
215807 101.334 -77.645 45.134
215808 101.347 -77.660 45.117
215809 101.378 -77.674 45.101
215810 101.408 -77.687 45.084
215811 101.438 -77.701 45.066
215812 101.463 -77.716 45.049
215813 101.488 -77.731 45.033
215814 101.513 -77.746 45.018
215815 101.537 -77.761 45.001
215816 101.562 -77.777 44.985
215817 101.587 -77.793 44.968
215818 101.612 -77.808 44.951
215819 101.637 -77.823 44.935
215820 101.662 -77.838 44.920
215821 101.687 -77.853 44.904
215822 101.711 -77.868 44.888
215823 101.737 -77.884 44.871
215824 101.833 -77.885 44.855
215825 101.958 -77.881 44.839
215826 101.989 -77.887 44.823
215827 101.995 -77.897 44.807
215828 101.999 -77.909 44.791
215829 101.986 -77.929 44.775
215830 101.974 -77.950 44.760
215831 101.950 -77.959 44.743
215832 101.921 -77.964 44.728
215833 101.892 -77.968 44.714
215834 101.941 -77.982 44.699
215835 101.997 -77.996 44.684
215836 102.043 -78.009 44.667
215837 102.079 -78.020 44.651
215838 102.113 -78.031 44.636
215839 102.148 -78.043 44.621
215840 102.183 -78.054 44.606
215841 102.213 -78.068 44.592
215842 102.243 -78.081 44.577
215843 102.274 -78.095 44.563
215844 102.304 -78.109 44.549
215845 102.334 -78.122 44.535
215846 102.360 -78.130 44.521
215847 102.382 -78.134 44.507
215848 102.407 -78.147 44.491
215849 102.437 -78.170 44.476
215850 102.467 -78.193 44.461
215851 102.498 -78.190 44.445
215852 102.530 -78.184 44.430
215853 102.561 -78.178 44.414
215854 102.594 -78.172 44.399
215855 102.625 -78.166 44.384
215856 102.657 -78.159 44.369
215857 102.689 -78.153 44.355
215858 102.721 -78.147 44.341
215859 102.736 -78.122 44.325
215860 102.761 -78.172 44.310
215861 102.816 -78.258 44.294
215862 102.858 -78.257 44.279
215863 102.899 -78.255 44.263
215864 102.941 -78.253 44.247
215865 102.983 -78.251 44.231
215866 103.025 -78.250 44.214
215867 103.067 -78.248 44.197
215868 103.108 -78.250 44.179
215869 103.144 -78.263 44.162
215870 103.181 -78.277 44.145
215871 103.219 -78.291 44.127
215872 103.255 -78.304 44.110
215873 103.292 -78.318 44.092
216156 93.325 -75.591 53.914
216157 93.344 -75.594 53.925
216158 93.362 -75.597 53.939
216159 93.380 -75.601 53.956
216160 93.399 -75.604 53.976
216161 93.420 -75.604 53.997
216162 93.444 -75.603 54.020
216163 93.467 -75.602 54.046
216164 93.490 -75.600 54.076
216165 93.513 -75.599 54.110
216166 93.537 -75.598 54.147
216167 93.571 -75.596 54.186
216168 93.614 -75.590 54.222
216169 93.667 -75.574 54.255
216170 93.707 -75.562 54.285
216171 93.737 -75.558 54.311
216172 93.767 -75.552 54.331
216173 93.796 -75.547 54.347
216174 93.826 -75.542 54.361
216175 93.855 -75.536 54.374
216176 93.885 -75.531 54.389
216177 93.911 -75.528 54.403
216178 93.933 -75.526 54.416
216179 93.956 -75.524 54.427
216180 93.978 -75.521 54.434
216181 94.002 -75.521 54.433
216182 94.026 -75.521 54.425
216183 94.050 -75.521 54.408
216184 94.074 -75.521 54.391
216185 94.099 -75.521 54.374
216186 94.123 -75.519 54.359
216187 94.147 -75.517 54.347
216188 94.171 -75.515 54.329
216189 94.195 -75.514 54.310
216190 94.219 -75.516 54.289
216191 94.243 -75.518 54.268
216192 94.267 -75.520 54.248
216193 94.293 -75.520 54.232
216194 94.319 -75.521 54.219
216195 94.344 -75.521 54.212
216196 94.366 -75.523 54.203
216197 94.378 -75.530 54.191
216198 94.389 -75.536 54.176
216199 94.400 -75.543 54.160
216200 94.412 -75.549 54.144
216201 94.423 -75.556 54.125
216202 94.435 -75.563 54.100
216203 94.448 -75.568 54.071
216204 94.462 -75.571 54.040
216205 94.476 -75.575 54.006
216206 94.489 -75.579 53.971
216207 94.503 -75.583 53.938
216208 94.521 -75.584 53.904
216209 94.545 -75.583 53.867
216210 94.570 -75.583 53.832
216211 94.596 -75.581 53.798
216212 94.621 -75.580 53.762
216213 94.646 -75.580 53.724
216214 94.672 -75.578 53.684
216215 94.698 -75.577 53.640
216216 94.719 -75.577 53.597
216217 94.709 -75.580 53.555
216218 94.700 -75.582 53.514
216219 94.691 -75.585 53.475
216220 94.681 -75.587 53.438
216221 94.672 -75.590 53.401
216222 94.663 -75.593 53.364
216223 94.661 -75.596 53.325
216224 94.670 -75.598 53.286
216225 94.679 -75.601 53.249
216226 94.687 -75.603 53.212
216227 94.696 -75.606 53.173
216228 94.705 -75.609 53.130
216229 94.714 -75.612 53.083
216230 94.722 -75.615 53.036
216231 94.732 -75.617 52.991
216232 94.741 -75.620 52.949
216233 94.749 -75.622 52.911
216234 94.775 -75.627 52.881
216235 94.805 -75.634 52.853
216236 94.835 -75.640 52.824
216237 94.865 -75.646 52.791
216238 94.896 -75.652 52.755
216239 94.926 -75.659 52.716
216240 94.954 -75.663 52.675
216241 94.979 -75.666 52.631
216242 95.005 -75.670 52.586
216243 95.032 -75.674 52.542
216244 95.066 -75.685 52.497
216245 95.100 -75.696 52.455
216246 95.134 -75.707 52.413
216247 95.168 -75.718 52.372
216248 95.202 -75.728 52.332
216249 95.236 -75.740 52.293
216250 95.277 -75.748 52.254
216251 95.334 -75.749 52.215
216252 95.392 -75.750 52.176
216253 95.449 -75.752 52.136
216254 95.495 -75.746 52.095
216255 95.507 -75.718 52.055
216256 95.520 -75.689 52.015
216257 95.532 -75.660 51.974
216258 95.543 -75.676 51.933
216259 95.554 -75.692 51.892
216260 95.565 -75.708 51.852
216261 95.577 -75.724 51.812
216262 95.588 -75.740 51.773
216263 95.599 -75.756 51.735
216264 95.610 -75.773 51.698
216265 95.621 -75.789 51.664
216266 95.632 -75.805 51.632
216267 95.645 -75.822 51.601
216268 95.676 -75.837 51.573
216269 95.722 -75.840 51.545
216270 95.774 -75.840 51.518
216271 95.825 -75.840 51.494
216272 95.873 -75.841 51.467
216273 95.909 -75.837 51.437
216274 95.947 -75.834 51.404
216275 95.984 -75.829 51.369
216276 96.022 -75.825 51.333
216277 96.060 -75.822 51.298
216278 96.115 -75.821 51.263
216279 96.182 -75.824 51.224
216280 96.202 -75.827 51.186
216281 96.205 -75.827 51.149
216282 96.207 -75.827 51.112
216283 96.210 -75.827 51.075
216284 96.213 -75.827 51.040
216285 96.229 -75.832 51.005
216286 96.249 -75.839 50.971
216287 96.270 -75.845 50.937
216288 96.290 -75.852 50.905
216289 96.310 -75.859 50.876
216290 96.331 -75.867 50.849
216291 96.353 -75.878 50.823
216292 96.375 -75.889 50.798
216293 96.397 -75.900 50.773
216294 96.419 -75.911 50.748
216295 96.442 -75.922 50.721
216296 96.463 -75.934 50.696
216297 96.481 -75.947 50.672
216298 96.498 -75.961 50.648
216299 96.514 -75.975 50.625
216300 96.530 -75.989 50.602
216301 96.540 -76.006 50.579
216302 96.539 -76.026 50.558
216303 96.539 -76.047 50.537
216304 96.538 -76.067 50.516
216305 96.548 -76.082 50.493
216306 96.557 -76.090 50.468
216307 96.593 -76.100 50.442
216308 96.628 -76.099 50.414
216309 96.658 -76.097 50.384
216310 96.684 -76.100 50.353
216311 96.711 -76.103 50.321
216312 96.738 -76.105 50.285
216313 96.764 -76.107 50.247
216314 96.785 -76.120 50.206
216315 96.804 -76.134 50.165
216316 96.823 -76.148 50.123
216317 96.846 -76.159 50.082
216318 96.874 -76.166 50.041
216319 96.901 -76.172 50.002
216320 96.930 -76.178 49.964
216321 96.957 -76.185 49.926
216322 96.985 -76.191 49.890
216323 97.013 -76.197 49.853
216324 97.041 -76.204 49.817
216325 97.069 -76.210 49.782
216326 97.097 -76.216 49.747
216327 97.125 -76.223 49.713
216328 97.163 -76.232 49.679
216329 97.210 -76.245 49.647
216330 97.201 -76.248 49.617
216331 97.217 -76.250 49.588
216332 97.260 -76.256 49.559
216333 97.302 -76.261 49.530
216334 97.345 -76.267 49.502
216335 97.387 -76.273 49.474
216336 97.430 -76.279 49.446
216337 97.472 -76.285 49.418
216338 97.514 -76.296 49.390
216339 97.541 -76.298 49.362
216340 97.557 -76.307 49.335
216341 97.573 -76.315 49.306
216342 97.589 -76.324 49.278
216343 97.605 -76.333 49.249
216344 97.618 -76.344 49.222
216345 97.623 -76.357 49.196
216346 97.628 -76.370 49.169
216347 97.632 -76.383 49.142
216348 97.636 -76.397 49.117
216349 97.641 -76.410 49.091
216350 97.645 -76.423 49.064
216351 97.660 -76.436 49.035
216352 97.683 -76.448 49.005
216353 97.707 -76.460 48.975
216354 97.731 -76.471 48.945
216355 97.755 -76.483 48.914
216356 97.805 -76.490 48.881
216357 97.874 -76.494 48.847
216358 97.943 -76.498 48.812
216359 98.012 -76.501 48.775
216360 98.066 -76.509 48.737
216361 98.110 -76.520 48.699
216362 98.153 -76.532 48.660
216363 98.197 -76.555 48.623
216364 98.240 -76.579 48.586
216365 98.283 -76.604 48.550
216366 98.272 -76.608 48.514
216367 98.275 -76.606 48.478
216368 98.285 -76.605 48.443
216369 98.298 -76.602 48.410
216370 98.319 -76.596 48.378
216371 98.339 -76.589 48.348
216372 98.365 -76.599 48.318
216373 98.388 -76.604 48.289
216374 98.409 -76.608 48.260
216375 98.426 -76.611 48.233
216376 98.448 -76.614 48.205
216377 98.477 -76.616 48.178
216378 98.505 -76.619 48.150
216379 98.533 -76.621 48.125
216380 98.562 -76.624 48.098
216381 98.591 -76.627 48.072
216382 98.619 -76.630 48.045
216383 98.648 -76.633 48.019
216384 98.676 -76.635 47.993
216385 98.705 -76.638 47.967
216386 98.722 -76.648 47.940
216387 98.728 -76.663 47.913
216388 98.735 -76.679 47.886
216389 98.742 -76.695 47.857
216390 98.749 -76.710 47.826
216391 98.763 -76.721 47.796
216392 98.777 -76.731 47.766
216393 98.794 -76.738 47.735
216394 98.819 -76.735 47.703
216395 98.845 -76.731 47.671
216396 98.871 -76.728 47.639
216397 98.884 -76.718 47.606
216398 98.897 -76.708 47.575
216399 98.909 -76.697 47.545
216400 98.921 -76.687 47.514
216401 98.933 -76.676 47.481
216402 98.948 -76.672 47.446
216403 98.968 -76.681 47.410
216404 99.000 -76.696 47.374
216405 99.046 -76.713 47.339
216406 99.094 -76.730 47.300
216407 99.140 -76.746 47.262
216408 99.187 -76.760 47.223
216409 99.233 -76.775 47.184
216410 99.265 -76.782 47.145
216411 99.283 -76.782 47.107
216412 99.301 -76.781 47.070
216413 99.319 -76.781 47.032
216414 99.337 -76.781 46.996
216415 99.355 -76.781 46.959
216416 99.372 -76.781 46.922
216417 99.391 -76.781 46.885
216418 99.408 -76.780 46.850
216419 99.425 -76.793 46.816
216420 99.442 -76.805 46.782
216421 99.460 -76.817 46.750
216422 99.477 -76.829 46.716
216423 99.493 -76.841 46.683
216424 99.517 -76.851 46.651
216425 99.546 -76.863 46.619
216426 99.563 -76.885 46.587
216427 99.580 -76.908 46.554
216428 99.596 -76.928 46.522
216429 99.605 -76.925 46.492
216430 99.614 -76.922 46.462
216431 99.623 -76.919 46.434
216432 99.632 -76.916 46.407
216433 99.641 -76.914 46.382
216434 99.650 -76.911 46.357
216435 99.659 -76.908 46.332
216436 99.668 -76.906 46.307
216437 99.678 -76.903 46.283
216438 99.687 -76.900 46.259
216439 99.696 -76.897 46.236
216440 99.714 -76.898 46.212
216441 99.735 -76.901 46.187
216442 99.757 -76.903 46.162
216443 99.779 -76.906 46.136
216444 99.810 -76.913 46.111
216445 99.842 -76.921 46.087
216446 99.873 -76.928 46.064
216447 99.905 -76.934 46.042
216448 99.946 -76.936 46.020
216449 99.988 -76.937 45.999
216450 100.030 -76.937 45.978
216451 100.071 -76.939 45.957
216452 100.113 -76.940 45.938
216453 100.155 -76.940 45.919
216454 100.195 -76.941 45.900
216455 100.223 -76.953 45.882
216456 100.246 -76.968 45.865
216457 100.269 -76.982 45.848
216458 100.292 -76.994 45.831
216459 100.317 -76.999 45.814
216460 100.339 -77.009 45.798
216461 100.359 -77.025 45.782
216462 100.379 -77.039 45.765
216463 100.397 -77.057 45.749
216464 100.415 -77.075 45.732
216465 100.433 -77.092 45.715
216466 100.451 -77.110 45.697
216467 100.469 -77.128 45.681
216468 100.478 -77.141 45.666
216469 100.487 -77.153 45.651
216470 100.496 -77.167 45.636
216471 100.505 -77.180 45.623
216472 100.514 -77.193 45.609
216473 100.523 -77.207 45.596
216474 100.540 -77.222 45.584
216475 100.560 -77.241 45.571
216476 100.580 -77.259 45.559
216477 100.601 -77.276 45.547
216478 100.621 -77.294 45.534
216479 100.641 -77.313 45.523
216480 100.661 -77.330 45.511
216481 100.681 -77.348 45.499
216482 100.701 -77.366 45.487
216483 100.721 -77.384 45.476
216484 100.742 -77.402 45.464
216485 100.763 -77.420 45.453
216486 100.790 -77.430 45.441
216487 100.817 -77.441 45.430
216488 100.838 -77.451 45.419
216489 100.859 -77.460 45.407
216490 100.881 -77.462 45.395
216491 100.903 -77.464 45.382
216492 100.924 -77.467 45.368
216493 100.946 -77.469 45.354
216494 100.968 -77.471 45.339
216495 100.992 -77.476 45.324
216496 101.021 -77.484 45.309
216497 101.049 -77.492 45.295
216498 101.088 -77.511 45.281
216499 101.126 -77.528 45.266
216500 101.161 -77.544 45.252
216501 101.187 -77.554 45.237
216502 101.212 -77.563 45.221
216503 101.238 -77.573 45.205
216504 101.263 -77.582 45.189
216505 101.288 -77.592 45.173
216506 101.307 -77.603 45.156
216507 101.311 -77.617 45.140
216508 101.314 -77.633 45.124
216509 101.331 -77.647 45.107
216510 101.362 -77.661 45.090
216511 101.393 -77.674 45.073
216512 101.423 -77.688 45.056
216513 101.454 -77.701 45.039
216514 101.485 -77.714 45.023
216515 101.516 -77.728 45.006
216516 101.542 -77.743 44.990
216517 101.567 -77.758 44.973
216518 101.592 -77.773 44.956
216519 101.617 -77.789 44.939
216520 101.642 -77.804 44.923
216521 101.667 -77.819 44.907
216522 101.692 -77.834 44.891
216523 101.716 -77.849 44.876
216524 101.741 -77.865 44.859
216525 101.843 -77.865 44.843
216526 101.969 -77.860 44.827
216527 102.026 -77.864 44.810
216528 102.022 -77.881 44.795
216529 102.009 -77.902 44.780
216530 101.996 -77.922 44.764
216531 101.974 -77.934 44.748
216532 101.945 -77.937 44.732
216533 101.916 -77.942 44.716
216534 101.901 -77.948 44.700
216535 101.958 -77.962 44.686
216536 102.013 -77.977 44.670
216537 102.066 -77.990 44.654
216538 102.101 -78.002 44.638
216539 102.135 -78.013 44.623
216540 102.165 -78.027 44.608
216541 102.196 -78.040 44.594
216542 102.226 -78.054 44.580
216543 102.256 -78.068 44.566
216544 102.287 -78.081 44.553
216545 102.317 -78.095 44.539
216546 102.347 -78.109 44.524
216547 102.371 -78.115 44.510
216548 102.393 -78.119 44.495
216549 102.415 -78.122 44.480
216550 102.444 -78.145 44.464
216551 102.474 -78.168 44.449
216552 102.503 -78.191 44.433
216553 102.534 -78.201 44.417
216554 102.565 -78.195 44.402
216555 102.597 -78.189 44.387
216556 102.629 -78.182 44.372
216557 102.661 -78.176 44.358
216558 102.693 -78.170 44.343
216559 102.725 -78.164 44.328
216560 102.745 -78.145 44.313
216561 102.772 -78.200 44.298
216562 102.822 -78.256 44.282
216563 102.864 -78.254 44.266
216564 102.906 -78.253 44.251
216565 102.948 -78.251 44.235
216566 102.989 -78.249 44.219
216567 103.032 -78.247 44.202
216568 103.074 -78.246 44.185
216569 103.115 -78.244 44.168
216570 103.157 -78.243 44.150
216571 103.198 -78.243 44.132
216572 103.235 -78.257 44.114
216573 103.272 -78.270 44.097
216857 93.317 -75.568 53.916
216858 93.336 -75.571 53.925
216859 93.354 -75.574 53.937
216860 93.372 -75.577 53.953
216861 93.396 -75.576 53.972
216862 93.419 -75.574 53.992
216863 93.442 -75.574 54.015
216864 93.466 -75.572 54.041
216865 93.489 -75.571 54.070
216866 93.512 -75.570 54.104
216867 93.536 -75.568 54.140
216868 93.570 -75.567 54.180
216869 93.609 -75.565 54.219
216870 93.661 -75.549 54.257
216871 93.714 -75.532 54.292
216872 93.748 -75.524 54.322
216873 93.778 -75.519 54.345
216874 93.807 -75.514 54.361
216875 93.837 -75.508 54.372
216876 93.865 -75.504 54.383
216877 93.887 -75.502 54.394
216878 93.910 -75.500 54.409
216879 93.932 -75.498 54.424
216880 93.955 -75.496 54.434
216881 93.977 -75.494 54.441
216882 94.001 -75.494 54.441
216883 94.025 -75.493 54.434
216884 94.049 -75.493 54.423
216885 94.073 -75.493 54.409
216886 94.097 -75.493 54.394
216887 94.121 -75.493 54.378
216888 94.145 -75.493 54.366
216889 94.170 -75.492 54.350
216890 94.194 -75.489 54.332
216891 94.218 -75.491 54.312
216892 94.242 -75.493 54.289
216893 94.268 -75.493 54.268
216894 94.293 -75.493 54.250
216895 94.319 -75.494 54.236
216896 94.345 -75.494 54.228
216897 94.368 -75.496 54.218
216898 94.380 -75.502 54.207
216899 94.391 -75.509 54.193
216900 94.403 -75.516 54.177
216901 94.416 -75.520 54.162
216902 94.429 -75.524 54.145
216903 94.443 -75.527 54.117
216904 94.457 -75.531 54.088
216905 94.471 -75.535 54.055
216906 94.484 -75.539 54.021
216907 94.498 -75.543 53.985
216908 94.511 -75.546 53.949
216909 94.529 -75.548 53.912
216910 94.546 -75.549 53.873
216911 94.571 -75.549 53.836
216912 94.597 -75.548 53.800
216913 94.622 -75.546 53.764
216914 94.648 -75.546 53.726
216915 94.673 -75.545 53.687
216916 94.699 -75.544 53.644
216917 94.721 -75.543 53.600
216918 94.712 -75.546 53.558
216919 94.702 -75.549 53.519
216920 94.693 -75.551 53.480
216921 94.684 -75.554 53.442
216922 94.674 -75.556 53.404
216923 94.665 -75.559 53.366
216924 94.655 -75.561 53.328
216925 94.652 -75.565 53.292
216926 94.662 -75.567 53.257
216927 94.670 -75.570 53.219
216928 94.679 -75.572 53.179
216929 94.688 -75.575 53.137
216930 94.697 -75.577 53.089
216931 94.705 -75.580 53.043
216932 94.714 -75.583 52.998
216933 94.723 -75.586 52.954
216934 94.753 -75.592 52.915
216935 94.783 -75.598 52.880
216936 94.814 -75.604 52.850
216937 94.844 -75.610 52.821
216938 94.874 -75.616 52.789
216939 94.904 -75.622 52.755
216940 94.934 -75.628 52.718
216941 94.961 -75.632 52.676
216942 94.986 -75.636 52.632
216943 95.012 -75.639 52.586
216944 95.038 -75.643 52.539
216945 95.071 -75.652 52.494
216946 95.105 -75.664 52.450
216947 95.139 -75.674 52.408
216948 95.174 -75.686 52.365
216949 95.207 -75.696 52.324
216950 95.242 -75.707 52.283
216951 95.276 -75.718 52.244
216952 95.313 -75.728 52.205
216953 95.370 -75.729 52.166
216954 95.427 -75.731 52.127
216955 95.485 -75.732 52.087
216956 95.498 -75.704 52.047
216957 95.510 -75.676 52.009
216958 95.522 -75.647 51.969
216959 95.534 -75.636 51.928
216960 95.545 -75.652 51.887
216961 95.556 -75.668 51.848
216962 95.567 -75.684 51.808
216963 95.579 -75.700 51.769
216964 95.590 -75.716 51.732
216965 95.601 -75.733 51.697
216966 95.616 -75.751 51.664
216967 95.640 -75.784 51.632
216968 95.659 -75.812 51.602
216969 95.700 -75.810 51.573
216970 95.726 -75.815 51.546
216971 95.778 -75.815 51.520
216972 95.830 -75.815 51.495
216973 95.878 -75.817 51.468
216974 95.905 -75.817 51.438
216975 95.948 -75.809 51.405
216976 95.991 -75.801 51.370
216977 96.034 -75.793 51.334
216978 96.077 -75.785 51.299
216979 96.122 -75.783 51.265
216980 96.168 -75.781 51.227
216981 96.217 -75.775 51.190
216982 96.203 -75.792 51.154
216983 96.206 -75.792 51.117
216984 96.209 -75.792 51.081
216985 96.223 -75.796 51.046
216986 96.243 -75.803 51.012
216987 96.263 -75.809 50.979
216988 96.284 -75.816 50.945
216989 96.304 -75.823 50.913
216990 96.324 -75.830 50.884
216991 96.344 -75.837 50.857
216992 96.367 -75.847 50.831
216993 96.389 -75.859 50.806
216994 96.411 -75.870 50.780
216995 96.433 -75.881 50.755
216996 96.455 -75.892 50.729
216997 96.475 -75.904 50.703
216998 96.491 -75.918 50.680
216999 96.508 -75.932 50.656
217000 96.524 -75.947 50.633
217001 96.540 -75.960 50.610
217002 96.555 -75.975 50.588
217003 96.555 -75.995 50.565
217004 96.554 -76.016 50.544
217005 96.554 -76.037 50.522
217006 96.553 -76.057 50.500
217007 96.559 -76.074 50.476
217008 96.587 -76.079 50.450
217009 96.622 -76.079 50.422
217010 96.649 -76.082 50.393
217011 96.676 -76.084 50.364
217012 96.703 -76.087 50.331
217013 96.730 -76.089 50.296
217014 96.754 -76.095 50.258
217015 96.773 -76.109 50.217
217016 96.798 -76.118 50.175
217017 96.826 -76.124 50.133
217018 96.854 -76.131 50.092
217019 96.882 -76.137 50.051
217020 96.910 -76.143 50.011
217021 96.938 -76.150 49.972
217022 96.966 -76.156 49.934
217023 96.994 -76.162 49.897
217024 97.022 -76.169 49.860
217025 97.049 -76.175 49.824
217026 97.078 -76.181 49.788
217027 97.105 -76.188 49.753
217028 97.133 -76.194 49.719
217029 97.161 -76.200 49.685
217030 97.208 -76.213 49.653
217031 97.195 -76.214 49.623
217032 97.238 -76.220 49.594
217033 97.281 -76.226 49.565
217034 97.323 -76.232 49.536
217035 97.366 -76.238 49.509
217036 97.408 -76.244 49.482
217037 97.451 -76.250 49.455
217038 97.494 -76.256 49.427
217039 97.533 -76.263 49.399
217040 97.557 -76.270 49.371
217041 97.573 -76.279 49.344
217042 97.589 -76.289 49.315
217043 97.605 -76.298 49.287
217044 97.621 -76.307 49.260
217045 97.633 -76.317 49.233
217046 97.638 -76.330 49.207
217047 97.642 -76.344 49.181
217048 97.647 -76.357 49.154
217049 97.651 -76.370 49.128
217050 97.656 -76.383 49.101
217051 97.660 -76.397 49.074
217052 97.665 -76.410 49.044
217053 97.689 -76.422 49.015
217054 97.713 -76.434 48.984
217055 97.737 -76.445 48.954
217056 97.760 -76.458 48.922
217057 97.829 -76.461 48.890
217058 97.898 -76.465 48.856
217059 97.967 -76.468 48.820
217060 98.018 -76.477 48.783
217061 98.062 -76.489 48.744
217062 98.105 -76.500 48.704
217063 98.149 -76.514 48.665
217064 98.192 -76.539 48.626
217065 98.235 -76.563 48.590
217066 98.278 -76.588 48.553
217067 98.295 -76.599 48.517
217068 98.283 -76.600 48.482
217069 98.303 -76.593 48.447
217070 98.323 -76.586 48.412
217071 98.344 -76.580 48.379
217072 98.364 -76.572 48.348
217073 98.393 -76.593 48.319
217074 98.416 -76.598 48.290
217075 98.439 -76.603 48.260
217076 98.460 -76.607 48.232
217077 98.478 -76.609 48.205
217078 98.496 -76.612 48.178
217079 98.513 -76.615 48.151
217080 98.541 -76.618 48.124
217081 98.570 -76.621 48.098
217082 98.598 -76.623 48.071
217083 98.627 -76.626 48.045
217084 98.656 -76.629 48.019
217085 98.684 -76.631 47.993
217086 98.713 -76.634 47.967
217087 98.728 -76.645 47.941
217088 98.735 -76.661 47.914
217089 98.741 -76.676 47.888
217090 98.748 -76.692 47.860
217091 98.756 -76.708 47.831
217092 98.764 -76.721 47.802
217093 98.779 -76.726 47.772
217094 98.804 -76.724 47.740
217095 98.830 -76.721 47.707
217096 98.856 -76.717 47.675
217097 98.881 -76.714 47.642
217098 98.902 -76.708 47.610
217099 98.914 -76.697 47.579
217100 98.926 -76.687 47.549
217101 98.939 -76.676 47.518
217102 98.951 -76.666 47.486
217103 98.967 -76.666 47.452
217104 98.987 -76.675 47.415
217105 99.016 -76.688 47.378
217106 99.053 -76.704 47.341
217107 99.101 -76.722 47.303
217108 99.150 -76.740 47.265
217109 99.196 -76.755 47.226
217110 99.243 -76.769 47.186
217111 99.278 -76.778 47.148
217112 99.296 -76.778 47.110
217113 99.313 -76.777 47.073
217114 99.332 -76.777 47.034
217115 99.349 -76.777 46.997
217116 99.367 -76.777 46.961
217117 99.385 -76.777 46.923
217118 99.403 -76.777 46.886
217119 99.421 -76.776 46.851
217120 99.439 -76.776 46.817
217121 99.456 -76.777 46.783
217122 99.474 -76.789 46.750
217123 99.491 -76.801 46.717
217124 99.508 -76.814 46.683
217125 99.535 -76.824 46.650
217126 99.552 -76.846 46.619
217127 99.569 -76.868 46.587
217128 99.586 -76.891 46.554
217129 99.600 -76.903 46.522
217130 99.610 -76.900 46.492
217131 99.619 -76.897 46.463
217132 99.628 -76.894 46.435
217133 99.637 -76.892 46.408
217134 99.646 -76.889 46.383
217135 99.654 -76.886 46.358
217136 99.663 -76.884 46.334
217137 99.673 -76.881 46.309
217138 99.682 -76.878 46.285
217139 99.696 -76.877 46.261
217140 99.717 -76.880 46.237
217141 99.739 -76.883 46.213
217142 99.761 -76.885 46.187
217143 99.782 -76.888 46.161
217144 99.805 -76.891 46.134
217145 99.837 -76.899 46.108
217146 99.868 -76.906 46.084
217147 99.900 -76.912 46.060
217148 99.931 -76.919 46.038
217149 99.962 -76.927 46.015
217150 99.997 -76.932 45.993
217151 100.039 -76.933 45.972
217152 100.080 -76.934 45.951
217153 100.122 -76.935 45.932
217154 100.164 -76.936 45.912
217155 100.206 -76.937 45.893
217156 100.239 -76.943 45.876
217157 100.262 -76.958 45.858
217158 100.284 -76.973 45.841
217159 100.308 -76.981 45.824
217160 100.333 -76.986 45.807
217161 100.354 -76.997 45.791
217162 100.375 -77.012 45.775
217163 100.395 -77.028 45.759
217164 100.412 -77.045 45.742
217165 100.430 -77.063 45.725
217166 100.449 -77.081 45.708
217167 100.467 -77.098 45.691
217168 100.484 -77.116 45.675
217169 100.497 -77.132 45.660
217170 100.506 -77.144 45.644
217171 100.515 -77.157 45.629
217172 100.525 -77.170 45.615
217173 100.534 -77.184 45.602
217174 100.543 -77.197 45.589
217175 100.561 -77.214 45.576
217176 100.581 -77.232 45.563
217177 100.602 -77.250 45.551
217178 100.622 -77.268 45.539
217179 100.642 -77.285 45.526
217180 100.662 -77.304 45.515
217181 100.682 -77.321 45.503
217182 100.702 -77.339 45.492
217183 100.722 -77.357 45.480
217184 100.743 -77.375 45.469
217185 100.765 -77.391 45.457
217186 100.792 -77.401 45.446
217187 100.819 -77.412 45.435
217188 100.846 -77.423 45.424
217189 100.868 -77.434 45.413
217190 100.889 -77.444 45.401
217191 100.909 -77.454 45.388
217192 100.931 -77.456 45.375
217193 100.953 -77.458 45.361
217194 100.975 -77.461 45.347
217195 100.997 -77.463 45.332
217196 101.019 -77.466 45.316
217197 101.048 -77.474 45.301
217198 101.075 -77.482 45.286
217199 101.104 -77.490 45.271
217200 101.134 -77.501 45.256
217201 101.173 -77.518 45.241
217202 101.203 -77.531 45.226
217203 101.229 -77.541 45.211
217204 101.255 -77.550 45.195
217205 101.279 -77.560 45.179
217206 101.283 -77.575 45.162
217207 101.286 -77.590 45.146
217208 101.290 -77.605 45.130
217209 101.294 -77.620 45.113
217210 101.316 -77.634 45.096
217211 101.346 -77.648 45.078
217212 101.377 -77.661 45.061
217213 101.408 -77.675 45.045
217214 101.439 -77.688 45.029
217215 101.469 -77.702 45.012
217216 101.500 -77.715 44.995
217217 101.531 -77.729 44.978
217218 101.561 -77.742 44.962
217219 101.592 -77.755 44.945
217220 101.622 -77.769 44.928
217221 101.646 -77.785 44.911
217222 101.672 -77.800 44.895
217223 101.696 -77.815 44.879
217224 101.722 -77.830 44.864
217225 101.746 -77.846 44.848
217226 101.854 -77.845 44.831
217227 101.979 -77.840 44.814
217228 102.017 -77.849 44.798
217229 102.032 -77.874 44.783
217230 102.019 -77.894 44.768
217231 101.998 -77.907 44.752
217232 101.974 -77.911 44.736
217233 101.961 -77.915 44.719
217234 101.972 -77.911 44.703
217235 102.013 -77.898 44.688
217236 102.051 -77.906 44.673
217237 102.075 -77.932 44.656
217238 102.082 -77.965 44.641
217239 102.118 -77.987 44.627
217240 102.148 -78.000 44.612
217241 102.178 -78.014 44.597
217242 102.208 -78.028 44.583
217243 102.239 -78.041 44.568
217244 102.269 -78.055 44.554
217245 102.300 -78.069 44.541
217246 102.330 -78.082 44.527
217247 102.360 -78.096 44.512
217248 102.382 -78.100 44.498
217249 102.404 -78.103 44.483
217250 102.426 -78.107 44.467
217251 102.451 -78.120 44.452
217252 102.481 -78.143 44.436
217253 102.511 -78.166 44.420
217254 102.540 -78.188 44.404
217255 102.570 -78.210 44.390
217256 102.601 -78.206 44.375
217257 102.633 -78.200 44.360
217258 102.665 -78.194 44.346
217259 102.696 -78.188 44.331
217260 102.728 -78.181 44.316
217261 102.754 -78.168 44.301
217262 102.783 -78.227 44.286
217263 102.829 -78.253 44.270
217264 102.871 -78.252 44.254
217265 102.912 -78.250 44.238
217266 102.954 -78.249 44.223
217267 102.996 -78.247 44.206
217268 103.038 -78.245 44.189
217269 103.080 -78.244 44.172
217270 103.122 -78.242 44.154
217271 103.164 -78.241 44.138
217272 103.206 -78.239 44.120
217273 103.247 -78.237 44.102
217274 103.289 -78.235 44.085
217559 93.328 -75.548 53.925
217560 93.348 -75.549 53.936
217561 93.371 -75.548 53.951
217562 93.395 -75.546 53.969
217563 93.418 -75.545 53.988
217564 93.441 -75.544 54.010
217565 93.465 -75.543 54.034
217566 93.488 -75.541 54.063
217567 93.511 -75.540 54.095
217568 93.535 -75.539 54.130
217569 93.569 -75.537 54.168
217570 93.608 -75.536 54.208
217571 93.656 -75.524 54.248
217572 93.709 -75.507 54.288
217573 93.759 -75.491 54.324
217574 93.789 -75.486 54.352
217575 93.819 -75.480 54.373
217576 93.841 -75.478 54.390
217577 93.864 -75.477 54.402
217578 93.886 -75.474 54.411
217579 93.908 -75.472 54.422
217580 93.931 -75.470 54.435
217581 93.953 -75.468 54.447
217582 93.976 -75.466 54.451
217583 94.000 -75.466 54.448
217584 94.024 -75.466 54.439
217585 94.048 -75.466 54.437
217586 94.072 -75.466 54.424
217587 94.096 -75.465 54.409
217588 94.120 -75.465 54.398
217589 94.144 -75.465 54.384
217590 94.168 -75.465 54.369
217591 94.192 -75.465 54.352
217592 94.217 -75.465 54.332
217593 94.243 -75.466 54.308
217594 94.269 -75.466 54.287
217595 94.294 -75.467 54.268
217596 94.320 -75.467 54.253
217597 94.345 -75.467 54.243
217598 94.370 -75.468 54.233
217599 94.384 -75.473 54.221
217600 94.397 -75.477 54.206
217601 94.411 -75.480 54.192
217602 94.425 -75.484 54.178
217603 94.438 -75.488 54.161
217604 94.452 -75.492 54.133
217605 94.466 -75.496 54.102
217606 94.479 -75.499 54.069
217607 94.493 -75.503 54.034
217608 94.507 -75.507 53.997
217609 94.521 -75.511 53.960
217610 94.537 -75.512 53.921
217611 94.555 -75.514 53.880
217612 94.572 -75.514 53.842
217613 94.598 -75.514 53.804
217614 94.624 -75.513 53.767
217615 94.649 -75.512 53.730
217616 94.675 -75.511 53.691
217617 94.700 -75.510 53.650
217618 94.723 -75.509 53.606
217619 94.714 -75.511 53.564
217620 94.705 -75.514 53.525
217621 94.695 -75.517 53.486
217622 94.686 -75.520 53.447
217623 94.677 -75.522 53.409
217624 94.667 -75.525 53.371
217625 94.657 -75.527 53.334
217626 94.648 -75.530 53.298
217627 94.644 -75.533 53.263
217628 94.653 -75.536 53.224
217629 94.662 -75.538 53.184
217630 94.671 -75.541 53.141
217631 94.680 -75.543 53.093
217632 94.688 -75.546 53.046
217633 94.701 -75.549 53.001
217634 94.731 -75.555 52.958
217635 94.761 -75.561 52.917
217636 94.791 -75.568 52.880
217637 94.822 -75.574 52.845
217638 94.852 -75.580 52.812
217639 94.882 -75.587 52.780
217640 94.912 -75.593 52.746
217641 94.942 -75.598 52.709
217642 94.968 -75.602 52.668
217643 94.993 -75.605 52.625
217644 95.019 -75.609 52.580
217645 95.045 -75.612 52.534
217646 95.076 -75.621 52.489
217647 95.111 -75.631 52.444
217648 95.145 -75.643 52.401
217649 95.179 -75.653 52.358
217650 95.213 -75.665 52.316
217651 95.247 -75.675 52.275
217652 95.281 -75.686 52.235
217653 95.315 -75.697 52.195
217654 95.349 -75.708 52.156
217655 95.406 -75.709 52.117
217656 95.463 -75.711 52.077
217657 95.488 -75.691 52.038
217658 95.500 -75.662 52.000
217659 95.513 -75.634 51.961
217660 95.525 -75.605 51.921
217661 95.536 -75.612 51.882
217662 95.547 -75.627 51.843
217663 95.559 -75.644 51.804
217664 95.570 -75.660 51.766
217665 95.587 -75.680 51.729
217666 95.617 -75.708 51.695
217667 95.629 -75.750 51.662
217668 95.641 -75.785 51.630
217669 95.683 -75.784 51.599
217670 95.725 -75.782 51.570
217671 95.767 -75.781 51.543
217672 95.782 -75.790 51.517
217673 95.834 -75.790 51.490
217674 95.886 -75.791 51.463
217675 95.900 -75.800 51.433
217676 95.932 -75.798 51.401
217677 95.975 -75.790 51.365
217678 96.018 -75.782 51.330
217679 96.061 -75.772 51.297
217680 96.108 -75.755 51.264
217681 96.154 -75.738 51.228
217682 96.200 -75.737 51.193
217683 96.206 -75.753 51.157
217684 96.205 -75.757 51.122
217685 96.216 -75.760 51.088
217686 96.237 -75.767 51.053
217687 96.257 -75.774 51.019
217688 96.277 -75.781 50.986
217689 96.298 -75.787 50.952
217690 96.318 -75.793 50.920
217691 96.338 -75.800 50.891
217692 96.359 -75.807 50.864
217693 96.380 -75.817 50.838
217694 96.402 -75.828 50.812
217695 96.424 -75.840 50.787
217696 96.446 -75.850 50.761
217697 96.468 -75.862 50.736
217698 96.485 -75.875 50.711
217699 96.502 -75.890 50.686
217700 96.518 -75.903 50.664
217701 96.534 -75.918 50.641
217702 96.550 -75.932 50.618
217703 96.566 -75.946 50.596
217704 96.571 -75.965 50.573
217705 96.570 -75.985 50.551
217706 96.570 -76.006 50.529
217707 96.570 -76.026 50.506
217708 96.590 -76.041 50.483
217709 96.659 -76.041 50.458
217710 96.700 -76.033 50.431
217711 96.706 -76.033 50.403
217712 96.716 -76.044 50.375
217713 96.726 -76.055 50.342
217714 96.736 -76.066 50.307
217715 96.751 -76.077 50.270
217716 96.779 -76.083 50.228
217717 96.806 -76.090 50.186
217718 96.834 -76.096 50.143
217719 96.862 -76.102 50.102
217720 96.890 -76.109 50.061
217721 96.918 -76.115 50.020
217722 96.946 -76.121 49.979
217723 96.974 -76.127 49.942
217724 97.002 -76.134 49.904
217725 97.030 -76.140 49.866
217726 97.058 -76.146 49.829
217727 97.086 -76.153 49.792
217728 97.113 -76.159 49.758
217729 97.142 -76.165 49.724
217730 97.169 -76.172 49.691
217731 97.205 -76.180 49.660
217732 97.217 -76.185 49.629
217733 97.259 -76.191 49.599
217734 97.302 -76.197 49.569
217735 97.344 -76.203 49.541
217736 97.387 -76.209 49.515
217737 97.426 -76.217 49.489
217738 97.464 -76.226 49.462
217739 97.502 -76.235 49.434
217740 97.532 -76.236 49.407
217741 97.565 -76.243 49.379
217742 97.588 -76.253 49.351
217743 97.604 -76.262 49.323
217744 97.621 -76.271 49.296
217745 97.637 -76.279 49.270
217746 97.648 -76.291 49.243
217747 97.653 -76.304 49.218
217748 97.657 -76.317 49.192
217749 97.661 -76.330 49.165
217750 97.666 -76.344 49.138
217751 97.670 -76.357 49.110
217752 97.675 -76.370 49.082
217753 97.679 -76.383 49.052
217754 97.695 -76.396 49.023
217755 97.718 -76.408 48.992
217756 97.742 -76.420 48.961
217757 97.784 -76.429 48.929
217758 97.853 -76.432 48.898
217759 97.922 -76.436 48.864
217760 97.970 -76.445 48.828
217761 98.013 -76.457 48.789
217762 98.057 -76.468 48.749
217763 98.100 -76.480 48.710
217764 98.144 -76.498 48.671
217765 98.187 -76.522 48.630
217766 98.230 -76.547 48.592
217767 98.281 -76.565 48.555
217768 98.266 -76.567 48.519
217769 98.288 -76.574 48.483
217770 98.319 -76.584 48.447
217771 98.349 -76.570 48.413
217772 98.369 -76.563 48.380
217773 98.399 -76.583 48.348
217774 98.422 -76.587 48.318
217775 98.445 -76.592 48.288
217776 98.467 -76.597 48.259
217777 98.490 -76.602 48.230
217778 98.511 -76.605 48.203
217779 98.529 -76.608 48.176
217780 98.547 -76.611 48.149
217781 98.565 -76.614 48.121
217782 98.583 -76.617 48.095
217783 98.606 -76.619 48.069
217784 98.635 -76.622 48.042
217785 98.664 -76.625 48.017
217786 98.692 -76.627 47.992
217787 98.721 -76.630 47.967
217788 98.734 -76.643 47.942
217789 98.741 -76.658 47.916
217790 98.748 -76.674 47.889
217791 98.756 -76.688 47.862
217792 98.768 -76.693 47.833
217793 98.781 -76.699 47.804
217794 98.795 -76.704 47.774
217795 98.816 -76.707 47.743
217796 98.841 -76.706 47.710
217797 98.866 -76.702 47.678
217798 98.892 -76.699 47.646
217799 98.918 -76.696 47.615
217800 98.932 -76.687 47.585
217801 98.944 -76.677 47.555
217802 98.956 -76.666 47.523
217803 98.969 -76.655 47.491
217804 98.987 -76.660 47.457
217805 99.006 -76.669 47.420
217806 99.033 -76.681 47.382
217807 99.065 -76.695 47.344
217808 99.109 -76.712 47.306
217809 99.157 -76.730 47.268
217810 99.206 -76.749 47.230
217811 99.253 -76.764 47.190
217812 99.290 -76.774 47.151
217813 99.308 -76.774 47.113
217814 99.326 -76.774 47.075
217815 99.344 -76.774 47.037
217816 99.362 -76.774 47.000
217817 99.380 -76.773 46.962
217818 99.398 -76.773 46.925
217819 99.416 -76.773 46.888
217820 99.433 -76.772 46.853
217821 99.451 -76.772 46.818
217822 99.469 -76.772 46.784
217823 99.487 -76.771 46.750
217824 99.505 -76.774 46.717
217825 99.525 -76.786 46.684
217826 99.542 -76.808 46.650
217827 99.559 -76.830 46.619
217828 99.576 -76.853 46.587
217829 99.593 -76.875 46.555
217830 99.605 -76.878 46.522
217831 99.614 -76.875 46.493
217832 99.623 -76.872 46.464
217833 99.632 -76.870 46.436
217834 99.641 -76.867 46.409
217835 99.650 -76.864 46.385
217836 99.659 -76.862 46.360
217837 99.668 -76.859 46.336
217838 99.678 -76.856 46.312
217839 99.700 -76.859 46.287
217840 99.721 -76.862 46.263
217841 99.743 -76.865 46.239
217842 99.764 -76.867 46.213
217843 99.786 -76.870 46.187
217844 99.807 -76.872 46.160
217845 99.832 -76.877 46.132
217846 99.863 -76.884 46.105
217847 99.894 -76.890 46.080
217848 99.926 -76.897 46.056
217849 99.957 -76.905 46.033
217850 99.988 -76.912 46.010
217851 100.020 -76.918 45.988
217852 100.051 -76.925 45.966
217853 100.089 -76.929 45.945
217854 100.131 -76.930 45.925
217855 100.173 -76.931 45.906
217856 100.214 -76.932 45.887
217857 100.254 -76.934 45.869
217858 100.277 -76.949 45.851
217859 100.299 -76.963 45.834
217860 100.324 -76.969 45.817
217861 100.349 -76.972 45.801
217862 100.370 -76.984 45.785
217863 100.391 -76.999 45.769
217864 100.410 -77.016 45.752
217865 100.427 -77.034 45.735
217866 100.445 -77.051 45.718
217867 100.464 -77.069 45.701
217868 100.482 -77.087 45.685
217869 100.499 -77.104 45.669
217870 100.517 -77.122 45.653
217871 100.526 -77.135 45.637
217872 100.535 -77.148 45.622
217873 100.544 -77.161 45.607
217874 100.553 -77.174 45.594
217875 100.562 -77.188 45.581
217876 100.582 -77.205 45.567
217877 100.602 -77.223 45.554
217878 100.622 -77.241 45.542
217879 100.642 -77.259 45.530
217880 100.662 -77.277 45.517
217881 100.683 -77.294 45.506
217882 100.703 -77.313 45.496
217883 100.723 -77.330 45.485
217884 100.743 -77.348 45.473
217885 100.767 -77.363 45.462
217886 100.794 -77.373 45.450
217887 100.822 -77.384 45.439
217888 100.849 -77.395 45.429
217889 100.876 -77.405 45.418
217890 100.898 -77.416 45.406
217891 100.919 -77.426 45.395
217892 100.939 -77.437 45.382
217893 100.960 -77.447 45.369
217894 100.982 -77.449 45.355
217895 101.003 -77.452 45.340
217896 101.025 -77.454 45.324
217897 101.047 -77.457 45.308
217898 101.075 -77.464 45.293
217899 101.103 -77.472 45.278
217900 101.130 -77.480 45.263
217901 101.159 -77.488 45.247
217902 101.186 -77.495 45.231
217903 101.219 -77.508 45.216
217904 101.246 -77.519 45.201
217905 101.255 -77.533 45.185
217906 101.259 -77.547 45.168
217907 101.263 -77.562 45.152
217908 101.267 -77.577 45.135
217909 101.270 -77.592 45.118
217910 101.274 -77.608 45.102
217911 101.300 -77.621 45.084
217912 101.331 -77.635 45.067
217913 101.361 -77.649 45.050
217914 101.392 -77.662 45.034
217915 101.423 -77.675 45.018
217916 101.454 -77.689 45.001
217917 101.484 -77.702 44.984
217918 101.515 -77.716 44.967
217919 101.546 -77.729 44.950
217920 101.576 -77.743 44.933
217921 101.607 -77.756 44.916
217922 101.638 -77.770 44.900
217923 101.669 -77.783 44.884
217924 101.699 -77.796 44.867
217925 101.726 -77.812 44.851
217926 101.751 -77.827 44.835
217927 101.845 -77.819 44.819
217928 101.934 -77.812 44.803
217929 101.957 -77.832 44.787
217930 101.983 -77.859 44.772
217931 102.009 -77.884 44.756
217932 101.998 -77.881 44.740
217933 102.004 -77.884 44.723
217934 102.009 -77.878 44.706
217935 102.021 -77.873 44.692
217936 102.042 -77.882 44.677
217937 102.036 -77.906 44.661
217938 102.050 -77.917 44.645
217939 102.081 -77.933 44.630
217940 102.130 -77.974 44.615
217941 102.161 -77.987 44.600
217942 102.191 -78.001 44.585
217943 102.222 -78.015 44.571
217944 102.252 -78.028 44.556
217945 102.282 -78.042 44.541
217946 102.313 -78.056 44.528
217947 102.343 -78.069 44.514
217948 102.372 -78.081 44.499
217949 102.393 -78.084 44.485
217950 102.415 -78.088 44.470
217951 102.437 -78.092 44.454
217952 102.459 -78.096 44.439
217953 102.488 -78.118 44.422
217954 102.518 -78.141 44.407
217955 102.547 -78.163 44.392
217956 102.577 -78.185 44.377
217957 102.606 -78.208 44.362
217958 102.636 -78.217 44.348
217959 102.669 -78.211 44.333
217960 102.700 -78.205 44.318
217961 102.732 -78.198 44.303
217962 102.763 -78.191 44.288
217963 102.794 -78.253 44.273
217964 102.835 -78.251 44.258
217965 102.877 -78.250 44.242
217966 102.919 -78.248 44.226
217967 102.961 -78.247 44.210
217968 103.003 -78.244 44.194
217969 103.045 -78.243 44.176
217970 103.087 -78.241 44.159
217971 103.128 -78.240 44.142
217972 103.170 -78.238 44.125
217973 103.212 -78.237 44.108
217974 103.254 -78.235 44.090
217975 103.296 -78.233 44.073
218218 92.394 -75.523 54.153
218219 92.418 -75.533 54.161
218220 92.440 -75.545 54.168
218221 92.454 -75.557 54.173
218222 92.465 -75.562 54.175
218223 92.486 -75.554 54.177
218224 92.506 -75.545 54.181
218260 93.324 -75.521 53.925
218261 93.347 -75.519 53.935
218262 93.370 -75.518 53.949
218263 93.394 -75.517 53.966
218264 93.417 -75.515 53.984
218265 93.440 -75.514 54.005
218266 93.464 -75.513 54.028
218267 93.487 -75.512 54.056
218268 93.510 -75.511 54.087
218269 93.534 -75.509 54.120
218270 93.568 -75.508 54.157
218271 93.607 -75.506 54.195
218272 93.650 -75.499 54.235
218273 93.703 -75.483 54.277
218274 93.756 -75.465 54.317
218275 93.795 -75.455 54.349
218276 93.818 -75.453 54.377
218277 93.840 -75.451 54.399
218278 93.862 -75.449 54.413
218279 93.885 -75.446 54.424
218280 93.907 -75.444 54.437
218281 93.929 -75.442 54.450
218282 93.952 -75.440 54.462
218283 93.974 -75.438 54.467
218284 93.998 -75.438 54.464
218285 94.023 -75.438 54.454
218286 94.047 -75.438 54.447
218287 94.071 -75.438 54.437
218288 94.095 -75.438 54.426
218289 94.119 -75.438 54.415
218290 94.142 -75.437 54.402
218291 94.166 -75.436 54.387
218292 94.189 -75.436 54.369
218293 94.215 -75.442 54.348
218294 94.241 -75.444 54.325
218295 94.267 -75.445 54.304
218296 94.292 -75.445 54.284
218297 94.318 -75.446 54.268
218298 94.343 -75.446 54.255
218299 94.368 -75.444 54.244
218300 94.390 -75.440 54.232
218301 94.406 -75.440 54.217
218302 94.420 -75.444 54.203
218303 94.433 -75.448 54.189
218304 94.447 -75.452 54.172
218305 94.461 -75.455 54.145
218306 94.475 -75.459 54.114
218307 94.488 -75.463 54.081
218308 94.502 -75.467 54.045
218309 94.516 -75.471 54.007
218310 94.530 -75.474 53.968
218311 94.546 -75.477 53.929
218312 94.563 -75.478 53.888
218313 94.581 -75.479 53.848
218314 94.599 -75.480 53.810
218315 94.625 -75.479 53.773
218316 94.650 -75.478 53.736
218317 94.676 -75.477 53.698
218318 94.701 -75.476 53.657
218319 94.726 -75.475 53.615
218320 94.716 -75.478 53.573
218321 94.707 -75.480 53.533
218322 94.698 -75.483 53.493
218323 94.688 -75.486 53.454
218324 94.679 -75.489 53.416
218325 94.670 -75.491 53.377
218326 94.660 -75.493 53.339
218327 94.650 -75.496 53.303
218328 94.641 -75.499 53.268
218329 94.636 -75.502 53.230
218330 94.645 -75.504 53.189
218331 94.653 -75.507 53.146
218332 94.663 -75.510 53.098
218333 94.679 -75.514 53.051
218334 94.709 -75.520 53.005
218335 94.740 -75.526 52.961
218336 94.770 -75.532 52.920
218337 94.800 -75.538 52.879
218338 94.830 -75.544 52.841
218339 94.860 -75.550 52.805
218340 94.891 -75.556 52.771
218341 94.921 -75.562 52.737
218342 94.949 -75.568 52.700
218343 94.974 -75.571 52.659
218344 95.000 -75.574 52.617
218345 95.026 -75.577 52.572
218346 95.052 -75.581 52.526
218347 95.082 -75.589 52.482
218348 95.116 -75.599 52.437
218349 95.150 -75.610 52.394
218350 95.184 -75.621 52.351
218351 95.218 -75.632 52.308
218352 95.252 -75.643 52.266
218353 95.286 -75.654 52.226
218354 95.321 -75.665 52.186
218355 95.354 -75.676 52.146
218356 95.389 -75.687 52.107
218357 95.442 -75.690 52.068
218358 95.479 -75.677 52.029
218359 95.491 -75.649 51.991
218360 95.503 -75.621 51.953
218361 95.516 -75.592 51.914
218362 95.527 -75.571 51.877
218363 95.539 -75.587 51.838
218364 95.557 -75.609 51.800
218365 95.588 -75.637 51.763
218366 95.618 -75.665 51.727
218367 95.617 -75.717 51.694
218368 95.624 -75.759 51.661
218369 95.666 -75.758 51.628
218370 95.708 -75.756 51.596
218371 95.750 -75.755 51.566
218372 95.791 -75.753 51.538
218373 95.833 -75.752 51.511
218374 95.838 -75.765 51.483
218375 95.890 -75.766 51.454
218376 95.914 -75.773 51.423
218377 95.922 -75.784 51.392
218378 95.959 -75.778 51.357
218379 96.002 -75.771 51.323
218380 96.048 -75.754 51.291
218381 96.094 -75.737 51.259
218382 96.140 -75.719 51.226
218383 96.186 -75.702 51.193
218384 96.205 -75.707 51.159
218385 96.210 -75.724 51.125
218386 96.230 -75.731 51.092
218387 96.250 -75.738 51.058
218388 96.271 -75.745 51.024
218389 96.291 -75.751 50.991
218390 96.311 -75.758 50.958
218391 96.332 -75.765 50.927
218392 96.352 -75.771 50.898
218393 96.372 -75.778 50.871
218394 96.393 -75.787 50.844
218395 96.415 -75.798 50.818
218396 96.438 -75.809 50.793
218397 96.460 -75.820 50.767
218398 96.479 -75.833 50.742
218399 96.496 -75.847 50.717
218400 96.512 -75.861 50.694
218401 96.528 -75.875 50.671
218402 96.544 -75.889 50.648
218403 96.560 -75.903 50.626
218404 96.576 -75.918 50.604
218405 96.587 -75.934 50.580
218406 96.586 -75.954 50.558
218407 96.586 -75.975 50.536
218408 96.589 -75.994 50.515
218409 96.659 -75.994 50.492
218410 96.709 -75.990 50.468
218411 96.749 -75.982 50.441
218412 96.783 -75.974 50.413
218413 96.800 -75.995 50.384
218414 96.837 -76.043 50.353
218415 96.810 -76.064 50.319
218416 96.772 -76.058 50.280
218417 96.799 -76.064 50.238
218418 96.825 -76.069 50.196
218419 96.852 -76.075 50.152
218420 96.879 -76.080 50.111
218421 96.906 -76.085 50.069
218422 96.933 -76.091 50.028
218423 96.960 -76.097 49.988
218424 96.986 -76.102 49.950
218425 97.013 -76.107 49.911
218426 97.040 -76.113 49.874
218427 97.067 -76.119 49.837
218428 97.094 -76.124 49.800
218429 97.122 -76.130 49.765
218430 97.150 -76.137 49.731
218431 97.178 -76.143 49.698
218432 97.206 -76.149 49.667
218433 97.238 -76.156 49.636
218434 97.279 -76.163 49.606
218435 97.317 -76.171 49.577
218436 97.356 -76.180 49.548
218437 97.394 -76.188 49.521
218438 97.432 -76.198 49.495
218439 97.470 -76.206 49.468
218440 97.501 -76.208 49.441
218441 97.531 -76.210 49.414
218442 97.564 -76.216 49.387
218443 97.597 -76.226 49.359
218444 97.620 -76.235 49.331
218445 97.637 -76.244 49.305
218446 97.653 -76.253 49.279
218447 97.663 -76.264 49.253
218448 97.667 -76.277 49.228
218449 97.672 -76.291 49.202
218450 97.676 -76.304 49.175
218451 97.680 -76.317 49.148
218452 97.685 -76.330 49.119
218453 97.689 -76.344 49.090
218454 97.694 -76.357 49.060
218455 97.700 -76.370 49.030
218456 97.724 -76.382 48.999
218457 97.748 -76.394 48.967
218458 97.808 -76.399 48.935
218459 97.877 -76.403 48.904
218460 97.922 -76.414 48.869
218461 97.965 -76.425 48.833
218462 98.009 -76.436 48.794
218463 98.052 -76.448 48.754
218464 98.096 -76.458 48.714
218465 98.139 -76.481 48.675
218466 98.182 -76.506 48.634
218467 98.232 -76.525 48.593
218468 98.263 -76.533 48.555
218469 98.261 -76.537 48.518
218470 98.293 -76.547 48.482
218471 98.325 -76.557 48.447
218472 98.357 -76.567 48.412
218473 98.394 -76.547 48.379
218474 98.427 -76.577 48.346
218475 98.450 -76.581 48.315
218476 98.473 -76.586 48.285
218477 98.496 -76.591 48.256
218478 98.519 -76.596 48.226
218479 98.541 -76.600 48.199
218480 98.563 -76.605 48.172
218481 98.581 -76.607 48.145
218482 98.598 -76.610 48.118
218483 98.617 -76.613 48.092
218484 98.634 -76.615 48.066
218485 98.652 -76.618 48.040
218486 98.672 -76.621 48.016
218487 98.700 -76.624 47.992
218488 98.729 -76.627 47.967
218489 98.740 -76.640 47.942
218490 98.747 -76.655 47.916
218491 98.760 -76.660 47.890
218492 98.773 -76.665 47.863
218493 98.786 -76.671 47.835
218494 98.798 -76.676 47.806
218495 98.811 -76.681 47.776
218496 98.832 -76.685 47.745
218497 98.854 -76.687 47.713
218498 98.877 -76.688 47.681
218499 98.903 -76.685 47.650
218500 98.928 -76.682 47.620
218501 98.950 -76.677 47.591
218502 98.962 -76.666 47.561
218503 98.974 -76.655 47.529
218504 98.987 -76.646 47.496
218505 99.006 -76.655 47.461
218506 99.025 -76.663 47.425
218507 99.050 -76.674 47.387
218508 99.082 -76.688 47.349
218509 99.117 -76.703 47.311
218510 99.165 -76.721 47.273
218511 99.213 -76.740 47.234
218512 99.261 -76.758 47.195
218513 99.303 -76.771 47.154
218514 99.321 -76.771 47.115
218515 99.339 -76.770 47.078
218516 99.357 -76.770 47.039
218517 99.375 -76.770 47.002
218518 99.393 -76.769 46.964
218519 99.411 -76.769 46.926
218520 99.429 -76.768 46.889
218521 99.447 -76.768 46.855
218522 99.466 -76.767 46.821
218523 99.484 -76.766 46.786
218524 99.502 -76.765 46.752
218525 99.521 -76.764 46.718
218526 99.532 -76.769 46.685
218527 99.549 -76.791 46.652
218528 99.566 -76.813 46.621
218529 99.583 -76.836 46.588
218530 99.600 -76.856 46.557
218531 99.609 -76.853 46.526
218532 99.618 -76.850 46.495
218533 99.627 -76.848 46.466
218534 99.636 -76.845 46.438
218535 99.645 -76.842 46.413
218536 99.654 -76.840 46.388
218537 99.663 -76.837 46.363
218538 99.682 -76.838 46.339
218539 99.703 -76.840 46.315
218540 99.725 -76.843 46.290
218541 99.747 -76.846 46.266
218542 99.768 -76.849 46.241
218543 99.789 -76.852 46.214
218544 99.811 -76.854 46.187
218545 99.832 -76.857 46.159
218546 99.858 -76.862 46.130
218547 99.889 -76.868 46.103
218548 99.920 -76.875 46.078
218549 99.952 -76.883 46.053
218550 99.983 -76.890 46.029
218551 100.014 -76.897 46.005
218552 100.046 -76.903 45.983
218553 100.077 -76.911 45.961
218554 100.109 -76.918 45.939
218555 100.140 -76.925 45.918
218556 100.182 -76.926 45.899
218557 100.223 -76.927 45.881
218558 100.265 -76.928 45.862
218559 100.292 -76.939 45.844
218560 100.315 -76.952 45.827
218561 100.340 -76.956 45.810
218562 100.365 -76.959 45.794
218563 100.387 -76.972 45.778
218564 100.407 -76.987 45.762
218565 100.425 -77.004 45.746
218566 100.443 -77.022 45.728
218567 100.461 -77.040 45.712
218568 100.479 -77.057 45.696
218569 100.497 -77.075 45.680
218570 100.515 -77.093 45.663
218571 100.533 -77.110 45.647
218572 100.545 -77.125 45.631
218573 100.554 -77.138 45.616
218574 100.563 -77.151 45.600
218575 100.572 -77.165 45.586
218576 100.582 -77.179 45.572
218577 100.603 -77.196 45.559
218578 100.623 -77.214 45.546
218579 100.643 -77.232 45.534
218580 100.663 -77.250 45.522
218581 100.683 -77.268 45.509
218582 100.703 -77.285 45.499
218583 100.723 -77.304 45.487
218584 100.744 -77.322 45.476
218585 100.769 -77.334 45.465
218586 100.796 -77.345 45.454
218587 100.824 -77.355 45.442
218588 100.851 -77.367 45.431
218589 100.878 -77.377 45.421
218590 100.905 -77.388 45.410
218591 100.928 -77.398 45.399
218592 100.949 -77.409 45.387
218593 100.970 -77.420 45.374
218594 100.990 -77.429 45.361
218595 101.010 -77.440 45.347
218596 101.032 -77.443 45.332
218597 101.054 -77.445 45.317
218598 101.076 -77.448 45.301
218599 101.102 -77.454 45.285
218600 101.130 -77.462 45.270
218601 101.158 -77.470 45.255
218602 101.186 -77.478 45.238
218603 101.213 -77.486 45.222
218604 101.236 -77.495 45.207
218605 101.234 -77.506 45.190
218606 101.235 -77.520 45.175
218607 101.239 -77.535 45.158
218608 101.243 -77.550 45.141
218609 101.247 -77.565 45.123
218610 101.250 -77.580 45.106
218611 101.254 -77.595 45.090
218612 101.284 -77.608 45.072
218613 101.315 -77.622 45.055
218614 101.346 -77.636 45.039
218615 101.376 -77.649 45.022
218616 101.407 -77.662 45.005
218617 101.438 -77.676 44.989
218618 101.469 -77.689 44.972
218619 101.499 -77.703 44.955
218620 101.530 -77.716 44.939
218621 101.561 -77.730 44.922
218622 101.592 -77.743 44.905
218623 101.622 -77.757 44.888
218624 101.653 -77.771 44.872
218625 101.688 -77.783 44.856
218626 101.724 -77.797 44.839
218627 101.764 -77.808 44.824
218628 101.835 -77.809 44.807
218629 101.926 -77.804 44.791
218630 101.944 -77.821 44.776
218631 101.958 -77.846 44.759
218632 101.983 -77.872 44.743
218633 101.976 -77.872 44.727
218634 101.965 -77.864 44.711
218635 101.995 -77.869 44.695
218636 102.012 -77.881 44.680
218637 102.011 -77.900 44.666
218638 102.040 -77.902 44.650
218639 102.069 -77.905 44.635
218640 102.094 -77.915 44.619
218641 102.139 -77.955 44.603
218642 102.174 -77.975 44.587
218643 102.204 -77.988 44.572
218644 102.235 -78.002 44.558
218645 102.265 -78.015 44.544
218646 102.295 -78.028 44.529
218647 102.325 -78.042 44.515
218648 102.356 -78.056 44.500
218649 102.383 -78.065 44.486
218650 102.404 -78.069 44.472
218651 102.426 -78.073 44.456
218652 102.448 -78.077 44.441
218653 102.470 -78.081 44.426
218654 102.495 -78.094 44.410
218655 102.525 -78.116 44.395
218656 102.554 -78.138 44.380
218657 102.584 -78.161 44.366
218658 102.613 -78.183 44.350
218659 102.643 -78.206 44.336
218660 102.672 -78.228 44.321
218661 102.704 -78.222 44.306
218662 102.736 -78.216 44.291
218663 102.767 -78.210 44.276
218664 102.800 -78.250 44.261
218665 102.842 -78.249 44.245
218666 102.884 -78.247 44.229
218667 102.926 -78.246 44.213
218668 102.968 -78.244 44.197
218669 103.010 -78.242 44.181
218670 103.051 -78.241 44.164
218671 103.093 -78.239 44.146
218672 103.135 -78.238 44.129
218673 103.177 -78.236 44.113
218674 103.219 -78.235 44.096
218675 103.260 -78.232 44.078
218918 92.359 -75.483 54.162
218919 92.382 -75.493 54.172
218920 92.405 -75.505 54.180
218921 92.428 -75.515 54.186
218922 92.447 -75.527 54.191
218923 92.455 -75.539 54.192
218924 92.475 -75.530 54.194
218925 92.496 -75.521 54.197
218926 92.516 -75.512 54.197
218927 92.536 -75.505 54.196
218928 92.555 -75.502 54.193
218961 93.323 -75.491 53.926
218962 93.346 -75.489 53.935
218963 93.369 -75.489 53.947
218964 93.392 -75.487 53.963
218965 93.416 -75.486 53.981
218966 93.439 -75.485 54.000
218967 93.462 -75.483 54.022
218968 93.486 -75.483 54.048
218969 93.509 -75.481 54.077
218970 93.532 -75.480 54.110
218971 93.567 -75.478 54.144
218972 93.606 -75.477 54.181
218973 93.645 -75.474 54.221
218974 93.698 -75.458 54.262
218975 93.750 -75.441 54.302
218976 93.788 -75.431 54.340
218977 93.812 -75.428 54.373
218978 93.836 -75.425 54.401
218979 93.860 -75.421 54.418
218980 93.884 -75.419 54.434
218981 93.906 -75.417 54.449
218982 93.928 -75.414 54.464
218983 93.951 -75.412 54.475
218984 93.973 -75.411 54.481
218985 93.997 -75.410 54.479
218986 94.021 -75.410 54.470
218987 94.045 -75.410 54.460
218988 94.069 -75.409 54.450
218989 94.092 -75.408 54.442
218990 94.115 -75.408 54.432
218991 94.138 -75.407 54.420
218992 94.161 -75.406 54.405
218993 94.186 -75.408 54.387
218994 94.212 -75.416 54.367
218995 94.239 -75.422 54.346
218996 94.264 -75.423 54.324
218997 94.290 -75.424 54.304
218998 94.316 -75.424 54.286
218999 94.341 -75.424 54.269
219000 94.366 -75.421 54.254
219001 94.390 -75.415 54.241
219002 94.409 -75.412 54.227
219003 94.426 -75.411 54.213
219004 94.442 -75.411 54.198
219005 94.456 -75.415 54.181
219006 94.470 -75.419 54.154
219007 94.483 -75.423 54.123
219008 94.497 -75.427 54.090
219009 94.511 -75.430 54.054
219010 94.525 -75.434 54.016
219011 94.538 -75.438 53.979
219012 94.554 -75.440 53.940
219013 94.571 -75.442 53.900
219014 94.589 -75.443 53.861
219015 94.607 -75.445 53.824
219016 94.626 -75.446 53.787
219017 94.651 -75.444 53.749
219018 94.677 -75.443 53.710
219019 94.703 -75.442 53.669
219020 94.728 -75.441 53.627
219021 94.719 -75.444 53.586
219022 94.709 -75.446 53.545
219023 94.700 -75.449 53.505
219024 94.691 -75.452 53.465
219025 94.681 -75.455 53.425
219026 94.672 -75.457 53.385
219027 94.662 -75.460 53.345
219028 94.653 -75.462 53.307
219029 94.643 -75.465 53.272
219030 94.634 -75.467 53.234
219031 94.628 -75.471 53.194
219032 94.636 -75.473 53.151
219033 94.657 -75.477 53.104
219034 94.687 -75.483 53.056
219035 94.717 -75.489 53.009
219036 94.748 -75.496 52.964
219037 94.778 -75.502 52.922
219038 94.808 -75.508 52.879
219039 94.838 -75.514 52.838
219040 94.868 -75.521 52.799
219041 94.899 -75.527 52.762
219042 94.929 -75.533 52.726
219043 94.956 -75.536 52.689
219044 94.981 -75.540 52.650
219045 95.007 -75.543 52.608
219046 95.033 -75.546 52.564
219047 95.058 -75.550 52.518
219048 95.088 -75.556 52.474
219049 95.121 -75.568 52.430
219050 95.156 -75.578 52.386
219051 95.190 -75.589 52.342
219052 95.224 -75.600 52.299
219053 95.258 -75.611 52.257
219054 95.292 -75.622 52.216
219055 95.326 -75.633 52.175
219056 95.360 -75.643 52.136
219057 95.394 -75.655 52.098
219058 95.428 -75.665 52.059
219059 95.469 -75.665 52.020
219060 95.481 -75.636 51.981
219061 95.494 -75.607 51.944
219062 95.510 -75.580 51.907
219063 95.531 -75.554 51.870
219064 95.558 -75.565 51.834
219065 95.589 -75.593 51.797
219066 95.607 -75.630 51.761
219067 95.606 -75.684 51.727
219068 95.607 -75.733 51.693
219069 95.649 -75.731 51.659
219070 95.691 -75.730 51.625
219071 95.732 -75.728 51.592
219072 95.774 -75.727 51.561
219073 95.816 -75.725 51.532
219074 95.858 -75.724 51.504
219075 95.899 -75.722 51.475
219076 95.894 -75.741 51.445
219077 95.928 -75.746 51.414
219078 95.937 -75.756 51.382
219079 95.944 -75.767 51.348
219080 95.988 -75.754 51.316
219081 96.036 -75.737 51.284
219082 96.086 -75.721 51.253
219083 96.135 -75.706 51.222
219084 96.185 -75.690 51.191
219085 96.221 -75.684 51.159
219086 96.224 -75.702 51.127
219087 96.243 -75.705 51.095
219088 96.264 -75.709 51.062
219089 96.285 -75.715 51.029
219090 96.305 -75.722 50.996
219091 96.325 -75.729 50.964
219092 96.345 -75.736 50.933
219093 96.366 -75.743 50.903
219094 96.386 -75.749 50.876
219095 96.407 -75.757 50.850
219096 96.429 -75.768 50.824
219097 96.451 -75.779 50.798
219098 96.473 -75.790 50.772
219099 96.489 -75.804 50.747
219100 96.506 -75.818 50.723
219101 96.522 -75.832 50.701
219102 96.538 -75.847 50.678
219103 96.554 -75.861 50.656
219104 96.570 -75.875 50.633
219105 96.586 -75.889 50.611
219106 96.602 -75.903 50.588
219107 96.602 -75.924 50.566
219108 96.601 -75.944 50.544
219109 96.658 -75.948 50.523
219110 96.702 -75.941 50.501
219111 96.743 -75.934 50.477
219112 96.801 -75.955 50.451
219113 96.869 -75.991 50.422
219114 96.936 -76.028 50.393
219115 96.960 -76.066 50.362
219116 96.841 -76.047 50.328
219117 96.793 -76.039 50.289
219118 96.819 -76.044 50.248
219119 96.847 -76.050 50.205
219120 96.873 -76.056 50.162
219121 96.900 -76.061 50.119
219122 96.927 -76.066 50.077
219123 96.954 -76.072 50.036
219124 96.980 -76.078 49.996
219125 97.007 -76.083 49.957
219126 97.034 -76.088 49.918
219127 97.061 -76.094 49.881
219128 97.088 -76.100 49.844
219129 97.115 -76.105 49.808
219130 97.142 -76.110 49.773
219131 97.168 -76.116 49.739
219132 97.192 -76.122 49.706
219133 97.217 -76.127 49.675
219134 97.247 -76.134 49.644
219135 97.286 -76.142 49.614
219136 97.324 -76.151 49.584
219137 97.362 -76.160 49.555
219138 97.400 -76.169 49.528
219139 97.439 -76.177 49.502
219140 97.469 -76.179 49.476
219141 97.500 -76.181 49.448
219142 97.530 -76.183 49.421
219143 97.562 -76.188 49.394
219144 97.595 -76.198 49.367
219145 97.629 -76.207 49.340
219146 97.652 -76.217 49.315
219147 97.668 -76.226 49.290
219148 97.677 -76.238 49.263
219149 97.682 -76.251 49.238
219150 97.686 -76.264 49.212
219151 97.691 -76.277 49.185
219152 97.696 -76.291 49.156
219153 97.700 -76.304 49.127
219154 97.704 -76.317 49.097
219155 97.709 -76.330 49.066
219156 97.713 -76.344 49.035
219157 97.730 -76.356 49.004
219158 97.762 -76.367 48.972
219159 97.830 -76.370 48.941
219160 97.874 -76.382 48.909
219161 97.917 -76.393 48.874
219162 97.960 -76.404 48.837
219163 98.004 -76.416 48.798
219164 98.048 -76.426 48.757
219165 98.091 -76.441 48.717
219166 98.134 -76.465 48.677
219167 98.183 -76.485 48.636
219168 98.240 -76.497 48.594
219169 98.234 -76.500 48.555
219170 98.267 -76.510 48.517
219171 98.298 -76.520 48.481
219172 98.330 -76.530 48.445
219173 98.362 -76.539 48.411
219174 98.394 -76.549 48.377
219175 98.456 -76.571 48.342
219176 98.479 -76.575 48.311
219177 98.501 -76.580 48.281
219178 98.524 -76.585 48.252
219179 98.547 -76.589 48.222
219180 98.570 -76.594 48.194
219181 98.593 -76.599 48.167
219182 98.615 -76.603 48.141
219183 98.632 -76.606 48.115
219184 98.650 -76.609 48.088
219185 98.668 -76.611 48.062
219186 98.686 -76.614 48.038
219187 98.704 -76.617 48.015
219188 98.721 -76.620 47.991
219189 98.739 -76.623 47.966
219190 98.751 -76.627 47.942
219191 98.765 -76.632 47.915
219192 98.777 -76.637 47.888
219193 98.790 -76.643 47.861
219194 98.803 -76.649 47.834
219195 98.816 -76.654 47.805
219196 98.829 -76.659 47.776
219197 98.848 -76.662 47.746
219198 98.869 -76.665 47.716
219199 98.891 -76.668 47.685
219200 98.913 -76.671 47.654
219201 98.939 -76.668 47.624
219202 98.964 -76.664 47.595
219203 98.980 -76.655 47.566
219204 98.992 -76.645 47.534
219205 99.006 -76.640 47.501
219206 99.025 -76.649 47.466
219207 99.045 -76.657 47.429
219208 99.067 -76.667 47.392
219209 99.099 -76.681 47.355
219210 99.131 -76.695 47.316
219211 99.172 -76.712 47.277
219212 99.220 -76.730 47.239
219213 99.269 -76.748 47.199
219214 99.308 -76.759 47.159
219215 99.337 -76.763 47.120
219216 99.355 -76.762 47.081
219217 99.373 -76.762 47.043
219218 99.392 -76.761 47.005
219219 99.410 -76.760 46.967
219220 99.429 -76.759 46.930
219221 99.447 -76.759 46.892
219222 99.466 -76.758 46.857
219223 99.484 -76.756 46.822
219224 99.502 -76.755 46.788
219225 99.520 -76.755 46.754
219226 99.532 -76.756 46.721
219227 99.541 -76.759 46.687
219228 99.556 -76.774 46.653
219229 99.573 -76.796 46.622
219230 99.590 -76.819 46.590
219231 99.604 -76.831 46.559
219232 99.613 -76.828 46.528
219233 99.622 -76.826 46.497
219234 99.631 -76.823 46.469
219235 99.640 -76.820 46.442
219236 99.649 -76.818 46.416
219237 99.664 -76.817 46.391
219238 99.686 -76.820 46.366
219239 99.707 -76.823 46.342
219240 99.728 -76.825 46.318
219241 99.750 -76.828 46.293
219242 99.772 -76.831 46.269
219243 99.793 -76.834 46.243
219244 99.815 -76.836 46.215
219245 99.836 -76.839 46.187
219246 99.857 -76.842 46.159
219247 99.884 -76.846 46.129
219248 99.915 -76.853 46.101
219249 99.946 -76.860 46.076
219250 99.978 -76.868 46.050
219251 100.009 -76.875 46.025
219252 100.041 -76.881 46.001
219253 100.072 -76.889 45.978
219254 100.104 -76.896 45.955
219255 100.135 -76.903 45.932
219256 100.167 -76.909 45.912
219257 100.198 -76.917 45.893
219258 100.232 -76.922 45.874
219259 100.274 -76.923 45.856
219260 100.308 -76.930 45.838
219261 100.331 -76.939 45.820
219262 100.356 -76.943 45.803
219263 100.381 -76.947 45.787
219264 100.403 -76.959 45.771
219265 100.422 -76.975 45.755
219266 100.440 -76.993 45.739
219267 100.458 -77.010 45.723
219268 100.476 -77.028 45.706
219269 100.494 -77.046 45.690
219270 100.512 -77.063 45.674
219271 100.530 -77.081 45.656
219272 100.548 -77.099 45.640
219273 100.565 -77.116 45.624
219274 100.574 -77.129 45.609
219275 100.583 -77.142 45.593
219276 100.592 -77.155 45.578
219277 100.603 -77.169 45.564
219278 100.623 -77.188 45.551
219279 100.644 -77.205 45.539
219280 100.664 -77.223 45.526
219281 100.684 -77.241 45.514
219282 100.704 -77.259 45.502
219283 100.724 -77.277 45.491
219284 100.745 -77.294 45.479
219285 100.772 -77.306 45.468
219286 100.799 -77.316 45.456
219287 100.826 -77.327 45.445
219288 100.853 -77.338 45.434
219289 100.881 -77.348 45.423
219290 100.908 -77.359 45.413
219291 100.935 -77.370 45.402
219292 100.959 -77.381 45.391
219293 100.979 -77.391 45.379
219294 101.000 -77.401 45.366
219295 101.020 -77.412 45.353
219296 101.041 -77.423 45.339
219297 101.061 -77.433 45.325
219298 101.082 -77.437 45.311
219299 101.105 -77.439 45.294
219300 101.129 -77.444 45.278
219301 101.157 -77.452 45.262
219302 101.185 -77.460 45.246
219303 101.213 -77.467 45.229
219304 101.230 -77.477 45.212
219305 101.228 -77.489 45.196
219306 101.226 -77.501 45.178
219307 101.224 -77.513 45.162
219308 101.222 -77.524 45.145
219309 101.222 -77.537 45.127
219310 101.227 -77.552 45.110
219311 101.231 -77.567 45.092
219312 101.239 -77.582 45.076
219313 101.269 -77.595 45.059
219314 101.300 -77.609 45.042
219315 101.330 -77.623 45.025
219316 101.361 -77.636 45.008
219317 101.392 -77.649 44.992
219318 101.422 -77.663 44.976
219319 101.453 -77.677 44.959
219320 101.487 -77.690 44.943
219321 101.523 -77.703 44.926
219322 101.559 -77.717 44.910
219323 101.595 -77.730 44.893
219324 101.631 -77.743 44.877
219325 101.668 -77.757 44.860
219326 101.704 -77.771 44.844
219327 101.746 -77.780 44.828
219328 101.790 -77.790 44.812
219329 101.828 -77.802 44.795
219330 101.919 -77.796 44.779
219331 101.943 -77.812 44.763
219332 101.942 -77.835 44.747
219333 101.959 -77.861 44.731
219334 101.959 -77.866 44.716
219335 101.969 -77.865 44.700
219336 101.982 -77.879 44.685
219337 102.001 -77.887 44.669
219338 102.030 -77.889 44.654
219339 102.059 -77.891 44.639
219340 102.088 -77.893 44.623
219341 102.113 -77.905 44.607
219342 102.139 -77.923 44.590
219343 102.187 -77.961 44.574
219344 102.217 -77.975 44.560
219345 102.247 -77.988 44.546
219346 102.278 -78.002 44.531
219347 102.308 -78.015 44.517
219348 102.338 -78.029 44.502
219349 102.369 -78.043 44.487
219350 102.394 -78.050 44.473
219351 102.415 -78.054 44.458
219352 102.437 -78.058 44.443
219353 102.459 -78.062 44.428
219354 102.481 -78.065 44.413
219355 102.503 -78.069 44.397
219356 102.532 -78.091 44.382
219357 102.561 -78.113 44.368
219358 102.591 -78.136 44.354
219359 102.620 -78.159 44.338
219360 102.650 -78.181 44.324
219361 102.679 -78.203 44.309
219362 102.709 -78.226 44.293
219363 102.740 -78.233 44.278
219364 102.771 -78.227 44.263
219365 102.807 -78.248 44.248
219366 102.848 -78.247 44.233
219367 102.890 -78.245 44.216
219368 102.932 -78.244 44.200
219369 102.974 -78.242 44.185
219370 103.016 -78.240 44.168
219371 103.058 -78.238 44.151
219372 103.100 -78.237 44.133
219373 103.141 -78.235 44.116
219374 103.183 -78.234 44.101
219375 103.225 -78.232 44.083
219376 103.267 -78.230 44.066
219618 92.323 -75.442 54.169
219619 92.346 -75.453 54.180
219620 92.369 -75.464 54.189
219621 92.392 -75.475 54.196
219622 92.415 -75.486 54.203
219623 92.438 -75.498 54.208
219624 92.448 -75.510 54.210
219625 92.465 -75.506 54.211
219626 92.485 -75.498 54.215
219627 92.505 -75.489 54.215
219628 92.525 -75.481 54.214
219629 92.545 -75.478 54.211
219630 92.564 -75.475 54.206
219631 92.583 -75.472 54.199
219639 92.823 -75.439 54.132
219640 92.854 -75.436 54.122
219641 92.884 -75.431 54.110
219642 92.916 -75.427 54.098
219643 92.946 -75.423 54.087
219644 92.973 -75.420 54.075
219645 92.992 -75.424 54.062
219646 93.010 -75.427 54.050
219662 93.321 -75.461 53.927
219663 93.344 -75.461 53.935
219664 93.368 -75.459 53.946
219665 93.391 -75.458 53.961
219666 93.414 -75.457 53.978
219667 93.438 -75.455 53.996
219668 93.461 -75.454 54.017
219669 93.484 -75.453 54.042
219670 93.508 -75.452 54.070
219671 93.531 -75.450 54.101
219672 93.566 -75.449 54.133
219673 93.605 -75.447 54.167
219674 93.644 -75.445 54.203
219675 93.692 -75.433 54.242
219676 93.745 -75.416 54.281
219677 93.781 -75.407 54.320
219678 93.805 -75.404 54.357
219679 93.829 -75.401 54.390
219680 93.853 -75.398 54.413
219681 93.877 -75.395 54.434
219682 93.901 -75.392 54.453
219683 93.925 -75.389 54.471
219684 93.948 -75.385 54.481
219685 93.972 -75.383 54.490
219686 93.995 -75.382 54.493
219687 94.018 -75.381 54.489
219688 94.041 -75.380 54.478
219689 94.064 -75.380 54.466
219690 94.087 -75.379 54.456
219691 94.110 -75.377 54.449
219692 94.133 -75.377 54.439
219693 94.156 -75.376 54.425
219694 94.183 -75.382 54.407
219695 94.209 -75.389 54.389
219696 94.236 -75.396 54.369
219697 94.262 -75.401 54.347
219698 94.288 -75.402 54.325
219699 94.314 -75.402 54.305
219700 94.339 -75.402 54.284
219701 94.363 -75.398 54.266
219702 94.387 -75.392 54.252
219703 94.411 -75.386 54.239
219704 94.428 -75.384 54.225
219705 94.445 -75.384 54.209
219706 94.462 -75.383 54.191
219707 94.479 -75.383 54.164
219708 94.492 -75.386 54.133
219709 94.506 -75.390 54.100
219710 94.520 -75.394 54.065
219711 94.534 -75.398 54.029
219712 94.547 -75.402 53.992
219713 94.562 -75.405 53.955
219714 94.580 -75.406 53.916
219715 94.598 -75.408 53.878
219716 94.615 -75.408 53.840
219717 94.633 -75.410 53.802
219718 94.653 -75.411 53.763
219719 94.678 -75.409 53.724
219720 94.704 -75.408 53.684
219721 94.730 -75.408 53.642
219722 94.721 -75.410 53.600
219723 94.712 -75.413 53.560
219724 94.702 -75.415 53.520
219725 94.693 -75.418 53.479
219726 94.684 -75.420 53.439
219727 94.674 -75.423 53.398
219728 94.665 -75.426 53.357
219729 94.655 -75.428 53.316
219730 94.646 -75.431 53.278
219731 94.636 -75.433 53.239
219732 94.627 -75.437 53.199
219733 94.635 -75.442 53.157
219734 94.666 -75.448 53.110
219735 94.696 -75.454 53.063
219736 94.726 -75.460 53.016
219737 94.756 -75.466 52.970
219738 94.786 -75.472 52.925
219739 94.817 -75.478 52.882
219740 94.847 -75.484 52.839
219741 94.877 -75.490 52.799
219742 94.907 -75.496 52.762
219743 94.937 -75.502 52.726
219744 94.963 -75.505 52.689
219745 94.988 -75.509 52.646
219746 95.014 -75.512 52.603
219747 95.040 -75.516 52.558
219748 95.065 -75.519 52.512
219749 95.093 -75.524 52.467
219750 95.127 -75.535 52.422
219751 95.161 -75.546 52.378
219752 95.195 -75.557 52.333
219753 95.229 -75.568 52.289
219754 95.263 -75.579 52.245
219755 95.297 -75.590 52.204
219756 95.331 -75.601 52.164
219757 95.366 -75.612 52.124
219758 95.400 -75.622 52.085
219759 95.433 -75.634 52.046
219760 95.462 -75.646 52.009
219761 95.486 -75.642 51.971
219762 95.517 -75.609 51.935
219763 95.543 -75.579 51.898
219764 95.564 -75.553 51.863
219765 95.590 -75.550 51.829
219766 95.595 -75.597 51.794
219767 95.594 -75.650 51.760
219768 95.593 -75.702 51.726
219769 95.631 -75.705 51.693
219770 95.673 -75.703 51.659
219771 95.715 -75.702 51.624
219772 95.757 -75.700 51.590
219773 95.799 -75.699 51.556
219774 95.840 -75.697 51.525
219775 95.882 -75.696 51.495
219776 95.924 -75.694 51.466
219777 95.966 -75.693 51.436
219778 95.943 -75.718 51.405
219779 95.962 -75.724 51.373
219780 95.975 -75.726 51.340
219781 95.990 -75.731 51.307
219782 96.036 -75.726 51.276
219783 96.085 -75.710 51.246
219784 96.135 -75.694 51.216
219785 96.184 -75.678 51.187
219786 96.234 -75.662 51.157
219787 96.238 -75.680 51.126
219788 96.256 -75.683 51.096
219789 96.276 -75.686 51.065
219790 96.298 -75.690 51.034
219791 96.319 -75.693 51.001
219792 96.339 -75.699 50.969
219793 96.359 -75.706 50.938
219794 96.380 -75.713 50.908
219795 96.400 -75.720 50.881
219796 96.420 -75.727 50.855
219797 96.442 -75.737 50.829
219798 96.464 -75.749 50.803
219799 96.483 -75.762 50.777
219800 96.500 -75.775 50.752
219801 96.516 -75.790 50.729
219802 96.532 -75.804 50.708
219803 96.548 -75.818 50.686
219804 96.564 -75.832 50.663
219805 96.580 -75.847 50.640
219806 96.596 -75.860 50.618
219807 96.612 -75.875 50.595
219808 96.617 -75.893 50.573
219809 96.654 -75.900 50.552
219810 96.695 -75.894 50.532
219811 96.763 -75.930 50.509
219812 96.830 -75.966 50.486
219813 96.877 -75.991 50.460
219814 96.909 -76.009 50.432
219815 96.941 -76.026 50.402
219816 96.973 -76.044 50.370
219817 96.873 -76.029 50.336
219818 96.814 -76.020 50.299
219819 96.840 -76.025 50.258
219820 96.867 -76.031 50.215
219821 96.894 -76.037 50.170
219822 96.921 -76.042 50.127
219823 96.948 -76.047 50.084
219824 96.974 -76.053 50.043
219825 97.002 -76.059 50.003
219826 97.028 -76.064 49.964
219827 97.055 -76.069 49.925
219828 97.081 -76.075 49.888
219829 97.106 -76.081 49.852
219830 97.130 -76.086 49.815
219831 97.155 -76.091 49.780
219832 97.179 -76.097 49.746
219833 97.204 -76.103 49.714
219834 97.229 -76.108 49.683
219835 97.254 -76.113 49.652
219836 97.293 -76.122 49.622
219837 97.331 -76.131 49.592
219838 97.369 -76.140 49.564
219839 97.407 -76.148 49.537
219840 97.438 -76.151 49.510
219841 97.468 -76.153 49.484
219842 97.499 -76.154 49.456
219843 97.529 -76.157 49.429
219844 97.561 -76.160 49.402
219845 97.594 -76.170 49.376
219846 97.628 -76.180 49.350
219847 97.661 -76.190 49.325
219848 97.684 -76.199 49.301
219849 97.692 -76.211 49.274
219850 97.697 -76.224 49.247
219851 97.701 -76.238 49.220
219852 97.706 -76.251 49.193
219853 97.710 -76.264 49.164
219854 97.715 -76.277 49.134
219855 97.719 -76.291 49.102
219856 97.724 -76.304 49.070
219857 97.728 -76.317 49.037
219858 97.735 -76.330 49.006
219859 97.782 -76.339 48.975
219860 97.825 -76.350 48.945
219861 97.869 -76.361 48.914
219862 97.913 -76.373 48.878
219863 97.956 -76.383 48.839
219864 98.000 -76.395 48.800
219865 98.043 -76.406 48.760
219866 98.086 -76.424 48.719
219867 98.135 -76.445 48.677
219868 98.191 -76.457 48.634
219869 98.231 -76.467 48.593
219870 98.240 -76.473 48.553
219871 98.272 -76.483 48.515
219872 98.303 -76.492 48.478
219873 98.336 -76.502 48.443
219874 98.366 -76.515 48.408
219875 98.395 -76.531 48.373
219876 98.455 -76.552 48.340
219877 98.488 -76.561 48.307
219878 98.520 -76.570 48.276
219879 98.553 -76.579 48.246
219880 98.575 -76.583 48.217
219881 98.598 -76.589 48.189
219882 98.621 -76.593 48.162
219883 98.644 -76.598 48.136
219884 98.666 -76.602 48.109
219885 98.684 -76.605 48.084
219886 98.702 -76.608 48.059
219887 98.719 -76.611 48.035
219888 98.737 -76.613 48.010
219889 98.754 -76.615 47.986
219890 98.762 -76.611 47.962
219891 98.771 -76.607 47.937
219892 98.782 -76.610 47.911
219893 98.795 -76.615 47.886
219894 98.808 -76.621 47.859
219895 98.820 -76.626 47.831
219896 98.834 -76.631 47.804
219897 98.846 -76.636 47.776
219898 98.863 -76.640 47.747
219899 98.885 -76.643 47.717
219900 98.907 -76.646 47.687
219901 98.928 -76.649 47.656
219902 98.950 -76.652 47.627
219903 98.975 -76.649 47.599
219904 98.997 -76.645 47.570
219905 99.010 -76.634 47.538
219906 99.025 -76.633 47.505
219907 99.045 -76.643 47.470
219908 99.065 -76.651 47.433
219909 99.084 -76.660 47.397
219910 99.116 -76.674 47.360
219911 99.148 -76.688 47.322
219912 99.180 -76.702 47.283
219913 99.222 -76.715 47.243
219914 99.262 -76.725 47.204
219915 99.300 -76.735 47.163
219916 99.339 -76.745 47.124
219917 99.373 -76.752 47.085
219918 99.392 -76.752 47.047
219919 99.410 -76.751 47.009
219920 99.428 -76.750 46.971
219921 99.447 -76.749 46.933
219922 99.465 -76.748 46.895
219923 99.484 -76.747 46.860
219924 99.502 -76.746 46.825
219925 99.520 -76.746 46.790
219926 99.533 -76.746 46.757
219927 99.542 -76.749 46.724
219928 99.551 -76.750 46.690
219929 99.562 -76.758 46.656
219930 99.579 -76.781 46.624
219931 99.596 -76.802 46.592
219932 99.608 -76.806 46.561
219933 99.617 -76.804 46.529
219934 99.626 -76.801 46.500
219935 99.635 -76.798 46.472
219936 99.646 -76.796 46.445
219937 99.667 -76.799 46.419
219938 99.689 -76.802 46.394
219939 99.711 -76.805 46.370
219940 99.732 -76.807 46.345
219941 99.754 -76.810 46.321
219942 99.775 -76.812 46.296
219943 99.797 -76.815 46.270
219944 99.818 -76.818 46.244
219945 99.840 -76.821 46.216
219946 99.861 -76.824 46.188
219947 99.883 -76.826 46.158
219948 99.910 -76.831 46.128
219949 99.941 -76.838 46.100
219950 99.973 -76.846 46.073
219951 100.004 -76.853 46.047
219952 100.036 -76.859 46.022
219953 100.067 -76.866 45.997
219954 100.099 -76.874 45.973
219955 100.130 -76.881 45.949
219956 100.161 -76.887 45.927
219957 100.193 -76.895 45.906
219958 100.224 -76.902 45.887
219959 100.256 -76.909 45.867
219960 100.287 -76.915 45.848
219961 100.323 -76.920 45.831
219962 100.347 -76.926 45.813
219963 100.372 -76.930 45.796
219964 100.397 -76.934 45.780
219965 100.418 -76.947 45.765
219966 100.437 -76.963 45.749
219967 100.455 -76.981 45.732
219968 100.473 -76.999 45.717
219969 100.491 -77.016 45.700
219970 100.509 -77.034 45.684
219971 100.527 -77.052 45.668
219972 100.545 -77.069 45.650
219973 100.563 -77.088 45.633
219974 100.581 -77.105 45.617
219975 100.593 -77.119 45.602
219976 100.602 -77.132 45.585
219977 100.611 -77.146 45.571
219978 100.624 -77.160 45.557
219979 100.644 -77.179 45.544
219980 100.665 -77.197 45.531
219981 100.685 -77.214 45.519
219982 100.705 -77.232 45.507
219983 100.725 -77.250 45.495
219984 100.747 -77.266 45.483
219985 100.774 -77.277 45.472
219986 100.801 -77.288 45.460
219987 100.828 -77.298 45.448
219988 100.855 -77.310 45.437
219989 100.882 -77.320 45.426
219990 100.910 -77.331 45.415
219991 100.937 -77.341 45.405
219992 100.964 -77.352 45.394
219993 100.989 -77.363 45.384
219994 101.009 -77.373 45.372
219995 101.030 -77.384 45.358
219996 101.050 -77.395 45.344
219997 101.071 -77.405 45.330
219998 101.091 -77.415 45.316
219999 101.112 -77.426 45.302
220000 101.133 -77.431 45.286
220001 101.156 -77.434 45.270
220002 101.184 -77.442 45.254
220003 101.212 -77.450 45.237
220004 101.223 -77.460 45.219
220005 101.221 -77.472 45.201
220006 101.219 -77.483 45.184
220007 101.218 -77.495 45.166
220008 101.216 -77.507 45.149
220009 101.214 -77.519 45.132
220010 101.213 -77.530 45.113
220011 101.211 -77.542 45.096
220012 101.210 -77.555 45.079
220013 101.223 -77.569 45.062
220014 101.254 -77.583 45.046
220015 101.286 -77.596 45.029
220016 101.322 -77.610 45.012
220017 101.358 -77.623 44.995
220018 101.394 -77.636 44.980
220019 101.430 -77.650 44.963
220020 101.467 -77.664 44.946
220021 101.502 -77.677 44.930
220022 101.539 -77.690 44.913
220023 101.575 -77.704 44.897
220024 101.611 -77.717 44.881
220025 101.647 -77.730 44.864
220026 101.685 -77.743 44.849
220027 101.729 -77.752 44.832
220028 101.771 -77.763 44.816
220029 101.815 -77.773 44.799
220030 101.849 -77.785 44.782
220031 101.911 -77.789 44.766
220032 101.942 -77.802 44.751
220033 101.941 -77.826 44.735
220034 101.939 -77.849 44.720
220035 101.943 -77.861 44.704
220036 101.963 -77.870 44.688
220037 101.991 -77.872 44.673
220038 102.021 -77.874 44.657
220039 102.049 -77.877 44.641
220040 102.078 -77.879 44.626
220041 102.107 -77.881 44.610
220042 102.131 -77.893 44.594
220043 102.155 -77.909 44.578
220044 102.198 -77.945 44.562
220045 102.230 -77.962 44.548
220046 102.260 -77.975 44.533
220047 102.290 -77.989 44.518
220048 102.321 -78.003 44.504
220049 102.351 -78.016 44.488
220050 102.382 -78.030 44.474
220051 102.405 -78.035 44.459
220052 102.427 -78.039 44.445
220053 102.448 -78.043 44.430
220054 102.470 -78.047 44.415
220055 102.492 -78.050 44.399
220056 102.514 -78.054 44.384
220057 102.539 -78.066 44.370
220058 102.569 -78.088 44.355
220059 102.598 -78.111 44.341
220060 102.627 -78.134 44.326
220061 102.657 -78.156 44.311
220062 102.686 -78.178 44.297
220063 102.716 -78.201 44.282
220064 102.746 -78.224 44.267
220065 102.775 -78.244 44.252
220066 102.813 -78.246 44.236
220067 102.855 -78.244 44.220
220068 102.897 -78.243 44.203
220069 102.939 -78.241 44.188
220070 102.981 -78.240 44.172
220071 103.023 -78.238 44.155
220072 103.065 -78.236 44.138
220073 103.106 -78.235 44.120
220074 103.148 -78.233 44.104
220075 103.190 -78.232 44.088
220076 103.232 -78.230 44.070
220077 103.274 -78.228 44.053
220317 92.265 -75.391 54.172
220318 92.287 -75.402 54.178
220319 92.310 -75.413 54.186
220320 92.334 -75.424 54.196
220321 92.357 -75.435 54.205
220322 92.380 -75.446 54.213
220323 92.403 -75.458 54.220
220324 92.426 -75.468 54.224
220325 92.441 -75.480 54.227
220326 92.454 -75.483 54.228
220327 92.475 -75.474 54.231
220328 92.495 -75.465 54.232
220329 92.515 -75.458 54.231
220330 92.534 -75.455 54.227
220331 92.554 -75.452 54.222
220332 92.573 -75.449 54.216
220333 92.600 -75.445 54.210
220334 92.631 -75.440 54.204
220339 92.785 -75.420 54.157
220340 92.816 -75.415 54.147
220341 92.847 -75.411 54.136
220342 92.877 -75.407 54.125
220343 92.908 -75.403 54.113
220344 92.939 -75.398 54.101
220345 92.965 -75.398 54.089
220346 92.984 -75.400 54.076
220347 93.002 -75.403 54.063
220348 93.020 -75.406 54.051
220349 93.039 -75.409 54.038
220363 93.320 -75.432 53.929
220364 93.343 -75.431 53.936
220365 93.366 -75.430 53.945
220366 93.390 -75.428 53.960
220367 93.413 -75.427 53.976
220368 93.436 -75.426 53.993
220369 93.460 -75.424 54.013
220370 93.483 -75.424 54.037
220371 93.506 -75.422 54.064
220372 93.530 -75.420 54.093
220373 93.565 -75.419 54.124
220374 93.604 -75.417 54.155
220375 93.643 -75.415 54.189
220376 93.687 -75.408 54.224
220377 93.740 -75.392 54.262
220378 93.774 -75.383 54.300
220379 93.798 -75.380 54.337
220380 93.822 -75.377 54.373
220381 93.846 -75.373 54.400
220382 93.869 -75.370 54.423
220383 93.894 -75.367 54.443
220384 93.917 -75.364 54.462
220385 93.942 -75.361 54.475
220386 93.968 -75.356 54.487
220387 93.990 -75.350 54.493
220388 94.013 -75.350 54.490
220389 94.037 -75.349 54.482
220390 94.060 -75.349 54.473
220391 94.083 -75.348 54.464
220392 94.106 -75.348 54.457
220393 94.129 -75.346 54.453
220394 94.152 -75.348 54.443
220395 94.180 -75.355 54.428
220396 94.206 -75.362 54.410
220397 94.233 -75.369 54.391
220398 94.260 -75.377 54.369
220399 94.286 -75.380 54.346
220400 94.312 -75.380 54.324
220401 94.337 -75.380 54.300
220402 94.361 -75.374 54.281
220403 94.385 -75.368 54.268
220404 94.408 -75.362 54.254
220405 94.430 -75.357 54.241
220406 94.448 -75.357 54.226
220407 94.464 -75.356 54.206
220408 94.481 -75.355 54.177
220409 94.498 -75.355 54.146
220410 94.515 -75.355 54.114
220411 94.529 -75.358 54.080
220412 94.542 -75.362 54.044
220413 94.556 -75.366 54.010
220414 94.571 -75.369 53.974
220415 94.589 -75.370 53.936
220416 94.606 -75.371 53.897
220417 94.623 -75.373 53.858
220418 94.641 -75.374 53.818
220419 94.659 -75.375 53.777
220420 94.680 -75.376 53.739
220421 94.705 -75.374 53.700
220422 94.731 -75.373 53.659
220423 94.723 -75.376 53.617
220424 94.714 -75.379 53.577
220425 94.705 -75.381 53.537
220426 94.695 -75.384 53.496
220427 94.686 -75.386 53.456
220428 94.677 -75.389 53.414
220429 94.667 -75.392 53.371
220430 94.657 -75.395 53.329
220431 94.651 -75.400 53.287
220432 94.644 -75.405 53.246
220433 94.637 -75.411 53.205
220434 94.651 -75.414 53.162
220435 94.674 -75.417 53.117
220436 94.704 -75.424 53.070
220437 94.734 -75.430 53.023
220438 94.764 -75.436 52.976
220439 94.794 -75.442 52.930
220440 94.825 -75.449 52.886
220441 94.855 -75.455 52.842
220442 94.885 -75.461 52.800
220443 94.915 -75.467 52.762
220444 94.944 -75.471 52.725
220445 94.969 -75.475 52.687
220446 94.995 -75.478 52.643
220447 95.021 -75.482 52.598
220448 95.047 -75.485 52.552
220449 95.072 -75.489 52.505
220450 95.098 -75.492 52.459
220451 95.132 -75.503 52.413
220452 95.166 -75.514 52.367
220453 95.200 -75.525 52.322
220454 95.235 -75.536 52.278
220455 95.268 -75.546 52.235
220456 95.303 -75.558 52.192
220457 95.337 -75.568 52.152
220458 95.371 -75.579 52.113
220459 95.405 -75.590 52.074
220460 95.436 -75.602 52.035
220461 95.457 -75.615 51.998
220462 95.477 -75.629 51.961
220463 95.498 -75.643 51.925
220464 95.528 -75.615 51.890
220465 95.559 -75.582 51.856
220466 95.584 -75.564 51.823
220467 95.583 -75.617 51.790
220468 95.582 -75.669 51.757
220469 95.614 -75.679 51.726
220470 95.656 -75.677 51.693
220471 95.698 -75.676 51.659
220472 95.740 -75.674 51.623
220473 95.781 -75.673 51.589
220474 95.823 -75.671 51.553
220475 95.865 -75.670 51.520
220476 95.907 -75.668 51.489
220477 95.949 -75.667 51.459
220478 95.990 -75.665 51.428
220479 96.024 -75.662 51.397
220480 96.016 -75.671 51.366
220481 96.008 -75.679 51.333
220482 96.003 -75.687 51.301
220483 96.046 -75.688 51.271
220484 96.089 -75.690 51.241
220485 96.135 -75.683 51.212
220486 96.184 -75.667 51.182
220487 96.234 -75.651 51.153
220488 96.253 -75.657 51.124
220489 96.270 -75.661 51.096
220490 96.288 -75.664 51.067
220491 96.309 -75.668 51.037
220492 96.330 -75.671 51.005
220493 96.352 -75.674 50.973
220494 96.373 -75.677 50.942
220495 96.393 -75.684 50.913
220496 96.414 -75.691 50.886
220497 96.434 -75.697 50.860
220498 96.456 -75.707 50.834
220499 96.477 -75.718 50.808
220500 96.493 -75.733 50.782
220501 96.510 -75.747 50.758
220502 96.526 -75.761 50.736
220503 96.542 -75.775 50.714
220504 96.558 -75.790 50.692
220505 96.574 -75.803 50.670
220506 96.590 -75.818 50.648
220507 96.606 -75.832 50.625
220508 96.622 -75.846 50.602
220509 96.643 -75.861 50.581
220510 96.674 -75.878 50.560
220511 96.707 -75.896 50.540
220512 96.738 -75.913 50.518
220513 96.770 -75.931 50.494
220514 96.802 -75.948 50.468
220515 96.834 -75.966 50.439
220516 96.866 -75.983 50.409
220517 96.898 -76.000 50.376
220518 96.904 -76.012 50.342
220519 96.834 -76.001 50.305
220520 96.861 -76.007 50.263
220521 96.888 -76.012 50.219
220522 96.915 -76.018 50.176
220523 96.942 -76.023 50.134
220524 96.968 -76.028 50.091
220525 96.994 -76.034 50.050
220526 97.019 -76.040 50.010
220527 97.043 -76.045 49.971
220528 97.068 -76.050 49.933
220529 97.092 -76.056 49.897
220530 97.117 -76.062 49.861
220531 97.142 -76.067 49.825
220532 97.166 -76.072 49.790
220533 97.191 -76.079 49.756
220534 97.216 -76.084 49.724
220535 97.240 -76.089 49.693
220536 97.265 -76.094 49.662
220537 97.299 -76.103 49.632
220538 97.337 -76.111 49.602
220539 97.376 -76.120 49.574
220540 97.406 -76.122 49.547
220541 97.437 -76.124 49.521
220542 97.467 -76.126 49.494
220543 97.498 -76.128 49.466
220544 97.528 -76.129 49.439
220545 97.560 -76.133 49.413
220546 97.593 -76.143 49.386
220547 97.627 -76.153 49.361
220548 97.660 -76.163 49.336
220549 97.693 -76.172 49.311
220550 97.699 -76.185 49.285
220551 97.701 -76.199 49.257
220552 97.704 -76.212 49.230
220553 97.706 -76.226 49.201
220554 97.709 -76.239 49.171
220555 97.712 -76.252 49.140
220556 97.714 -76.266 49.107
220557 97.717 -76.279 49.073
220558 97.719 -76.293 49.040
220559 97.744 -76.305 49.008
220560 97.785 -76.317 48.977
220561 97.822 -76.329 48.946
220562 97.864 -76.341 48.913
220563 97.908 -76.351 48.877
220564 97.951 -76.363 48.838
220565 97.995 -76.374 48.799
220566 98.038 -76.386 48.760
220567 98.086 -76.404 48.718
220568 98.143 -76.417 48.675
220569 98.199 -76.430 48.632
220570 98.213 -76.436 48.591
220571 98.245 -76.445 48.551
220572 98.277 -76.455 48.512
220573 98.309 -76.466 48.475
220574 98.338 -76.482 48.439
220575 98.368 -76.498 48.404
220576 98.397 -76.514 48.369
220577 98.452 -76.531 48.336
220578 98.485 -76.540 48.303
220579 98.517 -76.549 48.271
220580 98.550 -76.559 48.240
220581 98.583 -76.568 48.212
220582 98.616 -76.577 48.184
220583 98.648 -76.586 48.157
220584 98.672 -76.592 48.130
220585 98.695 -76.596 48.104
220586 98.718 -76.601 48.079
220587 98.735 -76.604 48.055
220588 98.753 -76.607 48.030
220589 98.769 -76.608 48.005
220590 98.777 -76.603 47.981
220591 98.785 -76.599 47.956
220592 98.793 -76.595 47.931
220593 98.801 -76.591 47.906
220594 98.812 -76.592 47.882
220595 98.825 -76.598 47.856
220596 98.838 -76.603 47.829
220597 98.851 -76.608 47.802
220598 98.864 -76.614 47.774
220599 98.879 -76.618 47.747
220600 98.901 -76.621 47.718
220601 98.922 -76.624 47.688
220602 98.944 -76.627 47.658
220603 98.966 -76.630 47.629
220604 98.987 -76.633 47.602
220605 99.011 -76.632 47.573
220606 99.027 -76.624 47.542
220607 99.045 -76.627 47.508
220608 99.065 -76.636 47.474
220609 99.084 -76.645 47.438
220610 99.103 -76.654 47.401
220611 99.133 -76.667 47.365
220612 99.164 -76.679 47.328
220613 99.195 -76.680 47.288
220614 99.225 -76.691 47.248
220615 99.255 -76.702 47.208
220616 99.292 -76.712 47.168
220617 99.332 -76.721 47.128
220618 99.370 -76.731 47.090
220619 99.409 -76.741 47.051
220620 99.428 -76.740 47.012
220621 99.446 -76.740 46.974
220622 99.465 -76.739 46.937
220623 99.483 -76.738 46.899
220624 99.501 -76.737 46.864
220625 99.520 -76.737 46.829
220626 99.535 -76.737 46.794
220627 99.544 -76.739 46.760
220628 99.553 -76.741 46.726
220629 99.561 -76.743 46.692
220630 99.570 -76.745 46.659
220631 99.586 -76.764 46.626
220632 99.603 -76.784 46.595
220633 99.612 -76.782 46.564
220634 99.621 -76.779 46.534
220635 99.630 -76.776 46.505
220636 99.650 -76.778 46.476
220637 99.672 -76.781 46.449
220638 99.693 -76.784 46.423
220639 99.714 -76.787 46.398
220640 99.736 -76.789 46.374
220641 99.757 -76.792 46.349
220642 99.779 -76.794 46.324
220643 99.800 -76.797 46.298
220644 99.822 -76.799 46.271
220645 99.843 -76.802 46.244
220646 99.865 -76.806 46.217
220647 99.887 -76.808 46.188
220648 99.908 -76.811 46.158
220649 99.936 -76.816 46.128
220650 99.968 -76.824 46.099
220651 99.999 -76.831 46.071
220652 100.031 -76.837 46.044
220653 100.062 -76.844 46.019
220654 100.093 -76.852 45.994
220655 100.125 -76.859 45.969
220656 100.156 -76.865 45.945
220657 100.187 -76.872 45.922
220658 100.219 -76.880 45.901
220659 100.250 -76.887 45.881
220660 100.282 -76.893 45.861
220661 100.313 -76.901 45.841
220662 100.333 -76.893 45.824
220663 100.363 -76.913 45.807
220664 100.388 -76.917 45.790
220665 100.413 -76.921 45.775
220666 100.434 -76.934 45.758
220667 100.453 -76.952 45.742
220668 100.470 -76.969 45.726
220669 100.488 -76.987 45.710
220670 100.506 -77.005 45.694
220671 100.525 -77.022 45.678
220672 100.542 -77.040 45.662
220673 100.560 -77.058 45.644
220674 100.578 -77.075 45.628
220675 100.596 -77.094 45.611
220676 100.613 -77.110 45.594
220677 100.622 -77.123 45.579
220678 100.631 -77.136 45.565
220679 100.645 -77.152 45.550
220680 100.665 -77.169 45.536
220681 100.685 -77.188 45.524
220682 100.705 -77.205 45.512
220683 100.725 -77.223 45.500
220684 100.749 -77.238 45.488
220685 100.776 -77.249 45.477
220686 100.804 -77.260 45.465
220687 100.831 -77.270 45.453
220688 100.858 -77.281 45.441
220689 100.885 -77.291 45.430
220690 100.912 -77.302 45.418
220691 100.939 -77.313 45.407
220692 100.967 -77.324 45.396
220693 100.994 -77.335 45.385
220694 101.019 -77.345 45.374
220695 101.040 -77.356 45.362
220696 101.060 -77.367 45.348
220697 101.080 -77.376 45.334
220698 101.101 -77.387 45.320
220699 101.121 -77.398 45.306
220700 101.142 -77.408 45.292
220701 101.162 -77.419 45.277
220702 101.184 -77.424 45.261
220703 101.211 -77.432 45.245
220704 101.216 -77.442 45.227
220705 101.215 -77.454 45.208
220706 101.213 -77.466 45.190
220707 101.211 -77.478 45.172
220708 101.210 -77.490 45.154
220709 101.209 -77.502 45.137
220710 101.211 -77.515 45.119
220711 101.212 -77.528 45.100
220712 101.213 -77.541 45.083
220713 101.219 -77.551 45.067
220714 101.237 -77.553 45.050
220715 101.265 -77.570 45.035
220716 101.302 -77.583 45.017
220717 101.338 -77.596 45.000
220718 101.374 -77.610 44.982
220719 101.410 -77.624 44.967
220720 101.446 -77.636 44.950
220721 101.482 -77.650 44.934
220722 101.519 -77.664 44.919
220723 101.555 -77.677 44.902
220724 101.591 -77.690 44.885
220725 101.627 -77.704 44.870
220726 101.668 -77.715 44.853
220727 101.710 -77.725 44.837
220728 101.754 -77.735 44.820
220729 101.796 -77.745 44.802
220730 101.840 -77.755 44.786
220731 101.880 -77.765 44.770
220732 101.904 -77.781 44.754
220733 101.941 -77.793 44.739
220734 101.940 -77.816 44.724
220735 101.939 -77.840 44.709
220736 101.953 -77.856 44.693
220737 101.982 -77.859 44.677
220738 102.010 -77.861 44.660
220739 102.039 -77.863 44.644
220740 102.068 -77.865 44.628
220741 102.097 -77.868 44.613
220742 102.126 -77.870 44.597
220743 102.150 -77.883 44.581
220744 102.174 -77.897 44.566
220745 102.197 -77.913 44.551
220746 102.243 -77.949 44.536
220747 102.273 -77.962 44.520
220748 102.304 -77.976 44.505
220749 102.334 -77.990 44.491
220750 102.364 -78.003 44.474
220751 102.394 -78.016 44.460
220752 102.416 -78.020 44.446
220753 102.438 -78.024 44.431
220754 102.459 -78.028 44.416
220755 102.481 -78.031 44.401
220756 102.503 -78.035 44.387
220757 102.525 -78.038 44.372
220758 102.546 -78.042 44.358
220759 102.576 -78.064 44.344
220760 102.605 -78.086 44.329
220761 102.634 -78.109 44.314
220762 102.664 -78.131 44.299
220763 102.694 -78.153 44.284
220764 102.723 -78.176 44.269
220765 102.753 -78.199 44.255
220766 102.783 -78.216 44.240
220767 102.822 -78.227 44.224
220768 102.863 -78.238 44.207
220769 102.903 -78.241 44.191
220770 102.946 -78.239 44.175
220771 102.987 -78.238 44.159
220772 103.029 -78.235 44.142
220773 103.071 -78.234 44.125
220774 103.113 -78.232 44.107
220775 103.155 -78.231 44.090
220776 103.196 -78.229 44.074
220777 103.239 -78.228 44.057
220778 103.281 -78.225 44.040
221017 92.229 -75.351 54.180
221018 92.252 -75.362 54.186
221019 92.275 -75.373 54.193
221020 92.298 -75.384 54.200
221021 92.321 -75.395 54.211
221022 92.344 -75.406 54.220
221023 92.367 -75.417 54.228
221024 92.390 -75.428 54.236
221025 92.413 -75.439 54.240
221026 92.434 -75.451 54.242
221027 92.444 -75.459 54.244
221028 92.464 -75.450 54.247
221029 92.485 -75.442 54.248
221030 92.505 -75.434 54.247
221031 92.524 -75.431 54.243
221032 92.543 -75.427 54.239
221033 92.563 -75.425 54.232
221034 92.593 -75.420 54.226
221035 92.624 -75.417 54.220
221036 92.654 -75.412 54.212
221037 92.686 -75.408 54.203
221039 92.747 -75.400 54.184
221040 92.778 -75.395 54.172
221041 92.809 -75.392 54.161
221042 92.840 -75.387 54.151
221043 92.870 -75.383 54.139
221044 92.901 -75.379 54.127
221045 92.932 -75.375 54.115
221046 92.957 -75.374 54.103
221047 92.975 -75.377 54.089
221048 92.994 -75.380 54.077
221049 93.012 -75.383 54.064
221050 93.030 -75.386 54.051
221051 93.049 -75.389 54.037
221052 93.067 -75.392 54.025
221064 93.319 -75.402 53.931
221065 93.342 -75.402 53.937
221066 93.365 -75.400 53.946
221067 93.389 -75.398 53.960
221068 93.412 -75.398 53.974
221069 93.435 -75.396 53.991
221070 93.459 -75.395 54.010
221071 93.482 -75.394 54.033
221072 93.505 -75.392 54.059
221073 93.529 -75.391 54.087
221074 93.564 -75.389 54.118
221075 93.603 -75.388 54.147
221076 93.642 -75.386 54.178
221077 93.681 -75.383 54.211
221078 93.734 -75.367 54.247
221079 93.767 -75.359 54.282
221080 93.791 -75.356 54.318
221081 93.815 -75.353 54.353
221082 93.839 -75.350 54.384
221083 93.862 -75.347 54.407
221084 93.887 -75.344 54.428
221085 93.913 -75.339 54.448
221086 93.939 -75.336 54.463
221087 93.964 -75.329 54.477
221088 93.984 -75.318 54.485
221089 94.007 -75.317 54.483
221090 94.031 -75.317 54.478
221091 94.054 -75.317 54.472
221092 94.077 -75.317 54.466
221093 94.101 -75.316 54.461
221094 94.124 -75.316 54.460
221095 94.149 -75.320 54.454
221096 94.177 -75.327 54.442
221097 94.203 -75.335 54.425
221098 94.230 -75.342 54.406
221099 94.257 -75.349 54.385
221100 94.283 -75.357 54.363
221101 94.310 -75.358 54.342
221102 94.334 -75.358 54.318
221103 94.358 -75.351 54.298
221104 94.382 -75.345 54.284
221105 94.406 -75.339 54.271
221106 94.429 -75.333 54.259
221107 94.450 -75.330 54.244
221108 94.467 -75.329 54.224
221109 94.483 -75.328 54.196
221110 94.500 -75.327 54.166
221111 94.518 -75.326 54.134
221112 94.535 -75.326 54.100
221113 94.551 -75.326 54.065
221114 94.565 -75.330 54.031
221115 94.579 -75.333 53.996
221116 94.597 -75.334 53.956
221117 94.614 -75.336 53.916
221118 94.632 -75.337 53.875
221119 94.649 -75.338 53.835
221120 94.667 -75.339 53.794
221121 94.685 -75.341 53.757
221122 94.706 -75.341 53.718
221123 94.732 -75.340 53.678
221124 94.726 -75.342 53.638
221125 94.716 -75.345 53.598
221126 94.707 -75.347 53.557
221127 94.698 -75.350 53.516
221128 94.688 -75.352 53.476
221129 94.681 -75.357 53.432
221130 94.674 -75.363 53.389
221131 94.668 -75.368 53.345
221132 94.661 -75.373 53.300
221133 94.654 -75.379 53.256
221134 94.652 -75.383 53.211
221135 94.670 -75.386 53.167
221136 94.688 -75.389 53.123
221137 94.712 -75.394 53.078
221138 94.743 -75.400 53.031
221139 94.773 -75.406 52.985
221140 94.803 -75.412 52.937
221141 94.833 -75.418 52.890
221142 94.863 -75.424 52.846
221143 94.894 -75.430 52.802
221144 94.924 -75.436 52.761
221145 94.951 -75.441 52.722
221146 94.976 -75.444 52.682
221147 95.002 -75.448 52.639
221148 95.028 -75.451 52.593
221149 95.053 -75.455 52.545
221150 95.079 -75.458 52.497
221151 95.105 -75.461 52.449
221152 95.137 -75.471 52.402
221153 95.172 -75.482 52.356
221154 95.206 -75.493 52.311
221155 95.240 -75.504 52.267
221156 95.274 -75.514 52.224
221157 95.308 -75.525 52.182
221158 95.342 -75.536 52.140
221159 95.376 -75.547 52.102
221160 95.410 -75.558 52.063
221161 95.430 -75.572 52.025
221162 95.451 -75.586 51.987
221163 95.472 -75.599 51.951
221164 95.493 -75.612 51.916
221165 95.514 -75.626 51.882
221166 95.539 -75.621 51.849
221167 95.569 -75.588 51.816
221168 95.570 -75.636 51.785
221169 95.597 -75.652 51.754
221170 95.639 -75.651 51.724
221171 95.681 -75.649 51.692
221172 95.723 -75.648 51.657
221173 95.764 -75.646 51.622
221174 95.804 -75.644 51.587
221175 95.839 -75.640 51.550
221176 95.874 -75.636 51.516
221177 95.909 -75.632 51.484
221178 95.945 -75.627 51.453
221179 95.980 -75.624 51.422
221180 96.017 -75.622 51.391
221181 96.048 -75.624 51.359
221182 96.040 -75.631 51.327
221183 96.032 -75.640 51.296
221184 96.058 -75.643 51.267
221185 96.101 -75.645 51.237
221186 96.144 -75.646 51.207
221187 96.187 -75.648 51.179
221188 96.234 -75.640 51.150
221189 96.268 -75.635 51.122
221190 96.285 -75.639 51.095
221191 96.303 -75.642 51.068
221192 96.320 -75.646 51.039
221193 96.341 -75.649 51.008
221194 96.363 -75.652 50.976
221195 96.385 -75.655 50.946
221196 96.406 -75.658 50.917
221197 96.428 -75.662 50.891
221198 96.448 -75.668 50.865
221199 96.469 -75.677 50.840
221200 96.487 -75.690 50.815
221201 96.504 -75.704 50.789
221202 96.520 -75.718 50.765
221203 96.536 -75.733 50.742
221204 96.552 -75.746 50.721
221205 96.569 -75.760 50.698
221206 96.585 -75.773 50.677
221207 96.601 -75.786 50.655
221208 96.618 -75.800 50.632
221209 96.633 -75.813 50.609
221210 96.644 -75.833 50.587
221211 96.660 -75.850 50.567
221212 96.683 -75.865 50.547
221213 96.705 -75.881 50.525
221214 96.728 -75.896 50.502
221215 96.750 -75.911 50.475
221216 96.772 -75.926 50.445
221217 96.795 -75.941 50.414
221218 96.823 -75.957 50.380
221219 96.855 -75.975 50.345
221220 96.856 -75.982 50.308
221221 96.882 -75.988 50.266
221222 96.907 -75.993 50.223
221223 96.932 -75.999 50.180
221224 96.956 -76.004 50.139
221225 96.981 -76.010 50.097
221226 97.006 -76.015 50.057
221227 97.030 -76.021 50.017
221228 97.054 -76.026 49.979
221229 97.079 -76.032 49.942
221230 97.104 -76.038 49.905
221231 97.129 -76.043 49.869
221232 97.153 -76.048 49.834
221233 97.178 -76.053 49.800
221234 97.203 -76.060 49.766
221235 97.227 -76.065 49.734
221236 97.252 -76.070 49.702
221237 97.276 -76.075 49.671
221238 97.306 -76.082 49.641
221239 97.345 -76.091 49.612
221240 97.375 -76.094 49.584
221241 97.405 -76.095 49.557
221242 97.436 -76.097 49.530
221243 97.466 -76.099 49.505
221244 97.497 -76.101 49.477
221245 97.527 -76.103 49.450
221246 97.558 -76.106 49.424
221247 97.592 -76.116 49.396
221248 97.625 -76.125 49.372
221249 97.658 -76.135 49.347
221250 97.660 -76.149 49.322
221251 97.647 -76.165 49.296
221252 97.649 -76.178 49.267
221253 97.652 -76.191 49.238
221254 97.654 -76.205 49.209
221255 97.657 -76.218 49.178
221256 97.660 -76.232 49.145
221257 97.678 -76.245 49.111
221258 97.706 -76.257 49.076
221259 97.733 -76.269 49.041
221260 97.761 -76.282 49.009
221261 97.804 -76.294 48.977
221262 97.849 -76.307 48.945
221263 97.887 -76.318 48.911
221264 97.924 -76.330 48.873
221265 97.961 -76.342 48.835
221266 97.999 -76.353 48.796
221267 98.037 -76.364 48.756
221268 98.094 -76.377 48.715
221269 98.150 -76.389 48.671
221270 98.200 -76.401 48.628
221271 98.219 -76.408 48.587
221272 98.250 -76.418 48.547
221273 98.281 -76.433 48.508
221274 98.310 -76.448 48.470
221275 98.340 -76.464 48.434
221276 98.369 -76.480 48.399
221277 98.407 -76.499 48.365
221278 98.449 -76.511 48.331
221279 98.482 -76.520 48.299
221280 98.515 -76.530 48.267
221281 98.548 -76.539 48.235
221282 98.580 -76.548 48.206
221283 98.613 -76.557 48.179
221284 98.645 -76.566 48.152
221285 98.678 -76.576 48.125
221286 98.711 -76.585 48.100
221287 98.743 -76.594 48.075
221288 98.769 -76.600 48.050
221289 98.783 -76.600 48.024
221290 98.791 -76.596 48.000
221291 98.799 -76.592 47.975
221292 98.808 -76.587 47.950
221293 98.816 -76.583 47.925
221294 98.824 -76.579 47.900
221295 98.833 -76.575 47.876
221296 98.843 -76.575 47.851
221297 98.856 -76.580 47.826
221298 98.868 -76.586 47.800
221299 98.881 -76.591 47.773
221300 98.895 -76.596 47.746
221301 98.916 -76.599 47.717
221302 98.938 -76.602 47.688
221303 98.959 -76.605 47.659
221304 98.981 -76.608 47.630
221305 99.003 -76.611 47.603
221306 99.024 -76.614 47.575
221307 99.045 -76.613 47.544
221308 99.064 -76.621 47.512
221309 99.084 -76.630 47.477
221310 99.103 -76.639 47.442
221311 99.122 -76.646 47.406
221312 99.148 -76.649 47.369
221313 99.179 -76.649 47.331
221314 99.209 -76.655 47.292
221315 99.239 -76.665 47.253
221316 99.269 -76.676 47.213
221317 99.299 -76.687 47.172
221318 99.329 -76.698 47.132
221319 99.362 -76.708 47.093
221320 99.402 -76.718 47.056
221321 99.440 -76.727 47.017
221322 99.465 -76.730 46.979
221323 99.483 -76.729 46.941
221324 99.501 -76.728 46.903
221325 99.519 -76.727 46.868
221326 99.536 -76.727 46.833
221327 99.545 -76.730 46.798
221328 99.554 -76.731 46.763
221329 99.563 -76.734 46.729
221330 99.572 -76.735 46.695
221331 99.581 -76.737 46.662
221332 99.593 -76.747 46.629
221333 99.607 -76.760 46.599
221334 99.616 -76.757 46.568
221335 99.632 -76.757 46.538
221336 99.653 -76.760 46.510
221337 99.675 -76.762 46.481
221338 99.697 -76.765 46.454
221339 99.718 -76.768 46.428
221340 99.739 -76.771 46.403
221341 99.761 -76.774 46.378
221342 99.782 -76.776 46.352
221343 99.804 -76.779 46.327
221344 99.826 -76.782 46.299
221345 99.847 -76.784 46.272
221346 99.868 -76.787 46.245
221347 99.890 -76.790 46.218
221348 99.912 -76.793 46.188
221349 99.933 -76.795 46.157
221350 99.963 -76.802 46.127
221351 99.994 -76.809 46.098
221352 100.025 -76.815 46.070
221353 100.057 -76.822 46.043
221354 100.088 -76.830 46.017
221355 100.119 -76.837 45.990
221356 100.151 -76.843 45.965
221357 100.182 -76.850 45.941
221358 100.213 -76.858 45.918
221359 100.245 -76.865 45.896
221360 100.276 -76.871 45.875
221361 100.304 -76.875 45.855
221362 100.319 -76.859 45.835
221363 100.339 -76.856 45.817
221364 100.374 -76.887 45.800
221365 100.404 -76.904 45.784
221366 100.429 -76.908 45.769
221367 100.450 -76.922 45.752
221368 100.468 -76.940 45.736
221369 100.485 -76.958 45.720
221370 100.503 -76.975 45.704
221371 100.522 -76.993 45.688
221372 100.540 -77.011 45.671
221373 100.557 -77.028 45.655
221374 100.575 -77.046 45.637
221375 100.594 -77.064 45.621
221376 100.612 -77.081 45.605
221377 100.629 -77.099 45.588
221378 100.641 -77.113 45.573
221379 100.650 -77.127 45.558
221380 100.665 -77.143 45.544
221381 100.686 -77.160 45.530
221382 100.706 -77.179 45.518
221383 100.726 -77.197 45.506
221384 100.751 -77.210 45.494
221385 100.778 -77.220 45.481
221386 100.805 -77.231 45.470
221387 100.833 -77.241 45.458
221388 100.860 -77.253 45.446
221389 100.887 -77.263 45.435
221390 100.914 -77.274 45.423
221391 100.941 -77.285 45.411
221392 100.969 -77.295 45.399
221393 100.996 -77.306 45.387
221394 101.023 -77.317 45.376
221395 101.049 -77.328 45.365
221396 101.070 -77.338 45.352
221397 101.090 -77.348 45.339
221398 101.111 -77.359 45.325
221399 101.131 -77.370 45.311
221400 101.151 -77.380 45.297
221401 101.172 -77.391 45.282
221402 101.193 -77.401 45.267
221403 101.213 -77.411 45.251
221404 101.217 -77.428 45.235
221405 101.218 -77.441 45.216
221406 101.220 -77.454 45.198
221407 101.221 -77.467 45.180
221408 101.222 -77.479 45.163
221409 101.223 -77.492 45.144
221410 101.225 -77.505 45.125
221411 101.227 -77.518 45.107
221412 101.228 -77.531 45.088
221413 101.241 -77.536 45.071
221414 101.258 -77.539 45.055
221415 101.276 -77.541 45.039
221416 101.281 -77.556 45.023
221417 101.318 -77.570 45.006
221418 101.353 -77.583 44.988
221419 101.390 -77.597 44.971
221420 101.426 -77.610 44.954
221421 101.462 -77.624 44.938
221422 101.498 -77.637 44.923
221423 101.534 -77.650 44.907
221424 101.570 -77.664 44.891
221425 101.607 -77.677 44.874
221426 101.650 -77.687 44.858
221427 101.693 -77.697 44.842
221428 101.736 -77.707 44.825
221429 101.779 -77.717 44.807
221430 101.822 -77.727 44.790
221431 101.865 -77.736 44.774
221432 101.908 -77.746 44.759
221433 101.928 -77.763 44.743
221434 101.941 -77.783 44.728
221435 101.942 -77.806 44.713
221436 101.944 -77.829 44.698
221437 101.972 -77.840 44.682
221438 102.001 -77.846 44.665
221439 102.030 -77.849 44.647
221440 102.058 -77.851 44.632
221441 102.087 -77.853 44.616
221442 102.116 -77.855 44.600
221443 102.145 -77.858 44.584
221444 102.169 -77.871 44.569
221445 102.192 -77.887 44.553
221446 102.216 -77.902 44.538
221447 102.256 -77.935 44.524
221448 102.286 -77.949 44.509
221449 102.317 -77.962 44.494
221450 102.347 -77.976 44.479
221451 102.377 -77.990 44.463
221452 102.405 -78.001 44.447
221453 102.427 -78.005 44.433
221454 102.449 -78.009 44.419
221455 102.470 -78.012 44.404
221456 102.492 -78.015 44.389
221457 102.514 -78.019 44.375
221458 102.536 -78.023 44.361
221459 102.557 -78.027 44.347
221460 102.583 -78.039 44.332
221461 102.612 -78.061 44.317
221462 102.642 -78.084 44.302
221463 102.671 -78.106 44.287
221464 102.701 -78.129 44.272
221465 102.730 -78.151 44.257
221466 102.761 -78.167 44.243
221467 102.795 -78.179 44.227
221468 102.834 -78.191 44.211
221469 102.874 -78.201 44.194
221470 102.914 -78.213 44.178
221471 102.954 -78.224 44.161
221472 102.994 -78.235 44.145
221473 103.036 -78.233 44.129
221474 103.078 -78.232 44.112
221475 103.119 -78.230 44.095
221476 103.161 -78.228 44.078
221477 103.204 -78.227 44.061
221478 103.245 -78.225 44.044
221718 92.216 -75.322 54.193
221719 92.239 -75.333 54.199
221720 92.262 -75.344 54.206
221721 92.285 -75.355 54.213
221722 92.308 -75.366 54.223
221723 92.332 -75.377 54.233
221724 92.355 -75.388 54.241
221725 92.377 -75.399 54.249
221726 92.401 -75.411 54.254
221727 92.424 -75.421 54.257
221728 92.434 -75.433 54.258
221729 92.454 -75.427 54.262
221730 92.474 -75.418 54.263
221731 92.494 -75.411 54.262
221732 92.513 -75.407 54.259
221733 92.533 -75.404 54.254
221734 92.555 -75.401 54.248
221735 92.586 -75.396 54.241
221736 92.617 -75.392 54.235
221737 92.647 -75.388 54.228
221738 92.678 -75.384 54.219
221739 92.709 -75.380 54.209
221740 92.740 -75.376 54.198
221741 92.771 -75.371 54.187
221742 92.802 -75.367 54.176
221743 92.833 -75.363 54.165
221744 92.863 -75.359 54.153
221745 92.894 -75.355 54.141
221746 92.925 -75.351 54.129
221747 92.949 -75.351 54.116
221748 92.967 -75.353 54.103
221749 92.986 -75.356 54.090
221750 93.004 -75.359 54.077
221751 93.022 -75.362 54.064
221752 93.041 -75.366 54.051
221753 93.059 -75.369 54.037
221754 93.077 -75.372 54.025
221765 93.318 -75.373 53.935
221766 93.341 -75.372 53.940
221767 93.364 -75.370 53.949
221768 93.388 -75.369 53.961
221769 93.411 -75.368 53.974
221770 93.434 -75.367 53.989
221771 93.458 -75.365 54.008
221772 93.481 -75.364 54.031
221773 93.504 -75.363 54.056
221774 93.528 -75.362 54.083
221775 93.563 -75.360 54.113
221776 93.602 -75.358 54.143
221777 93.641 -75.357 54.175
221778 93.680 -75.355 54.208
221779 93.728 -75.342 54.244
221780 93.760 -75.336 54.274
221781 93.783 -75.333 54.305
221782 93.807 -75.330 54.336
221783 93.831 -75.326 54.364
221784 93.857 -75.323 54.388
221785 93.884 -75.318 54.409
221786 93.910 -75.314 54.429
221787 93.936 -75.311 54.444
221788 93.960 -75.302 54.457
221789 93.978 -75.286 54.466
221790 94.001 -75.286 54.469
221791 94.025 -75.285 54.466
221792 94.048 -75.285 54.464
221793 94.071 -75.284 54.462
221794 94.097 -75.286 54.462
221795 94.122 -75.289 54.462
221796 94.147 -75.289 54.458
221797 94.172 -75.291 54.449
221798 94.199 -75.301 54.436
221799 94.226 -75.311 54.418
221800 94.254 -75.321 54.399
221801 94.280 -75.330 54.379
221802 94.307 -75.337 54.359
221803 94.332 -75.334 54.336
221804 94.356 -75.328 54.316
221805 94.380 -75.322 54.301
221806 94.403 -75.315 54.288
221807 94.427 -75.309 54.275
221808 94.451 -75.303 54.261
221809 94.469 -75.301 54.243
221810 94.486 -75.301 54.217
221811 94.503 -75.300 54.187
221812 94.520 -75.299 54.154
221813 94.537 -75.298 54.121
221814 94.554 -75.298 54.087
221815 94.571 -75.298 54.053
221816 94.588 -75.297 54.017
221817 94.605 -75.298 53.976
221818 94.623 -75.300 53.934
221819 94.640 -75.301 53.892
221820 94.658 -75.302 53.852
221821 94.676 -75.304 53.812
221822 94.693 -75.305 53.775
221823 94.710 -75.306 53.736
221824 94.733 -75.306 53.697
221825 94.729 -75.308 53.657
221826 94.719 -75.311 53.617
221827 94.711 -75.314 53.577
221828 94.704 -75.320 53.536
221829 94.697 -75.326 53.494
221830 94.691 -75.331 53.449
221831 94.684 -75.336 53.404
221832 94.678 -75.342 53.360
221833 94.671 -75.347 53.313
221834 94.665 -75.352 53.267
221835 94.671 -75.357 53.221
221836 94.688 -75.359 53.174
221837 94.706 -75.362 53.128
221838 94.724 -75.365 53.083
221839 94.751 -75.370 53.037
221840 94.781 -75.377 52.991
221841 94.811 -75.383 52.944
221842 94.841 -75.389 52.897
221843 94.872 -75.395 52.850
221844 94.902 -75.401 52.803
221845 94.932 -75.407 52.760
221846 94.958 -75.410 52.717
221847 94.983 -75.414 52.674
221848 95.009 -75.417 52.628
221849 95.035 -75.420 52.583
221850 95.060 -75.424 52.535
221851 95.086 -75.427 52.487
221852 95.112 -75.430 52.438
221853 95.143 -75.439 52.392
221854 95.177 -75.449 52.346
221855 95.211 -75.461 52.300
221856 95.245 -75.471 52.256
221857 95.279 -75.483 52.214
221858 95.313 -75.493 52.172
221859 95.347 -75.504 52.131
221860 95.382 -75.515 52.091
221861 95.404 -75.528 52.052
221862 95.425 -75.542 52.014
221863 95.446 -75.555 51.977
221864 95.467 -75.569 51.942
221865 95.487 -75.583 51.908
221866 95.509 -75.596 51.875
221867 95.529 -75.610 51.841
221868 95.550 -75.624 51.809
221869 95.580 -75.627 51.777
221870 95.617 -75.623 51.747
221871 95.652 -75.619 51.719
221872 95.688 -75.615 51.687
221873 95.723 -75.611 51.654
221874 95.758 -75.607 51.619
221875 95.794 -75.602 51.584
221876 95.829 -75.599 51.548
221877 95.865 -75.594 51.514
221878 95.900 -75.590 51.481
221879 95.935 -75.587 51.449
221880 95.970 -75.582 51.418
221881 96.008 -75.581 51.386
221882 96.045 -75.582 51.354
221883 96.072 -75.584 51.322
221884 96.064 -75.592 51.293
221885 96.071 -75.599 51.264
221886 96.114 -75.600 51.233
221887 96.157 -75.602 51.204
221888 96.200 -75.603 51.176
221889 96.243 -75.605 51.148
221890 96.281 -75.611 51.120
221891 96.300 -75.616 51.094
221892 96.317 -75.620 51.067
221893 96.335 -75.623 51.040
221894 96.353 -75.627 51.011
221895 96.374 -75.630 50.978
221896 96.396 -75.633 50.948
221897 96.417 -75.636 50.921
221898 96.439 -75.639 50.896
221899 96.461 -75.642 50.871
221900 96.481 -75.647 50.846
221901 96.498 -75.661 50.821
221902 96.515 -75.674 50.796
221903 96.531 -75.687 50.771
221904 96.547 -75.699 50.748
221905 96.564 -75.713 50.727
221906 96.581 -75.726 50.705
221907 96.597 -75.739 50.684
221908 96.614 -75.752 50.662
221909 96.630 -75.765 50.639
221910 96.641 -75.784 50.616
221911 96.652 -75.804 50.594
221912 96.663 -75.824 50.573
221913 96.680 -75.840 50.552
221914 96.703 -75.856 50.531
221915 96.725 -75.871 50.507
221916 96.747 -75.886 50.481
221917 96.770 -75.901 50.450
221918 96.792 -75.916 50.417
221919 96.815 -75.931 50.383
221920 96.837 -75.946 50.347
221921 96.877 -75.956 50.309
221922 96.901 -75.962 50.268
221923 96.925 -75.968 50.226
221924 96.949 -75.974 50.185
221925 96.973 -75.980 50.143
221926 96.998 -75.986 50.103
221927 97.022 -75.992 50.063
221928 97.046 -75.998 50.024
221929 97.070 -76.003 49.987
221930 97.094 -76.010 49.950
221931 97.118 -76.016 49.914
221932 97.143 -76.022 49.879
221933 97.167 -76.028 49.843
221934 97.191 -76.034 49.809
221935 97.215 -76.040 49.776
221936 97.239 -76.045 49.744
221937 97.263 -76.051 49.711
221938 97.288 -76.057 49.681
221939 97.313 -76.063 49.652
221940 97.344 -76.065 49.621
221941 97.374 -76.067 49.594
221942 97.405 -76.069 49.566
221943 97.435 -76.071 49.540
221944 97.466 -76.072 49.515
221945 97.496 -76.075 49.487
221946 97.527 -76.076 49.461
221947 97.557 -76.079 49.434
221948 97.590 -76.088 49.407
221949 97.624 -76.097 49.383
221950 97.641 -76.110 49.359
221951 97.615 -76.127 49.334
221952 97.595 -76.144 49.307
221953 97.597 -76.157 49.277
221954 97.600 -76.171 49.246
221955 97.611 -76.184 49.216
221956 97.639 -76.196 49.183
221957 97.667 -76.208 49.148
221958 97.695 -76.221 49.112
221959 97.722 -76.233 49.076
221960 97.750 -76.245 49.041
221961 97.778 -76.257 49.008
221962 97.820 -76.270 48.975
221963 97.866 -76.283 48.942
221964 97.912 -76.296 48.906
221965 97.951 -76.307 48.867
221966 97.989 -76.320 48.828
221967 98.026 -76.331 48.790
221968 98.047 -76.345 48.750
221969 98.101 -76.349 48.707
221970 98.158 -76.362 48.664
221971 98.192 -76.371 48.621
221972 98.223 -76.383 48.579
221973 98.253 -76.398 48.539
221974 98.283 -76.414 48.501
221975 98.312 -76.430 48.463
221976 98.342 -76.446 48.427
221977 98.371 -76.462 48.393
221978 98.414 -76.482 48.359
221979 98.447 -76.491 48.326
221980 98.480 -76.500 48.294
221981 98.512 -76.509 48.261
221982 98.545 -76.518 48.230
221983 98.577 -76.528 48.199
221984 98.610 -76.537 48.171
221985 98.643 -76.546 48.145
221986 98.675 -76.555 48.118
221987 98.708 -76.564 48.092
221988 98.740 -76.571 48.066
221989 98.773 -76.579 48.040
221990 98.794 -76.582 48.015
221991 98.811 -76.583 47.991
221992 98.822 -76.580 47.967
221993 98.831 -76.576 47.943
221994 98.839 -76.571 47.918
221995 98.847 -76.567 47.893
221996 98.855 -76.563 47.869
221997 98.863 -76.559 47.845
221998 98.873 -76.558 47.821
221999 98.886 -76.563 47.796
222000 98.899 -76.568 47.769
222001 98.912 -76.574 47.743
222002 98.932 -76.577 47.715
222003 98.953 -76.580 47.688
222004 98.975 -76.583 47.659
222005 98.997 -76.586 47.631
222006 99.018 -76.589 47.604
222007 99.040 -76.591 47.576
222008 99.060 -76.592 47.545
222009 99.079 -76.596 47.513
222010 99.098 -76.602 47.479
222011 99.117 -76.610 47.444
222012 99.136 -76.617 47.409
222013 99.162 -76.619 47.372
222014 99.193 -76.619 47.334
222015 99.222 -76.629 47.295
222016 99.253 -76.640 47.256
222017 99.283 -76.650 47.216
222018 99.313 -76.661 47.175
222019 99.343 -76.672 47.136
222020 99.373 -76.683 47.098
222021 99.404 -76.693 47.061
222022 99.433 -76.705 47.022
222023 99.472 -76.715 46.983
222024 99.501 -76.719 46.945
222025 99.519 -76.718 46.907
222026 99.537 -76.718 46.872
222027 99.546 -76.720 46.837
222028 99.555 -76.721 46.802
222029 99.564 -76.724 46.767
222030 99.573 -76.725 46.733
222031 99.582 -76.727 46.699
222032 99.591 -76.729 46.665
222033 99.600 -76.731 46.633
222034 99.614 -76.737 46.603
222035 99.636 -76.739 46.572
222036 99.657 -76.742 46.542
222037 99.679 -76.744 46.514
222038 99.700 -76.747 46.486
222039 99.722 -76.750 46.459
222040 99.743 -76.752 46.434
222041 99.765 -76.755 46.408
222042 99.786 -76.758 46.381
222043 99.807 -76.761 46.355
222044 99.829 -76.764 46.329
222045 99.851 -76.766 46.301
222046 99.872 -76.769 46.274
222047 99.894 -76.771 46.247
222048 99.915 -76.774 46.217
222049 99.936 -76.777 46.186
222050 99.958 -76.780 46.156
222051 99.989 -76.787 46.126
222052 100.020 -76.793 46.096
222053 100.051 -76.800 46.069
222054 100.083 -76.808 46.042
222055 100.114 -76.815 46.015
222056 100.145 -76.821 45.987
222057 100.177 -76.828 45.962
222058 100.208 -76.836 45.937
222059 100.240 -76.843 45.914
222060 100.271 -76.850 45.891
222061 100.293 -76.843 45.870
222062 100.307 -76.828 45.850
222063 100.321 -76.812 45.830
222064 100.346 -76.818 45.812
222065 100.381 -76.849 45.795
222066 100.416 -76.879 45.778
222067 100.445 -76.895 45.763
222068 100.465 -76.911 45.746
222069 100.483 -76.928 45.730
222070 100.501 -76.946 45.714
222071 100.519 -76.964 45.698
222072 100.537 -76.981 45.681
222073 100.555 -76.999 45.665
222074 100.573 -77.017 45.648
222075 100.591 -77.034 45.631
222076 100.609 -77.052 45.615
222077 100.627 -77.070 45.598
222078 100.644 -77.088 45.582
222079 100.661 -77.104 45.566
222080 100.670 -77.117 45.552
222081 100.687 -77.134 45.538
222082 100.707 -77.152 45.525
222083 100.727 -77.169 45.512
222084 100.754 -77.181 45.500
222085 100.781 -77.191 45.488
222086 100.808 -77.203 45.475
222087 100.835 -77.213 45.463
222088 100.862 -77.224 45.451
222089 100.890 -77.235 45.440
222090 100.917 -77.245 45.428
222091 100.944 -77.257 45.416
222092 100.971 -77.267 45.404
222093 100.998 -77.278 45.392
222094 101.025 -77.288 45.381
222095 101.053 -77.299 45.368
222096 101.079 -77.310 45.356
222097 101.100 -77.320 45.343
222098 101.120 -77.331 45.331
222099 101.141 -77.341 45.315
222100 101.162 -77.352 45.301
222101 101.182 -77.362 45.287
222102 101.202 -77.373 45.272
222103 101.218 -77.390 45.257
222104 101.229 -77.415 45.241
222105 101.233 -77.431 45.224
222106 101.234 -77.444 45.206
222107 101.236 -77.457 45.188
222108 101.237 -77.470 45.170
222109 101.238 -77.483 45.152
222110 101.240 -77.495 45.133
222111 101.241 -77.508 45.115
222112 101.244 -77.520 45.096
222113 101.262 -77.522 45.078
222114 101.279 -77.524 45.060
222115 101.297 -77.526 45.044
222116 101.315 -77.529 45.028
222117 101.332 -77.531 45.011
222118 101.333 -77.557 44.994
222119 101.370 -77.570 44.976
222120 101.405 -77.583 44.959
222121 101.442 -77.597 44.942
222122 101.478 -77.611 44.926
222123 101.514 -77.624 44.911
222124 101.550 -77.637 44.895
222125 101.589 -77.649 44.880
222126 101.632 -77.659 44.863
222127 101.675 -77.669 44.846
222128 101.718 -77.679 44.830
222129 101.761 -77.689 44.812
222130 101.804 -77.699 44.795
222131 101.842 -77.710 44.779
222132 101.879 -77.722 44.763
222133 101.915 -77.733 44.748
222134 101.930 -77.755 44.733
222135 101.946 -77.773 44.717
222136 101.948 -77.796 44.701
222137 101.963 -77.812 44.685
222138 101.991 -77.824 44.669
222139 102.020 -77.834 44.652
222140 102.049 -77.837 44.635
222141 102.078 -77.840 44.619
222142 102.106 -77.842 44.603
222143 102.135 -77.844 44.587
222144 102.164 -77.846 44.572
222145 102.187 -77.861 44.557
222146 102.211 -77.876 44.542
222147 102.234 -77.890 44.526
222148 102.257 -77.906 44.512
222149 102.299 -77.936 44.497
222150 102.329 -77.949 44.482
222151 102.360 -77.963 44.467
222152 102.390 -77.977 44.451
222153 102.416 -77.986 44.436
222154 102.438 -77.990 44.421
222155 102.460 -77.993 44.407
222156 102.481 -77.996 44.393
222157 102.504 -78.000 44.378
222158 102.525 -78.004 44.363
222159 102.547 -78.008 44.350
222160 102.569 -78.012 44.336
222161 102.591 -78.015 44.320
222162 102.619 -78.037 44.305
222163 102.649 -78.059 44.291
222164 102.678 -78.081 44.275
222165 102.708 -78.104 44.260
222166 102.739 -78.119 44.245
222167 102.769 -78.131 44.231
222168 102.807 -78.143 44.214
222169 102.846 -78.153 44.198
222170 102.886 -78.165 44.181
222171 102.926 -78.176 44.165
222172 102.966 -78.187 44.149
222173 103.006 -78.198 44.132
222174 103.046 -78.210 44.115
222175 103.086 -78.221 44.099
222176 103.126 -78.228 44.082
222177 103.168 -78.226 44.065
222178 103.210 -78.225 44.048
222179 103.252 -78.222 44.031
222419 92.204 -75.292 54.205
222420 92.227 -75.304 54.211
222421 92.249 -75.314 54.217
222422 92.273 -75.326 54.224
222423 92.296 -75.337 54.235
222424 92.319 -75.348 54.245
222425 92.342 -75.359 54.253
222426 92.365 -75.370 54.260
222427 92.388 -75.381 54.266
222428 92.411 -75.392 54.270
222429 92.428 -75.404 54.272
222430 92.443 -75.403 54.277
222431 92.463 -75.394 54.278
222432 92.484 -75.386 54.277
222433 92.503 -75.383 54.275
222434 92.522 -75.380 54.269
222435 92.548 -75.377 54.263
222436 92.579 -75.373 54.256
222437 92.609 -75.368 54.250
222438 92.640 -75.364 54.243
222439 92.671 -75.360 54.234
222440 92.702 -75.356 54.224
222441 92.733 -75.352 54.213
222442 92.764 -75.348 54.201
222443 92.794 -75.343 54.189
222444 92.826 -75.339 54.178
222445 92.856 -75.335 54.166
222446 92.887 -75.331 54.154
222447 92.918 -75.326 54.142
222448 92.941 -75.327 54.129
222449 92.959 -75.330 54.116
222450 92.978 -75.333 54.103
222451 92.996 -75.336 54.090
222452 93.014 -75.339 54.077
222453 93.032 -75.342 54.064
222454 93.051 -75.345 54.051
222455 93.069 -75.349 54.037
222456 93.087 -75.352 54.023
222466 93.317 -75.343 53.941
222467 93.340 -75.342 53.945
222468 93.363 -75.341 53.951
222469 93.387 -75.339 53.963
222470 93.410 -75.339 53.975
222471 93.433 -75.337 53.989
222472 93.457 -75.336 54.007
222473 93.480 -75.335 54.029
222474 93.503 -75.333 54.054
222475 93.526 -75.333 54.081
222476 93.562 -75.330 54.111
222477 93.601 -75.329 54.142
222478 93.640 -75.327 54.176
222479 93.679 -75.325 54.213
222480 93.723 -75.317 54.248
222481 93.753 -75.311 54.276
222482 93.776 -75.308 54.302
222483 93.801 -75.305 54.328
222484 93.828 -75.301 54.351
222485 93.854 -75.297 54.372
222486 93.880 -75.293 54.391
222487 93.907 -75.289 54.409
222488 93.933 -75.286 54.423
222489 93.956 -75.276 54.434
222490 93.974 -75.259 54.443
222491 93.995 -75.253 54.446
222492 94.018 -75.253 54.447
222493 94.044 -75.254 54.447
222494 94.069 -75.257 54.449
222495 94.096 -75.260 54.453
222496 94.121 -75.262 54.457
222497 94.146 -75.262 54.456
222498 94.170 -75.260 54.452
222499 94.195 -75.262 54.445
222500 94.222 -75.272 54.431
222501 94.249 -75.283 54.415
222502 94.276 -75.292 54.397
222503 94.303 -75.302 54.376
222504 94.329 -75.307 54.355
222505 94.353 -75.305 54.336
222506 94.377 -75.298 54.319
222507 94.401 -75.292 54.303
222508 94.424 -75.286 54.291
222509 94.448 -75.279 54.278
222510 94.471 -75.274 54.261
222511 94.488 -75.273 54.237
222512 94.505 -75.273 54.207
222513 94.522 -75.272 54.175
222514 94.539 -75.272 54.143
222515 94.556 -75.271 54.109
222516 94.572 -75.270 54.074
222517 94.586 -75.270 54.036
222518 94.600 -75.271 53.994
222519 94.619 -75.271 53.951
222520 94.638 -75.271 53.909
222521 94.657 -75.272 53.869
222522 94.676 -75.272 53.831
222523 94.696 -75.273 53.794
222524 94.715 -75.273 53.756
222525 94.734 -75.273 53.716
222526 94.734 -75.277 53.676
222527 94.728 -75.283 53.636
222528 94.721 -75.289 53.596
222529 94.714 -75.294 53.555
222530 94.707 -75.299 53.511
222531 94.701 -75.305 53.465
222532 94.694 -75.310 53.419
222533 94.688 -75.315 53.373
222534 94.681 -75.320 53.325
222535 94.675 -75.326 53.278
222536 94.689 -75.330 53.231
222537 94.707 -75.332 53.183
222538 94.724 -75.335 53.135
222539 94.743 -75.338 53.089
222540 94.760 -75.341 53.044
222541 94.789 -75.346 52.998
222542 94.820 -75.352 52.951
222543 94.850 -75.358 52.904
222544 94.880 -75.364 52.856
222545 94.910 -75.370 52.807
222546 94.938 -75.376 52.760
222547 94.965 -75.380 52.714
222548 94.990 -75.383 52.668
222549 95.016 -75.386 52.621
222550 95.042 -75.389 52.575
222551 95.067 -75.393 52.527
222552 95.093 -75.396 52.478
222553 95.119 -75.400 52.430
222554 95.148 -75.407 52.382
222555 95.182 -75.417 52.336
222556 95.216 -75.428 52.290
222557 95.251 -75.439 52.247
222558 95.284 -75.450 52.205
222559 95.319 -75.461 52.163
222560 95.353 -75.472 52.122
222561 95.378 -75.485 52.082
222562 95.399 -75.499 52.042
222563 95.420 -75.512 52.004
222564 95.441 -75.526 51.967
222565 95.461 -75.539 51.933
222566 95.482 -75.552 51.899
222567 95.503 -75.566 51.867
222568 95.525 -75.579 51.834
222569 95.549 -75.590 51.802
222570 95.577 -75.587 51.771
222571 95.607 -75.581 51.740
222572 95.642 -75.577 51.712
222573 95.678 -75.574 51.680
222574 95.713 -75.569 51.648
222575 95.748 -75.565 51.615
222576 95.784 -75.561 51.581
222577 95.819 -75.557 51.547
222578 95.854 -75.553 51.513
222579 95.890 -75.549 51.479
222580 95.925 -75.545 51.447
222581 95.961 -75.541 51.415
222582 95.998 -75.541 51.383
222583 96.035 -75.542 51.351
222584 96.073 -75.542 51.319
222585 96.096 -75.545 51.289
222586 96.088 -75.553 51.260
222587 96.126 -75.555 51.231
222588 96.169 -75.557 51.202
222589 96.212 -75.558 51.174
222590 96.253 -75.565 51.146
222591 96.289 -75.577 51.119
222592 96.310 -75.586 51.093
222593 96.330 -75.595 51.067
222594 96.349 -75.601 51.040
222595 96.367 -75.604 51.011
222596 96.386 -75.608 50.981
222597 96.407 -75.611 50.952
222598 96.429 -75.614 50.925
222599 96.450 -75.617 50.901
222600 96.471 -75.620 50.876
222601 96.490 -75.623 50.852
222602 96.509 -75.630 50.827
222603 96.527 -75.640 50.802
222604 96.543 -75.652 50.779
222605 96.560 -75.665 50.756
222606 96.577 -75.678 50.734
222607 96.593 -75.692 50.713
222608 96.609 -75.705 50.691
222609 96.626 -75.718 50.668
222610 96.638 -75.736 50.645
222611 96.649 -75.756 50.621
222612 96.659 -75.775 50.599
222613 96.670 -75.795 50.578
222614 96.681 -75.815 50.557
222615 96.700 -75.831 50.536
222616 96.723 -75.846 50.512
222617 96.745 -75.861 50.485
222618 96.767 -75.876 50.455
222619 96.790 -75.891 50.420
222620 96.812 -75.906 50.384
222621 96.834 -75.921 50.347
222622 96.893 -75.927 50.310
222623 96.924 -75.933 50.269
222624 96.948 -75.939 50.229
222625 96.972 -75.944 50.189
222626 96.997 -75.950 50.148
222627 97.021 -75.956 50.108
222628 97.045 -75.963 50.070
222629 97.069 -75.969 50.033
222630 97.093 -75.975 49.996
222631 97.117 -75.981 49.960
222632 97.142 -75.986 49.924
222633 97.165 -75.992 49.889
222634 97.189 -75.998 49.854
222635 97.214 -76.004 49.819
222636 97.238 -76.010 49.786
222637 97.262 -76.016 49.753
222638 97.286 -76.022 49.720
222639 97.307 -76.029 49.690
222640 97.330 -76.039 49.660
222641 97.370 -76.041 49.631
222642 97.384 -76.041 49.603
222643 97.404 -76.042 49.575
222644 97.434 -76.044 49.549
222645 97.465 -76.046 49.523
222646 97.495 -76.047 49.497
222647 97.526 -76.050 49.471
222648 97.556 -76.051 49.445
222649 97.589 -76.060 49.419
222650 97.621 -76.070 49.394
222651 97.596 -76.088 49.369
222652 97.571 -76.105 49.343
222653 97.546 -76.123 49.315
222654 97.572 -76.135 49.285
222655 97.600 -76.148 49.253
222656 97.628 -76.160 49.220
222657 97.656 -76.173 49.186
222658 97.683 -76.185 49.150
222659 97.711 -76.197 49.113
222660 97.739 -76.209 49.074
222661 97.767 -76.221 49.039
222662 97.795 -76.234 49.004
222663 97.837 -76.246 48.969
222664 97.883 -76.259 48.934
222665 97.929 -76.272 48.897
222666 97.974 -76.285 48.859
222667 98.016 -76.297 48.820
222668 98.054 -76.308 48.781
222669 98.072 -76.323 48.739
222670 98.096 -76.333 48.697
222671 98.166 -76.335 48.654
222672 98.196 -76.349 48.612
222673 98.225 -76.365 48.570
222674 98.254 -76.381 48.530
222675 98.284 -76.397 48.491
222676 98.314 -76.413 48.454
222677 98.343 -76.429 48.419
222678 98.373 -76.445 48.386
222679 98.412 -76.461 48.353
222680 98.444 -76.471 48.320
222681 98.477 -76.480 48.287
222682 98.509 -76.489 48.255
222683 98.542 -76.499 48.223
222684 98.575 -76.508 48.192
222685 98.607 -76.516 48.164
222686 98.640 -76.523 48.136
222687 98.672 -76.530 48.110
222688 98.704 -76.538 48.083
222689 98.736 -76.546 48.058
222690 98.769 -76.552 48.032
222691 98.791 -76.555 48.006
222692 98.808 -76.556 47.983
222693 98.826 -76.558 47.959
222694 98.843 -76.558 47.935
222695 98.860 -76.559 47.911
222696 98.870 -76.555 47.887
222697 98.878 -76.552 47.863
222698 98.886 -76.547 47.840
222699 98.895 -76.543 47.815
222700 98.903 -76.540 47.791
222701 98.916 -76.546 47.765
222702 98.929 -76.551 47.740
222703 98.947 -76.555 47.714
222704 98.969 -76.558 47.687
222705 98.991 -76.561 47.659
222706 99.012 -76.564 47.631
222707 99.033 -76.565 47.605
222708 99.053 -76.566 47.576
222709 99.073 -76.567 47.545
222710 99.092 -76.567 47.513
222711 99.112 -76.574 47.479
222712 99.131 -76.580 47.445
222713 99.149 -76.587 47.410
222714 99.176 -76.589 47.374
222715 99.206 -76.592 47.335
222716 99.236 -76.602 47.297
222717 99.267 -76.614 47.258
222718 99.297 -76.624 47.219
222719 99.327 -76.635 47.179
222720 99.357 -76.646 47.140
222721 99.387 -76.657 47.103
222722 99.417 -76.668 47.066
222723 99.447 -76.679 47.026
222724 99.478 -76.690 46.988
222725 99.508 -76.700 46.950
222726 99.538 -76.708 46.913
222727 99.548 -76.710 46.877
222728 99.558 -76.712 46.841
222729 99.569 -76.714 46.806
222730 99.581 -76.715 46.771
222731 99.592 -76.717 46.737
222732 99.604 -76.718 46.702
222733 99.616 -76.720 46.669
222734 99.627 -76.721 46.637
222735 99.642 -76.722 46.606
222736 99.661 -76.724 46.577
222737 99.683 -76.726 46.548
222738 99.704 -76.729 46.520
222739 99.725 -76.732 46.492
222740 99.747 -76.734 46.465
222741 99.768 -76.737 46.439
222742 99.790 -76.740 46.412
222743 99.811 -76.743 46.385
222744 99.833 -76.746 46.358
222745 99.854 -76.748 46.331
222746 99.876 -76.751 46.303
222747 99.897 -76.753 46.275
222748 99.919 -76.756 46.246
222749 99.940 -76.759 46.217
222750 99.962 -76.762 46.187
222751 99.983 -76.765 46.156
222752 100.015 -76.771 46.126
222753 100.046 -76.778 46.095
222754 100.077 -76.785 46.067
222755 100.109 -76.793 46.040
222756 100.140 -76.799 46.013
222757 100.171 -76.806 45.985
222758 100.203 -76.814 45.959
222759 100.234 -76.821 45.934
222760 100.266 -76.828 45.910
222761 100.281 -76.812 45.887
222762 100.295 -76.796 45.865
222763 100.310 -76.781 45.845
222764 100.324 -76.765 45.826
222765 100.352 -76.781 45.807
222766 100.388 -76.812 45.789
222767 100.423 -76.842 45.773
222768 100.456 -76.873 45.757
222769 100.480 -76.897 45.741
222770 100.498 -76.916 45.724
222771 100.516 -76.934 45.708
222772 100.534 -76.952 45.691
222773 100.552 -76.970 45.675
222774 100.570 -76.987 45.659
222775 100.588 -77.005 45.642
222776 100.606 -77.023 45.625
222777 100.624 -77.041 45.609
222778 100.642 -77.058 45.592
222779 100.660 -77.076 45.576
222780 100.678 -77.094 45.560
222781 100.689 -77.108 45.546
222782 100.707 -77.125 45.532
222783 100.728 -77.142 45.519
222784 100.756 -77.153 45.507
222785 100.783 -77.163 45.495
222786 100.810 -77.174 45.482
222787 100.837 -77.185 45.468
222788 100.864 -77.195 45.456
222789 100.892 -77.207 45.445
222790 100.919 -77.217 45.433
222791 100.946 -77.228 45.422
222792 100.973 -77.238 45.409
222793 101.000 -77.249 45.398
222794 101.028 -77.260 45.387
222795 101.055 -77.271 45.376
222796 101.082 -77.282 45.363
222797 101.109 -77.292 45.350
222798 101.130 -77.303 45.337
222799 101.150 -77.313 45.323
222800 101.171 -77.324 45.308
222801 101.191 -77.334 45.293
222802 101.212 -77.345 45.278
222803 101.223 -77.369 45.263
222804 101.234 -77.394 45.247
222805 101.245 -77.419 45.230
222806 101.249 -77.434 45.213
222807 101.250 -77.447 45.196
222808 101.251 -77.460 45.178
222809 101.252 -77.473 45.160
222810 101.254 -77.486 45.141
222811 101.255 -77.498 45.123
222812 101.265 -77.506 45.104
222813 101.283 -77.508 45.085
222814 101.301 -77.511 45.067
222815 101.318 -77.513 45.049
222816 101.336 -77.515 45.032
222817 101.353 -77.517 45.016
222818 101.371 -77.519 44.999
222819 101.389 -77.521 44.982
222820 101.385 -77.557 44.964
222821 101.421 -77.570 44.946
222822 101.458 -77.583 44.930
222823 101.493 -77.597 44.914
222824 101.530 -77.611 44.898
222825 101.571 -77.621 44.883
222826 101.614 -77.631 44.867
222827 101.657 -77.641 44.850
222828 101.700 -77.651 44.834
222829 101.737 -77.663 44.817
222830 101.773 -77.674 44.800
222831 101.810 -77.686 44.784
222832 101.846 -77.698 44.768
222833 101.882 -77.710 44.752
222834 101.918 -77.722 44.736
222835 101.932 -77.743 44.721
222836 101.946 -77.765 44.706
222837 101.954 -77.785 44.690
222838 101.983 -77.796 44.674
222839 102.011 -77.807 44.657
222840 102.039 -77.818 44.640
222841 102.068 -77.825 44.623
222842 102.097 -77.827 44.608
222843 102.125 -77.830 44.592
222844 102.154 -77.832 44.576
222845 102.183 -77.835 44.560
222846 102.206 -77.850 44.546
222847 102.230 -77.865 44.531
222848 102.253 -77.880 44.515
222849 102.276 -77.894 44.500
222850 102.312 -77.923 44.485
222851 102.342 -77.937 44.470
222852 102.373 -77.950 44.456
222853 102.403 -77.964 44.441
222854 102.427 -77.970 44.426
222855 102.449 -77.974 44.411
222856 102.471 -77.978 44.396
222857 102.493 -77.981 44.382
222858 102.515 -77.985 44.367
222859 102.536 -77.989 44.353
222860 102.558 -77.993 44.339
222861 102.580 -77.996 44.323
222862 102.602 -78.000 44.309
222863 102.626 -78.012 44.294
222864 102.656 -78.034 44.279
222865 102.685 -78.056 44.264
222866 102.716 -78.070 44.249
222867 102.747 -78.082 44.233
222868 102.778 -78.094 44.218
222869 102.818 -78.106 44.202
222870 102.858 -78.117 44.185
222871 102.898 -78.128 44.168
222872 102.938 -78.139 44.152
222873 102.978 -78.150 44.136
222874 103.018 -78.162 44.119
222875 103.057 -78.173 44.103
222876 103.098 -78.184 44.086
222877 103.138 -78.195 44.070
222878 103.177 -78.206 44.053
222879 103.217 -78.217 44.035
222880 103.258 -78.220 44.019
223120 92.191 -75.264 54.214
223121 92.214 -75.274 54.220
223122 92.237 -75.286 54.227
223123 92.260 -75.297 54.234
223124 92.283 -75.308 54.245
223125 92.306 -75.319 54.255
223126 92.330 -75.330 54.264
223127 92.352 -75.341 54.272
223128 92.375 -75.352 54.279
223129 92.399 -75.363 54.283
223130 92.421 -75.374 54.285
223131 92.433 -75.380 54.290
223132 92.453 -75.370 54.293
223133 92.473 -75.363 54.292
223134 92.493 -75.360 54.290
223135 92.512 -75.357 54.284
223136 92.541 -75.353 54.277
223137 92.572 -75.348 54.271
223138 92.602 -75.345 54.265
223139 92.633 -75.340 54.258
223140 92.664 -75.336 54.249
223141 92.695 -75.332 54.239
223142 92.726 -75.328 54.227
223143 92.757 -75.323 54.215
223144 92.787 -75.320 54.203
223145 92.818 -75.315 54.192
223146 92.849 -75.311 54.180
223147 92.880 -75.307 54.168
223148 92.911 -75.303 54.155
223149 92.933 -75.304 54.142
223150 92.951 -75.307 54.130
223151 92.970 -75.309 54.117
223152 92.988 -75.313 54.104
223153 93.006 -75.316 54.091
223154 93.024 -75.319 54.077
223155 93.043 -75.323 54.063
223156 93.061 -75.326 54.049
223157 93.082 -75.326 54.035
223167 93.315 -75.314 53.947
223168 93.339 -75.313 53.949
223169 93.362 -75.311 53.955
223170 93.385 -75.311 53.965
223171 93.409 -75.309 53.977
223172 93.432 -75.308 53.990
223173 93.455 -75.307 54.007
223174 93.478 -75.305 54.028
223175 93.502 -75.304 54.052
223176 93.525 -75.303 54.080
223177 93.561 -75.301 54.112
223178 93.600 -75.299 54.144
223179 93.639 -75.298 54.180
223180 93.678 -75.295 54.222
223181 93.717 -75.292 54.258
223182 93.745 -75.288 54.284
223183 93.772 -75.284 54.306
223184 93.798 -75.280 54.328
223185 93.825 -75.276 54.345
223186 93.851 -75.272 54.362
223187 93.877 -75.268 54.378
223188 93.904 -75.264 54.394
223189 93.930 -75.261 54.405
223190 93.952 -75.248 54.415
223191 93.970 -75.232 54.422
223192 93.990 -75.223 54.426
223193 94.016 -75.225 54.430
223194 94.042 -75.228 54.433
223195 94.068 -75.230 54.436
223196 94.094 -75.233 54.443
223197 94.120 -75.236 54.450
223198 94.145 -75.236 54.453
223199 94.169 -75.233 54.453
223200 94.192 -75.231 54.451
223201 94.217 -75.232 54.442
223202 94.244 -75.243 54.431
223203 94.271 -75.253 54.417
223204 94.298 -75.264 54.397
223205 94.325 -75.269 54.378
223206 94.349 -75.268 54.359
223207 94.373 -75.267 54.341
223208 94.398 -75.267 54.322
223209 94.422 -75.263 54.306
223210 94.445 -75.257 54.292
223211 94.469 -75.250 54.277
223212 94.490 -75.246 54.254
223213 94.507 -75.245 54.226
223214 94.524 -75.245 54.195
223215 94.541 -75.245 54.162
223216 94.556 -75.245 54.127
223217 94.570 -75.245 54.092
223218 94.584 -75.245 54.055
223219 94.599 -75.245 54.013
223220 94.613 -75.245 53.970
223221 94.630 -75.245 53.928
223222 94.649 -75.245 53.888
223223 94.668 -75.246 53.851
223224 94.688 -75.246 53.814
223225 94.707 -75.246 53.776
223226 94.728 -75.248 53.736
223227 94.744 -75.251 53.695
223228 94.738 -75.257 53.654
223229 94.731 -75.262 53.613
223230 94.724 -75.267 53.570
223231 94.717 -75.273 53.526
223232 94.711 -75.278 53.479
223233 94.704 -75.283 53.431
223234 94.698 -75.289 53.383
223235 94.691 -75.295 53.335
223236 94.689 -75.299 53.288
223237 94.707 -75.302 53.241
223238 94.725 -75.305 53.194
223239 94.743 -75.308 53.145
223240 94.761 -75.311 53.097
223241 94.779 -75.314 53.051
223242 94.797 -75.317 53.006
223243 94.828 -75.323 52.959
223244 94.858 -75.329 52.911
223245 94.888 -75.335 52.862
223246 94.918 -75.341 52.812
223247 94.946 -75.345 52.762
223248 94.971 -75.348 52.713
223249 94.997 -75.352 52.665
223250 95.023 -75.355 52.617
223251 95.049 -75.359 52.568
223252 95.074 -75.362 52.519
223253 95.100 -75.366 52.471
223254 95.126 -75.369 52.422
223255 95.154 -75.374 52.374
223256 95.188 -75.386 52.327
223257 95.222 -75.396 52.282
223258 95.256 -75.407 52.239
223259 95.290 -75.418 52.197
223260 95.324 -75.429 52.155
223261 95.352 -75.441 52.113
223262 95.373 -75.455 52.074
223263 95.394 -75.468 52.034
223264 95.414 -75.482 51.995
223265 95.435 -75.496 51.958
223266 95.456 -75.509 51.925
223267 95.477 -75.522 51.892
223268 95.501 -75.533 51.859
223269 95.526 -75.543 51.828
223270 95.550 -75.554 51.796
223271 95.578 -75.551 51.764
223272 95.606 -75.543 51.733
223273 95.635 -75.536 51.704
223274 95.668 -75.532 51.673
223275 95.703 -75.527 51.642
223276 95.739 -75.524 51.611
223277 95.774 -75.519 51.578
223278 95.809 -75.515 51.545
223279 95.845 -75.511 51.511
223280 95.880 -75.507 51.478
223281 95.915 -75.503 51.446
223282 95.952 -75.501 51.413
223283 95.989 -75.501 51.381
223284 96.026 -75.502 51.350
223285 96.063 -75.502 51.318
223286 96.101 -75.502 51.288
223287 96.120 -75.505 51.259
223288 96.138 -75.510 51.230
223289 96.182 -75.512 51.202
223290 96.223 -75.517 51.174
223291 96.259 -75.530 51.146
223292 96.295 -75.543 51.120
223293 96.318 -75.552 51.093
223294 96.339 -75.561 51.067
223295 96.359 -75.570 51.041
223296 96.380 -75.578 51.013
223297 96.399 -75.586 50.984
223298 96.418 -75.589 50.957
223299 96.440 -75.592 50.931
223300 96.459 -75.595 50.907
223301 96.479 -75.598 50.883
223302 96.498 -75.600 50.858
223303 96.517 -75.608 50.833
223304 96.535 -75.615 50.809
223305 96.553 -75.624 50.786
223306 96.572 -75.631 50.764
223307 96.589 -75.644 50.742
223308 96.605 -75.657 50.720
223309 96.622 -75.671 50.698
223310 96.635 -75.687 50.675
223311 96.646 -75.707 50.651
223312 96.656 -75.727 50.627
223313 96.667 -75.746 50.604
223314 96.678 -75.766 50.583
223315 96.688 -75.786 50.561
223316 96.699 -75.806 50.540
223317 96.721 -75.821 50.515
223318 96.743 -75.836 50.487
223319 96.765 -75.851 50.456
223320 96.788 -75.866 50.422
223321 96.810 -75.881 50.385
223322 96.842 -75.894 50.348
223323 96.906 -75.898 50.311
223324 96.947 -75.903 50.271
223325 96.971 -75.909 50.232
223326 96.996 -75.916 50.194
223327 97.019 -75.922 50.154
223328 97.043 -75.927 50.116
223329 97.068 -75.933 50.078
223330 97.092 -75.939 50.042
223331 97.116 -75.945 50.006
223332 97.140 -75.951 49.970
223333 97.164 -75.957 49.935
223334 97.188 -75.963 49.900
223335 97.213 -75.969 49.865
223336 97.237 -75.975 49.830
223337 97.259 -75.981 49.796
223338 97.281 -75.988 49.763
223339 97.302 -75.995 49.730
223340 97.323 -76.002 49.700
223341 97.348 -76.016 49.670
223342 97.384 -76.019 49.640
223343 97.427 -76.019 49.612
223344 97.451 -76.018 49.584
223345 97.460 -76.019 49.557
223346 97.468 -76.019 49.532
223347 97.494 -76.021 49.506
223348 97.525 -76.023 49.480
223349 97.555 -76.025 49.455
223350 97.588 -76.032 49.431
223351 97.577 -76.048 49.406
223352 97.569 -76.069 49.381
223353 97.581 -76.083 49.354
223354 97.582 -76.092 49.324
223355 97.590 -76.110 49.292
223356 97.616 -76.124 49.259
223357 97.645 -76.136 49.225
223358 97.672 -76.148 49.189
223359 97.700 -76.160 49.151
223360 97.728 -76.173 49.113
223361 97.756 -76.185 49.073
223362 97.784 -76.198 49.036
223363 97.811 -76.210 49.000
223364 97.854 -76.223 48.963
223365 97.899 -76.235 48.925
223366 97.945 -76.248 48.888
223367 97.991 -76.260 48.849
223368 98.036 -76.273 48.810
223369 98.081 -76.286 48.770
223370 98.096 -76.301 48.728
223371 98.096 -76.326 48.686
223372 98.180 -76.344 48.644
223373 98.211 -76.354 48.602
223374 98.236 -76.363 48.561
223375 98.262 -76.372 48.521
223376 98.287 -76.381 48.482
223377 98.315 -76.395 48.445
223378 98.345 -76.411 48.411
223379 98.375 -76.427 48.378
223380 98.409 -76.442 48.346
223381 98.441 -76.451 48.313
223382 98.474 -76.460 48.281
223383 98.506 -76.467 48.249
223384 98.539 -76.474 48.216
223385 98.571 -76.482 48.185
223386 98.603 -76.489 48.155
223387 98.636 -76.497 48.127
223388 98.668 -76.504 48.101
223389 98.701 -76.511 48.075
223390 98.733 -76.519 48.049
223391 98.766 -76.527 48.023
223392 98.788 -76.529 47.998
223393 98.805 -76.530 47.974
223394 98.822 -76.531 47.951
223395 98.840 -76.532 47.927
223396 98.857 -76.533 47.904
223397 98.874 -76.533 47.881
223398 98.892 -76.535 47.858
223399 98.909 -76.536 47.834
223400 98.917 -76.531 47.809
223401 98.925 -76.527 47.786
223402 98.934 -76.523 47.762
223403 98.946 -76.529 47.737
223404 98.963 -76.533 47.713
223405 98.985 -76.536 47.686
223406 99.006 -76.539 47.659
223407 99.026 -76.539 47.632
223408 99.046 -76.540 47.604
223409 99.066 -76.541 47.574
223410 99.086 -76.541 47.543
223411 99.106 -76.542 47.511
223412 99.125 -76.544 47.478
223413 99.144 -76.551 47.443
223414 99.163 -76.558 47.408
223415 99.190 -76.560 47.372
223416 99.220 -76.566 47.336
223417 99.251 -76.577 47.298
223418 99.281 -76.588 47.260
223419 99.311 -76.599 47.221
223420 99.341 -76.609 47.181
223421 99.371 -76.620 47.143
223422 99.401 -76.631 47.106
223423 99.431 -76.642 47.069
223424 99.462 -76.652 47.030
223425 99.491 -76.664 46.992
223426 99.521 -76.676 46.955
223427 99.552 -76.689 46.918
223428 99.582 -76.699 46.881
223429 99.594 -76.701 46.845
223430 99.606 -76.702 46.810
223431 99.617 -76.704 46.775
223432 99.629 -76.705 46.740
223433 99.640 -76.707 46.706
223434 99.652 -76.708 46.672
223435 99.663 -76.711 46.642
223436 99.679 -76.712 46.611
223437 99.695 -76.712 46.582
223438 99.711 -76.712 46.554
223439 99.729 -76.714 46.526
223440 99.751 -76.716 46.498
223441 99.772 -76.719 46.470
223442 99.793 -76.721 46.444
223443 99.815 -76.724 46.417
223444 99.837 -76.727 46.389
223445 99.858 -76.730 46.361
223446 99.879 -76.733 46.333
223447 99.901 -76.735 46.305
223448 99.922 -76.738 46.276
223449 99.944 -76.740 46.246
223450 99.966 -76.743 46.217
223451 99.987 -76.746 46.187
223452 100.009 -76.749 46.156
223453 100.041 -76.756 46.125
223454 100.072 -76.763 46.094
223455 100.104 -76.771 46.066
223456 100.135 -76.777 46.038
223457 100.167 -76.784 46.010
223458 100.198 -76.791 45.983
223459 100.229 -76.799 45.955
223460 100.255 -76.798 45.930
223461 100.269 -76.782 45.907
223462 100.283 -76.765 45.884
223463 100.298 -76.749 45.861
223464 100.312 -76.734 45.841
223465 100.326 -76.718 45.821
223466 100.359 -76.743 45.803
223467 100.394 -76.774 45.785
223468 100.429 -76.804 45.768
223469 100.456 -76.840 45.752
223470 100.483 -76.875 45.734
223471 100.506 -76.897 45.718
223472 100.528 -76.915 45.702
223473 100.551 -76.934 45.686
223474 100.570 -76.953 45.669
223475 100.587 -76.972 45.652
223476 100.605 -76.990 45.636
223477 100.622 -77.009 45.619
223478 100.639 -77.028 45.603
223479 100.657 -77.047 45.586
223480 100.675 -77.064 45.570
223481 100.693 -77.082 45.554
223482 100.708 -77.098 45.540
223483 100.731 -77.113 45.526
223484 100.758 -77.124 45.512
223485 100.785 -77.135 45.500
223486 100.813 -77.146 45.487
223487 100.840 -77.157 45.475
223488 100.867 -77.167 45.462
223489 100.894 -77.178 45.450
223490 100.921 -77.188 45.438
223491 100.949 -77.199 45.427
223492 100.976 -77.210 45.416
223493 101.003 -77.221 45.403
223494 101.030 -77.232 45.392
223495 101.057 -77.242 45.382
223496 101.085 -77.253 45.371
223497 101.112 -77.263 45.359
223498 101.139 -77.275 45.345
223499 101.160 -77.285 45.331
223500 101.181 -77.295 45.316
223501 101.201 -77.306 45.301
223502 101.218 -77.322 45.286
223503 101.229 -77.347 45.270
223504 101.240 -77.372 45.255
223505 101.250 -77.397 45.237
223506 101.261 -77.423 45.220
223507 101.264 -77.438 45.203
223508 101.265 -77.451 45.186
223509 101.267 -77.464 45.167
223510 101.268 -77.476 45.148
223511 101.269 -77.489 45.129
223512 101.286 -77.492 45.111
223513 101.304 -77.494 45.093
223514 101.322 -77.496 45.075
223515 101.339 -77.498 45.057
223516 101.357 -77.501 45.038
223517 101.375 -77.503 45.021
223518 101.392 -77.505 45.004
223519 101.410 -77.508 44.987
223520 101.428 -77.510 44.970
223521 101.446 -77.512 44.952
223522 101.437 -77.557 44.934
223523 101.473 -77.570 44.918
223524 101.510 -77.583 44.901
223525 101.553 -77.593 44.886
223526 101.596 -77.604 44.870
223527 101.632 -77.615 44.854
223528 101.668 -77.627 44.837
223529 101.704 -77.639 44.821
223530 101.741 -77.651 44.805
223531 101.777 -77.662 44.788
223532 101.813 -77.674 44.772
223533 101.849 -77.686 44.757
223534 101.886 -77.698 44.741
223535 101.920 -77.711 44.724
223536 101.934 -77.732 44.710
223537 101.948 -77.754 44.695
223538 101.973 -77.769 44.679
223539 102.002 -77.780 44.663
223540 102.030 -77.790 44.646
223541 102.058 -77.801 44.629
223542 102.087 -77.812 44.613
223543 102.116 -77.815 44.597
223544 102.144 -77.818 44.582
223545 102.173 -77.820 44.566
223546 102.201 -77.824 44.550
223547 102.225 -77.839 44.535
223548 102.248 -77.854 44.520
223549 102.271 -77.868 44.505
223550 102.295 -77.884 44.490
223551 102.318 -77.899 44.475
223552 102.356 -77.924 44.461
223553 102.386 -77.937 44.446
223554 102.416 -77.950 44.431
223555 102.439 -77.955 44.416
223556 102.460 -77.959 44.401
223557 102.482 -77.962 44.386
223558 102.504 -77.966 44.372
223559 102.526 -77.970 44.357
223560 102.547 -77.974 44.342
223561 102.569 -77.978 44.328
223562 102.591 -77.981 44.312
223563 102.613 -77.985 44.297
223564 102.634 -77.989 44.283
223565 102.663 -78.009 44.267
223566 102.694 -78.022 44.253
223567 102.725 -78.034 44.238
223568 102.756 -78.046 44.222
223569 102.790 -78.058 44.205
223570 102.830 -78.069 44.189
223571 102.870 -78.080 44.173
223572 102.910 -78.091 44.156
223573 102.950 -78.103 44.139
223574 102.990 -78.114 44.122
223575 103.030 -78.125 44.106
223576 103.070 -78.136 44.090
223577 103.109 -78.147 44.073
223578 103.149 -78.159 44.057
223579 103.189 -78.169 44.040
223580 103.229 -78.181 44.023
223821 92.178 -75.234 54.222
223822 92.202 -75.245 54.229
223823 92.224 -75.257 54.236
223824 92.247 -75.267 54.243
223825 92.271 -75.279 54.254
223826 92.294 -75.290 54.264
223827 92.317 -75.301 54.274
223828 92.340 -75.312 54.282
223829 92.363 -75.323 54.289
223830 92.386 -75.334 54.294
223831 92.409 -75.345 54.297
223832 92.422 -75.355 54.303
223833 92.443 -75.347 54.306
223834 92.463 -75.339 54.306
223835 92.482 -75.336 54.305
223836 92.503 -75.333 54.299
223837 92.533 -75.329 54.292
223838 92.564 -75.324 54.286
223839 92.595 -75.320 54.279
223840 92.626 -75.316 54.272
223841 92.657 -75.312 54.263
223842 92.688 -75.308 54.254
223843 92.719 -75.304 54.242
223844 92.749 -75.299 54.230
223845 92.780 -75.295 54.218
223846 92.811 -75.291 54.207
223847 92.842 -75.287 54.194
223848 92.873 -75.283 54.182
223849 92.904 -75.279 54.169
223850 92.925 -75.280 54.156
223851 92.943 -75.283 54.144
223852 92.961 -75.286 54.131
223853 92.980 -75.289 54.118
223854 92.998 -75.292 54.105
223855 93.016 -75.296 54.091
223856 93.035 -75.299 54.076
223857 93.058 -75.298 54.062
223858 93.081 -75.297 54.049
223859 93.104 -75.295 54.034
223867 93.291 -75.286 53.956
223868 93.314 -75.285 53.954
223869 93.337 -75.283 53.956
223870 93.361 -75.282 53.960
223871 93.384 -75.281 53.970
223872 93.407 -75.279 53.980
223873 93.431 -75.278 53.992
223874 93.454 -75.277 54.009
223875 93.477 -75.276 54.029
223876 93.500 -75.274 54.053
223877 93.524 -75.273 54.082
223878 93.560 -75.272 54.115
223879 93.599 -75.270 54.150
223880 93.638 -75.268 54.190
223881 93.677 -75.266 54.235
223882 93.713 -75.264 54.272
223883 93.742 -75.262 54.300
223884 93.769 -75.258 54.321
223885 93.795 -75.254 54.339
223886 93.822 -75.251 54.351
223887 93.848 -75.247 54.364
223888 93.874 -75.243 54.375
223889 93.901 -75.239 54.385
223890 93.927 -75.236 54.395
223891 93.947 -75.222 54.402
223892 93.968 -75.210 54.407
223893 93.991 -75.201 54.411
223894 94.015 -75.198 54.416
223895 94.041 -75.201 54.420
223896 94.067 -75.204 54.425
223897 94.093 -75.206 54.433
223898 94.119 -75.209 54.442
223899 94.144 -75.209 54.449
223900 94.168 -75.207 54.453
223901 94.191 -75.204 54.454
223902 94.215 -75.202 54.451
223903 94.240 -75.204 54.445
223904 94.267 -75.214 54.436
223905 94.294 -75.224 54.420
223906 94.320 -75.231 54.402
223907 94.345 -75.230 54.382
223908 94.369 -75.230 54.364
223909 94.394 -75.229 54.343
223910 94.418 -75.229 54.323
223911 94.442 -75.228 54.305
223912 94.467 -75.227 54.292
223913 94.491 -75.220 54.270
223914 94.509 -75.218 54.243
223915 94.526 -75.218 54.213
223916 94.540 -75.218 54.180
223917 94.554 -75.218 54.145
223918 94.568 -75.218 54.110
223919 94.582 -75.219 54.072
223920 94.597 -75.219 54.030
223921 94.611 -75.219 53.988
223922 94.625 -75.219 53.948
223923 94.641 -75.220 53.909
223924 94.661 -75.220 53.872
223925 94.680 -75.220 53.835
223926 94.699 -75.220 53.797
223927 94.722 -75.223 53.757
223928 94.747 -75.226 53.714
223929 94.748 -75.230 53.672
223930 94.741 -75.236 53.629
223931 94.735 -75.241 53.584
223932 94.728 -75.246 53.539
223933 94.721 -75.251 53.492
223934 94.714 -75.258 53.443
223935 94.708 -75.263 53.395
223936 94.701 -75.268 53.347
223937 94.708 -75.272 53.299
223938 94.725 -75.275 53.251
223939 94.744 -75.278 53.205
223940 94.761 -75.280 53.156
223941 94.779 -75.283 53.108
223942 94.797 -75.286 53.061
223943 94.815 -75.289 53.014
223944 94.836 -75.292 52.967
223945 94.866 -75.298 52.918
223946 94.897 -75.305 52.868
223947 94.927 -75.311 52.818
223948 94.953 -75.314 52.766
223949 94.978 -75.318 52.715
223950 95.004 -75.321 52.665
223951 95.030 -75.325 52.616
223952 95.055 -75.328 52.566
223953 95.082 -75.331 52.517
223954 95.107 -75.335 52.467
223955 95.133 -75.338 52.418
223956 95.159 -75.342 52.369
223957 95.193 -75.353 52.322
223958 95.228 -75.364 52.277
223959 95.261 -75.375 52.233
223960 95.296 -75.386 52.190
223961 95.326 -75.398 52.149
223962 95.346 -75.411 52.107
223963 95.367 -75.425 52.067
223964 95.388 -75.439 52.028
223965 95.409 -75.452 51.989
223966 95.430 -75.465 51.953
223967 95.454 -75.476 51.919
223968 95.478 -75.486 51.886
223969 95.502 -75.497 51.853
223970 95.527 -75.508 51.822
223971 95.551 -75.518 51.789
223972 95.579 -75.514 51.757
223973 95.608 -75.507 51.726
223974 95.636 -75.500 51.696
223975 95.665 -75.493 51.666
223976 95.693 -75.486 51.637
223977 95.729 -75.482 51.606
223978 95.764 -75.478 51.574
223979 95.800 -75.474 51.542
223980 95.835 -75.470 51.510
223981 95.870 -75.465 51.478
223982 95.906 -75.461 51.445
223983 95.942 -75.460 51.412
223984 95.979 -75.461 51.380
223985 96.017 -75.461 51.349
223986 96.054 -75.461 51.317
223987 96.091 -75.461 51.288
223988 96.128 -75.462 51.259
223989 96.151 -75.465 51.230
223990 96.192 -75.470 51.202
223991 96.229 -75.483 51.174
223992 96.265 -75.496 51.147
223993 96.301 -75.509 51.121
223994 96.327 -75.519 51.095
223995 96.347 -75.527 51.069
223996 96.368 -75.536 51.042
223997 96.388 -75.545 51.015
223998 96.408 -75.554 50.988
223999 96.429 -75.564 50.962
224000 96.448 -75.570 50.937
224001 96.467 -75.572 50.913
224002 96.486 -75.575 50.889
224003 96.506 -75.578 50.865
224004 96.525 -75.584 50.840
224005 96.543 -75.593 50.817
224006 96.562 -75.600 50.794
224007 96.580 -75.608 50.772
224008 96.599 -75.616 50.751
224009 96.617 -75.624 50.728
224010 96.631 -75.639 50.705
224011 96.643 -75.659 50.681
224012 96.653 -75.678 50.658
224013 96.664 -75.698 50.633
224014 96.674 -75.718 50.610
224015 96.685 -75.737 50.587
224016 96.696 -75.757 50.565
224017 96.707 -75.776 50.543
224018 96.718 -75.796 50.517
224019 96.740 -75.811 50.488
224020 96.763 -75.826 50.457
224021 96.785 -75.841 50.422
224022 96.808 -75.856 50.386
224023 96.855 -75.865 50.349
224024 96.920 -75.869 50.312
224025 96.970 -75.874 50.275
224026 96.994 -75.880 50.237
224027 97.018 -75.886 50.200
224028 97.042 -75.892 50.162
224029 97.067 -75.898 50.125
224030 97.091 -75.904 50.088
224031 97.115 -75.909 50.051
224032 97.139 -75.916 50.016
224033 97.163 -75.922 49.982
224034 97.187 -75.928 49.947
224035 97.211 -75.934 49.912
224036 97.233 -75.941 49.877
224037 97.254 -75.947 49.843
224038 97.276 -75.954 49.808
224039 97.297 -75.962 49.774
224040 97.318 -75.969 49.741
224041 97.341 -75.977 49.710
224042 97.365 -75.991 49.679
224043 97.398 -75.999 49.649
224044 97.441 -75.997 49.621
224045 97.484 -75.996 49.593
224046 97.519 -75.995 49.566
224047 97.527 -75.996 49.540
224048 97.535 -75.997 49.515
224049 97.543 -75.997 49.490
224050 97.554 -75.998 49.466
224051 97.591 -76.014 49.442
224052 97.603 -76.028 49.418
224053 97.614 -76.043 49.392
224054 97.627 -76.057 49.365
224055 97.615 -76.061 49.333
224056 97.616 -76.075 49.299
224057 97.634 -76.100 49.264
224058 97.661 -76.112 49.229
224059 97.689 -76.125 49.190
224060 97.717 -76.137 49.152
224061 97.745 -76.149 49.112
224062 97.773 -76.161 49.072
224063 97.800 -76.173 49.032
224064 97.828 -76.186 48.993
224065 97.870 -76.198 48.955
224066 97.916 -76.211 48.916
224067 97.962 -76.224 48.878
224068 98.007 -76.237 48.839
224069 98.053 -76.250 48.798
224070 98.099 -76.263 48.757
224071 98.150 -76.288 48.716
224072 98.144 -76.332 48.675
224073 98.194 -76.358 48.633
224074 98.233 -76.369 48.593
224075 98.258 -76.378 48.551
224076 98.284 -76.387 48.512
224077 98.309 -76.396 48.474
224078 98.334 -76.405 48.437
224079 98.359 -76.414 48.403
224080 98.382 -76.418 48.370
224081 98.407 -76.420 48.338
224082 98.438 -76.426 48.305
224083 98.471 -76.433 48.274
224084 98.503 -76.441 48.243
224085 98.535 -76.448 48.211
224086 98.568 -76.455 48.180
224087 98.600 -76.463 48.149
224088 98.632 -76.470 48.120
224089 98.665 -76.478 48.091
224090 98.697 -76.485 48.064
224091 98.729 -76.492 48.038
224092 98.762 -76.500 48.013
224093 98.784 -76.503 47.989
224094 98.802 -76.504 47.966
224095 98.819 -76.505 47.943
224096 98.837 -76.505 47.920
224097 98.854 -76.507 47.898
224098 98.871 -76.508 47.875
224099 98.888 -76.508 47.853
224100 98.906 -76.509 47.829
224101 98.923 -76.510 47.805
224102 98.941 -76.511 47.782
224103 98.956 -76.511 47.758
224104 98.964 -76.507 47.735
224105 98.979 -76.511 47.711
224106 99.000 -76.514 47.685
224107 99.020 -76.514 47.658
224108 99.040 -76.514 47.630
224109 99.059 -76.515 47.601
224110 99.079 -76.515 47.571
224111 99.099 -76.516 47.540
224112 99.119 -76.517 47.508
224113 99.138 -76.517 47.475
224114 99.157 -76.522 47.441
224115 99.177 -76.529 47.406
224116 99.204 -76.530 47.371
224117 99.234 -76.540 47.335
224118 99.265 -76.551 47.298
224119 99.295 -76.561 47.260
224120 99.325 -76.573 47.222
224121 99.355 -76.583 47.183
224122 99.385 -76.594 47.146
224123 99.415 -76.605 47.109
224124 99.444 -76.618 47.071
224125 99.474 -76.630 47.033
224126 99.503 -76.642 46.995
224127 99.533 -76.654 46.958
224128 99.565 -76.668 46.922
224129 99.600 -76.680 46.885
224130 99.630 -76.690 46.849
224131 99.642 -76.692 46.813
224132 99.653 -76.693 46.779
224133 99.665 -76.695 46.744
224134 99.677 -76.696 46.710
224135 99.688 -76.698 46.677
224136 99.700 -76.699 46.647
224137 99.714 -76.700 46.617
224138 99.731 -76.701 46.588
224139 99.747 -76.702 46.560
224140 99.763 -76.702 46.532
224141 99.779 -76.702 46.503
224142 99.797 -76.703 46.476
224143 99.819 -76.706 46.449
224144 99.840 -76.709 46.421
224145 99.862 -76.712 46.393
224146 99.883 -76.715 46.364
224147 99.905 -76.717 46.335
224148 99.926 -76.720 46.306
224149 99.947 -76.723 46.276
224150 99.969 -76.725 46.246
224151 99.991 -76.728 46.217
224152 100.012 -76.730 46.187
224153 100.036 -76.734 46.155
224154 100.067 -76.741 46.124
224155 100.099 -76.749 46.094
224156 100.130 -76.755 46.065
224157 100.161 -76.762 46.036
224158 100.193 -76.769 46.008
224159 100.224 -76.777 45.980
224160 100.243 -76.767 45.953
224161 100.258 -76.751 45.928
224162 100.272 -76.735 45.904
224163 100.286 -76.719 45.880
224164 100.300 -76.703 45.858
224165 100.315 -76.687 45.837
224166 100.331 -76.675 45.817
224167 100.366 -76.706 45.798
224168 100.401 -76.737 45.780
224169 100.428 -76.771 45.763
224170 100.456 -76.806 45.747
224171 100.482 -76.840 45.729
224172 100.509 -76.875 45.713
224173 100.532 -76.897 45.696
224174 100.554 -76.915 45.680
224175 100.577 -76.934 45.663
224176 100.599 -76.952 45.646
224177 100.622 -76.970 45.630
224178 100.641 -76.988 45.613
224179 100.659 -77.007 45.596
224180 100.676 -77.027 45.579
224181 100.694 -77.046 45.563
224182 100.711 -77.065 45.549
224183 100.732 -77.090 45.533
224184 100.760 -77.085 45.519
224185 100.787 -77.099 45.505
224186 100.815 -77.113 45.492
224187 100.842 -77.126 45.480
224188 100.869 -77.138 45.468
224189 100.896 -77.150 45.456
224190 100.923 -77.160 45.444
224191 100.951 -77.171 45.432
224192 100.978 -77.182 45.420
224193 101.005 -77.192 45.409
224194 101.032 -77.203 45.398
224195 101.059 -77.214 45.387
224196 101.087 -77.225 45.376
224197 101.114 -77.235 45.365
224198 101.141 -77.246 45.353
224199 101.168 -77.257 45.340
224200 101.190 -77.267 45.326
224201 101.211 -77.278 45.311
224202 101.223 -77.301 45.294
224203 101.234 -77.326 45.280
224204 101.245 -77.351 45.265
224205 101.256 -77.376 45.248
224206 101.266 -77.401 45.231
224207 101.277 -77.426 45.213
224208 101.280 -77.441 45.195
224209 101.281 -77.454 45.177
224210 101.282 -77.467 45.158
224211 101.290 -77.476 45.139
224212 101.308 -77.478 45.120
224213 101.325 -77.480 45.101
224214 101.343 -77.483 45.083
224215 101.360 -77.485 45.064
224216 101.378 -77.486 45.046
224217 101.396 -77.489 45.028
224218 101.413 -77.491 45.010
224219 101.431 -77.493 44.993
224220 101.449 -77.495 44.976
224221 101.467 -77.498 44.959
224222 101.484 -77.500 44.941
224223 101.502 -77.502 44.923
224224 101.490 -77.556 44.907
224225 101.527 -77.568 44.890
224226 101.563 -77.580 44.874
224227 101.599 -77.592 44.858
224228 101.635 -77.603 44.841
224229 101.672 -77.615 44.824
224230 101.708 -77.627 44.807
224231 101.744 -77.639 44.792
224232 101.780 -77.651 44.777
224233 101.817 -77.662 44.761
224234 101.853 -77.674 44.745
224235 101.889 -77.686 44.729
224236 101.921 -77.699 44.713
224237 101.936 -77.721 44.699
224238 101.949 -77.743 44.683
224239 101.987 -77.751 44.667
224240 102.020 -77.763 44.651
224241 102.049 -77.774 44.635
224242 102.078 -77.784 44.618
224243 102.106 -77.796 44.602
224244 102.134 -77.804 44.587
224245 102.164 -77.806 44.571
224246 102.192 -77.808 44.556
224247 102.220 -77.813 44.540
224248 102.243 -77.828 44.525
224249 102.267 -77.843 44.510
224250 102.290 -77.858 44.495
224251 102.313 -77.873 44.481
224252 102.337 -77.887 44.466
224253 102.369 -77.910 44.452
224254 102.399 -77.924 44.437
224255 102.428 -77.936 44.421
224256 102.450 -77.940 44.407
224257 102.471 -77.943 44.392
224258 102.493 -77.947 44.377
224259 102.515 -77.951 44.362
224260 102.537 -77.955 44.346
224261 102.558 -77.959 44.331
224262 102.580 -77.962 44.316
224263 102.602 -77.966 44.301
224264 102.624 -77.969 44.286
224265 102.647 -77.972 44.272
224266 102.674 -77.979 44.257
224267 102.703 -77.986 44.242
224268 102.734 -77.997 44.227
224269 102.764 -78.009 44.210
224270 102.802 -78.021 44.193
224271 102.842 -78.032 44.177
224272 102.882 -78.044 44.160
224273 102.922 -78.055 44.143
224274 102.962 -78.065 44.126
224275 103.001 -78.077 44.110
224276 103.042 -78.088 44.094
224277 103.082 -78.100 44.078
224278 103.121 -78.110 44.061
224279 103.161 -78.122 44.045
224280 103.201 -78.133 44.028
224281 103.241 -78.144 44.011
224522 92.166 -75.205 54.230
224523 92.189 -75.217 54.237
224524 92.212 -75.227 54.244
224525 92.235 -75.239 54.251
224526 92.258 -75.250 54.261
224527 92.281 -75.261 54.272
224528 92.304 -75.272 54.283
224529 92.327 -75.283 54.291
224530 92.350 -75.294 54.299
224531 92.373 -75.305 54.304
224532 92.397 -75.316 54.309
224533 92.415 -75.327 54.315
224534 92.432 -75.323 54.318
224535 92.452 -75.315 54.320
224536 92.472 -75.312 54.319
224537 92.496 -75.309 54.314
224538 92.526 -75.305 54.307
224539 92.557 -75.301 54.300
224540 92.588 -75.296 54.294
224541 92.619 -75.292 54.287
224542 92.650 -75.288 54.278
224543 92.681 -75.284 54.269
224544 92.712 -75.279 54.256
224545 92.742 -75.276 54.244
224546 92.773 -75.271 54.232
224547 92.804 -75.267 54.220
224548 92.835 -75.263 54.208
224549 92.866 -75.259 54.195
224550 92.897 -75.254 54.182
224551 92.917 -75.257 54.170
224552 92.935 -75.260 54.157
224553 92.953 -75.263 54.145
224554 92.972 -75.266 54.132
224555 92.990 -75.270 54.118
224556 93.010 -75.271 54.105
224557 93.033 -75.270 54.090
224558 93.056 -75.269 54.075
224559 93.080 -75.267 54.062
224560 93.103 -75.266 54.047
224568 93.289 -75.256 53.965
224569 93.313 -75.255 53.963
224570 93.336 -75.254 53.963
224571 93.359 -75.252 53.967
224572 93.383 -75.251 53.976
224573 93.406 -75.250 53.986
224574 93.429 -75.248 53.997
224575 93.452 -75.248 54.013
224576 93.476 -75.246 54.032
224577 93.499 -75.245 54.057
224578 93.522 -75.244 54.086
224579 93.559 -75.242 54.120
224580 93.598 -75.240 54.159
224581 93.637 -75.239 54.201
224582 93.675 -75.236 54.247
224583 93.707 -75.236 54.286
224584 93.738 -75.234 54.316
224585 93.766 -75.231 54.340
224586 93.792 -75.227 54.358
224587 93.819 -75.224 54.368
224588 93.845 -75.220 54.378
224589 93.871 -75.217 54.388
224590 93.898 -75.214 54.397
224591 93.924 -75.210 54.404
224592 93.947 -75.201 54.410
224593 93.970 -75.192 54.413
224594 93.992 -75.183 54.415
224595 94.015 -75.175 54.419
224596 94.040 -75.175 54.422
224597 94.066 -75.177 54.425
224598 94.092 -75.179 54.429
224599 94.118 -75.182 54.438
224600 94.143 -75.182 54.445
224601 94.167 -75.180 54.450
224602 94.190 -75.178 54.455
224603 94.214 -75.176 54.457
224604 94.237 -75.173 54.455
224605 94.262 -75.175 54.448
224606 94.289 -75.185 54.442
224607 94.316 -75.193 54.427
224608 94.340 -75.192 54.407
224609 94.365 -75.192 54.388
224610 94.389 -75.192 54.366
224611 94.414 -75.191 54.342
224612 94.438 -75.190 54.320
224613 94.463 -75.189 54.307
224614 94.487 -75.189 54.285
224615 94.510 -75.189 54.259
224616 94.525 -75.191 54.229
224617 94.538 -75.192 54.196
224618 94.553 -75.192 54.162
224619 94.567 -75.192 54.126
224620 94.581 -75.192 54.088
224621 94.595 -75.192 54.047
224622 94.609 -75.193 54.007
224623 94.623 -75.193 53.968
224624 94.637 -75.193 53.931
224625 94.652 -75.193 53.893
224626 94.672 -75.194 53.856
224627 94.691 -75.194 53.818
224628 94.716 -75.197 53.777
224629 94.741 -75.200 53.733
224630 94.758 -75.204 53.689
224631 94.751 -75.209 53.644
224632 94.745 -75.214 53.598
224633 94.738 -75.220 53.552
224634 94.731 -75.226 53.505
224635 94.724 -75.231 53.457
224636 94.718 -75.236 53.410
224637 94.711 -75.242 53.362
224638 94.726 -75.245 53.313
224639 94.744 -75.248 53.265
224640 94.762 -75.251 53.217
224641 94.780 -75.253 53.168
224642 94.797 -75.256 53.120
224643 94.815 -75.259 53.072
224644 94.833 -75.261 53.023
224645 94.851 -75.264 52.975
224646 94.875 -75.269 52.927
224647 94.905 -75.275 52.876
224648 94.934 -75.280 52.825
224649 94.960 -75.284 52.772
224650 94.985 -75.287 52.720
224651 95.011 -75.290 52.668
224652 95.037 -75.294 52.617
224653 95.062 -75.297 52.566
224654 95.088 -75.301 52.515
224655 95.114 -75.304 52.465
224656 95.139 -75.308 52.415
224657 95.166 -75.311 52.366
224658 95.199 -75.321 52.319
224659 95.233 -75.332 52.274
224660 95.267 -75.342 52.230
224661 95.300 -75.354 52.186
224662 95.320 -75.367 52.144
224663 95.341 -75.381 52.103
224664 95.362 -75.395 52.062
224665 95.383 -75.408 52.023
224666 95.407 -75.419 51.985
224667 95.431 -75.430 51.948
224668 95.455 -75.440 51.914
224669 95.479 -75.451 51.881
224670 95.503 -75.461 51.848
224671 95.528 -75.472 51.817
224672 95.552 -75.483 51.784
224673 95.580 -75.477 51.751
224674 95.609 -75.470 51.720
224675 95.637 -75.463 51.691
224676 95.666 -75.456 51.662
224677 95.694 -75.449 51.633
224678 95.723 -75.442 51.603
224679 95.754 -75.436 51.571
224680 95.790 -75.432 51.540
224681 95.825 -75.428 51.509
224682 95.861 -75.424 51.478
224683 95.896 -75.420 51.446
224684 95.933 -75.420 51.413
224685 95.970 -75.420 51.381
224686 96.008 -75.420 51.350
224687 96.045 -75.421 51.318
224688 96.082 -75.421 51.289
224689 96.119 -75.422 51.260
224690 96.161 -75.429 51.232
224691 96.198 -75.436 51.203
224692 96.235 -75.449 51.177
224693 96.271 -75.462 51.151
224694 96.307 -75.475 51.123
224695 96.335 -75.486 51.098
224696 96.356 -75.494 51.072
224697 96.376 -75.503 51.045
224698 96.396 -75.511 51.018
224699 96.416 -75.523 50.992
224700 96.436 -75.536 50.967
224701 96.456 -75.547 50.943
224702 96.475 -75.550 50.920
224703 96.495 -75.553 50.896
224704 96.514 -75.556 50.872
224705 96.533 -75.561 50.849
224706 96.551 -75.569 50.825
224707 96.570 -75.577 50.802
224708 96.588 -75.585 50.780
224709 96.607 -75.593 50.758
224710 96.625 -75.601 50.736
224711 96.641 -75.612 50.712
224712 96.651 -75.631 50.688
224713 96.661 -75.649 50.664
224714 96.671 -75.669 50.639
224715 96.682 -75.689 50.615
224716 96.693 -75.709 50.592
224717 96.703 -75.728 50.569
224718 96.714 -75.747 50.544
224719 96.725 -75.767 50.518
224720 96.738 -75.786 50.489
224721 96.760 -75.801 50.458
224722 96.783 -75.816 50.424
224723 96.805 -75.831 50.389
224724 96.868 -75.836 50.352
224725 96.933 -75.840 50.316
224726 96.993 -75.845 50.280
224727 97.017 -75.851 50.244
224728 97.041 -75.856 50.208
224729 97.066 -75.862 50.172
224730 97.090 -75.869 50.135
224731 97.114 -75.875 50.099
224732 97.138 -75.881 50.063
224733 97.162 -75.887 50.028
224734 97.185 -75.893 49.994
224735 97.207 -75.900 49.959
224736 97.228 -75.906 49.924
224737 97.249 -75.914 49.890
224738 97.271 -75.921 49.856
224739 97.292 -75.928 49.822
224740 97.314 -75.934 49.787
224741 97.335 -75.941 49.753
224742 97.359 -75.953 49.721
224743 97.383 -75.968 49.689
224744 97.413 -75.978 49.659
224745 97.456 -75.977 49.630
224746 97.499 -75.975 49.603
224747 97.541 -75.974 49.576
224748 97.584 -75.972 49.550
224749 97.594 -75.973 49.525
224750 97.602 -75.974 49.500
224751 97.628 -75.979 49.476
224752 97.638 -75.990 49.452
224753 97.649 -76.003 49.427
224754 97.660 -76.018 49.401
224755 97.672 -76.032 49.373
224756 97.649 -76.029 49.341
224757 97.643 -76.038 49.304
224758 97.660 -76.063 49.267
224759 97.678 -76.088 49.230
224760 97.706 -76.100 49.191
224761 97.734 -76.113 49.152
224762 97.761 -76.126 49.111
224763 97.789 -76.138 49.070
224764 97.817 -76.150 49.029
224765 97.845 -76.162 48.988
224766 97.887 -76.175 48.949
224767 97.933 -76.188 48.909
224768 97.979 -76.201 48.871
224769 98.024 -76.213 48.830
224770 98.070 -76.226 48.789
224771 98.119 -76.248 48.748
224772 98.174 -76.281 48.708
224773 98.229 -76.315 48.666
224774 98.215 -76.370 48.625
224775 98.255 -76.385 48.584
224776 98.281 -76.393 48.543
224777 98.304 -76.399 48.504
224778 98.326 -76.401 48.466
224779 98.348 -76.404 48.430
224780 98.370 -76.406 48.395
224781 98.392 -76.408 48.362
224782 98.415 -76.410 48.330
224783 98.442 -76.411 48.297
224784 98.469 -76.411 48.267
224785 98.499 -76.414 48.236
224786 98.531 -76.422 48.205
224787 98.564 -76.430 48.173
224788 98.596 -76.436 48.143
224789 98.629 -76.444 48.113
224790 98.661 -76.451 48.084
224791 98.694 -76.458 48.056
224792 98.726 -76.466 48.031
224793 98.758 -76.473 48.006
224794 98.781 -76.477 47.982
224795 98.798 -76.477 47.959
224796 98.816 -76.479 47.936
224797 98.833 -76.480 47.914
224798 98.851 -76.480 47.893
224799 98.868 -76.481 47.871
224800 98.885 -76.482 47.848
224801 98.903 -76.483 47.825
224802 98.920 -76.484 47.801
224803 98.937 -76.485 47.778
224804 98.955 -76.486 47.756
224805 98.972 -76.486 47.732
224806 98.996 -76.489 47.709
224807 99.013 -76.489 47.683
224808 99.033 -76.489 47.655
224809 99.053 -76.489 47.627
224810 99.072 -76.489 47.597
224811 99.092 -76.490 47.567
224812 99.112 -76.491 47.536
224813 99.132 -76.491 47.504
224814 99.151 -76.492 47.471
224815 99.171 -76.492 47.437
224816 99.190 -76.499 47.403
224817 99.218 -76.503 47.369
224818 99.249 -76.514 47.333
224819 99.278 -76.525 47.297
224820 99.308 -76.536 47.260
224821 99.339 -76.547 47.222
224822 99.368 -76.558 47.183
224823 99.398 -76.571 47.147
224824 99.427 -76.583 47.111
224825 99.456 -76.595 47.073
224826 99.486 -76.607 47.035
224827 99.515 -76.620 46.998
224828 99.546 -76.633 46.961
224829 99.579 -76.646 46.924
224830 99.614 -76.658 46.888
224831 99.649 -76.671 46.852
224832 99.678 -76.680 46.817
224833 99.690 -76.682 46.782
224834 99.701 -76.683 46.748
224835 99.713 -76.686 46.715
224836 99.724 -76.687 46.683
224837 99.736 -76.689 46.653
224838 99.751 -76.690 46.624
224839 99.767 -76.690 46.595
224840 99.783 -76.690 46.567
224841 99.799 -76.691 46.538
224842 99.816 -76.692 46.510
224843 99.832 -76.692 46.482
224844 99.848 -76.693 46.454
224845 99.865 -76.693 46.426
224846 99.887 -76.696 46.397
224847 99.908 -76.699 46.368
224848 99.930 -76.702 46.338
224849 99.951 -76.705 46.308
224850 99.973 -76.707 46.278
224851 99.994 -76.710 46.248
224852 100.016 -76.712 46.217
224853 100.037 -76.715 46.186
224854 100.062 -76.719 46.155
224855 100.093 -76.727 46.124
224856 100.125 -76.734 46.093
224857 100.156 -76.740 46.064
224858 100.188 -76.747 46.034
224859 100.217 -76.752 46.006
224860 100.231 -76.736 45.978
224861 100.246 -76.720 45.951
224862 100.260 -76.704 45.925
224863 100.274 -76.688 45.901
224864 100.288 -76.672 45.877
224865 100.303 -76.656 45.855
224866 100.317 -76.640 45.833
224867 100.338 -76.638 45.813
224868 100.372 -76.668 45.793
224869 100.401 -76.702 45.775
224870 100.428 -76.737 45.758
224871 100.455 -76.772 45.741
224872 100.482 -76.807 45.724
224873 100.508 -76.842 45.708
224874 100.535 -76.877 45.691
224875 100.558 -76.897 45.674
224876 100.580 -76.915 45.657
224877 100.603 -76.933 45.641
224878 100.625 -76.951 45.624
224879 100.647 -76.969 45.606
224880 100.670 -76.987 45.589
224881 100.692 -77.006 45.573
224882 100.713 -77.025 45.557
224883 100.731 -77.044 45.542
224884 100.758 -77.079 45.527
224885 100.789 -77.039 45.513
224886 100.816 -77.053 45.499
224887 100.843 -77.066 45.486
224888 100.870 -77.080 45.474
224889 100.898 -77.094 45.462
224890 100.925 -77.107 45.450
224891 100.953 -77.121 45.438
224892 100.980 -77.135 45.426
224893 101.007 -77.148 45.414
224894 101.034 -77.162 45.402
224895 101.061 -77.175 45.392
224896 101.089 -77.189 45.381
224897 101.116 -77.204 45.370
224898 101.143 -77.217 45.359
224899 101.171 -77.229 45.347
224900 101.198 -77.239 45.333
224901 101.217 -77.254 45.319
224902 101.229 -77.279 45.304
224903 101.239 -77.304 45.288
224904 101.250 -77.329 45.273
224905 101.261 -77.354 45.257
224906 101.271 -77.379 45.241
224907 101.282 -77.404 45.223
224908 101.293 -77.429 45.205
224909 101.296 -77.444 45.187
224910 101.297 -77.457 45.167
224911 101.311 -77.461 45.149
224912 101.329 -77.464 45.130
224913 101.346 -77.466 45.111
224914 101.364 -77.468 45.092
224915 101.382 -77.470 45.073
224916 101.399 -77.473 45.054
224917 101.417 -77.475 45.036
224918 101.434 -77.477 45.017
224919 101.450 -77.479 45.000
224920 101.466 -77.480 44.983
224921 101.481 -77.482 44.965
224922 101.496 -77.483 44.947
224923 101.512 -77.485 44.929
224924 101.527 -77.486 44.913
224925 101.505 -77.539 44.898
224926 101.530 -77.556 44.881
224927 101.566 -77.568 44.865
224928 101.603 -77.580 44.848
224929 101.639 -77.592 44.832
224930 101.675 -77.603 44.814
224931 101.711 -77.615 44.798
224932 101.748 -77.627 44.782
224933 101.784 -77.639 44.767
224934 101.820 -77.651 44.751
224935 101.856 -77.662 44.735
224936 101.893 -77.674 44.719
224937 101.923 -77.688 44.703
224938 101.937 -77.710 44.688
224939 101.958 -77.727 44.672
224940 102.001 -77.733 44.657
224941 102.034 -77.745 44.641
224942 102.066 -77.757 44.624
224943 102.097 -77.768 44.608
224944 102.125 -77.779 44.592
224945 102.154 -77.790 44.577
224946 102.183 -77.794 44.562
224947 102.211 -77.796 44.547
224948 102.239 -77.802 44.531
224949 102.262 -77.818 44.516
224950 102.286 -77.832 44.500
224951 102.309 -77.847 44.486
224952 102.332 -77.862 44.473
224953 102.356 -77.877 44.458
224954 102.379 -77.891 44.443
224955 102.411 -77.911 44.428
224956 102.439 -77.921 44.411
224957 102.461 -77.924 44.397
224958 102.482 -77.928 44.382
224959 102.504 -77.932 44.367
224960 102.526 -77.936 44.353
224961 102.548 -77.940 44.337
224962 102.571 -77.943 44.321
224963 102.595 -77.944 44.305
224964 102.619 -77.946 44.290
224965 102.641 -77.950 44.275
224966 102.660 -77.956 44.261
224967 102.689 -77.963 44.246
224968 102.718 -77.970 44.230
224969 102.747 -77.977 44.215
224970 102.779 -77.983 44.198
224971 102.816 -77.987 44.181
224972 102.854 -77.996 44.164
224973 102.894 -78.007 44.147
224974 102.934 -78.018 44.131
224975 102.974 -78.029 44.114
224976 103.014 -78.040 44.098
224977 103.053 -78.051 44.082
224978 103.093 -78.062 44.066
224979 103.134 -78.074 44.049
224980 103.173 -78.085 44.033
224981 103.213 -78.096 44.016
224982 103.253 -78.107 43.999
225223 92.153 -75.176 54.237
225224 92.176 -75.187 54.244
225225 92.199 -75.198 54.251
225226 92.222 -75.209 54.259
225227 92.245 -75.220 54.268
225228 92.269 -75.232 54.279
225229 92.292 -75.242 54.290
225230 92.314 -75.254 54.299
225231 92.338 -75.265 54.308
225232 92.361 -75.276 54.315
225233 92.384 -75.287 54.319
225234 92.407 -75.298 54.326
225235 92.422 -75.299 54.331
225236 92.442 -75.292 54.333
225237 92.461 -75.289 54.333
225238 92.489 -75.285 54.329
225239 92.519 -75.281 54.322
225240 92.550 -75.276 54.315
225241 92.581 -75.273 54.310
225242 92.612 -75.268 54.302
225243 92.643 -75.264 54.293
225244 92.673 -75.260 54.283
225245 92.704 -75.256 54.270
225246 92.735 -75.251 54.258
225247 92.766 -75.248 54.245
225248 92.797 -75.243 54.233
225249 92.828 -75.239 54.222
225250 92.859 -75.235 54.209
225251 92.889 -75.231 54.196
225252 92.908 -75.233 54.184
225253 92.927 -75.236 54.171
225254 92.945 -75.239 54.158
225255 92.964 -75.243 54.145
225256 92.985 -75.243 54.132
225257 93.008 -75.242 54.119
225258 93.032 -75.240 54.105
225259 93.055 -75.239 54.090
225260 93.078 -75.238 54.077
225261 93.102 -75.236 54.062
225262 93.125 -75.236 54.047
225269 93.288 -75.226 53.975
225270 93.311 -75.226 53.972
225271 93.335 -75.224 53.972
225272 93.358 -75.223 53.975
225273 93.381 -75.222 53.983
225274 93.405 -75.220 53.992
225275 93.428 -75.219 54.003
225276 93.451 -75.218 54.018
225277 93.475 -75.217 54.038
225278 93.498 -75.216 54.062
225279 93.521 -75.214 54.091
225280 93.558 -75.213 54.128
225281 93.597 -75.210 54.168
225282 93.636 -75.209 54.212
225283 93.669 -75.207 54.260
225284 93.700 -75.206 54.299
225285 93.731 -75.204 54.333
225286 93.763 -75.203 54.360
225287 93.789 -75.200 54.379
225288 93.816 -75.196 54.390
225289 93.842 -75.193 54.398
225290 93.868 -75.189 54.406
225291 93.895 -75.186 54.414
225292 93.921 -75.183 54.420
225293 93.947 -75.180 54.424
225294 93.971 -75.174 54.425
225295 93.994 -75.165 54.424
225296 94.016 -75.157 54.426
225297 94.039 -75.148 54.428
225298 94.065 -75.151 54.430
225299 94.091 -75.153 54.434
225300 94.117 -75.156 54.441
225301 94.142 -75.156 54.446
225302 94.166 -75.154 54.448
225303 94.189 -75.151 54.454
225304 94.212 -75.149 54.459
225305 94.236 -75.146 54.461
225306 94.260 -75.144 54.460
225307 94.284 -75.145 54.457
225308 94.312 -75.155 54.446
225309 94.336 -75.154 54.430
225310 94.360 -75.154 54.409
225311 94.385 -75.154 54.387
225312 94.409 -75.153 54.363
225313 94.435 -75.153 54.340
225314 94.461 -75.152 54.323
225315 94.486 -75.152 54.301
225316 94.509 -75.152 54.274
225317 94.524 -75.154 54.243
225318 94.538 -75.155 54.209
225319 94.552 -75.157 54.174
225320 94.565 -75.159 54.139
225321 94.579 -75.160 54.103
225322 94.594 -75.162 54.063
225323 94.608 -75.163 54.025
225324 94.622 -75.165 53.988
225325 94.636 -75.167 53.951
225326 94.650 -75.167 53.914
225327 94.664 -75.167 53.877
225328 94.686 -75.169 53.838
225329 94.710 -75.172 53.796
225330 94.735 -75.175 53.751
225331 94.760 -75.178 53.706
225332 94.761 -75.183 53.660
225333 94.755 -75.189 53.611
225334 94.748 -75.194 53.566
225335 94.741 -75.199 53.520
225336 94.735 -75.204 53.473
225337 94.728 -75.210 53.428
225338 94.726 -75.215 53.381
225339 94.745 -75.217 53.332
225340 94.762 -75.220 53.281
225341 94.780 -75.223 53.231
225342 94.798 -75.226 53.182
225343 94.816 -75.229 53.133
225344 94.834 -75.232 53.085
225345 94.851 -75.235 53.035
225346 94.870 -75.237 52.986
225347 94.887 -75.240 52.936
225348 94.913 -75.245 52.886
225349 94.941 -75.250 52.834
225350 94.967 -75.253 52.780
225351 94.992 -75.256 52.726
225352 95.018 -75.260 52.673
225353 95.044 -75.263 52.620
225354 95.069 -75.267 52.567
225355 95.095 -75.270 52.515
225356 95.121 -75.273 52.465
225357 95.146 -75.276 52.414
225358 95.172 -75.280 52.365
225359 95.204 -75.289 52.318
225360 95.238 -75.300 52.272
225361 95.272 -75.311 52.227
225362 95.294 -75.324 52.184
225363 95.315 -75.338 52.141
225364 95.336 -75.352 52.100
225365 95.359 -75.362 52.058
225366 95.384 -75.373 52.019
225367 95.408 -75.383 51.981
225368 95.432 -75.394 51.944
225369 95.456 -75.405 51.910
225370 95.480 -75.415 51.877
225371 95.505 -75.426 51.844
225372 95.529 -75.436 51.812
225373 95.553 -75.446 51.779
225374 95.582 -75.439 51.747
225375 95.610 -75.433 51.715
225376 95.639 -75.426 51.686
225377 95.667 -75.419 51.658
225378 95.696 -75.412 51.629
225379 95.725 -75.405 51.600
225380 95.753 -75.398 51.569
225381 95.781 -75.392 51.538
225382 95.815 -75.386 51.508
225383 95.850 -75.383 51.478
225384 95.886 -75.379 51.447
225385 95.923 -75.380 51.414
225386 95.961 -75.380 51.383
225387 95.998 -75.380 51.352
225388 96.039 -75.386 51.320
225389 96.081 -75.394 51.291
225390 96.122 -75.402 51.263
225391 96.165 -75.409 51.235
225392 96.203 -75.414 51.207
225393 96.240 -75.417 51.181
225394 96.277 -75.428 51.155
225395 96.313 -75.441 51.127
225396 96.343 -75.452 51.102
225397 96.364 -75.461 51.076
225398 96.384 -75.469 51.050
225399 96.404 -75.482 51.022
225400 96.424 -75.494 50.997
225401 96.444 -75.507 50.973
225402 96.464 -75.520 50.949
225403 96.483 -75.528 50.926
225404 96.503 -75.531 50.903
225405 96.522 -75.533 50.880
225406 96.541 -75.538 50.857
225407 96.559 -75.546 50.833
225408 96.578 -75.554 50.809
225409 96.596 -75.562 50.786
225410 96.615 -75.570 50.764
225411 96.635 -75.576 50.742
225412 96.651 -75.587 50.718
225413 96.661 -75.605 50.694
225414 96.671 -75.624 50.669
225415 96.681 -75.643 50.644
225416 96.691 -75.662 50.620
225417 96.701 -75.680 50.597
225418 96.711 -75.699 50.572
225419 96.722 -75.718 50.546
225420 96.733 -75.738 50.519
225421 96.743 -75.758 50.490
225422 96.758 -75.776 50.458
225423 96.781 -75.791 50.425
225424 96.817 -75.803 50.391
225425 96.881 -75.807 50.356
225426 96.946 -75.811 50.322
225427 97.010 -75.815 50.287
225428 97.040 -75.822 50.252
225429 97.064 -75.828 50.218
225430 97.088 -75.834 50.182
225431 97.112 -75.839 50.146
225432 97.137 -75.845 50.110
225433 97.158 -75.852 50.075
225434 97.180 -75.859 50.042
225435 97.202 -75.866 50.008
225436 97.223 -75.873 49.973
225437 97.244 -75.880 49.938
225438 97.266 -75.887 49.904
225439 97.287 -75.894 49.870
225440 97.309 -75.900 49.836
225441 97.330 -75.908 49.801
225442 97.352 -75.915 49.766
225443 97.376 -75.929 49.733
225444 97.401 -75.944 49.701
225445 97.428 -75.957 49.670
225446 97.470 -75.956 49.641
225447 97.513 -75.955 49.613
225448 97.556 -75.953 49.587
225449 97.598 -75.952 49.562
225450 97.641 -75.950 49.537
225451 97.661 -75.950 49.512
225452 97.677 -75.959 49.487
225453 97.686 -75.969 49.462
225454 97.696 -75.980 49.435
225455 97.706 -75.992 49.409
225456 97.710 -76.003 49.380
225457 97.682 -75.998 49.346
225458 97.669 -76.003 49.308
225459 97.687 -76.027 49.270
225460 97.705 -76.052 49.231
225461 97.723 -76.077 49.192
225462 97.750 -76.089 49.152
225463 97.778 -76.101 49.110
225464 97.806 -76.113 49.068
225465 97.834 -76.126 49.026
225466 97.862 -76.138 48.984
225467 97.903 -76.151 48.944
225468 97.949 -76.163 48.904
225469 97.995 -76.176 48.864
225470 98.040 -76.189 48.823
225471 98.089 -76.207 48.782
225472 98.144 -76.241 48.741
225473 98.199 -76.275 48.700
225474 98.253 -76.308 48.659
225475 98.308 -76.342 48.618
225476 98.286 -76.376 48.576
225477 98.292 -76.386 48.536
225478 98.314 -76.389 48.496
225479 98.336 -76.391 48.459
225480 98.358 -76.393 48.422
225481 98.380 -76.395 48.387
225482 98.402 -76.398 48.354
225483 98.424 -76.400 48.322
225484 98.450 -76.401 48.290
225485 98.478 -76.401 48.260
225486 98.505 -76.401 48.229
225487 98.532 -76.401 48.198
225488 98.560 -76.403 48.166
225489 98.593 -76.410 48.136
225490 98.625 -76.417 48.107
225491 98.657 -76.425 48.078
225492 98.690 -76.433 48.051
225493 98.722 -76.439 48.026
225494 98.755 -76.447 48.001
225495 98.778 -76.451 47.976
225496 98.795 -76.451 47.954
225497 98.812 -76.452 47.931
225498 98.830 -76.453 47.909
225499 98.847 -76.454 47.888
225500 98.865 -76.455 47.867
225501 98.882 -76.456 47.845
225502 98.900 -76.457 47.822
225503 98.917 -76.458 47.799
225504 98.941 -76.460 47.776
225505 98.967 -76.462 47.753
225506 98.993 -76.464 47.729
225507 99.017 -76.472 47.708
225508 99.031 -76.474 47.680
225509 99.044 -76.472 47.653
225510 99.064 -76.470 47.624
225511 99.084 -76.468 47.594
225512 99.105 -76.467 47.563
225513 99.125 -76.465 47.531
225514 99.145 -76.466 47.499
225515 99.164 -76.466 47.467
225516 99.184 -76.467 47.433
225517 99.204 -76.470 47.400
225518 99.232 -76.477 47.366
225519 99.262 -76.488 47.332
225520 99.292 -76.500 47.296
225521 99.322 -76.512 47.260
225522 99.351 -76.524 47.222
225523 99.380 -76.536 47.184
225524 99.409 -76.549 47.147
225525 99.438 -76.561 47.111
225526 99.468 -76.573 47.073
225527 99.497 -76.585 47.036
225528 99.528 -76.598 47.000
225529 99.559 -76.611 46.964
225530 99.592 -76.624 46.927
225531 99.628 -76.637 46.891
225532 99.663 -76.649 46.856
225533 99.698 -76.662 46.821
225534 99.726 -76.671 46.787
225535 99.737 -76.673 46.753
225536 99.749 -76.674 46.720
225537 99.760 -76.676 46.689
225538 99.772 -76.677 46.659
225539 99.787 -76.678 46.630
225540 99.803 -76.679 46.602
225541 99.819 -76.680 46.573
225542 99.835 -76.680 46.545
225543 99.852 -76.680 46.516
225544 99.868 -76.681 46.488
225545 99.884 -76.682 46.460
225546 99.900 -76.682 46.431
225547 99.917 -76.683 46.401
225548 99.933 -76.683 46.372
225549 99.955 -76.687 46.341
225550 99.977 -76.689 46.310
225551 99.998 -76.692 46.279
225552 100.019 -76.694 46.249
225553 100.041 -76.697 46.217
225554 100.062 -76.699 46.186
225555 100.088 -76.704 46.154
225556 100.120 -76.712 46.123
225557 100.151 -76.718 46.093
225558 100.182 -76.725 46.062
225559 100.205 -76.721 46.032
225560 100.220 -76.705 46.003
225561 100.234 -76.690 45.975
225562 100.248 -76.673 45.949
225563 100.263 -76.657 45.923
225564 100.277 -76.641 45.898
225565 100.291 -76.625 45.875
225566 100.305 -76.609 45.852
225567 100.320 -76.593 45.830
225568 100.344 -76.601 45.809
225569 100.374 -76.633 45.789
225570 100.401 -76.668 45.772
225571 100.428 -76.704 45.754
225572 100.455 -76.739 45.737
225573 100.481 -76.774 45.719
225574 100.508 -76.809 45.702
225575 100.535 -76.843 45.686
225576 100.561 -76.878 45.669
225577 100.583 -76.897 45.652
225578 100.606 -76.915 45.635
225579 100.628 -76.933 45.618
225580 100.651 -76.951 45.601
225581 100.673 -76.969 45.584
225582 100.696 -76.987 45.567
225583 100.718 -77.006 45.551
225584 100.749 -77.012 45.536
225585 100.784 -77.063 45.521
225586 100.818 -76.992 45.507
225587 100.845 -77.006 45.494
225588 100.872 -77.020 45.481
225589 100.899 -77.034 45.468
225590 100.926 -77.047 45.455
225591 100.954 -77.061 45.443
225592 100.981 -77.075 45.431
225593 101.008 -77.088 45.419
225594 101.036 -77.102 45.408
225595 101.063 -77.116 45.396
225596 101.090 -77.129 45.386
225597 101.118 -77.143 45.375
225598 101.132 -77.168 45.364
225599 101.153 -77.191 45.352
225600 101.182 -77.212 45.339
225601 101.211 -77.232 45.326
225602 101.229 -77.255 45.312
225603 101.241 -77.281 45.297
225604 101.253 -77.307 45.281
225605 101.265 -77.332 45.265
225606 101.277 -77.358 45.249
225607 101.287 -77.383 45.232
225608 101.298 -77.408 45.215
225609 101.309 -77.433 45.197
225610 101.314 -77.445 45.178
225611 101.332 -77.448 45.159
225612 101.350 -77.450 45.140
225613 101.366 -77.451 45.122
225614 101.382 -77.453 45.103
225615 101.397 -77.454 45.084
225616 101.412 -77.456 45.064
225617 101.428 -77.458 45.045
225618 101.443 -77.460 45.026
225619 101.459 -77.461 45.008
225620 101.474 -77.463 44.990
225621 101.489 -77.464 44.973
225622 101.505 -77.466 44.955
225623 101.520 -77.467 44.938
225624 101.536 -77.469 44.921
225625 101.551 -77.471 44.905
225626 101.533 -77.515 44.889
225627 101.553 -77.535 44.873
225628 101.572 -77.555 44.856
225629 101.606 -77.567 44.839
225630 101.642 -77.580 44.822
225631 101.679 -77.592 44.805
225632 101.715 -77.603 44.789
225633 101.751 -77.615 44.773
225634 101.787 -77.627 44.757
225635 101.823 -77.639 44.741
225636 101.859 -77.650 44.725
225637 101.896 -77.662 44.709
225638 101.925 -77.677 44.693
225639 101.939 -77.699 44.677
225640 101.971 -77.711 44.662
225641 102.014 -77.716 44.646
225642 102.048 -77.727 44.631
225643 102.081 -77.739 44.614
225644 102.113 -77.751 44.598
225645 102.145 -77.762 44.582
225646 102.173 -77.773 44.567
225647 102.201 -77.783 44.553
225648 102.231 -77.785 44.538
225649 102.257 -77.792 44.523
225650 102.280 -77.806 44.508
225651 102.304 -77.821 44.493
225652 102.327 -77.836 44.478
225653 102.350 -77.851 44.464
225654 102.374 -77.865 44.449
225655 102.397 -77.881 44.434
225656 102.424 -77.898 44.419
225657 102.450 -77.906 44.403
225658 102.472 -77.909 44.387
225659 102.493 -77.913 44.373
225660 102.517 -77.915 44.358
225661 102.541 -77.918 44.343
225662 102.566 -77.919 44.327
225663 102.590 -77.921 44.311
225664 102.615 -77.923 44.295
225665 102.635 -77.928 44.279
225666 102.653 -77.934 44.264
225667 102.675 -77.941 44.249
225668 102.704 -77.947 44.234
225669 102.733 -77.954 44.219
225670 102.763 -77.960 44.203
225671 102.801 -77.965 44.186
225672 102.837 -77.971 44.169
225673 102.874 -77.976 44.152
225674 102.911 -77.981 44.135
225675 102.948 -77.987 44.119
225676 102.986 -77.993 44.103
225677 103.026 -78.003 44.087
225678 103.066 -78.015 44.070
225679 103.105 -78.026 44.054
225680 103.145 -78.037 44.038
225681 103.185 -78.048 44.021
225682 103.225 -78.059 44.004
225683 103.265 -78.071 43.988
225924 92.141 -75.147 54.244
225925 92.164 -75.158 54.250
225926 92.187 -75.170 54.257
225927 92.210 -75.180 54.264
225928 92.233 -75.192 54.274
225929 92.256 -75.202 54.285
225930 92.279 -75.214 54.295
225931 92.302 -75.225 54.306
225932 92.325 -75.236 54.315
225933 92.348 -75.247 54.322
225934 92.371 -75.258 54.329
225935 92.394 -75.269 54.336
225936 92.411 -75.276 54.342
225937 92.431 -75.268 54.345
225938 92.451 -75.265 54.345
225939 92.482 -75.261 54.342
225940 92.512 -75.257 54.337
225941 92.543 -75.252 54.329
225942 92.574 -75.248 54.324
225943 92.605 -75.244 54.317
225944 92.636 -75.240 54.308
225945 92.666 -75.236 54.298
225946 92.697 -75.232 54.285
225947 92.728 -75.227 54.272
225948 92.759 -75.223 54.259
225949 92.790 -75.219 54.248
225950 92.820 -75.215 54.236
225951 92.851 -75.210 54.223
225952 92.882 -75.207 54.211
225953 92.900 -75.210 54.199
225954 92.919 -75.213 54.186
225955 92.937 -75.216 54.172
225956 92.960 -75.214 54.159
225957 92.984 -75.214 54.147
225958 93.007 -75.212 54.134
225959 93.030 -75.210 54.120
225960 93.054 -75.210 54.106
225961 93.077 -75.208 54.092
225962 93.100 -75.207 54.076
225963 93.124 -75.206 54.061
225964 93.147 -75.204 54.047
225970 93.287 -75.197 53.986
225971 93.310 -75.196 53.983
225972 93.334 -75.195 53.982
225973 93.357 -75.193 53.985
225974 93.380 -75.192 53.991
225975 93.404 -75.191 54.000
225976 93.427 -75.190 54.011
225977 93.450 -75.189 54.026
225978 93.474 -75.187 54.045
225979 93.497 -75.186 54.070
225980 93.520 -75.185 54.100
225981 93.557 -75.183 54.138
225982 93.595 -75.181 54.179
225983 93.630 -75.179 54.224
225984 93.661 -75.178 54.271
225985 93.693 -75.177 54.312
225986 93.724 -75.176 54.349
225987 93.756 -75.174 54.379
225988 93.786 -75.173 54.403
225989 93.813 -75.169 54.414
225990 93.839 -75.166 54.422
225991 93.865 -75.162 54.428
225992 93.892 -75.159 54.436
225993 93.917 -75.157 54.442
225994 93.943 -75.154 54.445
225995 93.969 -75.151 54.443
225996 93.995 -75.148 54.437
225997 94.017 -75.138 54.437
225998 94.040 -75.130 54.438
225999 94.064 -75.124 54.441
226000 94.090 -75.126 54.444
226001 94.116 -75.129 54.449
226002 94.140 -75.129 54.452
226003 94.164 -75.127 54.451
226004 94.188 -75.125 54.454
226005 94.211 -75.122 54.460
226006 94.235 -75.120 54.465
226007 94.259 -75.117 54.469
226008 94.282 -75.115 54.468
226009 94.307 -75.116 54.461
226010 94.332 -75.116 54.448
226011 94.357 -75.116 54.427
226012 94.383 -75.116 54.405
226013 94.408 -75.116 54.382
226014 94.434 -75.116 54.359
226015 94.460 -75.116 54.339
226016 94.486 -75.116 54.315
226017 94.508 -75.116 54.287
226018 94.523 -75.117 54.256
226019 94.537 -75.119 54.221
226020 94.551 -75.120 54.187
226021 94.565 -75.122 54.153
226022 94.578 -75.123 54.119
226023 94.593 -75.125 54.082
226024 94.607 -75.126 54.045
226025 94.621 -75.128 54.008
226026 94.635 -75.129 53.971
226027 94.649 -75.131 53.933
226028 94.660 -75.140 53.894
226029 94.681 -75.145 53.854
226030 94.704 -75.147 53.810
226031 94.729 -75.150 53.765
226032 94.754 -75.153 53.720
226033 94.771 -75.157 53.673
226034 94.765 -75.162 53.626
226035 94.758 -75.167 53.582
226036 94.752 -75.173 53.536
226037 94.745 -75.178 53.491
226038 94.738 -75.183 53.447
226039 94.745 -75.188 53.401
226040 94.763 -75.190 53.352
226041 94.781 -75.193 53.301
226042 94.799 -75.196 53.252
226043 94.816 -75.199 53.201
226044 94.834 -75.201 53.151
226045 94.852 -75.204 53.102
226046 94.870 -75.207 53.053
226047 94.888 -75.210 53.002
226048 94.906 -75.213 52.951
226049 94.923 -75.216 52.900
226050 94.948 -75.219 52.847
226051 94.973 -75.222 52.792
226052 94.999 -75.226 52.737
226053 95.025 -75.229 52.681
226054 95.051 -75.232 52.626
226055 95.076 -75.236 52.572
226056 95.102 -75.239 52.519
226057 95.128 -75.242 52.467
226058 95.154 -75.246 52.417
226059 95.179 -75.249 52.367
226060 95.209 -75.257 52.318
226061 95.244 -75.267 52.271
226062 95.268 -75.280 52.226
226063 95.288 -75.294 52.181
226064 95.312 -75.305 52.138
226065 95.336 -75.316 52.096
226066 95.360 -75.326 52.055
226067 95.385 -75.337 52.016
226068 95.409 -75.348 51.979
226069 95.433 -75.358 51.941
226070 95.457 -75.369 51.907
226071 95.481 -75.380 51.873
226072 95.506 -75.390 51.840
226073 95.530 -75.401 51.808
226074 95.554 -75.410 51.775
226075 95.583 -75.403 51.743
226076 95.611 -75.395 51.712
226077 95.640 -75.389 51.683
226078 95.669 -75.382 51.654
226079 95.697 -75.375 51.626
226080 95.726 -75.368 51.598
226081 95.754 -75.361 51.569
226082 95.783 -75.355 51.538
226083 95.812 -75.348 51.508
226084 95.841 -75.341 51.479
226085 95.877 -75.339 51.447
226086 95.917 -75.343 51.416
226087 95.959 -75.352 51.384
226088 96.001 -75.359 51.353
226089 96.043 -75.367 51.324
226090 96.085 -75.374 51.295
226091 96.126 -75.383 51.266
226092 96.168 -75.390 51.239
226093 96.207 -75.395 51.212
226094 96.245 -75.398 51.186
226095 96.282 -75.402 51.160
226096 96.319 -75.407 51.132
226097 96.352 -75.418 51.107
226098 96.372 -75.428 51.082
226099 96.392 -75.441 51.055
226100 96.412 -75.453 51.028
226101 96.432 -75.466 51.003
226102 96.452 -75.479 50.979
226103 96.471 -75.491 50.955
226104 96.491 -75.504 50.933
226105 96.511 -75.508 50.911
226106 96.530 -75.511 50.888
226107 96.549 -75.515 50.865
226108 96.568 -75.523 50.841
226109 96.586 -75.531 50.817
226110 96.605 -75.539 50.793
226111 96.623 -75.546 50.770
226112 96.645 -75.551 50.747
226113 96.661 -75.561 50.724
226114 96.671 -75.580 50.699
226115 96.681 -75.599 50.674
226116 96.691 -75.617 50.649
226117 96.701 -75.636 50.625
226118 96.711 -75.654 50.601
226119 96.721 -75.673 50.576
226120 96.731 -75.692 50.549
226121 96.741 -75.710 50.521
226122 96.751 -75.729 50.492
226123 96.761 -75.749 50.460
226124 96.778 -75.766 50.428
226125 96.830 -75.774 50.395
226126 96.894 -75.778 50.361
226127 96.959 -75.782 50.329
226128 97.023 -75.787 50.295
226129 97.063 -75.792 50.262
226130 97.087 -75.798 50.228
226131 97.110 -75.804 50.193
226132 97.132 -75.811 50.158
226133 97.154 -75.818 50.123
226134 97.175 -75.825 50.088
226135 97.196 -75.832 50.056
226136 97.218 -75.839 50.021
226137 97.239 -75.846 49.987
226138 97.261 -75.853 49.953
226139 97.282 -75.860 49.919
226140 97.304 -75.867 49.885
226141 97.325 -75.874 49.850
226142 97.347 -75.881 49.815
226143 97.369 -75.891 49.780
226144 97.394 -75.906 49.746
226145 97.419 -75.920 49.714
226146 97.444 -75.935 49.682
226147 97.485 -75.935 49.652
226148 97.527 -75.934 49.624
226149 97.570 -75.932 49.599
226150 97.613 -75.931 49.575
226151 97.656 -75.930 49.549
226152 97.696 -75.928 49.524
226153 97.712 -75.935 49.498
226154 97.723 -75.947 49.470
226155 97.734 -75.957 49.443
226156 97.745 -75.969 49.415
226157 97.743 -75.971 49.384
226158 97.716 -75.966 49.348
226159 97.696 -75.966 49.311
226160 97.713 -75.991 49.271
226161 97.731 -76.016 49.232
226162 97.749 -76.041 49.192
226163 97.767 -76.065 49.152
226164 97.795 -76.078 49.110
226165 97.823 -76.090 49.067
226166 97.851 -76.102 49.024
226167 97.878 -76.114 48.982
226168 97.920 -76.127 48.941
226169 97.966 -76.140 48.900
226170 98.012 -76.153 48.859
226171 98.058 -76.166 48.818
226172 98.113 -76.201 48.777
226173 98.168 -76.235 48.736
226174 98.223 -76.268 48.694
226175 98.273 -76.294 48.653
226176 98.318 -76.312 48.612
226177 98.344 -76.329 48.570
226178 98.355 -76.347 48.530
226179 98.323 -76.378 48.491
226180 98.346 -76.380 48.453
226181 98.367 -76.383 48.415
226182 98.389 -76.385 48.381
226183 98.412 -76.387 48.347
226184 98.434 -76.389 48.315
226185 98.458 -76.391 48.283
226186 98.486 -76.391 48.251
226187 98.513 -76.391 48.221
226188 98.540 -76.391 48.191
226189 98.567 -76.391 48.161
226190 98.594 -76.392 48.131
226191 98.622 -76.392 48.102
226192 98.654 -76.398 48.073
226193 98.686 -76.406 48.046
226194 98.719 -76.414 48.021
226195 98.751 -76.420 47.996
226196 98.774 -76.424 47.972
226197 98.792 -76.425 47.950
226198 98.809 -76.426 47.927
226199 98.827 -76.426 47.906
226200 98.844 -76.428 47.885
226201 98.861 -76.429 47.864
226202 98.885 -76.431 47.842
226203 98.912 -76.433 47.819
226204 98.937 -76.436 47.796
226205 98.963 -76.438 47.774
226206 98.990 -76.440 47.751
226207 99.015 -76.443 47.727
226208 99.040 -76.455 47.703
226209 99.059 -76.461 47.676
226210 99.067 -76.458 47.649
226211 99.075 -76.455 47.620
226212 99.095 -76.454 47.590
226213 99.116 -76.451 47.559
226214 99.136 -76.450 47.527
226215 99.156 -76.448 47.495
226216 99.177 -76.447 47.463
226217 99.197 -76.445 47.429
226218 99.217 -76.444 47.396
226219 99.245 -76.453 47.364
226220 99.274 -76.465 47.330
226221 99.304 -76.477 47.295
226222 99.333 -76.489 47.258
226223 99.362 -76.502 47.222
226224 99.392 -76.514 47.184
226225 99.421 -76.527 47.148
226226 99.450 -76.539 47.111
226227 99.480 -76.551 47.074
226228 99.510 -76.564 47.038
226229 99.541 -76.577 47.002
226230 99.572 -76.589 46.966
226231 99.607 -76.602 46.930
226232 99.641 -76.615 46.895
226233 99.677 -76.628 46.859
226234 99.712 -76.641 46.825
226235 99.747 -76.653 46.791
226236 99.773 -76.662 46.758
226237 99.785 -76.664 46.726
226238 99.796 -76.665 46.695
226239 99.808 -76.667 46.664
226240 99.823 -76.668 46.637
226241 99.839 -76.668 46.608
226242 99.855 -76.668 46.580
226243 99.871 -76.669 46.552
226244 99.888 -76.670 46.521
226245 99.904 -76.670 46.493
226246 99.920 -76.671 46.465
226247 99.936 -76.671 46.436
226248 99.952 -76.672 46.407
226249 99.969 -76.672 46.376
226250 99.985 -76.673 46.344
226251 100.002 -76.674 46.311
226252 100.023 -76.676 46.280
226253 100.045 -76.679 46.249
226254 100.066 -76.682 46.216
226255 100.087 -76.684 46.184
226256 100.114 -76.690 46.154
226257 100.146 -76.696 46.123
226258 100.177 -76.703 46.091
226259 100.194 -76.690 46.061
226260 100.208 -76.674 46.031
226261 100.222 -76.658 46.002
226262 100.237 -76.643 45.974
226263 100.251 -76.627 45.947
226264 100.265 -76.611 45.920
226265 100.279 -76.595 45.895
226266 100.293 -76.578 45.872
226267 100.308 -76.562 45.849
226268 100.322 -76.546 45.827
226269 100.347 -76.565 45.806
226270 100.373 -76.600 45.787
226271 100.401 -76.635 45.768
226272 100.427 -76.670 45.750
226273 100.454 -76.705 45.732
226274 100.481 -76.740 45.714
226275 100.507 -76.775 45.697
226276 100.534 -76.810 45.680
226277 100.561 -76.845 45.664
226278 100.587 -76.878 45.647
226279 100.609 -76.896 45.629
226280 100.632 -76.914 45.612
226281 100.654 -76.932 45.595
226282 100.677 -76.950 45.578
226283 100.699 -76.969 45.561
226284 100.728 -76.976 45.545
226285 100.770 -76.967 45.530
226286 100.811 -76.957 45.515
226287 100.846 -76.946 45.502
226288 100.874 -76.959 45.488
226289 100.901 -76.973 45.475
226290 100.928 -76.987 45.462
226291 100.956 -77.000 45.448
226292 100.983 -77.014 45.435
226293 101.008 -77.029 45.423
226294 101.019 -77.058 45.412
226295 101.030 -77.086 45.401
226296 101.047 -77.111 45.389
226297 101.076 -77.132 45.379
226298 101.105 -77.151 45.369
226299 101.134 -77.172 45.357
226300 101.164 -77.191 45.345
226301 101.193 -77.211 45.332
226302 101.222 -77.232 45.318
226303 101.238 -77.256 45.304
226304 101.250 -77.282 45.289
226305 101.262 -77.307 45.273
226306 101.274 -77.332 45.257
226307 101.286 -77.358 45.240
226308 101.298 -77.384 45.224
226309 101.310 -77.409 45.206
226310 101.324 -77.425 45.188
226311 101.343 -77.430 45.169
226312 101.359 -77.432 45.150
226313 101.375 -77.434 45.132
226314 101.390 -77.436 45.113
226315 101.406 -77.437 45.094
226316 101.421 -77.439 45.075
226317 101.437 -77.441 45.055
226318 101.452 -77.442 45.035
226319 101.467 -77.444 45.017
226320 101.483 -77.445 44.999
226321 101.498 -77.447 44.981
226322 101.513 -77.448 44.963
226323 101.529 -77.450 44.946
226324 101.544 -77.451 44.929
226325 101.560 -77.454 44.913
226326 101.575 -77.455 44.896
226327 101.560 -77.492 44.880
226328 101.580 -77.511 44.865
226329 101.600 -77.531 44.847
226330 101.620 -77.551 44.830
226331 101.645 -77.567 44.813
226332 101.682 -77.580 44.796
226333 101.718 -77.591 44.780
226334 101.754 -77.603 44.764
226335 101.790 -77.615 44.749
226336 101.827 -77.627 44.733
226337 101.863 -77.639 44.716
226338 101.899 -77.650 44.700
226339 101.927 -77.666 44.685
226340 101.941 -77.687 44.669
226341 101.983 -77.694 44.653
226342 102.027 -77.699 44.636
226343 102.062 -77.709 44.620
226344 102.095 -77.721 44.605
226345 102.127 -77.733 44.589
226346 102.160 -77.745 44.573
226347 102.192 -77.756 44.558
226348 102.221 -77.767 44.544
226349 102.249 -77.773 44.529
226350 102.276 -77.780 44.514
226351 102.299 -77.796 44.500
226352 102.323 -77.810 44.485
226353 102.346 -77.825 44.470
226354 102.369 -77.840 44.456
226355 102.393 -77.855 44.441
226356 102.416 -77.870 44.426
226357 102.439 -77.884 44.411
226358 102.464 -77.889 44.395
226359 102.488 -77.890 44.378
226360 102.512 -77.893 44.363
226361 102.537 -77.894 44.349
226362 102.561 -77.896 44.333
226363 102.586 -77.898 44.317
226364 102.610 -77.900 44.301
226365 102.629 -77.906 44.285
226366 102.647 -77.912 44.269
226367 102.665 -77.918 44.254
226368 102.690 -77.925 44.238
226369 102.719 -77.932 44.223
226370 102.748 -77.938 44.208
226371 102.785 -77.943 44.191
226372 102.822 -77.949 44.174
226373 102.859 -77.954 44.158
226374 102.896 -77.959 44.141
226375 102.933 -77.965 44.124
226376 102.970 -77.969 44.108
226377 103.007 -77.975 44.092
226378 103.044 -77.980 44.075
226379 103.081 -77.985 44.059
226380 103.118 -77.990 44.043
226381 103.157 -78.000 44.025
226382 103.197 -78.012 44.009
226383 103.237 -78.023 43.993
226625 92.128 -75.118 54.251
226626 92.151 -75.129 54.257
226627 92.174 -75.140 54.263
226628 92.197 -75.151 54.269
226629 92.220 -75.162 54.279
226630 92.243 -75.173 54.290
226631 92.267 -75.185 54.300
226632 92.289 -75.195 54.311
226633 92.312 -75.207 54.321
226634 92.336 -75.217 54.329
226635 92.359 -75.229 54.337
226636 92.382 -75.240 54.345
226637 92.402 -75.251 54.352
226638 92.421 -75.245 54.357
226639 92.443 -75.241 54.359
226640 92.475 -75.237 54.356
226641 92.505 -75.232 54.351
226642 92.536 -75.229 54.344
226643 92.567 -75.224 54.339
226644 92.597 -75.220 54.332
226645 92.629 -75.216 54.323
226646 92.659 -75.212 54.313
226647 92.690 -75.207 54.300
226648 92.721 -75.204 54.287
226649 92.752 -75.199 54.274
226650 92.783 -75.195 54.262
226651 92.813 -75.191 54.249
226652 92.844 -75.187 54.237
226653 92.874 -75.183 54.226
226654 92.892 -75.186 54.215
226655 92.913 -75.188 54.201
226656 92.936 -75.186 54.188
226657 92.959 -75.185 54.175
226658 92.983 -75.184 54.162
226659 93.006 -75.182 54.149
226660 93.029 -75.182 54.135
226661 93.053 -75.180 54.121
226662 93.076 -75.179 54.107
226663 93.099 -75.178 54.091
226664 93.123 -75.176 54.076
226665 93.146 -75.175 54.062
226670 93.263 -75.169 54.005
226671 93.286 -75.167 53.999
226672 93.309 -75.167 53.995
226673 93.333 -75.165 53.993
226674 93.356 -75.164 53.996
226675 93.379 -75.163 54.001
226676 93.403 -75.161 54.009
226677 93.426 -75.160 54.020
226678 93.449 -75.159 54.035
226679 93.473 -75.157 54.055
226680 93.496 -75.157 54.079
226681 93.519 -75.155 54.110
226682 93.556 -75.154 54.149
226683 93.592 -75.152 54.191
226684 93.623 -75.151 54.235
226685 93.654 -75.149 54.283
226686 93.686 -75.148 54.325
226687 93.717 -75.146 54.363
226688 93.749 -75.145 54.395
226689 93.780 -75.143 54.420
226690 93.810 -75.142 54.434
226691 93.836 -75.138 54.443
226692 93.862 -75.135 54.449
226693 93.888 -75.132 54.454
226694 93.914 -75.129 54.459
226695 93.940 -75.126 54.460
226696 93.966 -75.123 54.457
226697 93.992 -75.121 54.449
226698 94.017 -75.118 54.448
226699 94.042 -75.112 54.452
226700 94.064 -75.103 54.458
226701 94.088 -75.100 54.463
226702 94.115 -75.102 54.467
226703 94.139 -75.103 54.467
226704 94.163 -75.101 54.464
226705 94.187 -75.098 54.463
226706 94.210 -75.095 54.464
226707 94.234 -75.093 54.468
226708 94.258 -75.091 54.474
226709 94.281 -75.088 54.475
226710 94.307 -75.085 54.470
226711 94.331 -75.080 54.458
226712 94.356 -75.080 54.439
226713 94.382 -75.079 54.420
226714 94.408 -75.079 54.398
226715 94.433 -75.079 54.376
226716 94.459 -75.079 54.355
226717 94.485 -75.079 54.329
226718 94.507 -75.079 54.301
226719 94.522 -75.081 54.269
226720 94.536 -75.082 54.234
226721 94.550 -75.084 54.201
226722 94.564 -75.085 54.169
226723 94.578 -75.087 54.136
226724 94.592 -75.088 54.101
226725 94.606 -75.090 54.065
226726 94.620 -75.091 54.028
226727 94.631 -75.100 53.989
226728 94.641 -75.109 53.950
226729 94.659 -75.116 53.910
226730 94.680 -75.123 53.867
226731 94.700 -75.123 53.822
226732 94.723 -75.125 53.776
226733 94.748 -75.128 53.731
226734 94.772 -75.131 53.686
226735 94.775 -75.135 53.640
226736 94.768 -75.141 53.596
226737 94.762 -75.146 53.551
226738 94.755 -75.152 53.506
226739 94.748 -75.157 53.462
226740 94.763 -75.160 53.416
226741 94.781 -75.163 53.368
226742 94.799 -75.166 53.318
226743 94.817 -75.169 53.270
226744 94.835 -75.172 53.219
226745 94.852 -75.174 53.168
226746 94.871 -75.177 53.119
226747 94.888 -75.180 53.070
226748 94.906 -75.182 53.019
226749 94.922 -75.188 52.967
226750 94.930 -75.189 52.916
226751 94.957 -75.192 52.862
226752 94.982 -75.195 52.805
226753 95.007 -75.198 52.748
226754 95.033 -75.200 52.691
226755 95.058 -75.203 52.634
226756 95.084 -75.205 52.579
226757 95.109 -75.208 52.525
226758 95.135 -75.212 52.470
226759 95.161 -75.215 52.420
226760 95.186 -75.219 52.369
226761 95.215 -75.224 52.320
226762 95.242 -75.237 52.272
226763 95.265 -75.248 52.225
226764 95.289 -75.259 52.181
226765 95.313 -75.270 52.137
226766 95.337 -75.280 52.095
226767 95.361 -75.291 52.054
226768 95.386 -75.301 52.015
226769 95.410 -75.312 51.976
226770 95.434 -75.323 51.938
226771 95.458 -75.333 51.905
226772 95.482 -75.344 51.871
226773 95.507 -75.355 51.837
226774 95.531 -75.365 51.804
226775 95.555 -75.373 51.772
226776 95.584 -75.366 51.740
226777 95.613 -75.359 51.709
226778 95.641 -75.352 51.681
226779 95.670 -75.345 51.652
226780 95.698 -75.338 51.623
226781 95.727 -75.331 51.595
226782 95.756 -75.324 51.567
226783 95.784 -75.317 51.537
226784 95.813 -75.311 51.508
226785 95.845 -75.308 51.478
226786 95.883 -75.315 51.448
226787 95.921 -75.323 51.417
226788 95.963 -75.332 51.387
226789 96.005 -75.340 51.357
226790 96.046 -75.348 51.327
226791 96.089 -75.355 51.298
226792 96.130 -75.363 51.270
226793 96.172 -75.371 51.243
226794 96.211 -75.376 51.218
226795 96.249 -75.379 51.192
226796 96.286 -75.383 51.165
226797 96.324 -75.386 51.138
226798 96.360 -75.387 51.113
226799 96.380 -75.400 51.088
226800 96.400 -75.412 51.062
226801 96.419 -75.425 51.036
226802 96.439 -75.437 51.011
226803 96.459 -75.450 50.987
226804 96.479 -75.463 50.963
226805 96.499 -75.475 50.942
226806 96.519 -75.486 50.919
226807 96.538 -75.489 50.897
226808 96.557 -75.493 50.874
226809 96.576 -75.500 50.850
226810 96.594 -75.508 50.824
226811 96.613 -75.516 50.799
226812 96.633 -75.521 50.776
226813 96.654 -75.526 50.752
226814 96.671 -75.536 50.728
226815 96.681 -75.554 50.704
226816 96.691 -75.573 50.679
226817 96.701 -75.592 50.655
226818 96.711 -75.610 50.631
226819 96.721 -75.629 50.606
226820 96.731 -75.647 50.581
226821 96.741 -75.666 50.553
226822 96.751 -75.684 50.524
226823 96.760 -75.703 50.494
226824 96.770 -75.722 50.465
226825 96.784 -75.740 50.434
226826 96.842 -75.746 50.402
226827 96.907 -75.750 50.369
226828 96.971 -75.754 50.336
226829 97.036 -75.759 50.304
226830 97.084 -75.763 50.271
226831 97.106 -75.771 50.237
226832 97.127 -75.778 50.202
226833 97.149 -75.784 50.168
226834 97.170 -75.791 50.135
226835 97.191 -75.798 50.102
226836 97.213 -75.805 50.069
226837 97.234 -75.812 50.036
226838 97.256 -75.819 50.002
226839 97.277 -75.826 49.968
226840 97.299 -75.833 49.934
226841 97.320 -75.840 49.900
226842 97.342 -75.847 49.865
226843 97.363 -75.854 49.830
226844 97.387 -75.867 49.796
226845 97.412 -75.882 49.761
226846 97.437 -75.897 49.727
226847 97.461 -75.911 49.693
226848 97.499 -75.915 49.663
226849 97.542 -75.913 49.635
226850 97.585 -75.912 49.609
226851 97.628 -75.910 49.584
226852 97.668 -75.909 49.559
226853 97.699 -75.906 49.533
226854 97.723 -75.909 49.506
226855 97.734 -75.919 49.477
226856 97.745 -75.931 49.448
226857 97.756 -75.941 49.419
226858 97.759 -75.939 49.387
226859 97.749 -75.935 49.350
226860 97.722 -75.930 49.311
226861 97.740 -75.955 49.271
226862 97.757 -75.980 49.231
226863 97.776 -76.005 49.191
226864 97.793 -76.030 49.149
226865 97.812 -76.053 49.107
226866 97.840 -76.066 49.064
226867 97.867 -76.078 49.022
226868 97.895 -76.091 48.980
226869 97.937 -76.103 48.938
226870 97.983 -76.116 48.897
226871 98.028 -76.129 48.856
226872 98.082 -76.160 48.814
226873 98.137 -76.194 48.772
226874 98.191 -76.226 48.730
226875 98.236 -76.244 48.688
226876 98.282 -76.262 48.647
226877 98.326 -76.280 48.607
226878 98.359 -76.296 48.565
226879 98.371 -76.314 48.524
226880 98.383 -76.332 48.485
226881 98.394 -76.350 48.446
226882 98.377 -76.372 48.408
226883 98.400 -76.374 48.373
226884 98.421 -76.376 48.339
226885 98.443 -76.379 48.306
226886 98.467 -76.381 48.274
226887 98.494 -76.381 48.242
226888 98.521 -76.381 48.212
226889 98.549 -76.381 48.182
226890 98.576 -76.382 48.152
226891 98.603 -76.382 48.123
226892 98.630 -76.382 48.094
226893 98.657 -76.382 48.067
226894 98.685 -76.382 48.040
226895 98.715 -76.387 48.015
226896 98.747 -76.394 47.991
226897 98.771 -76.398 47.967
226898 98.788 -76.398 47.945
226899 98.806 -76.400 47.923
226900 98.830 -76.401 47.901
226901 98.856 -76.404 47.880
226902 98.882 -76.406 47.859
226903 98.908 -76.408 47.838
226904 98.934 -76.411 47.816
226905 98.960 -76.414 47.794
226906 98.987 -76.416 47.771
226907 99.012 -76.418 47.748
226908 99.038 -76.426 47.724
226909 99.062 -76.436 47.698
226910 99.085 -76.448 47.672
226911 99.094 -76.445 47.644
226912 99.102 -76.442 47.615
226913 99.110 -76.439 47.585
226914 99.126 -76.437 47.554
226915 99.146 -76.435 47.522
226916 99.167 -76.433 47.490
226917 99.187 -76.432 47.458
226918 99.207 -76.430 47.425
226919 99.227 -76.429 47.393
226920 99.257 -76.431 47.360
226921 99.286 -76.443 47.327
226922 99.315 -76.455 47.292
226923 99.345 -76.467 47.256
226924 99.374 -76.480 47.220
226925 99.403 -76.492 47.184
226926 99.432 -76.504 47.148
226927 99.462 -76.516 47.111
226928 99.492 -76.529 47.075
226929 99.523 -76.542 47.040
226930 99.554 -76.555 47.004
226931 99.585 -76.568 46.968
226932 99.620 -76.581 46.932
226933 99.655 -76.594 46.897
226934 99.691 -76.606 46.863
226935 99.725 -76.619 46.829
226936 99.761 -76.632 46.795
226937 99.796 -76.645 46.763
226938 99.821 -76.652 46.732
226939 99.833 -76.654 46.701
226940 99.844 -76.655 46.672
226941 99.859 -76.657 46.643
226942 99.875 -76.657 46.615
226943 99.891 -76.658 46.587
226944 99.908 -76.658 46.557
226945 99.924 -76.658 46.528
226946 99.940 -76.659 46.499
226947 99.956 -76.660 46.470
226948 99.973 -76.661 46.440
226949 99.989 -76.661 46.409
226950 100.005 -76.661 46.378
226951 100.021 -76.662 46.345
226952 100.037 -76.662 46.312
226953 100.054 -76.663 46.281
226954 100.070 -76.664 46.249
226955 100.091 -76.666 46.216
226956 100.113 -76.669 46.184
226957 100.140 -76.674 46.153
226958 100.168 -76.675 46.121
226959 100.182 -76.659 46.090
226960 100.196 -76.643 46.060
226961 100.210 -76.627 46.030
226962 100.225 -76.611 46.001
226963 100.239 -76.596 45.972
226964 100.253 -76.580 45.945
226965 100.270 -76.573 45.918
226966 100.287 -76.567 45.893
226967 100.304 -76.563 45.870
226968 100.322 -76.558 45.847
226969 100.339 -76.553 45.825
226970 100.359 -76.561 45.804
226971 100.380 -76.583 45.785
226972 100.402 -76.605 45.765
226973 100.427 -76.636 45.746
226974 100.454 -76.671 45.728
226975 100.480 -76.706 45.709
226976 100.507 -76.741 45.692
226977 100.534 -76.776 45.675
226978 100.560 -76.811 45.658
226979 100.587 -76.846 45.642
226980 100.613 -76.878 45.624
226981 100.635 -76.896 45.606
226982 100.657 -76.914 45.589
226983 100.680 -76.932 45.572
226984 100.709 -76.940 45.555
226985 100.750 -76.931 45.539
226986 100.791 -76.922 45.523
226987 100.833 -76.912 45.509
226988 100.874 -76.902 45.495
226989 100.896 -76.919 45.482
226990 100.906 -76.947 45.470
226991 100.917 -76.975 45.457
226992 100.928 -77.003 45.443
226993 100.941 -77.031 45.429
226994 100.970 -77.051 45.417
226995 100.999 -77.071 45.405
226996 101.029 -77.091 45.393
226997 101.057 -77.111 45.381
226998 101.087 -77.131 45.370
226999 101.116 -77.151 45.360
227000 101.145 -77.171 45.348
227001 101.174 -77.191 45.336
227002 101.203 -77.211 45.324
227003 101.232 -77.232 45.309
227004 101.247 -77.256 45.295
227005 101.259 -77.282 45.280
227006 101.271 -77.307 45.263
227007 101.283 -77.332 45.247
227008 101.295 -77.358 45.231
227009 101.308 -77.378 45.213
227010 101.323 -77.390 45.196
227011 101.340 -77.399 45.178
227012 101.360 -77.403 45.160
227013 101.379 -77.410 45.141
227014 101.398 -77.417 45.123
227015 101.414 -77.420 45.104
227016 101.429 -77.421 45.085
227017 101.445 -77.423 45.066
227018 101.460 -77.425 45.046
227019 101.476 -77.426 45.027
227020 101.491 -77.428 45.008
227021 101.506 -77.429 44.990
227022 101.522 -77.431 44.972
227023 101.538 -77.432 44.954
227024 101.553 -77.434 44.937
227025 101.568 -77.436 44.920
227026 101.584 -77.438 44.904
227027 101.599 -77.439 44.888
227028 101.615 -77.441 44.872
227029 101.607 -77.488 44.857
227030 101.627 -77.508 44.840
227031 101.647 -77.527 44.823
227032 101.667 -77.547 44.805
227033 101.687 -77.567 44.788
227034 101.722 -77.580 44.771
227035 101.758 -77.591 44.755
227036 101.794 -77.603 44.740
227037 101.830 -77.614 44.724
227038 101.867 -77.627 44.708
227039 101.903 -77.638 44.692
227040 101.928 -77.655 44.675
227041 101.952 -77.671 44.661
227042 101.995 -77.677 44.644
227043 102.039 -77.683 44.628
227044 102.076 -77.692 44.612
227045 102.109 -77.703 44.597
227046 102.141 -77.715 44.581
227047 102.174 -77.727 44.565
227048 102.206 -77.739 44.550
227049 102.239 -77.750 44.536
227050 102.268 -77.761 44.521
227051 102.295 -77.770 44.506
227052 102.318 -77.784 44.491
227053 102.341 -77.799 44.477
227054 102.364 -77.815 44.462
227055 102.383 -77.830 44.448
227056 102.402 -77.846 44.432
227057 102.420 -77.862 44.417
227058 102.448 -77.866 44.402
227059 102.476 -77.868 44.387
227060 102.505 -77.870 44.370
227061 102.532 -77.871 44.354
227062 102.556 -77.873 44.339
227063 102.581 -77.875 44.323
227064 102.605 -77.877 44.308
227065 102.623 -77.884 44.292
227066 102.641 -77.890 44.275
227067 102.658 -77.896 44.258
227068 102.676 -77.903 44.243
227069 102.704 -77.909 44.227
227070 102.734 -77.916 44.211
227071 102.770 -77.921 44.196
227072 102.807 -77.927 44.179
227073 102.843 -77.932 44.163
227074 102.881 -77.937 44.147
227075 102.917 -77.943 44.130
227076 102.955 -77.947 44.113
227077 102.992 -77.953 44.096
227078 103.028 -77.958 44.080
227079 103.066 -77.963 44.064
227080 103.102 -77.968 44.048
227081 103.140 -77.974 44.031
227082 103.176 -77.978 44.014
227083 103.213 -77.984 43.998
227084 103.250 -77.989 43.982
227327 92.139 -75.100 54.262
227328 92.161 -75.111 54.268
227329 92.184 -75.122 54.274
227330 92.208 -75.133 54.283
227331 92.231 -75.145 54.294
227332 92.254 -75.155 54.306
227333 92.277 -75.167 54.316
227334 92.300 -75.177 54.326
227335 92.323 -75.189 54.335
227336 92.346 -75.200 54.344
227337 92.369 -75.210 54.353
227338 92.392 -75.222 54.361
227339 92.411 -75.220 54.368
227340 92.436 -75.217 54.372
227341 92.467 -75.213 54.370
227342 92.498 -75.208 54.365
227343 92.529 -75.204 54.359
227344 92.560 -75.200 54.355
227345 92.590 -75.196 54.347
227346 92.622 -75.192 54.338
227347 92.652 -75.188 54.327
227348 92.683 -75.183 54.315
227349 92.714 -75.179 54.302
227350 92.745 -75.176 54.288
227351 92.776 -75.171 54.276
227352 92.806 -75.167 54.263
227353 92.837 -75.163 54.252
227354 92.866 -75.160 54.242
227355 92.888 -75.159 54.230
227356 92.912 -75.158 54.217
227357 92.935 -75.157 54.203
227358 92.958 -75.156 54.191
227359 92.982 -75.154 54.177
227360 93.005 -75.153 54.164
227361 93.028 -75.152 54.150
227362 93.052 -75.151 54.136
227363 93.075 -75.149 54.121
227364 93.098 -75.148 54.106
227365 93.122 -75.147 54.092
227366 93.145 -75.145 54.077
227367 93.168 -75.145 54.064
227370 93.238 -75.141 54.030
227371 93.262 -75.139 54.020
227372 93.285 -75.138 54.012
227373 93.308 -75.137 54.007
227374 93.331 -75.135 54.005
227375 93.355 -75.135 54.008
227376 93.378 -75.133 54.013
227377 93.401 -75.132 54.020
227378 93.425 -75.131 54.029
227379 93.448 -75.129 54.046
227380 93.471 -75.128 54.067
227381 93.495 -75.127 54.091
227382 93.518 -75.126 54.121
227383 93.554 -75.124 54.161
227384 93.585 -75.123 54.203
227385 93.616 -75.121 54.247
227386 93.648 -75.120 54.294
227387 93.679 -75.118 54.337
227388 93.711 -75.117 54.375
227389 93.743 -75.116 54.409
227390 93.775 -75.114 54.435
227391 93.807 -75.113 54.451
227392 93.833 -75.110 54.460
227393 93.858 -75.107 54.466
227394 93.885 -75.105 54.470
227395 93.910 -75.103 54.472
227396 93.936 -75.100 54.472
227397 93.962 -75.097 54.468
227398 93.988 -75.094 54.461
227399 94.014 -75.091 54.461
227400 94.040 -75.088 54.466
227401 94.066 -75.085 54.475
227402 94.088 -75.076 54.483
227403 94.114 -75.076 54.488
227404 94.138 -75.076 54.488
227405 94.162 -75.074 54.482
227406 94.186 -75.072 54.476
227407 94.209 -75.069 54.473
227408 94.234 -75.066 54.473
227409 94.259 -75.062 54.476
227410 94.284 -75.057 54.478
227411 94.306 -75.057 54.472
227412 94.330 -75.051 54.461
227413 94.356 -75.045 54.445
227414 94.382 -75.043 54.425
227415 94.407 -75.043 54.405
227416 94.433 -75.042 54.385
227417 94.459 -75.042 54.365
227418 94.484 -75.042 54.340
227419 94.506 -75.042 54.312
227420 94.521 -75.044 54.281
227421 94.535 -75.045 54.248
227422 94.549 -75.047 54.217
227423 94.563 -75.048 54.186
227424 94.577 -75.050 54.155
227425 94.591 -75.052 54.120
227426 94.603 -75.059 54.084
227427 94.613 -75.068 54.045
227428 94.623 -75.077 54.004
227429 94.636 -75.086 53.963
227430 94.657 -75.092 53.922
227431 94.678 -75.098 53.879
227432 94.699 -75.102 53.831
227433 94.718 -75.101 53.785
227434 94.742 -75.103 53.741
227435 94.767 -75.106 53.696
227436 94.785 -75.109 53.652
227437 94.778 -75.115 53.608
227438 94.772 -75.120 53.563
227439 94.765 -75.126 53.518
227440 94.764 -75.130 53.473
227441 94.782 -75.133 53.426
227442 94.800 -75.136 53.379
227443 94.818 -75.138 53.332
227444 94.835 -75.142 53.285
227445 94.853 -75.145 53.235
227446 94.871 -75.147 53.185
227447 94.889 -75.150 53.136
227448 94.905 -75.155 53.086
227449 94.920 -75.160 53.035
227450 94.936 -75.166 52.984
227451 94.935 -75.160 52.932
227452 94.964 -75.165 52.878
227453 94.992 -75.170 52.819
227454 95.018 -75.172 52.759
227455 95.043 -75.174 52.701
227456 95.068 -75.177 52.643
227457 95.093 -75.179 52.587
227458 95.118 -75.182 52.531
227459 95.143 -75.185 52.475
227460 95.169 -75.187 52.424
227461 95.194 -75.189 52.373
227462 95.218 -75.193 52.322
227463 95.242 -75.202 52.273
227464 95.266 -75.213 52.225
227465 95.290 -75.223 52.180
227466 95.314 -75.234 52.137
227467 95.338 -75.245 52.094
227468 95.362 -75.255 52.054
227469 95.387 -75.266 52.014
227470 95.411 -75.276 51.974
227471 95.435 -75.287 51.937
227472 95.459 -75.298 51.902
227473 95.483 -75.308 51.868
227474 95.508 -75.319 51.834
227475 95.532 -75.330 51.801
227476 95.557 -75.336 51.770
227477 95.586 -75.329 51.738
227478 95.614 -75.322 51.707
227479 95.642 -75.315 51.678
227480 95.671 -75.308 51.650
227481 95.700 -75.301 51.621
227482 95.728 -75.294 51.593
227483 95.757 -75.287 51.565
227484 95.785 -75.280 51.536
227485 95.819 -75.279 51.507
227486 95.860 -75.285 51.478
227487 95.897 -75.293 51.449
227488 95.933 -75.301 51.419
227489 95.970 -75.310 51.389
227490 96.011 -75.319 51.360
227491 96.051 -75.328 51.330
227492 96.092 -75.336 51.302
227493 96.134 -75.344 51.275
227494 96.176 -75.352 51.248
227495 96.215 -75.357 51.223
227496 96.253 -75.360 51.197
227497 96.292 -75.362 51.171
227498 96.331 -75.364 51.144
227499 96.362 -75.364 51.120
227500 96.387 -75.372 51.095
227501 96.407 -75.384 51.069
227502 96.427 -75.396 51.043
227503 96.447 -75.409 51.019
227504 96.467 -75.422 50.995
227505 96.486 -75.434 50.972
227506 96.507 -75.447 50.950
227507 96.526 -75.459 50.928
227508 96.546 -75.467 50.905
227509 96.565 -75.470 50.881
227510 96.584 -75.477 50.856
227511 96.602 -75.485 50.830
227512 96.621 -75.492 50.805
227513 96.642 -75.496 50.781
227514 96.663 -75.501 50.757
227515 96.681 -75.510 50.733
227516 96.691 -75.529 50.709
227517 96.701 -75.547 50.685
227518 96.711 -75.566 50.661
227519 96.721 -75.584 50.637
227520 96.731 -75.603 50.613
227521 96.741 -75.622 50.587
227522 96.751 -75.640 50.559
227523 96.760 -75.659 50.529
227524 96.770 -75.677 50.499
227525 96.784 -75.699 50.469
227526 96.805 -75.714 50.440
227527 96.853 -75.721 50.408
227528 96.917 -75.725 50.376
227529 96.981 -75.730 50.344
227530 97.046 -75.734 50.313
227531 97.095 -75.739 50.281
227532 97.119 -75.745 50.247
227533 97.143 -75.750 50.213
227534 97.165 -75.757 50.180
227535 97.186 -75.765 50.147
227536 97.208 -75.771 50.114
227537 97.229 -75.778 50.082
227538 97.251 -75.785 50.049
227539 97.273 -75.792 50.016
227540 97.294 -75.799 49.982
227541 97.315 -75.806 49.948
227542 97.336 -75.813 49.914
227543 97.358 -75.820 49.879
227544 97.380 -75.828 49.844
227545 97.405 -75.844 49.809
227546 97.430 -75.858 49.774
227547 97.455 -75.873 49.739
227548 97.479 -75.887 49.705
227549 97.514 -75.894 49.673
227550 97.557 -75.892 49.645
227551 97.599 -75.891 49.617
227552 97.640 -75.890 49.591
227553 97.671 -75.887 49.566
227554 97.703 -75.884 49.540
227555 97.734 -75.882 49.512
227556 97.745 -75.893 49.482
227557 97.756 -75.904 49.452
227558 97.764 -75.909 49.421
227559 97.767 -75.906 49.388
227560 97.770 -75.903 49.350
227561 97.755 -75.899 49.311
227562 97.766 -75.919 49.270
227563 97.784 -75.944 49.229
227564 97.802 -75.969 49.187
227565 97.819 -75.994 49.145
227566 97.838 -76.019 49.103
227567 97.856 -76.042 49.061
227568 97.884 -76.054 49.020
227569 97.912 -76.066 48.977
227570 97.953 -76.079 48.935
227571 97.999 -76.092 48.894
227572 98.052 -76.120 48.852
227573 98.106 -76.154 48.810
227574 98.154 -76.176 48.767
227575 98.200 -76.194 48.725
227576 98.246 -76.213 48.683
227577 98.292 -76.231 48.642
227578 98.335 -76.248 48.602
227579 98.375 -76.264 48.560
227580 98.387 -76.282 48.519
227581 98.398 -76.300 48.479
227582 98.410 -76.317 48.440
227583 98.422 -76.336 48.402
227584 98.423 -76.357 48.365
227585 98.431 -76.367 48.331
227586 98.453 -76.369 48.298
227587 98.475 -76.371 48.265
227588 98.502 -76.371 48.233
227589 98.529 -76.371 48.203
227590 98.557 -76.371 48.172
227591 98.584 -76.372 48.141
227592 98.611 -76.372 48.113
227593 98.638 -76.372 48.085
227594 98.665 -76.372 48.058
227595 98.693 -76.372 48.031
227596 98.720 -76.373 48.008
227597 98.747 -76.373 47.984
227598 98.775 -76.373 47.961
227599 98.801 -76.375 47.939
227600 98.827 -76.377 47.917
227601 98.853 -76.379 47.895
227602 98.879 -76.382 47.874
227603 98.905 -76.384 47.853
227604 98.931 -76.386 47.833
227605 98.957 -76.389 47.813
227606 98.983 -76.392 47.790
227607 99.009 -76.394 47.767
227608 99.035 -76.397 47.744
227609 99.059 -76.408 47.719
227610 99.083 -76.419 47.693
227611 99.108 -76.430 47.667
227612 99.121 -76.431 47.639
227613 99.129 -76.428 47.610
227614 99.137 -76.426 47.581
227615 99.145 -76.423 47.548
227616 99.157 -76.420 47.516
227617 99.178 -76.418 47.484
227618 99.198 -76.417 47.452
227619 99.218 -76.415 47.420
227620 99.238 -76.414 47.388
227621 99.268 -76.414 47.356
227622 99.298 -76.420 47.322
227623 99.327 -76.433 47.287
227624 99.356 -76.445 47.253
227625 99.385 -76.458 47.218
227626 99.415 -76.470 47.182
227627 99.444 -76.482 47.146
227628 99.474 -76.494 47.111
227629 99.505 -76.508 47.076
227630 99.536 -76.520 47.040
227631 99.567 -76.533 47.005
227632 99.598 -76.547 46.970
227633 99.634 -76.559 46.934
227634 99.669 -76.572 46.900
227635 99.704 -76.585 46.866
227636 99.739 -76.598 46.833
227637 99.775 -76.611 46.800
227638 99.809 -76.623 46.768
227639 99.845 -76.636 46.736
227640 99.869 -76.643 46.706
227641 99.880 -76.645 46.678
227642 99.895 -76.646 46.649
227643 99.911 -76.646 46.621
227644 99.927 -76.647 46.592
227645 99.943 -76.647 46.562
227646 99.960 -76.648 46.533
227647 99.976 -76.649 46.504
227648 99.992 -76.649 46.474
227649 100.008 -76.649 46.443
227650 100.025 -76.650 46.411
227651 100.041 -76.651 46.379
227652 100.057 -76.651 46.347
227653 100.073 -76.652 46.313
227654 100.089 -76.652 46.280
227655 100.106 -76.652 46.248
227656 100.122 -76.653 46.215
227657 100.138 -76.654 46.183
227658 100.156 -76.646 46.151
227659 100.174 -76.642 46.120
227660 100.191 -76.637 46.088
227661 100.208 -76.632 46.058
227662 100.226 -76.627 46.029
227663 100.243 -76.623 46.000
227664 100.261 -76.618 45.971
227665 100.278 -76.613 45.943
227666 100.295 -76.608 45.916
227667 100.313 -76.604 45.891
227668 100.330 -76.599 45.868
227669 100.347 -76.594 45.845
227670 100.364 -76.589 45.823
227671 100.382 -76.586 45.802
227672 100.404 -76.608 45.782
227673 100.425 -76.630 45.763
227674 100.446 -76.652 45.744
227675 100.468 -76.674 45.725
227676 100.489 -76.696 45.706
227677 100.510 -76.718 45.689
227678 100.533 -76.743 45.671
227679 100.560 -76.777 45.654
227680 100.586 -76.812 45.638
227681 100.613 -76.847 45.620
227682 100.638 -76.877 45.602
227683 100.661 -76.895 45.584
227684 100.689 -76.905 45.566
227685 100.730 -76.896 45.549
227686 100.772 -76.886 45.533
227687 100.796 -76.896 45.517
227688 100.817 -76.911 45.502
227689 100.838 -76.925 45.488
227690 100.837 -76.950 45.474
227691 100.864 -76.971 45.462
227692 100.893 -76.990 45.449
227693 100.922 -77.010 45.435
227694 100.952 -77.031 45.422
227695 100.980 -77.050 45.409
227696 101.009 -77.071 45.396
227697 101.039 -77.091 45.385
227698 101.068 -77.111 45.373
227699 101.097 -77.131 45.362
227700 101.126 -77.151 45.351
227701 101.155 -77.171 45.340
227702 101.184 -77.191 45.327
227703 101.213 -77.211 45.315
227704 101.243 -77.231 45.300
227705 101.257 -77.256 45.286
227706 101.268 -77.282 45.271
227707 101.280 -77.307 45.253
227708 101.293 -77.331 45.237
227709 101.308 -77.343 45.220
227710 101.322 -77.355 45.202
227711 101.337 -77.368 45.186
227712 101.356 -77.374 45.168
227713 101.378 -77.376 45.150
227714 101.396 -77.383 45.131
227715 101.414 -77.389 45.112
227716 101.433 -77.396 45.093
227717 101.452 -77.403 45.075
227718 101.469 -77.407 45.056
227719 101.484 -77.409 45.037
227720 101.500 -77.410 45.018
227721 101.515 -77.412 44.999
227722 101.531 -77.414 44.981
227723 101.546 -77.415 44.963
227724 101.561 -77.417 44.945
227725 101.577 -77.419 44.928
227726 101.592 -77.420 44.912
227727 101.608 -77.422 44.895
227728 101.623 -77.423 44.879
227729 101.638 -77.425 44.863
227730 101.634 -77.464 44.848
227731 101.654 -77.484 44.832
227732 101.674 -77.504 44.815
227733 101.694 -77.523 44.798
227734 101.714 -77.543 44.780
227735 101.734 -77.563 44.763
227736 101.761 -77.579 44.747
227737 101.798 -77.591 44.731
227738 101.834 -77.603 44.715
227739 101.870 -77.614 44.699
227740 101.906 -77.627 44.683
227741 101.930 -77.643 44.667
227742 101.964 -77.655 44.652
227743 102.007 -77.660 44.636
227744 102.051 -77.666 44.620
227745 102.089 -77.674 44.604
227746 102.121 -77.686 44.588
227747 102.154 -77.698 44.573
227748 102.186 -77.710 44.558
227749 102.219 -77.721 44.543
227750 102.251 -77.733 44.529
227751 102.284 -77.745 44.514
227752 102.307 -77.760 44.499
227753 102.326 -77.776 44.483
227754 102.345 -77.792 44.468
227755 102.364 -77.808 44.454
227756 102.383 -77.824 44.439
227757 102.402 -77.840 44.424
227758 102.429 -77.844 44.409
227759 102.458 -77.846 44.393
227760 102.486 -77.847 44.377
227761 102.515 -77.849 44.361
227762 102.544 -77.851 44.345
227763 102.573 -77.852 44.329
227764 102.599 -77.855 44.314
227765 102.616 -77.862 44.298
227766 102.634 -77.868 44.282
227767 102.652 -77.874 44.265
227768 102.670 -77.881 44.248
227769 102.691 -77.887 44.232
227770 102.720 -77.894 44.216
227771 102.754 -77.899 44.200
227772 102.791 -77.905 44.184
227773 102.828 -77.910 44.167
227774 102.865 -77.915 44.152
227775 102.902 -77.920 44.136
227776 102.939 -77.925 44.119
227777 102.976 -77.931 44.101
227778 103.013 -77.936 44.084
227779 103.050 -77.941 44.068
227780 103.087 -77.946 44.052
227781 103.124 -77.951 44.036
227782 103.161 -77.956 44.020
227783 103.198 -77.962 44.003
227784 103.235 -77.967 43.987
228028 92.126 -75.071 54.266
228029 92.149 -75.082 54.273
228030 92.172 -75.093 54.278
228031 92.195 -75.104 54.286
228032 92.218 -75.115 54.297
228033 92.241 -75.126 54.309
228034 92.264 -75.137 54.319
228035 92.287 -75.148 54.330
228036 92.310 -75.160 54.341
228037 92.334 -75.170 54.349
228038 92.357 -75.182 54.360
228039 92.379 -75.193 54.370
228040 92.400 -75.197 54.379
228041 92.429 -75.193 54.384
228042 92.460 -75.189 54.384
228043 92.491 -75.185 54.380
228044 92.522 -75.180 54.374
228045 92.553 -75.176 54.370
228046 92.583 -75.172 54.363
228047 92.614 -75.168 54.353
228048 92.645 -75.163 54.343
228049 92.676 -75.160 54.329
228050 92.708 -75.156 54.315
228051 92.739 -75.151 54.302
228052 92.771 -75.148 54.290
228053 92.802 -75.142 54.278
228054 92.834 -75.136 54.267
228055 92.863 -75.136 54.257
228056 92.888 -75.128 54.246
228057 92.912 -75.126 54.233
228058 92.935 -75.126 54.220
228059 92.958 -75.124 54.206
228060 92.982 -75.123 54.193
228061 93.005 -75.122 54.180
228062 93.028 -75.120 54.166
228063 93.051 -75.120 54.152
228064 93.074 -75.118 54.137
228065 93.098 -75.117 54.123
228066 93.121 -75.116 54.108
228067 93.144 -75.115 54.093
228068 93.167 -75.113 54.079
228071 93.237 -75.110 54.046
228072 93.261 -75.109 54.036
228073 93.284 -75.107 54.029
228074 93.307 -75.107 54.023
228075 93.331 -75.105 54.021
228076 93.354 -75.104 54.023
228077 93.377 -75.103 54.028
228078 93.401 -75.102 54.035
228079 93.424 -75.101 54.045
228080 93.447 -75.099 54.060
228081 93.470 -75.098 54.080
228082 93.494 -75.097 54.105
228083 93.517 -75.096 54.135
228084 93.548 -75.095 54.174
228085 93.580 -75.093 54.215
228086 93.612 -75.091 54.258
228087 93.644 -75.090 54.303
228088 93.677 -75.088 54.344
228089 93.709 -75.088 54.382
228090 93.741 -75.086 54.417
228091 93.773 -75.085 54.448
228092 93.805 -75.083 54.464
228093 93.831 -75.081 54.473
228094 93.856 -75.078 54.478
228095 93.882 -75.076 54.481
228096 93.908 -75.073 54.482
228097 93.933 -75.071 54.480
228098 93.959 -75.069 54.477
228099 93.985 -75.066 54.471
228100 94.010 -75.064 54.473
228101 94.037 -75.061 54.481
228102 94.062 -75.058 54.491
228103 94.088 -75.056 54.501
228104 94.113 -75.050 54.508
228105 94.137 -75.050 54.510
228106 94.161 -75.048 54.502
228107 94.186 -75.044 54.492
228108 94.211 -75.039 54.484
228109 94.237 -75.035 54.479
228110 94.262 -75.031 54.477
228111 94.287 -75.027 54.477
228112 94.305 -75.029 54.471
228113 94.330 -75.023 54.460
228114 94.355 -75.017 54.446
228115 94.381 -75.011 54.426
228116 94.407 -75.006 54.408
228117 94.432 -75.006 54.390
228118 94.458 -75.006 54.372
228119 94.484 -75.005 54.348
228120 94.506 -75.006 54.321
228121 94.520 -75.007 54.292
228122 94.534 -75.009 54.262
228123 94.548 -75.010 54.232
228124 94.562 -75.012 54.202
228125 94.574 -75.018 54.172
228126 94.584 -75.027 54.136
228127 94.595 -75.036 54.098
228128 94.605 -75.046 54.057
228129 94.616 -75.055 54.015
228130 94.634 -75.062 53.973
228131 94.655 -75.068 53.932
228132 94.677 -75.075 53.889
228133 94.698 -75.081 53.843
228134 94.718 -75.080 53.797
228135 94.738 -75.079 53.752
228136 94.761 -75.081 53.708
228137 94.785 -75.084 53.664
228138 94.788 -75.088 53.618
228139 94.782 -75.094 53.573
228140 94.775 -75.099 53.527
228141 94.782 -75.103 53.483
228142 94.800 -75.106 53.437
228143 94.818 -75.109 53.392
228144 94.836 -75.111 53.346
228145 94.853 -75.114 53.299
228146 94.872 -75.117 53.251
228147 94.888 -75.122 53.203
228148 94.903 -75.127 53.156
228149 94.919 -75.133 53.106
228150 94.934 -75.138 53.056
228151 94.946 -75.141 53.004
228152 94.939 -75.131 52.951
228153 94.970 -75.137 52.894
228154 95.002 -75.144 52.834
228155 95.027 -75.146 52.773
228156 95.052 -75.148 52.713
228157 95.077 -75.151 52.653
228158 95.103 -75.154 52.595
228159 95.128 -75.156 52.537
228160 95.154 -75.159 52.481
228161 95.179 -75.161 52.428
228162 95.204 -75.163 52.376
228163 95.230 -75.170 52.325
228164 95.249 -75.174 52.274
228165 95.266 -75.177 52.226
228166 95.290 -75.188 52.181
228167 95.315 -75.198 52.137
228168 95.339 -75.209 52.095
228169 95.364 -75.220 52.053
228170 95.388 -75.230 52.013
228171 95.412 -75.241 51.974
228172 95.436 -75.251 51.936
228173 95.460 -75.262 51.900
228174 95.484 -75.273 51.865
228175 95.509 -75.283 51.832
228176 95.533 -75.294 51.799
228177 95.558 -75.298 51.768
228178 95.587 -75.292 51.737
228179 95.615 -75.285 51.706
228180 95.644 -75.278 51.676
228181 95.673 -75.271 51.647
228182 95.701 -75.264 51.619
228183 95.730 -75.258 51.591
228184 95.758 -75.251 51.563
228185 95.794 -75.251 51.534
228186 95.835 -75.257 51.506
228187 95.876 -75.262 51.479
228188 95.911 -75.270 51.450
228189 95.947 -75.279 51.421
228190 95.982 -75.287 51.392
228191 96.019 -75.296 51.363
228192 96.059 -75.305 51.334
228193 96.100 -75.314 51.306
228194 96.140 -75.323 51.280
228195 96.180 -75.332 51.254
228196 96.220 -75.337 51.228
228197 96.260 -75.338 51.202
228198 96.300 -75.339 51.177
228199 96.333 -75.340 51.150
228200 96.364 -75.341 51.126
228201 96.394 -75.343 51.101
228202 96.414 -75.356 51.076
228203 96.434 -75.368 51.051
228204 96.454 -75.381 51.027
228205 96.474 -75.393 51.003
228206 96.494 -75.406 50.980
228207 96.514 -75.418 50.958
228208 96.534 -75.431 50.936
228209 96.553 -75.443 50.912
228210 96.573 -75.448 50.888
228211 96.592 -75.454 50.862
228212 96.610 -75.461 50.836
228213 96.630 -75.467 50.811
228214 96.652 -75.471 50.787
228215 96.673 -75.476 50.763
228216 96.691 -75.484 50.738
228217 96.701 -75.503 50.714
228218 96.711 -75.521 50.691
228219 96.721 -75.540 50.667
228220 96.731 -75.559 50.644
228221 96.741 -75.577 50.619
228222 96.751 -75.596 50.593
228223 96.760 -75.615 50.565
228224 96.772 -75.634 50.536
228225 96.786 -75.656 50.506
228226 96.802 -75.674 50.475
228227 96.826 -75.687 50.446
228228 96.862 -75.697 50.415
228229 96.927 -75.701 50.384
228230 96.992 -75.706 50.353
228231 97.056 -75.709 50.322
228232 97.105 -75.715 50.290
228233 97.129 -75.720 50.258
228234 97.153 -75.726 50.225
228235 97.177 -75.732 50.193
228236 97.202 -75.738 50.160
228237 97.225 -75.744 50.127
228238 97.246 -75.751 50.095
228239 97.267 -75.759 50.062
228240 97.289 -75.765 50.028
228241 97.310 -75.772 49.995
228242 97.332 -75.779 49.961
228243 97.353 -75.786 49.927
228244 97.375 -75.793 49.893
228245 97.398 -75.805 49.857
228246 97.423 -75.819 49.822
228247 97.448 -75.834 49.787
228248 97.472 -75.849 49.752
228249 97.497 -75.864 49.717
228250 97.528 -75.873 49.683
228251 97.571 -75.872 49.654
228252 97.613 -75.870 49.625
228253 97.644 -75.868 49.598
228254 97.675 -75.865 49.572
228255 97.706 -75.862 49.544
228256 97.737 -75.860 49.516
228257 97.756 -75.866 49.485
228258 97.767 -75.877 49.454
228259 97.771 -75.875 49.422
228260 97.775 -75.873 49.387
228261 97.778 -75.870 49.349
228262 97.781 -75.867 49.309
228263 97.793 -75.883 49.268
228264 97.810 -75.908 49.225
228265 97.828 -75.933 49.182
228266 97.846 -75.957 49.141
228267 97.864 -75.982 49.099
228268 97.882 -76.007 49.057
228269 97.901 -76.030 49.017
228270 97.929 -76.043 48.974
228271 97.970 -76.055 48.932
228272 98.021 -76.079 48.890
228273 98.072 -76.107 48.847
228274 98.118 -76.126 48.805
228275 98.164 -76.144 48.762
228276 98.210 -76.163 48.719
228277 98.256 -76.181 48.678
228278 98.302 -76.200 48.638
228279 98.343 -76.216 48.597
228280 98.383 -76.232 48.555
228281 98.403 -76.249 48.514
228282 98.414 -76.267 48.473
228283 98.426 -76.285 48.434
228284 98.438 -76.303 48.395
228285 98.449 -76.321 48.358
228286 98.461 -76.339 48.323
228287 98.463 -76.358 48.289
228288 98.485 -76.361 48.256
228289 98.510 -76.361 48.223
228290 98.537 -76.361 48.192
228291 98.565 -76.361 48.162
228292 98.592 -76.362 48.132
228293 98.619 -76.362 48.102
228294 98.649 -76.361 48.073
228295 98.681 -76.358 48.045
228296 98.713 -76.355 48.019
228297 98.744 -76.353 47.996
228298 98.772 -76.351 47.974
228299 98.798 -76.351 47.952
228300 98.824 -76.353 47.930
228301 98.850 -76.355 47.908
228302 98.876 -76.357 47.887
228303 98.902 -76.360 47.866
228304 98.928 -76.362 47.845
228305 98.954 -76.364 47.826
228306 98.980 -76.367 47.806
228307 99.006 -76.369 47.784
228308 99.032 -76.372 47.761
228309 99.057 -76.379 47.737
228310 99.081 -76.390 47.712
228311 99.106 -76.401 47.687
228312 99.130 -76.412 47.660
228313 99.148 -76.417 47.631
228314 99.156 -76.415 47.602
228315 99.164 -76.412 47.573
228316 99.172 -76.409 47.541
228317 99.181 -76.406 47.509
228318 99.189 -76.404 47.477
228319 99.208 -76.401 47.445
228320 99.229 -76.400 47.413
228321 99.249 -76.398 47.382
228322 99.279 -76.398 47.350
228323 99.309 -76.398 47.316
228324 99.339 -76.411 47.282
228325 99.368 -76.423 47.248
228326 99.397 -76.435 47.214
228327 99.427 -76.447 47.180
228328 99.456 -76.459 47.144
228329 99.487 -76.473 47.110
228330 99.518 -76.486 47.075
228331 99.549 -76.499 47.040
228332 99.579 -76.512 47.006
228333 99.612 -76.525 46.971
228334 99.647 -76.538 46.936
228335 99.683 -76.551 46.902
228336 99.718 -76.564 46.869
228337 99.753 -76.576 46.837
228338 99.788 -76.589 46.804
228339 99.824 -76.602 46.772
228340 99.859 -76.614 46.741
228341 99.894 -76.627 46.712
228342 99.917 -76.634 46.684
228343 99.931 -76.635 46.655
228344 99.947 -76.636 46.625
228345 99.963 -76.636 46.596
228346 99.980 -76.636 46.566
228347 99.996 -76.637 46.537
228348 100.012 -76.637 46.507
228349 100.028 -76.638 46.477
228350 100.045 -76.639 46.446
228351 100.061 -76.639 46.413
228352 100.074 -76.642 46.381
228353 100.088 -76.645 46.347
228354 100.102 -76.648 46.313
228355 100.115 -76.651 46.280
228356 100.129 -76.654 46.247
228357 100.142 -76.657 46.215
228358 100.157 -76.657 46.182
228359 100.176 -76.651 46.149
228360 100.196 -76.658 46.118
228361 100.215 -76.666 46.086
228362 100.233 -76.668 46.056
228363 100.251 -76.664 46.026
228364 100.268 -76.658 45.997
228365 100.286 -76.654 45.969
228366 100.303 -76.649 45.941
228367 100.321 -76.644 45.915
228368 100.338 -76.640 45.890
228369 100.355 -76.635 45.866
228370 100.372 -76.630 45.843
228371 100.390 -76.625 45.820
228372 100.407 -76.621 45.799
228373 100.427 -76.633 45.779
228374 100.449 -76.654 45.759
228375 100.470 -76.676 45.740
228376 100.491 -76.698 45.721
228377 100.512 -76.720 45.703
228378 100.534 -76.742 45.685
228379 100.555 -76.764 45.667
228380 100.576 -76.786 45.649
228381 100.598 -76.808 45.632
228382 100.619 -76.830 45.615
228383 100.641 -76.852 45.597
228384 100.668 -76.870 45.578
228385 100.691 -76.884 45.560
228386 100.719 -76.894 45.543
228387 100.747 -76.905 45.526
228388 100.775 -76.915 45.510
228389 100.804 -76.926 45.495
228390 100.831 -76.937 45.480
228391 100.849 -76.949 45.466
228392 100.878 -76.971 45.453
228393 100.907 -76.994 45.440
228394 100.935 -77.016 45.427
228395 100.964 -77.035 45.413
228396 100.991 -77.050 45.400
228397 101.020 -77.071 45.388
228398 101.049 -77.091 45.376
228399 101.078 -77.110 45.365
228400 101.107 -77.131 45.354
228401 101.136 -77.150 45.344
228402 101.165 -77.171 45.332
228403 101.195 -77.191 45.319
228404 101.224 -77.211 45.306
228405 101.253 -77.231 45.291
228406 101.266 -77.257 45.276
228407 101.277 -77.282 45.261
228408 101.292 -77.296 45.244
228409 101.307 -77.308 45.227
228410 101.322 -77.321 45.211
228411 101.337 -77.333 45.194
228412 101.351 -77.346 45.177
228413 101.373 -77.348 45.159
228414 101.395 -77.350 45.140
228415 101.413 -77.357 45.122
228416 101.431 -77.363 45.103
228417 101.450 -77.370 45.084
228418 101.469 -77.376 45.066
228419 101.487 -77.382 45.048
228420 101.506 -77.389 45.029
228421 101.524 -77.395 45.010
228422 101.539 -77.396 44.991
228423 101.555 -77.398 44.972
228424 101.570 -77.399 44.953
228425 101.586 -77.401 44.936
228426 101.601 -77.403 44.919
228427 101.616 -77.404 44.903
228428 101.632 -77.406 44.886
228429 101.647 -77.407 44.870
228430 101.663 -77.409 44.854
228431 101.662 -77.441 44.838
228432 101.681 -77.461 44.823
228433 101.701 -77.480 44.807
228434 101.722 -77.500 44.790
228435 101.741 -77.520 44.772
228436 101.761 -77.539 44.757
228437 101.781 -77.559 44.740
228438 101.801 -77.579 44.723
228439 101.837 -77.591 44.707
228440 101.874 -77.602 44.692
228441 101.910 -77.614 44.676
228442 101.933 -77.632 44.659
228443 101.976 -77.638 44.643
228444 102.020 -77.643 44.628
228445 102.077 -77.652 44.612
228446 102.097 -77.661 44.596
228447 102.129 -77.672 44.580
228448 102.162 -77.683 44.564
228449 102.194 -77.696 44.549
228450 102.227 -77.707 44.535
228451 102.259 -77.719 44.521
228452 102.292 -77.730 44.506
228453 102.313 -77.748 44.491
228454 102.328 -77.768 44.476
228455 102.345 -77.786 44.461
228456 102.365 -77.802 44.446
228457 102.383 -77.818 44.431
228458 102.410 -77.822 44.415
228459 102.439 -77.824 44.400
228460 102.468 -77.825 44.384
228461 102.496 -77.827 44.367
228462 102.522 -77.831 44.352
228463 102.548 -77.835 44.336
228464 102.575 -77.839 44.320
228465 102.601 -77.843 44.304
228466 102.627 -77.846 44.288
228467 102.646 -77.852 44.271
228468 102.664 -77.859 44.255
228469 102.681 -77.865 44.239
228470 102.705 -77.871 44.222
228471 102.739 -77.877 44.206
228472 102.776 -77.883 44.190
228473 102.813 -77.888 44.173
228474 102.850 -77.893 44.156
228475 102.887 -77.898 44.140
228476 102.924 -77.903 44.124
228477 102.961 -77.909 44.107
228478 102.998 -77.914 44.090
228479 103.035 -77.919 44.074
228480 103.072 -77.924 44.058
228481 103.109 -77.929 44.042
228482 103.146 -77.934 44.026
228483 103.182 -77.940 44.009
228484 103.220 -77.945 43.993
228729 92.113 -75.041 54.271
228730 92.136 -75.053 54.277
228731 92.159 -75.064 54.282
228732 92.182 -75.075 54.290
228733 92.206 -75.086 54.301
228734 92.229 -75.097 54.311
228735 92.251 -75.108 54.322
228736 92.275 -75.120 54.334
228737 92.298 -75.130 54.345
228738 92.321 -75.142 54.354
228739 92.344 -75.153 54.366
228740 92.367 -75.163 54.378
228741 92.391 -75.173 54.389
228742 92.423 -75.169 54.396
228743 92.454 -75.165 54.398
228744 92.486 -75.161 54.395
228745 92.517 -75.157 54.391
228746 92.549 -75.153 54.388
228747 92.580 -75.149 54.382
228748 92.611 -75.145 54.373
228749 92.643 -75.141 54.364
228750 92.674 -75.137 54.348
228751 92.706 -75.132 54.332
228752 92.738 -75.124 54.317
228753 92.770 -75.116 54.304
228754 92.802 -75.113 54.292
228755 92.831 -75.123 54.281
228756 92.859 -75.123 54.272
228757 92.888 -75.095 54.263
228758 92.912 -75.094 54.251
228759 92.935 -75.093 54.238
228760 92.958 -75.091 54.224
228761 92.982 -75.091 54.209
228762 93.005 -75.089 54.196
228763 93.028 -75.088 54.182
228764 93.052 -75.087 54.169
228765 93.075 -75.085 54.153
228766 93.098 -75.085 54.140
228767 93.122 -75.083 54.125
228768 93.145 -75.082 54.110
228769 93.168 -75.081 54.096
228770 93.191 -75.080 54.083
228771 93.215 -75.079 54.072
228772 93.238 -75.078 54.063
228773 93.261 -75.076 54.053
228774 93.284 -75.075 54.046
228775 93.308 -75.074 54.041
228776 93.331 -75.073 54.038
228777 93.354 -75.072 54.040
228778 93.377 -75.070 54.044
228779 93.401 -75.069 54.052
228780 93.425 -75.069 54.061
228781 93.450 -75.071 54.076
228782 93.477 -75.073 54.095
228783 93.499 -75.066 54.119
228784 93.522 -75.059 54.149
228785 93.551 -75.060 54.187
228786 93.579 -75.063 54.227
228787 93.610 -75.062 54.268
228788 93.642 -75.060 54.310
228789 93.675 -75.060 54.350
228790 93.707 -75.058 54.387
228791 93.739 -75.057 54.422
228792 93.771 -75.055 54.452
228793 93.802 -75.054 54.471
228794 93.828 -75.051 54.482
228795 93.854 -75.048 54.488
228796 93.880 -75.046 54.489
228797 93.905 -75.044 54.489
228798 93.931 -75.041 54.487
228799 93.957 -75.039 54.484
228800 93.982 -75.037 54.480
228801 94.008 -75.035 54.485
228802 94.034 -75.032 54.494
228803 94.060 -75.029 54.505
228804 94.085 -75.027 54.515
228805 94.111 -75.025 54.522
228806 94.138 -75.021 54.523
228807 94.164 -75.017 54.518
228808 94.189 -75.013 54.507
228809 94.214 -75.009 54.495
228810 94.240 -75.004 54.484
228811 94.265 -75.001 54.476
228812 94.288 -74.998 54.473
228813 94.305 -75.001 54.465
228814 94.330 -74.995 54.452
228815 94.355 -74.989 54.438
228816 94.381 -74.983 54.422
228817 94.406 -74.977 54.407
228818 94.432 -74.971 54.392
228819 94.458 -74.969 54.375
228820 94.483 -74.969 54.353
228821 94.505 -74.969 54.329
228822 94.519 -74.971 54.302
228823 94.533 -74.972 54.274
228824 94.545 -74.977 54.246
228825 94.556 -74.986 54.218
228826 94.566 -74.996 54.188
228827 94.577 -75.005 54.151
228828 94.588 -75.014 54.110
228829 94.598 -75.023 54.068
228830 94.612 -75.032 54.025
228831 94.633 -75.038 53.984
228832 94.654 -75.044 53.944
228833 94.676 -75.051 53.903
228834 94.697 -75.057 53.860
228835 94.717 -75.060 53.814
228836 94.737 -75.058 53.769
228837 94.757 -75.057 53.725
228838 94.779 -75.058 53.680
228839 94.799 -75.062 53.634
228840 94.792 -75.067 53.588
228841 94.785 -75.073 53.543
228842 94.801 -75.076 53.498
228843 94.819 -75.079 53.454
228844 94.836 -75.082 53.409
228845 94.854 -75.085 53.364
228846 94.871 -75.088 53.317
228847 94.886 -75.095 53.270
228848 94.901 -75.100 53.223
228849 94.917 -75.105 53.177
228850 94.932 -75.111 53.128
228851 94.948 -75.116 53.077
228852 94.950 -75.113 53.024
228853 94.944 -75.102 52.969
228854 94.976 -75.109 52.911
228855 95.007 -75.116 52.850
228856 95.037 -75.120 52.788
228857 95.062 -75.123 52.725
228858 95.088 -75.126 52.665
228859 95.113 -75.128 52.605
228860 95.138 -75.131 52.546
228861 95.163 -75.133 52.490
228862 95.188 -75.135 52.435
228863 95.213 -75.138 52.382
228864 95.241 -75.147 52.330
228865 95.265 -75.156 52.279
228866 95.280 -75.155 52.229
228867 95.294 -75.155 52.183
228868 95.316 -75.163 52.139
228869 95.340 -75.173 52.096
228870 95.364 -75.184 52.055
228871 95.388 -75.195 52.014
228872 95.412 -75.205 51.974
228873 95.437 -75.216 51.936
228874 95.461 -75.226 51.899
228875 95.485 -75.237 51.864
228876 95.509 -75.248 51.830
228877 95.533 -75.258 51.797
228878 95.559 -75.262 51.766
228879 95.588 -75.255 51.735
228880 95.617 -75.248 51.704
228881 95.645 -75.241 51.675
228882 95.674 -75.234 51.646
228883 95.702 -75.227 51.619
228884 95.731 -75.220 51.592
228885 95.768 -75.223 51.564
228886 95.809 -75.228 51.536
228887 95.850 -75.234 51.508
228888 95.890 -75.240 51.480
228889 95.925 -75.248 51.452
228890 95.961 -75.257 51.424
228891 95.997 -75.265 51.395
228892 96.033 -75.273 51.367
228893 96.068 -75.282 51.338
228894 96.108 -75.291 51.312
228895 96.149 -75.300 51.286
228896 96.189 -75.309 51.260
228897 96.229 -75.314 51.234
228898 96.269 -75.315 51.209
228899 96.304 -75.316 51.184
228900 96.335 -75.317 51.159
228901 96.366 -75.318 51.134
228902 96.397 -75.319 51.109
228903 96.421 -75.328 51.084
228904 96.441 -75.340 51.059
228905 96.461 -75.353 51.035
228906 96.481 -75.365 51.011
228907 96.502 -75.377 50.988
228908 96.522 -75.390 50.966
228909 96.541 -75.402 50.943
228910 96.561 -75.415 50.919
228911 96.581 -75.426 50.895
228912 96.600 -75.430 50.869
228913 96.619 -75.438 50.843
228914 96.640 -75.442 50.818
228915 96.661 -75.446 50.794
228916 96.682 -75.451 50.770
228917 96.701 -75.459 50.746
228918 96.711 -75.477 50.722
228919 96.721 -75.496 50.698
228920 96.731 -75.514 50.674
228921 96.741 -75.533 50.651
228922 96.751 -75.552 50.627
228923 96.760 -75.571 50.601
228924 96.774 -75.591 50.573
228925 96.788 -75.612 50.543
228926 96.801 -75.633 50.512
228927 96.823 -75.647 50.482
228928 96.848 -75.660 50.453
228929 96.872 -75.672 50.424
228930 96.937 -75.677 50.393
228931 97.001 -75.681 50.363
228932 97.066 -75.685 50.332
228933 97.115 -75.690 50.301
228934 97.139 -75.696 50.270
228935 97.163 -75.702 50.237
228936 97.187 -75.708 50.205
228937 97.212 -75.713 50.173
228938 97.236 -75.719 50.141
228939 97.260 -75.725 50.108
228940 97.284 -75.731 50.074
228941 97.305 -75.738 50.041
228942 97.327 -75.745 50.008
228943 97.348 -75.752 49.974
228944 97.370 -75.759 49.940
228945 97.391 -75.766 49.904
228946 97.416 -75.781 49.868
228947 97.441 -75.796 49.832
228948 97.465 -75.810 49.797
228949 97.490 -75.825 49.763
228950 97.515 -75.840 49.728
228951 97.543 -75.852 49.695
228952 97.585 -75.850 49.663
228953 97.616 -75.848 49.634
228954 97.648 -75.846 49.606
228955 97.679 -75.844 49.577
228956 97.710 -75.840 49.548
228957 97.741 -75.838 49.518
228958 97.767 -75.839 49.486
228959 97.776 -75.846 49.454
228960 97.779 -75.843 49.420
228961 97.782 -75.840 49.385
228962 97.793 -75.844 49.346
228963 97.803 -75.849 49.305
228964 97.816 -75.862 49.263
228965 97.835 -75.883 49.220
228966 97.853 -75.905 49.178
228967 97.872 -75.927 49.137
228968 97.890 -75.948 49.096
228969 97.908 -75.971 49.055
228970 97.926 -75.996 49.014
228971 97.946 -76.019 48.972
228972 97.990 -76.039 48.929
228973 98.036 -76.057 48.887
228974 98.082 -76.076 48.844
228975 98.128 -76.094 48.801
228976 98.174 -76.113 48.757
228977 98.220 -76.132 48.714
228978 98.266 -76.150 48.673
228979 98.311 -76.169 48.632
228980 98.352 -76.185 48.590
228981 98.392 -76.200 48.549
228982 98.418 -76.217 48.507
228983 98.430 -76.235 48.466
228984 98.442 -76.253 48.427
228985 98.453 -76.270 48.389
228986 98.465 -76.289 48.351
228987 98.477 -76.307 48.316
228988 98.489 -76.324 48.281
228989 98.500 -76.342 48.246
228990 98.519 -76.351 48.214
228991 98.551 -76.348 48.183
228992 98.583 -76.346 48.152
228993 98.615 -76.344 48.121
228994 98.647 -76.341 48.091
228995 98.678 -76.339 48.062
228996 98.710 -76.336 48.033
228997 98.742 -76.333 48.007
228998 98.770 -76.332 47.984
228999 98.795 -76.331 47.962
229000 98.821 -76.330 47.941
229001 98.847 -76.331 47.920
229002 98.873 -76.333 47.899
229003 98.899 -76.336 47.878
229004 98.925 -76.338 47.857
229005 98.951 -76.340 47.837
229006 98.977 -76.342 47.818
229007 99.003 -76.345 47.798
229008 99.029 -76.347 47.777
229009 99.055 -76.351 47.754
229010 99.079 -76.361 47.730
229011 99.103 -76.373 47.705
229012 99.128 -76.383 47.680
229013 99.152 -76.394 47.652
229014 99.175 -76.404 47.623
229015 99.183 -76.401 47.594
229016 99.191 -76.398 47.564
229017 99.200 -76.395 47.533
229018 99.208 -76.393 47.501
229019 99.216 -76.390 47.470
229020 99.224 -76.387 47.438
229021 99.239 -76.385 47.407
229022 99.261 -76.383 47.376
229023 99.290 -76.383 47.344
229024 99.321 -76.383 47.311
229025 99.350 -76.389 47.277
229026 99.379 -76.401 47.244
229027 99.409 -76.413 47.210
229028 99.438 -76.425 47.176
229029 99.469 -76.438 47.142
229030 99.500 -76.451 47.108
229031 99.531 -76.464 47.074
229032 99.561 -76.477 47.039
229033 99.592 -76.491 47.006
229034 99.626 -76.504 46.972
229035 99.661 -76.517 46.938
229036 99.696 -76.529 46.905
229037 99.731 -76.542 46.872
229038 99.767 -76.555 46.840
229039 99.802 -76.567 46.807
229040 99.837 -76.580 46.775
229041 99.872 -76.592 46.746
229042 99.908 -76.605 46.716
229043 99.942 -76.618 46.688
229044 99.967 -76.624 46.659
229045 99.981 -76.627 46.630
229046 99.994 -76.630 46.600
229047 100.008 -76.633 46.570
229048 100.021 -76.636 46.540
229049 100.035 -76.639 46.509
229050 100.049 -76.642 46.478
229051 100.062 -76.645 46.447
229052 100.076 -76.648 46.413
229053 100.089 -76.650 46.379
229054 100.103 -76.653 46.345
229055 100.117 -76.656 46.311
229056 100.130 -76.659 46.278
229057 100.143 -76.662 46.245
229058 100.157 -76.665 46.212
229059 100.174 -76.658 46.179
229060 100.194 -76.646 46.147
229061 100.213 -76.650 46.115
229062 100.232 -76.658 46.084
229063 100.252 -76.665 46.053
229064 100.272 -76.674 46.024
229065 100.291 -76.681 45.995
229066 100.311 -76.690 45.966
229067 100.328 -76.685 45.939
229068 100.345 -76.680 45.913
229069 100.363 -76.676 45.887
229070 100.380 -76.671 45.863
229071 100.398 -76.666 45.840
229072 100.415 -76.661 45.818
229073 100.432 -76.656 45.796
229074 100.451 -76.657 45.775
229075 100.472 -76.679 45.755
229076 100.493 -76.701 45.736
229077 100.514 -76.723 45.718
229078 100.536 -76.745 45.699
229079 100.557 -76.767 45.681
229080 100.578 -76.789 45.662
229081 100.600 -76.811 45.644
229082 100.621 -76.832 45.627
229083 100.645 -76.843 45.609
229084 100.669 -76.852 45.591
229085 100.694 -76.861 45.572
229086 100.722 -76.871 45.554
229087 100.750 -76.882 45.537
229088 100.778 -76.893 45.520
229089 100.806 -76.903 45.503
229090 100.834 -76.914 45.487
229091 100.862 -76.925 45.472
229092 100.876 -76.937 45.458
229093 100.891 -76.948 45.444
229094 100.917 -76.969 45.430
229095 100.946 -76.990 45.417
229096 100.974 -77.013 45.404
229097 101.003 -77.035 45.391
229098 101.032 -77.058 45.378
229099 101.061 -77.080 45.367
229100 101.090 -77.103 45.356
229101 101.119 -77.125 45.345
229102 101.148 -77.147 45.333
229103 101.177 -77.170 45.321
229104 101.205 -77.191 45.308
229105 101.234 -77.211 45.295
229106 101.263 -77.231 45.281
229107 101.277 -77.248 45.266
229108 101.291 -77.261 45.251
229109 101.307 -77.273 45.235
229110 101.321 -77.286 45.218
229111 101.336 -77.298 45.201
229112 101.351 -77.311 45.184
229113 101.366 -77.323 45.168
229114 101.389 -77.323 45.149
229115 101.411 -77.323 45.131
229116 101.430 -77.329 45.113
229117 101.449 -77.336 45.095
229118 101.467 -77.342 45.076
229119 101.486 -77.349 45.058
229120 101.504 -77.356 45.039
229121 101.523 -77.362 45.021
229122 101.541 -77.369 45.001
229123 101.560 -77.376 44.982
229124 101.579 -77.382 44.963
229125 101.594 -77.384 44.944
229126 101.610 -77.385 44.927
229127 101.625 -77.387 44.910
229128 101.640 -77.389 44.893
229129 101.656 -77.390 44.877
229130 101.671 -77.392 44.861
229131 101.687 -77.393 44.846
229132 101.689 -77.417 44.830
229133 101.708 -77.437 44.815
229134 101.729 -77.457 44.799
229135 101.749 -77.476 44.782
229136 101.768 -77.496 44.764
229137 101.788 -77.516 44.748
229138 101.809 -77.536 44.732
229139 101.829 -77.555 44.715
229140 101.848 -77.575 44.699
229141 101.877 -77.591 44.683
229142 101.913 -77.602 44.667
229143 101.945 -77.615 44.651
229144 101.992 -77.621 44.636
229145 102.049 -77.630 44.620
229146 102.089 -77.639 44.604
229147 102.109 -77.647 44.588
229148 102.137 -77.658 44.572
229149 102.169 -77.669 44.556
229150 102.202 -77.681 44.541
229151 102.234 -77.692 44.527
229152 102.267 -77.705 44.513
229153 102.299 -77.716 44.498
229154 102.320 -77.733 44.483
229155 102.335 -77.755 44.469
229156 102.349 -77.775 44.453
229157 102.365 -77.795 44.438
229158 102.390 -77.801 44.423
229159 102.416 -77.805 44.406
229160 102.443 -77.808 44.391
229161 102.469 -77.812 44.375
229162 102.495 -77.815 44.359
229163 102.522 -77.819 44.342
229164 102.548 -77.823 44.326
229165 102.574 -77.827 44.311
229166 102.600 -77.830 44.295
229167 102.626 -77.834 44.278
229168 102.653 -77.838 44.262
229169 102.675 -77.843 44.245
229170 102.693 -77.849 44.229
229171 102.724 -77.855 44.212
229172 102.760 -77.861 44.196
229173 102.798 -77.866 44.179
229174 102.834 -77.871 44.162
229175 102.872 -77.876 44.146
229176 102.908 -77.881 44.130
229177 102.945 -77.887 44.114
229178 102.982 -77.892 44.097
229179 103.019 -77.897 44.080
229180 103.056 -77.902 44.064
229181 103.093 -77.907 44.048
229182 103.130 -77.912 44.031
229183 103.167 -77.918 44.015
229184 103.204 -77.923 43.999
229185 103.241 -77.928 43.982
229430 92.101 -75.013 54.276
229431 92.125 -75.023 54.281
229432 92.149 -75.034 54.286
229433 92.173 -75.045 54.293
229434 92.198 -75.056 54.303
229435 92.222 -75.066 54.314
229436 92.246 -75.077 54.325
229437 92.271 -75.088 54.336
229438 92.295 -75.099 54.348
229439 92.320 -75.110 54.358
229440 92.344 -75.120 54.371
229441 92.368 -75.131 54.384
229442 92.393 -75.142 54.397
229443 92.419 -75.136 54.407
229444 92.451 -75.133 54.411
229445 92.482 -75.130 54.410
229446 92.514 -75.127 54.408
229447 92.545 -75.124 54.407
229448 92.577 -75.121 54.402
229449 92.608 -75.118 54.395
229450 92.641 -75.114 54.386
229451 92.673 -75.107 54.369
229452 92.706 -75.099 54.351
229453 92.738 -75.091 54.334
229454 92.769 -75.091 54.320
229455 92.798 -75.099 54.308
229456 92.827 -75.108 54.297
229457 92.856 -75.109 54.287
229458 92.886 -75.076 54.278
229459 92.912 -75.062 54.266
229460 92.936 -75.060 54.253
229461 92.959 -75.059 54.240
229462 92.982 -75.058 54.226
229463 93.005 -75.057 54.213
229464 93.029 -75.056 54.199
229465 93.052 -75.054 54.185
229466 93.075 -75.054 54.171
229467 93.098 -75.052 54.157
229468 93.122 -75.051 54.142
229469 93.145 -75.050 54.128
229470 93.168 -75.048 54.115
229471 93.192 -75.048 54.102
229472 93.215 -75.046 54.090
229473 93.238 -75.045 54.081
229474 93.262 -75.044 54.072
229475 93.285 -75.043 54.064
229476 93.308 -75.041 54.058
229477 93.332 -75.041 54.056
229478 93.358 -75.043 54.058
229479 93.383 -75.045 54.063
229480 93.410 -75.048 54.069
229481 93.435 -75.049 54.078
229482 93.459 -75.043 54.092
229483 93.482 -75.036 54.110
229484 93.504 -75.029 54.133
229485 93.528 -75.022 54.164
229486 93.555 -75.023 54.200
229487 93.584 -75.026 54.239
229488 93.613 -75.028 54.278
229489 93.642 -75.030 54.317
229490 93.673 -75.030 54.356
229491 93.705 -75.029 54.393
229492 93.737 -75.027 54.428
229493 93.769 -75.026 54.457
229494 93.800 -75.023 54.476
229495 93.826 -75.021 54.489
229496 93.851 -75.019 54.495
229497 93.877 -75.016 54.495
229498 93.903 -75.014 54.494
229499 93.928 -75.012 54.493
229500 93.954 -75.010 54.492
229501 93.980 -75.007 54.491
229502 94.006 -75.005 54.496
229503 94.032 -75.004 54.505
229504 94.058 -75.002 54.516
229505 94.084 -75.001 54.525
229506 94.110 -74.998 54.531
229507 94.139 -74.992 54.533
229508 94.167 -74.986 54.527
229509 94.192 -74.982 54.515
229510 94.217 -74.978 54.502
229511 94.243 -74.974 54.489
229512 94.268 -74.969 54.476
229513 94.287 -74.971 54.467
229514 94.304 -74.972 54.455
229515 94.330 -74.967 54.442
229516 94.355 -74.961 54.428
229517 94.381 -74.955 54.413
229518 94.406 -74.949 54.400
229519 94.431 -74.943 54.388
229520 94.457 -74.937 54.374
229521 94.482 -74.932 54.355
229522 94.504 -74.932 54.333
229523 94.517 -74.936 54.310
229524 94.528 -74.945 54.284
229525 94.538 -74.955 54.258
229526 94.548 -74.964 54.231
229527 94.559 -74.973 54.202
229528 94.569 -74.983 54.163
229529 94.579 -74.992 54.122
229530 94.590 -75.001 54.080
229531 94.611 -75.008 54.038
229532 94.632 -75.014 53.997
229533 94.653 -75.020 53.959
229534 94.674 -75.026 53.921
229535 94.695 -75.032 53.882
229536 94.716 -75.038 53.839
229537 94.736 -75.038 53.794
229538 94.756 -75.036 53.748
229539 94.776 -75.035 53.703
229540 94.799 -75.036 53.657
229541 94.802 -75.041 53.611
229542 94.801 -75.046 53.565
229543 94.819 -75.048 53.519
229544 94.837 -75.051 53.476
229545 94.853 -75.056 53.431
229546 94.869 -75.061 53.385
229547 94.884 -75.066 53.337
229548 94.900 -75.073 53.290
229549 94.915 -75.078 53.244
229550 94.931 -75.083 53.198
229551 94.946 -75.089 53.152
229552 94.961 -75.094 53.099
229553 94.955 -75.084 53.044
229554 94.949 -75.074 52.987
229555 94.981 -75.081 52.927
229556 95.013 -75.087 52.865
229557 95.045 -75.094 52.803
229558 95.072 -75.098 52.738
229559 95.097 -75.100 52.677
229560 95.122 -75.102 52.616
229561 95.148 -75.105 52.556
229562 95.173 -75.107 52.499
229563 95.198 -75.110 52.443
229564 95.224 -75.113 52.389
229565 95.251 -75.123 52.336
229566 95.278 -75.135 52.285
229567 95.296 -75.137 52.233
229568 95.311 -75.137 52.186
229569 95.325 -75.136 52.141
229570 95.341 -75.138 52.098
229571 95.365 -75.148 52.057
229572 95.389 -75.159 52.016
229573 95.413 -75.170 51.975
229574 95.438 -75.180 51.936
229575 95.462 -75.191 51.899
229576 95.486 -75.201 51.863
229577 95.510 -75.212 51.829
229578 95.534 -75.223 51.795
229579 95.561 -75.225 51.764
229580 95.589 -75.218 51.733
229581 95.618 -75.211 51.703
229582 95.647 -75.204 51.674
229583 95.675 -75.197 51.646
229584 95.703 -75.190 51.618
229585 95.743 -75.194 51.591
229586 95.783 -75.200 51.565
229587 95.824 -75.206 51.537
229588 95.865 -75.211 51.510
229589 95.904 -75.218 51.482
229590 95.940 -75.226 51.454
229591 95.976 -75.235 51.426
229592 96.012 -75.243 51.398
229593 96.047 -75.251 51.371
229594 96.083 -75.260 51.343
229595 96.118 -75.268 51.317
229596 96.157 -75.277 51.293
229597 96.197 -75.286 51.267
229598 96.237 -75.291 51.242
229599 96.274 -75.292 51.217
229600 96.306 -75.293 51.192
229601 96.337 -75.294 51.168
229602 96.369 -75.295 51.143
229603 96.400 -75.295 51.118
229604 96.428 -75.300 51.093
229605 96.448 -75.312 51.068
229606 96.468 -75.324 51.043
229607 96.488 -75.337 51.018
229608 96.509 -75.349 50.995
229609 96.529 -75.362 50.973
229610 96.549 -75.374 50.950
229611 96.569 -75.386 50.926
229612 96.589 -75.399 50.902
229613 96.608 -75.408 50.876
229614 96.628 -75.413 50.851
229615 96.650 -75.417 50.826
229616 96.671 -75.421 50.802
229617 96.692 -75.426 50.779
229618 96.711 -75.433 50.755
229619 96.721 -75.452 50.731
229620 96.731 -75.471 50.707
229621 96.741 -75.489 50.684
229622 96.751 -75.508 50.660
229623 96.762 -75.527 50.635
229624 96.776 -75.549 50.609
229625 96.789 -75.569 50.580
229626 96.803 -75.590 50.550
229627 96.820 -75.608 50.520
229628 96.844 -75.620 50.491
229629 96.869 -75.633 50.462
229630 96.892 -75.646 50.433
229631 96.947 -75.652 50.403
229632 97.011 -75.656 50.373
229633 97.076 -75.661 50.342
229634 97.124 -75.665 50.312
229635 97.149 -75.671 50.282
229636 97.173 -75.677 50.250
229637 97.197 -75.683 50.218
229638 97.222 -75.689 50.186
229639 97.246 -75.695 50.154
229640 97.271 -75.701 50.120
229641 97.295 -75.706 50.087
229642 97.319 -75.712 50.053
229643 97.343 -75.718 50.020
229644 97.365 -75.725 49.986
229645 97.386 -75.732 49.951
229646 97.409 -75.743 49.915
229647 97.434 -75.757 49.879
229648 97.459 -75.772 49.843
229649 97.484 -75.787 49.808
229650 97.508 -75.802 49.773
229651 97.533 -75.816 49.739
229652 97.558 -75.831 49.706
229653 97.589 -75.829 49.673
229654 97.620 -75.826 49.643
229655 97.651 -75.824 49.614
229656 97.682 -75.822 49.583
229657 97.714 -75.818 49.552
229658 97.745 -75.816 49.520
229659 97.776 -75.814 49.487
229660 97.786 -75.816 49.453
229661 97.797 -75.820 49.418
229662 97.808 -75.825 49.382
229663 97.819 -75.829 49.342
229664 97.829 -75.837 49.301
229665 97.840 -75.844 49.259
229666 97.858 -75.865 49.216
229667 97.877 -75.887 49.174
229668 97.895 -75.909 49.133
229669 97.914 -75.931 49.093
229670 97.932 -75.953 49.051
229671 97.950 -75.974 49.010
229672 97.968 -75.991 48.968
229673 98.000 -76.008 48.925
229674 98.045 -76.026 48.883
229675 98.092 -76.044 48.840
229676 98.138 -76.063 48.796
229677 98.183 -76.082 48.752
229678 98.229 -76.100 48.709
229679 98.275 -76.119 48.668
229680 98.320 -76.137 48.626
229681 98.360 -76.153 48.584
229682 98.400 -76.168 48.543
229683 98.434 -76.185 48.501
229684 98.446 -76.202 48.460
229685 98.457 -76.220 48.420
229686 98.469 -76.238 48.382
229687 98.481 -76.256 48.345
229688 98.493 -76.274 48.309
229689 98.504 -76.292 48.273
229690 98.516 -76.310 48.238
229691 98.549 -76.329 48.205
229692 98.580 -76.326 48.173
229693 98.612 -76.324 48.142
229694 98.644 -76.322 48.111
229695 98.676 -76.319 48.081
229696 98.708 -76.317 48.051
229697 98.739 -76.314 48.022
229698 98.767 -76.313 47.996
229699 98.793 -76.311 47.972
229700 98.818 -76.310 47.950
229701 98.844 -76.310 47.929
229702 98.870 -76.309 47.909
229703 98.896 -76.311 47.889
229704 98.922 -76.314 47.869
229705 98.948 -76.316 47.849
229706 98.974 -76.318 47.830
229707 99.000 -76.320 47.809
229708 99.026 -76.323 47.789
229709 99.052 -76.325 47.768
229710 99.077 -76.332 47.746
229711 99.101 -76.344 47.722
229712 99.125 -76.354 47.697
229713 99.149 -76.366 47.672
229714 99.173 -76.376 47.645
229715 99.198 -76.387 47.615
229716 99.210 -76.388 47.586
229717 99.218 -76.385 47.555
229718 99.226 -76.383 47.525
229719 99.235 -76.379 47.494
229720 99.243 -76.376 47.462
229721 99.251 -76.373 47.431
229722 99.259 -76.371 47.401
229723 99.272 -76.369 47.369
229724 99.302 -76.368 47.337
229725 99.332 -76.368 47.305
229726 99.362 -76.368 47.272
229727 99.392 -76.379 47.239
229728 99.421 -76.391 47.206
229729 99.451 -76.403 47.172
229730 99.482 -76.417 47.140
229731 99.512 -76.430 47.106
229732 99.544 -76.443 47.073
229733 99.574 -76.456 47.040
229734 99.605 -76.469 47.007
229735 99.639 -76.482 46.974
229736 99.675 -76.495 46.940
229737 99.710 -76.508 46.907
229738 99.745 -76.520 46.875
229739 99.780 -76.533 46.843
229740 99.816 -76.546 46.810
229741 99.851 -76.558 46.779
229742 99.886 -76.571 46.749
229743 99.920 -76.585 46.719
229744 99.952 -76.602 46.691
229745 99.972 -76.610 46.662
229746 99.990 -76.624 46.633
229747 100.009 -76.636 46.603
229748 100.023 -76.641 46.573
229749 100.037 -76.644 46.542
229750 100.050 -76.647 46.510
229751 100.064 -76.650 46.479
229752 100.077 -76.653 46.447
229753 100.091 -76.656 46.413
229754 100.105 -76.659 46.379
229755 100.118 -76.662 46.344
229756 100.131 -76.665 46.310
229757 100.145 -76.668 46.277
229758 100.158 -76.671 46.244
229759 100.173 -76.671 46.210
229760 100.192 -76.658 46.177
229761 100.211 -76.646 46.144
229762 100.230 -76.641 46.112
229763 100.250 -76.649 46.081
229764 100.270 -76.657 46.051
229765 100.289 -76.665 46.022
229766 100.309 -76.673 45.993
229767 100.328 -76.680 45.964
229768 100.348 -76.688 45.936
229769 100.367 -76.696 45.910
229770 100.387 -76.704 45.885
229771 100.405 -76.707 45.861
229772 100.423 -76.702 45.837
229773 100.440 -76.697 45.815
229774 100.458 -76.693 45.793
229775 100.475 -76.687 45.772
229776 100.495 -76.703 45.751
229777 100.516 -76.725 45.732
229778 100.538 -76.747 45.713
229779 100.559 -76.769 45.695
229780 100.581 -76.791 45.676
229781 100.604 -76.802 45.658
229782 100.628 -76.812 45.640
229783 100.652 -76.821 45.622
229784 100.676 -76.830 45.604
229785 100.700 -76.840 45.585
229786 100.725 -76.849 45.567
229787 100.753 -76.859 45.549
229788 100.781 -76.870 45.530
229789 100.809 -76.881 45.513
229790 100.837 -76.891 45.497
229791 100.865 -76.902 45.480
229792 100.890 -76.912 45.464
229793 100.905 -76.925 45.449
229794 100.920 -76.937 45.435
229795 100.934 -76.949 45.422
229796 100.956 -76.965 45.408
229797 100.984 -76.987 45.395
229798 101.013 -77.010 45.382
229799 101.042 -77.032 45.370
229800 101.071 -77.055 45.359
229801 101.100 -77.078 45.347
229802 101.129 -77.100 45.336
229803 101.158 -77.122 45.324
229804 101.187 -77.144 45.311
229805 101.215 -77.167 45.298
229806 101.244 -77.189 45.285
229807 101.273 -77.212 45.271
229808 101.291 -77.226 45.256
229809 101.305 -77.238 45.241
229810 101.321 -77.251 45.226
229811 101.335 -77.263 45.208
229812 101.350 -77.276 45.191
229813 101.365 -77.288 45.175
229814 101.382 -77.298 45.158
229815 101.405 -77.298 45.140
229816 101.428 -77.298 45.122
229817 101.447 -77.303 45.105
229818 101.466 -77.310 45.087
229819 101.484 -77.316 45.068
229820 101.503 -77.323 45.050
229821 101.522 -77.329 45.032
229822 101.540 -77.335 45.013
229823 101.558 -77.342 44.993
229824 101.577 -77.348 44.974
229825 101.596 -77.355 44.955
229826 101.614 -77.362 44.937
229827 101.633 -77.368 44.920
229828 101.649 -77.371 44.903
229829 101.664 -77.373 44.886
229830 101.680 -77.374 44.869
229831 101.695 -77.376 44.854
229832 101.710 -77.377 44.838
229833 101.716 -77.394 44.823
229834 101.736 -77.414 44.807
229835 101.756 -77.433 44.791
229836 101.776 -77.453 44.775
229837 101.796 -77.473 44.757
229838 101.816 -77.492 44.741
229839 101.836 -77.512 44.725
229840 101.856 -77.532 44.708
229841 101.876 -77.551 44.692
229842 101.896 -77.571 44.676
229843 101.917 -77.591 44.659
229844 101.965 -77.600 44.643
229845 102.022 -77.608 44.628
229846 102.079 -77.617 44.612
229847 102.102 -77.625 44.596
229848 102.122 -77.634 44.581
229849 102.144 -77.643 44.564
229850 102.177 -77.655 44.549
229851 102.209 -77.667 44.534
229852 102.242 -77.678 44.520
229853 102.274 -77.690 44.505
229854 102.307 -77.702 44.491
229855 102.327 -77.720 44.476
229856 102.342 -77.740 44.461
229857 102.357 -77.761 44.445
229858 102.373 -77.779 44.430
229859 102.399 -77.783 44.414
229860 102.424 -77.787 44.398
229861 102.450 -77.791 44.382
229862 102.475 -77.796 44.366
229863 102.501 -77.799 44.350
229864 102.527 -77.804 44.333
229865 102.552 -77.808 44.318
229866 102.578 -77.812 44.302
229867 102.603 -77.816 44.285
229868 102.629 -77.821 44.268
229869 102.655 -77.824 44.252
229870 102.680 -77.829 44.235
229871 102.709 -77.834 44.218
229872 102.745 -77.839 44.202
229873 102.782 -77.844 44.186
229874 102.819 -77.849 44.169
229875 102.856 -77.854 44.152
229876 102.893 -77.859 44.136
229877 102.930 -77.865 44.120
229878 102.967 -77.870 44.104
229879 103.004 -77.875 44.087
229880 103.041 -77.880 44.070
229881 103.078 -77.885 44.053
229882 103.115 -77.890 44.037
229883 103.152 -77.896 44.021
229884 103.189 -77.901 44.005
229885 103.226 -77.906 43.989
230132 92.127 -74.991 54.284
230133 92.151 -75.002 54.288
230134 92.175 -75.013 54.295
230135 92.200 -75.023 54.305
230136 92.224 -75.034 54.316
230137 92.248 -75.045 54.327
230138 92.273 -75.056 54.339
230139 92.297 -75.066 54.350
230140 92.321 -75.077 54.361
230141 92.346 -75.088 54.374
230142 92.370 -75.099 54.387
230143 92.394 -75.110 54.401
230144 92.416 -75.104 54.413
230145 92.447 -75.101 54.420
230146 92.479 -75.097 54.424
230147 92.510 -75.094 54.425
230148 92.542 -75.091 54.424
230149 92.573 -75.088 54.422
230150 92.606 -75.084 54.416
230151 92.641 -75.080 54.409
230152 92.674 -75.073 54.392
230153 92.706 -75.066 54.372
230154 92.736 -75.068 54.352
230155 92.765 -75.076 54.337
230156 92.794 -75.085 54.324
230157 92.823 -75.095 54.313
230158 92.852 -75.096 54.303
230159 92.882 -75.063 54.294
230160 92.913 -75.029 54.282
230161 92.936 -75.028 54.269
230162 92.959 -75.027 54.256
230163 92.983 -75.026 54.243
230164 93.006 -75.025 54.230
230165 93.029 -75.023 54.217
230166 93.052 -75.022 54.202
230167 93.076 -75.021 54.189
230168 93.099 -75.019 54.175
230169 93.122 -75.019 54.160
230170 93.145 -75.017 54.147
230171 93.169 -75.016 54.134
230172 93.192 -75.015 54.121
230173 93.215 -75.014 54.110
230174 93.239 -75.013 54.099
230175 93.265 -75.016 54.091
230176 93.291 -75.017 54.084
230177 93.317 -75.019 54.079
230178 93.343 -75.022 54.076
230179 93.368 -75.023 54.078
230180 93.394 -75.026 54.083
230181 93.418 -75.020 54.089
230182 93.441 -75.013 54.097
230183 93.464 -75.007 54.111
230184 93.487 -74.999 54.129
230185 93.509 -74.992 54.151
230186 93.533 -74.985 54.180
230187 93.560 -74.986 54.215
230188 93.589 -74.988 54.252
230189 93.618 -74.991 54.289
230190 93.647 -74.993 54.326
230191 93.676 -74.995 54.363
230192 93.704 -74.997 54.400
230193 93.735 -74.997 54.435
230194 93.768 -74.996 54.465
230195 93.797 -74.994 54.482
230196 93.823 -74.991 54.493
230197 93.849 -74.989 54.499
230198 93.874 -74.987 54.499
230199 93.901 -74.985 54.498
230200 93.926 -74.983 54.498
230201 93.953 -74.982 54.499
230202 93.979 -74.980 54.502
230203 94.004 -74.979 54.507
230204 94.031 -74.977 54.514
230205 94.057 -74.975 54.525
230206 94.082 -74.974 54.533
230207 94.110 -74.969 54.539
230208 94.139 -74.963 54.540
230209 94.167 -74.957 54.534
230210 94.195 -74.951 54.522
230211 94.220 -74.947 54.509
230212 94.246 -74.943 54.494
230213 94.270 -74.939 54.477
230214 94.286 -74.943 54.462
230215 94.304 -74.944 54.446
230216 94.329 -74.938 54.431
230217 94.355 -74.932 54.416
230218 94.380 -74.926 54.404
230219 94.406 -74.920 54.393
230220 94.431 -74.914 54.383
230221 94.457 -74.908 54.372
230222 94.482 -74.903 54.355
230223 94.498 -74.905 54.335
230224 94.509 -74.914 54.313
230225 94.520 -74.923 54.290
230226 94.530 -74.932 54.265
230227 94.541 -74.942 54.240
230228 94.551 -74.951 54.212
230229 94.562 -74.960 54.174
230230 94.572 -74.970 54.135
230231 94.588 -74.978 54.095
230232 94.609 -74.984 54.055
230233 94.630 -74.990 54.018
230234 94.651 -74.996 53.983
230235 94.673 -75.003 53.947
230236 94.694 -75.007 53.908
230237 94.713 -75.010 53.867
230238 94.734 -75.011 53.823
230239 94.755 -75.013 53.778
230240 94.775 -75.014 53.733
230241 94.795 -75.013 53.687
230242 94.812 -75.015 53.641
230243 94.820 -75.019 53.595
230244 94.836 -75.022 53.548
230245 94.852 -75.029 53.502
230246 94.867 -75.034 53.456
230247 94.883 -75.039 53.409
230248 94.898 -75.044 53.362
230249 94.914 -75.051 53.313
230250 94.929 -75.056 53.264
230251 94.945 -75.061 53.217
230252 94.960 -75.066 53.170
230253 94.965 -75.065 53.119
230254 94.959 -75.055 53.064
230255 94.955 -75.046 53.005
230256 94.987 -75.052 52.943
230257 95.019 -75.059 52.880
230258 95.051 -75.066 52.816
230259 95.082 -75.072 52.752
230260 95.107 -75.074 52.690
230261 95.132 -75.076 52.629
230262 95.158 -75.079 52.569
230263 95.183 -75.082 52.509
230264 95.208 -75.084 52.452
230265 95.234 -75.089 52.397
230266 95.261 -75.100 52.343
230267 95.289 -75.111 52.290
230268 95.313 -75.119 52.239
230269 95.327 -75.119 52.191
230270 95.341 -75.118 52.145
230271 95.355 -75.117 52.101
230272 95.370 -75.117 52.059
230273 95.390 -75.123 52.018
230274 95.414 -75.134 51.977
230275 95.439 -75.145 51.938
230276 95.463 -75.155 51.900
230277 95.487 -75.166 51.864
230278 95.511 -75.176 51.828
230279 95.535 -75.187 51.794
230280 95.562 -75.188 51.762
230281 95.591 -75.181 51.731
230282 95.619 -75.174 51.701
230283 95.648 -75.167 51.672
230284 95.676 -75.160 51.645
230285 95.717 -75.166 51.618
230286 95.758 -75.172 51.590
230287 95.799 -75.177 51.565
230288 95.839 -75.183 51.539
230289 95.881 -75.189 51.512
230290 95.918 -75.196 51.484
230291 95.954 -75.204 51.456
230292 95.990 -75.213 51.428
230293 96.026 -75.221 51.401
230294 96.061 -75.229 51.375
230295 96.097 -75.238 51.348
230296 96.133 -75.246 51.323
230297 96.169 -75.254 51.299
230298 96.206 -75.263 51.274
230299 96.245 -75.268 51.250
230300 96.276 -75.269 51.226
230301 96.308 -75.270 51.201
230302 96.339 -75.270 51.177
230303 96.371 -75.272 51.153
230304 96.402 -75.273 51.128
230305 96.433 -75.273 51.103
230306 96.455 -75.284 51.077
230307 96.475 -75.296 51.051
230308 96.496 -75.309 51.026
230309 96.515 -75.321 51.002
230310 96.535 -75.334 50.980
230311 96.555 -75.346 50.957
230312 96.576 -75.358 50.933
230313 96.596 -75.371 50.908
230314 96.616 -75.383 50.883
230315 96.638 -75.388 50.858
230316 96.659 -75.392 50.834
230317 96.680 -75.396 50.811
230318 96.701 -75.401 50.788
230319 96.721 -75.408 50.765
230320 96.731 -75.427 50.741
230321 96.741 -75.445 50.718
230322 96.751 -75.464 50.694
230323 96.764 -75.484 50.669
230324 96.778 -75.505 50.644
230325 96.791 -75.526 50.618
230326 96.804 -75.547 50.588
230327 96.818 -75.568 50.558
230328 96.842 -75.580 50.529
230329 96.866 -75.593 50.501
230330 96.890 -75.606 50.471
230331 96.914 -75.619 50.443
230332 96.957 -75.628 50.414
230333 97.021 -75.632 50.384
230334 97.086 -75.637 50.354
230335 97.135 -75.641 50.324
230336 97.159 -75.647 50.294
230337 97.183 -75.653 50.263
230338 97.208 -75.659 50.231
230339 97.232 -75.665 50.198
230340 97.256 -75.671 50.166
230341 97.281 -75.676 50.133
230342 97.305 -75.682 50.099
230343 97.329 -75.688 50.065
230344 97.353 -75.694 50.031
230345 97.378 -75.699 49.996
230346 97.402 -75.706 49.961
230347 97.425 -75.720 49.925
230348 97.447 -75.735 49.889
230349 97.470 -75.750 49.854
230350 97.493 -75.766 49.819
230351 97.516 -75.781 49.784
230352 97.538 -75.796 49.750
230353 97.569 -75.803 49.716
230354 97.600 -75.805 49.683
230355 97.632 -75.804 49.652
230356 97.663 -75.802 49.622
230357 97.693 -75.800 49.590
230358 97.725 -75.796 49.557
230359 97.755 -75.794 49.524
230360 97.787 -75.792 49.489
230361 97.813 -75.801 49.452
230362 97.824 -75.806 49.416
230363 97.835 -75.810 49.379
230364 97.846 -75.815 49.338
230365 97.854 -75.825 49.296
230366 97.865 -75.834 49.255
230367 97.882 -75.848 49.213
230368 97.900 -75.870 49.171
230369 97.919 -75.891 49.130
230370 97.937 -75.913 49.089
230371 97.955 -75.934 49.047
230372 97.973 -75.949 49.005
230373 97.992 -75.963 48.963
230374 98.010 -75.977 48.920
230375 98.056 -75.995 48.877
230376 98.101 -76.013 48.835
230377 98.147 -76.032 48.791
230378 98.193 -76.050 48.747
230379 98.239 -76.069 48.703
230380 98.285 -76.088 48.662
230381 98.329 -76.105 48.620
230382 98.368 -76.120 48.579
230383 98.409 -76.136 48.537
230384 98.448 -76.152 48.496
230385 98.461 -76.170 48.454
230386 98.473 -76.188 48.414
230387 98.485 -76.206 48.376
230388 98.497 -76.223 48.339
230389 98.508 -76.242 48.302
230390 98.520 -76.260 48.266
230391 98.544 -76.294 48.231
230392 98.578 -76.307 48.198
230393 98.609 -76.304 48.165
230394 98.642 -76.302 48.133
230395 98.673 -76.300 48.102
230396 98.705 -76.297 48.072
230397 98.737 -76.295 48.042
230398 98.765 -76.293 48.013
230399 98.790 -76.292 47.985
230400 98.816 -76.291 47.961
230401 98.842 -76.290 47.937
230402 98.867 -76.289 47.915
230403 98.893 -76.288 47.895
230404 98.919 -76.289 47.877
230405 98.945 -76.292 47.858
230406 98.971 -76.294 47.840
230407 98.997 -76.296 47.821
230408 99.023 -76.298 47.800
230409 99.049 -76.301 47.780
230410 99.075 -76.304 47.759
230411 99.099 -76.315 47.736
230412 99.123 -76.326 47.713
230413 99.147 -76.337 47.688
230414 99.171 -76.348 47.663
230415 99.196 -76.359 47.635
230416 99.220 -76.370 47.607
230417 99.237 -76.375 47.578
230418 99.245 -76.372 47.549
230419 99.254 -76.369 47.518
230420 99.262 -76.366 47.486
230421 99.270 -76.363 47.455
230422 99.278 -76.361 47.424
230423 99.286 -76.357 47.394
230424 99.303 -76.356 47.363
230425 99.325 -76.354 47.332
230426 99.346 -76.354 47.300
230427 99.373 -76.354 47.267
230428 99.403 -76.356 47.235
230429 99.433 -76.369 47.202
230430 99.464 -76.382 47.168
230431 99.495 -76.395 47.136
230432 99.525 -76.408 47.104
230433 99.556 -76.421 47.072
230434 99.587 -76.435 47.041
230435 99.618 -76.448 47.008
230436 99.653 -76.461 46.975
230437 99.688 -76.473 46.942
230438 99.723 -76.486 46.909
230439 99.759 -76.499 46.877
230440 99.794 -76.511 46.845
230441 99.829 -76.524 46.814
230442 99.863 -76.539 46.783
230443 99.895 -76.555 46.752
230444 99.927 -76.571 46.723
230445 99.954 -76.583 46.694
230446 99.973 -76.589 46.665
230447 99.992 -76.602 46.636
230448 100.010 -76.615 46.605
230449 100.028 -76.629 46.574
230450 100.047 -76.642 46.542
230451 100.065 -76.655 46.510
230452 100.079 -76.658 46.478
230453 100.092 -76.661 46.446
230454 100.106 -76.665 46.413
230455 100.119 -76.668 46.378
230456 100.133 -76.671 46.343
230457 100.146 -76.674 46.309
230458 100.160 -76.677 46.275
230459 100.174 -76.680 46.242
230460 100.191 -76.671 46.208
230461 100.210 -76.659 46.174
230462 100.229 -76.647 46.142
230463 100.248 -76.634 46.110
230464 100.267 -76.640 46.078
230465 100.287 -76.648 46.048
230466 100.307 -76.656 46.018
230467 100.326 -76.664 45.989
230468 100.346 -76.671 45.960
230469 100.365 -76.680 45.933
230470 100.385 -76.687 45.907
230471 100.404 -76.696 45.882
230472 100.424 -76.703 45.858
230473 100.443 -76.711 45.834
230474 100.463 -76.719 45.811
230475 100.482 -76.727 45.790
230476 100.500 -76.724 45.769
230477 100.519 -76.728 45.748
230478 100.540 -76.750 45.728
230479 100.563 -76.763 45.709
230480 100.587 -76.772 45.691
230481 100.611 -76.781 45.672
230482 100.635 -76.790 45.653
230483 100.659 -76.799 45.635
230484 100.683 -76.809 45.616
230485 100.707 -76.818 45.598
230486 100.731 -76.828 45.579
230487 100.756 -76.837 45.560
230488 100.784 -76.847 45.542
230489 100.812 -76.858 45.523
230490 100.840 -76.868 45.506
230491 100.868 -76.879 45.489
230492 100.896 -76.890 45.472
230493 100.918 -76.901 45.456
230494 100.933 -76.912 45.441
230495 100.948 -76.925 45.427
230496 100.963 -76.937 45.414
230497 100.977 -76.949 45.400
230498 100.994 -76.962 45.387
230499 101.023 -76.984 45.374
230500 101.052 -77.007 45.362
230501 101.081 -77.030 45.350
230502 101.110 -77.052 45.338
230503 101.139 -77.075 45.326
230504 101.168 -77.097 45.315
230505 101.196 -77.119 45.303
230506 101.226 -77.141 45.290
230507 101.254 -77.164 45.276
230508 101.283 -77.186 45.261
230509 101.305 -77.204 45.246
230510 101.320 -77.216 45.231
230511 101.335 -77.229 45.215
230512 101.349 -77.241 45.198
230513 101.364 -77.254 45.182
230514 101.379 -77.266 45.166
230515 101.398 -77.272 45.148
230516 101.421 -77.273 45.131
230517 101.444 -77.273 45.114
230518 101.464 -77.276 45.096
230519 101.483 -77.282 45.078
230520 101.501 -77.289 45.059
230521 101.520 -77.296 45.041
230522 101.539 -77.302 45.023
230523 101.557 -77.309 45.004
230524 101.575 -77.315 44.984
230525 101.594 -77.322 44.966
230526 101.613 -77.329 44.948
230527 101.631 -77.335 44.930
230528 101.650 -77.341 44.914
230529 101.668 -77.348 44.897
230530 101.687 -77.354 44.880
230531 101.704 -77.358 44.863
230532 101.719 -77.360 44.847
230533 101.735 -77.362 44.831
230534 101.746 -77.367 44.816
230535 101.763 -77.390 44.800
230536 101.783 -77.410 44.784
230537 101.803 -77.429 44.767
230538 101.823 -77.449 44.750
230539 101.843 -77.469 44.735
230540 101.863 -77.489 44.718
230541 101.883 -77.508 44.702
230542 101.903 -77.531 44.687
230543 101.922 -77.554 44.669
230544 101.942 -77.577 44.652
230545 101.994 -77.586 44.635
230546 102.051 -77.595 44.620
230547 102.094 -77.603 44.604
230548 102.114 -77.612 44.588
230549 102.134 -77.620 44.572
230550 102.155 -77.630 44.557
230551 102.185 -77.641 44.541
230552 102.217 -77.652 44.527
230553 102.249 -77.664 44.512
230554 102.282 -77.676 44.498
230555 102.314 -77.687 44.484
230556 102.334 -77.705 44.469
230557 102.349 -77.727 44.454
230558 102.364 -77.747 44.437
230559 102.384 -77.760 44.421
230560 102.409 -77.764 44.406
230561 102.435 -77.768 44.390
230562 102.460 -77.772 44.373
230563 102.486 -77.777 44.357
230564 102.512 -77.780 44.341
230565 102.537 -77.785 44.325
230566 102.562 -77.789 44.308
230567 102.588 -77.793 44.292
230568 102.614 -77.797 44.276
230569 102.640 -77.802 44.259
230570 102.665 -77.805 44.242
230571 102.697 -77.811 44.225
230572 102.733 -77.816 44.208
230573 102.769 -77.821 44.192
230574 102.806 -77.827 44.176
230575 102.842 -77.832 44.159
230576 102.878 -77.837 44.142
230577 102.914 -77.843 44.126
230578 102.952 -77.848 44.110
230579 102.988 -77.853 44.093
230580 103.026 -77.858 44.077
230581 103.063 -77.863 44.060
230582 103.099 -77.868 44.044
230583 103.137 -77.874 44.028
230584 103.173 -77.879 44.012
230585 103.211 -77.884 43.996
230834 92.153 -74.969 54.291
230835 92.177 -74.980 54.298
230836 92.202 -74.991 54.307
230837 92.226 -75.002 54.318
230838 92.250 -75.013 54.329
230839 92.275 -75.023 54.341
230840 92.299 -75.034 54.352
230841 92.323 -75.044 54.364
230842 92.348 -75.056 54.377
230843 92.372 -75.066 54.391
230844 92.397 -75.077 54.405
230845 92.417 -75.076 54.417
230846 92.443 -75.067 54.427
230847 92.475 -75.064 54.434
230848 92.506 -75.061 54.437
230849 92.537 -75.058 54.438
230850 92.571 -75.054 54.437
230851 92.605 -75.050 54.433
230852 92.640 -75.046 54.427
230853 92.674 -75.041 54.412
230854 92.704 -75.044 54.392
230855 92.733 -75.054 54.370
230856 92.762 -75.063 54.355
230857 92.790 -75.072 54.342
230858 92.819 -75.081 54.331
230859 92.848 -75.082 54.321
230860 92.878 -75.049 54.311
230861 92.909 -75.016 54.299
230862 92.936 -74.995 54.287
230863 92.959 -74.994 54.274
230864 92.983 -74.993 54.261
230865 93.006 -74.992 54.248
230866 93.029 -74.991 54.234
230867 93.053 -74.990 54.220
230868 93.076 -74.988 54.207
230869 93.099 -74.988 54.193
230870 93.123 -74.986 54.179
230871 93.146 -74.985 54.165
230872 93.172 -74.988 54.153
230873 93.198 -74.990 54.141
230874 93.224 -74.991 54.130
230875 93.250 -74.994 54.119
230876 93.276 -74.996 54.111
230877 93.301 -74.997 54.105
230878 93.327 -75.000 54.100
230879 93.353 -75.002 54.098
230880 93.377 -74.997 54.099
230881 93.400 -74.990 54.103
230882 93.423 -74.983 54.110
230883 93.445 -74.976 54.118
230884 93.469 -74.969 54.131
230885 93.492 -74.963 54.148
230886 93.514 -74.956 54.170
230887 93.538 -74.948 54.198
230888 93.565 -74.949 54.234
230889 93.594 -74.951 54.270
230890 93.623 -74.954 54.305
230891 93.652 -74.956 54.338
230892 93.681 -74.958 54.373
230893 93.709 -74.960 54.408
230894 93.738 -74.963 54.443
230895 93.767 -74.965 54.471
230896 93.795 -74.964 54.488
230897 93.821 -74.963 54.498
230898 93.847 -74.961 54.503
230899 93.873 -74.960 54.502
230900 93.899 -74.958 54.502
230901 93.925 -74.957 54.504
230902 93.952 -74.955 54.509
230903 93.977 -74.954 54.513
230904 94.003 -74.952 54.519
230905 94.030 -74.950 54.526
230906 94.055 -74.949 54.533
230907 94.082 -74.947 54.541
230908 94.110 -74.941 54.545
230909 94.139 -74.935 54.545
230910 94.168 -74.928 54.539
230911 94.196 -74.922 54.528
230912 94.223 -74.916 54.514
230913 94.249 -74.912 54.498
230914 94.269 -74.912 54.477
230915 94.285 -74.916 54.458
230916 94.303 -74.916 54.438
230917 94.329 -74.910 54.419
230918 94.354 -74.904 54.405
230919 94.380 -74.898 54.395
230920 94.405 -74.892 54.385
230921 94.431 -74.886 54.377
230922 94.454 -74.884 54.368
230923 94.474 -74.883 54.353
230924 94.487 -74.888 54.334
230925 94.500 -74.894 54.313
230926 94.512 -74.901 54.291
230927 94.523 -74.910 54.268
230928 94.533 -74.919 54.244
230929 94.544 -74.929 54.219
230930 94.554 -74.938 54.185
230931 94.565 -74.947 54.148
230932 94.586 -74.954 54.110
230933 94.608 -74.960 54.073
230934 94.629 -74.966 54.041
230935 94.650 -74.971 54.008
230936 94.670 -74.972 53.973
230937 94.690 -74.974 53.935
230938 94.710 -74.976 53.895
230939 94.731 -74.979 53.852
230940 94.751 -74.981 53.807
230941 94.772 -74.982 53.762
230942 94.792 -74.985 53.717
230943 94.815 -74.988 53.671
230944 94.834 -74.995 53.625
230945 94.850 -75.001 53.576
230946 94.865 -75.006 53.528
230947 94.881 -75.012 53.481
230948 94.897 -75.017 53.435
230949 94.912 -75.022 53.387
230950 94.927 -75.029 53.336
230951 94.943 -75.034 53.286
230952 94.959 -75.039 53.238
230953 94.974 -75.044 53.189
230954 94.970 -75.036 53.137
230955 94.964 -75.026 53.082
230956 94.961 -75.017 53.023
230957 94.992 -75.024 52.960
230958 95.025 -75.031 52.896
230959 95.056 -75.038 52.832
230960 95.089 -75.044 52.767
230961 95.117 -75.048 52.704
230962 95.142 -75.051 52.642
230963 95.167 -75.054 52.581
230964 95.193 -75.056 52.520
230965 95.218 -75.058 52.462
230966 95.244 -75.066 52.405
230967 95.272 -75.076 52.351
230968 95.299 -75.088 52.298
230969 95.327 -75.099 52.247
230970 95.343 -75.100 52.198
230971 95.357 -75.100 52.151
230972 95.372 -75.099 52.106
230973 95.386 -75.098 52.063
230974 95.400 -75.098 52.021
230975 95.415 -75.098 51.980
230976 95.440 -75.109 51.940
230977 95.464 -75.120 51.902
230978 95.488 -75.130 51.865
230979 95.512 -75.141 51.829
230980 95.536 -75.151 51.795
230981 95.563 -75.151 51.762
230982 95.592 -75.144 51.730
230983 95.620 -75.137 51.700
230984 95.651 -75.132 51.672
230985 95.691 -75.138 51.645
230986 95.732 -75.143 51.618
230987 95.773 -75.149 51.591
230988 95.814 -75.155 51.566
230989 95.855 -75.160 51.540
230990 95.896 -75.167 51.513
230991 95.933 -75.174 51.486
230992 95.969 -75.182 51.459
230993 96.005 -75.191 51.431
230994 96.040 -75.199 51.405
230995 96.076 -75.207 51.380
230996 96.112 -75.216 51.354
230997 96.147 -75.224 51.329
230998 96.183 -75.232 51.306
230999 96.219 -75.241 51.283
231000 96.249 -75.244 51.258
231001 96.279 -75.245 51.234
231002 96.310 -75.247 51.210
231003 96.341 -75.248 51.187
231004 96.373 -75.248 51.163
231005 96.404 -75.249 51.139
231006 96.435 -75.250 51.114
231007 96.462 -75.256 51.087
231008 96.482 -75.268 51.060
231009 96.502 -75.281 51.034
231010 96.522 -75.293 51.009
231011 96.542 -75.305 50.986
231012 96.562 -75.318 50.962
231013 96.581 -75.333 50.939
231014 96.599 -75.350 50.915
231015 96.622 -75.356 50.891
231016 96.645 -75.361 50.867
231017 96.668 -75.367 50.844
231018 96.690 -75.371 50.821
231019 96.711 -75.376 50.798
231020 96.731 -75.382 50.775
231021 96.741 -75.401 50.752
231022 96.752 -75.420 50.728
231023 96.766 -75.441 50.703
231024 96.779 -75.462 50.678
231025 96.793 -75.483 50.653
231026 96.806 -75.504 50.626
231027 96.819 -75.524 50.598
231028 96.839 -75.541 50.569
231029 96.863 -75.554 50.540
231030 96.887 -75.567 50.512
231031 96.911 -75.580 50.483
231032 96.935 -75.593 50.455
231033 96.966 -75.603 50.426
231034 97.031 -75.608 50.396
231035 97.096 -75.612 50.366
231036 97.145 -75.617 50.336
231037 97.169 -75.623 50.305
231038 97.193 -75.628 50.274
231039 97.218 -75.634 50.242
231040 97.242 -75.640 50.210
231041 97.266 -75.646 50.177
231042 97.291 -75.652 50.143
231043 97.315 -75.658 50.109
231044 97.339 -75.664 50.075
231045 97.363 -75.669 50.040
231046 97.388 -75.675 50.005
231047 97.405 -75.687 49.970
231048 97.428 -75.701 49.935
231049 97.450 -75.717 49.900
231050 97.473 -75.732 49.864
231051 97.496 -75.747 49.829
231052 97.519 -75.763 49.794
231053 97.547 -75.770 49.760
231054 97.580 -75.773 49.727
231055 97.611 -75.776 49.694
231056 97.643 -75.779 49.663
231057 97.675 -75.780 49.632
231058 97.706 -75.777 49.599
231059 97.737 -75.775 49.565
231060 97.768 -75.772 49.530
231061 97.799 -75.770 49.493
231062 97.837 -75.785 49.455
231063 97.851 -75.791 49.418
231064 97.862 -75.796 49.379
231065 97.872 -75.803 49.337
231066 97.879 -75.813 49.295
231067 97.889 -75.822 49.253
231068 97.905 -75.831 49.211
231069 97.924 -75.852 49.170
231070 97.942 -75.874 49.128
231071 97.960 -75.892 49.086
231072 97.979 -75.906 49.044
231073 97.997 -75.920 49.001
231074 98.015 -75.934 48.959
231075 98.033 -75.947 48.916
231076 98.065 -75.963 48.873
231077 98.111 -75.982 48.830
231078 98.157 -76.000 48.787
231079 98.203 -76.019 48.743
231080 98.249 -76.038 48.700
231081 98.295 -76.056 48.658
231082 98.337 -76.073 48.616
231083 98.377 -76.088 48.575
231084 98.417 -76.104 48.534
231085 98.457 -76.120 48.491
231086 98.477 -76.138 48.449
231087 98.489 -76.156 48.409
231088 98.501 -76.173 48.371
231089 98.512 -76.191 48.332
231090 98.524 -76.209 48.296
231091 98.536 -76.227 48.261
231092 98.575 -76.282 48.225
231093 98.607 -76.285 48.192
231094 98.639 -76.282 48.158
231095 98.671 -76.280 48.125
231096 98.703 -76.278 48.094
231097 98.734 -76.275 48.063
231098 98.762 -76.273 48.032
231099 98.788 -76.273 48.003
231100 98.813 -76.271 47.975
231101 98.839 -76.270 47.950
231102 98.865 -76.270 47.926
231103 98.890 -76.269 47.903
231104 98.916 -76.267 47.883
231105 98.942 -76.267 47.865
231106 98.968 -76.270 47.847
231107 98.994 -76.272 47.830
231108 99.020 -76.274 47.811
231109 99.046 -76.276 47.790
231110 99.072 -76.279 47.769
231111 99.097 -76.286 47.748
231112 99.121 -76.298 47.725
231113 99.145 -76.308 47.703
231114 99.169 -76.319 47.679
231115 99.194 -76.330 47.653
231116 99.218 -76.341 47.627
231117 99.242 -76.351 47.599
231118 99.264 -76.361 47.570
231119 99.272 -76.358 47.541
231120 99.281 -76.355 47.511
231121 99.289 -76.353 47.479
231122 99.297 -76.350 47.448
231123 99.305 -76.347 47.418
231124 99.314 -76.344 47.387
231125 99.337 -76.343 47.357
231126 99.358 -76.342 47.326
231127 99.380 -76.342 47.295
231128 99.402 -76.340 47.262
231129 99.422 -76.340 47.230
231130 99.446 -76.347 47.197
231131 99.477 -76.361 47.164
231132 99.507 -76.373 47.132
231133 99.539 -76.387 47.101
231134 99.569 -76.400 47.071
231135 99.600 -76.413 47.042
231136 99.632 -76.426 47.009
231137 99.667 -76.439 46.976
231138 99.702 -76.451 46.943
231139 99.737 -76.464 46.911
231140 99.772 -76.477 46.879
231141 99.805 -76.492 46.847
231142 99.838 -76.508 46.817
231143 99.869 -76.525 46.786
231144 99.902 -76.541 46.756
231145 99.933 -76.558 46.726
231146 99.955 -76.562 46.698
231147 99.975 -76.568 46.668
231148 99.993 -76.582 46.638
231149 100.011 -76.595 46.607
231150 100.030 -76.608 46.575
231151 100.048 -76.621 46.542
231152 100.066 -76.634 46.510
231153 100.084 -76.647 46.478
231154 100.103 -76.661 46.446
231155 100.121 -76.673 46.412
231156 100.134 -76.676 46.376
231157 100.148 -76.679 46.341
231158 100.161 -76.682 46.306
231159 100.175 -76.685 46.272
231160 100.190 -76.685 46.238
231161 100.209 -76.672 46.205
231162 100.227 -76.660 46.171
231163 100.247 -76.648 46.139
231164 100.266 -76.635 46.107
231165 100.285 -76.631 46.075
231166 100.304 -76.640 46.044
231167 100.324 -76.647 46.015
231168 100.344 -76.655 45.986
231169 100.363 -76.663 45.957
231170 100.382 -76.671 45.930
231171 100.402 -76.679 45.904
231172 100.422 -76.687 45.878
231173 100.441 -76.694 45.854
231174 100.461 -76.702 45.830
231175 100.480 -76.710 45.808
231176 100.500 -76.718 45.786
231177 100.519 -76.726 45.765
231178 100.546 -76.732 45.744
231179 100.570 -76.742 45.724
231180 100.594 -76.751 45.705
231181 100.618 -76.760 45.686
231182 100.642 -76.769 45.667
231183 100.666 -76.778 45.648
231184 100.690 -76.787 45.629
231185 100.714 -76.796 45.610
231186 100.738 -76.806 45.592
231187 100.762 -76.815 45.572
231188 100.787 -76.824 45.553
231189 100.815 -76.835 45.535
231190 100.843 -76.846 45.516
231191 100.871 -76.856 45.498
231192 100.899 -76.867 45.481
231193 100.927 -76.878 45.464
231194 100.947 -76.889 45.448
231195 100.962 -76.901 45.432
231196 100.976 -76.913 45.418
231197 100.991 -76.925 45.405
231198 101.006 -76.937 45.392
231199 101.021 -76.949 45.379
231200 101.035 -76.961 45.366
231201 101.062 -76.982 45.353
231202 101.091 -77.004 45.342
231203 101.120 -77.027 45.330
231204 101.148 -77.049 45.319
231205 101.178 -77.072 45.306
231206 101.206 -77.094 45.294
231207 101.236 -77.116 45.280
231208 101.264 -77.139 45.266
231209 101.293 -77.161 45.251
231210 101.319 -77.182 45.237
231211 101.334 -77.194 45.221
231212 101.349 -77.207 45.205
231213 101.364 -77.219 45.189
231214 101.379 -77.232 45.173
231215 101.394 -77.244 45.157
231216 101.415 -77.247 45.139
231217 101.438 -77.247 45.123
231218 101.460 -77.247 45.105
231219 101.481 -77.249 45.087
231220 101.500 -77.256 45.069
231221 101.519 -77.263 45.050
231222 101.537 -77.269 45.032
231223 101.556 -77.276 45.013
231224 101.574 -77.282 44.995
231225 101.593 -77.288 44.976
231226 101.611 -77.295 44.958
231227 101.630 -77.301 44.940
231228 101.649 -77.308 44.923
231229 101.667 -77.315 44.907
231230 101.685 -77.321 44.891
231231 101.704 -77.328 44.874
231232 101.723 -77.335 44.857
231233 101.741 -77.341 44.841
231234 101.759 -77.346 44.824
231235 101.774 -77.348 44.808
231236 101.790 -77.367 44.792
231237 101.810 -77.386 44.776
231238 101.830 -77.406 44.759
231239 101.850 -77.429 44.742
231240 101.870 -77.451 44.727
231241 101.889 -77.473 44.711
231242 101.909 -77.496 44.694
231243 101.928 -77.519 44.678
231244 101.948 -77.542 44.661
231245 101.968 -77.564 44.644
231246 102.024 -77.573 44.628
231247 102.081 -77.581 44.612
231248 102.106 -77.590 44.597
231249 102.126 -77.599 44.580
231250 102.147 -77.608 44.565
231251 102.167 -77.617 44.550
231252 102.192 -77.627 44.534
231253 102.225 -77.638 44.519
231254 102.257 -77.650 44.505
231255 102.290 -77.661 44.491
231256 102.322 -77.673 44.476
231257 102.341 -77.692 44.461
231258 102.356 -77.712 44.446
231259 102.371 -77.733 44.429
231260 102.394 -77.740 44.413
231261 102.419 -77.745 44.397
231262 102.445 -77.749 44.381
231263 102.471 -77.753 44.364
231264 102.496 -77.757 44.347
231265 102.522 -77.761 44.332
231266 102.547 -77.765 44.316
231267 102.573 -77.770 44.298
231268 102.599 -77.774 44.282
231269 102.624 -77.778 44.265
231270 102.650 -77.782 44.249
231271 102.685 -77.787 44.232
231272 102.721 -77.793 44.216
231273 102.757 -77.798 44.199
231274 102.794 -77.803 44.182
231275 102.830 -77.809 44.166
231276 102.867 -77.815 44.149
231277 102.903 -77.820 44.133
231278 102.939 -77.825 44.116
231279 102.975 -77.830 44.100
231280 103.012 -77.836 44.084
231281 103.048 -77.841 44.067
231282 103.085 -77.846 44.050
231283 103.121 -77.852 44.034
231284 103.158 -77.857 44.019
231285 103.195 -77.862 44.003
231535 92.155 -74.937 54.296
231536 92.179 -74.947 54.301
231537 92.204 -74.959 54.310
231538 92.228 -74.969 54.320
231539 92.252 -74.980 54.331
231540 92.277 -74.991 54.343
231541 92.301 -75.001 54.355
231542 92.325 -75.013 54.366
231543 92.350 -75.023 54.380
231544 92.374 -75.034 54.394
231545 92.399 -75.044 54.408
231546 92.422 -75.052 54.420
231547 92.439 -75.034 54.432
231548 92.471 -75.031 54.442
231549 92.502 -75.028 54.447
231550 92.535 -75.024 54.450
231551 92.570 -75.020 54.451
231552 92.605 -75.016 54.448
231553 92.639 -75.013 54.442
231554 92.672 -75.009 54.428
231555 92.701 -75.019 54.410
231556 92.729 -75.035 54.388
231557 92.758 -75.049 54.373
231558 92.786 -75.058 54.361
231559 92.815 -75.067 54.351
231560 92.844 -75.069 54.342
231561 92.874 -75.035 54.332
231562 92.905 -75.002 54.320
231563 92.935 -74.969 54.307
231564 92.960 -74.962 54.293
231565 92.983 -74.961 54.279
231566 93.007 -74.960 54.266
231567 93.030 -74.958 54.252
231568 93.054 -74.958 54.238
231569 93.079 -74.960 54.226
231570 93.106 -74.962 54.212
231571 93.131 -74.964 54.198
231572 93.157 -74.966 54.185
231573 93.183 -74.968 54.173
231574 93.209 -74.970 54.161
231575 93.234 -74.972 54.150
231576 93.261 -74.974 54.141
231577 93.286 -74.976 54.133
231578 93.312 -74.978 54.126
231579 93.336 -74.974 54.121
231580 93.359 -74.967 54.120
231581 93.382 -74.960 54.120
231582 93.405 -74.954 54.125
231583 93.428 -74.947 54.132
231584 93.450 -74.939 54.139
231585 93.474 -74.932 54.151
231586 93.496 -74.925 54.168
231587 93.519 -74.919 54.189
231588 93.543 -74.912 54.218
231589 93.570 -74.913 54.254
231590 93.599 -74.915 54.289
231591 93.628 -74.917 54.322
231592 93.657 -74.919 54.353
231593 93.686 -74.921 54.385
231594 93.714 -74.923 54.418
231595 93.743 -74.925 54.451
231596 93.773 -74.930 54.477
231597 93.803 -74.940 54.494
231598 93.821 -74.937 54.504
231599 93.846 -74.935 54.508
231600 93.872 -74.934 54.507
231601 93.898 -74.932 54.508
231602 93.924 -74.931 54.512
231603 93.950 -74.929 54.520
231604 93.976 -74.928 54.524
231605 94.002 -74.926 54.531
231606 94.028 -74.925 54.538
231607 94.054 -74.923 54.544
231608 94.082 -74.918 54.551
231609 94.111 -74.912 54.554
231610 94.139 -74.906 54.554
231611 94.168 -74.900 54.547
231612 94.196 -74.893 54.535
231613 94.224 -74.887 54.520
231614 94.252 -74.881 54.501
231615 94.268 -74.885 54.478
231616 94.284 -74.888 54.456
231617 94.303 -74.888 54.434
231618 94.329 -74.882 54.412
231619 94.354 -74.876 54.396
231620 94.380 -74.870 54.385
231621 94.403 -74.866 54.376
231622 94.425 -74.865 54.368
231623 94.447 -74.863 54.360
231624 94.463 -74.866 54.345
231625 94.476 -74.872 54.327
231626 94.489 -74.878 54.308
231627 94.501 -74.883 54.287
231628 94.514 -74.889 54.266
231629 94.526 -74.897 54.244
231630 94.536 -74.907 54.221
231631 94.547 -74.916 54.191
231632 94.564 -74.923 54.158
231633 94.585 -74.929 54.124
231634 94.606 -74.934 54.092
231635 94.626 -74.935 54.061
231636 94.646 -74.938 54.031
231637 94.666 -74.939 54.000
231638 94.686 -74.941 53.962
231639 94.706 -74.943 53.923
231640 94.727 -74.945 53.882
231641 94.747 -74.947 53.838
231642 94.768 -74.950 53.790
231643 94.794 -74.957 53.746
231644 94.822 -74.965 53.701
231645 94.848 -74.973 53.653
231646 94.864 -74.979 53.604
231647 94.879 -74.984 53.555
231648 94.895 -74.990 53.507
231649 94.910 -74.995 53.460
231650 94.926 -75.001 53.411
231651 94.941 -75.007 53.360
231652 94.957 -75.012 53.309
231653 94.972 -75.017 53.259
231654 94.980 -75.017 53.208
231655 94.974 -75.007 53.156
231656 94.968 -74.997 53.100
231657 94.966 -74.989 53.041
231658 94.998 -74.996 52.978
231659 95.030 -75.003 52.912
231660 95.062 -75.009 52.847
231661 95.094 -75.016 52.782
231662 95.126 -75.022 52.718
231663 95.152 -75.025 52.655
231664 95.177 -75.028 52.593
231665 95.202 -75.030 52.532
231666 95.228 -75.032 52.472
231667 95.255 -75.042 52.415
231668 95.282 -75.053 52.359
231669 95.310 -75.064 52.305
231670 95.337 -75.076 52.255
231671 95.359 -75.082 52.206
231672 95.374 -75.082 52.158
231673 95.388 -75.081 52.112
231674 95.402 -75.080 52.067
231675 95.417 -75.080 52.024
231676 95.431 -75.079 51.983
231677 95.446 -75.079 51.942
231678 95.465 -75.084 51.903
231679 95.489 -75.095 51.866
231680 95.513 -75.105 51.831
231681 95.537 -75.116 51.797
231682 95.564 -75.114 51.764
231683 95.593 -75.107 51.731
231684 95.625 -75.104 51.700
231685 95.666 -75.109 51.672
231686 95.706 -75.115 51.644
231687 95.748 -75.121 51.617
231688 95.789 -75.126 51.591
231689 95.829 -75.132 51.565
231690 95.870 -75.138 51.539
231691 95.911 -75.144 51.512
231692 95.948 -75.152 51.486
231693 95.983 -75.160 51.459
231694 96.019 -75.169 51.434
231695 96.054 -75.177 51.409
231696 96.090 -75.185 51.384
231697 96.126 -75.194 51.359
231698 96.162 -75.202 51.335
231699 96.197 -75.210 51.313
231700 96.229 -75.213 51.291
231701 96.260 -75.215 51.267
231702 96.290 -75.217 51.243
231703 96.317 -75.220 51.220
231704 96.345 -75.223 51.195
231705 96.375 -75.225 51.172
231706 96.406 -75.226 51.148
231707 96.437 -75.227 51.123
231708 96.469 -75.228 51.097
231709 96.489 -75.240 51.070
231710 96.509 -75.253 51.043
231711 96.529 -75.265 51.017
231712 96.548 -75.279 50.993
231713 96.566 -75.297 50.969
231714 96.584 -75.314 50.945
231715 96.604 -75.324 50.923
231716 96.627 -75.330 50.900
231717 96.650 -75.335 50.877
231718 96.673 -75.340 50.853
231719 96.696 -75.345 50.830
231720 96.719 -75.350 50.808
231721 96.741 -75.357 50.785
231722 96.754 -75.377 50.762
231723 96.767 -75.399 50.737
231724 96.781 -75.419 50.712
231725 96.795 -75.439 50.688
231726 96.808 -75.461 50.664
231727 96.821 -75.481 50.638
231728 96.836 -75.501 50.610
231729 96.860 -75.514 50.582
231730 96.884 -75.527 50.553
231731 96.908 -75.540 50.525
231732 96.932 -75.552 50.497
231733 96.956 -75.565 50.468
231734 96.980 -75.578 50.439
231735 97.041 -75.583 50.409
231736 97.105 -75.587 50.378
231737 97.155 -75.593 50.348
231738 97.179 -75.598 50.317
231739 97.204 -75.604 50.285
231740 97.228 -75.610 50.252
231741 97.252 -75.615 50.219
231742 97.276 -75.621 50.186
231743 97.301 -75.627 50.153
231744 97.325 -75.634 50.118
231745 97.349 -75.639 50.084
231746 97.373 -75.645 50.049
231747 97.391 -75.656 50.014
231748 97.408 -75.668 49.979
231749 97.431 -75.683 49.944
231750 97.453 -75.698 49.910
231751 97.476 -75.714 49.875
231752 97.499 -75.729 49.840
231753 97.527 -75.737 49.806
231754 97.559 -75.740 49.772
231755 97.590 -75.743 49.739
231756 97.623 -75.746 49.707
231757 97.654 -75.749 49.675
231758 97.686 -75.753 49.644
231759 97.718 -75.755 49.609
231760 97.749 -75.753 49.574
231761 97.780 -75.749 49.538
231762 97.811 -75.749 49.500
231763 97.849 -75.765 49.460
231764 97.878 -75.777 49.421
231765 97.889 -75.781 49.381
231766 97.897 -75.791 49.338
231767 97.905 -75.801 49.296
231768 97.913 -75.811 49.254
231769 97.929 -75.815 49.212
231770 97.947 -75.834 49.171
231771 97.965 -75.849 49.128
231772 97.984 -75.863 49.085
231773 98.002 -75.877 49.042
231774 98.020 -75.891 48.998
231775 98.038 -75.905 48.956
231776 98.057 -75.919 48.913
231777 98.075 -75.932 48.870
231778 98.121 -75.951 48.826
231779 98.167 -75.969 48.783
231780 98.213 -75.988 48.740
231781 98.258 -76.007 48.697
231782 98.304 -76.025 48.655
231783 98.345 -76.041 48.613
231784 98.385 -76.057 48.572
231785 98.425 -76.072 48.530
231786 98.465 -76.088 48.486
231787 98.493 -76.105 48.444
231788 98.505 -76.123 48.404
231789 98.516 -76.141 48.365
231790 98.528 -76.159 48.327
231791 98.540 -76.177 48.291
231792 98.565 -76.213 48.256
231793 98.604 -76.266 48.220
231794 98.637 -76.263 48.186
231795 98.668 -76.260 48.152
231796 98.700 -76.258 48.119
231797 98.732 -76.255 48.086
231798 98.760 -76.254 48.054
231799 98.785 -76.253 48.023
231800 98.811 -76.252 47.994
231801 98.837 -76.251 47.966
231802 98.862 -76.250 47.941
231803 98.888 -76.249 47.916
231804 98.914 -76.248 47.893
231805 98.939 -76.247 47.872
231806 98.964 -76.246 47.853
231807 98.991 -76.248 47.836
231808 99.017 -76.250 47.819
231809 99.043 -76.252 47.800
231810 99.069 -76.254 47.779
231811 99.094 -76.258 47.757
231812 99.119 -76.269 47.736
231813 99.143 -76.279 47.713
231814 99.167 -76.291 47.691
231815 99.191 -76.301 47.668
231816 99.215 -76.312 47.644
231817 99.239 -76.323 47.619
231818 99.264 -76.334 47.590
231819 99.288 -76.345 47.562
231820 99.299 -76.345 47.533
231821 99.307 -76.342 47.503
231822 99.316 -76.339 47.472
231823 99.324 -76.336 47.442
231824 99.332 -76.333 47.412
231825 99.348 -76.332 47.382
231826 99.370 -76.331 47.351
231827 99.392 -76.329 47.321
231828 99.414 -76.329 47.289
231829 99.433 -76.329 47.257
231830 99.448 -76.332 47.225
231831 99.464 -76.333 47.193
231832 99.492 -76.343 47.160
231833 99.521 -76.353 47.130
231834 99.551 -76.365 47.100
231835 99.582 -76.379 47.070
231836 99.613 -76.392 47.040
231837 99.645 -76.404 47.008
231838 99.681 -76.417 46.976
231839 99.716 -76.430 46.944
231840 99.749 -76.445 46.912
231841 99.780 -76.462 46.880
231842 99.812 -76.478 46.849
231843 99.844 -76.495 46.818
231844 99.876 -76.511 46.789
231845 99.909 -76.527 46.757
231846 99.936 -76.539 46.728
231847 99.956 -76.542 46.699
231848 99.976 -76.548 46.670
231849 99.994 -76.561 46.640
231850 100.012 -76.574 46.608
231851 100.031 -76.587 46.575
231852 100.049 -76.600 46.542
231853 100.067 -76.613 46.510
231854 100.085 -76.627 46.478
231855 100.104 -76.640 46.444
231856 100.122 -76.652 46.410
231857 100.140 -76.666 46.374
231858 100.159 -76.679 46.339
231859 100.176 -76.690 46.303
231860 100.190 -76.693 46.269
231861 100.207 -76.686 46.235
231862 100.226 -76.673 46.202
231863 100.246 -76.661 46.169
231864 100.264 -76.649 46.136
231865 100.283 -76.636 46.103
231866 100.302 -76.624 46.072
231867 100.322 -76.630 46.041
231868 100.341 -76.639 46.011
231869 100.361 -76.646 45.982
231870 100.381 -76.654 45.954
231871 100.400 -76.662 45.927
231872 100.420 -76.670 45.900
231873 100.439 -76.678 45.874
231874 100.459 -76.686 45.850
231875 100.478 -76.693 45.826
231876 100.498 -76.702 45.804
231877 100.517 -76.709 45.782
231878 100.537 -76.721 45.760
231879 100.572 -76.724 45.739
231880 100.601 -76.729 45.719
231881 100.625 -76.739 45.700
231882 100.649 -76.748 45.682
231883 100.673 -76.757 45.662
231884 100.697 -76.766 45.643
231885 100.720 -76.775 45.623
231886 100.745 -76.784 45.604
231887 100.768 -76.793 45.585
231888 100.792 -76.803 45.566
231889 100.818 -76.813 45.546
231890 100.846 -76.823 45.527
231891 100.874 -76.834 45.508
231892 100.902 -76.844 45.489
231893 100.930 -76.855 45.472
231894 100.958 -76.865 45.454
231895 100.975 -76.877 45.438
231896 100.990 -76.889 45.422
231897 101.004 -76.901 45.407
231898 101.019 -76.913 45.394
231899 101.034 -76.925 45.381
231900 101.049 -76.937 45.368
231901 101.063 -76.950 45.356
231902 101.078 -76.961 45.345
231903 101.101 -76.979 45.333
231904 101.130 -77.001 45.322
231905 101.159 -77.024 45.310
231906 101.188 -77.047 45.298
231907 101.216 -77.069 45.284
231908 101.245 -77.091 45.271
231909 101.274 -77.113 45.256
231910 101.303 -77.136 45.242
231911 101.332 -77.158 45.227
231912 101.348 -77.172 45.212
231913 101.363 -77.185 45.196
231914 101.378 -77.197 45.180
231915 101.393 -77.210 45.164
231916 101.408 -77.221 45.148
231917 101.431 -77.221 45.131
231918 101.454 -77.222 45.115
231919 101.476 -77.222 45.097
231920 101.498 -77.222 45.078
231921 101.517 -77.229 45.060
231922 101.536 -77.235 45.041
231923 101.554 -77.242 45.023
231924 101.572 -77.249 45.005
231925 101.591 -77.255 44.986
231926 101.610 -77.262 44.968
231927 101.628 -77.269 44.950
231928 101.647 -77.275 44.933
231929 101.666 -77.282 44.916
231930 101.684 -77.288 44.900
231931 101.702 -77.294 44.885
231932 101.721 -77.301 44.868
231933 101.740 -77.307 44.851
231934 101.758 -77.314 44.835
231935 101.777 -77.321 44.818
231936 101.796 -77.328 44.801
231937 101.817 -77.348 44.785
231938 101.836 -77.371 44.768
231939 101.856 -77.394 44.752
231940 101.876 -77.417 44.735
231941 101.895 -77.439 44.720
231942 101.915 -77.461 44.704
231943 101.935 -77.484 44.687
231944 101.954 -77.507 44.671
231945 101.974 -77.530 44.654
231946 101.996 -77.551 44.637
231947 102.053 -77.559 44.621
231948 102.098 -77.568 44.605
231949 102.119 -77.577 44.589
231950 102.139 -77.586 44.573
231951 102.159 -77.595 44.557
231952 102.180 -77.603 44.543
231953 102.200 -77.612 44.526
231954 102.232 -77.624 44.511
231955 102.265 -77.636 44.497
231956 102.297 -77.647 44.483
231957 102.330 -77.658 44.468
231958 102.348 -77.677 44.453
231959 102.363 -77.699 44.438
231960 102.379 -77.718 44.421
231961 102.404 -77.721 44.405
231962 102.430 -77.726 44.389
231963 102.455 -77.730 44.372
231964 102.481 -77.734 44.355
231965 102.507 -77.738 44.338
231966 102.532 -77.743 44.322
231967 102.558 -77.746 44.305
231968 102.583 -77.751 44.289
231969 102.609 -77.755 44.272
231970 102.637 -77.759 44.256
231971 102.673 -77.765 44.239
231972 102.710 -77.770 44.222
231973 102.746 -77.775 44.206
231974 102.782 -77.780 44.189
231975 102.818 -77.786 44.172
231976 102.855 -77.791 44.156
231977 102.891 -77.796 44.139
231978 102.927 -77.802 44.123
231979 102.964 -77.807 44.106
231980 103.000 -77.812 44.090
231981 103.036 -77.818 44.074
231982 103.073 -77.824 44.058
231983 103.109 -77.829 44.041
231984 103.146 -77.834 44.025
231985 103.182 -77.840 44.009
232237 92.181 -74.916 54.305
232238 92.205 -74.926 54.314
232239 92.230 -74.937 54.323
232240 92.254 -74.947 54.334
232241 92.278 -74.958 54.345
232242 92.303 -74.969 54.356
232243 92.327 -74.980 54.368
232244 92.352 -74.991 54.381
232245 92.376 -75.001 54.396
232246 92.400 -75.013 54.410
232247 92.425 -75.023 54.423
232248 92.439 -75.005 54.435
232249 92.467 -74.998 54.446
232250 92.500 -74.994 54.453
232251 92.535 -74.991 54.459
232252 92.568 -74.987 54.461
232253 92.601 -74.983 54.459
232254 92.634 -74.980 54.453
232255 92.667 -74.976 54.442
232256 92.699 -74.973 54.425
232257 92.728 -74.987 54.405
232258 92.756 -75.001 54.392
232259 92.784 -75.016 54.381
232260 92.813 -75.030 54.372
232261 92.842 -75.032 54.365
232262 92.872 -75.005 54.356
232263 92.902 -74.978 54.343
232264 92.932 -74.951 54.328
232265 92.961 -74.930 54.313
232266 92.987 -74.932 54.298
232267 93.013 -74.935 54.284
232268 93.039 -74.936 54.271
232269 93.064 -74.938 54.257
232270 93.090 -74.940 54.245
232271 93.116 -74.942 54.232
232272 93.142 -74.944 54.218
232273 93.167 -74.946 54.205
232274 93.194 -74.948 54.193
232275 93.219 -74.950 54.182
232276 93.245 -74.952 54.172
232277 93.271 -74.954 54.163
232278 93.295 -74.951 54.155
232279 93.318 -74.944 54.148
232280 93.341 -74.938 54.143
232281 93.364 -74.931 54.142
232282 93.387 -74.923 54.143
232283 93.410 -74.916 54.147
232284 93.433 -74.910 54.154
232285 93.455 -74.903 54.161
232286 93.479 -74.896 54.173
232287 93.501 -74.888 54.188
232288 93.524 -74.881 54.209
232289 93.547 -74.875 54.237
232290 93.575 -74.875 54.271
232291 93.604 -74.878 54.306
232292 93.633 -74.880 54.340
232293 93.662 -74.882 54.371
232294 93.691 -74.885 54.400
232295 93.719 -74.886 54.428
232296 93.748 -74.889 54.459
232297 93.781 -74.901 54.485
232298 93.814 -74.914 54.502
232299 93.834 -74.913 54.511
232300 93.852 -74.910 54.515
232301 93.871 -74.907 54.514
232302 93.897 -74.906 54.516
232303 93.923 -74.904 54.521
232304 93.949 -74.903 54.531
232305 93.975 -74.901 54.536
232306 94.001 -74.900 54.543
232307 94.027 -74.898 54.550
232308 94.054 -74.896 54.556
232309 94.082 -74.889 54.562
232310 94.111 -74.883 54.565
232311 94.139 -74.877 54.564
232312 94.168 -74.871 54.556
232313 94.196 -74.864 54.543
232314 94.223 -74.857 54.526
232315 94.247 -74.855 54.505
232316 94.267 -74.857 54.481
232317 94.283 -74.861 54.456
232318 94.303 -74.860 54.432
232319 94.329 -74.854 54.408
232320 94.353 -74.849 54.389
232321 94.374 -74.847 54.376
232322 94.396 -74.846 54.367
232323 94.418 -74.844 54.359
232324 94.439 -74.844 54.351
232325 94.452 -74.850 54.337
232326 94.465 -74.855 54.319
232327 94.477 -74.861 54.300
232328 94.490 -74.866 54.280
232329 94.503 -74.872 54.260
232330 94.516 -74.878 54.240
232331 94.529 -74.885 54.218
232332 94.542 -74.893 54.191
232333 94.562 -74.897 54.160
232334 94.582 -74.899 54.130
232335 94.602 -74.900 54.102
232336 94.622 -74.903 54.075
232337 94.642 -74.904 54.048
232338 94.662 -74.907 54.019
232339 94.682 -74.908 53.985
232340 94.702 -74.910 53.946
232341 94.723 -74.913 53.904
232342 94.745 -74.916 53.860
232343 94.773 -74.924 53.815
232344 94.801 -74.933 53.771
232345 94.829 -74.941 53.725
232346 94.857 -74.950 53.676
232347 94.878 -74.957 53.626
232348 94.893 -74.962 53.577
232349 94.909 -74.968 53.529
232350 94.924 -74.973 53.480
232351 94.939 -74.979 53.431
232352 94.955 -74.984 53.381
232353 94.971 -74.990 53.330
232354 94.986 -74.995 53.279
232355 94.985 -74.988 53.227
232356 94.979 -74.979 53.173
232357 94.973 -74.969 53.117
232358 94.972 -74.961 53.059
232359 95.004 -74.968 52.993
232360 95.036 -74.974 52.927
232361 95.068 -74.981 52.861
232362 95.100 -74.988 52.795
232363 95.132 -74.994 52.731
232364 95.162 -75.000 52.667
232365 95.187 -75.002 52.605
232366 95.212 -75.004 52.544
232367 95.238 -75.007 52.483
232368 95.265 -75.018 52.425
232369 95.293 -75.029 52.368
232370 95.320 -75.041 52.314
232371 95.347 -75.051 52.263
232372 95.375 -75.063 52.214
232373 95.390 -75.063 52.165
232374 95.404 -75.063 52.116
232375 95.419 -75.062 52.071
232376 95.433 -75.062 52.027
232377 95.448 -75.061 51.985
232378 95.462 -75.060 51.944
232379 95.476 -75.060 51.905
232380 95.490 -75.060 51.868
232381 95.514 -75.069 51.833
232382 95.538 -75.080 51.800
232383 95.566 -75.077 51.765
232384 95.599 -75.075 51.732
232385 95.640 -75.081 51.701
232386 95.681 -75.087 51.672
232387 95.722 -75.092 51.643
232388 95.763 -75.098 51.616
232389 95.804 -75.104 51.589
232390 95.844 -75.110 51.563
232391 95.885 -75.116 51.536
232392 95.926 -75.122 51.510
232393 95.962 -75.130 51.484
232394 95.997 -75.138 51.459
232395 96.033 -75.147 51.435
232396 96.069 -75.155 51.412
232397 96.105 -75.163 51.387
232398 96.141 -75.172 51.363
232399 96.176 -75.180 51.341
232400 96.207 -75.182 51.320
232401 96.239 -75.184 51.298
232402 96.270 -75.186 51.274
232403 96.302 -75.189 51.252
232404 96.330 -75.192 51.229
232405 96.358 -75.195 51.204
232406 96.385 -75.198 51.180
232407 96.413 -75.201 51.157
232408 96.440 -75.204 51.132
232409 96.471 -75.205 51.106
232410 96.496 -75.212 51.079
232411 96.515 -75.226 51.052
232412 96.533 -75.243 51.025
232413 96.551 -75.260 51.001
232414 96.569 -75.277 50.976
232415 96.587 -75.292 50.953
232416 96.610 -75.298 50.931
232417 96.633 -75.303 50.908
232418 96.656 -75.308 50.886
232419 96.679 -75.313 50.863
232420 96.702 -75.318 50.840
232421 96.726 -75.327 50.818
232422 96.751 -75.338 50.795
232423 96.769 -75.355 50.771
232424 96.783 -75.376 50.747
232425 96.796 -75.397 50.723
232426 96.810 -75.417 50.699
232427 96.823 -75.439 50.676
232428 96.836 -75.459 50.651
232429 96.857 -75.474 50.623
232430 96.881 -75.487 50.595
232431 96.905 -75.500 50.567
232432 96.930 -75.513 50.539
232433 96.953 -75.526 50.511
232434 96.977 -75.539 50.482
232435 97.002 -75.552 50.452
232436 97.051 -75.558 50.421
232437 97.115 -75.563 50.391
232438 97.165 -75.568 50.360
232439 97.189 -75.574 50.328
232440 97.214 -75.580 50.296
232441 97.238 -75.586 50.263
232442 97.262 -75.591 50.229
232443 97.286 -75.597 50.195
232444 97.311 -75.603 50.162
232445 97.335 -75.609 50.128
232446 97.359 -75.615 50.094
232447 97.377 -75.626 50.059
232448 97.394 -75.637 50.024
232449 97.412 -75.649 49.989
232450 97.434 -75.665 49.954
232451 97.456 -75.680 49.920
232452 97.479 -75.696 49.886
232453 97.506 -75.706 49.852
232454 97.538 -75.709 49.818
232455 97.570 -75.712 49.784
232456 97.601 -75.715 49.753
232457 97.633 -75.718 49.721
232458 97.665 -75.721 49.688
232459 97.697 -75.724 49.656
232460 97.729 -75.727 49.621
232461 97.760 -75.730 49.585
232462 97.792 -75.728 49.547
232463 97.824 -75.728 49.507
232464 97.862 -75.744 49.467
232465 97.899 -75.760 49.426
232466 97.915 -75.768 49.384
232467 97.923 -75.779 49.341
232468 97.930 -75.789 49.298
232469 97.935 -75.797 49.255
232470 97.947 -75.799 49.213
232471 97.969 -75.806 49.172
232472 97.989 -75.820 49.129
232473 98.007 -75.834 49.085
232474 98.025 -75.848 49.041
232475 98.043 -75.862 48.996
232476 98.062 -75.875 48.954
232477 98.080 -75.890 48.911
232478 98.099 -75.903 48.868
232479 98.131 -75.919 48.824
232480 98.176 -75.938 48.781
232481 98.222 -75.956 48.738
232482 98.268 -75.975 48.695
232483 98.313 -75.993 48.652
232484 98.353 -76.009 48.611
232485 98.393 -76.025 48.568
232486 98.434 -76.041 48.524
232487 98.474 -76.057 48.481
232488 98.509 -76.072 48.440
232489 98.520 -76.091 48.399
232490 98.532 -76.109 48.360
232491 98.544 -76.126 48.322
232492 98.556 -76.144 48.286
232493 98.596 -76.201 48.251
232494 98.634 -76.244 48.216
232495 98.666 -76.241 48.182
232496 98.698 -76.238 48.147
232497 98.729 -76.236 48.113
232498 98.757 -76.235 48.081
232499 98.783 -76.233 48.047
232500 98.808 -76.232 48.015
232501 98.834 -76.232 47.985
232502 98.860 -76.230 47.958
232503 98.885 -76.229 47.931
232504 98.911 -76.229 47.906
232505 98.937 -76.227 47.883
232506 98.962 -76.226 47.862
232507 98.988 -76.226 47.842
232508 99.014 -76.226 47.824
232509 99.040 -76.228 47.807
232510 99.066 -76.230 47.789
232511 99.092 -76.232 47.768
232512 99.117 -76.240 47.747
232513 99.141 -76.251 47.725
232514 99.165 -76.262 47.702
232515 99.189 -76.273 47.680
232516 99.213 -76.284 47.657
232517 99.237 -76.295 47.633
232518 99.262 -76.305 47.608
232519 99.286 -76.317 47.581
232520 99.310 -76.327 47.554
232521 99.327 -76.332 47.526
232522 99.335 -76.329 47.496
232523 99.343 -76.326 47.466
232524 99.351 -76.323 47.435
232525 99.360 -76.320 47.405
232526 99.382 -76.320 47.376
232527 99.404 -76.318 47.346
232528 99.426 -76.317 47.315
232529 99.444 -76.318 47.284
232530 99.459 -76.320 47.252
232531 99.475 -76.323 47.221
232532 99.490 -76.325 47.189
232533 99.509 -76.329 47.157
232534 99.538 -76.339 47.127
232535 99.567 -76.349 47.098
232536 99.596 -76.359 47.068
232537 99.626 -76.370 47.039
232538 99.659 -76.383 47.007
232539 99.691 -76.399 46.975
232540 99.723 -76.415 46.943
232541 99.756 -76.432 46.912
232542 99.787 -76.448 46.881
232543 99.820 -76.464 46.850
232544 99.851 -76.480 46.819
232545 99.884 -76.497 46.790
232546 99.915 -76.513 46.758
232547 99.937 -76.518 46.729
232548 99.957 -76.522 46.700
232549 99.977 -76.527 46.671
232550 99.995 -76.540 46.640
232551 100.013 -76.553 46.608
232552 100.031 -76.566 46.575
232553 100.050 -76.580 46.542
232554 100.068 -76.592 46.510
232555 100.087 -76.605 46.477
232556 100.105 -76.619 46.441
232557 100.123 -76.632 46.407
232558 100.142 -76.645 46.372
232559 100.160 -76.658 46.336
232560 100.178 -76.671 46.301
232561 100.207 -76.680 46.266
232562 100.225 -76.687 46.232
232563 100.244 -76.674 46.199
232564 100.263 -76.661 46.166
232565 100.282 -76.649 46.132
232566 100.301 -76.636 46.100
232567 100.320 -76.624 46.067
232568 100.339 -76.622 46.036
232569 100.359 -76.630 46.006
232570 100.379 -76.637 45.977
232571 100.398 -76.646 45.949
232572 100.418 -76.653 45.921
232573 100.437 -76.661 45.894
232574 100.457 -76.669 45.869
232575 100.476 -76.677 45.844
232576 100.496 -76.685 45.821
232577 100.515 -76.693 45.798
232578 100.535 -76.702 45.776
232579 100.554 -76.720 45.755
232580 100.598 -76.715 45.734
232581 100.632 -76.717 45.714
232582 100.656 -76.727 45.694
232583 100.680 -76.736 45.675
232584 100.704 -76.745 45.656
232585 100.727 -76.754 45.637
232586 100.752 -76.763 45.617
232587 100.775 -76.772 45.597
232588 100.799 -76.781 45.578
232589 100.823 -76.791 45.558
232590 100.849 -76.801 45.539
232591 100.877 -76.811 45.519
232592 100.905 -76.821 45.500
232593 100.933 -76.832 45.481
232594 100.961 -76.843 45.462
232595 100.989 -76.853 45.445
232596 101.003 -76.865 45.428
232597 101.018 -76.878 45.412
232598 101.033 -76.890 45.397
232599 101.048 -76.901 45.383
232600 101.062 -76.913 45.370
232601 101.077 -76.925 45.358
232602 101.092 -76.937 45.347
232603 101.107 -76.950 45.336
232604 101.121 -76.962 45.325
232605 101.140 -76.976 45.313
232606 101.169 -76.999 45.301
232607 101.197 -77.021 45.288
232608 101.227 -77.044 45.275
232609 101.255 -77.066 45.261
232610 101.284 -77.088 45.247
232611 101.313 -77.110 45.232
232612 101.342 -77.133 45.219
232613 101.362 -77.150 45.204
232614 101.378 -77.162 45.188
232615 101.392 -77.175 45.172
232616 101.407 -77.187 45.156
232617 101.424 -77.196 45.140
232618 101.447 -77.196 45.124
232619 101.470 -77.196 45.107
232620 101.492 -77.196 45.089
232621 101.516 -77.196 45.070
232622 101.534 -77.203 45.052
232623 101.553 -77.209 45.034
232624 101.571 -77.216 45.016
232625 101.590 -77.222 44.998
232626 101.608 -77.229 44.979
232627 101.627 -77.235 44.961
232628 101.645 -77.241 44.943
232629 101.664 -77.248 44.926
232630 101.682 -77.255 44.910
232631 101.701 -77.261 44.894
232632 101.720 -77.268 44.878
232633 101.738 -77.274 44.862
232634 101.757 -77.281 44.845
232635 101.775 -77.288 44.829
232636 101.795 -77.298 44.812
232637 101.814 -77.312 44.795
232638 101.837 -77.335 44.778
232639 101.860 -77.358 44.762
232640 101.882 -77.382 44.745
232641 101.902 -77.404 44.728
232642 101.921 -77.427 44.713
232643 101.941 -77.449 44.697
232644 101.961 -77.472 44.680
232645 101.980 -77.495 44.663
232646 102.000 -77.517 44.646
232647 102.026 -77.538 44.629
232648 102.083 -77.546 44.613
232649 102.111 -77.555 44.597
232650 102.131 -77.564 44.582
232651 102.152 -77.572 44.566
232652 102.172 -77.581 44.551
232653 102.192 -77.590 44.536
232654 102.212 -77.599 44.519
232655 102.240 -77.609 44.504
232656 102.272 -77.621 44.490
232657 102.305 -77.633 44.475
232658 102.337 -77.645 44.460
232659 102.355 -77.664 44.445
232660 102.370 -77.684 44.429
232661 102.389 -77.698 44.412
232662 102.414 -77.702 44.396
232663 102.440 -77.706 44.380
232664 102.466 -77.711 44.364
232665 102.491 -77.714 44.347
232666 102.517 -77.719 44.330
232667 102.542 -77.723 44.313
232668 102.568 -77.727 44.296
232669 102.594 -77.731 44.279
232670 102.625 -77.736 44.263
232671 102.662 -77.742 44.246
232672 102.698 -77.747 44.229
232673 102.734 -77.752 44.212
232674 102.770 -77.758 44.196
232675 102.807 -77.763 44.179
232676 102.843 -77.768 44.163
232677 102.879 -77.774 44.146
232678 102.915 -77.779 44.130
232679 102.952 -77.784 44.113
232680 102.988 -77.790 44.096
232681 103.025 -77.795 44.081
232682 103.061 -77.800 44.064
232683 103.097 -77.805 44.048
232684 103.134 -77.811 44.033
232685 103.170 -77.816 44.016
232686 103.206 -77.822 44.000
232938 92.182 -74.883 54.309
232939 92.207 -74.894 54.317
232940 92.231 -74.904 54.326
232941 92.256 -74.916 54.336
232942 92.280 -74.926 54.347
232943 92.304 -74.937 54.358
232944 92.329 -74.947 54.369
232945 92.353 -74.958 54.383
232946 92.378 -74.969 54.397
232947 92.402 -74.980 54.412
232948 92.427 -74.991 54.425
232949 92.444 -74.981 54.437
232950 92.464 -74.965 54.447
232951 92.497 -74.961 54.456
232952 92.530 -74.957 54.463
232953 92.564 -74.954 54.466
232954 92.596 -74.950 54.467
232955 92.630 -74.947 54.461
232956 92.662 -74.944 54.451
232957 92.695 -74.941 54.437
232958 92.726 -74.939 54.421
232959 92.755 -74.954 54.411
232960 92.783 -74.968 54.402
232961 92.811 -74.982 54.395
232962 92.841 -74.975 54.388
232963 92.871 -74.948 54.380
232964 92.901 -74.922 54.367
232965 92.932 -74.907 54.351
232966 92.966 -74.910 54.335
232967 92.997 -74.913 54.318
232968 93.023 -74.914 54.303
232969 93.049 -74.916 54.289
232970 93.075 -74.919 54.277
232971 93.100 -74.920 54.265
232972 93.127 -74.922 54.252
232973 93.152 -74.925 54.239
232974 93.178 -74.926 54.226
232975 93.204 -74.928 54.214
232976 93.230 -74.931 54.203
232977 93.255 -74.928 54.194
232978 93.277 -74.922 54.185
232979 93.300 -74.914 54.178
232980 93.323 -74.907 54.171
232981 93.346 -74.900 54.167
232982 93.369 -74.894 54.165
232983 93.392 -74.887 54.167
232984 93.415 -74.879 54.171
232985 93.438 -74.872 54.178
232986 93.461 -74.866 54.184
232987 93.484 -74.859 54.195
232988 93.506 -74.852 54.210
232989 93.530 -74.844 54.228
232990 93.552 -74.838 54.255
232991 93.580 -74.838 54.286
232992 93.609 -74.841 54.319
232993 93.638 -74.843 54.353
232994 93.666 -74.845 54.385
232995 93.696 -74.847 54.414
232996 93.724 -74.850 54.442
232997 93.757 -74.860 54.470
232998 93.790 -74.873 54.496
232999 93.823 -74.886 54.512
233000 93.847 -74.890 54.520
233001 93.865 -74.887 54.523
233002 93.883 -74.883 54.524
233003 93.900 -74.880 54.526
233004 93.922 -74.878 54.532
233005 93.947 -74.876 54.542
233006 93.974 -74.875 54.550
233007 94.000 -74.873 54.556
233008 94.026 -74.872 54.562
233009 94.054 -74.867 54.569
233010 94.082 -74.860 54.574
233011 94.111 -74.854 54.576
233012 94.139 -74.848 54.576
233013 94.166 -74.841 54.565
233014 94.193 -74.833 54.550
233015 94.218 -74.828 54.532
233016 94.240 -74.831 54.511
233017 94.261 -74.831 54.485
233018 94.282 -74.833 54.459
233019 94.303 -74.831 54.433
233020 94.325 -74.830 54.407
233021 94.346 -74.828 54.385
233022 94.368 -74.827 54.369
233023 94.390 -74.825 54.357
233024 94.411 -74.824 54.348
233025 94.427 -74.827 54.341
233026 94.440 -74.833 54.327
233027 94.453 -74.838 54.309
233028 94.466 -74.844 54.291
233029 94.479 -74.850 54.270
233030 94.492 -74.856 54.252
233031 94.505 -74.861 54.233
233032 94.522 -74.865 54.212
233033 94.541 -74.866 54.187
233034 94.560 -74.866 54.159
233035 94.579 -74.867 54.131
233036 94.598 -74.868 54.106
233037 94.618 -74.869 54.083
233038 94.638 -74.871 54.059
233039 94.659 -74.873 54.034
233040 94.678 -74.875 54.003
233041 94.699 -74.877 53.965
233042 94.724 -74.884 53.924
233043 94.753 -74.892 53.881
233044 94.780 -74.900 53.837
233045 94.809 -74.910 53.792
233046 94.836 -74.918 53.745
233047 94.864 -74.926 53.695
233048 94.892 -74.935 53.645
233049 94.907 -74.940 53.596
233050 94.922 -74.946 53.547
233051 94.938 -74.951 53.497
233052 94.954 -74.957 53.448
233053 94.969 -74.962 53.399
233054 94.984 -74.968 53.349
233055 94.995 -74.970 53.298
233056 94.989 -74.960 53.245
233057 94.983 -74.950 53.190
233058 94.978 -74.940 53.133
233059 94.978 -74.933 53.073
233060 95.010 -74.939 53.006
233061 95.042 -74.946 52.939
233062 95.073 -74.953 52.872
233063 95.106 -74.960 52.807
233064 95.137 -74.966 52.742
233065 95.169 -74.972 52.679
233066 95.197 -74.976 52.617
233067 95.222 -74.979 52.555
233068 95.248 -74.984 52.495
233069 95.275 -74.994 52.436
233070 95.303 -75.006 52.378
233071 95.330 -75.017 52.323
233072 95.358 -75.028 52.270
233073 95.385 -75.039 52.220
233074 95.406 -75.044 52.171
233075 95.421 -75.044 52.121
233076 95.435 -75.044 52.075
233077 95.450 -75.043 52.030
233078 95.464 -75.043 51.987
233079 95.478 -75.042 51.946
233080 95.492 -75.041 51.907
233081 95.507 -75.041 51.870
233082 95.521 -75.041 51.835
233083 95.539 -75.044 51.801
233084 95.574 -75.047 51.767
233085 95.615 -75.053 51.734
233086 95.656 -75.058 51.702
233087 95.696 -75.064 51.672
233088 95.737 -75.070 51.643
233089 95.778 -75.076 51.615
233090 95.819 -75.082 51.587
233091 95.860 -75.088 51.560
233092 95.901 -75.093 51.533
233093 95.941 -75.100 51.508
233094 95.976 -75.108 51.483
233095 96.012 -75.116 51.459
233096 96.048 -75.125 51.435
233097 96.083 -75.133 51.412
233098 96.119 -75.142 51.390
233099 96.155 -75.149 51.367
233100 96.186 -75.151 51.346
233101 96.218 -75.154 51.325
233102 96.249 -75.156 51.303
233103 96.281 -75.158 51.281
233104 96.312 -75.160 51.259
233105 96.343 -75.163 51.235
233106 96.371 -75.166 51.211
233107 96.398 -75.168 51.188
233108 96.426 -75.171 51.165
233109 96.453 -75.174 51.140
233110 96.480 -75.177 51.114
233111 96.500 -75.189 51.088
233112 96.518 -75.207 51.062
233113 96.536 -75.223 51.034
233114 96.553 -75.240 51.010
233115 96.571 -75.258 50.985
233116 96.593 -75.266 50.962
233117 96.616 -75.271 50.939
233118 96.639 -75.276 50.917
233119 96.662 -75.281 50.895
233120 96.685 -75.288 50.873
233121 96.710 -75.298 50.850
233122 96.735 -75.310 50.829
233123 96.760 -75.320 50.805
233124 96.785 -75.333 50.781
233125 96.798 -75.354 50.757
233126 96.811 -75.374 50.735
233127 96.825 -75.395 50.712
233128 96.838 -75.416 50.689
233129 96.855 -75.435 50.664
233130 96.879 -75.448 50.637
233131 96.902 -75.461 50.609
233132 96.927 -75.474 50.580
233133 96.951 -75.486 50.552
233134 96.974 -75.499 50.525
233135 96.999 -75.511 50.495
233136 97.023 -75.524 50.464
233137 97.061 -75.534 50.433
233138 97.125 -75.539 50.402
233139 97.175 -75.543 50.371
233140 97.199 -75.549 50.340
233141 97.224 -75.555 50.307
233142 97.248 -75.561 50.274
233143 97.272 -75.567 50.240
233144 97.296 -75.573 50.206
233145 97.321 -75.578 50.172
233146 97.345 -75.584 50.138
233147 97.363 -75.596 50.104
233148 97.380 -75.608 50.071
233149 97.397 -75.619 50.036
233150 97.414 -75.630 50.002
233151 97.437 -75.646 49.968
233152 97.459 -75.662 49.934
233153 97.485 -75.673 49.900
233154 97.517 -75.676 49.865
233155 97.548 -75.679 49.832
233156 97.581 -75.682 49.799
233157 97.612 -75.685 49.767
233158 97.644 -75.688 49.734
233159 97.676 -75.691 49.702
233160 97.708 -75.694 49.669
233161 97.740 -75.697 49.633
233162 97.772 -75.700 49.595
233163 97.804 -75.703 49.556
233164 97.837 -75.708 49.516
233165 97.870 -75.722 49.476
233166 97.899 -75.736 49.433
233167 97.902 -75.743 49.390
233168 97.907 -75.752 49.346
233169 97.926 -75.762 49.302
233170 97.945 -75.773 49.258
233171 97.959 -75.775 49.215
233172 97.991 -75.784 49.173
233173 98.009 -75.794 49.129
233174 98.030 -75.805 49.085
233175 98.049 -75.818 49.041
233176 98.067 -75.833 48.997
233177 98.085 -75.847 48.954
233178 98.104 -75.860 48.911
233179 98.122 -75.875 48.867
233180 98.141 -75.888 48.824
233181 98.186 -75.907 48.780
233182 98.232 -75.925 48.736
233183 98.278 -75.944 48.692
233184 98.322 -75.961 48.651
233185 98.362 -75.977 48.608
233186 98.402 -75.993 48.565
233187 98.442 -76.009 48.522
233188 98.482 -76.025 48.479
233189 98.522 -76.041 48.437
233190 98.536 -76.058 48.396
233191 98.548 -76.076 48.357
233192 98.560 -76.094 48.319
233193 98.586 -76.132 48.283
233194 98.627 -76.189 48.246
233195 98.663 -76.222 48.210
233196 98.695 -76.219 48.176
233197 98.727 -76.216 48.142
233198 98.755 -76.215 48.108
233199 98.780 -76.214 48.074
233200 98.806 -76.213 48.040
233201 98.832 -76.212 48.007
233202 98.857 -76.211 47.977
233203 98.883 -76.210 47.949
233204 98.909 -76.209 47.923
233205 98.934 -76.208 47.898
233206 98.959 -76.207 47.874
233207 98.985 -76.206 47.851
233208 99.011 -76.205 47.831
233209 99.037 -76.204 47.812
233210 99.063 -76.206 47.794
233211 99.088 -76.208 47.777
233212 99.115 -76.211 47.757
233213 99.139 -76.223 47.736
233214 99.163 -76.233 47.714
233215 99.187 -76.244 47.691
233216 99.211 -76.255 47.668
233217 99.235 -76.266 47.645
233218 99.260 -76.277 47.621
233219 99.284 -76.288 47.596
233220 99.308 -76.298 47.570
233221 99.332 -76.310 47.544
233222 99.354 -76.318 47.517
233223 99.362 -76.315 47.488
233224 99.370 -76.312 47.459
233225 99.378 -76.310 47.429
233226 99.394 -76.307 47.398
233227 99.416 -76.307 47.369
233228 99.438 -76.306 47.340
233229 99.455 -76.307 47.310
233230 99.470 -76.309 47.279
233231 99.485 -76.311 47.248
233232 99.500 -76.314 47.216
233233 99.516 -76.316 47.185
233234 99.531 -76.318 47.154
233235 99.555 -76.326 47.124
233236 99.584 -76.336 47.095
233237 99.614 -76.345 47.066
233238 99.642 -76.355 47.037
233239 99.666 -76.369 47.005
233240 99.698 -76.385 46.973
233241 99.730 -76.401 46.941
233242 99.762 -76.417 46.911
233243 99.794 -76.434 46.880
233244 99.826 -76.450 46.850
233245 99.858 -76.467 46.819
233246 99.890 -76.483 46.790
233247 99.918 -76.495 46.758
233248 99.938 -76.498 46.729
233249 99.958 -76.502 46.700
233250 99.978 -76.506 46.670
233251 99.996 -76.519 46.639
233252 100.014 -76.533 46.606
233253 100.033 -76.546 46.574
233254 100.051 -76.558 46.540
233255 100.069 -76.572 46.507
233256 100.088 -76.585 46.474
233257 100.106 -76.598 46.439
233258 100.125 -76.611 46.404
233259 100.143 -76.624 46.369
233260 100.161 -76.637 46.333
233261 100.198 -76.643 46.297
233262 100.227 -76.652 46.263
233263 100.245 -76.665 46.229
233264 100.262 -76.674 46.195
233265 100.281 -76.662 46.161
233266 100.300 -76.650 46.128
233267 100.319 -76.637 46.095
233268 100.338 -76.625 46.062
233269 100.357 -76.613 46.031
233270 100.376 -76.621 46.001
233271 100.396 -76.629 45.971
233272 100.415 -76.636 45.943
233273 100.435 -76.645 45.915
233274 100.455 -76.652 45.888
233275 100.474 -76.660 45.862
233276 100.494 -76.668 45.838
233277 100.513 -76.676 45.815
233278 100.533 -76.684 45.792
233279 100.552 -76.701 45.770
233280 100.578 -76.712 45.749
233281 100.623 -76.707 45.728
233282 100.662 -76.705 45.708
233283 100.687 -76.715 45.688
233284 100.710 -76.724 45.669
233285 100.735 -76.733 45.650
233286 100.758 -76.742 45.630
233287 100.782 -76.751 45.610
233288 100.806 -76.760 45.590
233289 100.830 -76.769 45.571
233290 100.854 -76.778 45.551
233291 100.880 -76.789 45.531
233292 100.908 -76.799 45.510
233293 100.936 -76.809 45.491
233294 100.964 -76.820 45.472
233295 100.992 -76.831 45.454
233296 101.017 -76.841 45.435
233297 101.032 -76.853 45.419
233298 101.047 -76.865 45.402
233299 101.061 -76.878 45.387
233300 101.076 -76.890 45.373
233301 101.091 -76.902 45.360
233302 101.106 -76.913 45.349
233303 101.120 -76.925 45.338
233304 101.135 -76.937 45.326
233305 101.149 -76.950 45.315
233306 101.165 -76.962 45.303
233307 101.179 -76.974 45.291
233308 101.207 -76.996 45.279
233309 101.236 -77.018 45.265
233310 101.265 -77.041 45.252
233311 101.294 -77.063 45.238
233312 101.323 -77.085 45.224
233313 101.352 -77.108 45.210
233314 101.376 -77.127 45.195
233315 101.392 -77.140 45.179
233316 101.406 -77.152 45.164
233317 101.421 -77.165 45.148
233318 101.441 -77.170 45.132
233319 101.463 -77.170 45.116
233320 101.486 -77.171 45.099
233321 101.509 -77.171 45.080
233322 101.532 -77.171 45.062
233323 101.551 -77.175 45.044
233324 101.570 -77.182 45.027
233325 101.588 -77.189 45.008
233326 101.607 -77.195 44.990
233327 101.625 -77.202 44.972
233328 101.644 -77.208 44.953
233329 101.663 -77.215 44.936
233330 101.681 -77.222 44.919
233331 101.699 -77.228 44.903
233332 101.718 -77.235 44.887
233333 101.737 -77.241 44.871
233334 101.755 -77.248 44.855
233335 101.774 -77.255 44.840
233336 101.793 -77.269 44.823
233337 101.813 -77.282 44.806
233338 101.833 -77.297 44.789
233339 101.856 -77.320 44.772
233340 101.880 -77.344 44.755
233341 101.903 -77.368 44.738
233342 101.926 -77.392 44.721
233343 101.947 -77.415 44.705
233344 101.967 -77.437 44.689
233345 101.986 -77.460 44.673
233346 102.006 -77.483 44.658
233347 102.026 -77.505 44.640
233348 102.055 -77.524 44.623
233349 102.103 -77.533 44.607
233350 102.123 -77.542 44.592
233351 102.144 -77.550 44.575
233352 102.164 -77.559 44.560
233353 102.184 -77.568 44.544
233354 102.204 -77.577 44.529
233355 102.225 -77.586 44.512
233356 102.248 -77.595 44.497
233357 102.280 -77.607 44.482
233358 102.313 -77.618 44.468
233359 102.345 -77.630 44.452
233360 102.362 -77.649 44.436
233361 102.377 -77.671 44.420
233362 102.399 -77.679 44.403
233363 102.425 -77.683 44.388
233364 102.451 -77.687 44.372
233365 102.476 -77.692 44.356
233366 102.501 -77.696 44.339
233367 102.527 -77.700 44.321
233368 102.553 -77.704 44.304
233369 102.579 -77.708 44.287
233370 102.613 -77.713 44.271
233371 102.650 -77.719 44.254
233372 102.686 -77.724 44.237
233373 102.722 -77.730 44.219
233374 102.759 -77.735 44.203
233375 102.795 -77.740 44.186
233376 102.831 -77.746 44.170
233377 102.868 -77.751 44.153
233378 102.904 -77.756 44.136
233379 102.940 -77.761 44.121
233380 102.976 -77.767 44.104
233381 103.013 -77.772 44.087
233382 103.049 -77.777 44.072
233383 103.086 -77.783 44.055
233384 103.122 -77.788 44.039
233385 103.158 -77.793 44.023
233386 103.194 -77.799 44.007
233640 92.209 -74.861 54.321
233641 92.233 -74.872 54.329
233642 92.258 -74.883 54.338
233643 92.282 -74.894 54.349
233644 92.306 -74.904 54.359
233645 92.331 -74.915 54.371
233646 92.355 -74.926 54.384
233647 92.379 -74.937 54.399
233648 92.404 -74.947 54.413
233649 92.428 -74.958 54.426
233650 92.449 -74.957 54.438
233651 92.463 -74.936 54.448
233652 92.493 -74.928 54.456
233653 92.525 -74.925 54.464
233654 92.559 -74.922 54.469
233655 92.592 -74.918 54.472
233656 92.625 -74.914 54.468
233657 92.658 -74.911 54.460
233658 92.690 -74.909 54.449
233659 92.722 -74.906 54.438
233660 92.754 -74.906 54.429
233661 92.782 -74.920 54.424
233662 92.810 -74.935 54.418
233663 92.840 -74.919 54.410
233664 92.870 -74.892 54.401
233665 92.902 -74.882 54.387
233666 92.935 -74.885 54.369
233667 92.969 -74.888 54.352
233668 93.003 -74.892 54.336
233669 93.034 -74.894 54.321
233670 93.060 -74.897 54.308
233671 93.085 -74.899 54.295
233672 93.111 -74.900 54.284
233673 93.137 -74.903 54.272
233674 93.163 -74.905 54.259
233675 93.189 -74.907 54.247
233676 93.213 -74.906 54.236
233677 93.236 -74.898 54.226
233678 93.259 -74.891 54.216
233679 93.282 -74.885 54.209
233680 93.305 -74.878 54.201
233681 93.328 -74.871 54.195
233682 93.351 -74.863 54.192
233683 93.374 -74.856 54.191
233684 93.397 -74.850 54.192
233685 93.420 -74.843 54.196
233686 93.443 -74.836 54.201
233687 93.466 -74.828 54.208
233688 93.489 -74.822 54.218
233689 93.511 -74.815 54.232
233690 93.535 -74.808 54.249
233691 93.557 -74.801 54.272
233692 93.585 -74.801 54.300
233693 93.614 -74.803 54.330
233694 93.643 -74.806 54.362
233695 93.672 -74.808 54.395
233696 93.700 -74.810 54.426
233697 93.732 -74.819 54.455
233698 93.765 -74.832 54.482
233699 93.798 -74.845 54.507
233700 93.831 -74.858 54.523
233701 93.860 -74.866 54.531
233702 93.877 -74.863 54.533
233703 93.895 -74.859 54.533
233704 93.913 -74.856 54.536
233705 93.931 -74.853 54.542
233706 93.948 -74.850 54.553
233707 93.973 -74.849 54.561
233708 93.998 -74.847 54.568
233709 94.025 -74.844 54.575
233710 94.054 -74.838 54.581
233711 94.081 -74.831 54.585
233712 94.109 -74.824 54.588
233713 94.135 -74.816 54.586
233714 94.162 -74.809 54.576
233715 94.189 -74.802 54.561
233716 94.210 -74.803 54.542
233717 94.232 -74.806 54.521
233718 94.253 -74.807 54.495
233719 94.274 -74.809 54.468
233720 94.296 -74.810 54.440
233721 94.318 -74.809 54.410
233722 94.339 -74.808 54.386
233723 94.361 -74.806 54.367
233724 94.383 -74.805 54.351
233725 94.403 -74.804 54.338
233726 94.416 -74.810 54.329
233727 94.429 -74.816 54.315
233728 94.442 -74.822 54.298
233729 94.455 -74.828 54.278
233730 94.469 -74.832 54.259
233731 94.486 -74.836 54.241
233732 94.503 -74.839 54.222
233733 94.521 -74.842 54.202
233734 94.542 -74.844 54.178
233735 94.561 -74.844 54.152
233736 94.580 -74.844 54.126
233737 94.600 -74.844 54.102
233738 94.619 -74.844 54.083
233739 94.638 -74.844 54.064
233740 94.657 -74.844 54.042
233741 94.676 -74.844 54.013
233742 94.704 -74.852 53.980
233743 94.732 -74.860 53.941
233744 94.760 -74.869 53.898
233745 94.788 -74.877 53.853
233746 94.816 -74.885 53.808
233747 94.844 -74.894 53.760
233748 94.872 -74.903 53.710
233749 94.900 -74.911 53.661
233750 94.921 -74.918 53.611
233751 94.936 -74.923 53.561
233752 94.952 -74.929 53.512
233753 94.967 -74.935 53.463
233754 94.983 -74.940 53.415
233755 94.998 -74.946 53.366
233756 95.000 -74.941 53.316
233757 94.994 -74.932 53.262
233758 94.988 -74.922 53.206
233759 94.982 -74.911 53.146
233760 94.983 -74.904 53.084
233761 95.015 -74.911 53.016
233762 95.047 -74.918 52.949
233763 95.079 -74.925 52.882
233764 95.111 -74.931 52.817
233765 95.143 -74.938 52.753
233766 95.175 -74.944 52.690
233767 95.206 -74.950 52.627
233768 95.232 -74.953 52.566
233769 95.258 -74.960 52.506
233770 95.285 -74.971 52.447
233771 95.313 -74.982 52.389
233772 95.341 -74.994 52.332
233773 95.368 -75.004 52.278
233774 95.396 -75.016 52.226
233775 95.423 -75.026 52.176
233776 95.437 -75.026 52.126
233777 95.451 -75.026 52.079
233778 95.466 -75.025 52.034
233779 95.480 -75.025 51.989
233780 95.494 -75.024 51.947
233781 95.509 -75.023 51.909
233782 95.523 -75.023 51.872
233783 95.540 -75.019 51.836
233784 95.557 -75.012 51.802
233785 95.589 -75.024 51.768
233786 95.630 -75.030 51.735
233787 95.671 -75.036 51.703
233788 95.711 -75.041 51.673
233789 95.752 -75.048 51.644
233790 95.794 -75.054 51.615
233791 95.834 -75.059 51.587
233792 95.875 -75.065 51.559
233793 95.916 -75.071 51.532
233794 95.955 -75.077 51.507
233795 95.990 -75.085 51.483
233796 96.026 -75.094 51.458
233797 96.062 -75.102 51.435
233798 96.098 -75.110 51.413
233799 96.133 -75.119 51.391
233800 96.165 -75.121 51.370
233801 96.196 -75.123 51.349
233802 96.228 -75.125 51.329
233803 96.259 -75.127 51.307
233804 96.291 -75.129 51.286
233805 96.322 -75.132 51.264
233806 96.354 -75.134 51.241
233807 96.383 -75.136 51.218
233808 96.411 -75.139 51.196
233809 96.438 -75.142 51.173
233810 96.466 -75.145 51.148
233811 96.488 -75.154 51.122
233812 96.503 -75.170 51.096
233813 96.520 -75.187 51.070
233814 96.538 -75.204 51.043
233815 96.556 -75.220 51.019
233816 96.576 -75.234 50.995
233817 96.598 -75.239 50.971
233818 96.621 -75.244 50.949
233819 96.644 -75.249 50.927
233820 96.669 -75.259 50.905
233821 96.694 -75.270 50.883
233822 96.719 -75.281 50.861
233823 96.744 -75.292 50.839
233824 96.769 -75.303 50.816
233825 96.795 -75.314 50.793
233826 96.813 -75.331 50.770
233827 96.826 -75.352 50.748
233828 96.840 -75.373 50.726
233829 96.854 -75.394 50.703
233830 96.876 -75.408 50.678
233831 96.900 -75.420 50.651
233832 96.924 -75.433 50.623
233833 96.948 -75.446 50.595
233834 96.972 -75.459 50.566
233835 96.996 -75.472 50.537
233836 97.020 -75.485 50.507
233837 97.044 -75.498 50.476
233838 97.071 -75.510 50.445
233839 97.136 -75.514 50.414
233840 97.185 -75.519 50.383
233841 97.209 -75.525 50.351
233842 97.234 -75.530 50.318
233843 97.258 -75.536 50.285
233844 97.282 -75.543 50.251
233845 97.306 -75.549 50.217
233846 97.331 -75.554 50.183
233847 97.348 -75.566 50.151
233848 97.365 -75.577 50.118
233849 97.383 -75.589 50.085
233850 97.400 -75.601 50.052
233851 97.417 -75.612 50.018
233852 97.440 -75.627 49.984
233853 97.464 -75.641 49.949
233854 97.496 -75.644 49.916
233855 97.528 -75.647 49.881
233856 97.560 -75.650 49.847
233857 97.591 -75.653 49.815
233858 97.624 -75.656 49.782
233859 97.655 -75.659 49.749
233860 97.687 -75.662 49.716
233861 97.719 -75.665 49.683
233862 97.743 -75.669 49.646
233863 97.768 -75.673 49.608
233864 97.793 -75.677 49.568
233865 97.819 -75.683 49.527
233866 97.852 -75.695 49.486
233867 97.878 -75.706 49.442
233868 97.897 -75.717 49.398
233869 97.917 -75.728 49.352
233870 97.936 -75.739 49.308
233871 97.954 -75.748 49.263
233872 97.978 -75.752 49.219
233873 98.013 -75.761 49.176
233874 98.030 -75.771 49.132
233875 98.049 -75.782 49.088
233876 98.070 -75.792 49.043
233877 98.090 -75.803 48.999
233878 98.109 -75.818 48.956
233879 98.127 -75.831 48.913
233880 98.146 -75.846 48.869
233881 98.164 -75.859 48.825
233882 98.196 -75.875 48.781
233883 98.242 -75.894 48.737
233884 98.288 -75.912 48.693
233885 98.330 -75.929 48.650
233886 98.370 -75.945 48.607
233887 98.410 -75.961 48.564
233888 98.450 -75.977 48.521
233889 98.490 -75.993 48.479
233890 98.530 -76.009 48.436
233891 98.552 -76.026 48.395
233892 98.564 -76.044 48.355
233893 98.577 -76.063 48.316
233894 98.617 -76.120 48.279
233895 98.657 -76.178 48.242
233896 98.693 -76.200 48.206
233897 98.724 -76.197 48.171
233898 98.752 -76.195 48.136
233899 98.778 -76.195 48.102
233900 98.803 -76.194 48.068
233901 98.829 -76.192 48.034
233902 98.855 -76.191 48.001
233903 98.880 -76.191 47.970
233904 98.906 -76.189 47.941
233905 98.932 -76.188 47.915
233906 98.957 -76.188 47.889
233907 98.983 -76.186 47.865
233908 99.008 -76.185 47.841
233909 99.034 -76.185 47.821
233910 99.060 -76.183 47.801
233911 99.085 -76.184 47.782
233912 99.112 -76.186 47.764
233913 99.136 -76.194 47.745
233914 99.160 -76.204 47.725
233915 99.185 -76.216 47.703
233916 99.209 -76.226 47.681
233917 99.233 -76.237 47.657
233918 99.257 -76.248 47.633
233919 99.281 -76.259 47.609
233920 99.305 -76.270 47.584
233921 99.330 -76.281 47.557
233922 99.354 -76.292 47.531
233923 99.378 -76.303 47.505
233924 99.389 -76.302 47.479
233925 99.397 -76.299 47.450
233926 99.406 -76.296 47.421
233927 99.428 -76.295 47.391
233928 99.450 -76.295 47.362
233929 99.466 -76.296 47.334
233930 99.481 -76.298 47.304
233931 99.496 -76.301 47.274
233932 99.511 -76.303 47.243
233933 99.526 -76.305 47.212
233934 99.542 -76.307 47.181
233935 99.557 -76.310 47.150
233936 99.572 -76.312 47.121
233937 99.602 -76.322 47.091
233938 99.631 -76.332 47.062
233939 99.659 -76.342 47.032
233940 99.679 -76.355 47.000
233941 99.705 -76.371 46.969
233942 99.737 -76.387 46.938
233943 99.769 -76.404 46.907
233944 99.801 -76.420 46.877
233945 99.833 -76.436 46.846
233946 99.865 -76.453 46.816
233947 99.897 -76.469 46.786
233948 99.919 -76.474 46.755
233949 99.939 -76.478 46.726
233950 99.959 -76.481 46.697
233951 99.979 -76.486 46.667
233952 99.997 -76.499 46.636
233953 100.016 -76.511 46.603
233954 100.034 -76.525 46.570
233955 100.052 -76.538 46.537
233956 100.071 -76.551 46.503
233957 100.089 -76.564 46.469
233958 100.108 -76.577 46.436
233959 100.126 -76.590 46.401
233960 100.149 -76.602 46.365
233961 100.187 -76.607 46.329
233962 100.225 -76.612 46.293
233963 100.247 -76.624 46.259
233964 100.264 -76.636 46.224
233965 100.281 -76.649 46.190
233966 100.298 -76.661 46.156
233967 100.318 -76.651 46.122
233968 100.337 -76.638 46.089
233969 100.355 -76.626 46.056
233970 100.374 -76.614 46.025
233971 100.394 -76.612 45.994
233972 100.413 -76.620 45.965
233973 100.433 -76.628 45.936
233974 100.453 -76.636 45.908
233975 100.472 -76.644 45.881
233976 100.491 -76.652 45.855
233977 100.511 -76.659 45.830
233978 100.531 -76.668 45.807
233979 100.550 -76.682 45.784
233980 100.569 -76.701 45.763
233981 100.604 -76.704 45.742
233982 100.649 -76.699 45.721
233983 100.694 -76.693 45.702
233984 100.717 -76.702 45.682
233985 100.741 -76.712 45.663
233986 100.765 -76.721 45.644
233987 100.789 -76.730 45.623
233988 100.813 -76.739 45.603
233989 100.837 -76.748 45.583
233990 100.861 -76.757 45.563
233991 100.885 -76.766 45.543
233992 100.911 -76.777 45.523
233993 100.939 -76.787 45.502
233994 100.967 -76.797 45.483
233995 100.995 -76.808 45.464
233996 101.022 -76.818 45.445
233997 101.045 -76.830 45.426
233998 101.060 -76.842 45.409
233999 101.075 -76.853 45.393
234000 101.090 -76.865 45.378
234001 101.104 -76.878 45.364
234002 101.119 -76.890 45.351
234003 101.134 -76.902 45.340
234004 101.148 -76.914 45.329
234005 101.163 -76.926 45.317
234006 101.178 -76.937 45.306
234007 101.193 -76.950 45.294
234008 101.207 -76.962 45.282
234009 101.222 -76.974 45.269
234010 101.246 -76.993 45.256
234011 101.275 -77.016 45.242
234012 101.304 -77.038 45.229
234013 101.333 -77.060 45.215
234014 101.362 -77.082 45.201
234015 101.391 -77.105 45.186
234016 101.406 -77.117 45.171
234017 101.421 -77.130 45.156
234018 101.435 -77.142 45.141
234019 101.457 -77.145 45.125
234020 101.479 -77.145 45.109
234021 101.502 -77.145 45.091
234022 101.525 -77.145 45.072
234023 101.548 -77.145 45.054
234024 101.568 -77.149 45.037
234025 101.587 -77.156 45.019
234026 101.605 -77.162 45.000
234027 101.624 -77.169 44.981
234028 101.642 -77.175 44.964
234029 101.661 -77.182 44.946
234030 101.680 -77.188 44.930
234031 101.698 -77.194 44.913
234032 101.716 -77.201 44.897
234033 101.735 -77.208 44.881
234034 101.754 -77.214 44.866
234035 101.773 -77.226 44.850
234036 101.793 -77.240 44.834
234037 101.813 -77.254 44.817
234038 101.832 -77.267 44.800
234039 101.852 -77.282 44.783
234040 101.876 -77.306 44.767
234041 101.899 -77.330 44.750
234042 101.922 -77.354 44.732
234043 101.945 -77.377 44.715
234044 101.969 -77.401 44.699
234045 101.992 -77.425 44.683
234046 102.013 -77.448 44.667
234047 102.032 -77.470 44.652
234048 102.052 -77.493 44.635
234049 102.085 -77.511 44.618
234050 102.115 -77.520 44.602
234051 102.135 -77.528 44.586
234052 102.156 -77.537 44.569
234053 102.176 -77.546 44.553
234054 102.196 -77.555 44.538
234055 102.217 -77.564 44.522
234056 102.237 -77.573 44.506
234057 102.257 -77.581 44.490
234058 102.287 -77.592 44.475
234059 102.320 -77.604 44.460
234060 102.352 -77.616 44.444
234061 102.369 -77.636 44.428
234062 102.384 -77.655 44.411
234063 102.410 -77.660 44.395
234064 102.436 -77.664 44.379
234065 102.461 -77.668 44.364
234066 102.486 -77.672 44.347
234067 102.512 -77.677 44.330
234068 102.538 -77.680 44.312
234069 102.565 -77.685 44.294
234070 102.602 -77.690 44.277
234071 102.638 -77.696 44.260
234072 102.674 -77.701 44.244
234073 102.711 -77.706 44.227
234074 102.747 -77.711 44.210
234075 102.783 -77.717 44.193
234076 102.820 -77.722 44.177
234077 102.856 -77.728 44.161
234078 102.892 -77.733 44.145
234079 102.929 -77.739 44.128
234080 102.965 -77.744 44.112
234081 103.001 -77.749 44.095
234082 103.037 -77.755 44.078
234083 103.074 -77.760 44.063
234084 103.110 -77.765 44.047
234085 103.147 -77.771 44.031
234086 103.183 -77.776 44.014
234341 92.211 -74.829 54.325
234342 92.235 -74.840 54.333
234343 92.260 -74.850 54.342
234344 92.284 -74.861 54.351
234345 92.308 -74.872 54.362
234346 92.333 -74.883 54.372
234347 92.357 -74.894 54.385
234348 92.381 -74.904 54.399
234349 92.406 -74.915 54.413
234350 92.430 -74.925 54.426
234351 92.453 -74.933 54.438
234352 92.466 -74.910 54.448
234353 92.492 -74.900 54.455
234354 92.521 -74.892 54.463
234355 92.554 -74.889 54.470
234356 92.587 -74.885 54.474
234357 92.621 -74.881 54.473
234358 92.653 -74.878 54.469
234359 92.685 -74.876 54.462
234360 92.717 -74.873 54.454
234361 92.749 -74.870 54.448
234362 92.780 -74.873 54.444
234363 92.808 -74.888 54.440
234364 92.839 -74.863 54.431
234365 92.871 -74.858 54.420
234366 92.905 -74.861 54.404
234367 92.939 -74.864 54.386
234368 92.972 -74.868 54.369
234369 93.006 -74.871 54.354
234370 93.040 -74.874 54.341
234371 93.070 -74.877 54.328
234372 93.096 -74.879 54.316
234373 93.122 -74.881 54.304
234374 93.148 -74.883 54.293
234375 93.172 -74.882 54.281
234376 93.196 -74.875 54.269
234377 93.218 -74.869 54.258
234378 93.241 -74.862 54.249
234379 93.264 -74.854 54.239
234380 93.287 -74.847 54.233
234381 93.310 -74.841 54.226
234382 93.333 -74.834 54.221
234383 93.356 -74.827 54.218
234384 93.379 -74.819 54.216
234385 93.402 -74.812 54.218
234386 93.425 -74.806 54.222
234387 93.448 -74.799 54.228
234388 93.471 -74.792 54.233
234389 93.494 -74.784 54.243
234390 93.516 -74.778 54.255
234391 93.540 -74.771 54.270
234392 93.562 -74.764 54.290
234393 93.590 -74.765 54.314
234394 93.619 -74.767 54.341
234395 93.648 -74.769 54.368
234396 93.677 -74.771 54.401
234397 93.708 -74.778 54.433
234398 93.741 -74.791 54.465
234399 93.774 -74.804 54.495
234400 93.806 -74.817 54.518
234401 93.839 -74.829 54.535
234402 93.872 -74.842 54.544
234403 93.891 -74.839 54.545
234404 93.908 -74.836 54.543
234405 93.926 -74.833 54.546
234406 93.944 -74.830 54.553
234407 93.962 -74.827 54.563
234408 93.979 -74.824 54.571
234409 93.997 -74.821 54.579
234410 94.024 -74.815 54.586
234411 94.051 -74.807 54.591
234412 94.078 -74.800 54.597
234413 94.105 -74.792 54.599
234414 94.131 -74.784 54.598
234415 94.158 -74.777 54.588
234416 94.181 -74.777 54.573
234417 94.203 -74.779 54.555
234418 94.224 -74.781 54.534
234419 94.246 -74.783 54.507
234420 94.267 -74.784 54.479
234421 94.289 -74.786 54.449
234422 94.311 -74.787 54.417
234423 94.333 -74.787 54.391
234424 94.354 -74.786 54.369
234425 94.376 -74.784 54.349
234426 94.392 -74.787 54.332
234427 94.405 -74.794 54.319
234428 94.418 -74.799 54.303
234429 94.433 -74.803 54.285
234430 94.451 -74.806 54.265
234431 94.468 -74.809 54.246
234432 94.485 -74.812 54.227
234433 94.503 -74.816 54.209
234434 94.522 -74.819 54.190
234435 94.543 -74.821 54.166
234436 94.563 -74.822 54.141
234437 94.582 -74.822 54.116
234438 94.601 -74.822 54.094
234439 94.620 -74.822 54.077
234440 94.639 -74.822 54.061
234441 94.659 -74.822 54.042
234442 94.683 -74.825 54.016
234443 94.711 -74.831 53.985
234444 94.739 -74.838 53.948
234445 94.767 -74.845 53.908
234446 94.795 -74.853 53.864
234447 94.823 -74.862 53.819
234448 94.851 -74.871 53.771
234449 94.879 -74.879 53.722
234450 94.907 -74.888 53.672
234451 94.934 -74.896 53.622
234452 94.950 -74.901 53.573
234453 94.966 -74.907 53.524
234454 94.981 -74.913 53.475
234455 94.996 -74.918 53.428
234456 95.011 -74.922 53.380
234457 95.004 -74.913 53.329
234458 94.998 -74.903 53.275
234459 94.993 -74.893 53.218
234460 94.987 -74.883 53.157
234461 94.989 -74.876 53.093
234462 95.021 -74.883 53.024
234463 95.053 -74.890 52.956
234464 95.085 -74.896 52.889
234465 95.117 -74.903 52.824
234466 95.148 -74.910 52.761
234467 95.181 -74.916 52.699
234468 95.212 -74.922 52.637
234469 95.242 -74.927 52.577
234470 95.268 -74.937 52.518
234471 95.296 -74.947 52.459
234472 95.324 -74.959 52.400
234473 95.351 -74.970 52.341
234474 95.379 -74.981 52.285
234475 95.406 -74.992 52.232
234476 95.433 -75.004 52.182
234477 95.453 -75.007 52.132
234478 95.468 -75.007 52.084
234479 95.482 -75.007 52.038
234480 95.496 -75.006 51.992
234481 95.512 -75.004 51.950
234482 95.529 -74.997 51.911
234483 95.547 -74.991 51.873
234484 95.565 -74.985 51.837
234485 95.583 -74.978 51.803
234486 95.605 -74.977 51.769
234487 95.645 -75.003 51.736
234488 95.686 -75.013 51.704
234489 95.727 -75.019 51.675
234490 95.768 -75.025 51.645
234491 95.809 -75.031 51.616
234492 95.849 -75.037 51.588
234493 95.890 -75.042 51.560
234494 95.931 -75.048 51.532
234495 95.969 -75.055 51.507
234496 96.005 -75.063 51.483
234497 96.041 -75.072 51.458
234498 96.077 -75.080 51.435
234499 96.112 -75.088 51.413
234500 96.143 -75.090 51.392
234501 96.175 -75.092 51.371
234502 96.206 -75.095 51.351
234503 96.238 -75.097 51.332
234504 96.269 -75.099 51.312
234505 96.301 -75.101 51.290
234506 96.332 -75.103 51.268
234507 96.364 -75.105 51.247
234508 96.395 -75.107 51.225
234509 96.424 -75.110 51.204
234510 96.451 -75.113 51.181
234511 96.476 -75.119 51.156
234512 96.491 -75.135 51.131
234513 96.507 -75.151 51.104
234514 96.523 -75.167 51.077
234515 96.541 -75.184 51.052
234516 96.558 -75.201 51.028
234517 96.581 -75.207 51.004
234518 96.604 -75.212 50.982
234519 96.627 -75.220 50.960
234520 96.653 -75.231 50.938
234521 96.678 -75.242 50.916
234522 96.703 -75.253 50.894
234523 96.729 -75.264 50.871
234524 96.754 -75.275 50.850
234525 96.779 -75.286 50.828
234526 96.804 -75.297 50.806
234527 96.828 -75.309 50.784
234528 96.841 -75.330 50.762
234529 96.855 -75.351 50.741
234530 96.873 -75.368 50.718
234531 96.897 -75.381 50.693
234532 96.921 -75.394 50.665
234533 96.945 -75.407 50.638
234534 96.969 -75.420 50.610
234535 96.993 -75.433 50.581
234536 97.017 -75.446 50.550
234537 97.041 -75.458 50.520
234538 97.065 -75.471 50.489
234539 97.089 -75.483 50.459
234540 97.145 -75.489 50.427
234541 97.195 -75.495 50.395
234542 97.219 -75.500 50.363
234543 97.244 -75.506 50.330
234544 97.268 -75.512 50.296
234545 97.292 -75.518 50.263
234546 97.316 -75.524 50.230
234547 97.333 -75.536 50.197
234548 97.351 -75.547 50.165
234549 97.368 -75.559 50.133
234550 97.385 -75.571 50.101
234551 97.403 -75.582 50.069
234552 97.420 -75.594 50.035
234553 97.443 -75.609 50.001
234554 97.475 -75.612 49.967
234555 97.507 -75.615 49.932
234556 97.539 -75.618 49.898
234557 97.571 -75.621 49.864
234558 97.597 -75.624 49.830
234559 97.621 -75.628 49.797
234560 97.646 -75.632 49.764
234561 97.671 -75.637 49.731
234562 97.696 -75.640 49.696
234563 97.720 -75.644 49.660
234564 97.745 -75.648 49.621
234565 97.770 -75.652 49.580
234566 97.796 -75.658 49.539
234567 97.828 -75.670 49.497
234568 97.886 -75.675 49.453
234569 97.908 -75.693 49.408
234570 97.927 -75.705 49.360
234571 97.947 -75.715 49.315
234572 97.963 -75.724 49.270
234573 97.996 -75.729 49.226
234574 98.035 -75.738 49.182
234575 98.053 -75.749 49.137
234576 98.070 -75.759 49.092
234577 98.088 -75.769 49.047
234578 98.110 -75.780 49.004
234579 98.131 -75.790 48.960
234580 98.151 -75.803 48.916
234581 98.169 -75.816 48.873
234582 98.187 -75.831 48.829
234583 98.206 -75.844 48.784
234584 98.251 -75.862 48.739
234585 98.297 -75.881 48.695
234586 98.339 -75.897 48.650
234587 98.378 -75.913 48.607
234588 98.419 -75.929 48.564
234589 98.458 -75.945 48.522
234590 98.499 -75.961 48.479
234591 98.538 -75.977 48.436
234592 98.568 -75.994 48.395
234593 98.579 -76.011 48.354
234594 98.607 -76.051 48.314
234595 98.648 -76.109 48.277
234596 98.688 -76.166 48.240
234597 98.722 -76.177 48.203
234598 98.749 -76.176 48.167
234599 98.775 -76.175 48.132
234600 98.801 -76.174 48.097
234601 98.827 -76.173 48.064
234602 98.852 -76.172 48.030
234603 98.878 -76.171 47.996
234604 98.904 -76.170 47.964
234605 98.929 -76.169 47.935
234606 98.954 -76.168 47.908
234607 98.980 -76.167 47.881
234608 99.006 -76.166 47.857
234609 99.031 -76.165 47.834
234610 99.057 -76.164 47.812
234611 99.083 -76.163 47.792
234612 99.109 -76.162 47.771
234613 99.134 -76.165 47.751
234614 99.158 -76.176 47.731
234615 99.183 -76.187 47.711
234616 99.207 -76.198 47.690
234617 99.231 -76.209 47.668
234618 99.255 -76.220 47.645
234619 99.279 -76.230 47.622
234620 99.303 -76.242 47.598
234621 99.328 -76.252 47.574
234622 99.352 -76.263 47.547
234623 99.376 -76.274 47.520
234624 99.400 -76.285 47.495
234625 99.416 -76.289 47.469
234626 99.424 -76.285 47.441
234627 99.440 -76.284 47.412
234628 99.462 -76.282 47.383
234629 99.477 -76.285 47.355
234630 99.492 -76.287 47.327
234631 99.507 -76.289 47.298
234632 99.522 -76.292 47.268
234633 99.538 -76.294 47.237
234634 99.553 -76.296 47.207
234635 99.568 -76.298 47.176
234636 99.583 -76.301 47.145
234637 99.598 -76.303 47.116
234638 99.619 -76.308 47.087
234639 99.648 -76.318 47.057
234640 99.676 -76.328 47.027
234641 99.693 -76.342 46.995
234642 99.712 -76.357 46.964
234643 99.744 -76.373 46.932
234644 99.776 -76.390 46.902
234645 99.808 -76.406 46.872
234646 99.840 -76.423 46.842
234647 99.872 -76.439 46.812
234648 99.900 -76.451 46.781
234649 99.920 -76.454 46.751
234650 99.940 -76.458 46.722
234651 99.960 -76.461 46.693
234652 99.980 -76.464 46.663
234653 99.999 -76.478 46.631
234654 100.017 -76.491 46.599
234655 100.035 -76.504 46.566
234656 100.054 -76.517 46.533
234657 100.072 -76.530 46.499
234658 100.090 -76.543 46.465
234659 100.109 -76.556 46.432
234660 100.138 -76.565 46.397
234661 100.176 -76.571 46.360
234662 100.213 -76.577 46.324
234663 100.250 -76.582 46.289
234664 100.267 -76.595 46.254
234665 100.284 -76.608 46.219
234666 100.301 -76.621 46.184
234667 100.319 -76.633 46.149
234668 100.336 -76.646 46.115
234669 100.354 -76.639 46.081
234670 100.373 -76.627 46.049
234671 100.392 -76.614 46.018
234672 100.411 -76.603 45.986
234673 100.431 -76.611 45.957
234674 100.451 -76.619 45.928
234675 100.470 -76.627 45.900
234676 100.489 -76.635 45.872
234677 100.509 -76.643 45.847
234678 100.528 -76.651 45.823
234679 100.548 -76.663 45.799
234680 100.567 -76.682 45.777
234681 100.586 -76.700 45.756
234682 100.630 -76.696 45.736
234683 100.675 -76.690 45.716
234684 100.719 -76.684 45.696
234685 100.748 -76.690 45.677
234686 100.772 -76.699 45.657
234687 100.796 -76.708 45.637
234688 100.820 -76.718 45.616
234689 100.844 -76.727 45.596
234690 100.868 -76.736 45.575
234691 100.892 -76.745 45.556
234692 100.916 -76.754 45.535
234693 100.942 -76.765 45.515
234694 100.970 -76.774 45.494
234695 100.998 -76.785 45.474
234696 101.025 -76.796 45.455
234697 101.054 -76.806 45.436
234698 101.073 -76.818 45.417
234699 101.089 -76.830 45.400
234700 101.103 -76.842 45.385
234701 101.118 -76.854 45.369
234702 101.133 -76.866 45.355
234703 101.147 -76.878 45.342
234704 101.162 -76.890 45.331
234705 101.177 -76.902 45.320
234706 101.192 -76.914 45.308
234707 101.206 -76.926 45.297
234708 101.221 -76.938 45.285
234709 101.236 -76.950 45.273
234710 101.251 -76.962 45.259
234711 101.265 -76.974 45.246
234712 101.285 -76.990 45.233
234713 101.314 -77.012 45.220
234714 101.343 -77.034 45.205
234715 101.372 -77.057 45.191
234716 101.401 -77.079 45.177
234717 101.420 -77.095 45.162
234718 101.435 -77.108 45.148
234719 101.450 -77.119 45.132
234720 101.473 -77.119 45.116
234721 101.496 -77.119 45.100
234722 101.519 -77.120 45.082
234723 101.541 -77.120 45.064
234724 101.564 -77.120 45.046
234725 101.586 -77.122 45.029
234726 101.604 -77.128 45.011
234727 101.622 -77.135 44.992
234728 101.641 -77.142 44.974
234729 101.659 -77.148 44.956
234730 101.678 -77.155 44.938
234731 101.697 -77.162 44.922
234732 101.715 -77.168 44.906
234733 101.734 -77.175 44.889
234734 101.753 -77.183 44.875
234735 101.772 -77.197 44.859
234736 101.792 -77.210 44.843
234737 101.812 -77.224 44.826
234738 101.831 -77.238 44.810
234739 101.851 -77.251 44.793
234740 101.872 -77.268 44.777
234741 101.895 -77.292 44.761
234742 101.918 -77.316 44.743
234743 101.942 -77.339 44.726
234744 101.965 -77.363 44.710
234745 101.988 -77.387 44.694
234746 102.011 -77.410 44.678
234747 102.034 -77.435 44.661
234748 102.057 -77.458 44.645
234749 102.078 -77.481 44.630
234750 102.108 -77.498 44.612
234751 102.128 -77.506 44.596
234752 102.148 -77.515 44.580
234753 102.168 -77.523 44.563
234754 102.189 -77.533 44.547
234755 102.209 -77.542 44.531
234756 102.229 -77.550 44.516
234757 102.250 -77.559 44.499
234758 102.270 -77.568 44.483
234759 102.296 -77.578 44.468
234760 102.328 -77.589 44.453
234761 102.361 -77.602 44.436
234762 102.376 -77.621 44.420
234763 102.395 -77.636 44.404
234764 102.420 -77.641 44.387
234765 102.446 -77.645 44.371
234766 102.471 -77.649 44.355
234767 102.497 -77.653 44.338
234768 102.523 -77.658 44.320
234769 102.553 -77.662 44.302
234770 102.590 -77.667 44.285
234771 102.626 -77.673 44.267
234772 102.663 -77.678 44.250
234773 102.699 -77.683 44.234
234774 102.735 -77.689 44.218
234775 102.771 -77.694 44.201
234776 102.808 -77.699 44.184
234777 102.844 -77.705 44.168
234778 102.880 -77.710 44.152
234779 102.917 -77.715 44.137
234780 102.953 -77.721 44.119
234781 102.989 -77.726 44.102
234782 103.026 -77.732 44.085
234783 103.062 -77.737 44.070
234784 103.098 -77.743 44.054
234785 103.135 -77.748 44.038
234786 103.171 -77.753 44.022
235043 92.237 -74.807 54.336
235044 92.262 -74.818 54.345
235045 92.286 -74.828 54.354
235046 92.310 -74.840 54.364
235047 92.335 -74.850 54.374
235048 92.359 -74.861 54.386
235049 92.383 -74.872 54.400
235050 92.408 -74.883 54.413
235051 92.432 -74.894 54.426
235052 92.456 -74.904 54.437
235053 92.471 -74.886 54.447
235054 92.494 -74.872 54.454
235055 92.520 -74.863 54.462
235056 92.550 -74.856 54.469
235057 92.583 -74.853 54.475
235058 92.616 -74.849 54.476
235059 92.648 -74.846 54.476
235060 92.681 -74.844 54.473
235061 92.712 -74.841 54.469
235062 92.744 -74.838 54.465
235063 92.777 -74.834 54.462
235064 92.807 -74.832 54.458
235065 92.841 -74.834 54.448
235066 92.874 -74.837 54.436
235067 92.908 -74.840 54.420
235068 92.942 -74.844 54.403
235069 92.976 -74.847 54.389
235070 93.009 -74.850 54.374
235071 93.043 -74.853 54.360
235072 93.076 -74.856 54.348
235073 93.107 -74.859 54.337
235074 93.132 -74.859 54.326
235075 93.155 -74.853 54.314
235076 93.178 -74.846 54.303
235077 93.201 -74.838 54.291
235078 93.223 -74.831 54.281
235079 93.247 -74.825 54.272
235080 93.269 -74.818 54.264
235081 93.292 -74.811 54.257
235082 93.315 -74.803 54.251
235083 93.338 -74.797 54.247
235084 93.361 -74.790 54.244
235085 93.384 -74.783 54.243
235086 93.407 -74.776 54.245
235087 93.430 -74.769 54.249
235088 93.453 -74.762 54.255
235089 93.476 -74.755 54.260
235090 93.499 -74.748 54.269
235091 93.521 -74.741 54.280
235092 93.545 -74.734 54.293
235093 93.567 -74.727 54.311
235094 93.595 -74.728 54.332
235095 93.624 -74.730 54.355
235096 93.653 -74.732 54.378
235097 93.683 -74.737 54.407
235098 93.716 -74.750 54.439
235099 93.749 -74.763 54.472
235100 93.782 -74.776 54.502
235101 93.815 -74.788 54.526
235102 93.848 -74.801 54.542
235103 93.881 -74.814 54.551
235104 93.903 -74.816 54.552
235105 93.921 -74.812 54.552
235106 93.939 -74.809 54.555
235107 93.957 -74.806 54.561
235108 93.975 -74.803 54.572
235109 93.992 -74.800 54.580
235110 94.009 -74.791 54.588
235111 94.027 -74.784 54.595
235112 94.047 -74.775 54.602
235113 94.074 -74.768 54.607
235114 94.101 -74.760 54.610
235115 94.127 -74.753 54.611
235116 94.152 -74.750 54.601
235117 94.174 -74.753 54.587
235118 94.195 -74.755 54.570
235119 94.217 -74.756 54.550
235120 94.238 -74.758 54.522
235121 94.260 -74.759 54.491
235122 94.282 -74.762 54.460
235123 94.303 -74.763 54.428
235124 94.326 -74.765 54.399
235125 94.347 -74.766 54.375
235126 94.368 -74.766 54.352
235127 94.381 -74.771 54.329
235128 94.398 -74.774 54.311
235129 94.415 -74.777 54.292
235130 94.433 -74.781 54.271
235131 94.450 -74.784 54.251
235132 94.468 -74.787 54.232
235133 94.485 -74.790 54.213
235134 94.502 -74.793 54.195
235135 94.523 -74.795 54.176
235136 94.544 -74.797 54.153
235137 94.564 -74.799 54.129
235138 94.583 -74.799 54.105
235139 94.603 -74.799 54.083
235140 94.622 -74.799 54.068
235141 94.641 -74.799 54.054
235142 94.662 -74.800 54.038
235143 94.690 -74.806 54.014
235144 94.718 -74.812 53.985
235145 94.746 -74.818 53.951
235146 94.774 -74.824 53.913
235147 94.802 -74.830 53.871
235148 94.830 -74.838 53.827
235149 94.858 -74.847 53.780
235150 94.886 -74.856 53.732
235151 94.914 -74.864 53.683
235152 94.942 -74.872 53.634
235153 94.964 -74.879 53.584
235154 94.979 -74.885 53.534
235155 94.995 -74.891 53.484
235156 95.011 -74.896 53.436
235157 95.015 -74.894 53.387
235158 95.009 -74.884 53.336
235159 95.003 -74.874 53.281
235160 94.997 -74.864 53.224
235161 94.991 -74.854 53.162
235162 94.994 -74.848 53.098
235163 95.027 -74.855 53.029
235164 95.058 -74.861 52.960
235165 95.091 -74.868 52.894
235166 95.122 -74.875 52.829
235167 95.155 -74.881 52.768
235168 95.186 -74.888 52.707
235169 95.218 -74.894 52.647
235170 95.251 -74.903 52.588
235171 95.278 -74.914 52.528
235172 95.306 -74.925 52.468
235173 95.334 -74.935 52.408
235174 95.361 -74.947 52.349
235175 95.389 -74.957 52.294
235176 95.416 -74.969 52.240
235177 95.444 -74.980 52.188
235178 95.470 -74.989 52.138
235179 95.484 -74.989 52.089
235180 95.501 -74.984 52.042
235181 95.519 -74.977 51.998
235182 95.536 -74.970 51.955
235183 95.554 -74.963 51.914
235184 95.572 -74.957 51.876
235185 95.590 -74.950 51.840
235186 95.608 -74.944 51.805
235187 95.625 -74.937 51.770
235188 95.661 -74.956 51.737
235189 95.701 -74.982 51.705
235190 95.742 -74.997 51.675
235191 95.783 -75.003 51.646
235192 95.824 -75.008 51.617
235193 95.865 -75.014 51.589
235194 95.905 -75.020 51.561
235195 95.947 -75.026 51.534
235196 95.984 -75.033 51.508
235197 96.019 -75.041 51.482
235198 96.055 -75.050 51.458
235199 96.090 -75.057 51.435
235200 96.122 -75.060 51.413
235201 96.153 -75.062 51.392
235202 96.185 -75.064 51.372
235203 96.216 -75.066 51.353
235204 96.248 -75.068 51.333
235205 96.279 -75.070 51.314
235206 96.311 -75.073 51.294
235207 96.342 -75.075 51.275
235208 96.374 -75.077 51.255
235209 96.405 -75.079 51.233
235210 96.437 -75.081 51.213
235211 96.464 -75.084 51.191
235212 96.480 -75.099 51.166
235213 96.495 -75.115 51.140
235214 96.510 -75.131 51.112
235215 96.526 -75.148 51.085
235216 96.543 -75.164 51.060
235217 96.564 -75.176 51.037
235218 96.587 -75.180 51.014
235219 96.612 -75.191 50.992
235220 96.637 -75.202 50.972
235221 96.662 -75.214 50.950
235222 96.687 -75.224 50.928
235223 96.713 -75.236 50.906
235224 96.738 -75.246 50.884
235225 96.763 -75.258 50.862
235226 96.788 -75.269 50.841
235227 96.813 -75.279 50.820
235228 96.838 -75.291 50.799
235229 96.857 -75.308 50.777
235230 96.870 -75.329 50.755
235231 96.894 -75.342 50.732
235232 96.918 -75.355 50.707
235233 96.942 -75.367 50.681
235234 96.966 -75.380 50.653
235235 96.990 -75.392 50.624
235236 97.014 -75.405 50.595
235237 97.038 -75.418 50.564
235238 97.062 -75.431 50.534
235239 97.086 -75.444 50.503
235240 97.110 -75.457 50.473
235241 97.155 -75.465 50.441
235242 97.205 -75.470 50.409
235243 97.229 -75.476 50.377
235244 97.254 -75.482 50.343
235245 97.278 -75.488 50.310
235246 97.302 -75.494 50.277
235247 97.319 -75.505 50.245
235248 97.336 -75.517 50.213
235249 97.354 -75.529 50.183
235250 97.371 -75.540 50.151
235251 97.388 -75.552 50.119
235252 97.405 -75.564 50.087
235253 97.423 -75.575 50.053
235254 97.450 -75.580 50.019
235255 97.475 -75.583 49.985
235256 97.500 -75.588 49.951
235257 97.524 -75.592 49.916
235258 97.549 -75.596 49.881
235259 97.574 -75.599 49.847
235260 97.598 -75.603 49.813
235261 97.623 -75.608 49.780
235262 97.648 -75.612 49.746
235263 97.672 -75.615 49.710
235264 97.697 -75.619 49.673
235265 97.722 -75.623 49.634
235266 97.746 -75.627 49.594
235267 97.773 -75.633 49.552
235268 97.830 -75.634 49.509
235269 97.891 -75.637 49.465
235270 97.918 -75.669 49.419
235271 97.938 -75.681 49.371
235272 97.957 -75.692 49.324
235273 97.973 -75.699 49.278
235274 98.015 -75.706 49.233
235275 98.057 -75.715 49.189
235276 98.075 -75.726 49.145
235277 98.092 -75.736 49.100
235278 98.109 -75.746 49.055
235279 98.128 -75.757 49.011
235280 98.149 -75.767 48.967
235281 98.171 -75.778 48.924
235282 98.192 -75.788 48.880
235283 98.211 -75.801 48.836
235284 98.229 -75.815 48.790
235285 98.262 -75.831 48.744
235286 98.307 -75.850 48.699
235287 98.347 -75.865 48.653
235288 98.387 -75.881 48.610
235289 98.427 -75.897 48.567
235290 98.467 -75.913 48.524
235291 98.507 -75.929 48.481
235292 98.547 -75.945 48.438
235293 98.583 -75.961 48.397
235294 98.598 -75.982 48.357
235295 98.638 -76.040 48.315
235296 98.678 -76.097 48.277
235297 98.719 -76.154 48.239
235298 98.747 -76.157 48.200
235299 98.773 -76.155 48.164
235300 98.798 -76.154 48.128
235301 98.824 -76.154 48.093
235302 98.850 -76.152 48.059
235303 98.875 -76.151 48.026
235304 98.901 -76.151 47.992
235305 98.927 -76.149 47.959
235306 98.952 -76.148 47.930
235307 98.978 -76.148 47.902
235308 99.003 -76.146 47.875
235309 99.029 -76.145 47.850
235310 99.055 -76.144 47.826
235311 99.080 -76.141 47.804
235312 99.105 -76.140 47.782
235313 99.130 -76.138 47.761
235314 99.155 -76.147 47.740
235315 99.181 -76.158 47.720
235316 99.205 -76.169 47.698
235317 99.229 -76.180 47.677
235318 99.253 -76.191 47.656
235319 99.277 -76.202 47.633
235320 99.301 -76.213 47.611
235321 99.326 -76.223 47.587
235322 99.350 -76.235 47.563
235323 99.374 -76.245 47.536
235324 99.398 -76.256 47.510
235325 99.422 -76.267 47.485
235326 99.443 -76.275 47.459
235327 99.452 -76.273 47.431
235328 99.473 -76.272 47.403
235329 99.488 -76.274 47.375
235330 99.503 -76.276 47.347
235331 99.518 -76.279 47.320
235332 99.534 -76.281 47.291
235333 99.549 -76.283 47.262
235334 99.564 -76.285 47.231
235335 99.579 -76.288 47.201
235336 99.594 -76.290 47.170
235337 99.610 -76.292 47.140
235338 99.625 -76.295 47.111
235339 99.640 -76.296 47.082
235340 99.665 -76.304 47.052
235341 99.691 -76.315 47.022
235342 99.708 -76.329 46.990
235343 99.725 -76.344 46.958
235344 99.751 -76.360 46.926
235345 99.783 -76.376 46.896
235346 99.815 -76.392 46.866
235347 99.847 -76.408 46.836
235348 99.879 -76.425 46.806
235349 99.901 -76.430 46.776
235350 99.921 -76.434 46.746
235351 99.941 -76.437 46.717
235352 99.961 -76.441 46.688
235353 99.981 -76.445 46.659
235354 100.000 -76.457 46.626
235355 100.018 -76.470 46.594
235356 100.037 -76.483 46.562
235357 100.055 -76.496 46.528
235358 100.073 -76.509 46.495
235359 100.091 -76.523 46.461
235360 100.126 -76.529 46.426
235361 100.164 -76.535 46.393
235362 100.202 -76.540 46.357
235363 100.240 -76.546 46.320
235364 100.270 -76.554 46.284
235365 100.287 -76.567 46.249
235366 100.304 -76.580 46.214
235367 100.321 -76.592 46.179
235368 100.338 -76.605 46.143
235369 100.355 -76.618 46.108
235370 100.372 -76.630 46.075
235371 100.391 -76.627 46.042
235372 100.410 -76.615 46.010
235373 100.429 -76.602 45.978
235374 100.449 -76.602 45.949
235375 100.468 -76.611 45.920
235376 100.487 -76.618 45.892
235377 100.507 -76.626 45.865
235378 100.527 -76.634 45.840
235379 100.546 -76.644 45.816
235380 100.565 -76.663 45.793
235381 100.584 -76.681 45.771
235382 100.611 -76.693 45.750
235383 100.655 -76.687 45.730
235384 100.700 -76.682 45.710
235385 100.745 -76.676 45.690
235386 100.779 -76.677 45.671
235387 100.803 -76.687 45.651
235388 100.827 -76.696 45.630
235389 100.851 -76.705 45.609
235390 100.875 -76.715 45.588
235391 100.899 -76.724 45.568
235392 100.922 -76.733 45.548
235393 100.947 -76.742 45.527
235394 100.973 -76.752 45.507
235395 101.001 -76.762 45.486
235396 101.029 -76.773 45.466
235397 101.057 -76.784 45.446
235398 101.085 -76.794 45.427
235399 101.102 -76.806 45.409
235400 101.117 -76.818 45.392
235401 101.131 -76.830 45.375
235402 101.146 -76.842 45.361
235403 101.161 -76.854 45.348
235404 101.176 -76.866 45.335
235405 101.190 -76.878 45.323
235406 101.205 -76.890 45.312
235407 101.220 -76.902 45.301
235408 101.235 -76.914 45.290
235409 101.249 -76.926 45.277
235410 101.264 -76.938 45.263
235411 101.279 -76.950 45.249
235412 101.293 -76.962 45.238
235413 101.308 -76.975 45.225
235414 101.324 -76.987 45.212
235415 101.353 -77.009 45.198
235416 101.382 -77.031 45.184
235417 101.411 -77.054 45.170
235418 101.434 -77.073 45.154
235419 101.449 -77.085 45.139
235420 101.466 -77.094 45.124
235421 101.489 -77.094 45.107
235422 101.512 -77.094 45.090
235423 101.535 -77.094 45.073
235424 101.557 -77.094 45.055
235425 101.580 -77.094 45.038
235426 101.603 -77.096 45.021
235427 101.621 -77.102 45.002
235428 101.639 -77.109 44.984
235429 101.658 -77.115 44.967
235430 101.677 -77.122 44.949
235431 101.695 -77.128 44.930
235432 101.713 -77.135 44.914
235433 101.732 -77.141 44.898
235434 101.752 -77.153 44.882
235435 101.771 -77.167 44.868
235436 101.791 -77.181 44.852
235437 101.811 -77.194 44.835
235438 101.831 -77.208 44.818
235439 101.850 -77.222 44.803
235440 101.870 -77.235 44.785
235441 101.891 -77.254 44.768
235442 101.914 -77.278 44.752
235443 101.938 -77.301 44.736
235444 101.961 -77.326 44.719
235445 101.984 -77.349 44.704
235446 102.007 -77.373 44.689
235447 102.031 -77.397 44.673
235448 102.054 -77.420 44.656
235449 102.077 -77.444 44.640
235450 102.100 -77.468 44.625
235451 102.120 -77.484 44.607
235452 102.140 -77.493 44.590
235453 102.161 -77.501 44.574
235454 102.181 -77.511 44.557
235455 102.201 -77.520 44.541
235456 102.222 -77.528 44.525
235457 102.242 -77.537 44.510
235458 102.262 -77.546 44.492
235459 102.282 -77.555 44.477
235460 102.303 -77.564 44.461
235461 102.335 -77.575 44.446
235462 102.368 -77.587 44.429
235463 102.383 -77.608 44.412
235464 102.405 -77.617 44.396
235465 102.430 -77.621 44.379
235466 102.456 -77.626 44.363
235467 102.482 -77.630 44.346
235468 102.507 -77.634 44.328
235469 102.542 -77.639 44.310
235470 102.578 -77.645 44.293
235471 102.614 -77.650 44.275
235472 102.650 -77.655 44.257
235473 102.687 -77.661 44.242
235474 102.724 -77.666 44.225
235475 102.760 -77.671 44.209
235476 102.796 -77.677 44.192
235477 102.832 -77.682 44.175
235478 102.869 -77.687 44.159
235479 102.905 -77.692 44.144
235480 102.941 -77.698 44.128
235481 102.977 -77.703 44.110
235482 103.014 -77.708 44.094
235483 103.050 -77.714 44.078
235484 103.087 -77.719 44.062
235485 103.123 -77.724 44.046
235486 103.159 -77.730 44.030
235744 92.239 -74.775 54.340
235745 92.263 -74.785 54.347
235746 92.288 -74.797 54.356
235747 92.312 -74.807 54.366
235748 92.337 -74.818 54.376
235749 92.361 -74.828 54.387
235750 92.385 -74.840 54.401
235751 92.410 -74.850 54.413
235752 92.434 -74.861 54.425
235753 92.458 -74.872 54.436
235754 92.476 -74.862 54.445
235755 92.495 -74.846 54.453
235756 92.522 -74.836 54.461
235757 92.549 -74.827 54.469
235758 92.578 -74.820 54.475
235759 92.611 -74.816 54.478
235760 92.643 -74.813 54.480
235761 92.675 -74.811 54.480
235762 92.708 -74.808 54.478
235763 92.740 -74.805 54.476
235764 92.773 -74.802 54.474
235765 92.806 -74.800 54.470
235766 92.840 -74.804 54.461
235767 92.874 -74.807 54.449
235768 92.908 -74.811 54.434
235769 92.941 -74.815 54.419
235770 92.975 -74.818 54.406
235771 93.009 -74.822 54.393
235772 93.043 -74.825 54.381
235773 93.077 -74.828 54.370
235774 93.110 -74.834 54.360
235775 93.136 -74.828 54.349
235776 93.159 -74.819 54.338
235777 93.183 -74.809 54.326
235778 93.206 -74.802 54.316
235779 93.228 -74.794 54.306
235780 93.252 -74.787 54.297
235781 93.274 -74.781 54.289
235782 93.297 -74.774 54.283
235783 93.320 -74.767 54.277
235784 93.343 -74.759 54.273
235785 93.366 -74.753 54.270
235786 93.389 -74.746 54.270
235787 93.412 -74.739 54.272
235788 93.435 -74.732 54.276
235789 93.458 -74.725 54.282
235790 93.481 -74.718 54.288
235791 93.504 -74.711 54.295
235792 93.526 -74.704 54.305
235793 93.550 -74.697 54.317
235794 93.572 -74.690 54.333
235795 93.600 -74.690 54.351
235796 93.629 -74.693 54.371
235797 93.658 -74.697 54.390
235798 93.692 -74.709 54.416
235799 93.724 -74.722 54.445
235800 93.758 -74.735 54.476
235801 93.790 -74.747 54.507
235802 93.824 -74.760 54.530
235803 93.856 -74.773 54.546
235804 93.889 -74.785 54.556
235805 93.916 -74.792 54.558
235806 93.934 -74.789 54.560
235807 93.952 -74.786 54.564
235808 93.970 -74.783 54.571
235809 93.987 -74.777 54.582
235810 94.004 -74.769 54.590
235811 94.022 -74.761 54.598
235812 94.040 -74.753 54.605
235813 94.057 -74.745 54.612
235814 94.074 -74.737 54.616
235815 94.097 -74.729 54.619
235816 94.122 -74.724 54.621
235817 94.144 -74.725 54.611
235818 94.166 -74.728 54.599
235819 94.188 -74.730 54.584
235820 94.209 -74.732 54.566
235821 94.231 -74.734 54.537
235822 94.253 -74.735 54.506
235823 94.275 -74.737 54.473
235824 94.296 -74.738 54.440
235825 94.319 -74.740 54.410
235826 94.342 -74.743 54.383
235827 94.363 -74.745 54.358
235828 94.380 -74.748 54.332
235829 94.398 -74.751 54.309
235830 94.415 -74.754 54.285
235831 94.432 -74.757 54.262
235832 94.450 -74.760 54.240
235833 94.467 -74.764 54.219
235834 94.484 -74.767 54.199
235835 94.503 -74.770 54.180
235836 94.524 -74.772 54.161
235837 94.545 -74.774 54.138
235838 94.566 -74.776 54.114
235839 94.585 -74.776 54.091
235840 94.604 -74.776 54.070
235841 94.623 -74.776 54.056
235842 94.642 -74.776 54.044
235843 94.668 -74.781 54.030
235844 94.696 -74.787 54.007
235845 94.724 -74.793 53.981
235846 94.753 -74.799 53.951
235847 94.781 -74.805 53.916
235848 94.809 -74.811 53.875
235849 94.837 -74.817 53.832
235850 94.865 -74.823 53.788
235851 94.894 -74.831 53.741
235852 94.921 -74.841 53.694
235853 94.950 -74.849 53.644
235854 94.977 -74.857 53.594
235855 94.993 -74.863 53.543
235856 95.008 -74.869 53.492
235857 95.024 -74.874 53.441
235858 95.020 -74.865 53.390
235859 95.014 -74.855 53.337
235860 95.007 -74.845 53.283
235861 95.002 -74.835 53.225
235862 94.996 -74.825 53.165
235863 95.000 -74.820 53.101
235864 95.032 -74.826 53.031
235865 95.064 -74.833 52.964
235866 95.096 -74.840 52.897
235867 95.128 -74.846 52.833
235868 95.160 -74.853 52.773
235869 95.192 -74.859 52.714
235870 95.225 -74.872 52.655
235871 95.259 -74.888 52.596
235872 95.287 -74.899 52.536
235873 95.315 -74.910 52.476
235874 95.342 -74.920 52.416
235875 95.370 -74.931 52.357
235876 95.398 -74.941 52.302
235877 95.425 -74.951 52.247
235878 95.453 -74.962 52.194
235879 95.480 -74.970 52.144
235880 95.508 -74.957 52.094
235881 95.526 -74.950 52.047
235882 95.544 -74.943 52.002
235883 95.561 -74.936 51.959
235884 95.580 -74.929 51.918
235885 95.597 -74.922 51.879
235886 95.615 -74.916 51.842
235887 95.633 -74.910 51.807
235888 95.651 -74.903 51.772
235889 95.677 -74.908 51.739
235890 95.717 -74.934 51.706
235891 95.758 -74.960 51.676
235892 95.798 -74.980 51.647
235893 95.839 -74.986 51.618
235894 95.880 -74.991 51.589
235895 95.921 -74.997 51.562
235896 95.962 -75.004 51.535
235897 95.998 -75.011 51.509
235898 96.034 -75.019 51.483
235899 96.068 -75.027 51.459
235900 96.100 -75.029 51.436
235901 96.132 -75.031 51.414
235902 96.163 -75.033 51.393
235903 96.195 -75.035 51.372
235904 96.226 -75.038 51.353
235905 96.258 -75.040 51.335
235906 96.290 -75.041 51.316
235907 96.321 -75.044 51.298
235908 96.353 -75.046 51.279
235909 96.384 -75.048 51.261
235910 96.415 -75.051 51.241
235911 96.447 -75.054 51.221
235912 96.469 -75.065 51.199
235913 96.484 -75.079 51.174
235914 96.499 -75.096 51.148
235915 96.514 -75.112 51.121
235916 96.529 -75.128 51.095
235917 96.546 -75.144 51.070
235918 96.571 -75.151 51.048
235919 96.596 -75.163 51.025
235920 96.621 -75.173 51.004
235921 96.646 -75.185 50.984
235922 96.671 -75.196 50.963
235923 96.696 -75.207 50.941
235924 96.722 -75.218 50.919
235925 96.747 -75.229 50.896
235926 96.772 -75.240 50.875
235927 96.797 -75.251 50.854
235928 96.822 -75.262 50.834
235929 96.848 -75.273 50.814
235930 96.872 -75.286 50.792
235931 96.891 -75.301 50.770
235932 96.915 -75.314 50.746
235933 96.939 -75.327 50.722
235934 96.963 -75.340 50.695
235935 96.987 -75.353 50.667
235936 97.011 -75.366 50.638
235937 97.035 -75.379 50.608
235938 97.060 -75.392 50.578
235939 97.083 -75.405 50.548
235940 97.107 -75.417 50.519
235941 97.132 -75.430 50.488
235942 97.165 -75.441 50.456
235943 97.215 -75.446 50.424
235944 97.239 -75.452 50.392
235945 97.264 -75.458 50.358
235946 97.288 -75.464 50.326
235947 97.305 -75.475 50.294
235948 97.322 -75.487 50.263
235949 97.339 -75.499 50.233
235950 97.357 -75.510 50.202
235951 97.374 -75.522 50.171
235952 97.391 -75.532 50.139
235953 97.408 -75.540 50.106
235954 97.426 -75.548 50.073
235955 97.453 -75.554 50.039
235956 97.476 -75.558 50.005
235957 97.499 -75.562 49.970
235958 97.521 -75.566 49.935
235959 97.544 -75.570 49.899
235960 97.567 -75.574 49.865
235961 97.590 -75.578 49.831
235962 97.612 -75.582 49.798
235963 97.636 -75.586 49.763
235964 97.658 -75.590 49.727
235965 97.681 -75.594 49.689
235966 97.704 -75.598 49.649
235967 97.727 -75.602 49.608
235968 97.760 -75.603 49.566
235969 97.834 -75.596 49.521
235970 97.896 -75.599 49.476
235971 97.923 -75.631 49.430
235972 97.948 -75.658 49.382
235973 97.967 -75.668 49.335
235974 97.989 -75.676 49.288
235975 98.034 -75.684 49.243
235976 98.079 -75.693 49.199
235977 98.097 -75.702 49.155
235978 98.114 -75.713 49.110
235979 98.131 -75.724 49.066
235980 98.149 -75.734 49.021
235981 98.167 -75.744 48.977
235982 98.189 -75.755 48.933
235983 98.210 -75.765 48.889
235984 98.232 -75.775 48.844
235985 98.252 -75.787 48.797
235986 98.271 -75.800 48.751
235987 98.315 -75.818 48.706
235988 98.355 -75.834 48.660
235989 98.395 -75.850 48.615
235990 98.435 -75.865 48.571
235991 98.476 -75.881 48.528
235992 98.515 -75.897 48.485
235993 98.556 -75.913 48.442
235994 98.595 -75.929 48.400
235995 98.629 -75.971 48.359
235996 98.664 -76.020 48.317
235997 98.695 -76.062 48.278
235998 98.726 -76.104 48.238
235999 98.760 -76.107 48.199
236000 98.785 -76.105 48.162
236001 98.810 -76.103 48.125
236002 98.836 -76.100 48.090
236003 98.861 -76.099 48.055
236004 98.886 -76.097 48.022
236005 98.912 -76.094 47.988
236006 98.937 -76.092 47.956
236007 98.962 -76.090 47.926
236008 98.987 -76.088 47.896
236009 99.012 -76.086 47.869
236010 99.038 -76.084 47.843
236011 99.063 -76.082 47.819
236012 99.088 -76.079 47.796
236013 99.106 -76.094 47.773
236014 99.124 -76.110 47.752
236015 99.148 -76.122 47.730
236016 99.175 -76.135 47.709
236017 99.203 -76.146 47.687
236018 99.231 -76.158 47.665
236019 99.258 -76.169 47.644
236020 99.286 -76.181 47.621
236021 99.313 -76.193 47.599
236022 99.341 -76.204 47.576
236023 99.369 -76.216 47.551
236024 99.396 -76.228 47.527
236025 99.420 -76.238 47.501
236026 99.444 -76.249 47.475
236027 99.468 -76.260 47.449
236028 99.484 -76.260 47.420
236029 99.499 -76.263 47.393
236030 99.514 -76.265 47.366
236031 99.529 -76.267 47.339
236032 99.545 -76.270 47.312
236033 99.560 -76.272 47.284
236034 99.575 -76.274 47.256
236035 99.590 -76.276 47.225
236036 99.605 -76.279 47.195
236037 99.621 -76.281 47.164
236038 99.636 -76.283 47.134
236039 99.651 -76.285 47.106
236040 99.666 -76.288 47.076
236041 99.682 -76.291 47.046
236042 99.706 -76.302 47.015
236043 99.723 -76.317 46.984
236044 99.740 -76.331 46.952
236045 99.758 -76.345 46.920
236046 99.790 -76.362 46.889
236047 99.822 -76.378 46.860
236048 99.854 -76.395 46.830
236049 99.882 -76.407 46.800
236050 99.902 -76.410 46.770
236051 99.922 -76.414 46.740
236052 99.942 -76.417 46.711
236053 99.963 -76.420 46.682
236054 99.983 -76.424 46.652
236055 100.001 -76.436 46.620
236056 100.019 -76.449 46.588
236057 100.037 -76.462 46.556
236058 100.056 -76.476 46.522
236059 100.077 -76.487 46.489
236060 100.115 -76.493 46.456
236061 100.153 -76.499 46.422
236062 100.191 -76.504 46.388
236063 100.228 -76.509 46.352
236064 100.267 -76.515 46.316
236065 100.290 -76.526 46.280
236066 100.307 -76.538 46.244
236067 100.324 -76.551 46.209
236068 100.341 -76.564 46.174
236069 100.358 -76.577 46.138
236070 100.375 -76.589 46.102
236071 100.393 -76.602 46.068
236072 100.410 -76.614 46.034
236073 100.428 -76.616 46.002
236074 100.446 -76.603 45.970
236075 100.466 -76.594 45.941
236076 100.485 -76.602 45.912
236077 100.505 -76.610 45.884
236078 100.525 -76.618 45.858
236079 100.544 -76.625 45.833
236080 100.563 -76.643 45.809
236081 100.582 -76.662 45.787
236082 100.601 -76.681 45.765
236083 100.636 -76.684 45.744
236084 100.681 -76.679 45.724
236085 100.725 -76.673 45.704
236086 100.770 -76.668 45.684
236087 100.810 -76.665 45.664
236088 100.834 -76.675 45.643
236089 100.858 -76.684 45.623
236090 100.882 -76.693 45.602
236091 100.906 -76.702 45.581
236092 100.929 -76.712 45.561
236093 100.954 -76.721 45.540
236094 100.977 -76.730 45.519
236095 101.004 -76.740 45.499
236096 101.032 -76.750 45.479
236097 101.060 -76.761 45.458
236098 101.088 -76.771 45.438
236099 101.116 -76.782 45.419
236100 101.130 -76.794 45.401
236101 101.145 -76.806 45.384
236102 101.159 -76.818 45.368
236103 101.175 -76.831 45.353
236104 101.189 -76.842 45.340
236105 101.204 -76.854 45.327
236106 101.218 -76.866 45.315
236107 101.234 -76.878 45.304
236108 101.248 -76.890 45.292
236109 101.263 -76.903 45.281
236110 101.278 -76.915 45.268
236111 101.292 -76.926 45.255
236112 101.307 -76.938 45.242
236113 101.322 -76.950 45.230
236114 101.337 -76.962 45.218
236115 101.351 -76.975 45.204
236116 101.366 -76.987 45.190
236117 101.392 -77.006 45.176
236118 101.420 -77.029 45.161
236119 101.449 -77.050 45.146
236120 101.463 -77.063 45.130
236121 101.482 -77.069 45.114
236122 101.505 -77.069 45.098
236123 101.528 -77.069 45.081
236124 101.551 -77.069 45.064
236125 101.573 -77.069 45.047
236126 101.596 -77.069 45.030
236127 101.619 -77.069 45.013
236128 101.638 -77.075 44.994
236129 101.656 -77.082 44.976
236130 101.675 -77.088 44.958
236131 101.694 -77.095 44.940
236132 101.712 -77.101 44.923
236133 101.731 -77.110 44.906
236134 101.751 -77.124 44.890
236135 101.770 -77.138 44.874
236136 101.790 -77.151 44.859
236137 101.810 -77.165 44.843
236138 101.830 -77.179 44.826
236139 101.849 -77.192 44.810
236140 101.869 -77.206 44.793
236141 101.889 -77.219 44.776
236142 101.911 -77.240 44.759
236143 101.934 -77.263 44.742
236144 101.957 -77.288 44.727
236145 101.980 -77.311 44.711
236146 102.004 -77.335 44.696
236147 102.027 -77.359 44.682
236148 102.050 -77.382 44.667
236149 102.073 -77.406 44.651
236150 102.096 -77.429 44.635
236151 102.118 -77.449 44.618
236152 102.139 -77.467 44.600
236153 102.157 -77.477 44.584
236154 102.174 -77.488 44.568
236155 102.193 -77.498 44.551
236156 102.214 -77.506 44.535
236157 102.234 -77.515 44.519
236158 102.254 -77.524 44.503
236159 102.274 -77.533 44.486
236160 102.295 -77.542 44.471
236161 102.315 -77.551 44.454
236162 102.343 -77.561 44.438
236163 102.376 -77.573 44.422
236164 102.390 -77.593 44.405
236165 102.415 -77.599 44.388
236166 102.441 -77.602 44.370
236167 102.467 -77.607 44.354
236168 102.493 -77.611 44.337
236169 102.530 -77.616 44.319
236170 102.567 -77.621 44.301
236171 102.603 -77.627 44.283
236172 102.639 -77.632 44.265
236173 102.675 -77.638 44.248
236174 102.712 -77.643 44.231
236175 102.748 -77.649 44.216
236176 102.784 -77.654 44.200
236177 102.821 -77.659 44.183
236178 102.857 -77.664 44.166
236179 102.893 -77.670 44.150
236180 102.930 -77.675 44.135
236181 102.966 -77.680 44.119
236182 103.002 -77.686 44.102
236183 103.038 -77.691 44.085
236184 103.075 -77.696 44.070
236185 103.111 -77.702 44.055
236186 103.148 -77.707 44.038
236446 92.265 -74.753 54.352
236447 92.289 -74.764 54.359
236448 92.314 -74.775 54.369
236449 92.338 -74.785 54.378
236450 92.363 -74.797 54.389
236451 92.387 -74.807 54.401
236452 92.411 -74.818 54.414
236453 92.436 -74.828 54.425
236454 92.460 -74.839 54.435
236455 92.481 -74.838 54.444
236456 92.497 -74.819 54.453
236457 92.523 -74.809 54.461
236458 92.550 -74.800 54.469
236459 92.577 -74.791 54.474
236460 92.606 -74.784 54.479
236461 92.639 -74.781 54.483
236462 92.671 -74.778 54.485
236463 92.705 -74.775 54.485
236464 92.738 -74.772 54.484
236465 92.771 -74.769 54.483
236466 92.804 -74.768 54.479
236467 92.838 -74.771 54.472
236468 92.872 -74.775 54.460
236469 92.906 -74.778 54.448
236470 92.939 -74.781 54.436
236471 92.973 -74.785 54.425
236472 93.007 -74.788 54.413
236473 93.041 -74.793 54.402
236474 93.074 -74.799 54.392
236475 93.107 -74.804 54.383
236476 93.139 -74.808 54.373
236477 93.162 -74.798 54.362
236478 93.186 -74.787 54.351
236479 93.209 -74.778 54.342
236480 93.232 -74.768 54.333
236481 93.256 -74.758 54.324
236482 93.279 -74.748 54.315
236483 93.302 -74.738 54.310
236484 93.325 -74.730 54.305
236485 93.348 -74.723 54.301
236486 93.371 -74.716 54.299
236487 93.394 -74.709 54.300
236488 93.417 -74.702 54.303
236489 93.440 -74.695 54.307
236490 93.463 -74.688 54.312
236491 93.486 -74.681 54.318
236492 93.508 -74.674 54.325
236493 93.532 -74.667 54.333
236494 93.555 -74.660 54.344
236495 93.577 -74.653 54.357
236496 93.605 -74.653 54.372
236497 93.634 -74.656 54.389
236498 93.667 -74.668 54.405
236499 93.700 -74.681 54.426
236500 93.733 -74.694 54.451
236501 93.766 -74.706 54.479
236502 93.799 -74.719 54.508
236503 93.832 -74.732 54.531
236504 93.865 -74.745 54.547
236505 93.898 -74.757 54.558
236506 93.929 -74.769 54.563
236507 93.947 -74.766 54.567
236508 93.965 -74.762 54.573
236509 93.982 -74.754 54.582
236510 93.999 -74.747 54.593
236511 94.017 -74.738 54.602
236512 94.035 -74.730 54.609
236513 94.052 -74.722 54.617
236514 94.069 -74.715 54.622
236515 94.085 -74.710 54.626
236516 94.101 -74.705 54.628
236517 94.118 -74.701 54.629
236518 94.139 -74.703 54.619
236519 94.159 -74.704 54.609
236520 94.180 -74.706 54.595
236521 94.202 -74.708 54.577
236522 94.225 -74.710 54.550
236523 94.249 -74.712 54.519
236524 94.272 -74.714 54.487
236525 94.295 -74.716 54.453
236526 94.318 -74.718 54.422
236527 94.341 -74.721 54.392
236528 94.361 -74.723 54.364
236529 94.380 -74.725 54.338
236530 94.397 -74.728 54.312
236531 94.414 -74.731 54.286
236532 94.432 -74.734 54.261
236533 94.449 -74.737 54.238
236534 94.467 -74.740 54.215
236535 94.484 -74.744 54.193
236536 94.503 -74.747 54.170
236537 94.525 -74.749 54.150
236538 94.546 -74.750 54.125
236539 94.567 -74.753 54.100
236540 94.586 -74.753 54.076
236541 94.606 -74.753 54.055
236542 94.625 -74.753 54.040
236543 94.647 -74.756 54.028
236544 94.675 -74.762 54.015
236545 94.703 -74.768 53.996
236546 94.732 -74.774 53.972
236547 94.760 -74.780 53.944
236548 94.788 -74.786 53.912
236549 94.816 -74.792 53.875
236550 94.844 -74.798 53.835
236551 94.873 -74.804 53.793
236552 94.901 -74.810 53.747
236553 94.929 -74.817 53.699
236554 94.957 -74.825 53.650
236555 94.985 -74.834 53.600
236556 95.007 -74.841 53.549
236557 95.023 -74.847 53.497
236558 95.030 -74.847 53.443
236559 95.024 -74.836 53.389
236560 95.018 -74.826 53.334
236561 95.013 -74.816 53.278
236562 95.006 -74.806 53.221
236563 95.000 -74.797 53.162
236564 95.006 -74.791 53.099
236565 95.038 -74.798 53.032
236566 95.070 -74.805 52.965
236567 95.102 -74.812 52.899
236568 95.134 -74.818 52.836
236569 95.166 -74.826 52.777
236570 95.199 -74.842 52.719
236571 95.233 -74.857 52.662
236572 95.267 -74.873 52.603
236573 95.296 -74.885 52.542
236574 95.323 -74.894 52.483
236575 95.351 -74.905 52.424
236576 95.378 -74.916 52.365
236577 95.406 -74.926 52.309
236578 95.433 -74.936 52.253
236579 95.461 -74.947 52.199
236580 95.487 -74.950 52.149
236581 95.512 -74.950 52.100
236582 95.545 -74.923 52.053
236583 95.569 -74.909 52.007
236584 95.587 -74.902 51.963
236585 95.604 -74.895 51.922
236586 95.622 -74.888 51.882
236587 95.640 -74.881 51.844
236588 95.658 -74.875 51.808
236589 95.676 -74.869 51.774
236590 95.693 -74.862 51.741
236591 95.733 -74.886 51.708
236592 95.773 -74.912 51.677
236593 95.814 -74.938 51.647
236594 95.854 -74.963 51.618
236595 95.895 -74.969 51.589
236596 95.936 -74.975 51.562
236597 95.977 -74.981 51.535
236598 96.013 -74.989 51.509
236599 96.047 -74.996 51.484
236600 96.079 -74.998 51.460
236601 96.110 -75.001 51.436
236602 96.142 -75.003 51.414
236603 96.173 -75.005 51.392
236604 96.205 -75.007 51.372
236605 96.236 -75.009 51.354
236606 96.268 -75.011 51.337
236607 96.300 -75.014 51.318
236608 96.331 -75.021 51.301
236609 96.363 -75.027 51.283
236610 96.395 -75.033 51.265
236611 96.427 -75.040 51.247
236612 96.458 -75.046 51.228
236613 96.474 -75.060 51.206
236614 96.488 -75.073 51.181
236615 96.504 -75.088 51.155
236616 96.518 -75.101 51.130
236617 96.533 -75.115 51.105
236618 96.556 -75.127 51.080
236619 96.586 -75.134 51.058
236620 96.611 -75.145 51.036
236621 96.637 -75.156 51.015
236622 96.662 -75.167 50.995
236623 96.687 -75.178 50.975
236624 96.713 -75.189 50.953
236625 96.738 -75.201 50.930
236626 96.763 -75.211 50.908
236627 96.789 -75.223 50.887
236628 96.814 -75.233 50.867
236629 96.839 -75.245 50.848
236630 96.865 -75.256 50.828
236631 96.890 -75.267 50.808
236632 96.909 -75.290 50.784
236633 96.935 -75.293 50.760
236634 96.960 -75.301 50.735
236635 96.984 -75.314 50.708
236636 97.009 -75.326 50.680
236637 97.032 -75.339 50.651
236638 97.056 -75.352 50.622
236639 97.081 -75.364 50.593
236640 97.105 -75.377 50.563
236641 97.129 -75.390 50.534
236642 97.153 -75.403 50.504
236643 97.177 -75.416 50.472
236644 97.225 -75.421 50.440
236645 97.249 -75.427 50.408
236646 97.273 -75.433 50.375
236647 97.290 -75.446 50.343
236648 97.308 -75.457 50.312
236649 97.325 -75.468 50.282
236650 97.342 -75.479 50.252
236651 97.359 -75.487 50.222
236652 97.376 -75.496 50.191
236653 97.393 -75.503 50.159
236654 97.409 -75.511 50.126
236655 97.430 -75.519 50.093
236656 97.458 -75.526 50.059
236657 97.484 -75.532 50.024
236658 97.507 -75.536 49.989
236659 97.529 -75.540 49.954
236660 97.552 -75.544 49.918
236661 97.575 -75.548 49.883
236662 97.598 -75.552 49.849
236663 97.620 -75.556 49.816
236664 97.643 -75.560 49.781
236665 97.666 -75.564 49.744
236666 97.689 -75.568 49.705
236667 97.712 -75.572 49.665
236668 97.734 -75.576 49.623
236669 97.780 -75.572 49.580
236670 97.844 -75.561 49.534
236671 97.901 -75.561 49.488
236672 97.928 -75.593 49.441
236673 97.955 -75.626 49.393
236674 97.978 -75.645 49.346
236675 98.007 -75.653 49.299
236676 98.053 -75.662 49.254
236677 98.097 -75.670 49.210
236678 98.119 -75.680 49.167
236679 98.136 -75.690 49.122
236680 98.153 -75.701 49.077
236681 98.171 -75.711 49.033
236682 98.188 -75.721 48.988
236683 98.207 -75.732 48.944
236684 98.229 -75.742 48.899
236685 98.250 -75.753 48.854
236686 98.271 -75.763 48.806
236687 98.293 -75.773 48.760
236688 98.329 -75.792 48.714
236689 98.374 -75.813 48.668
236690 98.420 -75.834 48.622
236691 98.465 -75.856 48.577
236692 98.510 -75.878 48.533
236693 98.556 -75.899 48.490
236694 98.601 -75.920 48.446
236695 98.619 -75.909 48.403
236696 98.629 -75.911 48.360
236697 98.658 -75.950 48.319
236698 98.690 -75.993 48.279
236699 98.720 -76.035 48.239
236700 98.761 -76.050 48.199
236701 98.794 -76.043 48.160
236702 98.819 -76.041 48.123
236703 98.844 -76.038 48.087
236704 98.869 -76.037 48.051
236705 98.895 -76.035 48.018
236706 98.920 -76.032 47.984
236707 98.945 -76.030 47.952
236708 98.970 -76.028 47.921
236709 98.996 -76.026 47.891
236710 99.021 -76.024 47.864
236711 99.046 -76.022 47.837
236712 99.064 -76.037 47.812
236713 99.081 -76.053 47.789
236714 99.098 -76.069 47.766
236715 99.115 -76.086 47.744
236716 99.140 -76.098 47.722
236717 99.168 -76.110 47.699
236718 99.196 -76.122 47.677
236719 99.223 -76.133 47.656
236720 99.251 -76.145 47.634
236721 99.279 -76.157 47.611
236722 99.306 -76.169 47.588
236723 99.334 -76.180 47.565
236724 99.361 -76.191 47.540
236725 99.389 -76.204 47.516
236726 99.417 -76.215 47.491
236727 99.445 -76.229 47.466
236728 99.474 -76.239 47.438
236729 99.498 -76.244 47.410
236730 99.513 -76.246 47.382
236731 99.529 -76.248 47.356
236732 99.545 -76.251 47.329
236733 99.560 -76.254 47.303
236734 99.576 -76.257 47.276
236735 99.592 -76.259 47.249
236736 99.608 -76.261 47.219
236737 99.623 -76.264 47.189
236738 99.639 -76.267 47.158
236739 99.654 -76.270 47.128
236740 99.670 -76.272 47.099
236741 99.686 -76.275 47.069
236742 99.702 -76.277 47.038
236743 99.721 -76.289 47.007
236744 99.738 -76.304 46.975
236745 99.755 -76.318 46.944
236746 99.772 -76.332 46.912
236747 99.797 -76.348 46.882
236748 99.829 -76.364 46.852
236749 99.861 -76.381 46.822
236750 99.883 -76.386 46.792
236751 99.903 -76.390 46.762
236752 99.923 -76.393 46.733
236753 99.943 -76.397 46.704
236754 99.964 -76.401 46.675
236755 99.984 -76.404 46.645
236756 100.002 -76.415 46.613
236757 100.020 -76.429 46.582
236758 100.039 -76.442 46.549
236759 100.066 -76.451 46.516
236760 100.104 -76.457 46.483
236761 100.142 -76.462 46.451
236762 100.180 -76.467 46.418
236763 100.217 -76.473 46.383
236764 100.255 -76.479 46.348
236765 100.292 -76.484 46.312
236766 100.310 -76.497 46.276
236767 100.327 -76.510 46.239
236768 100.344 -76.523 46.204
236769 100.361 -76.536 46.169
236770 100.378 -76.548 46.132
236771 100.395 -76.561 46.096
236772 100.412 -76.574 46.061
236773 100.429 -76.586 46.027
236774 100.446 -76.599 45.994
236775 100.465 -76.604 45.964
236776 100.483 -76.592 45.934
236777 100.503 -76.593 45.905
236778 100.522 -76.601 45.878
236779 100.542 -76.608 45.852
236780 100.561 -76.624 45.827
236781 100.580 -76.643 45.803
236782 100.599 -76.662 45.781
236783 100.618 -76.680 45.759
236784 100.662 -76.676 45.738
236785 100.706 -76.671 45.718
236786 100.751 -76.665 45.698
236787 100.796 -76.659 45.678
236788 100.841 -76.654 45.658
236789 100.864 -76.663 45.637
236790 100.889 -76.672 45.617
236791 100.912 -76.681 45.596
236792 100.936 -76.690 45.575
236793 100.960 -76.699 45.554
236794 100.984 -76.708 45.533
236795 101.008 -76.718 45.512
236796 101.035 -76.728 45.492
236797 101.063 -76.739 45.471
236798 101.091 -76.749 45.450
236799 101.119 -76.759 45.431
236800 101.144 -76.771 45.412
236801 101.159 -76.782 45.394
236802 101.173 -76.794 45.376
236803 101.188 -76.806 45.361
236804 101.203 -76.818 45.346
236805 101.217 -76.831 45.332
236806 101.232 -76.843 45.320
236807 101.247 -76.855 45.308
236808 101.262 -76.866 45.296
236809 101.276 -76.878 45.284
236810 101.291 -76.890 45.272
236811 101.306 -76.903 45.259
236812 101.321 -76.915 45.247
236813 101.335 -76.927 45.235
236814 101.350 -76.939 45.223
236815 101.365 -76.950 45.210
236816 101.379 -76.962 45.196
236817 101.394 -76.975 45.182
236818 101.409 -76.987 45.167
236819 101.430 -77.003 45.153
236820 101.459 -77.026 45.137
236821 101.477 -77.041 45.121
236822 101.499 -77.044 45.105
236823 101.522 -77.044 45.088
236824 101.544 -77.044 45.071
236825 101.567 -77.044 45.055
236826 101.590 -77.044 45.038
236827 101.613 -77.044 45.022
236828 101.635 -77.044 45.005
236829 101.655 -77.049 44.987
236830 101.674 -77.055 44.967
236831 101.692 -77.062 44.948
236832 101.711 -77.069 44.930
236833 101.731 -77.081 44.913
236834 101.750 -77.094 44.896
236835 101.770 -77.108 44.880
236836 101.790 -77.122 44.865
236837 101.809 -77.136 44.849
236838 101.829 -77.150 44.833
236839 101.848 -77.163 44.817
236840 101.869 -77.177 44.800
236841 101.888 -77.191 44.783
236842 101.908 -77.204 44.766
236843 101.930 -77.226 44.750
236844 101.954 -77.249 44.733
236845 101.977 -77.273 44.718
236846 102.000 -77.297 44.703
236847 102.023 -77.321 44.689
236848 102.046 -77.345 44.675
236849 102.069 -77.368 44.660
236850 102.092 -77.390 44.645
236851 102.114 -77.410 44.629
236852 102.135 -77.429 44.612
236853 102.157 -77.448 44.594
236854 102.175 -77.460 44.577
236855 102.192 -77.470 44.561
236856 102.210 -77.481 44.545
236857 102.228 -77.492 44.528
236858 102.246 -77.501 44.512
236859 102.267 -77.511 44.495
236860 102.287 -77.520 44.479
236861 102.307 -77.528 44.464
236862 102.327 -77.537 44.447
236863 102.351 -77.547 44.430
236864 102.383 -77.559 44.414
236865 102.400 -77.575 44.396
236866 102.426 -77.579 44.378
236867 102.451 -77.583 44.360
236868 102.482 -77.588 44.344
236869 102.518 -77.593 44.326
236870 102.554 -77.599 44.308
236871 102.591 -77.604 44.290
236872 102.627 -77.609 44.273
236873 102.664 -77.614 44.255
236874 102.700 -77.620 44.239
236875 102.736 -77.625 44.222
236876 102.772 -77.630 44.206
236877 102.809 -77.636 44.190
236878 102.845 -77.642 44.174
236879 102.882 -77.647 44.157
236880 102.918 -77.652 44.141
236881 102.954 -77.658 44.125
236882 102.991 -77.663 44.110
236883 103.027 -77.668 44.094
236884 103.063 -77.674 44.078
236885 103.099 -77.679 44.062
236886 103.135 -77.684 44.046
237147 92.267 -74.721 54.357
237148 92.291 -74.731 54.364
237149 92.315 -74.742 54.372
237150 92.340 -74.753 54.381
237151 92.364 -74.764 54.390
237152 92.389 -74.775 54.402
237153 92.413 -74.785 54.414
237154 92.437 -74.796 54.425
237155 92.462 -74.807 54.434
237156 92.485 -74.814 54.444
237157 92.499 -74.791 54.452
237158 92.525 -74.782 54.461
237159 92.552 -74.773 54.469
237160 92.578 -74.763 54.474
237161 92.605 -74.754 54.480
237162 92.636 -74.748 54.485
237163 92.669 -74.745 54.488
237164 92.702 -74.741 54.489
237165 92.735 -74.738 54.489
237166 92.769 -74.735 54.488
237167 92.802 -74.734 54.485
237168 92.836 -74.738 54.479
237169 92.869 -74.741 54.471
237170 92.903 -74.745 54.463
237171 92.937 -74.749 54.454
237172 92.971 -74.752 54.444
237173 93.004 -74.757 54.434
237174 93.038 -74.763 54.424
237175 93.071 -74.769 54.414
237176 93.104 -74.775 54.405
237177 93.137 -74.781 54.396
237178 93.165 -74.777 54.387
237179 93.188 -74.767 54.378
237180 93.211 -74.757 54.368
237181 93.235 -74.747 54.360
237182 93.258 -74.737 54.352
237183 93.281 -74.728 54.344
237184 93.305 -74.718 54.339
237185 93.328 -74.707 54.334
237186 93.351 -74.697 54.331
237187 93.375 -74.687 54.330
237188 93.398 -74.678 54.331
237189 93.422 -74.668 54.335
237190 93.445 -74.658 54.340
237191 93.468 -74.651 54.345
237192 93.491 -74.644 54.351
237193 93.513 -74.637 54.357
237194 93.537 -74.630 54.364
237195 93.559 -74.623 54.373
237196 93.582 -74.616 54.383
237197 93.610 -74.617 54.396
237198 93.642 -74.628 54.411
237199 93.676 -74.640 54.425
237200 93.708 -74.653 54.442
237201 93.742 -74.665 54.463
237202 93.774 -74.678 54.487
237203 93.807 -74.691 54.512
237204 93.840 -74.704 54.532
237205 93.873 -74.716 54.548
237206 93.906 -74.729 54.559
237207 93.939 -74.742 54.566
237208 93.960 -74.740 54.574
237209 93.977 -74.732 54.583
237210 93.993 -74.727 54.595
237211 94.009 -74.722 54.609
237212 94.025 -74.716 54.618
237213 94.041 -74.712 54.625
237214 94.056 -74.706 54.631
237215 94.072 -74.701 54.634
237216 94.088 -74.696 54.635
237217 94.104 -74.690 54.636
237218 94.120 -74.686 54.636
237219 94.138 -74.684 54.626
237220 94.159 -74.686 54.616
237221 94.180 -74.687 54.602
237222 94.201 -74.687 54.581
237223 94.224 -74.688 54.560
237224 94.247 -74.690 54.531
237225 94.270 -74.692 54.499
237226 94.293 -74.694 54.466
237227 94.317 -74.697 54.433
237228 94.339 -74.699 54.402
237229 94.359 -74.701 54.372
237230 94.378 -74.703 54.345
237231 94.397 -74.705 54.319
237232 94.414 -74.709 54.294
237233 94.431 -74.712 54.268
237234 94.449 -74.715 54.243
237235 94.466 -74.718 54.217
237236 94.483 -74.721 54.191
237237 94.504 -74.723 54.166
237238 94.526 -74.725 54.144
237239 94.547 -74.728 54.117
237240 94.568 -74.730 54.091
237241 94.588 -74.731 54.065
237242 94.607 -74.731 54.042
237243 94.626 -74.731 54.025
237244 94.653 -74.736 54.011
237245 94.682 -74.742 54.000
237246 94.710 -74.748 53.983
237247 94.738 -74.754 53.961
237248 94.766 -74.760 53.935
237249 94.794 -74.766 53.906
237250 94.823 -74.772 53.873
237251 94.851 -74.778 53.835
237252 94.879 -74.784 53.794
237253 94.907 -74.791 53.750
237254 94.935 -74.797 53.702
237255 94.964 -74.803 53.654
237256 94.992 -74.810 53.605
237257 95.020 -74.819 53.554
237258 95.036 -74.825 53.501
237259 95.035 -74.818 53.445
237260 95.029 -74.808 53.389
237261 95.023 -74.797 53.333
237262 95.017 -74.787 53.276
237263 95.011 -74.778 53.219
237264 95.005 -74.768 53.159
237265 95.012 -74.763 53.097
237266 95.043 -74.770 53.031
237267 95.075 -74.777 52.965
237268 95.107 -74.783 52.900
237269 95.140 -74.796 52.838
237270 95.174 -74.811 52.779
237271 95.207 -74.827 52.722
237272 95.241 -74.842 52.665
237273 95.275 -74.858 52.607
237274 95.304 -74.869 52.548
237275 95.331 -74.880 52.489
237276 95.359 -74.890 52.430
237277 95.387 -74.900 52.371
237278 95.414 -74.911 52.314
237279 95.442 -74.922 52.258
237280 95.470 -74.932 52.203
237281 95.494 -74.931 52.152
237282 95.520 -74.930 52.103
237283 95.547 -74.919 52.057
237284 95.580 -74.893 52.011
237285 95.612 -74.868 51.967
237286 95.629 -74.861 51.926
237287 95.648 -74.854 51.886
237288 95.665 -74.847 51.847
237289 95.683 -74.841 51.811
237290 95.701 -74.834 51.776
237291 95.719 -74.828 51.742
237292 95.749 -74.838 51.709
237293 95.789 -74.864 51.678
237294 95.829 -74.890 51.647
237295 95.870 -74.916 51.617
237296 95.910 -74.942 51.588
237297 95.951 -74.953 51.561
237298 95.991 -74.959 51.534
237299 96.026 -74.966 51.508
237300 96.057 -74.968 51.483
237301 96.089 -74.969 51.459
237302 96.120 -74.972 51.435
237303 96.152 -74.975 51.411
237304 96.184 -74.981 51.390
237305 96.216 -74.987 51.371
237306 96.247 -74.994 51.353
237307 96.279 -75.000 51.336
237308 96.311 -75.007 51.318
237309 96.342 -75.013 51.302
237310 96.374 -75.019 51.285
237311 96.406 -75.026 51.268
237312 96.438 -75.032 51.250
237313 96.464 -75.041 51.230
237314 96.479 -75.054 51.209
237315 96.493 -75.069 51.186
237316 96.509 -75.082 51.161
237317 96.525 -75.096 51.137
237318 96.545 -75.110 51.113
237319 96.566 -75.125 51.090
237320 96.596 -75.132 51.069
237321 96.626 -75.137 51.047
237322 96.657 -75.142 51.025
237323 96.687 -75.149 51.006
237324 96.712 -75.160 50.986
237325 96.738 -75.171 50.963
237326 96.763 -75.182 50.940
237327 96.788 -75.193 50.919
237328 96.814 -75.204 50.898
237329 96.839 -75.215 50.879
237330 96.864 -75.226 50.860
237331 96.889 -75.237 50.841
237332 96.915 -75.248 50.822
237333 96.921 -75.309 50.798
237334 96.947 -75.311 50.773
237335 96.973 -75.314 50.748
237336 96.999 -75.317 50.719
237337 97.025 -75.320 50.692
237338 97.051 -75.323 50.664
237339 97.078 -75.326 50.635
237340 97.102 -75.338 50.608
237341 97.126 -75.351 50.579
237342 97.150 -75.364 50.549
237343 97.174 -75.377 50.520
237344 97.198 -75.389 50.490
237345 97.235 -75.397 50.459
237346 97.258 -75.403 50.426
237347 97.276 -75.415 50.394
237348 97.293 -75.427 50.360
237349 97.310 -75.435 50.329
237350 97.327 -75.442 50.299
237351 97.344 -75.451 50.270
237352 97.361 -75.459 50.240
237353 97.378 -75.467 50.210
237354 97.394 -75.475 50.178
237355 97.412 -75.483 50.146
237356 97.433 -75.491 50.113
237357 97.463 -75.497 50.079
237358 97.489 -75.504 50.044
237359 97.515 -75.510 50.008
237360 97.537 -75.514 49.973
237361 97.560 -75.518 49.937
237362 97.583 -75.522 49.902
237363 97.606 -75.526 49.868
237364 97.629 -75.530 49.834
237365 97.651 -75.534 49.799
237366 97.674 -75.538 49.762
237367 97.697 -75.542 49.722
237368 97.720 -75.546 49.681
237369 97.742 -75.550 49.637
237370 97.799 -75.542 49.592
237371 97.864 -75.531 49.547
237372 97.909 -75.540 49.501
237373 97.935 -75.566 49.454
237374 97.961 -75.592 49.405
237375 97.987 -75.620 49.358
237376 98.026 -75.630 49.311
237377 98.071 -75.639 49.266
237378 98.116 -75.647 49.222
237379 98.141 -75.657 49.179
237380 98.158 -75.668 49.134
237381 98.175 -75.677 49.089
237382 98.193 -75.688 49.045
237383 98.210 -75.699 49.000
237384 98.227 -75.709 48.955
237385 98.247 -75.719 48.910
237386 98.268 -75.730 48.864
237387 98.290 -75.740 48.816
237388 98.325 -75.751 48.770
237389 98.359 -75.766 48.724
237390 98.380 -75.805 48.678
237391 98.400 -75.844 48.632
237392 98.421 -75.882 48.586
237393 98.492 -75.908 48.541
237394 98.573 -75.931 48.495
237395 98.645 -75.922 48.451
237396 98.637 -75.884 48.407
237397 98.648 -75.890 48.364
237398 98.660 -75.894 48.322
237399 98.684 -75.924 48.281
237400 98.715 -75.966 48.240
237401 98.748 -76.003 48.199
237402 98.802 -75.981 48.160
237403 98.827 -75.978 48.123
237404 98.852 -75.976 48.086
237405 98.878 -75.975 48.049
237406 98.903 -75.972 48.015
237407 98.928 -75.970 47.982
237408 98.953 -75.968 47.949
237409 98.979 -75.966 47.918
237410 99.004 -75.964 47.887
237411 99.021 -75.981 47.859
237412 99.038 -75.997 47.832
237413 99.055 -76.013 47.806
237414 99.072 -76.029 47.782
237415 99.089 -76.046 47.759
237416 99.107 -76.062 47.736
237417 99.133 -76.074 47.713
237418 99.160 -76.086 47.691
237419 99.188 -76.097 47.668
237420 99.216 -76.109 47.646
237421 99.243 -76.121 47.624
237422 99.271 -76.132 47.601
237423 99.299 -76.144 47.578
237424 99.327 -76.156 47.554
237425 99.355 -76.170 47.529
237426 99.383 -76.185 47.504
237427 99.412 -76.199 47.480
237428 99.441 -76.210 47.455
237429 99.471 -76.215 47.427
237430 99.501 -76.220 47.399
237431 99.518 -76.223 47.371
237432 99.534 -76.225 47.345
237433 99.549 -76.228 47.318
237434 99.565 -76.230 47.293
237435 99.581 -76.232 47.267
237436 99.596 -76.235 47.240
237437 99.612 -76.238 47.212
237438 99.628 -76.241 47.182
237439 99.643 -76.243 47.151
237440 99.659 -76.246 47.120
237441 99.675 -76.248 47.091
237442 99.691 -76.251 47.061
237443 99.715 -76.255 47.030
237444 99.739 -76.267 46.999
237445 99.754 -76.286 46.966
237446 99.770 -76.305 46.934
237447 99.787 -76.320 46.904
237448 99.804 -76.334 46.875
237449 99.836 -76.351 46.844
237450 99.864 -76.363 46.814
237451 99.884 -76.367 46.784
237452 99.904 -76.370 46.754
237453 99.924 -76.373 46.725
237454 99.944 -76.376 46.697
237455 99.965 -76.380 46.667
237456 99.985 -76.384 46.636
237457 100.003 -76.395 46.607
237458 100.021 -76.408 46.576
237459 100.055 -76.415 46.543
237460 100.092 -76.420 46.510
237461 100.130 -76.426 46.479
237462 100.168 -76.432 46.446
237463 100.206 -76.437 46.413
237464 100.244 -76.442 46.378
237465 100.282 -76.448 46.343
237466 100.312 -76.456 46.307
237467 100.330 -76.469 46.271
237468 100.347 -76.482 46.236
237469 100.364 -76.494 46.200
237470 100.381 -76.507 46.164
237471 100.398 -76.520 46.127
237472 100.415 -76.533 46.090
237473 100.432 -76.545 46.055
237474 100.450 -76.558 46.022
237475 100.467 -76.571 45.989
237476 100.484 -76.583 45.958
237477 100.501 -76.592 45.928
237478 100.520 -76.584 45.899
237479 100.540 -76.592 45.871
237480 100.559 -76.605 45.846
237481 100.578 -76.624 45.821
237482 100.597 -76.643 45.797
237483 100.616 -76.661 45.774
237484 100.642 -76.674 45.753
237485 100.687 -76.668 45.732
237486 100.732 -76.662 45.711
237487 100.777 -76.656 45.692
237488 100.821 -76.651 45.673
237489 100.866 -76.645 45.652
237490 100.896 -76.651 45.632
237491 100.919 -76.660 45.611
237492 100.943 -76.669 45.590
237493 100.967 -76.678 45.568
237494 100.991 -76.687 45.547
237495 101.015 -76.696 45.527
237496 101.039 -76.705 45.506
237497 101.066 -76.716 45.485
237498 101.094 -76.727 45.464
237499 101.122 -76.737 45.442
237500 101.149 -76.747 45.423
237501 101.172 -76.759 45.405
237502 101.187 -76.771 45.387
237503 101.202 -76.783 45.370
237504 101.216 -76.795 45.354
237505 101.231 -76.806 45.338
237506 101.246 -76.818 45.324
237507 101.261 -76.831 45.312
237508 101.275 -76.843 45.300
237509 101.290 -76.855 45.288
237510 101.305 -76.867 45.276
237511 101.320 -76.879 45.264
237512 101.334 -76.890 45.251
237513 101.349 -76.903 45.239
237514 101.363 -76.915 45.227
237515 101.379 -76.927 45.214
237516 101.393 -76.939 45.202
237517 101.408 -76.951 45.188
237518 101.423 -76.963 45.173
237519 101.438 -76.975 45.158
237520 101.452 -76.987 45.143
237521 101.469 -77.000 45.127
237522 101.492 -77.018 45.111
237523 101.515 -77.018 45.095
237524 101.538 -77.018 45.078
237525 101.560 -77.018 45.061
237526 101.583 -77.019 45.046
237527 101.606 -77.019 45.029
237528 101.629 -77.019 45.013
237529 101.652 -77.019 44.995
237530 101.672 -77.022 44.978
237531 101.691 -77.028 44.958
237532 101.710 -77.038 44.938
237533 101.730 -77.052 44.919
237534 101.749 -77.066 44.902
237535 101.769 -77.079 44.886
237536 101.788 -77.093 44.870
237537 101.809 -77.106 44.854
237538 101.828 -77.120 44.838
237539 101.848 -77.134 44.823
237540 101.868 -77.147 44.806
237541 101.887 -77.161 44.789
237542 101.907 -77.175 44.773
237543 101.926 -77.188 44.757
237544 101.950 -77.211 44.740
237545 101.973 -77.235 44.725
237546 101.996 -77.259 44.710
237547 102.019 -77.283 44.695
237548 102.042 -77.307 44.680
237549 102.065 -77.330 44.667
237550 102.088 -77.351 44.652
237551 102.110 -77.370 44.637
237552 102.131 -77.390 44.622
237553 102.153 -77.410 44.606
237554 102.175 -77.429 44.588
237555 102.194 -77.442 44.571
237556 102.211 -77.453 44.555
237557 102.229 -77.464 44.539
237558 102.246 -77.474 44.522
237559 102.264 -77.486 44.505
237560 102.281 -77.496 44.489
237561 102.300 -77.506 44.472
237562 102.320 -77.515 44.456
237563 102.340 -77.524 44.440
237564 102.360 -77.533 44.422
237565 102.389 -77.545 44.405
237566 102.410 -77.556 44.387
237567 102.436 -77.560 44.369
237568 102.470 -77.565 44.351
237569 102.507 -77.570 44.333
237570 102.543 -77.576 44.315
237571 102.579 -77.581 44.297
237572 102.615 -77.586 44.279
237573 102.652 -77.592 44.262
237574 102.688 -77.597 44.245
237575 102.725 -77.602 44.229
237576 102.761 -77.608 44.213
237577 102.797 -77.613 44.196
237578 102.833 -77.618 44.181
237579 102.870 -77.624 44.165
237580 102.906 -77.629 44.148
237581 102.942 -77.634 44.133
237582 102.979 -77.639 44.117
237583 103.015 -77.645 44.101
237584 103.051 -77.651 44.084
237585 103.088 -77.656 44.069
237586 103.124 -77.661 44.053
237849 92.293 -74.699 54.369
237850 92.317 -74.710 54.377
237851 92.342 -74.721 54.384
237852 92.366 -74.731 54.393
237853 92.390 -74.742 54.404
237854 92.415 -74.753 54.415
237855 92.439 -74.764 54.425
237856 92.464 -74.775 54.435
237857 92.488 -74.785 54.444
237858 92.503 -74.767 54.453
237859 92.527 -74.755 54.461
237860 92.554 -74.746 54.469
237861 92.580 -74.737 54.475
237862 92.606 -74.727 54.482
237863 92.636 -74.718 54.487
237864 92.667 -74.712 54.490
237865 92.700 -74.709 54.493
237866 92.733 -74.706 54.495
237867 92.766 -74.702 54.494
237868 92.800 -74.702 54.492
237869 92.834 -74.705 54.488
237870 92.867 -74.709 54.483
237871 92.901 -74.712 54.479
237872 92.935 -74.715 54.474
237873 92.968 -74.722 54.466
237874 93.001 -74.728 54.456
237875 93.035 -74.734 54.446
237876 93.068 -74.739 54.436
237877 93.101 -74.745 54.429
237878 93.134 -74.751 54.420
237879 93.167 -74.756 54.412
237880 93.191 -74.747 54.403
237881 93.214 -74.737 54.395
237882 93.237 -74.727 54.387
237883 93.261 -74.717 54.380
237884 93.284 -74.706 54.374
237885 93.307 -74.697 54.369
237886 93.331 -74.687 54.365
237887 93.354 -74.677 54.362
237888 93.377 -74.667 54.362
237889 93.401 -74.657 54.364
237890 93.424 -74.647 54.369
237891 93.447 -74.637 54.375
237892 93.471 -74.627 54.379
237893 93.494 -74.617 54.385
237894 93.517 -74.607 54.391
237895 93.541 -74.597 54.397
237896 93.564 -74.587 54.404
237897 93.587 -74.579 54.412
237898 93.618 -74.587 54.422
237899 93.651 -74.599 54.434
237900 93.683 -74.612 54.446
237901 93.714 -74.624 54.461
237902 93.747 -74.637 54.478
237903 93.778 -74.650 54.498
237904 93.810 -74.662 54.519
237905 93.842 -74.675 54.537
237906 93.873 -74.687 54.552
237907 93.905 -74.700 54.563
237908 93.937 -74.712 54.572
237909 93.962 -74.717 54.582
237910 93.979 -74.715 54.594
237911 93.996 -74.712 54.609
237912 94.012 -74.708 54.624
237913 94.028 -74.703 54.634
237914 94.044 -74.697 54.642
237915 94.059 -74.692 54.646
237916 94.075 -74.687 54.648
237917 94.091 -74.682 54.647
237918 94.107 -74.677 54.646
237919 94.123 -74.671 54.643
237920 94.139 -74.666 54.632
237921 94.159 -74.667 54.620
237922 94.180 -74.668 54.606
237923 94.201 -74.668 54.587
237924 94.223 -74.666 54.563
237925 94.246 -74.668 54.537
237926 94.269 -74.670 54.507
237927 94.292 -74.672 54.475
237928 94.316 -74.675 54.442
237929 94.337 -74.677 54.411
237930 94.356 -74.679 54.381
237931 94.376 -74.681 54.355
237932 94.395 -74.683 54.332
237933 94.413 -74.685 54.308
237934 94.430 -74.689 54.282
237935 94.448 -74.692 54.254
237936 94.465 -74.695 54.225
237937 94.484 -74.697 54.197
237938 94.505 -74.700 54.171
237939 94.527 -74.702 54.146
237940 94.548 -74.704 54.118
237941 94.569 -74.706 54.090
237942 94.590 -74.708 54.061
237943 94.609 -74.708 54.033
237944 94.632 -74.711 54.011
237945 94.661 -74.717 53.995
237946 94.689 -74.723 53.983
237947 94.717 -74.729 53.967
237948 94.745 -74.735 53.947
237949 94.773 -74.741 53.923
237950 94.802 -74.747 53.897
237951 94.830 -74.753 53.867
237952 94.858 -74.759 53.831
237953 94.886 -74.766 53.791
237954 94.914 -74.772 53.748
237955 94.943 -74.778 53.702
237956 94.971 -74.784 53.656
237957 94.999 -74.790 53.608
237958 95.027 -74.796 53.557
237959 95.045 -74.799 53.503
237960 95.039 -74.789 53.447
237961 95.033 -74.779 53.389
237962 95.028 -74.769 53.332
237963 95.022 -74.759 53.276
237964 95.016 -74.749 53.217
237965 95.010 -74.739 53.157
237966 95.017 -74.735 53.096
237967 95.049 -74.742 53.030
237968 95.081 -74.749 52.965
237969 95.115 -74.765 52.901
237970 95.148 -74.780 52.840
237971 95.182 -74.796 52.781
237972 95.215 -74.811 52.724
237973 95.249 -74.827 52.667
237974 95.282 -74.842 52.610
237975 95.312 -74.854 52.551
237976 95.340 -74.865 52.493
237977 95.368 -74.875 52.434
237978 95.395 -74.885 52.375
237979 95.423 -74.896 52.318
237980 95.451 -74.907 52.262
237981 95.477 -74.912 52.207
237982 95.501 -74.911 52.155
237983 95.527 -74.910 52.107
237984 95.551 -74.909 52.060
237985 95.583 -74.888 52.015
237986 95.615 -74.862 51.971
237987 95.648 -74.836 51.929
237988 95.672 -74.820 51.889
237989 95.690 -74.813 51.850
237990 95.708 -74.807 51.813
237991 95.726 -74.800 51.777
237992 95.744 -74.794 51.743
237993 95.764 -74.791 51.709
237994 95.805 -74.817 51.678
237995 95.845 -74.843 51.647
237996 95.886 -74.869 51.616
237997 95.926 -74.894 51.587
237998 95.967 -74.920 51.559
237999 96.005 -74.935 51.532
238000 96.036 -74.941 51.505
238001 96.067 -74.947 51.480
238002 96.100 -74.954 51.456
238003 96.131 -74.960 51.432
238004 96.163 -74.966 51.408
238005 96.195 -74.973 51.387
238006 96.227 -74.979 51.368
238007 96.258 -74.985 51.350
238008 96.290 -74.992 51.334
238009 96.322 -74.998 51.318
238010 96.354 -75.005 51.302
238011 96.385 -75.011 51.285
238012 96.417 -75.018 51.268
238013 96.449 -75.024 51.250
238014 96.469 -75.035 51.231
238015 96.484 -75.049 51.211
238016 96.499 -75.063 51.189
238017 96.517 -75.077 51.166
238018 96.537 -75.091 51.144
238019 96.558 -75.106 51.121
238020 96.579 -75.120 51.098
238021 96.605 -75.130 51.077
238022 96.635 -75.136 51.056
238023 96.666 -75.142 51.035
238024 96.696 -75.147 51.015
238025 96.727 -75.153 50.994
238026 96.757 -75.158 50.972
238027 96.788 -75.163 50.949
238028 96.813 -75.175 50.927
238029 96.838 -75.185 50.907
238030 96.864 -75.197 50.889
238031 96.889 -75.208 50.871
238032 96.914 -75.219 50.852
238033 96.940 -75.230 50.832
238034 96.932 -75.327 50.809
238035 96.958 -75.330 50.784
238036 96.984 -75.333 50.758
238037 97.010 -75.336 50.731
238038 97.036 -75.339 50.704
238039 97.063 -75.342 50.677
238040 97.089 -75.345 50.651
238041 97.115 -75.348 50.623
238042 97.141 -75.351 50.596
238043 97.167 -75.354 50.568
238044 97.193 -75.357 50.540
238045 97.219 -75.362 50.509
238046 97.244 -75.373 50.477
238047 97.261 -75.382 50.445
238048 97.278 -75.390 50.411
238049 97.295 -75.398 50.378
238050 97.312 -75.406 50.346
238051 97.329 -75.414 50.316
238052 97.346 -75.423 50.287
238053 97.363 -75.430 50.258
238054 97.379 -75.439 50.229
238055 97.396 -75.447 50.198
238056 97.413 -75.455 50.167
238057 97.436 -75.462 50.134
238058 97.467 -75.469 50.100
238059 97.494 -75.475 50.065
238060 97.519 -75.481 50.029
238061 97.545 -75.488 49.992
238062 97.568 -75.492 49.956
238063 97.591 -75.496 49.921
238064 97.613 -75.500 49.886
238065 97.637 -75.504 49.852
238066 97.659 -75.508 49.817
238067 97.682 -75.512 49.779
238068 97.705 -75.516 49.739
238069 97.728 -75.520 49.696
238070 97.754 -75.523 49.652
238071 97.819 -75.511 49.606
238072 97.876 -75.508 49.560
238073 97.915 -75.524 49.514
238074 97.944 -75.548 49.466
238075 97.969 -75.574 49.418
238076 97.996 -75.599 49.370
238077 98.042 -75.608 49.324
238078 98.088 -75.616 49.280
238079 98.135 -75.624 49.235
238080 98.163 -75.634 49.191
238081 98.180 -75.645 49.146
238082 98.198 -75.655 49.101
238083 98.215 -75.665 49.056
238084 98.232 -75.676 49.010
238085 98.249 -75.686 48.965
238086 98.267 -75.696 48.919
238087 98.291 -75.707 48.873
238088 98.327 -75.718 48.826
238089 98.363 -75.730 48.780
238090 98.400 -75.741 48.734
238091 98.425 -75.772 48.688
238092 98.445 -75.811 48.641
238093 98.465 -75.850 48.595
238094 98.486 -75.888 48.549
238095 98.506 -75.927 48.503
238096 98.658 -75.887 48.458
238097 98.656 -75.862 48.413
238098 98.667 -75.868 48.370
238099 98.679 -75.873 48.327
238100 98.691 -75.878 48.284
238101 98.710 -75.897 48.242
238102 98.741 -75.939 48.200
238103 98.789 -75.933 48.161
238104 98.836 -75.916 48.123
238105 98.861 -75.914 48.085
238106 98.886 -75.912 48.048
238107 98.911 -75.910 48.013
238108 98.936 -75.908 47.979
238109 98.961 -75.908 47.946
238110 98.978 -75.924 47.914
238111 98.995 -75.941 47.883
238112 99.012 -75.956 47.855
238113 99.029 -75.973 47.827
238114 99.047 -75.989 47.801
238115 99.064 -76.006 47.776
238116 99.081 -76.022 47.752
238117 99.098 -76.038 47.728
238118 99.126 -76.050 47.705
238119 99.153 -76.062 47.683
238120 99.181 -76.073 47.659
238121 99.208 -76.085 47.636
238122 99.236 -76.097 47.614
238123 99.265 -76.112 47.592
238124 99.293 -76.126 47.568
238125 99.322 -76.141 47.543
238126 99.350 -76.155 47.518
238127 99.378 -76.170 47.492
238128 99.408 -76.180 47.466
238129 99.438 -76.185 47.441
238130 99.468 -76.190 47.415
238131 99.498 -76.195 47.387
238132 99.522 -76.199 47.359
238133 99.539 -76.201 47.332
238134 99.554 -76.204 47.306
238135 99.570 -76.207 47.282
238136 99.585 -76.209 47.257
238137 99.601 -76.212 47.231
238138 99.617 -76.214 47.202
238139 99.633 -76.217 47.174
238140 99.648 -76.220 47.143
238141 99.664 -76.222 47.112
238142 99.680 -76.225 47.082
238143 99.702 -76.229 47.051
238144 99.732 -76.233 47.021
238145 99.757 -76.242 46.990
238146 99.773 -76.263 46.957
238147 99.788 -76.282 46.924
238148 99.803 -76.303 46.894
238149 99.819 -76.321 46.865
238150 99.842 -76.336 46.835
238151 99.865 -76.342 46.805
238152 99.885 -76.346 46.775
238153 99.905 -76.349 46.746
238154 99.925 -76.353 46.716
238155 99.945 -76.357 46.688
238156 99.966 -76.360 46.659
238157 99.986 -76.364 46.629
238158 100.005 -76.373 46.599
238159 100.043 -76.379 46.569
238160 100.081 -76.384 46.537
238161 100.119 -76.390 46.504
238162 100.157 -76.395 46.472
238163 100.195 -76.401 46.440
238164 100.232 -76.406 46.407
238165 100.270 -76.412 46.373
238166 100.308 -76.417 46.338
238167 100.332 -76.428 46.303
238168 100.349 -76.440 46.266
238169 100.366 -76.453 46.232
238170 100.384 -76.466 46.195
238171 100.401 -76.479 46.158
238172 100.418 -76.491 46.122
238173 100.435 -76.504 46.085
238174 100.452 -76.517 46.050
238175 100.469 -76.530 46.016
238176 100.486 -76.542 45.984
238177 100.503 -76.555 45.952
238178 100.521 -76.567 45.922
238179 100.538 -76.580 45.893
238180 100.557 -76.586 45.865
238181 100.576 -76.605 45.839
238182 100.595 -76.624 45.814
238183 100.614 -76.643 45.790
238184 100.633 -76.661 45.768
238185 100.668 -76.665 45.746
238186 100.713 -76.659 45.725
238187 100.758 -76.654 45.705
238188 100.802 -76.648 45.686
238189 100.847 -76.643 45.667
238190 100.892 -76.636 45.647
238191 100.926 -76.638 45.626
238192 100.950 -76.648 45.604
238193 100.974 -76.657 45.583
238194 100.998 -76.666 45.562
238195 101.022 -76.675 45.541
238196 101.046 -76.684 45.519
238197 101.070 -76.693 45.499
238198 101.097 -76.704 45.478
238199 101.124 -76.715 45.457
238200 101.152 -76.724 45.436
238201 101.180 -76.735 45.417
238202 101.200 -76.746 45.398
238203 101.215 -76.759 45.380
238204 101.230 -76.771 45.362
238205 101.245 -76.783 45.347
238206 101.259 -76.795 45.332
238207 101.274 -76.807 45.318
238208 101.289 -76.819 45.305
238209 101.304 -76.831 45.293
238210 101.318 -76.843 45.281
238211 101.333 -76.855 45.268
238212 101.348 -76.867 45.256
238213 101.362 -76.879 45.243
238214 101.377 -76.891 45.231
238215 101.392 -76.903 45.218
238216 101.407 -76.915 45.206
238217 101.421 -76.927 45.193
238218 101.436 -76.939 45.179
238219 101.451 -76.951 45.164
238220 101.466 -76.963 45.149
238221 101.480 -76.975 45.133
238222 101.495 -76.987 45.117
238223 101.513 -76.993 45.101
238224 101.535 -76.993 45.084
238225 101.557 -76.993 45.067
238226 101.579 -76.993 45.051
238227 101.600 -76.993 45.035
238228 101.622 -76.993 45.019
238229 101.645 -76.993 45.003
238230 101.668 -76.994 44.986
238231 101.689 -76.995 44.967
238232 101.709 -77.009 44.948
238233 101.729 -77.022 44.928
238234 101.748 -77.036 44.908
238235 101.768 -77.050 44.891
238236 101.788 -77.063 44.875
238237 101.808 -77.077 44.859
238238 101.827 -77.091 44.842
238239 101.847 -77.104 44.826
238240 101.867 -77.118 44.811
238241 101.886 -77.132 44.795
238242 101.906 -77.145 44.779
238243 101.926 -77.159 44.763
238244 101.946 -77.173 44.748
238245 101.969 -77.197 44.731
238246 101.992 -77.221 44.715
238247 102.016 -77.244 44.701
238248 102.039 -77.269 44.686
238249 102.062 -77.292 44.672
238250 102.084 -77.312 44.658
238251 102.106 -77.332 44.643
238252 102.127 -77.351 44.629
238253 102.149 -77.370 44.614
238254 102.171 -77.390 44.599
238255 102.192 -77.409 44.581
238256 102.212 -77.425 44.564
238257 102.230 -77.436 44.547
238258 102.247 -77.446 44.531
238259 102.265 -77.457 44.515
238260 102.282 -77.467 44.499
238261 102.300 -77.479 44.482
238262 102.318 -77.489 44.464
238263 102.335 -77.500 44.448
238264 102.352 -77.511 44.432
238265 102.373 -77.520 44.414
238266 102.396 -77.531 44.396
238267 102.422 -77.536 44.378
238268 102.458 -77.542 44.359
238269 102.495 -77.548 44.340
238270 102.531 -77.553 44.321
238271 102.568 -77.558 44.304
238272 102.604 -77.564 44.286
238273 102.640 -77.569 44.269
238274 102.676 -77.574 44.251
238275 102.713 -77.580 44.235
238276 102.749 -77.585 44.219
238277 102.786 -77.590 44.203
238278 102.822 -77.595 44.188
238279 102.858 -77.601 44.171
238280 102.894 -77.606 44.155
238281 102.930 -77.611 44.139
238282 102.967 -77.617 44.124
238283 103.003 -77.622 44.108
238284 103.039 -77.627 44.091
238285 103.076 -77.633 44.075
238551 92.319 -74.678 54.381
238552 92.344 -74.688 54.388
238553 92.368 -74.699 54.396
238554 92.392 -74.709 54.408
238555 92.417 -74.721 54.418
238556 92.441 -74.731 54.428
238557 92.465 -74.742 54.437
238558 92.490 -74.753 54.445
238559 92.508 -74.744 54.453
238560 92.528 -74.728 54.461
238561 92.555 -74.718 54.469
238562 92.582 -74.709 54.477
238563 92.609 -74.700 54.484
238564 92.637 -74.692 54.489
238565 92.668 -74.684 54.494
238566 92.698 -74.675 54.497
238567 92.731 -74.672 54.499
238568 92.764 -74.669 54.500
238569 92.797 -74.668 54.498
238570 92.831 -74.672 54.496
238571 92.865 -74.675 54.494
238572 92.899 -74.681 54.493
238573 92.932 -74.686 54.491
238574 92.965 -74.692 54.485
238575 92.998 -74.698 54.477
238576 93.031 -74.704 54.468
238577 93.065 -74.709 54.459
238578 93.098 -74.715 54.452
238579 93.132 -74.722 54.444
238580 93.165 -74.727 54.437
238581 93.194 -74.726 54.429
238582 93.217 -74.716 54.422
238583 93.240 -74.706 54.415
238584 93.264 -74.696 54.409
238585 93.287 -74.686 54.405
238586 93.310 -74.676 54.401
238587 93.334 -74.666 54.397
238588 93.357 -74.656 54.394
238589 93.380 -74.646 54.395
238590 93.404 -74.637 54.399
238591 93.427 -74.626 54.405
238592 93.450 -74.616 54.412
238593 93.474 -74.606 54.418
238594 93.497 -74.596 54.422
238595 93.520 -74.587 54.427
238596 93.544 -74.577 54.434
238597 93.567 -74.567 54.440
238598 93.589 -74.559 54.445
238599 93.612 -74.556 54.452
238600 93.644 -74.568 54.460
238601 93.676 -74.581 54.470
238602 93.707 -74.593 54.482
238603 93.739 -74.606 54.496
238604 93.771 -74.618 54.513
238605 93.802 -74.631 54.531
238606 93.834 -74.643 54.546
238607 93.866 -74.656 54.558
238608 93.898 -74.668 54.569
238609 93.929 -74.681 54.581
238610 93.956 -74.688 54.593
238611 93.977 -74.689 54.606
238612 93.995 -74.687 54.622
238613 94.012 -74.685 54.638
238614 94.029 -74.683 54.649
238615 94.046 -74.680 54.657
238616 94.062 -74.678 54.661
238617 94.078 -74.673 54.661
238618 94.095 -74.668 54.659
238619 94.110 -74.663 54.656
238620 94.126 -74.658 54.651
238621 94.142 -74.653 54.638
238622 94.159 -74.649 54.625
238623 94.180 -74.650 54.611
238624 94.201 -74.648 54.591
238625 94.222 -74.646 54.567
238626 94.245 -74.646 54.542
238627 94.268 -74.649 54.515
238628 94.291 -74.650 54.484
238629 94.315 -74.653 54.452
238630 94.335 -74.655 54.422
238631 94.354 -74.657 54.393
238632 94.373 -74.659 54.367
238633 94.393 -74.661 54.346
238634 94.412 -74.663 54.324
238635 94.430 -74.665 54.298
238636 94.448 -74.668 54.269
238637 94.465 -74.672 54.238
238638 94.485 -74.675 54.208
238639 94.506 -74.676 54.177
238640 94.528 -74.678 54.150
238641 94.549 -74.681 54.120
238642 94.570 -74.683 54.089
238643 94.591 -74.685 54.059
238644 94.611 -74.686 54.028
238645 94.639 -74.692 54.003
238646 94.668 -74.698 53.983
238647 94.696 -74.704 53.967
238648 94.724 -74.710 53.952
238649 94.752 -74.716 53.933
238650 94.780 -74.722 53.911
238651 94.809 -74.728 53.886
238652 94.836 -74.734 53.855
238653 94.863 -74.740 53.820
238654 94.890 -74.747 53.781
238655 94.917 -74.753 53.738
238656 94.944 -74.759 53.695
238657 94.970 -74.766 53.650
238658 94.997 -74.772 53.605
238659 95.023 -74.778 53.556
238660 95.031 -74.773 53.503
238661 95.030 -74.762 53.447
238662 95.029 -74.751 53.390
238663 95.028 -74.740 53.331
238664 95.026 -74.731 53.275
238665 95.020 -74.720 53.217
238666 95.014 -74.710 53.157
238667 95.023 -74.707 53.096
238668 95.056 -74.718 53.030
238669 95.089 -74.734 52.965
238670 95.123 -74.750 52.900
238671 95.157 -74.765 52.841
238672 95.190 -74.781 52.782
238673 95.224 -74.796 52.724
238674 95.257 -74.812 52.668
238675 95.290 -74.827 52.610
238676 95.321 -74.840 52.552
238677 95.348 -74.850 52.494
238678 95.376 -74.860 52.436
238679 95.403 -74.871 52.377
238680 95.431 -74.881 52.320
238681 95.459 -74.891 52.265
238682 95.484 -74.892 52.211
238683 95.509 -74.891 52.159
238684 95.534 -74.890 52.110
238685 95.558 -74.889 52.063
238686 95.585 -74.883 52.018
238687 95.618 -74.857 51.974
238688 95.651 -74.831 51.932
238689 95.683 -74.805 51.892
238690 95.716 -74.779 51.853
238691 95.733 -74.772 51.814
238692 95.751 -74.766 51.778
238693 95.768 -74.759 51.743
238694 95.787 -74.753 51.709
238695 95.820 -74.769 51.678
238696 95.861 -74.795 51.646
238697 95.901 -74.821 51.615
238698 95.942 -74.847 51.586
238699 95.982 -74.873 51.557
238700 96.018 -74.909 51.529
238701 96.050 -74.922 51.502
238702 96.080 -74.933 51.476
238703 96.111 -74.944 51.451
238704 96.143 -74.952 51.427
238705 96.174 -74.959 51.405
238706 96.206 -74.965 51.384
238707 96.238 -74.971 51.365
238708 96.269 -74.978 51.347
238709 96.301 -74.984 51.331
238710 96.333 -74.991 51.317
238711 96.365 -74.997 51.301
238712 96.396 -75.004 51.285
238713 96.428 -75.010 51.267
238714 96.459 -75.016 51.249
238715 96.474 -75.030 51.231
238716 96.489 -75.044 51.212
238717 96.509 -75.058 51.193
238718 96.529 -75.073 51.172
238719 96.550 -75.087 51.150
238720 96.571 -75.101 51.128
238721 96.591 -75.116 51.106
238722 96.613 -75.129 51.085
238723 96.644 -75.135 51.065
238724 96.674 -75.141 51.044
238725 96.705 -75.146 51.024
238726 96.735 -75.151 51.002
238727 96.766 -75.157 50.980
238728 96.796 -75.163 50.957
238729 96.827 -75.168 50.935
238730 96.857 -75.174 50.916
238731 96.888 -75.179 50.898
238732 96.914 -75.190 50.880
238733 96.939 -75.201 50.862
238734 96.964 -75.212 50.842
238735 96.943 -75.346 50.820
238736 96.969 -75.349 50.795
238737 96.995 -75.352 50.768
238738 97.021 -75.355 50.742
238739 97.047 -75.358 50.717
238740 97.074 -75.361 50.691
238741 97.100 -75.364 50.666
238742 97.125 -75.364 50.639
238743 97.150 -75.356 50.614
238744 97.175 -75.348 50.587
238745 97.199 -75.339 50.560
238746 97.222 -75.338 50.528
238747 97.245 -75.345 50.496
238748 97.263 -75.354 50.464
238749 97.280 -75.362 50.430
238750 97.297 -75.370 50.397
238751 97.314 -75.378 50.365
238752 97.330 -75.386 50.334
238753 97.348 -75.394 50.305
238754 97.364 -75.402 50.277
238755 97.381 -75.411 50.248
238756 97.398 -75.418 50.219
238757 97.415 -75.427 50.188
238758 97.440 -75.433 50.155
238759 97.470 -75.440 50.121
238760 97.499 -75.447 50.086
238761 97.524 -75.453 50.049
238762 97.550 -75.459 50.011
238763 97.576 -75.466 49.975
238764 97.599 -75.470 49.939
238765 97.621 -75.474 49.904
238766 97.644 -75.478 49.870
238767 97.667 -75.482 49.834
238768 97.690 -75.486 49.797
238769 97.713 -75.490 49.755
238770 97.735 -75.494 49.711
238771 97.774 -75.493 49.666
238772 97.839 -75.481 49.621
238773 97.882 -75.493 49.574
238774 97.921 -75.508 49.527
238775 97.952 -75.530 49.479
238776 97.979 -75.555 49.431
238777 98.009 -75.577 49.383
238778 98.056 -75.586 49.338
238779 98.102 -75.594 49.293
238780 98.143 -75.602 49.249
238781 98.164 -75.612 49.205
238782 98.178 -75.629 49.158
238783 98.195 -75.640 49.112
238784 98.212 -75.649 49.066
238785 98.229 -75.660 49.020
238786 98.246 -75.671 48.974
238787 98.264 -75.681 48.928
238788 98.309 -75.690 48.882
238789 98.365 -75.697 48.837
238790 98.402 -75.709 48.790
238791 98.438 -75.720 48.744
238792 98.469 -75.740 48.698
238793 98.490 -75.779 48.652
238794 98.510 -75.818 48.605
238795 98.530 -75.856 48.558
238796 98.528 -75.865 48.512
238797 98.622 -75.859 48.466
238798 98.675 -75.840 48.421
238799 98.687 -75.846 48.377
238800 98.699 -75.851 48.333
238801 98.710 -75.856 48.288
238802 98.722 -75.862 48.245
238803 98.735 -75.869 48.203
238804 98.776 -75.886 48.161
238805 98.830 -75.864 48.122
238806 98.869 -75.852 48.085
238807 98.895 -75.850 48.048
238808 98.918 -75.852 48.012
238809 98.935 -75.868 47.977
238810 98.952 -75.884 47.943
238811 98.970 -75.900 47.911
238812 98.987 -75.916 47.880
238813 99.004 -75.933 47.850
238814 99.021 -75.949 47.822
238815 99.039 -75.966 47.795
238816 99.056 -75.981 47.770
238817 99.073 -75.998 47.746
238818 99.090 -76.014 47.722
238819 99.118 -76.025 47.697
238820 99.146 -76.038 47.673
238821 99.174 -76.053 47.649
238822 99.203 -76.068 47.626
238823 99.231 -76.082 47.604
238824 99.260 -76.097 47.580
238825 99.288 -76.111 47.555
238826 99.317 -76.126 47.530
238827 99.346 -76.141 47.504
238828 99.374 -76.151 47.478
238829 99.405 -76.155 47.452
238830 99.434 -76.160 47.427
238831 99.465 -76.165 47.402
238832 99.494 -76.169 47.374
238833 99.524 -76.175 47.346
238834 99.543 -76.178 47.319
238835 99.559 -76.180 47.294
238836 99.574 -76.183 47.271
238837 99.590 -76.185 47.246
238838 99.606 -76.188 47.219
238839 99.622 -76.191 47.191
238840 99.637 -76.193 47.163
238841 99.653 -76.196 47.133
238842 99.668 -76.198 47.102
238843 99.689 -76.202 47.071
238844 99.719 -76.207 47.041
238845 99.749 -76.211 47.010
238846 99.775 -76.219 46.979
238847 99.791 -76.238 46.947
238848 99.806 -76.259 46.914
238849 99.822 -76.279 46.883
238850 99.837 -76.299 46.855
238851 99.854 -76.320 46.825
238852 99.874 -76.325 46.796
238853 99.895 -76.329 46.767
238854 99.915 -76.332 46.737
238855 99.936 -76.336 46.707
238856 99.956 -76.339 46.679
238857 99.977 -76.343 46.651
238858 99.999 -76.344 46.621
238859 100.031 -76.344 46.591
238860 100.070 -76.348 46.561
238861 100.108 -76.354 46.529
238862 100.145 -76.359 46.497
238863 100.184 -76.364 46.466
238864 100.221 -76.370 46.434
238865 100.259 -76.376 46.401
238866 100.297 -76.381 46.367
238867 100.335 -76.386 46.333
238868 100.352 -76.399 46.298
238869 100.369 -76.412 46.262
238870 100.387 -76.425 46.227
238871 100.404 -76.437 46.190
238872 100.421 -76.450 46.153
238873 100.438 -76.463 46.116
238874 100.455 -76.476 46.080
238875 100.472 -76.489 46.045
238876 100.489 -76.501 46.011
238877 100.506 -76.514 45.978
238878 100.524 -76.527 45.946
238879 100.540 -76.548 45.916
238880 100.558 -76.565 45.887
238881 100.574 -76.586 45.859
238882 100.593 -76.605 45.833
238883 100.612 -76.624 45.808
238884 100.631 -76.642 45.784
238885 100.650 -76.661 45.762
238886 100.694 -76.656 45.742
238887 100.738 -76.651 45.720
238888 100.783 -76.645 45.700
238889 100.828 -76.640 45.681
238890 100.873 -76.634 45.661
238891 100.917 -76.628 45.641
238892 100.957 -76.626 45.620
238893 100.981 -76.636 45.598
238894 101.005 -76.645 45.577
238895 101.029 -76.654 45.556
238896 101.053 -76.663 45.534
238897 101.076 -76.672 45.512
238898 101.101 -76.681 45.492
238899 101.127 -76.692 45.471
238900 101.156 -76.702 45.450
238901 101.183 -76.712 45.429
238902 101.211 -76.723 45.411
238903 101.229 -76.735 45.392
238904 101.244 -76.747 45.373
238905 101.258 -76.759 45.356
238906 101.273 -76.771 45.340
238907 101.288 -76.783 45.325
238908 101.302 -76.795 45.311
238909 101.317 -76.807 45.298
238910 101.332 -76.819 45.286
238911 101.347 -76.831 45.273
238912 101.361 -76.843 45.260
238913 101.376 -76.855 45.248
238914 101.391 -76.867 45.235
238915 101.406 -76.879 45.222
238916 101.420 -76.891 45.209
238917 101.435 -76.903 45.197
238918 101.450 -76.915 45.183
238919 101.465 -76.927 45.168
238920 101.479 -76.939 45.154
238921 101.494 -76.951 45.138
238922 101.509 -76.963 45.122
238923 101.528 -76.968 45.106
238924 101.549 -76.968 45.090
238925 101.571 -76.968 45.073
238926 101.593 -76.969 45.055
238927 101.614 -76.969 45.039
238928 101.636 -76.969 45.024
238929 101.658 -76.969 45.007
238930 101.679 -76.969 44.991
238931 101.699 -76.973 44.973
238932 101.715 -76.984 44.955
238933 101.731 -76.996 44.937
238934 101.748 -77.006 44.917
238935 101.767 -77.020 44.897
238936 101.787 -77.034 44.880
238937 101.807 -77.047 44.864
238938 101.826 -77.061 44.848
238939 101.846 -77.075 44.831
238940 101.866 -77.088 44.815
238941 101.886 -77.102 44.799
238942 101.905 -77.116 44.784
238943 101.925 -77.129 44.768
238944 101.945 -77.143 44.753
238945 101.965 -77.160 44.737
238946 101.988 -77.183 44.721
238947 102.012 -77.207 44.705
238948 102.035 -77.231 44.691
238949 102.058 -77.254 44.676
238950 102.080 -77.273 44.662
238951 102.101 -77.292 44.649
238952 102.123 -77.312 44.634
238953 102.145 -77.331 44.619
238954 102.167 -77.351 44.605
238955 102.188 -77.370 44.590
238956 102.210 -77.389 44.572
238957 102.231 -77.407 44.556
238958 102.249 -77.418 44.539
238959 102.266 -77.429 44.522
238960 102.283 -77.439 44.507
238961 102.301 -77.450 44.491
238962 102.319 -77.461 44.474
238963 102.336 -77.472 44.457
238964 102.354 -77.483 44.440
238965 102.371 -77.493 44.422
238966 102.389 -77.504 44.404
238967 102.408 -77.508 44.386
238968 102.444 -77.516 44.366
238969 102.482 -77.523 44.347
238970 102.519 -77.530 44.328
238971 102.556 -77.535 44.311
238972 102.592 -77.540 44.293
238973 102.628 -77.545 44.276
238974 102.665 -77.551 44.258
238975 102.701 -77.557 44.241
238976 102.737 -77.562 44.225
238977 102.773 -77.567 44.209
238978 102.810 -77.573 44.194
238979 102.846 -77.578 44.179
238980 102.883 -77.583 44.162
238981 102.919 -77.589 44.146
238982 102.955 -77.594 44.130
238983 102.992 -77.599 44.114
238984 103.028 -77.605 44.098
238985 103.064 -77.610 44.082
239253 92.346 -74.656 54.393
239254 92.370 -74.666 54.401
239255 92.394 -74.678 54.411
239256 92.419 -74.688 54.421
239257 92.443 -74.699 54.430
239258 92.467 -74.709 54.439
239259 92.492 -74.721 54.448
239260 92.512 -74.719 54.455
239261 92.530 -74.701 54.463
239262 92.557 -74.692 54.472
239263 92.583 -74.683 54.479
239264 92.610 -74.673 54.487
239265 92.639 -74.665 54.492
239266 92.669 -74.656 54.496
239267 92.699 -74.648 54.499
239268 92.729 -74.640 54.502
239269 92.762 -74.636 54.504
239270 92.795 -74.635 54.504
239271 92.829 -74.639 54.502
239272 92.862 -74.645 54.502
239273 92.896 -74.650 54.502
239274 92.929 -74.656 54.503
239275 92.962 -74.662 54.501
239276 92.995 -74.668 54.496
239277 93.029 -74.674 54.488
239278 93.062 -74.680 54.481
239279 93.095 -74.686 54.475
239280 93.129 -74.692 54.469
239281 93.162 -74.697 54.461
239282 93.195 -74.703 54.454
239283 93.220 -74.695 54.449
239284 93.243 -74.685 54.444
239285 93.266 -74.675 54.439
239286 93.290 -74.665 54.435
239287 93.313 -74.656 54.432
239288 93.336 -74.646 54.429
239289 93.360 -74.636 54.427
239290 93.383 -74.625 54.428
239291 93.406 -74.615 54.433
239292 93.430 -74.606 54.440
239293 93.453 -74.596 54.448
239294 93.476 -74.586 54.456
239295 93.500 -74.576 54.461
239296 93.523 -74.566 54.467
239297 93.547 -74.556 54.474
239298 93.570 -74.546 54.481
239299 93.591 -74.543 54.484
239300 93.613 -74.538 54.487
239301 93.636 -74.538 54.491
239302 93.668 -74.550 54.497
239303 93.700 -74.562 54.507
239304 93.731 -74.575 54.520
239305 93.763 -74.587 54.538
239306 93.795 -74.600 54.554
239307 93.827 -74.612 54.567
239308 93.858 -74.625 54.577
239309 93.890 -74.637 54.584
239310 93.922 -74.650 54.593
239311 93.950 -74.659 54.604
239312 93.972 -74.661 54.617
239313 93.993 -74.661 54.634
239314 94.011 -74.659 54.650
239315 94.028 -74.657 54.661
239316 94.045 -74.655 54.668
239317 94.061 -74.653 54.672
239318 94.078 -74.650 54.674
239319 94.095 -74.648 54.671
239320 94.112 -74.646 54.666
239321 94.129 -74.643 54.659
239322 94.145 -74.639 54.647
239323 94.161 -74.634 54.633
239324 94.180 -74.632 54.617
239325 94.201 -74.629 54.595
239326 94.222 -74.627 54.570
239327 94.244 -74.624 54.546
239328 94.267 -74.627 54.522
239329 94.290 -74.628 54.495
239330 94.314 -74.631 54.464
239331 94.333 -74.633 54.435
239332 94.352 -74.635 54.406
239333 94.371 -74.637 54.381
239334 94.391 -74.639 54.361
239335 94.410 -74.641 54.339
239336 94.429 -74.643 54.314
239337 94.447 -74.646 54.284
239338 94.465 -74.649 54.253
239339 94.486 -74.651 54.219
239340 94.507 -74.653 54.185
239341 94.529 -74.656 54.154
239342 94.550 -74.657 54.121
239343 94.571 -74.659 54.088
239344 94.592 -74.662 54.056
239345 94.615 -74.667 54.024
239346 94.642 -74.673 53.995
239347 94.669 -74.680 53.971
239348 94.696 -74.686 53.952
239349 94.722 -74.692 53.936
239350 94.749 -74.698 53.918
239351 94.776 -74.705 53.896
239352 94.803 -74.711 53.870
239353 94.830 -74.717 53.838
239354 94.857 -74.723 53.804
239355 94.884 -74.730 53.766
239356 94.911 -74.736 53.724
239357 94.937 -74.742 53.681
239358 94.964 -74.748 53.638
239359 94.988 -74.753 53.593
239360 95.012 -74.758 53.546
239361 95.011 -74.747 53.494
239362 95.010 -74.737 53.439
239363 95.009 -74.726 53.382
239364 95.008 -74.715 53.324
239365 95.007 -74.705 53.267
239366 95.006 -74.694 53.210
239367 95.005 -74.684 53.152
239368 95.030 -74.687 53.091
239369 95.063 -74.703 53.027
239370 95.097 -74.718 52.963
239371 95.131 -74.734 52.900
239372 95.164 -74.750 52.841
239373 95.198 -74.765 52.783
239374 95.232 -74.781 52.725
239375 95.265 -74.796 52.668
239376 95.299 -74.812 52.610
239377 95.329 -74.825 52.552
239378 95.356 -74.835 52.494
239379 95.384 -74.845 52.437
239380 95.412 -74.856 52.379
239381 95.440 -74.866 52.322
239382 95.466 -74.873 52.267
239383 95.491 -74.872 52.213
239384 95.516 -74.872 52.161
239385 95.541 -74.870 52.112
239386 95.566 -74.869 52.065
239387 95.591 -74.869 52.021
239388 95.620 -74.853 51.977
239389 95.653 -74.826 51.935
239390 95.686 -74.800 51.895
239391 95.719 -74.774 51.855
239392 95.751 -74.748 51.817
239393 95.776 -74.732 51.780
239394 95.794 -74.725 51.745
239395 95.812 -74.718 51.709
239396 95.836 -74.722 51.677
239397 95.877 -74.747 51.645
239398 95.917 -74.773 51.614
239399 95.958 -74.799 51.584
239400 95.994 -74.833 51.554
239401 96.030 -74.869 51.525
239402 96.065 -74.897 51.497
239403 96.096 -74.907 51.471
239404 96.126 -74.918 51.446
239405 96.157 -74.928 51.423
239406 96.188 -74.939 51.400
239407 96.219 -74.950 51.379
239408 96.249 -74.960 51.360
239409 96.281 -74.969 51.343
239410 96.312 -74.976 51.328
239411 96.344 -74.982 51.314
239412 96.376 -74.989 51.300
239413 96.407 -74.995 51.284
239414 96.439 -75.002 51.267
239415 96.464 -75.011 51.249
239416 96.480 -75.025 51.232
239417 96.501 -75.039 51.215
239418 96.521 -75.054 51.197
239419 96.542 -75.068 51.179
239420 96.562 -75.082 51.157
239421 96.583 -75.097 51.135
239422 96.604 -75.112 51.114
239423 96.624 -75.126 51.092
239424 96.653 -75.134 51.072
239425 96.683 -75.139 51.053
239426 96.714 -75.145 51.032
239427 96.744 -75.151 51.009
239428 96.774 -75.156 50.986
239429 96.805 -75.161 50.964
239430 96.835 -75.167 50.943
239431 96.866 -75.173 50.924
239432 96.896 -75.178 50.906
239433 96.927 -75.184 50.889
239434 96.957 -75.189 50.871
239435 96.987 -75.193 50.852
239436 96.954 -75.365 50.830
239437 96.979 -75.365 50.805
239438 97.004 -75.357 50.779
239439 97.028 -75.348 50.755
239440 97.053 -75.340 50.731
239441 97.078 -75.332 50.707
239442 97.102 -75.323 50.682
239443 97.126 -75.315 50.657
239444 97.151 -75.307 50.632
239445 97.176 -75.298 50.606
239446 97.198 -75.301 50.578
239447 97.219 -75.308 50.548
239448 97.241 -75.317 50.516
239449 97.265 -75.325 50.483
239450 97.282 -75.333 50.450
239451 97.299 -75.342 50.417
239452 97.315 -75.350 50.385
239453 97.332 -75.358 50.354
239454 97.349 -75.366 50.324
239455 97.366 -75.374 50.297
239456 97.383 -75.382 50.269
239457 97.400 -75.390 50.240
239458 97.417 -75.398 50.209
239459 97.443 -75.405 50.176
239460 97.473 -75.411 50.141
239461 97.504 -75.418 50.106
239462 97.529 -75.424 50.069
239463 97.555 -75.431 50.031
239464 97.581 -75.437 49.994
239465 97.606 -75.444 49.959
239466 97.630 -75.448 49.923
239467 97.652 -75.452 49.888
239468 97.675 -75.456 49.852
239469 97.698 -75.460 49.813
239470 97.721 -75.464 49.771
239471 97.743 -75.468 49.726
239472 97.794 -75.462 49.681
239473 97.848 -75.461 49.635
239474 97.887 -75.477 49.588
239475 97.927 -75.492 49.541
239476 97.961 -75.511 49.494
239477 97.987 -75.537 49.444
239478 98.023 -75.555 49.397
239479 98.069 -75.563 49.352
239480 98.114 -75.571 49.307
239481 98.139 -75.578 49.262
239482 98.164 -75.585 49.218
239483 98.178 -75.602 49.171
239484 98.191 -75.621 49.124
239485 98.205 -75.636 49.076
239486 98.223 -75.646 49.031
239487 98.240 -75.656 48.985
239488 98.257 -75.667 48.940
239489 98.315 -75.674 48.895
239490 98.374 -75.681 48.849
239491 98.433 -75.688 48.803
239492 98.476 -75.699 48.757
239493 98.512 -75.710 48.711
239494 98.534 -75.746 48.663
239495 98.555 -75.785 48.616
239496 98.555 -75.798 48.569
239497 98.544 -75.794 48.522
239498 98.643 -75.793 48.477
239499 98.689 -75.809 48.432
239500 98.701 -75.815 48.386
239501 98.713 -75.821 48.342
239502 98.725 -75.827 48.297
239503 98.737 -75.833 48.252
239504 98.750 -75.839 48.209
239505 98.774 -75.837 48.165
239506 98.826 -75.815 48.124
239507 98.876 -75.794 48.085
239508 98.895 -75.813 48.048
239509 98.912 -75.830 48.011
239510 98.929 -75.846 47.975
239511 98.947 -75.862 47.940
239512 98.964 -75.879 47.908
239513 98.982 -75.895 47.876
239514 98.999 -75.912 47.846
239515 99.016 -75.928 47.818
239516 99.034 -75.945 47.791
239517 99.052 -75.961 47.765
239518 99.069 -75.978 47.740
239519 99.086 -75.994 47.715
239520 99.113 -76.009 47.690
239521 99.141 -76.024 47.665
239522 99.170 -76.038 47.640
239523 99.198 -76.053 47.616
239524 99.227 -76.067 47.593
239525 99.255 -76.082 47.568
239526 99.284 -76.097 47.543
239527 99.312 -76.111 47.517
239528 99.342 -76.121 47.491
239529 99.371 -76.126 47.465
239530 99.402 -76.131 47.438
239531 99.431 -76.135 47.412
239532 99.462 -76.140 47.387
239533 99.491 -76.145 47.360
239534 99.521 -76.150 47.333
239535 99.548 -76.154 47.307
239536 99.563 -76.157 47.282
239537 99.579 -76.160 47.259
239538 99.595 -76.162 47.235
239539 99.611 -76.164 47.208
239540 99.626 -76.167 47.180
239541 99.642 -76.169 47.151
239542 99.657 -76.173 47.122
239543 99.676 -76.176 47.091
239544 99.706 -76.180 47.060
239545 99.735 -76.185 47.031
239546 99.765 -76.189 47.000
239547 99.794 -76.195 46.968
239548 99.809 -76.215 46.936
239549 99.825 -76.235 46.904
239550 99.840 -76.255 46.874
239551 99.856 -76.276 46.845
239552 99.875 -76.299 46.815
239553 99.896 -76.310 46.787
239554 99.916 -76.314 46.758
239555 99.936 -76.318 46.728
239556 99.957 -76.322 46.698
239557 99.977 -76.325 46.670
239558 100.001 -76.323 46.642
239559 100.025 -76.320 46.613
239560 100.053 -76.318 46.583
239561 100.093 -76.321 46.552
239562 100.134 -76.323 46.521
239563 100.172 -76.329 46.490
239564 100.210 -76.334 46.460
239565 100.248 -76.339 46.428
239566 100.286 -76.345 46.395
239567 100.324 -76.351 46.362
239568 100.355 -76.358 46.327
239569 100.372 -76.371 46.292
239570 100.389 -76.384 46.256
239571 100.406 -76.396 46.221
239572 100.423 -76.409 46.184
239573 100.440 -76.422 46.148
239574 100.458 -76.435 46.111
239575 100.475 -76.447 46.075
239576 100.492 -76.462 46.039
239577 100.508 -76.486 46.005
239578 100.524 -76.509 45.972
239579 100.540 -76.532 45.941
239580 100.559 -76.546 45.911
239581 100.579 -76.558 45.882
239582 100.591 -76.586 45.854
239583 100.610 -76.605 45.828
239584 100.629 -76.623 45.802
239585 100.648 -76.642 45.779
239586 100.675 -76.654 45.756
239587 100.719 -76.648 45.736
239588 100.764 -76.643 45.715
239589 100.809 -76.636 45.694
239590 100.853 -76.631 45.675
239591 100.898 -76.626 45.655
239592 100.943 -76.620 45.634
239593 100.988 -76.614 45.613
239594 101.012 -76.624 45.592
239595 101.036 -76.633 45.571
239596 101.060 -76.642 45.549
239597 101.084 -76.651 45.528
239598 101.108 -76.660 45.506
239599 101.131 -76.669 45.484
239600 101.159 -76.680 45.464
239601 101.186 -76.690 45.444
239602 101.214 -76.700 45.423
239603 101.242 -76.711 45.404
239604 101.257 -76.723 45.385
239605 101.272 -76.735 45.367
239606 101.286 -76.747 45.350
239607 101.302 -76.759 45.333
239608 101.316 -76.771 45.318
239609 101.331 -76.783 45.303
239610 101.345 -76.795 45.290
239611 101.360 -76.807 45.278
239612 101.375 -76.819 45.265
239613 101.390 -76.831 45.251
239614 101.404 -76.843 45.239
239615 101.419 -76.856 45.226
239616 101.434 -76.867 45.213
239617 101.449 -76.879 45.200
239618 101.463 -76.891 45.186
239619 101.478 -76.903 45.172
239620 101.493 -76.915 45.158
239621 101.507 -76.928 45.143
239622 101.522 -76.940 45.128
239623 101.542 -76.943 45.111
239624 101.563 -76.943 45.095
239625 101.585 -76.943 45.078
239626 101.607 -76.943 45.060
239627 101.628 -76.943 45.043
239628 101.650 -76.943 45.027
239629 101.672 -76.943 45.011
239630 101.693 -76.943 44.994
239631 101.710 -76.953 44.978
239632 101.727 -76.964 44.960
239633 101.743 -76.975 44.942
239634 101.759 -76.987 44.925
239635 101.776 -76.998 44.905
239636 101.792 -77.009 44.886
239637 101.809 -77.020 44.869
239638 101.826 -77.031 44.853
239639 101.845 -77.045 44.837
239640 101.865 -77.059 44.821
239641 101.885 -77.073 44.805
239642 101.904 -77.087 44.789
239643 101.924 -77.100 44.773
239644 101.944 -77.114 44.757
239645 101.964 -77.128 44.743
239646 101.985 -77.145 44.726
239647 102.008 -77.169 44.710
239648 102.031 -77.193 44.695
239649 102.054 -77.215 44.681
239650 102.076 -77.234 44.665
239651 102.097 -77.254 44.652
239652 102.119 -77.273 44.639
239653 102.140 -77.292 44.623
239654 102.163 -77.311 44.609
239655 102.184 -77.331 44.594
239656 102.206 -77.351 44.579
239657 102.228 -77.370 44.563
239658 102.249 -77.389 44.547
239659 102.267 -77.401 44.531
239660 102.285 -77.411 44.514
239661 102.302 -77.422 44.498
239662 102.320 -77.432 44.480
239663 102.337 -77.443 44.463
239664 102.355 -77.454 44.445
239665 102.373 -77.463 44.428
239666 102.393 -77.464 44.410
239667 102.412 -77.467 44.392
239668 102.433 -77.468 44.373
239669 102.452 -77.470 44.353
239670 102.486 -77.476 44.334
239671 102.524 -77.484 44.316
239672 102.562 -77.492 44.300
239673 102.600 -77.500 44.282
239674 102.638 -77.508 44.265
239675 102.676 -77.515 44.249
239676 102.715 -77.523 44.232
239677 102.752 -77.531 44.216
239678 102.790 -77.539 44.200
239679 102.828 -77.546 44.184
239680 102.867 -77.554 44.168
239681 102.904 -77.562 44.152
239682 102.943 -77.570 44.137
239683 102.980 -77.577 44.121
239684 103.016 -77.582 44.105
239685 103.052 -77.587 44.089
239955 92.372 -74.634 54.407
239956 92.396 -74.645 54.416
239957 92.421 -74.656 54.425
239958 92.445 -74.666 54.433
239959 92.470 -74.678 54.441
239960 92.494 -74.688 54.451
239961 92.517 -74.695 54.458
239962 92.532 -74.675 54.466
239963 92.559 -74.665 54.475
239964 92.585 -74.656 54.482
239965 92.612 -74.646 54.489
239966 92.640 -74.637 54.495
239967 92.670 -74.629 54.500
239968 92.701 -74.621 54.503
239969 92.731 -74.613 54.506
239970 92.761 -74.605 54.508
239971 92.793 -74.603 54.508
239972 92.826 -74.609 54.506
239973 92.859 -74.615 54.506
239974 92.893 -74.621 54.507
239975 92.926 -74.627 54.510
239976 92.959 -74.633 54.510
239977 92.993 -74.639 54.509
239978 93.026 -74.644 54.507
239979 93.059 -74.650 54.502
239980 93.092 -74.656 54.498
239981 93.126 -74.662 54.493
239982 93.159 -74.668 54.487
239983 93.192 -74.674 54.482
239984 93.222 -74.675 54.477
239985 93.246 -74.665 54.473
239986 93.269 -74.655 54.470
239987 93.292 -74.645 54.466
239988 93.315 -74.634 54.463
239989 93.339 -74.624 54.461
239990 93.362 -74.615 54.460
239991 93.386 -74.605 54.462
239992 93.409 -74.595 54.466
239993 93.432 -74.585 54.474
239994 93.456 -74.575 54.483
239995 93.479 -74.565 54.494
239996 93.502 -74.555 54.501
239997 93.526 -74.545 54.508
239998 93.549 -74.535 54.517
239999 93.571 -74.530 54.523
240000 93.592 -74.526 54.526
240001 93.614 -74.521 54.527
240002 93.635 -74.517 54.525
240003 93.660 -74.519 54.526
240004 93.692 -74.532 54.534
240005 93.724 -74.544 54.548
240006 93.756 -74.557 54.566
240007 93.787 -74.569 54.582
240008 93.819 -74.581 54.597
240009 93.851 -74.594 54.606
240010 93.883 -74.606 54.609
240011 93.915 -74.619 54.614
240012 93.944 -74.629 54.621
240013 93.966 -74.631 54.631
240014 93.988 -74.634 54.647
240015 94.008 -74.634 54.662
240016 94.027 -74.632 54.673
240017 94.044 -74.630 54.681
240018 94.060 -74.628 54.685
240019 94.077 -74.624 54.685
240020 94.094 -74.622 54.682
240021 94.111 -74.620 54.676
240022 94.128 -74.618 54.668
240023 94.145 -74.615 54.655
240024 94.162 -74.613 54.640
240025 94.179 -74.611 54.623
240026 94.201 -74.610 54.601
240027 94.222 -74.607 54.579
240028 94.244 -74.605 54.557
240029 94.266 -74.605 54.535
240030 94.289 -74.607 54.509
240031 94.311 -74.609 54.479
240032 94.331 -74.611 54.450
240033 94.350 -74.613 54.420
240034 94.369 -74.615 54.395
240035 94.389 -74.617 54.374
240036 94.408 -74.619 54.352
240037 94.427 -74.621 54.327
240038 94.446 -74.623 54.298
240039 94.466 -74.625 54.265
240040 94.487 -74.628 54.229
240041 94.508 -74.631 54.192
240042 94.529 -74.633 54.157
240043 94.550 -74.636 54.121
240044 94.571 -74.638 54.086
240045 94.591 -74.642 54.052
240046 94.609 -74.650 54.018
240047 94.636 -74.656 53.987
240048 94.663 -74.662 53.959
240049 94.690 -74.668 53.937
240050 94.716 -74.675 53.919
240051 94.744 -74.681 53.900
240052 94.770 -74.687 53.877
240053 94.797 -74.693 53.851
240054 94.824 -74.700 53.818
240055 94.851 -74.706 53.783
240056 94.878 -74.712 53.746
240057 94.905 -74.718 53.706
240058 94.929 -74.724 53.663
240059 94.953 -74.729 53.621
240060 94.977 -74.734 53.577
240061 94.991 -74.733 53.530
240062 94.990 -74.722 53.481
240063 94.989 -74.712 53.428
240064 94.989 -74.701 53.371
240065 94.988 -74.690 53.316
240066 94.987 -74.680 53.260
240067 94.986 -74.668 53.204
240068 95.009 -74.658 53.148
240069 95.041 -74.669 53.088
240070 95.071 -74.687 53.025
240071 95.105 -74.703 52.963
240072 95.139 -74.718 52.901
240073 95.172 -74.734 52.842
240074 95.206 -74.750 52.785
240075 95.240 -74.766 52.727
240076 95.273 -74.781 52.668
240077 95.307 -74.797 52.610
240078 95.337 -74.809 52.552
240079 95.365 -74.820 52.495
240080 95.393 -74.831 52.438
240081 95.420 -74.841 52.381
240082 95.448 -74.851 52.324
240083 95.474 -74.853 52.269
240084 95.498 -74.853 52.216
240085 95.523 -74.851 52.164
240086 95.548 -74.850 52.115
240087 95.573 -74.850 52.069
240088 95.598 -74.849 52.024
240089 95.623 -74.847 51.980
240090 95.656 -74.822 51.939
240091 95.688 -74.795 51.899
240092 95.721 -74.769 51.859
240093 95.753 -74.743 51.820
240094 95.786 -74.717 51.783
240095 95.819 -74.691 51.747
240096 95.836 -74.684 51.711
240097 95.854 -74.678 51.678
240098 95.892 -74.700 51.645
240099 95.933 -74.725 51.614
240100 95.971 -74.757 51.582
240101 96.007 -74.793 51.551
240102 96.043 -74.828 51.521
240103 96.079 -74.863 51.492
240104 96.111 -74.882 51.465
240105 96.142 -74.893 51.440
240106 96.173 -74.903 51.417
240107 96.203 -74.913 51.395
240108 96.234 -74.924 51.373
240109 96.265 -74.935 51.354
240110 96.296 -74.945 51.338
240111 96.326 -74.956 51.323
240112 96.357 -74.966 51.310
240113 96.388 -74.977 51.296
240114 96.418 -74.988 51.281
240115 96.450 -74.994 51.265
240116 96.472 -75.006 51.249
240117 96.493 -75.020 51.233
240118 96.513 -75.035 51.217
240119 96.534 -75.049 51.200
240120 96.554 -75.063 51.182
240121 96.575 -75.078 51.162
240122 96.596 -75.093 51.142
240123 96.616 -75.107 51.121
240124 96.637 -75.122 51.100
240125 96.661 -75.132 51.080
240126 96.692 -75.138 51.060
240127 96.722 -75.144 51.040
240128 96.753 -75.149 51.016
240129 96.783 -75.155 50.993
240130 96.814 -75.160 50.972
240131 96.844 -75.166 50.952
240132 96.875 -75.172 50.932
240133 96.905 -75.177 50.915
240134 96.934 -75.179 50.898
240135 96.957 -75.174 50.880
240136 96.980 -75.169 50.860
240137 96.956 -75.323 50.839
240138 96.980 -75.315 50.817
240139 97.005 -75.307 50.793
240140 97.029 -75.298 50.770
240141 97.054 -75.290 50.746
240142 97.079 -75.282 50.723
240143 97.103 -75.273 50.698
240144 97.127 -75.265 50.674
240145 97.152 -75.258 50.650
240146 97.173 -75.265 50.624
240147 97.195 -75.273 50.597
240148 97.216 -75.279 50.567
240149 97.238 -75.287 50.537
240150 97.262 -75.296 50.505
240151 97.284 -75.305 50.472
240152 97.300 -75.313 50.439
240153 97.317 -75.321 50.407
240154 97.334 -75.330 50.376
240155 97.351 -75.337 50.346
240156 97.368 -75.345 50.319
240157 97.385 -75.354 50.291
240158 97.402 -75.362 50.262
240159 97.419 -75.370 50.231
240160 97.446 -75.377 50.197
240161 97.477 -75.383 50.162
240162 97.508 -75.389 50.127
240163 97.534 -75.396 50.089
240164 97.560 -75.402 50.051
240165 97.586 -75.409 50.015
240166 97.611 -75.415 49.979
240167 97.637 -75.421 49.942
240168 97.660 -75.426 49.907
240169 97.683 -75.430 49.869
240170 97.706 -75.434 49.829
240171 97.728 -75.438 49.786
240172 97.751 -75.442 49.742
240173 97.814 -75.432 49.696
240174 97.854 -75.446 49.650
240175 97.893 -75.461 49.603
240176 97.933 -75.476 49.556
240177 97.970 -75.493 49.508
240178 97.996 -75.519 49.458
240179 98.036 -75.532 49.411
240180 98.083 -75.540 49.366
240181 98.110 -75.548 49.321
240182 98.135 -75.555 49.277
240183 98.160 -75.561 49.232
240184 98.179 -75.574 49.185
240185 98.192 -75.593 49.137
240186 98.204 -75.611 49.090
240187 98.217 -75.630 49.044
240188 98.233 -75.642 48.998
240189 98.262 -75.652 48.953
240190 98.320 -75.659 48.907
240191 98.380 -75.666 48.862
240192 98.439 -75.673 48.816
240193 98.498 -75.681 48.771
240194 98.551 -75.689 48.725
240195 98.579 -75.714 48.676
240196 98.582 -75.731 48.629
240197 98.571 -75.727 48.581
240198 98.560 -75.724 48.533
240199 98.619 -75.742 48.488
240200 98.686 -75.765 48.442
240201 98.710 -75.775 48.396
240202 98.722 -75.780 48.352
240203 98.734 -75.786 48.306
240204 98.746 -75.792 48.260
240205 98.759 -75.798 48.214
240206 98.799 -75.784 48.169
240207 98.852 -75.763 48.126
240208 98.879 -75.753 48.085
240209 98.897 -75.778 48.047
240210 98.914 -75.803 48.009
240211 98.931 -75.827 47.973
240212 98.948 -75.847 47.938
240213 98.966 -75.863 47.904
240214 98.983 -75.880 47.873
240215 99.001 -75.896 47.843
240216 99.018 -75.912 47.814
240217 99.035 -75.929 47.786
240218 99.053 -75.945 47.760
240219 99.070 -75.962 47.734
240220 99.088 -75.978 47.708
240221 99.108 -75.994 47.683
240222 99.137 -76.009 47.657
240223 99.165 -76.023 47.630
240224 99.194 -76.038 47.605
240225 99.222 -76.053 47.582
240226 99.251 -76.067 47.556
240227 99.279 -76.082 47.531
240228 99.308 -76.091 47.505
240229 99.338 -76.096 47.478
240230 99.368 -76.100 47.451
240231 99.398 -76.106 47.424
240232 99.428 -76.110 47.398
240233 99.458 -76.116 47.372
240234 99.488 -76.120 47.346
240235 99.518 -76.125 47.320
240236 99.548 -76.130 47.294
240237 99.568 -76.133 47.270
240238 99.584 -76.135 47.248
240239 99.600 -76.138 47.224
240240 99.615 -76.141 47.198
240241 99.631 -76.144 47.170
240242 99.647 -76.146 47.141
240243 99.663 -76.149 47.111
240244 99.693 -76.154 47.080
240245 99.722 -76.158 47.050
240246 99.752 -76.163 47.020
240247 99.781 -76.167 46.989
240248 99.811 -76.172 46.958
240249 99.827 -76.191 46.926
240250 99.843 -76.211 46.895
240251 99.858 -76.231 46.865
240252 99.877 -76.254 46.836
240253 99.897 -76.278 46.806
240254 99.916 -76.297 46.778
240255 99.937 -76.301 46.749
240256 99.958 -76.304 46.720
240257 99.978 -76.307 46.690
240258 100.003 -76.304 46.662
240259 100.028 -76.301 46.634
240260 100.052 -76.297 46.604
240261 100.077 -76.294 46.575
240262 100.116 -76.295 46.545
240263 100.156 -76.298 46.513
240264 100.196 -76.301 46.482
240265 100.237 -76.304 46.452
240266 100.274 -76.309 46.420
240267 100.312 -76.314 46.388
240268 100.350 -76.320 46.355
240269 100.374 -76.330 46.321
240270 100.392 -76.342 46.285
240271 100.409 -76.355 46.250
240272 100.426 -76.368 46.214
240273 100.443 -76.381 46.178
240274 100.460 -76.400 46.142
240275 100.476 -76.423 46.106
240276 100.492 -76.446 46.070
240277 100.508 -76.470 46.034
240278 100.524 -76.493 46.000
240279 100.540 -76.517 45.967
240280 100.561 -76.527 45.936
240281 100.582 -76.535 45.905
240282 100.597 -76.556 45.876
240283 100.608 -76.586 45.849
240284 100.627 -76.604 45.822
240285 100.646 -76.623 45.797
240286 100.665 -76.641 45.773
240287 100.700 -76.646 45.750
240288 100.745 -76.640 45.729
240289 100.789 -76.634 45.708
240290 100.834 -76.628 45.688
240291 100.879 -76.623 45.668
240292 100.924 -76.617 45.648
240293 100.968 -76.611 45.628
240294 101.013 -76.606 45.607
240295 101.043 -76.611 45.585
240296 101.066 -76.621 45.565
240297 101.091 -76.630 45.543
240298 101.114 -76.639 45.521
240299 101.138 -76.648 45.499
240300 101.162 -76.657 45.478
240301 101.189 -76.668 45.458
240302 101.217 -76.678 45.437
240303 101.245 -76.688 45.416
240304 101.271 -76.699 45.397
240305 101.285 -76.712 45.378
240306 101.300 -76.723 45.360
240307 101.315 -76.735 45.342
240308 101.330 -76.747 45.326
240309 101.344 -76.759 45.310
240310 101.359 -76.771 45.295
240311 101.374 -76.784 45.282
240312 101.389 -76.796 45.269
240313 101.403 -76.807 45.256
240314 101.418 -76.819 45.243
240315 101.433 -76.831 45.230
240316 101.448 -76.843 45.217
240317 101.462 -76.856 45.203
240318 101.477 -76.868 45.190
240319 101.492 -76.880 45.176
240320 101.506 -76.891 45.162
240321 101.521 -76.903 45.147
240322 101.536 -76.915 45.132
240323 101.556 -76.918 45.116
240324 101.578 -76.918 45.099
240325 101.599 -76.918 45.082
240326 101.621 -76.919 45.065
240327 101.642 -76.919 45.047
240328 101.664 -76.919 45.030
240329 101.686 -76.919 45.013
240330 101.706 -76.922 44.997
240331 101.722 -76.933 44.980
240332 101.739 -76.944 44.964
240333 101.755 -76.956 44.947
240334 101.771 -76.967 44.929
240335 101.787 -76.978 44.912
240336 101.804 -76.989 44.893
240337 101.820 -77.000 44.875
240338 101.837 -77.011 44.857
240339 101.853 -77.022 44.840
240340 101.870 -77.034 44.825
240341 101.886 -77.045 44.810
240342 101.904 -77.057 44.794
240343 101.923 -77.071 44.777
240344 101.943 -77.085 44.761
240345 101.963 -77.098 44.746
240346 101.982 -77.112 44.731
240347 102.004 -77.131 44.714
240348 102.027 -77.155 44.699
240349 102.049 -77.175 44.684
240350 102.071 -77.195 44.669
240351 102.093 -77.214 44.653
240352 102.115 -77.234 44.639
240353 102.136 -77.253 44.625
240354 102.158 -77.273 44.610
240355 102.180 -77.291 44.595
240356 102.202 -77.311 44.581
240357 102.224 -77.331 44.566
240358 102.245 -77.350 44.550
240359 102.267 -77.370 44.534
240360 102.286 -77.383 44.519
240361 102.304 -77.394 44.503
240362 102.321 -77.404 44.486
240363 102.338 -77.415 44.468
240364 102.358 -77.419 44.450
240365 102.378 -77.420 44.432
240366 102.397 -77.423 44.415
240367 102.417 -77.424 44.397
240368 102.437 -77.426 44.377
240369 102.457 -77.428 44.358
240370 102.477 -77.429 44.340
240371 102.497 -77.432 44.322
240372 102.528 -77.437 44.305
240373 102.567 -77.445 44.288
240374 102.604 -77.453 44.271
240375 102.643 -77.461 44.255
240376 102.680 -77.468 44.239
240377 102.719 -77.476 44.222
240378 102.756 -77.484 44.207
240379 102.795 -77.492 44.190
240380 102.832 -77.499 44.174
240381 102.871 -77.507 44.160
240382 102.908 -77.515 44.144
240383 102.947 -77.523 44.128
240384 102.984 -77.530 44.113
240385 103.023 -77.538 44.096
240657 92.398 -74.612 54.421
240658 92.422 -74.623 54.430
240659 92.447 -74.634 54.437
240660 92.471 -74.645 54.445
240661 92.496 -74.656 54.454
240662 92.520 -74.666 54.461
240663 92.534 -74.649 54.468
240664 92.560 -74.638 54.477
240665 92.587 -74.628 54.485
240666 92.614 -74.619 54.492
240667 92.641 -74.610 54.498
240668 92.671 -74.602 54.504
240669 92.702 -74.594 54.507
240670 92.732 -74.586 54.509
240671 92.762 -74.577 54.511
240672 92.792 -74.579 54.511
240673 92.824 -74.581 54.510
240674 92.856 -74.585 54.510
240675 92.890 -74.591 54.511
240676 92.923 -74.597 54.514
240677 92.956 -74.603 54.517
240678 92.990 -74.609 54.520
240679 93.023 -74.615 54.522
240680 93.056 -74.621 54.523
240681 93.089 -74.627 54.520
240682 93.123 -74.632 54.516
240683 93.156 -74.638 54.510
240684 93.189 -74.644 54.506
240685 93.223 -74.650 54.503
240686 93.249 -74.644 54.501
240687 93.272 -74.634 54.499
240688 93.295 -74.624 54.496
240689 93.319 -74.614 54.494
240690 93.342 -74.604 54.492
240691 93.365 -74.594 54.491
240692 93.389 -74.584 54.494
240693 93.412 -74.574 54.499
240694 93.435 -74.565 54.508
240695 93.459 -74.555 54.519
240696 93.482 -74.544 54.529
240697 93.505 -74.534 54.539
240698 93.529 -74.524 54.548
240699 93.551 -74.518 54.557
240700 93.572 -74.513 54.565
240701 93.594 -74.509 54.571
240702 93.615 -74.505 54.570
240703 93.637 -74.500 54.564
240704 93.658 -74.496 54.560
240705 93.685 -74.501 54.566
240706 93.716 -74.513 54.579
240707 93.749 -74.526 54.598
240708 93.780 -74.538 54.617
240709 93.812 -74.551 54.634
240710 93.844 -74.563 54.644
240711 93.875 -74.575 54.643
240712 93.907 -74.588 54.640
240713 93.938 -74.599 54.642
240714 93.960 -74.602 54.649
240715 93.982 -74.604 54.662
240716 94.003 -74.606 54.677
240717 94.024 -74.606 54.687
240718 94.042 -74.604 54.694
240719 94.059 -74.602 54.698
240720 94.076 -74.599 54.696
240721 94.093 -74.597 54.692
240722 94.110 -74.595 54.685
240723 94.127 -74.593 54.675
240724 94.144 -74.590 54.662
240725 94.161 -74.588 54.646
240726 94.179 -74.586 54.627
240727 94.200 -74.587 54.607
240728 94.222 -74.587 54.587
240729 94.244 -74.586 54.568
240730 94.265 -74.583 54.549
240731 94.288 -74.585 54.523
240732 94.309 -74.587 54.494
240733 94.328 -74.589 54.464
240734 94.348 -74.591 54.434
240735 94.367 -74.593 54.408
240736 94.386 -74.595 54.385
240737 94.406 -74.597 54.361
240738 94.425 -74.598 54.334
240739 94.445 -74.599 54.304
240740 94.466 -74.604 54.271
240741 94.487 -74.606 54.234
240742 94.508 -74.609 54.196
240743 94.529 -74.612 54.158
240744 94.550 -74.615 54.120
240745 94.571 -74.617 54.082
240746 94.589 -74.624 54.047
240747 94.607 -74.632 54.011
240748 94.630 -74.639 53.979
240749 94.657 -74.645 53.950
240750 94.684 -74.651 53.924
240751 94.710 -74.657 53.901
240752 94.738 -74.664 53.879
240753 94.764 -74.670 53.855
240754 94.791 -74.676 53.826
240755 94.818 -74.682 53.794
240756 94.845 -74.689 53.761
240757 94.871 -74.695 53.725
240758 94.895 -74.700 53.685
240759 94.919 -74.705 53.644
240760 94.943 -74.709 53.602
240761 94.966 -74.715 53.558
240762 94.971 -74.707 53.511
240763 94.970 -74.697 53.464
240764 94.969 -74.686 53.413
240765 94.969 -74.675 53.361
240766 94.968 -74.665 53.310
240767 94.969 -74.654 53.258
240768 94.993 -74.643 53.205
240769 95.018 -74.636 53.150
240770 95.051 -74.652 53.088
240771 95.086 -74.668 53.026
240772 95.113 -74.688 52.964
240773 95.146 -74.703 52.904
240774 95.180 -74.719 52.845
240775 95.214 -74.734 52.787
240776 95.247 -74.750 52.729
240777 95.281 -74.766 52.670
240778 95.315 -74.781 52.612
240779 95.346 -74.794 52.554
240780 95.374 -74.805 52.496
240781 95.401 -74.816 52.438
240782 95.428 -74.826 52.382
240783 95.456 -74.834 52.326
240784 95.481 -74.834 52.272
240785 95.506 -74.833 52.219
240786 95.531 -74.831 52.168
240787 95.555 -74.831 52.119
240788 95.580 -74.830 52.072
240789 95.605 -74.829 52.028
240790 95.630 -74.828 51.984
240791 95.658 -74.817 51.943
240792 95.691 -74.791 51.903
240793 95.724 -74.765 51.864
240794 95.756 -74.738 51.825
240795 95.789 -74.712 51.787
240796 95.821 -74.686 51.749
240797 95.854 -74.660 51.713
240798 95.880 -74.643 51.679
240799 95.908 -74.653 51.646
240800 95.947 -74.682 51.613
240801 95.983 -74.717 51.581
240802 96.019 -74.753 51.548
240803 96.055 -74.788 51.517
240804 96.091 -74.823 51.487
240805 96.127 -74.856 51.460
240806 96.157 -74.867 51.435
240807 96.188 -74.878 51.411
240808 96.219 -74.888 51.389
240809 96.249 -74.898 51.366
240810 96.280 -74.909 51.348
240811 96.311 -74.919 51.332
240812 96.341 -74.930 51.318
240813 96.372 -74.941 51.304
240814 96.403 -74.951 51.290
240815 96.433 -74.962 51.276
240816 96.457 -74.973 51.262
240817 96.481 -74.992 51.249
240818 96.504 -75.013 51.233
240819 96.526 -75.030 51.218
240820 96.546 -75.044 51.202
240821 96.567 -75.059 51.185
240822 96.588 -75.073 51.166
240823 96.608 -75.088 51.147
240824 96.629 -75.103 51.128
240825 96.650 -75.117 51.108
240826 96.670 -75.132 51.088
240827 96.700 -75.137 51.068
240828 96.731 -75.142 51.046
240829 96.761 -75.148 51.022
240830 96.792 -75.154 51.000
240831 96.822 -75.159 50.979
240832 96.853 -75.165 50.959
240833 96.881 -75.166 50.940
240834 96.904 -75.160 50.923
240835 96.927 -75.155 50.907
240836 96.951 -75.149 50.889
240837 96.974 -75.144 50.870
240838 96.957 -75.274 50.850
240839 96.981 -75.266 50.829
240840 97.006 -75.258 50.807
240841 97.030 -75.249 50.785
240842 97.055 -75.241 50.762
240843 97.080 -75.232 50.738
240844 97.104 -75.224 50.715
240845 97.127 -75.222 50.691
240846 97.149 -75.229 50.667
240847 97.170 -75.236 50.642
240848 97.192 -75.244 50.615
240849 97.213 -75.251 50.587
240850 97.235 -75.258 50.557
240851 97.259 -75.267 50.527
240852 97.283 -75.276 50.495
240853 97.302 -75.285 50.462
240854 97.319 -75.293 50.431
240855 97.336 -75.301 50.401
240856 97.353 -75.309 50.370
240857 97.370 -75.317 50.342
240858 97.387 -75.325 50.313
240859 97.403 -75.333 50.283
240860 97.421 -75.342 50.251
240861 97.450 -75.348 50.217
240862 97.480 -75.355 50.183
240863 97.511 -75.361 50.147
240864 97.539 -75.367 50.110
240865 97.565 -75.374 50.072
240866 97.591 -75.380 50.035
240867 97.616 -75.386 49.998
240868 97.642 -75.393 49.961
240869 97.667 -75.399 49.925
240870 97.691 -75.404 49.886
240871 97.714 -75.408 49.846
240872 97.736 -75.412 49.803
240873 97.769 -75.412 49.758
240874 97.820 -75.415 49.712
240875 97.860 -75.430 49.665
240876 97.899 -75.445 49.618
240877 97.938 -75.460 49.571
240878 97.978 -75.475 49.522
240879 98.004 -75.501 49.472
240880 98.050 -75.510 49.426
240881 98.081 -75.517 49.380
240882 98.106 -75.524 49.335
240883 98.131 -75.530 49.290
240884 98.156 -75.537 49.245
240885 98.181 -75.545 49.199
240886 98.194 -75.564 49.151
240887 98.206 -75.583 49.104
240888 98.218 -75.602 49.058
240889 98.230 -75.621 49.012
240890 98.272 -75.622 48.966
240891 98.331 -75.629 48.921
240892 98.390 -75.636 48.875
240893 98.450 -75.643 48.830
240894 98.509 -75.649 48.785
240895 98.568 -75.656 48.739
240896 98.603 -75.673 48.690
240897 98.587 -75.676 48.642
240898 98.572 -75.680 48.594
240899 98.557 -75.683 48.545
240900 98.597 -75.693 48.498
240901 98.661 -75.714 48.452
240902 98.718 -75.734 48.407
240903 98.730 -75.740 48.361
240904 98.743 -75.746 48.314
240905 98.755 -75.752 48.267
240906 98.773 -75.753 48.219
240907 98.826 -75.732 48.173
240908 98.859 -75.718 48.128
240909 98.881 -75.714 48.086
240910 98.899 -75.739 48.046
240911 98.916 -75.763 48.008
240912 98.933 -75.788 47.972
240913 98.950 -75.813 47.936
240914 98.968 -75.838 47.902
240915 98.985 -75.863 47.869
240916 99.002 -75.880 47.839
240917 99.019 -75.897 47.810
240918 99.037 -75.913 47.782
240919 99.054 -75.929 47.754
240920 99.072 -75.946 47.728
240921 99.089 -75.963 47.701
240922 99.106 -75.978 47.675
240923 99.132 -75.994 47.648
240924 99.160 -76.009 47.621
240925 99.189 -76.023 47.596
240926 99.217 -76.038 47.572
240927 99.246 -76.052 47.546
240928 99.275 -76.062 47.519
240929 99.305 -76.066 47.492
240930 99.335 -76.071 47.466
240931 99.365 -76.076 47.438
240932 99.395 -76.081 47.411
240933 99.425 -76.085 47.385
240934 99.455 -76.091 47.359
240935 99.485 -76.095 47.333
240936 99.515 -76.100 47.308
240937 99.545 -76.105 47.283
240938 99.573 -76.110 47.258
240939 99.588 -76.112 47.235
240940 99.604 -76.115 47.211
240941 99.620 -76.117 47.185
240942 99.636 -76.120 47.159
240943 99.651 -76.122 47.130
240944 99.680 -76.127 47.099
240945 99.709 -76.132 47.069
240946 99.739 -76.136 47.041
240947 99.768 -76.141 47.010
240948 99.798 -76.145 46.980
240949 99.828 -76.151 46.949
240950 99.846 -76.167 46.918
240951 99.861 -76.188 46.887
240952 99.879 -76.210 46.857
240953 99.899 -76.234 46.828
240954 99.918 -76.257 46.798
240955 99.937 -76.281 46.770
240956 99.958 -76.286 46.741
240957 99.980 -76.287 46.711
240958 100.005 -76.283 46.682
240959 100.030 -76.280 46.654
240960 100.054 -76.277 46.625
240961 100.079 -76.273 46.596
240962 100.104 -76.270 46.567
240963 100.137 -76.270 46.537
240964 100.177 -76.273 46.505
240965 100.218 -76.276 46.475
240966 100.259 -76.278 46.444
240967 100.299 -76.280 46.413
240968 100.339 -76.284 46.380
240969 100.376 -76.289 46.347
240970 100.395 -76.301 46.314
240971 100.412 -76.314 46.279
240972 100.428 -76.338 46.243
240973 100.444 -76.361 46.208
240974 100.460 -76.384 46.172
240975 100.476 -76.408 46.136
240976 100.492 -76.431 46.100
240977 100.508 -76.455 46.064
240978 100.524 -76.478 46.029
240979 100.541 -76.500 45.995
240980 100.562 -76.508 45.962
240981 100.584 -76.516 45.931
240982 100.606 -76.524 45.900
240983 100.615 -76.554 45.871
240984 100.625 -76.585 45.843
240985 100.644 -76.604 45.817
240986 100.664 -76.622 45.792
240987 100.682 -76.641 45.768
240988 100.725 -76.636 45.745
240989 100.770 -76.631 45.724
240990 100.815 -76.626 45.703
240991 100.860 -76.620 45.682
240992 100.905 -76.614 45.662
240993 100.949 -76.608 45.642
240994 100.994 -76.603 45.622
240995 101.039 -76.598 45.601
240996 101.073 -76.599 45.579
240997 101.098 -76.608 45.557
240998 101.121 -76.618 45.535
240999 101.145 -76.627 45.514
241000 101.169 -76.636 45.492
241001 101.193 -76.645 45.471
241002 101.220 -76.655 45.450
241003 101.248 -76.666 45.430
241004 101.276 -76.677 45.410
241005 101.299 -76.687 45.390
241006 101.314 -76.699 45.372
241007 101.329 -76.712 45.353
241008 101.343 -76.724 45.335
241009 101.358 -76.736 45.319
241010 101.373 -76.747 45.303
241011 101.388 -76.759 45.287
241012 101.402 -76.771 45.274
241013 101.417 -76.784 45.260
241014 101.431 -76.796 45.247
241015 101.447 -76.808 45.234
241016 101.461 -76.820 45.221
241017 101.476 -76.831 45.207
241018 101.490 -76.843 45.193
241019 101.505 -76.856 45.179
241020 101.520 -76.868 45.166
241021 101.535 -76.880 45.150
241022 101.549 -76.892 45.134
241023 101.570 -76.894 45.119
241024 101.592 -76.894 45.104
241025 101.613 -76.894 45.085
241026 101.635 -76.894 45.068
241027 101.656 -76.894 45.051
241028 101.678 -76.894 45.033
241029 101.700 -76.894 45.016
241030 101.718 -76.902 44.999
241031 101.734 -76.913 44.983
241032 101.750 -76.925 44.965
241033 101.767 -76.935 44.949
241034 101.783 -76.947 44.932
241035 101.799 -76.958 44.914
241036 101.816 -76.969 44.897
241037 101.832 -76.980 44.880
241038 101.848 -76.991 44.862
241039 101.865 -77.003 44.845
241040 101.881 -77.014 44.828
241041 101.898 -77.025 44.813
241042 101.914 -77.036 44.797
241043 101.930 -77.047 44.781
241044 101.947 -77.058 44.765
241045 101.963 -77.069 44.749
241046 101.982 -77.082 44.733
241047 102.001 -77.096 44.718
241048 102.024 -77.117 44.701
241049 102.045 -77.136 44.686
241050 102.067 -77.156 44.672
241051 102.089 -77.175 44.657
241052 102.111 -77.194 44.640
241053 102.132 -77.214 44.625
241054 102.154 -77.233 44.611
241055 102.176 -77.253 44.596
241056 102.198 -77.273 44.580
241057 102.220 -77.291 44.565
241058 102.241 -77.311 44.550
241059 102.263 -77.330 44.534
241060 102.284 -77.350 44.519
241061 102.305 -77.366 44.504
241062 102.323 -77.373 44.488
241063 102.343 -77.375 44.471
241064 102.363 -77.376 44.453
241065 102.382 -77.379 44.435
241066 102.402 -77.380 44.416
241067 102.422 -77.382 44.399
241068 102.442 -77.384 44.382
241069 102.462 -77.385 44.364
241070 102.481 -77.388 44.346
241071 102.501 -77.389 44.329
241072 102.521 -77.391 44.311
241073 102.541 -77.393 44.294
241074 102.571 -77.398 44.277
241075 102.608 -77.406 44.261
241076 102.647 -77.414 44.245
241077 102.684 -77.421 44.228
241078 102.723 -77.429 44.212
241079 102.760 -77.437 44.197
241080 102.799 -77.445 44.181
241081 102.836 -77.452 44.165
241082 102.875 -77.460 44.150
241083 102.912 -77.468 44.135
241084 102.951 -77.476 44.120
241085 102.988 -77.483 44.104
241359 92.424 -74.591 54.436
241360 92.448 -74.602 54.443
241361 92.473 -74.612 54.451
241362 92.497 -74.623 54.458
241363 92.522 -74.634 54.465
241364 92.540 -74.624 54.472
241365 92.562 -74.611 54.480
241366 92.588 -74.602 54.488
241367 92.615 -74.592 54.495
241368 92.643 -74.583 54.502
241369 92.673 -74.575 54.508
241370 92.703 -74.567 54.510
241371 92.733 -74.559 54.513
241372 92.764 -74.550 54.514
241373 92.793 -74.557 54.515
241374 92.825 -74.559 54.515
241375 92.856 -74.560 54.514
241376 92.887 -74.562 54.514
241377 92.920 -74.568 54.517
241378 92.953 -74.574 54.522
241379 92.987 -74.579 54.527
241380 93.020 -74.585 54.534
241381 93.053 -74.591 54.542
241382 93.086 -74.596 54.541
241383 93.120 -74.603 54.537
241384 93.153 -74.609 54.532
241385 93.187 -74.615 54.530
241386 93.220 -74.620 54.529
241387 93.251 -74.623 54.529
241388 93.275 -74.613 54.528
241389 93.298 -74.603 54.525
241390 93.321 -74.593 54.524
241391 93.345 -74.584 54.523
241392 93.368 -74.574 54.522
241393 93.391 -74.564 54.525
241394 93.415 -74.553 54.532
241395 93.438 -74.543 54.543
241396 93.461 -74.534 54.555
241397 93.485 -74.524 54.565
241398 93.508 -74.514 54.576
241399 93.531 -74.505 54.587
241400 93.552 -74.501 54.598
241401 93.574 -74.496 54.607
241402 93.595 -74.493 54.612
241403 93.617 -74.488 54.611
241404 93.638 -74.483 54.603
241405 93.659 -74.480 54.597
241406 93.681 -74.475 54.601
241407 93.709 -74.483 54.613
241408 93.741 -74.495 54.633
241409 93.773 -74.507 54.653
241410 93.804 -74.520 54.671
241411 93.836 -74.532 54.683
241412 93.868 -74.545 54.682
241413 93.900 -74.557 54.674
241414 93.931 -74.570 54.670
241415 93.954 -74.572 54.673
241416 93.976 -74.574 54.684
241417 93.998 -74.577 54.697
241418 94.018 -74.577 54.704
241419 94.039 -74.577 54.707
241420 94.058 -74.577 54.710
241421 94.075 -74.574 54.707
241422 94.092 -74.572 54.700
241423 94.109 -74.570 54.691
241424 94.126 -74.567 54.681
241425 94.143 -74.565 54.666
241426 94.159 -74.562 54.649
241427 94.178 -74.561 54.629
241428 94.200 -74.562 54.612
241429 94.221 -74.562 54.596
241430 94.243 -74.563 54.579
241431 94.265 -74.564 54.562
241432 94.287 -74.563 54.537
241433 94.307 -74.565 54.507
241434 94.326 -74.567 54.477
241435 94.346 -74.568 54.447
241436 94.365 -74.569 54.419
241437 94.385 -74.570 54.394
241438 94.405 -74.571 54.367
241439 94.424 -74.572 54.337
241440 94.444 -74.574 54.307
241441 94.466 -74.581 54.273
241442 94.487 -74.586 54.237
241443 94.508 -74.588 54.199
241444 94.529 -74.591 54.160
241445 94.550 -74.593 54.120
241446 94.569 -74.598 54.081
241447 94.587 -74.606 54.043
241448 94.604 -74.614 54.007
241449 94.624 -74.621 53.974
241450 94.651 -74.628 53.943
241451 94.678 -74.634 53.914
241452 94.704 -74.640 53.887
241453 94.732 -74.646 53.861
241454 94.758 -74.653 53.835
241455 94.785 -74.659 53.806
241456 94.812 -74.665 53.775
241457 94.837 -74.670 53.743
241458 94.860 -74.675 53.710
241459 94.884 -74.681 53.673
241460 94.908 -74.685 53.634
241461 94.931 -74.690 53.592
241462 94.952 -74.693 53.548
241463 94.951 -74.682 53.503
241464 94.950 -74.671 53.457
241465 94.949 -74.661 53.409
241466 94.948 -74.650 53.362
241467 94.953 -74.640 53.317
241468 94.976 -74.628 53.269
241469 94.999 -74.618 53.217
241470 95.028 -74.618 53.160
241471 95.061 -74.635 53.098
241472 95.096 -74.651 53.035
241473 95.127 -74.668 52.972
241474 95.155 -74.688 52.911
241475 95.188 -74.703 52.850
241476 95.222 -74.719 52.790
241477 95.255 -74.734 52.731
241478 95.289 -74.750 52.672
241479 95.322 -74.766 52.614
241480 95.354 -74.780 52.555
241481 95.382 -74.790 52.497
241482 95.409 -74.800 52.439
241483 95.437 -74.811 52.384
241484 95.463 -74.815 52.329
241485 95.488 -74.814 52.275
241486 95.513 -74.813 52.223
241487 95.538 -74.812 52.172
241488 95.562 -74.811 52.123
241489 95.588 -74.810 52.076
241490 95.612 -74.809 52.030
241491 95.637 -74.808 51.988
241492 95.662 -74.807 51.947
241493 95.693 -74.786 51.907
241494 95.726 -74.760 51.867
241495 95.758 -74.734 51.829
241496 95.791 -74.708 51.790
241497 95.824 -74.681 51.753
241498 95.856 -74.656 51.715
241499 95.889 -74.629 51.682
241500 95.923 -74.609 51.648
241501 95.959 -74.641 51.613
241502 95.995 -74.677 51.579
241503 96.031 -74.712 51.546
241504 96.067 -74.748 51.514
241505 96.103 -74.783 51.484
241506 96.139 -74.819 51.455
241507 96.173 -74.841 51.430
241508 96.203 -74.852 51.405
241509 96.234 -74.863 51.382
241510 96.265 -74.873 51.361
241511 96.296 -74.883 51.342
241512 96.326 -74.894 51.325
241513 96.357 -74.904 51.310
241514 96.388 -74.915 51.296
241515 96.413 -74.926 51.282
241516 96.438 -74.937 51.270
241517 96.462 -74.948 51.257
241518 96.485 -74.966 51.244
241519 96.508 -74.988 51.231
241520 96.531 -75.008 51.218
241521 96.554 -75.029 51.204
241522 96.578 -75.050 51.187
241523 96.600 -75.069 51.170
241524 96.621 -75.084 51.152
241525 96.642 -75.098 51.135
241526 96.662 -75.113 51.116
241527 96.682 -75.127 51.096
241528 96.709 -75.136 51.075
241529 96.740 -75.142 51.052
241530 96.770 -75.147 51.030
241531 96.801 -75.153 51.007
241532 96.827 -75.152 50.986
241533 96.851 -75.147 50.967
241534 96.874 -75.142 50.949
241535 96.897 -75.135 50.932
241536 96.921 -75.130 50.916
241537 96.944 -75.125 50.898
241538 96.967 -75.120 50.882
241539 96.958 -75.225 50.863
241540 96.982 -75.217 50.843
241541 97.007 -75.207 50.822
241542 97.031 -75.199 50.800
241543 97.056 -75.191 50.777
241544 97.081 -75.182 50.753
241545 97.103 -75.185 50.730
241546 97.124 -75.193 50.706
241547 97.146 -75.200 50.683
241548 97.167 -75.207 50.659
241549 97.189 -75.214 50.634
241550 97.210 -75.222 50.605
241551 97.232 -75.229 50.577
241552 97.255 -75.239 50.549
241553 97.280 -75.248 50.521
241554 97.304 -75.257 50.489
241555 97.321 -75.264 50.459
241556 97.337 -75.273 50.428
241557 97.355 -75.281 50.396
241558 97.372 -75.289 50.366
241559 97.388 -75.297 50.335
241560 97.405 -75.305 50.303
241561 97.422 -75.313 50.270
241562 97.453 -75.320 50.236
241563 97.484 -75.326 50.202
241564 97.514 -75.333 50.167
241565 97.544 -75.339 50.131
241566 97.570 -75.345 50.093
241567 97.595 -75.352 50.056
241568 97.621 -75.358 50.019
241569 97.647 -75.364 49.980
241570 97.672 -75.371 49.942
241571 97.698 -75.377 49.904
241572 97.722 -75.382 49.864
241573 97.744 -75.386 49.821
241574 97.787 -75.384 49.776
241575 97.826 -75.399 49.729
241576 97.866 -75.414 49.680
241577 97.905 -75.429 49.632
241578 97.944 -75.444 49.584
241579 97.984 -75.459 49.535
241580 98.017 -75.479 49.487
241581 98.053 -75.486 49.440
241582 98.077 -75.493 49.393
241583 98.102 -75.500 49.347
241584 98.128 -75.507 49.302
241585 98.152 -75.514 49.259
241586 98.177 -75.521 49.213
241587 98.195 -75.535 49.166
241588 98.207 -75.554 49.118
241589 98.219 -75.573 49.072
241590 98.248 -75.586 49.026
241591 98.298 -75.577 48.980
241592 98.359 -75.557 48.934
241593 98.419 -75.555 48.889
241594 98.478 -75.561 48.843
241595 98.537 -75.568 48.798
241596 98.596 -75.575 48.752
241597 98.593 -75.620 48.702
241598 98.581 -75.638 48.653
241599 98.565 -75.641 48.604
241600 98.560 -75.646 48.555
241601 98.600 -75.656 48.508
241602 98.641 -75.665 48.462
241603 98.703 -75.686 48.417
241604 98.739 -75.699 48.371
241605 98.751 -75.705 48.324
241606 98.764 -75.711 48.275
241607 98.799 -75.701 48.227
241608 98.838 -75.685 48.179
241609 98.861 -75.674 48.133
241610 98.883 -75.675 48.089
241611 98.901 -75.699 48.047
241612 98.918 -75.724 48.007
241613 98.935 -75.749 47.970
241614 98.952 -75.775 47.935
241615 98.970 -75.800 47.901
241616 98.987 -75.824 47.868
241617 99.004 -75.849 47.837
241618 99.021 -75.874 47.807
241619 99.038 -75.897 47.778
241620 99.056 -75.913 47.749
241621 99.073 -75.930 47.721
241622 99.090 -75.947 47.694
241623 99.108 -75.963 47.667
241624 99.127 -75.979 47.640
241625 99.156 -75.994 47.613
241626 99.185 -76.008 47.586
241627 99.213 -76.022 47.562
241628 99.242 -76.032 47.536
241629 99.272 -76.037 47.508
241630 99.302 -76.041 47.480
241631 99.332 -76.047 47.453
241632 99.362 -76.051 47.425
241633 99.392 -76.056 47.398
241634 99.422 -76.061 47.373
241635 99.452 -76.066 47.347
241636 99.482 -76.070 47.322
241637 99.512 -76.075 47.297
241638 99.542 -76.080 47.272
241639 99.572 -76.085 47.247
241640 99.593 -76.088 47.223
241641 99.609 -76.091 47.199
241642 99.625 -76.094 47.173
241643 99.640 -76.096 47.148
241644 99.666 -76.100 47.119
241645 99.696 -76.105 47.089
241646 99.726 -76.110 47.060
241647 99.755 -76.114 47.032
241648 99.785 -76.119 47.002
241649 99.815 -76.124 46.971
241650 99.844 -76.129 46.941
241651 99.864 -76.144 46.912
241652 99.881 -76.166 46.880
241653 99.901 -76.189 46.850
241654 99.920 -76.213 46.821
241655 99.939 -76.236 46.791
241656 99.959 -76.260 46.763
241657 99.982 -76.267 46.733
241658 100.007 -76.263 46.703
241659 100.031 -76.260 46.674
241660 100.056 -76.257 46.645
241661 100.081 -76.254 46.617
241662 100.106 -76.251 46.588
241663 100.130 -76.247 46.558
241664 100.159 -76.245 46.528
241665 100.200 -76.248 46.498
241666 100.240 -76.250 46.467
241667 100.280 -76.253 46.436
241668 100.321 -76.255 46.405
241669 100.361 -76.257 46.373
241670 100.395 -76.271 46.340
241671 100.412 -76.298 46.307
241672 100.428 -76.322 46.272
241673 100.444 -76.345 46.237
241674 100.460 -76.369 46.202
241675 100.476 -76.392 46.167
241676 100.492 -76.416 46.130
241677 100.508 -76.439 46.094
241678 100.524 -76.463 46.059
241679 100.542 -76.480 46.024
241680 100.564 -76.489 45.990
241681 100.585 -76.496 45.958
241682 100.607 -76.505 45.926
241683 100.625 -76.518 45.896
241684 100.633 -76.552 45.867
241685 100.642 -76.585 45.839
241686 100.662 -76.603 45.812
241687 100.680 -76.622 45.787
241688 100.706 -76.634 45.763
241689 100.751 -76.628 45.740
241690 100.796 -76.623 45.719
241691 100.841 -76.617 45.698
241692 100.885 -76.611 45.677
241693 100.930 -76.606 45.657
241694 100.975 -76.600 45.636
241695 101.019 -76.595 45.616
241696 101.064 -76.589 45.594
241697 101.104 -76.586 45.573
241698 101.128 -76.596 45.550
241699 101.152 -76.605 45.529
241700 101.176 -76.614 45.507
241701 101.200 -76.624 45.485
241702 101.224 -76.633 45.464
241703 101.251 -76.643 45.443
241704 101.279 -76.654 45.423
241705 101.307 -76.665 45.403
241706 101.327 -76.676 45.384
241707 101.342 -76.687 45.365
241708 101.357 -76.699 45.347
241709 101.372 -76.712 45.329
241710 101.386 -76.724 45.312
241711 101.401 -76.736 45.296
241712 101.416 -76.748 45.280
241713 101.430 -76.760 45.266
241714 101.445 -76.771 45.250
241715 101.460 -76.784 45.236
241716 101.475 -76.796 45.223
241717 101.489 -76.808 45.210
241718 101.504 -76.820 45.195
241719 101.519 -76.832 45.180
241720 101.534 -76.844 45.166
241721 101.548 -76.856 45.152
241722 101.563 -76.868 45.136
241723 101.584 -76.869 45.121
241724 101.606 -76.869 45.105
241725 101.627 -76.869 45.090
241726 101.649 -76.870 45.072
241727 101.671 -76.870 45.055
241728 101.692 -76.870 45.037
241729 101.713 -76.871 45.019
241730 101.730 -76.882 45.002
241731 101.746 -76.893 44.985
241732 101.762 -76.904 44.969
241733 101.778 -76.915 44.951
241734 101.795 -76.927 44.935
241735 101.811 -76.938 44.918
241736 101.828 -76.949 44.900
241737 101.844 -76.960 44.883
241738 101.861 -76.972 44.866
241739 101.877 -76.982 44.849
241740 101.893 -76.994 44.832
241741 101.909 -77.005 44.816
241742 101.926 -77.016 44.800
241743 101.942 -77.027 44.784
241744 101.959 -77.038 44.768
241745 101.975 -77.050 44.752
241746 101.991 -77.061 44.736
241747 102.007 -77.072 44.720
241748 102.021 -77.081 44.704
241749 102.043 -77.100 44.687
241750 102.064 -77.120 44.672
241751 102.087 -77.140 44.658
241752 102.108 -77.160 44.643
241753 102.130 -77.179 44.626
241754 102.152 -77.199 44.610
241755 102.174 -77.219 44.595
241756 102.196 -77.238 44.580
241757 102.218 -77.257 44.564
241758 102.239 -77.277 44.549
241759 102.261 -77.297 44.533
241760 102.283 -77.316 44.517
241761 102.304 -77.330 44.502
241762 102.326 -77.332 44.487
241763 102.347 -77.333 44.471
241764 102.368 -77.335 44.454
241765 102.387 -77.336 44.438
241766 102.407 -77.338 44.420
241767 102.427 -77.340 44.402
241768 102.447 -77.341 44.385
241769 102.466 -77.344 44.368
241770 102.486 -77.345 44.351
241771 102.506 -77.348 44.333
241772 102.526 -77.349 44.316
241773 102.545 -77.351 44.299
241774 102.565 -77.353 44.282
241775 102.585 -77.354 44.266
241776 102.612 -77.359 44.251
241777 102.650 -77.367 44.234
241778 102.688 -77.374 44.218
241779 102.727 -77.382 44.202
241780 102.764 -77.390 44.187
241781 102.803 -77.398 44.172
241782 102.840 -77.405 44.156
241783 102.879 -77.413 44.141
241784 102.916 -77.421 44.126
241785 102.954 -77.429 44.110
242061 92.450 -74.569 54.450
242062 92.475 -74.580 54.457
242063 92.499 -74.590 54.464
242064 92.523 -74.602 54.470
242065 92.544 -74.600 54.476
242066 92.564 -74.584 54.485
242067 92.590 -74.574 54.492
242068 92.617 -74.565 54.499
242069 92.644 -74.556 54.506
242070 92.674 -74.548 54.511
242071 92.704 -74.540 54.514
242072 92.734 -74.531 54.516
242073 92.765 -74.524 54.518
242074 92.795 -74.531 54.519
242075 92.826 -74.537 54.519
242076 92.856 -74.539 54.518
242077 92.887 -74.540 54.518
242078 92.919 -74.541 54.520
242079 92.950 -74.543 54.524
242080 92.984 -74.549 54.530
242081 93.017 -74.555 54.538
242082 93.050 -74.562 54.548
242083 93.084 -74.567 54.551
242084 93.117 -74.573 54.551
242085 93.150 -74.579 54.551
242086 93.184 -74.585 54.552
242087 93.217 -74.590 54.555
242088 93.250 -74.596 54.556
242089 93.277 -74.593 54.557
242090 93.300 -74.583 54.554
242091 93.324 -74.573 54.553
242092 93.347 -74.563 54.553
242093 93.370 -74.552 54.553
242094 93.394 -74.543 54.556
242095 93.417 -74.533 54.565
242096 93.440 -74.523 54.578
242097 93.464 -74.513 54.591
242098 93.487 -74.503 54.602
242099 93.510 -74.493 54.613
242100 93.532 -74.488 54.624
242101 93.554 -74.484 54.638
242102 93.575 -74.480 54.648
242103 93.596 -74.475 54.651
242104 93.618 -74.471 54.649
242105 93.639 -74.467 54.641
242106 93.660 -74.463 54.634
242107 93.682 -74.458 54.637
242108 93.704 -74.455 54.648
242109 93.733 -74.464 54.669
242110 93.765 -74.477 54.690
242111 93.797 -74.489 54.707
242112 93.829 -74.501 54.718
242113 93.860 -74.514 54.716
242114 93.892 -74.526 54.708
242115 93.924 -74.539 54.700
242116 93.947 -74.543 54.698
242117 93.970 -74.545 54.707
242118 93.992 -74.547 54.717
242119 94.013 -74.549 54.721
242120 94.034 -74.549 54.721
242121 94.054 -74.549 54.718
242122 94.074 -74.549 54.712
242123 94.091 -74.546 54.705
242124 94.108 -74.544 54.696
242125 94.125 -74.542 54.684
242126 94.142 -74.540 54.669
242127 94.158 -74.537 54.650
242128 94.177 -74.536 54.630
242129 94.199 -74.537 54.614
242130 94.220 -74.537 54.598
242131 94.243 -74.538 54.583
242132 94.264 -74.539 54.566
242133 94.285 -74.540 54.543
242134 94.306 -74.540 54.516
242135 94.325 -74.542 54.486
242136 94.345 -74.543 54.454
242137 94.364 -74.543 54.425
242138 94.384 -74.544 54.398
242139 94.404 -74.545 54.370
242140 94.423 -74.546 54.338
242141 94.444 -74.550 54.306
242142 94.465 -74.557 54.271
242143 94.487 -74.564 54.235
242144 94.508 -74.568 54.200
242145 94.529 -74.570 54.163
242146 94.550 -74.573 54.123
242147 94.567 -74.581 54.082
242148 94.585 -74.588 54.043
242149 94.602 -74.596 54.007
242150 94.620 -74.603 53.974
242151 94.645 -74.610 53.941
242152 94.672 -74.616 53.908
242153 94.699 -74.623 53.876
242154 94.725 -74.629 53.847
242155 94.752 -74.635 53.819
242156 94.778 -74.641 53.790
242157 94.802 -74.646 53.760
242158 94.826 -74.651 53.730
242159 94.849 -74.656 53.700
242160 94.873 -74.661 53.666
242161 94.897 -74.666 53.629
242162 94.921 -74.671 53.588
242163 94.931 -74.668 53.545
242164 94.930 -74.657 53.501
242165 94.930 -74.646 53.456
242166 94.929 -74.636 53.411
242167 94.936 -74.624 53.368
242168 94.960 -74.614 53.327
242169 94.983 -74.603 53.281
242170 95.006 -74.593 53.230
242171 95.038 -74.601 53.172
242172 95.072 -74.618 53.110
242173 95.106 -74.634 53.045
242174 95.140 -74.650 52.981
242175 95.167 -74.670 52.917
242176 95.196 -74.688 52.854
242177 95.230 -74.704 52.794
242178 95.263 -74.719 52.734
242179 95.297 -74.735 52.675
242180 95.330 -74.750 52.616
242181 95.362 -74.765 52.557
242182 95.390 -74.775 52.499
242183 95.418 -74.785 52.442
242184 95.446 -74.796 52.386
242185 95.470 -74.795 52.332
242186 95.495 -74.794 52.278
242187 95.520 -74.793 52.227
242188 95.545 -74.792 52.176
242189 95.570 -74.791 52.127
242190 95.595 -74.791 52.080
242191 95.619 -74.789 52.035
242192 95.644 -74.788 51.992
242193 95.669 -74.787 51.951
242194 95.696 -74.781 51.911
242195 95.728 -74.755 51.872
242196 95.761 -74.729 51.833
242197 95.794 -74.703 51.795
242198 95.826 -74.677 51.757
242199 95.859 -74.650 51.719
242200 95.892 -74.624 51.684
242201 95.929 -74.629 51.649
242202 95.968 -74.640 51.614
242203 96.006 -74.650 51.578
242204 96.043 -74.672 51.545
242205 96.080 -74.707 51.512
242206 96.116 -74.743 51.481
242207 96.152 -74.778 51.452
242208 96.188 -74.814 51.425
242209 96.219 -74.826 51.400
242210 96.250 -74.837 51.376
242211 96.281 -74.847 51.355
242212 96.311 -74.858 51.335
242213 96.342 -74.869 51.318
242214 96.370 -74.879 51.302
242215 96.394 -74.890 51.289
242216 96.418 -74.901 51.275
242217 96.442 -74.913 51.262
242218 96.466 -74.923 51.250
242219 96.489 -74.941 51.237
242220 96.513 -74.962 51.225
242221 96.536 -74.983 51.214
242222 96.559 -75.004 51.202
242223 96.582 -75.025 51.187
242224 96.605 -75.045 51.172
242225 96.628 -75.066 51.157
242226 96.652 -75.087 51.140
242227 96.674 -75.108 51.123
242228 96.695 -75.123 51.103
242229 96.718 -75.135 51.082
242230 96.748 -75.140 51.059
242231 96.774 -75.138 51.037
242232 96.798 -75.133 51.015
242233 96.821 -75.128 50.995
242234 96.844 -75.123 50.975
242235 96.868 -75.116 50.957
242236 96.891 -75.111 50.941
242237 96.914 -75.106 50.925
242238 96.938 -75.100 50.909
242239 96.961 -75.095 50.894
242240 96.959 -75.175 50.876
242241 96.983 -75.167 50.856
242242 97.008 -75.158 50.836
242243 97.032 -75.150 50.814
242244 97.057 -75.142 50.791
242245 97.079 -75.149 50.768
242246 97.100 -75.157 50.745
242247 97.121 -75.163 50.722
242248 97.143 -75.171 50.698
242249 97.164 -75.179 50.675
242250 97.186 -75.185 50.651
242251 97.207 -75.193 50.626
242252 97.229 -75.201 50.601
242253 97.252 -75.209 50.576
242254 97.277 -75.218 50.549
242255 97.301 -75.227 50.520
242256 97.322 -75.236 50.490
242257 97.339 -75.245 50.459
242258 97.357 -75.252 50.426
242259 97.373 -75.261 50.394
242260 97.390 -75.269 50.360
242261 97.407 -75.276 50.325
242262 97.426 -75.285 50.290
242263 97.456 -75.291 50.256
242264 97.487 -75.298 50.221
242265 97.517 -75.305 50.186
242266 97.548 -75.311 50.151
242267 97.575 -75.317 50.114
242268 97.600 -75.323 50.077
242269 97.626 -75.330 50.039
242270 97.652 -75.336 50.000
242271 97.677 -75.342 49.961
242272 97.703 -75.349 49.922
242273 97.729 -75.355 49.882
242274 97.752 -75.360 49.838
242275 97.793 -75.368 49.792
242276 97.832 -75.383 49.744
242277 97.871 -75.398 49.695
242278 97.911 -75.413 49.647
242279 97.950 -75.428 49.599
242280 97.989 -75.443 49.550
242281 98.023 -75.456 49.502
242282 98.049 -75.463 49.455
242283 98.074 -75.470 49.407
242284 98.098 -75.477 49.361
242285 98.124 -75.483 49.316
242286 98.149 -75.490 49.273
242287 98.174 -75.497 49.227
242288 98.196 -75.507 49.181
242289 98.208 -75.525 49.133
242290 98.236 -75.539 49.086
242291 98.272 -75.549 49.039
242292 98.324 -75.538 48.993
242293 98.385 -75.518 48.948
242294 98.446 -75.497 48.902
242295 98.506 -75.481 48.856
242296 98.566 -75.488 48.810
242297 98.589 -75.516 48.763
242298 98.583 -75.562 48.712
242299 98.575 -75.599 48.663
242300 98.559 -75.603 48.614
242301 98.563 -75.609 48.565
242302 98.603 -75.618 48.518
242303 98.644 -75.628 48.471
242304 98.685 -75.638 48.425
242305 98.745 -75.658 48.379
242306 98.761 -75.665 48.331
242307 98.773 -75.670 48.282
242308 98.816 -75.652 48.233
242309 98.841 -75.642 48.185
242310 98.864 -75.631 48.138
242311 98.885 -75.636 48.092
242312 98.903 -75.661 48.049
242313 98.920 -75.686 48.007
242314 98.937 -75.711 47.970
242315 98.954 -75.736 47.934
242316 98.972 -75.760 47.900
242317 98.989 -75.785 47.867
242318 99.006 -75.810 47.835
242319 99.023 -75.835 47.804
242320 99.040 -75.859 47.773
242321 99.058 -75.884 47.743
242322 99.075 -75.909 47.714
242323 99.092 -75.931 47.686
242324 99.110 -75.947 47.659
242325 99.127 -75.963 47.632
242326 99.151 -75.978 47.604
242327 99.180 -75.994 47.577
242328 99.209 -76.002 47.551
242329 99.239 -76.007 47.524
242330 99.269 -76.012 47.496
242331 99.299 -76.016 47.469
242332 99.329 -76.022 47.441
242333 99.359 -76.026 47.413
242334 99.389 -76.031 47.386
242335 99.419 -76.036 47.361
242336 99.449 -76.041 47.336
242337 99.479 -76.046 47.312
242338 99.509 -76.050 47.287
242339 99.539 -76.055 47.261
242340 99.569 -76.060 47.236
242341 99.598 -76.065 47.212
242342 99.614 -76.067 47.186
242343 99.629 -76.070 47.161
242344 99.653 -76.074 47.136
242345 99.683 -76.079 47.108
242346 99.713 -76.083 47.080
242347 99.742 -76.088 47.052
242348 99.772 -76.092 47.025
242349 99.801 -76.097 46.994
242350 99.831 -76.102 46.965
242351 99.860 -76.107 46.935
242352 99.883 -76.121 46.905
242353 99.903 -76.144 46.874
242354 99.922 -76.167 46.844
242355 99.941 -76.190 46.815
242356 99.960 -76.213 46.785
242357 99.979 -76.235 46.756
242358 100.003 -76.248 46.726
242359 100.034 -76.240 46.696
242360 100.058 -76.237 46.666
242361 100.083 -76.233 46.638
242362 100.108 -76.230 46.609
242363 100.132 -76.227 46.580
242364 100.157 -76.223 46.550
242365 100.182 -76.220 46.520
242366 100.221 -76.223 46.489
242367 100.262 -76.225 46.458
242368 100.302 -76.227 46.428
242369 100.343 -76.230 46.396
242370 100.383 -76.232 46.365
242371 100.409 -76.259 46.332
242372 100.427 -76.299 46.298
242373 100.444 -76.330 46.263
242374 100.460 -76.354 46.229
242375 100.476 -76.377 46.196
242376 100.492 -76.401 46.160
242377 100.508 -76.424 46.124
242378 100.524 -76.448 46.089
242379 100.544 -76.461 46.054
242380 100.565 -76.470 46.019
242381 100.587 -76.477 45.986
242382 100.609 -76.486 45.954
242383 100.630 -76.493 45.922
242384 100.644 -76.517 45.893
242385 100.652 -76.550 45.864
242386 100.660 -76.583 45.836
242387 100.679 -76.603 45.808
242388 100.697 -76.621 45.783
242389 100.732 -76.626 45.759
242390 100.777 -76.620 45.736
242391 100.821 -76.614 45.715
242392 100.866 -76.608 45.693
242393 100.911 -76.603 45.672
242394 100.956 -76.598 45.652
242395 101.000 -76.592 45.630
242396 101.045 -76.586 45.610
242397 101.090 -76.580 45.588
242398 101.135 -76.575 45.566
242399 101.159 -76.584 45.544
242400 101.183 -76.593 45.522
242401 101.207 -76.602 45.500
242402 101.231 -76.611 45.478
242403 101.255 -76.621 45.457
242404 101.282 -76.631 45.437
242405 101.310 -76.642 45.417
242406 101.338 -76.652 45.397
242407 101.356 -76.664 45.377
242408 101.371 -76.676 45.359
242409 101.385 -76.688 45.340
242410 101.400 -76.700 45.322
242411 101.415 -76.712 45.305
242412 101.429 -76.724 45.289
242413 101.444 -76.736 45.273
242414 101.459 -76.748 45.257
242415 101.474 -76.760 45.242
242416 101.488 -76.772 45.227
242417 101.503 -76.784 45.213
242418 101.518 -76.796 45.198
242419 101.533 -76.808 45.183
242420 101.547 -76.820 45.168
242421 101.562 -76.832 45.153
242422 101.576 -76.844 45.138
242423 101.598 -76.845 45.123
242424 101.620 -76.845 45.107
242425 101.641 -76.845 45.091
242426 101.663 -76.845 45.076
242427 101.684 -76.845 45.058
242428 101.706 -76.845 45.041
242429 101.725 -76.851 45.023
242430 101.741 -76.862 45.006
242431 101.758 -76.873 44.989
242432 101.774 -76.884 44.972
242433 101.790 -76.896 44.955
242434 101.807 -76.906 44.938
242435 101.823 -76.918 44.921
242436 101.839 -76.929 44.904
242437 101.856 -76.940 44.887
242438 101.872 -76.951 44.869
242439 101.889 -76.962 44.852
242440 101.905 -76.974 44.836
242441 101.921 -76.985 44.819
242442 101.938 -76.996 44.803
242443 101.954 -77.007 44.786
242444 101.970 -77.019 44.770
242445 101.987 -77.029 44.755
242446 102.003 -77.041 44.738
242447 102.006 -77.041 44.721
242448 102.021 -77.052 44.705
242449 102.043 -77.072 44.689
242450 102.064 -77.091 44.673
242451 102.087 -77.110 44.658
242452 102.108 -77.130 44.643
242453 102.130 -77.150 44.627
242454 102.152 -77.169 44.610
242455 102.174 -77.189 44.595
242456 102.196 -77.209 44.580
242457 102.218 -77.229 44.564
242458 102.239 -77.248 44.549
242459 102.261 -77.267 44.534
242460 102.283 -77.287 44.518
242461 102.301 -77.290 44.502
242462 102.324 -77.291 44.487
242463 102.347 -77.292 44.472
242464 102.370 -77.293 44.456
242465 102.392 -77.294 44.439
242466 102.412 -77.296 44.423
242467 102.432 -77.298 44.405
242468 102.451 -77.300 44.388
242469 102.471 -77.301 44.372
242470 102.491 -77.304 44.356
242471 102.511 -77.305 44.338
242472 102.530 -77.307 44.321
242473 102.550 -77.309 44.304
242474 102.570 -77.310 44.288
242475 102.590 -77.313 44.271
242476 102.610 -77.314 44.255
242477 102.629 -77.316 44.240
242478 102.655 -77.320 44.224
242479 102.692 -77.327 44.208
242480 102.731 -77.335 44.192
242481 102.768 -77.343 44.176
242482 102.806 -77.351 44.162
242483 102.844 -77.358 44.147
242484 102.882 -77.366 44.131
242763 92.477 -74.547 54.463
242764 92.501 -74.559 54.469
242765 92.525 -74.569 54.476
242766 92.549 -74.576 54.482
242767 92.565 -74.557 54.490
242768 92.592 -74.547 54.497
242769 92.619 -74.538 54.503
242770 92.645 -74.529 54.509
242771 92.675 -74.521 54.514
242772 92.706 -74.512 54.517
242773 92.736 -74.505 54.520
242774 92.766 -74.496 54.522
242775 92.796 -74.505 54.522
242776 92.826 -74.516 54.523
242777 92.857 -74.517 54.522
242778 92.888 -74.518 54.522
242779 92.919 -74.519 54.523
242780 92.950 -74.521 54.526
242781 92.982 -74.521 54.531
242782 93.014 -74.526 54.539
242783 93.048 -74.532 54.548
242784 93.081 -74.537 54.553
242785 93.114 -74.543 54.556
242786 93.147 -74.549 54.561
242787 93.181 -74.555 54.569
242788 93.214 -74.561 54.574
242789 93.247 -74.567 54.579
242790 93.280 -74.572 54.582
242791 93.303 -74.562 54.578
242792 93.327 -74.552 54.577
242793 93.350 -74.542 54.578
242794 93.373 -74.532 54.580
242795 93.397 -74.522 54.587
242796 93.420 -74.512 54.599
242797 93.443 -74.502 54.613
242798 93.467 -74.493 54.628
242799 93.490 -74.483 54.639
242800 93.512 -74.476 54.649
242801 93.534 -74.471 54.660
242802 93.555 -74.468 54.677
242803 93.576 -74.463 54.688
242804 93.598 -74.458 54.690
242805 93.619 -74.455 54.684
242806 93.641 -74.450 54.674
242807 93.662 -74.446 54.669
242808 93.684 -74.442 54.672
242809 93.705 -74.437 54.684
242810 93.726 -74.433 54.705
242811 93.758 -74.446 54.726
242812 93.789 -74.458 54.741
242813 93.821 -74.471 54.750
242814 93.853 -74.483 54.748
242815 93.885 -74.496 54.741
242816 93.917 -74.508 54.732
242817 93.941 -74.513 54.725
242818 93.964 -74.515 54.730
242819 93.985 -74.518 54.736
242820 94.007 -74.520 54.737
242821 94.030 -74.522 54.732
242822 94.050 -74.521 54.724
242823 94.070 -74.521 54.715
242824 94.090 -74.521 54.706
242825 94.107 -74.519 54.697
242826 94.124 -74.517 54.685
242827 94.141 -74.515 54.669
242828 94.157 -74.512 54.650
242829 94.177 -74.511 54.630
242830 94.198 -74.512 54.613
242831 94.220 -74.512 54.597
242832 94.242 -74.513 54.581
242833 94.263 -74.514 54.565
242834 94.284 -74.515 54.543
242835 94.305 -74.515 54.517
242836 94.324 -74.515 54.489
242837 94.344 -74.517 54.458
242838 94.363 -74.518 54.428
242839 94.383 -74.518 54.399
242840 94.403 -74.519 54.370
242841 94.422 -74.521 54.337
242842 94.443 -74.526 54.305
242843 94.465 -74.533 54.271
242844 94.486 -74.540 54.237
242845 94.508 -74.546 54.204
242846 94.529 -74.549 54.169
242847 94.548 -74.555 54.129
242848 94.565 -74.562 54.088
242849 94.583 -74.570 54.048
242850 94.600 -74.577 54.013
242851 94.618 -74.586 53.979
242852 94.639 -74.593 53.946
242853 94.666 -74.599 53.910
242854 94.693 -74.606 53.876
242855 94.719 -74.612 53.844
242856 94.744 -74.617 53.813
242857 94.767 -74.621 53.783
242858 94.791 -74.627 53.754
242859 94.815 -74.631 53.724
242860 94.839 -74.637 53.694
242861 94.862 -74.642 53.662
242862 94.886 -74.646 53.627
242863 94.910 -74.652 53.590
242864 94.911 -74.642 53.551
242865 94.910 -74.631 53.510
242866 94.910 -74.621 53.465
242867 94.920 -74.610 53.421
242868 94.944 -74.599 53.378
242869 94.967 -74.588 53.335
242870 94.990 -74.577 53.291
242871 95.014 -74.568 53.241
242872 95.048 -74.584 53.184
242873 95.082 -74.600 53.121
242874 95.116 -74.616 53.055
242875 95.150 -74.632 52.989
242876 95.183 -74.649 52.923
242877 95.207 -74.671 52.859
242878 95.238 -74.688 52.798
242879 95.271 -74.704 52.737
242880 95.305 -74.719 52.676
242881 95.338 -74.735 52.618
242882 95.371 -74.750 52.560
242883 95.399 -74.760 52.502
242884 95.426 -74.771 52.445
242885 95.453 -74.776 52.388
242886 95.477 -74.775 52.334
242887 95.502 -74.775 52.281
242888 95.527 -74.773 52.230
242889 95.552 -74.772 52.180
242890 95.577 -74.772 52.131
242891 95.602 -74.771 52.085
242892 95.627 -74.769 52.041
242893 95.652 -74.769 51.996
242894 95.676 -74.768 51.955
242895 95.701 -74.767 51.916
242896 95.731 -74.750 51.877
242897 95.763 -74.725 51.838
242898 95.796 -74.698 51.799
242899 95.829 -74.672 51.762
242900 95.862 -74.646 51.723
242901 95.897 -74.639 51.686
242902 95.936 -74.650 51.650
242903 95.974 -74.660 51.614
242904 96.013 -74.671 51.578
242905 96.051 -74.681 51.544
242906 96.090 -74.692 51.511
242907 96.128 -74.703 51.479
242908 96.164 -74.738 51.450
242909 96.200 -74.774 51.422
242910 96.234 -74.800 51.395
242911 96.265 -74.811 51.370
242912 96.296 -74.822 51.349
242913 96.326 -74.832 51.329
242914 96.350 -74.843 51.310
242915 96.374 -74.854 51.295
242916 96.398 -74.866 51.281
242917 96.422 -74.876 51.267
242918 96.446 -74.888 51.254
242919 96.470 -74.899 51.242
242920 96.494 -74.916 51.229
242921 96.517 -74.937 51.219
242922 96.540 -74.957 51.210
242923 96.564 -74.979 51.200
242924 96.587 -74.999 51.187
242925 96.609 -75.020 51.175
242926 96.632 -75.041 51.161
242927 96.656 -75.062 51.146
242928 96.679 -75.082 51.129
242929 96.704 -75.099 51.109
242930 96.730 -75.112 51.088
242931 96.748 -75.118 51.066
242932 96.768 -75.114 51.044
242933 96.791 -75.109 51.023
242934 96.814 -75.103 51.003
242935 96.838 -75.098 50.983
242936 96.861 -75.092 50.965
242937 96.884 -75.087 50.949
242938 96.908 -75.081 50.934
242939 96.931 -75.076 50.919
242940 96.954 -75.070 50.905
242941 96.960 -75.126 50.887
242942 96.984 -75.117 50.868
242943 97.009 -75.109 50.849
242944 97.032 -75.106 50.827
242945 97.054 -75.113 50.805
242946 97.075 -75.120 50.782
242947 97.097 -75.128 50.760
242948 97.118 -75.135 50.737
242949 97.140 -75.142 50.714
242950 97.161 -75.150 50.692
242951 97.183 -75.157 50.669
242952 97.204 -75.164 50.648
242953 97.226 -75.172 50.627
242954 97.249 -75.180 50.604
242955 97.273 -75.189 50.579
242956 97.298 -75.198 50.551
242957 97.322 -75.207 50.521
242958 97.341 -75.216 50.490
242959 97.358 -75.224 50.457
242960 97.375 -75.232 50.423
242961 97.392 -75.240 50.387
242962 97.409 -75.248 50.349
242963 97.429 -75.256 50.311
242964 97.459 -75.263 50.276
242965 97.490 -75.269 50.241
242966 97.521 -75.276 50.205
242967 97.551 -75.282 50.170
242968 97.580 -75.289 50.134
242969 97.605 -75.295 50.097
242970 97.631 -75.301 50.058
242971 97.657 -75.308 50.020
242972 97.682 -75.314 49.980
242973 97.708 -75.320 49.940
242974 97.734 -75.326 49.899
242975 97.763 -75.333 49.854
242976 97.815 -75.345 49.807
242977 97.871 -75.358 49.758
242978 97.916 -75.370 49.709
242979 97.960 -75.384 49.661
242980 98.006 -75.398 49.614
242981 98.051 -75.411 49.566
242982 98.073 -75.420 49.517
242983 98.089 -75.430 49.469
242984 98.106 -75.438 49.421
242985 98.122 -75.447 49.375
242986 98.139 -75.455 49.330
242987 98.155 -75.464 49.286
242988 98.172 -75.473 49.241
242989 98.195 -75.480 49.194
242990 98.224 -75.491 49.147
242991 98.261 -75.501 49.099
242992 98.296 -75.511 49.052
242993 98.351 -75.498 49.005
242994 98.412 -75.477 48.961
242995 98.474 -75.457 48.914
242996 98.535 -75.436 48.868
242997 98.586 -75.420 48.822
242998 98.579 -75.458 48.771
242999 98.573 -75.505 48.722
243000 98.566 -75.551 48.672
243001 98.553 -75.565 48.623
243002 98.566 -75.571 48.575
243003 98.606 -75.581 48.527
243004 98.647 -75.591 48.480
243005 98.688 -75.601 48.432
243006 98.728 -75.611 48.385
243007 98.769 -75.624 48.338
243008 98.796 -75.619 48.289
243009 98.819 -75.609 48.239
243010 98.844 -75.598 48.190
243011 98.867 -75.587 48.143
243012 98.887 -75.597 48.096
243013 98.905 -75.622 48.052
243014 98.922 -75.647 48.008
243015 98.939 -75.671 47.970
243016 98.956 -75.696 47.934
243017 98.974 -75.721 47.899
243018 98.991 -75.746 47.866
243019 99.008 -75.771 47.832
243020 99.025 -75.796 47.800
243021 99.042 -75.821 47.769
243022 99.060 -75.846 47.738
243023 99.077 -75.871 47.708
243024 99.094 -75.896 47.679
243025 99.111 -75.920 47.651
243026 99.128 -75.945 47.622
243027 99.147 -75.964 47.594
243028 99.176 -75.972 47.566
243029 99.206 -75.977 47.539
243030 99.236 -75.982 47.511
243031 99.266 -75.987 47.484
243032 99.296 -75.992 47.456
243033 99.326 -75.997 47.429
243034 99.356 -76.001 47.400
243035 99.385 -76.007 47.374
243036 99.416 -76.011 47.349
243037 99.445 -76.016 47.324
243038 99.476 -76.021 47.301
243039 99.505 -76.025 47.277
243040 99.536 -76.031 47.252
243041 99.565 -76.035 47.226
243042 99.595 -76.040 47.204
243043 99.618 -76.044 47.180
243044 99.640 -76.047 47.155
243045 99.670 -76.052 47.130
243046 99.700 -76.057 47.102
243047 99.729 -76.061 47.074
243048 99.759 -76.066 47.046
243049 99.788 -76.071 47.019
243050 99.818 -76.076 46.989
243051 99.848 -76.082 46.959
243052 99.877 -76.087 46.929
243053 99.903 -76.097 46.899
243054 99.923 -76.120 46.868
243055 99.942 -76.143 46.838
243056 99.961 -76.166 46.809
243057 99.980 -76.188 46.779
243058 99.999 -76.211 46.749
243059 100.019 -76.232 46.719
243060 100.053 -76.222 46.689
243061 100.085 -76.213 46.659
243062 100.110 -76.210 46.630
243063 100.134 -76.207 46.601
243064 100.159 -76.204 46.572
243065 100.184 -76.201 46.542
243066 100.208 -76.197 46.511
243067 100.243 -76.198 46.480
243068 100.284 -76.200 46.449
243069 100.324 -76.202 46.419
243070 100.364 -76.204 46.387
243071 100.405 -76.207 46.356
243072 100.423 -76.247 46.323
243073 100.441 -76.288 46.289
243074 100.459 -76.328 46.255
243075 100.476 -76.362 46.221
243076 100.492 -76.386 46.188
243077 100.508 -76.408 46.153
243078 100.524 -76.432 46.119
243079 100.545 -76.442 46.084
243080 100.567 -76.450 46.050
243081 100.588 -76.458 46.015
243082 100.610 -76.467 45.982
243083 100.631 -76.474 45.951
243084 100.653 -76.483 45.920
243085 100.662 -76.514 45.889
243086 100.670 -76.548 45.860
243087 100.678 -76.582 45.832
243088 100.695 -76.602 45.805
243089 100.714 -76.621 45.781
243090 100.758 -76.618 45.757
243091 100.802 -76.611 45.733
243092 100.847 -76.606 45.711
243093 100.892 -76.600 45.688
243094 100.936 -76.595 45.667
243095 100.981 -76.589 45.646
243096 101.026 -76.583 45.624
243097 101.071 -76.578 45.603
243098 101.115 -76.572 45.582
243099 101.160 -76.567 45.560
243100 101.190 -76.572 45.538
243101 101.214 -76.581 45.516
243102 101.238 -76.590 45.494
243103 101.262 -76.599 45.473
243104 101.285 -76.608 45.452
243105 101.313 -76.619 45.431
243106 101.341 -76.630 45.411
243107 101.369 -76.640 45.391
243108 101.384 -76.652 45.371
243109 101.399 -76.664 45.352
243110 101.413 -76.676 45.334
243111 101.428 -76.688 45.316
243112 101.443 -76.700 45.298
243113 101.458 -76.712 45.282
243114 101.472 -76.724 45.266
243115 101.487 -76.736 45.249
243116 101.502 -76.748 45.233
243117 101.516 -76.760 45.218
243118 101.531 -76.772 45.203
243119 101.546 -76.784 45.188
243120 101.561 -76.796 45.173
243121 101.575 -76.809 45.157
243122 101.591 -76.820 45.141
243123 101.612 -76.820 45.125
243124 101.634 -76.820 45.110
243125 101.655 -76.820 45.093
243126 101.677 -76.821 45.077
243127 101.698 -76.821 45.061
243128 101.720 -76.821 45.044
243129 101.737 -76.831 45.028
243130 101.753 -76.842 45.011
243131 101.769 -76.853 44.993
243132 101.786 -76.865 44.977
243133 101.802 -76.875 44.959
243134 101.819 -76.887 44.941
243135 101.835 -76.898 44.925
243136 101.851 -76.909 44.908
243137 101.868 -76.920 44.891
243138 101.884 -76.931 44.874
243139 101.900 -76.943 44.856
243140 101.917 -76.953 44.839
243141 101.933 -76.965 44.823
243142 101.950 -76.976 44.807
243143 101.966 -76.987 44.790
243144 101.982 -76.998 44.774
243145 101.990 -77.003 44.758
243146 101.994 -77.003 44.742
243147 102.002 -77.006 44.725
243148 102.021 -77.022 44.708
243149 102.043 -77.042 44.692
243150 102.064 -77.062 44.676
243151 102.087 -77.081 44.660
243152 102.108 -77.100 44.645
243153 102.130 -77.120 44.629
243154 102.152 -77.140 44.613
243155 102.174 -77.160 44.596
243156 102.196 -77.179 44.580
243157 102.218 -77.199 44.565
243158 102.239 -77.219 44.550
243159 102.261 -77.238 44.535
243160 102.281 -77.250 44.519
243161 102.299 -77.250 44.504
243162 102.321 -77.251 44.488
243163 102.344 -77.252 44.473
243164 102.367 -77.253 44.458
243165 102.390 -77.254 44.442
243166 102.413 -77.255 44.426
243167 102.436 -77.256 44.409
243168 102.456 -77.257 44.392
243169 102.476 -77.260 44.376
243170 102.495 -77.261 44.359
243171 102.515 -77.263 44.343
243172 102.535 -77.265 44.326
243173 102.555 -77.266 44.310
243174 102.575 -77.269 44.293
243175 102.595 -77.270 44.276
243176 102.614 -77.272 44.260
243177 102.634 -77.274 44.244
243178 102.654 -77.276 44.229
243179 102.674 -77.278 44.213
243180 102.696 -77.280 44.197
243181 102.734 -77.288 44.182
243182 102.772 -77.296 44.166
243183 102.810 -77.304 44.152
243465 92.503 -74.526 54.477
243466 92.527 -74.537 54.482
243467 92.552 -74.547 54.488
243468 92.567 -74.530 54.495
243469 92.593 -74.521 54.501
243470 92.620 -74.512 54.508
243471 92.647 -74.502 54.513
243472 92.676 -74.493 54.517
243473 92.707 -74.486 54.521
243474 92.737 -74.477 54.524
243475 92.767 -74.469 54.526
243476 92.797 -74.479 54.526
243477 92.827 -74.495 54.527
243478 92.858 -74.496 54.527
243479 92.889 -74.497 54.527
243480 92.920 -74.498 54.527
243481 92.951 -74.499 54.528
243482 92.983 -74.500 54.531
243483 93.013 -74.501 54.537
243484 93.045 -74.502 54.544
243485 93.078 -74.508 54.549
243486 93.111 -74.514 54.555
243487 93.144 -74.520 54.564
243488 93.178 -74.525 54.578
243489 93.211 -74.531 54.588
243490 93.244 -74.537 54.595
243491 93.277 -74.543 54.600
243492 93.306 -74.541 54.597
243493 93.329 -74.531 54.597
243494 93.353 -74.521 54.600
243495 93.376 -74.512 54.605
243496 93.400 -74.502 54.617
243497 93.423 -74.492 54.633
243498 93.446 -74.482 54.650
243499 93.470 -74.471 54.665
243500 93.492 -74.463 54.675
243501 93.513 -74.459 54.684
243502 93.535 -74.455 54.694
243503 93.556 -74.451 54.712
243504 93.578 -74.446 54.719
243505 93.599 -74.442 54.719
243506 93.621 -74.438 54.713
243507 93.642 -74.433 54.705
243508 93.663 -74.430 54.702
243509 93.685 -74.425 54.707
243510 93.706 -74.421 54.718
243511 93.727 -74.417 54.735
243512 93.751 -74.415 54.754
243513 93.782 -74.427 54.771
243514 93.814 -74.440 54.783
243515 93.846 -74.452 54.780
243516 93.877 -74.465 54.771
243517 93.909 -74.477 54.761
243518 93.935 -74.483 54.754
243519 93.957 -74.486 54.754
243520 93.979 -74.488 54.754
243521 94.001 -74.490 54.749
243522 94.024 -74.493 54.739
243523 94.045 -74.494 54.726
243524 94.065 -74.494 54.714
243525 94.085 -74.493 54.703
243526 94.105 -74.493 54.693
243527 94.123 -74.491 54.683
243528 94.139 -74.489 54.667
243529 94.156 -74.487 54.649
243530 94.176 -74.486 54.629
243531 94.198 -74.487 54.609
243532 94.219 -74.487 54.592
243533 94.241 -74.488 54.576
243534 94.261 -74.488 54.559
243535 94.282 -74.489 54.538
243536 94.303 -74.489 54.514
243537 94.323 -74.490 54.487
243538 94.343 -74.491 54.458
243539 94.362 -74.492 54.427
243540 94.382 -74.493 54.398
243541 94.402 -74.493 54.368
243542 94.421 -74.496 54.335
243543 94.443 -74.502 54.303
243544 94.465 -74.508 54.271
243545 94.486 -74.515 54.238
243546 94.507 -74.522 54.208
243547 94.529 -74.528 54.173
243548 94.546 -74.537 54.135
243549 94.563 -74.544 54.094
243550 94.581 -74.552 54.054
243551 94.598 -74.560 54.018
243552 94.616 -74.568 53.983
243553 94.633 -74.575 53.949
243554 94.660 -74.582 53.913
243555 94.686 -74.587 53.876
243556 94.709 -74.593 53.842
243557 94.733 -74.597 53.810
243558 94.757 -74.603 53.778
243559 94.780 -74.607 53.748
243560 94.804 -74.612 53.718
243561 94.828 -74.617 53.688
243562 94.851 -74.622 53.656
243563 94.876 -74.628 53.623
243564 94.892 -74.628 53.588
243565 94.891 -74.617 53.552
243566 94.890 -74.606 53.513
243567 94.904 -74.596 53.472
243568 94.927 -74.584 53.429
243569 94.951 -74.574 53.387
243570 94.974 -74.563 53.344
243571 94.997 -74.552 53.299
243572 95.024 -74.550 53.249
243573 95.058 -74.567 53.192
243574 95.092 -74.583 53.128
243575 95.126 -74.599 53.062
243576 95.160 -74.615 52.995
243577 95.194 -74.631 52.928
243578 95.224 -74.650 52.863
243579 95.248 -74.671 52.801
243580 95.279 -74.689 52.739
243581 95.313 -74.704 52.678
243582 95.346 -74.720 52.620
243583 95.379 -74.735 52.562
243584 95.407 -74.745 52.505
243585 95.435 -74.756 52.448
243586 95.460 -74.756 52.392
243587 95.485 -74.756 52.338
243588 95.510 -74.754 52.285
243589 95.534 -74.753 52.234
243590 95.559 -74.753 52.185
243591 95.584 -74.752 52.136
243592 95.609 -74.750 52.090
243593 95.634 -74.750 52.046
243594 95.659 -74.749 52.003
243595 95.684 -74.748 51.962
243596 95.708 -74.747 51.922
243597 95.733 -74.746 51.882
243598 95.766 -74.720 51.842
243599 95.799 -74.693 51.804
243600 95.831 -74.668 51.765
243601 95.865 -74.648 51.725
243602 95.904 -74.659 51.688
243603 95.942 -74.669 51.651
243604 95.980 -74.680 51.614
243605 96.019 -74.690 51.578
243606 96.057 -74.701 51.543
243607 96.096 -74.712 51.510
243608 96.134 -74.722 51.477
243609 96.173 -74.733 51.447
243610 96.212 -74.744 51.418
243611 96.249 -74.769 51.391
243612 96.281 -74.785 51.365
243613 96.306 -74.797 51.343
243614 96.330 -74.807 51.322
243615 96.355 -74.819 51.303
243616 96.379 -74.829 51.287
243617 96.402 -74.841 51.272
243618 96.427 -74.852 51.259
243619 96.451 -74.863 51.245
243620 96.475 -74.874 51.233
243621 96.499 -74.891 51.222
243622 96.522 -74.911 51.213
243623 96.544 -74.932 51.204
243624 96.568 -74.953 51.196
243625 96.591 -74.974 51.187
243626 96.617 -74.988 51.176
243627 96.643 -75.001 51.164
243628 96.670 -75.013 51.150
243629 96.696 -75.026 51.133
243630 96.722 -75.039 51.114
243631 96.749 -75.053 51.093
243632 96.776 -75.081 51.071
243633 96.789 -75.079 51.051
243634 96.803 -75.077 51.031
243635 96.817 -75.075 51.010
243636 96.831 -75.073 50.990
243637 96.855 -75.067 50.972
243638 96.878 -75.062 50.956
243639 96.901 -75.057 50.941
243640 96.925 -75.051 50.928
243641 96.948 -75.045 50.914
243642 96.961 -75.076 50.898
243643 96.985 -75.067 50.880
243644 97.008 -75.069 50.861
243645 97.029 -75.077 50.841
243646 97.051 -75.084 50.819
243647 97.072 -75.091 50.797
243648 97.094 -75.099 50.776
243649 97.115 -75.106 50.752
243650 97.137 -75.113 50.731
243651 97.158 -75.120 50.710
243652 97.180 -75.128 50.689
243653 97.201 -75.135 50.672
243654 97.223 -75.142 50.653
243655 97.245 -75.151 50.633
243656 97.270 -75.160 50.610
243657 97.294 -75.170 50.583
243658 97.318 -75.179 50.553
243659 97.343 -75.188 50.522
243660 97.360 -75.196 50.488
243661 97.377 -75.204 50.453
243662 97.394 -75.212 50.414
243663 97.410 -75.220 50.373
243664 97.432 -75.228 50.333
243665 97.463 -75.234 50.295
243666 97.494 -75.241 50.259
243667 97.524 -75.247 50.224
243668 97.555 -75.254 50.189
243669 97.585 -75.260 50.153
243670 97.610 -75.267 50.116
243671 97.636 -75.273 50.077
243672 97.662 -75.279 50.039
243673 97.687 -75.286 50.000
243674 97.713 -75.292 49.959
243675 97.747 -75.299 49.916
243676 97.781 -75.306 49.869
243677 97.824 -75.316 49.821
243678 97.879 -75.328 49.773
243679 97.935 -75.340 49.724
243680 97.990 -75.352 49.677
243681 98.046 -75.364 49.630
243682 98.095 -75.377 49.582
243683 98.115 -75.386 49.532
243684 98.129 -75.395 49.484
243685 98.143 -75.405 49.436
243686 98.156 -75.414 49.388
243687 98.169 -75.424 49.342
243688 98.183 -75.433 49.298
243689 98.197 -75.442 49.252
243690 98.210 -75.452 49.206
243691 98.249 -75.454 49.159
243692 98.298 -75.453 49.111
243693 98.347 -75.452 49.063
243694 98.395 -75.452 49.017
243695 98.443 -75.440 48.972
243696 98.489 -75.425 48.925
243697 98.533 -75.410 48.879
243698 98.578 -75.395 48.831
243699 98.575 -75.419 48.780
243700 98.563 -75.451 48.731
243701 98.556 -75.493 48.681
243702 98.547 -75.527 48.631
243703 98.568 -75.534 48.583
243704 98.609 -75.544 48.535
243705 98.650 -75.554 48.487
243706 98.690 -75.564 48.439
243707 98.731 -75.574 48.390
243708 98.771 -75.581 48.343
243709 98.799 -75.576 48.294
243710 98.822 -75.565 48.243
243711 98.847 -75.555 48.194
243712 98.870 -75.544 48.147
243713 98.889 -75.558 48.100
243714 98.907 -75.583 48.054
243715 98.924 -75.608 48.010
243716 98.941 -75.633 47.970
243717 98.958 -75.658 47.931
243718 98.976 -75.683 47.895
243719 98.993 -75.708 47.861
243720 99.010 -75.732 47.828
243721 99.027 -75.757 47.795
243722 99.045 -75.782 47.763
243723 99.062 -75.807 47.732
243724 99.079 -75.832 47.701
243725 99.096 -75.856 47.671
243726 99.113 -75.881 47.641
243727 99.133 -75.911 47.611
243728 99.163 -75.921 47.582
243729 99.195 -75.921 47.554
243730 99.224 -75.927 47.527
243731 99.253 -75.932 47.498
243732 99.282 -75.938 47.471
243733 99.311 -75.944 47.443
243734 99.340 -75.950 47.416
243735 99.369 -75.956 47.388
243736 99.398 -75.963 47.362
243737 99.427 -75.968 47.336
243738 99.456 -75.974 47.312
243739 99.485 -75.980 47.289
243740 99.514 -75.986 47.265
243741 99.543 -75.992 47.241
243742 99.572 -75.998 47.216
243743 99.602 -76.003 47.195
243744 99.626 -76.020 47.173
243745 99.657 -76.027 47.149
243746 99.687 -76.032 47.124
243747 99.717 -76.038 47.097
243748 99.747 -76.043 47.069
243749 99.776 -76.048 47.041
243750 99.806 -76.053 47.013
243751 99.836 -76.059 46.984
243752 99.865 -76.064 46.954
243753 99.895 -76.069 46.923
243754 99.919 -76.082 46.893
243755 99.943 -76.096 46.861
243756 99.962 -76.119 46.832
243757 99.981 -76.141 46.803
243758 100.000 -76.164 46.773
243759 100.019 -76.187 46.743
243760 100.039 -76.210 46.713
243761 100.069 -76.207 46.682
243762 100.104 -76.196 46.651
243763 100.137 -76.187 46.622
243764 100.161 -76.183 46.593
243765 100.186 -76.180 46.563
243766 100.210 -76.177 46.533
243767 100.235 -76.173 46.501
243768 100.265 -76.173 46.470
243769 100.305 -76.175 46.440
243770 100.346 -76.177 46.409
243771 100.387 -76.179 46.378
243772 100.420 -76.194 46.346
243773 100.437 -76.235 46.314
243774 100.455 -76.276 46.280
243775 100.473 -76.316 46.247
243776 100.491 -76.357 46.213
243777 100.508 -76.393 46.180
243778 100.525 -76.415 46.146
243779 100.547 -76.423 46.111
243780 100.568 -76.431 46.077
243781 100.590 -76.439 46.043
243782 100.611 -76.447 46.010
243783 100.633 -76.455 45.977
243784 100.654 -76.463 45.946
243785 100.672 -76.479 45.915
243786 100.680 -76.513 45.885
243787 100.688 -76.546 45.856
243788 100.696 -76.580 45.828
243789 100.712 -76.602 45.802
243790 100.738 -76.614 45.777
243791 100.783 -76.608 45.753
243792 100.828 -76.603 45.729
243793 100.873 -76.598 45.706
243794 100.917 -76.592 45.684
243795 100.962 -76.586 45.662
243796 101.007 -76.580 45.640
243797 101.051 -76.575 45.619
243798 101.096 -76.570 45.598
243799 101.141 -76.564 45.576
243800 101.186 -76.558 45.554
243801 101.220 -76.560 45.532
243802 101.245 -76.569 45.509
243803 101.268 -76.578 45.488
243804 101.292 -76.587 45.467
243805 101.316 -76.596 45.446
243806 101.344 -76.607 45.425
243807 101.372 -76.618 45.404
243808 101.398 -76.628 45.385
243809 101.412 -76.640 45.365
243810 101.427 -76.652 45.346
243811 101.442 -76.665 45.327
243812 101.457 -76.676 45.310
243813 101.471 -76.688 45.291
243814 101.486 -76.700 45.275
243815 101.501 -76.712 45.258
243816 101.516 -76.724 45.241
243817 101.530 -76.737 45.225
243818 101.545 -76.748 45.209
243819 101.560 -76.760 45.194
243820 101.574 -76.772 45.178
243821 101.589 -76.784 45.162
243822 101.604 -76.796 45.146
243823 101.626 -76.796 45.131
243824 101.647 -76.796 45.115
243825 101.669 -76.796 45.099
243826 101.691 -76.796 45.082
243827 101.712 -76.796 45.066
243828 101.733 -76.799 45.049
243829 101.749 -76.811 45.032
243830 101.765 -76.822 45.015
243831 101.781 -76.834 44.999
243832 101.798 -76.844 44.981
243833 101.814 -76.856 44.964
243834 101.830 -76.867 44.947
243835 101.847 -76.878 44.930
243836 101.863 -76.889 44.912
243837 101.880 -76.900 44.896
243838 101.896 -76.912 44.879
243839 101.912 -76.922 44.862
243840 101.928 -76.934 44.844
243841 101.945 -76.945 44.827
243842 101.961 -76.956 44.811
243843 101.978 -76.967 44.794
243844 101.976 -76.965 44.778
243845 101.983 -76.967 44.762
243846 101.990 -76.969 44.745
243847 101.999 -76.973 44.729
243848 102.021 -76.993 44.713
243849 102.043 -77.012 44.696
243850 102.064 -77.032 44.680
243851 102.087 -77.052 44.664
243852 102.108 -77.072 44.648
243853 102.130 -77.091 44.632
243854 102.152 -77.110 44.616
243855 102.174 -77.130 44.600
243856 102.196 -77.150 44.583
243857 102.218 -77.169 44.568
243858 102.239 -77.189 44.552
243859 102.261 -77.209 44.537
243860 102.278 -77.209 44.522
243861 102.296 -77.210 44.506
243862 102.319 -77.210 44.491
243863 102.342 -77.212 44.475
243864 102.365 -77.213 44.460
243865 102.388 -77.213 44.445
243866 102.411 -77.215 44.429
243867 102.434 -77.216 44.413
243868 102.457 -77.216 44.396
243869 102.480 -77.217 44.380
243870 102.501 -77.219 44.363
243871 102.520 -77.221 44.347
243872 102.540 -77.222 44.331
243873 102.559 -77.225 44.315
243874 102.580 -77.226 44.299
243875 102.599 -77.229 44.282
243876 102.619 -77.230 44.265
243877 102.639 -77.232 44.249
243878 102.659 -77.234 44.233
243879 102.678 -77.235 44.218
243880 102.698 -77.237 44.202
243881 102.718 -77.239 44.187
243882 102.738 -77.241 44.172
244167 92.529 -74.504 54.489
244168 92.554 -74.515 54.494
244169 92.571 -74.505 54.501
244170 92.595 -74.493 54.507
244171 92.622 -74.484 54.513
244172 92.648 -74.475 54.518
244173 92.678 -74.467 54.522
244174 92.708 -74.458 54.525
244175 92.738 -74.450 54.528
244176 92.769 -74.442 54.530
244177 92.798 -74.452 54.531
244178 92.828 -74.471 54.532
244179 92.859 -74.474 54.532
244180 92.890 -74.475 54.532
244181 92.921 -74.477 54.531
244182 92.952 -74.477 54.531
244183 92.983 -74.479 54.532
244184 93.014 -74.480 54.535
244185 93.046 -74.480 54.539
244186 93.076 -74.482 54.544
244187 93.108 -74.484 54.550
244188 93.141 -74.490 54.561
244189 93.174 -74.496 54.576
244190 93.208 -74.502 54.590
244191 93.241 -74.508 54.601
244192 93.275 -74.513 54.608
244193 93.308 -74.519 54.610
244194 93.332 -74.511 54.613
244195 93.355 -74.501 54.619
244196 93.379 -74.491 54.629
244197 93.402 -74.480 54.646
244198 93.425 -74.471 54.666
244199 93.449 -74.461 54.685
244200 93.472 -74.451 54.702
244201 93.494 -74.446 54.711
244202 93.515 -74.443 54.718
244203 93.537 -74.438 54.726
244204 93.558 -74.434 54.742
244205 93.579 -74.430 54.745
244206 93.601 -74.425 54.743
244207 93.622 -74.421 54.739
244208 93.643 -74.417 54.735
244209 93.665 -74.413 54.734
244210 93.686 -74.408 54.738
244211 93.708 -74.404 54.747
244212 93.729 -74.400 54.761
244213 93.751 -74.396 54.778
244214 93.775 -74.396 54.795
244215 93.806 -74.408 54.807
244216 93.838 -74.421 54.805
244217 93.870 -74.433 54.794
244218 93.902 -74.446 54.784
244219 93.929 -74.454 54.776
244220 93.951 -74.456 54.772
244221 93.973 -74.458 54.767
244222 93.995 -74.461 54.757
244223 94.017 -74.463 54.743
244224 94.039 -74.465 54.728
244225 94.060 -74.466 54.712
244226 94.080 -74.466 54.698
244227 94.101 -74.465 54.686
244228 94.121 -74.465 54.674
244229 94.138 -74.464 54.659
244230 94.155 -74.461 54.642
244231 94.175 -74.461 54.623
244232 94.197 -74.461 54.602
244233 94.218 -74.462 54.583
244234 94.239 -74.462 54.565
244235 94.260 -74.463 54.549
244236 94.281 -74.463 54.527
244237 94.301 -74.464 54.505
244238 94.322 -74.464 54.480
244239 94.342 -74.465 54.452
244240 94.361 -74.466 54.424
244241 94.381 -74.467 54.395
244242 94.401 -74.468 54.365
244243 94.421 -74.471 54.333
244244 94.442 -74.478 54.302
244245 94.464 -74.485 54.271
244246 94.485 -74.491 54.241
244247 94.507 -74.497 54.211
244248 94.532 -74.496 54.176
244249 94.549 -74.505 54.138
244250 94.566 -74.515 54.099
244251 94.582 -74.524 54.059
244252 94.599 -74.534 54.023
244253 94.616 -74.544 53.987
244254 94.632 -74.553 53.951
244255 94.651 -74.563 53.916
244256 94.675 -74.568 53.878
244257 94.698 -74.573 53.842
244258 94.722 -74.578 53.809
244259 94.746 -74.583 53.774
244260 94.770 -74.588 53.743
244261 94.793 -74.593 53.712
244262 94.817 -74.598 53.681
244263 94.841 -74.603 53.650
244264 94.864 -74.608 53.617
244265 94.872 -74.603 53.584
244266 94.871 -74.592 53.550
244267 94.888 -74.581 53.514
244268 94.911 -74.570 53.475
244269 94.934 -74.559 53.436
244270 94.957 -74.548 53.394
244271 94.981 -74.537 53.351
244272 95.004 -74.527 53.305
244273 95.034 -74.534 53.254
244274 95.068 -74.549 53.198
244275 95.102 -74.565 53.135
244276 95.136 -74.581 53.068
244277 95.170 -74.597 53.000
244278 95.204 -74.614 52.932
244279 95.238 -74.630 52.866
244280 95.264 -74.650 52.803
244281 95.288 -74.672 52.741
244282 95.321 -74.689 52.679
244283 95.354 -74.705 52.622
244284 95.388 -74.720 52.564
244285 95.415 -74.731 52.507
244286 95.442 -74.737 52.452
244287 95.467 -74.737 52.396
244288 95.492 -74.736 52.342
244289 95.517 -74.734 52.290
244290 95.542 -74.734 52.239
244291 95.566 -74.733 52.190
244292 95.591 -74.732 52.142
244293 95.616 -74.731 52.096
244294 95.641 -74.730 52.052
244295 95.666 -74.729 52.009
244296 95.691 -74.728 51.968
244297 95.716 -74.727 51.928
244298 95.741 -74.726 51.888
244299 95.768 -74.715 51.848
244300 95.801 -74.689 51.808
244301 95.834 -74.662 51.769
244302 95.872 -74.668 51.729
244303 95.910 -74.679 51.691
244304 95.949 -74.690 51.654
244305 95.987 -74.700 51.617
244306 96.026 -74.711 51.580
244307 96.064 -74.722 51.545
244308 96.102 -74.732 51.510
244309 96.141 -74.743 51.477
244310 96.179 -74.752 51.447
244311 96.217 -74.756 51.417
244312 96.254 -74.761 51.389
244313 96.275 -74.769 51.363
244314 96.301 -74.779 51.340
244315 96.326 -74.789 51.318
244316 96.352 -74.800 51.298
244317 96.377 -74.809 51.281
244318 96.403 -74.819 51.265
244319 96.429 -74.829 51.251
244320 96.454 -74.839 51.238
244321 96.479 -74.849 51.226
244322 96.504 -74.863 51.215
244323 96.530 -74.876 51.206
244324 96.556 -74.889 51.199
244325 96.583 -74.902 51.192
244326 96.609 -74.915 51.184
244327 96.636 -74.928 51.175
244328 96.662 -74.941 51.165
244329 96.688 -74.954 51.152
244330 96.715 -74.967 51.137
244331 96.741 -74.980 51.118
244332 96.767 -74.993 51.098
244333 96.795 -75.019 51.077
244334 96.817 -75.041 51.057
244335 96.831 -75.040 51.036
244336 96.844 -75.038 51.017
244337 96.859 -75.035 50.998
244338 96.872 -75.034 50.981
244339 96.886 -75.032 50.965
244340 96.900 -75.030 50.950
244341 96.919 -75.026 50.937
244342 96.942 -75.021 50.923
244343 96.962 -75.026 50.908
244344 96.983 -75.033 50.891
244345 97.005 -75.041 50.873
244346 97.026 -75.048 50.853
244347 97.048 -75.055 50.832
244348 97.069 -75.063 50.810
244349 97.091 -75.069 50.788
244350 97.112 -75.077 50.767
244351 97.134 -75.085 50.747
244352 97.155 -75.091 50.728
244353 97.177 -75.099 50.711
244354 97.198 -75.107 50.695
244355 97.220 -75.113 50.677
244356 97.242 -75.122 50.657
244357 97.266 -75.131 50.634
244358 97.291 -75.140 50.608
244359 97.315 -75.149 50.579
244360 97.339 -75.158 50.548
244361 97.362 -75.167 50.515
244362 97.379 -75.176 50.478
244363 97.395 -75.184 50.438
244364 97.413 -75.192 50.397
244365 97.436 -75.199 50.355
244366 97.466 -75.206 50.317
244367 97.497 -75.212 50.281
244368 97.527 -75.219 50.247
244369 97.558 -75.225 50.211
244370 97.589 -75.232 50.173
244371 97.615 -75.238 50.136
244372 97.641 -75.245 50.098
244373 97.666 -75.251 50.060
244374 97.697 -75.258 50.019
244375 97.731 -75.264 49.976
244376 97.764 -75.272 49.931
244377 97.798 -75.279 49.885
244378 97.832 -75.286 49.837
244379 97.888 -75.298 49.789
244380 97.943 -75.311 49.742
244381 97.999 -75.323 49.694
244382 98.054 -75.335 49.646
244383 98.088 -75.346 49.597
244384 98.122 -75.357 49.546
244385 98.144 -75.367 49.498
244386 98.157 -75.377 49.450
244387 98.171 -75.386 49.403
244388 98.184 -75.395 49.355
244389 98.198 -75.404 49.311
244390 98.211 -75.414 49.265
244391 98.225 -75.423 49.217
244392 98.262 -75.426 49.170
244393 98.311 -75.424 49.121
244394 98.360 -75.424 49.074
244395 98.408 -75.426 49.028
244396 98.436 -75.415 48.982
244397 98.481 -75.400 48.935
244398 98.526 -75.385 48.888
244399 98.571 -75.369 48.840
244400 98.574 -75.388 48.789
244401 98.563 -75.420 48.740
244402 98.551 -75.452 48.689
244403 98.539 -75.483 48.638
244404 98.570 -75.496 48.588
244405 98.609 -75.502 48.541
244406 98.649 -75.509 48.493
244407 98.688 -75.516 48.445
244408 98.727 -75.523 48.396
244409 98.767 -75.530 48.348
244410 98.802 -75.532 48.297
244411 98.826 -75.521 48.246
244412 98.849 -75.511 48.197
244413 98.873 -75.501 48.149
244414 98.891 -75.520 48.102
244415 98.909 -75.544 48.056
244416 98.926 -75.569 48.011
244417 98.943 -75.594 47.970
244418 98.960 -75.619 47.930
244419 98.978 -75.644 47.892
244420 98.995 -75.668 47.857
244421 99.012 -75.693 47.824
244422 99.029 -75.718 47.791
244423 99.047 -75.743 47.757
244424 99.064 -75.768 47.726
244425 99.081 -75.793 47.694
244426 99.102 -75.825 47.662
244427 99.125 -75.858 47.631
244428 99.152 -75.876 47.600
244429 99.190 -75.862 47.570
244430 99.224 -75.856 47.542
244431 99.253 -75.862 47.514
244432 99.282 -75.868 47.486
244433 99.311 -75.874 47.458
244434 99.340 -75.880 47.431
244435 99.369 -75.886 47.403
244436 99.398 -75.892 47.375
244437 99.427 -75.897 47.349
244438 99.456 -75.903 47.324
244439 99.485 -75.909 47.300
244440 99.514 -75.916 47.276
244441 99.543 -75.923 47.253
244442 99.567 -75.939 47.229
244443 99.590 -75.956 47.207
244444 99.615 -75.972 47.186
244445 99.641 -75.990 47.164
244446 99.666 -76.014 47.141
244447 99.705 -76.014 47.117
244448 99.734 -76.019 47.090
244449 99.764 -76.025 47.062
244450 99.794 -76.030 47.034
244451 99.824 -76.035 47.007
244452 99.853 -76.041 46.978
244453 99.883 -76.046 46.948
244454 99.910 -76.054 46.918
244455 99.934 -76.067 46.887
244456 99.959 -76.080 46.855
244457 99.982 -76.094 46.825
244458 100.001 -76.117 46.797
244459 100.020 -76.140 46.767
244460 100.040 -76.163 46.737
244461 100.059 -76.185 46.707
244462 100.085 -76.191 46.675
244463 100.120 -76.181 46.644
244464 100.154 -76.170 46.614
244465 100.188 -76.160 46.584
244466 100.213 -76.157 46.554
244467 100.238 -76.154 46.524
244468 100.262 -76.151 46.492
244469 100.287 -76.147 46.461
244470 100.327 -76.150 46.430
244471 100.368 -76.152 46.399
244472 100.408 -76.154 46.368
244473 100.433 -76.182 46.336
244474 100.452 -76.223 46.304
244475 100.469 -76.263 46.271
244476 100.487 -76.304 46.237
244477 100.505 -76.345 46.204
244478 100.525 -76.383 46.171
244479 100.548 -76.404 46.137
244480 100.570 -76.412 46.103
244481 100.591 -76.420 46.069
244482 100.613 -76.428 46.036
244483 100.634 -76.436 46.003
244484 100.656 -76.444 45.972
244485 100.678 -76.452 45.941
244486 100.690 -76.477 45.910
244487 100.698 -76.511 45.880
244488 100.706 -76.544 45.852
244489 100.714 -76.577 45.824
244490 100.729 -76.602 45.797
244491 100.764 -76.606 45.772
244492 100.809 -76.600 45.748
244493 100.853 -76.595 45.725
244494 100.898 -76.589 45.702
244495 100.943 -76.583 45.680
244496 100.988 -76.578 45.657
244497 101.032 -76.572 45.635
244498 101.077 -76.567 45.614
244499 101.122 -76.561 45.593
244500 101.167 -76.555 45.570
244501 101.211 -76.550 45.547
244502 101.252 -76.547 45.525
244503 101.275 -76.557 45.503
244504 101.299 -76.566 45.482
244505 101.323 -76.575 45.460
244506 101.347 -76.584 45.440
244507 101.375 -76.595 45.418
244508 101.403 -76.605 45.398
244509 101.426 -76.616 45.378
244510 101.441 -76.628 45.359
244511 101.455 -76.640 45.339
244512 101.470 -76.652 45.321
244513 101.485 -76.665 45.303
244514 101.499 -76.677 45.285
244515 101.515 -76.688 45.267
244516 101.529 -76.700 45.250
244517 101.544 -76.712 45.233
244518 101.558 -76.724 45.217
244519 101.573 -76.737 45.200
244520 101.588 -76.749 45.184
244521 101.603 -76.761 45.168
244522 101.618 -76.771 45.152
244523 101.640 -76.771 45.136
244524 101.662 -76.771 45.120
244525 101.683 -76.771 45.104
244526 101.705 -76.771 45.087
244527 101.727 -76.771 45.071
244528 101.744 -76.780 45.055
244529 101.760 -76.791 45.037
244530 101.777 -76.802 45.020
244531 101.793 -76.813 45.003
244532 101.810 -76.824 44.987
244533 101.826 -76.836 44.969
244534 101.842 -76.846 44.952
244535 101.858 -76.858 44.935
244536 101.875 -76.869 44.918
244537 101.891 -76.881 44.900
244538 101.908 -76.891 44.884
244539 101.924 -76.903 44.867
244540 101.941 -76.914 44.850
244541 101.957 -76.925 44.833
244542 101.970 -76.934 44.817
244543 101.964 -76.927 44.800
244544 101.971 -76.930 44.783
244545 101.979 -76.933 44.766
244546 101.986 -76.935 44.750
244547 101.999 -76.943 44.734
244548 102.021 -76.963 44.718
244549 102.043 -76.983 44.701
244550 102.064 -77.003 44.685
244551 102.087 -77.022 44.669
244552 102.108 -77.042 44.652
244553 102.130 -77.062 44.636
244554 102.152 -77.081 44.620
244555 102.174 -77.101 44.604
244556 102.196 -77.120 44.588
244557 102.218 -77.140 44.571
244558 102.239 -77.160 44.556
244559 102.259 -77.169 44.541
244560 102.276 -77.169 44.525
244561 102.293 -77.169 44.510
244562 102.316 -77.170 44.494
244563 102.339 -77.172 44.478
244564 102.362 -77.172 44.463
244565 102.385 -77.173 44.447
244566 102.408 -77.174 44.432
244567 102.431 -77.175 44.416
244568 102.454 -77.176 44.400
244569 102.477 -77.177 44.384
244570 102.501 -77.179 44.367
244571 102.523 -77.179 44.351
244572 102.545 -77.181 44.336
244573 102.564 -77.182 44.319
244574 102.584 -77.185 44.303
244575 102.604 -77.186 44.287
244576 102.624 -77.188 44.271
244577 102.644 -77.190 44.254
244578 102.663 -77.191 44.238
244579 102.683 -77.194 44.223
244580 102.703 -77.195 44.208
244581 102.723 -77.197 44.192
244582 102.743 -77.199 44.177
244869 92.555 -74.483 54.500
244870 92.576 -74.481 54.507
244871 92.597 -74.467 54.513
244872 92.624 -74.457 54.518
244873 92.650 -74.448 54.523
244874 92.679 -74.440 54.527
244875 92.709 -74.431 54.530
244876 92.739 -74.423 54.532
244877 92.770 -74.415 54.534
244878 92.800 -74.427 54.536
244879 92.830 -74.446 54.537
244880 92.860 -74.452 54.537
244881 92.891 -74.454 54.537
244882 92.922 -74.455 54.535
244883 92.953 -74.456 54.534
244884 92.984 -74.457 54.535
244885 93.015 -74.458 54.536
244886 93.046 -74.459 54.538
244887 93.077 -74.461 54.541
244888 93.109 -74.461 54.547
244889 93.140 -74.462 54.556
244890 93.172 -74.466 54.570
244891 93.205 -74.472 54.583
244892 93.238 -74.478 54.595
244893 93.272 -74.483 54.606
244894 93.305 -74.490 54.612
244895 93.335 -74.490 54.620
244896 93.358 -74.480 54.631
244897 93.381 -74.470 54.646
244898 93.405 -74.460 54.666
244899 93.428 -74.450 54.688
244900 93.451 -74.440 54.709
244901 93.474 -74.434 54.730
244902 93.495 -74.430 54.744
244903 93.516 -74.426 54.752
244904 93.538 -74.421 54.756
244905 93.559 -74.418 54.764
244906 93.580 -74.413 54.767
244907 93.602 -74.408 54.765
244908 93.623 -74.405 54.761
244909 93.645 -74.400 54.763
244910 93.666 -74.396 54.762
244911 93.688 -74.392 54.764
244912 93.709 -74.387 54.771
244913 93.730 -74.383 54.782
244914 93.752 -74.379 54.798
244915 93.773 -74.375 54.814
244916 93.799 -74.377 54.825
244917 93.831 -74.390 54.823
244918 93.862 -74.402 54.813
244919 93.894 -74.415 54.802
244920 93.923 -74.424 54.793
244921 93.945 -74.427 54.785
244922 93.967 -74.429 54.776
244923 93.989 -74.431 54.763
244924 94.011 -74.433 54.747
244925 94.033 -74.436 54.731
244926 94.055 -74.438 54.712
244927 94.075 -74.438 54.694
244928 94.096 -74.438 54.678
244929 94.116 -74.438 54.664
244930 94.136 -74.437 54.650
244931 94.154 -74.436 54.633
244932 94.175 -74.436 54.615
244933 94.195 -74.436 54.594
244934 94.216 -74.436 54.573
244935 94.237 -74.436 54.554
244936 94.258 -74.437 54.536
244937 94.279 -74.437 54.516
244938 94.300 -74.438 54.495
244939 94.321 -74.438 54.472
244940 94.341 -74.439 54.446
244941 94.360 -74.440 54.419
244942 94.380 -74.441 54.392
244943 94.400 -74.442 54.364
244944 94.420 -74.447 54.334
244945 94.442 -74.454 54.304
244946 94.464 -74.461 54.274
244947 94.485 -74.466 54.244
244948 94.510 -74.465 54.213
244949 94.534 -74.463 54.179
244950 94.553 -74.470 54.142
244951 94.570 -74.479 54.104
244952 94.586 -74.489 54.066
244953 94.603 -74.499 54.027
244954 94.620 -74.508 53.990
244955 94.636 -74.518 53.953
244956 94.652 -74.521 53.916
244957 94.672 -74.534 53.877
244958 94.692 -74.546 53.839
244959 94.712 -74.558 53.803
244960 94.735 -74.564 53.767
244961 94.759 -74.568 53.735
244962 94.782 -74.574 53.703
244963 94.806 -74.578 53.672
244964 94.830 -74.584 53.642
244965 94.852 -74.587 53.610
244966 94.851 -74.577 53.577
244967 94.871 -74.566 53.543
244968 94.894 -74.555 53.509
244969 94.918 -74.544 53.473
244970 94.941 -74.534 53.436
244971 94.964 -74.523 53.398
244972 94.987 -74.512 53.356
244973 95.011 -74.501 53.309
244974 95.044 -74.516 53.257
244975 95.078 -74.532 53.202
244976 95.112 -74.548 53.140
244977 95.146 -74.564 53.073
244978 95.180 -74.581 53.004
244979 95.214 -74.596 52.935
244980 95.248 -74.612 52.868
244981 95.280 -74.630 52.804
244982 95.305 -74.652 52.742
244983 95.329 -74.674 52.681
244984 95.362 -74.689 52.623
244985 95.396 -74.705 52.565
244986 95.424 -74.715 52.509
244987 95.450 -74.718 52.454
244988 95.474 -74.717 52.400
244989 95.499 -74.715 52.347
244990 95.524 -74.715 52.295
244991 95.549 -74.714 52.244
244992 95.573 -74.713 52.195
244993 95.599 -74.712 52.148
244994 95.623 -74.711 52.102
244995 95.648 -74.710 52.058
244996 95.673 -74.709 52.015
244997 95.698 -74.709 51.974
244998 95.723 -74.707 51.934
244999 95.748 -74.706 51.894
245000 95.772 -74.706 51.853
245001 95.804 -74.684 51.813
245002 95.839 -74.678 51.773
245003 95.878 -74.689 51.733
245004 95.916 -74.700 51.695
245005 95.955 -74.710 51.658
245006 95.993 -74.719 51.621
245007 96.030 -74.724 51.583
245008 96.068 -74.728 51.547
245009 96.105 -74.733 51.512
245010 96.143 -74.737 51.478
245011 96.180 -74.742 51.446
245012 96.217 -74.747 51.416
245013 96.254 -74.752 51.388
245014 96.267 -74.761 51.362
245015 96.284 -74.772 51.336
245016 96.309 -74.781 51.314
245017 96.334 -74.792 51.293
245018 96.360 -74.802 51.275
245019 96.386 -74.812 51.258
245020 96.411 -74.822 51.244
245021 96.437 -74.831 51.231
245022 96.481 -74.826 51.218
245023 96.508 -74.833 51.207
245024 96.534 -74.840 51.199
245025 96.560 -74.847 51.192
245026 96.589 -74.858 51.186
245027 96.617 -74.869 51.180
245028 96.645 -74.879 51.173
245029 96.674 -74.890 51.164
245030 96.702 -74.900 51.153
245031 96.730 -74.910 51.139
245032 96.759 -74.921 51.122
245033 96.786 -74.934 51.102
245034 96.812 -74.952 51.082
245035 96.842 -74.997 51.063
245036 96.859 -75.002 51.042
245037 96.872 -75.001 51.023
245038 96.886 -74.998 51.006
245039 96.900 -74.997 50.989
245040 96.914 -74.994 50.972
245041 96.928 -74.992 50.958
245042 96.941 -74.991 50.946
245043 96.947 -74.994 50.932
245044 96.957 -75.003 50.917
245045 96.980 -75.005 50.900
245046 97.002 -75.012 50.882
245047 97.023 -75.019 50.863
245048 97.045 -75.026 50.843
245049 97.066 -75.034 50.822
245050 97.088 -75.041 50.802
245051 97.109 -75.048 50.782
245052 97.131 -75.056 50.764
245053 97.152 -75.063 50.748
245054 97.174 -75.070 50.733
245055 97.195 -75.078 50.717
245056 97.217 -75.085 50.699
245057 97.239 -75.093 50.678
245058 97.263 -75.102 50.655
245059 97.288 -75.111 50.629
245060 97.312 -75.120 50.601
245061 97.336 -75.129 50.571
245062 97.360 -75.138 50.538
245063 97.380 -75.147 50.500
245064 97.397 -75.155 50.460
245065 97.414 -75.163 50.419
245066 97.439 -75.170 50.378
245067 97.469 -75.177 50.339
245068 97.500 -75.183 50.303
245069 97.531 -75.190 50.269
245070 97.561 -75.197 50.233
245071 97.592 -75.203 50.195
245072 97.620 -75.210 50.157
245073 97.647 -75.216 50.119
245074 97.681 -75.223 50.079
245075 97.715 -75.230 50.037
245076 97.748 -75.237 49.993
245077 97.782 -75.245 49.948
245078 97.815 -75.251 49.901
245079 97.849 -75.258 49.853
245080 97.896 -75.269 49.806
245081 97.952 -75.281 49.760
245082 98.007 -75.293 49.712
245083 98.048 -75.305 49.663
245084 98.081 -75.316 49.612
245085 98.115 -75.326 49.561
245086 98.148 -75.337 49.512
245087 98.172 -75.348 49.465
245088 98.185 -75.357 49.417
245089 98.199 -75.366 49.369
245090 98.212 -75.376 49.325
245091 98.226 -75.385 49.277
245092 98.239 -75.394 49.229
245093 98.275 -75.397 49.180
245094 98.323 -75.396 49.132
245095 98.372 -75.395 49.085
245096 98.418 -75.408 49.038
245097 98.439 -75.392 48.991
245098 98.474 -75.375 48.944
245099 98.519 -75.359 48.897
245100 98.564 -75.344 48.849
245101 98.574 -75.358 48.798
245102 98.562 -75.389 48.748
245103 98.551 -75.420 48.696
245104 98.547 -75.446 48.644
245105 98.574 -75.448 48.593
245106 98.605 -75.451 48.545
245107 98.645 -75.458 48.498
245108 98.684 -75.464 48.449
245109 98.723 -75.471 48.401
245110 98.763 -75.478 48.352
245111 98.802 -75.486 48.300
245112 98.829 -75.478 48.249
245113 98.852 -75.467 48.199
245114 98.876 -75.457 48.150
245115 98.894 -75.480 48.102
245116 98.911 -75.505 48.057
245117 98.928 -75.530 48.013
245118 98.945 -75.555 47.971
245119 98.962 -75.580 47.931
245120 98.980 -75.605 47.891
245121 98.997 -75.630 47.853
245122 99.014 -75.655 47.820
245123 99.031 -75.680 47.786
245124 99.049 -75.705 47.752
245125 99.071 -75.738 47.720
245126 99.093 -75.771 47.686
245127 99.116 -75.806 47.653
245128 99.142 -75.831 47.622
245129 99.179 -75.818 47.591
245130 99.216 -75.804 47.559
245131 99.253 -75.791 47.530
245132 99.282 -75.797 47.502
245133 99.311 -75.803 47.473
245134 99.340 -75.809 47.446
245135 99.369 -75.815 47.418
245136 99.398 -75.821 47.391
245137 99.427 -75.827 47.363
245138 99.456 -75.833 47.337
245139 99.483 -75.843 47.312
245140 99.507 -75.859 47.288
245141 99.532 -75.876 47.265
245142 99.556 -75.892 47.242
245143 99.579 -75.909 47.220
245144 99.604 -75.925 47.199
245145 99.629 -75.943 47.178
245146 99.655 -75.959 47.156
245147 99.678 -75.988 47.134
245148 99.708 -76.004 47.110
245149 99.749 -76.003 47.084
245150 99.782 -76.007 47.057
245151 99.811 -76.013 47.029
245152 99.841 -76.018 47.002
245153 99.871 -76.022 46.974
245154 99.901 -76.028 46.944
245155 99.926 -76.039 46.914
245156 99.950 -76.052 46.883
245157 99.974 -76.065 46.850
245158 99.998 -76.077 46.820
245159 100.021 -76.093 46.791
245160 100.041 -76.116 46.761
245161 100.060 -76.138 46.731
245162 100.079 -76.161 46.700
245163 100.101 -76.176 46.669
245164 100.136 -76.166 46.636
245165 100.170 -76.155 46.606
245166 100.204 -76.144 46.576
245167 100.239 -76.134 46.546
245168 100.264 -76.130 46.515
245169 100.289 -76.127 46.484
245170 100.314 -76.123 46.452
245171 100.349 -76.124 46.420
245172 100.390 -76.127 46.389
245173 100.430 -76.129 46.358
245174 100.447 -76.170 46.327
245175 100.466 -76.210 46.294
245176 100.483 -76.251 46.261
245177 100.501 -76.292 46.227
245178 100.523 -76.327 46.194
245179 100.546 -76.360 46.162
245180 100.569 -76.393 46.128
245181 100.592 -76.402 46.094
245182 100.614 -76.409 46.061
245183 100.636 -76.417 46.029
245184 100.657 -76.425 45.996
245185 100.679 -76.433 45.966
245186 100.700 -76.442 45.935
245187 100.708 -76.475 45.905
245188 100.716 -76.508 45.875
245189 100.725 -76.542 45.846
245190 100.733 -76.576 45.818
245191 100.747 -76.602 45.792
245192 100.789 -76.598 45.768
245193 100.834 -76.592 45.743
245194 100.879 -76.586 45.719
245195 100.924 -76.580 45.697
245196 100.968 -76.575 45.674
245197 101.013 -76.570 45.652
245198 101.058 -76.564 45.629
245199 101.103 -76.558 45.607
245200 101.147 -76.552 45.585
245201 101.192 -76.547 45.563
245202 101.237 -76.541 45.541
245203 101.281 -76.536 45.519
245204 101.306 -76.545 45.497
245205 101.330 -76.554 45.475
245206 101.354 -76.563 45.454
245207 101.378 -76.572 45.433
245208 101.406 -76.583 45.412
245209 101.434 -76.593 45.391
245210 101.454 -76.605 45.371
245211 101.469 -76.617 45.352
245212 101.484 -76.628 45.333
245213 101.498 -76.640 45.314
245214 101.513 -76.652 45.296
245215 101.528 -76.665 45.278
245216 101.543 -76.677 45.260
245217 101.557 -76.689 45.242
245218 101.572 -76.701 45.225
245219 101.587 -76.712 45.209
245220 101.602 -76.724 45.192
245221 101.616 -76.737 45.175
245222 101.632 -76.746 45.158
245223 101.654 -76.746 45.142
245224 101.676 -76.746 45.126
245225 101.697 -76.746 45.109
245226 101.719 -76.746 45.093
245227 101.740 -76.749 45.076
245228 101.756 -76.760 45.060
245229 101.772 -76.771 45.043
245230 101.788 -76.782 45.027
245231 101.805 -76.793 45.009
245232 101.821 -76.805 44.992
245233 101.838 -76.815 44.975
245234 101.854 -76.827 44.958
245235 101.871 -76.838 44.941
245236 101.887 -76.850 44.923
245237 101.903 -76.860 44.906
245238 101.919 -76.871 44.888
245239 101.936 -76.883 44.872
245240 101.952 -76.893 44.855
245241 101.957 -76.896 44.838
245242 101.952 -76.890 44.821
245243 101.960 -76.893 44.805
245244 101.967 -76.896 44.788
245245 101.975 -76.899 44.771
245246 101.983 -76.901 44.755
245247 101.999 -76.915 44.739
245248 102.021 -76.934 44.723
245249 102.043 -76.953 44.706
245250 102.064 -76.973 44.689
245251 102.087 -76.993 44.673
245252 102.108 -77.012 44.657
245253 102.130 -77.032 44.641
245254 102.152 -77.052 44.624
245255 102.174 -77.072 44.608
245256 102.196 -77.091 44.592
245257 102.218 -77.110 44.576
245258 102.239 -77.128 44.559
245259 102.256 -77.128 44.544
245260 102.274 -77.128 44.528
245261 102.291 -77.129 44.513
245262 102.314 -77.130 44.498
245263 102.337 -77.132 44.482
245264 102.360 -77.132 44.466
245265 102.383 -77.133 44.450
245266 102.406 -77.134 44.435
245267 102.429 -77.135 44.419
245268 102.452 -77.136 44.404
245269 102.475 -77.137 44.389
245270 102.498 -77.138 44.373
245271 102.521 -77.139 44.356
245272 102.544 -77.140 44.340
245273 102.567 -77.141 44.324
245274 102.589 -77.142 44.308
245275 102.609 -77.144 44.292
245276 102.628 -77.146 44.277
245277 102.649 -77.147 44.260
245278 102.668 -77.150 44.244
245279 102.688 -77.151 44.228
245280 102.708 -77.153 44.213
245281 102.728 -77.155 44.198
245571 92.580 -74.458 54.514
245572 92.598 -74.440 54.519
245573 92.625 -74.430 54.525
245574 92.652 -74.421 54.529
245575 92.681 -74.412 54.532
245576 92.711 -74.404 54.535
245577 92.741 -74.396 54.537
245578 92.771 -74.388 54.540
245579 92.801 -74.400 54.541
245580 92.831 -74.419 54.542
245581 92.861 -74.431 54.542
245582 92.892 -74.433 54.541
245583 92.923 -74.433 54.540
245584 92.954 -74.435 54.539
245585 92.985 -74.436 54.538
245586 93.016 -74.436 54.538
245587 93.047 -74.438 54.539
245588 93.078 -74.439 54.542
245589 93.110 -74.440 54.546
245590 93.140 -74.441 54.552
245591 93.171 -74.443 54.562
245592 93.203 -74.443 54.573
245593 93.235 -74.448 54.585
245594 93.269 -74.454 54.599
245595 93.302 -74.460 54.609
245596 93.335 -74.466 54.622
245597 93.361 -74.459 54.637
245598 93.384 -74.449 54.656
245599 93.408 -74.440 54.678
245600 93.431 -74.430 54.701
245601 93.453 -74.421 54.725
245602 93.475 -74.418 54.748
245603 93.496 -74.413 54.764
245604 93.518 -74.409 54.775
245605 93.539 -74.405 54.782
245606 93.561 -74.401 54.786
245607 93.582 -74.396 54.786
245608 93.604 -74.392 54.785
245609 93.625 -74.388 54.786
245610 93.646 -74.383 54.793
245611 93.667 -74.380 54.795
245612 93.689 -74.375 54.796
245613 93.710 -74.371 54.796
245614 93.732 -74.367 54.799
245615 93.753 -74.362 54.812
245616 93.775 -74.358 54.826
245617 93.796 -74.354 54.838
245618 93.823 -74.359 54.836
245619 93.855 -74.371 54.825
245620 93.887 -74.384 54.811
245621 93.917 -74.395 54.799
245622 93.939 -74.397 54.788
245623 93.961 -74.399 54.778
245624 93.983 -74.402 54.765
245625 94.005 -74.404 54.749
245626 94.027 -74.406 54.732
245627 94.049 -74.408 54.714
245628 94.071 -74.411 54.693
245629 94.091 -74.410 54.671
245630 94.111 -74.410 54.654
245631 94.131 -74.410 54.639
245632 94.151 -74.410 54.623
245633 94.173 -74.410 54.605
245634 94.194 -74.411 54.584
245635 94.215 -74.411 54.564
245636 94.236 -74.411 54.543
245637 94.257 -74.411 54.524
245638 94.277 -74.412 54.505
245639 94.298 -74.412 54.485
245640 94.319 -74.413 54.463
245641 94.340 -74.413 54.439
245642 94.359 -74.414 54.415
245643 94.379 -74.415 54.390
245644 94.399 -74.417 54.364
245645 94.420 -74.423 54.336
245646 94.441 -74.430 54.307
245647 94.464 -74.435 54.278
245648 94.488 -74.433 54.248
245649 94.512 -74.432 54.215
245650 94.537 -74.430 54.182
245651 94.557 -74.434 54.146
245652 94.574 -74.444 54.109
245653 94.591 -74.454 54.073
245654 94.607 -74.464 54.033
245655 94.624 -74.474 53.993
245656 94.640 -74.482 53.954
245657 94.655 -74.482 53.915
245658 94.674 -74.491 53.874
245659 94.694 -74.503 53.834
245660 94.714 -74.515 53.795
245661 94.734 -74.527 53.758
245662 94.753 -74.540 53.723
245663 94.773 -74.552 53.691
245664 94.795 -74.559 53.660
245665 94.819 -74.565 53.629
245666 94.832 -74.562 53.598
245667 94.855 -74.552 53.566
245668 94.878 -74.540 53.534
245669 94.901 -74.530 53.500
245670 94.924 -74.519 53.465
245671 94.948 -74.508 53.429
245672 94.971 -74.497 53.393
245673 94.994 -74.487 53.352
245674 95.021 -74.483 53.305
245675 95.054 -74.499 53.255
245676 95.089 -74.515 53.201
245677 95.122 -74.530 53.140
245678 95.157 -74.546 53.074
245679 95.190 -74.563 53.006
245680 95.225 -74.579 52.937
245681 95.259 -74.595 52.869
245682 95.293 -74.611 52.805
245683 95.321 -74.631 52.743
245684 95.345 -74.653 52.682
245685 95.370 -74.674 52.625
245686 95.404 -74.690 52.568
245687 95.431 -74.699 52.512
245688 95.457 -74.698 52.458
245689 95.481 -74.697 52.403
245690 95.507 -74.696 52.350
245691 95.531 -74.695 52.299
245692 95.556 -74.694 52.249
245693 95.581 -74.693 52.201
245694 95.606 -74.692 52.155
245695 95.630 -74.691 52.110
245696 95.656 -74.690 52.065
245697 95.680 -74.690 52.021
245698 95.705 -74.688 51.980
245699 95.730 -74.687 51.939
245700 95.755 -74.687 51.899
245701 95.779 -74.686 51.858
245702 95.807 -74.686 51.819
245703 95.844 -74.690 51.780
245704 95.882 -74.696 51.739
245705 95.919 -74.700 51.701
245706 95.957 -74.705 51.663
245707 95.994 -74.709 51.625
245708 96.031 -74.714 51.587
245709 96.068 -74.718 51.550
245710 96.106 -74.723 51.515
245711 96.143 -74.728 51.480
245712 96.180 -74.733 51.447
245713 96.218 -74.737 51.416
245714 96.255 -74.742 51.386
245715 96.266 -74.752 51.358
245716 96.266 -74.764 51.333
245717 96.292 -74.774 51.309
245718 96.317 -74.784 51.288
245719 96.343 -74.794 51.269
245720 96.368 -74.804 51.252
245721 96.394 -74.814 51.237
245722 96.419 -74.824 51.223
245723 96.485 -74.800 51.211
245724 96.511 -74.807 51.200
245725 96.537 -74.815 51.192
245726 96.564 -74.822 51.185
245727 96.589 -74.828 51.179
245728 96.615 -74.835 51.174
245729 96.643 -74.845 51.168
245730 96.672 -74.856 51.161
245731 96.700 -74.866 51.151
245732 96.728 -74.877 51.140
245733 96.757 -74.888 51.124
245734 96.785 -74.897 51.107
245735 96.813 -74.908 51.087
245736 96.859 -74.932 51.069
245737 96.886 -74.965 51.049
245738 96.900 -74.963 51.030
245739 96.914 -74.961 51.014
245740 96.927 -74.960 50.996
245741 96.934 -74.963 50.980
245742 96.940 -74.965 50.966
245743 96.946 -74.968 50.954
245744 96.953 -74.971 50.940
245745 96.965 -74.980 50.924
245746 96.975 -74.991 50.906
245747 96.993 -74.997 50.888
245748 97.016 -74.999 50.870
245749 97.039 -75.002 50.851
245750 97.063 -75.005 50.833
245751 97.085 -75.012 50.815
245752 97.106 -75.019 50.799
245753 97.127 -75.027 50.784
245754 97.149 -75.034 50.770
245755 97.171 -75.041 50.755
245756 97.192 -75.048 50.738
245757 97.214 -75.056 50.719
245758 97.236 -75.063 50.697
245759 97.260 -75.073 50.672
245760 97.284 -75.082 50.646
245761 97.308 -75.091 50.619
245762 97.333 -75.101 50.591
245763 97.357 -75.110 50.558
245764 97.381 -75.119 50.520
245765 97.399 -75.127 50.479
245766 97.416 -75.135 50.439
245767 97.442 -75.142 50.401
245768 97.473 -75.148 50.363
245769 97.503 -75.155 50.326
245770 97.534 -75.161 50.291
245771 97.565 -75.168 50.256
245772 97.595 -75.175 50.219
245773 97.629 -75.182 50.180
245774 97.664 -75.189 50.139
245775 97.699 -75.196 50.098
245776 97.732 -75.203 50.055
245777 97.766 -75.210 50.011
245778 97.799 -75.217 49.966
245779 97.833 -75.224 49.917
245780 97.866 -75.231 49.871
245781 97.905 -75.239 49.825
245782 97.960 -75.251 49.780
245783 98.007 -75.264 49.733
245784 98.041 -75.274 49.682
245785 98.075 -75.285 49.630
245786 98.108 -75.296 49.577
245787 98.142 -75.307 49.528
245788 98.175 -75.318 49.479
245789 98.200 -75.328 49.431
245790 98.213 -75.337 49.384
245791 98.227 -75.347 49.337
245792 98.240 -75.356 49.289
245793 98.253 -75.365 49.241
245794 98.287 -75.369 49.192
245795 98.337 -75.368 49.142
245796 98.385 -75.369 49.094
245797 98.426 -75.387 49.048
245798 98.445 -75.368 49.001
245799 98.466 -75.349 48.953
245800 98.511 -75.334 48.906
245801 98.556 -75.319 48.857
245802 98.574 -75.326 48.806
245803 98.562 -75.358 48.754
245804 98.550 -75.389 48.703
245805 98.561 -75.404 48.651
245806 98.588 -75.405 48.600
245807 98.616 -75.406 48.550
245808 98.644 -75.408 48.501
245809 98.680 -75.413 48.453
245810 98.719 -75.420 48.403
245811 98.758 -75.427 48.353
245812 98.797 -75.433 48.302
245813 98.832 -75.435 48.250
245814 98.855 -75.424 48.201
245815 98.878 -75.417 48.153
245816 98.896 -75.442 48.106
245817 98.913 -75.467 48.061
245818 98.930 -75.492 48.016
245819 98.947 -75.517 47.974
245820 98.964 -75.541 47.932
245821 98.982 -75.566 47.891
245822 98.999 -75.591 47.852
245823 99.017 -75.618 47.815
245824 99.040 -75.651 47.780
245825 99.062 -75.685 47.746
245826 99.085 -75.719 47.712
245827 99.108 -75.753 47.678
245828 99.131 -75.786 47.644
245829 99.168 -75.772 47.612
245830 99.205 -75.759 47.580
245831 99.242 -75.746 47.549
245832 99.279 -75.733 47.519
245833 99.310 -75.733 47.490
245834 99.340 -75.739 47.461
245835 99.369 -75.745 47.434
245836 99.398 -75.750 47.406
245837 99.424 -75.762 47.379
245838 99.448 -75.779 47.351
245839 99.472 -75.796 47.326
245840 99.496 -75.812 47.301
245841 99.520 -75.828 47.277
245842 99.544 -75.845 47.254
245843 99.568 -75.862 47.232
245844 99.592 -75.878 47.211
245845 99.618 -75.895 47.190
245846 99.644 -75.912 47.169
245847 99.670 -75.930 47.149
245848 99.693 -75.957 47.127
245849 99.715 -75.987 47.102
245850 99.751 -75.994 47.078
245851 99.791 -75.994 47.051
245852 99.829 -75.994 47.024
245853 99.859 -76.000 46.997
245854 99.889 -76.005 46.969
245855 99.917 -76.011 46.940
245856 99.941 -76.024 46.909
245857 99.966 -76.037 46.878
245858 99.990 -76.050 46.846
245859 100.014 -76.062 46.816
245860 100.038 -76.075 46.786
245861 100.061 -76.091 46.755
245862 100.080 -76.114 46.725
245863 100.099 -76.137 46.694
245864 100.118 -76.160 46.662
245865 100.151 -76.151 46.629
245866 100.186 -76.140 46.599
245867 100.220 -76.129 46.569
245868 100.255 -76.119 46.538
245869 100.289 -76.108 46.506
245870 100.316 -76.104 46.475
245871 100.340 -76.100 46.443
245872 100.371 -76.099 46.411
245873 100.411 -76.101 46.380
245874 100.444 -76.117 46.349
245875 100.462 -76.158 46.317
245876 100.480 -76.199 46.285
245877 100.498 -76.238 46.252
245878 100.522 -76.271 46.218
245879 100.544 -76.304 46.185
245880 100.567 -76.338 46.152
245881 100.591 -76.371 46.120
245882 100.613 -76.394 46.087
245883 100.636 -76.400 46.054
245884 100.658 -76.407 46.021
245885 100.681 -76.414 45.989
245886 100.702 -76.422 45.959
245887 100.718 -76.439 45.929
245888 100.727 -76.473 45.899
245889 100.735 -76.507 45.870
245890 100.743 -76.540 45.841
245891 100.751 -76.574 45.814
245892 100.770 -76.595 45.788
245893 100.815 -76.589 45.763
245894 100.860 -76.583 45.739
245895 100.905 -76.578 45.715
245896 100.949 -76.572 45.692
245897 100.994 -76.567 45.669
245898 101.039 -76.561 45.646
245899 101.084 -76.555 45.624
245900 101.128 -76.550 45.601
245901 101.173 -76.544 45.579
245902 101.218 -76.539 45.557
245903 101.262 -76.533 45.535
245904 101.307 -76.527 45.513
245905 101.337 -76.533 45.491
245906 101.361 -76.542 45.468
245907 101.385 -76.551 45.447
245908 101.409 -76.560 45.427
245909 101.437 -76.571 45.405
245910 101.465 -76.581 45.385
245911 101.482 -76.592 45.365
245912 101.497 -76.605 45.345
245913 101.512 -76.617 45.326
245914 101.527 -76.629 45.307
245915 101.541 -76.641 45.288
245916 101.556 -76.652 45.270
245917 101.571 -76.665 45.252
245918 101.586 -76.677 45.234
245919 101.600 -76.689 45.217
245920 101.615 -76.701 45.200
245921 101.630 -76.713 45.183
245922 101.646 -76.721 45.166
245923 101.668 -76.721 45.149
245924 101.690 -76.721 45.132
245925 101.711 -76.721 45.115
245926 101.733 -76.722 45.099
245927 101.752 -76.729 45.082
245928 101.768 -76.740 45.065
245929 101.784 -76.751 45.048
245930 101.801 -76.762 45.032
245931 101.817 -76.774 45.015
245932 101.833 -76.784 44.997
245933 101.849 -76.796 44.980
245934 101.866 -76.807 44.963
245935 101.882 -76.818 44.946
245936 101.899 -76.829 44.929
245937 101.915 -76.840 44.911
245938 101.932 -76.852 44.894
245939 101.948 -76.862 44.877
245940 101.943 -76.858 44.861
245941 101.941 -76.853 44.844
245942 101.948 -76.856 44.827
245943 101.956 -76.859 44.810
245944 101.963 -76.862 44.793
245945 101.971 -76.865 44.776
245946 101.979 -76.867 44.759
245947 101.999 -76.885 44.743
245948 102.021 -76.905 44.727
245949 102.043 -76.925 44.711
245950 102.064 -76.944 44.695
245951 102.087 -76.963 44.678
245952 102.108 -76.983 44.661
245953 102.130 -77.003 44.645
245954 102.152 -77.022 44.629
245955 102.174 -77.042 44.613
245956 102.196 -77.062 44.597
245957 102.218 -77.081 44.580
245958 102.236 -77.088 44.564
245959 102.254 -77.088 44.546
245960 102.271 -77.088 44.531
245961 102.289 -77.089 44.516
245962 102.311 -77.090 44.500
245963 102.334 -77.091 44.484
245964 102.357 -77.092 44.469
245965 102.380 -77.093 44.453
245966 102.403 -77.094 44.438
245967 102.427 -77.095 44.423
245968 102.449 -77.096 44.408
245969 102.472 -77.097 44.392
245970 102.495 -77.098 44.376
245971 102.518 -77.099 44.361
245972 102.541 -77.100 44.344
245973 102.564 -77.100 44.328
245974 102.587 -77.102 44.312
245975 102.610 -77.103 44.297
245976 102.633 -77.103 44.281
245977 102.653 -77.106 44.266
245978 102.673 -77.107 44.249
245979 102.692 -77.109 44.233
245980 102.713 -77.111 44.218
246273 92.600 -74.412 54.526
246274 92.627 -74.403 54.531
246275 92.653 -74.394 54.535
246276 92.682 -74.385 54.538
246277 92.712 -74.377 54.541
246278 92.742 -74.369 54.544
246279 92.772 -74.361 54.547
246280 92.802 -74.374 54.547
246281 92.832 -74.393 54.548
246282 92.862 -74.410 54.548
246283 92.893 -74.411 54.547
246284 92.924 -74.412 54.546
246285 92.955 -74.413 54.545
246286 92.986 -74.415 54.544
246287 93.017 -74.415 54.542
246288 93.048 -74.417 54.543
246289 93.079 -74.418 54.544
246290 93.110 -74.418 54.547
246291 93.141 -74.420 54.550
246292 93.172 -74.421 54.556
246293 93.204 -74.422 54.564
246294 93.234 -74.423 54.574
246295 93.266 -74.424 54.588
246296 93.299 -74.430 54.602
246297 93.332 -74.436 54.620
246298 93.363 -74.439 54.639
246299 93.387 -74.429 54.659
246300 93.410 -74.419 54.682
246301 93.433 -74.409 54.708
246302 93.455 -74.405 54.734
246303 93.476 -74.401 54.758
246304 93.498 -74.396 54.776
246305 93.519 -74.393 54.790
246306 93.541 -74.388 54.801
246307 93.562 -74.383 54.805
246308 93.583 -74.380 54.804
246309 93.605 -74.375 54.805
246310 93.626 -74.371 54.811
246311 93.647 -74.367 54.822
246312 93.669 -74.363 54.828
246313 93.690 -74.358 54.829
246314 93.712 -74.354 54.825
246315 93.733 -74.350 54.823
246316 93.755 -74.346 54.830
246317 93.776 -74.342 54.838
246318 93.797 -74.337 54.847
246319 93.819 -74.333 54.845
246320 93.848 -74.341 54.834
246321 93.880 -74.353 54.820
246322 93.911 -74.365 54.804
246323 93.933 -74.368 54.790
246324 93.955 -74.370 54.777
246325 93.977 -74.372 54.763
246326 93.999 -74.374 54.747
246327 94.021 -74.377 54.731
246328 94.043 -74.379 54.712
246329 94.065 -74.381 54.690
246330 94.086 -74.383 54.667
246331 94.107 -74.382 54.649
246332 94.127 -74.382 54.630
246333 94.149 -74.383 54.612
246334 94.174 -74.385 54.595
246335 94.194 -74.386 54.574
246336 94.213 -74.386 54.554
246337 94.234 -74.386 54.535
246338 94.255 -74.386 54.516
246339 94.276 -74.386 54.496
246340 94.296 -74.386 54.476
246341 94.318 -74.387 54.455
246342 94.338 -74.387 54.434
246343 94.358 -74.388 54.412
246344 94.378 -74.389 54.390
246345 94.398 -74.393 54.366
246346 94.420 -74.399 54.340
246347 94.442 -74.404 54.311
246348 94.466 -74.402 54.282
246349 94.490 -74.401 54.252
246350 94.515 -74.399 54.220
246351 94.539 -74.398 54.187
246352 94.561 -74.399 54.152
246353 94.578 -74.409 54.117
246354 94.595 -74.419 54.079
246355 94.611 -74.429 54.039
246356 94.628 -74.439 53.998
246357 94.644 -74.445 53.956
246358 94.660 -74.445 53.914
246359 94.676 -74.449 53.871
246360 94.696 -74.461 53.829
246361 94.715 -74.473 53.787
246362 94.736 -74.485 53.746
246363 94.755 -74.498 53.710
246364 94.775 -74.510 53.677
246365 94.794 -74.522 53.645
246366 94.814 -74.534 53.614
246367 94.838 -74.537 53.583
246368 94.861 -74.526 53.553
246369 94.885 -74.515 53.523
246370 94.908 -74.504 53.489
246371 94.931 -74.493 53.454
246372 94.955 -74.483 53.419
246373 94.978 -74.472 53.383
246374 95.001 -74.461 53.343
246375 95.031 -74.465 53.298
246376 95.064 -74.481 53.249
246377 95.099 -74.497 53.196
246378 95.132 -74.513 53.136
246379 95.167 -74.530 53.072
246380 95.201 -74.546 53.005
246381 95.235 -74.562 52.937
246382 95.269 -74.577 52.870
246383 95.303 -74.593 52.806
246384 95.337 -74.610 52.743
246385 95.361 -74.631 52.683
246386 95.386 -74.653 52.626
246387 95.412 -74.674 52.570
246388 95.439 -74.679 52.515
246389 95.464 -74.678 52.461
246390 95.488 -74.678 52.407
246391 95.514 -74.676 52.354
246392 95.538 -74.675 52.303
246393 95.564 -74.674 52.253
246394 95.590 -74.671 52.206
246395 95.617 -74.668 52.161
246396 95.643 -74.666 52.117
246397 95.669 -74.664 52.072
246398 95.695 -74.661 52.028
246399 95.722 -74.659 51.986
246400 95.748 -74.656 51.946
246401 95.774 -74.654 51.906
246402 95.801 -74.652 51.866
246403 95.827 -74.654 51.826
246404 95.852 -74.673 51.785
246405 95.883 -74.686 51.745
246406 95.920 -74.690 51.706
246407 95.957 -74.695 51.668
246408 95.994 -74.700 51.630
246409 96.032 -74.705 51.592
246410 96.069 -74.709 51.554
246411 96.106 -74.714 51.518
246412 96.144 -74.718 51.482
246413 96.181 -74.723 51.448
246414 96.219 -74.728 51.417
246415 96.256 -74.733 51.386
246416 96.264 -74.743 51.358
246417 96.258 -74.756 51.331
246418 96.274 -74.766 51.307
246419 96.300 -74.777 51.285
246420 96.326 -74.787 51.265
246421 96.351 -74.797 51.247
246422 96.377 -74.806 51.231
246423 96.463 -74.768 51.217
246424 96.489 -74.775 51.203
246425 96.515 -74.782 51.192
246426 96.541 -74.789 51.184
246427 96.567 -74.796 51.177
246428 96.593 -74.803 51.171
246429 96.619 -74.810 51.166
246430 96.645 -74.817 51.161
246431 96.671 -74.824 51.154
246432 96.698 -74.832 51.145
246433 96.726 -74.843 51.137
246434 96.755 -74.853 51.125
246435 96.783 -74.864 51.109
246436 96.811 -74.875 51.091
246437 96.852 -74.894 51.073
246438 96.899 -74.919 51.054
246439 96.920 -74.931 51.037
246440 96.926 -74.933 51.020
246441 96.933 -74.936 51.003
246442 96.939 -74.939 50.986
246443 96.946 -74.941 50.971
246444 96.952 -74.944 50.959
246445 96.959 -74.947 50.945
246446 96.972 -74.957 50.929
246447 96.985 -74.967 50.912
246448 96.993 -74.981 50.893
246449 97.004 -74.991 50.876
246450 97.029 -74.991 50.858
246451 97.052 -74.994 50.842
246452 97.076 -74.996 50.827
246453 97.099 -74.999 50.814
246454 97.122 -75.002 50.801
246455 97.146 -75.005 50.788
246456 97.167 -75.013 50.774
246457 97.189 -75.019 50.756
246458 97.211 -75.027 50.735
246459 97.232 -75.035 50.712
246460 97.256 -75.044 50.688
246461 97.281 -75.053 50.662
246462 97.305 -75.062 50.635
246463 97.329 -75.071 50.606
246464 97.354 -75.080 50.573
246465 97.378 -75.090 50.537
246466 97.401 -75.098 50.500
246467 97.418 -75.107 50.461
246468 97.446 -75.113 50.424
246469 97.476 -75.120 50.385
246470 97.507 -75.126 50.348
246471 97.538 -75.133 50.312
246472 97.572 -75.140 50.276
246473 97.605 -75.147 50.240
246474 97.639 -75.154 50.200
246475 97.673 -75.161 50.159
246476 97.709 -75.168 50.116
246477 97.744 -75.176 50.073
246478 97.780 -75.182 50.029
246479 97.815 -75.189 49.984
246480 97.850 -75.197 49.935
246481 97.884 -75.204 49.889
246482 97.918 -75.211 49.845
246483 97.967 -75.222 49.800
246484 98.001 -75.233 49.752
246485 98.034 -75.244 49.700
246486 98.068 -75.254 49.647
246487 98.101 -75.266 49.593
246488 98.135 -75.276 49.541
246489 98.168 -75.287 49.492
246490 98.202 -75.298 49.443
246491 98.228 -75.308 49.395
246492 98.241 -75.318 49.349
246493 98.255 -75.327 49.300
246494 98.268 -75.336 49.251
246495 98.300 -75.341 49.202
246496 98.349 -75.340 49.152
246497 98.395 -75.352 49.103
246498 98.432 -75.364 49.057
246499 98.452 -75.345 49.010
246500 98.472 -75.326 48.962
246501 98.504 -75.309 48.914
246502 98.549 -75.294 48.865
246503 98.573 -75.295 48.814
246504 98.561 -75.327 48.761
246505 98.550 -75.358 48.709
246506 98.575 -75.361 48.657
246507 98.603 -75.363 48.605
246508 98.631 -75.364 48.555
246509 98.658 -75.365 48.506
246510 98.686 -75.367 48.456
246511 98.715 -75.368 48.406
246512 98.754 -75.375 48.356
246513 98.793 -75.382 48.304
246514 98.833 -75.389 48.253
246515 98.858 -75.380 48.204
246516 98.880 -75.378 48.157
246517 98.898 -75.403 48.111
246518 98.915 -75.428 48.065
246519 98.932 -75.453 48.020
246520 98.949 -75.477 47.976
246521 98.967 -75.502 47.933
246522 98.986 -75.531 47.891
246523 99.009 -75.565 47.851
246524 99.031 -75.599 47.811
246525 99.054 -75.633 47.774
246526 99.077 -75.667 47.739
246527 99.099 -75.700 47.704
246528 99.122 -75.734 47.670
246529 99.157 -75.728 47.636
246530 99.194 -75.715 47.603
246531 99.231 -75.701 47.569
246532 99.269 -75.688 47.539
246533 99.305 -75.674 47.508
246534 99.340 -75.668 47.478
246535 99.365 -75.683 47.450
246536 99.389 -75.699 47.422
246537 99.413 -75.715 47.394
246538 99.437 -75.732 47.367
246539 99.461 -75.749 47.340
246540 99.485 -75.765 47.316
246541 99.509 -75.781 47.292
246542 99.533 -75.798 47.268
246543 99.557 -75.815 47.245
246544 99.581 -75.831 47.223
246545 99.606 -75.848 47.201
246546 99.632 -75.865 47.179
246547 99.658 -75.883 47.160
246548 99.685 -75.900 47.141
246549 99.708 -75.926 47.120
246550 99.729 -75.956 47.096
246551 99.752 -75.985 47.071
246552 99.792 -75.984 47.046
246553 99.833 -75.984 47.019
246554 99.873 -75.983 46.991
246555 99.906 -75.987 46.963
246556 99.933 -75.996 46.934
246557 99.957 -76.009 46.904
246558 99.981 -76.022 46.874
246559 100.005 -76.035 46.841
246560 100.030 -76.047 46.811
246561 100.054 -76.060 46.782
246562 100.078 -76.072 46.751
246563 100.100 -76.089 46.720
246564 100.119 -76.112 46.689
246565 100.138 -76.135 46.656
246566 100.168 -76.135 46.623
246567 100.202 -76.125 46.592
246568 100.237 -76.114 46.562
246569 100.271 -76.104 46.531
246570 100.305 -76.093 46.499
246571 100.340 -76.082 46.467
246572 100.367 -76.077 46.435
246573 100.393 -76.074 46.403
246574 100.433 -76.076 46.372
246575 100.458 -76.105 46.340
246576 100.476 -76.146 46.309
246577 100.496 -76.183 46.276
246578 100.519 -76.216 46.243
246579 100.543 -76.249 46.210
246580 100.566 -76.282 46.177
246581 100.588 -76.316 46.143
246582 100.612 -76.349 46.111
246583 100.635 -76.383 46.079
246584 100.657 -76.392 46.046
246585 100.680 -76.398 46.014
246586 100.702 -76.405 45.982
246587 100.724 -76.411 45.952
246588 100.737 -76.437 45.923
246589 100.745 -76.470 45.894
246590 100.754 -76.504 45.864
246591 100.762 -76.537 45.836
246592 100.770 -76.571 45.809
246593 100.795 -76.578 45.783
246594 100.841 -76.580 45.758
246595 100.885 -76.575 45.734
246596 100.930 -76.570 45.710
246597 100.975 -76.564 45.687
246598 101.019 -76.558 45.664
246599 101.064 -76.552 45.641
246600 101.109 -76.547 45.618
246601 101.154 -76.542 45.596
246602 101.198 -76.536 45.574
246603 101.243 -76.530 45.551
246604 101.288 -76.524 45.529
246605 101.333 -76.519 45.506
246606 101.368 -76.520 45.484
246607 101.392 -76.530 45.462
246608 101.416 -76.539 45.441
246609 101.440 -76.548 45.420
246610 101.468 -76.558 45.400
246611 101.496 -76.569 45.379
246612 101.511 -76.581 45.358
246613 101.526 -76.592 45.339
246614 101.540 -76.605 45.319
246615 101.555 -76.617 45.300
246616 101.570 -76.629 45.281
246617 101.585 -76.641 45.263
246618 101.599 -76.653 45.245
246619 101.614 -76.665 45.226
246620 101.629 -76.677 45.208
246621 101.643 -76.689 45.192
246622 101.661 -76.697 45.175
246623 101.682 -76.697 45.157
246624 101.704 -76.697 45.140
246625 101.726 -76.697 45.123
246626 101.747 -76.698 45.106
246627 101.763 -76.708 45.088
246628 101.779 -76.720 45.071
246629 101.796 -76.731 45.054
246630 101.812 -76.742 45.038
246631 101.829 -76.753 45.021
246632 101.845 -76.765 45.004
246633 101.862 -76.776 44.985
246634 101.878 -76.787 44.968
246635 101.894 -76.798 44.952
246636 101.910 -76.809 44.935
246637 101.927 -76.821 44.916
246638 101.937 -76.827 44.900
246639 101.929 -76.819 44.883
246640 101.929 -76.817 44.866
246641 101.937 -76.820 44.850
246642 101.944 -76.822 44.833
246643 101.952 -76.825 44.815
246644 101.960 -76.828 44.798
246645 101.967 -76.831 44.781
246646 101.977 -76.836 44.764
246647 101.999 -76.856 44.747
246648 102.021 -76.875 44.731
246649 102.043 -76.895 44.716
246650 102.064 -76.915 44.699
246651 102.087 -76.934 44.683
246652 102.108 -76.954 44.667
246653 102.130 -76.973 44.649
246654 102.152 -76.993 44.633
246655 102.174 -77.012 44.617
246656 102.196 -77.032 44.601
246657 102.217 -77.047 44.585
246658 102.234 -77.047 44.568
246659 102.251 -77.047 44.552
246660 102.269 -77.048 44.535
246661 102.286 -77.049 44.520
246662 102.309 -77.050 44.504
246663 102.332 -77.050 44.488
246664 102.355 -77.052 44.472
246665 102.378 -77.053 44.457
246666 102.401 -77.053 44.441
246667 102.424 -77.055 44.425
246668 102.447 -77.056 44.411
246669 102.470 -77.056 44.395
246670 102.493 -77.057 44.380
246671 102.516 -77.059 44.364
246672 102.539 -77.059 44.349
246673 102.562 -77.060 44.332
246674 102.585 -77.062 44.317
246675 102.608 -77.063 44.301
246676 102.631 -77.063 44.286
246677 102.654 -77.065 44.270
246678 102.677 -77.066 44.254
246679 102.697 -77.067 44.238
246975 92.629 -74.377 54.538
246976 92.655 -74.367 54.542
246977 92.683 -74.358 54.545
246978 92.713 -74.350 54.548
246979 92.743 -74.342 54.550
246980 92.774 -74.333 54.553
246981 92.803 -74.349 54.554
246982 92.833 -74.367 54.554
246983 92.863 -74.386 54.554
246984 92.894 -74.389 54.554
246985 92.925 -74.390 54.552
246986 92.956 -74.392 54.551
246987 92.987 -74.393 54.550
246988 93.018 -74.394 54.549
246989 93.049 -74.395 54.548
246990 93.080 -74.396 54.548
246991 93.111 -74.397 54.549
246992 93.142 -74.399 54.551
246993 93.174 -74.399 54.556
246994 93.204 -74.400 54.561
246995 93.235 -74.402 54.570
246996 93.267 -74.402 54.581
246997 93.297 -74.404 54.597
246998 93.330 -74.406 54.615
246999 93.363 -74.412 54.634
247000 93.390 -74.408 54.655
247001 93.413 -74.398 54.679
247002 93.435 -74.393 54.704
247003 93.456 -74.388 54.730
247004 93.478 -74.384 54.755
247005 93.499 -74.380 54.777
247006 93.520 -74.376 54.794
247007 93.542 -74.371 54.807
247008 93.563 -74.367 54.817
247009 93.585 -74.363 54.820
247010 93.606 -74.358 54.826
247011 93.628 -74.355 54.836
247012 93.649 -74.350 54.851
247013 93.671 -74.346 54.860
247014 93.692 -74.342 54.863
247015 93.713 -74.337 54.860
247016 93.734 -74.333 54.853
247017 93.756 -74.329 54.850
247018 93.777 -74.325 54.851
247019 93.799 -74.320 54.855
247020 93.820 -74.316 54.854
247021 93.842 -74.312 54.844
247022 93.872 -74.322 54.830
247023 93.904 -74.335 54.813
247024 93.926 -74.338 54.795
247025 93.948 -74.340 54.779
247026 93.971 -74.342 54.762
247027 93.993 -74.345 54.744
247028 94.014 -74.347 54.727
247029 94.037 -74.349 54.708
247030 94.059 -74.352 54.685
247031 94.081 -74.354 54.662
247032 94.102 -74.355 54.642
247033 94.124 -74.355 54.622
247034 94.148 -74.358 54.603
247035 94.172 -74.359 54.585
247036 94.194 -74.361 54.566
247037 94.214 -74.361 54.548
247038 94.234 -74.361 54.530
247039 94.254 -74.361 54.513
247040 94.274 -74.361 54.493
247041 94.295 -74.361 54.473
247042 94.316 -74.361 54.454
247043 94.337 -74.362 54.435
247044 94.357 -74.362 54.415
247045 94.377 -74.364 54.394
247046 94.398 -74.368 54.371
247047 94.420 -74.374 54.345
247048 94.444 -74.372 54.317
247049 94.469 -74.370 54.287
247050 94.493 -74.368 54.255
247051 94.518 -74.367 54.224
247052 94.542 -74.365 54.192
247053 94.565 -74.364 54.158
247054 94.582 -74.374 54.122
247055 94.599 -74.384 54.084
247056 94.616 -74.394 54.043
247057 94.632 -74.404 54.001
247058 94.648 -74.408 53.959
247059 94.663 -74.408 53.916
247060 94.678 -74.407 53.871
247061 94.698 -74.418 53.826
247062 94.717 -74.431 53.781
247063 94.737 -74.443 53.738
247064 94.757 -74.455 53.698
247065 94.777 -74.468 53.662
247066 94.796 -74.480 53.628
247067 94.815 -74.495 53.596
247068 94.826 -74.530 53.568
247069 94.852 -74.527 53.539
247070 94.892 -74.490 53.509
247071 94.915 -74.479 53.476
247072 94.938 -74.468 53.440
247073 94.962 -74.457 53.404
247074 94.985 -74.446 53.369
247075 95.008 -74.436 53.330
247076 95.041 -74.448 53.286
247077 95.074 -74.464 53.238
247078 95.109 -74.480 53.186
247079 95.143 -74.496 53.129
247080 95.177 -74.512 53.067
247081 95.211 -74.528 53.002
247082 95.245 -74.544 52.936
247083 95.279 -74.560 52.870
247084 95.313 -74.577 52.806
247085 95.347 -74.593 52.744
247086 95.378 -74.611 52.684
247087 95.402 -74.633 52.627
247088 95.424 -74.645 52.572
247089 95.451 -74.647 52.517
247090 95.479 -74.647 52.463
247091 95.507 -74.646 52.410
247092 95.533 -74.643 52.357
247093 95.559 -74.640 52.306
247094 95.586 -74.638 52.258
247095 95.612 -74.636 52.212
247096 95.638 -74.634 52.167
247097 95.665 -74.631 52.123
247098 95.691 -74.628 52.078
247099 95.717 -74.626 52.035
247100 95.744 -74.624 51.993
247101 95.770 -74.621 51.953
247102 95.796 -74.618 51.913
247103 95.822 -74.616 51.873
247104 95.848 -74.618 51.833
247105 95.874 -74.637 51.792
247106 95.899 -74.657 51.751
247107 95.925 -74.676 51.711
247108 95.958 -74.686 51.673
247109 95.995 -74.690 51.634
247110 96.033 -74.695 51.596
247111 96.070 -74.700 51.558
247112 96.107 -74.704 51.521
247113 96.145 -74.709 51.484
247114 96.182 -74.714 51.450
247115 96.219 -74.718 51.418
247116 96.256 -74.723 51.387
247117 96.263 -74.734 51.357
247118 96.257 -74.747 51.329
247119 96.257 -74.759 51.304
247120 96.283 -74.769 51.281
247121 96.308 -74.779 51.260
247122 96.334 -74.789 51.242
247123 96.370 -74.791 51.225
247124 96.466 -74.743 51.210
247125 96.493 -74.750 51.196
247126 96.518 -74.756 51.183
247127 96.544 -74.763 51.175
247128 96.571 -74.771 51.167
247129 96.596 -74.778 51.161
247130 96.622 -74.784 51.156
247131 96.649 -74.791 51.151
247132 96.675 -74.798 51.146
247133 96.700 -74.806 51.139
247134 96.727 -74.812 51.133
247135 96.753 -74.819 51.123
247136 96.781 -74.830 51.110
247137 96.809 -74.841 51.094
247138 96.846 -74.856 51.076
247139 96.888 -74.881 51.059
247140 96.926 -74.907 51.042
247141 96.932 -74.910 51.026
247142 96.938 -74.913 51.009
247143 96.945 -74.916 50.992
247144 96.951 -74.918 50.977
247145 96.958 -74.921 50.963
247146 96.966 -74.925 50.949
247147 96.979 -74.935 50.933
247148 96.992 -74.945 50.917
247149 97.003 -74.956 50.898
247150 97.011 -74.969 50.882
247151 97.019 -74.982 50.865
247152 97.040 -74.984 50.851
247153 97.065 -74.985 50.839
247154 97.088 -74.988 50.828
247155 97.112 -74.990 50.817
247156 97.135 -74.993 50.805
247157 97.159 -74.996 50.791
247158 97.182 -74.999 50.772
247159 97.206 -75.001 50.751
247160 97.229 -75.006 50.728
247161 97.253 -75.015 50.704
247162 97.278 -75.024 50.678
247163 97.302 -75.033 50.651
247164 97.326 -75.042 50.621
247165 97.351 -75.051 50.589
247166 97.375 -75.060 50.555
247167 97.399 -75.069 50.520
247168 97.420 -75.079 50.484
247169 97.449 -75.085 50.447
247170 97.480 -75.091 50.407
247171 97.514 -75.098 50.369
247172 97.548 -75.106 50.332
247173 97.582 -75.113 50.295
247174 97.616 -75.120 50.258
247175 97.650 -75.126 50.219
247176 97.683 -75.134 50.178
247177 97.718 -75.141 50.135
247178 97.753 -75.148 50.091
247179 97.789 -75.155 50.047
247180 97.824 -75.162 50.002
247181 97.860 -75.169 49.952
247182 97.895 -75.176 49.907
247183 97.931 -75.183 49.863
247184 97.962 -75.192 49.818
247185 97.994 -75.202 49.769
247186 98.027 -75.214 49.716
247187 98.061 -75.224 49.662
247188 98.095 -75.236 49.607
247189 98.128 -75.246 49.554
247190 98.162 -75.257 49.504
247191 98.196 -75.268 49.454
247192 98.229 -75.279 49.405
247193 98.256 -75.289 49.358
247194 98.270 -75.298 49.310
247195 98.283 -75.308 49.260
247196 98.313 -75.312 49.210
247197 98.362 -75.311 49.160
247198 98.405 -75.334 49.111
247199 98.439 -75.341 49.064
247200 98.458 -75.322 49.018
247201 98.479 -75.303 48.970
247202 98.498 -75.284 48.922
247203 98.541 -75.268 48.872
247204 98.573 -75.264 48.821
247205 98.561 -75.296 48.767
247206 98.562 -75.318 48.715
247207 98.590 -75.319 48.663
247208 98.618 -75.320 48.610
247209 98.645 -75.322 48.561
247210 98.673 -75.323 48.510
247211 98.701 -75.324 48.459
247212 98.728 -75.326 48.409
247213 98.756 -75.326 48.358
247214 98.789 -75.330 48.307
247215 98.829 -75.338 48.256
247216 98.861 -75.337 48.207
247217 98.882 -75.339 48.161
247218 98.900 -75.364 48.115
247219 98.917 -75.389 48.069
247220 98.934 -75.414 48.024
247221 98.955 -75.444 47.978
247222 98.978 -75.478 47.934
247223 99.000 -75.512 47.891
247224 99.023 -75.546 47.850
247225 99.046 -75.580 47.809
247226 99.068 -75.614 47.770
247227 99.091 -75.648 47.732
247228 99.114 -75.682 47.697
247229 99.146 -75.683 47.662
247230 99.184 -75.669 47.628
247231 99.221 -75.656 47.593
247232 99.258 -75.643 47.560
247233 99.295 -75.630 47.528
247234 99.327 -75.627 47.497
247235 99.354 -75.636 47.468
247236 99.377 -75.652 47.439
247237 99.402 -75.668 47.411
247238 99.426 -75.685 47.383
247239 99.450 -75.702 47.356
247240 99.474 -75.718 47.330
247241 99.498 -75.734 47.306
247242 99.522 -75.751 47.283
247243 99.546 -75.768 47.259
247244 99.570 -75.784 47.235
247245 99.594 -75.801 47.214
247246 99.621 -75.818 47.192
247247 99.647 -75.835 47.171
247248 99.673 -75.853 47.152
247249 99.699 -75.870 47.134
247250 99.722 -75.895 47.114
247251 99.745 -75.925 47.091
247252 99.767 -75.954 47.066
247253 99.794 -75.975 47.040
247254 99.835 -75.974 47.013
247255 99.875 -75.974 46.986
247256 99.915 -75.973 46.959
247257 99.947 -75.981 46.930
247258 99.973 -75.994 46.900
247259 99.997 -76.007 46.870
247260 100.021 -76.019 46.838
247261 100.045 -76.032 46.807
247262 100.069 -76.044 46.778
247263 100.093 -76.057 46.747
247264 100.118 -76.069 46.715
247265 100.139 -76.088 46.682
247266 100.158 -76.110 46.650
247267 100.184 -76.119 46.617
247268 100.218 -76.109 46.586
247269 100.253 -76.098 46.555
247270 100.287 -76.088 46.524
247271 100.321 -76.077 46.493
247272 100.356 -76.066 46.460
247273 100.390 -76.056 46.428
247274 100.418 -76.050 46.396
247275 100.454 -76.053 46.364
247276 100.472 -76.094 46.333
247277 100.494 -76.127 46.301
247278 100.517 -76.160 46.268
247279 100.541 -76.194 46.235
247280 100.564 -76.227 46.202
247281 100.587 -76.260 46.169
247282 100.610 -76.294 46.135
247283 100.633 -76.327 46.103
247284 100.656 -76.360 46.071
247285 100.679 -76.383 46.039
247286 100.701 -76.389 46.007
247287 100.724 -76.396 45.975
247288 100.746 -76.403 45.945
247289 100.756 -76.433 45.916
247290 100.764 -76.467 45.888
247291 100.772 -76.500 45.859
247292 100.781 -76.533 45.832
247293 100.787 -76.538 45.805
247294 100.821 -76.547 45.779
247295 100.862 -76.558 45.753
247296 100.911 -76.567 45.729
247297 100.956 -76.561 45.706
247298 101.000 -76.555 45.682
247299 101.045 -76.550 45.660
247300 101.090 -76.544 45.636
247301 101.135 -76.539 45.613
247302 101.179 -76.533 45.590
247303 101.224 -76.527 45.568
247304 101.269 -76.522 45.545
247305 101.314 -76.516 45.522
247306 101.358 -76.511 45.500
247307 101.399 -76.508 45.478
247308 101.422 -76.517 45.456
247309 101.447 -76.527 45.435
247310 101.471 -76.536 45.414
247311 101.498 -76.546 45.394
247312 101.525 -76.557 45.373
247313 101.539 -76.569 45.353
247314 101.554 -76.581 45.333
247315 101.568 -76.593 45.313
247316 101.584 -76.605 45.294
247317 101.598 -76.617 45.274
247318 101.613 -76.629 45.256
247319 101.627 -76.641 45.237
247320 101.642 -76.653 45.219
247321 101.657 -76.665 45.200
247322 101.675 -76.672 45.183
247323 101.696 -76.672 45.166
247324 101.718 -76.672 45.148
247325 101.740 -76.672 45.130
247326 101.759 -76.677 45.113
247327 101.775 -76.689 45.095
247328 101.792 -76.700 45.078
247329 101.808 -76.711 45.061
247330 101.824 -76.722 45.044
247331 101.840 -76.734 45.027
247332 101.857 -76.745 45.010
247333 101.873 -76.755 44.992
247334 101.890 -76.767 44.974
247335 101.906 -76.778 44.957
247336 101.922 -76.789 44.940
247337 101.923 -76.788 44.923
247338 101.916 -76.781 44.905
247339 101.917 -76.780 44.888
247340 101.925 -76.783 44.871
247341 101.933 -76.786 44.855
247342 101.941 -76.788 44.838
247343 101.948 -76.791 44.821
247344 101.956 -76.793 44.804
247345 101.963 -76.796 44.787
247346 101.977 -76.807 44.770
247347 101.999 -76.826 44.753
247348 102.021 -76.846 44.737
247349 102.043 -76.865 44.720
247350 102.064 -76.885 44.704
247351 102.087 -76.905 44.688
247352 102.108 -76.925 44.672
247353 102.130 -76.944 44.655
247354 102.152 -76.964 44.638
247355 102.174 -76.983 44.622
247356 102.196 -77.003 44.606
247357 102.214 -77.006 44.589
247358 102.231 -77.007 44.573
247359 102.249 -77.007 44.556
247360 102.266 -77.008 44.540
247361 102.283 -77.009 44.524
247362 102.306 -77.009 44.508
247363 102.329 -77.010 44.492
247364 102.352 -77.012 44.477
247365 102.375 -77.012 44.461
247366 102.398 -77.013 44.445
247367 102.421 -77.014 44.429
247368 102.444 -77.016 44.413
247369 102.467 -77.016 44.398
247370 102.490 -77.017 44.382
247371 102.513 -77.019 44.367
247372 102.536 -77.019 44.352
247373 102.559 -77.020 44.337
247374 102.582 -77.022 44.321
247375 102.605 -77.022 44.305
247376 102.628 -77.023 44.289
247377 102.651 -77.024 44.274
247378 102.674 -77.025 44.258
247678 92.684 -74.331 54.553
247679 92.714 -74.323 54.555
247680 92.744 -74.314 54.557
247681 92.775 -74.306 54.559
247682 92.805 -74.322 54.561
247683 92.834 -74.341 54.561
247684 92.864 -74.359 54.561
247685 92.895 -74.368 54.561
247686 92.926 -74.369 54.559
247687 92.956 -74.371 54.558
247688 92.988 -74.371 54.557
247689 93.019 -74.372 54.555
247690 93.050 -74.374 54.554
247691 93.081 -74.374 54.553
247692 93.112 -74.376 54.553
247693 93.143 -74.377 54.555
247694 93.174 -74.378 54.557
247695 93.205 -74.379 54.561
247696 93.236 -74.380 54.567
247697 93.267 -74.381 54.577
247698 93.298 -74.382 54.592
247699 93.330 -74.383 54.610
247700 93.361 -74.384 54.630
247701 93.393 -74.387 54.652
247702 93.415 -74.380 54.676
247703 93.436 -74.376 54.701
247704 93.458 -74.371 54.726
247705 93.479 -74.368 54.751
247706 93.500 -74.363 54.775
247707 93.522 -74.359 54.797
247708 93.543 -74.355 54.815
247709 93.565 -74.350 54.825
247710 93.586 -74.346 54.832
247711 93.608 -74.342 54.842
247712 93.629 -74.338 54.856
247713 93.650 -74.333 54.877
247714 93.672 -74.329 54.891
247715 93.693 -74.325 54.898
247716 93.714 -74.320 54.894
247717 93.736 -74.317 54.884
247718 93.757 -74.312 54.876
247719 93.779 -74.308 54.870
247720 93.800 -74.304 54.868
247721 93.822 -74.299 54.866
247722 93.843 -74.295 54.859
247723 93.864 -74.292 54.846
247724 93.896 -74.304 54.831
247725 93.920 -74.308 54.809
247726 93.942 -74.311 54.787
247727 93.965 -74.313 54.764
247728 93.986 -74.315 54.741
247729 94.008 -74.317 54.721
247730 94.031 -74.320 54.701
247731 94.053 -74.322 54.679
247732 94.076 -74.325 54.655
247733 94.101 -74.329 54.635
247734 94.123 -74.330 54.614
247735 94.146 -74.331 54.594
247736 94.170 -74.333 54.577
247737 94.194 -74.335 54.560
247738 94.215 -74.336 54.544
247739 94.235 -74.336 54.528
247740 94.255 -74.336 54.512
247741 94.275 -74.336 54.493
247742 94.294 -74.336 54.475
247743 94.314 -74.336 54.456
247744 94.335 -74.336 54.438
247745 94.356 -74.336 54.420
247746 94.376 -74.338 54.400
247747 94.398 -74.342 54.376
247748 94.422 -74.341 54.349
247749 94.447 -74.339 54.321
247750 94.471 -74.337 54.290
247751 94.495 -74.336 54.258
247752 94.520 -74.334 54.228
247753 94.544 -74.333 54.196
247754 94.568 -74.331 54.162
247755 94.586 -74.339 54.126
247756 94.603 -74.349 54.087
247757 94.620 -74.359 54.045
247758 94.636 -74.368 54.003
247759 94.652 -74.371 53.960
247760 94.667 -74.370 53.916
247761 94.682 -74.370 53.869
247762 94.700 -74.377 53.822
247763 94.719 -74.389 53.775
247764 94.739 -74.401 53.731
247765 94.759 -74.413 53.688
247766 94.779 -74.425 53.649
247767 94.799 -74.437 53.613
247768 94.814 -74.460 53.580
247769 94.824 -74.496 53.552
247770 94.835 -74.533 53.526
247771 94.866 -74.517 53.497
247772 94.908 -74.476 53.463
247773 94.945 -74.443 53.427
247774 94.969 -74.432 53.391
247775 94.992 -74.421 53.354
247776 95.017 -74.415 53.314
247777 95.051 -74.430 53.271
247778 95.085 -74.446 53.224
247779 95.119 -74.462 53.174
247780 95.153 -74.479 53.117
247781 95.187 -74.495 53.057
247782 95.221 -74.511 52.995
247783 95.255 -74.527 52.931
247784 95.289 -74.543 52.866
247785 95.321 -74.557 52.803
247786 95.354 -74.571 52.742
247787 95.388 -74.580 52.682
247788 95.413 -74.587 52.626
247789 95.435 -74.599 52.571
247790 95.461 -74.603 52.517
247791 95.489 -74.603 52.463
247792 95.519 -74.603 52.410
247793 95.547 -74.602 52.358
247794 95.576 -74.601 52.308
247795 95.605 -74.601 52.260
247796 95.634 -74.600 52.216
247797 95.660 -74.598 52.172
247798 95.686 -74.596 52.128
247799 95.712 -74.593 52.084
247800 95.739 -74.590 52.042
247801 95.765 -74.588 52.000
247802 95.792 -74.586 51.960
247803 95.818 -74.584 51.921
247804 95.844 -74.581 51.880
247805 95.870 -74.584 51.840
247806 95.896 -74.603 51.799
247807 95.921 -74.621 51.757
247808 95.947 -74.640 51.717
247809 95.972 -74.659 51.677
247810 95.997 -74.678 51.638
247811 96.033 -74.686 51.600
247812 96.071 -74.690 51.562
247813 96.108 -74.695 51.524
247814 96.145 -74.700 51.486
247815 96.182 -74.704 51.452
247816 96.220 -74.709 51.419
247817 96.257 -74.713 51.387
247818 96.261 -74.725 51.356
247819 96.256 -74.737 51.328
247820 96.250 -74.750 51.301
247821 96.265 -74.762 51.278
247822 96.291 -74.772 51.256
247823 96.317 -74.781 51.237
247824 96.444 -74.710 51.219
247825 96.470 -74.717 51.202
247826 96.496 -74.724 51.188
247827 96.522 -74.731 51.175
247828 96.548 -74.738 51.165
247829 96.574 -74.745 51.157
247830 96.600 -74.752 51.151
247831 96.626 -74.759 51.144
247832 96.652 -74.766 51.140
247833 96.678 -74.773 51.136
247834 96.704 -74.780 51.132
247835 96.730 -74.787 51.126
247836 96.756 -74.794 51.118
247837 96.782 -74.801 51.107
247838 96.808 -74.808 51.094
247839 96.838 -74.819 51.080
247840 96.876 -74.844 51.063
247841 96.913 -74.870 51.047
247842 96.938 -74.887 51.031
247843 96.944 -74.889 51.015
247844 96.951 -74.892 50.999
247845 96.957 -74.895 50.983
247846 96.963 -74.897 50.968
247847 96.973 -74.903 50.953
247848 96.986 -74.913 50.937
247849 96.999 -74.923 50.921
247850 97.012 -74.932 50.904
247851 97.021 -74.945 50.888
247852 97.029 -74.958 50.873
247853 97.037 -74.971 50.861
247854 97.052 -74.978 50.851
247855 97.077 -74.977 50.842
247856 97.101 -74.979 50.832
247857 97.124 -74.982 50.820
247858 97.148 -74.984 50.807
247859 97.171 -74.987 50.788
247860 97.198 -74.986 50.767
247861 97.229 -74.982 50.744
247862 97.253 -74.991 50.720
247863 97.277 -75.000 50.694
247864 97.301 -75.008 50.666
247865 97.325 -75.017 50.636
247866 97.349 -75.026 50.605
247867 97.373 -75.034 50.573
247868 97.397 -75.043 50.540
247869 97.425 -75.050 50.506
247870 97.457 -75.057 50.468
247871 97.491 -75.064 50.429
247872 97.525 -75.071 50.389
247873 97.559 -75.078 50.350
247874 97.592 -75.085 50.312
247875 97.626 -75.092 50.274
247876 97.660 -75.099 50.235
247877 97.694 -75.106 50.196
247878 97.728 -75.113 50.154
247879 97.762 -75.120 50.110
247880 97.798 -75.127 50.065
247881 97.833 -75.135 50.019
247882 97.869 -75.142 49.970
247883 97.904 -75.148 49.925
247884 97.938 -75.156 49.880
247885 97.968 -75.166 49.832
247886 97.999 -75.176 49.783
247887 98.029 -75.185 49.729
247888 98.060 -75.195 49.675
247889 98.090 -75.205 49.621
247890 98.122 -75.216 49.566
247891 98.155 -75.226 49.514
247892 98.189 -75.238 49.463
247893 98.222 -75.248 49.414
247894 98.256 -75.259 49.366
247895 98.284 -75.270 49.317
247896 98.297 -75.279 49.267
247897 98.326 -75.284 49.216
247898 98.372 -75.294 49.167
247899 98.414 -75.317 49.118
247900 98.445 -75.318 49.071
247901 98.465 -75.299 49.024
247902 98.485 -75.280 48.977
247903 98.505 -75.261 48.928
247904 98.534 -75.243 48.878
247905 98.572 -75.234 48.827
247906 98.561 -75.265 48.773
247907 98.576 -75.276 48.721
247908 98.604 -75.276 48.669
247909 98.632 -75.278 48.616
247910 98.659 -75.279 48.566
247911 98.687 -75.280 48.515
247912 98.715 -75.282 48.463
247913 98.742 -75.283 48.411
247914 98.770 -75.284 48.360
247915 98.798 -75.286 48.309
247916 98.826 -75.286 48.258
247917 98.864 -75.293 48.209
247918 98.884 -75.301 48.162
247919 98.902 -75.326 48.116
247920 98.924 -75.358 48.070
247921 98.946 -75.392 48.023
247922 98.969 -75.426 47.978
247923 98.992 -75.460 47.934
247924 99.014 -75.493 47.891
247925 99.037 -75.527 47.848
247926 99.060 -75.561 47.806
247927 99.082 -75.596 47.766
247928 99.106 -75.629 47.727
247929 99.136 -75.638 47.690
247930 99.173 -75.624 47.654
247931 99.210 -75.611 47.619
247932 99.247 -75.598 47.584
247933 99.281 -75.592 47.551
247934 99.312 -75.593 47.518
247935 99.343 -75.596 47.487
247936 99.368 -75.605 47.458
247937 99.391 -75.621 47.429
247938 99.415 -75.638 47.401
247939 99.439 -75.655 47.373
247940 99.463 -75.671 47.346
247941 99.487 -75.687 47.320
247942 99.511 -75.704 47.297
247943 99.535 -75.721 47.273
247944 99.559 -75.737 47.250
247945 99.583 -75.753 47.227
247946 99.609 -75.771 47.205
247947 99.635 -75.788 47.183
247948 99.661 -75.806 47.162
247949 99.687 -75.823 47.144
247950 99.713 -75.840 47.126
247951 99.737 -75.864 47.106
247952 99.759 -75.894 47.084
247953 99.781 -75.923 47.060
247954 99.803 -75.953 47.034
247955 99.836 -75.965 47.008
247956 99.876 -75.964 46.982
247957 99.917 -75.964 46.955
247958 99.947 -75.974 46.926
247959 99.976 -75.984 46.897
247960 100.006 -75.994 46.866
247961 100.035 -76.004 46.834
247962 100.061 -76.016 46.804
247963 100.085 -76.029 46.773
247964 100.109 -76.042 46.742
247965 100.133 -76.054 46.709
247966 100.157 -76.067 46.676
247967 100.179 -76.086 46.643
247968 100.200 -76.104 46.612
247969 100.234 -76.094 46.580
247970 100.269 -76.083 46.549
247971 100.303 -76.072 46.517
247972 100.337 -76.062 46.486
247973 100.371 -76.051 46.453
247974 100.406 -76.041 46.420
247975 100.440 -76.030 46.389
247976 100.470 -76.038 46.358
247977 100.493 -76.072 46.325
247978 100.516 -76.105 46.293
247979 100.539 -76.138 46.261
247980 100.562 -76.172 46.228
247981 100.585 -76.205 46.194
247982 100.608 -76.238 46.161
247983 100.631 -76.271 46.129
247984 100.654 -76.304 46.096
247985 100.677 -76.338 46.064
247986 100.700 -76.371 46.032
247987 100.723 -76.381 46.000
247988 100.745 -76.388 45.969
247989 100.767 -76.395 45.939
247990 100.775 -76.429 45.910
247991 100.783 -76.462 45.882
247992 100.791 -76.495 45.854
247993 100.798 -76.502 45.827
247994 100.805 -76.505 45.800
247995 100.846 -76.515 45.774
247996 100.888 -76.527 45.749
247997 100.929 -76.538 45.725
247998 100.981 -76.552 45.701
247999 101.026 -76.547 45.677
248000 101.071 -76.542 45.654
248001 101.115 -76.536 45.631
248002 101.160 -76.530 45.608
248003 101.205 -76.524 45.585
248004 101.250 -76.519 45.563
248005 101.294 -76.514 45.539
248006 101.339 -76.508 45.516
248007 101.384 -76.502 45.494
248008 101.428 -76.496 45.473
248009 101.454 -76.505 45.451
248010 101.477 -76.514 45.429
248011 101.501 -76.524 45.409
248012 101.530 -76.534 45.389
248013 101.553 -76.546 45.368
248014 101.567 -76.557 45.348
248015 101.582 -76.569 45.328
248016 101.597 -76.581 45.308
248017 101.612 -76.593 45.287
248018 101.626 -76.605 45.267
248019 101.641 -76.618 45.249
248020 101.656 -76.630 45.229
248021 101.671 -76.641 45.211
248022 101.689 -76.648 45.192
248023 101.710 -76.648 45.173
248024 101.732 -76.648 45.157
248025 101.754 -76.648 45.139
248026 101.770 -76.658 45.120
248027 101.787 -76.669 45.102
248028 101.803 -76.680 45.085
248029 101.820 -76.691 45.068
248030 101.836 -76.702 45.051
248031 101.852 -76.713 45.034
248032 101.869 -76.724 45.016
248033 101.885 -76.736 44.999
248034 101.901 -76.747 44.981
248035 101.917 -76.757 44.963
248036 101.910 -76.750 44.946
248037 101.903 -76.743 44.929
248038 101.906 -76.743 44.912
248039 101.913 -76.746 44.895
248040 101.921 -76.749 44.878
248041 101.929 -76.752 44.860
248042 101.937 -76.754 44.843
248043 101.944 -76.757 44.827
248044 101.952 -76.759 44.810
248045 101.959 -76.762 44.793
248046 101.977 -76.777 44.776
248047 101.999 -76.797 44.759
248048 102.021 -76.816 44.742
248049 102.043 -76.836 44.726
248050 102.064 -76.856 44.709
248051 102.087 -76.875 44.693
248052 102.108 -76.895 44.677
248053 102.130 -76.915 44.660
248054 102.152 -76.934 44.643
248055 102.174 -76.954 44.626
248056 102.194 -76.965 44.610
248057 102.211 -76.966 44.594
248058 102.229 -76.966 44.577
248059 102.246 -76.967 44.561
248060 102.264 -76.968 44.545
248061 102.281 -76.969 44.528
248062 102.304 -76.969 44.511
248063 102.327 -76.970 44.497
248064 102.350 -76.971 44.481
248065 102.373 -76.972 44.465
248066 102.396 -76.973 44.450
248067 102.419 -76.974 44.433
248068 102.442 -76.975 44.417
248069 102.465 -76.976 44.402
248070 102.488 -76.977 44.387
248071 102.511 -76.978 44.371
248072 102.534 -76.979 44.356
248073 102.557 -76.980 44.341
248074 102.580 -76.981 44.325
248075 102.603 -76.982 44.309
248076 102.626 -76.983 44.293
248380 92.716 -74.295 54.562
248381 92.746 -74.287 54.564
248382 92.776 -74.280 54.566
248383 92.806 -74.296 54.568
248384 92.836 -74.314 54.569
248385 92.865 -74.333 54.569
248386 92.896 -74.346 54.568
248387 92.926 -74.348 54.567
248388 92.957 -74.349 54.566
248389 92.989 -74.350 54.563
248390 93.019 -74.351 54.562
248391 93.051 -74.352 54.561
248392 93.082 -74.353 54.560
248393 93.113 -74.355 54.560
248394 93.144 -74.355 54.560
248395 93.175 -74.356 54.560
248396 93.206 -74.358 54.563
248397 93.237 -74.358 54.568
248398 93.268 -74.360 54.576
248399 93.299 -74.361 54.590
248400 93.331 -74.362 54.607
248401 93.361 -74.363 54.628
248402 93.392 -74.361 54.649
248403 93.416 -74.360 54.673
248404 93.438 -74.356 54.699
248405 93.459 -74.352 54.725
248406 93.481 -74.349 54.749
248407 93.502 -74.345 54.774
248408 93.523 -74.341 54.799
248409 93.545 -74.337 54.822
248410 93.566 -74.333 54.831
248411 93.587 -74.329 54.842
248412 93.609 -74.325 54.857
248413 93.630 -74.321 54.876
248414 93.652 -74.317 54.899
248415 93.673 -74.312 54.916
248416 93.695 -74.308 54.925
248417 93.716 -74.304 54.923
248418 93.737 -74.300 54.911
248419 93.759 -74.295 54.900
248420 93.780 -74.291 54.890
248421 93.801 -74.287 54.883
248422 93.823 -74.283 54.878
248423 93.844 -74.279 54.871
248424 93.866 -74.274 54.861
248425 93.889 -74.273 54.848
248426 93.914 -74.279 54.827
248427 93.936 -74.281 54.801
248428 93.958 -74.283 54.773
248429 93.981 -74.286 54.745
248430 94.005 -74.290 54.721
248431 94.028 -74.293 54.698
248432 94.052 -74.297 54.674
248433 94.075 -74.301 54.649
248434 94.109 -74.307 54.628
248435 94.126 -74.306 54.607
248436 94.145 -74.305 54.588
248437 94.169 -74.307 54.574
248438 94.193 -74.309 54.559
248439 94.216 -74.311 54.545
248440 94.236 -74.311 54.531
248441 94.256 -74.311 54.516
248442 94.275 -74.311 54.499
248443 94.295 -74.311 54.481
248444 94.315 -74.311 54.463
248445 94.335 -74.311 54.443
248446 94.355 -74.311 54.425
248447 94.376 -74.311 54.404
248448 94.401 -74.310 54.379
248449 94.425 -74.308 54.352
248450 94.449 -74.307 54.323
248451 94.474 -74.305 54.292
248452 94.498 -74.303 54.261
248453 94.522 -74.302 54.231
248454 94.547 -74.300 54.199
248455 94.571 -74.299 54.165
248456 94.591 -74.304 54.128
248457 94.607 -74.314 54.087
248458 94.624 -74.324 54.046
248459 94.640 -74.333 54.003
248460 94.655 -74.333 53.959
248461 94.671 -74.333 53.914
248462 94.686 -74.333 53.866
248463 94.702 -74.334 53.817
248464 94.721 -74.346 53.769
248465 94.741 -74.358 53.725
248466 94.761 -74.371 53.679
248467 94.780 -74.383 53.636
248468 94.801 -74.395 53.598
248469 94.813 -74.425 53.565
248470 94.823 -74.461 53.538
248471 94.834 -74.498 53.513
248472 94.852 -74.514 53.486
248473 94.888 -74.487 53.452
248474 94.924 -74.461 53.415
248475 94.964 -74.425 53.378
248476 94.998 -74.396 53.341
248477 95.027 -74.397 53.300
248478 95.061 -74.413 53.256
248479 95.095 -74.429 53.209
248480 95.127 -74.443 53.160
248481 95.160 -74.457 53.104
248482 95.192 -74.471 53.046
248483 95.225 -74.484 52.986
248484 95.258 -74.499 52.924
248485 95.290 -74.512 52.861
248486 95.323 -74.526 52.798
248487 95.355 -74.540 52.737
248488 95.392 -74.542 52.677
248489 95.424 -74.542 52.622
248490 95.446 -74.553 52.568
248491 95.471 -74.559 52.515
248492 95.499 -74.559 52.461
248493 95.529 -74.558 52.408
248494 95.557 -74.558 52.358
248495 95.586 -74.557 52.310
248496 95.615 -74.556 52.263
248497 95.644 -74.556 52.219
248498 95.673 -74.555 52.175
248499 95.701 -74.555 52.132
248500 95.731 -74.555 52.089
248501 95.759 -74.554 52.047
248502 95.787 -74.553 52.006
248503 95.813 -74.550 51.966
248504 95.839 -74.548 51.928
248505 95.866 -74.546 51.887
248506 95.892 -74.549 51.846
248507 95.917 -74.568 51.805
248508 95.943 -74.587 51.763
248509 95.968 -74.606 51.722
248510 95.994 -74.624 51.682
248511 96.020 -74.643 51.642
248512 96.045 -74.662 51.603
248513 96.071 -74.681 51.565
248514 96.108 -74.685 51.526
248515 96.146 -74.690 51.488
248516 96.183 -74.695 51.453
248517 96.221 -74.700 51.419
248518 96.258 -74.704 51.387
248519 96.260 -74.715 51.355
248520 96.254 -74.728 51.326
248521 96.249 -74.741 51.299
248522 96.248 -74.754 51.274
248523 96.274 -74.764 51.252
248524 96.372 -74.716 51.231
248525 96.447 -74.684 51.213
248526 96.473 -74.692 51.195
248527 96.500 -74.699 51.179
248528 96.525 -74.706 51.166
248529 96.551 -74.712 51.156
248530 96.578 -74.720 51.147
248531 96.603 -74.727 51.139
248532 96.629 -74.734 51.134
248533 96.656 -74.740 51.128
248534 96.682 -74.747 51.123
248535 96.708 -74.755 51.120
248536 96.733 -74.762 51.116
248537 96.758 -74.769 51.110
248538 96.783 -74.776 51.101
248539 96.808 -74.784 51.091
248540 96.834 -74.794 51.079
248541 96.863 -74.808 51.065
248542 96.900 -74.833 51.050
248543 96.937 -74.859 51.035
248544 96.950 -74.866 51.019
248545 96.956 -74.869 51.005
248546 96.963 -74.872 50.990
248547 96.969 -74.874 50.974
248548 96.979 -74.881 50.958
248549 96.993 -74.891 50.941
248550 97.006 -74.900 50.925
248551 97.019 -74.910 50.909
248552 97.031 -74.920 50.894
248553 97.039 -74.934 50.881
248554 97.047 -74.947 50.870
248555 97.055 -74.960 50.862
248556 97.064 -74.972 50.854
248557 97.088 -74.971 50.845
248558 97.113 -74.970 50.834
248559 97.137 -74.972 50.821
248560 97.168 -74.966 50.802
248561 97.202 -74.957 50.782
248562 97.230 -74.960 50.760
248563 97.254 -74.968 50.735
248564 97.278 -74.977 50.708
248565 97.302 -74.985 50.679
248566 97.326 -74.994 50.649
248567 97.350 -75.003 50.620
248568 97.374 -75.011 50.589
248569 97.406 -75.016 50.556
248570 97.443 -75.019 50.521
248571 97.476 -75.026 50.485
248572 97.508 -75.033 50.446
248573 97.540 -75.041 50.406
248574 97.571 -75.050 50.367
248575 97.603 -75.057 50.328
248576 97.637 -75.065 50.291
248577 97.670 -75.072 50.253
248578 97.704 -75.079 50.213
248579 97.738 -75.085 50.174
248580 97.772 -75.092 50.130
248581 97.807 -75.100 50.085
248582 97.843 -75.107 50.038
248583 97.878 -75.114 49.987
248584 97.911 -75.121 49.939
248585 97.943 -75.129 49.891
248586 97.973 -75.138 49.842
248587 98.004 -75.148 49.790
248588 98.034 -75.158 49.737
248589 98.065 -75.169 49.685
248590 98.095 -75.179 49.633
248591 98.126 -75.189 49.577
248592 98.156 -75.198 49.523
248593 98.186 -75.208 49.471
248594 98.217 -75.218 49.421
248595 98.249 -75.229 49.372
248596 98.283 -75.240 49.322
248597 98.312 -75.250 49.272
248598 98.339 -75.256 49.222
248599 98.382 -75.276 49.171
248600 98.424 -75.299 49.123
248601 98.452 -75.295 49.076
248602 98.472 -75.276 49.029
248603 98.492 -75.257 48.981
248604 98.512 -75.238 48.932
248605 98.531 -75.219 48.882
248606 98.572 -75.203 48.831
248607 98.563 -75.232 48.778
248608 98.591 -75.233 48.726
248609 98.619 -75.234 48.673
248610 98.646 -75.236 48.621
248611 98.674 -75.237 48.570
248612 98.702 -75.238 48.518
248613 98.729 -75.239 48.466
248614 98.757 -75.241 48.414
248615 98.785 -75.242 48.361
248616 98.812 -75.243 48.310
248617 98.840 -75.245 48.260
248618 98.867 -75.248 48.210
248619 98.892 -75.271 48.163
248620 98.915 -75.305 48.116
248621 98.938 -75.339 48.069
248622 98.960 -75.373 48.022
248623 98.983 -75.407 47.976
248624 99.006 -75.441 47.932
248625 99.029 -75.475 47.889
248626 99.052 -75.509 47.845
248627 99.074 -75.543 47.803
248628 99.097 -75.577 47.761
248629 99.125 -75.593 47.721
248630 99.162 -75.580 47.684
248631 99.199 -75.566 47.646
248632 99.235 -75.555 47.610
248633 99.266 -75.558 47.575
248634 99.297 -75.560 47.541
248635 99.328 -75.561 47.508
248636 99.358 -75.564 47.477
248637 99.383 -75.574 47.448
248638 99.404 -75.591 47.419
248639 99.427 -75.608 47.391
248640 99.451 -75.624 47.364
248641 99.476 -75.640 47.337
248642 99.500 -75.657 47.311
248643 99.524 -75.674 47.287
248644 99.548 -75.690 47.264
248645 99.572 -75.706 47.241
248646 99.597 -75.724 47.218
248647 99.624 -75.741 47.196
248648 99.650 -75.758 47.175
248649 99.676 -75.775 47.153
248650 99.702 -75.793 47.135
248651 99.728 -75.810 47.116
248652 99.752 -75.833 47.097
248653 99.774 -75.862 47.077
248654 99.796 -75.892 47.054
248655 99.818 -75.922 47.028
248656 99.840 -75.951 47.003
248657 99.878 -75.955 46.978
248658 99.918 -75.956 46.950
248659 99.947 -75.966 46.921
248660 99.977 -75.976 46.892
248661 100.006 -75.986 46.861
248662 100.035 -75.997 46.830
248663 100.064 -76.007 46.799
248664 100.094 -76.017 46.768
248665 100.123 -76.027 46.736
248666 100.149 -76.039 46.703
248667 100.173 -76.052 46.670
248668 100.197 -76.065 46.637
248669 100.218 -76.085 46.605
248670 100.250 -76.079 46.574
248671 100.284 -76.068 46.542
248672 100.319 -76.057 46.510
248673 100.353 -76.047 46.478
248674 100.387 -76.036 46.445
248675 100.419 -76.027 46.413
248676 100.452 -76.017 46.382
248677 100.481 -76.025 46.351
248678 100.508 -76.055 46.318
248679 100.535 -76.085 46.286
248680 100.560 -76.116 46.254
248681 100.583 -76.150 46.221
248682 100.606 -76.182 46.187
248683 100.629 -76.216 46.155
248684 100.652 -76.249 46.122
248685 100.676 -76.282 46.089
248686 100.698 -76.316 46.057
248687 100.721 -76.349 46.025
248688 100.745 -76.373 45.993
248689 100.767 -76.379 45.962
248690 100.786 -76.392 45.934
248691 100.794 -76.425 45.905
248692 100.802 -76.458 45.877
248693 100.809 -76.467 45.850
248694 100.816 -76.469 45.822
248695 100.831 -76.473 45.795
248696 100.872 -76.484 45.770
248697 100.913 -76.495 45.745
248698 100.955 -76.507 45.720
248699 100.996 -76.518 45.696
248700 101.050 -76.535 45.673
248701 101.096 -76.533 45.649
248702 101.141 -76.527 45.626
248703 101.186 -76.522 45.603
248704 101.231 -76.516 45.580
248705 101.275 -76.511 45.557
248706 101.320 -76.505 45.534
248707 101.365 -76.499 45.511
248708 101.409 -76.494 45.489
248709 101.454 -76.488 45.468
248710 101.484 -76.493 45.445
248711 101.508 -76.502 45.424
248712 101.533 -76.511 45.404
248713 101.560 -76.522 45.384
248714 101.581 -76.533 45.362
248715 101.596 -76.546 45.342
248716 101.611 -76.558 45.322
248717 101.625 -76.570 45.301
248718 101.640 -76.581 45.281
248719 101.655 -76.593 45.261
248720 101.670 -76.605 45.242
248721 101.684 -76.618 45.222
248722 101.703 -76.623 45.203
248723 101.725 -76.623 45.185
248724 101.746 -76.623 45.167
248725 101.766 -76.627 45.148
248726 101.782 -76.637 45.130
248727 101.799 -76.649 45.112
248728 101.815 -76.660 45.093
248729 101.831 -76.671 45.075
248730 101.848 -76.682 45.058
248731 101.864 -76.693 45.041
248732 101.881 -76.705 45.023
248733 101.897 -76.716 45.005
248734 101.903 -76.719 44.988
248735 101.896 -76.712 44.971
248736 101.889 -76.705 44.952
248737 101.894 -76.706 44.936
248738 101.902 -76.709 44.919
248739 101.909 -76.712 44.901
248740 101.917 -76.715 44.884
248741 101.925 -76.718 44.867
248742 101.933 -76.720 44.849
248743 101.940 -76.723 44.832
248744 101.948 -76.725 44.816
248745 101.956 -76.728 44.799
248746 101.977 -76.748 44.782
248747 101.999 -76.768 44.765
248748 102.021 -76.787 44.748
248749 102.043 -76.807 44.731
248750 102.065 -76.826 44.714
248751 102.087 -76.846 44.698
248752 102.108 -76.865 44.682
248753 102.130 -76.885 44.666
248754 102.152 -76.905 44.649
248755 102.174 -76.925 44.632
248756 102.191 -76.925 44.615
248757 102.209 -76.926 44.599
248758 102.227 -76.926 44.582
248759 102.244 -76.927 44.566
248760 102.261 -76.928 44.549
248761 102.279 -76.928 44.533
248762 102.301 -76.929 44.517
248763 102.324 -76.930 44.500
248764 102.347 -76.931 44.485
248765 102.370 -76.932 44.470
248766 102.393 -76.933 44.454
248767 102.416 -76.934 44.438
248768 102.439 -76.935 44.422
248769 102.462 -76.936 44.406
248770 102.485 -76.937 44.390
248771 102.508 -76.937 44.376
249082 92.747 -74.261 54.571
249083 92.778 -74.252 54.573
249084 92.807 -74.270 54.575
249085 92.837 -74.289 54.576
249086 92.866 -74.308 54.576
249087 92.896 -74.325 54.576
249088 92.927 -74.327 54.575
249089 92.958 -74.327 54.573
249090 92.989 -74.328 54.571
249091 93.020 -74.330 54.570
249092 93.052 -74.330 54.568
249093 93.083 -74.332 54.567
249094 93.114 -74.333 54.566
249095 93.145 -74.334 54.566
249096 93.176 -74.335 54.566
249097 93.207 -74.336 54.568
249098 93.238 -74.337 54.572
249099 93.269 -74.338 54.578
249100 93.300 -74.339 54.590
249101 93.331 -74.340 54.607
249102 93.362 -74.340 54.626
249103 93.389 -74.337 54.648
249104 93.412 -74.337 54.672
249105 93.436 -74.333 54.698
249106 93.461 -74.329 54.726
249107 93.483 -74.325 54.754
249108 93.504 -74.321 54.780
249109 93.525 -74.317 54.802
249110 93.547 -74.314 54.822
249111 93.568 -74.309 54.836
249112 93.589 -74.305 54.852
249113 93.611 -74.302 54.872
249114 93.632 -74.298 54.895
249115 93.653 -74.294 54.918
249116 93.675 -74.290 54.935
249117 93.696 -74.286 54.945
249118 93.718 -74.283 54.946
249119 93.739 -74.278 54.935
249120 93.760 -74.274 54.922
249121 93.782 -74.270 54.910
249122 93.803 -74.267 54.901
249123 93.825 -74.263 54.891
249124 93.846 -74.259 54.882
249125 93.867 -74.255 54.873
249126 93.889 -74.252 54.860
249127 93.911 -74.252 54.840
249128 93.934 -74.255 54.813
249129 93.958 -74.258 54.782
249130 93.981 -74.262 54.749
249131 94.004 -74.265 54.723
249132 94.028 -74.269 54.696
249133 94.051 -74.273 54.670
249134 94.074 -74.276 54.644
249135 94.112 -74.283 54.621
249136 94.134 -74.284 54.601
249137 94.152 -74.283 54.584
249138 94.170 -74.282 54.571
249139 94.191 -74.283 54.558
249140 94.215 -74.285 54.547
249141 94.237 -74.286 54.535
249142 94.257 -74.286 54.521
249143 94.276 -74.286 54.504
249144 94.296 -74.286 54.486
249145 94.316 -74.286 54.467
249146 94.336 -74.286 54.446
249147 94.356 -74.289 54.426
249148 94.379 -74.288 54.404
249149 94.403 -74.277 54.380
249150 94.427 -74.276 54.352
249151 94.452 -74.274 54.323
249152 94.476 -74.273 54.292
249153 94.500 -74.270 54.261
249154 94.525 -74.269 54.231
249155 94.549 -74.267 54.199
249156 94.573 -74.266 54.164
249157 94.595 -74.269 54.126
249158 94.611 -74.279 54.084
249159 94.628 -74.289 54.041
249160 94.644 -74.296 53.997
249161 94.660 -74.295 53.953
249162 94.675 -74.295 53.907
249163 94.690 -74.295 53.858
249164 94.705 -74.295 53.808
249165 94.723 -74.304 53.761
249166 94.743 -74.316 53.715
249167 94.763 -74.328 53.668
249168 94.782 -74.340 53.624
249169 94.802 -74.354 53.584
249170 94.812 -74.390 53.552
249171 94.822 -74.427 53.525
249172 94.834 -74.461 53.501
249173 94.856 -74.471 53.476
249174 94.892 -74.445 53.441
249175 94.927 -74.418 53.405
249176 94.963 -74.393 53.367
249177 94.998 -74.366 53.330
249178 95.031 -74.371 53.287
249179 95.063 -74.384 53.242
249180 95.096 -74.399 53.194
249181 95.128 -74.412 53.145
249182 95.161 -74.426 53.092
249183 95.194 -74.440 53.034
249184 95.226 -74.454 52.975
249185 95.259 -74.468 52.915
249186 95.291 -74.482 52.854
249187 95.324 -74.496 52.791
249188 95.356 -74.509 52.730
249189 95.395 -74.503 52.671
249190 95.435 -74.497 52.617
249191 95.457 -74.507 52.564
249192 95.481 -74.515 52.511
249193 95.510 -74.515 52.458
249194 95.539 -74.514 52.406
249195 95.567 -74.513 52.357
249196 95.596 -74.513 52.310
249197 95.625 -74.512 52.265
249198 95.654 -74.512 52.221
249199 95.683 -74.512 52.178
249200 95.711 -74.511 52.136
249201 95.741 -74.510 52.094
249202 95.769 -74.510 52.052
249203 95.798 -74.509 52.011
249204 95.827 -74.509 51.971
249205 95.856 -74.508 51.933
249206 95.885 -74.508 51.893
249207 95.912 -74.513 51.852
249208 95.939 -74.532 51.811
249209 95.965 -74.551 51.769
249210 95.990 -74.570 51.727
249211 96.016 -74.590 51.686
249212 96.041 -74.609 51.646
249213 96.067 -74.628 51.606
249214 96.092 -74.646 51.567
249215 96.118 -74.665 51.528
249216 96.147 -74.681 51.490
249217 96.184 -74.685 51.454
249218 96.221 -74.690 51.420
249219 96.258 -74.694 51.387
249220 96.259 -74.706 51.355
249221 96.253 -74.719 51.325
249222 96.247 -74.732 51.297
249223 96.242 -74.745 51.271
249224 96.259 -74.754 51.248
249225 96.425 -74.653 51.226
249226 96.451 -74.659 51.207
249227 96.477 -74.666 51.190
249228 96.503 -74.673 51.173
249229 96.529 -74.681 51.159
249230 96.555 -74.687 51.147
249231 96.580 -74.694 51.137
249232 96.605 -74.702 51.128
249233 96.630 -74.709 51.122
249234 96.655 -74.716 51.116
249235 96.680 -74.724 51.111
249236 96.705 -74.731 51.107
249237 96.730 -74.738 51.105
249238 96.755 -74.746 51.100
249239 96.780 -74.753 51.093
249240 96.805 -74.760 51.086
249241 96.832 -74.773 51.077
249242 96.860 -74.784 51.065
249243 96.887 -74.796 51.052
249244 96.925 -74.822 51.038
249245 96.956 -74.843 51.024
249246 96.962 -74.845 51.011
249247 96.968 -74.848 50.997
249248 96.975 -74.851 50.981
249249 96.986 -74.859 50.963
249250 97.000 -74.869 50.946
249251 97.013 -74.878 50.930
249252 97.026 -74.888 50.916
249253 97.039 -74.898 50.901
249254 97.049 -74.910 50.889
249255 97.058 -74.922 50.879
249256 97.065 -74.935 50.872
249257 97.073 -74.949 50.865
249258 97.081 -74.962 50.858
249259 97.100 -74.966 50.848
249260 97.139 -74.947 50.835
249261 97.172 -74.938 50.817
249262 97.206 -74.928 50.797
249263 97.231 -74.937 50.775
249264 97.254 -74.945 50.749
249265 97.279 -74.954 50.720
249266 97.302 -74.963 50.692
249267 97.326 -74.971 50.663
249268 97.350 -74.980 50.634
249269 97.387 -74.983 50.603
249270 97.424 -74.986 50.571
249271 97.461 -74.989 50.536
249272 97.496 -74.994 50.500
249273 97.528 -75.001 50.463
249274 97.559 -75.010 50.424
249275 97.591 -75.018 50.384
249276 97.623 -75.026 50.346
249277 97.654 -75.034 50.309
249278 97.685 -75.042 50.270
249279 97.717 -75.050 50.231
249280 97.748 -75.058 50.191
249281 97.782 -75.065 50.147
249282 97.816 -75.072 50.102
249283 97.852 -75.079 50.054
249284 97.885 -75.086 50.002
249285 97.917 -75.093 49.951
249286 97.948 -75.101 49.900
249287 97.979 -75.112 49.849
249288 98.009 -75.122 49.796
249289 98.040 -75.132 49.745
249290 98.070 -75.142 49.694
249291 98.100 -75.151 49.643
249292 98.131 -75.161 49.587
249293 98.161 -75.172 49.531
249294 98.192 -75.182 49.477
249295 98.222 -75.192 49.426
249296 98.253 -75.201 49.377
249297 98.284 -75.211 49.326
249298 98.314 -75.221 49.276
249299 98.349 -75.237 49.226
249300 98.391 -75.259 49.176
249301 98.434 -75.282 49.127
249302 98.459 -75.272 49.079
249303 98.479 -75.253 49.032
249304 98.499 -75.234 48.983
249305 98.518 -75.215 48.935
249306 98.538 -75.196 48.885
249307 98.564 -75.178 48.834
249308 98.577 -75.189 48.782
249309 98.605 -75.191 48.730
249310 98.633 -75.192 48.677
249311 98.661 -75.193 48.625
249312 98.689 -75.195 48.573
249313 98.715 -75.198 48.520
249314 98.742 -75.203 48.468
249315 98.770 -75.207 48.416
249316 98.796 -75.212 48.363
249317 98.823 -75.217 48.311
249318 98.850 -75.221 48.261
249319 98.877 -75.228 48.211
249320 98.903 -75.258 48.163
249321 98.928 -75.289 48.114
249322 98.952 -75.320 48.066
249323 98.975 -75.355 48.020
249324 98.998 -75.389 47.974
249325 99.020 -75.422 47.929
249326 99.043 -75.456 47.885
249327 99.066 -75.490 47.842
249328 99.088 -75.524 47.799
249329 99.114 -75.548 47.757
249330 99.151 -75.534 47.716
249331 99.189 -75.521 47.678
249332 99.220 -75.521 47.639
249333 99.251 -75.524 47.602
249334 99.282 -75.526 47.566
249335 99.312 -75.527 47.531
249336 99.343 -75.530 47.498
249337 99.374 -75.532 47.467
249338 99.398 -75.544 47.438
249339 99.418 -75.561 47.409
249340 99.440 -75.577 47.381
249341 99.465 -75.593 47.354
249342 99.489 -75.610 47.327
249343 99.512 -75.627 47.302
249344 99.537 -75.643 47.278
249345 99.561 -75.659 47.255
249346 99.586 -75.677 47.232
249347 99.612 -75.693 47.210
249348 99.638 -75.711 47.187
249349 99.664 -75.728 47.166
249350 99.690 -75.746 47.145
249351 99.716 -75.763 47.126
249352 99.743 -75.780 47.108
249353 99.767 -75.802 47.089
249354 99.789 -75.831 47.069
249355 99.811 -75.862 47.048
249356 99.833 -75.891 47.023
249357 99.855 -75.921 46.997
249358 99.880 -75.946 46.972
249359 99.918 -75.947 46.945
249360 99.947 -75.958 46.917
249361 99.977 -75.968 46.888
249362 100.006 -75.978 46.856
249363 100.036 -75.988 46.825
249364 100.065 -75.999 46.794
249365 100.094 -76.009 46.762
249366 100.123 -76.019 46.730
249367 100.153 -76.029 46.697
249368 100.182 -76.040 46.663
249369 100.211 -76.050 46.630
249370 100.237 -76.062 46.597
249371 100.266 -76.063 46.566
249372 100.300 -76.053 46.534
249373 100.332 -76.043 46.503
249374 100.364 -76.034 46.471
249375 100.396 -76.024 46.438
249376 100.428 -76.014 46.406
249377 100.461 -76.004 46.374
249378 100.490 -76.015 46.344
249379 100.517 -76.044 46.312
249380 100.544 -76.074 46.280
249381 100.571 -76.104 46.247
249382 100.598 -76.133 46.214
249383 100.625 -76.163 46.181
249384 100.650 -76.194 46.148
249385 100.674 -76.227 46.116
249386 100.697 -76.260 46.083
249387 100.720 -76.294 46.050
249388 100.743 -76.326 46.018
249389 100.766 -76.360 45.987
249390 100.788 -76.370 45.956
249391 100.805 -76.388 45.928
249392 100.814 -76.421 45.900
249393 100.820 -76.431 45.872
249394 100.827 -76.433 45.845
249395 100.833 -76.436 45.817
249396 100.856 -76.442 45.790
249397 100.898 -76.453 45.765
249398 100.939 -76.464 45.741
249399 100.981 -76.476 45.716
249400 101.022 -76.487 45.692
249401 101.064 -76.498 45.669
249402 101.117 -76.515 45.645
249403 101.167 -76.519 45.621
249404 101.211 -76.514 45.598
249405 101.256 -76.508 45.574
249406 101.301 -76.502 45.551
249407 101.345 -76.496 45.529
249408 101.390 -76.491 45.506
249409 101.435 -76.485 45.484
249410 101.480 -76.480 45.462
249411 101.515 -76.481 45.440
249412 101.539 -76.490 45.418
249413 101.563 -76.499 45.398
249414 101.592 -76.510 45.377
249415 101.609 -76.521 45.356
249416 101.624 -76.533 45.336
249417 101.639 -76.546 45.315
249418 101.653 -76.558 45.294
249419 101.668 -76.570 45.274
249420 101.683 -76.582 45.254
249421 101.698 -76.594 45.235
249422 101.717 -76.599 45.215
249423 101.739 -76.599 45.196
249424 101.760 -76.599 45.177
249425 101.778 -76.606 45.158
249426 101.794 -76.618 45.139
249427 101.811 -76.629 45.121
249428 101.827 -76.640 45.103
249429 101.843 -76.651 45.084
249430 101.859 -76.662 45.066
249431 101.876 -76.674 45.048
249432 101.892 -76.684 45.031
249433 101.890 -76.681 45.013
249434 101.883 -76.674 44.994
249435 101.876 -76.668 44.977
249436 101.883 -76.670 44.960
249437 101.890 -76.672 44.942
249438 101.898 -76.675 44.926
249439 101.906 -76.678 44.908
249440 101.913 -76.680 44.891
249441 101.921 -76.683 44.873
249442 101.928 -76.686 44.856
249443 101.936 -76.689 44.839
249444 101.944 -76.692 44.822
249445 101.956 -76.699 44.805
249446 101.977 -76.718 44.788
249447 101.999 -76.738 44.771
249448 102.021 -76.758 44.754
249449 102.043 -76.777 44.737
249450 102.065 -76.797 44.720
249451 102.087 -76.817 44.704
249452 102.108 -76.836 44.687
249453 102.130 -76.856 44.671
249454 102.152 -76.875 44.654
249455 102.172 -76.884 44.637
249456 102.189 -76.885 44.620
249457 102.206 -76.886 44.603
249458 102.224 -76.886 44.587
249459 102.241 -76.887 44.571
249460 102.259 -76.887 44.555
249461 102.276 -76.887 44.538
249462 102.299 -76.889 44.522
249463 102.322 -76.890 44.506
249464 102.345 -76.890 44.489
249465 102.368 -76.892 44.474
249466 102.391 -76.893 44.459
249467 102.414 -76.893 44.443
249784 92.779 -74.225 54.580
249785 92.808 -74.244 54.582
249786 92.838 -74.263 54.583
249787 92.868 -74.281 54.583
249788 92.898 -74.300 54.583
249789 92.928 -74.305 54.583
249790 92.959 -74.306 54.581
249791 92.990 -74.307 54.579
249792 93.021 -74.308 54.578
249793 93.053 -74.309 54.576
249794 93.083 -74.310 54.574
249795 93.115 -74.311 54.573
249796 93.146 -74.312 54.573
249797 93.177 -74.314 54.573
249798 93.208 -74.314 54.574
249799 93.239 -74.316 54.576
249800 93.270 -74.317 54.582
249801 93.301 -74.318 54.592
249802 93.332 -74.319 54.606
249803 93.364 -74.305 54.625
249804 93.385 -74.313 54.646
249805 93.408 -74.314 54.670
249806 93.432 -74.310 54.696
249807 93.457 -74.306 54.724
249808 93.481 -74.302 54.753
249809 93.505 -74.298 54.779
249810 93.527 -74.294 54.801
249811 93.549 -74.289 54.820
249812 93.570 -74.286 54.842
249813 93.591 -74.282 54.862
249814 93.613 -74.278 54.881
249815 93.634 -74.274 54.903
249816 93.655 -74.270 54.925
249817 93.677 -74.267 54.939
249818 93.698 -74.263 54.945
249819 93.719 -74.259 54.947
249820 93.741 -74.255 54.949
249821 93.762 -74.251 54.943
249822 93.784 -74.247 54.932
249823 93.805 -74.243 54.923
249824 93.827 -74.239 54.908
249825 93.848 -74.236 54.893
249826 93.869 -74.232 54.878
249827 93.891 -74.229 54.860
249828 93.912 -74.230 54.840
249829 93.933 -74.230 54.816
249830 93.957 -74.233 54.787
249831 93.980 -74.237 54.753
249832 94.004 -74.241 54.724
249833 94.027 -74.245 54.696
249834 94.051 -74.248 54.668
249835 94.074 -74.252 54.639
249836 94.115 -74.260 54.616
249837 94.141 -74.262 54.597
249838 94.159 -74.261 54.581
249839 94.177 -74.261 54.568
249840 94.195 -74.259 54.558
249841 94.213 -74.258 54.548
249842 94.237 -74.261 54.538
249843 94.257 -74.261 54.525
249844 94.277 -74.261 54.508
249845 94.297 -74.261 54.489
249846 94.317 -74.263 54.469
249847 94.337 -74.266 54.447
249848 94.358 -74.270 54.425
249849 94.382 -74.267 54.402
249850 94.407 -74.258 54.378
249851 94.430 -74.243 54.350
249852 94.454 -74.242 54.321
249853 94.478 -74.240 54.291
249854 94.503 -74.238 54.260
249855 94.527 -74.236 54.229
249856 94.551 -74.235 54.196
249857 94.576 -74.233 54.161
249858 94.599 -74.234 54.122
249859 94.615 -74.244 54.077
249860 94.632 -74.254 54.033
249861 94.648 -74.259 53.987
249862 94.663 -74.258 53.941
249863 94.678 -74.258 53.895
249864 94.693 -74.258 53.846
249865 94.708 -74.258 53.797
249866 94.725 -74.261 53.749
249867 94.745 -74.273 53.702
249868 94.765 -74.286 53.656
249869 94.784 -74.298 53.611
249870 94.801 -74.319 53.570
249871 94.811 -74.355 53.538
249872 94.822 -74.392 53.512
249873 94.836 -74.423 53.489
249874 94.859 -74.428 53.465
249875 94.895 -74.402 53.431
249876 94.930 -74.376 53.395
249877 94.966 -74.349 53.357
249878 95.000 -74.342 53.318
249879 95.032 -74.345 53.275
249880 95.065 -74.354 53.228
249881 95.097 -74.368 53.179
249882 95.130 -74.382 53.129
249883 95.162 -74.396 53.076
249884 95.195 -74.409 53.020
249885 95.228 -74.424 52.962
249886 95.260 -74.437 52.904
249887 95.293 -74.451 52.843
249888 95.325 -74.465 52.782
249889 95.360 -74.471 52.723
249890 95.399 -74.465 52.666
249891 95.438 -74.458 52.611
249892 95.468 -74.461 52.557
249893 95.491 -74.471 52.505
249894 95.520 -74.470 52.452
249895 95.549 -74.469 52.403
249896 95.578 -74.469 52.355
249897 95.606 -74.468 52.308
249898 95.635 -74.468 52.265
249899 95.664 -74.468 52.222
249900 95.693 -74.467 52.180
249901 95.722 -74.467 52.138
249902 95.751 -74.466 52.096
249903 95.779 -74.465 52.055
249904 95.808 -74.465 52.014
249905 95.837 -74.465 51.974
249906 95.867 -74.471 51.936
249907 95.895 -74.477 51.897
249908 95.905 -74.473 51.857
249909 95.928 -74.493 51.817
249910 95.956 -74.512 51.774
249911 95.988 -74.532 51.731
249912 96.020 -74.552 51.689
249913 96.052 -74.572 51.648
249914 96.085 -74.592 51.609
249915 96.114 -74.611 51.569
249916 96.139 -74.631 51.529
249917 96.165 -74.650 51.491
249918 96.191 -74.668 51.456
249919 96.222 -74.681 51.421
249920 96.259 -74.685 51.387
249921 96.257 -74.697 51.354
249922 96.252 -74.710 51.324
249923 96.246 -74.723 51.296
249924 96.315 -74.679 51.269
249925 96.390 -74.631 51.245
249926 96.428 -74.627 51.223
249927 96.452 -74.634 51.203
249928 96.477 -74.642 51.185
249929 96.502 -74.649 51.168
249930 96.527 -74.656 51.152
249931 96.552 -74.664 51.139
249932 96.577 -74.671 51.127
249933 96.602 -74.678 51.117
249934 96.627 -74.686 51.110
249935 96.652 -74.693 51.104
249936 96.677 -74.700 51.098
249937 96.702 -74.708 51.094
249938 96.727 -74.715 51.091
249939 96.752 -74.722 51.088
249940 96.777 -74.729 51.084
249941 96.802 -74.737 51.080
249942 96.830 -74.752 51.074
249943 96.859 -74.766 51.064
249944 96.884 -74.774 51.053
249945 96.911 -74.784 51.040
249946 96.949 -74.809 51.029
249947 96.968 -74.822 51.017
249948 96.974 -74.825 51.003
249949 96.980 -74.828 50.988
249950 96.994 -74.837 50.969
249951 97.007 -74.847 50.953
249952 97.020 -74.856 50.937
249953 97.033 -74.866 50.923
249954 97.046 -74.875 50.909
249955 97.059 -74.885 50.898
249956 97.068 -74.898 50.889
249957 97.076 -74.911 50.882
249958 97.083 -74.924 50.877
249959 97.091 -74.938 50.871
249960 97.100 -74.943 50.862
249961 97.143 -74.918 50.848
249962 97.176 -74.909 50.831
249963 97.207 -74.905 50.811
249964 97.231 -74.914 50.787
249965 97.255 -74.922 50.761
249966 97.279 -74.931 50.732
249967 97.303 -74.940 50.704
249968 97.330 -74.947 50.677
249969 97.368 -74.950 50.646
249970 97.405 -74.953 50.616
249971 97.442 -74.956 50.583
249972 97.479 -74.958 50.549
249973 97.516 -74.962 50.515
249974 97.547 -74.969 50.479
249975 97.579 -74.978 50.441
249976 97.610 -74.986 50.403
249977 97.642 -74.994 50.364
249978 97.673 -75.002 50.326
249979 97.705 -75.010 50.288
249980 97.736 -75.018 50.249
249981 97.768 -75.026 50.207
249982 97.800 -75.035 50.163
249983 97.831 -75.042 50.116
249984 97.858 -75.052 50.067
249985 97.890 -75.059 50.014
249986 97.923 -75.066 49.961
249987 97.956 -75.074 49.908
249988 97.986 -75.084 49.855
249989 98.017 -75.094 49.802
249990 98.048 -75.104 49.753
249991 98.078 -75.113 49.703
249992 98.108 -75.123 49.650
249993 98.139 -75.133 49.594
249994 98.170 -75.143 49.537
249995 98.201 -75.153 49.482
249996 98.231 -75.163 49.430
249997 98.262 -75.173 49.380
249998 98.292 -75.184 49.330
249999 98.323 -75.197 49.280
250000 98.360 -75.217 49.230
250001 98.401 -75.242 49.180
250002 98.444 -75.264 49.131
250003 98.465 -75.249 49.082
250004 98.485 -75.229 49.034
250005 98.505 -75.210 48.986
250006 98.525 -75.192 48.937
250007 98.545 -75.173 48.888
250008 98.565 -75.151 48.837
250009 98.591 -75.153 48.786
250010 98.618 -75.157 48.734
250011 98.645 -75.162 48.681
250012 98.671 -75.167 48.628
250013 98.699 -75.171 48.575
250014 98.725 -75.176 48.522
250015 98.752 -75.180 48.469
250016 98.779 -75.185 48.416
250017 98.806 -75.189 48.363
250018 98.834 -75.194 48.311
250019 98.860 -75.198 48.259
250020 98.886 -75.217 48.209
250021 98.912 -75.247 48.160
250022 98.937 -75.277 48.111
250023 98.962 -75.308 48.063
250024 98.988 -75.338 48.017
250025 99.012 -75.370 47.971
250026 99.034 -75.404 47.926
250027 99.057 -75.438 47.882
250028 99.080 -75.471 47.838
250029 99.103 -75.503 47.796
250030 99.141 -75.489 47.753
250031 99.174 -75.486 47.712
250032 99.205 -75.487 47.672
250033 99.235 -75.489 47.633
250034 99.267 -75.492 47.594
250035 99.297 -75.494 47.558
250036 99.327 -75.498 47.523
250037 99.357 -75.502 47.489
250038 99.387 -75.506 47.458
250039 99.412 -75.514 47.428
250040 99.433 -75.530 47.399
250041 99.453 -75.546 47.371
250042 99.478 -75.563 47.344
250043 99.501 -75.580 47.318
250044 99.525 -75.596 47.293
250045 99.550 -75.612 47.269
250046 99.574 -75.629 47.246
250047 99.600 -75.646 47.224
250048 99.627 -75.664 47.201
250049 99.652 -75.681 47.179
250050 99.679 -75.698 47.158
250051 99.705 -75.715 47.137
250052 99.731 -75.733 47.118
250053 99.757 -75.750 47.100
250054 99.782 -75.771 47.081
250055 99.804 -75.801 47.062
250056 99.826 -75.831 47.042
250057 99.848 -75.860 47.017
250058 99.870 -75.890 46.992
250059 99.892 -75.919 46.967
250060 99.918 -75.940 46.941
250061 99.947 -75.950 46.913
250062 99.978 -75.959 46.883
250063 100.007 -75.966 46.852
250064 100.037 -75.975 46.820
250065 100.068 -75.983 46.789
250066 100.098 -75.991 46.756
250067 100.128 -75.999 46.723
250068 100.158 -76.007 46.690
250069 100.188 -76.015 46.657
250070 100.218 -76.023 46.623
250071 100.249 -76.032 46.590
250072 100.277 -76.050 46.557
250073 100.309 -76.041 46.526
250074 100.341 -76.031 46.495
250075 100.373 -76.021 46.463
250076 100.405 -76.011 46.430
250077 100.437 -76.002 46.398
250078 100.470 -75.992 46.367
250079 100.499 -76.005 46.338
250080 100.526 -76.035 46.306
250081 100.553 -76.063 46.274
250082 100.580 -76.093 46.241
250083 100.607 -76.122 46.208
250084 100.634 -76.152 46.175
250085 100.661 -76.182 46.142
250086 100.688 -76.211 46.109
250087 100.715 -76.241 46.076
250088 100.741 -76.271 46.043
250089 100.764 -76.304 46.012
250090 100.787 -76.338 45.981
250091 100.810 -76.361 45.950
250092 100.824 -76.383 45.923
250093 100.831 -76.395 45.894
250094 100.838 -76.398 45.866
250095 100.844 -76.400 45.839
250096 100.850 -76.402 45.811
250097 100.882 -76.411 45.785
250098 100.923 -76.422 45.760
250099 100.965 -76.433 45.735
250100 101.006 -76.445 45.711
250101 101.048 -76.456 45.687
250102 101.090 -76.467 45.663
250103 101.131 -76.478 45.640
250104 101.184 -76.495 45.615
250105 101.237 -76.505 45.592
250106 101.281 -76.499 45.568
250107 101.326 -76.494 45.546
250108 101.371 -76.488 45.523
250109 101.416 -76.483 45.501
250110 101.461 -76.477 45.478
250111 101.505 -76.471 45.456
250112 101.546 -76.469 45.434
250113 101.570 -76.478 45.413
250114 101.595 -76.487 45.391
250115 101.622 -76.498 45.370
250116 101.638 -76.510 45.350
250117 101.653 -76.522 45.329
250118 101.667 -76.533 45.309
250119 101.682 -76.546 45.287
250120 101.697 -76.558 45.267
250121 101.711 -76.570 45.247
250122 101.731 -76.574 45.227
250123 101.752 -76.574 45.207
250124 101.773 -76.575 45.188
250130 101.872 -76.643 45.075
250131 101.883 -76.650 45.056
250132 101.876 -76.643 45.038
250133 101.869 -76.636 45.020
250134 101.864 -76.630 45.003
250135 101.871 -76.633 44.984
250136 101.879 -76.636 44.967
250137 101.887 -76.638 44.949
250138 101.894 -76.641 44.931
250139 101.902 -76.644 44.914
250140 101.909 -76.646 44.897
250141 101.917 -76.649 44.879
250142 101.924 -76.652 44.862
250143 101.933 -76.655 44.845
250144 101.940 -76.658 44.828
250145 101.956 -76.670 44.811
250146 101.977 -76.689 44.793
250147 101.999 -76.708 44.777
250148 102.021 -76.728 44.760
250149 102.043 -76.748 44.743
250150 102.065 -76.768 44.725
250151 102.087 -76.787 44.709
250152 102.108 -76.807 44.692
250153 102.130 -76.827 44.676
250154 102.152 -76.843 44.659
250155 102.169 -76.844 44.642
250156 102.186 -76.845 44.625
250157 102.204 -76.846 44.609
250158 102.222 -76.846 44.592
250159 102.239 -76.846 44.576
250160 102.256 -76.847 44.560
250161 102.274 -76.847 44.544
250162 102.296 -76.849 44.528
250488 92.869 -74.255 54.591
250489 92.899 -74.273 54.590
250490 92.929 -74.283 54.591
250491 92.960 -74.284 54.590
250492 92.991 -74.286 54.588
250493 93.022 -74.286 54.586
250494 93.053 -74.288 54.584
250495 93.084 -74.289 54.582
250496 93.116 -74.290 54.581
250497 93.146 -74.291 54.580
250498 93.178 -74.292 54.580
250499 93.209 -74.293 54.580
250500 93.240 -74.294 54.583
250501 93.271 -74.295 54.587
250502 93.302 -74.296 54.595
250503 93.335 -74.281 54.607
250504 93.360 -74.280 54.624
250505 93.381 -74.289 54.644
250506 93.404 -74.291 54.666
250507 93.428 -74.286 54.692
250508 93.452 -74.283 54.721
250509 93.477 -74.279 54.749
250510 93.501 -74.274 54.775
250511 93.525 -74.270 54.797
250512 93.550 -74.266 54.817
250513 93.572 -74.262 54.844
250514 93.593 -74.258 54.866
250515 93.615 -74.255 54.884
250516 93.636 -74.251 54.904
250517 93.657 -74.247 54.925
250518 93.679 -74.243 54.937
250519 93.700 -74.239 54.945
250520 93.721 -74.235 54.953
250521 93.743 -74.231 54.961
250522 93.764 -74.227 54.961
250523 93.785 -74.223 54.954
250524 93.807 -74.220 54.945
250525 93.828 -74.216 54.928
250526 93.850 -74.212 54.909
250527 93.871 -74.208 54.888
250528 93.893 -74.208 54.864
250529 93.914 -74.208 54.841
250530 93.935 -74.209 54.815
250531 93.957 -74.210 54.786
250532 93.980 -74.213 54.754
250533 94.003 -74.216 54.723
250534 94.027 -74.220 54.694
250535 94.050 -74.223 54.665
250536 94.073 -74.226 54.636
250537 94.118 -74.236 54.611
250538 94.149 -74.240 54.590
250539 94.167 -74.239 54.574
250540 94.185 -74.239 54.565
250541 94.202 -74.237 54.558
250542 94.220 -74.236 54.549
250543 94.238 -74.236 54.538
250544 94.258 -74.236 54.526
250545 94.278 -74.237 54.507
250546 94.298 -74.240 54.488
250547 94.319 -74.243 54.468
250548 94.339 -74.246 54.445
250549 94.359 -74.250 54.422
250550 94.384 -74.245 54.399
250551 94.410 -74.240 54.375
250552 94.433 -74.228 54.349
250553 94.457 -74.213 54.319
250554 94.481 -74.208 54.289
250555 94.505 -74.205 54.258
250556 94.530 -74.204 54.225
250557 94.554 -74.202 54.192
250558 94.578 -74.201 54.155
250559 94.603 -74.199 54.114
250560 94.619 -74.209 54.068
250561 94.636 -74.219 54.022
250562 94.652 -74.222 53.975
250563 94.667 -74.221 53.927
250564 94.682 -74.221 53.881
250565 94.697 -74.220 53.832
250566 94.712 -74.220 53.783
250567 94.728 -74.220 53.735
250568 94.747 -74.231 53.688
250569 94.767 -74.244 53.642
250570 94.786 -74.256 53.597
250571 94.800 -74.285 53.556
250572 94.810 -74.321 53.524
250573 94.823 -74.353 53.500
250574 94.837 -74.384 53.478
250575 94.862 -74.386 53.455
250576 94.898 -74.359 53.421
250577 94.934 -74.333 53.385
250578 94.969 -74.314 53.348
250579 95.001 -74.317 53.308
250580 95.034 -74.321 53.262
250581 95.066 -74.324 53.214
250582 95.099 -74.337 53.164
250583 95.131 -74.351 53.113
250584 95.164 -74.365 53.059
250585 95.196 -74.379 53.005
250586 95.229 -74.393 52.949
250587 95.261 -74.407 52.894
250588 95.294 -74.421 52.833
250589 95.326 -74.434 52.773
250590 95.364 -74.433 52.715
250591 95.403 -74.427 52.658
250592 95.442 -74.420 52.603
250593 95.479 -74.415 52.550
250594 95.501 -74.426 52.497
250595 95.530 -74.426 52.447
250596 95.559 -74.425 52.399
250597 95.588 -74.424 52.352
250598 95.616 -74.424 52.306
250599 95.645 -74.424 52.264
250600 95.674 -74.424 52.222
250601 95.703 -74.423 52.181
250602 95.732 -74.422 52.139
250603 95.761 -74.422 52.098
250604 95.790 -74.424 52.058
250605 95.819 -74.430 52.017
250606 95.848 -74.436 51.976
250607 95.878 -74.443 51.937
250608 95.892 -74.442 51.898
250609 95.902 -74.438 51.858
250610 95.922 -74.452 51.819
250611 95.946 -74.472 51.777
250612 95.970 -74.492 51.734
250613 96.001 -74.512 51.692
250614 96.033 -74.531 51.650
250615 96.065 -74.551 51.610
250616 96.097 -74.571 51.570
250617 96.129 -74.591 51.530
250618 96.161 -74.611 51.492
250619 96.193 -74.631 51.456
250620 96.225 -74.650 51.421
250621 96.259 -74.668 51.388
250622 96.285 -74.654 51.355
250623 96.311 -74.637 51.324
250624 96.336 -74.618 51.296
250625 96.363 -74.601 51.269
250626 96.389 -74.584 51.244
250627 96.423 -74.602 51.221
250628 96.450 -74.611 51.200
250629 96.474 -74.618 51.181
250630 96.500 -74.625 51.163
250631 96.524 -74.633 51.145
250632 96.549 -74.640 51.131
250633 96.574 -74.647 51.119
250634 96.599 -74.655 51.108
250635 96.624 -74.662 51.099
250636 96.649 -74.669 51.092
250637 96.674 -74.677 51.086
250638 96.699 -74.684 51.081
250639 96.724 -74.691 51.078
250640 96.749 -74.699 51.075
250641 96.774 -74.706 51.073
250642 96.799 -74.713 51.071
250643 96.829 -74.731 51.067
250644 96.858 -74.747 51.061
250645 96.884 -74.755 51.052
250646 96.909 -74.763 51.040
250647 96.936 -74.772 51.030
250648 96.973 -74.798 51.020
250649 96.980 -74.801 51.008
250650 96.987 -74.805 50.995
250651 97.000 -74.815 50.976
250652 97.013 -74.825 50.960
250653 97.026 -74.834 50.944
250654 97.039 -74.844 50.930
250655 97.052 -74.853 50.917
250656 97.066 -74.863 50.908
250657 97.078 -74.874 50.900
250658 97.086 -74.887 50.895
250659 97.094 -74.897 50.889
250660 97.102 -74.903 50.883
250661 97.110 -74.908 50.876
250662 97.146 -74.889 50.862
250663 97.180 -74.880 50.843
250664 97.208 -74.882 50.821
250665 97.232 -74.891 50.797
250666 97.256 -74.900 50.771
250667 97.280 -74.908 50.742
250668 97.311 -74.913 50.715
250669 97.349 -74.916 50.687
250670 97.385 -74.919 50.657
250671 97.423 -74.922 50.627
250672 97.460 -74.925 50.596
250673 97.497 -74.928 50.563
250674 97.534 -74.931 50.530
250675 97.567 -74.938 50.495
250676 97.598 -74.946 50.458
250677 97.630 -74.954 50.419
250678 97.661 -74.962 50.381
250679 97.693 -74.970 50.341
250680 97.724 -74.979 50.301
250681 97.756 -74.986 50.260
250682 97.788 -74.994 50.216
250683 97.819 -75.003 50.171
250684 97.845 -75.013 50.124
250685 97.868 -75.023 50.074
250686 97.896 -75.032 50.020
250687 97.929 -75.039 49.965
250688 97.964 -75.045 49.912
250689 97.995 -75.055 49.859
250690 98.025 -75.065 49.808
250691 98.056 -75.075 49.760
250692 98.087 -75.085 49.710
250693 98.117 -75.095 49.657
250694 98.148 -75.104 49.600
250695 98.178 -75.114 49.542
250696 98.209 -75.124 49.486
250697 98.239 -75.135 49.433
250698 98.271 -75.148 49.382
250699 98.302 -75.161 49.331
250700 98.333 -75.174 49.281
250701 98.371 -75.196 49.231
250702 98.413 -75.221 49.182
250703 98.452 -75.242 49.134
250704 98.472 -75.220 49.086
250705 98.492 -75.198 49.038
250706 98.512 -75.176 48.989
250707 98.531 -75.154 48.939
250708 98.551 -75.132 48.889
250709 98.573 -75.119 48.839
250710 98.600 -75.130 48.787
250711 98.628 -75.135 48.735
250712 98.655 -75.139 48.682
250713 98.681 -75.144 48.629
250714 98.709 -75.148 48.575
250715 98.735 -75.153 48.522
250716 98.763 -75.157 48.469
250717 98.789 -75.161 48.416
250718 98.816 -75.166 48.362
250719 98.843 -75.170 48.310
250720 98.870 -75.175 48.258
250721 98.896 -75.204 48.208
250722 98.921 -75.235 48.158
250723 98.946 -75.265 48.109
250724 98.972 -75.295 48.060
250725 98.997 -75.326 48.014
250726 99.023 -75.356 47.969
250727 99.048 -75.386 47.924
250728 99.072 -75.419 47.879
250729 99.094 -75.453 47.835
250730 99.128 -75.449 47.793
250731 99.159 -75.452 47.750
250732 99.189 -75.455 47.709
250733 99.219 -75.458 47.668
250734 99.250 -75.462 47.627
250735 99.279 -75.467 47.588
250736 99.309 -75.471 47.550
250737 99.340 -75.474 47.515
250738 99.369 -75.479 47.480
250739 99.398 -75.484 47.448
250740 99.422 -75.493 47.418
250741 99.447 -75.501 47.390
250742 99.468 -75.516 47.362
250743 99.490 -75.533 47.335
250744 99.514 -75.549 47.309
250745 99.539 -75.565 47.284
250746 99.563 -75.582 47.261
250747 99.589 -75.599 47.238
250748 99.615 -75.617 47.216
250749 99.641 -75.634 47.193
250750 99.667 -75.651 47.172
250751 99.693 -75.668 47.151
250752 99.719 -75.686 47.130
250753 99.746 -75.703 47.111
250754 99.771 -75.720 47.092
250755 99.796 -75.740 47.073
250756 99.819 -75.770 47.055
250757 99.841 -75.800 47.035
250758 99.863 -75.829 47.011
250759 99.885 -75.859 46.987
250760 99.908 -75.887 46.962
250761 99.931 -75.897 46.937
250762 99.961 -75.905 46.908
250763 99.991 -75.913 46.879
250764 100.021 -75.922 46.848
250765 100.051 -75.930 46.815
250766 100.081 -75.938 46.783
250767 100.112 -75.946 46.750
250768 100.142 -75.954 46.717
250769 100.172 -75.962 46.683
250770 100.202 -75.970 46.650
250771 100.232 -75.978 46.616
250772 100.261 -75.994 46.582
250773 100.289 -76.016 46.550
250774 100.318 -76.028 46.518
250775 100.350 -76.018 46.487
250776 100.382 -76.009 46.455
250777 100.414 -75.999 46.423
250778 100.446 -75.989 46.391
250779 100.479 -75.979 46.361
250780 100.507 -75.994 46.331
250781 100.535 -76.024 46.299
250782 100.562 -76.053 46.267
250783 100.588 -76.083 46.235
250784 100.616 -76.112 46.203
250785 100.643 -76.141 46.169
250786 100.670 -76.171 46.136
250787 100.697 -76.201 46.104
250788 100.724 -76.230 46.071
250789 100.751 -76.260 46.039
250790 100.778 -76.289 46.008
250791 100.805 -76.319 45.976
250792 100.831 -76.349 45.945
250793 100.842 -76.360 45.917
250794 100.848 -76.362 45.889
250795 100.855 -76.364 45.861
250796 100.861 -76.367 45.834
250797 100.868 -76.369 45.807
250798 100.908 -76.379 45.780
250799 100.950 -76.391 45.755
250800 100.991 -76.402 45.730
250801 101.032 -76.414 45.706
250802 101.074 -76.424 45.681
250803 101.115 -76.436 45.657
250804 101.157 -76.447 45.634
250805 101.198 -76.458 45.610
250806 101.252 -76.475 45.586
250807 101.307 -76.491 45.562
250808 101.352 -76.486 45.540
250809 101.397 -76.480 45.518
250810 101.442 -76.474 45.495
250811 101.486 -76.468 45.472
250812 101.531 -76.463 45.449
250813 101.575 -76.457 45.428
250814 101.601 -76.466 45.407
250815 101.625 -76.475 45.386
250816 101.652 -76.486 45.364
250817 101.666 -76.498 45.343
250818 101.681 -76.510 45.322
250819 101.695 -76.522 45.302
250820 101.710 -76.534 45.281
250821 101.725 -76.546 45.260
250822 101.745 -76.549 45.239
250823 101.766 -76.549 45.219
250824 101.785 -76.555 45.199
250825 101.802 -76.567 45.179
250840 101.905 -76.612 44.903
250841 101.913 -76.615 44.886
251189 92.870 -74.229 54.598
251190 92.900 -74.248 54.597
251191 92.930 -74.262 54.598
251192 92.961 -74.263 54.598
251193 92.992 -74.264 54.596
251194 93.023 -74.265 54.594
251195 93.054 -74.266 54.592
251196 93.085 -74.267 54.590
251197 93.116 -74.268 54.589
251198 93.147 -74.270 54.588
251199 93.179 -74.270 54.587
251200 93.210 -74.272 54.588
251201 93.240 -74.273 54.591
251202 93.272 -74.274 54.593
251203 93.304 -74.262 54.600
251204 93.335 -74.248 54.611
251205 93.356 -74.256 54.625
251206 93.377 -74.264 54.642
251207 93.400 -74.267 54.662
251208 93.424 -74.264 54.686
251209 93.448 -74.260 54.712
251210 93.473 -74.255 54.738
251211 93.497 -74.252 54.763
251212 93.521 -74.247 54.788
251213 93.546 -74.243 54.812
251214 93.570 -74.239 54.839
251215 93.595 -74.235 54.862
251216 93.617 -74.231 54.881
251217 93.638 -74.227 54.897
251218 93.659 -74.223 54.915
251219 93.681 -74.220 54.929
251220 93.702 -74.216 54.942
251221 93.723 -74.211 54.961
251222 93.745 -74.208 54.971
251223 93.766 -74.204 54.974
251224 93.787 -74.200 54.972
251225 93.809 -74.196 54.963
251226 93.830 -74.192 54.945
251227 93.852 -74.189 54.923
251228 93.873 -74.186 54.897
251229 93.895 -74.186 54.869
251230 93.916 -74.187 54.842
251231 93.937 -74.188 54.815
251232 93.959 -74.189 54.787
251233 93.980 -74.189 54.755
251234 94.003 -74.192 54.724
251235 94.026 -74.195 54.693
251236 94.050 -74.198 54.663
251237 94.074 -74.202 54.634
251238 94.121 -74.212 54.608
251239 94.157 -74.218 54.584
251240 94.175 -74.217 54.566
251241 94.192 -74.217 54.560
251242 94.210 -74.216 54.554
251243 94.228 -74.214 54.546
251244 94.244 -74.214 54.536
251245 94.260 -74.214 54.522
251246 94.280 -74.217 54.504
251247 94.300 -74.220 54.484
251248 94.321 -74.223 54.464
251249 94.341 -74.227 54.441
251250 94.361 -74.230 54.417
251251 94.387 -74.224 54.394
251252 94.412 -74.219 54.370
251253 94.437 -74.213 54.344
251254 94.461 -74.198 54.314
251255 94.484 -74.183 54.284
251256 94.508 -74.173 54.252
251257 94.532 -74.171 54.217
251258 94.556 -74.170 54.182
251259 94.581 -74.168 54.145
251260 94.605 -74.167 54.104
251261 94.623 -74.174 54.057
251262 94.640 -74.183 54.009
251263 94.655 -74.184 53.961
251264 94.671 -74.184 53.913
251265 94.686 -74.183 53.865
251266 94.701 -74.183 53.817
251267 94.716 -74.183 53.768
251268 94.731 -74.183 53.720
251269 94.749 -74.189 53.671
251270 94.768 -74.201 53.624
251271 94.788 -74.214 53.581
251272 94.799 -74.250 53.542
251273 94.811 -74.283 53.511
251274 94.825 -74.314 53.488
251275 94.839 -74.346 53.467
251276 94.866 -74.343 53.445
251277 94.901 -74.317 53.411
251278 94.937 -74.290 53.376
251279 94.970 -74.291 53.340
251280 95.002 -74.294 53.298
251281 95.035 -74.298 53.250
251282 95.067 -74.301 53.200
251283 95.100 -74.307 53.149
251284 95.132 -74.320 53.097
251285 95.165 -74.334 53.043
251286 95.197 -74.349 52.989
251287 95.230 -74.362 52.936
251288 95.262 -74.376 52.883
251289 95.295 -74.390 52.823
251290 95.328 -74.402 52.764
251291 95.368 -74.395 52.706
251292 95.406 -74.388 52.650
251293 95.446 -74.382 52.595
251294 95.485 -74.375 52.541
251295 95.512 -74.380 52.489
251296 95.540 -74.381 52.442
251297 95.569 -74.381 52.394
251298 95.598 -74.380 52.348
251299 95.626 -74.380 52.304
251300 95.656 -74.380 52.263
251301 95.684 -74.379 52.222
251302 95.713 -74.379 52.181
251303 95.742 -74.383 52.140
251304 95.772 -74.390 52.100
251305 95.801 -74.396 52.060
251306 95.830 -74.402 52.019
251307 95.860 -74.408 51.978
251308 95.880 -74.410 51.939
251309 95.890 -74.407 51.898
251310 95.899 -74.403 51.859
251311 95.916 -74.412 51.819
251312 95.940 -74.432 51.778
251313 95.964 -74.452 51.736
251314 95.987 -74.471 51.694
251315 96.014 -74.491 51.653
251316 96.045 -74.511 51.612
251317 96.078 -74.530 51.572
251318 96.110 -74.551 51.532
251319 96.142 -74.571 51.494
251320 96.174 -74.590 51.457
251321 96.209 -74.606 51.423
251322 96.254 -74.614 51.389
251323 96.270 -74.614 51.356
251324 96.297 -74.596 51.325
251325 96.323 -74.578 51.296
251326 96.349 -74.561 51.269
251327 96.376 -74.547 51.244
251328 96.410 -74.566 51.220
251329 96.445 -74.585 51.198
251330 96.471 -74.595 51.177
251331 96.497 -74.602 51.158
251332 96.521 -74.609 51.140
251333 96.546 -74.617 51.124
251334 96.572 -74.624 51.111
251335 96.596 -74.631 51.100
251336 96.621 -74.639 51.090
251337 96.646 -74.646 51.082
251338 96.671 -74.653 51.075
251339 96.696 -74.661 51.069
251340 96.721 -74.668 51.066
251341 96.746 -74.675 51.064
251342 96.771 -74.683 51.062
251343 96.797 -74.691 51.060
251344 96.827 -74.709 51.059
251345 96.858 -74.728 51.055
251346 96.883 -74.737 51.048
251347 96.909 -74.744 51.039
251348 96.935 -74.753 51.029
251349 96.960 -74.761 51.021
251350 96.985 -74.778 51.011
251351 96.994 -74.783 50.999
251352 97.007 -74.793 50.982
251353 97.020 -74.802 50.966
251354 97.033 -74.812 50.952
251355 97.046 -74.822 50.938
251356 97.060 -74.831 50.927
251357 97.073 -74.841 50.918
251358 97.085 -74.851 50.911
251359 97.096 -74.856 50.906
251360 97.105 -74.863 50.900
251361 97.113 -74.868 50.895
251362 97.121 -74.873 50.888
251363 97.150 -74.860 50.876
251364 97.184 -74.852 50.855
251365 97.209 -74.859 50.832
251366 97.232 -74.868 50.807
251367 97.256 -74.877 50.780
251368 97.292 -74.880 50.751
251369 97.329 -74.883 50.723
251370 97.366 -74.886 50.695
251371 97.403 -74.889 50.666
251372 97.441 -74.892 50.637
251373 97.477 -74.894 50.606
251374 97.515 -74.897 50.575
251375 97.552 -74.900 50.543
251376 97.586 -74.906 50.508
251377 97.618 -74.914 50.470
251378 97.649 -74.922 50.431
251379 97.681 -74.930 50.392
251380 97.712 -74.938 50.351
251381 97.744 -74.947 50.309
251382 97.776 -74.954 50.266
251383 97.807 -74.963 50.222
251384 97.833 -74.972 50.177
251385 97.856 -74.984 50.129
251386 97.879 -74.994 50.079
251387 97.902 -75.006 50.025
251388 97.936 -75.011 49.969
251389 97.973 -75.016 49.916
251390 98.003 -75.026 49.864
251391 98.034 -75.036 49.813
251392 98.065 -75.047 49.766
251393 98.095 -75.057 49.716
251394 98.126 -75.066 49.662
251395 98.156 -75.076 49.605
251396 98.187 -75.086 49.547
251397 98.218 -75.098 49.490
251398 98.249 -75.112 49.436
251399 98.280 -75.125 49.385
251400 98.311 -75.138 49.333
251401 98.343 -75.151 49.282
251402 98.382 -75.175 49.233
251403 98.424 -75.201 49.184
251404 98.458 -75.201 49.136
251405 98.479 -75.179 49.088
251406 98.498 -75.157 49.040
251407 98.518 -75.135 48.991
251408 98.538 -75.113 48.941
251409 98.558 -75.090 48.890
251410 98.583 -75.092 48.839
251411 98.610 -75.103 48.786
251412 98.638 -75.112 48.734
251413 98.664 -75.116 48.682
251414 98.692 -75.120 48.628
251415 98.718 -75.125 48.574
251416 98.745 -75.129 48.520
251417 98.773 -75.134 48.467
251418 98.799 -75.138 48.415
251419 98.827 -75.143 48.361
251420 98.853 -75.148 48.309
251421 98.880 -75.162 48.257
251422 98.905 -75.192 48.206
251423 98.931 -75.223 48.156
251424 98.956 -75.253 48.107
251425 98.981 -75.283 48.059
251426 99.007 -75.314 48.012
251427 99.032 -75.344 47.966
251428 99.058 -75.374 47.922
251429 99.083 -75.405 47.877
251430 99.112 -75.419 47.833
251431 99.141 -75.423 47.789
251432 99.171 -75.427 47.747
251433 99.202 -75.431 47.706
251434 99.232 -75.435 47.666
251435 99.262 -75.439 47.624
251436 99.292 -75.443 47.583
251437 99.322 -75.447 47.544
251438 99.352 -75.452 47.507
251439 99.382 -75.455 47.471
251440 99.409 -75.462 47.439
251441 99.433 -75.471 47.410
251442 99.457 -75.479 47.381
251443 99.482 -75.486 47.353
251444 99.503 -75.502 47.327
251445 99.527 -75.518 47.301
251446 99.551 -75.535 47.275
251447 99.577 -75.552 47.253
251448 99.604 -75.569 47.230
251449 99.630 -75.587 47.208
251450 99.655 -75.604 47.186
251451 99.682 -75.621 47.166
251452 99.708 -75.638 47.144
251453 99.734 -75.656 47.123
251454 99.760 -75.673 47.105
251455 99.786 -75.690 47.086
251456 99.811 -75.709 47.067
251457 99.834 -75.739 47.048
251458 99.855 -75.768 47.029
251459 99.877 -75.798 47.006
251460 99.900 -75.828 46.982
251461 99.946 -75.832 46.958
251462 99.951 -75.850 46.932
251463 99.975 -75.860 46.904
251464 100.005 -75.869 46.874
251465 100.035 -75.876 46.843
251466 100.065 -75.884 46.809
251467 100.095 -75.893 46.777
251468 100.125 -75.900 46.744
251469 100.155 -75.909 46.710
251470 100.186 -75.917 46.676
251471 100.216 -75.925 46.643
251472 100.246 -75.935 46.608
251473 100.273 -75.957 46.574
251474 100.301 -75.979 46.542
251475 100.329 -76.001 46.511
251476 100.359 -76.006 46.479
251477 100.391 -75.996 46.448
251478 100.423 -75.986 46.416
251479 100.456 -75.977 46.385
251480 100.488 -75.967 46.354
251481 100.516 -75.984 46.324
251482 100.543 -76.013 46.294
251483 100.570 -76.043 46.262
251484 100.598 -76.072 46.230
251485 100.625 -76.102 46.198
251486 100.651 -76.132 46.164
251487 100.679 -76.161 46.132
251488 100.706 -76.190 46.099
251489 100.733 -76.220 46.067
251490 100.760 -76.249 46.035
251491 100.787 -76.279 46.004
251492 100.816 -76.303 45.972
251493 100.847 -76.320 45.941
251494 100.859 -76.326 45.912
251495 100.866 -76.329 45.885
251496 100.873 -76.331 45.857
251497 100.879 -76.333 45.829
251498 100.892 -76.337 45.802
251499 100.933 -76.348 45.776
251500 100.975 -76.360 45.750
251501 101.016 -76.371 45.726
251502 101.058 -76.383 45.701
251503 101.100 -76.393 45.677
251504 101.141 -76.404 45.652
251505 101.183 -76.416 45.628
251506 101.224 -76.427 45.604
251507 101.266 -76.439 45.580
251508 101.319 -76.455 45.557
251509 101.376 -76.473 45.533
251510 101.422 -76.471 45.512
251511 101.467 -76.466 45.489
251512 101.512 -76.460 45.467
251513 101.556 -76.455 45.444
251514 101.601 -76.448 45.422
251515 101.631 -76.454 45.401
251516 101.656 -76.463 45.380
251517 101.680 -76.474 45.358
251518 101.694 -76.486 45.337
251519 101.709 -76.498 45.316
251520 101.724 -76.510 45.295
251521 101.739 -76.522 45.274
251522 101.759 -76.524 45.252
251523 101.780 -76.524 45.232
251524 101.797 -76.536 45.211
251525 101.813 -76.546 45.191
251526 101.830 -76.558 45.170
251889 92.842 -74.184 54.604
251890 92.871 -74.203 54.605
251891 92.901 -74.222 54.605
251892 92.931 -74.240 54.605
251893 92.962 -74.242 54.605
251894 92.993 -74.242 54.603
251895 93.024 -74.244 54.601
251896 93.055 -74.245 54.600
251897 93.086 -74.246 54.598
251898 93.117 -74.247 54.597
251899 93.148 -74.248 54.597
251900 93.180 -74.249 54.596
251901 93.210 -74.250 54.596
251902 93.241 -74.252 54.599
251903 93.274 -74.243 54.601
251904 93.307 -74.223 54.606
251905 93.332 -74.223 54.615
251906 93.353 -74.232 54.628
251907 93.374 -74.240 54.642
251908 93.396 -74.245 54.659
251909 93.420 -74.241 54.679
251910 93.444 -74.236 54.703
251911 93.469 -74.233 54.727
251912 93.493 -74.228 54.752
251913 93.518 -74.224 54.778
251914 93.542 -74.220 54.804
251915 93.566 -74.216 54.829
251916 93.591 -74.212 54.850
251917 93.615 -74.208 54.867
251918 93.639 -74.204 54.882
251919 93.661 -74.200 54.902
251920 93.683 -74.196 54.922
251921 93.704 -74.192 54.942
251922 93.725 -74.188 54.962
251923 93.747 -74.184 54.972
251924 93.768 -74.180 54.974
251925 93.789 -74.176 54.970
251926 93.811 -74.173 54.965
251927 93.832 -74.169 54.949
251928 93.853 -74.165 54.927
251929 93.875 -74.164 54.900
251930 93.897 -74.165 54.873
251931 93.918 -74.166 54.845
251932 93.939 -74.167 54.819
251933 93.961 -74.167 54.793
251934 93.982 -74.168 54.759
251935 94.004 -74.170 54.726
251936 94.026 -74.170 54.694
251937 94.049 -74.174 54.663
251938 94.077 -74.179 54.634
251939 94.124 -74.189 54.606
251940 94.165 -74.196 54.580
251941 94.182 -74.195 54.558
251942 94.200 -74.195 54.552
251943 94.214 -74.193 54.548
251944 94.228 -74.191 54.542
251945 94.243 -74.189 54.532
251946 94.261 -74.193 54.517
251947 94.281 -74.198 54.497
251948 94.302 -74.201 54.477
251949 94.322 -74.205 54.457
251950 94.342 -74.208 54.435
251951 94.364 -74.208 54.411
251952 94.390 -74.203 54.387
251953 94.415 -74.198 54.363
251954 94.440 -74.193 54.335
251955 94.465 -74.183 54.306
251956 94.488 -74.167 54.275
251957 94.511 -74.152 54.242
251958 94.535 -74.139 54.207
251959 94.559 -74.137 54.169
251960 94.583 -74.136 54.130
251961 94.608 -74.134 54.087
251962 94.628 -74.139 54.041
251963 94.644 -74.148 53.994
251964 94.660 -74.147 53.946
251965 94.675 -74.147 53.898
251966 94.690 -74.146 53.850
251967 94.705 -74.146 53.802
251968 94.720 -74.145 53.753
251969 94.735 -74.145 53.704
251970 94.751 -74.147 53.655
251971 94.770 -74.159 53.608
251972 94.787 -74.179 53.566
251973 94.799 -74.214 53.529
251974 94.813 -74.245 53.501
251975 94.827 -74.276 53.480
251976 94.841 -74.307 53.460
251977 94.869 -74.300 53.437
251978 94.905 -74.274 53.403
251979 94.938 -74.264 53.366
251980 94.971 -74.267 53.328
251981 95.003 -74.270 53.286
251982 95.036 -74.274 53.237
251983 95.068 -74.277 53.185
251984 95.101 -74.281 53.132
251985 95.133 -74.290 53.081
251986 95.166 -74.304 53.029
251987 95.198 -74.317 52.975
251988 95.231 -74.332 52.921
251989 95.264 -74.346 52.867
251990 95.296 -74.359 52.810
251991 95.332 -74.363 52.752
251992 95.371 -74.356 52.695
251993 95.410 -74.350 52.639
251994 95.449 -74.343 52.585
251995 95.488 -74.337 52.532
251996 95.523 -74.335 52.482
251997 95.550 -74.337 52.435
251998 95.579 -74.336 52.389
251999 95.608 -74.336 52.344
252000 95.636 -74.336 52.301
252001 95.666 -74.337 52.260
252002 95.695 -74.343 52.220
252003 95.725 -74.350 52.180
252004 95.754 -74.356 52.141
252005 95.783 -74.362 52.101
252006 95.812 -74.368 52.061
252007 95.842 -74.374 52.020
252008 95.867 -74.379 51.980
252009 95.877 -74.376 51.940
252010 95.887 -74.372 51.899
252011 95.896 -74.368 51.859
252012 95.910 -74.372 51.819
252013 95.934 -74.392 51.777
252014 95.958 -74.411 51.736
252015 95.981 -74.431 51.695
252016 96.005 -74.451 51.654
252017 96.029 -74.471 51.613
252018 96.058 -74.490 51.574
252019 96.090 -74.510 51.535
252020 96.122 -74.530 51.496
252021 96.160 -74.544 51.459
252022 96.204 -74.552 51.425
252023 96.249 -74.559 51.391
252024 96.257 -74.574 51.358
252025 96.283 -74.556 51.326
252026 96.309 -74.538 51.297
252027 96.335 -74.521 51.270
252028 96.363 -74.512 51.244
252029 96.397 -74.530 51.219
252030 96.432 -74.549 51.196
252031 96.466 -74.568 51.175
252032 96.493 -74.578 51.154
252033 96.519 -74.586 51.136
252034 96.543 -74.593 51.119
252035 96.569 -74.600 51.104
252036 96.593 -74.608 51.092
252037 96.618 -74.615 51.081
252038 96.643 -74.622 51.072
252039 96.668 -74.630 51.065
252040 96.693 -74.637 51.058
252041 96.718 -74.644 51.055
252042 96.743 -74.652 51.052
252043 96.768 -74.659 51.050
252044 96.795 -74.670 51.049
252045 96.825 -74.688 51.050
252046 96.856 -74.707 51.047
252047 96.883 -74.718 51.042
252048 96.908 -74.726 51.035
252049 96.934 -74.734 51.027
252050 96.959 -74.742 51.020
252051 96.985 -74.750 51.013
252052 97.001 -74.760 51.002
252053 97.014 -74.770 50.987
252054 97.026 -74.781 50.973
252055 97.035 -74.792 50.959
252056 97.045 -74.803 50.947
252057 97.055 -74.814 50.937
252058 97.067 -74.822 50.928
252059 97.083 -74.824 50.921
252060 97.099 -74.825 50.915
252061 97.115 -74.828 50.911
252062 97.124 -74.833 50.907
252063 97.132 -74.838 50.900
252064 97.154 -74.832 50.888
252065 97.185 -74.828 50.867
252066 97.209 -74.837 50.844
252067 97.236 -74.844 50.817
252068 97.273 -74.847 50.789
252069 97.310 -74.850 50.759
252070 97.347 -74.853 50.730
252071 97.384 -74.856 50.701
252072 97.422 -74.859 50.673
252073 97.458 -74.861 50.645
252074 97.496 -74.864 50.615
252075 97.533 -74.867 50.584
252076 97.570 -74.870 50.553
252077 97.606 -74.874 50.518
252078 97.637 -74.882 50.479
252079 97.669 -74.891 50.440
252080 97.701 -74.898 50.400
252081 97.732 -74.907 50.357
252082 97.764 -74.915 50.314
252083 97.795 -74.922 50.270
252084 97.821 -74.932 50.225
252085 97.844 -74.944 50.180
252086 97.867 -74.955 50.132
252087 97.890 -74.966 50.081
252088 97.913 -74.977 50.027
252089 97.944 -74.984 49.973
252090 97.982 -74.988 49.920
252091 98.012 -74.998 49.869
252092 98.042 -75.008 49.819
252093 98.073 -75.018 49.771
252094 98.103 -75.028 49.720
252095 98.134 -75.038 49.666
252096 98.165 -75.050 49.609
252097 98.196 -75.063 49.551
252098 98.227 -75.076 49.494
252099 98.258 -75.088 49.439
252100 98.290 -75.102 49.387
252101 98.321 -75.115 49.336
252102 98.353 -75.129 49.285
252103 98.394 -75.154 49.234
252104 98.435 -75.179 49.186
252105 98.465 -75.160 49.137
252106 98.485 -75.137 49.089
252107 98.505 -75.115 49.040
252108 98.524 -75.093 48.992
252109 98.545 -75.071 48.940
252110 98.565 -75.054 48.888
252111 98.593 -75.065 48.837
252112 98.620 -75.076 48.784
252113 98.647 -75.087 48.731
252114 98.674 -75.093 48.679
252115 98.702 -75.098 48.625
252116 98.728 -75.102 48.571
252117 98.756 -75.107 48.517
252118 98.782 -75.111 48.463
252119 98.809 -75.116 48.410
252120 98.837 -75.120 48.359
252121 98.863 -75.125 48.307
252122 98.889 -75.150 48.256
252123 98.915 -75.180 48.204
252124 98.940 -75.210 48.155
252125 98.966 -75.241 48.106
252126 98.991 -75.271 48.058
252127 99.016 -75.301 48.011
252128 99.042 -75.332 47.966
252129 99.067 -75.358 47.920
252130 99.091 -75.377 47.875
252131 99.124 -75.395 47.830
252132 99.154 -75.399 47.787
252133 99.184 -75.404 47.743
252134 99.214 -75.408 47.701
252135 99.244 -75.412 47.660
252136 99.274 -75.416 47.618
252137 99.304 -75.420 47.577
252138 99.334 -75.424 47.537
252139 99.364 -75.428 47.499
252140 99.394 -75.432 47.464
252141 99.419 -75.440 47.432
252142 99.443 -75.449 47.401
252143 99.468 -75.457 47.372
252144 99.493 -75.464 47.345
252145 99.517 -75.473 47.319
252146 99.540 -75.488 47.293
252147 99.566 -75.505 47.268
252148 99.592 -75.522 47.245
252149 99.618 -75.540 47.223
252150 99.644 -75.557 47.201
252151 99.670 -75.574 47.180
252152 99.696 -75.591 47.159
252153 99.722 -75.609 47.137
252154 99.748 -75.626 47.117
252155 99.774 -75.643 47.098
252156 99.800 -75.660 47.080
252157 99.826 -75.678 47.061
252158 99.848 -75.708 47.043
252159 99.870 -75.737 47.024
252160 99.893 -75.767 47.001
252161 99.932 -75.779 46.977
252162 99.969 -75.784 46.953
252163 99.974 -75.802 46.928
252164 99.988 -75.815 46.899
252165 100.018 -75.823 46.869
252166 100.049 -75.831 46.837
252167 100.079 -75.840 46.803
252168 100.109 -75.847 46.770
252169 100.139 -75.856 46.737
252170 100.169 -75.864 46.703
252171 100.199 -75.872 46.669
252172 100.229 -75.880 46.635
252173 100.258 -75.899 46.601
252174 100.285 -75.921 46.567
252175 100.314 -75.943 46.535
252176 100.341 -75.965 46.504
252177 100.369 -75.987 46.472
252178 100.400 -75.984 46.441
252179 100.432 -75.974 46.409
252180 100.465 -75.964 46.379
252181 100.497 -75.954 46.349
252182 100.525 -75.974 46.318
252183 100.552 -76.003 46.289
252184 100.579 -76.032 46.257
252185 100.606 -76.062 46.225
252186 100.633 -76.091 46.194
252187 100.661 -76.121 46.160
252188 100.687 -76.151 46.128
252189 100.714 -76.180 46.096
252190 100.742 -76.210 46.064
252191 100.770 -76.235 46.031
252192 100.804 -76.252 45.999
252193 100.836 -76.270 45.969
252194 100.859 -76.282 45.939
252195 100.876 -76.293 45.909
252196 100.883 -76.295 45.881
252197 100.890 -76.298 45.853
252198 100.896 -76.300 45.825
252199 100.918 -76.306 45.798
252200 100.960 -76.317 45.772
252201 101.001 -76.329 45.746
252202 101.042 -76.340 45.721
252203 101.084 -76.351 45.697
252204 101.125 -76.362 45.672
252205 101.167 -76.373 45.647
252206 101.208 -76.385 45.623
252207 101.250 -76.396 45.599
252208 101.291 -76.407 45.575
252209 101.333 -76.418 45.552
252210 101.386 -76.436 45.528
252211 101.443 -76.454 45.506
252212 101.492 -76.457 45.483
252213 101.537 -76.451 45.461
252214 101.582 -76.446 45.438
252215 101.627 -76.440 45.417
252216 101.662 -76.442 45.395
252217 101.687 -76.451 45.374
252218 101.708 -76.462 45.352
252219 101.723 -76.474 45.331
252220 101.738 -76.486 45.309
252221 101.752 -76.499 45.288
252222 101.773 -76.500 45.266
252223 101.793 -76.505 45.245
252224 101.809 -76.515 45.224
252225 101.825 -76.527 45.204
252226 101.841 -76.538 45.183
252227 101.850 -76.542 45.162
252589 92.813 -74.139 54.609
252590 92.843 -74.158 54.610
252591 92.873 -74.176 54.611
252592 92.902 -74.195 54.611
252593 92.932 -74.214 54.612
252594 92.963 -74.220 54.612
252595 92.994 -74.221 54.611
252596 93.025 -74.222 54.609
252597 93.056 -74.223 54.608
252598 93.087 -74.224 54.607
252599 93.118 -74.226 54.607
252600 93.149 -74.226 54.607
252601 93.180 -74.228 54.607
252602 93.211 -74.229 54.607
252603 93.243 -74.224 54.608
252604 93.277 -74.204 54.610
252605 93.306 -74.192 54.613
252606 93.328 -74.199 54.620
252607 93.349 -74.208 54.631
252608 93.370 -74.216 54.643
252609 93.392 -74.221 54.657
252610 93.416 -74.217 54.674
252611 93.441 -74.214 54.694
252612 93.465 -74.209 54.717
252613 93.489 -74.205 54.741
252614 93.514 -74.201 54.767
252615 93.538 -74.197 54.793
252616 93.562 -74.193 54.816
252617 93.587 -74.189 54.834
252618 93.611 -74.185 54.851
252619 93.636 -74.180 54.867
252620 93.660 -74.176 54.891
252621 93.684 -74.173 54.917
252622 93.706 -74.168 54.942
252623 93.727 -74.164 54.961
252624 93.749 -74.161 54.969
252625 93.770 -74.157 54.972
252626 93.791 -74.153 54.972
252627 93.813 -74.149 54.968
252628 93.834 -74.145 54.953
252629 93.855 -74.142 54.932
252630 93.877 -74.143 54.907
252631 93.899 -74.144 54.881
252632 93.920 -74.145 54.854
252633 93.941 -74.145 54.828
252634 93.963 -74.146 54.802
252635 93.984 -74.148 54.765
252636 94.006 -74.148 54.729
252637 94.027 -74.149 54.696
252638 94.049 -74.150 54.663
252639 94.080 -74.155 54.631
252640 94.127 -74.165 54.601
252641 94.171 -74.174 54.573
252642 94.185 -74.173 54.550
252643 94.199 -74.171 54.545
252644 94.213 -74.169 54.541
252645 94.227 -74.167 54.535
252646 94.243 -74.167 54.525
252647 94.263 -74.173 54.509
252648 94.283 -74.177 54.488
252649 94.303 -74.181 54.467
252650 94.324 -74.185 54.447
252651 94.344 -74.188 54.425
252652 94.366 -74.187 54.401
252653 94.392 -74.183 54.377
252654 94.417 -74.177 54.351
252655 94.443 -74.172 54.323
252656 94.468 -74.167 54.293
252657 94.492 -74.152 54.262
252658 94.515 -74.137 54.229
252659 94.538 -74.122 54.194
252660 94.561 -74.107 54.153
252661 94.585 -74.103 54.110
252662 94.610 -74.101 54.066
252663 94.632 -74.104 54.023
252664 94.648 -74.111 53.977
252665 94.663 -74.110 53.930
252666 94.678 -74.110 53.883
252667 94.693 -74.109 53.835
252668 94.708 -74.109 53.786
252669 94.723 -74.108 53.737
252670 94.739 -74.108 53.686
252671 94.754 -74.107 53.638
252672 94.772 -74.117 53.593
252673 94.786 -74.144 53.552
252674 94.800 -74.175 53.518
252675 94.814 -74.206 53.493
252676 94.829 -74.237 53.474
252677 94.843 -74.268 53.455
252678 94.873 -74.258 53.430
252679 94.908 -74.236 53.395
252680 94.940 -74.240 53.356
252681 94.972 -74.243 53.316
252682 95.005 -74.247 53.272
252683 95.037 -74.250 53.223
252684 95.070 -74.254 53.170
252685 95.102 -74.258 53.117
252686 95.135 -74.261 53.064
252687 95.167 -74.273 53.013
252688 95.200 -74.287 52.960
252689 95.233 -74.301 52.906
252690 95.265 -74.315 52.853
252691 95.298 -74.329 52.798
252692 95.336 -74.324 52.742
252693 95.375 -74.318 52.686
252694 95.414 -74.311 52.630
252695 95.453 -74.305 52.576
252696 95.492 -74.299 52.524
252697 95.531 -74.292 52.474
252698 95.560 -74.293 52.428
252699 95.589 -74.292 52.383
252700 95.618 -74.297 52.339
252701 95.648 -74.303 52.297
252702 95.677 -74.310 52.257
252703 95.706 -74.316 52.217
252704 95.736 -74.322 52.179
252705 95.765 -74.328 52.140
252706 95.795 -74.334 52.101
252707 95.824 -74.341 52.061
252708 95.853 -74.347 52.021
252709 95.865 -74.344 51.981
252710 95.874 -74.341 51.940
252711 95.884 -74.337 51.900
252712 95.894 -74.333 51.860
252713 95.904 -74.331 51.819
252714 95.928 -74.352 51.778
252715 95.952 -74.371 51.737
252716 95.975 -74.391 51.697
252717 95.999 -74.411 51.656
252718 96.023 -74.430 51.616
252719 96.046 -74.450 51.576
252720 96.071 -74.470 51.537
252721 96.111 -74.481 51.498
252722 96.155 -74.489 51.461
252723 96.200 -74.496 51.426
252724 96.244 -74.504 51.392
252725 96.251 -74.521 51.358
252726 96.269 -74.515 51.326
252727 96.295 -74.498 51.296
252728 96.321 -74.480 51.268
252729 96.350 -74.476 51.242
252730 96.385 -74.495 51.218
252731 96.419 -74.514 51.194
252732 96.453 -74.533 51.172
252733 96.488 -74.552 51.151
252734 96.516 -74.562 51.132
252735 96.540 -74.570 51.115
252736 96.566 -74.577 51.099
252737 96.591 -74.584 51.086
252738 96.615 -74.592 51.074
252739 96.641 -74.599 51.064
252740 96.665 -74.606 51.056
252741 96.690 -74.614 51.049
252742 96.715 -74.621 51.044
252743 96.740 -74.628 51.040
252744 96.765 -74.636 51.038
252745 96.793 -74.649 51.036
252746 96.823 -74.667 51.035
252747 96.854 -74.686 51.034
252748 96.882 -74.700 51.032
252749 96.907 -74.707 51.027
252750 96.933 -74.715 51.021
252751 96.959 -74.724 51.018
252752 96.982 -74.732 51.012
252753 96.996 -74.744 51.004
252754 97.006 -74.754 50.991
252755 97.016 -74.766 50.979
252756 97.025 -74.777 50.968
252757 97.035 -74.787 50.957
252758 97.050 -74.791 50.949
252759 97.067 -74.793 50.940
252760 97.083 -74.794 50.931
252761 97.099 -74.797 50.924
252762 97.115 -74.798 50.921
252763 97.131 -74.800 50.916
252764 97.143 -74.804 50.909
252765 97.158 -74.803 50.898
252766 97.186 -74.805 50.877
252767 97.217 -74.810 50.853
252768 97.254 -74.813 50.827
252769 97.291 -74.816 50.797
252770 97.328 -74.819 50.766
252771 97.365 -74.822 50.735
252772 97.402 -74.825 50.705
252773 97.439 -74.828 50.677
252774 97.476 -74.831 50.650
252775 97.514 -74.834 50.621
252776 97.550 -74.837 50.590
252777 97.588 -74.840 50.558
252778 97.625 -74.843 50.521
252779 97.657 -74.850 50.482
252780 97.688 -74.859 50.442
252781 97.720 -74.866 50.400
252782 97.751 -74.875 50.356
252783 97.783 -74.882 50.311
252784 97.809 -74.892 50.267
252785 97.832 -74.903 50.222
252786 97.855 -74.915 50.176
252787 97.878 -74.925 50.127
252788 97.904 -74.939 50.076
252789 97.929 -74.949 50.022
252790 97.952 -74.956 49.970
252791 97.990 -74.960 49.918
252792 98.020 -74.969 49.868
252793 98.051 -74.979 49.820
252794 98.082 -74.989 49.773
252795 98.112 -75.001 49.722
252796 98.144 -75.013 49.668
252797 98.174 -75.026 49.610
252798 98.206 -75.039 49.551
252799 98.237 -75.053 49.495
252800 98.268 -75.066 49.441
252801 98.299 -75.079 49.389
252802 98.330 -75.091 49.337
252803 98.364 -75.108 49.285
252804 98.405 -75.133 49.235
252805 98.443 -75.146 49.186
252806 98.472 -75.118 49.137
252807 98.491 -75.096 49.089
252808 98.511 -75.073 49.039
252809 98.531 -75.051 48.988
252810 98.551 -75.029 48.936
252811 98.575 -75.027 48.884
252812 98.602 -75.038 48.832
252813 98.630 -75.049 48.778
252814 98.657 -75.060 48.726
252815 98.685 -75.070 48.673
252816 98.712 -75.075 48.620
252817 98.738 -75.079 48.566
252818 98.766 -75.084 48.512
252819 98.792 -75.088 48.458
252820 98.819 -75.093 48.406
252821 98.846 -75.098 48.354
252822 98.873 -75.108 48.303
252823 98.899 -75.138 48.252
252824 98.924 -75.169 48.201
252825 98.949 -75.198 48.152
252826 98.975 -75.229 48.103
252827 99.000 -75.259 48.055
252828 99.025 -75.289 48.009
252829 99.050 -75.311 47.964
252830 99.075 -75.329 47.918
252831 99.099 -75.347 47.872
252832 99.134 -75.367 47.826
252833 99.166 -75.377 47.782
252834 99.196 -75.380 47.738
252835 99.226 -75.384 47.695
252836 99.257 -75.389 47.653
252837 99.286 -75.392 47.611
252838 99.316 -75.396 47.571
252839 99.347 -75.401 47.531
252840 99.376 -75.405 47.491
252841 99.405 -75.410 47.457
252842 99.430 -75.418 47.424
252843 99.454 -75.427 47.393
252844 99.479 -75.435 47.364
252845 99.503 -75.442 47.337
252846 99.527 -75.451 47.311
252847 99.558 -75.457 47.286
252848 99.580 -75.475 47.261
252849 99.607 -75.493 47.238
252850 99.632 -75.509 47.216
252851 99.658 -75.527 47.195
252852 99.685 -75.544 47.173
252853 99.711 -75.561 47.152
252854 99.736 -75.578 47.131
252855 99.763 -75.596 47.111
252856 99.789 -75.613 47.091
252857 99.815 -75.630 47.074
252858 99.841 -75.648 47.056
252859 99.863 -75.677 47.039
252860 99.885 -75.706 47.019
252861 99.916 -75.727 46.997
252862 99.972 -75.724 46.973
252863 99.991 -75.736 46.948
252864 99.996 -75.753 46.923
252865 100.002 -75.770 46.895
252866 100.032 -75.778 46.864
252867 100.062 -75.787 46.832
252868 100.092 -75.794 46.797
252869 100.123 -75.803 46.763
252870 100.152 -75.811 46.730
252871 100.183 -75.818 46.696
252872 100.213 -75.827 46.662
252873 100.242 -75.840 46.628
252874 100.270 -75.862 46.594
252875 100.298 -75.884 46.560
252876 100.326 -75.906 46.529
252877 100.353 -75.928 46.496
252878 100.382 -75.950 46.465
252879 100.409 -75.971 46.434
252880 100.441 -75.961 46.403
252881 100.474 -75.952 46.372
252882 100.506 -75.942 46.343
252883 100.534 -75.963 46.313
252884 100.561 -75.993 46.283
252885 100.588 -76.022 46.252
252886 100.615 -76.052 46.221
252887 100.642 -76.081 46.190
252888 100.669 -76.110 46.156
252889 100.696 -76.140 46.125
252890 100.724 -76.166 46.093
252891 100.758 -76.184 46.061
252892 100.790 -76.201 46.029
252893 100.823 -76.219 45.998
252894 100.855 -76.235 45.967
252895 100.872 -76.246 45.936
252896 100.890 -76.257 45.907
252897 100.901 -76.262 45.878
252898 100.907 -76.264 45.850
252899 100.913 -76.267 45.822
252900 100.944 -76.275 45.795
252901 100.985 -76.286 45.769
252902 101.027 -76.298 45.742
252903 101.068 -76.309 45.717
252904 101.110 -76.320 45.693
252905 101.151 -76.331 45.667
252906 101.193 -76.342 45.642
252907 101.234 -76.354 45.618
252908 101.276 -76.365 45.594
252909 101.317 -76.376 45.570
252910 101.358 -76.387 45.547
252911 101.400 -76.398 45.524
252912 101.454 -76.415 45.500
252913 101.510 -76.434 45.477
252914 101.563 -76.443 45.455
252915 101.608 -76.438 45.433
252916 101.653 -76.432 45.411
252917 101.693 -76.430 45.390
252918 101.718 -76.439 45.368
252919 101.736 -76.451 45.345
252920 101.751 -76.462 45.324
252921 101.766 -76.474 45.303
252922 101.787 -76.475 45.281
252923 101.804 -76.484 45.259
252924 101.821 -76.495 45.238
252925 101.837 -76.507 45.216
252926 101.836 -76.505 45.195
252927 101.829 -76.498 45.175
252928 101.822 -76.491 45.154
253289 92.785 -74.095 54.613
253290 92.814 -74.114 54.614
253291 92.844 -74.132 54.616
253292 92.874 -74.151 54.617
253293 92.904 -74.170 54.618
253294 92.933 -74.188 54.619
253295 92.964 -74.198 54.619
253296 92.995 -74.200 54.619
253297 93.025 -74.201 54.617
253298 93.057 -74.202 54.617
253299 93.088 -74.203 54.617
253300 93.119 -74.204 54.617
253301 93.150 -74.205 54.617
253302 93.181 -74.206 54.617
253303 93.212 -74.205 54.617
253304 93.246 -74.185 54.619
253305 93.280 -74.164 54.619
253306 93.303 -74.167 54.623
253307 93.324 -74.175 54.628
253308 93.345 -74.183 54.636
253309 93.366 -74.191 54.645
253310 93.388 -74.198 54.656
253311 93.412 -74.195 54.670
253312 93.437 -74.190 54.686
253313 93.461 -74.186 54.707
253314 93.485 -74.182 54.730
253315 93.510 -74.178 54.754
253316 93.534 -74.174 54.779
253317 93.559 -74.170 54.798
253318 93.583 -74.166 54.815
253319 93.607 -74.161 54.832
253320 93.632 -74.157 54.853
253321 93.656 -74.154 54.883
253322 93.680 -74.149 54.914
253323 93.705 -74.145 54.941
253324 93.729 -74.141 54.958
253325 93.751 -74.137 54.963
253326 93.772 -74.133 54.968
253327 93.793 -74.129 54.974
253328 93.815 -74.126 54.971
253329 93.836 -74.122 54.958
253330 93.857 -74.121 54.939
253331 93.879 -74.122 54.916
253332 93.900 -74.123 54.892
253333 93.922 -74.123 54.865
253334 93.943 -74.124 54.838
253335 93.965 -74.126 54.809
253336 93.986 -74.126 54.771
253337 94.008 -74.127 54.733
253338 94.029 -74.129 54.696
253339 94.050 -74.131 54.659
253340 94.087 -74.139 54.624
253341 94.136 -74.151 54.592
253342 94.159 -74.150 54.564
253343 94.182 -74.148 54.542
253344 94.198 -74.147 54.536
253345 94.212 -74.145 54.531
253346 94.225 -74.144 54.524
253347 94.245 -74.147 54.514
253348 94.265 -74.152 54.497
253349 94.285 -74.157 54.475
253350 94.305 -74.161 54.452
253351 94.325 -74.165 54.431
253352 94.346 -74.168 54.409
253353 94.369 -74.167 54.386
253354 94.395 -74.161 54.361
253355 94.420 -74.156 54.334
253356 94.445 -74.151 54.303
253357 94.471 -74.145 54.273
253358 94.495 -74.136 54.243
253359 94.519 -74.122 54.210
253360 94.542 -74.107 54.172
253361 94.565 -74.092 54.131
253362 94.589 -74.076 54.087
253363 94.613 -74.069 54.042
253364 94.636 -74.069 54.001
253365 94.651 -74.073 53.956
253366 94.667 -74.073 53.910
253367 94.682 -74.072 53.863
253368 94.697 -74.072 53.815
253369 94.712 -74.071 53.766
253370 94.728 -74.071 53.716
253371 94.743 -74.070 53.665
253372 94.758 -74.070 53.619
253373 94.774 -74.074 53.577
253374 94.788 -74.106 53.540
253375 94.802 -74.137 53.509
253376 94.816 -74.168 53.487
253377 94.830 -74.199 53.471
253378 94.844 -74.230 53.452
253379 94.876 -74.215 53.424
253380 94.909 -74.213 53.387
253381 94.942 -74.217 53.346
253382 94.974 -74.220 53.303
253383 95.006 -74.223 53.258
253384 95.039 -74.226 53.207
253385 95.071 -74.230 53.154
253386 95.104 -74.233 53.101
253387 95.136 -74.237 53.048
253388 95.169 -74.242 52.996
253389 95.201 -74.257 52.945
253390 95.234 -74.270 52.892
253391 95.266 -74.285 52.839
253392 95.301 -74.292 52.786
253393 95.339 -74.286 52.732
253394 95.379 -74.280 52.677
253395 95.418 -74.273 52.622
253396 95.457 -74.267 52.567
253397 95.496 -74.260 52.516
253398 95.534 -74.255 52.468
253399 95.571 -74.257 52.421
253400 95.600 -74.263 52.378
253401 95.630 -74.269 52.335
253402 95.659 -74.276 52.293
253403 95.688 -74.282 52.252
253404 95.718 -74.288 52.214
253405 95.747 -74.294 52.176
253406 95.776 -74.301 52.139
253407 95.806 -74.307 52.101
253408 95.835 -74.313 52.060
253409 95.852 -74.313 52.021
253410 95.862 -74.309 51.982
253411 95.872 -74.306 51.941
253412 95.881 -74.302 51.901
253413 95.891 -74.299 51.861
253414 95.901 -74.295 51.820
253415 95.922 -74.311 51.779
253416 95.946 -74.331 51.739
253417 95.969 -74.351 51.699
253418 95.993 -74.371 51.659
253419 96.017 -74.390 51.618
253420 96.040 -74.410 51.579
253421 96.066 -74.423 51.539
253422 96.106 -74.427 51.500
253423 96.151 -74.434 51.462
253424 96.195 -74.442 51.427
253425 96.239 -74.449 51.393
253426 96.246 -74.466 51.359
253427 96.255 -74.475 51.327
253428 96.282 -74.458 51.296
253429 96.308 -74.440 51.268
253430 96.338 -74.440 51.241
253431 96.372 -74.459 51.216
253432 96.406 -74.478 51.192
253433 96.441 -74.497 51.170
253434 96.475 -74.516 51.148
253435 96.509 -74.535 51.128
253436 96.538 -74.546 51.111
253437 96.562 -74.553 51.095
253438 96.588 -74.561 51.080
253439 96.612 -74.568 51.067
253440 96.638 -74.575 51.057
253441 96.662 -74.583 51.047
253442 96.687 -74.590 51.039
253443 96.713 -74.597 51.033
253444 96.737 -74.605 51.029
253445 96.762 -74.612 51.025
253446 96.791 -74.628 51.023
253447 96.822 -74.646 51.022
253448 96.852 -74.665 51.021
253449 96.881 -74.681 51.020
253450 96.907 -74.689 51.018
253451 96.933 -74.697 51.015
253452 96.954 -74.707 51.014
253453 96.975 -74.717 51.011
253454 96.986 -74.728 51.004
253455 96.996 -74.739 50.993
253456 97.006 -74.750 50.984
253457 97.018 -74.758 50.975
253458 97.034 -74.759 50.968
253459 97.050 -74.762 50.961
253460 97.066 -74.763 50.952
253461 97.082 -74.765 50.942
253462 97.098 -74.767 50.935
253463 97.114 -74.769 50.931
253464 97.131 -74.771 50.925
253465 97.147 -74.772 50.917
253466 97.162 -74.775 50.906
253467 97.197 -74.778 50.885
253468 97.235 -74.780 50.861
253469 97.272 -74.783 50.833
253470 97.309 -74.786 50.803
253471 97.346 -74.789 50.771
253472 97.383 -74.792 50.740
253473 97.420 -74.795 50.710
253474 97.457 -74.797 50.681
253475 97.495 -74.800 50.655
253476 97.531 -74.803 50.626
253477 97.569 -74.806 50.594
253478 97.606 -74.809 50.560
253479 97.643 -74.812 50.522
253480 97.676 -74.819 50.482
253481 97.708 -74.826 50.439
253482 97.739 -74.834 50.396
253483 97.771 -74.843 50.351
253484 97.798 -74.853 50.305
253485 97.820 -74.863 50.259
253486 97.844 -74.875 50.214
253487 97.869 -74.888 50.166
253488 97.896 -74.902 50.117
253489 97.921 -74.912 50.067
253490 97.944 -74.919 50.014
253491 97.967 -74.925 49.965
253492 97.998 -74.931 49.916
253493 98.029 -74.941 49.868
253494 98.060 -74.951 49.821
253495 98.091 -74.964 49.774
253496 98.122 -74.977 49.722
253497 98.153 -74.991 49.667
253498 98.184 -75.004 49.610
253499 98.215 -75.016 49.551
253500 98.246 -75.029 49.495
253501 98.278 -75.043 49.441
253502 98.308 -75.056 49.389
253503 98.340 -75.069 49.336
253504 98.375 -75.087 49.284
253505 98.416 -75.113 49.234
253506 98.450 -75.104 49.184
253507 98.478 -75.076 49.136
253508 98.498 -75.054 49.087
253509 98.517 -75.032 49.036
253510 98.537 -75.010 48.982
253511 98.557 -74.988 48.931
253512 98.585 -75.000 48.879
253513 98.612 -75.011 48.826
253514 98.640 -75.022 48.772
253515 98.667 -75.033 48.719
253516 98.695 -75.044 48.665
253517 98.722 -75.052 48.612
253518 98.748 -75.057 48.560
253519 98.776 -75.061 48.506
253520 98.802 -75.066 48.453
253521 98.830 -75.070 48.400
253522 98.856 -75.075 48.349
253523 98.882 -75.096 48.297
253524 98.908 -75.126 48.246
253525 98.933 -75.157 48.196
253526 98.958 -75.187 48.147
253527 98.984 -75.217 48.099
253528 99.009 -75.245 48.052
253529 99.034 -75.264 48.006
253530 99.059 -75.282 47.961
253531 99.083 -75.300 47.915
253532 99.109 -75.318 47.868
253533 99.144 -75.338 47.822
253534 99.179 -75.353 47.777
253535 99.209 -75.357 47.731
253536 99.239 -75.361 47.687
253537 99.269 -75.365 47.644
253538 99.299 -75.369 47.603
253539 99.329 -75.373 47.563
253540 99.359 -75.377 47.523
253541 99.389 -75.381 47.485
253542 99.416 -75.388 47.449
253543 99.440 -75.396 47.416
253544 99.465 -75.405 47.385
253545 99.489 -75.413 47.357
253546 99.514 -75.420 47.329
253547 99.542 -75.427 47.304
253548 99.578 -75.431 47.279
253549 99.611 -75.437 47.254
253550 99.627 -75.461 47.231
253551 99.650 -75.480 47.209
253552 99.674 -75.497 47.187
253553 99.699 -75.514 47.167
253554 99.725 -75.531 47.146
253555 99.751 -75.549 47.124
253556 99.777 -75.566 47.104
253557 99.803 -75.583 47.086
253558 99.830 -75.600 47.068
253559 99.855 -75.618 47.051
253560 99.878 -75.646 47.034
253561 99.901 -75.674 47.015
253562 99.956 -75.672 46.991
253563 100.008 -75.671 46.968
253564 100.013 -75.688 46.944
253565 100.018 -75.706 46.918
253566 100.022 -75.722 46.889
253567 100.046 -75.733 46.858
253568 100.076 -75.741 46.825
253569 100.106 -75.749 46.790
253570 100.136 -75.757 46.756
253571 100.167 -75.765 46.722
253572 100.196 -75.774 46.688
253573 100.226 -75.782 46.655
253574 100.254 -75.804 46.621
253575 100.282 -75.826 46.586
253576 100.310 -75.848 46.552
253577 100.338 -75.870 46.521
253578 100.366 -75.892 46.489
253579 100.394 -75.914 46.458
253580 100.421 -75.936 46.427
253581 100.451 -75.949 46.398
253582 100.483 -75.939 46.367
253583 100.515 -75.929 46.338
253584 100.543 -75.953 46.309
253585 100.570 -75.982 46.279
253586 100.597 -76.012 46.248
253587 100.624 -76.041 46.218
253588 100.651 -76.071 46.186
253589 100.679 -76.098 46.153
253590 100.712 -76.116 46.122
253591 100.745 -76.133 46.090
253592 100.777 -76.151 46.059
253593 100.810 -76.167 46.027
253594 100.843 -76.185 45.996
253595 100.868 -76.198 45.965
253596 100.885 -76.209 45.934
253597 100.902 -76.220 45.905
253598 100.918 -76.229 45.876
253599 100.924 -76.231 45.848
253600 100.931 -76.233 45.820
253601 100.969 -76.244 45.793
253602 101.011 -76.255 45.765
253603 101.052 -76.267 45.739
253604 101.094 -76.277 45.714
253605 101.135 -76.289 45.688
253606 101.177 -76.300 45.663
253607 101.218 -76.311 45.638
253608 101.260 -76.323 45.614
253609 101.302 -76.333 45.589
253610 101.343 -76.345 45.565
253611 101.385 -76.356 45.541
253612 101.426 -76.367 45.518
253613 101.467 -76.379 45.495
253614 101.521 -76.395 45.472
253615 101.578 -76.414 45.449
253616 101.633 -76.429 45.428
253617 101.678 -76.423 45.406
253618 101.723 -76.418 45.384
253619 101.749 -76.426 45.362
253620 101.764 -76.439 45.339
253621 101.779 -76.451 45.317
253622 101.799 -76.453 45.296
253623 101.816 -76.464 45.275
253624 101.830 -76.473 45.253
253625 101.823 -76.467 45.231
253626 101.816 -76.460 45.209
253627 101.809 -76.453 45.187
253628 101.802 -76.446 45.166
253629 101.806 -76.446 45.145
253989 92.756 -74.070 54.616
253990 92.786 -74.069 54.618
253991 92.816 -74.087 54.619
253992 92.846 -74.106 54.620
253993 92.875 -74.125 54.622
253994 92.905 -74.143 54.623
253995 92.934 -74.162 54.625
253996 92.965 -74.177 54.626
253997 92.996 -74.178 54.626
253998 93.026 -74.179 54.625
253999 93.058 -74.180 54.625
254000 93.089 -74.182 54.626
254001 93.120 -74.183 54.626
254002 93.151 -74.184 54.627
254003 93.182 -74.185 54.628
254004 93.215 -74.166 54.628
254005 93.249 -74.145 54.629
254006 93.278 -74.135 54.630
254007 93.299 -74.142 54.632
254008 93.321 -74.151 54.636
254009 93.342 -74.159 54.642
254010 93.363 -74.167 54.648
254011 93.384 -74.175 54.656
254012 93.408 -74.171 54.667
254013 93.433 -74.167 54.680
254014 93.457 -74.163 54.698
254015 93.482 -74.159 54.718
254016 93.506 -74.154 54.739
254017 93.530 -74.151 54.760
254018 93.555 -74.147 54.777
254019 93.579 -74.142 54.795
254020 93.603 -74.139 54.814
254021 93.628 -74.134 54.840
254022 93.652 -74.130 54.870
254023 93.677 -74.126 54.899
254024 93.701 -74.122 54.924
254025 93.725 -74.118 54.943
254026 93.750 -74.114 54.950
254027 93.774 -74.110 54.958
254028 93.795 -74.106 54.969
254029 93.817 -74.102 54.975
254030 93.838 -74.099 54.967
254031 93.859 -74.100 54.951
254032 93.881 -74.101 54.928
254033 93.902 -74.101 54.903
254034 93.924 -74.103 54.877
254035 93.945 -74.104 54.847
254036 93.967 -74.107 54.813
254037 93.988 -74.108 54.777
254038 94.009 -74.111 54.736
254039 94.031 -74.113 54.694
254040 94.052 -74.114 54.652
254041 94.098 -74.126 54.614
254042 94.126 -74.127 54.580
254043 94.148 -74.126 54.553
254044 94.170 -74.124 54.534
254045 94.192 -74.123 54.526
254046 94.210 -74.121 54.519
254047 94.226 -74.122 54.511
254048 94.247 -74.126 54.500
254049 94.267 -74.132 54.482
254050 94.287 -74.136 54.460
254051 94.307 -74.141 54.436
254052 94.327 -74.145 54.412
254053 94.347 -74.148 54.391
254054 94.371 -74.145 54.367
254055 94.397 -74.140 54.341
254056 94.422 -74.135 54.313
254057 94.448 -74.129 54.282
254058 94.473 -74.125 54.251
254059 94.499 -74.120 54.221
254060 94.523 -74.106 54.188
254061 94.546 -74.092 54.149
254062 94.569 -74.076 54.108
254063 94.593 -74.061 54.064
254064 94.616 -74.046 54.019
254065 94.639 -74.035 53.977
254066 94.655 -74.035 53.932
254067 94.671 -74.035 53.886
254068 94.686 -74.035 53.839
254069 94.701 -74.035 53.792
254070 94.716 -74.036 53.743
254071 94.731 -74.038 53.694
254072 94.746 -74.040 53.645
254073 94.761 -74.042 53.601
254074 94.780 -74.045 53.562
254075 94.793 -74.071 53.528
254076 94.804 -74.098 53.501
254077 94.818 -74.129 53.484
254078 94.832 -74.161 53.467
254079 94.846 -74.192 53.447
254080 94.878 -74.186 53.417
254081 94.910 -74.189 53.377
254082 94.943 -74.193 53.334
254083 94.975 -74.196 53.289
254084 95.007 -74.200 53.243
254085 95.040 -74.203 53.191
254086 95.072 -74.207 53.137
254087 95.105 -74.210 53.083
254088 95.137 -74.214 53.032
254089 95.170 -74.217 52.981
254090 95.202 -74.226 52.930
254091 95.235 -74.240 52.879
254092 95.267 -74.254 52.827
254093 95.304 -74.255 52.775
254094 95.343 -74.248 52.723
254095 95.382 -74.242 52.669
254096 95.421 -74.235 52.614
254097 95.459 -74.230 52.559
254098 95.492 -74.235 52.508
254099 95.534 -74.234 52.460
254100 95.575 -74.233 52.416
254101 95.612 -74.236 52.373
254102 95.641 -74.242 52.330
254103 95.670 -74.248 52.288
254104 95.700 -74.254 52.249
254105 95.729 -74.260 52.210
254106 95.758 -74.267 52.173
254107 95.788 -74.273 52.135
254108 95.817 -74.279 52.097
254109 95.839 -74.282 52.058
254110 95.849 -74.278 52.020
254111 95.859 -74.274 51.981
254112 95.869 -74.271 51.942
254113 95.879 -74.267 51.902
254114 95.888 -74.264 51.862
254115 95.898 -74.261 51.822
254116 95.916 -74.270 51.781
254117 95.940 -74.290 51.742
254118 95.963 -74.311 51.702
254119 95.987 -74.330 51.662
254120 96.011 -74.350 51.620
254121 96.034 -74.370 51.581
254122 96.065 -74.372 51.542
254123 96.098 -74.370 51.502
254124 96.146 -74.379 51.464
254125 96.190 -74.386 51.429
254126 96.234 -74.394 51.395
254127 96.241 -74.411 51.360
254128 96.245 -74.429 51.328
254129 96.268 -74.418 51.297
254130 96.294 -74.399 51.268
254131 96.325 -74.405 51.240
254132 96.359 -74.424 51.215
254133 96.394 -74.443 51.191
254134 96.428 -74.461 51.168
254135 96.462 -74.480 51.146
254136 96.497 -74.499 51.125
254137 96.531 -74.518 51.107
254138 96.559 -74.530 51.091
254139 96.585 -74.537 51.076
254140 96.610 -74.545 51.061
254141 96.635 -74.552 51.049
254142 96.660 -74.559 51.039
254143 96.684 -74.567 51.029
254144 96.710 -74.574 51.023
254145 96.734 -74.581 51.018
254146 96.759 -74.589 51.014
254147 96.790 -74.606 51.011
254148 96.820 -74.625 51.010
254149 96.851 -74.643 51.009
254150 96.881 -74.662 51.008
254151 96.905 -74.671 51.008
254152 96.926 -74.681 51.009
254153 96.947 -74.691 51.010
254154 96.966 -74.702 51.008
254155 96.976 -74.713 51.003
254156 96.986 -74.725 50.995
254157 97.002 -74.727 50.988
254158 97.018 -74.728 50.982
254159 97.034 -74.730 50.978
254160 97.050 -74.732 50.971
254161 97.066 -74.734 50.963
254162 97.082 -74.736 50.954
254163 97.098 -74.741 50.947
254164 97.114 -74.749 50.941
254165 97.131 -74.756 50.934
254166 97.147 -74.763 50.924
254167 97.163 -74.770 50.912
254168 97.197 -74.769 50.891
254169 97.240 -74.766 50.866
254170 97.283 -74.761 50.838
254171 97.326 -74.757 50.808
254172 97.364 -74.759 50.776
254173 97.401 -74.762 50.745
254174 97.438 -74.764 50.715
254175 97.475 -74.767 50.687
254176 97.512 -74.770 50.660
254177 97.549 -74.773 50.630
254178 97.587 -74.776 50.597
254179 97.624 -74.779 50.561
254180 97.661 -74.781 50.522
254181 97.696 -74.787 50.479
254182 97.727 -74.794 50.434
254183 97.759 -74.803 50.390
254184 97.786 -74.812 50.343
254185 97.809 -74.824 50.295
254186 97.833 -74.836 50.248
254187 97.861 -74.850 50.202
254188 97.887 -74.864 50.153
254189 97.913 -74.875 50.104
254190 97.936 -74.881 50.055
254191 97.959 -74.888 50.004
254192 97.983 -74.896 49.958
254193 98.007 -74.903 49.914
254194 98.038 -74.915 49.870
254195 98.069 -74.928 49.823
254196 98.100 -74.941 49.774
254197 98.131 -74.954 49.721
254198 98.162 -74.967 49.665
254199 98.194 -74.980 49.608
254200 98.225 -74.994 49.551
254201 98.256 -75.007 49.494
254202 98.287 -75.019 49.440
254203 98.318 -75.033 49.387
254204 98.349 -75.046 49.334
254205 98.386 -75.066 49.282
254206 98.427 -75.091 49.233
254207 98.457 -75.063 49.181
254208 98.484 -75.035 49.132
254209 98.504 -75.013 49.081
254210 98.524 -74.991 49.029
254211 98.544 -74.969 48.976
254212 98.567 -74.962 48.923
254213 98.594 -74.972 48.871
254214 98.622 -74.984 48.818
254215 98.649 -74.994 48.765
254216 98.676 -75.006 48.711
254217 98.704 -75.017 48.657
254218 98.731 -75.028 48.602
254219 98.759 -75.034 48.550
254220 98.786 -75.038 48.497
254221 98.812 -75.043 48.445
254222 98.840 -75.048 48.393
254223 98.866 -75.054 48.341
254224 98.891 -75.084 48.289
254225 98.917 -75.114 48.239
254226 98.942 -75.145 48.189
254227 98.968 -75.175 48.141
254228 98.993 -75.198 48.093
254229 99.017 -75.216 48.046
254230 99.042 -75.234 48.000
254231 99.067 -75.252 47.955
254232 99.091 -75.271 47.908
254233 99.119 -75.289 47.862
254234 99.154 -75.309 47.815
254235 99.191 -75.329 47.770
254236 99.221 -75.333 47.725
254237 99.251 -75.338 47.681
254238 99.281 -75.342 47.638
254239 99.311 -75.346 47.595
254240 99.341 -75.350 47.554
254241 99.371 -75.354 47.515
254242 99.402 -75.358 47.476
254243 99.426 -75.366 47.440
254244 99.451 -75.374 47.408
254245 99.475 -75.383 47.377
254246 99.500 -75.391 47.349
254247 99.525 -75.398 47.321
254248 99.561 -75.402 47.296
254249 99.597 -75.405 47.271
254250 99.633 -75.408 47.247
254251 99.660 -75.421 47.223
254252 99.673 -75.449 47.201
254253 99.696 -75.466 47.180
254254 99.720 -75.483 47.159
254255 99.743 -75.501 47.139
254256 99.767 -75.518 47.117
254257 99.792 -75.536 47.097
254258 99.818 -75.553 47.079
254259 99.844 -75.571 47.062
254260 99.870 -75.588 47.045
254261 99.893 -75.615 47.028
254262 99.940 -75.620 47.009
254263 99.995 -75.617 46.984
254264 100.031 -75.622 46.961
254265 100.035 -75.640 46.938
254266 100.040 -75.657 46.912
254267 100.045 -75.674 46.882
254268 100.059 -75.688 46.851
254269 100.089 -75.696 46.818
254270 100.120 -75.704 46.782
254271 100.150 -75.712 46.748
254272 100.180 -75.721 46.714
254273 100.210 -75.728 46.680
254274 100.239 -75.746 46.647
254275 100.267 -75.768 46.613
254276 100.294 -75.790 46.578
254277 100.322 -75.812 46.545
254278 100.350 -75.834 46.513
254279 100.378 -75.856 46.482
254280 100.406 -75.878 46.452
254281 100.433 -75.900 46.421
254282 100.462 -75.922 46.392
254283 100.492 -75.927 46.362
254284 100.524 -75.917 46.333
254285 100.551 -75.942 46.304
254286 100.578 -75.972 46.275
254287 100.606 -76.001 46.245
254288 100.633 -76.030 46.214
254289 100.665 -76.047 46.183
254290 100.699 -76.065 46.151
254291 100.732 -76.082 46.120
254292 100.764 -76.099 46.089
254293 100.797 -76.116 46.058
254294 100.830 -76.134 46.026
254295 100.863 -76.151 45.995
254296 100.881 -76.162 45.963
254297 100.898 -76.173 45.933
254298 100.915 -76.182 45.905
254299 100.932 -76.193 45.876
254300 100.942 -76.198 45.847
254301 100.954 -76.201 45.819
254302 100.995 -76.213 45.791
254303 101.037 -76.224 45.763
254304 101.078 -76.235 45.736
254305 101.120 -76.246 45.710
254306 101.161 -76.257 45.684
254307 101.203 -76.269 45.658
254308 101.244 -76.280 45.633
254309 101.286 -76.292 45.609
254310 101.327 -76.302 45.584
254311 101.369 -76.314 45.560
254312 101.410 -76.325 45.536
254313 101.452 -76.336 45.513
254314 101.493 -76.348 45.490
254315 101.535 -76.358 45.466
254316 101.589 -76.376 45.443
254317 101.645 -76.394 45.421
254318 101.702 -76.413 45.399
254319 101.748 -76.410 45.377
254320 101.778 -76.415 45.355
254321 101.793 -76.426 45.332
254322 101.812 -76.433 45.310
254323 101.817 -76.436 45.289
254324 101.810 -76.429 45.267
254325 101.802 -76.422 45.245
254326 101.796 -76.415 45.223
254327 101.788 -76.408 45.201
254328 101.794 -76.410 45.179
254329 101.802 -76.412 45.158
254330 101.810 -76.415 45.137
254689 92.727 -74.051 54.619
254690 92.758 -74.043 54.620
254691 92.787 -74.042 54.622
254692 92.817 -74.061 54.623
254693 92.847 -74.080 54.624
254694 92.876 -74.098 54.626
254695 92.906 -74.117 54.628
254696 92.936 -74.136 54.630
254697 92.966 -74.154 54.632
254698 92.996 -74.157 54.632
254699 93.027 -74.158 54.632
254700 93.059 -74.159 54.633
254701 93.089 -74.160 54.634
254702 93.121 -74.161 54.635
254703 93.152 -74.162 54.637
254704 93.185 -74.147 54.638
254705 93.218 -74.126 54.639
254706 93.252 -74.105 54.639
254707 93.274 -74.111 54.641
254708 93.295 -74.118 54.642
254709 93.317 -74.126 54.645
254710 93.338 -74.135 54.648
254711 93.359 -74.142 54.651
254712 93.380 -74.151 54.657
254713 93.405 -74.148 54.665
254714 93.429 -74.144 54.675
254715 93.453 -74.140 54.690
254716 93.478 -74.136 54.707
254717 93.502 -74.132 54.725
254718 93.526 -74.128 54.741
254719 93.551 -74.123 54.757
254720 93.575 -74.120 54.775
254721 93.600 -74.115 54.796
254722 93.624 -74.111 54.824
254723 93.648 -74.107 54.854
254724 93.673 -74.103 54.880
254725 93.697 -74.099 54.903
254726 93.721 -74.095 54.920
254727 93.746 -74.091 54.932
254728 93.770 -74.087 54.943
254729 93.794 -74.082 54.957
254730 93.819 -74.079 54.963
254731 93.840 -74.079 54.963
254732 93.861 -74.080 54.955
254733 93.883 -74.082 54.939
254734 93.904 -74.084 54.911
254735 93.925 -74.086 54.881
254736 93.946 -74.089 54.849
254737 93.968 -74.090 54.816
254738 93.989 -74.092 54.775
254739 94.010 -74.094 54.734
254740 94.032 -74.096 54.690
254741 94.059 -74.101 54.645
254742 94.092 -74.104 54.602
254743 94.115 -74.103 54.569
254744 94.136 -74.101 54.545
254745 94.158 -74.101 54.527
254746 94.181 -74.099 54.517
254747 94.208 -74.098 54.507
254748 94.233 -74.099 54.496
254749 94.249 -74.106 54.482
254750 94.268 -74.111 54.463
254751 94.288 -74.116 54.441
254752 94.309 -74.120 54.417
254753 94.329 -74.126 54.392
254754 94.349 -74.129 54.369
254755 94.374 -74.124 54.345
254756 94.400 -74.119 54.319
254757 94.425 -74.114 54.290
254758 94.451 -74.109 54.258
254759 94.476 -74.104 54.227
254760 94.501 -74.098 54.196
254761 94.527 -74.091 54.164
254762 94.550 -74.076 54.125
254763 94.573 -74.061 54.084
254764 94.596 -74.046 54.041
254765 94.620 -74.031 53.995
254766 94.641 -74.020 53.951
254767 94.658 -74.016 53.907
254768 94.673 -74.017 53.861
254769 94.688 -74.019 53.814
254770 94.703 -74.020 53.767
254771 94.718 -74.022 53.719
254772 94.733 -74.023 53.671
254773 94.748 -74.026 53.624
254774 94.763 -74.027 53.583
254775 94.783 -74.031 53.548
254776 94.812 -74.052 53.518
254777 94.823 -74.079 53.495
254778 94.833 -74.105 53.481
254779 94.843 -74.132 53.463
254780 94.853 -74.158 53.439
254781 94.878 -74.161 53.407
254782 94.911 -74.166 53.365
254783 94.944 -74.169 53.321
254784 94.976 -74.173 53.275
254785 95.009 -74.176 53.227
254786 95.041 -74.179 53.175
254787 95.073 -74.183 53.121
254788 95.106 -74.186 53.068
254789 95.138 -74.190 53.018
254790 95.171 -74.193 52.968
254791 95.203 -74.197 52.919
254792 95.236 -74.210 52.869
254793 95.269 -74.223 52.818
254794 95.308 -74.216 52.767
254795 95.347 -74.210 52.714
254796 95.385 -74.205 52.660
254797 95.415 -74.212 52.605
254798 95.452 -74.214 52.553
254799 95.493 -74.214 52.502
254800 95.534 -74.213 52.455
254801 95.576 -74.212 52.410
254802 95.617 -74.211 52.367
254803 95.653 -74.214 52.325
254804 95.682 -74.220 52.284
254805 95.711 -74.226 52.245
254806 95.740 -74.233 52.207
254807 95.770 -74.239 52.169
254808 95.799 -74.245 52.131
254809 95.827 -74.250 52.094
254810 95.837 -74.247 52.056
254811 95.846 -74.243 52.017
254812 95.856 -74.240 51.979
254813 95.866 -74.236 51.940
254814 95.876 -74.233 51.901
254815 95.885 -74.230 51.862
254816 95.895 -74.226 51.823
254817 95.910 -74.230 51.784
254818 95.934 -74.250 51.744
254819 95.957 -74.270 51.705
254820 95.981 -74.289 51.666
254821 96.005 -74.310 51.625
254822 96.031 -74.322 51.584
254823 96.064 -74.320 51.544
254824 96.097 -74.318 51.505
254825 96.136 -74.321 51.467
254826 96.185 -74.332 51.430
254827 96.229 -74.339 51.395
254828 96.236 -74.357 51.361
254829 96.240 -74.374 51.329
254830 96.254 -74.377 51.298
254831 96.280 -74.359 51.269
254832 96.312 -74.369 51.241
254833 96.346 -74.388 51.215
254834 96.381 -74.407 51.191
254835 96.415 -74.426 51.167
254836 96.450 -74.445 51.144
254837 96.484 -74.463 51.123
254838 96.518 -74.482 51.104
254839 96.552 -74.501 51.087
254840 96.582 -74.514 51.070
254841 96.607 -74.521 51.057
254842 96.631 -74.528 51.044
254843 96.657 -74.536 51.032
254844 96.681 -74.543 51.022
254845 96.707 -74.550 51.015
254846 96.732 -74.558 51.009
254847 96.757 -74.567 51.004
254848 96.788 -74.585 51.001
254849 96.818 -74.604 50.998
254850 96.849 -74.622 50.998
254851 96.875 -74.639 50.998
254852 96.897 -74.653 50.999
254853 96.919 -74.666 51.002
254854 96.940 -74.676 51.005
254855 96.956 -74.687 51.005
254856 96.969 -74.693 51.001
254857 96.985 -74.695 50.995
254858 97.001 -74.697 50.990
254859 97.017 -74.703 50.986
254860 97.033 -74.711 50.982
254861 97.050 -74.718 50.977
254862 97.066 -74.725 50.971
254863 97.082 -74.732 50.965
254864 97.098 -74.740 50.960
254865 97.114 -74.747 50.954
254866 97.131 -74.754 50.946
254867 97.147 -74.762 50.934
254868 97.163 -74.769 50.920
254869 97.190 -74.771 50.898
254870 97.233 -74.767 50.872
254871 97.276 -74.762 50.842
254872 97.318 -74.759 50.811
254873 97.361 -74.754 50.780
254874 97.404 -74.750 50.750
254875 97.446 -74.746 50.721
254876 97.489 -74.742 50.693
254877 97.530 -74.740 50.664
254878 97.567 -74.743 50.633
254879 97.604 -74.746 50.599
254880 97.642 -74.748 50.561
254881 97.678 -74.751 50.520
254882 97.715 -74.755 50.475
254883 97.747 -74.762 50.428
254884 97.774 -74.772 50.380
254885 97.798 -74.784 50.333
254886 97.825 -74.799 50.285
254887 97.852 -74.812 50.238
254888 97.879 -74.827 50.193
254889 97.904 -74.838 50.144
254890 97.928 -74.844 50.095
254891 97.951 -74.852 50.047
254892 97.974 -74.859 49.999
254893 97.995 -74.864 49.957
254894 98.016 -74.869 49.914
254895 98.048 -74.891 49.870
254896 98.079 -74.905 49.822
254897 98.109 -74.918 49.773
254898 98.141 -74.932 49.720
254899 98.172 -74.944 49.663
254900 98.203 -74.957 49.607
254901 98.234 -74.970 49.549
254902 98.265 -74.984 49.492
254903 98.296 -74.997 49.437
254904 98.327 -75.010 49.384
254905 98.359 -75.022 49.330
254906 98.397 -75.045 49.279
254907 98.434 -75.051 49.229
254908 98.463 -75.021 49.176
254909 98.491 -74.993 49.125
254910 98.511 -74.971 49.074
254911 98.530 -74.949 49.022
254912 98.550 -74.927 48.968
254913 98.577 -74.935 48.916
254914 98.604 -74.946 48.862
254915 98.632 -74.957 48.809
254916 98.659 -74.968 48.755
254917 98.687 -74.979 48.701
254918 98.714 -74.990 48.647
254919 98.741 -75.001 48.593
254920 98.769 -75.011 48.540
254921 98.795 -75.016 48.488
254922 98.822 -75.020 48.436
254923 98.850 -75.025 48.384
254924 98.876 -75.041 48.332
254925 98.901 -75.072 48.282
254926 98.926 -75.102 48.232
254927 98.952 -75.132 48.183
254928 98.977 -75.151 48.133
254929 99.001 -75.169 48.086
254930 99.026 -75.187 48.039
254931 99.051 -75.205 47.993
254932 99.075 -75.223 47.947
254933 99.100 -75.242 47.901
254934 99.129 -75.261 47.855
254935 99.165 -75.280 47.809
254936 99.201 -75.300 47.764
254937 99.233 -75.311 47.719
254938 99.264 -75.314 47.674
254939 99.293 -75.318 47.631
254940 99.324 -75.323 47.587
254941 99.354 -75.326 47.547
254942 99.383 -75.330 47.507
254943 99.412 -75.336 47.469
254944 99.437 -75.344 47.432
254945 99.462 -75.352 47.400
254946 99.486 -75.361 47.369
254947 99.510 -75.369 47.341
254948 99.545 -75.373 47.314
254949 99.581 -75.376 47.288
254950 99.617 -75.380 47.263
254951 99.652 -75.383 47.239
254952 99.688 -75.386 47.215
254953 99.709 -75.405 47.193
254954 99.723 -75.432 47.172
254955 99.742 -75.453 47.151
254956 99.766 -75.471 47.131
254957 99.789 -75.488 47.110
254958 99.812 -75.505 47.090
254959 99.836 -75.523 47.072
254960 99.860 -75.540 47.055
254961 99.885 -75.558 47.039
254962 99.924 -75.569 47.022
254963 99.971 -75.577 47.002
254964 100.018 -75.584 46.978
254965 100.049 -75.589 46.955
254966 100.057 -75.592 46.931
254967 100.062 -75.609 46.904
254968 100.067 -75.626 46.875
254969 100.073 -75.643 46.843
254970 100.103 -75.651 46.809
254971 100.133 -75.659 46.772
254972 100.163 -75.668 46.739
254973 100.194 -75.675 46.705
254974 100.223 -75.687 46.671
254975 100.251 -75.709 46.638
254976 100.279 -75.731 46.604
254977 100.307 -75.753 46.570
254978 100.335 -75.775 46.538
254979 100.362 -75.798 46.506
254980 100.390 -75.820 46.476
254981 100.418 -75.842 46.445
254982 100.446 -75.864 46.415
254983 100.474 -75.886 46.385
254984 100.501 -75.908 46.356
254985 100.533 -75.904 46.328
254986 100.560 -75.932 46.299
254987 100.587 -75.962 46.271
254988 100.620 -75.979 46.241
254989 100.653 -75.997 46.212
254990 100.686 -76.013 46.181
254991 100.718 -76.031 46.149
254992 100.751 -76.048 46.119
254993 100.784 -76.066 46.089
254994 100.817 -76.083 46.058
254995 100.850 -76.100 46.027
254996 100.876 -76.114 45.995
254997 100.893 -76.125 45.964
254998 100.910 -76.135 45.933
254999 100.928 -76.146 45.904
255000 100.945 -76.157 45.876
255001 100.959 -76.164 45.847
255002 100.979 -76.170 45.818
255003 101.021 -76.182 45.790
255004 101.062 -76.193 45.761
255005 101.104 -76.204 45.733
255006 101.145 -76.215 45.707
255007 101.187 -76.226 45.680
255008 101.229 -76.238 45.654
255009 101.270 -76.249 45.629
255010 101.312 -76.260 45.604
255011 101.353 -76.271 45.579
255012 101.395 -76.282 45.554
255013 101.436 -76.294 45.530
255014 101.477 -76.305 45.507
255015 101.519 -76.316 45.483
255016 101.560 -76.327 45.460
255017 101.602 -76.339 45.437
255018 101.656 -76.356 45.414
255019 101.713 -76.374 45.392
255020 101.770 -76.392 45.369
255021 101.807 -76.402 45.347
255022 101.803 -76.398 45.326
255023 101.796 -76.391 45.303
255024 101.789 -76.384 45.281
255025 101.782 -76.377 45.259
255026 101.775 -76.370 45.237
255027 101.782 -76.373 45.215
255028 101.790 -76.376 45.193
255029 101.798 -76.378 45.171
255030 101.806 -76.381 45.149
255031 101.813 -76.383 45.128
255389 92.702 -74.034 54.623
255390 92.728 -74.024 54.624
255391 92.759 -74.016 54.624
255392 92.789 -74.017 54.626
255393 92.818 -74.035 54.627
255394 92.848 -74.054 54.628
255395 92.878 -74.073 54.630
255396 92.907 -74.091 54.632
255397 92.937 -74.110 54.634
255398 92.967 -74.129 54.636
255399 92.997 -74.136 54.637
255400 93.028 -74.136 54.638
255401 93.059 -74.138 54.640
255402 93.090 -74.139 54.641
255403 93.122 -74.140 54.642
255404 93.154 -74.128 54.645
255405 93.188 -74.107 54.647
255406 93.221 -74.086 54.648
255407 93.250 -74.078 54.649
255408 93.271 -74.086 54.650
255409 93.292 -74.094 54.651
255410 93.313 -74.102 54.652
255411 93.334 -74.111 54.655
255412 93.355 -74.117 54.656
255413 93.379 -74.121 54.659
255414 93.406 -74.114 54.664
255415 93.430 -74.110 54.673
255416 93.454 -74.105 54.684
255417 93.479 -74.101 54.697
255418 93.503 -74.098 54.711
255419 93.527 -74.093 54.724
255420 93.552 -74.089 54.738
255421 93.576 -74.085 54.756
255422 93.600 -74.081 54.778
255423 93.625 -74.077 54.806
255424 93.649 -74.073 54.834
255425 93.674 -74.069 54.859
255426 93.698 -74.064 54.879
255427 93.722 -74.060 54.893
255428 93.747 -74.057 54.908
255429 93.771 -74.052 54.924
255430 93.796 -74.048 54.939
255431 93.819 -74.055 54.948
255432 93.840 -74.057 54.954
255433 93.861 -74.060 54.952
255434 93.883 -74.062 54.942
255435 93.904 -74.065 54.913
255436 93.926 -74.068 54.881
255437 93.947 -74.070 54.847
255438 93.969 -74.073 54.813
255439 93.990 -74.076 54.771
255440 94.011 -74.078 54.728
255441 94.033 -74.079 54.684
255442 94.059 -74.082 54.638
255443 94.082 -74.080 54.591
255444 94.109 -74.079 54.559
255445 94.135 -74.078 54.536
255446 94.161 -74.076 54.519
255447 94.188 -74.075 54.507
255448 94.217 -74.074 54.495
255449 94.248 -74.073 54.481
255450 94.269 -74.077 54.465
255451 94.285 -74.084 54.444
255452 94.301 -74.091 54.420
255453 94.318 -74.097 54.396
255454 94.334 -74.104 54.370
255455 94.351 -74.108 54.346
255456 94.377 -74.103 54.321
255457 94.402 -74.098 54.295
255458 94.428 -74.093 54.265
255459 94.453 -74.088 54.234
255460 94.479 -74.082 54.203
255461 94.504 -74.077 54.172
255462 94.530 -74.072 54.140
255463 94.553 -74.060 54.103
255464 94.577 -74.045 54.062
255465 94.600 -74.031 54.019
255466 94.621 -74.020 53.971
255467 94.641 -74.010 53.926
255468 94.660 -74.001 53.880
255469 94.675 -74.003 53.834
255470 94.690 -74.004 53.788
255471 94.705 -74.006 53.741
255472 94.720 -74.007 53.694
255473 94.735 -74.009 53.648
255474 94.750 -74.010 53.604
255475 94.765 -74.013 53.567
255476 94.785 -74.017 53.535
255477 94.832 -74.034 53.509
255478 94.842 -74.060 53.490
255479 94.852 -74.087 53.477
255480 94.862 -74.114 53.458
255481 94.873 -74.140 53.430
255482 94.892 -74.141 53.395
255483 94.913 -74.136 53.351
255484 94.940 -74.136 53.306
255485 94.974 -74.141 53.259
255486 95.007 -74.147 53.211
255487 95.041 -74.152 53.157
255488 95.074 -74.157 53.105
255489 95.107 -74.163 53.054
255490 95.140 -74.167 53.005
255491 95.172 -74.170 52.957
255492 95.205 -74.173 52.909
255493 95.237 -74.179 52.860
255494 95.272 -74.184 52.810
255495 95.310 -74.180 52.759
255496 95.340 -74.187 52.706
255497 95.370 -74.194 52.652
255498 95.411 -74.194 52.597
255499 95.452 -74.193 52.546
255500 95.494 -74.192 52.497
255501 95.535 -74.191 52.450
255502 95.577 -74.190 52.405
255503 95.618 -74.189 52.362
255504 95.660 -74.189 52.320
255505 95.693 -74.192 52.280
255506 95.723 -74.198 52.242
255507 95.752 -74.205 52.203
255508 95.781 -74.211 52.165
255509 95.811 -74.217 52.127
255510 95.824 -74.216 52.090
255511 95.834 -74.212 52.052
255512 95.843 -74.208 52.014
255513 95.853 -74.205 51.977
255514 95.863 -74.201 51.939
255515 95.873 -74.198 51.901
255516 95.883 -74.195 51.863
255517 95.892 -74.191 51.825
255518 95.904 -74.190 51.786
255519 95.927 -74.210 51.747
255520 95.951 -74.230 51.709
255521 95.975 -74.249 51.669
255522 95.999 -74.269 51.629
255523 96.030 -74.270 51.588
255524 96.063 -74.269 51.547
255525 96.096 -74.267 51.507
255526 96.129 -74.264 51.470
255527 96.180 -74.277 51.432
255528 96.225 -74.285 51.397
255529 96.232 -74.302 51.363
255530 96.236 -74.320 51.331
255531 96.240 -74.336 51.300
255532 96.266 -74.319 51.270
255533 96.300 -74.333 51.241
255534 96.334 -74.352 51.215
255535 96.368 -74.371 51.190
255536 96.402 -74.389 51.166
255537 96.437 -74.408 51.143
255538 96.471 -74.427 51.121
255539 96.506 -74.446 51.102
255540 96.540 -74.465 51.084
255541 96.574 -74.484 51.067
255542 96.604 -74.498 51.053
255543 96.629 -74.505 51.040
255544 96.654 -74.512 51.027
255545 96.679 -74.520 51.017
255546 96.703 -74.527 51.009
255547 96.729 -74.534 51.002
255548 96.755 -74.546 50.996
255549 96.786 -74.564 50.992
255550 96.816 -74.583 50.989
255551 96.844 -74.600 50.988
255552 96.866 -74.615 50.989
255553 96.888 -74.628 50.991
255554 96.910 -74.643 50.995
255555 96.932 -74.656 50.999
255556 96.952 -74.667 51.001
255557 96.969 -74.679 50.999
255558 96.985 -74.687 50.996
255559 97.002 -74.694 50.992
255560 97.018 -74.702 50.989
255561 97.034 -74.709 50.986
255562 97.050 -74.716 50.982
255563 97.067 -74.724 50.979
255564 97.083 -74.731 50.975
255565 97.099 -74.738 50.971
255566 97.115 -74.745 50.965
255567 97.131 -74.753 50.956
255568 97.148 -74.759 50.943
255569 97.164 -74.767 50.927
255570 97.183 -74.773 50.904
255571 97.226 -74.769 50.878
255572 97.268 -74.765 50.848
255573 97.311 -74.760 50.816
255574 97.354 -74.756 50.786
255575 97.396 -74.752 50.756
255576 97.439 -74.747 50.728
255577 97.482 -74.744 50.699
255578 97.525 -74.739 50.668
255579 97.567 -74.735 50.635
255580 97.610 -74.731 50.599
255581 97.653 -74.727 50.560
255582 97.696 -74.722 50.517
255583 97.734 -74.724 50.471
255584 97.763 -74.734 50.422
255585 97.790 -74.747 50.372
255586 97.817 -74.761 50.323
255587 97.844 -74.775 50.274
255588 97.870 -74.789 50.227
255589 97.896 -74.800 50.182
255590 97.919 -74.808 50.134
255591 97.943 -74.815 50.087
255592 97.964 -74.820 50.042
255593 97.984 -74.825 49.998
255594 98.005 -74.829 49.957
255595 98.026 -74.835 49.914
255596 98.057 -74.856 49.870
255597 98.088 -74.878 49.822
255598 98.119 -74.895 49.771
255599 98.150 -74.908 49.718
255600 98.181 -74.922 49.662
255601 98.213 -74.935 49.605
255602 98.243 -74.947 49.547
255603 98.275 -74.960 49.489
255604 98.306 -74.974 49.433
255605 98.337 -74.987 49.379
255606 98.368 -75.000 49.326
255607 98.409 -75.025 49.274
255608 98.441 -75.009 49.223
255609 98.471 -74.979 49.170
255610 98.497 -74.952 49.117
255611 98.517 -74.929 49.066
255612 98.537 -74.907 49.014
255613 98.559 -74.897 48.961
255614 98.586 -74.907 48.908
255615 98.614 -74.919 48.854
255616 98.641 -74.929 48.800
255617 98.669 -74.941 48.746
255618 98.696 -74.952 48.692
255619 98.723 -74.963 48.637
255620 98.751 -74.974 48.583
255621 98.778 -74.985 48.531
255622 98.805 -74.992 48.479
255623 98.833 -74.997 48.427
255624 98.859 -75.001 48.375
255625 98.885 -75.029 48.324
255626 98.911 -75.060 48.274
255627 98.935 -75.085 48.225
255628 98.960 -75.103 48.176
255629 98.985 -75.121 48.126
255630 99.009 -75.139 48.079
255631 99.034 -75.157 48.032
255632 99.059 -75.176 47.986
255633 99.083 -75.195 47.940
255634 99.108 -75.213 47.894
255635 99.139 -75.232 47.849
255636 99.175 -75.251 47.804
255637 99.211 -75.271 47.759
255638 99.246 -75.287 47.713
255639 99.276 -75.291 47.668
255640 99.306 -75.295 47.624
255641 99.336 -75.299 47.581
255642 99.366 -75.303 47.540
255643 99.396 -75.308 47.501
255644 99.423 -75.314 47.463
255645 99.447 -75.322 47.425
255646 99.472 -75.330 47.392
255647 99.496 -75.339 47.361
255648 99.529 -75.343 47.333
255649 99.565 -75.347 47.305
255650 99.600 -75.350 47.279
255651 99.636 -75.354 47.254
255652 99.672 -75.357 47.231
255653 99.708 -75.360 47.208
255654 99.744 -75.364 47.186
255655 99.759 -75.389 47.165
255656 99.772 -75.416 47.145
255657 99.788 -75.440 47.123
255658 99.811 -75.458 47.103
255659 99.835 -75.475 47.083
255660 99.859 -75.493 47.064
255661 99.882 -75.510 47.047
255662 99.911 -75.533 47.031
255663 99.946 -75.537 47.014
255664 99.993 -75.545 46.994
255665 100.041 -75.553 46.970
255666 100.067 -75.557 46.947
255667 100.076 -75.557 46.923
255668 100.084 -75.561 46.896
255669 100.089 -75.578 46.866
255670 100.093 -75.596 46.834
255671 100.117 -75.606 46.799
255672 100.147 -75.614 46.763
255673 100.177 -75.622 46.729
255674 100.207 -75.630 46.695
255675 100.235 -75.651 46.662
255676 100.263 -75.673 46.629
255677 100.291 -75.695 46.596
255678 100.319 -75.718 46.562
255679 100.347 -75.740 46.530
255680 100.374 -75.762 46.499
255681 100.402 -75.784 46.469
255682 100.430 -75.806 46.439
255683 100.458 -75.828 46.410
255684 100.486 -75.850 46.380
255685 100.514 -75.872 46.351
255686 100.542 -75.892 46.323
255687 100.574 -75.911 46.295
255688 100.607 -75.928 46.267
255689 100.640 -75.945 46.238
255690 100.673 -75.963 46.208
255691 100.705 -75.980 46.178
255692 100.739 -75.997 46.147
255693 100.771 -76.015 46.117
255694 100.804 -76.032 46.088
255695 100.837 -76.049 46.058
255696 100.869 -76.066 46.027
255697 100.889 -76.078 45.997
255698 100.906 -76.088 45.966
255699 100.923 -76.098 45.934
255700 100.940 -76.109 45.905
255701 100.958 -76.119 45.875
255702 100.975 -76.130 45.847
255703 101.005 -76.139 45.818
255704 101.047 -76.151 45.789
255705 101.088 -76.162 45.760
255706 101.130 -76.173 45.732
255707 101.171 -76.184 45.704
255708 101.213 -76.195 45.678
255709 101.254 -76.207 45.650
255710 101.296 -76.218 45.624
255711 101.337 -76.229 45.599
255712 101.379 -76.240 45.574
255713 101.420 -76.251 45.549
255714 101.462 -76.263 45.525
255715 101.503 -76.274 45.501
255716 101.545 -76.285 45.477
255717 101.586 -76.296 45.454
255718 101.628 -76.307 45.430
255719 101.669 -76.319 45.407
255720 101.723 -76.336 45.385
255721 101.738 -76.375 45.362
255722 101.782 -76.352 45.339
255723 101.775 -76.345 45.317
255724 101.768 -76.339 45.295
255725 101.763 -76.333 45.273
255726 101.771 -76.336 45.251
255727 101.779 -76.339 45.229
255728 101.786 -76.342 45.206
255729 101.794 -76.344 45.184
255730 101.802 -76.347 45.162
255731 101.809 -76.349 45.140
255732 101.817 -76.352 45.118
256089 92.683 -74.007 54.626
256090 92.703 -74.007 54.627
256091 92.730 -73.998 54.627
256092 92.760 -73.989 54.628
256093 92.790 -73.990 54.629
256094 92.819 -74.009 54.630
256095 92.849 -74.028 54.631
256096 92.879 -74.046 54.633
256097 92.909 -74.065 54.635
256098 92.938 -74.084 54.637
256099 92.968 -74.102 54.640
256100 92.998 -74.114 54.641
256101 93.029 -74.115 54.643
256102 93.060 -74.116 54.645
256103 93.091 -74.117 54.647
256104 93.124 -74.109 54.649
256105 93.157 -74.088 54.651
256106 93.191 -74.067 54.653
256107 93.224 -74.047 54.655
256108 93.246 -74.054 54.657
256109 93.267 -74.062 54.659
256110 93.288 -74.070 54.660
256111 93.309 -74.078 54.660
256112 93.330 -74.083 54.662
256113 93.350 -74.088 54.662
256114 93.380 -74.090 54.663
256115 93.409 -74.087 54.665
256116 93.431 -74.074 54.671
256117 93.455 -74.070 54.679
256118 93.480 -74.066 54.689
256119 93.504 -74.061 54.700
256120 93.529 -74.057 54.711
256121 93.553 -74.053 54.724
256122 93.577 -74.049 54.740
256123 93.602 -74.045 54.759
256124 93.626 -74.041 54.784
256125 93.650 -74.037 54.808
256126 93.675 -74.032 54.830
256127 93.699 -74.029 54.847
256128 93.723 -74.025 54.862
256129 93.748 -74.020 54.879
256130 93.785 -74.026 54.897
256131 93.807 -74.032 54.917
256132 93.819 -74.032 54.936
256133 93.840 -74.034 54.942
256134 93.861 -74.036 54.940
256135 93.883 -74.038 54.934
256136 93.904 -74.042 54.908
256137 93.925 -74.045 54.877
256138 93.947 -74.048 54.842
256139 93.969 -74.050 54.801
256140 93.990 -74.053 54.762
256141 94.011 -74.056 54.720
256142 94.033 -74.057 54.675
256143 94.062 -74.058 54.628
256144 94.089 -74.057 54.581
256145 94.115 -74.055 54.549
256146 94.142 -74.054 54.527
256147 94.168 -74.052 54.509
256148 94.196 -74.051 54.497
256149 94.227 -74.050 54.482
256150 94.258 -74.049 54.466
256151 94.289 -74.048 54.448
256152 94.306 -74.055 54.424
256153 94.322 -74.062 54.400
256154 94.338 -74.068 54.374
256155 94.354 -74.075 54.349
256156 94.367 -74.079 54.324
256157 94.376 -74.073 54.299
256158 94.402 -74.071 54.272
256159 94.429 -74.068 54.242
256160 94.455 -74.065 54.210
256161 94.481 -74.061 54.179
256162 94.506 -74.056 54.148
256163 94.532 -74.051 54.117
256164 94.557 -74.045 54.081
256165 94.580 -74.030 54.040
256166 94.601 -74.020 53.996
256167 94.621 -74.010 53.948
256168 94.641 -74.001 53.902
256169 94.661 -73.991 53.856
256170 94.677 -73.988 53.810
256171 94.692 -73.990 53.764
256172 94.707 -73.991 53.718
256173 94.722 -73.993 53.672
256174 94.737 -73.995 53.628
256175 94.752 -73.996 53.587
256176 94.767 -73.998 53.553
256177 94.788 -74.003 53.525
256178 94.835 -74.020 53.502
256179 94.861 -74.042 53.487
256180 94.872 -74.069 53.471
256181 94.882 -74.095 53.449
256182 94.892 -74.122 53.419
256183 94.910 -74.126 53.379
256184 94.931 -74.120 53.336
256185 94.953 -74.115 53.290
256186 94.974 -74.110 53.243
256187 95.002 -74.110 53.195
256188 95.035 -74.116 53.143
256189 95.069 -74.121 53.092
256190 95.103 -74.126 53.043
256191 95.136 -74.132 52.995
256192 95.170 -74.138 52.946
256193 95.203 -74.144 52.898
256194 95.237 -74.149 52.850
256195 95.266 -74.162 52.801
256196 95.296 -74.169 52.749
256197 95.329 -74.174 52.697
256198 95.370 -74.173 52.644
256199 95.411 -74.173 52.590
256200 95.453 -74.171 52.539
256201 95.494 -74.170 52.491
256202 95.536 -74.170 52.445
256203 95.578 -74.169 52.400
256204 95.619 -74.168 52.358
256205 95.661 -74.167 52.316
256206 95.702 -74.166 52.276
256207 95.734 -74.170 52.238
256208 95.763 -74.177 52.200
256209 95.793 -74.183 52.162
256210 95.812 -74.184 52.124
256211 95.821 -74.181 52.087
256212 95.831 -74.177 52.048
256213 95.841 -74.173 52.011
256214 95.850 -74.170 51.975
256215 95.861 -74.167 51.938
256216 95.870 -74.164 51.901
256217 95.880 -74.160 51.864
256218 95.890 -74.156 51.826
256219 95.899 -74.153 51.787
256220 95.921 -74.170 51.750
256221 95.945 -74.189 51.712
256222 95.969 -74.209 51.673
256223 95.995 -74.221 51.632
256224 96.029 -74.219 51.591
256225 96.062 -74.217 51.549
256226 96.095 -74.215 51.509
256227 96.128 -74.213 51.472
256228 96.161 -74.211 51.435
256229 96.220 -74.230 51.400
256230 96.227 -74.247 51.366
256231 96.231 -74.264 51.333
256232 96.235 -74.283 51.301
256233 96.252 -74.279 51.271
256234 96.287 -74.298 51.242
256235 96.321 -74.317 51.215
256236 96.356 -74.335 51.189
256237 96.390 -74.354 51.165
256238 96.424 -74.373 51.141
256239 96.458 -74.392 51.120
256240 96.493 -74.411 51.100
256241 96.527 -74.430 51.081
256242 96.561 -74.449 51.065
256243 96.596 -74.468 51.049
256244 96.626 -74.481 51.035
256245 96.651 -74.489 51.024
256246 96.676 -74.496 51.013
256247 96.700 -74.503 51.003
256248 96.726 -74.511 50.996
256249 96.754 -74.524 50.990
256250 96.784 -74.543 50.985
256251 96.814 -74.561 50.982
256252 96.835 -74.575 50.981
256253 96.858 -74.590 50.981
256254 96.879 -74.604 50.984
256255 96.900 -74.611 50.989
256256 96.921 -74.616 50.993
256257 96.946 -74.632 50.996
256258 96.963 -74.644 50.997
256259 96.980 -74.656 50.998
256260 96.997 -74.668 50.996
256261 97.014 -74.681 50.994
256262 97.031 -74.692 50.991
256263 97.048 -74.704 50.988
256264 97.066 -74.716 50.986
256265 97.083 -74.728 50.984
256266 97.099 -74.736 50.981
256267 97.115 -74.744 50.974
256268 97.132 -74.750 50.964
256269 97.148 -74.758 50.951
256270 97.164 -74.766 50.933
256271 97.180 -74.772 50.910
256272 97.219 -74.770 50.884
256273 97.261 -74.766 50.855
256274 97.304 -74.762 50.824
256275 97.347 -74.758 50.792
256276 97.389 -74.753 50.763
256277 97.432 -74.749 50.734
256278 97.474 -74.745 50.705
256279 97.520 -74.735 50.672
256280 97.567 -74.722 50.636
256281 97.613 -74.709 50.597
256282 97.660 -74.696 50.557
256283 97.708 -74.682 50.512
256284 97.743 -74.685 50.465
256285 97.776 -74.705 50.417
256286 97.808 -74.724 50.366
256287 97.835 -74.737 50.314
256288 97.862 -74.752 50.263
256289 97.888 -74.763 50.216
256290 97.911 -74.771 50.171
256291 97.932 -74.775 50.125
256292 97.953 -74.781 50.081
256293 97.973 -74.785 50.039
256294 97.994 -74.791 50.000
256295 98.014 -74.795 49.957
256296 98.035 -74.800 49.914
256297 98.067 -74.822 49.869
256298 98.097 -74.844 49.821
256299 98.129 -74.866 49.769
256300 98.160 -74.885 49.715
256301 98.191 -74.898 49.661
256302 98.222 -74.911 49.604
256303 98.253 -74.925 49.544
256304 98.284 -74.938 49.485
256305 98.315 -74.950 49.429
256306 98.347 -74.964 49.373
256307 98.379 -74.979 49.320
256308 98.419 -74.997 49.268
256309 98.448 -74.967 49.216
256310 98.477 -74.938 49.162
256311 98.504 -74.910 49.109
256312 98.523 -74.888 49.058
256313 98.543 -74.866 49.007
256314 98.569 -74.869 48.955
256315 98.596 -74.881 48.900
256316 98.624 -74.891 48.845
256317 98.651 -74.903 48.791
256318 98.678 -74.913 48.737
256319 98.706 -74.925 48.682
256320 98.733 -74.935 48.628
256321 98.761 -74.947 48.574
256322 98.788 -74.958 48.522
256323 98.815 -74.969 48.469
256324 98.843 -74.974 48.417
256325 98.869 -74.988 48.366
256326 98.895 -75.018 48.316
256327 98.919 -75.037 48.265
256328 98.944 -75.056 48.216
256329 98.969 -75.074 48.167
256330 98.993 -75.092 48.119
256331 99.018 -75.110 48.071
256332 99.043 -75.129 48.025
256333 99.067 -75.147 47.979
256334 99.092 -75.165 47.933
256335 99.117 -75.184 47.887
256336 99.149 -75.203 47.842
256337 99.186 -75.223 47.796
256338 99.221 -75.242 47.751
256339 99.257 -75.261 47.705
256340 99.288 -75.268 47.660
256341 99.318 -75.272 47.617
256342 99.348 -75.276 47.574
256343 99.378 -75.280 47.533
256344 99.409 -75.284 47.493
256345 99.433 -75.292 47.455
256346 99.458 -75.300 47.418
256347 99.483 -75.308 47.385
256348 99.512 -75.314 47.355
256349 99.548 -75.317 47.325
256350 99.584 -75.321 47.297
256351 99.620 -75.324 47.271
256352 99.656 -75.327 47.247
256353 99.692 -75.331 47.223
256354 99.727 -75.334 47.201
256355 99.763 -75.338 47.179
256356 99.795 -75.345 47.158
256357 99.808 -75.372 47.138
256358 99.822 -75.399 47.116
256359 99.835 -75.427 47.095
256360 99.857 -75.445 47.075
256361 99.883 -75.464 47.056
256362 99.913 -75.489 47.039
256363 99.940 -75.505 47.023
256364 99.968 -75.506 47.005
256365 100.015 -75.514 46.985
256366 100.063 -75.521 46.962
256367 100.085 -75.524 46.939
256368 100.094 -75.525 46.914
256369 100.103 -75.525 46.887
256370 100.111 -75.530 46.856
256371 100.116 -75.547 46.823
256372 100.131 -75.561 46.789
256373 100.161 -75.569 46.753
256374 100.191 -75.577 46.719
256375 100.220 -75.593 46.686
256376 100.248 -75.615 46.653
256377 100.275 -75.637 46.620
256378 100.303 -75.659 46.587
256379 100.331 -75.681 46.554
256380 100.359 -75.703 46.522
256381 100.387 -75.725 46.493
256382 100.415 -75.747 46.463
256383 100.442 -75.769 46.433
256384 100.471 -75.793 46.404
256385 100.499 -75.821 46.376
256386 100.528 -75.847 46.347
256387 100.556 -75.864 46.319
256388 100.596 -75.868 46.291
256389 100.628 -75.887 46.263
256390 100.661 -75.906 46.233
256391 100.693 -75.926 46.204
256392 100.725 -75.946 46.175
256393 100.758 -75.963 46.144
256394 100.791 -75.981 46.116
256395 100.824 -75.998 46.087
256396 100.857 -76.016 46.058
256397 100.884 -76.030 46.028
256398 100.902 -76.041 45.998
256399 100.919 -76.051 45.967
256400 100.936 -76.062 45.936
256401 100.955 -76.073 45.906
256402 100.974 -76.084 45.876
256403 100.991 -76.093 45.848
256404 101.016 -76.100 45.819
256405 101.058 -76.112 45.788
256406 101.100 -76.123 45.759
256407 101.141 -76.135 45.730
256408 101.183 -76.146 45.701
256409 101.225 -76.157 45.674
256410 101.267 -76.169 45.647
256411 101.308 -76.180 45.621
256412 101.350 -76.191 45.595
256413 101.392 -76.203 45.569
256414 101.433 -76.214 45.544
256415 101.475 -76.225 45.519
256416 101.516 -76.236 45.495
256417 101.558 -76.248 45.471
256418 101.600 -76.259 45.447
256419 101.641 -76.270 45.423
256420 101.684 -76.281 45.400
256421 101.715 -76.291 45.378
256422 101.733 -76.317 45.355
256423 101.755 -76.301 45.332
256424 101.752 -76.296 45.310
256425 101.759 -76.299 45.287
256426 101.767 -76.302 45.264
256427 101.775 -76.304 45.242
256428 101.782 -76.307 45.220
256429 101.790 -76.310 45.197
256430 101.798 -76.313 45.174
256431 101.806 -76.316 45.152
256432 101.813 -76.318 45.131
256433 101.821 -76.321 45.108
256789 92.660 -73.963 54.629
256790 92.684 -73.975 54.629
256791 92.707 -73.982 54.629
256792 92.732 -73.970 54.630
256793 92.761 -73.962 54.630
256794 92.791 -73.964 54.631
256795 92.821 -73.983 54.632
256796 92.851 -74.001 54.633
256797 92.880 -74.020 54.635
256798 92.910 -74.039 54.637
256799 92.939 -74.057 54.639
256800 92.969 -74.076 54.642
256801 92.999 -74.092 54.644
256802 93.030 -74.094 54.647
256803 93.061 -74.095 54.650
256804 93.093 -74.090 54.653
256805 93.127 -74.069 54.655
256806 93.160 -74.048 54.657
256807 93.194 -74.028 54.659
256808 93.221 -74.021 54.660
256809 93.242 -74.029 54.663
256810 93.263 -74.038 54.666
256811 93.284 -74.044 54.668
256812 93.304 -74.049 54.668
256813 93.325 -74.054 54.669
256814 93.348 -74.057 54.668
256815 93.380 -74.060 54.667
256816 93.413 -74.060 54.667
256817 93.435 -74.047 54.669
256818 93.458 -74.034 54.674
256819 93.482 -74.029 54.681
256820 93.506 -74.026 54.690
256821 93.530 -74.021 54.699
256822 93.555 -74.017 54.710
256823 93.579 -74.013 54.723
256824 93.604 -74.009 54.739
256825 93.628 -74.005 54.760
256826 93.652 -74.001 54.781
256827 93.677 -73.997 54.799
256828 93.701 -73.993 54.814
256829 93.725 -73.988 54.830
256830 93.775 -74.004 54.849
256831 93.794 -74.009 54.871
256832 93.804 -74.010 54.895
256833 93.819 -74.010 54.916
256834 93.840 -74.011 54.926
256835 93.860 -74.013 54.927
256836 93.882 -74.016 54.918
256837 93.904 -74.018 54.895
256838 93.925 -74.021 54.863
256839 93.946 -74.024 54.826
256840 93.968 -74.026 54.788
256841 93.990 -74.029 54.746
256842 94.011 -74.031 54.703
256843 94.037 -74.032 54.658
256844 94.067 -74.033 54.609
256845 94.096 -74.032 54.566
256846 94.122 -74.031 54.537
256847 94.148 -74.029 54.517
256848 94.175 -74.028 54.499
256849 94.205 -74.027 54.485
256850 94.237 -74.026 54.470
256851 94.268 -74.026 54.452
256852 94.299 -74.025 54.431
256853 94.326 -74.026 54.406
256854 94.342 -74.033 54.379
256855 94.358 -74.039 54.353
256856 94.374 -74.046 54.329
256857 94.387 -74.050 54.303
256858 94.385 -74.043 54.278
256859 94.401 -74.038 54.251
256860 94.427 -74.035 54.222
256861 94.454 -74.032 54.188
256862 94.480 -74.030 54.155
256863 94.506 -74.027 54.124
256864 94.533 -74.024 54.094
256865 94.559 -74.023 54.058
256866 94.581 -74.020 54.018
256867 94.601 -74.010 53.974
256868 94.621 -74.001 53.926
256869 94.640 -73.991 53.880
256870 94.661 -73.982 53.834
256871 94.679 -73.974 53.788
256872 94.694 -73.976 53.741
256873 94.709 -73.977 53.696
256874 94.724 -73.979 53.652
256875 94.739 -73.980 53.610
256876 94.754 -73.982 53.572
256877 94.769 -73.983 53.541
256878 94.791 -73.988 53.516
256879 94.838 -74.005 53.497
256880 94.881 -74.023 53.483
256881 94.891 -74.050 53.464
256882 94.901 -74.076 53.439
256883 94.911 -74.103 53.407
256884 94.928 -74.111 53.364
256885 94.950 -74.105 53.320
256886 94.971 -74.100 53.275
256887 94.992 -74.094 53.228
256888 95.014 -74.089 53.180
256889 95.035 -74.083 53.130
256890 95.063 -74.085 53.081
256891 95.097 -74.090 53.032
256892 95.131 -74.095 52.984
256893 95.164 -74.101 52.935
256894 95.198 -74.107 52.886
256895 95.231 -74.113 52.838
256896 95.254 -74.144 52.789
256897 95.288 -74.154 52.738
256898 95.329 -74.153 52.686
256899 95.371 -74.152 52.633
256900 95.412 -74.151 52.581
256901 95.454 -74.150 52.532
256902 95.495 -74.149 52.485
256903 95.537 -74.148 52.440
256904 95.578 -74.147 52.396
256905 95.620 -74.146 52.354
256906 95.661 -74.145 52.312
256907 95.702 -74.145 52.272
256908 95.744 -74.144 52.234
256909 95.774 -74.149 52.196
256910 95.799 -74.153 52.159
256911 95.809 -74.149 52.122
256912 95.819 -74.146 52.084
256913 95.828 -74.142 52.046
256914 95.838 -74.139 52.009
256915 95.848 -74.136 51.972
256916 95.858 -74.132 51.937
256917 95.868 -74.129 51.901
256918 95.877 -74.125 51.865
256919 95.887 -74.121 51.828
256920 95.897 -74.118 51.791
256921 95.915 -74.129 51.753
256922 95.939 -74.149 51.715
256923 95.963 -74.169 51.676
256924 95.994 -74.170 51.635
256925 96.028 -74.167 51.594
256926 96.060 -74.166 51.554
256927 96.094 -74.164 51.513
256928 96.127 -74.161 51.475
256929 96.160 -74.159 51.438
256930 96.209 -74.170 51.402
256931 96.222 -74.192 51.368
256932 96.226 -74.210 51.334
256933 96.230 -74.227 51.302
256934 96.240 -74.243 51.272
256935 96.274 -74.262 51.243
256936 96.308 -74.280 51.216
256937 96.343 -74.299 51.189
256938 96.377 -74.318 51.164
256939 96.411 -74.337 51.140
256940 96.446 -74.356 51.118
256941 96.480 -74.375 51.098
256942 96.514 -74.394 51.079
256943 96.548 -74.413 51.062
256944 96.583 -74.432 51.047
256945 96.617 -74.451 51.033
256946 96.648 -74.465 51.020
256947 96.673 -74.473 51.009
256948 96.698 -74.480 50.999
256949 96.723 -74.487 50.991
256950 96.752 -74.503 50.985
256951 96.783 -74.522 50.981
256952 96.805 -74.537 50.978
256953 96.826 -74.550 50.977
256954 96.848 -74.561 50.977
256955 96.869 -74.566 50.980
256956 96.889 -74.571 50.983
256957 96.910 -74.577 50.988
256958 96.939 -74.597 50.993
256959 96.956 -74.609 50.997
256960 96.973 -74.621 51.002
256961 96.991 -74.634 51.004
256962 97.008 -74.645 51.003
256963 97.025 -74.657 51.002
256964 97.042 -74.669 50.999
256965 97.060 -74.681 50.997
256966 97.077 -74.693 50.994
256967 97.094 -74.706 50.989
256968 97.111 -74.718 50.982
256969 97.128 -74.730 50.972
256970 97.145 -74.741 50.958
256971 97.162 -74.753 50.940
256972 97.179 -74.766 50.916
256973 97.216 -74.760 50.890
256974 97.263 -74.747 50.860
256975 97.310 -74.734 50.829
256976 97.357 -74.721 50.799
256977 97.404 -74.707 50.770
256978 97.451 -74.694 50.741
256979 97.498 -74.681 50.709
256980 97.544 -74.668 50.674
256981 97.592 -74.654 50.635
256982 97.639 -74.641 50.593
256983 97.685 -74.628 50.551
256984 97.721 -74.632 50.506
256985 97.747 -74.648 50.460
256986 97.778 -74.666 50.412
256987 97.815 -74.690 50.361
256988 97.852 -74.712 50.309
256989 97.880 -74.727 50.257
256990 97.900 -74.731 50.208
256991 97.921 -74.737 50.162
256992 97.942 -74.741 50.118
256993 97.962 -74.746 50.077
256994 97.983 -74.751 50.038
256995 98.004 -74.756 49.999
256996 98.024 -74.761 49.958
256997 98.045 -74.766 49.914
256998 98.076 -74.787 49.868
256999 98.107 -74.809 49.818
257000 98.138 -74.831 49.767
257001 98.169 -74.853 49.713
257002 98.200 -74.875 49.656
257003 98.231 -74.888 49.597
257004 98.263 -74.901 49.537
257005 98.294 -74.915 49.478
257006 98.325 -74.928 49.421
257007 98.356 -74.941 49.366
257008 98.390 -74.957 49.312
257009 98.425 -74.954 49.259
257010 98.454 -74.925 49.207
257011 98.484 -74.896 49.153
257012 98.510 -74.869 49.100
257013 98.530 -74.847 49.048
257014 98.551 -74.831 48.997
257015 98.579 -74.842 48.945
257016 98.606 -74.853 48.892
257017 98.633 -74.864 48.838
257018 98.661 -74.875 48.783
257019 98.688 -74.887 48.728
257020 98.716 -74.897 48.673
257021 98.743 -74.909 48.619
257022 98.771 -74.919 48.565
257023 98.798 -74.931 48.513
257024 98.825 -74.941 48.460
257025 98.853 -74.951 48.409
257026 98.878 -74.972 48.358
257027 98.903 -74.990 48.307
257028 98.927 -75.008 48.257
257029 98.952 -75.026 48.208
257030 98.977 -75.044 48.159
257031 99.001 -75.063 48.111
257032 99.026 -75.082 48.064
257033 99.051 -75.100 48.018
257034 99.075 -75.118 47.972
257035 99.100 -75.136 47.927
257036 99.125 -75.154 47.881
257037 99.160 -75.174 47.835
257038 99.196 -75.194 47.789
257039 99.231 -75.213 47.744
257040 99.268 -75.232 47.697
257041 99.300 -75.245 47.652
257042 99.331 -75.248 47.609
257043 99.361 -75.252 47.567
257044 99.391 -75.257 47.526
257045 99.420 -75.261 47.486
257046 99.444 -75.270 47.447
257047 99.469 -75.278 47.411
257048 99.496 -75.285 47.378
257049 99.532 -75.289 47.347
257050 99.568 -75.292 47.318
257051 99.604 -75.295 47.290
257052 99.639 -75.298 47.264
257053 99.676 -75.301 47.240
257054 99.711 -75.305 47.216
257055 99.747 -75.308 47.193
257056 99.783 -75.312 47.172
257057 99.819 -75.315 47.151
257058 99.845 -75.329 47.130
257059 99.858 -75.356 47.109
257060 99.871 -75.383 47.087
257061 99.887 -75.415 47.067
257062 99.914 -75.445 47.047
257063 99.944 -75.470 47.030
257064 99.969 -75.477 47.014
257065 99.991 -75.475 46.996
257066 100.037 -75.482 46.976
257067 100.085 -75.490 46.952
257068 100.104 -75.493 46.929
257069 100.113 -75.493 46.904
257070 100.121 -75.493 46.877
257071 100.130 -75.494 46.846
257072 100.138 -75.499 46.812
257073 100.144 -75.516 46.778
257074 100.174 -75.524 46.743
257075 100.204 -75.534 46.709
257076 100.232 -75.557 46.676
257077 100.260 -75.579 46.644
257078 100.288 -75.601 46.612
257079 100.316 -75.623 46.580
257080 100.343 -75.645 46.547
257081 100.371 -75.668 46.516
257082 100.400 -75.694 46.487
257083 100.429 -75.721 46.457
257084 100.458 -75.748 46.428
257085 100.486 -75.775 46.399
257086 100.515 -75.802 46.370
257087 100.544 -75.829 46.341
257088 100.571 -75.835 46.313
257089 100.612 -75.828 46.286
257090 100.650 -75.841 46.258
257091 100.682 -75.860 46.229
257092 100.715 -75.880 46.201
257093 100.747 -75.900 46.172
257094 100.780 -75.919 46.142
257095 100.812 -75.939 46.114
257096 100.845 -75.959 46.086
257097 100.877 -75.978 46.058
257098 100.901 -75.994 46.029
257099 100.920 -76.006 45.998
257100 100.939 -76.017 45.967
257101 100.958 -76.028 45.936
257102 100.977 -76.039 45.906
257103 100.993 -76.047 45.877
257104 101.008 -76.053 45.848
257105 101.023 -76.060 45.818
257106 101.062 -76.070 45.787
257107 101.104 -76.082 45.757
257108 101.145 -76.093 45.727
257109 101.187 -76.104 45.699
257110 101.229 -76.115 45.670
257111 101.270 -76.126 45.644
257112 101.312 -76.138 45.617
257113 101.353 -76.149 45.590
257114 101.395 -76.160 45.564
257115 101.437 -76.172 45.538
257116 101.479 -76.183 45.513
257117 101.521 -76.195 45.489
257118 101.562 -76.206 45.465
257119 101.605 -76.216 45.440
257120 101.649 -76.226 45.417
257121 101.683 -76.236 45.394
257122 101.709 -76.246 45.370
257123 101.740 -76.259 45.348
257124 101.748 -76.263 45.325
257125 101.756 -76.265 45.302
257126 101.763 -76.268 45.279
257127 101.771 -76.270 45.256
257128 101.778 -76.273 45.234
257129 101.786 -76.276 45.211
257130 101.794 -76.279 45.188
257131 101.802 -76.282 45.165
257132 101.809 -76.284 45.143
257133 101.817 -76.287 45.121
257134 101.824 -76.289 45.098
257489 92.637 -73.920 54.633
257490 92.661 -73.932 54.632
257491 92.686 -73.942 54.631
257492 92.710 -73.953 54.631
257493 92.733 -73.943 54.632
257494 92.763 -73.935 54.632
257495 92.792 -73.938 54.633
257496 92.822 -73.957 54.634
257497 92.852 -73.976 54.635
257498 92.881 -73.994 54.637
257499 92.911 -74.013 54.639
257500 92.941 -74.032 54.641
257501 92.971 -74.050 54.645
257502 93.000 -74.069 54.647
257503 93.031 -74.072 54.650
257504 93.062 -74.071 54.653
257505 93.096 -74.050 54.657
257506 93.129 -74.029 54.659
257507 93.163 -74.009 54.660
257508 93.196 -73.989 54.661
257509 93.217 -73.997 54.661
257510 93.238 -74.005 54.664
257511 93.259 -74.010 54.667
257512 93.279 -74.014 54.670
257513 93.300 -74.020 54.672
257514 93.320 -74.024 54.673
257515 93.348 -74.027 54.672
257516 93.381 -74.029 54.670
257517 93.414 -74.030 54.668
257518 93.438 -74.020 54.667
257519 93.461 -74.007 54.669
257520 93.483 -73.995 54.674
257521 93.507 -73.990 54.681
257522 93.532 -73.985 54.688
257523 93.556 -73.982 54.696
257524 93.580 -73.977 54.706
257525 93.605 -73.973 54.717
257526 93.629 -73.969 54.734
257527 93.653 -73.965 54.751
257528 93.678 -73.961 54.767
257529 93.707 -73.960 54.781
257530 93.765 -73.982 54.799
257531 93.782 -73.986 54.820
257532 93.792 -73.986 54.844
257533 93.802 -73.987 54.870
257534 93.819 -73.988 54.892
257535 93.840 -73.989 54.905
257536 93.861 -73.991 54.907
257537 93.882 -73.992 54.898
257538 93.903 -73.995 54.878
257539 93.925 -73.998 54.847
257540 93.946 -74.001 54.810
257541 93.968 -74.002 54.771
257542 93.990 -74.004 54.727
257543 94.011 -74.006 54.683
257544 94.040 -74.007 54.637
257545 94.071 -74.007 54.588
257546 94.102 -74.008 54.549
257547 94.128 -74.007 54.522
257548 94.155 -74.006 54.503
257549 94.184 -74.004 54.485
257550 94.215 -74.004 54.472
257551 94.247 -74.003 54.457
257552 94.278 -74.002 54.438
257553 94.309 -74.001 54.415
257554 94.340 -74.001 54.388
257555 94.362 -74.004 54.360
257556 94.379 -74.010 54.333
257557 94.395 -74.017 54.309
257558 94.407 -74.021 54.283
257559 94.405 -74.014 54.257
257560 94.404 -74.007 54.231
257561 94.426 -74.004 54.202
257562 94.454 -74.003 54.168
257563 94.481 -74.001 54.134
257564 94.508 -74.000 54.101
257565 94.535 -73.999 54.072
257566 94.562 -74.000 54.037
257567 94.582 -74.007 53.996
257568 94.601 -74.001 53.952
257569 94.621 -73.991 53.905
257570 94.640 -73.982 53.859
257571 94.661 -73.973 53.814
257572 94.680 -73.963 53.769
257573 94.696 -73.961 53.721
257574 94.711 -73.963 53.677
257575 94.726 -73.964 53.634
257576 94.742 -73.966 53.594
257577 94.756 -73.967 53.560
257578 94.771 -73.969 53.532
257579 94.794 -73.974 53.510
257580 94.840 -73.991 53.493
257581 94.887 -74.008 53.478
257582 94.910 -74.032 53.455
257583 94.920 -74.058 53.427
257584 94.930 -74.085 53.394
257585 94.947 -74.095 53.350
257586 94.968 -74.089 53.306
257587 94.989 -74.084 53.260
257588 95.011 -74.079 53.214
257589 95.032 -74.073 53.166
257590 95.054 -74.068 53.118
257591 95.075 -74.063 53.070
257592 95.097 -74.057 53.020
257593 95.125 -74.059 52.971
257594 95.159 -74.064 52.922
257595 95.192 -74.070 52.874
257596 95.226 -74.076 52.825
257597 95.239 -74.121 52.775
257598 95.277 -74.127 52.724
257599 95.319 -74.129 52.674
257600 95.361 -74.129 52.622
257601 95.404 -74.128 52.572
257602 95.446 -74.127 52.524
257603 95.489 -74.126 52.478
257604 95.532 -74.126 52.434
257605 95.574 -74.125 52.392
257606 95.617 -74.124 52.350
257607 95.659 -74.123 52.309
257608 95.702 -74.123 52.268
257609 95.745 -74.122 52.230
257610 95.786 -74.121 52.193
257611 95.796 -74.118 52.156
257612 95.806 -74.115 52.120
257613 95.816 -74.111 52.082
257614 95.825 -74.107 52.044
257615 95.835 -74.104 52.007
257616 95.845 -74.101 51.971
257617 95.855 -74.097 51.936
257618 95.865 -74.094 51.901
257619 95.874 -74.090 51.865
257620 95.884 -74.086 51.829
257621 95.894 -74.083 51.794
257622 95.909 -74.089 51.757
257623 95.933 -74.109 51.718
257624 95.960 -74.120 51.678
257625 95.993 -74.118 51.637
257626 96.026 -74.116 51.597
257627 96.059 -74.114 51.557
257628 96.093 -74.112 51.516
257629 96.125 -74.110 51.478
257630 96.156 -74.122 51.441
257631 96.185 -74.135 51.405
257632 96.217 -74.136 51.370
257633 96.253 -74.142 51.336
257634 96.256 -74.161 51.304
257635 96.253 -74.205 51.274
257636 96.272 -74.231 51.245
257637 96.304 -74.248 51.217
257638 96.335 -74.266 51.190
257639 96.366 -74.283 51.164
257640 96.399 -74.302 51.140
257641 96.433 -74.320 51.118
257642 96.467 -74.339 51.097
257643 96.502 -74.358 51.078
257644 96.536 -74.377 51.060
257645 96.570 -74.396 51.045
257646 96.604 -74.415 51.030
257647 96.639 -74.434 51.017
257648 96.670 -74.449 51.005
257649 96.695 -74.456 50.995
257650 96.720 -74.464 50.987
257651 96.750 -74.482 50.982
257652 96.774 -74.497 50.977
257653 96.796 -74.510 50.974
257654 96.816 -74.515 50.973
257655 96.837 -74.521 50.973
257656 96.858 -74.526 50.975
257657 96.879 -74.531 50.979
257658 96.904 -74.541 50.985
257659 96.933 -74.562 50.992
257660 96.950 -74.574 51.000
257661 96.967 -74.586 51.007
257662 96.984 -74.598 51.011
257663 97.002 -74.610 51.013
257664 97.019 -74.622 51.011
257665 97.036 -74.634 51.008
257666 97.053 -74.646 51.006
257667 97.071 -74.659 51.003
257668 97.088 -74.671 50.998
257669 97.105 -74.683 50.990
257670 97.122 -74.694 50.979
257671 97.139 -74.706 50.965
257672 97.167 -74.708 50.946
257673 97.202 -74.701 50.922
257674 97.241 -74.693 50.895
257675 97.288 -74.680 50.866
257676 97.335 -74.666 50.836
257677 97.382 -74.653 50.806
257678 97.429 -74.640 50.777
257679 97.476 -74.627 50.746
257680 97.523 -74.613 50.712
257681 97.570 -74.600 50.674
257682 97.616 -74.587 50.634
257683 97.663 -74.574 50.590
257684 97.698 -74.579 50.547
257685 97.724 -74.595 50.501
257686 97.751 -74.611 50.455
257687 97.779 -74.628 50.407
257688 97.816 -74.650 50.356
257689 97.853 -74.674 50.303
257690 97.884 -74.690 50.251
257691 97.907 -74.695 50.202
257692 97.929 -74.701 50.156
257693 97.951 -74.707 50.112
257694 97.972 -74.712 50.072
257695 97.993 -74.716 50.034
257696 98.013 -74.722 49.997
257697 98.034 -74.726 49.956
257698 98.055 -74.731 49.912
257699 98.086 -74.753 49.865
257700 98.117 -74.775 49.815
257701 98.148 -74.797 49.762
257702 98.179 -74.819 49.707
257703 98.210 -74.841 49.649
257704 98.241 -74.863 49.589
257705 98.272 -74.878 49.529
257706 98.303 -74.891 49.470
257707 98.334 -74.905 49.412
257708 98.365 -74.918 49.357
257709 98.402 -74.937 49.303
257710 98.432 -74.913 49.250
257711 98.461 -74.883 49.197
257712 98.491 -74.853 49.144
257713 98.516 -74.827 49.091
257714 98.536 -74.805 49.039
257715 98.561 -74.804 48.987
257716 98.588 -74.816 48.936
257717 98.616 -74.826 48.883
257718 98.643 -74.838 48.830
257719 98.670 -74.848 48.775
257720 98.698 -74.859 48.720
257721 98.725 -74.870 48.664
257722 98.752 -74.881 48.610
257723 98.780 -74.893 48.556
257724 98.807 -74.903 48.504
257725 98.834 -74.922 48.452
257726 98.863 -74.934 48.400
257727 98.886 -74.943 48.349
257728 98.911 -74.961 48.299
257729 98.936 -74.979 48.249
257730 98.960 -74.997 48.200
257731 98.985 -75.016 48.151
257732 99.010 -75.034 48.103
257733 99.034 -75.052 48.057
257734 99.059 -75.070 48.011
257735 99.084 -75.089 47.965
257736 99.109 -75.107 47.919
257737 99.134 -75.126 47.874
257738 99.170 -75.145 47.828
257739 99.206 -75.165 47.782
257740 99.242 -75.184 47.737
257741 99.278 -75.204 47.689
257742 99.313 -75.221 47.644
257743 99.343 -75.225 47.601
257744 99.373 -75.229 47.559
257745 99.403 -75.233 47.519
257746 99.430 -75.239 47.479
257747 99.454 -75.248 47.441
257748 99.480 -75.256 47.404
257749 99.516 -75.259 47.371
257750 99.552 -75.263 47.340
257751 99.587 -75.266 47.311
257752 99.623 -75.269 47.283
257753 99.659 -75.273 47.257
257754 99.695 -75.276 47.232
257755 99.730 -75.279 47.208
257756 99.766 -75.283 47.187
257757 99.802 -75.286 47.165
257758 99.838 -75.289 47.143
257759 99.874 -75.293 47.122
257760 99.894 -75.313 47.101
257761 99.907 -75.339 47.079
257762 99.926 -75.380 47.059
257763 99.947 -75.424 47.040
257764 99.976 -75.450 47.021
257765 99.998 -75.449 47.004
257766 100.020 -75.446 46.985
257767 100.060 -75.451 46.965
257768 100.107 -75.458 46.942
257769 100.122 -75.460 46.918
257770 100.131 -75.461 46.892
257771 100.139 -75.461 46.864
257772 100.148 -75.461 46.834
257773 100.157 -75.462 46.801
257774 100.164 -75.468 46.767
257775 100.188 -75.479 46.733
257776 100.216 -75.499 46.700
257777 100.244 -75.521 46.667
257778 100.272 -75.543 46.635
257779 100.300 -75.568 46.603
257780 100.329 -75.596 46.572
257781 100.358 -75.622 46.540
257782 100.387 -75.649 46.509
257783 100.416 -75.676 46.480
257784 100.444 -75.703 46.451
257785 100.473 -75.730 46.422
257786 100.502 -75.757 46.393
257787 100.531 -75.784 46.364
257788 100.559 -75.811 46.335
257789 100.586 -75.806 46.308
257790 100.622 -75.798 46.281
257791 100.671 -75.795 46.253
257792 100.704 -75.815 46.225
257793 100.737 -75.834 46.197
257794 100.769 -75.853 46.168
257795 100.802 -75.873 46.139
257796 100.834 -75.893 46.112
257797 100.865 -75.916 46.084
257798 100.897 -75.942 46.057
257799 100.922 -75.961 46.028
257800 100.941 -75.972 45.998
257801 100.961 -75.983 45.967
257802 100.980 -75.994 45.936
257803 100.995 -76.001 45.906
257804 101.010 -76.007 45.877
257805 101.025 -76.013 45.846
257806 101.040 -76.020 45.816
257807 101.065 -76.028 45.785
257808 101.107 -76.039 45.754
257809 101.149 -76.050 45.725
257810 101.191 -76.062 45.696
257811 101.232 -76.073 45.667
257812 101.274 -76.085 45.641
257813 101.316 -76.096 45.613
257814 101.357 -76.107 45.586
257815 101.399 -76.119 45.559
257816 101.441 -76.130 45.533
257817 101.482 -76.141 45.507
257818 101.526 -76.151 45.482
257819 101.569 -76.161 45.458
257820 101.613 -76.171 45.433
257821 101.650 -76.181 45.409
257822 101.677 -76.191 45.386
257823 101.708 -76.205 45.362
257824 101.744 -76.229 45.339
257825 101.752 -76.231 45.316
257826 101.759 -76.234 45.293
257827 101.767 -76.236 45.271
257828 101.774 -76.239 45.248
257829 101.782 -76.242 45.225
257830 101.790 -76.245 45.202
257831 101.798 -76.248 45.179
257832 101.805 -76.250 45.155
257833 101.813 -76.253 45.133
257834 101.821 -76.255 45.111
257835 101.828 -76.258 45.088
258189 92.615 -73.877 54.635
258190 92.639 -73.888 54.635
258191 92.663 -73.899 54.634
258192 92.688 -73.910 54.633
258193 92.712 -73.920 54.633
258194 92.735 -73.916 54.633
258195 92.764 -73.907 54.634
258196 92.794 -73.912 54.635
258197 92.824 -73.931 54.636
258198 92.853 -73.949 54.637
258199 92.882 -73.968 54.638
258200 92.912 -73.987 54.640
258201 92.942 -74.005 54.644
258202 92.972 -74.024 54.647
258203 93.001 -74.043 54.651
258204 93.032 -74.051 54.653
258205 93.065 -74.031 54.655
258206 93.099 -74.010 54.660
258207 93.132 -73.990 54.662
258208 93.166 -73.969 54.663
258209 93.192 -73.965 54.662
258210 93.213 -73.970 54.662
258211 93.233 -73.976 54.664
258212 93.254 -73.980 54.667
258213 93.274 -73.985 54.670
258214 93.295 -73.990 54.673
258215 93.316 -73.995 54.675
258216 93.349 -73.996 54.674
258217 93.382 -73.998 54.671
258218 93.415 -73.999 54.669
258219 93.442 -73.993 54.668
258220 93.465 -73.980 54.667
258221 93.487 -73.967 54.668
258222 93.509 -73.954 54.673
258223 93.534 -73.950 54.677
258224 93.558 -73.945 54.682
258225 93.582 -73.941 54.689
258226 93.607 -73.938 54.698
258227 93.631 -73.933 54.709
258228 93.655 -73.929 54.722
258229 93.697 -73.938 54.736
258230 93.754 -73.960 54.751
258231 93.770 -73.963 54.769
258232 93.780 -73.963 54.792
258233 93.790 -73.964 54.818
258234 93.800 -73.964 54.844
258235 93.819 -73.965 54.865
258236 93.840 -73.966 54.876
258237 93.861 -73.968 54.878
258238 93.882 -73.969 54.873
258239 93.903 -73.972 54.857
258240 93.924 -73.973 54.830
258241 93.946 -73.976 54.793
258242 93.968 -73.977 54.752
258243 93.989 -73.979 54.707
258244 94.013 -73.981 54.660
258245 94.044 -73.982 54.612
258246 94.075 -73.982 54.564
258247 94.106 -73.983 54.530
258248 94.135 -73.983 54.506
258249 94.163 -73.982 54.488
258250 94.194 -73.981 54.469
258251 94.225 -73.980 54.457
258252 94.256 -73.979 54.442
258253 94.287 -73.979 54.423
258254 94.319 -73.978 54.399
258255 94.350 -73.977 54.371
258256 94.381 -73.976 54.342
258257 94.399 -73.982 54.313
258258 94.415 -73.988 54.288
258259 94.427 -73.992 54.262
258260 94.425 -73.985 54.236
258261 94.422 -73.980 54.209
258262 94.429 -73.979 54.180
258263 94.457 -73.978 54.147
258264 94.484 -73.976 54.113
258265 94.511 -73.975 54.079
258266 94.539 -73.973 54.049
258267 94.563 -73.981 54.014
258268 94.584 -73.992 53.974
258269 94.602 -73.990 53.929
258270 94.621 -73.982 53.883
258271 94.640 -73.973 53.837
258272 94.660 -73.963 53.793
258273 94.680 -73.954 53.750
258274 94.699 -73.947 53.704
258275 94.713 -73.948 53.660
258276 94.729 -73.950 53.619
258277 94.744 -73.951 53.581
258278 94.759 -73.953 53.549
258279 94.773 -73.954 53.525
258280 94.796 -73.960 53.505
258281 94.843 -73.977 53.489
258282 94.890 -73.994 53.470
258283 94.929 -74.013 53.444
258284 94.939 -74.040 53.413
258285 94.950 -74.067 53.377
258286 94.965 -74.079 53.335
258287 94.986 -74.074 53.292
258288 95.007 -74.069 53.248
258289 95.029 -74.064 53.204
258290 95.050 -74.058 53.158
258291 95.072 -74.053 53.109
258292 95.093 -74.048 53.059
258293 95.115 -74.042 53.007
258294 95.136 -74.037 52.958
258295 95.158 -74.032 52.908
258296 95.187 -74.033 52.859
258297 95.220 -74.039 52.810
258298 95.228 -74.090 52.760
258299 95.267 -74.095 52.710
258300 95.307 -74.101 52.660
258301 95.347 -74.105 52.610
258302 95.390 -74.104 52.562
258303 95.432 -74.104 52.516
258304 95.475 -74.103 52.472
258305 95.517 -74.102 52.428
258306 95.560 -74.101 52.386
258307 95.603 -74.101 52.345
258308 95.645 -74.100 52.305
258309 95.688 -74.099 52.265
258310 95.731 -74.098 52.227
258311 95.766 -74.096 52.190
258312 95.779 -74.091 52.154
258313 95.794 -74.085 52.117
258314 95.807 -74.079 52.080
258315 95.821 -74.073 52.042
258316 95.832 -74.070 52.006
258317 95.842 -74.066 51.971
258318 95.852 -74.062 51.935
258319 95.862 -74.059 51.901
258320 95.872 -74.055 51.866
258321 95.881 -74.052 51.832
258322 95.891 -74.048 51.797
258323 95.903 -74.049 51.760
258324 95.927 -74.069 51.722
258325 95.959 -74.069 51.681
258326 95.992 -74.067 51.640
258327 96.025 -74.064 51.599
258328 96.057 -74.070 51.560
258329 96.087 -74.082 51.519
258330 96.116 -74.095 51.481
258331 96.146 -74.107 51.444
258332 96.176 -74.120 51.408
258333 96.206 -74.126 51.373
258334 96.238 -74.123 51.338
258335 96.271 -74.124 51.306
258336 96.275 -74.161 51.276
258337 96.277 -74.203 51.246
258338 96.308 -74.220 51.218
258339 96.339 -74.239 51.190
258340 96.370 -74.256 51.164
258341 96.401 -74.273 51.141
258342 96.433 -74.291 51.118
258343 96.464 -74.308 51.097
258344 96.495 -74.326 51.077
258345 96.526 -74.343 51.059
258346 96.557 -74.361 51.043
258347 96.592 -74.380 51.028
258348 96.626 -74.399 51.014
258349 96.661 -74.418 51.003
258350 96.692 -74.433 50.993
258351 96.718 -74.443 50.985
258352 96.743 -74.458 50.979
258353 96.764 -74.465 50.974
258354 96.785 -74.470 50.970
258355 96.805 -74.476 50.968
258356 96.826 -74.481 50.969
258357 96.847 -74.487 50.970
258358 96.868 -74.492 50.976
258359 96.899 -74.508 50.983
258360 96.927 -74.527 50.992
258361 96.944 -74.539 51.003
258362 96.961 -74.551 51.012
258363 96.978 -74.563 51.018
258364 96.996 -74.575 51.020
258365 97.013 -74.587 51.019
258366 97.030 -74.599 51.017
258367 97.047 -74.612 51.014
258368 97.064 -74.624 51.011
258369 97.082 -74.635 51.007
258370 97.099 -74.647 50.998
258371 97.124 -74.651 50.987
258372 97.159 -74.645 50.972
258373 97.194 -74.639 50.951
258374 97.230 -74.632 50.927
258375 97.266 -74.625 50.901
258376 97.313 -74.612 50.873
258377 97.360 -74.599 50.843
258378 97.407 -74.585 50.813
258379 97.454 -74.572 50.782
258380 97.501 -74.559 50.749
258381 97.547 -74.546 50.714
258382 97.594 -74.532 50.674
258383 97.642 -74.519 50.632
258384 97.674 -74.526 50.588
258385 97.702 -74.542 50.544
258386 97.728 -74.558 50.498
258387 97.755 -74.574 50.451
258388 97.782 -74.590 50.402
258389 97.818 -74.612 50.352
258390 97.855 -74.635 50.297
258391 97.885 -74.650 50.246
258392 97.908 -74.656 50.197
258393 97.931 -74.662 50.152
258394 97.953 -74.668 50.109
258395 97.975 -74.674 50.067
258396 97.998 -74.680 50.028
258397 98.021 -74.686 49.993
258398 98.043 -74.692 49.952
258399 98.064 -74.697 49.908
258400 98.095 -74.718 49.860
258401 98.126 -74.740 49.809
258402 98.157 -74.762 49.754
258403 98.189 -74.784 49.698
258404 98.219 -74.806 49.639
258405 98.250 -74.828 49.578
258406 98.282 -74.850 49.518
258407 98.312 -74.869 49.460
258408 98.344 -74.881 49.404
258409 98.375 -74.894 49.347
258410 98.409 -74.900 49.293
258411 98.439 -74.871 49.240
258412 98.468 -74.841 49.187
258413 98.498 -74.812 49.134
258414 98.523 -74.786 49.081
258415 98.543 -74.766 49.028
258416 98.571 -74.777 48.978
258417 98.598 -74.788 48.927
258418 98.626 -74.799 48.875
258419 98.653 -74.810 48.821
258420 98.680 -74.822 48.766
258421 98.708 -74.832 48.711
258422 98.735 -74.844 48.656
258423 98.762 -74.859 48.601
258424 98.787 -74.886 48.547
258425 98.812 -74.913 48.494
258426 98.837 -74.941 48.442
258427 98.875 -74.916 48.391
258428 98.899 -74.916 48.340
258429 98.919 -74.932 48.290
258430 98.944 -74.950 48.241
258431 98.969 -74.969 48.192
258432 98.993 -74.987 48.143
258433 99.018 -75.005 48.095
258434 99.043 -75.023 48.048
258435 99.067 -75.041 48.001
258436 99.092 -75.060 47.956
258437 99.117 -75.078 47.911
258438 99.144 -75.097 47.865
258439 99.180 -75.116 47.820
258440 99.216 -75.136 47.775
258441 99.252 -75.155 47.728
258442 99.288 -75.175 47.682
258443 99.324 -75.195 47.637
258444 99.355 -75.202 47.593
258445 99.385 -75.206 47.551
258446 99.416 -75.210 47.511
258447 99.441 -75.217 47.471
258448 99.465 -75.226 47.433
258449 99.499 -75.230 47.397
258450 99.536 -75.233 47.363
258451 99.571 -75.236 47.332
258452 99.607 -75.240 47.304
258453 99.643 -75.243 47.276
258454 99.679 -75.247 47.250
258455 99.714 -75.250 47.225
258456 99.750 -75.254 47.201
258457 99.786 -75.257 47.179
258458 99.822 -75.261 47.157
258459 99.857 -75.264 47.136
258460 99.894 -75.267 47.115
258461 99.929 -75.270 47.093
258462 99.945 -75.300 47.071
258463 99.966 -75.344 47.050
258464 99.977 -75.392 47.030
258465 100.005 -75.420 47.011
258466 100.027 -75.420 46.993
258467 100.049 -75.417 46.974
258468 100.082 -75.420 46.953
258469 100.129 -75.427 46.929
258470 100.140 -75.428 46.905
258471 100.149 -75.429 46.879
258472 100.157 -75.429 46.851
258473 100.167 -75.430 46.820
258474 100.175 -75.430 46.788
258475 100.184 -75.430 46.755
258476 100.201 -75.442 46.722
258477 100.230 -75.469 46.690
258478 100.259 -75.496 46.657
258479 100.287 -75.523 46.625
258480 100.316 -75.550 46.594
258481 100.345 -75.577 46.564
258482 100.373 -75.604 46.533
258483 100.403 -75.630 46.503
258484 100.431 -75.658 46.473
258485 100.460 -75.684 46.444
258486 100.489 -75.712 46.416
258487 100.517 -75.738 46.387
258488 100.546 -75.765 46.358
258489 100.575 -75.789 46.330
258490 100.601 -75.778 46.302
258491 100.633 -75.768 46.275
258492 100.682 -75.764 46.248
258493 100.726 -75.768 46.221
258494 100.759 -75.787 46.192
258495 100.791 -75.807 46.163
258496 100.823 -75.828 46.135
258497 100.854 -75.854 46.110
258498 100.885 -75.879 46.082
258499 100.916 -75.905 46.054
258500 100.945 -75.927 46.026
258501 100.964 -75.938 45.997
258502 100.983 -75.950 45.966
258503 100.997 -75.956 45.935
258504 101.012 -75.962 45.905
258505 101.027 -75.968 45.874
258506 101.042 -75.974 45.842
258507 101.057 -75.980 45.811
258508 101.072 -75.986 45.780
258509 101.111 -75.997 45.749
258510 101.152 -76.009 45.720
258511 101.194 -76.020 45.692
258512 101.236 -76.032 45.664
258513 101.278 -76.042 45.637
258514 101.319 -76.053 45.610
258515 101.361 -76.065 45.582
258516 101.403 -76.076 45.555
258517 101.447 -76.086 45.527
258518 101.490 -76.096 45.501
258519 101.534 -76.107 45.475
258520 101.577 -76.116 45.451
258521 101.618 -76.126 45.426
258522 101.645 -76.136 45.402
258523 101.677 -76.151 45.378
258524 101.734 -76.189 45.354
258525 101.748 -76.197 45.331
258526 101.755 -76.200 45.308
258527 101.763 -76.202 45.284
258528 101.771 -76.205 45.262
258529 101.778 -76.208 45.238
258530 101.786 -76.210 45.215
258531 101.793 -76.213 45.192
258532 101.801 -76.216 45.169
258533 101.809 -76.219 45.146
258534 101.817 -76.221 45.123
258535 101.824 -76.224 45.101
258536 101.832 -76.227 45.077
258889 92.592 -73.834 54.638
258890 92.617 -73.845 54.638
258891 92.641 -73.856 54.637
258892 92.665 -73.866 54.636
258893 92.690 -73.877 54.636
258894 92.714 -73.888 54.635
258895 92.737 -73.889 54.634
258896 92.765 -73.881 54.635
258897 92.795 -73.886 54.635
258898 92.825 -73.904 54.636
258899 92.854 -73.923 54.637
258900 92.884 -73.942 54.640
258901 92.914 -73.960 54.641
258902 92.943 -73.979 54.644
258903 92.973 -73.998 54.648
258904 93.002 -74.017 54.652
258905 93.029 -74.021 54.654
258906 93.062 -74.001 54.656
258907 93.096 -73.979 54.661
258908 93.131 -73.957 54.663
258909 93.165 -73.935 54.664
258910 93.188 -73.936 54.664
258911 93.208 -73.941 54.663
258912 93.228 -73.946 54.663
258913 93.249 -73.951 54.665
258914 93.269 -73.955 54.669
258915 93.290 -73.960 54.673
258916 93.317 -73.963 54.675
258917 93.350 -73.965 54.674
258918 93.382 -73.966 54.672
258919 93.416 -73.968 54.670
258920 93.446 -73.966 54.668
258921 93.468 -73.953 54.666
258922 93.490 -73.940 54.665
258923 93.513 -73.927 54.665
258924 93.535 -73.914 54.667
258925 93.559 -73.910 54.670
258926 93.583 -73.906 54.673
258927 93.608 -73.901 54.679
258928 93.632 -73.897 54.686
258929 93.686 -73.916 54.695
258930 93.744 -73.938 54.707
258931 93.758 -73.941 54.720
258932 93.768 -73.941 54.739
258933 93.778 -73.941 54.762
258934 93.788 -73.941 54.787
258935 93.798 -73.941 54.811
258936 93.819 -73.943 54.828
258937 93.840 -73.944 54.840
258938 93.861 -73.945 54.844
258939 93.882 -73.946 54.838
258940 93.903 -73.947 54.824
258941 93.924 -73.948 54.800
258942 93.946 -73.951 54.767
258943 93.968 -73.952 54.726
258944 93.989 -73.954 54.681
258945 94.016 -73.956 54.633
258946 94.048 -73.957 54.585
258947 94.078 -73.957 54.539
258948 94.110 -73.958 54.509
258949 94.140 -73.959 54.486
258950 94.172 -73.958 54.468
258951 94.203 -73.957 54.451
258952 94.235 -73.957 54.439
258953 94.266 -73.956 54.424
258954 94.297 -73.955 54.405
258955 94.329 -73.954 54.380
258956 94.360 -73.954 54.353
258957 94.391 -73.953 54.324
258958 94.420 -73.954 54.294
258959 94.436 -73.960 54.268
258960 94.448 -73.963 54.240
258961 94.445 -73.956 54.213
258962 94.440 -73.954 54.185
258963 94.434 -73.954 54.157
258964 94.460 -73.953 54.125
258965 94.487 -73.951 54.092
258966 94.514 -73.950 54.057
258967 94.542 -73.948 54.024
258968 94.564 -73.963 53.990
258969 94.586 -73.976 53.950
258970 94.604 -73.974 53.906
258971 94.621 -73.972 53.860
258972 94.640 -73.963 53.815
258973 94.660 -73.954 53.772
258974 94.680 -73.944 53.730
258975 94.700 -73.935 53.686
258976 94.716 -73.934 53.644
258977 94.731 -73.935 53.604
258978 94.746 -73.937 53.569
258979 94.761 -73.938 53.539
258980 94.776 -73.940 53.517
258981 94.800 -73.946 53.500
258982 94.846 -73.963 53.483
258983 94.893 -73.980 53.460
258984 94.939 -73.997 53.431
258985 94.959 -74.021 53.398
258986 94.969 -74.048 53.361
258987 94.983 -74.064 53.320
258988 95.004 -74.059 53.278
258989 95.026 -74.054 53.237
258990 95.047 -74.048 53.194
258991 95.068 -74.043 53.148
258992 95.090 -74.038 53.098
258993 95.112 -74.032 53.046
258994 95.133 -74.027 52.993
258995 95.155 -74.022 52.943
258996 95.176 -74.016 52.893
258997 95.197 -74.010 52.844
258998 95.218 -74.005 52.795
258999 95.218 -74.058 52.745
259000 95.257 -74.064 52.695
259001 95.297 -74.069 52.647
259002 95.336 -74.073 52.599
259003 95.376 -74.079 52.552
259004 95.418 -74.079 52.508
259005 95.461 -74.079 52.465
259006 95.503 -74.078 52.422
259007 95.546 -74.077 52.381
259008 95.589 -74.076 52.341
259009 95.631 -74.076 52.301
259010 95.674 -74.075 52.262
259011 95.716 -74.074 52.225
259012 95.744 -74.071 52.188
259013 95.758 -74.066 52.152
259014 95.772 -74.060 52.115
259015 95.785 -74.054 52.078
259016 95.800 -74.048 52.041
259017 95.813 -74.043 52.005
259018 95.827 -74.038 51.970
259019 95.841 -74.032 51.935
259020 95.855 -74.026 51.901
259021 95.869 -74.020 51.867
259022 95.879 -74.017 51.833
259023 95.888 -74.013 51.799
259024 95.898 -74.010 51.762
259025 95.925 -74.020 51.723
259026 95.958 -74.017 51.682
259027 95.988 -74.030 51.642
259028 96.018 -74.042 51.602
259029 96.047 -74.054 51.562
259030 96.077 -74.067 51.522
259031 96.107 -74.079 51.483
259032 96.137 -74.092 51.446
259033 96.167 -74.104 51.409
259034 96.196 -74.117 51.374
259035 96.228 -74.115 51.340
259036 96.261 -74.117 51.308
259037 96.294 -74.123 51.277
259038 96.295 -74.164 51.247
259039 96.312 -74.193 51.219
259040 96.343 -74.211 51.192
259041 96.375 -74.228 51.166
259042 96.406 -74.245 51.142
259043 96.437 -74.263 51.119
259044 96.468 -74.280 51.097
259045 96.499 -74.298 51.077
259046 96.530 -74.315 51.059
259047 96.561 -74.333 51.042
259048 96.593 -74.350 51.027
259049 96.623 -74.368 51.014
259050 96.655 -74.385 51.002
259051 96.686 -74.402 50.992
259052 96.712 -74.415 50.983
259053 96.732 -74.420 50.976
259054 96.753 -74.425 50.971
259055 96.774 -74.430 50.967
259056 96.795 -74.436 50.965
259057 96.815 -74.441 50.966
259058 96.836 -74.446 50.967
259059 96.857 -74.452 50.972
259060 96.895 -74.474 50.980
259061 96.921 -74.492 50.990
259062 96.938 -74.504 51.002
259063 96.955 -74.516 51.012
259064 96.972 -74.528 51.019
259065 96.990 -74.540 51.023
259066 97.007 -74.552 51.025
259067 97.024 -74.565 51.025
259068 97.041 -74.577 51.023
259069 97.058 -74.588 51.019
259070 97.081 -74.595 51.015
259071 97.116 -74.589 51.006
259072 97.151 -74.582 50.994
259073 97.186 -74.576 50.978
259074 97.222 -74.569 50.957
259075 97.257 -74.563 50.933
259076 97.293 -74.556 50.907
259077 97.338 -74.544 50.878
259078 97.385 -74.530 50.848
259079 97.432 -74.518 50.816
259080 97.478 -74.504 50.783
259081 97.526 -74.491 50.749
259082 97.573 -74.477 50.713
259083 97.619 -74.465 50.673
259084 97.651 -74.473 50.630
259085 97.678 -74.489 50.586
259086 97.705 -74.505 50.540
259087 97.732 -74.521 50.494
259088 97.759 -74.537 50.446
259089 97.786 -74.552 50.397
259090 97.819 -74.573 50.344
259091 97.857 -74.596 50.291
259092 97.887 -74.612 50.241
259093 97.910 -74.618 50.194
259094 97.932 -74.623 50.149
259095 97.954 -74.629 50.105
259096 97.977 -74.635 50.063
259097 98.000 -74.641 50.023
259098 98.022 -74.647 49.987
259099 98.044 -74.653 49.945
259100 98.074 -74.665 49.901
259101 98.107 -74.678 49.852
259102 98.138 -74.699 49.799
259103 98.168 -74.723 49.743
259104 98.199 -74.748 49.685
259105 98.229 -74.772 49.625
259106 98.260 -74.794 49.565
259107 98.291 -74.816 49.506
259108 98.322 -74.838 49.448
259109 98.353 -74.859 49.391
259110 98.384 -74.872 49.337
259111 98.416 -74.859 49.284
259112 98.445 -74.829 49.230
259113 98.475 -74.800 49.177
259114 98.504 -74.770 49.122
259115 98.529 -74.744 49.069
259116 98.553 -74.739 49.018
259117 98.580 -74.750 48.967
259118 98.608 -74.761 48.917
259119 98.635 -74.772 48.865
259120 98.662 -74.783 48.811
259121 98.690 -74.795 48.756
259122 98.715 -74.822 48.702
259123 98.740 -74.850 48.646
259124 98.765 -74.876 48.591
259125 98.790 -74.903 48.537
259126 98.815 -74.931 48.484
259127 98.844 -74.943 48.433
259128 98.888 -74.898 48.381
259129 98.915 -74.890 48.331
259130 98.931 -74.904 48.281
259131 98.952 -74.921 48.232
259132 98.977 -74.939 48.183
259133 99.002 -74.957 48.135
259134 99.026 -74.975 48.087
259135 99.051 -74.994 48.039
259136 99.076 -75.013 47.992
259137 99.100 -75.031 47.946
259138 99.125 -75.049 47.901
259139 99.154 -75.068 47.855
259140 99.191 -75.088 47.810
259141 99.226 -75.107 47.766
259142 99.263 -75.126 47.719
259143 99.298 -75.146 47.673
259144 99.334 -75.166 47.629
259145 99.368 -75.179 47.585
259146 99.398 -75.182 47.543
259147 99.427 -75.188 47.503
259148 99.451 -75.195 47.464
259149 99.483 -75.201 47.426
259150 99.519 -75.204 47.389
259151 99.555 -75.207 47.356
259152 99.590 -75.210 47.325
259153 99.627 -75.214 47.296
259154 99.662 -75.217 47.269
259155 99.698 -75.221 47.242
259156 99.734 -75.224 47.217
259157 99.770 -75.228 47.193
259158 99.805 -75.231 47.171
259159 99.841 -75.235 47.149
259160 99.877 -75.238 47.128
259161 99.905 -75.245 47.107
259162 99.922 -75.258 47.085
259163 99.932 -75.285 47.062
259164 99.951 -75.322 47.040
259165 100.003 -75.339 47.020
259166 100.054 -75.355 47.000
259167 100.056 -75.391 46.981
259168 100.078 -75.389 46.961
259169 100.104 -75.388 46.941
259170 100.149 -75.395 46.916
259171 100.158 -75.396 46.891
259172 100.167 -75.396 46.864
259173 100.176 -75.397 46.836
259174 100.185 -75.398 46.807
259175 100.196 -75.406 46.775
259176 100.213 -75.427 46.743
259177 100.229 -75.439 46.710
259178 100.246 -75.452 46.679
259179 100.274 -75.478 46.647
259180 100.303 -75.505 46.616
259181 100.332 -75.532 46.586
259182 100.360 -75.558 46.555
259183 100.389 -75.586 46.525
259184 100.418 -75.612 46.495
259185 100.447 -75.640 46.467
259186 100.476 -75.666 46.437
259187 100.504 -75.693 46.409
259188 100.533 -75.720 46.380
259189 100.562 -75.747 46.351
259190 100.590 -75.760 46.323
259191 100.616 -75.748 46.296
259192 100.643 -75.737 46.270
259193 100.693 -75.734 46.243
259194 100.742 -75.731 46.216
259195 100.780 -75.742 46.187
259196 100.812 -75.766 46.158
259197 100.843 -75.791 46.129
259198 100.874 -75.817 46.104
259199 100.905 -75.843 46.077
259200 100.936 -75.869 46.050
259201 100.967 -75.894 46.022
259202 100.985 -75.903 45.993
259203 101.000 -75.910 45.962
259204 101.015 -75.916 45.930
259205 101.029 -75.922 45.898
259206 101.044 -75.928 45.868
259207 101.059 -75.934 45.836
259208 101.074 -75.941 45.805
259209 101.089 -75.947 45.774
259210 101.115 -75.955 45.744
259211 101.156 -75.966 45.716
259212 101.198 -75.978 45.688
259213 101.240 -75.989 45.660
259214 101.281 -76.000 45.633
259215 101.324 -76.011 45.606
259216 101.368 -76.021 45.578
259217 101.411 -76.032 45.549
259218 101.454 -76.041 45.521
259219 101.498 -76.051 45.494
259220 101.541 -76.061 45.468
259221 101.585 -76.072 45.443
259222 101.613 -76.082 45.418
259223 101.645 -76.097 45.394
259224 101.702 -76.135 45.369
259225 101.744 -76.163 45.347
259226 101.752 -76.166 45.323
259227 101.759 -76.168 45.299
259228 101.767 -76.171 45.275
259229 101.774 -76.174 45.251
259230 101.782 -76.176 45.229
259231 101.790 -76.179 45.205
259232 101.798 -76.182 45.182
259233 101.805 -76.185 45.159
259234 101.813 -76.188 45.135
259235 101.820 -76.190 45.113
259236 101.828 -76.193 45.090
259589 92.570 -73.791 54.641
259590 92.594 -73.802 54.640
259591 92.619 -73.813 54.640
259592 92.643 -73.823 54.638
259593 92.667 -73.834 54.638
259594 92.692 -73.844 54.637
259595 92.716 -73.856 54.636
259596 92.739 -73.863 54.635
259597 92.766 -73.854 54.634
259598 92.796 -73.860 54.634
259599 92.826 -73.879 54.635
259600 92.855 -73.897 54.637
259601 92.885 -73.916 54.639
259602 92.915 -73.935 54.641
259603 92.944 -73.954 54.644
259604 92.974 -73.970 54.648
259605 93.003 -73.976 54.652
259606 93.031 -73.982 54.654
259607 93.054 -73.979 54.657
259608 93.080 -73.972 54.660
259609 93.115 -73.949 54.663
259610 93.149 -73.927 54.664
259611 93.183 -73.907 54.665
259612 93.203 -73.911 54.664
259613 93.223 -73.916 54.662
259614 93.244 -73.921 54.663
259615 93.265 -73.926 54.665
259616 93.285 -73.931 54.668
259617 93.318 -73.932 54.671
259618 93.350 -73.934 54.672
259619 93.383 -73.935 54.671
259620 93.416 -73.937 54.669
259621 93.449 -73.939 54.666
259622 93.472 -73.926 54.664
259623 93.494 -73.913 54.662
259624 93.516 -73.901 54.659
259625 93.539 -73.888 54.658
259626 93.561 -73.875 54.658
259627 93.585 -73.869 54.659
259628 93.618 -73.872 54.661
259629 93.676 -73.894 54.665
259630 93.733 -73.916 54.671
259631 93.746 -73.917 54.679
259632 93.756 -73.918 54.690
259633 93.766 -73.918 54.708
259634 93.776 -73.918 54.729
259635 93.786 -73.919 54.751
259636 93.798 -73.919 54.773
259637 93.819 -73.919 54.786
259638 93.840 -73.920 54.797
259639 93.861 -73.921 54.804
259640 93.882 -73.922 54.799
259641 93.903 -73.922 54.786
259642 93.924 -73.924 54.765
259643 93.946 -73.926 54.736
259644 93.968 -73.928 54.698
259645 93.989 -73.929 54.652
259646 94.020 -73.930 54.604
259647 94.051 -73.931 54.557
259648 94.082 -73.932 54.514
259649 94.114 -73.931 54.485
259650 94.146 -73.929 54.463
259651 94.176 -73.936 54.445
259652 94.209 -73.936 54.429
259653 94.242 -73.935 54.418
259654 94.275 -73.932 54.404
259655 94.307 -73.932 54.385
259656 94.338 -73.931 54.361
259657 94.369 -73.930 54.335
259658 94.401 -73.929 54.306
259659 94.432 -73.929 54.276
259660 94.456 -73.931 54.248
259661 94.468 -73.934 54.219
259662 94.464 -73.929 54.190
259663 94.458 -73.929 54.160
259664 94.453 -73.929 54.130
259665 94.463 -73.927 54.098
259666 94.490 -73.926 54.065
259667 94.518 -73.924 54.032
259668 94.543 -73.930 53.999
259669 94.565 -73.944 53.964
259670 94.587 -73.959 53.925
259671 94.605 -73.959 53.881
259672 94.623 -73.957 53.835
259673 94.640 -73.954 53.791
259674 94.660 -73.944 53.749
259675 94.680 -73.935 53.708
259676 94.700 -73.926 53.666
259677 94.718 -73.919 53.627
259678 94.733 -73.921 53.591
259679 94.748 -73.922 53.557
259680 94.763 -73.924 53.531
259681 94.778 -73.926 53.510
259682 94.802 -73.932 53.491
259683 94.849 -73.949 53.474
259684 94.896 -73.966 53.447
259685 94.943 -73.983 53.415
259686 94.978 -74.003 53.381
259687 94.988 -74.029 53.343
259688 95.001 -74.048 53.305
259689 95.023 -74.043 53.266
259690 95.044 -74.038 53.226
259691 95.065 -74.032 53.185
259692 95.087 -74.027 53.138
259693 95.108 -74.022 53.085
259694 95.130 -74.017 53.031
259695 95.151 -74.011 52.978
259696 95.173 -74.006 52.927
259697 95.194 -74.001 52.877
259698 95.180 -74.037 52.827
259699 95.187 -74.026 52.779
259700 95.217 -74.028 52.729
259701 95.250 -74.032 52.681
259702 95.287 -74.037 52.634
259703 95.326 -74.042 52.589
259704 95.366 -74.047 52.544
259705 95.406 -74.052 52.500
259706 95.447 -74.055 52.458
259707 95.489 -74.054 52.417
259708 95.532 -74.054 52.376
259709 95.574 -74.053 52.337
259710 95.617 -74.052 52.298
259711 95.660 -74.051 52.259
259712 95.702 -74.051 52.222
259713 95.722 -74.046 52.186
259714 95.736 -74.040 52.149
259715 95.750 -74.035 52.113
259716 95.764 -74.029 52.075
259717 95.778 -74.023 52.039
259718 95.792 -74.018 52.003
259719 95.806 -74.012 51.969
259720 95.819 -74.007 51.935
259721 95.833 -74.001 51.901
259722 95.847 -73.995 51.868
259723 95.861 -73.990 51.834
259724 95.875 -73.984 51.800
259725 95.897 -73.979 51.763
259726 95.924 -73.989 51.724
259727 95.949 -74.002 51.684
259728 95.978 -74.015 51.644
259729 96.008 -74.027 51.604
259730 96.038 -74.039 51.564
259731 96.068 -74.052 51.524
259732 96.098 -74.064 51.485
259733 96.127 -74.076 51.448
259734 96.158 -74.089 51.411
259735 96.187 -74.101 51.375
259736 96.218 -74.106 51.342
259737 96.251 -74.109 51.309
259738 96.284 -74.115 51.278
259739 96.317 -74.121 51.248
259740 96.317 -74.165 51.220
259741 96.347 -74.183 51.193
259742 96.379 -74.200 51.167
259743 96.410 -74.217 51.143
259744 96.441 -74.235 51.120
259745 96.472 -74.252 51.098
259746 96.504 -74.270 51.077
259747 96.534 -74.287 51.059
259748 96.566 -74.305 51.042
259749 96.597 -74.322 51.027
259750 96.628 -74.337 51.013
259751 96.660 -74.351 51.001
259752 96.691 -74.364 50.991
259753 96.716 -74.375 50.982
259754 96.728 -74.380 50.974
259755 96.742 -74.385 50.969
259756 96.763 -74.390 50.965
259757 96.784 -74.396 50.963
259758 96.804 -74.402 50.963
259759 96.825 -74.407 50.964
259760 96.846 -74.412 50.969
259761 96.890 -74.440 50.977
259762 96.914 -74.457 50.987
259763 96.932 -74.469 50.998
259764 96.949 -74.481 51.009
259765 96.966 -74.493 51.017
259766 96.983 -74.505 51.023
259767 97.001 -74.518 51.027
259768 97.018 -74.530 51.029
259769 97.037 -74.539 51.028
259770 97.073 -74.532 51.026
259771 97.108 -74.526 51.021
259772 97.144 -74.519 51.012
259773 97.179 -74.513 50.999
259774 97.214 -74.506 50.982
259775 97.249 -74.500 50.961
259776 97.285 -74.493 50.936
259777 97.320 -74.487 50.909
259778 97.363 -74.476 50.880
259779 97.409 -74.463 50.847
259780 97.457 -74.449 50.815
259781 97.504 -74.436 50.782
259782 97.550 -74.423 50.747
259783 97.597 -74.410 50.713
259784 97.628 -74.420 50.673
259785 97.655 -74.436 50.630
259786 97.682 -74.452 50.585
259787 97.709 -74.468 50.538
259788 97.736 -74.483 50.489
259789 97.763 -74.499 50.440
259790 97.790 -74.515 50.389
259791 97.821 -74.534 50.337
259792 97.858 -74.558 50.286
259793 97.888 -74.573 50.237
259794 97.911 -74.578 50.190
259795 97.933 -74.584 50.145
259796 97.956 -74.590 50.100
259797 97.979 -74.596 50.058
259798 98.001 -74.602 50.018
259799 98.023 -74.608 49.980
259800 98.052 -74.618 49.938
259801 98.085 -74.633 49.893
259802 98.118 -74.646 49.843
259803 98.150 -74.661 49.789
259804 98.183 -74.675 49.731
259805 98.214 -74.697 49.672
259806 98.244 -74.722 49.612
259807 98.274 -74.747 49.552
259808 98.304 -74.771 49.495
259809 98.335 -74.795 49.437
259810 98.365 -74.819 49.380
259811 98.380 -74.842 49.327
259812 98.423 -74.817 49.273
259813 98.452 -74.787 49.220
259814 98.482 -74.758 49.166
259815 98.511 -74.728 49.111
259816 98.535 -74.703 49.058
259817 98.563 -74.712 49.007
259818 98.590 -74.723 48.956
259819 98.618 -74.734 48.907
259820 98.643 -74.758 48.855
259821 98.668 -74.785 48.801
259822 98.693 -74.812 48.747
259823 98.718 -74.840 48.692
259824 98.743 -74.867 48.637
259825 98.768 -74.894 48.582
259826 98.793 -74.922 48.528
259827 98.818 -74.948 48.475
259828 98.856 -74.925 48.424
259829 98.900 -74.880 48.372
259830 98.930 -74.865 48.322
259831 98.947 -74.879 48.271
259832 98.964 -74.894 48.223
259833 98.985 -74.910 48.174
259834 99.010 -74.928 48.126
259835 99.034 -74.947 48.078
259836 99.059 -74.965 48.030
259837 99.084 -74.983 47.983
259838 99.109 -75.002 47.937
259839 99.133 -75.020 47.891
259840 99.165 -75.039 47.845
259841 99.201 -75.059 47.800
259842 99.237 -75.078 47.756
259843 99.273 -75.098 47.710
259844 99.308 -75.117 47.665
259845 99.345 -75.137 47.620
259846 99.380 -75.155 47.577
259847 99.410 -75.159 47.535
259848 99.437 -75.166 47.496
259849 99.467 -75.172 47.457
259850 99.503 -75.175 47.419
259851 99.539 -75.178 47.382
259852 99.574 -75.182 47.349
259853 99.610 -75.185 47.319
259854 99.646 -75.189 47.289
259855 99.682 -75.192 47.261
259856 99.717 -75.195 47.234
259857 99.754 -75.198 47.208
259858 99.772 -75.210 47.184
259859 99.789 -75.223 47.161
259860 99.806 -75.236 47.139
259861 99.831 -75.244 47.118
259862 99.862 -75.248 47.097
259863 99.893 -75.252 47.075
259864 99.928 -75.262 47.052
259865 99.985 -75.272 47.031
259866 100.041 -75.278 47.009
259867 100.093 -75.295 46.989
259868 100.109 -75.326 46.968
259869 100.107 -75.360 46.947
259870 100.129 -75.358 46.927
259871 100.168 -75.364 46.902
259872 100.176 -75.364 46.876
259873 100.185 -75.364 46.850
259874 100.194 -75.364 46.821
259875 100.213 -75.402 46.792
259876 100.230 -75.417 46.761
259877 100.247 -75.430 46.729
259878 100.263 -75.442 46.697
259879 100.279 -75.454 46.667
259880 100.295 -75.466 46.637
259881 100.319 -75.486 46.607
259882 100.347 -75.513 46.576
259883 100.376 -75.540 46.546
259884 100.405 -75.567 46.516
259885 100.433 -75.594 46.487
259886 100.462 -75.621 46.459
259887 100.491 -75.648 46.430
259888 100.520 -75.674 46.402
259889 100.549 -75.702 46.373
259890 100.577 -75.728 46.344
259891 100.605 -75.731 46.316
259892 100.631 -75.719 46.289
259893 100.657 -75.708 46.263
259894 100.703 -75.703 46.236
259895 100.752 -75.699 46.210
259896 100.798 -75.704 46.181
259897 100.830 -75.729 46.151
259898 100.861 -75.755 46.123
259899 100.892 -75.781 46.098
259900 100.923 -75.806 46.072
259901 100.955 -75.832 46.045
259902 100.985 -75.856 46.017
259903 101.002 -75.863 45.987
259904 101.017 -75.870 45.956
259905 101.032 -75.877 45.923
259906 101.046 -75.883 45.891
259907 101.061 -75.889 45.860
259908 101.076 -75.895 45.830
259909 101.091 -75.901 45.799
259910 101.106 -75.907 45.769
259911 101.120 -75.913 45.739
259912 101.160 -75.925 45.712
259913 101.202 -75.936 45.684
259914 101.245 -75.946 45.656
259915 101.288 -75.956 45.629
259916 101.332 -75.966 45.601
259917 101.375 -75.976 45.572
259918 101.418 -75.986 45.542
259919 101.462 -75.997 45.514
259920 101.505 -76.007 45.487
259921 101.549 -76.016 45.461
259922 101.581 -76.026 45.435
259923 101.614 -76.043 45.410
259924 101.671 -76.082 45.385
259925 101.728 -76.119 45.361
259926 101.748 -76.132 45.337
259927 101.755 -76.135 45.313
259928 101.763 -76.137 45.289
259929 101.770 -76.140 45.265
259930 101.778 -76.142 45.241
259931 101.786 -76.145 45.218
259932 101.793 -76.148 45.195
259933 101.801 -76.151 45.171
259934 101.809 -76.154 45.148
259935 101.816 -76.156 45.125
259936 101.824 -76.159 45.101
259937 101.832 -76.161 45.079
260290 92.571 -73.759 54.643
260291 92.596 -73.769 54.642
260292 92.620 -73.780 54.640
260293 92.645 -73.791 54.639
260294 92.669 -73.801 54.638
260295 92.693 -73.813 54.637
260296 92.718 -73.823 54.636
260297 92.742 -73.834 54.635
260298 92.768 -73.826 54.634
260299 92.797 -73.834 54.633
260300 92.827 -73.852 54.635
260301 92.857 -73.871 54.637
260302 92.886 -73.890 54.638
260303 92.916 -73.908 54.641
260304 92.945 -73.922 54.644
260305 92.974 -73.929 54.648
260306 93.003 -73.935 54.651
260307 93.031 -73.941 54.654
260308 93.056 -73.941 54.657
260309 93.079 -73.938 54.660
260310 93.102 -73.935 54.663
260311 93.133 -73.919 54.665
260312 93.167 -73.897 54.666
260313 93.198 -73.882 54.666
260314 93.219 -73.887 54.663
260315 93.239 -73.891 54.661
260316 93.260 -73.896 54.662
260317 93.285 -73.900 54.664
260318 93.318 -73.901 54.666
260319 93.351 -73.903 54.667
260320 93.384 -73.904 54.668
260321 93.417 -73.907 54.667
260322 93.450 -73.908 54.664
260323 93.475 -73.899 54.661
260324 93.498 -73.886 54.658
260325 93.520 -73.873 54.655
260326 93.542 -73.860 54.652
260327 93.565 -73.847 54.649
260328 93.614 -73.855 54.646
260329 93.665 -73.872 54.646
260330 93.723 -73.894 54.646
260331 93.733 -73.894 54.649
260332 93.743 -73.894 54.653
260333 93.752 -73.894 54.659
260334 93.761 -73.894 54.673
260335 93.770 -73.894 54.689
260336 93.780 -73.894 54.707
260337 93.798 -73.895 54.725
260338 93.820 -73.896 54.739
260339 93.840 -73.897 54.748
260340 93.861 -73.897 54.753
260341 93.883 -73.897 54.756
260342 93.904 -73.899 54.745
260343 93.924 -73.900 54.726
260344 93.946 -73.901 54.701
260345 93.968 -73.903 54.668
260346 93.993 -73.904 54.623
260347 94.024 -73.905 54.575
260348 94.055 -73.905 54.528
260349 94.087 -73.903 54.489
260350 94.119 -73.901 54.460
260351 94.150 -73.899 54.438
260352 94.175 -73.898 54.421
260353 94.213 -73.917 54.406
260354 94.246 -73.915 54.396
260355 94.279 -73.913 54.382
260356 94.312 -73.911 54.364
260357 94.344 -73.910 54.341
260358 94.378 -73.907 54.314
260359 94.410 -73.906 54.286
260360 94.442 -73.905 54.257
260361 94.473 -73.904 54.228
260362 94.488 -73.905 54.198
260363 94.482 -73.904 54.167
260364 94.476 -73.904 54.136
260365 94.470 -73.903 54.105
260366 94.466 -73.902 54.072
260367 94.493 -73.901 54.039
260368 94.521 -73.899 54.005
260369 94.544 -73.912 53.971
260370 94.566 -73.926 53.935
260371 94.588 -73.940 53.896
260372 94.607 -73.944 53.854
260373 94.625 -73.941 53.810
260374 94.642 -73.939 53.767
260375 94.660 -73.935 53.725
260376 94.680 -73.926 53.685
260377 94.699 -73.916 53.647
260378 94.719 -73.907 53.611
260379 94.735 -73.907 53.577
260380 94.750 -73.908 53.548
260381 94.765 -73.910 53.525
260382 94.780 -73.911 53.503
260383 94.805 -73.917 53.482
260384 94.852 -73.935 53.460
260385 94.898 -73.951 53.431
260386 94.945 -73.969 53.399
260387 94.992 -73.985 53.364
260388 95.008 -74.010 53.328
260389 95.020 -74.033 53.291
260390 95.041 -74.028 53.253
260391 95.062 -74.023 53.213
260392 95.084 -74.017 53.170
260393 95.105 -74.012 53.121
260394 95.127 -74.007 53.068
260395 95.148 -74.001 53.014
260396 95.170 -73.996 52.960
260397 95.182 -74.001 52.909
260398 95.174 -74.016 52.859
260399 95.181 -74.004 52.810
260400 95.192 -73.995 52.762
260401 95.224 -73.999 52.715
260402 95.257 -74.003 52.668
260403 95.289 -74.007 52.623
260404 95.322 -74.011 52.579
260405 95.356 -74.015 52.536
260406 95.396 -74.020 52.494
260407 95.435 -74.026 52.452
260408 95.475 -74.030 52.412
260409 95.518 -74.029 52.372
260410 95.560 -74.029 52.333
260411 95.603 -74.028 52.295
260412 95.645 -74.027 52.256
260413 95.687 -74.026 52.220
260414 95.700 -74.021 52.183
260415 95.714 -74.015 52.147
260416 95.728 -74.010 52.110
260417 95.742 -74.004 52.073
260418 95.756 -73.998 52.037
260419 95.770 -73.993 52.002
260420 95.784 -73.987 51.968
260421 95.798 -73.982 51.935
260422 95.812 -73.976 51.901
260423 95.836 -73.971 51.868
260424 95.862 -73.966 51.836
260425 95.887 -73.960 51.800
260426 95.918 -73.955 51.763
260427 95.933 -73.957 51.726
260428 95.966 -73.977 51.686
260429 95.995 -73.995 51.645
260430 96.019 -74.008 51.606
260431 96.043 -74.022 51.566
260432 96.067 -74.035 51.525
260433 96.092 -74.048 51.486
260434 96.118 -74.061 51.449
260435 96.148 -74.074 51.413
260436 96.178 -74.086 51.377
260437 96.208 -74.098 51.344
260438 96.241 -74.101 51.311
260439 96.274 -74.107 51.279
260440 96.308 -74.114 51.250
260441 96.337 -74.124 51.221
260442 96.352 -74.154 51.194
260443 96.383 -74.172 51.169
260444 96.414 -74.189 51.144
260445 96.445 -74.207 51.122
260446 96.476 -74.224 51.099
260447 96.508 -74.242 51.078
260448 96.539 -74.257 51.060
260449 96.571 -74.270 51.042
260450 96.602 -74.284 51.027
260451 96.634 -74.298 51.013
260452 96.666 -74.312 51.001
260453 96.697 -74.326 50.989
260454 96.722 -74.336 50.980
260455 96.734 -74.342 50.973
260456 96.746 -74.346 50.966
260457 96.758 -74.351 50.963
260458 96.773 -74.356 50.961
260459 96.794 -74.361 50.961
260460 96.814 -74.367 50.962
260461 96.835 -74.372 50.967
260462 96.886 -74.406 50.974
260463 96.908 -74.422 50.983
260464 96.926 -74.434 50.994
260465 96.943 -74.446 51.004
260466 96.960 -74.458 51.013
260467 96.977 -74.471 51.021
260468 96.995 -74.482 51.028
260469 97.030 -74.476 51.032
260470 97.065 -74.469 51.032
260471 97.100 -74.463 51.030
260472 97.136 -74.456 51.027
260473 97.171 -74.450 51.016
260474 97.206 -74.443 51.002
260475 97.241 -74.437 50.983
260476 97.277 -74.430 50.961
260477 97.312 -74.424 50.936
260478 97.348 -74.418 50.908
260479 97.388 -74.408 50.878
260480 97.435 -74.396 50.845
260481 97.481 -74.382 50.812
260482 97.528 -74.369 50.779
260483 97.575 -74.355 50.745
260484 97.605 -74.368 50.711
260485 97.632 -74.383 50.672
260486 97.659 -74.399 50.629
260487 97.686 -74.415 50.584
260488 97.713 -74.431 50.535
260489 97.740 -74.447 50.485
260490 97.767 -74.463 50.435
260491 97.794 -74.479 50.383
260492 97.822 -74.496 50.333
260493 97.860 -74.519 50.282
260494 97.890 -74.534 50.234
260495 97.912 -74.540 50.187
260496 97.935 -74.546 50.140
260497 97.957 -74.552 50.095
260498 97.980 -74.557 50.053
260499 98.002 -74.563 50.014
260500 98.029 -74.573 49.973
260501 98.063 -74.587 49.931
260502 98.095 -74.601 49.884
260503 98.128 -74.615 49.834
260504 98.161 -74.629 49.779
260505 98.194 -74.643 49.720
260506 98.227 -74.657 49.659
260507 98.260 -74.671 49.599
260508 98.290 -74.696 49.540
260509 98.319 -74.720 49.484
260510 98.350 -74.745 49.426
260511 98.304 -74.699 49.369
260512 98.305 -74.788 49.317
260513 98.325 -74.839 49.262
260514 98.397 -74.784 49.208
260515 98.468 -74.728 49.154
260516 98.518 -74.687 49.099
260517 98.545 -74.674 49.046
260518 98.571 -74.694 48.995
260519 98.596 -74.722 48.945
260520 98.621 -74.749 48.897
260521 98.646 -74.776 48.845
260522 98.671 -74.803 48.792
260523 98.696 -74.831 48.737
260524 98.721 -74.857 48.681
260525 98.746 -74.885 48.627
260526 98.771 -74.912 48.573
260527 98.796 -74.939 48.520
260528 98.825 -74.953 48.467
260529 98.869 -74.907 48.416
260530 98.913 -74.862 48.364
260531 98.946 -74.839 48.313
260532 98.963 -74.853 48.262
260533 98.980 -74.868 48.213
260534 98.997 -74.882 48.165
260535 99.018 -74.900 48.116
260536 99.043 -74.918 48.068
260537 99.068 -74.936 48.021
260538 99.092 -74.954 47.974
260539 99.117 -74.972 47.927
260540 99.142 -74.991 47.882
260541 99.175 -75.010 47.837
260542 99.211 -75.030 47.792
260543 99.247 -75.049 47.748
260544 99.283 -75.069 47.702
260545 99.319 -75.088 47.657
260546 99.355 -75.108 47.613
260547 99.391 -75.128 47.570
260548 99.423 -75.136 47.529
260549 99.450 -75.142 47.489
260550 99.487 -75.145 47.450
260551 99.522 -75.149 47.411
260552 99.558 -75.152 47.375
260553 99.594 -75.156 47.342
260554 99.621 -75.163 47.312
260555 99.639 -75.176 47.282
260556 99.663 -75.184 47.254
260557 99.694 -75.189 47.226
260558 99.725 -75.192 47.200
260559 99.756 -75.197 47.175
260560 99.787 -75.201 47.153
260561 99.819 -75.205 47.130
260562 99.849 -75.210 47.108
260563 99.880 -75.214 47.087
260564 99.912 -75.218 47.064
260565 99.949 -75.231 47.042
260566 100.000 -75.259 47.020
260567 100.056 -75.270 46.998
260568 100.117 -75.267 46.977
260569 100.154 -75.270 46.956
260570 100.157 -75.298 46.934
260571 100.158 -75.330 46.912
260572 100.186 -75.331 46.886
260573 100.194 -75.332 46.861
260574 100.211 -75.361 46.835
260575 100.230 -75.395 46.807
260576 100.247 -75.408 46.778
260577 100.263 -75.420 46.748
260578 100.280 -75.432 46.716
260579 100.296 -75.444 46.685
260580 100.313 -75.456 46.656
260581 100.329 -75.469 46.627
260582 100.345 -75.481 46.597
260583 100.363 -75.495 46.567
260584 100.392 -75.521 46.538
260585 100.420 -75.549 46.507
260586 100.449 -75.575 46.478
260587 100.478 -75.602 46.451
260588 100.506 -75.629 46.423
260589 100.536 -75.656 46.394
260590 100.564 -75.683 46.366
260591 100.593 -75.710 46.337
260592 100.619 -75.702 46.308
260593 100.646 -75.690 46.281
260594 100.672 -75.678 46.255
260595 100.713 -75.672 46.229
260596 100.763 -75.670 46.202
260597 100.812 -75.671 46.174
260598 100.846 -75.693 46.145
260599 100.878 -75.718 46.117
260600 100.909 -75.744 46.091
260601 100.940 -75.770 46.065
260602 100.971 -75.796 46.039
260603 101.002 -75.812 46.011
260604 101.023 -75.822 45.979
260605 101.037 -75.828 45.948
260606 101.052 -75.834 45.916
260607 101.066 -75.841 45.884
260608 101.080 -75.847 45.853
260609 101.095 -75.854 45.823
260610 101.110 -75.860 45.793
260611 101.124 -75.867 45.764
260612 101.138 -75.873 45.734
260613 101.165 -75.881 45.706
260614 101.209 -75.891 45.678
260615 101.253 -75.901 45.649
260616 101.296 -75.911 45.623
260617 101.339 -75.922 45.593
260618 101.383 -75.931 45.563
260619 101.426 -75.941 45.533
260620 101.470 -75.951 45.504
260621 101.513 -75.962 45.477
260622 101.548 -75.972 45.451
260623 101.583 -75.989 45.425
260624 101.639 -76.027 45.399
260625 101.696 -76.066 45.375
260626 101.744 -76.097 45.351
260627 101.751 -76.100 45.327
260628 101.759 -76.103 45.303
260629 101.767 -76.106 45.278
260630 101.774 -76.108 45.254
260631 101.782 -76.111 45.230
260632 101.790 -76.114 45.206
260633 101.797 -76.116 45.183
260634 101.805 -76.119 45.160
260635 101.813 -76.122 45.136
260636 101.820 -76.125 45.113
260637 101.828 -76.127 45.089
260638 101.835 -76.130 45.066
260990 92.549 -73.716 54.646
260991 92.573 -73.726 54.645
260992 92.598 -73.737 54.643
260993 92.622 -73.747 54.641
260994 92.646 -73.758 54.639
260995 92.671 -73.769 54.638
260996 92.695 -73.780 54.636
260997 92.720 -73.791 54.635
260998 92.744 -73.801 54.634
260999 92.769 -73.799 54.633
261000 92.799 -73.807 54.633
261001 92.829 -73.826 54.634
261002 92.858 -73.845 54.635
261003 92.887 -73.863 54.637
261004 92.917 -73.875 54.639
261005 92.945 -73.881 54.642
261006 92.974 -73.888 54.646
261007 93.003 -73.894 54.650
261008 93.031 -73.900 54.654
261009 93.058 -73.902 54.657
261010 93.081 -73.900 54.660
261011 93.104 -73.897 54.663
261012 93.127 -73.894 54.665
261013 93.151 -73.889 54.667
261014 93.186 -73.866 54.667
261015 93.214 -73.857 54.664
261016 93.234 -73.862 54.661
261017 93.255 -73.866 54.659
261018 93.286 -73.869 54.659
261019 93.319 -73.870 54.660
261020 93.352 -73.872 54.662
261021 93.384 -73.874 54.664
261022 93.418 -73.875 54.664
261023 93.450 -73.877 54.662
261024 93.479 -73.872 54.659
261025 93.501 -73.860 54.655
261026 93.523 -73.847 54.652
261027 93.546 -73.834 54.647
261028 93.602 -73.844 54.641
261029 93.666 -73.865 54.636
261030 93.694 -73.868 54.633
261031 93.718 -73.870 54.630
261032 93.727 -73.870 54.629
261033 93.736 -73.870 54.629
261034 93.746 -73.871 54.630
261035 93.755 -73.871 54.637
261036 93.764 -73.871 54.647
261037 93.777 -73.871 54.660
261038 93.798 -73.872 54.673
261039 93.820 -73.872 54.685
261040 93.841 -73.872 54.694
261041 93.862 -73.873 54.701
261042 93.883 -73.874 54.704
261043 93.904 -73.875 54.697
261044 93.925 -73.875 54.683
261045 93.946 -73.876 54.665
261046 93.967 -73.878 54.635
261047 93.997 -73.879 54.595
261048 94.029 -73.877 54.549
261049 94.061 -73.875 54.501
261050 94.092 -73.873 54.463
261051 94.124 -73.872 54.434
261052 94.154 -73.869 54.412
261053 94.179 -73.869 54.394
261054 94.204 -73.869 54.382
261055 94.249 -73.896 54.371
261056 94.282 -73.894 54.358
261057 94.315 -73.892 54.341
261058 94.348 -73.890 54.317
261059 94.381 -73.888 54.291
261060 94.414 -73.886 54.263
261061 94.447 -73.885 54.235
261062 94.480 -73.882 54.207
261063 94.498 -73.880 54.176
261064 94.500 -73.879 54.145
261065 94.494 -73.878 54.113
261066 94.488 -73.877 54.080
261067 94.483 -73.876 54.047
261068 94.497 -73.875 54.012
261069 94.523 -73.879 53.978
261070 94.545 -73.893 53.942
261071 94.567 -73.907 53.905
261072 94.589 -73.922 53.866
261073 94.609 -73.929 53.826
261074 94.626 -73.926 53.784
261075 94.644 -73.924 53.742
261076 94.661 -73.922 53.701
261077 94.680 -73.916 53.662
261078 94.699 -73.907 53.628
261079 94.719 -73.897 53.595
261080 94.737 -73.891 53.565
261081 94.752 -73.894 53.539
261082 94.767 -73.895 53.518
261083 94.782 -73.897 53.496
261084 94.808 -73.904 53.471
261085 94.854 -73.920 53.445
261086 94.901 -73.938 53.415
261087 94.948 -73.954 53.383
261088 94.994 -73.972 53.349
261089 95.027 -73.992 53.314
261090 95.038 -74.018 53.278
261091 95.059 -74.013 53.239
261092 95.080 -74.007 53.198
261093 95.102 -74.002 53.153
261094 95.123 -73.997 53.102
261095 95.145 -73.991 53.049
261096 95.166 -73.986 52.995
261097 95.160 -74.006 52.941
261098 95.167 -73.995 52.890
261099 95.175 -73.983 52.841
261100 95.182 -73.972 52.793
261101 95.199 -73.966 52.746
261102 95.232 -73.970 52.701
261103 95.264 -73.974 52.657
261104 95.297 -73.979 52.613
261105 95.329 -73.982 52.570
261106 95.361 -73.986 52.528
261107 95.394 -73.990 52.488
261108 95.426 -73.994 52.448
261109 95.465 -73.999 52.408
261110 95.505 -74.004 52.368
261111 95.546 -74.005 52.329
261112 95.589 -74.004 52.291
261113 95.631 -74.004 52.253
261114 95.665 -74.001 52.216
261115 95.679 -73.996 52.180
261116 95.693 -73.990 52.144
261117 95.706 -73.985 52.108
261118 95.721 -73.979 52.072
261119 95.734 -73.973 52.037
261120 95.750 -73.968 52.002
261121 95.775 -73.963 51.968
261122 95.801 -73.957 51.936
261123 95.826 -73.952 51.903
261124 95.851 -73.948 51.870
261125 95.877 -73.942 51.836
261126 95.902 -73.937 51.801
261127 95.945 -73.930 51.764
261128 95.940 -73.923 51.725
261129 95.973 -73.944 51.686
261130 96.007 -73.964 51.646
261131 96.039 -73.985 51.606
261132 96.066 -74.001 51.566
261133 96.091 -74.014 51.526
261134 96.114 -74.027 51.488
261135 96.138 -74.041 51.451
261136 96.163 -74.054 51.415
261137 96.187 -74.068 51.379
261138 96.211 -74.081 51.346
261139 96.237 -74.090 51.314
261140 96.264 -74.099 51.283
261141 96.298 -74.105 51.252
261142 96.331 -74.112 51.224
261143 96.356 -74.126 51.197
261144 96.387 -74.144 51.171
261145 96.418 -74.161 51.146
261146 96.450 -74.176 51.122
261147 96.481 -74.190 51.100
261148 96.513 -74.204 51.079
261149 96.545 -74.217 51.061
261150 96.577 -74.232 51.043
261151 96.608 -74.245 51.027
261152 96.640 -74.259 51.013
261153 96.671 -74.273 51.000
261154 96.703 -74.286 50.989
261155 96.729 -74.298 50.980
261156 96.740 -74.302 50.971
261157 96.752 -74.308 50.964
261158 96.764 -74.312 50.960
261159 96.777 -74.317 50.958
261160 96.789 -74.322 50.958
261161 96.804 -74.327 50.960
261162 96.824 -74.333 50.964
261163 96.881 -74.372 50.970
261164 96.902 -74.387 50.978
261165 96.920 -74.399 50.988
261166 96.937 -74.411 51.001
261167 96.954 -74.424 51.010
261168 96.986 -74.419 51.018
261169 97.022 -74.413 51.026
261170 97.058 -74.406 51.032
261171 97.093 -74.400 51.034
261172 97.128 -74.393 51.032
261173 97.163 -74.387 51.029
261174 97.198 -74.380 51.017
261175 97.234 -74.374 51.001
261176 97.269 -74.368 50.982
261177 97.304 -74.361 50.959
261178 97.339 -74.355 50.931
261179 97.375 -74.349 50.903
261180 97.413 -74.341 50.874
261181 97.459 -74.327 50.841
261182 97.507 -74.314 50.808
261183 97.554 -74.301 50.775
261184 97.582 -74.314 50.741
261185 97.609 -74.330 50.706
261186 97.636 -74.346 50.667
261187 97.663 -74.362 50.624
261188 97.690 -74.378 50.579
261189 97.717 -74.394 50.532
261190 97.744 -74.410 50.483
261191 97.771 -74.426 50.434
261192 97.798 -74.442 50.385
261193 97.825 -74.458 50.334
261194 97.861 -74.480 50.284
261195 97.891 -74.495 50.234
261196 97.914 -74.501 50.184
261197 97.936 -74.507 50.135
261198 97.959 -74.512 50.090
261199 97.982 -74.518 50.048
261200 98.007 -74.527 50.008
261201 98.040 -74.541 49.968
261202 98.073 -74.555 49.923
261203 98.105 -74.569 49.875
261204 98.139 -74.584 49.824
261205 98.171 -74.597 49.770
261206 98.204 -74.612 49.711
261207 98.237 -74.626 49.650
261208 98.270 -74.640 49.589
261209 98.303 -74.654 49.530
261210 98.327 -74.661 49.473
261211 98.322 -74.631 49.416
261212 98.283 -74.649 49.361
261213 98.230 -74.734 49.308
261214 98.231 -74.823 49.253
261215 98.279 -74.818 49.197
261216 98.351 -74.762 49.141
261217 98.521 -74.667 49.087
261218 98.560 -74.708 49.034
261219 98.574 -74.712 48.984
261220 98.599 -74.739 48.936
261221 98.624 -74.766 48.887
261222 98.649 -74.794 48.835
261223 98.674 -74.821 48.782
261224 98.699 -74.848 48.727
261225 98.724 -74.875 48.673
261226 98.749 -74.903 48.618
261227 98.774 -74.929 48.564
261228 98.799 -74.957 48.510
261229 98.837 -74.935 48.457
261230 98.881 -74.889 48.406
261231 98.925 -74.844 48.355
261232 98.962 -74.813 48.304
261233 98.979 -74.828 48.254
261234 98.996 -74.843 48.205
261235 99.012 -74.857 48.157
261236 99.029 -74.872 48.108
261237 99.051 -74.888 48.060
261238 99.076 -74.907 48.013
261239 99.100 -74.925 47.966
261240 99.125 -74.944 47.919
261241 99.150 -74.962 47.873
261242 99.185 -74.982 47.829
261243 99.221 -75.001 47.784
261244 99.257 -75.020 47.740
261245 99.293 -75.040 47.695
261246 99.329 -75.060 47.650
261247 99.365 -75.079 47.606
261248 99.401 -75.099 47.564
261249 99.434 -75.113 47.522
261250 99.470 -75.116 47.482
261251 99.495 -75.124 47.443
261252 99.526 -75.129 47.405
261253 99.558 -75.132 47.369
261254 99.588 -75.137 47.336
261255 99.620 -75.141 47.305
261256 99.651 -75.145 47.276
261257 99.682 -75.150 47.247
261258 99.713 -75.154 47.219
261259 99.744 -75.158 47.192
261260 99.775 -75.162 47.167
261261 99.806 -75.167 47.144
261262 99.837 -75.170 47.121
261263 99.868 -75.175 47.099
261264 99.900 -75.179 47.076
261265 99.930 -75.183 47.054
261266 99.971 -75.199 47.031
261267 100.021 -75.227 47.009
261268 100.072 -75.256 46.987
261269 100.127 -75.269 46.965
261270 100.158 -75.274 46.944
261271 100.174 -75.285 46.922
261272 100.190 -75.295 46.897
261273 100.209 -75.318 46.872
261274 100.229 -75.357 46.847
261275 100.248 -75.386 46.820
261276 100.264 -75.398 46.792
261277 100.280 -75.410 46.764
261278 100.296 -75.422 46.735
261279 100.313 -75.434 46.704
261280 100.329 -75.447 46.674
261281 100.346 -75.459 46.645
261282 100.362 -75.471 46.617
261283 100.379 -75.483 46.589
261284 100.395 -75.496 46.559
261285 100.411 -75.508 46.530
261286 100.436 -75.530 46.500
261287 100.465 -75.557 46.471
261288 100.493 -75.584 46.443
261289 100.522 -75.611 46.415
261290 100.551 -75.638 46.386
261291 100.580 -75.665 46.358
261292 100.608 -75.685 46.329
261293 100.634 -75.674 46.301
261294 100.661 -75.662 46.274
261295 100.687 -75.649 46.248
261296 100.724 -75.642 46.222
261297 100.773 -75.641 46.195
261298 100.823 -75.643 46.166
261299 100.863 -75.656 46.138
261300 100.894 -75.682 46.110
261301 100.925 -75.708 46.084
261302 100.957 -75.734 46.058
261303 100.988 -75.754 46.031
261304 101.018 -75.768 46.003
261305 101.043 -75.780 45.971
261306 101.058 -75.786 45.939
261307 101.072 -75.793 45.907
261308 101.087 -75.799 45.876
261309 101.101 -75.806 45.845
261310 101.115 -75.812 45.816
261311 101.130 -75.818 45.786
261312 101.144 -75.825 45.757
261313 101.158 -75.834 45.729
261314 101.181 -75.840 45.700
261315 101.217 -75.847 45.672
261316 101.260 -75.856 45.642
261317 101.304 -75.866 45.614
261318 101.347 -75.876 45.584
261319 101.391 -75.887 45.553
261320 101.434 -75.897 45.523
261321 101.477 -75.906 45.494
261322 101.516 -75.916 45.467
261323 101.552 -75.935 45.441
261324 101.608 -75.973 45.414
261325 101.665 -76.012 45.389
261326 101.721 -76.050 45.364
261327 101.747 -76.066 45.340
261328 101.755 -76.069 45.316
261329 101.763 -76.072 45.292
261330 101.770 -76.074 45.267
261331 101.778 -76.077 45.242
261332 101.785 -76.080 45.219
261333 101.793 -76.082 45.194
261334 101.801 -76.085 45.170
261335 101.809 -76.088 45.147
261336 101.816 -76.091 45.123
261337 101.824 -76.093 45.100
261338 101.831 -76.096 45.076
261339 101.839 -76.099 45.053
261690 92.526 -73.672 54.650
261691 92.551 -73.683 54.649
261692 92.575 -73.694 54.647
261693 92.599 -73.704 54.645
261694 92.624 -73.716 54.642
261695 92.648 -73.726 54.640
261696 92.672 -73.737 54.638
261697 92.697 -73.747 54.636
261698 92.721 -73.758 54.634
261699 92.746 -73.769 54.633
261700 92.770 -73.772 54.632
261701 92.800 -73.781 54.632
261702 92.830 -73.800 54.632
261703 92.859 -73.819 54.633
261704 92.888 -73.827 54.635
261705 92.917 -73.834 54.637
261706 92.945 -73.840 54.640
261707 92.974 -73.846 54.644
261708 93.003 -73.853 54.648
261709 93.031 -73.859 54.652
261710 93.059 -73.863 54.655
261711 93.082 -73.861 54.658
261712 93.106 -73.858 54.662
261713 93.129 -73.855 54.665
261714 93.152 -73.852 54.667
261715 93.175 -73.850 54.666
261716 93.204 -73.836 54.663
261717 93.229 -73.832 54.661
261718 93.254 -73.836 54.659
261719 93.286 -73.838 54.656
261720 93.320 -73.839 54.657
261721 93.352 -73.841 54.658
261722 93.386 -73.843 54.659
261723 93.418 -73.844 54.660
261724 93.451 -73.846 54.659
261725 93.483 -73.845 54.657
261726 93.505 -73.832 54.653
261727 93.527 -73.819 54.649
261728 93.586 -73.833 54.642
261729 93.627 -73.840 54.635
261730 93.658 -73.841 54.629
261731 93.687 -73.844 54.621
261732 93.711 -73.847 54.615
261733 93.721 -73.847 54.610
261734 93.730 -73.847 54.606
261735 93.739 -73.847 54.603
261736 93.749 -73.847 54.604
261737 93.758 -73.847 54.608
261738 93.778 -73.847 54.614
261739 93.798 -73.848 54.623
261740 93.820 -73.849 54.632
261741 93.841 -73.849 54.640
261742 93.862 -73.850 54.647
261743 93.883 -73.850 54.651
261744 93.904 -73.851 54.648
261745 93.925 -73.852 54.640
261746 93.946 -73.853 54.625
261747 93.968 -73.850 54.600
261748 94.002 -73.849 54.564
261749 94.034 -73.847 54.521
261750 94.066 -73.845 54.475
261751 94.098 -73.844 54.437
261752 94.130 -73.841 54.407
261753 94.158 -73.840 54.384
261754 94.183 -73.839 54.367
261755 94.207 -73.838 54.355
261756 94.244 -73.857 54.344
261757 94.285 -73.875 54.331
261758 94.319 -73.872 54.314
261759 94.351 -73.871 54.291
261760 94.385 -73.869 54.263
261761 94.417 -73.867 54.234
261762 94.450 -73.865 54.204
261763 94.478 -73.863 54.174
261764 94.483 -73.860 54.144
261765 94.488 -73.858 54.113
261766 94.494 -73.855 54.084
261767 94.499 -73.853 54.053
261768 94.501 -73.850 54.020
261769 94.500 -73.850 53.986
261770 94.524 -73.860 53.950
261771 94.546 -73.875 53.914
261772 94.568 -73.889 53.876
261773 94.590 -73.904 53.837
261774 94.611 -73.913 53.798
261775 94.628 -73.910 53.756
261776 94.646 -73.908 53.717
261777 94.663 -73.906 53.679
261778 94.681 -73.904 53.643
261779 94.699 -73.897 53.610
261780 94.719 -73.888 53.581
261781 94.739 -73.879 53.554
261782 94.754 -73.879 53.531
261783 94.769 -73.881 53.512
261784 94.784 -73.882 53.488
261785 94.811 -73.889 53.460
261786 94.857 -73.907 53.430
261787 94.904 -73.923 53.400
261788 94.951 -73.940 53.368
261789 94.997 -73.957 53.334
261790 95.044 -73.974 53.300
261791 95.057 -74.001 53.265
261792 95.077 -73.997 53.226
261793 95.099 -73.991 53.182
261794 95.120 -73.986 53.134
261795 95.141 -73.981 53.082
261796 95.146 -73.996 53.029
261797 95.154 -73.985 52.974
261798 95.161 -73.973 52.921
261799 95.168 -73.962 52.872
261800 95.176 -73.951 52.824
261801 95.183 -73.939 52.777
261802 95.206 -73.938 52.731
261803 95.239 -73.941 52.688
261804 95.271 -73.945 52.646
261805 95.304 -73.950 52.604
261806 95.336 -73.954 52.563
261807 95.369 -73.957 52.522
261808 95.401 -73.961 52.482
261809 95.433 -73.965 52.443
261810 95.466 -73.969 52.404
261811 95.498 -73.973 52.365
261812 95.534 -73.977 52.326
261813 95.574 -73.981 52.287
261814 95.617 -73.980 52.250
261815 95.643 -73.976 52.213
261816 95.657 -73.971 52.177
261817 95.671 -73.965 52.141
261818 95.689 -73.960 52.106
261819 95.714 -73.954 52.071
261820 95.740 -73.949 52.036
261821 95.765 -73.944 52.002
261822 95.790 -73.939 51.969
261823 95.816 -73.934 51.936
261824 95.841 -73.929 51.904
261825 95.866 -73.923 51.872
261826 95.891 -73.919 51.838
261827 95.917 -73.913 51.803
261828 95.968 -73.906 51.765
261829 95.950 -73.891 51.726
261830 95.980 -73.910 51.687
261831 96.014 -73.930 51.647
261832 96.046 -73.951 51.607
261833 96.080 -73.971 51.567
261834 96.112 -73.992 51.528
261835 96.137 -74.006 51.490
261836 96.162 -74.020 51.453
261837 96.186 -74.033 51.417
261838 96.210 -74.047 51.382
261839 96.229 -74.058 51.349
261840 96.257 -74.067 51.317
261841 96.285 -74.076 51.285
261842 96.312 -74.085 51.255
261843 96.340 -74.095 51.227
261844 96.362 -74.104 51.199
261845 96.393 -74.110 51.173
261846 96.424 -74.123 51.148
261847 96.456 -74.137 51.124
261848 96.487 -74.151 51.102
261849 96.519 -74.165 51.081
261850 96.551 -74.179 51.062
261851 96.582 -74.192 51.044
261852 96.614 -74.206 51.028
261853 96.646 -74.220 51.014
261854 96.677 -74.234 51.000
261855 96.709 -74.248 50.989
261856 96.735 -74.259 50.979
261857 96.747 -74.264 50.970
261858 96.758 -74.269 50.962
261859 96.770 -74.273 50.958
261860 96.783 -74.278 50.956
261861 96.795 -74.283 50.957
261862 96.807 -74.288 50.958
261863 96.822 -74.298 50.961
261864 96.876 -74.338 50.967
261865 96.896 -74.352 50.975
261866 96.913 -74.364 50.984
261867 96.944 -74.363 50.997
261868 96.979 -74.356 51.009
261869 97.014 -74.350 51.018
261870 97.049 -74.343 51.027
261871 97.085 -74.337 51.032
261872 97.120 -74.330 51.033
261873 97.155 -74.324 51.031
261874 97.190 -74.317 51.027
261875 97.226 -74.311 51.013
261876 97.261 -74.305 50.997
261877 97.297 -74.299 50.977
261878 97.332 -74.292 50.954
261879 97.367 -74.286 50.926
261880 97.402 -74.280 50.898
261881 97.438 -74.273 50.869
261882 97.485 -74.260 50.838
261883 97.531 -74.246 50.804
261884 97.559 -74.261 50.771
261885 97.586 -74.277 50.736
261886 97.613 -74.293 50.700
261887 97.640 -74.309 50.662
261888 97.667 -74.325 50.620
261889 97.694 -74.341 50.575
261890 97.721 -74.357 50.530
261891 97.748 -74.373 50.483
261892 97.775 -74.389 50.436
261893 97.802 -74.405 50.389
261894 97.828 -74.421 50.341
261895 97.863 -74.442 50.289
261896 97.892 -74.456 50.237
261897 97.915 -74.461 50.186
261898 97.938 -74.468 50.135
261899 97.960 -74.474 50.088
261900 97.985 -74.481 50.046
261901 98.018 -74.496 50.005
261902 98.051 -74.509 49.964
261903 98.083 -74.524 49.917
261904 98.116 -74.537 49.868
261905 98.149 -74.552 49.816
261906 98.182 -74.566 49.761
261907 98.215 -74.580 49.702
261908 98.247 -74.594 49.641
261909 98.279 -74.608 49.580
261910 98.309 -74.622 49.520
261911 98.328 -74.635 49.463
261912 98.345 -74.631 49.406
261913 98.425 -74.612 49.350
261914 98.337 -74.694 49.297
261915 98.247 -74.776 49.241
261916 98.179 -74.844 49.185
261917 98.428 -74.717 49.129
261918 98.600 -74.828 49.075
261919 98.600 -74.803 49.023
261920 98.600 -74.778 48.974
261921 98.602 -74.757 48.925
261922 98.627 -74.784 48.876
261923 98.652 -74.811 48.825
261924 98.677 -74.838 48.772
261925 98.702 -74.866 48.718
261926 98.727 -74.893 48.664
261927 98.752 -74.920 48.609
261928 98.777 -74.947 48.555
261929 98.806 -74.962 48.501
261930 98.850 -74.916 48.448
261931 98.894 -74.872 48.397
261932 98.937 -74.826 48.346
261933 98.978 -74.788 48.296
261934 98.995 -74.803 48.246
261935 99.011 -74.817 48.197
261936 99.028 -74.831 48.148
261937 99.045 -74.846 48.100
261938 99.062 -74.860 48.052
261939 99.084 -74.878 48.005
261940 99.109 -74.896 47.958
261941 99.134 -74.914 47.911
261942 99.159 -74.933 47.864
261943 99.196 -74.953 47.820
261944 99.231 -74.972 47.775
261945 99.268 -74.991 47.731
261946 99.303 -75.011 47.687
261947 99.340 -75.031 47.643
261948 99.375 -75.051 47.600
261949 99.411 -75.072 47.557
261950 99.448 -75.085 47.516
261951 99.482 -75.086 47.475
261952 99.514 -75.090 47.436
261953 99.546 -75.094 47.398
261954 99.576 -75.098 47.362
261955 99.608 -75.102 47.329
261956 99.639 -75.107 47.298
261957 99.669 -75.110 47.268
261958 99.701 -75.115 47.240
261959 99.732 -75.119 47.212
261960 99.763 -75.123 47.185
261961 99.794 -75.128 47.159
261962 99.825 -75.132 47.135
261963 99.856 -75.136 47.112
261964 99.887 -75.140 47.089
261965 99.918 -75.145 47.066
261966 99.949 -75.148 47.042
261967 99.992 -75.167 47.020
261968 100.043 -75.196 46.998
261969 100.093 -75.224 46.976
261970 100.143 -75.252 46.954
261971 100.178 -75.290 46.934
261972 100.202 -75.308 46.912
261973 100.225 -75.325 46.887
261974 100.247 -75.353 46.861
261975 100.264 -75.376 46.836
261976 100.281 -75.388 46.809
261977 100.297 -75.400 46.781
261978 100.314 -75.412 46.752
261979 100.330 -75.425 46.724
261980 100.346 -75.437 46.694
261981 100.362 -75.449 46.664
261982 100.379 -75.461 46.635
261983 100.395 -75.474 46.607
261984 100.412 -75.486 46.579
261985 100.428 -75.499 46.550
261986 100.444 -75.511 46.521
261987 100.461 -75.523 46.491
261988 100.480 -75.539 46.462
261989 100.509 -75.565 46.434
261990 100.538 -75.593 46.406
261991 100.566 -75.619 46.378
261992 100.595 -75.646 46.350
261993 100.623 -75.656 46.321
261994 100.649 -75.645 46.294
261995 100.676 -75.633 46.267
261996 100.702 -75.621 46.240
261997 100.735 -75.612 46.214
261998 100.784 -75.613 46.187
261999 100.833 -75.615 46.158
262000 100.880 -75.620 46.130
262001 100.911 -75.646 46.103
262002 100.942 -75.671 46.077
262003 100.973 -75.696 46.050
262004 101.004 -75.711 46.023
262005 101.034 -75.724 45.994
262006 101.063 -75.737 45.962
262007 101.078 -75.744 45.930
262008 101.092 -75.750 45.898
262009 101.107 -75.757 45.867
262010 101.121 -75.763 45.837
262011 101.135 -75.770 45.808
262012 101.150 -75.778 45.779
262013 101.164 -75.787 45.750
262014 101.178 -75.796 45.723
262015 101.200 -75.802 45.694
262016 101.231 -75.805 45.665
262017 101.268 -75.812 45.635
262018 101.311 -75.822 45.604
262019 101.355 -75.831 45.573
262020 101.398 -75.842 45.542
262021 101.442 -75.852 45.512
262022 101.484 -75.862 45.484
262023 101.520 -75.881 45.457
262024 101.576 -75.919 45.431
262025 101.633 -75.957 45.404
262026 101.690 -75.996 45.378
262027 101.744 -76.032 45.353
262028 101.751 -76.035 45.329
262029 101.759 -76.038 45.304
262030 101.766 -76.040 45.280
262031 101.774 -76.043 45.255
262032 101.782 -76.046 45.230
262033 101.790 -76.048 45.206
262034 101.797 -76.051 45.181
262035 101.805 -76.053 45.158
262036 101.812 -76.057 45.134
262037 101.820 -76.060 45.110
262038 101.828 -76.062 45.086
262039 101.835 -76.065 45.062
262040 101.846 -76.072 45.039
262390 92.504 -73.629 54.656
262391 92.528 -73.640 54.654
262392 92.553 -73.650 54.651
262393 92.577 -73.661 54.649
262394 92.601 -73.672 54.646
262395 92.626 -73.683 54.643
262396 92.650 -73.694 54.640
262397 92.674 -73.704 54.637
262398 92.699 -73.715 54.635
262399 92.723 -73.726 54.633
262400 92.747 -73.737 54.631
262401 92.772 -73.745 54.631
262402 92.801 -73.756 54.631
262403 92.831 -73.773 54.630
262404 92.859 -73.780 54.631
262405 92.888 -73.786 54.633
262406 92.917 -73.792 54.635
262407 92.945 -73.798 54.638
262408 92.974 -73.805 54.642
262409 93.003 -73.811 54.645
262410 93.031 -73.817 54.649
262411 93.060 -73.824 54.653
262412 93.084 -73.822 54.657
262413 93.108 -73.819 54.661
262414 93.131 -73.816 54.665
262415 93.153 -73.814 54.667
262416 93.177 -73.811 54.665
262417 93.200 -73.808 54.663
262418 93.223 -73.805 54.661
262419 93.254 -73.805 54.658
262420 93.287 -73.807 54.655
262421 93.320 -73.809 54.653
262422 93.353 -73.810 54.653
262423 93.386 -73.812 54.653
262424 93.419 -73.813 54.653
262425 93.452 -73.815 54.654
262426 93.485 -73.816 54.653
262427 93.508 -73.805 54.650
262428 93.556 -73.813 54.646
262429 93.587 -73.815 54.639
262430 93.618 -73.816 54.632
262431 93.649 -73.817 54.624
262432 93.679 -73.819 54.614
262433 93.705 -73.822 54.605
262434 93.714 -73.822 54.597
262435 93.723 -73.822 54.589
262436 93.733 -73.823 54.582
262437 93.742 -73.823 54.577
262438 93.757 -73.823 54.574
262439 93.778 -73.824 54.574
262440 93.799 -73.825 54.577
262441 93.820 -73.825 54.582
262442 93.841 -73.825 54.587
262443 93.862 -73.826 54.593
262444 93.883 -73.827 54.598
262445 93.904 -73.828 54.600
262446 93.925 -73.827 54.596
262447 93.947 -73.823 54.583
262448 93.969 -73.819 54.562
262449 94.000 -73.817 54.530
262450 94.039 -73.818 54.491
262451 94.071 -73.816 54.448
262452 94.103 -73.814 54.410
262453 94.135 -73.812 54.379
262454 94.163 -73.810 54.355
262455 94.187 -73.810 54.338
262456 94.211 -73.808 54.327
262457 94.236 -73.807 54.315
262458 94.284 -73.844 54.302
262459 94.322 -73.854 54.285
262460 94.355 -73.851 54.262
262461 94.388 -73.850 54.234
262462 94.421 -73.847 54.203
262463 94.454 -73.846 54.171
262464 94.463 -73.844 54.140
262465 94.468 -73.841 54.109
262466 94.473 -73.838 54.079
262467 94.478 -73.835 54.050
262468 94.484 -73.833 54.021
262469 94.491 -73.832 53.989
262470 94.506 -73.836 53.955
262471 94.525 -73.842 53.921
262472 94.547 -73.857 53.884
262473 94.569 -73.871 53.845
262474 94.591 -73.885 53.806
262475 94.613 -73.897 53.768
262476 94.630 -73.895 53.729
262477 94.647 -73.893 53.691
262478 94.665 -73.891 53.656
262479 94.682 -73.888 53.622
262480 94.700 -73.886 53.593
262481 94.719 -73.879 53.567
262482 94.739 -73.869 53.544
262483 94.756 -73.864 53.524
262484 94.771 -73.866 53.505
262485 94.786 -73.868 53.480
262486 94.813 -73.875 53.451
262487 94.860 -73.892 53.420
262488 94.907 -73.909 53.388
262489 94.954 -73.926 53.355
262490 95.000 -73.943 53.320
262491 95.047 -73.960 53.285
262492 95.076 -73.982 53.249
262493 95.096 -73.982 53.207
262494 95.117 -73.976 53.162
262495 95.138 -73.971 53.113
262496 95.139 -73.975 53.060
262497 95.147 -73.963 53.006
262498 95.155 -73.952 52.953
262499 95.162 -73.941 52.902
262500 95.169 -73.929 52.854
262501 95.176 -73.918 52.808
262502 95.184 -73.907 52.763
262503 95.213 -73.909 52.719
262504 95.246 -73.913 52.677
262505 95.278 -73.916 52.636
262506 95.311 -73.921 52.595
262507 95.343 -73.925 52.556
262508 95.376 -73.929 52.516
262509 95.408 -73.932 52.477
262510 95.441 -73.936 52.439
262511 95.473 -73.940 52.399
262512 95.505 -73.944 52.361
262513 95.538 -73.948 52.322
262514 95.570 -73.952 52.284
262515 95.603 -73.956 52.246
262516 95.628 -73.951 52.210
262517 95.654 -73.946 52.174
262518 95.679 -73.941 52.139
262519 95.704 -73.936 52.105
262520 95.729 -73.931 52.070
262521 95.755 -73.926 52.035
262522 95.780 -73.920 52.002
262523 95.805 -73.916 51.969
262524 95.830 -73.910 51.937
262525 95.856 -73.905 51.905
262526 95.881 -73.900 51.873
262527 95.906 -73.895 51.841
262528 95.934 -73.890 51.805
262529 95.987 -73.881 51.767
262530 95.969 -73.866 51.727
262531 95.987 -73.876 51.687
262532 96.021 -73.897 51.648
262533 96.053 -73.917 51.608
262534 96.087 -73.938 51.567
262535 96.120 -73.958 51.529
262536 96.153 -73.979 51.492
262537 96.185 -73.998 51.456
262538 96.209 -74.012 51.420
262539 96.233 -74.026 51.386
262540 96.249 -74.035 51.353
262541 96.276 -74.045 51.320
262542 96.304 -74.054 51.288
262543 96.325 -74.063 51.259
262544 96.341 -74.073 51.230
262545 96.367 -74.079 51.202
262546 96.395 -74.085 51.176
262547 96.430 -74.085 51.150
262548 96.462 -74.098 51.127
262549 96.493 -74.112 51.104
262550 96.525 -74.126 51.083
262551 96.556 -74.139 51.063
262552 96.588 -74.154 51.045
262553 96.620 -74.167 51.029
262554 96.651 -74.181 51.015
262555 96.683 -74.195 51.001
262556 96.715 -74.209 50.988
262557 96.741 -74.220 50.978
262558 96.753 -74.225 50.969
262559 96.764 -74.230 50.961
262560 96.777 -74.235 50.957
262561 96.789 -74.239 50.955
262562 96.801 -74.245 50.955
262563 96.813 -74.249 50.956
262564 96.840 -74.276 50.959
262565 96.872 -74.305 50.964
262566 96.900 -74.306 50.970
262567 96.936 -74.300 50.979
262568 96.971 -74.293 50.993
262569 97.007 -74.287 51.006
262570 97.042 -74.280 51.017
262571 97.077 -74.274 51.026
262572 97.112 -74.268 51.029
262573 97.148 -74.261 51.030
262574 97.183 -74.255 51.027
262575 97.218 -74.248 51.022
262576 97.253 -74.242 51.007
262577 97.289 -74.236 50.990
262578 97.324 -74.230 50.971
262579 97.359 -74.223 50.949
262580 97.394 -74.217 50.922
262581 97.430 -74.210 50.894
262582 97.465 -74.204 50.865
262583 97.509 -74.192 50.835
262584 97.536 -74.208 50.801
262585 97.563 -74.224 50.767
262586 97.590 -74.240 50.732
262587 97.617 -74.256 50.696
262588 97.644 -74.272 50.658
262589 97.671 -74.288 50.616
262590 97.698 -74.304 50.572
262591 97.725 -74.320 50.528
262592 97.752 -74.336 50.484
262593 97.779 -74.352 50.439
262594 97.806 -74.368 50.393
262595 97.832 -74.383 50.347
262596 97.864 -74.403 50.294
262597 97.894 -74.417 50.242
262598 97.917 -74.423 50.189
262599 97.939 -74.429 50.136
262600 97.962 -74.435 50.089
262601 97.995 -74.449 50.046
262602 98.028 -74.464 50.004
262603 98.061 -74.477 49.960
262604 98.094 -74.492 49.912
262605 98.127 -74.506 49.861
262606 98.159 -74.520 49.808
262607 98.192 -74.534 49.752
262608 98.222 -74.548 49.692
262609 98.252 -74.562 49.632
262610 98.283 -74.577 49.572
262611 98.313 -74.590 49.511
262612 98.329 -74.603 49.453
262613 98.307 -74.609 49.396
262614 98.337 -74.602 49.339
262615 98.324 -74.640 49.285
262616 98.229 -74.722 49.229
262617 98.250 -74.781 49.172
262618 98.468 -74.714 49.117
262619 98.640 -74.924 49.063
262620 98.640 -74.898 49.012
262621 98.640 -74.872 48.963
262622 98.640 -74.847 48.914
262623 98.639 -74.821 48.865
262624 98.655 -74.829 48.814
262625 98.680 -74.856 48.761
262626 98.705 -74.884 48.708
262627 98.730 -74.910 48.654
262628 98.756 -74.938 48.600
262629 98.780 -74.965 48.546
262630 98.818 -74.944 48.492
262631 98.862 -74.899 48.439
262632 98.906 -74.853 48.388
262633 98.950 -74.809 48.337
262634 98.994 -74.763 48.287
262635 99.010 -74.777 48.238
262636 99.027 -74.791 48.188
262637 99.044 -74.806 48.139
262638 99.061 -74.821 48.091
262639 99.077 -74.835 48.043
262640 99.094 -74.850 47.995
262641 99.117 -74.867 47.948
262642 99.142 -74.885 47.902
262643 99.170 -74.904 47.856
262644 99.206 -74.924 47.810
262645 99.242 -74.943 47.765
262646 99.278 -74.963 47.721
262647 99.313 -74.982 47.678
262648 99.349 -75.004 47.635
262649 99.383 -75.026 47.592
262650 99.420 -75.043 47.551
262651 99.458 -75.054 47.509
262652 99.494 -75.060 47.469
262653 99.528 -75.060 47.430
262654 99.563 -75.061 47.392
262655 99.595 -75.063 47.355
262656 99.626 -75.068 47.323
262657 99.657 -75.072 47.292
262658 99.689 -75.076 47.262
262659 99.720 -75.080 47.233
262660 99.751 -75.085 47.204
262661 99.782 -75.088 47.177
262662 99.813 -75.093 47.151
262663 99.844 -75.097 47.127
262664 99.875 -75.101 47.103
262665 99.906 -75.105 47.080
262666 99.937 -75.110 47.056
262667 99.968 -75.114 47.032
262668 100.013 -75.135 47.009
262669 100.064 -75.164 46.987
262670 100.114 -75.192 46.966
262671 100.153 -75.227 46.945
262672 100.187 -75.265 46.924
262673 100.221 -75.303 46.902
262674 100.255 -75.341 46.877
262675 100.279 -75.363 46.850
262676 100.297 -75.378 46.825
262677 100.314 -75.390 46.798
262678 100.330 -75.403 46.770
262679 100.347 -75.415 46.741
262680 100.363 -75.427 46.713
262681 100.379 -75.439 46.684
262682 100.396 -75.452 46.655
262683 100.412 -75.464 46.626
262684 100.428 -75.477 46.598
262685 100.445 -75.489 46.570
262686 100.462 -75.501 46.541
262687 100.478 -75.513 46.512
262688 100.494 -75.525 46.482
262689 100.510 -75.538 46.453
262690 100.527 -75.550 46.425
262691 100.553 -75.574 46.397
262692 100.582 -75.601 46.370
262693 100.611 -75.628 46.341
262694 100.638 -75.627 46.313
262695 100.664 -75.615 46.286
262696 100.690 -75.604 46.259
262697 100.717 -75.592 46.233
262698 100.749 -75.583 46.206
262699 100.794 -75.584 46.178
262700 100.844 -75.587 46.150
262701 100.893 -75.588 46.123
262702 100.928 -75.609 46.096
262703 100.959 -75.635 46.070
262704 100.990 -75.652 46.043
262705 101.020 -75.667 46.014
262706 101.051 -75.681 45.984
262707 101.081 -75.694 45.953
262708 101.098 -75.702 45.921
262709 101.113 -75.709 45.889
262710 101.127 -75.715 45.860
262711 101.141 -75.721 45.830
262712 101.156 -75.731 45.800
262713 101.170 -75.740 45.771
262714 101.183 -75.749 45.742
262715 101.197 -75.757 45.713
262716 101.219 -75.764 45.684
262717 101.250 -75.767 45.655
262718 101.281 -75.770 45.625
262719 101.319 -75.777 45.593
262720 101.362 -75.787 45.561
262721 101.406 -75.796 45.531
262722 101.449 -75.807 45.501
262723 101.489 -75.827 45.473
262724 101.545 -75.865 45.446
262725 101.602 -75.903 45.419
262726 101.659 -75.942 45.393
262727 101.715 -75.980 45.367
262728 101.747 -76.000 45.341
262729 101.755 -76.003 45.317
262730 101.763 -76.007 45.292
262731 101.770 -76.009 45.267
262732 101.778 -76.012 45.242
262733 101.785 -76.014 45.218
262734 101.793 -76.017 45.193
262735 101.801 -76.019 45.169
262736 101.809 -76.022 45.145
262737 101.816 -76.025 45.121
262738 101.824 -76.028 45.097
262739 101.831 -76.031 45.073
262740 101.846 -76.042 45.049
262741 101.869 -76.062 45.026
263090 92.465 -73.588 54.663
263091 92.492 -73.599 54.660
263092 92.518 -73.609 54.657
263093 92.545 -73.620 54.654
263094 92.571 -73.630 54.651
263095 92.597 -73.640 54.648
263096 92.624 -73.651 54.644
263097 92.650 -73.662 54.641
263098 92.676 -73.672 54.638
263099 92.701 -73.683 54.635
263100 92.725 -73.694 54.632
263101 92.749 -73.704 54.631
263102 92.774 -73.715 54.630
263103 92.802 -73.726 54.630
263104 92.831 -73.732 54.629
263105 92.859 -73.738 54.629
263106 92.888 -73.744 54.631
263107 92.917 -73.751 54.633
263108 92.945 -73.757 54.635
263109 92.974 -73.763 54.639
263110 93.003 -73.770 54.643
263111 93.031 -73.776 54.647
263112 93.060 -73.782 54.651
263113 93.086 -73.784 54.655
263114 93.109 -73.781 54.660
263115 93.132 -73.778 54.664
263116 93.155 -73.775 54.667
263117 93.179 -73.772 54.665
263118 93.202 -73.769 54.663
263119 93.227 -73.770 54.661
263120 93.259 -73.778 54.657
263121 93.292 -73.779 54.653
263122 93.324 -73.781 54.650
263123 93.357 -73.781 54.648
263124 93.389 -73.783 54.646
263125 93.422 -73.784 54.647
263126 93.454 -73.785 54.647
263127 93.487 -73.787 54.647
263128 93.516 -73.788 54.646
263129 93.547 -73.790 54.642
263130 93.578 -73.791 54.635
263131 93.610 -73.792 54.627
263132 93.641 -73.794 54.619
263133 93.672 -73.795 54.609
263134 93.698 -73.799 54.598
263135 93.708 -73.799 54.588
263136 93.717 -73.799 54.576
263137 93.726 -73.799 54.563
263138 93.735 -73.799 54.554
263139 93.757 -73.800 54.548
263140 93.778 -73.800 54.543
263141 93.799 -73.801 54.542
263142 93.820 -73.801 54.542
263143 93.841 -73.802 54.543
263144 93.862 -73.803 54.544
263145 93.883 -73.803 54.548
263146 93.905 -73.800 54.548
263147 93.927 -73.796 54.544
263148 93.949 -73.791 54.533
263149 93.971 -73.788 54.518
263150 93.993 -73.784 54.490
263151 94.032 -73.784 54.455
263152 94.076 -73.786 54.416
263153 94.108 -73.784 54.380
263154 94.140 -73.782 54.350
263155 94.167 -73.781 54.326
263156 94.192 -73.780 54.307
263157 94.216 -73.778 54.296
263158 94.240 -73.778 54.284
263159 94.272 -73.789 54.270
263160 94.325 -73.832 54.253
263161 94.358 -73.832 54.230
263162 94.391 -73.830 54.201
263163 94.424 -73.828 54.169
263164 94.442 -73.826 54.135
263165 94.448 -73.824 54.103
263166 94.453 -73.821 54.071
263167 94.458 -73.819 54.041
263168 94.469 -73.819 54.013
263169 94.484 -73.824 53.985
263170 94.499 -73.828 53.955
263171 94.514 -73.832 53.922
263172 94.529 -73.834 53.889
263173 94.548 -73.838 53.852
263174 94.570 -73.853 53.813
263175 94.592 -73.866 53.775
263176 94.614 -73.881 53.737
263177 94.632 -73.880 53.700
263178 94.649 -73.878 53.665
263179 94.667 -73.875 53.632
263180 94.684 -73.873 53.603
263181 94.702 -73.871 53.576
263182 94.719 -73.869 53.554
263183 94.739 -73.860 53.534
263184 94.759 -73.850 53.516
263185 94.773 -73.851 53.495
263186 94.788 -73.854 53.469
263187 94.816 -73.861 53.440
263188 94.863 -73.878 53.409
263189 94.909 -73.895 53.376
263190 94.956 -73.912 53.341
263191 95.003 -73.929 53.305
263192 95.050 -73.946 53.269
263193 95.096 -73.963 53.230
263194 95.114 -73.966 53.186
263195 95.126 -73.965 53.139
263196 95.133 -73.954 53.089
263197 95.140 -73.942 53.036
263198 95.148 -73.931 52.983
263199 95.155 -73.919 52.932
263200 95.163 -73.908 52.883
263201 95.170 -73.897 52.837
263202 95.177 -73.885 52.793
263203 95.188 -73.876 52.749
263204 95.220 -73.880 52.706
263205 95.253 -73.884 52.666
263206 95.285 -73.888 52.627
263207 95.318 -73.892 52.588
263208 95.350 -73.896 52.550
263209 95.383 -73.900 52.510
263210 95.415 -73.904 52.472
263211 95.448 -73.907 52.433
263212 95.480 -73.911 52.394
263213 95.513 -73.915 52.357
263214 95.545 -73.919 52.318
263215 95.577 -73.926 52.280
263216 95.608 -73.934 52.243
263217 95.640 -73.928 52.208
263218 95.668 -73.922 52.172
263219 95.693 -73.917 52.137
263220 95.719 -73.913 52.103
263221 95.744 -73.907 52.069
263222 95.769 -73.902 52.034
263223 95.795 -73.897 52.001
263224 95.820 -73.892 51.968
263225 95.845 -73.887 51.937
263226 95.871 -73.882 51.905
263227 95.896 -73.876 51.873
263228 95.921 -73.871 51.841
263229 95.960 -73.865 51.805
263230 96.007 -73.856 51.767
263231 95.989 -73.841 51.727
263232 95.994 -73.842 51.687
263233 96.028 -73.863 51.647
263234 96.060 -73.883 51.608
263235 96.094 -73.904 51.569
263236 96.127 -73.924 51.532
263237 96.160 -73.944 51.495
263238 96.193 -73.966 51.459
263239 96.226 -73.986 51.424
263240 96.256 -74.004 51.389
263241 96.269 -74.013 51.356
263242 96.289 -74.022 51.324
263243 96.305 -74.032 51.292
263244 96.320 -74.042 51.262
263245 96.346 -74.048 51.234
263246 96.372 -74.055 51.206
263247 96.399 -74.062 51.179
263248 96.432 -74.064 51.154
263249 96.466 -74.064 51.129
263250 96.499 -74.073 51.107
263251 96.531 -74.087 51.086
263252 96.562 -74.101 51.066
263253 96.594 -74.114 51.048
263254 96.625 -74.129 51.031
263255 96.657 -74.142 51.016
263256 96.689 -74.156 51.002
263257 96.721 -74.170 50.989
263258 96.747 -74.181 50.978
263259 96.759 -74.186 50.968
263260 96.771 -74.191 50.960
263261 96.783 -74.195 50.956
263262 96.795 -74.201 50.954
263263 96.807 -74.205 50.953
263264 96.831 -74.228 50.953
263265 96.858 -74.254 50.955
263266 96.890 -74.272 50.959
263267 96.925 -74.269 50.966
263268 96.960 -74.261 50.973
263269 96.996 -74.255 50.986
263270 97.031 -74.247 50.999
263271 97.067 -74.239 51.010
263272 97.102 -74.233 51.019
263273 97.137 -74.225 51.024
263274 97.173 -74.217 51.024
263275 97.208 -74.211 51.021
263276 97.243 -74.203 51.014
263277 97.279 -74.195 50.999
263278 97.314 -74.189 50.982
263279 97.350 -74.181 50.964
263280 97.385 -74.174 50.944
263281 97.420 -74.167 50.919
263282 97.456 -74.159 50.892
263283 97.486 -74.149 50.863
263284 97.511 -74.157 50.833
263285 97.540 -74.171 50.799
263286 97.567 -74.187 50.766
263287 97.594 -74.203 50.732
263288 97.621 -74.219 50.696
263289 97.648 -74.235 50.656
263290 97.675 -74.251 50.615
263291 97.702 -74.267 50.572
263292 97.729 -74.283 50.527
263293 97.755 -74.299 50.483
263294 97.783 -74.314 50.438
263295 97.809 -74.330 50.393
263296 97.837 -74.346 50.346
263297 97.866 -74.364 50.295
263298 97.895 -74.378 50.243
263299 97.918 -74.384 50.190
263300 97.940 -74.390 50.137
263301 97.973 -74.404 50.092
263302 98.006 -74.418 50.049
263303 98.038 -74.432 50.005
263304 98.071 -74.446 49.958
263305 98.104 -74.460 49.907
263306 98.135 -74.474 49.853
263307 98.165 -74.488 49.799
263308 98.196 -74.502 49.742
263309 98.226 -74.517 49.683
263310 98.256 -74.531 49.622
263311 98.287 -74.545 49.562
263312 98.317 -74.559 49.501
263313 98.329 -74.571 49.443
263314 98.308 -74.577 49.386
263315 98.287 -74.584 49.329
263316 98.304 -74.586 49.274
263317 98.250 -74.659 49.217
263318 98.334 -74.706 49.161
263319 98.417 -74.754 49.106
263320 98.598 -74.881 49.052
263321 98.647 -74.938 49.001
263322 98.680 -74.967 48.952
263323 98.679 -74.942 48.903
263324 98.679 -74.916 48.854
263325 98.679 -74.891 48.804
263326 98.684 -74.874 48.752
263327 98.709 -74.901 48.698
263328 98.733 -74.928 48.644
263329 98.759 -74.956 48.590
263330 98.787 -74.971 48.537
263331 98.831 -74.926 48.483
263332 98.874 -74.881 48.430
263333 98.918 -74.835 48.379
263334 98.962 -74.791 48.329
263335 99.006 -74.745 48.278
263336 99.026 -74.752 48.228
263337 99.043 -74.766 48.179
263338 99.060 -74.781 48.130
263339 99.077 -74.795 48.081
263340 99.093 -74.809 48.034
263341 99.110 -74.824 47.986
263342 99.127 -74.838 47.940
263343 99.150 -74.856 47.893
263344 99.180 -74.875 47.847
263345 99.216 -74.895 47.801
263346 99.252 -74.914 47.756
263347 99.287 -74.936 47.712
263348 99.322 -74.958 47.669
263349 99.357 -74.980 47.626
263350 99.392 -75.001 47.585
263351 99.430 -75.012 47.543
263352 99.469 -75.022 47.502
263353 99.506 -75.032 47.463
263354 99.541 -75.033 47.424
263355 99.575 -75.034 47.386
263356 99.609 -75.035 47.349
263357 99.643 -75.035 47.317
263358 99.677 -75.037 47.286
263359 99.707 -75.041 47.256
263360 99.738 -75.046 47.226
263361 99.770 -75.050 47.197
263362 99.801 -75.054 47.170
263363 99.832 -75.058 47.143
263364 99.863 -75.063 47.119
263365 99.894 -75.066 47.094
263366 99.925 -75.071 47.070
263367 99.956 -75.075 47.046
263368 99.987 -75.079 47.021
263369 100.034 -75.104 46.998
263370 100.085 -75.132 46.976
263371 100.130 -75.164 46.955
263372 100.162 -75.202 46.935
263373 100.196 -75.240 46.914
263374 100.230 -75.278 46.891
263375 100.263 -75.312 46.866
263376 100.291 -75.328 46.838
263377 100.310 -75.344 46.813
263378 100.329 -75.361 46.786
263379 100.347 -75.377 46.759
263380 100.366 -75.393 46.730
263381 100.385 -75.409 46.702
263382 100.403 -75.426 46.673
263383 100.422 -75.442 46.645
263384 100.440 -75.458 46.617
263385 100.460 -75.474 46.589
263386 100.478 -75.491 46.560
263387 100.495 -75.503 46.531
263388 100.511 -75.516 46.502
263389 100.528 -75.528 46.473
263390 100.544 -75.540 46.445
263391 100.560 -75.552 46.417
263392 100.576 -75.565 46.390
263393 100.598 -75.583 46.362
263394 100.626 -75.609 46.333
263395 100.653 -75.599 46.304
263396 100.680 -75.587 46.277
263397 100.707 -75.574 46.251
263398 100.735 -75.562 46.224
263399 100.767 -75.554 46.197
263400 100.811 -75.556 46.170
263401 100.855 -75.558 46.142
263402 100.904 -75.559 46.115
263403 100.945 -75.573 46.088
263404 100.975 -75.595 46.061
263405 101.006 -75.609 46.034
263406 101.037 -75.623 46.005
263407 101.067 -75.637 45.974
263408 101.098 -75.651 45.943
263409 101.118 -75.660 45.910
263410 101.133 -75.667 45.879
263411 101.147 -75.674 45.849
263412 101.161 -75.684 45.818
263413 101.175 -75.692 45.789
263414 101.189 -75.701 45.759
263415 101.203 -75.710 45.730
263416 101.217 -75.719 45.701
263417 101.239 -75.725 45.672
263418 101.269 -75.728 45.643
263419 101.301 -75.731 45.613
263420 101.331 -75.734 45.581
263421 101.370 -75.742 45.549
263422 101.413 -75.752 45.518
263423 101.458 -75.773 45.489
263424 101.514 -75.811 45.461
263425 101.570 -75.850 45.434
263426 101.627 -75.887 45.407
263427 101.684 -75.926 45.381
263428 101.740 -75.964 45.355
263429 101.751 -75.969 45.329
263430 101.759 -75.972 45.304
263431 101.766 -75.975 45.279
263432 101.774 -75.978 45.254
263433 101.781 -75.980 45.229
263434 101.789 -75.983 45.204
263435 101.797 -75.986 45.180
263436 101.805 -75.988 45.156
263437 101.812 -75.991 45.132
263438 101.820 -75.994 45.108
263439 101.827 -75.997 45.084
263440 101.846 -76.013 45.060
263441 101.869 -76.032 45.036
263442 101.890 -76.052 45.012
263790 92.419 -73.549 54.671
263791 92.445 -73.559 54.668
263792 92.472 -73.570 54.665
263793 92.498 -73.581 54.660
263794 92.524 -73.591 54.657
263795 92.551 -73.601 54.653
263796 92.577 -73.612 54.650
263797 92.603 -73.622 54.646
263798 92.630 -73.633 54.642
263799 92.656 -73.643 54.638
263800 92.683 -73.653 54.635
263801 92.709 -73.664 54.632
263802 92.740 -73.675 54.630
263803 92.771 -73.684 54.629
263804 92.804 -73.689 54.628
263805 92.831 -73.691 54.627
263806 92.859 -73.697 54.628
263807 92.888 -73.703 54.629
263808 92.917 -73.709 54.630
263809 92.945 -73.716 54.633
263810 92.974 -73.722 54.636
263811 93.003 -73.728 54.640
263812 93.031 -73.734 54.644
263813 93.060 -73.741 54.648
263814 93.087 -73.745 54.653
263815 93.111 -73.742 54.657
263816 93.134 -73.740 54.661
263817 93.157 -73.737 54.663
263818 93.180 -73.734 54.662
263819 93.203 -73.731 54.661
263820 93.233 -73.738 54.658
263821 93.266 -73.747 54.654
263822 93.298 -73.753 54.651
263823 93.330 -73.754 54.647
263824 93.363 -73.755 54.644
263825 93.395 -73.756 54.642
263826 93.428 -73.757 54.641
263827 93.460 -73.759 54.641
263828 93.491 -73.760 54.640
263829 93.513 -73.763 54.640
263830 93.539 -73.766 54.635
263831 93.570 -73.767 54.629
263832 93.601 -73.769 54.622
263833 93.632 -73.770 54.614
263834 93.663 -73.772 54.604
263835 93.690 -73.775 54.592
263836 93.701 -73.775 54.580
263837 93.711 -73.775 54.565
263838 93.720 -73.775 54.547
263839 93.736 -73.775 54.534
263840 93.757 -73.776 54.524
263841 93.778 -73.777 54.517
263842 93.799 -73.777 54.510
263843 93.820 -73.778 54.506
263844 93.841 -73.778 54.503
263845 93.863 -73.776 54.500
263846 93.885 -73.772 54.499
263847 93.907 -73.769 54.497
263848 93.929 -73.764 54.491
263849 93.951 -73.760 54.482
263850 93.973 -73.756 54.468
263851 93.995 -73.752 54.445
263852 94.019 -73.749 54.416
263853 94.065 -73.751 54.382
263854 94.111 -73.753 54.349
263855 94.145 -73.753 54.320
263856 94.172 -73.751 54.296
263857 94.196 -73.750 54.276
263858 94.220 -73.749 54.263
263859 94.245 -73.748 54.250
263860 94.269 -73.747 54.236
263861 94.313 -73.777 54.219
263862 94.362 -73.813 54.194
263863 94.395 -73.811 54.165
263864 94.422 -73.809 54.134
263865 94.427 -73.807 54.102
263866 94.432 -73.803 54.070
263867 94.447 -73.807 54.038
263868 94.462 -73.811 54.007
263869 94.477 -73.816 53.978
263870 94.492 -73.819 53.949
263871 94.507 -73.823 53.917
263872 94.522 -73.827 53.884
263873 94.537 -73.831 53.850
263874 94.551 -73.825 53.814
263875 94.571 -73.834 53.777
263876 94.593 -73.848 53.740
263877 94.615 -73.863 53.705
263878 94.634 -73.865 53.670
263879 94.651 -73.863 53.638
263880 94.669 -73.860 53.609
263881 94.686 -73.858 53.583
263882 94.703 -73.856 53.560
263883 94.721 -73.854 53.540
263884 94.739 -73.850 53.523
263885 94.759 -73.841 53.506
263886 94.776 -73.837 53.483
263887 94.790 -73.838 53.456
263888 94.819 -73.847 53.426
263889 94.865 -73.863 53.395
263890 94.912 -73.881 53.361
263891 94.959 -73.897 53.325
263892 95.005 -73.915 53.288
263893 95.052 -73.932 53.252
263894 95.099 -73.949 53.209
263895 95.119 -73.944 53.163
263896 95.127 -73.932 53.113
263897 95.134 -73.921 53.063
263898 95.141 -73.910 53.011
263899 95.149 -73.898 52.961
263900 95.156 -73.887 52.911
263901 95.164 -73.875 52.865
263902 95.171 -73.864 52.821
263903 95.178 -73.853 52.778
263904 95.195 -73.847 52.736
263905 95.228 -73.851 52.695
263906 95.260 -73.855 52.655
263907 95.293 -73.859 52.618
263908 95.325 -73.863 52.582
263909 95.357 -73.867 52.544
263910 95.390 -73.871 52.504
263911 95.422 -73.875 52.466
263912 95.455 -73.879 52.427
263913 95.487 -73.882 52.389
263914 95.520 -73.886 52.352
263915 95.552 -73.897 52.314
263916 95.584 -73.912 52.277
263917 95.612 -73.910 52.240
263918 95.643 -73.903 52.205
263919 95.677 -73.898 52.170
263920 95.708 -73.894 52.135
263921 95.734 -73.889 52.102
263922 95.759 -73.884 52.068
263923 95.784 -73.879 52.034
263924 95.810 -73.873 52.000
263925 95.835 -73.868 51.967
263926 95.861 -73.863 51.936
263927 95.886 -73.858 51.904
263928 95.911 -73.853 51.871
263929 95.936 -73.847 51.840
263930 95.986 -73.841 51.803
263931 96.027 -73.831 51.765
263932 96.009 -73.816 51.726
263933 96.002 -73.809 51.687
263934 96.035 -73.829 51.647
263935 96.067 -73.850 51.609
263936 96.101 -73.870 51.571
263937 96.134 -73.891 51.534
263938 96.167 -73.911 51.499
263939 96.200 -73.932 51.464
263940 96.232 -73.954 51.429
263941 96.261 -73.979 51.394
263942 96.268 -73.991 51.361
263943 96.284 -74.001 51.329
263944 96.300 -74.010 51.297
263945 96.324 -74.018 51.267
263946 96.351 -74.025 51.239
263947 96.378 -74.032 51.211
263948 96.404 -74.038 51.183
263949 96.434 -74.042 51.157
263950 96.468 -74.042 51.133
263951 96.503 -74.042 51.111
263952 96.536 -74.048 51.090
263953 96.568 -74.062 51.069
263954 96.600 -74.076 51.050
263955 96.631 -74.089 51.033
263956 96.663 -74.103 51.018
263957 96.694 -74.117 51.002
263958 96.726 -74.131 50.989
263959 96.753 -74.142 50.977
263960 96.765 -74.147 50.967
263961 96.777 -74.152 50.959
263962 96.789 -74.157 50.954
263963 96.801 -74.162 50.952
263964 96.822 -74.179 50.951
263965 96.849 -74.206 50.949
263966 96.882 -74.223 50.950
263967 96.914 -74.239 50.954
263968 96.948 -74.242 50.960
263969 96.984 -74.234 50.967
263970 97.019 -74.227 50.979
263971 97.054 -74.220 50.991
263972 97.090 -74.212 51.001
263973 97.125 -74.205 51.011
263974 97.161 -74.198 51.014
263975 97.196 -74.190 51.014
263976 97.231 -74.183 51.011
263977 97.267 -74.176 51.003
263978 97.302 -74.168 50.988
263979 97.337 -74.161 50.972
263980 97.373 -74.154 50.955
263981 97.408 -74.146 50.937
263982 97.440 -74.137 50.915
263983 97.462 -74.122 50.890
263984 97.486 -74.126 50.862
263985 97.510 -74.132 50.831
263986 97.534 -74.140 50.798
263987 97.564 -74.154 50.765
263988 97.593 -74.169 50.731
263989 97.622 -74.184 50.694
263990 97.651 -74.198 50.656
263991 97.678 -74.214 50.615
263992 97.706 -74.230 50.572
263993 97.733 -74.245 50.527
263994 97.759 -74.261 50.482
263995 97.787 -74.277 50.436
263996 97.813 -74.293 50.390
263997 97.841 -74.309 50.344
263998 97.867 -74.325 50.293
263999 97.897 -74.339 50.241
264000 97.919 -74.345 50.189
264001 97.950 -74.358 50.137
264002 97.983 -74.372 50.092
264003 98.016 -74.386 50.049
264004 98.048 -74.400 50.004
264005 98.078 -74.415 49.954
264006 98.108 -74.428 49.901
264007 98.139 -74.443 49.845
264008 98.169 -74.457 49.789
264009 98.200 -74.471 49.731
264010 98.230 -74.485 49.672
264011 98.260 -74.499 49.612
264012 98.291 -74.514 49.552
264013 98.321 -74.527 49.492
264014 98.329 -74.539 49.434
264015 98.308 -74.545 49.378
264016 98.287 -74.552 49.322
264017 98.266 -74.558 49.266
264018 98.307 -74.581 49.208
264019 98.360 -74.612 49.151
264020 98.414 -74.643 49.095
264021 98.464 -74.686 49.042
264022 98.514 -74.743 48.991
264023 98.564 -74.800 48.942
264024 98.613 -74.858 48.893
264025 98.662 -74.916 48.843
264026 98.712 -74.973 48.793
264027 98.719 -74.960 48.742
264028 98.719 -74.934 48.688
264029 98.736 -74.946 48.634
264030 98.762 -74.973 48.581
264031 98.799 -74.954 48.527
264032 98.843 -74.908 48.473
264033 98.886 -74.863 48.420
264034 98.931 -74.818 48.370
264035 98.975 -74.772 48.320
264036 99.018 -74.728 48.269
264037 99.042 -74.726 48.219
264038 99.059 -74.740 48.170
264039 99.076 -74.755 48.120
264040 99.092 -74.769 48.072
264041 99.109 -74.784 48.024
264042 99.126 -74.799 47.977
264043 99.143 -74.813 47.931
264044 99.159 -74.828 47.884
264045 99.190 -74.847 47.838
264046 99.225 -74.869 47.792
264047 99.260 -74.891 47.747
264048 99.295 -74.913 47.704
264049 99.330 -74.934 47.661
264050 99.365 -74.956 47.618
264051 99.402 -74.969 47.577
264052 99.440 -74.980 47.536
264053 99.479 -74.991 47.495
264054 99.517 -75.003 47.457
264055 99.553 -75.007 47.418
264056 99.587 -75.007 47.380
264057 99.621 -75.008 47.344
264058 99.655 -75.009 47.311
264059 99.689 -75.010 47.279
264060 99.723 -75.010 47.248
264061 99.758 -75.011 47.219
264062 99.788 -75.015 47.190
264063 99.820 -75.019 47.162
264064 99.851 -75.023 47.136
264065 99.881 -75.028 47.111
264066 99.913 -75.032 47.085
264067 99.944 -75.036 47.060
264068 99.975 -75.041 47.035
264069 100.006 -75.044 47.012
264070 100.056 -75.073 46.988
264071 100.106 -75.101 46.965
264072 100.139 -75.139 46.944
264073 100.171 -75.177 46.925
264074 100.205 -75.215 46.903
264075 100.239 -75.253 46.879
264076 100.269 -75.277 46.854
264077 100.297 -75.293 46.827
264078 100.322 -75.309 46.801
264079 100.341 -75.326 46.775
264080 100.359 -75.342 46.747
264081 100.378 -75.358 46.719
264082 100.397 -75.374 46.692
264083 100.415 -75.391 46.663
264084 100.434 -75.407 46.635
264085 100.453 -75.424 46.608
264086 100.471 -75.439 46.580
264087 100.490 -75.456 46.551
264088 100.508 -75.472 46.522
264089 100.527 -75.489 46.493
264090 100.546 -75.505 46.464
264091 100.565 -75.521 46.436
264092 100.583 -75.537 46.408
264093 100.602 -75.554 46.381
264094 100.621 -75.570 46.352
264095 100.642 -75.581 46.324
264096 100.670 -75.569 46.296
264097 100.697 -75.557 46.268
264098 100.725 -75.545 46.242
264099 100.752 -75.533 46.215
264100 100.786 -75.526 46.188
264101 100.830 -75.527 46.161
264102 100.874 -75.529 46.134
264103 100.918 -75.531 46.107
264104 100.961 -75.536 46.080
264105 100.992 -75.551 46.053
264106 101.022 -75.565 46.025
264107 101.053 -75.579 45.996
264108 101.084 -75.593 45.965
264109 101.114 -75.607 45.934
264110 101.139 -75.618 45.901
264111 101.153 -75.627 45.870
264112 101.167 -75.636 45.838
264113 101.181 -75.645 45.807
264114 101.195 -75.654 45.777
264115 101.209 -75.662 45.746
264116 101.222 -75.671 45.716
264117 101.237 -75.681 45.689
264118 101.258 -75.687 45.659
264119 101.289 -75.690 45.629
264120 101.320 -75.693 45.598
264121 101.351 -75.696 45.566
264122 101.382 -75.699 45.535
264123 101.426 -75.719 45.505
264124 101.483 -75.757 45.476
264125 101.539 -75.796 45.448
264126 101.596 -75.834 45.422
264127 101.652 -75.872 45.395
264128 101.708 -75.910 45.369
264129 101.747 -75.935 45.342
264130 101.755 -75.938 45.317
264131 101.762 -75.941 45.291
264132 101.770 -75.944 45.266
264133 101.778 -75.946 45.241
264134 101.785 -75.949 45.216
264135 101.793 -75.952 45.191
264136 101.801 -75.954 45.166
264137 101.808 -75.957 45.143
264138 101.816 -75.960 45.118
264139 101.825 -75.964 45.094
264140 101.846 -75.984 45.070
264141 101.869 -76.003 45.046
264142 101.890 -76.022 45.022
264143 101.910 -76.033 44.998
264491 92.399 -73.520 54.678
264492 92.425 -73.531 54.674
264493 92.451 -73.541 54.670
264494 92.478 -73.552 54.665
264495 92.504 -73.562 54.660
264496 92.531 -73.572 54.657
264497 92.557 -73.583 54.652
264498 92.583 -73.593 54.646
264499 92.613 -73.604 54.642
264500 92.644 -73.614 54.638
264501 92.674 -73.625 54.635
264502 92.705 -73.634 54.632
264503 92.735 -73.645 54.629
264504 92.766 -73.656 54.627
264505 92.807 -73.666 54.626
264506 92.838 -73.665 54.626
264507 92.860 -73.656 54.625
264508 92.888 -73.662 54.626
264509 92.917 -73.669 54.628
264510 92.945 -73.675 54.630
264511 92.974 -73.681 54.633
264512 93.003 -73.687 54.637
264513 93.031 -73.694 54.640
264514 93.060 -73.700 54.645
264515 93.089 -73.706 54.650
264516 93.113 -73.703 54.654
264517 93.136 -73.701 54.658
264518 93.159 -73.698 54.660
264519 93.182 -73.695 54.660
264520 93.208 -73.697 54.658
264521 93.240 -73.706 54.655
264522 93.272 -73.715 54.651
264523 93.304 -73.724 54.647
264524 93.337 -73.728 54.644
264525 93.369 -73.728 54.641
264526 93.402 -73.730 54.637
264527 93.434 -73.731 54.634
264528 93.467 -73.732 54.632
264529 93.493 -73.734 54.629
264530 93.514 -73.738 54.628
264531 93.537 -73.741 54.625
264532 93.561 -73.744 54.621
264533 93.592 -73.745 54.614
264534 93.623 -73.747 54.606
264535 93.654 -73.748 54.596
264536 93.683 -73.750 54.585
264537 93.695 -73.751 54.571
264538 93.704 -73.751 54.554
264539 93.715 -73.751 54.535
264540 93.736 -73.752 54.518
264541 93.757 -73.753 54.504
264542 93.778 -73.753 54.493
264543 93.799 -73.753 54.483
264544 93.821 -73.753 54.476
264545 93.842 -73.749 54.468
264546 93.864 -73.745 54.459
264547 93.887 -73.741 54.453
264548 93.908 -73.737 54.447
264549 93.930 -73.733 54.440
264550 93.953 -73.729 54.429
264551 93.975 -73.725 54.415
264552 93.996 -73.721 54.397
264553 94.018 -73.716 54.375
264554 94.052 -73.716 54.346
264555 94.098 -73.718 54.316
264556 94.142 -73.720 54.289
264557 94.173 -73.719 54.265
264558 94.200 -73.721 54.244
264559 94.224 -73.719 54.230
264560 94.249 -73.719 54.215
264561 94.273 -73.718 54.199
264562 94.300 -73.722 54.182
264563 94.353 -73.765 54.155
264564 94.398 -73.791 54.127
264565 94.410 -73.791 54.097
264566 94.425 -73.795 54.066
264567 94.440 -73.799 54.035
264568 94.455 -73.803 54.003
264569 94.470 -73.807 53.972
264570 94.485 -73.811 53.943
264571 94.500 -73.815 53.912
264572 94.515 -73.819 53.879
264573 94.530 -73.822 53.844
264574 94.545 -73.827 53.810
264575 94.560 -73.831 53.775
264576 94.572 -73.817 53.741
264577 94.594 -73.830 53.707
264578 94.616 -73.844 53.674
264579 94.636 -73.849 53.642
264580 94.653 -73.847 53.612
264581 94.671 -73.844 53.585
264582 94.688 -73.843 53.562
264583 94.705 -73.841 53.541
264584 94.723 -73.838 53.525
264585 94.740 -73.836 53.509
264586 94.758 -73.832 53.492
264587 94.778 -73.822 53.467
264588 94.793 -73.824 53.439
264589 94.822 -73.832 53.409
264590 94.868 -73.850 53.378
264591 94.915 -73.866 53.343
264592 94.962 -73.884 53.306
264593 95.008 -73.901 53.268
264594 95.050 -73.914 53.228
264595 95.084 -73.922 53.183
264596 95.116 -73.921 53.135
264597 95.128 -73.900 53.086
264598 95.135 -73.888 53.036
264599 95.142 -73.877 52.986
264600 95.150 -73.866 52.938
264601 95.157 -73.854 52.892
264602 95.165 -73.843 52.847
264603 95.172 -73.832 52.805
264604 95.179 -73.820 52.763
264605 95.202 -73.819 52.723
264606 95.235 -73.822 52.684
264607 95.267 -73.826 52.647
264608 95.300 -73.830 52.611
264609 95.332 -73.835 52.574
264610 95.365 -73.838 52.536
264611 95.397 -73.842 52.497
264612 95.429 -73.846 52.459
264613 95.462 -73.850 52.421
264614 95.494 -73.855 52.384
264615 95.526 -73.869 52.347
264616 95.558 -73.884 52.310
264617 95.590 -73.898 52.273
264618 95.616 -73.886 52.236
264619 95.646 -73.878 52.202
264620 95.680 -73.873 52.167
264621 95.715 -73.869 52.133
264622 95.749 -73.866 52.101
264623 95.774 -73.860 52.068
264624 95.800 -73.855 52.034
264625 95.825 -73.850 51.999
264626 95.850 -73.844 51.965
264627 95.875 -73.840 51.933
264628 95.901 -73.835 51.901
264629 95.926 -73.829 51.868
264630 95.951 -73.824 51.837
264631 96.013 -73.816 51.800
264632 96.047 -73.806 51.763
264633 96.029 -73.791 51.725
264634 96.011 -73.776 51.687
264635 96.042 -73.795 51.649
264636 96.075 -73.816 51.611
264637 96.108 -73.836 51.574
264638 96.141 -73.857 51.538
264639 96.169 -73.883 51.503
264640 96.198 -73.909 51.468
264641 96.227 -73.935 51.435
264642 96.256 -73.960 51.400
264643 96.263 -73.969 51.368
264644 96.279 -73.979 51.335
264645 96.303 -73.987 51.303
264646 96.329 -73.994 51.272
264647 96.356 -74.001 51.243
264648 96.383 -74.007 51.215
264649 96.409 -74.014 51.188
264650 96.436 -74.021 51.161
264651 96.470 -74.021 51.138
264652 96.505 -74.021 51.115
264653 96.540 -74.021 51.093
264654 96.574 -74.023 51.072
264655 96.605 -74.037 51.053
264656 96.637 -74.051 51.036
264657 96.669 -74.064 51.019
264658 96.700 -74.078 51.003
264659 96.732 -74.092 50.989
264660 96.759 -74.104 50.976
264661 96.771 -74.108 50.966
264662 96.783 -74.114 50.957
264663 96.795 -74.118 50.953
264664 96.813 -74.132 50.950
264665 96.840 -74.157 50.947
264666 96.873 -74.174 50.945
264667 96.905 -74.190 50.945
264668 96.938 -74.206 50.948
264669 96.972 -74.214 50.953
264670 97.007 -74.207 50.960
264671 97.043 -74.199 50.972
264672 97.078 -74.192 50.983
264673 97.113 -74.185 50.992
264674 97.149 -74.177 51.001
264675 97.184 -74.170 51.004
264676 97.220 -74.163 51.003
264677 97.255 -74.155 50.999
264678 97.290 -74.148 50.992
264679 97.326 -74.141 50.976
264680 97.361 -74.133 50.961
264681 97.393 -74.125 50.945
264682 97.416 -74.110 50.928
264683 97.439 -74.095 50.908
264684 97.462 -74.097 50.885
264685 97.486 -74.103 50.859
264686 97.510 -74.110 50.829
264687 97.533 -74.116 50.797
264688 97.558 -74.123 50.764
264689 97.587 -74.138 50.729
264690 97.616 -74.153 50.693
264691 97.646 -74.167 50.656
264692 97.675 -74.182 50.616
264693 97.704 -74.196 50.572
264694 97.733 -74.211 50.526
264695 97.762 -74.225 50.480
264696 97.790 -74.240 50.434
264697 97.817 -74.256 50.388
264698 97.844 -74.272 50.341
264699 97.871 -74.288 50.291
264700 97.898 -74.300 50.240
264701 97.928 -74.312 50.188
264702 97.961 -74.326 50.136
264703 97.991 -74.340 50.091
264704 98.021 -74.355 50.048
264705 98.052 -74.369 50.002
264706 98.082 -74.383 49.950
264707 98.112 -74.397 49.895
264708 98.143 -74.411 49.838
264709 98.173 -74.425 49.779
264710 98.204 -74.440 49.720
264711 98.234 -74.454 49.661
264712 98.264 -74.468 49.602
264713 98.294 -74.482 49.543
264714 98.325 -74.496 49.485
264715 98.329 -74.506 49.427
264716 98.308 -74.513 49.371
264717 98.287 -74.519 49.316
264718 98.282 -74.531 49.259
264719 98.358 -74.546 49.200
264720 98.412 -74.577 49.142
264721 98.465 -74.608 49.085
264722 98.512 -74.668 49.032
264723 98.542 -74.711 48.982
264724 98.570 -74.747 48.933
264725 98.597 -74.781 48.884
264726 98.625 -74.817 48.833
264727 98.652 -74.852 48.782
264728 98.679 -74.888 48.731
264729 98.707 -74.922 48.679
264730 98.735 -74.958 48.625
264731 98.768 -74.980 48.572
264732 98.811 -74.935 48.518
264733 98.855 -74.890 48.463
264734 98.899 -74.845 48.410
264735 98.943 -74.800 48.360
264736 98.987 -74.755 48.310
264737 99.030 -74.709 48.259
264738 99.058 -74.700 48.209
264739 99.075 -74.715 48.160
264740 99.091 -74.730 48.111
264741 99.108 -74.744 48.062
264742 99.125 -74.759 48.015
264743 99.142 -74.773 47.968
264744 99.158 -74.787 47.921
264745 99.175 -74.802 47.874
264746 99.196 -74.819 47.829
264747 99.233 -74.844 47.784
264748 99.268 -74.866 47.739
264749 99.303 -74.888 47.696
264750 99.338 -74.910 47.654
264751 99.374 -74.927 47.611
264752 99.413 -74.938 47.570
264753 99.451 -74.949 47.530
264754 99.489 -74.960 47.490
264755 99.527 -74.971 47.450
264756 99.565 -74.979 47.412
264757 99.599 -74.980 47.375
264758 99.633 -74.981 47.339
264759 99.667 -74.982 47.305
264760 99.701 -74.982 47.273
264761 99.735 -74.983 47.242
264762 99.770 -74.984 47.212
264763 99.803 -74.985 47.183
264764 99.838 -74.985 47.155
264765 99.869 -74.989 47.128
264766 99.901 -74.993 47.102
264767 99.932 -74.997 47.076
264768 99.963 -75.001 47.051
264769 99.994 -75.006 47.026
264770 100.027 -75.013 47.002
264771 100.077 -75.041 46.977
264772 100.115 -75.076 46.955
264773 100.148 -75.114 46.934
264774 100.181 -75.152 46.915
264775 100.214 -75.190 46.892
264776 100.248 -75.226 46.868
264777 100.276 -75.242 46.842
264778 100.304 -75.258 46.816
264779 100.332 -75.274 46.790
264780 100.352 -75.291 46.763
264781 100.371 -75.307 46.736
264782 100.390 -75.323 46.709
264783 100.408 -75.339 46.681
264784 100.427 -75.356 46.653
264785 100.446 -75.372 46.624
264786 100.465 -75.389 46.598
264787 100.483 -75.405 46.570
264788 100.502 -75.421 46.541
264789 100.521 -75.437 46.513
264790 100.539 -75.454 46.484
264791 100.558 -75.470 46.454
264792 100.576 -75.486 46.426
264793 100.595 -75.502 46.398
264794 100.614 -75.519 46.371
264795 100.632 -75.535 46.343
264796 100.659 -75.543 46.315
264797 100.687 -75.537 46.287
264798 100.715 -75.527 46.259
264799 100.742 -75.515 46.233
264800 100.770 -75.503 46.206
264801 100.805 -75.497 46.178
264802 100.848 -75.499 46.153
264803 100.892 -75.501 46.125
264804 100.936 -75.502 46.099
264805 100.981 -75.503 46.072
264806 101.025 -75.533 46.045
264807 101.055 -75.546 46.016
264808 101.086 -75.560 45.987
264809 101.116 -75.574 45.957
264810 101.146 -75.587 45.926
264811 101.155 -75.593 45.894
264812 101.173 -75.588 45.861
264813 101.186 -75.597 45.829
264814 101.200 -75.606 45.797
264815 101.214 -75.615 45.765
264816 101.229 -75.624 45.733
264817 101.242 -75.633 45.702
264818 101.256 -75.642 45.672
264819 101.277 -75.649 45.643
264820 101.308 -75.652 45.613
264821 101.339 -75.655 45.582
264822 101.370 -75.658 45.552
264823 101.403 -75.668 45.520
264824 101.451 -75.703 45.491
264825 101.508 -75.741 45.463
264826 101.564 -75.780 45.435
264827 101.621 -75.818 45.408
264828 101.677 -75.856 45.381
264829 101.734 -75.894 45.355
264830 101.751 -75.904 45.329
264831 101.758 -75.906 45.303
264832 101.766 -75.909 45.277
264833 101.774 -75.912 45.251
264834 101.781 -75.915 45.226
264835 101.789 -75.918 45.201
264836 101.796 -75.920 45.176
264837 101.804 -75.923 45.152
264838 101.812 -75.925 45.128
264839 101.825 -75.934 45.104
264840 101.846 -75.954 45.079
264841 101.869 -75.974 45.055
264842 101.890 -75.992 45.031
264843 101.907 -75.993 45.007
264844 101.925 -75.993 44.984
265191 92.353 -73.481 54.692
265192 92.379 -73.491 54.685
265193 92.405 -73.502 54.680
265194 92.432 -73.512 54.675
265195 92.458 -73.523 54.670
265196 92.486 -73.533 54.665
265197 92.517 -73.543 54.660
265198 92.548 -73.554 54.654
265199 92.578 -73.564 54.648
265200 92.609 -73.575 54.643
265201 92.639 -73.585 54.639
265202 92.670 -73.595 54.635
265203 92.701 -73.606 54.631
265204 92.731 -73.616 54.628
265205 92.762 -73.626 54.625
265206 92.804 -73.637 54.624
265207 92.847 -73.648 54.623
265208 92.871 -73.641 54.622
265209 92.894 -73.632 54.623
265210 92.917 -73.627 54.625
265211 92.945 -73.634 54.628
265212 92.974 -73.640 54.631
265213 93.003 -73.646 54.634
265214 93.031 -73.652 54.637
265215 93.060 -73.659 54.641
265216 93.089 -73.665 54.646
265217 93.114 -73.666 54.650
265218 93.137 -73.662 54.655
265219 93.160 -73.659 54.657
265220 93.184 -73.656 54.656
265221 93.214 -73.664 54.654
265222 93.247 -73.673 54.651
265223 93.278 -73.682 54.647
265224 93.310 -73.691 54.644
265225 93.343 -73.700 54.641
265226 93.375 -73.703 54.637
265227 93.408 -73.703 54.632
265228 93.440 -73.705 54.627
265229 93.473 -73.706 54.623
265230 93.494 -73.709 54.618
265231 93.516 -73.713 54.616
265232 93.538 -73.716 54.613
265233 93.560 -73.719 54.609
265234 93.583 -73.722 54.603
265235 93.615 -73.723 54.596
265236 93.646 -73.725 54.587
265237 93.675 -73.726 54.575
265238 93.689 -73.727 54.561
265239 93.698 -73.727 54.543
265240 93.715 -73.728 54.523
265241 93.736 -73.728 54.504
265242 93.757 -73.729 54.488
265243 93.778 -73.729 54.474
265244 93.800 -73.726 54.461
265245 93.822 -73.722 54.451
265246 93.844 -73.718 54.440
265247 93.866 -73.714 54.427
265248 93.888 -73.709 54.417
265249 93.910 -73.706 54.406
265250 93.932 -73.701 54.395
265251 93.954 -73.697 54.382
265252 93.976 -73.694 54.368
265253 93.998 -73.689 54.351
265254 94.020 -73.685 54.331
265255 94.042 -73.681 54.307
265256 94.084 -73.683 54.280
265257 94.120 -73.678 54.255
265258 94.156 -73.673 54.233
265259 94.185 -73.675 54.213
265260 94.213 -73.678 54.196
265261 94.242 -73.680 54.178
265262 94.270 -73.682 54.160
265263 94.299 -73.685 54.140
265264 94.327 -73.698 54.114
265265 94.359 -73.725 54.086
265266 94.392 -73.752 54.058
265267 94.424 -73.779 54.029
265268 94.448 -73.794 53.998
265269 94.463 -73.799 53.968
265270 94.478 -73.803 53.937
265271 94.493 -73.807 53.906
265272 94.508 -73.810 53.874
265273 94.523 -73.814 53.839
265274 94.538 -73.819 53.804
265275 94.553 -73.822 53.770
265276 94.568 -73.826 53.738
265277 94.582 -73.823 53.705
265278 94.595 -73.812 53.674
265279 94.617 -73.826 53.644
265280 94.637 -73.834 53.614
265281 94.655 -73.832 53.587
265282 94.673 -73.829 53.561
265283 94.690 -73.827 53.540
265284 94.707 -73.825 53.522
265285 94.724 -73.822 53.507
265286 94.742 -73.820 53.493
265287 94.760 -73.818 53.474
265288 94.778 -73.813 53.449
265289 94.795 -73.810 53.421
265290 94.824 -73.819 53.391
265291 94.871 -73.835 53.359
265292 94.918 -73.853 53.323
265293 94.965 -73.869 53.284
265294 94.999 -73.879 53.244
265295 95.034 -73.887 53.200
265296 95.068 -73.895 53.155
265297 95.102 -73.904 53.107
265298 95.127 -73.870 53.058
265299 95.136 -73.856 53.010
265300 95.143 -73.844 52.963
265301 95.150 -73.833 52.917
265302 95.158 -73.822 52.874
265303 95.166 -73.810 52.831
265304 95.173 -73.799 52.790
265305 95.180 -73.788 52.750
265306 95.210 -73.790 52.711
265307 95.242 -73.794 52.673
265308 95.274 -73.797 52.638
265309 95.307 -73.801 52.602
265310 95.339 -73.806 52.566
265311 95.372 -73.810 52.528
265312 95.404 -73.813 52.489
265313 95.437 -73.817 52.452
265314 95.469 -73.828 52.415
265315 95.500 -73.842 52.378
265316 95.533 -73.856 52.341
265317 95.564 -73.870 52.305
265318 95.594 -73.875 52.269
265319 95.621 -73.863 52.233
265320 95.649 -73.853 52.199
265321 95.683 -73.849 52.166
265322 95.718 -73.844 52.133
265323 95.752 -73.841 52.100
265324 95.787 -73.836 52.067
265325 95.814 -73.832 52.033
265326 95.840 -73.826 51.998
265327 95.865 -73.821 51.964
265328 95.890 -73.816 51.931
265329 95.915 -73.811 51.898
265330 95.941 -73.806 51.865
265331 95.976 -73.800 51.832
265332 96.039 -73.791 51.797
265333 96.066 -73.781 51.761
265334 96.048 -73.766 51.725
265335 96.031 -73.751 51.690
265336 96.049 -73.762 51.653
265337 96.078 -73.788 51.616
265338 96.106 -73.813 51.579
265339 96.135 -73.839 51.544
265340 96.164 -73.865 51.509
265341 96.193 -73.891 51.474
265342 96.222 -73.916 51.441
265343 96.242 -73.938 51.407
265344 96.258 -73.948 51.374
265345 96.281 -73.956 51.342
265346 96.308 -73.963 51.311
265347 96.334 -73.969 51.279
265348 96.361 -73.976 51.250
265349 96.388 -73.983 51.221
265350 96.414 -73.990 51.193
265351 96.441 -73.997 51.167
265352 96.472 -73.999 51.142
265353 96.507 -73.999 51.119
265354 96.542 -73.999 51.096
265355 96.576 -73.999 51.075
265356 96.611 -73.999 51.056
265357 96.643 -74.011 51.038
265358 96.674 -74.026 51.020
265359 96.706 -74.039 51.004
265360 96.738 -74.053 50.989
265361 96.765 -74.065 50.976
265362 96.777 -74.070 50.964
265363 96.789 -74.074 50.955
265364 96.804 -74.083 50.950
265365 96.831 -74.109 50.946
265366 96.864 -74.125 50.943
265367 96.897 -74.141 50.939
265368 96.930 -74.157 50.940
265369 96.962 -74.173 50.942
265370 96.996 -74.186 50.946
265371 97.031 -74.179 50.953
265372 97.066 -74.172 50.964
265373 97.101 -74.164 50.974
265374 97.137 -74.157 50.984
265375 97.172 -74.150 50.992
265376 97.208 -74.142 50.993
265377 97.243 -74.136 50.991
265378 97.279 -74.128 50.985
265379 97.314 -74.120 50.978
265380 97.348 -74.112 50.963
265381 97.370 -74.098 50.948
265382 97.393 -74.083 50.932
265383 97.416 -74.069 50.914
265384 97.439 -74.068 50.897
265385 97.462 -74.074 50.877
265386 97.486 -74.080 50.853
265387 97.510 -74.087 50.825
265388 97.533 -74.093 50.793
265389 97.557 -74.099 50.761
265390 97.582 -74.107 50.728
265391 97.611 -74.121 50.691
265392 97.640 -74.136 50.656
265393 97.669 -74.151 50.616
265394 97.699 -74.165 50.573
265395 97.728 -74.179 50.526
265396 97.757 -74.194 50.479
265397 97.786 -74.208 50.433
265398 97.815 -74.223 50.387
265399 97.845 -74.238 50.339
265400 97.874 -74.252 50.291
265401 97.904 -74.267 50.240
265402 97.934 -74.280 50.187
265403 97.965 -74.295 50.134
265404 97.995 -74.309 50.088
265405 98.025 -74.323 50.044
265406 98.056 -74.337 49.998
265407 98.086 -74.352 49.946
265408 98.117 -74.365 49.890
265409 98.147 -74.380 49.832
265410 98.177 -74.394 49.773
265411 98.208 -74.408 49.713
265412 98.238 -74.422 49.655
265413 98.268 -74.436 49.595
265414 98.298 -74.450 49.536
265415 98.329 -74.465 49.478
265416 98.330 -74.474 49.421
265417 98.309 -74.480 49.365
265418 98.288 -74.487 49.309
265419 98.317 -74.512 49.252
265420 98.386 -74.542 49.193
265421 98.463 -74.541 49.134
265422 98.517 -74.572 49.077
265423 98.563 -74.639 49.021
265424 98.600 -74.710 48.971
265425 98.628 -74.745 48.922
265426 98.656 -74.781 48.873
265427 98.683 -74.816 48.823
265428 98.711 -74.851 48.772
265429 98.738 -74.887 48.720
265430 98.759 -74.911 48.668
265431 98.769 -74.922 48.615
265432 98.782 -74.929 48.562
265433 98.825 -74.907 48.508
265434 98.867 -74.872 48.453
265435 98.912 -74.827 48.399
265436 98.955 -74.782 48.348
265437 98.999 -74.737 48.297
265438 99.043 -74.691 48.247
265439 99.074 -74.675 48.197
265440 99.090 -74.690 48.148
265441 99.108 -74.704 48.100
265442 99.124 -74.718 48.052
265443 99.141 -74.733 48.004
265444 99.157 -74.747 47.957
265445 99.174 -74.762 47.911
265446 99.191 -74.776 47.865
265447 99.207 -74.790 47.819
265448 99.237 -74.814 47.775
265449 99.276 -74.842 47.731
265450 99.311 -74.864 47.687
265451 99.346 -74.885 47.645
265452 99.384 -74.895 47.604
265453 99.423 -74.907 47.563
265454 99.461 -74.918 47.524
265455 99.499 -74.928 47.485
265456 99.538 -74.940 47.446
265457 99.576 -74.951 47.407
265458 99.611 -74.953 47.370
265459 99.645 -74.954 47.334
265460 99.679 -74.954 47.299
265461 99.713 -74.955 47.266
265462 99.748 -74.957 47.235
265463 99.782 -74.957 47.205
265464 99.816 -74.958 47.175
265465 99.850 -74.959 47.147
265466 99.884 -74.960 47.119
265467 99.918 -74.960 47.094
265468 99.950 -74.963 47.067
265469 99.982 -74.967 47.041
265470 100.013 -74.971 47.017
265471 100.048 -74.981 46.992
265472 100.091 -75.013 46.967
265473 100.124 -75.051 46.945
265474 100.157 -75.089 46.924
265475 100.190 -75.127 46.904
265476 100.223 -75.165 46.881
265477 100.254 -75.192 46.856
265478 100.282 -75.208 46.831
265479 100.310 -75.224 46.805
265480 100.338 -75.240 46.779
265481 100.364 -75.256 46.752
265482 100.383 -75.272 46.725
265483 100.402 -75.289 46.698
265484 100.420 -75.305 46.670
265485 100.439 -75.321 46.642
265486 100.458 -75.337 46.614
265487 100.476 -75.354 46.587
265488 100.495 -75.370 46.559
265489 100.513 -75.386 46.532
265490 100.533 -75.402 46.504
265491 100.551 -75.419 46.474
265492 100.570 -75.435 46.444
265493 100.588 -75.452 46.416
265494 100.607 -75.467 46.388
265495 100.626 -75.484 46.361
265496 100.650 -75.496 46.333
265497 100.678 -75.504 46.305
265498 100.705 -75.499 46.277
265499 100.733 -75.493 46.249
265500 100.760 -75.486 46.222
265501 100.787 -75.474 46.195
265502 100.823 -75.468 46.168
265503 100.867 -75.471 46.141
265504 100.912 -75.469 46.114
265505 100.957 -75.465 46.088
265506 101.001 -75.470 46.061
265507 101.047 -75.502 46.033
265508 101.093 -75.534 46.004
265509 101.125 -75.550 45.975
265510 101.154 -75.564 45.946
265511 101.161 -75.572 45.915
265512 101.169 -75.580 45.882
265513 101.187 -75.570 45.849
265514 101.206 -75.559 45.817
265515 101.220 -75.568 45.785
265516 101.234 -75.577 45.753
265517 101.248 -75.586 45.721
265518 101.262 -75.594 45.688
265519 101.276 -75.603 45.657
265520 101.296 -75.610 45.627
265521 101.327 -75.613 45.597
265522 101.358 -75.616 45.566
265523 101.389 -75.620 45.536
265524 101.429 -75.652 45.506
265525 101.476 -75.687 45.477
265526 101.533 -75.725 45.449
265527 101.590 -75.764 45.422
265528 101.646 -75.802 45.393
265529 101.702 -75.840 45.366
265530 101.747 -75.870 45.340
265531 101.755 -75.872 45.314
265532 101.762 -75.875 45.288
265533 101.770 -75.878 45.262
265534 101.777 -75.881 45.236
265535 101.785 -75.884 45.211
265536 101.793 -75.886 45.186
265537 101.801 -75.889 45.162
265538 101.808 -75.891 45.137
265539 101.825 -75.905 45.112
265540 101.846 -75.925 45.088
265541 101.869 -75.944 45.064
265542 101.888 -75.952 45.040
265543 101.905 -75.953 45.015
265544 101.922 -75.953 44.992
265545 101.940 -75.953 44.969
265892 92.333 -73.452 54.699
265893 92.359 -73.462 54.692
265894 92.390 -73.473 54.686
265895 92.421 -73.484 54.680
265896 92.451 -73.493 54.674
265897 92.482 -73.504 54.668
265898 92.513 -73.515 54.663
265899 92.544 -73.524 54.657
265900 92.574 -73.535 54.650
265901 92.604 -73.546 54.644
265902 92.635 -73.556 54.639
265903 92.666 -73.566 54.635
265904 92.697 -73.576 54.631
265905 92.727 -73.587 54.626
265906 92.758 -73.597 54.623
265907 92.801 -73.608 54.620
265908 92.844 -73.619 54.619
265909 92.882 -73.625 54.618
265910 92.904 -73.617 54.619
265911 92.927 -73.609 54.621
265912 92.949 -73.600 54.623
265913 92.974 -73.598 54.628
265914 93.003 -73.605 54.630
265915 93.031 -73.611 54.633
265916 93.060 -73.617 54.638
265917 93.089 -73.624 54.642
265918 93.116 -73.627 54.646
265919 93.139 -73.624 54.649
265920 93.162 -73.621 54.651
265921 93.189 -73.623 54.651
265922 93.221 -73.631 54.649
265923 93.253 -73.640 54.645
265924 93.285 -73.650 54.642
265925 93.317 -73.659 54.639
265926 93.349 -73.668 54.636
265927 93.381 -73.676 54.631
265928 93.414 -73.677 54.627
265929 93.446 -73.678 54.620
265930 93.474 -73.681 54.614
265931 93.496 -73.684 54.608
265932 93.517 -73.687 54.603
265933 93.540 -73.690 54.600
265934 93.561 -73.693 54.595
265935 93.583 -73.697 54.589
265936 93.606 -73.699 54.584
265937 93.637 -73.700 54.575
265938 93.667 -73.702 54.564
265939 93.682 -73.703 54.550
265940 93.694 -73.703 54.532
265941 93.715 -73.704 54.511
265942 93.736 -73.705 54.491
265943 93.758 -73.703 54.472
265944 93.780 -73.699 54.456
265945 93.802 -73.694 54.441
265946 93.824 -73.691 54.428
265947 93.846 -73.687 54.414
265948 93.868 -73.682 54.399
265949 93.890 -73.678 54.385
265950 93.912 -73.674 54.371
265951 93.934 -73.670 54.357
265952 93.956 -73.666 54.343
265953 93.978 -73.662 54.328
265954 94.000 -73.659 54.311
265955 94.021 -73.656 54.291
265956 94.042 -73.653 54.268
265957 94.071 -73.651 54.245
265958 94.110 -73.649 54.222
265959 94.146 -73.649 54.202
265960 94.175 -73.652 54.183
265961 94.203 -73.656 54.163
265962 94.233 -73.659 54.142
265963 94.261 -73.663 54.121
265964 94.285 -73.675 54.099
265965 94.309 -73.685 54.072
265966 94.333 -73.696 54.045
265967 94.357 -73.706 54.017
265968 94.381 -73.717 53.987
265969 94.405 -73.728 53.958
265970 94.429 -73.739 53.928
265971 94.461 -73.765 53.897
265972 94.493 -73.792 53.865
265973 94.516 -73.806 53.834
265974 94.531 -73.810 53.802
265975 94.546 -73.814 53.768
265976 94.561 -73.818 53.735
265977 94.576 -73.822 53.704
265978 94.591 -73.825 53.674
265979 94.604 -73.815 53.643
265980 94.618 -73.807 53.614
265981 94.639 -73.819 53.586
265982 94.657 -73.816 53.561
265983 94.674 -73.814 53.538
265984 94.692 -73.812 53.516
265985 94.709 -73.810 53.500
265986 94.726 -73.807 53.486
265987 94.744 -73.805 53.471
265988 94.761 -73.803 53.452
265989 94.779 -73.800 53.428
265990 94.797 -73.795 53.401
265991 94.827 -73.804 53.371
265992 94.874 -73.821 53.338
265993 94.915 -73.835 53.300
265994 94.950 -73.843 53.259
265995 94.984 -73.852 53.216
265996 95.018 -73.860 53.171
265997 95.052 -73.869 53.125
265998 95.087 -73.877 53.078
265999 95.116 -73.864 53.031
266000 95.137 -73.823 52.985
266001 95.144 -73.812 52.941
266002 95.151 -73.800 52.898
266003 95.159 -73.789 52.856
266004 95.167 -73.778 52.815
266005 95.174 -73.766 52.775
266006 95.184 -73.757 52.736
266007 95.217 -73.761 52.699
266008 95.249 -73.765 52.663
266009 95.282 -73.769 52.628
266010 95.314 -73.772 52.593
266011 95.346 -73.777 52.558
266012 95.379 -73.781 52.519
266013 95.411 -73.785 52.481
266014 95.443 -73.800 52.444
266015 95.475 -73.814 52.408
266016 95.507 -73.828 52.372
266017 95.539 -73.843 52.335
266018 95.571 -73.857 52.300
266019 95.599 -73.851 52.265
266020 95.625 -73.839 52.229
266021 95.652 -73.828 52.197
266022 95.687 -73.824 52.164
266023 95.721 -73.819 52.132
266024 95.755 -73.816 52.100
266025 95.790 -73.811 52.067
266026 95.824 -73.807 52.033
266027 95.854 -73.803 51.998
266028 95.880 -73.797 51.964
266029 95.905 -73.792 51.930
266030 95.931 -73.788 51.897
266031 95.956 -73.782 51.864
266032 96.002 -73.775 51.831
266033 96.066 -73.767 51.796
266034 96.086 -73.756 51.761
266035 96.068 -73.741 51.727
266036 96.051 -73.735 51.693
266037 96.052 -73.744 51.657
266038 96.076 -73.768 51.621
266039 96.102 -73.794 51.585
266040 96.130 -73.821 51.549
266041 96.159 -73.847 51.515
266042 96.188 -73.872 51.481
266043 96.217 -73.898 51.447
266044 96.237 -73.917 51.415
266045 96.259 -73.925 51.381
266046 96.286 -73.932 51.349
266047 96.313 -73.938 51.317
266048 96.339 -73.945 51.286
266049 96.366 -73.952 51.256
266050 96.393 -73.959 51.227
266051 96.419 -73.966 51.199
266052 96.446 -73.973 51.172
266053 96.475 -73.978 51.148
266054 96.509 -73.978 51.123
266055 96.544 -73.978 51.100
266056 96.578 -73.978 51.078
266057 96.613 -73.978 51.058
266058 96.648 -73.978 51.039
266059 96.680 -73.986 51.022
266060 96.712 -74.000 51.005
266061 96.743 -74.014 50.989
266062 96.771 -74.026 50.975
266063 96.783 -74.031 50.963
266064 96.795 -74.035 50.953
266065 96.823 -74.060 50.946
266066 96.856 -74.076 50.941
266067 96.888 -74.092 50.937
266068 96.921 -74.108 50.934
266069 96.954 -74.125 50.933
266070 96.986 -74.141 50.935
266071 97.019 -74.157 50.939
266072 97.054 -74.151 50.946
266073 97.090 -74.145 50.957
266074 97.125 -74.137 50.967
266075 97.160 -74.129 50.975
266076 97.196 -74.123 50.983
266077 97.231 -74.115 50.983
266078 97.266 -74.107 50.979
266079 97.302 -74.100 50.972
266080 97.324 -74.085 50.963
266081 97.347 -74.071 50.948
266082 97.369 -74.057 50.933
266083 97.392 -74.042 50.918
266084 97.415 -74.038 50.900
266085 97.439 -74.045 50.884
266086 97.462 -74.051 50.865
266087 97.486 -74.057 50.844
266088 97.510 -74.064 50.818
266089 97.533 -74.070 50.789
266090 97.558 -74.076 50.758
266091 97.581 -74.083 50.726
266092 97.605 -74.090 50.691
266093 97.635 -74.105 50.656
266094 97.664 -74.120 50.616
266095 97.693 -74.134 50.572
266096 97.722 -74.148 50.523
266097 97.751 -74.163 50.476
266098 97.782 -74.178 50.429
266099 97.813 -74.194 50.383
266100 97.845 -74.209 50.335
266101 97.876 -74.225 50.286
266102 97.907 -74.239 50.235
266103 97.938 -74.249 50.183
266104 97.968 -74.263 50.130
266105 97.999 -74.277 50.084
266106 98.029 -74.292 50.040
266107 98.060 -74.305 49.994
266108 98.090 -74.320 49.942
266109 98.121 -74.334 49.884
266110 98.151 -74.348 49.824
266111 98.181 -74.362 49.764
266112 98.212 -74.377 49.704
266113 98.242 -74.390 49.646
266114 98.272 -74.405 49.587
266115 98.302 -74.419 49.528
266116 98.333 -74.433 49.470
266117 98.330 -74.443 49.413
266118 98.309 -74.449 49.358
266119 98.289 -74.455 49.302
266120 98.353 -74.493 49.245
266121 98.418 -74.530 49.185
266122 98.508 -74.515 49.126
266123 98.569 -74.537 49.067
266124 98.613 -74.609 49.011
266125 98.655 -74.693 48.960
266126 98.687 -74.744 48.911
266127 98.714 -74.780 48.862
266128 98.742 -74.815 48.811
266129 98.756 -74.831 48.760
266130 98.767 -74.841 48.709
266131 98.777 -74.851 48.657
266132 98.788 -74.861 48.604
266133 98.798 -74.871 48.550
266134 98.840 -74.849 48.496
266135 98.882 -74.827 48.442
266136 98.924 -74.804 48.388
266137 98.968 -74.764 48.336
266138 99.011 -74.718 48.284
266139 99.056 -74.674 48.234
266140 99.090 -74.650 48.185
266141 99.106 -74.664 48.136
266142 99.123 -74.678 48.088
266143 99.140 -74.693 48.040
266144 99.156 -74.707 47.993
266145 99.173 -74.722 47.946
266146 99.190 -74.736 47.900
266147 99.206 -74.750 47.855
266148 99.222 -74.764 47.810
266149 99.239 -74.778 47.765
266150 99.278 -74.809 47.721
266151 99.318 -74.838 47.678
266152 99.356 -74.853 47.637
266153 99.395 -74.864 47.597
266154 99.433 -74.875 47.557
266155 99.472 -74.886 47.518
266156 99.510 -74.897 47.479
266157 99.548 -74.909 47.441
266158 99.586 -74.919 47.403
266159 99.623 -74.926 47.365
266160 99.657 -74.927 47.328
266161 99.691 -74.928 47.293
266162 99.725 -74.928 47.260
266163 99.760 -74.929 47.228
266164 99.794 -74.930 47.197
266165 99.828 -74.931 47.168
266166 99.862 -74.932 47.139
266167 99.896 -74.932 47.111
266168 99.930 -74.933 47.085
266169 99.964 -74.934 47.059
266170 99.998 -74.935 47.033
266171 100.031 -74.937 47.007
266172 100.067 -74.950 46.982
266173 100.101 -74.988 46.958
266174 100.133 -75.026 46.936
266175 100.166 -75.064 46.914
266176 100.199 -75.102 46.893
266177 100.232 -75.140 46.870
266178 100.261 -75.157 46.846
266179 100.289 -75.173 46.819
266180 100.317 -75.189 46.793
266181 100.345 -75.205 46.767
266182 100.372 -75.221 46.741
266183 100.395 -75.237 46.714
266184 100.414 -75.254 46.687
266185 100.432 -75.270 46.660
266186 100.451 -75.286 46.632
266187 100.470 -75.302 46.604
266188 100.488 -75.319 46.577
266189 100.507 -75.335 46.550
266190 100.525 -75.352 46.522
266191 100.544 -75.367 46.494
266192 100.563 -75.384 46.464
266193 100.581 -75.400 46.435
266194 100.600 -75.417 46.406
266195 100.619 -75.433 46.379
266196 100.640 -75.447 46.351
266197 100.668 -75.458 46.323
266198 100.695 -75.465 46.296
266199 100.722 -75.460 46.268
266200 100.750 -75.455 46.239
266201 100.777 -75.449 46.212
266202 100.805 -75.444 46.186
266203 100.842 -75.439 46.159
266204 100.887 -75.436 46.130
266205 100.932 -75.433 46.103
266206 100.977 -75.430 46.076
266207 101.022 -75.437 46.048
266208 101.067 -75.469 46.021
266209 101.113 -75.501 45.994
266210 101.156 -75.533 45.965
266211 101.169 -75.548 45.935
266212 101.177 -75.556 45.904
266213 101.184 -75.565 45.870
266214 101.200 -75.558 45.837
266215 101.219 -75.547 45.805
266216 101.238 -75.536 45.772
266217 101.254 -75.538 45.740
266218 101.268 -75.547 45.708
266219 101.281 -75.556 45.675
266220 101.296 -75.565 45.642
266221 101.316 -75.571 45.611
266222 101.347 -75.574 45.580
266223 101.378 -75.578 45.550
266224 101.416 -75.604 45.521
266225 101.456 -75.637 45.492
266226 101.501 -75.671 45.463
266227 101.558 -75.710 45.434
266228 101.615 -75.748 45.406
266229 101.671 -75.787 45.378
266230 101.728 -75.825 45.351
266231 101.751 -75.838 45.325
266232 101.758 -75.841 45.298
266233 101.766 -75.844 45.272
266234 101.773 -75.847 45.246
266235 101.781 -75.850 45.221
266236 101.789 -75.852 45.195
266237 101.796 -75.855 45.171
266238 101.804 -75.858 45.146
266239 101.825 -75.875 45.121
266240 101.846 -75.895 45.096
266241 101.868 -75.911 45.072
266242 101.885 -75.912 45.047
266243 101.902 -75.912 45.023
266244 101.920 -75.912 45.000
266245 101.938 -75.913 44.976
266592 92.294 -73.413 54.714
266593 92.324 -73.423 54.706
266594 92.355 -73.434 54.699
266595 92.386 -73.444 54.692
266596 92.417 -73.454 54.685
266597 92.447 -73.465 54.678
266598 92.478 -73.474 54.672
266599 92.508 -73.485 54.665
266600 92.539 -73.496 54.659
266601 92.570 -73.506 54.652
266602 92.600 -73.516 54.646
266603 92.631 -73.527 54.641
266604 92.662 -73.537 54.635
266605 92.693 -73.547 54.630
266606 92.723 -73.558 54.625
266607 92.755 -73.568 54.620
266608 92.798 -73.579 54.617
266609 92.841 -73.590 54.615
266610 92.884 -73.601 54.614
266611 92.915 -73.601 54.615
266612 92.938 -73.593 54.617
266613 92.960 -73.584 54.619
266614 92.982 -73.576 54.623
266615 93.005 -73.567 54.626
266616 93.031 -73.570 54.630
266617 93.060 -73.576 54.634
266618 93.089 -73.582 54.637
266619 93.118 -73.588 54.641
266620 93.141 -73.585 54.644
266621 93.164 -73.582 54.646
266622 93.195 -73.590 54.645
266623 93.227 -73.600 54.643
266624 93.259 -73.609 54.640
266625 93.291 -73.618 54.637
266626 93.324 -73.627 54.634
266627 93.355 -73.636 54.630
266628 93.388 -73.645 54.625
266629 93.420 -73.651 54.620
266630 93.452 -73.652 54.613
266631 93.476 -73.655 54.605
266632 93.497 -73.658 54.598
266633 93.519 -73.661 54.591
266634 93.541 -73.664 54.585
266635 93.563 -73.668 54.578
266636 93.584 -73.671 54.571
266637 93.606 -73.674 54.565
266638 93.628 -73.677 54.558
266639 93.659 -73.678 54.550
266640 93.676 -73.679 54.538
266641 93.694 -73.680 54.521
266642 93.715 -73.679 54.500
266643 93.737 -73.675 54.479
266644 93.760 -73.672 54.459
266645 93.781 -73.667 54.439
266646 93.803 -73.663 54.422
266647 93.826 -73.659 54.407
266648 93.847 -73.655 54.391
266649 93.869 -73.652 54.374
266650 93.890 -73.650 54.358
266651 93.911 -73.647 54.342
266652 93.932 -73.644 54.326
266653 93.954 -73.642 54.309
266654 93.975 -73.639 54.293
266655 93.996 -73.637 54.275
266656 94.017 -73.634 54.254
266657 94.038 -73.631 54.232
266658 94.067 -73.630 54.211
266659 94.105 -73.628 54.191
266660 94.137 -73.630 54.172
266661 94.166 -73.633 54.152
266662 94.195 -73.637 54.130
266663 94.221 -73.644 54.107
266664 94.245 -73.655 54.084
266665 94.269 -73.666 54.060
266666 94.293 -73.676 54.033
266667 94.317 -73.687 54.006
266668 94.341 -73.698 53.978
266669 94.365 -73.709 53.949
266670 94.389 -73.719 53.920
266671 94.413 -73.730 53.890
266672 94.437 -73.741 53.860
266673 94.461 -73.752 53.828
266674 94.485 -73.763 53.799
266675 94.509 -73.773 53.767
266676 94.533 -73.784 53.735
266677 94.562 -73.805 53.703
266678 94.584 -73.817 53.673
266679 94.599 -73.822 53.643
266680 94.613 -73.821 53.614
266681 94.625 -73.806 53.586
266682 94.641 -73.803 53.560
266683 94.659 -73.801 53.535
266684 94.676 -73.799 53.513
266685 94.694 -73.797 53.492
266686 94.711 -73.794 53.476
266687 94.729 -73.792 53.462
266688 94.746 -73.790 53.448
266689 94.763 -73.788 53.429
266690 94.781 -73.785 53.405
266691 94.798 -73.783 53.377
266692 94.830 -73.790 53.346
266693 94.865 -73.799 53.311
266694 94.899 -73.808 53.272
266695 94.933 -73.816 53.230
266696 94.968 -73.825 53.186
266697 95.002 -73.833 53.141
266698 95.036 -73.842 53.095
266699 95.070 -73.850 53.050
266700 95.105 -73.859 53.005
266701 95.127 -73.813 52.962
266702 95.145 -73.779 52.921
266703 95.153 -73.768 52.880
266704 95.160 -73.756 52.839
266705 95.167 -73.745 52.798
266706 95.175 -73.734 52.761
266707 95.192 -73.728 52.724
266708 95.224 -73.732 52.689
266709 95.257 -73.736 52.654
266710 95.289 -73.740 52.619
266711 95.321 -73.744 52.585
266712 95.354 -73.748 52.549
266713 95.386 -73.758 52.511
266714 95.418 -73.772 52.473
266715 95.450 -73.786 52.436
266716 95.482 -73.800 52.401
266717 95.514 -73.815 52.365
266718 95.545 -73.829 52.329
266719 95.577 -73.839 52.294
266720 95.603 -73.827 52.260
266721 95.629 -73.816 52.226
266722 95.656 -73.803 52.194
266723 95.690 -73.799 52.163
266724 95.724 -73.794 52.132
266725 95.758 -73.791 52.100
266726 95.793 -73.787 52.067
266727 95.827 -73.782 52.033
266728 95.861 -73.778 51.999
266729 95.895 -73.774 51.965
266730 95.920 -73.769 51.932
266731 95.946 -73.764 51.898
266732 95.971 -73.759 51.864
266733 96.029 -73.750 51.831
266734 96.092 -73.743 51.796
266735 96.106 -73.731 51.763
266736 96.089 -73.728 51.730
266737 96.072 -73.731 51.696
266738 96.076 -73.739 51.661
266739 96.084 -73.749 51.626
266740 96.107 -73.773 51.591
266741 96.133 -73.800 51.555
266742 96.159 -73.826 51.521
266743 96.185 -73.854 51.488
266744 96.212 -73.880 51.454
266745 96.238 -73.894 51.422
266746 96.264 -73.901 51.389
266747 96.291 -73.908 51.356
266748 96.318 -73.915 51.324
266749 96.344 -73.922 51.293
266750 96.371 -73.929 51.263
266751 96.398 -73.935 51.234
266752 96.425 -73.942 51.205
266753 96.451 -73.949 51.178
266754 96.477 -73.956 51.153
266755 96.511 -73.957 51.128
266756 96.546 -73.957 51.104
266757 96.581 -73.956 51.081
266758 96.615 -73.956 51.060
266759 96.650 -73.956 51.041
266760 96.684 -73.956 51.023
266761 96.718 -73.961 51.006
266762 96.749 -73.975 50.990
266763 96.777 -73.987 50.975
266764 96.789 -73.992 50.962
266765 96.814 -74.010 50.951
266766 96.847 -74.027 50.943
266767 96.880 -74.043 50.937
266768 96.912 -74.060 50.932
266769 96.945 -74.076 50.928
266770 96.978 -74.092 50.927
266771 97.011 -74.108 50.929
266772 97.043 -74.124 50.933
266773 97.078 -74.124 50.939
266774 97.113 -74.117 50.950
266775 97.149 -74.110 50.959
266776 97.184 -74.102 50.967
266777 97.219 -74.095 50.974
266778 97.255 -74.088 50.974
266779 97.278 -74.073 50.968
266780 97.301 -74.059 50.959
266781 97.323 -74.045 50.949
266782 97.346 -74.030 50.933
266783 97.369 -74.016 50.919
266784 97.391 -74.010 50.903
266785 97.415 -74.016 50.887
266786 97.439 -74.022 50.870
266787 97.463 -74.029 50.853
266788 97.487 -74.035 50.833
266789 97.510 -74.041 50.811
266790 97.534 -74.048 50.784
266791 97.558 -74.054 50.755
266792 97.581 -74.060 50.723
266793 97.605 -74.067 50.690
266794 97.629 -74.073 50.655
266795 97.659 -74.089 50.614
266796 97.690 -74.104 50.569
266797 97.722 -74.120 50.520
266798 97.753 -74.136 50.471
266799 97.784 -74.151 50.424
266800 97.815 -74.167 50.377
266801 97.847 -74.183 50.330
266802 97.878 -74.198 50.279
266803 97.910 -74.213 50.228
266804 97.941 -74.223 50.176
266805 97.972 -74.234 50.125
266806 98.003 -74.245 50.080
266807 98.033 -74.260 50.036
266808 98.064 -74.274 49.989
266809 98.094 -74.288 49.936
266810 98.125 -74.302 49.876
266811 98.155 -74.317 49.815
266812 98.185 -74.330 49.753
266813 98.215 -74.345 49.694
266814 98.246 -74.359 49.636
266815 98.276 -74.373 49.577
266816 98.306 -74.387 49.518
266817 98.337 -74.402 49.460
266818 98.331 -74.410 49.404
266819 98.309 -74.417 49.348
266820 98.324 -74.436 49.292
266821 98.388 -74.473 49.235
266822 98.453 -74.510 49.176
266823 98.535 -74.512 49.116
266824 98.621 -74.501 49.057
266825 98.663 -74.580 49.001
266826 98.705 -74.664 48.948
266827 98.743 -74.740 48.899
266828 98.754 -74.750 48.850
266829 98.764 -74.760 48.799
266830 98.775 -74.770 48.748
266831 98.785 -74.780 48.696
266832 98.796 -74.791 48.644
266833 98.806 -74.800 48.591
266834 98.816 -74.810 48.537
266835 98.856 -74.791 48.483
266836 98.898 -74.769 48.430
266837 98.940 -74.747 48.376
266838 98.982 -74.725 48.324
266839 99.024 -74.701 48.273
266840 99.068 -74.656 48.222
266841 99.106 -74.624 48.172
266842 99.122 -74.639 48.124
266843 99.139 -74.653 48.076
266844 99.155 -74.667 48.029
266845 99.172 -74.681 47.981
266846 99.189 -74.696 47.935
266847 99.205 -74.709 47.889
266848 99.224 -74.727 47.844
266849 99.243 -74.744 47.799
266850 99.262 -74.760 47.755
266851 99.305 -74.794 47.712
266852 99.343 -74.816 47.669
266853 99.370 -74.823 47.628
266854 99.405 -74.833 47.589
266855 99.443 -74.844 47.551
266856 99.482 -74.855 47.512
266857 99.520 -74.866 47.473
266858 99.559 -74.877 47.435
266859 99.597 -74.888 47.396
266860 99.635 -74.899 47.358
266861 99.669 -74.900 47.321
266862 99.703 -74.900 47.287
266863 99.737 -74.901 47.254
266864 99.772 -74.902 47.221
266865 99.806 -74.903 47.190
266866 99.840 -74.903 47.160
266867 99.874 -74.904 47.131
266868 99.908 -74.905 47.104
266869 99.942 -74.906 47.078
266870 99.976 -74.907 47.051
266871 100.010 -74.908 47.024
266872 100.045 -74.909 46.998
266873 100.078 -74.925 46.973
266874 100.111 -74.963 46.948
266875 100.144 -75.001 46.926
266876 100.177 -75.038 46.904
266877 100.210 -75.076 46.882
266878 100.241 -75.109 46.859
266879 100.267 -75.123 46.834
266880 100.295 -75.138 46.809
266881 100.323 -75.154 46.782
266882 100.351 -75.170 46.756
266883 100.379 -75.186 46.730
266884 100.407 -75.202 46.703
266885 100.425 -75.219 46.677
266886 100.444 -75.235 46.649
266887 100.463 -75.251 46.621
266888 100.481 -75.267 46.594
266889 100.500 -75.284 46.568
266890 100.519 -75.300 46.540
266891 100.538 -75.317 46.512
266892 100.556 -75.333 46.484
266893 100.575 -75.349 46.454
266894 100.594 -75.365 46.425
266895 100.612 -75.382 46.397
266896 100.631 -75.398 46.369
266897 100.658 -75.408 46.342
266898 100.685 -75.419 46.313
266899 100.713 -75.427 46.286
266900 100.740 -75.421 46.258
266901 100.768 -75.416 46.229
266902 100.795 -75.411 46.203
266903 100.823 -75.405 46.177
266904 100.863 -75.403 46.149
266905 100.907 -75.400 46.120
266906 100.952 -75.396 46.091
266907 100.997 -75.393 46.063
266908 101.042 -75.404 46.036
266909 101.088 -75.436 46.009
266910 101.125 -75.470 45.983
266911 101.156 -75.504 45.955
266912 101.177 -75.525 45.925
266913 101.192 -75.540 45.893
266914 101.199 -75.549 45.860
266915 101.213 -75.546 45.826
266916 101.232 -75.535 45.793
266917 101.251 -75.524 45.760
266918 101.270 -75.513 45.726
266919 101.287 -75.508 45.693
266920 101.301 -75.518 45.660
266921 101.315 -75.527 45.628
266922 101.335 -75.533 45.596
266923 101.366 -75.536 45.566
266924 101.402 -75.556 45.536
266925 101.442 -75.589 45.506
266926 101.482 -75.621 45.477
266927 101.527 -75.656 45.448
266928 101.583 -75.694 45.419
266929 101.640 -75.732 45.391
266930 101.696 -75.771 45.363
266931 101.747 -75.804 45.336
266932 101.754 -75.807 45.309
266933 101.762 -75.810 45.283
266934 101.770 -75.812 45.256
266935 101.777 -75.815 45.231
266936 101.785 -75.818 45.205
266937 101.793 -75.821 45.179
266938 101.803 -75.827 45.155
266939 101.825 -75.847 45.130
266940 101.846 -75.865 45.105
266941 101.865 -75.871 45.079
266942 101.883 -75.872 45.056
266943 101.900 -75.872 45.032
266944 101.917 -75.872 45.008
266945 101.935 -75.873 44.984
266946 101.952 -75.874 44.960
267293 92.290 -73.383 54.722
267294 92.320 -73.394 54.714
267295 92.351 -73.404 54.706
267296 92.382 -73.415 54.698
267297 92.412 -73.425 54.691
267298 92.443 -73.435 54.683
267299 92.474 -73.446 54.676
267300 92.504 -73.456 54.668
267301 92.535 -73.466 54.661
267302 92.566 -73.477 54.655
267303 92.596 -73.487 54.650
267304 92.627 -73.497 54.644
267305 92.658 -73.508 54.639
267306 92.688 -73.518 54.632
267307 92.719 -73.528 54.625
267308 92.752 -73.539 54.620
267309 92.795 -73.550 54.616
267310 92.838 -73.561 54.612
267311 92.880 -73.571 54.611
267312 92.923 -73.583 54.611
267313 92.949 -73.577 54.612
267314 92.971 -73.568 54.614
267315 92.993 -73.560 54.618
267316 93.016 -73.552 54.622
267317 93.038 -73.543 54.625
267318 93.060 -73.534 54.630
267319 93.089 -73.541 54.633
267320 93.118 -73.547 54.636
267321 93.142 -73.546 54.640
267322 93.169 -73.549 54.640
267323 93.202 -73.558 54.640
267324 93.233 -73.567 54.638
267325 93.266 -73.576 54.634
267326 93.298 -73.585 54.631
267327 93.330 -73.594 54.629
267328 93.362 -73.603 54.624
267329 93.394 -73.612 54.619
267330 93.426 -73.622 54.612
267331 93.455 -73.626 54.605
267332 93.477 -73.629 54.597
267333 93.499 -73.632 54.588
267334 93.521 -73.636 54.578
267335 93.543 -73.639 54.570
267336 93.564 -73.642 54.561
267337 93.586 -73.645 54.553
267338 93.608 -73.648 54.545
267339 93.630 -73.651 54.538
267340 93.651 -73.655 54.530
267341 93.673 -73.656 54.520
267342 93.695 -73.652 54.505
267343 93.717 -73.648 54.485
267344 93.738 -73.646 54.464
267345 93.759 -73.644 54.444
267346 93.781 -73.640 54.423
267347 93.802 -73.638 54.405
267348 93.823 -73.636 54.388
267349 93.844 -73.633 54.371
267350 93.865 -73.631 54.353
267351 93.887 -73.628 54.335
267352 93.908 -73.625 54.318
267353 93.929 -73.623 54.300
267354 93.950 -73.620 54.283
267355 93.971 -73.618 54.265
267356 93.992 -73.615 54.245
267357 94.013 -73.612 54.225
267358 94.035 -73.610 54.203
267359 94.063 -73.608 54.183
267360 94.100 -73.607 54.163
267361 94.128 -73.611 54.143
267362 94.157 -73.614 54.123
267363 94.182 -73.625 54.099
267364 94.206 -73.635 54.074
267365 94.229 -73.646 54.048
267366 94.254 -73.657 54.023
267367 94.278 -73.668 53.996
267368 94.301 -73.678 53.969
267369 94.326 -73.689 53.942
267370 94.350 -73.700 53.912
267371 94.373 -73.711 53.884
267372 94.398 -73.722 53.855
267373 94.421 -73.732 53.825
267374 94.445 -73.743 53.795
267375 94.470 -73.754 53.765
267376 94.493 -73.765 53.735
267377 94.518 -73.775 53.703
267378 94.542 -73.786 53.672
267379 94.566 -73.796 53.642
267380 94.592 -73.803 53.614
267381 94.617 -73.811 53.586
267382 94.631 -73.813 53.559
267383 94.640 -73.807 53.533
267384 94.652 -73.794 53.510
267385 94.690 -73.781 53.488
267386 94.708 -73.779 53.468
267387 94.726 -73.776 53.451
267388 94.744 -73.774 53.436
267389 94.762 -73.772 53.422
267390 94.780 -73.769 53.404
267391 94.796 -73.769 53.380
267392 94.811 -73.769 53.352
267393 94.825 -73.769 53.319
267394 94.849 -73.772 53.283
267395 94.884 -73.781 53.241
267396 94.918 -73.789 53.199
267397 94.952 -73.798 53.156
267398 94.986 -73.807 53.111
267399 95.021 -73.815 53.068
267400 95.055 -73.823 53.024
267401 95.089 -73.832 52.982
267402 95.116 -73.807 52.941
267403 95.139 -73.763 52.901
267404 95.154 -73.735 52.862
267405 95.161 -73.724 52.823
267406 95.168 -73.713 52.784
267407 95.176 -73.701 52.748
267408 95.199 -73.700 52.713
267409 95.231 -73.703 52.678
267410 95.264 -73.707 52.645
267411 95.296 -73.711 52.610
267412 95.328 -73.716 52.575
267413 95.360 -73.730 52.540
267414 95.392 -73.744 52.503
267415 95.424 -73.759 52.465
267416 95.456 -73.773 52.429
267417 95.488 -73.787 52.395
267418 95.520 -73.801 52.359
267419 95.552 -73.816 52.324
267420 95.581 -73.816 52.290
267421 95.607 -73.803 52.256
267422 95.634 -73.791 52.224
267423 95.660 -73.780 52.193
267424 95.693 -73.774 52.163
267425 95.727 -73.770 52.132
267426 95.761 -73.766 52.101
267427 95.796 -73.762 52.068
267428 95.830 -73.757 52.034
267429 95.865 -73.753 52.000
267430 95.899 -73.749 51.967
267431 95.933 -73.745 51.934
267432 95.960 -73.740 51.902
267433 95.991 -73.734 51.869
267434 96.055 -73.726 51.836
267435 96.118 -73.718 51.802
267436 96.126 -73.719 51.768
267437 96.110 -73.722 51.734
267438 96.093 -73.725 51.699
267439 96.099 -73.734 51.665
267440 96.107 -73.744 51.630
267441 96.115 -73.754 51.596
267442 96.138 -73.778 51.561
267443 96.164 -73.805 51.528
267444 96.190 -73.832 51.495
267445 96.219 -73.856 51.462
267446 96.248 -73.865 51.428
267447 96.270 -73.876 51.395
267448 96.296 -73.884 51.362
267449 96.323 -73.891 51.330
267450 96.349 -73.897 51.300
267451 96.376 -73.904 51.269
267452 96.403 -73.911 51.239
267453 96.430 -73.918 51.212
267454 96.456 -73.925 51.185
267455 96.482 -73.932 51.159
267456 96.513 -73.935 51.133
267457 96.548 -73.935 51.108
267458 96.583 -73.935 51.084
267459 96.617 -73.935 51.063
267460 96.652 -73.935 51.043
267461 96.686 -73.935 51.024
267462 96.721 -73.935 51.007
267463 96.755 -73.936 50.991
267464 96.783 -73.948 50.976
267465 96.806 -73.962 50.962
267466 96.838 -73.978 50.950
267467 96.871 -73.995 50.941
267468 96.904 -74.010 50.934
267469 96.937 -74.026 50.927
267470 96.969 -74.043 50.923
267471 97.002 -74.059 50.921
267472 97.035 -74.075 50.923
267473 97.068 -74.092 50.927
267474 97.101 -74.097 50.932
267475 97.137 -74.089 50.942
267476 97.172 -74.082 50.950
267477 97.208 -74.075 50.958
267478 97.232 -74.061 50.965
267479 97.254 -74.047 50.965
267480 97.277 -74.032 50.958
267481 97.300 -74.018 50.948
267482 97.322 -74.004 50.936
267483 97.345 -73.989 50.919
267484 97.368 -73.980 50.904
267485 97.392 -73.987 50.889
267486 97.416 -73.993 50.875
267487 97.439 -73.999 50.857
267488 97.463 -74.006 50.841
267489 97.487 -74.012 50.823
267490 97.510 -74.018 50.804
267491 97.534 -74.025 50.780
267492 97.558 -74.031 50.751
267493 97.581 -74.037 50.721
267494 97.601 -74.043 50.690
267495 97.633 -74.048 50.653
267496 97.664 -74.061 50.611
267497 97.692 -74.077 50.564
267498 97.724 -74.093 50.514
267499 97.755 -74.109 50.463
267500 97.787 -74.124 50.416
267501 97.818 -74.140 50.370
267502 97.850 -74.156 50.322
267503 97.881 -74.171 50.270
267504 97.912 -74.187 50.218
267505 97.943 -74.198 50.167
267506 97.974 -74.208 50.120
267507 98.006 -74.219 50.077
267508 98.037 -74.230 50.032
267509 98.068 -74.242 49.983
267510 98.098 -74.257 49.929
267511 98.129 -74.270 49.866
267512 98.159 -74.285 49.804
267513 98.189 -74.299 49.742
267514 98.219 -74.313 49.681
267515 98.250 -74.327 49.624
267516 98.280 -74.342 49.566
267517 98.310 -74.355 49.508
267518 98.341 -74.370 49.450
267519 98.331 -74.378 49.392
267520 98.310 -74.384 49.336
267521 98.360 -74.416 49.279
267522 98.424 -74.453 49.221
267523 98.489 -74.490 49.162
267524 98.563 -74.508 49.104
267525 98.639 -74.468 49.046
267526 98.672 -74.453 48.989
267527 98.706 -74.443 48.936
267528 98.740 -74.678 48.886
267529 98.770 -74.690 48.836
267530 98.782 -74.700 48.785
267531 98.793 -74.709 48.735
267532 98.803 -74.720 48.683
267533 98.814 -74.730 48.631
267534 98.825 -74.740 48.579
267535 98.835 -74.750 48.524
267536 98.871 -74.734 48.470
267537 98.913 -74.712 48.417
267538 98.955 -74.690 48.365
267539 98.997 -74.667 48.313
267540 99.040 -74.645 48.262
267541 99.081 -74.623 48.212
267542 99.121 -74.599 48.162
267543 99.139 -74.614 48.114
267544 99.158 -74.631 48.065
267545 99.177 -74.648 48.017
267546 99.196 -74.665 47.970
267547 99.215 -74.681 47.924
267548 99.234 -74.699 47.878
267549 99.253 -74.715 47.833
267550 99.272 -74.733 47.788
267551 99.293 -74.751 47.745
267552 99.337 -74.783 47.701
267553 99.365 -74.790 47.659
267554 99.393 -74.797 47.619
267555 99.421 -74.803 47.581
267556 99.454 -74.812 47.544
267557 99.492 -74.824 47.507
267558 99.531 -74.834 47.468
267559 99.569 -74.846 47.429
267560 99.607 -74.857 47.391
267561 99.645 -74.868 47.352
267562 99.681 -74.872 47.314
267563 99.715 -74.874 47.280
267564 99.750 -74.875 47.247
267565 99.784 -74.875 47.215
267566 99.818 -74.876 47.184
267567 99.852 -74.877 47.154
267568 99.886 -74.878 47.124
267569 99.920 -74.878 47.096
267570 99.955 -74.879 47.071
267571 99.988 -74.880 47.043
267572 100.022 -74.881 47.015
267573 100.056 -74.895 46.989
267574 100.089 -74.910 46.964
267575 100.123 -74.937 46.938
267576 100.155 -74.975 46.916
267577 100.189 -75.013 46.894
267578 100.221 -75.051 46.872
267579 100.251 -75.079 46.848
267580 100.276 -75.091 46.823
267581 100.301 -75.104 46.798
267582 100.330 -75.120 46.771
267583 100.357 -75.135 46.744
267584 100.385 -75.151 46.719
267585 100.413 -75.167 46.693
267586 100.437 -75.184 46.666
267587 100.456 -75.200 46.638
267588 100.475 -75.217 46.611
267589 100.493 -75.232 46.584
267590 100.512 -75.249 46.559
267591 100.531 -75.265 46.530
267592 100.549 -75.282 46.502
267593 100.568 -75.298 46.473
267594 100.586 -75.314 46.444
267595 100.606 -75.330 46.415
267596 100.624 -75.347 46.388
267597 100.648 -75.359 46.360
267598 100.676 -75.370 46.332
267599 100.703 -75.380 46.303
267600 100.731 -75.387 46.274
267601 100.758 -75.383 46.246
267602 100.785 -75.377 46.219
267603 100.813 -75.372 46.193
267604 100.845 -75.367 46.166
267605 100.888 -75.365 46.138
267606 100.931 -75.362 46.109
267607 100.974 -75.360 46.081
267608 101.017 -75.357 46.053
267609 101.061 -75.371 46.026
267610 101.093 -75.405 46.000
267611 101.125 -75.440 45.972
267612 101.156 -75.473 45.942
267613 101.177 -75.494 45.910
267614 101.198 -75.515 45.878
267615 101.214 -75.533 45.845
267616 101.227 -75.533 45.812
267617 101.245 -75.523 45.779
267618 101.264 -75.512 45.746
267619 101.283 -75.501 45.713
267620 101.302 -75.490 45.679
267621 101.321 -75.480 45.645
267622 101.335 -75.488 45.612
267623 101.354 -75.495 45.581
267624 101.388 -75.508 45.551
267625 101.428 -75.540 45.521
267626 101.468 -75.574 45.492
267627 101.508 -75.606 45.462
267628 101.552 -75.640 45.433
267629 101.608 -75.678 45.404
267630 101.665 -75.716 45.376
267631 101.722 -75.755 45.348
267632 101.751 -75.773 45.320
267633 101.758 -75.776 45.294
267634 101.766 -75.778 45.267
267635 101.773 -75.781 45.241
267636 101.781 -75.784 45.215
267637 101.788 -75.787 45.190
267638 101.803 -75.797 45.164
267639 101.825 -75.817 45.139
267640 101.845 -75.830 45.115
267641 101.863 -75.831 45.089
267642 101.880 -75.831 45.063
267643 101.897 -75.831 45.040
267644 101.915 -75.832 45.016
267645 101.933 -75.833 44.993
267646 101.950 -75.834 44.968
267994 92.286 -73.355 54.730
267995 92.316 -73.365 54.722
267996 92.347 -73.375 54.713
267997 92.377 -73.385 54.704
267998 92.408 -73.396 54.696
267999 92.439 -73.406 54.688
268000 92.470 -73.416 54.680
268001 92.500 -73.427 54.672
268002 92.531 -73.437 54.664
268003 92.562 -73.447 54.658
268004 92.592 -73.458 54.653
268005 92.623 -73.468 54.647
268006 92.653 -73.478 54.640
268007 92.684 -73.489 54.632
268008 92.715 -73.499 54.625
268009 92.748 -73.509 54.619
268010 92.791 -73.521 54.614
268011 92.835 -73.531 54.610
268012 92.877 -73.543 54.608
268013 92.920 -73.553 54.607
268014 92.960 -73.562 54.607
268015 92.982 -73.553 54.609
268016 93.004 -73.544 54.612
268017 93.027 -73.536 54.615
268018 93.049 -73.528 54.620
268019 93.071 -73.519 54.624
268020 93.094 -73.511 54.627
268021 93.118 -73.506 54.631
268022 93.144 -73.508 54.633
268023 93.176 -73.517 54.635
268024 93.208 -73.526 54.634
268025 93.240 -73.535 54.631
268026 93.272 -73.544 54.629
268027 93.305 -73.555 54.625
268028 93.338 -73.566 54.621
268029 93.371 -73.578 54.617
268030 93.404 -73.589 54.611
268031 93.436 -73.600 54.604
268032 93.462 -73.606 54.597
268033 93.484 -73.609 54.588
268034 93.506 -73.612 54.577
268035 93.528 -73.616 54.565
268036 93.549 -73.619 54.555
268037 93.571 -73.622 54.544
268038 93.592 -73.625 54.534
268039 93.615 -73.628 54.525
268040 93.636 -73.631 54.517
268041 93.657 -73.633 54.509
268042 93.677 -73.631 54.500
268043 93.697 -73.628 54.486
268044 93.717 -73.626 54.468
268045 93.737 -73.624 54.448
268046 93.757 -73.622 54.427
268047 93.777 -73.619 54.407
268048 93.798 -73.616 54.388
268049 93.819 -73.614 54.370
268050 93.840 -73.612 54.351
268051 93.861 -73.609 54.332
268052 93.883 -73.606 54.314
268053 93.904 -73.604 54.295
268054 93.925 -73.601 54.276
268055 93.946 -73.599 54.257
268056 93.968 -73.597 54.239
268057 93.989 -73.593 54.219
268058 94.010 -73.591 54.199
268059 94.031 -73.589 54.179
268060 94.059 -73.587 54.158
268061 94.090 -73.588 54.136
268062 94.118 -73.594 54.114
268063 94.142 -73.605 54.093
268064 94.166 -73.615 54.068
268065 94.190 -73.627 54.043
268066 94.214 -73.637 54.017
268067 94.238 -73.648 53.991
268068 94.262 -73.659 53.964
268069 94.286 -73.670 53.937
268070 94.310 -73.681 53.909
268071 94.334 -73.691 53.881
268072 94.359 -73.700 53.852
268073 94.385 -73.707 53.823
268074 94.410 -73.715 53.793
268075 94.436 -73.722 53.764
268076 94.462 -73.729 53.734
268077 94.488 -73.737 53.704
268078 94.513 -73.744 53.673
268079 94.539 -73.752 53.643
268080 94.565 -73.759 53.614
268081 94.591 -73.766 53.586
268082 94.616 -73.774 53.559
268083 94.642 -73.781 53.533
268084 94.649 -73.780 53.508
268085 94.656 -73.779 53.485
268086 94.663 -73.778 53.464
268087 94.689 -73.774 53.444
268088 94.733 -73.762 53.425
268089 94.769 -73.756 53.409
268090 94.784 -73.756 53.394
268091 94.799 -73.756 53.377
268092 94.814 -73.756 53.353
268093 94.828 -73.756 53.324
268094 94.842 -73.756 53.290
268095 94.857 -73.756 53.252
268096 94.871 -73.756 53.210
268097 94.902 -73.763 53.168
268098 94.936 -73.771 53.126
268099 94.970 -73.780 53.083
268100 95.004 -73.788 53.041
268101 95.039 -73.797 53.000
268102 95.073 -73.805 52.961
268103 95.105 -73.802 52.921
268104 95.128 -73.756 52.882
268105 95.150 -73.712 52.844
268106 95.162 -73.691 52.807
268107 95.169 -73.680 52.770
268108 95.177 -73.669 52.736
268109 95.206 -73.671 52.702
268110 95.239 -73.675 52.669
268111 95.271 -73.678 52.636
268112 95.303 -73.688 52.601
268113 95.335 -73.702 52.566
268114 95.367 -73.716 52.531
268115 95.399 -73.731 52.495
268116 95.430 -73.745 52.458
268117 95.463 -73.760 52.423
268118 95.494 -73.774 52.389
268119 95.527 -73.788 52.353
268120 95.558 -73.802 52.319
268121 95.585 -73.791 52.286
268122 95.612 -73.780 52.253
268123 95.638 -73.768 52.222
268124 95.665 -73.756 52.192
268125 95.696 -73.749 52.163
268126 95.730 -73.745 52.133
268127 95.764 -73.741 52.103
268128 95.799 -73.737 52.070
268129 95.833 -73.732 52.037
268130 95.868 -73.728 52.003
268131 95.902 -73.724 51.970
268132 95.936 -73.720 51.937
268133 95.970 -73.716 51.905
268134 96.022 -73.727 51.872
268135 96.094 -73.754 51.840
268136 96.111 -73.739 51.806
268137 96.128 -73.725 51.772
268138 96.130 -73.717 51.738
268139 96.115 -73.721 51.703
268140 96.123 -73.730 51.668
268141 96.131 -73.740 51.635
268142 96.138 -73.750 51.601
268143 96.147 -73.760 51.566
268144 96.169 -73.783 51.533
268145 96.195 -73.810 51.500
268146 96.240 -73.819 51.467
268147 96.262 -73.831 51.433
268148 96.285 -73.843 51.400
268149 96.306 -73.854 51.368
268150 96.328 -73.866 51.336
268151 96.355 -73.873 51.306
268152 96.381 -73.880 51.275
268153 96.408 -73.887 51.246
268154 96.435 -73.894 51.217
268155 96.461 -73.901 51.190
268156 96.487 -73.907 51.164
268157 96.516 -73.913 51.137
268158 96.550 -73.913 51.112
268159 96.585 -73.913 51.087
268160 96.619 -73.913 51.065
268161 96.654 -73.913 51.044
268162 96.688 -73.913 51.025
268163 96.723 -73.913 51.008
268164 96.757 -73.913 50.992
268165 96.791 -73.914 50.977
268166 96.825 -73.930 50.962
268167 96.861 -73.945 50.949
268168 96.895 -73.961 50.940
268169 96.928 -73.978 50.931
268170 96.961 -73.994 50.925
268171 96.994 -74.010 50.920
268172 97.026 -74.026 50.917
268173 97.058 -74.042 50.918
268174 97.091 -74.059 50.920
268175 97.125 -74.070 50.925
268176 97.160 -74.062 50.933
268177 97.186 -74.049 50.940
268178 97.209 -74.035 50.946
268179 97.231 -74.020 50.953
268180 97.254 -74.006 50.953
268181 97.277 -73.991 50.948
268182 97.299 -73.977 50.938
268183 97.321 -73.962 50.925
268184 97.345 -73.951 50.909
268185 97.368 -73.957 50.894
268186 97.392 -73.964 50.880
268187 97.416 -73.970 50.867
268188 97.439 -73.976 50.850
268189 97.463 -73.983 50.834
268190 97.487 -73.989 50.817
268191 97.510 -73.995 50.799
268192 97.533 -74.002 50.776
268193 97.552 -74.007 50.749
268194 97.572 -74.013 50.719
268195 97.599 -74.019 50.686
268196 97.640 -74.023 50.648
268197 97.676 -74.032 50.603
268198 97.703 -74.048 50.554
268199 97.729 -74.066 50.503
268200 97.758 -74.082 50.452
268201 97.789 -74.098 50.403
268202 97.821 -74.114 50.357
268203 97.852 -74.129 50.310
268204 97.883 -74.145 50.259
268205 97.915 -74.161 50.209
268206 97.946 -74.172 50.162
268207 97.977 -74.183 50.117
268208 98.008 -74.193 50.074
268209 98.039 -74.204 50.028
268210 98.071 -74.214 49.978
268211 98.102 -74.225 49.922
268212 98.132 -74.239 49.859
268213 98.163 -74.253 49.795
268214 98.193 -74.267 49.732
268215 98.223 -74.282 49.669
268216 98.254 -74.295 49.611
268217 98.284 -74.310 49.553
268218 98.314 -74.324 49.496
268219 98.345 -74.339 49.439
268220 98.332 -74.346 49.382
268221 98.331 -74.360 49.325
268222 98.395 -74.397 49.267
268223 98.459 -74.434 49.210
268224 98.524 -74.471 49.151
268225 98.571 -74.487 49.092
268226 98.623 -74.428 49.034
268227 98.655 -74.415 48.976
268228 98.707 -74.433 48.923
268229 98.736 -74.613 48.872
268230 98.765 -74.632 48.821
268231 98.795 -74.640 48.771
268232 98.811 -74.649 48.720
268233 98.822 -74.659 48.669
268234 98.833 -74.669 48.618
268235 98.843 -74.679 48.566
268236 98.854 -74.690 48.511
268237 98.886 -74.677 48.457
268238 98.929 -74.654 48.404
268239 98.971 -74.632 48.352
268240 99.013 -74.609 48.301
268241 99.055 -74.587 48.251
268242 99.096 -74.567 48.201
268243 99.129 -74.571 48.151
268244 99.153 -74.592 48.102
268245 99.177 -74.612 48.054
268246 99.201 -74.634 48.006
268247 99.225 -74.654 47.958
268248 99.244 -74.671 47.912
268249 99.264 -74.687 47.867
268250 99.283 -74.705 47.822
268251 99.301 -74.722 47.777
268252 99.329 -74.744 47.734
268253 99.360 -74.756 47.691
268254 99.387 -74.763 47.650
268255 99.416 -74.770 47.611
268256 99.443 -74.777 47.573
268257 99.472 -74.784 47.537
268258 99.502 -74.792 47.500
268259 99.541 -74.803 47.462
268260 99.579 -74.815 47.423
268261 99.618 -74.825 47.384
268262 99.656 -74.837 47.346
268263 99.693 -74.846 47.308
268264 99.727 -74.847 47.274
268265 99.762 -74.847 47.241
268266 99.796 -74.848 47.209
268267 99.830 -74.849 47.178
268268 99.864 -74.850 47.147
268269 99.898 -74.850 47.117
268270 99.932 -74.851 47.090
268271 99.967 -74.852 47.063
268272 100.000 -74.853 47.035
268273 100.034 -74.866 47.007
268274 100.067 -74.881 46.980
268275 100.101 -74.897 46.955
268276 100.134 -74.912 46.929
268277 100.167 -74.950 46.906
268278 100.200 -74.988 46.885
268279 100.233 -75.026 46.863
268280 100.261 -75.048 46.838
268281 100.286 -75.060 46.813
268282 100.311 -75.073 46.787
268283 100.336 -75.085 46.760
268284 100.364 -75.101 46.734
268285 100.392 -75.117 46.708
268286 100.420 -75.133 46.682
268287 100.448 -75.149 46.655
268288 100.468 -75.165 46.628
268289 100.487 -75.182 46.601
268290 100.505 -75.198 46.574
268291 100.524 -75.214 46.547
268292 100.543 -75.230 46.520
268293 100.561 -75.247 46.492
268294 100.580 -75.263 46.464
268295 100.599 -75.279 46.435
268296 100.617 -75.295 46.405
268297 100.639 -75.310 46.377
268298 100.666 -75.320 46.350
268299 100.693 -75.331 46.322
268300 100.721 -75.342 46.294
268301 100.748 -75.348 46.265
268302 100.776 -75.344 46.237
268303 100.803 -75.339 46.209
268304 100.831 -75.333 46.183
268305 100.870 -75.330 46.156
268306 100.913 -75.327 46.127
268307 100.956 -75.325 46.099
268308 100.998 -75.322 46.071
268309 101.037 -75.324 46.043
268310 101.061 -75.342 46.015
268311 101.094 -75.376 45.988
268312 101.126 -75.411 45.959
268313 101.155 -75.442 45.928
268314 101.176 -75.463 45.896
268315 101.197 -75.484 45.862
268316 101.218 -75.506 45.830
268317 101.238 -75.516 45.798
268318 101.258 -75.508 45.765
268319 101.277 -75.499 45.732
268320 101.296 -75.489 45.698
268321 101.315 -75.478 45.664
268322 101.334 -75.467 45.630
268323 101.353 -75.457 45.597
268324 101.375 -75.460 45.566
268325 101.415 -75.493 45.536
268326 101.455 -75.525 45.506
268327 101.494 -75.558 45.477
268328 101.534 -75.590 45.447
268329 101.577 -75.624 45.417
268330 101.633 -75.662 45.388
268331 101.690 -75.701 45.360
268332 101.747 -75.739 45.332
268333 101.754 -75.742 45.304
268334 101.762 -75.744 45.278
268335 101.769 -75.747 45.251
268336 101.777 -75.750 45.225
268337 101.785 -75.753 45.199
268338 101.803 -75.768 45.174
268339 101.825 -75.787 45.148
268340 101.842 -75.790 45.123
268341 101.860 -75.790 45.098
268342 101.878 -75.791 45.073
268343 101.895 -75.791 45.048
268344 101.912 -75.792 45.024
268345 101.930 -75.793 45.000
268346 101.947 -75.793 44.976
268347 101.965 -75.793 44.952
268695 92.281 -73.325 54.738
268696 92.312 -73.336 54.729
268697 92.343 -73.346 54.720
268698 92.373 -73.356 54.711
268699 92.404 -73.367 54.701
268700 92.435 -73.377 54.692
268701 92.465 -73.387 54.684
268702 92.496 -73.398 54.675
268703 92.526 -73.408 54.667
268704 92.557 -73.418 54.661
268705 92.588 -73.428 54.655
268706 92.619 -73.439 54.649
268707 92.649 -73.449 54.641
268708 92.680 -73.459 54.631
268709 92.711 -73.470 54.624
268710 92.745 -73.481 54.617
268711 92.788 -73.491 54.612
268712 92.831 -73.503 54.608
268713 92.874 -73.513 54.605
268714 92.917 -73.524 54.603
268715 92.960 -73.535 54.602
268716 92.993 -73.537 54.604
268717 93.015 -73.529 54.606
268718 93.038 -73.521 54.610
268719 93.060 -73.512 54.613
268720 93.082 -73.503 54.617
268721 93.105 -73.495 54.620
268722 93.127 -73.487 54.624
268723 93.154 -73.484 54.626
268724 93.186 -73.496 54.628
268725 93.219 -73.507 54.626
268726 93.252 -73.518 54.624
268727 93.285 -73.530 54.621
268728 93.318 -73.541 54.617
268729 93.351 -73.553 54.613
268730 93.384 -73.564 54.609
268731 93.417 -73.575 54.603
268732 93.450 -73.587 54.596
268733 93.472 -73.590 54.587
268734 93.493 -73.593 54.577
268735 93.515 -73.596 54.565
268736 93.537 -73.600 54.552
268737 93.559 -73.603 54.540
268738 93.580 -73.606 54.529
268739 93.602 -73.608 54.517
268740 93.622 -73.610 54.507
268741 93.643 -73.611 54.497
268742 93.663 -73.609 54.488
268743 93.683 -73.606 54.478
268744 93.703 -73.604 54.465
268745 93.723 -73.602 54.448
268746 93.744 -73.600 54.429
268747 93.764 -73.597 54.409
268748 93.783 -73.595 54.390
268749 93.803 -73.593 54.371
268750 93.824 -73.590 54.352
268751 93.844 -73.588 54.332
268752 93.863 -73.586 54.312
268753 93.884 -73.584 54.292
268754 93.904 -73.581 54.273
268755 93.924 -73.579 54.254
268756 93.944 -73.577 54.233
268757 93.964 -73.575 54.214
268758 93.985 -73.572 54.195
268759 94.006 -73.570 54.175
268760 94.028 -73.567 54.156
268761 94.053 -73.565 54.134
268762 94.078 -73.575 54.110
268763 94.102 -73.586 54.086
268764 94.126 -73.597 54.063
268765 94.152 -73.604 54.038
268766 94.178 -73.611 54.012
268767 94.203 -73.618 53.986
268768 94.229 -73.626 53.959
268769 94.255 -73.634 53.933
268770 94.281 -73.640 53.906
268771 94.307 -73.648 53.878
268772 94.332 -73.656 53.850
268773 94.358 -73.663 53.822
268774 94.384 -73.670 53.793
268775 94.409 -73.678 53.763
268776 94.435 -73.685 53.735
268777 94.461 -73.692 53.705
268778 94.486 -73.700 53.675
268779 94.512 -73.707 53.646
268780 94.538 -73.715 53.616
268781 94.564 -73.722 53.588
268782 94.590 -73.729 53.560
268783 94.615 -73.737 53.534
268784 94.641 -73.744 53.509
268785 94.660 -73.748 53.484
268786 94.667 -73.747 53.461
268787 94.674 -73.747 53.440
268788 94.683 -73.746 53.419
268789 94.711 -73.745 53.400
268790 94.754 -73.733 53.381
268791 94.790 -73.728 53.363
268792 94.805 -73.741 53.344
268793 94.829 -73.744 53.319
268794 94.845 -73.744 53.290
268795 94.859 -73.744 53.257
268796 94.874 -73.743 53.219
268797 94.888 -73.743 53.179
268798 94.903 -73.743 53.139
268799 94.920 -73.744 53.099
268800 94.955 -73.753 53.058
268801 94.989 -73.761 53.018
268802 95.023 -73.770 52.979
268803 95.057 -73.778 52.941
268804 95.092 -73.787 52.902
268805 95.116 -73.751 52.864
268806 95.139 -73.706 52.827
268807 95.162 -73.661 52.792
268808 95.170 -73.647 52.757
268809 95.181 -73.638 52.725
268810 95.213 -73.642 52.692
268811 95.246 -73.646 52.659
268812 95.277 -73.660 52.627
268813 95.310 -73.675 52.592
268814 95.341 -73.689 52.557
268815 95.373 -73.703 52.523
268816 95.405 -73.717 52.488
268817 95.437 -73.731 52.451
268818 95.469 -73.746 52.416
268819 95.501 -73.760 52.382
268820 95.533 -73.774 52.347
268821 95.563 -73.780 52.314
268822 95.590 -73.768 52.282
268823 95.616 -73.756 52.251
268824 95.642 -73.744 52.220
268825 95.669 -73.732 52.192
268826 95.699 -73.724 52.164
268827 95.733 -73.720 52.135
268828 95.767 -73.716 52.104
268829 95.802 -73.712 52.074
268830 95.836 -73.708 52.041
268831 95.871 -73.703 52.007
268832 95.905 -73.700 51.974
268833 95.943 -73.702 51.941
268834 96.008 -73.750 51.908
268835 96.060 -73.783 51.876
268836 96.077 -73.769 51.843
268837 96.094 -73.754 51.809
268838 96.111 -73.740 51.775
268839 96.128 -73.725 51.741
268840 96.138 -73.716 51.707
268841 96.147 -73.726 51.672
268842 96.154 -73.735 51.638
268843 96.162 -73.745 51.605
268844 96.170 -73.755 51.572
268845 96.178 -73.765 51.538
268846 96.201 -73.787 51.505
268847 96.255 -73.785 51.472
268848 96.277 -73.797 51.438
268849 96.299 -73.809 51.406
268850 96.321 -73.820 51.373
268851 96.343 -73.832 51.342
268852 96.365 -73.844 51.311
268853 96.387 -73.856 51.281
268854 96.413 -73.863 51.252
268855 96.440 -73.869 51.223
268856 96.466 -73.876 51.195
268857 96.493 -73.883 51.168
268858 96.519 -73.890 51.142
268859 96.552 -73.891 51.115
268860 96.587 -73.891 51.090
268861 96.621 -73.891 51.067
268862 96.656 -73.891 51.046
268863 96.690 -73.891 51.027
268864 96.725 -73.891 51.010
268865 96.758 -73.888 50.993
268866 96.791 -73.890 50.977
268867 96.820 -73.903 50.962
268868 96.854 -73.918 50.949
268869 96.890 -73.933 50.939
268870 96.926 -73.949 50.930
268871 96.961 -73.965 50.923
268872 96.997 -73.981 50.918
268873 97.032 -73.996 50.914
268874 97.068 -74.012 50.913
268875 97.103 -74.028 50.914
268876 97.138 -74.036 50.917
268877 97.163 -74.023 50.924
268878 97.185 -74.008 50.930
268879 97.208 -73.994 50.935
268880 97.230 -73.979 50.939
268881 97.253 -73.965 50.939
268882 97.276 -73.950 50.936
268883 97.298 -73.935 50.928
268884 97.321 -73.922 50.915
268885 97.345 -73.929 50.899
268886 97.368 -73.935 50.886
268887 97.392 -73.941 50.874
268888 97.416 -73.948 50.862
268889 97.439 -73.954 50.846
268890 97.463 -73.960 50.831
268891 97.485 -73.966 50.815
268892 97.504 -73.972 50.797
268893 97.523 -73.977 50.775
268894 97.543 -73.983 50.747
268895 97.567 -73.988 50.716
268896 97.606 -73.994 50.681
268897 97.646 -73.999 50.641
268898 97.686 -74.004 50.594
268899 97.714 -74.020 50.543
268900 97.740 -74.036 50.491
268901 97.767 -74.054 50.441
268902 97.793 -74.070 50.392
268903 97.823 -74.086 50.345
268904 97.855 -74.102 50.299
268905 97.886 -74.118 50.249
268906 97.918 -74.133 50.203
268907 97.949 -74.145 50.158
268908 97.980 -74.157 50.115
268909 98.011 -74.167 50.071
268910 98.042 -74.178 50.024
268911 98.073 -74.189 49.972
268912 98.104 -74.199 49.914
268913 98.136 -74.210 49.852
268914 98.167 -74.222 49.787
268915 98.197 -74.236 49.722
268916 98.227 -74.250 49.658
268917 98.257 -74.264 49.598
268918 98.288 -74.279 49.539
268919 98.318 -74.292 49.482
268920 98.349 -74.307 49.427
268921 98.332 -74.314 49.371
268922 98.366 -74.340 49.314
268923 98.431 -74.377 49.256
268924 98.495 -74.415 49.199
268925 98.528 -74.435 49.139
268926 98.567 -74.422 49.079
268927 98.606 -74.389 49.020
268928 98.649 -74.390 48.963
268929 98.708 -74.441 48.908
268930 98.733 -74.547 48.857
268931 98.759 -74.574 48.807
268932 98.790 -74.582 48.757
268933 98.820 -74.590 48.705
268934 98.840 -74.599 48.654
268935 98.851 -74.609 48.603
268936 98.861 -74.619 48.552
268937 98.872 -74.629 48.498
268938 98.902 -74.619 48.444
268939 98.944 -74.596 48.391
268940 98.986 -74.574 48.340
268941 99.028 -74.552 48.288
268942 99.068 -74.536 48.239
268943 99.100 -74.540 48.189
268944 99.133 -74.546 48.139
268945 99.158 -74.564 48.089
268946 99.183 -74.584 48.041
268947 99.207 -74.605 47.994
268948 99.231 -74.626 47.947
268949 99.255 -74.646 47.900
268950 99.279 -74.667 47.856
268951 99.303 -74.687 47.811
268952 99.328 -74.709 47.767
268953 99.354 -74.722 47.725
268954 99.382 -74.730 47.682
268955 99.410 -74.737 47.641
268956 99.438 -74.744 47.602
268957 99.466 -74.750 47.564
268958 99.494 -74.758 47.528
268959 99.522 -74.765 47.491
268960 99.551 -74.772 47.453
268961 99.589 -74.783 47.415
268962 99.628 -74.794 47.377
268963 99.666 -74.806 47.339
268964 99.704 -74.816 47.301
268965 99.739 -74.819 47.268
268966 99.774 -74.820 47.235
268967 99.808 -74.821 47.203
268968 99.842 -74.822 47.172
268969 99.876 -74.822 47.141
268970 99.910 -74.823 47.111
268971 99.944 -74.825 47.082
268972 99.979 -74.825 47.056
268973 100.012 -74.838 47.028
268974 100.045 -74.853 47.000
268975 100.079 -74.868 46.974
268976 100.112 -74.883 46.948
268977 100.145 -74.898 46.921
268978 100.179 -74.924 46.899
268979 100.212 -74.962 46.877
268980 100.245 -75.000 46.855
268981 100.271 -75.017 46.830
268982 100.296 -75.029 46.804
268983 100.321 -75.042 46.778
268984 100.346 -75.054 46.751
268985 100.370 -75.066 46.724
268986 100.398 -75.082 46.698
268987 100.426 -75.098 46.672
268988 100.454 -75.114 46.645
268989 100.480 -75.130 46.618
268990 100.498 -75.147 46.591
268991 100.517 -75.163 46.563
268992 100.536 -75.179 46.536
268993 100.554 -75.195 46.509
268994 100.573 -75.212 46.482
268995 100.592 -75.228 46.453
268996 100.611 -75.245 46.424
268997 100.629 -75.261 46.394
268998 100.656 -75.271 46.367
268999 100.684 -75.282 46.340
269000 100.711 -75.292 46.312
269001 100.739 -75.303 46.284
269002 100.766 -75.310 46.256
269003 100.793 -75.305 46.228
269004 100.821 -75.300 46.200
269005 100.852 -75.295 46.174
269006 100.895 -75.292 46.145
269007 100.937 -75.290 46.116
269008 100.980 -75.287 46.088
269009 101.018 -75.289 46.061
269010 101.042 -75.305 46.033
269011 101.066 -75.321 46.004
269012 101.094 -75.346 45.975
269013 101.126 -75.381 45.944
269014 101.155 -75.411 45.913
269015 101.176 -75.432 45.880
269016 101.196 -75.453 45.847
269017 101.214 -75.460 45.815
269018 101.231 -75.450 45.782
269019 101.251 -75.442 45.749
269020 101.271 -75.434 45.716
269021 101.291 -75.426 45.682
269022 101.311 -75.418 45.648
269023 101.336 -75.420 45.615
269024 101.370 -75.435 45.582
269025 101.403 -75.451 45.551
269026 101.441 -75.477 45.520
269027 101.481 -75.510 45.490
269028 101.521 -75.543 45.461
269029 101.560 -75.575 45.430
269030 101.602 -75.609 45.402
269031 101.659 -75.646 45.373
269032 101.715 -75.685 45.344
269033 101.750 -75.708 45.317
269034 101.758 -75.710 45.290
269035 101.766 -75.713 45.263
269036 101.773 -75.716 45.236
269037 101.781 -75.718 45.209
269038 101.803 -75.738 45.183
269039 101.823 -75.749 45.158
269040 101.840 -75.749 45.133
269041 101.857 -75.750 45.107
269042 101.875 -75.751 45.082
269043 101.893 -75.751 45.057
269044 101.910 -75.752 45.032
269045 101.927 -75.753 45.008
269046 101.945 -75.753 44.984
269047 101.962 -75.753 44.960
269396 92.277 -73.296 54.747
269397 92.308 -73.306 54.737
269398 92.339 -73.317 54.727
269399 92.369 -73.327 54.717
269400 92.400 -73.337 54.706
269401 92.430 -73.348 54.697
269402 92.461 -73.358 54.688
269403 92.492 -73.368 54.678
269404 92.522 -73.379 54.670
269405 92.553 -73.389 54.662
269406 92.584 -73.399 54.655
269407 92.615 -73.410 54.648
269408 92.645 -73.420 54.639
269409 92.676 -73.430 54.630
269410 92.706 -73.440 54.622
269411 92.742 -73.451 54.615
269412 92.785 -73.462 54.610
269413 92.828 -73.473 54.605
269414 92.871 -73.484 54.602
269415 92.914 -73.495 54.600
269416 92.956 -73.506 54.600
269417 93.000 -73.517 54.601
269418 93.026 -73.513 54.603
269419 93.049 -73.505 54.605
269420 93.071 -73.496 54.607
269421 93.093 -73.488 54.609
269422 93.116 -73.479 54.612
269423 93.138 -73.471 54.615
269424 93.163 -73.466 54.619
269425 93.202 -73.479 54.619
269426 93.234 -73.491 54.618
269427 93.266 -73.504 54.615
269428 93.298 -73.515 54.612
269429 93.331 -73.528 54.608
269430 93.364 -73.539 54.604
269431 93.397 -73.550 54.600
269432 93.430 -73.562 54.594
269433 93.459 -73.570 54.586
269434 93.481 -73.573 54.576
269435 93.503 -73.576 54.565
269436 93.524 -73.579 54.552
269437 93.546 -73.582 54.538
269438 93.567 -73.584 54.526
269439 93.588 -73.587 54.514
269440 93.609 -73.589 54.501
269441 93.630 -73.589 54.490
269442 93.649 -73.587 54.479
269443 93.670 -73.584 54.468
269444 93.690 -73.582 54.457
269445 93.710 -73.580 54.444
269446 93.730 -73.578 54.427
269447 93.750 -73.575 54.409
269448 93.770 -73.574 54.390
269449 93.790 -73.571 54.372
269450 93.810 -73.569 54.352
269451 93.830 -73.567 54.332
269452 93.850 -73.565 54.313
269453 93.870 -73.562 54.292
269454 93.890 -73.560 54.271
269455 93.910 -73.558 54.251
269456 93.930 -73.556 54.230
269457 93.950 -73.553 54.210
269458 93.970 -73.551 54.190
269459 93.990 -73.549 54.170
269460 94.010 -73.545 54.151
269461 94.029 -73.540 54.132
269462 94.048 -73.535 54.109
269463 94.073 -73.545 54.084
269464 94.100 -73.552 54.058
269465 94.125 -73.559 54.034
269466 94.151 -73.567 54.007
269467 94.177 -73.574 53.981
269468 94.202 -73.581 53.955
269469 94.228 -73.589 53.929
269470 94.254 -73.597 53.902
269471 94.279 -73.603 53.875
269472 94.305 -73.611 53.848
269473 94.331 -73.618 53.820
269474 94.357 -73.626 53.793
269475 94.383 -73.633 53.764
269476 94.408 -73.640 53.735
269477 94.434 -73.648 53.707
269478 94.460 -73.655 53.678
269479 94.485 -73.662 53.649
269480 94.511 -73.670 53.620
269481 94.537 -73.678 53.590
269482 94.562 -73.684 53.563
269483 94.588 -73.692 53.536
269484 94.614 -73.700 53.510
269485 94.640 -73.707 53.485
269486 94.666 -73.714 53.461
269487 94.678 -73.716 53.438
269488 94.685 -73.715 53.416
269489 94.692 -73.713 53.396
269490 94.706 -73.716 53.375
269491 94.732 -73.716 53.354
269492 94.775 -73.705 53.334
269493 94.812 -73.698 53.312
269494 94.826 -73.711 53.286
269495 94.841 -73.725 53.257
269496 94.864 -73.728 53.224
269497 94.889 -73.731 53.187
269498 94.905 -73.731 53.149
269499 94.920 -73.731 53.111
269500 94.934 -73.731 53.073
269501 94.949 -73.730 53.034
269502 94.973 -73.734 52.996
269503 95.007 -73.743 52.958
269504 95.041 -73.751 52.921
269505 95.075 -73.760 52.884
269506 95.105 -73.745 52.847
269507 95.128 -73.700 52.812
269508 95.150 -73.655 52.778
269509 95.171 -73.615 52.745
269510 95.188 -73.609 52.714
269511 95.220 -73.618 52.682
269512 95.252 -73.632 52.649
269513 95.284 -73.647 52.616
269514 95.316 -73.661 52.582
269515 95.348 -73.675 52.549
269516 95.380 -73.690 52.515
269517 95.412 -73.703 52.481
269518 95.444 -73.718 52.445
269519 95.475 -73.732 52.410
269520 95.508 -73.747 52.376
269521 95.539 -73.761 52.343
269522 95.567 -73.756 52.311
269523 95.594 -73.744 52.280
269524 95.620 -73.732 52.249
269525 95.647 -73.721 52.220
269526 95.673 -73.709 52.193
269527 95.702 -73.700 52.166
269528 95.736 -73.695 52.138
269529 95.770 -73.691 52.108
269530 95.805 -73.687 52.076
269531 95.839 -73.683 52.044
269532 95.874 -73.678 52.011
269533 95.932 -73.716 51.977
269534 95.986 -73.740 51.944
269535 96.020 -73.744 51.911
269536 96.050 -73.756 51.878
269537 96.075 -73.751 51.845
269538 96.091 -73.738 51.810
269539 96.107 -73.724 51.776
269540 96.124 -73.710 51.743
269541 96.141 -73.696 51.709
269542 96.144 -73.697 51.675
269543 96.151 -73.700 51.642
269544 96.159 -73.703 51.609
269545 96.167 -73.707 51.576
269546 96.177 -73.722 51.543
269547 96.225 -73.749 51.510
269548 96.269 -73.751 51.477
269549 96.291 -73.763 51.444
269550 96.313 -73.775 51.411
269551 96.335 -73.787 51.379
269552 96.357 -73.798 51.348
269553 96.379 -73.810 51.316
269554 96.401 -73.822 51.287
269555 96.423 -73.833 51.257
269556 96.445 -73.845 51.228
269557 96.471 -73.853 51.200
269558 96.498 -73.860 51.173
269559 96.524 -73.866 51.146
269560 96.554 -73.870 51.119
269561 96.589 -73.870 51.093
269562 96.623 -73.870 51.070
269563 96.658 -73.870 51.049
269564 96.692 -73.869 51.029
269565 96.726 -73.866 51.011
269566 96.759 -73.863 50.995
269567 96.791 -73.866 50.979
269568 96.820 -73.879 50.964
269569 96.849 -73.891 50.950
269570 96.883 -73.906 50.940
269571 96.919 -73.922 50.931
269572 96.954 -73.937 50.923
269573 96.990 -73.953 50.916
269574 97.025 -73.969 50.913
269575 97.061 -73.985 50.910
269576 97.096 -74.000 50.909
269577 97.129 -73.992 50.910
269578 97.161 -73.981 50.916
269579 97.184 -73.967 50.920
269580 97.207 -73.953 50.923
269581 97.230 -73.938 50.925
269582 97.252 -73.923 50.925
269583 97.275 -73.909 50.921
269584 97.297 -73.894 50.914
269585 97.321 -73.900 50.905
269586 97.345 -73.906 50.891
269587 97.368 -73.912 50.881
269588 97.392 -73.919 50.872
269589 97.416 -73.925 50.861
269590 97.436 -73.930 50.847
269591 97.455 -73.936 50.832
269592 97.474 -73.941 50.817
269593 97.494 -73.948 50.799
269594 97.513 -73.953 50.775
269595 97.533 -73.959 50.745
269596 97.573 -73.963 50.711
269597 97.613 -73.969 50.675
269598 97.653 -73.974 50.631
269599 97.693 -73.979 50.583
269600 97.726 -73.990 50.531
269601 97.752 -74.007 50.478
269602 97.779 -74.024 50.431
269603 97.805 -74.042 50.384
269604 97.831 -74.058 50.337
269605 97.858 -74.076 50.292
269606 97.888 -74.091 50.245
269607 97.920 -74.107 50.201
269608 97.951 -74.120 50.157
269609 97.983 -74.130 50.114
269610 98.014 -74.142 50.068
269611 98.045 -74.152 50.019
269612 98.076 -74.163 49.965
269613 98.107 -74.173 49.906
269614 98.138 -74.184 49.843
269615 98.169 -74.195 49.778
269616 98.201 -74.205 49.712
269617 98.231 -74.219 49.648
269618 98.262 -74.233 49.584
269619 98.292 -74.247 49.524
269620 98.322 -74.261 49.467
269621 98.353 -74.275 49.412
269622 98.337 -74.283 49.355
269623 98.402 -74.320 49.299
269624 98.466 -74.358 49.243
269625 98.490 -74.374 49.184
269626 98.536 -74.367 49.123
269627 98.564 -74.357 49.063
269628 98.592 -74.352 49.004
269629 98.651 -74.375 48.948
269630 98.707 -74.436 48.894
269631 98.729 -74.482 48.843
269632 98.754 -74.517 48.793
269633 98.784 -74.524 48.742
269634 98.815 -74.532 48.692
269635 98.845 -74.540 48.641
269636 98.869 -74.548 48.590
269637 98.880 -74.559 48.539
269638 98.890 -74.568 48.486
269639 98.918 -74.562 48.433
269640 98.959 -74.539 48.380
269641 99.002 -74.517 48.327
269642 99.040 -74.505 48.275
269643 99.072 -74.509 48.225
269644 99.104 -74.515 48.175
269645 99.137 -74.519 48.125
269646 99.163 -74.535 48.077
269647 99.188 -74.556 48.029
269648 99.212 -74.577 47.981
269649 99.236 -74.597 47.935
269650 99.260 -74.618 47.889
269651 99.284 -74.639 47.844
269652 99.308 -74.659 47.800
269653 99.334 -74.680 47.756
269654 99.373 -74.693 47.714
269655 99.403 -74.702 47.673
269656 99.432 -74.710 47.632
269657 99.461 -74.717 47.592
269658 99.489 -74.724 47.556
269659 99.516 -74.731 47.519
269660 99.545 -74.738 47.483
269661 99.573 -74.745 47.446
269662 99.600 -74.752 47.407
269663 99.638 -74.763 47.370
269664 99.677 -74.774 47.332
269665 99.715 -74.785 47.295
269666 99.752 -74.793 47.262
269667 99.786 -74.794 47.229
269668 99.820 -74.794 47.197
269669 99.854 -74.795 47.166
269670 99.888 -74.796 47.135
269671 99.922 -74.797 47.104
269672 99.957 -74.797 47.075
269673 99.990 -74.810 47.049
269674 100.023 -74.825 47.021
269675 100.057 -74.840 46.993
269676 100.090 -74.855 46.967
269677 100.124 -74.870 46.940
269678 100.157 -74.885 46.914
269679 100.190 -74.900 46.892
269680 100.223 -74.936 46.870
269681 100.256 -74.974 46.847
269682 100.281 -74.987 46.823
269683 100.306 -74.999 46.796
269684 100.331 -75.011 46.769
269685 100.356 -75.024 46.742
269686 100.381 -75.036 46.715
269687 100.405 -75.048 46.689
269688 100.432 -75.063 46.662
269689 100.461 -75.079 46.635
269690 100.489 -75.095 46.608
269691 100.510 -75.112 46.580
269692 100.529 -75.128 46.553
269693 100.548 -75.145 46.526
269694 100.566 -75.160 46.498
269695 100.585 -75.177 46.471
269696 100.604 -75.193 46.443
269697 100.622 -75.210 46.413
269698 100.647 -75.222 46.383
269699 100.674 -75.232 46.356
269700 100.701 -75.243 46.329
269701 100.729 -75.254 46.303
269702 100.756 -75.264 46.275
269703 100.783 -75.271 46.247
269704 100.811 -75.266 46.219
269705 100.838 -75.261 46.191
269706 100.876 -75.258 46.164
269707 100.919 -75.254 46.135
269708 100.962 -75.252 46.107
269709 100.999 -75.254 46.078
269710 101.024 -75.271 46.050
269711 101.048 -75.287 46.020
269712 101.072 -75.303 45.990
269713 101.096 -75.319 45.960
269714 101.127 -75.351 45.928
269715 101.154 -75.380 45.896
269716 101.175 -75.401 45.864
269717 101.192 -75.405 45.831
269718 101.207 -75.394 45.798
269719 101.225 -75.384 45.766
269720 101.244 -75.377 45.733
269721 101.267 -75.373 45.700
269722 101.301 -75.389 45.665
269723 101.334 -75.404 45.631
269724 101.368 -75.420 45.599
269725 101.402 -75.436 45.567
269726 101.435 -75.451 45.536
269727 101.469 -75.467 45.505
269728 101.507 -75.494 45.475
269729 101.547 -75.527 45.445
269730 101.587 -75.560 45.415
269731 101.627 -75.593 45.386
269732 101.684 -75.631 45.357
269733 101.740 -75.669 45.328
269734 101.754 -75.676 45.302
269735 101.762 -75.679 45.274
269736 101.769 -75.682 45.247
269737 101.781 -75.690 45.220
269738 101.802 -75.709 45.194
269739 101.820 -75.709 45.167
269740 101.838 -75.709 45.142
269741 101.855 -75.710 45.117
269742 101.873 -75.710 45.090
269743 101.890 -75.711 45.065
269744 101.907 -75.712 45.041
269745 101.925 -75.712 45.016
269746 101.943 -75.712 44.992
269747 101.960 -75.713 44.968
269748 101.977 -75.714 44.944
270097 92.273 -73.267 54.755
270098 92.304 -73.277 54.745
270099 92.335 -73.287 54.734
270100 92.365 -73.298 54.722
270101 92.395 -73.308 54.712
270102 92.426 -73.318 54.702
270103 92.457 -73.329 54.691
270104 92.488 -73.339 54.682
270105 92.518 -73.349 54.672
270106 92.549 -73.360 54.663
270107 92.580 -73.370 54.655
270108 92.611 -73.380 54.646
270109 92.641 -73.391 54.637
270110 92.671 -73.401 54.628
270111 92.702 -73.412 54.620
270112 92.739 -73.422 54.614
270113 92.782 -73.433 54.608
270114 92.825 -73.444 54.603
270115 92.868 -73.455 54.599
270116 92.911 -73.466 54.598
270117 92.953 -73.477 54.598
270118 92.996 -73.488 54.598
270119 93.038 -73.498 54.600
270120 93.060 -73.489 54.600
270121 93.082 -73.481 54.601
270122 93.105 -73.472 54.601
270123 93.127 -73.464 54.604
270124 93.149 -73.456 54.606
270125 93.173 -73.448 54.607
270126 93.214 -73.462 54.609
270127 93.249 -73.475 54.607
270128 93.281 -73.487 54.604
270129 93.313 -73.499 54.601
270130 93.346 -73.512 54.597
270131 93.378 -73.524 54.594
270132 93.410 -73.536 54.589
270133 93.443 -73.547 54.584
270134 93.469 -73.553 54.574
270135 93.490 -73.556 54.563
270136 93.511 -73.559 54.551
270137 93.532 -73.561 54.537
270138 93.553 -73.562 54.523
270139 93.574 -73.565 54.511
270140 93.595 -73.567 54.499
270141 93.616 -73.568 54.485
270142 93.636 -73.565 54.474
270143 93.656 -73.563 54.461
270144 93.676 -73.561 54.449
270145 93.696 -73.559 54.437
270146 93.716 -73.556 54.424
270147 93.736 -73.554 54.407
270148 93.756 -73.552 54.390
270149 93.776 -73.550 54.372
270150 93.796 -73.547 54.353
270151 93.817 -73.545 54.332
270152 93.836 -73.543 54.312
270153 93.856 -73.540 54.293
270154 93.876 -73.538 54.271
270155 93.897 -73.536 54.249
270156 93.917 -73.534 54.228
270157 93.936 -73.531 54.207
270158 93.956 -73.526 54.186
270159 93.974 -73.521 54.166
270160 93.993 -73.516 54.145
270161 94.012 -73.511 54.124
270162 94.031 -73.506 54.103
270163 94.050 -73.500 54.079
270164 94.072 -73.504 54.055
270165 94.098 -73.515 54.029
270166 94.124 -73.522 54.005
270167 94.149 -73.530 53.979
270168 94.176 -73.537 53.953
270169 94.201 -73.544 53.927
270170 94.227 -73.552 53.900
270171 94.253 -73.559 53.874
270172 94.278 -73.566 53.848
270173 94.304 -73.574 53.822
270174 94.330 -73.581 53.794
270175 94.355 -73.589 53.766
270176 94.381 -73.596 53.737
270177 94.407 -73.603 53.709
270178 94.433 -73.611 53.682
270179 94.459 -73.618 53.653
270180 94.484 -73.625 53.624
270181 94.510 -73.633 53.596
270182 94.536 -73.640 53.567
270183 94.561 -73.647 53.539
270184 94.587 -73.655 53.512
270185 94.613 -73.662 53.487
270186 94.638 -73.670 53.462
270187 94.664 -73.677 53.439
270188 94.688 -73.684 53.416
270189 94.695 -73.683 53.395
270190 94.702 -73.681 53.374
270191 94.709 -73.681 53.351
270192 94.730 -73.685 53.329
270193 94.753 -73.687 53.307
270194 94.796 -73.676 53.281
270195 94.834 -73.669 53.253
270196 94.848 -73.682 53.224
270197 94.862 -73.695 53.192
270198 94.877 -73.709 53.158
270199 94.899 -73.714 53.121
270200 94.924 -73.716 53.085
270201 94.950 -73.718 53.049
270202 94.966 -73.718 53.012
270203 94.980 -73.718 52.975
270204 94.995 -73.718 52.938
270205 95.026 -73.725 52.902
270206 95.060 -73.733 52.866
270207 95.094 -73.740 52.830
270208 95.116 -73.694 52.797
270209 95.139 -73.650 52.765
270210 95.162 -73.604 52.733
270211 95.194 -73.590 52.703
270212 95.227 -73.605 52.671
270213 95.259 -73.618 52.639
270214 95.290 -73.633 52.606
270215 95.322 -73.647 52.573
270216 95.354 -73.662 52.540
270217 95.386 -73.676 52.507
270218 95.418 -73.691 52.474
270219 95.450 -73.704 52.439
270220 95.482 -73.719 52.404
270221 95.514 -73.733 52.371
270222 95.545 -73.744 52.339
270223 95.571 -73.732 52.308
270224 95.598 -73.721 52.278
270225 95.624 -73.709 52.249
270226 95.651 -73.697 52.220
270227 95.678 -73.685 52.194
270228 95.705 -73.675 52.168
270229 95.739 -73.670 52.141
270230 95.774 -73.666 52.111
270231 95.808 -73.662 52.079
270232 95.857 -73.683 52.047
270233 95.908 -73.698 52.014
270234 95.948 -73.687 51.982
270235 95.978 -73.700 51.949
270236 96.008 -73.713 51.916
270237 96.038 -73.726 51.883
270238 96.074 -73.726 51.849
270239 96.089 -73.720 51.814
270240 96.105 -73.706 51.780
270241 96.121 -73.693 51.746
270242 96.138 -73.678 51.712
270243 96.138 -73.675 51.679
270244 96.146 -73.677 51.645
270245 96.153 -73.680 51.612
270246 96.161 -73.683 51.579
270247 96.177 -73.688 51.546
270248 96.250 -73.710 51.513
270249 96.285 -73.725 51.481
270250 96.306 -73.733 51.449
270251 96.328 -73.742 51.416
270252 96.349 -73.753 51.384
270253 96.372 -73.764 51.353
270254 96.394 -73.776 51.321
270255 96.415 -73.788 51.292
270256 96.438 -73.800 51.263
270257 96.460 -73.811 51.234
270258 96.481 -73.822 51.204
270259 96.504 -73.835 51.177
270260 96.530 -73.842 51.149
270261 96.556 -73.849 51.122
270262 96.591 -73.849 51.096
270263 96.625 -73.849 51.073
270264 96.659 -73.846 51.052
270265 96.692 -73.843 51.032
270266 96.726 -73.840 51.013
270267 96.759 -73.837 50.996
270268 96.791 -73.841 50.981
270269 96.820 -73.854 50.966
270270 96.849 -73.866 50.952
270271 96.878 -73.879 50.940
270272 96.911 -73.894 50.930
270273 96.947 -73.910 50.922
270274 96.982 -73.925 50.915
270275 97.018 -73.941 50.910
270276 97.054 -73.957 50.907
270277 97.088 -73.958 50.905
270278 97.119 -73.948 50.905
270279 97.152 -73.937 50.908
270280 97.183 -73.926 50.911
270281 97.206 -73.911 50.912
270282 97.229 -73.897 50.913
270283 97.251 -73.882 50.911
270284 97.274 -73.868 50.907
270285 97.297 -73.870 50.901
270286 97.321 -73.876 50.893
270287 97.345 -73.883 50.883
270288 97.367 -73.889 50.875
270289 97.387 -73.894 50.868
270290 97.406 -73.901 50.860
270291 97.426 -73.906 50.847
270292 97.445 -73.912 50.835
270293 97.464 -73.917 50.821
270294 97.484 -73.923 50.803
270295 97.504 -73.929 50.776
270296 97.540 -73.934 50.743
270297 97.580 -73.939 50.706
270298 97.620 -73.944 50.666
270299 97.660 -73.949 50.623
270300 97.700 -73.954 50.576
270301 97.738 -73.961 50.527
270302 97.764 -73.978 50.477
270303 97.791 -73.995 50.429
270304 97.817 -74.012 50.383
270305 97.843 -74.029 50.337
270306 97.870 -74.046 50.292
270307 97.896 -74.063 50.247
270308 97.923 -74.080 50.202
270309 97.954 -74.094 50.157
270310 97.985 -74.104 50.111
270311 98.016 -74.115 50.062
270312 98.048 -74.126 50.010
270313 98.079 -74.137 49.954
270314 98.110 -74.148 49.893
270315 98.141 -74.158 49.829
270316 98.172 -74.169 49.764
270317 98.203 -74.179 49.699
270318 98.234 -74.190 49.635
270319 98.266 -74.201 49.570
270320 98.296 -74.215 49.508
270321 98.326 -74.230 49.449
270322 98.354 -74.243 49.393
270323 98.373 -74.264 49.338
270324 98.433 -74.299 49.283
270325 98.452 -74.313 49.226
270326 98.486 -74.316 49.167
270327 98.533 -74.302 49.106
270328 98.566 -74.309 49.046
270329 98.603 -74.330 48.988
270330 98.650 -74.353 48.932
270331 98.702 -74.377 48.880
270332 98.725 -74.417 48.830
270333 98.748 -74.459 48.779
270334 98.779 -74.467 48.729
270335 98.809 -74.474 48.678
270336 98.840 -74.482 48.627
270337 98.870 -74.490 48.576
270338 98.898 -74.498 48.526
270339 98.909 -74.508 48.474
270340 98.933 -74.504 48.421
270341 98.975 -74.481 48.368
270342 99.012 -74.473 48.315
270343 99.045 -74.478 48.263
270344 99.076 -74.483 48.212
270345 99.109 -74.488 48.161
270346 99.141 -74.493 48.112
270347 99.168 -74.507 48.064
270348 99.193 -74.527 48.016
270349 99.217 -74.548 47.969
270350 99.241 -74.569 47.923
270351 99.265 -74.590 47.878
270352 99.290 -74.610 47.832
270353 99.314 -74.631 47.788
270354 99.349 -74.648 47.745
270355 99.390 -74.662 47.704
270356 99.421 -74.671 47.663
270357 99.450 -74.680 47.623
270358 99.481 -74.688 47.583
270359 99.510 -74.697 47.547
270360 99.539 -74.705 47.510
270361 99.567 -74.712 47.474
270362 99.595 -74.718 47.438
270363 99.623 -74.725 47.400
270364 99.651 -74.733 47.362
270365 99.687 -74.743 47.325
270366 99.725 -74.754 47.289
270367 99.763 -74.765 47.256
270368 99.798 -74.766 47.222
270369 99.832 -74.767 47.191
270370 99.866 -74.768 47.160
270371 99.900 -74.769 47.129
270372 99.934 -74.769 47.098
270373 99.968 -74.782 47.069
270374 100.001 -74.797 47.042
270375 100.034 -74.812 47.014
270376 100.068 -74.827 46.986
270377 100.102 -74.841 46.959
270378 100.135 -74.856 46.933
270379 100.168 -74.872 46.908
270380 100.202 -74.887 46.885
270381 100.235 -74.910 46.862
270382 100.266 -74.944 46.840
270383 100.291 -74.956 46.814
270384 100.316 -74.969 46.788
270385 100.341 -74.981 46.760
270386 100.366 -74.993 46.732
270387 100.391 -75.006 46.707
270388 100.415 -75.018 46.680
270389 100.440 -75.030 46.653
270390 100.467 -75.045 46.626
270391 100.495 -75.061 46.598
270392 100.522 -75.077 46.571
270393 100.541 -75.093 46.544
270394 100.560 -75.110 46.517
270395 100.578 -75.126 46.489
270396 100.597 -75.142 46.461
270397 100.616 -75.158 46.433
270398 100.637 -75.173 46.404
270399 100.665 -75.183 46.373
270400 100.692 -75.194 46.346
270401 100.719 -75.204 46.319
270402 100.747 -75.215 46.292
270403 100.774 -75.226 46.264
270404 100.801 -75.232 46.237
270405 100.829 -75.227 46.210
270406 100.858 -75.223 46.181
270407 100.901 -75.220 46.153
270408 100.944 -75.217 46.125
270409 100.981 -75.220 46.096
270410 101.005 -75.236 46.067
270411 101.029 -75.252 46.039
270412 101.053 -75.268 46.007
270413 101.077 -75.285 45.976
270414 101.102 -75.301 45.944
270415 101.128 -75.321 45.912
270416 101.153 -75.348 45.879
270417 101.171 -75.349 45.847
270418 101.185 -75.339 45.814
270419 101.201 -75.330 45.782
270420 101.231 -75.340 45.749
270421 101.265 -75.357 45.716
270422 101.299 -75.373 45.682
270423 101.332 -75.389 45.649
270424 101.366 -75.405 45.615
270425 101.400 -75.420 45.583
270426 101.433 -75.436 45.551
270427 101.467 -75.452 45.520
270428 101.501 -75.467 45.490
270429 101.535 -75.483 45.460
270430 101.573 -75.511 45.430
270431 101.613 -75.544 45.400
270432 101.653 -75.577 45.370
270433 101.709 -75.615 45.341
270434 101.750 -75.643 45.313
270435 101.758 -75.645 45.286
270436 101.765 -75.648 45.258
270437 101.781 -75.660 45.231
270438 101.800 -75.668 45.204
270439 101.818 -75.668 45.178
270440 101.835 -75.669 45.152
270441 101.852 -75.670 45.126
270442 101.870 -75.670 45.100
270443 101.888 -75.671 45.075
270444 101.905 -75.671 45.050
270445 101.922 -75.671 45.025
270446 101.940 -75.672 45.001
270447 101.957 -75.673 44.976
270448 101.975 -75.674 44.952
270798 92.269 -73.237 54.763
270799 92.299 -73.248 54.752
270800 92.330 -73.258 54.740
270801 92.361 -73.268 54.728
270802 92.391 -73.279 54.718
270803 92.422 -73.289 54.707
270804 92.453 -73.299 54.695
270805 92.484 -73.310 54.685
270806 92.514 -73.321 54.675
270807 92.545 -73.330 54.665
270808 92.575 -73.341 54.654
270809 92.606 -73.351 54.644
270810 92.637 -73.362 54.635
270811 92.667 -73.372 54.626
270812 92.698 -73.382 54.619
270813 92.736 -73.393 54.613
270814 92.779 -73.404 54.607
270815 92.821 -73.415 54.601
270816 92.864 -73.426 54.598
270817 92.907 -73.437 54.596
270818 92.950 -73.448 54.596
270819 92.993 -73.459 54.596
270820 93.036 -73.470 54.596
270821 93.071 -73.474 54.596
270822 93.093 -73.465 54.595
270823 93.116 -73.456 54.595
270824 93.138 -73.448 54.596
270825 93.160 -73.440 54.597
270826 93.183 -73.431 54.597
270827 93.223 -73.445 54.598
270828 93.264 -73.459 54.596
270829 93.297 -73.471 54.593
270830 93.329 -73.484 54.590
270831 93.361 -73.496 54.586
270832 93.394 -73.508 54.582
270833 93.425 -73.520 54.577
270834 93.455 -73.531 54.571
270835 93.476 -73.534 54.561
270836 93.497 -73.537 54.548
270837 93.518 -73.539 54.535
270838 93.540 -73.541 54.521
270839 93.561 -73.543 54.508
270840 93.582 -73.546 54.496
270841 93.603 -73.546 54.484
270842 93.622 -73.543 54.470
270843 93.642 -73.541 54.456
270844 93.662 -73.539 54.443
270845 93.683 -73.537 54.431
270846 93.703 -73.534 54.418
270847 93.722 -73.532 54.405
270848 93.743 -73.530 54.388
270849 93.763 -73.528 54.371
270850 93.783 -73.525 54.352
270851 93.802 -73.523 54.332
270852 93.823 -73.521 54.312
270853 93.843 -73.518 54.292
270854 93.863 -73.516 54.272
270855 93.883 -73.513 54.249
270856 93.901 -73.508 54.227
270857 93.920 -73.503 54.204
270858 93.939 -73.497 54.182
270859 93.958 -73.492 54.162
270860 93.977 -73.487 54.142
270861 93.995 -73.481 54.120
270862 94.014 -73.476 54.096
270863 94.033 -73.471 54.073
270864 94.052 -73.466 54.050
270865 94.071 -73.461 54.026
270866 94.097 -73.474 54.001
270867 94.123 -73.485 53.977
270868 94.148 -73.493 53.952
270869 94.174 -73.500 53.925
270870 94.200 -73.507 53.899
270871 94.225 -73.515 53.872
270872 94.252 -73.522 53.846
270873 94.277 -73.529 53.821
270874 94.303 -73.537 53.796
270875 94.329 -73.544 53.768
270876 94.354 -73.552 53.740
270877 94.380 -73.559 53.712
270878 94.406 -73.566 53.684
270879 94.431 -73.574 53.657
270880 94.457 -73.581 53.629
270881 94.483 -73.588 53.600
270882 94.509 -73.596 53.572
270883 94.535 -73.603 53.545
270884 94.560 -73.610 53.517
270885 94.586 -73.618 53.490
270886 94.612 -73.625 53.465
270887 94.637 -73.633 53.441
270888 94.663 -73.640 53.417
270889 94.689 -73.647 53.395
270890 94.706 -73.651 53.373
270891 94.713 -73.650 53.352
270892 94.720 -73.649 53.329
270893 94.733 -73.650 53.306
270894 94.754 -73.655 53.283
270895 94.774 -73.659 53.256
270896 94.817 -73.647 53.228
270897 94.856 -73.639 53.198
270898 94.870 -73.653 53.167
270899 94.885 -73.666 53.133
270900 94.899 -73.679 53.098
270901 94.913 -73.693 53.063
270902 94.934 -73.699 53.028
270903 94.959 -73.701 52.992
270904 94.984 -73.703 52.955
270905 95.010 -73.705 52.919
270906 95.027 -73.705 52.883
270907 95.044 -73.706 52.849
270908 95.078 -73.715 52.814
270909 95.105 -73.689 52.782
270910 95.128 -73.644 52.751
270911 95.150 -73.614 52.720
270912 95.177 -73.593 52.689
270913 95.219 -73.600 52.658
270914 95.263 -73.607 52.627
270915 95.297 -73.619 52.596
270916 95.329 -73.634 52.563
270917 95.360 -73.648 52.532
270918 95.393 -73.662 52.500
270919 95.424 -73.677 52.466
270920 95.457 -73.691 52.431
270921 95.488 -73.705 52.398
270922 95.520 -73.719 52.366
270923 95.549 -73.721 52.335
270924 95.576 -73.709 52.305
270925 95.602 -73.697 52.277
270926 95.629 -73.685 52.249
270927 95.656 -73.673 52.221
270928 95.682 -73.662 52.196
270929 95.708 -73.650 52.170
270930 95.743 -73.645 52.143
270931 95.782 -73.650 52.113
270932 95.829 -73.656 52.083
270933 95.870 -73.645 52.050
270934 95.906 -73.643 52.017
270935 95.936 -73.656 51.987
270936 95.966 -73.669 51.954
270937 95.995 -73.682 51.921
270938 96.025 -73.696 51.888
270939 96.061 -73.697 51.854
270940 96.090 -73.697 51.819
270941 96.103 -73.689 51.785
270942 96.119 -73.675 51.751
270943 96.135 -73.661 51.717
270944 96.141 -73.653 51.682
270945 96.141 -73.655 51.649
270946 96.148 -73.657 51.616
270947 96.156 -73.660 51.583
270948 96.215 -73.678 51.550
270949 96.262 -73.698 51.517
270950 96.303 -73.716 51.484
270951 96.324 -73.725 51.452
270952 96.346 -73.734 51.420
270953 96.368 -73.742 51.389
270954 96.389 -73.751 51.357
270955 96.411 -73.760 51.326
270956 96.432 -73.769 51.297
270957 96.454 -73.778 51.268
270958 96.475 -73.787 51.239
270959 96.497 -73.795 51.208
270960 96.519 -73.804 51.180
270961 96.540 -73.813 51.153
270962 96.562 -73.824 51.126
270963 96.592 -73.826 51.100
270964 96.626 -73.823 51.077
270965 96.660 -73.820 51.055
270966 96.693 -73.817 51.035
270967 96.727 -73.813 51.016
270968 96.760 -73.810 50.998
270969 96.791 -73.817 50.983
270970 96.820 -73.830 50.968
270971 96.849 -73.842 50.953
270972 96.878 -73.855 50.940
270973 96.907 -73.867 50.930
270974 96.940 -73.882 50.922
270975 96.975 -73.897 50.914
270976 97.011 -73.913 50.908
270977 97.046 -73.925 50.904
270978 97.078 -73.914 50.901
270979 97.110 -73.903 50.900
270980 97.142 -73.892 50.901
270981 97.174 -73.882 50.902
270982 97.205 -73.870 50.903
270983 97.228 -73.856 50.903
270984 97.250 -73.841 50.899
270985 97.274 -73.841 50.895
270986 97.298 -73.847 50.890
270987 97.318 -73.854 50.884
270988 97.337 -73.859 50.876
270989 97.357 -73.865 50.870
270990 97.377 -73.870 50.864
270991 97.396 -73.876 50.856
270992 97.416 -73.882 50.846
270993 97.435 -73.888 50.836
270994 97.455 -73.893 50.823
270995 97.474 -73.899 50.806
270996 97.507 -73.904 50.778
270997 97.547 -73.909 50.744
270998 97.587 -73.914 50.705
270999 97.627 -73.919 50.663
271000 97.667 -73.924 50.620
271001 97.707 -73.929 50.573
271002 97.747 -73.935 50.524
271003 97.776 -73.949 50.476
271004 97.802 -73.966 50.429
271005 97.829 -73.983 50.383
271006 97.855 -74.000 50.337
271007 97.881 -74.017 50.292
271008 97.908 -74.034 50.248
271009 97.934 -74.051 50.202
271010 97.968 -74.064 50.155
271011 98.005 -74.076 50.106
271012 98.032 -74.087 50.054
271013 98.060 -74.098 49.999
271014 98.087 -74.110 49.941
271015 98.114 -74.121 49.879
271016 98.144 -74.132 49.814
271017 98.175 -74.143 49.749
271018 98.206 -74.154 49.684
271019 98.237 -74.164 49.620
271020 98.268 -74.175 49.556
271021 98.299 -74.186 49.492
271022 98.331 -74.198 49.431
271023 98.354 -74.211 49.373
271024 98.394 -74.238 49.320
271025 98.414 -74.252 49.265
271026 98.434 -74.265 49.208
271027 98.498 -74.248 49.149
271028 98.539 -74.267 49.088
271029 98.576 -74.287 49.029
271030 98.614 -74.308 48.972
271031 98.651 -74.329 48.916
271032 98.692 -74.350 48.865
271033 98.719 -74.368 48.816
271034 98.744 -74.398 48.766
271035 98.773 -74.409 48.715
271036 98.804 -74.417 48.665
271037 98.835 -74.424 48.614
271038 98.865 -74.432 48.563
271039 98.896 -74.440 48.513
271040 98.926 -74.447 48.461
271041 98.948 -74.446 48.408
271042 98.984 -74.442 48.355
271043 99.016 -74.447 48.303
271044 99.049 -74.452 48.250
271045 99.081 -74.457 48.199
271046 99.113 -74.461 48.149
271047 99.145 -74.467 48.099
271048 99.173 -74.478 48.050
271049 99.198 -74.499 48.003
271050 99.222 -74.520 47.957
271051 99.246 -74.540 47.912
271052 99.271 -74.562 47.867
271053 99.295 -74.582 47.821
271054 99.323 -74.601 47.777
271055 99.364 -74.616 47.735
271056 99.405 -74.631 47.694
271057 99.439 -74.641 47.654
271058 99.469 -74.650 47.614
271059 99.498 -74.659 47.575
271060 99.528 -74.667 47.538
271061 99.558 -74.675 47.501
271062 99.588 -74.684 47.465
271063 99.618 -74.692 47.428
271064 99.646 -74.699 47.392
271065 99.674 -74.706 47.355
271066 99.702 -74.713 47.318
271067 99.735 -74.722 47.282
271068 99.774 -74.734 47.249
271069 99.810 -74.739 47.216
271070 99.844 -74.740 47.184
271071 99.878 -74.740 47.154
271072 99.912 -74.742 47.123
271073 99.946 -74.753 47.092
271074 99.979 -74.769 47.063
271075 100.012 -74.784 47.035
271076 100.046 -74.799 47.008
271077 100.079 -74.813 46.981
271078 100.113 -74.828 46.954
271079 100.146 -74.844 46.927
271080 100.180 -74.858 46.902
271081 100.213 -74.873 46.879
271082 100.247 -74.888 46.856
271083 100.276 -74.913 46.833
271084 100.301 -74.925 46.808
271085 100.326 -74.938 46.780
271086 100.351 -74.950 46.752
271087 100.375 -74.963 46.724
271088 100.401 -74.975 46.699
271089 100.425 -74.988 46.672
271090 100.450 -75.000 46.645
271091 100.475 -75.012 46.618
271092 100.501 -75.026 46.589
271093 100.530 -75.042 46.562
271094 100.553 -75.058 46.534
271095 100.571 -75.075 46.507
271096 100.590 -75.091 46.479
271097 100.609 -75.107 46.451
271098 100.627 -75.123 46.422
271099 100.655 -75.134 46.393
271100 100.682 -75.145 46.363
271101 100.709 -75.155 46.336
271102 100.737 -75.166 46.308
271103 100.764 -75.176 46.281
271104 100.791 -75.188 46.255
271105 100.819 -75.194 46.228
271106 100.846 -75.189 46.200
271107 100.882 -75.185 46.171
271108 100.925 -75.182 46.142
271109 100.962 -75.185 46.114
271110 100.986 -75.201 46.085
271111 101.010 -75.218 46.056
271112 101.035 -75.234 46.026
271113 101.059 -75.250 45.994
271114 101.083 -75.266 45.962
271115 101.107 -75.283 45.929
271116 101.131 -75.298 45.896
271117 101.155 -75.302 45.863
271118 101.176 -75.301 45.830
271119 101.198 -75.299 45.797
271120 101.226 -75.308 45.765
271121 101.260 -75.327 45.733
271122 101.294 -75.346 45.699
271123 101.329 -75.366 45.666
271124 101.363 -75.386 45.633
271125 101.398 -75.405 45.601
271126 101.431 -75.420 45.568
271127 101.465 -75.436 45.536
271128 101.499 -75.452 45.506
271129 101.533 -75.467 45.475
271130 101.566 -75.483 45.445
271131 101.600 -75.499 45.414
271132 101.639 -75.529 45.385
271133 101.679 -75.561 45.356
271134 101.734 -75.599 45.327
271135 101.754 -75.611 45.298
271136 101.761 -75.614 45.271
271137 101.780 -75.627 45.243
271138 101.798 -75.627 45.215
271139 101.815 -75.628 45.188
271140 101.833 -75.629 45.162
271141 101.850 -75.630 45.136
271142 101.868 -75.630 45.110
271143 101.885 -75.630 45.084
271144 101.902 -75.631 45.059
271145 101.920 -75.631 45.034
271146 101.938 -75.632 45.009
271147 101.955 -75.633 44.984
271148 101.972 -75.633 44.960
271149 101.990 -75.634 44.936
271499 92.265 -73.208 54.770
271500 92.295 -73.219 54.759
271501 92.326 -73.229 54.746
271502 92.357 -73.239 54.734
271503 92.387 -73.250 54.723
271504 92.418 -73.260 54.711
271505 92.449 -73.271 54.699
271506 92.479 -73.280 54.687
271507 92.510 -73.291 54.677
271508 92.541 -73.302 54.666
271509 92.571 -73.311 54.655
271510 92.602 -73.322 54.645
271511 92.633 -73.333 54.636
271512 92.663 -73.343 54.627
271513 92.694 -73.353 54.619
271514 92.732 -73.364 54.612
271515 92.775 -73.375 54.605
271516 92.818 -73.386 54.600
271517 92.861 -73.397 54.597
271518 92.904 -73.408 54.595
271519 92.947 -73.419 54.594
271520 92.990 -73.430 54.593
271521 93.033 -73.441 54.593
271522 93.076 -73.452 54.592
271523 93.105 -73.449 54.590
271524 93.127 -73.441 54.590
271525 93.149 -73.433 54.589
271526 93.171 -73.424 54.588
271527 93.194 -73.415 54.588
271528 93.233 -73.427 54.587
271529 93.274 -73.442 54.585
271530 93.312 -73.456 54.581
271531 93.344 -73.468 54.579
271532 93.376 -73.480 54.575
271533 93.409 -73.492 54.570
271534 93.441 -73.504 54.564
271535 93.463 -73.512 54.557
271536 93.484 -73.515 54.546
271537 93.505 -73.518 54.533
271538 93.526 -73.520 54.520
271539 93.547 -73.522 54.506
271540 93.568 -73.524 54.493
271541 93.589 -73.524 54.481
271542 93.609 -73.521 54.469
271543 93.629 -73.519 54.454
271544 93.649 -73.518 54.440
271545 93.669 -73.515 54.427
271546 93.689 -73.513 54.413
271547 93.709 -73.511 54.399
271548 93.729 -73.509 54.387
271549 93.749 -73.506 54.370
271550 93.769 -73.504 54.351
271551 93.789 -73.502 54.333
271552 93.809 -73.499 54.312
271553 93.828 -73.494 54.291
271554 93.847 -73.489 54.271
271555 93.866 -73.484 54.251
271556 93.885 -73.478 54.227
271557 93.903 -73.473 54.204
271558 93.922 -73.468 54.181
271559 93.941 -73.463 54.158
271560 93.960 -73.458 54.138
271561 93.979 -73.452 54.116
271562 93.998 -73.447 54.094
271563 94.016 -73.442 54.069
271564 94.035 -73.437 54.046
271565 94.054 -73.431 54.022
271566 94.073 -73.426 53.998
271567 94.095 -73.428 53.974
271568 94.121 -73.443 53.950
271569 94.147 -73.456 53.925
271570 94.173 -73.463 53.898
271571 94.199 -73.470 53.871
271572 94.224 -73.477 53.845
271573 94.250 -73.485 53.819
271574 94.276 -73.492 53.794
271575 94.301 -73.499 53.769
271576 94.328 -73.507 53.742
271577 94.353 -73.515 53.715
271578 94.379 -73.521 53.688
271579 94.405 -73.529 53.660
271580 94.430 -73.537 53.632
271581 94.456 -73.543 53.605
271582 94.482 -73.551 53.578
271583 94.507 -73.559 53.549
271584 94.533 -73.566 53.522
271585 94.559 -73.573 53.496
271586 94.585 -73.581 53.470
271587 94.611 -73.588 53.444
271588 94.636 -73.596 53.420
271589 94.662 -73.603 53.396
271590 94.688 -73.610 53.374
271591 94.713 -73.618 53.353
271592 94.724 -73.618 53.330
271593 94.731 -73.617 53.308
271594 94.738 -73.616 53.284
271595 94.757 -73.620 53.260
271596 94.777 -73.624 53.233
271597 94.797 -73.628 53.204
271598 94.838 -73.618 53.174
271599 94.878 -73.610 53.143
271600 94.892 -73.623 53.111
271601 94.906 -73.637 53.076
271602 94.921 -73.650 53.042
271603 94.935 -73.663 53.007
271604 94.950 -73.677 52.972
271605 94.969 -73.684 52.936
271606 94.994 -73.686 52.901
271607 95.020 -73.688 52.866
271608 95.045 -73.690 52.833
271609 95.070 -73.692 52.800
271610 95.094 -73.683 52.768
271611 95.116 -73.654 52.737
271612 95.138 -73.626 52.706
271613 95.160 -73.597 52.675
271614 95.202 -73.603 52.645
271615 95.245 -73.609 52.614
271616 95.288 -73.616 52.584
271617 95.331 -73.623 52.553
271618 95.367 -73.634 52.523
271619 95.399 -73.649 52.491
271620 95.431 -73.663 52.459
271621 95.463 -73.678 52.425
271622 95.495 -73.692 52.393
271623 95.527 -73.706 52.362
271624 95.554 -73.697 52.333
271625 95.580 -73.685 52.305
271626 95.607 -73.673 52.277
271627 95.633 -73.662 52.250
271628 95.660 -73.650 52.223
271629 95.686 -73.637 52.199
271630 95.712 -73.626 52.173
271631 95.751 -73.614 52.146
271632 95.791 -73.603 52.116
271633 95.831 -73.592 52.085
271634 95.864 -73.599 52.054
271635 95.894 -73.612 52.023
271636 95.923 -73.625 51.992
271637 95.953 -73.639 51.960
271638 95.983 -73.652 51.928
271639 96.013 -73.665 51.896
271640 96.048 -73.669 51.861
271641 96.085 -73.668 51.826
271642 96.104 -73.666 51.791
271643 96.117 -73.657 51.757
271644 96.133 -73.644 51.723
271645 96.145 -73.632 51.688
271646 96.136 -73.632 51.653
271647 96.143 -73.635 51.620
271648 96.179 -73.646 51.587
271649 96.226 -73.666 51.554
271650 96.273 -73.687 51.521
271651 96.320 -73.706 51.488
271652 96.342 -73.716 51.456
271653 96.364 -73.725 51.424
271654 96.386 -73.734 51.393
271655 96.407 -73.742 51.362
271656 96.429 -73.751 51.331
271657 96.450 -73.760 51.302
271658 96.472 -73.769 51.272
271659 96.493 -73.778 51.242
271660 96.515 -73.787 51.212
271661 96.537 -73.796 51.184
271662 96.553 -73.796 51.157
271663 96.567 -73.791 51.129
271664 96.621 -73.789 51.104
271665 96.658 -73.792 51.080
271666 96.691 -73.789 51.059
271667 96.726 -73.786 51.038
271668 96.759 -73.782 51.019
271669 96.793 -73.779 51.000
271670 96.836 -73.785 50.984
271671 96.855 -73.799 50.968
271672 96.866 -73.815 50.953
271673 96.878 -73.831 50.941
271674 96.907 -73.843 50.930
271675 96.937 -73.856 50.921
271676 96.969 -73.869 50.913
271677 97.004 -73.885 50.907
271678 97.037 -73.881 50.901
271679 97.069 -73.869 50.898
271680 97.101 -73.859 50.895
271681 97.133 -73.848 50.895
271682 97.165 -73.837 50.895
271683 97.197 -73.826 50.895
271684 97.227 -73.815 50.895
271685 97.250 -73.812 50.890
271686 97.270 -73.818 50.887
271687 97.289 -73.823 50.884
271688 97.308 -73.829 50.879
271689 97.328 -73.835 50.870
271690 97.347 -73.841 50.863
271691 97.367 -73.846 50.857
271692 97.386 -73.852 50.849
271693 97.406 -73.857 50.842
271694 97.425 -73.863 50.834
271695 97.445 -73.869 50.823
271696 97.474 -73.874 50.807
271697 97.514 -73.879 50.781
271698 97.554 -73.885 50.748
271699 97.594 -73.889 50.710
271700 97.634 -73.894 50.667
271701 97.674 -73.900 50.623
271702 97.714 -73.904 50.574
271703 97.754 -73.910 50.524
271704 97.788 -73.919 50.475
271705 97.814 -73.937 50.429
271706 97.841 -73.954 50.382
271707 97.867 -73.971 50.335
271708 97.893 -73.988 50.289
271709 97.920 -74.004 50.244
271710 97.946 -74.022 50.198
271711 97.986 -74.033 50.149
271712 98.027 -74.043 50.099
271713 98.069 -74.054 50.043
271714 98.101 -74.065 49.986
271715 98.128 -74.076 49.926
271716 98.156 -74.088 49.863
271717 98.183 -74.099 49.799
271718 98.210 -74.111 49.734
271719 98.237 -74.122 49.670
271720 98.265 -74.134 49.605
271721 98.292 -74.145 49.540
271722 98.319 -74.157 49.476
271723 98.347 -74.168 49.414
271724 98.368 -74.181 49.355
271725 98.397 -74.179 49.300
271726 98.423 -74.191 49.244
271727 98.449 -74.201 49.187
271728 98.503 -74.193 49.129
271729 98.557 -74.225 49.070
271730 98.593 -74.248 49.011
271731 98.629 -74.271 48.954
271732 98.665 -74.295 48.900
271733 98.693 -74.313 48.849
271734 98.720 -74.330 48.801
271735 98.745 -74.355 48.751
271736 98.772 -74.365 48.700
271737 98.802 -74.373 48.650
271738 98.833 -74.380 48.600
271739 98.863 -74.387 48.550
271740 98.894 -74.395 48.498
271741 98.924 -74.402 48.447
271742 98.956 -74.410 48.395
271743 98.988 -74.416 48.342
271744 99.020 -74.421 48.291
271745 99.053 -74.426 48.239
271746 99.085 -74.430 48.188
271747 99.117 -74.436 48.137
271748 99.149 -74.440 48.087
271749 99.179 -74.450 48.039
271750 99.203 -74.471 47.992
271751 99.227 -74.492 47.946
271752 99.252 -74.512 47.901
271753 99.276 -74.533 47.856
271754 99.300 -74.553 47.811
271755 99.338 -74.569 47.767
271756 99.379 -74.584 47.725
271757 99.420 -74.599 47.684
271758 99.457 -74.612 47.644
271759 99.487 -74.620 47.605
271760 99.516 -74.628 47.567
271761 99.546 -74.637 47.529
271762 99.576 -74.645 47.494
271763 99.606 -74.653 47.458
271764 99.636 -74.662 47.421
271765 99.665 -74.671 47.385
271766 99.695 -74.679 47.348
271767 99.724 -74.687 47.312
271768 99.752 -74.693 47.276
271769 99.784 -74.702 47.244
271770 99.822 -74.712 47.211
271771 99.856 -74.713 47.179
271772 99.890 -74.714 47.148
271773 99.924 -74.725 47.117
271774 99.957 -74.740 47.086
271775 99.991 -74.756 47.057
271776 100.024 -74.770 47.029
271777 100.057 -74.785 47.002
271778 100.091 -74.800 46.975
271779 100.124 -74.815 46.948
271780 100.157 -74.830 46.922
271781 100.191 -74.845 46.897
271782 100.224 -74.859 46.874
271783 100.258 -74.874 46.850
271784 100.286 -74.885 46.827
271785 100.311 -74.895 46.801
271786 100.336 -74.907 46.773
271787 100.361 -74.919 46.744
271788 100.385 -74.932 46.716
271789 100.411 -74.944 46.691
271790 100.435 -74.957 46.663
271791 100.460 -74.969 46.636
271792 100.485 -74.982 46.609
271793 100.510 -74.994 46.580
271794 100.536 -75.007 46.552
271795 100.564 -75.023 46.524
271796 100.583 -75.040 46.497
271797 100.602 -75.056 46.468
271798 100.621 -75.072 46.440
271799 100.645 -75.085 46.411
271800 100.673 -75.095 46.383
271801 100.700 -75.106 46.354
271802 100.727 -75.116 46.325
271803 100.755 -75.127 46.297
271804 100.782 -75.138 46.271
271805 100.809 -75.149 46.246
271806 100.837 -75.155 46.219
271807 100.864 -75.150 46.190
271808 100.907 -75.148 46.161
271809 100.944 -75.151 46.132
271810 100.968 -75.167 46.103
271811 100.992 -75.183 46.073
271812 101.016 -75.199 46.043
271813 101.040 -75.215 46.013
271814 101.064 -75.232 45.981
271815 101.092 -75.245 45.948
271816 101.121 -75.257 45.916
271817 101.149 -75.269 45.882
271818 101.173 -75.271 45.848
271819 101.195 -75.270 45.814
271820 101.220 -75.275 45.780
271821 101.255 -75.295 45.748
271822 101.289 -75.314 45.715
271823 101.324 -75.333 45.683
271824 101.358 -75.352 45.650
271825 101.393 -75.372 45.618
271826 101.427 -75.392 45.586
271827 101.462 -75.411 45.554
271828 101.496 -75.430 45.522
271829 101.531 -75.450 45.492
271830 101.565 -75.467 45.461
271831 101.599 -75.483 45.430
271832 101.632 -75.499 45.400
271833 101.666 -75.514 45.370
271834 101.705 -75.546 45.341
271835 101.750 -75.577 45.312
271836 101.759 -75.581 45.283
271837 101.778 -75.587 45.256
271838 101.795 -75.587 45.228
271839 101.813 -75.588 45.200
271840 101.830 -75.589 45.172
271841 101.847 -75.589 45.146
271842 101.865 -75.590 45.120
271843 101.883 -75.590 45.094
271844 101.900 -75.591 45.068
271845 101.917 -75.591 45.043
271846 101.935 -75.592 45.018
271847 101.952 -75.593 44.993
271848 101.970 -75.593 44.967
271849 101.987 -75.593 44.944
272200 92.261 -73.179 54.777
272201 92.291 -73.189 54.766
272202 92.322 -73.200 54.753
272203 92.353 -73.210 54.740
272204 92.383 -73.221 54.727
272205 92.414 -73.231 54.715
272206 92.444 -73.241 54.702
272207 92.475 -73.252 54.691
272208 92.506 -73.262 54.679
272209 92.536 -73.272 54.668
272210 92.567 -73.283 54.657
272211 92.598 -73.293 54.646
272212 92.629 -73.303 54.636
272213 92.659 -73.314 54.627
272214 92.690 -73.324 54.618
272215 92.729 -73.335 54.611
272216 92.772 -73.346 54.604
272217 92.815 -73.357 54.599
272218 92.858 -73.368 54.595
272219 92.901 -73.379 54.593
272220 92.944 -73.390 54.592
272221 92.987 -73.401 54.591
272222 93.029 -73.412 54.591
272223 93.072 -73.423 54.589
272224 93.115 -73.434 54.587
272225 93.138 -73.425 54.585
272226 93.160 -73.417 54.583
272227 93.183 -73.409 54.582
272228 93.205 -73.400 54.580
272229 93.242 -73.410 54.579
272230 93.283 -73.424 54.575
272231 93.324 -73.438 54.571
272232 93.360 -73.451 54.568
272233 93.392 -73.463 54.564
272234 93.424 -73.476 54.558
272235 93.452 -73.487 54.551
272236 93.471 -73.493 54.543
272237 93.491 -73.496 54.531
272238 93.512 -73.499 54.518
272239 93.534 -73.501 54.505
272240 93.555 -73.503 54.492
272241 93.575 -73.503 54.478
272242 93.595 -73.500 54.467
272243 93.615 -73.498 54.454
272244 93.635 -73.496 54.439
272245 93.655 -73.493 54.425
272246 93.675 -73.491 54.411
272247 93.695 -73.489 54.397
272248 93.715 -73.487 54.382
272249 93.735 -73.484 54.368
272250 93.755 -73.481 54.351
272251 93.774 -73.475 54.332
272252 93.792 -73.470 54.313
272253 93.812 -73.465 54.291
272254 93.830 -73.460 54.270
272255 93.849 -73.455 54.250
272256 93.868 -73.449 54.229
272257 93.887 -73.444 54.205
272258 93.906 -73.439 54.181
272259 93.924 -73.434 54.158
272260 93.943 -73.428 54.135
272261 93.962 -73.423 54.114
272262 93.981 -73.418 54.091
272263 94.000 -73.413 54.068
272264 94.018 -73.408 54.043
272265 94.038 -73.402 54.020
272266 94.056 -73.397 53.996
272267 94.075 -73.392 53.972
272268 94.094 -73.387 53.948
272269 94.119 -73.398 53.923
272270 94.145 -73.412 53.898
272271 94.172 -73.426 53.871
272272 94.198 -73.433 53.846
272273 94.223 -73.440 53.820
272274 94.249 -73.448 53.795
272275 94.275 -73.455 53.770
272276 94.300 -73.462 53.744
272277 94.326 -73.470 53.717
272278 94.352 -73.477 53.691
272279 94.378 -73.484 53.664
272280 94.403 -73.492 53.636
272281 94.429 -73.499 53.609
272282 94.455 -73.507 53.582
272283 94.481 -73.514 53.556
272284 94.506 -73.521 53.528
272285 94.532 -73.529 53.501
272286 94.558 -73.536 53.475
272287 94.583 -73.543 53.450
272288 94.609 -73.551 53.425
272289 94.635 -73.559 53.400
272290 94.661 -73.565 53.377
272291 94.687 -73.573 53.354
272292 94.712 -73.581 53.333
272293 94.735 -73.587 53.311
272294 94.742 -73.585 53.287
272295 94.749 -73.584 53.263
272296 94.760 -73.585 53.237
272297 94.780 -73.589 53.211
272298 94.801 -73.593 53.182
272299 94.821 -73.598 53.152
272300 94.859 -73.590 53.122
272301 94.899 -73.581 53.089
272302 94.914 -73.594 53.055
272303 94.928 -73.607 53.021
272304 94.943 -73.621 52.987
272305 94.957 -73.634 52.952
272306 94.972 -73.647 52.918
272307 94.986 -73.660 52.884
272308 95.004 -73.669 52.850
272309 95.030 -73.672 52.817
272310 95.055 -73.673 52.785
272311 95.079 -73.675 52.753
272312 95.104 -73.666 52.722
272313 95.126 -73.637 52.692
272314 95.148 -73.609 52.662
272315 95.184 -73.605 52.631
272316 95.227 -73.612 52.601
272317 95.270 -73.618 52.572
272318 95.313 -73.625 52.543
272319 95.356 -73.633 52.514
272320 95.399 -73.640 52.483
272321 95.438 -73.650 52.451
272322 95.469 -73.664 52.419
272323 95.501 -73.678 52.388
272324 95.532 -73.685 52.359
272325 95.558 -73.673 52.332
272326 95.585 -73.661 52.304
272327 95.611 -73.650 52.277
272328 95.638 -73.634 52.251
272329 95.666 -73.612 52.225
272330 95.693 -73.590 52.201
272331 95.721 -73.568 52.175
272332 95.753 -73.550 52.148
272333 95.792 -73.542 52.119
272334 95.822 -73.556 52.089
272335 95.851 -73.568 52.059
272336 95.881 -73.581 52.029
272337 95.911 -73.595 51.999
272338 95.941 -73.608 51.967
272339 95.971 -73.622 51.935
272340 96.001 -73.634 51.903
272341 96.035 -73.639 51.868
272342 96.072 -73.639 51.834
272343 96.108 -73.639 51.799
272344 96.117 -73.636 51.764
272345 96.131 -73.626 51.729
272346 96.148 -73.612 51.694
272347 96.141 -73.612 51.660
272348 96.143 -73.614 51.627
272349 96.190 -73.634 51.593
272350 96.237 -73.654 51.559
272351 96.285 -73.675 51.526
272352 96.331 -73.695 51.493
272353 96.361 -73.707 51.461
272354 96.382 -73.716 51.429
272355 96.404 -73.725 51.397
272356 96.426 -73.734 51.367
272357 96.447 -73.743 51.336
272358 96.469 -73.751 51.306
272359 96.490 -73.760 51.276
272360 96.512 -73.769 51.246
272361 96.526 -73.766 51.216
272362 96.539 -73.760 51.188
272363 96.552 -73.754 51.161
272364 96.576 -73.748 51.135
272365 96.648 -73.741 51.109
272366 96.689 -73.751 51.084
272367 96.727 -73.757 51.062
272368 96.761 -73.754 51.040
272369 96.795 -73.751 51.021
272370 96.831 -73.750 51.002
272371 96.876 -73.756 50.985
272372 96.920 -73.762 50.969
272373 96.934 -73.777 50.954
272374 96.945 -73.793 50.942
272375 96.957 -73.810 50.931
272376 96.968 -73.825 50.921
272377 96.979 -73.841 50.913
272378 96.996 -73.848 50.906
272379 97.028 -73.839 50.900
272380 97.061 -73.829 50.896
272381 97.092 -73.820 50.892
272382 97.124 -73.810 50.891
272383 97.157 -73.801 50.890
272384 97.189 -73.792 50.888
272385 97.221 -73.785 50.887
272386 97.240 -73.788 50.885
272387 97.259 -73.794 50.882
272388 97.279 -73.799 50.880
272389 97.298 -73.805 50.876
272390 97.318 -73.810 50.868
272391 97.337 -73.816 50.860
272392 97.357 -73.822 50.853
272393 97.376 -73.828 50.846
272394 97.396 -73.833 50.839
272395 97.416 -73.839 50.832
272396 97.441 -73.844 50.822
272397 97.480 -73.849 50.807
272398 97.521 -73.854 50.783
272399 97.561 -73.860 50.751
272400 97.600 -73.865 50.714
272401 97.641 -73.869 50.671
272402 97.680 -73.875 50.627
272403 97.721 -73.880 50.578
272404 97.760 -73.885 50.527
272405 97.800 -73.891 50.476
272406 97.826 -73.907 50.428
272407 97.853 -73.925 50.380
272408 97.879 -73.941 50.333
272409 97.905 -73.958 50.289
272410 97.932 -73.976 50.240
272411 97.962 -73.991 50.191
272412 98.003 -74.001 50.140
272413 98.045 -74.011 50.088
272414 98.087 -74.022 50.031
272415 98.129 -74.032 49.972
272416 98.169 -74.043 49.911
272417 98.197 -74.054 49.847
272418 98.224 -74.066 49.782
272419 98.251 -74.077 49.717
272420 98.279 -74.089 49.652
272421 98.306 -74.100 49.589
272422 98.334 -74.112 49.525
272423 98.364 -74.124 49.459
272424 98.370 -74.134 49.396
272425 98.375 -74.146 49.335
272426 98.416 -74.154 49.279
272427 98.450 -74.133 49.222
272428 98.480 -74.129 49.166
272429 98.508 -74.139 49.108
272430 98.566 -74.163 49.050
272431 98.610 -74.206 48.992
272432 98.646 -74.230 48.937
272433 98.677 -74.249 48.884
272434 98.704 -74.270 48.833
272435 98.727 -74.298 48.785
272436 98.751 -74.327 48.736
272437 98.774 -74.334 48.686
272438 98.804 -74.341 48.636
272439 98.835 -74.349 48.586
272440 98.865 -74.355 48.536
272441 98.896 -74.363 48.485
272442 98.927 -74.371 48.434
272443 98.959 -74.380 48.382
272444 98.992 -74.389 48.330
272445 99.024 -74.395 48.278
272446 99.057 -74.399 48.226
272447 99.089 -74.405 48.175
272448 99.121 -74.409 48.125
272449 99.153 -74.415 48.075
272450 99.184 -74.422 48.028
272451 99.208 -74.443 47.981
272452 99.232 -74.463 47.935
272453 99.257 -74.483 47.891
272454 99.281 -74.505 47.845
272455 99.312 -74.523 47.801
272456 99.353 -74.537 47.758
272457 99.394 -74.552 47.716
272458 99.435 -74.567 47.675
272459 99.475 -74.581 47.635
272460 99.505 -74.590 47.597
272461 99.535 -74.598 47.559
272462 99.564 -74.607 47.521
272463 99.594 -74.615 47.487
272464 99.624 -74.624 47.451
272465 99.654 -74.632 47.415
272466 99.684 -74.640 47.378
272467 99.713 -74.649 47.342
272468 99.743 -74.657 47.306
272469 99.773 -74.665 47.271
272470 99.802 -74.674 47.238
272471 99.833 -74.682 47.205
272472 99.868 -74.686 47.174
272473 99.902 -74.697 47.142
272474 99.935 -74.712 47.111
272475 99.969 -74.727 47.080
272476 100.002 -74.742 47.051
272477 100.036 -74.756 47.024
272478 100.069 -74.771 46.996
272479 100.102 -74.785 46.970
272480 100.136 -74.800 46.943
272481 100.169 -74.814 46.917
272482 100.203 -74.828 46.891
272483 100.236 -74.844 46.868
272484 100.269 -74.858 46.845
272485 100.294 -74.867 46.821
272486 100.320 -74.877 46.795
272487 100.345 -74.887 46.766
272488 100.370 -74.896 46.738
272489 100.395 -74.906 46.709
272490 100.420 -74.915 46.682
272491 100.445 -74.926 46.655
272492 100.470 -74.938 46.628
272493 100.495 -74.950 46.599
272494 100.520 -74.963 46.571
272495 100.545 -74.975 46.542
272496 100.570 -74.988 46.514
272497 100.595 -75.004 46.486
272498 100.614 -75.021 46.458
272499 100.635 -75.035 46.429
272500 100.662 -75.046 46.400
272501 100.690 -75.057 46.371
272502 100.717 -75.067 46.341
272503 100.745 -75.078 46.314
272504 100.772 -75.088 46.287
272505 100.799 -75.100 46.260
272506 100.827 -75.110 46.235
272507 100.854 -75.116 46.208
272508 100.889 -75.112 46.180
272509 100.925 -75.116 46.151
272510 100.949 -75.132 46.121
272511 100.973 -75.148 46.091
272512 100.997 -75.165 46.060
272513 101.025 -75.178 46.029
272514 101.053 -75.190 45.999
272515 101.082 -75.203 45.966
272516 101.111 -75.215 45.934
272517 101.140 -75.227 45.902
272518 101.169 -75.239 45.868
272519 101.191 -75.240 45.834
272520 101.215 -75.242 45.799
272521 101.250 -75.261 45.764
272522 101.284 -75.281 45.731
272523 101.319 -75.301 45.698
272524 101.353 -75.320 45.667
272525 101.388 -75.339 45.635
272526 101.422 -75.359 45.602
272527 101.457 -75.378 45.571
272528 101.491 -75.398 45.539
272529 101.526 -75.417 45.508
272530 101.560 -75.436 45.477
272531 101.595 -75.456 45.446
272532 101.629 -75.475 45.415
272533 101.663 -75.495 45.385
272534 101.698 -75.514 45.356
272535 101.732 -75.531 45.326
272536 101.758 -75.546 45.297
272537 101.775 -75.547 45.268
272538 101.793 -75.547 45.240
272539 101.810 -75.548 45.212
272540 101.828 -75.549 45.184
272541 101.845 -75.549 45.156
272542 101.863 -75.549 45.130
272543 101.880 -75.550 45.104
272544 101.898 -75.550 45.078
272545 101.915 -75.551 45.053
272546 101.933 -75.552 45.028
272547 101.950 -75.552 45.002
272548 101.967 -75.552 44.977
272549 101.985 -75.553 44.952
272550 102.003 -75.554 44.927
272901 92.257 -73.150 54.784
272902 92.287 -73.161 54.771
272903 92.317 -73.170 54.759
272904 92.348 -73.181 54.746
272905 92.379 -73.191 54.733
272906 92.410 -73.202 54.720
272907 92.440 -73.212 54.707
272908 92.471 -73.222 54.695
272909 92.502 -73.233 54.683
272910 92.532 -73.243 54.670
272911 92.563 -73.253 54.659
272912 92.593 -73.264 54.648
272913 92.624 -73.274 54.636
272914 92.655 -73.284 54.627
272915 92.686 -73.295 54.618
272916 92.726 -73.305 54.610
272917 92.769 -73.317 54.604
272918 92.812 -73.327 54.598
272919 92.855 -73.339 54.594
272920 92.898 -73.349 54.591
272921 92.940 -73.361 54.590
272922 92.984 -73.371 54.589
272923 93.026 -73.383 54.588
272924 93.069 -73.393 54.586
272925 93.112 -73.405 54.582
272926 93.149 -73.410 54.579
272927 93.171 -73.401 54.576
272928 93.194 -73.393 54.575
272929 93.216 -73.384 54.573
272930 93.252 -73.393 54.570
272931 93.293 -73.407 54.566
272932 93.334 -73.421 54.562
272933 93.374 -73.435 54.558
272934 93.408 -73.447 54.554
272935 93.440 -73.459 54.549
272936 93.460 -73.468 54.542
272937 93.478 -73.474 54.534
272938 93.499 -73.477 54.520
272939 93.520 -73.480 54.506
272940 93.541 -73.482 54.491
272941 93.561 -73.481 54.478
272942 93.581 -73.478 54.465
272943 93.602 -73.476 54.453
272944 93.622 -73.474 54.440
272945 93.642 -73.471 54.424
272946 93.661 -73.469 54.409
272947 93.682 -73.467 54.395
272948 93.701 -73.462 54.380
272949 93.719 -73.457 54.365
272950 93.738 -73.452 54.351
272951 93.757 -73.446 54.333
272952 93.776 -73.441 54.313
272953 93.795 -73.436 54.293
272954 93.814 -73.430 54.271
272955 93.833 -73.425 54.250
272956 93.851 -73.420 54.229
272957 93.870 -73.415 54.207
272958 93.889 -73.410 54.183
272959 93.908 -73.405 54.159
272960 93.927 -73.399 54.135
272961 93.945 -73.394 54.112
272962 93.965 -73.389 54.089
272963 93.983 -73.383 54.066
272964 94.002 -73.378 54.043
272965 94.021 -73.373 54.018
272966 94.040 -73.368 53.994
272967 94.059 -73.363 53.970
272968 94.077 -73.358 53.946
272969 94.097 -73.352 53.922
272970 94.117 -73.352 53.897
272971 94.144 -73.367 53.872
272972 94.170 -73.381 53.846
272973 94.196 -73.396 53.820
272974 94.222 -73.403 53.795
272975 94.248 -73.411 53.770
272976 94.274 -73.418 53.745
272977 94.299 -73.425 53.720
272978 94.325 -73.433 53.694
272979 94.351 -73.440 53.667
272980 94.377 -73.447 53.640
272981 94.402 -73.455 53.613
272982 94.428 -73.462 53.587
272983 94.454 -73.470 53.561
272984 94.479 -73.477 53.534
272985 94.505 -73.484 53.507
272986 94.531 -73.492 53.481
272987 94.557 -73.499 53.455
272988 94.582 -73.506 53.430
272989 94.608 -73.514 53.405
272990 94.634 -73.521 53.382
272991 94.660 -73.528 53.359
272992 94.685 -73.536 53.336
272993 94.711 -73.543 53.315
272994 94.737 -73.550 53.291
272995 94.753 -73.553 53.267
272996 94.760 -73.553 53.243
272997 94.767 -73.552 53.218
272998 94.783 -73.554 53.191
272999 94.804 -73.559 53.162
273000 94.824 -73.563 53.132
273001 94.845 -73.568 53.102
273002 94.881 -73.561 53.069
273003 94.921 -73.551 53.037
273004 94.936 -73.565 53.004
273005 94.950 -73.578 52.970
273006 94.965 -73.591 52.936
273007 94.979 -73.605 52.902
273008 94.993 -73.618 52.868
273009 95.008 -73.631 52.835
273010 95.022 -73.644 52.802
273011 95.039 -73.655 52.770
273012 95.064 -73.656 52.739
273013 95.090 -73.659 52.709
273014 95.114 -73.650 52.679
273015 95.136 -73.622 52.649
273016 95.166 -73.607 52.619
273017 95.209 -73.614 52.589
273018 95.252 -73.622 52.561
273019 95.296 -73.628 52.534
273020 95.338 -73.635 52.505
273021 95.382 -73.642 52.474
273022 95.424 -73.649 52.444
273023 95.467 -73.656 52.413
273024 95.508 -73.665 52.384
273025 95.536 -73.661 52.357
273026 95.563 -73.640 52.332
273027 95.591 -73.618 52.305
273028 95.618 -73.597 52.278
273029 95.645 -73.576 52.252
273030 95.672 -73.556 52.227
273031 95.699 -73.537 52.201
273032 95.726 -73.517 52.176
273033 95.753 -73.502 52.149
273034 95.779 -73.512 52.121
273035 95.809 -73.524 52.094
273036 95.839 -73.538 52.065
273037 95.869 -73.551 52.035
273038 95.899 -73.565 52.005
273039 95.929 -73.578 51.974
273040 95.959 -73.590 51.943
273041 95.988 -73.604 51.910
273042 96.022 -73.610 51.876
273043 96.058 -73.609 51.841
273044 96.095 -73.609 51.806
273045 96.121 -73.609 51.771
273046 96.130 -73.606 51.735
273047 96.145 -73.595 51.700
273048 96.144 -73.590 51.667
273049 96.164 -73.603 51.634
273050 96.203 -73.622 51.600
273051 96.248 -73.643 51.566
273052 96.296 -73.662 51.533
273053 96.342 -73.683 51.500
273054 96.379 -73.698 51.467
273055 96.400 -73.707 51.435
273056 96.422 -73.716 51.403
273057 96.444 -73.725 51.372
273058 96.465 -73.734 51.342
273059 96.486 -73.742 51.311
273060 96.499 -73.736 51.281
273061 96.512 -73.730 51.251
273062 96.525 -73.725 51.221
273063 96.538 -73.719 51.193
273064 96.551 -73.713 51.166
273065 96.593 -73.706 51.140
273066 96.669 -73.697 51.113
273067 96.716 -73.703 51.089
273068 96.757 -73.714 51.065
273069 96.797 -73.722 51.044
273070 96.830 -73.719 51.025
273071 96.870 -73.722 51.005
273072 96.914 -73.727 50.986
273073 96.959 -73.734 50.970
273074 96.965 -73.738 50.954
273075 96.959 -73.753 50.943
273076 96.953 -73.768 50.932
273077 96.947 -73.783 50.922
273078 96.964 -73.796 50.913
273079 96.994 -73.808 50.906
273080 97.022 -73.812 50.900
273081 97.053 -73.803 50.895
273082 97.086 -73.794 50.890
273083 97.117 -73.784 50.887
273084 97.150 -73.775 50.885
273085 97.182 -73.766 50.883
273086 97.216 -73.763 50.880
273087 97.244 -73.764 50.879
273088 97.249 -73.769 50.878
273089 97.269 -73.775 50.876
273090 97.289 -73.781 50.873
273091 97.308 -73.786 50.865
273092 97.327 -73.792 50.858
273093 97.347 -73.797 50.850
273094 97.367 -73.803 50.844
273095 97.386 -73.809 50.838
273096 97.407 -73.814 50.831
273097 97.447 -73.819 50.821
273098 97.488 -73.825 50.805
273099 97.527 -73.829 50.781
273100 97.568 -73.835 50.750
273101 97.607 -73.840 50.713
273102 97.647 -73.844 50.672
273103 97.687 -73.850 50.627
273104 97.727 -73.855 50.578
273105 97.768 -73.860 50.527
273106 97.807 -73.865 50.476
273107 97.838 -73.879 50.427
273108 97.864 -73.895 50.378
273109 97.891 -73.912 50.332
273110 97.917 -73.929 50.287
273111 97.943 -73.946 50.236
273112 97.980 -73.959 50.184
273113 98.021 -73.969 50.131
273114 98.063 -73.980 50.077
273115 98.104 -73.990 50.019
273116 98.146 -74.001 49.958
273117 98.187 -74.011 49.896
273118 98.229 -74.021 49.832
273119 98.266 -74.032 49.765
273120 98.293 -74.044 49.699
273121 98.320 -74.055 49.635
273122 98.347 -74.067 49.571
273123 98.386 -74.081 49.509
273124 98.382 -74.089 49.443
273125 98.376 -74.098 49.378
273126 98.381 -74.111 49.316
273127 98.428 -74.140 49.259
273128 98.464 -74.127 49.201
273129 98.499 -74.107 49.145
273130 98.533 -74.085 49.088
273131 98.573 -74.095 49.030
273132 98.625 -74.157 48.974
273133 98.662 -74.186 48.919
273134 98.686 -74.212 48.867
273135 98.710 -74.241 48.816
273136 98.733 -74.269 48.769
273137 98.757 -74.298 48.720
273138 98.780 -74.317 48.671
273139 98.806 -74.310 48.622
273140 98.836 -74.317 48.572
273141 98.867 -74.324 48.522
273142 98.898 -74.332 48.472
273143 98.931 -74.341 48.422
273144 98.963 -74.350 48.370
273145 98.996 -74.359 48.318
273146 99.028 -74.368 48.266
273147 99.061 -74.373 48.214
273148 99.093 -74.378 48.163
273149 99.125 -74.383 48.113
273150 99.157 -74.388 48.064
273151 99.189 -74.393 48.017
273152 99.213 -74.414 47.970
273153 99.238 -74.435 47.925
273154 99.262 -74.455 47.881
273155 99.286 -74.476 47.836
273156 99.327 -74.491 47.793
273157 99.368 -74.505 47.750
273158 99.409 -74.521 47.708
273159 99.449 -74.535 47.668
273160 99.491 -74.550 47.628
273161 99.523 -74.560 47.589
273162 99.553 -74.568 47.552
273163 99.582 -74.577 47.515
273164 99.612 -74.585 47.480
273165 99.642 -74.593 47.445
273166 99.672 -74.602 47.409
273167 99.702 -74.610 47.373
273168 99.731 -74.619 47.336
273169 99.761 -74.628 47.300
273170 99.791 -74.636 47.265
273171 99.821 -74.644 47.231
273172 99.851 -74.653 47.200
273173 99.894 -74.663 47.168
273174 99.913 -74.682 47.137
273175 99.946 -74.697 47.105
273176 99.980 -74.712 47.074
273177 100.013 -74.726 47.046
273178 100.047 -74.740 47.018
273179 100.080 -74.755 46.991
273180 100.114 -74.769 46.965
273181 100.147 -74.784 46.938
273182 100.181 -74.798 46.912
273183 100.214 -74.812 46.886
273184 100.248 -74.827 46.863
273185 100.278 -74.840 46.840
273186 100.303 -74.849 46.815
273187 100.329 -74.859 46.787
273188 100.354 -74.869 46.759
273189 100.379 -74.878 46.729
273190 100.404 -74.888 46.701
273191 100.429 -74.897 46.674
273192 100.454 -74.907 46.646
273193 100.479 -74.916 46.618
273194 100.504 -74.925 46.589
273195 100.530 -74.935 46.561
273196 100.555 -74.945 46.532
273197 100.579 -74.957 46.503
273198 100.605 -74.970 46.476
273199 100.626 -74.986 46.448
273200 100.653 -74.997 46.419
273201 100.681 -75.007 46.390
273202 100.708 -75.018 46.360
273203 100.735 -75.029 46.330
273204 100.763 -75.039 46.303
273205 100.790 -75.050 46.276
273206 100.817 -75.061 46.249
273207 100.845 -75.072 46.224
273208 100.872 -75.078 46.198
273209 100.906 -75.082 46.169
273210 100.930 -75.098 46.140
273211 100.958 -75.111 46.109
273212 100.986 -75.124 46.079
273213 101.015 -75.136 46.048
273214 101.044 -75.148 46.016
273215 101.073 -75.160 45.985
273216 101.102 -75.173 45.952
273217 101.130 -75.185 45.920
273218 101.159 -75.198 45.887
273219 101.188 -75.210 45.853
273220 101.210 -75.210 45.820
273221 101.245 -75.229 45.785
273222 101.279 -75.248 45.749
273223 101.314 -75.267 45.714
273224 101.348 -75.287 45.681
273225 101.383 -75.307 45.650
273226 101.417 -75.326 45.618
273227 101.452 -75.345 45.587
273228 101.486 -75.365 45.555
273229 101.520 -75.384 45.524
273230 101.555 -75.404 45.493
273231 101.589 -75.422 45.462
273232 101.622 -75.441 45.431
273233 101.656 -75.459 45.401
273234 101.689 -75.478 45.371
273235 101.723 -75.496 45.341
273236 101.755 -75.508 45.312
273237 101.773 -75.506 45.283
273238 101.790 -75.507 45.254
273239 101.808 -75.508 45.226
273240 101.825 -75.508 45.197
273241 101.843 -75.508 45.170
273242 101.860 -75.509 45.142
273243 101.878 -75.510 45.114
273244 101.895 -75.510 45.089
273245 101.912 -75.511 45.063
273246 101.930 -75.511 45.037
273247 101.948 -75.512 45.011
273248 101.965 -75.512 44.986
273249 101.982 -75.513 44.961
273250 102.000 -75.514 44.936
273602 92.252 -73.120 54.791
273603 92.283 -73.131 54.778
273604 92.313 -73.142 54.765
273605 92.344 -73.152 54.752
273606 92.375 -73.162 54.738
273607 92.406 -73.173 54.725
273608 92.436 -73.183 54.712
273609 92.467 -73.193 54.699
273610 92.497 -73.203 54.686
273611 92.528 -73.214 54.673
273612 92.559 -73.224 54.661
273613 92.589 -73.234 54.650
273614 92.620 -73.245 54.638
273615 92.651 -73.255 54.628
273616 92.682 -73.265 54.618
273617 92.723 -73.277 54.610
273618 92.766 -73.287 54.603
273619 92.808 -73.299 54.598
273620 92.852 -73.309 54.594
273621 92.895 -73.321 54.591
273622 92.937 -73.331 54.589
273623 92.980 -73.343 54.586
273624 93.023 -73.353 54.584
273625 93.066 -73.365 54.581
273626 93.109 -73.375 54.577
273627 93.152 -73.387 54.572
273628 93.183 -73.386 54.569
273629 93.205 -73.377 54.567
273630 93.227 -73.368 54.564
273631 93.262 -73.375 54.562
273632 93.302 -73.390 54.557
273633 93.343 -73.403 54.553
273634 93.384 -73.418 54.549
273635 93.423 -73.431 54.544
273636 93.449 -73.441 54.539
273637 93.468 -73.448 54.532
273638 93.486 -73.456 54.525
273639 93.506 -73.459 54.510
273640 93.528 -73.461 54.494
273641 93.548 -73.459 54.479
273642 93.568 -73.457 54.465
273643 93.588 -73.455 54.452
273644 93.608 -73.452 54.439
273645 93.627 -73.449 54.426
273646 93.646 -73.443 54.410
273647 93.665 -73.438 54.394
273648 93.684 -73.433 54.379
273649 93.703 -73.427 54.364
273650 93.721 -73.422 54.348
273651 93.741 -73.417 54.332
273652 93.759 -73.412 54.313
273653 93.778 -73.407 54.293
273654 93.797 -73.402 54.272
273655 93.816 -73.396 54.250
273656 93.835 -73.391 54.228
273657 93.853 -73.386 54.207
273658 93.872 -73.380 54.185
273659 93.891 -73.375 54.160
273660 93.910 -73.370 54.135
273661 93.929 -73.365 54.112
273662 93.948 -73.360 54.088
273663 93.967 -73.355 54.065
273664 93.985 -73.349 54.041
273665 94.004 -73.344 54.018
273666 94.023 -73.339 53.994
273667 94.042 -73.333 53.969
273668 94.061 -73.328 53.945
273669 94.080 -73.324 53.921
273670 94.099 -73.318 53.896
273671 94.117 -73.313 53.872
273672 94.142 -73.322 53.848
273673 94.168 -73.336 53.822
273674 94.195 -73.351 53.796
273675 94.221 -73.365 53.770
273676 94.247 -73.374 53.745
273677 94.272 -73.380 53.721
273678 94.298 -73.388 53.697
273679 94.324 -73.396 53.671
273680 94.350 -73.403 53.644
273681 94.376 -73.410 53.618
273685 94.478 -73.440 53.514
273686 94.504 -73.447 53.487
273687 94.530 -73.455 53.462
273688 94.555 -73.462 53.436
273689 94.581 -73.469 53.410
273690 94.607 -73.477 53.386
273691 94.633 -73.484 53.363
273692 94.659 -73.491 53.341
273693 94.684 -73.499 53.318
273694 94.710 -73.506 53.296
273695 94.736 -73.514 53.272
273696 94.761 -73.521 53.248
273697 94.770 -73.521 53.223
273698 94.777 -73.520 53.198
273699 94.786 -73.519 53.172
273700 94.807 -73.524 53.143
273701 94.827 -73.528 53.113
273702 94.848 -73.533 53.083
273703 94.868 -73.537 53.051
273704 94.902 -73.532 53.019
273705 94.943 -73.521 52.986
273706 94.958 -73.535 52.954
273707 94.972 -73.549 52.920
273708 94.986 -73.562 52.887
273709 95.001 -73.575 52.854
273710 95.015 -73.588 52.821
273711 95.030 -73.602 52.788
273712 95.044 -73.615 52.757
273713 95.058 -73.628 52.727
273714 95.074 -73.640 52.697
273715 95.100 -73.642 52.668
273716 95.124 -73.634 52.637
273717 95.148 -73.609 52.608
273718 95.191 -73.617 52.579
273719 95.235 -73.624 52.551
273720 95.277 -73.631 52.524
273721 95.321 -73.637 52.495
273722 95.364 -73.644 52.466
273723 95.408 -73.647 52.436
273724 95.455 -73.637 52.407
273725 95.502 -73.629 52.379
273726 95.534 -73.628 52.354
273727 95.561 -73.609 52.331
273728 95.588 -73.590 52.305
273729 95.615 -73.570 52.278
273730 95.642 -73.550 52.253
273731 95.669 -73.530 52.228
273732 95.695 -73.510 52.202
273733 95.722 -73.490 52.177
273734 95.750 -73.478 52.152
273735 95.775 -73.494 52.126
273736 95.797 -73.494 52.099
273737 95.827 -73.507 52.071
273738 95.857 -73.521 52.042
273739 95.887 -73.534 52.013
273740 95.916 -73.547 51.982
273741 95.946 -73.560 51.950
273742 95.976 -73.574 51.918
273743 96.009 -73.581 51.885
273744 96.045 -73.581 51.848
273745 96.082 -73.581 51.813
273746 96.118 -73.581 51.778
273747 96.135 -73.578 51.743
273748 96.144 -73.575 51.709
273749 96.153 -73.571 51.675
273750 96.191 -73.591 51.642
273751 96.230 -73.611 51.608
273752 96.268 -73.631 51.575
273753 96.307 -73.651 51.542
273754 96.354 -73.671 51.509
273755 96.397 -73.690 51.475
273756 96.418 -73.698 51.443
273757 96.440 -73.707 51.411
273758 96.459 -73.712 51.380
273759 96.472 -73.706 51.348
273760 96.485 -73.700 51.317
273761 96.498 -73.694 51.287
273762 96.511 -73.688 51.258
273763 96.524 -73.683 51.228
273764 96.536 -73.677 51.200
273765 96.549 -73.671 51.172
273766 96.611 -73.662 51.145
273767 96.686 -73.653 51.118
273768 96.743 -73.656 51.094
273769 96.784 -73.666 51.070
273770 96.825 -73.677 51.048
273771 96.866 -73.687 51.029
273772 96.905 -73.699 51.008
273773 96.943 -73.714 50.989
273774 96.955 -73.689 50.972
273775 96.940 -73.626 50.957
273776 96.914 -73.709 50.944
273777 96.908 -73.724 50.933
273778 96.934 -73.736 50.924
273779 96.963 -73.749 50.914
273780 96.992 -73.761 50.907
273781 97.022 -73.773 50.900
273782 97.049 -73.772 50.894
273783 97.079 -73.767 50.889
273784 97.111 -73.757 50.885
273785 97.143 -73.748 50.881
273786 97.175 -73.739 50.878
273787 97.210 -73.741 50.874
273788 97.244 -73.741 50.873
273789 97.252 -73.745 50.873
273790 97.259 -73.750 50.872
273791 97.279 -73.756 50.869
273792 97.298 -73.762 50.862
273793 97.317 -73.767 50.856
273794 97.337 -73.773 50.849
273795 97.357 -73.778 50.845
273796 97.376 -73.785 50.839
273797 97.414 -73.790 50.831
273798 97.454 -73.794 50.820
273799 97.494 -73.800 50.802
273800 97.534 -73.805 50.776
273801 97.574 -73.810 50.744
273802 97.614 -73.815 50.708
273803 97.654 -73.820 50.667
273804 97.694 -73.825 50.622
273805 97.734 -73.830 50.574
273806 97.774 -73.835 50.524
273807 97.814 -73.840 50.474
273808 97.850 -73.849 50.424
273809 97.876 -73.866 50.376
273810 97.902 -73.883 50.330
273811 97.929 -73.900 50.284
273812 97.955 -73.917 50.232
273813 97.997 -73.927 50.177
273814 98.039 -73.938 50.121
273815 98.080 -73.948 50.064
273816 98.122 -73.959 50.005
273817 98.164 -73.969 49.944
273818 98.206 -73.979 49.880
273819 98.247 -73.990 49.816
273820 98.289 -74.000 49.749
273821 98.331 -74.010 49.683
273822 98.369 -74.023 49.618
273823 98.400 -74.037 49.552
273824 98.394 -74.045 49.490
273825 98.388 -74.054 49.426
273826 98.382 -74.062 49.363
273827 98.388 -74.075 49.302
273828 98.435 -74.104 49.241
273829 98.479 -74.122 49.182
273830 98.512 -74.101 49.125
273831 98.547 -74.079 49.069
273832 98.591 -74.092 49.011
273833 98.640 -74.137 48.956
273834 98.674 -74.157 48.902
273835 98.692 -74.183 48.849
273836 98.716 -74.212 48.800
273837 98.739 -74.240 48.753
273838 98.763 -74.269 48.705
273839 98.786 -74.298 48.657
273840 98.807 -74.278 48.608
273841 98.838 -74.286 48.559
273842 98.869 -74.293 48.509
273843 98.902 -74.302 48.459
273844 98.934 -74.311 48.409
273845 98.967 -74.320 48.357
273846 99.000 -74.329 48.306
273847 99.032 -74.337 48.254
273848 99.065 -74.346 48.203
273849 99.097 -74.352 48.153
273850 99.129 -74.357 48.104
273851 99.161 -74.361 48.055
273852 99.194 -74.367 48.008
273853 99.219 -74.386 47.962
273854 99.243 -74.406 47.916
273855 99.267 -74.427 47.871
273856 99.300 -74.444 47.827
273857 99.342 -74.459 47.784
273858 99.382 -74.474 47.742
273859 99.424 -74.489 47.700
273860 99.465 -74.503 47.660
273861 99.505 -74.518 47.620
273862 99.541 -74.530 47.582
273863 99.571 -74.538 47.545
273864 99.600 -74.546 47.509
273865 99.630 -74.555 47.474
273866 99.660 -74.564 47.438
273867 99.690 -74.572 47.402
273868 99.720 -74.581 47.365
273869 99.750 -74.589 47.329
273870 99.779 -74.597 47.293
273871 99.809 -74.606 47.258
273872 99.842 -74.614 47.225
273873 99.905 -74.619 47.194
273874 99.969 -74.624 47.163
273875 99.987 -74.644 47.131
273876 100.001 -74.665 47.100
273877 100.015 -74.687 47.069
273878 100.029 -74.708 47.040
273879 100.059 -74.724 47.013
273880 100.092 -74.738 46.986
273881 100.126 -74.753 46.960
273882 100.159 -74.767 46.933
273883 100.192 -74.782 46.907
273884 100.226 -74.797 46.881
273885 100.259 -74.811 46.858
273886 100.287 -74.822 46.834
273887 100.312 -74.831 46.808
273888 100.337 -74.841 46.780
273889 100.362 -74.850 46.751
273890 100.388 -74.860 46.721
273891 100.413 -74.869 46.692
273892 100.437 -74.879 46.666
273893 100.463 -74.888 46.637
273894 100.488 -74.898 46.608
273895 100.513 -74.908 46.579
273896 100.538 -74.917 46.550
273897 100.563 -74.927 46.521
273898 100.591 -74.935 46.493
273899 100.617 -74.942 46.465
273900 100.643 -74.950 46.437
273901 100.668 -74.959 46.409
273902 100.696 -74.969 46.380
273903 100.723 -74.980 46.350
273904 100.751 -74.991 46.319
273905 100.779 -75.001 46.292
273906 100.807 -75.012 46.266
273907 100.834 -75.022 46.239
273908 100.861 -75.033 46.213
273909 100.890 -75.045 46.187
273910 100.919 -75.057 46.158
273911 100.948 -75.069 46.127
273912 100.977 -75.082 46.097
273913 101.005 -75.094 46.067
273914 101.034 -75.107 46.036
273915 101.063 -75.119 46.004
273916 101.092 -75.131 45.971
273917 101.121 -75.143 45.938
273918 101.149 -75.156 45.907
273919 101.178 -75.168 45.875
273920 101.206 -75.179 45.841
273921 101.240 -75.196 45.806
273922 101.274 -75.216 45.770
273923 101.308 -75.234 45.734
273924 101.341 -75.253 45.698
273925 101.375 -75.271 45.666
273926 101.409 -75.290 45.634
273927 101.443 -75.308 45.602
273928 101.476 -75.327 45.570
273929 101.510 -75.345 45.539
273930 101.543 -75.364 45.509
273931 101.577 -75.383 45.478
273932 101.611 -75.401 45.447
273933 101.644 -75.420 45.417
273934 101.678 -75.438 45.387
273935 101.712 -75.457 45.357
273936 101.745 -75.474 45.327
273937 101.770 -75.466 45.298
273938 101.788 -75.467 45.268
273939 101.805 -75.467 45.240
273940 101.823 -75.467 45.211
273941 101.840 -75.468 45.183
273942 101.857 -75.469 45.155
273943 101.875 -75.470 45.127
273944 101.893 -75.470 45.099
273945 101.910 -75.471 45.073
273946 101.927 -75.471 45.048
273947 101.945 -75.471 45.022
273948 101.962 -75.472 44.995
273949 101.980 -75.473 44.970
273950 101.997 -75.474 44.945
273951 102.015 -75.474 44.920
274305 92.309 -73.112 54.771
274306 92.340 -73.123 54.757
274307 92.371 -73.133 54.744
274308 92.402 -73.143 54.730
274309 92.432 -73.154 54.717
274310 92.462 -73.164 54.704
274311 92.493 -73.174 54.689
274312 92.524 -73.185 54.676
274313 92.555 -73.195 54.664
274314 92.585 -73.205 54.653
274315 92.616 -73.216 54.640
274316 92.647 -73.226 54.630
274317 92.677 -73.236 54.620
274318 92.720 -73.247 54.611
274319 92.763 -73.258 54.603
274320 92.805 -73.269 54.597
274321 92.848 -73.280 54.592
274322 92.891 -73.291 54.589
274323 92.934 -73.302 54.587
274324 92.977 -73.313 54.582
274325 93.020 -73.324 54.578
274326 93.063 -73.335 54.574
274327 93.106 -73.346 54.569
274328 93.148 -73.357 54.565
274329 93.192 -73.368 54.561
274330 93.216 -73.362 54.557
274331 93.238 -73.353 54.554
274332 93.271 -73.358 54.552
274333 93.312 -73.372 54.547
274334 93.353 -73.386 54.542
274335 93.394 -73.400 54.538
274336 93.431 -73.413 54.534
274337 93.455 -73.421 54.528
274338 93.475 -73.429 54.522
274339 93.493 -73.436 54.516
274340 93.514 -73.440 54.501
274341 93.534 -73.437 54.484
274342 93.554 -73.435 54.468
274343 93.573 -73.430 54.453
274344 93.592 -73.424 54.439
274345 93.611 -73.419 54.426
274346 93.630 -73.414 54.413
274347 93.648 -73.409 54.396
274348 93.667 -73.404 54.379
274349 93.686 -73.399 54.363
274350 93.705 -73.393 54.347
274351 93.724 -73.388 54.330
274352 93.743 -73.383 54.314
274353 93.762 -73.377 54.294
274354 93.780 -73.372 54.273
274355 93.799 -73.367 54.252
274356 93.818 -73.362 54.229
274357 93.837 -73.357 54.207
274358 93.856 -73.352 54.185
274359 93.874 -73.346 54.162
274360 93.894 -73.341 54.137
274361 93.912 -73.336 54.112
274362 93.931 -73.330 54.088
274363 93.950 -73.325 54.064
274364 93.969 -73.320 54.041
274365 93.988 -73.315 54.018
274366 94.006 -73.310 53.994
274367 94.026 -73.305 53.970
274368 94.044 -73.299 53.945
274369 94.063 -73.294 53.920
274370 94.082 -73.289 53.896
274371 94.101 -73.283 53.872
274372 94.120 -73.278 53.848
274373 94.140 -73.277 53.824
274374 94.166 -73.291 53.798
274375 94.193 -73.305 53.773
274376 94.219 -73.320 53.747
274377 94.246 -73.334 53.723
274378 94.271 -73.343 53.698
274379 94.297 -73.351 53.674
274380 94.323 -73.358 53.648
274389 94.554 -73.424 53.418
274390 94.580 -73.432 53.393
274391 94.606 -73.440 53.368
274392 94.631 -73.447 53.345
274393 94.657 -73.454 53.322
274394 94.683 -73.462 53.300
274395 94.709 -73.469 53.277
274396 94.735 -73.477 53.254
274397 94.760 -73.484 53.230
274398 94.781 -73.489 53.205
274399 94.788 -73.488 53.178
274400 94.795 -73.487 53.152
274401 94.810 -73.489 53.124
274402 94.830 -73.493 53.095
274403 94.851 -73.498 53.065
274404 94.872 -73.503 53.033
274405 94.892 -73.506 53.001
274406 94.923 -73.503 52.969
274407 94.965 -73.493 52.937
274408 94.979 -73.506 52.904
274409 94.994 -73.519 52.872
274410 95.008 -73.532 52.840
274411 95.023 -73.546 52.807
274412 95.037 -73.559 52.775
274413 95.051 -73.572 52.744
274414 95.066 -73.586 52.715
274415 95.080 -73.599 52.686
274416 95.095 -73.612 52.657
274417 95.110 -73.625 52.627
274418 95.134 -73.617 52.598
274419 95.174 -73.619 52.569
274420 95.217 -73.626 52.542
274421 95.262 -73.624 52.514
274422 95.309 -73.615 52.487
274423 95.356 -73.606 52.459
274424 95.403 -73.597 52.430
274425 95.450 -73.588 52.402
274426 95.497 -73.580 52.376
274427 95.530 -73.578 52.352
274428 95.558 -73.579 52.329
274429 95.585 -73.563 52.304
274430 95.612 -73.543 52.278
274431 95.638 -73.523 52.253
274432 95.665 -73.503 52.228
274433 95.692 -73.484 52.203
274434 95.719 -73.464 52.179
274435 95.747 -73.455 52.156
274436 95.777 -73.484 52.131
274437 95.798 -73.484 52.105
274438 95.819 -73.484 52.077
274439 95.844 -73.490 52.049
274440 95.874 -73.503 52.020
274441 95.904 -73.516 51.989
274442 95.934 -73.530 51.957
274443 95.964 -73.543 51.926
274444 95.996 -73.552 51.892
274445 96.032 -73.552 51.856
274446 96.069 -73.551 51.821
274447 96.105 -73.551 51.786
274448 96.139 -73.551 51.752
274449 96.148 -73.547 51.717
274450 96.164 -73.552 51.683
274451 96.205 -73.573 51.650
274452 96.247 -73.595 51.616
274453 96.289 -73.616 51.584
274454 96.331 -73.638 51.551
274455 96.373 -73.659 51.518
274456 96.411 -73.679 51.484
274457 96.432 -73.681 51.452
274458 96.445 -73.675 51.420
274459 96.458 -73.670 51.389
274460 96.471 -73.664 51.358
274461 96.483 -73.659 51.326
274462 96.497 -73.653 51.295
274463 96.510 -73.647 51.265
274464 96.522 -73.641 51.235
274465 96.535 -73.635 51.207
274466 96.554 -73.629 51.180
274467 96.629 -73.620 51.152
274468 96.704 -73.611 51.126
274469 96.761 -73.622 51.101
274470 96.808 -73.642 51.077
274471 96.849 -73.660 51.055
274472 96.888 -73.678 51.035
274473 96.928 -73.694 51.014
274474 96.942 -73.674 50.995
274475 96.928 -73.611 50.977
274476 96.913 -73.547 50.961
274477 96.892 -73.594 50.949
274478 96.901 -73.684 50.938
274479 96.932 -73.689 50.928
274480 96.961 -73.701 50.918
274481 96.991 -73.713 50.909
274482 97.020 -73.726 50.901
274483 97.049 -73.738 50.894
274484 97.077 -73.734 50.889
274485 97.104 -73.731 50.886
274486 97.136 -73.722 50.882
274487 97.169 -73.717 50.878
274488 97.204 -73.719 50.873
274489 97.238 -73.720 50.871
274490 97.263 -73.722 50.870
274491 97.260 -73.727 50.869
274492 97.268 -73.731 50.866
274493 97.288 -73.738 50.861
274494 97.308 -73.743 50.854
274495 97.327 -73.749 50.846
274496 97.347 -73.754 50.840
274497 97.381 -73.760 50.834
274498 97.421 -73.765 50.826
274499 97.461 -73.770 50.813
274500 97.501 -73.775 50.795
274501 97.541 -73.780 50.768
274502 97.581 -73.785 50.737
274503 97.621 -73.790 50.701
274504 97.661 -73.795 50.661
274505 97.701 -73.800 50.618
274506 97.741 -73.805 50.571
274507 97.781 -73.810 50.522
274508 97.821 -73.816 50.472
274509 97.861 -73.820 50.422
274510 97.888 -73.837 50.374
274511 97.915 -73.854 50.328
274512 97.941 -73.871 50.281
274513 97.973 -73.885 50.226
274514 98.015 -73.896 50.169
274515 98.057 -73.906 50.110
274516 98.098 -73.916 50.050
274517 98.140 -73.927 49.990
274518 98.181 -73.937 49.929
274519 98.223 -73.948 49.866
274520 98.265 -73.958 49.801
274521 98.303 -73.966 49.736
274522 98.340 -73.974 49.671
274523 98.409 -73.995 49.605
274524 98.406 -74.001 49.539
274525 98.400 -74.009 49.475
274526 98.394 -74.017 49.412
274527 98.388 -74.026 49.350
274528 98.394 -74.039 49.289
274529 98.441 -74.069 49.228
274530 98.488 -74.098 49.167
274531 98.526 -74.095 49.109
274532 98.561 -74.074 49.052
274533 98.610 -74.098 48.996
274534 98.665 -74.144 48.941
274535 98.701 -74.133 48.887
274536 98.714 -74.159 48.835
274537 98.728 -74.185 48.786
274538 98.745 -74.212 48.739
274539 98.769 -74.240 48.691
274540 98.792 -74.269 48.643
274541 98.813 -74.259 48.594
274542 98.840 -74.254 48.546
274543 98.873 -74.263 48.496
274544 98.906 -74.272 48.447
274545 98.938 -74.280 48.397
274546 98.971 -74.289 48.345
274547 99.003 -74.299 48.294
274548 99.036 -74.308 48.243
274549 99.069 -74.316 48.192
274550 99.101 -74.325 48.143
274551 99.133 -74.330 48.094
274552 99.165 -74.336 48.045
274553 99.198 -74.340 47.999
274554 99.224 -74.358 47.953
274555 99.248 -74.378 47.907
274556 99.275 -74.398 47.862
274557 99.315 -74.413 47.819
274558 99.356 -74.427 47.776
274559 99.398 -74.442 47.734
274560 99.438 -74.457 47.693
274561 99.479 -74.471 47.653
274562 99.520 -74.487 47.614
274563 99.559 -74.499 47.575
274564 99.589 -74.508 47.538
274565 99.619 -74.517 47.502
274566 99.648 -74.525 47.468
274567 99.678 -74.534 47.432
274568 99.708 -74.542 47.395
274569 99.738 -74.550 47.358
274570 99.768 -74.559 47.322
274571 99.797 -74.567 47.287
274572 99.839 -74.574 47.251
274573 99.903 -74.580 47.219
274574 99.966 -74.584 47.188
274575 100.030 -74.590 47.157
274576 100.057 -74.604 47.125
274577 100.079 -74.625 47.095
274578 100.093 -74.646 47.064
274579 100.108 -74.668 47.035
274580 100.122 -74.689 47.008
274581 100.136 -74.711 46.982
274582 100.150 -74.732 46.955
274583 100.171 -74.751 46.928
274584 100.204 -74.766 46.902
274585 100.238 -74.780 46.877
274586 100.270 -74.794 46.853
274587 100.295 -74.803 46.828
274588 100.321 -74.813 46.801
274589 100.346 -74.823 46.772
274590 100.371 -74.832 46.743
274591 100.396 -74.842 46.713
274592 100.422 -74.850 46.684
274593 100.450 -74.858 46.656
274594 100.476 -74.866 46.628
274595 100.503 -74.873 46.599
274596 100.530 -74.881 46.569
274597 100.557 -74.888 46.541
274598 100.583 -74.897 46.512
274599 100.611 -74.904 46.484
274600 100.637 -74.912 46.456
274601 100.665 -74.919 46.427
274602 100.687 -74.926 46.398
274603 100.708 -74.932 46.370
274604 100.735 -74.943 46.341
274605 100.763 -74.954 46.310
274606 100.790 -74.964 46.282
274607 100.818 -74.975 46.255
274608 100.845 -74.985 46.228
274609 100.874 -74.997 46.203
274610 100.908 -75.018 46.175
274611 100.938 -75.033 46.145
274612 100.967 -75.045 46.114
274613 100.996 -75.057 46.084
274614 101.025 -75.069 46.054
274615 101.053 -75.082 46.022
274616 101.082 -75.094 45.990
274617 101.111 -75.106 45.958
274618 101.140 -75.118 45.925
274619 101.169 -75.130 45.893
274620 101.197 -75.142 45.860
274621 101.229 -75.157 45.826
274622 101.263 -75.176 45.791
274623 101.296 -75.195 45.754
274624 101.330 -75.214 45.718
274625 101.364 -75.232 45.682
274626 101.398 -75.250 45.650
274627 101.431 -75.269 45.618
274628 101.465 -75.287 45.586
274629 101.498 -75.306 45.554
274630 101.532 -75.324 45.523
274631 101.566 -75.343 45.493
274632 101.600 -75.361 45.463
274633 101.633 -75.380 45.433
274634 101.667 -75.398 45.403
274635 101.700 -75.417 45.373
274636 101.734 -75.436 45.343
274637 101.764 -75.435 45.313
274638 101.785 -75.427 45.284
274639 101.802 -75.427 45.254
274640 101.820 -75.427 45.225
274641 101.838 -75.428 45.197
274642 101.855 -75.429 45.168
274643 101.873 -75.430 45.139
274644 101.890 -75.430 45.111
274645 101.907 -75.430 45.085
274646 101.925 -75.431 45.058
274647 101.943 -75.431 45.032
274648 101.960 -75.432 45.006
274649 101.977 -75.433 44.980
274650 101.995 -75.433 44.954
274651 102.012 -75.433 44.929
275008 92.366 -73.104 54.750
275009 92.397 -73.114 54.736
275010 92.428 -73.124 54.722
275011 92.458 -73.135 54.708
275012 92.489 -73.145 54.694
275013 92.520 -73.155 54.680
275014 92.551 -73.166 54.668
275015 92.581 -73.176 54.656
275016 92.612 -73.186 54.643
275017 92.642 -73.197 54.631
275018 92.673 -73.207 54.621
275019 92.716 -73.218 54.611
275020 92.759 -73.229 54.603
275021 92.802 -73.240 54.595
275022 92.845 -73.251 54.589
275023 92.888 -73.262 54.585
275024 92.931 -73.273 54.582
275025 92.974 -73.284 54.576
275026 93.017 -73.295 54.571
275027 93.060 -73.306 54.566
275028 93.102 -73.317 54.561
275029 93.145 -73.328 54.556
275030 93.188 -73.339 54.552
275031 93.227 -73.346 54.548
275032 93.249 -73.337 54.544
275033 93.281 -73.340 54.541
275034 93.322 -73.355 54.536
275035 93.362 -73.368 54.531
275036 93.399 -73.381 54.525
275037 93.423 -73.390 54.520
275038 93.447 -73.399 54.514
275039 93.472 -73.407 54.507
275040 93.497 -73.415 54.500
275041 93.519 -73.411 54.486
275042 93.538 -73.405 54.471
275043 93.556 -73.401 54.455
275044 93.575 -73.396 54.440
275045 93.594 -73.390 54.426
275046 93.613 -73.385 54.413
275047 93.632 -73.380 54.399
275048 93.651 -73.374 54.381
275049 93.670 -73.369 54.364
275050 93.688 -73.364 54.347
275051 93.707 -73.358 54.330
275052 93.726 -73.354 54.312
275053 93.745 -73.349 54.295
275054 93.764 -73.343 54.275
275055 93.783 -73.338 54.253
275056 93.801 -73.333 54.232
275057 93.820 -73.327 54.209
275058 93.839 -73.322 54.185
275059 93.858 -73.317 54.162
275060 93.877 -73.312 54.140
275061 93.896 -73.307 54.114
275062 93.915 -73.302 54.089
275063 93.933 -73.296 54.065
275064 93.952 -73.291 54.040
275065 93.971 -73.286 54.017
275066 93.990 -73.280 53.994
275067 94.009 -73.275 53.971
275068 94.028 -73.271 53.946
275069 94.047 -73.265 53.921
275070 94.065 -73.260 53.896
275071 94.084 -73.255 53.872
275072 94.103 -73.249 53.848
275073 94.122 -73.244 53.825
275074 94.141 -73.239 53.800
275075 94.165 -73.246 53.774
275076 94.191 -73.261 53.749
275077 94.217 -73.275 53.725
275078 94.244 -73.289 53.700
275079 94.270 -73.303 53.676
275091 94.579 -73.395 53.376
275092 94.605 -73.402 53.350
275093 94.630 -73.410 53.328
275094 94.656 -73.417 53.305
275095 94.682 -73.424 53.281
275096 94.707 -73.432 53.258
275097 94.734 -73.440 53.236
275098 94.759 -73.446 53.212
275099 94.785 -73.454 53.187
275100 94.799 -73.456 53.161
275101 94.806 -73.456 53.134
275102 94.813 -73.454 53.106
275103 94.833 -73.459 53.077
275104 94.854 -73.463 53.048
275105 94.875 -73.468 53.016
275106 94.895 -73.471 52.985
275107 94.915 -73.476 52.954
275108 94.944 -73.474 52.923
275109 94.987 -73.463 52.890
275110 95.001 -73.477 52.859
275111 95.016 -73.490 52.827
275112 95.030 -73.503 52.795
275113 95.045 -73.516 52.764
275114 95.059 -73.530 52.734
275115 95.073 -73.543 52.704
275116 95.088 -73.556 52.675
275117 95.102 -73.570 52.647
275118 95.117 -73.583 52.618
275119 95.131 -73.597 52.589
275120 95.164 -73.592 52.561
275121 95.210 -73.584 52.533
275122 95.257 -73.575 52.506
275123 95.305 -73.565 52.480
275124 95.351 -73.557 52.453
275125 95.398 -73.548 52.426
275126 95.446 -73.539 52.399
275127 95.492 -73.531 52.373
275128 95.525 -73.529 52.350
275129 95.553 -73.530 52.327
275130 95.581 -73.531 52.302
275131 95.608 -73.517 52.277
275132 95.635 -73.496 52.252
275133 95.662 -73.477 52.228
275134 95.689 -73.457 52.204
275135 95.716 -73.437 52.181
275136 95.744 -73.431 52.160
275137 95.774 -73.464 52.137
275138 95.799 -73.474 52.110
275139 95.820 -73.474 52.083
275140 95.842 -73.474 52.056
275141 95.863 -73.474 52.027
275142 95.892 -73.486 51.996
275143 95.921 -73.499 51.964
275144 95.952 -73.512 51.932
275145 95.983 -73.522 51.898
275146 96.020 -73.522 51.863
275147 96.056 -73.522 51.829
275148 96.092 -73.522 51.794
275149 96.129 -73.522 51.760
275150 96.151 -73.522 51.726
275151 96.158 -73.523 51.692
275152 96.200 -73.544 51.658
275153 96.242 -73.565 51.625
275154 96.284 -73.587 51.593
275155 96.325 -73.609 51.561
275156 96.368 -73.631 51.527
275157 96.408 -73.637 51.494
275158 96.426 -73.635 51.462
275159 96.444 -73.634 51.430
275160 96.456 -73.628 51.399
275161 96.469 -73.622 51.368
275162 96.482 -73.617 51.336
275163 96.495 -73.611 51.304
275164 96.508 -73.605 51.274
275165 96.521 -73.600 51.244
275166 96.534 -73.593 51.217
275167 96.572 -73.587 51.189
275168 96.647 -73.578 51.161
275169 96.705 -73.583 51.135
275170 96.752 -73.602 51.110
275171 96.802 -73.625 51.086
275172 96.852 -73.647 51.063
275173 96.902 -73.669 51.042
275174 96.933 -73.672 51.021
275175 96.924 -73.628 51.001
275176 96.914 -73.584 50.984
275177 96.905 -73.540 50.967
275178 96.897 -73.575 50.954
275179 96.907 -73.658 50.943
275180 96.929 -73.668 50.932
275181 96.957 -73.672 50.922
275182 96.987 -73.675 50.912
275183 97.019 -73.678 50.903
275184 97.048 -73.691 50.896
275185 97.077 -73.699 50.891
275186 97.104 -73.695 50.887
275187 97.134 -73.692 50.884
275188 97.163 -73.695 50.879
275189 97.198 -73.697 50.874
275190 97.233 -73.698 50.870
275191 97.267 -73.699 50.867
275192 97.271 -73.703 50.864
275193 97.267 -73.708 50.861
275194 97.278 -73.713 50.855
275195 97.298 -73.719 50.847
275196 97.317 -73.725 50.839
275197 97.348 -73.730 50.832
275198 97.388 -73.734 50.825
275199 97.428 -73.740 50.817
275200 97.468 -73.745 50.804
275201 97.508 -73.750 50.787
275202 97.548 -73.755 50.761
275203 97.588 -73.760 50.730
275204 97.628 -73.766 50.695
275205 97.668 -73.770 50.656
275206 97.708 -73.775 50.614
275207 97.748 -73.781 50.568
275208 97.788 -73.785 50.519
275209 97.828 -73.791 50.469
275210 97.868 -73.796 50.419
275211 97.900 -73.807 50.372
275212 97.926 -73.825 50.325
275213 97.953 -73.841 50.276
275214 97.991 -73.854 50.221
275215 98.032 -73.864 50.162
275216 98.074 -73.874 50.101
275217 98.115 -73.885 50.039
275218 98.157 -73.895 49.977
275219 98.199 -73.906 49.916
275220 98.237 -73.914 49.853
275221 98.274 -73.921 49.789
275222 98.310 -73.929 49.726
275223 98.361 -73.939 49.660
275224 98.397 -73.970 49.594
275225 98.405 -73.973 49.527
275226 98.406 -73.973 49.462
275227 98.400 -73.982 49.400
275228 98.394 -73.990 49.338
275229 98.401 -74.004 49.276
275230 98.448 -74.033 49.215
275231 98.495 -74.063 49.155
275232 98.541 -74.089 49.095
275233 98.578 -74.077 49.037
275234 98.638 -74.123 48.981
275235 98.685 -74.131 48.926
275236 98.724 -74.114 48.873
275237 98.740 -74.136 48.822
275238 98.754 -74.162 48.773
275239 98.768 -74.188 48.725
275240 98.781 -74.214 48.678
275241 98.798 -74.240 48.630
275242 98.820 -74.245 48.582
275243 98.854 -74.230 48.533
275244 98.880 -74.234 48.484
275245 98.909 -74.242 48.435
275246 98.942 -74.251 48.385
275247 98.975 -74.259 48.334
275248 99.007 -74.268 48.283
275249 99.040 -74.277 48.232
275250 99.072 -74.286 48.181
275251 99.105 -74.295 48.132
275252 99.138 -74.304 48.083
275253 99.170 -74.309 48.035
275254 99.202 -74.314 47.990
275255 99.229 -74.329 47.944
275256 99.253 -74.349 47.899
275257 99.289 -74.366 47.854
275258 99.331 -74.381 47.811
275259 99.371 -74.396 47.768
275260 99.412 -74.410 47.726
275261 99.453 -74.425 47.686
275262 99.494 -74.440 47.647
275263 99.535 -74.455 47.608
275264 99.576 -74.469 47.569
275265 99.607 -74.478 47.531
275266 99.637 -74.487 47.494
275267 99.666 -74.495 47.460
275268 99.696 -74.503 47.424
275269 99.726 -74.512 47.387
275270 99.756 -74.521 47.351
275271 99.786 -74.529 47.315
275272 99.837 -74.535 47.279
275273 99.900 -74.540 47.245
275274 99.964 -74.545 47.214
275275 100.000 -74.552 47.183
275276 100.031 -74.559 47.152
275277 100.060 -74.568 47.120
275278 100.081 -74.590 47.089
275279 100.103 -74.611 47.060
275280 100.125 -74.631 47.031
275281 100.146 -74.653 47.004
275282 100.168 -74.673 46.977
275283 100.190 -74.694 46.950
275284 100.212 -74.715 46.923
275285 100.233 -74.736 46.898
275286 100.255 -74.757 46.872
275287 100.280 -74.772 46.848
275288 100.308 -74.781 46.822
275289 100.335 -74.789 46.794
275290 100.362 -74.797 46.764
275291 100.388 -74.804 46.734
275292 100.416 -74.812 46.705
275293 100.442 -74.820 46.676
275294 100.470 -74.828 46.647
275295 100.496 -74.835 46.618
275296 100.523 -74.843 46.589
275297 100.550 -74.850 46.560
275298 100.577 -74.858 46.532
275299 100.604 -74.866 46.502
275300 100.631 -74.874 46.474
275301 100.657 -74.881 46.446
275302 100.685 -74.889 46.417
275303 100.710 -74.897 46.389
275304 100.732 -74.903 46.360
275305 100.753 -74.910 46.331
275306 100.774 -74.916 46.301
275307 100.802 -74.927 46.275
275308 100.830 -74.938 46.247
275309 100.857 -74.948 46.220
275310 100.891 -74.969 46.193
275311 100.924 -74.988 46.163
275312 100.957 -75.009 46.132
275313 100.986 -75.022 46.101
275314 101.015 -75.034 46.071
275315 101.044 -75.046 46.040
275316 101.073 -75.057 46.008
275317 101.102 -75.069 45.976
275318 101.130 -75.082 45.944
275319 101.159 -75.094 45.911
275320 101.188 -75.106 45.878
275321 101.217 -75.118 45.845
275322 101.251 -75.135 45.810
275323 101.285 -75.155 45.774
275324 101.319 -75.173 45.738
275325 101.352 -75.192 45.701
275326 101.386 -75.210 45.666
275327 101.420 -75.229 45.635
275328 101.454 -75.248 45.603
275329 101.487 -75.267 45.570
275330 101.521 -75.285 45.539
275331 101.555 -75.304 45.507
275332 101.588 -75.322 45.476
275333 101.622 -75.340 45.446
275334 101.656 -75.359 45.418
275335 101.689 -75.377 45.389
275336 101.723 -75.396 45.359
275337 101.754 -75.401 45.329
275338 101.783 -75.395 45.299
275339 101.801 -75.387 45.269
275340 101.818 -75.387 45.239
275341 101.835 -75.388 45.211
275342 101.853 -75.389 45.183
275343 101.870 -75.389 45.153
275344 101.888 -75.389 45.125
275345 101.905 -75.390 45.097
275346 101.922 -75.391 45.070
275347 101.940 -75.391 45.044
275348 101.957 -75.392 45.017
275349 101.975 -75.392 44.991
275350 101.992 -75.392 44.965
275351 102.010 -75.393 44.939
275352 102.027 -75.394 44.914
275710 92.393 -73.085 54.742
275711 92.424 -73.095 54.727
275712 92.454 -73.105 54.713
275713 92.485 -73.116 54.699
275714 92.516 -73.126 54.685
275715 92.546 -73.136 54.671
275716 92.577 -73.147 54.657
275717 92.607 -73.157 54.645
275718 92.638 -73.167 54.633
275719 92.670 -73.178 54.622
275720 92.713 -73.189 54.612
275721 92.756 -73.200 54.603
275722 92.799 -73.211 54.593
275723 92.842 -73.222 54.585
275724 92.885 -73.233 54.579
275725 92.928 -73.244 54.573
275726 92.971 -73.255 54.568
275727 93.013 -73.266 54.563
275728 93.056 -73.277 54.558
275729 93.099 -73.288 54.552
275730 93.142 -73.299 54.547
275731 93.185 -73.310 54.542
275732 93.228 -73.321 54.538
275733 93.260 -73.322 54.533
275734 93.290 -73.323 54.530
275735 93.331 -73.337 54.525
275736 93.366 -73.349 54.519
275737 93.391 -73.358 54.512
275738 93.415 -73.367 54.505
275739 93.439 -73.373 54.498
275740 93.464 -73.379 54.491
275741 93.491 -73.383 54.483
275742 93.518 -73.377 54.470
275743 93.540 -73.371 54.455
275744 93.559 -73.366 54.441
275745 93.577 -73.361 54.426
275746 93.596 -73.355 54.412
275747 93.615 -73.351 54.399
275748 93.634 -73.346 54.384
275749 93.653 -73.340 54.366
275750 93.672 -73.335 54.349
275751 93.691 -73.330 54.331
275752 93.709 -73.324 54.313
275753 93.728 -73.319 54.294
275754 93.747 -73.314 54.276
275755 93.766 -73.309 54.256
275756 93.785 -73.304 54.234
275757 93.803 -73.299 54.211
275758 93.823 -73.293 54.187
275759 93.841 -73.288 54.164
275760 93.860 -73.283 54.140
275761 93.879 -73.277 54.117
275762 93.898 -73.272 54.092
275763 93.917 -73.267 54.067
275764 93.935 -73.262 54.042
275765 93.955 -73.257 54.017
275766 93.973 -73.252 53.994
275767 93.992 -73.246 53.971
275768 94.011 -73.241 53.946
275769 94.030 -73.236 53.921
275770 94.049 -73.230 53.897
275771 94.067 -73.225 53.873
275772 94.086 -73.221 53.849
275773 94.105 -73.215 53.826
275774 94.124 -73.210 53.802
275775 94.143 -73.205 53.776
275776 94.163 -73.201 53.750
275777 94.189 -73.215 53.726
275778 94.215 -73.230 53.702
275779 94.242 -73.244 53.678
275795 94.655 -73.380 53.288
275796 94.681 -73.387 53.264
275797 94.706 -73.395 53.241
275798 94.732 -73.402 53.218
275799 94.758 -73.409 53.194
275800 94.783 -73.417 53.169
275801 94.810 -73.424 53.143
275802 94.817 -73.424 53.116
275803 94.824 -73.422 53.089
275804 94.836 -73.424 53.060
275805 94.857 -73.428 53.031
275806 94.878 -73.433 53.001
275807 94.898 -73.437 52.970
275808 94.918 -73.441 52.939
275809 94.939 -73.446 52.908
275810 94.966 -73.446 52.876
275811 95.008 -73.434 52.845
275812 95.023 -73.447 52.815
275813 95.038 -73.460 52.784
275814 95.052 -73.474 52.753
275815 95.066 -73.487 52.723
275816 95.081 -73.500 52.694
275817 95.095 -73.514 52.666
275818 95.110 -73.527 52.638
275819 95.124 -73.540 52.609
275820 95.138 -73.553 52.581
275821 95.158 -73.558 52.553
275822 95.204 -73.534 52.526
275823 95.253 -73.525 52.500
275824 95.300 -73.516 52.475
275825 95.346 -73.508 52.449
275826 95.394 -73.499 52.421
275827 95.441 -73.490 52.394
275828 95.487 -73.481 52.370
275829 95.520 -73.480 52.346
275830 95.548 -73.481 52.323
275831 95.576 -73.481 52.299
275832 95.605 -73.482 52.275
275833 95.632 -73.470 52.251
275834 95.659 -73.450 52.227
275835 95.685 -73.430 52.205
275836 95.712 -73.411 52.183
275837 95.741 -73.408 52.161
275838 95.771 -73.440 52.139
275839 95.800 -73.464 52.115
275840 95.822 -73.464 52.088
275841 95.843 -73.464 52.061
275842 95.864 -73.465 52.032
275843 95.885 -73.465 52.001
275844 95.909 -73.468 51.970
275845 95.939 -73.482 51.938
275846 95.970 -73.493 51.904
275847 96.007 -73.493 51.870
275848 96.043 -73.493 51.836
275849 96.079 -73.493 51.802
275850 96.115 -73.496 51.768
275851 96.151 -73.497 51.734
275852 96.160 -73.497 51.700
275853 96.195 -73.515 51.668
275854 96.237 -73.537 51.636
275855 96.278 -73.558 51.603
275856 96.320 -73.580 51.571
275857 96.366 -73.585 51.538
275858 96.398 -73.586 51.505
275859 96.415 -73.584 51.473
275860 96.433 -73.583 51.441
275861 96.450 -73.581 51.409
275862 96.467 -73.581 51.377
275863 96.481 -73.575 51.346
275864 96.494 -73.569 51.315
275865 96.507 -73.563 51.284
275866 96.520 -73.558 51.255
275867 96.533 -73.552 51.228
275868 96.590 -73.543 51.200
275869 96.651 -73.546 51.172
275870 96.695 -73.563 51.146
275871 96.742 -73.583 51.120
275872 96.792 -73.605 51.096
275873 96.842 -73.627 51.073
275874 96.892 -73.650 51.050
275875 96.920 -73.661 51.029
275876 96.921 -73.651 51.009
275877 96.920 -73.630 50.991
275878 96.911 -73.586 50.975
275879 96.902 -73.555 50.961
275880 96.913 -73.632 50.949
275881 96.935 -73.642 50.938
275882 96.955 -73.652 50.926
275883 96.980 -73.659 50.917
275884 97.012 -73.661 50.908
275885 97.043 -73.663 50.900
275886 97.074 -73.666 50.894
275887 97.117 -73.660 50.890
275888 97.156 -73.657 50.886
275889 97.192 -73.656 50.881
275890 97.209 -73.666 50.876
275891 97.227 -73.676 50.869
275892 97.261 -73.678 50.863
275893 97.282 -73.680 50.857
275894 97.279 -73.684 50.851
275895 97.276 -73.689 50.845
275896 97.288 -73.694 50.837
275897 97.315 -73.700 50.827
275898 97.355 -73.705 50.818
275899 97.395 -73.710 50.812
275900 97.435 -73.715 50.804
275901 97.475 -73.720 50.794
275902 97.515 -73.725 50.777
275903 97.555 -73.730 50.753
275904 97.595 -73.735 50.725
275905 97.635 -73.741 50.691
275906 97.675 -73.746 50.652
275907 97.715 -73.750 50.611
275908 97.755 -73.756 50.565
275909 97.795 -73.761 50.515
275910 97.835 -73.766 50.464
275911 97.875 -73.771 50.415
275912 97.912 -73.778 50.368
275913 97.938 -73.795 50.321
275914 97.967 -73.812 50.271
275915 98.008 -73.822 50.217
275916 98.050 -73.832 50.157
275917 98.092 -73.843 50.095
275918 98.134 -73.853 50.031
275919 98.172 -73.861 49.967
275920 98.208 -73.869 49.906
275921 98.244 -73.875 49.845
275922 98.281 -73.883 49.783
275923 98.317 -73.891 49.719
275924 98.383 -73.906 49.651
275925 98.385 -73.942 49.583
275926 98.390 -73.960 49.516
275927 98.401 -73.950 49.452
275928 98.406 -73.945 49.390
275929 98.401 -73.954 49.327
275930 98.407 -73.967 49.264
275931 98.454 -73.998 49.204
275932 98.501 -74.027 49.143
275933 98.548 -74.057 49.084
275934 98.597 -74.080 49.026
275935 98.650 -74.096 48.968
275936 98.703 -74.113 48.912
275937 98.744 -74.101 48.859
275938 98.767 -74.113 48.809
275939 98.780 -74.139 48.760
275940 98.794 -74.164 48.712
275941 98.807 -74.190 48.664
275942 98.820 -74.216 48.617
275943 98.836 -74.234 48.569
275944 98.872 -74.227 48.520
275945 98.908 -74.214 48.472
275946 98.934 -74.220 48.423
275947 98.960 -74.226 48.374
275948 98.987 -74.233 48.324
275949 99.013 -74.239 48.274
275950 99.043 -74.247 48.222
275951 99.076 -74.256 48.171
275952 99.109 -74.265 48.122
275953 99.141 -74.273 48.074
275954 99.174 -74.283 48.027
275955 99.206 -74.288 47.981
275956 99.234 -74.301 47.936
275957 99.263 -74.320 47.891
275958 99.304 -74.334 47.847
275959 99.345 -74.349 47.803
275960 99.386 -74.364 47.760
275961 99.427 -74.378 47.718
275962 99.468 -74.393 47.678
275963 99.509 -74.408 47.638
275964 99.550 -74.423 47.599
275965 99.591 -74.437 47.560
275966 99.625 -74.448 47.522
275967 99.655 -74.456 47.485
275968 99.685 -74.465 47.450
275969 99.714 -74.474 47.414
275970 99.744 -74.482 47.378
275971 99.774 -74.490 47.343
275972 99.834 -74.496 47.308
275973 99.898 -74.501 47.272
275974 99.938 -74.507 47.238
275975 99.969 -74.514 47.206
275976 100.000 -74.521 47.176
275977 100.031 -74.528 47.145
275978 100.061 -74.535 47.114
275979 100.084 -74.554 47.084
275980 100.106 -74.575 47.054
275981 100.127 -74.596 47.027
275982 100.149 -74.617 46.999
275983 100.171 -74.638 46.972
275984 100.192 -74.659 46.945
275985 100.214 -74.680 46.919
275986 100.236 -74.700 46.892
275987 100.258 -74.722 46.867
275988 100.286 -74.730 46.842
275989 100.317 -74.737 46.815
275990 100.347 -74.750 46.786
275991 100.378 -74.762 46.757
275992 100.409 -74.774 46.726
275993 100.436 -74.781 46.696
275994 100.463 -74.790 46.667
275995 100.490 -74.797 46.638
275996 100.516 -74.805 46.608
275997 100.543 -74.812 46.579
275998 100.570 -74.820 46.551
275999 100.597 -74.828 46.523
276000 100.624 -74.835 46.493
276001 100.651 -74.844 46.464
276002 100.678 -74.851 46.436
276003 100.705 -74.859 46.408
276004 100.732 -74.866 46.380
276005 100.754 -74.873 46.351
276006 100.775 -74.879 46.322
276007 100.796 -74.886 46.293
276008 100.818 -74.892 46.268
276009 100.841 -74.900 46.241
276010 100.874 -74.919 46.213
276011 100.907 -74.939 46.182
276012 100.940 -74.959 46.150
276013 100.973 -74.979 46.118
276014 101.005 -74.997 46.087
276015 101.034 -75.010 46.056
276016 101.063 -75.022 46.025
276017 101.092 -75.034 45.993
276018 101.121 -75.046 45.961
276019 101.149 -75.058 45.929
276020 101.178 -75.070 45.896
276021 101.207 -75.082 45.863
276022 101.236 -75.095 45.828
276023 101.272 -75.114 45.793
276024 101.308 -75.134 45.757
276025 101.341 -75.153 45.721
276026 101.375 -75.171 45.685
276027 101.409 -75.190 45.650
276028 101.443 -75.208 45.618
276029 101.476 -75.226 45.586
276030 101.510 -75.245 45.555
276031 101.543 -75.264 45.524
276032 101.577 -75.283 45.493
276033 101.611 -75.301 45.462
276034 101.644 -75.320 45.432
276035 101.678 -75.338 45.403
276036 101.712 -75.357 45.375
276037 101.744 -75.367 45.345
276038 101.773 -75.361 45.315
276039 101.802 -75.357 45.285
276040 101.815 -75.347 45.255
276041 101.833 -75.348 45.226
276042 101.850 -75.348 45.197
276043 101.868 -75.348 45.168
276044 101.885 -75.349 45.139
276045 101.903 -75.350 45.111
276046 101.920 -75.350 45.083
276047 101.938 -75.351 45.056
276048 101.955 -75.352 45.029
276049 101.972 -75.352 45.003
276050 101.990 -75.352 44.976
276051 102.007 -75.353 44.950
276052 102.025 -75.354 44.924
276412 92.420 -73.066 54.732
276413 92.450 -73.076 54.718
276414 92.481 -73.086 54.703
276415 92.511 -73.097 54.688
276416 92.542 -73.108 54.674
276417 92.573 -73.117 54.660
276418 92.603 -73.128 54.646
276419 92.634 -73.139 54.634
276420 92.667 -73.148 54.622
276421 92.710 -73.160 54.611
276422 92.753 -73.170 54.601
276423 92.796 -73.182 54.591
276424 92.839 -73.192 54.582
276425 92.881 -73.204 54.574
276426 92.924 -73.214 54.566
276427 92.968 -73.226 54.560
276428 93.010 -73.236 54.555
276429 93.053 -73.248 54.550
276430 93.096 -73.258 54.543
276431 93.139 -73.270 54.538
276432 93.182 -73.280 54.533
276433 93.225 -73.292 54.528
276434 93.268 -73.302 54.524
276435 93.300 -73.305 54.519
276436 93.334 -73.317 54.513
276437 93.358 -73.325 54.506
276438 93.382 -73.330 54.498
276439 93.407 -73.336 54.490
276440 93.431 -73.343 54.481
276441 93.456 -73.349 54.473
276442 93.487 -73.349 54.464
276443 93.513 -73.343 54.452
276444 93.539 -73.337 54.439
276445 93.561 -73.332 54.425
276446 93.580 -73.327 54.412
276447 93.599 -73.321 54.398
276448 93.618 -73.316 54.384
276449 93.636 -73.311 54.369
276450 93.655 -73.305 54.351
276451 93.674 -73.301 54.333
276452 93.693 -73.296 54.314
276453 93.712 -73.290 54.295
276459 93.825 -73.259 54.167
276460 93.844 -73.254 54.143
276461 93.862 -73.249 54.119
276462 93.882 -73.243 54.095
276463 93.900 -73.238 54.070
276464 93.919 -73.233 54.045
276465 93.938 -73.227 54.020
276466 93.957 -73.222 53.995
276467 93.976 -73.217 53.971
276468 93.994 -73.212 53.948
276469 94.013 -73.207 53.924
276470 94.032 -73.202 53.899
276471 94.051 -73.196 53.875
276472 94.070 -73.191 53.851
276473 94.088 -73.186 53.827
276474 94.108 -73.180 53.803
276475 94.126 -73.175 53.780
276476 94.145 -73.170 53.755
276477 94.164 -73.165 53.730
276478 94.187 -73.170 53.706
276500 94.757 -73.372 53.177
276501 94.782 -73.380 53.151
276502 94.808 -73.387 53.125
276503 94.828 -73.392 53.100
276504 94.835 -73.390 53.072
276505 94.842 -73.390 53.044
276506 94.860 -73.393 53.015
276507 94.881 -73.398 52.985
276508 94.901 -73.402 52.955
276509 94.922 -73.406 52.925
276510 94.942 -73.411 52.894
276511 94.963 -73.415 52.863
276512 94.987 -73.417 52.833
276513 95.030 -73.405 52.803
276514 95.045 -73.418 52.772
276515 95.059 -73.431 52.743
276516 95.074 -73.444 52.714
276517 95.089 -73.458 52.685
276518 95.103 -73.471 52.657
276519 95.117 -73.484 52.629
276520 95.131 -73.498 52.601
276521 95.146 -73.511 52.573
276522 95.161 -73.524 52.546
276523 95.198 -73.493 52.520
276524 95.245 -73.474 52.494
276525 95.294 -73.467 52.469
276526 95.342 -73.459 52.443
276527 95.389 -73.449 52.416
276528 95.436 -73.441 52.390
276529 95.483 -73.432 52.365
276530 95.515 -73.431 52.342
276531 95.543 -73.432 52.320
276532 95.571 -73.433 52.296
276533 95.600 -73.434 52.272
276534 95.628 -73.434 52.249
276535 95.656 -73.424 52.226
276536 95.682 -73.404 52.204
276537 95.709 -73.384 52.183
276538 95.738 -73.384 52.162
276539 95.768 -73.417 52.141
276540 95.800 -73.449 52.118
276541 95.823 -73.454 52.092
276542 95.844 -73.454 52.065
276543 95.865 -73.454 52.036
276544 95.887 -73.455 52.006
276545 95.908 -73.455 51.974
276546 95.929 -73.455 51.942
276547 95.957 -73.464 51.909
276548 95.993 -73.464 51.875
276549 96.029 -73.466 51.842
276550 96.065 -73.468 51.809
276551 96.101 -73.470 51.777
276552 96.136 -73.471 51.742
276553 96.162 -73.473 51.709
276554 96.189 -73.486 51.677
276555 96.231 -73.507 51.645
276556 96.273 -73.529 51.613
276557 96.319 -73.534 51.581
276558 96.366 -73.536 51.548
276559 96.387 -73.534 51.516
276560 96.404 -73.534 51.484
276561 96.422 -73.532 51.452
276562 96.440 -73.531 51.420
276563 96.457 -73.529 51.389
276564 96.474 -73.528 51.358
276565 96.491 -73.527 51.327
276566 96.506 -73.521 51.297
276567 96.518 -73.516 51.268
276568 96.532 -73.510 51.240
276569 96.597 -73.510 51.212
276570 96.641 -73.527 51.185
276571 96.685 -73.543 51.158
276572 96.733 -73.563 51.132
276573 96.783 -73.586 51.107
276574 96.833 -73.608 51.084
276575 96.881 -73.633 51.061
276576 96.904 -73.641 51.039
276577 96.905 -73.632 51.019
276578 96.906 -73.622 51.001
276579 96.909 -73.611 50.984
276580 96.907 -73.587 50.969
276581 96.920 -73.606 50.956
276582 96.941 -73.616 50.944
276583 96.961 -73.625 50.932
276584 96.982 -73.635 50.922
276585 97.004 -73.644 50.912
276586 97.035 -73.647 50.903
276587 97.075 -73.644 50.896
276588 97.124 -73.635 50.891
276589 97.174 -73.625 50.886
276590 97.217 -73.620 50.881
276591 97.244 -73.624 50.876
276592 97.261 -73.634 50.868
276593 97.278 -73.644 50.860
276594 97.292 -73.657 50.852
276595 97.288 -73.662 50.843
276596 97.285 -73.666 50.834
276597 97.282 -73.672 50.824
276598 97.321 -73.673 50.815
276599 97.362 -73.680 50.805
276600 97.401 -73.685 50.799
276601 97.442 -73.691 50.792
276602 97.481 -73.695 50.782
276603 97.522 -73.700 50.769
276604 97.562 -73.706 50.748
276605 97.602 -73.710 50.720
276606 97.642 -73.716 50.687
276607 97.681 -73.721 50.650
276608 97.722 -73.725 50.607
276609 97.761 -73.731 50.560
276610 97.802 -73.736 50.511
276611 97.842 -73.741 50.461
276612 97.881 -73.746 50.411
276613 97.922 -73.751 50.364
276614 97.950 -73.766 50.316
276615 97.985 -73.780 50.267
276616 98.026 -73.790 50.211
276617 98.068 -73.800 50.151
276618 98.106 -73.809 50.088
276619 98.142 -73.816 50.024
276620 98.178 -73.823 49.960
276621 98.215 -73.831 49.899
276622 98.251 -73.838 49.838
276623 98.288 -73.845 49.774
276624 98.324 -73.852 49.707
276625 98.371 -73.879 49.639
276626 98.373 -73.914 49.573
276627 98.376 -73.947 49.509
276628 98.387 -73.937 49.446
276629 98.397 -73.926 49.381
276630 98.407 -73.918 49.316
276631 98.414 -73.932 49.252
276632 98.460 -73.961 49.192
276633 98.532 -74.008 49.132
276634 98.590 -74.038 49.073
276635 98.629 -74.049 49.016
276636 98.667 -74.060 48.957
276637 98.713 -74.074 48.900
276638 98.765 -74.089 48.845
276639 98.793 -74.090 48.794
276640 98.807 -74.116 48.745
276641 98.820 -74.142 48.697
276642 98.823 -74.164 48.650
276643 98.827 -74.187 48.604
276644 98.851 -74.202 48.557
276645 98.888 -74.213 48.509
276646 98.927 -74.211 48.461
276647 98.961 -74.200 48.413
276648 98.987 -74.207 48.364
276649 99.014 -74.213 48.314
276650 99.040 -74.219 48.264
276651 99.067 -74.225 48.213
276652 99.093 -74.232 48.162
276653 99.120 -74.238 48.113
276654 99.146 -74.244 48.065
276655 99.178 -74.252 48.018
276656 99.210 -74.261 47.972
276657 99.239 -74.272 47.927
276658 99.278 -74.288 47.884
276659 99.319 -74.302 47.840
276660 99.360 -74.317 47.796
276661 99.401 -74.332 47.752
276662 99.442 -74.346 47.710
276663 99.483 -74.361 47.669
276664 99.524 -74.376 47.629
276665 99.565 -74.391 47.591
276666 99.606 -74.405 47.552
276667 99.643 -74.418 47.514
276668 99.673 -74.427 47.477
276669 99.703 -74.435 47.441
276670 99.732 -74.443 47.405
276671 99.769 -74.451 47.370
276672 99.832 -74.456 47.335
276673 99.876 -74.462 47.300
276674 99.907 -74.469 47.265
276675 99.938 -74.476 47.231
276676 99.969 -74.483 47.199
276677 99.999 -74.490 47.169
276678 100.030 -74.497 47.138
276679 100.061 -74.504 47.107
276680 100.086 -74.518 47.078
276681 100.108 -74.540 47.049
276682 100.130 -74.561 47.021
276683 100.151 -74.581 46.995
276684 100.173 -74.603 46.967
276685 100.195 -74.623 46.940
276686 100.216 -74.644 46.913
276687 100.238 -74.665 46.887
276688 100.260 -74.685 46.861
276689 100.291 -74.689 46.835
276690 100.322 -74.693 46.807
276691 100.353 -74.702 46.778
276692 100.384 -74.714 46.750
276693 100.414 -74.726 46.718
276694 100.444 -74.738 46.687
276695 100.475 -74.750 46.657
276696 100.506 -74.762 46.628
276697 100.537 -74.775 46.599
276698 100.563 -74.782 46.570
276699 100.591 -74.790 46.541
276700 100.617 -74.797 46.512
276701 100.644 -74.805 46.482
276702 100.671 -74.813 46.454
276703 100.698 -74.821 46.426
276704 100.725 -74.828 46.399
276705 100.752 -74.836 46.371
276706 100.777 -74.844 46.343
276707 100.798 -74.850 46.314
276708 100.820 -74.856 46.286
276709 100.841 -74.863 46.261
276710 100.862 -74.872 46.233
276711 100.890 -74.889 46.203
276712 100.923 -74.910 46.170
276713 100.957 -74.929 46.135
276714 100.990 -74.950 46.103
276715 101.023 -74.970 46.071
276716 101.054 -74.985 46.040
276717 101.082 -74.997 46.009
276718 101.111 -75.010 45.977
276719 101.140 -75.022 45.945
276720 101.169 -75.034 45.914
276721 101.198 -75.046 45.879
276722 101.227 -75.058 45.846
276723 101.255 -75.070 45.811
276724 101.293 -75.092 45.775
276725 101.330 -75.113 45.739
276726 101.364 -75.132 45.703
276727 101.397 -75.150 45.668
276728 101.431 -75.169 45.634
276729 101.465 -75.187 45.602
276730 101.498 -75.206 45.571
276731 101.532 -75.224 45.539
276732 101.566 -75.243 45.509
276733 101.600 -75.261 45.478
276734 101.633 -75.280 45.448
276735 101.667 -75.298 45.418
276736 101.700 -75.317 45.389
276737 101.734 -75.333 45.360
276738 101.763 -75.327 45.331
276739 101.792 -75.323 45.301
276740 101.815 -75.311 45.271
276741 101.830 -75.308 45.242
276742 101.848 -75.308 45.212
276743 101.865 -75.308 45.183
276744 101.883 -75.309 45.154
276745 101.900 -75.310 45.125
276746 101.917 -75.310 45.097
276747 101.935 -75.311 45.070
276748 101.953 -75.311 45.041
276749 101.970 -75.311 45.014
276750 101.987 -75.312 44.988
276751 102.005 -75.313 44.962
276752 102.022 -75.314 44.935
276753 102.040 -75.314 44.909
277113 92.415 -73.037 54.737
277114 92.446 -73.047 54.722
277115 92.477 -73.058 54.707
277116 92.507 -73.068 54.692
277117 92.538 -73.078 54.677
277118 92.569 -73.089 54.663
277119 92.599 -73.098 54.648
277120 92.630 -73.109 54.635
277121 92.664 -73.120 54.623
277122 92.707 -73.130 54.611
277123 92.749 -73.142 54.600
277124 92.792 -73.152 54.589
277125 92.836 -73.164 54.578
277126 92.878 -73.174 54.569
277127 92.921 -73.186 54.561
277128 92.964 -73.196 54.554
277129 93.007 -73.208 54.548
277130 93.050 -73.218 54.542
277131 93.093 -73.230 54.535
277132 93.136 -73.240 54.529
277133 93.179 -73.252 54.523
277134 93.221 -73.262 54.517
277135 93.264 -73.274 54.513
277136 93.301 -73.283 54.508
277137 93.326 -73.289 54.501
277138 93.350 -73.295 54.493
277139 93.374 -73.300 54.485
277140 93.399 -73.306 54.476
277141 93.423 -73.312 54.467
277142 93.448 -73.318 54.458
277143 93.482 -73.314 54.450
277144 93.508 -73.309 54.438
277145 93.534 -73.303 54.425
277146 93.560 -73.298 54.412
277147 93.582 -73.293 54.398
277148 93.601 -73.287 54.385
277149 93.620 -73.282 54.371
277150 93.638 -73.277 54.355
277151 93.657 -73.271 54.336
277152 93.676 -73.266 54.316
277153 93.695 -73.261 54.297
277154 93.714 -73.256 54.279
277204 94.833 -73.357 53.082
277205 94.845 -73.359 53.055
277206 94.852 -73.358 53.027
277207 94.863 -73.358 52.999
277208 94.884 -73.363 52.970
277209 94.904 -73.367 52.940
277210 94.925 -73.371 52.910
277211 94.945 -73.376 52.880
277212 94.966 -73.380 52.852
277213 94.986 -73.385 52.822
277214 95.008 -73.388 52.792
277215 95.051 -73.377 52.762
277216 95.067 -73.388 52.733
277217 95.082 -73.402 52.705
277218 95.096 -73.415 52.677
277219 95.110 -73.428 52.649
277220 95.125 -73.442 52.622
277221 95.139 -73.455 52.594
277222 95.154 -73.468 52.566
277223 95.168 -73.481 52.540
277224 95.194 -73.473 52.513
277225 95.238 -73.428 52.488
277226 95.285 -73.416 52.462
277227 95.334 -73.408 52.435
277228 95.383 -73.400 52.410
277229 95.431 -73.392 52.385
277230 95.478 -73.383 52.360
277231 95.510 -73.382 52.338
277232 95.538 -73.383 52.315
277233 95.566 -73.383 52.292
277234 95.595 -73.384 52.269
277235 95.623 -73.385 52.246
277236 95.651 -73.386 52.224
277237 95.679 -73.377 52.203
277238 95.706 -73.358 52.182
277239 95.735 -73.361 52.161
277240 95.765 -73.393 52.141
277241 95.800 -73.417 52.120
277242 95.837 -73.428 52.095
277243 95.858 -73.429 52.067
277244 95.879 -73.429 52.039
277245 95.900 -73.430 52.009
277246 95.920 -73.430 51.977
277247 95.942 -73.430 51.945
277248 95.962 -73.431 51.913
277249 95.988 -73.436 51.880
277250 96.015 -73.440 51.848
277251 96.051 -73.442 51.815
277252 96.087 -73.444 51.784
277253 96.122 -73.446 51.751
277254 96.158 -73.448 51.718
277255 96.184 -73.456 51.686
277256 96.226 -73.478 51.654
277257 96.272 -73.482 51.623
277258 96.319 -73.484 51.591
277259 96.359 -73.485 51.559
277260 96.377 -73.484 51.526
277261 96.394 -73.482 51.495
277262 96.411 -73.481 51.463
277263 96.429 -73.480 51.431
277264 96.446 -73.478 51.400
277265 96.463 -73.477 51.370
277266 96.481 -73.475 51.340
277267 96.499 -73.474 51.310
277268 96.516 -73.473 51.281
277269 96.543 -73.473 51.253
277270 96.587 -73.490 51.225
277271 96.631 -73.507 51.198
277272 96.675 -73.524 51.171
277273 96.723 -73.544 51.145
277274 96.773 -73.566 51.120
277275 96.822 -73.590 51.096
277276 96.869 -73.616 51.074
277277 96.889 -73.622 51.051
277278 96.890 -73.613 51.030
277279 96.890 -73.604 51.012
277280 96.892 -73.593 50.994
277281 96.895 -73.582 50.979
277282 96.929 -73.568 50.965
277283 96.949 -73.579 50.952
277284 96.969 -73.591 50.939
277285 96.990 -73.603 50.928
277286 97.011 -73.614 50.917
277287 97.043 -73.625 50.908
277288 97.091 -73.617 50.900
277289 97.139 -73.608 50.894
277290 97.186 -73.600 50.888
277291 97.234 -73.590 50.881
277292 97.279 -73.584 50.876
277293 97.339 -73.596 50.868
277294 97.315 -73.618 50.859
277295 97.293 -73.637 50.850
277296 97.289 -73.642 50.839
277297 97.283 -73.648 50.827
277298 97.291 -73.653 50.816
277299 97.329 -73.649 50.807
277300 97.368 -73.646 50.797
277301 97.407 -73.647 50.790
277302 97.447 -73.654 50.783
277303 97.488 -73.661 50.775
277304 97.527 -73.668 50.763
277305 97.568 -73.675 50.743
277306 97.608 -73.681 50.716
277307 97.648 -73.688 50.683
277308 97.688 -73.695 50.646
277309 97.728 -73.701 50.603
277310 97.769 -73.706 50.556
277311 97.808 -73.711 50.506
277312 97.849 -73.716 50.457
277313 97.888 -73.721 50.407
277314 97.928 -73.726 50.359
277315 97.962 -73.737 50.311
277316 98.002 -73.748 50.261
277317 98.040 -73.756 50.203
277318 98.077 -73.763 50.142
277319 98.113 -73.771 50.079
277320 98.149 -73.778 50.015
277321 98.185 -73.785 49.951
277322 98.222 -73.792 49.890
277323 98.258 -73.800 49.828
277324 98.295 -73.807 49.762
277325 98.342 -73.817 49.695
277326 98.360 -73.850 49.628
277327 98.362 -73.886 49.564
277328 98.364 -73.922 49.502
277329 98.372 -73.924 49.439
277330 98.383 -73.913 49.372
277331 98.393 -73.904 49.305
277332 98.429 -73.925 49.240
277333 98.533 -73.970 49.180
277334 98.589 -73.998 49.121
277335 98.628 -74.010 49.063
277336 98.667 -74.021 49.006
277337 98.705 -74.032 48.947
277338 98.744 -74.043 48.889
277339 98.783 -74.054 48.834
277340 98.820 -74.067 48.781
277341 98.825 -74.090 48.731
277342 98.827 -74.113 48.683
277343 98.830 -74.136 48.637
277344 98.833 -74.159 48.591
277345 98.865 -74.171 48.544
277346 98.903 -74.181 48.497
277347 98.942 -74.192 48.450
277348 98.981 -74.195 48.403
277349 99.014 -74.186 48.354
277350 99.041 -74.192 48.304
277351 99.067 -74.199 48.254
277352 99.093 -74.205 48.204
277353 99.120 -74.211 48.153
277354 99.146 -74.217 48.105
277355 99.173 -74.223 48.057
277356 99.199 -74.230 48.008
277357 99.226 -74.236 47.962
277358 99.249 -74.245 47.919
277359 99.292 -74.257 47.876
277360 99.334 -74.270 47.832
277361 99.375 -74.286 47.789
277362 99.416 -74.300 47.746
277363 99.457 -74.315 47.703
277364 99.498 -74.330 47.662
277365 99.539 -74.344 47.622
277366 99.580 -74.359 47.583
277367 99.621 -74.374 47.545
277368 99.661 -74.388 47.507
277369 99.691 -74.396 47.469
277370 99.721 -74.405 47.433
277371 99.766 -74.411 47.398
277372 99.815 -74.418 47.363
277373 99.845 -74.424 47.328
277374 99.876 -74.432 47.292
277375 99.907 -74.439 47.258
277376 99.937 -74.446 47.225
277377 99.968 -74.452 47.193
277378 99.999 -74.459 47.162
277379 100.030 -74.467 47.132
277380 100.060 -74.474 47.103
277381 100.089 -74.483 47.073
277382 100.111 -74.504 47.044
277383 100.132 -74.525 47.016
277384 100.154 -74.546 46.990
277385 100.176 -74.567 46.962
277386 100.197 -74.588 46.935
277387 100.219 -74.609 46.907
277388 100.241 -74.630 46.880
277389 100.266 -74.645 46.853
277390 100.296 -74.649 46.827
277391 100.327 -74.653 46.799
277392 100.358 -74.656 46.770
277393 100.389 -74.666 46.740
277394 100.419 -74.678 46.709
277395 100.450 -74.690 46.677
277396 100.481 -74.703 46.647
277397 100.511 -74.715 46.618
277398 100.542 -74.727 46.589
277399 100.572 -74.739 46.560
277400 100.603 -74.751 46.531
277401 100.634 -74.763 46.501
277402 100.665 -74.775 46.472
277403 100.691 -74.783 46.444
277404 100.718 -74.791 46.416
277405 100.745 -74.798 46.389
277406 100.772 -74.806 46.362
277407 100.798 -74.813 46.334
277408 100.821 -74.820 46.307
277409 100.842 -74.827 46.279
277410 100.862 -74.834 46.252
277411 100.886 -74.847 46.223
277412 100.913 -74.864 46.191
277413 100.940 -74.881 46.158
277414 100.973 -74.900 46.122
277415 101.006 -74.920 46.088
277416 101.039 -74.941 46.056
277417 101.072 -74.960 46.024
277418 101.102 -74.973 45.993
277419 101.130 -74.985 45.961
277420 101.159 -74.997 45.929
277421 101.188 -75.010 45.897
277422 101.217 -75.022 45.862
277423 101.246 -75.034 45.828
277424 101.276 -75.048 45.793
277425 101.315 -75.070 45.757
277426 101.352 -75.092 45.722
277427 101.386 -75.110 45.686
277428 101.420 -75.129 45.651
277429 101.454 -75.148 45.618
277430 101.487 -75.166 45.586
277431 101.521 -75.185 45.555
277432 101.554 -75.203 45.525
277433 101.588 -75.222 45.494
277434 101.622 -75.240 45.463
277435 101.656 -75.259 45.432
277436 101.689 -75.277 45.403
277437 101.723 -75.296 45.375
277438 101.753 -75.293 45.346
277439 101.782 -75.289 45.316
277440 101.811 -75.283 45.287
277441 101.828 -75.267 45.257
277442 101.845 -75.267 45.227
277443 101.863 -75.268 45.197
277444 101.880 -75.269 45.169
277445 101.898 -75.270 45.141
277446 101.915 -75.270 45.113
277447 101.933 -75.270 45.084
277448 101.950 -75.271 45.056
277449 101.967 -75.271 45.028
277450 101.985 -75.272 45.001
277451 102.003 -75.273 44.974
277452 102.020 -75.273 44.947
277453 102.037 -75.273 44.921
277814 92.411 -73.007 54.742
277815 92.442 -73.018 54.726
277816 92.473 -73.028 54.710
277817 92.503 -73.039 54.695
277818 92.534 -73.049 54.680
277819 92.564 -73.059 54.665
277820 92.595 -73.070 54.650
277821 92.626 -73.080 54.637
277822 92.661 -73.090 54.623
277823 92.704 -73.101 54.610
277824 92.746 -73.112 54.599
277825 92.789 -73.123 54.587
277826 92.832 -73.134 54.576
277827 92.875 -73.145 54.565
277828 92.918 -73.156 54.556
277829 92.961 -73.167 54.548
277830 93.004 -73.178 54.541
277831 93.047 -73.189 54.535
277832 93.089 -73.200 54.529
277833 93.132 -73.211 54.522
277834 93.176 -73.222 54.515
277835 93.218 -73.233 54.508
277836 93.261 -73.244 54.503
277837 93.293 -73.252 54.497
277838 93.318 -73.258 54.490
277839 93.342 -73.264 54.481
277840 93.366 -73.270 54.472
277841 93.391 -73.276 54.464
277842 93.415 -73.282 54.455
277843 93.440 -73.288 54.446
277844 93.478 -73.281 54.437
277845 93.503 -73.275 54.425
277846 93.529 -73.270 54.412
277847 93.555 -73.264 54.399
277848 93.581 -73.258 54.385
277849 93.603 -73.252 54.371
277850 93.622 -73.248 54.357
277851 93.641 -73.242 54.341
277852 93.659 -73.237 54.321
277853 93.679 -73.232 54.301
277854 93.697 -73.227 54.282
277908 94.871 -73.325 52.984
277909 94.887 -73.328 52.955
277910 94.907 -73.332 52.926
277911 94.928 -73.336 52.897
277912 94.948 -73.341 52.868
277913 94.969 -73.346 52.839
277914 94.989 -73.350 52.810
277915 95.010 -73.354 52.781
277916 95.030 -73.358 52.752
277917 95.072 -73.348 52.724
277918 95.089 -73.359 52.696
277919 95.103 -73.372 52.669
277920 95.118 -73.386 52.642
277921 95.132 -73.399 52.614
277922 95.146 -73.412 52.587
277923 95.161 -73.425 52.560
277924 95.175 -73.439 52.534
277925 95.190 -73.452 52.508
277926 95.234 -73.409 52.482
277927 95.278 -73.365 52.456
277928 95.326 -73.357 52.431
277929 95.375 -73.349 52.405
277930 95.423 -73.341 52.379
277931 95.473 -73.340 52.355
277932 95.506 -73.338 52.332
277933 95.533 -73.336 52.310
277934 95.561 -73.334 52.287
277935 95.590 -73.335 52.265
277936 95.618 -73.336 52.243
277937 95.647 -73.336 52.221
277938 95.675 -73.337 52.200
277939 95.702 -73.331 52.180
277940 95.732 -73.337 52.160
277941 95.767 -73.356 52.140
277942 95.806 -73.368 52.118
277943 95.843 -73.380 52.095
277944 95.881 -73.392 52.068
277945 95.902 -73.393 52.040
277946 95.923 -73.393 52.011
277947 95.944 -73.393 51.981
277948 95.965 -73.394 51.950
277949 95.985 -73.394 51.918
277950 96.012 -73.399 51.886
277951 96.038 -73.403 51.854
277952 96.064 -73.408 51.824
277953 96.091 -73.412 51.792
277954 96.118 -73.418 51.760
277955 96.145 -73.422 51.728
277956 96.179 -73.427 51.696
277957 96.225 -73.430 51.665
277958 96.272 -73.432 51.633
277959 96.319 -73.434 51.601
277960 96.348 -73.434 51.569
277961 96.366 -73.433 51.537
277962 96.383 -73.431 51.506
277963 96.400 -73.430 51.474
277964 96.418 -73.428 51.442
277965 96.436 -73.427 51.412
277966 96.453 -73.426 51.383
277967 96.470 -73.424 51.353
277968 96.489 -73.427 51.324
277969 96.516 -73.449 51.296
277970 96.541 -73.459 51.268
277971 96.577 -73.470 51.240
277972 96.621 -73.487 51.213
277973 96.665 -73.504 51.186
277974 96.714 -73.524 51.160
277975 96.763 -73.547 51.134
277976 96.811 -73.574 51.109
277977 96.858 -73.600 51.087
277978 96.874 -73.603 51.063
277979 96.875 -73.593 51.042
277980 96.875 -73.584 51.023
277981 96.876 -73.576 51.005
277982 96.886 -73.563 50.989
277983 96.926 -73.543 50.974
277984 96.960 -73.532 50.960
277985 96.980 -73.544 50.946
277986 97.025 -73.555 50.934
277987 97.066 -73.568 50.924
277988 97.100 -73.584 50.913
277989 97.147 -73.586 50.904
277990 97.185 -73.582 50.898
277991 97.224 -73.578 50.890
277992 97.262 -73.573 50.882
277993 97.345 -73.562 50.875
277994 97.338 -73.581 50.868
277995 97.309 -73.604 50.861
277996 97.294 -73.618 50.852
277997 97.287 -73.624 50.839
277998 97.273 -73.634 50.824
277999 97.299 -73.634 50.812
278000 97.337 -73.630 50.803
278001 97.376 -73.627 50.794
278002 97.415 -73.623 50.786
278003 97.453 -73.620 50.779
278004 97.493 -73.621 50.770
278005 97.532 -73.628 50.757
278006 97.573 -73.634 50.738
278007 97.613 -73.641 50.711
278008 97.653 -73.648 50.679
278009 97.693 -73.655 50.641
278010 97.734 -73.662 50.598
278011 97.774 -73.669 50.551
278012 97.814 -73.676 50.502
278013 97.854 -73.683 50.453
278014 97.894 -73.690 50.402
278015 97.935 -73.697 50.353
278016 97.974 -73.709 50.303
278017 98.011 -73.711 50.252
278018 98.047 -73.719 50.193
278019 98.084 -73.725 50.131
278020 98.120 -73.733 50.068
278021 98.156 -73.740 50.004
278022 98.193 -73.747 49.939
278023 98.229 -73.754 49.876
278024 98.265 -73.762 49.813
278025 98.301 -73.769 49.747
278026 98.346 -73.786 49.681
278027 98.348 -73.822 49.618
278028 98.350 -73.859 49.555
278029 98.352 -73.894 49.493
278030 98.358 -73.911 49.429
278031 98.368 -73.901 49.362
278032 98.364 -73.914 49.295
278033 98.500 -73.948 49.230
278034 98.571 -73.961 49.171
278035 98.604 -73.973 49.112
278036 98.640 -73.985 49.053
278037 98.675 -73.996 48.994
278038 98.711 -74.008 48.937
278039 98.747 -74.020 48.880
278040 98.783 -74.031 48.824
278041 98.827 -74.040 48.771
278042 98.831 -74.062 48.719
278043 98.834 -74.085 48.672
278044 98.837 -74.107 48.624
278045 98.843 -74.129 48.577
278046 98.880 -74.139 48.530
278047 98.918 -74.150 48.484
278048 98.955 -74.160 48.439
278049 98.995 -74.172 48.393
278050 99.035 -74.180 48.344
278051 99.068 -74.173 48.295
278052 99.094 -74.179 48.245
278053 99.121 -74.185 48.194
278054 99.147 -74.191 48.145
278055 99.173 -74.198 48.096
278056 99.200 -74.204 48.048
278057 99.226 -74.210 48.001
278058 99.238 -74.218 47.955
278059 99.252 -74.229 47.911
278060 99.295 -74.240 47.868
278061 99.338 -74.252 47.824
278062 99.381 -74.264 47.781
278063 99.425 -74.277 47.738
278064 99.468 -74.288 47.696
278065 99.511 -74.300 47.655
278066 99.554 -74.312 47.615
278067 99.595 -74.327 47.576
278068 99.636 -74.342 47.537
278069 99.677 -74.357 47.499
278070 99.709 -74.367 47.461
278071 99.753 -74.373 47.426
278072 99.783 -74.380 47.390
278073 99.814 -74.387 47.355
278074 99.845 -74.394 47.321
278075 99.875 -74.401 47.285
278076 99.906 -74.408 47.251
278077 99.937 -74.415 47.218
278078 99.968 -74.422 47.187
278079 99.998 -74.429 47.156
278080 100.028 -74.436 47.126
278081 100.059 -74.443 47.097
278082 100.090 -74.450 47.068
278083 100.113 -74.469 47.039
278084 100.135 -74.490 47.011
278085 100.156 -74.511 46.984
278086 100.178 -74.531 46.957
278087 100.200 -74.552 46.929
278088 100.221 -74.574 46.901
278089 100.243 -74.594 46.873
278090 100.271 -74.604 46.845
278091 100.301 -74.608 46.819
278092 100.333 -74.612 46.791
278093 100.363 -74.615 46.761
278094 100.394 -74.619 46.731
278095 100.425 -74.631 46.699
278096 100.456 -74.643 46.667
278097 100.486 -74.655 46.637
278098 100.517 -74.667 46.608
278099 100.548 -74.679 46.579
278100 100.578 -74.691 46.549
278101 100.609 -74.703 46.520
278102 100.639 -74.715 46.490
278103 100.670 -74.728 46.461
278104 100.701 -74.740 46.433
278105 100.731 -74.752 46.406
278106 100.762 -74.764 46.378
278107 100.792 -74.775 46.351
278108 100.819 -74.783 46.326
278109 100.844 -74.791 46.300
278110 100.864 -74.798 46.272
278111 100.884 -74.808 46.243
278112 100.909 -74.822 46.211
278113 100.936 -74.839 46.178
278114 100.964 -74.855 46.145
278115 100.991 -74.872 46.110
278116 101.022 -74.891 46.076
278117 101.055 -74.911 46.042
278118 101.089 -74.931 46.009
278119 101.121 -74.949 45.977
278120 101.150 -74.961 45.945
278121 101.179 -74.973 45.912
278122 101.207 -74.985 45.879
278123 101.236 -74.997 45.845
278124 101.265 -75.010 45.810
278125 101.298 -75.026 45.774
278126 101.337 -75.049 45.739
278127 101.375 -75.071 45.704
278128 101.409 -75.089 45.669
278129 101.443 -75.108 45.635
278130 101.476 -75.126 45.602
278131 101.510 -75.145 45.570
278132 101.543 -75.163 45.539
278133 101.577 -75.182 45.509
278134 101.611 -75.201 45.478
278135 101.644 -75.220 45.448
278136 101.678 -75.238 45.418
278137 101.711 -75.256 45.388
278138 101.742 -75.259 45.359
278139 101.771 -75.254 45.331
278140 101.801 -75.249 45.302
278141 101.830 -75.245 45.272
278142 101.843 -75.227 45.243
278143 101.860 -75.228 45.213
278144 101.878 -75.229 45.184
278145 101.895 -75.229 45.155
278146 101.912 -75.229 45.127
278147 101.930 -75.230 45.099
278148 101.948 -75.231 45.070
278149 101.965 -75.231 45.042
278150 101.982 -75.232 45.014
278151 102.000 -75.232 44.987
278152 102.017 -75.232 44.960
278153 102.035 -75.233 44.933
278154 102.052 -75.234 44.906
278515 92.407 -72.979 54.746
278516 92.438 -72.989 54.730
278517 92.469 -72.999 54.714
278518 92.499 -73.009 54.699
278519 92.529 -73.020 54.683
278520 92.560 -73.030 54.667
278521 92.591 -73.040 54.653
278522 92.622 -73.051 54.638
278523 92.657 -73.061 54.624
278524 92.700 -73.073 54.611
278525 92.743 -73.083 54.598
278526 92.786 -73.095 54.586
278527 92.829 -73.105 54.574
278528 92.872 -73.117 54.563
278529 92.915 -73.127 54.554
278530 92.957 -73.139 54.545
278531 93.001 -73.149 54.537
278532 93.044 -73.160 54.531
278533 93.086 -73.171 54.522
278534 93.129 -73.182 54.514
278535 93.172 -73.193 54.506
278536 93.215 -73.204 54.501
278537 93.258 -73.215 54.494
278538 93.285 -73.222 54.487
278539 93.309 -73.228 54.479
278540 93.334 -73.234 54.471
278541 93.358 -73.240 54.461
278542 93.383 -73.246 54.452
278543 93.407 -73.252 54.444
278544 93.435 -73.255 54.434
278545 93.473 -73.247 54.425
278546 93.498 -73.241 54.413
278547 93.524 -73.236 54.400
278548 93.550 -73.230 54.386
278549 93.576 -73.224 54.372
278550 93.602 -73.218 54.357
278551 93.624 -73.213 54.342
278552 93.643 -73.208 54.326
278553 93.662 -73.203 54.306
278554 93.681 -73.198 54.287
278611 94.910 -73.297 52.912
278612 94.931 -73.302 52.884
278613 94.951 -73.306 52.856
278614 94.972 -73.311 52.827
278615 94.992 -73.315 52.799
278616 95.013 -73.319 52.770
278617 95.034 -73.324 52.742
278618 95.054 -73.328 52.714
278619 95.093 -73.319 52.687
278620 95.111 -73.330 52.661
278621 95.125 -73.343 52.634
278622 95.139 -73.356 52.607
278623 95.154 -73.370 52.581
278624 95.169 -73.383 52.554
278625 95.183 -73.396 52.529
278626 95.197 -73.409 52.503
278627 95.230 -73.388 52.477
278628 95.274 -73.343 52.452
278629 95.319 -73.307 52.426
278630 95.369 -73.308 52.400
278631 95.418 -73.309 52.374
278632 95.468 -73.311 52.350
278633 95.503 -73.309 52.327
278634 95.531 -73.308 52.305
278635 95.559 -73.305 52.283
278636 95.587 -73.303 52.261
278637 95.615 -73.302 52.239
278638 95.643 -73.299 52.218
278639 95.671 -73.298 52.197
278640 95.699 -73.296 52.177
278641 95.735 -73.297 52.158
278642 95.773 -73.308 52.138
278643 95.811 -73.320 52.117
278644 95.849 -73.332 52.094
278645 95.888 -73.344 52.069
278646 95.925 -73.355 52.041
278647 95.946 -73.356 52.014
278648 95.967 -73.356 51.985
278649 95.988 -73.357 51.955
278650 96.009 -73.358 51.925
278651 96.034 -73.362 51.894
278652 96.061 -73.366 51.863
278653 96.088 -73.371 51.833
278654 96.114 -73.375 51.802
278655 96.141 -73.380 51.770
278656 96.167 -73.383 51.738
278657 96.185 -73.379 51.706
278658 96.226 -73.380 51.674
278659 96.272 -73.383 51.643
278660 96.319 -73.384 51.611
278661 96.337 -73.383 51.579
278662 96.355 -73.381 51.548
278663 96.373 -73.380 51.517
278664 96.390 -73.379 51.486
278665 96.407 -73.377 51.455
278666 96.425 -73.376 51.425
278667 96.445 -73.381 51.396
278668 96.471 -73.404 51.368
278669 96.498 -73.427 51.340
278670 96.525 -73.449 51.311
278671 96.550 -73.452 51.283
278672 96.575 -73.456 51.255
278673 96.611 -73.467 51.228
278674 96.655 -73.484 51.201
278675 96.704 -73.506 51.174
278676 96.752 -73.531 51.149
278677 96.799 -73.557 51.124
278678 96.847 -73.583 51.100
278679 96.859 -73.583 51.077
278680 96.860 -73.575 51.055
278681 96.860 -73.565 51.036
278682 96.857 -73.551 51.017
278683 96.873 -73.515 51.000
278684 96.913 -73.496 50.985
278685 96.959 -73.493 50.971
278686 97.000 -73.506 50.955
278687 97.035 -73.520 50.943
278688 97.070 -73.534 50.931
278689 97.104 -73.549 50.919
278690 97.160 -73.568 50.910
278691 97.208 -73.568 50.902
278692 97.246 -73.563 50.894
278693 97.319 -73.556 50.886
278694 97.307 -73.541 50.880
278695 97.314 -73.565 50.873
278696 97.303 -73.590 50.867
278697 97.292 -73.600 50.858
278698 97.277 -73.609 50.845
278699 97.269 -73.618 50.828
278700 97.308 -73.615 50.815
278701 97.346 -73.611 50.804
278702 97.385 -73.608 50.795
278703 97.423 -73.604 50.786
278704 97.462 -73.601 50.777
278705 97.500 -73.597 50.766
278706 97.539 -73.594 50.751
278707 97.578 -73.594 50.729
278708 97.618 -73.602 50.702
278709 97.658 -73.609 50.670
278710 97.699 -73.615 50.633
278711 97.739 -73.622 50.590
278712 97.779 -73.629 50.543
278713 97.819 -73.636 50.495
278714 97.859 -73.643 50.445
278715 97.899 -73.650 50.395
278716 97.940 -73.657 50.344
278717 97.978 -73.681 50.293
278718 98.016 -73.697 50.239
278719 98.054 -73.691 50.179
278720 98.090 -73.687 50.116
278721 98.127 -73.694 50.052
278722 98.163 -73.702 49.988
278723 98.200 -73.709 49.924
278724 98.236 -73.716 49.861
278725 98.272 -73.723 49.798
278726 98.308 -73.731 49.734
278727 98.335 -73.759 49.669
278728 98.337 -73.794 49.608
278729 98.339 -73.831 49.546
278730 98.341 -73.866 49.483
278731 98.343 -73.898 49.418
278732 98.348 -73.893 49.351
278733 98.447 -73.921 49.285
278734 98.510 -73.929 49.221
278735 98.574 -73.938 49.161
278736 98.594 -73.951 49.101
278737 98.630 -73.963 49.043
278738 98.666 -73.975 48.985
278739 98.702 -73.986 48.928
278740 98.737 -73.998 48.871
278741 98.773 -74.010 48.817
278742 98.828 -74.020 48.763
278743 98.838 -74.033 48.710
278744 98.841 -74.057 48.661
278745 98.844 -74.079 48.613
278746 98.858 -74.098 48.565
278747 98.895 -74.108 48.518
278748 98.933 -74.119 48.472
278749 98.971 -74.129 48.427
278750 99.008 -74.139 48.381
278751 99.048 -74.151 48.334
278752 99.089 -74.164 48.285
278753 99.121 -74.158 48.236
278754 99.148 -74.164 48.186
278755 99.174 -74.171 48.137
278756 99.200 -74.177 48.088
278757 99.227 -74.183 48.040
278758 99.237 -74.192 47.993
278759 99.242 -74.202 47.948
278760 99.254 -74.212 47.904
278761 99.297 -74.224 47.860
278762 99.341 -74.236 47.817
278763 99.384 -74.248 47.774
278764 99.427 -74.260 47.731
278765 99.471 -74.271 47.689
278766 99.514 -74.283 47.648
278767 99.557 -74.295 47.608
278768 99.600 -74.307 47.569
278769 99.644 -74.319 47.530
278770 99.695 -74.330 47.492
278771 99.721 -74.336 47.454
278772 99.752 -74.342 47.419
278773 99.783 -74.349 47.383
278774 99.814 -74.356 47.348
278775 99.844 -74.363 47.314
278776 99.875 -74.371 47.279
278777 99.906 -74.377 47.245
278778 99.936 -74.384 47.213
278779 99.967 -74.391 47.181
278780 99.997 -74.398 47.151
278781 100.028 -74.405 47.121
278782 100.059 -74.412 47.092
278783 100.089 -74.419 47.062
278784 100.116 -74.433 47.034
278785 100.137 -74.454 47.006
278786 100.159 -74.475 46.979
278787 100.181 -74.496 46.951
278788 100.202 -74.517 46.922
278789 100.224 -74.538 46.894
278790 100.246 -74.559 46.865
278791 100.276 -74.563 46.837
278792 100.307 -74.568 46.811
278793 100.338 -74.571 46.783
278794 100.369 -74.575 46.753
278795 100.400 -74.579 46.721
278796 100.430 -74.583 46.689
278797 100.461 -74.595 46.657
278798 100.492 -74.607 46.626
278799 100.522 -74.619 46.597
278800 100.553 -74.631 46.568
278801 100.583 -74.643 46.538
278802 100.614 -74.656 46.509
278803 100.645 -74.668 46.480
278804 100.676 -74.680 46.450
278805 100.706 -74.692 46.421
278806 100.737 -74.704 46.394
278807 100.767 -74.716 46.367
278808 100.798 -74.728 46.340
278809 100.829 -74.740 46.316
278810 100.864 -74.760 46.290
278811 100.886 -74.772 46.261
278812 100.905 -74.781 46.231
278813 100.932 -74.797 46.198
278814 100.960 -74.813 46.164
278815 100.987 -74.830 46.129
278816 101.014 -74.847 46.095
278817 101.042 -74.863 46.060
278818 101.072 -74.881 46.026
278819 101.105 -74.902 45.992
278820 101.138 -74.922 45.960
278821 101.169 -74.937 45.928
278822 101.198 -74.949 45.895
278823 101.227 -74.961 45.861
278824 101.255 -74.973 45.826
278825 101.284 -74.985 45.791
278826 101.320 -75.004 45.755
278827 101.358 -75.027 45.721
278828 101.397 -75.050 45.686
278829 101.431 -75.069 45.653
278830 101.465 -75.087 45.619
278831 101.498 -75.105 45.586
278832 101.532 -75.124 45.555
278833 101.566 -75.142 45.524
278834 101.599 -75.161 45.493
278835 101.633 -75.179 45.462
278836 101.667 -75.198 45.433
278837 101.700 -75.217 45.403
278838 101.732 -75.225 45.374
278839 101.761 -75.220 45.345
278840 101.790 -75.215 45.317
278841 101.820 -75.210 45.287
278842 101.849 -75.205 45.258
278843 101.858 -75.188 45.228
278844 101.875 -75.189 45.199
278845 101.893 -75.189 45.170
278846 101.910 -75.189 45.142
278847 101.927 -75.190 45.113
278848 101.945 -75.191 45.084
278849 101.962 -75.191 45.056
278850 101.980 -75.192 45.027
278851 101.997 -75.192 45.000
278852 102.015 -75.192 44.973
278853 102.032 -75.193 44.946
278854 102.050 -75.194 44.919
279217 92.433 -72.960 54.735
279218 92.464 -72.970 54.718
279219 92.495 -72.980 54.702
279220 92.525 -72.991 54.686
279221 92.556 -73.001 54.670
279222 92.587 -73.011 54.655
279223 92.618 -73.022 54.640
279224 92.654 -73.032 54.625
279225 92.697 -73.043 54.612
279226 92.740 -73.054 54.599
279227 92.783 -73.065 54.585
279228 92.826 -73.076 54.572
279229 92.869 -73.087 54.561
279230 92.912 -73.098 54.552
279231 92.954 -73.109 54.542
279232 92.997 -73.120 54.533
279233 93.040 -73.131 54.526
279234 93.083 -73.142 54.517
279235 93.126 -73.153 54.508
279236 93.169 -73.164 54.500
279237 93.212 -73.175 54.493
279238 93.252 -73.186 54.486
279239 93.277 -73.192 54.479
279240 93.301 -73.198 54.471
279241 93.326 -73.204 54.462
279242 93.350 -73.210 54.452
279243 93.375 -73.215 54.443
279244 93.399 -73.221 54.433
279245 93.432 -73.221 54.424
279246 93.468 -73.213 54.415
279247 93.494 -73.208 54.402
279248 93.519 -73.202 54.388
279249 93.545 -73.196 54.374
279250 93.571 -73.190 54.359
279251 93.596 -73.185 54.344
279252 93.623 -73.179 54.329
279253 93.645 -73.174 54.313
279254 93.664 -73.168 54.293
279314 94.955 -73.271 52.843
279315 94.975 -73.276 52.815
279316 94.995 -73.280 52.788
279317 95.016 -73.284 52.760
279318 95.037 -73.289 52.733
279319 95.057 -73.293 52.706
279320 95.077 -73.298 52.679
279321 95.114 -73.290 52.653
279322 95.133 -73.300 52.627
279323 95.147 -73.314 52.601
279324 95.162 -73.327 52.575
279325 95.176 -73.340 52.549
279326 95.190 -73.353 52.523
279327 95.205 -73.367 52.497
279328 95.226 -73.368 52.472
279329 95.270 -73.323 52.448
279330 95.314 -73.279 52.422
279331 95.364 -73.280 52.396
279332 95.413 -73.281 52.370
279333 95.462 -73.283 52.345
279334 95.498 -73.280 52.323
279335 95.528 -73.279 52.301
279336 95.556 -73.277 52.279
279337 95.585 -73.274 52.257
279338 95.612 -73.273 52.236
279339 95.640 -73.271 52.214
279340 95.668 -73.269 52.194
279341 95.696 -73.267 52.175
279342 95.733 -73.277 52.156
279343 95.773 -73.280 52.137
279344 95.817 -73.274 52.116
279345 95.855 -73.283 52.093
279346 95.894 -73.296 52.069
279347 95.932 -73.308 52.043
279348 95.969 -73.319 52.017
279349 95.990 -73.320 51.990
279350 96.011 -73.320 51.962
279351 96.032 -73.321 51.932
279352 96.057 -73.324 51.903
279353 96.084 -73.329 51.873
279354 96.110 -73.333 51.844
279355 96.137 -73.338 51.813
279356 96.163 -73.339 51.781
279357 96.189 -73.328 51.749
279358 96.201 -73.328 51.716
279359 96.226 -73.330 51.685
279360 96.272 -73.333 51.653
279361 96.310 -73.333 51.622
279362 96.327 -73.332 51.590
279363 96.344 -73.330 51.559
279364 96.362 -73.329 51.528
279365 96.379 -73.328 51.498
279366 96.400 -73.336 51.469
279367 96.427 -73.358 51.439
279368 96.454 -73.381 51.411
279369 96.480 -73.404 51.383
279370 96.508 -73.427 51.355
279371 96.534 -73.442 51.327
279372 96.559 -73.446 51.299
279373 96.584 -73.449 51.271
279374 96.609 -73.452 51.244
279375 96.646 -73.464 51.217
279376 96.693 -73.488 51.190
279377 96.741 -73.514 51.165
279378 96.788 -73.540 51.140
279379 96.835 -73.566 51.116
279380 96.843 -73.564 51.092
279381 96.844 -73.555 51.070
279382 96.823 -73.511 51.049
279383 96.812 -73.457 51.030
279384 96.855 -73.444 51.012
279385 96.903 -73.455 50.996
279386 96.953 -73.465 50.981
279387 96.996 -73.477 50.966
279388 97.035 -73.488 50.952
279389 97.073 -73.500 50.939
279390 97.108 -73.515 50.926
279391 97.143 -73.529 50.916
279392 97.230 -73.553 50.907
279393 97.268 -73.549 50.899
279394 97.279 -73.532 50.891
279395 97.272 -73.526 50.884
279396 97.280 -73.550 50.878
279397 97.295 -73.575 50.872
279398 97.282 -73.585 50.865
279399 97.267 -73.595 50.853
279400 97.278 -73.599 50.836
279401 97.316 -73.595 50.822
279402 97.355 -73.592 50.809
279403 97.393 -73.589 50.797
279404 97.432 -73.585 50.785
279405 97.470 -73.582 50.772
279406 97.509 -73.578 50.757
279407 97.547 -73.575 50.740
279408 97.586 -73.571 50.717
279409 97.624 -73.568 50.690
279410 97.663 -73.568 50.657
279411 97.704 -73.575 50.621
279412 97.744 -73.582 50.580
279413 97.784 -73.589 50.533
279414 97.824 -73.596 50.485
279415 97.864 -73.603 50.435
279416 97.904 -73.609 50.384
279417 97.943 -73.629 50.333
279418 97.982 -73.653 50.279
279419 98.020 -73.676 50.224
279420 98.058 -73.684 50.163
279421 98.095 -73.679 50.100
279422 98.133 -73.675 50.036
279423 98.170 -73.669 49.972
279424 98.206 -73.671 49.909
279425 98.243 -73.678 49.846
279426 98.279 -73.685 49.783
279427 98.320 -73.694 49.720
279428 98.322 -73.731 49.656
279429 98.324 -73.767 49.596
279430 98.326 -73.803 49.535
279431 98.329 -73.839 49.472
279432 98.331 -73.875 49.406
279433 98.381 -73.890 49.340
279434 98.444 -73.898 49.275
279435 98.507 -73.907 49.213
279436 98.571 -73.915 49.151
279437 98.596 -73.928 49.091
279438 98.621 -73.941 49.033
279439 98.656 -73.953 48.976
279440 98.692 -73.964 48.919
279441 98.727 -73.976 48.864
279442 98.781 -73.998 48.810
279443 98.830 -73.999 48.756
279444 98.844 -74.005 48.703
279445 98.847 -74.028 48.652
279446 98.850 -74.051 48.602
279447 98.872 -74.067 48.554
279448 98.910 -74.077 48.507
279449 98.948 -74.087 48.462
279450 98.985 -74.098 48.416
279451 99.023 -74.107 48.371
279452 99.061 -74.117 48.323
279453 99.101 -74.130 48.275
279454 99.142 -74.144 48.227
279455 99.174 -74.145 48.178
279456 99.201 -74.151 48.129
279457 99.227 -74.157 48.079
279458 99.236 -74.166 48.031
279459 99.241 -74.176 47.985
279460 99.246 -74.186 47.941
279461 99.257 -74.195 47.896
279462 99.300 -74.208 47.853
279463 99.343 -74.220 47.810
279464 99.386 -74.231 47.767
279465 99.430 -74.243 47.724
279466 99.473 -74.255 47.682
279467 99.516 -74.267 47.642
279468 99.564 -74.278 47.601
279469 99.625 -74.286 47.563
279470 99.685 -74.295 47.524
279471 99.707 -74.302 47.485
279472 99.730 -74.309 47.449
279473 99.755 -74.313 47.412
279474 99.782 -74.318 47.377
279475 99.813 -74.326 47.343
279476 99.844 -74.333 47.309
279477 99.874 -74.339 47.274
279478 99.905 -74.346 47.241
279479 99.935 -74.354 47.208
279480 99.966 -74.361 47.176
279481 99.997 -74.368 47.147
279482 100.028 -74.374 47.117
279483 100.058 -74.381 47.087
279484 100.089 -74.389 47.058
279485 100.118 -74.398 47.030
279486 100.140 -74.419 47.002
279487 100.161 -74.440 46.974
279488 100.183 -74.461 46.946
279489 100.205 -74.481 46.917
279490 100.226 -74.502 46.888
279491 100.251 -74.519 46.859
279492 100.281 -74.523 46.830
279493 100.312 -74.527 46.803
279494 100.343 -74.530 46.774
279495 100.374 -74.534 46.743
279496 100.405 -74.538 46.712
279497 100.436 -74.542 46.679
279498 100.467 -74.547 46.647
279499 100.497 -74.559 46.616
279500 100.528 -74.571 46.586
279501 100.558 -74.584 46.557
279502 100.589 -74.596 46.528
279503 100.620 -74.608 46.498
279504 100.650 -74.620 46.468
279505 100.681 -74.632 46.439
279506 100.711 -74.644 46.410
279507 100.742 -74.656 46.382
279508 100.773 -74.668 46.355
279509 100.804 -74.681 46.328
279510 100.834 -74.693 46.304
279511 100.882 -74.729 46.278
279512 100.907 -74.744 46.248
279513 100.928 -74.755 46.217
279514 100.956 -74.772 46.185
279515 100.983 -74.788 46.150
279516 101.010 -74.805 46.115
279517 101.038 -74.822 46.080
279518 101.065 -74.838 46.044
279519 101.092 -74.855 46.009
279520 101.121 -74.872 45.976
279521 101.155 -74.892 45.943
279522 101.188 -74.913 45.911
279523 101.217 -74.925 45.878
279524 101.246 -74.937 45.844
279525 101.275 -74.949 45.807
279526 101.304 -74.961 45.772
279527 101.341 -74.983 45.737
279528 101.379 -75.005 45.703
279529 101.418 -75.028 45.669
279530 101.454 -75.048 45.636
279531 101.487 -75.066 45.603
279532 101.521 -75.085 45.571
279533 101.554 -75.103 45.540
279534 101.588 -75.122 45.509
279535 101.622 -75.140 45.478
279536 101.656 -75.159 45.447
279537 101.689 -75.177 45.418
279538 101.722 -75.191 45.388
279539 101.751 -75.186 45.359
279540 101.780 -75.181 45.331
279541 101.810 -75.176 45.302
279542 101.839 -75.171 45.273
279543 101.863 -75.160 45.244
279544 101.873 -75.148 45.215
279545 101.890 -75.148 45.185
279546 101.908 -75.149 45.156
279547 101.925 -75.150 45.127
279548 101.943 -75.150 45.099
279549 101.960 -75.151 45.070
279550 101.977 -75.151 45.042
279551 101.995 -75.152 45.014
279552 102.013 -75.152 44.986
279553 102.030 -75.153 44.959
279554 102.047 -75.154 44.931
279555 102.065 -75.154 44.904
279918 92.429 -72.930 54.739
279919 92.460 -72.941 54.723
279920 92.491 -72.951 54.707
279921 92.521 -72.961 54.690
279922 92.552 -72.972 54.673
279923 92.582 -72.982 54.657
279924 92.613 -72.992 54.642
279925 92.651 -73.003 54.628
279926 92.694 -73.014 54.613
279927 92.737 -73.025 54.599
279928 92.780 -73.036 54.585
279929 92.823 -73.047 54.571
279930 92.865 -73.058 54.560
279931 92.908 -73.069 54.549
279932 92.951 -73.080 54.539
279933 92.994 -73.091 54.529
279934 93.037 -73.102 54.521
279935 93.080 -73.113 54.512
279936 93.123 -73.124 54.503
279937 93.165 -73.135 54.495
279938 93.209 -73.146 54.487
279939 93.244 -73.155 54.480
279940 93.269 -73.161 54.472
279941 93.293 -73.167 54.464
279942 93.318 -73.174 54.454
279943 93.342 -73.180 54.444
279944 93.367 -73.185 54.434
279945 93.391 -73.191 54.423
279946 93.430 -73.186 54.414
279947 93.463 -73.179 54.404
279948 93.489 -73.174 54.391
279949 93.514 -73.167 54.377
279950 93.540 -73.162 54.362
279951 93.566 -73.157 54.347
279952 93.592 -73.151 54.331
279953 93.618 -73.145 54.316
279954 93.643 -73.139 54.299
279955 93.666 -73.134 54.279
280017 94.998 -73.246 52.778
280018 95.019 -73.249 52.751
280019 95.040 -73.254 52.725
280020 95.060 -73.258 52.698
280021 95.080 -73.263 52.672
280022 95.101 -73.267 52.646
280023 95.135 -73.261 52.621
280024 95.155 -73.271 52.596
280025 95.169 -73.284 52.570
280026 95.183 -73.298 52.544
280027 95.198 -73.311 52.518
280028 95.212 -73.324 52.493
280029 95.227 -73.337 52.468
280030 95.266 -73.303 52.443
280031 95.310 -73.258 52.418
280032 95.358 -73.251 52.392
280033 95.408 -73.253 52.367
280034 95.455 -73.254 52.342
280035 95.492 -73.252 52.318
280036 95.526 -73.249 52.296
280037 95.554 -73.248 52.274
280038 95.582 -73.246 52.253
280039 95.610 -73.244 52.232
280040 95.638 -73.242 52.211
280041 95.666 -73.240 52.191
280042 95.695 -73.240 52.172
280043 95.733 -73.252 52.154
280044 95.771 -73.263 52.135
280045 95.812 -73.264 52.115
280046 95.855 -73.258 52.094
280047 95.898 -73.252 52.071
280048 95.938 -73.259 52.045
280049 95.976 -73.271 52.020
280050 96.013 -73.283 51.995
280051 96.034 -73.283 51.970
280052 96.055 -73.283 51.942
280053 96.080 -73.287 51.913
280054 96.107 -73.292 51.884
280055 96.133 -73.296 51.855
280056 96.160 -73.293 51.824
280057 96.186 -73.283 51.792
280058 96.210 -73.275 51.760
280059 96.217 -73.278 51.728
280060 96.226 -73.281 51.696
280061 96.272 -73.283 51.664
280062 96.299 -73.282 51.633
280063 96.316 -73.281 51.602
280064 96.333 -73.280 51.571
280065 96.356 -73.291 51.541
280066 96.383 -73.314 51.512
280067 96.409 -73.336 51.484
280068 96.436 -73.358 51.454
280069 96.463 -73.381 51.426
280070 96.490 -73.404 51.398
280071 96.517 -73.427 51.369
280072 96.542 -73.435 51.343
280073 96.568 -73.439 51.316
280074 96.593 -73.443 51.289
280075 96.617 -73.446 51.261
280076 96.639 -73.440 51.235
280077 96.682 -73.471 51.207
280078 96.729 -73.497 51.181
280079 96.776 -73.524 51.157
280080 96.823 -73.550 51.133
280081 96.828 -73.544 51.108
280082 96.789 -73.471 51.085
280083 96.751 -73.399 51.063
280084 96.799 -73.405 51.043
280085 96.849 -73.416 51.025
280086 96.897 -73.426 51.008
280087 96.946 -73.437 50.992
280088 96.992 -73.448 50.976
280089 97.030 -73.459 50.961
280090 97.069 -73.471 50.947
280091 97.108 -73.484 50.934
280092 97.147 -73.496 50.924
280093 97.182 -73.509 50.914
280094 97.251 -73.524 50.905
280095 97.240 -73.509 50.896
280096 97.236 -73.511 50.888
280097 97.259 -73.535 50.882
280098 97.283 -73.560 50.878
280099 97.272 -73.571 50.872
280100 97.257 -73.581 50.862
280101 97.286 -73.580 50.848
280102 97.324 -73.576 50.832
280103 97.363 -73.573 50.816
280104 97.401 -73.569 50.800
280105 97.440 -73.566 50.784
280106 97.478 -73.562 50.766
280107 97.517 -73.559 50.746
280108 97.556 -73.556 50.726
280109 97.594 -73.553 50.702
280110 97.633 -73.550 50.674
280111 97.671 -73.546 50.642
280112 97.710 -73.543 50.606
280113 97.749 -73.542 50.566
280114 97.789 -73.549 50.521
280115 97.829 -73.556 50.472
280116 97.869 -73.562 50.423
280117 97.909 -73.578 50.371
280118 97.947 -73.601 50.318
280119 97.986 -73.625 50.263
280120 98.024 -73.647 50.206
280121 98.062 -73.671 50.143
280122 98.100 -73.672 50.081
280123 98.137 -73.667 50.018
280124 98.174 -73.662 49.956
280125 98.212 -73.657 49.893
280126 98.249 -73.652 49.831
280127 98.286 -73.647 49.768
280128 98.309 -73.667 49.704
280129 98.314 -73.703 49.642
280130 98.322 -73.739 49.582
280131 98.331 -73.775 49.521
280132 98.338 -73.812 49.458
280133 98.347 -73.847 49.393
280134 98.381 -73.866 49.330
280135 98.441 -73.875 49.266
280136 98.504 -73.883 49.203
280137 98.567 -73.891 49.142
280138 98.605 -73.904 49.083
280139 98.612 -73.919 49.026
280140 98.646 -73.931 48.969
280141 98.682 -73.943 48.913
280142 98.718 -73.954 48.857
280143 98.802 -73.994 48.802
280144 98.831 -73.979 48.748
280145 98.851 -73.977 48.695
280146 98.854 -74.000 48.644
280147 98.857 -74.023 48.595
280148 98.887 -74.035 48.546
280149 98.925 -74.045 48.499
280150 98.962 -74.056 48.453
280151 99.000 -74.066 48.407
280152 99.038 -74.076 48.361
280153 99.075 -74.086 48.313
280154 99.113 -74.096 48.266
280155 99.154 -74.110 48.218
280156 99.196 -74.123 48.168
280157 99.228 -74.130 48.119
280158 99.235 -74.140 48.072
280159 99.240 -74.150 48.024
280160 99.245 -74.160 47.978
280161 99.251 -74.169 47.934
280162 99.259 -74.179 47.890
280163 99.302 -74.191 47.846
280164 99.346 -74.203 47.804
280165 99.389 -74.215 47.761
280166 99.432 -74.226 47.718
280167 99.493 -74.236 47.676
280168 99.554 -74.244 47.636
280169 99.615 -74.252 47.596
280170 99.672 -74.261 47.557
280171 99.695 -74.268 47.518
280172 99.717 -74.275 47.480
280173 99.740 -74.282 47.443
280174 99.764 -74.286 47.408
280175 99.788 -74.291 47.373
280176 99.812 -74.295 47.338
280177 99.843 -74.302 47.305
280178 99.873 -74.309 47.270
280179 99.904 -74.316 47.236
280180 99.935 -74.323 47.204
280181 99.966 -74.330 47.173
280182 99.996 -74.337 47.143
280183 100.027 -74.344 47.113
280184 100.058 -74.351 47.083
280185 100.088 -74.358 47.054
280186 100.119 -74.364 47.026
280187 100.142 -74.383 46.997
280188 100.164 -74.405 46.969
280189 100.186 -74.425 46.940
280190 100.207 -74.446 46.910
280191 100.229 -74.467 46.881
280192 100.256 -74.478 46.851
280193 100.286 -74.482 46.823
280194 100.318 -74.486 46.796
280195 100.348 -74.490 46.766
280196 100.379 -74.493 46.734
280197 100.410 -74.497 46.703
280198 100.441 -74.501 46.670
280199 100.472 -74.505 46.637
280200 100.503 -74.512 46.606
280201 100.533 -74.524 46.576
280202 100.564 -74.536 46.547
280203 100.595 -74.548 46.517
280204 100.625 -74.560 46.487
280205 100.656 -74.572 46.457
280206 100.686 -74.584 46.428
280207 100.717 -74.596 46.399
280208 100.748 -74.609 46.371
280209 100.778 -74.621 46.343
280210 100.809 -74.633 46.315
280211 100.844 -74.651 46.290
280212 100.901 -74.699 46.263
280213 100.928 -74.717 46.234
280214 100.952 -74.730 46.203
280215 100.979 -74.747 46.170
280216 101.006 -74.763 46.136
280217 101.034 -74.779 46.102
280218 101.061 -74.796 46.065
280219 101.088 -74.812 46.028
280220 101.116 -74.829 45.993
280221 101.143 -74.846 45.959
280222 101.171 -74.863 45.927
280223 101.204 -74.883 45.895
280224 101.236 -74.901 45.861
280225 101.265 -74.913 45.827
280226 101.294 -74.925 45.789
280227 101.324 -74.938 45.754
280228 101.362 -74.961 45.720
280229 101.401 -74.984 45.686
280230 101.440 -75.006 45.653
280231 101.476 -75.026 45.619
280232 101.510 -75.045 45.587
280233 101.543 -75.063 45.555
280234 101.577 -75.082 45.524
280235 101.611 -75.101 45.495
280236 101.644 -75.119 45.464
280237 101.678 -75.138 45.434
280238 101.711 -75.156 45.404
280239 101.741 -75.152 45.374
280240 101.770 -75.147 45.345
280241 101.799 -75.142 45.316
280242 101.829 -75.137 45.287
280243 101.858 -75.132 45.258
280244 101.871 -75.109 45.229
280245 101.888 -75.108 45.200
280246 101.905 -75.109 45.171
280247 101.922 -75.110 45.143
280248 101.940 -75.110 45.114
280249 101.958 -75.110 45.085
280250 101.975 -75.111 45.056
280251 101.992 -75.111 45.028
280252 102.010 -75.112 45.000
280253 102.027 -75.113 44.972
280254 102.045 -75.113 44.945
280255 102.062 -75.113 44.917
280619 92.425 -72.901 54.743
280620 92.456 -72.911 54.727
280621 92.487 -72.922 54.711
280622 92.517 -72.932 54.694
280623 92.548 -72.942 54.677
280624 92.578 -72.953 54.660
280625 92.609 -72.963 54.644
280626 92.648 -72.974 54.629
280627 92.691 -72.985 54.614
280628 92.733 -72.996 54.600
280629 92.776 -73.007 54.586
280630 92.819 -73.018 54.571
280631 92.862 -73.029 54.559
280632 92.905 -73.040 54.548
280633 92.948 -73.051 54.537
280634 92.991 -73.062 54.528
280635 93.034 -73.073 54.519
280636 93.077 -73.084 54.509
280637 93.120 -73.095 54.499
280638 93.162 -73.106 54.490
280639 93.205 -73.117 54.482
280640 93.236 -73.125 54.474
280641 93.261 -73.131 54.466
280642 93.285 -73.137 54.457
280643 93.309 -73.143 54.445
280644 93.334 -73.149 54.436
280645 93.358 -73.155 54.425
280646 93.383 -73.161 54.414
280647 93.428 -73.152 54.404
280648 93.458 -73.145 54.392
280649 93.484 -73.139 54.380
280650 93.509 -73.134 54.366
280651 93.536 -73.128 54.351
280652 93.561 -73.123 54.335
280653 93.587 -73.117 54.319
280654 93.613 -73.111 54.303
280655 93.638 -73.105 54.286
280720 95.043 -73.219 52.717
280721 95.063 -73.224 52.691
280722 95.084 -73.228 52.665
280723 95.104 -73.232 52.640
280724 95.125 -73.236 52.615
280725 95.157 -73.233 52.590
280726 95.176 -73.242 52.565
280727 95.191 -73.255 52.539
280728 95.205 -73.268 52.513
280729 95.219 -73.281 52.488
280730 95.234 -73.295 52.464
280731 95.262 -73.283 52.439
280732 95.306 -73.238 52.413
280733 95.353 -73.223 52.388
280734 95.403 -73.224 52.363
280735 95.449 -73.225 52.338
280736 95.485 -73.223 52.314
280737 95.522 -73.221 52.292
280738 95.551 -73.219 52.271
280739 95.580 -73.217 52.249
280740 95.608 -73.215 52.228
280741 95.635 -73.213 52.208
280742 95.664 -73.211 52.189
280743 95.695 -73.214 52.171
280744 95.733 -73.226 52.152
280745 95.771 -73.237 52.134
280746 95.809 -73.249 52.115
280747 95.851 -73.247 52.094
280748 95.894 -73.241 52.072
280749 95.937 -73.235 52.049
280750 95.980 -73.229 52.025
280751 96.020 -73.235 52.001
280752 96.057 -73.246 51.977
280753 96.079 -73.247 51.951
280754 96.103 -73.250 51.924
280755 96.129 -73.255 51.895
280756 96.156 -73.249 51.865
280757 96.182 -73.239 51.834
280758 96.208 -73.229 51.803
280759 96.226 -73.225 51.771
280760 96.232 -73.228 51.739
280761 96.239 -73.230 51.706
280762 96.270 -73.233 51.675
280763 96.288 -73.231 51.645
280764 96.312 -73.246 51.615
280765 96.338 -73.268 51.585
280766 96.365 -73.291 51.556
280767 96.392 -73.314 51.527
280768 96.418 -73.336 51.499
280769 96.446 -73.358 51.471
280770 96.472 -73.381 51.442
280771 96.499 -73.404 51.415
280772 96.526 -73.425 51.387
280773 96.551 -73.428 51.361
280774 96.576 -73.432 51.334
280775 96.601 -73.436 51.307
280776 96.608 -73.415 51.279
280777 96.637 -73.408 51.251
280778 96.677 -73.452 51.225
280779 96.723 -73.479 51.199
280780 96.768 -73.506 51.173
280781 96.812 -73.525 51.149
280782 96.794 -73.474 51.123
280783 96.768 -73.415 51.099
280784 96.758 -73.373 51.077
280785 96.793 -73.377 51.056
280786 96.842 -73.387 51.037
280787 96.891 -73.398 51.020
280788 96.940 -73.409 51.003
280789 96.987 -73.419 50.985
280790 97.026 -73.430 50.969
280791 97.065 -73.443 50.955
280792 97.104 -73.455 50.942
280793 97.142 -73.467 50.931
280794 97.181 -73.478 50.921
280795 97.197 -73.479 50.911
280796 97.206 -73.474 50.902
280797 97.223 -73.496 50.894
280798 97.247 -73.520 50.886
280799 97.271 -73.545 50.881
280800 97.262 -73.556 50.877
280801 97.256 -73.564 50.871
280802 97.295 -73.561 50.859
280803 97.333 -73.557 50.845
280804 97.372 -73.554 50.830
280805 97.410 -73.550 50.813
280806 97.449 -73.547 50.792
280807 97.487 -73.543 50.768
280808 97.526 -73.540 50.743
280809 97.564 -73.537 50.715
280810 97.603 -73.534 50.686
280811 97.641 -73.530 50.655
280812 97.680 -73.527 50.622
280813 97.718 -73.523 50.587
280814 97.757 -73.520 50.548
280815 97.795 -73.517 50.503
280816 97.834 -73.516 50.455
280817 97.874 -73.526 50.405
280818 97.913 -73.550 50.353
280819 97.951 -73.572 50.297
280820 97.989 -73.596 50.241
280821 98.028 -73.619 50.183
280822 98.066 -73.643 50.121
280823 98.104 -73.665 50.059
280824 98.142 -73.659 49.997
280825 98.179 -73.655 49.936
280826 98.216 -73.650 49.876
280827 98.253 -73.645 49.816
280828 98.291 -73.640 49.755
280829 98.322 -73.640 49.692
280830 98.330 -73.675 49.628
280831 98.338 -73.712 49.567
280832 98.346 -73.748 49.506
280833 98.354 -73.785 49.443
280834 98.368 -73.817 49.380
280835 98.398 -73.840 49.318
280836 98.439 -73.852 49.256
280837 98.501 -73.860 49.194
280838 98.564 -73.869 49.132
280839 98.615 -73.879 49.073
280840 98.621 -73.894 49.016
280841 98.637 -73.910 48.959
280842 98.672 -73.921 48.903
280843 98.708 -73.932 48.847
280844 98.803 -73.973 48.793
280845 98.832 -73.958 48.740
280846 98.858 -73.948 48.687
280847 98.861 -73.972 48.637
280848 98.864 -73.994 48.586
280849 98.902 -74.004 48.537
280850 98.940 -74.014 48.490
280851 98.978 -74.025 48.444
280852 99.015 -74.035 48.398
280853 99.053 -74.045 48.351
280854 99.090 -74.055 48.304
280855 99.128 -74.065 48.257
280856 99.166 -74.076 48.208
280857 99.207 -74.089 48.159
280858 99.240 -74.114 48.111
280859 99.252 -74.135 48.064
280860 99.250 -74.138 48.017
280861 99.250 -74.143 47.972
280862 99.255 -74.153 47.928
280863 99.262 -74.163 47.884
280864 99.305 -74.174 47.840
280865 99.361 -74.184 47.797
280866 99.422 -74.192 47.754
280867 99.483 -74.201 47.711
280868 99.544 -74.210 47.670
280869 99.605 -74.218 47.630
280870 99.659 -74.226 47.590
280871 99.682 -74.233 47.551
280872 99.704 -74.241 47.513
280873 99.727 -74.248 47.475
280874 99.750 -74.255 47.438
280875 99.774 -74.261 47.403
280876 99.798 -74.264 47.369
280877 99.822 -74.269 47.334
280878 99.846 -74.273 47.301
280879 99.873 -74.278 47.267
280880 99.904 -74.285 47.233
280881 99.934 -74.292 47.200
280882 99.965 -74.299 47.169
280883 99.996 -74.306 47.139
280884 100.027 -74.313 47.109
280885 100.057 -74.320 47.079
280886 100.087 -74.327 47.050
280887 100.118 -74.334 47.021
280888 100.144 -74.348 46.992
280889 100.166 -74.369 46.963
280890 100.188 -74.389 46.934
280891 100.210 -74.411 46.904
280892 100.231 -74.432 46.873
280893 100.261 -74.438 46.843
280894 100.292 -74.442 46.815
280895 100.323 -74.446 46.789
280896 100.354 -74.449 46.758
280897 100.385 -74.453 46.726
280898 100.415 -74.457 46.693
280899 100.446 -74.461 46.660
280900 100.477 -74.465 46.627
280901 100.508 -74.468 46.596
280902 100.539 -74.476 46.567
280903 100.569 -74.488 46.537
280904 100.600 -74.500 46.506
280905 100.631 -74.512 46.476
280906 100.662 -74.524 46.446
280907 100.692 -74.537 46.416
280908 100.722 -74.549 46.386
280909 100.753 -74.561 46.358
280910 100.784 -74.573 46.331
280911 100.815 -74.585 46.303
280912 100.862 -74.620 46.276
280913 100.918 -74.668 46.248
280914 100.950 -74.690 46.219
280915 100.975 -74.704 46.189
280916 101.002 -74.721 46.155
280917 101.030 -74.737 46.121
280918 101.057 -74.754 46.086
280919 101.084 -74.771 46.049
280920 101.112 -74.787 46.013
280921 101.139 -74.804 45.977
280922 101.166 -74.821 45.943
280923 101.194 -74.838 45.911
280924 101.221 -74.854 45.878
280925 101.254 -74.874 45.844
280926 101.284 -74.889 45.809
280927 101.313 -74.901 45.772
280928 101.345 -74.917 45.737
280929 101.384 -74.939 45.704
280930 101.423 -74.962 45.671
280931 101.462 -74.985 45.638
280932 101.498 -75.005 45.605
280933 101.532 -75.024 45.572
280934 101.565 -75.042 45.541
280935 101.599 -75.061 45.510
280936 101.633 -75.079 45.480
280937 101.667 -75.098 45.450
280938 101.700 -75.116 45.420
280939 101.731 -75.118 45.391
280940 101.760 -75.113 45.361
280941 101.789 -75.108 45.332
280942 101.818 -75.103 45.303
280943 101.847 -75.098 45.274
280944 101.877 -75.093 45.245
280945 101.885 -75.068 45.215
280946 101.903 -75.069 45.186
280947 101.920 -75.069 45.158
280948 101.938 -75.069 45.129
280949 101.955 -75.070 45.101
280950 101.972 -75.071 45.072
280951 101.990 -75.071 45.043
280952 102.007 -75.072 45.015
280953 102.025 -75.073 44.987
280954 102.042 -75.073 44.959
280955 102.060 -75.073 44.930
280956 102.077 -75.074 44.903
281320 92.421 -72.872 54.748
281321 92.451 -72.882 54.731
281322 92.482 -72.892 54.714
281323 92.513 -72.903 54.698
281324 92.544 -72.913 54.680
281325 92.574 -72.923 54.663
281326 92.605 -72.934 54.646
281327 92.645 -72.945 54.630
281328 92.688 -72.956 54.615
281329 92.730 -72.967 54.601
281330 92.773 -72.978 54.586
281331 92.816 -72.989 54.572
281332 92.859 -73.000 54.558
281333 92.902 -73.011 54.547
281334 92.945 -73.022 54.536
281335 92.988 -73.032 54.525
281336 93.030 -73.044 54.516
281337 93.073 -73.054 54.506
281338 93.117 -73.066 54.495
281339 93.159 -73.076 54.485
281340 93.201 -73.087 54.477
281341 93.228 -73.095 54.468
281342 93.253 -73.101 54.460
281343 93.277 -73.107 54.451
281344 93.301 -73.113 54.439
281345 93.326 -73.119 54.428
281346 93.350 -73.124 54.416
281347 93.378 -73.128 54.405
281348 93.426 -73.117 54.395
281349 93.453 -73.111 54.383
281350 93.479 -73.105 54.369
281351 93.505 -73.100 54.355
281352 93.531 -73.095 54.340
281353 93.556 -73.089 54.323
281354 93.582 -73.083 54.306
281355 93.608 -73.077 54.290
281421 95.046 -73.184 52.709
281422 95.066 -73.189 52.684
281423 95.087 -73.193 52.659
281424 95.107 -73.197 52.634
281425 95.128 -73.202 52.610
281426 95.148 -73.206 52.585
281427 95.178 -73.204 52.559
281428 95.198 -73.212 52.534
281429 95.213 -73.225 52.508
281430 95.227 -73.239 52.483
281431 95.242 -73.252 52.458
281432 95.258 -73.262 52.434
281433 95.302 -73.218 52.409
281434 95.348 -73.195 52.384
281435 95.398 -73.196 52.359
281436 95.443 -73.196 52.334
281437 95.479 -73.195 52.310
281438 95.516 -73.192 52.288
281439 95.549 -73.190 52.267
281440 95.577 -73.188 52.246
281441 95.605 -73.186 52.225
281442 95.633 -73.184 52.206
281443 95.661 -73.183 52.187
281444 95.695 -73.189 52.169
281445 95.734 -73.201 52.151
281446 95.771 -73.212 52.134
281447 95.810 -73.224 52.116
281448 95.847 -73.235 52.095
281449 95.890 -73.231 52.074
281450 95.933 -73.224 52.053
281451 95.976 -73.218 52.031
281452 96.019 -73.212 52.008
281453 96.062 -73.206 51.985
281454 96.102 -73.210 51.961
281455 96.126 -73.213 51.935
281456 96.152 -73.204 51.906
281457 96.178 -73.194 51.875
281458 96.204 -73.184 51.844
281459 96.231 -73.174 51.812
281460 96.241 -73.174 51.780
281461 96.248 -73.177 51.749
281462 96.255 -73.180 51.717
281463 96.269 -73.196 51.686
281464 96.294 -73.223 51.657
281465 96.321 -73.246 51.628
281466 96.347 -73.268 51.599
281467 96.374 -73.291 51.571
281468 96.401 -73.313 51.543
281469 96.428 -73.336 51.515
281470 96.455 -73.358 51.487
281471 96.481 -73.381 51.459
281472 96.508 -73.404 51.432
281473 96.534 -73.418 51.406
281474 96.559 -73.421 51.379
281475 96.584 -73.424 51.352
281476 96.585 -73.393 51.325
281477 96.607 -73.361 51.298
281478 96.635 -73.377 51.269
281479 96.671 -73.434 51.243
281480 96.717 -73.460 51.216
281481 96.762 -73.487 51.190
281482 96.798 -73.481 51.165
281483 96.797 -73.443 51.138
281484 96.771 -73.384 51.113
281485 96.771 -73.353 51.090
281486 96.802 -73.355 51.069
281487 96.836 -73.358 51.050
281488 96.885 -73.369 51.032
281489 96.934 -73.380 51.014
281490 96.983 -73.390 50.996
281491 97.022 -73.402 50.979
281492 97.061 -73.414 50.964
281493 97.099 -73.426 50.951
281494 97.138 -73.438 50.938
281495 97.166 -73.445 50.928
281496 97.186 -73.448 50.918
281497 97.195 -73.448 50.909
281498 97.214 -73.478 50.900
281499 97.235 -73.505 50.892
281500 97.259 -73.530 50.887
281501 97.252 -73.542 50.882
281502 97.264 -73.545 50.878
281503 97.303 -73.542 50.868
281504 97.342 -73.538 50.856
281505 97.380 -73.535 50.841
281506 97.419 -73.531 50.823
281507 97.457 -73.528 50.799
281508 97.496 -73.524 50.772
281509 97.534 -73.521 50.742
281510 97.573 -73.518 50.710
281511 97.611 -73.515 50.676
281512 97.650 -73.511 50.642
281513 97.688 -73.508 50.607
281514 97.727 -73.504 50.571
281515 97.766 -73.499 50.531
281516 97.804 -73.487 50.486
281517 97.843 -73.475 50.437
281518 97.886 -73.489 50.387
281519 97.916 -73.521 50.333
281520 97.955 -73.544 50.277
281521 97.993 -73.568 50.220
281522 98.031 -73.591 50.162
281523 98.070 -73.614 50.101
281524 98.108 -73.637 50.040
281525 98.146 -73.652 49.980
281526 98.183 -73.647 49.920
281527 98.221 -73.642 49.861
281528 98.258 -73.637 49.802
281529 98.295 -73.631 49.741
281530 98.334 -73.618 49.678
281531 98.346 -73.649 49.614
281532 98.354 -73.684 49.552
281533 98.362 -73.721 49.490
281534 98.370 -73.757 49.428
281535 98.388 -73.788 49.367
281536 98.415 -73.813 49.306
281537 98.455 -73.825 49.245
281538 98.498 -73.838 49.184
281539 98.561 -73.846 49.124
281540 98.624 -73.854 49.064
281541 98.631 -73.870 49.006
281542 98.637 -73.886 48.949
281543 98.662 -73.899 48.892
281544 98.725 -73.927 48.837
281545 98.805 -73.953 48.783
281546 98.834 -73.938 48.731
281547 98.862 -73.923 48.679
281548 98.867 -73.943 48.629
281549 98.879 -73.963 48.579
281550 98.917 -73.973 48.529
281551 98.954 -73.983 48.482
281552 98.992 -73.993 48.436
281553 99.030 -74.004 48.389
281554 99.067 -74.013 48.343
281555 99.105 -74.023 48.297
281556 99.143 -74.034 48.249
281557 99.181 -74.044 48.200
281558 99.219 -74.055 48.151
281559 99.246 -74.088 48.105
281560 99.271 -74.126 48.057
281561 99.286 -74.151 48.011
281562 99.284 -74.142 47.966
281563 99.283 -74.127 47.921
281564 99.290 -74.142 47.878
281565 99.351 -74.150 47.834
281566 99.412 -74.158 47.790
281567 99.473 -74.167 47.748
281568 99.534 -74.176 47.705
281569 99.594 -74.184 47.665
281570 99.646 -74.192 47.625
281571 99.668 -74.199 47.584
281572 99.692 -74.207 47.546
281573 99.714 -74.214 47.508
281574 99.737 -74.220 47.470
281575 99.760 -74.228 47.432
281576 99.783 -74.234 47.399
281577 99.807 -74.239 47.365
281578 99.832 -74.242 47.331
281579 99.856 -74.247 47.298
281580 99.880 -74.251 47.264
281581 99.905 -74.255 47.230
281582 99.934 -74.261 47.197
281583 99.965 -74.268 47.166
281584 99.995 -74.276 47.135
281585 100.025 -74.283 47.105
281586 100.056 -74.289 47.074
281587 100.087 -74.296 47.044
281588 100.118 -74.303 47.014
281589 100.147 -74.312 46.985
281590 100.168 -74.333 46.956
281591 100.190 -74.355 46.926
281592 100.212 -74.375 46.896
281593 100.235 -74.393 46.865
281594 100.266 -74.397 46.836
281595 100.297 -74.401 46.807
281596 100.328 -74.405 46.780
281597 100.359 -74.408 46.749
281598 100.390 -74.412 46.717
281599 100.421 -74.416 46.684
281600 100.452 -74.420 46.650
281601 100.482 -74.424 46.617
281602 100.513 -74.427 46.586
281603 100.544 -74.431 46.555
281604 100.575 -74.440 46.526
281605 100.606 -74.452 46.496
281606 100.636 -74.465 46.465
281607 100.667 -74.477 46.435
281608 100.697 -74.489 46.405
281609 100.728 -74.501 46.375
281610 100.759 -74.513 46.346
281611 100.789 -74.525 46.318
281612 100.823 -74.542 46.290
281613 100.880 -74.590 46.262
281614 100.936 -74.637 46.234
281615 100.971 -74.662 46.205
281616 100.998 -74.679 46.174
281617 101.025 -74.696 46.140
281618 101.053 -74.712 46.106
281619 101.080 -74.729 46.070
281620 101.108 -74.746 46.034
281621 101.135 -74.762 45.998
281622 101.162 -74.778 45.963
281623 101.190 -74.795 45.928
281624 101.217 -74.812 45.895
281625 101.244 -74.828 45.861
281626 101.271 -74.845 45.826
281627 101.304 -74.865 45.792
281628 101.332 -74.877 45.755
281629 101.367 -74.895 45.721
281630 101.406 -74.918 45.688
281631 101.445 -74.940 45.656
281632 101.483 -74.963 45.624
281633 101.521 -74.985 45.590
281634 101.554 -75.003 45.558
281635 101.588 -75.021 45.527
281636 101.622 -75.040 45.496
281637 101.656 -75.058 45.467
281638 101.689 -75.077 45.437
281639 101.721 -75.084 45.407
281640 101.750 -75.079 45.378
281641 101.779 -75.074 45.348
281642 101.808 -75.069 45.319
281643 101.837 -75.064 45.290
281644 101.867 -75.059 45.261
281645 101.896 -75.054 45.232
281646 101.900 -75.029 45.202
281647 101.917 -75.029 45.173
281648 101.935 -75.029 45.145
281649 101.953 -75.030 45.116
281650 101.970 -75.031 45.087
281651 101.987 -75.031 45.059
281652 102.005 -75.032 45.030
281653 102.022 -75.032 45.002
281654 102.040 -75.032 44.973
281655 102.057 -75.033 44.944
281656 102.075 -75.034 44.916
282022 92.447 -72.853 54.734
282023 92.478 -72.863 54.717
282024 92.509 -72.874 54.700
282025 92.540 -72.884 54.683
282026 92.570 -72.895 54.666
282027 92.601 -72.904 54.649
282028 92.641 -72.916 54.632
282029 92.684 -72.926 54.617
282030 92.727 -72.938 54.602
282031 92.770 -72.948 54.587
282032 92.813 -72.960 54.571
282033 92.856 -72.970 54.557
282034 92.899 -72.982 54.545
282035 92.941 -72.992 54.534
282036 92.985 -73.004 54.523
282037 93.027 -73.014 54.513
282038 93.070 -73.026 54.503
282039 93.113 -73.036 54.492
282040 93.156 -73.048 54.482
282041 93.194 -73.058 54.471
282042 93.220 -73.065 54.462
282043 93.245 -73.070 54.454
282044 93.269 -73.076 54.445
282045 93.293 -73.083 54.432
282046 93.318 -73.089 54.420
282047 93.342 -73.094 54.408
282048 93.376 -73.093 54.397
282049 93.422 -73.083 54.386
282050 93.448 -73.077 54.374
282051 93.474 -73.072 54.360
282052 93.500 -73.066 54.344
282053 93.525 -73.061 54.328
282054 93.551 -73.054 54.312
282055 93.577 -73.049 54.295
282122 95.049 -73.149 52.701
282123 95.069 -73.154 52.676
282124 95.090 -73.158 52.651
282125 95.110 -73.162 52.627
282126 95.131 -73.167 52.603
282127 95.151 -73.171 52.578
282128 95.172 -73.176 52.554
282129 95.199 -73.175 52.529
282130 95.220 -73.183 52.504
282131 95.235 -73.196 52.479
282132 95.249 -73.209 52.454
282133 95.263 -73.223 52.429
282134 95.298 -73.198 52.404
282135 95.343 -73.166 52.379
282136 95.393 -73.168 52.354
282137 95.436 -73.167 52.329
282138 95.473 -73.166 52.305
282139 95.510 -73.164 52.284
282140 95.546 -73.161 52.263
282141 95.574 -73.159 52.243
282142 95.603 -73.158 52.223
282143 95.630 -73.155 52.204
282144 95.659 -73.154 52.185
282145 95.695 -73.164 52.167
282146 95.734 -73.175 52.151
282147 95.771 -73.186 52.134
282148 95.810 -73.199 52.116
282149 95.847 -73.210 52.097
282150 95.886 -73.221 52.077
282151 95.929 -73.214 52.058
282152 95.972 -73.208 52.036
282153 96.015 -73.202 52.015
282154 96.058 -73.195 51.993
282155 96.101 -73.189 51.970
282156 96.136 -73.175 51.944
282157 96.168 -73.158 51.916
282158 96.200 -73.139 51.885
282159 96.227 -73.130 51.853
282160 96.250 -73.121 51.821
282161 96.252 -73.130 51.789
282162 96.249 -73.145 51.758
282163 96.246 -73.161 51.728
282164 96.272 -73.178 51.698
282165 96.316 -73.195 51.669
282166 96.342 -73.226 51.642
282167 96.362 -73.261 51.614
282168 96.384 -73.290 51.586
282169 96.410 -73.313 51.559
282170 96.437 -73.336 51.531
282171 96.464 -73.358 51.504
282172 96.490 -73.381 51.477
282173 96.517 -73.404 51.451
282174 96.543 -73.412 51.425
282175 96.559 -73.403 51.397
282176 96.560 -73.372 51.371
282177 96.567 -73.341 51.345
282178 96.607 -73.324 51.317
282179 96.633 -73.346 51.288
282180 96.665 -73.415 51.262
282181 96.711 -73.442 51.234
282182 96.750 -73.449 51.206
282183 96.783 -73.437 51.180
282184 96.800 -73.412 51.153
282185 96.774 -73.353 51.127
282186 96.785 -73.333 51.104
282187 96.816 -73.336 51.082
282188 96.847 -73.338 51.062
282189 96.879 -73.341 51.044
282190 96.928 -73.351 51.026
282191 96.976 -73.362 51.007
282192 97.018 -73.373 50.990
282193 97.056 -73.385 50.974
282194 97.095 -73.397 50.960
282195 97.124 -73.405 50.947
282196 97.152 -73.412 50.936
282197 97.175 -73.416 50.926
282198 97.188 -73.429 50.917
282199 97.206 -73.459 50.908
282200 97.224 -73.490 50.899
282201 97.247 -73.515 50.892
282202 97.242 -73.528 50.887
282203 97.273 -73.526 50.884
282204 97.312 -73.523 50.876
282205 97.350 -73.519 50.863
282206 97.389 -73.516 50.848
282207 97.427 -73.512 50.829
282208 97.466 -73.509 50.804
282209 97.504 -73.506 50.774
282210 97.543 -73.503 50.742
282211 97.581 -73.499 50.708
282212 97.620 -73.490 50.671
282213 97.659 -73.478 50.634
282214 97.698 -73.465 50.596
282215 97.736 -73.452 50.558
282216 97.775 -73.440 50.517
282217 97.814 -73.427 50.470
282218 97.855 -73.441 50.419
282219 97.898 -73.455 50.367
282220 97.944 -73.466 50.313
282221 97.974 -73.498 50.257
282222 97.999 -73.537 50.200
282223 98.035 -73.562 50.143
282224 98.074 -73.586 50.085
282225 98.112 -73.609 50.025
282226 98.150 -73.633 49.966
282227 98.188 -73.640 49.907
282228 98.225 -73.635 49.847
282229 98.263 -73.630 49.787
282230 98.301 -73.622 49.725
282231 98.340 -73.608 49.663
282232 98.362 -73.622 49.601
282233 98.370 -73.658 49.537
282234 98.378 -73.694 49.474
282235 98.386 -73.730 49.413
282236 98.409 -73.758 49.353
282237 98.435 -73.783 49.292
282238 98.472 -73.799 49.233
282239 98.513 -73.812 49.174
282240 98.558 -73.823 49.115
282241 98.621 -73.832 49.056
282242 98.640 -73.846 48.997
282243 98.647 -73.862 48.939
282244 98.653 -73.878 48.883
282245 98.754 -73.929 48.827
282246 98.806 -73.932 48.773
282247 98.835 -73.917 48.721
282248 98.864 -73.903 48.671
282249 98.874 -73.915 48.621
282250 98.894 -73.932 48.572
282251 98.932 -73.941 48.524
282252 98.970 -73.951 48.477
282253 99.007 -73.962 48.430
282254 99.045 -73.972 48.384
282255 99.082 -73.982 48.339
282256 99.120 -73.992 48.291
282257 99.157 -74.002 48.243
282258 99.196 -74.013 48.194
282259 99.224 -74.026 48.145
282260 99.245 -74.056 48.099
282261 99.272 -74.085 48.051
282262 99.349 -74.068 48.005
282263 99.323 -74.064 47.960
282264 99.332 -74.085 47.917
282265 99.343 -74.115 47.872
282266 99.402 -74.124 47.829
282267 99.463 -74.132 47.785
282268 99.523 -74.142 47.742
282269 99.584 -74.150 47.701
282270 99.633 -74.158 47.660
282271 99.656 -74.165 47.620
282272 99.679 -74.172 47.579
282273 99.702 -74.179 47.541
282274 99.724 -74.186 47.503
282275 99.747 -74.193 47.466
282276 99.770 -74.201 47.428
282277 99.793 -74.208 47.394
282278 99.817 -74.212 47.360
282279 99.841 -74.216 47.327
282280 99.865 -74.220 47.293
282281 99.890 -74.225 47.260
282282 99.914 -74.229 47.226
282283 99.938 -74.233 47.193
282284 99.964 -74.238 47.162
282285 99.994 -74.245 47.131
282286 100.025 -74.252 47.100
282287 100.056 -74.259 47.069
282288 100.086 -74.266 47.038
282289 100.117 -74.273 47.007
282290 100.148 -74.280 46.977
282291 100.171 -74.298 46.948
282292 100.193 -74.319 46.918
282293 100.214 -74.340 46.888
282294 100.241 -74.352 46.858
282295 100.271 -74.356 46.828
282296 100.302 -74.360 46.798
282297 100.333 -74.364 46.771
282298 100.364 -74.368 46.740
282299 100.395 -74.371 46.707
282300 100.426 -74.375 46.674
282301 100.457 -74.379 46.640
282302 100.488 -74.383 46.606
282303 100.519 -74.386 46.575
282304 100.550 -74.390 46.544
282305 100.580 -74.394 46.515
282306 100.611 -74.405 46.484
282307 100.642 -74.417 46.454
282308 100.672 -74.429 46.424
282309 100.703 -74.441 46.393
282310 100.734 -74.453 46.363
282311 100.764 -74.465 46.334
282312 100.795 -74.477 46.305
282313 100.841 -74.512 46.277
282314 100.898 -74.559 46.249
282315 100.955 -74.607 46.220
282316 100.994 -74.637 46.190
282317 101.022 -74.654 46.159
282318 101.049 -74.671 46.125
282319 101.076 -74.687 46.090
282320 101.104 -74.703 46.055
282321 101.131 -74.720 46.019
282322 101.158 -74.737 45.984
282323 101.186 -74.753 45.950
282324 101.213 -74.770 45.915
282325 101.240 -74.787 45.880
282326 101.267 -74.803 45.845
282327 101.295 -74.820 45.810
282328 101.322 -74.837 45.776
282329 101.351 -74.853 45.740
282330 101.389 -74.873 45.706
282331 101.427 -74.896 45.674
282332 101.466 -74.919 45.642
282333 101.504 -74.941 45.609
282334 101.543 -74.963 45.577
282335 101.577 -74.982 45.545
282336 101.611 -75.001 45.514
282337 101.644 -75.019 45.483
282338 101.678 -75.038 45.454
282339 101.710 -75.050 45.424
282340 101.740 -75.044 45.394
282341 101.769 -75.040 45.364
282342 101.798 -75.035 45.334
282343 101.827 -75.030 45.305
282344 101.856 -75.025 45.276
282345 101.886 -75.020 45.247
282346 101.911 -75.011 45.218
282347 101.915 -74.988 45.189
282348 101.933 -74.989 45.160
282349 101.950 -74.990 45.131
282350 101.968 -74.990 45.103
282351 101.985 -74.991 45.074
282352 102.003 -74.991 45.045
282353 102.020 -74.992 45.017
282354 102.037 -74.992 44.989
282355 102.055 -74.993 44.960
282356 102.072 -74.994 44.931
282357 102.090 -74.994 44.903
282723 92.443 -72.824 54.738
282724 92.474 -72.834 54.721
282725 92.505 -72.844 54.703
282726 92.535 -72.855 54.686
282727 92.566 -72.865 54.669
282728 92.596 -72.876 54.652
282729 92.638 -72.886 54.635
282730 92.681 -72.898 54.619
282731 92.724 -72.908 54.604
282732 92.767 -72.920 54.589
282733 92.809 -72.930 54.573
282734 92.853 -72.942 54.559
282735 92.896 -72.952 54.546
282736 92.938 -72.963 54.535
282737 92.981 -72.974 54.523
282738 93.024 -72.985 54.513
282739 93.067 -72.996 54.501
282740 93.110 -73.007 54.490
282741 93.153 -73.018 54.478
282742 93.186 -73.028 54.468
282743 93.212 -73.035 54.458
282744 93.236 -73.040 54.449
282745 93.261 -73.046 54.438
282746 93.285 -73.052 54.425
282747 93.310 -73.058 54.413
282748 93.334 -73.064 54.401
282749 93.374 -73.059 54.389
282750 93.418 -73.049 54.377
282751 93.443 -73.044 54.365
282752 93.469 -73.038 54.351
282753 93.495 -73.032 54.335
282754 93.520 -73.026 54.318
282755 93.547 -73.021 54.301
282756 93.572 -73.015 54.284
282823 95.052 -73.114 52.693
282824 95.072 -73.119 52.669
282825 95.093 -73.123 52.644
282826 95.113 -73.127 52.620
282827 95.134 -73.132 52.597
282828 95.155 -73.136 52.573
282829 95.175 -73.141 52.548
282830 95.195 -73.145 52.524
282831 95.220 -73.146 52.498
282832 95.242 -73.153 52.474
282833 95.256 -73.167 52.449
282834 95.271 -73.180 52.424
282835 95.294 -73.177 52.399
282836 95.338 -73.138 52.374
282837 95.388 -73.139 52.349
282838 95.430 -73.139 52.324
282839 95.466 -73.137 52.301
282840 95.503 -73.135 52.280
282841 95.540 -73.133 52.260
282842 95.572 -73.130 52.240
282843 95.600 -73.129 52.220
282844 95.628 -73.126 52.202
282845 95.658 -73.126 52.183
282846 95.695 -73.138 52.166
282847 95.734 -73.150 52.151
282848 95.771 -73.161 52.134
282849 95.810 -73.173 52.117
282850 95.847 -73.184 52.099
282851 95.886 -73.196 52.080
282852 95.924 -73.204 52.061
282853 95.968 -73.198 52.041
282854 96.011 -73.192 52.020
282855 96.053 -73.186 52.000
282856 96.096 -73.179 51.977
282857 96.128 -73.161 51.950
282858 96.160 -73.143 51.921
282859 96.192 -73.126 51.891
282860 96.224 -73.105 51.860
282861 96.247 -73.085 51.828
282862 96.243 -73.112 51.796
282863 96.240 -73.128 51.766
282864 96.237 -73.144 51.738
282865 96.273 -73.161 51.710
282866 96.317 -73.179 51.682
282867 96.361 -73.196 51.655
282868 96.395 -73.221 51.629
282869 96.414 -73.255 51.602
282870 96.434 -73.290 51.575
282871 96.453 -73.324 51.549
282872 96.473 -73.358 51.523
282873 96.500 -73.381 51.496
282874 96.526 -73.401 51.470
282875 96.535 -73.382 51.443
282876 96.536 -73.352 51.417
282877 96.537 -73.321 51.390
282878 96.573 -73.287 51.363
282879 96.605 -73.292 51.334
282880 96.631 -73.316 51.306
282881 96.661 -73.391 51.278
282882 96.703 -73.417 51.250
282883 96.736 -73.405 51.221
282884 96.769 -73.393 51.194
282885 96.802 -73.380 51.167
282886 96.778 -73.323 51.142
282887 96.798 -73.314 51.118
282888 96.829 -73.316 51.095
282889 96.860 -73.318 51.074
282890 96.891 -73.321 51.055
282891 96.923 -73.323 51.037
282892 96.970 -73.333 51.019
282893 97.014 -73.344 51.001
282894 97.052 -73.356 50.985
282895 97.082 -73.365 50.970
282896 97.109 -73.371 50.958
282897 97.137 -73.378 50.946
282898 97.163 -73.385 50.934
282899 97.180 -73.410 50.923
282900 97.197 -73.441 50.914
282901 97.215 -73.471 50.903
282902 97.236 -73.499 50.895
282903 97.243 -73.510 50.889
282904 97.282 -73.507 50.885
282905 97.320 -73.503 50.874
282906 97.359 -73.500 50.861
282907 97.397 -73.496 50.845
282908 97.436 -73.493 50.827
282909 97.474 -73.481 50.803
282910 97.513 -73.468 50.774
282911 97.552 -73.456 50.742
282912 97.591 -73.443 50.708
282913 97.630 -73.431 50.668
282914 97.668 -73.418 50.628
282915 97.707 -73.406 50.587
282916 97.746 -73.393 50.545
282917 97.785 -73.380 50.501
282918 97.825 -73.392 50.453
282919 97.867 -73.408 50.401
282920 97.911 -73.421 50.348
282921 97.956 -73.433 50.293
282922 98.002 -73.443 50.238
282923 98.034 -73.473 50.182
282924 98.059 -73.512 50.125
282925 98.083 -73.551 50.066
282926 98.116 -73.581 50.008
282927 98.154 -73.604 49.949
282928 98.193 -73.628 49.890
282929 98.230 -73.628 49.830
282930 98.267 -73.622 49.770
282931 98.306 -73.611 49.709
282932 98.345 -73.598 49.647
282933 98.378 -73.594 49.586
282934 98.385 -73.631 49.523
282935 98.393 -73.666 49.461
282936 98.403 -73.703 49.400
282937 98.429 -73.728 49.339
282938 98.456 -73.753 49.279
282939 98.489 -73.773 49.220
282940 98.530 -73.785 49.161
282941 98.571 -73.797 49.104
282942 98.618 -73.809 49.047
282943 98.649 -73.821 48.989
282944 98.656 -73.837 48.931
282945 98.704 -73.877 48.874
282946 98.779 -73.927 48.819
282947 98.807 -73.912 48.766
282948 98.836 -73.897 48.714
282949 98.865 -73.882 48.664
282950 98.880 -73.886 48.616
282951 98.909 -73.900 48.567
282952 98.946 -73.910 48.519
282953 98.984 -73.920 48.473
282954 99.022 -73.931 48.427
282955 99.060 -73.941 48.381
282956 99.104 -73.955 48.334
282957 99.139 -73.964 48.286
282958 99.169 -73.971 48.237
282959 99.200 -73.979 48.188
282960 99.224 -74.001 48.140
282961 99.245 -74.031 48.093
282962 99.273 -74.055 48.046
282963 99.312 -74.061 48.000
282964 99.359 -74.034 47.955
282965 99.385 -74.059 47.911
282966 99.397 -74.088 47.867
282967 99.453 -74.098 47.823
282968 99.514 -74.107 47.780
282969 99.575 -74.116 47.738
282970 99.620 -74.123 47.696
282971 99.643 -74.131 47.655
282972 99.666 -74.138 47.615
282973 99.689 -74.145 47.575
282974 99.711 -74.152 47.537
282975 99.734 -74.159 47.499
282976 99.757 -74.167 47.461
282977 99.780 -74.173 47.424
282978 99.803 -74.180 47.389
282979 99.827 -74.186 47.356
282980 99.851 -74.190 47.322
282981 99.875 -74.194 47.288
282982 99.899 -74.198 47.256
282983 99.923 -74.202 47.223
282984 99.948 -74.207 47.189
282985 99.972 -74.211 47.157
282986 99.996 -74.215 47.126
282987 100.025 -74.221 47.094
282988 100.055 -74.228 47.063
282989 100.086 -74.235 47.031
282990 100.117 -74.242 46.999
282991 100.147 -74.249 46.969
282992 100.174 -74.263 46.940
282993 100.195 -74.283 46.910
282994 100.217 -74.305 46.880
282995 100.246 -74.312 46.850
282996 100.277 -74.316 46.819
282997 100.308 -74.320 46.789
282998 100.339 -74.324 46.761
282999 100.369 -74.327 46.730
283000 100.400 -74.331 46.697
283001 100.431 -74.335 46.663
283002 100.462 -74.339 46.629
283003 100.493 -74.342 46.595
283004 100.524 -74.346 46.564
283005 100.555 -74.350 46.534
283006 100.585 -74.354 46.504
283007 100.617 -74.358 46.472
283008 100.647 -74.369 46.441
283009 100.678 -74.381 46.411
283010 100.708 -74.393 46.380
283011 100.739 -74.405 46.350
283012 100.770 -74.418 46.321
283013 100.802 -74.433 46.292
283014 100.859 -74.480 46.263
283015 100.916 -74.528 46.234
283016 100.973 -74.577 46.206
283017 101.015 -74.609 46.176
283018 101.041 -74.625 46.145
283019 101.068 -74.641 46.110
283020 101.095 -74.657 46.074
283021 101.122 -74.674 46.038
283022 101.149 -74.690 46.002
283023 101.176 -74.706 45.968
283024 101.203 -74.722 45.934
283025 101.230 -74.738 45.899
283026 101.257 -74.754 45.864
283027 101.284 -74.771 45.828
283028 101.311 -74.787 45.794
283029 101.338 -74.803 45.759
283030 101.364 -74.820 45.725
283031 101.410 -74.852 45.692
283032 101.449 -74.874 45.660
283033 101.487 -74.897 45.628
283034 101.526 -74.919 45.595
283035 101.565 -74.941 45.564
283036 101.599 -74.960 45.532
283037 101.633 -74.979 45.501
283038 101.667 -74.997 45.471
283039 101.700 -75.016 45.441
283040 101.730 -75.010 45.411
283041 101.759 -75.006 45.381
283042 101.788 -75.001 45.350
283043 101.817 -74.996 45.320
283044 101.846 -74.991 45.292
283045 101.876 -74.986 45.263
283046 101.905 -74.982 45.234
283047 101.919 -74.960 45.204
283048 101.930 -74.949 45.175
283049 101.948 -74.950 45.147
283050 101.965 -74.950 45.118
283051 101.982 -74.950 45.090
283052 102.000 -74.951 45.061
283053 102.018 -74.952 45.032
283054 102.035 -74.952 45.004
283055 102.052 -74.953 44.976
283056 102.070 -74.954 44.946
283057 102.087 -74.954 44.917
283425 92.470 -72.805 54.725
283426 92.500 -72.815 54.707
283427 92.531 -72.826 54.689
283428 92.562 -72.836 54.672
283429 92.592 -72.846 54.655
283430 92.635 -72.857 54.638
283431 92.678 -72.868 54.621
283432 92.721 -72.879 54.606
283433 92.764 -72.890 54.591
283434 92.806 -72.901 54.575
283435 92.849 -72.912 54.561
283436 92.892 -72.923 54.548
283437 92.935 -72.934 54.535
283438 92.978 -72.945 54.523
283439 93.021 -72.956 54.513
283440 93.064 -72.967 54.500
283441 93.107 -72.978 54.488
283442 93.149 -72.989 54.476
283443 93.179 -72.998 54.465
283444 93.204 -73.004 54.455
283445 93.228 -73.010 54.445
283446 93.253 -73.016 54.434
283447 93.277 -73.022 54.421
283448 93.302 -73.028 54.408
283449 93.326 -73.033 54.395
283450 93.371 -73.024 54.381
283451 93.413 -73.015 54.369
283452 93.438 -73.010 54.356
283453 93.464 -73.004 54.342
283454 93.490 -72.998 54.326
283524 95.055 -73.079 52.686
283525 95.075 -73.084 52.661
283526 95.096 -73.089 52.638
283527 95.116 -73.092 52.614
283528 95.137 -73.097 52.591
283529 95.158 -73.101 52.567
283530 95.178 -73.106 52.543
283531 95.198 -73.111 52.518
283532 95.219 -73.114 52.492
283533 95.242 -73.117 52.468
283534 95.264 -73.124 52.444
283535 95.278 -73.137 52.419
283536 95.293 -73.151 52.394
283537 95.334 -73.113 52.369
283538 95.383 -73.111 52.344
283539 95.423 -73.111 52.320
283540 95.460 -73.108 52.298
283541 95.497 -73.106 52.277
283542 95.533 -73.104 52.257
283543 95.569 -73.101 52.237
283544 95.598 -73.100 52.218
283545 95.626 -73.098 52.199
283546 95.658 -73.101 52.182
283547 95.695 -73.113 52.166
283548 95.734 -73.124 52.151
283549 95.772 -73.136 52.135
283550 95.810 -73.148 52.118
283551 95.848 -73.159 52.101
283552 95.886 -73.170 52.083
283553 95.924 -73.182 52.064
283554 95.963 -73.188 52.045
283555 96.007 -73.181 52.026
283556 96.049 -73.175 52.005
283557 96.088 -73.164 51.982
283558 96.120 -73.147 51.955
283559 96.153 -73.129 51.926
283560 96.182 -73.103 51.897
283561 96.209 -73.072 51.865
283562 96.238 -73.042 51.834
283563 96.234 -73.090 51.804
283564 96.230 -73.111 51.776
283565 96.231 -73.126 51.748
283566 96.274 -73.144 51.721
283567 96.318 -73.162 51.695
283568 96.362 -73.180 51.669
283569 96.406 -73.197 51.643
283570 96.448 -73.216 51.618
283571 96.467 -73.250 51.592
283572 96.487 -73.285 51.567
283573 96.501 -73.326 51.541
283574 96.515 -73.367 51.515
283575 96.513 -73.358 51.488
283576 96.512 -73.330 51.462
283577 96.513 -73.300 51.436
283578 96.533 -73.267 51.409
283579 96.576 -73.239 51.381
283580 96.603 -73.260 51.352
283581 96.630 -73.285 51.323
283582 96.659 -73.360 51.293
283583 96.689 -73.373 51.265
283584 96.722 -73.361 51.237
283585 96.755 -73.349 51.209
283586 96.788 -73.336 51.183
283587 96.781 -73.293 51.157
283588 96.811 -73.294 51.132
283589 96.842 -73.296 51.108
283590 96.873 -73.299 51.086
283591 96.904 -73.301 51.067
283592 96.936 -73.303 51.048
283593 96.967 -73.305 51.030
283594 97.010 -73.315 51.012
283595 97.040 -73.324 50.995
283596 97.068 -73.331 50.981
283597 97.094 -73.336 50.969
283598 97.120 -73.341 50.956
283599 97.154 -73.357 50.943
283600 97.173 -73.388 50.930
283601 97.190 -73.421 50.920
283602 97.208 -73.452 50.907
283603 97.225 -73.484 50.897
283604 97.252 -73.491 50.888
283605 97.290 -73.485 50.880
283606 97.329 -73.472 50.868
283607 97.368 -73.459 50.855
283608 97.406 -73.447 50.841
283609 97.446 -73.434 50.823
283610 97.485 -73.422 50.799
283611 97.523 -73.409 50.770
283612 97.562 -73.396 50.738
283613 97.601 -73.384 50.702
283614 97.640 -73.371 50.660
283615 97.678 -73.359 50.617
283616 97.717 -73.346 50.573
283617 97.756 -73.334 50.529
283618 97.796 -73.342 50.483
283619 97.838 -73.358 50.434
283620 97.879 -73.374 50.382
283621 97.923 -73.387 50.329
283622 97.968 -73.399 50.275
283623 98.014 -73.410 50.220
283624 98.060 -73.421 50.164
283625 98.093 -73.448 50.108
283626 98.118 -73.487 50.049
283627 98.143 -73.526 49.991
283628 98.167 -73.565 49.932
283629 98.196 -73.600 49.872
283630 98.235 -73.620 49.813
283631 98.272 -73.614 49.754
283632 98.311 -73.601 49.693
283633 98.350 -73.587 49.631
283634 98.388 -73.575 49.570
283635 98.402 -73.603 49.508
283636 98.410 -73.640 49.447
283637 98.423 -73.672 49.387
283638 98.450 -73.698 49.327
283639 98.477 -73.723 49.267
283640 98.506 -73.747 49.208
283641 98.547 -73.759 49.150
283642 98.588 -73.771 49.093
283643 98.629 -73.783 49.036
283644 98.673 -73.799 48.980
283645 98.724 -73.816 48.923
283646 98.776 -73.833 48.866
283647 98.806 -73.854 48.812
283648 98.818 -73.872 48.760
283649 98.838 -73.876 48.709
283650 98.866 -73.862 48.660
283651 98.887 -73.858 48.611
283652 98.924 -73.869 48.563
283653 98.961 -73.879 48.515
283654 99.006 -73.894 48.470
283655 99.051 -73.907 48.425
283656 99.081 -73.915 48.378
283657 99.112 -73.922 48.330
283658 99.143 -73.929 48.281
283659 99.174 -73.937 48.232
283660 99.204 -73.946 48.183
283661 99.225 -73.976 48.135
283662 99.249 -74.003 48.087
283663 99.289 -74.017 48.041
283664 99.330 -74.031 47.995
283665 99.367 -74.015 47.951
283666 99.439 -74.032 47.906
283667 99.450 -74.061 47.862
283668 99.504 -74.073 47.819
283669 99.565 -74.082 47.776
283670 99.608 -74.089 47.734
283671 99.630 -74.096 47.692
283672 99.653 -74.104 47.651
283673 99.676 -74.111 47.611
283674 99.699 -74.118 47.571
283675 99.721 -74.125 47.533
283676 99.744 -74.132 47.495
283677 99.767 -74.139 47.457
283678 99.790 -74.146 47.420
283679 99.812 -74.153 47.385
283680 99.836 -74.160 47.351
283681 99.860 -74.164 47.317
283682 99.885 -74.168 47.284
283683 99.909 -74.172 47.252
283684 99.933 -74.176 47.218
283685 99.957 -74.180 47.184
283686 99.982 -74.185 47.151
283687 100.006 -74.189 47.119
283688 100.030 -74.193 47.087
283689 100.055 -74.198 47.056
283690 100.085 -74.205 47.024
283691 100.116 -74.211 46.992
283692 100.146 -74.218 46.962
283693 100.176 -74.227 46.931
283694 100.198 -74.248 46.901
283695 100.220 -74.267 46.870
283696 100.251 -74.271 46.840
283697 100.282 -74.275 46.809
283698 100.313 -74.279 46.779
283699 100.344 -74.283 46.750
283700 100.374 -74.286 46.719
283701 100.406 -74.290 46.686
283702 100.436 -74.294 46.652
283703 100.467 -74.298 46.619
283704 100.498 -74.302 46.585
283705 100.529 -74.305 46.553
283706 100.560 -74.309 46.523
283707 100.591 -74.313 46.492
283708 100.622 -74.317 46.461
283709 100.652 -74.321 46.431
283710 100.683 -74.333 46.400
283711 100.714 -74.346 46.367
283712 100.745 -74.358 46.338
283713 100.775 -74.370 46.308
283714 100.821 -74.402 46.279
283715 100.878 -74.450 46.249
283716 100.934 -74.498 46.221
283717 100.973 -74.527 46.192
283718 101.011 -74.555 46.162
283719 101.047 -74.581 46.130
283720 101.074 -74.597 46.095
283721 101.101 -74.614 46.059
283722 101.128 -74.630 46.023
283723 101.155 -74.646 45.987
283724 101.182 -74.662 45.952
283725 101.209 -74.678 45.917
283726 101.236 -74.694 45.883
283727 101.263 -74.711 45.848
283728 101.289 -74.727 45.813
283729 101.317 -74.743 45.778
283730 101.343 -74.759 45.744
283731 101.371 -74.781 45.710
283732 101.397 -74.816 45.678
283733 101.470 -74.853 45.647
283734 101.509 -74.875 45.615
283735 101.548 -74.897 45.582
283736 101.587 -74.920 45.551
283737 101.622 -74.940 45.519
283738 101.656 -74.958 45.488
283739 101.689 -74.977 45.459
283740 101.719 -74.976 45.428
283741 101.748 -74.972 45.398
283742 101.777 -74.966 45.368
283743 101.807 -74.962 45.337
283744 101.836 -74.957 45.307
283745 101.865 -74.952 45.278
283746 101.894 -74.947 45.250
283747 101.923 -74.942 45.221
283748 101.927 -74.909 45.191
283749 101.945 -74.910 45.162
283750 101.963 -74.910 45.133
283751 101.980 -74.910 45.105
283752 101.997 -74.911 45.076
283753 102.015 -74.911 45.048
283754 102.032 -74.912 45.019
283755 102.050 -74.913 44.991
283756 102.067 -74.913 44.963
283757 102.085 -74.913 44.933
283758 102.102 -74.914 44.904
284127 92.496 -72.786 54.711
284128 92.527 -72.797 54.693
284129 92.558 -72.807 54.676
284130 92.589 -72.817 54.659
284131 92.632 -72.828 54.642
284132 92.674 -72.839 54.625
284133 92.717 -72.850 54.610
284134 92.761 -72.861 54.595
284135 92.803 -72.872 54.578
284136 92.846 -72.883 54.563
284137 92.889 -72.894 54.549
284138 92.932 -72.905 54.536
284139 92.975 -72.916 54.523
284140 93.018 -72.927 54.513
284141 93.061 -72.938 54.500
284142 93.103 -72.949 54.487
284143 93.143 -72.960 54.475
284144 93.171 -72.968 54.463
284145 93.196 -72.974 54.452
284146 93.220 -72.979 54.441
284147 93.245 -72.985 54.430
284148 93.269 -72.992 54.416
284149 93.294 -72.998 54.403
284150 93.322 -73.001 54.389
284151 93.369 -72.990 54.375
284225 95.058 -73.045 52.679
284226 95.078 -73.049 52.655
284227 95.099 -73.054 52.631
284228 95.119 -73.058 52.608
284229 95.140 -73.062 52.586
284230 95.161 -73.067 52.563
284231 95.181 -73.071 52.539
284232 95.201 -73.076 52.514
284233 95.222 -73.079 52.488
284234 95.243 -73.084 52.465
284235 95.263 -73.089 52.440
284236 95.286 -73.095 52.415
284237 95.300 -73.108 52.390
284238 95.330 -73.092 52.365
284239 95.378 -73.083 52.340
284240 95.417 -73.082 52.317
284241 95.454 -73.079 52.296
284242 95.490 -73.077 52.275
284243 95.527 -73.075 52.255
284244 95.564 -73.073 52.235
284245 95.595 -73.071 52.216
284246 95.623 -73.069 52.198
284247 95.658 -73.076 52.182
284248 95.696 -73.087 52.167
284249 95.734 -73.099 52.151
284250 95.772 -73.111 52.136
284251 95.810 -73.122 52.120
284252 95.848 -73.133 52.104
284253 95.886 -73.145 52.087
284254 95.924 -73.157 52.068
284255 95.962 -73.168 52.049
284256 96.003 -73.171 52.030
284257 96.045 -73.165 52.009
284258 96.080 -73.150 51.986
284259 96.112 -73.131 51.958
284260 96.140 -73.101 51.930
284261 96.168 -73.070 51.902
284262 96.196 -73.039 51.870
284263 96.224 -73.009 51.840
284264 96.227 -73.041 51.813
284265 96.220 -73.094 51.786
284266 96.232 -73.110 51.760
284267 96.275 -73.127 51.733
284268 96.319 -73.145 51.707
284269 96.366 -73.163 51.682
284270 96.415 -73.181 51.658
284271 96.465 -73.200 51.633
284272 96.477 -73.240 51.609
284273 96.490 -73.281 51.585
284274 96.504 -73.323 51.559
284275 96.503 -73.313 51.533
284276 96.498 -73.293 51.507
284277 96.493 -73.272 51.480
284278 96.493 -73.248 51.454
284279 96.539 -73.214 51.426
284280 96.574 -73.208 51.398
284281 96.601 -73.228 51.368
284282 96.628 -73.255 51.338
284283 96.630 -73.254 51.308
284284 96.632 -73.252 51.279
284285 96.677 -73.268 51.252
284286 96.731 -73.289 51.225
284287 96.774 -73.293 51.198
284288 96.793 -73.271 51.172
284289 96.824 -73.274 51.146
284290 96.856 -73.277 51.121
284291 96.887 -73.279 51.099
284292 96.918 -73.280 51.079
284293 96.949 -73.281 51.060
284294 96.979 -73.282 51.041
284295 97.002 -73.282 51.022
284296 97.022 -73.283 51.006
284297 97.048 -73.289 50.992
284298 97.075 -73.293 50.979
284299 97.107 -73.308 50.966
284300 97.146 -73.331 50.952
284301 97.168 -73.362 50.937
284302 97.185 -73.393 50.925
284303 97.202 -73.426 50.912
284304 97.238 -73.449 50.898
284305 97.261 -73.438 50.885
284306 97.300 -73.425 50.874
284307 97.339 -73.413 50.861
284308 97.378 -73.400 50.848
284309 97.417 -73.387 50.834
284310 97.455 -73.375 50.816
284311 97.494 -73.362 50.792
284312 97.533 -73.350 50.763
284313 97.572 -73.337 50.730
284314 97.610 -73.324 50.692
284315 97.649 -73.312 50.648
284316 97.688 -73.299 50.602
284317 97.727 -73.287 50.556
284318 97.767 -73.292 50.510
284319 97.808 -73.308 50.463
284320 97.850 -73.324 50.414
284321 97.890 -73.341 50.362
284322 97.935 -73.354 50.310
284323 97.980 -73.365 50.258
284324 98.026 -73.376 50.203
284325 98.072 -73.387 50.148
284326 98.117 -73.399 50.093
284327 98.153 -73.423 50.034
284328 98.177 -73.462 49.975
284329 98.202 -73.501 49.915
284330 98.227 -73.540 49.854
284331 98.251 -73.580 49.796
284332 98.277 -73.603 49.738
284333 98.316 -73.590 49.677
284334 98.355 -73.578 49.614
284335 98.394 -73.565 49.553
284336 98.417 -73.576 49.493
284337 98.425 -73.612 49.433
284338 98.443 -73.643 49.375
284339 98.470 -73.668 49.316
284340 98.497 -73.694 49.257
284341 98.524 -73.719 49.198
284342 98.563 -73.732 49.140
284343 98.604 -73.744 49.083
284344 98.649 -73.751 49.026
284345 98.728 -73.709 48.970
284346 98.780 -73.727 48.915
284347 98.829 -73.744 48.860
284348 98.839 -73.767 48.807
284349 98.849 -73.790 48.756
284350 98.861 -73.810 48.705
284351 98.881 -73.813 48.656
284352 98.911 -73.830 48.608
284353 98.954 -73.847 48.560
284354 98.993 -73.858 48.513
284355 99.024 -73.866 48.469
284356 99.055 -73.872 48.422
284357 99.086 -73.880 48.374
284358 99.117 -73.888 48.325
284359 99.148 -73.894 48.277
284360 99.179 -73.902 48.227
284361 99.205 -73.921 48.178
284362 99.225 -73.951 48.129
284363 99.265 -73.965 48.082
284364 99.306 -73.979 48.036
284365 99.347 -73.992 47.991
284366 99.387 -74.007 47.946
284367 99.454 -74.011 47.902
284368 99.504 -74.035 47.857
284369 99.555 -74.048 47.814
284370 99.594 -74.055 47.772
284371 99.617 -74.062 47.731
284372 99.640 -74.070 47.688
284373 99.663 -74.076 47.647
284374 99.686 -74.083 47.607
284375 99.709 -74.091 47.568
284376 99.731 -74.098 47.530
284377 99.754 -74.104 47.492
284378 99.777 -74.112 47.454
284379 99.800 -74.119 47.417
284380 99.823 -74.126 47.382
284381 99.845 -74.133 47.348
284382 99.870 -74.138 47.314
284383 99.894 -74.142 47.280
284384 99.918 -74.146 47.247
284385 99.942 -74.150 47.213
284386 99.967 -74.154 47.179
284387 99.991 -74.158 47.145
284388 100.015 -74.163 47.113
284389 100.040 -74.167 47.080
284390 100.064 -74.171 47.048
284391 100.088 -74.175 47.016
284392 100.115 -74.181 46.984
284393 100.146 -74.188 46.953
284394 100.177 -74.195 46.923
284395 100.200 -74.213 46.893
284396 100.225 -74.226 46.862
284397 100.256 -74.230 46.831
284398 100.287 -74.234 46.800
284399 100.318 -74.238 46.769
284400 100.349 -74.242 46.740
284401 100.380 -74.245 46.709
284402 100.411 -74.250 46.676
284403 100.442 -74.254 46.642
284404 100.473 -74.258 46.608
284405 100.503 -74.261 46.575
284406 100.535 -74.265 46.543
284407 100.565 -74.269 46.512
284408 100.596 -74.273 46.480
284409 100.627 -74.277 46.450
284410 100.658 -74.280 46.419
284411 100.689 -74.286 46.388
284412 100.719 -74.298 46.355
284413 100.750 -74.310 46.326
284414 100.782 -74.324 46.297
284415 100.839 -74.372 46.267
284416 100.892 -74.416 46.237
284417 100.930 -74.443 46.208
284418 100.968 -74.471 46.178
284419 101.005 -74.499 46.147
284420 101.043 -74.527 46.115
284421 101.080 -74.554 46.080
284422 101.107 -74.570 46.044
284423 101.134 -74.586 46.008
284424 101.161 -74.603 45.972
284425 101.188 -74.618 45.936
284426 101.214 -74.634 45.901
284427 101.242 -74.651 45.866
284428 101.268 -74.667 45.831
284429 101.296 -74.683 45.797
284430 101.322 -74.700 45.763
284431 101.349 -74.715 45.729
284432 101.376 -74.741 45.695
284433 101.403 -74.775 45.665
284434 101.429 -74.810 45.634
284435 101.516 -74.849 45.602
284436 101.569 -74.875 45.570
284437 101.608 -74.898 45.538
284438 101.644 -74.919 45.506
284439 101.678 -74.937 45.475
284440 101.709 -74.942 45.445
284441 101.738 -74.938 45.414
284442 101.767 -74.932 45.384
284443 101.796 -74.928 45.354
284444 101.826 -74.922 45.324
284445 101.855 -74.918 45.294
284446 101.884 -74.913 45.265
284447 101.913 -74.908 45.236
284448 101.943 -74.903 45.208
284449 101.943 -74.869 45.178
284450 101.960 -74.869 45.149
284451 101.977 -74.870 45.120
284452 101.995 -74.871 45.091
284453 102.013 -74.871 45.063
284454 102.030 -74.872 45.034
284455 102.047 -74.872 45.006
284456 102.065 -74.872 44.977
284457 102.082 -74.873 44.949
284458 102.100 -74.874 44.919
284830 92.554 -72.778 54.681
284831 92.585 -72.788 54.663
284832 92.629 -72.799 54.646
284833 92.671 -72.810 54.629
284834 92.714 -72.821 54.614
284835 92.757 -72.832 54.598
284836 92.800 -72.843 54.581
284837 92.843 -72.854 54.566
284838 92.886 -72.865 54.552
284839 92.929 -72.876 54.539
284840 92.972 -72.887 54.526
284841 93.014 -72.898 54.514
284842 93.057 -72.909 54.500
284843 93.100 -72.920 54.486
284844 93.136 -72.929 54.474
284845 93.163 -72.938 54.461
284846 93.188 -72.944 54.450
284847 93.212 -72.949 54.438
284848 93.236 -72.955 54.426
284849 93.261 -72.961 54.412
284925 95.040 -73.005 52.695
284926 95.061 -73.010 52.673
284927 95.082 -73.014 52.649
284928 95.102 -73.019 52.626
284929 95.122 -73.023 52.603
284930 95.143 -73.027 52.581
284931 95.164 -73.032 52.558
284932 95.184 -73.036 52.534
284933 95.204 -73.041 52.510
284934 95.225 -73.045 52.485
284935 95.246 -73.049 52.461
284936 95.266 -73.054 52.437
284937 95.287 -73.058 52.412
284938 95.308 -73.065 52.387
284939 95.326 -73.072 52.362
284940 95.373 -73.054 52.337
284941 95.411 -73.053 52.314
284942 95.448 -73.051 52.293
284943 95.484 -73.048 52.272
284944 95.521 -73.046 52.253
284945 95.557 -73.045 52.234
284946 95.593 -73.042 52.215
284947 95.621 -73.040 52.199
284948 95.658 -73.051 52.182
284949 95.696 -73.062 52.167
284950 95.734 -73.073 52.152
284951 95.772 -73.085 52.136
284952 95.810 -73.097 52.121
284953 95.848 -73.108 52.105
284954 95.886 -73.120 52.089
284955 95.924 -73.131 52.071
284956 95.962 -73.142 52.053
284957 96.000 -73.155 52.034
284958 96.040 -73.154 52.014
284959 96.070 -73.129 51.989
284960 96.098 -73.098 51.962
284961 96.126 -73.068 51.934
284962 96.154 -73.038 51.905
284963 96.182 -73.007 51.877
284964 96.210 -72.976 51.849
284965 96.220 -72.992 51.823
284966 96.214 -73.045 51.797
284967 96.245 -73.088 51.772
284968 96.293 -73.113 51.747
284969 96.342 -73.132 51.723
284970 96.391 -73.150 51.699
284971 96.441 -73.168 51.674
284972 96.475 -73.194 51.650
284973 96.480 -73.237 51.626
284974 96.493 -73.279 51.602
284975 96.491 -73.268 51.577
284976 96.486 -73.248 51.551
284977 96.482 -73.230 51.525
284978 96.479 -73.220 51.497
284979 96.508 -73.193 51.470
284980 96.549 -73.163 51.441
284981 96.584 -73.160 51.411
284982 96.615 -73.176 51.382
284983 96.645 -73.192 51.352
284984 96.649 -73.188 51.323
284985 96.636 -73.170 51.295
284986 96.653 -73.171 51.267
284987 96.716 -73.199 51.240
284988 96.780 -73.227 51.213
284989 96.799 -73.245 51.186
284990 96.835 -73.246 51.159
284991 96.867 -73.246 51.135
284992 96.897 -73.247 51.113
284993 96.928 -73.248 51.092
284994 96.959 -73.249 51.072
284995 96.986 -73.249 51.051
284996 97.001 -73.248 51.032
284997 97.016 -73.247 51.016
284998 97.031 -73.247 51.001
284999 97.062 -73.260 50.987
285000 97.099 -73.283 50.975
285001 97.138 -73.305 50.961
285002 97.162 -73.334 50.946
285003 97.179 -73.367 50.933
285004 97.217 -73.390 50.919
285005 97.261 -73.390 50.903
285006 97.275 -73.378 50.887
285007 97.310 -73.366 50.872
285008 97.349 -73.353 50.858
285009 97.387 -73.341 50.843
285010 97.426 -73.328 50.826
285011 97.465 -73.315 50.807
285012 97.504 -73.303 50.780
285013 97.543 -73.290 50.750
285014 97.582 -73.278 50.715
285015 97.620 -73.265 50.676
285016 97.659 -73.252 50.630
285017 97.698 -73.240 50.581
285018 97.738 -73.242 50.532
285019 97.779 -73.258 50.483
285020 97.820 -73.275 50.437
285021 97.861 -73.291 50.388
285022 97.902 -73.308 50.337
285023 97.947 -73.320 50.286
285024 97.993 -73.331 50.236
285025 98.038 -73.343 50.183
285026 98.084 -73.354 50.130
285027 98.129 -73.365 50.076
285028 98.175 -73.376 50.017
285029 98.213 -73.398 49.957
285030 98.237 -73.437 49.897
285031 98.262 -73.477 49.836
285032 98.286 -73.515 49.778
285033 98.307 -73.545 49.719
285034 98.326 -73.571 49.659
285035 98.360 -73.568 49.598
285036 98.399 -73.554 49.538
285037 98.433 -73.549 49.480
285038 98.441 -73.585 49.422
285039 98.464 -73.612 49.364
285040 98.491 -73.638 49.306
285041 98.517 -73.663 49.249
285042 98.544 -73.689 49.190
285043 98.580 -73.706 49.132
285044 98.621 -73.719 49.075
285045 98.703 -73.666 49.020
285046 98.784 -73.621 48.965
285047 98.835 -73.638 48.911
285048 98.863 -73.659 48.856
285049 98.873 -73.681 48.804
285050 98.883 -73.704 48.752
285051 98.893 -73.727 48.702
285052 98.935 -73.752 48.653
285053 98.998 -73.776 48.606
285054 98.968 -73.811 48.559
285055 98.998 -73.823 48.512
285056 99.029 -73.831 48.466
285057 99.060 -73.838 48.418
285058 99.091 -73.845 48.369
285059 99.122 -73.853 48.321
285060 99.153 -73.860 48.271
285061 99.184 -73.867 48.222
285062 99.205 -73.896 48.174
285063 99.241 -73.913 48.125
285064 99.282 -73.927 48.078
285065 99.323 -73.941 48.032
285066 99.363 -73.954 47.987
285067 99.404 -73.968 47.942
285068 99.471 -73.988 47.898
285069 99.544 -74.010 47.854
285070 99.582 -74.021 47.811
285071 99.605 -74.028 47.769
285072 99.627 -74.035 47.727
285073 99.650 -74.042 47.685
285074 99.673 -74.049 47.645
285075 99.696 -74.056 47.605
285076 99.718 -74.064 47.566
285077 99.742 -74.070 47.527
285078 99.764 -74.078 47.489
285079 99.787 -74.085 47.451
285080 99.810 -74.092 47.414
285081 99.833 -74.099 47.379
285082 99.855 -74.106 47.344
285083 99.879 -74.111 47.310
285084 99.904 -74.115 47.275
285085 99.928 -74.120 47.242
285086 99.952 -74.123 47.207
285087 99.976 -74.128 47.173
285088 100.000 -74.132 47.139
285089 100.025 -74.136 47.106
285090 100.049 -74.141 47.073
285091 100.073 -74.145 47.041
285092 100.098 -74.149 47.008
285093 100.122 -74.153 46.975
285094 100.146 -74.157 46.945
285095 100.176 -74.164 46.914
285096 100.203 -74.177 46.884
285097 100.231 -74.186 46.853
285098 100.262 -74.190 46.822
285099 100.292 -74.194 46.790
285100 100.324 -74.198 46.759
285101 100.354 -74.201 46.730
285102 100.385 -74.205 46.699
285103 100.416 -74.209 46.667
285104 100.447 -74.213 46.632
285105 100.478 -74.217 46.598
285106 100.509 -74.220 46.564
285107 100.540 -74.224 46.532
285108 100.570 -74.228 46.501
285109 100.602 -74.232 46.469
285110 100.632 -74.236 46.439
285111 100.663 -74.239 46.408
285112 100.694 -74.243 46.376
285113 100.725 -74.250 46.345
285114 100.756 -74.262 46.316
285115 100.800 -74.293 46.288
285116 100.848 -74.332 46.258
285117 100.886 -74.360 46.228
285118 100.924 -74.387 46.196
285119 100.962 -74.415 46.164
285120 100.999 -74.443 46.131
285121 101.037 -74.471 46.098
285122 101.075 -74.499 46.063
285123 101.113 -74.526 46.028
285124 101.140 -74.543 45.992
285125 101.167 -74.559 45.957
285126 101.193 -74.574 45.920
285127 101.220 -74.591 45.886
285128 101.247 -74.607 45.851
285129 101.274 -74.623 45.816
285130 101.301 -74.640 45.783
285131 101.328 -74.656 45.750
285132 101.355 -74.671 45.716
285133 101.382 -74.701 45.683
285134 101.409 -74.736 45.652
285135 101.435 -74.771 45.621
285136 101.462 -74.806 45.589
285137 101.560 -74.845 45.557
285138 101.629 -74.876 45.526
285139 101.667 -74.897 45.494
285140 101.699 -74.908 45.462
285141 101.728 -74.903 45.432
285142 101.757 -74.898 45.402
285143 101.786 -74.894 45.372
285144 101.816 -74.888 45.342
285145 101.845 -74.884 45.312
285146 101.874 -74.879 45.282
285147 101.903 -74.874 45.252
285148 101.933 -74.869 45.224
285149 101.959 -74.861 45.195
285150 101.958 -74.829 45.166
285151 101.975 -74.830 45.136
285152 101.992 -74.831 45.107
285153 102.010 -74.831 45.077
285154 102.027 -74.831 45.050
285155 102.045 -74.832 45.021
285156 102.062 -74.832 44.992
285157 102.080 -74.833 44.964
285158 102.097 -74.834 44.935
285159 102.115 -74.834 44.906
285626 95.043 -72.970 52.689
285627 95.064 -72.975 52.666
285628 95.085 -72.979 52.642
285629 95.105 -72.984 52.620
285630 95.125 -72.988 52.599
285631 95.146 -72.992 52.576
285632 95.167 -72.997 52.553
285633 95.187 -73.001 52.530
285634 95.207 -73.006 52.506
285635 95.228 -73.010 52.482
285636 95.249 -73.014 52.458
285637 95.269 -73.019 52.434
285638 95.290 -73.023 52.409
285639 95.310 -73.028 52.384
285640 95.330 -73.036 52.359
285641 95.368 -73.026 52.334
285642 95.404 -73.024 52.311
285643 95.441 -73.022 52.290
285644 95.478 -73.020 52.270
285645 95.514 -73.018 52.251
285646 95.551 -73.016 52.232
285647 95.588 -73.014 52.214
285648 95.620 -73.014 52.198
285649 95.658 -73.025 52.182
285650 95.696 -73.036 52.166
285651 95.734 -73.048 52.151
285652 95.772 -73.060 52.136
285653 95.810 -73.071 52.121
285654 95.848 -73.083 52.105
285655 95.887 -73.093 52.090
285656 95.926 -73.104 52.072
285657 95.965 -73.114 52.055
285658 96.004 -73.123 52.037
285659 96.031 -73.136 52.017
285660 96.057 -73.107 51.991
285661 96.085 -73.070 51.965
285662 96.113 -73.036 51.938
285663 96.141 -73.005 51.910
285664 96.169 -72.975 51.883
285665 96.197 -72.944 51.858
285666 96.213 -72.942 51.833
285667 96.238 -72.986 51.808
285668 96.283 -73.025 51.784
285669 96.327 -73.064 51.760
285670 96.371 -73.102 51.737
285671 96.416 -73.137 51.714
285672 96.466 -73.155 51.690
285673 96.477 -73.192 51.666
285674 96.482 -73.235 51.643
285675 96.480 -73.224 51.619
285676 96.477 -73.210 51.593
285677 96.474 -73.199 51.567
285678 96.472 -73.189 51.540
285679 96.482 -73.171 51.512
285680 96.524 -73.137 51.484
285681 96.554 -73.139 51.454
285682 96.582 -73.145 51.424
285683 96.625 -73.133 51.395
285684 96.664 -73.131 51.365
285685 96.694 -73.148 51.336
285686 96.681 -73.131 51.308
285687 96.723 -73.146 51.280
285688 96.771 -73.172 51.253
285689 96.782 -73.193 51.226
285690 96.796 -73.212 51.199
285691 96.833 -73.213 51.173
285692 96.871 -73.214 51.149
285693 96.908 -73.214 51.126
285694 96.939 -73.214 51.105
285695 96.969 -73.215 51.084
285696 96.985 -73.215 51.062
285697 97.001 -73.215 51.043
285698 97.022 -73.218 51.026
285699 97.042 -73.221 51.010
285700 97.066 -73.234 50.996
285701 97.092 -73.258 50.983
285702 97.130 -73.280 50.969
285703 97.157 -73.308 50.955
285704 97.196 -73.329 50.940
285705 97.261 -73.338 50.924
285706 97.283 -73.330 50.906
285707 97.297 -73.318 50.887
285708 97.320 -73.306 50.870
285709 97.359 -73.294 50.854
285710 97.397 -73.281 50.837
285711 97.436 -73.269 50.817
285712 97.475 -73.256 50.795
285713 97.514 -73.243 50.766
285714 97.552 -73.231 50.734
285715 97.591 -73.218 50.697
285716 97.630 -73.206 50.655
285717 97.669 -73.193 50.608
285718 97.709 -73.192 50.558
285719 97.750 -73.209 50.507
285720 97.791 -73.225 50.457
285721 97.832 -73.242 50.410
285722 97.873 -73.258 50.360
285723 97.914 -73.274 50.310
285724 97.959 -73.286 50.261
285725 98.005 -73.298 50.212
285726 98.050 -73.309 50.162
285727 98.096 -73.320 50.110
285728 98.141 -73.331 50.056
285729 98.187 -73.343 49.998
285730 98.233 -73.354 49.939
285731 98.272 -73.373 49.878
285732 98.297 -73.412 49.817
285733 98.321 -73.452 49.759
285734 98.341 -73.478 49.700
285735 98.360 -73.503 49.641
285736 98.380 -73.529 49.581
285737 98.404 -73.544 49.522
285738 98.443 -73.531 49.465
285739 98.458 -73.557 49.409
285740 98.485 -73.583 49.352
285741 98.511 -73.608 49.296
285742 98.538 -73.634 49.239
285743 98.565 -73.659 49.181
285744 98.597 -73.680 49.123
285745 98.665 -73.649 49.067
285746 98.757 -73.580 49.013
285747 98.839 -73.532 48.959
285748 98.886 -73.550 48.905
285749 98.896 -73.572 48.852
285750 98.906 -73.595 48.800
285751 98.906 -73.625 48.749
285752 98.885 -73.670 48.699
285753 98.963 -73.701 48.650
285754 99.013 -73.714 48.603
285755 98.994 -73.744 48.556
285756 99.007 -73.770 48.510
285757 99.034 -73.793 48.462
285758 99.065 -73.803 48.413
285759 99.095 -73.810 48.365
285760 99.127 -73.818 48.316
285761 99.157 -73.825 48.266
285762 99.185 -73.841 48.218
285763 99.217 -73.861 48.169
285764 99.258 -73.875 48.121
285765 99.299 -73.888 48.074
285766 99.351 -73.905 48.028
285767 99.419 -73.926 47.983
285768 99.468 -73.944 47.939
285769 99.510 -73.960 47.895
285770 99.552 -73.976 47.852
285771 99.592 -73.985 47.809
285772 99.616 -73.999 47.766
285773 99.637 -74.007 47.724
285774 99.660 -74.015 47.683
285775 99.683 -74.022 47.643
285776 99.706 -74.029 47.603
285777 99.728 -74.036 47.564
285778 99.751 -74.043 47.524
285779 99.774 -74.051 47.486
285780 99.797 -74.057 47.448
285781 99.820 -74.065 47.411
285782 99.843 -74.072 47.376
285783 99.865 -74.079 47.341
285784 99.889 -74.085 47.306
285785 99.913 -74.089 47.270
285786 99.937 -74.093 47.236
285787 99.962 -74.098 47.201
285788 99.986 -74.101 47.167
285789 100.010 -74.106 47.133
285790 100.034 -74.110 47.099
285791 100.059 -74.114 47.066
285792 100.083 -74.118 47.034
285793 100.107 -74.123 47.001
285794 100.131 -74.127 46.968
285795 100.155 -74.131 46.936
285796 100.180 -74.136 46.905
285797 100.205 -74.142 46.875
285798 100.236 -74.145 46.844
285799 100.267 -74.149 46.812
285800 100.298 -74.153 46.780
285801 100.329 -74.157 46.748
285802 100.359 -74.161 46.719
285803 100.391 -74.164 46.689
285804 100.421 -74.168 46.657
285805 100.452 -74.172 46.623
285806 100.483 -74.176 46.587
285807 100.514 -74.179 46.553
285808 100.545 -74.183 46.521
285809 100.576 -74.187 46.491
285810 100.607 -74.191 46.459
285811 100.637 -74.195 46.428
285812 100.668 -74.198 46.397
285813 100.699 -74.202 46.366
285814 100.731 -74.206 46.336
285815 100.762 -74.215 46.307
285816 100.804 -74.248 46.278
285817 100.842 -74.276 46.248
285818 100.880 -74.304 46.217
285819 100.918 -74.331 46.183
285820 100.956 -74.359 46.149
285821 100.993 -74.387 46.115
285822 101.032 -74.415 46.081
285823 101.069 -74.443 46.047
285824 101.107 -74.470 46.012
285825 101.145 -74.498 45.977
285826 101.173 -74.515 45.942
285827 101.199 -74.531 45.906
285828 101.227 -74.547 45.871
285829 101.253 -74.563 45.837
285830 101.280 -74.580 45.803
285831 101.307 -74.596 45.770
285832 101.334 -74.612 45.737
285833 101.361 -74.628 45.704
285834 101.388 -74.662 45.671
285835 101.414 -74.697 45.640
285836 101.441 -74.731 45.608
285837 101.468 -74.766 45.577
285838 101.504 -74.801 45.546
285839 101.603 -74.841 45.514
285840 101.688 -74.874 45.482
285841 101.717 -74.869 45.451
285842 101.747 -74.864 45.420
285843 101.776 -74.859 45.391
285844 101.806 -74.854 45.360
285845 101.835 -74.850 45.330
285846 101.864 -74.845 45.301
285847 101.893 -74.840 45.270
285848 101.922 -74.835 45.241
285849 101.952 -74.830 45.212
285850 101.967 -74.809 45.183
285851 101.973 -74.790 45.154
285852 101.990 -74.790 45.125
285853 102.007 -74.791 45.095
285854 102.025 -74.791 45.065
285855 102.042 -74.792 45.037
285856 102.060 -74.792 45.009
285857 102.078 -74.793 44.981
285858 102.095 -74.794 44.951
285859 102.112 -74.794 44.922
286327 95.049 -72.937 52.683
286328 95.067 -72.940 52.660
286329 95.088 -72.945 52.637
286330 95.108 -72.949 52.614
286331 95.128 -72.953 52.593
286332 95.149 -72.957 52.571
286333 95.170 -72.962 52.548
286334 95.190 -72.967 52.525
286335 95.211 -72.971 52.502
286336 95.231 -72.975 52.479
286337 95.252 -72.979 52.455
286338 95.272 -72.984 52.431
286339 95.293 -72.989 52.406
286340 95.313 -72.993 52.381
286341 95.334 -72.997 52.356
286342 95.361 -72.998 52.331
286343 95.398 -72.995 52.308
286344 95.435 -72.993 52.287
286345 95.471 -72.992 52.267
286346 95.508 -72.989 52.248
286347 95.546 -72.988 52.229
286348 95.586 -72.986 52.213
286349 95.622 -72.985 52.195
286350 95.661 -72.995 52.178
286351 95.700 -73.005 52.162
286352 95.740 -73.015 52.147
286353 95.778 -73.026 52.133
286354 95.818 -73.036 52.119
286355 95.856 -73.046 52.104
286356 95.896 -73.056 52.088
286357 95.935 -73.067 52.072
286358 95.974 -73.076 52.056
286359 96.007 -73.088 52.038
286360 96.025 -73.101 52.018
286361 96.046 -73.105 51.993
286362 96.074 -73.070 51.967
286363 96.101 -73.034 51.941
286364 96.129 -72.998 51.915
286365 96.157 -72.962 51.890
286366 96.184 -72.926 51.866
286367 96.213 -72.898 51.842
286368 96.255 -72.931 51.819
286369 96.298 -72.970 51.798
286370 96.341 -73.009 51.774
286371 96.385 -73.048 51.751
286372 96.428 -73.087 51.728
286373 96.471 -73.126 51.706
286374 96.477 -73.180 51.682
286375 96.472 -73.189 51.658
286376 96.469 -73.179 51.634
286377 96.467 -73.168 51.608
286378 96.464 -73.158 51.582
286379 96.461 -73.147 51.554
286380 96.499 -73.116 51.526
286381 96.531 -73.108 51.496
286382 96.559 -73.115 51.466
286383 96.587 -73.122 51.436
286384 96.623 -73.118 51.407
286385 96.667 -73.106 51.378
286386 96.710 -73.095 51.349
286387 96.730 -73.092 51.320
286388 96.759 -73.116 51.291
286389 96.769 -73.137 51.265
286390 96.780 -73.158 51.238
286391 96.792 -73.179 51.211
286392 96.829 -73.180 51.186
286393 96.868 -73.180 51.163
286394 96.905 -73.181 51.140
286395 96.943 -73.182 51.118
286396 96.975 -73.186 51.097
286397 96.995 -73.189 51.074
286398 97.016 -73.192 51.054
286399 97.036 -73.195 51.036
286400 97.057 -73.198 51.019
286401 97.079 -73.207 51.004
286402 97.105 -73.230 50.990
286403 97.131 -73.254 50.977
286404 97.174 -73.270 50.962
286405 97.240 -73.278 50.946
286406 97.291 -73.282 50.929
286407 97.305 -73.270 50.908
286408 97.319 -73.258 50.887
286409 97.333 -73.247 50.868
286410 97.368 -73.234 50.850
286411 97.407 -73.222 50.830
286412 97.446 -73.209 50.807
286413 97.485 -73.197 50.783
286414 97.524 -73.184 50.750
286415 97.563 -73.171 50.716
286416 97.601 -73.159 50.677
286417 97.640 -73.146 50.633
286418 97.679 -73.142 50.584
286419 97.721 -73.159 50.534
286420 97.761 -73.175 50.483
286421 97.803 -73.192 50.432
286422 97.844 -73.208 50.383
286423 97.885 -73.224 50.332
286424 97.926 -73.240 50.283
286425 97.971 -73.252 50.234
286426 98.017 -73.264 50.187
286427 98.062 -73.275 50.139
286428 98.108 -73.286 50.088
286429 98.153 -73.298 50.036
286430 98.199 -73.309 49.978
286431 98.245 -73.320 49.919
286432 98.290 -73.331 49.859
286433 98.332 -73.348 49.799
286434 98.356 -73.386 49.740
286435 98.375 -73.411 49.681
286436 98.394 -73.437 49.623
286437 98.414 -73.462 49.564
286438 98.433 -73.487 49.506
286439 98.452 -73.512 49.450
286440 98.478 -73.528 49.395
286441 98.505 -73.553 49.340
286442 98.532 -73.578 49.286
286443 98.559 -73.603 49.227
286444 98.587 -73.627 49.170
286445 98.627 -73.631 49.114
286446 98.719 -73.563 49.059
286447 98.811 -73.494 49.004
286448 98.895 -73.443 48.951
286449 98.920 -73.464 48.899
286450 98.921 -73.493 48.846
286451 98.901 -73.537 48.796
286452 98.880 -73.582 48.746
286453 98.860 -73.627 48.698
286454 98.950 -73.646 48.649
286455 99.000 -73.659 48.601
286456 99.049 -73.671 48.553
286457 99.018 -73.706 48.506
286458 99.046 -73.729 48.458
286459 99.073 -73.752 48.408
286460 99.100 -73.775 48.359
286461 99.131 -73.783 48.311
286462 99.162 -73.791 48.261
286463 99.194 -73.809 48.213
286464 99.234 -73.822 48.164
286465 99.298 -73.842 48.117
286466 99.339 -73.859 48.070
286467 99.381 -73.875 48.024
286468 99.423 -73.891 47.979
286469 99.465 -73.907 47.935
286470 99.506 -73.923 47.891
286471 99.548 -73.939 47.848
286472 99.588 -73.951 47.806
286473 99.628 -73.956 47.764
286474 99.660 -73.965 47.723
286475 99.673 -73.985 47.681
286476 99.693 -73.995 47.640
286477 99.716 -74.002 47.600
286478 99.738 -74.009 47.560
286479 99.761 -74.016 47.522
286480 99.784 -74.023 47.484
286481 99.807 -74.030 47.445
286482 99.830 -74.038 47.407
286483 99.852 -74.045 47.371
286484 99.875 -74.051 47.335
286485 99.898 -74.059 47.300
286486 99.922 -74.063 47.264
286487 99.947 -74.067 47.229
286488 99.971 -74.071 47.193
286489 99.995 -74.076 47.159
286490 100.019 -74.079 47.125
286491 100.044 -74.084 47.091
286492 100.068 -74.088 47.058
286493 100.092 -74.092 47.024
286494 100.117 -74.096 46.991
286495 100.141 -74.101 46.959
286496 100.165 -74.104 46.928
286497 100.194 -74.111 46.897
286498 100.226 -74.117 46.867
286499 100.257 -74.121 46.835
286500 100.287 -74.125 46.803
286501 100.318 -74.128 46.770
286502 100.349 -74.132 46.739
286503 100.379 -74.136 46.710
286504 100.410 -74.139 46.680
286505 100.441 -74.142 46.647
286506 100.472 -74.146 46.613
286507 100.502 -74.150 46.577
286508 100.533 -74.153 46.543
286509 100.563 -74.157 46.511
286510 100.594 -74.161 46.481
286511 100.625 -74.164 46.450
286512 100.655 -74.167 46.418
286513 100.686 -74.171 46.387
286514 100.717 -74.175 46.356
286515 100.747 -74.177 46.326
286516 100.775 -74.183 46.297
286517 100.805 -74.199 46.267
286518 100.836 -74.220 46.236
286519 100.874 -74.247 46.205
286520 100.912 -74.275 46.170
286521 100.950 -74.303 46.134
286522 100.988 -74.330 46.100
286523 101.025 -74.358 46.065
286524 101.063 -74.386 46.031
286525 101.101 -74.414 45.997
286526 101.139 -74.442 45.962
286527 101.177 -74.470 45.927
286528 101.205 -74.487 45.892
286529 101.232 -74.503 45.857
286530 101.259 -74.520 45.823
286531 101.286 -74.536 45.790
286532 101.313 -74.552 45.757
286533 101.340 -74.568 45.724
286534 101.367 -74.587 45.691
286535 101.393 -74.622 45.659
286536 101.420 -74.657 45.628
286537 101.447 -74.691 45.597
286538 101.474 -74.726 45.566
286539 101.575 -74.766 45.535
286540 101.653 -74.798 45.503
286541 101.699 -74.820 45.471
286542 101.737 -74.830 45.440
286543 101.766 -74.825 45.409
286544 101.795 -74.820 45.380
286545 101.824 -74.816 45.349
286546 101.853 -74.811 45.319
286547 101.883 -74.806 45.290
286548 101.912 -74.801 45.258
286549 101.941 -74.796 45.229
286550 101.970 -74.791 45.201
286551 101.975 -74.758 45.172
286552 101.987 -74.750 45.142
286553 102.005 -74.750 45.113
286554 102.023 -74.751 45.083
286555 102.040 -74.751 45.054
286556 102.057 -74.752 45.025
286557 102.075 -74.753 44.997
286558 102.092 -74.753 44.968
286559 102.110 -74.753 44.939
286560 102.127 -74.754 44.910
287028 95.060 -72.905 52.677
287029 95.070 -72.905 52.654
287030 95.091 -72.910 52.631
287031 95.111 -72.914 52.609
287032 95.132 -72.918 52.587
287033 95.152 -72.923 52.565
287034 95.173 -72.927 52.544
287035 95.193 -72.932 52.521
287036 95.214 -72.936 52.498
287037 95.234 -72.940 52.475
287038 95.255 -72.945 52.452
287039 95.275 -72.949 52.428
287040 95.296 -72.954 52.403
287041 95.316 -72.958 52.378
287042 95.337 -72.962 52.353
287043 95.365 -72.970 52.329
287044 95.410 -72.974 52.305
287045 95.449 -72.973 52.284
287046 95.488 -72.972 52.264
287047 95.528 -72.971 52.244
287048 95.567 -72.970 52.225
287049 95.606 -72.969 52.208
287050 95.633 -72.952 52.190
287051 95.670 -72.958 52.174
287052 95.709 -72.968 52.157
287053 95.749 -72.978 52.141
287054 95.788 -72.989 52.128
287055 95.827 -72.998 52.113
287056 95.866 -73.009 52.099
287057 95.905 -73.019 52.084
287058 95.944 -73.029 52.069
287059 95.982 -73.039 52.052
287060 96.001 -73.053 52.035
287061 96.020 -73.067 52.015
287062 96.039 -73.080 51.992
287063 96.062 -73.069 51.968
287064 96.090 -73.032 51.944
287065 96.117 -72.997 51.921
287066 96.145 -72.961 51.898
287067 96.173 -72.926 51.874
287068 96.201 -72.894 51.852
287069 96.235 -72.898 51.830
287070 96.276 -72.927 51.809
287071 96.319 -72.966 51.787
287072 96.363 -73.005 51.764
287073 96.406 -73.044 51.741
287074 96.449 -73.083 51.719
287075 96.467 -73.132 51.696
287076 96.461 -73.148 51.673
287077 96.459 -73.137 51.649
287078 96.456 -73.126 51.623
287079 96.453 -73.116 51.596
287080 96.473 -73.094 51.567
287081 96.509 -73.077 51.537
287082 96.537 -73.084 51.509
287083 96.565 -73.091 51.480
287084 96.592 -73.098 51.451
287085 96.621 -73.103 51.421
287086 96.665 -73.092 51.392
287087 96.708 -73.079 51.364
287088 96.747 -73.070 51.334
287089 96.757 -73.082 51.304
287090 96.768 -73.103 51.277
287091 96.779 -73.124 51.251
287092 96.789 -73.145 51.225
287093 96.826 -73.147 51.200
287094 96.864 -73.148 51.177
287095 96.901 -73.148 51.154
287096 96.943 -73.153 51.132
287097 96.982 -73.160 51.110
287098 97.010 -73.165 51.087
287099 97.030 -73.168 51.066
287100 97.051 -73.171 51.047
287101 97.072 -73.174 51.030
287102 97.093 -73.180 51.013
287103 97.118 -73.203 50.998
287104 97.145 -73.227 50.983
287105 97.195 -73.229 50.968
287106 97.262 -73.236 50.952
287107 97.306 -73.227 50.931
287108 97.327 -73.210 50.910
287109 97.341 -73.199 50.888
287110 97.355 -73.187 50.866
287111 97.378 -73.175 50.844
287112 97.417 -73.162 50.820
287113 97.456 -73.150 50.794
287114 97.495 -73.137 50.766
287115 97.533 -73.125 50.731
287116 97.572 -73.112 50.694
287117 97.611 -73.099 50.655
287118 97.650 -73.093 50.610
287119 97.691 -73.109 50.560
287120 97.732 -73.126 50.510
287121 97.774 -73.142 50.459
287122 97.815 -73.158 50.409
287123 97.856 -73.174 50.357
287124 97.897 -73.191 50.306
287125 97.938 -73.207 50.255
287126 97.983 -73.219 50.207
287127 98.029 -73.230 50.160
287128 98.074 -73.242 50.113
287129 98.120 -73.252 50.064
287130 98.165 -73.264 50.011
287131 98.211 -73.275 49.954
287132 98.257 -73.286 49.896
287133 98.302 -73.298 49.838
287134 98.348 -73.309 49.781
287135 98.387 -73.324 49.724
287136 98.414 -73.347 49.667
287137 98.435 -73.374 49.608
287138 98.456 -73.400 49.548
287139 98.477 -73.427 49.490
287140 98.498 -73.453 49.435
287141 98.518 -73.480 49.380
287142 98.545 -73.504 49.326
287143 98.574 -73.528 49.271
287144 98.590 -73.559 49.216
287145 98.596 -73.590 49.162
287146 98.687 -73.528 49.107
287147 98.777 -73.465 49.053
287148 98.867 -73.402 48.999
287149 98.936 -73.359 48.947
287150 98.916 -73.404 48.895
287151 98.896 -73.449 48.843
287152 98.875 -73.494 48.793
287153 98.855 -73.539 48.744
287154 98.877 -73.578 48.695
287155 98.936 -73.591 48.646
287156 98.986 -73.603 48.596
287157 99.035 -73.615 48.548
287158 99.046 -73.638 48.500
287159 99.057 -73.665 48.452
287160 99.084 -73.687 48.404
287161 99.112 -73.711 48.355
287162 99.139 -73.734 48.306
287163 99.169 -73.757 48.256
287164 99.210 -73.774 48.208
287165 99.252 -73.790 48.160
287166 99.294 -73.806 48.112
287167 99.336 -73.822 48.065
287168 99.377 -73.838 48.020
287169 99.419 -73.854 47.975
287170 99.461 -73.870 47.931
287171 99.502 -73.886 47.888
287172 99.544 -73.903 47.845
287173 99.585 -73.915 47.803
287174 99.625 -73.921 47.761
287175 99.664 -73.926 47.719
287176 99.703 -73.932 47.678
287177 99.717 -73.951 47.638
287178 99.729 -73.970 47.597
287179 99.749 -73.982 47.557
287180 99.771 -73.989 47.518
287181 99.794 -73.996 47.479
287182 99.817 -74.003 47.440
287183 99.840 -74.010 47.402
287184 99.862 -74.017 47.364
287185 99.886 -74.025 47.328
287186 99.908 -74.032 47.292
287187 99.932 -74.036 47.256
287188 99.956 -74.041 47.220
287189 99.981 -74.045 47.185
287190 100.005 -74.049 47.150
287191 100.029 -74.054 47.115
287192 100.053 -74.057 47.082
287193 100.077 -74.062 47.048
287194 100.102 -74.066 47.015
287195 100.126 -74.070 46.981
287196 100.153 -74.075 46.949
287197 100.185 -74.082 46.919
287198 100.217 -74.089 46.889
287199 100.249 -74.095 46.859
287200 100.280 -74.100 46.826
287201 100.311 -74.104 46.793
287202 100.342 -74.107 46.761
287203 100.372 -74.111 46.729
287204 100.403 -74.114 46.701
287205 100.434 -74.118 46.670
287206 100.465 -74.121 46.637
287207 100.495 -74.125 46.602
287208 100.526 -74.129 46.567
287209 100.557 -74.132 46.533
287210 100.587 -74.136 46.501
287211 100.618 -74.139 46.470
287212 100.649 -74.143 46.441
287213 100.680 -74.146 46.408
287214 100.709 -74.149 46.377
287215 100.738 -74.151 46.346
287216 100.766 -74.152 46.316
287217 100.794 -74.160 46.286
287218 100.824 -74.176 46.255
287219 100.853 -74.192 46.223
287220 100.882 -74.209 46.190
287221 100.911 -74.226 46.155
287222 100.944 -74.247 46.120
287223 100.982 -74.274 46.085
287224 101.019 -74.302 46.051
287225 101.057 -74.330 46.016
287226 101.095 -74.358 45.981
287227 101.133 -74.386 45.947
287228 101.171 -74.414 45.911
287229 101.209 -74.441 45.877
287230 101.238 -74.460 45.843
287231 101.265 -74.476 45.810
287232 101.292 -74.492 45.776
287233 101.319 -74.508 45.744
287234 101.346 -74.524 45.712
287235 101.373 -74.548 45.679
287236 101.399 -74.583 45.646
287237 101.426 -74.617 45.616
287238 101.454 -74.652 45.585
287239 101.556 -74.692 45.554
287240 101.616 -74.718 45.523
287241 101.661 -74.740 45.491
287242 101.706 -74.762 45.459
287243 101.751 -74.784 45.429
287244 101.785 -74.786 45.398
287245 101.814 -74.781 45.368
287246 101.843 -74.777 45.338
287247 101.873 -74.772 45.308
287248 101.902 -74.767 45.278
287249 101.931 -74.762 45.247
287250 101.960 -74.757 45.218
287251 101.989 -74.752 45.189
287252 101.985 -74.709 45.160
287253 102.003 -74.710 45.131
287254 102.020 -74.711 45.101
287255 102.037 -74.711 45.072
287256 102.055 -74.712 45.043
287257 102.073 -74.712 45.014
287258 102.090 -74.713 44.985
287259 102.107 -74.713 44.956
287260 102.125 -74.714 44.927
287261 102.142 -74.715 44.898
287729 95.070 -72.873 52.671
287730 95.077 -72.872 52.648
287731 95.094 -72.875 52.626
287732 95.114 -72.879 52.603
287733 95.135 -72.883 52.582
287734 95.155 -72.888 52.561
287735 95.176 -72.892 52.539
287736 95.196 -72.897 52.517
287737 95.217 -72.901 52.494
287738 95.237 -72.905 52.471
287739 95.258 -72.910 52.449
287740 95.278 -72.914 52.425
287741 95.299 -72.919 52.399
287742 95.320 -72.923 52.375
287743 95.340 -72.927 52.351
287744 95.369 -72.935 52.326
287745 95.426 -72.952 52.303
287746 95.469 -72.955 52.281
287747 95.509 -72.954 52.260
287748 95.548 -72.953 52.240
287749 95.587 -72.952 52.221
287750 95.624 -72.948 52.203
287751 95.648 -72.926 52.185
287752 95.679 -72.921 52.168
287753 95.719 -72.931 52.153
287754 95.758 -72.941 52.136
287755 95.797 -72.951 52.122
287756 95.836 -72.961 52.107
287757 95.875 -72.972 52.093
287758 95.914 -72.982 52.078
287759 95.953 -72.992 52.064
287760 95.977 -73.004 52.048
287761 95.996 -73.018 52.031
287762 96.015 -73.032 52.013
287763 96.033 -73.045 51.992
287764 96.052 -73.059 51.970
287765 96.078 -73.032 51.948
287766 96.106 -72.996 51.927
287767 96.133 -72.960 51.905
287768 96.161 -72.924 51.882
287769 96.189 -72.890 51.859
287770 96.223 -72.894 51.839
287771 96.257 -72.897 51.819
287772 96.297 -72.923 51.798
287773 96.340 -72.962 51.776
287774 96.384 -73.001 51.754
287775 96.428 -73.040 51.732
287776 96.457 -73.085 51.710
287777 96.451 -73.106 51.689
287778 96.448 -73.095 51.667
287779 96.446 -73.085 51.640
287780 96.448 -73.072 51.612
287781 96.486 -73.047 51.582
287782 96.514 -73.054 51.551
287783 96.542 -73.061 51.523
287784 96.570 -73.067 51.494
287785 96.597 -73.074 51.465
287786 96.625 -73.081 51.437
287787 96.663 -73.076 51.408
287788 96.706 -73.064 51.379
287789 96.745 -73.054 51.349
287790 96.756 -73.058 51.319
287791 96.766 -73.068 51.291
287792 96.777 -73.089 51.264
287793 96.788 -73.111 51.239
287794 96.822 -73.114 51.214
287795 96.860 -73.115 51.190
287796 96.901 -73.119 51.167
287797 96.945 -73.125 51.144
287798 96.986 -73.132 51.122
287799 97.016 -73.139 51.098
287800 97.045 -73.145 51.077
287801 97.066 -73.148 51.058
287802 97.086 -73.151 51.039
287803 97.106 -73.154 51.022
287804 97.133 -73.177 51.006
287805 97.162 -73.196 50.989
287806 97.221 -73.192 50.973
287807 97.282 -73.192 50.957
287808 97.314 -73.177 50.936
287809 97.341 -73.156 50.914
287810 97.363 -73.139 50.891
287811 97.377 -73.127 50.867
287812 97.391 -73.115 50.840
287813 97.427 -73.103 50.812
287814 97.465 -73.090 50.782
287815 97.505 -73.078 50.749
287816 97.543 -73.065 50.712
287817 97.582 -73.053 50.673
287818 97.621 -73.043 50.632
287819 97.662 -73.060 50.585
287820 97.703 -73.076 50.534
287821 97.744 -73.092 50.483
287822 97.786 -73.108 50.433
287823 97.826 -73.125 50.383
287824 97.868 -73.141 50.332
287825 97.909 -73.157 50.280
287826 97.950 -73.174 50.229
287827 97.995 -73.185 50.181
287828 98.041 -73.196 50.135
287829 98.086 -73.208 50.087
287830 98.132 -73.219 50.038
287831 98.177 -73.230 49.986
287832 98.223 -73.242 49.931
287833 98.269 -73.252 49.875
287834 98.314 -73.264 49.819
287835 98.358 -73.268 49.763
287836 98.393 -73.287 49.707
287837 98.421 -73.311 49.652
287838 98.448 -73.335 49.594
287839 98.475 -73.359 49.536
287840 98.503 -73.383 49.479
287841 98.530 -73.406 49.423
287842 98.557 -73.430 49.368
287843 98.584 -73.454 49.314
287844 98.570 -73.507 49.260
287845 98.565 -73.549 49.207
287846 98.652 -73.489 49.153
287847 98.726 -73.441 49.100
287848 98.799 -73.393 49.046
287849 98.873 -73.346 48.994
287850 98.902 -73.336 48.942
287851 98.889 -73.363 48.892
287852 98.870 -73.405 48.840
287853 98.850 -73.450 48.790
287854 98.829 -73.496 48.741
287855 98.873 -73.524 48.691
287856 98.923 -73.536 48.641
287857 98.972 -73.548 48.591
287858 99.022 -73.560 48.542
287859 99.071 -73.573 48.494
287860 99.068 -73.600 48.446
287861 99.095 -73.624 48.398
287862 99.123 -73.647 48.349
287863 99.150 -73.670 48.300
287864 99.176 -73.706 48.252
287865 99.207 -73.737 48.204
287866 99.249 -73.753 48.156
287867 99.290 -73.769 48.108
287868 99.332 -73.785 48.062
287869 99.373 -73.801 48.017
287870 99.415 -73.818 47.972
287871 99.457 -73.834 47.928
287872 99.499 -73.850 47.885
287873 99.541 -73.866 47.843
287874 99.582 -73.880 47.801
287875 99.621 -73.885 47.758
287876 99.660 -73.891 47.716
287877 99.700 -73.897 47.675
287878 99.739 -73.902 47.634
287879 99.762 -73.916 47.594
287880 99.774 -73.936 47.553
287881 99.786 -73.956 47.513
287882 99.804 -73.969 47.473
287883 99.827 -73.976 47.434
287884 99.850 -73.983 47.396
287885 99.872 -73.990 47.357
287886 99.895 -73.998 47.320
287887 99.918 -74.004 47.283
287888 99.941 -74.010 47.247
287889 99.966 -74.014 47.211
287890 99.990 -74.019 47.174
287891 100.014 -74.023 47.139
287892 100.039 -74.027 47.105
287893 100.063 -74.031 47.072
287894 100.087 -74.035 47.038
287895 100.113 -74.040 47.004
287896 100.145 -74.047 46.971
287897 100.177 -74.053 46.938
287898 100.209 -74.060 46.909
287899 100.241 -74.067 46.879
287900 100.273 -74.073 46.849
287901 100.304 -74.079 46.817
287902 100.335 -74.082 46.783
287903 100.366 -74.085 46.750
287904 100.397 -74.089 46.718
287905 100.427 -74.093 46.689
287906 100.458 -74.097 46.659
287907 100.488 -74.100 46.625
287908 100.519 -74.104 46.589
287909 100.550 -74.107 46.555
287910 100.580 -74.111 46.521
287911 100.611 -74.114 46.489
287912 100.642 -74.118 46.458
287913 100.671 -74.120 46.428
287914 100.700 -74.122 46.397
287915 100.728 -74.123 46.365
287916 100.757 -74.125 46.333
287917 100.785 -74.126 46.302
287918 100.814 -74.136 46.272
287919 100.843 -74.153 46.240
287920 100.873 -74.169 46.207
287921 100.901 -74.186 46.174
287922 100.930 -74.202 46.140
287923 100.960 -74.219 46.104
287924 100.989 -74.235 46.070
287925 101.018 -74.252 46.036
287926 101.051 -74.273 46.001
287927 101.090 -74.302 45.966
287928 101.127 -74.330 45.931
287929 101.165 -74.357 45.896
287930 101.203 -74.385 45.862
287931 101.241 -74.413 45.829
287932 101.271 -74.432 45.796
287933 101.298 -74.449 45.763
287934 101.325 -74.465 45.732
287935 101.351 -74.480 45.699
287936 101.378 -74.508 45.666
287937 101.405 -74.543 45.634
287938 101.435 -74.577 45.604
287939 101.533 -74.616 45.573
287940 101.578 -74.638 45.542
287941 101.623 -74.660 45.511
287942 101.669 -74.682 45.479
287943 101.714 -74.704 45.448
287944 101.759 -74.726 45.417
287945 101.804 -74.747 45.387
287946 101.833 -74.743 45.357
287947 101.863 -74.737 45.327
287948 101.892 -74.733 45.297
287949 101.921 -74.728 45.267
287950 101.950 -74.723 45.236
287951 101.979 -74.718 45.206
287952 102.007 -74.712 45.177
287953 102.000 -74.670 45.149
287954 102.018 -74.671 45.120
287955 102.035 -74.671 45.090
287956 102.052 -74.671 45.060
287957 102.070 -74.672 45.031
287958 102.087 -74.672 45.001
287959 102.105 -74.673 44.973
287960 102.122 -74.674 44.945
287961 102.139 -74.675 44.915
287962 102.162 -74.675 44.885
288430 95.071 -72.837 52.665
288431 95.089 -72.841 52.643
288432 95.097 -72.840 52.620
288433 95.117 -72.844 52.598
288434 95.138 -72.848 52.577
288435 95.158 -72.853 52.556
288436 95.179 -72.857 52.535
288437 95.199 -72.862 52.513
288438 95.220 -72.866 52.491
288439 95.240 -72.870 52.467
288440 95.261 -72.875 52.445
288441 95.281 -72.879 52.421
288442 95.302 -72.884 52.396
288443 95.323 -72.888 52.372
288444 95.343 -72.892 52.347
288445 95.374 -72.901 52.323
288446 95.430 -72.918 52.300
288447 95.488 -72.935 52.278
288448 95.529 -72.936 52.256
288449 95.568 -72.935 52.235
288450 95.607 -72.934 52.217
288451 95.639 -72.923 52.197
288452 95.663 -72.901 52.180
288453 95.688 -72.884 52.163
288454 95.728 -72.894 52.146
288455 95.766 -72.904 52.130
288456 95.806 -72.914 52.116
288457 95.845 -72.924 52.101
288458 95.884 -72.935 52.086
288459 95.923 -72.945 52.071
288460 95.953 -72.957 52.056
288461 95.972 -72.970 52.042
288462 95.990 -72.983 52.027
288463 96.009 -72.997 52.011
288464 96.028 -73.011 51.993
288465 96.047 -73.024 51.974
288466 96.067 -73.031 51.953
288467 96.095 -72.995 51.932
288468 96.122 -72.959 51.910
288469 96.150 -72.923 51.887
288470 96.177 -72.888 51.865
288471 96.212 -72.890 51.846
288472 96.246 -72.893 51.827
288473 96.279 -72.897 51.807
288474 96.319 -72.920 51.787
288475 96.362 -72.958 51.765
288476 96.405 -72.998 51.745
288477 96.447 -73.037 51.726
288478 96.441 -73.064 51.706
288479 96.438 -73.054 51.685
288480 96.435 -73.043 51.658
288481 96.464 -73.017 51.629
288482 96.491 -73.023 51.599
288483 96.519 -73.029 51.567
288484 96.547 -73.037 51.537
288485 96.575 -73.044 51.509
288486 96.602 -73.051 51.481
288487 96.630 -73.058 51.454
288488 96.661 -73.061 51.425
288489 96.704 -73.049 51.396
288490 96.743 -73.039 51.365
288491 96.754 -73.043 51.334
288492 96.765 -73.047 51.306
288493 96.776 -73.054 51.279
288494 96.787 -73.076 51.253
288495 96.818 -73.081 51.228
288496 96.858 -73.084 51.204
288497 96.902 -73.091 51.180
288498 96.946 -73.098 51.157
288499 96.990 -73.105 51.133
288500 97.020 -73.111 51.109
288501 97.050 -73.118 51.088
288502 97.080 -73.124 51.067
288503 97.100 -73.127 51.047
288504 97.121 -73.130 51.030
288505 97.146 -73.149 51.013
288506 97.187 -73.160 50.995
288507 97.245 -73.156 50.978
288508 97.291 -73.145 50.962
288509 97.322 -73.126 50.941
288510 97.349 -73.105 50.919
288511 97.375 -73.086 50.895
288512 97.399 -73.067 50.870
288513 97.413 -73.055 50.839
288514 97.437 -73.044 50.805
288515 97.475 -73.031 50.770
288516 97.514 -73.018 50.732
288517 97.553 -73.006 50.693
288518 97.592 -72.993 50.653
288519 97.633 -73.010 50.609
288520 97.674 -73.026 50.561
288521 97.715 -73.042 50.508
288522 97.756 -73.058 50.456
288523 97.797 -73.075 50.406
288524 97.839 -73.091 50.356
288525 97.880 -73.108 50.307
288526 97.921 -73.123 50.256
288527 97.962 -73.140 50.205
288528 98.007 -73.152 50.156
288529 98.053 -73.163 50.110
288530 98.098 -73.174 50.061
288531 98.144 -73.185 50.012
288532 98.190 -73.196 49.961
288533 98.235 -73.208 49.909
288534 98.281 -73.219 49.855
288535 98.324 -73.221 49.800
288536 98.367 -73.224 49.744
288537 98.400 -73.248 49.690
288538 98.427 -73.276 49.636
288539 98.454 -73.300 49.582
288540 98.481 -73.324 49.526
288541 98.508 -73.347 49.470
288542 98.536 -73.371 49.413
288543 98.563 -73.395 49.359
288544 98.585 -73.418 49.305
288545 98.571 -73.477 49.252
288546 98.573 -73.491 49.199
288547 98.646 -73.443 49.145
288548 98.720 -73.396 49.092
288549 98.793 -73.348 49.039
288550 98.867 -73.300 48.987
288551 98.866 -73.317 48.937
288552 98.854 -73.343 48.887
288553 98.841 -73.369 48.837
288554 98.824 -73.407 48.788
288555 98.803 -73.452 48.738
288556 98.859 -73.468 48.687
288557 98.909 -73.481 48.635
288558 98.958 -73.493 48.585
288559 99.008 -73.506 48.536
288560 99.058 -73.518 48.488
288561 99.098 -73.532 48.440
288562 99.106 -73.559 48.391
288563 99.134 -73.583 48.342
288564 99.161 -73.610 48.295
288565 99.186 -73.650 48.247
288566 99.211 -73.690 48.199
288567 99.244 -73.716 48.152
288568 99.286 -73.732 48.105
288569 99.328 -73.749 48.060
288570 99.370 -73.765 48.014
288571 99.412 -73.781 47.970
288572 99.453 -73.797 47.926
288573 99.495 -73.813 47.883
288574 99.537 -73.829 47.840
288575 99.578 -73.845 47.798
288576 99.618 -73.850 47.755
288577 99.657 -73.857 47.713
288578 99.696 -73.862 47.671
288579 99.736 -73.867 47.630
288580 99.775 -73.872 47.589
288581 99.806 -73.882 47.549
288582 99.819 -73.902 47.508
288583 99.831 -73.922 47.468
288584 99.843 -73.941 47.428
288585 99.860 -73.956 47.388
288586 99.883 -73.963 47.349
288587 99.905 -73.970 47.311
288588 99.928 -73.977 47.274
288589 99.951 -73.984 47.238
288590 99.975 -73.988 47.201
288591 99.999 -73.992 47.164
288592 100.024 -73.997 47.129
288593 100.048 -74.001 47.095
288594 100.072 -74.005 47.061
288595 100.104 -74.011 47.027
288596 100.136 -74.018 46.994
288597 100.168 -74.025 46.961
288598 100.200 -74.031 46.927
288599 100.232 -74.038 46.898
288600 100.264 -74.045 46.868
288601 100.296 -74.051 46.838
288602 100.328 -74.057 46.807
288603 100.359 -74.061 46.773
288604 100.390 -74.064 46.740
288605 100.420 -74.068 46.708
288606 100.451 -74.072 46.678
288607 100.482 -74.076 46.646
288608 100.512 -74.079 46.612
288609 100.543 -74.082 46.577
288610 100.574 -74.086 46.543
288611 100.605 -74.090 46.509
288612 100.634 -74.092 46.477
288613 100.662 -74.094 46.446
288614 100.691 -74.095 46.415
288615 100.719 -74.097 46.385
288616 100.748 -74.098 46.353
288617 100.776 -74.100 46.321
288618 100.805 -74.101 46.289
288619 100.833 -74.113 46.258
288620 100.862 -74.129 46.226
288621 100.892 -74.145 46.192
288622 100.921 -74.162 46.157
288623 100.950 -74.179 46.123
288624 100.979 -74.195 46.089
288625 101.008 -74.211 46.054
288626 101.037 -74.228 46.020
288627 101.066 -74.245 45.985
288628 101.096 -74.261 45.950
288629 101.125 -74.277 45.916
288630 101.159 -74.301 45.882
288631 101.197 -74.329 45.847
288632 101.235 -74.357 45.814
288633 101.273 -74.384 45.782
288634 101.304 -74.405 45.751
288635 101.330 -74.421 45.719
288636 101.357 -74.437 45.686
288637 101.384 -74.468 45.653
288638 101.417 -74.504 45.622
288639 101.495 -74.536 45.592
288640 101.540 -74.558 45.561
288641 101.586 -74.580 45.530
288642 101.631 -74.602 45.499
288643 101.676 -74.624 45.467
288644 101.722 -74.646 45.436
288645 101.766 -74.668 45.405
288646 101.812 -74.690 45.375
288647 101.852 -74.703 45.345
288648 101.882 -74.699 45.315
288649 101.911 -74.693 45.285
288650 101.940 -74.689 45.255
288651 101.969 -74.684 45.224
288652 101.998 -74.679 45.195
288653 102.015 -74.660 45.166
288654 102.015 -74.631 45.137
288655 102.032 -74.631 45.108
288656 102.050 -74.631 45.078
288657 102.067 -74.632 45.049
288658 102.085 -74.632 45.019
288659 102.102 -74.633 44.990
288660 102.120 -74.634 44.961
288661 102.137 -74.634 44.932
288662 102.159 -74.635 44.903
289131 95.070 -72.800 52.660
289132 95.096 -72.807 52.637
289133 95.106 -72.807 52.615
289134 95.120 -72.810 52.593
289135 95.141 -72.813 52.573
289136 95.161 -72.818 52.551
289137 95.182 -72.823 52.530
289138 95.202 -72.827 52.509
289139 95.223 -72.832 52.487
289140 95.244 -72.835 52.464
289141 95.264 -72.840 52.442
289142 95.284 -72.844 52.418
289143 95.305 -72.849 52.394
289144 95.326 -72.854 52.369
289145 95.346 -72.857 52.344
289146 95.378 -72.866 52.320
289147 95.435 -72.884 52.296
289148 95.492 -72.901 52.275
289149 95.549 -72.918 52.252
289150 95.589 -72.917 52.232
289151 95.628 -72.916 52.212
289152 95.654 -72.898 52.193
289153 95.678 -72.876 52.174
289154 95.701 -72.855 52.157
289155 95.737 -72.857 52.141
289156 95.775 -72.866 52.124
289157 95.815 -72.877 52.109
289158 95.854 -72.887 52.094
289159 95.893 -72.898 52.079
289160 95.928 -72.908 52.064
289161 95.948 -72.922 52.050
289162 95.966 -72.935 52.037
289163 95.985 -72.949 52.025
289164 96.004 -72.963 52.011
289165 96.023 -72.976 51.994
289166 96.041 -72.990 51.975
289167 96.060 -73.004 51.954
289168 96.083 -72.994 51.933
289169 96.111 -72.958 51.912
289170 96.138 -72.922 51.891
289171 96.166 -72.886 51.870
289172 96.199 -72.886 51.851
289173 96.234 -72.890 51.833
289174 96.268 -72.893 51.815
289175 96.302 -72.897 51.796
289176 96.340 -72.915 51.778
289177 96.383 -72.954 51.762
289178 96.427 -72.993 51.747
289179 96.430 -73.023 51.729
289180 96.427 -73.012 51.708
289181 96.438 -72.995 51.679
289182 96.469 -72.992 51.649
289183 96.497 -72.999 51.618
289184 96.525 -73.006 51.586
289185 96.552 -73.013 51.555
289186 96.580 -73.020 51.526
289187 96.608 -73.026 51.499
289188 96.636 -73.034 51.471
289189 96.663 -73.041 51.443
289190 96.702 -73.034 51.413
289191 96.741 -73.024 51.382
289192 96.752 -73.028 51.351
289193 96.763 -73.032 51.323
289194 96.774 -73.036 51.296
289195 96.785 -73.042 51.271
289196 96.816 -73.050 51.246
289197 96.860 -73.057 51.220
289198 96.903 -73.064 51.195
289199 96.947 -73.070 51.170
289200 96.991 -73.077 51.145
289201 97.024 -73.084 51.121
289202 97.054 -73.091 51.097
289203 97.085 -73.097 51.075
289204 97.115 -73.104 51.053
289205 97.136 -73.107 51.035
289206 97.160 -73.123 51.017
289207 97.212 -73.123 51.000
289208 97.254 -73.111 50.984
289209 97.287 -73.092 50.967
289210 97.320 -73.073 50.946
289211 97.353 -73.054 50.924
289212 97.384 -73.035 50.900
289213 97.410 -73.014 50.873
289214 97.435 -72.995 50.838
289215 97.449 -72.984 50.800
289216 97.485 -72.972 50.759
289217 97.524 -72.959 50.716
289218 97.563 -72.946 50.675
289219 97.604 -72.960 50.634
289220 97.645 -72.976 50.589
289221 97.686 -72.992 50.539
289222 97.727 -73.009 50.486
289223 97.768 -73.025 50.434
289224 97.809 -73.042 50.382
289225 97.851 -73.058 50.331
289226 97.891 -73.074 50.282
289227 97.933 -73.090 50.231
289228 97.973 -73.107 50.182
289229 98.019 -73.118 50.137
289230 98.065 -73.129 50.089
289231 98.110 -73.140 50.038
289232 98.156 -73.152 49.988
289233 98.202 -73.163 49.940
289234 98.247 -73.171 49.889
289235 98.290 -73.174 49.837
289236 98.333 -73.177 49.783
289237 98.376 -73.180 49.728
289238 98.406 -73.208 49.674
289239 98.432 -73.240 49.620
289240 98.460 -73.264 49.566
289241 98.487 -73.288 49.512
289242 98.514 -73.311 49.457
289243 98.541 -73.336 49.403
289244 98.569 -73.359 49.350
289245 98.584 -73.385 49.298
289246 98.577 -73.444 49.246
289247 98.566 -73.446 49.191
289248 98.640 -73.398 49.138
289249 98.713 -73.350 49.085
289250 98.787 -73.302 49.033
289251 98.843 -73.270 48.982
289252 98.830 -73.296 48.932
289253 98.818 -73.323 48.882
289254 98.805 -73.349 48.833
289255 98.793 -73.375 48.783
289256 98.796 -73.401 48.733
289257 98.846 -73.413 48.682
289258 98.896 -73.425 48.630
289259 98.945 -73.437 48.581
289260 98.994 -73.450 48.533
289261 99.044 -73.462 48.484
289262 99.093 -73.474 48.436
289263 99.118 -73.495 48.388
289264 99.145 -73.518 48.340
289265 99.170 -73.555 48.292
289266 99.196 -73.594 48.244
289267 99.220 -73.634 48.197
289268 99.245 -73.674 48.149
289269 99.283 -73.696 48.103
289270 99.324 -73.712 48.058
289271 99.366 -73.728 48.012
289272 99.408 -73.744 47.967
289273 99.449 -73.760 47.923
289274 99.491 -73.776 47.880
289275 99.533 -73.793 47.837
289276 99.574 -73.809 47.794
289277 99.614 -73.816 47.752
289278 99.653 -73.821 47.709
289279 99.693 -73.827 47.667
289280 99.732 -73.832 47.625
289281 99.772 -73.838 47.584
289282 99.811 -73.843 47.543
289283 99.851 -73.849 47.502
289284 99.863 -73.868 47.461
289285 99.875 -73.888 47.421
289286 99.888 -73.908 47.380
289287 99.899 -73.928 47.341
289288 99.915 -73.943 47.303
289289 99.938 -73.950 47.265
289290 99.961 -73.957 47.228
289291 99.985 -73.962 47.191
289292 100.009 -73.966 47.153
289293 100.033 -73.970 47.118
289294 100.064 -73.976 47.084
289295 100.096 -73.983 47.050
289296 100.128 -73.989 47.016
289297 100.159 -73.996 46.983
289298 100.192 -74.003 46.950
289299 100.223 -74.009 46.916
289300 100.256 -74.016 46.887
289301 100.287 -74.022 46.857
289302 100.320 -74.029 46.827
289303 100.352 -74.035 46.795
289304 100.382 -74.040 46.762
289305 100.413 -74.043 46.729
289306 100.444 -74.047 46.697
289307 100.475 -74.051 46.666
289308 100.505 -74.054 46.633
289309 100.536 -74.057 46.599
289310 100.567 -74.061 46.564
289311 100.596 -74.064 46.530
289312 100.624 -74.065 46.497
289313 100.653 -74.067 46.465
289314 100.682 -74.068 46.434
289315 100.710 -74.070 46.402
289316 100.739 -74.071 46.372
289317 100.767 -74.073 46.341
289318 100.795 -74.074 46.310
289319 100.824 -74.076 46.277
289320 100.853 -74.089 46.244
289321 100.882 -74.106 46.211
289322 100.911 -74.122 46.176
289323 100.940 -74.139 46.142
289324 100.969 -74.155 46.107
289325 100.998 -74.172 46.072
289326 101.028 -74.188 46.038
289327 101.057 -74.205 46.004
289328 101.086 -74.221 45.969
289329 101.115 -74.238 45.936
289330 101.144 -74.254 45.902
289331 101.173 -74.270 45.867
289332 101.202 -74.287 45.833
289333 101.232 -74.304 45.800
289334 101.267 -74.328 45.768
289335 101.305 -74.356 45.736
289336 101.336 -74.377 45.704
289337 101.363 -74.394 45.672
289338 101.398 -74.430 45.640
289339 101.458 -74.456 45.609
289340 101.502 -74.477 45.578
289341 101.548 -74.499 45.547
289342 101.593 -74.521 45.516
289343 101.638 -74.543 45.486
289344 101.684 -74.565 45.454
289345 101.729 -74.587 45.423
289346 101.774 -74.609 45.393
289347 101.819 -74.631 45.363
289348 101.865 -74.653 45.333
289349 101.900 -74.659 45.303
289350 101.929 -74.655 45.273
289351 101.959 -74.650 45.243
289352 101.988 -74.645 45.212
289353 102.017 -74.640 45.183
289354 102.023 -74.608 45.154
289355 102.030 -74.590 45.125
289356 102.047 -74.591 45.095
289357 102.065 -74.592 45.066
289358 102.083 -74.592 45.037
289359 102.100 -74.593 45.007
289360 102.117 -74.593 44.978
289361 102.135 -74.593 44.948
289362 102.157 -74.595 44.920
289363 102.179 -74.596 44.890
289832 95.069 -72.763 52.655
289833 95.095 -72.770 52.632
289834 95.117 -72.776 52.611
289835 95.124 -72.775 52.589
289836 95.144 -72.779 52.568
289837 95.165 -72.783 52.547
289838 95.185 -72.788 52.526
289839 95.205 -72.792 52.505
289840 95.226 -72.797 52.483
289841 95.247 -72.801 52.461
289842 95.267 -72.805 52.439
289843 95.287 -72.810 52.416
289844 95.308 -72.814 52.391
289845 95.329 -72.819 52.367
289846 95.349 -72.823 52.342
289847 95.382 -72.832 52.317
289848 95.440 -72.849 52.293
289849 95.496 -72.866 52.272
289850 95.554 -72.885 52.250
289851 95.609 -72.899 52.228
289852 95.644 -72.894 52.207
289853 95.668 -72.873 52.188
289854 95.692 -72.851 52.169
289855 95.716 -72.829 52.152
289856 95.746 -72.819 52.135
289857 95.784 -72.829 52.118
289858 95.824 -72.840 52.102
289859 95.863 -72.850 52.087
289860 95.902 -72.860 52.072
289861 95.923 -72.873 52.058
289862 95.942 -72.887 52.044
289863 95.961 -72.901 52.032
289864 95.980 -72.914 52.020
289865 95.999 -72.928 52.008
289866 96.017 -72.942 51.992
289867 96.036 -72.955 51.974
289868 96.055 -72.969 51.954
289869 96.074 -72.982 51.933
289870 96.099 -72.957 51.913
289871 96.127 -72.921 51.893
289872 96.155 -72.885 51.874
289873 96.188 -72.882 51.856
289874 96.222 -72.886 51.839
289875 96.256 -72.889 51.823
289876 96.290 -72.893 51.807
289877 96.324 -72.897 51.792
289878 96.361 -72.911 51.780
289879 96.404 -72.951 51.767
289880 96.419 -72.982 51.751
289881 96.417 -72.971 51.729
289882 96.447 -72.961 51.701
289883 96.474 -72.968 51.670
289884 96.502 -72.976 51.637
289885 96.530 -72.982 51.604
289886 96.557 -72.989 51.573
289887 96.585 -72.996 51.545
289888 96.613 -73.003 51.517
289889 96.641 -73.010 51.490
289890 96.668 -73.017 51.461
289891 96.699 -73.019 51.431
289892 96.740 -73.009 51.400
289893 96.751 -73.013 51.369
289894 96.761 -73.017 51.342
289895 96.772 -73.020 51.316
289896 96.793 -73.023 51.291
289897 96.824 -73.024 51.266
289898 96.861 -73.029 51.239
289899 96.905 -73.036 51.212
289900 96.949 -73.043 51.186
289901 96.993 -73.050 51.160
289902 97.027 -73.057 51.133
289903 97.058 -73.063 51.108
289904 97.089 -73.070 51.084
289905 97.119 -73.076 51.061
289906 97.150 -73.083 51.041
289907 97.178 -73.091 51.023
289908 97.218 -73.076 51.005
289909 97.250 -73.057 50.988
289910 97.283 -73.038 50.971
289911 97.316 -73.019 50.950
289912 97.349 -73.001 50.928
289913 97.381 -72.982 50.903
289914 97.414 -72.963 50.874
289915 97.445 -72.943 50.837
289916 97.471 -72.924 50.796
289917 97.495 -72.912 50.753
289918 97.534 -72.900 50.707
289919 97.575 -72.910 50.662
289920 97.615 -72.926 50.616
289921 97.657 -72.943 50.569
289922 97.698 -72.959 50.519
289923 97.739 -72.976 50.467
289924 97.780 -72.992 50.415
289925 97.821 -73.008 50.363
289926 97.862 -73.024 50.311
289927 97.903 -73.040 50.261
289928 97.944 -73.057 50.210
289929 97.986 -73.073 50.162
289930 98.031 -73.084 50.116
289931 98.077 -73.095 50.068
289932 98.123 -73.107 50.017
289933 98.168 -73.118 49.967
289934 98.212 -73.124 49.920
289935 98.255 -73.127 49.872
289936 98.298 -73.130 49.821
289937 98.342 -73.133 49.769
289938 98.384 -73.135 49.715
289939 98.413 -73.169 49.660
289940 98.438 -73.205 49.607
289941 98.465 -73.229 49.553
289942 98.493 -73.252 49.499
289943 98.520 -73.276 49.446
289944 98.548 -73.300 49.393
289945 98.574 -73.324 49.342
289946 98.584 -73.352 49.290
289947 98.583 -73.410 49.237
289948 98.570 -73.409 49.183
289949 98.634 -73.352 49.130
289950 98.707 -73.305 49.078
289951 98.781 -73.257 49.027
289952 98.807 -73.250 48.977
289953 98.794 -73.276 48.928
289954 98.782 -73.302 48.878
289955 98.770 -73.329 48.829
289956 98.772 -73.356 48.778
289957 98.808 -73.387 48.728
289958 98.848 -73.418 48.678
289959 98.892 -73.420 48.627
289960 98.938 -73.416 48.578
289961 98.985 -73.412 48.530
289962 99.030 -73.408 48.482
289963 99.080 -73.420 48.433
289964 99.129 -73.432 48.386
289965 99.155 -73.459 48.339
289966 99.181 -73.499 48.291
289967 99.205 -73.539 48.242
289968 99.230 -73.578 48.195
289969 99.255 -73.618 48.147
289970 99.280 -73.658 48.101
289971 99.321 -73.675 48.056
289972 99.362 -73.691 48.010
289973 99.404 -73.708 47.965
289974 99.445 -73.724 47.921
289975 99.487 -73.740 47.878
289976 99.529 -73.756 47.834
289977 99.571 -73.772 47.791
289978 99.611 -73.781 47.748
289979 99.650 -73.786 47.706
289980 99.690 -73.791 47.663
289981 99.729 -73.797 47.621
289982 99.768 -73.803 47.579
289983 99.808 -73.808 47.537
289984 99.847 -73.813 47.496
289985 99.887 -73.819 47.455
289986 99.908 -73.835 47.413
289987 99.920 -73.854 47.372
289988 99.932 -73.874 47.333
289989 99.944 -73.894 47.294
289990 99.956 -73.913 47.256
289991 99.971 -73.930 47.218
289992 99.994 -73.935 47.180
289993 100.023 -73.941 47.143
289994 100.055 -73.948 47.107
289995 100.087 -73.954 47.073
289996 100.119 -73.961 47.040
289997 100.151 -73.967 47.005
289998 100.183 -73.974 46.971
289999 100.215 -73.981 46.939
290000 100.247 -73.987 46.906
290001 100.279 -73.994 46.876
290002 100.311 -74.000 46.847
290003 100.343 -74.007 46.816
290004 100.375 -74.013 46.783
290005 100.407 -74.019 46.750
290006 100.437 -74.022 46.718
290007 100.468 -74.026 46.686
290008 100.498 -74.029 46.654
290009 100.529 -74.033 46.621
290010 100.558 -74.035 46.586
290011 100.587 -74.036 46.552
290012 100.615 -74.038 46.519
290013 100.644 -74.039 46.485
290014 100.672 -74.041 46.454
290015 100.701 -74.042 46.424
290016 100.729 -74.045 46.392
290017 100.758 -74.046 46.362
290018 100.786 -74.048 46.332
290019 100.815 -74.049 46.300
290020 100.843 -74.051 46.266
290021 100.873 -74.066 46.233
290022 100.901 -74.082 46.198
290023 100.930 -74.099 46.163
290024 100.960 -74.115 46.128
290025 100.989 -74.132 46.093
290026 101.018 -74.148 46.058
290027 101.047 -74.165 46.023
290028 101.076 -74.181 45.989
290029 101.105 -74.198 45.955
290030 101.134 -74.214 45.922
290031 101.164 -74.231 45.888
290032 101.193 -74.247 45.854
290033 101.221 -74.264 45.819
290034 101.251 -74.280 45.787
290035 101.280 -74.297 45.755
290036 101.309 -74.313 45.722
290037 101.338 -74.330 45.691
290038 101.375 -74.354 45.659
290039 101.420 -74.376 45.627
290040 101.465 -74.398 45.596
290041 101.510 -74.420 45.565
290042 101.555 -74.442 45.534
290043 101.601 -74.463 45.504
290044 101.646 -74.485 45.473
290045 101.691 -74.507 45.442
290046 101.736 -74.529 45.411
290047 101.781 -74.551 45.381
290048 101.827 -74.573 45.351
290049 101.872 -74.595 45.321
290050 101.917 -74.617 45.291
290051 101.949 -74.615 45.261
290052 101.978 -74.611 45.231
290053 102.007 -74.606 45.200
290054 102.036 -74.601 45.172
290055 102.031 -74.556 45.142
290056 102.045 -74.551 45.112
290057 102.062 -74.551 45.081
290058 102.080 -74.552 45.053
290059 102.097 -74.552 45.024
290060 102.115 -74.553 44.995
290061 102.132 -74.553 44.965
290062 102.154 -74.554 44.936
290063 102.177 -74.555 44.907
290064 102.200 -74.556 44.878
290533 95.068 -72.725 52.651
290534 95.094 -72.733 52.628
290535 95.119 -72.740 52.606
290536 95.135 -72.743 52.584
290537 95.147 -72.744 52.564
290538 95.168 -72.748 52.544
290539 95.188 -72.753 52.523
290540 95.208 -72.757 52.502
290541 95.229 -72.762 52.481
290542 95.250 -72.766 52.458
290543 95.270 -72.770 52.436
290544 95.290 -72.775 52.413
290545 95.311 -72.779 52.389
290546 95.332 -72.784 52.365
290547 95.352 -72.788 52.340
290548 95.387 -72.797 52.315
290549 95.444 -72.815 52.291
290550 95.501 -72.832 52.269
290551 95.558 -72.850 52.247
290552 95.615 -72.867 52.225
290553 95.659 -72.868 52.203
290554 95.683 -72.847 52.184
290555 95.707 -72.826 52.165
290556 95.731 -72.804 52.147
290557 95.755 -72.783 52.130
290558 95.794 -72.792 52.112
290559 95.833 -72.803 52.096
290560 95.872 -72.813 52.081
290561 95.899 -72.825 52.066
290562 95.918 -72.838 52.050
290563 95.937 -72.852 52.036
290564 95.956 -72.866 52.024
290565 95.974 -72.879 52.013
290566 95.993 -72.893 52.002
290567 96.012 -72.907 51.987
290568 96.031 -72.920 51.970
290569 96.049 -72.934 51.951
290570 96.068 -72.948 51.932
290571 96.088 -72.956 51.913
290572 96.115 -72.920 51.895
290573 96.143 -72.884 51.878
290574 96.176 -72.879 51.861
290575 96.210 -72.882 51.845
290576 96.244 -72.885 51.831
290577 96.278 -72.889 51.818
290578 96.312 -72.893 51.807
290579 96.346 -72.896 51.798
290580 96.382 -72.907 51.787
290581 96.409 -72.940 51.772
290582 96.424 -72.931 51.750
290583 96.452 -72.938 51.722
290584 96.479 -72.945 51.690
290585 96.507 -72.951 51.656
290586 96.535 -72.958 51.623
290587 96.562 -72.965 51.593
290588 96.590 -72.973 51.564
290589 96.618 -72.979 51.537
290590 96.646 -72.986 51.509
290591 96.673 -72.993 51.480
290592 96.701 -73.000 51.450
290593 96.738 -72.994 51.419
290594 96.749 -72.998 51.388
290595 96.760 -73.001 51.362
290596 96.789 -73.003 51.337
290597 96.820 -73.004 51.313
290598 96.852 -73.006 51.288
290599 96.884 -73.007 51.260
290600 96.914 -73.011 51.232
290601 96.950 -73.015 51.205
290602 96.994 -73.022 51.177
290603 97.031 -73.029 51.147
290604 97.062 -73.036 51.120
290605 97.092 -73.042 51.095
290606 97.123 -73.049 51.071
290607 97.158 -73.051 51.048
290608 97.190 -73.053 51.030
290609 97.221 -73.032 51.011
290610 97.253 -73.012 50.992
290611 97.285 -72.992 50.974
290612 97.316 -72.971 50.954
290613 97.348 -72.951 50.931
290614 97.379 -72.930 50.905
290615 97.410 -72.910 50.874
290616 97.443 -72.890 50.835
290617 97.475 -72.872 50.793
290618 97.507 -72.853 50.748
290619 97.545 -72.860 50.701
290620 97.586 -72.876 50.651
290621 97.628 -72.893 50.602
290622 97.669 -72.909 50.552
290623 97.710 -72.926 50.501
290624 97.751 -72.942 50.450
290625 97.792 -72.958 50.399
290626 97.833 -72.974 50.347
290627 97.874 -72.991 50.294
290628 97.915 -73.007 50.242
290629 97.956 -73.023 50.192
290630 97.998 -73.039 50.144
290631 98.043 -73.051 50.096
290632 98.089 -73.061 50.048
290633 98.135 -73.073 49.997
290634 98.178 -73.077 49.948
290635 98.221 -73.079 49.902
290636 98.264 -73.083 49.856
290637 98.307 -73.085 49.807
290638 98.350 -73.088 49.756
290639 98.393 -73.091 49.703
290640 98.419 -73.129 49.648
290641 98.444 -73.168 49.595
290642 98.472 -73.193 49.541
290643 98.499 -73.217 49.487
290644 98.526 -73.241 49.435
290645 98.554 -73.264 49.385
290646 98.574 -73.288 49.334
290647 98.583 -73.320 49.281
290648 98.589 -73.377 49.227
290649 98.576 -73.375 49.173
290650 98.627 -73.307 49.121
290651 98.701 -73.259 49.070
290652 98.774 -73.211 49.021
290653 98.771 -73.230 48.972
290654 98.759 -73.256 48.923
290655 98.746 -73.283 48.874
290656 98.750 -73.310 48.824
290657 98.797 -73.341 48.774
290658 98.825 -73.371 48.723
290659 98.865 -73.402 48.673
290660 98.904 -73.434 48.625
290661 98.946 -73.456 48.577
290662 98.992 -73.451 48.529
290663 99.038 -73.447 48.481
290664 99.083 -73.443 48.432
290665 99.129 -73.439 48.385
290666 99.164 -73.421 48.338
290667 99.190 -73.443 48.290
290668 99.215 -73.483 48.242
290669 99.240 -73.523 48.194
290670 99.265 -73.562 48.147
290671 99.290 -73.602 48.100
290672 99.317 -73.639 48.055
290673 99.358 -73.655 48.008
290674 99.400 -73.671 47.962
290675 99.442 -73.687 47.918
290676 99.484 -73.703 47.875
290677 99.525 -73.719 47.832
290678 99.567 -73.735 47.788
290679 99.608 -73.746 47.745
290680 99.647 -73.751 47.703
290681 99.686 -73.756 47.659
290682 99.725 -73.762 47.617
290683 99.765 -73.768 47.574
290684 99.804 -73.773 47.532
290685 99.844 -73.778 47.489
290686 99.883 -73.785 47.448
290687 99.923 -73.790 47.406
290688 99.952 -73.800 47.365
290689 99.965 -73.820 47.324
290690 99.977 -73.840 47.285
290691 99.989 -73.860 47.246
290692 100.001 -73.879 47.208
290693 100.015 -73.899 47.170
290694 100.043 -73.912 47.132
290695 100.076 -73.920 47.097
290696 100.109 -73.929 47.062
290697 100.142 -73.938 47.028
290698 100.174 -73.945 46.994
290699 100.207 -73.952 46.961
290700 100.239 -73.958 46.928
290701 100.271 -73.965 46.896
290702 100.302 -73.972 46.865
290703 100.335 -73.978 46.836
290704 100.366 -73.985 46.806
290705 100.399 -73.991 46.773
290706 100.430 -73.998 46.741
290707 100.461 -74.001 46.708
290708 100.492 -74.004 46.675
290709 100.521 -74.007 46.643
290710 100.549 -74.008 46.609
290711 100.577 -74.010 46.574
290712 100.606 -74.011 46.541
290713 100.635 -74.013 46.508
290714 100.663 -74.014 46.474
290715 100.692 -74.016 46.444
290716 100.720 -74.017 46.414
290717 100.749 -74.019 46.385
290718 100.777 -74.020 46.355
290719 100.805 -74.022 46.323
290720 100.834 -74.023 46.290
290721 100.862 -74.026 46.255
290722 100.892 -74.042 46.222
290723 100.921 -74.059 46.187
290724 100.950 -74.076 46.151
290725 100.979 -74.092 46.116
290726 101.008 -74.108 46.081
290727 101.037 -74.125 46.044
290728 101.066 -74.142 46.009
290729 101.096 -74.157 45.976
290730 101.125 -74.174 45.942
290731 101.153 -74.191 45.909
290732 101.183 -74.206 45.875
290733 101.212 -74.221 45.840
290734 101.241 -74.236 45.806
290735 101.270 -74.252 45.775
290736 101.299 -74.267 45.743
290737 101.328 -74.282 45.710
290738 101.357 -74.297 45.679
290739 101.388 -74.314 45.647
290740 101.428 -74.330 45.614
290741 101.473 -74.346 45.583
290742 101.518 -74.361 45.553
290743 101.563 -74.383 45.522
290744 101.608 -74.405 45.491
290745 101.653 -74.427 45.461
290746 101.699 -74.449 45.430
290747 101.744 -74.471 45.399
290748 101.789 -74.493 45.369
290749 101.834 -74.515 45.340
290750 101.880 -74.537 45.310
290751 101.925 -74.559 45.280
290752 101.968 -74.577 45.250
290753 101.997 -74.571 45.219
290754 102.026 -74.567 45.188
290755 102.055 -74.562 45.160
290756 102.042 -74.511 45.130
290757 102.060 -74.511 45.101
290758 102.078 -74.512 45.071
290759 102.095 -74.512 45.041
290760 102.112 -74.512 45.011
290761 102.130 -74.513 44.982
290762 102.152 -74.514 44.953
290763 102.174 -74.515 44.924
290764 102.197 -74.516 44.895
290765 102.221 -74.517 44.866
291235 95.093 -72.696 52.624
291236 95.118 -72.703 52.601
291237 95.144 -72.710 52.580
291238 95.153 -72.710 52.560
291239 95.171 -72.713 52.540
291240 95.191 -72.718 52.520
291241 95.211 -72.722 52.500
291242 95.232 -72.727 52.478
291243 95.253 -72.731 52.455
291244 95.273 -72.735 52.433
291245 95.294 -72.740 52.411
291246 95.314 -72.744 52.387
291247 95.335 -72.749 52.363
291248 95.355 -72.753 52.338
291249 95.391 -72.763 52.313
291250 95.449 -72.781 52.289
291251 95.506 -72.798 52.266
291252 95.563 -72.816 52.243
291253 95.620 -72.833 52.221
291254 95.656 -72.823 52.200
291255 95.691 -72.814 52.180
291256 95.722 -72.801 52.161
291257 95.746 -72.779 52.142
291258 95.770 -72.758 52.124
291259 95.803 -72.755 52.108
291260 95.842 -72.766 52.091
291261 95.875 -72.777 52.075
291262 95.894 -72.790 52.059
291263 95.912 -72.804 52.043
291264 95.932 -72.817 52.029
291265 95.950 -72.831 52.016
291266 95.969 -72.844 52.004
291267 95.987 -72.858 51.991
291268 96.007 -72.872 51.977
291269 96.025 -72.885 51.962
291270 96.044 -72.899 51.945
291271 96.063 -72.913 51.928
291272 96.082 -72.926 51.911
291273 96.104 -72.919 51.895
291274 96.131 -72.883 51.879
291275 96.164 -72.875 51.864
291276 96.198 -72.879 51.851
291277 96.232 -72.882 51.839
291278 96.266 -72.885 51.829
291279 96.300 -72.889 51.819
291280 96.334 -72.892 51.811
291281 96.369 -72.896 51.801
291282 96.402 -72.900 51.786
291283 96.430 -72.907 51.765
291284 96.457 -72.914 51.736
291285 96.485 -72.921 51.705
291286 96.513 -72.928 51.673
291287 96.540 -72.935 51.642
291288 96.568 -72.942 51.613
291289 96.596 -72.948 51.584
291290 96.623 -72.955 51.556
291291 96.651 -72.963 51.529
291292 96.678 -72.970 51.499
291293 96.706 -72.976 51.469
291294 96.736 -72.979 51.439
291295 96.754 -72.982 51.410
291296 96.786 -72.983 51.384
291297 96.817 -72.985 51.360
291298 96.849 -72.986 51.336
291299 96.880 -72.988 51.311
291300 96.912 -72.989 51.285
291301 96.944 -72.991 51.255
291302 96.974 -72.994 51.225
291303 97.004 -72.997 51.196
291304 97.036 -73.000 51.164
291305 97.072 -73.002 51.136
291306 97.108 -73.004 51.109
291307 97.144 -73.005 51.082
291308 97.179 -73.007 51.057
291309 97.210 -73.015 51.037
291310 97.237 -73.005 51.017
291311 97.269 -72.985 50.997
291312 97.301 -72.964 50.979
291313 97.332 -72.944 50.958
291314 97.364 -72.923 50.933
291315 97.395 -72.903 50.904
291316 97.427 -72.882 50.872
291317 97.458 -72.862 50.832
291318 97.490 -72.842 50.790
291319 97.522 -72.825 50.744
291320 97.560 -72.833 50.696
291321 97.597 -72.844 50.645
291322 97.639 -72.860 50.594
291323 97.680 -72.876 50.542
291324 97.722 -72.892 50.488
291325 97.762 -72.908 50.435
291326 97.804 -72.925 50.383
291327 97.845 -72.941 50.332
291328 97.886 -72.957 50.279
291329 97.927 -72.973 50.226
291330 97.968 -72.990 50.176
291331 98.010 -73.005 50.127
291332 98.056 -73.017 50.078
291333 98.101 -73.027 50.028
291334 98.144 -73.029 49.979
291335 98.187 -73.032 49.932
291336 98.230 -73.036 49.887
291337 98.273 -73.038 49.841
291338 98.316 -73.041 49.793
291339 98.359 -73.043 49.742
291340 98.400 -73.051 49.689
291341 98.425 -73.089 49.636
291342 98.451 -73.129 49.583
291343 98.478 -73.158 49.530
291344 98.505 -73.181 49.477
291345 98.532 -73.205 49.425
291346 98.560 -73.229 49.373
291347 98.573 -73.252 49.321
291348 98.583 -73.286 49.270
291349 98.594 -73.343 49.217
291350 98.582 -73.341 49.164
291351 98.621 -73.261 49.113
291352 98.695 -73.214 49.063
291353 98.747 -73.183 49.014
291354 98.735 -73.210 48.965
291355 98.723 -73.236 48.917
291356 98.729 -73.264 48.868
291357 98.777 -73.295 48.818
291358 98.823 -73.326 48.768
291359 98.842 -73.355 48.717
291360 98.882 -73.387 48.668
291361 98.921 -73.418 48.620
291362 98.961 -73.449 48.572
291363 99.000 -73.481 48.525
291364 99.045 -73.487 48.477
291365 99.090 -73.482 48.429
291366 99.137 -73.478 48.382
291367 99.170 -73.459 48.335
291368 99.198 -73.430 48.288
291369 99.225 -73.427 48.240
291370 99.250 -73.467 48.192
291371 99.275 -73.507 48.144
291372 99.300 -73.546 48.097
291373 99.325 -73.586 48.051
291374 99.355 -73.618 48.005
291375 99.397 -73.634 47.960
291376 99.438 -73.650 47.916
291377 99.480 -73.667 47.872
291378 99.521 -73.683 47.828
291379 99.563 -73.699 47.785
291380 99.604 -73.710 47.741
291381 99.643 -73.716 47.698
291382 99.683 -73.722 47.654
291383 99.722 -73.727 47.611
291384 99.762 -73.732 47.569
291385 99.801 -73.738 47.526
291386 99.840 -73.744 47.483
291387 99.880 -73.749 47.441
291388 99.919 -73.755 47.399
291389 99.959 -73.760 47.357
291390 99.997 -73.766 47.316
291391 100.009 -73.786 47.276
291392 100.021 -73.806 47.236
291393 100.033 -73.825 47.198
291394 100.050 -73.845 47.159
291395 100.067 -73.865 47.122
291396 100.091 -73.880 47.086
291397 100.124 -73.889 47.051
291398 100.157 -73.897 47.016
291399 100.190 -73.907 46.983
291400 100.223 -73.915 46.950
291401 100.256 -73.923 46.917
291402 100.289 -73.932 46.885
291403 100.322 -73.941 46.854
291404 100.355 -73.950 46.825
291405 100.388 -73.958 46.795
291406 100.421 -73.966 46.763
291407 100.454 -73.976 46.731
291408 100.483 -73.978 46.697
291409 100.511 -73.979 46.663
291410 100.540 -73.981 46.631
291411 100.568 -73.982 46.597
291412 100.597 -73.984 46.563
291413 100.625 -73.985 46.529
291414 100.654 -73.987 46.497
291415 100.683 -73.988 46.463
291416 100.711 -73.990 46.433
291417 100.740 -73.991 46.405
291418 100.768 -73.993 46.378
291419 100.796 -73.995 46.347
291420 100.825 -73.997 46.314
291421 100.853 -73.998 46.280
291422 100.882 -74.003 46.245
291423 100.911 -74.019 46.210
291424 100.940 -74.034 46.175
291425 100.969 -74.049 46.140
291426 100.998 -74.064 46.104
291427 101.028 -74.079 46.068
291428 101.057 -74.094 46.032
291429 101.086 -74.109 45.998
291430 101.115 -74.124 45.964
291431 101.144 -74.139 45.930
291432 101.173 -74.154 45.896
291433 101.202 -74.170 45.862
291434 101.231 -74.185 45.828
291435 101.260 -74.200 45.794
291436 101.289 -74.215 45.763
291437 101.319 -74.230 45.731
291438 101.348 -74.245 45.698
291439 101.377 -74.261 45.666
291440 101.408 -74.280 45.635
291441 101.439 -74.299 45.603
291442 101.483 -74.314 45.571
291443 101.528 -74.329 45.541
291444 101.572 -74.343 45.510
291445 101.617 -74.358 45.480
291446 101.662 -74.373 45.450
291447 101.706 -74.391 45.419
291448 101.751 -74.412 45.388
291449 101.796 -74.434 45.358
291450 101.842 -74.456 45.329
291451 101.887 -74.478 45.299
291452 101.933 -74.500 45.269
291453 101.977 -74.522 45.239
291454 102.016 -74.533 45.208
291455 102.045 -74.527 45.178
291456 102.063 -74.510 45.149
291457 102.057 -74.471 45.120
291458 102.075 -74.471 45.090
291459 102.092 -74.472 45.060
291460 102.110 -74.472 45.030
291461 102.127 -74.473 45.000
291462 102.149 -74.474 44.970
291463 102.172 -74.475 44.942
291464 102.195 -74.476 44.913
291465 102.218 -74.477 44.883
291466 102.241 -74.477 44.854
291937 95.117 -72.666 52.598
291938 95.143 -72.673 52.577
291939 95.164 -72.678 52.556
291940 95.174 -72.678 52.537
291941 95.194 -72.683 52.517
291942 95.214 -72.688 52.496
291943 95.235 -72.692 52.475
291944 95.256 -72.696 52.453
291945 95.276 -72.700 52.430
291946 95.297 -72.705 52.408
291947 95.317 -72.710 52.384
291948 95.338 -72.714 52.360
291949 95.358 -72.718 52.336
291950 95.396 -72.728 52.311
291951 95.453 -72.746 52.287
291952 95.510 -72.763 52.263
291953 95.567 -72.781 52.240
291954 95.619 -72.791 52.218
291955 95.652 -72.778 52.197
291956 95.687 -72.769 52.176
291957 95.722 -72.760 52.156
291958 95.757 -72.750 52.137
291959 95.784 -72.733 52.119
291960 95.812 -72.718 52.103
291961 95.851 -72.728 52.085
291962 95.870 -72.742 52.068
291963 95.888 -72.756 52.052
291964 95.907 -72.769 52.035
291965 95.926 -72.783 52.021
291966 95.945 -72.797 52.007
291967 95.963 -72.810 51.994
291968 95.982 -72.824 51.981
291969 96.001 -72.837 51.967
291970 96.020 -72.851 51.953
291971 96.039 -72.864 51.939
291972 96.057 -72.878 51.924
291973 96.076 -72.891 51.908
291974 96.095 -72.905 51.894
291975 96.120 -72.882 51.880
291976 96.152 -72.871 51.867
291977 96.186 -72.875 51.857
291978 96.220 -72.879 51.847
291979 96.254 -72.882 51.838
291980 96.289 -72.885 51.829
291981 96.325 -72.882 51.822
291982 96.362 -72.876 51.811
291983 96.399 -72.870 51.796
291984 96.428 -72.875 51.775
291985 96.456 -72.882 51.748
291986 96.484 -72.889 51.719
291987 96.512 -72.897 51.689
291988 96.540 -72.904 51.660
291989 96.568 -72.911 51.631
291990 96.596 -72.919 51.603
291991 96.624 -72.926 51.575
291992 96.652 -72.933 51.547
291993 96.680 -72.941 51.519
291994 96.708 -72.948 51.489
291995 96.743 -72.951 51.460
291996 96.774 -72.951 51.431
291997 96.807 -72.955 51.406
291998 96.840 -72.958 51.382
291999 96.873 -72.961 51.359
292000 96.905 -72.964 51.334
292001 96.938 -72.968 51.308
292002 96.970 -72.971 51.279
292003 97.003 -72.974 51.248
292004 97.032 -72.969 51.217
292005 97.058 -72.957 51.185
292006 97.093 -72.958 51.155
292007 97.129 -72.960 51.125
292008 97.165 -72.962 51.096
292009 97.201 -72.963 51.069
292010 97.232 -72.970 51.046
292011 97.253 -72.979 51.023
292012 97.285 -72.958 51.001
292013 97.317 -72.938 50.982
292014 97.348 -72.917 50.960
292015 97.380 -72.897 50.933
292016 97.411 -72.876 50.903
292017 97.443 -72.856 50.867
292018 97.474 -72.835 50.826
292019 97.506 -72.815 50.782
292020 97.540 -72.807 50.736
292021 97.578 -72.815 50.687
292022 97.615 -72.823 50.637
292023 97.647 -72.828 50.588
292024 97.688 -72.843 50.536
292025 97.730 -72.860 50.481
292026 97.772 -72.876 50.426
292027 97.814 -72.891 50.372
292028 97.856 -72.907 50.318
292029 97.898 -72.923 50.264
292030 97.939 -72.940 50.212
292031 97.980 -72.956 50.160
292032 98.022 -72.972 50.110
292033 98.067 -72.979 50.060
292034 98.109 -72.982 50.011
292035 98.153 -72.985 49.963
292036 98.196 -72.988 49.917
292037 98.238 -72.991 49.873
292038 98.282 -72.993 49.827
292039 98.324 -72.996 49.779
292040 98.368 -72.999 49.728
292041 98.406 -73.011 49.675
292042 98.432 -73.050 49.623
292043 98.457 -73.089 49.571
292044 98.484 -73.122 49.519
292045 98.511 -73.146 49.466
292046 98.538 -73.170 49.413
292047 98.563 -73.193 49.361
292048 98.573 -73.217 49.309
292049 98.582 -73.254 49.257
292050 98.593 -73.310 49.206
292051 98.588 -73.307 49.155
292052 98.615 -73.216 49.104
292053 98.688 -73.168 49.055
292054 98.712 -73.163 49.006
292055 98.700 -73.189 48.958
292056 98.708 -73.217 48.910
292057 98.756 -73.249 48.861
292058 98.802 -73.280 48.811
292059 98.838 -73.310 48.762
292060 98.859 -73.340 48.712
292061 98.899 -73.371 48.662
292062 98.938 -73.402 48.614
292063 98.978 -73.434 48.567
292064 99.017 -73.465 48.520
292065 99.057 -73.497 48.472
292066 99.098 -73.521 48.426
292067 99.144 -73.518 48.379
292068 99.176 -73.496 48.331
292069 99.204 -73.468 48.284
292070 99.232 -73.440 48.237
292071 99.260 -73.412 48.188
292072 99.285 -73.451 48.141
292073 99.309 -73.490 48.094
292074 99.335 -73.531 48.048
292075 99.359 -73.570 48.002
292076 99.393 -73.598 47.957
292077 99.434 -73.614 47.913
292078 99.476 -73.630 47.869
292079 99.518 -73.647 47.824
292080 99.560 -73.662 47.780
292081 99.600 -73.675 47.737
292082 99.640 -73.681 47.693
292083 99.680 -73.687 47.649
292084 99.719 -73.692 47.606
292085 99.758 -73.697 47.563
292086 99.797 -73.703 47.520
292087 99.837 -73.709 47.477
292088 99.876 -73.714 47.435
292089 99.916 -73.719 47.392
292090 99.955 -73.725 47.349
292091 99.994 -73.731 47.308
292092 100.034 -73.736 47.267
292093 100.054 -73.752 47.227
292094 100.066 -73.772 47.187
292095 100.084 -73.791 47.148
292096 100.102 -73.811 47.110
292097 100.119 -73.831 47.074
292098 100.139 -73.849 47.039
292099 100.172 -73.857 47.004
292100 100.205 -73.866 46.971
292101 100.238 -73.875 46.938
292102 100.271 -73.884 46.905
292103 100.304 -73.892 46.873
292104 100.337 -73.901 46.845
292105 100.370 -73.910 46.815
292106 100.403 -73.918 46.783
292107 100.436 -73.927 46.751
292108 100.466 -73.932 46.720
292109 100.485 -73.922 46.685
292110 100.511 -73.919 46.651
292111 100.540 -73.922 46.619
292112 100.568 -73.923 46.585
292113 100.597 -73.925 46.551
292114 100.625 -73.927 46.518
292115 100.654 -73.929 46.486
292116 100.683 -73.930 46.453
292117 100.711 -73.932 46.425
292118 100.740 -73.934 46.397
292119 100.768 -73.935 46.369
292120 100.797 -73.938 46.336
292121 100.826 -73.939 46.303
292122 100.854 -73.941 46.268
292123 100.893 -73.967 46.233
292124 100.930 -73.982 46.198
292125 100.960 -73.998 46.164
292126 100.989 -74.013 46.129
292127 101.018 -74.028 46.093
292128 101.047 -74.043 46.057
292129 101.076 -74.058 46.020
292130 101.105 -74.073 45.986
292131 101.134 -74.089 45.952
292132 101.163 -74.104 45.919
292133 101.192 -74.119 45.885
292134 101.221 -74.134 45.851
292135 101.251 -74.149 45.816
292136 101.280 -74.164 45.783
292137 101.309 -74.179 45.751
292138 101.338 -74.195 45.719
292139 101.367 -74.210 45.687
292140 101.397 -74.227 45.654
292141 101.428 -74.246 45.623
292142 101.460 -74.265 45.591
292143 101.494 -74.283 45.559
292144 101.538 -74.298 45.529
292145 101.583 -74.312 45.498
292146 101.627 -74.327 45.468
292147 101.672 -74.342 45.438
292148 101.716 -74.357 45.408
292149 101.761 -74.371 45.377
292150 101.805 -74.386 45.347
292151 101.849 -74.401 45.318
292152 101.895 -74.420 45.288
292153 101.940 -74.442 45.258
292154 101.985 -74.464 45.227
292155 102.030 -74.486 45.197
292156 102.064 -74.489 45.168
292157 102.071 -74.458 45.138
292158 102.073 -74.431 45.109
292159 102.090 -74.432 45.080
292160 102.107 -74.432 45.048
292161 102.125 -74.433 45.019
292162 102.147 -74.433 44.989
292163 102.170 -74.434 44.960
292164 102.192 -74.436 44.932
292165 102.216 -74.436 44.902
292166 102.239 -74.437 44.872
292167 102.261 -74.439 44.842
292639 95.141 -72.636 52.573
292640 95.168 -72.644 52.553
292641 95.181 -72.646 52.534
292642 95.197 -72.648 52.513
292643 95.217 -72.653 52.493
292644 95.238 -72.657 52.472
292645 95.259 -72.661 52.450
292646 95.279 -72.666 52.428
292647 95.300 -72.670 52.406
292648 95.320 -72.675 52.382
292649 95.341 -72.679 52.357
292650 95.361 -72.683 52.332
292651 95.400 -72.694 52.308
292652 95.457 -72.712 52.283
292653 95.515 -72.729 52.259
292654 95.571 -72.747 52.236
292655 95.614 -72.744 52.214
292656 95.648 -72.733 52.193
292657 95.683 -72.723 52.171
292658 95.718 -72.714 52.151
292659 95.753 -72.705 52.132
292660 95.788 -72.695 52.115
292661 95.823 -72.686 52.097
292662 95.845 -72.694 52.079
292663 95.864 -72.707 52.062
292664 95.883 -72.721 52.045
292665 95.902 -72.735 52.028
292666 95.920 -72.748 52.013
292667 95.939 -72.762 51.998
292668 95.958 -72.776 51.984
292669 95.977 -72.789 51.970
292670 95.995 -72.803 51.956
292671 96.015 -72.816 51.944
292672 96.033 -72.830 51.931
292673 96.052 -72.844 51.918
292674 96.071 -72.857 51.906
292675 96.090 -72.871 51.892
292676 96.109 -72.881 51.880
292677 96.140 -72.868 51.870
292678 96.174 -72.871 51.861
292679 96.209 -72.871 51.853
292680 96.246 -72.865 51.846
292681 96.284 -72.859 51.839
292682 96.320 -72.853 51.831
292683 96.358 -72.846 51.819
292684 96.394 -72.840 51.803
292685 96.426 -72.842 51.783
292686 96.454 -72.849 51.759
292687 96.482 -72.857 51.732
292688 96.510 -72.864 51.704
292689 96.538 -72.871 51.677
292690 96.566 -72.879 51.648
292691 96.594 -72.886 51.620
292692 96.622 -72.893 51.593
292693 96.650 -72.901 51.566
292694 96.678 -72.908 51.539
292695 96.719 -72.910 51.510
292696 96.754 -72.910 51.481
292697 96.785 -72.905 51.451
292698 96.817 -72.908 51.426
292699 96.850 -72.912 51.404
292700 96.883 -72.915 51.381
292701 96.915 -72.918 51.357
292702 96.948 -72.921 51.331
292703 96.981 -72.925 51.303
292704 97.014 -72.928 51.272
292705 97.058 -72.931 51.240
292706 97.084 -72.920 51.206
292707 97.114 -72.914 51.174
292708 97.150 -72.916 51.142
292709 97.186 -72.918 51.112
292710 97.222 -72.920 51.083
292711 97.253 -72.926 51.056
292712 97.272 -72.946 51.030
292713 97.301 -72.931 51.006
292714 97.333 -72.910 50.986
292715 97.364 -72.890 50.962
292716 97.396 -72.870 50.934
292717 97.427 -72.849 50.901
292718 97.459 -72.829 50.862
292719 97.490 -72.808 50.819
292720 97.522 -72.788 50.773
292721 97.558 -72.789 50.726
292722 97.596 -72.797 50.678
292723 97.634 -72.805 50.629
292724 97.671 -72.813 50.580
292725 97.704 -72.819 50.530
292726 97.738 -72.827 50.476
292727 97.780 -72.843 50.419
292728 97.822 -72.859 50.362
292729 97.864 -72.875 50.307
292730 97.906 -72.891 50.252
292731 97.948 -72.907 50.199
292732 97.990 -72.923 50.146
292733 98.030 -72.939 50.094
292734 98.075 -72.935 50.044
292735 98.119 -72.938 49.996
292736 98.161 -72.941 49.949
292737 98.204 -72.944 49.903
292738 98.247 -72.946 49.859
292739 98.290 -72.949 49.813
292740 98.334 -72.951 49.765
292741 98.376 -72.954 49.714
292742 98.413 -72.971 49.661
292743 98.438 -73.011 49.609
292744 98.463 -73.049 49.559
292745 98.489 -73.086 49.506
292746 98.517 -73.111 49.453
292747 98.544 -73.134 49.399
292748 98.563 -73.158 49.348
292749 98.572 -73.181 49.296
292750 98.582 -73.221 49.245
292751 98.593 -73.277 49.195
292752 98.594 -73.273 49.145
292753 98.608 -73.170 49.096
292754 98.682 -73.122 49.046
292755 98.676 -73.143 48.998
292756 98.687 -73.171 48.950
292757 98.734 -73.202 48.902
292758 98.782 -73.233 48.854
292759 98.829 -73.265 48.805
292760 98.852 -73.294 48.756
292761 98.876 -73.324 48.706
292762 98.916 -73.355 48.657
292763 98.955 -73.387 48.608
292764 98.995 -73.418 48.561
292765 99.034 -73.449 48.514
292766 99.074 -73.481 48.467
292767 99.114 -73.512 48.421
292768 99.153 -73.544 48.375
292769 99.182 -73.533 48.327
292770 99.210 -73.506 48.279
292771 99.237 -73.477 48.231
292772 99.266 -73.449 48.183
292773 99.293 -73.422 48.137
292774 99.320 -73.435 48.090
292775 99.344 -73.474 48.044
292776 99.369 -73.515 47.999
292777 99.394 -73.554 47.954
292778 99.431 -73.578 47.909
292779 99.472 -73.593 47.865
292780 99.514 -73.609 47.820
292781 99.556 -73.626 47.775
292782 99.597 -73.640 47.731
292783 99.637 -73.646 47.687
292784 99.676 -73.651 47.643
292785 99.715 -73.657 47.600
292786 99.755 -73.662 47.557
292787 99.794 -73.668 47.513
292788 99.834 -73.674 47.470
292789 99.873 -73.679 47.427
292790 99.912 -73.684 47.384
292791 99.951 -73.690 47.341
292792 99.991 -73.696 47.299
292793 100.031 -73.701 47.257
292794 100.070 -73.706 47.216
292795 100.101 -73.718 47.176
292796 100.118 -73.738 47.137
292797 100.136 -73.757 47.098
292798 100.153 -73.777 47.062
292799 100.171 -73.797 47.027
292800 100.189 -73.816 46.992
292801 100.220 -73.826 46.959
292802 100.253 -73.835 46.925
292803 100.286 -73.844 46.893
292804 100.319 -73.852 46.861
292805 100.352 -73.861 46.834
292806 100.385 -73.869 46.803
292807 100.418 -73.878 46.771
292808 100.451 -73.887 46.739
292809 100.476 -73.885 46.706
292810 100.495 -73.875 46.672
292811 100.514 -73.865 46.639
292812 100.539 -73.862 46.606
292813 100.567 -73.863 46.572
292814 100.596 -73.865 46.539
292815 100.625 -73.866 46.507
292816 100.653 -73.869 46.475
292817 100.682 -73.870 46.444
292818 100.710 -73.872 46.416
292819 100.739 -73.874 46.388
292820 100.768 -73.875 46.358
292821 100.796 -73.877 46.325
292822 100.825 -73.879 46.290
292823 100.853 -73.881 46.255
292824 100.901 -73.931 46.220
292825 100.939 -73.946 46.186
292826 100.977 -73.961 46.152
292827 101.008 -73.976 46.117
292828 101.037 -73.991 46.081
292829 101.066 -74.007 46.045
292830 101.095 -74.022 46.008
292831 101.124 -74.037 45.974
292832 101.153 -74.052 45.941
292833 101.183 -74.067 45.908
292834 101.211 -74.082 45.873
292835 101.241 -74.098 45.839
292836 101.270 -74.113 45.805
292837 101.299 -74.128 45.771
292838 101.328 -74.143 45.738
292839 101.357 -74.158 45.707
292840 101.387 -74.174 45.675
292841 101.417 -74.192 45.642
292842 101.449 -74.211 45.610
292843 101.480 -74.230 45.579
292844 101.510 -74.249 45.548
292845 101.549 -74.267 45.517
292846 101.593 -74.281 45.487
292847 101.638 -74.295 45.457
292848 101.682 -74.311 45.427
292849 101.727 -74.325 45.395
292850 101.771 -74.340 45.365
292851 101.816 -74.355 45.336
292852 101.860 -74.370 45.307
292853 101.904 -74.384 45.277
292854 101.949 -74.399 45.247
292855 101.993 -74.414 45.217
292856 102.038 -74.429 45.187
292857 102.083 -74.449 45.157
292858 102.079 -74.407 45.128
292859 102.088 -74.392 45.098
292860 102.105 -74.392 45.068
292861 102.122 -74.393 45.038
292862 102.144 -74.393 45.008
292863 102.167 -74.394 44.978
292864 102.190 -74.396 44.949
292865 102.213 -74.396 44.920
292866 102.236 -74.397 44.891
292867 102.259 -74.399 44.861
292868 102.282 -74.399 44.831
293340 95.140 -72.599 52.569
293341 95.166 -72.606 52.550
293342 95.192 -72.614 52.530
293343 95.200 -72.613 52.510
293344 95.220 -72.618 52.490
293345 95.241 -72.622 52.469
293346 95.262 -72.626 52.448
293347 95.282 -72.631 52.426
293348 95.303 -72.635 52.403
293349 95.323 -72.640 52.379
293350 95.344 -72.644 52.354
293351 95.365 -72.648 52.328
293352 95.404 -72.660 52.303
293353 95.462 -72.678 52.280
293354 95.519 -72.695 52.256
293355 95.576 -72.713 52.233
293356 95.610 -72.698 52.210
293357 95.644 -72.688 52.188
293358 95.679 -72.678 52.167
293359 95.714 -72.669 52.147
293360 95.749 -72.660 52.129
293361 95.784 -72.650 52.111
293362 95.817 -72.648 52.093
293363 95.838 -72.668 52.075
293364 95.850 -72.681 52.057
293365 95.869 -72.694 52.039
293366 95.888 -72.707 52.021
293367 95.908 -72.720 52.005
293368 95.927 -72.734 51.990
293369 95.946 -72.747 51.975
293370 95.966 -72.760 51.961
293371 95.985 -72.773 51.948
293372 96.004 -72.786 51.935
293373 96.023 -72.799 51.924
293374 96.043 -72.813 51.912
293375 96.062 -72.826 51.901
293376 96.081 -72.838 51.890
293377 96.100 -72.852 51.880
293378 96.131 -72.854 51.871
293379 96.168 -72.848 51.864
293380 96.205 -72.841 51.857
293381 96.242 -72.835 51.852
293382 96.279 -72.829 51.847
293383 96.316 -72.823 51.839
293384 96.353 -72.817 51.827
293385 96.390 -72.811 51.811
293386 96.424 -72.809 51.792
293387 96.452 -72.816 51.769
293388 96.480 -72.824 51.744
293389 96.508 -72.831 51.717
293390 96.536 -72.838 51.690
293391 96.564 -72.846 51.663
293392 96.592 -72.853 51.636
293393 96.620 -72.860 51.610
293394 96.648 -72.868 51.585
293395 96.690 -72.870 51.560
293396 96.733 -72.870 51.532
293397 96.766 -72.869 51.501
293398 96.797 -72.865 51.471
293399 96.827 -72.862 51.445
293400 96.860 -72.865 51.423
293401 96.893 -72.869 51.401
293402 96.926 -72.872 51.380
293403 96.958 -72.875 51.354
293404 96.991 -72.878 51.326
293405 97.024 -72.882 51.295
293406 97.064 -72.884 51.263
293407 97.109 -72.882 51.227
293408 97.136 -72.871 51.194
293409 97.172 -72.872 51.162
293410 97.208 -72.874 51.130
293411 97.243 -72.876 51.101
293412 97.276 -72.881 51.071
293413 97.294 -72.901 51.041
293414 97.317 -72.904 51.013
293415 97.349 -72.883 50.989
293416 97.380 -72.863 50.964
293417 97.412 -72.842 50.934
293418 97.443 -72.822 50.899
293419 97.475 -72.801 50.860
293420 97.506 -72.781 50.813
293421 97.538 -72.763 50.765
293422 97.576 -72.771 50.716
293423 97.614 -72.779 50.668
293424 97.652 -72.788 50.618
293425 97.689 -72.796 50.569
293426 97.727 -72.804 50.518
293427 97.761 -72.810 50.466
293428 97.791 -72.813 50.411
293429 97.830 -72.826 50.356
293430 97.872 -72.843 50.301
293431 97.914 -72.859 50.246
293432 97.956 -72.875 50.192
293433 97.997 -72.891 50.136
293434 98.036 -72.907 50.082
293435 98.075 -72.923 50.031
293436 98.120 -72.900 49.982
293437 98.170 -72.896 49.936
293438 98.213 -72.899 49.891
293439 98.256 -72.901 49.846
293440 98.299 -72.904 49.799
293441 98.342 -72.907 49.749
293442 98.385 -72.910 49.698
293443 98.419 -72.932 49.646
293444 98.444 -72.970 49.595
293445 98.470 -73.010 49.544
293446 98.496 -73.049 49.493
293447 98.522 -73.075 49.440
293448 98.550 -73.098 49.387
293449 98.562 -73.122 49.335
293450 98.571 -73.145 49.285
293451 98.581 -73.188 49.234
293452 98.592 -73.245 49.185
293453 98.600 -73.239 49.136
293454 98.608 -73.130 49.087
293455 98.653 -73.096 49.038
293456 98.666 -73.125 48.990
293457 98.713 -73.156 48.943
293458 98.761 -73.187 48.896
293459 98.807 -73.218 48.848
293460 98.855 -73.250 48.799
293461 98.867 -73.278 48.750
293462 98.893 -73.308 48.701
293463 98.933 -73.340 48.652
293464 98.973 -73.371 48.604
293465 99.012 -73.402 48.556
293466 99.052 -73.434 48.509
293467 99.091 -73.465 48.462
293468 99.120 -73.487 48.416
293469 99.157 -73.495 48.370
293470 99.193 -73.517 48.322
293471 99.217 -73.524 48.273
293472 99.243 -73.515 48.225
293473 99.271 -73.487 48.178
293474 99.299 -73.459 48.132
293475 99.327 -73.431 48.086
293476 99.354 -73.419 48.040
293477 99.379 -73.459 47.995
293478 99.404 -73.499 47.949
293479 99.429 -73.538 47.905
293480 99.469 -73.557 47.861
293481 99.510 -73.573 47.815
293482 99.552 -73.589 47.771
293483 99.593 -73.606 47.726
293484 99.633 -73.611 47.682
293485 99.673 -73.616 47.638
293486 99.712 -73.622 47.595
293487 99.751 -73.628 47.551
293488 99.791 -73.633 47.508
293489 99.830 -73.638 47.464
293490 99.869 -73.644 47.420
293491 99.909 -73.650 47.376
293492 99.948 -73.655 47.332
293493 99.988 -73.660 47.291
293494 100.027 -73.666 47.248
293495 100.061 -73.676 47.207
293496 100.098 -73.684 47.167
293497 100.135 -73.694 47.127
293498 100.170 -73.703 47.088
293499 100.187 -73.723 47.051
293500 100.205 -73.743 47.016
293501 100.223 -73.763 46.981
293502 100.241 -73.782 46.947
293503 100.268 -73.795 46.913
293504 100.300 -73.803 46.881
293505 100.334 -73.812 46.849
293506 100.366 -73.821 46.822
293507 100.400 -73.829 46.791
293508 100.432 -73.838 46.759
293509 100.466 -73.847 46.726
293510 100.486 -73.838 46.692
293511 100.505 -73.828 46.659
293512 100.525 -73.819 46.626
293513 100.544 -73.809 46.592
293514 100.567 -73.803 46.559
293515 100.596 -73.805 46.526
293516 100.624 -73.807 46.495
293517 100.652 -73.808 46.464
293518 100.681 -73.810 46.435
293519 100.710 -73.812 46.408
293520 100.739 -73.813 46.378
293521 100.767 -73.816 46.345
293522 100.795 -73.817 46.312
293523 100.824 -73.819 46.277
293524 100.853 -73.821 46.242
293525 100.909 -73.895 46.207
293526 100.947 -73.910 46.173
293527 100.985 -73.926 46.139
293528 101.023 -73.941 46.103
293529 101.056 -73.956 46.067
293530 101.086 -73.971 46.031
293531 101.115 -73.986 45.995
293532 101.143 -74.001 45.961
293533 101.173 -74.017 45.928
293534 101.202 -74.032 45.895
293535 101.231 -74.047 45.861
293536 101.260 -74.062 45.827
293537 101.289 -74.077 45.793
293538 101.318 -74.092 45.758
293539 101.347 -74.107 45.726
293540 101.376 -74.123 45.694
293541 101.407 -74.139 45.663
293542 101.438 -74.158 45.630
293543 101.469 -74.177 45.598
293544 101.500 -74.196 45.566
293545 101.531 -74.215 45.535
293546 101.562 -74.234 45.504
293547 101.604 -74.250 45.474
293548 101.648 -74.264 45.445
293549 101.693 -74.280 45.414
293550 101.737 -74.294 45.383
293551 101.781 -74.309 45.353
293552 101.826 -74.324 45.324
293553 101.871 -74.339 45.294
293554 101.915 -74.353 45.265
293555 101.959 -74.368 45.235
293556 102.004 -74.383 45.205
293557 102.048 -74.398 45.176
293558 102.091 -74.404 45.145
293559 102.087 -74.355 45.116
293560 102.102 -74.352 45.087
293561 102.120 -74.352 45.057
293562 102.141 -74.353 45.027
293563 102.165 -74.354 44.996
293564 102.187 -74.355 44.966
293565 102.210 -74.356 44.937
293566 102.234 -74.357 44.907
293567 102.256 -74.358 44.879
293568 102.279 -74.359 44.849
293569 102.303 -74.360 44.819
294042 95.165 -72.569 52.546
294043 95.191 -72.577 52.527
294044 95.210 -72.581 52.507
294045 95.224 -72.583 52.487
294046 95.244 -72.587 52.466
294047 95.265 -72.591 52.445
294048 95.285 -72.596 52.423
294049 95.306 -72.600 52.400
294050 95.326 -72.605 52.376
294051 95.347 -72.609 52.350
294052 95.368 -72.613 52.325
294053 95.409 -72.625 52.300
294054 95.466 -72.643 52.276
294055 95.523 -72.660 52.252
294056 95.578 -72.674 52.229
294057 95.606 -72.652 52.207
294058 95.640 -72.642 52.185
294059 95.676 -72.633 52.164
294060 95.710 -72.624 52.144
294061 95.746 -72.614 52.125
294062 95.780 -72.605 52.107
294063 95.810 -72.619 52.089
294064 95.839 -72.644 52.071
294065 95.847 -72.657 52.053
294066 95.855 -72.669 52.034
294067 95.872 -72.683 52.016
294068 95.891 -72.696 51.999
294069 95.910 -72.709 51.983
294070 95.930 -72.722 51.968
294071 95.949 -72.735 51.954
294072 95.968 -72.748 51.940
294073 95.987 -72.762 51.928
294074 96.007 -72.775 51.917
294075 96.026 -72.788 51.906
294076 96.045 -72.801 51.896
294077 96.064 -72.814 51.887
294078 96.097 -72.813 51.879
294079 96.131 -72.815 51.871
294080 96.165 -72.817 51.865
294081 96.201 -72.813 51.860
294082 96.238 -72.807 51.855
294083 96.275 -72.801 51.851
294084 96.312 -72.794 51.843
294085 96.349 -72.788 51.831
294086 96.386 -72.782 51.816
294087 96.422 -72.776 51.799
294088 96.450 -72.784 51.777
294089 96.478 -72.791 51.753
294090 96.506 -72.798 51.727
294091 96.534 -72.806 51.702
294092 96.562 -72.813 51.676
294093 96.590 -72.820 51.651
294094 96.618 -72.828 51.627
294095 96.661 -72.830 51.602
294096 96.712 -72.829 51.577
294097 96.745 -72.829 51.548
294098 96.778 -72.827 51.517
294099 96.809 -72.824 51.487
294100 96.839 -72.819 51.461
294101 96.871 -72.819 51.440
294102 96.903 -72.822 51.419
294103 96.936 -72.826 51.398
294104 96.968 -72.829 51.373
294105 97.001 -72.832 51.345
294106 97.034 -72.835 51.315
294107 97.071 -72.838 51.282
294108 97.127 -72.841 51.246
294109 97.162 -72.833 51.213
294110 97.193 -72.829 51.181
294111 97.229 -72.830 51.149
294112 97.264 -72.832 51.119
294113 97.298 -72.836 51.087
294114 97.315 -72.857 51.055
294115 97.333 -72.876 51.023
294116 97.365 -72.856 50.996
294117 97.396 -72.835 50.968
294118 97.428 -72.816 50.934
294119 97.459 -72.795 50.897
294120 97.491 -72.775 50.856
294121 97.522 -72.754 50.807
294122 97.557 -72.745 50.757
294123 97.594 -72.754 50.707
294124 97.632 -72.761 50.657
294125 97.670 -72.769 50.608
294126 97.708 -72.778 50.557
294127 97.745 -72.786 50.506
294128 97.783 -72.794 50.453
294129 97.818 -72.801 50.400
294130 97.848 -72.804 50.347
294131 97.880 -72.810 50.293
294132 97.922 -72.826 50.239
294133 97.963 -72.842 50.185
294134 98.002 -72.858 50.129
294135 98.041 -72.874 50.075
294136 98.081 -72.890 50.022
294137 98.120 -72.905 49.973
294138 98.136 -72.890 49.927
294139 98.222 -72.854 49.880
294140 98.265 -72.857 49.833
294141 98.308 -72.860 49.784
294142 98.351 -72.863 49.734
294143 98.394 -72.866 49.684
294144 98.425 -72.892 49.634
294145 98.451 -72.931 49.584
294146 98.477 -72.970 49.533
294147 98.502 -73.009 49.482
294148 98.528 -73.039 49.430
294149 98.552 -73.063 49.379
294150 98.562 -73.086 49.328
294151 98.571 -73.109 49.277
294152 98.581 -73.155 49.227
294153 98.592 -73.188 49.177
294154 98.603 -73.142 49.128
294155 98.615 -73.096 49.080
294156 98.645 -73.079 49.031
294157 98.692 -73.110 48.984
294158 98.739 -73.141 48.937
294159 98.787 -73.172 48.890
294160 98.834 -73.204 48.842
294161 98.880 -73.235 48.794
294162 98.881 -73.262 48.745
294163 98.910 -73.293 48.697
294164 98.950 -73.324 48.648
294165 98.989 -73.355 48.600
294166 99.029 -73.387 48.552
294167 99.069 -73.418 48.505
294168 99.086 -73.428 48.458
294169 99.090 -73.427 48.411
294170 99.162 -73.448 48.365
294171 99.200 -73.473 48.316
294172 99.230 -73.488 48.268
294173 99.255 -73.495 48.220
294174 99.279 -73.503 48.174
294175 99.305 -73.496 48.128
294176 99.333 -73.469 48.082
294177 99.360 -73.441 48.036
294178 99.389 -73.413 47.990
294179 99.414 -73.443 47.945
294180 99.439 -73.483 47.900
294181 99.465 -73.521 47.856
294182 99.506 -73.537 47.811
294183 99.548 -73.553 47.766
294184 99.590 -73.568 47.722
294185 99.630 -73.576 47.677
294186 99.669 -73.581 47.633
294187 99.708 -73.587 47.590
294188 99.748 -73.593 47.546
294189 99.787 -73.598 47.502
294190 99.827 -73.603 47.458
294191 99.866 -73.609 47.414
294192 99.906 -73.615 47.369
294193 99.945 -73.620 47.325
294194 99.981 -73.628 47.283
294195 100.015 -73.638 47.240
294196 100.050 -73.648 47.199
294197 100.087 -73.656 47.158
294198 100.124 -73.666 47.117
294199 100.161 -73.674 47.078
294200 100.198 -73.682 47.041
294201 100.235 -73.691 47.005
294202 100.257 -73.709 46.969
294203 100.274 -73.728 46.935
294204 100.292 -73.748 46.901
294205 100.316 -73.763 46.868
294206 100.349 -73.772 46.836
294207 100.382 -73.781 46.809
294208 100.415 -73.789 46.778
294209 100.447 -73.798 46.746
294210 100.477 -73.802 46.712
294211 100.496 -73.792 46.679
294212 100.515 -73.781 46.645
294213 100.535 -73.772 46.611
294214 100.554 -73.762 46.579
294215 100.573 -73.752 46.546
294216 100.595 -73.745 46.514
294217 100.623 -73.747 46.483
294218 100.652 -73.748 46.452
294219 100.680 -73.750 46.424
294220 100.709 -73.752 46.395
294221 100.738 -73.753 46.365
294222 100.766 -73.755 46.331
294223 100.795 -73.757 46.297
294224 100.824 -73.759 46.263
294225 100.852 -73.760 46.228
294226 100.917 -73.859 46.194
294227 100.955 -73.874 46.160
294228 100.993 -73.889 46.125
294229 101.031 -73.904 46.090
294230 101.069 -73.919 46.055
294231 101.105 -73.935 46.019
294232 101.134 -73.950 45.983
294233 101.163 -73.965 45.949
294234 101.192 -73.980 45.916
294235 101.221 -73.995 45.883
294236 101.250 -74.010 45.849
294237 101.279 -74.026 45.815
294238 101.309 -74.041 45.780
294239 101.337 -74.056 45.746
294240 101.367 -74.071 45.713
294241 101.396 -74.086 45.682
294242 101.427 -74.105 45.650
294243 101.458 -74.124 45.617
294244 101.489 -74.143 45.585
294245 101.520 -74.162 45.554
294246 101.551 -74.181 45.523
294247 101.582 -74.200 45.491
294248 101.614 -74.219 45.461
294249 101.659 -74.233 45.432
294250 101.703 -74.248 45.402
294251 101.748 -74.263 45.371
294252 101.792 -74.278 45.341
294253 101.836 -74.292 45.311
294254 101.881 -74.307 45.282
294255 101.925 -74.322 45.253
294256 101.970 -74.336 45.223
294257 102.014 -74.352 45.194
294258 102.059 -74.366 45.164
294259 102.094 -74.350 45.133
294260 102.100 -74.311 45.105
294261 102.117 -74.312 45.075
294262 102.139 -74.313 45.045
294263 102.162 -74.314 45.015
294264 102.185 -74.314 44.983
294265 102.208 -74.316 44.953
294266 102.231 -74.317 44.924
294267 102.254 -74.317 44.895
294268 102.277 -74.319 44.867
294269 102.300 -74.320 44.836
294270 102.323 -74.320 44.807
294744 95.190 -72.540 52.525
294745 95.215 -72.547 52.506
294746 95.228 -72.548 52.486
294747 95.247 -72.553 52.465
294748 95.268 -72.556 52.443
294749 95.288 -72.561 52.420
294750 95.309 -72.565 52.397
294751 95.329 -72.570 52.373
294752 95.350 -72.575 52.347
294753 95.371 -72.578 52.322
294754 95.413 -72.591 52.296
294755 95.471 -72.609 52.272
294756 95.528 -72.626 52.248
294757 95.573 -72.627 52.225
294758 95.602 -72.606 52.204
294759 95.637 -72.597 52.182
294760 95.672 -72.587 52.162
294761 95.707 -72.578 52.142
294762 95.742 -72.569 52.122
294763 95.775 -72.566 52.103
294764 95.803 -72.591 52.085
294765 95.832 -72.615 52.067
294766 95.850 -72.634 52.049
294767 95.857 -72.646 52.030
294768 95.865 -72.659 52.012
294769 95.875 -72.672 51.995
294770 95.894 -72.685 51.979
294771 95.913 -72.698 51.963
294772 95.932 -72.711 51.949
294773 95.952 -72.724 51.935
294774 95.971 -72.738 51.923
294775 95.990 -72.750 51.912
294776 96.010 -72.763 51.902
294777 96.029 -72.777 51.892
294778 96.063 -72.774 51.885
294779 96.098 -72.772 51.878
294780 96.131 -72.773 51.871
294781 96.165 -72.776 51.866
294782 96.199 -72.778 51.861
294783 96.234 -72.777 51.856
294784 96.271 -72.771 51.852
294785 96.308 -72.765 51.845
294786 96.345 -72.759 51.833
294787 96.382 -72.753 51.819
294788 96.419 -72.747 51.803
294789 96.449 -72.751 51.783
294790 96.476 -72.758 51.759
294791 96.505 -72.766 51.735
294792 96.532 -72.773 51.711
294793 96.560 -72.780 51.687
294794 96.588 -72.788 51.665
294795 96.631 -72.790 51.642
294796 96.682 -72.789 51.617
294797 96.724 -72.788 51.591
294798 96.757 -72.788 51.561
294799 96.790 -72.786 51.530
294800 96.821 -72.783 51.500
294801 96.851 -72.779 51.476
294802 96.881 -72.774 51.455
294803 96.913 -72.776 51.434
294804 96.946 -72.779 51.412
294805 96.979 -72.782 51.388
294806 97.012 -72.785 51.360
294807 97.044 -72.788 51.330
294808 97.078 -72.791 51.298
294809 97.135 -72.794 51.263
294810 97.187 -72.795 51.230
294811 97.214 -72.785 51.198
294812 97.250 -72.786 51.166
294813 97.286 -72.788 51.135
294814 97.319 -72.792 51.102
294815 97.337 -72.813 51.068
294816 97.355 -72.833 51.034
294817 97.381 -72.829 51.004
294818 97.413 -72.809 50.972
294819 97.444 -72.788 50.935
294820 97.475 -72.768 50.896
294821 97.507 -72.747 50.854
294822 97.538 -72.727 50.803
294823 97.575 -72.727 50.751
294824 97.612 -72.735 50.700
294825 97.650 -72.744 50.648
294826 97.688 -72.752 50.598
294827 97.726 -72.760 50.547
294828 97.763 -72.768 50.495
294829 97.801 -72.776 50.441
294830 97.839 -72.785 50.388
294831 97.875 -72.791 50.336
294832 97.905 -72.795 50.284
294833 97.935 -72.800 50.232
294834 97.968 -72.810 50.179
294835 98.008 -72.826 50.124
294836 98.047 -72.841 50.069
294837 98.086 -72.857 50.016
294838 98.126 -72.873 49.967
294839 98.162 -72.886 49.918
294840 98.174 -72.873 49.870
294841 98.245 -72.838 49.821
294842 98.316 -72.816 49.770
294843 98.359 -72.818 49.721
294844 98.403 -72.821 49.672
294845 98.432 -72.852 49.622
294846 98.457 -72.891 49.574
294847 98.483 -72.931 49.524
294848 98.508 -72.970 49.473
294849 98.534 -73.004 49.423
294850 98.552 -73.027 49.374
294851 98.561 -73.051 49.323
294852 98.570 -73.073 49.272
294853 98.581 -73.086 49.222
294854 98.594 -73.056 49.172
294855 98.610 -73.046 49.122
294856 98.627 -73.036 49.073
294857 98.658 -73.067 49.026
294858 98.718 -73.095 48.978
294859 98.766 -73.126 48.931
294860 98.812 -73.157 48.884
294861 98.860 -73.189 48.836
294862 98.895 -73.218 48.788
294863 98.896 -73.246 48.740
294864 98.927 -73.277 48.692
294865 98.967 -73.308 48.645
294866 99.006 -73.340 48.597
294867 99.046 -73.371 48.549
294868 99.052 -73.371 48.502
294869 99.056 -73.370 48.455
294870 99.072 -73.370 48.407
294871 99.167 -73.402 48.360
294872 99.205 -73.426 48.312
294873 99.242 -73.451 48.263
294874 99.268 -73.459 48.215
294875 99.292 -73.466 48.171
294876 99.316 -73.473 48.124
294877 99.341 -73.481 48.078
294878 99.366 -73.478 48.032
294879 99.394 -73.451 47.986
294880 99.422 -73.423 47.941
294881 99.448 -73.427 47.896
294882 99.474 -73.467 47.852
294883 99.503 -73.500 47.808
294884 99.545 -73.516 47.763
294885 99.586 -73.532 47.718
294886 99.626 -73.541 47.673
294887 99.665 -73.546 47.629
294888 99.705 -73.552 47.586
294889 99.745 -73.557 47.541
294890 99.784 -73.563 47.497
294891 99.823 -73.568 47.452
294892 99.863 -73.574 47.408
294893 99.900 -73.581 47.364
294894 99.934 -73.591 47.320
294895 99.969 -73.601 47.276
294896 100.003 -73.611 47.233
294897 100.039 -73.620 47.191
294898 100.076 -73.628 47.149
294899 100.113 -73.637 47.108
294900 100.150 -73.646 47.068
294901 100.187 -73.654 47.031
294902 100.224 -73.663 46.995
294903 100.261 -73.672 46.960
294904 100.298 -73.680 46.925
294905 100.326 -73.694 46.891
294906 100.344 -73.714 46.858
294907 100.363 -73.732 46.826
294908 100.397 -73.741 46.797
294909 100.429 -73.750 46.767
294910 100.463 -73.758 46.734
294911 100.487 -73.755 46.700
294912 100.506 -73.745 46.666
294913 100.525 -73.735 46.633
294914 100.545 -73.725 46.600
294915 100.564 -73.716 46.567
294916 100.583 -73.706 46.534
294917 100.603 -73.696 46.502
294918 100.622 -73.687 46.471
294919 100.651 -73.688 46.440
294920 100.680 -73.690 46.412
294921 100.708 -73.692 46.382
294922 100.737 -73.694 46.351
294923 100.765 -73.695 46.318
294924 100.794 -73.697 46.284
294925 100.823 -73.699 46.250
294926 100.851 -73.700 46.215
294927 100.925 -73.822 46.180
294928 100.964 -73.838 46.147
294929 101.001 -73.853 46.112
294930 101.039 -73.869 46.077
294931 101.077 -73.884 46.042
294932 101.115 -73.899 46.007
294933 101.153 -73.914 45.972
294934 101.182 -73.929 45.937
294935 101.211 -73.944 45.904
294936 101.241 -73.960 45.871
294937 101.269 -73.975 45.837
294938 101.299 -73.990 45.803
294939 101.328 -74.005 45.768
294940 101.357 -74.020 45.733
294941 101.386 -74.035 45.701
294942 101.416 -74.052 45.669
294943 101.447 -74.071 45.637
294944 101.478 -74.090 45.605
294945 101.509 -74.109 45.573
294946 101.540 -74.128 45.541
294947 101.571 -74.147 45.510
294948 101.602 -74.166 45.478
294949 101.633 -74.185 45.448
294950 101.669 -74.202 45.419
294951 101.713 -74.217 45.388
294952 101.758 -74.232 45.358
294953 101.802 -74.246 45.329
294954 101.847 -74.261 45.299
294955 101.891 -74.276 45.270
294956 101.936 -74.291 45.240
294957 101.980 -74.305 45.212
294958 102.025 -74.320 45.182
294959 102.066 -74.323 45.152
294960 102.098 -74.295 45.121
294961 102.115 -74.272 45.092
294962 102.136 -74.273 45.062
294963 102.160 -74.273 45.032
294964 102.182 -74.274 45.002
294965 102.205 -74.276 44.970
294966 102.229 -74.277 44.941
294967 102.251 -74.277 44.911
294968 102.274 -74.279 44.882
294969 102.298 -74.280 44.854
294970 102.321 -74.280 44.824
294971 102.343 -74.281 44.794
295446 95.214 -72.509 52.504
295447 95.239 -72.516 52.485
295448 95.250 -72.518 52.464
295449 95.271 -72.522 52.442
295450 95.291 -72.526 52.419
295451 95.312 -72.531 52.395
295452 95.332 -72.535 52.371
295453 95.353 -72.540 52.345
295454 95.374 -72.544 52.319
295455 95.418 -72.556 52.294
295456 95.475 -72.574 52.268
295457 95.532 -72.592 52.245
295458 95.568 -72.581 52.222
295459 95.598 -72.561 52.200
295460 95.633 -72.552 52.179
295461 95.668 -72.542 52.159
295462 95.703 -72.533 52.139
295463 95.738 -72.524 52.120
295464 95.767 -72.537 52.102
295465 95.795 -72.562 52.084
295466 95.824 -72.586 52.065
295467 95.852 -72.610 52.046
295468 95.860 -72.622 52.028
295469 95.867 -72.635 52.010
295470 95.874 -72.648 51.992
295471 95.881 -72.660 51.976
295472 95.896 -72.673 51.960
295473 95.915 -72.687 51.945
295474 95.935 -72.700 51.931
295475 95.954 -72.713 51.919
295476 95.973 -72.726 51.908
295477 95.994 -72.738 51.898
295478 96.029 -72.735 51.890
295479 96.063 -72.732 51.883
295480 96.098 -72.730 51.877
295481 96.132 -72.732 51.871
295482 96.166 -72.735 51.865
295483 96.199 -72.737 51.860
295484 96.233 -72.739 51.855
295485 96.267 -72.741 51.849
295486 96.304 -72.735 51.840
295487 96.340 -72.729 51.829
295488 96.378 -72.723 51.815
295489 96.414 -72.717 51.801
295490 96.447 -72.719 51.782
295491 96.474 -72.725 51.760
295492 96.503 -72.733 51.737
295493 96.531 -72.741 51.716
295494 96.558 -72.747 51.694
295495 96.602 -72.750 51.674
295496 96.653 -72.749 51.652
295497 96.702 -72.748 51.627
295498 96.736 -72.747 51.599
295499 96.768 -72.746 51.569
295500 96.802 -72.745 51.539
295501 96.833 -72.742 51.510
295502 96.863 -72.738 51.487
295503 96.893 -72.734 51.465
295504 96.924 -72.729 51.443
295505 96.956 -72.732 51.419
295506 96.989 -72.735 51.394
295507 97.022 -72.738 51.367
295508 97.054 -72.742 51.339
295509 97.087 -72.745 51.309
295510 97.141 -72.747 51.276
295511 97.197 -72.750 51.244
295512 97.239 -72.746 51.212
295513 97.271 -72.742 51.180
295514 97.307 -72.744 51.146
295515 97.342 -72.747 51.113
295516 97.359 -72.768 51.077
295517 97.377 -72.788 51.042
295518 97.397 -72.802 51.011
295519 97.429 -72.782 50.976
295520 97.460 -72.761 50.938
295521 97.492 -72.741 50.897
295522 97.523 -72.720 50.852
295523 97.555 -72.701 50.802
295524 97.593 -72.710 50.749
295525 97.631 -72.717 50.696
295526 97.668 -72.725 50.642
295527 97.706 -72.734 50.592
295528 97.744 -72.742 50.539
295529 97.781 -72.750 50.485
295530 97.819 -72.758 50.429
295531 97.857 -72.766 50.377
295532 97.894 -72.775 50.324
295533 97.932 -72.782 50.273
295534 97.962 -72.788 50.224
295535 97.994 -72.797 50.174
295536 98.025 -72.804 50.120
295537 98.057 -72.813 50.066
295538 98.092 -72.824 50.013
295539 98.131 -72.840 49.960
295540 98.170 -72.856 49.910
295541 98.205 -72.867 49.861
295542 98.217 -72.854 49.813
295543 98.261 -72.829 49.762
295544 98.354 -72.787 49.712
295545 98.411 -72.776 49.662
295546 98.438 -72.813 49.613
295547 98.463 -72.852 49.565
295548 98.489 -72.891 49.517
295549 98.515 -72.930 49.468
295550 98.541 -72.968 49.420
295551 98.552 -72.992 49.371
295552 98.565 -73.017 49.320
295553 98.579 -73.041 49.269
295554 98.595 -73.039 49.217
295555 98.611 -73.029 49.167
295556 98.628 -73.019 49.117
295557 98.645 -73.012 49.069
295558 98.675 -73.038 49.021
295559 98.713 -73.090 48.974
295560 98.792 -73.111 48.927
295561 98.839 -73.142 48.879
295562 98.886 -73.174 48.832
295563 98.909 -73.202 48.784
295564 98.910 -73.230 48.736
295565 98.944 -73.261 48.688
295566 98.984 -73.292 48.641
295567 99.013 -73.314 48.594
295568 99.018 -73.313 48.545
295569 99.022 -73.311 48.499
295570 99.027 -73.311 48.452
295571 99.135 -73.330 48.404
295572 99.173 -73.355 48.355
295573 99.210 -73.380 48.307
295574 99.248 -73.404 48.259
295575 99.281 -73.423 48.211
295576 99.305 -73.430 48.166
295577 99.330 -73.437 48.120
295578 99.354 -73.444 48.074
295579 99.378 -73.452 48.029
295580 99.403 -73.459 47.984
295581 99.428 -73.460 47.938
295582 99.455 -73.432 47.893
295583 99.483 -73.411 47.849
295584 99.508 -73.450 47.804
295585 99.541 -73.480 47.759
295586 99.582 -73.496 47.714
295587 99.623 -73.506 47.670
295588 99.662 -73.512 47.626
295589 99.702 -73.517 47.582
295590 99.741 -73.522 47.538
295591 99.780 -73.528 47.493
295592 99.819 -73.534 47.449
295593 99.853 -73.544 47.404
295594 99.888 -73.553 47.359
295595 99.922 -73.563 47.315
295596 99.957 -73.573 47.271
295597 99.991 -73.583 47.228
295598 100.028 -73.591 47.185
295599 100.065 -73.600 47.143
295600 100.103 -73.609 47.101
295601 100.139 -73.618 47.061
295602 100.177 -73.626 47.023
295603 100.213 -73.634 46.987
295604 100.251 -73.644 46.951
295605 100.287 -73.652 46.916
295606 100.325 -73.661 46.881
295607 100.361 -73.669 46.848
295608 100.396 -73.680 46.815
295609 100.413 -73.700 46.786
295610 100.444 -73.709 46.755
295611 100.477 -73.718 46.723
295612 100.497 -73.709 46.689
295613 100.516 -73.699 46.654
295614 100.536 -73.688 46.621
295615 100.555 -73.678 46.588
295616 100.574 -73.669 46.555
295617 100.593 -73.659 46.523
295618 100.613 -73.649 46.491
295619 100.632 -73.639 46.459
295620 100.651 -73.629 46.428
295621 100.679 -73.630 46.399
295622 100.707 -73.631 46.370
295623 100.736 -73.634 46.338
295624 100.765 -73.635 46.306
295625 100.793 -73.637 46.273
295626 100.822 -73.638 46.238
295627 100.848 -73.650 46.202
295628 100.933 -73.787 46.167
295629 100.972 -73.802 46.134
295630 101.009 -73.817 46.100
295631 101.048 -73.832 46.065
295632 101.086 -73.847 46.030
295633 101.123 -73.863 45.996
295634 101.162 -73.878 45.961
295635 101.199 -73.893 45.927
295636 101.231 -73.908 45.893
295637 101.260 -73.923 45.860
295638 101.289 -73.938 45.826
295639 101.318 -73.954 45.791
295640 101.347 -73.969 45.757
295641 101.376 -73.984 45.722
295642 101.405 -73.999 45.690
295643 101.436 -74.018 45.658
295644 101.467 -74.037 45.625
295645 101.498 -74.056 45.593
295646 101.529 -74.075 45.560
295647 101.560 -74.094 45.529
295648 101.592 -74.113 45.498
295649 101.622 -74.132 45.467
295650 101.653 -74.151 45.436
295651 101.685 -74.170 45.406
295652 101.724 -74.186 45.376
295653 101.768 -74.201 45.345
295654 101.813 -74.215 45.316
295655 101.857 -74.230 45.287
295656 101.902 -74.245 45.257
295657 101.946 -74.260 45.229
295658 101.991 -74.274 45.200
295659 102.035 -74.289 45.170
295660 102.069 -74.269 45.140
295661 102.102 -74.242 45.110
295662 102.132 -74.233 45.080
295663 102.157 -74.233 45.049
295664 102.180 -74.234 45.019
295665 102.203 -74.236 44.989
295666 102.226 -74.236 44.958
295667 102.249 -74.237 44.928
295668 102.272 -74.239 44.899
295669 102.295 -74.239 44.869
295670 102.318 -74.240 44.841
295671 102.341 -74.241 44.811
295672 102.364 -74.242 44.781
296148 95.239 -72.480 52.483
296149 95.256 -72.484 52.462
296150 95.274 -72.487 52.440
296151 95.295 -72.491 52.417
296152 95.315 -72.496 52.393
296153 95.335 -72.500 52.369
296154 95.356 -72.504 52.342
296155 95.377 -72.509 52.317
296156 95.422 -72.522 52.291
296157 95.479 -72.540 52.265
296158 95.536 -72.556 52.242
296159 95.564 -72.534 52.220
296160 95.595 -72.515 52.197
296161 95.629 -72.506 52.176
296162 95.665 -72.497 52.156
296163 95.699 -72.487 52.136
296164 95.733 -72.484 52.118
296165 95.759 -72.508 52.100
296166 95.788 -72.532 52.082
296167 95.817 -72.557 52.063
296168 95.846 -72.582 52.044
296169 95.862 -72.599 52.026
296170 95.869 -72.612 52.008
296171 95.876 -72.625 51.990
296172 95.884 -72.637 51.974
296173 95.891 -72.650 51.958
296174 95.899 -72.663 51.942
296175 95.918 -72.675 51.928
296176 95.938 -72.688 51.916
296177 95.960 -72.698 51.905
296178 95.995 -72.695 51.895
296179 96.030 -72.692 51.887
296180 96.064 -72.690 51.881
296181 96.099 -72.688 51.875
296182 96.132 -72.691 51.869
296183 96.166 -72.693 51.864
296184 96.200 -72.695 51.858
296185 96.234 -72.697 51.851
296186 96.267 -72.700 51.844
296187 96.301 -72.702 51.833
296188 96.336 -72.700 51.822
296189 96.373 -72.694 51.809
296190 96.410 -72.688 51.796
296191 96.445 -72.685 51.777
296192 96.473 -72.693 51.756
296193 96.501 -72.700 51.735
296194 96.529 -72.707 51.717
296195 96.573 -72.709 51.698
296196 96.624 -72.709 51.680
296197 96.674 -72.708 51.659
296198 96.714 -72.707 51.634
296199 96.747 -72.707 51.606
296200 96.781 -72.705 51.576
296201 96.814 -72.704 51.547
296202 96.846 -72.701 51.519
296203 96.875 -72.697 51.495
296204 96.905 -72.693 51.471
296205 96.936 -72.688 51.447
296206 96.966 -72.686 51.423
296207 96.999 -72.689 51.397
296208 97.032 -72.692 51.371
296209 97.065 -72.695 51.345
296210 97.097 -72.699 51.317
296211 97.148 -72.701 51.286
296212 97.205 -72.704 51.255
296213 97.261 -72.707 51.223
296214 97.293 -72.698 51.189
296215 97.328 -72.700 51.153
296216 97.363 -72.703 51.117
296217 97.381 -72.723 51.082
296218 97.399 -72.744 51.047
296219 97.417 -72.764 51.014
296220 97.445 -72.754 50.978
296221 97.476 -72.734 50.938
296222 97.508 -72.713 50.895
296223 97.539 -72.693 50.850
296224 97.573 -72.683 50.801
296225 97.611 -72.691 50.748
296226 97.649 -72.700 50.693
296227 97.686 -72.708 50.638
296228 97.724 -72.716 50.587
296229 97.761 -72.724 50.536
296230 97.799 -72.732 50.482
296231 97.837 -72.741 50.425
296232 97.875 -72.748 50.370
296233 97.913 -72.757 50.317
296234 97.947 -72.758 50.265
296235 97.983 -72.768 50.215
296236 98.016 -72.779 50.166
296237 98.050 -72.789 50.113
296238 98.083 -72.801 50.060
296239 98.115 -72.810 50.007
296240 98.147 -72.818 49.955
296241 98.178 -72.826 49.906
296242 98.215 -72.839 49.858
296243 98.248 -72.848 49.809
296244 98.260 -72.835 49.759
296245 98.360 -72.790 49.708
296246 98.415 -72.756 49.658
296247 98.441 -72.775 49.609
296248 98.466 -72.813 49.561
296249 98.492 -72.853 49.514
296250 98.517 -72.892 49.466
296251 98.543 -72.931 49.417
296252 98.568 -72.961 49.368
296253 98.581 -72.986 49.317
296254 98.595 -73.011 49.266
296255 98.612 -73.012 49.215
296256 98.628 -73.002 49.165
296257 98.644 -72.992 49.115
296258 98.664 -72.990 49.065
296259 98.693 -73.008 49.016
296260 98.731 -73.061 48.969
296261 98.769 -73.114 48.922
296262 98.865 -73.127 48.874
296263 98.912 -73.158 48.826
296264 98.924 -73.186 48.779
296265 98.925 -73.214 48.731
296266 98.961 -73.245 48.684
296267 98.980 -73.256 48.637
296268 98.984 -73.255 48.590
296269 98.989 -73.254 48.541
296270 98.993 -73.252 48.494
296271 98.998 -73.252 48.447
296272 99.140 -73.283 48.400
296273 99.178 -73.308 48.351
296274 99.215 -73.333 48.303
296275 99.253 -73.358 48.255
296276 99.290 -73.382 48.208
296277 99.319 -73.394 48.162
296278 99.343 -73.401 48.116
296279 99.367 -73.409 48.071
296280 99.392 -73.415 48.026
296281 99.416 -73.423 47.981
296282 99.440 -73.430 47.935
296283 99.465 -73.437 47.890
296284 99.489 -73.442 47.846
296285 99.517 -73.414 47.801
296286 99.543 -73.434 47.756
296287 99.579 -73.459 47.712
296288 99.620 -73.471 47.668
296289 99.659 -73.476 47.623
296290 99.698 -73.482 47.579
296291 99.737 -73.487 47.534
296292 99.773 -73.496 47.490
296293 99.807 -73.506 47.445
296294 99.842 -73.516 47.400
296295 99.876 -73.526 47.355
296296 99.911 -73.536 47.310
296297 99.945 -73.546 47.266
296298 99.981 -73.555 47.223
296299 100.017 -73.563 47.180
296300 100.055 -73.572 47.137
296301 100.091 -73.581 47.095
296302 100.129 -73.589 47.055
296303 100.165 -73.598 47.017
296304 100.203 -73.606 46.979
296305 100.240 -73.615 46.943
296306 100.277 -73.624 46.908
296307 100.314 -73.632 46.872
296308 100.351 -73.641 46.838
296309 100.388 -73.650 46.805
296310 100.424 -73.659 46.774
296311 100.462 -73.667 46.744
296312 100.488 -73.672 46.712
296313 100.507 -73.662 46.677
296314 100.526 -73.652 46.643
296315 100.546 -73.642 46.610
296316 100.565 -73.632 46.577
296317 100.584 -73.622 46.544
296318 100.603 -73.612 46.512
296319 100.623 -73.603 46.480
296320 100.642 -73.592 46.447
296321 100.661 -73.582 46.416
296322 100.680 -73.572 46.386
296323 100.707 -73.571 46.357
296324 100.736 -73.573 46.326
296325 100.764 -73.575 46.293
296326 100.792 -73.577 46.260
296327 100.821 -73.578 46.226
296328 100.857 -73.616 46.191
296329 100.941 -73.750 46.157
296330 100.980 -73.766 46.124
296331 101.018 -73.781 46.089
296332 101.056 -73.796 46.055
296333 101.094 -73.811 46.021
296334 101.131 -73.826 45.987
296335 101.170 -73.842 45.951
296336 101.207 -73.857 45.918
296337 101.246 -73.872 45.884
296338 101.279 -73.888 45.850
296339 101.308 -73.903 45.815
296340 101.337 -73.918 45.781
296341 101.367 -73.933 45.747
296342 101.395 -73.948 45.712
296343 101.425 -73.965 45.680
296344 101.456 -73.984 45.647
296345 101.487 -74.003 45.614
296346 101.519 -74.022 45.581
296347 101.550 -74.041 45.549
296348 101.581 -74.060 45.517
296349 101.612 -74.078 45.486
296350 101.643 -74.097 45.456
296351 101.674 -74.116 45.424
296352 101.705 -74.135 45.394
296353 101.736 -74.154 45.364
296354 101.779 -74.170 45.334
296355 101.823 -74.184 45.304
296356 101.868 -74.199 45.275
296357 101.912 -74.214 45.246
296358 101.957 -74.228 45.217
296359 102.001 -74.243 45.189
296360 102.041 -74.242 45.158
296361 102.074 -74.214 45.128
296362 102.112 -74.195 45.098
296363 102.139 -74.196 45.068
296364 102.164 -74.197 45.037
296365 102.189 -74.198 45.006
296366 102.214 -74.198 44.976
296367 102.239 -74.198 44.946
296368 102.263 -74.199 44.915
296369 102.288 -74.200 44.886
296370 102.313 -74.201 44.857
296371 102.338 -74.201 44.828
296372 102.362 -74.202 44.798
296373 102.385 -74.203 44.769
296848 95.212 -72.435 52.502
296849 95.238 -72.443 52.481
296850 95.263 -72.450 52.461
296851 95.277 -72.452 52.438
296852 95.298 -72.456 52.414
296853 95.318 -72.461 52.390
296854 95.338 -72.465 52.365
296855 95.359 -72.469 52.339
296856 95.380 -72.474 52.314
296857 95.427 -72.488 52.288
296858 95.484 -72.506 52.263
296859 95.532 -72.510 52.239
296860 95.560 -72.487 52.216
296861 95.591 -72.471 52.195
296862 95.626 -72.461 52.174
296863 95.661 -72.452 52.154
296864 95.696 -72.443 52.134
296865 95.725 -72.455 52.116
296866 95.751 -72.479 52.097
296867 95.780 -72.503 52.079
296868 95.810 -72.528 52.061
296869 95.839 -72.553 52.043
296870 95.864 -72.575 52.025
296871 95.871 -72.588 52.007
296872 95.879 -72.601 51.990
296873 95.886 -72.613 51.973
296874 95.893 -72.626 51.956
296875 95.900 -72.639 51.940
296876 95.908 -72.651 51.926
296877 95.926 -72.659 51.913
296878 95.961 -72.656 51.902
296879 95.995 -72.653 51.892
296880 96.030 -72.650 51.885
296881 96.065 -72.647 51.878
296882 96.099 -72.647 51.872
296883 96.133 -72.650 51.867
296884 96.167 -72.652 51.862
296885 96.200 -72.654 51.855
296886 96.234 -72.656 51.846
296887 96.268 -72.659 51.837
296888 96.302 -72.661 51.825
296889 96.335 -72.663 51.814
296890 96.369 -72.665 51.802
296891 96.406 -72.659 51.789
296892 96.443 -72.653 51.770
296893 96.471 -72.660 51.751
296894 96.499 -72.667 51.732
296895 96.544 -72.669 51.715
296896 96.595 -72.669 51.699
296897 96.645 -72.668 51.683
296898 96.693 -72.667 51.663
296899 96.726 -72.666 51.639
296900 96.759 -72.665 51.612
296901 96.792 -72.664 51.583
296902 96.825 -72.663 51.555
296903 96.858 -72.661 51.528
296904 96.887 -72.656 51.503
296905 96.918 -72.652 51.477
296906 96.948 -72.647 51.452
296907 96.978 -72.643 51.428
296908 97.009 -72.643 51.401
296909 97.042 -72.646 51.376
296910 97.075 -72.649 51.351
296911 97.107 -72.652 51.324
296912 97.155 -72.655 51.295
296913 97.211 -72.657 51.264
296914 97.268 -72.660 51.231
296915 97.317 -72.660 51.196
296916 97.350 -72.656 51.158
296917 97.385 -72.658 51.120
296918 97.403 -72.678 51.085
296919 97.421 -72.699 51.050
296920 97.439 -72.719 51.015
296921 97.461 -72.728 50.978
296922 97.493 -72.707 50.937
296923 97.524 -72.687 50.893
296924 97.556 -72.666 50.847
296925 97.591 -72.666 50.798
296926 97.629 -72.673 50.746
296927 97.667 -72.681 50.691
296928 97.705 -72.690 50.636
296929 97.742 -72.698 50.583
296930 97.780 -72.707 50.532
296931 97.817 -72.714 50.479
296932 97.855 -72.722 50.422
296933 97.889 -72.725 50.367
296934 97.923 -72.725 50.314
296935 97.957 -72.727 50.262
296936 97.998 -72.732 50.210
296937 98.032 -72.750 50.160
296938 98.065 -72.761 50.108
296939 98.098 -72.772 50.055
296940 98.132 -72.782 50.003
296941 98.165 -72.793 49.953
296942 98.198 -72.804 49.905
296943 98.231 -72.815 49.857
296944 98.265 -72.826 49.808
296945 98.305 -72.819 49.757
296946 98.360 -72.786 49.706
296947 98.414 -72.753 49.657
296948 98.445 -72.749 49.608
296949 98.467 -72.776 49.560
296950 98.493 -72.816 49.514
296951 98.518 -72.854 49.465
296952 98.544 -72.894 49.415
296953 98.584 -72.931 49.365
296954 98.597 -72.956 49.315
296955 98.610 -72.980 49.264
296956 98.629 -72.985 49.214
296957 98.645 -72.975 49.164
296958 98.661 -72.964 49.113
296959 98.684 -72.967 49.062
296960 98.710 -72.979 49.012
296961 98.748 -73.032 48.964
296962 98.786 -73.085 48.916
296963 98.824 -73.138 48.868
296964 98.937 -73.143 48.820
296965 98.938 -73.170 48.773
296966 98.939 -73.198 48.726
296967 98.946 -73.199 48.680
296968 98.950 -73.197 48.633
296969 98.955 -73.196 48.585
296970 98.961 -73.195 48.537
296971 98.968 -73.195 48.489
296972 99.051 -73.205 48.442
296973 99.145 -73.236 48.395
296974 99.183 -73.261 48.346
296975 99.221 -73.286 48.298
296976 99.258 -73.311 48.251
296977 99.296 -73.336 48.204
296978 99.332 -73.358 48.158
296979 99.356 -73.365 48.113
296980 99.380 -73.372 48.068
296981 99.405 -73.380 48.023
296982 99.429 -73.387 47.978
296983 99.453 -73.394 47.932
296984 99.478 -73.401 47.888
296985 99.502 -73.409 47.843
296986 99.526 -73.415 47.799
296987 99.551 -73.422 47.754
296988 99.578 -73.418 47.710
296989 99.616 -73.436 47.666
296990 99.655 -73.441 47.621
296991 99.692 -73.449 47.576
296992 99.726 -73.459 47.531
296993 99.761 -73.469 47.487
296994 99.795 -73.479 47.443
296995 99.830 -73.488 47.397
296996 99.864 -73.498 47.351
296997 99.899 -73.508 47.306
296998 99.933 -73.518 47.262
296999 99.970 -73.527 47.218
297000 100.007 -73.535 47.175
297001 100.044 -73.543 47.132
297002 100.081 -73.553 47.090
297003 100.118 -73.561 47.049
297004 100.155 -73.570 47.010
297005 100.192 -73.578 46.973
297006 100.229 -73.587 46.936
297007 100.266 -73.596 46.899
297008 100.303 -73.604 46.864
297009 100.342 -73.613 46.829
297010 100.382 -73.622 46.795
297011 100.422 -73.631 46.763
297012 100.462 -73.640 46.732
297013 100.491 -73.647 46.700
297014 100.512 -73.631 46.666
297015 100.534 -73.612 46.633
297016 100.556 -73.596 46.598
297017 100.575 -73.585 46.565
297018 100.594 -73.575 46.533
297019 100.613 -73.565 46.501
297020 100.633 -73.556 46.469
297021 100.652 -73.546 46.436
297022 100.671 -73.536 46.404
297023 100.690 -73.526 46.373
297024 100.710 -73.516 46.343
297025 100.735 -73.513 46.312
297026 100.763 -73.515 46.280
297027 100.792 -73.517 46.247
297028 100.821 -73.518 46.214
297029 100.867 -73.583 46.180
297030 100.950 -73.714 46.146
297031 100.988 -73.729 46.113
297032 101.025 -73.745 46.079
297033 101.064 -73.760 46.045
297034 101.102 -73.775 46.011
297035 101.140 -73.791 45.977
297036 101.178 -73.806 45.942
297037 101.215 -73.821 45.908
297038 101.254 -73.836 45.874
297039 101.291 -73.851 45.840
297040 101.327 -73.866 45.805
297041 101.356 -73.882 45.771
297042 101.386 -73.897 45.737
297043 101.415 -73.912 45.702
297044 101.446 -73.930 45.670
297045 101.477 -73.949 45.637
297046 101.507 -73.968 45.604
297047 101.539 -73.987 45.571
297048 101.570 -74.006 45.539
297049 101.601 -74.025 45.506
297050 101.632 -74.044 45.475
297051 101.663 -74.063 45.444
297052 101.694 -74.082 45.413
297053 101.725 -74.101 45.383
297054 101.756 -74.120 45.353
297055 101.790 -74.138 45.324
297056 101.834 -74.153 45.294
297057 101.878 -74.167 45.264
297058 101.922 -74.183 45.235
297059 101.967 -74.197 45.206
297060 102.012 -74.212 45.178
297061 102.045 -74.187 45.147
297062 102.083 -74.151 45.116
297063 102.122 -74.159 45.085
297064 102.147 -74.160 45.055
297065 102.172 -74.160 45.025
297066 102.197 -74.161 44.994
297067 102.222 -74.161 44.964
297068 102.246 -74.162 44.934
297069 102.271 -74.163 44.904
297070 102.296 -74.164 44.874
297071 102.321 -74.164 44.845
297072 102.345 -74.164 44.815
297073 102.370 -74.165 44.787
297074 102.395 -74.166 44.757
297549 95.211 -72.398 52.499
297550 95.237 -72.406 52.478
297551 95.262 -72.413 52.457
297552 95.285 -72.419 52.434
297553 95.301 -72.421 52.411
297554 95.321 -72.426 52.387
297555 95.342 -72.431 52.363
297556 95.362 -72.434 52.336
297557 95.383 -72.439 52.311
297558 95.431 -72.453 52.285
297559 95.488 -72.471 52.260
297560 95.527 -72.463 52.236
297561 95.555 -72.441 52.214
297562 95.587 -72.425 52.192
297563 95.622 -72.416 52.170
297564 95.657 -72.406 52.151
297565 95.691 -72.402 52.131
297566 95.717 -72.426 52.112
297567 95.744 -72.450 52.094
297568 95.773 -72.475 52.077
297569 95.802 -72.500 52.060
297570 95.831 -72.524 52.042
297571 95.861 -72.549 52.023
297572 95.874 -72.565 52.006
297573 95.881 -72.578 51.989
297574 95.888 -72.590 51.972
297575 95.895 -72.603 51.956
297576 95.903 -72.616 51.940
297577 95.910 -72.628 51.926
297578 95.935 -72.625 51.913
297579 95.969 -72.622 51.900
297580 96.004 -72.619 51.890
297581 96.038 -72.616 51.882
297582 96.074 -72.611 51.875
297583 96.112 -72.606 51.869
297584 96.144 -72.607 51.864
297585 96.175 -72.609 51.858
297586 96.206 -72.613 51.849
297587 96.238 -72.615 51.839
297588 96.269 -72.617 51.828
297589 96.302 -72.619 51.817
297590 96.335 -72.622 51.806
297591 96.369 -72.624 51.794
297592 96.403 -72.626 51.781
297593 96.439 -72.623 51.762
297594 96.469 -72.628 51.743
297595 96.515 -72.628 51.726
297596 96.566 -72.628 51.711
297597 96.616 -72.628 51.697
297598 96.667 -72.627 51.683
297599 96.705 -72.626 51.665
297600 96.738 -72.625 51.642
297601 96.771 -72.624 51.616
297602 96.804 -72.623 51.589
297603 96.837 -72.622 51.563
297604 96.870 -72.620 51.537
297605 96.900 -72.616 51.513
297606 96.930 -72.611 51.486
297607 96.960 -72.606 51.460
297608 96.990 -72.603 51.435
297609 97.020 -72.598 51.409
297610 97.052 -72.600 51.385
297611 97.085 -72.603 51.360
297612 97.117 -72.606 51.333
297613 97.161 -72.609 51.303
297614 97.218 -72.611 51.272
297615 97.275 -72.614 51.237
297616 97.331 -72.616 51.200
297617 97.371 -72.613 51.161
297618 97.406 -72.615 51.123
297619 97.425 -72.634 51.085
297620 97.443 -72.654 51.048
297621 97.461 -72.675 51.012
297622 97.478 -72.695 50.975
297623 97.509 -72.680 50.935
297624 97.540 -72.660 50.891
297625 97.572 -72.639 50.844
297626 97.609 -72.647 50.795
297627 97.647 -72.656 50.744
297628 97.685 -72.664 50.690
297629 97.723 -72.672 50.634
297630 97.760 -72.680 50.580
297631 97.798 -72.688 50.528
297632 97.832 -72.691 50.475
297633 97.866 -72.692 50.421
297634 97.900 -72.693 50.367
297635 97.934 -72.694 50.316
297636 97.967 -72.696 50.265
297637 98.008 -72.700 50.213
297638 98.052 -72.707 50.159
297639 98.081 -72.732 50.106
297640 98.114 -72.743 50.054
297641 98.147 -72.754 50.005
297642 98.180 -72.765 49.957
297643 98.214 -72.776 49.907
297644 98.247 -72.786 49.858
297645 98.280 -72.797 49.808
297646 98.311 -72.806 49.756
297647 98.359 -72.782 49.708
297648 98.414 -72.749 49.659
297649 98.451 -72.737 49.610
297650 98.471 -72.748 49.561
297651 98.493 -72.778 49.514
297652 98.519 -72.817 49.465
297653 98.545 -72.857 49.415
297654 98.587 -72.895 49.364
297655 98.613 -72.925 49.314
297656 98.627 -72.950 49.264
297657 98.646 -72.957 49.213
297658 98.662 -72.948 49.163
297659 98.678 -72.938 49.111
297660 98.703 -72.945 49.059
297661 98.728 -72.952 49.009
297662 98.765 -73.003 48.959
297663 98.803 -73.055 48.910
297664 98.842 -73.108 48.862
297665 98.878 -73.158 48.815
297666 98.913 -73.160 48.768
297667 98.947 -73.161 48.720
297668 98.982 -73.163 48.673
297669 99.003 -73.163 48.627
297670 99.009 -73.162 48.580
297671 99.029 -73.163 48.532
297672 99.092 -73.167 48.484
297673 99.126 -73.178 48.437
297674 99.151 -73.190 48.389
297675 99.188 -73.214 48.342
297676 99.226 -73.239 48.294
297677 99.263 -73.264 48.247
297678 99.301 -73.289 48.201
297679 99.338 -73.314 48.156
297680 99.369 -73.329 48.110
297681 99.394 -73.336 48.065
297682 99.418 -73.343 48.021
297683 99.442 -73.351 47.975
297684 99.467 -73.358 47.929
297685 99.491 -73.365 47.885
297686 99.515 -73.372 47.841
297687 99.540 -73.379 47.797
297688 99.564 -73.387 47.752
297689 99.588 -73.393 47.708
297690 99.614 -73.403 47.664
297691 99.645 -73.412 47.619
297692 99.680 -73.421 47.574
297693 99.714 -73.431 47.529
297694 99.749 -73.441 47.484
297695 99.783 -73.451 47.439
297696 99.818 -73.461 47.393
297697 99.852 -73.471 47.347
297698 99.887 -73.481 47.302
297699 99.922 -73.490 47.258
297700 99.959 -73.498 47.214
297701 99.996 -73.507 47.170
297702 100.033 -73.515 47.127
297703 100.070 -73.524 47.084
297704 100.107 -73.533 47.044
297705 100.147 -73.542 47.004
297706 100.187 -73.551 46.966
297707 100.227 -73.559 46.928
297708 100.267 -73.568 46.891
297709 100.307 -73.578 46.855
297710 100.347 -73.587 46.820
297711 100.387 -73.595 46.785
297712 100.426 -73.604 46.752
297713 100.466 -73.613 46.720
297714 100.495 -73.621 46.688
297715 100.511 -73.620 46.655
297716 100.533 -73.601 46.621
297717 100.555 -73.582 46.587
297718 100.577 -73.562 46.554
297719 100.600 -73.543 46.522
297720 100.622 -73.524 46.490
297721 100.642 -73.509 46.458
297722 100.662 -73.499 46.426
297723 100.681 -73.489 46.394
297724 100.700 -73.479 46.361
297725 100.719 -73.469 46.330
297726 100.739 -73.459 46.299
297727 100.762 -73.455 46.267
297728 100.791 -73.456 46.235
297729 100.820 -73.459 46.201
297730 100.876 -73.550 46.168
297731 100.958 -73.678 46.135
297732 100.996 -73.694 46.102
297733 101.034 -73.709 46.068
297734 101.072 -73.724 46.035
297735 101.110 -73.739 46.001
297736 101.148 -73.754 45.966
297737 101.186 -73.769 45.933
297738 101.223 -73.785 45.898
297739 101.262 -73.800 45.865
297740 101.299 -73.815 45.831
297741 101.338 -73.831 45.796
297742 101.376 -73.846 45.762
297743 101.405 -73.861 45.727
297744 101.434 -73.877 45.693
297745 101.466 -73.896 45.660
297746 101.497 -73.915 45.627
297747 101.528 -73.934 45.594
297748 101.559 -73.953 45.560
297749 101.590 -73.972 45.529
297750 101.621 -73.991 45.495
297751 101.652 -74.010 45.463
297752 101.683 -74.029 45.433
297753 101.714 -74.048 45.401
297754 101.745 -74.067 45.371
297755 101.776 -74.085 45.341
297756 101.808 -74.104 45.311
297757 101.844 -74.122 45.280
297758 101.889 -74.136 45.251
297759 101.933 -74.151 45.222
297760 101.977 -74.166 45.193
297761 102.016 -74.160 45.164
297762 102.053 -74.113 45.133
297763 102.096 -74.117 45.103
297764 102.130 -74.123 45.072
297765 102.155 -74.123 45.042
297766 102.180 -74.123 45.012
297767 102.205 -74.124 44.982
297768 102.230 -74.125 44.952
297769 102.254 -74.126 44.922
297770 102.279 -74.126 44.892
297771 102.304 -74.126 44.862
297772 102.328 -74.127 44.833
297773 102.353 -74.128 44.803
297774 102.378 -74.129 44.775
297775 102.403 -74.129 44.745
298249 95.183 -72.347 52.516
298250 95.209 -72.361 52.496
298251 95.235 -72.368 52.475
298252 95.261 -72.376 52.454
298253 95.287 -72.384 52.431
298254 95.304 -72.387 52.408
298255 95.324 -72.391 52.384
298256 95.345 -72.396 52.360
298257 95.365 -72.399 52.333
298258 95.386 -72.404 52.308
298259 95.436 -72.419 52.282
298260 95.493 -72.437 52.257
298261 95.523 -72.417 52.234
298262 95.551 -72.395 52.211
298263 95.583 -72.380 52.189
298264 95.618 -72.371 52.167
298265 95.653 -72.362 52.148
298266 95.683 -72.372 52.128
298267 95.709 -72.396 52.109
298268 95.736 -72.421 52.092
298269 95.765 -72.446 52.075
298270 95.795 -72.471 52.058
298271 95.824 -72.495 52.040
298272 95.853 -72.520 52.022
298273 95.876 -72.541 52.005
298274 95.883 -72.554 51.988
298275 95.890 -72.566 51.972
298276 95.898 -72.579 51.956
298277 95.905 -72.592 51.941
298278 95.912 -72.602 51.926
298279 95.944 -72.591 51.912
298280 95.978 -72.587 51.899
298281 96.013 -72.584 51.888
298282 96.048 -72.581 51.879
298283 96.086 -72.575 51.871
298284 96.123 -72.569 51.865
298285 96.161 -72.564 51.859
298286 96.192 -72.567 51.852
298287 96.223 -72.569 51.844
298288 96.254 -72.572 51.833
298289 96.286 -72.575 51.821
298290 96.317 -72.577 51.809
298291 96.348 -72.579 51.797
298292 96.380 -72.582 51.785
298293 96.411 -72.584 51.773
298294 96.442 -72.587 51.754
298295 96.485 -72.588 51.737
298296 96.536 -72.587 51.722
298297 96.587 -72.587 51.707
298298 96.638 -72.587 51.694
298299 96.683 -72.586 51.680
298300 96.717 -72.585 51.662
298301 96.750 -72.584 51.641
298302 96.783 -72.583 51.616
298303 96.816 -72.582 51.592
298304 96.849 -72.581 51.568
298305 96.882 -72.579 51.544
298306 96.912 -72.575 51.520
298307 96.942 -72.570 51.494
298308 96.972 -72.566 51.468
298309 97.002 -72.562 51.442
298310 97.032 -72.557 51.417
298311 97.063 -72.553 51.393
298312 97.095 -72.556 51.368
298313 97.128 -72.559 51.341
298314 97.168 -72.562 51.310
298315 97.225 -72.565 51.278
298316 97.281 -72.568 51.243
298317 97.338 -72.570 51.205
298318 97.394 -72.572 51.166
298319 97.428 -72.571 51.127
298320 97.447 -72.590 51.088
298321 97.465 -72.610 51.049
298322 97.483 -72.631 51.010
298323 97.501 -72.651 50.974
298324 97.525 -72.653 50.933
298325 97.556 -72.632 50.889
298326 97.590 -72.622 50.842
298327 97.628 -72.629 50.792
298328 97.665 -72.638 50.741
298329 97.703 -72.646 50.689
298330 97.741 -72.654 50.634
298331 97.776 -72.657 50.583
298332 97.809 -72.658 50.533
298333 97.843 -72.660 50.480
298334 97.877 -72.660 50.426
298335 97.911 -72.662 50.372
298336 97.944 -72.663 50.321
298337 97.978 -72.664 50.269
298338 98.019 -72.669 50.216
298339 98.063 -72.676 50.163
298340 98.106 -72.683 50.110
298341 98.130 -72.715 50.058
298342 98.163 -72.725 50.008
298343 98.196 -72.736 49.960
298344 98.229 -72.747 49.909
298345 98.263 -72.757 49.859
298346 98.294 -72.767 49.809
298347 98.325 -72.776 49.759
298348 98.359 -72.779 49.710
298349 98.414 -72.745 49.660
298350 98.458 -72.724 49.610
298351 98.477 -72.736 49.562
298352 98.496 -72.747 49.515
298353 98.519 -72.780 49.466
298354 98.546 -72.819 49.416
298355 98.588 -72.857 49.365
298356 98.629 -72.895 49.315
298357 98.643 -72.920 49.264
298358 98.662 -72.929 49.213
298359 98.679 -72.920 49.163
298360 98.698 -72.916 49.109
298361 98.723 -72.923 49.057
298362 98.747 -72.930 49.006
298363 98.783 -72.973 48.955
298364 98.820 -73.026 48.907
298365 98.859 -73.079 48.858
298366 98.885 -73.104 48.810
298367 98.919 -73.110 48.763
298368 98.953 -73.113 48.716
298369 98.989 -73.116 48.669
298370 99.023 -73.120 48.622
298371 99.062 -73.123 48.576
298372 99.100 -73.127 48.528
298373 99.132 -73.144 48.480
298374 99.154 -73.154 48.432
298375 99.177 -73.164 48.385
298376 99.198 -73.173 48.338
298377 99.231 -73.192 48.290
298378 99.269 -73.217 48.244
298379 99.306 -73.242 48.198
298380 99.344 -73.267 48.153
298381 99.381 -73.292 48.108
298382 99.407 -73.300 48.063
298383 99.431 -73.307 48.019
298384 99.455 -73.314 47.973
298385 99.480 -73.321 47.927
298386 99.504 -73.329 47.883
298387 99.528 -73.336 47.839
298388 99.553 -73.343 47.795
298389 99.577 -73.350 47.750
298390 99.603 -73.360 47.706
298391 99.631 -73.378 47.661
298392 99.661 -73.391 47.616
298393 99.691 -73.399 47.571
298394 99.721 -73.409 47.526
298395 99.752 -73.417 47.481
298396 99.782 -73.426 47.435
298397 99.813 -73.435 47.390
298398 99.843 -73.443 47.344
298399 99.875 -73.452 47.299
298400 99.913 -73.462 47.255
298401 99.952 -73.471 47.210
298402 99.992 -73.479 47.166
298403 100.032 -73.488 47.122
298404 100.072 -73.497 47.079
298405 100.112 -73.506 47.038
298406 100.152 -73.515 46.998
298407 100.192 -73.524 46.959
298408 100.231 -73.533 46.920
298409 100.271 -73.542 46.884
298410 100.312 -73.551 46.848
298411 100.351 -73.559 46.811
298412 100.391 -73.568 46.776
298413 100.431 -73.578 46.742
298414 100.471 -73.587 46.709
298415 100.498 -73.593 46.676
298416 100.513 -73.600 46.643
298417 100.532 -73.590 46.610
298418 100.554 -73.571 46.577
298419 100.576 -73.552 46.544
298420 100.599 -73.532 46.512
298421 100.621 -73.513 46.480
298422 100.643 -73.494 46.449
298423 100.665 -73.474 46.417
298424 100.687 -73.456 46.385
298425 100.709 -73.437 46.352
298426 100.729 -73.423 46.321
298427 100.749 -73.413 46.289
298428 100.768 -73.402 46.257
298429 100.790 -73.396 46.224
298430 100.819 -73.398 46.191
298431 100.885 -73.515 46.158
298432 100.966 -73.642 46.125
298433 101.004 -73.657 46.092
298434 101.042 -73.672 46.059
298435 101.080 -73.687 46.025
298436 101.118 -73.703 45.991
298437 101.156 -73.718 45.957
298438 101.194 -73.734 45.923
298439 101.232 -73.749 45.889
298440 101.270 -73.764 45.855
298441 101.308 -73.779 45.822
298442 101.346 -73.794 45.787
298443 101.384 -73.810 45.752
298444 101.421 -73.825 45.718
298445 101.455 -73.843 45.683
298446 101.486 -73.862 45.650
298447 101.517 -73.881 45.617
298448 101.548 -73.900 45.583
298449 101.579 -73.919 45.550
298450 101.610 -73.938 45.517
298451 101.641 -73.957 45.484
298452 101.672 -73.976 45.452
298453 101.703 -73.995 45.421
298454 101.735 -74.013 45.389
298455 101.765 -74.032 45.359
298456 101.796 -74.051 45.328
298457 101.828 -74.070 45.298
298458 101.858 -74.089 45.267
298459 101.899 -74.105 45.237
298460 101.944 -74.120 45.209
298461 101.987 -74.133 45.180
298462 102.023 -74.091 45.150
298463 102.066 -74.073 45.119
298464 102.109 -74.082 45.088
298465 102.138 -74.085 45.058
298466 102.163 -74.086 45.028
298467 102.188 -74.087 44.999
298468 102.212 -74.088 44.970
298469 102.237 -74.088 44.940
298470 102.262 -74.089 44.910
298471 102.287 -74.089 44.878
298472 102.312 -74.090 44.850
298473 102.336 -74.091 44.821
298474 102.361 -74.092 44.791
298475 102.386 -74.092 44.762
298949 95.155 -72.287 52.532
298950 95.181 -72.302 52.512
298951 95.208 -72.316 52.491
298952 95.234 -72.330 52.470
298953 95.260 -72.339 52.449
298954 95.285 -72.346 52.427
298955 95.311 -72.353 52.404
298956 95.327 -72.356 52.380
298957 95.348 -72.361 52.355
298958 95.368 -72.365 52.330
298959 95.389 -72.369 52.304
298960 95.440 -72.385 52.278
298961 95.490 -72.393 52.254
298962 95.519 -72.371 52.231
298963 95.547 -72.348 52.208
298964 95.580 -72.335 52.186
298965 95.615 -72.325 52.165
298966 95.649 -72.319 52.144
298967 95.675 -72.343 52.125
298968 95.701 -72.368 52.106
298969 95.729 -72.392 52.089
298970 95.758 -72.417 52.073
298971 95.788 -72.442 52.055
298972 95.816 -72.466 52.038
298973 95.845 -72.491 52.021
298974 95.875 -72.515 52.005
298975 95.885 -72.531 51.987
298976 95.893 -72.543 51.971
298977 95.900 -72.556 51.956
298978 95.907 -72.568 51.941
298979 95.918 -72.560 51.925
298980 95.953 -72.557 51.911
298981 95.987 -72.554 51.898
298982 96.022 -72.551 51.885
298983 96.059 -72.545 51.876
298984 96.097 -72.540 51.868
298985 96.135 -72.534 51.861
298986 96.173 -72.528 51.854
298987 96.209 -72.524 51.847
298988 96.240 -72.526 51.838
298989 96.271 -72.529 51.827
298990 96.303 -72.531 51.814
298991 96.334 -72.534 51.802
298992 96.365 -72.537 51.790
298993 96.397 -72.539 51.777
298994 96.428 -72.541 51.764
298995 96.458 -72.540 51.747
298996 96.507 -72.547 51.732
298997 96.557 -72.547 51.718
298998 96.608 -72.547 51.703
298999 96.659 -72.546 51.690
299000 96.695 -72.545 51.674
299001 96.729 -72.544 51.656
299002 96.761 -72.543 51.636
299003 96.795 -72.542 51.614
299004 96.828 -72.541 51.593
299005 96.861 -72.540 51.571
299006 96.894 -72.538 51.549
299007 96.924 -72.534 51.526
299008 96.954 -72.530 51.500
299009 96.984 -72.525 51.474
299010 97.014 -72.521 51.448
299011 97.044 -72.516 51.423
299012 97.075 -72.512 51.399
299013 97.105 -72.509 51.374
299014 97.138 -72.513 51.347
299015 97.174 -72.516 51.315
299016 97.231 -72.519 51.283
299017 97.288 -72.521 51.248
299018 97.345 -72.524 51.211
299019 97.401 -72.526 51.172
299020 97.449 -72.527 51.135
299021 97.469 -72.545 51.095
299022 97.487 -72.565 51.054
299023 97.505 -72.586 51.012
299024 97.523 -72.606 50.973
299025 97.541 -72.625 50.932
299026 97.572 -72.606 50.887
299027 97.608 -72.603 50.841
299028 97.646 -72.612 50.790
299029 97.683 -72.620 50.740
299030 97.718 -72.623 50.690
299031 97.752 -72.625 50.638
299032 97.786 -72.625 50.590
299033 97.819 -72.627 50.540
299034 97.854 -72.628 50.488
299035 97.887 -72.629 50.434
299036 97.921 -72.631 50.380
299037 97.955 -72.631 50.327
299038 97.989 -72.633 50.274
299039 98.029 -72.638 50.221
299040 98.073 -72.644 50.169
299041 98.117 -72.651 50.115
299042 98.161 -72.658 50.063
299043 98.178 -72.697 50.013
299044 98.211 -72.707 49.963
299045 98.244 -72.719 49.912
299046 98.278 -72.729 49.861
299047 98.309 -72.738 49.812
299048 98.340 -72.747 49.763
299049 98.370 -72.755 49.713
299050 98.414 -72.741 49.662
299051 98.465 -72.711 49.612
299052 98.484 -72.723 49.564
299053 98.503 -72.735 49.516
299054 98.521 -72.747 49.467
299055 98.547 -72.782 49.417
299056 98.589 -72.819 49.367
299057 98.632 -72.858 49.316
299058 98.659 -72.889 49.265
299059 98.678 -72.898 49.213
299060 98.696 -72.893 49.162
299061 98.717 -72.894 49.108
299062 98.742 -72.901 49.056
299063 98.767 -72.908 49.004
299064 98.800 -72.945 48.953
299065 98.838 -72.997 48.904
299066 98.868 -73.031 48.855
299067 98.893 -73.050 48.806
299068 98.924 -73.059 48.758
299069 98.959 -73.062 48.712
299070 98.994 -73.065 48.666
299071 99.030 -73.068 48.619
299072 99.069 -73.071 48.572
299073 99.103 -73.086 48.525
299074 99.137 -73.103 48.477
299075 99.170 -73.121 48.428
299076 99.204 -73.139 48.381
299077 99.226 -73.148 48.335
299078 99.249 -73.158 48.288
299079 99.274 -73.171 48.241
299080 99.311 -73.195 48.196
299081 99.349 -73.221 48.151
299082 99.386 -73.245 48.105
299083 99.420 -73.264 48.061
299084 99.444 -73.271 48.016
299085 99.468 -73.278 47.971
299086 99.493 -73.286 47.925
299087 99.517 -73.293 47.881
299088 99.542 -73.300 47.838
299089 99.566 -73.307 47.793
299090 99.591 -73.317 47.749
299091 99.620 -73.335 47.704
299092 99.648 -73.352 47.659
299093 99.677 -73.370 47.613
299094 99.708 -73.379 47.569
299095 99.738 -73.387 47.523
299096 99.768 -73.396 47.478
299097 99.799 -73.405 47.433
299098 99.829 -73.414 47.387
299099 99.860 -73.423 47.341
299100 99.890 -73.431 47.295
299101 99.922 -73.437 47.251
299102 99.957 -73.443 47.206
299103 99.997 -73.452 47.162
299104 100.037 -73.462 47.118
299105 100.077 -73.471 47.075
299106 100.117 -73.479 47.033
299107 100.156 -73.488 46.993
299108 100.196 -73.497 46.954
299109 100.236 -73.506 46.915
299110 100.276 -73.515 46.877
299111 100.316 -73.524 46.840
299112 100.356 -73.533 46.803
299113 100.396 -73.542 46.766
299114 100.436 -73.550 46.732
299115 100.476 -73.559 46.699
299116 100.502 -73.567 46.666
299117 100.516 -73.573 46.632
299118 100.531 -73.579 46.599
299119 100.553 -73.559 46.567
299120 100.575 -73.540 46.535
299121 100.598 -73.521 46.503
299122 100.620 -73.502 46.473
299123 100.642 -73.483 46.441
299124 100.664 -73.464 46.409
299125 100.686 -73.444 46.377
299126 100.708 -73.425 46.344
299127 100.731 -73.406 46.313
299128 100.753 -73.387 46.282
299129 100.775 -73.368 46.249
299130 100.797 -73.348 46.216
299131 100.818 -73.338 46.183
299132 100.894 -73.482 46.149
299133 100.974 -73.606 46.116
299134 101.012 -73.621 46.083
299135 101.050 -73.637 46.050
299136 101.088 -73.652 46.016
299137 101.126 -73.667 45.982
299138 101.164 -73.682 45.949
299139 101.201 -73.697 45.914
299140 101.238 -73.713 45.880
299141 101.275 -73.729 45.846
299142 101.312 -73.744 45.812
299143 101.348 -73.760 45.778
299144 101.386 -73.776 45.743
299145 101.424 -73.788 45.709
299146 101.463 -73.797 45.674
299147 101.498 -73.815 45.640
299148 101.531 -73.836 45.606
299149 101.563 -73.857 45.573
299150 101.596 -73.878 45.539
299151 101.628 -73.899 45.505
299152 101.660 -73.920 45.472
299153 101.692 -73.941 45.440
299154 101.724 -73.960 45.409
299155 101.755 -73.979 45.377
299156 101.785 -73.998 45.347
299157 101.817 -74.017 45.316
299158 101.848 -74.036 45.284
299159 101.879 -74.055 45.253
299160 101.910 -74.074 45.224
299161 101.954 -74.089 45.195
299162 101.993 -74.068 45.165
299163 102.036 -74.029 45.135
299164 102.079 -74.039 45.105
299165 102.121 -74.048 45.074
299166 102.146 -74.049 45.044
299167 102.171 -74.050 45.015
299168 102.195 -74.051 44.985
299169 102.221 -74.051 44.956
299170 102.245 -74.051 44.926
299171 102.270 -74.052 44.896
299172 102.295 -74.053 44.866
299173 102.319 -74.054 44.837
299174 102.344 -74.054 44.808
299175 102.369 -74.054 44.778
299176 102.394 -74.055 44.750
299650 95.153 -72.243 52.528
299651 95.179 -72.257 52.508
299652 95.206 -72.271 52.487
299653 95.232 -72.285 52.466
299654 95.259 -72.299 52.445
299655 95.284 -72.309 52.422
299656 95.310 -72.316 52.399
299657 95.331 -72.321 52.376
299658 95.351 -72.326 52.351
299659 95.371 -72.330 52.326
299660 95.392 -72.334 52.301
299661 95.445 -72.350 52.276
299662 95.486 -72.346 52.252
299663 95.514 -72.324 52.230
299664 95.542 -72.302 52.207
299665 95.576 -72.290 52.184
299666 95.611 -72.280 52.163
299667 95.640 -72.290 52.142
299668 95.667 -72.315 52.123
299669 95.693 -72.339 52.105
299670 95.722 -72.363 52.087
299671 95.751 -72.388 52.071
299672 95.780 -72.412 52.054
299673 95.809 -72.437 52.037
299674 95.838 -72.462 52.020
299675 95.868 -72.487 52.005
299676 95.888 -72.507 51.988
299677 95.895 -72.519 51.972
299678 95.902 -72.532 51.956
299679 95.909 -72.544 51.941
299680 95.927 -72.527 51.926
299681 95.962 -72.524 51.911
299682 95.996 -72.521 51.898
299683 96.033 -72.515 51.884
299684 96.071 -72.509 51.874
299685 96.109 -72.503 51.865
299686 96.147 -72.497 51.856
299687 96.184 -72.492 51.848
299688 96.222 -72.486 51.841
299689 96.257 -72.484 51.833
299690 96.288 -72.486 51.822
299691 96.320 -72.488 51.808
299692 96.351 -72.491 51.797
299693 96.382 -72.493 51.785
299694 96.413 -72.495 51.772
299695 96.443 -72.494 51.757
299696 96.473 -72.493 51.742
299697 96.520 -72.500 51.728
299698 96.579 -72.506 51.713
299699 96.630 -72.506 51.698
299700 96.674 -72.505 51.683
299701 96.707 -72.504 51.665
299702 96.740 -72.503 51.647
299703 96.773 -72.502 51.629
299704 96.807 -72.501 51.610
299705 96.839 -72.500 51.590
299706 96.873 -72.499 51.570
299707 96.906 -72.498 51.551
299708 96.936 -72.493 51.530
299709 96.966 -72.489 51.504
299710 96.996 -72.484 51.478
299711 97.026 -72.480 51.453
299712 97.056 -72.475 51.427
299713 97.087 -72.472 51.402
299714 97.117 -72.467 51.377
299715 97.148 -72.466 51.350
299716 97.181 -72.470 51.319
299717 97.238 -72.472 51.285
299718 97.295 -72.475 51.251
299719 97.351 -72.478 51.216
299720 97.408 -72.480 51.182
299721 97.459 -72.485 51.147
299722 97.489 -72.501 51.108
299723 97.509 -72.521 51.065
299724 97.527 -72.541 51.020
299725 97.544 -72.562 50.976
299726 97.563 -72.582 50.931
299727 97.588 -72.578 50.885
299728 97.626 -72.585 50.839
299729 97.661 -72.590 50.790
299730 97.695 -72.591 50.742
299731 97.729 -72.592 50.694
299732 97.762 -72.593 50.646
299733 97.796 -72.594 50.600
299734 97.830 -72.596 50.550
299735 97.864 -72.597 50.497
299736 97.897 -72.598 50.442
299737 97.932 -72.599 50.388
299738 97.965 -72.600 50.335
299739 97.999 -72.602 50.282
299740 98.039 -72.606 50.229
299741 98.083 -72.613 50.177
299742 98.127 -72.620 50.122
299743 98.171 -72.627 50.070
299744 98.194 -72.668 50.019
299745 98.227 -72.679 49.967
299746 98.260 -72.690 49.915
299747 98.292 -72.700 49.865
299748 98.323 -72.708 49.816
299749 98.354 -72.717 49.767
299750 98.384 -72.725 49.717
299751 98.415 -72.735 49.666
299752 98.467 -72.704 49.617
299753 98.491 -72.710 49.568
299754 98.510 -72.722 49.520
299755 98.528 -72.734 49.469
299756 98.551 -72.747 49.417
299757 98.592 -72.784 49.367
299758 98.633 -72.820 49.316
299759 98.674 -72.859 49.264
299760 98.695 -72.867 49.212
299761 98.713 -72.866 49.160
299762 98.737 -72.872 49.107
299763 98.761 -72.879 49.055
299764 98.786 -72.885 49.004
299765 98.817 -72.915 48.953
299766 98.851 -72.957 48.902
299767 98.875 -72.977 48.852
299768 98.900 -72.997 48.804
299769 98.930 -73.007 48.756
299770 98.964 -73.011 48.709
299771 99.000 -73.014 48.662
299772 99.037 -73.017 48.615
299773 99.074 -73.026 48.569
299774 99.107 -73.044 48.522
299775 99.141 -73.061 48.475
299776 99.174 -73.079 48.427
299777 99.208 -73.097 48.380
299778 99.241 -73.115 48.333
299779 99.275 -73.133 48.286
299780 99.299 -73.144 48.240
299781 99.321 -73.153 48.195
299782 99.354 -73.174 48.149
299783 99.392 -73.199 48.104
299784 99.429 -73.223 48.059
299785 99.457 -73.235 48.015
299786 99.481 -73.242 47.970
299787 99.506 -73.249 47.925
299788 99.530 -73.257 47.880
299789 99.555 -73.264 47.836
299790 99.580 -73.274 47.791
299791 99.609 -73.292 47.748
299792 99.637 -73.309 47.703
299793 99.665 -73.327 47.657
299794 99.694 -73.346 47.612
299795 99.724 -73.358 47.567
299796 99.754 -73.367 47.522
299797 99.785 -73.375 47.476
299798 99.815 -73.384 47.430
299799 99.846 -73.393 47.385
299800 99.876 -73.402 47.339
299801 99.907 -73.410 47.293
299802 99.938 -73.415 47.249
299803 99.970 -73.421 47.204
299804 100.001 -73.426 47.159
299805 100.042 -73.435 47.115
299806 100.081 -73.443 47.071
299807 100.121 -73.452 47.029
299808 100.161 -73.462 46.988
299809 100.201 -73.471 46.949
299810 100.241 -73.479 46.910
299811 100.281 -73.488 46.871
299812 100.321 -73.497 46.833
299813 100.360 -73.506 46.795
299814 100.401 -73.515 46.758
299815 100.440 -73.524 46.724
299816 100.480 -73.533 46.690
299817 100.506 -73.540 46.656
299818 100.521 -73.546 46.623
299819 100.535 -73.552 46.590
299820 100.552 -73.549 46.558
299821 100.574 -73.529 46.527
299822 100.597 -73.510 46.496
299823 100.619 -73.491 46.466
299824 100.641 -73.471 46.434
299825 100.663 -73.452 46.403
299826 100.685 -73.434 46.372
299827 100.707 -73.414 46.339
299828 100.729 -73.395 46.308
299829 100.752 -73.376 46.276
299830 100.774 -73.356 46.243
299831 100.788 -73.333 46.209
299832 100.799 -73.331 46.175
299833 101.008 -73.579 46.141
299834 100.981 -73.571 46.107
299835 101.018 -73.586 46.074
299836 101.054 -73.602 46.041
299837 101.092 -73.618 46.008
299838 101.128 -73.633 45.974
299839 101.165 -73.649 45.940
299840 101.202 -73.665 45.905
299841 101.239 -73.680 45.871
299842 101.276 -73.696 45.837
299843 101.313 -73.712 45.803
299844 101.349 -73.727 45.769
299845 101.388 -73.739 45.735
299846 101.427 -73.750 45.700
299847 101.466 -73.760 45.665
299848 101.506 -73.769 45.631
299849 101.541 -73.785 45.597
299850 101.574 -73.807 45.563
299851 101.606 -73.828 45.529
299852 101.638 -73.849 45.494
299853 101.671 -73.870 45.460
299854 101.703 -73.891 45.428
299855 101.736 -73.912 45.396
299856 101.768 -73.933 45.365
299857 101.801 -73.954 45.335
299858 101.833 -73.976 45.304
299859 101.865 -73.997 45.272
299860 101.897 -74.018 45.240
299861 101.930 -74.039 45.211
299862 101.964 -74.046 45.181
299863 102.006 -73.986 45.150
299864 102.049 -73.995 45.120
299865 102.092 -74.005 45.090
299866 102.129 -74.012 45.059
299867 102.154 -74.013 45.030
299868 102.179 -74.013 45.000
299869 102.203 -74.013 44.971
299870 102.228 -74.014 44.941
299871 102.253 -74.015 44.912
299872 102.277 -74.016 44.882
299873 102.303 -74.016 44.853
299874 102.327 -74.017 44.824
299875 102.352 -74.017 44.795
299876 102.377 -74.018 44.766
300350 95.125 -72.183 52.544
300351 95.151 -72.197 52.524
300352 95.178 -72.211 52.503
300353 95.204 -72.226 52.483
300354 95.230 -72.240 52.462
300355 95.257 -72.255 52.440
300356 95.283 -72.269 52.417
300357 95.309 -72.279 52.395
300358 95.335 -72.287 52.372
300359 95.354 -72.291 52.348
300360 95.375 -72.295 52.324
300361 95.395 -72.299 52.299
300362 95.449 -72.316 52.275
300363 95.481 -72.299 52.251
300364 95.510 -72.277 52.229
300365 95.538 -72.255 52.206
300366 95.572 -72.244 52.183
300367 95.606 -72.237 52.162
300368 95.632 -72.262 52.141
300369 95.659 -72.286 52.121
300370 95.685 -72.309 52.103
300371 95.714 -72.334 52.086
300372 95.743 -72.359 52.070
300373 95.772 -72.384 52.054
300374 95.802 -72.409 52.037
300375 95.831 -72.434 52.021
300376 95.860 -72.458 52.005
300377 95.889 -72.483 51.989
300378 95.897 -72.496 51.973
300379 95.904 -72.509 51.957
300380 95.912 -72.521 51.942
300381 95.937 -72.493 51.927
300382 95.971 -72.490 51.912
300383 96.008 -72.485 51.898
300384 96.045 -72.479 51.885
300385 96.083 -72.473 51.874
300386 96.120 -72.468 51.863
300387 96.158 -72.462 51.853
300388 96.196 -72.456 51.845
300389 96.234 -72.450 51.837
300390 96.271 -72.444 51.829
300391 96.305 -72.443 51.819
300392 96.336 -72.446 51.807
300393 96.368 -72.448 51.795
300394 96.398 -72.449 51.783
300395 96.429 -72.448 51.770
300396 96.459 -72.446 51.755
300397 96.489 -72.446 51.739
300398 96.533 -72.451 51.724
300399 96.601 -72.465 51.708
300400 96.651 -72.465 51.692
300401 96.686 -72.464 51.676
300402 96.719 -72.463 51.658
300403 96.752 -72.462 51.638
300404 96.785 -72.461 51.620
300405 96.818 -72.460 51.602
300406 96.852 -72.459 51.584
300407 96.885 -72.458 51.567
300408 96.918 -72.457 51.549
300409 96.948 -72.453 51.529
300410 96.978 -72.448 51.504
300411 97.008 -72.443 51.478
300412 97.038 -72.439 51.453
300413 97.069 -72.435 51.428
300414 97.099 -72.431 51.402
300415 97.129 -72.426 51.376
300416 97.159 -72.421 51.348
300417 97.191 -72.423 51.316
300418 97.245 -72.426 51.285
300419 97.301 -72.428 51.254
300420 97.358 -72.431 51.223
300421 97.389 -72.446 51.192
300422 97.418 -72.462 51.158
300423 97.448 -72.479 51.119
300424 97.477 -72.496 51.075
300425 97.506 -72.512 51.029
300426 97.536 -72.528 50.983
300427 97.568 -72.544 50.937
300428 97.604 -72.554 50.892
300429 97.638 -72.556 50.848
300430 97.672 -72.558 50.803
300431 97.706 -72.559 50.755
300432 97.739 -72.560 50.707
300433 97.773 -72.562 50.659
300434 97.807 -72.563 50.611
300435 97.841 -72.564 50.560
300436 97.874 -72.565 50.507
300437 97.908 -72.566 50.452
300438 97.942 -72.568 50.399
300439 97.975 -72.569 50.346
300440 98.010 -72.570 50.293
300441 98.050 -72.575 50.240
300442 98.094 -72.581 50.187
300443 98.138 -72.588 50.132
300444 98.181 -72.595 50.078
300445 98.226 -72.602 50.026
300446 98.242 -72.650 49.972
300447 98.276 -72.661 49.921
300448 98.306 -72.670 49.871
300449 98.337 -72.678 49.822
300450 98.368 -72.688 49.772
300451 98.398 -72.696 49.722
300452 98.429 -72.705 49.671
300453 98.467 -72.700 49.621
300454 98.498 -72.698 49.571
300455 98.516 -72.710 49.521
300456 98.539 -72.718 49.468
300457 98.567 -72.725 49.416
300458 98.607 -72.762 49.366
300459 98.648 -72.799 49.315
300460 98.680 -72.820 49.263
300461 98.708 -72.835 49.210
300462 98.731 -72.842 49.158
300463 98.756 -72.850 49.105
300464 98.781 -72.857 49.053
300465 98.805 -72.863 49.002
300466 98.834 -72.884 48.952
300467 98.858 -72.904 48.901
300468 98.883 -72.923 48.851
300469 98.907 -72.943 48.803
300470 98.935 -72.957 48.755
300471 98.970 -72.960 48.708
300472 99.005 -72.963 48.661
300473 99.044 -72.967 48.614
300474 99.077 -72.984 48.568
300475 99.111 -73.002 48.522
300476 99.145 -73.020 48.475
300477 99.178 -73.038 48.428
300478 99.212 -73.055 48.380
300479 99.245 -73.073 48.333
300480 99.279 -73.091 48.286
300481 99.312 -73.109 48.240
300482 99.346 -73.127 48.194
300483 99.371 -73.139 48.148
300484 99.397 -73.152 48.104
300485 99.434 -73.177 48.059
300486 99.470 -73.199 48.014
300487 99.494 -73.206 47.970
300488 99.519 -73.214 47.924
300489 99.543 -73.221 47.880
300490 99.569 -73.230 47.835
300491 99.597 -73.249 47.790
300492 99.626 -73.266 47.746
300493 99.654 -73.284 47.701
300494 99.683 -73.302 47.656
300495 99.711 -73.320 47.611
300496 99.740 -73.337 47.566
300497 99.771 -73.346 47.520
300498 99.801 -73.355 47.474
300499 99.832 -73.363 47.429
300500 99.862 -73.372 47.384
300501 99.893 -73.381 47.337
300502 99.923 -73.389 47.292
300503 99.955 -73.393 47.247
300504 99.986 -73.399 47.203
300505 100.018 -73.404 47.157
300506 100.050 -73.409 47.112
300507 100.086 -73.417 47.068
300508 100.126 -73.426 47.025
300509 100.165 -73.434 46.984
300510 100.206 -73.443 46.944
300511 100.246 -73.452 46.905
300512 100.285 -73.462 46.865
300513 100.325 -73.471 46.826
300514 100.365 -73.479 46.788
300515 100.405 -73.488 46.751
300516 100.445 -73.497 46.716
300517 100.485 -73.506 46.682
300518 100.510 -73.513 46.648
300519 100.525 -73.519 46.615
300520 100.539 -73.525 46.583
300521 100.553 -73.531 46.551
300522 100.573 -73.518 46.521
300523 100.596 -73.499 46.491
300524 100.618 -73.480 46.461
300525 100.640 -73.461 46.430
300526 100.662 -73.441 46.400
300527 100.684 -73.422 46.370
300528 100.706 -73.403 46.340
300529 100.728 -73.383 46.306
300530 100.751 -73.365 46.273
300531 100.755 -73.349 46.239
300532 100.789 -73.362 46.204
300533 100.823 -73.374 46.170
300534 100.935 -73.531 46.135
300535 100.978 -73.556 46.101
300536 101.016 -73.562 46.068
300537 101.055 -73.568 46.034
300538 101.092 -73.584 46.000
300539 101.129 -73.600 45.967
300540 101.166 -73.615 45.933
300541 101.203 -73.631 45.898
300542 101.240 -73.647 45.864
300543 101.276 -73.662 45.830
300544 101.313 -73.678 45.796
300545 101.351 -73.691 45.762
300546 101.391 -73.701 45.727
300547 101.430 -73.711 45.692
300548 101.469 -73.722 45.657
300549 101.509 -73.731 45.622
300550 101.548 -73.741 45.588
300551 101.585 -73.756 45.553
300552 101.617 -73.777 45.519
300553 101.649 -73.798 45.484
300554 101.682 -73.819 45.450
300555 101.714 -73.841 45.417
300556 101.746 -73.862 45.385
300557 101.778 -73.883 45.353
300558 101.811 -73.904 45.323
300559 101.843 -73.926 45.292
300560 101.876 -73.946 45.260
300561 101.908 -73.969 45.228
300562 101.937 -73.998 45.198
300563 101.976 -73.949 45.167
300564 102.019 -73.952 45.136
300565 102.062 -73.961 45.105
300566 102.105 -73.971 45.075
300567 102.137 -73.975 45.045
300568 102.162 -73.976 45.016
300569 102.186 -73.976 44.987
300570 102.211 -73.977 44.957
300571 102.236 -73.978 44.928
300572 102.261 -73.979 44.898
300573 102.286 -73.979 44.869
300574 102.310 -73.979 44.840
300575 102.335 -73.980 44.812
300576 102.360 -73.981 44.783
301051 95.123 -72.138 52.541
301052 95.149 -72.152 52.520
301053 95.176 -72.167 52.499
301054 95.202 -72.181 52.479
301055 95.229 -72.196 52.458
301056 95.255 -72.209 52.436
301057 95.281 -72.224 52.413
301058 95.308 -72.238 52.391
301059 95.334 -72.249 52.369
301060 95.359 -72.257 52.346
301061 95.378 -72.260 52.322
301062 95.398 -72.265 52.298
301063 95.449 -72.275 52.274
301064 95.477 -72.253 52.251
301065 95.506 -72.231 52.228
301066 95.534 -72.208 52.205
301067 95.568 -72.199 52.183
301068 95.599 -72.208 52.161
301069 95.624 -72.232 52.141
301070 95.651 -72.256 52.121
301071 95.678 -72.280 52.103
301072 95.707 -72.305 52.087
301073 95.736 -72.330 52.070
301074 95.765 -72.355 52.053
301075 95.794 -72.380 52.038
301076 95.823 -72.404 52.023
301077 95.852 -72.429 52.006
301078 95.882 -72.454 51.990
301079 95.899 -72.472 51.975
301080 95.907 -72.485 51.960
301081 95.914 -72.472 51.945
301082 95.945 -72.459 51.930
301083 95.981 -72.455 51.915
301084 96.019 -72.449 51.900
301085 96.056 -72.443 51.886
301086 96.094 -72.437 51.874
301087 96.132 -72.431 51.863
301088 96.170 -72.425 51.852
301089 96.207 -72.420 51.843
301090 96.245 -72.414 51.834
301091 96.283 -72.408 51.826
301092 96.321 -72.402 51.816
301093 96.354 -72.403 51.806
301094 96.384 -72.403 51.795
301095 96.414 -72.402 51.783
301096 96.444 -72.400 51.770
301097 96.474 -72.399 51.756
301098 96.505 -72.398 51.738
301099 96.546 -72.403 51.721
301100 96.616 -72.419 51.704
301101 96.665 -72.424 51.687
301102 96.698 -72.423 51.669
301103 96.731 -72.421 51.650
301104 96.764 -72.421 51.630
301105 96.797 -72.420 51.611
301106 96.830 -72.418 51.593
301107 96.863 -72.418 51.577
301108 96.896 -72.417 51.561
301109 96.930 -72.415 51.545
301110 96.960 -72.412 51.525
301111 96.991 -72.407 51.501
301112 97.020 -72.403 51.475
301113 97.050 -72.399 51.450
301114 97.081 -72.394 51.425
301115 97.111 -72.390 51.399
301116 97.141 -72.385 51.371
301117 97.171 -72.381 51.343
301118 97.202 -72.377 51.312
301119 97.251 -72.380 51.284
301120 97.289 -72.392 51.257
301121 97.318 -72.408 51.229
301122 97.351 -72.424 51.199
301123 97.391 -72.440 51.166
301124 97.430 -72.456 51.126
301125 97.469 -72.472 51.082
301126 97.509 -72.489 51.037
301127 97.548 -72.505 50.991
301128 97.584 -72.512 50.945
301129 97.617 -72.515 50.901
301130 97.648 -72.517 50.860
301131 97.679 -72.520 50.816
301132 97.713 -72.522 50.770
301133 97.747 -72.523 50.721
301134 97.781 -72.525 50.672
301135 97.815 -72.527 50.622
301136 97.849 -72.528 50.570
301137 97.883 -72.530 50.517
301138 97.917 -72.532 50.463
301139 97.951 -72.534 50.410
301140 97.985 -72.535 50.357
301141 98.019 -72.537 50.304
301142 98.055 -72.544 50.251
301143 98.099 -72.551 50.197
301144 98.144 -72.558 50.142
301145 98.189 -72.564 50.087
301146 98.234 -72.571 50.032
301147 98.280 -72.578 49.978
301148 98.290 -72.632 49.927
301149 98.321 -72.641 49.877
301150 98.352 -72.650 49.828
301151 98.382 -72.658 49.777
301152 98.413 -72.667 49.726
301153 98.444 -72.675 49.675
301154 98.475 -72.685 49.624
301155 98.504 -72.685 49.573
301156 98.529 -72.693 49.520
301157 98.556 -72.698 49.467
301158 98.583 -72.704 49.414
301159 98.624 -72.741 49.363
301160 98.658 -72.767 49.312
301161 98.686 -72.781 49.261
301162 98.718 -72.797 49.208
301163 98.733 -72.804 49.156
301164 98.758 -72.811 49.103
301165 98.783 -72.819 49.052
301166 98.808 -72.826 49.001
301167 98.847 -72.838 48.951
301168 98.865 -72.850 48.900
301169 98.890 -72.870 48.851
301170 98.915 -72.889 48.804
301171 98.941 -72.906 48.756
301172 98.976 -72.909 48.708
301173 99.012 -72.912 48.662
301174 99.048 -72.925 48.616
301175 99.081 -72.942 48.568
301176 99.115 -72.960 48.523
301177 99.148 -72.978 48.477
301178 99.182 -72.996 48.429
301179 99.215 -73.014 48.381
301180 99.249 -73.032 48.333
301181 99.283 -73.049 48.286
301182 99.316 -73.067 48.240
301183 99.350 -73.086 48.194
301184 99.383 -73.103 48.148
301185 99.417 -73.121 48.104
301186 99.443 -73.133 48.060
301187 99.477 -73.155 48.015
301188 99.507 -73.170 47.969
301189 99.532 -73.177 47.924
301190 99.557 -73.187 47.880
301191 99.586 -73.205 47.836
301192 99.615 -73.223 47.790
301193 99.643 -73.241 47.745
301194 99.672 -73.258 47.700
301195 99.700 -73.277 47.655
301196 99.729 -73.295 47.610
301197 99.758 -73.312 47.564
301198 99.787 -73.325 47.519
301199 99.818 -73.333 47.474
301200 99.848 -73.343 47.428
301201 99.878 -73.352 47.383
301202 99.909 -73.360 47.336
301203 99.940 -73.367 47.290
301204 99.972 -73.372 47.245
301205 100.003 -73.377 47.201
301206 100.034 -73.383 47.155
301207 100.066 -73.388 47.111
301208 100.098 -73.393 47.067
301209 100.130 -73.399 47.022
301210 100.171 -73.408 46.982
301211 100.210 -73.417 46.941
301212 100.250 -73.426 46.900
301213 100.290 -73.434 46.861
301214 100.330 -73.443 46.821
301215 100.370 -73.452 46.782
301216 100.410 -73.462 46.745
301217 100.450 -73.470 46.710
301218 100.489 -73.479 46.676
301219 100.514 -73.487 46.643
301220 100.528 -73.493 46.610
301221 100.543 -73.498 46.578
301222 100.557 -73.504 46.546
301223 100.572 -73.507 46.516
301224 100.595 -73.488 46.488
301225 100.617 -73.469 46.460
301226 100.639 -73.449 46.431
301227 100.661 -73.430 46.401
301228 100.683 -73.411 46.369
301229 100.705 -73.392 46.337
301230 100.711 -73.367 46.304
301231 100.745 -73.379 46.270
301232 100.778 -73.392 46.235
301233 100.812 -73.404 46.199
301234 100.873 -73.455 46.164
301235 100.936 -73.534 46.130
301236 100.975 -73.541 46.096
301237 101.013 -73.547 46.062
301238 101.052 -73.554 46.029
301239 101.091 -73.561 45.995
301240 101.129 -73.568 45.961
301241 101.167 -73.582 45.927
301242 101.203 -73.598 45.892
301243 101.240 -73.614 45.858
301244 101.277 -73.630 45.824
301245 101.315 -73.643 45.790
301246 101.354 -73.653 45.755
301247 101.394 -73.663 45.720
301248 101.433 -73.673 45.685
301249 101.472 -73.683 45.649
301250 101.512 -73.694 45.614
301251 101.551 -73.703 45.580
301252 101.591 -73.713 45.545
301253 101.627 -73.727 45.510
301254 101.660 -73.748 45.475
301255 101.692 -73.769 45.440
301256 101.725 -73.791 45.408
301257 101.757 -73.811 45.375
301258 101.790 -73.832 45.342
301259 101.822 -73.854 45.311
301260 101.853 -73.877 45.280
301261 101.883 -73.906 45.248
301262 101.912 -73.935 45.216
301263 101.947 -73.927 45.184
301264 101.989 -73.908 45.153
301265 102.032 -73.918 45.123
301266 102.076 -73.928 45.093
301267 102.118 -73.937 45.062
301268 102.144 -73.938 45.032
301269 102.170 -73.939 45.002
301270 102.194 -73.940 44.974
301271 102.219 -73.941 44.945
301272 102.244 -73.941 44.915
301273 102.268 -73.941 44.886
301274 102.293 -73.942 44.857
301275 102.318 -73.943 44.828
301276 102.343 -73.944 44.800
301752 95.121 -72.092 52.537
301753 95.147 -72.107 52.516
301754 95.174 -72.121 52.495
301755 95.200 -72.136 52.474
301756 95.227 -72.150 52.453
301757 95.253 -72.164 52.432
301758 95.279 -72.179 52.409
301759 95.306 -72.193 52.386
301760 95.332 -72.208 52.365
301761 95.358 -72.220 52.342
301762 95.381 -72.225 52.319
301763 95.401 -72.230 52.296
301764 95.445 -72.229 52.273
301765 95.473 -72.207 52.249
301766 95.501 -72.184 52.227
301767 95.530 -72.163 52.205
301768 95.564 -72.155 52.184
301769 95.591 -72.179 52.161
301770 95.617 -72.203 52.141
301771 95.642 -72.227 52.122
301772 95.670 -72.252 52.104
301773 95.699 -72.277 52.086
301774 95.729 -72.301 52.069
301775 95.758 -72.326 52.054
301776 95.787 -72.351 52.039
301777 95.816 -72.375 52.025
301778 95.845 -72.400 52.008
301779 95.874 -72.425 51.993
301780 95.902 -72.449 51.978
301781 95.909 -72.462 51.963
301782 95.920 -72.428 51.947
301783 95.955 -72.424 51.933
301784 95.993 -72.419 51.919
301785 96.031 -72.413 51.905
301786 96.068 -72.407 51.891
301787 96.106 -72.401 51.878
301788 96.144 -72.396 51.865
301789 96.181 -72.390 51.853
301790 96.219 -72.384 51.843
301791 96.257 -72.377 51.833
301792 96.295 -72.370 51.824
301793 96.333 -72.362 51.815
301794 96.369 -72.356 51.805
301795 96.400 -72.352 51.794
301796 96.431 -72.349 51.782
301797 96.461 -72.348 51.771
301798 96.491 -72.347 51.758
301799 96.522 -72.346 51.738
301800 96.552 -72.344 51.720
301801 96.608 -72.354 51.701
301802 96.668 -72.365 51.683
301803 96.716 -72.377 51.663
301804 96.743 -72.381 51.644
301805 96.776 -72.380 51.624
301806 96.809 -72.378 51.603
301807 96.842 -72.377 51.585
301808 96.875 -72.377 51.569
301809 96.908 -72.375 51.552
301810 96.941 -72.374 51.536
301811 96.972 -72.371 51.517
301812 97.003 -72.367 51.493
301813 97.032 -72.362 51.468
301814 97.063 -72.358 51.442
301815 97.093 -72.353 51.416
301816 97.123 -72.349 51.389
301817 97.153 -72.344 51.362
301818 97.183 -72.340 51.333
301819 97.216 -72.340 51.307
301820 97.253 -72.352 51.283
301821 97.293 -72.368 51.258
301822 97.332 -72.385 51.231
301823 97.371 -72.401 51.203
301824 97.411 -72.417 51.169
301825 97.450 -72.433 51.130
301826 97.490 -72.450 51.087
301827 97.529 -72.465 51.043
301828 97.564 -72.470 50.998
301829 97.596 -72.472 50.954
301830 97.630 -72.475 50.911
301831 97.662 -72.477 50.872
301832 97.692 -72.479 50.829
301833 97.723 -72.482 50.783
301834 97.754 -72.484 50.735
301835 97.788 -72.487 50.685
301836 97.822 -72.488 50.634
301837 97.857 -72.490 50.582
301838 97.890 -72.492 50.528
301839 97.925 -72.493 50.473
301840 97.958 -72.495 50.419
301841 97.993 -72.497 50.368
301842 98.026 -72.499 50.315
301843 98.062 -72.506 50.261
301844 98.097 -72.517 50.206
301845 98.133 -72.527 50.150
301846 98.169 -72.537 50.094
301847 98.211 -72.544 50.039
301848 98.256 -72.551 49.985
301849 98.301 -72.558 49.933
301850 98.320 -72.591 49.884
301851 98.352 -72.601 49.834
301852 98.384 -72.612 49.781
301853 98.416 -72.622 49.728
301854 98.450 -72.634 49.676
301855 98.487 -72.650 49.623
301856 98.509 -72.666 49.572
301857 98.546 -72.672 49.517
301858 98.572 -72.678 49.463
301859 98.599 -72.684 49.410
301860 98.636 -72.713 49.359
301861 98.667 -72.728 49.307
301862 98.702 -72.745 49.255
301863 98.725 -72.757 49.204
301864 98.729 -72.760 49.153
301865 98.754 -72.768 49.102
301866 98.779 -72.776 49.051
301867 98.815 -72.786 49.000
301868 98.867 -72.801 48.951
301869 98.881 -72.807 48.901
301870 98.897 -72.816 48.854
301871 98.922 -72.835 48.807
301872 98.946 -72.855 48.759
301873 98.981 -72.858 48.712
301874 99.018 -72.865 48.666
301875 99.052 -72.883 48.619
301876 99.085 -72.901 48.572
301877 99.119 -72.919 48.525
301878 99.152 -72.936 48.479
301879 99.186 -72.954 48.431
301880 99.219 -72.972 48.383
301881 99.253 -72.990 48.334
301882 99.286 -73.008 48.287
301883 99.320 -73.026 48.240
301884 99.354 -73.044 48.194
301885 99.387 -73.061 48.148
301886 99.421 -73.079 48.104
301887 99.454 -73.097 48.059
301888 99.488 -73.115 48.014
301889 99.519 -73.133 47.969
301890 99.546 -73.144 47.924
301891 99.575 -73.162 47.879
301892 99.603 -73.180 47.834
301893 99.632 -73.198 47.789
301894 99.660 -73.215 47.744
301895 99.689 -73.233 47.699
301896 99.717 -73.252 47.654
301897 99.746 -73.269 47.608
301898 99.775 -73.287 47.562
301899 99.803 -73.304 47.517
301900 99.834 -73.313 47.473
301901 99.864 -73.321 47.427
301902 99.895 -73.330 47.382
301903 99.925 -73.340 47.335
301904 99.957 -73.346 47.289
301905 99.988 -73.351 47.245
301906 100.019 -73.356 47.199
301907 100.051 -73.362 47.154
301908 100.083 -73.366 47.109
301909 100.114 -73.371 47.065
301910 100.146 -73.377 47.019
301911 100.177 -73.382 46.978
301912 100.215 -73.390 46.937
301913 100.255 -73.399 46.896
301914 100.295 -73.408 46.856
301915 100.335 -73.417 46.816
301916 100.374 -73.426 46.777
301917 100.414 -73.434 46.740
301918 100.454 -73.443 46.705
301919 100.494 -73.452 46.671
301920 100.517 -73.459 46.637
301921 100.532 -73.465 46.605
301922 100.547 -73.471 46.573
301923 100.561 -73.477 46.541
301924 100.575 -73.484 46.513
301925 100.594 -73.477 46.486
301926 100.616 -73.458 46.460
301927 100.638 -73.439 46.431
301928 100.660 -73.419 46.401
301929 100.682 -73.400 46.368
301930 100.700 -73.396 46.333
301931 100.734 -73.409 46.301
301932 100.768 -73.421 46.266
301933 100.802 -73.434 46.230
301934 100.835 -73.446 46.195
301935 100.919 -73.528 46.159
301936 100.933 -73.520 46.124
301937 100.972 -73.526 46.090
301938 101.011 -73.533 46.057
301939 101.049 -73.540 46.025
301940 101.088 -73.546 45.991
301941 101.127 -73.553 45.956
301942 101.165 -73.560 45.922
301943 101.204 -73.567 45.888
301944 101.241 -73.581 45.854
301945 101.279 -73.594 45.820
301946 101.318 -73.605 45.785
301947 101.357 -73.615 45.750
301948 101.397 -73.625 45.714
301949 101.436 -73.635 45.679
301950 101.475 -73.645 45.644
301951 101.515 -73.655 45.609
301952 101.554 -73.665 45.574
301953 101.594 -73.675 45.539
301954 101.633 -73.685 45.504
301955 101.671 -73.697 45.468
301956 101.703 -73.719 45.433
301957 101.735 -73.740 45.399
301958 101.768 -73.761 45.367
301959 101.799 -73.785 45.334
301960 101.828 -73.814 45.301
301961 101.858 -73.843 45.269
301962 101.888 -73.871 45.237
301963 101.917 -73.900 45.205
301964 101.959 -73.865 45.173
301965 102.003 -73.874 45.143
301966 102.045 -73.884 45.112
301967 102.089 -73.894 45.081
301968 102.128 -73.901 45.051
301969 102.153 -73.902 45.020
301970 102.177 -73.903 44.991
301971 102.202 -73.903 44.962
301972 102.227 -73.904 44.933
301973 102.252 -73.904 44.904
301974 102.276 -73.905 44.876
301975 102.301 -73.906 44.847
301976 102.326 -73.906 44.818
302453 95.119 -72.048 52.533
302454 95.145 -72.062 52.512
302455 95.172 -72.077 52.491
302456 95.198 -72.091 52.470
302457 95.225 -72.105 52.448
302458 95.251 -72.120 52.427
302459 95.277 -72.133 52.404
302460 95.304 -72.148 52.382
302461 95.330 -72.162 52.361
302462 95.356 -72.177 52.339
302463 95.383 -72.190 52.316
302464 95.405 -72.196 52.294
302465 95.440 -72.183 52.271
302466 95.468 -72.160 52.248
302467 95.496 -72.138 52.227
302468 95.526 -72.117 52.205
302469 95.556 -72.126 52.184
302470 95.583 -72.150 52.162
302471 95.609 -72.174 52.141
302472 95.635 -72.199 52.122
302473 95.663 -72.223 52.104
302474 95.692 -72.248 52.086
302475 95.721 -72.272 52.070
302476 95.750 -72.297 52.055
302477 95.779 -72.322 52.040
302478 95.809 -72.346 52.027
302479 95.838 -72.371 52.012
302480 95.867 -72.396 51.997
302481 95.896 -72.421 51.982
302482 95.911 -72.438 51.967
302483 95.929 -72.395 51.953
302484 95.967 -72.389 51.938
302485 96.005 -72.383 51.924
302486 96.042 -72.376 51.911
302487 96.081 -72.368 51.896
302488 96.119 -72.361 51.882
302489 96.157 -72.353 51.868
302490 96.195 -72.346 51.855
302491 96.233 -72.338 51.844
302492 96.271 -72.330 51.834
302493 96.309 -72.322 51.824
302494 96.347 -72.315 51.814
302495 96.382 -72.309 51.803
302496 96.413 -72.306 51.792
302497 96.445 -72.302 51.781
302498 96.476 -72.299 51.769
302499 96.507 -72.296 51.756
302500 96.538 -72.293 51.738
302501 96.570 -72.290 51.719
302502 96.601 -72.287 51.701
302503 96.652 -72.294 51.682
302504 96.709 -72.302 51.660
302505 96.765 -72.312 51.639
302506 96.803 -72.327 51.618
302507 96.821 -72.337 51.596
302508 96.854 -72.337 51.577
302509 96.887 -72.336 51.558
302510 96.920 -72.334 51.540
302511 96.953 -72.334 51.521
302512 96.984 -72.330 51.501
302513 97.015 -72.326 51.478
302514 97.044 -72.321 51.453
302515 97.075 -72.317 51.428
302516 97.105 -72.312 51.402
302517 97.136 -72.309 51.375
302518 97.171 -72.312 51.349
302519 97.204 -72.313 51.324
302520 97.234 -72.313 51.300
302521 97.274 -72.329 51.278
302522 97.313 -72.345 51.254
302523 97.352 -72.362 51.229
302524 97.391 -72.377 51.202
302525 97.431 -72.393 51.169
302526 97.470 -72.410 51.131
302527 97.510 -72.426 51.090
302528 97.543 -72.428 51.048
302529 97.576 -72.431 51.004
302530 97.609 -72.432 50.962
302531 97.642 -72.434 50.922
302532 97.675 -72.437 50.882
302533 97.707 -72.439 50.838
302534 97.737 -72.442 50.792
302535 97.767 -72.445 50.744
302536 97.797 -72.447 50.694
302537 97.829 -72.450 50.643
302538 97.864 -72.452 50.591
302539 97.898 -72.453 50.537
302540 97.932 -72.455 50.481
302541 97.966 -72.457 50.427
302542 98.000 -72.459 50.373
302543 98.034 -72.460 50.319
302544 98.069 -72.469 50.263
302545 98.105 -72.479 50.209
302546 98.141 -72.489 50.153
302547 98.176 -72.499 50.097
302548 98.212 -72.509 50.041
302549 98.248 -72.519 49.988
302550 98.284 -72.529 49.937
302551 98.320 -72.542 49.886
302552 98.358 -72.558 49.835
302553 98.395 -72.574 49.781
302554 98.432 -72.590 49.728
302555 98.469 -72.606 49.674
302556 98.507 -72.622 49.621
302557 98.498 -72.642 49.568
302558 98.535 -72.649 49.515
302559 98.573 -72.656 49.461
302560 98.610 -72.663 49.407
302561 98.651 -72.678 49.355
302562 98.686 -72.694 49.303
302563 98.721 -72.711 49.251
302564 98.724 -72.715 49.200
302565 98.726 -72.718 49.149
302566 98.750 -72.725 49.100
302567 98.775 -72.732 49.050
302568 98.833 -72.749 49.001
302569 98.887 -72.764 48.953
302570 98.901 -72.770 48.904
302571 98.915 -72.776 48.858
302572 98.929 -72.782 48.810
302573 98.953 -72.801 48.763
302574 98.988 -72.807 48.717
302575 99.021 -72.823 48.670
302576 99.055 -72.841 48.623
302577 99.089 -72.859 48.576
302578 99.122 -72.877 48.528
302579 99.156 -72.895 48.480
302580 99.190 -72.913 48.432
302581 99.223 -72.930 48.384
302582 99.257 -72.948 48.335
302583 99.290 -72.967 48.288
302584 99.324 -72.984 48.241
302585 99.357 -73.003 48.194
302586 99.391 -73.023 48.148
302587 99.425 -73.043 48.104
302588 99.459 -73.064 48.059
302589 99.493 -73.083 48.014
302590 99.528 -73.100 47.969
302591 99.562 -73.117 47.923
302592 99.592 -73.136 47.878
302593 99.621 -73.155 47.834
302594 99.649 -73.172 47.789
302595 99.678 -73.190 47.743
302596 99.706 -73.208 47.698
302597 99.735 -73.226 47.653
302598 99.764 -73.244 47.607
302599 99.792 -73.261 47.562
302600 99.821 -73.280 47.517
302601 99.850 -73.292 47.472
302602 99.880 -73.301 47.427
302603 99.911 -73.310 47.381
302604 99.941 -73.318 47.335
302605 99.973 -73.324 47.289
302606 100.004 -73.329 47.244
302607 100.036 -73.334 47.198
302608 100.068 -73.340 47.152
302609 100.100 -73.345 47.107
302610 100.131 -73.350 47.063
302611 100.162 -73.355 47.017
302612 100.194 -73.361 46.975
302613 100.226 -73.366 46.933
302614 100.260 -73.372 46.892
302615 100.299 -73.381 46.851
302616 100.339 -73.390 46.811
302617 100.379 -73.399 46.773
302618 100.419 -73.408 46.735
302619 100.459 -73.417 46.700
302620 100.499 -73.426 46.665
302621 100.522 -73.433 46.632
302622 100.536 -73.439 46.600
302623 100.550 -73.445 46.568
302624 100.565 -73.450 46.538
302625 100.579 -73.456 46.511
302626 100.594 -73.462 46.485
302627 100.615 -73.446 46.459
302628 100.637 -73.427 46.429
302629 100.656 -73.415 46.398
302630 100.690 -73.427 46.364
302631 100.723 -73.440 46.329
302632 100.757 -73.452 46.297
302633 100.791 -73.465 46.262
302634 100.825 -73.477 46.226
302635 100.858 -73.490 46.191
302636 100.893 -73.496 46.155
302637 100.933 -73.502 46.120
302638 100.973 -73.506 46.087
302639 101.012 -73.512 46.054
302640 101.052 -73.516 46.019
302641 101.092 -73.521 45.985
302642 101.132 -73.526 45.951
302643 101.172 -73.531 45.917
302644 101.211 -73.536 45.882
302645 101.251 -73.541 45.849
302646 101.289 -73.552 45.815
302647 101.328 -73.562 45.780
302648 101.367 -73.573 45.745
302649 101.405 -73.584 45.710
302650 101.443 -73.594 45.674
302651 101.481 -73.605 45.639
302652 101.519 -73.615 45.603
302653 101.558 -73.626 45.568
302654 101.597 -73.637 45.533
302655 101.636 -73.647 45.497
302656 101.675 -73.657 45.461
302657 101.713 -73.669 45.426
302658 101.745 -73.694 45.392
302659 101.774 -73.722 45.359
302660 101.804 -73.751 45.326
302661 101.833 -73.780 45.292
302662 101.863 -73.808 45.259
302663 101.893 -73.837 45.227
302664 101.929 -73.821 45.195
302665 101.972 -73.831 45.162
302666 102.016 -73.841 45.132
302667 102.058 -73.850 45.101
302668 102.102 -73.860 45.070
302669 102.135 -73.865 45.040
302670 102.160 -73.865 45.009
302671 102.185 -73.866 44.980
302672 102.210 -73.866 44.951
302673 102.235 -73.867 44.922
302674 102.259 -73.868 44.894
302675 102.284 -73.869 44.865
302676 102.309 -73.869 44.837
302677 102.334 -73.869 44.808
303154 95.117 -72.002 52.529
303155 95.143 -72.017 52.508
303156 95.170 -72.031 52.486
303157 95.196 -72.045 52.465
303158 95.222 -72.060 52.443
303159 95.249 -72.074 52.422
303160 95.275 -72.089 52.400
303161 95.302 -72.103 52.379
303162 95.328 -72.117 52.358
303163 95.354 -72.132 52.336
303164 95.381 -72.146 52.314
303165 95.407 -72.158 52.292
303166 95.436 -72.136 52.269
303167 95.464 -72.114 52.247
303168 95.492 -72.092 52.227
303169 95.522 -72.073 52.206
303170 95.548 -72.097 52.185
303171 95.574 -72.121 52.164
303172 95.601 -72.146 52.143
303173 95.627 -72.169 52.123
303174 95.656 -72.194 52.105
303175 95.685 -72.219 52.087
303176 95.713 -72.243 52.071
303177 95.743 -72.268 52.056
303178 95.772 -72.293 52.042
303179 95.801 -72.318 52.029
303180 95.830 -72.343 52.015
303181 95.860 -72.367 52.001
303182 95.889 -72.392 51.987
303183 95.915 -72.398 51.973
303184 95.942 -72.352 51.959
303185 95.980 -72.344 51.945
303186 96.019 -72.337 51.930
303187 96.057 -72.329 51.917
303188 96.095 -72.321 51.902
303189 96.133 -72.314 51.887
303190 96.171 -72.305 51.873
303191 96.209 -72.298 51.859
303192 96.247 -72.290 51.846
303193 96.286 -72.283 51.835
303194 96.324 -72.275 51.824
303195 96.362 -72.268 51.812
303196 96.395 -72.263 51.801
303197 96.427 -72.259 51.790
303198 96.458 -72.256 51.779
303199 96.489 -72.252 51.766
303200 96.521 -72.249 51.753
303201 96.552 -72.246 51.736
303202 96.583 -72.243 51.718
303203 96.613 -72.245 51.700
303204 96.654 -72.250 51.681
303205 96.702 -72.247 51.658
303206 96.752 -72.244 51.635
303207 96.806 -72.249 51.612
303208 96.863 -72.258 51.589
303209 96.890 -72.276 51.568
303210 96.899 -72.294 51.548
303211 96.932 -72.293 51.527
303212 96.965 -72.293 51.506
303213 96.997 -72.290 51.485
303214 97.027 -72.285 51.462
303215 97.056 -72.280 51.438
303216 97.090 -72.280 51.414
303217 97.125 -72.283 51.388
303218 97.160 -72.286 51.362
303219 97.191 -72.285 51.337
303220 97.221 -72.285 51.315
303221 97.254 -72.290 51.294
303222 97.294 -72.305 51.273
303223 97.333 -72.322 51.251
303224 97.372 -72.338 51.227
303225 97.412 -72.354 51.200
303226 97.451 -72.371 51.168
303227 97.490 -72.384 51.132
303228 97.522 -72.387 51.092
303229 97.556 -72.388 51.051
303230 97.589 -72.390 51.010
303231 97.621 -72.393 50.970
303232 97.655 -72.394 50.930
303233 97.688 -72.396 50.890
303234 97.720 -72.399 50.845
303235 97.751 -72.402 50.799
303236 97.781 -72.404 50.751
303237 97.811 -72.407 50.701
303238 97.842 -72.410 50.650
303239 97.872 -72.413 50.597
303240 97.905 -72.415 50.542
303241 97.939 -72.417 50.486
303242 97.973 -72.418 50.431
303243 98.007 -72.420 50.375
303244 98.041 -72.422 50.319
303245 98.077 -72.431 50.263
303246 98.112 -72.442 50.208
303247 98.148 -72.452 50.152
303248 98.184 -72.462 50.096
303249 98.220 -72.472 50.041
303250 98.259 -72.484 49.988
303251 98.299 -72.498 49.938
303252 98.340 -72.512 49.887
303253 98.377 -72.528 49.834
303254 98.415 -72.544 49.780
303255 98.452 -72.560 49.726
303256 98.489 -72.576 49.673
303257 98.526 -72.592 49.619
303258 98.486 -72.619 49.566
303259 98.523 -72.625 49.513
303260 98.561 -72.633 49.458
303261 98.610 -72.638 49.404
303262 98.668 -72.644 49.352
303263 98.705 -72.660 49.299
303264 98.720 -72.669 49.248
303265 98.722 -72.672 49.197
303266 98.725 -72.675 49.147
303267 98.746 -72.681 49.099
303268 98.790 -72.694 49.050
303269 98.850 -72.711 49.003
303270 98.908 -72.728 48.956
303271 98.921 -72.733 48.908
303272 98.935 -72.738 48.861
303273 98.948 -72.744 48.814
303274 98.962 -72.750 48.766
303275 98.992 -72.764 48.721
303276 99.025 -72.782 48.674
303277 99.059 -72.802 48.626
303278 99.093 -72.822 48.578
303279 99.127 -72.842 48.531
303280 99.161 -72.862 48.482
303281 99.195 -72.882 48.434
303282 99.228 -72.903 48.385
303283 99.262 -72.923 48.336
303284 99.296 -72.943 48.288
303285 99.330 -72.963 48.241
303286 99.364 -72.983 48.194
303287 99.398 -73.004 48.149
303288 99.431 -73.024 48.104
303289 99.465 -73.044 48.059
303290 99.501 -73.061 48.014
303291 99.541 -73.068 47.968
303292 99.576 -73.082 47.922
303293 99.604 -73.106 47.878
303294 99.636 -73.126 47.834
303295 99.666 -73.147 47.788
303296 99.695 -73.164 47.743
303297 99.723 -73.183 47.697
303298 99.752 -73.201 47.651
303299 99.781 -73.218 47.605
303300 99.809 -73.236 47.561
303301 99.838 -73.254 47.516
303302 99.866 -73.271 47.471
303303 99.897 -73.280 47.426
303304 99.927 -73.289 47.381
303305 99.958 -73.297 47.334
303306 99.990 -73.302 47.288
303307 100.021 -73.308 47.242
303308 100.053 -73.313 47.197
303309 100.084 -73.318 47.151
303310 100.116 -73.324 47.106
303311 100.147 -73.329 47.061
303312 100.179 -73.334 47.016
303313 100.211 -73.339 46.972
303314 100.243 -73.344 46.930
303315 100.274 -73.349 46.889
303316 100.305 -73.355 46.848
303317 100.344 -73.363 46.807
303318 100.384 -73.372 46.768
303319 100.424 -73.381 46.731
303320 100.464 -73.390 46.695
303321 100.503 -73.399 46.661
303322 100.525 -73.405 46.629
303323 100.540 -73.412 46.596
303324 100.554 -73.418 46.565
303325 100.568 -73.424 46.536
303326 100.583 -73.430 46.510
303327 100.598 -73.436 46.484
303328 100.614 -73.436 46.458
303329 100.645 -73.444 46.427
303330 100.679 -73.442 46.395
303331 100.711 -73.439 46.361
303332 100.745 -73.437 46.326
303333 100.777 -73.434 46.292
303334 100.810 -73.431 46.257
303335 100.843 -73.428 46.222
303336 100.878 -73.432 46.187
303337 100.916 -73.446 46.151
303338 100.954 -73.450 46.116
303339 100.993 -73.456 46.082
303340 101.033 -73.460 46.049
303341 101.073 -73.465 46.014
303342 101.113 -73.470 45.980
303343 101.152 -73.475 45.946
303344 101.192 -73.480 45.912
303345 101.232 -73.485 45.877
303346 101.271 -73.492 45.844
303347 101.310 -73.503 45.810
303348 101.348 -73.513 45.775
303349 101.387 -73.524 45.740
303350 101.425 -73.534 45.705
303351 101.463 -73.545 45.669
303352 101.501 -73.556 45.633
303353 101.540 -73.566 45.598
303354 101.579 -73.577 45.563
303355 101.617 -73.588 45.526
303356 101.655 -73.599 45.490
303357 101.693 -73.611 45.454
303358 101.721 -73.637 45.419
303359 101.750 -73.662 45.385
303360 101.779 -73.688 45.351
303361 101.809 -73.716 45.318
303362 101.838 -73.745 45.283
303363 101.868 -73.774 45.250
303364 101.900 -73.785 45.217
303365 101.943 -73.787 45.185
303366 101.985 -73.797 45.152
303367 102.029 -73.807 45.120
303368 102.071 -73.816 45.089
303369 102.115 -73.825 45.060
303370 102.144 -73.828 45.029
303371 102.168 -73.828 44.999
303372 102.193 -73.829 44.970
303373 102.218 -73.830 44.941
303374 102.242 -73.831 44.912
303375 102.267 -73.831 44.883
303376 102.292 -73.832 44.855
303377 102.317 -73.832 44.827
303855 95.115 -71.958 52.524
303856 95.141 -71.972 52.504
303857 95.168 -71.986 52.482
303858 95.194 -72.001 52.460
303859 95.220 -72.015 52.439
303860 95.247 -72.030 52.418
303861 95.273 -72.044 52.396
303862 95.300 -72.058 52.375
303863 95.326 -72.072 52.354
303864 95.352 -72.086 52.333
303865 95.379 -72.101 52.312
303866 95.392 -72.130 52.290
303867 95.404 -72.102 52.268
303868 95.438 -72.080 52.246
303869 95.471 -72.062 52.226
303870 95.509 -72.056 52.206
303871 95.540 -72.068 52.186
303872 95.566 -72.092 52.166
303873 95.593 -72.116 52.146
303874 95.619 -72.140 52.126
303875 95.648 -72.165 52.107
303876 95.677 -72.189 52.089
303877 95.706 -72.215 52.073
303878 95.736 -72.240 52.057
303879 95.765 -72.264 52.043
303880 95.794 -72.287 52.031
303881 95.824 -72.309 52.019
303882 95.853 -72.330 52.006
303883 95.883 -72.352 51.993
303884 95.921 -72.356 51.979
303885 95.948 -72.312 51.966
303886 95.995 -72.297 51.951
303887 96.033 -72.290 51.937
303888 96.071 -72.281 51.923
303889 96.109 -72.274 51.908
303890 96.148 -72.266 51.893
303891 96.185 -72.258 51.879
303892 96.224 -72.251 51.865
303893 96.262 -72.243 51.852
303894 96.300 -72.236 51.839
303895 96.338 -72.228 51.826
303896 96.376 -72.220 51.812
303897 96.408 -72.216 51.800
303898 96.440 -72.213 51.787
303899 96.471 -72.209 51.775
303900 96.503 -72.206 51.762
303901 96.534 -72.202 51.749
303902 96.565 -72.199 51.732
303903 96.596 -72.198 51.715
303904 96.624 -72.213 51.698
303905 96.663 -72.220 51.679
303906 96.713 -72.217 51.655
303907 96.761 -72.214 51.631
303908 96.810 -72.211 51.607
303909 96.860 -72.208 51.581
303910 96.908 -72.205 51.558
303911 96.960 -72.205 51.537
303912 96.977 -72.226 51.516
303913 96.977 -72.251 51.494
303914 97.009 -72.249 51.472
303915 97.043 -72.251 51.448
303916 97.079 -72.254 51.424
303917 97.114 -72.256 51.398
303918 97.149 -72.258 51.373
303919 97.178 -72.258 51.349
303920 97.209 -72.257 51.327
303921 97.238 -72.257 51.307
303922 97.274 -72.266 51.287
303923 97.314 -72.283 51.267
303924 97.353 -72.299 51.247
303925 97.392 -72.315 51.225
303926 97.432 -72.331 51.200
303927 97.469 -72.342 51.168
303928 97.502 -72.344 51.132
303929 97.535 -72.346 51.093
303930 97.568 -72.348 51.055
303931 97.601 -72.350 51.016
303932 97.634 -72.352 50.976
303933 97.667 -72.354 50.936
303934 97.701 -72.356 50.895
303935 97.733 -72.359 50.850
303936 97.764 -72.361 50.803
303937 97.795 -72.364 50.754
303938 97.825 -72.367 50.703
303939 97.856 -72.370 50.652
303940 97.886 -72.372 50.599
303941 97.917 -72.375 50.544
303942 97.947 -72.378 50.487
303943 97.981 -72.380 50.430
303944 98.014 -72.382 50.372
303945 98.049 -72.384 50.315
303946 98.084 -72.394 50.258
303947 98.120 -72.404 50.203
303948 98.156 -72.414 50.147
303949 98.196 -72.428 50.092
303950 98.236 -72.441 50.037
303951 98.277 -72.455 49.986
303952 98.318 -72.469 49.937
303953 98.358 -72.484 49.886
303954 98.396 -72.500 49.834
303955 98.435 -72.515 49.779
303956 98.472 -72.531 49.726
303957 98.509 -72.547 49.672
303958 98.547 -72.563 49.618
303959 98.474 -72.595 49.565
303960 98.512 -72.603 49.510
303961 98.554 -72.609 49.456
303962 98.611 -72.614 49.402
303963 98.668 -72.619 49.349
303964 98.717 -72.624 49.296
303965 98.719 -72.626 49.246
303966 98.721 -72.629 49.197
303967 98.723 -72.632 49.148
303968 98.747 -72.640 49.101
303969 98.807 -72.656 49.053
303970 98.867 -72.674 49.005
303971 98.927 -72.691 48.958
303972 98.941 -72.698 48.911
303973 98.952 -72.708 48.864
303974 98.963 -72.719 48.816
303975 98.977 -72.729 48.770
303976 98.998 -72.742 48.723
303977 99.031 -72.763 48.676
303978 99.066 -72.782 48.628
303979 99.099 -72.803 48.580
303980 99.133 -72.823 48.531
303981 99.167 -72.843 48.483
303982 99.201 -72.863 48.434
303983 99.234 -72.883 48.385
303984 99.269 -72.904 48.335
303985 99.302 -72.924 48.287
303986 99.336 -72.944 48.240
303987 99.370 -72.964 48.194
303988 99.404 -72.985 48.149
303989 99.438 -73.004 48.103
303990 99.474 -73.020 48.058
303991 99.513 -73.028 48.012
303992 99.553 -73.036 47.966
303993 99.592 -73.043 47.921
303994 99.615 -73.075 47.876
303995 99.647 -73.095 47.832
303996 99.678 -73.116 47.786
303997 99.710 -73.136 47.741
303998 99.741 -73.157 47.696
303999 99.769 -73.175 47.651
304000 99.798 -73.193 47.605
304001 99.827 -73.211 47.560
304002 99.855 -73.229 47.515
304003 99.884 -73.246 47.470
304004 99.913 -73.259 47.425
304005 99.944 -73.268 47.379
304006 99.975 -73.276 47.333
304007 100.006 -73.281 47.287
304008 100.037 -73.286 47.242
304009 100.069 -73.291 47.196
304010 100.101 -73.296 47.151
304011 100.133 -73.302 47.105
304012 100.164 -73.307 47.060
304013 100.196 -73.312 47.014
304014 100.227 -73.318 46.971
304015 100.259 -73.323 46.928
304016 100.290 -73.328 46.886
304017 100.322 -73.333 46.845
304018 100.354 -73.339 46.804
304019 100.388 -73.346 46.765
304020 100.428 -73.354 46.727
304021 100.468 -73.363 46.692
304022 100.508 -73.372 46.658
304023 100.529 -73.379 46.625
304024 100.544 -73.385 46.593
304025 100.558 -73.391 46.562
304026 100.572 -73.397 46.534
304027 100.587 -73.402 46.508
304028 100.609 -73.392 46.482
304029 100.632 -73.378 46.455
304030 100.663 -73.374 46.425
304031 100.696 -73.371 46.391
304032 100.729 -73.368 46.357
304033 100.762 -73.365 46.323
304034 100.795 -73.363 46.288
304035 100.828 -73.360 46.253
304036 100.863 -73.365 46.218
304037 100.903 -73.382 46.182
304038 100.939 -73.394 46.147
304039 100.974 -73.399 46.112
304040 101.014 -73.404 46.077
304041 101.054 -73.409 46.043
304042 101.094 -73.414 46.009
304043 101.133 -73.419 45.976
304044 101.173 -73.424 45.942
304045 101.213 -73.429 45.907
304046 101.253 -73.434 45.873
304047 101.291 -73.443 45.839
304048 101.330 -73.453 45.805
304049 101.369 -73.464 45.770
304050 101.407 -73.474 45.736
304051 101.445 -73.485 45.700
304052 101.483 -73.496 45.663
304053 101.522 -73.506 45.628
304054 101.560 -73.518 45.593
304055 101.599 -73.528 45.557
304056 101.637 -73.539 45.521
304057 101.670 -73.557 45.484
304058 101.699 -73.584 45.447
304059 101.727 -73.610 45.412
304060 101.756 -73.634 45.378
304061 101.785 -73.659 45.343
304062 101.814 -73.684 45.309
304063 101.843 -73.711 45.275
304064 101.873 -73.740 45.242
304065 101.913 -73.744 45.208
304066 101.956 -73.753 45.174
304067 101.999 -73.763 45.142
304068 102.042 -73.772 45.110
304069 102.085 -73.781 45.079
304070 102.126 -73.791 45.049
304071 102.151 -73.791 45.019
304072 102.176 -73.792 44.989
304073 102.201 -73.793 44.960
304074 102.226 -73.793 44.931
304075 102.250 -73.794 44.902
304076 102.275 -73.794 44.873
304077 102.300 -73.795 44.845
304556 95.113 -71.912 52.519
304557 95.140 -71.926 52.499
304558 95.166 -71.941 52.478
304559 95.192 -71.955 52.456
304560 95.219 -71.970 52.435
304561 95.245 -71.984 52.414
304562 95.271 -71.998 52.393
304563 95.298 -72.013 52.371
304564 95.324 -72.027 52.351
304565 95.351 -72.042 52.330
304566 95.377 -72.056 52.309
304567 95.376 -72.095 52.288
304568 95.366 -72.106 52.267
304569 95.398 -72.090 52.246
304570 95.436 -72.086 52.225
304571 95.474 -72.085 52.206
304572 95.512 -72.083 52.187
304573 95.550 -72.080 52.167
304574 95.585 -72.087 52.148
304575 95.611 -72.111 52.129
304576 95.640 -72.136 52.110
304577 95.670 -72.159 52.092
304578 95.699 -72.180 52.075
304579 95.729 -72.202 52.060
304580 95.758 -72.224 52.047
304581 95.788 -72.246 52.034
304582 95.818 -72.268 52.022
304583 95.847 -72.290 52.010
304584 95.880 -72.299 51.998
304585 95.925 -72.308 51.985
304586 95.952 -72.273 51.972
304587 96.002 -72.256 51.959
304588 96.047 -72.242 51.945
304589 96.086 -72.234 51.932
304590 96.123 -72.227 51.916
304591 96.162 -72.219 51.902
304592 96.200 -72.211 51.888
304593 96.238 -72.204 51.873
304594 96.276 -72.196 51.859
304595 96.314 -72.188 51.846
304596 96.353 -72.180 51.831
304597 96.390 -72.173 51.815
304598 96.422 -72.170 51.800
304599 96.453 -72.166 51.786
304600 96.484 -72.163 51.772
304601 96.516 -72.159 51.758
304602 96.547 -72.156 51.744
304603 96.579 -72.152 51.727
304604 96.607 -72.167 51.709
304605 96.636 -72.182 51.692
304606 96.673 -72.189 51.674
304607 96.722 -72.186 51.652
304608 96.771 -72.183 51.629
304609 96.820 -72.180 51.604
304610 96.869 -72.177 51.578
304611 96.918 -72.174 51.554
304612 96.967 -72.171 51.532
304613 97.007 -72.168 51.510
304614 97.004 -72.193 51.487
304615 97.017 -72.215 51.465
304616 97.061 -72.223 51.441
304617 97.095 -72.225 51.415
304618 97.130 -72.227 51.390
304619 97.165 -72.230 51.366
304620 97.196 -72.230 51.343
304621 97.226 -72.229 51.322
304622 97.256 -72.229 51.303
304623 97.294 -72.243 51.283
304624 97.334 -72.259 51.263
304625 97.373 -72.275 51.243
304626 97.413 -72.292 51.222
304627 97.449 -72.300 51.198
304628 97.481 -72.302 51.169
304629 97.515 -72.304 51.134
304630 97.547 -72.306 51.096
304631 97.581 -72.309 51.057
304632 97.614 -72.310 51.019
304633 97.647 -72.312 50.979
304634 97.680 -72.315 50.937
304635 97.713 -72.316 50.895
304636 97.746 -72.318 50.848
304637 97.779 -72.321 50.800
304638 97.809 -72.324 50.750
304639 97.840 -72.326 50.698
304640 97.870 -72.329 50.646
304641 97.900 -72.332 50.594
304642 97.930 -72.335 50.540
304643 97.961 -72.338 50.484
304644 97.991 -72.340 50.427
304645 98.022 -72.343 50.370
304646 98.056 -72.346 50.313
304647 98.092 -72.357 50.256
304648 98.133 -72.371 50.201
304649 98.173 -72.384 50.145
304650 98.214 -72.399 50.089
304651 98.254 -72.412 50.036
304652 98.295 -72.426 49.985
304653 98.336 -72.440 49.936
304654 98.376 -72.454 49.886
304655 98.415 -72.470 49.834
304656 98.453 -72.486 49.780
304657 98.492 -72.502 49.726
304658 98.529 -72.518 49.671
304659 98.566 -72.534 49.617
304660 98.462 -72.572 49.563
304661 98.500 -72.579 49.510
304662 98.554 -72.584 49.456
304663 98.612 -72.590 49.402
304664 98.678 -72.594 49.349
304665 98.726 -72.597 49.298
304666 98.725 -72.596 49.248
304667 98.725 -72.595 49.200
304668 98.725 -72.594 49.151
304669 98.764 -72.609 49.104
304670 98.823 -72.631 49.056
304671 98.883 -72.652 49.009
304672 98.938 -72.672 48.962
304673 98.949 -72.682 48.914
304674 98.960 -72.693 48.866
304675 98.972 -72.703 48.819
304676 98.992 -72.715 48.772
304677 99.012 -72.728 48.725
304678 99.038 -72.743 48.677
304679 99.072 -72.763 48.630
304680 99.106 -72.784 48.582
304681 99.139 -72.804 48.533
304682 99.173 -72.824 48.484
304683 99.207 -72.844 48.435
304684 99.241 -72.864 48.385
304685 99.275 -72.885 48.336
304686 99.308 -72.904 48.288
304687 99.342 -72.925 48.240
304688 99.376 -72.945 48.194
304689 99.410 -72.965 48.148
304690 99.446 -72.981 48.103
304691 99.486 -72.988 48.058
304692 99.525 -72.995 48.012
304693 99.565 -73.003 47.966
304694 99.605 -73.011 47.920
304695 99.631 -73.037 47.876
304696 99.658 -73.064 47.831
304697 99.689 -73.085 47.786
304698 99.721 -73.105 47.741
304699 99.752 -73.126 47.696
304700 99.783 -73.146 47.650
304701 99.815 -73.167 47.605
304702 99.844 -73.186 47.560
304703 99.872 -73.203 47.515
304704 99.901 -73.221 47.470
304705 99.930 -73.239 47.425
304706 99.960 -73.247 47.379
304707 99.991 -73.254 47.332
304708 100.023 -73.259 47.287
304709 100.054 -73.264 47.242
304710 100.086 -73.270 47.197
304711 100.118 -73.275 47.151
304712 100.149 -73.280 47.105
304713 100.181 -73.286 47.059
304714 100.212 -73.291 47.013
304715 100.244 -73.296 46.970
304716 100.275 -73.302 46.927
304717 100.307 -73.307 46.885
304718 100.339 -73.311 46.843
304719 100.370 -73.317 46.802
304720 100.402 -73.322 46.762
304721 100.433 -73.327 46.725
304722 100.473 -73.336 46.689
304723 100.513 -73.346 46.654
304724 100.533 -73.352 46.622
304725 100.547 -73.358 46.591
304726 100.562 -73.364 46.561
304727 100.584 -73.353 46.532
304728 100.607 -73.340 46.504
304729 100.630 -73.326 46.477
304730 100.653 -73.311 46.451
304731 100.681 -73.302 46.420
304732 100.714 -73.300 46.388
304733 100.747 -73.297 46.354
304734 100.779 -73.294 46.319
304735 100.813 -73.292 46.284
304736 100.848 -73.298 46.249
304737 100.888 -73.314 46.213
304738 100.927 -73.332 46.178
304739 100.964 -73.343 46.143
304740 100.996 -73.348 46.107
304741 101.035 -73.353 46.073
304742 101.075 -73.358 46.039
304743 101.115 -73.363 46.005
304744 101.155 -73.368 45.971
304745 101.194 -73.373 45.938
304746 101.234 -73.377 45.903
304747 101.274 -73.383 45.868
304748 101.312 -73.393 45.834
304749 101.350 -73.404 45.800
304750 101.389 -73.415 45.766
304751 101.427 -73.425 45.731
304752 101.465 -73.436 45.695
304753 101.503 -73.447 45.659
304754 101.542 -73.458 45.624
304755 101.581 -73.468 45.588
304756 101.619 -73.479 45.552
304757 101.648 -73.504 45.515
304758 101.677 -73.531 45.478
304759 101.705 -73.558 45.441
304760 101.734 -73.583 45.406
304761 101.762 -73.607 45.371
304762 101.791 -73.632 45.336
304763 101.820 -73.656 45.301
304764 101.849 -73.681 45.267
304765 101.883 -73.700 45.233
304766 101.926 -73.709 45.198
304767 101.969 -73.719 45.164
304768 102.012 -73.728 45.132
304769 102.055 -73.738 45.101
304770 102.098 -73.748 45.070
304771 102.134 -73.754 45.039
304772 102.159 -73.755 45.009
304773 102.184 -73.755 44.980
304774 102.208 -73.756 44.951
304775 102.233 -73.756 44.922
304776 102.258 -73.757 44.893
304777 102.283 -73.758 44.864
305257 95.111 -71.867 52.514
305258 95.138 -71.882 52.494
305259 95.164 -71.896 52.473
305260 95.191 -71.911 52.452
305261 95.217 -71.925 52.431
305262 95.243 -71.939 52.410
305263 95.270 -71.954 52.389
305264 95.296 -71.968 52.369
305265 95.322 -71.982 52.349
305266 95.349 -71.996 52.327
305267 95.365 -72.014 52.307
305268 95.359 -72.059 52.287
305269 95.355 -72.078 52.266
305270 95.370 -72.089 52.245
305271 95.408 -72.086 52.226
305272 95.447 -72.083 52.207
305273 95.485 -72.080 52.189
305274 95.523 -72.077 52.169
305275 95.563 -72.074 52.151
305276 95.605 -72.075 52.133
305277 95.634 -72.096 52.113
305278 95.664 -72.118 52.095
305279 95.693 -72.140 52.079
305280 95.723 -72.161 52.064
305281 95.752 -72.183 52.051
305282 95.782 -72.205 52.039
305283 95.812 -72.227 52.027
305284 95.840 -72.245 52.015
305285 95.884 -72.251 52.004
305286 95.929 -72.260 51.992
305287 95.961 -72.238 51.980
305288 96.007 -72.217 51.967
305289 96.055 -72.199 51.954
305290 96.100 -72.187 51.942
305291 96.138 -72.180 51.926
305292 96.176 -72.171 51.911
305293 96.214 -72.164 51.897
305294 96.252 -72.156 51.883
305295 96.291 -72.149 51.869
305296 96.329 -72.141 51.855
305297 96.367 -72.133 51.839
305298 96.404 -72.127 51.821
305299 96.435 -72.123 51.803
305300 96.466 -72.120 51.787
305301 96.498 -72.116 51.771
305302 96.529 -72.113 51.756
305303 96.560 -72.109 51.741
305304 96.590 -72.120 51.723
305305 96.618 -72.135 51.704
305306 96.647 -72.150 51.686
305307 96.683 -72.158 51.668
305308 96.732 -72.155 51.648
305309 96.781 -72.153 51.626
305310 96.830 -72.150 51.603
305311 96.879 -72.147 51.578
305312 96.928 -72.144 51.554
305313 96.977 -72.141 51.532
305314 97.003 -72.137 51.510
305315 97.009 -72.152 51.487
305316 97.015 -72.175 51.464
305317 97.066 -72.187 51.438
305318 97.111 -72.196 51.412
305319 97.146 -72.199 51.386
305320 97.181 -72.201 51.362
305321 97.214 -72.202 51.340
305322 97.243 -72.202 51.320
305323 97.275 -72.204 51.300
305324 97.314 -72.220 51.279
305325 97.354 -72.236 51.259
305326 97.393 -72.252 51.239
305327 97.428 -72.258 51.218
305328 97.461 -72.260 51.195
305329 97.494 -72.262 51.167
305330 97.527 -72.265 51.135
305331 97.561 -72.266 51.098
305332 97.593 -72.268 51.059
305333 97.627 -72.271 51.020
305334 97.659 -72.272 50.979
305335 97.692 -72.274 50.937
305336 97.726 -72.277 50.893
305337 97.758 -72.278 50.843
305338 97.792 -72.280 50.793
305339 97.823 -72.283 50.742
305340 97.853 -72.286 50.690
305341 97.884 -72.289 50.639
305342 97.914 -72.292 50.588
305343 97.944 -72.294 50.534
305344 97.974 -72.297 50.478
305345 98.005 -72.300 50.423
305346 98.035 -72.303 50.367
305347 98.048 -72.321 50.310
305348 98.107 -72.329 50.255
305349 98.151 -72.342 50.200
305350 98.192 -72.356 50.144
305351 98.232 -72.369 50.088
305352 98.273 -72.384 50.035
305353 98.313 -72.397 49.985
305354 98.354 -72.411 49.936
305355 98.394 -72.424 49.886
305356 98.433 -72.441 49.834
305357 98.472 -72.457 49.780
305358 98.510 -72.473 49.726
305359 98.549 -72.489 49.671
305360 98.569 -72.508 49.616
305361 98.451 -72.549 49.563
305362 98.497 -72.555 49.511
305363 98.559 -72.559 49.457
305364 98.627 -72.563 49.404
305365 98.694 -72.568 49.351
305366 98.734 -72.569 49.301
305367 98.734 -72.569 49.252
305368 98.734 -72.568 49.204
305369 98.754 -72.579 49.155
305370 98.783 -72.596 49.108
305371 98.839 -72.617 49.060
305372 98.898 -72.638 49.013
305373 98.946 -72.656 48.966
305374 98.957 -72.667 48.918
305375 98.969 -72.677 48.869
305376 98.988 -72.689 48.821
305377 99.008 -72.701 48.775
305378 99.027 -72.713 48.727
305379 99.048 -72.725 48.679
305380 99.078 -72.744 48.632
305381 99.112 -72.764 48.584
305382 99.146 -72.785 48.534
305383 99.180 -72.804 48.485
305384 99.213 -72.825 48.435
305385 99.247 -72.845 48.386
305386 99.281 -72.865 48.337
305387 99.315 -72.885 48.289
305388 99.349 -72.906 48.241
305389 99.382 -72.926 48.194
305390 99.419 -72.941 48.148
305391 99.458 -72.948 48.103
305392 99.498 -72.956 48.057
305393 99.538 -72.963 48.011
305394 99.577 -72.970 47.966
305395 99.617 -72.978 47.920
305396 99.649 -72.997 47.876
305397 99.669 -73.032 47.831
305398 99.700 -73.054 47.787
305399 99.732 -73.074 47.741
305400 99.763 -73.095 47.696
305401 99.794 -73.115 47.650
305402 99.826 -73.136 47.605
305403 99.857 -73.156 47.561
305404 99.889 -73.177 47.515
305405 99.918 -73.195 47.470
305406 99.947 -73.214 47.425
305407 99.977 -73.227 47.379
305408 100.008 -73.233 47.333
305409 100.039 -73.238 47.288
305410 100.071 -73.243 47.243
305411 100.103 -73.249 47.197
305412 100.134 -73.254 47.152
305413 100.165 -73.259 47.106
305414 100.197 -73.264 47.060
305415 100.229 -73.269 47.015
305416 100.261 -73.274 46.971
305417 100.292 -73.280 46.928
305418 100.324 -73.285 46.886
305419 100.355 -73.290 46.844
305420 100.387 -73.296 46.803
305421 100.418 -73.301 46.762
305422 100.450 -73.306 46.723
305423 100.482 -73.311 46.688
305424 100.517 -73.318 46.654
305425 100.537 -73.325 46.622
305426 100.559 -73.314 46.591
305427 100.582 -73.300 46.562
305428 100.605 -73.286 46.533
305429 100.629 -73.273 46.506
305430 100.652 -73.259 46.478
305431 100.675 -73.246 46.449
305432 100.698 -73.232 46.417
305433 100.731 -73.229 46.384
305434 100.764 -73.226 46.349
305435 100.797 -73.224 46.315
305436 100.833 -73.230 46.280
305437 100.873 -73.247 46.245
305438 100.912 -73.264 46.210
305439 100.952 -73.281 46.174
305440 100.987 -73.292 46.139
305441 101.020 -73.297 46.104
305442 101.056 -73.302 46.070
305443 101.096 -73.307 46.036
305444 101.135 -73.311 46.002
305445 101.175 -73.317 45.969
305446 101.215 -73.321 45.935
305447 101.255 -73.327 45.900
305448 101.294 -73.333 45.865
305449 101.332 -73.344 45.831
305450 101.371 -73.355 45.797
305451 101.409 -73.365 45.762
305452 101.447 -73.376 45.727
305453 101.486 -73.387 45.692
305454 101.524 -73.398 45.656
305455 101.562 -73.409 45.620
305456 101.598 -73.424 45.584
305457 101.626 -73.451 45.547
305458 101.655 -73.477 45.510
305459 101.683 -73.505 45.473
305460 101.711 -73.531 45.436
305461 101.740 -73.556 45.400
305462 101.769 -73.581 45.364
305463 101.798 -73.605 45.329
305464 101.826 -73.629 45.293
305465 101.855 -73.653 45.258
305466 101.896 -73.666 45.224
305467 101.939 -73.675 45.189
305468 101.982 -73.685 45.155
305469 102.025 -73.694 45.123
305470 102.068 -73.704 45.092
305471 102.111 -73.714 45.061
305472 102.142 -73.718 45.030
305473 102.167 -73.718 44.999
305474 102.192 -73.719 44.970
305475 102.217 -73.719 44.940
305476 102.241 -73.720 44.912
305477 102.266 -73.721 44.885
305478 102.290 -73.721 44.856
305958 95.110 -71.823 52.511
305959 95.136 -71.836 52.490
305960 95.162 -71.851 52.469
305961 95.189 -71.865 52.449
305962 95.215 -71.879 52.428
305963 95.242 -71.894 52.407
305964 95.268 -71.908 52.386
305965 95.294 -71.923 52.366
305966 95.321 -71.937 52.347
305967 95.347 -71.951 52.327
305968 95.342 -71.973 52.307
305969 95.342 -72.023 52.287
305970 95.344 -72.050 52.267
305971 95.349 -72.058 52.246
305972 95.388 -72.055 52.227
305973 95.426 -72.052 52.208
305974 95.464 -72.049 52.190
305975 95.509 -72.046 52.172
305976 95.561 -72.043 52.154
305977 95.601 -72.040 52.136
305978 95.628 -72.055 52.118
305979 95.658 -72.077 52.100
305980 95.687 -72.099 52.083
305981 95.717 -72.121 52.068
305982 95.747 -72.142 52.055
305983 95.776 -72.164 52.043
305984 95.806 -72.186 52.031
305985 95.842 -72.195 52.020
305986 95.887 -72.203 52.010
305987 95.933 -72.212 51.999
305988 95.971 -72.206 51.987
305989 96.012 -72.177 51.975
305990 96.060 -72.160 51.963
305991 96.110 -72.143 51.951
305992 96.152 -72.132 51.937
305993 96.190 -72.124 51.922
305994 96.229 -72.117 51.907
305995 96.267 -72.109 51.893
305996 96.305 -72.102 51.880
305997 96.343 -72.094 51.865
305998 96.381 -72.086 51.849
305999 96.417 -72.080 51.831
306000 96.448 -72.077 51.812
306001 96.480 -72.073 51.793
306002 96.511 -72.070 51.774
306003 96.542 -72.066 51.757
306004 96.572 -72.073 51.742
306005 96.601 -72.088 51.722
306006 96.629 -72.104 51.701
306007 96.658 -72.119 51.681
306008 96.692 -72.128 51.661
306009 96.742 -72.125 51.643
306010 96.791 -72.123 51.624
306011 96.839 -72.120 51.603
306012 96.889 -72.117 51.580
306013 96.938 -72.114 51.560
306014 96.979 -72.111 51.538
306015 97.000 -72.106 51.516
306016 97.013 -72.111 51.492
306017 97.015 -72.136 51.469
306018 97.057 -72.143 51.441
306019 97.114 -72.159 51.413
306020 97.162 -72.170 51.385
306021 97.197 -72.172 51.362
306022 97.231 -72.174 51.340
306023 97.261 -72.174 51.319
306024 97.295 -72.180 51.299
306025 97.334 -72.196 51.277
306026 97.374 -72.213 51.255
306027 97.407 -72.216 51.234
306028 97.441 -72.218 51.213
306029 97.474 -72.221 51.191
306030 97.507 -72.222 51.164
306031 97.540 -72.224 51.133
306032 97.573 -72.227 51.098
306033 97.606 -72.228 51.060
306034 97.639 -72.230 51.020
306035 97.672 -72.233 50.978
306036 97.705 -72.234 50.933
306037 97.738 -72.236 50.887
306038 97.771 -72.239 50.835
306039 97.804 -72.240 50.782
306040 97.837 -72.243 50.730
306041 97.867 -72.246 50.681
306042 97.897 -72.249 50.631
306043 97.928 -72.251 50.579
306044 97.958 -72.254 50.526
306045 97.989 -72.257 50.472
306046 98.008 -72.268 50.418
306047 98.010 -72.290 50.363
306048 98.015 -72.311 50.308
306049 98.074 -72.318 50.254
306050 98.134 -72.325 50.199
306051 98.193 -72.332 50.143
306052 98.250 -72.340 50.088
306053 98.291 -72.354 50.034
306054 98.332 -72.368 49.986
306055 98.372 -72.382 49.936
306056 98.413 -72.396 49.885
306057 98.451 -72.412 49.833
306058 98.490 -72.428 49.778
306059 98.528 -72.443 49.724
306060 98.567 -72.460 49.671
306061 98.558 -72.481 49.617
306062 98.445 -72.522 49.565
306063 98.508 -72.529 49.510
306064 98.575 -72.533 49.458
306065 98.643 -72.537 49.407
306066 98.710 -72.541 49.356
306067 98.743 -72.543 49.306
306068 98.743 -72.541 49.257
306069 98.759 -72.550 49.208
306070 98.788 -72.568 49.160
306071 98.817 -72.584 49.112
306072 98.854 -72.603 49.065
306073 98.913 -72.624 49.018
306074 98.955 -72.641 48.971
306075 98.966 -72.651 48.922
306076 98.984 -72.663 48.873
306077 99.003 -72.675 48.825
306078 99.023 -72.688 48.777
306079 99.043 -72.700 48.729
306080 99.063 -72.712 48.682
306081 99.084 -72.725 48.634
306082 99.118 -72.745 48.585
306083 99.152 -72.766 48.535
306084 99.186 -72.785 48.485
306085 99.220 -72.806 48.436
306086 99.254 -72.826 48.387
306087 99.287 -72.846 48.337
306088 99.321 -72.866 48.289
306089 99.355 -72.886 48.242
306090 99.392 -72.901 48.195
306091 99.431 -72.908 48.149
306092 99.471 -72.916 48.104
306093 99.510 -72.923 48.057
306094 99.550 -72.931 48.012
306095 99.589 -72.938 47.967
306096 99.629 -72.946 47.922
306097 99.666 -72.957 47.877
306098 99.687 -72.992 47.832
306099 99.712 -73.023 47.787
306100 99.743 -73.043 47.742
306101 99.774 -73.064 47.697
306102 99.806 -73.084 47.651
306103 99.837 -73.105 47.606
306104 99.868 -73.125 47.562
306105 99.900 -73.145 47.517
306106 99.931 -73.166 47.472
306107 99.963 -73.187 47.427
306108 99.993 -73.205 47.380
306109 100.025 -73.211 47.334
306110 100.056 -73.216 47.289
306111 100.087 -73.221 47.245
306112 100.119 -73.227 47.199
306113 100.151 -73.232 47.154
306114 100.182 -73.236 47.109
306115 100.213 -73.241 47.063
306116 100.245 -73.245 47.018
306117 100.276 -73.249 46.973
306118 100.308 -73.253 46.930
306119 100.339 -73.257 46.888
306120 100.370 -73.261 46.846
306121 100.402 -73.264 46.804
306122 100.433 -73.269 46.764
306123 100.465 -73.273 46.725
306124 100.496 -73.277 46.689
306125 100.527 -73.280 46.656
306126 100.556 -73.273 46.623
306127 100.580 -73.252 46.592
306128 100.604 -73.233 46.562
306129 100.627 -73.220 46.534
306130 100.650 -73.206 46.506
306131 100.674 -73.192 46.478
306132 100.697 -73.179 46.447
306133 100.720 -73.165 46.415
306134 100.749 -73.158 46.381
306135 100.782 -73.155 46.346
306136 100.818 -73.163 46.311
306137 100.857 -73.180 46.276
306138 100.897 -73.197 46.242
306139 100.936 -73.214 46.207
306140 100.976 -73.231 46.171
306141 101.011 -73.241 46.136
306142 101.044 -73.246 46.102
306143 101.077 -73.251 46.068
306144 101.116 -73.255 46.034
306145 101.156 -73.261 46.000
306146 101.196 -73.265 45.967
306147 101.236 -73.271 45.933
306148 101.275 -73.275 45.898
306149 101.314 -73.284 45.863
306150 101.352 -73.295 45.829
306151 101.391 -73.305 45.795
306152 101.429 -73.317 45.760
306153 101.468 -73.327 45.725
306154 101.506 -73.338 45.690
306155 101.544 -73.349 45.654
306156 101.575 -73.371 45.618
306157 101.604 -73.398 45.581
306158 101.632 -73.424 45.543
306159 101.661 -73.452 45.506
306160 101.689 -73.478 45.468
306161 101.717 -73.505 45.431
306162 101.746 -73.529 45.395
306163 101.775 -73.553 45.359
306164 101.804 -73.578 45.323
306165 101.833 -73.603 45.287
306166 101.868 -73.621 45.251
306167 101.909 -73.632 45.216
306168 101.952 -73.641 45.182
306169 101.995 -73.651 45.149
306170 102.038 -73.661 45.117
306171 102.081 -73.670 45.085
306172 102.124 -73.680 45.054
306173 102.150 -73.681 45.023
306174 102.175 -73.681 44.992
306175 102.199 -73.682 44.962
306176 102.224 -73.683 44.933
306177 102.249 -73.683 44.904
306178 102.274 -73.684 44.876
306659 95.108 -71.777 52.508
306660 95.134 -71.791 52.486
306661 95.161 -71.806 52.466
306662 95.187 -71.820 52.446
306663 95.213 -71.835 52.425
306664 95.240 -71.849 52.405
306665 95.266 -71.864 52.384
306666 95.293 -71.878 52.364
306667 95.319 -71.892 52.346
306668 95.345 -71.906 52.326
306669 95.320 -71.934 52.306
306670 95.325 -71.988 52.287
306671 95.332 -72.022 52.267
306672 95.328 -72.028 52.247
306673 95.367 -72.025 52.228
306674 95.405 -72.022 52.210
306675 95.455 -72.018 52.192
306676 95.507 -72.015 52.175
306677 95.559 -72.012 52.157
306678 95.598 -72.009 52.140
306679 95.622 -72.015 52.122
306680 95.652 -72.036 52.104
306681 95.681 -72.058 52.088
306682 95.711 -72.080 52.073
306683 95.741 -72.102 52.060
306684 95.770 -72.124 52.047
306685 95.800 -72.138 52.036
306686 95.845 -72.147 52.026
306687 95.891 -72.156 52.015
306688 95.937 -72.164 52.005
306689 95.982 -72.174 51.994
306690 96.016 -72.137 51.984
306691 96.065 -72.121 51.973
306692 96.114 -72.104 51.962
306693 96.164 -72.087 51.948
306694 96.204 -72.077 51.934
306695 96.243 -72.070 51.920
306696 96.281 -72.061 51.906
306697 96.319 -72.054 51.893
306698 96.358 -72.046 51.879
306699 96.395 -72.039 51.863
306700 96.431 -72.033 51.845
306701 96.462 -72.030 51.825
306702 96.493 -72.026 51.805
306703 96.524 -72.023 51.785
306704 96.554 -72.026 51.767
306705 96.583 -72.042 51.749
306706 96.612 -72.057 51.727
306707 96.640 -72.072 51.705
306708 96.669 -72.087 51.683
306709 96.702 -72.098 51.661
306710 96.751 -72.095 51.643
306711 96.800 -72.092 51.625
306712 96.850 -72.089 51.606
306713 96.898 -72.086 51.587
306714 96.947 -72.083 51.568
306715 96.976 -72.080 51.547
306716 96.997 -72.075 51.525
306717 97.017 -72.071 51.503
306718 97.015 -72.098 51.479
306719 97.054 -72.103 51.451
306720 97.105 -72.115 51.422
306721 97.163 -72.131 51.392
306722 97.213 -72.143 51.365
306723 97.248 -72.146 51.340
306724 97.279 -72.146 51.318
306725 97.315 -72.157 51.296
306726 97.354 -72.172 51.273
306727 97.387 -72.174 51.250
306728 97.420 -72.176 51.227
306729 97.453 -72.178 51.204
306730 97.487 -72.180 51.180
306731 97.519 -72.182 51.155
306732 97.552 -72.184 51.125
306733 97.586 -72.186 51.092
306734 97.618 -72.188 51.055
306735 97.652 -72.190 51.014
306736 97.684 -72.193 50.970
306737 97.718 -72.194 50.925
306738 97.751 -72.196 50.877
306739 97.784 -72.199 50.826
306740 97.817 -72.200 50.774
306741 97.850 -72.202 50.722
306742 97.881 -72.205 50.673
306743 97.912 -72.208 50.622
306744 97.942 -72.211 50.571
306745 97.972 -72.214 50.518
306746 97.979 -72.233 50.463
306747 97.981 -72.254 50.409
306748 97.984 -72.276 50.355
306749 98.005 -72.293 50.301
306750 98.048 -72.300 50.247
306751 98.104 -72.310 50.194
306752 98.161 -72.321 50.140
306753 98.219 -72.328 50.087
306754 98.279 -72.336 50.035
306755 98.338 -72.343 49.986
306756 98.390 -72.352 49.935
306757 98.431 -72.367 49.883
306758 98.469 -72.382 49.831
306759 98.505 -72.397 49.777
306760 98.536 -72.409 49.722
306761 98.567 -72.420 49.669
306762 98.504 -72.451 49.617
306763 98.492 -72.476 49.565
306764 98.540 -72.493 49.514
306765 98.591 -72.506 49.463
306766 98.659 -72.510 49.412
306767 98.726 -72.515 49.362
306768 98.752 -72.515 49.314
306769 98.765 -72.522 49.264
306770 98.794 -72.539 49.215
306771 98.823 -72.556 49.167
306772 98.852 -72.573 49.118
306773 98.881 -72.591 49.071
306774 98.929 -72.609 49.024
306775 98.963 -72.625 48.976
306776 98.979 -72.637 48.927
306777 98.999 -72.649 48.878
306778 99.019 -72.661 48.828
306779 99.039 -72.673 48.779
306780 99.059 -72.685 48.730
306781 99.078 -72.698 48.683
306782 99.098 -72.710 48.635
306783 99.124 -72.726 48.586
306784 99.158 -72.746 48.536
306785 99.192 -72.766 48.486
306786 99.226 -72.786 48.437
306787 99.260 -72.807 48.388
306788 99.293 -72.827 48.339
306789 99.328 -72.847 48.290
306790 99.364 -72.861 48.243
306791 99.404 -72.869 48.196
306792 99.443 -72.876 48.150
306793 99.483 -72.884 48.104
306794 99.522 -72.891 48.058
306795 99.562 -72.898 48.012
306796 99.602 -72.906 47.967
306797 99.641 -72.913 47.922
306798 99.681 -72.921 47.877
306799 99.704 -72.952 47.833
306800 99.725 -72.989 47.788
306801 99.754 -73.011 47.743
306802 99.785 -73.032 47.698
306803 99.817 -73.053 47.653
306804 99.848 -73.073 47.607
306805 99.879 -73.094 47.563
306806 99.911 -73.114 47.518
306807 99.942 -73.135 47.473
306808 99.974 -73.155 47.429
306809 100.005 -73.173 47.382
306810 100.037 -73.167 47.336
306811 100.068 -73.171 47.291
306812 100.100 -73.176 47.247
306813 100.131 -73.180 47.201
306814 100.162 -73.183 47.157
306815 100.194 -73.187 47.112
306816 100.225 -73.192 47.066
306817 100.256 -73.195 47.021
306818 100.288 -73.199 46.977
306819 100.319 -73.203 46.934
306820 100.350 -73.208 46.891
306821 100.382 -73.211 46.849
306822 100.413 -73.215 46.807
306823 100.444 -73.219 46.767
306824 100.476 -73.224 46.728
306825 100.507 -73.227 46.692
306826 100.540 -73.236 46.659
306827 100.573 -73.239 46.625
306828 100.597 -73.218 46.593
306829 100.621 -73.198 46.562
306830 100.645 -73.177 46.534
306831 100.670 -73.155 46.505
306832 100.694 -73.134 46.476
306833 100.718 -73.114 46.445
306834 100.742 -73.098 46.414
306835 100.766 -73.087 46.380
306836 100.804 -73.096 46.344
306837 100.843 -73.113 46.308
306838 100.882 -73.130 46.273
306839 100.921 -73.147 46.239
306840 100.961 -73.164 46.203
306841 101.000 -73.181 46.168
306842 101.035 -73.190 46.135
306843 101.068 -73.195 46.100
306844 101.101 -73.199 46.065
306845 101.137 -73.205 46.032
306846 101.177 -73.209 45.999
306847 101.216 -73.214 45.966
306848 101.256 -73.219 45.932
306849 101.296 -73.224 45.897
306850 101.334 -73.235 45.863
306851 101.373 -73.246 45.829
306852 101.411 -73.257 45.795
306853 101.450 -73.267 45.760
306854 101.488 -73.278 45.725
306855 101.525 -73.291 45.689
306856 101.553 -73.318 45.652
306857 101.582 -73.345 45.616
306858 101.610 -73.371 45.579
306859 101.638 -73.398 45.541
306860 101.667 -73.425 45.504
306861 101.695 -73.452 45.466
306862 101.724 -73.477 45.429
306863 101.752 -73.502 45.391
306864 101.781 -73.526 45.355
306865 101.810 -73.551 45.319
306866 101.840 -73.574 45.282
306867 101.880 -73.587 45.245
306868 101.922 -73.598 45.210
306869 101.965 -73.607 45.176
306870 102.009 -73.617 45.143
306871 102.051 -73.627 45.111
306872 102.095 -73.636 45.079
306873 102.133 -73.644 45.048
306874 102.158 -73.644 45.017
306875 102.182 -73.645 44.986
306876 102.207 -73.646 44.955
306877 102.232 -73.646 44.926
306878 102.257 -73.647 44.898
307360 95.106 -71.732 52.504
307361 95.132 -71.747 52.483
307362 95.159 -71.760 52.463
307363 95.185 -71.775 52.443
307364 95.211 -71.789 52.422
307365 95.238 -71.804 52.402
307366 95.264 -71.818 52.382
307367 95.290 -71.832 52.363
307368 95.317 -71.847 52.345
307369 95.326 -71.866 52.325
307370 95.298 -71.894 52.306
307371 95.309 -71.952 52.288
307372 95.317 -71.990 52.269
307373 95.308 -71.997 52.249
307374 95.349 -71.994 52.230
307375 95.401 -71.991 52.212
307376 95.453 -71.987 52.195
307377 95.506 -71.984 52.178
307378 95.557 -71.981 52.161
307379 95.595 -71.978 52.143
307380 95.617 -71.975 52.127
307381 95.646 -71.995 52.109
307382 95.675 -72.017 52.093
307383 95.705 -72.039 52.079
307384 95.735 -72.061 52.065
307385 95.764 -72.083 52.052
307386 95.804 -72.090 52.042
307387 95.849 -72.099 52.031
307388 95.895 -72.108 52.021
307389 95.940 -72.117 52.012
307390 95.986 -72.126 52.003
307391 96.021 -72.098 51.993
307392 96.071 -72.081 51.984
307393 96.119 -72.064 51.973
307394 96.169 -72.048 51.961
307395 96.218 -72.031 51.947
307396 96.257 -72.022 51.933
307397 96.295 -72.014 51.920
307398 96.333 -72.007 51.907
307399 96.372 -71.999 51.894
307400 96.410 -71.992 51.878
307401 96.444 -71.986 51.860
307402 96.475 -71.983 51.840
307403 96.507 -71.980 51.820
307404 96.537 -71.980 51.801
307405 96.566 -71.995 51.780
307406 96.594 -72.010 51.760
307407 96.622 -72.025 51.736
307408 96.651 -72.040 51.712
307409 96.679 -72.056 51.688
307410 96.712 -72.067 51.665
307411 96.761 -72.064 51.645
307412 96.810 -72.061 51.627
307413 96.859 -72.058 51.611
307414 96.908 -72.056 51.593
307415 96.952 -72.052 51.577
307416 96.972 -72.048 51.557
307417 96.993 -72.044 51.536
307418 97.013 -72.040 51.515
307419 97.016 -72.059 51.491
307420 97.054 -72.064 51.463
307421 97.097 -72.071 51.432
307422 97.154 -72.087 51.401
307423 97.211 -72.103 51.369
307424 97.263 -72.117 51.341
307425 97.296 -72.117 51.316
307426 97.334 -72.130 51.292
307427 97.367 -72.132 51.268
307428 97.400 -72.134 51.245
307429 97.433 -72.136 51.220
307430 97.466 -72.138 51.195
307431 97.499 -72.140 51.169
307432 97.532 -72.142 51.144
307433 97.565 -72.144 51.115
307434 97.598 -72.146 51.083
307435 97.631 -72.149 51.047
307436 97.664 -72.150 51.006
307437 97.697 -72.152 50.963
307438 97.730 -72.155 50.918
307439 97.764 -72.156 50.870
307440 97.796 -72.158 50.819
307441 97.829 -72.161 50.767
307442 97.863 -72.162 50.715
307443 97.895 -72.164 50.665
307444 97.926 -72.168 50.614
307445 97.948 -72.175 50.563
307446 97.950 -72.197 50.511
307447 97.964 -72.216 50.456
307448 97.980 -72.234 50.402
307449 97.999 -72.249 50.348
307450 98.033 -72.265 50.294
307451 98.058 -72.270 50.241
307452 98.098 -72.275 50.189
307453 98.155 -72.285 50.136
307454 98.211 -72.295 50.085
307455 98.267 -72.305 50.034
307456 98.323 -72.315 49.985
307457 98.380 -72.325 49.933
307458 98.431 -72.334 49.881
307459 98.473 -72.344 49.829
307460 98.503 -72.356 49.776
307461 98.534 -72.367 49.722
307462 98.565 -72.379 49.670
307463 98.492 -72.414 49.619
307464 98.539 -72.430 49.568
307465 98.587 -72.446 49.520
307466 98.635 -72.463 49.469
307467 98.683 -72.479 49.418
307468 98.742 -72.488 49.369
307469 98.770 -72.493 49.321
307470 98.799 -72.511 49.272
307471 98.829 -72.528 49.223
307472 98.857 -72.545 49.174
307473 98.886 -72.562 49.126
307474 98.916 -72.579 49.078
307475 98.945 -72.596 49.029
307476 98.975 -72.610 48.981
307477 98.995 -72.623 48.931
307478 99.015 -72.635 48.881
307479 99.034 -72.647 48.831
307480 99.054 -72.660 48.781
307481 99.074 -72.672 48.732
307482 99.094 -72.684 48.683
307483 99.114 -72.696 48.635
307484 99.134 -72.709 48.587
307485 99.164 -72.727 48.537
307486 99.198 -72.747 48.487
307487 99.232 -72.767 48.438
307488 99.266 -72.788 48.388
307489 99.300 -72.807 48.340
307490 99.337 -72.821 48.291
307491 99.376 -72.829 48.244
307492 99.416 -72.836 48.197
307493 99.455 -72.844 48.151
307494 99.495 -72.851 48.105
307495 99.535 -72.859 48.059
307496 99.574 -72.866 48.013
307497 99.614 -72.873 47.967
307498 99.653 -72.881 47.922
307499 99.693 -72.888 47.878
307500 99.722 -72.912 47.834
307501 99.742 -72.948 47.790
307502 99.765 -72.980 47.744
307503 99.797 -73.001 47.699
307504 99.828 -73.021 47.654
307505 99.859 -73.042 47.609
307506 99.891 -73.063 47.565
307507 99.922 -73.083 47.520
307508 99.954 -73.104 47.475
307509 99.985 -73.124 47.430
307510 100.016 -73.136 47.384
307511 100.048 -73.126 47.339
307512 100.079 -73.122 47.294
307513 100.111 -73.126 47.250
307514 100.142 -73.130 47.205
307515 100.174 -73.134 47.161
307516 100.205 -73.138 47.116
307517 100.237 -73.142 47.071
307518 100.268 -73.146 47.026
307519 100.299 -73.150 46.982
307520 100.331 -73.154 46.940
307521 100.362 -73.158 46.897
307522 100.393 -73.162 46.854
307523 100.424 -73.166 46.812
307524 100.456 -73.170 46.771
307525 100.487 -73.174 46.732
307526 100.519 -73.178 46.695
307527 100.554 -73.191 46.661
307528 100.591 -73.207 46.629
307529 100.615 -73.186 46.597
307530 100.639 -73.164 46.565
307531 100.663 -73.143 46.534
307532 100.687 -73.123 46.504
307533 100.711 -73.101 46.474
307534 100.736 -73.080 46.444
307535 100.760 -73.059 46.411
307536 100.788 -73.045 46.377
307537 100.828 -73.045 46.342
307538 100.867 -73.063 46.306
307539 100.906 -73.079 46.272
307540 100.946 -73.097 46.237
307541 100.985 -73.114 46.201
307542 101.025 -73.131 46.167
307543 101.059 -73.139 46.133
307544 101.092 -73.144 46.098
307545 101.124 -73.148 46.064
307546 101.158 -73.153 46.031
307547 101.197 -73.158 45.998
307548 101.237 -73.163 45.965
307549 101.277 -73.168 45.931
307550 101.316 -73.176 45.897
307551 101.354 -73.186 45.862
307552 101.393 -73.197 45.829
307553 101.431 -73.208 45.794
307554 101.470 -73.218 45.760
307555 101.502 -73.238 45.725
307556 101.531 -73.264 45.689
307557 101.559 -73.291 45.651
307558 101.588 -73.318 45.614
307559 101.616 -73.345 45.577
307560 101.644 -73.372 45.539
307561 101.673 -73.399 45.501
307562 101.701 -73.426 45.464
307563 101.730 -73.450 45.426
307564 101.759 -73.475 45.388
307565 101.787 -73.499 45.351
307566 101.816 -73.524 45.314
307567 101.853 -73.540 45.277
307568 101.893 -73.554 45.240
307569 101.936 -73.564 45.206
307570 101.978 -73.574 45.171
307571 102.022 -73.583 45.137
307572 102.064 -73.593 45.106
307573 102.108 -73.602 45.074
307574 102.140 -73.607 45.043
307575 102.166 -73.608 45.012
307576 102.190 -73.608 44.981
307577 102.215 -73.609 44.951
307578 102.240 -73.609 44.921
308061 95.104 -71.687 52.500
308062 95.130 -71.701 52.479
308063 95.157 -71.716 52.460
308064 95.183 -71.730 52.440
308065 95.209 -71.744 52.420
308066 95.236 -71.759 52.400
308067 95.262 -71.773 52.380
308068 95.288 -71.788 52.361
308069 95.315 -71.802 52.343
308070 95.304 -71.826 52.325
308071 95.275 -71.854 52.306
308072 95.291 -71.917 52.288
308073 95.301 -71.958 52.269
308074 95.296 -71.964 52.250
308075 95.347 -71.963 52.232
308076 95.399 -71.960 52.214
308077 95.451 -71.956 52.197
308078 95.503 -71.953 52.180
308079 95.555 -71.950 52.163
308080 95.592 -71.947 52.147
308081 95.614 -71.944 52.131
308082 95.640 -71.955 52.114
308083 95.670 -71.976 52.099
308084 95.699 -71.998 52.084
308085 95.729 -72.020 52.070
308086 95.762 -72.034 52.058
308087 95.808 -72.042 52.047
308088 95.853 -72.052 52.038
308089 95.899 -72.061 52.028
308090 95.944 -72.070 52.019
308091 95.989 -72.079 52.011
308092 96.026 -72.058 52.002
308093 96.075 -72.042 51.994
308094 96.124 -72.025 51.985
308095 96.174 -72.008 51.973
308096 96.223 -71.992 51.960
308097 96.271 -71.975 51.948
308098 96.310 -71.967 51.936
308099 96.348 -71.959 51.923
308100 96.386 -71.951 51.910
308101 96.424 -71.944 51.895
308102 96.457 -71.939 51.878
308103 96.488 -71.936 51.859
308104 96.520 -71.933 51.841
308105 96.548 -71.948 51.821
308106 96.577 -71.963 51.799
308107 96.605 -71.979 51.776
308108 96.633 -71.994 51.750
308109 96.662 -72.009 51.723
308110 96.690 -72.024 51.697
308111 96.722 -72.037 51.672
308112 96.771 -72.034 51.650
308113 96.820 -72.031 51.633
308114 96.869 -72.028 51.616
308115 96.918 -72.025 51.600
308116 96.948 -72.022 51.585
308117 96.969 -72.017 51.568
308118 96.990 -72.014 51.549
308119 97.010 -72.009 51.529
308120 97.020 -72.018 51.505
308121 97.054 -72.026 51.475
308122 97.093 -72.030 51.443
308123 97.145 -72.043 51.409
308124 97.203 -72.059 51.376
308125 97.259 -72.075 51.344
308126 97.312 -72.090 51.314
308127 97.346 -72.092 51.288
308128 97.379 -72.094 51.263
308129 97.412 -72.096 51.238
308130 97.445 -72.098 51.212
308131 97.478 -72.100 51.184
308132 97.511 -72.102 51.158
308133 97.544 -72.104 51.132
308134 97.577 -72.105 51.103
308135 97.610 -72.108 51.071
308136 97.643 -72.110 51.037
308137 97.676 -72.111 50.998
308138 97.709 -72.114 50.956
308139 97.742 -72.115 50.912
308140 97.776 -72.117 50.866
308141 97.809 -72.119 50.814
308142 97.842 -72.121 50.762
308143 97.875 -72.123 50.709
308144 97.908 -72.125 50.657
308145 97.930 -72.138 50.607
308146 97.945 -72.155 50.558
308147 97.961 -72.174 50.506
308148 97.977 -72.192 50.452
308149 97.992 -72.210 50.396
308150 98.034 -72.202 50.341
308151 98.062 -72.238 50.288
308152 98.086 -72.244 50.235
308153 98.111 -72.245 50.185
308154 98.149 -72.250 50.133
308155 98.205 -72.260 50.082
308156 98.261 -72.270 50.032
308157 98.317 -72.280 49.982
308158 98.365 -72.289 49.930
308159 98.411 -72.296 49.877
308160 98.461 -72.305 49.825
308161 98.501 -72.315 49.776
308162 98.532 -72.326 49.726
308163 98.512 -72.348 49.676
308164 98.539 -72.368 49.626
308165 98.587 -72.384 49.576
308166 98.634 -72.400 49.526
308167 98.682 -72.417 49.475
308168 98.726 -72.434 49.424
308169 98.758 -72.450 49.375
308170 98.789 -72.468 49.327
308171 98.817 -72.484 49.279
308172 98.846 -72.500 49.230
308173 98.874 -72.517 49.182
308174 98.903 -72.533 49.133
308175 98.932 -72.550 49.084
308176 98.966 -72.569 49.035
308177 98.988 -72.587 48.985
308178 99.009 -72.605 48.935
308179 99.030 -72.621 48.884
308180 99.050 -72.633 48.834
308181 99.070 -72.646 48.784
308182 99.090 -72.658 48.734
308183 99.110 -72.670 48.685
308184 99.130 -72.682 48.637
308185 99.149 -72.694 48.588
308186 99.171 -72.708 48.537
308187 99.205 -72.728 48.488
308188 99.238 -72.748 48.438
308189 99.272 -72.768 48.388
308190 99.309 -72.782 48.339
308191 99.349 -72.789 48.291
308192 99.389 -72.797 48.244
308193 99.428 -72.804 48.197
308194 99.468 -72.811 48.151
308195 99.507 -72.819 48.106
308196 99.547 -72.826 48.060
308197 99.586 -72.834 48.014
308198 99.626 -72.841 47.969
308199 99.665 -72.849 47.923
308200 99.705 -72.856 47.878
308201 99.739 -72.872 47.834
308202 99.760 -72.908 47.790
308203 99.780 -72.944 47.745
308204 99.808 -72.970 47.700
308205 99.839 -72.990 47.655
308206 99.870 -73.011 47.609
308207 99.902 -73.031 47.566
308208 99.933 -73.052 47.522
308209 99.965 -73.073 47.477
308210 99.996 -73.093 47.432
308211 100.028 -73.098 47.386
308212 100.059 -73.089 47.341
308213 100.091 -73.079 47.297
308214 100.122 -73.076 47.253
308215 100.154 -73.081 47.209
308216 100.185 -73.085 47.165
308217 100.216 -73.089 47.120
308218 100.248 -73.092 47.075
308219 100.279 -73.097 47.031
308220 100.311 -73.101 46.988
308221 100.342 -73.105 46.945
308222 100.373 -73.108 46.902
308223 100.405 -73.113 46.860
308224 100.436 -73.117 46.817
308225 100.467 -73.120 46.776
308226 100.499 -73.125 46.737
308227 100.530 -73.129 46.699
308228 100.567 -73.147 46.665
308229 100.607 -73.170 46.632
308230 100.632 -73.152 46.600
308231 100.656 -73.132 46.568
308232 100.680 -73.111 46.536
308233 100.704 -73.089 46.505
308234 100.728 -73.068 46.474
308235 100.753 -73.048 46.443
308236 100.777 -73.026 46.410
308237 100.813 -73.022 46.375
308238 100.852 -73.022 46.340
308239 100.892 -73.022 46.304
308240 100.931 -73.029 46.270
308241 100.970 -73.047 46.234
308242 101.009 -73.064 46.200
308243 101.049 -73.080 46.165
308244 101.083 -73.088 46.130
308245 101.116 -73.092 46.096
308246 101.148 -73.098 46.063
308247 101.181 -73.102 46.030
308248 101.218 -73.107 45.997
308249 101.258 -73.112 45.964
308250 101.298 -73.117 45.930
308251 101.336 -73.126 45.896
308252 101.375 -73.137 45.862
308253 101.413 -73.148 45.828
308254 101.452 -73.158 45.794
308255 101.480 -73.184 45.760
308256 101.509 -73.211 45.724
308257 101.537 -73.238 45.688
308258 101.565 -73.265 45.650
308259 101.594 -73.292 45.613
308260 101.622 -73.319 45.575
308261 101.650 -73.346 45.537
308262 101.679 -73.372 45.499
308263 101.707 -73.399 45.461
308264 101.736 -73.424 45.423
308265 101.765 -73.448 45.385
308266 101.794 -73.472 45.347
308267 101.826 -73.494 45.309
308268 101.866 -73.507 45.272
308269 101.906 -73.520 45.236
308270 101.949 -73.530 45.202
308271 101.991 -73.540 45.167
308272 102.035 -73.549 45.133
308273 102.078 -73.559 45.101
308274 102.121 -73.568 45.069
308275 102.149 -73.570 45.037
308276 102.173 -73.571 45.007
308277 102.198 -73.571 44.977
308278 102.223 -73.572 44.946
308762 95.102 -71.642 52.496
308763 95.128 -71.657 52.475
308764 95.155 -71.671 52.456
308765 95.181 -71.685 52.437
308766 95.207 -71.699 52.417
308767 95.234 -71.713 52.398
308768 95.260 -71.728 52.378
308769 95.286 -71.742 52.360
308770 95.309 -71.757 52.342
308771 95.281 -71.786 52.324
308772 95.253 -71.814 52.306
308773 95.275 -71.881 52.288
308774 95.293 -71.916 52.270
308775 95.299 -71.917 52.251
308776 95.347 -71.926 52.233
308777 95.397 -71.928 52.216
308778 95.449 -71.925 52.199
308779 95.501 -71.922 52.183
308780 95.553 -71.919 52.166
308781 95.588 -71.916 52.150
308782 95.610 -71.913 52.135
308783 95.634 -71.914 52.119
308784 95.664 -71.936 52.104
308785 95.693 -71.958 52.089
308786 95.723 -71.980 52.076
308787 95.766 -71.986 52.064
308788 95.812 -71.995 52.053
308789 95.857 -72.004 52.044
308790 95.902 -72.013 52.035
308791 95.948 -72.022 52.027
308792 95.993 -72.031 52.020
308793 96.031 -72.019 52.012
308794 96.080 -72.002 52.005
308795 96.129 -71.986 51.998
308796 96.178 -71.969 51.988
308797 96.228 -71.951 51.977
308798 96.277 -71.935 51.965
308799 96.324 -71.920 51.953
308800 96.362 -71.912 51.940
308801 96.400 -71.904 51.925
308802 96.439 -71.897 51.910
308803 96.470 -71.893 51.893
308804 96.502 -71.889 51.876
308805 96.531 -71.901 51.859
308806 96.559 -71.917 51.840
308807 96.588 -71.932 51.818
308808 96.616 -71.947 51.794
308809 96.645 -71.962 51.766
308810 96.673 -71.977 51.737
308811 96.701 -71.992 51.708
308812 96.732 -72.007 51.681
308813 96.781 -72.004 51.658
308814 96.829 -72.001 51.640
308815 96.879 -71.998 51.623
308816 96.925 -71.995 51.605
308817 96.945 -71.991 51.590
308818 96.965 -71.986 51.576
308819 96.986 -71.983 51.561
308820 97.007 -71.978 51.543
308821 97.025 -71.976 51.520
308822 97.054 -71.987 51.489
308823 97.093 -71.992 51.454
308824 97.137 -71.999 51.419
308825 97.193 -72.015 51.384
308826 97.253 -72.034 51.349
308827 97.317 -72.057 51.317
308828 97.356 -72.065 51.288
308829 97.385 -72.067 51.258
308830 97.418 -72.070 51.231
308831 97.451 -72.071 51.203
308832 97.484 -72.073 51.176
308833 97.517 -72.075 51.149
308834 97.550 -72.077 51.123
308835 97.583 -72.079 51.094
308836 97.616 -72.081 51.063
308837 97.649 -72.083 51.030
308838 97.682 -72.085 50.993
308839 97.715 -72.087 50.952
308840 97.748 -72.089 50.908
308841 97.782 -72.091 50.863
308842 97.815 -72.092 50.810
308843 97.848 -72.095 50.758
308844 97.881 -72.096 50.706
308845 97.914 -72.099 50.654
308846 97.937 -72.114 50.602
308847 97.957 -72.132 50.552
308848 97.973 -72.150 50.502
308849 97.989 -72.168 50.450
308850 98.025 -72.165 50.397
308851 98.069 -72.158 50.343
308852 98.090 -72.211 50.291
308853 98.114 -72.219 50.241
308854 98.139 -72.220 50.193
308855 98.163 -72.221 50.142
308856 98.199 -72.224 50.089
308857 98.255 -72.235 50.035
308858 98.299 -72.243 49.983
308859 98.344 -72.251 49.931
308860 98.390 -72.258 49.880
308861 98.441 -72.267 49.831
308862 98.492 -72.274 49.783
308863 98.530 -72.285 49.733
308864 98.538 -72.305 49.683
308865 98.586 -72.321 49.633
308866 98.627 -72.338 49.582
308867 98.659 -72.356 49.532
308868 98.692 -72.373 49.481
308869 98.724 -72.390 49.430
308870 98.757 -72.407 49.381
308871 98.789 -72.424 49.333
308872 98.819 -72.441 49.284
308873 98.848 -72.458 49.236
308874 98.877 -72.475 49.187
308875 98.912 -72.493 49.138
308876 98.945 -72.513 49.088
308877 98.977 -72.532 49.039
308878 98.998 -72.550 48.988
308879 99.019 -72.568 48.937
308880 99.040 -72.586 48.887
308881 99.062 -72.603 48.837
308882 99.083 -72.622 48.786
308883 99.104 -72.639 48.737
308884 99.125 -72.656 48.687
308885 99.145 -72.669 48.638
308886 99.165 -72.681 48.589
308887 99.185 -72.693 48.538
308888 99.211 -72.709 48.489
308889 99.244 -72.728 48.439
308890 99.282 -72.741 48.389
308891 99.322 -72.749 48.340
308892 99.361 -72.757 48.292
308893 99.401 -72.764 48.244
308894 99.440 -72.772 48.198
308895 99.480 -72.779 48.152
308896 99.519 -72.786 48.106
308897 99.559 -72.794 48.060
308898 99.598 -72.801 48.015
308899 99.638 -72.809 47.970
308900 99.678 -72.816 47.924
308901 99.717 -72.824 47.879
308902 99.757 -72.832 47.835
308903 99.777 -72.868 47.790
308904 99.798 -72.904 47.745
308905 99.819 -72.938 47.701
308906 99.850 -72.959 47.655
308907 99.881 -72.979 47.610
308908 99.913 -73.000 47.567
308909 99.944 -73.020 47.523
308910 99.976 -73.042 47.479
308911 100.007 -73.062 47.434
308912 100.039 -73.060 47.388
308913 100.070 -73.051 47.343
308914 100.102 -73.042 47.300
308915 100.134 -73.032 47.256
308916 100.165 -73.032 47.213
308917 100.197 -73.036 47.169
308918 100.228 -73.039 47.124
308919 100.259 -73.043 47.080
308920 100.290 -73.048 47.036
308921 100.322 -73.051 46.993
308922 100.353 -73.055 46.950
308923 100.385 -73.059 46.908
308924 100.416 -73.064 46.866
308925 100.447 -73.067 46.824
308926 100.479 -73.071 46.782
308927 100.510 -73.075 46.743
308928 100.542 -73.079 46.705
308929 100.581 -73.102 46.670
308930 100.620 -73.125 46.636
308931 100.649 -73.120 46.603
308932 100.674 -73.098 46.572
308933 100.698 -73.077 46.540
308934 100.722 -73.057 46.509
308935 100.746 -73.036 46.477
308936 100.770 -73.014 46.443
308937 100.798 -72.999 46.410
308938 100.838 -72.999 46.373
308939 100.876 -72.999 46.337
308940 100.916 -72.999 46.302
308941 100.955 -72.999 46.267
308942 100.994 -72.999 46.232
308943 101.034 -73.014 46.198
308944 101.073 -73.030 46.164
308945 101.107 -73.037 46.128
308946 101.139 -73.042 46.094
308947 101.172 -73.046 46.061
308948 101.205 -73.051 46.029
308949 101.239 -73.056 45.997
308950 101.279 -73.061 45.963
308951 101.318 -73.067 45.929
308952 101.356 -73.077 45.895
308953 101.395 -73.088 45.861
308954 101.430 -73.105 45.827
308955 101.458 -73.131 45.793
308956 101.486 -73.158 45.758
308957 101.515 -73.185 45.722
308958 101.543 -73.212 45.686
308959 101.571 -73.239 45.649
308960 101.600 -73.265 45.611
308961 101.628 -73.293 45.573
308962 101.656 -73.319 45.535
308963 101.685 -73.346 45.496
308964 101.713 -73.372 45.459
308965 101.743 -73.396 45.421
308966 101.771 -73.421 45.382
308967 101.800 -73.445 45.344
308968 101.838 -73.460 45.306
308969 101.878 -73.474 45.269
308970 101.918 -73.487 45.233
308971 101.962 -73.496 45.198
308972 102.004 -73.506 45.163
308973 102.048 -73.515 45.129
308974 102.091 -73.524 45.097
308975 102.131 -73.533 45.065
308976 102.157 -73.534 45.033
308977 102.181 -73.534 45.002
308978 102.206 -73.535 44.973
308979 102.231 -73.536 44.942
309463 95.100 -71.597 52.491
309464 95.126 -71.611 52.472
309465 95.153 -71.625 52.453
309466 95.179 -71.640 52.434
309467 95.205 -71.654 52.414
309468 95.232 -71.669 52.395
309469 95.258 -71.683 52.377
309470 95.284 -71.697 52.358
309471 95.287 -71.718 52.340
309472 95.259 -71.746 52.323
309473 95.231 -71.774 52.306
309474 95.258 -71.845 52.289
309475 95.289 -71.869 52.271
309476 95.301 -71.871 52.252
309477 95.349 -71.880 52.234
309478 95.397 -71.890 52.217
309479 95.447 -71.894 52.201
309480 95.499 -71.891 52.186
309481 95.551 -71.888 52.170
309482 95.585 -71.885 52.154
309483 95.607 -71.882 52.138
309484 95.629 -71.879 52.124
309485 95.658 -71.895 52.109
309486 95.687 -71.917 52.094
309487 95.725 -71.929 52.081
309488 95.770 -71.939 52.070
309489 95.815 -71.948 52.060
309490 95.861 -71.957 52.051
309491 95.906 -71.965 52.043
309492 95.952 -71.974 52.036
309493 95.997 -71.983 52.029
309494 96.038 -71.982 52.023
309495 96.085 -71.963 52.018
309496 96.134 -71.946 52.013
309497 96.183 -71.929 52.004
309498 96.233 -71.912 51.994
309499 96.282 -71.895 51.983
309500 96.331 -71.879 51.971
309501 96.377 -71.865 51.957
309502 96.414 -71.857 51.942
309503 96.452 -71.850 51.925
309504 96.483 -71.846 51.908
309505 96.513 -71.854 51.890
309506 96.542 -71.870 51.874
309507 96.570 -71.885 51.855
309508 96.599 -71.900 51.835
309509 96.627 -71.916 51.810
309510 96.656 -71.931 51.781
309511 96.684 -71.946 51.750
309512 96.713 -71.961 51.719
309513 96.741 -71.976 51.691
309514 96.790 -71.973 51.667
309515 96.839 -71.970 51.647
309516 96.888 -71.967 51.628
309517 96.921 -71.964 51.609
309518 96.942 -71.960 51.594
309519 96.962 -71.955 51.581
309520 96.982 -71.951 51.569
309521 97.003 -71.947 51.554
309522 97.023 -71.943 51.530
309523 97.054 -71.948 51.499
309524 97.093 -71.953 51.464
309525 97.132 -71.958 51.428
309526 97.187 -71.973 51.392
309527 97.251 -71.997 51.356
309528 97.314 -72.020 51.322
309529 97.372 -72.039 51.290
309530 97.399 -72.041 51.258
309531 97.427 -72.043 51.228
309532 97.457 -72.045 51.199
309533 97.490 -72.047 51.171
309534 97.523 -72.048 51.144
309535 97.556 -72.051 51.117
309536 97.589 -72.052 51.087
309537 97.622 -72.055 51.055
309538 97.655 -72.057 51.022
309539 97.688 -72.058 50.986
309540 97.721 -72.061 50.946
309541 97.754 -72.062 50.903
309542 97.788 -72.064 50.857
309543 97.821 -72.066 50.806
309544 97.854 -72.068 50.754
309545 97.887 -72.070 50.702
309546 97.917 -72.076 50.649
309547 97.940 -72.092 50.597
309548 97.963 -72.109 50.547
309549 97.985 -72.127 50.497
309550 98.016 -72.129 50.447
309551 98.061 -72.117 50.396
309552 98.095 -72.130 50.345
309553 98.109 -72.180 50.296
309554 98.132 -72.190 50.249
309555 98.156 -72.191 50.201
309556 98.180 -72.192 50.150
309557 98.206 -72.194 50.096
309558 98.234 -72.197 50.041
309559 98.278 -72.205 49.988
309560 98.322 -72.213 49.935
309561 98.370 -72.221 49.885
309562 98.421 -72.229 49.837
309563 98.472 -72.236 49.790
309564 98.528 -72.248 49.740
309565 98.564 -72.262 49.690
309566 98.596 -72.279 49.639
309567 98.629 -72.296 49.587
309568 98.661 -72.314 49.537
309569 98.694 -72.331 49.486
309570 98.726 -72.348 49.435
309571 98.759 -72.365 49.385
309572 98.791 -72.383 49.337
309573 98.824 -72.399 49.288
309574 98.858 -72.419 49.240
309575 98.891 -72.438 49.191
309576 98.925 -72.457 49.142
309577 98.959 -72.477 49.091
309578 98.986 -72.495 49.041
309579 99.007 -72.513 48.990
309580 99.029 -72.531 48.939
309581 99.050 -72.549 48.889
309582 99.071 -72.567 48.838
309583 99.092 -72.584 48.788
309584 99.114 -72.603 48.738
309585 99.135 -72.620 48.689
309586 99.156 -72.638 48.639
309587 99.178 -72.656 48.589
309588 99.199 -72.674 48.539
309589 99.220 -72.691 48.490
309590 99.255 -72.702 48.440
309591 99.294 -72.710 48.390
309592 99.334 -72.717 48.342
309593 99.373 -72.724 48.293
309594 99.413 -72.732 48.245
309595 99.452 -72.739 48.198
309596 99.492 -72.747 48.152
309597 99.532 -72.754 48.107
309598 99.571 -72.762 48.061
309599 99.611 -72.769 48.015
309600 99.650 -72.776 47.970
309601 99.690 -72.784 47.925
309602 99.730 -72.791 47.881
309603 99.769 -72.799 47.836
309604 99.795 -72.828 47.790
309605 99.815 -72.863 47.745
309606 99.836 -72.899 47.701
309607 99.861 -72.928 47.656
309608 99.893 -72.948 47.611
309609 99.924 -72.969 47.567
309610 99.956 -72.989 47.523
309611 99.987 -73.010 47.480
309612 100.018 -73.030 47.435
309613 100.050 -73.023 47.389
309614 100.082 -73.014 47.345
309615 100.113 -73.004 47.302
309616 100.145 -72.995 47.258
309617 100.177 -72.985 47.216
309618 100.208 -72.986 47.172
309619 100.240 -72.990 47.129
309620 100.271 -72.994 47.085
309621 100.302 -72.998 47.041
309622 100.333 -73.002 46.998
309623 100.365 -73.006 46.956
309624 100.396 -73.010 46.914
309625 100.427 -73.014 46.871
309626 100.459 -73.018 46.830
309627 100.490 -73.022 46.789
309628 100.522 -73.026 46.750
309629 100.555 -73.035 46.712
309630 100.594 -73.058 46.677
309631 100.633 -73.081 46.643
309632 100.667 -73.086 46.610
309633 100.691 -73.066 46.577
309634 100.715 -73.045 46.545
309635 100.739 -73.023 46.512
309636 100.763 -73.002 46.478
309637 100.787 -72.982 46.443
309638 100.823 -72.976 46.408
309639 100.862 -72.976 46.371
309640 100.901 -72.976 46.335
309641 100.940 -72.976 46.300
309642 100.980 -72.976 46.264
309643 101.019 -72.976 46.230
309644 101.058 -72.976 46.196
309645 101.098 -72.980 46.162
309646 101.131 -72.985 46.126
309647 101.164 -72.991 46.092
309648 101.196 -72.995 46.060
309649 101.229 -73.000 46.028
309650 101.262 -73.005 45.995
309651 101.299 -73.010 45.962
309652 101.338 -73.017 45.928
309653 101.377 -73.028 45.894
309654 101.407 -73.051 45.860
309655 101.436 -73.078 45.827
309656 101.464 -73.105 45.792
309657 101.493 -73.132 45.756
309658 101.521 -73.158 45.721
309659 101.549 -73.186 45.684
309660 101.578 -73.212 45.646
309661 101.606 -73.239 45.609
309662 101.634 -73.266 45.571
309663 101.663 -73.293 45.533
309664 101.691 -73.320 45.494
309665 101.720 -73.345 45.456
309666 101.749 -73.369 45.418
309667 101.777 -73.394 45.379
309668 101.811 -73.414 45.341
309669 101.851 -73.427 45.303
309670 101.891 -73.440 45.266
309671 101.932 -73.452 45.230
309672 101.975 -73.462 45.194
309673 102.018 -73.471 45.160
309674 102.061 -73.481 45.126
309675 102.104 -73.490 45.093
309676 102.139 -73.496 45.060
309677 102.164 -73.497 45.029
309678 102.189 -73.498 44.998
309679 102.214 -73.498 44.968
310164 95.098 -71.552 52.488
310165 95.124 -71.566 52.469
310166 95.151 -71.581 52.450
310167 95.177 -71.595 52.431
310168 95.203 -71.609 52.412
310169 95.230 -71.623 52.393
310170 95.256 -71.638 52.375
310171 95.283 -71.652 52.357
310172 95.265 -71.678 52.339
310173 95.237 -71.707 52.322
310174 95.209 -71.735 52.306
310175 95.238 -71.795 52.289
310176 95.285 -71.821 52.271
310177 95.304 -71.825 52.254
310178 95.352 -71.834 52.237
310179 95.400 -71.844 52.221
310180 95.448 -71.853 52.205
310181 95.497 -71.860 52.190
310182 95.549 -71.857 52.175
310183 95.582 -71.854 52.160
310184 95.604 -71.851 52.144
310185 95.626 -71.848 52.128
310186 95.652 -71.854 52.114
310187 95.683 -71.873 52.101
310188 95.728 -71.882 52.088
310189 95.773 -71.891 52.077
310190 95.819 -71.900 52.068
310191 95.865 -71.909 52.059
310192 95.910 -71.918 52.051
310193 95.955 -71.926 52.046
310194 96.001 -71.936 52.040
310195 96.046 -71.945 52.036
310196 96.090 -71.923 52.033
310197 96.139 -71.907 52.029
310198 96.188 -71.889 52.022
310199 96.238 -71.873 52.013
310200 96.287 -71.856 52.003
310201 96.336 -71.839 51.991
310202 96.385 -71.823 51.975
310203 96.429 -71.810 51.959
310204 96.466 -71.803 51.942
310205 96.496 -71.807 51.922
310206 96.524 -71.823 51.902
310207 96.553 -71.838 51.885
310208 96.581 -71.854 51.867
310209 96.610 -71.869 51.847
310210 96.638 -71.884 51.824
310211 96.667 -71.899 51.794
310212 96.695 -71.914 51.763
310213 96.724 -71.929 51.730
310214 96.752 -71.945 51.702
310215 96.800 -71.943 51.677
310216 96.849 -71.940 51.654
310217 96.897 -71.937 51.633
310218 96.918 -71.932 51.612
310219 96.938 -71.929 51.597
310220 96.958 -71.925 51.584
310221 96.979 -71.920 51.572
310222 97.000 -71.917 51.559
310223 97.028 -71.914 51.536
310224 97.069 -71.920 51.507
310225 97.109 -71.926 51.473
310226 97.149 -71.932 51.436
310227 97.185 -71.936 51.400
310228 97.248 -71.960 51.364
310229 97.312 -71.983 51.329
310230 97.375 -72.005 51.295
310231 97.415 -72.014 51.261
310232 97.442 -72.017 51.229
310233 97.470 -72.018 51.199
310234 97.497 -72.020 51.169
310235 97.529 -72.022 51.141
310236 97.562 -72.024 51.112
310237 97.595 -72.026 51.081
310238 97.628 -72.028 51.047
310239 97.661 -72.030 51.013
310240 97.695 -72.032 50.977
310241 97.727 -72.034 50.938
310242 97.760 -72.036 50.896
310243 97.794 -72.038 50.851
310244 97.827 -72.039 50.802
310245 97.860 -72.042 50.750
310246 97.893 -72.043 50.697
310247 97.920 -72.054 50.644
310248 97.943 -72.070 50.592
310249 97.966 -72.087 50.541
310250 97.990 -72.101 50.491
310251 98.016 -72.108 50.443
310252 98.056 -72.101 50.394
310253 98.096 -72.094 50.348
310254 98.109 -72.144 50.303
310255 98.132 -72.155 50.258
310256 98.156 -72.156 50.210
310257 98.180 -72.157 50.158
310258 98.218 -72.173 50.103
310259 98.248 -72.178 50.048
310260 98.270 -72.174 49.996
310261 98.301 -72.175 49.944
310262 98.350 -72.183 49.893
310263 98.422 -72.202 49.844
310264 98.494 -72.221 49.795
310265 98.530 -72.223 49.745
310266 98.569 -72.227 49.694
310267 98.610 -72.243 49.642
310268 98.643 -72.261 49.591
310269 98.675 -72.277 49.539
310270 98.708 -72.295 49.489
310271 98.740 -72.312 49.438
310272 98.773 -72.330 49.388
310273 98.805 -72.346 49.338
310274 98.838 -72.364 49.290
310275 98.871 -72.383 49.241
310276 98.905 -72.402 49.193
310277 98.939 -72.421 49.143
310278 98.973 -72.440 49.093
310279 98.996 -72.459 49.042
310280 99.017 -72.476 48.991
310281 99.039 -72.494 48.940
310282 99.060 -72.512 48.888
310283 99.081 -72.530 48.838
310284 99.102 -72.548 48.789
310285 99.124 -72.565 48.738
310286 99.145 -72.584 48.689
310287 99.166 -72.601 48.639
310288 99.188 -72.619 48.590
310289 99.209 -72.637 48.540
310290 99.231 -72.653 48.491
310291 99.273 -72.666 48.441
310292 99.306 -72.677 48.391
310293 99.346 -72.685 48.343
310294 99.385 -72.692 48.295
310295 99.425 -72.700 48.247
310296 99.465 -72.707 48.200
310297 99.504 -72.714 48.154
310298 99.544 -72.722 48.109
310299 99.583 -72.729 48.063
310300 99.623 -72.737 48.017
310301 99.663 -72.744 47.971
310302 99.702 -72.752 47.926
310303 99.742 -72.759 47.882
310304 99.782 -72.766 47.837
310305 99.812 -72.788 47.792
310306 99.833 -72.823 47.747
310307 99.853 -72.859 47.703
310308 99.873 -72.895 47.658
310309 99.904 -72.917 47.613
310310 99.935 -72.938 47.569
310311 99.967 -72.958 47.526
310312 99.998 -72.979 47.482
310313 100.030 -72.994 47.437
310314 100.061 -72.985 47.393
310315 100.093 -72.976 47.348
310316 100.125 -72.967 47.305
310317 100.156 -72.957 47.262
310318 100.188 -72.948 47.220
310319 100.219 -72.939 47.177
310320 100.251 -72.941 47.134
310321 100.282 -72.945 47.091
310322 100.314 -72.948 47.047
310323 100.345 -72.953 47.005
310324 100.376 -72.957 46.963
310325 100.408 -72.960 46.922
310326 100.439 -72.964 46.879
310327 100.470 -72.969 46.838
310328 100.501 -72.973 46.797
310329 100.533 -72.976 46.758
310330 100.568 -72.991 46.720
310331 100.608 -73.014 46.684
310332 100.647 -73.036 46.650
310333 100.684 -73.054 46.617
310334 100.708 -73.032 46.584
310335 100.733 -73.011 46.549
310336 100.757 -72.991 46.515
310337 100.781 -72.970 46.480
310338 100.808 -72.953 46.444
310339 100.847 -72.953 46.408
310340 100.887 -72.953 46.370
310341 100.926 -72.953 46.334
310342 100.965 -72.953 46.298
310343 101.004 -72.953 46.263
310344 101.044 -72.953 46.228
310345 101.080 -72.957 46.194
310346 101.109 -72.973 46.160
310347 101.144 -72.969 46.126
310348 101.181 -72.960 46.091
310349 101.218 -72.951 46.059
310350 101.253 -72.949 46.027
310351 101.285 -72.954 45.994
310352 101.320 -72.959 45.961
310353 101.357 -72.971 45.927
310354 101.385 -72.998 45.893
310355 101.414 -73.025 45.859
310356 101.442 -73.051 45.826
310357 101.470 -73.079 45.790
310358 101.499 -73.105 45.755
310359 101.527 -73.133 45.720
310360 101.555 -73.159 45.682
310361 101.584 -73.186 45.644
310362 101.612 -73.213 45.606
310363 101.641 -73.239 45.569
310364 101.669 -73.267 45.531
310365 101.697 -73.293 45.493
310366 101.726 -73.318 45.454
310367 101.755 -73.343 45.414
310368 101.784 -73.367 45.376
310369 101.824 -73.380 45.337
310370 101.864 -73.393 45.300
310371 101.904 -73.406 45.263
310372 101.945 -73.418 45.226
310373 101.988 -73.427 45.191
310374 102.031 -73.437 45.156
310375 102.074 -73.447 45.123
310376 102.117 -73.456 45.089
310377 102.147 -73.460 45.057
310378 102.172 -73.460 45.026
310379 102.197 -73.461 44.995
310865 95.096 -71.506 52.485
310866 95.123 -71.521 52.466
310867 95.149 -71.535 52.448
310868 95.175 -71.550 52.429
310869 95.202 -71.564 52.410
310870 95.228 -71.578 52.391
310871 95.255 -71.593 52.373
310872 95.270 -71.610 52.356
310873 95.242 -71.638 52.339
310874 95.214 -71.666 52.322
310875 95.186 -71.694 52.306
310876 95.217 -71.739 52.290
310877 95.265 -71.766 52.273
310878 95.306 -71.779 52.256
310879 95.354 -71.788 52.240
310880 95.403 -71.798 52.225
310881 95.451 -71.807 52.209
310882 95.499 -71.817 52.194
310883 95.548 -71.826 52.180
310884 95.579 -71.823 52.166
310885 95.601 -71.820 52.150
310886 95.623 -71.817 52.135
310887 95.645 -71.814 52.121
310888 95.687 -71.826 52.108
310889 95.732 -71.835 52.095
310890 95.777 -71.843 52.085
310891 95.823 -71.852 52.076
310892 95.868 -71.861 52.068
310893 95.914 -71.870 52.061
310894 95.959 -71.879 52.057
310895 96.005 -71.888 52.053
310896 96.050 -71.897 52.051
310897 96.095 -71.884 52.048
310898 96.144 -71.867 52.046
310899 96.193 -71.850 52.039
310900 96.242 -71.833 52.032
310901 96.292 -71.817 52.022
310902 96.341 -71.800 52.011
310903 96.390 -71.783 51.996
310904 96.439 -71.766 51.980
310905 96.478 -71.761 51.962
310906 96.507 -71.776 51.940
310907 96.535 -71.791 51.918
310908 96.564 -71.807 51.898
310909 96.592 -71.822 51.877
310910 96.620 -71.837 51.855
310911 96.649 -71.852 51.832
310912 96.678 -71.868 51.804
310913 96.706 -71.883 51.774
310914 96.735 -71.898 51.744
310915 96.763 -71.914 51.715
310916 96.810 -71.913 51.688
310917 96.859 -71.910 51.664
310918 96.893 -71.906 51.641
310919 96.914 -71.902 51.619
310920 96.935 -71.898 51.600
310921 96.955 -71.894 51.583
310922 96.975 -71.889 51.568
310923 97.013 -71.890 51.553
310924 97.052 -71.892 51.532
310925 97.094 -71.898 51.505
310926 97.135 -71.904 51.473
310927 97.170 -71.909 51.440
310928 97.205 -71.914 51.406
310929 97.246 -71.923 51.373
310930 97.310 -71.945 51.339
310931 97.373 -71.968 51.303
310932 97.431 -71.988 51.266
310933 97.458 -71.990 51.232
310934 97.486 -71.992 51.202
310935 97.513 -71.994 51.173
310936 97.540 -71.995 51.144
310937 97.568 -71.998 51.115
310938 97.601 -71.999 51.081
310939 97.634 -72.001 51.044
310940 97.667 -72.004 51.006
310941 97.701 -72.005 50.968
310942 97.733 -72.008 50.928
310943 97.767 -72.009 50.887
310944 97.800 -72.011 50.843
310945 97.832 -72.013 50.794
310946 97.866 -72.015 50.745
310947 97.899 -72.017 50.694
310948 97.923 -72.032 50.643
310949 97.946 -72.048 50.590
310950 97.969 -72.064 50.542
310951 97.995 -72.072 50.495
310952 98.020 -72.080 50.449
310953 98.045 -72.089 50.404
310954 98.085 -72.081 50.359
310955 98.110 -72.108 50.314
310956 98.131 -72.121 50.269
310957 98.155 -72.122 50.220
310958 98.191 -72.135 50.167
310959 98.230 -72.152 50.112
310960 98.263 -72.160 50.056
310961 98.285 -72.155 50.002
310962 98.308 -72.152 49.949
310963 98.386 -72.174 49.898
310964 98.458 -72.193 49.849
310965 98.496 -72.196 49.799
310966 98.532 -72.198 49.749
310967 98.570 -72.200 49.698
310968 98.611 -72.210 49.645
310969 98.655 -72.225 49.593
310970 98.690 -72.242 49.542
310971 98.722 -72.259 49.490
310972 98.755 -72.277 49.440
310973 98.787 -72.293 49.389
310974 98.819 -72.311 49.339
310975 98.852 -72.328 49.290
310976 98.885 -72.346 49.241
310977 98.919 -72.365 49.192
310978 98.953 -72.385 49.143
310979 98.984 -72.404 49.093
310980 99.006 -72.421 49.042
310981 99.027 -72.440 48.991
310982 99.048 -72.457 48.941
310983 99.069 -72.475 48.889
310984 99.091 -72.493 48.839
310985 99.112 -72.511 48.789
310986 99.133 -72.529 48.738
310987 99.155 -72.547 48.689
310988 99.177 -72.564 48.639
310989 99.199 -72.581 48.590
310990 99.221 -72.597 48.540
310991 99.243 -72.614 48.491
310992 99.287 -72.626 48.441
310993 99.334 -72.638 48.392
310994 99.364 -72.650 48.344
310995 99.398 -72.660 48.296
310996 99.437 -72.667 48.248
310997 99.477 -72.675 48.202
310998 99.516 -72.682 48.157
310999 99.556 -72.689 48.111
311000 99.596 -72.697 48.064
311001 99.635 -72.704 48.018
311002 99.675 -72.712 47.972
311003 99.715 -72.719 47.927
311004 99.754 -72.727 47.883
311005 99.794 -72.735 47.839
311006 99.830 -72.747 47.794
311007 99.850 -72.783 47.749
311008 99.871 -72.819 47.705
311009 99.891 -72.855 47.660
311010 99.915 -72.885 47.614
311011 99.946 -72.907 47.572
311012 99.978 -72.927 47.528
311013 100.009 -72.948 47.484
311014 100.041 -72.956 47.440
311015 100.072 -72.947 47.396
311016 100.104 -72.938 47.352
311017 100.136 -72.929 47.309
311018 100.168 -72.920 47.267
311019 100.199 -72.910 47.225
311020 100.230 -72.901 47.183
311021 100.262 -72.892 47.140
311022 100.293 -72.895 47.097
311023 100.325 -72.899 47.055
311024 100.356 -72.903 47.013
311025 100.388 -72.907 46.972
311026 100.419 -72.911 46.930
311027 100.451 -72.915 46.889
311028 100.482 -72.920 46.847
311029 100.513 -72.923 46.806
311030 100.544 -72.927 46.767
311031 100.582 -72.946 46.729
311032 100.621 -72.970 46.693
311033 100.661 -72.992 46.659
311034 100.700 -73.015 46.625
311035 100.725 -72.999 46.591
311036 100.750 -72.979 46.554
311037 100.774 -72.957 46.518
311038 100.798 -72.936 46.481
311039 100.833 -72.930 46.444
311040 100.872 -72.930 46.408
311041 100.911 -72.930 46.370
311042 100.951 -72.930 46.334
311043 100.990 -72.930 46.298
311044 101.029 -72.930 46.262
311045 101.060 -72.942 46.228
311046 101.089 -72.957 46.194
311047 101.117 -72.973 46.160
311048 101.150 -72.975 46.126
311049 101.187 -72.967 46.092
311050 101.224 -72.957 46.059
311051 101.261 -72.949 46.027
311052 101.298 -72.940 45.993
311053 101.339 -72.927 45.961
311054 101.363 -72.945 45.927
311055 101.392 -72.972 45.893
311056 101.420 -72.998 45.859
311057 101.448 -73.026 45.825
311058 101.477 -73.052 45.790
311059 101.505 -73.079 45.754
311060 101.533 -73.106 45.718
311061 101.562 -73.133 45.681
311062 101.590 -73.160 45.644
311063 101.618 -73.186 45.606
311064 101.647 -73.214 45.568
311065 101.675 -73.240 45.530
311066 101.704 -73.266 45.491
311067 101.733 -73.291 45.452
311068 101.761 -73.315 45.412
311069 101.796 -73.333 45.373
311070 101.836 -73.346 45.335
311071 101.877 -73.360 45.298
311072 101.916 -73.373 45.261
311073 101.958 -73.384 45.225
311074 102.001 -73.394 45.189
311075 102.044 -73.403 45.154
311076 102.087 -73.413 45.120
311077 102.130 -73.422 45.087
311078 102.155 -73.423 45.055
311079 102.180 -73.424 45.024
311566 95.094 -71.462 52.482
311567 95.121 -71.476 52.463
311568 95.147 -71.491 52.445
311569 95.174 -71.505 52.427
311570 95.200 -71.519 52.408
311571 95.226 -71.533 52.390
311572 95.253 -71.547 52.372
311573 95.248 -71.570 52.355
311574 95.220 -71.598 52.338
311575 95.192 -71.626 52.323
311576 95.164 -71.655 52.307
311577 95.200 -71.682 52.291
311578 95.252 -71.705 52.276
311579 95.305 -71.726 52.260
311580 95.352 -71.735 52.244
311581 95.401 -71.745 52.229
311582 95.449 -71.754 52.214
311583 95.497 -71.764 52.200
311584 95.546 -71.773 52.186
311585 95.584 -71.783 52.172
311586 95.606 -71.782 52.157
311587 95.629 -71.779 52.143
311588 95.651 -71.776 52.128
311589 95.685 -71.779 52.115
311590 95.731 -71.787 52.104
311591 95.778 -71.796 52.093
311592 95.824 -71.805 52.085
311593 95.871 -71.813 52.078
311594 95.917 -71.823 52.072
311595 95.963 -71.832 52.068
311596 96.008 -71.841 52.066
311597 96.054 -71.849 52.066
311598 96.100 -71.845 52.065
311599 96.149 -71.827 52.063
311600 96.198 -71.810 52.057
311601 96.247 -71.793 52.050
311602 96.293 -71.773 52.041
311603 96.338 -71.751 52.031
311604 96.384 -71.731 52.018
311605 96.439 -71.719 52.002
311606 96.488 -71.724 51.983
311607 96.518 -71.742 51.960
311608 96.546 -71.760 51.936
311609 96.575 -71.775 51.913
311610 96.603 -71.791 51.890
311611 96.631 -71.806 51.867
311612 96.660 -71.821 51.844
311613 96.688 -71.836 51.817
311614 96.717 -71.851 51.789
311615 96.746 -71.867 51.759
311616 96.774 -71.882 51.731
311617 96.820 -71.882 51.703
311618 96.869 -71.879 51.677
311619 96.890 -71.875 51.651
311620 96.910 -71.871 51.626
311621 96.931 -71.867 51.603
311622 96.958 -71.864 51.583
311623 96.998 -71.866 51.565
311624 97.038 -71.867 51.546
311625 97.078 -71.870 51.526
311626 97.119 -71.877 51.500
311627 97.155 -71.882 51.471
311628 97.190 -71.886 51.441
311629 97.226 -71.892 51.410
311630 97.261 -71.896 51.378
311631 97.307 -71.908 51.345
311632 97.371 -71.931 51.310
311633 97.434 -71.954 51.271
311634 97.474 -71.964 51.235
311635 97.502 -71.965 51.203
311636 97.529 -71.967 51.175
311637 97.557 -71.969 51.146
311638 97.584 -71.971 51.116
311639 97.611 -71.973 51.081
311640 97.640 -71.975 51.042
311641 97.673 -71.977 51.002
311642 97.707 -71.979 50.960
311643 97.739 -71.981 50.918
311644 97.773 -71.983 50.876
311645 97.806 -71.985 50.832
311646 97.839 -71.986 50.784
311647 97.872 -71.989 50.737
311648 97.903 -71.993 50.689
311649 97.926 -72.010 50.639
311650 97.949 -72.026 50.590
311651 97.974 -72.035 50.546
311652 98.000 -72.043 50.503
311653 98.025 -72.052 50.458
311654 98.050 -72.060 50.415
311655 98.075 -72.068 50.371
311656 98.110 -72.070 50.326
311657 98.131 -72.087 50.280
311658 98.164 -72.096 50.230
311659 98.204 -72.113 50.175
311660 98.243 -72.130 50.119
311661 98.278 -72.141 50.062
311662 98.299 -72.137 50.005
311663 98.350 -72.146 49.953
311664 98.423 -72.166 49.902
311665 98.460 -72.168 49.853
311666 98.498 -72.171 49.803
311667 98.534 -72.173 49.752
311668 98.572 -72.175 49.699
311669 98.609 -72.177 49.646
311670 98.653 -72.192 49.593
311671 98.697 -72.208 49.541
311672 98.736 -72.224 49.490
311673 98.769 -72.240 49.439
311674 98.801 -72.258 49.389
311675 98.834 -72.275 49.338
311676 98.866 -72.293 49.289
311677 98.899 -72.310 49.240
311678 98.933 -72.329 49.191
311679 98.967 -72.349 49.141
311680 98.994 -72.367 49.091
311681 99.015 -72.385 49.041
311682 99.037 -72.403 48.990
311683 99.058 -72.421 48.941
311684 99.079 -72.438 48.889
311685 99.100 -72.456 48.839
311686 99.122 -72.474 48.789
311687 99.143 -72.491 48.738
311688 99.166 -72.508 48.688
311689 99.188 -72.525 48.638
311690 99.210 -72.541 48.589
311691 99.233 -72.558 48.539
311692 99.255 -72.575 48.490
311693 99.301 -72.587 48.441
311694 99.348 -72.599 48.393
311695 99.395 -72.610 48.345
311696 99.424 -72.622 48.297
311697 99.451 -72.634 48.250
311698 99.489 -72.642 48.204
311699 99.529 -72.650 48.158
311700 99.568 -72.657 48.112
311701 99.608 -72.665 48.066
311702 99.648 -72.672 48.019
311703 99.687 -72.679 47.973
311704 99.727 -72.687 47.927
311705 99.767 -72.694 47.884
311706 99.806 -72.702 47.840
311707 99.846 -72.710 47.795
311708 99.868 -72.743 47.750
311709 99.888 -72.779 47.706
311710 99.909 -72.815 47.661
311711 99.929 -72.851 47.617
311712 99.958 -72.875 47.574
311713 99.989 -72.895 47.531
311714 100.020 -72.916 47.486
311715 100.052 -72.919 47.442
311716 100.083 -72.910 47.400
311717 100.115 -72.901 47.356
311718 100.147 -72.891 47.314
311719 100.179 -72.882 47.273
311720 100.210 -72.873 47.231
311721 100.242 -72.863 47.189
311722 100.273 -72.854 47.147
311723 100.305 -72.846 47.105
311724 100.336 -72.850 47.063
311725 100.368 -72.854 47.022
311726 100.399 -72.858 46.981
311727 100.430 -72.862 46.940
311728 100.462 -72.866 46.899
311729 100.493 -72.870 46.857
311730 100.525 -72.874 46.817
311731 100.556 -72.879 46.779
311732 100.596 -72.902 46.742
311733 100.635 -72.925 46.705
311734 100.674 -72.948 46.669
311735 100.713 -72.971 46.633
311736 100.743 -72.967 46.596
311737 100.767 -72.945 46.557
311738 100.791 -72.924 46.521
311739 100.818 -72.907 46.484
311740 100.857 -72.907 46.447
311741 100.897 -72.907 46.411
311742 100.936 -72.907 46.373
311743 100.975 -72.907 46.337
311744 101.011 -72.911 46.301
311745 101.040 -72.927 46.265
311746 101.068 -72.942 46.230
311747 101.097 -72.958 46.196
311748 101.125 -72.973 46.162
311749 101.156 -72.981 46.127
311750 101.193 -72.973 46.093
311751 101.230 -72.963 46.061
311752 101.267 -72.955 46.028
311753 101.309 -72.942 45.994
311754 101.356 -72.923 45.961
311755 101.374 -72.929 45.927
311756 101.398 -72.945 45.893
311757 101.426 -72.972 45.859
311758 101.455 -72.999 45.825
311759 101.483 -73.026 45.790
311760 101.511 -73.053 45.754
311761 101.540 -73.079 45.718
311762 101.568 -73.107 45.681
311763 101.596 -73.133 45.644
311764 101.625 -73.160 45.606
311765 101.653 -73.187 45.567
311766 101.681 -73.214 45.529
311767 101.710 -73.239 45.490
311768 101.739 -73.264 45.450
311769 101.769 -73.286 45.411
311770 101.809 -73.300 45.372
311771 101.849 -73.313 45.334
311772 101.889 -73.326 45.297
311773 101.929 -73.340 45.260
311774 101.971 -73.350 45.224
311775 102.014 -73.360 45.188
311776 102.057 -73.369 45.153
311777 102.100 -73.379 45.119
311778 102.138 -73.386 45.086
311779 102.163 -73.387 45.054
311780 102.188 -73.387 45.023
312267 95.093 -71.416 52.478
312268 95.119 -71.431 52.460
312269 95.145 -71.445 52.442
312270 95.172 -71.459 52.423
312271 95.198 -71.474 52.406
312272 95.225 -71.488 52.388
312273 95.251 -71.503 52.371
312274 95.226 -71.530 52.355
312275 95.198 -71.558 52.339
312276 95.170 -71.587 52.323
312277 95.164 -71.612 52.308
312278 95.194 -71.618 52.293
312279 95.246 -71.640 52.278
312280 95.299 -71.663 52.263
312281 95.349 -71.681 52.248
312282 95.397 -71.690 52.234
312283 95.446 -71.700 52.220
312284 95.494 -71.709 52.206
312285 95.542 -71.719 52.191
312286 95.586 -71.728 52.177
312287 95.613 -71.738 52.162
312288 95.637 -71.738 52.149
312289 95.660 -71.735 52.136
312290 95.682 -71.732 52.124
312291 95.727 -71.740 52.112
312292 95.773 -71.749 52.102
312293 95.820 -71.758 52.095
312294 95.866 -71.766 52.088
312295 95.913 -71.776 52.081
312296 95.959 -71.785 52.079
312297 96.006 -71.793 52.079
312298 96.052 -71.802 52.080
312299 96.094 -71.791 52.081
312300 96.140 -71.770 52.081
312301 96.186 -71.750 52.076
312302 96.232 -71.729 52.069
312303 96.277 -71.708 52.060
312304 96.327 -71.691 52.051
312305 96.386 -71.684 52.040
312306 96.446 -71.677 52.025
312307 96.495 -71.685 52.006
312308 96.530 -71.701 51.981
312309 96.560 -71.719 51.957
312310 96.587 -71.739 51.931
312311 96.614 -71.759 51.907
312312 96.643 -71.774 51.883
312313 96.671 -71.789 51.860
312314 96.699 -71.804 51.834
312315 96.728 -71.820 51.806
312316 96.756 -71.835 51.777
312317 96.785 -71.851 51.749
312318 96.829 -71.851 51.720
312319 96.866 -71.848 51.692
312320 96.887 -71.844 51.663
312321 96.907 -71.840 51.634
312322 96.944 -71.841 51.608
312323 96.983 -71.842 51.584
312324 97.023 -71.844 51.561
312325 97.063 -71.845 51.539
312326 97.103 -71.849 51.517
312327 97.140 -71.854 51.492
312328 97.175 -71.860 51.466
312329 97.211 -71.864 51.439
312330 97.246 -71.869 51.410
312331 97.281 -71.873 51.380
312332 97.316 -71.879 51.349
312333 97.368 -71.894 51.314
312334 97.432 -71.917 51.274
312335 97.490 -71.937 51.236
312336 97.517 -71.939 51.203
312337 97.545 -71.941 51.174
312338 97.572 -71.942 51.146
312339 97.600 -71.945 51.116
312340 97.628 -71.946 51.081
312341 97.655 -71.948 51.041
312342 97.682 -71.951 50.999
312343 97.713 -71.952 50.954
312344 97.745 -71.954 50.909
312345 97.779 -71.956 50.865
312346 97.812 -71.958 50.820
312347 97.845 -71.960 50.774
312348 97.878 -71.962 50.728
312349 97.906 -71.971 50.681
312350 97.929 -71.988 50.634
312351 97.954 -71.997 50.590
312352 97.979 -72.005 50.552
312353 98.004 -72.014 50.513
312354 98.029 -72.022 50.470
312355 98.055 -72.030 50.427
312356 98.080 -72.039 50.383
312357 98.105 -72.047 50.337
312358 98.137 -72.058 50.289
312359 98.176 -72.075 50.238
312360 98.216 -72.092 50.183
312361 98.255 -72.108 50.124
312362 98.292 -72.123 50.064
312363 98.314 -72.118 50.006
312364 98.388 -72.139 49.954
312365 98.425 -72.141 49.905
312366 98.462 -72.143 49.855
312367 98.500 -72.146 49.805
312368 98.536 -72.148 49.753
312369 98.574 -72.150 49.700
312370 98.611 -72.152 49.645
312371 98.650 -72.159 49.590
312372 98.695 -72.174 49.538
312373 98.738 -72.189 49.486
312374 98.782 -72.205 49.435
312375 98.815 -72.222 49.386
312376 98.848 -72.240 49.335
312377 98.880 -72.256 49.285
312378 98.913 -72.274 49.237
312379 98.947 -72.293 49.188
312380 98.981 -72.312 49.138
312381 99.004 -72.330 49.088
312382 99.025 -72.348 49.038
312383 99.047 -72.366 48.988
312384 99.068 -72.384 48.939
312385 99.089 -72.402 48.889
312386 99.111 -72.419 48.839
312387 99.133 -72.436 48.789
312388 99.155 -72.453 48.738
312389 99.178 -72.469 48.686
312390 99.200 -72.486 48.636
312391 99.222 -72.503 48.586
312392 99.244 -72.519 48.536
312393 99.268 -72.535 48.489
312394 99.315 -72.547 48.441
312395 99.362 -72.559 48.393
312396 99.409 -72.571 48.346
312397 99.455 -72.582 48.299
312398 99.484 -72.594 48.252
312399 99.511 -72.606 48.205
312400 99.541 -72.617 48.159
312401 99.581 -72.625 48.113
312402 99.620 -72.632 48.067
312403 99.660 -72.640 48.021
312404 99.700 -72.647 47.975
312405 99.739 -72.654 47.929
312406 99.779 -72.662 47.885
312407 99.819 -72.669 47.841
312408 99.858 -72.677 47.797
312409 99.886 -72.703 47.752
312410 99.906 -72.738 47.708
312411 99.926 -72.775 47.663
312412 99.946 -72.810 47.619
312413 99.969 -72.844 47.576
312414 100.000 -72.864 47.533
312415 100.031 -72.885 47.489
312416 100.063 -72.881 47.446
312417 100.095 -72.872 47.404
312418 100.126 -72.863 47.360
312419 100.158 -72.854 47.318
312420 100.190 -72.844 47.278
312421 100.221 -72.835 47.236
312422 100.253 -72.826 47.195
312423 100.284 -72.816 47.154
312424 100.316 -72.807 47.112
312425 100.348 -72.801 47.071
312426 100.379 -72.804 47.031
312427 100.411 -72.808 46.991
312428 100.442 -72.813 46.951
312429 100.473 -72.816 46.911
312430 100.504 -72.820 46.869
312431 100.536 -72.825 46.829
312432 100.570 -72.835 46.791
312433 100.609 -72.857 46.754
312434 100.648 -72.880 46.717
312435 100.687 -72.904 46.681
312436 100.726 -72.926 46.644
312437 100.760 -72.933 46.606
312438 100.784 -72.913 46.566
312439 100.809 -72.891 46.528
312440 100.843 -72.884 46.490
312441 100.882 -72.884 46.452
312442 100.921 -72.884 46.415
312443 100.960 -72.884 46.377
312444 100.991 -72.896 46.340
312445 101.019 -72.912 46.304
312446 101.048 -72.928 46.268
312447 101.076 -72.943 46.233
312448 101.105 -72.959 46.198
312449 101.133 -72.974 46.164
312450 101.162 -72.987 46.129
312451 101.199 -72.979 46.095
312452 101.236 -72.970 46.062
312453 101.278 -72.957 46.029
312454 101.328 -72.936 45.995
312455 101.374 -72.919 45.962
312456 101.391 -72.924 45.928
312457 101.409 -72.930 45.894
312458 101.432 -72.946 45.860
312459 101.461 -72.973 45.825
312460 101.489 -72.999 45.790
312461 101.517 -73.026 45.754
312462 101.546 -73.053 45.718
312463 101.574 -73.080 45.682
312464 101.603 -73.107 45.644
312465 101.631 -73.134 45.607
312466 101.659 -73.161 45.568
312467 101.688 -73.187 45.528
312468 101.716 -73.212 45.490
312469 101.745 -73.236 45.450
312470 101.782 -73.253 45.410
312471 101.822 -73.266 45.371
312472 101.862 -73.280 45.333
312473 101.902 -73.293 45.296
312474 101.942 -73.305 45.260
312475 101.984 -73.316 45.224
312476 102.028 -73.326 45.189
312477 102.070 -73.335 45.153
312478 102.114 -73.345 45.119
312479 102.146 -73.349 45.087
312480 102.171 -73.350 45.054
312968 95.091 -71.371 52.474
312969 95.117 -71.386 52.456
312970 95.143 -71.400 52.438
312971 95.170 -71.415 52.421
312972 95.196 -71.429 52.404
312973 95.222 -71.444 52.387
312974 95.232 -71.462 52.371
312975 95.203 -71.491 52.355
312976 95.176 -71.519 52.339
312977 95.147 -71.547 52.324
312978 95.178 -71.566 52.310
312979 95.215 -71.584 52.295
312980 95.239 -71.575 52.281
312981 95.292 -71.599 52.266
312982 95.344 -71.622 52.252
312983 95.394 -71.635 52.239
312984 95.442 -71.644 52.225
312985 95.490 -71.654 52.211
312986 95.539 -71.663 52.196
312987 95.587 -71.673 52.182
312988 95.616 -71.683 52.169
312989 95.643 -71.692 52.156
312990 95.668 -71.694 52.143
312991 95.690 -71.691 52.131
312992 95.722 -71.693 52.120
312993 95.768 -71.702 52.111
312994 95.815 -71.711 52.104
312995 95.861 -71.719 52.098
312996 95.908 -71.729 52.093
312997 95.954 -71.738 52.090
312998 96.000 -71.745 52.092
312999 96.036 -71.741 52.095
313000 96.079 -71.727 52.098
313001 96.125 -71.706 52.099
313002 96.171 -71.685 52.095
313003 96.217 -71.664 52.089
313004 96.274 -71.657 52.082
313005 96.334 -71.650 52.074
313006 96.393 -71.642 52.065
313007 96.452 -71.635 52.052
313008 96.502 -71.645 52.033
313009 96.537 -71.662 52.007
313010 96.572 -71.679 51.980
313011 96.601 -71.697 51.952
313012 96.628 -71.716 51.926
313013 96.656 -71.736 51.902
313014 96.683 -71.756 51.879
313015 96.711 -71.773 51.853
313016 96.739 -71.788 51.827
313017 96.767 -71.804 51.799
313018 96.796 -71.819 51.770
313019 96.839 -71.821 51.741
313020 96.863 -71.817 51.711
313021 96.889 -71.815 51.680
313022 96.929 -71.817 51.648
313023 96.968 -71.818 51.617
313024 97.008 -71.820 51.588
313025 97.047 -71.821 51.562
313026 97.087 -71.823 51.536
313027 97.125 -71.827 51.513
313028 97.161 -71.832 51.487
313029 97.196 -71.837 51.460
313030 97.231 -71.842 51.433
313031 97.266 -71.847 51.405
313032 97.302 -71.851 51.376
313033 97.337 -71.856 51.345
313034 97.372 -71.861 51.311
313035 97.429 -71.879 51.271
313036 97.493 -71.903 51.235
313037 97.533 -71.912 51.202
313038 97.561 -71.914 51.172
313039 97.588 -71.916 51.145
313040 97.615 -71.918 51.116
313041 97.643 -71.920 51.081
313042 97.671 -71.922 51.041
313043 97.698 -71.924 50.998
313044 97.726 -71.926 50.953
313045 97.753 -71.928 50.909
313046 97.785 -71.929 50.864
313047 97.818 -71.932 50.818
313048 97.851 -71.933 50.771
313049 97.884 -71.936 50.725
313050 97.909 -71.949 50.679
313051 97.934 -71.960 50.635
313052 97.959 -71.968 50.594
313053 97.984 -71.976 50.556
313054 98.009 -71.985 50.518
313055 98.034 -71.993 50.478
313056 98.060 -72.001 50.434
313057 98.087 -72.010 50.389
313058 98.114 -72.018 50.342
313059 98.147 -72.032 50.292
313060 98.188 -72.052 50.240
313061 98.228 -72.070 50.182
313062 98.268 -72.087 50.122
313063 98.307 -72.104 50.062
313064 98.353 -72.111 50.004
313065 98.390 -72.114 49.951
313066 98.427 -72.116 49.900
313067 98.464 -72.118 49.849
313068 98.502 -72.121 49.798
313069 98.539 -72.123 49.746
313070 98.576 -72.125 49.693
313071 98.613 -72.127 49.641
313072 98.650 -72.130 49.587
313073 98.692 -72.141 49.535
313074 98.736 -72.156 49.484
313075 98.780 -72.172 49.433
313076 98.824 -72.187 49.383
313077 98.862 -72.203 49.332
313078 98.895 -72.221 49.283
313079 98.927 -72.238 49.235
313080 98.961 -72.256 49.187
313081 98.992 -72.275 49.137
313082 99.013 -72.293 49.087
313083 99.035 -72.311 49.037
313084 99.056 -72.329 48.987
313085 99.078 -72.346 48.937
313086 99.100 -72.363 48.886
313087 99.122 -72.380 48.837
313088 99.145 -72.396 48.787
313089 99.167 -72.413 48.736
313090 99.189 -72.430 48.685
313091 99.212 -72.446 48.634
313092 99.234 -72.463 48.585
313093 99.257 -72.480 48.535
313094 99.282 -72.496 48.488
313095 99.329 -72.508 48.440
313096 99.376 -72.519 48.393
313097 99.423 -72.531 48.346
313098 99.470 -72.543 48.300
313099 99.516 -72.555 48.253
313100 99.545 -72.566 48.207
313101 99.571 -72.578 48.161
313102 99.598 -72.591 48.115
313103 99.633 -72.600 48.068
313104 99.672 -72.607 48.022
313105 99.712 -72.615 47.976
313106 99.752 -72.622 47.931
313107 99.791 -72.630 47.887
313108 99.831 -72.637 47.842
313109 99.870 -72.644 47.798
313110 99.903 -72.663 47.754
313111 99.923 -72.698 47.710
313112 99.944 -72.735 47.665
313113 99.964 -72.770 47.621
313114 99.985 -72.806 47.579
313115 100.011 -72.833 47.536
313116 100.043 -72.853 47.492
313117 100.074 -72.843 47.449
313118 100.106 -72.834 47.407
313119 100.138 -72.825 47.364
313120 100.170 -72.816 47.323
313121 100.201 -72.807 47.283
313122 100.232 -72.797 47.242
313123 100.264 -72.788 47.202
313124 100.296 -72.779 47.161
313125 100.328 -72.770 47.120
313126 100.359 -72.761 47.079
313127 100.391 -72.755 47.039
313128 100.422 -72.759 47.000
313129 100.454 -72.763 46.961
313130 100.485 -72.767 46.922
313131 100.516 -72.771 46.880
313132 100.547 -72.775 46.841
313133 100.583 -72.790 46.804
313134 100.623 -72.813 46.767
313135 100.662 -72.836 46.731
313136 100.701 -72.859 46.694
313137 100.740 -72.882 46.656
313138 100.778 -72.901 46.617
313139 100.802 -72.879 46.577
313140 100.828 -72.861 46.537
313141 100.867 -72.861 46.498
313142 100.906 -72.861 46.459
313143 100.943 -72.866 46.420
313144 100.971 -72.882 46.382
313145 100.999 -72.897 46.344
313146 101.028 -72.913 46.308
313147 101.056 -72.928 46.272
313148 101.085 -72.944 46.236
313149 101.113 -72.960 46.200
313150 101.141 -72.975 46.166
313151 101.170 -72.991 46.132
313152 101.205 -72.985 46.097
313153 101.247 -72.972 46.064
313154 101.296 -72.951 46.030
313155 101.346 -72.930 45.996
313156 101.391 -72.914 45.962
313157 101.408 -72.920 45.928
313158 101.426 -72.926 45.894
313159 101.444 -72.931 45.860
313160 101.467 -72.946 45.825
313161 101.495 -72.973 45.790
313162 101.524 -73.000 45.754
313163 101.552 -73.027 45.718
313164 101.580 -73.054 45.681
313165 101.609 -73.080 45.644
313166 101.637 -73.108 45.606
313167 101.665 -73.134 45.568
313168 101.694 -73.161 45.529
313169 101.723 -73.186 45.489
313170 101.755 -73.206 45.450
313171 101.795 -73.220 45.410
313172 101.835 -73.233 45.372
313173 101.875 -73.246 45.334
313174 101.915 -73.259 45.298
313175 101.955 -73.272 45.261
313176 101.997 -73.282 45.225
313177 102.041 -73.292 45.190
313178 102.084 -73.302 45.154
313179 102.127 -73.311 45.120
313180 102.153 -73.313 45.088
313669 95.089 -71.326 52.470
313670 95.115 -71.340 52.452
313671 95.141 -71.355 52.435
313672 95.168 -71.369 52.419
313673 95.194 -71.384 52.402
313674 95.220 -71.398 52.386
313675 95.209 -71.422 52.370
313676 95.181 -71.450 52.355
313677 95.154 -71.478 52.339
313678 95.157 -71.502 52.325
313679 95.193 -71.520 52.311
313680 95.231 -71.538 52.298
313681 95.264 -71.550 52.283
313682 95.285 -71.535 52.269
313683 95.338 -71.558 52.256
313684 95.390 -71.581 52.243
313685 95.439 -71.590 52.229
313686 95.487 -71.600 52.216
313687 95.535 -71.609 52.202
313688 95.584 -71.619 52.188
313689 95.618 -71.628 52.175
313690 95.646 -71.638 52.163
313691 95.673 -71.647 52.150
313692 95.698 -71.651 52.138
313693 95.721 -71.648 52.128
313694 95.763 -71.655 52.120
313695 95.810 -71.663 52.113
313696 95.856 -71.672 52.108
313697 95.903 -71.682 52.104
313698 95.947 -71.688 52.102
313699 95.983 -71.684 52.105
313700 96.020 -71.681 52.110
313701 96.064 -71.663 52.115
313702 96.110 -71.642 52.117
313703 96.162 -71.629 52.114
313704 96.222 -71.622 52.110
313705 96.281 -71.615 52.104
313706 96.340 -71.607 52.097
313707 96.399 -71.600 52.089
313708 96.459 -71.594 52.077
313709 96.508 -71.606 52.058
313710 96.543 -71.622 52.033
313711 96.578 -71.639 52.004
313712 96.613 -71.655 51.973
313713 96.643 -71.674 51.944
313714 96.670 -71.694 51.918
313715 96.697 -71.713 51.896
313716 96.724 -71.733 51.870
313717 96.751 -71.753 51.844
313718 96.779 -71.773 51.818
313719 96.807 -71.788 51.790
313720 96.838 -71.791 51.761
313721 96.874 -71.791 51.730
313722 96.914 -71.792 51.697
313723 96.954 -71.794 51.663
313724 96.993 -71.795 51.628
313725 97.033 -71.797 51.595
313726 97.072 -71.797 51.565
313727 97.110 -71.799 51.537
313728 97.146 -71.805 51.512
313729 97.181 -71.810 51.484
313730 97.217 -71.814 51.455
313731 97.252 -71.820 51.428
313732 97.287 -71.824 51.400
313733 97.322 -71.829 51.370
313734 97.358 -71.834 51.339
313735 97.393 -71.838 51.306
313736 97.428 -71.843 51.268
313737 97.491 -71.866 51.233
313738 97.549 -71.885 51.200
313739 97.577 -71.888 51.170
313740 97.604 -71.890 51.142
313741 97.632 -71.892 51.114
313742 97.659 -71.894 51.081
313743 97.686 -71.895 51.042
313744 97.714 -71.898 51.000
313745 97.741 -71.899 50.955
313746 97.769 -71.901 50.910
313747 97.796 -71.903 50.864
313748 97.824 -71.905 50.819
313749 97.857 -71.907 50.774
313750 97.889 -71.911 50.729
313751 97.913 -71.922 50.685
313752 97.938 -71.930 50.641
313753 97.963 -71.939 50.600
313754 97.991 -71.948 50.562
313755 98.017 -71.956 50.523
313756 98.044 -71.964 50.481
313757 98.072 -71.973 50.435
313758 98.098 -71.981 50.388
313759 98.126 -71.989 50.339
313760 98.155 -71.998 50.288
313761 98.196 -72.019 50.236
313762 98.237 -72.039 50.179
313763 98.278 -72.060 50.120
313764 98.318 -72.080 50.061
313765 98.355 -72.086 50.006
313766 98.392 -72.089 49.954
313767 98.430 -72.091 49.902
313768 98.466 -72.093 49.851
313769 98.504 -72.095 49.799
313770 98.541 -72.098 49.746
313771 98.578 -72.100 49.694
313772 98.615 -72.102 49.642
313773 98.652 -72.105 49.590
313774 98.690 -72.108 49.538
313775 98.734 -72.124 49.486
313776 98.778 -72.139 49.435
313777 98.822 -72.154 49.384
313778 98.866 -72.169 49.333
313779 98.909 -72.185 49.284
313780 98.941 -72.202 49.235
313781 98.975 -72.220 49.186
313782 99.002 -72.239 49.137
313783 99.023 -72.256 49.086
313784 99.045 -72.274 49.036
313785 99.067 -72.290 48.987
313786 99.089 -72.307 48.935
313787 99.112 -72.324 48.884
313788 99.134 -72.340 48.834
313789 99.156 -72.357 48.784
313790 99.179 -72.374 48.734
313791 99.201 -72.390 48.683
313792 99.223 -72.407 48.634
313793 99.246 -72.424 48.584
313794 99.268 -72.440 48.535
313795 99.296 -72.456 48.487
313796 99.343 -72.468 48.440
313797 99.390 -72.480 48.393
313798 99.436 -72.491 48.346
313799 99.484 -72.503 48.300
313800 99.531 -72.515 48.254
313801 99.577 -72.527 48.208
313802 99.605 -72.539 48.163
313803 99.632 -72.550 48.116
313804 99.658 -72.562 48.069
313805 99.685 -72.575 48.023
313806 99.724 -72.582 47.976
313807 99.764 -72.590 47.931
313808 99.803 -72.597 47.888
313809 99.843 -72.605 47.844
313810 99.883 -72.613 47.799
313811 99.920 -72.622 47.756
313812 99.941 -72.658 47.712
313813 99.961 -72.694 47.667
313814 99.982 -72.730 47.624
313815 100.002 -72.766 47.582
313816 100.022 -72.802 47.538
313817 100.054 -72.815 47.495
313818 100.085 -72.806 47.453
313819 100.117 -72.797 47.411
313820 100.149 -72.788 47.369
313821 100.181 -72.779 47.328
313822 100.212 -72.769 47.288
313823 100.244 -72.760 47.248
313824 100.275 -72.750 47.209
313825 100.307 -72.741 47.168
313826 100.339 -72.732 47.127
313827 100.370 -72.723 47.087
313828 100.402 -72.714 47.048
313829 100.433 -72.710 47.009
313830 100.465 -72.714 46.970
313831 100.496 -72.718 46.932
313832 100.528 -72.722 46.894
313833 100.559 -72.725 46.856
313834 100.597 -72.746 46.820
313835 100.636 -72.769 46.784
313836 100.675 -72.791 46.744
313837 100.714 -72.815 46.707
313838 100.754 -72.838 46.668
313839 100.793 -72.860 46.628
313840 100.819 -72.847 46.588
313841 100.852 -72.838 46.546
313842 100.892 -72.838 46.506
313843 100.922 -72.851 46.467
313844 100.951 -72.866 46.428
313845 100.979 -72.882 46.389
313846 101.007 -72.898 46.351
313847 101.036 -72.913 46.313
313848 101.064 -72.929 46.277
313849 101.093 -72.945 46.241
313850 101.121 -72.960 46.205
313851 101.149 -72.976 46.171
313852 101.178 -72.992 46.137
313853 101.215 -72.987 46.101
313854 101.265 -72.966 46.067
313855 101.315 -72.945 46.033
313856 101.365 -72.925 45.999
313857 101.408 -72.910 45.964
313858 101.425 -72.916 45.929
313859 101.444 -72.921 45.895
313860 101.461 -72.927 45.860
313861 101.479 -72.932 45.826
313862 101.501 -72.947 45.790
313863 101.530 -72.973 45.754
313864 101.558 -73.001 45.718
313865 101.587 -73.027 45.681
313866 101.615 -73.054 45.644
313867 101.643 -73.081 45.606
313868 101.672 -73.108 45.568
313869 101.700 -73.134 45.529
313870 101.729 -73.158 45.489
313871 101.767 -73.173 45.450
313872 101.808 -73.186 45.411
313873 101.847 -73.199 45.373
313874 101.888 -73.212 45.335
313875 101.927 -73.226 45.299
313876 101.968 -73.239 45.263
313877 102.010 -73.248 45.227
313878 102.054 -73.258 45.191
313879 102.097 -73.267 45.156
313880 102.137 -73.275 45.121
314370 95.087 -71.281 52.467
314371 95.113 -71.296 52.450
314372 95.139 -71.310 52.433
314373 95.166 -71.324 52.417
314374 95.192 -71.339 52.401
314375 95.215 -71.354 52.385
314376 95.187 -71.382 52.370
314377 95.159 -71.410 52.355
314378 95.134 -71.438 52.341
314379 95.172 -71.456 52.327
314380 95.208 -71.474 52.313
314381 95.245 -71.492 52.300
314382 95.282 -71.510 52.286
314383 95.306 -71.504 52.274
314384 95.331 -71.493 52.261
314385 95.384 -71.516 52.248
314386 95.435 -71.535 52.235
314387 95.483 -71.544 52.222
314388 95.532 -71.554 52.208
314389 95.580 -71.563 52.195
314390 95.621 -71.573 52.183
314391 95.649 -71.582 52.169
314392 95.676 -71.592 52.157
314393 95.703 -71.602 52.146
314394 95.729 -71.607 52.135
314395 95.758 -71.608 52.128
314396 95.805 -71.616 52.122
314397 95.851 -71.625 52.117
314398 95.894 -71.630 52.114
314399 95.930 -71.626 52.113
314400 95.966 -71.623 52.119
314401 96.003 -71.619 52.124
314402 96.050 -71.600 52.129
314403 96.109 -71.594 52.133
314404 96.169 -71.587 52.134
314405 96.228 -71.580 52.132
314406 96.288 -71.572 52.128
314407 96.346 -71.566 52.122
314408 96.406 -71.559 52.113
314409 96.468 -71.553 52.099
314410 96.515 -71.567 52.080
314411 96.550 -71.583 52.056
314412 96.585 -71.600 52.028
314413 96.620 -71.616 51.995
314414 96.655 -71.632 51.963
314415 96.684 -71.651 51.933
314416 96.711 -71.671 51.911
314417 96.738 -71.691 51.885
314418 96.765 -71.710 51.859
314419 96.793 -71.730 51.834
314420 96.820 -71.750 51.807
314421 96.855 -71.757 51.779
314422 96.893 -71.756 51.748
314423 96.932 -71.756 51.715
314424 96.970 -71.755 51.677
314425 97.009 -71.755 51.640
314426 97.047 -71.754 51.604
314427 97.084 -71.755 51.570
314428 97.120 -71.761 51.541
314429 97.160 -71.773 51.513
314430 97.199 -71.784 51.483
314431 97.237 -71.792 51.453
314432 97.273 -71.797 51.424
314433 97.308 -71.801 51.395
314434 97.343 -71.807 51.363
314435 97.378 -71.811 51.332
314436 97.413 -71.816 51.300
314437 97.448 -71.821 51.264
314438 97.488 -71.829 51.230
314439 97.551 -71.851 51.198
314440 97.592 -71.861 51.167
314441 97.620 -71.864 51.138
314442 97.647 -71.865 51.110
314443 97.675 -71.867 51.079
314444 97.703 -71.869 51.043
314445 97.730 -71.871 51.002
314446 97.757 -71.873 50.958
314447 97.785 -71.875 50.911
314448 97.812 -71.876 50.866
314449 97.840 -71.879 50.823
314450 97.867 -71.880 50.782
314451 97.894 -71.885 50.739
314452 97.921 -71.893 50.695
314453 97.948 -71.901 50.651
314454 97.975 -71.910 50.609
314455 98.002 -71.918 50.569
314456 98.029 -71.926 50.528
314457 98.056 -71.936 50.482
314458 98.083 -71.944 50.432
314459 98.110 -71.952 50.381
314460 98.138 -71.960 50.330
314461 98.165 -71.967 50.279
314462 98.205 -71.986 50.228
314463 98.245 -72.007 50.173
314464 98.286 -72.027 50.117
314465 98.326 -72.048 50.062
314466 98.364 -72.062 50.011
314467 98.396 -72.064 49.961
314468 98.432 -72.066 49.910
314469 98.468 -72.068 49.858
314470 98.506 -72.070 49.807
314471 98.543 -72.073 49.752
314472 98.580 -72.075 49.700
314473 98.617 -72.077 49.649
314474 98.654 -72.080 49.597
314475 98.691 -72.082 49.546
314476 98.732 -72.090 49.493
314477 98.776 -72.105 49.440
314478 98.819 -72.121 49.388
314479 98.864 -72.136 49.337
314480 98.908 -72.152 49.287
314481 98.951 -72.167 49.236
314482 98.989 -72.183 49.186
314483 99.012 -72.202 49.138
314484 99.034 -72.218 49.087
314485 99.057 -72.235 49.038
314486 99.079 -72.252 48.988
314487 99.101 -72.268 48.936
314488 99.124 -72.285 48.883
314489 99.146 -72.302 48.831
314490 99.168 -72.318 48.781
314491 99.191 -72.335 48.731
314492 99.213 -72.352 48.682
314493 99.235 -72.368 48.632
314494 99.258 -72.385 48.583
314495 99.280 -72.402 48.534
314496 99.310 -72.417 48.486
314497 99.357 -72.428 48.440
314498 99.404 -72.440 48.393
314499 99.450 -72.452 48.347
314500 99.497 -72.464 48.301
314501 99.544 -72.475 48.255
314502 99.591 -72.487 48.209
314503 99.638 -72.499 48.164
314504 99.665 -72.511 48.117
314505 99.692 -72.523 48.071
314506 99.719 -72.535 48.024
314507 99.746 -72.547 47.978
314508 99.776 -72.557 47.932
314509 99.816 -72.565 47.888
314510 99.855 -72.572 47.844
314511 99.895 -72.580 47.800
314512 99.934 -72.587 47.756
314513 99.959 -72.618 47.713
314514 99.979 -72.653 47.669
314515 99.999 -72.690 47.626
314516 100.019 -72.725 47.584
314517 100.040 -72.762 47.541
314518 100.065 -72.777 47.499
314519 100.097 -72.768 47.456
314520 100.128 -72.759 47.414
314521 100.160 -72.750 47.373
314522 100.192 -72.741 47.333
314523 100.223 -72.732 47.294
314524 100.255 -72.722 47.254
314525 100.286 -72.713 47.215
314526 100.318 -72.703 47.175
314527 100.350 -72.694 47.135
314528 100.382 -72.685 47.095
314529 100.413 -72.676 47.058
314530 100.444 -72.667 47.019
314531 100.476 -72.664 46.981
314532 100.507 -72.669 46.943
314533 100.539 -72.672 46.905
314534 100.571 -72.678 46.869
314535 100.610 -72.701 46.834
314536 100.649 -72.724 46.798
314537 100.689 -72.747 46.758
314538 100.728 -72.770 46.720
314539 100.767 -72.793 46.680
314540 100.806 -72.816 46.639
314541 100.838 -72.815 46.598
314542 100.874 -72.820 46.556
314543 100.902 -72.836 46.516
314544 100.930 -72.851 46.476
314545 100.959 -72.867 46.436
314546 100.987 -72.883 46.397
314547 101.015 -72.898 46.357
314548 101.044 -72.911 46.319
314549 101.073 -72.923 46.283
314550 101.102 -72.935 46.247
314551 101.130 -72.948 46.211
314552 101.159 -72.960 46.175
314553 101.191 -72.972 46.141
314554 101.234 -72.981 46.105
314555 101.284 -72.960 46.070
314556 101.334 -72.940 46.036
314557 101.384 -72.919 46.001
314558 101.425 -72.906 45.966
314559 101.443 -72.911 45.931
314560 101.461 -72.917 45.896
314561 101.478 -72.923 45.861
314562 101.496 -72.928 45.826
314563 101.514 -72.934 45.791
314564 101.536 -72.948 45.755
314565 101.564 -72.974 45.719
314566 101.593 -73.001 45.682
314567 101.621 -73.028 45.643
314568 101.650 -73.054 45.605
314569 101.678 -73.082 45.568
314570 101.706 -73.107 45.529
314571 101.740 -73.126 45.490
314572 101.780 -73.139 45.451
314573 101.820 -73.152 45.412
314574 101.861 -73.166 45.374
314575 101.900 -73.179 45.337
314576 101.941 -73.192 45.300
314577 101.981 -73.205 45.264
314578 102.024 -73.214 45.228
314579 102.067 -73.224 45.192
314580 102.110 -73.233 45.157
314581 102.144 -73.239 45.123
315071 95.085 -71.236 52.465
315072 95.111 -71.250 52.448
315073 95.137 -71.265 52.432
315074 95.164 -71.279 52.416
315075 95.190 -71.293 52.401
315076 95.193 -71.314 52.385
315077 95.165 -71.343 52.370
315078 95.137 -71.371 52.356
315079 95.149 -71.393 52.342
315080 95.187 -71.410 52.330
315081 95.224 -71.428 52.316
315082 95.260 -71.447 52.303
315083 95.298 -71.465 52.290
315084 95.334 -71.482 52.277
315085 95.348 -71.458 52.265
315086 95.377 -71.452 52.253
315087 95.430 -71.475 52.240
315088 95.480 -71.490 52.227
315089 95.528 -71.499 52.214
315090 95.577 -71.509 52.201
315091 95.623 -71.518 52.189
315092 95.651 -71.528 52.176
315093 95.678 -71.538 52.165
315094 95.705 -71.547 52.154
315095 95.733 -71.556 52.144
315096 95.759 -71.564 52.136
315097 95.800 -71.569 52.130
315098 95.841 -71.572 52.126
315099 95.877 -71.569 52.123
315100 95.913 -71.566 52.123
315101 95.954 -71.563 52.128
315102 95.998 -71.563 52.134
315103 96.057 -71.559 52.140
315104 96.116 -71.552 52.146
315105 96.175 -71.545 52.149
315106 96.235 -71.538 52.148
315107 96.294 -71.531 52.145
315108 96.353 -71.524 52.139
315109 96.412 -71.517 52.129
315110 96.482 -71.512 52.115
315111 96.521 -71.528 52.096
315112 96.556 -71.544 52.073
315113 96.591 -71.560 52.044
315114 96.626 -71.577 52.011
315115 96.661 -71.594 51.978
315116 96.696 -71.610 51.947
315117 96.725 -71.629 51.921
315118 96.752 -71.648 51.894
315119 96.780 -71.668 51.867
315120 96.807 -71.688 51.842
315121 96.834 -71.707 51.817
315122 96.869 -71.713 51.791
315123 96.907 -71.713 51.761
315124 96.946 -71.713 51.728
315125 96.985 -71.712 51.690
315126 97.023 -71.712 51.651
315127 97.060 -71.713 51.613
315128 97.093 -71.716 51.576
315129 97.130 -71.724 51.544
315130 97.170 -71.735 51.515
315131 97.210 -71.747 51.484
315132 97.249 -71.758 51.453
315133 97.289 -71.770 51.422
315134 97.328 -71.779 51.391
315135 97.363 -71.784 51.360
315136 97.398 -71.789 51.329
315137 97.434 -71.794 51.298
315138 97.469 -71.798 51.263
315139 97.504 -71.803 51.230
315140 97.549 -71.814 51.197
315141 97.608 -71.835 51.165
315142 97.636 -71.837 51.134
315143 97.663 -71.838 51.106
315144 97.691 -71.841 51.075
315145 97.718 -71.842 51.039
315146 97.746 -71.845 50.999
315147 97.773 -71.846 50.956
315148 97.801 -71.848 50.912
315149 97.828 -71.850 50.868
315150 97.855 -71.852 50.827
315151 97.880 -71.853 50.788
315152 97.906 -71.856 50.747
315153 97.932 -71.864 50.705
315154 97.959 -71.873 50.661
315155 97.987 -71.881 50.618
315156 98.013 -71.889 50.574
315157 98.040 -71.898 50.527
315158 98.067 -71.906 50.477
315159 98.094 -71.915 50.426
315160 98.122 -71.923 50.376
315161 98.149 -71.931 50.326
315162 98.177 -71.939 50.277
315163 98.213 -71.954 50.227
315164 98.253 -71.974 50.175
315165 98.294 -71.995 50.123
315166 98.335 -72.015 50.071
315167 98.376 -72.039 50.022
315168 98.408 -72.040 49.974
315169 98.440 -72.042 49.924
315170 98.472 -72.043 49.873
315171 98.508 -72.045 49.820
315172 98.545 -72.048 49.766
315173 98.582 -72.050 49.714
315174 98.619 -72.052 49.662
315175 98.656 -72.055 49.610
315176 98.693 -72.057 49.558
315177 98.730 -72.059 49.504
315178 98.774 -72.073 49.450
315179 98.817 -72.088 49.395
315180 98.861 -72.103 49.343
315181 98.906 -72.118 49.292
315182 98.949 -72.133 49.242
315183 98.991 -72.149 49.191
315184 99.023 -72.162 49.140
315185 99.046 -72.179 49.089
315186 99.068 -72.196 49.039
315187 99.090 -72.212 48.988
315188 99.113 -72.229 48.937
315189 99.135 -72.246 48.883
315190 99.158 -72.262 48.831
315191 99.180 -72.279 48.781
315192 99.202 -72.296 48.731
315193 99.225 -72.312 48.681
315194 99.247 -72.329 48.632
315195 99.269 -72.346 48.583
315196 99.292 -72.362 48.535
315197 99.324 -72.377 48.488
315198 99.370 -72.389 48.441
315199 99.418 -72.400 48.394
315200 99.465 -72.412 48.348
315201 99.511 -72.424 48.301
315202 99.558 -72.436 48.255
315203 99.605 -72.447 48.209
315204 99.652 -72.459 48.163
315205 99.699 -72.472 48.118
315206 99.725 -72.483 48.072
315207 99.752 -72.495 48.026
315208 99.779 -72.507 47.979
315209 99.806 -72.519 47.934
315210 99.833 -72.531 47.889
315211 99.867 -72.540 47.845
315212 99.907 -72.547 47.802
315213 99.946 -72.555 47.758
315214 99.976 -72.578 47.714
315215 99.996 -72.613 47.671
315216 100.017 -72.650 47.628
315217 100.037 -72.685 47.587
315218 100.056 -72.717 47.545
315219 100.081 -72.733 47.502
315220 100.113 -72.735 47.460
315221 100.143 -72.724 47.418
315222 100.174 -72.714 47.377
315223 100.204 -72.704 47.338
315224 100.234 -72.694 47.299
315225 100.266 -72.685 47.260
315226 100.298 -72.675 47.221
315227 100.329 -72.666 47.182
315228 100.361 -72.657 47.143
315229 100.393 -72.648 47.104
315230 100.424 -72.639 47.068
315231 100.456 -72.629 47.030
315232 100.487 -72.620 46.993
315233 100.519 -72.619 46.956
315234 100.550 -72.623 46.918
315235 100.584 -72.634 46.883
315236 100.624 -72.657 46.847
315237 100.663 -72.680 46.810
315238 100.702 -72.703 46.771
315239 100.742 -72.725 46.732
315240 100.780 -72.749 46.691
315241 100.819 -72.772 46.650
315242 100.854 -72.795 46.608
315243 100.883 -72.807 46.566
315244 100.912 -72.819 46.525
315245 100.940 -72.832 46.485
315246 100.970 -72.844 46.445
315247 100.998 -72.857 46.404
315248 101.027 -72.869 46.364
315249 101.056 -72.882 46.326
315250 101.085 -72.894 46.290
315251 101.114 -72.906 46.253
315252 101.142 -72.919 46.217
315253 101.174 -72.930 46.180
315254 101.216 -72.939 46.144
315255 101.259 -72.949 46.108
315256 101.303 -72.955 46.073
315257 101.353 -72.934 46.038
315258 101.403 -72.913 46.003
315259 101.442 -72.901 45.969
315260 101.460 -72.907 45.933
315261 101.478 -72.913 45.898
315262 101.495 -72.918 45.862
315263 101.513 -72.924 45.828
315264 101.531 -72.929 45.792
315265 101.549 -72.935 45.757
315266 101.570 -72.948 45.720
315267 101.599 -72.975 45.682
315268 101.627 -73.001 45.644
315269 101.656 -73.029 45.606
315270 101.684 -73.055 45.568
315271 101.713 -73.079 45.529
315272 101.753 -73.092 45.490
315273 101.793 -73.106 45.452
315274 101.833 -73.119 45.414
315275 101.873 -73.132 45.376
315276 101.913 -73.145 45.339
315277 101.953 -73.158 45.302
315278 101.994 -73.170 45.266
315279 102.037 -73.180 45.230
315280 102.080 -73.189 45.195
315281 102.123 -73.199 45.160
315772 95.083 -71.191 52.462
315773 95.109 -71.205 52.447
315774 95.136 -71.220 52.432
315775 95.162 -71.234 52.417
315776 95.188 -71.249 52.402
315777 95.171 -71.274 52.386
315778 95.142 -71.303 52.372
315779 95.128 -71.329 52.358
315780 95.165 -71.346 52.345
315781 95.201 -71.365 52.333
315782 95.239 -71.383 52.320
315783 95.275 -71.400 52.307
315784 95.313 -71.419 52.294
315785 95.349 -71.437 52.282
315786 95.376 -71.436 52.270
315787 95.390 -71.412 52.258
315788 95.423 -71.411 52.246
315789 95.476 -71.434 52.233
315790 95.525 -71.444 52.220
315791 95.573 -71.454 52.208
315792 95.621 -71.463 52.196
315793 95.654 -71.473 52.184
315794 95.681 -71.482 52.172
315795 95.708 -71.492 52.162
315796 95.736 -71.502 52.152
315797 95.763 -71.511 52.144
315798 95.796 -71.516 52.138
315799 95.836 -71.513 52.134
315800 95.877 -71.511 52.132
315801 95.918 -71.509 52.132
315802 95.959 -71.506 52.137
315803 96.008 -71.511 52.142
315804 96.063 -71.517 52.148
315805 96.122 -71.510 52.157
315806 96.182 -71.503 52.161
315807 96.241 -71.496 52.161
315808 96.301 -71.489 52.158
315809 96.360 -71.482 52.153
315810 96.419 -71.475 52.142
315811 96.493 -71.472 52.127
315812 96.528 -71.488 52.109
315813 96.563 -71.505 52.085
315814 96.598 -71.521 52.057
315815 96.633 -71.538 52.026
315816 96.668 -71.554 51.993
315817 96.703 -71.571 51.960
315818 96.738 -71.587 51.933
315819 96.766 -71.606 51.903
315820 96.794 -71.625 51.875
315821 96.821 -71.645 51.849
315822 96.848 -71.665 51.825
315823 96.883 -71.670 51.799
315824 96.922 -71.670 51.770
315825 96.961 -71.669 51.738
315826 97.000 -71.669 51.699
315827 97.035 -71.670 51.660
315828 97.068 -71.673 51.622
315829 97.101 -71.675 51.584
315830 97.140 -71.686 51.551
315831 97.180 -71.697 51.520
315832 97.220 -71.709 51.488
315833 97.259 -71.721 51.455
315834 97.299 -71.732 51.423
315835 97.339 -71.744 51.391
315836 97.379 -71.755 51.360
315837 97.419 -71.766 51.329
315838 97.454 -71.771 51.299
315839 97.490 -71.776 51.265
315840 97.525 -71.781 51.231
315841 97.560 -71.785 51.198
315842 97.610 -71.800 51.165
315843 97.652 -71.810 51.133
315844 97.679 -71.812 51.103
315845 97.707 -71.814 51.071
315846 97.734 -71.816 51.035
315847 97.761 -71.818 50.997
315848 97.789 -71.820 50.955
315849 97.816 -71.822 50.913
315850 97.842 -71.823 50.872
315851 97.867 -71.824 50.835
315852 97.892 -71.826 50.797
315853 97.917 -71.827 50.757
315854 97.944 -71.835 50.715
315855 97.971 -71.844 50.671
315856 97.998 -71.852 50.626
315857 98.025 -71.860 50.579
315858 98.052 -71.869 50.530
315859 98.079 -71.877 50.479
315860 98.105 -71.885 50.429
315861 98.133 -71.894 50.381
315862 98.161 -71.901 50.333
315863 98.189 -71.909 50.284
315864 98.221 -71.920 50.235
315865 98.262 -71.941 50.184
315866 98.302 -71.961 50.134
315867 98.343 -71.982 50.085
315868 98.384 -72.006 50.035
315869 98.420 -72.017 49.988
315870 98.452 -72.018 49.938
315871 98.484 -72.020 49.886
315872 98.516 -72.021 49.832
315873 98.548 -72.023 49.779
315874 98.584 -72.025 49.727
315875 98.621 -72.027 49.676
315876 98.658 -72.030 49.624
315877 98.696 -72.032 49.571
315878 98.732 -72.034 49.516
315879 98.771 -72.039 49.460
315880 98.815 -72.055 49.404
315881 98.859 -72.070 49.351
315882 98.903 -72.086 49.300
315883 98.947 -72.101 49.248
315884 98.984 -72.114 49.196
315885 99.018 -72.128 49.143
315886 99.052 -72.142 49.091
315887 99.080 -72.156 49.040
315888 99.102 -72.173 48.989
315889 99.125 -72.189 48.936
315890 99.147 -72.206 48.883
315891 99.169 -72.223 48.831
315892 99.192 -72.240 48.781
315893 99.214 -72.256 48.730
315894 99.236 -72.273 48.681
315895 99.259 -72.290 48.632
315896 99.281 -72.306 48.584
315897 99.303 -72.323 48.536
315898 99.338 -72.337 48.490
315899 99.384 -72.349 48.442
315900 99.431 -72.361 48.395
315901 99.478 -72.373 48.349
315902 99.525 -72.384 48.302
315903 99.572 -72.396 48.255
315904 99.619 -72.408 48.209
315905 99.665 -72.420 48.163
315906 99.713 -72.431 48.118
315907 99.759 -72.443 48.073
315908 99.786 -72.456 48.027
315909 99.812 -72.467 47.981
315910 99.839 -72.479 47.936
315911 99.866 -72.491 47.891
315912 99.893 -72.503 47.847
315913 99.920 -72.515 47.804
315914 99.959 -72.522 47.760
315915 99.993 -72.537 47.716
315916 100.014 -72.573 47.673
315917 100.034 -72.608 47.631
315918 100.053 -72.639 47.590
315919 100.074 -72.663 47.548
315920 100.100 -72.678 47.506
315921 100.130 -72.698 47.464
315922 100.161 -72.692 47.423
315923 100.191 -72.681 47.382
315924 100.221 -72.672 47.342
315925 100.252 -72.661 47.304
315926 100.282 -72.651 47.266
315927 100.313 -72.641 47.227
315928 100.343 -72.631 47.189
315929 100.374 -72.621 47.151
315930 100.404 -72.610 47.114
315931 100.435 -72.601 47.078
315932 100.467 -72.592 47.042
315933 100.499 -72.583 47.006
315934 100.530 -72.574 46.970
315935 100.562 -72.574 46.935
315936 100.598 -72.590 46.895
315937 100.637 -72.613 46.858
315938 100.677 -72.635 46.822
315939 100.714 -72.660 46.784
315940 100.751 -72.685 46.745
315941 100.788 -72.710 46.704
315942 100.826 -72.735 46.661
315943 100.863 -72.760 46.618
315944 100.895 -72.778 46.576
315945 100.924 -72.790 46.534
315946 100.953 -72.803 46.494
315947 100.981 -72.815 46.454
315948 101.010 -72.827 46.413
315949 101.039 -72.839 46.372
315950 101.068 -72.852 46.334
315951 101.097 -72.864 46.297
315952 101.125 -72.876 46.259
315953 101.156 -72.888 46.223
315954 101.198 -72.898 46.186
315955 101.242 -72.907 46.150
315956 101.284 -72.916 46.114
315957 101.327 -72.926 46.078
315958 101.372 -72.929 46.043
315959 101.421 -72.908 46.007
315960 101.459 -72.897 45.972
315961 101.477 -72.903 45.936
315962 101.495 -72.908 45.901
315963 101.513 -72.914 45.865
315964 101.531 -72.920 45.830
315965 101.548 -72.925 45.794
315966 101.566 -72.931 45.758
315967 101.584 -72.936 45.722
315968 101.605 -72.948 45.684
315969 101.633 -72.975 45.646
315970 101.662 -73.002 45.608
315971 101.690 -73.028 45.570
315972 101.726 -73.046 45.530
315973 101.766 -73.059 45.492
315974 101.806 -73.072 45.454
315975 101.846 -73.086 45.416
315976 101.886 -73.098 45.378
315977 101.926 -73.111 45.341
315978 101.966 -73.125 45.305
315979 102.007 -73.136 45.269
315980 102.050 -73.146 45.233
315981 102.093 -73.155 45.198
315982 102.135 -73.165 45.163
316473 95.081 -71.146 52.463
316474 95.107 -71.160 52.447
316475 95.134 -71.174 52.433
316476 95.160 -71.189 52.418
316477 95.176 -71.206 52.403
316478 95.148 -71.234 52.389
316479 95.120 -71.262 52.376
316480 95.143 -71.283 52.363
316481 95.180 -71.301 52.350
316482 95.216 -71.319 52.337
316483 95.254 -71.337 52.325
316484 95.290 -71.355 52.313
316485 95.327 -71.373 52.300
316486 95.365 -71.391 52.288
316487 95.401 -71.409 52.276
316488 95.418 -71.390 52.264
316489 95.432 -71.365 52.251
316490 95.470 -71.370 52.240
316491 95.521 -71.390 52.227
316492 95.569 -71.399 52.215
316493 95.618 -71.409 52.204
316494 95.656 -71.418 52.192
316495 95.683 -71.428 52.180
316496 95.710 -71.437 52.169
316497 95.738 -71.447 52.160
316498 95.765 -71.456 52.152
316499 95.800 -71.459 52.147
316500 95.841 -71.456 52.143
316501 95.882 -71.454 52.140
316502 95.922 -71.452 52.140
316503 95.963 -71.449 52.144
316504 96.017 -71.458 52.148
316505 96.072 -71.468 52.155
316506 96.129 -71.469 52.164
316507 96.188 -71.462 52.168
316508 96.248 -71.454 52.169
316509 96.307 -71.447 52.168
316510 96.366 -71.440 52.163
316511 96.426 -71.434 52.151
316512 96.499 -71.433 52.136
316513 96.534 -71.449 52.118
316514 96.570 -71.466 52.095
316515 96.604 -71.482 52.068
316516 96.640 -71.498 52.039
316517 96.674 -71.515 52.007
316518 96.710 -71.531 51.974
316519 96.744 -71.548 51.944
316520 96.780 -71.564 51.914
316521 96.808 -71.583 51.884
316522 96.835 -71.603 51.857
316523 96.862 -71.622 51.832
316524 96.898 -71.628 51.804
316525 96.937 -71.627 51.776
316526 96.975 -71.627 51.745
316527 97.011 -71.628 51.707
316528 97.043 -71.631 51.669
316529 97.077 -71.634 51.631
316530 97.110 -71.638 51.593
316531 97.150 -71.649 51.559
316532 97.190 -71.660 51.528
316533 97.230 -71.672 51.494
316534 97.270 -71.683 51.459
316535 97.309 -71.694 51.426
316536 97.349 -71.706 51.393
316537 97.389 -71.717 51.363
316538 97.429 -71.729 51.333
316539 97.469 -71.740 51.303
316540 97.509 -71.751 51.269
316541 97.545 -71.758 51.235
316542 97.580 -71.763 51.201
316543 97.615 -71.768 51.167
316544 97.667 -71.784 51.134
316545 97.695 -71.785 51.102
316546 97.722 -71.788 51.068
316547 97.750 -71.789 51.032
316548 97.778 -71.791 50.994
316549 97.803 -71.793 50.954
316550 97.828 -71.794 50.915
316551 97.854 -71.795 50.878
316552 97.878 -71.797 50.844
316553 97.903 -71.798 50.807
316554 97.929 -71.799 50.767
316555 97.955 -71.807 50.725
316556 97.982 -71.815 50.680
316557 98.009 -71.823 50.633
316558 98.036 -71.832 50.584
316559 98.063 -71.840 50.534
316560 98.090 -71.848 50.485
316561 98.117 -71.857 50.438
316562 98.145 -71.864 50.393
316563 98.173 -71.873 50.347
316564 98.201 -71.880 50.299
316565 98.229 -71.888 50.250
316566 98.270 -71.908 50.199
316567 98.310 -71.929 50.150
316568 98.351 -71.949 50.100
316569 98.392 -71.973 50.051
316570 98.432 -71.993 50.002
316571 98.464 -71.995 49.951
316572 98.496 -71.996 49.897
316573 98.528 -71.998 49.842
316574 98.560 -71.999 49.792
316575 98.592 -72.001 49.741
316576 98.624 -72.002 49.689
316577 98.660 -72.005 49.637
316578 98.698 -72.007 49.583
316579 98.734 -72.009 49.527
316580 98.772 -72.011 49.471
316581 98.813 -72.022 49.415
316582 98.857 -72.037 49.360
316583 98.901 -72.052 49.308
316584 98.943 -72.067 49.255
316585 98.977 -72.081 49.201
316586 99.011 -72.094 49.147
316587 99.045 -72.108 49.093
316588 99.079 -72.121 49.041
316589 99.113 -72.134 48.989
316590 99.137 -72.151 48.935
316591 99.159 -72.168 48.883
316592 99.181 -72.184 48.831
316593 99.204 -72.201 48.781
316594 99.226 -72.218 48.730
316595 99.248 -72.234 48.681
316596 99.271 -72.251 48.632
316597 99.293 -72.268 48.584
316598 99.315 -72.284 48.536
316599 99.352 -72.298 48.490
316600 99.399 -72.309 48.444
316601 99.445 -72.321 48.397
316602 99.492 -72.333 48.350
316603 99.539 -72.345 48.302
316604 99.586 -72.356 48.255
316605 99.633 -72.368 48.208
316606 99.680 -72.381 48.163
316607 99.726 -72.392 48.118
316608 99.773 -72.404 48.073
316609 99.820 -72.415 48.029
316610 99.846 -72.428 47.984
316611 99.873 -72.440 47.937
316612 99.900 -72.451 47.894
316613 99.926 -72.463 47.850
316614 99.953 -72.475 47.806
316615 99.980 -72.487 47.762
316616 100.010 -72.498 47.719
316617 100.031 -72.530 47.676
316618 100.050 -72.561 47.634
316619 100.069 -72.592 47.593
316620 100.093 -72.609 47.551
316621 100.118 -72.625 47.510
316622 100.142 -72.640 47.469
316623 100.178 -72.660 47.428
316624 100.209 -72.649 47.387
316625 100.239 -72.639 47.348
316626 100.270 -72.628 47.310
316627 100.300 -72.619 47.272
316628 100.330 -72.609 47.234
316629 100.361 -72.598 47.197
316630 100.391 -72.588 47.160
316631 100.422 -72.578 47.124
316632 100.452 -72.568 47.088
316633 100.482 -72.558 47.054
316634 100.513 -72.547 47.020
316635 100.543 -72.537 46.985
316636 100.574 -72.527 46.951
316637 100.609 -72.547 46.911
316638 100.646 -72.572 46.872
316639 100.684 -72.598 46.834
316640 100.721 -72.623 46.795
316641 100.758 -72.648 46.756
316642 100.795 -72.673 46.715
316643 100.833 -72.698 46.672
316644 100.870 -72.723 46.629
316645 100.907 -72.748 46.587
316646 100.936 -72.760 46.545
316647 100.965 -72.773 46.504
316648 100.993 -72.785 46.464
316649 101.022 -72.797 46.422
316650 101.051 -72.810 46.382
316651 101.080 -72.823 46.343
316652 101.109 -72.835 46.305
316653 101.138 -72.847 46.267
316654 101.181 -72.857 46.230
316655 101.223 -72.866 46.193
316656 101.266 -72.875 46.156
316657 101.309 -72.884 46.120
316658 101.352 -72.893 46.084
316659 101.395 -72.903 46.048
316660 101.440 -72.902 46.012
316661 101.476 -72.893 45.976
316662 101.494 -72.898 45.940
316663 101.512 -72.904 45.904
316664 101.530 -72.910 45.868
316665 101.548 -72.915 45.832
316666 101.565 -72.921 45.797
316667 101.583 -72.926 45.760
316668 101.601 -72.932 45.724
316669 101.619 -72.938 45.687
316670 101.639 -72.949 45.648
316671 101.668 -72.976 45.611
316672 101.698 -72.999 45.572
316673 101.739 -73.012 45.532
316674 101.778 -73.026 45.495
316675 101.819 -73.039 45.456
316676 101.858 -73.051 45.419
316677 101.899 -73.065 45.381
316678 101.939 -73.078 45.344
316679 101.979 -73.091 45.308
316680 102.020 -73.102 45.272
316681 102.063 -73.112 45.237
316682 102.106 -73.122 45.202
316683 102.144 -73.128 45.167
317174 95.079 -71.101 52.465
317175 95.106 -71.115 52.449
317176 95.132 -71.130 52.435
317177 95.158 -71.144 52.421
317178 95.154 -71.166 52.406
317179 95.126 -71.194 52.393
317180 95.121 -71.219 52.380
317181 95.158 -71.237 52.367
317182 95.195 -71.255 52.355
317183 95.232 -71.273 52.342
317184 95.269 -71.291 52.331
317185 95.306 -71.309 52.319
317186 95.342 -71.327 52.306
317187 95.380 -71.345 52.295
317188 95.416 -71.363 52.282
317189 95.446 -71.368 52.270
317190 95.460 -71.343 52.258
317191 95.475 -71.320 52.246
317192 95.516 -71.328 52.235
317193 95.566 -71.344 52.223
317194 95.614 -71.353 52.212
317195 95.659 -71.363 52.199
317196 95.686 -71.373 52.188
317197 95.713 -71.382 52.178
317198 95.741 -71.392 52.169
317199 95.770 -71.400 52.161
317200 95.805 -71.403 52.155
317201 95.846 -71.400 52.150
317202 95.887 -71.397 52.147
317203 95.927 -71.395 52.148
317204 95.972 -71.396 52.149
317205 96.027 -71.406 52.152
317206 96.082 -71.415 52.158
317207 96.136 -71.425 52.165
317208 96.195 -71.419 52.168
317209 96.254 -71.412 52.169
317210 96.314 -71.406 52.168
317211 96.373 -71.399 52.164
317212 96.432 -71.392 52.154
317213 96.506 -71.393 52.141
317214 96.541 -71.409 52.124
317215 96.576 -71.426 52.104
317216 96.611 -71.443 52.079
317217 96.646 -71.459 52.052
317218 96.681 -71.475 52.023
317219 96.716 -71.492 51.992
317220 96.751 -71.509 51.964
317221 96.786 -71.525 51.932
317222 96.821 -71.541 51.900
317223 96.849 -71.561 51.869
317224 96.876 -71.580 51.840
317225 96.912 -71.585 51.812
317226 96.951 -71.584 51.784
317227 96.986 -71.585 51.754
317228 97.019 -71.588 51.719
317229 97.052 -71.591 51.682
317230 97.085 -71.594 51.643
317231 97.120 -71.600 51.606
317232 97.160 -71.612 51.571
317233 97.200 -71.623 51.539
317234 97.240 -71.635 51.505
317235 97.280 -71.646 51.469
317236 97.319 -71.657 51.433
317237 97.359 -71.669 51.401
317238 97.399 -71.680 51.370
317239 97.439 -71.691 51.341
317240 97.479 -71.703 51.311
317241 97.519 -71.714 51.277
317242 97.559 -71.726 51.242
317243 97.598 -71.737 51.207
317244 97.636 -71.745 51.173
317245 97.671 -71.751 51.139
317246 97.711 -71.759 51.106
317247 97.738 -71.761 51.071
317248 97.765 -71.763 51.034
317249 97.790 -71.764 50.995
317250 97.815 -71.765 50.958
317251 97.840 -71.766 50.921
317252 97.865 -71.768 50.886
317253 97.890 -71.769 50.851
317254 97.915 -71.770 50.815
317255 97.940 -71.771 50.776
317256 97.966 -71.778 50.734
317257 97.994 -71.786 50.691
317258 98.020 -71.795 50.644
317259 98.048 -71.803 50.595
317260 98.075 -71.811 50.545
317261 98.101 -71.820 50.496
317262 98.129 -71.828 50.453
317263 98.157 -71.835 50.409
317264 98.184 -71.843 50.364
317265 98.213 -71.851 50.317
317266 98.240 -71.859 50.270
317267 98.278 -71.876 50.220
317268 98.318 -71.896 50.169
317269 98.359 -71.917 50.117
317270 98.401 -71.940 50.064
317271 98.442 -71.964 50.011
317272 98.476 -71.971 49.959
317273 98.508 -71.973 49.907
317274 98.540 -71.974 49.855
317275 98.572 -71.976 49.805
317276 98.604 -71.977 49.753
317277 98.636 -71.979 49.700
317278 98.668 -71.980 49.647
317279 98.700 -71.982 49.591
317280 98.736 -71.984 49.535
317281 98.774 -71.986 49.479
317282 98.811 -71.989 49.423
317283 98.855 -72.004 49.366
317284 98.899 -72.019 49.312
317285 98.936 -72.033 49.259
317286 98.971 -72.047 49.205
317287 99.004 -72.060 49.150
317288 99.038 -72.073 49.096
317289 99.072 -72.087 49.042
317290 99.106 -72.101 48.989
317291 99.139 -72.114 48.935
317292 99.170 -72.128 48.883
317293 99.193 -72.145 48.832
317294 99.215 -72.161 48.781
317295 99.237 -72.178 48.730
317296 99.260 -72.195 48.682
317297 99.282 -72.211 48.633
317298 99.305 -72.228 48.584
317299 99.327 -72.245 48.536
317300 99.365 -72.258 48.490
317301 99.412 -72.270 48.444
317302 99.459 -72.282 48.397
317303 99.506 -72.293 48.350
317304 99.553 -72.305 48.302
317305 99.600 -72.317 48.255
317306 99.647 -72.329 48.209
317307 99.694 -72.340 48.164
317308 99.740 -72.352 48.119
317309 99.787 -72.364 48.075
317310 99.834 -72.376 48.030
317311 99.879 -72.387 47.985
317312 99.907 -72.399 47.939
317313 99.933 -72.412 47.895
317314 99.960 -72.424 47.852
317315 99.981 -72.437 47.809
317316 99.999 -72.449 47.765
317317 100.027 -72.453 47.723
317318 100.046 -72.483 47.680
317319 100.065 -72.514 47.638
317320 100.086 -72.539 47.597
317321 100.112 -72.555 47.555
317322 100.136 -72.570 47.514
317323 100.161 -72.586 47.473
317324 100.192 -72.606 47.433
317325 100.226 -72.616 47.393
317326 100.257 -72.606 47.354
317327 100.287 -72.597 47.316
317328 100.318 -72.586 47.279
317329 100.348 -72.576 47.241
317330 100.378 -72.565 47.205
317331 100.409 -72.556 47.169
317332 100.439 -72.546 47.133
317333 100.470 -72.535 47.099
317334 100.500 -72.525 47.066
317335 100.530 -72.515 47.033
317336 100.561 -72.505 46.999
317337 100.592 -72.503 46.965
317338 100.624 -72.522 46.926
317339 100.656 -72.540 46.886
317340 100.691 -72.560 46.847
317341 100.728 -72.585 46.807
317342 100.765 -72.611 46.767
317343 100.802 -72.636 46.726
317344 100.839 -72.661 46.683
317345 100.876 -72.686 46.640
317346 100.914 -72.711 46.599
317347 100.948 -72.731 46.557
317348 100.977 -72.744 46.515
317349 101.005 -72.756 46.473
317350 101.034 -72.768 46.431
317351 101.063 -72.781 46.392
317352 101.092 -72.793 46.353
317353 101.121 -72.805 46.314
317354 101.162 -72.815 46.276
317355 101.205 -72.824 46.238
317356 101.249 -72.833 46.200
317357 101.291 -72.842 46.162
317358 101.334 -72.852 46.126
317359 101.377 -72.861 46.090
317360 101.420 -72.870 46.053
317361 101.463 -72.879 46.017
317362 101.494 -72.888 45.981
317363 101.512 -72.894 45.945
317364 101.529 -72.900 45.908
317365 101.547 -72.905 45.872
317366 101.565 -72.911 45.836
317367 101.583 -72.916 45.801
317368 101.600 -72.922 45.764
317369 101.618 -72.928 45.727
317370 101.636 -72.933 45.689
317371 101.653 -72.939 45.651
317372 101.674 -72.949 45.613
317373 101.711 -72.966 45.575
317374 101.751 -72.979 45.536
317375 101.792 -72.992 45.499
317376 101.831 -73.005 45.461
317377 101.872 -73.018 45.423
317378 101.911 -73.032 45.386
317379 101.951 -73.045 45.349
317380 101.991 -73.058 45.313
317381 102.033 -73.068 45.277
317382 102.077 -73.078 45.242
317383 102.119 -73.088 45.208
317873 95.025 -71.027 52.501
317874 95.051 -71.042 52.484
317875 95.077 -71.056 52.468
317876 95.104 -71.071 52.453
317877 95.130 -71.084 52.438
317878 95.157 -71.099 52.425
317879 95.132 -71.127 52.411
317880 95.104 -71.155 52.398
317881 95.136 -71.174 52.385
317882 95.173 -71.191 52.372
317883 95.210 -71.209 52.360
317884 95.247 -71.227 52.348
317885 95.283 -71.246 52.337
317886 95.321 -71.263 52.325
317887 95.357 -71.281 52.313
317888 95.395 -71.299 52.301
317889 95.431 -71.318 52.289
317890 95.468 -71.335 52.277
317891 95.488 -71.321 52.265
317892 95.502 -71.298 52.253
317893 95.517 -71.274 52.242
317894 95.562 -71.287 52.230
317895 95.611 -71.299 52.219
317896 95.659 -71.309 52.207
317897 95.688 -71.318 52.196
317898 95.716 -71.328 52.187
317899 95.743 -71.337 52.177
317900 95.775 -71.343 52.169
317901 95.810 -71.346 52.163
317902 95.851 -71.343 52.159
317903 95.892 -71.340 52.155
317904 95.933 -71.338 52.156
317905 95.982 -71.343 52.156
317906 96.036 -71.353 52.158
317907 96.091 -71.362 52.162
317908 96.146 -71.372 52.167
317909 96.201 -71.378 52.168
317910 96.261 -71.371 52.167
317911 96.320 -71.364 52.165
317912 96.379 -71.357 52.162
317913 96.439 -71.350 52.154
317914 96.512 -71.354 52.144
317915 96.547 -71.371 52.130
317916 96.583 -71.387 52.112
317917 96.617 -71.403 52.091
317918 96.653 -71.420 52.066
317919 96.687 -71.437 52.039
317920 96.723 -71.453 52.012
317921 96.757 -71.469 51.984
317922 96.793 -71.486 51.951
317923 96.827 -71.502 51.917
317924 96.863 -71.519 51.883
317925 96.890 -71.538 51.852
317926 96.927 -71.541 51.823
317927 96.962 -71.544 51.794
317928 96.995 -71.546 51.764
317929 97.027 -71.549 51.731
317930 97.061 -71.551 51.694
317931 97.093 -71.554 51.656
317932 97.131 -71.563 51.619
317933 97.170 -71.574 51.584
317934 97.210 -71.585 51.551
317935 97.250 -71.597 51.516
317936 97.290 -71.608 51.479
317937 97.329 -71.619 51.443
317938 97.369 -71.631 51.410
317939 97.409 -71.642 51.378
317940 97.449 -71.654 51.349
317941 97.489 -71.665 51.319
317942 97.529 -71.676 51.285
317943 97.569 -71.688 51.249
317944 97.608 -71.700 51.214
317945 97.648 -71.711 51.179
317946 97.688 -71.722 51.145
317947 97.727 -71.732 51.111
317948 97.751 -71.734 51.076
317949 97.777 -71.735 51.040
317950 97.802 -71.736 51.003
317951 97.827 -71.738 50.967
317952 97.852 -71.738 50.932
317953 97.877 -71.740 50.897
317954 97.902 -71.741 50.862
317955 97.927 -71.742 50.824
317956 97.952 -71.744 50.784
317957 97.978 -71.748 50.743
317958 98.005 -71.757 50.698
317959 98.032 -71.766 50.653
317960 98.059 -71.774 50.605
317961 98.086 -71.782 50.556
317962 98.113 -71.791 50.509
317963 98.141 -71.798 50.467
317964 98.168 -71.806 50.426
317965 98.197 -71.814 50.383
317966 98.224 -71.822 50.336
317967 98.252 -71.829 50.289
317968 98.286 -71.843 50.238
317969 98.326 -71.864 50.185
317970 98.367 -71.884 50.131
317971 98.409 -71.907 50.076
317972 98.450 -71.932 50.021
317973 98.488 -71.948 49.967
317974 98.520 -71.949 49.915
317975 98.552 -71.951 49.865
317976 98.584 -71.952 49.813
317977 98.616 -71.954 49.761
317978 98.648 -71.955 49.708
317979 98.680 -71.957 49.654
317980 98.712 -71.958 49.597
317981 98.744 -71.960 49.541
317982 98.776 -71.961 49.486
317983 98.813 -71.964 49.430
317984 98.853 -71.971 49.373
317985 98.896 -71.986 49.317
317986 98.930 -71.999 49.262
317987 98.963 -72.013 49.207
317988 98.997 -72.026 49.153
317989 99.031 -72.039 49.098
317990 99.065 -72.053 49.044
317991 99.099 -72.067 48.990
317992 99.133 -72.080 48.936
317993 99.166 -72.093 48.884
317994 99.200 -72.107 48.832
317995 99.227 -72.122 48.781
317996 99.250 -72.139 48.730
317997 99.272 -72.155 48.682
317998 99.294 -72.172 48.634
317999 99.316 -72.189 48.585
318000 99.339 -72.205 48.535
318001 99.379 -72.218 48.489
318002 99.426 -72.230 48.443
318003 99.473 -72.242 48.396
318004 99.520 -72.254 48.349
318005 99.567 -72.265 48.302
318006 99.614 -72.277 48.256
318007 99.660 -72.290 48.210
318008 99.707 -72.301 48.165
318009 99.754 -72.313 48.120
318010 99.801 -72.324 48.076
318011 99.848 -72.337 48.031
318012 99.895 -72.348 47.986
318013 99.930 -72.362 47.940
318014 99.947 -72.376 47.896
318015 99.964 -72.390 47.854
318016 99.986 -72.393 47.811
318017 100.010 -72.395 47.769
318018 100.043 -72.405 47.726
318019 100.062 -72.436 47.683
318020 100.081 -72.467 47.641
318021 100.105 -72.484 47.600
318022 100.130 -72.500 47.559
318023 100.154 -72.515 47.518
318024 100.179 -72.531 47.478
318025 100.205 -72.547 47.438
318026 100.242 -72.572 47.399
318027 100.274 -72.574 47.360
318028 100.304 -72.564 47.323
318029 100.335 -72.553 47.286
318030 100.365 -72.544 47.249
318031 100.396 -72.534 47.213
318032 100.426 -72.523 47.177
318033 100.457 -72.513 47.142
318034 100.487 -72.503 47.110
318035 100.517 -72.493 47.077
318036 100.548 -72.483 47.045
318037 100.578 -72.472 47.012
318038 100.610 -72.482 46.978
318039 100.642 -72.500 46.940
318040 100.675 -72.519 46.901
318041 100.707 -72.537 46.862
318042 100.739 -72.554 46.822
318043 100.772 -72.573 46.780
318044 100.809 -72.599 46.738
318045 100.846 -72.624 46.695
318046 100.884 -72.649 46.652
318047 100.921 -72.674 46.610
318048 100.958 -72.699 46.567
318049 100.988 -72.714 46.526
318050 101.018 -72.726 46.484
318051 101.046 -72.738 46.442
318052 101.075 -72.751 46.402
318053 101.104 -72.763 46.362
318054 101.145 -72.773 46.324
318055 101.188 -72.782 46.285
318056 101.231 -72.791 46.246
318057 101.273 -72.801 46.208
318058 101.316 -72.810 46.170
318059 101.359 -72.819 46.133
318060 101.400 -72.827 46.097
318061 101.439 -72.835 46.060
318062 101.478 -72.841 46.023
318063 101.513 -72.852 45.987
318064 101.533 -72.860 45.951
318065 101.550 -72.866 45.914
318066 101.568 -72.873 45.877
318067 101.586 -72.879 45.841
318068 101.604 -72.886 45.804
318069 101.621 -72.892 45.768
318070 101.639 -72.899 45.730
318071 101.656 -72.906 45.692
318072 101.674 -72.912 45.655
318073 101.692 -72.919 45.617
318074 101.724 -72.932 45.579
318075 101.764 -72.945 45.541
318076 101.804 -72.958 45.503
318077 101.844 -72.972 45.466
318078 101.884 -72.985 45.429
318079 101.924 -72.998 45.391
318080 101.964 -73.011 45.354
318081 102.004 -73.024 45.319
318082 102.046 -73.035 45.284
318083 102.090 -73.044 45.248
318084 102.132 -73.054 45.214
318571 94.944 -70.939 52.556
318572 94.970 -70.953 52.538
318573 94.996 -70.967 52.521
318574 95.023 -70.982 52.504
318575 95.049 -70.996 52.487
318576 95.075 -71.011 52.472
318577 95.102 -71.025 52.458
318578 95.128 -71.039 52.444
318579 95.137 -71.058 52.430
318580 95.110 -71.086 52.417
318581 95.114 -71.110 52.404
318582 95.151 -71.127 52.391
318583 95.188 -71.146 52.379
318584 95.225 -71.164 52.366
318585 95.262 -71.181 52.354
318586 95.299 -71.199 52.343
318587 95.336 -71.218 52.331
318588 95.373 -71.236 52.319
318589 95.409 -71.253 52.307
318590 95.447 -71.271 52.295
318591 95.483 -71.290 52.283
318592 95.516 -71.300 52.273
318593 95.531 -71.276 52.261
318594 95.545 -71.252 52.249
318595 95.559 -71.227 52.237
318596 95.607 -71.244 52.226
318597 95.656 -71.253 52.215
318598 95.691 -71.263 52.204
318599 95.719 -71.272 52.194
318600 95.746 -71.282 52.184
318601 95.780 -71.285 52.177
318602 95.815 -71.289 52.172
318603 95.855 -71.287 52.168
318604 95.896 -71.284 52.164
318605 95.937 -71.281 52.163
318606 95.991 -71.290 52.165
318607 96.046 -71.300 52.166
318608 96.101 -71.310 52.167
318609 96.156 -71.319 52.170
318610 96.210 -71.329 52.168
318611 96.267 -71.329 52.165
318612 96.327 -71.322 52.162
318613 96.386 -71.315 52.158
318614 96.445 -71.309 52.152
318615 96.519 -71.315 52.144
318616 96.554 -71.331 52.134
318617 96.589 -71.348 52.121
318618 96.624 -71.364 52.102
318619 96.659 -71.381 52.080
318620 96.694 -71.397 52.057
318621 96.729 -71.413 52.032
318622 96.764 -71.430 52.005
318623 96.799 -71.447 51.971
318624 96.834 -71.463 51.935
318625 96.869 -71.479 51.899
318626 96.904 -71.496 51.865
318627 96.937 -71.501 51.835
318628 96.970 -71.503 51.805
318629 97.003 -71.506 51.775
318630 97.036 -71.509 51.742
318631 97.069 -71.512 51.706
318632 97.102 -71.515 51.669
318633 97.141 -71.525 51.632
318634 97.180 -71.537 51.596
318635 97.220 -71.548 51.563
318636 97.260 -71.560 51.527
318637 97.300 -71.571 51.491
318638 97.339 -71.582 51.455
318639 97.379 -71.594 51.419
318640 97.419 -71.605 51.387
318641 97.459 -71.616 51.357
318642 97.499 -71.628 51.328
318643 97.539 -71.639 51.294
318644 97.579 -71.650 51.258
318645 97.618 -71.662 51.223
318646 97.655 -71.674 51.188
318647 97.690 -71.686 51.152
318648 97.726 -71.698 51.119
318649 97.762 -71.706 51.085
318650 97.788 -71.707 51.049
318651 97.813 -71.709 51.016
318652 97.839 -71.710 50.981
318653 97.864 -71.711 50.947
318654 97.888 -71.713 50.912
318655 97.914 -71.713 50.876
318656 97.939 -71.715 50.836
318657 97.964 -71.716 50.795
318658 97.990 -71.719 50.752
318659 98.016 -71.728 50.707
318660 98.043 -71.736 50.662
318661 98.070 -71.745 50.615
318662 98.097 -71.754 50.568
318663 98.125 -71.761 50.522
318664 98.152 -71.770 50.482
318665 98.180 -71.777 50.442
318666 98.208 -71.785 50.400
318667 98.236 -71.793 50.355
318668 98.264 -71.801 50.306
318669 98.294 -71.810 50.253
318670 98.335 -71.830 50.198
318671 98.375 -71.851 50.142
318672 98.417 -71.874 50.086
318673 98.458 -71.898 50.030
318674 98.500 -71.923 49.975
318675 98.532 -71.926 49.922
318676 98.564 -71.927 49.871
318677 98.596 -71.929 49.819
318678 98.628 -71.930 49.767
318679 98.660 -71.932 49.713
318680 98.692 -71.933 49.658
318681 98.724 -71.935 49.602
318682 98.756 -71.936 49.546
318683 98.788 -71.938 49.490
318684 98.820 -71.939 49.435
318685 98.852 -71.941 49.379
318686 98.889 -71.952 49.321
318687 98.923 -71.966 49.264
318688 98.956 -71.979 49.209
318689 98.991 -71.992 49.154
318690 99.024 -72.006 49.099
318691 99.058 -72.019 49.046
318692 99.092 -72.033 48.992
318693 99.126 -72.046 48.938
318694 99.160 -72.059 48.884
318695 99.194 -72.073 48.832
318696 99.227 -72.086 48.782
318697 99.261 -72.100 48.731
318698 99.284 -72.117 48.683
318699 99.306 -72.133 48.634
318700 99.328 -72.150 48.585
318701 99.351 -72.167 48.536
318702 99.393 -72.179 48.488
318703 99.440 -72.191 48.441
318704 99.487 -72.202 48.395
318705 99.534 -72.215 48.349
318706 99.581 -72.226 48.302
318707 99.628 -72.238 48.256
318708 99.675 -72.249 48.210
318709 99.721 -72.262 48.165
318710 99.766 -72.274 48.122
318711 99.807 -72.288 48.077
318712 99.849 -72.302 48.031
318713 99.890 -72.315 47.986
318714 99.929 -72.329 47.941
318715 99.949 -72.338 47.899
318716 99.973 -72.339 47.856
318717 99.997 -72.340 47.814
318718 100.021 -72.342 47.771
318719 100.059 -72.358 47.729
318720 100.078 -72.389 47.686
318721 100.099 -72.415 47.644
318722 100.124 -72.431 47.604
318723 100.148 -72.446 47.563
318724 100.173 -72.462 47.522
318725 100.198 -72.477 47.482
318726 100.222 -72.493 47.441
318727 100.254 -72.513 47.403
318728 100.291 -72.537 47.365
318729 100.322 -72.531 47.329
318730 100.352 -72.521 47.293
318731 100.382 -72.511 47.256
318732 100.413 -72.501 47.220
318733 100.443 -72.490 47.186
318734 100.474 -72.481 47.152
318735 100.504 -72.470 47.121
318736 100.535 -72.460 47.091
318737 100.565 -72.450 47.059
318738 100.596 -72.443 47.026
318739 100.628 -72.462 46.991
318740 100.661 -72.479 46.954
318741 100.693 -72.497 46.915
318742 100.725 -72.515 46.876
318743 100.757 -72.534 46.836
318744 100.789 -72.551 46.794
318745 100.822 -72.569 46.752
318746 100.854 -72.587 46.709
318747 100.890 -72.612 46.665
318748 100.927 -72.637 46.623
318749 100.965 -72.662 46.580
318750 101.000 -72.685 46.537
318751 101.029 -72.697 46.495
318752 101.058 -72.709 46.453
318753 101.087 -72.722 46.411
318754 101.127 -72.732 46.372
318755 101.170 -72.741 46.333
318756 101.213 -72.750 46.294
318757 101.256 -72.759 46.255
318758 101.296 -72.767 46.216
318759 101.335 -72.775 46.178
318760 101.375 -72.782 46.140
318761 101.413 -72.789 46.104
318762 101.452 -72.796 46.067
318763 101.491 -72.803 46.030
318764 101.526 -72.813 45.993
318765 101.554 -72.826 45.956
318766 101.572 -72.832 45.919
318767 101.590 -72.838 45.882
318768 101.607 -72.845 45.845
318769 101.625 -72.851 45.808
318770 101.642 -72.858 45.771
318771 101.660 -72.865 45.734
318772 101.678 -72.871 45.696
318773 101.696 -72.878 45.659
318774 101.712 -72.885 45.621
318775 101.737 -72.898 45.583
318776 101.777 -72.912 45.545
318777 101.817 -72.925 45.507
318778 101.857 -72.938 45.472
318779 101.897 -72.951 45.434
318780 101.937 -72.964 45.397
318781 101.977 -72.977 45.361
318782 102.017 -72.991 45.326
318783 102.060 -73.001 45.291
318784 102.103 -73.010 45.256
318785 102.142 -73.017 45.222
319269 94.884 -70.908 52.612
319270 94.903 -70.902 52.592
319271 94.922 -70.897 52.574
319272 94.942 -70.894 52.558
319273 94.968 -70.908 52.540
319274 94.994 -70.923 52.524
319275 95.021 -70.937 52.508
319276 95.047 -70.952 52.492
319277 95.073 -70.966 52.477
319278 95.100 -70.980 52.464
319279 95.126 -70.995 52.450
319280 95.115 -71.018 52.437
319281 95.092 -71.046 52.424
319282 95.129 -71.064 52.411
319283 95.166 -71.082 52.398
319284 95.203 -71.100 52.386
319285 95.240 -71.118 52.373
319286 95.277 -71.136 52.361
319287 95.314 -71.154 52.350
319288 95.351 -71.172 52.338
319289 95.388 -71.190 52.327
319290 95.424 -71.208 52.314
319291 95.462 -71.226 52.303
319292 95.498 -71.244 52.291
319293 95.535 -71.262 52.279
319294 95.558 -71.254 52.267
319295 95.572 -71.230 52.256
319296 95.587 -71.205 52.245
319297 95.601 -71.181 52.234
319298 95.652 -71.199 52.222
319299 95.693 -71.208 52.212
319300 95.721 -71.218 52.201
319301 95.751 -71.224 52.192
319302 95.786 -71.228 52.185
319303 95.821 -71.231 52.180
319304 95.861 -71.230 52.176
319305 95.901 -71.227 52.172
319306 95.946 -71.228 52.173
319307 96.001 -71.238 52.174
319308 96.055 -71.247 52.175
319309 96.110 -71.257 52.176
319310 96.165 -71.267 52.177
319311 96.220 -71.277 52.176
319312 96.274 -71.286 52.172
319313 96.333 -71.281 52.168
319314 96.392 -71.274 52.163
319315 96.453 -71.268 52.158
319316 96.525 -71.275 52.152
319317 96.560 -71.292 52.144
319318 96.596 -71.309 52.133
319319 96.630 -71.325 52.114
319320 96.666 -71.341 52.094
319321 96.700 -71.358 52.071
319322 96.736 -71.375 52.047
319323 96.770 -71.390 52.022
319324 96.806 -71.407 51.989
319325 96.841 -71.424 51.953
319326 96.878 -71.442 51.916
319327 96.913 -71.454 51.882
319328 96.946 -71.461 51.849
319329 96.978 -71.464 51.818
319330 97.011 -71.467 51.788
319331 97.044 -71.469 51.754
319332 97.077 -71.472 51.719
319333 97.111 -71.476 51.683
319334 97.151 -71.487 51.646
319335 97.190 -71.499 51.611
319336 97.230 -71.510 51.577
319337 97.270 -71.522 51.540
319338 97.310 -71.533 51.503
319339 97.350 -71.544 51.466
319340 97.389 -71.556 51.430
319341 97.429 -71.567 51.398
319342 97.469 -71.578 51.367
319343 97.509 -71.591 51.338
319344 97.546 -71.602 51.304
319345 97.582 -71.614 51.270
319346 97.617 -71.626 51.235
319347 97.653 -71.638 51.199
319348 97.688 -71.650 51.164
319349 97.724 -71.663 51.131
319350 97.760 -71.671 51.097
319351 97.797 -71.678 51.064
319352 97.825 -71.681 51.034
319353 97.850 -71.682 51.000
319354 97.875 -71.683 50.966
319355 97.900 -71.685 50.932
319356 97.925 -71.686 50.896
319357 97.950 -71.687 50.855
319358 97.975 -71.688 50.812
319359 98.001 -71.691 50.768
319360 98.028 -71.699 50.722
319361 98.055 -71.707 50.676
319362 98.082 -71.716 50.630
319363 98.108 -71.725 50.584
319364 98.136 -71.732 50.538
319365 98.164 -71.740 50.497
319366 98.192 -71.748 50.456
319367 98.220 -71.756 50.412
319368 98.248 -71.763 50.367
319369 98.276 -71.771 50.318
319370 98.304 -71.779 50.264
319371 98.343 -71.798 50.208
319372 98.383 -71.818 50.151
319373 98.425 -71.842 50.093
319374 98.466 -71.866 50.037
319375 98.508 -71.890 49.983
319376 98.544 -71.902 49.929
319377 98.576 -71.904 49.877
319378 98.608 -71.905 49.824
319379 98.640 -71.907 49.770
319380 98.672 -71.908 49.714
319381 98.704 -71.910 49.659
319382 98.736 -71.911 49.602
319383 98.768 -71.913 49.547
319384 98.800 -71.914 49.491
319385 98.832 -71.914 49.437
319386 98.863 -71.913 49.380
319387 98.892 -71.920 49.324
319388 98.922 -71.931 49.267
319389 98.954 -71.945 49.210
319390 98.987 -71.958 49.156
319391 99.019 -71.972 49.101
319392 99.051 -71.986 49.047
319393 99.085 -71.998 48.993
319394 99.119 -72.012 48.940
319395 99.153 -72.026 48.886
319396 99.187 -72.039 48.834
319397 99.220 -72.052 48.784
319398 99.254 -72.066 48.733
319399 99.288 -72.080 48.684
319400 99.317 -72.094 48.635
319401 99.340 -72.111 48.586
319402 99.362 -72.127 48.536
319403 99.407 -72.139 48.489
319404 99.454 -72.151 48.442
319405 99.501 -72.163 48.395
319406 99.548 -72.174 48.348
319407 99.594 -72.186 48.302
319408 99.636 -72.200 48.256
319409 99.678 -72.214 48.210
319410 99.719 -72.227 48.166
319411 99.761 -72.241 48.121
319412 99.802 -72.255 48.077
319413 99.843 -72.268 48.032
319414 99.884 -72.281 47.987
319415 99.926 -72.283 47.942
319416 99.960 -72.284 47.901
319417 99.984 -72.286 47.858
319418 100.008 -72.287 47.815
319419 100.034 -72.290 47.773
319420 100.074 -72.311 47.731
319421 100.093 -72.342 47.689
319422 100.117 -72.361 47.647
319423 100.142 -72.376 47.608
319424 100.167 -72.392 47.567
319425 100.191 -72.407 47.527
319426 100.216 -72.423 47.486
319427 100.241 -72.438 47.446
319428 100.267 -72.455 47.408
319429 100.303 -72.479 47.371
319430 100.339 -72.499 47.335
319431 100.370 -72.489 47.299
319432 100.400 -72.478 47.264
319433 100.430 -72.468 47.228
319434 100.461 -72.458 47.195
319435 100.491 -72.448 47.163
319436 100.522 -72.438 47.133
319437 100.552 -72.428 47.105
319438 100.582 -72.418 47.073
319439 100.614 -72.422 47.039
319440 100.646 -72.440 47.004
319441 100.679 -72.459 46.967
319442 100.711 -72.477 46.929
319443 100.743 -72.494 46.889
319444 100.775 -72.512 46.849
319445 100.808 -72.531 46.807
319446 100.840 -72.548 46.765
319447 100.872 -72.566 46.722
319448 100.904 -72.584 46.678
319449 100.936 -72.602 46.636
319450 100.972 -72.625 46.592
319451 101.009 -72.650 46.549
319452 101.041 -72.667 46.507
319453 101.070 -72.680 46.463
319454 101.109 -72.690 46.421
319455 101.152 -72.699 46.381
319456 101.193 -72.707 46.342
319457 101.232 -72.715 46.302
319458 101.271 -72.722 46.264
319459 101.310 -72.728 46.225
319460 101.349 -72.736 46.186
319461 101.388 -72.743 46.148
319462 101.426 -72.750 46.111
319463 101.465 -72.757 46.074
319464 101.504 -72.764 46.037
319465 101.539 -72.776 45.999
319466 101.571 -72.788 45.962
319467 101.594 -72.797 45.924
319468 101.611 -72.804 45.887
319469 101.629 -72.810 45.850
319470 101.646 -72.817 45.813
319471 101.664 -72.824 45.776
319472 101.682 -72.830 45.738
319473 101.699 -72.837 45.701
319474 101.717 -72.844 45.664
319475 101.734 -72.852 45.626
319476 101.750 -72.865 45.588
319477 101.790 -72.878 45.551
319478 101.830 -72.891 45.514
319479 101.870 -72.904 45.477
319480 101.910 -72.917 45.440
319481 101.950 -72.931 45.404
319482 101.989 -72.944 45.368
319483 102.030 -72.957 45.333
319484 102.073 -72.967 45.298
319485 102.116 -72.976 45.263
319969 94.860 -70.879 52.632
319970 94.886 -70.873 52.613
319971 94.905 -70.867 52.594
319972 94.924 -70.862 52.576
319973 94.943 -70.858 52.560
319974 94.966 -70.863 52.544
319975 94.992 -70.877 52.528
319976 95.019 -70.892 52.512
319977 95.045 -70.906 52.497
319978 95.071 -70.920 52.484
319979 95.098 -70.935 52.471
319980 95.121 -70.950 52.458
319981 95.093 -70.979 52.445
319982 95.107 -71.000 52.432
319983 95.144 -71.018 52.420
319984 95.181 -71.036 52.407
319985 95.218 -71.054 52.393
319986 95.255 -71.072 52.381
319987 95.292 -71.090 52.369
319988 95.329 -71.108 52.358
319989 95.366 -71.126 52.346
319990 95.403 -71.144 52.334
319991 95.440 -71.162 52.322
319992 95.476 -71.180 52.310
319993 95.514 -71.198 52.299
319994 95.550 -71.216 52.286
319995 95.586 -71.232 52.275
319996 95.601 -71.208 52.264
319997 95.615 -71.183 52.254
319998 95.629 -71.159 52.242
319999 95.647 -71.140 52.230
320000 95.696 -71.153 52.219
320001 95.724 -71.163 52.209
320002 95.757 -71.168 52.200
320003 95.792 -71.171 52.193
320004 95.826 -71.174 52.188
320005 95.866 -71.173 52.184
320006 95.906 -71.171 52.180
320007 95.956 -71.176 52.181
320008 96.011 -71.185 52.183
320009 96.065 -71.195 52.185
320010 96.120 -71.205 52.186
320011 96.175 -71.214 52.188
320012 96.229 -71.224 52.186
320013 96.284 -71.234 52.183
320014 96.339 -71.239 52.178
320015 96.399 -71.232 52.174
320016 96.466 -71.227 52.169
320017 96.532 -71.237 52.164
320018 96.567 -71.252 52.157
320019 96.602 -71.269 52.146
320020 96.637 -71.286 52.128
320021 96.672 -71.302 52.108
320022 96.707 -71.318 52.086
320023 96.742 -71.335 52.062
320024 96.779 -71.353 52.038
320025 96.816 -71.371 52.006
320026 96.853 -71.389 51.971
320027 96.889 -71.403 51.935
320028 96.922 -71.410 51.899
320029 96.954 -71.418 51.864
320030 96.987 -71.425 51.832
320031 97.020 -71.428 51.801
320032 97.052 -71.430 51.767
320033 97.085 -71.433 51.732
320034 97.121 -71.439 51.697
320035 97.161 -71.450 51.661
320036 97.201 -71.462 51.625
320037 97.240 -71.473 51.590
320038 97.280 -71.484 51.553
320039 97.320 -71.496 51.515
320040 97.360 -71.507 51.478
320041 97.399 -71.519 51.443
320042 97.438 -71.530 51.411
320043 97.473 -71.542 51.380
320044 97.509 -71.554 51.350
320045 97.544 -71.566 51.317
320046 97.580 -71.578 51.282
320047 97.615 -71.591 51.248
320048 97.651 -71.603 51.213
320049 97.686 -71.615 51.179
320050 97.722 -71.627 51.146
320051 97.758 -71.636 51.113
320052 97.795 -71.643 51.081
320053 97.831 -71.650 51.052
320054 97.862 -71.654 51.019
320055 97.887 -71.656 50.986
320056 97.912 -71.657 50.954
320057 97.937 -71.658 50.919
320058 97.962 -71.660 50.877
320059 97.987 -71.660 50.832
320060 98.012 -71.662 50.786
320061 98.039 -71.670 50.739
320062 98.066 -71.679 50.692
320063 98.093 -71.687 50.646
320064 98.120 -71.695 50.601
320065 98.148 -71.703 50.556
320066 98.176 -71.711 50.515
320067 98.204 -71.719 50.471
320068 98.232 -71.726 50.426
320069 98.260 -71.735 50.380
320070 98.288 -71.742 50.330
320071 98.315 -71.750 50.275
320072 98.351 -71.765 50.217
320073 98.391 -71.785 50.159
320074 98.433 -71.808 50.102
320075 98.475 -71.833 50.046
320076 98.516 -71.857 49.991
320077 98.556 -71.879 49.937
320078 98.588 -71.880 49.883
320079 98.620 -71.882 49.829
320080 98.652 -71.883 49.774
320081 98.684 -71.885 49.718
320082 98.716 -71.886 49.661
320083 98.748 -71.887 49.604
320084 98.780 -71.885 49.547
320085 98.811 -71.884 49.492
320086 98.843 -71.883 49.436
320087 98.874 -71.882 49.380
320088 98.905 -71.885 49.325
320089 98.931 -71.901 49.269
320090 98.958 -71.910 49.212
320091 98.991 -71.923 49.158
320092 99.023 -71.937 49.103
320093 99.055 -71.951 49.048
320094 99.087 -71.964 48.995
320095 99.120 -71.978 48.942
320096 99.152 -71.991 48.888
320097 99.184 -72.005 48.837
320098 99.217 -72.018 48.786
320099 99.249 -72.032 48.736
320100 99.281 -72.045 48.687
320101 99.315 -72.059 48.637
320102 99.349 -72.072 48.588
320103 99.374 -72.088 48.538
320104 99.421 -72.100 48.490
320105 99.465 -72.112 48.443
320106 99.507 -72.126 48.395
320107 99.548 -72.139 48.348
320108 99.589 -72.153 48.302
320109 99.630 -72.166 48.256
320110 99.671 -72.180 48.211
320111 99.712 -72.193 48.166
320112 99.753 -72.206 48.122
320113 99.794 -72.219 48.077
320114 99.836 -72.228 48.032
320115 99.878 -72.229 47.988
320116 99.921 -72.230 47.944
320117 99.963 -72.231 47.902
320118 99.995 -72.232 47.860
320119 100.019 -72.233 47.817
320120 100.051 -72.241 47.775
320121 100.090 -72.264 47.733
320122 100.111 -72.291 47.691
320123 100.135 -72.306 47.651
320124 100.160 -72.322 47.612
320125 100.185 -72.337 47.571
320126 100.210 -72.353 47.531
320127 100.234 -72.368 47.492
320128 100.259 -72.384 47.452
320129 100.284 -72.399 47.413
320130 100.316 -72.421 47.376
320131 100.353 -72.444 47.341
320132 100.387 -72.456 47.306
320133 100.418 -72.446 47.271
320134 100.448 -72.436 47.237
320135 100.478 -72.426 47.204
320136 100.509 -72.415 47.174
320137 100.539 -72.406 47.145
320138 100.570 -72.395 47.119
320139 100.600 -72.385 47.088
320140 100.632 -72.402 47.053
320141 100.665 -72.419 47.017
320142 100.697 -72.437 46.979
320143 100.729 -72.456 46.941
320144 100.761 -72.474 46.902
320145 100.793 -72.491 46.862
320146 100.826 -72.509 46.818
320147 100.858 -72.528 46.776
320148 100.890 -72.545 46.733
320149 100.922 -72.563 46.689
320150 100.955 -72.581 46.647
320151 100.987 -72.599 46.603
320152 101.019 -72.617 46.560
320153 101.053 -72.638 46.517
320154 101.090 -72.647 46.474
320155 101.129 -72.655 46.432
320156 101.168 -72.662 46.391
320157 101.206 -72.669 46.351
320158 101.246 -72.676 46.312
320159 101.284 -72.683 46.273
320160 101.323 -72.690 46.234
320161 101.362 -72.697 46.195
320162 101.401 -72.704 46.157
320163 101.440 -72.712 46.119
320164 101.478 -72.719 46.082
320165 101.517 -72.725 46.044
320166 101.551 -72.737 46.006
320167 101.584 -72.750 45.968
320168 101.615 -72.763 45.931
320169 101.632 -72.770 45.894
320170 101.650 -72.776 45.856
320171 101.668 -72.783 45.819
320172 101.686 -72.790 45.781
320173 101.703 -72.796 45.744
320174 101.721 -72.803 45.707
320175 101.739 -72.810 45.669
320176 101.754 -72.819 45.632
320177 101.768 -72.832 45.594
320178 101.802 -72.844 45.557
320179 101.842 -72.857 45.520
320180 101.882 -72.871 45.484
320181 101.922 -72.884 45.447
320182 101.962 -72.898 45.411
320183 102.003 -72.910 45.376
320184 102.043 -72.923 45.340
320185 102.086 -72.932 45.305
320186 102.129 -72.942 45.270
320670 94.855 -70.845 52.634
320671 94.881 -70.839 52.616
320672 94.907 -70.833 52.598
320673 94.926 -70.828 52.581
320674 94.945 -70.823 52.566
320675 94.964 -70.818 52.550
320676 94.990 -70.832 52.535
320677 95.017 -70.847 52.521
320678 95.043 -70.861 52.506
320679 95.069 -70.876 52.493
320680 95.096 -70.890 52.480
320681 95.099 -70.911 52.467
320682 95.085 -70.936 52.455
320683 95.122 -70.955 52.442
320684 95.159 -70.972 52.429
320685 95.196 -70.990 52.416
320686 95.233 -71.008 52.402
320687 95.270 -71.027 52.390
320688 95.307 -71.044 52.378
320689 95.344 -71.062 52.366
320690 95.381 -71.080 52.354
320691 95.418 -71.099 52.342
320692 95.455 -71.116 52.330
320693 95.491 -71.134 52.318
320694 95.529 -71.152 52.306
320695 95.565 -71.171 52.293
320696 95.602 -71.188 52.283
320697 95.628 -71.186 52.272
320698 95.642 -71.162 52.261
320699 95.657 -71.137 52.250
320700 95.671 -71.114 52.238
320701 95.693 -71.099 52.227
320702 95.728 -71.107 52.217
320703 95.762 -71.110 52.208
320704 95.797 -71.113 52.200
320705 95.831 -71.117 52.195
320706 95.870 -71.116 52.192
320707 95.911 -71.114 52.188
320708 95.966 -71.123 52.189
320709 96.020 -71.133 52.192
320710 96.075 -71.143 52.195
320711 96.129 -71.152 52.199
320712 96.184 -71.162 52.201
320713 96.239 -71.171 52.200
320714 96.294 -71.181 52.196
320715 96.348 -71.190 52.192
320716 96.405 -71.190 52.189
320717 96.479 -71.186 52.185
320718 96.538 -71.197 52.179
320719 96.574 -71.214 52.171
320720 96.609 -71.230 52.160
320721 96.644 -71.246 52.142
320722 96.679 -71.263 52.123
320723 96.717 -71.281 52.101
320724 96.754 -71.299 52.077
320725 96.791 -71.318 52.054
320726 96.828 -71.337 52.023
320727 96.865 -71.353 51.989
320728 96.897 -71.359 51.953
320729 96.930 -71.367 51.916
320730 96.963 -71.374 51.881
320731 96.996 -71.381 51.847
320732 97.028 -71.387 51.814
320733 97.061 -71.390 51.780
320734 97.094 -71.393 51.746
320735 97.131 -71.401 51.711
320736 97.171 -71.412 51.675
320737 97.211 -71.424 51.638
320738 97.250 -71.435 51.602
320739 97.290 -71.447 51.565
320740 97.330 -71.458 51.527
320741 97.365 -71.470 51.491
320742 97.401 -71.482 51.457
320743 97.436 -71.494 51.426
320744 97.472 -71.506 51.396
320745 97.507 -71.519 51.365
320746 97.542 -71.531 51.331
320747 97.578 -71.543 51.296
320748 97.613 -71.555 51.262
320749 97.649 -71.567 51.228
320750 97.684 -71.579 51.196
320751 97.720 -71.591 51.163
320752 97.756 -71.601 51.130
320753 97.792 -71.609 51.098
320754 97.829 -71.616 51.068
320755 97.865 -71.622 51.037
320756 97.898 -71.628 51.007
320757 97.924 -71.629 50.977
320758 97.949 -71.631 50.943
320759 97.974 -71.632 50.899
320760 97.999 -71.633 50.852
320761 98.024 -71.634 50.805
320762 98.051 -71.641 50.756
320763 98.078 -71.650 50.708
320764 98.104 -71.658 50.663
320765 98.132 -71.666 50.618
320766 98.160 -71.674 50.574
320767 98.188 -71.682 50.533
320768 98.216 -71.689 50.489
320769 98.244 -71.697 50.442
320770 98.272 -71.705 50.393
320771 98.299 -71.713 50.342
320772 98.327 -71.721 50.284
320773 98.359 -71.732 50.226
320774 98.400 -71.752 50.169
320775 98.441 -71.776 50.112
320776 98.483 -71.800 50.056
320777 98.524 -71.825 50.000
320778 98.566 -71.849 49.945
320779 98.600 -71.857 49.889
320780 98.632 -71.858 49.834
320781 98.664 -71.860 49.779
320782 98.696 -71.859 49.722
320783 98.727 -71.857 49.663
320784 98.759 -71.856 49.606
320785 98.791 -71.854 49.548
320786 98.822 -71.853 49.491
320787 98.854 -71.852 49.435
320788 98.886 -71.851 49.379
320789 98.917 -71.851 49.325
320790 98.943 -71.867 49.270
320791 98.970 -71.882 49.214
320792 98.995 -71.889 49.160
320793 99.027 -71.902 49.104
320794 99.059 -71.916 49.050
320795 99.092 -71.929 48.997
320796 99.124 -71.943 48.944
320797 99.156 -71.957 48.891
320798 99.189 -71.970 48.840
320799 99.221 -71.983 48.789
320800 99.253 -71.997 48.738
320801 99.285 -72.011 48.690
320802 99.317 -72.024 48.640
320803 99.350 -72.038 48.590
320804 99.380 -72.051 48.541
320805 99.417 -72.064 48.491
320806 99.458 -72.077 48.444
320807 99.499 -72.091 48.396
320808 99.540 -72.104 48.349
320809 99.581 -72.117 48.303
320810 99.622 -72.130 48.257
320811 99.663 -72.144 48.211
320812 99.704 -72.157 48.167
320813 99.746 -72.171 48.123
320814 99.787 -72.174 48.079
320815 99.830 -72.175 48.035
320816 99.872 -72.176 47.990
320817 99.915 -72.177 47.945
320818 99.957 -72.178 47.903
320819 99.999 -72.179 47.861
320820 100.030 -72.180 47.819
320821 100.069 -72.193 47.778
320822 100.106 -72.217 47.736
320823 100.129 -72.236 47.694
320824 100.154 -72.252 47.654
320825 100.179 -72.268 47.615
320826 100.203 -72.283 47.576
320827 100.228 -72.299 47.536
320828 100.253 -72.314 47.497
320829 100.278 -72.330 47.458
320830 100.302 -72.345 47.419
320831 100.329 -72.362 47.382
320832 100.366 -72.386 47.347
320833 100.402 -72.410 47.313
320834 100.435 -72.413 47.279
320835 100.466 -72.403 47.245
320836 100.496 -72.393 47.214
320837 100.527 -72.383 47.185
320838 100.557 -72.373 47.158
320839 100.587 -72.362 47.131
320840 100.618 -72.362 47.099
320841 100.650 -72.381 47.065
320842 100.683 -72.399 47.029
320843 100.715 -72.416 46.990
320844 100.747 -72.434 46.952
320845 100.779 -72.453 46.912
320846 100.812 -72.471 46.870
320847 100.844 -72.488 46.827
320848 100.876 -72.506 46.784
320849 100.908 -72.525 46.741
320850 100.940 -72.542 46.698
320851 100.973 -72.560 46.656
320852 101.005 -72.578 46.612
320853 101.036 -72.587 46.569
320854 101.067 -72.596 46.527
320855 101.103 -72.609 46.484
320856 101.142 -72.616 46.442
320857 101.181 -72.623 46.401
320858 101.220 -72.630 46.361
320859 101.259 -72.638 46.322
320860 101.298 -72.644 46.282
320861 101.336 -72.652 46.243
320862 101.375 -72.659 46.205
320863 101.414 -72.666 46.166
320864 101.453 -72.673 46.128
320865 101.492 -72.680 46.090
320866 101.531 -72.687 46.052
320867 101.564 -72.699 46.014
320868 101.597 -72.712 45.976
320869 101.628 -72.725 45.938
320870 101.654 -72.735 45.901
320871 101.672 -72.742 45.863
320872 101.689 -72.749 45.826
320873 101.707 -72.755 45.788
320874 101.725 -72.762 45.751
320875 101.742 -72.769 45.714
320876 101.760 -72.775 45.675
320877 101.775 -72.785 45.637
320878 101.789 -72.798 45.600
320879 101.815 -72.811 45.563
320880 101.855 -72.824 45.526
320881 101.895 -72.838 45.491
320882 101.935 -72.851 45.455
320883 101.975 -72.863 45.419
320884 102.015 -72.877 45.384
320885 102.056 -72.888 45.349
320886 102.099 -72.898 45.314
321370 94.825 -70.817 52.654
321371 94.850 -70.811 52.638
321372 94.876 -70.805 52.620
321373 94.902 -70.799 52.603
321374 94.928 -70.794 52.587
321375 94.947 -70.789 52.572
321376 94.966 -70.783 52.557
321377 94.988 -70.787 52.543
321378 95.015 -70.801 52.529
321379 95.041 -70.816 52.515
321380 95.068 -70.830 52.502
321381 95.094 -70.845 52.489
321382 95.076 -70.870 52.477
321383 95.100 -70.891 52.465
321384 95.137 -70.908 52.452
321385 95.174 -70.927 52.439
321386 95.211 -70.945 52.426
321387 95.248 -70.962 52.413
321388 95.285 -70.980 52.400
321389 95.322 -70.999 52.388
321390 95.359 -71.017 52.376
321391 95.396 -71.034 52.363
321392 95.432 -71.052 52.350
321393 95.470 -71.071 52.338
321394 95.507 -71.089 52.326
321395 95.544 -71.106 52.314
321396 95.581 -71.124 52.302
321397 95.617 -71.143 52.291
321398 95.655 -71.161 52.279
321399 95.671 -71.140 52.268
321400 95.685 -71.116 52.257
321401 95.699 -71.092 52.246
321402 95.713 -71.068 52.235
321403 95.733 -71.049 52.224
321404 95.767 -71.052 52.215
321405 95.802 -71.056 52.208
321406 95.837 -71.059 52.203
321407 95.875 -71.059 52.200
321408 95.920 -71.061 52.196
321409 95.975 -71.071 52.196
321410 96.030 -71.080 52.198
321411 96.085 -71.089 52.202
321412 96.139 -71.099 52.208
321413 96.194 -71.109 52.214
321414 96.248 -71.118 52.214
321415 96.303 -71.128 52.213
321416 96.358 -71.138 52.210
321417 96.412 -71.148 52.207
321418 96.493 -71.145 52.202
321419 96.545 -71.158 52.195
321420 96.580 -71.174 52.186
321421 96.617 -71.192 52.171
321422 96.654 -71.210 52.152
321423 96.691 -71.229 52.132
321424 96.729 -71.247 52.111
321425 96.766 -71.265 52.089
321426 96.803 -71.284 52.066
321427 96.840 -71.302 52.036
321428 96.873 -71.309 52.003
321429 96.906 -71.316 51.968
321430 96.939 -71.323 51.931
321431 96.971 -71.331 51.896
321432 97.004 -71.337 51.863
321433 97.036 -71.345 51.831
321434 97.069 -71.351 51.795
321435 97.102 -71.353 51.762
321436 97.141 -71.364 51.726
321437 97.181 -71.375 51.689
321438 97.221 -71.387 51.651
321439 97.257 -71.399 51.615
321440 97.293 -71.411 51.577
321441 97.328 -71.423 51.540
321442 97.363 -71.435 51.506
321443 97.399 -71.447 51.475
321444 97.434 -71.459 51.444
321445 97.470 -71.471 51.414
321446 97.505 -71.483 51.384
321447 97.541 -71.495 51.348
321448 97.576 -71.507 51.313
321449 97.611 -71.519 51.279
321450 97.647 -71.531 51.246
321451 97.682 -71.544 51.214
321452 97.718 -71.556 51.182
321453 97.754 -71.567 51.148
321454 97.790 -71.574 51.115
321455 97.827 -71.581 51.084
321456 97.863 -71.588 51.055
321457 97.900 -71.595 51.026
321458 97.935 -71.602 50.996
321459 97.960 -71.603 50.963
321460 97.986 -71.604 50.917
321461 98.011 -71.605 50.870
321462 98.035 -71.607 50.820
321463 98.062 -71.613 50.770
321464 98.089 -71.621 50.723
321465 98.116 -71.629 50.680
321466 98.144 -71.637 50.639
321467 98.172 -71.644 50.597
321468 98.200 -71.653 50.554
321469 98.228 -71.660 50.508
321470 98.255 -71.668 50.459
321471 98.283 -71.676 50.407
321472 98.311 -71.684 50.353
321473 98.339 -71.691 50.295
321474 98.367 -71.699 50.238
321475 98.408 -71.719 50.181
321476 98.449 -71.743 50.124
321477 98.491 -71.767 50.067
321478 98.532 -71.791 50.009
321479 98.574 -71.816 49.952
321480 98.612 -71.832 49.895
321481 98.644 -71.830 49.838
321482 98.675 -71.829 49.780
321483 98.707 -71.828 49.722
321484 98.738 -71.826 49.662
321485 98.770 -71.825 49.606
321486 98.802 -71.823 49.548
321487 98.834 -71.822 49.491
321488 98.865 -71.821 49.435
321489 98.897 -71.820 49.380
321490 98.929 -71.818 49.325
321491 98.955 -71.833 49.270
321492 98.982 -71.848 49.216
321493 99.009 -71.864 49.161
321494 99.035 -71.879 49.106
321495 99.064 -71.881 49.052
321496 99.096 -71.895 48.999
321497 99.128 -71.908 48.946
321498 99.160 -71.922 48.893
321499 99.193 -71.936 48.842
321500 99.225 -71.949 48.792
321501 99.257 -71.963 48.743
321502 99.289 -71.976 48.694
321503 99.322 -71.989 48.644
321504 99.354 -72.003 48.593
321505 99.382 -72.016 48.543
321506 99.410 -72.029 48.494
321507 99.450 -72.042 48.446
321508 99.491 -72.055 48.398
321509 99.532 -72.068 48.351
321510 99.573 -72.082 48.304
321511 99.615 -72.095 48.258
321512 99.655 -72.108 48.212
321513 99.697 -72.119 48.168
321514 99.739 -72.120 48.124
321515 99.782 -72.121 48.080
321516 99.824 -72.122 48.036
321517 99.866 -72.123 47.992
321518 99.909 -72.124 47.948
321519 99.951 -72.125 47.905
321520 99.994 -72.126 47.863
321521 100.041 -72.129 47.822
321522 100.086 -72.146 47.781
321523 100.123 -72.167 47.739
321524 100.147 -72.182 47.697
321525 100.173 -72.198 47.657
321526 100.197 -72.213 47.619
321527 100.222 -72.229 47.580
321528 100.247 -72.244 47.541
321529 100.271 -72.260 47.502
321530 100.296 -72.275 47.463
321531 100.321 -72.291 47.425
321532 100.346 -72.306 47.389
321533 100.379 -72.327 47.354
321534 100.415 -72.352 47.320
321535 100.452 -72.376 47.287
321536 100.483 -72.371 47.254
321537 100.513 -72.361 47.223
321538 100.544 -72.350 47.194
321539 100.574 -72.340 47.168
321540 100.605 -72.330 47.139
321541 100.636 -72.342 47.108
321542 100.668 -72.359 47.075
321543 100.701 -72.377 47.039
321544 100.733 -72.396 47.000
321545 100.765 -72.413 46.961
321546 100.797 -72.431 46.921
321547 100.830 -72.450 46.879
321548 100.862 -72.468 46.835
321549 100.894 -72.485 46.792
321550 100.926 -72.503 46.749
321551 100.958 -72.515 46.705
321552 100.989 -72.524 46.664
321553 101.019 -72.532 46.621
321554 101.050 -72.541 46.579
321555 101.083 -72.552 46.537
321556 101.118 -72.566 46.494
321557 101.156 -72.578 46.453
321558 101.194 -72.584 46.412
321559 101.233 -72.591 46.371
321560 101.272 -72.599 46.332
321561 101.311 -72.606 46.292
321562 101.349 -72.613 46.254
321563 101.388 -72.620 46.215
321564 101.427 -72.627 46.177
321565 101.466 -72.634 46.138
321566 101.505 -72.641 46.100
321567 101.544 -72.648 46.061
321568 101.577 -72.660 46.023
321569 101.609 -72.674 45.985
321570 101.641 -72.687 45.947
321571 101.674 -72.700 45.909
321572 101.693 -72.708 45.871
321573 101.711 -72.714 45.835
321574 101.729 -72.721 45.797
321575 101.746 -72.728 45.759
321576 101.764 -72.734 45.721
321577 101.781 -72.741 45.683
321578 101.796 -72.752 45.646
321579 101.810 -72.765 45.609
321580 101.828 -72.778 45.572
321581 101.868 -72.791 45.535
321582 101.908 -72.804 45.500
321583 101.948 -72.817 45.463
321584 101.988 -72.830 45.427
321585 102.028 -72.843 45.393
321586 102.069 -72.855 45.358
321587 102.112 -72.864 45.323
322071 94.820 -70.782 52.658
322072 94.845 -70.777 52.642
322073 94.871 -70.771 52.626
322074 94.897 -70.766 52.609
322075 94.923 -70.760 52.593
322076 94.949 -70.754 52.578
322077 94.968 -70.749 52.565
322078 94.987 -70.744 52.551
322079 95.013 -70.757 52.538
322080 95.039 -70.771 52.524
322081 95.066 -70.785 52.511
322082 95.082 -70.802 52.500
322083 95.078 -70.826 52.488
322084 95.115 -70.845 52.476
322085 95.152 -70.863 52.463
322086 95.189 -70.881 52.450
322087 95.226 -70.898 52.438
322088 95.263 -70.917 52.424
322089 95.300 -70.935 52.411
322090 95.337 -70.953 52.398
322091 95.374 -70.970 52.386
322092 95.411 -70.989 52.373
322093 95.448 -71.007 52.359
322094 95.485 -71.025 52.346
322095 95.522 -71.042 52.335
322096 95.558 -71.061 52.323
322097 95.596 -71.079 52.311
322098 95.632 -71.097 52.299
322099 95.670 -71.115 52.287
322100 95.698 -71.118 52.276
322101 95.712 -71.094 52.265
322102 95.727 -71.070 52.254
322103 95.747 -71.041 52.242
322104 95.773 -71.008 52.232
322105 95.776 -70.997 52.221
322106 95.808 -70.999 52.215
322107 95.842 -71.002 52.210
322108 95.880 -71.002 52.207
322109 95.930 -71.008 52.204
322110 95.985 -71.017 52.203
322111 96.039 -71.027 52.203
322112 96.094 -71.037 52.207
322113 96.149 -71.046 52.214
322114 96.203 -71.056 52.222
322115 96.258 -71.066 52.224
322116 96.313 -71.076 52.223
322117 96.367 -71.086 52.221
322118 96.421 -71.101 52.217
322119 96.489 -71.111 52.212
322120 96.555 -71.121 52.204
322121 96.593 -71.140 52.193
322122 96.630 -71.158 52.176
322123 96.667 -71.176 52.156
322124 96.704 -71.194 52.136
322125 96.741 -71.212 52.116
322126 96.779 -71.231 52.095
322127 96.816 -71.249 52.074
322128 96.849 -71.259 52.046
322129 96.882 -71.266 52.014
322130 96.914 -71.273 51.980
322131 96.947 -71.280 51.944
322132 96.980 -71.287 51.910
322133 97.012 -71.294 51.878
322134 97.045 -71.301 51.846
322135 97.078 -71.309 51.811
322136 97.111 -71.315 51.776
322137 97.149 -71.327 51.739
322138 97.184 -71.339 51.701
322139 97.220 -71.351 51.664
322140 97.255 -71.363 51.627
322141 97.291 -71.375 51.591
322142 97.326 -71.387 51.556
322143 97.362 -71.400 51.526
322144 97.397 -71.412 51.496
322145 97.432 -71.424 51.467
322146 97.468 -71.436 51.438
322147 97.503 -71.448 51.407
322148 97.539 -71.460 51.370
322149 97.574 -71.472 51.334
322150 97.609 -71.484 51.300
322151 97.645 -71.496 51.267
322152 97.680 -71.508 51.235
322153 97.716 -71.520 51.202
322154 97.751 -71.532 51.168
322155 97.788 -71.539 51.134
322156 97.824 -71.546 51.100
322157 97.861 -71.553 51.071
322158 97.898 -71.560 51.042
322159 97.934 -71.567 51.012
322160 97.971 -71.575 50.978
322161 97.997 -71.576 50.934
322162 98.022 -71.578 50.886
322163 98.048 -71.578 50.836
322164 98.073 -71.584 50.785
322165 98.100 -71.592 50.738
322166 98.128 -71.600 50.697
322167 98.156 -71.608 50.657
322168 98.183 -71.616 50.616
322169 98.212 -71.623 50.572
322170 98.239 -71.632 50.525
322171 98.267 -71.639 50.475
322172 98.295 -71.647 50.421
322173 98.323 -71.654 50.365
322174 98.351 -71.663 50.307
322175 98.379 -71.670 50.249
322176 98.416 -71.687 50.191
322177 98.457 -71.710 50.132
322178 98.499 -71.735 50.073
322179 98.540 -71.759 50.015
322180 98.582 -71.779 49.956
322181 98.621 -71.788 49.899
322182 98.654 -71.795 49.841
322183 98.687 -71.798 49.782
322184 98.718 -71.797 49.722
322185 98.750 -71.795 49.663
322186 98.782 -71.794 49.606
322187 98.813 -71.792 49.548
322188 98.845 -71.791 49.491
322189 98.877 -71.790 49.435
322190 98.908 -71.788 49.379
322191 98.940 -71.787 49.325
322192 98.968 -71.799 49.270
322193 98.994 -71.814 49.216
322194 99.021 -71.829 49.161
322195 99.048 -71.845 49.108
322196 99.074 -71.860 49.054
322197 99.100 -71.860 49.001
322198 99.132 -71.873 48.949
322199 99.164 -71.887 48.897
322200 99.197 -71.901 48.846
322201 99.229 -71.914 48.796
322202 99.261 -71.928 48.746
322203 99.293 -71.942 48.697
322204 99.326 -71.954 48.647
322205 99.356 -71.968 48.595
322206 99.383 -71.981 48.544
322207 99.411 -71.993 48.496
322208 99.442 -72.006 48.447
322209 99.484 -72.020 48.399
322210 99.524 -72.033 48.352
322211 99.566 -72.046 48.305
322212 99.607 -72.059 48.259
322213 99.649 -72.064 48.214
322214 99.691 -72.066 48.170
322215 99.733 -72.067 48.125
322216 99.776 -72.067 48.082
322217 99.818 -72.068 48.038
322218 99.864 -72.070 47.994
322219 99.914 -72.074 47.950
322220 99.964 -72.078 47.908
322221 100.014 -72.081 47.866
322222 100.060 -72.093 47.825
322223 100.105 -72.106 47.783
322224 100.139 -72.117 47.742
322225 100.166 -72.128 47.701
322226 100.191 -72.143 47.661
322227 100.215 -72.159 47.622
322228 100.240 -72.174 47.584
322229 100.265 -72.190 47.546
322230 100.290 -72.205 47.507
322231 100.315 -72.221 47.468
322232 100.339 -72.236 47.432
322233 100.364 -72.252 47.396
322234 100.391 -72.269 47.361
322235 100.428 -72.293 47.327
322236 100.465 -72.318 47.295
322237 100.500 -72.338 47.263
322238 100.531 -72.328 47.230
322239 100.561 -72.318 47.200
322240 100.592 -72.308 47.173
322241 100.622 -72.302 47.144
322242 100.654 -72.321 47.114
322243 100.687 -72.339 47.082
322244 100.719 -72.356 47.047
322245 100.751 -72.374 47.008
322246 100.783 -72.393 46.970
322247 100.816 -72.410 46.930
322248 100.848 -72.428 46.888
322249 100.879 -72.443 46.843
322250 100.911 -72.451 46.800
322251 100.941 -72.460 46.756
322252 100.972 -72.468 46.713
322253 101.003 -72.478 46.671
322254 101.034 -72.486 46.629
322255 101.065 -72.495 46.587
322256 101.099 -72.507 46.546
322257 101.134 -72.522 46.504
322258 101.170 -72.537 46.464
322259 101.207 -72.546 46.424
322260 101.246 -72.553 46.384
322261 101.285 -72.560 46.344
322262 101.324 -72.567 46.305
322263 101.363 -72.575 46.266
322264 101.402 -72.581 46.227
322265 101.441 -72.588 46.188
322266 101.479 -72.596 46.149
322267 101.518 -72.603 46.110
322268 101.557 -72.609 46.072
322269 101.590 -72.622 46.033
322270 101.622 -72.635 45.995
322271 101.654 -72.649 45.956
322272 101.686 -72.662 45.918
322273 101.714 -72.673 45.880
322274 101.732 -72.680 45.844
322275 101.750 -72.687 45.806
322276 101.768 -72.693 45.768
322277 101.785 -72.700 45.730
322278 101.803 -72.707 45.692
322279 101.817 -72.719 45.655
322280 101.831 -72.732 45.618
322281 101.845 -72.744 45.581
322282 101.881 -72.757 45.544
322283 101.920 -72.770 45.509
322284 101.961 -72.783 45.473
322285 102.000 -72.797 45.437
322286 102.041 -72.810 45.402
322287 102.083 -72.821 45.368
322288 102.125 -72.830 45.333
322771 94.789 -70.754 52.678
322772 94.815 -70.748 52.662
322773 94.840 -70.743 52.647
322774 94.866 -70.737 52.631
322775 94.892 -70.732 52.615
322776 94.918 -70.726 52.600
322777 94.944 -70.720 52.587
322778 94.970 -70.714 52.574
322779 94.989 -70.709 52.561
322780 95.011 -70.711 52.548
322781 95.038 -70.726 52.535
322782 95.064 -70.740 52.522
322783 95.060 -70.763 52.510
322784 95.094 -70.781 52.499
322785 95.130 -70.799 52.487
322786 95.167 -70.817 52.475
322787 95.204 -70.835 52.463
322788 95.241 -70.853 52.451
322789 95.278 -70.871 52.436
322790 95.315 -70.889 52.424
322791 95.352 -70.907 52.410
322792 95.389 -70.925 52.397
322793 95.426 -70.943 52.384
322794 95.463 -70.961 52.370
322795 95.500 -70.979 52.357
322796 95.537 -70.997 52.344
322797 95.573 -71.015 52.332
322798 95.611 -71.033 52.319
322799 95.648 -71.051 52.307
322800 95.684 -71.069 52.295
322801 95.722 -71.087 52.284
322802 95.741 -71.072 52.273
322803 95.758 -71.046 52.261
322804 95.785 -71.011 52.250
322805 95.813 -70.977 52.239
322806 95.823 -70.958 52.228
322807 95.827 -70.948 52.222
322808 95.848 -70.945 52.216
322809 95.885 -70.945 52.212
322810 95.940 -70.955 52.210
322811 95.994 -70.965 52.208
322812 96.049 -70.975 52.207
322813 96.104 -70.984 52.210
322814 96.158 -70.999 52.216
322815 96.213 -71.013 52.225
322816 96.267 -71.027 52.227
322817 96.321 -71.042 52.227
322818 96.376 -71.057 52.225
322819 96.430 -71.071 52.221
322820 96.485 -71.084 52.214
322821 96.563 -71.091 52.205
322822 96.608 -71.107 52.193
322823 96.646 -71.125 52.175
322824 96.683 -71.143 52.155
322825 96.720 -71.162 52.136
322826 96.757 -71.180 52.117
322827 96.795 -71.198 52.096
322828 96.828 -71.215 52.076
322829 96.858 -71.225 52.051
322830 96.891 -71.232 52.022
322831 96.924 -71.239 51.989
322832 96.956 -71.246 51.955
322833 96.989 -71.252 51.923
322834 97.022 -71.259 51.892
322835 97.054 -71.266 51.860
322836 97.088 -71.272 51.825
322837 97.125 -71.276 51.788
322838 97.154 -71.290 51.751
322839 97.182 -71.303 51.713
322840 97.218 -71.315 51.676
322841 97.253 -71.328 51.640
322842 97.289 -71.340 51.606
322843 97.324 -71.352 51.574
322844 97.360 -71.364 51.547
322845 97.395 -71.376 51.518
322846 97.430 -71.388 51.491
322847 97.466 -71.400 51.463
322848 97.501 -71.412 51.432
322849 97.537 -71.425 51.394
322850 97.572 -71.437 51.358
322851 97.608 -71.449 51.322
322852 97.643 -71.461 51.289
322853 97.678 -71.473 51.257
322854 97.714 -71.485 51.224
322855 97.749 -71.497 51.189
322856 97.786 -71.504 51.153
322857 97.822 -71.512 51.118
322858 97.859 -71.519 51.085
322859 97.895 -71.525 51.055
322860 97.932 -71.533 51.025
322861 97.968 -71.540 50.991
322862 98.005 -71.547 50.948
322863 98.034 -71.550 50.900
322864 98.059 -71.551 50.851
322865 98.085 -71.555 50.800
322866 98.111 -71.563 50.752
322867 98.140 -71.571 50.712
322868 98.167 -71.578 50.674
322869 98.196 -71.586 50.633
322870 98.223 -71.594 50.590
322871 98.251 -71.602 50.541
322872 98.279 -71.610 50.489
322873 98.307 -71.618 50.435
322874 98.335 -71.625 50.378
322875 98.363 -71.633 50.319
322876 98.390 -71.641 50.260
322877 98.424 -71.654 50.201
322878 98.465 -71.677 50.141
322879 98.507 -71.701 50.080
322880 98.548 -71.725 50.019
322881 98.591 -71.733 49.960
322882 98.630 -71.741 49.901
322883 98.663 -71.748 49.844
322884 98.696 -71.754 49.784
322885 98.729 -71.761 49.724
322886 98.761 -71.764 49.665
322887 98.793 -71.763 49.607
322888 98.825 -71.761 49.549
322889 98.856 -71.760 49.492
322890 98.888 -71.759 49.435
322891 98.920 -71.757 49.379
322892 98.951 -71.756 49.325
322893 98.980 -71.766 49.270
322894 99.006 -71.781 49.216
322895 99.033 -71.796 49.162
322896 99.060 -71.811 49.109
322897 99.086 -71.826 49.056
322898 99.113 -71.842 49.004
322899 99.136 -71.839 48.953
322900 99.169 -71.852 48.901
322901 99.201 -71.866 48.851
322902 99.233 -71.879 48.800
322903 99.266 -71.893 48.749
322904 99.298 -71.907 48.699
322905 99.330 -71.920 48.649
322906 99.357 -71.932 48.598
322907 99.385 -71.945 48.546
322908 99.413 -71.958 48.497
322909 99.440 -71.971 48.448
322910 99.476 -71.984 48.401
322911 99.517 -71.997 48.354
322912 99.558 -72.010 48.306
322913 99.600 -72.011 48.260
322914 99.643 -72.011 48.216
322915 99.687 -72.013 48.172
322916 99.737 -72.017 48.127
322917 99.787 -72.020 48.083
322918 99.837 -72.023 48.040
322919 99.887 -72.027 47.996
322920 99.937 -72.031 47.952
322921 99.987 -72.035 47.910
322922 100.034 -72.045 47.868
322923 100.079 -72.058 47.827
322924 100.117 -72.070 47.786
322925 100.151 -72.082 47.745
322926 100.179 -72.090 47.704
322927 100.206 -72.099 47.665
322928 100.233 -72.106 47.627
322929 100.259 -72.120 47.589
322930 100.283 -72.136 47.551
322931 100.308 -72.151 47.513
322932 100.333 -72.167 47.475
322933 100.358 -72.182 47.438
322934 100.382 -72.198 47.402
322935 100.407 -72.213 47.368
322936 100.440 -72.235 47.334
322937 100.477 -72.259 47.301
322938 100.514 -72.283 47.270
322939 100.548 -72.296 47.238
322940 100.579 -72.286 47.208
322941 100.609 -72.275 47.179
322942 100.640 -72.282 47.150
322943 100.673 -72.299 47.121
322944 100.705 -72.318 47.089
322945 100.737 -72.336 47.053
322946 100.769 -72.353 47.015
322947 100.802 -72.371 46.976
322948 100.833 -72.379 46.936
322949 100.863 -72.388 46.895
322950 100.894 -72.396 46.852
322951 100.925 -72.406 46.809
322952 100.956 -72.414 46.766
322953 100.987 -72.423 46.723
322954 101.018 -72.431 46.681
322955 101.049 -72.440 46.639
322956 101.079 -72.449 46.598
322957 101.114 -72.462 46.558
322958 101.150 -72.478 46.517
322959 101.185 -72.492 46.478
322960 101.221 -72.506 46.438
322961 101.259 -72.515 46.398
322962 101.299 -72.522 46.358
322963 101.337 -72.528 46.318
322964 101.376 -72.536 46.280
322965 101.415 -72.543 46.241
322966 101.454 -72.550 46.201
322967 101.492 -72.557 46.161
322968 101.531 -72.564 46.122
322969 101.570 -72.571 46.083
322970 101.603 -72.584 46.044
322971 101.635 -72.597 46.006
322972 101.667 -72.610 45.967
322973 101.699 -72.624 45.928
322974 101.732 -72.637 45.890
322975 101.754 -72.646 45.853
322976 101.771 -72.653 45.815
322977 101.789 -72.659 45.777
322978 101.807 -72.666 45.739
322979 101.824 -72.673 45.702
322980 101.838 -72.685 45.665
322981 101.852 -72.698 45.627
322982 101.866 -72.710 45.590
322983 101.893 -72.723 45.553
322984 101.934 -72.737 45.519
322985 101.973 -72.750 45.483
322986 102.013 -72.763 45.448
322987 102.053 -72.776 45.413
322988 102.096 -72.787 45.378
322989 102.138 -72.796 45.344
322990 102.164 -72.797 45.309
323371 91.895 -71.027 55.313
323372 91.926 -71.037 55.299
323373 91.956 -71.047 55.283
323374 91.987 -71.058 55.267
323375 92.017 -71.068 55.250
323376 92.048 -71.078 55.235
323377 92.079 -71.089 55.218
323378 92.109 -71.099 55.201
323379 92.140 -71.109 55.183
323380 92.171 -71.120 55.165
323381 92.202 -71.130 55.146
323382 92.237 -71.140 55.127
323383 92.278 -71.151 55.108
323384 92.305 -71.159 55.088
323385 92.333 -71.168 55.067
323386 92.360 -71.176 55.046
323387 92.388 -71.184 55.024
323388 92.415 -71.193 55.002
323389 92.443 -71.201 54.978
323390 92.471 -71.209 54.955
323391 92.498 -71.218 54.930
323392 92.531 -71.218 54.903
323393 92.579 -71.207 54.875
323472 94.784 -70.720 52.682
323473 94.810 -70.715 52.667
323474 94.836 -70.709 52.652
323475 94.861 -70.704 52.637
323476 94.887 -70.698 52.621
323477 94.913 -70.692 52.607
323478 94.939 -70.686 52.595
323479 94.965 -70.681 52.583
323480 94.990 -70.675 52.570
323481 95.011 -70.670 52.558
323482 95.036 -70.681 52.546
323483 95.062 -70.695 52.534
323484 95.071 -70.717 52.522
323485 95.108 -70.735 52.511
323486 95.145 -70.753 52.499
323487 95.182 -70.771 52.487
323488 95.219 -70.789 52.476
323489 95.256 -70.807 52.464
323490 95.293 -70.825 52.450
323491 95.330 -70.843 52.436
323492 95.367 -70.861 52.422
323493 95.404 -70.880 52.409
323494 95.441 -70.897 52.395
323495 95.478 -70.915 52.381
323496 95.515 -70.933 52.367
323497 95.552 -70.952 52.354
323498 95.589 -70.969 52.341
323499 95.625 -70.987 52.328
323500 95.663 -71.005 52.316
323501 95.699 -71.024 52.304
323502 95.737 -71.041 52.292
323503 95.768 -71.050 52.281
323504 95.796 -71.015 52.269
323505 95.823 -70.981 52.257
323506 95.850 -70.947 52.245
323507 95.870 -70.920 52.234
323508 95.874 -70.909 52.226
323509 95.877 -70.898 52.220
323510 95.894 -70.896 52.216
323511 95.949 -70.911 52.212
323512 96.003 -70.925 52.210
323513 96.058 -70.939 52.208
323514 96.112 -70.954 52.208
323515 96.167 -70.969 52.212
323516 96.221 -70.983 52.218
323517 96.275 -70.998 52.218
323518 96.330 -71.012 52.217
323519 96.384 -71.027 52.215
323520 96.439 -71.041 52.211
323521 96.493 -71.054 52.204
323522 96.548 -71.068 52.194
323523 96.624 -71.074 52.182
323524 96.662 -71.093 52.164
323525 96.699 -71.111 52.145
323526 96.736 -71.129 52.126
323527 96.773 -71.147 52.108
323528 96.810 -71.165 52.089
323529 96.839 -71.178 52.071
323530 96.868 -71.191 52.049
323531 96.900 -71.199 52.024
323532 96.933 -71.205 51.995
323533 96.965 -71.212 51.963
323534 96.998 -71.218 51.932
323535 97.031 -71.225 51.901
323536 97.066 -71.231 51.869
323537 97.104 -71.233 51.834
323538 97.142 -71.237 51.797
323539 97.171 -71.250 51.760
323540 97.199 -71.264 51.723
323541 97.228 -71.278 51.687
323542 97.256 -71.291 51.653
323543 97.287 -71.304 51.622
323544 97.322 -71.316 51.594
323545 97.358 -71.328 51.569
323546 97.393 -71.340 51.543
323547 97.429 -71.353 51.518
323548 97.464 -71.365 51.491
323549 97.499 -71.377 51.460
323550 97.535 -71.389 51.422
323551 97.570 -71.401 51.384
323552 97.606 -71.413 51.348
323553 97.641 -71.425 51.314
323554 97.677 -71.437 51.282
323555 97.712 -71.450 51.248
323556 97.747 -71.462 51.211
323557 97.784 -71.470 51.173
323558 97.820 -71.477 51.136
323559 97.857 -71.484 51.100
323560 97.893 -71.491 51.066
323561 97.930 -71.498 51.033
323562 97.966 -71.505 50.999
323563 98.003 -71.512 50.956
323564 98.039 -71.519 50.909
323565 98.071 -71.523 50.860
323566 98.096 -71.525 50.810
323567 98.124 -71.534 50.766
323568 98.151 -71.541 50.725
323569 98.179 -71.550 50.686
323570 98.207 -71.557 50.646
323571 98.235 -71.565 50.604
323572 98.263 -71.573 50.556
323573 98.291 -71.581 50.504
323574 98.319 -71.588 50.448
323575 98.347 -71.596 50.390
323576 98.374 -71.604 50.332
323577 98.403 -71.612 50.272
323578 98.432 -71.621 50.212
323579 98.473 -71.644 50.151
323580 98.515 -71.669 50.088
323581 98.557 -71.678 50.026
323582 98.600 -71.686 49.965
323583 98.638 -71.694 49.904
323584 98.671 -71.701 49.846
323585 98.704 -71.707 49.787
323586 98.737 -71.714 49.729
323587 98.771 -71.721 49.670
323588 98.803 -71.728 49.612
323589 98.836 -71.730 49.553
323590 98.867 -71.729 49.495
323591 98.899 -71.728 49.438
323592 98.931 -71.726 49.382
323593 98.962 -71.725 49.326
323594 98.992 -71.732 49.271
323595 99.018 -71.747 49.216
323596 99.045 -71.762 49.162
323597 99.072 -71.777 49.110
323598 99.098 -71.792 49.058
323599 99.125 -71.807 49.006
323600 99.152 -71.823 48.954
323601 99.178 -71.838 48.903
323602 99.205 -71.832 48.853
323603 99.237 -71.845 48.803
323604 99.270 -71.858 48.752
323605 99.302 -71.872 48.702
323606 99.331 -71.885 48.651
323607 99.359 -71.898 48.600
323608 99.386 -71.911 48.548
323609 99.414 -71.923 48.499
323610 99.442 -71.936 48.450
323611 99.470 -71.948 48.402
323612 99.510 -71.955 48.355
323613 99.560 -71.959 48.308
323614 99.610 -71.963 48.262
323615 99.660 -71.966 48.217
323616 99.710 -71.970 48.173
323617 99.760 -71.973 48.128
323618 99.810 -71.977 48.084
323619 99.860 -71.980 48.041
323620 99.910 -71.984 47.998
323621 99.960 -71.988 47.955
323622 100.007 -71.996 47.912
323623 100.053 -72.009 47.870
323624 100.090 -72.021 47.830
323625 100.127 -72.033 47.789
323626 100.163 -72.046 47.748
323627 100.190 -72.055 47.708
323628 100.217 -72.062 47.669
323629 100.245 -72.070 47.631
323630 100.272 -72.079 47.593
323631 100.299 -72.087 47.556
323632 100.327 -72.097 47.518
323633 100.351 -72.112 47.481
323634 100.376 -72.128 47.443
323635 100.401 -72.143 47.408
323636 100.426 -72.159 47.374
323637 100.454 -72.177 47.340
323638 100.490 -72.201 47.308
323639 100.527 -72.225 47.277
323640 100.563 -72.249 47.247
323641 100.596 -72.253 47.217
323642 100.627 -72.243 47.188
323643 100.659 -72.261 47.159
323644 100.691 -72.279 47.128
323645 100.723 -72.296 47.095
323646 100.754 -72.307 47.058
323647 100.785 -72.315 47.021
323648 100.816 -72.324 46.982
323649 100.847 -72.333 46.942
323650 100.878 -72.342 46.902
323651 100.909 -72.350 46.861
323652 100.939 -72.359 46.819
323653 100.971 -72.368 46.776
323654 101.001 -72.377 46.734
323655 101.032 -72.385 46.693
323656 101.063 -72.394 46.652
323657 101.095 -72.403 46.611
323658 101.130 -72.418 46.571
323659 101.165 -72.433 46.531
323660 101.201 -72.448 46.492
323661 101.237 -72.462 46.452
323662 101.273 -72.476 46.412
323663 101.312 -72.483 46.373
323664 101.350 -72.490 46.333
323665 101.389 -72.497 46.294
323666 101.428 -72.504 46.254
323667 101.467 -72.511 46.214
323668 101.505 -72.519 46.173
323669 101.545 -72.525 46.135
323670 101.584 -72.532 46.095
323671 101.616 -72.546 46.056
323672 101.648 -72.559 46.017
323673 101.680 -72.572 45.978
323674 101.712 -72.585 45.940
323675 101.744 -72.599 45.902
323676 101.775 -72.612 45.863
323677 101.793 -72.618 45.825
323678 101.811 -72.625 45.788
323679 101.828 -72.631 45.750
323680 101.845 -72.640 45.712
323681 101.859 -72.652 45.675
323682 101.873 -72.665 45.638
323683 101.887 -72.677 45.601
323684 101.906 -72.690 45.565
323685 101.946 -72.703 45.529
323686 101.986 -72.716 45.494
323687 102.026 -72.729 45.459
323688 102.066 -72.743 45.425
323689 102.109 -72.753 45.390
323690 102.147 -72.760 45.355
323691 102.172 -72.760 45.322
323692 102.197 -72.761 45.289
324069 91.799 -70.967 55.356
324070 91.829 -70.977 55.341
324071 91.860 -70.987 55.327
324072 91.891 -70.998 55.313
324073 91.921 -71.008 55.299
324074 91.952 -71.018 55.284
324075 91.983 -71.028 55.267
324076 92.013 -71.039 55.252
324077 92.044 -71.049 55.237
324078 92.075 -71.059 55.220
324079 92.105 -71.070 55.204
324080 92.136 -71.080 55.186
324081 92.167 -71.090 55.168
324082 92.197 -71.101 55.149
324083 92.234 -71.111 55.131
324084 92.270 -71.121 55.113
324085 92.298 -71.130 55.093
324086 92.325 -71.138 55.074
324087 92.353 -71.146 55.053
324088 92.380 -71.155 55.031
324089 92.408 -71.163 55.009
324090 92.435 -71.171 54.987
324091 92.463 -71.179 54.962
324092 92.490 -71.187 54.938
324093 92.529 -71.183 54.912
324094 92.577 -71.172 54.884
324095 92.624 -71.162 54.855
324096 92.671 -71.151 54.825
324097 92.719 -71.140 54.795
324172 94.753 -70.692 52.699
324173 94.779 -70.686 52.686
324174 94.805 -70.681 52.673
324175 94.831 -70.675 52.658
324176 94.856 -70.670 52.643
324177 94.882 -70.663 52.629
324178 94.908 -70.658 52.614
324179 94.934 -70.653 52.603
324180 94.960 -70.647 52.592
324181 94.985 -70.641 52.580
324182 95.011 -70.635 52.569
324183 95.034 -70.635 52.558
324184 95.050 -70.654 52.546
324185 95.087 -70.672 52.535
324186 95.123 -70.689 52.523
324187 95.161 -70.707 52.512
324188 95.197 -70.726 52.500
324189 95.234 -70.743 52.488
324190 95.271 -70.761 52.477
324191 95.308 -70.779 52.463
324192 95.345 -70.798 52.449
324193 95.382 -70.815 52.435
324194 95.419 -70.833 52.421
324195 95.456 -70.851 52.407
324196 95.494 -70.870 52.391
324197 95.531 -70.889 52.378
324198 95.569 -70.907 52.364
324199 95.606 -70.925 52.350
324200 95.644 -70.944 52.338
324201 95.682 -70.962 52.325
324202 95.719 -70.980 52.313
324203 95.757 -70.999 52.300
324204 95.791 -71.006 52.288
324205 95.824 -70.988 52.275
324206 95.861 -70.952 52.263
324207 95.889 -70.917 52.251
324208 95.916 -70.883 52.240
324209 95.920 -70.871 52.230
324210 95.916 -70.862 52.224
324211 95.917 -70.866 52.218
324212 95.957 -70.880 52.214
324213 96.012 -70.895 52.210
324214 96.066 -70.910 52.207
324215 96.120 -70.924 52.205
324216 96.175 -70.939 52.206
324217 96.229 -70.953 52.208
324218 96.284 -70.967 52.207
324219 96.338 -70.982 52.205
324220 96.392 -70.996 52.202
324221 96.447 -71.011 52.198
324222 96.502 -71.024 52.190
324223 96.557 -71.037 52.180
324224 96.612 -71.050 52.167
324225 96.677 -71.060 52.151
324226 96.715 -71.078 52.132
324227 96.752 -71.096 52.115
324228 96.789 -71.115 52.097
324229 96.821 -71.130 52.079
324230 96.850 -71.143 52.063
324231 96.878 -71.155 52.044
324232 96.909 -71.165 52.022
324233 96.942 -71.171 51.996
324234 96.975 -71.178 51.965
324235 97.007 -71.185 51.935
324236 97.043 -71.190 51.905
324237 97.083 -71.192 51.873
324238 97.121 -71.193 51.838
324239 97.159 -71.197 51.802
324240 97.188 -71.211 51.767
324241 97.217 -71.224 51.732
324242 97.245 -71.238 51.697
324243 97.274 -71.252 51.665
324244 97.302 -71.265 51.637
324245 97.331 -71.278 51.611
324246 97.360 -71.292 51.590
324247 97.391 -71.305 51.568
324248 97.427 -71.317 51.544
324249 97.462 -71.329 51.518
324250 97.498 -71.341 51.488
324251 97.533 -71.353 51.450
324252 97.568 -71.365 51.411
324253 97.604 -71.378 51.374
324254 97.639 -71.390 51.340
324255 97.675 -71.402 51.307
324256 97.710 -71.414 51.272
324257 97.746 -71.426 51.234
324258 97.782 -71.435 51.195
324259 97.818 -71.442 51.156
324260 97.855 -71.449 51.117
324261 97.891 -71.456 51.078
324262 97.928 -71.463 51.040
324263 97.964 -71.470 51.004
324264 98.001 -71.478 50.961
324265 98.037 -71.484 50.915
324266 98.074 -71.491 50.868
324267 98.107 -71.497 50.823
324268 98.135 -71.505 50.781
324269 98.163 -71.513 50.741
324270 98.191 -71.520 50.701
324271 98.219 -71.528 50.660
324272 98.247 -71.536 50.617
324273 98.275 -71.544 50.569
324274 98.303 -71.551 50.517
324275 98.331 -71.560 50.462
324276 98.358 -71.567 50.404
324277 98.386 -71.575 50.345
324278 98.414 -71.582 50.285
324279 98.442 -71.591 50.223
324280 98.482 -71.611 50.162
324281 98.524 -71.623 50.098
324282 98.566 -71.631 50.035
324283 98.608 -71.639 49.973
324284 98.647 -71.647 49.914
324285 98.680 -71.654 49.854
324286 98.713 -71.660 49.795
324287 98.746 -71.667 49.736
324288 98.779 -71.674 49.677
324289 98.812 -71.681 49.618
324290 98.845 -71.688 49.559
324291 98.878 -71.694 49.500
324292 98.911 -71.697 49.442
324293 98.942 -71.695 49.386
324294 98.974 -71.694 49.329
324295 99.004 -71.698 49.273
324296 99.030 -71.713 49.217
324297 99.057 -71.728 49.163
324298 99.084 -71.743 49.111
324299 99.111 -71.758 49.059
324300 99.137 -71.773 49.007
324301 99.164 -71.788 48.955
324302 99.190 -71.804 48.904
324303 99.217 -71.819 48.854
324304 99.241 -71.810 48.805
324305 99.274 -71.824 48.754
324306 99.305 -71.837 48.703
324307 99.333 -71.850 48.652
324308 99.360 -71.863 48.601
324309 99.388 -71.875 48.550
324310 99.420 -71.886 48.501
324311 99.453 -71.897 48.452
324312 99.487 -71.907 48.404
324313 99.533 -71.912 48.357
324314 99.583 -71.916 48.310
324315 99.633 -71.920 48.264
324316 99.683 -71.923 48.219
324317 99.733 -71.926 48.175
324318 99.783 -71.930 48.129
324319 99.833 -71.933 48.086
324320 99.883 -71.937 48.043
324321 99.932 -71.941 48.000
324322 99.981 -71.947 47.957
324323 100.026 -71.961 47.914
324324 100.065 -71.973 47.874
324325 100.101 -71.985 47.833
324326 100.137 -71.997 47.793
324327 100.173 -72.010 47.752
324328 100.202 -72.018 47.712
324329 100.229 -72.026 47.673
324330 100.257 -72.035 47.635
324331 100.284 -72.043 47.597
324332 100.312 -72.052 47.561
324333 100.339 -72.060 47.524
324334 100.366 -72.068 47.487
324335 100.394 -72.077 47.449
324336 100.419 -72.089 47.415
324337 100.444 -72.105 47.380
324338 100.469 -72.121 47.346
324339 100.503 -72.142 47.315
324340 100.540 -72.167 47.285
324341 100.576 -72.191 47.256
324342 100.613 -72.215 47.227
324343 100.645 -72.221 47.199
324344 100.676 -72.235 47.170
324345 100.707 -72.243 47.136
324346 100.738 -72.252 47.101
324347 100.769 -72.261 47.063
324348 100.800 -72.270 47.024
324349 100.831 -72.278 46.987
324350 100.861 -72.287 46.948
324351 100.893 -72.296 46.908
324352 100.923 -72.304 46.867
324353 100.954 -72.313 46.827
324354 100.985 -72.321 46.787
324355 101.016 -72.330 46.748
324356 101.047 -72.339 46.707
324357 101.078 -72.348 46.668
324358 101.111 -72.359 46.627
324359 101.146 -72.374 46.587
324360 101.182 -72.389 46.547
324361 101.217 -72.403 46.507
324362 101.252 -72.418 46.466
324363 101.288 -72.433 46.426
324364 101.325 -72.444 46.387
324365 101.363 -72.451 46.348
324366 101.402 -72.459 46.308
324367 101.441 -72.465 46.268
324368 101.480 -72.472 46.228
324369 101.519 -72.480 46.187
324370 101.558 -72.487 46.147
324371 101.597 -72.494 46.107
324372 101.628 -72.507 46.067
324373 101.661 -72.521 46.028
324374 101.693 -72.534 45.989
324375 101.725 -72.547 45.951
324376 101.757 -72.560 45.912
324377 101.789 -72.574 45.873
324378 101.814 -72.584 45.836
324379 101.832 -72.591 45.799
324380 101.850 -72.597 45.761
324381 101.866 -72.606 45.723
324382 101.880 -72.619 45.685
324383 101.894 -72.631 45.649
324384 101.907 -72.644 45.613
324385 101.921 -72.656 45.577
324386 101.959 -72.669 45.541
324387 101.999 -72.683 45.506
324388 102.039 -72.696 45.471
324389 102.079 -72.709 45.437
324390 102.122 -72.719 45.402
324391 102.155 -72.723 45.368
324392 102.180 -72.724 45.334
324393 102.204 -72.725 45.302
324394 102.229 -72.725 45.268
324395 102.254 -72.726 45.235
324770 91.795 -70.937 55.354
324771 91.825 -70.948 55.341
324772 91.856 -70.958 55.327
324773 91.886 -70.968 55.312
324774 91.917 -70.979 55.299
324775 91.948 -70.989 55.284
324776 91.979 -70.999 55.268
324777 92.009 -71.010 55.253
324778 92.040 -71.020 55.238
324779 92.071 -71.030 55.222
324780 92.101 -71.041 55.207
324781 92.132 -71.051 55.190
324782 92.162 -71.061 55.172
324783 92.193 -71.071 55.153
324784 92.230 -71.083 55.136
324785 92.263 -71.092 55.118
324786 92.290 -71.099 55.099
324787 92.317 -71.108 55.080
324788 92.345 -71.116 55.059
324789 92.373 -71.124 55.038
324790 92.400 -71.133 55.016
324791 92.428 -71.141 54.994
324792 92.455 -71.149 54.970
324793 92.482 -71.157 54.946
324794 92.527 -71.149 54.920
324795 92.574 -71.138 54.892
324796 92.622 -71.127 54.863
324797 92.669 -71.117 54.833
324798 92.717 -71.106 54.803
324799 92.764 -71.095 54.770
324800 92.811 -71.084 54.737
324873 94.749 -70.658 52.704
324874 94.774 -70.653 52.691
324875 94.800 -70.647 52.678
324876 94.826 -70.641 52.665
324877 94.852 -70.635 52.651
324878 94.878 -70.630 52.637
324879 94.903 -70.624 52.623
324880 94.929 -70.619 52.613
324881 94.955 -70.613 52.602
324882 94.981 -70.607 52.590
324883 95.006 -70.601 52.579
324884 95.032 -70.596 52.569
324885 95.065 -70.607 52.558
324886 95.102 -70.626 52.547
324887 95.140 -70.644 52.537
324888 95.177 -70.663 52.526
324889 95.215 -70.682 52.514
324890 95.253 -70.700 52.503
324891 95.290 -70.718 52.490
324892 95.328 -70.737 52.477
324893 95.366 -70.755 52.462
324894 95.403 -70.773 52.448
324895 95.441 -70.792 52.434
324896 95.478 -70.811 52.419
324897 95.516 -70.829 52.403
324898 95.553 -70.848 52.387
324899 95.591 -70.866 52.373
324900 95.628 -70.884 52.359
324901 95.666 -70.903 52.346
324902 95.703 -70.921 52.334
324903 95.741 -70.939 52.321
324904 95.779 -70.958 52.308
324905 95.809 -70.955 52.295
324906 95.838 -70.949 52.282
324907 95.873 -70.928 52.268
324908 95.913 -70.891 52.255
324909 95.954 -70.853 52.244
324910 95.960 -70.835 52.233
324911 95.943 -70.829 52.226
324912 95.948 -70.835 52.220
324913 95.973 -70.850 52.214
324914 96.020 -70.865 52.209
324915 96.075 -70.880 52.205
324916 96.129 -70.894 52.201
324917 96.183 -70.908 52.199
324918 96.238 -70.923 52.198
324919 96.292 -70.937 52.194
324920 96.346 -70.952 52.191
324921 96.401 -70.966 52.188
324922 96.455 -70.980 52.183
324923 96.510 -70.993 52.175
324924 96.565 -71.006 52.163
324925 96.620 -71.020 52.151
324926 96.675 -71.033 52.137
324927 96.730 -71.046 52.119
324928 96.767 -71.064 52.101
324929 96.802 -71.080 52.083
324930 96.831 -71.093 52.067
324931 96.860 -71.106 52.054
324932 96.889 -71.119 52.037
324933 96.919 -71.130 52.017
324934 96.951 -71.137 51.992
324935 96.984 -71.144 51.964
324936 97.021 -71.148 51.935
324937 97.061 -71.150 51.905
324938 97.099 -71.152 51.873
324939 97.139 -71.154 51.839
324940 97.176 -71.158 51.805
324941 97.205 -71.171 51.771
324942 97.233 -71.185 51.738
324943 97.262 -71.199 51.706
324944 97.291 -71.212 51.677
324945 97.319 -71.226 51.651
324946 97.348 -71.239 51.628
324947 97.377 -71.252 51.610
324948 97.405 -71.266 51.591
324949 97.434 -71.280 51.569
324950 97.462 -71.293 51.544
324951 97.496 -71.306 51.515
324952 97.531 -71.318 51.477
324953 97.567 -71.330 51.438
324954 97.602 -71.342 51.401
324955 97.637 -71.354 51.365
324956 97.673 -71.366 51.332
324957 97.708 -71.378 51.296
324958 97.744 -71.390 51.258
324959 97.780 -71.400 51.216
324960 97.816 -71.407 51.176
324961 97.853 -71.415 51.134
324962 97.889 -71.422 51.091
324963 97.926 -71.428 51.049
324964 97.962 -71.436 51.010
324965 97.999 -71.443 50.967
324966 98.035 -71.450 50.923
324967 98.073 -71.454 50.879
324968 98.108 -71.458 50.837
324969 98.137 -71.466 50.796
324970 98.166 -71.475 50.756
324971 98.195 -71.484 50.716
324972 98.223 -71.492 50.673
324973 98.252 -71.500 50.629
324974 98.281 -71.509 50.580
324975 98.309 -71.517 50.528
324976 98.338 -71.526 50.474
324977 98.367 -71.534 50.417
324978 98.395 -71.543 50.357
324979 98.424 -71.551 50.296
324980 98.452 -71.560 50.234
324981 98.490 -71.568 50.172
324982 98.532 -71.576 50.108
324983 98.575 -71.585 50.044
324984 98.618 -71.593 49.983
324985 98.656 -71.600 49.925
324986 98.689 -71.607 49.866
324987 98.722 -71.613 49.805
324988 98.755 -71.620 49.744
324989 98.788 -71.627 49.684
324990 98.821 -71.634 49.624
324991 98.854 -71.641 49.565
324992 98.887 -71.647 49.506
324993 98.920 -71.654 49.447
324994 98.953 -71.661 49.390
324995 98.985 -71.663 49.332
324996 99.016 -71.664 49.276
324997 99.043 -71.679 49.220
324998 99.069 -71.694 49.165
324999 99.096 -71.710 49.112
325000 99.123 -71.725 49.060
325001 99.149 -71.740 49.007
325002 99.176 -71.755 48.956
325003 99.203 -71.770 48.906
325004 99.229 -71.785 48.856
325005 99.256 -71.801 48.806
325006 99.278 -71.789 48.756
325007 99.307 -71.801 48.704
325008 99.341 -71.812 48.653
325009 99.375 -71.822 48.602
325010 99.409 -71.832 48.551
325011 99.443 -71.843 48.503
325012 99.477 -71.853 48.455
325013 99.511 -71.864 48.407
325014 99.556 -71.869 48.359
325015 99.606 -71.873 48.313
325016 99.656 -71.876 48.265
325017 99.706 -71.879 48.221
325018 99.756 -71.883 48.177
325019 99.805 -71.886 48.132
325020 99.855 -71.890 48.089
325021 99.905 -71.894 48.045
325022 99.955 -71.898 48.002
325023 100.000 -71.911 47.959
325024 100.039 -71.924 47.918
325025 100.075 -71.936 47.877
325026 100.111 -71.948 47.838
325027 100.147 -71.961 47.798
325028 100.183 -71.973 47.756
325029 100.214 -71.983 47.716
325030 100.241 -71.991 47.678
325031 100.268 -71.999 47.641
325032 100.296 -72.008 47.603
325033 100.323 -72.016 47.567
325034 100.350 -72.024 47.530
325035 100.378 -72.033 47.493
325036 100.405 -72.041 47.455
325037 100.432 -72.049 47.421
325038 100.460 -72.058 47.387
325039 100.487 -72.066 47.354
325040 100.515 -72.084 47.322
325041 100.552 -72.108 47.291
325042 100.587 -72.138 47.262
325043 100.623 -72.169 47.234
325044 100.657 -72.183 47.206
325045 100.691 -72.189 47.176
325046 100.721 -72.197 47.143
325047 100.753 -72.206 47.108
325048 100.783 -72.215 47.071
325049 100.814 -72.224 47.033
325050 100.845 -72.232 46.996
325051 100.876 -72.241 46.958
325052 100.907 -72.249 46.918
325053 100.938 -72.258 46.879
325054 100.969 -72.267 46.839
325055 100.999 -72.276 46.799
325056 101.031 -72.284 46.760
325057 101.061 -72.293 46.722
325058 101.092 -72.302 46.682
325059 101.126 -72.315 46.642
325060 101.162 -72.330 46.601
325061 101.197 -72.344 46.561
325062 101.233 -72.359 46.520
325063 101.268 -72.374 46.481
325064 101.304 -72.388 46.441
325065 101.339 -72.403 46.401
325066 101.376 -72.412 46.362
325067 101.415 -72.420 46.321
325068 101.455 -72.427 46.281
325069 101.493 -72.434 46.241
325070 101.532 -72.441 46.200
325071 101.571 -72.448 46.159
325072 101.609 -72.456 46.119
325073 101.641 -72.469 46.080
325074 101.674 -72.482 46.040
325075 101.706 -72.496 46.001
325076 101.738 -72.509 45.962
325077 101.770 -72.522 45.923
325078 101.802 -72.535 45.884
325079 101.834 -72.549 45.847
325080 101.853 -72.556 45.809
325081 101.871 -72.562 45.772
325082 101.887 -72.572 45.734
325083 101.900 -72.585 45.697
325084 101.914 -72.598 45.660
325085 101.928 -72.610 45.625
325086 101.942 -72.623 45.589
325087 101.972 -72.636 45.553
325088 102.012 -72.649 45.518
325089 102.052 -72.663 45.483
325090 102.092 -72.675 45.449
325091 102.135 -72.685 45.415
325092 102.163 -72.687 45.380
325093 102.187 -72.688 45.347
325094 102.212 -72.688 45.314
325095 102.237 -72.688 45.281
325096 102.262 -72.689 45.248
325097 102.287 -72.690 45.217
325098 102.311 -72.691 45.185
325471 91.790 -70.908 55.353
325472 91.821 -70.918 55.340
325473 91.852 -70.929 55.327
325474 91.882 -70.939 55.313
325475 91.913 -70.949 55.300
325476 91.944 -70.960 55.284
325477 91.975 -70.970 55.269
325478 92.005 -70.980 55.255
325479 92.035 -70.991 55.239
325480 92.066 -71.001 55.225
325481 92.097 -71.011 55.210
325482 92.128 -71.022 55.194
325483 92.158 -71.032 55.175
325484 92.189 -71.042 55.158
325485 92.227 -71.053 55.141
325486 92.255 -71.061 55.123
325487 92.283 -71.070 55.104
325488 92.310 -71.078 55.086
325489 92.338 -71.086 55.066
325490 92.365 -71.095 55.045
325491 92.392 -71.103 55.024
325492 92.420 -71.111 55.001
325493 92.447 -71.119 54.978
325494 92.478 -71.124 54.954
325495 92.525 -71.114 54.928
325496 92.572 -71.103 54.900
325497 92.620 -71.093 54.871
325498 92.667 -71.082 54.842
325499 92.714 -71.071 54.811
325500 92.762 -71.061 54.778
325501 92.809 -71.050 54.745
325502 92.857 -71.039 54.709
325503 92.904 -71.029 54.673
325574 94.744 -70.624 52.709
325575 94.769 -70.619 52.697
325576 94.795 -70.613 52.684
325577 94.821 -70.607 52.672
325578 94.847 -70.601 52.658
325579 94.875 -70.595 52.645
325580 94.905 -70.587 52.633
325581 94.935 -70.579 52.623
325582 94.965 -70.571 52.612
325583 94.995 -70.563 52.601
325584 95.025 -70.555 52.591
325585 95.055 -70.547 52.581
325586 95.088 -70.566 52.570
325587 95.124 -70.585 52.560
325588 95.162 -70.604 52.551
325589 95.199 -70.622 52.540
325590 95.237 -70.641 52.528
325591 95.274 -70.659 52.516
325592 95.312 -70.677 52.503
325593 95.350 -70.696 52.490
325594 95.387 -70.714 52.475
325595 95.425 -70.732 52.460
325596 95.463 -70.751 52.446
325597 95.500 -70.770 52.431
325598 95.538 -70.788 52.415
325599 95.576 -70.807 52.399
325600 95.613 -70.825 52.384
325601 95.651 -70.843 52.370
325602 95.688 -70.862 52.356
325603 95.726 -70.880 52.342
325604 95.763 -70.898 52.329
325605 95.798 -70.909 52.316
325606 95.827 -70.904 52.302
325607 95.856 -70.898 52.289
325608 95.886 -70.892 52.275
325609 95.921 -70.869 52.261
325610 95.962 -70.831 52.249
325611 95.987 -70.802 52.238
325612 95.970 -70.797 52.230
325613 95.979 -70.804 52.223
325614 96.004 -70.820 52.215
325615 96.029 -70.835 52.208
325616 96.083 -70.849 52.201
325617 96.137 -70.864 52.194
325618 96.191 -70.878 52.188
325619 96.246 -70.892 52.185
325620 96.301 -70.907 52.182
325621 96.355 -70.921 52.179
325622 96.409 -70.936 52.177
325623 96.464 -70.949 52.171
325624 96.519 -70.963 52.161
325625 96.574 -70.976 52.149
325626 96.628 -70.989 52.135
325627 96.683 -71.002 52.120
325628 96.739 -71.015 52.103
325629 96.784 -71.032 52.087
325630 96.813 -71.045 52.070
325631 96.842 -71.058 52.055
325632 96.871 -71.071 52.042
325633 96.899 -71.083 52.026
325634 96.928 -71.096 52.007
325635 96.960 -71.104 51.985
325636 96.999 -71.106 51.957
325637 97.038 -71.108 51.928
325638 97.077 -71.111 51.900
325639 97.116 -71.113 51.871
325640 97.156 -71.115 51.836
325641 97.193 -71.119 51.805
325642 97.222 -71.132 51.773
325643 97.250 -71.146 51.743
325644 97.279 -71.159 51.714
325645 97.308 -71.173 51.690
325646 97.336 -71.187 51.666
325647 97.365 -71.199 51.644
325648 97.394 -71.213 51.630
325649 97.422 -71.227 51.611
325650 97.451 -71.240 51.592
325651 97.479 -71.254 51.569
325652 97.508 -71.267 51.541
325653 97.537 -71.281 51.505
325654 97.566 -71.294 51.465
325655 97.600 -71.306 51.426
325656 97.636 -71.318 51.390
325657 97.671 -71.331 51.355
325658 97.706 -71.343 51.319
325659 97.742 -71.355 51.280
325660 97.778 -71.365 51.237
325661 97.814 -71.373 51.194
325662 97.851 -71.380 51.150
325663 97.887 -71.387 51.105
325664 97.924 -71.393 51.062
325665 97.960 -71.401 51.021
325666 97.998 -71.404 50.979
325667 98.035 -71.406 50.937
325668 98.073 -71.409 50.894
325669 98.106 -71.415 50.853
325670 98.135 -71.424 50.812
325671 98.163 -71.432 50.772
325672 98.192 -71.440 50.730
325673 98.221 -71.450 50.686
325674 98.249 -71.458 50.640
325675 98.278 -71.466 50.591
325676 98.306 -71.475 50.538
325677 98.335 -71.483 50.483
325678 98.364 -71.492 50.427
325679 98.392 -71.500 50.366
325680 98.423 -71.509 50.303
325681 98.463 -71.516 50.241
325682 98.503 -71.522 50.178
325683 98.543 -71.530 50.114
325684 98.584 -71.538 50.053
325685 98.627 -71.546 49.994
325686 98.664 -71.553 49.937
325687 98.698 -71.560 49.877
325688 98.730 -71.567 49.815
325689 98.764 -71.574 49.753
325690 98.797 -71.581 49.693
325691 98.830 -71.588 49.632
325692 98.863 -71.594 49.572
325693 98.896 -71.601 49.511
325694 98.929 -71.608 49.452
325695 98.962 -71.615 49.393
325696 98.995 -71.622 49.335
325697 99.027 -71.629 49.279
325698 99.052 -71.644 49.223
325699 99.078 -71.660 49.166
325700 99.104 -71.674 49.113
325701 99.130 -71.689 49.060
325702 99.156 -71.704 49.008
325703 99.182 -71.719 48.956
325704 99.208 -71.733 48.906
325705 99.233 -71.748 48.856
325706 99.259 -71.763 48.807
325707 99.296 -71.774 48.757
325708 99.331 -71.758 48.705
325709 99.364 -71.768 48.654
325710 99.399 -71.779 48.604
325711 99.432 -71.788 48.553
325712 99.466 -71.799 48.504
325713 99.500 -71.810 48.457
325714 99.534 -71.820 48.409
325715 99.579 -71.826 48.362
325716 99.629 -71.829 48.315
325717 99.679 -71.832 48.269
325718 99.728 -71.836 48.224
325719 99.778 -71.839 48.180
325720 99.828 -71.843 48.136
325721 99.878 -71.847 48.093
325722 99.928 -71.851 48.050
325723 99.974 -71.863 48.006
325724 100.013 -71.876 47.964
325725 100.049 -71.888 47.923
325726 100.085 -71.900 47.883
325727 100.121 -71.912 47.843
325728 100.157 -71.924 47.802
325729 100.193 -71.936 47.762
325730 100.225 -71.947 47.722
325731 100.253 -71.955 47.684
325732 100.280 -71.964 47.647
325733 100.308 -71.972 47.610
325734 100.335 -71.980 47.573
325735 100.362 -71.989 47.536
325736 100.390 -71.997 47.500
325737 100.417 -72.005 47.461
325738 100.444 -72.014 47.427
325739 100.472 -72.022 47.394
325740 100.499 -72.030 47.361
325741 100.525 -72.042 47.328
325742 100.560 -72.073 47.297
325743 100.596 -72.105 47.266
325744 100.631 -72.135 47.237
325745 100.665 -72.147 47.209
325746 100.699 -72.152 47.178
325747 100.733 -72.156 47.145
325748 100.766 -72.161 47.111
325749 100.798 -72.168 47.076
325750 100.829 -72.177 47.040
325751 100.860 -72.186 47.003
325752 100.891 -72.195 46.966
325753 100.921 -72.203 46.928
325754 100.953 -72.212 46.891
325755 100.983 -72.221 46.851
325756 101.014 -72.230 46.811
325757 101.045 -72.238 46.772
325758 101.076 -72.247 46.734
325759 101.107 -72.255 46.695
325760 101.142 -72.271 46.655
325761 101.178 -72.285 46.615
325762 101.213 -72.299 46.575
325763 101.249 -72.315 46.534
325764 101.284 -72.329 46.495
325765 101.320 -72.343 46.456
325766 101.355 -72.359 46.416
325767 101.390 -72.373 46.375
325768 101.429 -72.381 46.334
325769 101.468 -72.388 46.294
325770 101.506 -72.395 46.253
325771 101.545 -72.403 46.212
325772 101.584 -72.409 46.171
325773 101.622 -72.418 46.132
325774 101.655 -72.431 46.093
325775 101.686 -72.444 46.053
325776 101.718 -72.457 46.012
325777 101.751 -72.471 45.973
325778 101.783 -72.484 45.935
325779 101.815 -72.497 45.896
325780 101.847 -72.510 45.859
325781 101.875 -72.522 45.821
325782 101.893 -72.528 45.783
325783 101.907 -72.539 45.745
325784 101.921 -72.552 45.708
325785 101.936 -72.565 45.673
325786 101.949 -72.577 45.637
325787 101.963 -72.590 45.601
325788 101.984 -72.603 45.565
325789 102.025 -72.616 45.530
325790 102.064 -72.628 45.496
325791 102.105 -72.641 45.461
325792 102.146 -72.650 45.427
325793 102.171 -72.650 45.392
325794 102.195 -72.650 45.359
325795 102.220 -72.651 45.327
325796 102.245 -72.652 45.294
325797 102.269 -72.653 45.262
325798 102.295 -72.653 45.230
325799 102.319 -72.653 45.198
325800 102.344 -72.654 45.167
326172 91.786 -70.879 55.352
326173 91.817 -70.889 55.339
326174 91.848 -70.899 55.326
326175 91.878 -70.910 55.313
326176 91.909 -70.920 55.300
326177 91.939 -70.930 55.285
326178 91.970 -70.941 55.271
326179 92.001 -70.952 55.256
326180 92.031 -70.961 55.242
326181 92.062 -70.972 55.228
326182 92.093 -70.982 55.213
326183 92.124 -70.992 55.198
326184 92.154 -71.003 55.183
326185 92.185 -71.013 55.165
326186 92.220 -71.024 55.147
326187 92.247 -71.032 55.128
326188 92.275 -71.039 55.111
326189 92.302 -71.048 55.092
326190 92.330 -71.056 55.073
326191 92.357 -71.064 55.052
326192 92.385 -71.073 55.031
326193 92.413 -71.081 55.009
326194 92.440 -71.089 54.986
326195 92.475 -71.090 54.962
326196 92.522 -71.080 54.936
326197 92.570 -71.069 54.908
326198 92.618 -71.058 54.879
326199 92.665 -71.048 54.849
326200 92.712 -71.036 54.818
326201 92.760 -71.026 54.786
326202 92.807 -71.015 54.752
326203 92.854 -71.005 54.717
326204 92.902 -70.994 54.679
326205 92.934 -70.987 54.641
326206 92.960 -70.981 54.603
326207 92.986 -70.976 54.565
326275 94.750 -70.584 52.715
326276 94.780 -70.576 52.704
326277 94.811 -70.569 52.691
326278 94.840 -70.560 52.679
326279 94.871 -70.553 52.666
326280 94.901 -70.544 52.654
326281 94.930 -70.537 52.642
326282 94.961 -70.529 52.633
326283 94.991 -70.521 52.623
326284 95.021 -70.513 52.613
326285 95.051 -70.505 52.604
326286 95.076 -70.502 52.594
326287 95.112 -70.522 52.583
326288 95.147 -70.543 52.573
326289 95.184 -70.563 52.564
326290 95.222 -70.581 52.553
326291 95.259 -70.600 52.542
326292 95.297 -70.618 52.530
326293 95.334 -70.636 52.516
326294 95.372 -70.655 52.502
326295 95.409 -70.673 52.488
326296 95.447 -70.691 52.473
326297 95.484 -70.710 52.458
326298 95.522 -70.729 52.443
326299 95.560 -70.747 52.427
326300 95.597 -70.766 52.410
326301 95.635 -70.784 52.395
326302 95.672 -70.802 52.380
326303 95.710 -70.821 52.365
326304 95.748 -70.839 52.350
326305 95.785 -70.858 52.336
326306 95.816 -70.858 52.323
326307 95.846 -70.852 52.309
326308 95.875 -70.847 52.295
326309 95.904 -70.841 52.280
326310 95.934 -70.836 52.266
326311 95.970 -70.809 52.253
326312 96.011 -70.771 52.242
326313 95.997 -70.764 52.233
326314 96.010 -70.773 52.226
326315 96.035 -70.789 52.217
326316 96.060 -70.804 52.208
326317 96.091 -70.819 52.199
326318 96.146 -70.833 52.190
326319 96.200 -70.848 52.181
326320 96.254 -70.862 52.176
326321 96.309 -70.876 52.171
326322 96.363 -70.891 52.166
326323 96.417 -70.905 52.162
326324 96.472 -70.919 52.156
326325 96.527 -70.933 52.146
326326 96.582 -70.945 52.134
326327 96.637 -70.958 52.119
326328 96.692 -70.972 52.104
326329 96.747 -70.985 52.088
326330 96.792 -70.997 52.072
326331 96.820 -71.010 52.056
326332 96.850 -71.023 52.042
326333 96.878 -71.036 52.031
326334 96.907 -71.049 52.015
326335 96.936 -71.061 51.997
326336 96.975 -71.068 51.975
326337 97.014 -71.071 51.949
326338 97.052 -71.074 51.920
326339 97.091 -71.077 51.892
326340 97.131 -71.079 51.865
326341 97.169 -71.082 51.833
326342 97.202 -71.084 51.803
326343 97.229 -71.096 51.775
326344 97.259 -71.108 51.748
326345 97.289 -71.121 51.724
326346 97.319 -71.135 51.703
326347 97.349 -71.148 51.681
326348 97.379 -71.161 51.661
326349 97.409 -71.174 51.647
326350 97.439 -71.187 51.629
326351 97.468 -71.201 51.611
326352 97.497 -71.215 51.589
326353 97.525 -71.227 51.562
326354 97.554 -71.241 51.527
326355 97.583 -71.255 51.488
326356 97.611 -71.268 51.450
326357 97.640 -71.282 51.413
326358 97.669 -71.295 51.378
326359 97.705 -71.307 51.341
326360 97.740 -71.319 51.301
326361 97.776 -71.331 51.258
326362 97.812 -71.338 51.213
326363 97.849 -71.345 51.167
326364 97.885 -71.351 51.121
326365 97.923 -71.353 51.077
326366 97.960 -71.356 51.037
326367 97.998 -71.359 50.997
326368 98.035 -71.362 50.954
326369 98.073 -71.365 50.912
326370 98.103 -71.372 50.871
326371 98.132 -71.381 50.829
326372 98.160 -71.390 50.787
326373 98.189 -71.398 50.743
326374 98.218 -71.406 50.698
326375 98.246 -71.415 50.650
326376 98.275 -71.424 50.600
326377 98.304 -71.432 50.546
326378 98.333 -71.440 50.489
326379 98.361 -71.449 50.431
326380 98.397 -71.456 50.370
326381 98.437 -71.464 50.307
326382 98.477 -71.471 50.244
326383 98.516 -71.478 50.184
326384 98.556 -71.485 50.123
326385 98.596 -71.492 50.064
326386 98.636 -71.499 50.006
326387 98.673 -71.506 49.948
326388 98.706 -71.513 49.887
326389 98.739 -71.521 49.825
326390 98.771 -71.528 49.763
326391 98.804 -71.535 49.701
326392 98.837 -71.543 49.639
326393 98.869 -71.550 49.577
326394 98.902 -71.557 49.516
326395 98.934 -71.565 49.456
326396 98.967 -71.572 49.397
326397 98.999 -71.579 49.339
326398 99.032 -71.587 49.281
326399 99.052 -71.606 49.224
326400 99.078 -71.620 49.167
326401 99.103 -71.635 49.114
326402 99.129 -71.650 49.061
326403 99.155 -71.665 49.009
326404 99.181 -71.680 48.958
326405 99.207 -71.694 48.907
326406 99.233 -71.710 48.857
326407 99.265 -71.722 48.807
326408 99.310 -71.729 48.756
326409 99.355 -71.737 48.705
326410 99.388 -71.725 48.655
326411 99.422 -71.735 48.605
326412 99.455 -71.745 48.555
326413 99.489 -71.755 48.506
326414 99.523 -71.766 48.460
326415 99.557 -71.776 48.412
326416 99.602 -71.782 48.365
326417 99.651 -71.785 48.319
326418 99.701 -71.789 48.273
326419 99.751 -71.793 48.229
326420 99.801 -71.796 48.185
326421 99.851 -71.800 48.141
326422 99.901 -71.804 48.098
326423 99.947 -71.814 48.055
326424 99.987 -71.826 48.012
326425 100.023 -71.839 47.970
326426 100.059 -71.851 47.930
326427 100.095 -71.864 47.889
326428 100.131 -71.876 47.849
326429 100.168 -71.888 47.808
326430 100.203 -71.900 47.768
326431 100.237 -71.911 47.728
326432 100.265 -71.920 47.691
326433 100.292 -71.928 47.654
326434 100.319 -71.936 47.617
326435 100.347 -71.945 47.580
326436 100.374 -71.953 47.543
326437 100.401 -71.961 47.506
326438 100.429 -71.970 47.469
326439 100.456 -71.977 47.435
326440 100.483 -71.986 47.401
326441 100.510 -71.995 47.367
326442 100.534 -72.010 47.335
326443 100.568 -72.039 47.302
326444 100.603 -72.070 47.269
326445 100.638 -72.102 47.238
326446 100.673 -72.111 47.208
326447 100.706 -72.115 47.177
326448 100.740 -72.120 47.144
326449 100.774 -72.124 47.111
326450 100.808 -72.129 47.078
326451 100.842 -72.133 47.044
326452 100.875 -72.140 47.009
326453 100.905 -72.149 46.974
326454 100.936 -72.158 46.938
326455 100.967 -72.166 46.902
326456 100.998 -72.175 46.863
326457 101.029 -72.183 46.823
326458 101.059 -72.193 46.783
326459 101.091 -72.201 46.745
326460 101.122 -72.211 46.707
326461 101.158 -72.226 46.668
326462 101.193 -72.240 46.628
326463 101.229 -72.255 46.588
326464 101.264 -72.270 46.548
326465 101.300 -72.284 46.510
326466 101.335 -72.299 46.470
326467 101.371 -72.314 46.429
326468 101.406 -72.328 46.389
326469 101.442 -72.343 46.347
326470 101.481 -72.349 46.306
326471 101.519 -72.357 46.264
326472 101.558 -72.364 46.225
326473 101.597 -72.371 46.184
326474 101.635 -72.379 46.145
326475 101.667 -72.393 46.107
326476 101.699 -72.406 46.067
326477 101.732 -72.419 46.027
326478 101.763 -72.432 45.987
326479 101.796 -72.446 45.948
326480 101.828 -72.459 45.909
326481 101.860 -72.472 45.871
326482 101.892 -72.485 45.833
326483 101.914 -72.494 45.795
326484 101.928 -72.506 45.758
326485 101.942 -72.519 45.720
326486 101.956 -72.531 45.685
326487 101.970 -72.544 45.649
326488 101.984 -72.556 45.613
326489 101.998 -72.569 45.578
326490 102.037 -72.582 45.543
326491 102.077 -72.595 45.508
326492 102.118 -72.607 45.474
326493 102.153 -72.613 45.439
326494 102.178 -72.613 45.405
326495 102.203 -72.614 45.372
326496 102.228 -72.615 45.339
326497 102.253 -72.616 45.307
326498 102.277 -72.616 45.275
326499 102.302 -72.616 45.243
326500 102.327 -72.617 45.211
326501 102.352 -72.618 45.180
326873 91.782 -70.850 55.350
326874 91.813 -70.860 55.338
326875 91.844 -70.870 55.325
326876 91.874 -70.880 55.312
326877 91.905 -70.891 55.300
326878 91.935 -70.901 55.286
326879 91.966 -70.911 55.272
326880 91.997 -70.922 55.258
326881 92.027 -70.933 55.245
326882 92.058 -70.942 55.230
326883 92.089 -70.953 55.216
326884 92.119 -70.964 55.203
326885 92.150 -70.974 55.189
326886 92.180 -70.984 55.172
326887 92.212 -70.993 55.154
326888 92.240 -71.002 55.135
326889 92.267 -71.010 55.117
326890 92.295 -71.018 55.099
326891 92.322 -71.027 55.080
326892 92.350 -71.035 55.059
326893 92.377 -71.043 55.038
326894 92.405 -71.051 55.017
326895 92.433 -71.059 54.994
326896 92.473 -71.055 54.970
326897 92.520 -71.045 54.945
326898 92.568 -71.034 54.917
326899 92.615 -71.024 54.888
326900 92.662 -71.013 54.858
326901 92.710 -71.002 54.826
326902 92.758 -70.992 54.793
326903 92.805 -70.981 54.759
326904 92.852 -70.970 54.723
326905 92.900 -70.960 54.685
326906 92.929 -70.953 54.647
326907 92.955 -70.948 54.608
326908 92.981 -70.942 54.569
326909 93.006 -70.936 54.531
326910 93.032 -70.930 54.494
326975 94.746 -70.542 52.735
326976 94.776 -70.534 52.722
326977 94.806 -70.526 52.711
326978 94.836 -70.518 52.700
326979 94.866 -70.510 52.687
326980 94.897 -70.502 52.675
326981 94.926 -70.494 52.664
326982 94.957 -70.486 52.653
326983 94.987 -70.478 52.644
326984 95.017 -70.470 52.634
326985 95.047 -70.463 52.625
326986 95.077 -70.454 52.617
326987 95.100 -70.459 52.608
326988 95.136 -70.479 52.597
326989 95.172 -70.500 52.587
326990 95.207 -70.520 52.578
326991 95.244 -70.540 52.567
326992 95.281 -70.559 52.556
326993 95.319 -70.577 52.544
326994 95.356 -70.595 52.530
326995 95.394 -70.614 52.516
326996 95.431 -70.632 52.502
326997 95.469 -70.651 52.487
326998 95.507 -70.670 52.471
326999 95.544 -70.688 52.454
327000 95.582 -70.706 52.438
327001 95.619 -70.725 52.421
327002 95.657 -70.743 52.405
327003 95.694 -70.761 52.390
327004 95.732 -70.780 52.374
327005 95.769 -70.798 52.359
327006 95.806 -70.813 52.344
327007 95.835 -70.807 52.330
327008 95.864 -70.801 52.315
327009 95.893 -70.795 52.300
327010 95.922 -70.790 52.285
327011 95.952 -70.785 52.271
327012 95.981 -70.779 52.258
327013 96.019 -70.749 52.247
327014 96.024 -70.732 52.238
327015 96.041 -70.743 52.229
327016 96.066 -70.758 52.220
327017 96.091 -70.773 52.209
327018 96.116 -70.789 52.198
327019 96.154 -70.803 52.187
327020 96.208 -70.817 52.177
327021 96.263 -70.832 52.169
327022 96.317 -70.846 52.160
327023 96.372 -70.861 52.153
327024 96.426 -70.876 52.146
327025 96.481 -70.889 52.140
327026 96.536 -70.902 52.129
327027 96.591 -70.915 52.117
327028 96.646 -70.928 52.104
327029 96.700 -70.942 52.088
327030 96.750 -70.953 52.072
327031 96.794 -70.964 52.057
327032 96.823 -70.977 52.042
327033 96.852 -70.990 52.030
327034 96.881 -71.003 52.019
327035 96.909 -71.016 52.004
327036 96.943 -71.028 51.985
327037 96.987 -71.036 51.965
327038 97.026 -71.039 51.938
327039 97.065 -71.042 51.911
327040 97.104 -71.045 51.885
327041 97.143 -71.048 51.858
327042 97.181 -71.050 51.829
327043 97.210 -71.052 51.801
327044 97.234 -71.059 51.776
327045 97.263 -71.072 51.754
327046 97.294 -71.085 51.735
327047 97.323 -71.099 51.716
327048 97.354 -71.111 51.696
327049 97.383 -71.125 51.676
327050 97.414 -71.138 51.661
327051 97.443 -71.151 51.645
327052 97.473 -71.165 51.628
327053 97.503 -71.177 51.606
327054 97.533 -71.190 51.580
327055 97.563 -71.204 51.546
327056 97.593 -71.217 51.509
327057 97.623 -71.230 51.471
327058 97.653 -71.243 51.435
327059 97.683 -71.256 51.400
327060 97.713 -71.270 51.361
327061 97.743 -71.283 51.320
327062 97.774 -71.296 51.276
327063 97.810 -71.301 51.230
327064 97.848 -71.303 51.183
327065 97.885 -71.306 51.137
327066 97.923 -71.309 51.094
327067 97.961 -71.312 51.056
327068 97.998 -71.315 51.015
327069 98.036 -71.318 50.973
327070 98.072 -71.321 50.931
327071 98.100 -71.330 50.890
327072 98.129 -71.338 50.847
327073 98.158 -71.346 50.803
327074 98.186 -71.356 50.757
327075 98.215 -71.364 50.710
327076 98.244 -71.372 50.660
327077 98.273 -71.381 50.608
327078 98.301 -71.389 50.552
327079 98.331 -71.397 50.494
327080 98.371 -71.405 50.434
327081 98.411 -71.412 50.372
327082 98.450 -71.419 50.309
327083 98.491 -71.426 50.248
327084 98.530 -71.433 50.192
327085 98.570 -71.440 50.133
327086 98.608 -71.447 50.075
327087 98.644 -71.454 50.016
327088 98.681 -71.459 49.958
327089 98.711 -71.471 49.897
327090 98.743 -71.478 49.835
327091 98.776 -71.486 49.772
327092 98.808 -71.493 49.709
327093 98.841 -71.500 49.646
327094 98.873 -71.508 49.583
327095 98.906 -71.515 49.521
327096 98.939 -71.522 49.460
327097 98.971 -71.530 49.401
327098 99.004 -71.537 49.342
327099 99.037 -71.544 49.283
327100 99.059 -71.560 49.225
327101 99.077 -71.582 49.168
327102 99.103 -71.597 49.115
327103 99.129 -71.611 49.063
327104 99.155 -71.626 49.010
327105 99.181 -71.641 48.959
327106 99.207 -71.656 48.909
327107 99.234 -71.670 48.859
327108 99.279 -71.678 48.807
327109 99.324 -71.685 48.756
327110 99.369 -71.693 48.707
327111 99.415 -71.701 48.656
327112 99.445 -71.691 48.606
327113 99.479 -71.701 48.557
327114 99.513 -71.712 48.510
327115 99.547 -71.722 48.463
327116 99.580 -71.732 48.416
327117 99.624 -71.738 48.369
327118 99.674 -71.742 48.324
327119 99.724 -71.746 48.278
327120 99.774 -71.749 48.234
327121 99.824 -71.753 48.190
327122 99.873 -71.757 48.147
327123 99.921 -71.765 48.105
327124 99.961 -71.778 48.062
327125 99.997 -71.790 48.019
327126 100.033 -71.802 47.977
327127 100.069 -71.815 47.937
327128 100.106 -71.827 47.896
327129 100.141 -71.839 47.856
327130 100.177 -71.851 47.816
327131 100.214 -71.864 47.777
327132 100.249 -71.876 47.738
327133 100.276 -71.884 47.700
327134 100.303 -71.892 47.664
327135 100.331 -71.901 47.627
327136 100.358 -71.909 47.588
327137 100.385 -71.917 47.551
327138 100.413 -71.926 47.515
327139 100.440 -71.933 47.480
327140 100.468 -71.942 47.444
327141 100.495 -71.950 47.409
327142 100.520 -71.963 47.375
327143 100.543 -71.978 47.340
327144 100.576 -72.006 47.306
327145 100.611 -72.036 47.272
327146 100.646 -72.067 47.238
327147 100.680 -72.075 47.206
327148 100.714 -72.080 47.174
327149 100.748 -72.084 47.142
327150 100.782 -72.089 47.109
327151 100.816 -72.093 47.077
327152 100.849 -72.098 47.044
327153 100.883 -72.102 47.010
327154 100.917 -72.106 46.976
327155 100.951 -72.111 46.942
327156 100.981 -72.120 46.907
327157 101.012 -72.129 46.870
327158 101.043 -72.137 46.831
327159 101.074 -72.146 46.792
327160 101.105 -72.155 46.754
327161 101.138 -72.167 46.717
327162 101.174 -72.181 46.679
327163 101.209 -72.196 46.641
327164 101.245 -72.211 46.601
327165 101.280 -72.225 46.561
327166 101.316 -72.240 46.523
327167 101.351 -72.255 46.483
327168 101.387 -72.269 46.441
327169 101.422 -72.284 46.401
327170 101.458 -72.299 46.360
327171 101.494 -72.311 46.319
327172 101.533 -72.318 46.278
327173 101.572 -72.325 46.238
327174 101.611 -72.332 46.197
327175 101.648 -72.340 46.159
327176 101.680 -72.354 46.120
327177 101.712 -72.368 46.081
327178 101.744 -72.381 46.042
327179 101.776 -72.394 46.001
327180 101.809 -72.407 45.962
327181 101.840 -72.421 45.922
327182 101.873 -72.434 45.884
327183 101.905 -72.447 45.846
327184 101.936 -72.460 45.808
327185 101.949 -72.472 45.771
327186 101.963 -72.485 45.733
327187 101.977 -72.498 45.697
327188 101.991 -72.510 45.661
327189 102.005 -72.523 45.626
327190 102.019 -72.536 45.590
327191 102.050 -72.549 45.556
327192 102.090 -72.562 45.521
327193 102.131 -72.573 45.486
327194 102.162 -72.576 45.452
327195 102.186 -72.577 45.418
327196 102.211 -72.578 45.385
327197 102.236 -72.578 45.352
327198 102.260 -72.578 45.320
327199 102.285 -72.579 45.287
327200 102.310 -72.580 45.255
327201 102.335 -72.581 45.224
327202 102.360 -72.581 45.193
327574 91.772 -70.818 55.348
327575 91.803 -70.829 55.336
327576 91.835 -70.839 55.324
327577 91.865 -70.850 55.312
327578 91.896 -70.861 55.301
327579 91.928 -70.870 55.287
327580 91.959 -70.881 55.272
327581 91.990 -70.892 55.258
327582 92.021 -70.902 55.245
327583 92.053 -70.913 55.232
327584 92.084 -70.923 55.219
327585 92.115 -70.934 55.207
327586 92.146 -70.945 55.194
327587 92.176 -70.955 55.179
327588 92.205 -70.964 55.162
327589 92.233 -70.971 55.143
327590 92.262 -70.979 55.124
327591 92.290 -70.986 55.106
327592 92.318 -70.994 55.087
327593 92.347 -71.002 55.067
327594 92.375 -71.009 55.047
327595 92.403 -71.017 55.026
327596 92.437 -71.021 55.003
327597 92.478 -71.021 54.979
327598 92.518 -71.011 54.954
327599 92.566 -71.000 54.926
327600 92.613 -70.989 54.897
327601 92.660 -70.978 54.867
327602 92.708 -70.967 54.835
327603 92.755 -70.957 54.801
327604 92.802 -70.946 54.767
327605 92.850 -70.936 54.731
327606 92.897 -70.925 54.693
327607 92.924 -70.919 54.654
327608 92.950 -70.914 54.614
327609 92.976 -70.908 54.575
327610 93.002 -70.902 54.536
327611 93.027 -70.897 54.498
327612 93.053 -70.891 54.460
327613 93.079 -70.886 54.425
327676 94.772 -70.491 52.741
327677 94.802 -70.484 52.729
327678 94.832 -70.475 52.719
327679 94.862 -70.468 52.708
327680 94.892 -70.460 52.696
327681 94.922 -70.452 52.685
327682 94.953 -70.444 52.674
327683 94.982 -70.436 52.664
327684 95.013 -70.428 52.655
327685 95.043 -70.420 52.646
327686 95.072 -70.412 52.638
327687 95.103 -70.404 52.631
327688 95.124 -70.416 52.622
327689 95.160 -70.436 52.612
327690 95.196 -70.456 52.603
327691 95.232 -70.477 52.594
327692 95.268 -70.497 52.583
327693 95.304 -70.517 52.572
327694 95.340 -70.536 52.559
327695 95.378 -70.554 52.545
327696 95.416 -70.573 52.530
327697 95.453 -70.591 52.516
327698 95.491 -70.610 52.500
327699 95.529 -70.629 52.483
327700 95.566 -70.647 52.466
327701 95.604 -70.665 52.450
327702 95.641 -70.684 52.433
327703 95.679 -70.702 52.416
327704 95.716 -70.720 52.400
327705 95.754 -70.739 52.383
327706 95.792 -70.757 52.368
327707 95.824 -70.761 52.353
327708 95.853 -70.756 52.337
327709 95.883 -70.750 52.322
327710 95.912 -70.745 52.307
327711 95.941 -70.739 52.292
327712 95.970 -70.733 52.277
327713 96.000 -70.728 52.263
327714 96.029 -70.722 52.251
327715 96.050 -70.699 52.241
327716 96.072 -70.712 52.232
327717 96.097 -70.727 52.223
327718 96.122 -70.742 52.213
327719 96.147 -70.757 52.203
327720 96.172 -70.773 52.191
327721 96.217 -70.787 52.180
327722 96.271 -70.802 52.169
327723 96.326 -70.817 52.155
327724 96.380 -70.831 52.142
327725 96.435 -70.845 52.131
327726 96.489 -70.858 52.122
327727 96.544 -70.871 52.111
327728 96.599 -70.885 52.100
327729 96.654 -70.898 52.087
327730 96.703 -70.909 52.072
327731 96.751 -70.920 52.057
327732 96.796 -70.932 52.043
327733 96.825 -70.945 52.029
327734 96.854 -70.958 52.018
327735 96.883 -70.970 52.006
327736 96.912 -70.983 51.991
327737 96.953 -70.993 51.972
327738 96.998 -71.002 51.952
327739 97.039 -71.008 51.927
327740 97.078 -71.011 51.903
327741 97.116 -71.014 51.880
327742 97.155 -71.016 51.856
327743 97.194 -71.019 51.828
327744 97.219 -71.021 51.803
327745 97.238 -71.023 51.781
327746 97.268 -71.036 51.763
327747 97.298 -71.049 51.747
327748 97.328 -71.062 51.731
327749 97.358 -71.075 51.711
327750 97.388 -71.089 51.690
327751 97.418 -71.102 51.673
327752 97.448 -71.115 51.657
327753 97.477 -71.128 51.639
327754 97.508 -71.141 51.618
327755 97.537 -71.154 51.594
327756 97.568 -71.168 51.560
327757 97.597 -71.180 51.525
327758 97.628 -71.194 51.489
327759 97.657 -71.207 51.453
327760 97.687 -71.220 51.417
327761 97.717 -71.233 51.379
327762 97.746 -71.246 51.337
327763 97.778 -71.257 51.291
327764 97.811 -71.256 51.243
327765 97.848 -71.259 51.196
327766 97.886 -71.262 51.152
327767 97.924 -71.265 51.110
327768 97.961 -71.267 51.074
327769 97.999 -71.270 51.033
327770 98.036 -71.273 50.993
327771 98.069 -71.279 50.951
327772 98.098 -71.287 50.908
327773 98.127 -71.296 50.865
327774 98.155 -71.304 50.820
327775 98.183 -71.312 50.772
327776 98.212 -71.321 50.723
327777 98.241 -71.330 50.670
327778 98.270 -71.338 50.614
327779 98.305 -71.346 50.557
327780 98.345 -71.353 50.497
327781 98.385 -71.360 50.436
327782 98.425 -71.367 50.374
327783 98.464 -71.375 50.314
327784 98.504 -71.381 50.258
327785 98.541 -71.388 50.204
327786 98.577 -71.395 50.145
327787 98.614 -71.402 50.085
327788 98.651 -71.407 50.025
327789 98.690 -71.411 49.966
327790 98.720 -71.422 49.904
327791 98.750 -71.433 49.842
327792 98.781 -71.444 49.778
327793 98.813 -71.451 49.713
327794 98.846 -71.458 49.651
327795 98.878 -71.466 49.588
327796 98.911 -71.473 49.526
327797 98.943 -71.480 49.464
327798 98.976 -71.487 49.404
327799 99.008 -71.495 49.343
327800 99.041 -71.502 49.283
327801 99.067 -71.516 49.225
327802 99.084 -71.538 49.171
327803 99.102 -71.557 49.117
327804 99.129 -71.572 49.064
327805 99.154 -71.587 49.012
327806 99.180 -71.602 48.962
327807 99.206 -71.616 48.910
327808 99.248 -71.625 48.859
327809 99.293 -71.633 48.809
327810 99.338 -71.641 48.758
327811 99.383 -71.649 48.708
327812 99.428 -71.657 48.659
327813 99.468 -71.647 48.609
327814 99.502 -71.658 48.560
327815 99.536 -71.668 48.514
327816 99.570 -71.679 48.467
327817 99.604 -71.689 48.420
327818 99.647 -71.695 48.374
327819 99.697 -71.699 48.329
327820 99.747 -71.702 48.284
327821 99.796 -71.706 48.240
327822 99.846 -71.710 48.198
327823 99.895 -71.716 48.155
327824 99.935 -71.729 48.113
327825 99.971 -71.741 48.070
327826 100.007 -71.754 48.027
327827 100.044 -71.766 47.985
327828 100.079 -71.779 47.945
327829 100.116 -71.791 47.905
327830 100.151 -71.803 47.865
327831 100.188 -71.815 47.826
327832 100.224 -71.827 47.787
327833 100.260 -71.839 47.748
327834 100.288 -71.848 47.710
327835 100.316 -71.857 47.673
327836 100.343 -71.865 47.636
327837 100.370 -71.873 47.598
327838 100.398 -71.881 47.561
327839 100.425 -71.889 47.525
327840 100.452 -71.898 47.489
327841 100.480 -71.906 47.453
327842 100.506 -71.917 47.417
327843 100.529 -71.931 47.381
327844 100.552 -71.946 47.345
327845 100.583 -71.972 47.310
327846 100.619 -72.003 47.275
327847 100.654 -72.033 47.241
327848 100.688 -72.039 47.207
327849 100.721 -72.043 47.175
327850 100.756 -72.048 47.142
327851 100.789 -72.052 47.110
327852 100.823 -72.057 47.078
327853 100.857 -72.061 47.045
327854 100.891 -72.066 47.013
327855 100.925 -72.070 46.980
327856 100.959 -72.075 46.946
327857 100.992 -72.080 46.911
327858 101.026 -72.084 46.875
327859 101.058 -72.092 46.837
327860 101.089 -72.100 46.800
327861 101.120 -72.109 46.762
327862 101.154 -72.122 46.726
327863 101.190 -72.137 46.689
327864 101.225 -72.152 46.652
327865 101.260 -72.166 46.613
327866 101.296 -72.181 46.574
327867 101.331 -72.196 46.535
327868 101.367 -72.211 46.494
327869 101.402 -72.225 46.453
327870 101.438 -72.240 46.413
327871 101.473 -72.255 46.372
327872 101.509 -72.269 46.333
327873 101.546 -72.280 46.293
327874 101.585 -72.287 46.251
327875 101.624 -72.293 46.212
327876 101.661 -72.302 46.172
327877 101.693 -72.315 46.133
327878 101.725 -72.329 46.094
327879 101.757 -72.343 46.056
327880 101.789 -72.356 46.016
327881 101.821 -72.369 45.977
327882 101.853 -72.382 45.938
327883 101.886 -72.396 45.899
327884 101.910 -72.410 45.860
327885 101.934 -72.424 45.822
327886 101.962 -72.439 45.784
327887 101.984 -72.452 45.747
327888 101.998 -72.465 45.710
327889 102.012 -72.477 45.674
327890 102.026 -72.490 45.639
327891 102.040 -72.502 45.604
327892 102.062 -72.515 45.569
327893 102.103 -72.528 45.534
327894 102.144 -72.539 45.499
327895 102.169 -72.540 45.465
327896 102.194 -72.540 45.432
327897 102.219 -72.540 45.398
327898 102.244 -72.541 45.365
327899 102.268 -72.542 45.332
327900 102.293 -72.543 45.300
327901 102.318 -72.544 45.269
327902 102.342 -72.544 45.237
327903 102.368 -72.544 45.206
328275 91.760 -70.785 55.345
328276 91.791 -70.796 55.334
328277 91.822 -70.807 55.323
328278 91.854 -70.817 55.311
328279 91.884 -70.828 55.300
328280 91.916 -70.839 55.286
328281 91.947 -70.849 55.273
328282 91.978 -70.860 55.259
328283 92.009 -70.870 55.246
328284 92.040 -70.880 55.233
328285 92.072 -70.891 55.221
328286 92.103 -70.901 55.209
328287 92.134 -70.912 55.197
328288 92.166 -70.923 55.184
328289 92.203 -70.930 55.168
328290 92.231 -70.937 55.151
328291 92.260 -70.945 55.133
328292 92.288 -70.952 55.116
328293 92.316 -70.960 55.097
328294 92.345 -70.967 55.078
328295 92.373 -70.976 55.057
328296 92.402 -70.983 55.036
328297 92.443 -70.982 55.013
328298 92.484 -70.981 54.989
328299 92.523 -70.976 54.965
328300 92.563 -70.965 54.936
328301 92.611 -70.955 54.906
328302 92.658 -70.944 54.876
328303 92.705 -70.933 54.844
328304 92.753 -70.923 54.810
328305 92.800 -70.912 54.775
328306 92.848 -70.901 54.739
328307 92.894 -70.891 54.701
328308 92.919 -70.886 54.661
328309 92.945 -70.880 54.621
328310 92.971 -70.874 54.580
328311 92.997 -70.868 54.540
328312 93.022 -70.863 54.503
328313 93.048 -70.857 54.464
328314 93.074 -70.851 54.428
328315 93.100 -70.845 54.394
328377 94.798 -70.441 52.748
328378 94.828 -70.433 52.736
328379 94.858 -70.425 52.727
328380 94.888 -70.417 52.716
328381 94.918 -70.409 52.706
328382 94.948 -70.401 52.695
328383 94.978 -70.394 52.684
328384 95.008 -70.385 52.675
328385 95.038 -70.378 52.666
328386 95.068 -70.369 52.659
328387 95.099 -70.362 52.652
328388 95.128 -70.353 52.645
328389 95.148 -70.372 52.636
328390 95.184 -70.393 52.627
328391 95.220 -70.413 52.619
328392 95.256 -70.434 52.611
328393 95.292 -70.453 52.600
328394 95.328 -70.474 52.588
328395 95.364 -70.494 52.574
328396 95.400 -70.513 52.560
328397 95.438 -70.532 52.545
328398 95.475 -70.550 52.530
328399 95.513 -70.569 52.513
328400 95.550 -70.588 52.496
328401 95.588 -70.606 52.479
328402 95.626 -70.624 52.461
328403 95.663 -70.643 52.444
328404 95.701 -70.661 52.427
328405 95.738 -70.679 52.410
328406 95.776 -70.698 52.393
328407 95.813 -70.716 52.377
328408 95.842 -70.710 52.361
328409 95.872 -70.704 52.345
328410 95.901 -70.699 52.330
328411 95.931 -70.693 52.314
328412 95.960 -70.688 52.298
328413 95.989 -70.682 52.283
328414 96.018 -70.676 52.268
328415 96.047 -70.671 52.256
328416 96.076 -70.666 52.245
328417 96.102 -70.682 52.236
328418 96.127 -70.697 52.226
328419 96.153 -70.712 52.216
328420 96.178 -70.727 52.205
328421 96.203 -70.742 52.194
328422 96.228 -70.757 52.182
328423 96.279 -70.772 52.169
328424 96.334 -70.786 52.152
328425 96.388 -70.801 52.135
328426 96.443 -70.815 52.121
328427 96.498 -70.828 52.109
328428 96.553 -70.841 52.097
328429 96.608 -70.854 52.086
328430 96.657 -70.866 52.073
328431 96.704 -70.877 52.059
328432 96.752 -70.888 52.045
328433 96.799 -70.899 52.031
328434 96.827 -70.912 52.017
328435 96.856 -70.925 52.006
328436 96.885 -70.938 51.994
328437 96.918 -70.950 51.978
328438 96.963 -70.959 51.961
328439 97.007 -70.968 51.942
328440 97.051 -70.977 51.921
328441 97.090 -70.980 51.899
328442 97.129 -70.982 51.879
328443 97.168 -70.985 51.858
328444 97.207 -70.988 51.834
328445 97.228 -70.989 51.812
328446 97.246 -70.990 51.792
328447 97.273 -70.999 51.775
328448 97.302 -71.013 51.760
328449 97.332 -71.026 51.744
328450 97.362 -71.039 51.724
328451 97.392 -71.052 51.704
328452 97.422 -71.065 51.687
328453 97.452 -71.078 51.670
328454 97.482 -71.092 51.652
328455 97.512 -71.105 51.632
328456 97.542 -71.118 51.608
328457 97.572 -71.131 51.574
328458 97.602 -71.144 51.539
328459 97.632 -71.157 51.504
328460 97.661 -71.170 51.468
328461 97.690 -71.182 51.432
328462 97.720 -71.195 51.393
328463 97.749 -71.207 51.350
328464 97.782 -71.218 51.303
328465 97.813 -71.214 51.255
328466 97.849 -71.215 51.208
328467 97.886 -71.217 51.165
328468 97.924 -71.220 51.126
328469 97.961 -71.223 51.091
328470 97.999 -71.225 51.051
328471 98.036 -71.228 51.011
328472 98.066 -71.236 50.969
328473 98.095 -71.244 50.925
328474 98.124 -71.253 50.881
328475 98.152 -71.262 50.835
328476 98.181 -71.270 50.786
328477 98.210 -71.278 50.734
328478 98.239 -71.287 50.679
328479 98.279 -71.294 50.621
328480 98.319 -71.301 50.562
328481 98.359 -71.308 50.501
328482 98.398 -71.315 50.440
328483 98.438 -71.322 50.379
328484 98.474 -71.329 50.322
328485 98.510 -71.336 50.267
328486 98.547 -71.343 50.213
328487 98.583 -71.350 50.154
328488 98.619 -71.356 50.093
328489 98.658 -71.360 50.033
328490 98.698 -71.363 49.971
328491 98.729 -71.373 49.909
328492 98.760 -71.384 49.845
328493 98.789 -71.395 49.781
328494 98.819 -71.406 49.717
328495 98.850 -71.416 49.654
328496 98.883 -71.423 49.591
328497 98.915 -71.431 49.529
328498 98.948 -71.438 49.467
328499 98.981 -71.445 49.406
328500 99.013 -71.453 49.345
328501 99.046 -71.460 49.285
328502 99.074 -71.471 49.227
328503 99.091 -71.492 49.172
328504 99.109 -71.514 49.119
328505 99.128 -71.533 49.066
328506 99.154 -71.548 49.014
328507 99.180 -71.563 48.964
328508 99.216 -71.574 48.912
328509 99.262 -71.582 48.861
328510 99.306 -71.589 48.811
328511 99.352 -71.597 48.760
328512 99.397 -71.604 48.711
328513 99.442 -71.612 48.661
328514 99.487 -71.619 48.612
328515 99.525 -71.614 48.564
328516 99.559 -71.625 48.517
328517 99.593 -71.635 48.471
328518 99.627 -71.645 48.425
328519 99.669 -71.652 48.379
328520 99.719 -71.655 48.334
328521 99.769 -71.659 48.290
328522 99.819 -71.663 48.248
328523 99.868 -71.668 48.206
328524 99.909 -71.681 48.164
328525 99.945 -71.693 48.122
328526 99.982 -71.705 48.080
328527 100.017 -71.717 48.037
328528 100.054 -71.729 47.994
328529 100.089 -71.741 47.954
328530 100.126 -71.754 47.914
328531 100.162 -71.766 47.875
328532 100.198 -71.779 47.837
328533 100.234 -71.791 47.798
328534 100.270 -71.803 47.758
328535 100.300 -71.813 47.719
328536 100.327 -71.821 47.681
328537 100.354 -71.829 47.646
328538 100.382 -71.837 47.608
328539 100.409 -71.845 47.571
328540 100.436 -71.854 47.535
328541 100.464 -71.862 47.498
328542 100.491 -71.870 47.461
328543 100.515 -71.885 47.424
328544 100.539 -71.899 47.387
328545 100.562 -71.914 47.350
328546 100.592 -71.938 47.314
328547 100.627 -71.969 47.279
328548 100.662 -71.998 47.245
328549 100.695 -72.002 47.211
328550 100.729 -72.007 47.178
328551 100.763 -72.011 47.145
328552 100.797 -72.016 47.112
328553 100.831 -72.020 47.080
328554 100.864 -72.025 47.047
328555 100.899 -72.030 47.018
328556 100.932 -72.034 46.985
328557 100.966 -72.039 46.950
328558 101.000 -72.043 46.915
328559 101.034 -72.048 46.879
328560 101.067 -72.052 46.844
328561 101.102 -72.057 46.807
328562 101.134 -72.063 46.770
328563 101.170 -72.078 46.733
328564 101.205 -72.092 46.697
328565 101.241 -72.107 46.661
328566 101.276 -72.122 46.623
328567 101.312 -72.136 46.585
328568 101.347 -72.152 46.545
328569 101.383 -72.166 46.505
328570 101.418 -72.180 46.466
328571 101.454 -72.196 46.427
328572 101.489 -72.210 46.387
328573 101.525 -72.224 46.346
328574 101.560 -72.240 46.305
328575 101.598 -72.248 46.265
328576 101.637 -72.255 46.226
328577 101.674 -72.264 46.185
328578 101.706 -72.277 46.145
328579 101.738 -72.290 46.105
328580 101.770 -72.304 46.067
328581 101.802 -72.318 46.028
328582 101.834 -72.330 45.989
328583 101.859 -72.345 45.951
328584 101.883 -72.360 45.913
328585 101.906 -72.374 45.874
328586 101.929 -72.390 45.836
328587 101.954 -72.404 45.798
328588 101.985 -72.418 45.761
328589 102.017 -72.431 45.724
328590 102.033 -72.443 45.688
328591 102.047 -72.456 45.652
328592 102.060 -72.468 45.617
328593 102.076 -72.481 45.582
328594 102.115 -72.494 45.547
328595 102.152 -72.503 45.512
328596 102.177 -72.503 45.478
328597 102.202 -72.503 45.445
328598 102.227 -72.504 45.411
328599 102.251 -72.505 45.377
328600 102.276 -72.506 45.345
328601 102.301 -72.506 45.313
328602 102.326 -72.506 45.282
328603 102.350 -72.507 45.250
328604 102.375 -72.508 45.219
328976 91.748 -70.754 55.342
328977 91.779 -70.764 55.332
328978 91.810 -70.775 55.321
328979 91.842 -70.785 55.309
328980 91.873 -70.795 55.298
328981 91.904 -70.806 55.286
328982 91.935 -70.817 55.273
328983 91.966 -70.827 55.260
328984 91.997 -70.838 55.246
328985 92.028 -70.848 55.235
328986 92.060 -70.859 55.223
328987 92.091 -70.870 55.211
328988 92.122 -70.880 55.200
328989 92.160 -70.898 55.189
328990 92.201 -70.895 55.176
328991 92.229 -70.903 55.160
328992 92.258 -70.911 55.144
328993 92.286 -70.918 55.127
328994 92.314 -70.927 55.108
328995 92.343 -70.934 55.089
328996 92.371 -70.942 55.069
328997 92.409 -70.942 55.047
328998 92.449 -70.942 55.025
328999 92.491 -70.941 55.001
329000 92.531 -70.940 54.976
329001 92.568 -70.931 54.948
329002 92.609 -70.920 54.918
329003 92.656 -70.909 54.887
329004 92.703 -70.898 54.854
329005 92.751 -70.888 54.820
329006 92.798 -70.877 54.785
329007 92.845 -70.867 54.748
329008 92.888 -70.857 54.709
329009 92.915 -70.851 54.669
329010 92.940 -70.845 54.628
329011 92.966 -70.840 54.587
329012 92.992 -70.835 54.547
329013 93.017 -70.829 54.507
329014 93.044 -70.823 54.469
329015 93.069 -70.817 54.432
329016 93.095 -70.812 54.397
329017 93.121 -70.806 54.363
329078 94.824 -70.391 52.756
329079 94.854 -70.383 52.745
329080 94.884 -70.375 52.735
329081 94.914 -70.367 52.725
329082 94.944 -70.359 52.714
329083 94.974 -70.351 52.704
329084 95.004 -70.343 52.695
329085 95.034 -70.335 52.686
329086 95.064 -70.327 52.679
329087 95.095 -70.319 52.673
329088 95.124 -70.311 52.667
329089 95.155 -70.303 52.661
329090 95.173 -70.329 52.653
329091 95.208 -70.350 52.644
329092 95.244 -70.370 52.636
329093 95.280 -70.391 52.629
329094 95.316 -70.410 52.618
329095 95.352 -70.431 52.604
329096 95.388 -70.451 52.591
329097 95.423 -70.471 52.577
329098 95.460 -70.491 52.561
329099 95.497 -70.510 52.545
329100 95.535 -70.528 52.527
329101 95.572 -70.547 52.509
329102 95.610 -70.565 52.491
329103 95.648 -70.583 52.474
329104 95.685 -70.602 52.456
329105 95.723 -70.620 52.438
329106 95.760 -70.638 52.420
329107 95.798 -70.657 52.402
329108 95.832 -70.665 52.385
329109 95.861 -70.659 52.369
329110 95.890 -70.654 52.353
329111 95.919 -70.648 52.337
329112 95.949 -70.642 52.321
329113 95.978 -70.637 52.304
329114 96.007 -70.631 52.289
329115 96.037 -70.626 52.273
329116 96.060 -70.632 52.260
329117 96.083 -70.641 52.249
329118 96.114 -70.652 52.238
329119 96.153 -70.666 52.229
329120 96.183 -70.681 52.218
329121 96.208 -70.696 52.207
329122 96.234 -70.711 52.195
329123 96.259 -70.726 52.183
329124 96.288 -70.742 52.169
329125 96.342 -70.756 52.152
329126 96.397 -70.770 52.133
329127 96.451 -70.784 52.115
329128 96.506 -70.798 52.100
329129 96.561 -70.811 52.087
329130 96.610 -70.822 52.074
329131 96.658 -70.833 52.061
329132 96.705 -70.845 52.047
329133 96.753 -70.855 52.034
329134 96.801 -70.867 52.020
329135 96.830 -70.880 52.007
329136 96.859 -70.892 51.995
329137 96.887 -70.905 51.983
329138 96.928 -70.915 51.969
329139 96.972 -70.924 51.953
329140 97.017 -70.933 51.936
329141 97.062 -70.942 51.918
329142 97.103 -70.948 51.900
329143 97.142 -70.951 51.883
329144 97.180 -70.953 51.866
329145 97.219 -70.956 51.845
329146 97.237 -70.958 51.825
329147 97.255 -70.958 51.807
329148 97.277 -70.963 51.789
329149 97.307 -70.977 51.773
329150 97.336 -70.989 51.756
329151 97.367 -71.002 51.737
329152 97.396 -71.016 51.718
329153 97.427 -71.029 51.701
329154 97.456 -71.042 51.683
329155 97.487 -71.055 51.666
329156 97.516 -71.068 51.646
329157 97.546 -71.081 51.621
329158 97.575 -71.093 51.587
329159 97.605 -71.106 51.552
329160 97.634 -71.119 51.517
329161 97.664 -71.131 51.481
329162 97.693 -71.144 51.443
329163 97.723 -71.156 51.403
329164 97.754 -71.168 51.360
329165 97.787 -71.178 51.313
329166 97.818 -71.176 51.265
329167 97.849 -71.170 51.219
329168 97.886 -71.172 51.178
329169 97.924 -71.175 51.141
329170 97.961 -71.178 51.107
329171 97.999 -71.180 51.068
329172 98.035 -71.185 51.028
329173 98.064 -71.193 50.986
329174 98.092 -71.202 50.941
329175 98.121 -71.210 50.896
329176 98.150 -71.219 50.848
329177 98.178 -71.227 50.798
329178 98.213 -71.235 50.744
329179 98.253 -71.242 50.686
329180 98.293 -71.249 50.627
329181 98.333 -71.256 50.566
329182 98.371 -71.263 50.504
329183 98.407 -71.270 50.443
329184 98.443 -71.277 50.385
329185 98.480 -71.284 50.330
329186 98.516 -71.290 50.276
329187 98.552 -71.297 50.220
329188 98.588 -71.304 50.160
329189 98.625 -71.310 50.100
329190 98.665 -71.313 50.037
329191 98.705 -71.316 49.974
329192 98.739 -71.324 49.911
329193 98.769 -71.335 49.847
329194 98.798 -71.346 49.783
329195 98.829 -71.357 49.719
329196 98.858 -71.368 49.655
329197 98.888 -71.380 49.592
329198 98.920 -71.388 49.530
329199 98.952 -71.396 49.469
329200 98.985 -71.403 49.406
329201 99.017 -71.410 49.346
329202 99.050 -71.418 49.286
329203 99.082 -71.426 49.228
329204 99.098 -71.447 49.173
329205 99.116 -71.469 49.121
329206 99.133 -71.491 49.068
329207 99.153 -71.509 49.017
329208 99.185 -71.522 48.966
329209 99.230 -71.530 48.915
329210 99.275 -71.538 48.864
329211 99.321 -71.545 48.813
329212 99.366 -71.553 48.763
329213 99.411 -71.560 48.714
329214 99.456 -71.568 48.664
329215 99.501 -71.575 48.615
329216 99.547 -71.583 48.567
329217 99.582 -71.581 48.521
329218 99.616 -71.591 48.475
329219 99.650 -71.602 48.429
329220 99.692 -71.609 48.384
329221 99.742 -71.612 48.340
329222 99.792 -71.616 48.298
329223 99.842 -71.619 48.256
329224 99.884 -71.632 48.214
329225 99.920 -71.644 48.172
329226 99.956 -71.657 48.131
329227 99.992 -71.669 48.090
329228 100.028 -71.681 48.048
329229 100.064 -71.693 48.006
329230 100.100 -71.705 47.965
329231 100.136 -71.718 47.926
329232 100.172 -71.730 47.887
329233 100.208 -71.742 47.848
329234 100.244 -71.754 47.808
329235 100.280 -71.766 47.767
329236 100.312 -71.777 47.728
329237 100.339 -71.785 47.691
329238 100.366 -71.793 47.656
329239 100.394 -71.801 47.619
329240 100.421 -71.810 47.582
329241 100.449 -71.818 47.544
329242 100.476 -71.826 47.505
329243 100.501 -71.838 47.468
329244 100.525 -71.853 47.432
329245 100.548 -71.868 47.395
329246 100.571 -71.882 47.357
329247 100.599 -71.904 47.320
329248 100.634 -71.935 47.285
329249 100.669 -71.961 47.251
329250 100.703 -71.966 47.217
329251 100.737 -71.970 47.184
329252 100.771 -71.975 47.150
329253 100.805 -71.980 47.116
329254 100.838 -71.984 47.084
329255 100.873 -71.989 47.050
329256 100.906 -71.993 47.022
329257 100.940 -71.998 46.991
329258 100.974 -72.002 46.957
329259 101.007 -72.007 46.923
329260 101.041 -72.011 46.886
329261 101.075 -72.016 46.849
329262 101.115 -72.021 46.813
329263 101.149 -72.031 46.776
329264 101.185 -72.041 46.739
329265 101.220 -72.053 46.704
329266 101.256 -72.066 46.668
329267 101.292 -72.078 46.630
329268 101.328 -72.092 46.594
329269 101.363 -72.107 46.555
329270 101.398 -72.121 46.517
329271 101.434 -72.136 46.479
329272 101.469 -72.151 46.441
329273 101.505 -72.165 46.400
329274 101.540 -72.180 46.360
329275 101.576 -72.195 46.319
329276 101.611 -72.209 46.278
329277 101.650 -72.216 46.239
329278 101.686 -72.226 46.198
329279 101.718 -72.239 46.156
329280 101.751 -72.252 46.117
329281 101.783 -72.265 46.079
329282 101.809 -72.280 46.041
329283 101.832 -72.295 46.003
329284 101.856 -72.309 45.965
329285 101.879 -72.324 45.928
329286 101.902 -72.339 45.889
329287 101.925 -72.354 45.851
329288 101.949 -72.368 45.814
329289 101.977 -72.383 45.776
329290 102.009 -72.396 45.738
329291 102.040 -72.409 45.702
329292 102.067 -72.423 45.666
329293 102.082 -72.435 45.631
329294 102.095 -72.448 45.596
329295 102.128 -72.461 45.561
329296 102.160 -72.465 45.525
329297 102.185 -72.466 45.490
329298 102.209 -72.467 45.458
329299 102.234 -72.468 45.423
329300 102.259 -72.468 45.390
329301 102.284 -72.468 45.358
329302 102.309 -72.469 45.326
329303 102.333 -72.470 45.295
329304 102.358 -72.471 45.263
329305 102.383 -72.472 45.232
329677 91.736 -70.721 55.340
329678 91.767 -70.732 55.329
329679 91.798 -70.742 55.318
329680 91.829 -70.753 55.306
329681 91.861 -70.764 55.296
329682 91.892 -70.774 55.284
329683 91.923 -70.785 55.272
329684 91.954 -70.795 55.260
329685 91.985 -70.805 55.248
329686 92.016 -70.816 55.236
329687 92.048 -70.826 55.225
329688 92.079 -70.837 55.215
329689 92.110 -70.848 55.204
329690 92.153 -70.872 55.195
329691 92.201 -70.852 55.184
329692 92.227 -70.869 55.172
329693 92.255 -70.876 55.157
329694 92.284 -70.885 55.140
329695 92.312 -70.892 55.121
329696 92.341 -70.900 55.102
329697 92.374 -70.904 55.082
329698 92.415 -70.903 55.060
329699 92.456 -70.902 55.037
329700 92.497 -70.901 55.014
329701 92.538 -70.900 54.989
329702 92.577 -70.896 54.961
329703 92.613 -70.886 54.931
329704 92.653 -70.875 54.900
329705 92.701 -70.864 54.866
329706 92.748 -70.854 54.831
329707 92.796 -70.843 54.795
329708 92.843 -70.832 54.757
329709 92.884 -70.823 54.717
329710 92.910 -70.817 54.678
329711 92.935 -70.812 54.636
329712 92.961 -70.806 54.595
329713 92.987 -70.801 54.554
329714 93.013 -70.795 54.514
329715 93.039 -70.789 54.474
329716 93.064 -70.783 54.436
329717 93.090 -70.778 54.400
329718 93.116 -70.772 54.366
329719 93.142 -70.767 54.334
329779 94.850 -70.341 52.763
329780 94.880 -70.332 52.753
329781 94.910 -70.325 52.744
329782 94.938 -70.317 52.734
329783 94.968 -70.309 52.724
329784 94.996 -70.301 52.715
329785 95.025 -70.293 52.706
329786 95.054 -70.286 52.697
329787 95.083 -70.278 52.692
329788 95.112 -70.271 52.687
329789 95.144 -70.269 52.683
329790 95.179 -70.268 52.677
329791 95.198 -70.291 52.671
329792 95.233 -70.306 52.663
329793 95.268 -70.327 52.655
329794 95.305 -70.347 52.647
329795 95.340 -70.367 52.637
329796 95.376 -70.388 52.623
329797 95.412 -70.408 52.609
329798 95.448 -70.428 52.594
329799 95.483 -70.448 52.578
329800 95.520 -70.469 52.560
329801 95.557 -70.487 52.542
329802 95.595 -70.506 52.523
329803 95.632 -70.524 52.504
329804 95.670 -70.542 52.486
329805 95.707 -70.561 52.468
329806 95.745 -70.579 52.449
329807 95.782 -70.597 52.431
329808 95.820 -70.616 52.413
329809 95.850 -70.613 52.395
329810 95.880 -70.608 52.378
329811 95.909 -70.602 52.362
329812 95.938 -70.597 52.345
329813 95.967 -70.591 52.328
329814 95.996 -70.585 52.311
329815 96.022 -70.588 52.296
329816 96.044 -70.597 52.280
329817 96.066 -70.606 52.267
329818 96.089 -70.615 52.254
329819 96.113 -70.625 52.243
329820 96.152 -70.638 52.231
329821 96.191 -70.652 52.219
329822 96.230 -70.666 52.208
329823 96.265 -70.680 52.196
329824 96.290 -70.695 52.183
329825 96.315 -70.710 52.169
329826 96.350 -70.726 52.152
329827 96.405 -70.740 52.133
329828 96.460 -70.754 52.115
329829 96.515 -70.767 52.096
329830 96.563 -70.779 52.083
329831 96.611 -70.789 52.070
329832 96.659 -70.801 52.057
329833 96.707 -70.812 52.041
329834 96.754 -70.823 52.027
329835 96.802 -70.834 52.014
329836 96.832 -70.847 52.000
329837 96.861 -70.860 51.988
329838 96.893 -70.872 51.978
329839 96.938 -70.881 51.966
329840 96.982 -70.890 51.952
329841 97.027 -70.899 51.939
329842 97.072 -70.908 51.924
329843 97.115 -70.917 51.910
329844 97.154 -70.919 51.896
329845 97.193 -70.922 51.883
329846 97.228 -70.924 51.863
329847 97.246 -70.926 51.844
329848 97.264 -70.927 51.825
329849 97.282 -70.928 51.805
329850 97.311 -70.940 51.787
329851 97.341 -70.953 51.769
329852 97.371 -70.967 51.751
329853 97.401 -70.980 51.731
329854 97.431 -70.992 51.714
329855 97.460 -71.005 51.697
329856 97.490 -71.017 51.679
329857 97.519 -71.030 51.658
329858 97.548 -71.042 51.634
329859 97.578 -71.055 51.599
329860 97.607 -71.068 51.563
329861 97.637 -71.080 51.526
329862 97.667 -71.093 51.488
329863 97.696 -71.105 51.449
329864 97.726 -71.118 51.408
329865 97.758 -71.128 51.365
329866 97.792 -71.139 51.318
329867 97.823 -71.137 51.271
329868 97.853 -71.130 51.227
329869 97.887 -71.127 51.188
329870 97.924 -71.130 51.154
329871 97.962 -71.133 51.121
329872 98.000 -71.136 51.083
329873 98.032 -71.142 51.043
329874 98.061 -71.151 51.000
329875 98.089 -71.159 50.955
329876 98.118 -71.168 50.908
329877 98.147 -71.176 50.858
329878 98.187 -71.183 50.805
329879 98.227 -71.190 50.750
329880 98.267 -71.197 50.690
329881 98.304 -71.204 50.628
329882 98.340 -71.211 50.567
329883 98.376 -71.218 50.505
329884 98.413 -71.224 50.445
329885 98.448 -71.231 50.391
329886 98.485 -71.238 50.338
329887 98.521 -71.245 50.283
329888 98.557 -71.252 50.226
329889 98.593 -71.259 50.165
329890 98.632 -71.263 50.102
329891 98.672 -71.266 50.038
329892 98.712 -71.268 49.973
329893 98.748 -71.274 49.909
329894 98.778 -71.286 49.845
329895 98.808 -71.296 49.781
329896 98.838 -71.308 49.718
329897 98.868 -71.319 49.655
329898 98.898 -71.331 49.593
329899 98.928 -71.341 49.530
329900 98.957 -71.353 49.467
329901 98.990 -71.361 49.405
329902 99.022 -71.368 49.345
329903 99.055 -71.375 49.286
329904 99.087 -71.383 49.230
329905 99.106 -71.403 49.177
329906 99.123 -71.424 49.124
329907 99.140 -71.446 49.072
329908 99.160 -71.466 49.021
329909 99.199 -71.478 48.970
329910 99.244 -71.485 48.917
329911 99.289 -71.493 48.866
329912 99.335 -71.500 48.816
329913 99.379 -71.508 48.766
329914 99.425 -71.516 48.716
329915 99.470 -71.524 48.667
329916 99.515 -71.531 48.619
329917 99.560 -71.539 48.571
329918 99.606 -71.547 48.526
329919 99.640 -71.547 48.480
329920 99.674 -71.558 48.436
329921 99.715 -71.565 48.392
329922 99.765 -71.569 48.348
329923 99.815 -71.572 48.306
329924 99.857 -71.583 48.265
329925 99.894 -71.595 48.224
329926 99.930 -71.608 48.184
329927 99.966 -71.620 48.143
329928 100.002 -71.632 48.101
329929 100.038 -71.644 48.059
329930 100.074 -71.657 48.018
329931 100.110 -71.669 47.978
329932 100.146 -71.681 47.939
329933 100.182 -71.694 47.899
329934 100.218 -71.706 47.859
329935 100.254 -71.718 47.818
329936 100.290 -71.730 47.778
329937 100.323 -71.741 47.739
329938 100.351 -71.749 47.702
329939 100.378 -71.757 47.667
329940 100.405 -71.766 47.629
329941 100.433 -71.774 47.591
329942 100.460 -71.782 47.552
329943 100.487 -71.791 47.513
329944 100.510 -71.806 47.476
329945 100.534 -71.821 47.440
329946 100.557 -71.836 47.402
329947 100.580 -71.851 47.364
329948 100.607 -71.870 47.326
329949 100.642 -71.901 47.291
329950 100.677 -71.925 47.256
329951 100.711 -71.929 47.221
329952 100.745 -71.934 47.188
329953 100.778 -71.939 47.153
329954 100.812 -71.943 47.120
329955 100.846 -71.948 47.087
329956 100.880 -71.952 47.055
329957 100.914 -71.957 47.027
329958 100.948 -71.961 46.997
329959 100.981 -71.966 46.964
329960 101.015 -71.970 46.930
329961 101.049 -71.975 46.893
329962 101.097 -71.981 46.855
329963 101.140 -71.991 46.819
329964 101.164 -72.002 46.783
329965 101.200 -72.008 46.746
329966 101.236 -72.020 46.711
329967 101.271 -72.033 46.675
329968 101.307 -72.045 46.637
329969 101.342 -72.058 46.602
329970 101.378 -72.070 46.565
329971 101.414 -72.083 46.528
329972 101.450 -72.095 46.491
329973 101.485 -72.107 46.453
329974 101.521 -72.121 46.412
329975 101.556 -72.136 46.373
329976 101.592 -72.150 46.332
329977 101.627 -72.165 46.291
329978 101.664 -72.177 46.251
329979 101.699 -72.187 46.209
329980 101.732 -72.201 46.168
329981 101.759 -72.215 46.129
329982 101.782 -72.230 46.093
329983 101.805 -72.244 46.056
329984 101.829 -72.259 46.019
329985 101.852 -72.274 45.981
329986 101.875 -72.289 45.944
329987 101.902 -72.305 45.906
329988 101.929 -72.321 45.869
329989 101.956 -72.338 45.831
329990 101.980 -72.351 45.793
329991 102.006 -72.363 45.755
329992 102.032 -72.374 45.718
329993 102.063 -72.388 45.682
329994 102.095 -72.402 45.646
329995 102.116 -72.415 45.610
329996 102.141 -72.428 45.574
329997 102.168 -72.429 45.539
329998 102.193 -72.430 45.504
329999 102.218 -72.431 45.470
330000 102.242 -72.431 45.436
330001 102.267 -72.431 45.403
330002 102.292 -72.432 45.371
330003 102.317 -72.433 45.339
330004 102.341 -72.434 45.307
330005 102.366 -72.434 45.276
330006 102.392 -72.434 45.244
330007 102.418 -72.434 45.212
330379 91.755 -70.699 55.327
330380 91.786 -70.710 55.316
330381 91.817 -70.720 55.305
330382 91.849 -70.731 55.295
330383 91.880 -70.742 55.283
330384 91.911 -70.752 55.272
330385 91.942 -70.763 55.260
330386 91.973 -70.773 55.248
330387 92.004 -70.784 55.237
330388 92.035 -70.795 55.226
330389 92.067 -70.804 55.216
330390 92.098 -70.815 55.207
330391 92.146 -70.842 55.199
330392 92.197 -70.818 55.192
330393 92.225 -70.835 55.184
330394 92.253 -70.843 55.172
330395 92.282 -70.851 55.157
330396 92.310 -70.858 55.139
330397 92.340 -70.865 55.118
330398 92.380 -70.864 55.097
330399 92.422 -70.864 55.074
330400 92.462 -70.862 55.051
330401 92.504 -70.861 55.028
330402 92.545 -70.861 55.003
330403 92.585 -70.860 54.975
330404 92.623 -70.851 54.945
330405 92.658 -70.841 54.914
330406 92.699 -70.829 54.880
330407 92.746 -70.819 54.842
330408 92.793 -70.808 54.805
330409 92.841 -70.798 54.767
330410 92.879 -70.789 54.727
330411 92.905 -70.784 54.687
330412 92.930 -70.778 54.645
330413 92.956 -70.773 54.603
330414 92.982 -70.767 54.562
330415 93.008 -70.761 54.521
330416 93.034 -70.755 54.481
330417 93.059 -70.750 54.442
330418 93.085 -70.744 54.404
330419 93.111 -70.739 54.369
330420 93.137 -70.732 54.336
330421 93.162 -70.727 54.305
330479 94.831 -70.302 52.782
330480 94.859 -70.294 52.771
330481 94.888 -70.287 52.761
330482 94.917 -70.279 52.751
330483 94.946 -70.271 52.743
330484 94.975 -70.263 52.734
330485 95.004 -70.256 52.726
330486 95.033 -70.248 52.718
330487 95.065 -70.248 52.709
330488 95.098 -70.249 52.705
330489 95.131 -70.250 52.702
330490 95.164 -70.250 52.698
330491 95.194 -70.251 52.694
330492 95.225 -70.268 52.689
330493 95.259 -70.281 52.682
330494 95.295 -70.295 52.674
330495 95.329 -70.309 52.666
330496 95.365 -70.324 52.656
330497 95.400 -70.344 52.643
330498 95.436 -70.365 52.628
330499 95.472 -70.384 52.612
330500 95.508 -70.405 52.595
330501 95.544 -70.425 52.576
330502 95.580 -70.446 52.556
330503 95.616 -70.465 52.537
330504 95.654 -70.483 52.518
330505 95.692 -70.501 52.499
330506 95.729 -70.520 52.480
330507 95.767 -70.538 52.461
330508 95.804 -70.557 52.442
330509 95.839 -70.568 52.424
330510 95.869 -70.563 52.405
330511 95.898 -70.557 52.388
330512 95.927 -70.551 52.370
330513 95.957 -70.546 52.354
330514 95.983 -70.544 52.336
330515 96.006 -70.554 52.319
330516 96.028 -70.563 52.303
330517 96.050 -70.572 52.288
330518 96.073 -70.580 52.274
330519 96.095 -70.589 52.260
330520 96.118 -70.598 52.248
330521 96.152 -70.610 52.235
330522 96.190 -70.625 52.222
330523 96.229 -70.638 52.209
330524 96.268 -70.652 52.196
330525 96.307 -70.666 52.183
330526 96.345 -70.680 52.170
330527 96.371 -70.695 52.153
330528 96.413 -70.710 52.135
330529 96.468 -70.723 52.115
330530 96.516 -70.735 52.096
330531 96.564 -70.746 52.081
330532 96.612 -70.757 52.067
330533 96.660 -70.768 52.054
330534 96.708 -70.779 52.038
330535 96.755 -70.790 52.024
330536 96.803 -70.801 52.011
330537 96.834 -70.814 51.999
330538 96.863 -70.827 51.987
330539 96.903 -70.837 51.978
330540 96.948 -70.846 51.967
330541 96.992 -70.855 51.956
330542 97.037 -70.864 51.945
330543 97.081 -70.874 51.935
330544 97.126 -70.883 51.926
330545 97.166 -70.888 51.917
330546 97.205 -70.890 51.905
330547 97.237 -70.892 51.887
330548 97.255 -70.894 51.866
330549 97.273 -70.895 51.845
330550 97.291 -70.896 51.823
330551 97.315 -70.904 51.804
330552 97.345 -70.917 51.784
330553 97.375 -70.929 51.763
330554 97.404 -70.942 51.743
330555 97.434 -70.955 51.727
330556 97.463 -70.967 51.710
330557 97.493 -70.980 51.690
330558 97.522 -70.992 51.668
330559 97.551 -71.005 51.643
330560 97.581 -71.017 51.609
330561 97.610 -71.030 51.571
330562 97.640 -71.042 51.531
330563 97.669 -71.055 51.492
330564 97.699 -71.068 51.451
330565 97.730 -71.079 51.409
330566 97.764 -71.089 51.366
330567 97.796 -71.099 51.321
330568 97.828 -71.099 51.277
330569 97.858 -71.093 51.236
330570 97.888 -71.085 51.200
330571 97.925 -71.086 51.166
330572 97.962 -71.089 51.134
330573 98.000 -71.092 51.095
330574 98.029 -71.099 51.053
330575 98.058 -71.108 51.009
330576 98.087 -71.116 50.962
330577 98.121 -71.124 50.913
330578 98.161 -71.131 50.861
330579 98.201 -71.138 50.807
330580 98.237 -71.145 50.749
330581 98.273 -71.152 50.689
330582 98.309 -71.159 50.626
330583 98.345 -71.166 50.564
330584 98.381 -71.173 50.504
330585 98.418 -71.180 50.447
330586 98.454 -71.187 50.395
330587 98.490 -71.193 50.342
330588 98.526 -71.200 50.287
330589 98.563 -71.207 50.228
330590 98.599 -71.213 50.166
330591 98.639 -71.216 50.102
330592 98.678 -71.218 50.037
330593 98.719 -71.221 49.973
330594 98.757 -71.225 49.908
330595 98.787 -71.237 49.843
330596 98.817 -71.248 49.780
330597 98.847 -71.259 49.716
330598 98.877 -71.270 49.653
330599 98.907 -71.281 49.591
330600 98.937 -71.293 49.529
330601 98.967 -71.303 49.466
330602 98.997 -71.315 49.405
330603 99.027 -71.326 49.345
330604 99.059 -71.333 49.287
330605 99.092 -71.340 49.232
330606 99.114 -71.358 49.180
330607 99.131 -71.379 49.127
330608 99.153 -71.398 49.076
330609 99.180 -71.415 49.024
330610 99.212 -71.430 48.972
330611 99.257 -71.438 48.919
330612 99.303 -71.447 48.868
330613 99.348 -71.456 48.819
330614 99.394 -71.464 48.769
330615 99.439 -71.472 48.719
330616 99.484 -71.479 48.671
330617 99.529 -71.487 48.623
330618 99.574 -71.494 48.576
330619 99.620 -71.502 48.532
330620 99.663 -71.504 48.487
330621 99.697 -71.514 48.443
330622 99.738 -71.522 48.400
330623 99.788 -71.525 48.357
330624 99.832 -71.535 48.315
330625 99.868 -71.547 48.275
330626 99.904 -71.559 48.235
330627 99.940 -71.571 48.196
330628 99.976 -71.584 48.155
330629 100.012 -71.596 48.114
330630 100.048 -71.608 48.072
330631 100.084 -71.620 48.031
330632 100.120 -71.632 47.992
330633 100.156 -71.644 47.952
330634 100.192 -71.657 47.912
330635 100.228 -71.669 47.871
330636 100.264 -71.682 47.830
330637 100.300 -71.694 47.790
330638 100.335 -71.705 47.751
330639 100.362 -71.713 47.713
330640 100.390 -71.722 47.677
330641 100.417 -71.730 47.639
330642 100.444 -71.738 47.600
330643 100.472 -71.747 47.561
330644 100.497 -71.759 47.521
330645 100.520 -71.774 47.484
330646 100.543 -71.789 47.448
330647 100.566 -71.804 47.410
330648 100.590 -71.820 47.371
330649 100.615 -71.836 47.332
330650 100.650 -71.867 47.296
330651 100.685 -71.889 47.261
330652 100.718 -71.893 47.226
330653 100.752 -71.898 47.192
330654 100.786 -71.902 47.157
330655 100.820 -71.907 47.123
330656 100.854 -71.911 47.091
330657 100.888 -71.916 47.062
330658 100.921 -71.920 47.034
330659 100.955 -71.925 47.003
330660 100.989 -71.929 46.970
330661 101.025 -71.934 46.936
330662 101.078 -71.942 46.900
330663 101.132 -71.949 46.863
330664 101.152 -71.963 46.827
330665 101.178 -71.973 46.791
330666 101.214 -71.980 46.755
330667 101.250 -71.987 46.719
330668 101.286 -71.999 46.682
330669 101.321 -72.012 46.645
330670 101.357 -72.024 46.609
330671 101.393 -72.037 46.572
330672 101.428 -72.049 46.536
330673 101.464 -72.061 46.500
330674 101.500 -72.074 46.462
330675 101.536 -72.086 46.421
330676 101.571 -72.099 46.381
330677 101.607 -72.111 46.342
330678 101.642 -72.124 46.302
330679 101.678 -72.136 46.261
330680 101.709 -72.150 46.220
330681 101.736 -72.167 46.180
330682 101.763 -72.183 46.141
330683 101.790 -72.200 46.104
330684 101.817 -72.217 46.068
330685 101.844 -72.233 46.032
330686 101.871 -72.249 45.995
330687 101.898 -72.266 45.959
330688 101.925 -72.283 45.922
330689 101.952 -72.299 45.885
330690 101.979 -72.315 45.847
330691 102.001 -72.325 45.809
330692 102.026 -72.336 45.770
330693 102.051 -72.348 45.734
330694 102.077 -72.359 45.697
330695 102.102 -72.371 45.661
330696 102.127 -72.383 45.624
330697 102.153 -72.390 45.587
330698 102.179 -72.390 45.551
330699 102.206 -72.390 45.516
330700 102.232 -72.390 45.482
330701 102.258 -72.390 45.448
330702 102.284 -72.390 45.416
330703 102.311 -72.390 45.383
330704 102.337 -72.389 45.351
330705 102.364 -72.389 45.319
330706 102.390 -72.389 45.288
330707 102.416 -72.389 45.256
330708 102.443 -72.388 45.224
331081 91.774 -70.678 55.314
331082 91.805 -70.688 55.304
331083 91.837 -70.699 55.293
331084 91.868 -70.709 55.282
331085 91.899 -70.720 55.271
331086 91.930 -70.730 55.260
331087 91.961 -70.741 55.248
331088 91.992 -70.751 55.237
331089 92.023 -70.762 55.227
331090 92.055 -70.773 55.217
331091 92.086 -70.783 55.209
331092 92.139 -70.813 55.202
331093 92.190 -70.789 55.196
331094 92.225 -70.795 55.192
331095 92.251 -70.809 55.185
331096 92.280 -70.817 55.173
331097 92.308 -70.824 55.156
331098 92.346 -70.826 55.135
331099 92.387 -70.824 55.112
331100 92.428 -70.823 55.089
331101 92.469 -70.823 55.065
331102 92.510 -70.822 55.041
331103 92.551 -70.821 55.016
331104 92.592 -70.820 54.989
331105 92.632 -70.817 54.960
331106 92.667 -70.806 54.928
331107 92.703 -70.795 54.894
331108 92.744 -70.785 54.856
331109 92.791 -70.774 54.819
331110 92.839 -70.764 54.781
331111 92.874 -70.755 54.739
331112 92.900 -70.750 54.697
331113 92.926 -70.744 54.655
331114 92.951 -70.739 54.613
331115 92.977 -70.732 54.570
331116 93.003 -70.727 54.529
331117 93.028 -70.721 54.488
331118 93.055 -70.716 54.449
331119 93.080 -70.710 54.410
331120 93.106 -70.704 54.373
331121 93.132 -70.699 54.338
331122 93.157 -70.693 54.306
331123 93.184 -70.688 54.276
331180 94.838 -70.256 52.788
331181 94.867 -70.249 52.779
331182 94.896 -70.241 52.770
331183 94.925 -70.233 52.763
331184 94.954 -70.225 52.754
331185 94.986 -70.226 52.746
331186 95.020 -70.227 52.738
331187 95.052 -70.228 52.730
331188 95.086 -70.228 52.723
331189 95.119 -70.229 52.719
331190 95.151 -70.230 52.716
331191 95.185 -70.231 52.713
331192 95.216 -70.231 52.710
331193 95.252 -70.244 52.707
331194 95.286 -70.258 52.701
331195 95.321 -70.272 52.694
331196 95.356 -70.285 52.687
331197 95.391 -70.299 52.675
331198 95.426 -70.313 52.662
331199 95.461 -70.327 52.647
331200 95.496 -70.341 52.631
331201 95.532 -70.362 52.613
331202 95.568 -70.382 52.592
331203 95.604 -70.402 52.571
331204 95.639 -70.422 52.551
331205 95.676 -70.442 52.531
331206 95.713 -70.460 52.512
331207 95.751 -70.479 52.492
331208 95.789 -70.497 52.473
331209 95.826 -70.516 52.455
331210 95.858 -70.517 52.435
331211 95.887 -70.511 52.416
331212 95.916 -70.506 52.398
331213 95.945 -70.501 52.380
331214 95.967 -70.510 52.363
331215 95.990 -70.519 52.345
331216 96.012 -70.528 52.328
331217 96.035 -70.537 52.310
331218 96.057 -70.546 52.295
331219 96.080 -70.554 52.281
331220 96.102 -70.563 52.267
331221 96.124 -70.572 52.253
331222 96.151 -70.582 52.239
331223 96.190 -70.597 52.225
331224 96.229 -70.610 52.210
331225 96.267 -70.624 52.196
331226 96.306 -70.638 52.184
331227 96.345 -70.652 52.170
331228 96.384 -70.666 52.154
331229 96.424 -70.679 52.136
331230 96.470 -70.691 52.117
331231 96.517 -70.702 52.098
331232 96.565 -70.713 52.082
331233 96.613 -70.724 52.066
331234 96.661 -70.735 52.052
331235 96.709 -70.747 52.037
331236 96.756 -70.757 52.024
331237 96.804 -70.769 52.011
331238 96.837 -70.782 52.000
331239 96.868 -70.794 51.990
331240 96.913 -70.803 51.982
331241 96.957 -70.812 51.972
331242 97.002 -70.821 51.962
331243 97.046 -70.830 51.955
331244 97.091 -70.839 51.950
331245 97.136 -70.848 51.945
331246 97.179 -70.856 51.939
331247 97.218 -70.859 51.929
331248 97.246 -70.861 51.911
331249 97.264 -70.862 51.888
331250 97.282 -70.864 51.865
331251 97.298 -70.864 51.842
331252 97.318 -70.871 51.820
331253 97.348 -70.878 51.797
331254 97.377 -70.891 51.775
331255 97.407 -70.904 51.755
331256 97.436 -70.916 51.738
331257 97.466 -70.929 51.720
331258 97.495 -70.942 51.700
331259 97.525 -70.954 51.677
331260 97.555 -70.967 51.650
331261 97.584 -70.979 51.616
331262 97.613 -70.992 51.576
331263 97.643 -71.005 51.536
331264 97.672 -71.017 51.494
331265 97.702 -71.029 51.452
331266 97.735 -71.039 51.411
331267 97.768 -71.049 51.368
331268 97.801 -71.060 51.325
331269 97.833 -71.061 51.283
331270 97.863 -71.054 51.244
331271 97.893 -71.047 51.209
331272 97.925 -71.041 51.177
331273 97.962 -71.044 51.145
331274 97.998 -71.049 51.105
331275 98.027 -71.057 51.061
331276 98.056 -71.065 51.015
331277 98.095 -71.072 50.966
331278 98.134 -71.080 50.915
331279 98.170 -71.086 50.861
331280 98.206 -71.093 50.805
331281 98.242 -71.100 50.746
331282 98.278 -71.107 50.685
331283 98.314 -71.114 50.623
331284 98.351 -71.121 50.562
331285 98.387 -71.127 50.504
331286 98.423 -71.134 50.450
331287 98.459 -71.141 50.399
331288 98.495 -71.148 50.345
331289 98.531 -71.155 50.289
331290 98.568 -71.162 50.229
331291 98.606 -71.166 50.166
331292 98.646 -71.168 50.101
331293 98.686 -71.171 50.036
331294 98.725 -71.174 49.972
331295 98.765 -71.177 49.906
331296 98.796 -71.187 49.842
331297 98.827 -71.199 49.778
331298 98.856 -71.210 49.714
331299 98.886 -71.221 49.651
331300 98.916 -71.232 49.589
331301 98.946 -71.243 49.528
331302 98.976 -71.255 49.466
331303 99.006 -71.265 49.405
331304 99.036 -71.277 49.346
331305 99.066 -71.288 49.289
331306 99.096 -71.298 49.235
331307 99.121 -71.312 49.183
331308 99.146 -71.331 49.131
331309 99.173 -71.347 49.079
331310 99.201 -71.364 49.028
331311 99.228 -71.381 48.975
331312 99.270 -71.390 48.921
331313 99.316 -71.400 48.871
331314 99.361 -71.408 48.821
331315 99.407 -71.416 48.772
331316 99.452 -71.425 48.723
331317 99.497 -71.434 48.675
331318 99.543 -71.442 48.628
331319 99.588 -71.450 48.582
331320 99.633 -71.458 48.538
331321 99.679 -71.466 48.494
331322 99.720 -71.471 48.451
331323 99.761 -71.478 48.408
331324 99.806 -71.486 48.365
331325 99.842 -71.498 48.324
331326 99.878 -71.510 48.286
331327 99.914 -71.522 48.247
331328 99.950 -71.534 48.209
331329 99.986 -71.547 48.168
331330 100.022 -71.560 48.127
331331 100.058 -71.572 48.085
331332 100.094 -71.584 48.045
331333 100.130 -71.596 48.006
331334 100.167 -71.608 47.966
331335 100.202 -71.621 47.926
331336 100.239 -71.633 47.885
331337 100.275 -71.645 47.844
331338 100.311 -71.657 47.803
331339 100.347 -71.669 47.764
331340 100.374 -71.678 47.724
331341 100.402 -71.686 47.686
331342 100.429 -71.694 47.647
331343 100.456 -71.703 47.608
331344 100.483 -71.712 47.569
331345 100.506 -71.727 47.529
331346 100.530 -71.742 47.492
331347 100.553 -71.757 47.456
331348 100.576 -71.773 47.418
331349 100.599 -71.788 47.379
331350 100.622 -71.803 47.340
331351 100.657 -71.833 47.303
331352 100.692 -71.852 47.266
331353 100.726 -71.857 47.230
331354 100.760 -71.861 47.195
331355 100.793 -71.866 47.162
331356 100.828 -71.870 47.128
331357 100.861 -71.875 47.096
331358 100.895 -71.879 47.066
331359 100.929 -71.884 47.037
331360 100.963 -71.889 47.005
331361 101.006 -71.895 46.971
331362 101.059 -71.902 46.938
331363 101.113 -71.909 46.904
331364 101.148 -71.920 46.869
331365 101.165 -71.935 46.835
331366 101.193 -71.945 46.800
331367 101.229 -71.951 46.765
331368 101.265 -71.956 46.729
331369 101.301 -71.967 46.691
331370 101.336 -71.979 46.654
331371 101.372 -71.991 46.617
331372 101.407 -72.004 46.580
331373 101.443 -72.016 46.543
331374 101.479 -72.029 46.506
331375 101.515 -72.041 46.469
331376 101.550 -72.053 46.429
331377 101.586 -72.066 46.389
331378 101.621 -72.078 46.349
331379 101.657 -72.091 46.310
331380 101.693 -72.103 46.270
331381 101.729 -72.121 46.230
331382 101.758 -72.141 46.191
331383 101.786 -72.161 46.152
331384 101.814 -72.177 46.116
331385 101.840 -72.194 46.080
331386 101.868 -72.210 46.044
331387 101.894 -72.227 46.008
331388 101.921 -72.243 45.972
331389 101.948 -72.260 45.936
331390 101.975 -72.276 45.899
331391 102.001 -72.292 45.862
331392 102.022 -72.299 45.823
331393 102.045 -72.309 45.785
331394 102.071 -72.321 45.748
331395 102.096 -72.333 45.711
331396 102.120 -72.340 45.674
331397 102.136 -72.338 45.637
331398 102.166 -72.340 45.600
331399 102.196 -72.341 45.563
331400 102.226 -72.343 45.527
331401 102.256 -72.344 45.493
331402 102.283 -72.344 45.461
331403 102.309 -72.344 45.429
331404 102.335 -72.344 45.396
331405 102.362 -72.343 45.363
331406 102.388 -72.343 45.330
331407 102.414 -72.343 45.299
331408 102.441 -72.343 45.268
331409 102.467 -72.343 45.237
331783 91.793 -70.656 55.302
331784 91.824 -70.666 55.291
331785 91.856 -70.677 55.280
331786 91.887 -70.688 55.269
331787 91.918 -70.698 55.258
331788 91.949 -70.709 55.248
331789 91.981 -70.719 55.238
331790 92.011 -70.729 55.227
331791 92.042 -70.740 55.218
331792 92.078 -70.754 55.210
331793 92.132 -70.783 55.203
331794 92.183 -70.759 55.199
331795 92.225 -70.751 55.197
331796 92.249 -70.775 55.196
331797 92.278 -70.782 55.188
331798 92.312 -70.786 55.172
331799 92.353 -70.785 55.151
331800 92.393 -70.785 55.128
331801 92.435 -70.784 55.104
331802 92.476 -70.783 55.079
331803 92.516 -70.782 55.054
331804 92.558 -70.781 55.028
331805 92.598 -70.780 55.001
331806 92.640 -70.779 54.973
331807 92.677 -70.772 54.942
331808 92.713 -70.761 54.907
331809 92.748 -70.750 54.870
331810 92.789 -70.739 54.833
331811 92.836 -70.729 54.794
331812 92.869 -70.722 54.752
331813 92.895 -70.716 54.709
331814 92.921 -70.710 54.667
331815 92.946 -70.704 54.625
331816 92.972 -70.699 54.582
331817 92.998 -70.693 54.540
331818 93.024 -70.688 54.498
331819 93.050 -70.682 54.457
331820 93.075 -70.676 54.417
331821 93.101 -70.670 54.380
331822 93.127 -70.665 54.343
331823 93.155 -70.658 54.309
331824 93.185 -70.650 54.277
331825 93.215 -70.642 54.246
331881 94.845 -70.211 52.795
331882 94.875 -70.204 52.786
331883 94.908 -70.205 52.779
331884 94.941 -70.206 52.772
331885 94.974 -70.206 52.764
331886 95.007 -70.206 52.757
331887 95.040 -70.207 52.750
331888 95.073 -70.208 52.743
331889 95.106 -70.209 52.737
331890 95.139 -70.209 52.734
331891 95.172 -70.210 52.731
331892 95.207 -70.209 52.728
331893 95.244 -70.207 52.726
331894 95.278 -70.221 52.724
331895 95.313 -70.234 52.718
331896 95.348 -70.248 52.712
331897 95.383 -70.262 52.705
331898 95.418 -70.275 52.694
331899 95.453 -70.289 52.680
331900 95.488 -70.303 52.665
331901 95.523 -70.317 52.649
331902 95.558 -70.331 52.630
331903 95.593 -70.344 52.609
331904 95.628 -70.359 52.587
331905 95.664 -70.379 52.565
331906 95.699 -70.400 52.545
331907 95.736 -70.419 52.524
331908 95.773 -70.438 52.505
331909 95.811 -70.456 52.485
331910 95.847 -70.472 52.466
331911 95.876 -70.466 52.447
331912 95.906 -70.460 52.428
331913 95.929 -70.466 52.409
331914 95.952 -70.475 52.390
331915 95.974 -70.485 52.374
331916 95.996 -70.494 52.355
331917 96.019 -70.502 52.337
331918 96.041 -70.511 52.319
331919 96.063 -70.520 52.304
331920 96.086 -70.529 52.289
331921 96.108 -70.538 52.275
331922 96.130 -70.547 52.261
331923 96.154 -70.555 52.248
331924 96.191 -70.568 52.232
331925 96.232 -70.581 52.215
331926 96.271 -70.594 52.200
331927 96.311 -70.607 52.187
331928 96.352 -70.621 52.172
331929 96.391 -70.635 52.157
331930 96.431 -70.648 52.140
331931 96.467 -70.651 52.122
331932 96.517 -70.666 52.103
331933 96.566 -70.680 52.084
331934 96.614 -70.691 52.068
331935 96.662 -70.703 52.054
331936 96.710 -70.714 52.039
331937 96.757 -70.726 52.025
331938 96.805 -70.736 52.014
331939 96.839 -70.749 52.004
331940 96.878 -70.759 51.996
331941 96.923 -70.768 51.989
331942 96.967 -70.777 51.981
331943 97.012 -70.787 51.974
331944 97.056 -70.796 51.970
331945 97.101 -70.805 51.968
331946 97.145 -70.814 51.966
331947 97.190 -70.823 51.961
331948 97.230 -70.828 51.951
331949 97.254 -70.829 51.932
331950 97.273 -70.830 51.907
331951 97.287 -70.832 51.882
331952 97.302 -70.833 51.857
331953 97.324 -70.842 51.832
331954 97.348 -70.850 51.808
331955 97.380 -70.854 51.786
331956 97.409 -70.865 51.765
331957 97.439 -70.878 51.748
331958 97.468 -70.891 51.728
331959 97.498 -70.903 51.707
331960 97.528 -70.916 51.682
331961 97.557 -70.928 51.654
331962 97.587 -70.941 51.619
331963 97.616 -70.954 51.580
331964 97.646 -70.966 51.538
331965 97.675 -70.979 51.496
331966 97.707 -70.989 51.454
331967 97.740 -71.000 51.414
331968 97.773 -71.010 51.374
331969 97.806 -71.021 51.333
331970 97.838 -71.023 51.291
331971 97.868 -71.016 51.251
331972 97.898 -71.008 51.216
331973 97.929 -71.002 51.184
331974 97.963 -70.999 51.152
331975 97.995 -71.005 51.111
331976 98.029 -71.014 51.066
331977 98.067 -71.021 51.018
331978 98.103 -71.027 50.966
331979 98.139 -71.034 50.913
331980 98.175 -71.041 50.858
331981 98.211 -71.048 50.801
331982 98.247 -71.055 50.741
331983 98.284 -71.061 50.680
331984 98.320 -71.068 50.620
331985 98.356 -71.075 50.561
331986 98.392 -71.082 50.506
331987 98.428 -71.089 50.455
331988 98.464 -71.096 50.401
331989 98.501 -71.102 50.347
331990 98.537 -71.109 50.289
331991 98.573 -71.116 50.227
331992 98.613 -71.119 50.164
331993 98.653 -71.121 50.098
331994 98.693 -71.124 50.033
331995 98.732 -71.127 49.968
331996 98.772 -71.130 49.904
331997 98.806 -71.138 49.839
331998 98.836 -71.149 49.776
331999 98.866 -71.161 49.713
332000 98.896 -71.172 49.649
332001 98.926 -71.183 49.588
332002 98.955 -71.194 49.527
332003 98.985 -71.205 49.466
332004 99.015 -71.217 49.406
332005 99.045 -71.227 49.348
332006 99.075 -71.239 49.291
332007 99.105 -71.250 49.238
332008 99.137 -71.264 49.186
332009 99.167 -71.280 49.134
332010 99.194 -71.296 49.082
332011 99.221 -71.313 49.030
332012 99.249 -71.330 48.977
332013 99.284 -71.343 48.924
332014 99.329 -71.352 48.874
332015 99.374 -71.360 48.825
332016 99.420 -71.368 48.775
332017 99.465 -71.378 48.728
332018 99.510 -71.386 48.680
332019 99.556 -71.394 48.633
332020 99.602 -71.403 48.589
332021 99.647 -71.412 48.545
332022 99.692 -71.420 48.502
332023 99.737 -71.429 48.459
332024 99.780 -71.437 48.417
332025 99.816 -71.450 48.374
332026 99.852 -71.462 48.335
332027 99.888 -71.474 48.296
332028 99.924 -71.486 48.258
332029 99.960 -71.498 48.221
332030 99.996 -71.510 48.181
332031 100.032 -71.523 48.141
332032 100.068 -71.535 48.101
332033 100.105 -71.547 48.061
332034 100.140 -71.560 48.022
332035 100.177 -71.572 47.982
332036 100.212 -71.584 47.941
332037 100.249 -71.597 47.900
332038 100.285 -71.609 47.858
332039 100.321 -71.621 47.816
332040 100.357 -71.633 47.775
332041 100.386 -71.642 47.735
332042 100.413 -71.650 47.695
332043 100.440 -71.659 47.656
332044 100.468 -71.667 47.617
332045 100.492 -71.680 47.577
332046 100.515 -71.695 47.538
332047 100.539 -71.710 47.501
332048 100.562 -71.726 47.463
332049 100.585 -71.741 47.425
332050 100.609 -71.756 47.385
332051 100.632 -71.771 47.346
332052 100.665 -71.799 47.308
332053 100.700 -71.817 47.271
332054 100.734 -71.821 47.236
332055 100.767 -71.826 47.200
332056 100.802 -71.830 47.166
332057 100.835 -71.834 47.133
332058 100.869 -71.838 47.102
332059 100.903 -71.843 47.071
332060 100.936 -71.848 47.039
332061 100.987 -71.854 47.006
332062 101.041 -71.862 46.973
332063 101.095 -71.870 46.942
332064 101.144 -71.878 46.910
332065 101.160 -71.892 46.877
332066 101.177 -71.907 46.844
332067 101.207 -71.916 46.810
332068 101.243 -71.921 46.775
332069 101.279 -71.927 46.739
332070 101.315 -71.933 46.701
332071 101.351 -71.945 46.664
332072 101.387 -71.958 46.626
332073 101.422 -71.970 46.588
332074 101.458 -71.983 46.551
332075 101.493 -71.995 46.513
332076 101.529 -72.008 46.475
332077 101.565 -72.020 46.436
332078 101.601 -72.033 46.397
332079 101.636 -72.045 46.357
332080 101.672 -72.058 46.318
332081 101.709 -72.071 46.279
332082 101.748 -72.089 46.239
332083 101.776 -72.110 46.201
332084 101.805 -72.130 46.164
332085 101.834 -72.150 46.127
332086 101.863 -72.170 46.091
332087 101.891 -72.187 46.055
332088 101.917 -72.204 46.020
332089 101.945 -72.221 45.983
332090 101.971 -72.237 45.948
332091 101.998 -72.253 45.911
332092 102.023 -72.266 45.873
332093 102.044 -72.274 45.835
332094 102.065 -72.283 45.796
332095 102.089 -72.288 45.759
332096 102.112 -72.291 45.721
332097 102.120 -72.286 45.683
332098 102.150 -72.288 45.646
332099 102.180 -72.290 45.609
332100 102.210 -72.291 45.573
332101 102.240 -72.293 45.538
332102 102.270 -72.294 45.504
332103 102.300 -72.296 45.471
332104 102.330 -72.297 45.438
332105 102.360 -72.299 45.406
332106 102.386 -72.298 45.374
332107 102.412 -72.298 45.341
332108 102.439 -72.298 45.310
332109 102.465 -72.298 45.278
332110 102.491 -72.297 45.248
332485 91.812 -70.634 55.288
332486 91.844 -70.644 55.278
332487 91.875 -70.655 55.267
332488 91.906 -70.666 55.256
332489 91.937 -70.676 55.246
332490 91.968 -70.687 55.237
332491 92.000 -70.698 55.228
332492 92.030 -70.708 55.220
332493 92.072 -70.726 55.211
332494 92.125 -70.754 55.204
332495 92.176 -70.729 55.200
332496 92.226 -70.707 55.199
332497 92.248 -70.739 55.202
332498 92.277 -70.748 55.197
332499 92.318 -70.747 55.185
332500 92.359 -70.746 55.167
332501 92.400 -70.745 55.146
332502 92.441 -70.744 55.120
332503 92.482 -70.743 55.093
332504 92.523 -70.742 55.066
332505 92.564 -70.742 55.038
332506 92.605 -70.741 55.012
332507 92.646 -70.739 54.985
332508 92.687 -70.738 54.956
332509 92.722 -70.726 54.920
332510 92.758 -70.716 54.883
332511 92.793 -70.705 54.846
332512 92.834 -70.695 54.807
332513 92.864 -70.688 54.765
332514 92.890 -70.682 54.722
332515 92.916 -70.676 54.680
332516 92.941 -70.670 54.638
332517 92.970 -70.663 54.595
332518 93.000 -70.655 54.552
332519 93.030 -70.648 54.509
332520 93.060 -70.639 54.467
332521 93.090 -70.632 54.426
332522 93.120 -70.624 54.387
332523 93.149 -70.616 54.350
332524 93.178 -70.609 54.314
332525 93.206 -70.601 54.280
332526 93.235 -70.593 54.248
332527 93.264 -70.585 54.218
332581 94.829 -70.183 52.808
332582 94.862 -70.184 52.802
332583 94.895 -70.184 52.794
332584 94.928 -70.185 52.786
332585 94.961 -70.186 52.780
332586 94.994 -70.187 52.773
332587 95.027 -70.187 52.767
332588 95.060 -70.188 52.762
332589 95.093 -70.189 52.756
332590 95.126 -70.189 52.751
332591 95.159 -70.190 52.748
332592 95.192 -70.190 52.746
332593 95.234 -70.183 52.744
332594 95.270 -70.183 52.743
332595 95.306 -70.197 52.740
332596 95.340 -70.211 52.735
332597 95.375 -70.225 52.729
332598 95.410 -70.238 52.723
332599 95.445 -70.252 52.712
332600 95.480 -70.265 52.699
332601 95.515 -70.279 52.683
332602 95.550 -70.293 52.665
332603 95.585 -70.307 52.647
332604 95.620 -70.321 52.626
332605 95.655 -70.334 52.603
332606 95.690 -70.348 52.579
332607 95.725 -70.362 52.558
332608 95.760 -70.377 52.537
332609 95.796 -70.397 52.517
332610 95.833 -70.416 52.498
332611 95.866 -70.420 52.477
332612 95.891 -70.423 52.459
332613 95.913 -70.432 52.440
332614 95.936 -70.441 52.421
332615 95.958 -70.450 52.402
332616 95.981 -70.459 52.384
332617 96.004 -70.466 52.365
332618 96.027 -70.475 52.346
332619 96.050 -70.483 52.329
332620 96.074 -70.491 52.313
332621 96.097 -70.499 52.298
332622 96.120 -70.507 52.284
332623 96.143 -70.516 52.270
332624 96.166 -70.524 52.256
332625 96.202 -70.536 52.240
332626 96.240 -70.549 52.222
332627 96.279 -70.563 52.206
332628 96.319 -70.576 52.193
332629 96.360 -70.589 52.179
332630 96.399 -70.603 52.163
332631 96.439 -70.614 52.146
332632 96.474 -70.616 52.127
332633 96.514 -70.623 52.109
332634 96.563 -70.638 52.091
332635 96.612 -70.653 52.075
332636 96.662 -70.667 52.060
332637 96.711 -70.682 52.044
332638 96.758 -70.693 52.032
332639 96.806 -70.704 52.022
332640 96.843 -70.716 52.014
332641 96.888 -70.725 52.008
332642 96.933 -70.734 52.004
332643 96.977 -70.743 51.998
332644 97.022 -70.752 51.992
332645 97.066 -70.761 51.988
332646 97.111 -70.770 51.987
332647 97.155 -70.779 51.984
332648 97.199 -70.789 51.979
332649 97.243 -70.796 51.969
332650 97.263 -70.798 51.949
332651 97.277 -70.798 51.923
332652 97.290 -70.799 51.895
332653 97.308 -70.804 51.867
332654 97.330 -70.813 51.841
332655 97.353 -70.822 51.817
332656 97.379 -70.829 51.795
332657 97.411 -70.832 51.775
332658 97.442 -70.840 51.757
332659 97.471 -70.852 51.735
332660 97.501 -70.865 51.712
332661 97.530 -70.877 51.686
332662 97.560 -70.890 51.656
332663 97.589 -70.903 51.618
332664 97.619 -70.915 51.578
332665 97.648 -70.928 51.538
332666 97.678 -70.940 51.497
332667 97.712 -70.950 51.458
332668 97.745 -70.961 51.420
332669 97.778 -70.970 51.380
332670 97.811 -70.980 51.338
332671 97.843 -70.985 51.296
332672 97.873 -70.977 51.258
332673 97.903 -70.970 51.223
332674 97.934 -70.963 51.190
332675 97.964 -70.956 51.155
332676 98.000 -70.961 51.113
332677 98.036 -70.968 51.066
332678 98.072 -70.975 51.016
332679 98.108 -70.982 50.963
332680 98.144 -70.989 50.909
332681 98.180 -70.995 50.853
332682 98.217 -71.002 50.795
332683 98.253 -71.009 50.735
332684 98.289 -71.016 50.675
332685 98.325 -71.023 50.617
332686 98.361 -71.030 50.562
332687 98.397 -71.036 50.510
332688 98.434 -71.043 50.459
332689 98.470 -71.050 50.404
332690 98.506 -71.057 50.347
332691 98.542 -71.064 50.289
332692 98.580 -71.069 50.227
332693 98.620 -71.071 50.162
332694 98.660 -71.074 50.096
332695 98.700 -71.077 50.030
332696 98.739 -71.080 49.964
332697 98.779 -71.083 49.900
332698 98.815 -71.089 49.836
332699 98.845 -71.100 49.773
332700 98.875 -71.111 49.710
332701 98.905 -71.123 49.648
332702 98.935 -71.133 49.586
332703 98.964 -71.145 49.526
332704 98.995 -71.156 49.466
332705 99.024 -71.168 49.407
332706 99.055 -71.179 49.349
332707 99.084 -71.190 49.293
332708 99.117 -71.204 49.239
332709 99.149 -71.218 49.188
332710 99.182 -71.232 49.136
332711 99.214 -71.246 49.085
332712 99.242 -71.262 49.032
332713 99.269 -71.279 48.980
332714 99.297 -71.295 48.927
332715 99.342 -71.304 48.877
332716 99.387 -71.312 48.829
332717 99.433 -71.321 48.780
332718 99.478 -71.330 48.732
332719 99.523 -71.338 48.685
332720 99.569 -71.346 48.640
332721 99.615 -71.356 48.596
332722 99.659 -71.364 48.553
332723 99.699 -71.371 48.510
332724 99.740 -71.378 48.468
332725 99.783 -71.393 48.427
332726 99.822 -71.409 48.385
332727 99.861 -71.424 48.346
332728 99.898 -71.437 48.308
332729 99.934 -71.450 48.271
332730 99.970 -71.462 48.234
332731 100.006 -71.474 48.195
332732 100.043 -71.487 48.156
332733 100.078 -71.499 48.116
332734 100.115 -71.511 48.076
332735 100.150 -71.523 48.037
332736 100.187 -71.535 47.997
332737 100.223 -71.547 47.956
332738 100.259 -71.560 47.913
332739 100.295 -71.572 47.870
332740 100.331 -71.585 47.827
332741 100.367 -71.597 47.785
332742 100.398 -71.607 47.744
332743 100.425 -71.615 47.704
332744 100.453 -71.623 47.664
332745 100.478 -71.634 47.624
332746 100.501 -71.649 47.584
332747 100.525 -71.663 47.545
332748 100.548 -71.679 47.507
332749 100.571 -71.694 47.469
332750 100.595 -71.709 47.430
332751 100.618 -71.724 47.391
332752 100.641 -71.739 47.352
332753 100.673 -71.766 47.314
332754 100.707 -71.780 47.277
332755 100.741 -71.785 47.242
332756 100.775 -71.789 47.206
332757 100.809 -71.794 47.171
332758 100.843 -71.798 47.138
332759 100.876 -71.803 47.107
332760 100.915 -71.807 47.075
332761 100.969 -71.815 47.042
332762 101.022 -71.823 47.009
332763 101.076 -71.829 46.977
332764 101.130 -71.837 46.947
332765 101.156 -71.850 46.919
332766 101.172 -71.864 46.887
332767 101.189 -71.879 46.853
332768 101.221 -71.886 46.819
332769 101.258 -71.892 46.785
332770 101.293 -71.898 46.748
332771 101.330 -71.904 46.711
332772 101.365 -71.913 46.673
332773 101.401 -71.925 46.635
332774 101.437 -71.938 46.599
332775 101.473 -71.950 46.561
332776 101.509 -71.962 46.523
332777 101.544 -71.975 46.485
332778 101.580 -71.987 46.447
332779 101.615 -72.000 46.407
332780 101.651 -72.012 46.368
332781 101.687 -72.024 46.329
332782 101.727 -72.040 46.289
332783 101.766 -72.058 46.250
332784 101.795 -72.078 46.212
332785 101.824 -72.099 46.175
332786 101.852 -72.118 46.138
332787 101.882 -72.139 46.102
332788 101.910 -72.158 46.065
332789 101.939 -72.179 46.029
332790 101.968 -72.198 45.992
332791 101.994 -72.215 45.956
332792 102.022 -72.230 45.919
332793 102.044 -72.241 45.881
332794 102.058 -72.236 45.843
332795 102.081 -72.240 45.803
332796 102.102 -72.242 45.765
332797 102.104 -72.234 45.726
332798 102.134 -72.236 45.689
332799 102.164 -72.237 45.653
332800 102.194 -72.239 45.617
332801 102.224 -72.240 45.582
332802 102.254 -72.242 45.548
332803 102.284 -72.243 45.514
332804 102.314 -72.245 45.481
332805 102.344 -72.246 45.449
332806 102.374 -72.248 45.416
332807 102.404 -72.249 45.384
332808 102.434 -72.251 45.352
332809 102.463 -72.252 45.321
332810 102.489 -72.252 45.289
333186 91.800 -70.602 55.285
333187 91.832 -70.613 55.275
333188 91.863 -70.623 55.264
333189 91.894 -70.633 55.254
333190 91.925 -70.644 55.244
333191 91.956 -70.654 55.235
333192 91.988 -70.665 55.227
333193 92.019 -70.676 55.220
333194 92.066 -70.698 55.211
333195 92.119 -70.725 55.205
333196 92.169 -70.701 55.201
333197 92.220 -70.676 55.200
333198 92.248 -70.695 55.204
333199 92.281 -70.707 55.201
333200 92.324 -70.707 55.192
333201 92.366 -70.706 55.176
333202 92.407 -70.705 55.156
333203 92.448 -70.704 55.131
333204 92.489 -70.704 55.104
333205 92.529 -70.703 55.076
333206 92.571 -70.701 55.048
333207 92.612 -70.701 55.021
333208 92.652 -70.700 54.994
333209 92.694 -70.699 54.965
333210 92.731 -70.692 54.930
333211 92.767 -70.682 54.894
333212 92.803 -70.671 54.857
333213 92.835 -70.659 54.818
333214 92.867 -70.647 54.778
333215 92.896 -70.639 54.736
333216 92.925 -70.632 54.693
333217 92.954 -70.624 54.651
333218 92.983 -70.616 54.609
333219 93.012 -70.609 54.565
333220 93.041 -70.601 54.522
333221 93.069 -70.594 54.479
333222 93.098 -70.586 54.436
333223 93.127 -70.579 54.395
333224 93.156 -70.571 54.357
333225 93.185 -70.563 54.320
333226 93.214 -70.555 54.285
333227 93.243 -70.547 54.252
333228 93.272 -70.540 54.220
333229 93.300 -70.532 54.189
333282 94.849 -70.163 52.813
333283 94.883 -70.164 52.807
333284 94.915 -70.165 52.800
333285 94.949 -70.165 52.793
333286 94.982 -70.166 52.787
333287 95.015 -70.167 52.782
333288 95.048 -70.168 52.777
333289 95.080 -70.168 52.772
333290 95.114 -70.169 52.768
333291 95.146 -70.170 52.764
333292 95.180 -70.171 52.761
333293 95.212 -70.171 52.760
333294 95.261 -70.157 52.759
333295 95.298 -70.159 52.757
333296 95.332 -70.173 52.754
333297 95.367 -70.187 52.750
333298 95.402 -70.201 52.746
333299 95.437 -70.215 52.740
333300 95.472 -70.228 52.729
333301 95.507 -70.242 52.715
333302 95.542 -70.256 52.698
333303 95.577 -70.269 52.680
333304 95.612 -70.284 52.663
333305 95.647 -70.297 52.641
333306 95.682 -70.311 52.617
333307 95.716 -70.325 52.594
333308 95.751 -70.338 52.571
333309 95.787 -70.352 52.550
333310 95.821 -70.366 52.529
333311 95.851 -70.378 52.510
333312 95.877 -70.387 52.490
333313 95.900 -70.394 52.471
333314 95.923 -70.403 52.452
333315 95.947 -70.411 52.433
333316 95.970 -70.419 52.414
333317 95.993 -70.427 52.394
333318 96.016 -70.435 52.374
333319 96.039 -70.444 52.356
333320 96.062 -70.452 52.339
333321 96.086 -70.460 52.323
333322 96.109 -70.468 52.308
333323 96.132 -70.476 52.294
333324 96.155 -70.485 52.279
333325 96.180 -70.493 52.264
333326 96.216 -70.505 52.248
333327 96.252 -70.518 52.231
333328 96.288 -70.531 52.215
333329 96.327 -70.544 52.201
333330 96.368 -70.557 52.188
333331 96.407 -70.571 52.172
333332 96.446 -70.582 52.154
333333 96.481 -70.584 52.135
333334 96.517 -70.586 52.117
333335 96.560 -70.595 52.100
333336 96.609 -70.610 52.083
333337 96.658 -70.625 52.068
333338 96.708 -70.640 52.052
333339 96.757 -70.654 52.041
333340 96.806 -70.670 52.033
333341 96.853 -70.681 52.027
333342 96.897 -70.690 52.024
333343 96.942 -70.699 52.022
333344 96.986 -70.708 52.019
333345 97.030 -70.717 52.013
333346 97.075 -70.726 52.008
333347 97.119 -70.735 52.005
333348 97.163 -70.744 52.001
333349 97.208 -70.753 51.993
333350 97.250 -70.761 51.983
333351 97.265 -70.765 51.963
333352 97.279 -70.766 51.936
333353 97.293 -70.767 51.906
333354 97.314 -70.775 51.876
333355 97.336 -70.784 51.848
333356 97.359 -70.793 51.825
333357 97.381 -70.801 51.804
333358 97.410 -70.807 51.784
333359 97.442 -70.811 51.765
333360 97.474 -70.815 51.742
333361 97.504 -70.826 51.717
333362 97.533 -70.839 51.689
333363 97.563 -70.852 51.656
333364 97.592 -70.864 51.616
333365 97.621 -70.877 51.577
333366 97.651 -70.890 51.538
333367 97.683 -70.901 51.499
333368 97.716 -70.911 51.462
333369 97.749 -70.921 51.424
333370 97.783 -70.931 51.384
333371 97.815 -70.941 51.341
333372 97.848 -70.946 51.300
333373 97.878 -70.939 51.264
333374 97.909 -70.932 51.229
333375 97.939 -70.925 51.193
333376 97.971 -70.920 51.155
333377 98.005 -70.916 51.112
333378 98.041 -70.923 51.063
333379 98.077 -70.930 51.012
333380 98.113 -70.936 50.959
333381 98.150 -70.943 50.904
333382 98.185 -70.950 50.848
333383 98.222 -70.957 50.790
333384 98.258 -70.964 50.731
333385 98.294 -70.970 50.672
333386 98.331 -70.977 50.616
333387 98.367 -70.984 50.563
333388 98.403 -70.991 50.512
333389 98.439 -70.998 50.461
333390 98.475 -71.005 50.404
333391 98.511 -71.011 50.347
333392 98.548 -71.018 50.287
333393 98.587 -71.021 50.226
333394 98.627 -71.024 50.160
333395 98.667 -71.027 50.094
333396 98.707 -71.030 50.027
333397 98.746 -71.033 49.962
333398 98.786 -71.036 49.897
333399 98.825 -71.040 49.833
333400 98.854 -71.051 49.769
333401 98.884 -71.062 49.707
333402 98.914 -71.074 49.646
333403 98.944 -71.085 49.585
333404 98.974 -71.096 49.524
333405 99.004 -71.107 49.465
333406 99.034 -71.118 49.406
333407 99.064 -71.130 49.349
333408 99.096 -71.144 49.294
333409 99.129 -71.158 49.240
333410 99.161 -71.171 49.189
333411 99.193 -71.186 49.138
333412 99.225 -71.199 49.086
333413 99.258 -71.214 49.034
333414 99.290 -71.227 48.982
333415 99.317 -71.244 48.930
333416 99.355 -71.256 48.881
333417 99.401 -71.265 48.833
333418 99.446 -71.274 48.786
333419 99.491 -71.282 48.738
333420 99.535 -71.290 48.691
333421 99.575 -71.297 48.647
333422 99.615 -71.304 48.603
333423 99.656 -71.312 48.561
333424 99.696 -71.318 48.519
333425 99.738 -71.329 48.478
333426 99.784 -71.346 48.437
333427 99.823 -71.362 48.397
333428 99.862 -71.377 48.358
333429 99.901 -71.392 48.322
333430 99.939 -71.407 48.285
333431 99.978 -71.422 48.247
333432 100.016 -71.437 48.208
333433 100.053 -71.450 48.170
333434 100.088 -71.462 48.130
333435 100.125 -71.475 48.090
333436 100.161 -71.487 48.052
333437 100.197 -71.499 48.011
333438 100.233 -71.511 47.968
333439 100.269 -71.523 47.924
333440 100.305 -71.536 47.880
333441 100.341 -71.548 47.835
333442 100.377 -71.560 47.793
333443 100.409 -71.571 47.752
333444 100.437 -71.579 47.712
333445 100.464 -71.588 47.671
333446 100.488 -71.602 47.631
333447 100.511 -71.617 47.590
333448 100.534 -71.632 47.550
333449 100.557 -71.647 47.511
333450 100.580 -71.662 47.472
333451 100.604 -71.677 47.434
333452 100.627 -71.692 47.396
333453 100.650 -71.707 47.357
333454 100.681 -71.732 47.320
333455 100.715 -71.744 47.284
333456 100.749 -71.748 47.249
333457 100.783 -71.753 47.213
333458 100.817 -71.757 47.178
333459 100.850 -71.762 47.145
333460 100.897 -71.768 47.112
333461 100.950 -71.776 47.078
333462 101.004 -71.782 47.045
333463 101.057 -71.790 47.014
333464 101.111 -71.798 46.983
333465 101.151 -71.807 46.954
333466 101.168 -71.822 46.924
333467 101.184 -71.836 46.893
333468 101.201 -71.851 46.860
333469 101.236 -71.857 46.827
333470 101.272 -71.864 46.792
333471 101.308 -71.870 46.756
333472 101.344 -71.876 46.719
333473 101.380 -71.882 46.682
333474 101.416 -71.892 46.646
333475 101.452 -71.904 46.610
333476 101.487 -71.917 46.573
333477 101.523 -71.929 46.535
333478 101.559 -71.942 46.497
333479 101.595 -71.954 46.459
333480 101.630 -71.967 46.419
333481 101.666 -71.979 46.379
333482 101.702 -71.992 46.340
333483 101.744 -72.009 46.300
333484 101.784 -72.026 46.261
333485 101.814 -72.047 46.223
333486 101.842 -72.067 46.185
333487 101.871 -72.087 46.148
333488 101.900 -72.107 46.111
333489 101.929 -72.127 46.074
333490 101.958 -72.147 46.036
333491 101.986 -72.168 45.999
333492 102.016 -72.187 45.961
333493 102.043 -72.205 45.924
333494 102.050 -72.188 45.886
333495 102.073 -72.191 45.847
333496 102.088 -72.190 45.808
333497 102.090 -72.183 45.769
333498 102.118 -72.183 45.731
333499 102.148 -72.185 45.695
333500 102.178 -72.186 45.660
333501 102.208 -72.188 45.625
333502 102.238 -72.189 45.591
333503 102.268 -72.191 45.557
333504 102.298 -72.193 45.524
333505 102.328 -72.195 45.492
333506 102.358 -72.196 45.459
333507 102.388 -72.198 45.427
333508 102.418 -72.199 45.394
333509 102.448 -72.201 45.363
333510 102.478 -72.202 45.332
333511 102.508 -72.204 45.300
333888 91.819 -70.580 55.272
333889 91.851 -70.591 55.261
333890 91.882 -70.601 55.251
333891 91.913 -70.612 55.242
333892 91.944 -70.623 55.233
333893 91.976 -70.632 55.226
333894 92.007 -70.643 55.218
333895 92.060 -70.669 55.211
333896 92.111 -70.695 55.205
333897 92.162 -70.671 55.200
333898 92.213 -70.647 55.199
333899 92.249 -70.651 55.203
333900 92.278 -70.666 55.202
333901 92.315 -70.664 55.194
333902 92.359 -70.663 55.180
333903 92.402 -70.663 55.162
333904 92.446 -70.663 55.138
333905 92.490 -70.663 55.111
333906 92.533 -70.663 55.083
333907 92.577 -70.662 55.056
333908 92.618 -70.661 55.029
333909 92.659 -70.660 55.001
333910 92.700 -70.659 54.971
333911 92.741 -70.658 54.938
333912 92.777 -70.648 54.904
333913 92.806 -70.635 54.868
333914 92.834 -70.621 54.830
333915 92.873 -70.604 54.792
333916 92.904 -70.594 54.751
333917 92.932 -70.586 54.709
333918 92.961 -70.579 54.667
333919 92.990 -70.571 54.625
333920 93.019 -70.563 54.580
333921 93.048 -70.556 54.536
333922 93.077 -70.548 54.493
333923 93.106 -70.541 54.449
333924 93.135 -70.532 54.407
333925 93.163 -70.525 54.366
333926 93.193 -70.517 54.328
333927 93.221 -70.510 54.292
333928 93.250 -70.502 54.256
333929 93.279 -70.494 54.222
333930 93.308 -70.487 54.190
333931 93.337 -70.479 54.159
333982 94.837 -70.143 52.824
333983 94.870 -70.144 52.817
333984 94.903 -70.145 52.811
333985 94.936 -70.146 52.804
333986 94.969 -70.146 52.799
333987 95.002 -70.146 52.794
333988 95.035 -70.147 52.790
333989 95.068 -70.148 52.786
333990 95.101 -70.149 52.783
333991 95.134 -70.149 52.779
333992 95.167 -70.150 52.777
333993 95.200 -70.151 52.775
333994 95.237 -70.148 52.773
333995 95.287 -70.131 52.772
333996 95.324 -70.136 52.770
333997 95.359 -70.149 52.768
333998 95.394 -70.163 52.765
333999 95.429 -70.178 52.762
334000 95.464 -70.191 52.758
334001 95.499 -70.205 52.747
334002 95.534 -70.218 52.733
334003 95.569 -70.232 52.715
334004 95.604 -70.246 52.696
334005 95.638 -70.259 52.676
334006 95.674 -70.274 52.653
334007 95.705 -70.285 52.630
334008 95.737 -70.297 52.606
334009 95.768 -70.309 52.584
334010 95.800 -70.321 52.563
334011 95.831 -70.333 52.542
334012 95.853 -70.347 52.523
334013 95.879 -70.356 52.505
334014 95.908 -70.364 52.486
334015 95.936 -70.372 52.467
334016 95.959 -70.380 52.448
334017 95.982 -70.388 52.428
334018 96.005 -70.396 52.407
334019 96.028 -70.404 52.386
334020 96.051 -70.413 52.367
334021 96.075 -70.420 52.350
334022 96.098 -70.428 52.333
334023 96.121 -70.437 52.319
334024 96.144 -70.445 52.303
334025 96.167 -70.453 52.287
334026 96.193 -70.462 52.271
334027 96.229 -70.475 52.255
334028 96.265 -70.487 52.239
334029 96.301 -70.500 52.223
334030 96.337 -70.513 52.210
334031 96.375 -70.525 52.199
334032 96.415 -70.538 52.185
334033 96.454 -70.548 52.168
334034 96.489 -70.550 52.150
334035 96.524 -70.553 52.131
334036 96.559 -70.555 52.112
334037 96.606 -70.567 52.095
334038 96.655 -70.582 52.080
334039 96.704 -70.597 52.065
334040 96.756 -70.613 52.056
334041 96.811 -70.630 52.049
334042 96.859 -70.642 52.044
334043 96.901 -70.651 52.044
334044 96.945 -70.660 52.045
334045 96.990 -70.669 52.043
334046 97.034 -70.678 52.037
334047 97.078 -70.687 52.030
334048 97.122 -70.696 52.023
334049 97.167 -70.705 52.014
334050 97.209 -70.713 52.004
334051 97.248 -70.722 51.991
334052 97.266 -70.730 51.969
334053 97.282 -70.734 51.943
334054 97.297 -70.737 51.914
334055 97.320 -70.746 51.883
334056 97.342 -70.754 51.854
334057 97.365 -70.764 51.832
334058 97.387 -70.773 51.812
334059 97.409 -70.782 51.791
334060 97.441 -70.786 51.771
334061 97.473 -70.790 51.747
334062 97.505 -70.794 51.720
334063 97.536 -70.801 51.689
334064 97.566 -70.814 51.656
334065 97.595 -70.826 51.616
334066 97.625 -70.839 51.577
334067 97.655 -70.851 51.539
334068 97.688 -70.861 51.503
334069 97.721 -70.871 51.465
334070 97.754 -70.881 51.427
334071 97.787 -70.892 51.385
334072 97.820 -70.901 51.343
334073 97.853 -70.908 51.305
334074 97.883 -70.901 51.269
334075 97.913 -70.894 51.232
334076 97.945 -70.888 51.194
334077 97.979 -70.883 51.154
334078 98.012 -70.878 51.108
334079 98.046 -70.877 51.058
334080 98.083 -70.884 51.006
334081 98.119 -70.891 50.953
334082 98.155 -70.898 50.899
334083 98.191 -70.905 50.844
334084 98.227 -70.911 50.788
334085 98.264 -70.918 50.730
334086 98.300 -70.925 50.673
334087 98.336 -70.932 50.618
334088 98.372 -70.939 50.565
334089 98.408 -70.945 50.513
334090 98.444 -70.952 50.460
334091 98.481 -70.959 50.402
334092 98.517 -70.966 50.343
334093 98.554 -70.971 50.284
334094 98.594 -70.974 50.221
334095 98.634 -70.977 50.157
334096 98.673 -70.980 50.092
334097 98.714 -70.983 50.026
334098 98.754 -70.986 49.961
334099 98.793 -70.989 49.894
334100 98.833 -70.992 49.829
334101 98.864 -71.002 49.766
334102 98.894 -71.013 49.704
334103 98.923 -71.024 49.643
334104 98.953 -71.036 49.582
334105 98.983 -71.047 49.522
334106 99.013 -71.058 49.463
334107 99.044 -71.070 49.405
334108 99.076 -71.083 49.348
334109 99.108 -71.098 49.293
334110 99.140 -71.111 49.240
334111 99.173 -71.125 49.189
334112 99.205 -71.140 49.139
334113 99.237 -71.153 49.087
334114 99.270 -71.168 49.036
334115 99.302 -71.181 48.984
334116 99.334 -71.196 48.933
334117 99.368 -71.209 48.884
334118 99.410 -71.216 48.836
334119 99.450 -71.223 48.790
334120 99.491 -71.231 48.744
334121 99.531 -71.237 48.698
334122 99.572 -71.245 48.654
334123 99.612 -71.252 48.611
334124 99.652 -71.259 48.570
334125 99.693 -71.266 48.528
334126 99.739 -71.282 48.487
334127 99.786 -71.300 48.446
334128 99.825 -71.315 48.407
334129 99.863 -71.331 48.368
334130 99.902 -71.346 48.332
334131 99.940 -71.361 48.297
334132 99.979 -71.376 48.259
334133 100.018 -71.391 48.221
334134 100.057 -71.406 48.182
334135 100.095 -71.422 48.141
334136 100.134 -71.437 48.100
334137 100.171 -71.450 48.061
334138 100.207 -71.462 48.018
334139 100.243 -71.475 47.975
334140 100.279 -71.487 47.932
334141 100.315 -71.500 47.888
334142 100.351 -71.512 47.844
334143 100.387 -71.524 47.801
334144 100.421 -71.535 47.759
334145 100.449 -71.544 47.718
334146 100.474 -71.555 47.678
334147 100.497 -71.570 47.637
334148 100.521 -71.585 47.596
334149 100.544 -71.600 47.556
334150 100.567 -71.615 47.517
334151 100.590 -71.630 47.479
334152 100.613 -71.645 47.441
334153 100.636 -71.660 47.404
334154 100.660 -71.675 47.366
334155 100.689 -71.697 47.329
334156 100.723 -71.707 47.293
334157 100.757 -71.712 47.257
334158 100.790 -71.716 47.222
334159 100.824 -71.721 47.187
334160 100.878 -71.729 47.152
334161 100.931 -71.735 47.118
334162 100.985 -71.743 47.084
334163 101.039 -71.750 47.051
334164 101.093 -71.757 47.020
334165 101.146 -71.765 46.989
334166 101.164 -71.779 46.958
334167 101.180 -71.794 46.928
334168 101.196 -71.809 46.896
334169 101.214 -71.823 46.864
334170 101.251 -71.829 46.831
334171 101.286 -71.835 46.796
334172 101.323 -71.841 46.761
334173 101.358 -71.847 46.724
334174 101.395 -71.852 46.689
334175 101.431 -71.859 46.655
334176 101.466 -71.871 46.619
334177 101.502 -71.884 46.584
334178 101.538 -71.896 46.547
334179 101.573 -71.908 46.509
334180 101.609 -71.921 46.471
334181 101.645 -71.933 46.431
334182 101.680 -71.946 46.391
334183 101.719 -71.961 46.351
334184 101.761 -71.978 46.312
334185 101.803 -71.995 46.272
334186 101.832 -72.015 46.233
334187 101.861 -72.036 46.195
334188 101.890 -72.055 46.157
334189 101.918 -72.076 46.119
334190 101.948 -72.095 46.081
334191 101.976 -72.116 46.042
334192 102.005 -72.136 46.004
334193 102.033 -72.155 45.963
334194 102.042 -72.139 45.925
334195 102.065 -72.143 45.887
334196 102.074 -72.139 45.847
334197 102.076 -72.131 45.810
334198 102.102 -72.131 45.773
334199 102.132 -72.133 45.737
334200 102.162 -72.135 45.702
334201 102.192 -72.136 45.667
334202 102.222 -72.138 45.633
334203 102.252 -72.139 45.600
334204 102.282 -72.141 45.566
334205 102.312 -72.142 45.533
334206 102.342 -72.144 45.501
334207 102.372 -72.146 45.469
334208 102.402 -72.147 45.436
334209 102.432 -72.149 45.404
334210 102.462 -72.150 45.372
334211 102.492 -72.152 45.340
334212 102.522 -72.153 45.309
334589 91.808 -70.547 55.268
334590 91.839 -70.558 55.258
334591 91.870 -70.569 55.248
334592 91.901 -70.579 55.240
334593 91.932 -70.590 55.232
334594 91.963 -70.601 55.224
334595 91.996 -70.612 55.217
334596 92.054 -70.641 55.210
334597 92.104 -70.666 55.204
334598 92.155 -70.641 55.199
334599 92.206 -70.617 55.196
334600 92.249 -70.607 55.198
334601 92.277 -70.627 55.198
334602 92.314 -70.624 55.192
334603 92.351 -70.622 55.180
334604 92.392 -70.620 55.161
334605 92.436 -70.620 55.139
334606 92.480 -70.619 55.114
334607 92.523 -70.619 55.087
334608 92.567 -70.619 55.063
334609 92.611 -70.619 55.036
334610 92.655 -70.618 55.007
334611 92.699 -70.618 54.977
334612 92.742 -70.617 54.945
334613 92.778 -70.610 54.913
334614 92.805 -70.597 54.879
334615 92.833 -70.584 54.843
334616 92.872 -70.571 54.807
334617 92.911 -70.548 54.769
334618 92.940 -70.541 54.727
334619 92.969 -70.533 54.685
334620 92.998 -70.525 54.642
334621 93.026 -70.518 54.597
334622 93.056 -70.510 54.552
334623 93.084 -70.502 54.506
334624 93.114 -70.494 54.461
334625 93.142 -70.487 54.418
334626 93.171 -70.479 54.376
334627 93.200 -70.472 54.336
334628 93.229 -70.464 54.299
334629 93.258 -70.456 54.262
334630 93.287 -70.449 54.227
334631 93.315 -70.441 54.193
334632 93.344 -70.434 54.161
334682 94.824 -70.123 52.832
334683 94.857 -70.124 52.826
334684 94.890 -70.124 52.819
334685 94.923 -70.125 52.813
334686 94.956 -70.126 52.808
334687 94.989 -70.127 52.804
334688 95.022 -70.127 52.800
334689 95.055 -70.128 52.798
334690 95.088 -70.129 52.796
334691 95.121 -70.129 52.794
334692 95.154 -70.130 52.791
334693 95.187 -70.130 52.790
334694 95.220 -70.131 52.788
334695 95.263 -70.123 52.787
334696 95.314 -70.106 52.786
334697 95.351 -70.112 52.783
334698 95.386 -70.126 52.780
334699 95.421 -70.140 52.778
334700 95.456 -70.153 52.776
334701 95.491 -70.168 52.775
334702 95.526 -70.181 52.765
334703 95.558 -70.193 52.751
334704 95.590 -70.206 52.733
334705 95.621 -70.217 52.712
334706 95.653 -70.229 52.690
334707 95.684 -70.241 52.666
334708 95.716 -70.253 52.642
334709 95.747 -70.265 52.619
334710 95.779 -70.277 52.596
334711 95.810 -70.288 52.576
334712 95.834 -70.303 52.556
334713 95.854 -70.317 52.537
334714 95.880 -70.327 52.521
334715 95.908 -70.334 52.502
334716 95.937 -70.342 52.483
334717 95.965 -70.349 52.463
334718 95.993 -70.356 52.443
334719 96.018 -70.365 52.421
334720 96.041 -70.373 52.399
334721 96.064 -70.381 52.378
334722 96.087 -70.389 52.360
334723 96.110 -70.397 52.345
334724 96.133 -70.406 52.329
334725 96.157 -70.413 52.312
334726 96.180 -70.422 52.295
334727 96.206 -70.431 52.278
334728 96.242 -70.444 52.263
334729 96.278 -70.456 52.247
334730 96.315 -70.469 52.232
334731 96.350 -70.481 52.221
334732 96.387 -70.494 52.211
334733 96.424 -70.507 52.198
334734 96.461 -70.516 52.182
334735 96.497 -70.517 52.164
334736 96.532 -70.519 52.146
334737 96.567 -70.522 52.127
334738 96.603 -70.525 52.110
334739 96.654 -70.541 52.095
334740 96.709 -70.557 52.082
334741 96.763 -70.575 52.072
334742 96.818 -70.592 52.067
334743 96.867 -70.605 52.066
334744 96.906 -70.613 52.068
334745 96.949 -70.621 52.071
334746 96.993 -70.630 52.069
334747 97.037 -70.639 52.063
334748 97.082 -70.648 52.052
334749 97.126 -70.657 52.041
334750 97.167 -70.666 52.027
334751 97.207 -70.673 52.013
334752 97.245 -70.682 51.997
334753 97.264 -70.691 51.973
334754 97.283 -70.699 51.946
334755 97.303 -70.707 51.918
334756 97.326 -70.717 51.887
334757 97.349 -70.726 51.860
334758 97.371 -70.735 51.838
334759 97.393 -70.744 51.817
334760 97.416 -70.753 51.795
334761 97.440 -70.760 51.773
334762 97.472 -70.764 51.748
334763 97.504 -70.768 51.720
334764 97.536 -70.773 51.688
334765 97.568 -70.776 51.654
334766 97.598 -70.789 51.616
334767 97.628 -70.801 51.577
334768 97.660 -70.811 51.539
334769 97.693 -70.822 51.502
334770 97.726 -70.832 51.463
334771 97.759 -70.842 51.424
334772 97.792 -70.852 51.383
334773 97.825 -70.862 51.343
334774 97.858 -70.870 51.308
334775 97.888 -70.863 51.271
334776 97.920 -70.857 51.233
334777 97.953 -70.851 51.192
334778 97.986 -70.847 51.150
334779 98.019 -70.842 51.101
334780 98.053 -70.837 51.050
334781 98.088 -70.839 50.998
334782 98.124 -70.845 50.946
334783 98.160 -70.852 50.893
334784 98.197 -70.859 50.839
334785 98.232 -70.866 50.784
334786 98.269 -70.873 50.727
334787 98.305 -70.880 50.671
334788 98.341 -70.887 50.617
334789 98.377 -70.894 50.563
334790 98.414 -70.901 50.509
334791 98.449 -70.908 50.454
334792 98.486 -70.914 50.394
334793 98.522 -70.921 50.335
334794 98.561 -70.924 50.275
334795 98.601 -70.927 50.214
334796 98.641 -70.930 50.151
334797 98.680 -70.933 50.087
334798 98.720 -70.936 50.023
334799 98.760 -70.939 49.957
334800 98.800 -70.942 49.891
334801 98.840 -70.944 49.826
334802 98.873 -70.953 49.762
334803 98.903 -70.964 49.700
334804 98.933 -70.975 49.638
334805 98.962 -70.986 49.578
334806 98.993 -70.998 49.519
334807 99.023 -71.009 49.460
334808 99.055 -71.024 49.402
334809 99.088 -71.037 49.347
334810 99.120 -71.052 49.293
334811 99.152 -71.065 49.240
334812 99.185 -71.080 49.190
334813 99.217 -71.093 49.139
334814 99.249 -71.107 49.088
334815 99.281 -71.121 49.037
334816 99.313 -71.137 48.986
334817 99.345 -71.154 48.936
334818 99.377 -71.169 48.887
334819 99.413 -71.173 48.840
334820 99.452 -71.179 48.794
334821 99.492 -71.186 48.749
334822 99.532 -71.192 48.705
334823 99.572 -71.199 48.661
334824 99.613 -71.205 48.619
334825 99.653 -71.212 48.578
334826 99.693 -71.218 48.538
334827 99.739 -71.235 48.497
334828 99.786 -71.253 48.457
334829 99.826 -71.268 48.418
334830 99.864 -71.284 48.379
334831 99.903 -71.299 48.343
334832 99.941 -71.314 48.307
334833 99.980 -71.330 48.269
334834 100.019 -71.345 48.230
334835 100.058 -71.360 48.191
334836 100.097 -71.375 48.150
334837 100.135 -71.390 48.108
334838 100.174 -71.406 48.069
334839 100.212 -71.421 48.026
334840 100.251 -71.436 47.983
334841 100.289 -71.450 47.939
334842 100.325 -71.462 47.896
334843 100.361 -71.475 47.852
334844 100.398 -71.487 47.810
334845 100.433 -71.500 47.767
334846 100.460 -71.508 47.725
334847 100.483 -71.523 47.685
334848 100.506 -71.538 47.644
334849 100.530 -71.553 47.603
334850 100.553 -71.569 47.563
334851 100.576 -71.583 47.525
334852 100.600 -71.598 47.488
334853 100.623 -71.613 47.450
334854 100.646 -71.629 47.413
334855 100.669 -71.644 47.376
334856 100.696 -71.664 47.339
334857 100.731 -71.671 47.302
334858 100.764 -71.675 47.267
334859 100.805 -71.681 47.231
334860 100.859 -71.688 47.196
334861 100.913 -71.696 47.162
334862 100.967 -71.703 47.127
334863 101.020 -71.710 47.092
334864 101.074 -71.718 47.059
334865 101.128 -71.725 47.027
334866 101.159 -71.737 46.994
334867 101.176 -71.751 46.961
334868 101.192 -71.766 46.929
334869 101.208 -71.781 46.896
334870 101.229 -71.794 46.863
334871 101.265 -71.800 46.830
334872 101.301 -71.806 46.796
334873 101.337 -71.812 46.761
334874 101.373 -71.817 46.726
334875 101.409 -71.823 46.692
334876 101.445 -71.829 46.660
334877 101.481 -71.838 46.626
334878 101.516 -71.851 46.593
334879 101.552 -71.863 46.558
334880 101.588 -71.876 46.520
334881 101.624 -71.888 46.482
334882 101.659 -71.901 46.442
334883 101.695 -71.913 46.402
334884 101.736 -71.929 46.363
334885 101.778 -71.947 46.322
334886 101.820 -71.964 46.282
334887 101.850 -71.984 46.241
334888 101.880 -72.004 46.201
334889 101.908 -72.024 46.163
334890 101.937 -72.044 46.123
334891 101.966 -72.064 46.083
334892 101.995 -72.084 46.043
334893 102.024 -72.105 46.003
334894 102.034 -72.091 45.964
334895 102.057 -72.095 45.927
334896 102.060 -72.087 45.889
334897 102.062 -72.080 45.851
334898 102.086 -72.080 45.815
334899 102.116 -72.081 45.779
334900 102.146 -72.083 45.744
334901 102.176 -72.084 45.710
334902 102.206 -72.086 45.676
334903 102.236 -72.087 45.642
334904 102.266 -72.089 45.608
334905 102.296 -72.090 45.575
334906 102.326 -72.092 45.543
334907 102.356 -72.093 45.510
334908 102.386 -72.095 45.478
334909 102.416 -72.096 45.446
334910 102.446 -72.098 45.413
334911 102.476 -72.100 45.381
334912 102.506 -72.102 45.350
335291 91.827 -70.526 55.254
335292 91.858 -70.537 55.245
335293 91.889 -70.547 55.237
335294 91.920 -70.557 55.229
335295 91.951 -70.568 55.221
335296 91.989 -70.583 55.214
335297 92.047 -70.616 55.207
335298 92.097 -70.636 55.201
335299 92.148 -70.612 55.196
335300 92.199 -70.588 55.192
335301 92.250 -70.563 55.192
335302 92.276 -70.587 55.191
335303 92.313 -70.584 55.187
335304 92.350 -70.582 55.176
335305 92.387 -70.579 55.158
335306 92.426 -70.577 55.136
335307 92.470 -70.576 55.112
335308 92.513 -70.576 55.089
335309 92.557 -70.576 55.065
335310 92.601 -70.576 55.038
335311 92.645 -70.575 55.011
335312 92.689 -70.575 54.982
335313 92.732 -70.575 54.954
335314 92.777 -70.573 54.923
335315 92.804 -70.560 54.891
335316 92.832 -70.546 54.859
335317 92.865 -70.533 54.825
335318 92.914 -70.515 54.788
335319 92.947 -70.495 54.748
335320 92.977 -70.488 54.705
335321 93.005 -70.480 54.662
335322 93.034 -70.472 54.615
335323 93.063 -70.464 54.567
335324 93.092 -70.456 54.520
335325 93.121 -70.449 54.474
335326 93.150 -70.441 54.429
335327 93.179 -70.434 54.386
335328 93.208 -70.426 54.345
335329 93.236 -70.419 54.306
335330 93.265 -70.411 54.268
335331 93.294 -70.403 54.231
335332 93.323 -70.396 54.196
335333 93.352 -70.388 54.163
335383 94.845 -70.103 52.832
335384 94.878 -70.104 52.827
335385 94.911 -70.105 52.822
335386 94.944 -70.105 52.816
335387 94.977 -70.106 52.812
335388 95.010 -70.107 52.808
335389 95.043 -70.108 52.806
335390 95.075 -70.109 52.806
335391 95.109 -70.109 52.805
335392 95.141 -70.110 52.804
335393 95.175 -70.111 52.802
335394 95.207 -70.112 52.801
335395 95.241 -70.112 52.800
335396 95.290 -70.097 52.799
335397 95.341 -70.080 52.797
335398 95.378 -70.089 52.794
335399 95.411 -70.102 52.791
335400 95.443 -70.113 52.789
335401 95.474 -70.125 52.789
335402 95.506 -70.137 52.789
335403 95.537 -70.149 52.783
335404 95.568 -70.161 52.769
335405 95.600 -70.173 52.751
335406 95.632 -70.184 52.730
335407 95.663 -70.196 52.704
335408 95.695 -70.209 52.680
335409 95.726 -70.220 52.655
335410 95.758 -70.232 52.631
335411 95.790 -70.244 52.609
335412 95.815 -70.258 52.589
335413 95.835 -70.272 52.569
335414 95.855 -70.287 52.552
335415 95.881 -70.297 52.536
335416 95.909 -70.305 52.518
335417 95.938 -70.312 52.499
335418 95.966 -70.319 52.480
335419 95.994 -70.327 52.461
335420 96.022 -70.334 52.436
335421 96.051 -70.342 52.412
335422 96.076 -70.350 52.391
335423 96.099 -70.358 52.371
335424 96.122 -70.366 52.356
335425 96.146 -70.374 52.338
335426 96.169 -70.382 52.321
335427 96.192 -70.391 52.304
335428 96.220 -70.400 52.286
335429 96.256 -70.413 52.271
335430 96.292 -70.425 52.255
335431 96.328 -70.438 52.241
335432 96.364 -70.450 52.231
335433 96.400 -70.463 52.223
335434 96.436 -70.474 52.211
335435 96.471 -70.481 52.196
335436 96.507 -70.483 52.179
335437 96.542 -70.485 52.160
335438 96.578 -70.487 52.143
335439 96.616 -70.492 52.126
335440 96.661 -70.503 52.111
335441 96.716 -70.519 52.099
335442 96.770 -70.536 52.090
335443 96.825 -70.554 52.087
335444 96.875 -70.567 52.089
335445 96.914 -70.576 52.093
335446 96.954 -70.583 52.096
335447 96.997 -70.591 52.094
335448 97.041 -70.601 52.087
335449 97.085 -70.610 52.074
335450 97.126 -70.618 52.058
335451 97.166 -70.626 52.041
335452 97.206 -70.634 52.023
335453 97.243 -70.642 52.002
335454 97.262 -70.651 51.976
335455 97.283 -70.662 51.949
335456 97.307 -70.681 51.920
335457 97.328 -70.691 51.891
335458 97.352 -70.700 51.865
335459 97.375 -70.707 51.844
335460 97.398 -70.716 51.822
335461 97.421 -70.724 51.799
335462 97.444 -70.732 51.775
335463 97.471 -70.739 51.748
335464 97.503 -70.743 51.719
335465 97.535 -70.747 51.686
335466 97.567 -70.751 51.652
335467 97.599 -70.755 51.614
335468 97.632 -70.762 51.575
335469 97.664 -70.772 51.537
335470 97.698 -70.782 51.498
335471 97.731 -70.792 51.458
335472 97.764 -70.802 51.418
335473 97.797 -70.813 51.379
335474 97.830 -70.823 51.342
335475 97.863 -70.832 51.307
335476 97.894 -70.825 51.270
335477 97.927 -70.820 51.230
335478 97.960 -70.815 51.187
335479 97.994 -70.811 51.143
335480 98.027 -70.805 51.093
335481 98.060 -70.801 51.041
335482 98.094 -70.795 50.990
335483 98.129 -70.801 50.938
335484 98.165 -70.807 50.887
335485 98.202 -70.814 50.834
335486 98.238 -70.821 50.781
335487 98.274 -70.828 50.726
335488 98.310 -70.835 50.671
335489 98.347 -70.842 50.616
335490 98.382 -70.848 50.561
335491 98.419 -70.855 50.505
335492 98.455 -70.862 50.447
335493 98.491 -70.869 50.386
335494 98.528 -70.874 50.326
335495 98.568 -70.877 50.266
335496 98.608 -70.880 50.205
335497 98.648 -70.883 50.143
335498 98.688 -70.886 50.080
335499 98.727 -70.889 50.017
335500 98.767 -70.892 49.952
335501 98.807 -70.894 49.886
335502 98.847 -70.897 49.821
335503 98.882 -70.904 49.757
335504 98.912 -70.915 49.695
335505 98.942 -70.926 49.633
335506 98.972 -70.937 49.573
335507 99.003 -70.949 49.515
335508 99.035 -70.964 49.457
335509 99.067 -70.977 49.400
335510 99.099 -70.991 49.346
335511 99.132 -71.005 49.292
335512 99.164 -71.019 49.240
335513 99.196 -71.033 49.190
335514 99.228 -71.050 49.140
335515 99.260 -71.067 49.089
335516 99.292 -71.083 49.039
335517 99.325 -71.099 48.989
335518 99.356 -71.116 48.938
335519 99.389 -71.130 48.891
335520 99.423 -71.136 48.845
335521 99.458 -71.139 48.799
335522 99.495 -71.143 48.755
335523 99.535 -71.150 48.712
335524 99.575 -71.156 48.669
335525 99.616 -71.163 48.627
335526 99.656 -71.170 48.587
335527 99.696 -71.177 48.547
335528 99.739 -71.188 48.508
335529 99.786 -71.206 48.469
335530 99.827 -71.222 48.430
335531 99.865 -71.237 48.392
335532 99.904 -71.252 48.354
335533 99.943 -71.268 48.316
335534 99.982 -71.283 48.279
335535 100.020 -71.298 48.239
335536 100.059 -71.313 48.199
335537 100.098 -71.328 48.157
335538 100.136 -71.343 48.115
335539 100.175 -71.359 48.076
335540 100.214 -71.375 48.034
335541 100.252 -71.390 47.992
335542 100.291 -71.405 47.948
335543 100.330 -71.420 47.903
335544 100.368 -71.435 47.860
335545 100.407 -71.450 47.817
335546 100.443 -71.463 47.776
335547 100.469 -71.476 47.735
335548 100.492 -71.491 47.695
335549 100.516 -71.506 47.654
335550 100.539 -71.522 47.613
335551 100.562 -71.537 47.573
335552 100.585 -71.552 47.534
335553 100.609 -71.566 47.498
335554 100.632 -71.582 47.461
335555 100.655 -71.597 47.424
335556 100.679 -71.612 47.387
335557 100.704 -71.630 47.350
335558 100.738 -71.635 47.314
335559 100.787 -71.641 47.279
335560 100.841 -71.649 47.243
335561 100.895 -71.656 47.208
335562 100.948 -71.663 47.173
335563 101.002 -71.671 47.138
335564 101.055 -71.678 47.103
335565 101.109 -71.685 47.070
335566 101.155 -71.694 47.036
335567 101.171 -71.709 47.003
335568 101.188 -71.723 46.969
335569 101.204 -71.738 46.936
335570 101.220 -71.753 46.902
335571 101.244 -71.765 46.868
335572 101.279 -71.771 46.834
335573 101.316 -71.776 46.799
335574 101.351 -71.782 46.764
335575 101.388 -71.788 46.729
335576 101.424 -71.795 46.695
335577 101.460 -71.801 46.662
335578 101.496 -71.807 46.629
335579 101.532 -71.817 46.597
335580 101.567 -71.830 46.564
335581 101.603 -71.842 46.528
335582 101.638 -71.855 46.490
335583 101.674 -71.867 46.450
335584 101.711 -71.881 46.411
335585 101.753 -71.898 46.372
335586 101.796 -71.916 46.331
335587 101.837 -71.933 46.289
335588 101.869 -71.953 46.248
335589 101.898 -71.973 46.208
335590 101.927 -71.992 46.167
335591 101.956 -72.013 46.126
335592 101.984 -72.033 46.085
335593 102.013 -72.053 46.045
335594 102.026 -72.043 46.005
335595 102.044 -72.044 45.967
335596 102.046 -72.036 45.930
335597 102.048 -72.029 45.893
335598 102.069 -72.027 45.857
335599 102.100 -72.029 45.821
335600 102.129 -72.030 45.786
335601 102.160 -72.032 45.751
335602 102.189 -72.033 45.718
335603 102.220 -72.035 45.684
335604 102.250 -72.036 45.650
335605 102.279 -72.039 45.617
335606 102.310 -72.040 45.584
335607 102.339 -72.042 45.553
335608 102.370 -72.043 45.520
335609 102.400 -72.045 45.488
335610 102.430 -72.046 45.456
335611 102.460 -72.048 45.424
335612 102.489 -72.049 45.391
335992 91.815 -70.494 55.249
335993 91.846 -70.504 55.240
335994 91.877 -70.515 55.233
335995 91.908 -70.525 55.225
335996 91.939 -70.536 55.218
335997 91.983 -70.555 55.211
335998 92.040 -70.591 55.204
335999 92.091 -70.607 55.197
336000 92.141 -70.582 55.192
336001 92.192 -70.558 55.187
336002 92.243 -70.534 55.185
336003 92.275 -70.547 55.183
336004 92.311 -70.544 55.179
336005 92.349 -70.541 55.169
336006 92.386 -70.539 55.153
336007 92.423 -70.536 55.133
336008 92.460 -70.534 55.110
336009 92.503 -70.533 55.090
336010 92.547 -70.532 55.068
336011 92.591 -70.532 55.045
336012 92.635 -70.532 55.020
336013 92.678 -70.532 54.994
336014 92.722 -70.532 54.967
336015 92.768 -70.530 54.937
336016 92.803 -70.522 54.907
336017 92.831 -70.509 54.877
336018 92.859 -70.495 54.844
336019 92.916 -70.483 54.808
336020 92.952 -70.456 54.769
336021 92.984 -70.442 54.727
336022 93.013 -70.434 54.682
336023 93.042 -70.426 54.634
336024 93.071 -70.419 54.584
336025 93.099 -70.411 54.535
336026 93.128 -70.403 54.488
336027 93.157 -70.396 54.442
336028 93.186 -70.388 54.398
336029 93.215 -70.381 54.355
336030 93.244 -70.373 54.314
336031 93.273 -70.366 54.275
336032 93.302 -70.358 54.237
336033 93.331 -70.350 54.201
336034 93.359 -70.342 54.166
336083 94.832 -70.083 52.838
336084 94.865 -70.084 52.833
336085 94.898 -70.085 52.828
336086 94.931 -70.086 52.823
336087 94.964 -70.086 52.819
336088 94.997 -70.087 52.815
336089 95.030 -70.087 52.813
336090 95.063 -70.088 52.812
336091 95.096 -70.089 52.813
336092 95.129 -70.090 52.814
336093 95.162 -70.090 52.815
336094 95.195 -70.091 52.814
336095 95.228 -70.092 52.813
336096 95.265 -70.088 52.812
336097 95.317 -70.071 52.811
336098 95.362 -70.058 52.809
336099 95.390 -70.057 52.805
336100 95.421 -70.069 52.802
336101 95.453 -70.080 52.799
336102 95.484 -70.093 52.799
336103 95.516 -70.105 52.802
336104 95.548 -70.116 52.799
336105 95.579 -70.128 52.788
336106 95.611 -70.140 52.770
336107 95.642 -70.152 52.749
336108 95.674 -70.164 52.721
336109 95.705 -70.176 52.696
336110 95.737 -70.187 52.670
336111 95.768 -70.199 52.645
336112 95.796 -70.212 52.623
336113 95.816 -70.228 52.602
336114 95.836 -70.242 52.583
336115 95.856 -70.257 52.565
336116 95.881 -70.267 52.552
336117 95.910 -70.275 52.533
336118 95.938 -70.282 52.514
336119 95.966 -70.290 52.495
336120 95.994 -70.297 52.477
336121 96.023 -70.305 52.451
336122 96.051 -70.312 52.425
336123 96.080 -70.319 52.402
336124 96.108 -70.327 52.381
336125 96.135 -70.334 52.363
336126 96.158 -70.343 52.345
336127 96.181 -70.351 52.328
336128 96.204 -70.359 52.313
336129 96.233 -70.369 52.295
336130 96.269 -70.382 52.278
336131 96.305 -70.394 52.262
336132 96.341 -70.407 52.248
336133 96.377 -70.417 52.240
336134 96.413 -70.427 52.234
336135 96.449 -70.437 52.223
336136 96.484 -70.444 52.208
336137 96.520 -70.447 52.190
336138 96.555 -70.449 52.173
336139 96.591 -70.451 52.156
336140 96.636 -70.462 52.140
336141 96.681 -70.472 52.126
336142 96.726 -70.483 52.115
336143 96.778 -70.498 52.110
336144 96.832 -70.515 52.109
336145 96.883 -70.530 52.111
336146 96.923 -70.538 52.115
336147 96.962 -70.545 52.119
336148 97.002 -70.553 52.116
336149 97.044 -70.562 52.108
336150 97.085 -70.569 52.095
336151 97.124 -70.578 52.075
336152 97.164 -70.586 52.055
336153 97.204 -70.594 52.034
336154 97.241 -70.602 52.010
336155 97.259 -70.611 51.981
336156 97.286 -70.629 51.952
336157 97.313 -70.650 51.923
336158 97.331 -70.665 51.894
336159 97.352 -70.676 51.869
336160 97.375 -70.684 51.848
336161 97.398 -70.691 51.826
336162 97.422 -70.700 51.801
336163 97.445 -70.708 51.777
336164 97.471 -70.715 51.747
336165 97.504 -70.719 51.717
336166 97.535 -70.723 51.684
336167 97.568 -70.727 51.649
336168 97.599 -70.731 51.610
336169 97.627 -70.735 51.570
336170 97.659 -70.742 51.530
336171 97.693 -70.751 51.489
336172 97.728 -70.760 51.449
336173 97.762 -70.769 51.410
336174 97.797 -70.777 51.374
336175 97.831 -70.786 51.339
336176 97.867 -70.792 51.304
336177 97.901 -70.789 51.266
336178 97.935 -70.784 51.224
336179 97.968 -70.779 51.179
336180 98.001 -70.774 51.132
336181 98.034 -70.769 51.082
336182 98.068 -70.764 51.031
336183 98.101 -70.759 50.981
336184 98.135 -70.755 50.931
336185 98.171 -70.762 50.882
336186 98.207 -70.769 50.831
336187 98.243 -70.776 50.779
336188 98.279 -70.782 50.726
336189 98.315 -70.789 50.671
336190 98.352 -70.796 50.615
336191 98.388 -70.803 50.559
336192 98.424 -70.810 50.501
336193 98.460 -70.817 50.440
336194 98.496 -70.823 50.378
336195 98.535 -70.827 50.316
336196 98.575 -70.830 50.255
336197 98.615 -70.833 50.194
336198 98.655 -70.836 50.132
336199 98.695 -70.839 50.070
336200 98.734 -70.842 50.007
336201 98.774 -70.845 49.944
336202 98.814 -70.847 49.879
336203 98.854 -70.850 49.814
336204 98.891 -70.854 49.751
336205 98.921 -70.866 49.689
336206 98.951 -70.877 49.628
336207 98.982 -70.889 49.569
336208 99.014 -70.903 49.511
336209 99.047 -70.917 49.454
336210 99.079 -70.931 49.399
336211 99.111 -70.945 49.345
336212 99.143 -70.962 49.292
336213 99.175 -70.979 49.241
336214 99.207 -70.995 49.191
336215 99.239 -71.012 49.139
336216 99.271 -71.029 49.089
336217 99.303 -71.046 49.040
336218 99.335 -71.062 48.991
336219 99.367 -71.078 48.942
336220 99.400 -71.092 48.895
336221 99.433 -71.099 48.850
336222 99.469 -71.102 48.805
336223 99.503 -71.105 48.762
336224 99.539 -71.108 48.719
336225 99.578 -71.115 48.677
336226 99.619 -71.121 48.637
336227 99.659 -71.127 48.597
336228 99.699 -71.134 48.558
336229 99.739 -71.141 48.521
336230 99.786 -71.159 48.483
336231 99.828 -71.175 48.445
336232 99.866 -71.190 48.406
336233 99.905 -71.205 48.369
336234 99.944 -71.221 48.331
336235 99.983 -71.237 48.293
336236 100.021 -71.252 48.254
336237 100.060 -71.267 48.214
336238 100.099 -71.282 48.173
336239 100.137 -71.297 48.131
336240 100.176 -71.312 48.090
336241 100.215 -71.328 48.047
336242 100.254 -71.343 48.004
336243 100.292 -71.358 47.960
336244 100.331 -71.373 47.915
336245 100.369 -71.388 47.872
336246 100.408 -71.407 47.831
336247 100.444 -71.425 47.790
336248 100.468 -71.441 47.748
336249 100.492 -71.456 47.707
336250 100.516 -71.472 47.666
336251 100.541 -71.487 47.625
336252 100.565 -71.503 47.585
336253 100.590 -71.518 47.547
336254 100.614 -71.533 47.510
336255 100.638 -71.549 47.474
336256 100.662 -71.564 47.437
336257 100.686 -71.579 47.399
336258 100.714 -71.594 47.363
336259 100.769 -71.602 47.327
336260 100.822 -71.609 47.292
336261 100.876 -71.616 47.256
336262 100.929 -71.624 47.221
336263 100.983 -71.631 47.185
336264 101.037 -71.638 47.150
336265 101.091 -71.645 47.115
336266 101.144 -71.653 47.082
336267 101.167 -71.666 47.047
336268 101.183 -71.681 47.013
336269 101.200 -71.696 46.979
336270 101.216 -71.710 46.945
336271 101.233 -71.725 46.910
336272 101.258 -71.736 46.876
336273 101.294 -71.741 46.841
336274 101.330 -71.748 46.805
336275 101.366 -71.754 46.770
336276 101.402 -71.760 46.734
336277 101.438 -71.766 46.699
336278 101.474 -71.772 46.665
336279 101.510 -71.778 46.632
336280 101.546 -71.785 46.600
336281 101.582 -71.797 46.567
336282 101.618 -71.809 46.533
336283 101.653 -71.822 46.494
336284 101.689 -71.834 46.456
336285 101.729 -71.850 46.418
336286 101.770 -71.867 46.378
336287 101.813 -71.885 46.337
336288 101.855 -71.902 46.295
336289 101.888 -71.921 46.254
336290 101.916 -71.942 46.213
336291 101.946 -71.961 46.172
336292 101.974 -71.982 46.130
336293 102.003 -72.001 46.088
336294 102.018 -71.995 46.049
336295 102.030 -71.992 46.010
336296 102.032 -71.985 45.972
336297 102.034 -71.977 45.936
336298 102.053 -71.975 45.900
336299 102.084 -71.977 45.864
336300 102.113 -71.979 45.828
336301 102.144 -71.980 45.794
336302 102.173 -71.982 45.760
336303 102.203 -71.983 45.726
336304 102.234 -71.985 45.693
336305 102.263 -71.986 45.659
336306 102.293 -71.988 45.626
336307 102.323 -71.989 45.594
336308 102.353 -71.991 45.562
336309 102.384 -71.992 45.529
336310 102.413 -71.994 45.497
336311 102.444 -71.995 45.464
336693 91.803 -70.461 55.245
336694 91.834 -70.472 55.236
336695 91.865 -70.482 55.229
336696 91.896 -70.493 55.221
336697 91.927 -70.503 55.214
336698 91.977 -70.527 55.207
336699 92.033 -70.566 55.201
336700 92.084 -70.578 55.193
336701 92.135 -70.554 55.186
336702 92.185 -70.529 55.182
336703 92.236 -70.504 55.178
336704 92.274 -70.506 55.175
336705 92.310 -70.504 55.170
336706 92.348 -70.501 55.161
336707 92.385 -70.499 55.147
336708 92.422 -70.496 55.129
336709 92.459 -70.494 55.108
336710 92.496 -70.491 55.089
336711 92.536 -70.490 55.072
336712 92.580 -70.489 55.054
336713 92.624 -70.489 55.034
336714 92.668 -70.488 55.011
336715 92.713 -70.488 54.984
336716 92.759 -70.487 54.955
336717 92.802 -70.485 54.926
336718 92.830 -70.471 54.897
336719 92.857 -70.458 54.865
336720 92.909 -70.445 54.829
336721 92.954 -70.425 54.791
336722 92.991 -70.397 54.749
336723 93.020 -70.388 54.703
336724 93.049 -70.381 54.653
336725 93.078 -70.373 54.601
336726 93.107 -70.366 54.550
336727 93.136 -70.358 54.501
336728 93.165 -70.350 54.454
336729 93.194 -70.343 54.409
336730 93.222 -70.335 54.365
336731 93.252 -70.328 54.323
336732 93.280 -70.320 54.282
336733 93.309 -70.312 54.244
336734 93.338 -70.304 54.208
336735 93.367 -70.297 54.172
336783 94.819 -70.063 52.844
336784 94.852 -70.064 52.838
336785 94.885 -70.065 52.833
336786 94.918 -70.065 52.827
336787 94.951 -70.066 52.822
336788 94.984 -70.067 52.821
336789 95.018 -70.068 52.819
336790 95.050 -70.068 52.817
336791 95.084 -70.069 52.818
336792 95.116 -70.069 52.821
336793 95.149 -70.070 52.823
336794 95.182 -70.071 52.825
336795 95.215 -70.071 52.827
336796 95.248 -70.072 52.827
336797 95.292 -70.062 52.826
336798 95.343 -70.046 52.824
336799 95.379 -70.037 52.821
336800 95.412 -70.031 52.816
336801 95.432 -70.037 52.811
336802 95.464 -70.048 52.809
336803 95.495 -70.060 52.809
336804 95.527 -70.072 52.813
336805 95.558 -70.083 52.814
336806 95.590 -70.096 52.805
336807 95.621 -70.108 52.789
336808 95.653 -70.119 52.768
336809 95.684 -70.131 52.738
336810 95.716 -70.143 52.712
336811 95.747 -70.155 52.686
336812 95.777 -70.168 52.659
336813 95.797 -70.182 52.636
336814 95.817 -70.197 52.615
336815 95.837 -70.212 52.596
336816 95.858 -70.227 52.579
336817 95.882 -70.237 52.567
336818 95.910 -70.245 52.547
336819 95.939 -70.253 52.528
336820 95.967 -70.260 52.509
336821 95.995 -70.268 52.490
336822 96.024 -70.275 52.464
336823 96.052 -70.282 52.438
336824 96.080 -70.290 52.412
336825 96.109 -70.297 52.390
336826 96.137 -70.305 52.370
336827 96.165 -70.312 52.352
336828 96.193 -70.320 52.335
336829 96.217 -70.328 52.320
336830 96.247 -70.338 52.303
336831 96.283 -70.350 52.286
336832 96.319 -70.360 52.270
336833 96.355 -70.370 52.257
336834 96.390 -70.380 52.248
336835 96.426 -70.390 52.242
336836 96.462 -70.400 52.230
336837 96.498 -70.409 52.216
336838 96.533 -70.411 52.199
336839 96.569 -70.413 52.183
336840 96.611 -70.422 52.167
336841 96.656 -70.432 52.153
336842 96.701 -70.443 52.140
336843 96.746 -70.453 52.129
336844 96.791 -70.464 52.126
336845 96.839 -70.477 52.127
336846 96.891 -70.492 52.131
336847 96.931 -70.500 52.134
336848 96.970 -70.508 52.137
336849 97.009 -70.516 52.133
336850 97.047 -70.524 52.124
336851 97.084 -70.530 52.111
336852 97.123 -70.538 52.091
336853 97.163 -70.546 52.070
336854 97.203 -70.554 52.047
336855 97.238 -70.563 52.023
336856 97.261 -70.576 51.991
336857 97.290 -70.597 51.960
336858 97.318 -70.619 51.930
336859 97.337 -70.634 51.901
336860 97.355 -70.649 51.876
336861 97.376 -70.660 51.854
336862 97.399 -70.668 51.830
336863 97.423 -70.676 51.803
336864 97.446 -70.684 51.775
336865 97.472 -70.691 51.744
336866 97.504 -70.695 51.712
336867 97.536 -70.699 51.678
336868 97.568 -70.703 51.642
336869 97.596 -70.707 51.602
336870 97.623 -70.711 51.563
336871 97.650 -70.715 51.524
336872 97.685 -70.724 51.484
336873 97.719 -70.732 51.445
336874 97.754 -70.742 51.407
336875 97.791 -70.745 51.370
336876 97.828 -70.746 51.333
336877 97.867 -70.747 51.296
336878 97.904 -70.748 51.257
336879 97.942 -70.748 51.214
336880 97.975 -70.742 51.168
336881 98.009 -70.738 51.121
336882 98.042 -70.732 51.070
336883 98.075 -70.728 51.019
336884 98.108 -70.723 50.967
336885 98.142 -70.718 50.917
336886 98.176 -70.717 50.870
336887 98.212 -70.723 50.822
336888 98.248 -70.730 50.772
336889 98.285 -70.737 50.720
336890 98.321 -70.744 50.665
336891 98.357 -70.751 50.608
336892 98.393 -70.757 50.551
336893 98.429 -70.764 50.493
336894 98.465 -70.771 50.432
336895 98.502 -70.777 50.370
336896 98.542 -70.780 50.308
336897 98.582 -70.783 50.244
336898 98.622 -70.785 50.181
336899 98.662 -70.789 50.120
336900 98.702 -70.792 50.058
336901 98.741 -70.795 49.996
336902 98.781 -70.797 49.933
336903 98.821 -70.800 49.869
336904 98.861 -70.803 49.806
336905 98.901 -70.806 49.745
336906 98.931 -70.817 49.686
336907 98.961 -70.829 49.626
336908 98.994 -70.843 49.567
336909 99.026 -70.858 49.509
336910 99.058 -70.874 49.452
336911 99.090 -70.891 49.398
336912 99.122 -70.908 49.344
336913 99.154 -70.924 49.292
336914 99.186 -70.941 49.240
336915 99.218 -70.958 49.190
336916 99.250 -70.974 49.140
336917 99.282 -70.991 49.090
336918 99.314 -71.008 49.042
336919 99.346 -71.024 48.994
336920 99.378 -71.041 48.945
336921 99.410 -71.053 48.900
336922 99.444 -71.061 48.855
336923 99.479 -71.065 48.812
336924 99.514 -71.068 48.769
336925 99.549 -71.071 48.727
336926 99.584 -71.074 48.686
336927 99.622 -71.079 48.647
336928 99.662 -71.086 48.608
336929 99.702 -71.092 48.570
336930 99.742 -71.099 48.534
336931 99.786 -71.112 48.498
336932 99.829 -71.129 48.461
336933 99.868 -71.144 48.423
336934 99.907 -71.159 48.387
336935 99.945 -71.174 48.350
336936 99.984 -71.190 48.312
336937 100.022 -71.205 48.273
336938 100.061 -71.220 48.234
336939 100.100 -71.235 48.193
336940 100.139 -71.250 48.150
336941 100.177 -71.266 48.107
336942 100.216 -71.281 48.062
336943 100.255 -71.296 48.017
336944 100.293 -71.315 47.972
336945 100.331 -71.334 47.927
336946 100.369 -71.353 47.884
336947 100.408 -71.371 47.843
336948 100.437 -71.388 47.802
336949 100.462 -71.403 47.761
336950 100.486 -71.419 47.720
336951 100.510 -71.434 47.678
336952 100.535 -71.450 47.637
336953 100.559 -71.465 47.598
336954 100.583 -71.481 47.561
336955 100.607 -71.496 47.523
336956 100.631 -71.511 47.487
336957 100.656 -71.526 47.450
336958 100.694 -71.535 47.413
336959 100.732 -71.544 47.377
336960 100.789 -71.558 47.341
336961 100.856 -71.575 47.306
336962 100.911 -71.584 47.270
336963 100.965 -71.591 47.235
336964 101.018 -71.598 47.199
336965 101.072 -71.606 47.164
336966 101.125 -71.613 47.128
336967 101.162 -71.624 47.094
336968 101.179 -71.638 47.060
336969 101.195 -71.654 47.025
336970 101.212 -71.668 46.991
336971 101.228 -71.682 46.956
336972 101.245 -71.697 46.921
336973 101.272 -71.707 46.885
336974 101.309 -71.713 46.850
336975 101.344 -71.719 46.814
336976 101.381 -71.725 46.779
336977 101.417 -71.731 46.743
336978 101.453 -71.737 46.707
336979 101.489 -71.742 46.671
336980 101.525 -71.748 46.636
336981 101.561 -71.754 46.601
336982 101.597 -71.763 46.566
336983 101.632 -71.776 46.530
336984 101.668 -71.788 46.495
336985 101.704 -71.801 46.457
336986 101.746 -71.819 46.418
336987 101.788 -71.836 46.379
336988 101.830 -71.854 46.339
336989 101.872 -71.871 46.299
336990 101.906 -71.890 46.258
336991 101.935 -71.910 46.217
336992 101.964 -71.930 46.176
336993 101.993 -71.950 46.134
336994 102.010 -71.946 46.093
336995 102.016 -71.941 46.054
336996 102.018 -71.933 46.016
336997 102.020 -71.926 45.978
336998 102.037 -71.923 45.942
336999 102.067 -71.925 45.906
337000 102.097 -71.926 45.871
337001 102.127 -71.928 45.835
337002 102.157 -71.929 45.801
337003 102.187 -71.931 45.767
337004 102.218 -71.933 45.733
337005 102.247 -71.934 45.700
337006 102.277 -71.936 45.667
337007 102.307 -71.937 45.634
337008 102.337 -71.939 45.602
337009 102.368 -71.940 45.570
337010 102.397 -71.942 45.537
337011 102.427 -71.943 45.505
337394 91.791 -70.429 55.240
337395 91.822 -70.440 55.232
337396 91.853 -70.450 55.224
337397 91.884 -70.461 55.217
337398 91.915 -70.471 55.210
337399 91.971 -70.499 55.203
337400 92.026 -70.542 55.197
337401 92.077 -70.548 55.189
337402 92.128 -70.524 55.182
337403 92.178 -70.500 55.177
337404 92.229 -70.475 55.172
337405 92.274 -70.462 55.167
337406 92.309 -70.464 55.161
337407 92.346 -70.461 55.152
337408 92.383 -70.459 55.139
337409 92.421 -70.456 55.124
337410 92.458 -70.453 55.109
337411 92.495 -70.451 55.095
337412 92.532 -70.448 55.083
337413 92.570 -70.446 55.071
337414 92.614 -70.446 55.056
337415 92.658 -70.445 55.036
337416 92.703 -70.444 55.007
337417 92.749 -70.444 54.976
337418 92.794 -70.443 54.946
337419 92.829 -70.434 54.917
337420 92.856 -70.420 54.888
337421 92.902 -70.407 54.854
337422 92.956 -70.394 54.814
337423 92.993 -70.366 54.771
337424 93.028 -70.343 54.723
337425 93.057 -70.335 54.671
337426 93.085 -70.328 54.618
337427 93.115 -70.320 54.565
337428 93.143 -70.312 54.514
337429 93.172 -70.305 54.466
337430 93.201 -70.297 54.420
337431 93.230 -70.290 54.375
337432 93.259 -70.281 54.333
337433 93.288 -70.274 54.291
337434 93.317 -70.266 54.253
337435 93.346 -70.259 54.216
337436 93.374 -70.251 54.179
337437 93.403 -70.243 54.145
337484 94.840 -70.043 52.842
337485 94.873 -70.044 52.836
337486 94.906 -70.045 52.831
337487 94.938 -70.046 52.826
337488 94.972 -70.046 52.823
337489 95.004 -70.047 52.822
337490 95.038 -70.048 52.821
337491 95.070 -70.049 52.821
337492 95.104 -70.049 52.824
337493 95.137 -70.050 52.828
337494 95.170 -70.051 52.833
337495 95.203 -70.052 52.837
337496 95.236 -70.052 52.841
337497 95.269 -70.053 52.843
337498 95.319 -70.037 52.841
337499 95.360 -70.026 52.837
337500 95.396 -70.018 52.831
337501 95.431 -70.009 52.826
337502 95.455 -70.010 52.821
337503 95.474 -70.015 52.818
337504 95.506 -70.027 52.817
337505 95.537 -70.040 52.822
337506 95.568 -70.051 52.824
337507 95.600 -70.063 52.820
337508 95.631 -70.075 52.806
337509 95.663 -70.087 52.784
337510 95.695 -70.099 52.756
337511 95.726 -70.111 52.728
337512 95.758 -70.123 52.701
337513 95.778 -70.137 52.673
337514 95.798 -70.152 52.648
337515 95.818 -70.167 52.627
337516 95.839 -70.181 52.608
337517 95.859 -70.196 52.593
337518 95.883 -70.208 52.580
337519 95.911 -70.215 52.561
337520 95.939 -70.223 52.540
337521 95.968 -70.231 52.520
337522 95.996 -70.237 52.502
337523 96.024 -70.245 52.477
337524 96.052 -70.253 52.449
337525 96.081 -70.260 52.422
337526 96.109 -70.268 52.400
337527 96.137 -70.275 52.378
337528 96.166 -70.282 52.359
337529 96.196 -70.288 52.342
337530 96.227 -70.293 52.325
337531 96.260 -70.304 52.309
337532 96.296 -70.313 52.293
337533 96.331 -70.323 52.278
337534 96.367 -70.333 52.265
337535 96.403 -70.343 52.255
337536 96.439 -70.353 52.248
337537 96.475 -70.362 52.237
337538 96.511 -70.372 52.223
337539 96.546 -70.375 52.208
337540 96.586 -70.381 52.193
337541 96.631 -70.391 52.178
337542 96.676 -70.402 52.164
337543 96.721 -70.413 52.153
337544 96.766 -70.423 52.143
337545 96.811 -70.434 52.141
337546 96.856 -70.444 52.144
337547 96.899 -70.455 52.148
337548 96.938 -70.463 52.150
337549 96.977 -70.472 52.152
337550 97.016 -70.480 52.148
337551 97.054 -70.488 52.139
337552 97.091 -70.494 52.126
337553 97.127 -70.500 52.106
337554 97.163 -70.507 52.084
337555 97.201 -70.514 52.060
337556 97.236 -70.522 52.035
337557 97.264 -70.544 52.003
337558 97.293 -70.565 51.970
337559 97.322 -70.586 51.939
337560 97.343 -70.603 51.910
337561 97.361 -70.618 51.884
337562 97.379 -70.633 51.861
337563 97.400 -70.644 51.835
337564 97.424 -70.652 51.806
337565 97.447 -70.660 51.775
337566 97.472 -70.667 51.743
337567 97.504 -70.671 51.709
337568 97.536 -70.676 51.673
337569 97.566 -70.679 51.637
337570 97.593 -70.683 51.597
337571 97.620 -70.687 51.558
337572 97.647 -70.691 51.520
337573 97.676 -70.697 51.480
337574 97.714 -70.699 51.442
337575 97.752 -70.699 51.404
337576 97.790 -70.700 51.365
337577 97.828 -70.701 51.325
337578 97.866 -70.701 51.286
337579 97.904 -70.702 51.245
337580 97.942 -70.702 51.201
337581 97.980 -70.703 51.155
337582 98.016 -70.701 51.109
337583 98.050 -70.696 51.058
337584 98.083 -70.691 51.007
337585 98.116 -70.686 50.955
337586 98.149 -70.682 50.904
337587 98.183 -70.676 50.857
337588 98.218 -70.678 50.810
337589 98.254 -70.685 50.761
337590 98.290 -70.691 50.710
337591 98.326 -70.698 50.656
337592 98.362 -70.705 50.600
337593 98.398 -70.712 50.543
337594 98.435 -70.719 50.484
337595 98.471 -70.726 50.423
337596 98.509 -70.730 50.362
337597 98.549 -70.733 50.299
337598 98.589 -70.735 50.235
337599 98.629 -70.739 50.173
337600 98.668 -70.742 50.111
337601 98.708 -70.745 50.049
337602 98.748 -70.747 49.986
337603 98.788 -70.750 49.923
337604 98.828 -70.753 49.860
337605 98.868 -70.756 49.800
337606 98.908 -70.758 49.741
337607 98.941 -70.770 49.683
337608 98.973 -70.786 49.623
337609 99.005 -70.803 49.564
337610 99.037 -70.820 49.506
337611 99.069 -70.836 49.450
337612 99.101 -70.853 49.396
337613 99.133 -70.870 49.343
337614 99.165 -70.886 49.291
337615 99.197 -70.903 49.240
337616 99.229 -70.920 49.189
337617 99.261 -70.936 49.140
337618 99.293 -70.953 49.091
337619 99.325 -70.970 49.043
337620 99.357 -70.986 48.996
337621 99.389 -71.002 48.949
337622 99.421 -71.014 48.904
337623 99.454 -71.025 48.861
337624 99.489 -71.028 48.819
337625 99.524 -71.031 48.777
337626 99.559 -71.034 48.735
337627 99.594 -71.037 48.695
337628 99.630 -71.040 48.657
337629 99.664 -71.043 48.621
337630 99.705 -71.050 48.583
337631 99.745 -71.056 48.548
337632 99.786 -71.065 48.513
337633 99.830 -71.082 48.478
337634 99.869 -71.097 48.443
337635 99.908 -71.112 48.407
337636 99.946 -71.127 48.371
337637 99.985 -71.143 48.335
337638 100.024 -71.158 48.296
337639 100.062 -71.174 48.257
337640 100.101 -71.189 48.214
337641 100.140 -71.204 48.169
337642 100.178 -71.222 48.124
337643 100.216 -71.241 48.077
337644 100.255 -71.260 48.031
337645 100.293 -71.279 47.985
337646 100.331 -71.298 47.940
337647 100.369 -71.317 47.895
337648 100.407 -71.335 47.854
337649 100.431 -71.351 47.813
337650 100.455 -71.366 47.771
337651 100.480 -71.381 47.732
337652 100.504 -71.397 47.692
337653 100.528 -71.412 47.651
337654 100.552 -71.428 47.612
337655 100.576 -71.443 47.575
337656 100.601 -71.459 47.538
337657 100.637 -71.468 47.501
337658 100.675 -71.476 47.465
337659 100.712 -71.485 47.428
337660 100.750 -71.494 47.391
337661 100.796 -71.505 47.355
337662 100.863 -71.522 47.320
337663 100.930 -71.540 47.284
337664 100.997 -71.557 47.250
337665 101.053 -71.566 47.215
337666 101.107 -71.573 47.180
337667 101.158 -71.582 47.145
337668 101.175 -71.596 47.110
337669 101.191 -71.610 47.076
337670 101.207 -71.625 47.041
337671 101.224 -71.640 47.006
337672 101.240 -71.655 46.971
337673 101.257 -71.669 46.935
337674 101.287 -71.678 46.899
337675 101.323 -71.684 46.864
337676 101.359 -71.690 46.827
337677 101.395 -71.696 46.791
337678 101.431 -71.702 46.754
337679 101.467 -71.707 46.717
337680 101.503 -71.713 46.680
337681 101.539 -71.719 46.644
337682 101.575 -71.726 46.607
337683 101.611 -71.732 46.570
337684 101.647 -71.743 46.533
337685 101.683 -71.756 46.496
337686 101.721 -71.770 46.458
337687 101.763 -71.787 46.419
337688 101.805 -71.804 46.380
337689 101.847 -71.822 46.341
337690 101.889 -71.840 46.302
337691 101.925 -71.858 46.261
337692 101.954 -71.879 46.221
337693 101.982 -71.898 46.181
337694 102.000 -71.898 46.139
337695 102.002 -71.890 46.099
337696 102.004 -71.882 46.059
337697 102.006 -71.874 46.021
337698 102.021 -71.871 45.984
337699 102.051 -71.873 45.948
337700 102.081 -71.874 45.912
337701 102.111 -71.876 45.877
337702 102.141 -71.877 45.842
337703 102.171 -71.879 45.807
337704 102.201 -71.880 45.774
337705 102.231 -71.882 45.740
337706 102.261 -71.884 45.707
337707 102.291 -71.885 45.675
337708 102.321 -71.887 45.643
337709 102.351 -71.889 45.610
337710 102.381 -71.890 45.578
338096 91.810 -70.407 55.227
338097 91.841 -70.418 55.219
338098 91.872 -70.428 55.212
338099 91.907 -70.441 55.205
338100 91.965 -70.471 55.199
338101 92.019 -70.517 55.192
338102 92.070 -70.519 55.185
338103 92.121 -70.494 55.177
338104 92.171 -70.470 55.171
338105 92.222 -70.446 55.166
338106 92.273 -70.422 55.160
338107 92.308 -70.424 55.153
338108 92.345 -70.421 55.144
338109 92.382 -70.419 55.133
338110 92.419 -70.416 55.122
338111 92.456 -70.413 55.112
338112 92.494 -70.411 55.102
338113 92.531 -70.408 55.095
338114 92.568 -70.406 55.086
338115 92.605 -70.403 55.074
338116 92.648 -70.402 55.057
338117 92.694 -70.401 55.030
338118 92.739 -70.400 55.000
338119 92.785 -70.400 54.968
338120 92.827 -70.396 54.940
338121 92.854 -70.383 54.910
338122 92.896 -70.370 54.876
338123 92.952 -70.357 54.836
338124 92.995 -70.334 54.789
338125 93.031 -70.307 54.740
338126 93.064 -70.290 54.686
338127 93.093 -70.282 54.633
338128 93.122 -70.275 54.580
338129 93.151 -70.267 54.528
338130 93.180 -70.259 54.479
338131 93.209 -70.252 54.432
338132 93.237 -70.243 54.387
338133 93.266 -70.236 54.344
338134 93.295 -70.228 54.302
338135 93.324 -70.221 54.262
338136 93.353 -70.213 54.224
338137 93.382 -70.206 54.187
338138 93.411 -70.198 54.152
338184 94.827 -70.024 52.849
338185 94.860 -70.024 52.840
338186 94.893 -70.025 52.834
338187 94.926 -70.026 52.829
338188 94.959 -70.026 52.825
338189 94.992 -70.027 52.823
338190 95.025 -70.027 52.822
338191 95.058 -70.028 52.823
338192 95.091 -70.029 52.825
338193 95.124 -70.030 52.829
338194 95.157 -70.030 52.836
338195 95.190 -70.031 52.844
338196 95.223 -70.032 52.851
338197 95.256 -70.033 52.856
338198 95.295 -70.028 52.860
338199 95.341 -70.014 52.857
338200 95.377 -70.005 52.852
338201 95.412 -69.997 52.844
338202 95.448 -69.990 52.836
338203 95.478 -69.985 52.831
338204 95.497 -69.990 52.826
338205 95.516 -69.995 52.825
338206 95.547 -70.007 52.828
338207 95.579 -70.018 52.831
338208 95.611 -70.030 52.829
338209 95.642 -70.043 52.818
338210 95.674 -70.054 52.797
338211 95.705 -70.066 52.769
338212 95.737 -70.078 52.742
338213 95.759 -70.093 52.715
338214 95.779 -70.107 52.687
338215 95.800 -70.122 52.660
338216 95.820 -70.137 52.639
338217 95.840 -70.151 52.620
338218 95.860 -70.166 52.605
338219 95.883 -70.178 52.592
338220 95.911 -70.186 52.573
338221 95.940 -70.193 52.553
338222 95.968 -70.200 52.531
338223 95.997 -70.208 52.512
338224 96.028 -70.213 52.488
338225 96.058 -70.218 52.460
338226 96.090 -70.224 52.432
338227 96.120 -70.229 52.409
338228 96.151 -70.234 52.387
338229 96.182 -70.240 52.367
338230 96.212 -70.246 52.349
338231 96.241 -70.259 52.332
338232 96.274 -70.269 52.315
338233 96.309 -70.278 52.300
338234 96.345 -70.287 52.286
338235 96.381 -70.297 52.272
338236 96.416 -70.306 52.261
338237 96.452 -70.315 52.253
338238 96.488 -70.325 52.244
338239 96.524 -70.334 52.232
338240 96.561 -70.341 52.219
338241 96.606 -70.351 52.204
338242 96.651 -70.362 52.190
338243 96.695 -70.371 52.177
338244 96.740 -70.378 52.165
338245 96.784 -70.384 52.156
338246 96.828 -70.391 52.155
338247 96.869 -70.401 52.158
338248 96.907 -70.416 52.161
338249 96.945 -70.427 52.162
338250 96.983 -70.435 52.164
338251 97.022 -70.444 52.161
338252 97.061 -70.452 52.153
338253 97.097 -70.458 52.138
338254 97.134 -70.465 52.118
338255 97.170 -70.471 52.096
338256 97.208 -70.478 52.071
338257 97.240 -70.491 52.045
338258 97.268 -70.511 52.015
338259 97.297 -70.532 51.982
338260 97.325 -70.554 51.950
338261 97.349 -70.572 51.919
338262 97.367 -70.587 51.893
338263 97.385 -70.602 51.867
338264 97.403 -70.617 51.840
338265 97.425 -70.628 51.810
338266 97.448 -70.636 51.778
338267 97.472 -70.644 51.743
338268 97.504 -70.648 51.707
338269 97.535 -70.651 51.671
338270 97.563 -70.655 51.636
338271 97.590 -70.660 51.596
338272 97.617 -70.663 51.557
338273 97.651 -70.660 51.519
338274 97.684 -70.657 51.479
338275 97.719 -70.656 51.441
338276 97.753 -70.654 51.401
338277 97.790 -70.654 51.358
338278 97.827 -70.655 51.316
338279 97.866 -70.656 51.276
338280 97.904 -70.656 51.231
338281 97.942 -70.657 51.186
338282 97.980 -70.657 51.141
338283 98.018 -70.658 51.096
338284 98.056 -70.659 51.047
338285 98.090 -70.655 50.996
338286 98.124 -70.650 50.943
338287 98.157 -70.645 50.890
338288 98.190 -70.640 50.841
338289 98.224 -70.635 50.794
338290 98.259 -70.639 50.747
338291 98.295 -70.646 50.698
338292 98.332 -70.653 50.646
338293 98.368 -70.660 50.592
338294 98.404 -70.666 50.534
338295 98.440 -70.673 50.474
338296 98.477 -70.680 50.414
338297 98.516 -70.683 50.352
338298 98.556 -70.685 50.290
338299 98.596 -70.688 50.229
338300 98.636 -70.691 50.169
338301 98.675 -70.695 50.106
338302 98.715 -70.697 50.043
338303 98.755 -70.700 49.979
338304 98.795 -70.703 49.916
338305 98.835 -70.706 49.854
338306 98.874 -70.712 49.795
338307 98.912 -70.721 49.737
338308 98.951 -70.732 49.679
338309 98.984 -70.748 49.620
338310 99.016 -70.765 49.561
338311 99.048 -70.782 49.504
338312 99.080 -70.798 49.449
338313 99.112 -70.815 49.394
338314 99.144 -70.832 49.341
338315 99.176 -70.848 49.290
338316 99.208 -70.865 49.239
338317 99.240 -70.882 49.189
338318 99.272 -70.898 49.140
338319 99.304 -70.915 49.091
338320 99.336 -70.932 49.044
338321 99.368 -70.948 48.999
338322 99.400 -70.964 48.952
338323 99.432 -70.976 48.909
338324 99.465 -70.987 48.867
338325 99.499 -70.991 48.826
338326 99.535 -70.994 48.784
338327 99.570 -70.997 48.745
338328 99.605 -71.000 48.706
338329 99.640 -71.003 48.668
338330 99.675 -71.006 48.632
338331 99.710 -71.009 48.597
338332 99.748 -71.014 48.564
338333 99.788 -71.021 48.531
338334 99.832 -71.036 48.498
338335 99.870 -71.051 48.464
338336 99.909 -71.066 48.429
338337 99.947 -71.081 48.393
338338 99.986 -71.096 48.356
338339 100.025 -71.111 48.315
338340 100.063 -71.130 48.274
338341 100.102 -71.149 48.230
338342 100.140 -71.168 48.184
338343 100.178 -71.187 48.137
338344 100.216 -71.205 48.090
338345 100.254 -71.224 48.044
338346 100.292 -71.243 47.999
338347 100.331 -71.262 47.953
338348 100.369 -71.281 47.909
338349 100.401 -71.298 47.869
338350 100.425 -71.313 47.828
338351 100.449 -71.329 47.787
338352 100.473 -71.344 47.747
338353 100.497 -71.359 47.707
338354 100.522 -71.375 47.667
338355 100.546 -71.390 47.627
338356 100.580 -71.400 47.591
338357 100.618 -71.409 47.554
338358 100.656 -71.418 47.517
338359 100.694 -71.426 47.481
338360 100.731 -71.435 47.444
338361 100.769 -71.444 47.407
338362 100.807 -71.453 47.372
338363 100.870 -71.469 47.337
338364 100.937 -71.486 47.302
338365 101.004 -71.503 47.267
338366 101.071 -71.521 47.233
338367 101.138 -71.538 47.198
338368 101.170 -71.553 47.163
338369 101.187 -71.568 47.128
338370 101.203 -71.583 47.094
338371 101.219 -71.597 47.059
338372 101.236 -71.612 47.024
338373 101.252 -71.627 46.989
338374 101.269 -71.641 46.952
338375 101.302 -71.649 46.916
338376 101.337 -71.655 46.880
338377 101.374 -71.661 46.843
338378 101.410 -71.667 46.805
338379 101.446 -71.672 46.767
338380 101.482 -71.679 46.729
338381 101.518 -71.685 46.691
338382 101.554 -71.691 46.654
338383 101.590 -71.697 46.616
338384 101.626 -71.703 46.578
338385 101.662 -71.710 46.539
338386 101.697 -71.722 46.500
338387 101.738 -71.738 46.462
338388 101.780 -71.756 46.422
338389 101.822 -71.773 46.383
338390 101.865 -71.791 46.344
338391 101.906 -71.808 46.305
338392 101.943 -71.827 46.266
338393 101.972 -71.847 46.226
338394 101.986 -71.846 46.186
338395 101.988 -71.838 46.145
338396 101.990 -71.831 46.105
338397 101.992 -71.823 46.065
338398 102.005 -71.819 46.027
338399 102.035 -71.820 45.989
338400 102.065 -71.822 45.954
338401 102.095 -71.824 45.918
338402 102.125 -71.826 45.883
338403 102.155 -71.827 45.848
338404 102.185 -71.829 45.814
338405 102.215 -71.830 45.780
338406 102.245 -71.832 45.747
338407 102.275 -71.833 45.715
338408 102.305 -71.835 45.682
338409 102.335 -71.836 45.649
338410 102.365 -71.838 45.616
338797 91.798 -70.375 55.222
338798 91.829 -70.385 55.215
338799 91.860 -70.396 55.208
338800 91.900 -70.413 55.202
338801 91.959 -70.442 55.195
338802 92.012 -70.493 55.188
338803 92.063 -70.489 55.180
338804 92.113 -70.465 55.173
338805 92.164 -70.441 55.166
338806 92.215 -70.416 55.160
338807 92.266 -70.392 55.153
338808 92.307 -70.384 55.147
338809 92.344 -70.381 55.139
338810 92.381 -70.378 55.130
338811 92.418 -70.375 55.123
338812 92.456 -70.372 55.117
338813 92.493 -70.368 55.112
338814 92.530 -70.364 55.107
338815 92.567 -70.362 55.099
338816 92.602 -70.362 55.088
338817 92.639 -70.361 55.075
338818 92.684 -70.358 55.052
338819 92.729 -70.357 55.022
338820 92.775 -70.356 54.991
338821 92.821 -70.355 54.963
338822 92.853 -70.345 54.932
338823 92.889 -70.332 54.897
338824 92.945 -70.319 54.856
338825 92.997 -70.303 54.807
338826 93.034 -70.276 54.756
338827 93.070 -70.249 54.703
338828 93.100 -70.237 54.649
338829 93.130 -70.229 54.597
338830 93.158 -70.221 54.542
338831 93.187 -70.213 54.492
338832 93.216 -70.206 54.444
338833 93.245 -70.198 54.399
338834 93.274 -70.190 54.355
338835 93.303 -70.183 54.312
338836 93.332 -70.175 54.272
338837 93.360 -70.168 54.234
338838 93.390 -70.160 54.196
338839 93.418 -70.152 54.161
338840 93.447 -70.145 54.125
338885 94.847 -70.004 52.847
338886 94.881 -70.005 52.838
338887 94.913 -70.005 52.832
338888 94.947 -70.006 52.827
338889 94.979 -70.007 52.824
338890 95.013 -70.008 52.822
338891 95.045 -70.008 52.822
338892 95.078 -70.009 52.824
338893 95.111 -70.009 52.828
338894 95.144 -70.010 52.835
338895 95.177 -70.011 52.845
338896 95.210 -70.011 52.856
338897 95.243 -70.012 52.867
338898 95.276 -70.013 52.874
338899 95.321 -70.002 52.876
338900 95.358 -69.993 52.874
338901 95.393 -69.986 52.867
338902 95.429 -69.977 52.857
338903 95.464 -69.969 52.848
338904 95.499 -69.961 52.840
338905 95.520 -69.964 52.835
338906 95.539 -69.969 52.832
338907 95.558 -69.974 52.832
338908 95.590 -69.986 52.836
338909 95.621 -69.998 52.836
338910 95.653 -70.010 52.827
338911 95.684 -70.021 52.807
338912 95.716 -70.033 52.780
338913 95.740 -70.047 52.754
338914 95.760 -70.062 52.728
338915 95.780 -70.077 52.699
338916 95.801 -70.092 52.672
338917 95.821 -70.106 52.650
338918 95.841 -70.121 52.632
338919 95.864 -70.134 52.617
338920 95.890 -70.143 52.604
338921 95.921 -70.149 52.585
338922 95.952 -70.154 52.564
338923 95.982 -70.159 52.543
338924 96.013 -70.165 52.523
338925 96.044 -70.170 52.499
338926 96.075 -70.175 52.471
338927 96.105 -70.181 52.443
338928 96.136 -70.187 52.419
338929 96.167 -70.192 52.397
338930 96.197 -70.200 52.376
338931 96.226 -70.212 52.357
338932 96.255 -70.224 52.339
338933 96.288 -70.234 52.322
338934 96.323 -70.243 52.308
338935 96.359 -70.253 52.294
338936 96.394 -70.262 52.281
338937 96.430 -70.271 52.269
338938 96.465 -70.280 52.262
338939 96.501 -70.290 52.253
338940 96.537 -70.299 52.244
338941 96.579 -70.301 52.233
338942 96.623 -70.308 52.218
338943 96.667 -70.315 52.203
338944 96.711 -70.321 52.190
338945 96.755 -70.328 52.177
338946 96.799 -70.334 52.167
338947 96.838 -70.347 52.166
338948 96.876 -70.362 52.168
338949 96.914 -70.377 52.171
338950 96.953 -70.389 52.172
338951 96.991 -70.398 52.172
338952 97.029 -70.407 52.168
338953 97.068 -70.416 52.159
338954 97.104 -70.422 52.144
338955 97.142 -70.430 52.123
338956 97.180 -70.438 52.101
338957 97.219 -70.447 52.077
338958 97.248 -70.461 52.051
338959 97.272 -70.478 52.022
338960 97.300 -70.500 51.990
338961 97.329 -70.522 51.957
338962 97.355 -70.541 51.927
338963 97.373 -70.557 51.899
338964 97.391 -70.571 51.873
338965 97.409 -70.586 51.845
338966 97.427 -70.601 51.814
338967 97.448 -70.612 51.780
338968 97.472 -70.619 51.745
338969 97.505 -70.623 51.709
338970 97.532 -70.628 51.672
338971 97.560 -70.632 51.635
338972 97.593 -70.628 51.593
338973 97.628 -70.623 51.554
338974 97.663 -70.617 51.514
338975 97.696 -70.618 51.473
338976 97.730 -70.616 51.434
338977 97.764 -70.615 51.392
338978 97.799 -70.613 51.348
338979 97.833 -70.613 51.303
338980 97.867 -70.611 51.259
338981 97.903 -70.610 51.216
338982 97.941 -70.611 51.172
338983 97.980 -70.612 51.126
338984 98.017 -70.613 51.081
338985 98.056 -70.613 51.032
338986 98.093 -70.613 50.981
338987 98.131 -70.613 50.930
338988 98.164 -70.609 50.878
338989 98.198 -70.604 50.827
338990 98.231 -70.599 50.779
338991 98.265 -70.594 50.731
338992 98.301 -70.601 50.682
338993 98.337 -70.608 50.632
338994 98.373 -70.615 50.579
338995 98.409 -70.622 50.523
338996 98.445 -70.629 50.465
338997 98.484 -70.633 50.407
338998 98.523 -70.635 50.346
338999 98.563 -70.638 50.285
339000 98.603 -70.641 50.225
339001 98.643 -70.644 50.163
339002 98.683 -70.647 50.099
339003 98.722 -70.650 50.036
339004 98.762 -70.653 49.974
339005 98.801 -70.660 49.911
339006 98.839 -70.670 49.849
339007 98.877 -70.679 49.791
339008 98.916 -70.688 49.733
339009 98.957 -70.701 49.675
339010 98.994 -70.712 49.616
339011 99.027 -70.727 49.557
339012 99.059 -70.744 49.500
339013 99.091 -70.760 49.445
339014 99.123 -70.777 49.391
339015 99.155 -70.794 49.339
339016 99.187 -70.811 49.288
339017 99.219 -70.827 49.238
339018 99.251 -70.844 49.189
339019 99.283 -70.861 49.140
339020 99.314 -70.877 49.093
339021 99.347 -70.894 49.046
339022 99.378 -70.911 49.001
339023 99.411 -70.925 48.956
339024 99.443 -70.937 48.913
339025 99.476 -70.948 48.872
339026 99.510 -70.954 48.832
339027 99.545 -70.957 48.792
339028 99.580 -70.960 48.753
339029 99.615 -70.963 48.716
339030 99.650 -70.966 48.680
339031 99.685 -70.969 48.645
339032 99.720 -70.972 48.611
339033 99.756 -70.975 48.578
339034 99.791 -70.979 48.547
339035 99.833 -70.989 48.516
339036 99.871 -71.004 48.483
339037 99.910 -71.019 48.448
339038 99.948 -71.037 48.411
339039 99.987 -71.056 48.372
339040 100.025 -71.075 48.330
339041 100.063 -71.094 48.287
339042 100.101 -71.113 48.242
339043 100.139 -71.132 48.195
339044 100.177 -71.151 48.150
339045 100.216 -71.169 48.103
339046 100.254 -71.188 48.058
339047 100.292 -71.207 48.013
339048 100.330 -71.226 47.969
339049 100.368 -71.245 47.925
339050 100.394 -71.261 47.884
339051 100.418 -71.276 47.844
339052 100.442 -71.291 47.803
339053 100.467 -71.307 47.763
339054 100.491 -71.322 47.724
339055 100.524 -71.333 47.683
339056 100.562 -71.342 47.644
339057 100.600 -71.350 47.608
339058 100.637 -71.359 47.571
339059 100.675 -71.368 47.534
339060 100.712 -71.377 47.498
339061 100.750 -71.385 47.461
339062 100.787 -71.394 47.424
339063 100.825 -71.403 47.389
339064 100.877 -71.415 47.355
339065 100.945 -71.433 47.321
339066 101.011 -71.450 47.286
339067 101.078 -71.468 47.252
339068 101.143 -71.484 47.218
339069 101.173 -71.497 47.183
339070 101.192 -71.522 47.149
339071 101.212 -71.547 47.115
339072 101.232 -71.569 47.080
339073 101.248 -71.585 47.045
339074 101.264 -71.599 47.009
339075 101.281 -71.613 46.972
339076 101.316 -71.620 46.935
339077 101.352 -71.626 46.898
339078 101.388 -71.632 46.860
339079 101.424 -71.638 46.822
339080 101.460 -71.644 46.782
339081 101.496 -71.650 46.743
339082 101.532 -71.656 46.704
339083 101.568 -71.662 46.665
339084 101.604 -71.668 46.627
339085 101.640 -71.673 46.589
339086 101.676 -71.679 46.550
339087 101.713 -71.690 46.509
339088 101.755 -71.707 46.470
339089 101.798 -71.725 46.430
339090 101.839 -71.742 46.390
339091 101.882 -71.760 46.350
339092 101.923 -71.777 46.312
339093 101.962 -71.795 46.272
339094 101.972 -71.795 46.232
339095 101.974 -71.787 46.191
339096 101.976 -71.779 46.151
339097 101.978 -71.772 46.111
339098 101.989 -71.767 46.071
339099 102.019 -71.769 46.033
339100 102.049 -71.770 45.996
339101 102.079 -71.772 45.961
339102 102.109 -71.773 45.925
339103 102.139 -71.775 45.889
339104 102.169 -71.776 45.855
339105 102.199 -71.778 45.820
339106 102.229 -71.779 45.787
339107 102.259 -71.781 45.754
339108 102.289 -71.782 45.721
339109 102.319 -71.784 45.689
339498 91.786 -70.343 55.218
339499 91.817 -70.353 55.210
339500 91.848 -70.364 55.204
339501 91.894 -70.385 55.198
339502 91.953 -70.414 55.192
339503 92.005 -70.468 55.184
339504 92.056 -70.460 55.176
339505 92.106 -70.436 55.169
339506 92.157 -70.411 55.162
339507 92.208 -70.387 55.156
339508 92.259 -70.362 55.149
339509 92.308 -70.334 55.142
339510 92.345 -70.330 55.135
339511 92.383 -70.327 55.128
339512 92.419 -70.325 55.123
339513 92.455 -70.324 55.120
339514 92.490 -70.324 55.117
339515 92.526 -70.323 55.113
339516 92.561 -70.323 55.106
339517 92.597 -70.322 55.097
339518 92.632 -70.322 55.084
339519 92.675 -70.317 55.064
339520 92.720 -70.314 55.039
339521 92.766 -70.312 55.011
339522 92.811 -70.312 54.984
339523 92.852 -70.308 54.952
339524 92.882 -70.294 54.916
339525 92.939 -70.282 54.873
339526 92.996 -70.269 54.823
339527 93.036 -70.244 54.772
339528 93.072 -70.217 54.720
339529 93.108 -70.191 54.668
339530 93.137 -70.184 54.616
339531 93.166 -70.175 54.558
339532 93.195 -70.168 54.505
339533 93.224 -70.160 54.456
339534 93.253 -70.152 54.410
339535 93.281 -70.145 54.366
339536 93.310 -70.137 54.323
339537 93.339 -70.130 54.283
339538 93.368 -70.122 54.244
339539 93.397 -70.115 54.206
339540 93.426 -70.107 54.169
339541 93.454 -70.099 54.134
339542 93.484 -70.091 54.098
339585 94.835 -69.983 52.857
339586 94.867 -69.984 52.846
339587 94.901 -69.985 52.836
339588 94.933 -69.986 52.829
339589 94.967 -69.986 52.825
339590 95.000 -69.987 52.822
339591 95.033 -69.988 52.821
339592 95.066 -69.989 52.821
339593 95.099 -69.990 52.825
339594 95.132 -69.990 52.831
339595 95.165 -69.991 52.841
339596 95.198 -69.992 52.856
339597 95.231 -69.993 52.874
339598 95.264 -69.993 52.890
339599 95.297 -69.993 52.898
339600 95.339 -69.982 52.893
339601 95.375 -69.974 52.887
339602 95.410 -69.965 52.880
339603 95.445 -69.958 52.871
339604 95.481 -69.949 52.860
339605 95.516 -69.941 52.850
339606 95.544 -69.939 52.843
339607 95.562 -69.944 52.839
339608 95.581 -69.949 52.837
339609 95.600 -69.954 52.840
339610 95.631 -69.965 52.840
339611 95.663 -69.977 52.833
339612 95.694 -69.990 52.816
339613 95.721 -70.002 52.790
339614 95.741 -70.017 52.764
339615 95.761 -70.032 52.737
339616 95.782 -70.046 52.710
339617 95.804 -70.060 52.683
339618 95.830 -70.070 52.661
339619 95.856 -70.079 52.643
339620 95.882 -70.089 52.628
339621 95.908 -70.099 52.615
339622 95.937 -70.105 52.596
339623 95.968 -70.112 52.575
339624 95.999 -70.117 52.555
339625 96.029 -70.122 52.536
339626 96.060 -70.127 52.511
339627 96.091 -70.133 52.482
339628 96.121 -70.138 52.454
339629 96.153 -70.143 52.431
339630 96.182 -70.153 52.408
339631 96.211 -70.165 52.386
339632 96.240 -70.177 52.366
339633 96.268 -70.189 52.348
339634 96.301 -70.199 52.330
339635 96.336 -70.209 52.316
339636 96.372 -70.218 52.303
339637 96.408 -70.227 52.290
339638 96.444 -70.236 52.279
339639 96.479 -70.246 52.271
339640 96.515 -70.255 52.264
339641 96.554 -70.255 52.256
339642 96.595 -70.253 52.246
339643 96.638 -70.258 52.232
339644 96.682 -70.265 52.217
339645 96.726 -70.271 52.203
339646 96.769 -70.278 52.189
339647 96.808 -70.293 52.178
339648 96.846 -70.309 52.176
339649 96.884 -70.323 52.178
339650 96.922 -70.338 52.181
339651 96.960 -70.352 52.181
339652 96.998 -70.361 52.179
339653 97.036 -70.370 52.173
339654 97.076 -70.381 52.161
339655 97.115 -70.391 52.145
339656 97.154 -70.399 52.123
339657 97.191 -70.406 52.100
339658 97.230 -70.415 52.077
339659 97.255 -70.431 52.053
339660 97.280 -70.450 52.024
339661 97.304 -70.468 51.993
339662 97.332 -70.489 51.962
339663 97.361 -70.510 51.932
339664 97.379 -70.525 51.903
339665 97.397 -70.541 51.876
339666 97.415 -70.555 51.847
339667 97.433 -70.570 51.815
339668 97.451 -70.585 51.781
339669 97.473 -70.596 51.746
339670 97.502 -70.600 51.710
339671 97.535 -70.597 51.672
339672 97.570 -70.591 51.633
339673 97.605 -70.585 51.589
339674 97.641 -70.580 51.547
339675 97.675 -70.576 51.506
339676 97.708 -70.579 51.464
339677 97.742 -70.577 51.425
339678 97.776 -70.576 51.383
339679 97.810 -70.574 51.339
339680 97.845 -70.573 51.292
339681 97.879 -70.572 51.245
339682 97.913 -70.570 51.201
339683 97.947 -70.569 51.156
339684 97.981 -70.567 51.111
339685 98.017 -70.566 51.066
339686 98.055 -70.567 51.017
339687 98.093 -70.568 50.968
339688 98.131 -70.569 50.918
339689 98.169 -70.569 50.867
339690 98.205 -70.567 50.816
339691 98.238 -70.563 50.766
339692 98.272 -70.557 50.715
339693 98.306 -70.556 50.665
339694 98.342 -70.563 50.616
339695 98.378 -70.569 50.565
339696 98.415 -70.576 50.511
339697 98.451 -70.583 50.455
339698 98.491 -70.585 50.399
339699 98.530 -70.588 50.338
339700 98.570 -70.591 50.278
339701 98.610 -70.594 50.217
339702 98.650 -70.597 50.154
339703 98.690 -70.600 50.091
339704 98.728 -70.609 50.029
339705 98.766 -70.618 49.967
339706 98.804 -70.627 49.905
339707 98.843 -70.636 49.845
339708 98.880 -70.645 49.787
339709 98.920 -70.657 49.730
339710 98.961 -70.669 49.671
339711 99.000 -70.680 49.612
339712 99.036 -70.691 49.553
339713 99.070 -70.706 49.496
339714 99.101 -70.723 49.442
339715 99.134 -70.739 49.389
339716 99.165 -70.756 49.338
339717 99.198 -70.773 49.286
339718 99.229 -70.789 49.236
339719 99.262 -70.806 49.188
339720 99.293 -70.823 49.140
339721 99.326 -70.839 49.094
339722 99.357 -70.856 49.048
339723 99.390 -70.873 49.004
339724 99.422 -70.886 48.960
339725 99.454 -70.898 48.918
339726 99.486 -70.910 48.878
339727 99.520 -70.917 48.839
339728 99.555 -70.920 48.800
339729 99.590 -70.923 48.763
339730 99.626 -70.927 48.726
339731 99.660 -70.930 48.691
339732 99.696 -70.933 48.659
339733 99.731 -70.936 48.625
339734 99.766 -70.939 48.593
339735 99.801 -70.942 48.563
339736 99.836 -70.946 48.532
339737 99.872 -70.964 48.500
339738 99.910 -70.983 48.464
339739 99.948 -71.002 48.426
339740 99.986 -71.021 48.385
339741 100.025 -71.039 48.342
339742 100.063 -71.058 48.298
339743 100.101 -71.077 48.252
339744 100.139 -71.096 48.206
339745 100.177 -71.115 48.162
339746 100.215 -71.133 48.117
339747 100.254 -71.152 48.072
339748 100.292 -71.171 48.028
339749 100.330 -71.190 47.985
339750 100.363 -71.208 47.942
339751 100.388 -71.223 47.901
339752 100.412 -71.239 47.861
339753 100.436 -71.254 47.822
339754 100.468 -71.266 47.781
339755 100.505 -71.274 47.741
339756 100.543 -71.283 47.701
339757 100.580 -71.292 47.662
339758 100.618 -71.300 47.626
339759 100.656 -71.309 47.590
339760 100.693 -71.318 47.553
339761 100.731 -71.327 47.516
339762 100.769 -71.335 47.480
339763 100.807 -71.344 47.444
339764 100.844 -71.353 47.409
339765 100.885 -71.362 47.375
339766 100.952 -71.380 47.341
339767 101.019 -71.397 47.308
339768 101.086 -71.414 47.275
339769 101.136 -71.429 47.242
339770 101.166 -71.440 47.209
339771 101.192 -71.457 47.177
339772 101.212 -71.482 47.142
339773 101.232 -71.507 47.108
339774 101.252 -71.532 47.073
339775 101.272 -71.556 47.036
339776 101.293 -71.581 47.000
339777 101.330 -71.588 46.962
339778 101.366 -71.593 46.922
339779 101.402 -71.598 46.883
339780 101.438 -71.603 46.843
339781 101.474 -71.608 46.802
339782 101.510 -71.613 46.761
339783 101.545 -71.619 46.721
339784 101.581 -71.623 46.681
339785 101.617 -71.629 46.640
339786 101.653 -71.634 46.601
339787 101.689 -71.639 46.561
339788 101.726 -71.647 46.520
339789 101.768 -71.665 46.480
339790 101.811 -71.684 46.440
339791 101.853 -71.702 46.399
339792 101.896 -71.721 46.358
339793 101.939 -71.739 46.318
339794 101.958 -71.743 46.277
339795 101.960 -71.735 46.237
339796 101.962 -71.728 46.196
339797 101.964 -71.720 46.156
339798 101.973 -71.715 46.117
339799 102.003 -71.716 46.078
339800 102.033 -71.718 46.039
339801 102.063 -71.719 46.003
339802 102.093 -71.721 45.968
339803 102.123 -71.722 45.931
339804 102.153 -71.724 45.896
339805 102.183 -71.726 45.861
339806 102.213 -71.727 45.827
339807 102.243 -71.729 45.794
339808 102.273 -71.731 45.761
339809 102.303 -71.732 45.728
340200 91.805 -70.321 55.207
340201 91.836 -70.331 55.200
340202 91.888 -70.356 55.194
340203 91.947 -70.386 55.188
340204 91.998 -70.444 55.182
340205 92.049 -70.431 55.174
340206 92.100 -70.406 55.166
340207 92.150 -70.382 55.159
340208 92.201 -70.358 55.152
340209 92.252 -70.333 55.145
340210 92.304 -70.296 55.138
340211 92.343 -70.284 55.131
340212 92.378 -70.285 55.125
340213 92.414 -70.285 55.121
340214 92.449 -70.284 55.119
340215 92.485 -70.284 55.117
340216 92.520 -70.284 55.113
340217 92.556 -70.284 55.107
340218 92.591 -70.283 55.099
340219 92.627 -70.283 55.088
340220 92.668 -70.279 55.071
340221 92.712 -70.274 55.049
340222 92.756 -70.269 55.024
340223 92.801 -70.268 54.998
340224 92.847 -70.268 54.966
340225 92.878 -70.257 54.931
340226 92.932 -70.244 54.890
340227 92.989 -70.231 54.841
340228 93.038 -70.213 54.791
340229 93.074 -70.186 54.741
340230 93.111 -70.159 54.689
340231 93.144 -70.137 54.635
340232 93.174 -70.130 54.574
340233 93.202 -70.122 54.519
340234 93.231 -70.115 54.468
340235 93.260 -70.107 54.420
340236 93.289 -70.099 54.376
340237 93.318 -70.092 54.334
340238 93.347 -70.084 54.293
340239 93.375 -70.077 54.254
340240 93.405 -70.069 54.215
340241 93.433 -70.061 54.179
340242 93.462 -70.053 54.143
340243 93.491 -70.046 54.107
340244 93.520 -70.038 54.071
340286 94.855 -69.964 52.856
340287 94.888 -69.965 52.844
340288 94.921 -69.966 52.834
340289 94.954 -69.966 52.827
340290 94.987 -69.967 52.823
340291 95.020 -69.968 52.820
340292 95.053 -69.968 52.820
340293 95.086 -69.969 52.821
340294 95.119 -69.970 52.826
340295 95.152 -69.971 52.834
340296 95.185 -69.971 52.847
340297 95.218 -69.972 52.867
340298 95.251 -69.973 52.890
340299 95.284 -69.974 52.911
340300 95.321 -69.970 52.922
340301 95.356 -69.961 52.914
340302 95.391 -69.953 52.906
340303 95.426 -69.946 52.896
340304 95.462 -69.937 52.885
340305 95.497 -69.929 52.873
340306 95.533 -69.921 52.862
340307 95.567 -69.914 52.853
340308 95.586 -69.918 52.847
340309 95.605 -69.924 52.843
340310 95.623 -69.928 52.844
340311 95.642 -69.933 52.843
340312 95.674 -69.945 52.836
340313 95.702 -69.957 52.820
340314 95.723 -69.972 52.796
340315 95.744 -69.986 52.771
340316 95.770 -69.996 52.745
340317 95.796 -70.005 52.719
340318 95.822 -70.015 52.695
340319 95.848 -70.024 52.673
340320 95.875 -70.034 52.654
340321 95.901 -70.044 52.638
340322 95.926 -70.053 52.623
340323 95.953 -70.063 52.605
340324 95.984 -70.068 52.586
340325 96.015 -70.074 52.567
340326 96.045 -70.079 52.548
340327 96.076 -70.084 52.521
340328 96.107 -70.090 52.494
340329 96.137 -70.096 52.467
340330 96.167 -70.106 52.444
340331 96.196 -70.118 52.422
340332 96.225 -70.130 52.398
340333 96.253 -70.142 52.376
340334 96.283 -70.154 52.357
340335 96.315 -70.165 52.340
340336 96.350 -70.174 52.325
340337 96.386 -70.183 52.312
340338 96.421 -70.192 52.300
340339 96.457 -70.202 52.290
340340 96.493 -70.211 52.281
340341 96.532 -70.212 52.274
340342 96.572 -70.210 52.266
340343 96.613 -70.209 52.257
340344 96.654 -70.208 52.245
340345 96.697 -70.215 52.231
340346 96.739 -70.225 52.216
340347 96.777 -70.240 52.202
340348 96.815 -70.254 52.189
340349 96.854 -70.269 52.186
340350 96.891 -70.284 52.188
340351 96.930 -70.299 52.191
340352 96.967 -70.314 52.190
340353 97.006 -70.324 52.184
340354 97.046 -70.335 52.174
340355 97.086 -70.347 52.160
340356 97.126 -70.359 52.142
340357 97.165 -70.367 52.118
340358 97.203 -70.375 52.094
340359 97.239 -70.384 52.071
340360 97.263 -70.402 52.050
340361 97.287 -70.420 52.023
340362 97.311 -70.438 51.993
340363 97.336 -70.456 51.963
340364 97.365 -70.478 51.935
340365 97.385 -70.494 51.905
340366 97.403 -70.510 51.878
340367 97.421 -70.525 51.847
340368 97.439 -70.539 51.814
340369 97.457 -70.554 51.781
340370 97.477 -70.565 51.745
340371 97.513 -70.560 51.710
340372 97.548 -70.554 51.672
340373 97.583 -70.549 51.631
340374 97.618 -70.543 51.583
340375 97.653 -70.538 51.538
340376 97.687 -70.537 51.494
340377 97.719 -70.538 51.453
340378 97.753 -70.538 51.414
340379 97.788 -70.536 51.374
340380 97.822 -70.535 51.330
340381 97.856 -70.534 51.283
340382 97.890 -70.532 51.234
340383 97.924 -70.531 51.185
340384 97.958 -70.529 51.140
340385 97.993 -70.528 51.096
340386 98.027 -70.526 51.051
340387 98.061 -70.525 51.002
340388 98.095 -70.524 50.956
340389 98.131 -70.523 50.908
340390 98.169 -70.523 50.859
340391 98.207 -70.524 50.806
340392 98.245 -70.525 50.753
340393 98.279 -70.521 50.700
340394 98.313 -70.516 50.648
340395 98.348 -70.517 50.599
340396 98.383 -70.524 50.549
340397 98.420 -70.531 50.498
340398 98.457 -70.536 50.444
340399 98.497 -70.538 50.388
340400 98.537 -70.541 50.329
340401 98.577 -70.544 50.267
340402 98.617 -70.547 50.205
340403 98.655 -70.557 50.143
340404 98.693 -70.566 50.081
340405 98.731 -70.575 50.020
340406 98.770 -70.584 49.959
340407 98.807 -70.594 49.899
340408 98.846 -70.603 49.841
340409 98.884 -70.612 49.784
340410 98.925 -70.624 49.726
340411 98.966 -70.636 49.666
340412 99.006 -70.649 49.607
340413 99.042 -70.660 49.549
340414 99.079 -70.670 49.493
340415 99.113 -70.685 49.440
340416 99.144 -70.701 49.387
340417 99.177 -70.718 49.336
340418 99.208 -70.735 49.285
340419 99.240 -70.751 49.236
340420 99.272 -70.768 49.188
340421 99.304 -70.785 49.141
340422 99.336 -70.801 49.095
340423 99.368 -70.818 49.050
340424 99.400 -70.835 49.006
340425 99.432 -70.848 48.963
340426 99.465 -70.859 48.924
340427 99.497 -70.871 48.885
340428 99.531 -70.880 48.846
340429 99.566 -70.883 48.809
340430 99.601 -70.886 48.773
340431 99.636 -70.889 48.737
340432 99.671 -70.892 48.703
340433 99.706 -70.895 48.673
340434 99.741 -70.898 48.640
340435 99.776 -70.901 48.609
340436 99.813 -70.908 48.579
340437 99.848 -70.916 48.549
340438 99.880 -70.926 48.516
340439 99.915 -70.942 48.478
340440 99.952 -70.963 48.438
340441 99.988 -70.983 48.396
340442 100.025 -71.003 48.353
340443 100.062 -71.022 48.308
340444 100.101 -71.041 48.265
340445 100.139 -71.060 48.222
340446 100.177 -71.079 48.179
340447 100.215 -71.098 48.135
340448 100.254 -71.117 48.092
340449 100.292 -71.136 48.049
340450 100.330 -71.155 48.006
340451 100.357 -71.171 47.963
340452 100.381 -71.186 47.922
340453 100.411 -71.199 47.882
340454 100.449 -71.207 47.841
340455 100.486 -71.216 47.800
340456 100.524 -71.224 47.761
340457 100.561 -71.234 47.722
340458 100.599 -71.242 47.683
340459 100.637 -71.251 47.647
340460 100.675 -71.259 47.610
340461 100.712 -71.268 47.574
340462 100.750 -71.277 47.537
340463 100.787 -71.285 47.500
340464 100.825 -71.294 47.465
340465 100.863 -71.303 47.431
340466 100.900 -71.312 47.397
340467 100.959 -71.326 47.364
340468 101.025 -71.343 47.332
340469 101.093 -71.361 47.299
340470 101.129 -71.373 47.267
340471 101.159 -71.384 47.235
340472 101.189 -71.396 47.202
340473 101.211 -71.418 47.170
340474 101.232 -71.443 47.136
340475 101.251 -71.468 47.100
340476 101.271 -71.492 47.063
340477 101.301 -71.513 47.027
340478 101.339 -71.528 46.988
340479 101.378 -71.543 46.947
340480 101.415 -71.553 46.906
340481 101.451 -71.558 46.864
340482 101.487 -71.563 46.822
340483 101.523 -71.569 46.780
340484 101.558 -71.573 46.738
340485 101.594 -71.578 46.697
340486 101.630 -71.584 46.655
340487 101.666 -71.588 46.614
340488 101.702 -71.594 46.573
340489 101.740 -71.603 46.533
340490 101.783 -71.622 46.491
340491 101.825 -71.641 46.450
340492 101.868 -71.659 46.409
340493 101.913 -71.680 46.365
340494 101.944 -71.692 46.323
340495 101.946 -71.685 46.282
340496 101.948 -71.677 46.241
340497 101.950 -71.669 46.202
340498 101.957 -71.663 46.161
340499 101.987 -71.664 46.123
340500 102.017 -71.666 46.084
340501 102.047 -71.668 46.046
340502 102.077 -71.669 46.010
340503 102.107 -71.671 45.974
340504 102.137 -71.672 45.938
340505 102.167 -71.674 45.903
340506 102.197 -71.675 45.868
340507 102.227 -71.677 45.833
340508 102.257 -71.679 45.800
340509 102.287 -71.680 45.767
340901 91.793 -70.289 55.204
340902 91.824 -70.300 55.198
340903 91.882 -70.328 55.192
340904 91.941 -70.359 55.186
340905 91.991 -70.419 55.179
340906 92.042 -70.401 55.172
340907 92.093 -70.377 55.164
340908 92.144 -70.353 55.157
340909 92.194 -70.328 55.150
340910 92.245 -70.303 55.142
340911 92.296 -70.265 55.135
340912 92.336 -70.235 55.126
340913 92.372 -70.234 55.120
340914 92.408 -70.240 55.116
340915 92.444 -70.245 55.114
340916 92.479 -70.245 55.113
340917 92.514 -70.244 55.110
340918 92.550 -70.244 55.104
340919 92.585 -70.243 55.097
340920 92.621 -70.243 55.089
340921 92.661 -70.240 55.074
340922 92.705 -70.235 55.054
340923 92.748 -70.231 55.030
340924 92.792 -70.225 55.005
340925 92.838 -70.225 54.975
340926 92.877 -70.219 54.942
340927 92.926 -70.206 54.904
340928 92.982 -70.193 54.858
340929 93.039 -70.181 54.809
340930 93.076 -70.155 54.759
340931 93.113 -70.127 54.708
340932 93.149 -70.100 54.651
340933 93.181 -70.084 54.590
340934 93.210 -70.077 54.532
340935 93.238 -70.069 54.479
340936 93.268 -70.062 54.430
340937 93.296 -70.054 54.386
340938 93.325 -70.046 54.343
340939 93.354 -70.039 54.302
340940 93.383 -70.031 54.264
340941 93.412 -70.023 54.225
340942 93.441 -70.015 54.189
340943 93.470 -70.008 54.152
340944 93.499 -70.000 54.116
340945 93.528 -69.993 54.080
340946 93.556 -69.985 54.044
340986 94.842 -69.944 52.867
340987 94.876 -69.945 52.854
340988 94.908 -69.946 52.843
340989 94.942 -69.946 52.833
340990 94.974 -69.947 52.826
340991 95.007 -69.948 52.822
340992 95.040 -69.949 52.819
340993 95.073 -69.949 52.820
340994 95.106 -69.949 52.822
340995 95.139 -69.950 52.827
340996 95.172 -69.951 52.837
340997 95.204 -69.948 52.854
340998 95.235 -69.945 52.877
340999 95.267 -69.942 52.905
341000 95.299 -69.939 52.930
341001 95.333 -69.934 52.943
341002 95.370 -69.930 52.937
341003 95.406 -69.925 52.926
341004 95.442 -69.921 52.914
341005 95.478 -69.916 52.901
341006 95.514 -69.909 52.887
341007 95.549 -69.901 52.875
341008 95.585 -69.893 52.864
341009 95.610 -69.893 52.856
341010 95.628 -69.898 52.851
341011 95.647 -69.902 52.849
341012 95.666 -69.908 52.846
341013 95.684 -69.911 52.839
341014 95.710 -69.921 52.825
341015 95.737 -69.930 52.801
341016 95.763 -69.940 52.777
341017 95.789 -69.950 52.754
341018 95.815 -69.960 52.730
341019 95.841 -69.970 52.707
341020 95.867 -69.979 52.685
341021 95.893 -69.989 52.665
341022 95.919 -69.999 52.649
341023 95.945 -70.008 52.633
341024 95.971 -70.018 52.614
341025 96.000 -70.026 52.597
341026 96.031 -70.031 52.579
341027 96.061 -70.037 52.559
341028 96.092 -70.042 52.532
341029 96.123 -70.047 52.505
341030 96.152 -70.059 52.479
341031 96.181 -70.071 52.458
341032 96.209 -70.083 52.436
341033 96.238 -70.096 52.411
341034 96.267 -70.107 52.387
341035 96.296 -70.119 52.368
341036 96.328 -70.130 52.350
341037 96.364 -70.139 52.335
341038 96.399 -70.148 52.322
341039 96.435 -70.158 52.310
341040 96.471 -70.167 52.299
341041 96.509 -70.169 52.290
341042 96.550 -70.168 52.283
341043 96.590 -70.166 52.275
341044 96.630 -70.164 52.266
341045 96.671 -70.162 52.254
341046 96.709 -70.172 52.241
341047 96.747 -70.187 52.227
341048 96.785 -70.203 52.212
341049 96.823 -70.218 52.197
341050 96.861 -70.233 52.193
341051 96.899 -70.248 52.194
341052 96.937 -70.263 52.196
341053 96.975 -70.278 52.194
341054 97.015 -70.290 52.186
341055 97.055 -70.301 52.172
341056 97.096 -70.313 52.154
341057 97.137 -70.325 52.133
341058 97.176 -70.335 52.108
341059 97.214 -70.344 52.085
341060 97.247 -70.355 52.064
341061 97.271 -70.373 52.044
341062 97.295 -70.391 52.017
341063 97.319 -70.409 51.989
341064 97.343 -70.426 51.961
341065 97.368 -70.446 51.934
341066 97.391 -70.463 51.906
341067 97.409 -70.478 51.877
341068 97.427 -70.494 51.844
341069 97.448 -70.504 51.809
341070 97.469 -70.513 51.774
341071 97.491 -70.522 51.737
341072 97.525 -70.517 51.702
341073 97.561 -70.512 51.665
341074 97.595 -70.507 51.623
341075 97.631 -70.500 51.575
341076 97.666 -70.495 51.527
341077 97.699 -70.497 51.482
341078 97.731 -70.499 51.441
341079 97.765 -70.498 51.402
341080 97.799 -70.497 51.364
341081 97.833 -70.495 51.322
341082 97.867 -70.494 51.276
341083 97.901 -70.493 51.227
341084 97.936 -70.491 51.177
341085 97.970 -70.490 51.128
341086 98.004 -70.488 51.080
341087 98.038 -70.487 51.033
341088 98.073 -70.485 50.986
341089 98.107 -70.485 50.941
341090 98.141 -70.483 50.897
341091 98.175 -70.481 50.850
341092 98.210 -70.480 50.800
341093 98.244 -70.479 50.746
341094 98.283 -70.480 50.691
341095 98.320 -70.480 50.636
341096 98.354 -70.475 50.582
341097 98.389 -70.478 50.531
341098 98.425 -70.485 50.479
341099 98.464 -70.488 50.425
341100 98.504 -70.491 50.369
341101 98.544 -70.496 50.310
341102 98.582 -70.505 50.249
341103 98.620 -70.514 50.189
341104 98.658 -70.523 50.129
341105 98.697 -70.532 50.069
341106 98.735 -70.541 50.010
341107 98.773 -70.550 49.952
341108 98.811 -70.560 49.894
341109 98.849 -70.569 49.836
341110 98.888 -70.579 49.779
341111 98.929 -70.591 49.721
341112 98.970 -70.604 49.662
341113 99.010 -70.616 49.603
341114 99.048 -70.628 49.546
341115 99.085 -70.639 49.491
341116 99.121 -70.650 49.437
341117 99.155 -70.663 49.386
341118 99.187 -70.680 49.334
341119 99.219 -70.697 49.284
341120 99.251 -70.713 49.236
341121 99.283 -70.730 49.188
341122 99.315 -70.747 49.141
341123 99.347 -70.764 49.096
341124 99.379 -70.780 49.052
341125 99.411 -70.797 49.009
341126 99.443 -70.809 48.967
341127 99.476 -70.820 48.929
341128 99.508 -70.832 48.891
341129 99.541 -70.843 48.853
341130 99.576 -70.846 48.817
341131 99.611 -70.849 48.781
341132 99.646 -70.852 48.747
341133 99.682 -70.855 48.716
341134 99.716 -70.858 48.686
341135 99.752 -70.861 48.655
341136 99.789 -70.868 48.624
341137 99.826 -70.876 48.594
341138 99.860 -70.886 48.565
341139 99.892 -70.895 48.530
341140 99.924 -70.906 48.492
341141 99.959 -70.921 48.451
341142 99.995 -70.941 48.409
341143 100.032 -70.961 48.367
341144 100.068 -70.982 48.323
341145 100.105 -71.002 48.281
341146 100.141 -71.022 48.240
341147 100.177 -71.042 48.198
341148 100.215 -71.062 48.155
341149 100.253 -71.081 48.113
341150 100.291 -71.099 48.071
341151 100.326 -71.118 48.028
341152 100.354 -71.131 47.986
341153 100.392 -71.140 47.945
341154 100.429 -71.149 47.904
341155 100.467 -71.157 47.863
341156 100.505 -71.166 47.822
341157 100.543 -71.174 47.782
341158 100.580 -71.183 47.743
341159 100.618 -71.192 47.705
341160 100.655 -71.201 47.669
341161 100.693 -71.209 47.632
341162 100.731 -71.218 47.595
341163 100.768 -71.227 47.559
341164 100.806 -71.236 47.523
341165 100.844 -71.244 47.488
341166 100.882 -71.253 47.454
341167 100.919 -71.262 47.420
341168 100.966 -71.273 47.388
341169 101.033 -71.290 47.356
341170 101.092 -71.306 47.324
341171 101.122 -71.318 47.292
341172 101.152 -71.329 47.260
341173 101.182 -71.340 47.226
341174 101.211 -71.353 47.196
341175 101.231 -71.378 47.162
341176 101.251 -71.403 47.127
341177 101.271 -71.428 47.090
341178 101.308 -71.444 47.052
341179 101.346 -71.459 47.013
341180 101.385 -71.474 46.972
341181 101.423 -71.489 46.930
341182 101.462 -71.504 46.886
341183 101.500 -71.518 46.843
341184 101.536 -71.523 46.799
341185 101.571 -71.528 46.756
341186 101.608 -71.533 46.714
341187 101.643 -71.538 46.671
341188 101.679 -71.544 46.629
341189 101.715 -71.548 46.588
341190 101.754 -71.560 46.548
341191 101.797 -71.579 46.504
341192 101.839 -71.597 46.461
341193 101.889 -71.622 46.416
341194 101.934 -71.645 46.371
341195 101.933 -71.633 46.329
341196 101.935 -71.625 46.288
341197 101.936 -71.618 46.247
341198 101.941 -71.611 46.208
341199 101.971 -71.613 46.168
341200 102.000 -71.614 46.130
341201 102.031 -71.616 46.091
341202 102.060 -71.617 46.053
341203 102.091 -71.619 46.017
341204 102.121 -71.620 45.980
341205 102.150 -71.622 45.944
341206 102.181 -71.623 45.909
341207 102.210 -71.625 45.875
341208 102.241 -71.626 45.840
341209 102.270 -71.628 45.807
341602 91.781 -70.256 55.203
341603 91.818 -70.271 55.196
341604 91.876 -70.300 55.191
341605 91.934 -70.334 55.185
341606 91.985 -70.394 55.178
341607 92.035 -70.372 55.170
341608 92.086 -70.347 55.162
341609 92.137 -70.323 55.155
341610 92.188 -70.299 55.147
341611 92.238 -70.271 55.139
341612 92.287 -70.234 55.132
341613 92.326 -70.205 55.123
341614 92.365 -70.178 55.115
341615 92.402 -70.184 55.109
341616 92.437 -70.190 55.104
341617 92.473 -70.196 55.102
341618 92.509 -70.203 55.099
341619 92.545 -70.205 55.094
341620 92.580 -70.204 55.089
341621 92.616 -70.204 55.082
341622 92.653 -70.202 55.070
341623 92.697 -70.197 55.051
341624 92.741 -70.192 55.027
341625 92.785 -70.187 55.003
341626 92.829 -70.182 54.977
341627 92.873 -70.180 54.949
341628 92.919 -70.168 54.916
341629 92.976 -70.156 54.874
341630 93.032 -70.143 54.825
341631 93.078 -70.123 54.776
341632 93.115 -70.096 54.723
341633 93.151 -70.068 54.665
341634 93.188 -70.041 54.604
341635 93.217 -70.031 54.545
341636 93.246 -70.024 54.489
341637 93.275 -70.016 54.439
341638 93.304 -70.008 54.395
341639 93.333 -70.001 54.352
341640 93.362 -69.993 54.311
341641 93.391 -69.985 54.273
341642 93.419 -69.977 54.235
341643 93.448 -69.970 54.198
341644 93.477 -69.962 54.162
341645 93.506 -69.955 54.125
341646 93.535 -69.947 54.090
341647 93.564 -69.939 54.054
341687 94.862 -69.922 52.869
341688 94.894 -69.920 52.856
341689 94.925 -69.917 52.844
341690 94.957 -69.914 52.834
341691 94.988 -69.911 52.827
341692 95.020 -69.908 52.823
341693 95.051 -69.905 52.820
341694 95.083 -69.902 52.820
341695 95.114 -69.899 52.823
341696 95.146 -69.895 52.830
341697 95.177 -69.892 52.840
341698 95.209 -69.889 52.858
341699 95.241 -69.886 52.882
341700 95.272 -69.883 52.911
341701 95.305 -69.880 52.938
341702 95.341 -69.876 52.955
341703 95.377 -69.871 52.956
341704 95.413 -69.866 52.948
341705 95.450 -69.861 52.935
341706 95.485 -69.857 52.920
341707 95.522 -69.852 52.904
341708 95.558 -69.850 52.888
341709 95.594 -69.855 52.876
341710 95.629 -69.859 52.866
341711 95.649 -69.863 52.858
341712 95.667 -69.867 52.855
341713 95.686 -69.871 52.850
341714 95.704 -69.875 52.842
341715 95.729 -69.883 52.829
341716 95.755 -69.891 52.806
341717 95.781 -69.899 52.784
341718 95.808 -69.907 52.763
341719 95.833 -69.915 52.744
341720 95.860 -69.924 52.720
341721 95.886 -69.934 52.696
341722 95.912 -69.944 52.676
341723 95.938 -69.953 52.660
341724 95.964 -69.963 52.643
341725 95.990 -69.973 52.625
341726 96.016 -69.983 52.607
341727 96.047 -69.988 52.590
341728 96.078 -69.993 52.569
341729 96.108 -70.000 52.542
341730 96.136 -70.012 52.515
341731 96.166 -70.024 52.491
341732 96.194 -70.037 52.472
341733 96.223 -70.049 52.449
341734 96.252 -70.061 52.424
341735 96.281 -70.072 52.399
341736 96.310 -70.084 52.378
341737 96.342 -70.095 52.361
341738 96.378 -70.104 52.346
341739 96.413 -70.114 52.332
341740 96.449 -70.123 52.319
341741 96.487 -70.127 52.308
341742 96.527 -70.125 52.299
341743 96.568 -70.123 52.291
341744 96.608 -70.121 52.282
341745 96.645 -70.125 52.273
341746 96.680 -70.132 52.263
341747 96.718 -70.143 52.250
341748 96.755 -70.159 52.236
341749 96.794 -70.174 52.221
341750 96.832 -70.189 52.205
341751 96.870 -70.204 52.199
341752 96.908 -70.219 52.198
341753 96.946 -70.232 52.199
341754 96.982 -70.240 52.195
341755 97.022 -70.252 52.184
341756 97.064 -70.265 52.168
341757 97.106 -70.279 52.146
341758 97.147 -70.291 52.123
341759 97.187 -70.303 52.100
341760 97.225 -70.312 52.079
341761 97.254 -70.325 52.060
341762 97.279 -70.344 52.040
341763 97.302 -70.362 52.013
341764 97.326 -70.379 51.987
341765 97.351 -70.397 51.961
341766 97.374 -70.415 51.935
341767 97.397 -70.433 51.906
341768 97.418 -70.443 51.876
341769 97.440 -70.452 51.841
341770 97.461 -70.460 51.803
341771 97.483 -70.469 51.767
341772 97.505 -70.478 51.729
341773 97.538 -70.475 51.694
341774 97.573 -70.469 51.657
341775 97.608 -70.464 51.615
341776 97.643 -70.458 51.569
341777 97.677 -70.456 51.521
341778 97.710 -70.457 51.474
341779 97.742 -70.460 51.431
341780 97.776 -70.459 51.392
341781 97.810 -70.457 51.354
341782 97.845 -70.456 51.313
341783 97.879 -70.454 51.268
341784 97.913 -70.453 51.220
341785 97.947 -70.452 51.169
341786 97.982 -70.450 51.118
341787 98.016 -70.449 51.068
341788 98.050 -70.447 51.019
341789 98.084 -70.446 50.971
341790 98.119 -70.445 50.927
341791 98.153 -70.444 50.884
341792 98.186 -70.442 50.840
341793 98.221 -70.441 50.790
341794 98.255 -70.439 50.737
341795 98.289 -70.438 50.682
341796 98.323 -70.437 50.625
341797 98.358 -70.435 50.569
341798 98.394 -70.434 50.515
341799 98.432 -70.438 50.462
341800 98.471 -70.444 50.408
341801 98.509 -70.453 50.352
341802 98.547 -70.463 50.293
341803 98.585 -70.472 50.234
341804 98.624 -70.481 50.176
341805 98.662 -70.490 50.118
341806 98.700 -70.499 50.061
341807 98.738 -70.508 50.003
341808 98.776 -70.517 49.946
341809 98.814 -70.526 49.889
341810 98.853 -70.535 49.832
341811 98.892 -70.547 49.774
341812 98.933 -70.560 49.716
341813 98.974 -70.572 49.657
341814 99.014 -70.585 49.599
341815 99.054 -70.597 49.542
341816 99.091 -70.607 49.488
341817 99.127 -70.618 49.435
341818 99.163 -70.629 49.384
341819 99.198 -70.642 49.333
341820 99.230 -70.659 49.283
341821 99.262 -70.676 49.235
341822 99.294 -70.692 49.187
341823 99.326 -70.709 49.142
341824 99.358 -70.726 49.098
341825 99.390 -70.742 49.054
341826 99.422 -70.758 49.012
341827 99.454 -70.770 48.971
341828 99.487 -70.782 48.934
341829 99.519 -70.794 48.897
341830 99.552 -70.805 48.860
341831 99.586 -70.809 48.824
341832 99.622 -70.812 48.790
341833 99.656 -70.815 48.758
341834 99.692 -70.818 48.729
341835 99.727 -70.822 48.700
341836 99.765 -70.829 48.671
341837 99.802 -70.837 48.640
341838 99.839 -70.845 48.609
341839 99.871 -70.855 48.579
341840 99.904 -70.865 48.543
341841 99.936 -70.876 48.505
341842 99.969 -70.886 48.465
341843 100.002 -70.899 48.423
341844 100.039 -70.920 48.381
341845 100.075 -70.940 48.340
341846 100.112 -70.960 48.299
341847 100.148 -70.980 48.259
341848 100.185 -71.001 48.218
341849 100.221 -71.021 48.177
341850 100.258 -71.041 48.135
341851 100.294 -71.061 48.093
341852 100.332 -71.074 48.052
341853 100.371 -71.083 48.010
341854 100.410 -71.090 47.968
341855 100.449 -71.099 47.927
341856 100.486 -71.108 47.886
341857 100.524 -71.116 47.846
341858 100.561 -71.124 47.804
341859 100.599 -71.133 47.764
341860 100.636 -71.142 47.727
341861 100.674 -71.151 47.692
341862 100.712 -71.159 47.655
341863 100.750 -71.168 47.618
341864 100.787 -71.177 47.583
341865 100.825 -71.186 47.548
341866 100.862 -71.194 47.514
341867 100.900 -71.203 47.479
341868 100.938 -71.212 47.445
341869 100.976 -71.221 47.412
341870 101.040 -71.237 47.380
341871 101.085 -71.251 47.347
341872 101.115 -71.262 47.315
341873 101.145 -71.273 47.283
341874 101.176 -71.284 47.249
341875 101.205 -71.296 47.218
341876 101.231 -71.314 47.185
341877 101.251 -71.339 47.150
341878 101.276 -71.361 47.114
341879 101.314 -71.376 47.075
341880 101.353 -71.391 47.034
341881 101.392 -71.406 46.993
341882 101.430 -71.421 46.950
341883 101.469 -71.436 46.906
341884 101.507 -71.451 46.862
341885 101.546 -71.466 46.817
341886 101.585 -71.481 46.773
341887 101.621 -71.488 46.730
341888 101.656 -71.494 46.688
341889 101.692 -71.498 46.645
341890 101.729 -71.503 46.602
341891 101.768 -71.517 46.558
341892 101.811 -71.536 46.513
341893 101.864 -71.565 46.468
341894 101.920 -71.598 46.422
341895 101.918 -71.582 46.377
341896 101.920 -71.574 46.335
341897 101.922 -71.566 46.294
341898 101.924 -71.559 46.254
341899 101.955 -71.560 46.214
341900 101.984 -71.562 46.174
341901 102.015 -71.563 46.136
341902 102.044 -71.565 46.097
341903 102.075 -71.566 46.059
341904 102.105 -71.568 46.023
341905 102.134 -71.569 45.986
341906 102.165 -71.571 45.951
341907 102.194 -71.572 45.915
341908 102.225 -71.575 45.881
341909 102.254 -71.576 45.846
342304 91.812 -70.243 55.196
342305 91.870 -70.272 55.190
342306 91.927 -70.309 55.184
342307 91.978 -70.367 55.177
342308 92.028 -70.343 55.168
342309 92.079 -70.318 55.161
342310 92.130 -70.293 55.153
342311 92.180 -70.269 55.144
342312 92.232 -70.239 55.137
342313 92.278 -70.205 55.128
342314 92.317 -70.174 55.119
342315 92.357 -70.145 55.109
342316 92.395 -70.127 55.101
342317 92.431 -70.134 55.095
342318 92.466 -70.140 55.090
342319 92.502 -70.146 55.085
342320 92.538 -70.152 55.080
342321 92.574 -70.159 55.075
342322 92.610 -70.165 55.070
342323 92.646 -70.164 55.060
342324 92.690 -70.159 55.043
342325 92.733 -70.153 55.021
342326 92.777 -70.149 54.996
342327 92.821 -70.143 54.971
342328 92.865 -70.139 54.944
342329 92.913 -70.131 54.914
342330 92.969 -70.118 54.878
342331 93.026 -70.105 54.834
342332 93.080 -70.092 54.785
342333 93.117 -70.065 54.731
342334 93.153 -70.037 54.673
342335 93.190 -70.010 54.612
342336 93.225 -69.986 54.554
342337 93.254 -69.978 54.500
342338 93.282 -69.971 54.449
342339 93.311 -69.963 54.404
342340 93.340 -69.955 54.361
342341 93.369 -69.947 54.320
342342 93.398 -69.939 54.281
342343 93.427 -69.932 54.244
342344 93.456 -69.924 54.208
342345 93.485 -69.917 54.172
342346 93.513 -69.909 54.135
342347 93.543 -69.902 54.100
342348 93.571 -69.894 54.065
342388 94.867 -69.864 52.873
342389 94.899 -69.861 52.859
342390 94.930 -69.858 52.847
342391 94.962 -69.855 52.836
342392 94.993 -69.852 52.829
342393 95.025 -69.849 52.824
342394 95.056 -69.846 52.821
342395 95.088 -69.843 52.821
342396 95.119 -69.840 52.825
342397 95.151 -69.838 52.832
342398 95.182 -69.835 52.843
342399 95.214 -69.832 52.861
342400 95.246 -69.829 52.885
342401 95.277 -69.826 52.914
342402 95.312 -69.821 52.942
342403 95.348 -69.817 52.961
342404 95.385 -69.812 52.967
342405 95.420 -69.808 52.963
342406 95.457 -69.803 52.952
342407 95.492 -69.805 52.938
342408 95.528 -69.809 52.920
342409 95.564 -69.813 52.903
342410 95.600 -69.817 52.890
342411 95.635 -69.822 52.878
342412 95.669 -69.827 52.869
342413 95.687 -69.830 52.862
342414 95.706 -69.835 52.856
342415 95.725 -69.839 52.849
342416 95.743 -69.842 52.836
342417 95.774 -69.853 52.814
342418 95.801 -69.861 52.795
342419 95.827 -69.870 52.776
342420 95.852 -69.877 52.756
342421 95.879 -69.886 52.733
342422 95.905 -69.894 52.710
342423 95.931 -69.902 52.690
342424 95.957 -69.910 52.674
342425 95.983 -69.918 52.657
342426 96.009 -69.927 52.637
342427 96.035 -69.937 52.618
342428 96.062 -69.946 52.600
342429 96.093 -69.953 52.580
342430 96.121 -69.965 52.554
342431 96.151 -69.977 52.526
342432 96.179 -69.990 52.502
342433 96.208 -70.002 52.483
342434 96.237 -70.014 52.458
342435 96.266 -70.026 52.432
342436 96.295 -70.037 52.408
342437 96.324 -70.049 52.389
342438 96.356 -70.060 52.373
342439 96.391 -70.070 52.357
342440 96.427 -70.079 52.341
342441 96.464 -70.083 52.328
342442 96.505 -70.082 52.318
342443 96.545 -70.080 52.309
342444 96.584 -70.082 52.299
342445 96.619 -70.089 52.289
342446 96.654 -70.096 52.280
342447 96.689 -70.102 52.269
342448 96.726 -70.115 52.257
342449 96.764 -70.130 52.244
342450 96.803 -70.145 52.228
342451 96.840 -70.160 52.211
342452 96.879 -70.175 52.202
342453 96.915 -70.187 52.199
342454 96.952 -70.194 52.198
342455 96.988 -70.202 52.192
342456 97.028 -70.214 52.179
342457 97.071 -70.228 52.160
342458 97.113 -70.242 52.137
342459 97.155 -70.256 52.115
342460 97.197 -70.269 52.096
342461 97.236 -70.280 52.078
342462 97.262 -70.297 52.059
342463 97.286 -70.314 52.040
342464 97.310 -70.332 52.012
342465 97.334 -70.346 51.987
342466 97.358 -70.359 51.962
342467 97.382 -70.373 51.936
342468 97.408 -70.391 51.906
342469 97.432 -70.400 51.874
342470 97.453 -70.409 51.836
342471 97.474 -70.417 51.797
342472 97.496 -70.426 51.760
342473 97.518 -70.435 51.721
342474 97.550 -70.432 51.685
342475 97.586 -70.427 51.648
342476 97.620 -70.421 51.608
342477 97.656 -70.416 51.563
342478 97.689 -70.416 51.517
342479 97.722 -70.418 51.470
342480 97.754 -70.419 51.424
342481 97.788 -70.419 51.382
342482 97.822 -70.418 51.343
342483 97.856 -70.416 51.303
342484 97.890 -70.415 51.259
342485 97.925 -70.414 51.212
342486 97.959 -70.413 51.162
342487 97.993 -70.411 51.111
342488 98.027 -70.409 51.059
342489 98.062 -70.408 51.009
342490 98.096 -70.406 50.958
342491 98.130 -70.405 50.912
342492 98.164 -70.404 50.869
342493 98.198 -70.403 50.827
342494 98.232 -70.401 50.777
342495 98.267 -70.400 50.726
342496 98.301 -70.398 50.671
342497 98.335 -70.397 50.615
342498 98.369 -70.396 50.558
342499 98.403 -70.395 50.502
342500 98.436 -70.402 50.447
342501 98.474 -70.411 50.392
342502 98.512 -70.420 50.337
342503 98.551 -70.429 50.279
342504 98.589 -70.438 50.221
342505 98.627 -70.447 50.166
342506 98.665 -70.456 50.112
342507 98.703 -70.466 50.055
342508 98.741 -70.475 49.999
342509 98.780 -70.484 49.942
342510 98.818 -70.493 49.884
342511 98.856 -70.503 49.828
342512 98.897 -70.515 49.770
342513 98.937 -70.527 49.712
342514 98.978 -70.540 49.654
342515 99.019 -70.552 49.594
342516 99.059 -70.564 49.539
342517 99.097 -70.576 49.485
342518 99.133 -70.586 49.433
342519 99.169 -70.597 49.381
342520 99.206 -70.608 49.331
342521 99.241 -70.621 49.282
342522 99.273 -70.638 49.233
342523 99.305 -70.654 49.187
342524 99.337 -70.671 49.143
342525 99.369 -70.688 49.099
342526 99.401 -70.704 49.056
342527 99.433 -70.720 49.014
342528 99.466 -70.731 48.975
342529 99.498 -70.743 48.938
342530 99.530 -70.754 48.903
342531 99.563 -70.767 48.867
342532 99.597 -70.772 48.833
342533 99.632 -70.775 48.801
342534 99.667 -70.778 48.772
342535 99.703 -70.783 48.743
342536 99.740 -70.791 48.715
342537 99.778 -70.798 48.687
342538 99.816 -70.806 48.656
342539 99.851 -70.815 48.624
342540 99.884 -70.825 48.591
342541 99.916 -70.835 48.556
342542 99.948 -70.845 48.517
342543 99.980 -70.855 48.478
342544 100.012 -70.865 48.437
342545 100.046 -70.878 48.396
342546 100.082 -70.898 48.357
342547 100.119 -70.918 48.318
342548 100.155 -70.939 48.279
342549 100.192 -70.959 48.240
342550 100.228 -70.979 48.198
342551 100.266 -70.999 48.157
342552 100.307 -71.015 48.116
342553 100.346 -71.030 48.075
342554 100.379 -71.039 48.033
342555 100.414 -71.048 47.991
342556 100.454 -71.055 47.951
342557 100.493 -71.064 47.911
342558 100.532 -71.071 47.870
342559 100.571 -71.079 47.830
342560 100.611 -71.087 47.790
342561 100.650 -71.095 47.753
342562 100.689 -71.103 47.717
342563 100.728 -71.111 47.680
342564 100.768 -71.118 47.644
342565 100.806 -71.127 47.609
342566 100.844 -71.136 47.574
342567 100.881 -71.144 47.539
342568 100.919 -71.153 47.505
342569 100.957 -71.162 47.471
342570 100.994 -71.171 47.436
342571 101.047 -71.183 47.404
342572 101.078 -71.195 47.370
342573 101.108 -71.206 47.337
342574 101.138 -71.218 47.305
342575 101.169 -71.229 47.271
342576 101.198 -71.240 47.240
342577 101.229 -71.252 47.207
342578 101.250 -71.274 47.172
342579 101.283 -71.293 47.133
342580 101.321 -71.308 47.093
342581 101.360 -71.322 47.052
342582 101.398 -71.337 47.010
342583 101.437 -71.353 46.968
342584 101.476 -71.368 46.924
342585 101.514 -71.383 46.879
342586 101.553 -71.398 46.833
342587 101.591 -71.413 46.788
342588 101.630 -71.428 46.744
342589 101.669 -71.443 46.702
342590 101.706 -71.453 46.658
342591 101.742 -71.458 46.613
342592 101.783 -71.475 46.566
342593 101.840 -71.507 46.520
342594 101.896 -71.541 46.473
342595 101.904 -71.530 46.427
342596 101.906 -71.522 46.383
342597 101.908 -71.515 46.340
342598 101.910 -71.507 46.300
342599 101.939 -71.508 46.260
342600 101.968 -71.509 46.220
342601 101.998 -71.511 46.180
342602 102.028 -71.513 46.141
342603 102.058 -71.515 46.103
342604 102.088 -71.516 46.065
342605 102.118 -71.518 46.029
342606 102.149 -71.519 45.993
342607 102.178 -71.521 45.957
342608 102.208 -71.522 45.921
342609 102.238 -71.524 45.886
343006 91.864 -70.244 55.190
343007 91.920 -70.284 55.184
343008 91.970 -70.337 55.176
343009 92.021 -70.313 55.168
343010 92.072 -70.289 55.160
343011 92.123 -70.265 55.151
343012 92.173 -70.240 55.142
343013 92.225 -70.208 55.134
343014 92.269 -70.174 55.125
343015 92.308 -70.145 55.114
343016 92.348 -70.115 55.104
343017 92.387 -70.085 55.094
343018 92.424 -70.077 55.086
343019 92.460 -70.083 55.077
343020 92.496 -70.090 55.071
343021 92.531 -70.096 55.065
343022 92.568 -70.102 55.060
343023 92.603 -70.109 55.055
343024 92.639 -70.115 55.046
343025 92.684 -70.114 55.031
343026 92.727 -70.109 55.012
343027 92.771 -70.103 54.987
343028 92.815 -70.098 54.961
343029 92.859 -70.092 54.936
343030 92.910 -70.087 54.908
343031 92.963 -70.080 54.876
343032 93.019 -70.068 54.834
343033 93.076 -70.055 54.788
343034 93.119 -70.033 54.736
343035 93.155 -70.005 54.679
343036 93.192 -69.978 54.618
343037 93.228 -69.951 54.561
343038 93.261 -69.933 54.508
343039 93.290 -69.924 54.456
343040 93.319 -69.917 54.411
343041 93.348 -69.909 54.368
343042 93.376 -69.902 54.328
343043 93.406 -69.894 54.289
343044 93.434 -69.886 54.253
343045 93.464 -69.879 54.216
343046 93.492 -69.871 54.181
343047 93.521 -69.864 54.146
343048 93.550 -69.857 54.110
343049 93.583 -69.858 54.075
343088 94.841 -69.809 52.893
343089 94.872 -69.806 52.879
343090 94.904 -69.804 52.864
343091 94.935 -69.801 52.851
343092 94.967 -69.798 52.840
343093 94.998 -69.795 52.832
343094 95.030 -69.792 52.827
343095 95.061 -69.789 52.824
343096 95.093 -69.786 52.824
343097 95.124 -69.783 52.827
343098 95.156 -69.780 52.835
343099 95.188 -69.776 52.846
343100 95.219 -69.773 52.863
343101 95.251 -69.770 52.888
343102 95.283 -69.767 52.916
343103 95.320 -69.763 52.944
343104 95.356 -69.758 52.965
343105 95.392 -69.754 52.974
343106 95.427 -69.758 52.975
343107 95.463 -69.763 52.967
343108 95.499 -69.767 52.953
343109 95.535 -69.772 52.935
343110 95.570 -69.776 52.919
343111 95.606 -69.781 52.905
343112 95.641 -69.785 52.892
343113 95.677 -69.789 52.881
343114 95.707 -69.794 52.872
343115 95.726 -69.798 52.865
343116 95.745 -69.802 52.858
343117 95.763 -69.806 52.845
343118 95.788 -69.813 52.825
343119 95.819 -69.823 52.807
343120 95.846 -69.832 52.790
343121 95.872 -69.840 52.771
343122 95.898 -69.848 52.748
343123 95.924 -69.856 52.724
343124 95.950 -69.864 52.703
343125 95.976 -69.873 52.687
343126 96.002 -69.881 52.670
343127 96.028 -69.889 52.649
343128 96.054 -69.897 52.629
343129 96.080 -69.905 52.611
343130 96.107 -69.918 52.591
343131 96.135 -69.930 52.564
343132 96.164 -69.942 52.536
343133 96.193 -69.955 52.511
343134 96.222 -69.967 52.492
343135 96.251 -69.979 52.467
343136 96.280 -69.991 52.442
343137 96.309 -70.002 52.419
343138 96.337 -70.015 52.402
343139 96.369 -70.026 52.386
343140 96.405 -70.035 52.369
343141 96.442 -70.041 52.353
343142 96.482 -70.040 52.339
343143 96.522 -70.039 52.328
343144 96.557 -70.046 52.319
343145 96.592 -70.053 52.308
343146 96.627 -70.060 52.297
343147 96.663 -70.067 52.288
343148 96.698 -70.074 52.276
343149 96.735 -70.087 52.264
343150 96.773 -70.102 52.250
343151 96.811 -70.116 52.234
343152 96.850 -70.131 52.216
343153 96.886 -70.140 52.205
343154 96.922 -70.149 52.199
343155 96.958 -70.156 52.195
343156 96.995 -70.164 52.186
343157 97.034 -70.175 52.172
343158 97.077 -70.190 52.154
343159 97.119 -70.204 52.135
343160 97.161 -70.218 52.117
343161 97.204 -70.232 52.101
343162 97.245 -70.243 52.084
343163 97.270 -70.254 52.066
343164 97.294 -70.267 52.046
343165 97.318 -70.280 52.018
343166 97.343 -70.293 51.991
343167 97.366 -70.309 51.965
343168 97.388 -70.334 51.937
343169 97.416 -70.347 51.907
343170 97.445 -70.356 51.875
343171 97.466 -70.365 51.837
343172 97.488 -70.374 51.796
343173 97.510 -70.383 51.756
343174 97.531 -70.391 51.717
343175 97.563 -70.390 51.679
343176 97.598 -70.384 51.640
343177 97.634 -70.378 51.601
343178 97.668 -70.374 51.557
343179 97.701 -70.376 51.512
343180 97.733 -70.378 51.467
343181 97.766 -70.380 51.422
343182 97.799 -70.380 51.379
343183 97.833 -70.378 51.339
343184 97.868 -70.377 51.297
343185 97.902 -70.375 51.252
343186 97.936 -70.374 51.205
343187 97.970 -70.373 51.155
343188 98.005 -70.372 51.104
343189 98.038 -70.370 51.054
343190 98.073 -70.369 51.004
343191 98.107 -70.367 50.952
343192 98.141 -70.366 50.904
343193 98.175 -70.365 50.858
343194 98.210 -70.363 50.812
343195 98.244 -70.362 50.762
343196 98.278 -70.360 50.710
343197 98.312 -70.359 50.656
343198 98.347 -70.357 50.601
343199 98.380 -70.357 50.546
343200 98.410 -70.365 50.490
343201 98.440 -70.370 50.436
343202 98.478 -70.378 50.382
343203 98.516 -70.387 50.328
343204 98.554 -70.396 50.273
343205 98.592 -70.405 50.218
343206 98.630 -70.414 50.164
343207 98.668 -70.423 50.109
343208 98.707 -70.432 50.052
343209 98.745 -70.441 49.995
343210 98.783 -70.450 49.938
343211 98.821 -70.460 49.880
343212 98.860 -70.470 49.822
343213 98.901 -70.482 49.764
343214 98.942 -70.495 49.706
343215 98.982 -70.507 49.648
343216 99.023 -70.519 49.591
343217 99.064 -70.532 49.536
343218 99.103 -70.544 49.483
343219 99.139 -70.555 49.431
343220 99.176 -70.566 49.380
343221 99.212 -70.576 49.330
343222 99.249 -70.588 49.281
343223 99.284 -70.600 49.233
343224 99.316 -70.616 49.187
343225 99.348 -70.633 49.143
343226 99.380 -70.650 49.100
343227 99.412 -70.666 49.057
343228 99.444 -70.680 49.017
343229 99.476 -70.692 48.979
343230 99.509 -70.704 48.943
343231 99.541 -70.716 48.908
343232 99.573 -70.728 48.874
343233 99.607 -70.735 48.841
343234 99.642 -70.739 48.810
343235 99.679 -70.745 48.781
343236 99.716 -70.752 48.754
343237 99.754 -70.760 48.728
343238 99.791 -70.767 48.701
343239 99.829 -70.775 48.670
343240 99.863 -70.785 48.639
343241 99.895 -70.795 48.606
343242 99.927 -70.804 48.569
343243 99.960 -70.815 48.529
343244 99.992 -70.825 48.489
343245 100.025 -70.835 48.450
343246 100.057 -70.845 48.411
343247 100.089 -70.857 48.376
343248 100.126 -70.877 48.337
343249 100.162 -70.897 48.299
343250 100.199 -70.917 48.261
343251 100.237 -70.936 48.219
343252 100.279 -70.953 48.179
343253 100.321 -70.970 48.138
343254 100.361 -70.985 48.097
343255 100.394 -70.994 48.056
343256 100.428 -71.003 48.014
343257 100.461 -71.012 47.974
343258 100.497 -71.021 47.935
343259 100.537 -71.029 47.893
343260 100.576 -71.036 47.854
343261 100.615 -71.045 47.815
343262 100.654 -71.052 47.778
343263 100.693 -71.060 47.741
343264 100.733 -71.068 47.704
343265 100.772 -71.076 47.668
343266 100.811 -71.084 47.633
343267 100.850 -71.092 47.599
343268 100.890 -71.099 47.565
343269 100.929 -71.108 47.530
343270 100.968 -71.115 47.495
343271 101.007 -71.124 47.461
343272 101.045 -71.132 47.427
343273 101.074 -71.143 47.392
343274 101.104 -71.153 47.359
343275 101.133 -71.164 47.325
343276 101.162 -71.174 47.292
343277 101.192 -71.184 47.260
343278 101.221 -71.196 47.227
343279 101.251 -71.209 47.191
343280 101.289 -71.224 47.151
343281 101.328 -71.240 47.109
343282 101.367 -71.254 47.066
343283 101.406 -71.270 47.025
343284 101.445 -71.286 46.983
343285 101.484 -71.303 46.939
343286 101.523 -71.318 46.894
343287 101.562 -71.335 46.847
343288 101.602 -71.351 46.800
343289 101.640 -71.368 46.756
343290 101.680 -71.384 46.713
343291 101.719 -71.400 46.667
343292 101.759 -71.417 46.619
343293 101.815 -71.450 46.572
343294 101.872 -71.483 46.524
343295 101.894 -71.483 46.477
343296 101.893 -71.472 46.431
343297 101.894 -71.463 46.388
343298 101.896 -71.456 46.345
343299 101.922 -71.456 46.305
343300 101.952 -71.458 46.265
343301 101.982 -71.459 46.225
343302 102.012 -71.461 46.186
343303 102.042 -71.462 46.146
343304 102.072 -71.464 46.108
343305 102.102 -71.466 46.070
343306 102.132 -71.467 46.033
343307 102.162 -71.469 45.997
343308 102.192 -71.470 45.961
343309 102.222 -71.472 45.925
343708 91.913 -70.260 55.183
343709 91.963 -70.308 55.176
343710 92.014 -70.284 55.167
343711 92.065 -70.259 55.158
343712 92.116 -70.235 55.149
343713 92.166 -70.211 55.140
343714 92.218 -70.176 55.130
343715 92.260 -70.144 55.120
343716 92.299 -70.115 55.110
343717 92.339 -70.085 55.099
343718 92.378 -70.055 55.087
343719 92.417 -70.026 55.077
343720 92.453 -70.027 55.067
343721 92.489 -70.033 55.057
343722 92.525 -70.040 55.049
343723 92.561 -70.046 55.042
343724 92.597 -70.052 55.038
343725 92.633 -70.058 55.029
343726 92.677 -70.065 55.015
343727 92.721 -70.060 54.996
343728 92.765 -70.055 54.975
343729 92.809 -70.049 54.950
343730 92.856 -70.043 54.925
343731 92.909 -70.038 54.899
343732 92.961 -70.032 54.868
343733 93.015 -70.027 54.827
343734 93.069 -70.018 54.784
343735 93.121 -70.002 54.738
343736 93.157 -69.974 54.682
343737 93.194 -69.947 54.622
343738 93.230 -69.920 54.566
343739 93.267 -69.892 54.513
343740 93.297 -69.879 54.462
343741 93.327 -69.871 54.416
343742 93.355 -69.864 54.375
343743 93.384 -69.856 54.335
343744 93.413 -69.848 54.297
343745 93.442 -69.841 54.260
343746 93.472 -69.835 54.224
343747 93.505 -69.836 54.189
343748 93.538 -69.836 54.155
343749 93.571 -69.837 54.119
343750 93.604 -69.838 54.085
343789 94.846 -69.751 52.899
343790 94.877 -69.748 52.884
343791 94.909 -69.745 52.869
343792 94.941 -69.742 52.856
343793 94.972 -69.739 52.844
343794 95.003 -69.736 52.837
343795 95.035 -69.733 52.832
343796 95.066 -69.730 52.829
343797 95.098 -69.727 52.828
343798 95.130 -69.725 52.832
343799 95.161 -69.722 52.838
343800 95.193 -69.719 52.848
343801 95.224 -69.716 52.866
343802 95.256 -69.713 52.890
343803 95.291 -69.708 52.917
343804 95.327 -69.708 52.946
343805 95.362 -69.713 52.969
343806 95.398 -69.717 52.981
343807 95.433 -69.722 52.983
343808 95.469 -69.726 52.977
343809 95.505 -69.730 52.964
343810 95.541 -69.735 52.947
343811 95.577 -69.739 52.934
343812 95.612 -69.744 52.922
343813 95.648 -69.748 52.908
343814 95.683 -69.753 52.897
343815 95.719 -69.757 52.885
343816 95.746 -69.761 52.876
343817 95.764 -69.765 52.869
343818 95.783 -69.770 52.857
343819 95.802 -69.773 52.840
343820 95.832 -69.783 52.822
343821 95.864 -69.794 52.805
343822 95.891 -69.803 52.788
343823 95.917 -69.811 52.765
343824 95.943 -69.819 52.739
343825 95.969 -69.827 52.716
343826 95.995 -69.836 52.699
343827 96.021 -69.843 52.682
343828 96.047 -69.852 52.660
343829 96.073 -69.860 52.639
343830 96.099 -69.869 52.621
343831 96.126 -69.881 52.601
343832 96.154 -69.894 52.574
343833 96.181 -69.907 52.546
343834 96.208 -69.919 52.521
343835 96.236 -69.932 52.500
343836 96.265 -69.944 52.476
343837 96.294 -69.956 52.452
343838 96.322 -69.968 52.432
343839 96.352 -69.980 52.415
343840 96.383 -69.991 52.401
343841 96.419 -69.999 52.383
343842 96.459 -69.997 52.365
343843 96.495 -70.003 52.350
343844 96.530 -70.010 52.339
343845 96.566 -70.017 52.329
343846 96.601 -70.024 52.318
343847 96.636 -70.030 52.307
343848 96.671 -70.038 52.297
343849 96.707 -70.045 52.285
343850 96.744 -70.058 52.272
343851 96.782 -70.073 52.258
343852 96.820 -70.087 52.242
343853 96.856 -70.095 52.222
343854 96.892 -70.102 52.209
343855 96.929 -70.111 52.199
343856 96.965 -70.118 52.191
343857 97.001 -70.126 52.180
343858 97.041 -70.137 52.167
343859 97.083 -70.152 52.151
343860 97.125 -70.165 52.134
343861 97.168 -70.180 52.120
343862 97.210 -70.194 52.106
343863 97.249 -70.194 52.090
343864 97.278 -70.201 52.073
343865 97.302 -70.214 52.053
343866 97.326 -70.227 52.024
343867 97.350 -70.246 51.995
343868 97.372 -70.270 51.967
343869 97.394 -70.294 51.938
343870 97.424 -70.303 51.908
343871 97.455 -70.312 51.876
343872 97.480 -70.322 51.839
343873 97.502 -70.331 51.797
343874 97.523 -70.339 51.755
343875 97.545 -70.348 51.715
343876 97.576 -70.347 51.675
343877 97.611 -70.342 51.635
343878 97.646 -70.336 51.596
343879 97.680 -70.334 51.552
343880 97.712 -70.336 51.508
343881 97.745 -70.338 51.465
343882 97.778 -70.340 51.421
343883 97.811 -70.341 51.378
343884 97.845 -70.339 51.337
343885 97.879 -70.337 51.294
343886 97.914 -70.336 51.248
343887 97.947 -70.334 51.201
343888 97.982 -70.334 51.153
343889 98.016 -70.332 51.104
343890 98.050 -70.331 51.055
343891 98.084 -70.329 51.006
343892 98.119 -70.328 50.953
343893 98.153 -70.326 50.902
343894 98.187 -70.325 50.852
343895 98.221 -70.324 50.803
343896 98.255 -70.322 50.752
343897 98.290 -70.321 50.700
343898 98.324 -70.319 50.648
343899 98.357 -70.319 50.594
343900 98.387 -70.327 50.540
343901 98.417 -70.334 50.486
343902 98.447 -70.342 50.432
343903 98.481 -70.344 50.378
343904 98.519 -70.353 50.324
343905 98.558 -70.362 50.270
343906 98.595 -70.372 50.216
343907 98.634 -70.381 50.161
343908 98.672 -70.390 50.105
343909 98.710 -70.399 50.048
343910 98.748 -70.408 49.990
343911 98.787 -70.417 49.933
343912 98.825 -70.426 49.875
343913 98.865 -70.438 49.816
343914 98.906 -70.450 49.759
343915 98.946 -70.463 49.701
343916 98.987 -70.475 49.644
343917 99.027 -70.488 49.587
343918 99.068 -70.500 49.534
343919 99.109 -70.513 49.481
343920 99.145 -70.523 49.429
343921 99.182 -70.534 49.379
343922 99.218 -70.545 49.329
343923 99.255 -70.556 49.280
343924 99.291 -70.566 49.233
343925 99.327 -70.579 49.188
343926 99.359 -70.595 49.144
343927 99.391 -70.612 49.101
343928 99.423 -70.629 49.059
343929 99.455 -70.641 49.020
343930 99.487 -70.654 48.983
343931 99.519 -70.666 48.947
343932 99.552 -70.677 48.913
343933 99.584 -70.689 48.881
343934 99.618 -70.698 48.848
343935 99.655 -70.706 48.818
343936 99.693 -70.713 48.790
343937 99.730 -70.721 48.763
343938 99.768 -70.729 48.738
343939 99.805 -70.736 48.713
343940 99.842 -70.744 48.684
343941 99.874 -70.754 48.653
343942 99.907 -70.764 48.620
343943 99.939 -70.774 48.582
343944 99.972 -70.785 48.541
343945 100.004 -70.795 48.501
343946 100.036 -70.804 48.463
343947 100.068 -70.815 48.426
343948 100.101 -70.825 48.393
343949 100.133 -70.836 48.356
343950 100.170 -70.855 48.318
343951 100.209 -70.874 48.281
343952 100.250 -70.890 48.239
343953 100.292 -70.907 48.199
343954 100.334 -70.923 48.160
343955 100.375 -70.940 48.119
343956 100.409 -70.949 48.078
343957 100.442 -70.958 48.037
343958 100.476 -70.967 47.997
343959 100.509 -70.977 47.957
343960 100.543 -70.986 47.916
343961 100.580 -70.994 47.877
343962 100.619 -71.002 47.838
343963 100.658 -71.010 47.801
343964 100.698 -71.017 47.765
343965 100.737 -71.026 47.727
343966 100.776 -71.033 47.691
343967 100.816 -71.041 47.656
343968 100.855 -71.049 47.622
343969 100.894 -71.057 47.589
343970 100.933 -71.065 47.553
343971 100.973 -71.073 47.519
343972 101.011 -71.081 47.484
343973 101.047 -71.091 47.448
343974 101.079 -71.101 47.413
343975 101.109 -71.111 47.378
343976 101.138 -71.122 47.345
343977 101.168 -71.133 47.311
343978 101.197 -71.143 47.277
343979 101.227 -71.153 47.243
343980 101.260 -71.169 47.207
343981 101.313 -71.190 47.168
343982 101.352 -71.206 47.125
343983 101.392 -71.222 47.081
343984 101.431 -71.239 47.036
343985 101.470 -71.255 46.992
343986 101.509 -71.271 46.948
343987 101.548 -71.287 46.902
343988 101.588 -71.304 46.857
343989 101.627 -71.320 46.811
343990 101.666 -71.337 46.766
343991 101.705 -71.353 46.719
343992 101.744 -71.369 46.671
343993 101.790 -71.392 46.624
343994 101.847 -71.425 46.577
343995 101.896 -71.451 46.528
343996 101.879 -71.420 46.482
343997 101.881 -71.412 46.437
343998 101.883 -71.405 46.393
343999 101.906 -71.404 46.352
344000 101.936 -71.406 46.311
344001 101.966 -71.407 46.270
344002 101.996 -71.409 46.230
344003 102.026 -71.410 46.191
344004 102.056 -71.412 46.151
344005 102.086 -71.413 46.113
344006 102.116 -71.415 46.075
344007 102.146 -71.416 46.037
344008 102.176 -71.418 46.001
344009 102.206 -71.420 45.965
344409 91.906 -70.235 55.182
344410 91.957 -70.278 55.175
344411 92.007 -70.254 55.166
344412 92.058 -70.230 55.156
344413 92.109 -70.206 55.147
344414 92.159 -70.181 55.137
344415 92.211 -70.144 55.126
344416 92.250 -70.115 55.116
344417 92.290 -70.084 55.104
344418 92.330 -70.055 55.093
344419 92.369 -70.025 55.080
344420 92.408 -69.996 55.067
344421 92.447 -69.971 55.056
344422 92.483 -69.977 55.045
344423 92.518 -69.983 55.034
344424 92.555 -69.990 55.025
344425 92.590 -69.996 55.018
344426 92.626 -70.002 55.007
344427 92.670 -70.011 54.993
344428 92.715 -70.011 54.978
344429 92.759 -70.006 54.958
344430 92.803 -70.001 54.933
344431 92.855 -69.995 54.907
344432 92.908 -69.990 54.882
344433 92.960 -69.983 54.850
344434 93.014 -69.978 54.812
344435 93.067 -69.972 54.771
344436 93.120 -69.966 54.726
344437 93.159 -69.943 54.677
344438 93.196 -69.916 54.621
344439 93.232 -69.889 54.566
344440 93.269 -69.861 54.515
344441 93.305 -69.834 54.466
344442 93.334 -69.826 54.420
344443 93.363 -69.818 54.379
344444 93.393 -69.814 54.341
344445 93.426 -69.814 54.303
344446 93.459 -69.814 54.267
344447 93.492 -69.815 54.231
344448 93.525 -69.816 54.196
344449 93.558 -69.816 54.162
344450 93.589 -69.813 54.128
344451 93.621 -69.810 54.094
344489 94.819 -69.701 52.921
344490 94.850 -69.698 52.905
344491 94.882 -69.696 52.890
344492 94.913 -69.694 52.875
344493 94.945 -69.691 52.862
344494 94.976 -69.689 52.850
344495 95.007 -69.686 52.842
344496 95.039 -69.684 52.837
344497 95.070 -69.682 52.834
344498 95.102 -69.679 52.833
344499 95.133 -69.677 52.836
344500 95.165 -69.675 52.842
344501 95.196 -69.672 52.851
344502 95.228 -69.670 52.870
344503 95.267 -69.669 52.894
344504 95.297 -69.667 52.921
344505 95.333 -69.671 52.950
344506 95.369 -69.676 52.975
344507 95.404 -69.680 52.988
344508 95.440 -69.685 52.991
344509 95.475 -69.689 52.986
344510 95.511 -69.694 52.974
344511 95.547 -69.698 52.959
344512 95.583 -69.702 52.949
344513 95.618 -69.707 52.939
344514 95.654 -69.711 52.927
344515 95.689 -69.716 52.917
344516 95.725 -69.720 52.905
344517 95.761 -69.725 52.894
344518 95.784 -69.729 52.883
344519 95.803 -69.733 52.872
344520 95.822 -69.737 52.854
344521 95.846 -69.744 52.836
344522 95.877 -69.754 52.821
344523 95.908 -69.764 52.807
344524 95.936 -69.773 52.785
344525 95.962 -69.782 52.758
344526 95.988 -69.790 52.732
344527 96.014 -69.798 52.711
344528 96.040 -69.806 52.693
344529 96.066 -69.814 52.671
344530 96.092 -69.822 52.649
344531 96.118 -69.832 52.629
344532 96.146 -69.845 52.609
344533 96.173 -69.857 52.583
344534 96.200 -69.870 52.556
344535 96.228 -69.882 52.531
344536 96.255 -69.895 52.509
344537 96.282 -69.908 52.486
344538 96.309 -69.920 52.465
344539 96.337 -69.933 52.446
344540 96.365 -69.945 52.431
344541 96.396 -69.956 52.416
344542 96.434 -69.960 52.397
344543 96.469 -69.967 52.378
344544 96.504 -69.974 52.362
344545 96.539 -69.981 52.351
344546 96.574 -69.988 52.341
344547 96.609 -69.995 52.331
344548 96.645 -70.002 52.319
344549 96.680 -70.008 52.308
344550 96.715 -70.015 52.296
344551 96.753 -70.029 52.283
344552 96.790 -70.042 52.268
344553 96.826 -70.049 52.252
344554 96.863 -70.057 52.230
344555 96.899 -70.065 52.213
344556 96.935 -70.073 52.199
344557 96.971 -70.080 52.189
344558 97.008 -70.088 52.176
344559 97.047 -70.099 52.164
344560 97.089 -70.114 52.151
344561 97.132 -70.127 52.137
344562 97.174 -70.142 52.123
344563 97.214 -70.148 52.109
344564 97.252 -70.146 52.095
344565 97.287 -70.149 52.079
344566 97.310 -70.162 52.060
344567 97.333 -70.183 52.031
344568 97.356 -70.206 52.000
344569 97.378 -70.231 51.969
344570 97.402 -70.251 51.939
344571 97.432 -70.260 51.909
344572 97.463 -70.269 51.877
344573 97.493 -70.278 51.840
344574 97.515 -70.287 51.799
344575 97.536 -70.296 51.758
344576 97.558 -70.305 51.715
344577 97.588 -70.305 51.674
344578 97.624 -70.300 51.633
344579 97.659 -70.293 51.593
344580 97.691 -70.294 51.551
344581 97.724 -70.297 51.508
344582 97.756 -70.298 51.465
344583 97.789 -70.300 51.423
344584 97.822 -70.301 51.380
344585 97.857 -70.300 51.338
344586 97.890 -70.298 51.293
344587 97.925 -70.297 51.247
344588 97.959 -70.295 51.201
344589 97.993 -70.294 51.155
344590 98.027 -70.293 51.109
344591 98.062 -70.291 51.063
344592 98.096 -70.290 51.013
344593 98.130 -70.288 50.959
344594 98.164 -70.287 50.905
344595 98.199 -70.285 50.852
344596 98.233 -70.284 50.798
344597 98.267 -70.283 50.747
344598 98.301 -70.281 50.696
344599 98.335 -70.281 50.646
344600 98.364 -70.289 50.594
344601 98.394 -70.297 50.541
344602 98.424 -70.304 50.488
344603 98.454 -70.312 50.434
344604 98.485 -70.311 50.380
344605 98.522 -70.320 50.326
344606 98.561 -70.329 50.270
344607 98.599 -70.338 50.214
344608 98.637 -70.347 50.157
344609 98.675 -70.356 50.101
344610 98.714 -70.366 50.043
344611 98.752 -70.375 49.985
344612 98.790 -70.384 49.927
344613 98.829 -70.394 49.869
344614 98.869 -70.406 49.811
344615 98.910 -70.418 49.754
344616 98.950 -70.431 49.697
344617 98.991 -70.443 49.640
344618 99.031 -70.455 49.583
344619 99.072 -70.468 49.531
344620 99.113 -70.480 49.479
344621 99.151 -70.491 49.427
344622 99.188 -70.503 49.376
344623 99.224 -70.513 49.328
344624 99.261 -70.524 49.280
344625 99.297 -70.535 49.233
344626 99.334 -70.546 49.189
344627 99.369 -70.557 49.146
344628 99.402 -70.574 49.102
344629 99.433 -70.591 49.062
344630 99.466 -70.603 49.024
344631 99.498 -70.615 48.986
344632 99.531 -70.626 48.951
344633 99.564 -70.638 48.918
344634 99.603 -70.645 48.886
344635 99.634 -70.658 48.856
344636 99.666 -70.670 48.826
344637 99.704 -70.678 48.797
344638 99.742 -70.685 48.771
344639 99.779 -70.694 48.748
344640 99.817 -70.701 48.724
344641 99.853 -70.710 48.697
344642 99.886 -70.720 48.666
344643 99.918 -70.730 48.631
344644 99.950 -70.741 48.591
344645 99.982 -70.751 48.552
344646 100.015 -70.761 48.514
344647 100.047 -70.772 48.478
344648 100.079 -70.782 48.442
344649 100.112 -70.792 48.408
344650 100.144 -70.802 48.372
344651 100.180 -70.812 48.336
344652 100.222 -70.828 48.297
344653 100.264 -70.845 48.256
344654 100.305 -70.861 48.219
344655 100.347 -70.877 48.180
344656 100.388 -70.894 48.140
344657 100.423 -70.905 48.097
344658 100.457 -70.914 48.059
344659 100.490 -70.923 48.020
344660 100.524 -70.932 47.979
344661 100.557 -70.941 47.937
344662 100.591 -70.950 47.900
344663 100.624 -70.959 47.862
344664 100.663 -70.967 47.824
344665 100.702 -70.975 47.788
344666 100.742 -70.983 47.749
344667 100.781 -70.991 47.712
344668 100.820 -70.999 47.678
344669 100.859 -71.007 47.644
344670 100.899 -71.014 47.611
344671 100.938 -71.023 47.576
344672 100.977 -71.030 47.541
344673 101.013 -71.039 47.506
344674 101.048 -71.050 47.469
344675 101.084 -71.060 47.435
344676 101.115 -71.071 47.399
344677 101.144 -71.080 47.364
344678 101.173 -71.091 47.330
344679 101.202 -71.102 47.295
344680 101.233 -71.115 47.258
344681 101.268 -71.133 47.221
344682 101.320 -71.153 47.181
344683 101.373 -71.174 47.138
344684 101.417 -71.191 47.093
344685 101.456 -71.208 47.047
344686 101.495 -71.224 47.002
344687 101.534 -71.240 46.958
344688 101.573 -71.256 46.913
344689 101.613 -71.273 46.868
344690 101.652 -71.289 46.822
344691 101.691 -71.306 46.776
344692 101.730 -71.321 46.726
344693 101.769 -71.338 46.678
344694 101.823 -71.368 46.629
344695 101.879 -71.400 46.582
344696 101.865 -71.368 46.533
344697 101.867 -71.361 46.488
344698 101.869 -71.353 46.443
344699 101.890 -71.352 46.399
344700 101.920 -71.353 46.357
344701 101.950 -71.355 46.316
344702 101.980 -71.356 46.276
344703 102.010 -71.359 46.235
344704 102.040 -71.360 46.195
344705 102.070 -71.362 46.157
344706 102.100 -71.363 46.117
344707 102.130 -71.365 46.079
344708 102.160 -71.366 46.043
344709 102.190 -71.368 46.007
345111 91.950 -70.250 55.173
345112 92.001 -70.225 55.164
345113 92.051 -70.200 55.154
345114 92.102 -70.176 55.143
345115 92.153 -70.151 55.132
345116 92.202 -70.114 55.122
345117 92.241 -70.084 55.111
345118 92.281 -70.055 55.098
345119 92.320 -70.025 55.085
345120 92.360 -69.995 55.072
345121 92.399 -69.965 55.058
345122 92.438 -69.936 55.045
345123 92.476 -69.921 55.033
345124 92.512 -69.927 55.019
345125 92.548 -69.933 55.008
345126 92.584 -69.939 54.997
345127 92.620 -69.946 54.984
345128 92.663 -69.955 54.970
345129 92.709 -69.964 54.956
345130 92.753 -69.958 54.936
345131 92.801 -69.952 54.911
345132 92.854 -69.946 54.885
345133 92.907 -69.941 54.860
345134 92.959 -69.935 54.827
345135 93.011 -69.929 54.790
345136 93.061 -69.922 54.750
345137 93.112 -69.915 54.707
345138 93.151 -69.893 54.661
345139 93.189 -69.868 54.610
345140 93.227 -69.844 54.559
345141 93.265 -69.819 54.510
345142 93.302 -69.795 54.464
345143 93.339 -69.789 54.422
345144 93.371 -69.788 54.382
345145 93.403 -69.786 54.343
345146 93.434 -69.783 54.306
345147 93.466 -69.781 54.270
345148 93.497 -69.779 54.235
345149 93.529 -69.776 54.201
345150 93.560 -69.773 54.167
345151 93.591 -69.771 54.135
345152 93.623 -69.769 54.101
345190 94.818 -69.679 52.929
345191 94.849 -69.676 52.912
345192 94.881 -69.674 52.896
345193 94.912 -69.672 52.882
345194 94.944 -69.670 52.869
345195 94.975 -69.667 52.857
345196 95.006 -69.664 52.848
345197 95.038 -69.662 52.842
345198 95.069 -69.660 52.839
345199 95.101 -69.657 52.838
345200 95.132 -69.655 52.840
345201 95.164 -69.653 52.846
345202 95.195 -69.651 52.856
345203 95.228 -69.648 52.876
345204 95.289 -69.653 52.899
345205 95.317 -69.648 52.926
345206 95.345 -69.642 52.954
345207 95.375 -69.639 52.980
345208 95.410 -69.643 52.994
345209 95.446 -69.648 52.998
345210 95.482 -69.652 52.994
345211 95.518 -69.657 52.983
345212 95.553 -69.661 52.970
345213 95.589 -69.666 52.963
345214 95.624 -69.670 52.956
345215 95.660 -69.674 52.947
345216 95.696 -69.679 52.938
345217 95.731 -69.683 52.927
345218 95.767 -69.688 52.915
345219 95.803 -69.692 52.901
345220 95.823 -69.696 52.888
345221 95.841 -69.700 52.869
345222 95.860 -69.704 52.851
345223 95.891 -69.714 52.836
345224 95.922 -69.725 52.824
345225 95.953 -69.736 52.802
345226 95.981 -69.744 52.776
345227 96.007 -69.752 52.750
345228 96.033 -69.761 52.727
345229 96.059 -69.769 52.707
345230 96.085 -69.776 52.683
345231 96.111 -69.785 52.660
345232 96.138 -69.795 52.638
345233 96.165 -69.808 52.619
345234 96.192 -69.820 52.594
345235 96.220 -69.833 52.569
345236 96.247 -69.845 52.546
345237 96.274 -69.858 52.523
345238 96.302 -69.870 52.502
345239 96.329 -69.883 52.481
345240 96.356 -69.895 52.462
345241 96.382 -69.908 52.447
345242 96.413 -69.918 52.431
345243 96.450 -69.925 52.412
345244 96.484 -69.933 52.391
345245 96.518 -69.941 52.374
345246 96.551 -69.949 52.363
345247 96.586 -69.957 52.354
345248 96.619 -69.964 52.344
345249 96.653 -69.973 52.333
345250 96.688 -69.980 52.320
345251 96.724 -69.986 52.309
345252 96.760 -69.996 52.295
345253 96.797 -70.004 52.279
345254 96.833 -70.011 52.264
345255 96.869 -70.019 52.242
345256 96.905 -70.027 52.223
345257 96.942 -70.035 52.207
345258 96.978 -70.043 52.195
345259 97.014 -70.050 52.182
345260 97.053 -70.062 52.168
345261 97.096 -70.075 52.154
345262 97.138 -70.090 52.141
345263 97.180 -70.102 52.129
345264 97.218 -70.100 52.119
345265 97.256 -70.098 52.107
345266 97.294 -70.096 52.092
345267 97.317 -70.119 52.073
345268 97.339 -70.143 52.043
345269 97.362 -70.167 52.010
345270 97.384 -70.190 51.976
345271 97.410 -70.207 51.941
345272 97.441 -70.216 51.908
345273 97.471 -70.225 51.876
345274 97.502 -70.234 51.840
345275 97.528 -70.243 51.801
345276 97.550 -70.253 51.760
345277 97.572 -70.262 51.721
345278 97.601 -70.262 51.682
345279 97.636 -70.257 51.644
345280 97.670 -70.253 51.607
345281 97.703 -70.255 51.566
345282 97.736 -70.256 51.524
345283 97.768 -70.259 51.482
345284 97.801 -70.260 51.439
345285 97.833 -70.262 51.396
345286 97.868 -70.260 51.352
345287 97.902 -70.259 51.308
345288 97.936 -70.257 51.262
345289 97.970 -70.256 51.218
345290 98.005 -70.254 51.174
345291 98.039 -70.253 51.128
345292 98.073 -70.252 51.080
345293 98.107 -70.250 51.030
345294 98.142 -70.249 50.974
345295 98.176 -70.247 50.918
345296 98.210 -70.246 50.864
345297 98.244 -70.245 50.810
345298 98.278 -70.243 50.759
345299 98.312 -70.243 50.709
345300 98.342 -70.251 50.658
345301 98.371 -70.259 50.605
345302 98.402 -70.267 50.552
345303 98.431 -70.275 50.497
345304 98.461 -70.282 50.442
345305 98.492 -70.284 50.387
345306 98.526 -70.287 50.331
345307 98.564 -70.296 50.273
345308 98.602 -70.305 50.215
345309 98.641 -70.314 50.157
345310 98.679 -70.323 50.099
345311 98.717 -70.332 50.042
345312 98.755 -70.341 49.984
345313 98.793 -70.350 49.926
345314 98.833 -70.361 49.868
345315 98.873 -70.373 49.809
345316 98.914 -70.386 49.751
345317 98.954 -70.398 49.694
345318 98.995 -70.410 49.637
345319 99.035 -70.423 49.582
345320 99.076 -70.435 49.530
345321 99.117 -70.447 49.478
345322 99.157 -70.460 49.426
345323 99.194 -70.471 49.375
345324 99.230 -70.481 49.327
345325 99.267 -70.493 49.279
345326 99.303 -70.503 49.234
345327 99.340 -70.514 49.190
345328 99.376 -70.525 49.147
345329 99.412 -70.536 49.104
345330 99.444 -70.552 49.065
345331 99.477 -70.564 49.027
345332 99.512 -70.574 48.990
345333 99.551 -70.581 48.956
345334 99.590 -70.588 48.924
345335 99.622 -70.601 48.893
345336 99.653 -70.614 48.863
345337 99.684 -70.628 48.833
345338 99.715 -70.641 48.805
345339 99.753 -70.648 48.778
345340 99.790 -70.656 48.754
345341 99.828 -70.664 48.732
345342 99.863 -70.673 48.706
345343 99.896 -70.683 48.676
345344 99.928 -70.694 48.642
345345 99.960 -70.704 48.603
345346 99.992 -70.714 48.564
345347 100.025 -70.724 48.528
345348 100.057 -70.735 48.492
345349 100.090 -70.744 48.457
345350 100.123 -70.754 48.422
345351 100.156 -70.763 48.387
345352 100.192 -70.773 48.352
345353 100.234 -70.784 48.314
345354 100.277 -70.798 48.274
345355 100.318 -70.815 48.237
345356 100.360 -70.832 48.199
345357 100.402 -70.848 48.159
345358 100.438 -70.860 48.117
345359 100.471 -70.869 48.079
345360 100.505 -70.878 48.041
345361 100.538 -70.887 48.001
345362 100.572 -70.896 47.960
345363 100.605 -70.905 47.924
345364 100.639 -70.914 47.886
345365 100.673 -70.923 47.848
345366 100.706 -70.933 47.811
345367 100.746 -70.940 47.772
345368 100.785 -70.948 47.736
345369 100.824 -70.956 47.701
345370 100.863 -70.964 47.667
345371 100.903 -70.972 47.633
345372 100.942 -70.980 47.598
345373 100.980 -70.989 47.563
345374 101.015 -70.999 47.528
345375 101.050 -71.009 47.490
345376 101.085 -71.019 47.457
345377 101.120 -71.029 47.421
345378 101.149 -71.039 47.384
345379 101.179 -71.050 47.349
345380 101.208 -71.061 47.312
345381 101.240 -71.078 47.273
345382 101.275 -71.096 47.234
345383 101.328 -71.116 47.193
345384 101.380 -71.136 47.149
345385 101.433 -71.157 47.104
345386 101.481 -71.176 47.058
345387 101.520 -71.193 47.013
345388 101.559 -71.209 46.969
345389 101.599 -71.225 46.925
345390 101.638 -71.241 46.880
345391 101.677 -71.258 46.833
345392 101.716 -71.274 46.785
345393 101.755 -71.290 46.736
345394 101.798 -71.310 46.686
345395 101.854 -71.343 46.636
345396 101.853 -71.321 46.587
345397 101.852 -71.309 46.539
345398 101.854 -71.302 46.493
345399 101.874 -71.300 46.448
345400 101.904 -71.302 46.405
345401 101.934 -71.303 46.362
345402 101.964 -71.305 46.321
345403 101.994 -71.306 46.281
345404 102.024 -71.308 46.241
345405 102.054 -71.309 46.201
345406 102.084 -71.311 46.161
345407 102.114 -71.312 46.122
345408 102.144 -71.314 46.084
345409 102.174 -71.315 46.048
345814 92.044 -70.171 55.152
345815 92.095 -70.146 55.140
345816 92.146 -70.119 55.129
345817 92.193 -70.084 55.117
345818 92.232 -70.054 55.105
345819 92.272 -70.024 55.092
345820 92.311 -69.995 55.078
345821 92.351 -69.965 55.064
345822 92.390 -69.936 55.049
345823 92.429 -69.905 55.035
345824 92.469 -69.876 55.020
345825 92.506 -69.870 55.006
345826 92.542 -69.877 54.991
345827 92.577 -69.883 54.978
345828 92.613 -69.889 54.962
345829 92.656 -69.898 54.946
345830 92.702 -69.908 54.930
345831 92.747 -69.910 54.909
345832 92.800 -69.904 54.885
345833 92.851 -69.897 54.859
345834 92.901 -69.891 54.832
345835 92.951 -69.884 54.799
345836 93.002 -69.877 54.763
345837 93.053 -69.870 54.725
345838 93.103 -69.864 54.684
345839 93.142 -69.841 54.640
345840 93.180 -69.816 54.594
345841 93.217 -69.792 54.548
345842 93.255 -69.767 54.503
345843 93.293 -69.751 54.461
345844 93.329 -69.754 54.421
345845 93.365 -69.758 54.382
345846 93.401 -69.762 54.344
345847 93.434 -69.761 54.308
345848 93.465 -69.759 54.272
345849 93.497 -69.757 54.238
345850 93.528 -69.754 54.205
345851 93.560 -69.752 54.172
345852 93.591 -69.749 54.139
345853 93.622 -69.747 54.106
345890 94.786 -69.660 52.955
345891 94.817 -69.657 52.938
345892 94.849 -69.654 52.921
345893 94.880 -69.652 52.905
345894 94.912 -69.650 52.890
345895 94.943 -69.648 52.876
345896 94.974 -69.645 52.864
345897 95.006 -69.643 52.855
345898 95.037 -69.641 52.849
345899 95.069 -69.639 52.845
345900 95.100 -69.635 52.844
345901 95.132 -69.633 52.845
345902 95.163 -69.631 52.851
345903 95.195 -69.629 52.861
345904 95.269 -69.638 52.881
345905 95.312 -69.636 52.904
345906 95.340 -69.631 52.929
345907 95.368 -69.626 52.957
345908 95.395 -69.620 52.983
345909 95.423 -69.615 52.998
345910 95.452 -69.611 53.004
345911 95.488 -69.615 53.003
345912 95.524 -69.620 52.993
345913 95.559 -69.624 52.981
345914 95.595 -69.629 52.976
345915 95.630 -69.633 52.973
345916 95.666 -69.638 52.968
345917 95.702 -69.642 52.961
345918 95.738 -69.646 52.952
345919 95.773 -69.651 52.939
345920 95.809 -69.655 52.923
345921 95.843 -69.660 52.906
345922 95.862 -69.664 52.886
345923 95.880 -69.667 52.867
345924 95.904 -69.675 52.850
345925 95.936 -69.685 52.838
345926 95.967 -69.695 52.817
345927 95.998 -69.706 52.793
345928 96.026 -69.715 52.767
345929 96.052 -69.723 52.743
345930 96.079 -69.731 52.721
345931 96.104 -69.739 52.696
345932 96.130 -69.748 52.671
345933 96.157 -69.758 52.649
345934 96.184 -69.770 52.629
345935 96.212 -69.783 52.606
345936 96.239 -69.795 52.583
345937 96.266 -69.808 52.561
345938 96.294 -69.821 52.540
345939 96.321 -69.833 52.520
345940 96.347 -69.846 52.498
345941 96.371 -69.858 52.478
345942 96.394 -69.871 52.463
345943 96.427 -69.880 52.446
345944 96.472 -69.885 52.425
345945 96.508 -69.892 52.404
345946 96.542 -69.900 52.385
345947 96.576 -69.908 52.374
345948 96.610 -69.916 52.367
345949 96.644 -69.924 52.358
345950 96.675 -69.932 52.346
345951 96.703 -69.940 52.333
345952 96.735 -69.951 52.321
345953 96.768 -69.959 52.307
345954 96.803 -69.966 52.291
345955 96.839 -69.974 52.277
345956 96.875 -69.981 52.255
345957 96.911 -69.990 52.235
345958 96.948 -69.997 52.219
345959 96.984 -70.005 52.207
345960 97.020 -70.012 52.194
345961 97.060 -70.024 52.180
345962 97.102 -70.037 52.164
345963 97.144 -70.052 52.149
345964 97.183 -70.054 52.137
345965 97.222 -70.052 52.129
345966 97.259 -70.052 52.118
345967 97.295 -70.062 52.103
345968 97.323 -70.079 52.084
345969 97.345 -70.103 52.054
345970 97.368 -70.127 52.019
345971 97.390 -70.151 51.983
345972 97.418 -70.164 51.946
345973 97.449 -70.173 51.912
345974 97.479 -70.182 51.879
345975 97.510 -70.191 51.844
345976 97.540 -70.200 51.807
345977 97.564 -70.209 51.769
345978 97.585 -70.218 51.732
345979 97.613 -70.220 51.696
345980 97.649 -70.214 51.660
345981 97.682 -70.213 51.626
345982 97.715 -70.215 51.586
345983 97.747 -70.217 51.545
345984 97.780 -70.219 51.503
345985 97.812 -70.221 51.459
345986 97.845 -70.222 51.416
345987 97.879 -70.221 51.372
345988 97.914 -70.219 51.327
345989 97.948 -70.218 51.281
345990 97.982 -70.216 51.237
345991 98.016 -70.215 51.193
345992 98.051 -70.214 51.146
345993 98.085 -70.212 51.097
345994 98.119 -70.211 51.048
345995 98.153 -70.209 50.992
345996 98.187 -70.208 50.936
345997 98.221 -70.206 50.880
345998 98.255 -70.206 50.826
345999 98.289 -70.206 50.774
346000 98.319 -70.214 50.724
346001 98.349 -70.221 50.671
346002 98.378 -70.229 50.616
346003 98.409 -70.237 50.561
346004 98.438 -70.244 50.506
346005 98.468 -70.252 50.449
346006 98.498 -70.258 50.392
346007 98.529 -70.253 50.335
346008 98.568 -70.262 50.276
346009 98.606 -70.272 50.217
346010 98.644 -70.281 50.159
346011 98.683 -70.290 50.101
346012 98.720 -70.299 50.043
346013 98.759 -70.308 49.984
346014 98.797 -70.317 49.925
346015 98.837 -70.328 49.866
346016 98.877 -70.341 49.808
346017 98.918 -70.353 49.750
346018 98.959 -70.366 49.692
346019 98.999 -70.378 49.636
346020 99.040 -70.391 49.580
346021 99.080 -70.403 49.529
346022 99.121 -70.416 49.476
346023 99.162 -70.428 49.425
346024 99.200 -70.439 49.374
346025 99.236 -70.450 49.326
346026 99.273 -70.461 49.279
346027 99.309 -70.472 49.234
346028 99.346 -70.482 49.190
346029 99.382 -70.494 49.147
346030 99.419 -70.504 49.106
346031 99.460 -70.510 49.067
346032 99.499 -70.516 49.030
346033 99.538 -70.524 48.993
346034 99.577 -70.531 48.961
346035 99.610 -70.543 48.929
346036 99.641 -70.557 48.898
346037 99.671 -70.569 48.869
346038 99.702 -70.583 48.840
346039 99.733 -70.597 48.812
346040 99.764 -70.610 48.786
346041 99.801 -70.619 48.761
346042 99.839 -70.626 48.739
346043 99.873 -70.636 48.714
346044 99.905 -70.646 48.685
346045 99.938 -70.657 48.651
346046 99.970 -70.666 48.614
346047 100.003 -70.676 48.577
346048 100.036 -70.686 48.542
346049 100.069 -70.695 48.507
346050 100.102 -70.705 48.472
346051 100.135 -70.714 48.436
346052 100.168 -70.723 48.401
346053 100.203 -70.733 48.367
346054 100.246 -70.745 48.331
346055 100.289 -70.756 48.294
346056 100.332 -70.769 48.255
346057 100.373 -70.785 48.216
346058 100.415 -70.802 48.177
346059 100.452 -70.814 48.137
346060 100.486 -70.824 48.099
346061 100.519 -70.833 48.061
346062 100.553 -70.842 48.022
346063 100.586 -70.851 47.985
346064 100.620 -70.861 47.948
346065 100.653 -70.870 47.911
346066 100.687 -70.879 47.873
346067 100.720 -70.888 47.835
346068 100.754 -70.897 47.797
346069 100.789 -70.906 47.761
346070 100.829 -70.914 47.725
346071 100.868 -70.921 47.691
346072 100.907 -70.930 47.656
346073 100.946 -70.937 47.620
346074 100.981 -70.948 47.585
346075 101.016 -70.958 47.548
346076 101.052 -70.967 47.511
346077 101.087 -70.978 47.477
346078 101.122 -70.988 47.441
346079 101.155 -70.998 47.405
346080 101.184 -71.008 47.368
346081 101.215 -71.024 47.329
346082 101.247 -71.041 47.287
346083 101.283 -71.059 47.244
346084 101.335 -71.080 47.201
346085 101.388 -71.100 47.157
346086 101.441 -71.121 47.112
346087 101.493 -71.141 47.068
346088 101.545 -71.162 47.024
346089 101.585 -71.177 46.980
346090 101.624 -71.194 46.935
346091 101.663 -71.210 46.890
346092 101.702 -71.227 46.843
346093 101.741 -71.243 46.794
346094 101.780 -71.259 46.743
346095 101.830 -71.285 46.693
346096 101.856 -71.289 46.642
346097 101.839 -71.259 46.593
346098 101.841 -71.250 46.544
346099 101.858 -71.248 46.499
346100 101.888 -71.249 46.454
346101 101.918 -71.251 46.411
346102 101.948 -71.252 46.368
346103 101.978 -71.254 46.326
346104 102.007 -71.256 46.286
346105 102.038 -71.257 46.246
346106 102.068 -71.259 46.206
346107 102.098 -71.260 46.166
346108 102.128 -71.262 46.128
346109 102.158 -71.263 46.090
346516 92.088 -70.118 55.137
346517 92.139 -70.087 55.125
346518 92.184 -70.054 55.112
346519 92.223 -70.024 55.099
346520 92.263 -69.994 55.085
346521 92.302 -69.964 55.070
346522 92.342 -69.935 55.055
346523 92.381 -69.905 55.039
346524 92.420 -69.876 55.024
346525 92.459 -69.845 55.008
346526 92.499 -69.816 54.992
346527 92.535 -69.820 54.975
346528 92.571 -69.827 54.960
346529 92.606 -69.833 54.942
346530 92.648 -69.841 54.923
346531 92.695 -69.851 54.903
346532 92.741 -69.858 54.881
346533 92.792 -69.853 54.856
346534 92.842 -69.846 54.829
346535 92.893 -69.839 54.801
346536 92.943 -69.833 54.770
346537 92.993 -69.827 54.735
346538 93.044 -69.820 54.699
346539 93.094 -69.813 54.661
346540 93.132 -69.789 54.619
346541 93.170 -69.764 54.578
346542 93.208 -69.739 54.536
346543 93.246 -69.715 54.495
346544 93.282 -69.717 54.455
346545 93.318 -69.720 54.417
346546 93.354 -69.724 54.379
346547 93.391 -69.728 54.343
346548 93.427 -69.731 54.307
346549 93.463 -69.735 54.273
346550 93.496 -69.735 54.240
346551 93.528 -69.733 54.207
346552 93.559 -69.730 54.175
346553 93.590 -69.728 54.142
346554 93.622 -69.726 54.110
346590 94.754 -69.640 52.985
346591 94.785 -69.638 52.965
346592 94.817 -69.635 52.947
346593 94.848 -69.633 52.930
346594 94.880 -69.631 52.914
346595 94.911 -69.628 52.899
346596 94.943 -69.626 52.885
346597 94.974 -69.623 52.873
346598 95.005 -69.621 52.864
346599 95.037 -69.619 52.858
346600 95.068 -69.617 52.853
346601 95.100 -69.614 52.852
346602 95.131 -69.612 52.853
346603 95.163 -69.609 52.858
346604 95.218 -69.613 52.867
346605 95.307 -69.626 52.885
346606 95.334 -69.620 52.906
346607 95.362 -69.615 52.930
346608 95.390 -69.609 52.956
346609 95.418 -69.604 52.981
346610 95.446 -69.598 52.998
346611 95.474 -69.593 53.007
346612 95.501 -69.588 53.010
346613 95.530 -69.583 53.003
346614 95.565 -69.587 52.995
346615 95.601 -69.591 52.989
346616 95.637 -69.596 52.987
346617 95.672 -69.601 52.986
346618 95.708 -69.605 52.983
346619 95.744 -69.610 52.977
346620 95.779 -69.614 52.966
346621 95.815 -69.618 52.950
346622 95.850 -69.623 52.929
346623 95.882 -69.627 52.906
346624 95.900 -69.631 52.884
346625 95.919 -69.635 52.863
346626 95.949 -69.645 52.850
346627 95.980 -69.656 52.831
346628 96.012 -69.666 52.807
346629 96.043 -69.676 52.782
346630 96.072 -69.686 52.759
346631 96.098 -69.694 52.734
346632 96.123 -69.702 52.707
346633 96.150 -69.710 52.682
346634 96.176 -69.720 52.659
346635 96.203 -69.733 52.639
346636 96.231 -69.746 52.619
346637 96.258 -69.758 52.599
346638 96.286 -69.771 52.578
346639 96.313 -69.784 52.559
346640 96.336 -69.796 52.537
346641 96.360 -69.809 52.515
346642 96.384 -69.821 52.494
346643 96.407 -69.834 52.477
346644 96.441 -69.843 52.458
346645 96.485 -69.848 52.437
346646 96.531 -69.852 52.416
346647 96.567 -69.860 52.396
346648 96.595 -69.867 52.386
346649 96.624 -69.876 52.379
346650 96.653 -69.884 52.370
346651 96.682 -69.892 52.357
346652 96.713 -69.902 52.344
346653 96.746 -69.914 52.332
346654 96.780 -69.924 52.318
346655 96.813 -69.930 52.304
346656 96.847 -69.936 52.290
346657 96.882 -69.943 52.270
346658 96.918 -69.952 52.251
346659 96.954 -69.959 52.236
346660 96.991 -69.967 52.226
346661 97.027 -69.975 52.214
346662 97.066 -69.986 52.198
346663 97.108 -69.999 52.180
346664 97.149 -70.008 52.160
346665 97.187 -70.005 52.146
346666 97.224 -70.008 52.138
346667 97.260 -70.018 52.126
346668 97.296 -70.027 52.111
346669 97.329 -70.040 52.093
346670 97.351 -70.064 52.061
346671 97.374 -70.087 52.027
346672 97.396 -70.111 51.990
346673 97.427 -70.121 51.953
346674 97.457 -70.130 51.919
346675 97.488 -70.139 51.886
346676 97.518 -70.148 51.852
346677 97.548 -70.157 51.817
346678 97.577 -70.166 51.783
346679 97.598 -70.174 51.748
346680 97.626 -70.178 51.714
346681 97.661 -70.171 51.681
346682 97.694 -70.174 51.648
346683 97.726 -70.175 51.609
346684 97.759 -70.178 51.568
346685 97.791 -70.179 51.526
346686 97.824 -70.181 51.482
346687 97.857 -70.183 51.437
346688 97.891 -70.181 51.393
346689 97.925 -70.180 51.348
346690 97.959 -70.178 51.302
346691 97.994 -70.177 51.257
346692 98.028 -70.175 51.210
346693 98.062 -70.174 51.162
346694 98.096 -70.173 51.114
346695 98.130 -70.171 51.064
346696 98.164 -70.170 51.010
346697 98.199 -70.168 50.954
346698 98.233 -70.167 50.899
346699 98.266 -70.168 50.844
346700 98.296 -70.176 50.790
346701 98.326 -70.184 50.739
346702 98.356 -70.191 50.684
346703 98.385 -70.199 50.627
346704 98.416 -70.206 50.570
346705 98.445 -70.214 50.513
346706 98.476 -70.221 50.455
346707 98.505 -70.229 50.397
346708 98.536 -70.226 50.340
346709 98.571 -70.229 50.280
346710 98.609 -70.238 50.221
346711 98.647 -70.247 50.163
346712 98.686 -70.256 50.105
346713 98.724 -70.265 50.045
346714 98.762 -70.275 49.985
346715 98.800 -70.284 49.925
346716 98.841 -70.297 49.866
346717 98.882 -70.309 49.807
346718 98.922 -70.322 49.749
346719 98.963 -70.334 49.691
346720 99.004 -70.346 49.635
346721 99.044 -70.359 49.579
346722 99.085 -70.371 49.528
346723 99.125 -70.383 49.475
346724 99.166 -70.396 49.423
346725 99.206 -70.408 49.373
346726 99.242 -70.419 49.325
346727 99.279 -70.429 49.278
346728 99.315 -70.441 49.233
346729 99.352 -70.451 49.190
346730 99.389 -70.462 49.147
346731 99.425 -70.472 49.108
346732 99.473 -70.472 49.070
346733 99.520 -70.471 49.033
346734 99.564 -70.474 48.997
346735 99.598 -70.485 48.964
346736 99.629 -70.498 48.932
346737 99.660 -70.512 48.902
346738 99.691 -70.525 48.873
346739 99.721 -70.539 48.845
346740 99.752 -70.553 48.819
346741 99.783 -70.566 48.794
346742 99.814 -70.580 48.771
346743 99.850 -70.589 48.749
346744 99.883 -70.599 48.723
346745 99.916 -70.609 48.692
346746 99.948 -70.619 48.658
346747 99.982 -70.628 48.622
346748 100.015 -70.637 48.588
346749 100.048 -70.647 48.555
346750 100.081 -70.656 48.522
346751 100.114 -70.666 48.487
346752 100.147 -70.675 48.450
346753 100.181 -70.685 48.415
346754 100.214 -70.694 48.382
346755 100.258 -70.705 48.348
346756 100.300 -70.717 48.312
346757 100.343 -70.728 48.273
346758 100.387 -70.739 48.234
346759 100.428 -70.756 48.195
346760 100.467 -70.770 48.158
346761 100.500 -70.779 48.121
346762 100.534 -70.788 48.082
346763 100.567 -70.798 48.043
346764 100.601 -70.807 48.006
346765 100.634 -70.816 47.971
346766 100.668 -70.825 47.934
346767 100.701 -70.834 47.896
346768 100.735 -70.843 47.859
346769 100.768 -70.852 47.819
346770 100.802 -70.861 47.781
346771 100.836 -70.870 47.746
346772 100.872 -70.879 47.713
346773 100.911 -70.887 47.678
346774 100.948 -70.896 47.642
346775 100.983 -70.906 47.606
346776 101.018 -70.917 47.568
346777 101.053 -70.927 47.530
346778 101.089 -70.936 47.495
346779 101.123 -70.947 47.458
346780 101.159 -70.957 47.422
346781 101.191 -70.970 47.384
346782 101.222 -70.987 47.344
346783 101.253 -71.005 47.301
346784 101.290 -71.022 47.258
346785 101.343 -71.042 47.214
346786 101.395 -71.063 47.168
346787 101.448 -71.083 47.123
346788 101.500 -71.104 47.078
346789 101.553 -71.124 47.034
346790 101.606 -71.145 46.990
346791 101.649 -71.162 46.944
346792 101.688 -71.179 46.897
346793 101.727 -71.195 46.850
346794 101.766 -71.212 46.800
346795 101.806 -71.227 46.749
346796 101.858 -71.257 46.698
346797 101.825 -71.207 46.648
346798 101.827 -71.199 46.598
346799 101.841 -71.196 46.549
346800 101.872 -71.197 46.504
346801 101.902 -71.199 46.459
346802 101.932 -71.200 46.416
346803 101.962 -71.202 46.373
346804 101.991 -71.204 46.331
346805 102.022 -71.205 46.290
346806 102.052 -71.207 46.250
346807 102.082 -71.209 46.210
346808 102.112 -71.210 46.171
346809 102.141 -71.212 46.133
347218 92.133 -70.055 55.119
347219 92.175 -70.024 55.107
347220 92.214 -69.994 55.092
347221 92.253 -69.964 55.077
347222 92.293 -69.935 55.061
347223 92.333 -69.905 55.045
347224 92.372 -69.875 55.028
347225 92.411 -69.845 55.012
347226 92.450 -69.816 54.996
347227 92.490 -69.786 54.980
347228 92.528 -69.764 54.960
347229 92.564 -69.770 54.943
347230 92.600 -69.776 54.924
347231 92.641 -69.784 54.904
347232 92.687 -69.794 54.881
347233 92.735 -69.798 54.857
347234 92.783 -69.801 54.832
347235 92.834 -69.795 54.805
347236 92.884 -69.789 54.777
347237 92.934 -69.782 54.746
347238 92.985 -69.775 54.713
347239 93.035 -69.768 54.678
347240 93.085 -69.761 54.643
347241 93.123 -69.736 54.604
347242 93.160 -69.712 54.565
347243 93.199 -69.687 54.526
347244 93.235 -69.678 54.488
347245 93.272 -69.682 54.449
347246 93.308 -69.686 54.413
347247 93.344 -69.689 54.377
347248 93.380 -69.693 54.341
347249 93.416 -69.697 54.306
347250 93.452 -69.701 54.273
347251 93.488 -69.704 54.241
347252 93.524 -69.708 54.209
347253 93.558 -69.708 54.177
347254 93.589 -69.706 54.145
347255 93.621 -69.704 54.114
347291 94.753 -69.618 52.996
347292 94.784 -69.616 52.975
347293 94.816 -69.613 52.958
347294 94.847 -69.611 52.941
347295 94.879 -69.609 52.925
347296 94.910 -69.607 52.910
347297 94.942 -69.604 52.894
347298 94.973 -69.602 52.883
347299 95.004 -69.599 52.874
347300 95.036 -69.597 52.867
347301 95.067 -69.595 52.863
347302 95.099 -69.592 52.861
347303 95.130 -69.590 52.863
347304 95.168 -69.589 52.867
347305 95.259 -69.602 52.875
347306 95.329 -69.609 52.889
347307 95.357 -69.604 52.909
347308 95.385 -69.598 52.931
347309 95.413 -69.593 52.953
347310 95.441 -69.588 52.977
347311 95.468 -69.582 52.996
347312 95.496 -69.576 53.009
347313 95.524 -69.571 53.015
347314 95.552 -69.566 53.012
347315 95.580 -69.560 53.007
347316 95.607 -69.555 53.002
347317 95.643 -69.559 53.000
347318 95.679 -69.563 53.001
347319 95.714 -69.568 53.001
347320 95.750 -69.573 52.998
347321 95.785 -69.577 52.990
347322 95.821 -69.582 52.976
347323 95.857 -69.586 52.954
347324 95.893 -69.590 52.928
347325 95.920 -69.595 52.902
347326 95.939 -69.598 52.878
347327 95.963 -69.605 52.862
347328 95.994 -69.616 52.844
347329 96.025 -69.626 52.822
347330 96.056 -69.636 52.798
347331 96.088 -69.647 52.774
347332 96.117 -69.657 52.746
347333 96.143 -69.664 52.720
347334 96.169 -69.673 52.694
347335 96.195 -69.684 52.671
347336 96.223 -69.696 52.651
347337 96.250 -69.709 52.632
347338 96.277 -69.722 52.613
347339 96.302 -69.734 52.594
347340 96.326 -69.747 52.575
347341 96.349 -69.759 52.554
347342 96.373 -69.772 52.531
347343 96.397 -69.784 52.508
347344 96.420 -69.797 52.490
347345 96.454 -69.806 52.469
347346 96.499 -69.811 52.447
347347 96.544 -69.814 52.426
347348 96.574 -69.820 52.407
347349 96.603 -69.827 52.395
347350 96.631 -69.836 52.387
347351 96.661 -69.843 52.377
347352 96.691 -69.853 52.364
347353 96.725 -69.865 52.350
347354 96.758 -69.878 52.338
347355 96.791 -69.889 52.325
347356 96.824 -69.895 52.312
347357 96.858 -69.901 52.299
347358 96.891 -69.907 52.279
347359 96.925 -69.914 52.263
347360 96.961 -69.921 52.251
347361 96.997 -69.929 52.243
347362 97.033 -69.937 52.232
347363 97.072 -69.947 52.218
347364 97.114 -69.961 52.198
347365 97.153 -69.959 52.174
347366 97.189 -69.965 52.156
347367 97.225 -69.974 52.143
347368 97.261 -69.984 52.129
347369 97.297 -69.993 52.113
347370 97.332 -70.003 52.094
347371 97.357 -70.024 52.065
347372 97.380 -70.048 52.033
347373 97.404 -70.068 51.998
347374 97.435 -70.077 51.963
347375 97.465 -70.086 51.931
347376 97.496 -70.095 51.900
347377 97.526 -70.104 51.867
347378 97.557 -70.113 51.833
347379 97.587 -70.122 51.800
347380 97.612 -70.131 51.766
347381 97.639 -70.135 51.732
347382 97.673 -70.132 51.699
347383 97.706 -70.134 51.665
347384 97.738 -70.136 51.626
347385 97.771 -70.137 51.585
347386 97.803 -70.140 51.542
347387 97.836 -70.141 51.498
347388 97.868 -70.143 51.453
347389 97.902 -70.142 51.410
347390 97.937 -70.140 51.364
347391 97.971 -70.139 51.316
347392 98.005 -70.137 51.270
347393 98.039 -70.136 51.222
347394 98.073 -70.134 51.174
347395 98.107 -70.134 51.126
347396 98.142 -70.132 51.077
347397 98.176 -70.130 51.022
347398 98.210 -70.129 50.967
347399 98.243 -70.130 50.911
347400 98.273 -70.138 50.857
347401 98.303 -70.146 50.802
347402 98.333 -70.153 50.748
347403 98.363 -70.161 50.692
347404 98.393 -70.168 50.634
347405 98.423 -70.176 50.576
347406 98.452 -70.184 50.518
347407 98.483 -70.191 50.459
347408 98.512 -70.199 50.401
347409 98.543 -70.199 50.343
347410 98.575 -70.196 50.285
347411 98.613 -70.205 50.226
347412 98.651 -70.214 50.168
347413 98.689 -70.223 50.109
347414 98.727 -70.232 50.049
347415 98.766 -70.241 49.988
347416 98.805 -70.252 49.927
347417 98.846 -70.264 49.866
347418 98.886 -70.277 49.806
347419 98.927 -70.289 49.747
347420 98.967 -70.301 49.689
347421 99.008 -70.314 49.632
347422 99.049 -70.326 49.577
347423 99.089 -70.338 49.525
347424 99.130 -70.351 49.473
347425 99.170 -70.363 49.422
347426 99.211 -70.375 49.372
347427 99.249 -70.387 49.324
347428 99.285 -70.398 49.278
347429 99.322 -70.409 49.233
347430 99.358 -70.419 49.190
347431 99.395 -70.430 49.148
347432 99.438 -70.433 49.108
347433 99.486 -70.433 49.070
347434 99.533 -70.432 49.033
347435 99.577 -70.435 48.998
347436 99.614 -70.453 48.966
347437 99.645 -70.466 48.933
347438 99.676 -70.478 48.902
347439 99.707 -70.491 48.874
347440 99.738 -70.504 48.847
347441 99.769 -70.517 48.823
347442 99.800 -70.530 48.799
347443 99.831 -70.543 48.777
347444 99.862 -70.556 48.754
347445 99.895 -70.560 48.726
347446 99.927 -70.569 48.695
347447 99.961 -70.579 48.663
347448 99.994 -70.588 48.628
347449 100.027 -70.598 48.599
347450 100.060 -70.607 48.567
347451 100.093 -70.616 48.534
347452 100.126 -70.626 48.499
347453 100.159 -70.635 48.460
347454 100.193 -70.645 48.426
347455 100.226 -70.654 48.395
347456 100.269 -70.666 48.363
347457 100.312 -70.677 48.327
347458 100.355 -70.688 48.289
347459 100.398 -70.700 48.251
347460 100.441 -70.711 48.213
347461 100.481 -70.725 48.176
347462 100.515 -70.734 48.139
347463 100.548 -70.743 48.099
347464 100.582 -70.752 48.061
347465 100.615 -70.761 48.026
347466 100.649 -70.771 47.991
347467 100.682 -70.780 47.954
347468 100.716 -70.789 47.915
347469 100.750 -70.798 47.877
347470 100.783 -70.807 47.838
347471 100.818 -70.817 47.800
347472 100.852 -70.827 47.765
347473 100.887 -70.837 47.732
347474 100.918 -70.846 47.697
347475 100.954 -70.855 47.661
347476 100.990 -70.866 47.624
347477 101.026 -70.876 47.586
347478 101.062 -70.886 47.548
347479 101.099 -70.896 47.510
347480 101.134 -70.906 47.473
347481 101.170 -70.917 47.436
347482 101.198 -70.933 47.398
347483 101.229 -70.951 47.357
347484 101.260 -70.967 47.315
347485 101.298 -70.986 47.272
347486 101.350 -71.006 47.227
347487 101.403 -71.027 47.180
347488 101.456 -71.047 47.134
347489 101.508 -71.068 47.089
347490 101.560 -71.088 47.044
347491 101.613 -71.108 46.999
347492 101.666 -71.129 46.951
347493 101.713 -71.148 46.902
347494 101.752 -71.164 46.853
347495 101.792 -71.180 46.803
347496 101.837 -71.203 46.754
347497 101.813 -71.158 46.703
347498 101.813 -71.148 46.652
347499 101.825 -71.144 46.602
347500 101.855 -71.146 46.554
347501 101.886 -71.147 46.507
347502 101.915 -71.149 46.463
347503 101.946 -71.150 46.419
347504 101.975 -71.152 46.377
347505 102.006 -71.153 46.335
347506 102.036 -71.155 46.295
347507 102.065 -71.156 46.254
347508 102.096 -71.158 46.214
347509 102.125 -71.159 46.173
347510 102.156 -71.161 46.135
347920 92.166 -69.993 55.101
347921 92.205 -69.964 55.085
347922 92.244 -69.934 55.070
347923 92.284 -69.905 55.053
347924 92.323 -69.875 55.037
347925 92.363 -69.845 55.020
347926 92.402 -69.815 55.002
347927 92.441 -69.786 54.985
347928 92.481 -69.756 54.967
347929 92.520 -69.726 54.947
347930 92.558 -69.714 54.929
347931 92.593 -69.720 54.909
347932 92.634 -69.728 54.888
347933 92.680 -69.737 54.864
347934 92.728 -69.739 54.840
347935 92.776 -69.742 54.813
347936 92.825 -69.744 54.785
347937 92.875 -69.737 54.758
347938 92.925 -69.730 54.727
347939 92.976 -69.724 54.695
347940 93.026 -69.717 54.662
347941 93.075 -69.709 54.628
347942 93.113 -69.684 54.591
347943 93.151 -69.660 54.554
347944 93.189 -69.640 54.518
347945 93.225 -69.644 54.481
347946 93.261 -69.648 54.444
347947 93.297 -69.651 54.409
347948 93.333 -69.654 54.374
347949 93.369 -69.658 54.339
347950 93.406 -69.662 54.305
347951 93.442 -69.666 54.272
347952 93.478 -69.670 54.240
347953 93.514 -69.673 54.208
347954 93.550 -69.677 54.177
347955 93.586 -69.681 54.146
347956 93.620 -69.682 54.116
347991 94.721 -69.599 53.031
347992 94.752 -69.597 53.009
347993 94.784 -69.595 52.987
347994 94.815 -69.591 52.970
347995 94.847 -69.589 52.953
347996 94.878 -69.587 52.937
347997 94.910 -69.585 52.922
347998 94.941 -69.582 52.907
347999 94.972 -69.580 52.895
348000 95.004 -69.578 52.885
348001 95.035 -69.576 52.879
348002 95.067 -69.573 52.874
348003 95.098 -69.570 52.873
348004 95.130 -69.568 52.874
348005 95.209 -69.578 52.878
348006 95.301 -69.591 52.885
348007 95.352 -69.593 52.896
348008 95.380 -69.588 52.912
348009 95.407 -69.582 52.930
348010 95.436 -69.576 52.951
348011 95.463 -69.571 52.974
348012 95.491 -69.566 52.993
348013 95.519 -69.560 53.007
348014 95.546 -69.555 53.016
348015 95.574 -69.550 53.018
348016 95.602 -69.544 53.017
348017 95.630 -69.538 53.015
348018 95.658 -69.533 53.013
348019 95.686 -69.528 53.012
348020 95.721 -69.531 53.015
348021 95.756 -69.535 53.016
348022 95.792 -69.540 53.013
348023 95.827 -69.544 53.003
348024 95.863 -69.549 52.980
348025 95.899 -69.554 52.952
348026 95.935 -69.557 52.923
348027 95.959 -69.562 52.895
348028 95.977 -69.566 52.874
348029 96.008 -69.576 52.857
348030 96.039 -69.586 52.837
348031 96.070 -69.597 52.813
348032 96.101 -69.607 52.787
348033 96.132 -69.617 52.759
348034 96.162 -69.627 52.733
348035 96.188 -69.635 52.708
348036 96.215 -69.647 52.684
348037 96.242 -69.660 52.664
348038 96.268 -69.672 52.646
348039 96.292 -69.685 52.629
348040 96.315 -69.697 52.611
348041 96.338 -69.710 52.591
348042 96.362 -69.722 52.570
348043 96.386 -69.735 52.547
348044 96.409 -69.747 52.524
348045 96.433 -69.760 52.503
348046 96.468 -69.768 52.481
348047 96.513 -69.773 52.458
348048 96.551 -69.776 52.437
348049 96.581 -69.780 52.419
348050 96.610 -69.787 52.406
348051 96.639 -69.795 52.395
348052 96.669 -69.805 52.384
348053 96.702 -69.817 52.370
348054 96.736 -69.829 52.356
348055 96.769 -69.842 52.344
348056 96.803 -69.853 52.332
348057 96.836 -69.860 52.319
348058 96.869 -69.866 52.306
348059 96.903 -69.872 52.288
348060 96.936 -69.878 52.274
348061 96.969 -69.884 52.265
348062 97.004 -69.891 52.258
348063 97.040 -69.899 52.249
348064 97.079 -69.909 52.235
348065 97.118 -69.913 52.214
348066 97.154 -69.922 52.188
348067 97.190 -69.931 52.164
348068 97.226 -69.940 52.147
348069 97.261 -69.950 52.131
348070 97.297 -69.959 52.115
348071 97.333 -69.969 52.096
348072 97.363 -69.984 52.069
348073 97.386 -70.008 52.039
348074 97.413 -70.024 52.006
348075 97.443 -70.033 51.972
348076 97.473 -70.043 51.942
348077 97.504 -70.052 51.913
348078 97.534 -70.061 51.881
348079 97.565 -70.070 51.849
348080 97.595 -70.079 51.816
348081 97.626 -70.088 51.782
348082 97.651 -70.093 51.748
348083 97.684 -70.092 51.714
348084 97.717 -70.094 51.679
348085 97.749 -70.096 51.638
348086 97.782 -70.098 51.596
348087 97.815 -70.099 51.552
348088 97.847 -70.102 51.508
348089 97.880 -70.103 51.463
348090 97.914 -70.102 51.419
348091 97.948 -70.101 51.372
348092 97.982 -70.099 51.323
348093 98.016 -70.098 51.276
348094 98.051 -70.096 51.227
348095 98.085 -70.095 51.180
348096 98.119 -70.094 51.132
348097 98.153 -70.093 51.083
348098 98.187 -70.091 51.029
348099 98.220 -70.093 50.974
348100 98.250 -70.100 50.919
348101 98.280 -70.108 50.863
348102 98.310 -70.115 50.807
348103 98.340 -70.123 50.751
348104 98.370 -70.130 50.695
348105 98.400 -70.138 50.636
348106 98.430 -70.146 50.578
348107 98.459 -70.153 50.519
348108 98.490 -70.161 50.460
348109 98.519 -70.168 50.402
348110 98.550 -70.172 50.346
348111 98.580 -70.167 50.288
348112 98.616 -70.171 50.229
348113 98.654 -70.181 50.170
348114 98.693 -70.190 50.110
348115 98.731 -70.199 50.048
348116 98.769 -70.208 49.987
348117 98.809 -70.219 49.926
348118 98.850 -70.232 49.865
348119 98.890 -70.244 49.805
348120 98.931 -70.256 49.745
348121 98.972 -70.269 49.687
348122 99.012 -70.281 49.630
348123 99.053 -70.293 49.575
348124 99.093 -70.306 49.522
348125 99.134 -70.319 49.471
348126 99.174 -70.331 49.420
348127 99.215 -70.344 49.371
348128 99.255 -70.356 49.323
348129 99.291 -70.366 49.277
348130 99.328 -70.377 49.233
348131 99.364 -70.388 49.190
348132 99.404 -70.395 49.148
348133 99.451 -70.394 49.108
348134 99.499 -70.394 49.070
348135 99.541 -70.398 49.033
348136 99.585 -70.402 48.999
348137 99.628 -70.419 48.966
348138 99.659 -70.441 48.933
348139 99.691 -70.454 48.902
348140 99.721 -70.467 48.874
348141 99.753 -70.480 48.847
348142 99.783 -70.493 48.823
348143 99.815 -70.506 48.801
348144 99.845 -70.519 48.779
348145 99.876 -70.532 48.756
348146 99.908 -70.532 48.727
348147 99.940 -70.532 48.697
348148 99.973 -70.540 48.666
348149 100.006 -70.549 48.634
348150 100.039 -70.559 48.608
348151 100.072 -70.568 48.578
348152 100.105 -70.578 48.545
348153 100.138 -70.587 48.508
348154 100.171 -70.597 48.469
348155 100.204 -70.606 48.436
348156 100.238 -70.615 48.406
348157 100.280 -70.626 48.376
348158 100.324 -70.638 48.341
348159 100.366 -70.649 48.302
348160 100.409 -70.660 48.266
348161 100.453 -70.672 48.230
348162 100.498 -70.684 48.193
348163 100.538 -70.695 48.153
348164 100.572 -70.704 48.113
348165 100.607 -70.713 48.077
348166 100.641 -70.723 48.044
348167 100.676 -70.733 48.010
348168 100.710 -70.743 47.971
348169 100.745 -70.753 47.931
348170 100.779 -70.763 47.890
348171 100.814 -70.773 47.855
348172 100.848 -70.782 47.818
348173 100.882 -70.792 47.782
348174 100.908 -70.808 47.748
348175 100.937 -70.816 47.713
348176 100.970 -70.820 47.676
348177 101.004 -70.825 47.639
348178 101.040 -70.836 47.601
348179 101.076 -70.845 47.562
348180 101.112 -70.855 47.523
348181 101.148 -70.866 47.484
348182 101.179 -70.880 47.446
348183 101.206 -70.897 47.408
348184 101.236 -70.914 47.369
348185 101.267 -70.931 47.327
348186 101.305 -70.948 47.284
348187 101.358 -70.969 47.239
348188 101.410 -70.989 47.191
348189 101.463 -71.010 47.143
348190 101.516 -71.030 47.095
348191 101.568 -71.051 47.047
348192 101.620 -71.071 47.000
348193 101.673 -71.092 46.953
348194 101.726 -71.112 46.906
348195 101.777 -71.133 46.859
348196 101.817 -71.149 46.811
348197 101.816 -71.127 46.760
348198 101.799 -71.096 46.707
348199 101.809 -71.092 46.655
348200 101.839 -71.093 46.604
348201 101.870 -71.095 46.556
348202 101.899 -71.096 46.509
348203 101.929 -71.098 46.465
348204 101.959 -71.099 46.421
348205 101.989 -71.101 46.379
348206 102.020 -71.102 46.338
348207 102.049 -71.104 46.297
348208 102.080 -71.105 46.257
348209 102.109 -71.107 46.216
348210 102.139 -71.108 46.176
348621 92.157 -69.964 55.095
348622 92.196 -69.934 55.079
348623 92.235 -69.904 55.062
348624 92.275 -69.874 55.045
348625 92.314 -69.845 55.028
348626 92.354 -69.815 55.010
348627 92.393 -69.786 54.992
348628 92.432 -69.755 54.974
348629 92.472 -69.726 54.956
348630 92.511 -69.696 54.935
348631 92.550 -69.667 54.915
348632 92.587 -69.664 54.895
348633 92.626 -69.671 54.875
348634 92.673 -69.678 54.851
348635 92.722 -69.680 54.826
348636 92.770 -69.682 54.800
348637 92.818 -69.685 54.771
348638 92.866 -69.686 54.743
348639 92.917 -69.679 54.713
348640 92.967 -69.673 54.683
348641 93.017 -69.666 54.651
348642 93.066 -69.657 54.617
348643 93.104 -69.632 54.581
348644 93.142 -69.607 54.546
348645 93.178 -69.605 54.510
348646 93.214 -69.609 54.474
348647 93.251 -69.613 54.439
348648 93.287 -69.617 54.404
348649 93.323 -69.620 54.369
348650 93.359 -69.624 54.336
348651 93.395 -69.628 54.302
348652 93.431 -69.632 54.271
348653 93.468 -69.635 54.238
348654 93.503 -69.639 54.207
348655 93.540 -69.642 54.177
348656 93.576 -69.646 54.146
348657 93.612 -69.650 54.116
348692 94.720 -69.577 53.044
348693 94.752 -69.575 53.023
348694 94.783 -69.573 53.001
348695 94.815 -69.570 52.984
348696 94.846 -69.568 52.967
348697 94.878 -69.565 52.951
348698 94.909 -69.563 52.936
348699 94.941 -69.560 52.920
348700 94.972 -69.558 52.908
348701 95.003 -69.556 52.899
348702 95.035 -69.554 52.893
348703 95.066 -69.551 52.888
348704 95.098 -69.549 52.886
348705 95.158 -69.554 52.887
348706 95.250 -69.567 52.890
348707 95.342 -69.580 52.896
348708 95.375 -69.576 52.906
348709 95.402 -69.571 52.918
348710 95.430 -69.566 52.932
348711 95.458 -69.560 52.950
348712 95.486 -69.555 52.970
348713 95.514 -69.549 52.988
348714 95.541 -69.544 53.003
348715 95.569 -69.538 53.013
348716 95.597 -69.533 53.020
348717 95.625 -69.528 53.026
348718 95.653 -69.523 53.028
348719 95.680 -69.517 53.028
348720 95.708 -69.511 53.026
348721 95.736 -69.506 53.027
348722 95.764 -69.501 53.030
348723 95.798 -69.503 53.031
348724 95.834 -69.507 53.026
348725 95.869 -69.512 53.008
348726 95.905 -69.516 52.979
348727 95.941 -69.521 52.947
348728 95.976 -69.526 52.915
348729 95.997 -69.529 52.888
348730 96.021 -69.536 52.871
348731 96.052 -69.547 52.850
348732 96.084 -69.557 52.825
348733 96.115 -69.567 52.798
348734 96.146 -69.578 52.772
348735 96.177 -69.588 52.748
348736 96.207 -69.598 52.723
348737 96.233 -69.610 52.699
348738 96.257 -69.623 52.678
348739 96.281 -69.635 52.661
348740 96.304 -69.648 52.645
348741 96.328 -69.660 52.627
348742 96.352 -69.673 52.609
348743 96.375 -69.685 52.586
348744 96.399 -69.698 52.563
348745 96.422 -69.710 52.541
348746 96.446 -69.723 52.517
348747 96.481 -69.731 52.495
348748 96.526 -69.736 52.472
348749 96.557 -69.738 52.451
348750 96.588 -69.740 52.433
348751 96.617 -69.747 52.418
348752 96.647 -69.756 52.405
348753 96.680 -69.768 52.391
348754 96.714 -69.780 52.376
348755 96.747 -69.792 52.361
348756 96.781 -69.805 52.350
348757 96.814 -69.817 52.338
348758 96.848 -69.824 52.325
348759 96.881 -69.830 52.313
348760 96.914 -69.837 52.297
348761 96.948 -69.843 52.286
348762 96.981 -69.849 52.278
348763 97.014 -69.855 52.273
348764 97.047 -69.861 52.266
348765 97.083 -69.869 52.250
348766 97.119 -69.878 52.229
348767 97.155 -69.888 52.202
348768 97.190 -69.897 52.173
348769 97.227 -69.907 52.151
348770 97.262 -69.916 52.132
348771 97.298 -69.925 52.115
348772 97.334 -69.935 52.098
348773 97.369 -69.945 52.073
348774 97.392 -69.968 52.045
348775 97.421 -69.980 52.013
348776 97.451 -69.990 51.981
348777 97.481 -69.999 51.953
348778 97.512 -70.008 51.925
348779 97.542 -70.017 51.895
348780 97.573 -70.026 51.864
348781 97.603 -70.035 51.831
348782 97.634 -70.044 51.796
348783 97.664 -70.050 51.761
348784 97.696 -70.052 51.727
348785 97.729 -70.054 51.690
348786 97.761 -70.056 51.646
348787 97.794 -70.058 51.602
348788 97.826 -70.060 51.557
348789 97.859 -70.062 51.512
348790 97.891 -70.064 51.468
348791 97.925 -70.063 51.422
348792 97.959 -70.062 51.375
348793 97.994 -70.060 51.325
348794 98.028 -70.058 51.277
348795 98.062 -70.057 51.228
348796 98.096 -70.055 51.180
348797 98.131 -70.055 51.133
348798 98.165 -70.053 51.085
348799 98.198 -70.055 51.032
348800 98.227 -70.062 50.977
348801 98.257 -70.070 50.921
348802 98.287 -70.077 50.864
348803 98.317 -70.085 50.806
348804 98.347 -70.093 50.751
348805 98.377 -70.100 50.693
348806 98.407 -70.108 50.634
348807 98.437 -70.115 50.576
348808 98.466 -70.123 50.517
348809 98.497 -70.131 50.459
348810 98.526 -70.139 50.402
348811 98.557 -70.146 50.347
348812 98.587 -70.140 50.290
348813 98.620 -70.138 50.230
348814 98.658 -70.147 50.169
348815 98.696 -70.156 50.108
348816 98.734 -70.165 50.045
348817 98.773 -70.174 49.984
348818 98.813 -70.187 49.923
348819 98.854 -70.199 49.863
348820 98.895 -70.212 49.802
348821 98.935 -70.225 49.743
348822 98.976 -70.237 49.685
348823 99.016 -70.250 49.628
348824 99.057 -70.262 49.572
348825 99.097 -70.274 49.520
348826 99.138 -70.287 49.469
348827 99.179 -70.299 49.419
348828 99.219 -70.311 49.370
348829 99.260 -70.324 49.322
348830 99.297 -70.334 49.276
348831 99.334 -70.346 49.232
348832 99.370 -70.356 49.190
348833 99.417 -70.356 49.148
348834 99.463 -70.357 49.107
348835 99.505 -70.361 49.069
348836 99.548 -70.366 49.033
348837 99.594 -70.369 48.998
348838 99.642 -70.375 48.965
348839 99.674 -70.414 48.932
348840 99.705 -70.429 48.901
348841 99.736 -70.442 48.872
348842 99.767 -70.455 48.845
348843 99.798 -70.468 48.820
348844 99.829 -70.481 48.797
348845 99.860 -70.494 48.777
348846 99.891 -70.507 48.756
348847 99.922 -70.508 48.729
348848 99.954 -70.508 48.700
348849 99.986 -70.508 48.671
348850 100.018 -70.510 48.640
348851 100.051 -70.519 48.614
348852 100.084 -70.529 48.585
348853 100.117 -70.538 48.552
348854 100.150 -70.547 48.516
348855 100.184 -70.557 48.476
348856 100.216 -70.566 48.445
348857 100.250 -70.576 48.417
348858 100.292 -70.587 48.388
348859 100.335 -70.598 48.354
348860 100.378 -70.610 48.316
348861 100.421 -70.621 48.279
348862 100.467 -70.633 48.243
348863 100.515 -70.646 48.204
348864 100.565 -70.659 48.164
348865 100.603 -70.669 48.126
348866 100.637 -70.679 48.090
348867 100.671 -70.688 48.057
348868 100.706 -70.698 48.020
348869 100.741 -70.708 47.980
348870 100.775 -70.718 47.941
348871 100.810 -70.727 47.904
348872 100.844 -70.737 47.870
348873 100.879 -70.747 47.833
348874 100.901 -70.765 47.797
348875 100.923 -70.782 47.761
348876 100.956 -70.785 47.724
348877 100.989 -70.789 47.685
348878 101.022 -70.792 47.647
348879 101.054 -70.796 47.609
348880 101.090 -70.805 47.570
348881 101.126 -70.815 47.532
348882 101.160 -70.827 47.493
348883 101.188 -70.844 47.454
348884 101.215 -70.861 47.416
348885 101.243 -70.877 47.377
348886 101.274 -70.894 47.336
348887 101.313 -70.912 47.293
348888 101.365 -70.933 47.247
348889 101.418 -70.953 47.199
348890 101.471 -70.974 47.150
348891 101.523 -70.994 47.101
348892 101.575 -71.014 47.053
348893 101.628 -71.035 47.004
348894 101.681 -71.055 46.958
348895 101.733 -71.076 46.911
348896 101.785 -71.096 46.864
348897 101.818 -71.095 46.816
348898 101.785 -71.046 46.763
348899 101.793 -71.039 46.709
348900 101.823 -71.041 46.656
348901 101.853 -71.042 46.605
348902 101.883 -71.044 46.557
348903 101.913 -71.046 46.511
348904 101.943 -71.047 46.466
348905 101.973 -71.049 46.423
348906 102.004 -71.051 46.381
348907 102.033 -71.052 46.339
348908 102.063 -71.054 46.299
348909 102.093 -71.055 46.258
348910 102.123 -71.057 46.217
349323 92.187 -69.904 55.072
349324 92.226 -69.874 55.055
349325 92.266 -69.845 55.037
349326 92.305 -69.814 55.018
349327 92.345 -69.785 55.000
349328 92.384 -69.755 54.981
349329 92.423 -69.726 54.963
349330 92.462 -69.696 54.944
349331 92.502 -69.666 54.924
349332 92.541 -69.636 54.903
349333 92.580 -69.607 54.882
349334 92.619 -69.614 54.861
349335 92.667 -69.618 54.837
349336 92.715 -69.621 54.813
349337 92.763 -69.623 54.787
349338 92.811 -69.626 54.758
349339 92.859 -69.628 54.730
349340 92.908 -69.629 54.701
349341 92.958 -69.622 54.671
349342 93.009 -69.615 54.640
349343 93.057 -69.604 54.607
349344 93.094 -69.579 54.573
349345 93.132 -69.567 54.538
349346 93.167 -69.571 54.502
349347 93.204 -69.575 54.466
349348 93.240 -69.579 54.431
349349 93.276 -69.582 54.397
349350 93.312 -69.585 54.364
349351 93.348 -69.589 54.331
349352 93.384 -69.593 54.297
349353 93.421 -69.597 54.265
349354 93.457 -69.601 54.234
349355 93.493 -69.604 54.204
349356 93.529 -69.608 54.173
349357 93.565 -69.612 54.144
349358 93.602 -69.616 54.115
349392 94.688 -69.558 53.079
349393 94.719 -69.555 53.058
349394 94.751 -69.553 53.037
349395 94.782 -69.551 53.016
349396 94.814 -69.548 52.999
349397 94.845 -69.546 52.982
349398 94.877 -69.544 52.966
349399 94.908 -69.541 52.950
349400 94.939 -69.539 52.935
349401 94.971 -69.536 52.924
349402 95.002 -69.534 52.915
349403 95.034 -69.532 52.909
349404 95.065 -69.529 52.904
349405 95.107 -69.529 52.901
349406 95.199 -69.543 52.901
349407 95.291 -69.557 52.903
349408 95.369 -69.566 52.908
349409 95.397 -69.560 52.917
349410 95.425 -69.554 52.926
349411 95.453 -69.549 52.936
349412 95.480 -69.544 52.949
349413 95.508 -69.538 52.967
349414 95.536 -69.533 52.983
349415 95.564 -69.528 52.997
349416 95.592 -69.523 53.009
349417 95.619 -69.516 53.021
349418 95.648 -69.511 53.031
349419 95.675 -69.506 53.038
349420 95.703 -69.501 53.042
349421 95.731 -69.495 53.041
349422 95.758 -69.490 53.040
349423 95.787 -69.485 53.045
349424 95.814 -69.479 53.049
349425 95.842 -69.473 53.048
349426 95.876 -69.475 53.032
349427 95.911 -69.479 53.004
349428 95.947 -69.484 52.971
349429 95.982 -69.488 52.937
349430 96.018 -69.493 52.905
349431 96.036 -69.497 52.884
349432 96.066 -69.507 52.862
349433 96.097 -69.517 52.838
349434 96.128 -69.528 52.810
349435 96.160 -69.538 52.786
349436 96.191 -69.548 52.763
349437 96.218 -69.562 52.738
349438 96.244 -69.574 52.713
349439 96.270 -69.585 52.691
349440 96.294 -69.598 52.675
349441 96.317 -69.610 52.660
349442 96.341 -69.623 52.643
349443 96.364 -69.635 52.624
349444 96.388 -69.648 52.603
349445 96.411 -69.660 52.580
349446 96.435 -69.673 52.558
349447 96.459 -69.686 52.537
349448 96.495 -69.694 52.515
349449 96.534 -69.697 52.494
349450 96.565 -69.699 52.472
349451 96.595 -69.701 52.452
349452 96.625 -69.707 52.436
349453 96.658 -69.719 52.420
349454 96.692 -69.732 52.404
349455 96.725 -69.744 52.387
349456 96.759 -69.756 52.373
349457 96.792 -69.768 52.360
349458 96.825 -69.781 52.349
349459 96.859 -69.789 52.337
349460 96.892 -69.795 52.325
349461 96.926 -69.801 52.309
349462 96.959 -69.808 52.299
349463 96.992 -69.814 52.292
349464 97.025 -69.820 52.287
349465 97.056 -69.827 52.277
349466 97.088 -69.834 52.261
349467 97.123 -69.843 52.239
349468 97.158 -69.853 52.211
349469 97.192 -69.863 52.180
349470 97.227 -69.873 52.155
349471 97.263 -69.882 52.134
349472 97.299 -69.891 52.116
349473 97.335 -69.901 52.100
349474 97.371 -69.910 52.076
349475 97.398 -69.928 52.047
349476 97.429 -69.937 52.017
349477 97.459 -69.946 51.987
349478 97.490 -69.955 51.960
349479 97.520 -69.964 51.934
349480 97.550 -69.974 51.906
349481 97.581 -69.983 51.874
349482 97.611 -69.993 51.840
349483 97.641 -70.003 51.806
349484 97.673 -70.010 51.770
349485 97.708 -70.013 51.732
349486 97.740 -70.015 51.693
349487 97.773 -70.017 51.647
349488 97.805 -70.018 51.601
349489 97.838 -70.020 51.555
349490 97.870 -70.022 51.510
349491 97.903 -70.024 51.464
349492 97.937 -70.024 51.420
349493 97.971 -70.022 51.373
349494 98.005 -70.021 51.324
349495 98.039 -70.019 51.274
349496 98.074 -70.018 51.224
349497 98.108 -70.016 51.175
349498 98.142 -70.015 51.127
349499 98.174 -70.017 51.079
349500 98.205 -70.024 51.026
349501 98.234 -70.032 50.972
349502 98.265 -70.040 50.916
349503 98.294 -70.047 50.858
349504 98.324 -70.055 50.800
349505 98.354 -70.062 50.744
349506 98.384 -70.070 50.686
349507 98.414 -70.078 50.627
349508 98.444 -70.086 50.569
349509 98.474 -70.093 50.513
349510 98.504 -70.101 50.457
349511 98.533 -70.109 50.402
349512 98.564 -70.116 50.346
349513 98.594 -70.113 50.288
349514 98.625 -70.108 50.227
349515 98.661 -70.114 50.165
349516 98.700 -70.123 50.102
349517 98.737 -70.132 50.040
349518 98.777 -70.143 49.978
349519 98.817 -70.155 49.918
349520 98.858 -70.168 49.857
349521 98.899 -70.180 49.797
349522 98.939 -70.192 49.739
349523 98.980 -70.205 49.682
349524 99.020 -70.217 49.625
349525 99.061 -70.229 49.571
349526 99.102 -70.242 49.518
349527 99.142 -70.254 49.467
349528 99.183 -70.266 49.417
349529 99.224 -70.279 49.369
349530 99.264 -70.291 49.321
349531 99.303 -70.303 49.275
349532 99.340 -70.314 49.231
349533 99.383 -70.318 49.188
349534 99.427 -70.320 49.146
349535 99.470 -70.324 49.107
349536 99.512 -70.328 49.068
349537 99.554 -70.332 49.030
349538 99.602 -70.335 48.995
349539 99.652 -70.338 48.961
349540 99.688 -70.370 48.930
349541 99.719 -70.404 48.899
349542 99.750 -70.417 48.870
349543 99.781 -70.430 48.842
349544 99.812 -70.443 48.818
349545 99.843 -70.456 48.796
349546 99.874 -70.469 48.775
349547 99.905 -70.481 48.754
349548 99.937 -70.484 48.728
349549 99.969 -70.484 48.701
349550 100.000 -70.484 48.673
349551 100.032 -70.484 48.644
349552 100.064 -70.484 48.617
349553 100.096 -70.490 48.587
349554 100.129 -70.499 48.555
349555 100.162 -70.508 48.518
349556 100.195 -70.518 48.483
349557 100.228 -70.527 48.453
349558 100.262 -70.537 48.425
349559 100.303 -70.547 48.397
349560 100.346 -70.559 48.364
349561 100.390 -70.570 48.326
349562 100.434 -70.582 48.290
349563 100.484 -70.595 48.254
349564 100.533 -70.608 48.213
349565 100.582 -70.620 48.173
349566 100.631 -70.633 48.136
349567 100.668 -70.644 48.102
349568 100.702 -70.654 48.068
349569 100.737 -70.663 48.030
349570 100.771 -70.673 47.991
349571 100.805 -70.682 47.954
349572 100.840 -70.692 47.918
349573 100.874 -70.703 47.884
349574 100.895 -70.722 47.847
349575 100.915 -70.742 47.810
349576 100.943 -70.752 47.773
349577 100.975 -70.755 47.735
349578 101.008 -70.759 47.694
349579 101.041 -70.762 47.654
349580 101.073 -70.766 47.615
349581 101.106 -70.769 47.577
349582 101.140 -70.775 47.539
349583 101.169 -70.790 47.499
349584 101.197 -70.807 47.461
349585 101.225 -70.823 47.423
349586 101.252 -70.840 47.384
349587 101.281 -70.857 47.343
349588 101.321 -70.876 47.299
349589 101.373 -70.896 47.253
349590 101.425 -70.917 47.205
349591 101.478 -70.936 47.155
349592 101.531 -70.957 47.107
349593 101.583 -70.977 47.059
349594 101.633 -70.991 47.011
349595 101.684 -71.005 46.964
349596 101.735 -71.020 46.917
349597 101.785 -71.034 46.868
349598 101.773 -70.996 46.818
349599 101.777 -70.988 46.765
349600 101.807 -70.989 46.709
349601 101.837 -70.991 46.656
349602 101.867 -70.992 46.605
349603 101.897 -70.994 46.557
349604 101.927 -70.995 46.512
349605 101.957 -70.997 46.468
349606 101.987 -70.999 46.425
349607 102.017 -71.000 46.383
349608 102.047 -71.002 46.341
349609 102.077 -71.003 46.300
349610 102.107 -71.005 46.260
349611 102.137 -71.006 46.220
350024 92.178 -69.874 55.065
350025 92.217 -69.844 55.048
350026 92.257 -69.814 55.029
350027 92.296 -69.785 55.009
350028 92.336 -69.754 54.991
350029 92.375 -69.725 54.971
350030 92.414 -69.695 54.952
350031 92.453 -69.666 54.933
350032 92.493 -69.636 54.912
350033 92.532 -69.606 54.890
350034 92.571 -69.576 54.869
350035 92.612 -69.557 54.847
350036 92.660 -69.559 54.825
350037 92.708 -69.562 54.800
350038 92.757 -69.564 54.774
350039 92.804 -69.566 54.746
350040 92.853 -69.569 54.717
350041 92.901 -69.571 54.689
350042 92.950 -69.570 54.660
350043 93.000 -69.564 54.629
350044 93.047 -69.552 54.598
350045 93.085 -69.529 54.564
350046 93.121 -69.532 54.529
350047 93.157 -69.536 54.492
350048 93.193 -69.540 54.457
350049 93.229 -69.544 54.423
350050 93.266 -69.548 54.389
350051 93.302 -69.551 54.356
350052 93.338 -69.555 54.324
350053 93.374 -69.559 54.290
350054 93.410 -69.563 54.259
350055 93.446 -69.566 54.228
350056 93.482 -69.570 54.198
350057 93.518 -69.573 54.168
350092 94.656 -69.538 53.115
350093 94.687 -69.536 53.094
350094 94.719 -69.534 53.073
350095 94.750 -69.532 53.052
350096 94.782 -69.529 53.032
350097 94.813 -69.526 53.015
350098 94.845 -69.524 52.999
350099 94.876 -69.522 52.983
350100 94.907 -69.519 52.968
350101 94.939 -69.517 52.954
350102 94.970 -69.515 52.942
350103 95.002 -69.513 52.933
350104 95.033 -69.510 52.926
350105 95.065 -69.507 52.922
350106 95.148 -69.519 52.919
350107 95.241 -69.532 52.917
350108 95.333 -69.545 52.918
350109 95.392 -69.549 52.921
350110 95.419 -69.544 52.929
350111 95.448 -69.538 52.938
350112 95.475 -69.533 52.945
350113 95.503 -69.527 52.950
350114 95.531 -69.522 52.964
350115 95.558 -69.516 52.978
350116 95.587 -69.511 52.993
350117 95.614 -69.506 53.006
350118 95.642 -69.501 53.020
350119 95.670 -69.495 53.033
350120 95.698 -69.490 53.045
350121 95.726 -69.484 53.052
350122 95.753 -69.479 53.055
350123 95.781 -69.473 53.055
350124 95.809 -69.468 53.060
350125 95.837 -69.463 53.066
350126 95.865 -69.457 53.066
350127 95.892 -69.452 53.053
350128 95.920 -69.446 53.027
350129 95.953 -69.447 52.995
350130 95.988 -69.451 52.959
350131 96.024 -69.456 52.924
350132 96.056 -69.460 52.898
350133 96.080 -69.467 52.874
350134 96.111 -69.477 52.850
350135 96.143 -69.488 52.822
350136 96.173 -69.498 52.799
350137 96.200 -69.513 52.776
350138 96.223 -69.529 52.752
350139 96.248 -69.541 52.726
350140 96.274 -69.552 52.706
350141 96.301 -69.563 52.691
350142 96.327 -69.575 52.676
350143 96.354 -69.585 52.660
350144 96.377 -69.598 52.640
350145 96.401 -69.610 52.620
350146 96.425 -69.623 52.599
350147 96.448 -69.636 52.578
350148 96.472 -69.648 52.560
350149 96.508 -69.656 52.539
350150 96.541 -69.659 52.518
350151 96.571 -69.661 52.496
350152 96.603 -69.665 52.473
350153 96.636 -69.673 52.455
350154 96.669 -69.685 52.437
350155 96.703 -69.696 52.419
350156 96.737 -69.707 52.401
350157 96.770 -69.720 52.386
350158 96.804 -69.732 52.374
350159 96.837 -69.745 52.362
350160 96.870 -69.754 52.350
350161 96.904 -69.761 52.337
350162 96.937 -69.767 52.321
350163 96.969 -69.773 52.311
350164 97.001 -69.780 52.303
350165 97.032 -69.787 52.296
350166 97.064 -69.794 52.284
350167 97.096 -69.800 52.268
350168 97.129 -69.808 52.246
350169 97.164 -69.818 52.218
350170 97.198 -69.827 52.188
350171 97.233 -69.837 52.160
350172 97.268 -69.847 52.136
350173 97.303 -69.857 52.116
350174 97.338 -69.866 52.100
350175 97.373 -69.876 52.075
350176 97.406 -69.885 52.047
350177 97.436 -69.895 52.018
350178 97.466 -69.905 51.990
350179 97.496 -69.914 51.964
350180 97.526 -69.924 51.940
350181 97.557 -69.934 51.912
350182 97.587 -69.944 51.880
350183 97.616 -69.954 51.846
350184 97.647 -69.964 51.810
350185 97.681 -69.969 51.773
350186 97.717 -69.972 51.734
350187 97.752 -69.975 51.694
350188 97.785 -69.977 51.648
350189 97.817 -69.979 51.602
350190 97.850 -69.980 51.556
350191 97.882 -69.983 51.512
350192 97.915 -69.984 51.468
350193 97.948 -69.983 51.425
350194 97.983 -69.983 51.378
350195 98.017 -69.981 51.329
350196 98.051 -69.980 51.280
350197 98.085 -69.978 51.230
350198 98.120 -69.977 51.179
350199 98.152 -69.979 51.127
350200 98.181 -69.986 51.076
350201 98.212 -69.994 51.021
350202 98.241 -70.002 50.965
350203 98.272 -70.009 50.909
350204 98.301 -70.017 50.853
350205 98.332 -70.025 50.797
350206 98.361 -70.033 50.742
350207 98.391 -70.040 50.685
350208 98.421 -70.048 50.627
350209 98.451 -70.055 50.568
350210 98.481 -70.063 50.510
350211 98.511 -70.071 50.455
350212 98.541 -70.078 50.399
350213 98.571 -70.086 50.344
350214 98.601 -70.087 50.284
350215 98.632 -70.081 50.222
350216 98.664 -70.080 50.160
350217 98.703 -70.090 50.097
350218 98.741 -70.099 50.035
350219 98.781 -70.110 49.974
350220 98.822 -70.123 49.913
350221 98.862 -70.135 49.853
350222 98.903 -70.147 49.793
350223 98.944 -70.160 49.736
350224 98.984 -70.172 49.679
350225 99.025 -70.184 49.623
350226 99.065 -70.197 49.569
350227 99.106 -70.209 49.517
350228 99.147 -70.221 49.466
350229 99.187 -70.234 49.416
350230 99.228 -70.246 49.368
350231 99.269 -70.259 49.321
350232 99.309 -70.272 49.274
350233 99.349 -70.280 49.229
350234 99.392 -70.283 49.187
350235 99.434 -70.287 49.145
350236 99.476 -70.291 49.105
350237 99.518 -70.295 49.066
350238 99.560 -70.299 49.028
350239 99.611 -70.302 48.991
350240 99.660 -70.305 48.958
350241 99.701 -70.326 48.926
350242 99.733 -70.365 48.896
350243 99.765 -70.392 48.866
350244 99.795 -70.405 48.838
350245 99.827 -70.418 48.814
350246 99.857 -70.431 48.793
350247 99.889 -70.444 48.772
350248 99.919 -70.456 48.750
350249 99.951 -70.460 48.725
350250 99.983 -70.460 48.699
350251 100.014 -70.460 48.672
350252 100.046 -70.460 48.644
350253 100.078 -70.460 48.617
350254 100.110 -70.460 48.587
350255 100.141 -70.460 48.554
350256 100.174 -70.469 48.518
350257 100.207 -70.478 48.488
350258 100.241 -70.488 48.460
350259 100.274 -70.497 48.432
350260 100.315 -70.508 48.405
350261 100.358 -70.519 48.372
350262 100.402 -70.532 48.335
350263 100.452 -70.544 48.299
350264 100.501 -70.557 48.262
350265 100.550 -70.569 48.220
350266 100.599 -70.582 48.181
350267 100.648 -70.595 48.146
350268 100.697 -70.608 48.112
350269 100.733 -70.618 48.078
350270 100.767 -70.628 48.039
350271 100.802 -70.638 48.002
350272 100.836 -70.648 47.967
350273 100.867 -70.660 47.931
350274 100.888 -70.679 47.897
350275 100.909 -70.698 47.859
350276 100.930 -70.717 47.821
350277 100.962 -70.722 47.783
350278 100.994 -70.725 47.745
350279 101.027 -70.729 47.704
350280 101.060 -70.732 47.661
350281 101.093 -70.735 47.620
350282 101.125 -70.739 47.581
350283 101.152 -70.743 47.543
350284 101.179 -70.757 47.502
350285 101.206 -70.773 47.463
350286 101.234 -70.788 47.426
350287 101.261 -70.804 47.388
350288 101.289 -70.820 47.347
350289 101.328 -70.839 47.304
350290 101.380 -70.858 47.258
350291 101.430 -70.872 47.209
350292 101.481 -70.886 47.160
350293 101.532 -70.901 47.111
350294 101.583 -70.915 47.062
350295 101.633 -70.930 47.015
350296 101.684 -70.944 46.968
350297 101.735 -70.958 46.920
350298 101.775 -70.964 46.870
350299 101.761 -70.936 46.818
350300 101.791 -70.937 46.764
350301 101.821 -70.939 46.707
350302 101.851 -70.940 46.654
350303 101.881 -70.942 46.604
350304 101.911 -70.943 46.557
350305 101.941 -70.945 46.513
350306 101.971 -70.946 46.468
350307 102.001 -70.948 46.425
350308 102.031 -70.949 46.383
350309 102.061 -70.951 46.341
350310 102.091 -70.952 46.301
350311 102.121 -70.954 46.261
350725 92.169 -69.844 55.058
350726 92.208 -69.814 55.040
350727 92.247 -69.784 55.020
350728 92.287 -69.754 55.000
350729 92.326 -69.725 54.981
350730 92.366 -69.695 54.961
350731 92.405 -69.665 54.941
350732 92.444 -69.635 54.921
350733 92.484 -69.606 54.900
350734 92.523 -69.576 54.877
350735 92.562 -69.547 54.855
350736 92.607 -69.532 54.832
350737 92.655 -69.524 54.809
350738 92.702 -69.516 54.785
350739 92.750 -69.510 54.759
350740 92.798 -69.507 54.733
350741 92.846 -69.510 54.705
350742 92.894 -69.512 54.676
350743 92.942 -69.514 54.647
350744 92.992 -69.513 54.618
350745 93.038 -69.500 54.588
350746 93.074 -69.494 54.553
350747 93.111 -69.498 54.518
350748 93.147 -69.502 54.482
350749 93.183 -69.506 54.446
350750 93.219 -69.509 54.413
350751 93.255 -69.513 54.379
350752 93.291 -69.516 54.347
350753 93.327 -69.520 54.315
350754 93.363 -69.524 54.283
350755 93.400 -69.528 54.252
350756 93.436 -69.532 54.221
350757 93.472 -69.535 54.191
350792 94.624 -69.519 53.151
350793 94.655 -69.516 53.130
350794 94.687 -69.514 53.110
350795 94.718 -69.512 53.089
350796 94.750 -69.510 53.069
350797 94.781 -69.507 53.049
350798 94.813 -69.505 53.033
350799 94.844 -69.502 53.017
350800 94.876 -69.500 53.002
350801 94.907 -69.497 52.987
350802 94.938 -69.495 52.973
350803 94.970 -69.493 52.962
350804 95.001 -69.491 52.953
350805 95.033 -69.488 52.946
350806 95.098 -69.494 52.942
350807 95.190 -69.508 52.939
350808 95.282 -69.521 52.937
350809 95.374 -69.535 52.936
350810 95.414 -69.532 52.937
350811 95.442 -69.527 52.943
350812 95.470 -69.522 52.950
350813 95.498 -69.516 52.956
350814 95.526 -69.511 52.961
350815 95.553 -69.506 52.970
350816 95.581 -69.501 52.980
350817 95.609 -69.494 52.991
350818 95.637 -69.489 53.004
350819 95.665 -69.484 53.019
350820 95.692 -69.479 53.034
350821 95.720 -69.473 53.048
350822 95.748 -69.468 53.059
350823 95.776 -69.463 53.066
350824 95.804 -69.457 53.068
350825 95.831 -69.451 53.073
350826 95.860 -69.446 53.078
350827 95.887 -69.441 53.079
350828 95.915 -69.435 53.069
350829 95.943 -69.430 53.047
350830 95.970 -69.425 53.017
350831 95.999 -69.419 52.982
350832 96.030 -69.419 52.946
350833 96.066 -69.423 52.916
350834 96.095 -69.428 52.889
350835 96.125 -69.438 52.863
350836 96.156 -69.448 52.834
350837 96.183 -69.462 52.810
350838 96.203 -69.480 52.786
350839 96.227 -69.495 52.763
350840 96.252 -69.508 52.739
350841 96.278 -69.519 52.721
350842 96.305 -69.530 52.709
350843 96.331 -69.541 52.696
350844 96.358 -69.553 52.680
350845 96.384 -69.564 52.661
350846 96.410 -69.576 52.641
350847 96.437 -69.586 52.621
350848 96.461 -69.599 52.602
350849 96.484 -69.611 52.588
350850 96.517 -69.618 52.566
350851 96.548 -69.622 52.544
350852 96.581 -69.628 52.521
350853 96.613 -69.634 52.496
350854 96.646 -69.642 52.474
350855 96.679 -69.654 52.454
350856 96.713 -69.665 52.435
350857 96.747 -69.676 52.416
350858 96.781 -69.687 52.402
350859 96.814 -69.698 52.389
350860 96.848 -69.710 52.377
350861 96.881 -69.720 52.364
350862 96.912 -69.726 52.349
350863 96.945 -69.733 52.333
350864 96.976 -69.740 52.322
350865 97.008 -69.746 52.312
350866 97.040 -69.753 52.302
350867 97.072 -69.760 52.288
350868 97.103 -69.767 52.272
350869 97.135 -69.773 52.250
350870 97.170 -69.783 52.223
350871 97.205 -69.792 52.195
350872 97.239 -69.801 52.165
350873 97.275 -69.811 52.138
350874 97.309 -69.821 52.116
350875 97.344 -69.831 52.097
350876 97.379 -69.840 52.072
350877 97.411 -69.848 52.045
350878 97.442 -69.855 52.016
350879 97.472 -69.865 51.989
350880 97.502 -69.875 51.965
350881 97.532 -69.885 51.942
350882 97.563 -69.895 51.915
350883 97.592 -69.905 51.883
350884 97.623 -69.914 51.848
350885 97.653 -69.924 51.810
350886 97.688 -69.928 51.772
350887 97.725 -69.931 51.734
350888 97.761 -69.934 51.694
350889 97.796 -69.937 51.649
350890 97.828 -69.939 51.604
350891 97.861 -69.941 51.559
350892 97.893 -69.942 51.517
350893 97.926 -69.945 51.476
350894 97.960 -69.944 51.434
350895 97.994 -69.943 51.388
350896 98.028 -69.942 51.339
350897 98.062 -69.940 51.290
350898 98.096 -69.939 51.239
350899 98.129 -69.941 51.186
350900 98.159 -69.949 51.131
350901 98.189 -69.956 51.075
350902 98.219 -69.964 51.018
350903 98.248 -69.972 50.961
350904 98.279 -69.980 50.905
350905 98.308 -69.987 50.849
350906 98.339 -69.995 50.797
350907 98.368 -70.002 50.744
350908 98.398 -70.010 50.688
350909 98.428 -70.018 50.629
350910 98.458 -70.025 50.569
350911 98.488 -70.033 50.510
350912 98.518 -70.040 50.452
350913 98.548 -70.048 50.396
350914 98.578 -70.055 50.341
350915 98.608 -70.060 50.279
350916 98.639 -70.055 50.217
350917 98.669 -70.049 50.155
350918 98.706 -70.056 50.093
350919 98.745 -70.065 50.030
350920 98.786 -70.078 49.969
350921 98.826 -70.090 49.908
350922 98.867 -70.102 49.849
350923 98.907 -70.115 49.790
350924 98.948 -70.127 49.732
350925 98.989 -70.140 49.676
350926 99.029 -70.152 49.621
350927 99.070 -70.165 49.568
350928 99.111 -70.177 49.517
350929 99.151 -70.190 49.466
350930 99.192 -70.202 49.416
350931 99.232 -70.215 49.367
350932 99.277 -70.228 49.320
350933 99.316 -70.240 49.273
350934 99.356 -70.246 49.228
350935 99.398 -70.250 49.185
350936 99.440 -70.254 49.144
350937 99.482 -70.258 49.104
350938 99.524 -70.262 49.064
350939 99.569 -70.265 49.025
350940 99.619 -70.268 48.988
350941 99.668 -70.272 48.954
350942 99.715 -70.282 48.923
350943 99.747 -70.321 48.891
350944 99.779 -70.359 48.861
350945 99.810 -70.380 48.832
350946 99.841 -70.393 48.809
350947 99.872 -70.406 48.787
350948 99.903 -70.419 48.765
350949 99.934 -70.431 48.743
350950 99.965 -70.435 48.719
350951 99.997 -70.435 48.693
350952 100.028 -70.435 48.667
350953 100.060 -70.435 48.639
350954 100.092 -70.435 48.612
350955 100.124 -70.436 48.582
350956 100.155 -70.436 48.549
350957 100.187 -70.436 48.517
350958 100.219 -70.439 48.490
350959 100.253 -70.449 48.464
350960 100.286 -70.458 48.438
350961 100.326 -70.469 48.411
350962 100.370 -70.481 48.379
350963 100.419 -70.494 48.343
350964 100.469 -70.506 48.306
350965 100.518 -70.519 48.268
350966 100.567 -70.532 48.228
350967 100.616 -70.544 48.194
350968 100.665 -70.557 48.159
350969 100.714 -70.570 48.124
350970 100.763 -70.583 48.089
350971 100.797 -70.593 48.051
350972 100.832 -70.603 48.016
350973 100.860 -70.616 47.980
350974 100.881 -70.635 47.944
350975 100.902 -70.655 47.907
350976 100.923 -70.674 47.869
350977 100.948 -70.688 47.829
350978 100.981 -70.691 47.789
350979 101.013 -70.695 47.749
350980 101.046 -70.698 47.707
350981 101.079 -70.701 47.664
350982 101.108 -70.705 47.622
350983 101.135 -70.710 47.581
350984 101.162 -70.713 47.543
350985 101.190 -70.727 47.502
350986 101.217 -70.742 47.464
350987 101.245 -70.758 47.427
350988 101.272 -70.773 47.389
350989 101.299 -70.789 47.348
350990 101.332 -70.804 47.304
350991 101.381 -70.811 47.258
350992 101.431 -70.816 47.209
350993 101.481 -70.825 47.160
350994 101.532 -70.839 47.111
350995 101.583 -70.854 47.062
350996 101.633 -70.868 47.015
350997 101.684 -70.883 46.969
350998 101.734 -70.897 46.920
350999 101.745 -70.883 46.869
351000 101.775 -70.885 46.816
351001 101.805 -70.886 46.759
351002 101.835 -70.888 46.703
351003 101.865 -70.889 46.651
351004 101.895 -70.891 46.603
351005 101.925 -70.892 46.557
351006 101.955 -70.895 46.512
351007 101.985 -70.896 46.469
351008 102.015 -70.898 46.426
351009 102.045 -70.899 46.383
351010 102.075 -70.901 46.342
351011 102.105 -70.902 46.302
351427 92.199 -69.784 55.031
351428 92.238 -69.754 55.012
351429 92.278 -69.724 54.991
351430 92.317 -69.695 54.971
351431 92.357 -69.665 54.951
351432 92.396 -69.635 54.930
351433 92.435 -69.605 54.909
351434 92.475 -69.576 54.887
351435 92.514 -69.546 54.864
351436 92.556 -69.523 54.841
351437 92.603 -69.515 54.817
351438 92.651 -69.508 54.793
351439 92.698 -69.501 54.769
351440 92.746 -69.493 54.744
351441 92.793 -69.485 54.717
351442 92.841 -69.478 54.689
351443 92.888 -69.470 54.661
351444 92.936 -69.463 54.632
351445 92.984 -69.458 54.604
351446 93.027 -69.456 54.573
351447 93.064 -69.460 54.540
351448 93.100 -69.463 54.505
351449 93.136 -69.467 54.469
351450 93.172 -69.471 54.434
351451 93.208 -69.475 54.401
351452 93.245 -69.479 54.368
351453 93.281 -69.482 54.336
351454 93.317 -69.486 54.304
351455 93.353 -69.490 54.272
351456 93.389 -69.493 54.242
351457 93.425 -69.497 54.212
351458 93.462 -69.501 54.183
351493 94.623 -69.497 53.165
351494 94.654 -69.495 53.146
351495 94.686 -69.492 53.126
351496 94.717 -69.490 53.107
351497 94.749 -69.488 53.087
351498 94.780 -69.485 53.067
351499 94.812 -69.483 53.052
351500 94.843 -69.481 53.037
351501 94.875 -69.479 53.021
351502 94.906 -69.476 53.006
351503 94.937 -69.473 52.994
351504 94.969 -69.471 52.982
351505 95.000 -69.469 52.974
351506 95.047 -69.470 52.968
351507 95.139 -69.484 52.964
351508 95.231 -69.497 52.961
351509 95.323 -69.510 52.958
351510 95.409 -69.522 52.957
351511 95.437 -69.516 52.957
351512 95.465 -69.511 52.960
351513 95.492 -69.506 52.965
351514 95.520 -69.500 52.971
351515 95.548 -69.494 52.978
351516 95.576 -69.489 52.983
351517 95.604 -69.484 52.990
351518 95.631 -69.479 52.999
351519 95.660 -69.473 53.010
351520 95.687 -69.468 53.023
351521 95.715 -69.462 53.037
351522 95.743 -69.457 53.050
351523 95.770 -69.451 53.063
351524 95.799 -69.446 53.074
351525 95.826 -69.441 53.079
351526 95.854 -69.435 53.085
351527 95.882 -69.430 53.089
351528 95.910 -69.425 53.090
351529 95.938 -69.419 53.080
351530 95.965 -69.413 53.060
351531 95.993 -69.408 53.032
351532 96.021 -69.403 52.998
351533 96.049 -69.397 52.961
351534 96.077 -69.392 52.930
351535 96.108 -69.391 52.902
351536 96.138 -69.397 52.875
351537 96.165 -69.412 52.847
351538 96.186 -69.430 52.820
351539 96.208 -69.447 52.796
351540 96.232 -69.462 52.774
351541 96.256 -69.475 52.754
351542 96.283 -69.486 52.740
351543 96.309 -69.497 52.729
351544 96.335 -69.508 52.717
351545 96.362 -69.519 52.702
351546 96.388 -69.531 52.683
351547 96.415 -69.542 52.662
351548 96.441 -69.554 52.643
351549 96.467 -69.565 52.627
351550 96.494 -69.576 52.611
351551 96.526 -69.584 52.590
351552 96.558 -69.590 52.567
351553 96.590 -69.596 52.542
351554 96.623 -69.603 52.518
351555 96.656 -69.611 52.495
351556 96.689 -69.623 52.474
351557 96.723 -69.634 52.454
351558 96.757 -69.645 52.435
351559 96.791 -69.656 52.419
351560 96.824 -69.667 52.403
351561 96.855 -69.671 52.387
351562 96.888 -69.682 52.371
351563 96.920 -69.692 52.354
351564 96.952 -69.699 52.339
351565 96.984 -69.706 52.325
351566 97.016 -69.713 52.313
351567 97.047 -69.720 52.300
351568 97.079 -69.726 52.284
351569 97.111 -69.733 52.267
351570 97.143 -69.739 52.246
351571 97.176 -69.747 52.220
351572 97.211 -69.757 52.192
351573 97.246 -69.767 52.163
351574 97.281 -69.776 52.135
351575 97.315 -69.786 52.109
351576 97.350 -69.795 52.088
351577 97.384 -69.805 52.062
351578 97.416 -69.811 52.035
351579 97.447 -69.818 52.007
351580 97.478 -69.826 51.981
351581 97.508 -69.836 51.959
351582 97.538 -69.845 51.938
351583 97.568 -69.855 51.912
351584 97.598 -69.865 51.881
351585 97.629 -69.875 51.845
351586 97.660 -69.883 51.806
351587 97.696 -69.887 51.768
351588 97.732 -69.890 51.730
351589 97.769 -69.893 51.691
351590 97.805 -69.896 51.648
351591 97.840 -69.899 51.605
351592 97.873 -69.901 51.564
351593 97.905 -69.903 51.524
351594 97.938 -69.905 51.484
351595 97.971 -69.905 51.443
351596 98.005 -69.903 51.398
351597 98.039 -69.902 51.350
351598 98.074 -69.901 51.300
351599 98.106 -69.903 51.248
351600 98.136 -69.911 51.194
351601 98.166 -69.919 51.138
351602 98.196 -69.927 51.083
351603 98.226 -69.934 51.025
351604 98.255 -69.942 50.967
351605 98.286 -69.949 50.911
351606 98.315 -69.957 50.855
351607 98.346 -69.964 50.803
351608 98.375 -69.972 50.748
351609 98.405 -69.980 50.691
351610 98.435 -69.987 50.632
351611 98.465 -69.995 50.570
351612 98.495 -70.002 50.510
351613 98.525 -70.010 50.450
351614 98.555 -70.018 50.393
351615 98.585 -70.025 50.337
351616 98.615 -70.033 50.274
351617 98.645 -70.027 50.213
351618 98.676 -70.022 50.150
351619 98.710 -70.023 50.087
351620 98.749 -70.033 50.025
351621 98.790 -70.046 49.965
351622 98.831 -70.058 49.905
351623 98.871 -70.071 49.845
351624 98.912 -70.083 49.788
351625 98.952 -70.096 49.731
351626 98.994 -70.108 49.675
351627 99.039 -70.121 49.620
351628 99.084 -70.135 49.567
351629 99.129 -70.149 49.515
351630 99.174 -70.162 49.464
351631 99.219 -70.176 49.414
351632 99.256 -70.187 49.366
351633 99.289 -70.198 49.318
351634 99.324 -70.206 49.271
351635 99.362 -70.212 49.226
351636 99.405 -70.217 49.184
351637 99.446 -70.221 49.143
351638 99.489 -70.225 49.102
351639 99.531 -70.229 49.062
351640 99.577 -70.232 49.023
351641 99.627 -70.235 48.985
351642 99.677 -70.238 48.950
351643 99.727 -70.241 48.917
351644 99.760 -70.277 48.884
351645 99.792 -70.315 48.852
351646 99.824 -70.354 48.823
351647 99.855 -70.369 48.797
351648 99.886 -70.381 48.773
351649 99.917 -70.394 48.750
351650 99.948 -70.407 48.728
351651 99.980 -70.411 48.704
351652 100.011 -70.411 48.680
351653 100.043 -70.412 48.656
351654 100.075 -70.412 48.630
351655 100.107 -70.412 48.603
351656 100.138 -70.412 48.575
351657 100.170 -70.412 48.545
351658 100.202 -70.412 48.515
351659 100.233 -70.412 48.490
351660 100.265 -70.412 48.465
351661 100.298 -70.419 48.440
351662 100.338 -70.430 48.413
351663 100.388 -70.443 48.382
351664 100.436 -70.455 48.347
351665 100.486 -70.468 48.310
351666 100.535 -70.481 48.273
351667 100.584 -70.494 48.235
351668 100.633 -70.507 48.202
351669 100.683 -70.519 48.169
351670 100.732 -70.532 48.134
351671 100.781 -70.545 48.098
351672 100.828 -70.557 48.061
351673 100.854 -70.573 48.025
351674 100.875 -70.593 47.988
351675 100.896 -70.612 47.951
351676 100.916 -70.632 47.912
351677 100.937 -70.651 47.873
351678 100.967 -70.658 47.832
351679 101.000 -70.661 47.791
351680 101.033 -70.665 47.750
351681 101.064 -70.668 47.707
351682 101.092 -70.673 47.664
351683 101.119 -70.676 47.621
351684 101.146 -70.680 47.579
351685 101.173 -70.685 47.539
351686 101.200 -70.697 47.500
351687 101.228 -70.712 47.463
351688 101.255 -70.728 47.426
351689 101.282 -70.743 47.388
351690 101.312 -70.758 47.346
351691 101.346 -70.774 47.301
351692 101.383 -70.788 47.255
351693 101.433 -70.793 47.206
351694 101.483 -70.798 47.157
351695 101.533 -70.804 47.108
351696 101.584 -70.809 47.060
351697 101.633 -70.815 47.013
351698 101.684 -70.821 46.967
351699 101.733 -70.834 46.917
351700 101.759 -70.833 46.865
351701 101.789 -70.835 46.810
351702 101.819 -70.836 46.752
351703 101.849 -70.838 46.697
351704 101.879 -70.839 46.647
351705 101.909 -70.841 46.600
351706 101.939 -70.842 46.556
351707 101.969 -70.844 46.511
351708 101.999 -70.845 46.468
351709 102.029 -70.847 46.426
351710 102.059 -70.848 46.383
351711 102.089 -70.850 46.342
351712 102.119 -70.851 46.302
352128 92.190 -69.754 55.024
352129 92.229 -69.724 55.003
352130 92.269 -69.695 54.982
352131 92.308 -69.664 54.961
352132 92.348 -69.635 54.940
352133 92.387 -69.605 54.918
352134 92.426 -69.576 54.896
352135 92.465 -69.546 54.873
352136 92.505 -69.516 54.850
352137 92.552 -69.507 54.826
352138 92.599 -69.499 54.802
352139 92.647 -69.491 54.777
352140 92.694 -69.484 54.753
352141 92.742 -69.476 54.728
352142 92.789 -69.469 54.701
352143 92.837 -69.462 54.672
352144 92.884 -69.454 54.645
352145 92.932 -69.447 54.617
352146 92.979 -69.438 54.588
352147 93.021 -69.430 54.557
352148 93.056 -69.431 54.524
352149 93.090 -69.432 54.490
352150 93.126 -69.433 54.455
352151 93.162 -69.436 54.421
352152 93.198 -69.440 54.388
352153 93.234 -69.444 54.356
352154 93.270 -69.447 54.324
352155 93.306 -69.451 54.292
352156 93.342 -69.455 54.261
352157 93.378 -69.459 54.231
352158 93.415 -69.463 54.202
352159 93.451 -69.466 54.173
352193 94.591 -69.478 53.199
352194 94.622 -69.476 53.180
352195 94.654 -69.473 53.162
352196 94.685 -69.471 53.143
352197 94.717 -69.469 53.124
352198 94.748 -69.466 53.105
352199 94.780 -69.463 53.087
352200 94.811 -69.461 53.072
352201 94.842 -69.459 53.057
352202 94.874 -69.457 53.043
352203 94.905 -69.454 53.029
352204 94.937 -69.452 53.017
352205 94.968 -69.450 53.007
352206 95.000 -69.447 52.999
352207 95.089 -69.460 52.993
352208 95.181 -69.472 52.990
352209 95.273 -69.486 52.987
352210 95.365 -69.500 52.983
352211 95.431 -69.505 52.981
352212 95.460 -69.500 52.980
352213 95.487 -69.494 52.983
352214 95.515 -69.489 52.987
352215 95.543 -69.484 52.994
352216 95.570 -69.479 53.002
352217 95.599 -69.473 53.007
352218 95.626 -69.467 53.012
352219 95.654 -69.462 53.018
352220 95.682 -69.457 53.026
352221 95.710 -69.451 53.033
352222 95.738 -69.446 53.042
352223 95.765 -69.441 53.053
352224 95.793 -69.435 53.065
352225 95.820 -69.429 53.079
352226 95.848 -69.424 53.088
352227 95.875 -69.418 53.095
352228 95.902 -69.413 53.099
352229 95.930 -69.407 53.099
352230 95.957 -69.401 53.089
352231 95.984 -69.395 53.069
352232 96.012 -69.390 53.042
352233 96.039 -69.384 53.009
352234 96.066 -69.378 52.972
352235 96.093 -69.372 52.941
352236 96.120 -69.366 52.913
352237 96.148 -69.361 52.887
352238 96.169 -69.380 52.859
352239 96.189 -69.398 52.831
352240 96.213 -69.413 52.807
352241 96.236 -69.428 52.788
352242 96.261 -69.441 52.773
352243 96.287 -69.453 52.763
352244 96.314 -69.464 52.752
352245 96.340 -69.476 52.740
352246 96.366 -69.487 52.726
352247 96.393 -69.498 52.707
352248 96.419 -69.509 52.685
352249 96.445 -69.520 52.666
352250 96.472 -69.532 52.650
352251 96.507 -69.544 52.633
352252 96.537 -69.551 52.612
352253 96.568 -69.559 52.588
352254 96.600 -69.565 52.563
352255 96.632 -69.571 52.539
352256 96.666 -69.580 52.515
352257 96.699 -69.591 52.493
352258 96.733 -69.603 52.472
352259 96.767 -69.614 52.451
352260 96.799 -69.620 52.432
352261 96.830 -69.625 52.413
352262 96.861 -69.629 52.393
352263 96.894 -69.644 52.373
352264 96.928 -69.659 52.355
352265 96.960 -69.666 52.339
352266 96.992 -69.673 52.323
352267 97.023 -69.679 52.308
352268 97.055 -69.686 52.293
352269 97.087 -69.692 52.275
352270 97.118 -69.699 52.257
352271 97.150 -69.705 52.235
352272 97.182 -69.712 52.210
352273 97.217 -69.722 52.182
352274 97.252 -69.731 52.153
352275 97.287 -69.741 52.124
352276 97.321 -69.751 52.097
352277 97.356 -69.760 52.074
352278 97.388 -69.767 52.048
352279 97.420 -69.775 52.021
352280 97.452 -69.782 51.995
352281 97.483 -69.789 51.969
352282 97.514 -69.796 51.950
352283 97.544 -69.806 51.930
352284 97.574 -69.816 51.906
352285 97.604 -69.826 51.875
352286 97.634 -69.836 51.839
352287 97.667 -69.842 51.802
352288 97.704 -69.845 51.764
352289 97.740 -69.848 51.726
352290 97.776 -69.852 51.689
352291 97.812 -69.855 51.648
352292 97.849 -69.858 51.607
352293 97.884 -69.861 51.568
352294 97.917 -69.863 51.530
352295 97.949 -69.865 51.491
352296 97.983 -69.865 51.450
352297 98.017 -69.864 51.406
352298 98.051 -69.863 51.357
352299 98.083 -69.866 51.307
352300 98.113 -69.874 51.253
352301 98.143 -69.881 51.198
352302 98.173 -69.889 51.142
352303 98.203 -69.896 51.086
352304 98.233 -69.904 51.027
352305 98.263 -69.911 50.971
352306 98.293 -69.919 50.915
352307 98.322 -69.927 50.860
352308 98.353 -69.934 50.807
352309 98.382 -69.942 50.750
352310 98.412 -69.949 50.692
352311 98.442 -69.957 50.632
352312 98.472 -69.964 50.570
352313 98.502 -69.972 50.510
352314 98.532 -69.980 50.451
352315 98.562 -69.988 50.392
352316 98.592 -69.996 50.335
352317 98.622 -70.003 50.273
352318 98.652 -70.001 50.212
352319 98.683 -69.996 50.150
352320 98.714 -69.990 50.086
352321 98.756 -70.002 50.024
352322 98.801 -70.015 49.963
352323 98.846 -70.029 49.903
352324 98.891 -70.043 49.843
352325 98.936 -70.056 49.786
352326 98.981 -70.070 49.730
352327 99.026 -70.083 49.674
352328 99.071 -70.097 49.618
352329 99.116 -70.111 49.565
352330 99.161 -70.124 49.513
352331 99.196 -70.135 49.463
352332 99.229 -70.145 49.413
352333 99.262 -70.156 49.365
352334 99.294 -70.165 49.316
352335 99.332 -70.172 49.269
352336 99.370 -70.179 49.224
352337 99.411 -70.184 49.182
352338 99.453 -70.187 49.140
352339 99.495 -70.192 49.099
352340 99.538 -70.196 49.059
352341 99.585 -70.199 49.020
352342 99.635 -70.202 48.981
352343 99.686 -70.205 48.946
352344 99.735 -70.208 48.911
352345 99.774 -70.233 48.877
352346 99.806 -70.272 48.844
352347 99.838 -70.310 48.814
352348 99.869 -70.344 48.785
352349 99.901 -70.356 48.758
352350 99.931 -70.369 48.734
352351 99.963 -70.382 48.711
352352 99.994 -70.388 48.687
352353 100.025 -70.388 48.665
352354 100.057 -70.388 48.643
352355 100.089 -70.388 48.620
352356 100.121 -70.388 48.593
352357 100.152 -70.388 48.567
352358 100.184 -70.388 48.539
352359 100.216 -70.388 48.511
352360 100.248 -70.388 48.486
352361 100.279 -70.388 48.462
352362 100.311 -70.386 48.437
352363 100.355 -70.395 48.411
352364 100.404 -70.408 48.381
352365 100.453 -70.421 48.348
352366 100.502 -70.434 48.312
352367 100.551 -70.447 48.276
352368 100.600 -70.460 48.240
352369 100.649 -70.474 48.207
352370 100.698 -70.487 48.174
352371 100.748 -70.500 48.139
352372 100.797 -70.513 48.103
352373 100.845 -70.526 48.068
352374 100.867 -70.549 48.030
352375 100.889 -70.569 47.992
352376 100.910 -70.588 47.953
352377 100.931 -70.607 47.912
352378 100.954 -70.624 47.872
352379 100.986 -70.628 47.831
352380 101.019 -70.631 47.789
352381 101.048 -70.635 47.748
352382 101.075 -70.639 47.705
352383 101.102 -70.643 47.661
352384 101.129 -70.647 47.617
352385 101.156 -70.651 47.573
352386 101.183 -70.655 47.533
352387 101.211 -70.666 47.496
352388 101.238 -70.682 47.460
352389 101.266 -70.698 47.423
352390 101.293 -70.713 47.384
352391 101.325 -70.728 47.339
352392 101.360 -70.743 47.294
352393 101.394 -70.759 47.248
352394 101.435 -70.770 47.200
352395 101.486 -70.775 47.151
352396 101.536 -70.781 47.101
352397 101.586 -70.786 47.052
352398 101.636 -70.792 47.004
352399 101.686 -70.797 46.957
352400 101.732 -70.796 46.906
352401 101.778 -70.795 46.853
352402 101.809 -70.792 46.798
352403 101.835 -70.789 46.742
352404 101.863 -70.787 46.691
352405 101.893 -70.789 46.642
352406 101.922 -70.790 46.596
352407 101.953 -70.792 46.552
352408 101.983 -70.793 46.508
352409 102.013 -70.795 46.466
352410 102.043 -70.796 46.424
352411 102.073 -70.798 46.382
352412 102.103 -70.799 46.341
352830 92.220 -69.694 54.995
352831 92.260 -69.664 54.973
352832 92.299 -69.635 54.952
352833 92.338 -69.605 54.930
352834 92.378 -69.575 54.907
352835 92.417 -69.545 54.885
352836 92.456 -69.516 54.861
352837 92.500 -69.497 54.837
352838 92.548 -69.490 54.813
352839 92.595 -69.482 54.788
352840 92.643 -69.475 54.762
352841 92.690 -69.467 54.737
352842 92.738 -69.460 54.711
352843 92.785 -69.453 54.684
352844 92.833 -69.445 54.656
352845 92.880 -69.438 54.629
352846 92.928 -69.429 54.600
352847 92.972 -69.418 54.571
352848 93.015 -69.408 54.540
352849 93.050 -69.409 54.507
352850 93.085 -69.410 54.474
352851 93.120 -69.410 54.440
352852 93.155 -69.411 54.407
352853 93.190 -69.412 54.375
352854 93.225 -69.413 54.343
352855 93.260 -69.413 54.311
352856 93.296 -69.417 54.279
352857 93.332 -69.420 54.249
352858 93.368 -69.424 54.220
352859 93.404 -69.428 54.191
352860 93.440 -69.432 54.162
352893 94.559 -69.459 53.234
352894 94.591 -69.456 53.214
352895 94.622 -69.454 53.196
352896 94.653 -69.451 53.178
352897 94.685 -69.449 53.160
352898 94.716 -69.447 53.142
352899 94.748 -69.444 53.124
352900 94.779 -69.442 53.108
352901 94.811 -69.440 53.093
352902 94.842 -69.437 53.079
352903 94.874 -69.435 53.066
352904 94.905 -69.432 53.053
352905 94.936 -69.430 53.042
352906 94.968 -69.428 53.033
352907 95.038 -69.435 53.026
352908 95.130 -69.449 53.021
352909 95.222 -69.462 53.019
352910 95.314 -69.476 53.016
352911 95.399 -69.482 53.013
352912 95.441 -69.476 53.010
352913 95.466 -69.474 53.008
352914 95.494 -69.469 53.009
352915 95.522 -69.463 53.013
352916 95.549 -69.458 53.018
352917 95.577 -69.453 53.026
352918 95.604 -69.447 53.034
352919 95.632 -69.442 53.040
352920 95.659 -69.436 53.045
352921 95.686 -69.431 53.051
352922 95.713 -69.425 53.058
352923 95.741 -69.419 53.066
352924 95.768 -69.413 53.075
352925 95.796 -69.408 53.086
352926 95.823 -69.402 53.095
352927 95.850 -69.397 53.104
352928 95.877 -69.391 53.110
352929 95.904 -69.385 53.112
352930 95.932 -69.379 53.107
352931 95.959 -69.374 53.094
352932 95.986 -69.368 53.072
352933 96.014 -69.363 53.044
352934 96.041 -69.356 53.011
352935 96.068 -69.351 52.977
352936 96.096 -69.345 52.948
352937 96.123 -69.340 52.920
352938 96.150 -69.334 52.894
352939 96.172 -69.348 52.869
352940 96.194 -69.365 52.844
352941 96.218 -69.380 52.823
352942 96.241 -69.394 52.808
352943 96.265 -69.409 52.799
352944 96.292 -69.420 52.789
352945 96.318 -69.432 52.778
352946 96.344 -69.442 52.766
352947 96.371 -69.454 52.751
352948 96.397 -69.465 52.732
352949 96.424 -69.476 52.709
352950 96.451 -69.488 52.688
352951 96.485 -69.500 52.671
352952 96.519 -69.511 52.653
352953 96.549 -69.519 52.631
352954 96.580 -69.526 52.607
352955 96.610 -69.534 52.582
352956 96.643 -69.540 52.558
352957 96.676 -69.549 52.535
352958 96.710 -69.560 52.512
352959 96.742 -69.568 52.488
352960 96.773 -69.573 52.465
352961 96.804 -69.578 52.441
352962 96.835 -69.583 52.419
352963 96.867 -69.589 52.395
352964 96.901 -69.605 52.372
352965 96.935 -69.622 52.353
352966 96.967 -69.632 52.335
352967 96.999 -69.639 52.316
352968 97.031 -69.645 52.298
352969 97.063 -69.652 52.281
352970 97.094 -69.659 52.261
352971 97.126 -69.665 52.242
352972 97.158 -69.671 52.221
352973 97.189 -69.678 52.195
352974 97.223 -69.686 52.168
352975 97.258 -69.696 52.139
352976 97.293 -69.705 52.109
352977 97.327 -69.715 52.082
352978 97.362 -69.724 52.058
352979 97.393 -69.731 52.032
352980 97.425 -69.738 52.006
352981 97.456 -69.745 51.981
352982 97.488 -69.751 51.957
352983 97.519 -69.758 51.940
352984 97.550 -69.767 51.921
352985 97.580 -69.776 51.898
352986 97.610 -69.786 51.869
352987 97.640 -69.796 51.834
352988 97.675 -69.801 51.798
352989 97.711 -69.805 51.761
352990 97.747 -69.808 51.724
352991 97.784 -69.811 51.688
352992 97.820 -69.814 51.647
352993 97.856 -69.817 51.609
352994 97.892 -69.820 51.572
352995 97.929 -69.823 51.535
352996 97.961 -69.825 51.496
352997 97.994 -69.826 51.455
352998 98.028 -69.824 51.410
352999 98.061 -69.828 51.361
353000 98.090 -69.836 51.310
353001 98.120 -69.843 51.256
353002 98.150 -69.851 51.200
353003 98.180 -69.858 51.143
353004 98.210 -69.866 51.087
353005 98.240 -69.874 51.028
353006 98.270 -69.881 50.973
353007 98.300 -69.889 50.920
353008 98.329 -69.896 50.864
353009 98.360 -69.904 50.808
353010 98.389 -69.911 50.750
353011 98.420 -69.919 50.691
353012 98.449 -69.927 50.631
353013 98.479 -69.935 50.570
353014 98.509 -69.942 50.511
353015 98.539 -69.949 50.452
353016 98.570 -69.956 50.393
353017 98.601 -69.964 50.335
353018 98.632 -69.971 50.274
353019 98.667 -69.973 50.212
353020 98.702 -69.966 50.150
353021 98.736 -69.960 50.087
353022 98.784 -69.975 50.024
353023 98.832 -69.990 49.963
353024 98.878 -70.005 49.902
353025 98.923 -70.018 49.842
353026 98.968 -70.032 49.785
353027 99.013 -70.046 49.728
353028 99.058 -70.059 49.672
353029 99.103 -70.073 49.617
353030 99.136 -70.083 49.564
353031 99.168 -70.093 49.512
353032 99.201 -70.103 49.461
353033 99.234 -70.113 49.412
353034 99.267 -70.123 49.362
353035 99.303 -70.131 49.313
353036 99.340 -70.138 49.266
353037 99.378 -70.145 49.222
353038 99.417 -70.150 49.179
353039 99.460 -70.155 49.136
353040 99.502 -70.159 49.095
353041 99.544 -70.162 49.056
353042 99.593 -70.165 49.016
353043 99.644 -70.168 48.977
353044 99.694 -70.171 48.941
353045 99.744 -70.174 48.905
353046 99.788 -70.189 48.870
353047 99.820 -70.228 48.836
353048 99.851 -70.267 48.805
353049 99.884 -70.306 48.773
353050 99.915 -70.331 48.745
353051 99.946 -70.344 48.718
353052 99.977 -70.357 48.693
353053 100.008 -70.363 48.670
353054 100.040 -70.363 48.648
353055 100.072 -70.363 48.627
353056 100.104 -70.363 48.607
353057 100.135 -70.363 48.583
353058 100.167 -70.363 48.557
353059 100.199 -70.363 48.531
353060 100.230 -70.363 48.504
353061 100.262 -70.362 48.477
353062 100.292 -70.356 48.452
353063 100.328 -70.356 48.426
353064 100.371 -70.366 48.401
353065 100.419 -70.378 48.376
353066 100.469 -70.391 48.346
353067 100.517 -70.405 48.312
353068 100.567 -70.418 48.276
353069 100.616 -70.431 48.239
353070 100.665 -70.444 48.206
353071 100.714 -70.457 48.172
353072 100.763 -70.470 48.136
353073 100.812 -70.484 48.099
353074 100.856 -70.497 48.063
353075 100.870 -70.515 48.023
353076 100.900 -70.542 47.984
353077 100.924 -70.564 47.944
353078 100.945 -70.584 47.902
353079 100.973 -70.594 47.861
353080 101.004 -70.597 47.821
353081 101.031 -70.602 47.780
353082 101.058 -70.606 47.738
353083 101.086 -70.610 47.695
353084 101.113 -70.614 47.652
353085 101.140 -70.618 47.608
353086 101.167 -70.622 47.565
353087 101.194 -70.626 47.526
353088 101.221 -70.636 47.491
353089 101.249 -70.651 47.455
353090 101.276 -70.667 47.417
353091 101.305 -70.682 47.375
353092 101.339 -70.698 47.329
353093 101.374 -70.713 47.283
353094 101.408 -70.729 47.237
353095 101.443 -70.744 47.190
353096 101.488 -70.752 47.141
353097 101.538 -70.757 47.090
353098 101.588 -70.763 47.039
353099 101.638 -70.768 46.989
353100 101.684 -70.767 46.940
353101 101.730 -70.765 46.888
353102 101.775 -70.764 46.836
353103 101.821 -70.763 46.783
353104 101.866 -70.761 46.730
353105 101.892 -70.758 46.682
353106 101.919 -70.756 46.635
353107 101.946 -70.753 46.590
353108 101.973 -70.751 46.547
353109 102.000 -70.748 46.504
353110 102.027 -70.745 46.462
353111 102.056 -70.746 46.420
353112 102.087 -70.748 46.379
353531 92.211 -69.664 54.987
353532 92.250 -69.634 54.965
353533 92.290 -69.604 54.942
353534 92.329 -69.575 54.920
353535 92.369 -69.545 54.897
353536 92.408 -69.515 54.875
353537 92.448 -69.488 54.850
353538 92.496 -69.481 54.825
353539 92.544 -69.473 54.799
353540 92.591 -69.466 54.774
353541 92.639 -69.459 54.748
353542 92.686 -69.451 54.722
353543 92.734 -69.444 54.696
353544 92.781 -69.436 54.668
353545 92.829 -69.429 54.640
353546 92.876 -69.421 54.612
353547 92.921 -69.409 54.583
353548 92.966 -69.397 54.554
353549 93.010 -69.385 54.523
353550 93.045 -69.386 54.491
353551 93.079 -69.388 54.459
353552 93.115 -69.388 54.425
353553 93.149 -69.389 54.392
353554 93.185 -69.390 54.361
353555 93.219 -69.391 54.329
353556 93.254 -69.391 54.299
353557 93.289 -69.392 54.268
353558 93.324 -69.393 54.237
353559 93.359 -69.394 54.208
353560 93.394 -69.394 54.180
353561 93.430 -69.397 54.151
353594 94.558 -69.437 53.249
353595 94.590 -69.435 53.229
353596 94.621 -69.432 53.213
353597 94.652 -69.429 53.195
353598 94.684 -69.427 53.178
353599 94.715 -69.425 53.161
353600 94.747 -69.422 53.144
353601 94.778 -69.420 53.129
353602 94.810 -69.418 53.114
353603 94.841 -69.416 53.102
353604 94.873 -69.413 53.090
353605 94.904 -69.410 53.078
353606 94.935 -69.408 53.068
353607 94.987 -69.411 53.060
353608 95.079 -69.425 53.054
353609 95.171 -69.438 53.051
353610 95.263 -69.451 53.050
353611 95.345 -69.454 53.048
353612 95.407 -69.439 53.045
353613 95.443 -69.434 53.042
353614 95.469 -69.433 53.039
353615 95.494 -69.433 53.039
353616 95.521 -69.432 53.040
353617 95.548 -69.427 53.044
353618 95.576 -69.422 53.050
353619 95.603 -69.416 53.059
353620 95.631 -69.411 53.066
353621 95.659 -69.406 53.072
353622 95.686 -69.400 53.078
353623 95.714 -69.395 53.084
353624 95.742 -69.391 53.092
353625 95.769 -69.385 53.101
353626 95.797 -69.380 53.109
353627 95.825 -69.375 53.116
353628 95.852 -69.369 53.123
353629 95.880 -69.363 53.127
353630 95.907 -69.358 53.127
353631 95.934 -69.352 53.119
353632 95.961 -69.347 53.102
353633 95.988 -69.341 53.078
353634 96.016 -69.335 53.048
353635 96.043 -69.329 53.016
353636 96.071 -69.324 52.984
353637 96.098 -69.318 52.955
353638 96.125 -69.313 52.929
353639 96.152 -69.306 52.904
353640 96.176 -69.316 52.882
353641 96.199 -69.331 52.861
353642 96.222 -69.347 52.844
353643 96.246 -69.361 52.832
353644 96.269 -69.375 52.827
353645 96.296 -69.387 52.817
353646 96.322 -69.398 52.804
353647 96.348 -69.410 52.790
353648 96.375 -69.421 52.777
353649 96.401 -69.432 52.757
353650 96.430 -69.444 52.734
353651 96.464 -69.456 52.712
353652 96.499 -69.468 52.692
353653 96.531 -69.478 52.672
353654 96.561 -69.485 52.648
353655 96.592 -69.493 52.625
353656 96.622 -69.501 52.601
353657 96.653 -69.508 52.578
353658 96.685 -69.516 52.553
353659 96.717 -69.522 52.527
353660 96.748 -69.526 52.500
353661 96.779 -69.531 52.474
353662 96.810 -69.536 52.448
353663 96.841 -69.541 52.421
353664 96.873 -69.551 52.395
353665 96.907 -69.567 52.369
353666 96.941 -69.583 52.347
353667 96.974 -69.598 52.326
353668 97.007 -69.605 52.305
353669 97.038 -69.612 52.285
353670 97.070 -69.619 52.264
353671 97.102 -69.625 52.245
353672 97.134 -69.631 52.224
353673 97.165 -69.638 52.201
353674 97.196 -69.644 52.175
353675 97.229 -69.651 52.148
353676 97.264 -69.660 52.119
353677 97.299 -69.670 52.092
353678 97.333 -69.679 52.066
353679 97.366 -69.688 52.043
353680 97.397 -69.695 52.017
353681 97.429 -69.701 51.990
353682 97.461 -69.708 51.965
353683 97.492 -69.715 51.943
353684 97.524 -69.722 51.927
353685 97.555 -69.729 51.910
353686 97.586 -69.737 51.889
353687 97.616 -69.747 51.862
353688 97.646 -69.757 51.829
353689 97.682 -69.761 51.795
353690 97.719 -69.764 51.759
353691 97.755 -69.767 51.723
353692 97.791 -69.770 51.688
353693 97.827 -69.773 51.648
353694 97.864 -69.776 51.611
353695 97.900 -69.780 51.575
353696 97.936 -69.783 51.538
353697 97.972 -69.786 51.499
353698 98.006 -69.786 51.457
353699 98.037 -69.790 51.411
353700 98.068 -69.798 51.361
353701 98.097 -69.805 51.307
353702 98.127 -69.813 51.252
353703 98.157 -69.820 51.196
353704 98.187 -69.828 51.141
353705 98.217 -69.836 51.086
353706 98.248 -69.842 51.031
353707 98.279 -69.849 50.977
353708 98.309 -69.857 50.922
353709 98.340 -69.864 50.866
353710 98.371 -69.870 50.807
353711 98.402 -69.877 50.748
353712 98.433 -69.884 50.689
353713 98.463 -69.891 50.630
353714 98.494 -69.899 50.570
353715 98.525 -69.905 50.511
353716 98.556 -69.912 50.452
353717 98.586 -69.919 50.393
353718 98.617 -69.926 50.334
353719 98.650 -69.935 50.272
353720 98.691 -69.942 50.210
353721 98.725 -69.936 50.146
353722 98.761 -69.933 50.083
353723 98.809 -69.947 50.020
353724 98.856 -69.962 49.959
353725 98.904 -69.977 49.899
353726 98.952 -69.992 49.841
353727 99.000 -70.007 49.784
353728 99.043 -70.021 49.727
353729 99.076 -70.030 49.671
353730 99.109 -70.040 49.616
353731 99.141 -70.051 49.563
353732 99.174 -70.061 49.511
353733 99.207 -70.071 49.460
353734 99.239 -70.081 49.410
353735 99.273 -70.090 49.360
353736 99.311 -70.097 49.312
353737 99.349 -70.104 49.265
353738 99.386 -70.111 49.220
353739 99.424 -70.118 49.176
353740 99.466 -70.121 49.133
353741 99.508 -70.125 49.091
353742 99.552 -70.129 49.051
353743 99.602 -70.132 49.011
353744 99.652 -70.135 48.972
353745 99.702 -70.138 48.936
353746 99.752 -70.141 48.898
353747 99.801 -70.146 48.863
353748 99.833 -70.184 48.828
353749 99.865 -70.223 48.795
353750 99.897 -70.262 48.764
353751 99.929 -70.300 48.734
353752 99.960 -70.319 48.705
353753 99.991 -70.332 48.678
353754 100.022 -70.339 48.653
353755 100.054 -70.339 48.629
353756 100.086 -70.339 48.607
353757 100.118 -70.339 48.586
353758 100.149 -70.339 48.563
353759 100.181 -70.339 48.539
353760 100.213 -70.339 48.514
353761 100.244 -70.333 48.488
353762 100.274 -70.327 48.460
353763 100.304 -70.320 48.434
353764 100.346 -70.328 48.410
353765 100.390 -70.338 48.387
353766 100.435 -70.349 48.365
353767 100.484 -70.362 48.337
353768 100.533 -70.375 48.305
353769 100.582 -70.388 48.268
353770 100.631 -70.401 48.230
353771 100.680 -70.415 48.196
353772 100.729 -70.428 48.162
353773 100.778 -70.441 48.126
353774 100.828 -70.454 48.089
353775 100.866 -70.468 48.053
353776 100.878 -70.484 48.012
353777 100.903 -70.508 47.972
353778 100.932 -70.535 47.932
353779 100.960 -70.560 47.892
353780 100.988 -70.564 47.849
353781 101.015 -70.569 47.809
353782 101.042 -70.572 47.768
353783 101.069 -70.576 47.726
353784 101.096 -70.581 47.683
353785 101.123 -70.585 47.641
353786 101.150 -70.589 47.597
353787 101.178 -70.593 47.554
353788 101.204 -70.597 47.517
353789 101.232 -70.606 47.483
353790 101.259 -70.621 47.447
353791 101.286 -70.637 47.407
353792 101.318 -70.652 47.364
353793 101.353 -70.667 47.317
353794 101.388 -70.682 47.271
353795 101.422 -70.698 47.225
353796 101.457 -70.713 47.178
353797 101.491 -70.729 47.130
353798 101.540 -70.735 47.078
353799 101.590 -70.739 47.025
353800 101.636 -70.738 46.972
353801 101.682 -70.736 46.920
353802 101.728 -70.735 46.868
353803 101.773 -70.733 46.817
353804 101.819 -70.732 46.768
353805 101.865 -70.730 46.718
353806 101.911 -70.729 46.672
353807 101.949 -70.727 46.628
353808 101.976 -70.724 46.584
353809 102.003 -70.722 46.541
353810 102.030 -70.719 46.498
353811 102.057 -70.717 46.456
353812 102.084 -70.713 46.415
353813 102.111 -70.711 46.375
354233 92.241 -69.604 54.956
354234 92.281 -69.574 54.934
354235 92.320 -69.544 54.911
354236 92.359 -69.515 54.888
354237 92.399 -69.485 54.864
354238 92.444 -69.472 54.839
354239 92.492 -69.465 54.813
354240 92.540 -69.457 54.787
354241 92.587 -69.450 54.762
354242 92.635 -69.442 54.735
354243 92.682 -69.435 54.708
354244 92.730 -69.427 54.681
354245 92.777 -69.419 54.653
354246 92.825 -69.412 54.624
354247 92.869 -69.400 54.596
354248 92.914 -69.388 54.567
354249 92.958 -69.376 54.538
354250 93.003 -69.365 54.506
354251 93.039 -69.364 54.475
354252 93.074 -69.365 54.443
354253 93.109 -69.366 54.410
354254 93.144 -69.366 54.378
354255 93.179 -69.367 54.346
354256 93.214 -69.368 54.315
354257 93.249 -69.369 54.284
354258 93.284 -69.370 54.254
354259 93.319 -69.371 54.224
354260 93.354 -69.372 54.195
354261 93.389 -69.372 54.167
354262 93.424 -69.373 54.140
354294 94.526 -69.417 53.282
354295 94.557 -69.415 53.263
354296 94.589 -69.413 53.245
354297 94.620 -69.410 53.229
354298 94.652 -69.408 53.213
354299 94.683 -69.406 53.196
354300 94.715 -69.403 53.180
354301 94.746 -69.400 53.165
354302 94.777 -69.398 53.150
354303 94.809 -69.396 53.137
354304 94.840 -69.394 53.125
354305 94.872 -69.391 53.115
354306 94.903 -69.389 53.104
354307 94.936 -69.387 53.096
354308 95.029 -69.400 53.090
354309 95.120 -69.413 53.085
354310 95.212 -69.427 53.084
354311 95.291 -69.427 53.084
354312 95.352 -69.411 53.083
354313 95.414 -69.396 53.082
354314 95.446 -69.392 53.079
354315 95.472 -69.392 53.075
354316 95.497 -69.391 53.072
354317 95.523 -69.391 53.070
354318 95.549 -69.391 53.070
354319 95.574 -69.391 53.071
354320 95.602 -69.385 53.078
354321 95.630 -69.380 53.088
354322 95.658 -69.375 53.100
354323 95.685 -69.369 53.108
354324 95.712 -69.364 53.112
354325 95.740 -69.359 53.120
354326 95.768 -69.353 53.129
354327 95.796 -69.348 53.136
354328 95.823 -69.343 53.142
354329 95.851 -69.338 53.146
354330 95.879 -69.333 53.147
354331 95.906 -69.328 53.143
354332 95.934 -69.322 53.133
354333 95.962 -69.317 53.110
354334 95.989 -69.312 53.083
354335 96.017 -69.306 53.054
354336 96.044 -69.301 53.023
354337 96.072 -69.296 52.994
354338 96.100 -69.291 52.965
354339 96.127 -69.285 52.939
354340 96.155 -69.280 52.917
354341 96.180 -69.284 52.898
354342 96.203 -69.298 52.881
354343 96.227 -69.313 52.867
354344 96.250 -69.328 52.857
354345 96.274 -69.342 52.854
354346 96.300 -69.354 52.845
354347 96.326 -69.365 52.831
354348 96.353 -69.376 52.816
354349 96.379 -69.388 52.802
354350 96.408 -69.399 52.783
354351 96.443 -69.412 52.759
354352 96.477 -69.424 52.735
354353 96.511 -69.436 52.713
354354 96.543 -69.445 52.691
354355 96.573 -69.453 52.667
354356 96.604 -69.460 52.644
354357 96.634 -69.468 52.620
354358 96.663 -69.476 52.596
354359 96.693 -69.482 52.568
354360 96.723 -69.485 52.540
354361 96.754 -69.488 52.511
354362 96.785 -69.490 52.481
354363 96.816 -69.494 52.452
354364 96.847 -69.499 52.422
354365 96.880 -69.513 52.392
354366 96.913 -69.529 52.364
354367 96.948 -69.544 52.340
354368 96.981 -69.560 52.316
354369 97.014 -69.571 52.293
354370 97.046 -69.578 52.271
354371 97.078 -69.585 52.249
354372 97.109 -69.591 52.228
354373 97.141 -69.598 52.207
354374 97.172 -69.604 52.183
354375 97.204 -69.610 52.156
354376 97.236 -69.617 52.130
354377 97.270 -69.625 52.102
354378 97.305 -69.635 52.075
354379 97.339 -69.644 52.051
354380 97.371 -69.651 52.029
354381 97.402 -69.658 52.002
354382 97.434 -69.665 51.974
354383 97.465 -69.672 51.949
354384 97.497 -69.679 51.927
354385 97.528 -69.686 51.912
354386 97.560 -69.692 51.897
354387 97.591 -69.699 51.878
354388 97.622 -69.707 51.854
354389 97.653 -69.716 51.823
354390 97.690 -69.719 51.792
354391 97.726 -69.723 51.757
354392 97.762 -69.726 51.722
354393 97.799 -69.729 51.687
354394 97.835 -69.732 51.648
354395 97.871 -69.735 51.611
354396 97.908 -69.738 51.574
354397 97.945 -69.742 51.536
354398 97.983 -69.745 51.496
354399 98.018 -69.750 51.454
354400 98.049 -69.757 51.407
354401 98.079 -69.764 51.355
354402 98.110 -69.770 51.301
354403 98.141 -69.777 51.246
354404 98.172 -69.785 51.191
354405 98.203 -69.792 51.137
354406 98.233 -69.798 51.086
354407 98.264 -69.805 51.032
354408 98.295 -69.812 50.979
354409 98.325 -69.819 50.924
354410 98.356 -69.827 50.866
354411 98.387 -69.833 50.806
354412 98.418 -69.840 50.748
354413 98.449 -69.847 50.689
354414 98.480 -69.854 50.629
354415 98.510 -69.861 50.569
354416 98.541 -69.868 50.510
354417 98.572 -69.875 50.451
354418 98.602 -69.882 50.391
354419 98.634 -69.889 50.331
354420 98.672 -69.902 50.266
354421 98.714 -69.911 50.203
354422 98.748 -69.905 50.140
354423 98.786 -69.905 50.077
354424 98.834 -69.919 50.014
354425 98.881 -69.934 49.955
354426 98.929 -69.949 49.897
354427 98.977 -69.964 49.840
354428 99.023 -69.979 49.784
354429 99.046 -69.991 49.727
354430 99.079 -70.000 49.670
354431 99.113 -70.010 49.615
354432 99.146 -70.019 49.561
354433 99.179 -70.029 49.509
354434 99.212 -70.039 49.458
354435 99.245 -70.049 49.408
354436 99.281 -70.056 49.358
354437 99.319 -70.063 49.310
354438 99.357 -70.070 49.263
354439 99.395 -70.077 49.217
354440 99.432 -70.083 49.173
354441 99.473 -70.088 49.130
354442 99.514 -70.093 49.088
354443 99.560 -70.096 49.047
354444 99.610 -70.099 49.006
354445 99.660 -70.102 48.967
354446 99.710 -70.105 48.929
354447 99.760 -70.108 48.892
354448 99.810 -70.111 48.856
354449 99.847 -70.140 48.821
354450 99.879 -70.179 48.787
354451 99.911 -70.218 48.755
354452 99.943 -70.256 48.724
354453 99.975 -70.294 48.694
354454 100.005 -70.307 48.665
354455 100.037 -70.315 48.638
354456 100.068 -70.315 48.612
354457 100.100 -70.315 48.588
354458 100.132 -70.315 48.564
354459 100.163 -70.315 48.541
354460 100.194 -70.309 48.518
354461 100.225 -70.303 48.493
354462 100.256 -70.297 48.468
354463 100.286 -70.290 48.441
354464 100.322 -70.290 48.415
354465 100.365 -70.300 48.392
354466 100.409 -70.310 48.371
354467 100.453 -70.320 48.352
354468 100.499 -70.332 48.326
354469 100.548 -70.346 48.294
354470 100.598 -70.359 48.258
354471 100.646 -70.372 48.218
354472 100.696 -70.385 48.183
354473 100.745 -70.398 48.148
354474 100.794 -70.411 48.112
354475 100.843 -70.425 48.075
354476 100.877 -70.438 48.039
354477 100.888 -70.454 48.000
354478 100.905 -70.474 47.959
354479 100.935 -70.501 47.919
354480 100.965 -70.526 47.881
354481 100.996 -70.533 47.838
354482 101.025 -70.539 47.796
354483 101.052 -70.544 47.754
354484 101.079 -70.547 47.712
354485 101.107 -70.551 47.669
354486 101.134 -70.556 47.628
354487 101.161 -70.560 47.585
354488 101.188 -70.564 47.543
354489 101.215 -70.568 47.504
354490 101.242 -70.576 47.470
354491 101.270 -70.591 47.434
354492 101.298 -70.607 47.394
354493 101.332 -70.622 47.349
354494 101.367 -70.637 47.301
354495 101.401 -70.652 47.258
354496 101.436 -70.667 47.213
354497 101.470 -70.682 47.166
354498 101.505 -70.698 47.118
354499 101.542 -70.710 47.067
354500 101.588 -70.708 47.013
354501 101.633 -70.707 46.958
354502 101.679 -70.706 46.904
354503 101.725 -70.704 46.853
354504 101.771 -70.703 46.804
354505 101.817 -70.701 46.757
354506 101.863 -70.700 46.709
354507 101.908 -70.698 46.664
354508 101.954 -70.697 46.620
354509 102.000 -70.695 46.577
354510 102.033 -70.693 46.535
354511 102.060 -70.691 46.492
354512 102.087 -70.688 46.451
354513 102.114 -70.685 46.411
354935 92.272 -69.544 54.926
354936 92.311 -69.515 54.902
354937 92.350 -69.485 54.878
354938 92.392 -69.463 54.854
354939 92.440 -69.456 54.828
354940 92.488 -69.448 54.802
354941 92.535 -69.441 54.775
354942 92.583 -69.433 54.749
354943 92.631 -69.425 54.722
354944 92.678 -69.418 54.694
354945 92.725 -69.411 54.667
354946 92.773 -69.403 54.638
354947 92.818 -69.391 54.610
354948 92.862 -69.379 54.581
354949 92.907 -69.368 54.551
354950 92.952 -69.356 54.521
354951 92.996 -69.344 54.490
354952 93.034 -69.342 54.458
354953 93.069 -69.343 54.426
354954 93.103 -69.344 54.395
354955 93.138 -69.344 54.362
354956 93.174 -69.345 54.331
354957 93.208 -69.346 54.300
354958 93.243 -69.347 54.269
354959 93.278 -69.347 54.239
354960 93.313 -69.348 54.210
354961 93.348 -69.349 54.181
354962 93.383 -69.350 54.154
354963 93.418 -69.350 54.128
354994 94.494 -69.398 53.315
354995 94.526 -69.396 53.297
354996 94.557 -69.393 53.279
354997 94.588 -69.391 53.261
354998 94.620 -69.388 53.246
354999 94.651 -69.386 53.230
355000 94.683 -69.384 53.215
355001 94.714 -69.382 53.200
355002 94.746 -69.379 53.186
355003 94.777 -69.377 53.172
355004 94.809 -69.374 53.160
355005 94.840 -69.372 53.150
355006 94.871 -69.369 53.140
355007 94.903 -69.367 53.131
355008 94.978 -69.376 53.124
355009 95.070 -69.389 53.119
355010 95.162 -69.403 53.116
355011 95.237 -69.400 53.116
355012 95.298 -69.384 53.117
355013 95.359 -69.368 53.118
355014 95.421 -69.353 53.118
355015 95.449 -69.351 53.117
355016 95.475 -69.350 53.112
355017 95.500 -69.350 53.110
355018 95.526 -69.350 53.109
355019 95.552 -69.350 53.108
355020 95.578 -69.349 53.108
355021 95.603 -69.349 53.109
355022 95.628 -69.348 53.112
355023 95.656 -69.343 53.118
355024 95.684 -69.338 53.127
355025 95.711 -69.333 53.135
355026 95.739 -69.328 53.146
355027 95.767 -69.322 53.156
355028 95.795 -69.317 53.163
355029 95.822 -69.312 53.169
355030 95.850 -69.306 53.171
355031 95.877 -69.301 53.168
355032 95.905 -69.296 53.158
355033 95.933 -69.291 53.142
355034 95.960 -69.285 53.115
355035 95.988 -69.280 53.086
355036 96.016 -69.275 53.059
355037 96.043 -69.270 53.033
355038 96.071 -69.265 53.004
355039 96.099 -69.259 52.979
355040 96.126 -69.254 52.956
355041 96.154 -69.249 52.935
355042 96.182 -69.251 52.920
355043 96.208 -69.265 52.903
355044 96.232 -69.279 52.891
355045 96.255 -69.294 52.883
355046 96.278 -69.309 52.879
355047 96.304 -69.321 52.871
355048 96.331 -69.332 52.857
355049 96.357 -69.344 52.842
355050 96.387 -69.355 52.826
355051 96.421 -69.367 52.807
355052 96.455 -69.380 52.784
355053 96.490 -69.392 52.759
355054 96.524 -69.404 52.735
355055 96.554 -69.412 52.712
355056 96.585 -69.419 52.688
355057 96.615 -69.427 52.664
355058 96.644 -69.435 52.639
355059 96.673 -69.443 52.611
355060 96.702 -69.450 52.581
355061 96.732 -69.455 52.550
355062 96.763 -69.457 52.519
355063 96.793 -69.460 52.486
355064 96.824 -69.463 52.454
355065 96.855 -69.466 52.421
355066 96.887 -69.475 52.388
355067 96.920 -69.490 52.358
355068 96.954 -69.506 52.332
355069 96.988 -69.523 52.306
355070 97.022 -69.538 52.281
355071 97.053 -69.544 52.258
355072 97.085 -69.551 52.235
355073 97.117 -69.557 52.213
355074 97.148 -69.563 52.191
355075 97.180 -69.570 52.166
355076 97.212 -69.576 52.139
355077 97.243 -69.582 52.114
355078 97.276 -69.590 52.086
355079 97.311 -69.600 52.060
355080 97.344 -69.607 52.037
355081 97.375 -69.614 52.015
355082 97.407 -69.621 51.986
355083 97.438 -69.628 51.958
355084 97.470 -69.635 51.932
355085 97.501 -69.642 51.911
355086 97.533 -69.648 51.897
355087 97.564 -69.656 51.883
355088 97.596 -69.663 51.867
355089 97.628 -69.670 51.845
355090 97.661 -69.675 51.817
355091 97.698 -69.678 51.787
355092 97.736 -69.682 51.755
355093 97.774 -69.686 51.720
355094 97.812 -69.689 51.685
355095 97.850 -69.692 51.646
355096 97.888 -69.695 51.608
355097 97.926 -69.699 51.571
355098 97.964 -69.703 51.532
355099 97.990 -69.708 51.490
355100 98.023 -69.715 51.447
355101 98.065 -69.720 51.399
355102 98.095 -69.726 51.347
355103 98.127 -69.733 51.292
355104 98.157 -69.740 51.238
355105 98.188 -69.747 51.185
355106 98.219 -69.754 51.134
355107 98.249 -69.761 51.085
355108 98.280 -69.768 51.032
355109 98.311 -69.775 50.978
355110 98.342 -69.782 50.923
355111 98.373 -69.789 50.864
355112 98.404 -69.796 50.805
355113 98.434 -69.803 50.747
355114 98.465 -69.810 50.687
355115 98.496 -69.817 50.626
355116 98.526 -69.823 50.567
355117 98.557 -69.830 50.507
355118 98.588 -69.838 50.447
355119 98.619 -69.845 50.386
355120 98.650 -69.852 50.324
355121 98.695 -69.869 50.258
355122 98.737 -69.881 50.194
355123 98.772 -69.875 50.132
355124 98.811 -69.877 50.070
355125 98.859 -69.891 50.008
355126 98.906 -69.906 49.951
355127 98.954 -69.921 49.895
355128 99.002 -69.936 49.840
355129 99.031 -69.949 49.783
355130 99.055 -69.961 49.725
355131 99.082 -69.971 49.669
355132 99.115 -69.981 49.613
355133 99.148 -69.990 49.559
355134 99.182 -70.000 49.507
355135 99.215 -70.009 49.456
355136 99.248 -70.019 49.406
355137 99.288 -70.024 49.356
355138 99.327 -70.029 49.308
355139 99.365 -70.036 49.261
355140 99.403 -70.043 49.215
355141 99.440 -70.049 49.170
355142 99.479 -70.055 49.127
355143 99.521 -70.059 49.084
355144 99.569 -70.062 49.042
355145 99.619 -70.065 49.002
355146 99.668 -70.068 48.961
355147 99.719 -70.071 48.922
355148 99.769 -70.074 48.886
355149 99.819 -70.077 48.849
355150 99.861 -70.096 48.814
355151 99.893 -70.135 48.780
355152 99.924 -70.174 48.747
355153 99.957 -70.212 48.715
355154 99.988 -70.251 48.683
355155 100.020 -70.283 48.653
355156 100.051 -70.290 48.625
355157 100.083 -70.291 48.597
355158 100.115 -70.291 48.569
355159 100.145 -70.286 48.542
355160 100.176 -70.280 48.518
355161 100.207 -70.274 48.493
355162 100.237 -70.267 48.468
355163 100.268 -70.261 48.444
355164 100.298 -70.255 48.419
355165 100.341 -70.263 48.396
355166 100.384 -70.273 48.375
355167 100.428 -70.283 48.357
355168 100.471 -70.293 48.338
355169 100.514 -70.303 48.311
355170 100.564 -70.316 48.281
355171 100.613 -70.329 48.246
355172 100.662 -70.342 48.207
355173 100.711 -70.356 48.169
355174 100.760 -70.369 48.133
355175 100.810 -70.381 48.097
355176 100.858 -70.395 48.061
355177 100.887 -70.409 48.025
355178 100.899 -70.425 47.986
355179 100.910 -70.441 47.947
355180 100.938 -70.466 47.908
355181 100.968 -70.490 47.868
355182 100.999 -70.497 47.828
355183 101.030 -70.505 47.786
355184 101.060 -70.512 47.743
355185 101.090 -70.519 47.699
355186 101.117 -70.522 47.658
355187 101.144 -70.526 47.615
355188 101.172 -70.531 47.572
355189 101.199 -70.535 47.530
355190 101.226 -70.539 47.489
355191 101.253 -70.545 47.452
355192 101.280 -70.561 47.415
355193 101.312 -70.576 47.374
355194 101.346 -70.591 47.329
355195 101.381 -70.607 47.284
355196 101.415 -70.622 47.242
355197 101.450 -70.637 47.197
355198 101.484 -70.654 47.150
355199 101.518 -70.667 47.103
355200 101.554 -70.671 47.054
355201 101.591 -70.675 47.002
355202 101.631 -70.676 46.947
355203 101.677 -70.675 46.892
355204 101.723 -70.673 46.841
355205 101.769 -70.672 46.793
355206 101.815 -70.671 46.746
355207 101.861 -70.670 46.701
355208 101.906 -70.668 46.656
355209 101.952 -70.666 46.613
355210 101.998 -70.665 46.570
355211 102.044 -70.663 46.528
355212 102.090 -70.662 46.486
355213 102.117 -70.660 46.446
355637 92.302 -69.485 54.894
355638 92.341 -69.455 54.869
355639 92.388 -69.447 54.844
355640 92.436 -69.439 54.818
355641 92.484 -69.432 54.791
355642 92.531 -69.424 54.764
355643 92.579 -69.417 54.737
355644 92.627 -69.410 54.709
355645 92.674 -69.402 54.681
355646 92.721 -69.394 54.654
355647 92.766 -69.382 54.624
355648 92.811 -69.371 54.595
355649 92.856 -69.359 54.565
355650 92.900 -69.347 54.535
355651 92.945 -69.335 54.505
355652 92.990 -69.323 54.474
355653 93.028 -69.319 54.442
355654 93.063 -69.320 54.410
355655 93.098 -69.321 54.378
355656 93.133 -69.322 54.346
355657 93.168 -69.322 54.315
355658 93.203 -69.324 54.284
355659 93.238 -69.325 54.253
355660 93.273 -69.325 54.224
355661 93.307 -69.326 54.195
355662 93.343 -69.327 54.166
355663 93.377 -69.328 54.140
355664 93.413 -69.328 54.115
355694 94.462 -69.378 53.346
355695 94.493 -69.376 53.328
355696 94.525 -69.374 53.311
355697 94.556 -69.372 53.294
355698 94.588 -69.369 53.277
355699 94.619 -69.366 53.263
355700 94.650 -69.364 53.248
355701 94.682 -69.362 53.234
355702 94.713 -69.360 53.220
355703 94.745 -69.357 53.207
355704 94.776 -69.355 53.194
355705 94.808 -69.353 53.184
355706 94.839 -69.350 53.175
355707 94.871 -69.347 53.166
355708 94.927 -69.352 53.159
355709 95.019 -69.365 53.153
355710 95.111 -69.378 53.149
355711 95.182 -69.372 53.146
355712 95.244 -69.356 53.148
355713 95.306 -69.341 53.150
355714 95.367 -69.325 53.152
355715 95.426 -69.310 53.154
355716 95.452 -69.309 53.153
355717 95.477 -69.309 53.149
355718 95.503 -69.309 53.147
355719 95.529 -69.309 53.147
355720 95.554 -69.308 53.148
355721 95.580 -69.308 53.149
355722 95.606 -69.307 53.146
355723 95.631 -69.306 53.144
355724 95.657 -69.306 53.144
355725 95.683 -69.306 53.148
355726 95.710 -69.301 53.156
355727 95.738 -69.296 53.168
355728 95.765 -69.291 53.180
355729 95.793 -69.285 53.189
355730 95.821 -69.280 53.195
355731 95.848 -69.275 53.194
355732 95.876 -69.270 53.187
355733 95.904 -69.265 53.174
355734 95.932 -69.259 53.154
355735 95.959 -69.254 53.128
355736 95.987 -69.249 53.102
355737 96.015 -69.244 53.077
355738 96.042 -69.238 53.052
355739 96.070 -69.233 53.025
355740 96.097 -69.228 53.000
355741 96.125 -69.222 52.976
355742 96.153 -69.218 52.955
355743 96.180 -69.221 52.944
355744 96.208 -69.234 52.930
355745 96.236 -69.246 52.921
355746 96.260 -69.261 52.915
355747 96.283 -69.275 52.909
355748 96.309 -69.288 52.897
355749 96.335 -69.299 52.883
355750 96.365 -69.311 52.867
355751 96.399 -69.323 52.850
355752 96.434 -69.335 52.830
355753 96.468 -69.348 52.808
355754 96.503 -69.360 52.783
355755 96.536 -69.371 52.759
355756 96.567 -69.378 52.736
355757 96.595 -69.387 52.710
355758 96.624 -69.394 52.683
355759 96.654 -69.402 52.655
355760 96.683 -69.410 52.624
355761 96.712 -69.417 52.591
355762 96.741 -69.425 52.558
355763 96.771 -69.428 52.525
355764 96.802 -69.430 52.489
355765 96.832 -69.433 52.455
355766 96.863 -69.436 52.420
355767 96.895 -69.445 52.387
355768 96.928 -69.456 52.356
355769 96.960 -69.468 52.328
355770 96.995 -69.484 52.301
355771 97.028 -69.500 52.275
355772 97.061 -69.510 52.251
355773 97.093 -69.517 52.227
355774 97.124 -69.523 52.204
355775 97.156 -69.530 52.181
355776 97.187 -69.536 52.156
355777 97.219 -69.542 52.129
355778 97.251 -69.549 52.102
355779 97.283 -69.555 52.075
355780 97.317 -69.564 52.049
355781 97.349 -69.571 52.023
355782 97.380 -69.578 51.999
355783 97.412 -69.585 51.969
355784 97.443 -69.591 51.941
355785 97.474 -69.598 51.916
355786 97.506 -69.606 51.895
355787 97.537 -69.613 51.881
355788 97.568 -69.620 51.869
355789 97.599 -69.628 51.853
355790 97.633 -69.633 51.832
355791 97.674 -69.636 51.807
355792 97.718 -69.639 51.779
355793 97.755 -69.643 51.748
355794 97.794 -69.647 51.714
355795 97.831 -69.650 51.680
355796 97.870 -69.654 51.640
355797 97.908 -69.657 51.601
355798 97.945 -69.660 51.562
355799 97.968 -69.667 51.523
355800 97.991 -69.673 51.481
355801 98.028 -69.680 51.436
355802 98.073 -69.685 51.387
355803 98.112 -69.689 51.334
355804 98.143 -69.696 51.281
355805 98.173 -69.703 51.230
355806 98.204 -69.710 51.179
355807 98.235 -69.717 51.130
355808 98.266 -69.724 51.081
355809 98.296 -69.731 51.028
355810 98.327 -69.738 50.974
355811 98.358 -69.745 50.917
355812 98.389 -69.751 50.859
355813 98.420 -69.758 50.800
355814 98.450 -69.766 50.741
355815 98.481 -69.773 50.681
355816 98.512 -69.780 50.619
355817 98.543 -69.786 50.559
355818 98.574 -69.793 50.498
355819 98.604 -69.800 50.436
355820 98.635 -69.808 50.374
355821 98.673 -69.819 50.310
355822 98.717 -69.836 50.246
355823 98.761 -69.851 50.183
355824 98.795 -69.845 50.121
355825 98.836 -69.848 50.062
355826 98.883 -69.864 50.005
355827 98.931 -69.878 49.949
355828 98.979 -69.893 49.894
355829 99.017 -69.907 49.838
355830 99.040 -69.919 49.780
355831 99.063 -69.930 49.722
355832 99.086 -69.942 49.666
355833 99.118 -69.952 49.611
355834 99.151 -69.961 49.557
355835 99.184 -69.971 49.505
355836 99.217 -69.980 49.453
355837 99.251 -69.990 49.403
355838 99.290 -69.995 49.354
355839 99.331 -69.999 49.305
355840 99.371 -70.003 49.257
355841 99.411 -70.008 49.211
355842 99.448 -70.015 49.166
355843 99.486 -70.022 49.121
355844 99.527 -70.026 49.079
355845 99.577 -70.029 49.037
355846 99.627 -70.032 48.996
355847 99.677 -70.035 48.955
355848 99.727 -70.038 48.917
355849 99.777 -70.041 48.879
355850 99.827 -70.044 48.842
355851 99.874 -70.052 48.806
355852 99.906 -70.091 48.772
355853 99.938 -70.130 48.737
355854 99.970 -70.168 48.704
355855 100.002 -70.207 48.672
355856 100.034 -70.246 48.641
355857 100.065 -70.267 48.612
355858 100.097 -70.262 48.582
355859 100.127 -70.256 48.553
355860 100.157 -70.250 48.526
355861 100.188 -70.244 48.500
355862 100.219 -70.237 48.475
355863 100.250 -70.231 48.450
355864 100.280 -70.225 48.426
355865 100.316 -70.225 48.401
355866 100.359 -70.235 48.379
355867 100.403 -70.246 48.358
355868 100.446 -70.256 48.339
355869 100.490 -70.265 48.321
355870 100.534 -70.275 48.294
355871 100.579 -70.287 48.265
355872 100.628 -70.300 48.231
355873 100.678 -70.312 48.194
355874 100.726 -70.326 48.156
355875 100.776 -70.339 48.119
355876 100.825 -70.352 48.083
355877 100.874 -70.366 48.047
355878 100.898 -70.380 48.011
355879 100.909 -70.396 47.972
355880 100.920 -70.411 47.934
355881 100.940 -70.432 47.895
355882 100.971 -70.454 47.855
355883 101.001 -70.461 47.816
355884 101.032 -70.469 47.774
355885 101.063 -70.476 47.731
355886 101.094 -70.483 47.688
355887 101.125 -70.491 47.649
355888 101.155 -70.497 47.607
355889 101.182 -70.501 47.564
355890 101.209 -70.506 47.521
355891 101.236 -70.510 47.478
355892 101.263 -70.515 47.437
355893 101.291 -70.531 47.396
355894 101.325 -70.546 47.353
355895 101.360 -70.561 47.308
355896 101.394 -70.576 47.266
355897 101.428 -70.594 47.223
355898 101.461 -70.613 47.179
355899 101.496 -70.623 47.133
355900 101.533 -70.627 47.086
355901 101.569 -70.631 47.039
355902 101.605 -70.635 46.989
355903 101.642 -70.638 46.936
355904 101.678 -70.643 46.881
355905 101.721 -70.643 46.831
355906 101.767 -70.641 46.783
355907 101.813 -70.640 46.737
355908 101.858 -70.638 46.692
355909 101.904 -70.637 46.649
355910 101.950 -70.635 46.605
355911 101.996 -70.635 46.563
355912 102.042 -70.633 46.522
355913 102.088 -70.632 46.481
355914 102.133 -70.630 46.441
356339 92.337 -69.438 54.859
356340 92.384 -69.430 54.835
356341 92.432 -69.422 54.809
356342 92.480 -69.416 54.781
356343 92.527 -69.408 54.753
356344 92.575 -69.400 54.726
356345 92.622 -69.393 54.698
356346 92.670 -69.385 54.669
356347 92.715 -69.374 54.641
356348 92.760 -69.362 54.611
356349 92.804 -69.350 54.581
356350 92.849 -69.338 54.551
356351 92.894 -69.326 54.520
356352 92.938 -69.315 54.490
356353 92.983 -69.303 54.459
356354 93.022 -69.297 54.427
356355 93.058 -69.298 54.394
356356 93.092 -69.299 54.361
356357 93.127 -69.300 54.330
356358 93.162 -69.300 54.299
356359 93.197 -69.301 54.268
356360 93.232 -69.302 54.238
356361 93.267 -69.303 54.209
356362 93.302 -69.303 54.180
356363 93.337 -69.304 54.152
356364 93.372 -69.305 54.126
356365 93.407 -69.306 54.101
356395 94.461 -69.356 53.359
356396 94.492 -69.354 53.342
356397 94.524 -69.352 53.326
356398 94.555 -69.350 53.310
356399 94.587 -69.347 53.294
356400 94.618 -69.345 53.280
356401 94.650 -69.343 53.267
356402 94.681 -69.340 53.253
356403 94.712 -69.338 53.240
356404 94.744 -69.335 53.228
356405 94.775 -69.333 53.216
356406 94.807 -69.331 53.206
356407 94.838 -69.328 53.199
356408 94.876 -69.328 53.193
356409 94.968 -69.341 53.187
356410 95.060 -69.354 53.182
356411 95.128 -69.344 53.179
356412 95.189 -69.328 53.178
356413 95.251 -69.313 53.180
356414 95.313 -69.297 53.182
356415 95.374 -69.281 53.185
356416 95.429 -69.269 53.186
356417 95.455 -69.269 53.186
356418 95.480 -69.268 53.182
356419 95.506 -69.268 53.182
356420 95.532 -69.267 53.184
356421 95.557 -69.267 53.187
356422 95.583 -69.266 53.190
356423 95.609 -69.266 53.189
356424 95.634 -69.266 53.185
356425 95.660 -69.265 53.178
356426 95.686 -69.265 53.171
356427 95.711 -69.264 53.175
356428 95.737 -69.264 53.188
356429 95.764 -69.259 53.202
356430 95.792 -69.254 53.214
356431 95.820 -69.249 53.218
356432 95.847 -69.244 53.214
356433 95.875 -69.238 53.205
356434 95.903 -69.233 53.189
356435 95.930 -69.228 53.167
356436 95.958 -69.222 53.146
356437 95.985 -69.218 53.125
356438 96.013 -69.212 53.101
356439 96.041 -69.207 53.075
356440 96.068 -69.202 53.049
356441 96.096 -69.197 53.022
356442 96.124 -69.191 52.998
356443 96.152 -69.186 52.978
356444 96.179 -69.190 52.970
356445 96.207 -69.203 52.960
356446 96.235 -69.215 52.953
356447 96.262 -69.228 52.947
356448 96.288 -69.242 52.939
356449 96.313 -69.255 52.924
356450 96.344 -69.267 52.909
356451 96.378 -69.279 52.892
356452 96.412 -69.291 52.874
356453 96.447 -69.303 52.853
356454 96.481 -69.316 52.831
356455 96.517 -69.328 52.807
356456 96.547 -69.338 52.781
356457 96.576 -69.346 52.757
356458 96.605 -69.353 52.729
356459 96.634 -69.361 52.699
356460 96.663 -69.369 52.667
356461 96.692 -69.377 52.633
356462 96.721 -69.385 52.599
356463 96.750 -69.392 52.564
356464 96.780 -69.397 52.528
356465 96.811 -69.400 52.490
356466 96.841 -69.403 52.454
356467 96.872 -69.406 52.419
356468 96.904 -69.415 52.385
356469 96.937 -69.426 52.355
356470 96.969 -69.437 52.327
356471 97.002 -69.447 52.299
356472 97.035 -69.461 52.272
356473 97.069 -69.477 52.246
356474 97.100 -69.484 52.221
356475 97.132 -69.490 52.196
356476 97.163 -69.496 52.172
356477 97.195 -69.502 52.146
356478 97.227 -69.509 52.118
356479 97.258 -69.515 52.090
356480 97.290 -69.521 52.063
356481 97.321 -69.528 52.035
356482 97.352 -69.535 52.007
356483 97.383 -69.543 51.980
356484 97.414 -69.551 51.950
356485 97.445 -69.558 51.921
356486 97.476 -69.565 51.896
356487 97.508 -69.573 51.876
356488 97.538 -69.580 51.862
356489 97.570 -69.587 51.850
356490 97.605 -69.591 51.835
356491 97.640 -69.595 51.815
356492 97.683 -69.598 51.792
356493 97.727 -69.601 51.764
356494 97.771 -69.604 51.735
356495 97.813 -69.608 51.703
356496 97.852 -69.611 51.670
356497 97.889 -69.615 51.631
356498 97.924 -69.619 51.592
356499 97.946 -69.625 51.551
356500 97.969 -69.631 51.510
356501 97.992 -69.637 51.467
356502 98.033 -69.645 51.421
356503 98.078 -69.650 51.373
356504 98.122 -69.654 51.322
356505 98.159 -69.659 51.271
356506 98.190 -69.666 51.221
356507 98.220 -69.673 51.173
356508 98.251 -69.679 51.124
356509 98.282 -69.686 51.074
356510 98.313 -69.694 51.020
356511 98.344 -69.701 50.964
356512 98.374 -69.707 50.906
356513 98.405 -69.714 50.847
356514 98.436 -69.721 50.789
356515 98.466 -69.728 50.730
356516 98.497 -69.736 50.669
356517 98.528 -69.742 50.608
356518 98.559 -69.749 50.546
356519 98.590 -69.756 50.483
356520 98.621 -69.763 50.420
356521 98.651 -69.770 50.358
356522 98.695 -69.786 50.295
356523 98.739 -69.803 50.232
356524 98.784 -69.820 50.171
356525 98.818 -69.814 50.112
356526 98.861 -69.820 50.056
356527 98.909 -69.836 50.000
356528 98.956 -69.850 49.945
356529 99.002 -69.865 49.889
356530 99.025 -69.877 49.833
356531 99.048 -69.889 49.774
356532 99.071 -69.901 49.717
356533 99.094 -69.912 49.661
356534 99.120 -69.924 49.606
356535 99.153 -69.933 49.552
356536 99.187 -69.942 49.500
356537 99.220 -69.952 49.449
356538 99.254 -69.961 49.400
356539 99.292 -69.967 49.351
356540 99.333 -69.971 49.301
356541 99.374 -69.974 49.253
356542 99.414 -69.979 49.207
356543 99.455 -69.983 49.162
356544 99.494 -69.988 49.116
356545 99.535 -69.993 49.073
356546 99.585 -69.996 49.031
356547 99.635 -69.999 48.990
356548 99.685 -70.002 48.949
356549 99.735 -70.005 48.911
356550 99.785 -70.008 48.872
356551 99.835 -70.011 48.835
356552 99.886 -70.014 48.798
356553 99.920 -70.047 48.762
356554 99.952 -70.086 48.728
356555 99.984 -70.124 48.694
356556 100.015 -70.163 48.662
356557 100.047 -70.203 48.630
356558 100.075 -70.237 48.600
356559 100.106 -70.231 48.569
356560 100.136 -70.225 48.540
356561 100.168 -70.218 48.511
356562 100.198 -70.212 48.485
356563 100.229 -70.205 48.460
356564 100.260 -70.198 48.435
356565 100.290 -70.192 48.410
356566 100.335 -70.198 48.385
356567 100.379 -70.208 48.363
356568 100.422 -70.218 48.342
356569 100.466 -70.228 48.323
356570 100.509 -70.237 48.304
356571 100.552 -70.247 48.278
356572 100.596 -70.257 48.249
356573 100.644 -70.270 48.216
356574 100.693 -70.283 48.181
356575 100.742 -70.297 48.144
356576 100.791 -70.309 48.107
356577 100.840 -70.322 48.070
356578 100.889 -70.336 48.034
356579 100.908 -70.351 47.997
356580 100.920 -70.366 47.959
356581 100.931 -70.382 47.921
356582 100.943 -70.398 47.883
356583 100.973 -70.418 47.844
356584 101.004 -70.425 47.803
356585 101.035 -70.433 47.762
356586 101.066 -70.440 47.721
356587 101.097 -70.447 47.680
356588 101.127 -70.455 47.643
356589 101.158 -70.462 47.602
356590 101.189 -70.469 47.560
356591 101.219 -70.476 47.517
356592 101.247 -70.481 47.472
356593 101.274 -70.485 47.424
356594 101.305 -70.500 47.378
356595 101.339 -70.516 47.332
356596 101.373 -70.533 47.287
356597 101.406 -70.552 47.245
356598 101.439 -70.571 47.202
356599 101.474 -70.579 47.157
356600 101.511 -70.583 47.112
356601 101.547 -70.587 47.067
356602 101.584 -70.591 47.021
356603 101.620 -70.595 46.972
356604 101.656 -70.599 46.921
356605 101.693 -70.603 46.867
356606 101.729 -70.607 46.819
356607 101.765 -70.610 46.773
356608 101.810 -70.610 46.728
356609 101.856 -70.608 46.684
356610 101.902 -70.607 46.641
356611 101.948 -70.605 46.598
356612 101.993 -70.604 46.556
356613 102.039 -70.602 46.515
356614 102.085 -70.601 46.474
357041 92.380 -69.414 54.826
357042 92.428 -69.407 54.800
357043 92.476 -69.399 54.772
357044 92.523 -69.391 54.743
357045 92.571 -69.384 54.716
357046 92.618 -69.376 54.687
357047 92.663 -69.365 54.657
357048 92.708 -69.353 54.629
357049 92.753 -69.341 54.598
357050 92.797 -69.329 54.568
357051 92.842 -69.318 54.537
357052 92.886 -69.306 54.507
357053 92.931 -69.294 54.475
357054 92.976 -69.282 54.444
357055 93.017 -69.275 54.412
357056 93.052 -69.275 54.378
357057 93.087 -69.276 54.346
357058 93.122 -69.277 54.314
357059 93.157 -69.278 54.282
357060 93.192 -69.279 54.252
357061 93.227 -69.280 54.222
357062 93.262 -69.281 54.193
357063 93.296 -69.281 54.165
357064 93.332 -69.282 54.139
357065 93.366 -69.283 54.113
357095 94.425 -69.332 53.388
357096 94.461 -69.335 53.372
357097 94.492 -69.333 53.356
357098 94.523 -69.330 53.341
357099 94.555 -69.328 53.325
357100 94.586 -69.325 53.310
357101 94.618 -69.323 53.297
357102 94.649 -69.321 53.285
357103 94.681 -69.319 53.273
357104 94.712 -69.316 53.261
357105 94.744 -69.314 53.249
357106 94.775 -69.311 53.238
357107 94.806 -69.309 53.230
357108 94.838 -69.306 53.224
357109 94.918 -69.317 53.220
357110 95.010 -69.330 53.214
357111 95.073 -69.317 53.211
357112 95.135 -69.301 53.209
357113 95.197 -69.285 53.209
357114 95.258 -69.269 53.211
357115 95.320 -69.254 53.214
357116 95.382 -69.238 53.215
357117 95.432 -69.228 53.216
357118 95.458 -69.227 53.216
357119 95.483 -69.227 53.213
357120 95.509 -69.226 53.213
357121 95.535 -69.225 53.216
357122 95.560 -69.225 53.219
357123 95.586 -69.225 53.222
357124 95.612 -69.225 53.223
357125 95.637 -69.224 53.220
357126 95.663 -69.224 53.214
357127 95.689 -69.223 53.209
357128 95.714 -69.223 53.204
357129 95.740 -69.222 53.208
357130 95.765 -69.222 53.218
357131 95.791 -69.222 53.228
357132 95.818 -69.218 53.231
357133 95.846 -69.212 53.227
357134 95.874 -69.207 53.217
357135 95.901 -69.202 53.203
357136 95.929 -69.197 53.184
357137 95.957 -69.191 53.167
357138 95.984 -69.186 53.148
357139 96.012 -69.181 53.126
357140 96.040 -69.175 53.100
357141 96.067 -69.170 53.070
357142 96.095 -69.165 53.042
357143 96.122 -69.160 53.019
357144 96.150 -69.155 53.004
357145 96.178 -69.160 53.001
357146 96.206 -69.172 52.994
357147 96.234 -69.185 52.987
357148 96.261 -69.198 52.978
357149 96.289 -69.210 52.966
357150 96.324 -69.223 52.951
357151 96.360 -69.236 52.935
357152 96.395 -69.248 52.918
357153 96.431 -69.261 52.899
357154 96.466 -69.274 52.877
357155 96.501 -69.286 52.854
357156 96.527 -69.295 52.829
357157 96.552 -69.303 52.802
357158 96.584 -69.312 52.776
357159 96.615 -69.321 52.744
357160 96.644 -69.328 52.711
357161 96.673 -69.336 52.677
357162 96.702 -69.344 52.640
357163 96.731 -69.351 52.604
357164 96.760 -69.360 52.567
357165 96.789 -69.367 52.529
357166 96.819 -69.371 52.488
357167 96.850 -69.373 52.452
357168 96.881 -69.376 52.417
357169 96.912 -69.385 52.384
357170 96.945 -69.396 52.355
357171 96.977 -69.407 52.327
357172 97.010 -69.418 52.298
357173 97.043 -69.429 52.270
357174 97.075 -69.440 52.242
357175 97.108 -69.450 52.216
357176 97.139 -69.456 52.190
357177 97.170 -69.463 52.163
357178 97.201 -69.469 52.136
357179 97.231 -69.476 52.107
357180 97.261 -69.482 52.079
357181 97.292 -69.488 52.049
357182 97.323 -69.495 52.019
357183 97.354 -69.503 51.990
357184 97.385 -69.510 51.960
357185 97.416 -69.517 51.930
357186 97.447 -69.525 51.901
357187 97.478 -69.532 51.875
357188 97.510 -69.540 51.855
357189 97.542 -69.545 51.841
357190 97.577 -69.550 51.829
357191 97.612 -69.554 51.815
357192 97.649 -69.557 51.795
357193 97.693 -69.560 51.772
357194 97.736 -69.563 51.746
357195 97.780 -69.566 51.718
357196 97.823 -69.570 51.688
357197 97.867 -69.573 51.658
357198 97.902 -69.577 51.620
357199 97.925 -69.583 51.580
357200 97.947 -69.589 51.539
357201 97.970 -69.595 51.496
357202 97.993 -69.601 51.452
357203 98.038 -69.610 51.407
357204 98.084 -69.616 51.359
357205 98.128 -69.619 51.309
357206 98.172 -69.623 51.260
357207 98.206 -69.629 51.212
357208 98.236 -69.635 51.163
357209 98.267 -69.642 51.114
357210 98.298 -69.649 51.063
357211 98.329 -69.656 51.007
357212 98.360 -69.664 50.950
357213 98.390 -69.670 50.891
357214 98.421 -69.677 50.831
357215 98.452 -69.684 50.773
357216 98.483 -69.691 50.714
357217 98.514 -69.698 50.654
357218 98.545 -69.705 50.592
357219 98.575 -69.712 50.529
357220 98.606 -69.719 50.466
357221 98.637 -69.726 50.403
357222 98.673 -69.736 50.341
357223 98.717 -69.754 50.279
357224 98.762 -69.770 50.218
357225 98.806 -69.787 50.160
357226 98.842 -69.784 50.104
357227 98.886 -69.792 50.050
357228 98.933 -69.808 49.994
357229 98.981 -69.823 49.938
357230 99.010 -69.835 49.883
357231 99.033 -69.847 49.826
357232 99.056 -69.858 49.767
357233 99.079 -69.870 49.711
357234 99.102 -69.883 49.655
357235 99.126 -69.894 49.601
357236 99.156 -69.905 49.547
357237 99.190 -69.914 49.495
357238 99.223 -69.924 49.445
357239 99.256 -69.933 49.395
357240 99.295 -69.938 49.346
357241 99.335 -69.942 49.296
357242 99.376 -69.946 49.249
357243 99.416 -69.950 49.203
357244 99.457 -69.954 49.157
357245 99.497 -69.958 49.112
357246 99.542 -69.965 49.068
357247 99.592 -69.965 49.026
357248 99.644 -69.965 48.984
357249 99.694 -69.968 48.942
357250 99.744 -69.971 48.904
357251 99.794 -69.974 48.865
357252 99.844 -69.977 48.827
357253 99.894 -69.980 48.790
357254 99.933 -70.003 48.753
357255 99.966 -70.042 48.719
357256 99.997 -70.080 48.685
357257 100.029 -70.121 48.653
357258 100.060 -70.161 48.621
357259 100.085 -70.190 48.589
357260 100.113 -70.204 48.558
357261 100.144 -70.197 48.528
357262 100.174 -70.190 48.498
357263 100.205 -70.184 48.472
357264 100.237 -70.177 48.447
357265 100.267 -70.171 48.421
357266 100.300 -70.165 48.396
357267 100.348 -70.173 48.371
357268 100.397 -70.181 48.348
357269 100.441 -70.190 48.328
357270 100.484 -70.200 48.308
357271 100.528 -70.210 48.289
357272 100.571 -70.220 48.264
357273 100.615 -70.230 48.235
357274 100.659 -70.240 48.203
357275 100.708 -70.253 48.169
357276 100.757 -70.267 48.135
357277 100.807 -70.280 48.097
357278 100.855 -70.293 48.059
357279 100.905 -70.306 48.023
357280 100.919 -70.322 47.986
357281 100.930 -70.337 47.948
357282 100.941 -70.353 47.912
357283 100.953 -70.369 47.876
357284 100.976 -70.382 47.839
357285 101.007 -70.390 47.798
357286 101.038 -70.397 47.757
357287 101.068 -70.404 47.716
357288 101.099 -70.412 47.675
357289 101.130 -70.419 47.637
357290 101.161 -70.426 47.596
357291 101.192 -70.433 47.553
357292 101.222 -70.441 47.509
357293 101.253 -70.448 47.464
357294 101.284 -70.455 47.417
357295 101.317 -70.473 47.369
357296 101.350 -70.492 47.320
357297 101.384 -70.511 47.272
357298 101.417 -70.530 47.226
357299 101.453 -70.535 47.180
357300 101.489 -70.539 47.136
357301 101.526 -70.544 47.091
357302 101.562 -70.547 47.047
357303 101.598 -70.551 47.001
357304 101.634 -70.555 46.953
357305 101.671 -70.559 46.903
357306 101.707 -70.563 46.853
357307 101.744 -70.566 46.807
357308 101.780 -70.570 46.762
357309 101.816 -70.574 46.717
357310 101.854 -70.577 46.674
357311 101.900 -70.576 46.632
357312 101.946 -70.575 46.590
357313 101.991 -70.573 46.549
357314 102.037 -70.572 46.508
357743 92.424 -69.390 54.791
357744 92.472 -69.382 54.763
357745 92.519 -69.375 54.734
357746 92.567 -69.368 54.706
357747 92.612 -69.356 54.677
357748 92.656 -69.344 54.647
357749 92.701 -69.333 54.617
357750 92.746 -69.321 54.587
357751 92.790 -69.309 54.556
357752 92.835 -69.297 54.525
357753 92.880 -69.285 54.494
357754 92.924 -69.273 54.463
357755 92.969 -69.262 54.431
357756 93.011 -69.253 54.398
357757 93.047 -69.253 54.364
357758 93.081 -69.254 54.332
357759 93.116 -69.255 54.300
357760 93.151 -69.256 54.269
357761 93.186 -69.256 54.238
357762 93.221 -69.257 54.208
357763 93.256 -69.258 54.180
357764 93.291 -69.259 54.152
357765 93.326 -69.260 54.125
357796 94.415 -69.298 53.401
357797 94.451 -69.302 53.385
357798 94.487 -69.306 53.370
357799 94.523 -69.309 53.355
357800 94.554 -69.306 53.340
357801 94.585 -69.303 53.325
357802 94.617 -69.301 53.314
357803 94.648 -69.299 53.302
357804 94.680 -69.297 53.291
357805 94.711 -69.294 53.280
357806 94.743 -69.292 53.271
357807 94.774 -69.290 53.261
357808 94.806 -69.288 53.254
357809 94.867 -69.293 53.250
357810 94.958 -69.305 53.246
357811 95.020 -69.289 53.241
357812 95.081 -69.273 53.239
357813 95.142 -69.258 53.238
357814 95.204 -69.242 53.238
357815 95.266 -69.226 53.241
357816 95.327 -69.211 53.244
357817 95.389 -69.195 53.245
357818 95.435 -69.186 53.244
357819 95.461 -69.186 53.245
357820 95.486 -69.185 53.243
357821 95.512 -69.184 53.244
357822 95.538 -69.184 53.246
357823 95.563 -69.184 53.249
357824 95.589 -69.184 53.252
357825 95.614 -69.183 53.254
357826 95.640 -69.183 53.252
357827 95.666 -69.182 53.247
357828 95.691 -69.182 53.241
357829 95.717 -69.181 53.234
357830 95.743 -69.181 53.232
357831 95.768 -69.181 53.236
357832 95.794 -69.180 53.242
357833 95.820 -69.180 53.242
357834 95.845 -69.179 53.236
357835 95.873 -69.175 53.225
357836 95.900 -69.170 53.211
357837 95.928 -69.165 53.194
357838 95.956 -69.160 53.180
357839 95.983 -69.155 53.164
357840 96.011 -69.150 53.144
357841 96.038 -69.144 53.120
357842 96.066 -69.139 53.091
357843 96.094 -69.134 53.065
357844 96.121 -69.128 53.045
357845 96.149 -69.123 53.032
357846 96.177 -69.130 53.029
357847 96.204 -69.142 53.023
357848 96.232 -69.155 53.013
357849 96.260 -69.167 52.999
357850 96.291 -69.177 52.984
357851 96.329 -69.186 52.968
357852 96.373 -69.203 52.951
357853 96.415 -69.219 52.934
357854 96.451 -69.231 52.916
357855 96.486 -69.244 52.893
357856 96.518 -69.255 52.867
357857 96.529 -69.258 52.839
357858 96.556 -69.266 52.810
357859 96.588 -69.275 52.780
357860 96.620 -69.285 52.747
357861 96.652 -69.294 52.713
357862 96.682 -69.303 52.676
357863 96.712 -69.311 52.639
357864 96.741 -69.319 52.602
357865 96.770 -69.326 52.564
357866 96.799 -69.334 52.526
357867 96.828 -69.341 52.487
357868 96.859 -69.344 52.450
357869 96.889 -69.346 52.415
357870 96.922 -69.355 52.383
357871 96.954 -69.366 52.355
357872 96.986 -69.377 52.327
357873 97.019 -69.388 52.298
357874 97.051 -69.399 52.268
357875 97.083 -69.408 52.239
357876 97.111 -69.416 52.212
357877 97.141 -69.423 52.184
357878 97.172 -69.429 52.156
357879 97.203 -69.436 52.127
357880 97.233 -69.442 52.096
357881 97.263 -69.448 52.064
357882 97.294 -69.455 52.033
357883 97.325 -69.463 52.002
357884 97.356 -69.469 51.970
357885 97.387 -69.477 51.940
357886 97.418 -69.485 51.909
357887 97.449 -69.492 51.880
357888 97.480 -69.500 51.855
357889 97.515 -69.504 51.834
357890 97.549 -69.508 51.819
357891 97.584 -69.512 51.806
357892 97.619 -69.516 51.790
357893 97.659 -69.519 51.771
357894 97.702 -69.523 51.748
357895 97.746 -69.526 51.723
357896 97.789 -69.529 51.698
357897 97.833 -69.532 51.672
357898 97.874 -69.535 51.644
357899 97.903 -69.541 51.607
357900 97.926 -69.548 51.568
357901 97.948 -69.554 51.527
357902 97.971 -69.560 51.483
357903 97.996 -69.566 51.439
357904 98.043 -69.575 51.393
357905 98.089 -69.582 51.345
357906 98.133 -69.585 51.296
357907 98.177 -69.588 51.247
357908 98.221 -69.591 51.199
357909 98.253 -69.598 51.150
357910 98.284 -69.605 51.099
357911 98.314 -69.612 51.047
357912 98.345 -69.619 50.989
357913 98.376 -69.626 50.930
357914 98.407 -69.633 50.870
357915 98.437 -69.640 50.810
357916 98.468 -69.647 50.750
357917 98.499 -69.654 50.692
357918 98.530 -69.660 50.633
357919 98.561 -69.667 50.572
357920 98.591 -69.675 50.510
357921 98.622 -69.682 50.448
357922 98.653 -69.689 50.386
357923 98.695 -69.704 50.325
357924 98.739 -69.720 50.266
357925 98.784 -69.737 50.209
357926 98.829 -69.754 50.151
357927 98.865 -69.754 50.096
357928 98.911 -69.764 50.042
357929 98.958 -69.780 49.987
357930 98.996 -69.793 49.930
357931 99.019 -69.805 49.873
357932 99.042 -69.817 49.816
357933 99.065 -69.829 49.759
357934 99.088 -69.840 49.704
357935 99.111 -69.852 49.649
357936 99.134 -69.864 49.595
357937 99.159 -69.876 49.542
357938 99.192 -69.885 49.490
357939 99.225 -69.895 49.439
357940 99.259 -69.904 49.389
357941 99.297 -69.910 49.340
357942 99.338 -69.914 49.292
357943 99.378 -69.917 49.244
357944 99.419 -69.921 49.198
357945 99.459 -69.926 49.152
357946 99.501 -69.931 49.107
357947 99.548 -69.941 49.063
357948 99.593 -69.950 49.019
357949 99.645 -69.949 48.976
357950 99.697 -69.947 48.934
357951 99.749 -69.945 48.895
357952 99.801 -69.943 48.857
357953 99.852 -69.944 48.819
357954 99.902 -69.947 48.781
357955 99.947 -69.959 48.744
357956 99.979 -69.999 48.709
357957 100.010 -70.039 48.674
357958 100.042 -70.080 48.642
357959 100.070 -70.114 48.611
357960 100.096 -70.143 48.579
357961 100.121 -70.173 48.548
357962 100.151 -70.169 48.517
357963 100.182 -70.162 48.486
357964 100.213 -70.156 48.460
357965 100.244 -70.149 48.434
357966 100.274 -70.143 48.408
357967 100.312 -70.141 48.384
357968 100.360 -70.149 48.359
357969 100.409 -70.156 48.338
357970 100.457 -70.164 48.318
357971 100.503 -70.172 48.297
357972 100.547 -70.183 48.274
357973 100.590 -70.193 48.249
357974 100.634 -70.203 48.219
357975 100.677 -70.212 48.188
357976 100.724 -70.224 48.155
357977 100.773 -70.237 48.122
357978 100.822 -70.250 48.085
357979 100.871 -70.263 48.050
357980 100.918 -70.277 48.015
357981 100.929 -70.292 47.978
357982 100.941 -70.308 47.940
357983 100.952 -70.324 47.904
357984 100.964 -70.339 47.868
357985 100.989 -70.350 47.831
357986 101.015 -70.361 47.791
357987 101.046 -70.369 47.750
357988 101.077 -70.376 47.708
357989 101.108 -70.384 47.667
357990 101.139 -70.391 47.627
357991 101.170 -70.399 47.584
357992 101.201 -70.406 47.540
357993 101.232 -70.414 47.495
357994 101.263 -70.422 47.451
357995 101.295 -70.432 47.404
357996 101.328 -70.451 47.355
357997 101.361 -70.470 47.305
357998 101.395 -70.488 47.255
357999 101.431 -70.491 47.206
358000 101.467 -70.496 47.160
358001 101.503 -70.500 47.115
358002 101.540 -70.503 47.071
358003 101.576 -70.507 47.026
358004 101.613 -70.511 46.980
358005 101.649 -70.515 46.933
358006 101.685 -70.519 46.885
358007 101.722 -70.522 46.839
358008 101.758 -70.526 46.794
358009 101.795 -70.530 46.750
358010 101.831 -70.534 46.707
358011 101.868 -70.538 46.664
358012 101.904 -70.542 46.622
358013 101.944 -70.544 46.581
358014 101.989 -70.542 46.541
358015 102.035 -70.541 46.500
358446 92.515 -69.359 54.727
358447 92.560 -69.347 54.697
358448 92.605 -69.336 54.668
358449 92.650 -69.324 54.639
358450 92.694 -69.312 54.607
358451 92.739 -69.300 54.577
358452 92.784 -69.288 54.546
358453 92.828 -69.276 54.514
358454 92.873 -69.265 54.483
358455 92.918 -69.253 54.451
358456 92.962 -69.240 54.418
358457 93.006 -69.230 54.385
358458 93.041 -69.231 54.352
358459 93.076 -69.232 54.320
358460 93.111 -69.233 54.288
358461 93.146 -69.234 54.256
358462 93.181 -69.234 54.226
358463 93.216 -69.235 54.196
358464 93.251 -69.236 54.168
358465 93.285 -69.237 54.140
358497 94.404 -69.263 53.413
358498 94.440 -69.267 53.398
358499 94.476 -69.271 53.383
358500 94.512 -69.275 53.369
358501 94.549 -69.278 53.356
358502 94.585 -69.282 53.342
358503 94.616 -69.280 53.331
358504 94.648 -69.278 53.320
358505 94.679 -69.275 53.310
358506 94.710 -69.272 53.300
358507 94.742 -69.270 53.290
358508 94.773 -69.268 53.282
358509 94.816 -69.269 53.277
358510 94.904 -69.277 53.275
358511 94.965 -69.262 53.272
358512 95.027 -69.246 53.267
358513 95.088 -69.230 53.264
358514 95.150 -69.215 53.262
358515 95.211 -69.199 53.263
358516 95.273 -69.183 53.267
358517 95.334 -69.168 53.270
358518 95.396 -69.152 53.271
358519 95.438 -69.144 53.272
358520 95.463 -69.144 53.273
358521 95.489 -69.143 53.273
358522 95.515 -69.143 53.273
358523 95.540 -69.143 53.275
358524 95.566 -69.143 53.278
358525 95.592 -69.142 53.281
358526 95.617 -69.142 53.281
358527 95.643 -69.141 53.280
358528 95.669 -69.141 53.276
358529 95.694 -69.140 53.269
358530 95.720 -69.140 53.261
358531 95.746 -69.140 53.258
358532 95.771 -69.139 53.257
358533 95.797 -69.139 53.256
358534 95.823 -69.138 53.251
358535 95.848 -69.138 53.241
358536 95.874 -69.137 53.229
358537 95.900 -69.137 53.215
358538 95.926 -69.134 53.201
358539 95.954 -69.128 53.189
358540 95.982 -69.123 53.175
358541 96.010 -69.118 53.159
358542 96.037 -69.112 53.137
358543 96.065 -69.107 53.111
358544 96.093 -69.103 53.088
358545 96.120 -69.097 53.070
358546 96.148 -69.092 53.060
358547 96.176 -69.099 53.055
358548 96.203 -69.112 53.047
358549 96.231 -69.124 53.033
358550 96.261 -69.134 53.015
358551 96.296 -69.140 52.996
358552 96.332 -69.148 52.977
358553 96.377 -69.165 52.960
358554 96.420 -69.181 52.944
358555 96.464 -69.199 52.927
358556 96.506 -69.214 52.902
358557 96.520 -69.218 52.874
358558 96.532 -69.221 52.843
358559 96.560 -69.229 52.812
358560 96.592 -69.239 52.779
358561 96.624 -69.249 52.745
358562 96.656 -69.258 52.710
358563 96.688 -69.268 52.673
358564 96.720 -69.277 52.635
358565 96.750 -69.286 52.597
358566 96.780 -69.294 52.559
358567 96.808 -69.301 52.520
358568 96.837 -69.309 52.482
358569 96.868 -69.313 52.446
358570 96.898 -69.316 52.413
358571 96.930 -69.325 52.382
358572 96.962 -69.336 52.353
358573 96.995 -69.347 52.325
358574 97.028 -69.358 52.295
358575 97.058 -69.366 52.264
358576 97.088 -69.373 52.233
358577 97.116 -69.381 52.204
358578 97.144 -69.389 52.174
358579 97.174 -69.396 52.144
358580 97.204 -69.403 52.113
358581 97.235 -69.409 52.081
358582 97.265 -69.416 52.048
358583 97.296 -69.422 52.016
358584 97.327 -69.429 51.983
358585 97.358 -69.437 51.951
358586 97.389 -69.444 51.921
358587 97.420 -69.452 51.890
358588 97.453 -69.458 51.861
358589 97.488 -69.462 51.836
358590 97.522 -69.466 51.815
358591 97.557 -69.470 51.798
358592 97.591 -69.475 51.782
358593 97.626 -69.479 51.766
358594 97.668 -69.482 51.746
358595 97.712 -69.485 51.725
358596 97.755 -69.488 51.704
358597 97.799 -69.491 51.682
358598 97.843 -69.494 51.658
358599 97.873 -69.500 51.632
358600 97.901 -69.506 51.596
358601 97.927 -69.512 51.557
358602 97.949 -69.518 51.514
358603 97.972 -69.524 51.470
358604 98.001 -69.531 51.426
358605 98.048 -69.539 51.380
358606 98.094 -69.547 51.332
358607 98.139 -69.551 51.282
358608 98.182 -69.554 51.233
358609 98.227 -69.557 51.184
358610 98.269 -69.561 51.134
358611 98.300 -69.568 51.082
358612 98.331 -69.575 51.028
358613 98.361 -69.582 50.970
358614 98.392 -69.588 50.910
358615 98.423 -69.595 50.850
358616 98.454 -69.603 50.790
358617 98.485 -69.610 50.730
358618 98.515 -69.617 50.672
358619 98.546 -69.623 50.613
358620 98.577 -69.630 50.553
358621 98.607 -69.637 50.492
358622 98.638 -69.645 50.431
358623 98.673 -69.654 50.371
358624 98.717 -69.671 50.312
358625 98.762 -69.688 50.255
358626 98.806 -69.704 50.199
358627 98.851 -69.721 50.142
358628 98.889 -69.723 50.087
358629 98.936 -69.736 50.032
358630 98.981 -69.751 49.977
358631 99.004 -69.763 49.920
358632 99.027 -69.775 49.864
358633 99.050 -69.786 49.807
358634 99.073 -69.798 49.752
358635 99.096 -69.811 49.697
358636 99.119 -69.822 49.643
358637 99.142 -69.834 49.589
358638 99.165 -69.845 49.536
358639 99.195 -69.856 49.485
358640 99.228 -69.866 49.435
358641 99.261 -69.875 49.385
358642 99.299 -69.881 49.336
358643 99.340 -69.885 49.288
358644 99.380 -69.889 49.240
358645 99.421 -69.893 49.194
358646 99.462 -69.898 49.148
358647 99.507 -69.907 49.102
358648 99.554 -69.916 49.058
358649 99.600 -69.926 49.014
358650 99.646 -69.934 48.970
358651 99.698 -69.932 48.927
358652 99.750 -69.930 48.888
358653 99.802 -69.929 48.849
358654 99.854 -69.927 48.810
358655 99.906 -69.925 48.772
358656 99.958 -69.922 48.735
358657 99.992 -69.957 48.698
358658 100.023 -69.998 48.664
358659 100.054 -70.038 48.632
358660 100.080 -70.068 48.601
358661 100.106 -70.097 48.570
358662 100.131 -70.127 48.538
358663 100.158 -70.141 48.506
358664 100.189 -70.134 48.476
358665 100.220 -70.128 48.449
358666 100.251 -70.121 48.423
358667 100.282 -70.115 48.397
358668 100.324 -70.118 48.372
358669 100.372 -70.124 48.348
358670 100.421 -70.132 48.328
358671 100.469 -70.140 48.308
358672 100.517 -70.147 48.286
358673 100.565 -70.155 48.260
358674 100.609 -70.165 48.234
358675 100.652 -70.174 48.206
358676 100.696 -70.184 48.175
358677 100.740 -70.194 48.142
358678 100.788 -70.208 48.108
358679 100.837 -70.221 48.075
358680 100.887 -70.234 48.041
358681 100.929 -70.247 48.006
358682 100.940 -70.263 47.969
358683 100.952 -70.278 47.931
358684 100.963 -70.294 47.895
358685 100.986 -70.306 47.859
358686 101.011 -70.317 47.822
358687 101.037 -70.328 47.781
358688 101.063 -70.339 47.740
358689 101.089 -70.350 47.698
358690 101.118 -70.359 47.656
358691 101.149 -70.366 47.614
358692 101.180 -70.374 47.568
358693 101.211 -70.381 47.523
358694 101.243 -70.389 47.478
358695 101.273 -70.397 47.432
358696 101.305 -70.409 47.386
358697 101.339 -70.428 47.337
358698 101.373 -70.444 47.287
358699 101.409 -70.448 47.237
358700 101.446 -70.452 47.187
358701 101.482 -70.456 47.140
358702 101.518 -70.460 47.094
358703 101.555 -70.463 47.050
358704 101.591 -70.467 47.005
358705 101.627 -70.471 46.959
358706 101.664 -70.475 46.914
358707 101.700 -70.478 46.869
358708 101.737 -70.482 46.824
358709 101.773 -70.486 46.780
358710 101.809 -70.491 46.737
358711 101.845 -70.494 46.695
358712 101.882 -70.498 46.654
358713 101.918 -70.502 46.612
358714 101.955 -70.506 46.572
358715 101.991 -70.510 46.531
359149 92.598 -69.316 54.661
359150 92.643 -69.303 54.632
359151 92.688 -69.291 54.600
359152 92.732 -69.280 54.569
359153 92.777 -69.268 54.538
359154 92.821 -69.256 54.506
359155 92.866 -69.244 54.474
359156 92.911 -69.232 54.440
359157 92.955 -69.220 54.408
359158 93.000 -69.209 54.375
359159 93.036 -69.209 54.341
359160 93.070 -69.209 54.309
359161 93.105 -69.210 54.277
359162 93.140 -69.211 54.246
359163 93.175 -69.212 54.215
359164 93.210 -69.212 54.186
359165 93.245 -69.213 54.157
359197 94.363 -69.240 53.438
359198 94.398 -69.240 53.424
359199 94.433 -69.241 53.410
359200 94.468 -69.242 53.396
359201 94.503 -69.243 53.383
359202 94.538 -69.244 53.370
359203 94.574 -69.247 53.357
359204 94.610 -69.251 53.347
359205 94.646 -69.255 53.337
359206 94.679 -69.253 53.327
359207 94.710 -69.251 53.318
359208 94.741 -69.248 53.311
359209 94.773 -69.246 53.303
359210 94.849 -69.250 53.299
359211 94.911 -69.234 53.297
359212 94.972 -69.219 53.294
359213 95.034 -69.203 53.289
359214 95.096 -69.187 53.285
359215 95.157 -69.172 53.283
359216 95.218 -69.156 53.285
359217 95.280 -69.140 53.289
359218 95.342 -69.125 53.292
359219 95.403 -69.109 53.294
359220 95.441 -69.103 53.295
359221 95.466 -69.103 53.297
359222 95.492 -69.103 53.296
359223 95.518 -69.102 53.296
359224 95.543 -69.102 53.295
359225 95.569 -69.101 53.296
359226 95.595 -69.101 53.296
359227 95.620 -69.100 53.295
359228 95.646 -69.100 53.293
359229 95.672 -69.099 53.290
359230 95.697 -69.099 53.285
359231 95.723 -69.099 53.278
359232 95.749 -69.098 53.273
359233 95.774 -69.098 53.270
359234 95.800 -69.097 53.265
359235 95.826 -69.097 53.256
359236 95.851 -69.096 53.242
359237 95.877 -69.096 53.228
359238 95.902 -69.096 53.215
359239 95.928 -69.096 53.203
359240 95.954 -69.095 53.193
359241 95.981 -69.092 53.182
359242 96.009 -69.087 53.168
359243 96.036 -69.081 53.150
359244 96.063 -69.076 53.128
359245 96.091 -69.071 53.109
359246 96.119 -69.065 53.095
359247 96.147 -69.060 53.086
359248 96.174 -69.068 53.078
359249 96.202 -69.081 53.065
359250 96.231 -69.093 53.048
359251 96.266 -69.098 53.025
359252 96.301 -69.103 52.999
359253 96.336 -69.110 52.978
359254 96.380 -69.127 52.962
359255 96.424 -69.143 52.948
359256 96.468 -69.161 52.932
359257 96.507 -69.176 52.906
359258 96.522 -69.181 52.876
359259 96.534 -69.184 52.843
359260 96.564 -69.193 52.807
359261 96.596 -69.203 52.773
359262 96.628 -69.212 52.739
359263 96.660 -69.222 52.703
359264 96.692 -69.231 52.666
359265 96.724 -69.240 52.628
359266 96.756 -69.250 52.590
359267 96.788 -69.260 52.551
359268 96.818 -69.269 52.513
359269 96.847 -69.276 52.475
359270 96.876 -69.284 52.442
359271 96.907 -69.287 52.411
359272 96.939 -69.295 52.381
359273 96.971 -69.306 52.350
359274 97.004 -69.317 52.321
359275 97.033 -69.324 52.289
359276 97.063 -69.331 52.256
359277 97.092 -69.338 52.225
359278 97.120 -69.346 52.194
359279 97.148 -69.354 52.162
359280 97.176 -69.363 52.130
359281 97.206 -69.369 52.098
359282 97.237 -69.376 52.065
359283 97.267 -69.382 52.032
359284 97.298 -69.389 51.999
359285 97.329 -69.397 51.965
359286 97.360 -69.404 51.934
359287 97.391 -69.412 51.904
359288 97.425 -69.416 51.874
359289 97.460 -69.420 51.845
359290 97.495 -69.425 51.820
359291 97.529 -69.429 51.798
359292 97.564 -69.433 51.779
359293 97.598 -69.437 51.762
359294 97.634 -69.441 51.744
359295 97.677 -69.444 51.725
359296 97.721 -69.447 51.706
359297 97.764 -69.450 51.688
359298 97.808 -69.454 51.668
359299 97.843 -69.458 51.646
359300 97.871 -69.464 51.620
359301 97.900 -69.470 51.585
359302 97.928 -69.476 51.545
359303 97.950 -69.482 51.503
359304 97.973 -69.488 51.459
359305 98.006 -69.496 51.414
359306 98.053 -69.504 51.367
359307 98.100 -69.513 51.319
359308 98.144 -69.516 51.269
359309 98.188 -69.519 51.218
359310 98.232 -69.523 51.168
359311 98.276 -69.526 51.117
359312 98.316 -69.531 51.064
359313 98.347 -69.538 51.009
359314 98.377 -69.544 50.951
359315 98.408 -69.551 50.891
359316 98.439 -69.558 50.831
359317 98.470 -69.565 50.771
359318 98.501 -69.573 50.712
359319 98.531 -69.579 50.654
359320 98.562 -69.586 50.595
359321 98.593 -69.593 50.535
359322 98.624 -69.600 50.475
359323 98.654 -69.607 50.415
359324 98.695 -69.621 50.357
359325 98.739 -69.638 50.300
359326 98.784 -69.654 50.245
359327 98.829 -69.672 50.189
359328 98.873 -69.689 50.133
359329 98.913 -69.694 50.077
359330 98.960 -69.708 50.022
359331 98.989 -69.721 49.967
359332 99.012 -69.733 49.911
359333 99.035 -69.745 49.855
359334 99.058 -69.757 49.799
359335 99.081 -69.768 49.745
359336 99.104 -69.780 49.691
359337 99.127 -69.792 49.637
359338 99.150 -69.804 49.584
359339 99.173 -69.816 49.531
359340 99.197 -69.827 49.479
359341 99.230 -69.837 49.430
359342 99.264 -69.846 49.381
359343 99.301 -69.853 49.332
359344 99.342 -69.857 49.284
359345 99.382 -69.861 49.237
359346 99.423 -69.864 49.190
359347 99.468 -69.873 49.144
359348 99.513 -69.883 49.098
359349 99.560 -69.892 49.053
359350 99.606 -69.901 49.009
359351 99.651 -69.911 48.965
359352 99.699 -69.917 48.921
359353 99.751 -69.916 48.880
359354 99.803 -69.914 48.841
359355 99.855 -69.912 48.801
359356 99.907 -69.910 48.762
359357 99.959 -69.905 48.725
359358 100.004 -69.916 48.688
359359 100.036 -69.956 48.654
359360 100.064 -69.992 48.622
359361 100.090 -70.021 48.591
359362 100.116 -70.051 48.561
359363 100.141 -70.080 48.528
359364 100.167 -70.110 48.496
359365 100.197 -70.107 48.466
359366 100.227 -70.100 48.440
359367 100.258 -70.093 48.413
359368 100.289 -70.087 48.387
359369 100.336 -70.093 48.362
359370 100.385 -70.101 48.339
359371 100.433 -70.109 48.318
359372 100.481 -70.115 48.298
359373 100.529 -70.123 48.276
359374 100.577 -70.130 48.250
359375 100.626 -70.138 48.223
359376 100.671 -70.147 48.195
359377 100.715 -70.157 48.165
359378 100.758 -70.167 48.132
359379 100.804 -70.178 48.098
359380 100.853 -70.191 48.064
359381 100.902 -70.204 48.030
359382 100.939 -70.218 47.994
359383 100.951 -70.234 47.957
359384 100.962 -70.250 47.919
359385 100.982 -70.262 47.882
359386 101.007 -70.273 47.845
359387 101.033 -70.284 47.806
359388 101.059 -70.295 47.766
359389 101.085 -70.306 47.726
359390 101.110 -70.317 47.684
359391 101.136 -70.328 47.641
359392 101.162 -70.339 47.598
359393 101.191 -70.348 47.552
359394 101.221 -70.356 47.505
359395 101.253 -70.363 47.458
359396 101.284 -70.371 47.411
359397 101.317 -70.388 47.362
359398 101.351 -70.400 47.313
359399 101.387 -70.404 47.264
359400 101.424 -70.408 47.215
359401 101.460 -70.412 47.165
359402 101.496 -70.416 47.118
359403 101.533 -70.419 47.073
359404 101.569 -70.423 47.029
359405 101.606 -70.427 46.985
359406 101.642 -70.431 46.941
359407 101.678 -70.434 46.897
359408 101.714 -70.439 46.853
359409 101.751 -70.443 46.810
359410 101.787 -70.447 46.768
359411 101.824 -70.450 46.726
359412 101.860 -70.454 46.684
359413 101.896 -70.458 46.643
359414 101.933 -70.462 46.602
359415 101.969 -70.466 46.563
359854 92.770 -69.247 54.531
359855 92.814 -69.235 54.499
359856 92.859 -69.223 54.467
359857 92.904 -69.212 54.433
359858 92.948 -69.200 54.400
359859 92.993 -69.187 54.367
359860 93.030 -69.187 54.334
359861 93.065 -69.187 54.301
359862 93.100 -69.188 54.269
359863 93.135 -69.189 54.237
359864 93.169 -69.190 54.206
359865 93.205 -69.190 54.177
359898 94.358 -69.218 53.448
359899 94.393 -69.219 53.435
359900 94.428 -69.219 53.422
359901 94.463 -69.220 53.408
359902 94.498 -69.221 53.395
359903 94.533 -69.222 53.383
359904 94.567 -69.222 53.372
359905 94.603 -69.223 53.362
359906 94.637 -69.224 53.353
359907 94.673 -69.225 53.344
359908 94.708 -69.228 53.336
359909 94.741 -69.227 53.332
359910 94.795 -69.222 53.325
359911 94.856 -69.206 53.320
359912 94.918 -69.190 53.316
359913 94.980 -69.175 53.313
359914 95.041 -69.159 53.307
359915 95.103 -69.143 53.303
359916 95.165 -69.128 53.301
359917 95.226 -69.112 53.304
359918 95.287 -69.096 53.308
359919 95.349 -69.081 53.311
359920 95.411 -69.065 53.312
359921 95.444 -69.062 53.314
359922 95.469 -69.062 53.315
359923 95.495 -69.061 53.313
359924 95.521 -69.061 53.312
359925 95.546 -69.060 53.309
359926 95.572 -69.060 53.306
359927 95.598 -69.059 53.302
359928 95.623 -69.059 53.301
359929 95.649 -69.059 53.299
359930 95.675 -69.058 53.295
359931 95.700 -69.058 53.292
359932 95.726 -69.057 53.286
359933 95.751 -69.057 53.281
359934 95.777 -69.056 53.276
359935 95.803 -69.056 53.269
359936 95.828 -69.056 53.258
359937 95.854 -69.056 53.242
359938 95.880 -69.055 53.225
359939 95.905 -69.054 53.210
359940 95.931 -69.054 53.200
359941 95.957 -69.053 53.191
359942 95.982 -69.053 53.183
359943 96.008 -69.052 53.173
359944 96.035 -69.049 53.158
359945 96.062 -69.045 53.138
359946 96.090 -69.040 53.122
359947 96.118 -69.034 53.110
359948 96.146 -69.029 53.101
359949 96.173 -69.038 53.092
359950 96.201 -69.050 53.076
359951 96.236 -69.056 53.054
359952 96.270 -69.062 53.029
359953 96.305 -69.067 53.000
359954 96.340 -69.072 52.976
359955 96.384 -69.089 52.958
359956 96.428 -69.106 52.944
359957 96.470 -69.122 52.929
359958 96.504 -69.137 52.906
359959 96.524 -69.144 52.877
359960 96.536 -69.147 52.843
359961 96.568 -69.156 52.805
359962 96.600 -69.166 52.768
359963 96.632 -69.176 52.731
359964 96.664 -69.185 52.693
359965 96.696 -69.195 52.656
359966 96.728 -69.204 52.617
359967 96.760 -69.214 52.580
359968 96.792 -69.224 52.542
359969 96.824 -69.233 52.505
359970 96.856 -69.243 52.468
359971 96.886 -69.250 52.436
359972 96.915 -69.256 52.405
359973 96.947 -69.265 52.374
359974 96.979 -69.275 52.343
359975 97.009 -69.282 52.312
359976 97.038 -69.289 52.278
359977 97.068 -69.296 52.245
359978 97.097 -69.303 52.213
359979 97.125 -69.311 52.180
359980 97.153 -69.319 52.147
359981 97.181 -69.328 52.115
359982 97.209 -69.336 52.082
359983 97.238 -69.343 52.048
359984 97.269 -69.350 52.015
359985 97.300 -69.356 51.982
359986 97.331 -69.364 51.950
359987 97.363 -69.370 51.919
359988 97.397 -69.375 51.890
359989 97.432 -69.378 51.860
359990 97.467 -69.383 51.832
359991 97.502 -69.387 51.807
359992 97.536 -69.391 51.786
359993 97.571 -69.395 51.766
359994 97.605 -69.400 51.749
359995 97.643 -69.403 51.730
359996 97.686 -69.407 51.711
359997 97.730 -69.410 51.693
359998 97.774 -69.413 51.676
359999 97.813 -69.416 51.656
360000 97.842 -69.422 51.635
360001 97.870 -69.429 51.608
360002 97.899 -69.435 51.574
360003 97.928 -69.441 51.535
360004 97.951 -69.447 51.493
360005 97.974 -69.453 51.450
360006 98.011 -69.460 51.404
360007 98.058 -69.469 51.357
360008 98.105 -69.478 51.307
360009 98.149 -69.482 51.256
360010 98.194 -69.485 51.204
360011 98.237 -69.488 51.152
360012 98.282 -69.491 51.099
360013 98.326 -69.495 51.046
360014 98.363 -69.501 50.992
360015 98.393 -69.507 50.933
360016 98.425 -69.514 50.874
360017 98.455 -69.521 50.815
360018 98.486 -69.528 50.756
360019 98.517 -69.535 50.698
360020 98.548 -69.542 50.639
360021 98.578 -69.549 50.579
360022 98.609 -69.556 50.520
360023 98.640 -69.563 50.460
360024 98.673 -69.571 50.401
360025 98.717 -69.588 50.344
360026 98.762 -69.605 50.288
360027 98.806 -69.622 50.233
360028 98.851 -69.639 50.178
360029 98.896 -69.656 50.123
360030 98.938 -69.666 50.068
360031 98.975 -69.679 50.013
360032 98.998 -69.691 49.958
360033 99.021 -69.703 49.902
360034 99.044 -69.714 49.847
360035 99.067 -69.726 49.793
360036 99.090 -69.739 49.739
360037 99.113 -69.750 49.685
360038 99.136 -69.762 49.631
360039 99.159 -69.774 49.578
360040 99.182 -69.786 49.525
360041 99.205 -69.798 49.475
360042 99.233 -69.808 49.426
360043 99.266 -69.817 49.377
360044 99.303 -69.824 49.328
360045 99.344 -69.828 49.281
360046 99.384 -69.833 49.233
360047 99.428 -69.839 49.187
360048 99.474 -69.848 49.141
360049 99.519 -69.858 49.094
360050 99.566 -69.867 49.049
360051 99.612 -69.877 49.004
360052 99.657 -69.886 48.960
360053 99.704 -69.895 48.916
360054 99.752 -69.901 48.875
360055 99.803 -69.899 48.833
360056 99.855 -69.898 48.793
360057 99.908 -69.894 48.754
360058 99.960 -69.889 48.715
360059 100.013 -69.884 48.679
360060 100.049 -69.915 48.645
360061 100.074 -69.946 48.612
360062 100.100 -69.975 48.580
360063 100.126 -70.005 48.549
360064 100.151 -70.034 48.518
360065 100.177 -70.063 48.486
360066 100.204 -70.079 48.456
360067 100.234 -70.072 48.429
360068 100.266 -70.065 48.402
360069 100.299 -70.062 48.376
360070 100.348 -70.069 48.351
360071 100.396 -70.077 48.328
360072 100.444 -70.084 48.307
360073 100.493 -70.092 48.286
360074 100.541 -70.099 48.264
360075 100.590 -70.106 48.238
360076 100.638 -70.114 48.211
360077 100.686 -70.121 48.182
360078 100.734 -70.130 48.151
360079 100.777 -70.140 48.118
360080 100.821 -70.149 48.085
360081 100.868 -70.162 48.050
360082 100.917 -70.174 48.014
360083 100.950 -70.189 47.977
360084 100.961 -70.205 47.940
360085 100.978 -70.218 47.902
360086 101.004 -70.229 47.864
360087 101.030 -70.240 47.827
360088 101.055 -70.251 47.788
360089 101.081 -70.262 47.749
360090 101.107 -70.273 47.709
360091 101.132 -70.284 47.667
360092 101.158 -70.295 47.624
360093 101.184 -70.306 47.580
360094 101.209 -70.317 47.534
360095 101.235 -70.328 47.487
360096 101.263 -70.337 47.440
360097 101.294 -70.347 47.392
360098 101.329 -70.356 47.342
360099 101.365 -70.360 47.292
360100 101.402 -70.364 47.242
360101 101.439 -70.368 47.192
360102 101.475 -70.372 47.144
360103 101.511 -70.375 47.097
360104 101.547 -70.379 47.053
360105 101.584 -70.383 47.009
360106 101.620 -70.387 46.966
360107 101.656 -70.391 46.923
360108 101.693 -70.395 46.881
360109 101.729 -70.399 46.839
360110 101.766 -70.403 46.797
360111 101.802 -70.406 46.756
360112 101.838 -70.410 46.714
360113 101.875 -70.414 46.673
360114 101.911 -70.418 46.633
360115 101.948 -70.422 46.592
360116 101.984 -70.425 46.554
360599 94.352 -69.195 53.456
360600 94.387 -69.196 53.444
360601 94.422 -69.197 53.432
360602 94.457 -69.197 53.419
360603 94.492 -69.199 53.407
360604 94.527 -69.200 53.395
360605 94.562 -69.200 53.384
360606 94.597 -69.201 53.375
360607 94.632 -69.202 53.367
360608 94.667 -69.203 53.359
360609 94.702 -69.203 53.353
360610 94.755 -69.171 53.350
360611 94.808 -69.166 53.344
360612 94.864 -69.163 53.337
360613 94.925 -69.147 53.332
360614 94.987 -69.132 53.327
360615 95.049 -69.116 53.321
360616 95.110 -69.100 53.317
360617 95.172 -69.085 53.316
360618 95.233 -69.069 53.318
360619 95.295 -69.053 53.321
360620 95.356 -69.038 53.322
360621 95.418 -69.022 53.323
360622 95.447 -69.021 53.325
360623 95.472 -69.020 53.325
360624 95.498 -69.020 53.323
360625 95.524 -69.019 53.320
360626 95.549 -69.019 53.315
360627 95.574 -69.018 53.308
360628 95.600 -69.018 53.301
360629 95.626 -69.018 53.298
360630 95.652 -69.017 53.296
360631 95.677 -69.017 53.294
360632 95.703 -69.016 53.292
360633 95.729 -69.016 53.287
360634 95.754 -69.015 53.282
360635 95.780 -69.015 53.277
360636 95.806 -69.015 53.270
360637 95.831 -69.015 53.259
360638 95.857 -69.014 53.242
360639 95.883 -69.013 53.222
360640 95.908 -69.013 53.204
360641 95.934 -69.012 53.195
360642 95.960 -69.012 53.188
360643 95.985 -69.012 53.183
360644 96.011 -69.012 53.175
360645 96.037 -69.011 53.162
360646 96.062 -69.011 53.145
360647 96.089 -69.008 53.132
360648 96.116 -69.002 53.122
360649 96.144 -68.997 53.114
360650 96.172 -69.008 53.102
360651 96.206 -69.014 53.083
360652 96.241 -69.019 53.058
360653 96.275 -69.024 53.030
360654 96.310 -69.031 53.000
360655 96.345 -69.036 52.974
360656 96.388 -69.051 52.955
360657 96.432 -69.068 52.941
360658 96.468 -69.083 52.925
360659 96.501 -69.097 52.904
360660 96.527 -69.107 52.876
360661 96.540 -69.111 52.842
360662 96.572 -69.120 52.803
360663 96.604 -69.130 52.762
360664 96.636 -69.140 52.723
360665 96.668 -69.149 52.684
360666 96.700 -69.159 52.645
360667 96.732 -69.168 52.606
360668 96.764 -69.178 52.568
360669 96.796 -69.187 52.532
360670 96.828 -69.197 52.495
360671 96.860 -69.206 52.459
360672 96.892 -69.215 52.428
360673 96.925 -69.225 52.397
360674 96.955 -69.233 52.366
360675 96.984 -69.240 52.334
360676 97.013 -69.247 52.302
360677 97.042 -69.254 52.267
360678 97.072 -69.261 52.232
360679 97.101 -69.268 52.199
360680 97.129 -69.276 52.166
360681 97.157 -69.284 52.132
360682 97.185 -69.293 52.100
360683 97.213 -69.301 52.066
360684 97.241 -69.309 52.032
360685 97.271 -69.316 51.999
360686 97.302 -69.323 51.968
360687 97.335 -69.328 51.937
360688 97.370 -69.333 51.907
360689 97.404 -69.337 51.878
360690 97.439 -69.341 51.850
360691 97.474 -69.345 51.824
360692 97.509 -69.350 51.800
360693 97.543 -69.353 51.779
360694 97.578 -69.358 51.761
360695 97.612 -69.362 51.744
360696 97.652 -69.366 51.726
360697 97.696 -69.369 51.707
360698 97.740 -69.372 51.688
360699 97.783 -69.375 51.669
360700 97.811 -69.381 51.649
360701 97.840 -69.387 51.627
360702 97.869 -69.393 51.601
360703 97.898 -69.399 51.566
360704 97.927 -69.405 51.527
360705 97.952 -69.411 51.486
360706 97.975 -69.417 51.442
360707 98.016 -69.425 51.396
360708 98.063 -69.434 51.349
360709 98.110 -69.442 51.299
360710 98.155 -69.447 51.246
360711 98.199 -69.450 51.193
360712 98.243 -69.454 51.140
360713 98.287 -69.457 51.086
360714 98.331 -69.460 51.032
360715 98.375 -69.464 50.977
360716 98.410 -69.470 50.918
360717 98.441 -69.477 50.860
360718 98.472 -69.484 50.801
360719 98.502 -69.491 50.743
360720 98.533 -69.497 50.684
360721 98.564 -69.504 50.625
360722 98.594 -69.512 50.566
360723 98.626 -69.519 50.507
360724 98.656 -69.526 50.448
360725 98.695 -69.538 50.390
360726 98.740 -69.555 50.334
360727 98.784 -69.572 50.279
360728 98.829 -69.589 50.224
360729 98.873 -69.606 50.170
360730 98.918 -69.623 50.115
360731 98.960 -69.637 50.060
360732 98.983 -69.649 50.005
360733 99.006 -69.661 49.950
360734 99.029 -69.673 49.896
360735 99.052 -69.685 49.841
360736 99.075 -69.696 49.787
360737 99.098 -69.708 49.732
360738 99.121 -69.720 49.678
360739 99.144 -69.732 49.625
360740 99.167 -69.744 49.573
360741 99.190 -69.755 49.521
360742 99.213 -69.767 49.472
360743 99.236 -69.780 49.422
360744 99.269 -69.789 49.373
360745 99.305 -69.795 49.325
360746 99.346 -69.800 49.277
360747 99.387 -69.805 49.230
360748 99.434 -69.814 49.184
360749 99.480 -69.823 49.138
360750 99.526 -69.833 49.091
360751 99.572 -69.842 49.046
360752 99.618 -69.852 49.000
360753 99.664 -69.861 48.955
360754 99.710 -69.870 48.911
360755 99.756 -69.880 48.869
360756 99.804 -69.885 48.827
360757 99.856 -69.882 48.786
360758 99.909 -69.877 48.746
360759 99.962 -69.873 48.706
360760 100.014 -69.867 48.670
360761 100.059 -69.870 48.635
360762 100.084 -69.899 48.602
360763 100.110 -69.929 48.569
360764 100.136 -69.958 48.537
360765 100.161 -69.987 48.506
360766 100.187 -70.017 48.476
360767 100.212 -70.046 48.445
360768 100.242 -70.044 48.418
360769 100.273 -70.038 48.391
360770 100.312 -70.037 48.364
360771 100.360 -70.045 48.339
360772 100.408 -70.052 48.315
360773 100.457 -70.060 48.293
360774 100.505 -70.068 48.271
360775 100.553 -70.075 48.247
360776 100.602 -70.083 48.222
360777 100.650 -70.090 48.195
360778 100.698 -70.097 48.166
360779 100.747 -70.105 48.134
360780 100.795 -70.112 48.101
360781 100.840 -70.121 48.069
360782 100.884 -70.132 48.033
360783 100.933 -70.145 47.995
360784 100.960 -70.159 47.957
360785 100.975 -70.174 47.920
360786 101.000 -70.185 47.882
360787 101.026 -70.196 47.845
360788 101.052 -70.207 47.807
360789 101.077 -70.218 47.769
360790 101.103 -70.229 47.730
360791 101.129 -70.240 47.690
360792 101.155 -70.251 47.648
360793 101.180 -70.262 47.604
360794 101.205 -70.273 47.561
360795 101.231 -70.284 47.515
360796 101.251 -70.296 47.468
360797 101.270 -70.308 47.422
360798 101.306 -70.312 47.375
360799 101.344 -70.316 47.325
360800 101.380 -70.320 47.273
360801 101.416 -70.324 47.221
360802 101.453 -70.328 47.171
360803 101.489 -70.331 47.123
360804 101.526 -70.335 47.077
360805 101.562 -70.339 47.034
360806 101.598 -70.344 46.991
360807 101.635 -70.347 46.949
360808 101.671 -70.351 46.907
360809 101.707 -70.355 46.866
360810 101.744 -70.359 46.825
360811 101.780 -70.362 46.784
360812 101.817 -70.366 46.743
360813 101.853 -70.370 46.703
360814 101.889 -70.374 46.662
360815 101.925 -70.378 46.622
360816 101.962 -70.381 46.583
361300 94.347 -69.173 53.465
361301 94.382 -69.174 53.452
361302 94.417 -69.175 53.441
361303 94.452 -69.175 53.429
361304 94.487 -69.176 53.417
361305 94.522 -69.177 53.406
361306 94.557 -69.178 53.395
361307 94.592 -69.178 53.387
361308 94.626 -69.179 53.379
361309 94.679 -69.148 53.372
361310 94.732 -69.115 53.367
361311 94.784 -69.081 53.364
361312 94.837 -69.065 53.358
361313 94.890 -69.072 53.351
361314 94.943 -69.078 53.345
361315 94.995 -69.085 53.339
361316 95.056 -69.073 53.333
361317 95.117 -69.057 53.329
361318 95.179 -69.042 53.326
361319 95.241 -69.026 53.325
361320 95.302 -69.010 53.326
361321 95.364 -68.994 53.326
361322 95.423 -68.980 53.326
361323 95.449 -68.979 53.326
361324 95.475 -68.979 53.327
361325 95.500 -68.978 53.323
361326 95.526 -68.978 53.318
361327 95.552 -68.977 53.312
361328 95.578 -68.977 53.307
361329 95.603 -68.977 53.302
361330 95.629 -68.976 53.301
361331 95.655 -68.976 53.301
361332 95.680 -68.975 53.299
361333 95.706 -68.975 53.295
361334 95.732 -68.974 53.287
361335 95.757 -68.974 53.279
361336 95.783 -68.974 53.273
361337 95.809 -68.973 53.269
361338 95.834 -68.973 53.260
361339 95.860 -68.972 53.244
361340 95.886 -68.972 53.224
361341 95.911 -68.971 53.201
361342 95.937 -68.971 53.190
361343 95.963 -68.971 53.186
361344 95.988 -68.971 53.182
361345 96.014 -68.970 53.175
361346 96.039 -68.970 53.162
361347 96.065 -68.969 53.148
361348 96.091 -68.968 53.139
361349 96.116 -68.968 53.132
361350 96.143 -68.966 53.124
361351 96.176 -68.972 53.110
361352 96.210 -68.977 53.087
361353 96.245 -68.983 53.059
361354 96.280 -68.988 53.028
361355 96.315 -68.993 52.998
361356 96.349 -68.999 52.972
361357 96.391 -69.013 52.953
361358 96.432 -69.029 52.938
361359 96.465 -69.043 52.921
361360 96.499 -69.058 52.901
361361 96.529 -69.070 52.874
361362 96.544 -69.074 52.840
361363 96.576 -69.084 52.801
361364 96.608 -69.093 52.758
361365 96.640 -69.103 52.715
361366 96.672 -69.112 52.674
361367 96.704 -69.122 52.634
361368 96.736 -69.131 52.594
361369 96.768 -69.141 52.556
361370 96.800 -69.151 52.520
361371 96.832 -69.160 52.484
361372 96.865 -69.170 52.448
361373 96.900 -69.181 52.417
361374 96.934 -69.192 52.387
361375 96.965 -69.200 52.357
361376 96.993 -69.206 52.325
361377 97.020 -69.213 52.290
361378 97.048 -69.219 52.254
361379 97.077 -69.226 52.219
361380 97.106 -69.233 52.186
361381 97.134 -69.241 52.152
361382 97.162 -69.250 52.118
361383 97.189 -69.258 52.084
361384 97.218 -69.266 52.050
361385 97.246 -69.274 52.016
361386 97.274 -69.282 51.984
361387 97.308 -69.287 51.953
361388 97.342 -69.291 51.924
361389 97.377 -69.295 51.895
361390 97.412 -69.300 51.867
361391 97.446 -69.303 51.842
361392 97.481 -69.308 51.819
361393 97.515 -69.312 51.797
361394 97.550 -69.316 51.777
361395 97.585 -69.320 51.761
361396 97.619 -69.325 51.745
361397 97.662 -69.328 51.728
361398 97.706 -69.331 51.709
361399 97.749 -69.334 51.687
361400 97.782 -69.339 51.666
361401 97.810 -69.345 51.645
361402 97.839 -69.351 51.622
361403 97.868 -69.357 51.595
361404 97.896 -69.363 51.559
361405 97.925 -69.369 51.520
361406 97.954 -69.375 51.479
361407 97.977 -69.382 51.435
361408 98.021 -69.390 51.388
361409 98.068 -69.399 51.340
361410 98.115 -69.407 51.290
361411 98.160 -69.413 51.237
361412 98.204 -69.416 51.183
361413 98.248 -69.419 51.128
361414 98.293 -69.423 51.074
361415 98.337 -69.426 51.019
361416 98.381 -69.429 50.964
361417 98.425 -69.433 50.905
361418 98.457 -69.440 50.847
361419 98.488 -69.447 50.788
361420 98.518 -69.454 50.730
361421 98.549 -69.460 50.671
361422 98.580 -69.467 50.613
361423 98.611 -69.474 50.554
361424 98.642 -69.482 50.495
361425 98.673 -69.488 50.438
361426 98.717 -69.505 50.381
361427 98.762 -69.523 50.325
361428 98.806 -69.539 50.271
361429 98.851 -69.556 50.216
361430 98.896 -69.573 50.162
361431 98.934 -69.587 50.107
361432 98.961 -69.601 50.053
361433 98.986 -69.611 49.998
361434 99.011 -69.624 49.943
361435 99.036 -69.640 49.889
361436 99.060 -69.654 49.835
361437 99.083 -69.667 49.780
361438 99.106 -69.678 49.725
361439 99.129 -69.690 49.672
361440 99.152 -69.702 49.620
361441 99.175 -69.714 49.568
361442 99.198 -69.726 49.517
361443 99.221 -69.737 49.468
361444 99.244 -69.749 49.419
361445 99.271 -69.760 49.370
361446 99.307 -69.767 49.322
361447 99.348 -69.771 49.274
361448 99.394 -69.780 49.227
361449 99.440 -69.790 49.181
361450 99.486 -69.799 49.135
361451 99.532 -69.808 49.088
361452 99.578 -69.818 49.042
361453 99.624 -69.827 48.997
361454 99.670 -69.837 48.951
361455 99.716 -69.846 48.906
361456 99.762 -69.855 48.864
361457 99.808 -69.862 48.822
361458 99.854 -69.875 48.780
361459 99.905 -69.876 48.739
361460 99.956 -69.877 48.699
361461 100.006 -69.878 48.661
361462 100.054 -69.885 48.626
361463 100.102 -69.892 48.591
361464 100.125 -69.895 48.558
361465 100.145 -69.912 48.526
361466 100.171 -69.941 48.494
361467 100.197 -69.971 48.463
361468 100.222 -70.000 48.434
361469 100.250 -70.017 48.406
361470 100.280 -70.010 48.379
361471 100.324 -70.013 48.351
361472 100.372 -70.021 48.324
361473 100.420 -70.028 48.299
361474 100.469 -70.036 48.275
361475 100.517 -70.043 48.251
361476 100.565 -70.051 48.226
361477 100.614 -70.058 48.201
361478 100.662 -70.066 48.174
361479 100.710 -70.074 48.145
361480 100.759 -70.080 48.113
361481 100.807 -70.088 48.080
361482 100.855 -70.096 48.048
361483 100.902 -70.104 48.011
361484 100.948 -70.115 47.973
361485 100.971 -70.130 47.936
361486 100.997 -70.141 47.898
361487 101.022 -70.152 47.861
361488 101.048 -70.163 47.824
361489 101.073 -70.174 47.786
361490 101.099 -70.185 47.748
361491 101.120 -70.197 47.709
361492 101.139 -70.209 47.668
361493 101.158 -70.221 47.626
361494 101.177 -70.233 47.582
361495 101.196 -70.244 47.537
361496 101.214 -70.256 47.490
361497 101.241 -70.263 47.444
361498 101.276 -70.264 47.397
361499 101.316 -70.268 47.351
361500 101.354 -70.274 47.301
361501 101.393 -70.279 47.250
361502 101.431 -70.284 47.200
361503 101.467 -70.287 47.151
361504 101.504 -70.291 47.105
361505 101.540 -70.296 47.060
361506 101.576 -70.300 47.017
361507 101.613 -70.303 46.976
361508 101.650 -70.307 46.934
361509 101.686 -70.311 46.893
361510 101.722 -70.315 46.852
361511 101.758 -70.319 46.812
361512 101.795 -70.322 46.772
361513 101.831 -70.326 46.731
361514 101.868 -70.330 46.692
361515 101.904 -70.334 46.652
361516 101.940 -70.338 46.613
362001 94.341 -69.150 53.472
362002 94.377 -69.151 53.461
362003 94.411 -69.152 53.449
362004 94.446 -69.153 53.438
362005 94.481 -69.154 53.427
362006 94.516 -69.155 53.416
362007 94.551 -69.156 53.406
362008 94.603 -69.125 53.397
362009 94.655 -69.091 53.390
362010 94.708 -69.059 53.383
362011 94.761 -69.025 53.378
362012 94.815 -68.992 53.374
362013 94.867 -68.965 53.369
362014 94.920 -68.971 53.363
362015 94.973 -68.978 53.357
362016 95.026 -68.984 53.351
362017 95.078 -68.991 53.345
362018 95.131 -68.998 53.340
362019 95.186 -68.998 53.335
362020 95.248 -68.983 53.332
362021 95.309 -68.967 53.331
362022 95.371 -68.951 53.329
362023 95.426 -68.938 53.328
362024 95.452 -68.938 53.326
362025 95.478 -68.937 53.326
362026 95.503 -68.936 53.322
362027 95.529 -68.936 53.318
362028 95.555 -68.936 53.315
362029 95.581 -68.936 53.313
362030 95.606 -68.935 53.311
362031 95.632 -68.935 53.309
362032 95.658 -68.934 53.307
362033 95.683 -68.934 53.302
362034 95.709 -68.933 53.294
362035 95.735 -68.933 53.286
362036 95.760 -68.933 53.279
362037 95.786 -68.932 53.275
362038 95.812 -68.932 53.272
362039 95.837 -68.931 53.262
362040 95.863 -68.931 53.246
362041 95.888 -68.930 53.225
362042 95.914 -68.931 53.201
362043 95.940 -68.932 53.190
362044 95.966 -68.933 53.185
362045 95.992 -68.933 53.181
362046 96.019 -68.935 53.174
362047 96.044 -68.936 53.162
362048 96.071 -68.936 53.149
362049 96.097 -68.937 53.142
362050 96.122 -68.938 53.139
362051 96.144 -68.936 53.133
362052 96.179 -68.940 53.117
362053 96.215 -68.943 53.090
362054 96.250 -68.946 53.059
362055 96.285 -68.952 53.026
362056 96.320 -68.957 52.994
362057 96.355 -68.962 52.970
362058 96.395 -68.975 52.951
362059 96.430 -68.990 52.933
362060 96.463 -69.004 52.912
362061 96.496 -69.018 52.891
362062 96.529 -69.032 52.866
362063 96.548 -69.038 52.836
362064 96.580 -69.047 52.799
362065 96.612 -69.057 52.754
362066 96.644 -69.067 52.708
362067 96.676 -69.076 52.663
362068 96.708 -69.086 52.620
362069 96.740 -69.095 52.581
362070 96.773 -69.106 52.543
362071 96.808 -69.116 52.507
362072 96.842 -69.128 52.473
362073 96.877 -69.138 52.439
362074 96.911 -69.150 52.408
362075 96.946 -69.160 52.377
362076 96.977 -69.168 52.346
362077 97.005 -69.175 52.313
362078 97.032 -69.181 52.278
362079 97.060 -69.187 52.240
362080 97.088 -69.193 52.205
362081 97.115 -69.198 52.171
362082 97.142 -69.203 52.138
362083 97.171 -69.212 52.103
362084 97.199 -69.219 52.069
362085 97.227 -69.228 52.036
362086 97.255 -69.236 52.003
362087 97.285 -69.243 51.970
362088 97.316 -69.249 51.939
362089 97.349 -69.253 51.911
362090 97.384 -69.258 51.884
362091 97.419 -69.262 51.861
362092 97.453 -69.266 51.838
362093 97.488 -69.270 51.817
362094 97.522 -69.275 51.798
362095 97.557 -69.278 51.782
362096 97.592 -69.283 51.767
362097 97.628 -69.287 51.753
362098 97.671 -69.290 51.737
362099 97.715 -69.293 51.715
362100 97.751 -69.297 51.693
362101 97.780 -69.303 51.671
362102 97.809 -69.309 51.648
362103 97.838 -69.316 51.622
362104 97.866 -69.322 51.592
362105 97.895 -69.328 51.554
362106 97.924 -69.334 51.514
362107 97.953 -69.340 51.471
362108 97.979 -69.346 51.427
362109 98.026 -69.355 51.379
362110 98.073 -69.363 51.332
362111 98.120 -69.372 51.281
362112 98.166 -69.378 51.228
362113 98.210 -69.382 51.172
362114 98.254 -69.385 51.118
362115 98.298 -69.388 51.063
362116 98.342 -69.392 51.007
362117 98.386 -69.395 50.951
362118 98.430 -69.398 50.893
362119 98.473 -69.402 50.834
362120 98.504 -69.410 50.776
362121 98.534 -69.416 50.717
362122 98.566 -69.423 50.658
362123 98.596 -69.430 50.601
362124 98.627 -69.437 50.543
362125 98.658 -69.444 50.486
362126 98.695 -69.456 50.429
362127 98.740 -69.472 50.374
362128 98.784 -69.489 50.318
362129 98.829 -69.507 50.264
362130 98.873 -69.523 50.209
362131 98.904 -69.533 50.154
362132 98.926 -69.540 50.099
362133 98.955 -69.558 50.045
362134 98.983 -69.573 49.990
362135 99.008 -69.583 49.936
362136 99.032 -69.593 49.881
362137 99.058 -69.603 49.827
362138 99.083 -69.618 49.772
362139 99.108 -69.635 49.719
362140 99.133 -69.651 49.667
362141 99.159 -69.668 49.615
362142 99.184 -69.684 49.564
362143 99.207 -69.695 49.514
362144 99.230 -69.707 49.465
362145 99.253 -69.719 49.416
362146 99.276 -69.731 49.367
362147 99.310 -69.739 49.319
362148 99.354 -69.746 49.271
362149 99.400 -69.756 49.223
362150 99.446 -69.765 49.177
362151 99.492 -69.775 49.132
362152 99.538 -69.782 49.086
362153 99.583 -69.785 49.039
362154 99.629 -69.788 48.993
362155 99.675 -69.792 48.947
362156 99.724 -69.804 48.901
362157 99.773 -69.817 48.858
362158 99.814 -69.843 48.815
362159 99.854 -69.870 48.773
362160 99.897 -69.888 48.731
362161 99.947 -69.889 48.691
362162 99.997 -69.892 48.652
362163 100.045 -69.899 48.616
362164 100.092 -69.907 48.581
362165 100.140 -69.914 48.547
362166 100.174 -69.919 48.514
362167 100.190 -69.921 48.482
362168 100.207 -69.924 48.450
362169 100.232 -69.954 48.419
362170 100.258 -69.983 48.392
362171 100.287 -69.982 48.366
362172 100.336 -69.990 48.339
362173 100.384 -69.997 48.311
362174 100.432 -70.004 48.284
362175 100.481 -70.011 48.257
362176 100.529 -70.019 48.230
362177 100.577 -70.027 48.203
362178 100.626 -70.034 48.175
362179 100.674 -70.042 48.147
362180 100.722 -70.049 48.117
362181 100.771 -70.057 48.086
362182 100.819 -70.065 48.054
362183 100.867 -70.071 48.023
362184 100.915 -70.079 47.988
362185 100.964 -70.087 47.951
362186 100.988 -70.098 47.913
362187 101.007 -70.110 47.874
362188 101.027 -70.121 47.837
362189 101.046 -70.134 47.801
362190 101.064 -70.146 47.764
362191 101.084 -70.157 47.725
362192 101.103 -70.169 47.685
362193 101.122 -70.181 47.644
362194 101.140 -70.193 47.601
362195 101.159 -70.205 47.556
362196 101.183 -70.213 47.510
362197 101.212 -70.218 47.464
362198 101.245 -70.220 47.417
362199 101.287 -70.221 47.370
362200 101.326 -70.226 47.323
362201 101.364 -70.231 47.275
362202 101.403 -70.237 47.226
362203 101.442 -70.242 47.178
362204 101.481 -70.247 47.132
362205 101.519 -70.252 47.087
362206 101.555 -70.256 47.043
362207 101.591 -70.259 47.001
362208 101.627 -70.263 46.960
362209 101.664 -70.267 46.920
362210 101.700 -70.271 46.879
362211 101.737 -70.275 46.839
362212 101.773 -70.278 46.799
362213 101.809 -70.282 46.760
362214 101.846 -70.287 46.720
362215 101.882 -70.290 46.682
362216 101.918 -70.294 46.643
362217 101.955 -70.298 46.603
362703 94.371 -69.129 53.467
362704 94.406 -69.130 53.456
362705 94.440 -69.131 53.445
362706 94.476 -69.131 53.434
362707 94.527 -69.102 53.424
362708 94.580 -69.068 53.415
362709 94.633 -69.035 53.406
362710 94.686 -69.002 53.399
362711 94.739 -68.969 53.392
362712 94.791 -68.936 53.386
362713 94.844 -68.902 53.380
362714 94.897 -68.870 53.376
362715 94.950 -68.871 53.371
362716 95.003 -68.877 53.367
362717 95.056 -68.884 53.365
362718 95.108 -68.890 53.358
362719 95.161 -68.897 53.351
362720 95.214 -68.903 53.345
362721 95.267 -68.910 53.340
362722 95.320 -68.917 53.338
362723 95.378 -68.908 53.334
362724 95.429 -68.897 53.331
362725 95.455 -68.896 53.327
362726 95.481 -68.896 53.325
362727 95.507 -68.896 53.321
362728 95.532 -68.895 53.321
362729 95.558 -68.895 53.324
362730 95.584 -68.894 53.328
362731 95.609 -68.895 53.326
362732 95.635 -68.896 53.320
362733 95.662 -68.897 53.312
362734 95.687 -68.898 53.302
362735 95.713 -68.899 53.291
362736 95.740 -68.899 53.285
362737 95.766 -68.901 53.281
362738 95.792 -68.902 53.280
362739 95.818 -68.902 53.275
362740 95.844 -68.903 53.262
362741 95.870 -68.905 53.245
362742 95.896 -68.905 53.224
362743 95.922 -68.906 53.201
362744 95.948 -68.907 53.189
362745 95.974 -68.908 53.183
362746 96.000 -68.909 53.179
362747 96.026 -68.910 53.173
362748 96.052 -68.911 53.161
362749 96.078 -68.911 53.148
362750 96.104 -68.913 53.144
362751 96.127 -68.912 53.144
362752 96.147 -68.908 53.138
362753 96.182 -68.911 53.121
362754 96.218 -68.915 53.093
362755 96.253 -68.918 53.059
362756 96.288 -68.921 53.023
362757 96.324 -68.924 52.990
362758 96.359 -68.927 52.967
362759 96.393 -68.936 52.947
362760 96.427 -68.950 52.927
362761 96.460 -68.965 52.903
362762 96.493 -68.979 52.882
362763 96.527 -68.993 52.859
362764 96.552 -69.002 52.832
362765 96.584 -69.011 52.796
362766 96.616 -69.021 52.751
362767 96.648 -69.031 52.703
362768 96.682 -69.040 52.656
362769 96.716 -69.052 52.610
362770 96.751 -69.062 52.569
362771 96.785 -69.074 52.531
362772 96.819 -69.084 52.496
362773 96.854 -69.095 52.463
362774 96.888 -69.106 52.429
362775 96.923 -69.117 52.397
362776 96.957 -69.128 52.365
362777 96.990 -69.137 52.333
362778 97.017 -69.143 52.299
362779 97.045 -69.150 52.264
362780 97.073 -69.156 52.226
362781 97.100 -69.162 52.190
362782 97.127 -69.165 52.157
362783 97.153 -69.168 52.123
362784 97.180 -69.173 52.088
362785 97.209 -69.181 52.055
362786 97.237 -69.189 52.022
362787 97.265 -69.197 51.989
362788 97.296 -69.203 51.957
362789 97.327 -69.209 51.928
362790 97.359 -69.215 51.902
362791 97.391 -69.220 51.878
362792 97.426 -69.225 51.859
362793 97.460 -69.228 51.840
362794 97.495 -69.233 51.823
362795 97.529 -69.237 51.809
362796 97.564 -69.241 51.796
362797 97.599 -69.245 51.783
362798 97.637 -69.249 51.768
362799 97.680 -69.252 51.750
362800 97.721 -69.256 51.728
362801 97.750 -69.262 51.704
362802 97.779 -69.268 51.679
362803 97.808 -69.274 51.651
362804 97.837 -69.280 51.621
362805 97.865 -69.286 51.588
362806 97.894 -69.292 51.548
362807 97.923 -69.298 51.506
362808 97.951 -69.304 51.462
362809 97.984 -69.311 51.417
362810 98.031 -69.319 51.369
362811 98.078 -69.328 51.321
362812 98.125 -69.337 51.271
362813 98.171 -69.344 51.218
362814 98.215 -69.347 51.162
362815 98.260 -69.351 51.107
362816 98.303 -69.354 51.051
362817 98.348 -69.357 50.995
362818 98.391 -69.361 50.939
362819 98.436 -69.364 50.880
362820 98.480 -69.368 50.821
362821 98.520 -69.372 50.762
362822 98.551 -69.379 50.705
362823 98.582 -69.386 50.648
362824 98.613 -69.393 50.592
362825 98.643 -69.400 50.536
362826 98.674 -69.407 50.480
362827 98.717 -69.422 50.424
362828 98.762 -69.440 50.369
362829 98.806 -69.457 50.314
362830 98.851 -69.473 50.258
362831 98.873 -69.480 50.203
362832 98.896 -69.486 50.147
362833 98.918 -69.494 50.091
362834 98.949 -69.516 50.037
362835 98.979 -69.535 49.983
362836 99.004 -69.545 49.929
362837 99.029 -69.555 49.875
362838 99.055 -69.565 49.821
362839 99.079 -69.575 49.768
362840 99.104 -69.585 49.715
362841 99.130 -69.596 49.663
362842 99.155 -69.613 49.612
362843 99.180 -69.629 49.561
362844 99.205 -69.646 49.511
362845 99.231 -69.662 49.462
362846 99.256 -69.679 49.413
362847 99.281 -69.695 49.364
362848 99.314 -69.707 49.316
362849 99.359 -69.710 49.267
362850 99.405 -69.713 49.220
362851 99.451 -69.716 49.173
362852 99.497 -69.719 49.127
362853 99.543 -69.721 49.082
362854 99.591 -69.733 49.035
362855 99.641 -69.745 48.988
362856 99.690 -69.758 48.942
362857 99.737 -69.773 48.896
362858 99.777 -69.798 48.852
362859 99.818 -69.825 48.809
362860 99.858 -69.852 48.767
362861 99.898 -69.877 48.725
362862 99.940 -69.900 48.684
362863 99.988 -69.907 48.644
362864 100.035 -69.914 48.607
362865 100.083 -69.921 48.571
362866 100.131 -69.929 48.536
362867 100.178 -69.936 48.502
362868 100.223 -69.942 48.469
362869 100.240 -69.945 48.436
362870 100.256 -69.947 48.405
362871 100.272 -69.949 48.378
362872 100.299 -69.958 48.352
362873 100.348 -69.965 48.325
362874 100.394 -69.972 48.298
362875 100.439 -69.980 48.271
362876 100.485 -69.987 48.241
362877 100.531 -69.994 48.212
362878 100.577 -70.002 48.182
362879 100.623 -70.008 48.152
362880 100.668 -70.016 48.122
362881 100.714 -70.024 48.092
362882 100.760 -70.030 48.060
362883 100.806 -70.038 48.029
362884 100.852 -70.045 47.998
362885 100.898 -70.052 47.963
362886 100.944 -70.060 47.927
362887 100.971 -70.070 47.889
362888 100.990 -70.082 47.851
362889 101.009 -70.094 47.813
362890 101.029 -70.105 47.777
362891 101.047 -70.118 47.741
362892 101.066 -70.130 47.702
362893 101.086 -70.141 47.661
362894 101.105 -70.153 47.619
362895 101.125 -70.164 47.574
362896 101.154 -70.168 47.529
362897 101.183 -70.173 47.483
362898 101.214 -70.176 47.437
362899 101.257 -70.174 47.388
362900 101.297 -70.178 47.340
362901 101.336 -70.184 47.294
362902 101.375 -70.189 47.248
362903 101.413 -70.194 47.202
362904 101.452 -70.199 47.157
362905 101.491 -70.204 47.113
362906 101.530 -70.209 47.068
362907 101.568 -70.215 47.027
362908 101.606 -70.219 46.986
362909 101.642 -70.223 46.945
362910 101.678 -70.227 46.905
362911 101.715 -70.231 46.866
362912 101.751 -70.234 46.826
362913 101.787 -70.239 46.786
362914 101.824 -70.243 46.747
362915 101.861 -70.246 46.709
362916 101.897 -70.250 46.670
362917 101.933 -70.254 46.632
363404 94.365 -69.106 53.472
363405 94.400 -69.108 53.462
363406 94.451 -69.078 53.451
363407 94.504 -69.046 53.441
363408 94.557 -69.012 53.432
363409 94.610 -68.979 53.423
363410 94.663 -68.946 53.414
363411 94.715 -68.913 53.406
363412 94.768 -68.880 53.399
363413 94.821 -68.846 53.393
363414 94.875 -68.813 53.387
363415 94.927 -68.780 53.382
363416 94.980 -68.770 53.379
363417 95.033 -68.777 53.377
363418 95.086 -68.783 53.377
363419 95.138 -68.790 53.370
363420 95.191 -68.796 53.364
363421 95.244 -68.803 53.357
363422 95.286 -68.819 53.351
363423 95.328 -68.836 53.347
363424 95.370 -68.853 53.344
363425 95.428 -68.861 53.341
363426 95.461 -68.864 53.336
363427 95.487 -68.866 53.333
363428 95.513 -68.867 53.332
363429 95.539 -68.868 53.333
363430 95.565 -68.868 53.336
363431 95.591 -68.870 53.336
363432 95.617 -68.871 53.330
363433 95.643 -68.871 53.321
363434 95.669 -68.872 53.309
363435 95.695 -68.873 53.296
363436 95.721 -68.874 53.290
363437 95.747 -68.875 53.289
363438 95.773 -68.876 53.287
363439 95.799 -68.877 53.282
363440 95.825 -68.878 53.274
363441 95.851 -68.879 53.257
363442 95.877 -68.880 53.239
363443 95.903 -68.880 53.219
363444 95.930 -68.881 53.194
363445 95.955 -68.883 53.184
363446 95.981 -68.883 53.180
363447 96.008 -68.884 53.177
363448 96.034 -68.885 53.171
363449 96.059 -68.886 53.158
363450 96.086 -68.887 53.146
363451 96.111 -68.887 53.142
363452 96.128 -68.883 53.144
363453 96.150 -68.880 53.139
363454 96.185 -68.883 53.122
363455 96.220 -68.886 53.094
363456 96.256 -68.889 53.059
363457 96.291 -68.893 53.021
363458 96.326 -68.896 52.987
363459 96.361 -68.900 52.964
363460 96.394 -68.908 52.943
363461 96.427 -68.919 52.921
363462 96.459 -68.930 52.895
363463 96.491 -68.941 52.873
363464 96.524 -68.953 52.852
363465 96.556 -68.965 52.826
363466 96.590 -68.976 52.792
363467 96.624 -68.987 52.748
363468 96.659 -68.998 52.699
363469 96.693 -69.009 52.649
363470 96.728 -69.019 52.602
363471 96.762 -69.031 52.559
363472 96.797 -69.041 52.521
363473 96.831 -69.052 52.487
363474 96.866 -69.063 52.454
363475 96.900 -69.074 52.419
363476 96.935 -69.085 52.385
363477 96.969 -69.096 52.352
363478 97.002 -69.106 52.319
363479 97.029 -69.112 52.285
363480 97.057 -69.118 52.249
363481 97.085 -69.125 52.211
363482 97.112 -69.131 52.176
363483 97.139 -69.133 52.142
363484 97.165 -69.135 52.108
363485 97.191 -69.137 52.074
363486 97.219 -69.143 52.042
363487 97.247 -69.150 52.010
363488 97.275 -69.159 51.976
363489 97.306 -69.165 51.947
363490 97.338 -69.170 51.920
363491 97.369 -69.175 51.896
363492 97.400 -69.181 51.874
363493 97.433 -69.187 51.860
363494 97.467 -69.191 51.845
363495 97.502 -69.195 51.833
363496 97.536 -69.199 51.823
363497 97.571 -69.203 51.813
363498 97.606 -69.207 51.801
363499 97.646 -69.211 51.785
363500 97.690 -69.214 51.765
363501 97.720 -69.220 51.742
363502 97.749 -69.226 51.716
363503 97.778 -69.232 51.687
363504 97.806 -69.238 51.655
363505 97.835 -69.244 51.621
363506 97.864 -69.250 51.585
363507 97.892 -69.256 51.542
363508 97.923 -69.262 51.499
363509 97.957 -69.269 51.454
363510 97.995 -69.276 51.407
363511 98.035 -69.284 51.358
363512 98.083 -69.293 51.310
363513 98.130 -69.302 51.260
363514 98.176 -69.310 51.207
363515 98.221 -69.313 51.151
363516 98.265 -69.316 51.096
363517 98.309 -69.320 51.040
363518 98.353 -69.323 50.984
363519 98.397 -69.327 50.928
363520 98.441 -69.330 50.869
363521 98.485 -69.333 50.810
363522 98.529 -69.337 50.752
363523 98.567 -69.341 50.695
363524 98.598 -69.348 50.641
363525 98.629 -69.356 50.586
363526 98.659 -69.363 50.531
363527 98.695 -69.373 50.475
363528 98.740 -69.390 50.420
363529 98.784 -69.407 50.364
363530 98.821 -69.420 50.309
363531 98.843 -69.426 50.253
363532 98.865 -69.433 50.197
363533 98.887 -69.439 50.140
363534 98.912 -69.451 50.084
363535 98.943 -69.472 50.030
363536 98.975 -69.494 49.977
363537 99.001 -69.508 49.922
363538 99.026 -69.518 49.869
363539 99.051 -69.527 49.817
363540 99.076 -69.537 49.764
363541 99.101 -69.547 49.712
363542 99.126 -69.557 49.660
363543 99.151 -69.566 49.610
363544 99.177 -69.576 49.559
363545 99.202 -69.591 49.509
363546 99.227 -69.607 49.460
363547 99.252 -69.623 49.411
363548 99.277 -69.640 49.362
363549 99.319 -69.646 49.313
363550 99.364 -69.649 49.265
363551 99.410 -69.652 49.216
363552 99.458 -69.661 49.169
363553 99.507 -69.674 49.123
363554 99.557 -69.686 49.077
363555 99.606 -69.698 49.031
363556 99.655 -69.711 48.985
363557 99.701 -69.728 48.938
363558 99.742 -69.754 48.892
363559 99.782 -69.780 48.848
363560 99.822 -69.807 48.804
363561 99.862 -69.833 48.761
363562 99.907 -69.853 48.718
363563 99.956 -69.867 48.677
363564 100.000 -69.883 48.636
363565 100.043 -69.898 48.598
363566 100.086 -69.913 48.561
363567 100.130 -69.928 48.525
363568 100.173 -69.943 48.489
363569 100.216 -69.958 48.455
363570 100.264 -69.964 48.422
363571 100.289 -69.968 48.390
363572 100.288 -69.946 48.363
363573 100.307 -69.938 48.336
363574 100.343 -69.942 48.310
363575 100.380 -69.947 48.285
363576 100.418 -69.952 48.258
363577 100.464 -69.960 48.227
363578 100.509 -69.968 48.197
363579 100.555 -69.974 48.165
363580 100.601 -69.982 48.132
363581 100.647 -69.989 48.100
363582 100.693 -69.996 48.068
363583 100.739 -70.004 48.037
363584 100.784 -70.011 48.006
363585 100.830 -70.018 47.971
363586 100.876 -70.025 47.937
363587 100.922 -70.033 47.901
363588 100.954 -70.043 47.865
363589 100.973 -70.054 47.827
363590 100.992 -70.066 47.789
363591 101.011 -70.078 47.753
363592 101.030 -70.090 47.716
363593 101.049 -70.102 47.678
363594 101.068 -70.113 47.636
363595 101.096 -70.119 47.591
363596 101.125 -70.124 47.546
363597 101.155 -70.128 47.501
363598 101.184 -70.132 47.457
363599 101.223 -70.130 47.410
363600 101.268 -70.130 47.362
363601 101.307 -70.136 47.316
363602 101.346 -70.141 47.270
363603 101.385 -70.146 47.225
363604 101.423 -70.151 47.181
363605 101.462 -70.156 47.138
363606 101.501 -70.162 47.095
363607 101.539 -70.167 47.051
363608 101.578 -70.172 47.011
363609 101.617 -70.178 46.970
363610 101.656 -70.183 46.930
363611 101.693 -70.187 46.890
363612 101.730 -70.191 46.851
363613 101.766 -70.195 46.813
363614 101.802 -70.199 46.774
363615 101.838 -70.203 46.735
363616 101.875 -70.206 46.698
363617 101.911 -70.210 46.659
364105 94.375 -69.056 53.477
364106 94.428 -69.022 53.467
364107 94.481 -68.990 53.458
364108 94.534 -68.956 53.448
364109 94.586 -68.923 53.438
364110 94.639 -68.889 53.429
364111 94.693 -68.857 53.420
364112 94.746 -68.824 53.413
364113 94.799 -68.790 53.406
364114 94.851 -68.757 53.399
364115 94.904 -68.723 53.393
364116 94.957 -68.692 53.388
364117 94.991 -68.696 53.386
364118 95.027 -68.711 53.386
364119 95.068 -68.728 53.387
364120 95.110 -68.745 53.381
364121 95.152 -68.762 53.375
364122 95.194 -68.779 53.369
364123 95.235 -68.796 53.362
364124 95.277 -68.813 53.357
364125 95.319 -68.830 53.353
364126 95.398 -68.824 53.349
364127 95.457 -68.833 53.344
364128 95.494 -68.841 53.341
364129 95.520 -68.842 53.341
364130 95.546 -68.842 53.343
364131 95.572 -68.844 53.345
364132 95.599 -68.845 53.341
364133 95.624 -68.846 53.331
364134 95.651 -68.846 53.320
364135 95.677 -68.848 53.308
364136 95.702 -68.849 53.296
364137 95.729 -68.849 53.296
364138 95.755 -68.850 53.298
364139 95.780 -68.851 53.295
364140 95.807 -68.852 53.286
364141 95.833 -68.853 53.272
364142 95.859 -68.854 53.251
364143 95.885 -68.855 53.231
364144 95.911 -68.856 53.210
364145 95.937 -68.857 53.187
364146 95.963 -68.858 53.175
364147 95.989 -68.858 53.172
364148 96.015 -68.859 53.171
364149 96.041 -68.861 53.165
364150 96.067 -68.861 53.152
364151 96.093 -68.862 53.140
364152 96.112 -68.858 53.138
364153 96.129 -68.854 53.141
364154 96.152 -68.852 53.137
364155 96.187 -68.855 53.121
364156 96.223 -68.858 53.094
364157 96.258 -68.861 53.059
364158 96.294 -68.864 53.022
364159 96.328 -68.869 52.985
364160 96.362 -68.874 52.957
364161 96.395 -68.881 52.933
364162 96.428 -68.893 52.909
364163 96.460 -68.904 52.882
364164 96.493 -68.915 52.861
364165 96.523 -68.927 52.840
364166 96.564 -68.934 52.815
364167 96.601 -68.943 52.784
364168 96.636 -68.955 52.740
364169 96.670 -68.965 52.691
364170 96.705 -68.977 52.641
364171 96.739 -68.987 52.594
364172 96.774 -68.998 52.552
364173 96.808 -69.009 52.516
364174 96.842 -69.020 52.481
364175 96.877 -69.031 52.446
364176 96.911 -69.042 52.409
364177 96.946 -69.052 52.374
364178 96.980 -69.064 52.340
364179 97.014 -69.074 52.305
364180 97.042 -69.080 52.269
364181 97.070 -69.087 52.234
364182 97.097 -69.093 52.197
364183 97.124 -69.098 52.162
364184 97.151 -69.100 52.128
364185 97.177 -69.103 52.096
364186 97.203 -69.104 52.062
364187 97.229 -69.106 52.031
364188 97.257 -69.112 52.000
364189 97.286 -69.119 51.968
364190 97.317 -69.125 51.940
364191 97.349 -69.131 51.915
364192 97.380 -69.137 51.894
364193 97.412 -69.142 51.877
364194 97.443 -69.147 51.865
364195 97.474 -69.153 51.855
364196 97.509 -69.157 51.848
364197 97.543 -69.162 51.841
364198 97.578 -69.165 51.832
364199 97.613 -69.170 51.819
364200 97.656 -69.173 51.803
364201 97.690 -69.178 51.782
364202 97.719 -69.184 51.756
364203 97.747 -69.190 51.726
364204 97.776 -69.197 51.695
364205 97.805 -69.203 51.660
364206 97.835 -69.209 51.623
364207 97.867 -69.214 51.583
364208 97.899 -69.220 51.537
364209 97.933 -69.226 51.492
364210 97.971 -69.234 51.445
364211 98.009 -69.242 51.397
364212 98.047 -69.250 51.347
364213 98.088 -69.258 51.298
364214 98.135 -69.266 51.247
364215 98.181 -69.275 51.194
364216 98.226 -69.279 51.140
364217 98.270 -69.282 51.085
364218 98.314 -69.285 51.029
364219 98.358 -69.289 50.973
364220 98.403 -69.292 50.917
364221 98.447 -69.296 50.859
364222 98.491 -69.299 50.801
364223 98.535 -69.302 50.744
364224 98.579 -69.306 50.690
364225 98.614 -69.311 50.636
364226 98.645 -69.318 50.582
364227 98.675 -69.325 50.528
364228 98.717 -69.340 50.472
364229 98.762 -69.357 50.415
364230 98.791 -69.366 50.360
364231 98.812 -69.373 50.304
364232 98.835 -69.379 50.247
364233 98.857 -69.386 50.191
364234 98.878 -69.392 50.134
364235 98.906 -69.408 50.080
364236 98.937 -69.430 50.026
364237 98.969 -69.452 49.972
364238 98.998 -69.470 49.918
364239 99.023 -69.480 49.865
364240 99.048 -69.490 49.813
364241 99.073 -69.500 49.761
364242 99.098 -69.510 49.709
364243 99.123 -69.519 49.658
364244 99.148 -69.529 49.608
364245 99.173 -69.539 49.558
364246 99.198 -69.549 49.508
364247 99.223 -69.558 49.458
364248 99.248 -69.569 49.409
364249 99.277 -69.582 49.360
364250 99.325 -69.590 49.311
364251 99.374 -69.603 49.262
364252 99.424 -69.615 49.213
364253 99.473 -69.627 49.166
364254 99.522 -69.640 49.120
364255 99.571 -69.652 49.073
364256 99.621 -69.664 49.027
364257 99.665 -69.684 48.981
364258 99.706 -69.710 48.934
364259 99.746 -69.736 48.888
364260 99.786 -69.762 48.844
364261 99.829 -69.785 48.799
364262 99.878 -69.799 48.756
364263 99.927 -69.813 48.713
364264 99.976 -69.827 48.670
364265 100.021 -69.842 48.628
364266 100.065 -69.858 48.588
364267 100.108 -69.873 48.551
364268 100.151 -69.888 48.514
364269 100.194 -69.903 48.477
364270 100.237 -69.919 48.442
364271 100.280 -69.934 48.408
364272 100.290 -69.944 48.375
364273 100.287 -69.919 48.346
364274 100.317 -69.919 48.317
364275 100.353 -69.924 48.290
364276 100.390 -69.929 48.264
364277 100.427 -69.933 48.239
364278 100.464 -69.939 48.209
364279 100.500 -69.943 48.177
364280 100.537 -69.949 48.144
364281 100.579 -69.955 48.111
364282 100.625 -69.962 48.079
364283 100.671 -69.969 48.047
364284 100.717 -69.977 48.014
364285 100.763 -69.983 47.981
364286 100.808 -69.991 47.948
364287 100.854 -69.999 47.912
364288 100.900 -70.005 47.877
364289 100.937 -70.015 47.841
364290 100.956 -70.026 47.803
364291 100.975 -70.038 47.764
364292 100.994 -70.050 47.728
364293 101.013 -70.062 47.692
364294 101.038 -70.069 47.654
364295 101.067 -70.074 47.613
364296 101.096 -70.079 47.567
364297 101.125 -70.083 47.520
364298 101.155 -70.088 47.474
364299 101.193 -70.087 47.430
364300 101.238 -70.084 47.384
364301 101.278 -70.088 47.339
364302 101.317 -70.093 47.293
364303 101.356 -70.098 47.249
364304 101.395 -70.103 47.205
364305 101.433 -70.109 47.162
364306 101.472 -70.114 47.120
364307 101.510 -70.119 47.078
364308 101.549 -70.124 47.036
364309 101.588 -70.130 46.996
364310 101.627 -70.134 46.955
364311 101.666 -70.140 46.915
364312 101.704 -70.145 46.876
364313 101.743 -70.150 46.838
364314 101.780 -70.155 46.800
364315 101.817 -70.159 46.762
364316 101.853 -70.162 46.724
364807 94.458 -68.933 53.471
364808 94.511 -68.900 53.462
364809 94.564 -68.867 53.453
364810 94.617 -68.833 53.443
364811 94.670 -68.800 53.434
364812 94.722 -68.767 53.426
364813 94.775 -68.734 53.419
364814 94.828 -68.701 53.411
364815 94.876 -68.684 53.404
364816 94.922 -68.674 53.398
364817 94.967 -68.665 53.394
364818 94.987 -68.689 53.391
364819 95.012 -68.703 53.392
364820 95.039 -68.714 53.393
364821 95.065 -68.724 53.390
364822 95.101 -68.739 53.385
364823 95.142 -68.755 53.379
364824 95.184 -68.773 53.373
364825 95.226 -68.789 53.366
364826 95.280 -68.799 53.360
364827 95.362 -68.792 53.354
364828 95.434 -68.792 53.350
364829 95.484 -68.806 53.348
364830 95.528 -68.817 53.349
364831 95.554 -68.818 53.351
364832 95.580 -68.819 53.351
364833 95.606 -68.820 53.344
364834 95.632 -68.821 53.330
364835 95.658 -68.822 53.318
364836 95.684 -68.823 53.308
364837 95.710 -68.824 53.300
364838 95.736 -68.824 53.306
364839 95.762 -68.826 53.309
364840 95.788 -68.827 53.304
364841 95.814 -68.827 53.291
364842 95.840 -68.828 53.270
364843 95.867 -68.829 53.246
364844 95.892 -68.830 53.223
364845 95.918 -68.831 53.201
364846 95.945 -68.832 53.179
364847 95.970 -68.833 53.164
364848 95.996 -68.834 53.162
364849 96.023 -68.835 53.161
364850 96.049 -68.836 53.155
364851 96.075 -68.836 53.142
364852 96.095 -68.834 53.132
364853 96.113 -68.830 53.131
364854 96.130 -68.825 53.135
364855 96.155 -68.824 53.132
364856 96.190 -68.827 53.117
364857 96.226 -68.830 53.091
364858 96.260 -68.833 53.058
364859 96.295 -68.838 53.022
364860 96.329 -68.842 52.983
364861 96.363 -68.847 52.952
364862 96.396 -68.855 52.924
364863 96.429 -68.866 52.897
364864 96.461 -68.877 52.870
364865 96.490 -68.891 52.847
364866 96.526 -68.901 52.825
364867 96.568 -68.905 52.802
364868 96.608 -68.914 52.771
364869 96.648 -68.922 52.729
364870 96.682 -68.933 52.682
364871 96.717 -68.944 52.633
364872 96.751 -68.955 52.587
364873 96.786 -68.966 52.545
364874 96.820 -68.977 52.509
364875 96.855 -68.988 52.473
364876 96.889 -68.999 52.438
364877 96.923 -69.010 52.399
364878 96.958 -69.021 52.363
364879 96.992 -69.031 52.326
364880 97.026 -69.042 52.290
364881 97.054 -69.049 52.254
364882 97.082 -69.055 52.220
364883 97.109 -69.061 52.184
364884 97.137 -69.065 52.149
364885 97.163 -69.068 52.115
364886 97.189 -69.069 52.083
364887 97.215 -69.071 52.051
364888 97.241 -69.074 52.021
364889 97.267 -69.076 51.992
364890 97.297 -69.080 51.963
364891 97.328 -69.086 51.937
364892 97.360 -69.091 51.916
364893 97.391 -69.097 51.898
364894 97.423 -69.103 51.885
364895 97.454 -69.109 51.877
364896 97.485 -69.114 51.871
364897 97.517 -69.119 51.865
364898 97.550 -69.124 51.858
364899 97.585 -69.128 51.849
364900 97.621 -69.132 51.836
364901 97.660 -69.136 51.818
364902 97.689 -69.143 51.795
364903 97.718 -69.149 51.768
364904 97.746 -69.155 51.736
364905 97.778 -69.160 51.701
364906 97.810 -69.166 51.664
364907 97.843 -69.172 51.623
364908 97.875 -69.178 51.581
364909 97.910 -69.184 51.534
364910 97.947 -69.192 51.487
364911 97.986 -69.200 51.439
364912 98.023 -69.207 51.390
364913 98.061 -69.215 51.340
364914 98.099 -69.223 51.290
364915 98.140 -69.231 51.239
364916 98.186 -69.240 51.185
364917 98.232 -69.244 51.131
364918 98.276 -69.248 51.075
364919 98.320 -69.251 51.020
364920 98.364 -69.255 50.964
364921 98.408 -69.258 50.908
364922 98.452 -69.261 50.852
364923 98.496 -69.265 50.796
364924 98.540 -69.268 50.741
364925 98.584 -69.271 50.688
364926 98.629 -69.275 50.635
364927 98.661 -69.281 50.581
364928 98.695 -69.291 50.525
364929 98.738 -69.306 50.469
364930 98.761 -69.313 50.412
364931 98.782 -69.319 50.355
364932 98.804 -69.326 50.299
364933 98.826 -69.332 50.243
364934 98.848 -69.339 50.186
364935 98.870 -69.345 50.131
364936 98.901 -69.366 50.077
364937 98.932 -69.388 50.023
364938 98.962 -69.409 49.969
364939 98.994 -69.431 49.916
364940 99.020 -69.442 49.864
364941 99.045 -69.452 49.813
364942 99.070 -69.462 49.761
364943 99.095 -69.472 49.709
364944 99.120 -69.482 49.658
364945 99.145 -69.491 49.608
364946 99.170 -69.501 49.557
364947 99.195 -69.511 49.507
364948 99.220 -69.521 49.456
364949 99.259 -69.527 49.407
364950 99.290 -69.544 49.358
364951 99.340 -69.556 49.309
364952 99.389 -69.569 49.260
364953 99.438 -69.581 49.211
364954 99.488 -69.593 49.164
364955 99.537 -69.606 49.117
364956 99.586 -69.618 49.070
364957 99.630 -69.639 49.024
364958 99.670 -69.666 48.978
364959 99.710 -69.692 48.931
364960 99.752 -69.716 48.885
364961 99.800 -69.730 48.840
364962 99.849 -69.745 48.795
364963 99.899 -69.759 48.751
364964 99.947 -69.773 48.707
364965 99.996 -69.788 48.664
364966 100.043 -69.802 48.621
364967 100.086 -69.817 48.581
364968 100.129 -69.833 48.542
364969 100.173 -69.848 48.504
364970 100.215 -69.864 48.467
364971 100.259 -69.879 48.430
364972 100.275 -69.894 48.395
364973 100.287 -69.910 48.361
364974 100.290 -69.895 48.331
364975 100.327 -69.900 48.300
364976 100.364 -69.905 48.271
364977 100.401 -69.910 48.243
364978 100.437 -69.914 48.216
364979 100.474 -69.920 48.186
364980 100.510 -69.924 48.155
364981 100.547 -69.930 48.123
364982 100.583 -69.934 48.091
364983 100.620 -69.939 48.059
364984 100.657 -69.944 48.026
364985 100.695 -69.949 47.992
364986 100.741 -69.957 47.957
364987 100.786 -69.964 47.924
364988 100.832 -69.971 47.888
364989 100.878 -69.979 47.853
364990 100.920 -69.986 47.817
364991 100.939 -69.999 47.779
364992 100.958 -70.010 47.739
364993 100.980 -70.020 47.703
364994 101.009 -70.024 47.667
364995 101.038 -70.029 47.629
364996 101.067 -70.033 47.588
364997 101.097 -70.038 47.542
364998 101.125 -70.043 47.495
364999 101.162 -70.043 47.449
365000 101.204 -70.040 47.404
365001 101.250 -70.040 47.361
365002 101.288 -70.045 47.317
365003 101.327 -70.050 47.273
365004 101.366 -70.055 47.230
365005 101.405 -70.061 47.186
365006 101.443 -70.066 47.144
365007 101.482 -70.071 47.103
365008 101.521 -70.077 47.062
365009 101.559 -70.081 47.021
365010 101.598 -70.087 46.981
365011 101.637 -70.092 46.941
365012 101.676 -70.097 46.902
365013 101.714 -70.102 46.864
365014 101.753 -70.108 46.826
365015 101.792 -70.113 46.787
365016 101.830 -70.118 46.750
365433 91.379 -69.196 54.860
365434 91.437 -69.225 54.852
365435 91.495 -69.256 54.845
365436 91.545 -69.316 54.837
365437 91.596 -69.375 54.828
365508 94.488 -68.843 53.474
365509 94.541 -68.811 53.465
365510 94.594 -68.777 53.456
365511 94.646 -68.744 53.447
365512 94.699 -68.711 53.438
365513 94.750 -68.685 53.431
365514 94.796 -68.676 53.423
365515 94.841 -68.666 53.415
365516 94.887 -68.656 53.408
365517 94.932 -68.647 53.402
365518 94.965 -68.654 53.398
365519 94.985 -68.678 53.395
365520 95.007 -68.700 53.395
365521 95.033 -68.710 53.395
365522 95.060 -68.720 53.394
365523 95.086 -68.731 53.391
365524 95.113 -68.742 53.385
365525 95.139 -68.752 53.379
365526 95.175 -68.766 53.373
365527 95.245 -68.767 53.365
365528 95.326 -68.759 53.357
365529 95.408 -68.752 53.350
365530 95.462 -68.764 53.348
365531 95.513 -68.780 53.348
365532 95.561 -68.793 53.349
365533 95.588 -68.794 53.348
365534 95.614 -68.796 53.340
365535 95.639 -68.796 53.326
365536 95.666 -68.797 53.313
365537 95.692 -68.798 53.305
365538 95.718 -68.799 53.306
365539 95.744 -68.800 53.316
365540 95.770 -68.801 53.319
365541 95.796 -68.802 53.314
365542 95.822 -68.802 53.298
365543 95.848 -68.804 53.275
365544 95.874 -68.805 53.248
365545 95.900 -68.805 53.219
365546 95.926 -68.806 53.192
365547 95.952 -68.808 53.169
365548 95.978 -68.808 53.154
365549 96.004 -68.809 53.149
365550 96.030 -68.810 53.147
365551 96.056 -68.811 53.141
365552 96.079 -68.810 53.129
365553 96.096 -68.805 53.121
365554 96.113 -68.801 53.122
365555 96.130 -68.796 53.127
365556 96.158 -68.796 53.122
365557 96.193 -68.799 53.109
365558 96.228 -68.803 53.086
365559 96.262 -68.808 53.055
365560 96.296 -68.811 53.020
365561 96.330 -68.816 52.981
365562 96.364 -68.821 52.948
365563 96.398 -68.828 52.916
365564 96.430 -68.839 52.885
365565 96.458 -68.854 52.858
365566 96.487 -68.868 52.831
365567 96.530 -68.872 52.808
365568 96.572 -68.877 52.785
365569 96.612 -68.885 52.756
365570 96.653 -68.893 52.717
365571 96.693 -68.901 52.671
365572 96.728 -68.912 52.624
365573 96.762 -68.923 52.580
365574 96.797 -68.934 52.538
365575 96.831 -68.945 52.501
365576 96.866 -68.955 52.464
365577 96.900 -68.967 52.428
365578 96.935 -68.977 52.388
365579 96.969 -68.989 52.351
365580 97.004 -68.999 52.312
365581 97.038 -69.010 52.275
365582 97.066 -69.017 52.239
365583 97.094 -69.023 52.206
365584 97.121 -69.029 52.171
365585 97.148 -69.033 52.136
365586 97.174 -69.035 52.103
365587 97.201 -69.037 52.071
365588 97.227 -69.039 52.040
365589 97.253 -69.041 52.012
365590 97.278 -69.041 51.986
365591 97.304 -69.043 51.960
365592 97.336 -69.048 51.937
365593 97.368 -69.053 51.919
365594 97.400 -69.059 51.905
365595 97.433 -69.064 51.894
365596 97.465 -69.069 51.890
365597 97.496 -69.074 51.886
365598 97.527 -69.081 51.881
365599 97.559 -69.086 51.874
365600 97.592 -69.090 51.864
365601 97.630 -69.094 51.850
365602 97.659 -69.100 51.832
365603 97.689 -69.106 51.807
365604 97.722 -69.112 51.778
365605 97.754 -69.118 51.744
365606 97.786 -69.124 51.707
365607 97.818 -69.129 51.667
365608 97.851 -69.135 51.624
365609 97.886 -69.142 51.581
365610 97.924 -69.150 51.532
365611 97.962 -69.157 51.483
365612 98.000 -69.165 51.434
365613 98.037 -69.173 51.384
365614 98.076 -69.181 51.334
365615 98.113 -69.189 51.284
365616 98.152 -69.197 51.233
365617 98.192 -69.205 51.178
365618 98.237 -69.210 51.124
365619 98.281 -69.214 51.068
365620 98.325 -69.217 51.012
365621 98.369 -69.220 50.957
365622 98.414 -69.224 50.902
365623 98.457 -69.227 50.847
365624 98.502 -69.230 50.794
365625 98.546 -69.234 50.741
365626 98.590 -69.237 50.688
365627 98.634 -69.240 50.634
365628 98.677 -69.244 50.580
365629 98.708 -69.253 50.524
365630 98.730 -69.259 50.467
365631 98.752 -69.266 50.410
365632 98.774 -69.272 50.353
365633 98.796 -69.279 50.297
365634 98.818 -69.285 50.241
365635 98.840 -69.292 50.184
365636 98.863 -69.301 50.129
365637 98.895 -69.323 50.076
365638 98.926 -69.344 50.022
365639 98.957 -69.366 49.968
365640 98.988 -69.388 49.916
365641 99.016 -69.404 49.865
365642 99.042 -69.414 49.813
365643 99.067 -69.424 49.762
365644 99.091 -69.434 49.711
365645 99.117 -69.444 49.659
365646 99.142 -69.454 49.608
365647 99.167 -69.463 49.557
365648 99.192 -69.473 49.505
365649 99.248 -69.476 49.454
365650 99.284 -69.490 49.405
365651 99.305 -69.510 49.355
365652 99.354 -69.523 49.306
365653 99.404 -69.535 49.258
365654 99.453 -69.547 49.210
365655 99.502 -69.560 49.162
365656 99.552 -69.572 49.115
365657 99.593 -69.595 49.068
365658 99.634 -69.621 49.021
365659 99.674 -69.648 48.975
365660 99.723 -69.662 48.929
365661 99.772 -69.676 48.882
365662 99.821 -69.691 48.837
365663 99.870 -69.705 48.792
365664 99.919 -69.719 48.747
365665 99.968 -69.733 48.702
365666 100.017 -69.748 48.659
365667 100.065 -69.762 48.615
365668 100.108 -69.777 48.574
365669 100.151 -69.793 48.535
365670 100.194 -69.808 48.495
365671 100.237 -69.823 48.456
365672 100.258 -69.839 48.419
365673 100.269 -69.855 48.383
365674 100.281 -69.870 48.348
365675 100.301 -69.877 48.316
365676 100.337 -69.881 48.284
365677 100.374 -69.886 48.252
365678 100.411 -69.891 48.221
365679 100.447 -69.896 48.191
365680 100.484 -69.901 48.161
365681 100.521 -69.906 48.131
365682 100.557 -69.911 48.101
365683 100.594 -69.916 48.071
365684 100.631 -69.921 48.040
365685 100.668 -69.926 48.007
365686 100.704 -69.930 47.971
365687 100.741 -69.935 47.934
365688 100.777 -69.940 47.900
365689 100.814 -69.945 47.865
365690 100.856 -69.952 47.831
365691 100.902 -69.958 47.795
365692 100.922 -69.971 47.756
365693 100.951 -69.975 47.714
365694 100.980 -69.980 47.675
365695 101.009 -69.984 47.637
365696 101.039 -69.989 47.598
365697 101.067 -69.993 47.557
365698 101.097 -69.998 47.513
365699 101.132 -69.999 47.470
365700 101.172 -69.996 47.427
365701 101.218 -69.994 47.383
365702 101.260 -69.997 47.341
365703 101.299 -70.002 47.298
365704 101.337 -70.008 47.255
365705 101.376 -70.013 47.212
365706 101.414 -70.018 47.170
365707 101.453 -70.024 47.129
365708 101.492 -70.028 47.088
365709 101.531 -70.033 47.047
365710 101.569 -70.039 47.007
365711 101.608 -70.044 46.966
365712 101.647 -70.049 46.928
365713 101.686 -70.055 46.889
365714 101.724 -70.060 46.851
365715 101.763 -70.065 46.813
366133 91.314 -69.138 54.859
366134 91.373 -69.167 54.853
366135 91.431 -69.197 54.846
366136 91.488 -69.231 54.839
366137 91.539 -69.291 54.832
366138 91.589 -69.350 54.824
366139 91.640 -69.410 54.814
366140 91.691 -69.391 54.803
366141 91.733 -69.360 54.791
366142 91.772 -69.331 54.777
366143 91.811 -69.301 54.762
366144 91.852 -69.275 54.747
366210 94.571 -68.721 53.467
366211 94.624 -68.688 53.458
366212 94.670 -68.676 53.450
366213 94.715 -68.667 53.440
366214 94.761 -68.658 53.433
366215 94.807 -68.648 53.426
366216 94.852 -68.639 53.419
366217 94.898 -68.629 53.412
366218 94.942 -68.620 53.406
366219 94.963 -68.644 53.401
366220 94.983 -68.667 53.398
366221 95.004 -68.691 53.396
366222 95.028 -68.707 53.396
366223 95.054 -68.717 53.396
366224 95.081 -68.727 53.393
366225 95.107 -68.738 53.388
366226 95.134 -68.747 53.382
366227 95.169 -68.735 53.376
366228 95.209 -68.734 53.368
366229 95.291 -68.727 53.359
366230 95.373 -68.720 53.350
366231 95.440 -68.723 53.346
366232 95.490 -68.738 53.345
366233 95.540 -68.752 53.344
366234 95.591 -68.767 53.341
366235 95.621 -68.770 53.333
366236 95.647 -68.771 53.322
366237 95.673 -68.772 53.311
366238 95.699 -68.774 53.303
366239 95.725 -68.774 53.306
366240 95.751 -68.775 53.318
366241 95.777 -68.776 53.321
366242 95.803 -68.777 53.315
366243 95.829 -68.778 53.299
366244 95.855 -68.779 53.274
366245 95.882 -68.780 53.244
366246 95.907 -68.780 53.213
366247 95.934 -68.782 53.183
366248 95.960 -68.783 53.159
366249 95.985 -68.783 53.144
366250 96.012 -68.784 53.137
366251 96.038 -68.786 53.133
366252 96.062 -68.786 53.126
366253 96.080 -68.781 53.115
366254 96.097 -68.777 53.109
366255 96.114 -68.772 53.109
366256 96.131 -68.767 53.113
366257 96.160 -68.767 53.109
366258 96.194 -68.772 53.095
366259 96.229 -68.777 53.073
366260 96.263 -68.781 53.045
366261 96.297 -68.786 53.013
366262 96.331 -68.790 52.978
366263 96.365 -68.795 52.946
366264 96.397 -68.803 52.915
366265 96.426 -68.817 52.884
366266 96.454 -68.832 52.851
366267 96.490 -68.840 52.819
366268 96.533 -68.844 52.791
366269 96.576 -68.849 52.764
366270 96.617 -68.856 52.737
366271 96.657 -68.864 52.701
366272 96.698 -68.872 52.660
366273 96.739 -68.880 52.615
366274 96.774 -68.891 52.570
366275 96.809 -68.902 52.529
366276 96.843 -68.913 52.491
366277 96.878 -68.924 52.453
366278 96.912 -68.934 52.415
366279 96.946 -68.946 52.376
366280 96.981 -68.956 52.337
366281 97.015 -68.968 52.298
366282 97.050 -68.978 52.261
366283 97.079 -68.985 52.225
366284 97.106 -68.991 52.192
366285 97.134 -68.998 52.157
366286 97.160 -69.000 52.123
366287 97.186 -69.002 52.090
366288 97.213 -69.004 52.059
366289 97.239 -69.006 52.031
366290 97.264 -69.007 52.007
366291 97.286 -69.003 51.983
366292 97.312 -69.005 51.960
366293 97.344 -69.011 51.940
366294 97.375 -69.016 51.925
366295 97.407 -69.021 51.914
366296 97.440 -69.027 51.907
366297 97.471 -69.032 51.904
366298 97.504 -69.037 51.899
366299 97.536 -69.043 51.893
366300 97.568 -69.048 51.886
366301 97.600 -69.052 51.875
366302 97.633 -69.059 51.861
366303 97.665 -69.064 51.842
366304 97.697 -69.070 51.817
366305 97.729 -69.075 51.786
366306 97.762 -69.081 51.750
366307 97.794 -69.087 51.710
366308 97.826 -69.093 51.669
366309 97.862 -69.099 51.626
366310 97.900 -69.107 51.581
366311 97.938 -69.115 51.532
366312 97.976 -69.123 51.481
366313 98.014 -69.131 51.430
366314 98.052 -69.139 51.380
366315 98.090 -69.147 51.328
366316 98.128 -69.154 51.278
366317 98.166 -69.162 51.227
366318 98.203 -69.169 51.173
366319 98.243 -69.174 51.119
366320 98.287 -69.177 51.065
366321 98.332 -69.180 51.009
366322 98.376 -69.182 50.952
366323 98.420 -69.185 50.898
366324 98.464 -69.188 50.846
366325 98.509 -69.191 50.795
366326 98.553 -69.194 50.743
366327 98.597 -69.197 50.691
366328 98.641 -69.200 50.635
366329 98.683 -69.209 50.580
366330 98.700 -69.206 50.525
366331 98.722 -69.212 50.467
366332 98.743 -69.219 50.409
366333 98.766 -69.225 50.353
366334 98.787 -69.231 50.297
366335 98.809 -69.238 50.241
366336 98.831 -69.244 50.186
366337 98.858 -69.259 50.130
366338 98.888 -69.280 50.076
366339 98.920 -69.302 50.023
366340 98.951 -69.323 49.970
366341 98.982 -69.345 49.918
366342 99.013 -69.366 49.866
366343 99.039 -69.377 49.815
366344 99.063 -69.387 49.764
366345 99.088 -69.397 49.713
366346 99.114 -69.407 49.661
366347 99.138 -69.416 49.610
366348 99.174 -69.423 49.557
366349 99.238 -69.424 49.505
366350 99.284 -69.434 49.454
366351 99.302 -69.454 49.403
366352 99.321 -69.476 49.353
366353 99.369 -69.488 49.304
366354 99.418 -69.501 49.256
366355 99.468 -69.513 49.208
366356 99.517 -69.526 49.161
366357 99.558 -69.551 49.113
366358 99.598 -69.577 49.066
366359 99.645 -69.594 49.018
366360 99.694 -69.608 48.972
366361 99.743 -69.623 48.926
366362 99.792 -69.636 48.880
366363 99.841 -69.651 48.834
366364 99.890 -69.665 48.789
366365 99.939 -69.679 48.744
366366 99.988 -69.694 48.699
366367 100.037 -69.708 48.654
366368 100.086 -69.722 48.610
366369 100.129 -69.738 48.569
366370 100.173 -69.753 48.528
366371 100.215 -69.768 48.488
366372 100.240 -69.783 48.448
366373 100.252 -69.799 48.410
366374 100.264 -69.814 48.373
366375 100.282 -69.822 48.337
366376 100.310 -69.827 48.303
366377 100.346 -69.833 48.270
366378 100.383 -69.839 48.237
366379 100.420 -69.844 48.205
366380 100.456 -69.849 48.173
366381 100.493 -69.855 48.141
366382 100.530 -69.861 48.111
366383 100.566 -69.866 48.080
366384 100.603 -69.872 48.049
366385 100.640 -69.877 48.017
366386 100.677 -69.883 47.983
366387 100.713 -69.889 47.948
366388 100.750 -69.894 47.912
366389 100.786 -69.899 47.878
366390 100.823 -69.905 47.842
366391 100.860 -69.911 47.808
366392 100.893 -69.925 47.772
366393 100.922 -69.930 47.732
366394 100.952 -69.935 47.691
366395 100.980 -69.939 47.652
366396 101.009 -69.944 47.613
366397 101.039 -69.949 47.574
366398 101.068 -69.953 47.533
366399 101.102 -69.955 47.490
366400 101.141 -69.952 47.448
366401 101.184 -69.950 47.407
366402 101.231 -69.949 47.365
366403 101.270 -69.955 47.323
366404 101.309 -69.960 47.281
366405 101.347 -69.965 47.239
366406 101.386 -69.971 47.197
366407 101.424 -69.975 47.155
366408 101.463 -69.980 47.114
366409 101.502 -69.986 47.073
366410 101.541 -69.991 47.032
366411 101.580 -69.996 46.993
366412 101.618 -70.002 46.953
366413 101.657 -70.007 46.915
366414 101.695 -70.012 46.877
366832 91.191 -69.051 54.860
366833 91.250 -69.081 54.856
366834 91.308 -69.109 54.852
366835 91.367 -69.139 54.846
366836 91.425 -69.168 54.841
366837 91.481 -69.206 54.835
366838 91.532 -69.266 54.828
366839 91.583 -69.326 54.820
366840 91.633 -69.386 54.810
366841 91.684 -69.360 54.800
366842 91.724 -69.330 54.788
366843 91.763 -69.300 54.775
366844 91.802 -69.271 54.760
366845 91.848 -69.259 54.745
366846 91.896 -69.251 54.729
366911 94.589 -68.668 53.469
366912 94.635 -68.659 53.460
366913 94.680 -68.649 53.452
366914 94.726 -68.639 53.443
366915 94.771 -68.630 53.435
366916 94.817 -68.620 53.428
366917 94.862 -68.611 53.421
366918 94.908 -68.601 53.414
366919 94.940 -68.610 53.408
366920 94.961 -68.633 53.402
366921 94.981 -68.657 53.398
366922 95.002 -68.681 53.397
366923 95.023 -68.703 53.397
366924 95.049 -68.714 53.395
366925 95.075 -68.724 53.392
366926 95.109 -68.717 53.388
366927 95.144 -68.704 53.382
366928 95.179 -68.692 53.375
366929 95.213 -68.688 53.367
366930 95.255 -68.694 53.358
366931 95.337 -68.687 53.349
366932 95.417 -68.681 53.343
366933 95.467 -68.696 53.339
366934 95.518 -68.711 53.336
366935 95.568 -68.726 53.332
366936 95.619 -68.740 53.325
366937 95.655 -68.747 53.319
366938 95.681 -68.748 53.310
366939 95.707 -68.749 53.302
366940 95.733 -68.749 53.302
366941 95.759 -68.750 53.313
366942 95.785 -68.752 53.316
366943 95.811 -68.752 53.310
366944 95.837 -68.753 53.296
366945 95.863 -68.754 53.270
366946 95.889 -68.755 53.239
366947 95.915 -68.756 53.206
366948 95.941 -68.757 53.175
366949 95.967 -68.758 53.152
366950 95.993 -68.758 53.137
366951 96.019 -68.760 53.128
366952 96.045 -68.761 53.121
366953 96.063 -68.757 53.113
366954 96.081 -68.752 53.103
366955 96.098 -68.747 53.097
366956 96.115 -68.742 53.095
366957 96.135 -68.739 53.096
366958 96.164 -68.742 53.092
366959 96.193 -68.750 53.078
366960 96.224 -68.757 53.057
366961 96.257 -68.763 53.031
366962 96.291 -68.768 53.002
366963 96.324 -68.774 52.971
366964 96.358 -68.779 52.940
366965 96.391 -68.785 52.911
366966 96.421 -68.796 52.880
366967 96.452 -68.808 52.844
366968 96.495 -68.812 52.809
366969 96.537 -68.816 52.776
366970 96.580 -68.820 52.745
366971 96.621 -68.828 52.717
366972 96.662 -68.836 52.684
366973 96.702 -68.844 52.647
366974 96.743 -68.852 52.605
366975 96.784 -68.860 52.560
366976 96.820 -68.870 52.519
366977 96.855 -68.880 52.480
366978 96.889 -68.892 52.442
366979 96.924 -68.902 52.403
366980 96.958 -68.913 52.363
366981 96.993 -68.924 52.323
366982 97.027 -68.935 52.284
366983 97.062 -68.946 52.247
366984 97.091 -68.953 52.211
366985 97.118 -68.960 52.178
366986 97.146 -68.965 52.144
366987 97.172 -68.968 52.110
366988 97.198 -68.969 52.079
366989 97.225 -68.971 52.049
366990 97.250 -68.974 52.024
366991 97.272 -68.970 52.002
366992 97.294 -68.966 51.982
366993 97.320 -68.968 51.962
366994 97.352 -68.974 51.947
366995 97.382 -68.980 51.936
366996 97.414 -68.985 51.929
366997 97.446 -68.990 51.924
366998 97.478 -68.995 51.921
366999 97.511 -69.000 51.914
367000 97.543 -69.005 51.905
367001 97.575 -69.011 51.896
367002 97.607 -69.016 51.882
367003 97.640 -69.021 51.866
367004 97.673 -69.027 51.847
367005 97.705 -69.033 51.822
367006 97.737 -69.039 51.790
367007 97.770 -69.044 51.755
367008 97.802 -69.050 51.715
367009 97.839 -69.057 51.674
367010 97.876 -69.065 51.630
367011 97.915 -69.073 51.586
367012 97.952 -69.081 51.535
367013 97.991 -69.089 51.484
367014 98.028 -69.096 51.432
367015 98.065 -69.103 51.379
367016 98.101 -69.109 51.326
367017 98.137 -69.115 51.273
367018 98.173 -69.121 51.219
367019 98.209 -69.127 51.164
367020 98.252 -69.130 51.110
367021 98.296 -69.133 51.057
367022 98.341 -69.136 51.003
367023 98.384 -69.139 50.949
367024 98.429 -69.142 50.898
367025 98.473 -69.144 50.848
367026 98.517 -69.147 50.797
367027 98.562 -69.150 50.745
367028 98.606 -69.153 50.692
367029 98.647 -69.165 50.636
367030 98.687 -69.178 50.581
367031 98.706 -69.176 50.525
367032 98.715 -69.168 50.468
367033 98.735 -69.172 50.411
367034 98.757 -69.178 50.355
367035 98.779 -69.184 50.300
367036 98.801 -69.191 50.245
367037 98.823 -69.197 50.190
367038 98.852 -69.215 50.134
367039 98.883 -69.237 50.079
367040 98.914 -69.259 50.025
367041 98.945 -69.281 49.972
367042 98.976 -69.302 49.919
367043 99.007 -69.324 49.868
367044 99.035 -69.339 49.817
367045 99.060 -69.349 49.766
367046 99.085 -69.359 49.714
367047 99.110 -69.369 49.662
367048 99.163 -69.372 49.609
367049 99.227 -69.372 49.557
367050 99.285 -69.378 49.505
367051 99.303 -69.398 49.453
367052 99.321 -69.419 49.402
367053 99.339 -69.441 49.352
367054 99.384 -69.454 49.303
367055 99.433 -69.466 49.254
367056 99.482 -69.480 49.206
367057 99.522 -69.506 49.159
367058 99.567 -69.526 49.111
367059 99.616 -69.539 49.063
367060 99.665 -69.554 49.016
367061 99.714 -69.568 48.969
367062 99.763 -69.582 48.922
367063 99.812 -69.597 48.876
367064 99.861 -69.611 48.830
367065 99.910 -69.626 48.786
367066 99.960 -69.639 48.740
367067 100.008 -69.654 48.695
367068 100.057 -69.668 48.650
367069 100.106 -69.682 48.606
367070 100.151 -69.698 48.564
367071 100.194 -69.713 48.522
367072 100.222 -69.728 48.481
367073 100.234 -69.744 48.440
367074 100.251 -69.754 48.401
367075 100.270 -69.758 48.363
367076 100.289 -69.764 48.326
367077 100.318 -69.769 48.291
367078 100.354 -69.775 48.257
367079 100.391 -69.780 48.223
367080 100.428 -69.786 48.190
367081 100.465 -69.792 48.156
367082 100.501 -69.797 48.124
367083 100.538 -69.802 48.092
367084 100.575 -69.808 48.061
367085 100.611 -69.814 48.029
367086 100.648 -69.820 47.996
367087 100.685 -69.825 47.961
367088 100.721 -69.830 47.926
367089 100.758 -69.836 47.891
367090 100.795 -69.842 47.856
367091 100.833 -69.865 47.821
367092 100.864 -69.880 47.786
367093 100.894 -69.885 47.749
367094 100.922 -69.889 47.710
367095 100.952 -69.894 47.670
367096 100.981 -69.899 47.631
367097 101.009 -69.904 47.592
367098 101.039 -69.908 47.552
367099 101.071 -69.911 47.512
367100 101.111 -69.908 47.471
367101 101.150 -69.906 47.430
367102 101.199 -69.905 47.389
367103 101.241 -69.907 47.348
367104 101.280 -69.912 47.306
367105 101.319 -69.917 47.265
367106 101.357 -69.923 47.224
367107 101.396 -69.927 47.182
367108 101.434 -69.933 47.141
367109 101.473 -69.938 47.099
367110 101.512 -69.943 47.059
367111 101.551 -69.949 47.019
367112 101.590 -69.954 46.980
367113 101.628 -69.959 46.942
367114 101.667 -69.964 46.903
367532 91.127 -68.993 54.858
367533 91.185 -69.023 54.853
367534 91.244 -69.052 54.849
367535 91.302 -69.081 54.846
367536 91.361 -69.111 54.842
367537 91.419 -69.140 54.837
367538 91.474 -69.181 54.832
367539 91.525 -69.241 54.826
367540 91.576 -69.301 54.818
367541 91.627 -69.357 54.809
367542 91.675 -69.330 54.798
367543 91.715 -69.300 54.786
367544 91.754 -69.270 54.773
367545 91.797 -69.250 54.758
367546 91.844 -69.242 54.743
367547 91.891 -69.234 54.727
367548 91.936 -69.222 54.712
367613 94.645 -68.631 53.462
367614 94.691 -68.622 53.453
367615 94.737 -68.612 53.444
367616 94.782 -68.603 53.437
367617 94.828 -68.593 53.429
367618 94.873 -68.583 53.422
367619 94.917 -68.576 53.415
367620 94.938 -68.599 53.409
367621 94.959 -68.623 53.403
367622 94.979 -68.646 53.398
367623 95.000 -68.670 53.396
367624 95.021 -68.694 53.396
367625 95.049 -68.698 53.393
367626 95.084 -68.686 53.389
367627 95.119 -68.673 53.384
367628 95.154 -68.661 53.377
367629 95.189 -68.648 53.370
367630 95.224 -68.641 53.362
367631 95.255 -68.649 53.353
367632 95.302 -68.654 53.344
367633 95.384 -68.647 53.337
367634 95.445 -68.654 53.332
367635 95.495 -68.669 53.329
367636 95.546 -68.684 53.326
367637 95.597 -68.698 53.322
367638 95.647 -68.714 53.316
367639 95.688 -68.723 53.306
367640 95.714 -68.723 53.296
367641 95.740 -68.725 53.292
367642 95.766 -68.726 53.301
367643 95.793 -68.727 53.304
367644 95.818 -68.727 53.300
367645 95.844 -68.729 53.288
367646 95.871 -68.730 53.263
367647 95.897 -68.730 53.231
367648 95.922 -68.731 53.199
367649 95.949 -68.732 53.170
367650 95.975 -68.733 53.148
367651 96.001 -68.734 53.133
367652 96.027 -68.735 53.122
367653 96.047 -68.732 53.114
367654 96.064 -68.727 53.104
367655 96.081 -68.723 53.094
367656 96.098 -68.718 53.085
367657 96.122 -68.716 53.080
367658 96.151 -68.714 53.078
367659 96.179 -68.714 53.070
367660 96.208 -68.722 53.057
367661 96.238 -68.730 53.038
367662 96.268 -68.737 53.014
367663 96.297 -68.745 52.989
367664 96.327 -68.752 52.961
367665 96.356 -68.760 52.932
367666 96.386 -68.767 52.903
367667 96.416 -68.774 52.873
367668 96.456 -68.780 52.837
367669 96.499 -68.783 52.800
367670 96.541 -68.787 52.764
367671 96.584 -68.792 52.729
367672 96.625 -68.799 52.697
367673 96.666 -68.808 52.667
367674 96.707 -68.815 52.632
367675 96.747 -68.824 52.592
367676 96.788 -68.831 52.549
367677 96.828 -68.839 52.508
367678 96.866 -68.849 52.469
367679 96.901 -68.859 52.430
367680 96.935 -68.871 52.390
367681 96.970 -68.881 52.348
367682 97.004 -68.892 52.308
367683 97.038 -68.903 52.269
367684 97.073 -68.914 52.233
367685 97.103 -68.922 52.198
367686 97.131 -68.928 52.164
367687 97.158 -68.933 52.131
367688 97.184 -68.934 52.099
367689 97.210 -68.936 52.069
367690 97.236 -68.939 52.042
367691 97.259 -68.936 52.019
367692 97.281 -68.932 51.999
367693 97.302 -68.929 51.982
367694 97.329 -68.930 51.966
367695 97.360 -68.936 51.955
367696 97.390 -68.942 51.949
367697 97.422 -68.948 51.944
367698 97.453 -68.953 51.941
367699 97.486 -68.958 51.938
367700 97.518 -68.964 51.928
367701 97.549 -68.969 51.916
367702 97.582 -68.974 51.904
367703 97.614 -68.979 51.888
367704 97.647 -68.985 51.870
367705 97.680 -68.990 51.849
367706 97.713 -68.996 51.824
367707 97.745 -69.002 51.792
367708 97.778 -69.008 51.757
367709 97.815 -69.015 51.718
367710 97.853 -69.023 51.676
367711 97.889 -69.030 51.632
367712 97.926 -69.036 51.587
367713 97.962 -69.042 51.536
367714 97.998 -69.047 51.484
367715 98.034 -69.053 51.430
367716 98.071 -69.059 51.376
367717 98.107 -69.065 51.322
367718 98.143 -69.071 51.267
367719 98.179 -69.078 51.211
367720 98.216 -69.084 51.155
367721 98.261 -69.087 51.101
367722 98.305 -69.089 51.048
367723 98.349 -69.092 50.996
367724 98.393 -69.095 50.946
367725 98.438 -69.098 50.899
367726 98.482 -69.101 50.851
367727 98.526 -69.103 50.800
367728 98.570 -69.108 50.748
367729 98.610 -69.121 50.694
367730 98.651 -69.133 50.638
367731 98.691 -69.146 50.583
367732 98.713 -69.146 50.527
367733 98.722 -69.137 50.470
367734 98.731 -69.130 50.413
367735 98.748 -69.131 50.357
367736 98.771 -69.137 50.302
367737 98.792 -69.144 50.248
367738 98.815 -69.151 50.194
367739 98.846 -69.173 50.136
367740 98.877 -69.194 50.081
367741 98.908 -69.216 50.026
367742 98.939 -69.237 49.972
367743 98.971 -69.259 49.919
367744 99.002 -69.281 49.868
367745 99.032 -69.301 49.817
367746 99.057 -69.311 49.766
367747 99.089 -69.319 49.714
367748 99.153 -69.320 49.661
367749 99.217 -69.321 49.608
367750 99.285 -69.322 49.555
367751 99.303 -69.342 49.503
367752 99.321 -69.363 49.451
367753 99.339 -69.385 49.400
367754 99.357 -69.406 49.350
367755 99.399 -69.420 49.301
367756 99.446 -69.435 49.252
367757 99.490 -69.457 49.204
367758 99.539 -69.471 49.157
367759 99.587 -69.485 49.109
367760 99.637 -69.500 49.061
367761 99.686 -69.514 49.013
367762 99.734 -69.529 48.966
367763 99.784 -69.542 48.919
367764 99.833 -69.557 48.872
367765 99.881 -69.571 48.827
367766 99.931 -69.585 48.782
367767 99.980 -69.600 48.737
367768 100.028 -69.614 48.691
367769 100.077 -69.628 48.647
367770 100.127 -69.642 48.601
367771 100.172 -69.657 48.558
367772 100.205 -69.673 48.516
367773 100.219 -69.686 48.474
367774 100.239 -69.690 48.433
367775 100.258 -69.695 48.393
367776 100.277 -69.701 48.354
367777 100.296 -69.706 48.316
367778 100.326 -69.711 48.280
367779 100.363 -69.717 48.245
367780 100.400 -69.722 48.210
367781 100.436 -69.728 48.176
367782 100.473 -69.733 48.142
367783 100.509 -69.739 48.108
367784 100.546 -69.745 48.074
367785 100.583 -69.750 48.042
367786 100.620 -69.756 48.011
367787 100.656 -69.761 47.976
367788 100.693 -69.767 47.941
367789 100.730 -69.773 47.906
367790 100.766 -69.789 47.871
367791 100.804 -69.810 47.836
367792 100.834 -69.825 47.801
367793 100.864 -69.840 47.765
367794 100.894 -69.845 47.728
367795 100.923 -69.849 47.690
367796 100.952 -69.854 47.651
367797 100.981 -69.858 47.613
367798 101.010 -69.864 47.574
367799 101.041 -69.867 47.535
367800 101.080 -69.864 47.495
367801 101.120 -69.862 47.455
367802 101.165 -69.861 47.414
367803 101.212 -69.859 47.374
367804 101.251 -69.864 47.333
367805 101.290 -69.870 47.292
367806 101.328 -69.874 47.251
367807 101.367 -69.880 47.209
367808 101.406 -69.885 47.168
367809 101.445 -69.890 47.126
367810 101.483 -69.895 47.087
367811 101.522 -69.901 47.047
367812 101.561 -69.906 47.007
367813 101.599 -69.911 46.968
368232 91.075 -68.945 54.852
368233 91.121 -68.965 54.850
368234 91.179 -68.995 54.846
368235 91.238 -69.024 54.843
368236 91.296 -69.053 54.840
368237 91.355 -69.083 54.838
368238 91.413 -69.112 54.834
368239 91.467 -69.157 54.829
368240 91.518 -69.217 54.824
368241 91.569 -69.277 54.816
368242 91.619 -69.328 54.807
368243 91.666 -69.300 54.796
368244 91.706 -69.270 54.784
368245 91.745 -69.240 54.771
368246 91.793 -69.234 54.756
368247 91.840 -69.225 54.741
368248 91.884 -69.213 54.725
368249 91.929 -69.202 54.710
368250 91.974 -69.190 54.694
368267 92.733 -68.988 54.324
368268 92.777 -68.976 54.297
368269 92.822 -68.964 54.269
368314 94.656 -68.604 53.463
368315 94.702 -68.595 53.453
368316 94.747 -68.585 53.445
368317 94.793 -68.575 53.438
368318 94.838 -68.566 53.430
368319 94.884 -68.556 53.423
368320 94.915 -68.565 53.416
368321 94.936 -68.589 53.409
368322 94.957 -68.612 53.403
368323 94.977 -68.636 53.397
368324 94.998 -68.660 53.394
368325 95.021 -68.674 53.394
368326 95.051 -68.663 53.391
368327 95.090 -68.647 53.386
368328 95.128 -68.631 53.380
368329 95.164 -68.618 53.372
368330 95.198 -68.606 53.364
368331 95.234 -68.594 53.356
368332 95.266 -68.602 53.347
368333 95.298 -68.611 53.339
368334 95.348 -68.614 53.332
368335 95.422 -68.612 53.326
368336 95.473 -68.627 53.321
368337 95.523 -68.642 53.317
368338 95.574 -68.657 53.314
368339 95.624 -68.671 53.308
368340 95.675 -68.686 53.299
368341 95.722 -68.699 53.287
368342 95.748 -68.700 53.282
368343 95.774 -68.701 53.289
368344 95.800 -68.701 53.292
368345 95.826 -68.703 53.288
368346 95.852 -68.704 53.278
368347 95.878 -68.705 53.255
368348 95.904 -68.705 53.224
368349 95.930 -68.707 53.193
368350 95.956 -68.708 53.167
368351 95.982 -68.708 53.148
368352 96.008 -68.709 53.136
368353 96.030 -68.708 53.128
368354 96.047 -68.703 53.120
368355 96.065 -68.698 53.108
368356 96.082 -68.694 53.094
368357 96.110 -68.692 53.081
368358 96.138 -68.692 53.070
368359 96.167 -68.690 53.061
368360 96.195 -68.689 53.050
368361 96.224 -68.694 53.036
368362 96.254 -68.701 53.018
368363 96.283 -68.709 52.997
368364 96.313 -68.717 52.975
368365 96.342 -68.724 52.950
368366 96.372 -68.732 52.922
368367 96.401 -68.739 52.893
368368 96.431 -68.747 52.862
368369 96.461 -68.752 52.827
368370 96.503 -68.755 52.791
368371 96.545 -68.759 52.754
368372 96.588 -68.763 52.717
368373 96.629 -68.770 52.681
368374 96.670 -68.779 52.649
368375 96.711 -68.786 52.614
368376 96.751 -68.795 52.576
368377 96.792 -68.802 52.535
368378 96.832 -68.811 52.493
368379 96.873 -68.818 52.454
368380 96.912 -68.827 52.415
368381 96.947 -68.838 52.375
368382 96.981 -68.849 52.333
368383 97.016 -68.860 52.293
368384 97.050 -68.871 52.255
368385 97.085 -68.882 52.219
368386 97.115 -68.890 52.184
368387 97.143 -68.896 52.151
368388 97.170 -68.900 52.119
368389 97.196 -68.902 52.088
368390 97.222 -68.904 52.060
368391 97.245 -68.902 52.036
368392 97.267 -68.899 52.016
368393 97.289 -68.895 51.999
368394 97.310 -68.891 51.984
368395 97.337 -68.893 51.972
368396 97.368 -68.899 51.966
368397 97.398 -68.904 51.963
368398 97.430 -68.910 51.960
368399 97.460 -68.916 51.957
368400 97.492 -68.921 51.952
368401 97.524 -68.927 51.940
368402 97.557 -68.932 51.925
368403 97.589 -68.937 51.909
368404 97.621 -68.943 51.891
368405 97.654 -68.948 51.871
368406 97.687 -68.954 51.849
368407 97.720 -68.959 51.823
368408 97.753 -68.965 51.791
368409 97.787 -68.968 51.756
368410 97.823 -68.974 51.718
368411 97.860 -68.980 51.676
368412 97.896 -68.987 51.631
368413 97.932 -68.993 51.584
368414 97.968 -68.999 51.533
368415 98.004 -69.005 51.479
368416 98.040 -69.011 51.425
368417 98.077 -69.017 51.369
368418 98.113 -69.023 51.314
368419 98.149 -69.029 51.259
368420 98.185 -69.035 51.202
368421 98.225 -69.040 51.146
368422 98.269 -69.043 51.090
368423 98.313 -69.046 51.039
368424 98.358 -69.049 50.990
368425 98.402 -69.051 50.944
368426 98.446 -69.054 50.901
368427 98.491 -69.057 50.854
368428 98.533 -69.064 50.804
368429 98.574 -69.076 50.752
368430 98.614 -69.089 50.698
368431 98.655 -69.101 50.642
368432 98.695 -69.114 50.586
368433 98.719 -69.115 50.530
368434 98.728 -69.107 50.472
368435 98.737 -69.099 50.415
368436 98.746 -69.091 50.359
368437 98.762 -69.090 50.304
368438 98.784 -69.097 50.249
368439 98.809 -69.109 50.194
368440 98.840 -69.130 50.137
368441 98.871 -69.152 50.080
368442 98.903 -69.173 50.024
368443 98.933 -69.195 49.970
368444 98.964 -69.217 49.916
368445 98.996 -69.238 49.865
368446 99.027 -69.260 49.814
368447 99.078 -69.268 49.763
368448 99.142 -69.269 49.710
368449 99.206 -69.269 49.657
368450 99.285 -69.265 49.605
368451 99.303 -69.286 49.552
368452 99.322 -69.307 49.499
368453 99.340 -69.328 49.448
368454 99.358 -69.350 49.397
368455 99.376 -69.370 49.347
368456 99.412 -69.388 49.298
368457 99.461 -69.403 49.249
368458 99.510 -69.417 49.200
368459 99.559 -69.432 49.153
368460 99.608 -69.445 49.106
368461 99.657 -69.460 49.058
368462 99.706 -69.474 49.010
368463 99.755 -69.488 48.964
368464 99.804 -69.503 48.917
368465 99.853 -69.517 48.871
368466 99.902 -69.532 48.825
368467 99.951 -69.545 48.779
368468 100.000 -69.560 48.733
368469 100.049 -69.574 48.688
368470 100.098 -69.588 48.644
368471 100.147 -69.603 48.597
368472 100.188 -69.617 48.554
368473 100.207 -69.622 48.511
368474 100.226 -69.627 48.469
368475 100.246 -69.632 48.427
368476 100.265 -69.638 48.386
368477 100.284 -69.642 48.345
368478 100.303 -69.648 48.306
368479 100.335 -69.653 48.270
368480 100.371 -69.659 48.234
368481 100.408 -69.664 48.198
368482 100.444 -69.670 48.164
368483 100.481 -69.676 48.129
368484 100.518 -69.681 48.094
368485 100.555 -69.686 48.061
368486 100.592 -69.692 48.028
368487 100.628 -69.698 47.995
368488 100.665 -69.703 47.961
368489 100.701 -69.720 47.926
368490 100.743 -69.758 47.890
368491 100.773 -69.754 47.855
368492 100.805 -69.770 47.820
368493 100.836 -69.795 47.785
368494 100.864 -69.800 47.749
368495 100.894 -69.805 47.712
368496 100.923 -69.809 47.674
368497 100.952 -69.814 47.635
368498 100.981 -69.818 47.597
368499 101.010 -69.823 47.559
368500 101.050 -69.820 47.520
368501 101.090 -69.818 47.480
368502 101.131 -69.817 47.440
368503 101.179 -69.814 47.399
368504 101.222 -69.817 47.359
368505 101.261 -69.821 47.319
368506 101.299 -69.827 47.278
368507 101.338 -69.832 47.237
368508 101.377 -69.837 47.196
368509 101.416 -69.842 47.154
368510 101.455 -69.848 47.113
368511 101.493 -69.853 47.074
368512 101.532 -69.858 47.035
368513 101.570 -69.863 46.996
368931 91.001 -68.891 54.850
368932 91.032 -68.902 54.846
368933 91.063 -68.912 54.844
368934 91.115 -68.937 54.842
368935 91.173 -68.967 54.840
368936 91.232 -68.996 54.838
368937 91.290 -69.025 54.836
368938 91.349 -69.055 54.834
368939 91.406 -69.084 54.831
368940 91.460 -69.132 54.827
368941 91.511 -69.192 54.822
368942 91.562 -69.252 54.814
368943 91.612 -69.298 54.805
368944 91.657 -69.269 54.794
368945 91.696 -69.240 54.782
368946 91.741 -69.225 54.770
368947 91.788 -69.216 54.755
368948 91.833 -69.205 54.739
368949 91.877 -69.193 54.724
368950 91.922 -69.181 54.708
368951 91.967 -69.169 54.693
368952 92.011 -69.157 54.678
368967 92.681 -68.979 54.355
368968 92.726 -68.967 54.328
368969 92.771 -68.955 54.301
368970 92.815 -68.943 54.274
368971 92.860 -68.931 54.246
368972 92.905 -68.920 54.220
368973 92.949 -68.908 54.192
368974 92.993 -68.897 54.164
368975 93.028 -68.898 54.136
368976 93.063 -68.899 54.109
368977 93.098 -68.899 54.081
368978 93.133 -68.900 54.055
369016 94.712 -68.567 53.455
369017 94.758 -68.557 53.446
369018 94.804 -68.548 53.438
369019 94.849 -68.538 53.431
369020 94.893 -68.531 53.423
369021 94.913 -68.554 53.416
369022 94.934 -68.578 53.410
369023 94.955 -68.601 53.402
369024 94.975 -68.626 53.396
369025 94.996 -68.649 53.391
369026 95.026 -68.639 53.390
369027 95.057 -68.627 53.386
369028 95.091 -68.614 53.382
369029 95.129 -68.598 53.376
369030 95.168 -68.582 53.368
369031 95.206 -68.566 53.358
369032 95.243 -68.551 53.349
369033 95.276 -68.554 53.340
369034 95.308 -68.564 53.333
369035 95.340 -68.573 53.327
369036 95.394 -68.574 53.320
369037 95.450 -68.586 53.314
369038 95.501 -68.600 53.308
369039 95.551 -68.615 53.302
369040 95.602 -68.629 53.297
369041 95.653 -68.645 53.288
369042 95.703 -68.659 53.278
369043 95.753 -68.674 53.274
369044 95.781 -68.676 53.279
369045 95.808 -68.677 53.280
369046 95.833 -68.678 53.277
369047 95.860 -68.679 53.268
369048 95.886 -68.680 53.248
369049 95.912 -68.681 53.217
369050 95.938 -68.682 53.189
369051 95.964 -68.683 53.168
369052 95.990 -68.683 53.154
369053 96.014 -68.683 53.147
369054 96.031 -68.679 53.140
369055 96.048 -68.674 53.132
369056 96.069 -68.670 53.118
369057 96.098 -68.670 53.103
369058 96.126 -68.668 53.087
369059 96.154 -68.667 53.071
369060 96.182 -68.666 53.055
369061 96.211 -68.665 53.038
369062 96.240 -68.666 53.022
369063 96.269 -68.673 53.003
369064 96.299 -68.681 52.981
369065 96.328 -68.689 52.961
369066 96.358 -68.696 52.937
369067 96.387 -68.704 52.911
369068 96.416 -68.711 52.882
369069 96.446 -68.718 52.850
369070 96.476 -68.723 52.815
369071 96.507 -68.727 52.778
369072 96.549 -68.731 52.741
369073 96.592 -68.735 52.704
369074 96.633 -68.742 52.666
369075 96.674 -68.750 52.632
369076 96.715 -68.758 52.596
369077 96.755 -68.766 52.558
369078 96.796 -68.774 52.519
369079 96.837 -68.782 52.478
369080 96.877 -68.790 52.438
369081 96.918 -68.798 52.399
369082 96.958 -68.806 52.360
369083 96.993 -68.817 52.319
369084 97.027 -68.828 52.280
369085 97.062 -68.839 52.243
369086 97.096 -68.850 52.207
369087 97.127 -68.858 52.172
369088 97.155 -68.864 52.141
369089 97.182 -68.867 52.110
369090 97.208 -68.869 52.082
369091 97.232 -68.868 52.056
369092 97.253 -68.864 52.034
369093 97.275 -68.861 52.017
369094 97.297 -68.858 52.004
369095 97.319 -68.854 51.992
369096 97.345 -68.855 51.983
369097 97.376 -68.861 51.981
369098 97.406 -68.867 51.980
369099 97.438 -68.873 51.978
369100 97.468 -68.879 51.972
369101 97.499 -68.884 51.965
369102 97.531 -68.889 51.948
369103 97.564 -68.895 51.930
369104 97.596 -68.900 51.910
369105 97.628 -68.905 51.887
369106 97.661 -68.911 51.865
369107 97.694 -68.917 51.842
369108 97.727 -68.922 51.815
369109 97.761 -68.926 51.782
369110 97.796 -68.929 51.747
369111 97.830 -68.933 51.707
369112 97.866 -68.938 51.664
369113 97.902 -68.944 51.619
369114 97.938 -68.950 51.571
369115 97.974 -68.956 51.519
369116 98.011 -68.962 51.465
369117 98.046 -68.968 51.409
369118 98.083 -68.974 51.354
369119 98.119 -68.980 51.299
369120 98.155 -68.987 51.244
369121 98.192 -68.993 51.190
369122 98.234 -68.996 51.137
369123 98.278 -68.999 51.086
369124 98.322 -69.002 51.036
369125 98.366 -69.005 50.988
369126 98.411 -69.008 50.943
369127 98.455 -69.011 50.900
369128 98.497 -69.019 50.853
369129 98.537 -69.032 50.803
369130 98.578 -69.044 50.751
369131 98.618 -69.057 50.696
369132 98.659 -69.069 50.641
369133 98.699 -69.082 50.585
369134 98.725 -69.085 50.529
369135 98.734 -69.077 50.472
369136 98.744 -69.069 50.414
369137 98.753 -69.061 50.358
369138 98.762 -69.053 50.302
369139 98.776 -69.050 50.246
369140 98.803 -69.065 50.191
369141 98.834 -69.087 50.134
369142 98.865 -69.109 50.078
369143 98.912 -69.134 50.022
369144 98.959 -69.160 49.967
369145 99.005 -69.186 49.913
369146 99.052 -69.212 49.861
369147 99.103 -69.230 49.810
369148 99.165 -69.230 49.758
369149 99.202 -69.236 49.706
369150 99.225 -69.245 49.654
369151 99.278 -69.248 49.601
369152 99.297 -69.269 49.549
369153 99.315 -69.290 49.497
369154 99.334 -69.310 49.447
369155 99.354 -69.329 49.396
369156 99.384 -69.339 49.346
369157 99.421 -69.349 49.296
369158 99.473 -69.357 49.247
369159 99.522 -69.372 49.198
369160 99.571 -69.386 49.152
369161 99.620 -69.400 49.105
369162 99.669 -69.415 49.058
369163 99.718 -69.429 49.010
369164 99.767 -69.443 48.964
369165 99.816 -69.457 48.917
369166 99.865 -69.472 48.870
369167 99.914 -69.486 48.824
369168 99.963 -69.501 48.777
369169 100.011 -69.514 48.732
369170 100.061 -69.529 48.686
369171 100.110 -69.543 48.640
369172 100.158 -69.557 48.594
369173 100.186 -69.566 48.550
369174 100.214 -69.564 48.507
369175 100.233 -69.569 48.463
369176 100.253 -69.574 48.421
369177 100.272 -69.579 48.379
369178 100.291 -69.585 48.338
369179 100.310 -69.589 48.298
369180 100.343 -69.595 48.260
369181 100.379 -69.601 48.224
369182 100.416 -69.606 48.188
369183 100.453 -69.612 48.152
369184 100.490 -69.617 48.117
369185 100.526 -69.623 48.082
369186 100.563 -69.629 48.049
369187 100.600 -69.634 48.015
369188 100.636 -69.651 47.982
369189 100.673 -69.668 47.947
369190 100.713 -69.702 47.912
369191 100.744 -69.699 47.877
369192 100.775 -69.715 47.841
369193 100.807 -69.751 47.807
369194 100.836 -69.755 47.771
369195 100.865 -69.760 47.735
369196 100.894 -69.764 47.698
369197 100.923 -69.769 47.660
369198 100.952 -69.773 47.621
369199 100.981 -69.778 47.583
369200 101.020 -69.776 47.545
369201 101.059 -69.774 47.506
369202 101.099 -69.773 47.467
369203 101.145 -69.770 47.426
369204 101.194 -69.768 47.385
369205 101.233 -69.773 47.345
369206 101.271 -69.779 47.305
369207 101.310 -69.784 47.264
369208 101.348 -69.789 47.224
369209 101.387 -69.795 47.183
369210 101.426 -69.800 47.142
369211 101.465 -69.805 47.101
369212 101.503 -69.810 47.061
369631 90.957 -68.849 54.845
369632 90.989 -68.859 54.841
369633 91.020 -68.869 54.838
369634 91.051 -68.880 54.837
369635 91.109 -68.909 54.836
369636 91.167 -68.938 54.835
369637 91.226 -68.968 54.834
369638 91.284 -68.997 54.834
369639 91.343 -69.026 54.832
369640 91.400 -69.056 54.830
369641 91.453 -69.108 54.827
369642 91.504 -69.168 54.822
369643 91.555 -69.227 54.814
369644 91.605 -69.269 54.805
369645 91.648 -69.240 54.794
369646 91.689 -69.215 54.781
369647 91.737 -69.208 54.768
369648 91.781 -69.196 54.754
369649 91.826 -69.184 54.739
369650 91.871 -69.172 54.723
369651 91.915 -69.160 54.708
369652 91.960 -69.149 54.693
369653 92.005 -69.137 54.677
369666 92.585 -68.982 54.410
369667 92.630 -68.971 54.385
369668 92.674 -68.958 54.359
369669 92.719 -68.946 54.333
369670 92.764 -68.935 54.306
369671 92.808 -68.923 54.280
369672 92.853 -68.911 54.253
369673 92.898 -68.899 54.227
369674 92.942 -68.887 54.200
369675 92.987 -68.875 54.172
369676 93.022 -68.875 54.145
369677 93.057 -68.877 54.118
369678 93.092 -68.877 54.090
369679 93.127 -68.878 54.064
369680 93.162 -68.879 54.037
369717 94.723 -68.539 53.456
369718 94.768 -68.530 53.446
369719 94.814 -68.520 53.438
369720 94.859 -68.511 53.430
369721 94.891 -68.520 53.423
369722 94.911 -68.544 53.416
369723 94.932 -68.567 53.409
369724 94.953 -68.592 53.401
369725 94.973 -68.615 53.394
369726 95.001 -68.614 53.389
369727 95.031 -68.602 53.385
369728 95.062 -68.592 53.381
369729 95.093 -68.580 53.376
369730 95.130 -68.565 53.370
369731 95.168 -68.549 53.362
369732 95.207 -68.532 53.353
369733 95.245 -68.517 53.343
369734 95.277 -68.505 53.334
369735 95.312 -68.514 53.328
369736 95.348 -68.524 53.321
369737 95.383 -68.534 53.313
369738 95.427 -68.544 53.306
369739 95.478 -68.558 53.298
369740 95.529 -68.573 53.289
369741 95.579 -68.588 53.282
369742 95.630 -68.603 53.275
369743 95.680 -68.617 53.269
369744 95.731 -68.633 53.269
369745 95.781 -68.647 53.273
369746 95.815 -68.652 53.273
369747 95.841 -68.653 53.269
369748 95.867 -68.654 53.259
369749 95.893 -68.655 53.240
369750 95.919 -68.656 53.212
369751 95.945 -68.657 53.188
369752 95.971 -68.658 53.173
369753 95.997 -68.659 53.170
369754 96.015 -68.654 53.167
369755 96.032 -68.650 53.161
369756 96.057 -68.648 53.151
369757 96.085 -68.646 53.136
369758 96.113 -68.645 53.120
369759 96.142 -68.644 53.104
369760 96.170 -68.643 53.084
369761 96.198 -68.642 53.062
369762 96.227 -68.641 53.036
369763 96.255 -68.639 53.014
369764 96.285 -68.645 52.991
369765 96.314 -68.653 52.969
369766 96.343 -68.661 52.948
369767 96.373 -68.668 52.925
369768 96.402 -68.676 52.898
369769 96.432 -68.683 52.868
369770 96.462 -68.689 52.836
369771 96.491 -68.694 52.801
369772 96.522 -68.698 52.765
369773 96.553 -68.702 52.728
369774 96.596 -68.706 52.691
369775 96.638 -68.714 52.653
369776 96.678 -68.721 52.616
369777 96.719 -68.730 52.579
369778 96.759 -68.737 52.541
369779 96.800 -68.745 52.502
369780 96.841 -68.753 52.462
369781 96.882 -68.761 52.423
369782 96.922 -68.769 52.384
369783 96.963 -68.777 52.346
369784 97.004 -68.785 52.306
369785 97.039 -68.796 52.269
369786 97.073 -68.807 52.232
369787 97.108 -68.817 52.196
369788 97.140 -68.827 52.162
369789 97.167 -68.833 52.132
369790 97.193 -68.834 52.104
369791 97.218 -68.834 52.078
369792 97.240 -68.831 52.055
369793 97.261 -68.827 52.036
369794 97.284 -68.824 52.022
369795 97.305 -68.820 52.011
369796 97.327 -68.816 52.003
369797 97.353 -68.817 51.996
369798 97.384 -68.823 51.997
369799 97.415 -68.829 51.997
369800 97.446 -68.835 51.995
369801 97.476 -68.841 51.988
369802 97.507 -68.847 51.976
369803 97.538 -68.853 51.957
369804 97.571 -68.858 51.934
369805 97.602 -68.864 51.910
369806 97.635 -68.869 51.884
369807 97.668 -68.874 51.859
369808 97.702 -68.880 51.833
369809 97.736 -68.883 51.804
369810 97.770 -68.886 51.769
369811 97.804 -68.889 51.733
369812 97.839 -68.893 51.692
369813 97.873 -68.896 51.648
369814 97.908 -68.902 51.602
369815 97.944 -68.908 51.554
369816 97.981 -68.913 51.500
369817 98.017 -68.919 51.446
369818 98.053 -68.925 51.391
369819 98.089 -68.931 51.336
369820 98.126 -68.937 51.282
369821 98.161 -68.943 51.230
369822 98.198 -68.949 51.179
369823 98.242 -68.952 51.129
369824 98.287 -68.955 51.081
369825 98.331 -68.958 51.033
369826 98.375 -68.961 50.987
369827 98.419 -68.964 50.942
369828 98.460 -68.975 50.898
369829 98.501 -68.987 50.850
369830 98.541 -69.000 50.799
369831 98.582 -69.012 50.746
369832 98.622 -69.025 50.691
369833 98.663 -69.037 50.637
369834 98.703 -69.050 50.582
369835 98.732 -69.055 50.526
369836 98.741 -69.046 50.469
369837 98.750 -69.039 50.412
369838 98.760 -69.031 50.355
369839 98.769 -69.023 50.299
369840 98.778 -69.015 50.243
369841 98.812 -69.029 50.188
369842 98.855 -69.056 50.132
369843 98.910 -69.082 50.076
369844 98.966 -69.109 50.020
369845 99.021 -69.137 49.965
369846 99.076 -69.164 49.910
369847 99.132 -69.190 49.857
369848 99.158 -69.201 49.805
369849 99.182 -69.211 49.754
369850 99.206 -69.220 49.702
369851 99.230 -69.229 49.651
369852 99.257 -69.237 49.599
369853 99.285 -69.254 49.548
369854 99.306 -69.273 49.497
369855 99.336 -69.284 49.446
369856 99.366 -69.294 49.395
369857 99.397 -69.304 49.345
369858 99.427 -69.314 49.295
369859 99.478 -69.322 49.245
369860 99.529 -69.330 49.197
369861 99.581 -69.340 49.151
369862 99.630 -69.354 49.105
369863 99.679 -69.368 49.058
369864 99.728 -69.382 49.010
369865 99.777 -69.397 48.964
369866 99.826 -69.411 48.917
369867 99.874 -69.425 48.870
369868 99.924 -69.440 48.823
369869 99.973 -69.454 48.776
369870 100.021 -69.468 48.730
369871 100.070 -69.482 48.684
369872 100.120 -69.497 48.638
369873 100.152 -69.507 48.590
369874 100.177 -69.516 48.546
369875 100.208 -69.515 48.502
369876 100.240 -69.511 48.458
369877 100.260 -69.516 48.415
369878 100.279 -69.521 48.373
369879 100.298 -69.526 48.332
369880 100.317 -69.532 48.291
369881 100.351 -69.537 48.251
369882 100.388 -69.543 48.215
369883 100.424 -69.548 48.178
369884 100.461 -69.554 48.142
369885 100.498 -69.560 48.107
369886 100.535 -69.565 48.071
369887 100.571 -69.582 48.037
369888 100.608 -69.599 48.003
369889 100.651 -69.645 47.969
369890 100.684 -69.647 47.935
369891 100.714 -69.644 47.901
369892 100.745 -69.660 47.865
369893 100.777 -69.700 47.829
369894 100.807 -69.710 47.795
369895 100.836 -69.715 47.759
369896 100.865 -69.720 47.723
369897 100.894 -69.724 47.686
369898 100.923 -69.729 47.648
369899 100.952 -69.733 47.609
369900 100.989 -69.733 47.571
369901 101.029 -69.730 47.533
369902 101.068 -69.729 47.495
369903 101.112 -69.726 47.455
369904 101.160 -69.724 47.415
369905 101.203 -69.726 47.374
369906 101.242 -69.731 47.334
369907 101.281 -69.736 47.294
369908 101.320 -69.742 47.253
369909 101.358 -69.747 47.213
369910 101.397 -69.751 47.172
369911 101.436 -69.757 47.131
370330 90.883 -68.795 54.843
370331 90.914 -68.805 54.837
370332 90.945 -68.816 54.834
370333 90.976 -68.827 54.832
370334 91.008 -68.837 54.830
370335 91.045 -68.852 54.830
370336 91.103 -68.881 54.830
370337 91.161 -68.910 54.830
370338 91.220 -68.940 54.831
370339 91.278 -68.969 54.832
370340 91.336 -68.998 54.831
370341 91.394 -69.027 54.830
370342 91.446 -69.083 54.828
370343 91.497 -69.143 54.823
370344 91.548 -69.203 54.815
370345 91.599 -69.240 54.805
370346 91.639 -69.209 54.794
370347 91.685 -69.199 54.781
370348 91.730 -69.187 54.767
370349 91.775 -69.175 54.752
370350 91.819 -69.163 54.737
370351 91.864 -69.152 54.721
370352 91.909 -69.140 54.705
370353 91.953 -69.128 54.691
370354 91.998 -69.116 54.676
370366 92.533 -68.974 54.437
370367 92.578 -68.962 54.413
370368 92.623 -68.949 54.389
370369 92.667 -68.938 54.364
370370 92.712 -68.926 54.339
370371 92.757 -68.914 54.312
370372 92.801 -68.902 54.286
370373 92.846 -68.890 54.261
370374 92.891 -68.878 54.236
370375 92.935 -68.867 54.209
370376 92.980 -68.855 54.181
370377 93.017 -68.853 54.154
370378 93.052 -68.854 54.127
370379 93.087 -68.855 54.100
370380 93.122 -68.855 54.074
370381 93.157 -68.856 54.046
370382 93.192 -68.857 54.018
370418 94.734 -68.512 53.456
370419 94.779 -68.503 53.446
370420 94.825 -68.493 53.437
370421 94.868 -68.486 53.430
370422 94.889 -68.510 53.423
370423 94.909 -68.533 53.415
370424 94.930 -68.557 53.407
370425 94.951 -68.581 53.399
370426 94.976 -68.589 53.392
370427 95.006 -68.577 53.386
370428 95.037 -68.567 53.380
370429 95.067 -68.556 53.375
370430 95.098 -68.545 53.370
370431 95.131 -68.532 53.363
370432 95.169 -68.516 53.356
370433 95.207 -68.500 53.347
370434 95.246 -68.483 53.337
370435 95.280 -68.471 53.328
370436 95.307 -68.464 53.321
370437 95.341 -68.472 53.315
370438 95.377 -68.482 53.306
370439 95.416 -68.494 53.298
370440 95.462 -68.512 53.289
370441 95.508 -68.530 53.279
370442 95.557 -68.546 53.271
370443 95.607 -68.561 53.264
370444 95.658 -68.576 53.260
370445 95.708 -68.591 53.262
370446 95.759 -68.605 53.264
370447 95.809 -68.620 53.262
370448 95.848 -68.629 53.257
370449 95.875 -68.629 53.246
370450 95.901 -68.630 53.227
370451 95.927 -68.631 53.205
370452 95.953 -68.633 53.190
370453 95.979 -68.633 53.185
370454 95.998 -68.630 53.188
370455 96.016 -68.626 53.186
370456 96.044 -68.624 53.178
370457 96.073 -68.623 53.166
370458 96.101 -68.622 53.148
370459 96.129 -68.621 53.132
370460 96.158 -68.620 53.115
370461 96.186 -68.619 53.096
370462 96.215 -68.617 53.075
370463 96.243 -68.617 53.048
370464 96.271 -68.615 53.020
370465 96.300 -68.618 52.992
370466 96.329 -68.626 52.964
370467 96.359 -68.633 52.940
370468 96.388 -68.641 52.916
370469 96.418 -68.648 52.889
370470 96.447 -68.656 52.859
370471 96.477 -68.661 52.826
370472 96.507 -68.665 52.791
370473 96.537 -68.670 52.754
370474 96.567 -68.674 52.717
370475 96.600 -68.678 52.680
370476 96.642 -68.685 52.642
370477 96.682 -68.693 52.604
370478 96.723 -68.701 52.565
370479 96.763 -68.709 52.525
370480 96.804 -68.717 52.485
370481 96.845 -68.725 52.446
370482 96.886 -68.733 52.408
370483 96.927 -68.741 52.369
370484 96.967 -68.749 52.332
370485 97.008 -68.757 52.295
370486 97.048 -68.764 52.259
370487 97.085 -68.774 52.223
370488 97.119 -68.786 52.188
370489 97.152 -68.796 52.155
370490 97.179 -68.799 52.127
370491 97.205 -68.801 52.100
370492 97.226 -68.797 52.076
370493 97.248 -68.793 52.056
370494 97.270 -68.789 52.041
370495 97.292 -68.786 52.030
370496 97.313 -68.783 52.023
370497 97.335 -68.779 52.018
370498 97.361 -68.780 52.013
370499 97.392 -68.786 52.015
370500 97.423 -68.792 52.014
370501 97.454 -68.798 52.011
370502 97.485 -68.803 52.002
370503 97.516 -68.809 51.988
370504 97.546 -68.815 51.965
370505 97.577 -68.821 51.939
370506 97.609 -68.827 51.910
370507 97.642 -68.832 51.882
370508 97.676 -68.836 51.853
370509 97.710 -68.839 51.823
370510 97.744 -68.843 51.790
370511 97.779 -68.846 51.754
370512 97.813 -68.850 51.715
370513 97.847 -68.853 51.673
370514 97.881 -68.857 51.629
370515 97.916 -68.860 51.581
370516 97.950 -68.864 51.530
370517 97.987 -68.871 51.476
370518 98.023 -68.877 51.422
370519 98.059 -68.883 51.369
370520 98.095 -68.889 51.316
370521 98.132 -68.895 51.265
370522 98.168 -68.901 51.219
370523 98.207 -68.905 51.172
370524 98.251 -68.908 51.124
370525 98.295 -68.911 51.076
370526 98.340 -68.915 51.031
370527 98.383 -68.918 50.985
370528 98.424 -68.930 50.939
370529 98.464 -68.943 50.893
370530 98.505 -68.955 50.843
370531 98.546 -68.968 50.791
370532 98.586 -68.980 50.738
370533 98.626 -68.993 50.684
370534 98.667 -69.005 50.630
370535 98.707 -69.018 50.576
370536 98.738 -69.024 50.521
370537 98.747 -69.016 50.466
370538 98.757 -69.009 50.410
370539 98.766 -69.000 50.352
370540 98.775 -68.993 50.295
370541 98.793 -68.990 50.240
370542 98.841 -69.003 50.186
370543 98.875 -69.021 50.130
370544 98.892 -69.044 50.074
370545 98.929 -69.069 50.018
370546 98.984 -69.096 49.963
370547 99.039 -69.123 49.907
370548 99.082 -69.143 49.853
370549 99.114 -69.157 49.800
370550 99.146 -69.171 49.749
370551 99.178 -69.185 49.699
370552 99.210 -69.199 49.649
370553 99.259 -69.215 49.599
370554 99.288 -69.228 49.548
370555 99.319 -69.237 49.496
370556 99.349 -69.248 49.445
370557 99.379 -69.258 49.395
370558 99.409 -69.269 49.344
370559 99.439 -69.279 49.294
370560 99.482 -69.288 49.246
370561 99.534 -69.296 49.198
370562 99.586 -69.303 49.152
370563 99.638 -69.311 49.106
370564 99.689 -69.322 49.059
370565 99.738 -69.336 49.011
370566 99.787 -69.350 48.965
370567 99.836 -69.365 48.918
370568 99.885 -69.378 48.870
370569 99.934 -69.393 48.822
370570 99.983 -69.407 48.776
370571 100.031 -69.422 48.729
370572 100.080 -69.436 48.682
370573 100.119 -69.448 48.636
370574 100.145 -69.457 48.588
370575 100.171 -69.466 48.542
370576 100.199 -69.469 48.498
370577 100.231 -69.465 48.454
370578 100.263 -69.461 48.410
370579 100.286 -69.463 48.368
370580 100.305 -69.469 48.325
370581 100.324 -69.473 48.284
370582 100.359 -69.479 48.246
370583 100.396 -69.485 48.208
370584 100.433 -69.490 48.170
370585 100.470 -69.496 48.134
370586 100.506 -69.513 48.098
370587 100.543 -69.531 48.061
370588 100.579 -69.548 48.027
370589 100.623 -69.595 47.993
370590 100.653 -69.591 47.959
370591 100.684 -69.588 47.925
370592 100.715 -69.606 47.890
370593 100.748 -69.645 47.854
370594 100.778 -69.665 47.819
370595 100.807 -69.670 47.785
370596 100.836 -69.674 47.749
370597 100.865 -69.679 47.713
370598 100.894 -69.684 47.676
370599 100.923 -69.689 47.638
370600 100.959 -69.689 47.600
370601 100.998 -69.686 47.562
370602 101.038 -69.685 47.524
370603 101.078 -69.682 47.485
370604 101.126 -69.680 47.445
370605 101.174 -69.679 47.405
370606 101.213 -69.683 47.365
370607 101.252 -69.689 47.325
370608 101.291 -69.694 47.284
370609 101.330 -69.699 47.244
370610 101.369 -69.704 47.203
371031 90.871 -68.763 54.832
371032 90.902 -68.774 54.827
371033 90.933 -68.783 54.825
371034 90.964 -68.794 54.824
371035 90.996 -68.805 54.824
371036 91.038 -68.824 54.825
371037 91.097 -68.852 54.826
371038 91.155 -68.882 54.827
371039 91.214 -68.911 54.828
371040 91.272 -68.940 54.831
371041 91.330 -68.970 54.831
371042 91.388 -68.999 54.830
371043 91.440 -69.058 54.828
371044 91.490 -69.118 54.823
371045 91.541 -69.178 54.815
371046 91.590 -69.209 54.805
371047 91.634 -69.190 54.793
371048 91.678 -69.178 54.780
371049 91.723 -69.166 54.766
371050 91.768 -69.155 54.750
371051 91.812 -69.143 54.734
371052 91.857 -69.131 54.718
371053 91.901 -69.119 54.703
371054 91.946 -69.107 54.688
371055 91.991 -69.095 54.673
371065 92.437 -68.977 54.487
371066 92.482 -68.965 54.464
371067 92.527 -68.953 54.441
371068 92.571 -68.941 54.418
371069 92.616 -68.929 54.395
371070 92.661 -68.918 54.369
371071 92.705 -68.905 54.345
371072 92.750 -68.893 54.319
371073 92.795 -68.882 54.294
371074 92.839 -68.870 54.270
371075 92.884 -68.858 54.244
371076 92.929 -68.846 54.218
371077 92.973 -68.834 54.192
371078 93.011 -68.831 54.165
371079 93.046 -68.832 54.138
371080 93.081 -68.833 54.111
371081 93.116 -68.833 54.084
371082 93.151 -68.834 54.057
371083 93.186 -68.835 54.029
371084 93.221 -68.836 54.001
371119 94.744 -68.485 53.455
371120 94.790 -68.475 53.445
371121 94.835 -68.466 53.437
371122 94.866 -68.476 53.430
371123 94.887 -68.499 53.421
371124 94.907 -68.523 53.413
371125 94.928 -68.547 53.405
371126 94.950 -68.564 53.397
371127 94.981 -68.553 53.389
371128 95.011 -68.542 53.382
371129 95.042 -68.531 53.375
371130 95.072 -68.520 53.370
371131 95.103 -68.509 53.364
371132 95.133 -68.498 53.357
371133 95.170 -68.483 53.347
371134 95.208 -68.466 53.338
371135 95.247 -68.451 53.330
371136 95.283 -68.436 53.321
371137 95.310 -68.429 53.314
371138 95.337 -68.423 53.308
371139 95.371 -68.429 53.299
371140 95.409 -68.441 53.290
371141 95.455 -68.460 53.281
371142 95.501 -68.477 53.270
371143 95.547 -68.495 53.262
371144 95.593 -68.513 53.256
371145 95.639 -68.532 53.253
371146 95.686 -68.548 53.253
371147 95.736 -68.564 53.252
371148 95.787 -68.578 53.248
371149 95.837 -68.593 53.242
371150 95.882 -68.604 53.232
371151 95.908 -68.606 53.213
371152 95.935 -68.607 53.196
371153 95.960 -68.607 53.189
371154 95.982 -68.605 53.193
371155 96.004 -68.602 53.200
371156 96.032 -68.601 53.198
371157 96.060 -68.600 53.189
371158 96.089 -68.599 53.175
371159 96.117 -68.598 53.156
371160 96.146 -68.597 53.139
371161 96.174 -68.595 53.121
371162 96.202 -68.595 53.102
371163 96.230 -68.593 53.081
371164 96.259 -68.592 53.055
371165 96.287 -68.591 53.024
371166 96.315 -68.590 52.991
371167 96.345 -68.598 52.960
371168 96.374 -68.605 52.933
371169 96.404 -68.613 52.908
371170 96.433 -68.620 52.881
371171 96.463 -68.628 52.850
371172 96.493 -68.632 52.817
371173 96.523 -68.636 52.781
371174 96.552 -68.641 52.744
371175 96.583 -68.645 52.706
371176 96.612 -68.649 52.670
371177 96.641 -68.651 52.632
371178 96.683 -68.661 52.593
371179 96.726 -68.671 52.552
371180 96.768 -68.680 52.512
371181 96.808 -68.688 52.472
371182 96.849 -68.696 52.433
371183 96.890 -68.704 52.394
371184 96.931 -68.712 52.357
371185 96.971 -68.720 52.320
371186 97.012 -68.728 52.283
371187 97.052 -68.736 52.247
371188 97.093 -68.744 52.213
371189 97.131 -68.753 52.179
371190 97.165 -68.764 52.148
371191 97.191 -68.767 52.124
371192 97.213 -68.763 52.100
371193 97.235 -68.760 52.079
371194 97.256 -68.756 52.062
371195 97.278 -68.752 52.050
371196 97.300 -68.749 52.042
371197 97.321 -68.745 52.038
371198 97.344 -68.742 52.035
371199 97.369 -68.742 52.033
371200 97.400 -68.748 52.034
371201 97.431 -68.754 52.032
371202 97.462 -68.760 52.026
371203 97.493 -68.766 52.015
371204 97.524 -68.772 51.996
371205 97.555 -68.778 51.971
371206 97.585 -68.784 51.943
371207 97.616 -68.789 51.912
371208 97.650 -68.793 51.881
371209 97.684 -68.797 51.848
371210 97.719 -68.800 51.815
371211 97.753 -68.804 51.778
371212 97.787 -68.807 51.738
371213 97.821 -68.811 51.697
371214 97.856 -68.814 51.654
371215 97.889 -68.817 51.608
371216 97.924 -68.821 51.559
371217 97.958 -68.824 51.509
371218 97.993 -68.828 51.455
371219 98.029 -68.834 51.403
371220 98.065 -68.840 51.351
371221 98.101 -68.846 51.301
371222 98.138 -68.852 51.254
371223 98.174 -68.858 51.209
371224 98.215 -68.862 51.163
371225 98.260 -68.865 51.116
371226 98.304 -68.868 51.069
371227 98.347 -68.874 51.023
371228 98.387 -68.886 50.976
371229 98.428 -68.899 50.928
371230 98.468 -68.911 50.882
371231 98.509 -68.924 50.832
371232 98.550 -68.936 50.779
371233 98.590 -68.949 50.725
371234 98.630 -68.962 50.672
371235 98.671 -68.974 50.618
371236 98.711 -68.987 50.564
371237 98.744 -68.994 50.509
371238 98.754 -68.986 50.455
371239 98.763 -68.978 50.400
371240 98.772 -68.970 50.343
371241 98.781 -68.962 50.288
371242 98.812 -68.966 50.233
371243 98.859 -68.980 50.178
371244 98.907 -68.993 50.123
371245 98.938 -69.012 50.068
371246 98.955 -69.035 50.014
371247 98.972 -69.059 49.960
371248 98.998 -69.081 49.905
371249 99.048 -69.097 49.852
371250 99.100 -69.115 49.799
371251 99.152 -69.131 49.747
371252 99.204 -69.149 49.697
371253 99.242 -69.167 49.648
371254 99.265 -69.185 49.599
371255 99.293 -69.193 49.549
371256 99.331 -69.202 49.498
371257 99.361 -69.212 49.447
371258 99.391 -69.223 49.396
371259 99.421 -69.233 49.346
371260 99.451 -69.244 49.296
371261 99.486 -69.253 49.247
371262 99.539 -69.261 49.200
371263 99.590 -69.269 49.153
371264 99.642 -69.277 49.107
371265 99.695 -69.284 49.060
371266 99.747 -69.292 49.014
371267 99.797 -69.304 48.967
371268 99.846 -69.318 48.919
371269 99.895 -69.332 48.871
371270 99.943 -69.347 48.825
371271 99.993 -69.361 48.777
371272 100.042 -69.375 48.729
371273 100.086 -69.388 48.682
371274 100.112 -69.397 48.635
371275 100.137 -69.407 48.586
371276 100.163 -69.416 48.540
371277 100.191 -69.422 48.495
371278 100.222 -69.418 48.450
371279 100.254 -69.414 48.407
371280 100.285 -69.410 48.363
371281 100.312 -69.410 48.320
371282 100.331 -69.416 48.278
371283 100.368 -69.421 48.240
371284 100.405 -69.427 48.201
371285 100.441 -69.444 48.163
371286 100.477 -69.462 48.126
371287 100.514 -69.479 48.090
371288 100.559 -69.529 48.053
371289 100.594 -69.540 48.018
371290 100.624 -69.536 47.984
371291 100.654 -69.533 47.950
371292 100.686 -69.551 47.916
371293 100.717 -69.591 47.881
371294 100.749 -69.620 47.845
371295 100.778 -69.625 47.810
371296 100.807 -69.629 47.776
371297 100.836 -69.634 47.741
371298 100.865 -69.639 47.704
371299 100.894 -69.643 47.667
371300 100.928 -69.645 47.629
371301 100.968 -69.642 47.591
371302 101.008 -69.641 47.553
371303 101.047 -69.639 47.515
371304 101.093 -69.636 47.476
371305 101.140 -69.635 47.436
371306 101.185 -69.635 47.397
371307 101.223 -69.641 47.356
371308 101.262 -69.646 47.316
371309 101.301 -69.651 47.275
371731 90.828 -68.720 54.827
371732 90.859 -68.730 54.822
371733 90.890 -68.741 54.818
371734 90.922 -68.752 54.816
371735 90.952 -68.762 54.816
371736 90.983 -68.773 54.818
371737 91.032 -68.796 54.821
371738 91.091 -68.824 54.823
371739 91.149 -68.854 54.825
371740 91.208 -68.883 54.828
371741 91.266 -68.912 54.831
371742 91.324 -68.942 54.832
371743 91.382 -68.974 54.831
371744 91.433 -69.034 54.830
371745 91.484 -69.093 54.824
371746 91.534 -69.153 54.815
371747 91.582 -69.181 54.803
371748 91.627 -69.170 54.791
371749 91.671 -69.158 54.778
371750 91.716 -69.146 54.763
371751 91.761 -69.134 54.747
371752 91.805 -69.122 54.731
371753 91.850 -69.110 54.714
371754 91.895 -69.099 54.699
371755 91.939 -69.087 54.685
371756 91.984 -69.074 54.670
371765 92.386 -68.968 54.512
371766 92.431 -68.956 54.490
371767 92.475 -68.944 54.468
371768 92.520 -68.932 54.445
371769 92.565 -68.921 54.423
371770 92.609 -68.908 54.401
371771 92.654 -68.896 54.376
371772 92.699 -68.885 54.353
371773 92.743 -68.873 54.328
371774 92.788 -68.861 54.304
371775 92.833 -68.849 54.280
371776 92.877 -68.837 54.255
371777 92.922 -68.825 54.229
371778 92.967 -68.814 54.202
371779 93.006 -68.808 54.176
371780 93.041 -68.809 54.149
371781 93.076 -68.810 54.123
371782 93.111 -68.811 54.096
371783 93.146 -68.811 54.068
371784 93.181 -68.813 54.040
371785 93.215 -68.814 54.012
371820 94.755 -68.457 53.454
371821 94.801 -68.448 53.444
371822 94.843 -68.441 53.436
371823 94.864 -68.465 53.428
371824 94.885 -68.488 53.420
371825 94.905 -68.513 53.411
371826 94.926 -68.536 53.403
371827 94.956 -68.528 53.394
371828 94.986 -68.517 53.386
371829 95.017 -68.506 53.378
371830 95.047 -68.495 53.370
371831 95.077 -68.484 53.364
371832 95.108 -68.473 53.357
371833 95.138 -68.462 53.349
371834 95.171 -68.450 53.340
371835 95.209 -68.434 53.332
371836 95.248 -68.417 53.324
371837 95.286 -68.402 53.315
371838 95.313 -68.395 53.306
371839 95.339 -68.388 53.299
371840 95.366 -68.382 53.290
371841 95.402 -68.389 53.282
371842 95.448 -68.407 53.273
371843 95.493 -68.425 53.263
371844 95.540 -68.443 53.256
371845 95.586 -68.461 53.250
371846 95.631 -68.479 53.245
371847 95.678 -68.497 53.241
371848 95.724 -68.515 53.237
371849 95.769 -68.532 53.231
371850 95.816 -68.551 53.224
371851 95.865 -68.566 53.215
371852 95.916 -68.581 53.198
371853 95.942 -68.582 53.185
371854 95.965 -68.581 53.183
371855 95.991 -68.579 53.191
371856 96.020 -68.578 53.201
371857 96.048 -68.577 53.200
371858 96.076 -68.576 53.191
371859 96.105 -68.575 53.177
371860 96.133 -68.573 53.158
371861 96.161 -68.573 53.139
371862 96.189 -68.571 53.121
371863 96.218 -68.570 53.102
371864 96.246 -68.569 53.082
371865 96.274 -68.568 53.056
371866 96.303 -68.567 53.023
371867 96.331 -68.566 52.990
371868 96.360 -68.570 52.957
371869 96.390 -68.577 52.928
371870 96.419 -68.585 52.902
371871 96.449 -68.592 52.873
371872 96.478 -68.599 52.842
371873 96.508 -68.603 52.808
371874 96.538 -68.607 52.771
371875 96.568 -68.612 52.733
371876 96.598 -68.617 52.695
371877 96.624 -68.620 52.660
371878 96.650 -68.622 52.621
371879 96.678 -68.620 52.581
371880 96.712 -68.622 52.541
371881 96.754 -68.633 52.500
371882 96.797 -68.642 52.461
371883 96.840 -68.653 52.421
371884 96.883 -68.664 52.383
371885 96.926 -68.673 52.346
371886 96.969 -68.684 52.309
371887 97.012 -68.695 52.271
371888 97.054 -68.705 52.237
371889 97.097 -68.715 52.203
371890 97.138 -68.723 52.172
371891 97.177 -68.732 52.144
371892 97.199 -68.730 52.123
371893 97.221 -68.726 52.102
371894 97.243 -68.722 52.084
371895 97.264 -68.718 52.070
371896 97.286 -68.715 52.061
371897 97.308 -68.711 52.056
371898 97.330 -68.708 52.054
371899 97.352 -68.704 52.054
371900 97.377 -68.705 52.053
371901 97.408 -68.711 52.053
371902 97.439 -68.717 52.049
371903 97.470 -68.723 52.040
371904 97.501 -68.728 52.026
371905 97.532 -68.734 52.005
371906 97.563 -68.740 51.976
371907 97.594 -68.746 51.945
371908 97.626 -68.752 51.913
371909 97.659 -68.755 51.879
371910 97.693 -68.758 51.843
371911 97.727 -68.761 51.807
371912 97.761 -68.764 51.767
371913 97.796 -68.767 51.725
371914 97.830 -68.770 51.681
371915 97.864 -68.774 51.636
371916 97.898 -68.777 51.589
371917 97.932 -68.781 51.540
371918 97.967 -68.784 51.490
371919 98.001 -68.788 51.438
371920 98.035 -68.792 51.387
371921 98.071 -68.797 51.337
371922 98.107 -68.803 51.289
371923 98.144 -68.809 51.243
371924 98.180 -68.815 51.199
371925 98.224 -68.818 51.153
371926 98.268 -68.821 51.106
371927 98.310 -68.830 51.059
371928 98.351 -68.842 51.012
371929 98.391 -68.855 50.965
371930 98.432 -68.867 50.917
371931 98.473 -68.880 50.870
371932 98.513 -68.892 50.819
371933 98.554 -68.905 50.766
371934 98.594 -68.917 50.712
371935 98.635 -68.930 50.659
371936 98.675 -68.942 50.605
371937 98.715 -68.955 50.551
371938 98.751 -68.964 50.497
371939 98.760 -68.955 50.442
371940 98.769 -68.948 50.388
371941 98.779 -68.940 50.333
371942 98.788 -68.932 50.279
371943 98.831 -68.943 50.224
371944 98.878 -68.956 50.169
371945 98.925 -68.970 50.114
371946 98.973 -68.983 50.060
371947 99.001 -69.002 50.007
371948 99.017 -69.026 49.954
371949 99.044 -69.047 49.901
371950 99.096 -69.065 49.848
371951 99.148 -69.082 49.796
371952 99.200 -69.099 49.745
371953 99.226 -69.118 49.695
371954 99.249 -69.137 49.648
371955 99.269 -69.153 49.600
371956 99.290 -69.162 49.551
371957 99.338 -69.168 49.500
371958 99.373 -69.177 49.449
371959 99.404 -69.187 49.398
371960 99.434 -69.197 49.349
371961 99.464 -69.208 49.300
371962 99.494 -69.219 49.250
371963 99.543 -69.226 49.202
371964 99.595 -69.234 49.155
371965 99.647 -69.242 49.109
371966 99.699 -69.250 49.062
371967 99.751 -69.258 49.016
371968 99.803 -69.266 48.969
371969 99.855 -69.273 48.920
371970 99.905 -69.286 48.873
371971 99.954 -69.300 48.827
371972 100.003 -69.315 48.779
371973 100.052 -69.328 48.730
371974 100.079 -69.338 48.682
371975 100.105 -69.347 48.635
371976 100.130 -69.357 48.586
371977 100.155 -69.366 48.539
371978 100.182 -69.375 48.493
371979 100.213 -69.371 48.447
371980 100.245 -69.368 48.403
371981 100.276 -69.364 48.358
371982 100.308 -69.360 48.314
371983 100.339 -69.358 48.273
371984 100.376 -69.375 48.234
371985 100.412 -69.393 48.194
371986 100.449 -69.410 48.156
371987 100.485 -69.427 48.119
371988 100.533 -69.488 48.082
371989 100.564 -69.485 48.045
371990 100.594 -69.481 48.010
371991 100.624 -69.478 47.976
371992 100.655 -69.497 47.943
371993 100.688 -69.535 47.908
371994 100.720 -69.575 47.874
371995 100.749 -69.580 47.837
371996 100.778 -69.585 47.802
371997 100.807 -69.589 47.768
371998 100.836 -69.594 47.732
371999 100.865 -69.598 47.696
372000 100.898 -69.601 47.659
372001 100.937 -69.598 47.621
372002 100.977 -69.597 47.583
372003 101.017 -69.595 47.545
372004 101.058 -69.592 47.507
372005 101.107 -69.591 47.468
372006 101.155 -69.588 47.429
372007 101.195 -69.593 47.390
372008 101.234 -69.598 47.348
372432 90.816 -68.687 54.815
372433 90.847 -68.698 54.812
372434 90.878 -68.708 54.810
372435 90.909 -68.719 54.809
372436 90.941 -68.730 54.810
372437 90.971 -68.740 54.812
372438 91.026 -68.767 54.817
372439 91.085 -68.796 54.820
372440 91.143 -68.826 54.824
372441 91.202 -68.855 54.827
372442 91.260 -68.884 54.832
372443 91.318 -68.914 54.834
372444 91.375 -68.949 54.834
372445 91.426 -69.009 54.832
372446 91.476 -69.068 54.825
372447 91.527 -69.128 54.815
372448 91.576 -69.148 54.802
372449 91.620 -69.143 54.789
372450 91.665 -69.137 54.775
372451 91.709 -69.125 54.759
372452 91.754 -69.113 54.743
372453 91.799 -69.102 54.727
372454 91.843 -69.090 54.710
372455 91.888 -69.078 54.695
372456 91.933 -69.066 54.681
372457 91.977 -69.054 54.667
372464 92.290 -68.971 54.556
372465 92.335 -68.959 54.536
372466 92.379 -68.947 54.515
372467 92.424 -68.935 54.493
372468 92.469 -68.924 54.472
372469 92.513 -68.911 54.451
372470 92.558 -68.899 54.429
372471 92.602 -68.888 54.407
372472 92.647 -68.876 54.384
372473 92.692 -68.864 54.360
372474 92.736 -68.852 54.337
372475 92.781 -68.840 54.314
372476 92.826 -68.828 54.291
372477 92.870 -68.817 54.267
372478 92.915 -68.805 54.241
372479 92.959 -68.793 54.214
372480 93.000 -68.786 54.188
372481 93.035 -68.787 54.161
372482 93.070 -68.788 54.135
372483 93.105 -68.789 54.108
372484 93.140 -68.789 54.080
372485 93.175 -68.790 54.051
372486 93.210 -68.791 54.023
372487 93.245 -68.792 53.994
372521 94.765 -68.430 53.453
372522 94.811 -68.420 53.443
372523 94.841 -68.431 53.435
372524 94.862 -68.454 53.427
372525 94.883 -68.478 53.418
372526 94.903 -68.502 53.409
372527 94.930 -68.503 53.400
372528 94.961 -68.492 53.391
372529 94.991 -68.481 53.382
372530 95.022 -68.470 53.374
372531 95.052 -68.459 53.365
372532 95.083 -68.448 53.358
372533 95.113 -68.437 53.351
372534 95.144 -68.426 53.343
372535 95.174 -68.415 53.335
372536 95.210 -68.401 53.327
372537 95.249 -68.385 53.318
372538 95.287 -68.369 53.309
372539 95.315 -68.360 53.300
372540 95.342 -68.354 53.291
372541 95.369 -68.347 53.282
372542 95.398 -68.346 53.273
372543 95.440 -68.357 53.265
372544 95.485 -68.375 53.257
372545 95.532 -68.393 53.250
372546 95.578 -68.410 53.245
372547 95.623 -68.429 53.239
372548 95.670 -68.446 53.233
372549 95.715 -68.464 53.225
372550 95.762 -68.482 53.217
372551 95.808 -68.500 53.208
372552 95.854 -68.517 53.197
372553 95.900 -68.535 53.183
372554 95.946 -68.553 53.174
372555 95.979 -68.556 53.173
372556 96.007 -68.555 53.180
372557 96.036 -68.554 53.189
372558 96.064 -68.553 53.189
372559 96.092 -68.551 53.183
372560 96.120 -68.551 53.170
372561 96.149 -68.549 53.152
372562 96.177 -68.548 53.133
372563 96.205 -68.547 53.114
372564 96.234 -68.546 53.094
372565 96.262 -68.545 53.073
372566 96.291 -68.544 53.048
372567 96.319 -68.542 53.016
372568 96.347 -68.542 52.982
372569 96.376 -68.542 52.948
372570 96.405 -68.549 52.921
372571 96.435 -68.557 52.895
372572 96.464 -68.564 52.866
372573 96.494 -68.570 52.835
372574 96.524 -68.574 52.799
372575 96.554 -68.579 52.761
372576 96.584 -68.583 52.723
372577 96.610 -68.586 52.685
372578 96.636 -68.589 52.648
372579 96.661 -68.592 52.609
372580 96.689 -68.592 52.570
372581 96.717 -68.590 52.530
372582 96.746 -68.589 52.490
372583 96.783 -68.594 52.451
372584 96.825 -68.604 52.413
372585 96.868 -68.614 52.375
372586 96.911 -68.625 52.338
372587 96.954 -68.635 52.301
372588 96.997 -68.645 52.263
372589 97.040 -68.656 52.229
372590 97.085 -68.667 52.197
372591 97.134 -68.683 52.168
372592 97.183 -68.698 52.144
372593 97.207 -68.693 52.125
372594 97.229 -68.689 52.106
372595 97.251 -68.685 52.090
372596 97.273 -68.681 52.079
372597 97.295 -68.677 52.073
372598 97.316 -68.674 52.072
372599 97.338 -68.670 52.073
372600 97.360 -68.667 52.074
372601 97.385 -68.667 52.073
372602 97.416 -68.673 52.071
372603 97.447 -68.679 52.064
372604 97.478 -68.685 52.052
372605 97.509 -68.691 52.036
372606 97.540 -68.697 52.013
372607 97.572 -68.702 51.981
372608 97.604 -68.708 51.946
372609 97.636 -68.714 51.911
372610 97.669 -68.717 51.875
372611 97.703 -68.720 51.837
372612 97.736 -68.723 51.798
372613 97.770 -68.725 51.758
372614 97.804 -68.728 51.714
372615 97.839 -68.731 51.668
372616 97.872 -68.735 51.620
372617 97.907 -68.738 51.572
372618 97.941 -68.741 51.523
372619 97.975 -68.745 51.475
372620 98.009 -68.749 51.425
372621 98.044 -68.752 51.375
372622 98.078 -68.755 51.327
372623 98.114 -68.761 51.279
372624 98.150 -68.767 51.232
372625 98.189 -68.772 51.188
372626 98.233 -68.775 51.143
372627 98.274 -68.785 51.094
372628 98.314 -68.798 51.048
372629 98.355 -68.810 50.999
372630 98.395 -68.823 50.951
372631 98.436 -68.835 50.904
372632 98.477 -68.848 50.855
372633 98.517 -68.860 50.804
372634 98.558 -68.873 50.751
372635 98.598 -68.885 50.698
372636 98.639 -68.898 50.645
372637 98.679 -68.910 50.591
372638 98.719 -68.923 50.537
372639 98.757 -68.933 50.483
372640 98.767 -68.925 50.429
372641 98.776 -68.918 50.376
372642 98.785 -68.909 50.322
372643 98.803 -68.907 50.267
372644 98.850 -68.920 50.212
372645 98.897 -68.933 50.157
372646 98.944 -68.946 50.103
372647 98.991 -68.959 50.051
372648 99.039 -68.973 49.998
372649 99.064 -68.993 49.946
372650 99.091 -69.015 49.894
372651 99.144 -69.032 49.843
372652 99.187 -69.049 49.792
372653 99.210 -69.068 49.743
372654 99.232 -69.087 49.695
372655 99.255 -69.107 49.649
372656 99.274 -69.120 49.602
372657 99.290 -69.131 49.553
372658 99.335 -69.137 49.503
372659 99.382 -69.143 49.452
372660 99.416 -69.152 49.402
372661 99.446 -69.162 49.353
372662 99.476 -69.172 49.304
372663 99.506 -69.183 49.255
372664 99.547 -69.192 49.204
372665 99.599 -69.200 49.157
372666 99.651 -69.208 49.111
372667 99.703 -69.215 49.065
372668 99.755 -69.223 49.018
372669 99.807 -69.231 48.971
372670 99.859 -69.239 48.923
372671 99.911 -69.247 48.875
372672 99.964 -69.254 48.829
372673 100.012 -69.268 48.781
372674 100.046 -69.279 48.733
372675 100.071 -69.288 48.684
372676 100.097 -69.297 48.636
372677 100.123 -69.306 48.587
372678 100.148 -69.316 48.539
372679 100.174 -69.325 48.492
372680 100.204 -69.325 48.446
372681 100.236 -69.321 48.400
372682 100.267 -69.317 48.355
372683 100.305 -69.316 48.310
372684 100.347 -69.324 48.268
372685 100.384 -69.341 48.228
372686 100.420 -69.358 48.188
372687 100.467 -69.413 48.150
372688 100.503 -69.432 48.112
372689 100.534 -69.429 48.076
372690 100.564 -69.425 48.038
372691 100.595 -69.422 48.004
372692 100.626 -69.442 47.970
372693 100.658 -69.481 47.935
372694 100.690 -69.520 47.901
372695 100.720 -69.535 47.865
372696 100.749 -69.540 47.830
372697 100.778 -69.544 47.794
372698 100.808 -69.549 47.760
372699 100.836 -69.554 47.725
372700 100.867 -69.557 47.688
372701 100.907 -69.555 47.652
372702 100.947 -69.553 47.614
372703 100.986 -69.551 47.576
372704 101.026 -69.548 47.539
372705 101.073 -69.547 47.501
372706 101.121 -69.544 47.462
372707 101.166 -69.544 47.422
373132 90.773 -68.645 54.809
373133 90.804 -68.655 54.804
373134 90.835 -68.666 54.802
373135 90.866 -68.676 54.801
373136 90.897 -68.687 54.802
373137 90.929 -68.697 54.804
373138 90.962 -68.710 54.807
373139 91.020 -68.739 54.812
373140 91.079 -68.768 54.818
373141 91.137 -68.798 54.823
373142 91.195 -68.827 54.828
373143 91.254 -68.856 54.833
373144 91.312 -68.885 54.837
373145 91.368 -68.924 54.838
373146 91.419 -68.984 54.835
373147 91.469 -69.044 54.826
373148 91.520 -69.092 54.814
373149 91.570 -69.090 54.801
373150 91.613 -69.090 54.786
373151 91.658 -69.086 54.771
373152 91.702 -69.080 54.754
373153 91.747 -69.074 54.738
373154 91.792 -69.068 54.722
373155 91.837 -69.063 54.705
373156 91.881 -69.057 54.690
373157 91.926 -69.046 54.677
373158 91.970 -69.034 54.663
373159 92.015 -69.021 54.650
373164 92.238 -68.962 54.574
373165 92.283 -68.950 54.556
373166 92.328 -68.939 54.537
373167 92.372 -68.927 54.518
373168 92.417 -68.915 54.497
373169 92.461 -68.903 54.477
373170 92.506 -68.891 54.456
373171 92.551 -68.879 54.435
373172 92.595 -68.868 54.415
373173 92.640 -68.855 54.392
373174 92.685 -68.843 54.370
373175 92.729 -68.832 54.348
373176 92.774 -68.820 54.326
373177 92.819 -68.808 54.303
373178 92.863 -68.796 54.279
373179 92.908 -68.784 54.253
373180 92.953 -68.772 54.227
373181 92.995 -68.764 54.200
373182 93.029 -68.764 54.174
373183 93.065 -68.765 54.147
373184 93.099 -68.766 54.120
373185 93.135 -68.767 54.092
373186 93.169 -68.768 54.064
373187 93.205 -68.769 54.034
373188 93.239 -68.770 54.005
373189 93.274 -68.770 53.977
373222 94.776 -68.402 53.453
373223 94.819 -68.397 53.443
373224 94.839 -68.420 53.434
373225 94.860 -68.444 53.425
373226 94.881 -68.468 53.417
373227 94.905 -68.478 53.407
373228 94.935 -68.467 53.397
373229 94.966 -68.456 53.388
373230 94.996 -68.445 53.378
373231 95.027 -68.434 53.369
373232 95.057 -68.423 53.360
373233 95.088 -68.412 53.353
373234 95.118 -68.401 53.346
373235 95.149 -68.390 53.340
373236 95.180 -68.379 53.332
373237 95.211 -68.368 53.323
373238 95.249 -68.352 53.314
373239 95.288 -68.335 53.304
373240 95.318 -68.326 53.295
373241 95.345 -68.319 53.285
373242 95.372 -68.313 53.275
373243 95.402 -68.318 53.266
373244 95.443 -68.329 53.259
373245 95.483 -68.340 53.252
373246 95.524 -68.350 53.245
373247 95.568 -68.365 53.240
373248 95.614 -68.383 53.234
373249 95.660 -68.401 53.226
373250 95.706 -68.419 53.216
373251 95.752 -68.436 53.205
373252 95.798 -68.454 53.193
373253 95.844 -68.472 53.182
373254 95.890 -68.490 53.171
373255 95.936 -68.507 53.165
373256 95.974 -68.515 53.164
373257 96.011 -68.520 53.168
373258 96.047 -68.526 53.174
373259 96.080 -68.529 53.175
373260 96.108 -68.527 53.171
373261 96.136 -68.526 53.161
373262 96.165 -68.525 53.146
373263 96.193 -68.524 53.127
373264 96.222 -68.523 53.107
373265 96.250 -68.522 53.086
373266 96.278 -68.520 53.063
373267 96.306 -68.520 53.037
373268 96.335 -68.518 53.006
373269 96.363 -68.517 52.972
373270 96.391 -68.516 52.939
373271 96.420 -68.521 52.912
373272 96.450 -68.529 52.885
373273 96.479 -68.536 52.857
373274 96.510 -68.542 52.826
373275 96.539 -68.546 52.791
373276 96.569 -68.550 52.752
373277 96.596 -68.554 52.713
373278 96.621 -68.557 52.674
373279 96.647 -68.560 52.636
373280 96.673 -68.563 52.598
373281 96.699 -68.564 52.559
373282 96.728 -68.562 52.520
373283 96.756 -68.560 52.481
373284 96.785 -68.558 52.443
373285 96.813 -68.557 52.407
373286 96.854 -68.565 52.370
373287 96.896 -68.576 52.332
373288 96.942 -68.588 52.295
373289 96.991 -68.603 52.256
373290 97.040 -68.618 52.223
373291 97.089 -68.633 52.192
373292 97.138 -68.648 52.166
373293 97.187 -68.664 52.144
373294 97.214 -68.661 52.127
373295 97.236 -68.655 52.111
373296 97.259 -68.649 52.099
373297 97.281 -68.644 52.092
373298 97.303 -68.640 52.090
373299 97.324 -68.636 52.091
373300 97.346 -68.633 52.094
373301 97.368 -68.629 52.096
373302 97.393 -68.629 52.094
373303 97.424 -68.636 52.088
373304 97.455 -68.642 52.077
373305 97.486 -68.647 52.061
373306 97.518 -68.653 52.040
373307 97.549 -68.658 52.012
373308 97.582 -68.664 51.978
373309 97.613 -68.670 51.942
373310 97.646 -68.675 51.906
373311 97.678 -68.680 51.868
373312 97.712 -68.683 51.830
373313 97.746 -68.685 51.792
373314 97.780 -68.687 51.750
373315 97.813 -68.690 51.706
373316 97.847 -68.692 51.658
373317 97.881 -68.695 51.610
373318 97.915 -68.698 51.562
373319 97.949 -68.701 51.515
373320 97.984 -68.705 51.467
373321 98.018 -68.708 51.417
373322 98.052 -68.712 51.368
373323 98.087 -68.716 51.319
373324 98.121 -68.719 51.271
373325 98.156 -68.724 51.223
373326 98.197 -68.728 51.177
373327 98.237 -68.741 51.130
373328 98.278 -68.753 51.080
373329 98.318 -68.766 51.032
373330 98.359 -68.778 50.983
373331 98.400 -68.791 50.935
373332 98.440 -68.803 50.887
373333 98.481 -68.816 50.839
373334 98.521 -68.828 50.786
373335 98.562 -68.841 50.733
373336 98.602 -68.853 50.681
373337 98.643 -68.866 50.629
373338 98.683 -68.878 50.576
373339 98.723 -68.891 50.523
373340 98.764 -68.903 50.470
373341 98.773 -68.895 50.417
373342 98.782 -68.887 50.363
373343 98.791 -68.879 50.309
373344 98.822 -68.883 50.255
373345 98.869 -68.896 50.202
373346 98.916 -68.910 50.148
373347 98.963 -68.923 50.095
373348 99.010 -68.936 50.044
373349 99.057 -68.949 49.992
373350 99.104 -68.963 49.940
373351 99.140 -68.982 49.890
373352 99.170 -69.000 49.841
373353 99.193 -69.020 49.790
373354 99.216 -69.039 49.740
373355 99.239 -69.058 49.692
373356 99.262 -69.077 49.647
373357 99.278 -69.087 49.602
373358 99.295 -69.098 49.555
373359 99.331 -69.106 49.506
373360 99.379 -69.111 49.456
373361 99.427 -69.117 49.407
373362 99.458 -69.127 49.357
373363 99.488 -69.137 49.307
373364 99.518 -69.147 49.258
373365 99.551 -69.157 49.209
373366 99.604 -69.165 49.162
373367 99.655 -69.173 49.115
373368 99.708 -69.181 49.069
373369 99.760 -69.189 49.022
373370 99.811 -69.197 48.975
373371 99.864 -69.204 48.927
373372 99.916 -69.212 48.879
373373 99.968 -69.220 48.832
373374 100.017 -69.226 48.785
373375 100.043 -69.232 48.736
373376 100.065 -69.239 48.687
373377 100.089 -69.247 48.638
373378 100.115 -69.256 48.589
373379 100.141 -69.266 48.540
373380 100.167 -69.275 48.491
373381 100.195 -69.278 48.444
373382 100.227 -69.274 48.398
373383 100.270 -69.275 48.351
373384 100.316 -69.278 48.306
373385 100.355 -69.289 48.263
373386 100.392 -69.306 48.223
373387 100.443 -69.380 48.183
373388 100.474 -69.377 48.144
373389 100.504 -69.374 48.107
373390 100.534 -69.370 48.070
373391 100.565 -69.367 48.032
373392 100.596 -69.388 47.998
373393 100.628 -69.426 47.963
373394 100.661 -69.466 47.928
373395 100.691 -69.490 47.894
373396 100.720 -69.494 47.858
373397 100.749 -69.499 47.823
373398 100.778 -69.504 47.787
373399 100.808 -69.509 47.753
373400 100.837 -69.513 47.718
373401 100.877 -69.511 47.682
373402 100.916 -69.509 47.645
373403 100.956 -69.507 47.608
373404 100.996 -69.504 47.570
373405 101.039 -69.503 47.533
373406 101.088 -69.501 47.495
373833 90.760 -68.612 54.799
373834 90.792 -68.623 54.795
373835 90.823 -68.633 54.793
373836 90.854 -68.644 54.793
373837 90.885 -68.654 54.795
373838 90.917 -68.665 54.798
373839 90.956 -68.681 54.804
373840 91.014 -68.711 54.810
373841 91.073 -68.740 54.816
373842 91.131 -68.769 54.822
373843 91.189 -68.799 54.830
373844 91.248 -68.828 54.837
373845 91.306 -68.857 54.842
373846 91.361 -68.899 54.843
373847 91.412 -68.959 54.838
373848 91.462 -69.019 54.826
373849 91.513 -69.034 54.812
373850 91.563 -69.033 54.797
373851 91.608 -69.032 54.782
373852 91.651 -69.033 54.766
373853 91.695 -69.028 54.749
373854 91.740 -69.023 54.732
373855 91.785 -69.017 54.715
373856 91.829 -69.012 54.700
373857 91.874 -69.005 54.685
373858 91.919 -69.000 54.672
373859 91.963 -68.994 54.659
373860 92.008 -68.989 54.646
373864 92.187 -68.953 54.588
373865 92.231 -68.942 54.573
373866 92.276 -68.930 54.556
373867 92.321 -68.918 54.539
373868 92.365 -68.906 54.520
373869 92.410 -68.894 54.502
373870 92.455 -68.882 54.481
373871 92.499 -68.871 54.461
373872 92.544 -68.858 54.442
373873 92.589 -68.846 54.422
373874 92.633 -68.835 54.401
373875 92.678 -68.823 54.380
373876 92.723 -68.811 54.359
373877 92.767 -68.799 54.338
373878 92.812 -68.787 54.316
373879 92.857 -68.775 54.292
373880 92.901 -68.764 54.267
373881 92.946 -68.752 54.241
373882 92.989 -68.742 54.214
373883 93.024 -68.742 54.187
373884 93.059 -68.743 54.160
373885 93.094 -68.744 54.132
373886 93.129 -68.745 54.105
373887 93.164 -68.745 54.075
373888 93.199 -68.746 54.047
373889 93.234 -68.747 54.018
373890 93.269 -68.748 53.990
373891 93.312 -68.733 53.962
373923 94.787 -68.375 53.453
373924 94.817 -68.386 53.444
373925 94.837 -68.410 53.434
373926 94.858 -68.433 53.424
373927 94.880 -68.453 53.415
373928 94.910 -68.442 53.405
373929 94.941 -68.431 53.395
373930 94.971 -68.420 53.384
373931 95.001 -68.409 53.374
373932 95.032 -68.398 53.365
373933 95.062 -68.387 53.356
373934 95.093 -68.376 53.348
373935 95.124 -68.366 53.340
373936 95.154 -68.354 53.333
373937 95.185 -68.344 53.327
373938 95.215 -68.332 53.319
373939 95.250 -68.319 53.311
373940 95.288 -68.303 53.301
373941 95.321 -68.291 53.291
373942 95.348 -68.285 53.280
373943 95.375 -68.278 53.270
373944 95.406 -68.290 53.262
373945 95.446 -68.301 53.256
373946 95.486 -68.312 53.249
373947 95.527 -68.322 53.241
373948 95.568 -68.333 53.234
373949 95.609 -68.344 53.227
373950 95.651 -68.355 53.218
373951 95.696 -68.373 53.207
373952 95.743 -68.391 53.196
373953 95.789 -68.409 53.184
373954 95.835 -68.426 53.173
373955 95.881 -68.445 53.164
373956 95.926 -68.462 53.158
373957 95.965 -68.470 53.156
373958 96.002 -68.476 53.155
373959 96.038 -68.482 53.157
373960 96.075 -68.487 53.158
373961 96.111 -68.492 53.156
373962 96.148 -68.498 53.150
373963 96.181 -68.501 53.140
373964 96.209 -68.500 53.121
373965 96.237 -68.498 53.099
373966 96.266 -68.498 53.077
373967 96.294 -68.496 53.052
373968 96.322 -68.495 53.024
373969 96.350 -68.494 52.992
373970 96.379 -68.493 52.959
373971 96.407 -68.492 52.928
373972 96.436 -68.493 52.900
373973 96.465 -68.501 52.874
373974 96.495 -68.508 52.846
373975 96.525 -68.513 52.814
373976 96.555 -68.517 52.781
373977 96.582 -68.520 52.742
373978 96.608 -68.523 52.702
373979 96.633 -68.526 52.662
373980 96.659 -68.529 52.624
373981 96.684 -68.532 52.586
373982 96.710 -68.535 52.548
373983 96.738 -68.533 52.510
373984 96.766 -68.532 52.472
373985 96.795 -68.529 52.436
373986 96.822 -68.528 52.401
373987 96.851 -68.526 52.365
373988 96.895 -68.536 52.326
373989 96.946 -68.553 52.288
373990 96.995 -68.568 52.251
373991 97.043 -68.583 52.218
373992 97.093 -68.598 52.189
373993 97.142 -68.614 52.165
373994 97.191 -68.629 52.145
373995 97.221 -68.629 52.130
373996 97.243 -68.623 52.119
373997 97.265 -68.617 52.110
373998 97.288 -68.611 52.107
373999 97.310 -68.604 52.108
374000 97.333 -68.599 52.112
374001 97.355 -68.595 52.116
374002 97.376 -68.592 52.118
374003 97.401 -68.592 52.114
374004 97.433 -68.598 52.103
374005 97.463 -68.604 52.087
374006 97.496 -68.609 52.066
374007 97.527 -68.615 52.041
374008 97.560 -68.620 52.009
374009 97.591 -68.626 51.973
374010 97.624 -68.632 51.936
374011 97.655 -68.637 51.899
374012 97.687 -68.642 51.862
374013 97.722 -68.645 51.824
374014 97.755 -68.647 51.785
374015 97.789 -68.650 51.744
374016 97.823 -68.652 51.699
374017 97.857 -68.655 51.650
374018 97.890 -68.658 51.603
374019 97.925 -68.660 51.555
374020 97.958 -68.663 51.508
374021 97.992 -68.666 51.460
374022 98.026 -68.669 51.410
374023 98.061 -68.673 51.360
374024 98.095 -68.676 51.310
374025 98.131 -68.681 51.261
374026 98.167 -68.686 51.213
374027 98.201 -68.696 51.165
374028 98.241 -68.709 51.116
374029 98.282 -68.721 51.065
374030 98.322 -68.734 51.015
374031 98.363 -68.746 50.965
374032 98.404 -68.759 50.916
374033 98.444 -68.771 50.867
374034 98.485 -68.784 50.818
374035 98.525 -68.796 50.765
374036 98.566 -68.809 50.714
374037 98.606 -68.821 50.663
374038 98.647 -68.834 50.612
374039 98.687 -68.846 50.561
374040 98.727 -68.859 50.510
374041 98.768 -68.871 50.458
374042 98.779 -68.864 50.405
374043 98.788 -68.857 50.351
374044 98.798 -68.849 50.297
374045 98.841 -68.860 50.244
374046 98.888 -68.873 50.192
374047 98.935 -68.886 50.139
374048 98.982 -68.899 50.089
374049 99.029 -68.913 50.038
374050 99.076 -68.926 49.987
374051 99.123 -68.938 49.936
374052 99.154 -68.952 49.887
374053 99.177 -68.971 49.838
374054 99.200 -68.990 49.787
374055 99.222 -69.009 49.738
374056 99.245 -69.027 49.691
374057 99.266 -69.045 49.645
374058 99.283 -69.056 49.600
374059 99.299 -69.066 49.554
374060 99.328 -69.074 49.506
374061 99.376 -69.080 49.458
374062 99.424 -69.086 49.410
374063 99.471 -69.091 49.360
374064 99.501 -69.102 49.311
374065 99.531 -69.112 49.261
374066 99.561 -69.122 49.214
374067 99.608 -69.131 49.167
374068 99.660 -69.138 49.121
374069 99.712 -69.147 49.074
374070 99.764 -69.154 49.027
374071 99.816 -69.162 48.980
374072 99.868 -69.170 48.932
374073 99.920 -69.178 48.883
374074 99.970 -69.184 48.836
374075 100.016 -69.190 48.789
374076 100.051 -69.195 48.740
374077 100.073 -69.201 48.690
374078 100.096 -69.208 48.641
374079 100.117 -69.214 48.591
374080 100.139 -69.220 48.541
374081 100.161 -69.227 48.491
374082 100.189 -69.232 48.442
374083 100.234 -69.234 48.395
374084 100.280 -69.237 48.348
374085 100.326 -69.239 48.302
374086 100.374 -69.298 48.259
374087 100.413 -69.325 48.218
374088 100.443 -69.322 48.178
374089 100.474 -69.319 48.140
374090 100.504 -69.315 48.102
374091 100.535 -69.312 48.064
374092 100.566 -69.333 48.027
374093 100.599 -69.372 47.991
374094 100.631 -69.411 47.956
374095 100.662 -69.445 47.922
374096 100.691 -69.450 47.887
374097 100.720 -69.454 47.852
374098 100.750 -69.459 47.816
374099 100.778 -69.463 47.781
374100 100.808 -69.469 47.746
374101 100.846 -69.467 47.711
374102 100.886 -69.465 47.675
374103 100.925 -69.463 47.639
374104 100.965 -69.460 47.602
374105 101.005 -69.459 47.564
374534 90.749 -68.580 54.789
374535 90.780 -68.591 54.785
374536 90.811 -68.601 54.785
374537 90.842 -68.611 54.786
374538 90.873 -68.622 54.789
374539 90.904 -68.633 54.794
374540 90.950 -68.653 54.800
374541 91.008 -68.683 54.807
374542 91.067 -68.712 54.814
374543 91.125 -68.741 54.822
374544 91.183 -68.770 54.831
374545 91.242 -68.799 54.840
374546 91.300 -68.829 54.845
374547 91.354 -68.875 54.845
374548 91.405 -68.935 54.839
374549 91.455 -68.979 54.824
374550 91.506 -68.977 54.808
374551 91.556 -68.975 54.792
374552 91.604 -68.974 54.776
374553 91.646 -68.975 54.760
374554 91.688 -68.976 54.743
374555 91.733 -68.971 54.726
374556 91.778 -68.965 54.709
374557 91.822 -68.960 54.693
374558 91.867 -68.954 54.679
374559 91.912 -68.949 54.666
374560 91.956 -68.943 54.654
374561 92.001 -68.937 54.641
374564 92.135 -68.921 54.602
374565 92.180 -68.915 54.588
374566 92.225 -68.909 54.573
374567 92.269 -68.903 54.557
374568 92.314 -68.897 54.541
374569 92.359 -68.886 54.524
374570 92.403 -68.874 54.506
374571 92.448 -68.861 54.486
374572 92.493 -68.850 54.467
374573 92.537 -68.838 54.448
374574 92.582 -68.826 54.431
374575 92.627 -68.814 54.411
374576 92.671 -68.802 54.391
374577 92.716 -68.790 54.371
374578 92.761 -68.779 54.351
374579 92.805 -68.767 54.330
374580 92.850 -68.755 54.307
374581 92.895 -68.743 54.281
374582 92.939 -68.731 54.254
374583 92.984 -68.719 54.227
374584 93.019 -68.720 54.201
374585 93.054 -68.720 54.173
374586 93.088 -68.722 54.145
374587 93.124 -68.723 54.117
374588 93.158 -68.723 54.088
374589 93.194 -68.724 54.059
374590 93.236 -68.710 54.030
374591 93.289 -68.676 54.002
374592 93.339 -68.653 53.974
374593 93.384 -68.644 53.947
374597 93.567 -68.605 53.845
374598 93.613 -68.596 53.822
374599 93.658 -68.586 53.799
374600 93.704 -68.577 53.778
374601 93.749 -68.567 53.758
374602 93.795 -68.557 53.738
374603 93.840 -68.548 53.719
374604 93.886 -68.538 53.701
374605 93.931 -68.529 53.684
374606 93.977 -68.520 53.669
374607 94.023 -68.510 53.653
374608 94.068 -68.501 53.638
374609 94.114 -68.491 53.625
374610 94.159 -68.482 53.612
374611 94.205 -68.472 53.597
374624 94.794 -68.352 53.453
374625 94.815 -68.376 53.444
374626 94.835 -68.399 53.434
374627 94.855 -68.423 53.424
374628 94.885 -68.417 53.413
374629 94.915 -68.406 53.403
374630 94.946 -68.395 53.392
374631 94.976 -68.385 53.381
374632 95.007 -68.373 53.370
374633 95.037 -68.363 53.361
374634 95.068 -68.351 53.351
374635 95.098 -68.341 53.342
374636 95.129 -68.329 53.333
374637 95.160 -68.319 53.325
374638 95.190 -68.307 53.319
374639 95.220 -68.297 53.312
374640 95.251 -68.286 53.305
374641 95.289 -68.269 53.296
374642 95.324 -68.256 53.285
374643 95.350 -68.250 53.274
374644 95.379 -68.248 53.265
374645 95.411 -68.262 53.257
374646 95.449 -68.273 53.252
374647 95.489 -68.284 53.245
374648 95.531 -68.295 53.237
374649 95.571 -68.306 53.229
374650 95.612 -68.316 53.220
374651 95.653 -68.327 53.210
374652 95.694 -68.338 53.200
374653 95.735 -68.348 53.190
374654 95.779 -68.363 53.179
374655 95.825 -68.381 53.169
374656 95.871 -68.399 53.162
374657 95.917 -68.416 53.156
374658 95.956 -68.426 53.151
374659 95.993 -68.431 53.147
374660 96.029 -68.437 53.146
374661 96.066 -68.442 53.145
374662 96.102 -68.448 53.142
374663 96.138 -68.454 53.138
374664 96.175 -68.459 53.127
374665 96.212 -68.464 53.110
374666 96.248 -68.470 53.091
374667 96.282 -68.473 53.069
374668 96.310 -68.472 53.044
374669 96.338 -68.471 53.013
374670 96.367 -68.470 52.979
374671 96.395 -68.469 52.945
374672 96.423 -68.467 52.913
374673 96.452 -68.466 52.887
374674 96.481 -68.473 52.860
374675 96.511 -68.479 52.831
374676 96.540 -68.484 52.801
374677 96.568 -68.488 52.767
374678 96.594 -68.491 52.728
374679 96.619 -68.494 52.689
374680 96.645 -68.497 52.650
374681 96.668 -68.498 52.612
374682 96.690 -68.499 52.574
374683 96.719 -68.504 52.536
374684 96.749 -68.507 52.499
374685 96.775 -68.501 52.463
374686 96.803 -68.499 52.427
374687 96.831 -68.498 52.393
374688 96.859 -68.495 52.356
374689 96.887 -68.494 52.319
374690 96.940 -68.511 52.282
374691 96.993 -68.529 52.246
374692 97.047 -68.548 52.215
374693 97.096 -68.564 52.189
374694 97.146 -68.579 52.166
374695 97.194 -68.594 52.148
374696 97.227 -68.596 52.137
374697 97.250 -68.590 52.129
374698 97.272 -68.584 52.124
374699 97.295 -68.578 52.123
374700 97.317 -68.572 52.128
374701 97.339 -68.566 52.136
374702 97.362 -68.560 52.140
374703 97.384 -68.554 52.139
374704 97.409 -68.554 52.133
374705 97.441 -68.560 52.116
374706 97.473 -68.566 52.095
374707 97.505 -68.571 52.069
374708 97.537 -68.576 52.040
374709 97.569 -68.582 52.004
374710 97.601 -68.588 51.966
374711 97.633 -68.593 51.929
374712 97.665 -68.599 51.892
374713 97.697 -68.604 51.856
374714 97.731 -68.607 51.818
374715 97.765 -68.610 51.779
374716 97.799 -68.612 51.738
374717 97.832 -68.615 51.693
374718 97.866 -68.617 51.644
374719 97.900 -68.620 51.597
374720 97.934 -68.623 51.551
374721 97.968 -68.625 51.503
374722 98.005 -68.630 51.453
374723 98.042 -68.636 51.403
374724 98.079 -68.642 51.352
374725 98.115 -68.647 51.300
374726 98.151 -68.652 51.249
374727 98.172 -68.656 51.201
374728 98.215 -68.664 51.152
374729 98.250 -68.676 51.100
374730 98.286 -68.689 51.047
374731 98.326 -68.702 50.996
374732 98.367 -68.714 50.944
374733 98.408 -68.727 50.893
374734 98.448 -68.739 50.843
374735 98.489 -68.752 50.793
374736 98.529 -68.764 50.743
374737 98.570 -68.777 50.694
374738 98.610 -68.789 50.644
374739 98.651 -68.802 50.594
374740 98.691 -68.814 50.545
374741 98.731 -68.827 50.497
374742 98.772 -68.839 50.446
374743 98.786 -68.834 50.393
374744 98.795 -68.827 50.339
374745 98.813 -68.824 50.286
374746 98.860 -68.836 50.234
374747 98.907 -68.850 50.182
374748 98.954 -68.863 50.131
374749 99.001 -68.877 50.084
374750 99.048 -68.889 50.034
374751 99.090 -68.896 49.984
374752 99.120 -68.887 49.934
374753 99.157 -68.918 49.885
374754 99.183 -68.940 49.834
374755 99.206 -68.959 49.784
374756 99.229 -68.979 49.736
374757 99.252 -68.998 49.690
374758 99.271 -69.012 49.642
374759 99.287 -69.023 49.597
374760 99.304 -69.034 49.550
374761 99.325 -69.043 49.503
374762 99.372 -69.049 49.457
374763 99.420 -69.054 49.411
374764 99.468 -69.060 49.363
374765 99.513 -69.066 49.314
374766 99.543 -69.077 49.266
374767 99.573 -69.087 49.220
374768 99.612 -69.096 49.173
374769 99.664 -69.104 49.127
374770 99.716 -69.112 49.080
374771 99.768 -69.119 49.034
374772 99.820 -69.128 48.986
374773 99.872 -69.135 48.938
374774 99.924 -69.143 48.889
374775 99.970 -69.148 48.841
374776 100.015 -69.153 48.792
374777 100.060 -69.158 48.743
374778 100.082 -69.164 48.693
374779 100.104 -69.171 48.644
374780 100.126 -69.177 48.592
374781 100.148 -69.183 48.543
374782 100.170 -69.189 48.492
374783 100.210 -69.208 48.442
374784 100.261 -69.234 48.392
374785 100.307 -69.238 48.345
374786 100.353 -69.273 48.299
374787 100.384 -69.269 48.256
374788 100.414 -69.266 48.214
374789 100.444 -69.263 48.175
374790 100.475 -69.259 48.136
374791 100.505 -69.256 48.097
374792 100.537 -69.278 48.059
374793 100.569 -69.317 48.022
374794 100.601 -69.356 47.987
374795 100.633 -69.395 47.951
374796 100.662 -69.405 47.916
374797 100.692 -69.410 47.882
374798 100.720 -69.414 47.846
374799 100.750 -69.419 47.811
374800 100.779 -69.423 47.776
374801 100.816 -69.423 47.741
374802 100.855 -69.421 47.705
374803 100.895 -69.419 47.670
374804 100.935 -69.417 47.634
375234 90.706 -68.537 54.787
375235 90.737 -68.548 54.782
375236 90.768 -68.558 54.778
375237 90.799 -68.569 54.778
375238 90.830 -68.579 54.780
375239 90.861 -68.590 54.785
375240 90.892 -68.601 54.790
375241 90.944 -68.625 54.797
375242 91.002 -68.654 54.804
375243 91.061 -68.683 54.812
375244 91.119 -68.713 54.823
375245 91.177 -68.742 54.833
375246 91.235 -68.771 54.843
375247 91.294 -68.801 54.848
375248 91.347 -68.850 54.846
375249 91.398 -68.910 54.837
375250 91.448 -68.921 54.821
375251 91.498 -68.920 54.803
375252 91.548 -68.918 54.786
375253 91.599 -68.916 54.770
375254 91.642 -68.917 54.753
375255 91.684 -68.918 54.736
375256 91.726 -68.918 54.719
375257 91.770 -68.914 54.702
375258 91.815 -68.908 54.687
375259 91.860 -68.902 54.673
375260 91.905 -68.897 54.660
375261 91.949 -68.892 54.648
375262 91.994 -68.886 54.636
375264 92.083 -68.874 54.611
375265 92.128 -68.869 54.599
375266 92.173 -68.863 54.587
375267 92.217 -68.858 54.573
375268 92.262 -68.852 54.558
375269 92.307 -68.846 54.542
375270 92.352 -68.840 54.526
375271 92.397 -68.835 54.510
375272 92.441 -68.829 54.491
375273 92.486 -68.824 54.474
375274 92.530 -68.817 54.457
375275 92.575 -68.805 54.440
375276 92.620 -68.793 54.421
375277 92.664 -68.782 54.403
375278 92.709 -68.770 54.384
375279 92.754 -68.758 54.365
375280 92.798 -68.746 54.344
375281 92.843 -68.734 54.323
375282 92.887 -68.722 54.297
375283 92.932 -68.711 54.269
375284 92.977 -68.698 54.241
375285 93.013 -68.698 54.214
375286 93.048 -68.698 54.186
375287 93.083 -68.699 54.158
375288 93.118 -68.700 54.130
375289 93.160 -68.687 54.101
375290 93.213 -68.654 54.071
375291 93.259 -68.645 54.042
375292 93.304 -68.636 54.014
375293 93.350 -68.626 53.987
375294 93.396 -68.617 53.960
375296 93.487 -68.597 53.908
375297 93.532 -68.588 53.883
375298 93.578 -68.578 53.859
375299 93.623 -68.569 53.836
375300 93.669 -68.559 53.813
375301 93.714 -68.549 53.792
375302 93.760 -68.540 53.772
375303 93.805 -68.530 53.752
375304 93.851 -68.521 53.733
375305 93.897 -68.511 53.715
375306 93.942 -68.501 53.697
375307 93.988 -68.492 53.681
375308 94.034 -68.482 53.664
375309 94.079 -68.473 53.649
375310 94.125 -68.463 53.635
375311 94.170 -68.454 53.620
375312 94.216 -68.445 53.606
375313 94.261 -68.435 53.593
375325 94.792 -68.341 53.454
375326 94.813 -68.365 53.443
375327 94.833 -68.389 53.433
375328 94.859 -68.392 53.423
375329 94.890 -68.382 53.412
375330 94.920 -68.370 53.401
375331 94.951 -68.360 53.389
375332 94.981 -68.348 53.378
375333 95.012 -68.338 53.366
375334 95.043 -68.326 53.356
375335 95.073 -68.316 53.346
375336 95.104 -68.304 53.336
375337 95.134 -68.294 53.325
375338 95.165 -68.283 53.315
375339 95.195 -68.272 53.308
375340 95.226 -68.261 53.302
375341 95.256 -68.250 53.296
375342 95.290 -68.237 53.288
375343 95.327 -68.222 53.277
375344 95.353 -68.216 53.267
375345 95.383 -68.220 53.259
375346 95.415 -68.234 53.251
375347 95.452 -68.246 53.247
375348 95.492 -68.256 53.240
375349 95.534 -68.267 53.232
375350 95.574 -68.278 53.224
375351 95.615 -68.288 53.214
375352 95.657 -68.299 53.204
375353 95.697 -68.310 53.194
375354 95.738 -68.320 53.184
375355 95.779 -68.332 53.174
375356 95.820 -68.342 53.166
375357 95.862 -68.353 53.159
375358 95.907 -68.371 53.153
375359 95.948 -68.381 53.147
375360 95.984 -68.387 53.138
375361 96.021 -68.392 53.133
375362 96.057 -68.398 53.129
375363 96.093 -68.403 53.126
375364 96.130 -68.409 53.121
375365 96.166 -68.414 53.111
375366 96.203 -68.419 53.096
375367 96.239 -68.426 53.077
375368 96.276 -68.431 53.056
375369 96.312 -68.436 53.029
375370 96.348 -68.442 52.997
375371 96.383 -68.445 52.962
375372 96.411 -68.444 52.928
375373 96.439 -68.443 52.897
375374 96.467 -68.441 52.870
375375 96.497 -68.445 52.843
375376 96.526 -68.451 52.815
375377 96.554 -68.454 52.784
375378 96.580 -68.457 52.751
375379 96.604 -68.460 52.713
375380 96.627 -68.461 52.675
375381 96.650 -68.463 52.637
375382 96.672 -68.463 52.599
375383 96.697 -68.466 52.562
375384 96.728 -68.471 52.525
375385 96.759 -68.476 52.489
375386 96.789 -68.479 52.453
375387 96.814 -68.473 52.419
375388 96.839 -68.467 52.384
375389 96.868 -68.466 52.349
375390 96.895 -68.463 52.312
375391 96.931 -68.467 52.276
375392 96.984 -68.486 52.241
375393 97.038 -68.504 52.213
375394 97.092 -68.523 52.189
375395 97.146 -68.541 52.170
375396 97.198 -68.559 52.153
375397 97.234 -68.564 52.145
375398 97.256 -68.557 52.140
375399 97.279 -68.551 52.139
375400 97.301 -68.546 52.142
375401 97.324 -68.541 52.149
375402 97.347 -68.536 52.155
375403 97.369 -68.531 52.157
375404 97.391 -68.526 52.152
375405 97.417 -68.520 52.141
375406 97.450 -68.523 52.120
375407 97.483 -68.527 52.095
375408 97.515 -68.533 52.065
375409 97.547 -68.538 52.032
375410 97.579 -68.544 51.996
375411 97.611 -68.550 51.958
375412 97.643 -68.555 51.920
375413 97.675 -68.560 51.883
375414 97.707 -68.567 51.847
375415 97.740 -68.570 51.810
375416 97.774 -68.572 51.772
375417 97.808 -68.575 51.732
375418 97.842 -68.577 51.686
375419 97.876 -68.579 51.638
375420 97.910 -68.582 51.591
375421 97.947 -68.588 51.543
375422 97.986 -68.595 51.494
375423 98.024 -68.601 51.444
375424 98.062 -68.608 51.391
375425 98.098 -68.614 51.337
375426 98.135 -68.619 51.283
375427 98.171 -68.624 51.230
375428 98.176 -68.626 51.179
375429 98.224 -68.633 51.130
375430 98.264 -68.645 51.079
375431 98.300 -68.657 51.026
375432 98.335 -68.670 50.972
375433 98.371 -68.683 50.919
375434 98.412 -68.695 50.866
375435 98.452 -68.708 50.814
375436 98.493 -68.720 50.763
375437 98.533 -68.733 50.715
375438 98.574 -68.745 50.667
375439 98.614 -68.758 50.620
375440 98.655 -68.770 50.573
375441 98.695 -68.783 50.526
375442 98.735 -68.796 50.479
375443 98.776 -68.808 50.429
375444 98.792 -68.804 50.378
375445 98.801 -68.796 50.328
375446 98.832 -68.800 50.276
375447 98.879 -68.813 50.224
375448 98.926 -68.827 50.174
375449 98.973 -68.839 50.125
375450 99.020 -68.853 50.079
375451 99.058 -68.855 50.029
375452 99.087 -68.846 49.979
375453 99.120 -68.847 49.930
375454 99.157 -68.882 49.881
375455 99.190 -68.911 49.830
375456 99.212 -68.930 49.780
375457 99.235 -68.949 49.731
375458 99.258 -68.968 49.684
375459 99.275 -68.980 49.637
375460 99.292 -68.990 49.591
375461 99.308 -69.001 49.545
375462 99.325 -69.012 49.499
375463 99.369 -69.018 49.455
375464 99.417 -69.023 49.408
375465 99.465 -69.029 49.361
375466 99.512 -69.034 49.314
375467 99.555 -69.041 49.267
375468 99.585 -69.052 49.222
375469 99.617 -69.062 49.177
375470 99.668 -69.069 49.131
375471 99.720 -69.078 49.086
375472 99.773 -69.085 49.040
375473 99.825 -69.093 48.992
375474 99.876 -69.101 48.943
375475 99.923 -69.106 48.893
375476 99.969 -69.111 48.845
375477 100.014 -69.116 48.795
375478 100.060 -69.121 48.745
375479 100.090 -69.127 48.695
375480 100.112 -69.133 48.644
375481 100.134 -69.140 48.592
375482 100.156 -69.146 48.542
375483 100.179 -69.153 48.491
375484 100.230 -69.178 48.440
375485 100.281 -69.205 48.390
375486 100.323 -69.218 48.342
375487 100.354 -69.214 48.296
375488 100.384 -69.211 48.252
375489 100.414 -69.208 48.211
375490 100.445 -69.204 48.171
375491 100.475 -69.201 48.132
375492 100.507 -69.224 48.093
375493 100.539 -69.262 48.055
375494 100.571 -69.301 48.018
375495 100.603 -69.341 47.982
375496 100.633 -69.360 47.947
375497 100.662 -69.365 47.912
375498 100.692 -69.369 47.877
375499 100.721 -69.374 47.841
375500 100.750 -69.378 47.806
375501 100.785 -69.379 47.771
375502 100.825 -69.377 47.736
375503 100.865 -69.375 47.701
375935 90.693 -68.505 54.779
375936 90.725 -68.516 54.774
375937 90.756 -68.526 54.772
375938 90.787 -68.536 54.772
375939 90.818 -68.547 54.776
375940 90.849 -68.557 54.781
375941 90.880 -68.568 54.787
375942 90.938 -68.597 54.794
375943 90.996 -68.626 54.802
375944 91.054 -68.655 54.812
375945 91.113 -68.685 54.821
375946 91.171 -68.714 54.834
375947 91.229 -68.743 54.842
375948 91.288 -68.773 54.845
375949 91.340 -68.826 54.842
375950 91.391 -68.866 54.832
375951 91.441 -68.864 54.814
375952 91.491 -68.862 54.797
375953 91.541 -68.861 54.780
375954 91.591 -68.859 54.762
375955 91.637 -68.859 54.746
375956 91.679 -68.860 54.730
375957 91.721 -68.861 54.713
375958 91.764 -68.861 54.695
375959 91.808 -68.857 54.680
375960 91.853 -68.852 54.666
375961 91.897 -68.846 54.653
375962 91.942 -68.840 54.642
375963 91.987 -68.834 54.631
375964 92.031 -68.829 54.620
375965 92.076 -68.823 54.608
375966 92.121 -68.817 54.596
375967 92.165 -68.811 54.585
375968 92.210 -68.806 54.572
375969 92.255 -68.800 54.559
375970 92.300 -68.795 54.544
375971 92.345 -68.789 54.528
375972 92.389 -68.783 54.513
375973 92.434 -68.777 54.497
375974 92.479 -68.772 54.481
375975 92.523 -68.766 54.465
375976 92.568 -68.761 54.450
375977 92.613 -68.755 54.434
375978 92.657 -68.749 54.416
375979 92.702 -68.744 54.398
375980 92.747 -68.737 54.379
375981 92.791 -68.726 54.358
375982 92.836 -68.714 54.336
375983 92.881 -68.701 54.310
375984 92.925 -68.690 54.282
375985 92.970 -68.678 54.254
375986 93.008 -68.675 54.225
375987 93.043 -68.676 54.197
375988 93.084 -68.664 54.169
375989 93.133 -68.646 54.141
375990 93.178 -68.637 54.112
375991 93.224 -68.627 54.083
375992 93.269 -68.617 54.054
375993 93.315 -68.608 54.027
375994 93.360 -68.598 54.000
375995 93.406 -68.589 53.974
375996 93.451 -68.579 53.947
375997 93.497 -68.570 53.922
375998 93.543 -68.560 53.897
375999 93.588 -68.551 53.874
376000 93.634 -68.542 53.851
376001 93.680 -68.532 53.828
376002 93.725 -68.522 53.807
376003 93.771 -68.513 53.787
376004 93.816 -68.503 53.767
376005 93.862 -68.493 53.748
376006 93.907 -68.484 53.729
376007 93.953 -68.474 53.710
376008 93.998 -68.465 53.693
376009 94.044 -68.455 53.677
376010 94.089 -68.445 53.661
376011 94.135 -68.436 53.645
376012 94.181 -68.426 53.629
376013 94.226 -68.417 53.614
376014 94.272 -68.407 53.601
376015 94.318 -68.398 53.587
376016 94.363 -68.388 53.572
376027 94.811 -68.354 53.444
376028 94.834 -68.367 53.433
376029 94.864 -68.357 53.423
376030 94.895 -68.345 53.411
376031 94.926 -68.335 53.399
376032 94.956 -68.323 53.387
376033 94.987 -68.313 53.375
376034 95.017 -68.301 53.362
376035 95.048 -68.291 53.351
376036 95.078 -68.280 53.341
376037 95.109 -68.269 53.329
376038 95.139 -68.258 53.318
376039 95.170 -68.247 53.308
376040 95.200 -68.236 53.298
376041 95.231 -68.225 53.290
376042 95.261 -68.214 53.284
376043 95.292 -68.203 53.277
376044 95.329 -68.188 53.267
376045 95.356 -68.181 53.258
376046 95.388 -68.192 53.249
376047 95.420 -68.205 53.242
376048 95.455 -68.218 53.238
376049 95.496 -68.228 53.233
376050 95.537 -68.239 53.227
376051 95.578 -68.250 53.219
376052 95.618 -68.260 53.210
376053 95.660 -68.271 53.200
376054 95.700 -68.282 53.190
376055 95.741 -68.293 53.179
376056 95.782 -68.303 53.169
376057 95.823 -68.314 53.161
376058 95.864 -68.325 53.156
376059 95.905 -68.334 53.149
376060 95.943 -68.340 53.141
376061 95.979 -68.345 53.129
376062 96.015 -68.350 53.119
376063 96.051 -68.355 53.111
376064 96.087 -68.360 53.106
376065 96.122 -68.365 53.100
376066 96.158 -68.370 53.091
376067 96.194 -68.376 53.076
376068 96.231 -68.381 53.058
376069 96.267 -68.386 53.036
376070 96.303 -68.391 53.009
376071 96.340 -68.398 52.977
376072 96.376 -68.403 52.943
376073 96.413 -68.408 52.910
376074 96.449 -68.414 52.879
376075 96.483 -68.417 52.852
376076 96.512 -68.416 52.825
376077 96.540 -68.422 52.797
376078 96.564 -68.423 52.766
376079 96.586 -68.425 52.734
376080 96.609 -68.426 52.697
376081 96.631 -68.427 52.660
376082 96.654 -68.428 52.623
376083 96.676 -68.429 52.586
376084 96.707 -68.434 52.550
376085 96.737 -68.438 52.514
376086 96.768 -68.443 52.478
376087 96.799 -68.448 52.442
376088 96.828 -68.450 52.410
376089 96.853 -68.445 52.377
376090 96.879 -68.439 52.342
376091 96.904 -68.433 52.306
376092 96.932 -68.432 52.271
376093 96.976 -68.442 52.239
376094 97.030 -68.460 52.213
376095 97.083 -68.479 52.191
376096 97.139 -68.504 52.174
376097 97.198 -68.529 52.160
376098 97.239 -68.534 52.154
376099 97.263 -68.532 52.152
376100 97.286 -68.526 52.154
376101 97.308 -68.521 52.160
376102 97.331 -68.517 52.167
376103 97.354 -68.511 52.171
376104 97.376 -68.506 52.169
376105 97.398 -68.501 52.160
376106 97.423 -68.495 52.144
376107 97.454 -68.492 52.119
376108 97.488 -68.495 52.090
376109 97.522 -68.498 52.058
376110 97.556 -68.501 52.024
376111 97.589 -68.506 51.987
376112 97.621 -68.511 51.949
376113 97.653 -68.517 51.911
376114 97.685 -68.523 51.875
376115 97.717 -68.528 51.838
376116 97.750 -68.532 51.802
376117 97.784 -68.535 51.765
376118 97.817 -68.537 51.725
376119 97.852 -68.539 51.680
376120 97.889 -68.546 51.631
376121 97.928 -68.552 51.583
376122 97.966 -68.559 51.533
376123 98.005 -68.566 51.482
376124 98.043 -68.573 51.432
376125 98.082 -68.579 51.376
376126 98.118 -68.585 51.320
376127 98.154 -68.590 51.265
376128 98.176 -68.593 51.210
376129 98.181 -68.595 51.158
376130 98.231 -68.603 51.107
376131 98.278 -68.612 51.056
376132 98.313 -68.625 51.003
376133 98.349 -68.638 50.949
376134 98.384 -68.650 50.896
376135 98.420 -68.663 50.842
376136 98.456 -68.676 50.790
376137 98.497 -68.689 50.740
376138 98.537 -68.701 50.690
376139 98.578 -68.714 50.642
376140 98.618 -68.726 50.595
376141 98.659 -68.739 50.549
376142 98.699 -68.751 50.503
376143 98.740 -68.763 50.458
376144 98.780 -68.776 50.410
376145 98.798 -68.774 50.362
376146 98.808 -68.766 50.313
376147 98.851 -68.777 50.263
376148 98.898 -68.790 50.214
376149 98.945 -68.803 50.165
376150 98.992 -68.817 50.120
376151 99.025 -68.812 50.073
376152 99.055 -68.804 50.025
376153 99.085 -68.795 49.975
376154 99.120 -68.811 49.926
376155 99.158 -68.846 49.878
376156 99.196 -68.880 49.828
376157 99.219 -68.899 49.778
376158 99.241 -68.918 49.728
376159 99.264 -68.936 49.680
376160 99.280 -68.947 49.632
376161 99.296 -68.958 49.586
376162 99.313 -68.968 49.540
376163 99.329 -68.979 49.495
376164 99.365 -68.987 49.451
376165 99.413 -68.992 49.405
376166 99.462 -68.997 49.359
376167 99.509 -69.003 49.313
376168 99.557 -69.009 49.267
376169 99.598 -69.016 49.223
376170 99.628 -69.026 49.179
376171 99.673 -69.035 49.135
376172 99.725 -69.043 49.091
376173 99.777 -69.050 49.045
376174 99.829 -69.059 48.996
376175 99.876 -69.064 48.947
376176 99.922 -69.069 48.897
376177 99.968 -69.074 48.848
376178 100.013 -69.079 48.798
376179 100.059 -69.084 48.747
376180 100.099 -69.090 48.696
376181 100.121 -69.096 48.644
376182 100.143 -69.102 48.591
376183 100.164 -69.109 48.540
376184 100.199 -69.123 48.489
376185 100.250 -69.149 48.438
376186 100.288 -69.186 48.388
376187 100.324 -69.159 48.339
376188 100.354 -69.156 48.293
376189 100.385 -69.153 48.249
376190 100.415 -69.149 48.208
376191 100.445 -69.146 48.168
376192 100.477 -69.169 48.128
376193 100.509 -69.208 48.089
376194 100.541 -69.247 48.052
376195 100.573 -69.286 48.014
376196 100.605 -69.315 47.977
376197 100.634 -69.320 47.942
376198 100.662 -69.325 47.908
376199 100.692 -69.329 47.872
376200 100.721 -69.334 47.836
376201 100.755 -69.335 47.802
376202 100.795 -69.333 47.766
376636 90.681 -68.473 54.771
376637 90.713 -68.483 54.768
376638 90.744 -68.494 54.766
376639 90.775 -68.504 54.767
376640 90.806 -68.515 54.771
376641 90.837 -68.526 54.777
376642 90.874 -68.539 54.784
376643 90.932 -68.569 54.792
376644 90.990 -68.598 54.801
376645 91.048 -68.627 54.810
376646 91.107 -68.657 54.819
376647 91.165 -68.686 54.832
376648 91.223 -68.715 54.838
376649 91.282 -68.745 54.840
376650 91.333 -68.801 54.836
376651 91.383 -68.808 54.824
376652 91.434 -68.807 54.807
376653 91.484 -68.805 54.790
376654 91.534 -68.804 54.773
376655 91.584 -68.802 54.756
376656 91.633 -68.801 54.738
376657 91.674 -68.802 54.722
376658 91.717 -68.802 54.705
376659 91.759 -68.803 54.688
376660 91.801 -68.804 54.673
376661 91.845 -68.800 54.659
376662 91.890 -68.794 54.647
376663 91.935 -68.789 54.635
376664 91.980 -68.783 54.625
376665 92.024 -68.777 54.615
376666 92.069 -68.771 54.604
376667 92.113 -68.766 54.593
376668 92.158 -68.760 54.583
376669 92.203 -68.755 54.571
376670 92.248 -68.749 54.559
376671 92.293 -68.743 54.545
376672 92.337 -68.737 54.530
376673 92.382 -68.732 54.516
376674 92.427 -68.727 54.502
376675 92.472 -68.720 54.488
376676 92.516 -68.715 54.474
376677 92.561 -68.709 54.461
376678 92.605 -68.704 54.447
376679 92.650 -68.698 54.430
376680 92.695 -68.692 54.411
376681 92.740 -68.686 54.392
376682 92.784 -68.681 54.370
376683 92.829 -68.675 54.348
376684 92.874 -68.670 54.322
376685 92.919 -68.664 54.294
376686 92.964 -68.658 54.265
376687 93.006 -68.648 54.236
376688 93.052 -68.638 54.208
376689 93.097 -68.629 54.180
376690 93.143 -68.619 54.152
376691 93.189 -68.609 54.123
376692 93.234 -68.600 54.094
376693 93.280 -68.590 54.066
376694 93.326 -68.581 54.039
376695 93.371 -68.571 54.012
376696 93.417 -68.561 53.987
376697 93.462 -68.552 53.961
376698 93.508 -68.542 53.936
376699 93.553 -68.533 53.912
376700 93.599 -68.523 53.889
376701 93.644 -68.513 53.866
376702 93.690 -68.504 53.844
376703 93.735 -68.495 53.823
376704 93.781 -68.485 53.803
376705 93.827 -68.476 53.782
376706 93.872 -68.466 53.763
376707 93.918 -68.457 53.744
376708 93.964 -68.447 53.725
376709 94.009 -68.437 53.707
376710 94.055 -68.428 53.689
376711 94.100 -68.418 53.673
376712 94.146 -68.409 53.657
376713 94.191 -68.399 53.640
376714 94.237 -68.389 53.624
376715 94.282 -68.380 53.609
376716 94.328 -68.370 53.594
376717 94.373 -68.361 53.579
376718 94.419 -68.351 53.565
376729 94.839 -68.332 53.434
376730 94.870 -68.320 53.423
376731 94.900 -68.310 53.411
376732 94.931 -68.299 53.398
376733 94.961 -68.288 53.385
376734 94.992 -68.277 53.373
376735 95.023 -68.266 53.359
376736 95.053 -68.255 53.347
376737 95.084 -68.244 53.335
376738 95.114 -68.233 53.323
376739 95.144 -68.222 53.310
376740 95.175 -68.211 53.299
376741 95.205 -68.200 53.288
376742 95.236 -68.189 53.279
376743 95.267 -68.178 53.271
376744 95.297 -68.167 53.264
376745 95.330 -68.155 53.255
376746 95.360 -68.150 53.246
376747 95.392 -68.163 53.237
376748 95.424 -68.177 53.230
376749 95.458 -68.190 53.226
376750 95.499 -68.200 53.221
376751 95.540 -68.211 53.216
376752 95.581 -68.222 53.208
376753 95.622 -68.233 53.199
376754 95.663 -68.244 53.190
376755 95.703 -68.254 53.180
376756 95.745 -68.265 53.170
376757 95.785 -68.275 53.160
376758 95.826 -68.285 53.153
376759 95.867 -68.292 53.144
376760 95.908 -68.300 53.135
376761 95.944 -68.305 53.125
376762 95.982 -68.310 53.112
376763 96.019 -68.316 53.101
376764 96.054 -68.320 53.093
376765 96.090 -68.325 53.085
376766 96.126 -68.331 53.076
376767 96.162 -68.335 53.064
376768 96.198 -68.341 53.049
376769 96.234 -68.346 53.030
376770 96.269 -68.350 53.008
376771 96.305 -68.356 52.984
376772 96.341 -68.360 52.952
376773 96.377 -68.366 52.919
376774 96.412 -68.371 52.888
376775 96.449 -68.376 52.858
376776 96.484 -68.381 52.832
376777 96.520 -68.385 52.805
376778 96.545 -68.388 52.777
376779 96.568 -68.389 52.747
376780 96.590 -68.390 52.714
376781 96.613 -68.391 52.680
376782 96.636 -68.392 52.644
376783 96.658 -68.394 52.609
376784 96.684 -68.397 52.573
376785 96.715 -68.401 52.537
376786 96.746 -68.406 52.503
376787 96.777 -68.410 52.467
376788 96.807 -68.415 52.432
376789 96.838 -68.420 52.401
376790 96.867 -68.422 52.368
376791 96.892 -68.416 52.335
376792 96.918 -68.410 52.302
376793 96.943 -68.405 52.270
376794 96.968 -68.400 52.239
376795 97.026 -68.433 52.215
376796 97.086 -68.476 52.195
376797 97.145 -68.491 52.179
376798 97.202 -68.491 52.167
376799 97.236 -68.497 52.163
376800 97.270 -68.506 52.165
376801 97.293 -68.502 52.170
376802 97.315 -68.497 52.177
376803 97.338 -68.492 52.183
376804 97.361 -68.486 52.183
376805 97.383 -68.481 52.176
376806 97.405 -68.476 52.162
376807 97.429 -68.470 52.142
376808 97.459 -68.463 52.114
376809 97.493 -68.463 52.082
376810 97.526 -68.466 52.048
376811 97.560 -68.470 52.013
376812 97.594 -68.473 51.976
376813 97.628 -68.477 51.939
376814 97.661 -68.480 51.902
376815 97.695 -68.485 51.866
376816 97.727 -68.490 51.829
376817 97.759 -68.495 51.794
376818 97.793 -68.497 51.758
376819 97.832 -68.504 51.719
376820 97.870 -68.510 51.673
376821 97.909 -68.517 51.624
376822 97.947 -68.523 51.572
376823 97.986 -68.530 51.520
376824 98.024 -68.537 51.468
376825 98.063 -68.543 51.416
376826 98.101 -68.550 51.359
376827 98.138 -68.556 51.302
376828 98.174 -68.561 51.246
376829 98.180 -68.563 51.190
376830 98.189 -68.565 51.136
376831 98.238 -68.573 51.084
376832 98.289 -68.580 51.032
376833 98.327 -68.592 50.979
376834 98.363 -68.605 50.927
376835 98.398 -68.618 50.874
376836 98.434 -68.631 50.822
376837 98.469 -68.643 50.770
376838 98.505 -68.656 50.720
376839 98.541 -68.669 50.668
376840 98.582 -68.682 50.618
376841 98.622 -68.694 50.570
376842 98.663 -68.707 50.525
376843 98.703 -68.719 50.480
376844 98.744 -68.731 50.436
376845 98.784 -68.744 50.391
376846 98.805 -68.743 50.344
376847 98.822 -68.740 50.297
376848 98.869 -68.754 50.249
376849 98.917 -68.767 50.203
376850 98.963 -68.780 50.158
376851 98.993 -68.770 50.114
376852 99.023 -68.761 50.068
376853 99.053 -68.753 50.021
376854 99.083 -68.744 49.972
376855 99.120 -68.775 49.924
376856 99.158 -68.809 49.876
376857 99.196 -68.844 49.827
376858 99.225 -68.870 49.778
376859 99.248 -68.889 49.728
376860 99.268 -68.904 49.678
376861 99.284 -68.915 49.628
376862 99.301 -68.925 49.580
376863 99.317 -68.936 49.535
376864 99.334 -68.946 49.490
376865 99.362 -68.955 49.445
376866 99.410 -68.961 49.402
376867 99.458 -68.966 49.357
376868 99.506 -68.972 49.311
376869 99.554 -68.977 49.267
376870 99.602 -68.983 49.224
376871 99.640 -68.991 49.181
376872 99.677 -69.000 49.138
376873 99.729 -69.009 49.094
376874 99.781 -69.016 49.047
376875 99.830 -69.022 48.999
376876 99.875 -69.027 48.950
376877 99.921 -69.032 48.900
376878 99.967 -69.037 48.850
376879 100.012 -69.043 48.800
376880 100.058 -69.047 48.748
376881 100.104 -69.052 48.695
376882 100.129 -69.059 48.642
376883 100.151 -69.065 48.590
376884 100.173 -69.071 48.536
376885 100.219 -69.093 48.485
376886 100.266 -69.123 48.435
376887 100.294 -69.103 48.383
376888 100.325 -69.100 48.336
376889 100.355 -69.097 48.290
376890 100.385 -69.093 48.247
376891 100.416 -69.090 48.206
376892 100.447 -69.115 48.165
376893 100.479 -69.153 48.126
376894 100.511 -69.192 48.086
376895 100.544 -69.231 48.048
376896 100.575 -69.270 48.011
376897 100.605 -69.275 47.975
376898 100.634 -69.279 47.939
376899 100.663 -69.284 47.903
376900 100.692 -69.289 47.867
376901 100.724 -69.291 47.833
377337 90.669 -68.440 54.766
377338 90.700 -68.451 54.764
377339 90.732 -68.461 54.763
377340 90.763 -68.472 54.765
377341 90.794 -68.482 54.769
377342 90.825 -68.493 54.774
377343 90.867 -68.511 54.781
377344 90.926 -68.541 54.790
377345 90.984 -68.570 54.799
377346 91.042 -68.599 54.808
377347 91.101 -68.628 54.818
377348 91.159 -68.658 54.827
377349 91.217 -68.687 54.833
377350 91.276 -68.717 54.833
377351 91.326 -68.753 54.829
377352 91.376 -68.752 54.815
377353 91.427 -68.750 54.800
377354 91.476 -68.748 54.783
377355 91.527 -68.746 54.766
377356 91.577 -68.745 54.749
377357 91.627 -68.743 54.731
377358 91.670 -68.743 54.714
377359 91.712 -68.744 54.697
377360 91.754 -68.745 54.680
377361 91.797 -68.745 54.665
377362 91.839 -68.747 54.652
377363 91.883 -68.742 54.640
377364 91.928 -68.737 54.629
377365 91.972 -68.731 54.619
377366 92.017 -68.726 54.610
377367 92.062 -68.720 54.600
377368 92.106 -68.714 54.590
377369 92.151 -68.708 54.580
377370 92.196 -68.703 54.569
377371 92.241 -68.697 54.558
377372 92.285 -68.692 54.546
377373 92.330 -68.686 54.533
377374 92.375 -68.680 54.520
377375 92.420 -68.675 54.507
377376 92.464 -68.669 54.495
377377 92.509 -68.664 54.484
377378 92.554 -68.658 54.474
377379 92.598 -68.652 54.459
377380 92.643 -68.646 54.443
377381 92.688 -68.641 54.424
377382 92.732 -68.635 54.403
377383 92.777 -68.629 54.380
377384 92.822 -68.623 54.357
377385 92.867 -68.618 54.331
377386 92.912 -68.612 54.303
377387 92.960 -68.610 54.275
377388 93.010 -68.607 54.246
377389 93.060 -68.604 54.218
377390 93.108 -68.601 54.190
377391 93.154 -68.592 54.162
377392 93.199 -68.582 54.133
377393 93.245 -68.573 54.105
377394 93.290 -68.563 54.077
377395 93.336 -68.553 54.050
377396 93.381 -68.544 54.023
377397 93.427 -68.534 53.999
377398 93.473 -68.525 53.974
377399 93.518 -68.515 53.950
377400 93.564 -68.505 53.927
377401 93.610 -68.496 53.904
377402 93.655 -68.486 53.882
377403 93.701 -68.477 53.860
377404 93.746 -68.467 53.839
377405 93.792 -68.457 53.819
377406 93.837 -68.448 53.798
377407 93.883 -68.438 53.778
377408 93.928 -68.429 53.759
377409 93.974 -68.419 53.740
377410 94.019 -68.410 53.721
377411 94.065 -68.401 53.703
377412 94.111 -68.391 53.686
377413 94.156 -68.381 53.669
377414 94.202 -68.372 53.651
377415 94.248 -68.362 53.635
377416 94.293 -68.353 53.618
377417 94.339 -68.343 53.603
377418 94.384 -68.333 53.587
377419 94.430 -68.324 53.571
377420 94.475 -68.314 53.557
377421 94.521 -68.305 53.542
377430 94.844 -68.296 53.435
377431 94.875 -68.285 53.423
377432 94.906 -68.274 53.410
377433 94.936 -68.263 53.398
377434 94.967 -68.252 53.385
377435 94.997 -68.241 53.371
377436 95.028 -68.230 53.357
377437 95.058 -68.219 53.344
377438 95.089 -68.208 53.330
377439 95.119 -68.197 53.316
377440 95.149 -68.186 53.303
377441 95.180 -68.175 53.291
377442 95.211 -68.164 53.278
377443 95.241 -68.153 53.268
377444 95.272 -68.142 53.258
377445 95.303 -68.131 53.249
377446 95.333 -68.121 53.241
377447 95.365 -68.122 53.232
377448 95.397 -68.135 53.223
377449 95.429 -68.149 53.216
377450 95.461 -68.162 53.211
377451 95.502 -68.173 53.206
377452 95.543 -68.184 53.200
377453 95.584 -68.194 53.193
377454 95.625 -68.205 53.184
377455 95.666 -68.216 53.175
377456 95.706 -68.226 53.167
377457 95.748 -68.235 53.158
377458 95.788 -68.243 53.148
377459 95.829 -68.250 53.140
377460 95.870 -68.258 53.129
377461 95.910 -68.266 53.118
377462 95.945 -68.270 53.106
377463 95.983 -68.275 53.093
377464 96.021 -68.281 53.082
377465 96.058 -68.286 53.074
377466 96.094 -68.291 53.065
377467 96.130 -68.296 53.053
377468 96.166 -68.301 53.039
377469 96.201 -68.306 53.022
377470 96.237 -68.311 53.002
377471 96.273 -68.316 52.980
377472 96.309 -68.321 52.956
377473 96.345 -68.326 52.926
377474 96.381 -68.331 52.896
377475 96.416 -68.336 52.867
377476 96.452 -68.341 52.838
377477 96.488 -68.346 52.812
377478 96.524 -68.351 52.785
377479 96.549 -68.353 52.756
377480 96.572 -68.354 52.726
377481 96.594 -68.356 52.695
377482 96.617 -68.357 52.662
377483 96.640 -68.358 52.628
377484 96.663 -68.360 52.594
377485 96.693 -68.364 52.560
377486 96.724 -68.369 52.526
377487 96.755 -68.373 52.492
377488 96.786 -68.378 52.458
377489 96.816 -68.383 52.424
377490 96.847 -68.388 52.393
377491 96.878 -68.392 52.362
377492 96.906 -68.393 52.329
377493 96.932 -68.388 52.298
377494 96.957 -68.382 52.268
377495 96.982 -68.376 52.240
377496 97.033 -68.446 52.217
377497 97.092 -68.453 52.199
377498 97.150 -68.453 52.185
377499 97.199 -68.454 52.176
377500 97.234 -68.459 52.175
377501 97.267 -68.468 52.178
377502 97.300 -68.477 52.184
377503 97.322 -68.472 52.191
377504 97.345 -68.466 52.192
377505 97.368 -68.461 52.186
377506 97.390 -68.456 52.174
377507 97.413 -68.451 52.155
377508 97.435 -68.446 52.131
377509 97.466 -68.438 52.102
377510 97.497 -68.432 52.069
377511 97.530 -68.435 52.035
377512 97.564 -68.438 52.000
377513 97.598 -68.442 51.964
377514 97.632 -68.445 51.928
377515 97.665 -68.449 51.891
377516 97.699 -68.452 51.855
377517 97.733 -68.456 51.819
377518 97.768 -68.460 51.785
377519 97.810 -68.467 51.748
377520 97.851 -68.475 51.708
377521 97.889 -68.481 51.661
377522 97.928 -68.488 51.609
377523 97.966 -68.495 51.557
377524 98.005 -68.501 51.504
377525 98.043 -68.507 51.451
377526 98.082 -68.514 51.397
377527 98.120 -68.521 51.338
377528 98.158 -68.527 51.280
377529 98.180 -68.531 51.223
377530 98.184 -68.532 51.167
377531 98.196 -68.535 51.112
377532 98.246 -68.542 51.059
377533 98.296 -68.551 51.007
377534 98.342 -68.560 50.955
377535 98.377 -68.573 50.903
377536 98.413 -68.586 50.852
377537 98.448 -68.598 50.801
377538 98.483 -68.611 50.750
377539 98.519 -68.623 50.701
377540 98.554 -68.636 50.650
377541 98.589 -68.649 50.601
377542 98.626 -68.662 50.555
377543 98.667 -68.675 50.510
377544 98.707 -68.687 50.465
377545 98.748 -68.699 50.422
377546 98.788 -68.712 50.376
377547 98.811 -68.713 50.329
377548 98.841 -68.717 50.285
377549 98.888 -68.730 50.240
377550 98.931 -68.737 50.196
377551 98.960 -68.729 50.153
377552 98.991 -68.720 50.109
377553 99.020 -68.711 50.062
377554 99.050 -68.702 50.015
377555 99.082 -68.705 49.967
377556 99.120 -68.739 49.920
377557 99.158 -68.774 49.872
377558 99.196 -68.808 49.823
377559 99.231 -68.839 49.774
377560 99.255 -68.858 49.724
377561 99.273 -68.871 49.674
377562 99.289 -68.882 49.624
377563 99.305 -68.893 49.577
377564 99.322 -68.903 49.530
377565 99.338 -68.914 49.483
377566 99.359 -68.924 49.439
377567 99.407 -68.930 49.396
377568 99.454 -68.935 49.353
377569 99.502 -68.940 49.310
377570 99.550 -68.946 49.266
377571 99.598 -68.952 49.224
377572 99.646 -68.957 49.182
377573 99.683 -68.965 49.139
377574 99.733 -68.974 49.094
377575 99.783 -68.980 49.047
377576 99.829 -68.985 48.999
377577 99.874 -68.990 48.950
377578 99.920 -68.996 48.900
377579 99.966 -69.000 48.851
377580 100.012 -69.005 48.799
377581 100.057 -69.011 48.747
377582 100.103 -69.015 48.694
377583 100.137 -69.021 48.640
377584 100.159 -69.027 48.587
377585 100.188 -69.038 48.534
377586 100.239 -69.064 48.482
377587 100.264 -69.048 48.431
377588 100.294 -69.045 48.380
377589 100.325 -69.042 48.332
377590 100.355 -69.038 48.287
377591 100.386 -69.035 48.244
377592 100.417 -69.060 48.203
377593 100.450 -69.099 48.163
377594 100.482 -69.137 48.123
377595 100.514 -69.177 48.084
377596 100.546 -69.215 48.045
377597 100.576 -69.230 48.007
377598 100.605 -69.234 47.971
377599 100.634 -69.239 47.936
377600 100.663 -69.244 47.900
377601 100.694 -69.247 47.864
378038 90.657 -68.408 54.762
378039 90.688 -68.419 54.760
378040 90.720 -68.429 54.761
378041 90.751 -68.439 54.763
378042 90.782 -68.450 54.767
378043 90.813 -68.460 54.772
378044 90.861 -68.483 54.780
378045 90.920 -68.512 54.789
378046 90.978 -68.542 54.799
378047 91.036 -68.571 54.807
378048 91.095 -68.600 54.816
378049 91.153 -68.629 54.823
378050 91.211 -68.659 54.828
378051 91.269 -68.692 54.828
378052 91.319 -68.696 54.821
378053 91.369 -68.694 54.808
378054 91.419 -68.692 54.792
378055 91.469 -68.691 54.776
378056 91.520 -68.689 54.759
378057 91.570 -68.687 54.742
378058 91.620 -68.686 54.724
378059 91.665 -68.686 54.707
378060 91.708 -68.686 54.691
378061 91.750 -68.687 54.673
378062 91.792 -68.688 54.658
378063 91.834 -68.689 54.645
378064 91.876 -68.689 54.633
378065 91.921 -68.686 54.622
378066 91.965 -68.680 54.613
378067 92.010 -68.674 54.605
378068 92.055 -68.668 54.596
378069 92.099 -68.663 54.587
378070 92.144 -68.658 54.578
378071 92.189 -68.651 54.568
378072 92.233 -68.646 54.558
378073 92.278 -68.640 54.547
378074 92.323 -68.635 54.536
378075 92.368 -68.629 54.524
378076 92.412 -68.623 54.513
378077 92.457 -68.617 54.503
378078 92.502 -68.612 54.493
378079 92.547 -68.606 54.483
378080 92.591 -68.601 54.469
378081 92.636 -68.595 54.452
378082 92.681 -68.589 54.432
378083 92.725 -68.583 54.411
378084 92.770 -68.578 54.388
378085 92.815 -68.572 54.364
378086 92.862 -68.568 54.339
378087 92.911 -68.566 54.311
378088 92.960 -68.563 54.282
378089 93.010 -68.560 54.253
378090 93.060 -68.557 54.225
378091 93.110 -68.554 54.197
378092 93.158 -68.552 54.170
378093 93.207 -68.549 54.142
378094 93.256 -68.545 54.114
378095 93.301 -68.535 54.087
378096 93.347 -68.526 54.061
378097 93.393 -68.517 54.035
378098 93.438 -68.507 54.012
378099 93.484 -68.497 53.988
378100 93.529 -68.488 53.964
378101 93.575 -68.478 53.941
378102 93.620 -68.469 53.919
378103 93.666 -68.459 53.898
378104 93.711 -68.449 53.876
378105 93.757 -68.440 53.855
378106 93.802 -68.430 53.835
378107 93.848 -68.421 53.814
378108 93.894 -68.411 53.794
378109 93.939 -68.401 53.775
378110 93.985 -68.392 53.755
378111 94.031 -68.382 53.736
378112 94.076 -68.373 53.717
378113 94.122 -68.363 53.700
378114 94.167 -68.354 53.682
378115 94.213 -68.344 53.663
378116 94.258 -68.335 53.646
378117 94.304 -68.325 53.629
378118 94.349 -68.316 53.611
378119 94.395 -68.306 53.595
378120 94.440 -68.297 53.578
378121 94.486 -68.287 53.563
378122 94.532 -68.277 53.548
378123 94.577 -68.268 53.533
378132 94.880 -68.249 53.424
378133 94.911 -68.238 53.411
378134 94.942 -68.227 53.397
378135 94.972 -68.216 53.383
378136 95.002 -68.205 53.368
378137 95.033 -68.194 53.354
378138 95.063 -68.183 53.340
378139 95.094 -68.172 53.325
378140 95.124 -68.161 53.310
378141 95.155 -68.150 53.296
378142 95.185 -68.139 53.283
378143 95.216 -68.128 53.269
378144 95.247 -68.118 53.256
378145 95.277 -68.106 53.246
378146 95.306 -68.095 53.237
378147 95.331 -68.083 53.228
378148 95.368 -68.094 53.219
378149 95.401 -68.107 53.210
378150 95.433 -68.120 53.200
378151 95.466 -68.134 53.194
378152 95.505 -68.145 53.187
378153 95.546 -68.156 53.181
378154 95.587 -68.166 53.173
378155 95.628 -68.177 53.164
378156 95.669 -68.186 53.155
378157 95.709 -68.194 53.148
378158 95.750 -68.201 53.140
378159 95.791 -68.209 53.131
378160 95.832 -68.216 53.121
378161 95.873 -68.224 53.109
378162 95.913 -68.231 53.097
378163 95.946 -68.235 53.085
378164 95.984 -68.241 53.072
378165 96.022 -68.246 53.064
378166 96.059 -68.251 53.055
378167 96.097 -68.256 53.045
378168 96.134 -68.261 53.032
378169 96.170 -68.266 53.015
378170 96.205 -68.272 52.996
378171 96.241 -68.276 52.975
378172 96.277 -68.282 52.951
378173 96.313 -68.287 52.927
378174 96.348 -68.291 52.901
378175 96.384 -68.297 52.873
378176 96.420 -68.301 52.846
378177 96.456 -68.307 52.819
378178 96.492 -68.312 52.792
378179 96.528 -68.316 52.764
378180 96.553 -68.319 52.735
378181 96.576 -68.320 52.706
378182 96.599 -68.321 52.676
378183 96.621 -68.322 52.644
378184 96.644 -68.323 52.612
378185 96.671 -68.327 52.580
378186 96.702 -68.332 52.548
378187 96.733 -68.336 52.515
378188 96.764 -68.341 52.481
378189 96.795 -68.346 52.448
378190 96.825 -68.350 52.416
378191 96.856 -68.355 52.386
378192 96.887 -68.360 52.356
378193 96.918 -68.364 52.324
378194 96.946 -68.364 52.294
378195 96.971 -68.359 52.265
378196 96.997 -68.354 52.239
378197 97.037 -68.398 52.219
378198 97.097 -68.414 52.203
378199 97.155 -68.414 52.191
378200 97.196 -68.416 52.184
378201 97.232 -68.421 52.186
378202 97.264 -68.430 52.190
378203 97.298 -68.439 52.197
378204 97.329 -68.447 52.202
378205 97.352 -68.441 52.199
378206 97.375 -68.436 52.188
378207 97.397 -68.432 52.170
378208 97.420 -68.426 52.147
378209 97.442 -68.421 52.120
378210 97.472 -68.413 52.089
378211 97.503 -68.406 52.056
378212 97.534 -68.404 52.022
378213 97.568 -68.407 51.987
378214 97.602 -68.410 51.951
378215 97.636 -68.414 51.915
378216 97.670 -68.417 51.879
378217 97.704 -68.421 51.844
378218 97.737 -68.424 51.810
378219 97.772 -68.429 51.775
378220 97.811 -68.435 51.738
378221 97.853 -68.443 51.697
378222 97.895 -68.450 51.649
378223 97.937 -68.457 51.596
378224 97.979 -68.464 51.543
378225 98.020 -68.472 51.490
378226 98.062 -68.479 51.435
378227 98.101 -68.485 51.381
378228 98.140 -68.492 51.323
378229 98.177 -68.498 51.265
378230 98.184 -68.500 51.209
378231 98.189 -68.501 51.151
378232 98.203 -68.504 51.096
378233 98.253 -68.512 51.043
378234 98.303 -68.520 50.989
378235 98.353 -68.528 50.936
378236 98.391 -68.540 50.883
378237 98.426 -68.553 50.832
378238 98.462 -68.566 50.782
378239 98.497 -68.579 50.733
378240 98.533 -68.592 50.686
378241 98.568 -68.604 50.638
378242 98.603 -68.617 50.592
378243 98.639 -68.629 50.547
378244 98.674 -68.642 50.502
378245 98.711 -68.655 50.456
378246 98.752 -68.667 50.412
378247 98.792 -68.680 50.366
378248 98.817 -68.683 50.321
378249 98.860 -68.694 50.278
378250 98.899 -68.695 50.234
378251 98.928 -68.686 50.190
378252 98.958 -68.678 50.147
378253 98.988 -68.669 50.103
378254 99.018 -68.660 50.055
378255 99.048 -68.651 50.007
378256 99.082 -68.668 49.960
378257 99.121 -68.702 49.913
378258 99.158 -68.737 49.865
378259 99.197 -68.771 49.817
378260 99.234 -68.806 49.769
378261 99.261 -68.828 49.720
378262 99.277 -68.839 49.670
378263 99.293 -68.850 49.621
378264 99.310 -68.861 49.573
378265 99.327 -68.871 49.525
378266 99.343 -68.882 49.478
378267 99.359 -68.893 49.433
378268 99.403 -68.899 49.390
378269 99.451 -68.904 49.348
378270 99.499 -68.909 49.306
378271 99.547 -68.915 49.264
378272 99.595 -68.921 49.223
378273 99.643 -68.926 49.181
378274 99.691 -68.931 49.138
378275 99.736 -68.939 49.094
378276 99.782 -68.943 49.047
378277 99.828 -68.949 48.998
378278 99.874 -68.954 48.949
378279 99.919 -68.958 48.900
378280 99.965 -68.964 48.850
378281 100.011 -68.969 48.798
378282 100.057 -68.974 48.745
378283 100.102 -68.979 48.692
378284 100.146 -68.984 48.638
378285 100.168 -68.990 48.583
378286 100.208 -69.009 48.531
378287 100.223 -69.046 48.478
378288 100.265 -68.990 48.427
378289 100.295 -68.987 48.377
378290 100.326 -68.983 48.330
378291 100.356 -68.980 48.285
378292 100.388 -69.005 48.242
378293 100.420 -69.044 48.201
378294 100.452 -69.083 48.160
378295 100.484 -69.121 48.120
378296 100.516 -69.161 48.081
378297 100.547 -69.185 48.042
378298 100.576 -69.190 48.003
378299 100.605 -69.194 47.968
378300 100.634 -69.199 47.932
378301 100.664 -69.203 47.897
378739 90.645 -68.376 54.759
378740 90.676 -68.386 54.757
378741 90.708 -68.397 54.759
378742 90.739 -68.407 54.762
378743 90.770 -68.418 54.766
378744 90.801 -68.429 54.772
378745 90.855 -68.455 54.780
378746 90.913 -68.484 54.790
378747 90.972 -68.513 54.799
378748 91.030 -68.542 54.807
378749 91.089 -68.572 54.814
378750 91.147 -68.601 54.821
378751 91.205 -68.630 54.824
378752 91.261 -68.643 54.822
378753 91.312 -68.639 54.814
378754 91.362 -68.637 54.801
378755 91.412 -68.636 54.785
378756 91.462 -68.633 54.769
378757 91.512 -68.632 54.751
378758 91.563 -68.630 54.734
378759 91.612 -68.628 54.716
378760 91.661 -68.627 54.699
378761 91.703 -68.628 54.683
378762 91.745 -68.629 54.667
378763 91.787 -68.629 54.652
378764 91.829 -68.630 54.637
378765 91.872 -68.631 54.626
378766 91.914 -68.632 54.616
378767 91.958 -68.628 54.608
378768 92.003 -68.623 54.600
378769 92.048 -68.617 54.592
378770 92.092 -68.611 54.584
378771 92.137 -68.606 54.575
378772 92.181 -68.600 54.567
378773 92.226 -68.595 54.558
378774 92.271 -68.589 54.548
378775 92.316 -68.583 54.538
378776 92.360 -68.577 54.528
378777 92.405 -68.572 54.519
378778 92.450 -68.566 54.510
378779 92.495 -68.560 54.501
378780 92.540 -68.554 54.490
378781 92.584 -68.549 54.475
378782 92.629 -68.543 54.458
378783 92.673 -68.538 54.438
378784 92.718 -68.532 54.417
378785 92.763 -68.526 54.394
378786 92.812 -68.524 54.370
378787 92.862 -68.521 54.344
378788 92.911 -68.519 54.316
378789 92.960 -68.517 54.287
378790 93.010 -68.513 54.259
378791 93.060 -68.510 54.231
378792 93.110 -68.508 54.204
378793 93.159 -68.505 54.177
378794 93.209 -68.502 54.150
378795 93.258 -68.500 54.123
378796 93.306 -68.497 54.097
378797 93.355 -68.494 54.071
378798 93.403 -68.488 54.047
378799 93.448 -68.479 54.024
378800 93.494 -68.470 54.001
378801 93.540 -68.460 53.978
378802 93.585 -68.451 53.956
378803 93.631 -68.441 53.934
378804 93.677 -68.432 53.913
378805 93.722 -68.422 53.893
378806 93.768 -68.413 53.872
378807 93.813 -68.403 53.851
378808 93.859 -68.393 53.831
378809 93.904 -68.384 53.811
378810 93.950 -68.374 53.791
378811 93.995 -68.365 53.771
378812 94.041 -68.355 53.751
378813 94.086 -68.345 53.733
378814 94.132 -68.336 53.714
378815 94.178 -68.326 53.695
378816 94.223 -68.317 53.677
378817 94.269 -68.307 53.658
378818 94.315 -68.297 53.640
378819 94.360 -68.288 53.621
378820 94.406 -68.278 53.603
378821 94.451 -68.269 53.586
378822 94.497 -68.260 53.569
378823 94.542 -68.250 53.554
378824 94.588 -68.241 53.539
378825 94.633 -68.231 53.524
378826 94.679 -68.221 53.511
378834 94.916 -68.202 53.412
378835 94.947 -68.191 53.397
378836 94.977 -68.180 53.381
378837 95.007 -68.169 53.366
378838 95.038 -68.158 53.351
378839 95.068 -68.147 53.336
378840 95.099 -68.136 53.320
378841 95.130 -68.125 53.305
378842 95.160 -68.115 53.290
378843 95.191 -68.103 53.275
378844 95.221 -68.093 53.260
378845 95.247 -68.080 53.246
378846 95.272 -68.067 53.235
378847 95.297 -68.053 53.224
378848 95.331 -68.056 53.214
378849 95.368 -68.065 53.205
378850 95.403 -68.079 53.195
378851 95.438 -68.092 53.186
378852 95.470 -68.106 53.178
378853 95.508 -68.117 53.169
378854 95.549 -68.128 53.161
378855 95.590 -68.137 53.152
378856 95.631 -68.144 53.142
378857 95.672 -68.152 53.133
378858 95.712 -68.159 53.124
378859 95.753 -68.167 53.116
378860 95.794 -68.175 53.106
378861 95.834 -68.182 53.096
378862 95.875 -68.190 53.085
378863 95.916 -68.197 53.074
378864 95.947 -68.200 53.063
378865 95.985 -68.206 53.052
378866 96.023 -68.211 53.046
378867 96.060 -68.216 53.038
378868 96.098 -68.222 53.027
378869 96.136 -68.227 53.013
378870 96.173 -68.231 52.995
378871 96.209 -68.237 52.976
378872 96.245 -68.242 52.953
378873 96.281 -68.247 52.928
378874 96.317 -68.252 52.904
378875 96.353 -68.257 52.878
378876 96.388 -68.262 52.851
378877 96.424 -68.267 52.825
378878 96.460 -68.272 52.798
378879 96.496 -68.277 52.772
378880 96.531 -68.282 52.743
378881 96.558 -68.285 52.714
378882 96.580 -68.285 52.685
378883 96.603 -68.287 52.658
378884 96.625 -68.288 52.628
378885 96.650 -68.290 52.597
378886 96.680 -68.294 52.566
378887 96.711 -68.299 52.535
378888 96.742 -68.303 52.503
378889 96.772 -68.308 52.471
378890 96.803 -68.313 52.439
378891 96.834 -68.318 52.408
378892 96.865 -68.322 52.380
378893 96.896 -68.327 52.350
378894 96.927 -68.332 52.320
378895 96.957 -68.337 52.290
378896 96.985 -68.336 52.263
378897 97.010 -68.330 52.238
378898 97.041 -68.343 52.220
378899 97.101 -68.376 52.206
378900 97.158 -68.376 52.196
378901 97.194 -68.379 52.192
378902 97.229 -68.383 52.195
378903 97.262 -68.392 52.201
378904 97.295 -68.402 52.206
378905 97.328 -68.411 52.210
378906 97.359 -68.417 52.203
378907 97.382 -68.412 52.186
378908 97.404 -68.407 52.164
378909 97.427 -68.401 52.138
378910 97.449 -68.396 52.108
378911 97.478 -68.389 52.075
378912 97.509 -68.382 52.042
378913 97.539 -68.374 52.008
378914 97.573 -68.376 51.972
378915 97.606 -68.379 51.937
378916 97.640 -68.382 51.901
378917 97.674 -68.385 51.865
378918 97.708 -68.389 51.832
378919 97.742 -68.394 51.800
378920 97.777 -68.398 51.765
378921 97.813 -68.404 51.728
378922 97.855 -68.410 51.686
378923 97.896 -68.418 51.637
378924 97.938 -68.425 51.584
378925 97.980 -68.432 51.531
378926 98.022 -68.439 51.477
378927 98.064 -68.447 51.422
378928 98.105 -68.454 51.368
378929 98.147 -68.461 51.312
378930 98.173 -68.466 51.256
378931 98.189 -68.470 51.200
378932 98.193 -68.471 51.142
378933 98.210 -68.474 51.086
378934 98.260 -68.482 51.032
378935 98.310 -68.490 50.977
378936 98.361 -68.498 50.921
378937 98.405 -68.507 50.866
378938 98.440 -68.520 50.816
378939 98.476 -68.533 50.767
378940 98.511 -68.546 50.720
378941 98.547 -68.559 50.675
378942 98.582 -68.572 50.631
378943 98.618 -68.585 50.587
378944 98.653 -68.598 50.543
378945 98.689 -68.610 50.497
378946 98.724 -68.623 50.451
378947 98.759 -68.636 50.405
378948 98.796 -68.648 50.361
378949 98.832 -68.657 50.317
378950 98.866 -68.653 50.275
378951 98.896 -68.645 50.230
378952 98.926 -68.636 50.186
378953 98.955 -68.626 50.141
378954 98.986 -68.618 50.095
378955 99.015 -68.609 50.047
378956 99.046 -68.601 49.998
378957 99.083 -68.632 49.950
378958 99.121 -68.667 49.902
378959 99.159 -68.701 49.855
378960 99.197 -68.736 49.809
378961 99.235 -68.770 49.762
378962 99.265 -68.796 49.714
378963 99.282 -68.807 49.666
378964 99.298 -68.817 49.618
378965 99.314 -68.828 49.569
378966 99.331 -68.839 49.520
378967 99.347 -68.849 49.473
378968 99.364 -68.860 49.427
378969 99.400 -68.868 49.384
378970 99.448 -68.873 49.343
378971 99.496 -68.878 49.302
378972 99.544 -68.883 49.260
378973 99.591 -68.889 49.219
378974 99.645 -68.893 49.178
378975 99.701 -68.896 49.136
378976 99.744 -68.904 49.093
378977 99.794 -68.909 49.047
378978 99.836 -68.913 48.997
378979 99.877 -68.918 48.948
378980 99.919 -68.922 48.898
378981 99.964 -68.927 48.847
378982 100.010 -68.932 48.796
378983 100.056 -68.937 48.743
378984 100.102 -68.942 48.689
378985 100.147 -68.947 48.634
378986 100.177 -68.953 48.580
378987 100.193 -69.009 48.527
378988 100.235 -68.934 48.475
378989 100.265 -68.931 48.423
378990 100.296 -68.927 48.374
378991 100.326 -68.924 48.328
378992 100.358 -68.951 48.283
378993 100.390 -68.990 48.240
378994 100.422 -69.028 48.198
378995 100.454 -69.067 48.157
378996 100.486 -69.106 48.118
378997 100.518 -69.140 48.079
378998 100.547 -69.145 48.039
378999 100.576 -69.150 48.001
379000 100.605 -69.154 47.964
379001 100.634 -69.159 47.928
379440 90.634 -68.344 54.759
379441 90.664 -68.354 54.758
379442 90.695 -68.364 54.759
379443 90.727 -68.375 54.762
379444 90.758 -68.385 54.768
379445 90.791 -68.398 54.774
379446 90.849 -68.426 54.782
379447 90.907 -68.456 54.791
379448 90.966 -68.485 54.799
379449 91.024 -68.514 54.806
379450 91.083 -68.544 54.813
379451 91.141 -68.573 54.819
379452 91.202 -68.605 54.820
379453 91.253 -68.595 54.817
379454 91.304 -68.583 54.807
379455 91.355 -68.579 54.795
379456 91.405 -68.578 54.779
379457 91.455 -68.576 54.762
379458 91.505 -68.575 54.744
379459 91.555 -68.573 54.727
379460 91.605 -68.571 54.709
379461 91.656 -68.570 54.692
379462 91.699 -68.570 54.676
379463 91.741 -68.570 54.660
379464 91.783 -68.572 54.645
379465 91.825 -68.573 54.632
379466 91.867 -68.573 54.620
379467 91.910 -68.574 54.611
379468 91.952 -68.575 54.603
379469 91.996 -68.571 54.595
379470 92.040 -68.566 54.588
379471 92.085 -68.560 54.580
379472 92.130 -68.554 54.573
379473 92.174 -68.548 54.566
379474 92.219 -68.543 54.558
379475 92.264 -68.537 54.550
379476 92.308 -68.532 54.540
379477 92.353 -68.526 54.532
379478 92.398 -68.520 54.524
379479 92.443 -68.514 54.516
379480 92.488 -68.509 54.507
379481 92.532 -68.503 54.495
379482 92.577 -68.498 54.479
379483 92.622 -68.492 54.461
379484 92.666 -68.486 54.442
379485 92.714 -68.482 54.421
379486 92.763 -68.480 54.398
379487 92.812 -68.477 54.375
379488 92.862 -68.475 54.349
379489 92.911 -68.472 54.320
379490 92.960 -68.470 54.292
379491 93.010 -68.466 54.264
379492 93.060 -68.464 54.237
379493 93.110 -68.461 54.209
379494 93.160 -68.458 54.183
379495 93.210 -68.456 54.156
379496 93.259 -68.453 54.130
379497 93.307 -68.450 54.105
379498 93.356 -68.448 54.081
379499 93.405 -68.445 54.058
379500 93.454 -68.441 54.036
379501 93.503 -68.439 54.013
379502 93.550 -68.432 53.991
379503 93.596 -68.423 53.970
379504 93.641 -68.413 53.949
379505 93.687 -68.404 53.929
379506 93.732 -68.394 53.909
379507 93.778 -68.385 53.889
379508 93.824 -68.376 53.868
379509 93.869 -68.366 53.848
379510 93.915 -68.357 53.828
379511 93.961 -68.347 53.808
379512 94.006 -68.337 53.788
379513 94.052 -68.328 53.768
379514 94.097 -68.318 53.749
379515 94.143 -68.309 53.729
379516 94.188 -68.299 53.709
379517 94.234 -68.289 53.689
379518 94.279 -68.280 53.670
379519 94.325 -68.270 53.651
379520 94.370 -68.261 53.632
379521 94.416 -68.251 53.613
379522 94.462 -68.241 53.594
379523 94.507 -68.232 53.577
379524 94.553 -68.222 53.560
379525 94.599 -68.213 53.544
379526 94.644 -68.203 53.529
379527 94.690 -68.194 53.515
379528 94.718 -68.207 53.502
379536 94.952 -68.155 53.397
379537 94.982 -68.144 53.381
379538 95.013 -68.133 53.365
379539 95.043 -68.122 53.348
379540 95.074 -68.112 53.332
379541 95.104 -68.100 53.316
379542 95.135 -68.090 53.300
379543 95.164 -68.078 53.284
379544 95.188 -68.065 53.267
379545 95.213 -68.052 53.252
379546 95.238 -68.038 53.237
379547 95.263 -68.025 53.224
379548 95.293 -68.020 53.212
379549 95.330 -68.028 53.200
379550 95.367 -68.037 53.190
379551 95.402 -68.050 53.181
379552 95.438 -68.064 53.172
379553 95.473 -68.078 53.163
379554 95.511 -68.087 53.152
379555 95.552 -68.095 53.142
379556 95.593 -68.103 53.132
379557 95.633 -68.110 53.121
379558 95.674 -68.118 53.111
379559 95.715 -68.125 53.101
379560 95.756 -68.133 53.092
379561 95.797 -68.141 53.082
379562 95.837 -68.148 53.071
379563 95.878 -68.156 53.061
379564 95.918 -68.163 53.051
379565 95.949 -68.166 53.041
379566 95.986 -68.171 53.032
379567 96.024 -68.176 53.026
379568 96.061 -68.181 53.018
379569 96.099 -68.187 53.007
379570 96.137 -68.192 52.993
379571 96.175 -68.197 52.975
379572 96.213 -68.202 52.956
379573 96.249 -68.207 52.934
379574 96.285 -68.213 52.909
379575 96.320 -68.217 52.885
379576 96.356 -68.222 52.858
379577 96.392 -68.228 52.831
379578 96.428 -68.232 52.803
379579 96.464 -68.238 52.777
379580 96.500 -68.242 52.751
379581 96.535 -68.247 52.723
379582 96.562 -68.250 52.694
379583 96.585 -68.251 52.667
379584 96.607 -68.252 52.640
379585 96.629 -68.253 52.612
379586 96.659 -68.257 52.583
379587 96.689 -68.262 52.553
379588 96.720 -68.266 52.523
379589 96.751 -68.271 52.492
379590 96.782 -68.276 52.461
379591 96.812 -68.281 52.430
379592 96.843 -68.285 52.399
379593 96.874 -68.290 52.371
379594 96.904 -68.294 52.343
379595 96.935 -68.300 52.314
379596 96.966 -68.304 52.286
379597 96.997 -68.309 52.260
379598 97.024 -68.307 52.239
379599 97.049 -68.302 52.224
379600 97.106 -68.337 52.212
379601 97.155 -68.338 52.203
379602 97.191 -68.341 52.200
379603 97.227 -68.345 52.202
379604 97.260 -68.355 52.207
379605 97.293 -68.364 52.212
379606 97.326 -68.373 52.213
379607 97.359 -68.382 52.202
379608 97.389 -68.387 52.183
379609 97.411 -68.382 52.158
379610 97.434 -68.376 52.130
379611 97.456 -68.371 52.098
379612 97.485 -68.364 52.064
379613 97.515 -68.357 52.031
379614 97.545 -68.350 51.996
379615 97.577 -68.344 51.960
379616 97.610 -68.347 51.925
379617 97.644 -68.350 51.889
379618 97.678 -68.354 51.853
379619 97.712 -68.358 51.821
379620 97.747 -68.363 51.790
379621 97.781 -68.367 51.756
379622 97.816 -68.372 51.718
379623 97.856 -68.379 51.677
379624 97.898 -68.385 51.629
379625 97.940 -68.393 51.578
379626 97.982 -68.400 51.524
379627 98.023 -68.407 51.471
379628 98.065 -68.415 51.418
379629 98.105 -68.422 51.366
379630 98.127 -68.426 51.311
379631 98.150 -68.431 51.254
379632 98.172 -68.435 51.197
379633 98.195 -68.440 51.140
379634 98.217 -68.444 51.083
379635 98.267 -68.452 51.028
379636 98.315 -68.460 50.973
379637 98.361 -68.470 50.917
379638 98.408 -68.479 50.863
379639 98.445 -68.491 50.812
379640 98.481 -68.504 50.764
379641 98.516 -68.517 50.717
379642 98.552 -68.529 50.671
379643 98.587 -68.542 50.628
379644 98.623 -68.555 50.584
379645 98.658 -68.568 50.540
379646 98.694 -68.580 50.494
379647 98.729 -68.593 50.447
379648 98.764 -68.606 50.401
379649 98.800 -68.619 50.355
379650 98.834 -68.617 50.311
379651 98.863 -68.602 50.267
379652 98.894 -68.594 50.223
379653 98.923 -68.585 50.178
379654 98.953 -68.576 50.132
379655 98.983 -68.567 50.085
379656 99.013 -68.558 50.036
379657 99.045 -68.561 49.987
379658 99.083 -68.595 49.937
379659 99.121 -68.630 49.889
379660 99.159 -68.664 49.842
379661 99.201 -68.705 49.797
379662 99.244 -68.744 49.751
379663 99.279 -68.758 49.705
379664 99.300 -68.770 49.658
379665 99.312 -68.781 49.611
379666 99.325 -68.793 49.562
379667 99.338 -68.805 49.514
379668 99.352 -68.817 49.467
379669 99.368 -68.827 49.423
379670 99.397 -68.836 49.379
379671 99.444 -68.842 49.337
379672 99.500 -68.845 49.296
379673 99.556 -68.848 49.255
379674 99.613 -68.851 49.214
379675 99.669 -68.854 49.173
379676 99.715 -68.858 49.132
379677 99.749 -68.871 49.089
379678 99.802 -68.876 49.044
379679 99.856 -68.880 48.995
379680 99.905 -68.885 48.946
379681 99.946 -68.889 48.897
379682 99.987 -68.894 48.846
379683 100.028 -68.899 48.795
379684 100.070 -68.902 48.742
379685 100.111 -68.907 48.688
379686 100.153 -68.911 48.633
379687 100.171 -68.946 48.576
379688 100.201 -68.898 48.523
379689 100.235 -68.876 48.471
379690 100.266 -68.872 48.421
379691 100.296 -68.869 48.372
379692 100.328 -68.896 48.326
379693 100.360 -68.935 48.281
379694 100.392 -68.974 48.238
379695 100.424 -69.012 48.196
379696 100.457 -69.052 48.155
379697 100.489 -69.090 48.115
379698 100.518 -69.099 48.076
379699 100.547 -69.105 48.036
379700 100.576 -69.109 47.998
379701 100.605 -69.114 47.961
380140 90.590 -68.300 54.762
380141 90.621 -68.311 54.760
380142 90.653 -68.322 54.760
380143 90.683 -68.332 54.761
380144 90.715 -68.343 54.765
380145 90.746 -68.354 54.771
380146 90.785 -68.369 54.778
380147 90.843 -68.398 54.786
380148 90.901 -68.428 54.794
380149 90.960 -68.457 54.802
380150 91.018 -68.486 54.810
380151 91.077 -68.516 54.815
380152 91.135 -68.545 54.818
380153 91.193 -68.557 54.817
380154 91.244 -68.545 54.811
380155 91.296 -68.535 54.801
380156 91.347 -68.524 54.788
380157 91.397 -68.521 54.772
380158 91.448 -68.519 54.755
380159 91.498 -68.517 54.738
380160 91.548 -68.516 54.721
380161 91.598 -68.513 54.703
380162 91.648 -68.512 54.686
380163 91.694 -68.512 54.669
380164 91.736 -68.513 54.653
380165 91.778 -68.513 54.639
380166 91.820 -68.514 54.627
380167 91.863 -68.515 54.615
380168 91.905 -68.516 54.605
380169 91.947 -68.517 54.598
380170 91.989 -68.517 54.591
380171 92.033 -68.514 54.584
380172 92.078 -68.508 54.577
380173 92.123 -68.503 54.571
380174 92.167 -68.497 54.565
380175 92.212 -68.492 54.559
380176 92.257 -68.485 54.551
380177 92.301 -68.480 54.543
380178 92.346 -68.474 54.535
380179 92.391 -68.469 54.527
380180 92.436 -68.463 54.519
380181 92.480 -68.457 54.509
380182 92.525 -68.452 54.497
380183 92.570 -68.446 54.480
380184 92.615 -68.441 54.463
380185 92.664 -68.438 54.444
380186 92.714 -68.435 54.423
380187 92.763 -68.433 54.400
380188 92.812 -68.431 54.377
380189 92.862 -68.428 54.351
380190 92.911 -68.426 54.323
380191 92.960 -68.423 54.295
380192 93.010 -68.420 54.268
380193 93.060 -68.417 54.241
380194 93.110 -68.414 54.214
380195 93.160 -68.412 54.188
380196 93.210 -68.409 54.162
380197 93.260 -68.406 54.136
380198 93.308 -68.404 54.113
380199 93.357 -68.401 54.090
380200 93.407 -68.398 54.067
380201 93.455 -68.395 54.046
380202 93.504 -68.392 54.025
380203 93.553 -68.389 54.004
380204 93.602 -68.387 53.983
380205 93.651 -68.384 53.964
380206 93.698 -68.376 53.944
380207 93.744 -68.367 53.924
380208 93.789 -68.357 53.905
380209 93.835 -68.348 53.885
380210 93.880 -68.338 53.865
380211 93.926 -68.329 53.845
380212 93.971 -68.319 53.825
380213 94.017 -68.310 53.806
380214 94.062 -68.300 53.785
380215 94.108 -68.291 53.765
380216 94.153 -68.281 53.745
380217 94.199 -68.272 53.724
380218 94.245 -68.262 53.703
380219 94.290 -68.253 53.683
380220 94.336 -68.243 53.663
380221 94.382 -68.233 53.643
380222 94.427 -68.224 53.623
380223 94.473 -68.214 53.604
380224 94.518 -68.205 53.585
380225 94.564 -68.195 53.567
380226 94.609 -68.185 53.550
380227 94.655 -68.176 53.535
380228 94.695 -68.173 53.520
380229 94.715 -68.197 53.506
380230 94.743 -68.196 53.494
380231 94.774 -68.185 53.480
380237 94.957 -68.119 53.397
380238 94.987 -68.109 53.381
380239 95.018 -68.097 53.364
380240 95.048 -68.087 53.347
380241 95.079 -68.075 53.330
380242 95.105 -68.062 53.314
380243 95.130 -68.050 53.297
380244 95.155 -68.037 53.279
380245 95.180 -68.024 53.261
380246 95.204 -68.010 53.245
380247 95.229 -67.997 53.229
380248 95.254 -67.984 53.215
380249 95.292 -67.992 53.201
380250 95.330 -68.000 53.188
380251 95.367 -68.009 53.177
380252 95.402 -68.022 53.167
380253 95.438 -68.036 53.157
380254 95.473 -68.049 53.147
380255 95.514 -68.053 53.136
380256 95.555 -68.061 53.126
380257 95.596 -68.068 53.115
380258 95.636 -68.076 53.103
380259 95.677 -68.084 53.091
380260 95.718 -68.091 53.080
380261 95.758 -68.099 53.069
380262 95.799 -68.106 53.058
380263 95.840 -68.114 53.049
380264 95.881 -68.121 53.039
380265 95.920 -68.128 53.029
380266 95.951 -68.131 53.020
380267 95.986 -68.136 53.011
380268 96.025 -68.141 53.003
380269 96.062 -68.147 52.994
380270 96.100 -68.152 52.983
380271 96.138 -68.157 52.969
380272 96.176 -68.162 52.955
380273 96.214 -68.167 52.936
380274 96.251 -68.172 52.915
380275 96.289 -68.178 52.891
380276 96.324 -68.183 52.867
380277 96.360 -68.188 52.839
380278 96.396 -68.193 52.811
380279 96.432 -68.198 52.782
380280 96.467 -68.203 52.755
380281 96.503 -68.208 52.730
380282 96.539 -68.213 52.703
380283 96.566 -68.216 52.676
380284 96.589 -68.216 52.650
380285 96.611 -68.218 52.623
380286 96.637 -68.220 52.597
380287 96.667 -68.225 52.569
380288 96.698 -68.229 52.541
380289 96.729 -68.234 52.510
380290 96.759 -68.239 52.480
380291 96.791 -68.244 52.452
380292 96.821 -68.248 52.422
380293 96.852 -68.253 52.392
380294 96.883 -68.257 52.363
380295 96.913 -68.263 52.336
380296 96.944 -68.267 52.309
380297 96.975 -68.272 52.283
380298 97.006 -68.276 52.259
380299 97.036 -68.281 52.242
380300 97.064 -68.278 52.228
380301 97.111 -68.298 52.218
380302 97.153 -68.300 52.210
380303 97.189 -68.303 52.208
380304 97.224 -68.308 52.210
380305 97.257 -68.317 52.214
380306 97.291 -68.326 52.216
380307 97.324 -68.335 52.215
380308 97.357 -68.344 52.202
380309 97.390 -68.354 52.181
380310 97.418 -68.357 52.154
380311 97.441 -68.351 52.124
380312 97.463 -68.347 52.089
380313 97.491 -68.340 52.054
380314 97.521 -68.332 52.020
380315 97.552 -68.325 51.985
380316 97.582 -68.317 51.949
380317 97.615 -68.316 51.915
380318 97.649 -68.319 51.880
380319 97.682 -68.322 51.846
380320 97.717 -68.327 51.815
380321 97.751 -68.332 51.783
380322 97.786 -68.336 51.749
380323 97.820 -68.341 51.711
380324 97.858 -68.347 51.669
380325 97.899 -68.354 51.622
380326 97.941 -68.361 51.571
380327 97.983 -68.368 51.518
380328 98.025 -68.376 51.466
380329 98.059 -68.382 51.415
380330 98.082 -68.386 51.363
380331 98.104 -68.391 51.308
380332 98.126 -68.395 51.252
380333 98.151 -68.401 51.194
380334 98.185 -68.410 51.137
380335 98.218 -68.416 51.080
380336 98.257 -68.427 51.025
380337 98.304 -68.436 50.969
380338 98.350 -68.445 50.914
380339 98.396 -68.454 50.861
380340 98.442 -68.464 50.811
380341 98.478 -68.477 50.762
380342 98.513 -68.490 50.715
380343 98.549 -68.503 50.670
380344 98.584 -68.515 50.626
380345 98.619 -68.528 50.582
380346 98.655 -68.541 50.537
380347 98.690 -68.554 50.490
380348 98.725 -68.567 50.442
380349 98.761 -68.579 50.394
380350 98.798 -68.587 50.348
380351 98.833 -68.580 50.304
380352 98.862 -68.565 50.260
380353 98.891 -68.551 50.215
380354 98.921 -68.535 50.169
380355 98.951 -68.525 50.121
380356 98.981 -68.517 50.074
380357 99.011 -68.507 50.025
380358 99.046 -68.525 49.975
380359 99.083 -68.560 49.925
380360 99.127 -68.601 49.877
380361 99.176 -68.650 49.829
380362 99.219 -68.689 49.784
380363 99.250 -68.707 49.739
380364 99.285 -68.720 49.693
380365 99.325 -68.730 49.647
380366 99.338 -68.742 49.601
380367 99.350 -68.754 49.553
380368 99.363 -68.766 49.506
380369 99.375 -68.778 49.460
380370 99.388 -68.790 49.417
380371 99.411 -68.799 49.373
380372 99.468 -68.803 49.331
380373 99.523 -68.806 49.290
380374 99.580 -68.809 49.248
380375 99.636 -68.812 49.208
380376 99.692 -68.815 49.168
380377 99.726 -68.820 49.127
380378 99.754 -68.839 49.084
380379 99.808 -68.843 49.040
380380 99.861 -68.848 48.992
380381 99.915 -68.852 48.944
380382 99.969 -68.857 48.895
380383 100.014 -68.861 48.845
380384 100.056 -68.866 48.794
380385 100.097 -68.871 48.740
380386 100.139 -68.874 48.686
380387 100.148 -68.890 48.630
380388 100.158 -68.906 48.572
380389 100.206 -68.821 48.519
380390 100.236 -68.817 48.468
380391 100.267 -68.814 48.419
380392 100.298 -68.842 48.371
380393 100.330 -68.880 48.325
380394 100.362 -68.919 48.280
380395 100.395 -68.958 48.236
380396 100.427 -68.997 48.195
380397 100.459 -69.036 48.153
380398 100.489 -69.055 48.113
380399 100.518 -69.059 48.073
380400 100.547 -69.064 48.034
380401 100.576 -69.068 47.995
380841 90.578 -68.268 54.764
380842 90.609 -68.278 54.762
380843 90.641 -68.289 54.763
380844 90.672 -68.300 54.765
380845 90.702 -68.310 54.770
380846 90.734 -68.321 54.776
380847 90.779 -68.341 54.783
380848 90.837 -68.370 54.791
380849 90.895 -68.400 54.798
380850 90.954 -68.429 54.806
380851 91.012 -68.458 54.813
380852 91.071 -68.487 54.818
380853 91.134 -68.519 54.819
380854 91.184 -68.507 54.816
380855 91.236 -68.497 54.808
380856 91.287 -68.486 54.797
380857 91.338 -68.476 54.783
380858 91.390 -68.465 54.767
380859 91.440 -68.462 54.750
380860 91.491 -68.460 54.733
380861 91.541 -68.458 54.715
380862 91.591 -68.457 54.698
380863 91.641 -68.455 54.681
380864 91.689 -68.454 54.665
380865 91.731 -68.454 54.649
380866 91.774 -68.455 54.635
380867 91.816 -68.456 54.622
380868 91.858 -68.457 54.611
380869 91.900 -68.457 54.602
380870 91.942 -68.458 54.595
380871 91.985 -68.460 54.588
380872 92.027 -68.460 54.582
380873 92.071 -68.457 54.576
380874 92.116 -68.451 54.571
380875 92.160 -68.445 54.565
380876 92.205 -68.440 54.559
380877 92.249 -68.434 54.553
380878 92.294 -68.429 54.546
380879 92.339 -68.423 54.538
380880 92.384 -68.417 54.530
380881 92.428 -68.412 54.521
380882 92.473 -68.406 54.510
380883 92.518 -68.401 54.498
380884 92.566 -68.397 54.481
380885 92.615 -68.394 54.463
380886 92.664 -68.391 54.444
380887 92.714 -68.389 54.423
380888 92.763 -68.386 54.401
380889 92.812 -68.384 54.378
380890 92.862 -68.382 54.353
380891 92.911 -68.379 54.325
380892 92.960 -68.376 54.297
380893 93.010 -68.373 54.270
380894 93.060 -68.371 54.244
380895 93.110 -68.368 54.217
380896 93.160 -68.365 54.192
380897 93.210 -68.362 54.166
380898 93.260 -68.360 54.142
380899 93.310 -68.357 54.119
380900 93.359 -68.354 54.097
380901 93.408 -68.351 54.076
380902 93.457 -68.348 54.056
380903 93.505 -68.345 54.036
380904 93.554 -68.343 54.016
380905 93.603 -68.340 53.997
380906 93.652 -68.337 53.978
380907 93.701 -68.335 53.959
380908 93.750 -68.332 53.940
380909 93.799 -68.329 53.921
380910 93.845 -68.320 53.901
380911 93.891 -68.311 53.881
380912 93.936 -68.301 53.862
380913 93.982 -68.292 53.842
380914 94.028 -68.282 53.824
380915 94.073 -68.272 53.803
380916 94.119 -68.263 53.782
380917 94.164 -68.253 53.760
380918 94.210 -68.244 53.739
380919 94.255 -68.235 53.718
380920 94.301 -68.225 53.697
380921 94.346 -68.216 53.676
380922 94.392 -68.206 53.655
380923 94.437 -68.197 53.633
380924 94.483 -68.187 53.613
380925 94.529 -68.177 53.594
380926 94.574 -68.168 53.575
380927 94.620 -68.158 53.556
380928 94.666 -68.149 53.540
380929 94.693 -68.162 53.525
380930 94.718 -68.171 53.511
380931 94.749 -68.160 53.497
380932 94.779 -68.149 53.484
380933 94.810 -68.138 53.471
380939 94.993 -68.072 53.381
380940 95.021 -68.061 53.363
380941 95.046 -68.048 53.345
380942 95.071 -68.034 53.328
380943 95.096 -68.021 53.310
380944 95.121 -68.009 53.293
380945 95.145 -67.995 53.275
380946 95.171 -67.982 53.256
380947 95.195 -67.969 53.239
380948 95.220 -67.956 53.222
380949 95.253 -67.956 53.207
380950 95.291 -67.965 53.192
380951 95.329 -67.973 53.178
380952 95.367 -67.981 53.165
380953 95.402 -67.994 53.153
380954 95.437 -68.008 53.142
380955 95.474 -68.018 53.131
380956 95.516 -68.021 53.120
380957 95.557 -68.027 53.109
380958 95.598 -68.034 53.097
380959 95.639 -68.041 53.085
380960 95.680 -68.049 53.073
380961 95.721 -68.056 53.061
380962 95.761 -68.064 53.050
380963 95.802 -68.072 53.039
380964 95.842 -68.079 53.028
380965 95.883 -68.087 53.017
380966 95.922 -68.094 53.006
380967 95.953 -68.097 52.996
380968 95.987 -68.101 52.986
380969 96.025 -68.106 52.977
380970 96.063 -68.112 52.966
380971 96.101 -68.117 52.954
380972 96.139 -68.122 52.941
380973 96.177 -68.127 52.928
380974 96.215 -68.132 52.911
380975 96.252 -68.137 52.890
380976 96.290 -68.143 52.867
380977 96.328 -68.148 52.843
380978 96.364 -68.153 52.817
380979 96.400 -68.158 52.789
380980 96.436 -68.163 52.761
380981 96.471 -68.169 52.735
380982 96.507 -68.173 52.710
380983 96.543 -68.178 52.684
380984 96.571 -68.181 52.658
380985 96.593 -68.182 52.633
380986 96.615 -68.183 52.608
380987 96.646 -68.188 52.583
380988 96.676 -68.192 52.557
380989 96.707 -68.197 52.528
380990 96.738 -68.202 52.497
380991 96.768 -68.206 52.469
380992 96.799 -68.211 52.442
380993 96.830 -68.216 52.414
380994 96.861 -68.220 52.386
380995 96.891 -68.225 52.357
380996 96.922 -68.230 52.330
380997 96.953 -68.235 52.305
380998 96.984 -68.239 52.282
380999 97.015 -68.244 52.262
381000 97.045 -68.248 52.246
381001 97.076 -68.253 52.234
381002 97.114 -68.260 52.225
381003 97.150 -68.263 52.218
381004 97.186 -68.266 52.216
381005 97.222 -68.270 52.218
381006 97.255 -68.279 52.219
381007 97.288 -68.288 52.219
381008 97.321 -68.297 52.216
381009 97.354 -68.307 52.201
381010 97.387 -68.316 52.179
381011 97.421 -68.325 52.152
381012 97.448 -68.327 52.119
381013 97.470 -68.322 52.082
381014 97.497 -68.316 52.045
381015 97.527 -68.308 52.009
381016 97.558 -68.300 51.973
381017 97.589 -68.293 51.937
381018 97.619 -68.285 51.905
381019 97.653 -68.288 51.874
381020 97.687 -68.291 51.842
381021 97.721 -68.296 51.810
381022 97.756 -68.300 51.778
381023 97.791 -68.305 51.742
381024 97.825 -68.310 51.704
381025 97.860 -68.314 51.662
381026 97.901 -68.322 51.614
381027 97.943 -68.329 51.564
381028 97.985 -68.336 51.513
381029 98.013 -68.341 51.462
381030 98.036 -68.346 51.412
381031 98.058 -68.350 51.360
381032 98.091 -68.362 51.305
381033 98.126 -68.374 51.249
381034 98.159 -68.379 51.192
381035 98.191 -68.374 51.134
381036 98.224 -68.387 51.076
381037 98.257 -68.400 51.021
381038 98.293 -68.412 50.966
381039 98.339 -68.421 50.911
381040 98.385 -68.430 50.861
381041 98.432 -68.439 50.811
381042 98.474 -68.450 50.761
381043 98.510 -68.463 50.713
381044 98.545 -68.476 50.668
381045 98.580 -68.488 50.624
381046 98.616 -68.501 50.579
381047 98.651 -68.514 50.533
381048 98.687 -68.527 50.485
381049 98.722 -68.540 50.436
381050 98.759 -68.551 50.387
381051 98.798 -68.553 50.341
381052 98.831 -68.544 50.297
381053 98.860 -68.529 50.254
381054 98.890 -68.514 50.207
381055 98.920 -68.499 50.159
381056 98.949 -68.485 50.110
381057 98.979 -68.470 50.062
381058 99.008 -68.457 50.012
381059 99.054 -68.498 49.963
381060 99.102 -68.547 49.914
381061 99.151 -68.595 49.866
381062 99.195 -68.635 49.817
381063 99.224 -68.652 49.771
381064 99.254 -68.670 49.725
381065 99.290 -68.683 49.680
381066 99.331 -68.692 49.634
381067 99.363 -68.703 49.589
381068 99.375 -68.715 49.544
381069 99.388 -68.727 49.499
381070 99.401 -68.739 49.453
381071 99.413 -68.751 49.409
381072 99.435 -68.761 49.365
381073 99.491 -68.764 49.325
381074 99.547 -68.767 49.284
381075 99.604 -68.770 49.242
381076 99.660 -68.773 49.203
381077 99.703 -68.777 49.163
381078 99.735 -68.785 49.122
381079 99.761 -68.805 49.079
381080 99.813 -68.811 49.035
381081 99.867 -68.815 48.988
381082 99.920 -68.820 48.940
381083 99.974 -68.824 48.892
381084 100.028 -68.829 48.844
381085 100.081 -68.833 48.792
381086 100.124 -68.838 48.738
381087 100.125 -68.857 48.683
381088 100.121 -68.893 48.627
381089 100.176 -68.765 48.570
381090 100.206 -68.761 48.517
381091 100.237 -68.758 48.466
381092 100.268 -68.787 48.417
381093 100.300 -68.826 48.371
381094 100.333 -68.864 48.324
381095 100.365 -68.903 48.279
381096 100.397 -68.942 48.235
381097 100.429 -68.981 48.193
381098 100.460 -69.010 48.151
381099 100.489 -69.015 48.110
381100 100.519 -69.019 48.071
381101 100.547 -69.024 48.031
381541 90.535 -68.225 54.770
381542 90.566 -68.236 54.767
381543 90.597 -68.247 54.766
381544 90.628 -68.257 54.768
381545 90.660 -68.268 54.771
381546 90.691 -68.278 54.776
381547 90.722 -68.288 54.783
381548 90.773 -68.313 54.790
381549 90.831 -68.342 54.798
381550 90.889 -68.371 54.805
381551 90.948 -68.401 54.812
381552 91.006 -68.430 54.817
381553 91.067 -68.462 54.820
381554 91.125 -68.470 54.820
381555 91.176 -68.459 54.815
381556 91.228 -68.448 54.806
381557 91.279 -68.438 54.794
381558 91.330 -68.427 54.779
381559 91.382 -68.416 54.763
381560 91.433 -68.406 54.746
381561 91.484 -68.403 54.729
381562 91.533 -68.401 54.710
381563 91.584 -68.399 54.693
381564 91.634 -68.398 54.677
381565 91.684 -68.396 54.661
381566 91.727 -68.397 54.645
381567 91.769 -68.398 54.631
381568 91.811 -68.398 54.619
381569 91.854 -68.399 54.609
381570 91.895 -68.400 54.600
381571 91.938 -68.401 54.593
381572 91.980 -68.401 54.587
381573 92.022 -68.402 54.582
381574 92.065 -68.403 54.577
381575 92.108 -68.400 54.572
381576 92.153 -68.394 54.567
381577 92.198 -68.388 54.561
381578 92.242 -68.383 54.555
381579 92.287 -68.377 54.548
381580 92.332 -68.372 54.541
381581 92.376 -68.366 54.532
381582 92.421 -68.360 54.521
381583 92.467 -68.355 54.510
381584 92.516 -68.353 54.496
381585 92.566 -68.350 54.480
381586 92.615 -68.347 54.462
381587 92.664 -68.345 54.443
381588 92.714 -68.342 54.421
381589 92.763 -68.340 54.400
381590 92.812 -68.337 54.377
381591 92.862 -68.335 54.353
381592 92.911 -68.332 54.327
381593 92.960 -68.329 54.299
381594 93.010 -68.327 54.272
381595 93.060 -68.324 54.246
381596 93.110 -68.321 54.220
381597 93.160 -68.318 54.195
381598 93.210 -68.316 54.170
381599 93.260 -68.313 54.147
381600 93.310 -68.310 54.125
381601 93.360 -68.307 54.103
381602 93.409 -68.304 54.084
381603 93.458 -68.301 54.065
381604 93.506 -68.299 54.046
381605 93.555 -68.296 54.028
381606 93.605 -68.293 54.010
381607 93.653 -68.291 53.992
381608 93.702 -68.288 53.973
381609 93.751 -68.285 53.955
381610 93.800 -68.282 53.937
381611 93.849 -68.279 53.917
381612 93.898 -68.277 53.898
381613 93.947 -68.274 53.879
381614 93.992 -68.264 53.860
381615 94.038 -68.255 53.841
381616 94.083 -68.245 53.821
381617 94.129 -68.236 53.801
381618 94.175 -68.226 53.779
381619 94.220 -68.216 53.757
381620 94.266 -68.207 53.734
381621 94.312 -68.197 53.711
381622 94.357 -68.188 53.689
381623 94.403 -68.178 53.666
381624 94.448 -68.169 53.644
381625 94.494 -68.159 53.622
381626 94.539 -68.150 53.602
381627 94.585 -68.141 53.582
381628 94.630 -68.131 53.564
381629 94.671 -68.128 53.546
381630 94.693 -68.146 53.530
381631 94.723 -68.135 53.515
381632 94.754 -68.125 53.501
381633 94.784 -68.113 53.487
381634 94.815 -68.103 53.473
381635 94.845 -68.091 53.459
381636 94.876 -68.081 53.445
381641 95.013 -68.019 53.362
381642 95.037 -68.006 53.344
381643 95.062 -67.993 53.325
381644 95.087 -67.981 53.307
381645 95.112 -67.967 53.289
381646 95.137 -67.954 53.271
381647 95.162 -67.941 53.251
381648 95.187 -67.928 53.233
381649 95.215 -67.921 53.215
381650 95.253 -67.929 53.199
381651 95.290 -67.937 53.183
381652 95.328 -67.945 53.169
381653 95.366 -67.953 53.155
381654 95.402 -67.966 53.141
381655 95.437 -67.979 53.128
381656 95.476 -67.987 53.116
381657 95.518 -67.989 53.104
381658 95.559 -67.994 53.092
381659 95.601 -68.001 53.079
381660 95.641 -68.007 53.067
381661 95.682 -68.015 53.055
381662 95.723 -68.022 53.042
381663 95.764 -68.030 53.030
381664 95.805 -68.037 53.018
381665 95.845 -68.045 53.006
381666 95.886 -68.053 52.994
381667 95.924 -68.059 52.983
381668 95.955 -68.062 52.972
381669 95.988 -68.066 52.961
381670 96.026 -68.072 52.950
381671 96.064 -68.077 52.939
381672 96.102 -68.081 52.926
381673 96.139 -68.087 52.912
381674 96.178 -68.092 52.899
381675 96.216 -68.097 52.883
381676 96.253 -68.103 52.863
381677 96.291 -68.108 52.841
381678 96.329 -68.113 52.818
381679 96.367 -68.119 52.793
381680 96.403 -68.124 52.767
381681 96.439 -68.128 52.741
381682 96.475 -68.134 52.715
381683 96.511 -68.138 52.690
381684 96.547 -68.144 52.665
381685 96.575 -68.147 52.642
381686 96.597 -68.147 52.618
381687 96.623 -68.150 52.594
381688 96.654 -68.155 52.569
381689 96.685 -68.159 52.543
381690 96.716 -68.165 52.514
381691 96.747 -68.169 52.485
381692 96.778 -68.174 52.458
381693 96.808 -68.178 52.433
381694 96.839 -68.183 52.407
381695 96.870 -68.188 52.379
381696 96.900 -68.193 52.351
381697 96.931 -68.197 52.326
381698 96.962 -68.202 52.304
381699 96.993 -68.206 52.285
381700 97.023 -68.211 52.267
381701 97.053 -68.216 52.253
381702 97.081 -68.222 52.243
381703 97.113 -68.225 52.235
381704 97.148 -68.225 52.229
381705 97.184 -68.228 52.228
381706 97.220 -68.232 52.229
381707 97.253 -68.241 52.228
381708 97.286 -68.250 52.225
381709 97.319 -68.260 52.218
381710 97.352 -68.269 52.202
381711 97.385 -68.278 52.179
381712 97.418 -68.288 52.150
381713 97.451 -68.297 52.117
381714 97.477 -68.297 52.078
381715 97.503 -68.291 52.039
381716 97.534 -68.283 52.001
381717 97.564 -68.275 51.965
381718 97.595 -68.268 51.932
381719 97.626 -68.260 51.901
381720 97.657 -68.256 51.870
381721 97.691 -68.260 51.837
381722 97.726 -68.265 51.803
381723 97.760 -68.269 51.769
381724 97.795 -68.274 51.732
381725 97.829 -68.279 51.695
381726 97.864 -68.284 51.653
381727 97.902 -68.290 51.606
381728 97.944 -68.297 51.557
381729 97.967 -68.301 51.508
381730 97.998 -68.311 51.458
381731 98.032 -68.323 51.408
381732 98.067 -68.335 51.356
381733 98.101 -68.347 51.300
381734 98.134 -68.348 51.243
381735 98.165 -68.342 51.186
381736 98.198 -68.345 51.128
381737 98.231 -68.358 51.070
381738 98.265 -68.370 51.016
381739 98.298 -68.383 50.963
381740 98.331 -68.396 50.909
381741 98.374 -68.406 50.858
381742 98.421 -68.415 50.807
381743 98.467 -68.425 50.758
381744 98.506 -68.436 50.710
381745 98.541 -68.449 50.664
381746 98.577 -68.462 50.618
381747 98.613 -68.474 50.571
381748 98.648 -68.487 50.523
381749 98.684 -68.500 50.475
381750 98.719 -68.513 50.427
381751 98.759 -68.516 50.381
381752 98.798 -68.519 50.334
381753 98.830 -68.507 50.289
381754 98.859 -68.492 50.244
381755 98.888 -68.478 50.196
381756 98.918 -68.463 50.148
381757 98.950 -68.451 50.099
381758 98.989 -68.448 50.050
381759 99.029 -68.446 50.002
381760 99.077 -68.492 49.953
381761 99.126 -68.541 49.904
381762 99.169 -68.579 49.855
381763 99.199 -68.598 49.808
381764 99.229 -68.615 49.762
381765 99.259 -68.633 49.716
381766 99.296 -68.645 49.670
381767 99.337 -68.654 49.625
381768 99.376 -68.664 49.581
381769 99.401 -68.676 49.535
381770 99.413 -68.687 49.490
381771 99.426 -68.699 49.446
381772 99.438 -68.711 49.402
381773 99.458 -68.722 49.358
381774 99.515 -68.725 49.317
381775 99.571 -68.728 49.277
381776 99.627 -68.731 49.235
381777 99.680 -68.734 49.196
381778 99.714 -68.739 49.156
381779 99.745 -68.750 49.115
381780 99.770 -68.770 49.073
381781 99.819 -68.777 49.030
381782 99.872 -68.783 48.984
381783 99.925 -68.787 48.936
381784 99.979 -68.792 48.887
381785 100.033 -68.796 48.839
381786 100.086 -68.801 48.788
381787 100.108 -68.802 48.735
381788 100.098 -68.843 48.680
381789 100.136 -68.758 48.624
381790 100.177 -68.706 48.567
381791 100.207 -68.703 48.515
381792 100.239 -68.732 48.465
381793 100.271 -68.771 48.416
381794 100.303 -68.810 48.369
381795 100.335 -68.849 48.323
381796 100.367 -68.887 48.277
381797 100.399 -68.927 48.234
381798 100.431 -68.965 48.192
381799 100.461 -68.970 48.150
381800 100.489 -68.974 48.109
381801 100.519 -68.979 48.069
382241 90.492 -68.182 54.777
382242 90.523 -68.193 54.774
382243 90.554 -68.203 54.772
382244 90.585 -68.214 54.773
382245 90.616 -68.225 54.776
382246 90.648 -68.235 54.779
382247 90.679 -68.246 54.785
382248 90.710 -68.256 54.791
382249 90.766 -68.285 54.798
382250 90.825 -68.314 54.806
382251 90.883 -68.343 54.813
382252 90.942 -68.372 54.818
382253 91.000 -68.401 54.823
382254 91.065 -68.432 54.824
382255 91.116 -68.421 54.822
382256 91.168 -68.410 54.816
382257 91.219 -68.400 54.805
382258 91.271 -68.389 54.791
382259 91.322 -68.379 54.777
382260 91.373 -68.368 54.760
382261 91.425 -68.357 54.742
382262 91.476 -68.346 54.725
382263 91.526 -68.344 54.706
382264 91.577 -68.342 54.689
382265 91.627 -68.341 54.673
382266 91.677 -68.338 54.657
382267 91.722 -68.338 54.643
382268 91.765 -68.339 54.629
382269 91.807 -68.340 54.618
382270 91.849 -68.341 54.608
382271 91.891 -68.341 54.600
382272 91.933 -68.342 54.594
382273 91.976 -68.343 54.589
382274 92.018 -68.344 54.584
382275 92.060 -68.344 54.580
382276 92.102 -68.345 54.575
382277 92.146 -68.343 54.570
382278 92.191 -68.337 54.565
382279 92.235 -68.332 54.559
382280 92.280 -68.325 54.551
382281 92.324 -68.320 54.543
382282 92.369 -68.314 54.534
382283 92.418 -68.311 54.522
382284 92.467 -68.309 54.510
382285 92.516 -68.306 54.495
382286 92.566 -68.303 54.479
382287 92.615 -68.300 54.460
382288 92.664 -68.298 54.440
382289 92.714 -68.296 54.419
382290 92.763 -68.293 54.398
382291 92.812 -68.291 54.375
382292 92.861 -68.288 54.351
382293 92.911 -68.285 54.325
382294 92.960 -68.283 54.299
382295 93.010 -68.280 54.273
382296 93.060 -68.277 54.247
382297 93.110 -68.275 54.222
382298 93.160 -68.272 54.198
382299 93.210 -68.269 54.174
382300 93.260 -68.266 54.151
382301 93.310 -68.263 54.130
382302 93.360 -68.260 54.110
382303 93.410 -68.257 54.092
382304 93.459 -68.255 54.074
382305 93.508 -68.252 54.057
382306 93.557 -68.249 54.039
382307 93.606 -68.247 54.022
382308 93.654 -68.244 54.005
382309 93.703 -68.241 53.987
382310 93.752 -68.238 53.970
382311 93.801 -68.235 53.953
382312 93.850 -68.232 53.931
382313 93.899 -68.230 53.914
382314 93.948 -68.227 53.896
382315 93.997 -68.225 53.878
382316 94.046 -68.222 53.859
382317 94.094 -68.218 53.840
382318 94.140 -68.208 53.820
382319 94.185 -68.199 53.799
382320 94.231 -68.189 53.775
382321 94.276 -68.180 53.751
382322 94.322 -68.170 53.727
382323 94.367 -68.160 53.703
382324 94.413 -68.151 53.678
382325 94.459 -68.141 53.655
382326 94.504 -68.132 53.632
382327 94.550 -68.122 53.610
382328 94.596 -68.112 53.589
382329 94.641 -68.103 53.569
382330 94.668 -68.118 53.551
382331 94.698 -68.110 53.534
382332 94.729 -68.100 53.518
382333 94.759 -68.088 53.503
382334 94.789 -68.078 53.489
382335 94.820 -68.066 53.474
382336 94.850 -68.056 53.460
382337 94.879 -68.043 53.446
382338 94.904 -68.031 53.432
382339 94.929 -68.018 53.415
382340 94.954 -68.005 53.398
382341 94.979 -67.991 53.380
382342 95.003 -67.978 53.361
382343 95.028 -67.965 53.342
382344 95.053 -67.952 53.323
382345 95.078 -67.939 53.304
382346 95.103 -67.926 53.286
382347 95.128 -67.913 53.266
382348 95.153 -67.899 53.247
382349 95.178 -67.887 53.228
382350 95.214 -67.893 53.210
382351 95.252 -67.902 53.193
382352 95.290 -67.909 53.176
382353 95.328 -67.918 53.161
382354 95.366 -67.926 53.145
382355 95.402 -67.937 53.130
382356 95.437 -67.951 53.116
382357 95.478 -67.955 53.103
382358 95.520 -67.957 53.090
382359 95.562 -67.961 53.077
382360 95.603 -67.968 53.063
382361 95.644 -67.974 53.049
382362 95.685 -67.981 53.036
382363 95.726 -67.988 53.022
382364 95.766 -67.996 53.009
382365 95.807 -68.003 52.997
382366 95.848 -68.011 52.983
382367 95.889 -68.018 52.971
382368 95.926 -68.025 52.959
382369 95.956 -68.028 52.947
382370 95.989 -68.031 52.935
382371 96.027 -68.037 52.923
382372 96.065 -68.042 52.911
382373 96.103 -68.046 52.898
382374 96.141 -68.052 52.883
382375 96.178 -68.057 52.869
382376 96.216 -68.062 52.853
382377 96.254 -68.068 52.835
382378 96.292 -68.073 52.815
382379 96.330 -68.078 52.793
382380 96.368 -68.084 52.769
382381 96.405 -68.089 52.745
382382 96.443 -68.094 52.721
382383 96.479 -68.099 52.695
382384 96.515 -68.104 52.671
382385 96.550 -68.109 52.648
382386 96.579 -68.112 52.626
382387 96.602 -68.113 52.604
382388 96.632 -68.118 52.580
382389 96.663 -68.122 52.555
382390 96.694 -68.128 52.529
382391 96.725 -68.132 52.501
382392 96.755 -68.137 52.473
382393 96.786 -68.141 52.449
382394 96.817 -68.146 52.424
382395 96.848 -68.150 52.399
382396 96.879 -68.156 52.372
382397 96.909 -68.160 52.346
382398 96.938 -68.166 52.323
382399 96.966 -68.171 52.304
382400 96.994 -68.176 52.287
382401 97.022 -68.181 52.271
382402 97.050 -68.186 52.259
382403 97.082 -68.194 52.251
382404 97.115 -68.197 52.245
382405 97.147 -68.192 52.241
382406 97.181 -68.191 52.241
382407 97.217 -68.194 52.241
382408 97.250 -68.203 52.238
382409 97.283 -68.213 52.231
382410 97.316 -68.222 52.220
382411 97.350 -68.231 52.202
382412 97.383 -68.241 52.177
382413 97.416 -68.250 52.147
382414 97.449 -68.260 52.113
382415 97.482 -68.269 52.073
382416 97.509 -68.266 52.033
382417 97.540 -68.259 51.994
382418 97.571 -68.250 51.958
382419 97.601 -68.241 51.926
382420 97.633 -68.231 51.895
382421 97.665 -68.229 51.863
382422 97.700 -68.240 51.830
382423 97.734 -68.243 51.794
382424 97.768 -68.246 51.759
382425 97.802 -68.249 51.722
382426 97.836 -68.252 51.685
382427 97.870 -68.255 51.644
382428 97.903 -68.260 51.596
382429 97.938 -68.272 51.548
382430 97.973 -68.284 51.499
382431 98.007 -68.296 51.450
382432 98.042 -68.308 51.399
382433 98.077 -68.320 51.347
382434 98.108 -68.317 51.293
382435 98.140 -68.312 51.237
382436 98.171 -68.306 51.180
382437 98.204 -68.316 51.123
382438 98.238 -68.329 51.067
382439 98.271 -68.341 51.014
382440 98.304 -68.354 50.960
382441 98.338 -68.367 50.907
382442 98.371 -68.380 50.855
382443 98.410 -68.391 50.804
382444 98.456 -68.401 50.755
382445 98.503 -68.410 50.707
382446 98.538 -68.422 50.659
382447 98.574 -68.435 50.611
382448 98.609 -68.448 50.563
382449 98.645 -68.460 50.515
382450 98.680 -68.473 50.466
382451 98.718 -68.479 50.419
382452 98.758 -68.482 50.373
382453 98.798 -68.485 50.327
382454 98.828 -68.471 50.280
382455 98.858 -68.457 50.235
382456 98.896 -68.452 50.186
382457 98.935 -68.449 50.137
382458 98.974 -68.446 50.088
382459 99.013 -68.444 50.039
382460 99.053 -68.441 49.990
382461 99.101 -68.486 49.941
382462 99.144 -68.525 49.893
382463 99.174 -68.543 49.845
382464 99.204 -68.560 49.798
382465 99.234 -68.578 49.752
382466 99.264 -68.595 49.705
382467 99.302 -68.607 49.660
382468 99.342 -68.617 49.615
382469 99.382 -68.626 49.571
382470 99.422 -68.636 49.525
382471 99.438 -68.648 49.481
382472 99.451 -68.660 49.438
382473 99.464 -68.672 49.394
382474 99.482 -68.683 49.350
382475 99.539 -68.686 49.309
382476 99.594 -68.689 49.268
382477 99.651 -68.692 49.228
382478 99.691 -68.695 49.188
382479 99.725 -68.700 49.149
382480 99.754 -68.715 49.109
382481 99.779 -68.736 49.067
382482 99.824 -68.745 49.023
382483 99.877 -68.749 48.978
382484 99.931 -68.754 48.930
382485 99.984 -68.759 48.882
382486 100.038 -68.764 48.833
382487 100.085 -68.747 48.782
382488 100.098 -68.734 48.730
382489 100.092 -68.766 48.676
382490 100.147 -68.651 48.620
382491 100.177 -68.648 48.564
382492 100.209 -68.677 48.512
382493 100.241 -68.717 48.463
382494 100.273 -68.755 48.415
382495 100.305 -68.794 48.367
382496 100.337 -68.833 48.322
382497 100.369 -68.872 48.276
382498 100.402 -68.911 48.233
382499 100.431 -68.925 48.190
382500 100.461 -68.930 48.149
382501 100.490 -68.934 48.108
382941 90.449 -68.140 54.785
382942 90.480 -68.150 54.781
382943 90.511 -68.161 54.779
382944 90.542 -68.172 54.778
382945 90.573 -68.182 54.780
382946 90.604 -68.192 54.784
382947 90.636 -68.203 54.788
382948 90.667 -68.213 54.794
382949 90.702 -68.227 54.801
382950 90.760 -68.256 54.808
382951 90.819 -68.285 54.815
382952 90.877 -68.315 54.821
382953 90.936 -68.344 54.826
382954 91.002 -68.385 54.830
382955 91.057 -68.383 54.829
382956 91.108 -68.372 54.826
382957 91.160 -68.362 54.817
382958 91.211 -68.351 54.805
382959 91.262 -68.341 54.790
382960 91.313 -68.330 54.775
382961 91.365 -68.319 54.758
382962 91.416 -68.308 54.740
382963 91.467 -68.297 54.722
382964 91.519 -68.287 54.704
382965 91.569 -68.285 54.687
382966 91.619 -68.283 54.672
382967 91.669 -68.282 54.657
382968 91.718 -68.280 54.643
382969 91.760 -68.281 54.631
382970 91.802 -68.282 54.620
382971 91.844 -68.282 54.611
382972 91.886 -68.283 54.604
382973 91.929 -68.285 54.598
382974 91.971 -68.285 54.593
382975 92.013 -68.286 54.588
382976 92.056 -68.287 54.583
382977 92.097 -68.288 54.579
382978 92.140 -68.288 54.575
382979 92.183 -68.285 54.571
382980 92.228 -68.280 54.565
382981 92.273 -68.274 54.556
382982 92.319 -68.269 54.546
382983 92.369 -68.267 54.535
382984 92.418 -68.265 54.524
382985 92.467 -68.262 54.510
382986 92.516 -68.260 54.495
382987 92.566 -68.256 54.477
382988 92.615 -68.254 54.458
382989 92.664 -68.251 54.438
382990 92.714 -68.249 54.417
382991 92.763 -68.247 54.396
382992 92.812 -68.244 54.374
382993 92.861 -68.241 54.350
382994 92.911 -68.238 54.324
382995 92.960 -68.236 54.298
382996 93.010 -68.233 54.273
382997 93.060 -68.231 54.248
382998 93.110 -68.228 54.224
382999 93.160 -68.225 54.201
383000 93.210 -68.222 54.178
383001 93.260 -68.219 54.157
383002 93.310 -68.216 54.136
383003 93.360 -68.213 54.117
383004 93.410 -68.210 54.099
383005 93.460 -68.208 54.083
383006 93.509 -68.205 54.067
383007 93.558 -68.203 54.050
383008 93.607 -68.200 54.035
383009 93.655 -68.197 54.019
383010 93.704 -68.194 54.002
383011 93.753 -68.191 53.983
383012 93.802 -68.188 53.965
383013 93.851 -68.186 53.946
383014 93.900 -68.183 53.927
383015 93.949 -68.180 53.911
383016 93.998 -68.178 53.895
383017 94.047 -68.175 53.877
383018 94.096 -68.172 53.859
383019 94.145 -68.169 53.839
383020 94.194 -68.166 53.818
383021 94.242 -68.162 53.794
383022 94.287 -68.152 53.769
383023 94.333 -68.143 53.744
383024 94.379 -68.133 53.719
383025 94.424 -68.124 53.691
383026 94.470 -68.114 53.666
383027 94.515 -68.104 53.641
383028 94.561 -68.095 53.618
383029 94.606 -68.085 53.596
383030 94.645 -68.084 53.575
383031 94.673 -68.085 53.556
383032 94.703 -68.075 53.538
383033 94.734 -68.063 53.521
383034 94.764 -68.053 53.505
383035 94.794 -68.041 53.490
383036 94.820 -68.029 53.476
383037 94.845 -68.015 53.461
383038 94.870 -68.003 53.446
383039 94.895 -67.990 53.432
383040 94.920 -67.976 53.415
383041 94.945 -67.963 53.398
383042 94.970 -67.950 53.380
383043 94.994 -67.937 53.361
383044 95.020 -67.924 53.340
383045 95.044 -67.911 53.321
383046 95.069 -67.898 53.302
383047 95.094 -67.884 53.282
383048 95.119 -67.871 53.263
383049 95.144 -67.858 53.243
383050 95.176 -67.858 53.224
383051 95.214 -67.866 53.205
383052 95.252 -67.874 53.187
383053 95.289 -67.882 53.170
383054 95.327 -67.890 53.154
383055 95.365 -67.898 53.137
383056 95.401 -67.909 53.122
383057 95.438 -67.921 53.107
383058 95.480 -67.923 53.092
383059 95.522 -67.924 53.078
383060 95.564 -67.927 53.063
383061 95.605 -67.934 53.048
383062 95.647 -67.941 53.033
383063 95.687 -67.948 53.018
383064 95.729 -67.954 53.004
383065 95.769 -67.962 52.990
383066 95.810 -67.969 52.976
383067 95.851 -67.977 52.962
383068 95.891 -67.984 52.949
383069 95.928 -67.990 52.936
383070 95.958 -67.993 52.923
383071 95.990 -67.996 52.911
383072 96.028 -68.002 52.898
383073 96.066 -68.006 52.885
383074 96.104 -68.012 52.872
383075 96.142 -68.017 52.857
383076 96.179 -68.022 52.842
383077 96.217 -68.028 52.826
383078 96.255 -68.033 52.809
383079 96.293 -68.038 52.790
383080 96.331 -68.043 52.770
383081 96.369 -68.049 52.748
383082 96.406 -68.054 52.725
383083 96.444 -68.059 52.702
383084 96.482 -68.065 52.678
383085 96.519 -68.069 52.654
383086 96.554 -68.075 52.633
383087 96.583 -68.078 52.612
383088 96.611 -68.081 52.590
383089 96.642 -68.085 52.566
383090 96.672 -68.090 52.541
383091 96.703 -68.095 52.515
383092 96.734 -68.100 52.488
383093 96.764 -68.104 52.463
383094 96.794 -68.109 52.440
383095 96.822 -68.115 52.416
383096 96.850 -68.120 52.392
383097 96.878 -68.125 52.365
383098 96.906 -68.130 52.341
383099 96.935 -68.135 52.321
383100 96.963 -68.141 52.304
383101 96.991 -68.146 52.290
383102 97.019 -68.151 52.275
383103 97.051 -68.159 52.264
383104 97.084 -68.167 52.258
383105 97.117 -68.170 52.254
383106 97.149 -68.165 52.253
383107 97.181 -68.159 52.255
383108 97.215 -68.157 52.254
383109 97.248 -68.166 52.248
383110 97.281 -68.175 52.237
383111 97.314 -68.184 52.223
383112 97.347 -68.194 52.201
383113 97.380 -68.202 52.174
383114 97.413 -68.210 52.142
383115 97.446 -68.219 52.107
383116 97.479 -68.228 52.067
383117 97.516 -68.238 52.027
383118 97.547 -68.227 51.988
383119 97.578 -68.217 51.952
383120 97.609 -68.207 51.919
383121 97.640 -68.197 51.888
383122 97.673 -68.201 51.855
383123 97.709 -68.218 51.820
383124 97.743 -68.225 51.784
383125 97.777 -68.228 51.750
383126 97.811 -68.231 51.712
383127 97.845 -68.235 51.674
383128 97.879 -68.238 51.633
383129 97.914 -68.244 51.586
383130 97.948 -68.256 51.537
383131 97.983 -68.269 51.488
383132 98.017 -68.281 51.438
383133 98.052 -68.292 51.388
383134 98.083 -68.286 51.336
383135 98.114 -68.281 51.283
383136 98.146 -68.275 51.229
383137 98.177 -68.274 51.174
383138 98.211 -68.287 51.118
383139 98.244 -68.300 51.064
383140 98.278 -68.312 51.011
383141 98.311 -68.325 50.957
383142 98.345 -68.338 50.904
383143 98.378 -68.350 50.852
383144 98.411 -68.363 50.800
383145 98.445 -68.376 50.751
383146 98.492 -68.385 50.702
383147 98.535 -68.395 50.653
383148 98.571 -68.408 50.603
383149 98.606 -68.421 50.555
383150 98.641 -68.434 50.506
383151 98.678 -68.442 50.457
383152 98.718 -68.445 50.411
383153 98.758 -68.448 50.365
383154 98.797 -68.451 50.319
383155 98.839 -68.451 50.272
383156 98.880 -68.450 50.225
383157 98.920 -68.447 50.177
383158 98.959 -68.444 50.127
383159 98.998 -68.441 50.078
383160 99.038 -68.438 50.029
383161 99.077 -68.436 49.979
383162 99.119 -68.470 49.930
383163 99.149 -68.488 49.883
383164 99.179 -68.506 49.837
383165 99.209 -68.523 49.789
383166 99.239 -68.541 49.742
383167 99.269 -68.559 49.695
383168 99.308 -68.569 49.650
383169 99.348 -68.579 49.605
383170 99.388 -68.589 49.560
383171 99.428 -68.598 49.515
383172 99.464 -68.609 49.472
383173 99.477 -68.620 49.430
383174 99.489 -68.633 49.386
383175 99.506 -68.644 49.342
383176 99.562 -68.647 49.300
383177 99.618 -68.650 49.259
383178 99.668 -68.653 49.219
383179 99.702 -68.658 49.180
383180 99.737 -68.661 49.141
383181 99.763 -68.680 49.101
383182 99.788 -68.700 49.060
383183 99.829 -68.712 49.016
383184 99.883 -68.717 48.970
383185 99.936 -68.721 48.924
383186 99.990 -68.726 48.876
383187 100.043 -68.730 48.826
383188 100.077 -68.672 48.776
383189 100.087 -68.667 48.723
383190 100.114 -68.609 48.670
383191 100.147 -68.592 48.616
383192 100.179 -68.623 48.560
383193 100.211 -68.662 48.509
383194 100.243 -68.701 48.460
383195 100.275 -68.739 48.412
383196 100.308 -68.778 48.365
383197 100.340 -68.817 48.321
383198 100.372 -68.856 48.275
383199 100.403 -68.880 48.232
383200 100.431 -68.884 48.190
383201 100.461 -68.889 48.148
383642 90.437 -68.107 54.789
383643 90.468 -68.118 54.786
383644 90.499 -68.128 54.785
383645 90.530 -68.139 54.786
383646 90.561 -68.150 54.789
383647 90.592 -68.160 54.793
383648 90.623 -68.171 54.798
383649 90.655 -68.181 54.805
383650 90.696 -68.199 54.811
383651 90.754 -68.228 54.818
383652 90.813 -68.257 54.825
383653 90.871 -68.287 54.831
383654 90.931 -68.318 54.835
383655 90.997 -68.345 54.838
383656 91.048 -68.335 54.836
383657 91.100 -68.324 54.830
383658 91.151 -68.313 54.820
383659 91.203 -68.303 54.805
383660 91.254 -68.292 54.790
383661 91.305 -68.282 54.773
383662 91.357 -68.270 54.756
383663 91.408 -68.260 54.738
383664 91.459 -68.249 54.720
383665 91.511 -68.238 54.703
383666 91.562 -68.228 54.687
383667 91.612 -68.225 54.672
383668 91.662 -68.224 54.659
383669 91.713 -68.222 54.646
383670 91.755 -68.223 54.634
383671 91.797 -68.224 54.624
383672 91.840 -68.225 54.616
383673 91.882 -68.225 54.610
383674 91.924 -68.226 54.605
383675 91.966 -68.227 54.601
383676 92.009 -68.228 54.596
383677 92.051 -68.228 54.593
383678 92.093 -68.229 54.589
383679 92.135 -68.230 54.585
383680 92.177 -68.231 54.580
383681 92.221 -68.228 54.574
383682 92.270 -68.225 54.564
383683 92.319 -68.223 54.551
383684 92.368 -68.220 54.538
383685 92.418 -68.218 54.524
383686 92.467 -68.216 54.510
383687 92.516 -68.213 54.494
383688 92.566 -68.210 54.475
383689 92.615 -68.207 54.456
383690 92.664 -68.205 54.437
383691 92.714 -68.202 54.416
383692 92.763 -68.200 54.394
383693 92.812 -68.197 54.372
383694 92.861 -68.194 54.349
383695 92.911 -68.192 54.323
383696 92.960 -68.189 54.298
383697 93.009 -68.187 54.274
383698 93.060 -68.184 54.250
383699 93.110 -68.181 54.226
383700 93.160 -68.178 54.205
383701 93.210 -68.175 54.183
383702 93.260 -68.172 54.162
383703 93.310 -68.169 54.142
383704 93.360 -68.167 54.123
383705 93.410 -68.164 54.107
383706 93.461 -68.161 54.092
383707 93.510 -68.158 54.076
383708 93.559 -68.156 54.061
383709 93.608 -68.153 54.047
383710 93.656 -68.150 54.032
383711 93.706 -68.147 54.015
383712 93.755 -68.144 53.996
383713 93.803 -68.142 53.978
383714 93.852 -68.139 53.958
383715 93.901 -68.136 53.939
383716 93.950 -68.134 53.922
383717 93.999 -68.131 53.908
383718 94.048 -68.128 53.892
383719 94.097 -68.125 53.875
383720 94.146 -68.122 53.856
383721 94.195 -68.120 53.835
383722 94.244 -68.117 53.812
383723 94.292 -68.114 53.787
383724 94.341 -68.112 53.762
383725 94.389 -68.106 53.736
383726 94.434 -68.096 53.708
383727 94.480 -68.087 53.679
383728 94.526 -68.077 53.651
383729 94.571 -68.068 53.626
383730 94.617 -68.058 53.603
383731 94.647 -68.060 53.581
383732 94.678 -68.050 53.561
383733 94.708 -68.038 53.542
383734 94.737 -68.027 53.524
383735 94.762 -68.014 53.508
383736 94.786 -68.000 53.492
383737 94.812 -67.987 53.477
383738 94.836 -67.974 53.462
383739 94.861 -67.962 53.446
383740 94.886 -67.948 53.431
383741 94.911 -67.935 53.415
383742 94.936 -67.922 53.398
383743 94.961 -67.909 53.379
383744 94.986 -67.896 53.360
383745 95.011 -67.883 53.339
383746 95.035 -67.870 53.320
383747 95.060 -67.856 53.300
383748 95.085 -67.843 53.279
383749 95.110 -67.830 53.259
383750 95.138 -67.822 53.240
383751 95.176 -67.830 53.220
383752 95.213 -67.838 53.201
383753 95.251 -67.846 53.182
383754 95.289 -67.855 53.165
383755 95.327 -67.862 53.147
383756 95.365 -67.871 53.130
383757 95.401 -67.881 53.114
383758 95.440 -67.889 53.098
383759 95.482 -67.890 53.083
383760 95.525 -67.893 53.067
383761 95.566 -67.895 53.051
383762 95.608 -67.901 53.034
383763 95.649 -67.908 53.018
383764 95.690 -67.915 53.002
383765 95.731 -67.921 52.987
383766 95.772 -67.928 52.972
383767 95.813 -67.935 52.956
383768 95.853 -67.943 52.943
383769 95.894 -67.950 52.929
383770 95.930 -67.956 52.915
383771 95.960 -67.959 52.901
383772 95.991 -67.962 52.888
383773 96.029 -67.966 52.874
383774 96.067 -67.971 52.861
383775 96.105 -67.977 52.847
383776 96.143 -67.982 52.833
383777 96.180 -67.987 52.817
383778 96.218 -67.993 52.801
383779 96.256 -67.998 52.785
383780 96.294 -68.003 52.767
383781 96.331 -68.009 52.748
383782 96.370 -68.014 52.727
383783 96.407 -68.019 52.706
383784 96.445 -68.025 52.684
383785 96.483 -68.029 52.660
383786 96.521 -68.034 52.637
383787 96.558 -68.040 52.617
383788 96.589 -68.043 52.596
383789 96.619 -68.048 52.576
383790 96.650 -68.053 52.552
383791 96.678 -68.059 52.527
383792 96.707 -68.064 52.502
383793 96.735 -68.068 52.476
383794 96.763 -68.074 52.453
383795 96.791 -68.079 52.432
383796 96.819 -68.084 52.409
383797 96.848 -68.090 52.386
383798 96.876 -68.095 52.361
383799 96.904 -68.100 52.337
383800 96.932 -68.106 52.319
383801 96.960 -68.111 52.304
383802 96.988 -68.116 52.292
383803 97.020 -68.124 52.280
383804 97.053 -68.132 52.272
383805 97.086 -68.141 52.266
383806 97.119 -68.143 52.264
383807 97.151 -68.137 52.265
383808 97.183 -68.132 52.265
383809 97.218 -68.128 52.262
383810 97.249 -68.131 52.254
383811 97.280 -68.134 52.240
383812 97.311 -68.138 52.223
383813 97.344 -68.147 52.198
383814 97.377 -68.156 52.169
383815 97.410 -68.164 52.136
383816 97.446 -68.175 52.099
383817 97.484 -68.186 52.057
383818 97.521 -68.198 52.017
383819 97.554 -68.194 51.978
383820 97.585 -68.184 51.942
383821 97.616 -68.174 51.909
383822 97.647 -68.164 51.878
383823 97.681 -68.173 51.844
383824 97.717 -68.190 51.809
383825 97.753 -68.206 51.775
383826 97.787 -68.210 51.739
383827 97.821 -68.214 51.700
383828 97.855 -68.217 51.661
383829 97.889 -68.220 51.619
383830 97.923 -68.230 51.570
383831 97.958 -68.242 51.522
383832 97.993 -68.254 51.473
383833 98.026 -68.261 51.423
383834 98.058 -68.256 51.373
383835 98.089 -68.250 51.325
383836 98.120 -68.244 51.273
383837 98.152 -68.238 51.220
383838 98.184 -68.244 51.164
383839 98.218 -68.257 51.108
383840 98.251 -68.270 51.054
383841 98.285 -68.283 51.002
383842 98.318 -68.296 50.948
383843 98.351 -68.309 50.895
383844 98.385 -68.321 50.844
383845 98.418 -68.334 50.792
383846 98.451 -68.347 50.742
383847 98.485 -68.360 50.693
383848 98.527 -68.370 50.644
383849 98.567 -68.381 50.594
383850 98.602 -68.394 50.544
383851 98.639 -68.405 50.495
383852 98.678 -68.408 50.447
383853 98.717 -68.411 50.401
383854 98.755 -68.415 50.355
383855 98.791 -68.419 50.310
383856 98.830 -68.422 50.264
383857 98.872 -68.421 50.217
383858 98.916 -68.421 50.169
383859 98.958 -68.421 50.120
383860 99.001 -68.421 50.071
383861 99.044 -68.421 50.022
383862 99.088 -68.422 49.973
383863 99.124 -68.433 49.924
383864 99.154 -68.451 49.875
383865 99.184 -68.469 49.827
383866 99.214 -68.486 49.781
383867 99.244 -68.504 49.735
383868 99.274 -68.522 49.690
383869 99.314 -68.532 49.645
383870 99.354 -68.541 49.599
383871 99.394 -68.551 49.555
383872 99.434 -68.560 49.510
383873 99.474 -68.570 49.465
383874 99.502 -68.582 49.421
383875 99.514 -68.593 49.375
383876 99.529 -68.604 49.331
383877 99.586 -68.607 49.288
383878 99.642 -68.611 49.248
383879 99.679 -68.615 49.208
383880 99.713 -68.619 49.169
383881 99.747 -68.626 49.131
383882 99.772 -68.645 49.091
383883 99.797 -68.665 49.049
383884 99.834 -68.680 49.006
383885 99.888 -68.684 48.962
383886 99.941 -68.689 48.916
383887 99.995 -68.693 48.868
383888 100.037 -68.661 48.819
383889 100.070 -68.598 48.769
383890 100.077 -68.601 48.717
383891 100.118 -68.537 48.664
383892 100.149 -68.568 48.610
383893 100.181 -68.607 48.555
383894 100.213 -68.646 48.505
383895 100.246 -68.685 48.457
383896 100.278 -68.723 48.409
383897 100.310 -68.762 48.363
383898 100.342 -68.802 48.318
383899 100.373 -68.835 48.273
383900 100.403 -68.839 48.229
383901 100.432 -68.844 48.187
383902 100.461 -68.849 48.145
384342 90.393 -68.065 54.798
384343 90.425 -68.075 54.794
384344 90.456 -68.086 54.792
384345 90.487 -68.097 54.792
384346 90.518 -68.106 54.795
384347 90.549 -68.117 54.799
384348 90.580 -68.128 54.804
384349 90.611 -68.138 54.810
384350 90.643 -68.149 54.816
384351 90.690 -68.171 54.824
384352 90.748 -68.200 54.830
384353 90.807 -68.229 54.837
384354 90.865 -68.259 54.843
384355 90.937 -68.307 54.848
384356 90.989 -68.297 54.848
384357 91.040 -68.286 54.844
384358 91.091 -68.275 54.835
384359 91.143 -68.265 54.822
384360 91.194 -68.254 54.807
384361 91.245 -68.244 54.791
384362 91.297 -68.232 54.774
384363 91.348 -68.222 54.756
384364 91.399 -68.211 54.738
384365 91.451 -68.200 54.720
384366 91.502 -68.190 54.704
384367 91.554 -68.179 54.688
384368 91.605 -68.169 54.674
384369 91.655 -68.167 54.661
384370 91.705 -68.165 54.650
384371 91.750 -68.165 54.640
384372 91.793 -68.166 54.630
384373 91.835 -68.166 54.623
384374 91.877 -68.167 54.618
384375 91.920 -68.168 54.615
384376 91.962 -68.169 54.611
384377 92.004 -68.169 54.607
384378 92.046 -68.170 54.605
384379 92.088 -68.171 54.602
384380 92.131 -68.172 54.597
384381 92.171 -68.181 54.590
384382 92.212 -68.191 54.583
384383 92.258 -68.193 54.570
384384 92.307 -68.191 54.555
384385 92.356 -68.188 54.540
384386 92.406 -68.184 54.525
384387 92.455 -68.182 54.509
384388 92.504 -68.179 54.492
384389 92.554 -68.177 54.473
384390 92.603 -68.174 54.454
384391 92.653 -68.172 54.434
384392 92.702 -68.169 54.414
384393 92.752 -68.166 54.392
384394 92.801 -68.163 54.370
384395 92.850 -68.160 54.348
384396 92.901 -68.156 54.323
384397 92.959 -68.146 54.299
384398 93.006 -68.141 54.276
384399 93.059 -68.137 54.252
384400 93.110 -68.134 54.230
384401 93.160 -68.131 54.207
384402 93.210 -68.128 54.187
384403 93.260 -68.126 54.167
384404 93.310 -68.123 54.148
384405 93.360 -68.120 54.130
384406 93.410 -68.117 54.115
384407 93.461 -68.115 54.101
384408 93.510 -68.112 54.086
384409 93.560 -68.109 54.072
384410 93.609 -68.106 54.059
384411 93.658 -68.103 54.043
384412 93.707 -68.100 54.026
384413 93.756 -68.098 54.008
384414 93.804 -68.095 53.990
384415 93.853 -68.092 53.969
384416 93.902 -68.090 53.949
384417 93.951 -68.087 53.933
384418 94.000 -68.084 53.918
384419 94.049 -68.081 53.905
384420 94.098 -68.078 53.888
384421 94.147 -68.076 53.870
384422 94.196 -68.073 53.851
384423 94.245 -68.070 53.828
384424 94.294 -68.068 53.803
384425 94.343 -68.065 53.777
384426 94.392 -68.062 53.751
384427 94.440 -68.059 53.722
384428 94.489 -68.056 53.691
384429 94.536 -68.050 53.662
384430 94.582 -68.040 53.634
384431 94.622 -68.036 53.609
384432 94.652 -68.025 53.587
384433 94.678 -68.012 53.565
384434 94.703 -67.999 53.545
384435 94.728 -67.986 53.527
384436 94.753 -67.972 53.510
384437 94.778 -67.959 53.494
384438 94.803 -67.946 53.478
384439 94.827 -67.933 53.462
384440 94.852 -67.920 53.447
384441 94.877 -67.907 53.432
384442 94.902 -67.894 53.415
384443 94.927 -67.880 53.398
384444 94.952 -67.868 53.379
384445 94.977 -67.855 53.360
384446 95.001 -67.841 53.339
384447 95.027 -67.828 53.318
384448 95.051 -67.815 53.298
384449 95.076 -67.802 53.277
384450 95.101 -67.789 53.258
384451 95.137 -67.795 53.238
384452 95.175 -67.802 53.217
384453 95.213 -67.811 53.197
384454 95.251 -67.819 53.178
384455 95.288 -67.827 53.161
384456 95.326 -67.835 53.142
384457 95.364 -67.843 53.124
384458 95.401 -67.853 53.107
384459 95.442 -67.856 53.090
384460 95.484 -67.858 53.075
384461 95.527 -67.861 53.058
384462 95.568 -67.863 53.040
384463 95.610 -67.868 53.022
384464 95.651 -67.874 53.004
384465 95.692 -67.881 52.987
384466 95.733 -67.888 52.970
384467 95.774 -67.895 52.954
384468 95.815 -67.902 52.938
384469 95.856 -67.909 52.924
384470 95.897 -67.916 52.909
384471 95.932 -67.921 52.895
384472 95.962 -67.924 52.881
384473 95.992 -67.927 52.866
384474 96.030 -67.931 52.852
384475 96.067 -67.937 52.838
384476 96.106 -67.942 52.824
384477 96.144 -67.947 52.810
384478 96.181 -67.952 52.794
384479 96.219 -67.958 52.778
384480 96.257 -67.963 52.762
384481 96.295 -67.968 52.745
384482 96.332 -67.974 52.726
384483 96.370 -67.979 52.706
384484 96.408 -67.984 52.686
384485 96.446 -67.990 52.665
384486 96.484 -67.994 52.643
384487 96.521 -67.999 52.620
384488 96.556 -68.006 52.600
384489 96.591 -68.012 52.580
384490 96.619 -68.018 52.560
384491 96.647 -68.023 52.538
384492 96.675 -68.028 52.513
384493 96.703 -68.034 52.489
384494 96.732 -68.039 52.465
384495 96.760 -68.044 52.443
384496 96.788 -68.050 52.423
384497 96.816 -68.055 52.402
384498 96.844 -68.060 52.380
384499 96.873 -68.065 52.356
384500 96.901 -68.071 52.333
384501 96.929 -68.076 52.317
384502 96.957 -68.081 52.304
384503 96.989 -68.088 52.295
384504 97.022 -68.097 52.286
384505 97.055 -68.106 52.280
384506 97.088 -68.114 52.276
384507 97.121 -68.115 52.274
384508 97.153 -68.109 52.275
384509 97.185 -68.104 52.273
384510 97.225 -68.103 52.267
384511 97.256 -68.106 52.256
384512 97.287 -68.109 52.239
384513 97.317 -68.112 52.219
384514 97.348 -68.115 52.192
384515 97.378 -68.119 52.162
384516 97.413 -68.122 52.127
384517 97.451 -68.133 52.090
384518 97.488 -68.145 52.047
384519 97.526 -68.156 52.007
384520 97.562 -68.161 51.969
384521 97.592 -68.151 51.935
384522 97.623 -68.141 51.903
384523 97.654 -68.131 51.871
384524 97.689 -68.145 51.836
384525 97.725 -68.162 51.801
384526 97.761 -68.178 51.764
384527 97.797 -68.193 51.725
384528 97.830 -68.196 51.686
384529 97.865 -68.199 51.645
384530 97.898 -68.203 51.602
384531 97.933 -68.215 51.555
384532 97.967 -68.227 51.506
384533 98.001 -68.230 51.457
384534 98.032 -68.225 51.407
384535 98.063 -68.219 51.359
384536 98.095 -68.213 51.312
384537 98.126 -68.207 51.261
384538 98.158 -68.203 51.208
384539 98.191 -68.216 51.154
384540 98.224 -68.228 51.098
384541 98.258 -68.241 51.044
384542 98.291 -68.253 50.991
384543 98.324 -68.266 50.938
384544 98.358 -68.279 50.884
384545 98.391 -68.292 50.834
384546 98.425 -68.305 50.782
384547 98.458 -68.318 50.731
384548 98.492 -68.330 50.682
384549 98.525 -68.343 50.632
384550 98.562 -68.355 50.584
384551 98.599 -68.367 50.534
384552 98.638 -68.371 50.485
384553 98.675 -68.375 50.437
384554 98.712 -68.379 50.390
384555 98.749 -68.383 50.344
384556 98.786 -68.387 50.299
384557 98.822 -68.391 50.254
384558 98.863 -68.392 50.208
384559 98.906 -68.391 50.160
384560 98.949 -68.391 50.111
384561 98.992 -68.391 50.063
384562 99.035 -68.392 50.014
384563 99.081 -68.395 49.965
384564 99.127 -68.398 49.917
384565 99.159 -68.414 49.868
384566 99.189 -68.432 49.821
384567 99.219 -68.449 49.776
384568 99.249 -68.467 49.732
384569 99.280 -68.484 49.686
384570 99.320 -68.494 49.640
384571 99.360 -68.504 49.594
384572 99.400 -68.513 49.548
384573 99.440 -68.523 49.502
384574 99.480 -68.533 49.456
384575 99.520 -68.542 49.410
384576 99.540 -68.554 49.364
384577 99.553 -68.566 49.320
384578 99.610 -68.569 49.278
384579 99.656 -68.573 49.237
384580 99.690 -68.576 49.198
384581 99.724 -68.580 49.158
384582 99.756 -68.591 49.120
384583 99.781 -68.611 49.080
384584 99.807 -68.630 49.039
384585 99.840 -68.646 48.997
384586 99.893 -68.651 48.953
384587 99.947 -68.656 48.907
384588 99.997 -68.651 48.860
384589 100.030 -68.586 48.811
384590 100.062 -68.523 48.761
384591 100.087 -68.482 48.710
384592 100.120 -68.513 48.657
384593 100.151 -68.552 48.604
384594 100.184 -68.592 48.550
384595 100.216 -68.630 48.501
384596 100.248 -68.669 48.453
384597 100.280 -68.708 48.407
384598 100.312 -68.747 48.361
384599 100.344 -68.786 48.315
384600 100.373 -68.795 48.271
384601 100.403 -68.799 48.227
384602 100.432 -68.804 48.184
384603 100.469 -68.803 48.142
385042 90.350 -68.021 54.805
385043 90.381 -68.032 54.801
385044 90.412 -68.043 54.799
385045 90.444 -68.053 54.798
385046 90.475 -68.064 54.800
385047 90.506 -68.075 54.804
385048 90.537 -68.085 54.810
385049 90.569 -68.096 54.816
385050 90.599 -68.106 54.823
385051 90.630 -68.116 54.829
385052 90.684 -68.143 54.837
385053 90.742 -68.172 54.844
385054 90.801 -68.201 54.851
385055 90.866 -68.241 54.857
385056 90.929 -68.259 54.861
385057 90.980 -68.248 54.858
385058 91.032 -68.238 54.852
385059 91.083 -68.227 54.841
385060 91.135 -68.216 54.826
385061 91.186 -68.206 54.810
385062 91.237 -68.194 54.793
385063 91.289 -68.184 54.776
385064 91.340 -68.173 54.758
385065 91.391 -68.162 54.739
385066 91.442 -68.152 54.723
385067 91.494 -68.141 54.708
385068 91.545 -68.131 54.693
385069 91.596 -68.120 54.680
385070 91.648 -68.109 54.668
385071 91.698 -68.108 54.658
385072 91.746 -68.106 54.649
385073 91.788 -68.107 54.641
385074 91.830 -68.109 54.635
385075 91.873 -68.109 54.630
385076 91.915 -68.110 54.627
385077 91.957 -68.111 54.624
385078 91.999 -68.112 54.622
385079 92.041 -68.113 54.621
385080 92.082 -68.123 54.617
385081 92.122 -68.133 54.610
385082 92.162 -68.142 54.601
385083 92.203 -68.152 54.590
385084 92.243 -68.162 54.576
385085 92.290 -68.162 54.560
385086 92.340 -68.159 54.543
385087 92.389 -68.156 54.526
385088 92.439 -68.153 54.509
385089 92.488 -68.151 54.491
385090 92.537 -68.148 54.472
385091 92.587 -68.146 54.452
385092 92.636 -68.143 54.431
385093 92.686 -68.141 54.411
385094 92.735 -68.137 54.390
385095 92.792 -68.128 54.368
385096 92.850 -68.117 54.346
385097 92.909 -68.106 54.322
385098 92.967 -68.096 54.299
385099 93.009 -68.093 54.277
385100 93.053 -68.091 54.254
385101 93.105 -68.087 54.233
385102 93.157 -68.084 54.212
385103 93.209 -68.082 54.192
385104 93.260 -68.079 54.172
385105 93.310 -68.076 54.154
385106 93.360 -68.073 54.137
385107 93.410 -68.071 54.122
385108 93.460 -68.068 54.108
385109 93.510 -68.065 54.095
385110 93.560 -68.062 54.081
385111 93.610 -68.059 54.067
385112 93.659 -68.056 54.052
385113 93.708 -68.053 54.035
385114 93.757 -68.051 54.017
385115 93.805 -68.048 53.999
385116 93.854 -68.046 53.978
385117 93.904 -68.043 53.959
385118 93.953 -68.040 53.942
385119 94.001 -68.037 53.928
385120 94.050 -68.034 53.914
385121 94.099 -68.031 53.899
385122 94.148 -68.029 53.882
385123 94.197 -68.026 53.863
385124 94.246 -68.024 53.841
385125 94.295 -68.021 53.815
385126 94.344 -68.018 53.790
385127 94.393 -68.015 53.764
385128 94.441 -68.012 53.734
385129 94.490 -68.009 53.702
385130 94.539 -68.007 53.671
385131 94.584 -68.006 53.642
385132 94.619 -67.997 53.616
385133 94.644 -67.984 53.592
385134 94.669 -67.971 53.569
385135 94.694 -67.957 53.549
385136 94.719 -67.944 53.530
385137 94.744 -67.931 53.512
385138 94.769 -67.918 53.495
385139 94.793 -67.905 53.478
385140 94.819 -67.892 53.462
385141 94.843 -67.879 53.447
385142 94.868 -67.865 53.432
385143 94.893 -67.852 53.415
385144 94.918 -67.840 53.398
385145 94.943 -67.827 53.379
385146 94.968 -67.813 53.359
385147 94.993 -67.800 53.338
385148 95.018 -67.787 53.318
385149 95.042 -67.774 53.297
385150 95.067 -67.761 53.276
385151 95.099 -67.759 53.256
385152 95.137 -67.767 53.235
385153 95.175 -67.775 53.215
385154 95.212 -67.783 53.194
385155 95.250 -67.791 53.175
385156 95.288 -67.799 53.156
385157 95.326 -67.808 53.137
385158 95.364 -67.816 53.119
385159 95.402 -67.822 53.100
385160 95.445 -67.823 53.083
385161 95.487 -67.824 53.066
385162 95.530 -67.825 53.049
385163 95.572 -67.827 53.030
385164 95.613 -67.832 53.011
385165 95.654 -67.840 52.993
385166 95.695 -67.846 52.975
385167 95.736 -67.854 52.957
385168 95.776 -67.861 52.940
385169 95.818 -67.868 52.925
385170 95.859 -67.875 52.909
385171 95.900 -67.881 52.893
385172 95.934 -67.887 52.878
385173 95.964 -67.889 52.863
385174 95.994 -67.892 52.849
385175 96.031 -67.896 52.833
385176 96.068 -67.902 52.818
385177 96.106 -67.907 52.803
385178 96.144 -67.912 52.788
385179 96.182 -67.918 52.772
385180 96.220 -67.923 52.756
385181 96.258 -67.928 52.739
385182 96.296 -67.934 52.722
385183 96.333 -67.939 52.704
385184 96.371 -67.944 52.685
385185 96.408 -67.949 52.666
385186 96.443 -67.956 52.646
385187 96.478 -67.962 52.624
385188 96.514 -67.968 52.604
385189 96.549 -67.974 52.584
385190 96.584 -67.981 52.563
385191 96.616 -67.988 52.544
385192 96.644 -67.993 52.523
385193 96.672 -67.999 52.499
385194 96.700 -68.004 52.475
385195 96.729 -68.009 52.453
385196 96.757 -68.015 52.433
385197 96.785 -68.020 52.413
385198 96.813 -68.025 52.393
385199 96.841 -68.030 52.373
385200 96.870 -68.035 52.351
385201 96.898 -68.040 52.330
385202 96.926 -68.046 52.316
385203 96.957 -68.053 52.306
385204 96.991 -68.062 52.299
385205 97.024 -68.070 52.292
385206 97.057 -68.079 52.287
385207 97.090 -68.087 52.285
385208 97.123 -68.087 52.285
385209 97.155 -68.082 52.284
385210 97.187 -68.077 52.277
385211 97.232 -68.078 52.269
385212 97.262 -68.081 52.255
385213 97.293 -68.084 52.236
385214 97.324 -68.087 52.212
385215 97.355 -68.090 52.184
385216 97.387 -68.091 52.152
385217 97.421 -68.090 52.117
385218 97.455 -68.092 52.080
385219 97.493 -68.103 52.037
385220 97.531 -68.115 51.997
385221 97.569 -68.127 51.961
385222 97.599 -68.118 51.928
385223 97.631 -68.108 51.898
385224 97.662 -68.101 51.865
385225 97.698 -68.117 51.830
385226 97.733 -68.134 51.792
385227 97.769 -68.150 51.752
385228 97.805 -68.166 51.710
385229 97.840 -68.178 51.670
385230 97.874 -68.181 51.629
385231 97.908 -68.188 51.586
385232 97.943 -68.200 51.539
385233 97.975 -68.200 51.490
385234 98.006 -68.194 51.441
385235 98.038 -68.188 51.392
385236 98.069 -68.182 51.345
385237 98.100 -68.176 51.299
385238 98.132 -68.171 51.249
385239 98.164 -68.173 51.196
385240 98.198 -68.186 51.142
385241 98.231 -68.199 51.088
385242 98.265 -68.212 51.033
385243 98.298 -68.225 50.979
385244 98.332 -68.238 50.926
385245 98.365 -68.250 50.873
385246 98.398 -68.263 50.823
385247 98.432 -68.275 50.771
385248 98.465 -68.288 50.721
385249 98.498 -68.301 50.671
385250 98.531 -68.314 50.621
385251 98.565 -68.327 50.573
385252 98.601 -68.332 50.525
385253 98.633 -68.339 50.476
385254 98.670 -68.343 50.426
385255 98.707 -68.347 50.379
385256 98.743 -68.350 50.332
385257 98.780 -68.354 50.287
385258 98.817 -68.358 50.243
385259 98.854 -68.362 50.198
385260 98.897 -68.363 50.151
385261 98.939 -68.362 50.102
385262 98.983 -68.363 50.054
385263 99.028 -68.366 50.006
385264 99.074 -68.369 49.957
385265 99.120 -68.372 49.910
385266 99.164 -68.377 49.862
385267 99.194 -68.394 49.816
385268 99.224 -68.412 49.774
385269 99.254 -68.430 49.729
385270 99.286 -68.446 49.683
385271 99.326 -68.456 49.635
385272 99.366 -68.466 49.588
385273 99.406 -68.476 49.539
385274 99.446 -68.485 49.492
385275 99.486 -68.495 49.445
385276 99.526 -68.505 49.399
385277 99.565 -68.515 49.352
385278 99.577 -68.526 49.309
385279 99.632 -68.530 49.268
385280 99.667 -68.534 49.227
385281 99.701 -68.538 49.187
385282 99.736 -68.542 49.147
385283 99.765 -68.556 49.108
385284 99.790 -68.576 49.069
385285 99.816 -68.595 49.028
385286 99.845 -68.614 48.988
385287 99.899 -68.618 48.944
385288 99.952 -68.623 48.898
385289 99.990 -68.576 48.851
385290 100.022 -68.512 48.803
385291 100.055 -68.448 48.752
385292 100.089 -68.459 48.701
385293 100.122 -68.498 48.649
385294 100.154 -68.537 48.596
385295 100.186 -68.576 48.544
385296 100.218 -68.614 48.497
385297 100.250 -68.653 48.450
385298 100.282 -68.692 48.403
385299 100.315 -68.731 48.357
385300 100.345 -68.750 48.312
385301 100.374 -68.755 48.268
385302 100.403 -68.759 48.225
385303 100.439 -68.759 48.182
385742 90.307 -67.979 54.814
385743 90.338 -67.990 54.809
385744 90.369 -68.000 54.806
385745 90.400 -68.010 54.805
385746 90.432 -68.021 54.807
385747 90.463 -68.031 54.811
385748 90.494 -68.042 54.815
385749 90.525 -68.053 54.822
385750 90.556 -68.063 54.829
385751 90.588 -68.074 54.837
385752 90.620 -68.085 54.845
385753 90.678 -68.114 54.852
385754 90.736 -68.144 54.860
385755 90.795 -68.174 54.867
385756 90.869 -68.221 54.872
385757 90.921 -68.210 54.875
385758 90.972 -68.200 54.870
385759 91.023 -68.189 54.860
385760 91.075 -68.178 54.847
385761 91.126 -68.168 54.831
385762 91.177 -68.156 54.813
385763 91.229 -68.146 54.796
385764 91.280 -68.135 54.779
385765 91.331 -68.125 54.761
385766 91.383 -68.114 54.742
385767 91.434 -68.103 54.727
385768 91.486 -68.093 54.713
385769 91.537 -68.082 54.699
385770 91.588 -68.071 54.687
385771 91.640 -68.060 54.677
385772 91.690 -68.050 54.669
385773 91.741 -68.049 54.661
385774 91.784 -68.050 54.655
385775 91.826 -68.050 54.649
385776 91.868 -68.051 54.644
385777 91.911 -68.052 54.642
385778 91.952 -68.056 54.641
385779 91.992 -68.065 54.638
385780 92.033 -68.075 54.636
385781 92.073 -68.084 54.629
385782 92.113 -68.094 54.620
385783 92.154 -68.103 54.609
385784 92.195 -68.112 54.595
385785 92.235 -68.122 54.580
385786 92.275 -68.132 54.562
385787 92.323 -68.131 54.544
385788 92.373 -68.128 54.526
385789 92.422 -68.125 54.508
385790 92.472 -68.122 54.490
385791 92.521 -68.120 54.471
385792 92.571 -68.117 54.450
385793 92.625 -68.110 54.429
385794 92.683 -68.100 54.408
385795 92.741 -68.089 54.386
385796 92.799 -68.078 54.365
385797 92.858 -68.067 54.343
385798 92.916 -68.056 54.321
385799 92.974 -68.045 54.299
385800 93.012 -68.045 54.278
385801 93.047 -68.046 54.256
385802 93.099 -68.042 54.236
385803 93.152 -68.037 54.217
385804 93.204 -68.034 54.197
385805 93.256 -68.032 54.179
385806 93.307 -68.029 54.161
385807 93.359 -68.027 54.144
385808 93.410 -68.024 54.130
385809 93.460 -68.021 54.116
385810 93.510 -68.018 54.103
385811 93.560 -68.015 54.090
385812 93.611 -68.012 54.075
385813 93.660 -68.009 54.060
385814 93.709 -68.007 54.044
385815 93.758 -68.004 54.026
385816 93.806 -68.002 54.007
385817 93.856 -67.999 53.987
385818 93.905 -67.996 53.968
385819 93.954 -67.993 53.950
385820 94.002 -67.990 53.936
385821 94.051 -67.987 53.921
385822 94.100 -67.985 53.906
385823 94.149 -67.982 53.889
385824 94.198 -67.979 53.870
385825 94.247 -67.977 53.849
385826 94.296 -67.974 53.825
385827 94.345 -67.971 53.800
385828 94.390 -67.971 53.772
385829 94.436 -67.972 53.741
385830 94.482 -67.972 53.708
385831 94.519 -67.967 53.677
385832 94.555 -67.961 53.649
385833 94.592 -67.955 53.622
385834 94.628 -67.948 53.596
385835 94.657 -67.932 53.574
385836 94.685 -67.916 53.552
385837 94.710 -67.903 53.532
385838 94.735 -67.890 53.513
385839 94.760 -67.877 53.496
385840 94.785 -67.864 53.479
385841 94.810 -67.851 53.462
385842 94.834 -67.837 53.446
385843 94.859 -67.824 53.431
385844 94.884 -67.811 53.414
385845 94.909 -67.798 53.397
385846 94.934 -67.785 53.379
385847 94.959 -67.772 53.359
385848 94.984 -67.759 53.339
385849 95.008 -67.746 53.318
385850 95.034 -67.733 53.297
385851 95.060 -67.724 53.276
385852 95.099 -67.731 53.255
385853 95.136 -67.739 53.234
385854 95.174 -67.748 53.213
385855 95.212 -67.755 53.192
385856 95.250 -67.764 53.172
385857 95.287 -67.772 53.152
385858 95.326 -67.777 53.133
385859 95.367 -67.778 53.114
385860 95.408 -67.779 53.095
385861 95.450 -67.780 53.077
385862 95.492 -67.781 53.058
385863 95.535 -67.783 53.040
385864 95.577 -67.786 53.021
385865 95.617 -67.793 53.002
385866 95.658 -67.800 52.983
385867 95.699 -67.807 52.964
385868 95.740 -67.815 52.945
385869 95.781 -67.822 52.928
385870 95.822 -67.829 52.912
385871 95.863 -67.837 52.895
385872 95.903 -67.846 52.879
385873 95.937 -67.852 52.863
385874 95.966 -67.855 52.848
385875 95.996 -67.857 52.833
385876 96.032 -67.862 52.816
385877 96.070 -67.867 52.800
385878 96.107 -67.872 52.785
385879 96.145 -67.877 52.770
385880 96.183 -67.883 52.754
385881 96.221 -67.888 52.737
385882 96.259 -67.893 52.720
385883 96.295 -67.899 52.703
385884 96.330 -67.905 52.684
385885 96.365 -67.912 52.666
385886 96.400 -67.918 52.648
385887 96.436 -67.924 52.628
385888 96.470 -67.931 52.607
385889 96.506 -67.937 52.587
385890 96.541 -67.943 52.567
385891 96.576 -67.950 52.547
385892 96.610 -67.957 52.527
385893 96.642 -67.964 52.507
385894 96.670 -67.968 52.485
385895 96.698 -67.974 52.462
385896 96.726 -67.979 52.441
385897 96.754 -67.984 52.421
385898 96.783 -67.990 52.402
385899 96.811 -67.995 52.384
385900 96.839 -68.000 52.365
385901 96.867 -68.006 52.344
385902 96.895 -68.011 52.327
385903 96.926 -68.018 52.315
385904 96.959 -68.027 52.306
385905 96.993 -68.035 52.300
385906 97.026 -68.043 52.296
385907 97.059 -68.053 52.294
385908 97.092 -68.061 52.293
385909 97.125 -68.060 52.292
385910 97.157 -68.055 52.289
385911 97.195 -68.052 52.280
385912 97.239 -68.053 52.269
385913 97.270 -68.056 52.252
385914 97.300 -68.059 52.229
385915 97.330 -68.062 52.201
385916 97.362 -68.064 52.171
385917 97.396 -68.063 52.137
385918 97.430 -68.062 52.102
385919 97.463 -68.061 52.065
385920 97.498 -68.062 52.024
385921 97.535 -68.074 51.987
385922 97.573 -68.086 51.952
385923 97.607 -68.084 51.920
385924 97.638 -68.075 51.890
385925 97.670 -68.073 51.858
385926 97.706 -68.089 51.820
385927 97.741 -68.106 51.779
385928 97.777 -68.122 51.738
385929 97.813 -68.137 51.694
385930 97.849 -68.154 51.653
385931 97.884 -68.164 51.612
385932 97.918 -68.173 51.571
385933 97.950 -68.169 51.525
385934 97.981 -68.162 51.475
385935 98.012 -68.157 51.427
385936 98.044 -68.151 51.379
385937 98.075 -68.146 51.333
385938 98.106 -68.140 51.286
385939 98.138 -68.134 51.235
385940 98.171 -68.144 51.183
385941 98.205 -68.157 51.129
385942 98.238 -68.169 51.075
385943 98.271 -68.182 51.021
385944 98.305 -68.195 50.967
385945 98.338 -68.208 50.914
385946 98.371 -68.221 50.861
385947 98.405 -68.234 50.811
385948 98.438 -68.247 50.760
385949 98.472 -68.259 50.711
385950 98.505 -68.272 50.663
385951 98.538 -68.282 50.613
385952 98.573 -68.291 50.563
385953 98.607 -68.300 50.515
385954 98.647 -68.297 50.467
385955 98.688 -68.288 50.419
385956 98.718 -68.299 50.369
385957 98.747 -68.310 50.322
385958 98.776 -68.321 50.276
385959 98.812 -68.326 50.231
385960 98.849 -68.330 50.188
385961 98.887 -68.333 50.143
385962 98.930 -68.333 50.096
385963 98.976 -68.336 50.049
385964 99.022 -68.339 50.002
385965 99.068 -68.342 49.953
385966 99.113 -68.346 49.904
385967 99.159 -68.349 49.858
385968 99.199 -68.357 49.813
385969 99.228 -68.375 49.769
385970 99.259 -68.393 49.723
385971 99.291 -68.408 49.675
385972 99.332 -68.418 49.627
385973 99.371 -68.428 49.578
385974 99.412 -68.438 49.528
385975 99.451 -68.448 49.481
385976 99.492 -68.457 49.434
385977 99.532 -68.467 49.387
385978 99.572 -68.477 49.341
385979 99.603 -68.488 49.297
385980 99.643 -68.492 49.257
385981 99.678 -68.495 49.217
385982 99.712 -68.500 49.177
385983 99.747 -68.504 49.136
385984 99.774 -68.520 49.097
385985 99.800 -68.541 49.058
385986 99.825 -68.560 49.018
385987 99.851 -68.580 48.980
385988 99.904 -68.586 48.936
385989 99.949 -68.565 48.890
385990 99.982 -68.501 48.842
385991 100.020 -68.446 48.793
385992 100.060 -68.405 48.742
385993 100.092 -68.443 48.692
385994 100.124 -68.482 48.640
385995 100.156 -68.521 48.589
385996 100.188 -68.560 48.538
385997 100.220 -68.598 48.491
385998 100.253 -68.638 48.445
385999 100.285 -68.676 48.399
386000 100.316 -68.705 48.353
386001 100.345 -68.710 48.309
386002 100.374 -68.714 48.264
386003 100.408 -68.715 48.221
386004 100.448 -68.714 48.179
386443 90.295 -67.946 54.817
386444 90.326 -67.957 54.814
386445 90.358 -67.968 54.812
386446 90.388 -67.978 54.814
386447 90.419 -67.989 54.817
386448 90.451 -67.999 54.822
386449 90.482 -68.010 54.828
386450 90.513 -68.020 54.836
386451 90.544 -68.031 54.844
386452 90.576 -68.041 54.853
386453 90.614 -68.057 54.862
386454 90.672 -68.086 54.870
386455 90.730 -68.115 54.878
386456 90.802 -68.164 54.885
386457 90.870 -68.167 54.889
386458 90.923 -68.158 54.888
386459 90.972 -68.148 54.881
386460 91.023 -68.137 54.869
386461 91.073 -68.128 54.854
386462 91.122 -68.117 54.836
386463 91.172 -68.107 54.818
386464 91.223 -68.097 54.801
386465 91.273 -68.087 54.783
386466 91.323 -68.076 54.766
386467 91.374 -68.065 54.747
386468 91.426 -68.055 54.733
386469 91.477 -68.044 54.719
386470 91.528 -68.033 54.707
386471 91.580 -68.022 54.696
386472 91.631 -68.012 54.689
386473 91.682 -68.001 54.682
386474 91.734 -67.991 54.676
386475 91.779 -67.991 54.670
386476 91.821 -67.992 54.665
386477 91.863 -67.997 54.661
386478 91.903 -68.007 54.659
386479 91.943 -68.016 54.657
386480 91.984 -68.026 54.653
386481 92.024 -68.035 54.648
386482 92.065 -68.045 54.639
386483 92.105 -68.055 54.628
386484 92.146 -68.064 54.614
386485 92.186 -68.074 54.599
386486 92.226 -68.083 54.582
386487 92.267 -68.093 54.564
386488 92.307 -68.103 54.546
386489 92.357 -68.100 54.527
386490 92.406 -68.097 54.507
386491 92.457 -68.093 54.488
386492 92.515 -68.081 54.468
386493 92.574 -68.071 54.448
386494 92.632 -68.060 54.426
386495 92.690 -68.050 54.405
386496 92.748 -68.038 54.383
386497 92.807 -68.028 54.362
386498 92.865 -68.017 54.341
386499 92.923 -68.006 54.320
386500 92.980 -67.996 54.299
386501 93.014 -67.997 54.279
386502 93.049 -67.999 54.260
386503 93.092 -67.997 54.241
386504 93.145 -67.992 54.223
386505 93.199 -67.987 54.204
386506 93.251 -67.985 54.186
386507 93.302 -67.982 54.169
386508 93.354 -67.979 54.152
386509 93.406 -67.977 54.137
386510 93.458 -67.974 54.125
386511 93.510 -67.971 54.112
386512 93.560 -67.968 54.097
386513 93.611 -67.966 54.083
386514 93.660 -67.963 54.068
386515 93.710 -67.960 54.052
386516 93.759 -67.957 54.035
386517 93.808 -67.955 54.016
386518 93.857 -67.952 53.996
386519 93.906 -67.949 53.976
386520 93.955 -67.946 53.958
386521 94.003 -67.943 53.942
386522 94.052 -67.941 53.928
386523 94.102 -67.938 53.912
386524 94.149 -67.937 53.895
386525 94.195 -67.937 53.877
386526 94.241 -67.937 53.857
386527 94.286 -67.937 53.834
386528 94.332 -67.937 53.809
386529 94.378 -67.937 53.780
386530 94.418 -67.934 53.748
386531 94.454 -67.927 53.714
386532 94.490 -67.921 53.683
386533 94.527 -67.915 53.654
386534 94.563 -67.909 53.627
386535 94.600 -67.903 53.601
386536 94.636 -67.897 53.577
386537 94.666 -67.883 53.556
386538 94.694 -67.867 53.535
386539 94.723 -67.851 53.515
386540 94.751 -67.836 53.498
386541 94.776 -67.822 53.480
386542 94.801 -67.809 53.463
386543 94.826 -67.796 53.446
386544 94.850 -67.783 53.431
386545 94.875 -67.770 53.414
386546 94.900 -67.757 53.396
386547 94.925 -67.744 53.378
386548 94.950 -67.730 53.359
386549 94.975 -67.717 53.339
386550 95.000 -67.705 53.318
386551 95.025 -67.692 53.297
386552 95.060 -67.695 53.276
386553 95.098 -67.704 53.254
386554 95.135 -67.712 53.232
386555 95.173 -67.720 53.211
386556 95.211 -67.728 53.191
386557 95.251 -67.732 53.171
386558 95.291 -67.733 53.149
386559 95.332 -67.734 53.129
386560 95.373 -67.735 53.110
386561 95.413 -67.736 53.090
386562 95.455 -67.737 53.072
386563 95.497 -67.739 53.052
386564 95.540 -67.740 53.032
386565 95.581 -67.746 53.013
386566 95.622 -67.753 52.994
386567 95.663 -67.761 52.974
386568 95.703 -67.768 52.954
386569 95.744 -67.775 52.934
386570 95.785 -67.784 52.916
386571 95.826 -67.793 52.898
386572 95.867 -67.802 52.881
386573 95.907 -67.811 52.865
386574 95.942 -67.818 52.849
386575 95.969 -67.820 52.834
386576 95.998 -67.823 52.818
386577 96.033 -67.827 52.801
386578 96.071 -67.832 52.785
386579 96.108 -67.837 52.769
386580 96.146 -67.843 52.753
386581 96.181 -67.849 52.737
386582 96.217 -67.855 52.720
386583 96.252 -67.862 52.703
386584 96.287 -67.868 52.686
386585 96.322 -67.874 52.668
386586 96.357 -67.880 52.650
386587 96.392 -67.887 52.631
386588 96.428 -67.893 52.611
386589 96.463 -67.899 52.591
386590 96.498 -67.905 52.570
386591 96.533 -67.912 52.551
386592 96.568 -67.919 52.531
386593 96.602 -67.926 52.511
386594 96.637 -67.933 52.491
386595 96.667 -67.939 52.470
386596 96.695 -67.944 52.448
386597 96.723 -67.949 52.428
386598 96.751 -67.955 52.409
386599 96.780 -67.960 52.391
386600 96.808 -67.965 52.373
386601 96.836 -67.971 52.355
386602 96.864 -67.976 52.337
386603 96.895 -67.983 52.323
386604 96.928 -67.991 52.312
386605 96.961 -68.000 52.305
386606 96.995 -68.009 52.301
386607 97.028 -68.017 52.299
386608 97.061 -68.026 52.299
386609 97.094 -68.034 52.298
386610 97.126 -68.033 52.295
386611 97.159 -68.028 52.289
386612 97.203 -68.028 52.279
386613 97.245 -68.028 52.266
386614 97.276 -68.031 52.247
386615 97.307 -68.035 52.221
386616 97.337 -68.037 52.190
386617 97.371 -68.037 52.158
386618 97.405 -68.035 52.122
386619 97.438 -68.034 52.085
386620 97.472 -68.033 52.049
386621 97.506 -68.031 52.011
386622 97.540 -68.033 51.977
386623 97.578 -68.045 51.945
386624 97.614 -68.052 51.914
386625 97.645 -68.041 51.882
386626 97.678 -68.045 51.848
386627 97.714 -68.061 51.808
386628 97.749 -68.078 51.766
386629 97.785 -68.094 51.722
386630 97.821 -68.110 51.679
386631 97.857 -68.126 51.637
386632 97.892 -68.139 51.597
386633 97.924 -68.137 51.555
386634 97.955 -68.131 51.509
386635 97.987 -68.126 51.459
386636 98.018 -68.120 51.411
386637 98.050 -68.115 51.365
386638 98.081 -68.109 51.319
386639 98.112 -68.103 51.273
386640 98.144 -68.102 51.222
386641 98.178 -68.115 51.170
386642 98.211 -68.128 51.116
386643 98.244 -68.141 51.063
386644 98.278 -68.153 51.011
386645 98.311 -68.166 50.959
386646 98.345 -68.178 50.907
386647 98.378 -68.191 50.854
386648 98.412 -68.204 50.803
386649 98.445 -68.217 50.752
386650 98.479 -68.228 50.703
386651 98.512 -68.237 50.655
386652 98.546 -68.246 50.607
386653 98.580 -68.254 50.556
386654 98.616 -68.264 50.507
386655 98.652 -68.271 50.459
386656 98.693 -68.263 50.411
386657 98.734 -68.253 50.363
386658 98.774 -68.247 50.315
386659 98.803 -68.258 50.268
386660 98.833 -68.269 50.223
386661 98.862 -68.281 50.179
386662 98.891 -68.291 50.136
386663 98.925 -68.306 50.091
386664 98.970 -68.310 50.044
386665 99.015 -68.313 49.997
386666 99.061 -68.316 49.949
386667 99.106 -68.319 49.901
386668 99.152 -68.322 49.854
386669 99.198 -68.325 49.809
386670 99.233 -68.338 49.763
386671 99.263 -68.356 49.715
386672 99.297 -68.370 49.666
386673 99.337 -68.380 49.617
386674 99.377 -68.390 49.568
386675 99.417 -68.400 49.518
386676 99.457 -68.410 49.471
386677 99.497 -68.419 49.423
386678 99.538 -68.429 49.376
386679 99.577 -68.438 49.330
386680 99.611 -68.447 49.286
386681 99.655 -68.453 49.246
386682 99.689 -68.457 49.206
386683 99.723 -68.461 49.166
386684 99.758 -68.466 49.126
386685 99.783 -68.485 49.086
386686 99.809 -68.506 49.048
386687 99.835 -68.526 49.009
386688 99.860 -68.545 48.970
386689 99.909 -68.553 48.927
386690 99.942 -68.490 48.880
386691 99.985 -68.444 48.832
386692 100.028 -68.398 48.783
386693 100.062 -68.388 48.732
386694 100.094 -68.428 48.682
386695 100.126 -68.466 48.632
386696 100.158 -68.505 48.581
386697 100.191 -68.544 48.532
386698 100.223 -68.582 48.486
386699 100.255 -68.622 48.440
386700 100.287 -68.660 48.394
386701 100.316 -68.664 48.349
386702 100.345 -68.669 48.305
386703 100.378 -68.671 48.260
386704 100.417 -68.670 48.217
386705 100.457 -68.667 48.175
387143 90.252 -67.904 54.826
387144 90.283 -67.915 54.822
387145 90.314 -67.924 54.820
387146 90.345 -67.935 54.820
387147 90.377 -67.946 54.823
387148 90.407 -67.956 54.828
387149 90.439 -67.967 54.834
387150 90.470 -67.978 54.842
387151 90.501 -67.988 54.851
387152 90.532 -67.999 54.860
387153 90.564 -68.009 54.869
387154 90.607 -68.028 54.879
387155 90.666 -68.058 54.888
387156 90.731 -68.097 54.896
387157 90.801 -68.112 54.902
387158 90.868 -68.109 54.903
387159 90.929 -68.104 54.899
387160 90.978 -68.094 54.890
387161 91.029 -68.084 54.876
387162 91.079 -68.074 54.860
387163 91.129 -68.063 54.842
387164 91.179 -68.053 54.823
387165 91.229 -68.043 54.807
387166 91.279 -68.033 54.789
387167 91.328 -68.022 54.772
387168 91.379 -68.012 54.755
387169 91.429 -68.002 54.741
387170 91.478 -67.992 54.729
387171 91.529 -67.981 54.718
387172 91.579 -67.971 54.709
387173 91.629 -67.961 54.703
387174 91.679 -67.951 54.698
387175 91.729 -67.940 54.693
387176 91.774 -67.936 54.688
387177 91.815 -67.946 54.683
387178 91.855 -67.955 54.679
387179 91.895 -67.965 54.676
387180 91.936 -67.974 54.673
387181 91.977 -67.984 54.666
387182 92.017 -67.993 54.658
387183 92.057 -68.003 54.647
387184 92.098 -68.012 54.633
387185 92.138 -68.021 54.617
387186 92.178 -68.031 54.603
387187 92.219 -68.040 54.584
387188 92.260 -68.050 54.565
387189 92.300 -68.059 54.547
387190 92.348 -68.064 54.527
387191 92.406 -68.053 54.506
387192 92.464 -68.043 54.487
387193 92.523 -68.031 54.467
387194 92.581 -68.021 54.445
387195 92.639 -68.010 54.424
387196 92.698 -67.999 54.402
387197 92.756 -67.988 54.380
387198 92.814 -67.978 54.359
387199 92.872 -67.967 54.339
387200 92.931 -67.956 54.319
387201 92.983 -67.949 54.300
387202 93.017 -67.949 54.282
387203 93.052 -67.951 54.264
387204 93.087 -67.952 54.247
387205 93.139 -67.947 54.232
387206 93.192 -67.943 54.215
387207 93.245 -67.937 54.198
387208 93.297 -67.935 54.179
387209 93.349 -67.932 54.162
387210 93.401 -67.930 54.147
387211 93.453 -67.927 54.132
387212 93.505 -67.924 54.119
387213 93.557 -67.921 54.106
387214 93.609 -67.919 54.091
387215 93.660 -67.916 54.076
387216 93.710 -67.913 54.060
387217 93.760 -67.911 54.043
387218 93.809 -67.908 54.024
387219 93.858 -67.905 54.004
387220 93.907 -67.902 53.984
387221 93.954 -67.901 53.965
387222 94.000 -67.901 53.947
387223 94.046 -67.902 53.933
387224 94.091 -67.902 53.918
387225 94.137 -67.902 53.902
387226 94.183 -67.902 53.884
387227 94.228 -67.902 53.864
387228 94.274 -67.903 53.842
387229 94.317 -67.901 53.816
387230 94.353 -67.895 53.787
387231 94.389 -67.889 53.755
387232 94.425 -67.883 53.721
387233 94.462 -67.877 53.688
387234 94.498 -67.871 53.658
387235 94.535 -67.865 53.630
387236 94.571 -67.858 53.605
387237 94.607 -67.852 53.581
387238 94.643 -67.846 53.559
387239 94.675 -67.834 53.538
387240 94.704 -67.818 53.518
387241 94.732 -67.802 53.500
387242 94.761 -67.786 53.482
387243 94.789 -67.770 53.464
387244 94.817 -67.755 53.447
387245 94.841 -67.742 53.431
387246 94.866 -67.729 53.414
387247 94.891 -67.716 53.396
387248 94.916 -67.702 53.378
387249 94.941 -67.689 53.359
387250 94.966 -67.677 53.339
387251 94.991 -67.663 53.319
387252 95.022 -67.660 53.298
387253 95.059 -67.668 53.277
387254 95.097 -67.677 53.254
387255 95.135 -67.684 53.232
387256 95.175 -67.686 53.211
387257 95.216 -67.688 53.190
387258 95.257 -67.689 53.170
387259 95.298 -67.690 53.148
387260 95.338 -67.691 53.127
387261 95.379 -67.692 53.107
387262 95.419 -67.693 53.086
387263 95.460 -67.695 53.067
387264 95.502 -67.695 53.047
387265 95.544 -67.699 53.026
387266 95.585 -67.707 53.006
387267 95.626 -67.714 52.987
387268 95.667 -67.722 52.967
387269 95.707 -67.731 52.947
387270 95.748 -67.740 52.927
387271 95.789 -67.749 52.908
387272 95.830 -67.758 52.889
387273 95.871 -67.768 52.872
387274 95.911 -67.777 52.855
387275 95.947 -67.783 52.839
387276 95.974 -67.786 52.822
387277 96.001 -67.788 52.806
387278 96.033 -67.792 52.789
387279 96.068 -67.799 52.773
387280 96.103 -67.805 52.757
387281 96.138 -67.811 52.740
387282 96.174 -67.817 52.724
387283 96.208 -67.824 52.707
387284 96.244 -67.830 52.690
387285 96.279 -67.836 52.673
387286 96.314 -67.843 52.655
387287 96.349 -67.849 52.636
387288 96.384 -67.855 52.617
387289 96.419 -67.861 52.597
387290 96.455 -67.868 52.576
387291 96.490 -67.874 52.555
387292 96.525 -67.880 52.534
387293 96.559 -67.887 52.515
387294 96.594 -67.895 52.496
387295 96.628 -67.902 52.475
387296 96.663 -67.909 52.454
387297 96.692 -67.915 52.434
387298 96.720 -67.920 52.415
387299 96.748 -67.925 52.397
387300 96.777 -67.931 52.378
387301 96.805 -67.935 52.361
387302 96.833 -67.940 52.344
387303 96.864 -67.948 52.329
387304 96.897 -67.956 52.318
387305 96.930 -67.965 52.308
387306 96.963 -67.974 52.302
387307 96.997 -67.982 52.299
387308 97.030 -67.990 52.301
387309 97.063 -67.999 52.302
387310 97.096 -68.008 52.299
387311 97.129 -68.006 52.294
387312 97.164 -68.002 52.286
387313 97.211 -68.003 52.275
387314 97.252 -68.004 52.260
387315 97.283 -68.007 52.239
387316 97.313 -68.010 52.212
387317 97.347 -68.009 52.178
387318 97.380 -68.009 52.144
387319 97.414 -68.007 52.106
387320 97.447 -68.006 52.068
387321 97.481 -68.005 52.033
387322 97.514 -68.003 51.997
387323 97.548 -68.003 51.966
387324 97.582 -68.003 51.937
387325 97.620 -68.015 51.907
387326 97.652 -68.009 51.873
387327 97.686 -68.017 51.837
387328 97.722 -68.034 51.796
387329 97.758 -68.050 51.751
387330 97.794 -68.065 51.706
387331 97.828 -68.077 51.663
387332 97.862 -68.085 51.622
387333 97.896 -68.094 51.582
387334 97.930 -68.101 51.540
387335 97.961 -68.095 51.492
387336 97.993 -68.090 51.443
387337 98.024 -68.084 51.396
387338 98.056 -68.078 51.350
387339 98.087 -68.072 51.306
387340 98.119 -68.066 51.260
387341 98.151 -68.073 51.210
387342 98.184 -68.086 51.157
387343 98.218 -68.099 51.104
387344 98.251 -68.111 51.052
387345 98.285 -68.124 51.001
387346 98.318 -68.137 50.952
387347 98.352 -68.150 50.901
387348 98.385 -68.162 50.848
387349 98.418 -68.175 50.798
387350 98.452 -68.183 50.747
387351 98.486 -68.192 50.697
387352 98.520 -68.200 50.648
387353 98.554 -68.209 50.601
387354 98.588 -68.218 50.551
387355 98.624 -68.228 50.501
387356 98.659 -68.238 50.453
387357 98.698 -68.236 50.405
387358 98.739 -68.227 50.358
387359 98.781 -68.219 50.311
387360 98.821 -68.209 50.263
387361 98.859 -68.206 50.215
387362 98.888 -68.218 50.170
387363 98.918 -68.228 50.129
387364 98.952 -68.244 50.085
387365 98.987 -68.258 50.039
387366 99.021 -68.273 49.992
387367 99.056 -68.288 49.944
387368 99.099 -68.293 49.897
387369 99.145 -68.296 49.849
387370 99.191 -68.299 49.802
387371 99.236 -68.303 49.755
387372 99.268 -68.319 49.706
387373 99.303 -68.332 49.657
387374 99.343 -68.342 49.608
387375 99.383 -68.352 49.558
387376 99.423 -68.362 49.509
387377 99.464 -68.372 49.460
387378 99.498 -68.379 49.412
387379 99.532 -68.388 49.365
387380 99.566 -68.395 49.320
387381 99.600 -68.403 49.276
387382 99.647 -68.410 49.235
387383 99.694 -68.417 49.194
387384 99.735 -68.423 49.154
387385 99.767 -68.431 49.115
387386 99.792 -68.451 49.076
387387 99.818 -68.471 49.038
387388 99.844 -68.491 49.000
387389 99.869 -68.510 48.960
387390 99.907 -68.488 48.917
387391 99.950 -68.442 48.870
387392 99.994 -68.397 48.822
387393 100.037 -68.351 48.773
387394 100.064 -68.372 48.723
387395 100.097 -68.412 48.673
387396 100.129 -68.451 48.623
387397 100.161 -68.489 48.574
387398 100.193 -68.528 48.526
387399 100.225 -68.567 48.481
387400 100.257 -68.606 48.435
387401 100.287 -68.620 48.389
387402 100.316 -68.624 48.344
387403 100.347 -68.627 48.300
387404 100.387 -68.626 48.256
387405 100.427 -68.623 48.212
387844 90.239 -67.872 54.831
387845 90.271 -67.882 54.828
387846 90.302 -67.893 54.827
387847 90.333 -67.903 54.829
387848 90.365 -67.914 54.833
387849 90.396 -67.924 54.840
387850 90.427 -67.934 54.848
387851 90.458 -67.945 54.857
387852 90.489 -67.956 54.867
387853 90.520 -67.966 54.877
387854 90.551 -67.977 54.887
387855 90.601 -68.000 54.897
387856 90.660 -68.030 54.908
387857 90.732 -68.059 54.914
387858 90.799 -68.056 54.917
387859 90.866 -68.053 54.916
387860 90.933 -68.050 54.910
387861 90.984 -68.040 54.898
387862 91.035 -68.030 54.883
387863 91.085 -68.020 54.866
387864 91.135 -68.009 54.847
387865 91.185 -67.999 54.829
387866 91.235 -67.989 54.813
387867 91.285 -67.979 54.796
387868 91.334 -67.968 54.779
387869 91.385 -67.959 54.764
387870 91.435 -67.948 54.750
387871 91.485 -67.938 54.740
387872 91.535 -67.927 54.731
387873 91.585 -67.918 54.723
387874 91.635 -67.907 54.719
387875 91.685 -67.897 54.715
387876 91.730 -67.888 54.711
387877 91.772 -67.880 54.706
387878 91.812 -67.887 54.703
387879 91.853 -67.897 54.697
387880 91.893 -67.906 54.693
387881 91.934 -67.916 54.687
387882 91.974 -67.925 54.678
387883 92.014 -67.935 54.666
387884 92.055 -67.944 54.653
387885 92.095 -67.954 54.638
387886 92.136 -67.963 54.622
387887 92.176 -67.973 54.606
387888 92.217 -67.983 54.586
387889 92.257 -67.992 54.566
387890 92.306 -67.997 54.546
387891 92.360 -68.000 54.526
387892 92.414 -68.003 54.506
387893 92.472 -67.992 54.487
387894 92.530 -67.981 54.465
387895 92.588 -67.971 54.443
387896 92.647 -67.960 54.420
387897 92.705 -67.949 54.398
387898 92.764 -67.938 54.377
387899 92.821 -67.927 54.357
387900 92.880 -67.917 54.338
387901 92.938 -67.905 54.319
387902 92.985 -67.901 54.302
387903 93.020 -67.902 54.286
387904 93.055 -67.903 54.271
387905 93.090 -67.905 54.257
387906 93.132 -67.903 54.244
387907 93.185 -67.898 54.230
387908 93.238 -67.893 54.212
387909 93.292 -67.888 54.193
387910 93.344 -67.885 54.174
387911 93.396 -67.882 54.156
387912 93.447 -67.880 54.140
387913 93.499 -67.877 54.126
387914 93.551 -67.874 54.113
387915 93.603 -67.871 54.098
387916 93.655 -67.869 54.084
387917 93.707 -67.866 54.069
387918 93.758 -67.864 54.052
387919 93.804 -67.865 54.033
387920 93.850 -67.866 54.012
387921 93.896 -67.867 53.991
387922 93.942 -67.867 53.972
387923 93.988 -67.868 53.954
387924 94.034 -67.868 53.939
387925 94.079 -67.868 53.924
387926 94.125 -67.868 53.908
387927 94.171 -67.868 53.890
387928 94.215 -67.868 53.870
387929 94.252 -67.862 53.849
387930 94.288 -67.855 53.824
387931 94.324 -67.849 53.795
387932 94.360 -67.843 53.762
387933 94.397 -67.837 53.728
387934 94.433 -67.831 53.694
387935 94.470 -67.825 53.663
387936 94.506 -67.819 53.634
387937 94.542 -67.813 53.610
387938 94.578 -67.807 53.587
387939 94.615 -67.801 53.564
387940 94.651 -67.795 53.543
387941 94.684 -67.785 53.523
387942 94.713 -67.769 53.503
387943 94.742 -67.752 53.485
387944 94.770 -67.736 53.467
387945 94.799 -67.721 53.449
387946 94.827 -67.705 53.432
387947 94.856 -67.689 53.415
387948 94.882 -67.674 53.397
387949 94.907 -67.661 53.378
387950 94.932 -67.648 53.359
387951 94.957 -67.635 53.340
387952 94.983 -67.624 53.319
387953 95.021 -67.633 53.298
387954 95.059 -67.640 53.277
387955 95.100 -67.642 53.255
387956 95.140 -67.642 53.233
387957 95.181 -67.644 53.211
387958 95.222 -67.645 53.189
387959 95.262 -67.646 53.169
387960 95.303 -67.647 53.147
387961 95.344 -67.648 53.126
387962 95.385 -67.649 53.105
387963 95.425 -67.651 53.084
387964 95.466 -67.652 53.064
387965 95.508 -67.653 53.043
387966 95.548 -67.661 53.022
387967 95.589 -67.669 53.002
387968 95.630 -67.678 52.981
387969 95.671 -67.687 52.961
387970 95.711 -67.696 52.941
387971 95.752 -67.705 52.921
387972 95.793 -67.714 52.901
387973 95.834 -67.724 52.882
387974 95.875 -67.733 52.865
387975 95.915 -67.742 52.847
387976 95.952 -67.749 52.830
387977 95.979 -67.751 52.812
387978 96.006 -67.754 52.796
387979 96.034 -67.761 52.779
387980 96.068 -67.768 52.763
387981 96.101 -67.774 52.747
387982 96.134 -67.780 52.730
387983 96.168 -67.786 52.714
387984 96.201 -67.792 52.698
387985 96.236 -67.798 52.680
387986 96.271 -67.805 52.663
387987 96.306 -67.811 52.644
387988 96.341 -67.817 52.625
387989 96.377 -67.823 52.605
387990 96.411 -67.830 52.584
387991 96.447 -67.836 52.563
387992 96.482 -67.842 52.542
387993 96.517 -67.849 52.521
387994 96.551 -67.856 52.501
387995 96.586 -67.863 52.481
387996 96.620 -67.871 52.461
387997 96.655 -67.877 52.441
387998 96.689 -67.884 52.421
387999 96.718 -67.890 52.403
388000 96.746 -67.895 52.385
388001 96.774 -67.900 52.367
388002 96.802 -67.905 52.351
388003 96.833 -67.912 52.335
388004 96.866 -67.921 52.321
388005 96.899 -67.930 52.310
388006 96.932 -67.938 52.302
388007 96.965 -67.947 52.296
388008 96.999 -67.956 52.294
388009 97.032 -67.964 52.298
388010 97.065 -67.973 52.297
388011 97.098 -67.981 52.293
388012 97.131 -67.978 52.286
388013 97.172 -67.977 52.278
388014 97.218 -67.978 52.267
388015 97.259 -67.979 52.251
388016 97.290 -67.982 52.230
388017 97.322 -67.983 52.201
388018 97.356 -67.982 52.167
388019 97.389 -67.981 52.131
388020 97.423 -67.979 52.092
388021 97.456 -67.978 52.053
388022 97.490 -67.977 52.018
388023 97.523 -67.976 51.988
388024 97.557 -67.974 51.957
388025 97.591 -67.973 51.927
388026 97.625 -67.974 51.896
388027 97.659 -67.975 51.863
388028 97.695 -67.989 51.825
388029 97.730 -68.006 51.783
388030 97.764 -68.015 51.737
388031 97.799 -68.023 51.692
388032 97.832 -68.031 51.647
388033 97.867 -68.040 51.607
388034 97.900 -68.049 51.568
388035 97.935 -68.057 51.525
388036 97.967 -68.059 51.478
388037 97.999 -68.053 51.428
388038 98.030 -68.047 51.381
388039 98.062 -68.041 51.336
388040 98.093 -68.036 51.294
388041 98.125 -68.031 51.248
388042 98.158 -68.043 51.198
388043 98.191 -68.056 51.147
388044 98.225 -68.069 51.094
388045 98.258 -68.082 51.041
388046 98.291 -68.095 50.991
388047 98.325 -68.108 50.943
388048 98.358 -68.120 50.895
388049 98.392 -68.130 50.844
388050 98.426 -68.138 50.795
388051 98.460 -68.147 50.744
388052 98.494 -68.156 50.693
388053 98.527 -68.164 50.643
388054 98.561 -68.173 50.595
388055 98.596 -68.182 50.546
388056 98.632 -68.192 50.498
388057 98.667 -68.202 50.450
388058 98.703 -68.209 50.403
388059 98.744 -68.201 50.356
388060 98.785 -68.192 50.308
388061 98.827 -68.183 50.258
388062 98.868 -68.174 50.208
388063 98.909 -68.166 50.161
388064 98.945 -68.166 50.119
388065 98.980 -68.181 50.075
388066 99.014 -68.196 50.030
388067 99.049 -68.211 49.984
388068 99.083 -68.225 49.938
388069 99.118 -68.241 49.889
388070 99.152 -68.255 49.841
388071 99.187 -68.270 49.793
388072 99.230 -68.276 49.745
388073 99.273 -68.282 49.697
388074 99.309 -68.294 49.648
388075 99.349 -68.304 49.598
388076 99.384 -68.313 49.549
388077 99.419 -68.320 49.500
388078 99.453 -68.328 49.451
388079 99.487 -68.336 49.403
388080 99.521 -68.344 49.356
388081 99.556 -68.351 49.310
388082 99.590 -68.359 49.266
388083 99.636 -68.366 49.225
388084 99.684 -68.373 49.184
388085 99.731 -68.381 49.144
388086 99.763 -68.366 49.105
388087 99.791 -68.394 49.066
388088 99.820 -68.423 49.028
388089 99.849 -68.451 48.989
388090 99.876 -68.469 48.948
388091 99.916 -68.440 48.904
388092 99.959 -68.394 48.858
388093 100.002 -68.349 48.812
388094 100.035 -68.319 48.763
388095 100.067 -68.357 48.714
388096 100.099 -68.396 48.664
388097 100.131 -68.435 48.614
388098 100.163 -68.473 48.566
388099 100.195 -68.513 48.519
388100 100.227 -68.551 48.475
388101 100.258 -68.575 48.429
388102 100.287 -68.579 48.384
388103 100.317 -68.583 48.339
388104 100.356 -68.582 48.295
388105 100.396 -68.579 48.251
388106 100.436 -68.577 48.208
388545 90.219 -67.843 54.835
388546 90.253 -67.852 54.835
388547 90.288 -67.862 54.835
388548 90.321 -67.871 54.839
388549 90.353 -67.881 54.844
388550 90.384 -67.892 54.852
388551 90.415 -67.902 54.862
388552 90.446 -67.913 54.872
388553 90.477 -67.924 54.883
388554 90.508 -67.934 54.894
388555 90.539 -67.944 54.905
388556 90.595 -67.972 54.915
388557 90.664 -68.004 54.924
388558 90.730 -68.001 54.929
388559 90.797 -67.998 54.929
388560 90.864 -67.995 54.925
388561 90.931 -67.992 54.917
388562 90.991 -67.987 54.904
388563 91.041 -67.976 54.888
388564 91.091 -67.966 54.871
388565 91.141 -67.956 54.854
388566 91.191 -67.946 54.836
388567 91.241 -67.935 54.819
388568 91.291 -67.925 54.803
388569 91.341 -67.915 54.787
388570 91.391 -67.905 54.773
388571 91.441 -67.894 54.761
388572 91.491 -67.884 54.751
388573 91.541 -67.874 54.744
388574 91.591 -67.864 54.739
388575 91.641 -67.853 54.736
388576 91.686 -67.845 54.733
388577 91.728 -67.837 54.730
388578 91.769 -67.830 54.725
388579 91.810 -67.830 54.721
388580 91.850 -67.839 54.715
388581 91.891 -67.849 54.708
388582 91.931 -67.858 54.700
388583 91.971 -67.868 54.689
388584 92.012 -67.877 54.674
388585 92.053 -67.887 54.660
388586 92.093 -67.896 54.644
388587 92.133 -67.905 54.627
388588 92.174 -67.915 54.609
388589 92.214 -67.924 54.588
388590 92.265 -67.929 54.567
388591 92.318 -67.932 54.546
388592 92.372 -67.935 54.525
388593 92.426 -67.938 54.505
388594 92.480 -67.941 54.485
388595 92.537 -67.931 54.464
388596 92.596 -67.921 54.442
388597 92.654 -67.910 54.418
388598 92.713 -67.899 54.397
388599 92.771 -67.888 54.376
388600 92.829 -67.877 54.356
388601 92.887 -67.866 54.338
388602 92.945 -67.855 54.321
388603 92.988 -67.852 54.305
388604 93.023 -67.854 54.292
388605 93.058 -67.855 54.281
388606 93.092 -67.856 54.271
388607 93.127 -67.858 54.261
388608 93.179 -67.853 54.248
388609 93.231 -67.849 54.231
388610 93.285 -67.843 54.210
388611 93.338 -67.838 54.187
388612 93.391 -67.835 54.166
388613 93.442 -67.833 54.148
388614 93.494 -67.830 54.133
388615 93.546 -67.827 54.119
388616 93.598 -67.824 54.105
388617 93.650 -67.822 54.092
388618 93.697 -67.818 54.077
388619 93.743 -67.814 54.061
388620 93.789 -67.812 54.042
388621 93.836 -67.817 54.021
388622 93.882 -67.822 53.999
388623 93.928 -67.827 53.978
388624 93.975 -67.832 53.960
388625 94.021 -67.833 53.944
388626 94.067 -67.833 53.929
388627 94.113 -67.834 53.913
388628 94.150 -67.829 53.895
388629 94.187 -67.823 53.876
388630 94.223 -67.817 53.854
388631 94.259 -67.811 53.830
388632 94.295 -67.805 53.803
388633 94.332 -67.799 53.771
388634 94.368 -67.793 53.736
388635 94.404 -67.786 53.701
388636 94.441 -67.780 53.668
388637 94.477 -67.774 53.640
388638 94.513 -67.768 53.615
388639 94.550 -67.761 53.592
388640 94.586 -67.755 53.570
388641 94.623 -67.749 53.550
388642 94.660 -67.743 53.529
388643 94.696 -67.736 53.509
388644 94.723 -67.720 53.489
388645 94.751 -67.704 53.470
388646 94.779 -67.687 53.452
388647 94.808 -67.671 53.434
388648 94.837 -67.655 53.416
388649 94.865 -67.639 53.397
388650 94.894 -67.623 53.379
388651 94.922 -67.608 53.360
388652 94.948 -67.594 53.341
388653 94.984 -67.595 53.321
388654 95.024 -67.596 53.300
388655 95.065 -67.598 53.278
388656 95.106 -67.598 53.256
388657 95.146 -67.600 53.234
388658 95.187 -67.602 53.212
388659 95.228 -67.604 53.189
388660 95.268 -67.605 53.168
388661 95.308 -67.608 53.147
388662 95.349 -67.610 53.126
388663 95.389 -67.612 53.104
388664 95.429 -67.614 53.083
388665 95.470 -67.616 53.062
388666 95.512 -67.617 53.041
388667 95.552 -67.625 53.020
388668 95.593 -67.634 53.000
388669 95.634 -67.643 52.978
388670 95.675 -67.652 52.957
388671 95.715 -67.661 52.937
388672 95.756 -67.670 52.916
388673 95.797 -67.680 52.896
388674 95.838 -67.689 52.877
388675 95.879 -67.698 52.859
388676 95.919 -67.707 52.841
388677 95.957 -67.714 52.823
388678 95.984 -67.717 52.805
388679 96.011 -67.719 52.788
388680 96.037 -67.730 52.771
388681 96.071 -67.736 52.754
388682 96.104 -67.743 52.738
388683 96.137 -67.749 52.722
388684 96.171 -67.755 52.707
388685 96.204 -67.761 52.690
388686 96.238 -67.768 52.672
388687 96.271 -67.774 52.654
388688 96.305 -67.780 52.634
388689 96.338 -67.786 52.614
388690 96.372 -67.792 52.593
388691 96.405 -67.798 52.572
388692 96.439 -67.804 52.550
388693 96.474 -67.811 52.530
388694 96.509 -67.818 52.508
388695 96.543 -67.825 52.487
388696 96.578 -67.832 52.467
388697 96.612 -67.840 52.447
388698 96.647 -67.846 52.428
388699 96.682 -67.853 52.409
388700 96.715 -67.860 52.390
388701 96.743 -67.865 52.372
388702 96.771 -67.871 52.356
388703 96.802 -67.877 52.340
388704 96.835 -67.886 52.326
388705 96.868 -67.894 52.312
388706 96.901 -67.903 52.302
388707 96.934 -67.912 52.294
388708 96.967 -67.920 52.289
388709 97.001 -67.929 52.287
388710 97.034 -67.937 52.290
388711 97.067 -67.946 52.289
388712 97.100 -67.955 52.283
388713 97.134 -67.951 52.276
388714 97.180 -67.952 52.268
388715 97.226 -67.952 52.257
388716 97.265 -67.954 52.241
388717 97.297 -67.956 52.219
388718 97.331 -67.955 52.189
388719 97.364 -67.954 52.155
388720 97.398 -67.952 52.117
388721 97.432 -67.952 52.078
388722 97.465 -67.950 52.039
388723 97.499 -67.949 52.004
388724 97.532 -67.948 51.975
388725 97.566 -67.946 51.947
388726 97.599 -67.946 51.917
388727 97.633 -67.944 51.886
388728 97.667 -67.944 51.851
388729 97.701 -67.952 51.813
388730 97.735 -67.961 51.769
388731 97.769 -67.969 51.723
388732 97.803 -67.978 51.677
388733 97.837 -67.987 51.632
388734 97.871 -67.995 51.593
388735 97.905 -68.003 51.553
388736 97.939 -68.012 51.511
388737 97.973 -68.020 51.464
388738 98.005 -68.016 51.415
388739 98.036 -68.010 51.368
388740 98.067 -68.005 51.324
388741 98.099 -67.999 51.281
388742 98.131 -68.002 51.235
388743 98.164 -68.015 51.186
388744 98.198 -68.028 51.135
388745 98.231 -68.040 51.083
388746 98.265 -68.053 51.031
388747 98.298 -68.065 50.981
388748 98.332 -68.076 50.934
388749 98.366 -68.084 50.887
388750 98.400 -68.094 50.839
388751 98.433 -68.102 50.790
388752 98.467 -68.110 50.740
388753 98.501 -68.119 50.690
388754 98.535 -68.128 50.640
388755 98.569 -68.137 50.591
388756 98.604 -68.146 50.544
388757 98.640 -68.156 50.496
388758 98.675 -68.166 50.449
388759 98.711 -68.175 50.402
388760 98.749 -68.175 50.353
388761 98.790 -68.166 50.304
388762 98.832 -68.157 50.254
388763 98.873 -68.148 50.203
388764 98.914 -68.139 50.154
388765 98.956 -68.133 50.110
388766 98.993 -68.144 50.066
388767 99.028 -68.157 50.020
388768 99.064 -68.170 49.972
388769 99.099 -68.184 49.924
388770 99.135 -68.197 49.877
388771 99.170 -68.209 49.829
388772 99.206 -68.223 49.781
388773 99.241 -68.236 49.732
388774 99.277 -68.248 49.684
388775 99.308 -68.256 49.636
388776 99.340 -68.261 49.588
388777 99.374 -68.269 49.539
388778 99.409 -68.277 49.490
388779 99.443 -68.285 49.442
388780 99.477 -68.292 49.394
388781 99.511 -68.300 49.347
388782 99.546 -68.307 49.301
388783 99.580 -68.316 49.258
388784 99.626 -68.322 49.216
388785 99.673 -68.330 49.175
388786 99.713 -68.323 49.134
388787 99.759 -68.344 49.094
388788 99.800 -68.366 49.055
388789 99.829 -68.392 49.017
388790 99.857 -68.408 48.977
388791 99.885 -68.424 48.935
388792 99.924 -68.393 48.891
388793 99.968 -68.347 48.846
388794 100.011 -68.302 48.800
388795 100.037 -68.303 48.753
388796 100.069 -68.341 48.704
388797 100.101 -68.380 48.655
388798 100.133 -68.419 48.606
388799 100.165 -68.458 48.558
388800 100.198 -68.497 48.512
388801 100.229 -68.530 48.468
388802 100.258 -68.535 48.423
388803 100.287 -68.539 48.378
388804 100.326 -68.538 48.334
388805 100.366 -68.535 48.290
388806 100.405 -68.533 48.246
388807 100.445 -68.532 48.203
389245 90.164 -67.805 54.843
389246 90.199 -67.815 54.842
389247 90.233 -67.824 54.843
389248 90.268 -67.833 54.844
389249 90.303 -67.841 54.850
389250 90.337 -67.850 54.856
389251 90.372 -67.859 54.865
389252 90.403 -67.870 54.876
389253 90.434 -67.880 54.886
389254 90.465 -67.891 54.899
389255 90.496 -67.902 54.911
389256 90.531 -67.915 54.923
389257 90.595 -67.949 54.933
389258 90.662 -67.946 54.940
389259 90.728 -67.943 54.942
389260 90.795 -67.941 54.940
389261 90.862 -67.938 54.934
389262 90.929 -67.935 54.923
389263 90.996 -67.932 54.909
389264 91.047 -67.922 54.893
389265 91.097 -67.912 54.876
389266 91.147 -67.902 54.860
389267 91.197 -67.892 54.843
389268 91.247 -67.881 54.827
389269 91.297 -67.871 54.813
389270 91.347 -67.861 54.798
389271 91.397 -67.851 54.785
389272 91.447 -67.840 54.775
389273 91.497 -67.830 54.766
389274 91.547 -67.820 54.759
389275 91.597 -67.810 54.756
389276 91.643 -67.801 54.754
389277 91.684 -67.793 54.752
389278 91.725 -67.786 54.748
389279 91.767 -67.778 54.745
389280 91.807 -67.771 54.739
389281 91.848 -67.781 54.730
389282 91.888 -67.791 54.722
389283 91.929 -67.800 54.712
389284 91.969 -67.810 54.699
389285 92.010 -67.819 54.683
389286 92.050 -67.829 54.667
389287 92.090 -67.838 54.650
389288 92.131 -67.848 54.632
389289 92.171 -67.857 54.612
389290 92.223 -67.861 54.591
389291 92.277 -67.864 54.568
389292 92.331 -67.867 54.546
389293 92.384 -67.870 54.524
389294 92.438 -67.873 54.504
389295 92.492 -67.876 54.484
389296 92.546 -67.878 54.463
389297 92.603 -67.871 54.441
389298 92.662 -67.859 54.418
389299 92.720 -67.849 54.397
389300 92.778 -67.838 54.376
389301 92.837 -67.827 54.356
389302 92.895 -67.816 54.338
389303 92.953 -67.805 54.322
389304 92.991 -67.805 54.310
389305 93.026 -67.806 54.300
389306 93.061 -67.808 54.292
389307 93.095 -67.808 54.285
389308 93.130 -67.810 54.276
389309 93.171 -67.809 54.264
389310 93.225 -67.804 54.247
389311 93.278 -67.799 54.225
389312 93.331 -67.794 54.200
389313 93.384 -67.789 54.176
389314 93.437 -67.785 54.156
389315 93.489 -67.783 54.138
389316 93.541 -67.780 54.124
389317 93.589 -67.777 54.111
389318 93.635 -67.772 54.099
389319 93.681 -67.768 54.085
389320 93.727 -67.764 54.068
389321 93.773 -67.761 54.049
389322 93.820 -67.764 54.028
389323 93.866 -67.769 54.006
389324 93.913 -67.774 53.984
389325 93.960 -67.779 53.965
389326 94.006 -67.784 53.948
389327 94.053 -67.789 53.932
389328 94.092 -67.787 53.916
389329 94.130 -67.781 53.900
389330 94.167 -67.775 53.880
389331 94.203 -67.768 53.859
389332 94.240 -67.762 53.836
389333 94.277 -67.756 53.810
389334 94.314 -67.750 53.780
389335 94.351 -67.743 53.744
389336 94.388 -67.737 53.708
389337 94.424 -67.731 53.674
389338 94.461 -67.725 53.646
389339 94.498 -67.718 53.621
389340 94.535 -67.712 53.598
389341 94.572 -67.706 53.577
389342 94.609 -67.700 53.556
389343 94.650 -67.697 53.536
389344 94.690 -67.697 53.515
389345 94.723 -67.692 53.495
389346 94.751 -67.676 53.475
389347 94.779 -67.660 53.456
389348 94.807 -67.643 53.437
389349 94.834 -67.627 53.418
389350 94.862 -67.611 53.400
389351 94.890 -67.595 53.382
389352 94.917 -67.579 53.363
389353 94.947 -67.564 53.343
389354 94.987 -67.566 53.323
389355 95.027 -67.568 53.302
389356 95.068 -67.570 53.280
389357 95.108 -67.573 53.258
389358 95.149 -67.575 53.235
389359 95.189 -67.577 53.212
389360 95.230 -67.580 53.190
389361 95.270 -67.581 53.168
389362 95.311 -67.583 53.147
389363 95.351 -67.586 53.126
389364 95.392 -67.588 53.105
389365 95.429 -67.583 53.083
389366 95.469 -67.580 53.061
389367 95.517 -67.583 53.041
389368 95.557 -67.591 53.020
389369 95.597 -67.599 52.999
389370 95.638 -67.608 52.978
389371 95.679 -67.617 52.957
389372 95.720 -67.627 52.935
389373 95.760 -67.636 52.915
389374 95.801 -67.645 52.895
389375 95.842 -67.654 52.875
389376 95.883 -67.663 52.856
389377 95.923 -67.672 52.837
389378 95.962 -67.680 52.819
389379 95.989 -67.682 52.800
389380 96.016 -67.686 52.782
389381 96.040 -67.700 52.765
389382 96.074 -67.706 52.748
389383 96.107 -67.712 52.732
389384 96.141 -67.718 52.715
389385 96.174 -67.724 52.699
389386 96.207 -67.730 52.682
389387 96.241 -67.736 52.663
389388 96.274 -67.743 52.644
389389 96.308 -67.749 52.624
389390 96.341 -67.755 52.603
389391 96.375 -67.761 52.582
389392 96.408 -67.768 52.561
389393 96.440 -67.770 52.539
389394 96.473 -67.774 52.517
389395 96.506 -67.782 52.495
389396 96.539 -67.790 52.473
389397 96.572 -67.799 52.453
389398 96.605 -67.807 52.434
389399 96.639 -67.815 52.415
389400 96.674 -67.822 52.397
389401 96.709 -67.829 52.379
389402 96.740 -67.835 52.360
389403 96.770 -67.842 52.344
389404 96.804 -67.851 52.329
389405 96.837 -67.859 52.315
389406 96.870 -67.868 52.303
389407 96.903 -67.877 52.293
389408 96.937 -67.885 52.285
389409 96.970 -67.893 52.280
389410 97.003 -67.902 52.277
389411 97.036 -67.911 52.278
389412 97.069 -67.919 52.276
389413 97.102 -67.928 52.271
389414 97.142 -67.926 52.264
389415 97.187 -67.927 52.258
389416 97.234 -67.927 52.246
389417 97.273 -67.930 52.230
389418 97.306 -67.928 52.206
389419 97.339 -67.927 52.176
389420 97.373 -67.926 52.141
389421 97.407 -67.924 52.103
389422 97.441 -67.923 52.064
389423 97.475 -67.920 52.026
389424 97.509 -67.917 51.990
389425 97.543 -67.914 51.961
389426 97.578 -67.911 51.934
389427 97.612 -67.908 51.906
389428 97.646 -67.905 51.874
389429 97.675 -67.907 51.839
389430 97.711 -67.912 51.800
389431 97.741 -67.918 51.756
389432 97.773 -67.924 51.709
389433 97.807 -67.933 51.662
389434 97.841 -67.941 51.617
389435 97.875 -67.949 51.577
389436 97.909 -67.958 51.538
389437 97.943 -67.966 51.496
389438 97.977 -67.974 51.449
389439 98.011 -67.980 51.401
389440 98.042 -67.974 51.356
389441 98.073 -67.968 51.313
389442 98.105 -67.962 51.268
389443 98.138 -67.972 51.222
389444 98.171 -67.985 51.173
389445 98.205 -67.998 51.122
389446 98.238 -68.011 51.071
389447 98.272 -68.022 51.020
389448 98.305 -68.031 50.970
389449 98.339 -68.040 50.924
389450 98.373 -68.048 50.878
389451 98.407 -68.057 50.832
389452 98.441 -68.065 50.783
389453 98.475 -68.074 50.734
389454 98.509 -68.083 50.685
389455 98.543 -68.091 50.637
389456 98.576 -68.100 50.589
389457 98.612 -68.110 50.541
389458 98.647 -68.119 50.494
389459 98.683 -68.129 50.447
389460 98.718 -68.139 50.400
389461 98.754 -68.148 50.350
389462 98.795 -68.140 50.300
389463 98.837 -68.131 50.249
389464 98.878 -68.122 50.197
389465 98.919 -68.113 50.147
389466 98.961 -68.109 50.101
389467 99.000 -68.116 50.055
389468 99.036 -68.130 50.008
389469 99.071 -68.143 49.958
389470 99.107 -68.156 49.908
389471 99.142 -68.169 49.862
389472 99.178 -68.182 49.815
389473 99.214 -68.195 49.768
389474 99.249 -68.203 49.719
389475 99.282 -68.204 49.670
389476 99.313 -68.212 49.624
389477 99.344 -68.231 49.576
389478 99.373 -68.234 49.528
389479 99.403 -68.237 49.479
389480 99.432 -68.241 49.432
389481 99.467 -68.248 49.385
389482 99.501 -68.256 49.338
389483 99.535 -68.264 49.293
389484 99.569 -68.272 49.251
389485 99.615 -68.279 49.209
389486 99.658 -68.282 49.167
389487 99.703 -68.303 49.124
389488 99.752 -68.321 49.083
389489 99.806 -68.335 49.045
389490 99.837 -68.347 49.006
389491 99.864 -68.363 48.965
389492 99.893 -68.379 48.923
389493 99.933 -68.345 48.879
389494 99.976 -68.300 48.835
389495 100.012 -68.266 48.789
389496 100.039 -68.287 48.743
389497 100.071 -68.325 48.694
389498 100.104 -68.364 48.645
389499 100.136 -68.403 48.598
389500 100.168 -68.442 48.551
389501 100.200 -68.481 48.505
389502 100.229 -68.489 48.461
389503 100.259 -68.495 48.417
389504 100.296 -68.494 48.372
389505 100.335 -68.492 48.328
389506 100.375 -68.489 48.284
389507 100.415 -68.488 48.241
389946 90.144 -67.777 54.850
389947 90.179 -67.786 54.849
389948 90.214 -67.795 54.852
389949 90.248 -67.804 54.855
389950 90.283 -67.812 54.861
389951 90.317 -67.821 54.869
389952 90.352 -67.830 54.878
389953 90.386 -67.840 54.889
389954 90.421 -67.849 54.902
389955 90.453 -67.858 54.915
389956 90.484 -67.869 54.928
389957 90.525 -67.887 54.941
389958 90.593 -67.893 54.949
389959 90.660 -67.890 54.954
389960 90.726 -67.887 54.954
389961 90.793 -67.884 54.950
389962 90.860 -67.880 54.943
389963 90.927 -67.877 54.928
389964 90.994 -67.874 54.913
389965 91.053 -67.868 54.897
389966 91.103 -67.858 54.881
389967 91.153 -67.848 54.866
389968 91.203 -67.838 54.851
389969 91.253 -67.827 54.836
389970 91.303 -67.818 54.823
389971 91.353 -67.807 54.811
389972 91.403 -67.797 54.799
389973 91.453 -67.786 54.791
389974 91.503 -67.777 54.784
389975 91.553 -67.766 54.778
389976 91.599 -67.758 54.775
389977 91.640 -67.750 54.774
389978 91.681 -67.743 54.771
389979 91.722 -67.735 54.767
389980 91.764 -67.727 54.762
389981 91.805 -67.720 54.754
389982 91.846 -67.724 54.745
389983 91.886 -67.733 54.735
389984 91.926 -67.743 54.723
389985 91.967 -67.752 54.708
389986 92.007 -67.761 54.692
389987 92.048 -67.771 54.674
389988 92.088 -67.780 54.656
389989 92.129 -67.790 54.638
389990 92.181 -67.793 54.617
389991 92.235 -67.796 54.594
389992 92.289 -67.799 54.570
389993 92.343 -67.802 54.547
389994 92.397 -67.805 54.525
389995 92.451 -67.808 54.504
389996 92.504 -67.811 54.484
389997 92.558 -67.813 54.463
389998 92.612 -67.816 54.441
389999 92.669 -67.809 54.419
390000 92.727 -67.799 54.397
390001 92.786 -67.788 54.376
390002 92.844 -67.777 54.355
390003 92.902 -67.766 54.338
390004 92.959 -67.756 54.325
390005 92.994 -67.757 54.315
390006 93.028 -67.758 54.308
390007 93.063 -67.760 54.303
390008 93.098 -67.761 54.297
390009 93.133 -67.762 54.289
390010 93.168 -67.764 54.277
390011 93.218 -67.759 54.260
390012 93.271 -67.755 54.236
390013 93.325 -67.749 54.210
390014 93.378 -67.744 54.186
390015 93.431 -67.739 54.164
390016 93.481 -67.735 54.145
390017 93.528 -67.731 54.130
390018 93.574 -67.727 54.118
390019 93.620 -67.723 54.106
390020 93.666 -67.719 54.094
390021 93.712 -67.715 54.077
390022 93.758 -67.711 54.057
390023 93.804 -67.711 54.036
390024 93.851 -67.716 54.014
390025 93.898 -67.721 53.991
390026 93.944 -67.726 53.971
390027 93.991 -67.731 53.953
390028 94.037 -67.736 53.936
390029 94.078 -67.738 53.919
390030 94.115 -67.732 53.901
390031 94.152 -67.725 53.881
390032 94.189 -67.719 53.860
390033 94.225 -67.713 53.838
390034 94.263 -67.707 53.815
390035 94.299 -67.700 53.785
390036 94.336 -67.694 53.750
390037 94.373 -67.688 53.714
390038 94.410 -67.682 53.680
390039 94.447 -67.675 53.652
390040 94.484 -67.669 53.625
390041 94.521 -67.663 53.603
390042 94.557 -67.657 53.583
390043 94.597 -67.652 53.562
390044 94.647 -67.658 53.541
390045 94.682 -67.655 53.521
390046 94.717 -67.654 53.502
390047 94.751 -67.648 53.482
390048 94.779 -67.632 53.461
390049 94.807 -67.616 53.442
390050 94.834 -67.599 53.422
390051 94.862 -67.583 53.404
390052 94.890 -67.567 53.386
390053 94.917 -67.551 53.367
390054 94.949 -67.539 53.347
390055 94.989 -67.542 53.326
390056 95.030 -67.544 53.304
390057 95.070 -67.546 53.281
390058 95.111 -67.548 53.259
390059 95.151 -67.551 53.236
390060 95.192 -67.553 53.213
390061 95.232 -67.555 53.191
390062 95.272 -67.557 53.169
390063 95.313 -67.559 53.148
390064 95.351 -67.556 53.126
390065 95.389 -67.551 53.105
390066 95.426 -67.545 53.083
390067 95.473 -67.547 53.062
390068 95.520 -67.550 53.041
390069 95.563 -67.556 53.020
390070 95.603 -67.564 52.999
390071 95.642 -67.573 52.978
390072 95.683 -67.583 52.957
390073 95.724 -67.592 52.936
390074 95.764 -67.601 52.915
390075 95.805 -67.610 52.895
390076 95.846 -67.619 52.874
390077 95.887 -67.628 52.855
390078 95.927 -67.637 52.836
390079 95.966 -67.645 52.816
390080 95.994 -67.648 52.797
390081 96.020 -67.655 52.779
390082 96.044 -67.668 52.761
390083 96.077 -67.675 52.743
390084 96.110 -67.681 52.725
390085 96.144 -67.687 52.708
390086 96.177 -67.693 52.691
390087 96.210 -67.699 52.672
390088 96.244 -67.706 52.653
390089 96.277 -67.712 52.633
390090 96.311 -67.718 52.614
390091 96.344 -67.724 52.593
390092 96.377 -67.728 52.571
390093 96.409 -67.730 52.550
390094 96.441 -67.733 52.528
390095 96.474 -67.736 52.506
390096 96.507 -67.743 52.483
390097 96.540 -67.751 52.461
390098 96.573 -67.760 52.441
390099 96.606 -67.768 52.422
390100 96.640 -67.777 52.402
390101 96.672 -67.785 52.383
390102 96.705 -67.793 52.365
390103 96.739 -67.800 52.348
390104 96.772 -67.809 52.331
390105 96.805 -67.819 52.316
390106 96.838 -67.830 52.303
390107 96.872 -67.840 52.291
390108 96.905 -67.849 52.282
390109 96.939 -67.858 52.274
390110 96.972 -67.867 52.268
390111 97.005 -67.876 52.263
390112 97.038 -67.884 52.261
390113 97.071 -67.893 52.258
390114 97.104 -67.901 52.253
390115 97.149 -67.901 52.249
390116 97.195 -67.902 52.243
390117 97.241 -67.898 52.231
390118 97.283 -67.896 52.214
390119 97.317 -67.893 52.190
390120 97.352 -67.890 52.160
390121 97.386 -67.887 52.125
390122 97.420 -67.884 52.088
390123 97.454 -67.881 52.049
390124 97.489 -67.878 52.010
390125 97.523 -67.875 51.975
390126 97.557 -67.872 51.945
390127 97.591 -67.869 51.918
390128 97.625 -67.866 51.891
390129 97.654 -67.868 51.860
390130 97.682 -67.871 51.825
390131 97.725 -67.875 51.785
390132 97.755 -67.880 51.741
390133 97.786 -67.886 51.693
390134 97.816 -67.891 51.644
390135 97.847 -67.896 51.600
390136 97.879 -67.904 51.560
390137 97.914 -67.913 51.520
390138 97.947 -67.921 51.479
390139 97.982 -67.930 51.432
390140 98.016 -67.938 51.385
390141 98.048 -67.937 51.342
390142 98.079 -67.931 51.299
390143 98.111 -67.931 51.256
390144 98.145 -67.943 51.211
390145 98.178 -67.956 51.162
390146 98.211 -67.968 51.110
390147 98.245 -67.977 51.059
390148 98.279 -67.986 51.008
390149 98.313 -67.994 50.958
390150 98.347 -68.003 50.912
390151 98.381 -68.012 50.867
390152 98.415 -68.021 50.822
390153 98.448 -68.029 50.774
390154 98.482 -68.037 50.726
390155 98.516 -68.046 50.678
390156 98.550 -68.055 50.630
390157 98.584 -68.064 50.584
390158 98.620 -68.074 50.536
390159 98.655 -68.084 50.488
390160 98.691 -68.094 50.441
390161 98.726 -68.103 50.394
390162 98.762 -68.112 50.346
390163 98.800 -68.113 50.296
390164 98.842 -68.104 50.244
390165 98.883 -68.096 50.192
390166 98.924 -68.087 50.139
390167 98.967 -68.084 50.089
390168 99.008 -68.089 50.042
390169 99.044 -68.102 49.994
390170 99.079 -68.115 49.942
390171 99.115 -68.128 49.891
390172 99.150 -68.141 49.844
390173 99.186 -68.154 49.799
390174 99.220 -68.159 49.754
390175 99.254 -68.160 49.708
390176 99.287 -68.161 49.660
390177 99.320 -68.162 49.613
390178 99.356 -68.185 49.565
390179 99.380 -68.206 49.517
390180 99.410 -68.209 49.470
390181 99.439 -68.212 49.422
390182 99.468 -68.215 49.375
390183 99.497 -68.218 49.329
390184 99.526 -68.221 49.286
390185 99.559 -68.228 49.245
390186 99.604 -68.240 49.203
390187 99.659 -68.253 49.160
390188 99.713 -68.268 49.116
390189 99.768 -68.282 49.074
390190 99.812 -68.293 49.034
390191 99.845 -68.303 48.994
390192 99.872 -68.319 48.953
390193 99.900 -68.335 48.911
390194 99.941 -68.298 48.868
390195 99.985 -68.253 48.823
390196 100.010 -68.234 48.777
390197 100.042 -68.271 48.731
390198 100.074 -68.310 48.684
390199 100.106 -68.348 48.637
390200 100.138 -68.388 48.590
390201 100.170 -68.426 48.543
390202 100.200 -68.445 48.498
390203 100.229 -68.449 48.454
390204 100.265 -68.450 48.410
390205 100.305 -68.448 48.366
390206 100.344 -68.445 48.322
390207 100.384 -68.444 48.279
390208 100.424 -68.441 48.236
390646 90.095 -67.730 54.858
390647 90.129 -67.740 54.857
390648 90.163 -67.750 54.858
390649 90.197 -67.759 54.861
390650 90.232 -67.769 54.866
390651 90.266 -67.779 54.873
390652 90.300 -67.788 54.881
390653 90.334 -67.798 54.891
390654 90.369 -67.808 54.903
390655 90.403 -67.817 54.916
390656 90.437 -67.827 54.930
390657 90.471 -67.836 54.945
390658 90.524 -67.838 54.957
390659 90.591 -67.835 54.965
390660 90.658 -67.832 54.967
390661 90.724 -67.829 54.965
390662 90.791 -67.826 54.959
390663 90.858 -67.823 54.949
390664 90.925 -67.821 54.934
390665 90.992 -67.818 54.918
390666 91.058 -67.815 54.903
390667 91.109 -67.805 54.889
390668 91.159 -67.794 54.874
390669 91.209 -67.784 54.859
390670 91.259 -67.774 54.846
390671 91.309 -67.764 54.835
390672 91.359 -67.753 54.825
390673 91.409 -67.743 54.816
390674 91.459 -67.733 54.809
390675 91.509 -67.723 54.803
390676 91.555 -67.714 54.798
390677 91.596 -67.706 54.795
390678 91.637 -67.699 54.794
390679 91.678 -67.691 54.790
390680 91.720 -67.683 54.785
390681 91.761 -67.676 54.778
390682 91.802 -67.668 54.769
390683 91.843 -67.665 54.759
390684 91.883 -67.675 54.747
390685 91.924 -67.684 54.734
390686 91.964 -67.694 54.717
390687 92.005 -67.703 54.700
390688 92.046 -67.713 54.682
390689 92.086 -67.722 54.663
390690 92.140 -67.725 54.643
390691 92.194 -67.728 54.622
390692 92.247 -67.730 54.598
390693 92.301 -67.733 54.574
390694 92.355 -67.736 54.551
390695 92.409 -67.739 54.529
390696 92.463 -67.743 54.508
390697 92.517 -67.746 54.487
390698 92.571 -67.748 54.466
390699 92.625 -67.751 54.443
390700 92.678 -67.754 54.420
390701 92.735 -67.749 54.398
390702 92.793 -67.738 54.377
390703 92.851 -67.727 54.358
390704 92.910 -67.716 54.341
390705 92.962 -67.708 54.329
390706 92.997 -67.709 54.321
390707 93.031 -67.711 54.316
390708 93.066 -67.712 54.311
390709 93.101 -67.713 54.305
390710 93.136 -67.714 54.296
390711 93.172 -67.712 54.284
390712 93.221 -67.708 54.266
390713 93.275 -67.702 54.243
390714 93.329 -67.697 54.217
390715 93.382 -67.692 54.194
390716 93.436 -67.687 54.172
390717 93.476 -67.686 54.152
390718 93.512 -67.681 54.136
390719 93.558 -67.677 54.125
390720 93.605 -67.674 54.114
390721 93.650 -67.670 54.102
390722 93.697 -67.665 54.086
390723 93.743 -67.661 54.066
390724 93.789 -67.658 54.044
390725 93.835 -67.663 54.021
390726 93.882 -67.668 53.997
390727 93.928 -67.673 53.976
390728 93.975 -67.678 53.957
390729 94.021 -67.683 53.938
390730 94.063 -67.687 53.921
390731 94.101 -67.682 53.902
390732 94.137 -67.676 53.882
390733 94.175 -67.670 53.861
390734 94.211 -67.664 53.839
390735 94.248 -67.657 53.816
390736 94.285 -67.651 53.787
390737 94.322 -67.645 53.754
390738 94.359 -67.639 53.720
390739 94.396 -67.632 53.687
390740 94.432 -67.626 53.657
390741 94.469 -67.620 53.630
390742 94.506 -67.614 53.607
390743 94.544 -67.608 53.587
390744 94.594 -67.613 53.567
390745 94.639 -67.616 53.547
390746 94.675 -67.614 53.529
390747 94.710 -67.612 53.511
390748 94.746 -67.610 53.493
390749 94.779 -67.605 53.472
390750 94.807 -67.588 53.451
390751 94.834 -67.572 53.431
390752 94.862 -67.555 53.413
390753 94.890 -67.539 53.395
390754 94.917 -67.523 53.376
390755 94.951 -67.515 53.354
390756 94.992 -67.517 53.331
390757 95.032 -67.520 53.308
390758 95.072 -67.522 53.286
390759 95.113 -67.524 53.262
390760 95.154 -67.526 53.238
390761 95.194 -67.529 53.215
390762 95.235 -67.531 53.192
390763 95.273 -67.529 53.170
390764 95.311 -67.523 53.149
390765 95.349 -67.518 53.127
390766 95.386 -67.513 53.106
390767 95.429 -67.511 53.085
390768 95.476 -67.514 53.064
390769 95.523 -67.517 53.043
390770 95.569 -67.521 53.022
390771 95.608 -67.530 53.001
390772 95.648 -67.539 52.979
390773 95.687 -67.548 52.958
390774 95.728 -67.557 52.937
390775 95.768 -67.566 52.916
390776 95.809 -67.575 52.896
390777 95.850 -67.584 52.875
390778 95.891 -67.593 52.856
390779 95.932 -67.602 52.836
390780 95.971 -67.611 52.816
390781 95.999 -67.614 52.796
390782 96.025 -67.623 52.776
390783 96.049 -67.636 52.758
390784 96.080 -67.645 52.739
390785 96.113 -67.651 52.720
390786 96.147 -67.657 52.701
390787 96.180 -67.663 52.683
390788 96.214 -67.669 52.663
390789 96.247 -67.675 52.643
390790 96.281 -67.681 52.623
390791 96.313 -67.686 52.604
390792 96.345 -67.688 52.583
390793 96.377 -67.690 52.561
390794 96.409 -67.692 52.540
390795 96.442 -67.695 52.517
390796 96.475 -67.699 52.495
390797 96.508 -67.703 52.472
390798 96.541 -67.712 52.450
390799 96.574 -67.721 52.430
390800 96.607 -67.729 52.410
390801 96.641 -67.737 52.390
390802 96.673 -67.746 52.371
390803 96.707 -67.754 52.353
390804 96.740 -67.758 52.336
390805 96.773 -67.763 52.320
390806 96.806 -67.773 52.305
390807 96.840 -67.783 52.291
390808 96.873 -67.793 52.279
390809 96.906 -67.804 52.270
390810 96.940 -67.814 52.262
390811 96.973 -67.824 52.254
390812 97.006 -67.834 52.248
390813 97.039 -67.844 52.244
390814 97.073 -67.854 52.240
390815 97.111 -67.860 52.235
390816 97.157 -67.858 52.229
390817 97.203 -67.853 52.222
390818 97.249 -67.849 52.209
390819 97.296 -67.855 52.191
390820 97.331 -67.852 52.168
390821 97.365 -67.849 52.138
390822 97.399 -67.846 52.104
390823 97.434 -67.843 52.068
390824 97.468 -67.840 52.030
390825 97.502 -67.837 51.992
390826 97.536 -67.833 51.957
390827 97.570 -67.830 51.928
390828 97.604 -67.827 51.901
390829 97.634 -67.830 51.872
390830 97.662 -67.832 51.842
390831 97.690 -67.834 51.808
390832 97.737 -67.838 51.769
390833 97.770 -67.843 51.725
390834 97.800 -67.849 51.678
390835 97.830 -67.854 51.629
390836 97.861 -67.860 51.585
390837 97.891 -67.865 51.544
390838 97.922 -67.871 51.503
390839 97.952 -67.876 51.460
390840 97.986 -67.884 51.415
390841 98.020 -67.893 51.371
390842 98.054 -67.900 51.329
390843 98.085 -67.895 51.288
390844 98.118 -67.901 51.245
390845 98.151 -67.914 51.201
390846 98.185 -67.924 51.151
390847 98.219 -67.932 51.099
390848 98.253 -67.940 51.048
390849 98.287 -67.949 50.997
390850 98.320 -67.958 50.946
390851 98.354 -67.967 50.899
390852 98.388 -67.975 50.853
390853 98.422 -67.984 50.807
390854 98.456 -67.993 50.760
390855 98.490 -68.001 50.712
390856 98.524 -68.010 50.665
390857 98.558 -68.018 50.618
390858 98.592 -68.028 50.573
390859 98.628 -68.037 50.526
390860 98.663 -68.047 50.478
390861 98.699 -68.057 50.430
390862 98.734 -68.067 50.382
390863 98.770 -68.077 50.335
390864 98.805 -68.087 50.287
390865 98.847 -68.078 50.237
390866 98.887 -68.069 50.186
390867 98.929 -68.060 50.133
390868 98.972 -68.060 50.080
390869 99.015 -68.062 50.030
390870 99.051 -68.075 49.979
390871 99.087 -68.087 49.927
390872 99.122 -68.100 49.876
390873 99.158 -68.114 49.829
390874 99.192 -68.115 49.784
390875 99.225 -68.115 49.740
390876 99.259 -68.116 49.695
390877 99.292 -68.117 49.649
390878 99.326 -68.118 49.603
390879 99.353 -68.135 49.557
390880 99.403 -68.159 49.510
390881 99.417 -68.181 49.462
390882 99.446 -68.184 49.413
390883 99.475 -68.188 49.366
390884 99.504 -68.191 49.321
390885 99.533 -68.194 49.280
390886 99.570 -68.193 49.240
390887 99.621 -68.200 49.197
390888 99.675 -68.214 49.153
390889 99.730 -68.228 49.109
390890 99.784 -68.241 49.065
390891 99.819 -68.251 49.023
390892 99.851 -68.260 48.982
390893 99.880 -68.274 48.941
390894 99.908 -68.290 48.899
390895 99.950 -68.250 48.857
390896 99.989 -68.212 48.812
390897 100.012 -68.216 48.767
390898 100.044 -68.255 48.721
390899 100.076 -68.294 48.675
390900 100.108 -68.333 48.628
390901 100.140 -68.372 48.582
390902 100.171 -68.400 48.536
390903 100.201 -68.404 48.491
390904 100.235 -68.406 48.448
390905 100.274 -68.404 48.403
390906 100.314 -68.401 48.359
390907 100.354 -68.400 48.316
390908 100.393 -68.398 48.273
390909 100.433 -68.395 48.230
391347 90.087 -67.683 54.864
391348 90.121 -67.692 54.863
391349 90.155 -67.702 54.866
391350 90.189 -67.711 54.870
391351 90.224 -67.721 54.877
391352 90.258 -67.730 54.885
391353 90.292 -67.739 54.894
391354 90.326 -67.749 54.905
391355 90.361 -67.759 54.916
391356 90.395 -67.768 54.929
391357 90.429 -67.778 54.945
391358 90.470 -67.782 54.960
391359 90.530 -67.776 54.972
391360 90.591 -67.777 54.978
391361 90.656 -67.775 54.979
391362 90.722 -67.772 54.975
391363 90.789 -67.769 54.967
391364 90.856 -67.766 54.955
391365 90.923 -67.763 54.940
391366 90.990 -67.760 54.925
391367 91.056 -67.757 54.910
391368 91.115 -67.751 54.897
391369 91.165 -67.740 54.883
391370 91.215 -67.730 54.871
391371 91.265 -67.720 54.860
391372 91.315 -67.710 54.851
391373 91.365 -67.699 54.845
391374 91.415 -67.689 54.839
391375 91.465 -67.679 54.833
391376 91.511 -67.670 54.827
391377 91.552 -67.663 54.821
391378 91.593 -67.655 54.818
391379 91.635 -67.648 54.814
391380 91.676 -67.640 54.808
391381 91.717 -67.633 54.801
391382 91.758 -67.625 54.793
391383 91.799 -67.617 54.782
391384 91.841 -67.610 54.771
391385 91.881 -67.617 54.759
391386 91.922 -67.627 54.744
391387 91.962 -67.636 54.727
391388 92.003 -67.645 54.709
391389 92.044 -67.654 54.690
391390 92.098 -67.657 54.669
391391 92.152 -67.660 54.649
391392 92.206 -67.663 54.626
391393 92.260 -67.665 54.603
391394 92.314 -67.668 54.579
391395 92.368 -67.671 54.556
391396 92.422 -67.674 54.534
391397 92.475 -67.677 54.512
391398 92.529 -67.680 54.491
391399 92.583 -67.683 54.470
391400 92.637 -67.686 54.446
391401 92.691 -67.689 54.422
391402 92.744 -67.692 54.399
391403 92.800 -67.687 54.378
391404 92.859 -67.677 54.359
391405 92.917 -67.666 54.343
391406 92.965 -67.661 54.332
391407 93.000 -67.661 54.325
391408 93.035 -67.663 54.321
391409 93.069 -67.664 54.316
391410 93.105 -67.663 54.308
391411 93.141 -67.661 54.299
391412 93.177 -67.660 54.285
391413 93.226 -67.655 54.267
391414 93.280 -67.650 54.244
391415 93.334 -67.645 54.222
391416 93.387 -67.639 54.199
391417 93.439 -67.635 54.178
391418 93.476 -67.639 54.158
391419 93.507 -67.633 54.143
391420 93.543 -67.628 54.132
391421 93.589 -67.623 54.123
391422 93.635 -67.620 54.112
391423 93.681 -67.616 54.095
391424 93.727 -67.612 54.075
391425 93.773 -67.608 54.052
391426 93.820 -67.610 54.027
391427 93.866 -67.615 54.002
391428 93.913 -67.620 53.980
391429 93.960 -67.625 53.960
391430 94.005 -67.630 53.940
391431 94.047 -67.634 53.920
391432 94.086 -67.633 53.902
391433 94.123 -67.627 53.881
391434 94.160 -67.620 53.861
391435 94.197 -67.614 53.839
391436 94.234 -67.608 53.815
391437 94.270 -67.602 53.787
391438 94.308 -67.595 53.757
391439 94.344 -67.589 53.725
391440 94.381 -67.583 53.694
391441 94.418 -67.576 53.663
391442 94.455 -67.570 53.636
391443 94.492 -67.564 53.612
391444 94.541 -67.568 53.591
391445 94.591 -67.573 53.571
391446 94.632 -67.574 53.552
391447 94.668 -67.572 53.535
391448 94.703 -67.570 53.519
391449 94.738 -67.568 53.501
391450 94.773 -67.566 53.482
391451 94.807 -67.561 53.461
391452 94.834 -67.544 53.443
391453 94.862 -67.528 53.426
391454 94.890 -67.512 53.409
391455 94.918 -67.495 53.387
391456 94.954 -67.491 53.363
391457 94.994 -67.493 53.338
391458 95.035 -67.495 53.313
391459 95.075 -67.498 53.289
391460 95.115 -67.500 53.264
391461 95.156 -67.502 53.240
391462 95.195 -67.502 53.217
391463 95.233 -67.496 53.194
391464 95.271 -67.491 53.172
391465 95.308 -67.486 53.150
391466 95.346 -67.480 53.129
391467 95.385 -67.476 53.108
391468 95.432 -67.479 53.087
391469 95.479 -67.482 53.066
391470 95.527 -67.484 53.045
391471 95.574 -67.487 53.025
391472 95.614 -67.495 53.004
391473 95.654 -67.504 52.984
391474 95.693 -67.513 52.963
391475 95.733 -67.522 52.941
391476 95.772 -67.531 52.920
391477 95.813 -67.540 52.900
391478 95.854 -67.549 52.879
391479 95.895 -67.558 52.859
391480 95.936 -67.567 52.838
391481 95.976 -67.576 52.817
391482 96.004 -67.579 52.797
391483 96.029 -67.591 52.776
391484 96.053 -67.605 52.756
391485 96.083 -67.614 52.737
391486 96.116 -67.620 52.717
391487 96.150 -67.626 52.696
391488 96.183 -67.632 52.676
391489 96.217 -67.638 52.655
391490 96.250 -67.644 52.634
391491 96.282 -67.646 52.615
391492 96.314 -67.648 52.594
391493 96.346 -67.651 52.573
391494 96.378 -67.653 52.551
391495 96.410 -67.655 52.530
391496 96.443 -67.658 52.506
391497 96.476 -67.661 52.484
391498 96.509 -67.664 52.462
391499 96.542 -67.673 52.440
391500 96.575 -67.681 52.420
391501 96.608 -67.689 52.400
391502 96.642 -67.698 52.380
391503 96.674 -67.707 52.361
391504 96.708 -67.712 52.343
391505 96.741 -67.715 52.326
391506 96.774 -67.718 52.309
391507 96.808 -67.727 52.293
391508 96.841 -67.737 52.280
391509 96.874 -67.747 52.268
391510 96.907 -67.758 52.258
391511 96.941 -67.768 52.249
391512 96.974 -67.777 52.240
391513 97.007 -67.788 52.233
391514 97.044 -67.796 52.227
391515 97.083 -67.802 52.222
391516 97.122 -67.808 52.216
391517 97.164 -67.808 52.207
391518 97.211 -67.805 52.197
391519 97.256 -67.802 52.183
391520 97.304 -67.809 52.164
391521 97.345 -67.811 52.141
391522 97.379 -67.808 52.112
391523 97.413 -67.805 52.079
391524 97.447 -67.802 52.043
391525 97.481 -67.799 52.007
391526 97.515 -67.796 51.971
391527 97.549 -67.793 51.938
391528 97.584 -67.790 51.909
391529 97.613 -67.791 51.881
391530 97.642 -67.793 51.852
391531 97.670 -67.796 51.822
391532 97.699 -67.798 51.788
391533 97.743 -67.802 51.751
391534 97.785 -67.806 51.709
391535 97.815 -67.811 51.662
391536 97.845 -67.817 51.615
391537 97.875 -67.823 51.570
391538 97.906 -67.828 51.529
391539 97.936 -67.833 51.487
391540 97.966 -67.839 51.443
391541 97.997 -67.844 51.400
391542 98.027 -67.849 51.358
391543 98.058 -67.856 51.317
391544 98.091 -67.859 51.277
391545 98.125 -67.870 51.235
391546 98.159 -67.878 51.190
391547 98.193 -67.887 51.140
391548 98.226 -67.896 51.089
391549 98.261 -67.904 51.037
391550 98.294 -67.913 50.985
391551 98.328 -67.921 50.935
391552 98.362 -67.931 50.886
391553 98.396 -67.939 50.838
391554 98.430 -67.948 50.791
391555 98.463 -67.956 50.745
391556 98.497 -67.965 50.698
391557 98.531 -67.974 50.651
391558 98.565 -67.982 50.605
391559 98.600 -67.992 50.560
391560 98.636 -68.002 50.513
391561 98.671 -68.012 50.465
391562 98.707 -68.021 50.417
391563 98.742 -68.031 50.369
391564 98.778 -68.040 50.322
391565 98.813 -68.050 50.276
391566 98.851 -68.052 50.229
391567 98.892 -68.043 50.179
391568 98.934 -68.034 50.128
391569 98.977 -68.037 50.074
391570 99.020 -68.038 50.021
391571 99.059 -68.046 49.968
391572 99.094 -68.060 49.916
391573 99.130 -68.070 49.865
391574 99.163 -68.071 49.816
391575 99.197 -68.072 49.769
391576 99.230 -68.072 49.725
391577 99.264 -68.073 49.682
391578 99.297 -68.074 49.637
391579 99.331 -68.075 49.593
391580 99.360 -68.085 49.548
391581 99.392 -68.109 49.503
391582 99.449 -68.134 49.455
391583 99.453 -68.156 49.405
391584 99.482 -68.159 49.357
391585 99.511 -68.162 49.314
391586 99.546 -68.162 49.275
391587 99.589 -68.158 49.234
391588 99.637 -68.160 49.191
391589 99.692 -68.174 49.147
391590 99.746 -68.188 49.102
391591 99.792 -68.200 49.058
391592 99.825 -68.209 49.014
391593 99.857 -68.218 48.971
391594 99.888 -68.229 48.929
391595 99.916 -68.245 48.888
391596 99.959 -68.203 48.846
391597 99.987 -68.180 48.801
391598 100.014 -68.200 48.757
391599 100.046 -68.239 48.712
391600 100.078 -68.278 48.665
391601 100.111 -68.317 48.620
391602 100.142 -68.355 48.575
391603 100.171 -68.360 48.529
391604 100.204 -68.362 48.484
391605 100.244 -68.360 48.441
391606 100.284 -68.357 48.397
391607 100.323 -68.356 48.353
391608 100.363 -68.354 48.310
391609 100.403 -68.351 48.267
391610 100.442 -68.350 48.224
392047 90.044 -67.624 54.872
392048 90.079 -67.634 54.871
392049 90.113 -67.643 54.872
392050 90.147 -67.653 54.876
392051 90.181 -67.663 54.881
392052 90.216 -67.672 54.889
392053 90.250 -67.682 54.897
392054 90.284 -67.692 54.907
392055 90.318 -67.701 54.918
392056 90.353 -67.711 54.930
392057 90.387 -67.720 54.943
392058 90.425 -67.728 54.958
392059 90.478 -67.717 54.973
392060 90.538 -67.712 54.984
392061 90.599 -67.713 54.990
392062 90.660 -67.713 54.989
392063 90.722 -67.714 54.983
392064 90.787 -67.711 54.973
392065 90.854 -67.708 54.960
392066 90.921 -67.705 54.946
392067 90.988 -67.703 54.932
392068 91.054 -67.700 54.918
392069 91.121 -67.697 54.906
392070 91.171 -67.686 54.894
392071 91.221 -67.677 54.884
392072 91.271 -67.666 54.875
392073 91.321 -67.656 54.869
392074 91.371 -67.645 54.866
392075 91.421 -67.636 54.863
392076 91.467 -67.627 54.858
392077 91.508 -67.619 54.851
392078 91.549 -67.611 54.844
392079 91.591 -67.604 54.840
392080 91.632 -67.596 54.835
392081 91.673 -67.589 54.827
392082 91.714 -67.581 54.818
392083 91.755 -67.573 54.808
392084 91.797 -67.566 54.796
392085 91.838 -67.559 54.783
392086 91.879 -67.559 54.770
392087 91.919 -67.568 54.754
392088 91.960 -67.578 54.736
392089 92.003 -67.586 54.718
392090 92.057 -67.589 54.697
392091 92.111 -67.592 54.676
392092 92.165 -67.595 54.655
392093 92.218 -67.598 54.633
392094 92.272 -67.601 54.609
392095 92.326 -67.603 54.585
392096 92.380 -67.606 54.562
392097 92.434 -67.609 54.539
392098 92.488 -67.612 54.518
392099 92.542 -67.615 54.497
392100 92.595 -67.618 54.474
392101 92.649 -67.620 54.448
392102 92.703 -67.623 54.424
392103 92.757 -67.627 54.401
392104 92.810 -67.630 54.379
392105 92.866 -67.627 54.360
392106 92.925 -67.616 54.345
392107 92.968 -67.613 54.335
392108 93.002 -67.614 54.328
392109 93.038 -67.614 54.323
392110 93.074 -67.612 54.318
392111 93.110 -67.610 54.309
392112 93.146 -67.608 54.300
392113 93.182 -67.607 54.286
392114 93.231 -67.602 54.268
392115 93.285 -67.597 54.246
392116 93.339 -67.592 54.225
392117 93.392 -67.586 54.203
392118 93.439 -67.586 54.182
392119 93.476 -67.590 54.163
392120 93.508 -67.586 54.149
392121 93.539 -67.580 54.138
392122 93.573 -67.574 54.129
392123 93.620 -67.570 54.119
392124 93.666 -67.567 54.103
392125 93.712 -67.562 54.082
392126 93.758 -67.558 54.059
392127 93.804 -67.558 54.034
392128 93.851 -67.562 54.008
392129 93.897 -67.567 53.985
392130 93.944 -67.572 53.964
392131 93.989 -67.576 53.943
392132 94.031 -67.581 53.923
392133 94.072 -67.583 53.902
392134 94.109 -67.577 53.882
392135 94.145 -67.570 53.860
392136 94.182 -67.564 53.837
392137 94.219 -67.558 53.813
392138 94.256 -67.552 53.787
392139 94.293 -67.545 53.759
392140 94.330 -67.539 53.731
392141 94.366 -67.533 53.701
392142 94.403 -67.527 53.670
392143 94.440 -67.520 53.641
392144 94.488 -67.523 53.617
392145 94.538 -67.529 53.595
392146 94.588 -67.533 53.575
392147 94.625 -67.533 53.557
392148 94.660 -67.530 53.541
392149 94.695 -67.528 53.525
392150 94.731 -67.526 53.507
392151 94.766 -67.524 53.489
392152 94.801 -67.523 53.472
392153 94.834 -67.517 53.457
392154 94.862 -67.501 53.443
392155 94.890 -67.484 53.426
392156 94.918 -67.468 53.401
392157 94.956 -67.467 53.373
392158 94.996 -67.469 53.345
392159 95.037 -67.471 53.318
392160 95.077 -67.473 53.291
392161 95.118 -67.475 53.267
392162 95.155 -67.470 53.243
392163 95.193 -67.464 53.219
392164 95.231 -67.458 53.196
392165 95.268 -67.453 53.174
392166 95.306 -67.448 53.152
392167 95.343 -67.442 53.132
392168 95.388 -67.443 53.111
392169 95.436 -67.446 53.090
392170 95.483 -67.448 53.069
392171 95.530 -67.451 53.049
392172 95.578 -67.454 53.028
392173 95.620 -67.460 53.008
392174 95.660 -67.470 52.988
392175 95.699 -67.478 52.968
392176 95.739 -67.487 52.947
392177 95.778 -67.496 52.926
392178 95.817 -67.505 52.906
392179 95.858 -67.514 52.884
392180 95.899 -67.523 52.864
392181 95.940 -67.532 52.842
392182 95.980 -67.541 52.821
392183 96.009 -67.545 52.800
392184 96.033 -67.559 52.778
392185 96.057 -67.573 52.757
392186 96.086 -67.583 52.736
392187 96.119 -67.589 52.715
392188 96.153 -67.595 52.693
392189 96.186 -67.602 52.671
392190 96.219 -67.604 52.650
392191 96.251 -67.607 52.628
392192 96.283 -67.609 52.607
392193 96.315 -67.611 52.586
392194 96.347 -67.614 52.564
392195 96.379 -67.616 52.543
392196 96.411 -67.618 52.521
392197 96.444 -67.621 52.499
392198 96.477 -67.624 52.476
392199 96.510 -67.627 52.454
392200 96.543 -67.633 52.433
392201 96.576 -67.642 52.413
392202 96.609 -67.651 52.392
392203 96.643 -67.659 52.372
392204 96.675 -67.667 52.352
392205 96.709 -67.670 52.334
392206 96.742 -67.673 52.316
392207 96.776 -67.676 52.299
392208 96.809 -67.681 52.283
392209 96.842 -67.691 52.270
392210 96.875 -67.701 52.256
392211 96.908 -67.711 52.245
392212 96.942 -67.721 52.236
392213 96.976 -67.730 52.226
392214 97.015 -67.736 52.217
392215 97.054 -67.743 52.210
392216 97.093 -67.749 52.202
392217 97.133 -67.755 52.192
392218 97.172 -67.760 52.181
392219 97.218 -67.756 52.168
392220 97.264 -67.757 52.151
392221 97.311 -67.763 52.132
392222 97.358 -67.769 52.109
392223 97.392 -67.766 52.080
392224 97.426 -67.763 52.048
392225 97.460 -67.760 52.014
392226 97.495 -67.757 51.979
392227 97.529 -67.754 51.946
392228 97.563 -67.751 51.915
392229 97.593 -67.752 51.887
392230 97.621 -67.755 51.858
392231 97.650 -67.757 51.829
392232 97.678 -67.759 51.798
392233 97.707 -67.761 51.766
392234 97.750 -67.765 51.731
392235 97.799 -67.769 51.690
392236 97.829 -67.774 51.646
392237 97.860 -67.780 51.601
392238 97.890 -67.786 51.555
392239 97.920 -67.791 51.512
392240 97.950 -67.796 51.470
392241 97.981 -67.802 51.430
392242 98.011 -67.807 51.389
392243 98.041 -67.813 51.349
392244 98.072 -67.818 51.308
392245 98.099 -67.824 51.267
392246 98.133 -67.833 51.224
392247 98.166 -67.842 51.179
392248 98.200 -67.851 51.129
392249 98.234 -67.859 51.078
392250 98.268 -67.868 51.027
392251 98.302 -67.877 50.975
392252 98.336 -67.885 50.923
392253 98.369 -67.894 50.873
392254 98.404 -67.902 50.824
392255 98.437 -67.912 50.777
392256 98.471 -67.920 50.730
392257 98.505 -67.928 50.684
392258 98.539 -67.937 50.638
392259 98.573 -67.946 50.591
392260 98.608 -67.956 50.545
392261 98.644 -67.965 50.499
392262 98.679 -67.975 50.452
392263 98.715 -67.985 50.403
392264 98.750 -67.995 50.354
392265 98.786 -68.005 50.307
392266 98.821 -68.014 50.262
392267 98.857 -68.024 50.217
392268 98.898 -68.017 50.172
392269 98.939 -68.010 50.123
392270 98.982 -68.012 50.071
392271 99.025 -68.014 50.018
392272 99.067 -68.019 49.966
392273 99.101 -68.025 49.913
392274 99.135 -68.026 49.860
392275 99.168 -68.027 49.807
392276 99.202 -68.028 49.757
392277 99.235 -68.028 49.711
392278 99.269 -68.029 49.669
392279 99.302 -68.030 49.627
392280 99.336 -68.031 49.583
392281 99.367 -68.036 49.539
392282 99.394 -68.059 49.494
392283 99.438 -68.082 49.446
392284 99.496 -68.107 49.396
392285 99.489 -68.131 49.351
392286 99.522 -68.132 49.309
392287 99.566 -68.128 49.270
392288 99.609 -68.124 49.230
392289 99.653 -68.121 49.187
392290 99.708 -68.134 49.141
392291 99.762 -68.148 49.095
392292 99.798 -68.158 49.050
392293 99.831 -68.167 49.006
392294 99.863 -68.175 48.963
392295 99.896 -68.184 48.920
392296 99.924 -68.200 48.879
392297 99.966 -68.157 48.836
392298 99.985 -68.148 48.791
392299 100.016 -68.184 48.747
392300 100.049 -68.224 48.703
392301 100.081 -68.263 48.657
392302 100.113 -68.301 48.612
392303 100.143 -68.315 48.567
392304 100.174 -68.318 48.522
392305 100.214 -68.316 48.477
392306 100.253 -68.313 48.435
392307 100.293 -68.312 48.391
392308 100.332 -68.310 48.348
392309 100.372 -68.307 48.304
392310 100.412 -68.306 48.261
392311 100.451 -68.303 48.218
392748 90.036 -67.576 54.878
392749 90.071 -67.586 54.878
392750 90.105 -67.595 54.881
392751 90.140 -67.605 54.885
392752 90.174 -67.614 54.892
392753 90.208 -67.623 54.900
392754 90.242 -67.633 54.909
392755 90.276 -67.643 54.919
392756 90.311 -67.652 54.930
392757 90.345 -67.662 54.942
392758 90.381 -67.670 54.956
392759 90.429 -67.670 54.970
392760 90.486 -67.653 54.983
392761 90.546 -67.648 54.993
392762 90.607 -67.649 54.997
392763 90.668 -67.650 54.994
392764 90.730 -67.650 54.988
392765 90.791 -67.651 54.977
392766 90.852 -67.652 54.964
392767 90.919 -67.648 54.950
392768 90.986 -67.645 54.938
392769 91.052 -67.642 54.926
392770 91.119 -67.639 54.915
392771 91.177 -67.633 54.906
392772 91.227 -67.623 54.898
392773 91.277 -67.613 54.892
392774 91.327 -67.602 54.889
392775 91.377 -67.592 54.889
392776 91.423 -67.583 54.888
392777 91.464 -67.576 54.883
392778 91.505 -67.568 54.876
392779 91.546 -67.561 54.867
392780 91.588 -67.553 54.862
392781 91.629 -67.545 54.855
392782 91.670 -67.538 54.847
392783 91.712 -67.530 54.836
392784 91.753 -67.523 54.824
392785 91.794 -67.515 54.811
392786 91.835 -67.507 54.796
392787 91.876 -67.501 54.781
392788 91.917 -67.511 54.765
392789 91.961 -67.517 54.746
392790 92.015 -67.520 54.727
392791 92.069 -67.523 54.706
392792 92.123 -67.526 54.684
392793 92.177 -67.529 54.662
392794 92.231 -67.533 54.640
392795 92.285 -67.536 54.616
392796 92.339 -67.538 54.592
392797 92.392 -67.541 54.570
392798 92.446 -67.544 54.547
392799 92.500 -67.547 54.525
392800 92.554 -67.550 54.502
392801 92.607 -67.553 54.478
392802 92.661 -67.555 54.453
392803 92.715 -67.558 54.428
392804 92.769 -67.561 54.404
392805 92.823 -67.564 54.383
392806 92.877 -67.567 54.365
392807 92.932 -67.566 54.350
392808 92.971 -67.564 54.338
392809 93.007 -67.562 54.330
392810 93.043 -67.561 54.323
392811 93.079 -67.559 54.315
392812 93.115 -67.558 54.307
392813 93.151 -67.556 54.298
392814 93.187 -67.555 54.286
392815 93.236 -67.550 54.270
392816 93.290 -67.545 54.249
392817 93.344 -67.539 54.228
392818 93.397 -67.534 54.206
392819 93.438 -67.536 54.185
392820 93.476 -67.540 54.167
392821 93.509 -67.539 54.154
392822 93.540 -67.533 54.144
392823 93.570 -67.526 54.135
392824 93.604 -67.521 54.124
392825 93.650 -67.517 54.108
392826 93.697 -67.513 54.088
392827 93.743 -67.509 54.065
392828 93.789 -67.505 54.039
392829 93.835 -67.509 54.013
392830 93.882 -67.514 53.989
392831 93.928 -67.519 53.968
392832 93.973 -67.523 53.946
392833 94.014 -67.528 53.925
392834 94.056 -67.533 53.904
392835 94.094 -67.527 53.883
392836 94.131 -67.521 53.861
392837 94.168 -67.515 53.839
392838 94.205 -67.509 53.815
392839 94.242 -67.502 53.790
392840 94.278 -67.496 53.764
392841 94.315 -67.490 53.736
392842 94.352 -67.483 53.707
392843 94.389 -67.477 53.675
392844 94.435 -67.479 53.647
392845 94.485 -67.484 53.623
392846 94.535 -67.489 53.601
392847 94.582 -67.492 53.581
392848 94.617 -67.490 53.564
392849 94.652 -67.489 53.549
392850 94.688 -67.486 53.532
392851 94.723 -67.485 53.515
392852 94.758 -67.482 53.498
392853 94.794 -67.481 53.484
392854 94.829 -67.479 53.472
392855 94.862 -67.473 53.460
392856 94.890 -67.457 53.444
392857 94.918 -67.441 53.417
392858 94.958 -67.442 53.385
392859 94.999 -67.445 53.352
392860 95.039 -67.447 53.322
392861 95.077 -67.442 53.294
392862 95.115 -67.437 53.269
392863 95.153 -67.432 53.244
392864 95.190 -67.426 53.221
392865 95.228 -67.420 53.199
392866 95.265 -67.415 53.176
392867 95.303 -67.410 53.155
392868 95.344 -67.407 53.134
392869 95.392 -67.410 53.114
392870 95.439 -67.413 53.094
392871 95.486 -67.416 53.073
392872 95.533 -67.419 53.053
392873 95.581 -67.421 53.032
392874 95.626 -67.426 53.013
392875 95.665 -67.435 52.993
392876 95.705 -67.444 52.973
392877 95.744 -67.452 52.952
392878 95.784 -67.461 52.932
392879 95.823 -67.470 52.912
392880 95.863 -67.479 52.890
392881 95.903 -67.489 52.870
392882 95.944 -67.497 52.848
392883 95.984 -67.506 52.826
392884 96.013 -67.513 52.804
392885 96.037 -67.527 52.782
392886 96.061 -67.541 52.759
392887 96.089 -67.552 52.737
392888 96.122 -67.558 52.713
392889 96.155 -67.562 52.691
392890 96.187 -67.564 52.668
392891 96.220 -67.567 52.645
392892 96.251 -67.569 52.623
392893 96.284 -67.571 52.600
392894 96.316 -67.574 52.579
392895 96.347 -67.576 52.557
392896 96.380 -67.579 52.535
392897 96.412 -67.581 52.514
392898 96.445 -67.584 52.492
392899 96.478 -67.587 52.470
392900 96.511 -67.590 52.448
392901 96.544 -67.595 52.427
392902 96.577 -67.603 52.406
392903 96.610 -67.611 52.385
392904 96.644 -67.620 52.364
392905 96.677 -67.624 52.345
392906 96.710 -67.627 52.326
392907 96.743 -67.630 52.308
392908 96.777 -67.633 52.291
392909 96.810 -67.636 52.275
392910 96.843 -67.645 52.261
392911 96.876 -67.655 52.246
392912 96.909 -67.665 52.234
392913 96.948 -67.672 52.223
392914 96.986 -67.677 52.213
392915 97.026 -67.683 52.203
392916 97.065 -67.689 52.193
392917 97.104 -67.695 52.182
392918 97.143 -67.702 52.171
392919 97.182 -67.708 52.158
392920 97.225 -67.707 52.143
392921 97.272 -67.711 52.125
392922 97.319 -67.717 52.105
392923 97.366 -67.724 52.082
392924 97.405 -67.724 52.053
392925 97.440 -67.721 52.021
392926 97.474 -67.718 51.988
392927 97.508 -67.715 51.955
392928 97.542 -67.712 51.923
392929 97.573 -67.713 51.893
392930 97.601 -67.715 51.863
392931 97.630 -67.718 51.834
392932 97.658 -67.721 51.805
392933 97.687 -67.723 51.775
392934 97.715 -67.725 51.743
392935 97.756 -67.729 51.710
392936 97.806 -67.733 51.671
392937 97.844 -67.737 51.629
392938 97.874 -67.743 51.585
392939 97.904 -67.749 51.541
392940 97.935 -67.754 51.497
392941 97.965 -67.759 51.456
392942 97.995 -67.764 51.417
392943 98.025 -67.770 51.378
392944 98.056 -67.776 51.338
392945 98.082 -67.777 51.297
392946 98.108 -67.781 51.255
392947 98.143 -67.797 51.211
392948 98.176 -67.806 51.167
392949 98.209 -67.815 51.117
392950 98.242 -67.823 51.067
392951 98.276 -67.832 51.017
392952 98.309 -67.840 50.966
392953 98.343 -67.849 50.914
392954 98.377 -67.858 50.863
392955 98.411 -67.866 50.813
392956 98.445 -67.875 50.763
392957 98.479 -67.884 50.716
392958 98.512 -67.892 50.670
392959 98.547 -67.901 50.623
392960 98.581 -67.910 50.576
392961 98.616 -67.920 50.531
392962 98.652 -67.930 50.484
392963 98.688 -67.939 50.437
392964 98.723 -67.949 50.388
392965 98.759 -67.959 50.339
392966 98.794 -67.968 50.292
392967 98.830 -67.978 50.248
392968 98.865 -67.988 50.205
392969 98.903 -67.990 50.161
392970 98.944 -67.986 50.115
392971 98.988 -67.988 50.065
392972 99.032 -67.990 50.014
392973 99.081 -67.990 49.962
392974 99.113 -67.985 49.909
392975 99.140 -67.983 49.855
392976 99.173 -67.984 49.802
392977 99.207 -67.984 49.751
392978 99.240 -67.985 49.702
392979 99.274 -67.986 49.656
392980 99.307 -67.987 49.614
392981 99.341 -67.987 49.572
392982 99.374 -67.988 49.528
392983 99.401 -68.009 49.482
392984 99.427 -68.031 49.435
392985 99.484 -68.056 49.386
392986 99.534 -68.082 49.343
392987 99.548 -68.094 49.303
392988 99.585 -68.094 49.265
392989 99.629 -68.089 49.223
392990 99.672 -68.085 49.179
392991 99.724 -68.094 49.133
392992 99.771 -68.106 49.087
392993 99.804 -68.115 49.041
392994 99.837 -68.125 48.998
392995 99.869 -68.134 48.955
392996 99.902 -68.143 48.913
392997 99.932 -68.154 48.871
392998 99.964 -68.126 48.827
392999 99.987 -68.130 48.783
393000 100.019 -68.169 48.739
393001 100.051 -68.208 48.695
393002 100.083 -68.247 48.649
393003 100.114 -68.269 48.605
393004 100.143 -68.274 48.560
393005 100.183 -68.272 48.516
393006 100.223 -68.269 48.471
393007 100.262 -68.268 48.429
393008 100.302 -68.266 48.386
393009 100.342 -68.263 48.342
393010 100.381 -68.262 48.298
393011 100.421 -68.260 48.255
393012 100.461 -68.257 48.212
393449 90.029 -67.527 54.884
393450 90.063 -67.537 54.885
393451 90.097 -67.547 54.889
393452 90.131 -67.556 54.894
393453 90.166 -67.566 54.901
393454 90.200 -67.576 54.910
393455 90.234 -67.585 54.920
393456 90.268 -67.595 54.930
393457 90.303 -67.604 54.940
393458 90.337 -67.614 54.953
393459 90.384 -67.617 54.968
393460 90.437 -67.606 54.981
393461 90.494 -67.589 54.992
393462 90.554 -67.585 55.000
393463 90.615 -67.586 55.002
393464 90.676 -67.586 54.999
393465 90.738 -67.586 54.991
393466 90.799 -67.587 54.980
393467 90.860 -67.588 54.968
393468 90.921 -67.589 54.955
393469 90.983 -67.588 54.944
393470 91.050 -67.586 54.934
393471 91.117 -67.583 54.925
393472 91.183 -67.580 54.920
393473 91.233 -67.569 54.916
393474 91.283 -67.559 54.913
393475 91.333 -67.548 54.912
393476 91.379 -67.539 54.913
393477 91.420 -67.532 54.911
393478 91.461 -67.524 54.905
393479 91.503 -67.517 54.897
393480 91.544 -67.509 54.888
393481 91.585 -67.501 54.883
393482 91.627 -67.494 54.876
393483 91.668 -67.487 54.866
393484 91.709 -67.479 54.854
393485 91.750 -67.472 54.840
393486 91.791 -67.464 54.826
393487 91.832 -67.457 54.810
393488 91.874 -67.449 54.793
393489 91.920 -67.450 54.777
393490 91.974 -67.453 54.757
393491 92.028 -67.455 54.736
393492 92.082 -67.458 54.715
393493 92.135 -67.462 54.693
393494 92.188 -67.466 54.670
393495 92.241 -67.470 54.648
393496 92.295 -67.473 54.624
393497 92.348 -67.477 54.601
393498 92.401 -67.481 54.579
393499 92.454 -67.485 54.556
393500 92.508 -67.489 54.533
393501 92.561 -67.492 54.509
393502 92.614 -67.496 54.483
393503 92.667 -67.500 54.459
393504 92.721 -67.504 54.433
393505 92.774 -67.507 54.410
393506 92.827 -67.511 54.389
393507 92.880 -67.515 54.370
393508 92.937 -67.520 54.356
393509 92.976 -67.511 54.344
393510 93.012 -67.510 54.335
393511 93.048 -67.508 54.326
393512 93.084 -67.507 54.316
393513 93.120 -67.504 54.308
393514 93.156 -67.503 54.299
393515 93.192 -67.501 54.287
393516 93.241 -67.497 54.270
393517 93.295 -67.492 54.251
393518 93.349 -67.486 54.228
393519 93.401 -67.482 54.207
393520 93.438 -67.486 54.187
393521 93.475 -67.491 54.169
393522 93.510 -67.492 54.158
393523 93.541 -67.486 54.149
393524 93.571 -67.479 54.139
393525 93.602 -67.473 54.128
393526 93.635 -67.467 54.112
393527 93.681 -67.464 54.091
393528 93.727 -67.460 54.068
393529 93.773 -67.455 54.042
393530 93.820 -67.456 54.017
393531 93.866 -67.461 53.993
393532 93.913 -67.466 53.972
393533 93.957 -67.470 53.950
393534 93.998 -67.475 53.928
393535 94.040 -67.479 53.908
393536 94.079 -67.478 53.886
393537 94.117 -67.472 53.864
393538 94.153 -67.466 53.842
393539 94.190 -67.459 53.818
393540 94.227 -67.453 53.794
393541 94.264 -67.447 53.768
393542 94.301 -67.440 53.742
393543 94.338 -67.434 53.713
393544 94.382 -67.435 53.681
393545 94.432 -67.439 53.653
393546 94.482 -67.445 53.629
393547 94.532 -67.449 53.608
393548 94.574 -67.451 53.588
393549 94.610 -67.448 53.572
393550 94.645 -67.447 53.556
393551 94.680 -67.445 53.540
393552 94.715 -67.443 53.524
393553 94.751 -67.441 53.508
393554 94.786 -67.439 53.497
393555 94.822 -67.437 53.487
393556 94.857 -67.435 53.475
393557 94.890 -67.429 53.458
393558 94.920 -67.417 53.430
393559 94.961 -67.418 53.395
393560 94.999 -67.415 53.359
393561 95.037 -67.410 53.326
393562 95.074 -67.404 53.298
393563 95.112 -67.399 53.272
393564 95.149 -67.393 53.247
393565 95.187 -67.388 53.224
393566 95.225 -67.382 53.202
393567 95.262 -67.377 53.181
393568 95.301 -67.372 53.160
393569 95.348 -67.375 53.140
393570 95.395 -67.378 53.120
393571 95.442 -67.380 53.099
393572 95.489 -67.383 53.079
393573 95.537 -67.385 53.059
393574 95.584 -67.388 53.039
393575 95.631 -67.392 53.019
393576 95.671 -67.400 52.999
393577 95.710 -67.409 52.979
393578 95.750 -67.418 52.959
393579 95.790 -67.426 52.940
393580 95.829 -67.435 52.918
393581 95.869 -67.445 52.898
393582 95.908 -67.454 52.876
393583 95.948 -67.462 52.855
393584 95.988 -67.471 52.833
393585 96.017 -67.481 52.811
393586 96.042 -67.495 52.787
393587 96.066 -67.509 52.763
393588 96.092 -67.520 52.738
393589 96.124 -67.523 52.715
393590 96.156 -67.525 52.690
393591 96.188 -67.527 52.667
393592 96.220 -67.529 52.643
393593 96.252 -67.532 52.620
393594 96.284 -67.534 52.595
393595 96.316 -67.536 52.573
393596 96.348 -67.539 52.551
393597 96.381 -67.542 52.529
393598 96.413 -67.544 52.507
393599 96.446 -67.547 52.485
393600 96.479 -67.550 52.464
393601 96.512 -67.553 52.443
393602 96.545 -67.556 52.421
393603 96.578 -67.564 52.400
393604 96.611 -67.572 52.379
393605 96.645 -67.579 52.359
393606 96.678 -67.582 52.339
393607 96.711 -67.585 52.319
393608 96.744 -67.588 52.301
393609 96.778 -67.591 52.285
393610 96.811 -67.594 52.269
393611 96.844 -67.598 52.254
393612 96.880 -67.607 52.237
393613 96.919 -67.613 52.223
393614 96.958 -67.619 52.211
393615 96.997 -67.625 52.200
393616 97.036 -67.631 52.189
393617 97.075 -67.636 52.177
393618 97.114 -67.642 52.164
393619 97.153 -67.648 52.151
393620 97.192 -67.655 52.137
393621 97.233 -67.660 52.121
393622 97.280 -67.666 52.102
393623 97.327 -67.672 52.082
393624 97.374 -67.678 52.058
393625 97.419 -67.683 52.029
393626 97.453 -67.680 51.998
393627 97.488 -67.677 51.966
393628 97.522 -67.674 51.934
393629 97.552 -67.674 51.902
393630 97.581 -67.677 51.872
393631 97.609 -67.679 51.841
393632 97.638 -67.682 51.812
393633 97.666 -67.684 51.782
393634 97.695 -67.686 51.752
393635 97.724 -67.689 51.721
393636 97.762 -67.692 51.688
393637 97.812 -67.696 51.651
393638 97.858 -67.701 51.611
393639 97.888 -67.706 51.569
393640 97.919 -67.711 51.526
393641 97.949 -67.717 51.482
393642 97.979 -67.722 51.442
393643 98.010 -67.727 51.403
393644 98.040 -67.733 51.365
393645 98.068 -67.736 51.324
393646 98.089 -67.734 51.283
393647 98.116 -67.739 51.241
393648 98.152 -67.758 51.198
393649 98.186 -67.770 51.153
393650 98.219 -67.779 51.105
393651 98.252 -67.787 51.056
393652 98.286 -67.796 51.007
393653 98.319 -67.804 50.955
393654 98.352 -67.813 50.904
393655 98.385 -67.821 50.853
393656 98.418 -67.830 50.801
393657 98.452 -67.839 50.750
393658 98.486 -67.847 50.703
393659 98.520 -67.855 50.657
393660 98.554 -67.865 50.610
393661 98.589 -67.874 50.562
393662 98.624 -67.884 50.516
393663 98.660 -67.893 50.468
393664 98.695 -67.903 50.421
393665 98.731 -67.913 50.372
393666 98.766 -67.923 50.324
393667 98.802 -67.932 50.277
393668 98.837 -67.942 50.234
393669 98.873 -67.952 50.192
393670 98.909 -67.962 50.149
393671 98.950 -67.962 50.105
393672 99.000 -67.962 50.056
393673 99.049 -67.961 50.006
393674 99.098 -67.961 49.955
393675 99.148 -67.961 49.903
393676 99.183 -67.957 49.849
393677 99.196 -67.948 49.797
393678 99.212 -67.940 49.745
393679 99.245 -67.941 49.695
393680 99.279 -67.942 49.645
393681 99.312 -67.943 49.603
393682 99.346 -67.943 49.560
393683 99.379 -67.944 49.516
393684 99.408 -67.959 49.470
393685 99.434 -67.982 49.423
393686 99.473 -68.006 49.377
393687 99.545 -68.030 49.335
393688 99.586 -68.037 49.296
393689 99.626 -68.043 49.258
393690 99.653 -68.053 49.215
393691 99.692 -68.050 49.169
393692 99.740 -68.055 49.123
393693 99.777 -68.065 49.078
393694 99.810 -68.074 49.032
393695 99.843 -68.083 48.989
393696 99.875 -68.091 48.947
393697 99.908 -68.100 48.905
393698 99.941 -68.106 48.863
393699 99.962 -68.094 48.819
393700 99.989 -68.114 48.776
393701 100.021 -68.153 48.732
393702 100.053 -68.192 48.688
393703 100.085 -68.225 48.643
393704 100.114 -68.229 48.599
393705 100.153 -68.228 48.554
393706 100.192 -68.226 48.510
393707 100.232 -68.224 48.466
393708 100.272 -68.222 48.423
393709 100.311 -68.219 48.380
393710 100.351 -68.218 48.337
393711 100.391 -68.216 48.293
393712 100.430 -68.213 48.250
393713 100.470 -68.212 48.207
394150 90.021 -67.479 54.889
394151 90.055 -67.489 54.891
394152 90.089 -67.498 54.896
394153 90.123 -67.507 54.903
394154 90.158 -67.517 54.910
394155 90.192 -67.527 54.920
394156 90.226 -67.536 54.930
394157 90.260 -67.546 54.940
394158 90.295 -67.556 54.950
394159 90.340 -67.560 54.964
394160 90.388 -67.559 54.978
394161 90.445 -67.542 54.991
394162 90.502 -67.524 55.000
394163 90.562 -67.521 55.007
394164 90.623 -67.522 55.006
394165 90.684 -67.523 55.002
394166 90.746 -67.523 54.994
394167 90.807 -67.523 54.983
394168 90.868 -67.524 54.971
394169 90.929 -67.525 54.960
394170 90.991 -67.525 54.950
394171 91.051 -67.526 54.942
394172 91.115 -67.525 54.937
394173 91.182 -67.522 54.935
394174 91.239 -67.515 54.935
394175 91.289 -67.505 54.935
394176 91.335 -67.496 54.936
394177 91.376 -67.489 54.936
394178 91.418 -67.481 54.931
394179 91.459 -67.473 54.924
394180 91.500 -67.466 54.915
394181 91.541 -67.458 54.907
394182 91.582 -67.451 54.901
394183 91.624 -67.443 54.894
394184 91.665 -67.435 54.884
394185 91.707 -67.429 54.871
394186 91.748 -67.422 54.856
394187 91.790 -67.415 54.841
394188 91.831 -67.407 54.824
394189 91.873 -67.401 54.806
394190 91.920 -67.404 54.788
394191 91.972 -67.407 54.768
394192 92.026 -67.411 54.748
394193 92.079 -67.415 54.727
394194 92.133 -67.419 54.705
394195 92.186 -67.423 54.681
394196 92.239 -67.426 54.657
394197 92.292 -67.430 54.634
394198 92.345 -67.434 54.611
394199 92.399 -67.438 54.588
394200 92.452 -67.442 54.565
394201 92.505 -67.445 54.541
394202 92.558 -67.449 54.516
394203 92.612 -67.453 54.490
394204 92.665 -67.457 54.464
394205 92.718 -67.460 54.440
394206 92.771 -67.464 54.417
394207 92.825 -67.468 54.396
394208 92.883 -67.475 54.377
394209 92.939 -67.476 54.363
394210 92.981 -67.459 54.352
394211 93.017 -67.457 54.342
394212 93.053 -67.455 54.332
394213 93.089 -67.454 54.321
394214 93.125 -67.452 54.310
394215 93.161 -67.451 54.300
394216 93.197 -67.449 54.286
394217 93.247 -67.445 54.269
394218 93.300 -67.439 54.249
394219 93.354 -67.434 54.227
394220 93.401 -67.432 54.205
394221 93.437 -67.437 54.187
394222 93.475 -67.442 54.172
394223 93.511 -67.445 54.161
394224 93.542 -67.439 54.152
394225 93.572 -67.432 54.141
394226 93.603 -67.426 54.129
394227 93.633 -67.420 54.113
394228 93.665 -67.413 54.093
394229 93.712 -67.410 54.069
394230 93.758 -67.406 54.045
394231 93.804 -67.403 54.021
394232 93.851 -67.408 53.998
394233 93.897 -67.413 53.975
394234 93.940 -67.417 53.954
394235 93.982 -67.422 53.933
394236 94.024 -67.426 53.910
394237 94.065 -67.429 53.889
394238 94.102 -67.422 53.866
394239 94.139 -67.416 53.844
394240 94.176 -67.410 53.821
394241 94.212 -67.404 53.797
394242 94.250 -67.397 53.773
394243 94.286 -67.391 53.747
394244 94.329 -67.390 53.720
394245 94.379 -67.395 53.688
394246 94.429 -67.400 53.659
394247 94.479 -67.404 53.635
394248 94.530 -67.410 53.614
394249 94.567 -67.409 53.596
394250 94.602 -67.407 53.580
394251 94.637 -67.405 53.564
394252 94.673 -67.403 53.549
394253 94.708 -67.401 53.534
394254 94.744 -67.399 53.521
394255 94.779 -67.397 53.511
394256 94.814 -67.395 53.500
394257 94.849 -67.393 53.487
394258 94.885 -67.392 53.468
394259 94.921 -67.388 53.439
394260 94.959 -67.382 53.402
394261 94.996 -67.377 53.364
394262 95.034 -67.372 53.329
394263 95.071 -67.366 53.301
394264 95.109 -67.361 53.276
394265 95.147 -67.355 53.251
394266 95.184 -67.350 53.227
394267 95.222 -67.344 53.207
394268 95.260 -67.339 53.186
394269 95.304 -67.339 53.166
394270 95.351 -67.342 53.146
394271 95.398 -67.344 53.126
394272 95.446 -67.348 53.106
394273 95.493 -67.351 53.085
394274 95.540 -67.353 53.066
394275 95.587 -67.356 53.046
394276 95.634 -67.359 53.027
394277 95.677 -67.365 53.007
394278 95.716 -67.374 52.987
394279 95.756 -67.383 52.967
394280 95.796 -67.392 52.948
394281 95.835 -67.401 52.926
394282 95.875 -67.410 52.906
394283 95.914 -67.419 52.884
394284 95.953 -67.428 52.862
394285 95.993 -67.436 52.840
394286 96.022 -67.446 52.817
394287 96.047 -67.457 52.792
394288 96.072 -67.468 52.767
394289 96.097 -67.479 52.743
394290 96.124 -67.486 52.718
394291 96.157 -67.488 52.693
394292 96.189 -67.490 52.668
394293 96.221 -67.492 52.643
394294 96.253 -67.495 52.618
394295 96.285 -67.497 52.593
394296 96.317 -67.499 52.570
394297 96.349 -67.501 52.547
394298 96.382 -67.504 52.525
394299 96.414 -67.507 52.503
394300 96.448 -67.510 52.481
394301 96.480 -67.513 52.460
394302 96.513 -67.516 52.438
394303 96.546 -67.519 52.417
394304 96.579 -67.524 52.396
394305 96.612 -67.533 52.375
394306 96.646 -67.536 52.355
394307 96.679 -67.539 52.335
394308 96.712 -67.542 52.315
394309 96.745 -67.545 52.297
394310 96.779 -67.548 52.281
394311 96.812 -67.553 52.265
394312 96.847 -67.556 52.248
394313 96.886 -67.559 52.231
394314 96.925 -67.566 52.216
394315 96.964 -67.572 52.201
394316 97.003 -67.578 52.188
394317 97.042 -67.584 52.175
394318 97.081 -67.590 52.162
394319 97.120 -67.596 52.148
394320 97.159 -67.602 52.134
394321 97.197 -67.608 52.119
394322 97.239 -67.614 52.103
394323 97.287 -67.620 52.083
394324 97.334 -67.627 52.063
394325 97.381 -67.633 52.040
394326 97.429 -67.639 52.012
394327 97.467 -67.639 51.984
394328 97.501 -67.636 51.952
394329 97.532 -67.636 51.919
394330 97.561 -67.638 51.886
394331 97.589 -67.640 51.854
394332 97.617 -67.642 51.823
394333 97.646 -67.645 51.793
394334 97.675 -67.648 51.762
394335 97.703 -67.650 51.732
394336 97.732 -67.652 51.701
394337 97.769 -67.655 51.668
394338 97.819 -67.660 51.632
394339 97.869 -67.664 51.593
394340 97.902 -67.669 51.552
394341 97.933 -67.674 51.509
394342 97.963 -67.680 51.465
394343 97.994 -67.685 51.425
394344 98.024 -67.690 51.387
394345 98.053 -67.695 51.348
394346 98.075 -67.692 51.307
394347 98.097 -67.691 51.267
394348 98.125 -67.696 51.225
394349 98.160 -67.716 51.182
394350 98.196 -67.734 51.138
394351 98.229 -67.743 51.090
394352 98.263 -67.751 51.042
394353 98.295 -67.760 50.993
394354 98.329 -67.768 50.943
394355 98.362 -67.777 50.893
394356 98.395 -67.785 50.841
394357 98.428 -67.794 50.789
394358 98.461 -67.802 50.738
394359 98.495 -67.811 50.691
394360 98.528 -67.820 50.645
394361 98.561 -67.828 50.599
394362 98.597 -67.838 50.551
394363 98.632 -67.847 50.502
394364 98.667 -67.857 50.452
394365 98.702 -67.865 50.403
394366 98.737 -67.874 50.356
394367 98.772 -67.883 50.309
394368 98.806 -67.891 50.262
394369 98.841 -67.900 50.219
394370 98.876 -67.909 50.177
394371 98.915 -67.921 50.135
394372 98.966 -67.928 50.090
394373 99.016 -67.933 50.041
394374 99.066 -67.933 49.993
394375 99.115 -67.932 49.944
394376 99.164 -67.932 49.893
394377 99.214 -67.931 49.842
394378 99.253 -67.929 49.789
394379 99.267 -67.921 49.739
394380 99.280 -67.912 49.689
394381 99.294 -67.903 49.639
394382 99.317 -67.899 49.593
394383 99.351 -67.900 49.549
394384 99.384 -67.901 49.504
394385 99.415 -67.909 49.459
394386 99.441 -67.932 49.413
394387 99.481 -67.955 49.369
394388 99.547 -67.979 49.327
394389 99.593 -67.988 49.288
394390 99.633 -67.994 49.248
394391 99.673 -68.000 49.204
394392 99.712 -68.006 49.159
394393 99.748 -68.015 49.114
394394 99.779 -68.025 49.069
394395 99.810 -68.035 49.024
394396 99.842 -68.045 48.981
394397 99.873 -68.055 48.938
394398 99.905 -68.065 48.896
394399 99.941 -68.072 48.854
394400 99.960 -68.063 48.811
394401 99.991 -68.099 48.768
394402 100.023 -68.137 48.725
394403 100.056 -68.176 48.681
394404 100.085 -68.184 48.636
394405 100.122 -68.184 48.592
394406 100.162 -68.182 48.548
394407 100.202 -68.180 48.503
394408 100.241 -68.178 48.460
394409 100.281 -68.175 48.416
394410 100.320 -68.174 48.374
394411 100.360 -68.172 48.331
394412 100.400 -68.169 48.288
394413 100.439 -68.168 48.244
394414 100.479 -68.166 48.201
394851 90.013 -67.431 54.894
394852 90.047 -67.440 54.896
394853 90.081 -67.450 54.902
394854 90.115 -67.460 54.910
394855 90.150 -67.469 54.919
394856 90.184 -67.479 54.928
394857 90.218 -67.488 54.938
394858 90.252 -67.498 54.948
394859 90.296 -67.503 54.959
394860 90.342 -67.507 54.972
394861 90.396 -67.495 54.987
394862 90.453 -67.477 54.998
394863 90.510 -67.460 55.005
394864 90.570 -67.458 55.011
394865 90.632 -67.458 55.009
394866 90.692 -67.459 55.004
394867 90.754 -67.460 54.996
394868 90.815 -67.460 54.985
394869 90.876 -67.460 54.975
394870 90.937 -67.461 54.965
394871 90.999 -67.461 54.958
394872 91.060 -67.462 54.954
394873 91.121 -67.463 54.952
394874 91.182 -67.464 54.954
394875 91.245 -67.461 54.956
394876 91.291 -67.452 54.958
394877 91.332 -67.445 54.960
394878 91.374 -67.438 54.957
394879 91.416 -67.430 54.950
394880 91.457 -67.423 54.942
394881 91.499 -67.417 54.932
394882 91.540 -67.410 54.923
394883 91.582 -67.403 54.918
394884 91.624 -67.395 54.910
394885 91.665 -67.388 54.900
394886 91.707 -67.382 54.887
394887 91.749 -67.375 54.872
394888 91.790 -67.367 54.855
394889 91.832 -67.360 54.837
394890 91.875 -67.357 54.818
394891 91.919 -67.360 54.800
394892 91.970 -67.364 54.780
394893 92.023 -67.368 54.760
394894 92.077 -67.372 54.739
394895 92.130 -67.376 54.717
394896 92.183 -67.379 54.692
394897 92.236 -67.383 54.668
394898 92.290 -67.387 54.644
394899 92.343 -67.391 54.620
394900 92.396 -67.395 54.597
394901 92.449 -67.398 54.573
394902 92.502 -67.402 54.549
394903 92.556 -67.406 54.524
394904 92.609 -67.410 54.500
394905 92.662 -67.413 54.475
394906 92.715 -67.417 54.450
394907 92.769 -67.421 54.426
394908 92.827 -67.428 54.404
394909 92.890 -67.439 54.386
394910 92.941 -67.433 54.373
394911 92.982 -67.416 54.363
394912 93.021 -67.404 54.353
394913 93.058 -67.403 54.342
394914 93.093 -67.401 54.330
394915 93.130 -67.399 54.317
394916 93.165 -67.398 54.302
394917 93.202 -67.396 54.286
394918 93.252 -67.392 54.266
394919 93.305 -67.386 54.245
394920 93.359 -67.381 54.223
394921 93.400 -67.383 54.203
394922 93.437 -67.388 54.186
394923 93.474 -67.392 54.173
394924 93.511 -67.396 54.163
394925 93.543 -67.392 54.153
394926 93.573 -67.385 54.142
394927 93.604 -67.379 54.129
394928 93.634 -67.373 54.113
394929 93.664 -67.366 54.093
394930 93.696 -67.360 54.071
394931 93.743 -67.357 54.047
394932 93.789 -67.352 54.023
394933 93.835 -67.355 54.001
394934 93.882 -67.360 53.979
394935 93.924 -67.364 53.958
394936 93.966 -67.369 53.937
394937 94.008 -67.373 53.915
394938 94.050 -67.377 53.893
394939 94.087 -67.373 53.871
394940 94.124 -67.366 53.849
394941 94.161 -67.360 53.827
394942 94.198 -67.354 53.803
394943 94.235 -67.348 53.779
394944 94.276 -67.345 53.753
394945 94.326 -67.350 53.726
394946 94.377 -67.355 53.696
394947 94.426 -67.360 53.668
394948 94.477 -67.365 53.644
394949 94.524 -67.369 53.622
394950 94.559 -67.367 53.602
394951 94.595 -67.365 53.586
394952 94.630 -67.363 53.571
394953 94.666 -67.361 53.557
394954 94.701 -67.359 53.545
394955 94.736 -67.357 53.533
394956 94.771 -67.355 53.521
394957 94.807 -67.354 53.507
394958 94.842 -67.351 53.491
394959 94.882 -67.351 53.470
394960 94.928 -67.354 53.440
394961 94.956 -67.344 53.403
394962 94.993 -67.339 53.366
394963 95.031 -67.334 53.331
394964 95.069 -67.329 53.303
394965 95.107 -67.323 53.278
394966 95.144 -67.317 53.254
394967 95.182 -67.312 53.231
394968 95.219 -67.307 53.211
394969 95.260 -67.304 53.190
394970 95.307 -67.307 53.171
394971 95.354 -67.309 53.152
394972 95.402 -67.312 53.132
394973 95.449 -67.315 53.113
394974 95.496 -67.317 53.092
394975 95.543 -67.320 53.073
394976 95.591 -67.323 53.054
394977 95.638 -67.326 53.034
394978 95.683 -67.330 53.015
394979 95.723 -67.339 52.995
394980 95.762 -67.348 52.975
394981 95.801 -67.357 52.956
394982 95.841 -67.366 52.935
394983 95.880 -67.375 52.915
394984 95.920 -67.384 52.893
394985 95.957 -67.394 52.870
394986 95.994 -67.404 52.847
394987 96.028 -67.404 52.823
394988 96.052 -67.414 52.798
394989 96.077 -67.426 52.773
394990 96.102 -67.436 52.748
394991 96.129 -67.444 52.721
394992 96.158 -67.450 52.696
394993 96.189 -67.453 52.670
394994 96.222 -67.455 52.643
394995 96.254 -67.457 52.617
394996 96.286 -67.460 52.593
394997 96.318 -67.462 52.568
394998 96.350 -67.464 52.545
394999 96.383 -67.467 52.522
395000 96.415 -67.470 52.500
395001 96.449 -67.473 52.478
395002 96.481 -67.476 52.456
395003 96.514 -67.479 52.435
395004 96.547 -67.482 52.413
395005 96.580 -67.486 52.393
395006 96.613 -67.491 52.372
395007 96.647 -67.494 52.352
395008 96.680 -67.498 52.332
395009 96.714 -67.502 52.312
395010 96.747 -67.506 52.294
395011 96.781 -67.510 52.278
395012 96.814 -67.514 52.261
395013 96.850 -67.517 52.245
395014 96.890 -67.517 52.227
395015 96.930 -67.521 52.210
395016 96.969 -67.527 52.193
395017 97.008 -67.533 52.178
395018 97.046 -67.539 52.164
395019 97.085 -67.545 52.149
395020 97.124 -67.551 52.134
395021 97.163 -67.558 52.118
395022 97.202 -67.564 52.102
395023 97.245 -67.570 52.086
395024 97.293 -67.576 52.068
395025 97.341 -67.581 52.048
395026 97.388 -67.587 52.025
395027 97.436 -67.593 51.999
395028 97.480 -67.597 51.971
395029 97.512 -67.597 51.940
395030 97.540 -67.599 51.907
395031 97.569 -67.602 51.873
395032 97.597 -67.604 51.839
395033 97.626 -67.606 51.806
395034 97.654 -67.609 51.774
395035 97.683 -67.611 51.743
395036 97.712 -67.614 51.711
395037 97.740 -67.616 51.680
395038 97.775 -67.619 51.646
395039 97.825 -67.623 51.610
395040 97.875 -67.627 51.572
395041 97.917 -67.632 51.532
395042 97.947 -67.637 51.491
395043 97.978 -67.642 51.449
395044 98.008 -67.648 51.409
395045 98.038 -67.654 51.371
395046 98.061 -67.652 51.329
395047 98.083 -67.650 51.288
395048 98.105 -67.648 51.247
395049 98.133 -67.655 51.207
395050 98.169 -67.674 51.165
395051 98.205 -67.692 51.121
395052 98.239 -67.707 51.075
395053 98.272 -67.715 51.028
395054 98.305 -67.724 50.980
395055 98.339 -67.732 50.931
395056 98.372 -67.741 50.880
395057 98.405 -67.749 50.829
395058 98.438 -67.758 50.778
395059 98.472 -67.767 50.727
395060 98.505 -67.775 50.679
395061 98.537 -67.783 50.632
395062 98.558 -67.786 50.586
395063 98.596 -67.792 50.539
395064 98.632 -67.799 50.491
395065 98.667 -67.808 50.442
395066 98.702 -67.816 50.394
395067 98.736 -67.825 50.346
395068 98.771 -67.833 50.300
395069 98.806 -67.842 50.254
395070 98.841 -67.851 50.208
395071 98.878 -67.870 50.164
395072 98.929 -67.877 50.119
395073 98.980 -67.884 50.075
395074 99.031 -67.892 50.026
395075 99.082 -67.899 49.977
395076 99.132 -67.904 49.929
395077 99.182 -67.903 49.880
395078 99.231 -67.903 49.831
395079 99.280 -67.902 49.781
395080 99.324 -67.901 49.731
395081 99.337 -67.893 49.681
395082 99.351 -67.884 49.632
395083 99.364 -67.875 49.584
395084 99.377 -67.867 49.538
395085 99.391 -67.858 49.494
395086 99.422 -67.860 49.450
395087 99.460 -67.883 49.404
395088 99.502 -67.905 49.361
395089 99.550 -67.927 49.319
395090 99.600 -67.940 49.278
395091 99.640 -67.946 49.236
395092 99.688 -67.944 49.192
395093 99.713 -67.964 49.147
395094 99.738 -67.983 49.104
395095 99.768 -67.996 49.060
395096 99.799 -68.006 49.016
395097 99.831 -68.015 48.973
395098 99.862 -68.025 48.930
395099 99.905 -68.037 48.888
395100 99.939 -68.040 48.846
395101 99.962 -68.044 48.804
395102 99.994 -68.083 48.761
395103 100.026 -68.122 48.719
395104 100.056 -68.140 48.676
395105 100.092 -68.140 48.630
395106 100.131 -68.138 48.586
395107 100.171 -68.136 48.541
395108 100.211 -68.134 48.497
395109 100.250 -68.131 48.453
395110 100.290 -68.130 48.410
395111 100.330 -68.128 48.368
395112 100.369 -68.125 48.325
395113 100.409 -68.124 48.283
395114 100.449 -68.122 48.239
395115 100.488 -68.119 48.196
395552 90.005 -67.382 54.898
395553 90.039 -67.392 54.902
395554 90.074 -67.401 54.908
395555 90.108 -67.411 54.916
395556 90.142 -67.420 54.926
395557 90.176 -67.430 54.935
395558 90.211 -67.440 54.945
395559 90.252 -67.446 54.956
395560 90.298 -67.450 54.966
395561 90.347 -67.448 54.980
395562 90.404 -67.430 54.993
395563 90.461 -67.413 55.004
395564 90.517 -67.395 55.010
395565 90.578 -67.395 55.015
395566 90.640 -67.395 55.011
395567 90.700 -67.395 55.005
395568 90.762 -67.396 54.998
395569 90.823 -67.396 54.988
395570 90.884 -67.397 54.980
395571 90.945 -67.398 54.973
395572 91.007 -67.398 54.968
395573 91.068 -67.398 54.967
395574 91.129 -67.399 54.969
395575 91.190 -67.400 54.974
395576 91.248 -67.407 54.979
395577 91.291 -67.404 54.983
395578 91.332 -67.398 54.983
395579 91.374 -67.390 54.979
395580 91.416 -67.383 54.971
395581 91.457 -67.376 54.960
395582 91.499 -67.370 54.949
395583 91.540 -67.362 54.939
395584 91.582 -67.355 54.932
395585 91.624 -67.348 54.924
395586 91.666 -67.341 54.914
395587 91.707 -67.335 54.901
395588 91.749 -67.327 54.885
395589 91.791 -67.320 54.868
395590 91.832 -67.313 54.850
395591 91.876 -67.314 54.831
395592 91.921 -67.318 54.812
395593 91.967 -67.321 54.792
395594 92.021 -67.325 54.772
395595 92.074 -67.329 54.751
395596 92.127 -67.332 54.729
395597 92.180 -67.336 54.704
395598 92.234 -67.340 54.680
395599 92.287 -67.344 54.655
395600 92.340 -67.348 54.630
395601 92.393 -67.351 54.606
395602 92.447 -67.355 54.583
395603 92.500 -67.359 54.558
395604 92.553 -67.363 54.533
395605 92.606 -67.366 54.510
395606 92.659 -67.370 54.485
395607 92.713 -67.374 54.461
395608 92.771 -67.381 54.437
395609 92.834 -67.392 54.416
395610 92.897 -67.402 54.398
395611 92.942 -67.390 54.385
395612 92.984 -67.373 54.375
395613 93.025 -67.355 54.366
395614 93.063 -67.350 54.356
395615 93.098 -67.348 54.344
395616 93.135 -67.347 54.330
395617 93.170 -67.345 54.313
395618 93.207 -67.344 54.293
395619 93.257 -67.339 54.270
395620 93.310 -67.334 54.246
395621 93.363 -67.329 54.223
395622 93.400 -67.333 54.203
395623 93.437 -67.338 54.186
395624 93.474 -67.342 54.175
395625 93.511 -67.347 54.164
395626 93.544 -67.345 54.153
395627 93.574 -67.338 54.141
395628 93.605 -67.332 54.127
395629 93.635 -67.326 54.111
395630 93.665 -67.319 54.093
395631 93.696 -67.313 54.072
395632 93.727 -67.307 54.048
395633 93.773 -67.303 54.024
395634 93.820 -67.302 54.003
395635 93.866 -67.307 53.983
395636 93.908 -67.311 53.963
395637 93.949 -67.316 53.942
395638 93.992 -67.320 53.920
395639 94.034 -67.324 53.899
395640 94.073 -67.323 53.878
395641 94.110 -67.317 53.857
395642 94.147 -67.310 53.835
395643 94.184 -67.304 53.812
395644 94.223 -67.301 53.787
395645 94.273 -67.306 53.760
395646 94.324 -67.310 53.733
395647 94.373 -67.316 53.704
395648 94.423 -67.320 53.678
395649 94.474 -67.326 53.654
395650 94.517 -67.327 53.631
395651 94.552 -67.326 53.610
395652 94.587 -67.323 53.593
395653 94.623 -67.321 53.578
395654 94.658 -67.319 53.565
395655 94.693 -67.317 53.552
395656 94.729 -67.316 53.540
395657 94.764 -67.313 53.525
395658 94.802 -67.312 53.509
395659 94.841 -67.312 53.489
395660 94.882 -67.311 53.465
395661 94.936 -67.319 53.434
395662 94.963 -67.310 53.399
395663 94.991 -67.301 53.365
395664 95.029 -67.296 53.332
395665 95.066 -67.290 53.305
395666 95.104 -67.285 53.280
395667 95.141 -67.279 53.258
395668 95.179 -67.274 53.236
395669 95.216 -67.269 53.216
395670 95.263 -67.271 53.197
395671 95.311 -67.274 53.178
395672 95.358 -67.276 53.159
395673 95.405 -67.279 53.140
395674 95.452 -67.282 53.122
395675 95.499 -67.285 53.102
395676 95.547 -67.288 53.083
395677 95.594 -67.290 53.064
395678 95.641 -67.293 53.045
395679 95.688 -67.296 53.025
395680 95.728 -67.304 53.005
395681 95.768 -67.313 52.986
395682 95.807 -67.323 52.966
395683 95.846 -67.332 52.945
395684 95.885 -67.341 52.925
395685 95.921 -67.351 52.902
395686 95.959 -67.361 52.879
395687 95.995 -67.370 52.855
395688 96.030 -67.366 52.831
395689 96.058 -67.372 52.805
395690 96.083 -67.382 52.779
395691 96.108 -67.393 52.754
395692 96.135 -67.401 52.726
395693 96.162 -67.408 52.700
395694 96.190 -67.415 52.673
395695 96.223 -67.417 52.645
395696 96.254 -67.420 52.618
395697 96.287 -67.423 52.593
395698 96.319 -67.425 52.568
395699 96.351 -67.427 52.543
395700 96.384 -67.430 52.520
395701 96.417 -67.433 52.497
395702 96.450 -67.435 52.475
395703 96.482 -67.439 52.453
395704 96.516 -67.442 52.432
395705 96.548 -67.445 52.410
395706 96.581 -67.448 52.390
395707 96.615 -67.451 52.370
395708 96.649 -67.455 52.350
395709 96.682 -67.459 52.330
395710 96.716 -67.463 52.310
395711 96.749 -67.467 52.292
395712 96.783 -67.471 52.275
395713 96.816 -67.475 52.260
395714 96.854 -67.477 52.244
395715 96.894 -67.476 52.225
395716 96.934 -67.476 52.205
395717 96.973 -67.482 52.187
395718 97.012 -67.489 52.170
395719 97.051 -67.495 52.154
395720 97.090 -67.501 52.138
395721 97.129 -67.507 52.122
395722 97.168 -67.513 52.105
395723 97.207 -67.519 52.088
395724 97.251 -67.525 52.070
395725 97.299 -67.531 52.053
395726 97.347 -67.536 52.035
395727 97.394 -67.542 52.012
395728 97.442 -67.548 51.987
395729 97.488 -67.555 51.959
395730 97.519 -67.560 51.929
395731 97.548 -67.563 51.896
395732 97.577 -67.565 51.861
395733 97.605 -67.567 51.827
395734 97.634 -67.570 51.792
395735 97.663 -67.573 51.758
395736 97.691 -67.575 51.726
395737 97.720 -67.577 51.694
395738 97.748 -67.580 51.661
395739 97.781 -67.583 51.628
395740 97.831 -67.586 51.591
395741 97.881 -67.591 51.552
395742 97.932 -67.595 51.514
395743 97.962 -67.600 51.474
395744 97.992 -67.605 51.432
395745 98.022 -67.611 51.393
395746 98.046 -67.611 51.354
395747 98.069 -67.608 51.311
395748 98.091 -67.607 51.269
395749 98.113 -67.605 51.228
395750 98.141 -67.612 51.187
395751 98.177 -67.631 51.147
395752 98.213 -67.651 51.104
395753 98.249 -67.670 51.059
395754 98.282 -67.679 51.013
395755 98.315 -67.688 50.966
395756 98.349 -67.696 50.918
395757 98.382 -67.705 50.866
395758 98.415 -67.714 50.816
395759 98.448 -67.722 50.766
395760 98.481 -67.730 50.716
395761 98.510 -67.737 50.667
395762 98.530 -67.740 50.619
395763 98.553 -67.743 50.574
395764 98.590 -67.749 50.528
395765 98.628 -67.754 50.481
395766 98.665 -67.759 50.434
395767 98.701 -67.767 50.388
395768 98.736 -67.776 50.341
395769 98.771 -67.784 50.294
395770 98.805 -67.793 50.249
395771 98.841 -67.818 50.201
395772 98.891 -67.826 50.153
395773 98.943 -67.833 50.107
395774 98.994 -67.840 50.061
395775 99.045 -67.848 50.011
395776 99.095 -67.855 49.963
395777 99.146 -67.862 49.915
395778 99.198 -67.870 49.866
395779 99.248 -67.874 49.819
395780 99.297 -67.874 49.772
395781 99.347 -67.874 49.723
395782 99.394 -67.873 49.673
395783 99.408 -67.865 49.625
395784 99.421 -67.856 49.576
395785 99.434 -67.847 49.531
395786 99.448 -67.839 49.487
395787 99.465 -67.833 49.443
395788 99.503 -67.840 49.398
395789 99.542 -67.847 49.355
395790 99.572 -67.871 49.312
395791 99.630 -67.871 49.269
395792 99.655 -67.891 49.225
395793 99.680 -67.910 49.181
395794 99.705 -67.930 49.137
395795 99.730 -67.949 49.093
395796 99.757 -67.966 49.051
395797 99.788 -67.976 49.007
395798 99.821 -67.986 48.964
395799 99.864 -67.997 48.922
395800 99.907 -68.009 48.880
395801 99.937 -68.009 48.838
395802 99.964 -68.028 48.796
395803 99.996 -68.067 48.753
395804 100.027 -68.095 48.712
395805 100.061 -68.096 48.670
395806 100.101 -68.094 48.624
395807 100.141 -68.092 48.579
395808 100.180 -68.090 48.534
395809 100.220 -68.087 48.490
395810 100.260 -68.086 48.447
395811 100.299 -68.084 48.405
395812 100.339 -68.081 48.362
395813 100.379 -68.080 48.320
395814 100.418 -68.078 48.277
395815 100.458 -68.075 48.234
395816 100.497 -68.074 48.192
396253 89.997 -67.334 54.901
396254 90.031 -67.344 54.906
396255 90.065 -67.353 54.912
396256 90.100 -67.363 54.920
396257 90.134 -67.372 54.931
396258 90.168 -67.382 54.941
396259 90.208 -67.388 54.951
396260 90.254 -67.392 54.961
396261 90.301 -67.397 54.972
396262 90.355 -67.383 54.986
396263 90.412 -67.366 54.999
396264 90.468 -67.348 55.008
396265 90.525 -67.331 55.014
396266 90.586 -67.331 55.017
396267 90.648 -67.332 55.013
396268 90.709 -67.332 55.007
396269 90.770 -67.332 55.001
396270 90.831 -67.333 54.992
396271 90.892 -67.333 54.987
396272 90.953 -67.334 54.982
396273 91.015 -67.335 54.980
396274 91.076 -67.335 54.982
396275 91.135 -67.341 54.988
396276 91.194 -67.348 54.996
396277 91.248 -67.355 55.002
396278 91.291 -67.357 55.006
396279 91.332 -67.350 55.004
396280 91.374 -67.343 54.999
396281 91.416 -67.336 54.989
396282 91.457 -67.329 54.977
396283 91.499 -67.322 54.964
396284 91.541 -67.315 54.955
396285 91.583 -67.308 54.946
396286 91.624 -67.301 54.937
396287 91.666 -67.294 54.928
396288 91.708 -67.287 54.915
396289 91.749 -67.280 54.900
396290 91.791 -67.273 54.882
396291 91.833 -67.268 54.864
396292 91.878 -67.271 54.845
396293 91.922 -67.275 54.826
396294 91.967 -67.279 54.804
396295 92.018 -67.282 54.784
396296 92.071 -67.286 54.763
396297 92.125 -67.289 54.741
396298 92.178 -67.293 54.716
396299 92.231 -67.297 54.691
396300 92.284 -67.301 54.665
396301 92.338 -67.304 54.640
396302 92.391 -67.308 54.617
396303 92.444 -67.312 54.593
396304 92.497 -67.316 54.568
396305 92.551 -67.319 54.543
396306 92.604 -67.323 54.521
396307 92.657 -67.327 54.496
396308 92.715 -67.334 54.473
396309 92.778 -67.344 54.451
396310 92.841 -67.355 54.430
396311 92.903 -67.363 54.413
396312 92.944 -67.346 54.399
396313 92.986 -67.329 54.389
396314 93.027 -67.311 54.380
396315 93.067 -67.297 54.372
396316 93.103 -67.296 54.360
396317 93.139 -67.294 54.344
396318 93.176 -67.292 54.325
396319 93.211 -67.291 54.301
396320 93.262 -67.286 54.275
396321 93.315 -67.281 54.248
396322 93.362 -67.279 54.224
396323 93.399 -67.284 54.204
396324 93.436 -67.288 54.187
396325 93.473 -67.293 54.174
396326 93.510 -67.297 54.162
396327 93.545 -67.298 54.151
396328 93.575 -67.291 54.138
396329 93.606 -67.285 54.125
396330 93.636 -67.279 54.109
396331 93.667 -67.272 54.092
396332 93.697 -67.266 54.072
396333 93.727 -67.260 54.048
396334 93.758 -67.254 54.025
396335 93.804 -67.250 54.006
396336 93.850 -67.254 53.987
396337 93.892 -67.258 53.968
396338 93.933 -67.263 53.949
396339 93.976 -67.267 53.928
396340 94.017 -67.271 53.907
396341 94.059 -67.274 53.886
396342 94.096 -67.267 53.866
396343 94.132 -67.261 53.845
396344 94.171 -67.256 53.821
396345 94.220 -67.261 53.796
396346 94.271 -67.266 53.769
396347 94.321 -67.271 53.742
396348 94.370 -67.276 53.715
396349 94.421 -67.281 53.689
396350 94.471 -67.286 53.664
396351 94.509 -67.285 53.641
396352 94.545 -67.284 53.619
396353 94.580 -67.282 53.601
396354 94.615 -67.279 53.585
396355 94.650 -67.278 53.571
396356 94.686 -67.276 53.558
396357 94.721 -67.274 53.543
396358 94.761 -67.273 53.525
396359 94.801 -67.272 53.506
396360 94.841 -67.272 53.483
396361 94.883 -67.273 53.457
396362 94.939 -67.282 53.426
396363 94.975 -67.279 53.394
396364 94.996 -67.266 53.362
396365 95.026 -67.258 53.333
396366 95.063 -67.252 53.306
396367 95.101 -67.247 53.284
396368 95.138 -67.241 53.262
396369 95.176 -67.236 53.242
396370 95.219 -67.235 53.223
396371 95.267 -67.238 53.204
396372 95.314 -67.241 53.186
396373 95.361 -67.244 53.167
396374 95.408 -67.247 53.149
396375 95.456 -67.249 53.132
396376 95.503 -67.252 53.113
396377 95.550 -67.255 53.094
396378 95.597 -67.257 53.075
396379 95.644 -67.260 53.056
396380 95.692 -67.263 53.036
396381 95.734 -67.269 53.017
396382 95.773 -67.279 52.997
396383 95.812 -67.288 52.976
396384 95.849 -67.298 52.957
396385 95.886 -67.309 52.935
396386 95.923 -67.319 52.913
396387 95.960 -67.329 52.889
396388 95.996 -67.335 52.866
396389 96.031 -67.331 52.841
396390 96.063 -67.329 52.815
396391 96.088 -67.340 52.788
396392 96.113 -67.351 52.761
396393 96.141 -67.358 52.734
396394 96.168 -67.366 52.707
396395 96.195 -67.373 52.678
396396 96.223 -67.380 52.650
396397 96.255 -67.382 52.622
396398 96.288 -67.385 52.595
396399 96.319 -67.387 52.569
396400 96.352 -67.390 52.544
396401 96.385 -67.393 52.520
396402 96.418 -67.396 52.497
396403 96.451 -67.398 52.474
396404 96.483 -67.401 52.452
396405 96.516 -67.405 52.430
396406 96.549 -67.409 52.409
396407 96.583 -67.411 52.390
396408 96.617 -67.412 52.370
396409 96.651 -67.416 52.350
396410 96.684 -67.420 52.330
396411 96.718 -67.424 52.312
396412 96.751 -67.428 52.293
396413 96.785 -67.432 52.277
396414 96.818 -67.436 52.260
396415 96.857 -67.437 52.243
396416 96.897 -67.437 52.225
396417 96.938 -67.437 52.205
396418 96.977 -67.438 52.186
396419 97.017 -67.444 52.168
396420 97.055 -67.450 52.150
396421 97.094 -67.456 52.134
396422 97.133 -67.462 52.115
396423 97.172 -67.468 52.097
396424 97.211 -67.475 52.078
396425 97.256 -67.480 52.059
396426 97.304 -67.486 52.040
396427 97.352 -67.492 52.021
396428 97.400 -67.498 52.000
396429 97.447 -67.504 51.975
396430 97.488 -67.511 51.947
396431 97.518 -67.517 51.917
396432 97.550 -67.522 51.885
396433 97.583 -67.527 51.851
396434 97.614 -67.531 51.819
396435 97.642 -67.534 51.784
396436 97.671 -67.536 51.750
396437 97.699 -67.539 51.717
396438 97.728 -67.541 51.684
396439 97.756 -67.543 51.652
396440 97.788 -67.546 51.616
396441 97.838 -67.550 51.578
396442 97.888 -67.554 51.538
396443 97.938 -67.558 51.498
396444 97.976 -67.563 51.458
396445 98.007 -67.568 51.417
396446 98.032 -67.570 51.377
396447 98.054 -67.567 51.338
396448 98.076 -67.565 51.295
396449 98.098 -67.564 51.253
396450 98.121 -67.561 51.211
396451 98.150 -67.570 51.169
396452 98.185 -67.589 51.128
396453 98.221 -67.608 51.088
396454 98.257 -67.627 51.043
396455 98.292 -67.643 50.997
396456 98.325 -67.652 50.950
396457 98.358 -67.661 50.902
396458 98.391 -67.669 50.851
396459 98.425 -67.677 50.801
396460 98.458 -67.686 50.751
396461 98.482 -67.690 50.703
396462 98.502 -67.693 50.656
396463 98.522 -67.696 50.609
396464 98.547 -67.700 50.563
396465 98.584 -67.705 50.519
396466 98.622 -67.711 50.475
396467 98.660 -67.716 50.431
396468 98.697 -67.721 50.386
396469 98.735 -67.727 50.340
396470 98.770 -67.735 50.294
396471 98.803 -67.767 50.248
396472 98.855 -67.774 50.198
396473 98.906 -67.782 50.148
396474 98.956 -67.789 50.098
396475 99.007 -67.796 50.049
396476 99.058 -67.804 49.999
396477 99.109 -67.811 49.951
396478 99.160 -67.818 49.903
396479 99.211 -67.826 49.855
396480 99.262 -67.833 49.809
396481 99.313 -67.840 49.761
396482 99.364 -67.846 49.713
396483 99.413 -67.845 49.664
396484 99.462 -67.845 49.616
396485 99.478 -67.837 49.572
396486 99.491 -67.828 49.528
396487 99.505 -67.819 49.485
396488 99.528 -67.817 49.443
396489 99.567 -67.824 49.399
396490 99.605 -67.831 49.352
396491 99.635 -67.837 49.306
396492 99.657 -67.840 49.261
396493 99.672 -67.858 49.218
396494 99.687 -67.877 49.173
396495 99.701 -67.896 49.129
396496 99.722 -67.915 49.085
396497 99.747 -67.934 49.042
396498 99.780 -67.946 48.998
396499 99.823 -67.958 48.955
396500 99.866 -67.969 48.913
396501 99.909 -67.981 48.871
396502 99.935 -67.978 48.829
396503 99.966 -68.012 48.787
396504 99.998 -68.050 48.745
396505 100.031 -68.052 48.703
396506 100.071 -68.050 48.661
396507 100.110 -68.048 48.616
396508 100.150 -68.046 48.572
396509 100.190 -68.044 48.527
396510 100.229 -68.042 48.483
396511 100.269 -68.040 48.440
396512 100.308 -68.037 48.399
396513 100.348 -68.036 48.356
396514 100.388 -68.034 48.314
396515 100.427 -68.031 48.271
396516 100.467 -68.030 48.228
396517 100.506 -68.028 48.186
396954 89.989 -67.285 54.905
396955 90.023 -67.295 54.910
396956 90.057 -67.304 54.916
396957 90.092 -67.314 54.924
396958 90.126 -67.324 54.935
396959 90.164 -67.332 54.945
396960 90.211 -67.335 54.955
396961 90.257 -67.339 54.965
396962 90.306 -67.336 54.976
396963 90.363 -67.319 54.989
396964 90.419 -67.301 55.002
396965 90.476 -67.284 55.012
396966 90.533 -67.266 55.018
396967 90.594 -67.267 55.019
396968 90.656 -67.268 55.016
396969 90.717 -67.268 55.011
396970 90.778 -67.269 55.007
396971 90.839 -67.269 55.000
396972 90.900 -67.269 54.995
396973 90.961 -67.270 54.993
396974 91.021 -67.275 54.995
396975 91.080 -67.282 55.000
396976 91.139 -67.288 55.009
396977 91.198 -67.296 55.017
396978 91.248 -67.304 55.024
396979 91.291 -67.310 55.027
396980 91.333 -67.303 55.024
396981 91.375 -67.296 55.017
396982 91.416 -67.288 55.006
396983 91.458 -67.282 54.993
396984 91.500 -67.275 54.980
396985 91.541 -67.268 54.971
396986 91.583 -67.261 54.961
396987 91.625 -67.254 54.952
396988 91.666 -67.247 54.942
396989 91.708 -67.240 54.930
396990 91.749 -67.233 54.915
396991 91.791 -67.225 54.897
396992 91.835 -67.225 54.878
396993 91.879 -67.229 54.858
396994 91.924 -67.232 54.839
396995 91.968 -67.235 54.817
396996 92.015 -67.239 54.795
396997 92.069 -67.243 54.774
396998 92.122 -67.247 54.751
396999 92.175 -67.251 54.727
397000 92.228 -67.254 54.702
397001 92.282 -67.258 54.677
397002 92.335 -67.262 54.652
397003 92.388 -67.266 54.629
397004 92.441 -67.269 54.605
397005 92.495 -67.273 54.580
397006 92.548 -67.276 54.556
397007 92.601 -67.280 54.534
397008 92.658 -67.287 54.510
397009 92.722 -67.297 54.488
397010 92.785 -67.308 54.468
397011 92.848 -67.319 54.449
397012 92.905 -67.320 54.430
397013 92.946 -67.303 54.417
397014 92.988 -67.285 54.407
397015 93.029 -67.268 54.399
397016 93.070 -67.251 54.392
397017 93.109 -67.243 54.378
397018 93.144 -67.241 54.361
397019 93.181 -67.240 54.339
397020 93.216 -67.238 54.311
397021 93.267 -67.233 54.280
397022 93.321 -67.228 54.252
397023 93.362 -67.230 54.227
397024 93.399 -67.235 54.205
397025 93.436 -67.239 54.188
397026 93.473 -67.243 54.174
397027 93.510 -67.247 54.160
397028 93.546 -67.251 54.148
397029 93.577 -67.245 54.135
397030 93.607 -67.238 54.122
397031 93.637 -67.232 54.107
397032 93.668 -67.225 54.090
397033 93.698 -67.219 54.071
397034 93.729 -67.213 54.048
397035 93.759 -67.207 54.027
397036 93.789 -67.200 54.009
397037 93.834 -67.201 53.992
397038 93.875 -67.205 53.974
397039 93.917 -67.210 53.955
397040 93.959 -67.214 53.936
397041 94.001 -67.218 53.916
397042 94.043 -67.222 53.897
397043 94.081 -67.218 53.877
397044 94.118 -67.212 53.856
397045 94.168 -67.216 53.832
397046 94.217 -67.222 53.806
397047 94.268 -67.226 53.780
397048 94.318 -67.232 53.753
397049 94.368 -67.236 53.728
397050 94.418 -67.241 53.702
397051 94.466 -67.246 53.677
397052 94.502 -67.244 53.653
397053 94.537 -67.241 53.629
397054 94.572 -67.240 53.609
397055 94.608 -67.238 53.592
397056 94.643 -67.236 53.577
397057 94.680 -67.235 53.561
397058 94.720 -67.234 53.543
397059 94.761 -67.234 53.522
397060 94.801 -67.233 53.499
397061 94.841 -67.233 53.475
397062 94.886 -67.236 53.447
397063 94.943 -67.245 53.417
397064 94.987 -67.247 53.388
397065 95.009 -67.235 53.360
397066 95.031 -67.223 53.334
397067 95.060 -67.214 53.309
397068 95.098 -67.209 53.288
397069 95.136 -67.204 53.268
397070 95.176 -67.200 53.249
397071 95.223 -67.203 53.231
397072 95.270 -67.206 53.212
397073 95.317 -67.208 53.194
397074 95.365 -67.211 53.177
397075 95.412 -67.213 53.160
397076 95.459 -67.216 53.142
397077 95.507 -67.219 53.124
397078 95.553 -67.222 53.106
397079 95.601 -67.225 53.087
397080 95.648 -67.228 53.068
397081 95.695 -67.230 53.048
397082 95.740 -67.235 53.029
397083 95.776 -67.245 53.009
397084 95.813 -67.256 52.988
397085 95.850 -67.266 52.969
397086 95.887 -67.276 52.946
397087 95.924 -67.287 52.924
397088 95.961 -67.297 52.901
397089 95.997 -67.300 52.878
397090 96.032 -67.295 52.852
397091 96.066 -67.291 52.825
397092 96.094 -67.297 52.798
397093 96.119 -67.308 52.771
397094 96.146 -67.316 52.743
397095 96.173 -67.323 52.714
397096 96.200 -67.331 52.685
397097 96.228 -67.338 52.656
397098 96.256 -67.345 52.627
397099 96.288 -67.348 52.599
397100 96.320 -67.350 52.573
397101 96.353 -67.353 52.546
397102 96.386 -67.356 52.521
397103 96.419 -67.359 52.498
397104 96.452 -67.363 52.475
397105 96.484 -67.366 52.452
397106 96.517 -67.370 52.430
397107 96.550 -67.374 52.409
397108 96.584 -67.375 52.389
397109 96.619 -67.374 52.370
397110 96.653 -67.377 52.351
397111 96.686 -67.381 52.331
397112 96.720 -67.385 52.313
397113 96.753 -67.389 52.295
397114 96.787 -67.394 52.279
397115 96.820 -67.398 52.263
397116 96.860 -67.398 52.245
397117 96.900 -67.398 52.226
397118 96.941 -67.397 52.207
397119 96.981 -67.397 52.188
397120 97.021 -67.399 52.169
397121 97.060 -67.405 52.151
397122 97.099 -67.411 52.133
397123 97.138 -67.417 52.114
397124 97.177 -67.423 52.094
397125 97.216 -67.430 52.075
397126 97.262 -67.435 52.054
397127 97.310 -67.442 52.033
397128 97.358 -67.447 52.012
397129 97.406 -67.453 51.989
397130 97.449 -67.460 51.964
397131 97.488 -67.467 51.936
397132 97.516 -67.474 51.907
397133 97.549 -67.479 51.876
397134 97.581 -67.485 51.845
397135 97.614 -67.489 51.814
397136 97.646 -67.495 51.781
397137 97.678 -67.499 51.747
397138 97.708 -67.502 51.714
397139 97.736 -67.504 51.681
397140 97.764 -67.507 51.649
397141 97.794 -67.510 51.614
397142 97.844 -67.514 51.576
397143 97.894 -67.517 51.536
397144 97.944 -67.522 51.495
397145 97.991 -67.526 51.455
397146 98.018 -67.529 51.414
397147 98.040 -67.526 51.374
397148 98.062 -67.524 51.334
397149 98.084 -67.522 51.292
397150 98.106 -67.520 51.250
397151 98.129 -67.518 51.206
397152 98.158 -67.527 51.162
397153 98.194 -67.547 51.117
397154 98.230 -67.566 51.075
397155 98.266 -67.585 51.029
397156 98.301 -67.605 50.983
397157 98.335 -67.616 50.935
397158 98.368 -67.624 50.887
397159 98.402 -67.633 50.839
397160 98.434 -67.641 50.793
397161 98.454 -67.644 50.747
397162 98.474 -67.647 50.702
397163 98.494 -67.650 50.656
397164 98.514 -67.653 50.609
397165 98.541 -67.657 50.563
397166 98.579 -67.662 50.519
397167 98.617 -67.667 50.475
397168 98.654 -67.673 50.431
397169 98.692 -67.678 50.386
397170 98.732 -67.686 50.340
397171 98.766 -67.712 50.293
397172 98.812 -67.723 50.246
397173 98.865 -67.730 50.196
397174 98.917 -67.737 50.146
397175 98.969 -67.745 50.096
397176 99.021 -67.752 50.047
397177 99.072 -67.760 49.997
397178 99.123 -67.767 49.948
397179 99.174 -67.774 49.899
397180 99.225 -67.782 49.849
397181 99.276 -67.789 49.801
397182 99.327 -67.796 49.754
397183 99.378 -67.804 49.707
397184 99.429 -67.811 49.660
397185 99.479 -67.816 49.615
397186 99.528 -67.816 49.571
397187 99.548 -67.808 49.527
397188 99.562 -67.800 49.484
397189 99.591 -67.802 49.441
397190 99.630 -67.809 49.396
397191 99.654 -67.812 49.348
397192 99.676 -67.815 49.301
397193 99.698 -67.818 49.256
397194 99.717 -67.827 49.212
397195 99.732 -67.846 49.166
397196 99.747 -67.865 49.122
397197 99.761 -67.884 49.077
397198 99.772 -67.902 49.033
397199 99.782 -67.918 48.989
397200 99.825 -67.930 48.946
397201 99.868 -67.941 48.904
397202 99.911 -67.952 48.862
397203 99.936 -67.958 48.820
397204 99.969 -67.996 48.778
397205 100.001 -68.008 48.736
397206 100.040 -68.006 48.694
397207 100.080 -68.004 48.651
397208 100.119 -68.002 48.607
397209 100.159 -68.000 48.563
397210 100.199 -67.998 48.520
397211 100.238 -67.996 48.476
397212 100.278 -67.993 48.433
397213 100.318 -67.992 48.392
397214 100.357 -67.990 48.349
397215 100.397 -67.987 48.307
397216 100.436 -67.986 48.265
397217 100.476 -67.984 48.222
397218 100.516 -67.981 48.179
397655 89.981 -67.237 54.908
397656 90.016 -67.247 54.913
397657 90.050 -67.256 54.920
397658 90.084 -67.266 54.929
397659 90.120 -67.274 54.939
397660 90.167 -67.279 54.948
397661 90.213 -67.282 54.958
397662 90.259 -67.286 54.968
397663 90.314 -67.272 54.979
397664 90.371 -67.254 54.991
397665 90.428 -67.237 55.003
397666 90.484 -67.219 55.014
397667 90.541 -67.203 55.021
397668 90.602 -67.204 55.022
397669 90.664 -67.204 55.020
397670 90.725 -67.204 55.017
397671 90.786 -67.205 55.013
397672 90.847 -67.206 55.008
397673 90.907 -67.209 55.005
397674 90.966 -67.216 55.006
397675 91.025 -67.223 55.011
397676 91.084 -67.230 55.020
397677 91.143 -67.237 55.031
397678 91.202 -67.244 55.038
397679 91.247 -67.251 55.042
397680 91.291 -67.260 55.043
397681 91.333 -67.256 55.039
397682 91.375 -67.248 55.030
397683 91.417 -67.241 55.019
397684 91.458 -67.235 55.008
397685 91.500 -67.228 54.997
397686 91.541 -67.220 54.987
397687 91.583 -67.213 54.977
397688 91.625 -67.207 54.967
397689 91.666 -67.200 54.956
397690 91.708 -67.193 54.944
397691 91.750 -67.185 54.929
397692 91.792 -67.178 54.911
397693 91.836 -67.182 54.892
397694 91.881 -67.185 54.872
397695 91.925 -67.189 54.851
397696 91.970 -67.192 54.828
397697 92.014 -67.196 54.806
397698 92.066 -67.200 54.784
397699 92.120 -67.204 54.761
397700 92.172 -67.207 54.737
397701 92.226 -67.211 54.713
397702 92.279 -67.215 54.689
397703 92.333 -67.219 54.664
397704 92.385 -67.222 54.641
397705 92.439 -67.226 54.617
397706 92.492 -67.230 54.593
397707 92.545 -67.234 54.571
397708 92.602 -67.240 54.550
397709 92.665 -67.251 54.527
397710 92.729 -67.261 54.507
397711 92.792 -67.272 54.487
397712 92.855 -67.282 54.468
397713 92.906 -67.276 54.449
397714 92.948 -67.260 54.434
397715 92.989 -67.242 54.423
397716 93.030 -67.225 54.416
397717 93.072 -67.207 54.409
397718 93.114 -67.191 54.395
397719 93.149 -67.188 54.376
397720 93.185 -67.187 54.351
397721 93.221 -67.185 54.322
397722 93.272 -67.181 54.289
397723 93.324 -67.176 54.259
397724 93.361 -67.181 54.233
397725 93.398 -67.185 54.211
397726 93.435 -67.189 54.191
397727 93.472 -67.194 54.175
397728 93.509 -67.198 54.160
397729 93.546 -67.203 54.146
397730 93.578 -67.198 54.132
397731 93.608 -67.191 54.118
397732 93.639 -67.185 54.103
397733 93.669 -67.178 54.087
397734 93.699 -67.172 54.069
397735 93.730 -67.166 54.048
397736 93.760 -67.160 54.030
397737 93.791 -67.153 54.013
397738 93.826 -67.153 53.997
397739 93.864 -67.155 53.979
397740 93.903 -67.157 53.961
397741 93.943 -67.160 53.943
397742 93.985 -67.165 53.926
397743 94.027 -67.169 53.909
397744 94.066 -67.169 53.890
397745 94.115 -67.172 53.868
397746 94.165 -67.177 53.843
397747 94.215 -67.182 53.818
397748 94.265 -67.187 53.792
397749 94.315 -67.191 53.766
397750 94.365 -67.197 53.742
397751 94.415 -67.202 53.717
397752 94.459 -67.204 53.692
397753 94.494 -67.202 53.667
397754 94.530 -67.200 53.642
397755 94.565 -67.198 53.620
397756 94.600 -67.196 53.600
397757 94.640 -67.195 53.583
397758 94.680 -67.195 53.565
397759 94.720 -67.194 53.544
397760 94.760 -67.194 53.520
397761 94.800 -67.194 53.495
397762 94.840 -67.194 53.470
397763 94.890 -67.199 53.443
397764 94.947 -67.208 53.415
397765 94.999 -67.215 53.387
397766 95.021 -67.203 53.362
397767 95.043 -67.191 53.338
397768 95.064 -67.179 53.315
397769 95.096 -67.171 53.295
397770 95.133 -67.166 53.276
397771 95.179 -67.167 53.258
397772 95.227 -67.170 53.240
397773 95.273 -67.172 53.223
397774 95.321 -67.175 53.206
397775 95.368 -67.178 53.189
397776 95.415 -67.181 53.172
397777 95.462 -67.184 53.155
397778 95.509 -67.185 53.138
397779 95.555 -67.188 53.120
397780 95.601 -67.190 53.101
397781 95.648 -67.192 53.083
397782 95.694 -67.194 53.063
397783 95.738 -67.204 53.043
397784 95.777 -67.213 53.023
397785 95.815 -67.223 53.002
397786 95.851 -67.234 52.981
397787 95.888 -67.244 52.959
397788 95.925 -67.254 52.938
397789 95.962 -67.264 52.915
397790 95.997 -67.265 52.890
397791 96.032 -67.260 52.864
397792 96.067 -67.256 52.837
397793 96.099 -67.255 52.809
397794 96.124 -67.266 52.781
397795 96.152 -67.273 52.752
397796 96.179 -67.281 52.724
397797 96.206 -67.288 52.694
397798 96.234 -67.296 52.664
397799 96.261 -67.304 52.634
397800 96.289 -67.310 52.605
397801 96.321 -67.313 52.577
397802 96.354 -67.316 52.550
397803 96.387 -67.320 52.525
397804 96.419 -67.324 52.500
397805 96.452 -67.328 52.477
397806 96.485 -67.332 52.454
397807 96.518 -67.335 52.431
397808 96.551 -67.338 52.410
397809 96.586 -67.338 52.390
397810 96.620 -67.338 52.371
397811 96.655 -67.338 52.352
397812 96.688 -67.342 52.333
397813 96.722 -67.347 52.314
397814 96.755 -67.351 52.298
397815 96.789 -67.354 52.282
397816 96.823 -67.358 52.266
397817 96.863 -67.358 52.248
397818 96.903 -67.357 52.229
397819 96.944 -67.357 52.210
397820 96.984 -67.357 52.191
397821 97.025 -67.357 52.172
397822 97.065 -67.360 52.154
397823 97.103 -67.366 52.135
397824 97.142 -67.373 52.115
397825 97.181 -67.379 52.094
397826 97.221 -67.385 52.074
397827 97.268 -67.391 52.052
397828 97.316 -67.397 52.030
397829 97.364 -67.402 52.006
397830 97.408 -67.409 51.981
397831 97.451 -67.417 51.956
397832 97.488 -67.423 51.928
397833 97.516 -67.431 51.899
397834 97.547 -67.437 51.870
397835 97.580 -67.442 51.842
397836 97.612 -67.447 51.812
397837 97.645 -67.452 51.779
397838 97.677 -67.457 51.745
397839 97.710 -67.462 51.712
397840 97.742 -67.467 51.680
397841 97.773 -67.471 51.648
397842 97.801 -67.473 51.614
397843 97.850 -67.477 51.576
397844 97.900 -67.481 51.536
397845 97.951 -67.484 51.495
397846 97.999 -67.487 51.455
397847 98.024 -67.494 51.415
397848 98.046 -67.491 51.374
397849 98.068 -67.488 51.333
397850 98.091 -67.485 51.291
397851 98.113 -67.482 51.248
397852 98.135 -67.479 51.203
397853 98.166 -67.486 51.157
397854 98.202 -67.504 51.110
397855 98.238 -67.523 51.064
397856 98.274 -67.543 51.018
397857 98.310 -67.562 50.971
397858 98.345 -67.580 50.923
397859 98.378 -67.589 50.877
397860 98.406 -67.595 50.831
397861 98.426 -67.598 50.789
397862 98.446 -67.601 50.747
397863 98.466 -67.604 50.703
397864 98.486 -67.606 50.659
397865 98.506 -67.609 50.613
397866 98.535 -67.614 50.568
397867 98.573 -67.619 50.522
397868 98.611 -67.624 50.477
397869 98.650 -67.630 50.433
397870 98.692 -67.639 50.387
397871 98.734 -67.648 50.340
397872 98.768 -67.677 50.293
397873 98.816 -67.679 50.244
397874 98.868 -67.686 50.194
397875 98.921 -67.693 50.144
397876 98.973 -67.700 50.094
397877 99.025 -67.708 50.045
397878 99.077 -67.715 49.996
397879 99.130 -67.723 49.946
397880 99.182 -67.730 49.897
397881 99.234 -67.737 49.847
397882 99.286 -67.745 49.797
397883 99.339 -67.752 49.750
397884 99.391 -67.759 49.703
397885 99.442 -67.767 49.658
397886 99.493 -67.774 49.613
397887 99.545 -67.782 49.569
397888 99.595 -67.787 49.525
397889 99.619 -67.780 49.481
397890 99.650 -67.786 49.437
397891 99.673 -67.788 49.391
397892 99.695 -67.791 49.344
397893 99.717 -67.793 49.297
397894 99.739 -67.796 49.251
397895 99.761 -67.799 49.206
397896 99.777 -67.816 49.160
397897 99.788 -67.834 49.115
397898 99.797 -67.852 49.071
397899 99.806 -67.869 49.026
397900 99.809 -67.887 48.982
397901 99.827 -67.902 48.939
397902 99.870 -67.913 48.896
397903 99.912 -67.924 48.854
397904 99.939 -67.942 48.812
397905 99.970 -67.964 48.769
397906 100.010 -67.962 48.727
397907 100.049 -67.960 48.684
397908 100.089 -67.958 48.641
397909 100.129 -67.956 48.598
397910 100.168 -67.954 48.555
397911 100.208 -67.952 48.512
397912 100.248 -67.949 48.469
397913 100.287 -67.948 48.425
397914 100.327 -67.946 48.383
397915 100.366 -67.943 48.342
397916 100.406 -67.942 48.300
397917 100.445 -67.940 48.258
397918 100.485 -67.937 48.216
398356 89.973 -67.188 54.910
398357 90.008 -67.198 54.915
398358 90.042 -67.208 54.923
398359 90.077 -67.217 54.931
398360 90.123 -67.221 54.942
398361 90.169 -67.225 54.951
398362 90.215 -67.229 54.960
398363 90.265 -67.225 54.970
398364 90.322 -67.207 54.980
398365 90.379 -67.190 54.992
398366 90.435 -67.172 55.004
398367 90.492 -67.155 55.015
398368 90.549 -67.139 55.023
398369 90.610 -67.140 55.024
398370 90.672 -67.141 55.024
398371 90.733 -67.141 55.023
398372 90.794 -67.143 55.020
398373 90.853 -67.150 55.016
398374 90.911 -67.157 55.016
398375 90.970 -67.164 55.020
398376 91.029 -67.171 55.029
398377 91.088 -67.178 55.041
398378 91.147 -67.185 55.051
398379 91.204 -67.192 55.057
398380 91.247 -67.200 55.059
398381 91.291 -67.208 55.058
398382 91.333 -67.208 55.052
398383 91.375 -67.201 55.043
398384 91.417 -67.194 55.033
398385 91.458 -67.188 55.023
398386 91.500 -67.180 55.013
398387 91.542 -67.173 55.003
398388 91.583 -67.166 54.993
398389 91.625 -67.160 54.983
398390 91.667 -67.152 54.972
398391 91.709 -67.145 54.959
398392 91.750 -67.138 54.942
398393 91.793 -67.135 54.924
398394 91.838 -67.139 54.905
398395 91.882 -67.142 54.886
398396 91.927 -67.146 54.864
398397 91.971 -67.150 54.841
398398 92.016 -67.153 54.818
398399 92.064 -67.157 54.795
398400 92.117 -67.160 54.771
398401 92.170 -67.164 54.749
398402 92.223 -67.168 54.725
398403 92.277 -67.172 54.702
398404 92.330 -67.175 54.678
398405 92.383 -67.179 54.654
398406 92.436 -67.183 54.631
398407 92.490 -67.187 54.608
398408 92.546 -67.193 54.586
398409 92.609 -67.204 54.567
398410 92.673 -67.214 54.547
398411 92.736 -67.225 54.527
398412 92.799 -67.235 54.507
398413 92.863 -67.245 54.487
398414 92.908 -67.233 54.468
398415 92.949 -67.216 54.451
398416 92.991 -67.198 54.439
398417 93.032 -67.181 54.430
398418 93.074 -67.163 54.424
398419 93.115 -67.147 54.409
398420 93.154 -67.136 54.389
398421 93.190 -67.135 54.364
398422 93.226 -67.132 54.334
398423 93.277 -67.128 54.302
398424 93.324 -67.126 54.271
398425 93.361 -67.131 54.244
398426 93.398 -67.135 54.220
398427 93.435 -67.140 54.198
398428 93.472 -67.144 54.180
398429 93.509 -67.149 54.162
398430 93.546 -67.153 54.145
398431 93.579 -67.151 54.130
398432 93.609 -67.144 54.116
398433 93.640 -67.138 54.101
398434 93.670 -67.131 54.084
398435 93.701 -67.125 54.067
398436 93.731 -67.119 54.048
398437 93.761 -67.113 54.032
398438 93.795 -67.112 54.016
398439 93.829 -67.111 54.000
398440 93.867 -67.113 53.985
398441 93.906 -67.116 53.969
398442 93.945 -67.119 53.954
398443 93.984 -67.121 53.939
398444 94.023 -67.123 53.922
398445 94.065 -67.127 53.903
398446 94.112 -67.132 53.880
398447 94.162 -67.137 53.856
398448 94.212 -67.142 53.831
398449 94.262 -67.147 53.806
398450 94.312 -67.152 53.782
398451 94.362 -67.157 53.758
398452 94.412 -67.162 53.733
398453 94.452 -67.162 53.708
398454 94.487 -67.160 53.683
398455 94.522 -67.158 53.657
398456 94.559 -67.157 53.632
398457 94.600 -67.157 53.610
398458 94.639 -67.156 53.590
398459 94.680 -67.156 53.570
398460 94.719 -67.155 53.546
398461 94.760 -67.155 53.520
398462 94.800 -67.154 53.494
398463 94.840 -67.154 53.468
398464 94.894 -67.162 53.442
398465 94.951 -67.171 53.416
398466 95.007 -67.181 53.390
398467 95.033 -67.171 53.365
398468 95.055 -67.160 53.342
398469 95.077 -67.147 53.321
398470 95.099 -67.135 53.302
398471 95.135 -67.131 53.285
398472 95.182 -67.134 53.267
398473 95.228 -67.136 53.249
398474 95.274 -67.138 53.234
398475 95.321 -67.140 53.218
398476 95.367 -67.142 53.202
398477 95.413 -67.144 53.185
398478 95.460 -67.146 53.169
398479 95.507 -67.148 53.153
398480 95.553 -67.150 53.135
398481 95.599 -67.152 53.117
398482 95.645 -67.154 53.098
398483 95.690 -67.163 53.079
398484 95.734 -67.172 53.059
398485 95.777 -67.181 53.039
398486 95.816 -67.191 53.018
398487 95.852 -67.201 52.996
398488 95.889 -67.212 52.973
398489 95.926 -67.222 52.952
398490 95.963 -67.232 52.929
398491 95.999 -67.229 52.904
398492 96.033 -67.225 52.878
398493 96.068 -67.220 52.850
398494 96.103 -67.216 52.822
398495 96.130 -67.223 52.793
398496 96.157 -67.231 52.764
398497 96.184 -67.238 52.734
398498 96.212 -67.246 52.704
398499 96.239 -67.254 52.674
398500 96.266 -67.261 52.643
398501 96.294 -67.269 52.614
398502 96.322 -67.277 52.585
398503 96.355 -67.281 52.558
398504 96.388 -67.285 52.531
398505 96.420 -67.289 52.506
398506 96.453 -67.293 52.481
398507 96.486 -67.297 52.458
398508 96.519 -67.301 52.436
398509 96.552 -67.302 52.416
398510 96.587 -67.301 52.395
398511 96.621 -67.301 52.376
398512 96.656 -67.301 52.357
398513 96.690 -67.304 52.338
398514 96.724 -67.307 52.320
398515 96.757 -67.312 52.303
398516 96.791 -67.316 52.287
398517 96.826 -67.319 52.270
398518 96.867 -67.319 52.252
398519 96.907 -67.318 52.233
398520 96.947 -67.318 52.215
398521 96.987 -67.317 52.195
398522 97.028 -67.317 52.176
398523 97.069 -67.317 52.159
398524 97.108 -67.322 52.139
398525 97.147 -67.329 52.118
398526 97.185 -67.335 52.096
398527 97.226 -67.341 52.074
398528 97.274 -67.346 52.053
398529 97.322 -67.352 52.031
398530 97.367 -67.358 52.007
398531 97.409 -67.366 51.982
398532 97.453 -67.373 51.955
398533 97.488 -67.380 51.927
398534 97.516 -67.387 51.897
398535 97.546 -67.394 51.868
398536 97.578 -67.399 51.840
398537 97.611 -67.404 51.811
398538 97.643 -67.409 51.778
398539 97.676 -67.414 51.744
398540 97.708 -67.419 51.711
398541 97.740 -67.424 51.679
398542 97.773 -67.429 51.647
398543 97.808 -67.433 51.614
398544 97.859 -67.434 51.577
398545 97.910 -67.435 51.538
398546 97.961 -67.437 51.498
398547 97.993 -67.445 51.458
398548 98.019 -67.455 51.417
398549 98.045 -67.465 51.376
398550 98.070 -67.471 51.335
398551 98.092 -67.468 51.291
398552 98.114 -67.465 51.247
398553 98.137 -67.462 51.202
398554 98.159 -67.459 51.155
398555 98.195 -67.473 51.106
398556 98.234 -67.489 51.057
398557 98.273 -67.507 51.009
398558 98.312 -67.524 50.962
398559 98.351 -67.541 50.917
398560 98.378 -67.548 50.873
398561 98.398 -67.551 50.829
398562 98.418 -67.554 50.788
398563 98.438 -67.557 50.749
398564 98.458 -67.560 50.708
398565 98.479 -67.563 50.665
398566 98.499 -67.566 50.622
398567 98.530 -67.570 50.576
398568 98.568 -67.576 50.528
398569 98.609 -67.585 50.481
398570 98.652 -67.594 50.434
398571 98.694 -67.602 50.388
398572 98.736 -67.612 50.341
398573 98.773 -67.636 50.292
398574 98.822 -67.637 50.241
398575 98.872 -67.641 50.192
398576 98.925 -67.648 50.141
398577 98.977 -67.656 50.091
398578 99.029 -67.663 50.043
398579 99.081 -67.670 49.995
398580 99.134 -67.678 49.946
398581 99.186 -67.686 49.898
398582 99.238 -67.692 49.849
398583 99.290 -67.700 49.799
398584 99.343 -67.708 49.750
398585 99.395 -67.715 49.703
398586 99.447 -67.722 49.656
398587 99.499 -67.730 49.610
398588 99.552 -67.737 49.566
398589 99.604 -67.745 49.521
398590 99.639 -67.755 49.475
398591 99.663 -67.758 49.430
398592 99.684 -67.765 49.385
398593 99.705 -67.772 49.337
398594 99.726 -67.778 49.291
398595 99.748 -67.785 49.245
398596 99.769 -67.790 49.200
398597 99.807 -67.802 49.156
398598 99.824 -67.801 49.110
398599 99.833 -67.819 49.065
398600 99.842 -67.837 49.022
398601 99.848 -67.854 48.978
398602 99.839 -67.872 48.934
398603 99.872 -67.885 48.891
398604 99.910 -67.892 48.847
398605 99.940 -67.920 48.804
398606 99.979 -67.918 48.761
398607 100.019 -67.916 48.717
398608 100.059 -67.914 48.674
398609 100.098 -67.912 48.631
398610 100.138 -67.910 48.589
398611 100.177 -67.908 48.546
398612 100.217 -67.905 48.503
398613 100.257 -67.904 48.461
398614 100.296 -67.902 48.418
398615 100.336 -67.899 48.375
398616 100.375 -67.898 48.334
398617 100.415 -67.896 48.293
398618 100.455 -67.893 48.251
398619 100.497 -67.892 48.209
399057 89.965 -67.140 54.912
399058 90.000 -67.150 54.917
399059 90.034 -67.160 54.924
399060 90.079 -67.164 54.932
399061 90.125 -67.168 54.942
399062 90.171 -67.172 54.951
399063 90.218 -67.175 54.961
399064 90.273 -67.160 54.970
399065 90.330 -67.143 54.980
399066 90.386 -67.125 54.992
399067 90.443 -67.108 55.003
399068 90.500 -67.091 55.014
399069 90.557 -67.076 55.023
399070 90.618 -67.076 55.025
399071 90.679 -67.078 55.027
399072 90.738 -67.084 55.027
399073 90.798 -67.091 55.026
399074 90.857 -67.098 55.025
399075 90.916 -67.105 55.027
399076 90.974 -67.112 55.034
399077 91.033 -67.119 55.046
399078 91.092 -67.126 55.060
399079 91.151 -67.133 55.069
399080 91.203 -67.140 55.074
399081 91.247 -67.148 55.074
399082 91.291 -67.157 55.071
399083 91.333 -67.161 55.064
399084 91.375 -67.154 55.056
399085 91.417 -67.147 55.047
399086 91.459 -67.140 55.038
399087 91.500 -67.133 55.028
399088 91.542 -67.126 55.018
399089 91.584 -67.119 55.009
399090 91.626 -67.112 54.999
399091 91.667 -67.105 54.987
399092 91.709 -67.098 54.972
399093 91.750 -67.091 54.955
399094 91.794 -67.092 54.937
399095 91.839 -67.096 54.918
399096 91.883 -67.100 54.899
399097 91.928 -67.103 54.876
399098 91.973 -67.106 54.854
399099 92.017 -67.110 54.831
399100 92.062 -67.113 54.807
399101 92.114 -67.117 54.784
399102 92.167 -67.121 54.761
399103 92.221 -67.125 54.739
399104 92.274 -67.128 54.716
399105 92.327 -67.132 54.693
399106 92.380 -67.136 54.669
399107 92.434 -67.140 54.648
399108 92.490 -67.146 54.627
399109 92.554 -67.157 54.607
399110 92.617 -67.167 54.592
399111 92.680 -67.178 54.572
399112 92.743 -67.188 54.550
399113 92.806 -67.198 54.528
399114 92.868 -67.207 54.509
399115 92.910 -67.190 54.489
399116 92.951 -67.172 54.470
399117 92.993 -67.155 54.455
399118 93.034 -67.138 54.443
399119 93.075 -67.120 54.434
399120 93.117 -67.103 54.420
399121 93.158 -67.085 54.401
399122 93.195 -67.081 54.377
399123 93.231 -67.080 54.350
399124 93.282 -67.075 54.318
399125 93.324 -67.077 54.286
399126 93.360 -67.081 54.257
399127 93.398 -67.086 54.232
399128 93.434 -67.091 54.208
399129 93.472 -67.095 54.186
399130 93.508 -67.099 54.165
399131 93.546 -67.103 54.146
399132 93.580 -67.104 54.129
399133 93.611 -67.097 54.114
399134 93.641 -67.091 54.098
399135 93.671 -67.084 54.082
399136 93.702 -67.078 54.065
399137 93.732 -67.072 54.049
399138 93.766 -67.071 54.035
399139 93.799 -67.071 54.020
399140 93.834 -67.071 54.005
399141 93.870 -67.072 53.991
399142 93.909 -67.075 53.979
399143 93.948 -67.077 53.965
399144 93.987 -67.079 53.951
399145 94.026 -67.081 53.934
399146 94.071 -67.087 53.914
399147 94.117 -67.092 53.892
399148 94.162 -67.097 53.870
399149 94.209 -67.103 53.847
399150 94.259 -67.107 53.823
399151 94.309 -67.113 53.799
399152 94.359 -67.118 53.776
399153 94.409 -67.122 53.751
399154 94.444 -67.120 53.725
399155 94.479 -67.119 53.700
399156 94.519 -67.118 53.674
399157 94.559 -67.117 53.648
399158 94.599 -67.117 53.623
399159 94.639 -67.116 53.598
399160 94.679 -67.116 53.575
399161 94.719 -67.116 53.552
399162 94.759 -67.116 53.527
399163 94.800 -67.115 53.502
399164 94.840 -67.115 53.475
399165 94.897 -67.125 53.448
399166 94.954 -67.134 53.421
399167 95.008 -67.143 53.395
399168 95.033 -67.133 53.369
399169 95.053 -67.120 53.346
399170 95.073 -67.107 53.327
399171 95.094 -67.094 53.310
399172 95.129 -67.094 53.292
399173 95.179 -67.096 53.276
399174 95.226 -67.098 53.259
399175 95.272 -67.100 53.244
399176 95.318 -67.103 53.230
399177 95.365 -67.104 53.215
399178 95.411 -67.106 53.199
399179 95.458 -67.109 53.184
399180 95.504 -67.111 53.168
399181 95.550 -67.113 53.151
399182 95.597 -67.115 53.133
399183 95.641 -67.122 53.115
399184 95.685 -67.131 53.096
399185 95.729 -67.141 53.076
399186 95.773 -67.150 53.055
399187 95.817 -67.159 53.033
399188 95.854 -67.169 53.011
399189 95.891 -67.179 52.988
399190 95.927 -67.190 52.966
399191 95.964 -67.199 52.944
399192 95.999 -67.194 52.919
399193 96.034 -67.190 52.892
399194 96.068 -67.185 52.865
399195 96.104 -67.181 52.837
399196 96.135 -67.181 52.807
399197 96.163 -67.188 52.778
399198 96.190 -67.196 52.746
399199 96.217 -67.204 52.716
399200 96.244 -67.213 52.686
399201 96.271 -67.222 52.654
399202 96.298 -67.232 52.624
399203 96.325 -67.240 52.595
399204 96.356 -67.246 52.566
399205 96.388 -67.250 52.538
399206 96.421 -67.254 52.513
399207 96.454 -67.258 52.487
399208 96.486 -67.262 52.464
399209 96.519 -67.266 52.443
399210 96.554 -67.265 52.423
399211 96.589 -67.265 52.402
399212 96.623 -67.264 52.383
399213 96.658 -67.264 52.364
399214 96.692 -67.265 52.345
399215 96.726 -67.269 52.327
399216 96.759 -67.273 52.310
399217 96.793 -67.277 52.293
399218 96.829 -67.279 52.276
399219 96.870 -67.279 52.258
399220 96.910 -67.279 52.239
399221 96.951 -67.278 52.220
399222 96.991 -67.278 52.202
399223 97.031 -67.278 52.183
399224 97.072 -67.277 52.165
399225 97.112 -67.277 52.146
399226 97.151 -67.284 52.125
399227 97.190 -67.290 52.104
399228 97.232 -67.296 52.082
399229 97.280 -67.301 52.059
399230 97.325 -67.308 52.036
399231 97.369 -67.315 52.012
399232 97.412 -67.322 51.987
399233 97.454 -67.329 51.960
399234 97.488 -67.336 51.930
399235 97.516 -67.344 51.901
399236 97.545 -67.351 51.871
399237 97.577 -67.357 51.842
399238 97.609 -67.361 51.811
399239 97.642 -67.366 51.779
399240 97.674 -67.371 51.746
399241 97.707 -67.376 51.712
399242 97.739 -67.382 51.679
399243 97.772 -67.386 51.648
399244 97.818 -67.385 51.614
399245 97.869 -67.387 51.578
399246 97.920 -67.388 51.539
399247 97.961 -67.394 51.499
399248 97.987 -67.404 51.460
399249 98.013 -67.413 51.418
399250 98.039 -67.423 51.377
399251 98.065 -67.433 51.335
399252 98.091 -67.444 51.291
399253 98.116 -67.448 51.245
399254 98.138 -67.445 51.199
399255 98.161 -67.442 51.152
399256 98.185 -67.442 51.104
399257 98.224 -67.459 51.056
399258 98.263 -67.476 51.008
399259 98.302 -67.494 50.961
399260 98.334 -67.504 50.915
399261 98.356 -67.507 50.870
399262 98.379 -67.510 50.829
399263 98.401 -67.512 50.790
399264 98.424 -67.514 50.752
399265 98.446 -67.517 50.711
399266 98.468 -67.520 50.669
399267 98.491 -67.523 50.625
399268 98.527 -67.529 50.579
399269 98.570 -67.539 50.532
399270 98.612 -67.548 50.484
399271 98.654 -67.557 50.437
399272 98.697 -67.566 50.388
399273 98.734 -67.583 50.340
399274 98.779 -67.594 50.290
399275 98.828 -67.595 50.240
399276 98.876 -67.596 50.190
399277 98.929 -67.604 50.140
399278 98.981 -67.611 50.091
399279 99.033 -67.619 50.041
399280 99.085 -67.626 49.993
399281 99.138 -67.633 49.944
399282 99.190 -67.641 49.896
399283 99.242 -67.648 49.847
399284 99.294 -67.655 49.797
399285 99.347 -67.663 49.748
399286 99.399 -67.670 49.700
399287 99.451 -67.678 49.652
399288 99.503 -67.685 49.605
399289 99.556 -67.692 49.559
399290 99.608 -67.700 49.513
399291 99.612 -67.716 49.467
399292 99.637 -67.714 49.421
399293 99.658 -67.721 49.375
399294 99.680 -67.727 49.329
399295 99.701 -67.734 49.283
399296 99.722 -67.740 49.237
399297 99.759 -67.780 49.194
399298 99.796 -67.819 49.150
399299 99.832 -67.837 49.105
399300 99.851 -67.827 49.061
399301 99.870 -67.819 49.017
399302 99.886 -67.821 48.973
399303 99.877 -67.840 48.930
399304 99.874 -67.856 48.885
399305 99.911 -67.871 48.841
399306 99.949 -67.874 48.796
399307 99.989 -67.872 48.752
399308 100.028 -67.870 48.707
399309 100.068 -67.868 48.664
399310 100.107 -67.866 48.621
399311 100.147 -67.864 48.579
399312 100.187 -67.862 48.536
399313 100.226 -67.860 48.494
399314 100.266 -67.858 48.452
399315 100.305 -67.855 48.409
399316 100.345 -67.854 48.367
399317 100.385 -67.852 48.326
399318 100.424 -67.849 48.285
399319 100.464 -67.848 48.243
399757 89.923 -67.082 54.909
399758 89.958 -67.092 54.912
399759 89.992 -67.101 54.917
399760 90.035 -67.106 54.923
399761 90.081 -67.110 54.932
399762 90.127 -67.115 54.941
399763 90.174 -67.119 54.951
399764 90.224 -67.113 54.960
399765 90.281 -67.096 54.969
399766 90.337 -67.078 54.980
399767 90.394 -67.061 54.990
399768 90.451 -67.044 55.002
399769 90.508 -67.025 55.013
399770 90.566 -67.012 55.021
399771 90.624 -67.019 55.026
399772 90.683 -67.025 55.029
399773 90.742 -67.032 55.031
399774 90.801 -67.039 55.031
399775 90.860 -67.046 55.033
399776 90.920 -67.053 55.038
399777 90.978 -67.060 55.048
399778 91.037 -67.067 55.063
399779 91.096 -67.074 55.075
399780 91.155 -67.081 55.083
399781 91.203 -67.088 55.086
399782 91.246 -67.097 55.085
399783 91.290 -67.105 55.081
399784 91.334 -67.113 55.074
399785 91.376 -67.106 55.068
399786 91.418 -67.100 55.060
399787 91.459 -67.093 55.051
399788 91.501 -67.086 55.041
399789 91.542 -67.078 55.031
399790 91.584 -67.072 55.022
399791 91.626 -67.065 55.012
399792 91.667 -67.058 55.000
399793 91.709 -67.051 54.984
399794 91.751 -67.046 54.967
399795 91.796 -67.050 54.949
399796 91.841 -67.053 54.931
399797 91.885 -67.056 54.911
399798 91.930 -67.059 54.889
399799 91.975 -67.063 54.867
399800 92.019 -67.067 54.844
399801 92.064 -67.070 54.820
399802 92.111 -67.074 54.797
399803 92.165 -67.078 54.775
399804 92.218 -67.081 54.753
399805 92.271 -67.085 54.731
399806 92.324 -67.089 54.709
399807 92.378 -67.093 54.686
399808 92.434 -67.099 54.665
399809 92.497 -67.109 54.646
399810 92.561 -67.120 54.629
399811 92.624 -67.131 54.616
399812 92.687 -67.141 54.598
399813 92.751 -67.151 54.577
399814 92.813 -67.162 54.555
399815 92.870 -67.163 54.533
399816 92.912 -67.146 54.511
399817 92.953 -67.128 54.491
399818 92.994 -67.112 54.473
399819 93.036 -67.094 54.456
399820 93.077 -67.077 54.446
399821 93.119 -67.059 54.431
399822 93.160 -67.042 54.412
399823 93.200 -67.029 54.388
399824 93.236 -67.027 54.360
399825 93.286 -67.023 54.329
399826 93.323 -67.028 54.300
399827 93.360 -67.032 54.272
399828 93.397 -67.037 54.245
399829 93.434 -67.041 54.218
399830 93.471 -67.045 54.195
399831 93.508 -67.050 54.172
399832 93.545 -67.054 54.150
399833 93.581 -67.057 54.130
399834 93.612 -67.050 54.114
399835 93.642 -67.044 54.097
399836 93.673 -67.037 54.080
399837 93.703 -67.031 54.066
399838 93.736 -67.031 54.052
399839 93.770 -67.030 54.038
399840 93.804 -67.030 54.024
399841 93.838 -67.030 54.012
399842 93.874 -67.031 54.000
399843 93.913 -67.033 53.989
399844 93.952 -67.035 53.977
399845 93.990 -67.037 53.963
399846 94.032 -67.041 53.946
399847 94.077 -67.047 53.925
399848 94.123 -67.052 53.905
399849 94.169 -67.057 53.884
399850 94.214 -67.062 53.863
399851 94.260 -67.068 53.840
399852 94.306 -67.073 53.817
399853 94.356 -67.078 53.795
399854 94.401 -67.081 53.769
399855 94.438 -67.079 53.743
399856 94.479 -67.078 53.717
399857 94.519 -67.078 53.691
399858 94.559 -67.078 53.664
399859 94.599 -67.078 53.637
399860 94.639 -67.077 53.610
399861 94.679 -67.077 53.585
399862 94.719 -67.076 53.561
399863 94.759 -67.076 53.536
399864 94.799 -67.075 53.511
399865 94.844 -67.078 53.484
399866 94.901 -67.088 53.456
399867 94.953 -67.095 53.427
399868 95.005 -67.103 53.400
399869 95.026 -67.091 53.374
399870 95.046 -67.078 53.351
399871 95.066 -67.065 53.333
399872 95.092 -67.058 53.317
399873 95.123 -67.056 53.300
399874 95.171 -67.058 53.285
399875 95.222 -67.061 53.270
399876 95.270 -67.063 53.256
399877 95.316 -67.065 53.242
399878 95.362 -67.067 53.229
399879 95.409 -67.069 53.215
399880 95.455 -67.072 53.200
399881 95.502 -67.073 53.184
399882 95.548 -67.075 53.166
399883 95.593 -67.082 53.149
399884 95.637 -67.091 53.131
399885 95.681 -67.100 53.111
399886 95.725 -67.109 53.091
399887 95.768 -67.119 53.070
399888 95.813 -67.128 53.048
399889 95.855 -67.137 53.025
399890 95.892 -67.147 53.003
399891 95.928 -67.157 52.982
399892 95.965 -67.164 52.959
399893 96.000 -67.160 52.935
399894 96.035 -67.155 52.909
399895 96.070 -67.150 52.881
399896 96.104 -67.145 52.852
399897 96.138 -67.142 52.822
399898 96.168 -67.148 52.792
399899 96.195 -67.157 52.760
399900 96.222 -67.166 52.729
399901 96.249 -67.175 52.698
399902 96.276 -67.184 52.666
399903 96.303 -67.193 52.635
399904 96.330 -67.202 52.605
399905 96.357 -67.210 52.575
399906 96.389 -67.215 52.547
399907 96.422 -67.219 52.522
399908 96.455 -67.222 52.496
399909 96.487 -67.226 52.472
399910 96.521 -67.229 52.451
399911 96.555 -67.229 52.430
399912 96.590 -67.229 52.410
399913 96.624 -67.228 52.391
399914 96.659 -67.227 52.372
399915 96.694 -67.227 52.353
399916 96.728 -67.230 52.334
399917 96.761 -67.234 52.316
399918 96.795 -67.238 52.299
399919 96.832 -67.239 52.281
399920 96.873 -67.239 52.263
399921 96.913 -67.239 52.245
399922 96.954 -67.238 52.227
399923 96.994 -67.238 52.209
399924 97.035 -67.238 52.191
399925 97.075 -67.238 52.172
399926 97.115 -67.238 52.154
399927 97.156 -67.239 52.134
399928 97.194 -67.245 52.114
399929 97.238 -67.251 52.093
399930 97.285 -67.257 52.068
399931 97.327 -67.264 52.044
399932 97.370 -67.272 52.020
399933 97.414 -67.279 51.995
399934 97.456 -67.285 51.966
399935 97.488 -67.293 51.936
399936 97.516 -67.301 51.906
399937 97.545 -67.307 51.875
399938 97.575 -67.313 51.845
399939 97.608 -67.319 51.813
399940 97.640 -67.324 51.781
399941 97.673 -67.329 51.747
399942 97.705 -67.334 51.714
399943 97.738 -67.338 51.681
399944 97.778 -67.340 51.649
399945 97.828 -67.338 51.615
399946 97.879 -67.340 51.579
399947 97.929 -67.341 51.540
399948 97.955 -67.352 51.501
399949 97.981 -67.362 51.461
399950 98.007 -67.372 51.420
399951 98.033 -67.382 51.378
399952 98.060 -67.392 51.335
399953 98.086 -67.402 51.290
399954 98.112 -67.412 51.244
399955 98.138 -67.422 51.197
399956 98.162 -67.426 51.150
399957 98.184 -67.423 51.103
399958 98.214 -67.429 51.056
399959 98.253 -67.446 51.009
399960 98.290 -67.461 50.961
399961 98.312 -67.464 50.914
399962 98.335 -67.466 50.867
399963 98.357 -67.469 50.827
399964 98.379 -67.471 50.789
399965 98.405 -67.474 50.751
399966 98.437 -67.477 50.710
399967 98.469 -67.480 50.669
399968 98.499 -67.486 50.625
399969 98.530 -67.492 50.580
399970 98.572 -67.501 50.533
399971 98.615 -67.511 50.484
399972 98.656 -67.520 50.437
399973 98.699 -67.529 50.388
399974 98.735 -67.551 50.338
399975 98.784 -67.552 50.288
399976 98.833 -67.553 50.238
399977 98.882 -67.555 50.188
399978 98.932 -67.559 50.138
399979 98.985 -67.567 50.089
399980 99.037 -67.574 50.039
399981 99.089 -67.582 49.989
399982 99.141 -67.589 49.941
399983 99.194 -67.596 49.892
399984 99.246 -67.604 49.842
399985 99.298 -67.611 49.793
399986 99.350 -67.618 49.744
399987 99.403 -67.626 49.696
399988 99.455 -67.633 49.647
399989 99.507 -67.641 49.598
399990 99.559 -67.648 49.551
399991 99.584 -67.661 49.504
399992 99.584 -67.677 49.458
399993 99.588 -67.691 49.411
399994 99.633 -67.676 49.365
399995 99.654 -67.683 49.320
399996 99.683 -67.705 49.274
399997 99.720 -67.745 49.230
399998 99.757 -67.784 49.187
399999 99.793 -67.824 49.143
400000 99.831 -67.863 49.099
400001 99.853 -67.871 49.055
400002 99.862 -67.865 49.011
400003 99.871 -67.857 48.967
400004 99.881 -67.846 48.924
400005 99.891 -67.834 48.879
400006 99.918 -67.830 48.834
400007 99.958 -67.828 48.788
400008 99.998 -67.826 48.743
400009 100.037 -67.824 48.697
400010 100.077 -67.822 48.654
400011 100.117 -67.820 48.611
400012 100.156 -67.818 48.569
400013 100.196 -67.816 48.527
400014 100.235 -67.814 48.485
400015 100.275 -67.811 48.443
400016 100.315 -67.810 48.401
400017 100.354 -67.808 48.359
400018 100.394 -67.805 48.318
400019 100.433 -67.804 48.277
400020 100.477 -67.802 48.236
400458 89.915 -67.034 54.908
400459 89.950 -67.044 54.912
400460 89.991 -67.050 54.916
400461 90.037 -67.053 54.922
400462 90.084 -67.057 54.931
400463 90.129 -67.061 54.940
400464 90.176 -67.066 54.948
400465 90.232 -67.049 54.957
400466 90.289 -67.031 54.967
400467 90.345 -67.014 54.977
400468 90.402 -66.997 54.987
400469 90.459 -66.979 54.998
400470 90.516 -66.961 55.010
400471 90.571 -66.961 55.019
400472 90.628 -66.966 55.027
400473 90.687 -66.973 55.032
400474 90.746 -66.981 55.035
400475 90.805 -66.987 55.036
400476 90.864 -66.994 55.040
400477 90.923 -67.001 55.049
400478 90.982 -67.008 55.062
400479 91.041 -67.015 55.078
400480 91.100 -67.022 55.090
400481 91.159 -67.028 55.096
400482 91.203 -67.037 55.098
400483 91.246 -67.045 55.097
400484 91.290 -67.053 55.093
400485 91.333 -67.061 55.086
400486 91.376 -67.059 55.080
400487 91.418 -67.053 55.073
400488 91.459 -67.046 55.065
400489 91.501 -67.038 55.054
400490 91.542 -67.031 55.045
400491 91.584 -67.025 55.036
400492 91.626 -67.018 55.026
400493 91.668 -67.011 55.013
400494 91.709 -67.003 54.997
400495 91.752 -67.003 54.980
400496 91.797 -67.006 54.962
400497 91.842 -67.009 54.943
400498 91.886 -67.013 54.922
400499 91.931 -67.017 54.901
400500 91.976 -67.020 54.879
400501 92.020 -67.024 54.856
400502 92.065 -67.028 54.832
400503 92.110 -67.031 54.810
400504 92.162 -67.034 54.788
400505 92.215 -67.038 54.767
400506 92.269 -67.042 54.747
400507 92.322 -67.046 54.725
400508 92.378 -67.052 54.703
400509 92.441 -67.062 54.683
400510 92.505 -67.073 54.666
400511 92.568 -67.083 54.652
400512 92.631 -67.094 54.640
400513 92.694 -67.104 54.625
400514 92.758 -67.115 54.607
400515 92.821 -67.125 54.584
400516 92.872 -67.120 54.560
400517 92.913 -67.103 54.537
400518 92.954 -67.085 54.515
400519 92.996 -67.068 54.494
400520 93.038 -67.050 54.474
400521 93.079 -67.033 54.460
400522 93.121 -67.016 54.442
400523 93.162 -66.999 54.422
400524 93.203 -66.981 54.398
400525 93.241 -66.975 54.370
400526 93.285 -66.974 54.340
400527 93.323 -66.978 54.313
400528 93.359 -66.982 54.286
400529 93.397 -66.987 54.259
400530 93.433 -66.991 54.230
400531 93.471 -66.996 54.204
400532 93.507 -67.000 54.178
400533 93.545 -67.005 54.155
400534 93.581 -67.009 54.133
400535 93.613 -67.003 54.116
400536 93.643 -66.997 54.100
400537 93.674 -66.990 54.085
400538 93.707 -66.990 54.071
400539 93.741 -66.990 54.057
400540 93.775 -66.990 54.044
400541 93.809 -66.990 54.033
400542 93.843 -66.989 54.023
400543 93.877 -66.989 54.013
400544 93.916 -66.991 54.002
400545 93.955 -66.994 53.989
400546 93.993 -66.996 53.974
400547 94.038 -67.001 53.956
400548 94.083 -67.006 53.936
400549 94.129 -67.012 53.918
400550 94.175 -67.017 53.898
400551 94.220 -67.022 53.878
400552 94.266 -67.028 53.856
400553 94.312 -67.033 53.835
400554 94.357 -67.038 53.813
400555 94.392 -67.039 53.788
400556 94.432 -67.039 53.762
400557 94.474 -67.039 53.735
400558 94.514 -67.038 53.708
400559 94.556 -67.038 53.680
400560 94.597 -67.038 53.653
400561 94.638 -67.037 53.625
400562 94.679 -67.037 53.598
400563 94.718 -67.037 53.573
400564 94.759 -67.037 53.547
400565 94.799 -67.036 53.521
400566 94.846 -67.041 53.493
400567 94.898 -67.048 53.463
400568 94.950 -67.056 53.433
400569 94.998 -67.061 53.405
400570 95.019 -67.048 53.380
400571 95.039 -67.035 53.357
400572 95.059 -67.023 53.341
400573 95.090 -67.021 53.324
400574 95.121 -67.019 53.309
400575 95.162 -67.020 53.295
400576 95.212 -67.022 53.282
400577 95.263 -67.025 53.268
400578 95.314 -67.028 53.256
400579 95.360 -67.030 53.244
400580 95.407 -67.032 53.232
400581 95.453 -67.034 53.217
400582 95.499 -67.036 53.200
400583 95.545 -67.042 53.183
400584 95.589 -67.051 53.165
400585 95.632 -67.059 53.147
400586 95.676 -67.069 53.126
400587 95.720 -67.078 53.106
400588 95.764 -67.087 53.085
400589 95.808 -67.096 53.063
400590 95.852 -67.105 53.040
400591 95.893 -67.115 53.018
400592 95.930 -67.125 52.997
400593 95.966 -67.128 52.974
400594 96.001 -67.124 52.951
400595 96.035 -67.119 52.924
400596 96.070 -67.116 52.896
400597 96.102 -67.114 52.867
400598 96.134 -67.113 52.838
400599 96.165 -67.115 52.808
400600 96.195 -67.122 52.776
400601 96.226 -67.128 52.745
400602 96.253 -67.136 52.714
400603 96.281 -67.145 52.681
400604 96.308 -67.154 52.649
400605 96.335 -67.163 52.618
400606 96.362 -67.172 52.588
400607 96.390 -67.180 52.559
400608 96.422 -67.184 52.532
400609 96.455 -67.188 52.506
400610 96.488 -67.191 52.482
400611 96.522 -67.192 52.460
400612 96.557 -67.192 52.440
400613 96.591 -67.191 52.421
400614 96.626 -67.191 52.402
400615 96.661 -67.191 52.383
400616 96.695 -67.191 52.363
400617 96.729 -67.191 52.343
400618 96.763 -67.195 52.325
400619 96.797 -67.199 52.306
400620 96.836 -67.200 52.288
400621 96.876 -67.200 52.270
400622 96.916 -67.199 52.251
400623 96.957 -67.199 52.234
400624 96.998 -67.198 52.217
400625 97.038 -67.198 52.199
400626 97.079 -67.198 52.180
400627 97.119 -67.197 52.163
400628 97.159 -67.197 52.145
400629 97.199 -67.200 52.126
400630 97.243 -67.207 52.106
400631 97.286 -67.213 52.081
400632 97.329 -67.221 52.056
400633 97.372 -67.228 52.031
400634 97.415 -67.235 52.004
400635 97.458 -67.242 51.974
400636 97.488 -67.250 51.943
400637 97.516 -67.257 51.911
400638 97.545 -67.264 51.880
400639 97.574 -67.271 51.849
400640 97.606 -67.276 51.817
400641 97.639 -67.281 51.783
400642 97.671 -67.286 51.749
400643 97.704 -67.291 51.716
400644 97.737 -67.295 51.685
400645 97.788 -67.291 51.652
400646 97.839 -67.291 51.618
400647 97.890 -67.293 51.581
400648 97.923 -67.300 51.543
400649 97.949 -67.310 51.502
400650 97.975 -67.320 51.462
400651 98.002 -67.330 51.420
400652 98.028 -67.340 51.378
400653 98.054 -67.350 51.335
400654 98.080 -67.360 51.288
400655 98.106 -67.370 51.243
400656 98.132 -67.380 51.197
400657 98.158 -67.390 51.151
400658 98.184 -67.400 51.104
400659 98.208 -67.404 51.059
400660 98.243 -67.416 51.012
400661 98.268 -67.420 50.963
400662 98.290 -67.423 50.913
400663 98.322 -67.426 50.866
400664 98.355 -67.429 50.822
400665 98.387 -67.432 50.783
400666 98.419 -67.435 50.745
400667 98.451 -67.439 50.705
400668 98.482 -67.443 50.664
400669 98.509 -67.450 50.622
400670 98.536 -67.456 50.577
400671 98.574 -67.464 50.531
400672 98.617 -67.473 50.483
400673 98.659 -67.482 50.435
400674 98.701 -67.492 50.386
400675 98.741 -67.509 50.335
400676 98.790 -67.510 50.284
400677 98.839 -67.511 50.235
400678 98.887 -67.513 50.185
400679 98.936 -67.515 50.136
400680 98.989 -67.522 50.086
400681 99.041 -67.529 50.036
400682 99.093 -67.537 49.985
400683 99.145 -67.545 49.935
400684 99.198 -67.551 49.885
400685 99.250 -67.559 49.836
400686 99.302 -67.567 49.787
400687 99.354 -67.574 49.739
400688 99.407 -67.581 49.690
400689 99.459 -67.589 49.641
400690 99.511 -67.596 49.591
400691 99.556 -67.605 49.542
400692 99.556 -67.622 49.496
400693 99.556 -67.639 49.449
400694 99.557 -67.655 49.403
400695 99.596 -67.649 49.357
400696 99.644 -67.658 49.312
400697 99.681 -67.692 49.266
400698 99.718 -67.727 49.222
400699 99.756 -67.762 49.179
400700 99.793 -67.796 49.135
400701 99.835 -67.829 49.092
400702 99.845 -67.818 49.048
400703 99.855 -67.807 49.004
400704 99.864 -67.796 48.959
400705 99.874 -67.784 48.915
400706 99.894 -67.783 48.870
400707 99.928 -67.784 48.825
400708 99.967 -67.782 48.779
400709 100.007 -67.780 48.732
400710 100.047 -67.778 48.687
400711 100.086 -67.776 48.644
400712 100.126 -67.774 48.602
400713 100.165 -67.772 48.560
400714 100.205 -67.770 48.518
400715 100.245 -67.768 48.476
400716 100.284 -67.766 48.434
400717 100.324 -67.764 48.392
400718 100.363 -67.761 48.352
400719 100.403 -67.760 48.310
400720 100.443 -67.758 48.269
401159 89.907 -66.985 54.908
401160 89.947 -66.992 54.911
401161 89.993 -66.997 54.916
401162 90.040 -67.000 54.921
401163 90.086 -67.004 54.929
401164 90.132 -67.008 54.937
401165 90.183 -67.002 54.946
401166 90.240 -66.984 54.954
401167 90.297 -66.967 54.963
401168 90.353 -66.950 54.973
401169 90.410 -66.932 54.983
401170 90.466 -66.914 54.994
401171 90.523 -66.907 55.005
401172 90.578 -66.912 55.015
401173 90.634 -66.915 55.024
401174 90.691 -66.921 55.031
401175 90.750 -66.928 55.036
401176 90.809 -66.935 55.039
401177 90.868 -66.942 55.046
401178 90.927 -66.949 55.058
401179 90.986 -66.956 55.074
401180 91.045 -66.963 55.090
401181 91.104 -66.970 55.101
401182 91.159 -66.977 55.107
401183 91.203 -66.985 55.109
401184 91.246 -66.994 55.108
401185 91.290 -67.001 55.106
401186 91.333 -67.009 55.099
401187 91.376 -67.012 55.093
401188 91.418 -67.006 55.086
401189 91.459 -66.998 55.077
401190 91.501 -66.991 55.068
401191 91.543 -66.984 55.059
401192 91.585 -66.978 55.049
401193 91.626 -66.970 55.039
401194 91.668 -66.963 55.026
401195 91.710 -66.956 55.009
401196 91.754 -66.959 54.993
401197 91.799 -66.963 54.975
401198 91.844 -66.967 54.954
401199 91.888 -66.970 54.932
401200 91.933 -66.974 54.911
401201 91.978 -66.978 54.889
401202 92.022 -66.981 54.867
401203 92.067 -66.984 54.844
401204 92.111 -66.988 54.822
401205 92.159 -66.992 54.801
401206 92.213 -66.996 54.782
401207 92.266 -67.000 54.762
401208 92.322 -67.005 54.741
401209 92.385 -67.015 54.720
401210 92.448 -67.026 54.703
401211 92.512 -67.036 54.688
401212 92.575 -67.047 54.678
401213 92.638 -67.057 54.668
401214 92.702 -67.068 54.654
401215 92.765 -67.078 54.637
401216 92.828 -67.089 54.617
401217 92.873 -67.077 54.593
401218 92.915 -67.059 54.567
401219 92.956 -67.042 54.544
401220 92.998 -67.025 54.520
401221 93.039 -67.007 54.496
401222 93.081 -66.990 54.474
401223 93.122 -66.972 54.454
401224 93.163 -66.955 54.432
401225 93.205 -66.937 54.407
401226 93.246 -66.921 54.380
401227 93.285 -66.924 54.353
401228 93.322 -66.928 54.326
401229 93.359 -66.933 54.299
401230 93.396 -66.937 54.272
401231 93.433 -66.942 54.242
401232 93.470 -66.946 54.213
401233 93.507 -66.950 54.186
401234 93.544 -66.955 54.161
401235 93.581 -66.959 54.139
401236 93.614 -66.956 54.121
401237 93.644 -66.950 54.106
401238 93.678 -66.950 54.092
401239 93.712 -66.949 54.078
401240 93.746 -66.949 54.065
401241 93.780 -66.949 54.054
401242 93.814 -66.949 54.045
401243 93.848 -66.949 54.036
401244 93.882 -66.948 54.027
401245 93.919 -66.950 54.016
401246 93.958 -66.953 54.003
401247 93.999 -66.956 53.986
401248 94.044 -66.961 53.968
401249 94.090 -66.966 53.950
401250 94.135 -66.972 53.932
401251 94.181 -66.977 53.914
401252 94.226 -66.982 53.895
401253 94.271 -66.987 53.875
401254 94.314 -66.992 53.853
401255 94.356 -66.997 53.830
401256 94.394 -67.000 53.806
401257 94.415 -66.997 53.780
401258 94.439 -66.994 53.752
401259 94.480 -66.994 53.725
401260 94.521 -66.994 53.697
401261 94.562 -66.993 53.668
401262 94.603 -66.993 53.640
401263 94.644 -66.993 53.613
401264 94.687 -66.993 53.586
401265 94.739 -66.995 53.560
401266 94.791 -66.997 53.532
401267 94.843 -67.001 53.501
401268 94.895 -67.009 53.469
401269 94.948 -67.017 53.437
401270 94.991 -67.019 53.410
401271 95.011 -67.006 53.386
401272 95.031 -66.994 53.366
401273 95.057 -66.986 53.350
401274 95.089 -66.984 53.334
401275 95.120 -66.983 53.319
401276 95.154 -66.982 53.305
401277 95.204 -66.984 53.294
401278 95.254 -66.987 53.282
401279 95.305 -66.990 53.270
401280 95.355 -66.992 53.260
401281 95.404 -66.994 53.249
401282 95.451 -66.997 53.235
401283 95.496 -67.001 53.218
401284 95.540 -67.010 53.199
401285 95.584 -67.019 53.181
401286 95.628 -67.028 53.161
401287 95.672 -67.037 53.141
401288 95.716 -67.047 53.121
401289 95.760 -67.056 53.100
401290 95.804 -67.065 53.078
401291 95.847 -67.074 53.054
401292 95.891 -67.083 53.033
401293 95.931 -67.093 53.011
401294 95.967 -67.094 52.989
401295 96.002 -67.089 52.966
401296 96.033 -67.088 52.939
401297 96.065 -67.086 52.911
401298 96.098 -67.084 52.883
401299 96.129 -67.083 52.854
401300 96.161 -67.083 52.824
401301 96.192 -67.089 52.793
401302 96.223 -67.096 52.762
401303 96.253 -67.102 52.730
401304 96.284 -67.108 52.696
401305 96.312 -67.116 52.664
401306 96.339 -67.125 52.633
401307 96.367 -67.134 52.602
401308 96.393 -67.143 52.573
401309 96.424 -67.149 52.545
401310 96.456 -67.153 52.518
401311 96.489 -67.157 52.493
401312 96.524 -67.156 52.470
401313 96.558 -67.155 52.450
401314 96.593 -67.155 52.432
401315 96.627 -67.154 52.414
401316 96.662 -67.154 52.395
401317 96.696 -67.153 52.375
401318 96.731 -67.153 52.354
401319 96.765 -67.157 52.335
401320 96.799 -67.160 52.316
401321 96.839 -67.160 52.297
401322 96.880 -67.160 52.279
401323 96.920 -67.160 52.259
401324 96.960 -67.160 52.242
401325 97.001 -67.159 52.225
401326 97.041 -67.159 52.207
401327 97.082 -67.158 52.190
401328 97.122 -67.158 52.173
401329 97.163 -67.158 52.156
401330 97.203 -67.157 52.138
401331 97.245 -67.163 52.118
401332 97.288 -67.170 52.094
401333 97.331 -67.177 52.069
401334 97.374 -67.185 52.043
401335 97.417 -67.191 52.015
401336 97.459 -67.198 51.984
401337 97.488 -67.206 51.952
401338 97.516 -67.213 51.919
401339 97.545 -67.220 51.886
401340 97.574 -67.228 51.854
401341 97.605 -67.233 51.821
401342 97.637 -67.238 51.787
401343 97.670 -67.243 51.753
401344 97.702 -67.248 51.720
401345 97.747 -67.247 51.687
401346 97.798 -67.243 51.654
401347 97.849 -67.244 51.620
401348 97.891 -67.248 51.584
401349 97.918 -67.258 51.545
401350 97.944 -67.269 51.505
401351 97.969 -67.279 51.463
401352 97.996 -67.288 51.420
401353 98.022 -67.298 51.377
401354 98.048 -67.308 51.333
401355 98.074 -67.319 51.287
401356 98.100 -67.329 51.241
401357 98.126 -67.338 51.196
401358 98.152 -67.348 51.149
401359 98.178 -67.358 51.102
401360 98.208 -67.369 51.057
401361 98.240 -67.377 51.009
401362 98.273 -67.381 50.960
401363 98.305 -67.384 50.911
401364 98.337 -67.387 50.864
401365 98.369 -67.390 50.819
401366 98.402 -67.394 50.777
401367 98.434 -67.397 50.736
401368 98.465 -67.401 50.695
401369 98.492 -67.407 50.654
401370 98.519 -67.413 50.613
401371 98.547 -67.420 50.570
401372 98.577 -67.428 50.526
401373 98.619 -67.436 50.481
401374 98.661 -67.445 50.432
401375 98.703 -67.455 50.383
401376 98.746 -67.467 50.333
401377 98.795 -67.468 50.282
401378 98.844 -67.470 50.233
401379 98.892 -67.471 50.182
401380 98.941 -67.472 50.133
401381 98.993 -67.478 50.083
401382 99.045 -67.485 50.033
401383 99.097 -67.492 49.980
401384 99.149 -67.500 49.929
401385 99.202 -67.508 49.879
401386 99.254 -67.514 49.829
401387 99.306 -67.522 49.779
401388 99.358 -67.529 49.730
401389 99.411 -67.537 49.681
401390 99.463 -67.544 49.631
401391 99.515 -67.551 49.583
401392 99.528 -67.566 49.534
401393 99.528 -67.583 49.487
401394 99.528 -67.600 49.441
401395 99.556 -67.617 49.395
401396 99.594 -67.632 49.348
401397 99.642 -67.626 49.303
401398 99.681 -67.655 49.258
401399 99.718 -67.689 49.213
401400 99.756 -67.724 49.170
401401 99.793 -67.759 49.126
401402 99.828 -67.768 49.082
401403 99.838 -67.757 49.038
401404 99.848 -67.746 48.993
401405 99.861 -67.733 48.948
401406 99.885 -67.731 48.904
401407 99.899 -67.739 48.858
401408 99.937 -67.738 48.813
401409 99.977 -67.736 48.767
401410 100.016 -67.734 48.721
401411 100.056 -67.732 48.676
401412 100.095 -67.730 48.633
401413 100.135 -67.728 48.591
401414 100.174 -67.726 48.549
401415 100.214 -67.724 48.508
401416 100.254 -67.722 48.466
401417 100.293 -67.720 48.424
401418 100.333 -67.717 48.382
401419 100.372 -67.716 48.343
401420 100.412 -67.714 48.302
401421 100.458 -67.711 48.261
401860 89.903 -66.935 54.907
401861 89.950 -66.939 54.910
401862 89.996 -66.943 54.915
401863 90.042 -66.947 54.919
401864 90.088 -66.951 54.927
401865 90.134 -66.955 54.935
401866 90.191 -66.937 54.943
401867 90.247 -66.920 54.951
401868 90.304 -66.903 54.960
401869 90.361 -66.885 54.969
401870 90.417 -66.868 54.979
401871 90.474 -66.853 54.989
401872 90.530 -66.858 54.999
401873 90.586 -66.862 55.009
401874 90.641 -66.866 55.019
401875 90.696 -66.871 55.027
401876 90.754 -66.876 55.034
401877 90.813 -66.884 55.040
401878 90.872 -66.890 55.049
401879 90.931 -66.897 55.065
401880 90.990 -66.904 55.082
401881 91.049 -66.911 55.098
401882 91.108 -66.918 55.109
401883 91.158 -66.925 55.116
401884 91.202 -66.934 55.119
401885 91.246 -66.941 55.119
401886 91.289 -66.950 55.117
401887 91.333 -66.958 55.111
401888 91.376 -66.965 55.105
401889 91.418 -66.958 55.098
401890 91.460 -66.951 55.091
401891 91.502 -66.944 55.082
401892 91.543 -66.937 55.071
401893 91.585 -66.930 55.061
401894 91.627 -66.923 55.051
401895 91.668 -66.916 55.038
401896 91.711 -66.913 55.022
401897 91.755 -66.917 55.005
401898 91.800 -66.920 54.986
401899 91.845 -66.924 54.964
401900 91.889 -66.928 54.941
401901 91.934 -66.931 54.920
401902 91.979 -66.934 54.898
401903 92.023 -66.938 54.877
401904 92.068 -66.941 54.856
401905 92.112 -66.945 54.834
401906 92.157 -66.949 54.815
401907 92.210 -66.953 54.796
401908 92.266 -66.958 54.778
401909 92.329 -66.968 54.756
401910 92.392 -66.979 54.737
401911 92.455 -66.989 54.722
401912 92.519 -67.000 54.711
401913 92.582 -67.010 54.703
401914 92.645 -67.021 54.695
401915 92.709 -67.031 54.684
401916 92.772 -67.042 54.669
401917 92.834 -67.050 54.651
401918 92.875 -67.033 54.627
401919 92.917 -67.015 54.602
401920 92.958 -66.998 54.578
401921 93.000 -66.981 54.553
401922 93.041 -66.964 54.526
401923 93.082 -66.946 54.499
401924 93.124 -66.929 54.472
401925 93.165 -66.912 54.446
401926 93.207 -66.894 54.420
401927 93.246 -66.879 54.393
401928 93.283 -66.877 54.365
401929 93.320 -66.881 54.337
401930 93.357 -66.884 54.310
401931 93.395 -66.888 54.282
401932 93.432 -66.893 54.253
401933 93.470 -66.896 54.225
401934 93.506 -66.901 54.198
401935 93.544 -66.906 54.173
401936 93.580 -66.910 54.151
401937 93.615 -66.909 54.135
401938 93.648 -66.909 54.119
401939 93.683 -66.909 54.103
401940 93.716 -66.909 54.088
401941 93.751 -66.909 54.075
401942 93.784 -66.908 54.066
401943 93.819 -66.908 54.058
401944 93.852 -66.908 54.050
401945 93.887 -66.908 54.042
401946 93.922 -66.909 54.031
401947 93.961 -66.911 54.017
401948 94.005 -66.915 54.000
401949 94.051 -66.921 53.982
401950 94.096 -66.926 53.964
401951 94.142 -66.931 53.947
401952 94.185 -66.936 53.931
401953 94.228 -66.941 53.913
401954 94.271 -66.946 53.893
401955 94.314 -66.950 53.872
401956 94.356 -66.956 53.849
401957 94.399 -66.960 53.824
401958 94.420 -66.957 53.798
401959 94.441 -66.954 53.769
401960 94.463 -66.951 53.741
401961 94.487 -66.948 53.713
401962 94.529 -66.948 53.684
401963 94.580 -66.950 53.655
401964 94.632 -66.953 53.627
401965 94.684 -66.955 53.599
401966 94.736 -66.957 53.570
401967 94.788 -66.959 53.540
401968 94.840 -66.962 53.507
401969 94.892 -66.969 53.474
401970 94.945 -66.978 53.444
401971 94.984 -66.977 53.418
401972 95.004 -66.964 53.395
401973 95.024 -66.951 53.376
401974 95.055 -66.950 53.360
401975 95.087 -66.948 53.345
401976 95.118 -66.946 53.330
401977 95.149 -66.945 53.316
401978 95.195 -66.946 53.307
401979 95.246 -66.949 53.296
401980 95.296 -66.951 53.286
401981 95.346 -66.954 53.276
401982 95.397 -66.956 53.266
401983 95.448 -66.963 53.252
401984 95.491 -66.969 53.234
401985 95.535 -66.978 53.214
401986 95.579 -66.987 53.194
401987 95.623 -66.997 53.174
401988 95.667 -67.006 53.154
401989 95.711 -67.015 53.134
401990 95.755 -67.024 53.113
401991 95.799 -67.033 53.091
401992 95.843 -67.042 53.068
401993 95.887 -67.051 53.046
401994 95.931 -67.060 53.025
401995 95.965 -67.061 53.003
401996 95.997 -67.059 52.980
401997 96.029 -67.058 52.953
401998 96.061 -67.056 52.925
401999 96.093 -67.055 52.898
402000 96.125 -67.053 52.870
402001 96.157 -67.053 52.840
402002 96.188 -67.057 52.810
402003 96.219 -67.063 52.779
402004 96.249 -67.070 52.746
402005 96.279 -67.076 52.712
402006 96.310 -67.082 52.680
402007 96.340 -67.089 52.648
402008 96.371 -67.095 52.617
402009 96.398 -67.104 52.587
402010 96.425 -67.113 52.559
402011 96.457 -67.118 52.532
402012 96.490 -67.119 52.506
402013 96.525 -67.119 52.483
402014 96.559 -67.119 52.461
402015 96.594 -67.119 52.444
402016 96.629 -67.118 52.426
402017 96.663 -67.118 52.407
402018 96.698 -67.117 52.388
402019 96.733 -67.116 52.366
402020 96.767 -67.118 52.346
402021 96.802 -67.121 52.326
402022 96.842 -67.120 52.307
402023 96.883 -67.120 52.288
402024 96.923 -67.120 52.268
402025 96.964 -67.119 52.251
402026 97.004 -67.119 52.233
402027 97.044 -67.119 52.215
402028 97.083 -67.119 52.199
402029 97.122 -67.119 52.182
402030 97.162 -67.119 52.165
402031 97.204 -67.120 52.148
402032 97.246 -67.123 52.128
402033 97.290 -67.126 52.106
402034 97.333 -67.134 52.081
402035 97.376 -67.141 52.055
402036 97.419 -67.148 52.026
402037 97.459 -67.155 51.994
402038 97.488 -67.163 51.962
402039 97.516 -67.169 51.927
402040 97.545 -67.177 51.893
402041 97.574 -67.184 51.860
402042 97.603 -67.191 51.826
402043 97.636 -67.196 51.791
402044 97.668 -67.200 51.757
402045 97.707 -67.203 51.723
402046 97.757 -67.197 51.689
402047 97.808 -67.195 51.655
402048 97.860 -67.197 51.621
402049 97.885 -67.207 51.584
402050 97.912 -67.216 51.546
402051 97.938 -67.227 51.505
402052 97.964 -67.237 51.462
402053 97.990 -67.247 51.419
402054 98.016 -67.257 51.376
402055 98.042 -67.266 51.330
402056 98.068 -67.277 51.285
402057 98.094 -67.287 51.239
402058 98.125 -67.297 51.193
402059 98.157 -67.307 51.146
402060 98.190 -67.318 51.099
402061 98.222 -67.329 51.053
402062 98.255 -67.339 51.004
402063 98.287 -67.342 50.955
402064 98.319 -67.345 50.907
402065 98.352 -67.348 50.861
402066 98.384 -67.351 50.816
402067 98.416 -67.355 50.772
402068 98.448 -67.358 50.730
402069 98.476 -67.365 50.688
402070 98.503 -67.372 50.646
402071 98.530 -67.378 50.605
402072 98.557 -67.385 50.564
402073 98.584 -67.392 50.522
402074 98.621 -67.400 50.479
402075 98.663 -67.408 50.430
402076 98.703 -67.424 50.381
402077 98.751 -67.426 50.331
402078 98.800 -67.426 50.280
402079 98.849 -67.428 50.231
402080 98.898 -67.429 50.180
402081 98.947 -67.430 50.130
402082 98.997 -67.433 50.080
402083 99.049 -67.440 50.028
402084 99.101 -67.448 49.976
402085 99.153 -67.455 49.924
402086 99.206 -67.463 49.873
402087 99.258 -67.470 49.822
402088 99.310 -67.477 49.771
402089 99.362 -67.485 49.721
402090 99.415 -67.492 49.672
402091 99.467 -67.500 49.622
402092 99.500 -67.511 49.574
402093 99.500 -67.527 49.526
402094 99.500 -67.544 49.478
402095 99.524 -67.561 49.432
402096 99.568 -67.579 49.386
402097 99.614 -67.589 49.339
402098 99.632 -67.617 49.294
402099 99.681 -67.617 49.248
402100 99.718 -67.652 49.204
402101 99.755 -67.687 49.160
402102 99.808 -67.717 49.116
402103 99.821 -67.707 49.071
402104 99.831 -67.695 49.026
402105 99.851 -67.683 48.981
402106 99.877 -67.679 48.936
402107 99.885 -67.687 48.891
402108 99.906 -67.694 48.846
402109 99.946 -67.692 48.800
402110 99.986 -67.690 48.754
402111 100.025 -67.688 48.709
402112 100.065 -67.686 48.665
402113 100.105 -67.684 48.622
402114 100.144 -67.682 48.580
402115 100.184 -67.680 48.539
402116 100.223 -67.678 48.497
402117 100.263 -67.676 48.456
402118 100.302 -67.674 48.414
402119 100.342 -67.672 48.373
402120 100.382 -67.670 48.333
402121 100.424 -67.667 48.293
402561 89.906 -66.882 54.906
402562 89.952 -66.886 54.909
402563 89.998 -66.890 54.913
402564 90.044 -66.893 54.918
402565 90.090 -66.897 54.925
402566 90.142 -66.890 54.932
402567 90.198 -66.873 54.940
402568 90.255 -66.856 54.949
402569 90.312 -66.838 54.957
402570 90.369 -66.821 54.965
402571 90.426 -66.802 54.974
402572 90.481 -66.804 54.983
402573 90.537 -66.808 54.992
402574 90.593 -66.812 55.001
402575 90.648 -66.817 55.011
402576 90.704 -66.821 55.021
402577 90.759 -66.825 55.030
402578 90.817 -66.831 55.039
402579 90.876 -66.838 55.052
402580 90.935 -66.845 55.070
402581 90.994 -66.852 55.089
402582 91.053 -66.859 55.103
402583 91.112 -66.866 55.115
402584 91.158 -66.874 55.123
402585 91.202 -66.882 55.128
402586 91.245 -66.890 55.130
402587 91.289 -66.898 55.128
402588 91.332 -66.906 55.123
402589 91.376 -66.914 55.117
402590 91.419 -66.911 55.110
402591 91.460 -66.904 55.103
402592 91.502 -66.896 55.094
402593 91.543 -66.890 55.082
402594 91.585 -66.883 55.070
402595 91.627 -66.876 55.059
402596 91.668 -66.869 55.048
402597 91.713 -66.870 55.031
402598 91.757 -66.874 55.014
402599 91.802 -66.878 54.994
402600 91.847 -66.881 54.972
402601 91.891 -66.884 54.950
402602 91.936 -66.888 54.929
402603 91.981 -66.891 54.908
402604 92.025 -66.895 54.887
402605 92.070 -66.899 54.867
402606 92.114 -66.902 54.847
402607 92.159 -66.906 54.829
402608 92.210 -66.911 54.811
402609 92.273 -66.921 54.792
402610 92.336 -66.932 54.770
402611 92.400 -66.942 54.754
402612 92.463 -66.953 54.742
402613 92.526 -66.963 54.734
402614 92.589 -66.974 54.728
402615 92.653 -66.984 54.722
402616 92.716 -66.995 54.714
402617 92.779 -67.006 54.702
402618 92.836 -67.007 54.684
402619 92.877 -66.990 54.662
402620 92.918 -66.972 54.638
402621 92.960 -66.955 54.614
402622 93.001 -66.937 54.588
402623 93.043 -66.920 54.559
402624 93.084 -66.903 54.528
402625 93.126 -66.885 54.496
402626 93.166 -66.868 54.466
402627 93.205 -66.855 54.437
402628 93.242 -66.840 54.408
402629 93.278 -66.831 54.379
402630 93.315 -66.834 54.350
402631 93.353 -66.839 54.321
402632 93.391 -66.843 54.293
402633 93.428 -66.846 54.265
402634 93.466 -66.851 54.239
402635 93.503 -66.855 54.213
402636 93.540 -66.859 54.190
402637 93.578 -66.862 54.168
402638 93.616 -66.868 54.151
402639 93.652 -66.869 54.133
402640 93.687 -66.868 54.116
402641 93.721 -66.868 54.101
402642 93.755 -66.868 54.089
402643 93.789 -66.868 54.080
402644 93.823 -66.867 54.073
402645 93.857 -66.867 54.066
402646 93.891 -66.867 54.058
402647 93.925 -66.867 54.047
402648 93.966 -66.870 54.032
402649 94.011 -66.875 54.015
402650 94.057 -66.881 53.997
402651 94.099 -66.885 53.979
402652 94.142 -66.890 53.963
402653 94.185 -66.895 53.948
402654 94.227 -66.900 53.931
402655 94.270 -66.905 53.912
402656 94.313 -66.909 53.892
402657 94.356 -66.914 53.868
402658 94.399 -66.919 53.843
402659 94.426 -66.918 53.816
402660 94.447 -66.915 53.787
402661 94.469 -66.912 53.758
402662 94.490 -66.909 53.729
402663 94.530 -66.910 53.700
402664 94.578 -66.912 53.670
402665 94.630 -66.915 53.641
402666 94.682 -66.917 53.611
402667 94.734 -66.919 53.580
402668 94.785 -66.921 53.547
402669 94.837 -66.924 53.514
402670 94.890 -66.930 53.481
402671 94.942 -66.938 53.452
402672 94.977 -66.934 53.426
402673 94.996 -66.921 53.404
402674 95.023 -66.915 53.386
402675 95.054 -66.913 53.371
402676 95.085 -66.912 53.356
402677 95.116 -66.909 53.342
402678 95.147 -66.908 53.329
402679 95.186 -66.908 53.321
402680 95.237 -66.911 53.312
402681 95.287 -66.913 53.303
402682 95.338 -66.915 53.294
402683 95.388 -66.918 53.284
402684 95.447 -66.947 53.268
402685 95.487 -66.943 53.248
402686 95.528 -66.941 53.228
402687 95.571 -66.950 53.207
402688 95.616 -66.960 53.187
402689 95.660 -66.969 53.167
402690 95.703 -66.978 53.147
402691 95.748 -66.987 53.126
402692 95.792 -66.997 53.105
402693 95.835 -67.006 53.082
402694 95.880 -67.015 53.061
402695 95.922 -67.024 53.039
402696 95.959 -67.028 53.017
402697 95.993 -67.030 52.994
402698 96.025 -67.028 52.967
402699 96.057 -67.027 52.940
402700 96.089 -67.026 52.913
402701 96.121 -67.025 52.886
402702 96.153 -67.023 52.857
402703 96.184 -67.025 52.827
402704 96.215 -67.031 52.796
402705 96.245 -67.037 52.764
402706 96.276 -67.044 52.731
402707 96.306 -67.050 52.698
402708 96.337 -67.056 52.666
402709 96.367 -67.063 52.634
402710 96.398 -67.069 52.604
402711 96.429 -67.075 52.576
402712 96.458 -67.083 52.548
402713 96.492 -67.083 52.521
402714 96.527 -67.082 52.496
402715 96.561 -67.082 52.474
402716 96.596 -67.081 52.455
402717 96.630 -67.081 52.437
402718 96.665 -67.081 52.418
402719 96.699 -67.081 52.399
402720 96.734 -67.080 52.379
402721 96.768 -67.080 52.360
402722 96.805 -67.081 52.341
402723 96.846 -67.081 52.321
402724 96.886 -67.081 52.300
402725 96.925 -67.080 52.279
402726 96.964 -67.080 52.260
402727 97.003 -67.080 52.241
402728 97.042 -67.080 52.223
402729 97.082 -67.080 52.205
402730 97.121 -67.080 52.189
402731 97.162 -67.081 52.171
402732 97.205 -67.083 52.154
402733 97.247 -67.086 52.136
402734 97.291 -67.089 52.115
402735 97.334 -67.092 52.092
402736 97.378 -67.097 52.066
402737 97.421 -67.104 52.037
402738 97.458 -67.112 52.004
402739 97.487 -67.119 51.972
402740 97.516 -67.126 51.937
402741 97.544 -67.133 51.902
402742 97.573 -67.141 51.867
402743 97.602 -67.147 51.832
402744 97.635 -67.153 51.797
402745 97.667 -67.158 51.761
402746 97.717 -67.153 51.725
402747 97.768 -67.149 51.690
402748 97.819 -67.148 51.655
402749 97.854 -67.155 51.620
402750 97.880 -67.165 51.583
402751 97.906 -67.175 51.544
402752 97.932 -67.185 51.502
402753 97.958 -67.195 51.461
402754 97.984 -67.205 51.419
402755 98.010 -67.215 51.375
402756 98.042 -67.225 51.329
402757 98.074 -67.236 51.282
402758 98.107 -67.247 51.236
402759 98.139 -67.257 51.189
402760 98.172 -67.267 51.141
402761 98.204 -67.278 51.093
402762 98.237 -67.288 51.046
402763 98.270 -67.299 50.997
402764 98.302 -67.304 50.949
402765 98.334 -67.307 50.902
402766 98.366 -67.310 50.857
402767 98.398 -67.313 50.813
402768 98.431 -67.316 50.770
402769 98.459 -67.323 50.728
402770 98.486 -67.329 50.684
402771 98.513 -67.335 50.641
402772 98.540 -67.342 50.599
402773 98.567 -67.349 50.559
402774 98.594 -67.356 50.518
402775 98.624 -67.363 50.476
402776 98.666 -67.372 50.427
402777 98.708 -67.382 50.379
402778 98.757 -67.383 50.330
402779 98.806 -67.385 50.280
402780 98.855 -67.386 50.230
402781 98.904 -67.387 50.178
402782 98.952 -67.388 50.127
402783 99.001 -67.390 50.076
402784 99.053 -67.396 50.023
402785 99.105 -67.403 49.971
402786 99.157 -67.410 49.920
402787 99.209 -67.418 49.867
402788 99.262 -67.426 49.814
402789 99.314 -67.433 49.763
402790 99.366 -67.440 49.713
402791 99.418 -67.448 49.663
402792 99.471 -67.455 49.613
402793 99.472 -67.472 49.565
402794 99.473 -67.489 49.516
402795 99.492 -67.505 49.470
402796 99.536 -67.523 49.423
402797 99.579 -67.540 49.376
402798 99.624 -67.555 49.330
402799 99.663 -67.561 49.285
402800 99.675 -67.595 49.239
402801 99.718 -67.614 49.194
402802 99.755 -67.649 49.149
402803 99.804 -67.657 49.104
402804 99.814 -67.645 49.058
402805 99.841 -67.632 49.013
402806 99.869 -67.627 48.968
402807 99.877 -67.635 48.923
402808 99.890 -67.643 48.877
402809 99.916 -67.648 48.833
402810 99.955 -67.646 48.787
402811 99.995 -67.644 48.741
402812 100.034 -67.642 48.697
402813 100.074 -67.640 48.654
402814 100.114 -67.638 48.611
402815 100.153 -67.636 48.570
402816 100.193 -67.634 48.529
402817 100.232 -67.632 48.487
402818 100.272 -67.630 48.447
402819 100.312 -67.628 48.405
402820 100.351 -67.626 48.363
402821 100.391 -67.623 48.324
402822 100.438 -67.622 48.283
403262 89.908 -66.829 54.907
403263 89.954 -66.833 54.908
403264 90.001 -66.837 54.911
403265 90.046 -66.840 54.915
403266 90.093 -66.843 54.922
403267 90.150 -66.826 54.929
403268 90.206 -66.809 54.937
403269 90.263 -66.791 54.945
403270 90.320 -66.774 54.952
403271 90.377 -66.756 54.960
403272 90.433 -66.750 54.968
403273 90.488 -66.754 54.977
403274 90.544 -66.758 54.986
403275 90.600 -66.763 54.996
403276 90.655 -66.767 55.005
403277 90.711 -66.771 55.015
403278 90.766 -66.775 55.025
403279 90.822 -66.780 55.036
403280 90.880 -66.786 55.052
403281 90.939 -66.793 55.072
403282 90.998 -66.800 55.092
403283 91.057 -66.807 55.108
403284 91.114 -66.814 55.118
403285 91.158 -66.822 55.128
403286 91.202 -66.830 55.136
403287 91.245 -66.838 55.141
403288 91.289 -66.846 55.140
403289 91.332 -66.854 55.135
403290 91.376 -66.862 55.129
403291 91.419 -66.864 55.122
403292 91.460 -66.856 55.114
403293 91.502 -66.849 55.104
403294 91.544 -66.843 55.092
403295 91.585 -66.836 55.079
403296 91.627 -66.828 55.067
403297 91.669 -66.824 55.056
403298 91.714 -66.827 55.039
403299 91.758 -66.831 55.020
403300 91.803 -66.834 55.000
403301 91.848 -66.838 54.979
403302 91.892 -66.841 54.958
403303 91.937 -66.845 54.937
403304 91.982 -66.849 54.917
403305 92.026 -66.852 54.897
403306 92.071 -66.856 54.878
403307 92.116 -66.859 54.859
403308 92.162 -66.865 54.841
403309 92.220 -66.874 54.823
403310 92.280 -66.885 54.805
403311 92.344 -66.895 54.785
403312 92.407 -66.906 54.769
403313 92.470 -66.916 54.758
403314 92.533 -66.927 54.752
403315 92.596 -66.937 54.751
403316 92.660 -66.948 54.749
403317 92.723 -66.959 54.743
403318 92.786 -66.968 54.733
403319 92.837 -66.963 54.718
403320 92.879 -66.946 54.696
403321 92.920 -66.929 54.674
403322 92.961 -66.912 54.650
403323 93.003 -66.894 54.624
403324 93.045 -66.877 54.594
403325 93.086 -66.859 54.560
403326 93.125 -66.844 54.525
403327 93.163 -66.830 54.490
403328 93.202 -66.815 54.457
403329 93.239 -66.802 54.425
403330 93.276 -66.791 54.394
403331 93.311 -66.789 54.363
403332 93.349 -66.793 54.335
403333 93.386 -66.797 54.308
403334 93.424 -66.801 54.281
403335 93.461 -66.805 54.256
403336 93.498 -66.809 54.232
403337 93.536 -66.813 54.208
403338 93.573 -66.817 54.186
403339 93.612 -66.823 54.166
403340 93.651 -66.831 54.147
403341 93.687 -66.832 54.129
403342 93.721 -66.831 54.115
403343 93.756 -66.830 54.104
403344 93.791 -66.829 54.095
403345 93.826 -66.828 54.088
403346 93.861 -66.827 54.082
403347 93.896 -66.827 54.073
403348 93.930 -66.826 54.063
403349 93.971 -66.829 54.048
403350 94.013 -66.834 54.031
403351 94.056 -66.839 54.014
403352 94.099 -66.844 53.995
403353 94.142 -66.849 53.979
403354 94.184 -66.853 53.964
403355 94.227 -66.859 53.949
403356 94.270 -66.863 53.932
403357 94.313 -66.868 53.912
403358 94.355 -66.873 53.889
403359 94.398 -66.878 53.863
403360 94.431 -66.879 53.834
403361 94.453 -66.876 53.804
403362 94.474 -66.873 53.773
403363 94.502 -66.871 53.743
403364 94.542 -66.873 53.714
403365 94.582 -66.874 53.684
403366 94.627 -66.877 53.654
403367 94.679 -66.879 53.623
403368 94.731 -66.881 53.590
403369 94.783 -66.884 53.555
403370 94.835 -66.886 53.521
403371 94.887 -66.890 53.489
403372 94.939 -66.899 53.460
403373 94.969 -66.892 53.435
403374 94.989 -66.879 53.414
403375 95.021 -66.878 53.396
403376 95.052 -66.876 53.381
403377 95.083 -66.874 53.367
403378 95.115 -66.873 53.354
403379 95.146 -66.871 53.344
403380 95.178 -66.870 53.336
403381 95.228 -66.873 53.329
403382 95.278 -66.875 53.322
403383 95.329 -66.878 53.313
403384 95.383 -66.892 53.301
403385 95.445 -66.932 53.283
403386 95.485 -66.928 53.262
403387 95.526 -66.924 53.240
403388 95.566 -66.920 53.220
403389 95.608 -66.923 53.199
403390 95.652 -66.932 53.179
403391 95.696 -66.941 53.160
403392 95.740 -66.950 53.141
403393 95.784 -66.959 53.120
403394 95.828 -66.969 53.098
403395 95.872 -66.978 53.077
403396 95.913 -66.986 53.055
403397 95.950 -66.992 53.032
403398 95.985 -66.996 53.008
403399 96.021 -67.000 52.983
403400 96.052 -66.998 52.956
403401 96.085 -66.997 52.929
403402 96.117 -66.995 52.902
403403 96.149 -66.994 52.873
403404 96.181 -66.993 52.844
403405 96.211 -66.999 52.814
403406 96.242 -67.006 52.782
403407 96.272 -67.012 52.750
403408 96.303 -67.019 52.717
403409 96.333 -67.025 52.684
403410 96.364 -67.031 52.653
403411 96.394 -67.037 52.622
403412 96.425 -67.044 52.593
403413 96.457 -67.045 52.565
403414 96.487 -67.044 52.537
403415 96.522 -67.043 52.511
403416 96.557 -67.043 52.488
403417 96.592 -67.042 52.468
403418 96.626 -67.042 52.449
403419 96.661 -67.042 52.430
403420 96.696 -67.041 52.411
403421 96.731 -67.041 52.392
403422 96.766 -67.041 52.374
403423 96.806 -67.041 52.355
403424 96.845 -67.041 52.335
403425 96.884 -67.041 52.313
403426 96.923 -67.041 52.291
403427 96.962 -67.041 52.271
403428 97.002 -67.041 52.252
403429 97.041 -67.041 52.233
403430 97.080 -67.041 52.214
403431 97.120 -67.041 52.197
403432 97.163 -67.044 52.179
403433 97.206 -67.047 52.162
403434 97.248 -67.049 52.145
403435 97.292 -67.052 52.125
403436 97.335 -67.055 52.103
403437 97.379 -67.058 52.077
403438 97.423 -67.061 52.047
403439 97.458 -67.068 52.015
403440 97.487 -67.075 51.981
403441 97.516 -67.082 51.945
403442 97.544 -67.090 51.909
403443 97.573 -67.097 51.874
403444 97.602 -67.104 51.839
403445 97.633 -67.110 51.802
403446 97.676 -67.109 51.766
403447 97.727 -67.105 51.730
403448 97.778 -67.100 51.693
403449 97.822 -67.103 51.656
403450 97.848 -67.113 51.619
403451 97.874 -67.123 51.580
403452 97.900 -67.134 51.540
403453 97.926 -67.144 51.497
403454 97.959 -67.154 51.455
403455 97.991 -67.164 51.413
403456 98.024 -67.175 51.369
403457 98.056 -67.185 51.324
403458 98.089 -67.196 51.277
403459 98.122 -67.207 51.230
403460 98.154 -67.217 51.183
403461 98.186 -67.228 51.136
403462 98.219 -67.238 51.087
403463 98.251 -67.248 51.039
403464 98.284 -67.259 50.991
403465 98.316 -67.265 50.945
403466 98.349 -67.268 50.900
403467 98.381 -67.271 50.857
403468 98.413 -67.275 50.814
403469 98.442 -67.280 50.772
403470 98.469 -67.287 50.730
403471 98.496 -67.293 50.686
403472 98.523 -67.300 50.643
403473 98.551 -67.307 50.600
403474 98.578 -67.313 50.557
403475 98.605 -67.319 50.516
403476 98.632 -67.326 50.473
403477 98.668 -67.335 50.427
403478 98.714 -67.341 50.380
403479 98.763 -67.341 50.331
403480 98.811 -67.343 50.281
403481 98.860 -67.344 50.230
403482 98.909 -67.345 50.178
403483 98.958 -67.347 50.126
403484 99.007 -67.348 50.073
403485 99.057 -67.351 50.020
403486 99.109 -67.359 49.968
403487 99.161 -67.366 49.915
403488 99.213 -67.373 49.861
403489 99.266 -67.381 49.808
403490 99.318 -67.388 49.756
403491 99.370 -67.396 49.705
403492 99.422 -67.403 49.655
403493 99.444 -67.416 49.606
403494 99.444 -67.433 49.557
403495 99.461 -67.450 49.509
403496 99.504 -67.467 49.462
403497 99.548 -67.485 49.415
403498 99.591 -67.501 49.368
403499 99.635 -67.519 49.323
403500 99.686 -67.514 49.276
403501 99.701 -67.545 49.230
403502 99.717 -67.577 49.183
403503 99.780 -67.604 49.138
403504 99.797 -67.595 49.092
403505 99.831 -67.581 49.046
403506 99.861 -67.574 49.000
403507 99.869 -67.583 48.955
403508 99.877 -67.591 48.909
403509 99.895 -67.599 48.864
403510 99.925 -67.602 48.819
403511 99.965 -67.600 48.774
403512 100.004 -67.598 48.729
403513 100.044 -67.596 48.686
403514 100.083 -67.594 48.643
403515 100.123 -67.592 48.600
403516 100.162 -67.590 48.559
403517 100.202 -67.588 48.518
403518 100.242 -67.586 48.477
403519 100.281 -67.584 48.436
403520 100.321 -67.582 48.394
403521 100.360 -67.580 48.354
403522 100.405 -67.578 48.314
403964 89.957 -66.779 54.908
403965 90.003 -66.784 54.910
403966 90.049 -66.787 54.913
403967 90.101 -66.779 54.919
403968 90.158 -66.762 54.926
403969 90.214 -66.744 54.933
403970 90.271 -66.727 54.941
403971 90.328 -66.709 54.948
403972 90.384 -66.696 54.955
403973 90.440 -66.700 54.963
403974 90.496 -66.705 54.972
403975 90.551 -66.708 54.981
403976 90.607 -66.713 54.991
403977 90.662 -66.718 55.000
403978 90.718 -66.721 55.009
403979 90.773 -66.726 55.020
403980 90.829 -66.730 55.032
403981 90.885 -66.734 55.049
403982 90.943 -66.741 55.069
403983 91.002 -66.748 55.091
403984 91.061 -66.755 55.108
403985 91.114 -66.762 55.120
403986 91.158 -66.771 55.131
403987 91.201 -66.778 55.140
403988 91.245 -66.787 55.148
403989 91.289 -66.795 55.149
403990 91.332 -66.802 55.147
403991 91.376 -66.811 55.141
403992 91.419 -66.816 55.134
403993 91.461 -66.809 55.125
403994 91.502 -66.802 55.113
403995 91.544 -66.796 55.100
403996 91.586 -66.788 55.087
403997 91.628 -66.781 55.074
403998 91.671 -66.780 55.061
403999 91.716 -66.784 55.044
404000 91.760 -66.788 55.026
404001 91.805 -66.791 55.006
404002 91.850 -66.795 54.985
404003 91.894 -66.799 54.965
404004 91.939 -66.802 54.945
404005 91.983 -66.805 54.926
404006 92.028 -66.809 54.907
404007 92.073 -66.813 54.888
404008 92.117 -66.819 54.871
404009 92.175 -66.829 54.853
404010 92.233 -66.839 54.835
404011 92.292 -66.849 54.818
404012 92.351 -66.859 54.800
404013 92.414 -66.869 54.785
404014 92.477 -66.880 54.774
404015 92.541 -66.890 54.769
404016 92.603 -66.901 54.770
404017 92.667 -66.911 54.771
404018 92.730 -66.921 54.768
404019 92.793 -66.932 54.760
404020 92.839 -66.920 54.747
404021 92.880 -66.903 54.728
404022 92.922 -66.885 54.706
404023 92.964 -66.868 54.681
404024 93.005 -66.851 54.654
404025 93.046 -66.834 54.622
404026 93.084 -66.819 54.588
404027 93.122 -66.805 54.553
404028 93.160 -66.791 54.517
404029 93.198 -66.777 54.482
404030 93.236 -66.763 54.448
404031 93.274 -66.750 54.415
404032 93.308 -66.746 54.384
404033 93.344 -66.747 54.357
404034 93.382 -66.751 54.332
404035 93.419 -66.755 54.306
404036 93.457 -66.759 54.281
404037 93.494 -66.763 54.254
404038 93.532 -66.767 54.228
404039 93.569 -66.771 54.203
404040 93.608 -66.777 54.180
404041 93.647 -66.786 54.159
404042 93.687 -66.795 54.142
404043 93.722 -66.794 54.129
404044 93.757 -66.793 54.119
404045 93.791 -66.793 54.111
404046 93.826 -66.792 54.105
404047 93.861 -66.791 54.098
404048 93.896 -66.790 54.089
404049 93.932 -66.784 54.078
404050 93.972 -66.789 54.064
404051 94.014 -66.793 54.047
404052 94.056 -66.798 54.029
404053 94.099 -66.802 54.011
404054 94.141 -66.808 53.995
404055 94.184 -66.812 53.981
404056 94.227 -66.817 53.966
404057 94.269 -66.822 53.951
404058 94.312 -66.827 53.931
404059 94.355 -66.832 53.908
404060 94.398 -66.837 53.882
404061 94.436 -66.840 53.853
404062 94.458 -66.837 53.822
404063 94.479 -66.834 53.789
404064 94.515 -66.834 53.758
404065 94.555 -66.836 53.728
404066 94.595 -66.837 53.698
404067 94.635 -66.840 53.667
404068 94.676 -66.841 53.635
404069 94.728 -66.843 53.600
404070 94.780 -66.846 53.564
404071 94.832 -66.848 53.530
404072 94.884 -66.852 53.498
404073 94.936 -66.859 53.469
404074 94.962 -66.850 53.445
404075 94.988 -66.843 53.424
404076 95.019 -66.841 53.407
404077 95.050 -66.840 53.392
404078 95.082 -66.838 53.380
404079 95.113 -66.837 53.369
404080 95.144 -66.835 53.359
404081 95.175 -66.834 53.352
404082 95.219 -66.834 53.346
404083 95.270 -66.837 53.339
404084 95.320 -66.840 53.329
404085 95.383 -66.878 53.315
404086 95.443 -66.914 53.294
404087 95.484 -66.912 53.272
404088 95.524 -66.909 53.251
404089 95.564 -66.905 53.231
404090 95.605 -66.901 53.212
404091 95.645 -66.896 53.193
404092 95.689 -66.905 53.175
404093 95.733 -66.914 53.157
404094 95.776 -66.923 53.138
404095 95.821 -66.932 53.118
404096 95.865 -66.941 53.096
404097 95.903 -66.948 53.073
404098 95.940 -66.954 53.049
404099 95.976 -66.959 53.025
404100 96.012 -66.963 52.999
404101 96.047 -66.967 52.973
404102 96.081 -66.967 52.945
404103 96.112 -66.965 52.918
404104 96.145 -66.964 52.890
404105 96.176 -66.962 52.863
404106 96.207 -66.967 52.833
404107 96.238 -66.973 52.802
404108 96.268 -66.980 52.769
404109 96.299 -66.986 52.736
404110 96.329 -66.993 52.703
404111 96.360 -66.999 52.672
404112 96.391 -67.004 52.641
404113 96.426 -67.006 52.612
404114 96.456 -67.006 52.583
404115 96.485 -67.004 52.555
404116 96.514 -67.002 52.527
404117 96.549 -67.002 52.503
404118 96.584 -67.002 52.482
404119 96.619 -67.001 52.462
404120 96.654 -67.001 52.443
404121 96.689 -67.001 52.423
404122 96.724 -67.000 52.406
404123 96.762 -67.001 52.388
404124 96.804 -67.003 52.369
404125 96.843 -67.003 52.349
404126 96.882 -67.003 52.327
404127 96.922 -67.002 52.305
404128 96.961 -67.002 52.283
404129 97.000 -67.002 52.263
404130 97.039 -67.002 52.243
404131 97.079 -67.002 52.223
404132 97.121 -67.004 52.206
404133 97.164 -67.007 52.188
404134 97.207 -67.009 52.171
404135 97.249 -67.012 52.156
404136 97.293 -67.015 52.137
404137 97.336 -67.018 52.115
404138 97.380 -67.021 52.088
404139 97.426 -67.019 52.058
404140 97.467 -67.023 52.025
404141 97.495 -67.031 51.990
404142 97.522 -67.038 51.953
404143 97.549 -67.046 51.916
404144 97.577 -67.053 51.881
404145 97.604 -67.060 51.845
404146 97.637 -67.065 51.808
404147 97.687 -67.060 51.771
404148 97.738 -67.056 51.734
404149 97.789 -67.053 51.696
404150 97.816 -67.062 51.657
404151 97.843 -67.072 51.618
404152 97.876 -67.082 51.578
404153 97.908 -67.093 51.536
404154 97.941 -67.103 51.493
404155 97.973 -67.114 51.450
404156 98.006 -67.125 51.406
404157 98.038 -67.135 51.363
404158 98.071 -67.145 51.318
404159 98.103 -67.156 51.271
404160 98.136 -67.166 51.224
404161 98.168 -67.177 51.177
404162 98.201 -67.188 51.130
404163 98.233 -67.198 51.082
404164 98.266 -67.208 51.034
404165 98.298 -67.219 50.988
404166 98.331 -67.226 50.944
404167 98.363 -67.229 50.901
404168 98.395 -67.233 50.859
404169 98.425 -67.238 50.817
404170 98.452 -67.244 50.775
404171 98.480 -67.251 50.732
404172 98.507 -67.257 50.689
404173 98.534 -67.264 50.645
404174 98.561 -67.271 50.600
404175 98.588 -67.278 50.557
404176 98.616 -67.284 50.515
404177 98.643 -67.291 50.472
404178 98.670 -67.297 50.427
404179 98.719 -67.299 50.380
404180 98.768 -67.300 50.332
404181 98.817 -67.301 50.281
404182 98.866 -67.303 50.229
404183 98.915 -67.304 50.177
404184 98.963 -67.305 50.124
404185 99.012 -67.306 50.070
404186 99.061 -67.307 50.017
404187 99.113 -67.314 49.964
404188 99.165 -67.322 49.909
404189 99.217 -67.329 49.855
404190 99.270 -67.336 49.802
404191 99.322 -67.344 49.750
404192 99.374 -67.351 49.698
404193 99.416 -67.360 49.648
404194 99.416 -67.377 49.599
404195 99.429 -67.395 49.550
404196 99.472 -67.411 49.501
404197 99.516 -67.429 49.453
404198 99.559 -67.446 49.407
404199 99.603 -67.464 49.361
404200 99.646 -67.481 49.315
404201 99.696 -67.480 49.268
404202 99.732 -67.489 49.220
404203 99.739 -67.529 49.173
404204 99.780 -67.545 49.127
404205 99.821 -67.529 49.081
404206 99.853 -67.522 49.034
404207 99.861 -67.530 48.988
404208 99.869 -67.539 48.942
404209 99.880 -67.547 48.897
404210 99.901 -67.555 48.851
404211 99.934 -67.556 48.806
404212 99.974 -67.555 48.762
404213 100.013 -67.552 48.717
404214 100.053 -67.550 48.674
404215 100.092 -67.548 48.631
404216 100.132 -67.546 48.589
404217 100.171 -67.544 48.548
404218 100.211 -67.542 48.507
404219 100.251 -67.540 48.467
404220 100.290 -67.538 48.426
404221 100.330 -67.536 48.384
404222 100.371 -67.534 48.344
404223 100.419 -67.532 48.305
404666 90.005 -66.730 54.909
404667 90.052 -66.732 54.912
404668 90.109 -66.715 54.918
404669 90.165 -66.697 54.924
404670 90.222 -66.680 54.930
404671 90.279 -66.662 54.937
404672 90.335 -66.644 54.945
404673 90.391 -66.646 54.952
404674 90.447 -66.651 54.959
404675 90.503 -66.655 54.967
404676 90.558 -66.659 54.976
404677 90.614 -66.664 54.985
404678 90.669 -66.668 54.994
404679 90.725 -66.672 55.003
404680 90.781 -66.677 55.014
404681 90.836 -66.680 55.028
404682 90.892 -66.685 55.043
404683 90.947 -66.690 55.063
404684 91.006 -66.696 55.086
404685 91.065 -66.703 55.105
404686 91.113 -66.711 55.119
404687 91.157 -66.718 55.131
404688 91.201 -66.727 55.141
404689 91.244 -66.735 55.152
404690 91.288 -66.743 55.155
404691 91.332 -66.751 55.155
404692 91.375 -66.759 55.151
404693 91.419 -66.767 55.143
404694 91.461 -66.762 55.133
404695 91.503 -66.755 55.119
404696 91.544 -66.748 55.106
404697 91.586 -66.741 55.093
404698 91.628 -66.734 55.078
404699 91.672 -66.738 55.064
404700 91.717 -66.741 55.047
404701 91.761 -66.745 55.028
404702 91.806 -66.749 55.009
404703 91.851 -66.752 54.990
404704 91.895 -66.755 54.971
404705 91.940 -66.759 54.953
404706 91.985 -66.762 54.935
404707 92.029 -66.768 54.917
404708 92.074 -66.774 54.899
404709 92.130 -66.783 54.882
404710 92.189 -66.793 54.865
404711 92.247 -66.803 54.847
404712 92.305 -66.813 54.832
404713 92.364 -66.823 54.815
404714 92.422 -66.833 54.801
404715 92.484 -66.843 54.791
404716 92.548 -66.854 54.785
404717 92.611 -66.864 54.786
404718 92.674 -66.874 54.790
404719 92.737 -66.885 54.789
404720 92.799 -66.894 54.783
404721 92.841 -66.877 54.772
404722 92.882 -66.859 54.755
404723 92.924 -66.842 54.732
404724 92.965 -66.824 54.706
404725 93.005 -66.809 54.678
404726 93.043 -66.795 54.646
404727 93.081 -66.780 54.612
404728 93.119 -66.766 54.577
404729 93.157 -66.752 54.541
404730 93.195 -66.738 54.505
404731 93.233 -66.724 54.471
404732 93.271 -66.710 54.437
404733 93.305 -66.705 54.406
404734 93.340 -66.701 54.378
404735 93.377 -66.705 54.353
404736 93.415 -66.709 54.328
404737 93.452 -66.713 54.301
404738 93.490 -66.718 54.273
404739 93.527 -66.721 54.245
404740 93.565 -66.725 54.219
404741 93.603 -66.731 54.195
404742 93.643 -66.740 54.175
404743 93.683 -66.749 54.159
404744 93.722 -66.758 54.146
404745 93.757 -66.757 54.136
404746 93.792 -66.756 54.128
404747 93.827 -66.755 54.121
404748 93.862 -66.751 54.114
404749 93.898 -66.746 54.105
404750 93.935 -66.743 54.093
404751 93.975 -66.748 54.079
404752 94.016 -66.752 54.062
404753 94.058 -66.757 54.044
404754 94.100 -66.762 54.026
404755 94.141 -66.766 54.010
404756 94.184 -66.771 53.996
404757 94.226 -66.776 53.982
404758 94.269 -66.780 53.966
404759 94.312 -66.786 53.948
404760 94.354 -66.790 53.926
404761 94.397 -66.796 53.901
404762 94.440 -66.800 53.872
404763 94.463 -66.798 53.840
404764 94.487 -66.796 53.806
404765 94.528 -66.797 53.773
404766 94.567 -66.799 53.742
404767 94.608 -66.800 53.712
404768 94.647 -66.802 53.680
404769 94.688 -66.804 53.647
404770 94.728 -66.805 53.611
404771 94.777 -66.808 53.574
404772 94.829 -66.810 53.539
404773 94.881 -66.812 53.507
404774 94.933 -66.820 53.479
404775 94.955 -66.808 53.456
404776 94.986 -66.806 53.436
404777 95.017 -66.805 53.419
404778 95.048 -66.803 53.405
404779 95.080 -66.802 53.394
404780 95.111 -66.800 53.384
404781 95.142 -66.798 53.376
404782 95.174 -66.796 53.369
404783 95.211 -66.796 53.363
404784 95.261 -66.799 53.355
404785 95.316 -66.815 53.343
404786 95.377 -66.853 53.325
404787 95.425 -66.866 53.303
404788 95.472 -66.878 53.280
404789 95.516 -66.881 53.260
404790 95.560 -66.884 53.241
404791 95.603 -66.885 53.223
404792 95.643 -66.881 53.206
404793 95.684 -66.878 53.189
404794 95.725 -66.878 53.172
404795 95.769 -66.887 53.156
404796 95.813 -66.896 53.137
404797 95.857 -66.905 53.116
404798 95.893 -66.911 53.093
404799 95.930 -66.917 53.069
404800 95.966 -66.923 53.044
404801 96.003 -66.927 53.018
404802 96.038 -66.931 52.992
404803 96.074 -66.934 52.964
404804 96.108 -66.936 52.937
404805 96.140 -66.935 52.910
404806 96.172 -66.934 52.883
404807 96.203 -66.935 52.853
404808 96.234 -66.941 52.822
404809 96.264 -66.947 52.790
404810 96.295 -66.954 52.757
404811 96.326 -66.960 52.724
404812 96.358 -66.965 52.693
404813 96.393 -66.967 52.662
404814 96.427 -66.968 52.632
404815 96.455 -66.966 52.602
404816 96.484 -66.965 52.573
404817 96.513 -66.962 52.546
404818 96.542 -66.961 52.520
404819 96.577 -66.961 52.498
404820 96.612 -66.960 52.478
404821 96.647 -66.960 52.457
404822 96.681 -66.960 52.436
404823 96.719 -66.960 52.420
404824 96.760 -66.962 52.402
404825 96.801 -66.963 52.383
404826 96.841 -66.963 52.364
404827 96.881 -66.963 52.342
404828 96.920 -66.963 52.317
404829 96.959 -66.963 52.295
404830 96.998 -66.962 52.274
404831 97.037 -66.962 52.254
404832 97.080 -66.965 52.234
404833 97.122 -66.968 52.216
404834 97.165 -66.970 52.199
404835 97.208 -66.972 52.184
404836 97.250 -66.975 52.169
404837 97.294 -66.978 52.151
404838 97.337 -66.981 52.128
404839 97.384 -66.977 52.100
404840 97.433 -66.971 52.067
404841 97.480 -66.978 52.033
404842 97.508 -66.986 51.998
404843 97.535 -66.994 51.962
404844 97.563 -67.000 51.925
404845 97.590 -67.008 51.888
404846 97.621 -67.013 51.852
404847 97.666 -67.010 51.814
404848 97.712 -67.007 51.776
404849 97.757 -67.004 51.738
404850 97.792 -67.012 51.699
404851 97.824 -67.022 51.659
404852 97.858 -67.031 51.618
404853 97.890 -67.042 51.576
404854 97.923 -67.053 51.533
404855 97.955 -67.063 51.489
404856 97.988 -67.074 51.444
404857 98.020 -67.084 51.400
404858 98.053 -67.095 51.356
404859 98.085 -67.105 51.311
404860 98.118 -67.116 51.263
404861 98.150 -67.126 51.217
404862 98.183 -67.137 51.169
404863 98.215 -67.147 51.122
404864 98.248 -67.158 51.076
404865 98.281 -67.169 51.031
404866 98.313 -67.178 50.989
404867 98.346 -67.188 50.947
404868 98.378 -67.191 50.905
404869 98.409 -67.195 50.863
404870 98.436 -67.202 50.822
404871 98.463 -67.209 50.780
404872 98.490 -67.215 50.736
404873 98.517 -67.222 50.691
404874 98.545 -67.229 50.648
404875 98.571 -67.235 50.602
404876 98.597 -67.237 50.558
404877 98.624 -67.238 50.515
404878 98.650 -67.241 50.472
404879 98.678 -67.256 50.427
404880 98.726 -67.257 50.380
404881 98.774 -67.258 50.331
404882 98.822 -67.260 50.280
404883 98.871 -67.260 50.227
404884 98.920 -67.262 50.174
404885 98.969 -67.263 50.121
404886 99.017 -67.264 50.067
404887 99.066 -67.266 50.012
404888 99.117 -67.269 49.958
404889 99.169 -67.277 49.904
404890 99.221 -67.285 49.850
404891 99.274 -67.291 49.797
404892 99.326 -67.299 49.744
404893 99.378 -67.307 49.692
404894 99.388 -67.322 49.641
404895 99.397 -67.338 49.592
404896 99.440 -67.356 49.543
404897 99.484 -67.373 49.493
404898 99.527 -67.391 49.445
404899 99.571 -67.407 49.399
404900 99.615 -67.425 49.352
404901 99.658 -67.442 49.306
404902 99.706 -67.446 49.259
404903 99.758 -67.440 49.211
404904 99.770 -67.474 49.163
404905 99.811 -67.479 49.116
404906 99.845 -67.469 49.070
404907 99.853 -67.478 49.023
404908 99.861 -67.486 48.976
404909 99.869 -67.495 48.930
404910 99.886 -67.503 48.884
404911 99.907 -67.511 48.839
404912 99.943 -67.511 48.793
404913 99.983 -67.508 48.749
404914 100.022 -67.506 48.705
404915 100.062 -67.504 48.663
404916 100.102 -67.502 48.620
404917 100.141 -67.500 48.579
404918 100.181 -67.498 48.539
404919 100.220 -67.496 48.497
404920 100.260 -67.494 48.457
404921 100.299 -67.492 48.416
404922 100.339 -67.490 48.375
404923 100.385 -67.488 48.334
405368 90.060 -66.668 54.911
405369 90.116 -66.650 54.916
405370 90.173 -66.633 54.922
405371 90.230 -66.615 54.927
405372 90.287 -66.598 54.934
405373 90.343 -66.592 54.941
405374 90.398 -66.597 54.947
405375 90.454 -66.601 54.954
405376 90.510 -66.605 54.962
405377 90.565 -66.610 54.971
405378 90.621 -66.614 54.980
405379 90.676 -66.618 54.989
405380 90.732 -66.623 54.999
405381 90.788 -66.627 55.009
405382 90.843 -66.631 55.022
405383 90.899 -66.636 55.037
405384 90.954 -66.639 55.055
405385 91.010 -66.644 55.076
405386 91.069 -66.651 55.097
405387 91.113 -66.659 55.113
405388 91.157 -66.667 55.127
405389 91.201 -66.675 55.139
405390 91.244 -66.683 55.152
405391 91.288 -66.691 55.158
405392 91.331 -66.699 55.160
405393 91.375 -66.707 55.158
405394 91.419 -66.715 55.151
405395 91.461 -66.715 55.139
405396 91.503 -66.708 55.125
405397 91.544 -66.701 55.111
405398 91.586 -66.694 55.097
405399 91.629 -66.691 55.081
405400 91.674 -66.695 55.066
405401 91.718 -66.699 55.049
405402 91.763 -66.702 55.030
405403 91.808 -66.705 55.012
405404 91.852 -66.709 54.994
405405 91.897 -66.712 54.976
405406 91.942 -66.717 54.960
405407 91.986 -66.723 54.944
405408 92.030 -66.728 54.927
405409 92.085 -66.737 54.911
405410 92.143 -66.747 54.895
405411 92.202 -66.757 54.878
405412 92.260 -66.767 54.863
405413 92.318 -66.777 54.848
405414 92.377 -66.787 54.832
405415 92.435 -66.797 54.819
405416 92.494 -66.807 54.808
405417 92.555 -66.817 54.801
405418 92.618 -66.827 54.799
405419 92.682 -66.838 54.801
405420 92.744 -66.849 54.802
405421 92.801 -66.850 54.797
405422 92.843 -66.833 54.786
405423 92.884 -66.816 54.770
405424 92.925 -66.799 54.747
405425 92.964 -66.784 54.722
405426 93.002 -66.770 54.695
405427 93.040 -66.755 54.664
405428 93.078 -66.742 54.632
405429 93.116 -66.727 54.597
405430 93.154 -66.713 54.561
405431 93.192 -66.699 54.526
405432 93.230 -66.685 54.493
405433 93.268 -66.671 54.459
405434 93.303 -66.664 54.428
405435 93.337 -66.659 54.399
405436 93.373 -66.659 54.373
405437 93.411 -66.663 54.348
405438 93.448 -66.668 54.320
405439 93.485 -66.671 54.291
405440 93.523 -66.675 54.262
405441 93.560 -66.680 54.235
405442 93.599 -66.686 54.211
405443 93.638 -66.694 54.192
405444 93.678 -66.703 54.177
405445 93.718 -66.712 54.165
405446 93.758 -66.721 54.155
405447 93.793 -66.718 54.146
405448 93.828 -66.712 54.138
405449 93.864 -66.708 54.130
405450 93.900 -66.702 54.119
405451 93.938 -66.703 54.107
405452 93.979 -66.708 54.092
405453 94.019 -66.711 54.076
405454 94.060 -66.717 54.058
405455 94.102 -66.721 54.040
405456 94.144 -66.726 54.025
405457 94.186 -66.730 54.010
405458 94.227 -66.735 53.996
405459 94.269 -66.740 53.980
405460 94.312 -66.744 53.962
405461 94.354 -66.749 53.940
405462 94.397 -66.754 53.915
405463 94.439 -66.759 53.887
405464 94.469 -66.758 53.856
405465 94.500 -66.758 53.822
405466 94.540 -66.760 53.788
405467 94.580 -66.762 53.756
405468 94.620 -66.763 53.725
405469 94.660 -66.765 53.693
405470 94.700 -66.767 53.659
405471 94.740 -66.768 53.622
405472 94.780 -66.770 53.585
405473 94.827 -66.772 53.549
405474 94.879 -66.774 53.519
405475 94.927 -66.778 53.491
405476 94.953 -66.771 53.468
405477 94.984 -66.770 53.448
405478 95.016 -66.768 53.431
405479 95.047 -66.766 53.419
405480 95.078 -66.765 53.408
405481 95.109 -66.763 53.399
405482 95.140 -66.762 53.393
405483 95.172 -66.760 53.387
405484 95.203 -66.758 53.381
405485 95.252 -66.761 53.370
405486 95.311 -66.792 53.353
405487 95.359 -66.805 53.332
405488 95.407 -66.818 53.309
405489 95.456 -66.832 53.288
405490 95.502 -66.842 53.269
405491 95.546 -66.845 53.251
405492 95.590 -66.848 53.234
405493 95.634 -66.851 53.218
405494 95.678 -66.854 53.202
405495 95.722 -66.857 53.187
405496 95.763 -66.854 53.172
405497 95.806 -66.859 53.155
405498 95.847 -66.867 53.136
405499 95.884 -66.873 53.114
405500 95.920 -66.879 53.089
405501 95.957 -66.885 53.065
405502 95.993 -66.890 53.038
405503 96.029 -66.894 53.011
405504 96.064 -66.898 52.983
405505 96.100 -66.902 52.956
405506 96.136 -66.906 52.930
405507 96.168 -66.904 52.903
405508 96.200 -66.903 52.874
405509 96.230 -66.909 52.844
405510 96.261 -66.915 52.811
405511 96.292 -66.921 52.778
405512 96.325 -66.925 52.745
405513 96.360 -66.928 52.714
405514 96.395 -66.930 52.683
405515 96.426 -66.929 52.652
405516 96.454 -66.927 52.622
405517 96.483 -66.925 52.593
405518 96.512 -66.924 52.565
405519 96.541 -66.921 52.540
405520 96.570 -66.920 52.517
405521 96.604 -66.919 52.495
405522 96.639 -66.919 52.473
405523 96.675 -66.919 52.451
405524 96.716 -66.921 52.433
405525 96.757 -66.921 52.416
405526 96.799 -66.923 52.398
405527 96.839 -66.925 52.378
405528 96.879 -66.924 52.357
405529 96.918 -66.924 52.331
405530 96.957 -66.924 52.308
405531 96.996 -66.924 52.286
405532 97.038 -66.926 52.265
405533 97.081 -66.928 52.244
405534 97.124 -66.931 52.227
405535 97.166 -66.933 52.211
405536 97.209 -66.935 52.195
405537 97.252 -66.937 52.180
405538 97.295 -66.940 52.163
405539 97.342 -66.935 52.138
405540 97.392 -66.924 52.110
405541 97.441 -66.925 52.078
405542 97.490 -66.934 52.043
405543 97.521 -66.941 52.006
405544 97.549 -66.948 51.970
405545 97.576 -66.956 51.933
405546 97.605 -66.962 51.895
405547 97.651 -66.959 51.857
405548 97.696 -66.956 51.819
405549 97.734 -66.961 51.780
405550 97.768 -66.971 51.740
405551 97.802 -66.981 51.700
405552 97.835 -66.990 51.660
405553 97.868 -67.000 51.618
405554 97.901 -67.009 51.575
405555 97.934 -67.019 51.531
405556 97.967 -67.028 51.486
405557 98.000 -67.038 51.441
405558 98.033 -67.047 51.396
405559 98.066 -67.057 51.351
405560 98.099 -67.066 51.305
405561 98.132 -67.075 51.257
405562 98.165 -67.086 51.211
405563 98.198 -67.097 51.165
405564 98.230 -67.107 51.119
405565 98.263 -67.118 51.074
405566 98.295 -67.128 51.032
405567 98.327 -67.139 50.992
405568 98.360 -67.149 50.951
405569 98.392 -67.153 50.911
405570 98.419 -67.160 50.868
405571 98.446 -67.166 50.827
405572 98.473 -67.172 50.784
405573 98.499 -67.174 50.740
405574 98.525 -67.175 50.695
405575 98.552 -67.178 50.650
405576 98.578 -67.179 50.604
405577 98.604 -67.181 50.560
405578 98.630 -67.183 50.516
405579 98.655 -67.192 50.471
405580 98.687 -67.214 50.426
405581 98.734 -67.216 50.377
405582 98.781 -67.216 50.327
405583 98.829 -67.218 50.276
405584 98.876 -67.219 50.222
405585 98.925 -67.220 50.170
405586 98.974 -67.222 50.116
405587 99.023 -67.222 50.062
405588 99.072 -67.224 50.007
405589 99.121 -67.225 49.953
405590 99.173 -67.232 49.899
405591 99.225 -67.240 49.845
405592 99.277 -67.247 49.792
405593 99.330 -67.254 49.739
405594 99.360 -67.266 49.686
405595 99.365 -67.283 49.635
405596 99.409 -67.301 49.585
405597 99.452 -67.317 49.534
405598 99.496 -67.335 49.486
405599 99.540 -67.352 49.437
405600 99.583 -67.370 49.390
405601 99.627 -67.386 49.344
405602 99.670 -67.404 49.297
405603 99.717 -67.412 49.249
405604 99.768 -67.406 49.201
405605 99.802 -67.418 49.153
405606 99.837 -67.417 49.106
405607 99.845 -67.426 49.058
405608 99.853 -67.434 49.011
405609 99.861 -67.442 48.964
405610 99.871 -67.451 48.917
405611 99.892 -67.459 48.871
405612 99.913 -67.467 48.826
405613 99.952 -67.464 48.781
405614 99.992 -67.462 48.738
405615 100.031 -67.460 48.695
405616 100.071 -67.458 48.652
405617 100.111 -67.456 48.610
405618 100.150 -67.454 48.569
405619 100.190 -67.452 48.528
405620 100.229 -67.450 48.487
405621 100.269 -67.448 48.446
405622 100.309 -67.446 48.406
405623 100.351 -67.444 48.365
405624 100.400 -67.442 48.324
406070 90.124 -66.586 54.915
406071 90.181 -66.568 54.920
406072 90.238 -66.551 54.925
406073 90.294 -66.539 54.931
406074 90.350 -66.543 54.937
406075 90.405 -66.547 54.944
406076 90.461 -66.552 54.951
406077 90.517 -66.556 54.959
406078 90.572 -66.560 54.967
406079 90.628 -66.564 54.975
406080 90.683 -66.569 54.984
406081 90.739 -66.573 54.994
406082 90.795 -66.577 55.004
406083 90.850 -66.582 55.016
406084 90.906 -66.586 55.030
406085 90.961 -66.590 55.046
406086 91.017 -66.595 55.064
406087 91.070 -66.599 55.084
406088 91.113 -66.607 55.101
406089 91.157 -66.615 55.116
406090 91.200 -66.624 55.130
406091 91.244 -66.631 55.144
406092 91.288 -66.639 55.153
406093 91.331 -66.648 55.157
406094 91.375 -66.655 55.157
406095 91.419 -66.664 55.152
406096 91.461 -66.668 55.142
406097 91.503 -66.661 55.128
406098 91.545 -66.654 55.113
406099 91.587 -66.646 55.098
406100 91.631 -66.649 55.082
406101 91.675 -66.652 55.067
406102 91.720 -66.655 55.050
406103 91.765 -66.659 55.032
406104 91.809 -66.662 55.014
406105 91.854 -66.666 54.997
406106 91.898 -66.672 54.981
406107 91.943 -66.677 54.967
406108 91.988 -66.683 54.953
406109 92.040 -66.691 54.938
406110 92.098 -66.702 54.923
406111 92.157 -66.711 54.908
406112 92.215 -66.721 54.893
406113 92.274 -66.731 54.879
406114 92.332 -66.741 54.866
406115 92.390 -66.751 54.851
406116 92.448 -66.761 54.838
406117 92.507 -66.771 54.827
406118 92.565 -66.781 54.819
406119 92.625 -66.791 54.814
406120 92.689 -66.802 54.812
406121 92.752 -66.812 54.810
406122 92.803 -66.807 54.804
406123 92.844 -66.790 54.793
406124 92.884 -66.774 54.776
406125 92.922 -66.759 54.755
406126 92.960 -66.746 54.731
406127 92.999 -66.731 54.705
406128 93.037 -66.717 54.676
406129 93.075 -66.703 54.646
406130 93.113 -66.689 54.614
406131 93.151 -66.674 54.581
406132 93.189 -66.661 54.547
406133 93.227 -66.646 54.515
406134 93.265 -66.632 54.482
406135 93.300 -66.624 54.452
406136 93.335 -66.619 54.423
406137 93.369 -66.614 54.395
406138 93.406 -66.617 54.367
406139 93.443 -66.621 54.338
406140 93.481 -66.626 54.310
406141 93.518 -66.630 54.281
406142 93.556 -66.633 54.253
406143 93.594 -66.640 54.229
406144 93.634 -66.649 54.210
406145 93.674 -66.658 54.196
406146 93.714 -66.666 54.185
406147 93.754 -66.674 54.174
406148 93.794 -66.674 54.165
406149 93.830 -66.669 54.156
406150 93.866 -66.664 54.146
406151 93.902 -66.658 54.134
406152 93.941 -66.662 54.121
406153 93.982 -66.667 54.106
406154 94.022 -66.671 54.089
406155 94.063 -66.676 54.072
406156 94.105 -66.680 54.055
406157 94.146 -66.685 54.040
406158 94.188 -66.690 54.025
406159 94.229 -66.694 54.010
406160 94.271 -66.699 53.995
406161 94.313 -66.704 53.975
406162 94.355 -66.708 53.953
406163 94.397 -66.713 53.928
406164 94.439 -66.718 53.900
406165 94.474 -66.720 53.870
406166 94.512 -66.721 53.837
406167 94.553 -66.723 53.802
406168 94.593 -66.724 53.768
406169 94.633 -66.726 53.737
406170 94.673 -66.727 53.704
406171 94.713 -66.729 53.669
406172 94.753 -66.731 53.633
406173 94.793 -66.733 53.596
406174 94.833 -66.734 53.561
406175 94.876 -66.736 53.531
406176 94.920 -66.736 53.504
406177 94.951 -66.734 53.482
406178 94.982 -66.733 53.463
406179 95.014 -66.731 53.448
406180 95.045 -66.730 53.436
406181 95.076 -66.728 53.425
406182 95.108 -66.727 53.417
406183 95.139 -66.725 53.410
406184 95.170 -66.724 53.404
406185 95.201 -66.722 53.394
406186 95.245 -66.730 53.379
406187 95.293 -66.744 53.360
406188 95.341 -66.758 53.336
406189 95.390 -66.771 53.313
406190 95.438 -66.784 53.294
406191 95.486 -66.798 53.276
406192 95.532 -66.805 53.260
406193 95.576 -66.809 53.244
406194 95.620 -66.812 53.230
406195 95.664 -66.815 53.215
406196 95.708 -66.818 53.201
406197 95.752 -66.821 53.187
406198 95.796 -66.824 53.171
406199 95.837 -66.830 53.154
406200 95.874 -66.836 53.134
406201 95.910 -66.842 53.111
406202 95.947 -66.848 53.088
406203 95.983 -66.853 53.060
406204 96.019 -66.858 53.032
406205 96.055 -66.862 53.003
406206 96.091 -66.865 52.974
406207 96.127 -66.869 52.949
406208 96.163 -66.873 52.923
406209 96.195 -66.873 52.896
406210 96.227 -66.877 52.866
406211 96.257 -66.884 52.832
406212 96.292 -66.886 52.800
406213 96.327 -66.888 52.767
406214 96.362 -66.890 52.735
406215 96.397 -66.891 52.703
406216 96.426 -66.890 52.673
406217 96.454 -66.888 52.642
406218 96.482 -66.886 52.612
406219 96.511 -66.884 52.584
406220 96.540 -66.882 52.560
406221 96.569 -66.881 52.536
406222 96.598 -66.879 52.514
406223 96.631 -66.878 52.491
406224 96.672 -66.880 52.468
406225 96.714 -66.881 52.448
406226 96.755 -66.882 52.429
406227 96.796 -66.884 52.411
406228 96.837 -66.885 52.392
406229 96.877 -66.885 52.372
406230 96.916 -66.885 52.345
406231 96.955 -66.885 52.321
406232 96.997 -66.887 52.298
406233 97.039 -66.889 52.276
406234 97.082 -66.891 52.256
406235 97.125 -66.893 52.238
406236 97.167 -66.896 52.222
406237 97.210 -66.898 52.206
406238 97.253 -66.900 52.191
406239 97.300 -66.893 52.173
406240 97.350 -66.882 52.148
406241 97.400 -66.872 52.120
406242 97.449 -66.881 52.088
406243 97.498 -66.888 52.053
406244 97.535 -66.896 52.016
406245 97.563 -66.903 51.979
406246 97.590 -66.911 51.941
406247 97.635 -66.908 51.902
406248 97.677 -66.909 51.863
406249 97.711 -66.918 51.823
406250 97.744 -66.929 51.782
406251 97.778 -66.939 51.741
406252 97.812 -66.950 51.700
406253 97.845 -66.959 51.659
406254 97.878 -66.968 51.617
406255 97.911 -66.978 51.574
406256 97.944 -66.987 51.529
406257 97.978 -66.997 51.484
406258 98.010 -67.006 51.439
406259 98.043 -67.016 51.393
406260 98.077 -67.025 51.347
406261 98.109 -67.034 51.300
406262 98.143 -67.044 51.253
406263 98.175 -67.053 51.207
406264 98.209 -67.063 51.161
406265 98.242 -67.072 51.116
406266 98.275 -67.082 51.072
406267 98.308 -67.091 51.033
406268 98.341 -67.101 50.995
406269 98.373 -67.110 50.956
406270 98.401 -67.111 50.916
406271 98.427 -67.113 50.874
406272 98.453 -67.114 50.832
406273 98.480 -67.116 50.789
406274 98.506 -67.118 50.744
406275 98.532 -67.120 50.699
406276 98.558 -67.122 50.653
406277 98.584 -67.123 50.607
406278 98.610 -67.125 50.561
406279 98.637 -67.127 50.516
406280 98.659 -67.157 50.470
406281 98.695 -67.173 50.423
406282 98.742 -67.174 50.373
406283 98.790 -67.175 50.322
406284 98.837 -67.176 50.271
406285 98.884 -67.177 50.218
406286 98.931 -67.178 50.165
406287 98.980 -67.179 50.111
406288 99.028 -67.181 50.056
406289 99.077 -67.182 50.001
406290 99.126 -67.183 49.947
406291 99.177 -67.188 49.893
406292 99.229 -67.195 49.839
406293 99.281 -67.203 49.786
406294 99.332 -67.210 49.733
406295 99.334 -67.227 49.681
406296 99.377 -67.244 49.630
406297 99.421 -67.262 49.579
406298 99.464 -67.279 49.529
406299 99.508 -67.296 49.479
406300 99.551 -67.313 49.431
406301 99.595 -67.331 49.383
406302 99.638 -67.348 49.335
406303 99.682 -67.366 49.287
406304 99.727 -67.378 49.239
406305 99.779 -67.372 49.191
406306 99.828 -67.365 49.143
406307 99.837 -67.373 49.095
406308 99.845 -67.382 49.047
406309 99.853 -67.390 49.000
406310 99.861 -67.399 48.952
406311 99.877 -67.407 48.905
406312 99.897 -67.415 48.859
406313 99.922 -67.420 48.814
406314 99.962 -67.418 48.770
406315 100.001 -67.417 48.727
406316 100.041 -67.414 48.684
406317 100.080 -67.412 48.641
406318 100.120 -67.410 48.599
406319 100.159 -67.408 48.558
406320 100.199 -67.406 48.517
406321 100.239 -67.404 48.476
406322 100.278 -67.402 48.436
406323 100.318 -67.400 48.396
406324 100.366 -67.398 48.355
406772 90.189 -66.504 54.918
406773 90.246 -66.486 54.923
406774 90.301 -66.489 54.930
406775 90.357 -66.493 54.936
406776 90.412 -66.498 54.942
406777 90.468 -66.502 54.949
406778 90.524 -66.506 54.955
406779 90.579 -66.511 54.964
406780 90.635 -66.515 54.972
406781 90.690 -66.519 54.980
406782 90.746 -66.523 54.989
406783 90.802 -66.528 54.999
406784 90.858 -66.532 55.010
406785 90.913 -66.536 55.023
406786 90.968 -66.541 55.037
406787 91.021 -66.548 55.052
406788 91.069 -66.557 55.071
406789 91.115 -66.564 55.088
406790 91.157 -66.564 55.103
406791 91.200 -66.572 55.118
406792 91.244 -66.580 55.131
406793 91.287 -66.588 55.141
406794 91.331 -66.596 55.147
406795 91.375 -66.604 55.149
406796 91.418 -66.612 55.147
406797 91.462 -66.620 55.138
406798 91.504 -66.614 55.126
406799 91.545 -66.606 55.113
406800 91.587 -66.602 55.098
406801 91.632 -66.605 55.084
406802 91.677 -66.609 55.068
406803 91.721 -66.612 55.051
406804 91.766 -66.616 55.034
406805 91.811 -66.621 55.018
406806 91.855 -66.627 55.002
406807 91.899 -66.632 54.987
406808 91.944 -66.638 54.975
406809 91.995 -66.646 54.962
406810 92.054 -66.655 54.950
406811 92.112 -66.665 54.936
406812 92.170 -66.675 54.922
406813 92.228 -66.686 54.909
406814 92.287 -66.696 54.897
406815 92.345 -66.705 54.886
406816 92.404 -66.715 54.872
406817 92.462 -66.725 54.859
406818 92.520 -66.735 54.847
406819 92.579 -66.745 54.838
406820 92.637 -66.755 54.831
406821 92.696 -66.765 54.825
406822 92.759 -66.776 54.819
406823 92.804 -66.764 54.811
406824 92.843 -66.749 54.799
406825 92.881 -66.735 54.782
406826 92.919 -66.721 54.763
406827 92.957 -66.706 54.740
406828 92.995 -66.693 54.716
406829 93.034 -66.678 54.689
406830 93.072 -66.664 54.662
406831 93.110 -66.650 54.634
406832 93.148 -66.636 54.604
406833 93.186 -66.621 54.573
406834 93.224 -66.608 54.541
406835 93.262 -66.593 54.509
406836 93.298 -66.583 54.478
406837 93.332 -66.579 54.447
406838 93.366 -66.574 54.417
406839 93.402 -66.572 54.387
406840 93.439 -66.576 54.357
406841 93.477 -66.580 54.327
406842 93.514 -66.583 54.297
406843 93.552 -66.588 54.270
406844 93.590 -66.594 54.248
406845 93.630 -66.603 54.231
406846 93.670 -66.611 54.217
406847 93.709 -66.621 54.208
406848 93.751 -66.625 54.197
406849 93.794 -66.627 54.187
406850 93.832 -66.625 54.177
406851 93.867 -66.620 54.166
406852 93.905 -66.617 54.153
406853 93.945 -66.622 54.139
406854 93.985 -66.627 54.124
406855 94.026 -66.631 54.106
406856 94.066 -66.635 54.088
406857 94.107 -66.639 54.071
406858 94.148 -66.645 54.056
406859 94.190 -66.649 54.041
406860 94.232 -66.654 54.026
406861 94.274 -66.658 54.010
406862 94.316 -66.663 53.989
406863 94.357 -66.668 53.965
406864 94.399 -66.672 53.940
406865 94.440 -66.677 53.913
406866 94.480 -66.680 53.883
406867 94.521 -66.682 53.850
406868 94.560 -66.683 53.815
406869 94.601 -66.685 53.780
406870 94.640 -66.687 53.747
406871 94.681 -66.689 53.714
406872 94.720 -66.690 53.679
406873 94.761 -66.692 53.643
406874 94.801 -66.693 53.607
406875 94.841 -66.696 53.574
406876 94.881 -66.697 53.545
406877 94.920 -66.703 53.519
406878 94.953 -66.705 53.497
406879 94.985 -66.704 53.480
406880 95.017 -66.703 53.466
406881 95.048 -66.702 53.454
406882 95.080 -66.702 53.444
406883 95.112 -66.701 53.436
406884 95.143 -66.700 53.428
406885 95.175 -66.700 53.418
406886 95.207 -66.699 53.404
406887 95.229 -66.694 53.386
406888 95.276 -66.699 53.365
406889 95.324 -66.709 53.340
406890 95.372 -66.723 53.318
406891 95.420 -66.737 53.299
406892 95.469 -66.750 53.283
406893 95.517 -66.764 53.269
406894 95.562 -66.769 53.255
406895 95.606 -66.772 53.242
406896 95.650 -66.775 53.228
406897 95.694 -66.778 53.214
406898 95.738 -66.782 53.201
406899 95.783 -66.787 53.186
406900 95.827 -66.794 53.169
406901 95.864 -66.799 53.151
406902 95.901 -66.804 53.130
406903 95.937 -66.810 53.107
406904 95.974 -66.816 53.081
406905 96.010 -66.821 53.053
406906 96.046 -66.824 53.025
406907 96.082 -66.829 52.996
406908 96.117 -66.833 52.969
406909 96.153 -66.837 52.944
406910 96.189 -66.840 52.917
406911 96.224 -66.844 52.887
406912 96.258 -66.845 52.854
406913 96.293 -66.846 52.821
406914 96.327 -66.846 52.788
406915 96.362 -66.847 52.756
406916 96.398 -66.849 52.725
406917 96.430 -66.850 52.694
406918 96.458 -66.849 52.662
406919 96.486 -66.847 52.632
406920 96.515 -66.845 52.606
406921 96.543 -66.843 52.580
406922 96.572 -66.841 52.557
406923 96.600 -66.840 52.535
406924 96.630 -66.837 52.512
406925 96.670 -66.840 52.486
406926 96.711 -66.841 52.463
406927 96.752 -66.843 52.442
406928 96.794 -66.844 52.423
406929 96.835 -66.845 52.405
406930 96.875 -66.846 52.384
406931 96.914 -66.846 52.358
406932 96.955 -66.847 52.333
406933 96.998 -66.849 52.310
406934 97.041 -66.852 52.290
406935 97.083 -66.854 52.269
406936 97.127 -66.853 52.251
406937 97.173 -66.849 52.233
406938 97.219 -66.846 52.217
406939 97.264 -66.841 52.201
406940 97.310 -66.837 52.182
406941 97.358 -66.828 52.158
406942 97.407 -66.827 52.130
406943 97.456 -66.835 52.098
406944 97.505 -66.843 52.063
406945 97.548 -66.851 52.027
406946 97.576 -66.859 51.990
406947 97.619 -66.856 51.950
406948 97.653 -66.866 51.908
406949 97.687 -66.877 51.869
406950 97.721 -66.887 51.827
406951 97.754 -66.897 51.785
406952 97.788 -66.907 51.742
406953 97.822 -66.918 51.699
406954 97.856 -66.928 51.657
406955 97.888 -66.937 51.616
406956 97.922 -66.946 51.573
406957 97.954 -66.956 51.528
406958 97.988 -66.965 51.483
406959 98.021 -66.975 51.437
406960 98.054 -66.984 51.391
406961 98.087 -66.994 51.344
406962 98.120 -67.003 51.297
406963 98.153 -67.012 51.250
406964 98.186 -67.022 51.204
406965 98.219 -67.031 51.158
406966 98.252 -67.041 51.113
406967 98.285 -67.050 51.072
406968 98.314 -67.059 51.035
406969 98.340 -67.066 50.998
406970 98.365 -67.071 50.960
406971 98.390 -67.069 50.921
406972 98.421 -67.067 50.879
406973 98.451 -67.066 50.837
406974 98.482 -67.064 50.793
406975 98.512 -67.062 50.748
406976 98.538 -67.064 50.702
406977 98.565 -67.066 50.656
406978 98.591 -67.068 50.609
406979 98.617 -67.069 50.562
406980 98.641 -67.087 50.515
406981 98.663 -67.121 50.469
406982 98.704 -67.131 50.419
406983 98.750 -67.132 50.368
406984 98.798 -67.134 50.317
406985 98.845 -67.135 50.265
406986 98.892 -67.136 50.213
406987 98.939 -67.137 50.159
406988 98.987 -67.138 50.105
406989 99.034 -67.139 50.051
406990 99.083 -67.141 49.996
406991 99.132 -67.141 49.940
406992 99.181 -67.144 49.886
406993 99.233 -67.150 49.834
406994 99.285 -67.158 49.781
406995 99.304 -67.172 49.729
406996 99.345 -67.189 49.677
406997 99.389 -67.206 49.625
406998 99.432 -67.223 49.574
406999 99.476 -67.241 49.525
407000 99.519 -67.258 49.475
407001 99.563 -67.276 49.426
407002 99.607 -67.292 49.377
407003 99.650 -67.310 49.327
407004 99.694 -67.327 49.278
407005 99.737 -67.344 49.229
407006 99.789 -67.338 49.179
407007 99.826 -67.325 49.132
407008 99.837 -67.329 49.085
407009 99.845 -67.338 49.037
407010 99.853 -67.346 48.989
407011 99.862 -67.355 48.941
407012 99.883 -67.363 48.893
407013 99.903 -67.370 48.847
407014 99.931 -67.374 48.804
407015 99.971 -67.373 48.760
407016 100.010 -67.370 48.716
407017 100.050 -67.368 48.674
407018 100.089 -67.366 48.630
407019 100.129 -67.364 48.589
407020 100.168 -67.362 48.548
407021 100.208 -67.360 48.507
407022 100.248 -67.358 48.467
407023 100.287 -67.356 48.425
407024 100.332 -67.354 48.386
407474 90.253 -66.436 54.923
407475 90.308 -66.439 54.930
407476 90.364 -66.444 54.936
407477 90.419 -66.448 54.942
407478 90.475 -66.452 54.948
407479 90.531 -66.457 54.955
407480 90.586 -66.461 54.962
407481 90.642 -66.465 54.970
407482 90.697 -66.470 54.978
407483 90.753 -66.474 54.986
407484 90.809 -66.478 54.994
407485 90.865 -66.483 55.005
407486 90.920 -66.487 55.016
407487 90.971 -66.495 55.028
407488 91.020 -66.504 55.041
407489 91.068 -66.514 55.057
407490 91.117 -66.523 55.073
407491 91.161 -66.527 55.088
407492 91.203 -66.528 55.102
407493 91.244 -66.530 55.115
407494 91.287 -66.536 55.125
407495 91.330 -66.544 55.133
407496 91.374 -66.552 55.137
407497 91.418 -66.561 55.137
407498 91.461 -66.568 55.130
407499 91.504 -66.566 55.121
407500 91.545 -66.559 55.110
407501 91.589 -66.559 55.096
407502 91.634 -66.562 55.083
407503 91.678 -66.566 55.068
407504 91.723 -66.570 55.052
407505 91.768 -66.576 55.036
407506 91.812 -66.581 55.021
407507 91.857 -66.587 55.007
407508 91.901 -66.592 54.994
407509 91.950 -66.600 54.983
407510 92.008 -66.610 54.972
407511 92.067 -66.620 54.960
407512 92.125 -66.630 54.948
407513 92.183 -66.639 54.935
407514 92.242 -66.649 54.924
407515 92.300 -66.659 54.915
407516 92.359 -66.670 54.907
407517 92.417 -66.680 54.894
407518 92.476 -66.690 54.881
407519 92.533 -66.699 54.868
407520 92.592 -66.709 54.859
407521 92.650 -66.719 54.854
407522 92.709 -66.730 54.847
407523 92.763 -66.737 54.837
407524 92.802 -66.724 54.823
407525 92.840 -66.710 54.809
407526 92.878 -66.696 54.793
407527 92.916 -66.682 54.774
407528 92.954 -66.668 54.753
407529 92.992 -66.653 54.730
407530 93.030 -66.639 54.704
407531 93.069 -66.625 54.681
407532 93.107 -66.611 54.657
407533 93.145 -66.597 54.632
407534 93.183 -66.583 54.603
407535 93.221 -66.568 54.570
407536 93.259 -66.555 54.537
407537 93.296 -66.543 54.504
407538 93.330 -66.538 54.471
407539 93.364 -66.533 54.439
407540 93.398 -66.528 54.407
407541 93.435 -66.530 54.375
407542 93.472 -66.534 54.344
407543 93.510 -66.538 54.314
407544 93.547 -66.542 54.289
407545 93.586 -66.549 54.269
407546 93.626 -66.558 54.253
407547 93.665 -66.566 54.240
407548 93.706 -66.574 54.231
407549 93.748 -66.577 54.221
407550 93.791 -66.579 54.211
407551 93.834 -66.581 54.201
407552 93.869 -66.577 54.189
407553 93.908 -66.577 54.174
407554 93.948 -66.581 54.159
407555 93.988 -66.586 54.142
407556 94.029 -66.590 54.123
407557 94.069 -66.595 54.104
407558 94.109 -66.599 54.087
407559 94.151 -66.604 54.072
407560 94.193 -66.608 54.057
407561 94.235 -66.613 54.041
407562 94.276 -66.617 54.023
407563 94.318 -66.622 54.002
407564 94.359 -66.627 53.977
407565 94.401 -66.632 53.950
407566 94.438 -66.635 53.921
407567 94.470 -66.636 53.891
407568 94.518 -66.639 53.857
407569 94.558 -66.641 53.824
407570 94.598 -66.643 53.791
407571 94.638 -66.645 53.758
407572 94.678 -66.646 53.724
407573 94.718 -66.648 53.690
407574 94.758 -66.649 53.655
407575 94.798 -66.651 53.621
407576 94.838 -66.653 53.590
407577 94.878 -66.655 53.562
407578 94.917 -66.667 53.538
407579 94.954 -66.675 53.517
407580 94.991 -66.683 53.499
407581 95.024 -66.685 53.486
407582 95.056 -66.684 53.474
407583 95.088 -66.683 53.464
407584 95.119 -66.683 53.456
407585 95.151 -66.682 53.446
407586 95.183 -66.681 53.431
407587 95.211 -66.679 53.413
407588 95.232 -66.673 53.392
407589 95.260 -66.671 53.367
407590 95.308 -66.676 53.344
407591 95.355 -66.681 53.324
407592 95.403 -66.689 53.305
407593 95.451 -66.702 53.290
407594 95.499 -66.716 53.279
407595 95.548 -66.730 53.266
407596 95.592 -66.733 53.254
407597 95.636 -66.736 53.241
407598 95.680 -66.740 53.228
407599 95.725 -66.744 53.214
407600 95.771 -66.751 53.199
407601 95.817 -66.758 53.183
407602 95.854 -66.763 53.166
407603 95.891 -66.768 53.147
407604 95.927 -66.773 53.124
407605 95.964 -66.778 53.100
407606 96.001 -66.784 53.074
407607 96.036 -66.788 53.046
407608 96.072 -66.792 53.018
407609 96.108 -66.796 52.990
407610 96.146 -66.799 52.964
407611 96.184 -66.801 52.937
407612 96.224 -66.802 52.908
407613 96.259 -66.802 52.875
407614 96.293 -66.803 52.842
407615 96.328 -66.804 52.809
407616 96.363 -66.805 52.778
407617 96.399 -66.808 52.747
407618 96.435 -66.811 52.715
407619 96.463 -66.809 52.682
407620 96.491 -66.808 52.652
407621 96.520 -66.805 52.626
407622 96.548 -66.804 52.600
407623 96.576 -66.802 52.578
407624 96.604 -66.797 52.555
407625 96.635 -66.793 52.531
407626 96.674 -66.796 52.505
407627 96.713 -66.798 52.479
407628 96.752 -66.801 52.455
407629 96.792 -66.804 52.434
407630 96.832 -66.805 52.414
407631 96.873 -66.807 52.394
407632 96.914 -66.808 52.369
407633 96.957 -66.810 52.345
407634 97.002 -66.806 52.322
407635 97.047 -66.802 52.301
407636 97.093 -66.799 52.282
407637 97.139 -66.794 52.263
407638 97.184 -66.790 52.245
407639 97.230 -66.787 52.227
407640 97.276 -66.782 52.210
407641 97.321 -66.778 52.190
407642 97.367 -66.774 52.167
407643 97.415 -66.782 52.138
407644 97.464 -66.790 52.106
407645 97.513 -66.798 52.071
407646 97.562 -66.806 52.035
407647 97.596 -66.814 51.996
407648 97.630 -66.824 51.955
407649 97.663 -66.834 51.913
407650 97.697 -66.845 51.873
407651 97.731 -66.855 51.829
407652 97.764 -66.865 51.785
407653 97.798 -66.875 51.741
407654 97.832 -66.886 51.696
407655 97.866 -66.896 51.654
407656 97.899 -66.906 51.612
407657 97.932 -66.915 51.569
407658 97.965 -66.925 51.525
407659 97.998 -66.934 51.480
407660 98.031 -66.943 51.434
407661 98.064 -66.953 51.387
407662 98.097 -66.962 51.341
407663 98.130 -66.972 51.295
407664 98.163 -66.981 51.249
407665 98.196 -66.990 51.203
407666 98.229 -67.000 51.158
407667 98.256 -67.008 51.114
407668 98.282 -67.015 51.072
407669 98.307 -67.022 51.034
407670 98.333 -67.028 50.997
407671 98.357 -67.031 50.961
407672 98.379 -67.028 50.923
407673 98.410 -67.026 50.882
407674 98.440 -67.025 50.839
407675 98.470 -67.023 50.795
407676 98.500 -67.022 50.751
407677 98.530 -67.020 50.705
407678 98.561 -67.019 50.658
407679 98.591 -67.017 50.611
407680 98.623 -67.017 50.563
407681 98.645 -67.051 50.514
407682 98.667 -67.085 50.465
407683 98.712 -67.090 50.415
407684 98.759 -67.091 50.364
407685 98.806 -67.092 50.312
407686 98.853 -67.094 50.260
407687 98.901 -67.094 50.206
407688 98.948 -67.095 50.153
407689 98.995 -67.097 50.099
407690 99.042 -67.097 50.044
407691 99.089 -67.099 49.990
407692 99.137 -67.100 49.936
407693 99.186 -67.101 49.882
407694 99.237 -67.106 49.828
407695 99.276 -67.116 49.776
407696 99.313 -67.133 49.724
407697 99.357 -67.150 49.673
407698 99.401 -67.168 49.621
407699 99.444 -67.185 49.571
407700 99.488 -67.202 49.520
407701 99.532 -67.219 49.469
407702 99.575 -67.237 49.419
407703 99.619 -67.254 49.369
407704 99.662 -67.272 49.318
407705 99.706 -67.288 49.268
407706 99.749 -67.306 49.220
407707 99.798 -67.304 49.170
407708 99.824 -67.285 49.122
407709 99.837 -67.285 49.075
407710 99.845 -67.294 49.027
407711 99.853 -67.303 48.978
407712 99.868 -67.310 48.931
407713 99.888 -67.319 48.884
407714 99.908 -67.326 48.837
407715 99.940 -67.329 48.794
407716 99.980 -67.326 48.750
407717 100.019 -67.324 48.707
407718 100.059 -67.323 48.664
407719 100.099 -67.320 48.621
407720 100.138 -67.318 48.579
407721 100.177 -67.316 48.538
407722 100.217 -67.314 48.497
407723 100.257 -67.312 48.457
407724 100.298 -67.310 48.417
407725 100.347 -67.308 48.378
408177 90.371 -66.395 54.936
408178 90.427 -66.398 54.942
408179 90.482 -66.403 54.948
408180 90.538 -66.407 54.955
408181 90.593 -66.411 54.962
408182 90.649 -66.416 54.969
408183 90.705 -66.420 54.976
408184 90.760 -66.424 54.983
408185 90.816 -66.429 54.991
408186 90.872 -66.433 55.000
408187 90.922 -66.442 55.010
408188 90.970 -66.451 55.021
408189 91.019 -66.461 55.031
408190 91.067 -66.470 55.044
408191 91.116 -66.480 55.057
408192 91.164 -66.489 55.071
408193 91.207 -66.492 55.085
408194 91.248 -66.492 55.096
408195 91.289 -66.493 55.107
408196 91.331 -66.494 55.115
408197 91.374 -66.501 55.121
408198 91.418 -66.508 55.122
408199 91.461 -66.517 55.119
408200 91.504 -66.519 55.113
408201 91.546 -66.512 55.104
408202 91.590 -66.516 55.093
408203 91.635 -66.520 55.081
408204 91.679 -66.524 55.067
408205 91.724 -66.530 55.051
408206 91.769 -66.536 55.036
408207 91.813 -66.542 55.023
408208 91.858 -66.547 55.011
408209 91.905 -66.554 55.001
408210 91.963 -66.564 54.992
408211 92.022 -66.574 54.982
408212 92.080 -66.584 54.971
408213 92.139 -66.594 54.959
408214 92.197 -66.604 54.949
408215 92.255 -66.614 54.940
408216 92.313 -66.624 54.932
408217 92.372 -66.633 54.926
408218 92.430 -66.643 54.914
408219 92.489 -66.654 54.901
408220 92.546 -66.664 54.889
408221 92.601 -66.677 54.881
408222 92.654 -66.684 54.877
408223 92.704 -66.687 54.872
408224 92.751 -66.690 54.861
408225 92.795 -66.682 54.846
408226 92.837 -66.671 54.829
408227 92.875 -66.657 54.812
408228 92.913 -66.643 54.792
408229 92.951 -66.629 54.771
408230 92.989 -66.614 54.748
408231 93.027 -66.601 54.724
408232 93.065 -66.586 54.704
408233 93.103 -66.572 54.683
408234 93.142 -66.558 54.661
408235 93.180 -66.544 54.632
408236 93.218 -66.530 54.599
408237 93.256 -66.516 54.564
408238 93.293 -66.503 54.528
408239 93.328 -66.498 54.493
408240 93.362 -66.492 54.459
408241 93.396 -66.487 54.425
408242 93.430 -66.484 54.392
408243 93.468 -66.488 54.362
408244 93.505 -66.492 54.331
408245 93.543 -66.496 54.310
408246 93.581 -66.503 54.292
408247 93.621 -66.511 54.277
408248 93.661 -66.520 54.264
408249 93.703 -66.525 54.255
408250 93.746 -66.527 54.246
408251 93.788 -66.530 54.236
408252 93.831 -66.533 54.225
408253 93.871 -66.533 54.213
408254 93.911 -66.536 54.197
408255 93.952 -66.541 54.180
408256 93.992 -66.545 54.161
408257 94.032 -66.550 54.140
408258 94.072 -66.554 54.121
408259 94.113 -66.558 54.103
408260 94.153 -66.563 54.088
408261 94.195 -66.567 54.072
408262 94.237 -66.573 54.055
408263 94.279 -66.577 54.036
408264 94.320 -66.582 54.013
408265 94.362 -66.586 53.987
408266 94.400 -66.589 53.959
408267 94.430 -66.591 53.928
408268 94.461 -66.592 53.897
408269 94.511 -66.596 53.864
408270 94.555 -66.599 53.833
408271 94.595 -66.600 53.803
408272 94.635 -66.602 53.773
408273 94.675 -66.604 53.739
408274 94.715 -66.605 53.705
408275 94.755 -66.607 53.671
408276 94.795 -66.608 53.637
408277 94.835 -66.610 53.607
408278 94.876 -66.619 53.581
408279 94.915 -66.630 53.558
408280 94.952 -66.639 53.538
408281 94.988 -66.647 53.522
408282 95.025 -66.655 53.508
408283 95.062 -66.664 53.497
408284 95.095 -66.665 53.487
408285 95.127 -66.664 53.476
408286 95.159 -66.664 53.462
408287 95.190 -66.663 53.444
408288 95.213 -66.658 53.422
408289 95.234 -66.652 53.398
408290 95.255 -66.646 53.372
408291 95.293 -66.647 53.350
408292 95.340 -66.653 53.330
408293 95.387 -66.658 53.313
408294 95.435 -66.664 53.299
408295 95.482 -66.669 53.288
408296 95.530 -66.682 53.277
408297 95.578 -66.694 53.267
408298 95.622 -66.697 53.255
408299 95.667 -66.701 53.241
408300 95.712 -66.708 53.226
408301 95.758 -66.715 53.210
408302 95.804 -66.722 53.195
408303 95.844 -66.727 53.178
408304 95.881 -66.733 53.161
408305 95.917 -66.737 53.140
408306 95.954 -66.743 53.118
408307 95.991 -66.747 53.093
408308 96.029 -66.750 53.066
408309 96.067 -66.752 53.039
408310 96.105 -66.755 53.011
408311 96.143 -66.757 52.985
408312 96.181 -66.759 52.957
408313 96.222 -66.760 52.927
408314 96.259 -66.760 52.894
408315 96.294 -66.761 52.862
408316 96.328 -66.762 52.830
408317 96.364 -66.763 52.799
408318 96.400 -66.767 52.770
408319 96.436 -66.770 52.736
408320 96.468 -66.771 52.702
408321 96.496 -66.768 52.672
408322 96.524 -66.767 52.645
408323 96.553 -66.765 52.620
408324 96.581 -66.761 52.599
408325 96.609 -66.754 52.575
408326 96.639 -66.749 52.550
408327 96.678 -66.752 52.524
408328 96.718 -66.754 52.495
408329 96.757 -66.757 52.468
408330 96.796 -66.759 52.444
408331 96.835 -66.762 52.423
408332 96.876 -66.762 52.404
408333 96.920 -66.758 52.380
408334 96.965 -66.755 52.356
408335 97.011 -66.752 52.333
408336 97.056 -66.748 52.312
408337 97.102 -66.744 52.293
408338 97.147 -66.740 52.275
408339 97.192 -66.737 52.257
408340 97.238 -66.733 52.238
408341 97.283 -66.729 52.220
408342 97.327 -66.730 52.198
408343 97.369 -66.738 52.173
408344 97.413 -66.744 52.144
408345 97.462 -66.752 52.110
408346 97.511 -66.759 52.075
408347 97.561 -66.767 52.038
408348 97.599 -66.779 51.999
408349 97.638 -66.791 51.958
408350 97.673 -66.802 51.917
408351 97.707 -66.813 51.875
408352 97.741 -66.823 51.831
408353 97.775 -66.834 51.786
408354 97.808 -66.843 51.741
408355 97.842 -66.854 51.695
408356 97.875 -66.864 51.652
408357 97.909 -66.874 51.608
408358 97.942 -66.884 51.565
408359 97.975 -66.893 51.521
408360 98.008 -66.903 51.477
408361 98.041 -66.912 51.432
408362 98.074 -66.921 51.386
408363 98.107 -66.931 51.340
408364 98.140 -66.940 51.294
408365 98.172 -66.950 51.248
408366 98.198 -66.956 51.203
408367 98.223 -66.963 51.158
408368 98.248 -66.970 51.114
408369 98.274 -66.978 51.072
408370 98.300 -66.984 51.034
408371 98.325 -66.991 50.997
408372 98.349 -66.990 50.961
408373 98.371 -66.987 50.924
408374 98.397 -66.985 50.885
408375 98.428 -66.984 50.842
408376 98.458 -66.982 50.798
408377 98.489 -66.981 50.753
408378 98.519 -66.979 50.707
408379 98.553 -66.981 50.659
408380 98.588 -66.985 50.612
408381 98.623 -66.990 50.563
408382 98.649 -67.015 50.513
408383 98.673 -67.047 50.462
408384 98.720 -67.048 50.411
408385 98.767 -67.050 50.359
408386 98.814 -67.050 50.307
408387 98.862 -67.052 50.254
408388 98.909 -67.053 50.201
408389 98.956 -67.054 50.147
408390 99.003 -67.055 50.093
408391 99.051 -67.056 50.039
408392 99.098 -67.057 49.985
408393 99.145 -67.058 49.931
408394 99.192 -67.059 49.877
408395 99.241 -67.062 49.824
408396 99.282 -67.078 49.772
408397 99.325 -67.095 49.720
408398 99.369 -67.112 49.669
408399 99.412 -67.129 49.618
408400 99.456 -67.147 49.567
408401 99.500 -67.164 49.516
408402 99.543 -67.182 49.464
408403 99.587 -67.198 49.413
408404 99.630 -67.216 49.362
408405 99.674 -67.233 49.310
408406 99.717 -67.251 49.260
408407 99.761 -67.267 49.211
408408 99.796 -67.263 49.161
408409 99.821 -67.245 49.112
408410 99.837 -67.241 49.065
408411 99.845 -67.251 49.017
408412 99.853 -67.259 48.969
408413 99.873 -67.266 48.921
408414 99.894 -67.275 48.874
408415 99.914 -67.282 48.828
408416 99.949 -67.282 48.785
408417 99.989 -67.280 48.740
408418 100.028 -67.279 48.698
408419 100.068 -67.276 48.655
408420 100.108 -67.274 48.612
408421 100.147 -67.272 48.570
408422 100.187 -67.270 48.529
408423 100.226 -67.268 48.488
408424 100.266 -67.266 48.448
408425 100.313 -67.264 48.409
408879 90.434 -66.349 54.942
408880 90.489 -66.354 54.948
408881 90.545 -66.357 54.954
408882 90.600 -66.362 54.961
408883 90.656 -66.367 54.968
408884 90.712 -66.370 54.974
408885 90.767 -66.375 54.980
408886 90.823 -66.379 54.988
408887 90.872 -66.389 54.996
408888 90.921 -66.398 55.005
408889 90.969 -66.407 55.014
408890 91.018 -66.417 55.022
408891 91.067 -66.426 55.033
408892 91.115 -66.436 55.044
408893 91.163 -66.446 55.056
408894 91.211 -66.454 55.069
408895 91.252 -66.455 55.079
408896 91.294 -66.456 55.090
408897 91.335 -66.457 55.099
408898 91.376 -66.458 55.106
408899 91.418 -66.459 55.109
408900 91.461 -66.465 55.108
408901 91.504 -66.472 55.103
408902 91.547 -66.470 55.096
408903 91.592 -66.473 55.087
408904 91.637 -66.480 55.076
408905 91.681 -66.485 55.064
408906 91.726 -66.491 55.051
408907 91.770 -66.496 55.039
408908 91.814 -66.502 55.029
408909 91.860 -66.508 55.020
408910 91.918 -66.518 55.011
408911 91.977 -66.528 55.002
408912 92.035 -66.538 54.993
408913 92.093 -66.548 54.982
408914 92.152 -66.558 54.972
408915 92.210 -66.568 54.963
408916 92.269 -66.578 54.955
408917 92.326 -66.589 54.949
408918 92.381 -66.602 54.943
408919 92.437 -66.614 54.932
408920 92.492 -66.627 54.919
408921 92.545 -66.633 54.909
408922 92.596 -66.633 54.903
408923 92.646 -66.636 54.901
408924 92.694 -66.646 54.897
408925 92.740 -66.643 54.888
408926 92.783 -66.635 54.873
408927 92.827 -66.626 54.854
408928 92.871 -66.617 54.833
408929 92.910 -66.604 54.812
408930 92.948 -66.590 54.789
408931 92.986 -66.576 54.767
408932 93.024 -66.561 54.744
408933 93.062 -66.548 54.726
408934 93.100 -66.533 54.707
408935 93.138 -66.519 54.686
408936 93.177 -66.505 54.657
408937 93.215 -66.491 54.625
408938 93.253 -66.476 54.588
408939 93.291 -66.463 54.549
408940 93.325 -66.457 54.513
408941 93.359 -66.452 54.478
408942 93.394 -66.447 54.443
408943 93.427 -66.442 54.411
408944 93.464 -66.442 54.382
408945 93.501 -66.446 54.354
408946 93.539 -66.450 54.333
408947 93.577 -66.457 54.316
408948 93.617 -66.466 54.303
408949 93.657 -66.473 54.291
408950 93.700 -66.476 54.281
408951 93.743 -66.479 54.270
408952 93.785 -66.481 54.260
408953 93.828 -66.484 54.248
408954 93.872 -66.490 54.236
408955 93.914 -66.495 54.220
408956 93.955 -66.500 54.201
408957 93.995 -66.505 54.180
408958 94.035 -66.509 54.158
408959 94.075 -66.514 54.139
408960 94.116 -66.518 54.120
408961 94.156 -66.523 54.103
408962 94.198 -66.527 54.086
408963 94.240 -66.532 54.068
408964 94.281 -66.536 54.046
408965 94.323 -66.541 54.023
408966 94.361 -66.545 53.995
408967 94.391 -66.546 53.965
408968 94.422 -66.548 53.934
408969 94.452 -66.549 53.903
408970 94.501 -66.552 53.871
408971 94.552 -66.556 53.842
408972 94.592 -66.558 53.816
408973 94.632 -66.559 53.788
408974 94.672 -66.561 53.754
408975 94.712 -66.563 53.721
408976 94.752 -66.564 53.687
408977 94.792 -66.566 53.655
408978 94.832 -66.568 53.626
408979 94.875 -66.583 53.601
408980 94.912 -66.593 53.580
408981 94.949 -66.602 53.561
408982 94.986 -66.611 53.545
408983 95.023 -66.619 53.531
408984 95.059 -66.627 53.519
408985 95.096 -66.636 53.507
408986 95.133 -66.644 53.494
408987 95.166 -66.646 53.476
408988 95.195 -66.644 53.455
408989 95.216 -66.638 53.430
408990 95.237 -66.632 53.404
408991 95.257 -66.626 53.378
408992 95.278 -66.620 53.356
408993 95.324 -66.624 53.337
408994 95.372 -66.630 53.321
408995 95.419 -66.635 53.308
408996 95.467 -66.640 53.298
408997 95.514 -66.646 53.288
408998 95.561 -66.652 53.279
408999 95.608 -66.658 53.268
409000 95.654 -66.665 53.254
409001 95.700 -66.672 53.239
409002 95.746 -66.679 53.223
409003 95.792 -66.686 53.206
409004 95.834 -66.693 53.189
409005 95.871 -66.697 53.171
409006 95.908 -66.702 53.151
409007 95.944 -66.707 53.130
409008 95.984 -66.709 53.108
409009 96.026 -66.709 53.084
409010 96.064 -66.711 53.059
409011 96.102 -66.714 53.032
409012 96.141 -66.716 53.005
409013 96.179 -66.718 52.978
409014 96.220 -66.718 52.948
409015 96.259 -66.718 52.914
409016 96.294 -66.718 52.881
409017 96.329 -66.719 52.849
409018 96.365 -66.722 52.818
409019 96.400 -66.725 52.788
409020 96.437 -66.728 52.755
409021 96.473 -66.731 52.722
409022 96.501 -66.729 52.692
409023 96.529 -66.727 52.665
409024 96.557 -66.724 52.640
409025 96.586 -66.717 52.618
409026 96.614 -66.710 52.595
409027 96.644 -66.705 52.569
409028 96.683 -66.707 52.542
409029 96.722 -66.710 52.511
409030 96.761 -66.713 52.482
409031 96.801 -66.715 52.455
409032 96.842 -66.712 52.430
409033 96.883 -66.710 52.411
409034 96.928 -66.706 52.388
409035 96.973 -66.702 52.365
409036 97.018 -66.699 52.343
409037 97.064 -66.696 52.323
409038 97.109 -66.692 52.304
409039 97.154 -66.688 52.286
409040 97.199 -66.684 52.268
409041 97.245 -66.680 52.249
409042 97.288 -66.685 52.228
409043 97.330 -66.696 52.204
409044 97.371 -66.702 52.177
409045 97.411 -66.707 52.146
409046 97.459 -66.715 52.111
409047 97.508 -66.722 52.073
409048 97.558 -66.730 52.036
409049 97.596 -66.742 51.998
409050 97.634 -66.754 51.958
409051 97.672 -66.766 51.917
409052 97.711 -66.778 51.874
409053 97.749 -66.790 51.831
409054 97.785 -66.802 51.786
409055 97.818 -66.812 51.740
409056 97.852 -66.822 51.695
409057 97.885 -66.832 51.649
409058 97.919 -66.843 51.604
409059 97.952 -66.853 51.560
409060 97.986 -66.862 51.517
409061 98.018 -66.871 51.474
409062 98.052 -66.881 51.431
409063 98.085 -66.890 51.386
409064 98.113 -66.898 51.340
409065 98.139 -66.905 51.294
409066 98.164 -66.912 51.249
409067 98.190 -66.919 51.203
409068 98.216 -66.926 51.157
409069 98.241 -66.933 51.114
409070 98.267 -66.940 51.071
409071 98.292 -66.947 51.034
409072 98.317 -66.953 50.998
409073 98.340 -66.950 50.962
409074 98.363 -66.947 50.925
409075 98.386 -66.944 50.887
409076 98.417 -66.943 50.845
409077 98.447 -66.941 50.800
409078 98.482 -66.946 50.754
409079 98.518 -66.950 50.707
409080 98.553 -66.954 50.659
409081 98.588 -66.958 50.611
409082 98.624 -66.962 50.562
409083 98.653 -66.980 50.511
409084 98.681 -67.006 50.460
409085 98.728 -67.007 50.407
409086 98.776 -67.008 50.354
409087 98.823 -67.009 50.302
409088 98.870 -67.010 50.249
409089 98.917 -67.011 50.196
409090 98.964 -67.012 50.142
409091 99.012 -67.013 50.089
409092 99.059 -67.015 50.034
409093 99.106 -67.015 49.980
409094 99.153 -67.018 49.926
409095 99.199 -67.019 49.873
409096 99.244 -67.025 49.820
409097 99.277 -67.030 49.768
409098 99.313 -67.040 49.716
409099 99.356 -67.057 49.666
409100 99.400 -67.074 49.614
409101 99.443 -67.091 49.563
409102 99.486 -67.108 49.511
409103 99.529 -67.125 49.459
409104 99.573 -67.143 49.407
409105 99.617 -67.160 49.355
409106 99.660 -67.177 49.304
409107 99.703 -67.194 49.252
409108 99.747 -67.211 49.202
409109 99.794 -67.223 49.153
409110 99.819 -67.205 49.104
409111 99.837 -67.198 49.055
409112 99.845 -67.207 49.007
409113 99.858 -67.215 48.959
409114 99.879 -67.222 48.912
409115 99.899 -67.230 48.865
409116 99.920 -67.238 48.820
409117 99.959 -67.236 48.776
409118 99.998 -67.235 48.732
409119 100.037 -67.232 48.689
409120 100.077 -67.230 48.646
409121 100.117 -67.229 48.603
409122 100.156 -67.226 48.562
409123 100.196 -67.224 48.521
409124 100.235 -67.222 48.481
409125 100.279 -67.220 48.440
409126 100.327 -67.218 48.401
409582 90.552 -66.308 54.955
409583 90.608 -66.313 54.961
409584 90.663 -66.317 54.967
409585 90.719 -66.321 54.973
409586 90.774 -66.326 54.979
409587 90.822 -66.335 54.986
409588 90.871 -66.345 54.993
409589 90.920 -66.354 55.000
409590 90.968 -66.364 55.008
409591 91.017 -66.373 55.014
409592 91.066 -66.383 55.023
409593 91.114 -66.393 55.033
409594 91.162 -66.402 55.043
409595 91.211 -66.412 55.055
409596 91.257 -66.418 55.065
409597 91.298 -66.419 55.074
409598 91.340 -66.420 55.084
409599 91.381 -66.421 55.092
409600 91.422 -66.422 55.097
409601 91.463 -66.423 55.097
409602 91.505 -66.423 55.094
409603 91.548 -66.428 55.089
409604 91.593 -66.434 55.081
409605 91.638 -66.440 55.072
409606 91.682 -66.445 55.062
409607 91.727 -66.451 55.052
409608 91.772 -66.457 55.043
409609 91.816 -66.463 55.036
409610 91.873 -66.472 55.029
409611 91.932 -66.483 55.021
409612 91.990 -66.492 55.012
409613 92.049 -66.502 55.004
409614 92.107 -66.512 54.994
409615 92.162 -66.525 54.984
409616 92.217 -66.538 54.976
409617 92.273 -66.550 54.969
409618 92.328 -66.563 54.963
409619 92.383 -66.576 54.957
409620 92.436 -66.581 54.947
409621 92.487 -66.581 54.936
409622 92.537 -66.580 54.927
409623 92.587 -66.584 54.922
409624 92.636 -66.594 54.923
409625 92.684 -66.604 54.920
409626 92.728 -66.597 54.913
409627 92.772 -66.588 54.901
409628 92.816 -66.580 54.880
409629 92.860 -66.570 54.857
409630 92.904 -66.562 54.832
409631 92.945 -66.551 54.807
409632 92.983 -66.537 54.784
409633 93.021 -66.523 54.763
409634 93.059 -66.508 54.747
409635 93.097 -66.495 54.729
409636 93.135 -66.480 54.708
409637 93.174 -66.466 54.679
409638 93.211 -66.452 54.647
409639 93.250 -66.438 54.608
409640 93.288 -66.423 54.569
409641 93.323 -66.417 54.531
409642 93.357 -66.411 54.496
409643 93.391 -66.407 54.462
409644 93.425 -66.401 54.431
409645 93.459 -66.396 54.402
409646 93.497 -66.401 54.376
409647 93.534 -66.404 54.356
409648 93.573 -66.411 54.341
409649 93.613 -66.420 54.328
409650 93.654 -66.424 54.316
409651 93.697 -66.427 54.306
409652 93.740 -66.429 54.294
409653 93.783 -66.433 54.281
409654 93.828 -66.441 54.269
409655 93.872 -66.448 54.255
409656 93.915 -66.455 54.239
409657 93.956 -66.460 54.221
409658 93.997 -66.464 54.200
409659 94.039 -66.469 54.177
409660 94.079 -66.473 54.156
409661 94.119 -66.477 54.136
409662 94.159 -66.482 54.118
409663 94.200 -66.486 54.100
409664 94.242 -66.491 54.080
409665 94.283 -66.495 54.054
409666 94.322 -66.499 54.027
409667 94.352 -66.501 53.999
409668 94.383 -66.502 53.969
409669 94.413 -66.504 53.938
409670 94.444 -66.505 53.907
409671 94.491 -66.508 53.879
409672 94.543 -66.512 53.852
409673 94.589 -66.515 53.826
409674 94.629 -66.517 53.799
409675 94.669 -66.518 53.768
409676 94.709 -66.520 53.736
409677 94.749 -66.522 53.704
409678 94.789 -66.523 53.673
409679 94.831 -66.533 53.646
409680 94.873 -66.548 53.624
409681 94.910 -66.557 53.604
409682 94.947 -66.565 53.586
409683 94.983 -66.574 53.570
409684 95.020 -66.582 53.554
409685 95.057 -66.591 53.539
409686 95.094 -66.599 53.524
409687 95.130 -66.608 53.508
409688 95.166 -66.615 53.488
409689 95.194 -66.618 53.464
409690 95.218 -66.617 53.437
409691 95.239 -66.611 53.410
409692 95.260 -66.605 53.384
409693 95.280 -66.599 53.363
409694 95.309 -66.596 53.345
409695 95.356 -66.602 53.330
409696 95.404 -66.607 53.317
409697 95.451 -66.612 53.308
409698 95.498 -66.617 53.299
409699 95.554 -66.617 53.290
409700 95.600 -66.622 53.280
409701 95.642 -66.630 53.268
409702 95.688 -66.636 53.254
409703 95.734 -66.643 53.236
409704 95.779 -66.650 53.218
409705 95.824 -66.657 53.201
409706 95.861 -66.661 53.182
409707 95.898 -66.667 53.163
409708 95.937 -66.669 53.143
409709 95.979 -66.669 53.122
409710 96.021 -66.670 53.100
409711 96.061 -66.671 53.077
409712 96.100 -66.673 53.052
409713 96.137 -66.675 53.026
409714 96.176 -66.677 52.998
409715 96.218 -66.676 52.968
409716 96.259 -66.675 52.934
409717 96.295 -66.676 52.902
409718 96.329 -66.677 52.869
409719 96.366 -66.680 52.836
409720 96.401 -66.683 52.803
409721 96.438 -66.686 52.771
409722 96.474 -66.690 52.740
409723 96.506 -66.690 52.711
409724 96.534 -66.686 52.684
409725 96.562 -66.680 52.659
409726 96.591 -66.673 52.636
409727 96.619 -66.666 52.613
409728 96.648 -66.661 52.587
409729 96.687 -66.663 52.557
409730 96.727 -66.666 52.525
409731 96.767 -66.665 52.494
409732 96.808 -66.662 52.465
409733 96.850 -66.660 52.439
409734 96.891 -66.657 52.419
409735 96.935 -66.654 52.396
409736 96.980 -66.650 52.374
409737 97.026 -66.646 52.353
409738 97.071 -66.643 52.333
409739 97.116 -66.639 52.314
409740 97.162 -66.636 52.296
409741 97.207 -66.632 52.276
409742 97.249 -66.640 52.255
409743 97.291 -66.651 52.233
409744 97.332 -66.661 52.206
409745 97.373 -66.667 52.176
409746 97.413 -66.671 52.143
409747 97.456 -66.677 52.107
409748 97.505 -66.684 52.068
409749 97.554 -66.693 52.029
409750 97.592 -66.705 51.990
409751 97.631 -66.717 51.950
409752 97.669 -66.729 51.909
409753 97.707 -66.741 51.868
409754 97.745 -66.753 51.824
409755 97.784 -66.765 51.780
409756 97.822 -66.777 51.735
409757 97.860 -66.790 51.690
409758 97.895 -66.800 51.645
409759 97.929 -66.811 51.600
409760 97.962 -66.821 51.555
409761 97.996 -66.830 51.510
409762 98.029 -66.840 51.467
409763 98.055 -66.846 51.424
409764 98.080 -66.854 51.379
409765 98.106 -66.861 51.335
409766 98.131 -66.868 51.291
409767 98.157 -66.874 51.246
409768 98.182 -66.882 51.202
409769 98.208 -66.889 51.158
409770 98.233 -66.896 51.116
409771 98.259 -66.903 51.074
409772 98.285 -66.910 51.034
409773 98.309 -66.913 50.997
409774 98.332 -66.910 50.960
409775 98.354 -66.907 50.923
409776 98.378 -66.904 50.886
409777 98.413 -66.909 50.845
409778 98.447 -66.914 50.801
409779 98.483 -66.918 50.755
409780 98.518 -66.922 50.708
409781 98.553 -66.926 50.660
409782 98.589 -66.931 50.610
409783 98.624 -66.934 50.560
409784 98.657 -66.944 50.509
409785 98.690 -66.965 50.456
409786 98.737 -66.965 50.404
409787 98.784 -66.966 50.350
409788 98.831 -66.968 50.297
409789 98.878 -66.968 50.243
409790 98.926 -66.970 50.189
409791 98.973 -66.972 50.136
409792 99.019 -66.973 50.082
409793 99.065 -66.975 50.028
409794 99.111 -66.978 49.974
409795 99.157 -66.980 49.921
409796 99.203 -66.981 49.868
409797 99.248 -66.990 49.815
409798 99.291 -67.003 49.763
409799 99.320 -67.006 49.710
409800 99.341 -67.002 49.659
409801 99.378 -67.012 49.608
409802 99.421 -67.030 49.556
409803 99.465 -67.047 49.504
409804 99.508 -67.064 49.451
409805 99.551 -67.081 49.400
409806 99.594 -67.098 49.347
409807 99.638 -67.116 49.295
409808 99.682 -67.132 49.244
409809 99.725 -67.150 49.194
409810 99.777 -67.165 49.144
409811 99.817 -67.165 49.095
409812 99.837 -67.154 49.046
409813 99.845 -67.163 48.998
409814 99.864 -67.170 48.951
409815 99.885 -67.178 48.904
409816 99.905 -67.186 48.858
409817 99.928 -67.192 48.812
409818 99.968 -67.191 48.768
409819 100.007 -67.188 48.725
409820 100.047 -67.186 48.682
409821 100.086 -67.185 48.639
409822 100.126 -67.182 48.596
409823 100.165 -67.180 48.554
409824 100.205 -67.178 48.513
409825 100.245 -67.176 48.473
409826 100.293 -67.174 48.433
409827 100.342 -67.172 48.393
410284 90.615 -66.263 54.961
410285 90.670 -66.267 54.966
410286 90.724 -66.273 54.972
410287 90.773 -66.282 54.978
410288 90.821 -66.292 54.983
410289 90.870 -66.301 54.990
410290 90.919 -66.311 54.996
410291 90.967 -66.320 55.003
410292 91.016 -66.330 55.010
410293 91.065 -66.340 55.017
410294 91.113 -66.349 55.026
410295 91.162 -66.359 55.036
410296 91.210 -66.369 55.048
410297 91.258 -66.378 55.058
410298 91.302 -66.381 55.066
410299 91.343 -66.382 55.073
410300 91.383 -66.382 55.081
410301 91.424 -66.382 55.086
410302 91.465 -66.383 55.088
410303 91.506 -66.383 55.086
410304 91.551 -66.397 55.083
410305 91.595 -66.401 55.076
410306 91.639 -66.402 55.067
410307 91.683 -66.406 55.059
410308 91.728 -66.411 55.053
410309 91.773 -66.417 55.048
410310 91.828 -66.426 55.044
410311 91.887 -66.436 55.038
410312 91.943 -66.448 55.031
410313 91.998 -66.461 55.023
410314 92.054 -66.474 55.016
410315 92.108 -66.486 55.007
410316 92.164 -66.499 54.998
410317 92.219 -66.512 54.990
410318 92.275 -66.525 54.983
410319 92.327 -66.530 54.976
410320 92.378 -66.530 54.970
410321 92.429 -66.529 54.961
410322 92.479 -66.529 54.951
410323 92.529 -66.533 54.943
410324 92.577 -66.542 54.940
410325 92.626 -66.552 54.941
410326 92.673 -66.559 54.940
410327 92.717 -66.550 54.934
410328 92.761 -66.542 54.922
410329 92.804 -66.533 54.901
410330 92.848 -66.524 54.877
410331 92.892 -66.515 54.851
410332 92.936 -66.507 54.825
410333 92.980 -66.498 54.799
410334 93.018 -66.484 54.780
410335 93.056 -66.470 54.763
410336 93.094 -66.456 54.746
410337 93.132 -66.442 54.723
410338 93.170 -66.427 54.695
410339 93.208 -66.414 54.660
410340 93.247 -66.399 54.622
410341 93.284 -66.385 54.583
410342 93.320 -66.376 54.548
410343 93.354 -66.371 54.513
410344 93.389 -66.366 54.480
410345 93.423 -66.361 54.450
410346 93.457 -66.356 54.423
410347 93.492 -66.354 54.397
410348 93.530 -66.358 54.380
410349 93.569 -66.366 54.366
410350 93.609 -66.373 54.353
410351 93.652 -66.376 54.341
410352 93.694 -66.378 54.330
410353 93.739 -66.384 54.316
410354 93.783 -66.392 54.302
410355 93.827 -66.399 54.288
410356 93.872 -66.407 54.272
410357 93.916 -66.414 54.256
410358 93.958 -66.419 54.238
410359 93.998 -66.423 54.218
410360 94.040 -66.428 54.195
410361 94.081 -66.433 54.172
410362 94.122 -66.437 54.152
410363 94.163 -66.442 54.133
410364 94.203 -66.446 54.114
410365 94.244 -66.451 54.092
410366 94.283 -66.454 54.064
410367 94.314 -66.456 54.035
410368 94.344 -66.458 54.005
410369 94.374 -66.459 53.975
410370 94.405 -66.461 53.944
410371 94.436 -66.461 53.914
410372 94.481 -66.464 53.887
410373 94.533 -66.468 53.861
410374 94.584 -66.472 53.835
410375 94.626 -66.474 53.809
410376 94.667 -66.476 53.779
410377 94.706 -66.477 53.750
410378 94.747 -66.479 53.721
410379 94.786 -66.483 53.693
410380 94.829 -66.498 53.670
410381 94.871 -66.512 53.650
410382 94.907 -66.520 53.631
410383 94.944 -66.529 53.613
410384 94.981 -66.537 53.596
410385 95.018 -66.545 53.578
410386 95.054 -66.554 53.560
410387 95.090 -66.562 53.540
410388 95.120 -66.569 53.518
410389 95.151 -66.576 53.495
410390 95.178 -66.577 53.469
410391 95.204 -66.577 53.442
410392 95.231 -66.578 53.415
410393 95.258 -66.579 53.391
410394 95.283 -66.578 53.371
410395 95.304 -66.572 53.354
410396 95.341 -66.573 53.340
410397 95.388 -66.578 53.329
410398 95.443 -66.579 53.318
410399 95.500 -66.577 53.309
410400 95.556 -66.577 53.301
410401 95.596 -66.585 53.293
410402 95.633 -66.595 53.283
410403 95.677 -66.601 53.267
410404 95.722 -66.607 53.249
410405 95.767 -66.614 53.232
410406 95.813 -66.621 53.215
410407 95.851 -66.626 53.196
410408 95.890 -66.630 53.175
410409 95.932 -66.630 53.155
410410 95.973 -66.630 53.135
410411 96.015 -66.630 53.115
410412 96.057 -66.630 53.094
410413 96.097 -66.631 53.071
410414 96.135 -66.633 53.046
410415 96.174 -66.635 53.019
410416 96.215 -66.634 52.988
410417 96.256 -66.633 52.955
410418 96.295 -66.633 52.923
410419 96.330 -66.636 52.888
410420 96.367 -66.639 52.853
410421 96.402 -66.642 52.816
410422 96.439 -66.645 52.786
410423 96.475 -66.648 52.757
410424 96.510 -66.648 52.730
410425 96.538 -66.642 52.703
410426 96.567 -66.636 52.676
410427 96.596 -66.629 52.652
410428 96.624 -66.623 52.629
410429 96.653 -66.616 52.603
410430 96.692 -66.618 52.572
410431 96.734 -66.615 52.538
410432 96.775 -66.612 52.506
410433 96.816 -66.610 52.475
410434 96.858 -66.607 52.449
410435 96.899 -66.605 52.427
410436 96.942 -66.602 52.403
410437 96.987 -66.598 52.381
410438 97.033 -66.594 52.360
410439 97.078 -66.590 52.341
410440 97.123 -66.586 52.322
410441 97.168 -66.585 52.304
410442 97.210 -66.596 52.284
410443 97.251 -66.606 52.261
410444 97.293 -66.617 52.236
410445 97.334 -66.627 52.207
410446 97.375 -66.631 52.174
410447 97.415 -66.636 52.140
410448 97.455 -66.640 52.102
410449 97.502 -66.647 52.063
410450 97.550 -66.655 52.022
410451 97.589 -66.667 51.982
410452 97.627 -66.679 51.942
410453 97.665 -66.691 51.901
410454 97.704 -66.703 51.861
410455 97.742 -66.715 51.817
410456 97.780 -66.727 51.774
410457 97.818 -66.740 51.729
410458 97.857 -66.752 51.684
410459 97.895 -66.765 51.640
410460 97.933 -66.777 51.596
410461 97.971 -66.788 51.551
410462 97.996 -66.796 51.506
410463 98.021 -66.802 51.461
410464 98.047 -66.809 51.418
410465 98.073 -66.816 51.375
410466 98.098 -66.824 51.332
410467 98.124 -66.831 51.291
410468 98.149 -66.837 51.248
410469 98.175 -66.844 51.206
410470 98.201 -66.852 51.163
410471 98.226 -66.859 51.121
410472 98.251 -66.865 51.078
410473 98.277 -66.872 51.036
410474 98.301 -66.872 50.996
410475 98.323 -66.869 50.957
410476 98.349 -66.866 50.920
410477 98.383 -66.862 50.883
410478 98.421 -66.864 50.842
410479 98.456 -66.871 50.799
410480 98.489 -66.880 50.755
410481 98.521 -66.889 50.708
410482 98.555 -66.897 50.660
410483 98.589 -66.903 50.610
410484 98.624 -66.907 50.559
410485 98.660 -66.911 50.507
410486 98.698 -66.923 50.455
410487 98.745 -66.924 50.402
410488 98.792 -66.925 50.348
410489 98.839 -66.928 50.294
410490 98.885 -66.929 50.239
410491 98.931 -66.931 50.184
410492 98.978 -66.934 50.131
410493 99.023 -66.935 50.077
410494 99.070 -66.937 50.024
410495 99.116 -66.940 49.970
410496 99.162 -66.942 49.917
410497 99.208 -66.944 49.863
410498 99.251 -66.955 49.810
410499 99.294 -66.968 49.757
410500 99.338 -66.981 49.704
410501 99.363 -66.981 49.653
410502 99.384 -66.978 49.601
410503 99.405 -66.974 49.549
410504 99.443 -66.985 49.497
410505 99.486 -67.003 49.443
410506 99.529 -67.019 49.392
410507 99.573 -67.037 49.340
410508 99.616 -67.053 49.287
410509 99.659 -67.071 49.236
410510 99.705 -67.088 49.186
410511 99.757 -67.103 49.136
410512 99.809 -67.118 49.086
410513 99.837 -67.110 49.038
410514 99.849 -67.119 48.990
410515 99.869 -67.126 48.943
410516 99.890 -67.134 48.897
410517 99.910 -67.142 48.851
410518 99.937 -67.147 48.805
410519 99.977 -67.144 48.762
410520 100.016 -67.142 48.718
410521 100.056 -67.141 48.675
410522 100.096 -67.138 48.632
410523 100.135 -67.136 48.589
410524 100.175 -67.135 48.547
410525 100.214 -67.132 48.506
410526 100.260 -67.130 48.466
410527 100.308 -67.128 48.426
410528 100.356 -67.126 48.386
410986 90.675 -66.219 54.965
410987 90.723 -66.229 54.971
410988 90.772 -66.238 54.976
410989 90.820 -66.248 54.981
410990 90.869 -66.258 54.987
410991 90.918 -66.267 54.993
410992 90.966 -66.277 54.999
410993 91.015 -66.287 55.005
410994 91.064 -66.296 55.012
410995 91.112 -66.306 55.020
410996 91.161 -66.316 55.029
410997 91.209 -66.323 55.041
410998 91.257 -66.330 55.051
410999 91.303 -66.342 55.059
411000 91.344 -66.342 55.066
411001 91.384 -66.342 55.072
411002 91.425 -66.342 55.077
411003 91.466 -66.343 55.079
411004 91.507 -66.344 55.079
411005 91.553 -66.358 55.076
411006 91.598 -66.373 55.071
411007 91.642 -66.378 55.064
411008 91.686 -66.379 55.059
411009 91.730 -66.379 55.056
411010 91.779 -66.385 55.053
411011 91.834 -66.398 55.050
411012 91.889 -66.411 55.046
411013 91.944 -66.423 55.041
411014 92.000 -66.436 55.035
411015 92.055 -66.448 55.029
411016 92.110 -66.461 55.021
411017 92.166 -66.474 55.013
411018 92.218 -66.478 55.005
411019 92.269 -66.477 54.997
411020 92.319 -66.477 54.988
411021 92.370 -66.476 54.981
411022 92.421 -66.476 54.973
411023 92.471 -66.482 54.965
411024 92.519 -66.492 54.958
411025 92.567 -66.501 54.956
411026 92.616 -66.511 54.957
411027 92.662 -66.512 54.957
411028 92.706 -66.503 54.951
411029 92.749 -66.495 54.939
411030 92.793 -66.486 54.918
411031 92.837 -66.477 54.894
411032 92.880 -66.468 54.867
411033 92.924 -66.460 54.840
411034 92.968 -66.451 54.815
411035 93.012 -66.442 54.796
411036 93.053 -66.431 54.780
411037 93.091 -66.417 54.762
411038 93.129 -66.403 54.739
411039 93.167 -66.389 54.708
411040 93.205 -66.374 54.671
411041 93.243 -66.361 54.633
411042 93.281 -66.346 54.595
411043 93.318 -66.335 54.559
411044 93.352 -66.331 54.527
411045 93.386 -66.326 54.496
411046 93.420 -66.320 54.468
411047 93.454 -66.316 54.443
411048 93.488 -66.310 54.420
411049 93.525 -66.313 54.403
411050 93.564 -66.320 54.388
411051 93.607 -66.324 54.375
411052 93.649 -66.327 54.363
411053 93.694 -66.335 54.351
411054 93.738 -66.342 54.338
411055 93.783 -66.350 54.323
411056 93.827 -66.357 54.308
411057 93.871 -66.365 54.292
411058 93.916 -66.372 54.274
411059 93.959 -66.378 54.253
411060 94.000 -66.382 54.231
411061 94.041 -66.387 54.209
411062 94.082 -66.392 54.186
411063 94.123 -66.396 54.167
411064 94.164 -66.401 54.147
411065 94.205 -66.405 54.127
411066 94.245 -66.410 54.105
411067 94.275 -66.411 54.077
411068 94.305 -66.412 54.046
411069 94.336 -66.414 54.015
411070 94.366 -66.415 53.983
411071 94.397 -66.417 53.952
411072 94.427 -66.418 53.923
411073 94.471 -66.421 53.896
411074 94.523 -66.424 53.871
411075 94.574 -66.428 53.845
411076 94.623 -66.432 53.819
411077 94.664 -66.433 53.791
411078 94.703 -66.435 53.764
411079 94.744 -66.436 53.739
411080 94.785 -66.447 53.716
411081 94.827 -66.462 53.697
411082 94.868 -66.476 53.678
411083 94.905 -66.484 53.659
411084 94.942 -66.492 53.641
411085 94.978 -66.501 53.621
411086 95.014 -66.509 53.601
411087 95.044 -66.515 53.579
411088 95.075 -66.522 53.555
411089 95.106 -66.529 53.529
411090 95.135 -66.533 53.503
411091 95.162 -66.534 53.475
411092 95.188 -66.535 53.448
411093 95.215 -66.536 53.422
411094 95.242 -66.537 53.399
411095 95.268 -66.538 53.380
411096 95.295 -66.539 53.364
411097 95.333 -66.540 53.351
411098 95.389 -66.539 53.341
411099 95.445 -66.538 53.330
411100 95.502 -66.537 53.322
411101 95.558 -66.536 53.314
411102 95.592 -66.548 53.305
411103 95.624 -66.560 53.295
411104 95.668 -66.567 53.279
411105 95.712 -66.573 53.261
411106 95.756 -66.579 53.244
411107 95.801 -66.585 53.227
411108 95.842 -66.589 53.209
411109 95.884 -66.589 53.188
411110 95.926 -66.589 53.168
411111 95.968 -66.590 53.148
411112 96.010 -66.590 53.128
411113 96.051 -66.590 53.108
411114 96.093 -66.590 53.087
411115 96.132 -66.592 53.063
411116 96.172 -66.593 53.036
411117 96.213 -66.592 53.006
411118 96.254 -66.592 52.973
411119 96.295 -66.591 52.939
411120 96.332 -66.592 52.904
411121 96.369 -66.595 52.869
411122 96.405 -66.597 52.832
411123 96.442 -66.599 52.801
411124 96.478 -66.601 52.773
411125 96.510 -66.597 52.746
411126 96.538 -66.591 52.718
411127 96.568 -66.586 52.693
411128 96.597 -66.580 52.668
411129 96.625 -66.574 52.644
411130 96.658 -66.571 52.618
411131 96.700 -66.566 52.585
411132 96.741 -66.563 52.552
411133 96.783 -66.560 52.518
411134 96.824 -66.558 52.486
411135 96.866 -66.555 52.458
411136 96.907 -66.552 52.434
411137 96.949 -66.549 52.411
411138 96.995 -66.545 52.388
411139 97.040 -66.542 52.366
411140 97.086 -66.538 52.346
411141 97.129 -66.541 52.328
411142 97.171 -66.552 52.311
411143 97.213 -66.561 52.291
411144 97.254 -66.572 52.266
411145 97.296 -66.583 52.237
411146 97.336 -66.591 52.205
411147 97.377 -66.596 52.171
411148 97.417 -66.600 52.135
411149 97.457 -66.605 52.097
411150 97.499 -66.609 52.058
411151 97.547 -66.617 52.016
411152 97.586 -66.630 51.975
411153 97.624 -66.642 51.933
411154 97.662 -66.654 51.892
411155 97.700 -66.666 51.852
411156 97.738 -66.678 51.810
411157 97.777 -66.690 51.767
411158 97.815 -66.703 51.723
411159 97.854 -66.716 51.678
411160 97.893 -66.729 51.634
411161 97.932 -66.741 51.591
411162 97.968 -66.748 51.547
411163 97.997 -66.754 51.503
411164 98.020 -66.762 51.457
411165 98.042 -66.771 51.414
411166 98.065 -66.779 51.372
411167 98.091 -66.786 51.332
411168 98.116 -66.793 51.293
411169 98.142 -66.800 51.253
411170 98.167 -66.807 51.213
411171 98.193 -66.814 51.171
411172 98.219 -66.821 51.128
411173 98.244 -66.828 51.084
411174 98.270 -66.835 51.038
411175 98.292 -66.832 50.995
411176 98.319 -66.829 50.954
411177 98.353 -66.825 50.914
411178 98.388 -66.821 50.877
411179 98.426 -66.823 50.837
411180 98.465 -66.826 50.796
411181 98.499 -66.834 50.753
411182 98.532 -66.843 50.707
411183 98.565 -66.851 50.659
411184 98.598 -66.860 50.609
411185 98.631 -66.868 50.558
411186 98.663 -66.877 50.506
411187 98.705 -66.883 50.455
411188 98.751 -66.885 50.402
411189 98.797 -66.887 50.347
411190 98.844 -66.890 50.292
411191 98.890 -66.891 50.236
411192 98.936 -66.893 50.181
411193 98.982 -66.896 50.127
411194 99.028 -66.897 50.074
411195 99.075 -66.900 50.021
411196 99.121 -66.902 49.967
411197 99.166 -66.904 49.913
411198 99.212 -66.906 49.859
411199 99.255 -66.920 49.805
411200 99.298 -66.934 49.752
411201 99.341 -66.946 49.698
411202 99.384 -66.960 49.646
411203 99.407 -66.957 49.594
411204 99.427 -66.953 49.542
411205 99.448 -66.950 49.490
411206 99.469 -66.945 49.436
411207 99.508 -66.958 49.384
411208 99.551 -66.975 49.333
411209 99.594 -66.992 49.281
411210 99.638 -67.009 49.229
411211 99.686 -67.025 49.179
411212 99.737 -67.041 49.129
411213 99.789 -67.056 49.079
411214 99.837 -67.066 49.031
411215 99.855 -67.075 48.984
411216 99.875 -67.082 48.937
411217 99.896 -67.090 48.891
411218 99.916 -67.098 48.844
411219 99.946 -67.100 48.798
411220 99.986 -67.098 48.755
411221 100.025 -67.097 48.712
411222 100.065 -67.094 48.668
411223 100.105 -67.092 48.624
411224 100.144 -67.091 48.582
411225 100.184 -67.088 48.541
411226 100.226 -67.086 48.500
411227 100.274 -67.084 48.459
411228 100.322 -67.082 48.419
411229 100.370 -67.080 48.378
411687 90.674 -66.176 54.965
411688 90.723 -66.185 54.970
411689 90.771 -66.195 54.975
411690 90.819 -66.205 54.980
411691 90.868 -66.214 54.985
411692 90.917 -66.224 54.991
411693 90.965 -66.234 54.997
411694 91.014 -66.243 55.002
411695 91.063 -66.253 55.007
411696 91.111 -66.262 55.014
411697 91.159 -66.268 55.023
411698 91.208 -66.274 55.034
411699 91.255 -66.283 55.045
411700 91.301 -66.295 55.054
411701 91.345 -66.301 55.062
411702 91.385 -66.302 55.066
411703 91.426 -66.303 55.070
411704 91.467 -66.303 55.073
411705 91.508 -66.305 55.074
411706 91.554 -66.320 55.072
411707 91.599 -66.335 55.067
411708 91.645 -66.349 55.063
411709 91.689 -66.354 55.061
411710 91.733 -66.355 55.059
411711 91.781 -66.360 55.057
411712 91.836 -66.372 55.055
411713 91.891 -66.385 55.053
411714 91.946 -66.398 55.050
411715 92.002 -66.411 55.047
411716 92.057 -66.423 55.042
411717 92.109 -66.426 55.035
411718 92.160 -66.426 55.028
411719 92.210 -66.426 55.020
411720 92.261 -66.425 55.011
411721 92.312 -66.425 55.002
411722 92.363 -66.424 54.994
411723 92.412 -66.430 54.986
411724 92.460 -66.440 54.978
411725 92.509 -66.450 54.972
411726 92.557 -66.460 54.972
411727 92.605 -66.469 54.972
411728 92.650 -66.465 54.970
411729 92.694 -66.457 54.964
411730 92.738 -66.448 54.950
411731 92.782 -66.439 54.929
411732 92.826 -66.430 54.904
411733 92.869 -66.422 54.878
411734 92.913 -66.413 54.852
411735 92.957 -66.404 54.830
411736 93.001 -66.395 54.812
411737 93.044 -66.387 54.795
411738 93.088 -66.378 54.777
411739 93.126 -66.364 54.754
411740 93.164 -66.350 54.720
411741 93.202 -66.335 54.683
411742 93.240 -66.321 54.644
411743 93.278 -66.307 54.606
411744 93.315 -66.295 54.571
411745 93.350 -66.290 54.540
411746 93.383 -66.285 54.512
411747 93.418 -66.280 54.486
411748 93.452 -66.275 54.463
411749 93.486 -66.270 54.441
411750 93.521 -66.267 54.425
411751 93.561 -66.273 54.411
411752 93.605 -66.278 54.398
411753 93.649 -66.285 54.385
411754 93.694 -66.293 54.373
411755 93.738 -66.301 54.358
411756 93.782 -66.307 54.343
411757 93.827 -66.315 54.328
411758 93.871 -66.323 54.311
411759 93.916 -66.330 54.291
411760 93.960 -66.338 54.269
411761 94.001 -66.342 54.246
411762 94.042 -66.347 54.222
411763 94.083 -66.351 54.200
411764 94.124 -66.356 54.179
411765 94.166 -66.361 54.160
411766 94.206 -66.365 54.139
411767 94.246 -66.370 54.116
411768 94.285 -66.376 54.088
411769 94.312 -66.376 54.057
411770 94.339 -66.376 54.026
411771 94.367 -66.376 53.995
411772 94.394 -66.376 53.964
411773 94.421 -66.376 53.936
411774 94.461 -66.377 53.909
411775 94.512 -66.381 53.883
411776 94.564 -66.384 53.858
411777 94.615 -66.388 53.832
411778 94.661 -66.391 53.806
411779 94.701 -66.392 53.782
411780 94.741 -66.396 53.761
411781 94.783 -66.411 53.744
411782 94.825 -66.426 53.726
411783 94.865 -66.439 53.707
411784 94.902 -66.448 53.688
411785 94.937 -66.455 53.668
411786 94.968 -66.462 53.646
411787 94.999 -66.469 53.622
411788 95.030 -66.476 53.597
411789 95.060 -66.482 53.570
411790 95.091 -66.489 53.542
411791 95.119 -66.492 53.512
411792 95.146 -66.492 53.484
411793 95.173 -66.493 53.456
411794 95.199 -66.494 53.430
411795 95.226 -66.495 53.408
411796 95.261 -66.499 53.390
411797 95.296 -66.503 53.375
411798 95.339 -66.503 53.363
411799 95.391 -66.498 53.353
411800 95.447 -66.497 53.343
411801 95.504 -66.496 53.335
411802 95.557 -66.497 53.326
411803 95.589 -66.510 53.317
411804 95.620 -66.523 53.306
411805 95.659 -66.532 53.291
411806 95.703 -66.538 53.273
411807 95.748 -66.544 53.255
411808 95.795 -66.548 53.237
411809 95.839 -66.551 53.220
411810 95.881 -66.551 53.200
411811 95.922 -66.551 53.180
411812 95.963 -66.551 53.160
411813 96.005 -66.551 53.140
411814 96.046 -66.551 53.121
411815 96.088 -66.551 53.100
411816 96.129 -66.552 53.076
411817 96.170 -66.552 53.049
411818 96.212 -66.550 53.020
411819 96.254 -66.548 52.986
411820 96.297 -66.545 52.951
411821 96.335 -66.545 52.916
411822 96.372 -66.548 52.881
411823 96.409 -66.549 52.846
411824 96.446 -66.552 52.815
411825 96.480 -66.552 52.787
411826 96.509 -66.546 52.760
411827 96.538 -66.540 52.734
411828 96.568 -66.535 52.710
411829 96.597 -66.530 52.685
411830 96.630 -66.527 52.661
411831 96.664 -66.524 52.634
411832 96.703 -66.520 52.602
411833 96.747 -66.514 52.568
411834 96.790 -66.508 52.533
411835 96.832 -66.505 52.498
411836 96.873 -66.502 52.467
411837 96.915 -66.499 52.439
411838 96.957 -66.496 52.415
411839 97.002 -66.492 52.392
411840 97.047 -66.489 52.370
411841 97.090 -66.496 52.351
411842 97.132 -66.507 52.332
411843 97.173 -66.517 52.314
411844 97.215 -66.528 52.293
411845 97.256 -66.538 52.268
411846 97.298 -66.549 52.239
411847 97.338 -66.555 52.206
411848 97.379 -66.560 52.170
411849 97.419 -66.564 52.133
411850 97.459 -66.568 52.095
411851 97.500 -66.573 52.054
411852 97.544 -66.580 52.013
411853 97.582 -66.592 51.970
411854 97.620 -66.605 51.927
411855 97.659 -66.617 51.885
411856 97.698 -66.630 51.844
411857 97.737 -66.643 51.802
411858 97.776 -66.656 51.760
411859 97.815 -66.669 51.718
411860 97.854 -66.682 51.675
411861 97.892 -66.694 51.632
411862 97.931 -66.702 51.591
411863 97.965 -66.708 51.548
411864 97.998 -66.712 51.504
411865 98.023 -66.720 51.460
411866 98.045 -66.728 51.417
411867 98.068 -66.737 51.376
411868 98.091 -66.745 51.338
411869 98.113 -66.754 51.301
411870 98.136 -66.762 51.261
411871 98.160 -66.770 51.220
411872 98.185 -66.777 51.178
411873 98.211 -66.784 51.135
411874 98.237 -66.791 51.090
411875 98.261 -66.795 51.044
411876 98.289 -66.791 50.996
411877 98.324 -66.787 50.951
411878 98.358 -66.784 50.910
411879 98.393 -66.780 50.870
411880 98.431 -66.782 50.830
411881 98.471 -66.785 50.790
411882 98.509 -66.788 50.748
411883 98.542 -66.796 50.704
411884 98.575 -66.805 50.658
411885 98.608 -66.814 50.609
411886 98.641 -66.822 50.559
411887 98.674 -66.831 50.507
411888 98.715 -66.838 50.455
411889 98.758 -66.846 50.402
411890 98.802 -66.849 50.348
411891 98.848 -66.852 50.293
411892 98.895 -66.853 50.237
411893 98.941 -66.856 50.182
411894 98.987 -66.858 50.128
411895 99.033 -66.860 50.073
411896 99.079 -66.862 50.019
411897 99.125 -66.864 49.966
411898 99.170 -66.866 49.912
411899 99.215 -66.872 49.858
411900 99.259 -66.885 49.804
411901 99.301 -66.899 49.750
411902 99.345 -66.912 49.696
411903 99.388 -66.925 49.641
411904 99.429 -66.937 49.588
411905 99.449 -66.933 49.536
411906 99.471 -66.929 49.483
411907 99.491 -66.925 49.430
411908 99.512 -66.921 49.378
411909 99.533 -66.917 49.326
411910 99.573 -66.931 49.275
411911 99.616 -66.948 49.223
411912 99.666 -66.964 49.172
411913 99.718 -66.979 49.122
411914 99.778 -67.001 49.073
411915 99.838 -67.022 49.025
411916 99.860 -67.030 48.978
411917 99.881 -67.038 48.932
411918 99.901 -67.046 48.885
411919 99.921 -67.053 48.838
411920 99.956 -67.054 48.792
411921 99.995 -67.053 48.749
411922 100.034 -67.050 48.705
411923 100.074 -67.048 48.662
411924 100.114 -67.047 48.619
411925 100.153 -67.044 48.576
411926 100.193 -67.042 48.535
411927 100.240 -67.040 48.494
411928 100.288 -67.038 48.453
411929 100.337 -67.036 48.413
411930 100.385 -67.034 48.372
412388 90.673 -66.132 54.966
412389 90.722 -66.142 54.970
412390 90.770 -66.152 54.976
412391 90.818 -66.161 54.980
412392 90.867 -66.171 54.984
412393 90.916 -66.181 54.989
412394 90.964 -66.190 54.993
412395 91.013 -66.200 54.999
412396 91.062 -66.207 55.003
412397 91.110 -66.213 55.009
412398 91.158 -66.219 55.017
412399 91.207 -66.226 55.027
412400 91.253 -66.236 55.037
412401 91.299 -66.248 55.047
412402 91.345 -66.259 55.056
412403 91.386 -66.262 55.062
412404 91.427 -66.263 55.068
412405 91.468 -66.263 55.072
412406 91.510 -66.266 55.073
412407 91.555 -66.281 55.071
412408 91.601 -66.296 55.066
412409 91.646 -66.310 55.064
412410 91.692 -66.326 55.064
412411 91.736 -66.331 55.063
412412 91.782 -66.334 55.060
412413 91.838 -66.347 55.060
412414 91.893 -66.360 55.060
412415 91.948 -66.373 55.059
412416 92.000 -66.374 55.057
412417 92.051 -66.374 55.055
412418 92.101 -66.374 55.049
412419 92.152 -66.373 55.042
412420 92.203 -66.373 55.034
412421 92.253 -66.373 55.026
412422 92.304 -66.373 55.017
412423 92.353 -66.379 55.009
412424 92.402 -66.389 54.999
412425 92.450 -66.398 54.991
412426 92.499 -66.408 54.986
412427 92.547 -66.418 54.986
412428 92.595 -66.427 54.985
412429 92.639 -66.419 54.982
412430 92.683 -66.410 54.974
412431 92.726 -66.401 54.956
412432 92.770 -66.392 54.933
412433 92.814 -66.384 54.908
412434 92.858 -66.375 54.882
412435 92.902 -66.367 54.857
412436 92.945 -66.357 54.836
412437 92.989 -66.349 54.820
412438 93.033 -66.340 54.805
412439 93.077 -66.332 54.787
412440 93.121 -66.323 54.763
412441 93.161 -66.311 54.731
412442 93.199 -66.297 54.695
412443 93.237 -66.282 54.657
412444 93.275 -66.269 54.620
412445 93.313 -66.255 54.586
412446 93.347 -66.250 54.556
412447 93.381 -66.245 54.529
412448 93.415 -66.240 54.505
412449 93.449 -66.235 54.483
412450 93.484 -66.229 54.462
412451 93.518 -66.224 54.446
412452 93.560 -66.229 54.433
412453 93.605 -66.235 54.420
412454 93.649 -66.243 54.407
412455 93.693 -66.251 54.394
412456 93.737 -66.258 54.379
412457 93.782 -66.266 54.363
412458 93.827 -66.273 54.347
412459 93.871 -66.281 54.330
412460 93.916 -66.288 54.309
412461 93.960 -66.296 54.286
412462 94.002 -66.301 54.262
412463 94.043 -66.306 54.238
412464 94.084 -66.310 54.214
412465 94.125 -66.315 54.192
412466 94.166 -66.320 54.171
412467 94.205 -66.326 54.149
412468 94.246 -66.331 54.125
412469 94.285 -66.336 54.096
412470 94.325 -66.342 54.066
412471 94.352 -66.342 54.035
412472 94.380 -66.342 54.005
412473 94.407 -66.342 53.977
412474 94.434 -66.342 53.951
412475 94.473 -66.342 53.924
412476 94.514 -66.341 53.898
412477 94.555 -66.341 53.872
412478 94.605 -66.344 53.845
412479 94.657 -66.348 53.821
412480 94.698 -66.350 53.800
412481 94.739 -66.361 53.784
412482 94.781 -66.376 53.769
412483 94.823 -66.391 53.753
412484 94.861 -66.402 53.734
412485 94.892 -66.409 53.715
412486 94.923 -66.416 53.694
412487 94.954 -66.422 53.669
412488 94.984 -66.429 53.644
412489 95.015 -66.436 53.617
412490 95.046 -66.442 53.588
412491 95.076 -66.448 53.559
412492 95.103 -66.449 53.528
412493 95.130 -66.450 53.498
412494 95.157 -66.451 53.470
412495 95.192 -66.455 53.444
412496 95.228 -66.460 53.421
412497 95.263 -66.464 53.404
412498 95.299 -66.469 53.389
412499 95.346 -66.466 53.376
412500 95.397 -66.461 53.367
412501 95.450 -66.457 53.358
412502 95.506 -66.455 53.350
412503 95.553 -66.460 53.339
412504 95.585 -66.473 53.328
412505 95.616 -66.486 53.317
412506 95.652 -66.496 53.302
412507 95.698 -66.501 53.285
412508 95.745 -66.505 53.266
412509 95.792 -66.509 53.247
412510 95.838 -66.513 53.230
412511 95.879 -66.513 53.212
412512 95.920 -66.513 53.192
412513 95.962 -66.513 53.172
412514 96.003 -66.513 53.151
412515 96.044 -66.512 53.132
412516 96.085 -66.512 53.111
412517 96.127 -66.511 53.086
412518 96.170 -66.507 53.059
412519 96.213 -66.505 53.030
412520 96.255 -66.502 52.996
412521 96.298 -66.500 52.961
412522 96.339 -66.498 52.926
412523 96.376 -66.501 52.891
412524 96.412 -66.502 52.859
412525 96.449 -66.505 52.829
412526 96.480 -66.501 52.801
412527 96.509 -66.495 52.775
412528 96.538 -66.489 52.751
412529 96.570 -66.486 52.726
412530 96.603 -66.483 52.702
412531 96.637 -66.480 52.677
412532 96.670 -66.478 52.650
412533 96.707 -66.474 52.620
412534 96.750 -66.468 52.584
412535 96.794 -66.463 52.547
412536 96.837 -66.457 52.510
412537 96.880 -66.451 52.475
412538 96.923 -66.447 52.445
412539 96.964 -66.444 52.419
412540 97.009 -66.441 52.396
412541 97.051 -66.451 52.374
412542 97.092 -66.462 52.354
412543 97.134 -66.473 52.335
412544 97.176 -66.483 52.315
412545 97.217 -66.494 52.293
412546 97.259 -66.504 52.268
412547 97.300 -66.514 52.239
412548 97.341 -66.519 52.206
412549 97.381 -66.523 52.169
412550 97.421 -66.528 52.131
412551 97.461 -66.533 52.092
412552 97.502 -66.537 52.051
412553 97.541 -66.543 52.009
412554 97.580 -66.557 51.966
412555 97.619 -66.570 51.923
412556 97.658 -66.583 51.880
412557 97.697 -66.596 51.840
412558 97.736 -66.609 51.798
412559 97.775 -66.622 51.757
412560 97.814 -66.635 51.716
412561 97.853 -66.647 51.674
412562 97.891 -66.655 51.635
412563 97.929 -66.664 51.595
412564 97.962 -66.668 51.553
412565 97.996 -66.672 51.509
412566 98.025 -66.677 51.465
412567 98.048 -66.686 51.422
412568 98.071 -66.694 51.380
412569 98.093 -66.702 51.342
412570 98.116 -66.711 51.305
412571 98.139 -66.720 51.265
412572 98.161 -66.728 51.224
412573 98.184 -66.737 51.182
412574 98.207 -66.745 51.140
412575 98.229 -66.753 51.095
412576 98.260 -66.754 51.048
412577 98.294 -66.750 50.998
412578 98.329 -66.746 50.951
412579 98.363 -66.742 50.908
412580 98.398 -66.738 50.867
412581 98.436 -66.740 50.828
412582 98.476 -66.743 50.788
412583 98.514 -66.746 50.747
412584 98.552 -66.750 50.703
412585 98.585 -66.759 50.658
412586 98.618 -66.768 50.609
412587 98.651 -66.776 50.559
412588 98.685 -66.785 50.507
412589 98.727 -66.792 50.455
412590 98.769 -66.799 50.401
412591 98.811 -66.806 50.347
412592 98.853 -66.814 50.292
412593 98.900 -66.815 50.237
412594 98.946 -66.818 50.182
412595 98.992 -66.820 50.128
412596 99.038 -66.822 50.073
412597 99.083 -66.824 50.019
412598 99.129 -66.826 49.965
412599 99.174 -66.828 49.911
412600 99.219 -66.837 49.857
412601 99.262 -66.850 49.802
412602 99.305 -66.864 49.747
412603 99.348 -66.878 49.693
412604 99.392 -66.890 49.640
412605 99.435 -66.904 49.586
412606 99.472 -66.912 49.533
412607 99.493 -66.909 49.480
412608 99.513 -66.905 49.426
412609 99.535 -66.901 49.374
412610 99.555 -66.896 49.322
412611 99.576 -66.893 49.270
412612 99.597 -66.888 49.218
412613 99.661 -66.914 49.166
412614 99.725 -66.939 49.115
412615 99.790 -66.964 49.067
412616 99.841 -66.978 49.019
412617 99.866 -66.986 48.973
412618 99.887 -66.994 48.926
412619 99.907 -67.002 48.879
412620 99.927 -67.009 48.832
412621 99.965 -67.009 48.786
412622 100.004 -67.006 48.743
412623 100.044 -67.004 48.699
412624 100.083 -67.003 48.656
412625 100.123 -67.000 48.614
412626 100.163 -66.998 48.571
412627 100.206 -66.996 48.529
412628 100.255 -66.994 48.488
412629 100.303 -66.992 48.448
412630 100.351 -66.990 48.408
412631 100.399 -66.987 48.367
413089 90.672 -66.089 54.967
413090 90.721 -66.099 54.971
413091 90.769 -66.108 54.976
413092 90.818 -66.118 54.980
413093 90.866 -66.128 54.983
413094 90.915 -66.137 54.988
413095 90.963 -66.145 54.991
413096 91.012 -66.151 54.995
413097 91.060 -66.157 54.999
413098 91.108 -66.164 55.004
413099 91.157 -66.170 55.010
413100 91.205 -66.177 55.018
413101 91.251 -66.189 55.027
413102 91.296 -66.201 55.037
413103 91.342 -66.212 55.047
413104 91.387 -66.222 55.054
413105 91.428 -66.223 55.063
413106 91.469 -66.223 55.068
413107 91.511 -66.227 55.071
413108 91.557 -66.242 55.071
413109 91.602 -66.257 55.070
413110 91.648 -66.272 55.067
413111 91.693 -66.287 55.066
413112 91.738 -66.301 55.065
413113 91.784 -66.309 55.064
413114 91.840 -66.321 55.064
413115 91.891 -66.323 55.064
413116 91.941 -66.323 55.064
413117 91.992 -66.322 55.064
413118 92.043 -66.322 55.064
413119 92.094 -66.322 55.062
413120 92.144 -66.321 55.056
413121 92.195 -66.321 55.049
413122 92.246 -66.320 55.041
413123 92.295 -66.327 55.032
413124 92.343 -66.337 55.024
413125 92.391 -66.347 55.013
413126 92.440 -66.357 55.004
413127 92.489 -66.367 55.000
413128 92.537 -66.376 54.999
413129 92.584 -66.380 54.998
413130 92.628 -66.372 54.993
413131 92.671 -66.363 54.982
413132 92.715 -66.354 54.961
413133 92.759 -66.345 54.936
413134 92.803 -66.337 54.909
413135 92.847 -66.329 54.883
413136 92.890 -66.320 54.859
413137 92.934 -66.311 54.839
413138 92.978 -66.302 54.825
413139 93.021 -66.294 54.810
413140 93.065 -66.285 54.793
413141 93.109 -66.276 54.771
413142 93.153 -66.267 54.741
413143 93.196 -66.258 54.706
413144 93.234 -66.244 54.671
413145 93.272 -66.229 54.635
413146 93.310 -66.216 54.602
413147 93.345 -66.210 54.573
413148 93.379 -66.204 54.546
413149 93.413 -66.199 54.522
413150 93.447 -66.194 54.500
413151 93.483 -66.191 54.481
413152 93.521 -66.191 54.466
413153 93.566 -66.191 54.452
413154 93.609 -66.197 54.439
413155 93.652 -66.204 54.428
413156 93.696 -66.210 54.415
413157 93.740 -66.217 54.399
413158 93.783 -66.225 54.383
413159 93.827 -66.232 54.366
413160 93.871 -66.239 54.348
413161 93.915 -66.246 54.328
413162 93.960 -66.254 54.305
413163 94.003 -66.260 54.282
413164 94.045 -66.265 54.257
413165 94.085 -66.270 54.231
413166 94.125 -66.276 54.205
413167 94.165 -66.281 54.180
413168 94.205 -66.286 54.156
413169 94.245 -66.292 54.131
413170 94.285 -66.297 54.102
413171 94.325 -66.302 54.072
413172 94.365 -66.307 54.043
413173 94.392 -66.307 54.015
413174 94.420 -66.307 53.990
413175 94.451 -66.307 53.967
413176 94.491 -66.307 53.941
413177 94.532 -66.307 53.914
413178 94.572 -66.307 53.886
413179 94.613 -66.307 53.858
413180 94.653 -66.307 53.835
413181 94.695 -66.310 53.818
413182 94.736 -66.325 53.805
413183 94.776 -66.340 53.792
413184 94.816 -66.355 53.777
413185 94.854 -66.369 53.760
413186 94.883 -66.373 53.740
413187 94.911 -66.378 53.718
413188 94.939 -66.382 53.692
413189 94.970 -66.389 53.667
413190 95.000 -66.395 53.640
413191 95.032 -66.402 53.611
413192 95.060 -66.407 53.581
413193 95.088 -66.407 53.549
413194 95.124 -66.412 53.517
413195 95.159 -66.417 53.488
413196 95.195 -66.421 53.461
413197 95.231 -66.425 53.438
413198 95.266 -66.429 53.420
413199 95.302 -66.434 53.404
413200 95.352 -66.429 53.391
413201 95.403 -66.423 53.382
413202 95.454 -66.418 53.373
413203 95.508 -66.415 53.365
413204 95.549 -66.423 53.353
413205 95.581 -66.437 53.340
413206 95.611 -66.454 53.327
413207 95.650 -66.463 53.312
413208 95.695 -66.461 53.295
413209 95.742 -66.466 53.277
413210 95.789 -66.470 53.259
413211 95.836 -66.475 53.241
413212 95.877 -66.475 53.223
413213 95.918 -66.475 53.203
413214 95.960 -66.474 53.182
413215 96.001 -66.474 53.160
413216 96.042 -66.474 53.138
413217 96.085 -66.472 53.116
413218 96.128 -66.467 53.092
413219 96.171 -66.463 53.065
413220 96.214 -66.460 53.034
413221 96.256 -66.458 53.000
413222 96.298 -66.455 52.967
413223 96.341 -66.453 52.934
413224 96.380 -66.453 52.902
413225 96.416 -66.455 52.873
413226 96.451 -66.455 52.843
413227 96.480 -66.450 52.816
413228 96.509 -66.444 52.791
413229 96.542 -66.442 52.767
413230 96.576 -66.439 52.741
413231 96.609 -66.436 52.718
413232 96.643 -66.434 52.693
413233 96.677 -66.431 52.667
413234 96.710 -66.429 52.636
413235 96.754 -66.423 52.600
413236 96.797 -66.417 52.561
413237 96.840 -66.411 52.521
413238 96.884 -66.405 52.483
413239 96.927 -66.400 52.451
413240 96.971 -66.397 52.422
413241 97.012 -66.407 52.398
413242 97.053 -66.417 52.376
413243 97.095 -66.428 52.356
413244 97.137 -66.439 52.336
413245 97.178 -66.449 52.317
413246 97.220 -66.460 52.293
413247 97.261 -66.470 52.265
413248 97.302 -66.479 52.236
413249 97.343 -66.483 52.203
413250 97.383 -66.488 52.167
413251 97.423 -66.492 52.128
413252 97.463 -66.497 52.088
413253 97.502 -66.502 52.047
413254 97.540 -66.510 52.005
413255 97.580 -66.523 51.962
413256 97.618 -66.536 51.919
413257 97.658 -66.549 51.878
413258 97.697 -66.562 51.838
413259 97.736 -66.575 51.797
413260 97.775 -66.589 51.757
413261 97.813 -66.601 51.717
413262 97.852 -66.608 51.677
413263 97.890 -66.617 51.640
413264 97.926 -66.624 51.602
413265 97.959 -66.627 51.560
413266 97.993 -66.632 51.517
413267 98.025 -66.636 51.472
413268 98.051 -66.643 51.427
413269 98.074 -66.652 51.385
413270 98.096 -66.660 51.345
413271 98.119 -66.669 51.308
413272 98.142 -66.677 51.268
413273 98.164 -66.686 51.227
413274 98.187 -66.694 51.185
413275 98.210 -66.702 51.143
413276 98.232 -66.711 51.098
413277 98.265 -66.712 51.051
413278 98.299 -66.708 51.001
413279 98.334 -66.705 50.952
413280 98.368 -66.701 50.908
413281 98.403 -66.697 50.866
413282 98.441 -66.699 50.827
413283 98.481 -66.702 50.787
413284 98.519 -66.705 50.746
413285 98.559 -66.708 50.702
413286 98.595 -66.713 50.657
413287 98.628 -66.721 50.608
413288 98.661 -66.730 50.558
413289 98.696 -66.739 50.506
413290 98.738 -66.746 50.454
413291 98.781 -66.753 50.400
413292 98.822 -66.760 50.345
413293 98.865 -66.768 50.291
413294 98.908 -66.774 50.237
413295 98.950 -66.780 50.182
413296 98.996 -66.782 50.128
413297 99.042 -66.784 50.073
413298 99.087 -66.787 50.018
413299 99.133 -66.788 49.963
413300 99.179 -66.790 49.909
413301 99.222 -66.802 49.855
413302 99.266 -66.816 49.800
413303 99.309 -66.829 49.745
413304 99.352 -66.843 49.691
413305 99.395 -66.856 49.639
413306 99.438 -66.869 49.586
413307 99.482 -66.883 49.532
413308 99.515 -66.888 49.477
413309 99.536 -66.884 49.423
413310 99.557 -66.881 49.371
413311 99.577 -66.877 49.318
413312 99.588 -66.856 49.265
413313 99.609 -66.851 49.212
413314 99.673 -66.877 49.160
413315 99.737 -66.902 49.110
413316 99.799 -66.925 49.061
413317 99.845 -66.934 49.014
413318 99.871 -66.942 48.967
413319 99.892 -66.950 48.921
413320 99.912 -66.958 48.873
413321 99.934 -66.965 48.826
413322 99.974 -66.962 48.781
413323 100.013 -66.960 48.737
413324 100.053 -66.959 48.693
413325 100.093 -66.956 48.650
413326 100.132 -66.954 48.607
413327 100.173 -66.953 48.566
413328 100.221 -66.950 48.525
413329 100.269 -66.948 48.484
413330 100.317 -66.946 48.443
413331 100.365 -66.943 48.403
413790 90.671 -66.046 54.968
413791 90.720 -66.055 54.972
413792 90.768 -66.065 54.976
413793 90.817 -66.075 54.980
413794 90.865 -66.084 54.983
413795 90.913 -66.090 54.987
413796 90.962 -66.097 54.990
413797 91.011 -66.103 54.993
413798 91.059 -66.109 54.996
413799 91.107 -66.115 55.000
413800 91.156 -66.121 55.004
413801 91.203 -66.130 55.009
413802 91.248 -66.141 55.017
413803 91.294 -66.154 55.026
413804 91.340 -66.165 55.035
413805 91.386 -66.176 55.044
413806 91.429 -66.182 55.055
413807 91.470 -66.183 55.063
413808 91.513 -66.188 55.067
413809 91.558 -66.204 55.069
413810 91.603 -66.219 55.070
413811 91.649 -66.233 55.067
413812 91.695 -66.248 55.065
413813 91.740 -66.263 55.065
413814 91.784 -66.270 55.065
413815 91.832 -66.271 55.065
413816 91.883 -66.270 55.065
413817 91.934 -66.270 55.066
413818 91.985 -66.270 55.067
413819 92.035 -66.270 55.069
413820 92.086 -66.270 55.068
413821 92.137 -66.269 55.066
413822 92.188 -66.269 55.062
413823 92.236 -66.276 55.057
413824 92.285 -66.286 55.050
413825 92.333 -66.295 55.040
413826 92.381 -66.305 55.029
413827 92.430 -66.315 55.019
413828 92.479 -66.325 55.014
413829 92.527 -66.335 55.012
413830 92.572 -66.334 55.010
413831 92.616 -66.325 55.003
413832 92.660 -66.317 54.988
413833 92.704 -66.307 54.965
413834 92.747 -66.299 54.937
413835 92.791 -66.290 54.909
413836 92.835 -66.282 54.883
413837 92.879 -66.273 54.859
413838 92.923 -66.264 54.840
413839 92.967 -66.255 54.826
413840 93.010 -66.247 54.813
413841 93.054 -66.238 54.798
413842 93.097 -66.229 54.779
413843 93.141 -66.220 54.749
413844 93.185 -66.212 54.717
413845 93.229 -66.203 54.684
413846 93.269 -66.191 54.651
413847 93.307 -66.177 54.618
413848 93.342 -66.169 54.590
413849 93.376 -66.164 54.564
413850 93.411 -66.159 54.540
413851 93.448 -66.158 54.519
413852 93.486 -66.157 54.501
413853 93.526 -66.157 54.486
413854 93.570 -66.157 54.472
413855 93.615 -66.159 54.459
413856 93.658 -66.166 54.448
413857 93.702 -66.172 54.435
413858 93.745 -66.179 54.420
413859 93.788 -66.186 54.404
413860 93.832 -66.193 54.387
413861 93.875 -66.200 54.368
413862 93.919 -66.207 54.347
413863 93.962 -66.213 54.324
413864 94.005 -66.220 54.299
413865 94.045 -66.226 54.273
413866 94.084 -66.231 54.246
413867 94.124 -66.236 54.217
413868 94.165 -66.241 54.189
413869 94.205 -66.247 54.163
413870 94.245 -66.252 54.136
413871 94.285 -66.257 54.108
413872 94.325 -66.263 54.080
413873 94.365 -66.268 54.053
413874 94.405 -66.273 54.028
413875 94.432 -66.273 54.006
413876 94.468 -66.273 53.983
413877 94.508 -66.273 53.956
413878 94.549 -66.273 53.926
413879 94.586 -66.274 53.896
413880 94.621 -66.276 53.870
413881 94.655 -66.279 53.850
413882 94.689 -66.282 53.835
413883 94.726 -66.289 53.823
413884 94.766 -66.304 53.813
413885 94.807 -66.320 53.799
413886 94.846 -66.335 53.783
413887 94.878 -66.342 53.763
413888 94.906 -66.346 53.741
413889 94.934 -66.351 53.715
413890 94.963 -66.356 53.691
413891 94.991 -66.361 53.666
413892 95.020 -66.365 53.637
413893 95.055 -66.369 53.607
413894 95.091 -66.373 53.573
413895 95.126 -66.377 53.540
413896 95.162 -66.382 53.509
413897 95.197 -66.386 53.480
413898 95.233 -66.390 53.457
413899 95.269 -66.395 53.438
413900 95.308 -66.397 53.421
413901 95.359 -66.392 53.407
413902 95.410 -66.386 53.397
413903 95.461 -66.381 53.389
413904 95.512 -66.376 53.379
413905 95.544 -66.391 53.366
413906 95.574 -66.408 53.352
413907 95.605 -66.426 53.337
413908 95.648 -66.429 53.322
413909 95.693 -66.428 53.306
413910 95.739 -66.427 53.288
413911 95.786 -66.432 53.270
413912 95.833 -66.436 53.251
413913 95.876 -66.436 53.233
413914 95.917 -66.436 53.212
413915 95.958 -66.436 53.190
413916 95.999 -66.436 53.166
413917 96.042 -66.433 53.144
413918 96.085 -66.429 53.121
413919 96.128 -66.424 53.096
413920 96.172 -66.420 53.069
413921 96.215 -66.416 53.039
413922 96.257 -66.413 53.007
413923 96.299 -66.411 52.976
413924 96.341 -66.407 52.947
413925 96.383 -66.406 52.919
413926 96.420 -66.408 52.890
413927 96.451 -66.404 52.860
413928 96.482 -66.400 52.831
413929 96.515 -66.398 52.804
413930 96.549 -66.395 52.781
413931 96.582 -66.392 52.756
413932 96.616 -66.390 52.731
413933 96.649 -66.387 52.707
413934 96.683 -66.385 52.681
413935 96.717 -66.382 52.651
413936 96.757 -66.377 52.615
413937 96.801 -66.371 52.575
413938 96.844 -66.366 52.533
413939 96.887 -66.360 52.493
413940 96.931 -66.361 52.457
413941 96.975 -66.364 52.426
413942 97.016 -66.373 52.399
413943 97.056 -66.384 52.375
413944 97.097 -66.394 52.354
413945 97.139 -66.404 52.334
413946 97.180 -66.415 52.315
413947 97.222 -66.426 52.291
413948 97.263 -66.436 52.264
413949 97.304 -66.443 52.234
413950 97.345 -66.448 52.199
413951 97.385 -66.452 52.163
413952 97.424 -66.457 52.125
413953 97.463 -66.462 52.085
413954 97.501 -66.467 52.043
413955 97.540 -66.476 52.001
413956 97.579 -66.489 51.960
413957 97.618 -66.502 51.919
413958 97.657 -66.515 51.879
413959 97.696 -66.528 51.841
413960 97.735 -66.542 51.803
413961 97.774 -66.554 51.764
413962 97.812 -66.561 51.725
413963 97.851 -66.570 51.686
413964 97.889 -66.578 51.647
413965 97.924 -66.583 51.607
413966 97.956 -66.588 51.565
413967 97.990 -66.592 51.520
413968 98.023 -66.596 51.475
413969 98.054 -66.601 51.432
413970 98.077 -66.609 51.390
413971 98.099 -66.617 51.349
413972 98.122 -66.627 51.310
413973 98.145 -66.635 51.271
413974 98.167 -66.643 51.230
413975 98.190 -66.652 51.187
413976 98.213 -66.660 51.142
413977 98.235 -66.668 51.096
413978 98.270 -66.669 51.047
413979 98.304 -66.667 50.997
413980 98.339 -66.663 50.949
413981 98.373 -66.659 50.905
413982 98.408 -66.655 50.864
413983 98.447 -66.658 50.824
413984 98.486 -66.661 50.785
413985 98.524 -66.664 50.744
413986 98.564 -66.667 50.700
413987 98.602 -66.670 50.655
413988 98.639 -66.676 50.606
413989 98.671 -66.684 50.556
413990 98.708 -66.693 50.505
413991 98.750 -66.699 50.452
413992 98.792 -66.707 50.399
413993 98.834 -66.714 50.345
413994 98.884 -66.722 50.291
413995 98.932 -66.730 50.237
413996 98.958 -66.740 50.183
413997 99.000 -66.744 50.127
413998 99.046 -66.746 50.071
413999 99.092 -66.749 50.016
414000 99.138 -66.751 49.961
414001 99.183 -66.755 49.905
414002 99.226 -66.768 49.850
414003 99.269 -66.781 49.795
414004 99.312 -66.794 49.742
414005 99.356 -66.808 49.690
414006 99.399 -66.821 49.637
414007 99.442 -66.834 49.584
414008 99.485 -66.848 49.531
414009 99.528 -66.861 49.477
414010 99.558 -66.864 49.421
414011 99.573 -66.849 49.367
414012 99.582 -66.825 49.313
414013 99.591 -66.802 49.260
414014 99.620 -66.814 49.207
414015 99.685 -66.840 49.155
414016 99.749 -66.865 49.105
414017 99.803 -66.881 49.056
414018 99.849 -66.890 49.008
414019 99.877 -66.898 48.960
414020 99.898 -66.906 48.914
414021 99.918 -66.913 48.867
414022 99.943 -66.918 48.821
414023 99.983 -66.916 48.775
414024 100.022 -66.915 48.731
414025 100.062 -66.912 48.688
414026 100.102 -66.910 48.646
414027 100.141 -66.909 48.603
414028 100.187 -66.906 48.561
414029 100.235 -66.904 48.519
414030 100.283 -66.902 48.478
414031 100.332 -66.900 48.438
414032 100.380 -66.898 48.398
414491 90.670 -66.002 54.969
414492 90.719 -66.012 54.973
414493 90.767 -66.022 54.976
414494 90.816 -66.029 54.979
414495 90.864 -66.035 54.983
414496 90.912 -66.041 54.985
414497 90.961 -66.047 54.988
414498 91.009 -66.053 54.990
414499 91.058 -66.060 54.992
414500 91.106 -66.066 54.995
414501 91.155 -66.072 54.998
414502 91.201 -66.083 55.003
414503 91.246 -66.094 55.007
414504 91.292 -66.106 55.014
414505 91.338 -66.118 55.022
414506 91.384 -66.129 55.032
414507 91.430 -66.141 55.045
414508 91.471 -66.143 55.054
414509 91.514 -66.150 55.060
414510 91.560 -66.165 55.065
414511 91.605 -66.179 55.066
414512 91.650 -66.194 55.064
414513 91.694 -66.201 55.062
414514 91.737 -66.207 55.062
414515 91.781 -66.213 55.063
414516 91.825 -66.219 55.063
414517 91.875 -66.218 55.064
414518 91.926 -66.218 55.065
414519 91.977 -66.217 55.067
414520 92.028 -66.217 55.071
414521 92.078 -66.216 55.071
414522 92.129 -66.216 55.071
414523 92.178 -66.225 55.070
414524 92.226 -66.235 55.068
414525 92.275 -66.245 55.062
414526 92.323 -66.254 55.053
414527 92.371 -66.263 55.043
414528 92.420 -66.273 55.033
414529 92.468 -66.283 55.027
414530 92.517 -66.293 55.025
414531 92.561 -66.287 55.021
414532 92.605 -66.279 55.011
414533 92.649 -66.270 54.993
414534 92.693 -66.261 54.966
414535 92.736 -66.252 54.936
414536 92.780 -66.244 54.908
414537 92.824 -66.235 54.882
414538 92.867 -66.226 54.858
414539 92.911 -66.217 54.840
414540 92.955 -66.209 54.826
414541 92.999 -66.200 54.813
414542 93.043 -66.191 54.800
414543 93.086 -66.182 54.783
414544 93.130 -66.174 54.757
414545 93.174 -66.165 54.727
414546 93.218 -66.157 54.696
414547 93.262 -66.147 54.664
414548 93.304 -66.138 54.635
414549 93.340 -66.129 54.609
414550 93.376 -66.126 54.583
414551 93.414 -66.125 54.560
414552 93.451 -66.124 54.540
414553 93.489 -66.123 54.522
414554 93.531 -66.123 54.506
414555 93.575 -66.123 54.493
414556 93.620 -66.123 54.480
414557 93.664 -66.127 54.467
414558 93.707 -66.134 54.455
414559 93.751 -66.141 54.441
414560 93.794 -66.147 54.425
414561 93.838 -66.154 54.407
414562 93.881 -66.162 54.388
414563 93.924 -66.169 54.366
414564 93.968 -66.176 54.341
414565 94.010 -66.182 54.316
414566 94.049 -66.187 54.288
414567 94.089 -66.192 54.260
414568 94.128 -66.197 54.230
414569 94.168 -66.202 54.201
414570 94.207 -66.207 54.173
414571 94.247 -66.213 54.146
414572 94.287 -66.218 54.118
414573 94.326 -66.223 54.091
414574 94.366 -66.229 54.066
414575 94.405 -66.234 54.041
414576 94.445 -66.239 54.019
414577 94.479 -66.241 53.994
414578 94.513 -66.244 53.965
414579 94.548 -66.247 53.934
414580 94.582 -66.249 53.904
414581 94.616 -66.251 53.880
414582 94.650 -66.254 53.863
414583 94.685 -66.257 53.850
414584 94.719 -66.259 53.839
414585 94.757 -66.269 53.831
414586 94.797 -66.284 53.818
414587 94.837 -66.299 53.802
414588 94.873 -66.310 53.783
414589 94.901 -66.315 53.762
414590 94.930 -66.320 53.738
414591 94.958 -66.324 53.716
414592 94.987 -66.329 53.692
414593 95.017 -66.333 53.664
414594 95.055 -66.335 53.635
414595 95.094 -66.339 53.601
414596 95.129 -66.342 53.567
414597 95.165 -66.347 53.535
414598 95.200 -66.351 53.505
414599 95.236 -66.356 53.480
414600 95.271 -66.360 53.458
414601 95.315 -66.360 53.439
414602 95.366 -66.354 53.424
414603 95.416 -66.349 53.413
414604 95.465 -66.346 53.403
414605 95.503 -66.351 53.392
414606 95.538 -66.362 53.379
414607 95.568 -66.380 53.363
414608 95.600 -66.397 53.347
414609 95.645 -66.395 53.331
414610 95.691 -66.394 53.315
414611 95.737 -66.393 53.298
414612 95.783 -66.392 53.280
414613 95.830 -66.397 53.261
414614 95.874 -66.398 53.242
414615 95.915 -66.398 53.220
414616 95.956 -66.398 53.196
414617 95.999 -66.395 53.172
414618 96.042 -66.390 53.148
414619 96.086 -66.386 53.124
414620 96.129 -66.381 53.099
414621 96.172 -66.376 53.072
414622 96.216 -66.373 53.044
414623 96.258 -66.370 53.016
414624 96.300 -66.367 52.988
414625 96.342 -66.364 52.963
414626 96.385 -66.361 52.938
414627 96.422 -66.359 52.909
414628 96.455 -66.356 52.877
414629 96.488 -66.354 52.846
414630 96.522 -66.351 52.817
414631 96.555 -66.348 52.793
414632 96.589 -66.346 52.769
414633 96.622 -66.343 52.744
414634 96.656 -66.341 52.719
414635 96.689 -66.339 52.693
414636 96.723 -66.335 52.664
414637 96.760 -66.332 52.629
414638 96.804 -66.326 52.588
414639 96.848 -66.323 52.546
414640 96.892 -66.325 52.503
414641 96.936 -66.327 52.465
414642 96.979 -66.330 52.430
414643 97.020 -66.340 52.400
414644 97.061 -66.351 52.374
414645 97.101 -66.361 52.351
414646 97.142 -66.370 52.332
414647 97.183 -66.381 52.312
414648 97.224 -66.392 52.289
414649 97.266 -66.402 52.262
414650 97.306 -66.407 52.232
414651 97.346 -66.412 52.198
414652 97.385 -66.417 52.161
414653 97.424 -66.422 52.124
414654 97.462 -66.427 52.084
414655 97.501 -66.433 52.042
414656 97.539 -66.442 52.000
414657 97.579 -66.455 51.960
414658 97.617 -66.468 51.921
414659 97.657 -66.482 51.884
414660 97.696 -66.495 51.848
414661 97.734 -66.507 51.810
414662 97.773 -66.515 51.771
414663 97.811 -66.523 51.733
414664 97.850 -66.531 51.694
414665 97.887 -66.539 51.653
414666 97.921 -66.543 51.612
414667 97.953 -66.548 51.568
414668 97.987 -66.552 51.523
414669 98.020 -66.556 51.479
414670 98.053 -66.560 51.437
414671 98.080 -66.567 51.395
414672 98.102 -66.575 51.354
414673 98.125 -66.584 51.314
414674 98.148 -66.592 51.273
414675 98.170 -66.601 51.231
414676 98.193 -66.609 51.186
414677 98.215 -66.617 51.140
414678 98.239 -66.625 51.091
414679 98.274 -66.625 51.042
414680 98.309 -66.625 50.992
414681 98.344 -66.622 50.944
414682 98.378 -66.617 50.899
414683 98.413 -66.614 50.857
414684 98.452 -66.617 50.818
414685 98.491 -66.620 50.778
414686 98.530 -66.623 50.737
414687 98.569 -66.626 50.694
414688 98.607 -66.629 50.651
414689 98.647 -66.632 50.604
414690 98.681 -66.638 50.555
414691 98.720 -66.646 50.505
414692 98.762 -66.653 50.452
414693 98.810 -66.661 50.399
414694 98.861 -66.669 50.345
414695 98.912 -66.677 50.291
414696 98.951 -66.686 50.238
414697 98.978 -66.696 50.183
414698 99.005 -66.706 50.126
414699 99.050 -66.708 50.070
414700 99.096 -66.711 50.014
414701 99.142 -66.713 49.958
414702 99.187 -66.720 49.902
414703 99.229 -66.733 49.847
414704 99.273 -66.746 49.793
414705 99.316 -66.760 49.740
414706 99.359 -66.773 49.689
414707 99.403 -66.787 49.636
414708 99.445 -66.799 49.582
414709 99.489 -66.813 49.529
414710 99.532 -66.821 49.474
414711 99.566 -66.819 49.418
414712 99.575 -66.795 49.363
414713 99.584 -66.771 49.310
414714 99.595 -66.762 49.256
414715 99.632 -66.777 49.202
414716 99.696 -66.802 49.150
414717 99.760 -66.827 49.100
414718 99.806 -66.837 49.050
414719 99.853 -66.846 49.002
414720 99.883 -66.853 48.954
414721 99.903 -66.862 48.908
414722 99.923 -66.869 48.862
414723 99.952 -66.872 48.815
414724 99.992 -66.871 48.769
414725 100.032 -66.868 48.726
414726 100.071 -66.866 48.684
414727 100.111 -66.865 48.641
414728 100.153 -66.862 48.598
414729 100.201 -66.860 48.556
414730 100.250 -66.858 48.513
414731 100.298 -66.856 48.474
414732 100.346 -66.854 48.434
414733 100.394 -66.852 48.394
415192 90.669 -65.959 54.970
415193 90.718 -65.968 54.974
415194 90.766 -65.974 54.976
415195 90.814 -65.980 54.979
415196 90.863 -65.986 54.981
415197 90.911 -65.992 54.984
415198 90.960 -65.998 54.985
415199 91.008 -66.005 54.987
415200 91.057 -66.011 54.988
415201 91.105 -66.017 54.990
415202 91.153 -66.025 54.991
415203 91.199 -66.036 54.995
415204 91.244 -66.047 54.999
415205 91.290 -66.059 55.004
415206 91.336 -66.070 55.011
415207 91.381 -66.082 55.022
415208 91.427 -66.094 55.036
415209 91.472 -66.103 55.046
415210 91.516 -66.111 55.053
415211 91.561 -66.126 55.058
415212 91.604 -66.133 55.059
415213 91.648 -66.139 55.059
415214 91.692 -66.144 55.059
415215 91.735 -66.151 55.059
415216 91.778 -66.157 55.060
415217 91.821 -66.163 55.059
415218 91.868 -66.166 55.059
415219 91.919 -66.166 55.060
415220 91.969 -66.165 55.062
415221 92.020 -66.165 55.067
415222 92.071 -66.164 55.069
415223 92.120 -66.173 55.071
415224 92.168 -66.183 55.073
415225 92.216 -66.193 55.074
415226 92.265 -66.203 55.070
415227 92.313 -66.213 55.063
415228 92.361 -66.223 55.054
415229 92.410 -66.232 55.044
415230 92.458 -66.241 55.037
415231 92.506 -66.249 55.034
415232 92.550 -66.240 55.028
415233 92.593 -66.232 55.015
415234 92.637 -66.223 54.995
415235 92.681 -66.214 54.967
415236 92.725 -66.206 54.936
415237 92.769 -66.197 54.908
415238 92.812 -66.188 54.882
415239 92.856 -66.179 54.858
415240 92.900 -66.171 54.840
415241 92.943 -66.162 54.825
415242 92.987 -66.154 54.813
415243 93.031 -66.144 54.801
415244 93.075 -66.136 54.785
415245 93.119 -66.127 54.762
415246 93.162 -66.119 54.734
415247 93.206 -66.110 54.704
415248 93.250 -66.101 54.674
415249 93.296 -66.097 54.647
415250 93.342 -66.093 54.623
415251 93.379 -66.092 54.600
415252 93.417 -66.091 54.579
415253 93.454 -66.091 54.562
415254 93.492 -66.090 54.546
415255 93.536 -66.090 54.529
415256 93.580 -66.089 54.514
415257 93.625 -66.089 54.499
415258 93.670 -66.089 54.486
415259 93.713 -66.096 54.474
415260 93.756 -66.103 54.461
415261 93.800 -66.110 54.446
415262 93.843 -66.116 54.429
415263 93.887 -66.123 54.408
415264 93.930 -66.130 54.383
415265 93.974 -66.137 54.357
415266 94.014 -66.142 54.330
415267 94.054 -66.147 54.302
415268 94.093 -66.153 54.273
415269 94.133 -66.158 54.244
415270 94.172 -66.163 54.216
415271 94.212 -66.169 54.187
415272 94.251 -66.174 54.162
415273 94.291 -66.179 54.134
415274 94.331 -66.185 54.106
415275 94.370 -66.190 54.079
415276 94.410 -66.194 54.054
415277 94.439 -66.204 54.029
415278 94.474 -66.211 54.002
415279 94.509 -66.219 53.972
415280 94.544 -66.222 53.941
415281 94.578 -66.224 53.912
415282 94.612 -66.226 53.891
415283 94.646 -66.229 53.875
415284 94.681 -66.232 53.863
415285 94.715 -66.234 53.855
415286 94.749 -66.237 53.847
415287 94.787 -66.248 53.836
415288 94.828 -66.263 53.820
415289 94.868 -66.279 53.802
415290 94.897 -66.283 53.781
415291 94.925 -66.288 53.760
415292 94.954 -66.292 53.739
415293 94.982 -66.297 53.716
415294 95.016 -66.301 53.690
415295 95.054 -66.304 53.662
415296 95.093 -66.306 53.629
415297 95.131 -66.308 53.595
415298 95.168 -66.312 53.561
415299 95.203 -66.317 53.530
415300 95.239 -66.321 53.503
415301 95.274 -66.326 53.480
415302 95.321 -66.323 53.459
415303 95.372 -66.317 53.442
415304 95.418 -66.316 53.427
415305 95.457 -66.321 53.416
415306 95.495 -66.326 53.404
415307 95.532 -66.333 53.391
415308 95.563 -66.351 53.376
415309 95.598 -66.364 53.360
415310 95.643 -66.362 53.343
415311 95.689 -66.361 53.326
415312 95.735 -66.359 53.308
415313 95.781 -66.358 53.290
415314 95.827 -66.358 53.270
415315 95.872 -66.361 53.249
415316 95.913 -66.361 53.225
415317 95.956 -66.356 53.200
415318 96.000 -66.351 53.176
415319 96.043 -66.348 53.150
415320 96.086 -66.343 53.126
415321 96.129 -66.339 53.101
415322 96.173 -66.334 53.076
415323 96.217 -66.329 53.052
415324 96.259 -66.326 53.027
415325 96.301 -66.323 53.004
415326 96.343 -66.320 52.980
415327 96.386 -66.317 52.955
415328 96.425 -66.314 52.926
415329 96.459 -66.310 52.893
415330 96.493 -66.308 52.859
415331 96.526 -66.305 52.828
415332 96.560 -66.303 52.803
415333 96.594 -66.300 52.781
415334 96.628 -66.297 52.757
415335 96.662 -66.295 52.731
415336 96.695 -66.292 52.705
415337 96.729 -66.289 52.676
415338 96.764 -66.286 52.642
415339 96.808 -66.287 52.602
415340 96.853 -66.289 52.559
415341 96.896 -66.291 52.515
415342 96.941 -66.293 52.473
415343 96.984 -66.297 52.436
415344 97.024 -66.307 52.402
415345 97.065 -66.317 52.373
415346 97.106 -66.327 52.348
415347 97.146 -66.337 52.327
415348 97.187 -66.348 52.309
415349 97.227 -66.357 52.287
415350 97.268 -66.367 52.261
415351 97.307 -66.372 52.231
415352 97.346 -66.377 52.198
415353 97.384 -66.382 52.162
415354 97.423 -66.387 52.125
415355 97.461 -66.392 52.085
415356 97.500 -66.397 52.041
415357 97.539 -66.408 52.001
415358 97.578 -66.421 51.962
415359 97.617 -66.435 51.926
415360 97.656 -66.448 51.891
415361 97.695 -66.460 51.856
415362 97.733 -66.468 51.817
415363 97.772 -66.476 51.779
415364 97.810 -66.484 51.741
415365 97.848 -66.492 51.700
415366 97.884 -66.499 51.659
415367 97.918 -66.503 51.616
415368 97.950 -66.508 51.571
415369 97.984 -66.511 51.526
415370 98.017 -66.516 51.484
415371 98.050 -66.520 51.443
415372 98.082 -66.524 51.401
415373 98.105 -66.533 51.360
415374 98.128 -66.542 51.318
415375 98.150 -66.550 51.275
415376 98.173 -66.558 51.231
415377 98.196 -66.567 51.184
415378 98.218 -66.575 51.136
415379 98.244 -66.581 51.085
415380 98.279 -66.581 51.036
415381 98.314 -66.581 50.987
415382 98.349 -66.580 50.939
415383 98.383 -66.577 50.892
415384 98.418 -66.573 50.850
415385 98.457 -66.576 50.810
415386 98.496 -66.579 50.770
415387 98.535 -66.582 50.729
415388 98.574 -66.585 50.688
415389 98.613 -66.587 50.646
415390 98.652 -66.590 50.601
415391 98.691 -66.593 50.553
415392 98.735 -66.600 50.505
415393 98.786 -66.608 50.453
415394 98.837 -66.616 50.400
415395 98.888 -66.624 50.346
415396 98.939 -66.632 50.292
415397 98.971 -66.642 50.238
415398 98.997 -66.652 50.183
415399 99.023 -66.663 50.125
415400 99.055 -66.671 50.068
415401 99.100 -66.673 50.011
415402 99.146 -66.675 49.955
415403 99.190 -66.685 49.900
415404 99.233 -66.699 49.846
415405 99.276 -66.711 49.792
415406 99.320 -66.725 49.740
415407 99.363 -66.738 49.688
415408 99.406 -66.750 49.635
415409 99.448 -66.755 49.580
415410 99.491 -66.762 49.525
415411 99.534 -66.767 49.470
415412 99.569 -66.765 49.414
415413 99.578 -66.740 49.360
415414 99.587 -66.716 49.306
415415 99.598 -66.722 49.252
415416 99.644 -66.740 49.197
415417 99.708 -66.765 49.146
415418 99.764 -66.784 49.094
415419 99.810 -66.793 49.045
415420 99.856 -66.801 48.996
415421 99.888 -66.809 48.949
415422 99.909 -66.817 48.903
415423 99.929 -66.825 48.856
415424 99.962 -66.827 48.809
415425 100.001 -66.824 48.764
415426 100.041 -66.822 48.721
415427 100.081 -66.821 48.679
415428 100.120 -66.818 48.636
415429 100.168 -66.816 48.593
415430 100.216 -66.814 48.551
415431 100.264 -66.812 48.509
415432 100.312 -66.810 48.469
415433 100.360 -66.808 48.430
415892 90.619 -65.906 54.968
415893 90.668 -65.912 54.971
415894 90.716 -65.919 54.974
415895 90.765 -65.925 54.976
415896 90.813 -65.931 54.978
415897 90.862 -65.937 54.980
415898 90.910 -65.944 54.981
415899 90.958 -65.950 54.982
415900 91.007 -65.956 54.982
415901 91.055 -65.962 54.983
415902 91.104 -65.968 54.983
415903 91.151 -65.978 54.985
415904 91.196 -65.989 54.987
415905 91.242 -66.000 54.991
415906 91.288 -66.012 54.995
415907 91.333 -66.023 55.001
415908 91.379 -66.035 55.012
415909 91.425 -66.047 55.025
415910 91.471 -66.058 55.036
415911 91.515 -66.065 55.045
415912 91.559 -66.070 55.050
415913 91.602 -66.076 55.052
415914 91.645 -66.082 55.053
415915 91.688 -66.088 55.055
415916 91.732 -66.094 55.057
415917 91.775 -66.100 55.057
415918 91.818 -66.106 55.057
415919 91.862 -66.112 55.057
415920 91.911 -66.113 55.056
415921 91.962 -66.113 55.057
415922 92.012 -66.113 55.059
415923 92.061 -66.122 55.062
415924 92.109 -66.132 55.066
415925 92.158 -66.141 55.069
415926 92.206 -66.151 55.069
415927 92.254 -66.161 55.066
415928 92.303 -66.171 55.062
415929 92.351 -66.181 55.056
415930 92.400 -66.191 55.048
415931 92.448 -66.200 55.042
415932 92.495 -66.202 55.037
415933 92.538 -66.194 55.028
415934 92.582 -66.185 55.014
415935 92.626 -66.176 54.994
415936 92.669 -66.167 54.966
415937 92.713 -66.159 54.936
415938 92.757 -66.150 54.908
415939 92.801 -66.141 54.883
415940 92.845 -66.132 54.859
415941 92.888 -66.124 54.840
415942 92.932 -66.115 54.825
415943 92.976 -66.107 54.813
415944 93.020 -66.097 54.801
415945 93.064 -66.089 54.787
415946 93.108 -66.081 54.765
415947 93.151 -66.072 54.739
415948 93.198 -66.069 54.711
415949 93.247 -66.071 54.682
415950 93.295 -66.066 54.658
415951 93.343 -66.060 54.637
415952 93.382 -66.058 54.617
415953 93.420 -66.057 54.599
415954 93.458 -66.056 54.584
415955 93.496 -66.056 54.569
415956 93.541 -66.056 54.552
415957 93.585 -66.056 54.536
415958 93.630 -66.056 54.520
415959 93.675 -66.056 54.507
415960 93.718 -66.057 54.495
415961 93.762 -66.064 54.481
415962 93.805 -66.071 54.467
415963 93.849 -66.078 54.450
415964 93.893 -66.085 54.429
415965 93.936 -66.091 54.404
415966 93.979 -66.098 54.376
415967 94.018 -66.103 54.348
415968 94.058 -66.109 54.318
415969 94.098 -66.114 54.290
415970 94.137 -66.119 54.262
415971 94.177 -66.125 54.234
415972 94.216 -66.130 54.206
415973 94.256 -66.135 54.180
415974 94.295 -66.140 54.150
415975 94.335 -66.144 54.120
415976 94.374 -66.148 54.090
415977 94.409 -66.154 54.063
415978 94.432 -66.169 54.034
415979 94.468 -66.176 54.005
415980 94.503 -66.183 53.976
415981 94.538 -66.191 53.946
415982 94.573 -66.198 53.921
415983 94.608 -66.202 53.902
415984 94.642 -66.204 53.888
415985 94.676 -66.207 53.877
415986 94.710 -66.210 53.869
415987 94.745 -66.212 53.862
415988 94.779 -66.214 53.852
415989 94.818 -66.228 53.837
415990 94.858 -66.242 53.819
415991 94.892 -66.251 53.800
415992 94.920 -66.257 53.780
415993 94.949 -66.261 53.760
415994 94.977 -66.266 53.739
415995 95.014 -66.269 53.715
415996 95.053 -66.271 53.687
415997 95.091 -66.273 53.655
415998 95.130 -66.276 53.622
415999 95.168 -66.279 53.588
416000 95.206 -66.282 53.554
416001 95.241 -66.286 53.525
416002 95.277 -66.291 53.501
416003 95.328 -66.285 53.479
416004 95.372 -66.286 53.458
416005 95.410 -66.292 53.440
416006 95.448 -66.297 53.428
416007 95.486 -66.302 53.417
416008 95.525 -66.307 53.404
416009 95.557 -66.323 53.390
416010 95.596 -66.330 53.374
416011 95.642 -66.329 53.356
416012 95.688 -66.327 53.338
416013 95.734 -66.326 53.319
416014 95.779 -66.324 53.300
416015 95.825 -66.323 53.278
416016 95.870 -66.323 53.257
416017 95.913 -66.318 53.233
416018 95.957 -66.314 53.207
416019 96.000 -66.309 53.182
416020 96.043 -66.304 53.157
416021 96.087 -66.300 53.132
416022 96.130 -66.295 53.108
416023 96.174 -66.291 53.083
416024 96.217 -66.286 53.061
416025 96.260 -66.282 53.039
416026 96.302 -66.279 53.017
416027 96.344 -66.276 52.995
416028 96.387 -66.273 52.971
416029 96.425 -66.270 52.941
416030 96.460 -66.266 52.909
416031 96.494 -66.264 52.874
416032 96.528 -66.261 52.841
416033 96.562 -66.259 52.815
416034 96.596 -66.256 52.793
416035 96.629 -66.253 52.768
416036 96.663 -66.251 52.742
416037 96.697 -66.248 52.716
416038 96.731 -66.245 52.687
416039 96.770 -66.250 52.653
416040 96.813 -66.253 52.613
416041 96.857 -66.255 52.570
416042 96.901 -66.257 52.525
416043 96.945 -66.259 52.480
416044 96.988 -66.263 52.441
416045 97.029 -66.273 52.407
416046 97.070 -66.284 52.376
416047 97.110 -66.294 52.348
416048 97.151 -66.304 52.324
416049 97.191 -66.314 52.304
416050 97.232 -66.324 52.283
416051 97.269 -66.331 52.258
416052 97.306 -66.337 52.229
416053 97.345 -66.342 52.197
416054 97.384 -66.347 52.162
416055 97.423 -66.352 52.124
416056 97.461 -66.357 52.085
416057 97.500 -66.362 52.044
416058 97.538 -66.375 52.005
416059 97.578 -66.388 51.969
416060 97.616 -66.401 51.934
416061 97.655 -66.413 51.899
416062 97.694 -66.421 51.866
416063 97.732 -66.429 51.827
416064 97.771 -66.437 51.788
416065 97.809 -66.445 51.749
416066 97.847 -66.454 51.707
416067 97.881 -66.459 51.666
416068 97.915 -66.463 51.622
416069 97.948 -66.467 51.576
416070 97.981 -66.472 51.532
416071 98.014 -66.476 51.489
416072 98.047 -66.480 51.448
416073 98.080 -66.484 51.406
416074 98.108 -66.490 51.364
416075 98.131 -66.499 51.321
416076 98.153 -66.508 51.276
416077 98.176 -66.516 51.229
416078 98.199 -66.524 51.180
416079 98.221 -66.533 51.130
416080 98.249 -66.538 51.080
416081 98.284 -66.537 51.030
416082 98.318 -66.537 50.981
416083 98.353 -66.537 50.934
416084 98.388 -66.535 50.889
416085 98.423 -66.532 50.846
416086 98.462 -66.535 50.804
416087 98.501 -66.538 50.763
416088 98.540 -66.540 50.722
416089 98.579 -66.543 50.681
416090 98.618 -66.546 50.639
416091 98.659 -66.549 50.597
416092 98.700 -66.552 50.551
416093 98.745 -66.556 50.503
416094 98.796 -66.564 50.453
416095 98.848 -66.572 50.400
416096 98.899 -66.580 50.346
416097 98.947 -66.588 50.292
416098 98.982 -66.598 50.238
416099 99.016 -66.608 50.182
416100 99.042 -66.618 50.124
416101 99.068 -66.629 50.066
416102 99.104 -66.635 50.010
416103 99.150 -66.637 49.954
416104 99.194 -66.650 49.900
416105 99.237 -66.664 49.845
416106 99.280 -66.677 49.791
416107 99.323 -66.685 49.738
416108 99.366 -66.690 49.685
416109 99.408 -66.696 49.632
416110 99.451 -66.701 49.578
416111 99.494 -66.706 49.522
416112 99.537 -66.712 49.465
416113 99.571 -66.710 49.409
416114 99.580 -66.686 49.355
416115 99.590 -66.676 49.301
416116 99.602 -66.682 49.247
416117 99.655 -66.703 49.194
416118 99.719 -66.728 49.142
416119 99.768 -66.740 49.090
416120 99.814 -66.748 49.040
416121 99.860 -66.757 48.992
416122 99.894 -66.765 48.945
416123 99.914 -66.773 48.898
416124 99.934 -66.781 48.851
416125 99.971 -66.780 48.805
416126 100.010 -66.778 48.759
416127 100.050 -66.777 48.717
416128 100.090 -66.774 48.674
416129 100.134 -66.772 48.632
416130 100.182 -66.770 48.589
416131 100.230 -66.768 48.547
416132 100.278 -66.766 48.506
416133 100.327 -66.764 48.466
416134 100.375 -66.762 48.426
416564 89.167 -65.750 55.027
416565 89.213 -65.754 55.016
416592 90.570 -65.851 54.966
416593 90.618 -65.857 54.969
416594 90.667 -65.863 54.972
416595 90.715 -65.869 54.975
416596 90.763 -65.876 54.977
416597 90.812 -65.882 54.978
416598 90.861 -65.888 54.978
416599 90.909 -65.894 54.979
416600 90.957 -65.900 54.979
416601 91.006 -65.906 54.979
416602 91.054 -65.912 54.978
416603 91.103 -65.919 54.978
416604 91.149 -65.930 54.979
416605 91.194 -65.942 54.980
416606 91.240 -65.953 54.983
416607 91.286 -65.965 54.987
416608 91.331 -65.976 54.991
416609 91.377 -65.988 54.998
416610 91.423 -65.999 55.010
416611 91.468 -66.010 55.022
416612 91.511 -66.021 55.032
416613 91.554 -66.026 55.039
416614 91.598 -66.029 55.044
416615 91.642 -66.033 55.047
416616 91.685 -66.036 55.050
416617 91.729 -66.040 55.054
416618 91.773 -66.044 55.054
416619 91.816 -66.050 55.055
416620 91.859 -66.056 55.055
416621 91.904 -66.061 55.053
416622 91.954 -66.061 55.052
416623 92.002 -66.071 55.052
416624 92.051 -66.081 55.055
416625 92.099 -66.091 55.058
416626 92.148 -66.100 55.061
416627 92.196 -66.110 55.060
416628 92.244 -66.119 55.058
416629 92.293 -66.129 55.055
416630 92.341 -66.139 55.051
416631 92.389 -66.149 55.045
416632 92.438 -66.159 55.040
416633 92.483 -66.156 55.033
416634 92.527 -66.147 55.023
416635 92.571 -66.138 55.009
416636 92.615 -66.129 54.991
416637 92.658 -66.121 54.964
416638 92.702 -66.112 54.936
416639 92.746 -66.103 54.909
416640 92.789 -66.094 54.885
416641 92.833 -66.086 54.863
416642 92.877 -66.077 54.843
416643 92.921 -66.069 54.826
416644 92.965 -66.060 54.813
416645 93.008 -66.051 54.801
416646 93.052 -66.042 54.786
416647 93.100 -66.042 54.766
416648 93.149 -66.044 54.742
416649 93.198 -66.045 54.715
416650 93.246 -66.042 54.690
416651 93.294 -66.035 54.669
416652 93.341 -66.028 54.650
416653 93.385 -66.025 54.634
416654 93.423 -66.024 54.620
416655 93.460 -66.023 54.607
416656 93.501 -66.022 54.591
416657 93.546 -66.022 54.575
416658 93.590 -66.022 54.559
416659 93.635 -66.022 54.542
416660 93.679 -66.022 54.529
416661 93.724 -66.022 54.516
416662 93.768 -66.026 54.502
416663 93.811 -66.033 54.488
416664 93.854 -66.040 54.471
416665 93.898 -66.047 54.450
416666 93.941 -66.053 54.425
416667 93.983 -66.060 54.397
416668 94.023 -66.065 54.368
416669 94.062 -66.070 54.338
416670 94.102 -66.075 54.310
416671 94.141 -66.081 54.282
416672 94.181 -66.085 54.254
416673 94.220 -66.091 54.225
416674 94.260 -66.096 54.197
416675 94.299 -66.099 54.167
416676 94.339 -66.103 54.136
416677 94.379 -66.107 54.104
416678 94.401 -66.120 54.073
416679 94.426 -66.133 54.042
416680 94.461 -66.141 54.013
416681 94.496 -66.148 53.983
416682 94.532 -66.156 53.955
416683 94.567 -66.163 53.933
416684 94.602 -66.171 53.915
416685 94.637 -66.179 53.902
416686 94.672 -66.182 53.892
416687 94.706 -66.185 53.886
416688 94.741 -66.187 53.879
416689 94.775 -66.189 53.869
416690 94.809 -66.192 53.855
416691 94.849 -66.207 53.838
416692 94.887 -66.220 53.818
416693 94.915 -66.225 53.799
416694 94.944 -66.229 53.780
416695 94.975 -66.234 53.759
416696 95.013 -66.236 53.736
416697 95.052 -66.238 53.709
416698 95.090 -66.241 53.678
416699 95.129 -66.244 53.645
416700 95.167 -66.246 53.611
416701 95.206 -66.249 53.578
416702 95.244 -66.251 53.547
416703 95.283 -66.254 53.521
416704 95.325 -66.256 53.497
416705 95.363 -66.262 53.474
416706 95.401 -66.267 53.453
416707 95.440 -66.273 53.440
416708 95.478 -66.278 53.429
416709 95.517 -66.283 53.418
416710 95.551 -66.295 53.405
416711 95.594 -66.297 53.390
416712 95.640 -66.295 53.372
416713 95.686 -66.294 53.352
416714 95.732 -66.292 53.331
416715 95.779 -66.288 53.310
416716 95.826 -66.285 53.288
416717 95.874 -66.282 53.266
416718 95.916 -66.276 53.241
416719 95.958 -66.270 53.216
416720 96.001 -66.266 53.190
416721 96.044 -66.261 53.166
416722 96.088 -66.257 53.141
416723 96.131 -66.252 53.117
416724 96.174 -66.248 53.094
416725 96.218 -66.243 53.071
416726 96.261 -66.238 53.051
416727 96.304 -66.235 53.030
416728 96.346 -66.232 53.008
416729 96.387 -66.229 52.984
416730 96.425 -66.226 52.955
416731 96.462 -66.223 52.923
416732 96.496 -66.220 52.889
416733 96.530 -66.217 52.856
416734 96.564 -66.215 52.827
416735 96.597 -66.212 52.803
416736 96.631 -66.209 52.778
416737 96.665 -66.207 52.752
416738 96.699 -66.204 52.726
416739 96.737 -66.207 52.696
416740 96.777 -66.214 52.661
416741 96.818 -66.219 52.622
416742 96.862 -66.221 52.579
416743 96.906 -66.223 52.533
416744 96.950 -66.224 52.488
416745 96.993 -66.230 52.448
416746 97.033 -66.240 52.412
416747 97.074 -66.251 52.380
416748 97.114 -66.260 52.350
416749 97.155 -66.270 52.323
416750 97.195 -66.281 52.300
416751 97.233 -66.288 52.277
416752 97.271 -66.294 52.253
416753 97.307 -66.301 52.226
416754 97.345 -66.307 52.194
416755 97.383 -66.312 52.160
416756 97.422 -66.317 52.123
416757 97.460 -66.323 52.085
416758 97.499 -66.328 52.046
416759 97.538 -66.341 52.010
416760 97.577 -66.354 51.976
416761 97.616 -66.366 51.942
416762 97.654 -66.374 51.908
416763 97.692 -66.382 51.875
416764 97.731 -66.390 51.837
416765 97.769 -66.398 51.797
416766 97.808 -66.407 51.757
416767 97.846 -66.414 51.714
416768 97.878 -66.419 51.670
416769 97.912 -66.423 51.626
416770 97.945 -66.427 51.582
416771 97.978 -66.432 51.537
416772 98.011 -66.436 51.495
416773 98.044 -66.440 51.453
416774 98.077 -66.444 51.411
416775 98.110 -66.448 51.367
416776 98.134 -66.457 51.322
416777 98.156 -66.465 51.276
416778 98.179 -66.473 51.226
416779 98.202 -66.482 51.176
416780 98.224 -66.490 51.125
416781 98.253 -66.494 51.074
416782 98.288 -66.494 51.025
416783 98.323 -66.493 50.977
416784 98.358 -66.493 50.931
416785 98.393 -66.493 50.887
416786 98.428 -66.491 50.844
416787 98.467 -66.493 50.801
416788 98.506 -66.496 50.759
416789 98.546 -66.499 50.716
416790 98.587 -66.502 50.675
416791 98.628 -66.505 50.635
416792 98.669 -66.508 50.593
416793 98.710 -66.510 50.549
416794 98.752 -66.513 50.501
416795 98.803 -66.520 50.452
416796 98.855 -66.529 50.399
416797 98.906 -66.536 50.345
416798 98.946 -66.545 50.290
416799 98.981 -66.555 50.234
416800 99.015 -66.565 50.178
416801 99.050 -66.575 50.121
416802 99.084 -66.585 50.065
416803 99.114 -66.596 50.009
416804 99.154 -66.602 49.953
416805 99.197 -66.614 49.898
416806 99.240 -66.620 49.844
416807 99.283 -66.625 49.789
416808 99.326 -66.630 49.734
416809 99.368 -66.636 49.681
416810 99.411 -66.641 49.627
416811 99.453 -66.646 49.572
416812 99.496 -66.652 49.516
416813 99.539 -66.657 49.461
416814 99.573 -66.655 49.405
416815 99.582 -66.631 49.350
416816 99.594 -66.636 49.296
416817 99.606 -66.642 49.243
416818 99.667 -66.666 49.190
416819 99.725 -66.686 49.138
416820 99.772 -66.695 49.086
416821 99.818 -66.704 49.036
416822 99.864 -66.713 48.987
416823 99.900 -66.721 48.940
416824 99.920 -66.729 48.893
416825 99.940 -66.737 48.847
416826 99.980 -66.734 48.800
416827 100.020 -66.733 48.755
416828 100.059 -66.730 48.713
416829 100.100 -66.728 48.670
416830 100.148 -66.726 48.628
416831 100.196 -66.724 48.586
416832 100.245 -66.722 48.544
416833 100.293 -66.720 48.502
416834 100.341 -66.718 48.463
416835 100.389 -66.716 48.423
417262 89.041 -65.680 55.063
417263 89.077 -65.689 55.054
417264 89.123 -65.693 55.043
417265 89.169 -65.697 55.033
417266 89.215 -65.701 55.022
417267 89.262 -65.705 55.011
417268 89.308 -65.709 55.001
417269 89.354 -65.712 54.990
417270 89.400 -65.716 54.979
417291 90.472 -65.790 54.962
417292 90.520 -65.796 54.965
417293 90.569 -65.802 54.968
417294 90.617 -65.809 54.971
417295 90.665 -65.815 54.973
417296 90.714 -65.821 54.976
417297 90.762 -65.827 54.977
417298 90.811 -65.833 54.977
417299 90.859 -65.839 54.977
417300 90.908 -65.845 54.976
417301 90.956 -65.852 54.976
417302 91.005 -65.858 54.975
417303 91.053 -65.864 54.973
417304 91.101 -65.872 54.973
417305 91.147 -65.883 54.973
417306 91.192 -65.894 54.973
417307 91.238 -65.906 54.975
417308 91.284 -65.918 54.978
417309 91.329 -65.929 54.981
417310 91.375 -65.941 54.986
417311 91.421 -65.952 54.994
417312 91.464 -65.963 55.004
417313 91.507 -65.974 55.013
417314 91.550 -65.984 55.022
417315 91.593 -65.988 55.031
417316 91.637 -65.991 55.039
417317 91.681 -65.994 55.045
417318 91.724 -65.997 55.048
417319 91.768 -66.001 55.049
417320 91.812 -66.004 55.050
417321 91.856 -66.008 55.050
417322 91.899 -66.011 55.049
417323 91.944 -66.019 55.048
417324 91.992 -66.029 55.047
417325 92.040 -66.039 55.047
417326 92.089 -66.049 55.049
417327 92.138 -66.059 55.050
417328 92.186 -66.068 55.049
417329 92.234 -66.078 55.047
417330 92.283 -66.088 55.044
417331 92.331 -66.097 55.041
417332 92.379 -66.107 55.037
417333 92.428 -66.117 55.033
417334 92.472 -66.109 55.026
417335 92.515 -66.100 55.015
417336 92.559 -66.091 55.001
417337 92.603 -66.083 54.985
417338 92.647 -66.074 54.962
417339 92.691 -66.066 54.937
417340 92.734 -66.056 54.912
417341 92.778 -66.048 54.889
417342 92.822 -66.039 54.867
417343 92.866 -66.031 54.847
417344 92.910 -66.022 54.830
417345 92.953 -66.013 54.814
417346 93.002 -66.015 54.802
417347 93.051 -66.016 54.787
417348 93.100 -66.018 54.768
417349 93.149 -66.019 54.745
417350 93.197 -66.017 54.723
417351 93.245 -66.010 54.701
417352 93.292 -66.004 54.685
417353 93.340 -65.997 54.671
417354 93.388 -65.991 54.659
417355 93.426 -65.990 54.648
417356 93.463 -65.989 54.635
417357 93.506 -65.989 54.619
417358 93.551 -65.988 54.602
417359 93.595 -65.988 54.584
417360 93.640 -65.988 54.566
417361 93.684 -65.988 54.552
417362 93.729 -65.988 54.539
417363 93.773 -65.988 54.525
417364 93.817 -65.994 54.510
417365 93.860 -66.001 54.493
417366 93.904 -66.008 54.471
417367 93.947 -66.015 54.446
417368 93.987 -66.021 54.419
417369 94.027 -66.025 54.391
417370 94.066 -66.031 54.362
417371 94.106 -66.036 54.333
417372 94.145 -66.041 54.303
417373 94.185 -66.047 54.273
417374 94.224 -66.050 54.244
417375 94.264 -66.053 54.215
417376 94.303 -66.057 54.185
417377 94.343 -66.061 54.153
417378 94.373 -66.070 54.121
417379 94.392 -66.087 54.086
417380 94.420 -66.098 54.054
417381 94.455 -66.106 54.023
417382 94.490 -66.113 53.994
417383 94.525 -66.120 53.967
417384 94.560 -66.128 53.948
417385 94.596 -66.135 53.931
417386 94.631 -66.143 53.918
417387 94.666 -66.151 53.910
417388 94.701 -66.158 53.904
417389 94.736 -66.162 53.897
417390 94.770 -66.165 53.887
417391 94.805 -66.167 53.873
417392 94.840 -66.171 53.856
417393 94.880 -66.186 53.837
417394 94.911 -66.193 53.819
417395 94.939 -66.198 53.800
417396 94.973 -66.201 53.778
417397 95.012 -66.204 53.755
417398 95.050 -66.207 53.729
417399 95.089 -66.209 53.698
417400 95.127 -66.211 53.665
417401 95.166 -66.214 53.631
417402 95.204 -66.216 53.597
417403 95.245 -66.218 53.565
417404 95.284 -66.221 53.538
417405 95.321 -66.227 53.512
417406 95.358 -66.234 53.488
417407 95.395 -66.240 53.467
417408 95.432 -66.247 53.452
417409 95.470 -66.253 53.442
417410 95.508 -66.258 53.432
417411 95.546 -66.264 53.421
417412 95.593 -66.262 53.406
417413 95.640 -66.258 53.388
417414 95.687 -66.255 53.368
417415 95.735 -66.251 53.346
417416 95.782 -66.248 53.323
417417 95.829 -66.245 53.300
417418 95.876 -66.241 53.276
417419 95.920 -66.237 53.251
417420 95.962 -66.231 53.226
417421 96.004 -66.225 53.200
417422 96.046 -66.219 53.176
417423 96.088 -66.213 53.153
417424 96.131 -66.209 53.129
417425 96.175 -66.204 53.106
417426 96.218 -66.200 53.083
417427 96.261 -66.195 53.062
417428 96.305 -66.191 53.041
417429 96.347 -66.188 53.019
417430 96.386 -66.185 52.995
417431 96.424 -66.182 52.968
417432 96.462 -66.179 52.937
417433 96.498 -66.176 52.904
417434 96.531 -66.173 52.870
417435 96.565 -66.171 52.836
417436 96.599 -66.168 52.811
417437 96.633 -66.165 52.786
417438 96.667 -66.163 52.760
417439 96.704 -66.165 52.732
417440 96.744 -66.172 52.702
417441 96.784 -66.178 52.666
417442 96.823 -66.184 52.627
417443 96.867 -66.186 52.584
417444 96.910 -66.188 52.540
417445 96.955 -66.190 52.496
417446 96.997 -66.197 52.454
417447 97.037 -66.207 52.418
417448 97.078 -66.217 52.386
417449 97.118 -66.227 52.356
417450 97.159 -66.237 52.325
417451 97.197 -66.245 52.297
417452 97.234 -66.251 52.272
417453 97.272 -66.257 52.248
417454 97.309 -66.264 52.222
417455 97.346 -66.270 52.190
417456 97.383 -66.277 52.157
417457 97.421 -66.282 52.122
417458 97.460 -66.287 52.085
417459 97.499 -66.294 52.048
417460 97.537 -66.307 52.015
417461 97.577 -66.319 51.982
417462 97.615 -66.327 51.950
417463 97.653 -66.335 51.917
417464 97.691 -66.344 51.884
417465 97.730 -66.351 51.847
417466 97.768 -66.360 51.806
417467 97.806 -66.368 51.764
417468 97.843 -66.375 51.720
417469 97.876 -66.379 51.675
417470 97.909 -66.383 51.631
417471 97.942 -66.387 51.587
417472 97.975 -66.392 51.543
417473 98.008 -66.395 51.501
417474 98.041 -66.400 51.459
417475 98.074 -66.404 51.415
417476 98.107 -66.408 51.370
417477 98.136 -66.414 51.323
417478 98.159 -66.423 51.274
417479 98.181 -66.431 51.224
417480 98.204 -66.439 51.172
417481 98.227 -66.448 51.120
417482 98.258 -66.450 51.069
417483 98.293 -66.450 51.022
417484 98.328 -66.450 50.976
417485 98.363 -66.449 50.930
417486 98.397 -66.449 50.887
417487 98.433 -66.449 50.843
417488 98.474 -66.452 50.800
417489 98.515 -66.454 50.756
417490 98.556 -66.458 50.713
417491 98.597 -66.460 50.671
417492 98.638 -66.463 50.632
417493 98.679 -66.466 50.591
417494 98.720 -66.468 50.546
417495 98.762 -66.471 50.499
417496 98.810 -66.477 50.450
417497 98.861 -66.485 50.398
417498 98.911 -66.493 50.343
417499 98.945 -66.503 50.287
417500 98.980 -66.513 50.230
417501 99.014 -66.523 50.173
417502 99.049 -66.533 50.118
417503 99.083 -66.542 50.063
417504 99.119 -66.555 50.007
417505 99.158 -66.564 49.951
417506 99.200 -66.565 49.896
417507 99.242 -66.565 49.841
417508 99.285 -66.570 49.785
417509 99.328 -66.576 49.729
417510 99.370 -66.581 49.676
417511 99.413 -66.586 49.621
417512 99.456 -66.592 49.565
417513 99.499 -66.597 49.510
417514 99.542 -66.602 49.456
417515 99.576 -66.601 49.400
417516 99.586 -66.590 49.346
417517 99.598 -66.596 49.292
417518 99.615 -66.604 49.239
417519 99.679 -66.629 49.186
417520 99.729 -66.642 49.135
417521 99.775 -66.651 49.083
417522 99.822 -66.660 49.032
417523 99.868 -66.668 48.982
417524 99.905 -66.677 48.935
417525 99.925 -66.685 48.888
417526 99.950 -66.690 48.842
417527 99.989 -66.689 48.796
417528 100.029 -66.686 48.751
417529 100.068 -66.684 48.709
417530 100.114 -66.683 48.667
417531 100.162 -66.680 48.624
417532 100.211 -66.678 48.583
417533 100.259 -66.676 48.541
417534 100.307 -66.674 48.498
417535 100.355 -66.672 48.458
417960 88.930 -65.603 55.100
417961 88.965 -65.613 55.090
417962 88.999 -65.622 55.080
417963 89.033 -65.632 55.070
417964 89.079 -65.636 55.061
417965 89.125 -65.640 55.051
417966 89.172 -65.643 55.041
417967 89.217 -65.648 55.030
417968 89.264 -65.652 55.018
417969 89.310 -65.655 55.009
417970 89.356 -65.659 54.997
417971 89.402 -65.663 54.986
417972 89.449 -65.667 54.975
417973 89.495 -65.671 54.964
417974 89.541 -65.675 54.954
417975 89.587 -65.679 54.945
417991 90.422 -65.734 54.962
417992 90.470 -65.740 54.964
417993 90.519 -65.747 54.967
417994 90.568 -65.753 54.970
417995 90.616 -65.759 54.973
417996 90.664 -65.765 54.975
417997 90.713 -65.771 54.976
417998 90.761 -65.778 54.977
417999 90.810 -65.784 54.976
418000 90.858 -65.790 54.975
418001 90.906 -65.796 54.973
418002 90.955 -65.803 54.972
418003 91.004 -65.809 54.970
418004 91.052 -65.815 54.969
418005 91.098 -65.825 54.968
418006 91.144 -65.836 54.966
418007 91.190 -65.847 54.965
418008 91.236 -65.859 54.966
418009 91.282 -65.871 54.968
418010 91.327 -65.882 54.970
418011 91.373 -65.894 54.974
418012 91.417 -65.905 54.980
418013 91.460 -65.916 54.987
418014 91.503 -65.926 54.995
418015 91.546 -65.937 55.005
418016 91.589 -65.945 55.017
418017 91.633 -65.948 55.028
418018 91.676 -65.952 55.037
418019 91.720 -65.955 55.041
418020 91.764 -65.959 55.042
418021 91.807 -65.962 55.043
418022 91.851 -65.966 55.044
418023 91.890 -65.974 55.044
418024 91.934 -65.983 55.044
418025 91.982 -65.988 55.043
418026 92.030 -65.997 55.041
418027 92.079 -66.007 55.040
418028 92.128 -66.017 55.039
418029 92.176 -66.027 55.036
418030 92.224 -66.036 55.034
418031 92.273 -66.046 55.032
418032 92.321 -66.056 55.030
418033 92.369 -66.066 55.027
418034 92.417 -66.071 55.021
418035 92.460 -66.062 55.014
418036 92.504 -66.053 55.004
418037 92.548 -66.044 54.992
418038 92.592 -66.036 54.977
418039 92.635 -66.027 54.958
418040 92.679 -66.019 54.938
418041 92.723 -66.010 54.916
418042 92.767 -66.001 54.894
418043 92.810 -65.992 54.873
418044 92.855 -65.986 54.853
418045 92.904 -65.988 54.835
418046 92.953 -65.989 54.817
418047 93.002 -65.991 54.802
418048 93.051 -65.991 54.788
418049 93.100 -65.993 54.769
418050 93.148 -65.993 54.750
418051 93.196 -65.986 54.732
418052 93.243 -65.979 54.714
418053 93.291 -65.972 54.702
418054 93.339 -65.966 54.693
418055 93.386 -65.959 54.686
418056 93.429 -65.956 54.677
418057 93.467 -65.955 54.665
418058 93.511 -65.955 54.648
418059 93.556 -65.955 54.630
418060 93.600 -65.955 54.610
418061 93.645 -65.955 54.591
418062 93.689 -65.954 54.576
418063 93.733 -65.954 54.562
418064 93.778 -65.954 54.548
418065 93.823 -65.956 54.533
418066 93.866 -65.963 54.514
418067 93.909 -65.970 54.492
418068 93.952 -65.976 54.469
418069 93.992 -65.981 54.443
418070 94.031 -65.987 54.415
418071 94.071 -65.992 54.385
418072 94.110 -65.997 54.354
418073 94.150 -66.001 54.324
418074 94.189 -66.005 54.294
418075 94.229 -66.008 54.265
418076 94.268 -66.012 54.235
418077 94.308 -66.016 54.204
418078 94.346 -66.019 54.171
418079 94.364 -66.036 54.138
418080 94.383 -66.053 54.104
418081 94.413 -66.063 54.072
418082 94.449 -66.070 54.040
418083 94.484 -66.078 54.011
418084 94.519 -66.085 53.984
418085 94.554 -66.093 53.965
418086 94.590 -66.100 53.950
418087 94.625 -66.108 53.939
418088 94.660 -66.115 53.930
418089 94.695 -66.122 53.924
418090 94.731 -66.130 53.916
418091 94.766 -66.138 53.905
418092 94.800 -66.142 53.891
418093 94.834 -66.144 53.875
418094 94.871 -66.151 53.857
418095 94.906 -66.162 53.840
418096 94.934 -66.166 53.820
418097 94.972 -66.169 53.798
418098 95.011 -66.172 53.774
418099 95.049 -66.174 53.748
418100 95.088 -66.176 53.717
418101 95.127 -66.179 53.683
418102 95.168 -66.179 53.648
418103 95.209 -66.179 53.614
418104 95.249 -66.180 53.581
418105 95.288 -66.184 53.554
418106 95.326 -66.190 53.527
418107 95.363 -66.197 53.501
418108 95.400 -66.203 53.480
418109 95.437 -66.210 53.464
418110 95.474 -66.216 53.454
418111 95.513 -66.222 53.446
418112 95.555 -66.226 53.436
418113 95.596 -66.222 53.422
418114 95.643 -66.218 53.404
418115 95.690 -66.215 53.383
418116 95.738 -66.211 53.361
418117 95.785 -66.208 53.337
418118 95.832 -66.204 53.312
418119 95.879 -66.201 53.287
418120 95.924 -66.197 53.262
418121 95.967 -66.191 53.236
418122 96.009 -66.185 53.210
418123 96.050 -66.180 53.186
418124 96.093 -66.174 53.163
418125 96.134 -66.168 53.140
418126 96.177 -66.163 53.117
418127 96.219 -66.157 53.094
418128 96.262 -66.152 53.073
418129 96.305 -66.147 53.051
418130 96.348 -66.144 53.029
418131 96.386 -66.141 53.005
418132 96.424 -66.138 52.979
418133 96.462 -66.135 52.949
418134 96.499 -66.132 52.916
418135 96.533 -66.129 52.882
418136 96.567 -66.127 52.847
418137 96.601 -66.124 52.818
418138 96.635 -66.121 52.791
418139 96.671 -66.123 52.764
418140 96.711 -66.129 52.735
418141 96.750 -66.135 52.703
418142 96.790 -66.141 52.667
418143 96.830 -66.147 52.629
418144 96.871 -66.152 52.587
418145 96.915 -66.154 52.543
418146 96.959 -66.156 52.500
418147 97.001 -66.163 52.460
418148 97.042 -66.173 52.424
418149 97.082 -66.184 52.392
418150 97.123 -66.194 52.362
418151 97.161 -66.201 52.330
418152 97.198 -66.207 52.299
418153 97.236 -66.214 52.271
418154 97.273 -66.220 52.245
418155 97.310 -66.227 52.217
418156 97.347 -66.234 52.186
418157 97.384 -66.241 52.154
418158 97.421 -66.247 52.121
418159 97.459 -66.252 52.088
418160 97.498 -66.260 52.054
418161 97.537 -66.272 52.020
418162 97.575 -66.280 51.986
418163 97.613 -66.288 51.953
418164 97.652 -66.297 51.922
418165 97.690 -66.304 51.890
418166 97.729 -66.313 51.854
418167 97.767 -66.321 51.814
418168 97.805 -66.329 51.772
418169 97.840 -66.335 51.729
418170 97.873 -66.339 51.682
418171 97.906 -66.343 51.637
418172 97.939 -66.347 51.593
418173 97.972 -66.351 51.549
418174 98.005 -66.356 51.507
418175 98.038 -66.360 51.462
418176 98.071 -66.364 51.417
418177 98.104 -66.368 51.371
418178 98.137 -66.373 51.323
418179 98.162 -66.380 51.272
418180 98.184 -66.389 51.220
418181 98.207 -66.397 51.168
418182 98.230 -66.405 51.117
418183 98.263 -66.406 51.067
418184 98.298 -66.406 51.023
418185 98.333 -66.406 50.977
418186 98.368 -66.406 50.931
418187 98.403 -66.406 50.887
418188 98.439 -66.409 50.843
418189 98.482 -66.411 50.798
418190 98.524 -66.414 50.754
418191 98.566 -66.416 50.710
418192 98.607 -66.418 50.667
418193 98.648 -66.421 50.628
418194 98.690 -66.424 50.587
418195 98.730 -66.426 50.543
418196 98.772 -66.429 50.497
418197 98.817 -66.434 50.447
418198 98.868 -66.442 50.395
418199 98.910 -66.451 50.341
418200 98.944 -66.461 50.285
418201 98.979 -66.470 50.228
418202 99.013 -66.480 50.170
418203 99.050 -66.493 50.113
418204 99.087 -66.507 50.058
418205 99.125 -66.521 50.003
418206 99.163 -66.535 49.948
418207 99.205 -66.535 49.892
418208 99.247 -66.535 49.835
418209 99.289 -66.536 49.779
418210 99.332 -66.536 49.724
418211 99.374 -66.536 49.670
418212 99.416 -66.536 49.615
418213 99.458 -66.537 49.559
418214 99.501 -66.542 49.504
418215 99.544 -66.548 49.449
418216 99.578 -66.546 49.395
418217 99.590 -66.550 49.341
418218 99.602 -66.556 49.288
418219 99.626 -66.567 49.235
418220 99.687 -66.589 49.182
418221 99.733 -66.598 49.130
418222 99.779 -66.606 49.079
418223 99.826 -66.615 49.028
418224 99.871 -66.624 48.979
418225 99.911 -66.633 48.931
418226 99.931 -66.641 48.884
418227 99.959 -66.645 48.838
418228 99.998 -66.643 48.793
418229 100.038 -66.640 48.749
418230 100.080 -66.639 48.706
418231 100.129 -66.636 48.664
418232 100.177 -66.634 48.622
418233 100.225 -66.632 48.580
418234 100.273 -66.630 48.538
418235 100.322 -66.628 48.497
418236 100.370 -66.626 48.457
418658 88.820 -65.526 55.144
418659 88.854 -65.536 55.131
418660 88.888 -65.545 55.120
418661 88.922 -65.555 55.108
418662 88.957 -65.564 55.098
418663 88.991 -65.574 55.088
418664 89.035 -65.579 55.079
418665 89.082 -65.583 55.070
418666 89.128 -65.586 55.060
418667 89.174 -65.590 55.050
418668 89.220 -65.594 55.039
418669 89.266 -65.598 55.027
418670 89.312 -65.602 55.017
418671 89.358 -65.606 55.006
418672 89.405 -65.610 54.995
418673 89.451 -65.614 54.984
418674 89.497 -65.618 54.973
418675 89.543 -65.622 54.962
418676 89.590 -65.626 54.952
418677 89.641 -65.618 54.945
418692 90.421 -65.686 54.965
418693 90.469 -65.692 54.967
418694 90.518 -65.698 54.969
418695 90.566 -65.704 54.972
418696 90.615 -65.710 54.975
418697 90.663 -65.716 54.976
418698 90.712 -65.723 54.976
418699 90.760 -65.729 54.977
418700 90.808 -65.735 54.976
418701 90.857 -65.741 54.974
418702 90.905 -65.747 54.972
418703 90.954 -65.753 54.970
418704 91.002 -65.759 54.967
418705 91.050 -65.766 54.964
418706 91.096 -65.778 54.962
418707 91.142 -65.789 54.960
418708 91.188 -65.800 54.959
418709 91.234 -65.812 54.958
418710 91.280 -65.823 54.959
418711 91.325 -65.835 54.960
418712 91.370 -65.847 54.963
418713 91.413 -65.857 54.967
418714 91.456 -65.868 54.973
418715 91.499 -65.878 54.978
418716 91.542 -65.889 54.987
418717 91.585 -65.900 55.000
418718 91.628 -65.906 55.014
418719 91.672 -65.909 55.026
418720 91.716 -65.913 55.032
418721 91.759 -65.916 55.031
418722 91.803 -65.920 55.033
418723 91.847 -65.923 55.035
418724 91.881 -65.937 55.038
418725 91.924 -65.947 55.040
418726 91.972 -65.953 55.039
418727 92.020 -65.958 55.036
418728 92.069 -65.966 55.034
418729 92.118 -65.975 55.033
418730 92.166 -65.985 55.030
418731 92.214 -65.995 55.027
418732 92.263 -66.004 55.025
418733 92.311 -66.014 55.022
418734 92.359 -66.024 55.017
418735 92.405 -66.024 55.009
418736 92.449 -66.016 55.002
418737 92.493 -66.006 54.992
418738 92.536 -65.998 54.981
418739 92.580 -65.989 54.969
418740 92.624 -65.981 54.952
418741 92.668 -65.972 54.933
418742 92.712 -65.963 54.914
418743 92.758 -65.959 54.895
418744 92.806 -65.960 54.875
418745 92.855 -65.962 54.856
418746 92.904 -65.963 54.837
418747 92.953 -65.964 54.819
418748 93.002 -65.966 54.803
418749 93.051 -65.967 54.788
418750 93.099 -65.968 54.772
418751 93.147 -65.961 54.757
418752 93.195 -65.954 54.744
418753 93.242 -65.948 54.730
418754 93.290 -65.941 54.722
418755 93.338 -65.934 54.717
418756 93.385 -65.928 54.713
418757 93.431 -65.922 54.707
418758 93.472 -65.922 54.696
418759 93.516 -65.922 54.679
418760 93.561 -65.922 54.659
418761 93.605 -65.921 54.638
418762 93.649 -65.921 54.617
418763 93.694 -65.921 54.600
418764 93.739 -65.921 54.586
418765 93.783 -65.921 54.572
418766 93.828 -65.920 54.556
418767 93.871 -65.925 54.536
418768 93.915 -65.931 54.514
418769 93.957 -65.937 54.492
418770 93.996 -65.943 54.467
418771 94.036 -65.948 54.438
418772 94.075 -65.952 54.407
418773 94.115 -65.956 54.376
418774 94.154 -65.959 54.345
418775 94.194 -65.963 54.315
418776 94.233 -65.966 54.287
418777 94.273 -65.970 54.255
418778 94.312 -65.973 54.222
418779 94.337 -65.986 54.189
418780 94.355 -66.002 54.156
418781 94.374 -66.019 54.124
418782 94.407 -66.028 54.091
418783 94.442 -66.035 54.060
418784 94.477 -66.043 54.030
418785 94.512 -66.050 54.003
418786 94.548 -66.057 53.985
418787 94.583 -66.065 53.972
418788 94.618 -66.072 53.962
418789 94.653 -66.080 53.954
418790 94.689 -66.088 53.948
418791 94.724 -66.095 53.940
418792 94.759 -66.103 53.929
418793 94.794 -66.110 53.915
418794 94.830 -66.117 53.899
418795 94.864 -66.122 53.882
418796 94.901 -66.130 53.863
418797 94.932 -66.135 53.841
418798 94.971 -66.137 53.817
418799 95.010 -66.139 53.792
418800 95.051 -66.140 53.762
418801 95.092 -66.141 53.731
418802 95.132 -66.141 53.698
418803 95.173 -66.141 53.663
418804 95.213 -66.142 53.628
418805 95.254 -66.143 53.594
418806 95.293 -66.147 53.566
418807 95.330 -66.153 53.539
418808 95.368 -66.160 53.513
418809 95.405 -66.166 53.491
418810 95.442 -66.172 53.476
418811 95.482 -66.177 53.466
418812 95.524 -66.182 53.459
418813 95.565 -66.186 53.451
418814 95.607 -66.184 53.439
418815 95.648 -66.179 53.421
418816 95.693 -66.175 53.399
418817 95.740 -66.171 53.376
418818 95.788 -66.168 53.351
418819 95.835 -66.164 53.325
418820 95.882 -66.161 53.299
418821 95.929 -66.157 53.273
418822 95.971 -66.152 53.247
418823 96.013 -66.146 53.220
418824 96.055 -66.140 53.196
418825 96.097 -66.135 53.173
418826 96.139 -66.129 53.150
418827 96.181 -66.122 53.127
418828 96.223 -66.117 53.104
418829 96.265 -66.111 53.082
418830 96.307 -66.106 53.059
418831 96.347 -66.100 53.037
418832 96.385 -66.097 53.014
418833 96.424 -66.094 52.988
418834 96.461 -66.091 52.959
418835 96.500 -66.088 52.927
418836 96.535 -66.085 52.892
418837 96.569 -66.083 52.856
418838 96.602 -66.080 52.825
418839 96.638 -66.081 52.796
418840 96.678 -66.087 52.767
418841 96.718 -66.093 52.737
418842 96.757 -66.099 52.704
418843 96.797 -66.105 52.668
418844 96.837 -66.111 52.631
418845 96.876 -66.117 52.590
418846 96.920 -66.120 52.548
418847 96.964 -66.122 52.507
418848 97.006 -66.130 52.467
418849 97.046 -66.140 52.431
418850 97.087 -66.150 52.399
418851 97.125 -66.158 52.368
418852 97.162 -66.164 52.336
418853 97.199 -66.171 52.304
418854 97.237 -66.177 52.273
418855 97.274 -66.184 52.243
418856 97.311 -66.191 52.215
418857 97.348 -66.197 52.184
418858 97.385 -66.204 52.152
418859 97.422 -66.210 52.122
418860 97.459 -66.217 52.092
418861 97.498 -66.225 52.060
418862 97.536 -66.233 52.025
418863 97.574 -66.241 51.990
418864 97.612 -66.250 51.956
418865 97.651 -66.258 51.925
418866 97.689 -66.266 51.893
418867 97.727 -66.274 51.858
418868 97.766 -66.282 51.820
418869 97.803 -66.290 51.779
418870 97.837 -66.295 51.736
418871 97.870 -66.298 51.690
418872 97.903 -66.303 51.645
418873 97.936 -66.307 51.600
418874 97.969 -66.311 51.555
418875 98.002 -66.316 51.510
418876 98.035 -66.320 51.463
418877 98.068 -66.324 51.416
418878 98.101 -66.328 51.368
418879 98.134 -66.332 51.319
418880 98.165 -66.338 51.267
418881 98.187 -66.346 51.215
418882 98.210 -66.354 51.164
418883 98.233 -66.363 51.115
418884 98.268 -66.363 51.067
418885 98.303 -66.362 51.023
418886 98.338 -66.362 50.977
418887 98.372 -66.362 50.932
418888 98.409 -66.366 50.887
418889 98.445 -66.369 50.842
418890 98.487 -66.371 50.796
418891 98.529 -66.373 50.750
418892 98.572 -66.376 50.704
418893 98.615 -66.378 50.659
418894 98.658 -66.379 50.619
418895 98.700 -66.382 50.578
418896 98.741 -66.385 50.536
418897 98.782 -66.388 50.490
418898 98.823 -66.391 50.442
418899 98.875 -66.398 50.391
418900 98.909 -66.408 50.338
418901 98.943 -66.418 50.283
418902 98.980 -66.431 50.225
418903 99.018 -66.445 50.166
418904 99.056 -66.460 50.108
418905 99.093 -66.474 50.052
418906 99.131 -66.489 49.996
418907 99.168 -66.503 49.941
418908 99.210 -66.505 49.884
418909 99.252 -66.505 49.828
418910 99.294 -66.505 49.772
418911 99.337 -66.505 49.718
418912 99.379 -66.506 49.663
418913 99.421 -66.506 49.609
418914 99.464 -66.506 49.554
418915 99.505 -66.507 49.499
418916 99.548 -66.507 49.444
418917 99.581 -66.505 49.391
418918 99.593 -66.511 49.337
418919 99.606 -66.516 49.284
418920 99.638 -66.530 49.231
418921 99.691 -66.545 49.179
418922 99.737 -66.553 49.126
418923 99.783 -66.562 49.075
418924 99.829 -66.571 49.025
418925 99.875 -66.580 48.976
418926 99.916 -66.589 48.928
418927 99.936 -66.596 48.882
418928 99.968 -66.599 48.836
418929 100.008 -66.596 48.791
418930 100.047 -66.595 48.747
418931 100.095 -66.592 48.704
418932 100.143 -66.590 48.662
418933 100.191 -66.588 48.620
418934 100.240 -66.586 48.578
418935 100.288 -66.584 48.536
418936 100.336 -66.582 48.495
418937 100.384 -66.580 48.455
419356 88.709 -65.449 55.192
419357 88.744 -65.458 55.177
419358 88.778 -65.468 55.164
419359 88.812 -65.477 55.151
419360 88.846 -65.487 55.139
419361 88.881 -65.497 55.128
419362 88.915 -65.506 55.117
419363 88.949 -65.516 55.107
419364 88.991 -65.521 55.098
419365 89.038 -65.525 55.090
419366 89.084 -65.530 55.080
419367 89.130 -65.533 55.070
419368 89.176 -65.537 55.060
419369 89.223 -65.541 55.049
419370 89.268 -65.545 55.038
419371 89.315 -65.549 55.027
419372 89.361 -65.553 55.016
419373 89.407 -65.557 55.005
419374 89.453 -65.561 54.993
419375 89.499 -65.564 54.982
419376 89.545 -65.568 54.970
419377 89.593 -65.571 54.960
419378 89.649 -65.553 54.951
419379 89.705 -65.542 54.944
419392 90.371 -65.630 54.963
419393 90.419 -65.637 54.966
419394 90.468 -65.643 54.970
419395 90.517 -65.649 54.972
419396 90.565 -65.655 54.974
419397 90.613 -65.662 54.976
419398 90.662 -65.668 54.977
419399 90.711 -65.674 54.977
419400 90.759 -65.680 54.976
419401 90.807 -65.686 54.975
419402 90.856 -65.692 54.973
419403 90.904 -65.699 54.971
419404 90.953 -65.705 54.967
419405 91.001 -65.711 54.964
419406 91.048 -65.719 54.960
419407 91.094 -65.731 54.957
419408 91.140 -65.742 54.954
419409 91.186 -65.753 54.952
419410 91.232 -65.765 54.950
419411 91.277 -65.776 54.949
419412 91.323 -65.787 54.950
419413 91.366 -65.799 54.952
419414 91.410 -65.809 54.955
419415 91.452 -65.820 54.959
419416 91.495 -65.831 54.964
419417 91.538 -65.842 54.972
419418 91.581 -65.853 54.983
419419 91.624 -65.863 54.996
419420 91.667 -65.867 55.008
419421 91.711 -65.871 55.014
419422 91.755 -65.874 55.016
419423 91.798 -65.878 55.020
419424 91.837 -65.887 55.025
419425 91.871 -65.901 55.030
419426 91.914 -65.911 55.033
419427 91.962 -65.916 55.034
419428 92.010 -65.922 55.033
419429 92.059 -65.928 55.032
419430 92.107 -65.934 55.031
419431 92.156 -65.944 55.028
419432 92.204 -65.953 55.025
419433 92.252 -65.963 55.021
419434 92.301 -65.972 55.016
419435 92.349 -65.982 55.007
419436 92.394 -65.978 54.998
419437 92.438 -65.969 54.989
419438 92.481 -65.960 54.980
419439 92.525 -65.951 54.970
419440 92.569 -65.943 54.960
419441 92.613 -65.934 54.944
419442 92.660 -65.931 54.928
419443 92.709 -65.933 54.911
419444 92.758 -65.934 54.893
419445 92.806 -65.936 54.874
419446 92.855 -65.937 54.856
419447 92.904 -65.938 54.837
419448 92.953 -65.940 54.820
419449 93.002 -65.941 54.804
419450 93.051 -65.943 54.791
419451 93.098 -65.937 54.779
419452 93.146 -65.930 54.768
419453 93.194 -65.923 54.758
419454 93.241 -65.916 54.748
419455 93.289 -65.910 54.741
419456 93.336 -65.903 54.738
419457 93.383 -65.897 54.736
419458 93.431 -65.890 54.732
419459 93.477 -65.888 54.723
419460 93.521 -65.887 54.707
419461 93.565 -65.887 54.687
419462 93.610 -65.887 54.666
419463 93.654 -65.887 54.644
419464 93.699 -65.887 54.624
419465 93.744 -65.887 54.609
419466 93.788 -65.887 54.594
419467 93.833 -65.887 54.579
419468 93.877 -65.887 54.560
419469 93.921 -65.894 54.539
419470 93.961 -65.899 54.515
419471 94.000 -65.903 54.489
419472 94.040 -65.906 54.460
419473 94.079 -65.910 54.429
419474 94.119 -65.914 54.396
419475 94.158 -65.917 54.366
419476 94.198 -65.921 54.337
419477 94.238 -65.925 54.309
419478 94.277 -65.928 54.275
419479 94.310 -65.935 54.240
419480 94.328 -65.952 54.206
419481 94.347 -65.969 54.175
419482 94.365 -65.984 54.142
419483 94.401 -65.992 54.110
419484 94.436 -66.000 54.080
419485 94.471 -66.007 54.050
419486 94.506 -66.015 54.024
419487 94.541 -66.022 54.007
419488 94.576 -66.030 53.996
419489 94.612 -66.038 53.986
419490 94.647 -66.045 53.980
419491 94.682 -66.052 53.976
419492 94.717 -66.060 53.967
419493 94.753 -66.067 53.956
419494 94.788 -66.075 53.942
419495 94.823 -66.082 53.925
419496 94.858 -66.090 53.908
419497 94.894 -66.097 53.888
419498 94.933 -66.100 53.865
419499 94.974 -66.101 53.838
419500 95.015 -66.101 53.809
419501 95.055 -66.102 53.777
419502 95.096 -66.103 53.743
419503 95.137 -66.103 53.709
419504 95.178 -66.104 53.674
419505 95.218 -66.104 53.639
419506 95.259 -66.105 53.606
419507 95.298 -66.110 53.576
419508 95.335 -66.116 53.548
419509 95.372 -66.122 53.523
419510 95.410 -66.129 53.502
419511 95.452 -66.133 53.489
419512 95.493 -66.138 53.480
419513 95.535 -66.142 53.473
419514 95.577 -66.147 53.465
419515 95.618 -66.146 53.452
419516 95.659 -66.141 53.434
419517 95.699 -66.135 53.413
419518 95.743 -66.131 53.389
419519 95.791 -66.128 53.363
419520 95.838 -66.124 53.337
419521 95.885 -66.121 53.310
419522 95.932 -66.117 53.283
419523 95.975 -66.112 53.256
419524 96.018 -66.107 53.227
419525 96.059 -66.100 53.203
419526 96.102 -66.095 53.180
419527 96.144 -66.089 53.158
419528 96.185 -66.083 53.136
419529 96.228 -66.078 53.112
419530 96.269 -66.072 53.089
419531 96.311 -66.068 53.066
419532 96.350 -66.063 53.043
419533 96.387 -66.057 53.020
419534 96.423 -66.051 52.994
419535 96.461 -66.047 52.966
419536 96.499 -66.044 52.934
419537 96.536 -66.041 52.899
419538 96.570 -66.039 52.864
419539 96.605 -66.038 52.831
419540 96.645 -66.044 52.800
419541 96.685 -66.050 52.770
419542 96.725 -66.056 52.740
419543 96.764 -66.063 52.706
419544 96.804 -66.069 52.671
419545 96.843 -66.075 52.633
419546 96.883 -66.082 52.594
419547 96.925 -66.086 52.553
419548 96.969 -66.088 52.515
419549 97.010 -66.097 52.476
419550 97.050 -66.107 52.439
419551 97.089 -66.115 52.405
419552 97.126 -66.121 52.375
419553 97.164 -66.128 52.343
419554 97.201 -66.134 52.311
419555 97.238 -66.141 52.277
419556 97.276 -66.147 52.243
419557 97.312 -66.154 52.213
419558 97.350 -66.160 52.182
419559 97.386 -66.167 52.151
419560 97.424 -66.173 52.123
419561 97.460 -66.180 52.097
419562 97.496 -66.186 52.066
419563 97.534 -66.194 52.031
419564 97.573 -66.203 51.994
419565 97.611 -66.211 51.959
419566 97.650 -66.219 51.926
419567 97.688 -66.227 51.895
419568 97.726 -66.235 51.861
419569 97.764 -66.244 51.825
419570 97.801 -66.251 51.784
419571 97.833 -66.254 51.742
419572 97.867 -66.259 51.697
419573 97.900 -66.263 51.652
419574 97.933 -66.267 51.605
419575 97.966 -66.271 51.559
419576 97.999 -66.276 51.511
419577 98.032 -66.279 51.463
419578 98.065 -66.284 51.413
419579 98.098 -66.288 51.363
419580 98.132 -66.292 51.314
419581 98.164 -66.297 51.262
419582 98.191 -66.304 51.211
419583 98.213 -66.312 51.162
419584 98.238 -66.319 51.115
419585 98.273 -66.319 51.069
419586 98.307 -66.319 51.024
419587 98.343 -66.319 50.978
419588 98.379 -66.322 50.932
419589 98.415 -66.326 50.886
419590 98.451 -66.329 50.840
419591 98.492 -66.332 50.794
419592 98.534 -66.333 50.746
419593 98.577 -66.335 50.697
419594 98.620 -66.338 50.649
419595 98.663 -66.340 50.607
419596 98.706 -66.342 50.566
419597 98.748 -66.344 50.524
419598 98.791 -66.346 50.480
419599 98.833 -66.348 50.435
419600 98.874 -66.356 50.385
419601 98.911 -66.370 50.334
419602 98.948 -66.384 50.279
419603 98.986 -66.398 50.222
419604 99.024 -66.413 50.163
419605 99.061 -66.427 50.103
419606 99.099 -66.442 50.045
419607 99.137 -66.455 49.988
419608 99.174 -66.470 49.932
419609 99.215 -66.475 49.876
419610 99.257 -66.475 49.821
419611 99.299 -66.476 49.765
419612 99.342 -66.476 49.711
419613 99.383 -66.476 49.657
419614 99.426 -66.476 49.603
419615 99.468 -66.476 49.549
419616 99.510 -66.470 49.494
419617 99.552 -66.466 49.439
419618 99.578 -66.467 49.387
419619 99.597 -66.470 49.333
419620 99.609 -66.476 49.280
419621 99.648 -66.492 49.228
419622 99.694 -66.500 49.176
419623 99.740 -66.509 49.123
419624 99.787 -66.518 49.071
419625 99.833 -66.527 49.022
419626 99.879 -66.536 48.974
419627 99.922 -66.545 48.926
419628 99.942 -66.552 48.880
419629 99.977 -66.552 48.834
419630 100.017 -66.551 48.789
419631 100.061 -66.549 48.745
419632 100.109 -66.546 48.703
419633 100.157 -66.544 48.660
419634 100.206 -66.542 48.618
419635 100.254 -66.540 48.576
419636 100.302 -66.538 48.535
419637 100.350 -66.536 48.493
419638 100.399 -66.534 48.453
420055 88.633 -65.381 55.230
420056 88.667 -65.391 55.213
420057 88.701 -65.400 55.196
420058 88.736 -65.410 55.183
420059 88.770 -65.420 55.170
420060 88.804 -65.429 55.158
420061 88.838 -65.439 55.148
420062 88.873 -65.449 55.138
420063 88.907 -65.458 55.128
420064 88.948 -65.464 55.118
420065 88.994 -65.468 55.109
420066 89.040 -65.472 55.101
420067 89.086 -65.476 55.092
420068 89.132 -65.480 55.082
420069 89.179 -65.484 55.072
420070 89.225 -65.488 55.061
420071 89.271 -65.492 55.049
420072 89.317 -65.496 55.038
420073 89.363 -65.499 55.027
420074 89.409 -65.503 55.015
420075 89.456 -65.508 55.004
420076 89.502 -65.511 54.991
420077 89.548 -65.515 54.979
420078 89.600 -65.506 54.968
420079 89.657 -65.489 54.959
420080 89.712 -65.492 54.951
420093 90.370 -65.581 54.965
420094 90.418 -65.587 54.968
420095 90.467 -65.594 54.971
420096 90.515 -65.600 54.974
420097 90.564 -65.606 54.976
420098 90.612 -65.612 54.978
420099 90.661 -65.618 54.979
420100 90.709 -65.624 54.978
420101 90.757 -65.630 54.977
420102 90.806 -65.637 54.976
420103 90.855 -65.643 54.973
420104 90.903 -65.649 54.969
420105 90.951 -65.655 54.964
420106 91.000 -65.662 54.959
420107 91.046 -65.671 54.955
420108 91.092 -65.684 54.951
420109 91.138 -65.695 54.947
420110 91.183 -65.706 54.944
420111 91.229 -65.718 54.942
420112 91.275 -65.729 54.940
420113 91.319 -65.740 54.940
420114 91.363 -65.751 54.941
420115 91.405 -65.762 54.943
420116 91.448 -65.773 54.947
420117 91.491 -65.784 54.951
420118 91.534 -65.794 54.958
420119 91.577 -65.805 54.967
420120 91.620 -65.815 54.978
420121 91.663 -65.825 54.988
420122 91.707 -65.828 54.994
420123 91.750 -65.832 54.999
420124 91.793 -65.837 55.004
420125 91.827 -65.851 55.009
420126 91.862 -65.865 55.015
420127 91.903 -65.875 55.019
420128 91.952 -65.881 55.022
420129 92.000 -65.886 55.022
420130 92.049 -65.892 55.022
420131 92.097 -65.897 55.023
420132 92.145 -65.903 55.022
420133 92.194 -65.912 55.019
420134 92.242 -65.922 55.014
420135 92.291 -65.931 55.008
420136 92.339 -65.939 54.998
420137 92.382 -65.931 54.987
420138 92.426 -65.922 54.977
420139 92.470 -65.913 54.967
420140 92.514 -65.904 54.958
420141 92.562 -65.904 54.949
420142 92.611 -65.906 54.935
420143 92.659 -65.907 54.921
420144 92.708 -65.909 54.906
420145 92.757 -65.910 54.889
420146 92.806 -65.911 54.871
420147 92.855 -65.912 54.854
420148 92.904 -65.914 54.836
420149 92.953 -65.916 54.820
420150 93.002 -65.917 54.807
420151 93.050 -65.912 54.795
420152 93.097 -65.905 54.787
420153 93.145 -65.898 54.779
420154 93.192 -65.892 54.772
420155 93.240 -65.885 54.764
420156 93.287 -65.878 54.760
420157 93.335 -65.872 54.758
420158 93.383 -65.866 54.756
420159 93.432 -65.862 54.753
420160 93.476 -65.862 54.745
420161 93.520 -65.862 54.731
420162 93.566 -65.861 54.713
420163 93.611 -65.859 54.692
420164 93.656 -65.859 54.670
420165 93.701 -65.857 54.649
420166 93.747 -65.856 54.632
420167 93.791 -65.855 54.617
420168 93.837 -65.854 54.600
420169 93.882 -65.853 54.583
420170 93.925 -65.854 54.561
420171 93.965 -65.858 54.536
420172 94.004 -65.862 54.509
420173 94.044 -65.865 54.480
420174 94.083 -65.868 54.448
420175 94.123 -65.872 54.418
420176 94.163 -65.875 54.387
420177 94.202 -65.879 54.357
420178 94.242 -65.882 54.326
420179 94.281 -65.886 54.293
420180 94.301 -65.902 54.258
420181 94.320 -65.918 54.224
420182 94.338 -65.934 54.191
420183 94.359 -65.950 54.157
420184 94.394 -65.957 54.126
420185 94.429 -65.965 54.096
420186 94.465 -65.972 54.069
420187 94.500 -65.980 54.045
420188 94.535 -65.987 54.031
420189 94.570 -65.994 54.021
420190 94.606 -66.002 54.014
420191 94.641 -66.010 54.009
420192 94.676 -66.017 54.005
420193 94.711 -66.025 53.998
420194 94.747 -66.032 53.986
420195 94.781 -66.040 53.972
420196 94.817 -66.047 53.955
420197 94.856 -66.051 53.935
420198 94.896 -66.056 53.915
420199 94.938 -66.057 53.890
420200 94.979 -66.060 53.860
420201 95.020 -66.062 53.826
420202 95.060 -66.064 53.791
420203 95.101 -66.066 53.755
420204 95.142 -66.066 53.719
420205 95.183 -66.066 53.683
420206 95.224 -66.067 53.648
420207 95.264 -66.068 53.615
420208 95.302 -66.072 53.585
420209 95.339 -66.079 53.557
420210 95.380 -66.084 53.533
420211 95.421 -66.088 53.513
420212 95.463 -66.093 53.502
420213 95.505 -66.097 53.494
420214 95.546 -66.102 53.487
420215 95.588 -66.107 53.478
420216 95.629 -66.108 53.465
420217 95.670 -66.103 53.448
420218 95.710 -66.097 53.426
420219 95.751 -66.092 53.401
420220 95.794 -66.087 53.375
420221 95.841 -66.084 53.348
420222 95.888 -66.081 53.320
420223 95.935 -66.077 53.292
420224 95.980 -66.072 53.264
420225 96.022 -66.067 53.235
420226 96.064 -66.061 53.210
420227 96.106 -66.055 53.187
420228 96.148 -66.050 53.165
420229 96.190 -66.044 53.143
420230 96.232 -66.038 53.119
420231 96.273 -66.035 53.095
420232 96.315 -66.031 53.072
420233 96.354 -66.026 53.048
420234 96.390 -66.020 53.023
420235 96.427 -66.014 52.995
420236 96.463 -66.008 52.967
420237 96.500 -66.003 52.936
420238 96.537 -65.997 52.903
420239 96.573 -65.996 52.870
420240 96.612 -66.002 52.837
420241 96.652 -66.008 52.805
420242 96.691 -66.014 52.775
420243 96.731 -66.020 52.745
420244 96.771 -66.026 52.712
420245 96.811 -66.033 52.677
420246 96.850 -66.039 52.640
420247 96.890 -66.045 52.601
420248 96.930 -66.051 52.561
420249 96.973 -66.053 52.523
420250 97.014 -66.063 52.484
420251 97.053 -66.072 52.447
420252 97.090 -66.078 52.412
420253 97.128 -66.085 52.380
420254 97.165 -66.091 52.349
420255 97.203 -66.097 52.317
420256 97.240 -66.103 52.284
420257 97.277 -66.110 52.248
420258 97.314 -66.116 52.214
420259 97.351 -66.123 52.181
420260 97.388 -66.130 52.152
420261 97.425 -66.137 52.125
420262 97.461 -66.143 52.100
420263 97.497 -66.149 52.069
420264 97.533 -66.156 52.035
420265 97.572 -66.164 51.999
420266 97.610 -66.172 51.963
420267 97.648 -66.180 51.928
420268 97.686 -66.188 51.897
420269 97.725 -66.197 51.864
420270 97.763 -66.205 51.827
420271 97.798 -66.210 51.788
420272 97.830 -66.214 51.746
420273 97.864 -66.219 51.701
420274 97.897 -66.223 51.655
420275 97.930 -66.227 51.608
420276 97.963 -66.231 51.560
420277 97.996 -66.235 51.510
420278 98.029 -66.239 51.460
420279 98.062 -66.244 51.410
420280 98.095 -66.248 51.359
420281 98.129 -66.252 51.309
420282 98.161 -66.257 51.259
420283 98.193 -66.261 51.210
420284 98.216 -66.270 51.163
420285 98.243 -66.275 51.118
420286 98.278 -66.275 51.071
420287 98.312 -66.276 51.025
420288 98.349 -66.279 50.979
420289 98.385 -66.282 50.932
420290 98.421 -66.285 50.886
420291 98.457 -66.288 50.839
420292 98.497 -66.292 50.790
420293 98.539 -66.294 50.742
420294 98.582 -66.296 50.692
420295 98.625 -66.298 50.641
420296 98.668 -66.300 50.593
420297 98.711 -66.302 50.550
420298 98.754 -66.304 50.506
420299 98.796 -66.306 50.464
420300 98.839 -66.310 50.422
420301 98.879 -66.318 50.376
420302 98.917 -66.335 50.326
420303 98.954 -66.351 50.273
420304 98.992 -66.366 50.217
420305 99.029 -66.379 50.159
420306 99.067 -66.394 50.100
420307 99.105 -66.408 50.041
420308 99.142 -66.423 49.983
420309 99.180 -66.437 49.925
420310 99.220 -66.445 49.869
420311 99.262 -66.445 49.814
420312 99.304 -66.445 49.759
420313 99.346 -66.445 49.705
420314 99.389 -66.445 49.651
420315 99.430 -66.440 49.597
420316 99.472 -66.435 49.544
420317 99.513 -66.429 49.490
420318 99.554 -66.425 49.436
420319 99.574 -66.429 49.383
420320 99.595 -66.432 49.330
420321 99.613 -66.436 49.278
420322 99.652 -66.447 49.226
420323 99.698 -66.456 49.173
420324 99.745 -66.465 49.120
420325 99.790 -66.473 49.069
420326 99.837 -66.483 49.020
420327 99.883 -66.492 48.972
420328 99.927 -66.501 48.925
420329 99.947 -66.508 48.878
420330 99.986 -66.507 48.833
420331 100.027 -66.505 48.788
420332 100.075 -66.502 48.744
420333 100.124 -66.500 48.702
420334 100.172 -66.498 48.658
420335 100.220 -66.496 48.616
420336 100.268 -66.494 48.575
420337 100.317 -66.492 48.533
420338 100.365 -66.490 48.491
420339 100.413 -66.488 48.452
420753 88.522 -65.304 55.283
420754 88.556 -65.314 55.265
420755 88.591 -65.323 55.249
420756 88.625 -65.333 55.234
420757 88.659 -65.342 55.218
420758 88.693 -65.352 55.203
420759 88.728 -65.361 55.189
420760 88.762 -65.371 55.177
420761 88.796 -65.381 55.166
420762 88.830 -65.390 55.158
420763 88.864 -65.400 55.149
420764 88.904 -65.407 55.140
420765 88.950 -65.411 55.130
420766 88.996 -65.415 55.121
420767 89.042 -65.419 55.113
420768 89.089 -65.423 55.105
420769 89.134 -65.427 55.095
420770 89.181 -65.430 55.084
420771 89.227 -65.435 55.074
420772 89.273 -65.439 55.061
420773 89.319 -65.442 55.050
420774 89.366 -65.446 55.039
420775 89.412 -65.450 55.027
420776 89.458 -65.454 55.016
420777 89.504 -65.458 55.002
420778 89.551 -65.459 54.989
420779 89.608 -65.442 54.978
420780 89.664 -65.438 54.968
420781 89.720 -65.442 54.959
420782 89.775 -65.447 54.951
420793 90.320 -65.527 54.960
420794 90.369 -65.533 54.966
420795 90.417 -65.539 54.970
420796 90.466 -65.545 54.972
420797 90.514 -65.551 54.975
420798 90.563 -65.557 54.978
420799 90.611 -65.563 54.979
420800 90.660 -65.570 54.979
420801 90.708 -65.576 54.979
420802 90.756 -65.582 54.979
420803 90.805 -65.588 54.977
420804 90.854 -65.594 54.972
420805 90.902 -65.600 54.967
420806 90.950 -65.606 54.962
420807 90.999 -65.613 54.956
420808 91.044 -65.624 54.951
420809 91.090 -65.636 54.946
420810 91.136 -65.648 54.942
420811 91.181 -65.659 54.938
420812 91.227 -65.671 54.935
420813 91.273 -65.682 54.932
420814 91.316 -65.693 54.929
420815 91.359 -65.703 54.930
420816 91.401 -65.715 54.932
420817 91.444 -65.725 54.935
420818 91.488 -65.736 54.938
420819 91.530 -65.746 54.944
420820 91.573 -65.757 54.951
420821 91.616 -65.768 54.959
420822 91.659 -65.779 54.968
420823 91.702 -65.786 54.974
420824 91.746 -65.790 54.980
420825 91.784 -65.800 54.986
420826 91.818 -65.815 54.991
420827 91.852 -65.829 54.996
420828 91.893 -65.839 55.001
420829 91.941 -65.844 55.004
420830 91.990 -65.850 55.006
420831 92.038 -65.856 55.008
420832 92.087 -65.862 55.010
420833 92.135 -65.867 55.011
420834 92.183 -65.873 55.009
420835 92.232 -65.880 55.003
420836 92.281 -65.890 54.996
420837 92.327 -65.893 54.985
420838 92.371 -65.884 54.973
420839 92.415 -65.875 54.962
420840 92.464 -65.877 54.953
420841 92.513 -65.878 54.944
420842 92.562 -65.880 54.936
420843 92.611 -65.881 54.923
420844 92.659 -65.883 54.910
420845 92.708 -65.884 54.897
420846 92.757 -65.885 54.883
420847 92.806 -65.887 54.868
420848 92.855 -65.888 54.853
420849 92.904 -65.890 54.838
420850 92.952 -65.891 54.824
420851 93.001 -65.887 54.812
420852 93.048 -65.881 54.803
420853 93.096 -65.874 54.796
420854 93.143 -65.867 54.791
420855 93.191 -65.860 54.785
420856 93.238 -65.854 54.779
420857 93.287 -65.848 54.776
420858 93.336 -65.844 54.774
420859 93.384 -65.840 54.772
420860 93.431 -65.837 54.769
420861 93.475 -65.837 54.764
420862 93.519 -65.837 54.751
420863 93.564 -65.837 54.735
420864 93.609 -65.835 54.716
420865 93.654 -65.834 54.694
420866 93.699 -65.834 54.673
420867 93.745 -65.832 54.655
420868 93.790 -65.831 54.638
420869 93.835 -65.830 54.621
420870 93.880 -65.829 54.603
420871 93.925 -65.824 54.580
420872 93.969 -65.816 54.554
420873 94.008 -65.819 54.526
420874 94.048 -65.823 54.497
420875 94.087 -65.827 54.466
420876 94.127 -65.830 54.436
420877 94.167 -65.834 54.405
420878 94.206 -65.837 54.373
420879 94.246 -65.840 54.342
420880 94.274 -65.851 54.309
420881 94.292 -65.868 54.275
420882 94.311 -65.884 54.242
420883 94.329 -65.900 54.209
420884 94.352 -65.914 54.175
420885 94.388 -65.922 54.144
420886 94.423 -65.929 54.115
420887 94.458 -65.937 54.088
420888 94.493 -65.944 54.067
420889 94.529 -65.952 54.054
420890 94.564 -65.959 54.046
420891 94.599 -65.967 54.042
420892 94.634 -65.975 54.038
420893 94.670 -65.982 54.035
420894 94.705 -65.989 54.027
420895 94.740 -65.997 54.016
420896 94.780 -66.001 54.001
420897 94.820 -66.005 53.983
420898 94.859 -66.009 53.962
420899 94.900 -66.012 53.939
420900 94.943 -66.013 53.911
420901 94.984 -66.016 53.879
420902 95.025 -66.018 53.843
420903 95.065 -66.021 53.803
420904 95.106 -66.023 53.764
420905 95.147 -66.025 53.727
420906 95.187 -66.028 53.692
420907 95.228 -66.030 53.656
420908 95.269 -66.030 53.623
420909 95.308 -66.035 53.592
420910 95.349 -66.040 53.564
420911 95.391 -66.044 53.541
420912 95.432 -66.049 53.525
420913 95.474 -66.053 53.516
420914 95.516 -66.058 53.509
420915 95.557 -66.063 53.500
420916 95.599 -66.067 53.489
420917 95.640 -66.070 53.479
420918 95.681 -66.065 53.461
420919 95.722 -66.060 53.439
420920 95.762 -66.054 53.414
420921 95.803 -66.049 53.387
420922 95.844 -66.044 53.359
420923 95.891 -66.040 53.329
420924 95.938 -66.037 53.299
420925 95.984 -66.033 53.271
420926 96.026 -66.027 53.242
420927 96.068 -66.022 53.217
420928 96.110 -66.016 53.195
420929 96.153 -66.010 53.172
420930 96.194 -66.005 53.150
420931 96.236 -66.001 53.127
420932 96.277 -65.998 53.102
420933 96.319 -65.994 53.077
420934 96.358 -65.989 53.051
420935 96.394 -65.983 53.024
420936 96.431 -65.977 52.995
420937 96.467 -65.972 52.965
420938 96.503 -65.966 52.935
420939 96.542 -65.958 52.904
420940 96.578 -65.959 52.872
420941 96.617 -65.966 52.840
420942 96.658 -65.972 52.808
420943 96.697 -65.978 52.778
420944 96.738 -65.984 52.748
420945 96.778 -65.991 52.717
420946 96.817 -65.997 52.682
420947 96.857 -66.003 52.646
420948 96.897 -66.009 52.608
420949 96.937 -66.015 52.569
420950 96.978 -66.019 52.531
420951 97.017 -66.028 52.492
420952 97.054 -66.035 52.454
420953 97.092 -66.041 52.418
420954 97.129 -66.047 52.385
420955 97.166 -66.054 52.353
420956 97.204 -66.060 52.322
420957 97.241 -66.067 52.288
420958 97.278 -66.073 52.253
420959 97.315 -66.080 52.217
420960 97.352 -66.087 52.184
420961 97.389 -66.093 52.154
420962 97.425 -66.100 52.126
420963 97.461 -66.106 52.101
420964 97.497 -66.112 52.071
420965 97.533 -66.118 52.037
420966 97.571 -66.125 52.002
420967 97.609 -66.133 51.966
420968 97.647 -66.141 51.931
420969 97.685 -66.150 51.899
420970 97.724 -66.158 51.865
420971 97.761 -66.166 51.829
420972 97.795 -66.170 51.790
420973 97.828 -66.174 51.747
420974 97.861 -66.179 51.702
420975 97.894 -66.182 51.656
420976 97.927 -66.187 51.607
420977 97.960 -66.191 51.558
420978 97.993 -66.195 51.506
420979 98.026 -66.200 51.456
420980 98.059 -66.204 51.406
420981 98.092 -66.208 51.356
420982 98.126 -66.212 51.307
420983 98.159 -66.216 51.259
420984 98.192 -66.220 51.211
420985 98.219 -66.227 51.166
420986 98.247 -66.232 51.120
420987 98.283 -66.232 51.072
420988 98.319 -66.235 51.025
420989 98.355 -66.238 50.978
420990 98.391 -66.242 50.932
420991 98.428 -66.245 50.885
420992 98.464 -66.249 50.836
420993 98.502 -66.252 50.787
420994 98.545 -66.254 50.737
420995 98.588 -66.256 50.686
420996 98.631 -66.258 50.634
420997 98.673 -66.260 50.583
420998 98.716 -66.262 50.535
420999 98.759 -66.265 50.489
421000 98.801 -66.270 50.444
421001 98.844 -66.274 50.402
421002 98.885 -66.281 50.356
421003 98.923 -66.293 50.309
421004 98.961 -66.310 50.257
421005 98.998 -66.326 50.203
421006 99.035 -66.343 50.147
421007 99.073 -66.360 50.091
421008 99.111 -66.376 50.035
421009 99.148 -66.390 49.978
421010 99.186 -66.404 49.921
421011 99.224 -66.415 49.866
421012 99.267 -66.415 49.813
421013 99.309 -66.414 49.758
421014 99.351 -66.409 49.703
421015 99.392 -66.404 49.649
421016 99.434 -66.399 49.595
421017 99.476 -66.394 49.541
421018 99.517 -66.389 49.487
421019 99.550 -66.387 49.433
421020 99.571 -66.391 49.380
421021 99.591 -66.394 49.328
421022 99.615 -66.398 49.275
421023 99.655 -66.403 49.223
421024 99.702 -66.412 49.171
421025 99.748 -66.420 49.119
421026 99.794 -66.429 49.068
421027 99.841 -66.439 49.019
421028 99.887 -66.448 48.972
421029 99.933 -66.456 48.924
421030 99.956 -66.463 48.877
421031 99.996 -66.461 48.832
421032 100.042 -66.458 48.787
421033 100.090 -66.456 48.743
421034 100.138 -66.454 48.700
421035 100.186 -66.452 48.657
421036 100.234 -66.450 48.615
421037 100.283 -66.448 48.574
421038 100.331 -66.446 48.532
421039 100.379 -66.444 48.490
421040 100.427 -66.442 48.451
421451 88.412 -65.227 55.341
421452 88.446 -65.236 55.321
421453 88.480 -65.246 55.302
421454 88.515 -65.256 55.283
421455 88.549 -65.265 55.267
421456 88.583 -65.275 55.253
421457 88.617 -65.284 55.239
421458 88.651 -65.294 55.224
421459 88.686 -65.304 55.211
421460 88.720 -65.313 55.198
421461 88.754 -65.323 55.187
421462 88.788 -65.333 55.177
421463 88.823 -65.342 55.168
421464 88.860 -65.350 55.162
421465 88.906 -65.354 55.153
421466 88.952 -65.358 55.144
421467 88.998 -65.361 55.134
421468 89.045 -65.366 55.127
421469 89.091 -65.370 55.118
421470 89.137 -65.373 55.108
421471 89.183 -65.377 55.098
421472 89.229 -65.381 55.087
421473 89.275 -65.385 55.074
421474 89.322 -65.389 55.063
421475 89.368 -65.393 55.051
421476 89.414 -65.397 55.040
421477 89.460 -65.401 55.028
421478 89.506 -65.405 55.013
421479 89.559 -65.395 55.000
421480 89.616 -65.384 54.987
421481 89.671 -65.389 54.977
421482 89.727 -65.393 54.967
421483 89.782 -65.397 54.958
421494 90.319 -65.477 54.961
421495 90.368 -65.483 54.966
421496 90.416 -65.489 54.971
421497 90.465 -65.496 54.974
421498 90.513 -65.502 54.976
421499 90.561 -65.508 54.979
421500 90.610 -65.514 54.980
421501 90.658 -65.521 54.980
421502 90.707 -65.527 54.980
421503 90.755 -65.533 54.979
421504 90.804 -65.539 54.977
421505 90.852 -65.546 54.971
421506 90.900 -65.552 54.965
421507 90.949 -65.558 54.959
421508 90.997 -65.566 54.954
421509 91.042 -65.577 54.947
421510 91.088 -65.589 54.941
421511 91.134 -65.600 54.936
421512 91.179 -65.612 54.931
421513 91.225 -65.624 54.927
421514 91.269 -65.634 54.923
421515 91.312 -65.645 54.921
421516 91.355 -65.656 54.921
421517 91.397 -65.667 54.922
421518 91.441 -65.677 54.924
421519 91.484 -65.688 54.927
421520 91.526 -65.699 54.932
421521 91.569 -65.710 54.938
421522 91.612 -65.721 54.946
421523 91.655 -65.731 54.953
421524 91.698 -65.742 54.960
421525 91.739 -65.749 54.964
421526 91.774 -65.764 54.968
421527 91.808 -65.778 54.971
421528 91.843 -65.793 54.974
421529 91.883 -65.803 54.978
421530 91.931 -65.809 54.981
421531 91.980 -65.814 54.986
421532 92.028 -65.820 54.990
421533 92.077 -65.825 54.993
421534 92.125 -65.831 54.995
421535 92.173 -65.837 54.994
421536 92.222 -65.843 54.989
421537 92.271 -65.848 54.982
421538 92.317 -65.848 54.970
421539 92.366 -65.850 54.957
421540 92.415 -65.851 54.947
421541 92.463 -65.853 54.938
421542 92.513 -65.854 54.929
421543 92.562 -65.856 54.922
421544 92.611 -65.856 54.911
421545 92.659 -65.858 54.900
421546 92.708 -65.859 54.889
421547 92.757 -65.861 54.877
421548 92.806 -65.862 54.864
421549 92.855 -65.864 54.851
421550 92.904 -65.865 54.839
421551 92.952 -65.862 54.827
421552 92.999 -65.856 54.819
421553 93.047 -65.849 54.812
421554 93.094 -65.842 54.808
421555 93.142 -65.836 54.803
421556 93.190 -65.830 54.798
421557 93.239 -65.826 54.791
421558 93.288 -65.821 54.787
421559 93.337 -65.817 54.785
421560 93.386 -65.812 54.783
421561 93.430 -65.812 54.781
421562 93.474 -65.812 54.777
421563 93.518 -65.812 54.767
421564 93.562 -65.812 54.753
421565 93.607 -65.812 54.736
421566 93.652 -65.810 54.717
421567 93.698 -65.809 54.697
421568 93.743 -65.809 54.679
421569 93.788 -65.807 54.659
421570 93.833 -65.806 54.638
421571 93.878 -65.805 54.617
421572 93.924 -65.799 54.594
421573 93.970 -65.784 54.568
421574 94.013 -65.778 54.541
421575 94.052 -65.781 54.514
421576 94.092 -65.784 54.483
421577 94.131 -65.788 54.451
421578 94.171 -65.791 54.419
421579 94.210 -65.795 54.387
421580 94.247 -65.801 54.356
421581 94.265 -65.817 54.325
421582 94.283 -65.834 54.293
421583 94.301 -65.850 54.261
421584 94.320 -65.867 54.228
421585 94.346 -65.879 54.195
421586 94.382 -65.887 54.163
421587 94.417 -65.894 54.134
421588 94.452 -65.902 54.108
421589 94.487 -65.909 54.089
421590 94.522 -65.916 54.077
421591 94.557 -65.924 54.070
421592 94.593 -65.931 54.067
421593 94.628 -65.939 54.066
421594 94.664 -65.947 54.063
421595 94.703 -65.950 54.056
421596 94.743 -65.954 54.044
421597 94.783 -65.958 54.028
421598 94.823 -65.963 54.008
421599 94.862 -65.966 53.985
421600 94.904 -65.969 53.959
421601 94.947 -65.970 53.929
421602 94.988 -65.972 53.895
421603 95.029 -65.975 53.856
421604 95.070 -65.977 53.815
421605 95.111 -65.979 53.774
421606 95.151 -65.981 53.735
421607 95.192 -65.984 53.699
421608 95.233 -65.987 53.663
421609 95.274 -65.990 53.629
421610 95.318 -65.994 53.598
421611 95.360 -66.000 53.571
421612 95.402 -66.004 53.550
421613 95.444 -66.009 53.538
421614 95.485 -66.013 53.531
421615 95.527 -66.018 53.524
421616 95.568 -66.022 53.515
421617 95.610 -66.027 53.504
421618 95.651 -66.031 53.491
421619 95.692 -66.028 53.474
421620 95.733 -66.022 53.452
421621 95.774 -66.016 53.427
421622 95.815 -66.011 53.399
421623 95.855 -66.006 53.369
421624 95.896 -66.000 53.338
421625 95.941 -65.997 53.308
421626 95.988 -65.993 53.280
421627 96.031 -65.988 53.251
421628 96.073 -65.981 53.226
421629 96.115 -65.976 53.202
421630 96.157 -65.972 53.179
421631 96.198 -65.969 53.157
421632 96.240 -65.965 53.133
421633 96.282 -65.961 53.108
421634 96.323 -65.957 53.081
421635 96.361 -65.952 53.053
421636 96.398 -65.946 53.023
421637 96.434 -65.941 52.992
421638 96.471 -65.934 52.961
421639 96.512 -65.925 52.930
421640 96.553 -65.916 52.899
421641 96.589 -65.918 52.868
421642 96.620 -65.930 52.838
421643 96.660 -65.936 52.807
421644 96.700 -65.942 52.777
421645 96.740 -65.948 52.749
421646 96.781 -65.954 52.719
421647 96.821 -65.960 52.685
421648 96.861 -65.966 52.650
421649 96.901 -65.972 52.613
421650 96.942 -65.979 52.576
421651 96.981 -65.984 52.538
421652 97.019 -65.991 52.499
421653 97.056 -65.998 52.460
421654 97.093 -66.004 52.422
421655 97.131 -66.011 52.388
421656 97.168 -66.017 52.356
421657 97.205 -66.024 52.325
421658 97.242 -66.030 52.292
421659 97.280 -66.037 52.257
421660 97.316 -66.043 52.221
421661 97.354 -66.050 52.187
421662 97.389 -66.056 52.155
421663 97.426 -66.062 52.127
421664 97.461 -66.069 52.101
421665 97.498 -66.075 52.070
421666 97.533 -66.081 52.038
421667 97.570 -66.087 52.003
421668 97.607 -66.094 51.968
421669 97.646 -66.103 51.933
421670 97.684 -66.111 51.901
421671 97.723 -66.119 51.866
421672 97.758 -66.125 51.829
421673 97.792 -66.130 51.789
421674 97.825 -66.135 51.746
421675 97.858 -66.138 51.701
421676 97.891 -66.143 51.654
421677 97.924 -66.147 51.605
421678 97.957 -66.151 51.555
421679 97.990 -66.155 51.502
421680 98.023 -66.160 51.452
421681 98.057 -66.163 51.404
421682 98.089 -66.168 51.356
421683 98.123 -66.172 51.307
421684 98.156 -66.176 51.260
421685 98.189 -66.181 51.214
421686 98.222 -66.185 51.168
421687 98.252 -66.188 51.121
421688 98.289 -66.191 51.073
421689 98.325 -66.195 51.024
421690 98.361 -66.198 50.978
421691 98.397 -66.202 50.931
421692 98.434 -66.205 50.882
421693 98.470 -66.209 50.833
421694 98.507 -66.212 50.783
421695 98.550 -66.214 50.731
421696 98.593 -66.216 50.679
421697 98.636 -66.218 50.626
421698 98.678 -66.220 50.573
421699 98.721 -66.225 50.523
421700 98.764 -66.229 50.474
421701 98.807 -66.234 50.427
421702 98.850 -66.238 50.381
421703 98.890 -66.245 50.335
421704 98.929 -66.252 50.288
421705 98.967 -66.268 50.239
421706 99.004 -66.285 50.187
421707 99.042 -66.301 50.135
421708 99.079 -66.318 50.082
421709 99.117 -66.335 50.028
421710 99.154 -66.351 49.974
421711 99.192 -66.368 49.919
421712 99.229 -66.382 49.866
421713 99.271 -66.378 49.814
421714 99.312 -66.373 49.759
421715 99.354 -66.368 49.705
421716 99.396 -66.363 49.650
421717 99.437 -66.358 49.595
421718 99.479 -66.353 49.541
421719 99.521 -66.348 49.486
421720 99.547 -66.349 49.432
421721 99.567 -66.353 49.378
421722 99.592 -66.356 49.326
421723 99.631 -66.361 49.274
421724 99.670 -66.364 49.222
421725 99.710 -66.370 49.170
421726 99.752 -66.376 49.119
421727 99.798 -66.386 49.068
421728 99.844 -66.394 49.019
421729 99.891 -66.403 48.972
421730 99.937 -66.412 48.924
421731 99.965 -66.417 48.877
421732 100.008 -66.414 48.831
421733 100.056 -66.412 48.786
421734 100.104 -66.411 48.741
421735 100.152 -66.408 48.699
421736 100.201 -66.406 48.657
421737 100.249 -66.404 48.615
421738 100.297 -66.402 48.573
421739 100.345 -66.400 48.531
421740 100.394 -66.398 48.490
421741 100.442 -66.395 48.450
422152 88.404 -65.179 55.335
422153 88.438 -65.188 55.317
422154 88.472 -65.198 55.300
422155 88.506 -65.207 55.283
422156 88.541 -65.217 55.269
422157 88.575 -65.226 55.256
422158 88.609 -65.236 55.244
422159 88.643 -65.245 55.231
422160 88.678 -65.255 55.220
422161 88.712 -65.265 55.208
422162 88.746 -65.274 55.199
422163 88.780 -65.284 55.190
422164 88.816 -65.293 55.182
422165 88.862 -65.297 55.175
422166 88.908 -65.301 55.168
422167 88.955 -65.304 55.159
422168 89.000 -65.308 55.150
422169 89.047 -65.312 55.142
422170 89.093 -65.317 55.132
422171 89.139 -65.320 55.123
422172 89.185 -65.324 55.112
422173 89.232 -65.328 55.099
422174 89.278 -65.332 55.088
422175 89.324 -65.336 55.076
422176 89.370 -65.340 55.064
422177 89.416 -65.344 55.052
422178 89.462 -65.348 55.039
422179 89.511 -65.348 55.025
422180 89.567 -65.330 55.011
422181 89.623 -65.335 54.997
422182 89.678 -65.339 54.986
422183 89.734 -65.343 54.975
422184 89.785 -65.351 54.965
422185 89.833 -65.361 54.957
422195 90.318 -65.428 54.962
422196 90.367 -65.434 54.967
422197 90.415 -65.441 54.971
422198 90.463 -65.447 54.975
422199 90.512 -65.453 54.978
422200 90.561 -65.459 54.980
422201 90.609 -65.465 54.982
422202 90.657 -65.471 54.981
422203 90.706 -65.477 54.980
422204 90.754 -65.484 54.978
422205 90.803 -65.490 54.974
422206 90.851 -65.496 54.968
422207 90.899 -65.502 54.963
422208 90.948 -65.508 54.957
422209 90.994 -65.519 54.950
422210 91.040 -65.530 54.943
422211 91.086 -65.542 54.937
422212 91.132 -65.553 54.930
422213 91.177 -65.564 54.925
422214 91.222 -65.576 54.920
422215 91.265 -65.586 54.915
422216 91.308 -65.598 54.913
422217 91.351 -65.608 54.913
422218 91.394 -65.619 54.913
422219 91.437 -65.630 54.914
422220 91.480 -65.640 54.916
422221 91.522 -65.652 54.921
422222 91.565 -65.662 54.927
422223 91.608 -65.673 54.933
422224 91.651 -65.684 54.941
422225 91.694 -65.696 54.949
422226 91.732 -65.712 54.954
422227 91.764 -65.727 54.957
422228 91.799 -65.742 54.958
422229 91.833 -65.756 54.959
422230 91.873 -65.767 54.960
422231 91.921 -65.773 54.961
422232 91.969 -65.778 54.964
422233 92.018 -65.784 54.969
422234 92.066 -65.790 54.972
422235 92.115 -65.795 54.975
422236 92.163 -65.801 54.974
422237 92.214 -65.803 54.970
422238 92.264 -65.805 54.964
422239 92.315 -65.815 54.954
422240 92.366 -65.825 54.941
422241 92.415 -65.827 54.930
422242 92.463 -65.828 54.921
422243 92.512 -65.829 54.914
422244 92.561 -65.831 54.908
422245 92.610 -65.832 54.899
422246 92.659 -65.834 54.890
422247 92.708 -65.835 54.881
422248 92.757 -65.837 54.871
422249 92.806 -65.837 54.859
422250 92.855 -65.839 54.848
422251 92.903 -65.837 54.838
422252 92.950 -65.831 54.829
422253 92.998 -65.824 54.825
422254 93.046 -65.818 54.821
422255 93.094 -65.812 54.818
422256 93.142 -65.807 54.815
422257 93.191 -65.803 54.809
422258 93.240 -65.799 54.803
422259 93.289 -65.794 54.798
422260 93.338 -65.790 54.795
422261 93.386 -65.786 54.793
422262 93.430 -65.787 54.791
422263 93.474 -65.787 54.788
422264 93.517 -65.787 54.779
422265 93.561 -65.787 54.767
422266 93.605 -65.787 54.752
422267 93.650 -65.787 54.736
422268 93.696 -65.785 54.718
422269 93.741 -65.784 54.698
422270 93.786 -65.783 54.678
422271 93.831 -65.782 54.655
422272 93.876 -65.781 54.632
422273 93.922 -65.774 54.607
422274 93.968 -65.759 54.581
422275 94.014 -65.744 54.554
422276 94.057 -65.739 54.527
422277 94.096 -65.743 54.496
422278 94.136 -65.746 54.464
422279 94.175 -65.749 54.432
422280 94.215 -65.753 54.402
422281 94.238 -65.767 54.375
422282 94.256 -65.784 54.345
422283 94.274 -65.800 54.314
422284 94.293 -65.816 54.282
422285 94.311 -65.833 54.248
422286 94.340 -65.844 54.214
422287 94.375 -65.851 54.183
422288 94.410 -65.859 54.154
422289 94.445 -65.866 54.130
422290 94.481 -65.874 54.111
422291 94.516 -65.881 54.099
422292 94.551 -65.889 54.092
422293 94.587 -65.896 54.090
422294 94.627 -65.900 54.090
422295 94.667 -65.903 54.090
422296 94.706 -65.908 54.083
422297 94.746 -65.912 54.070
422298 94.786 -65.916 54.052
422299 94.826 -65.920 54.029
422300 94.865 -65.924 54.004
422301 94.908 -65.925 53.977
422302 94.951 -65.926 53.945
422303 94.993 -65.928 53.908
422304 95.034 -65.931 53.868
422305 95.075 -65.933 53.826
422306 95.115 -65.935 53.784
422307 95.156 -65.938 53.744
422308 95.197 -65.941 53.706
422309 95.238 -65.943 53.670
422310 95.281 -65.950 53.636
422311 95.323 -65.949 53.605
422312 95.367 -65.956 53.579
422313 95.411 -65.963 53.560
422314 95.454 -65.969 53.550
422315 95.496 -65.974 53.544
422316 95.538 -65.978 53.538
422317 95.579 -65.983 53.530
422318 95.621 -65.988 53.519
422319 95.662 -65.992 53.506
422320 95.703 -65.990 53.488
422321 95.744 -65.984 53.466
422322 95.785 -65.979 53.440
422323 95.826 -65.974 53.410
422324 95.867 -65.968 53.380
422325 95.907 -65.963 53.349
422326 95.948 -65.957 53.318
422327 95.991 -65.953 53.289
422328 96.035 -65.948 53.260
422329 96.077 -65.943 53.235
422330 96.119 -65.939 53.210
422331 96.161 -65.935 53.186
422332 96.202 -65.931 53.164
422333 96.244 -65.928 53.140
422334 96.286 -65.925 53.114
422335 96.327 -65.921 53.085
422336 96.365 -65.915 53.055
422337 96.401 -65.909 53.022
422338 96.442 -65.901 52.990
422339 96.483 -65.892 52.957
422340 96.524 -65.884 52.924
422341 96.565 -65.875 52.892
422342 96.601 -65.876 52.861
422343 96.631 -65.888 52.830
422344 96.663 -65.900 52.801
422345 96.702 -65.906 52.772
422346 96.743 -65.912 52.743
422347 96.783 -65.918 52.712
422348 96.823 -65.925 52.679
422349 96.864 -65.931 52.644
422350 96.904 -65.937 52.608
422351 96.944 -65.943 52.571
422352 96.983 -65.948 52.535
422353 97.021 -65.954 52.496
422354 97.058 -65.960 52.457
422355 97.095 -65.967 52.421
422356 97.132 -65.974 52.386
422357 97.169 -65.981 52.354
422358 97.206 -65.987 52.324
422359 97.243 -65.994 52.292
422360 97.281 -66.000 52.256
422361 97.318 -66.006 52.221
422362 97.354 -66.013 52.187
422363 97.390 -66.019 52.154
422364 97.426 -66.025 52.124
422365 97.462 -66.031 52.097
422366 97.498 -66.037 52.065
422367 97.534 -66.044 52.033
422368 97.570 -66.050 52.000
422369 97.606 -66.056 51.966
422370 97.645 -66.064 51.933
422371 97.683 -66.072 51.901
422372 97.721 -66.081 51.864
422373 97.755 -66.086 51.825
422374 97.789 -66.090 51.784
422375 97.822 -66.094 51.740
422376 97.855 -66.098 51.696
422377 97.888 -66.103 51.650
422378 97.921 -66.107 51.602
422379 97.954 -66.111 51.552
422380 97.987 -66.115 51.501
422381 98.020 -66.119 51.452
422382 98.054 -66.123 51.406
422383 98.086 -66.128 51.360
422384 98.120 -66.132 51.312
422385 98.153 -66.136 51.263
422386 98.187 -66.141 51.216
422387 98.222 -66.145 51.169
422388 98.262 -66.154 51.122
422389 98.296 -66.154 51.073
422390 98.332 -66.155 51.024
422391 98.368 -66.158 50.976
422392 98.404 -66.162 50.928
422393 98.440 -66.165 50.877
422394 98.477 -66.169 50.827
422395 98.513 -66.172 50.775
422396 98.555 -66.174 50.723
422397 98.598 -66.176 50.670
422398 98.641 -66.180 50.618
422399 98.684 -66.185 50.565
422400 98.726 -66.189 50.513
422401 98.769 -66.193 50.463
422402 98.812 -66.198 50.416
422403 98.855 -66.202 50.366
422404 98.895 -66.209 50.318
422405 98.934 -66.216 50.269
422406 98.974 -66.226 50.221
422407 99.011 -66.242 50.172
422408 99.048 -66.259 50.123
422409 99.086 -66.276 50.072
422410 99.123 -66.293 50.022
422411 99.160 -66.310 49.971
422412 99.198 -66.326 49.921
422413 99.237 -66.331 49.869
422414 99.277 -66.332 49.817
422415 99.316 -66.332 49.763
422416 99.358 -66.327 49.708
422417 99.400 -66.322 49.653
422418 99.441 -66.317 49.598
422419 99.483 -66.312 49.543
422420 99.522 -66.308 49.488
422421 99.543 -66.311 49.434
422422 99.569 -66.315 49.379
422423 99.609 -66.319 49.327
422424 99.647 -66.323 49.274
422425 99.686 -66.327 49.223
422426 99.725 -66.332 49.172
422427 99.766 -66.338 49.120
422428 99.806 -66.343 49.069
422429 99.848 -66.350 49.020
422430 99.894 -66.359 48.971
422431 99.940 -66.368 48.924
422432 99.974 -66.370 48.877
422433 100.022 -66.369 48.831
422434 100.070 -66.367 48.786
422435 100.119 -66.364 48.742
422436 100.167 -66.362 48.699
422437 100.215 -66.360 48.656
422438 100.263 -66.358 48.614
422439 100.312 -66.356 48.572
422440 100.360 -66.354 48.530
422441 100.408 -66.351 48.488
422442 100.456 -66.350 48.450
422853 88.396 -65.130 55.331
422854 88.430 -65.140 55.315
422855 88.464 -65.149 55.299
422856 88.498 -65.159 55.283
422857 88.533 -65.168 55.270
422858 88.567 -65.178 55.259
422859 88.601 -65.188 55.248
422860 88.635 -65.197 55.239
422861 88.670 -65.207 55.230
422862 88.704 -65.217 55.220
422863 88.738 -65.226 55.212
422864 88.772 -65.235 55.204
422865 88.818 -65.239 55.197
422866 88.864 -65.243 55.190
422867 88.911 -65.248 55.183
422868 88.957 -65.251 55.176
422869 89.003 -65.255 55.167
422870 89.049 -65.259 55.159
422871 89.095 -65.263 55.149
422872 89.141 -65.267 55.138
422873 89.188 -65.271 55.127
422874 89.234 -65.275 55.114
422875 89.280 -65.279 55.102
422876 89.326 -65.282 55.089
422877 89.373 -65.286 55.077
422878 89.418 -65.290 55.065
422879 89.465 -65.295 55.050
422880 89.518 -65.283 55.036
422881 89.574 -65.281 55.022
422882 89.630 -65.286 55.008
422883 89.685 -65.289 54.996
422884 89.735 -65.298 54.984
422885 89.784 -65.308 54.973
422886 89.832 -65.317 54.963
422896 90.317 -65.380 54.963
422897 90.365 -65.386 54.968
422898 90.414 -65.392 54.972
422899 90.462 -65.398 54.976
422900 90.511 -65.404 54.979
422901 90.559 -65.410 54.981
422902 90.607 -65.417 54.982
422903 90.656 -65.423 54.982
422904 90.705 -65.429 54.980
422905 90.753 -65.435 54.978
422906 90.801 -65.441 54.973
422907 90.850 -65.447 54.967
422908 90.898 -65.453 54.961
422909 90.946 -65.460 54.955
422910 90.992 -65.472 54.948
422911 91.038 -65.483 54.941
422912 91.084 -65.495 54.934
422913 91.130 -65.506 54.927
422914 91.175 -65.517 54.920
422915 91.218 -65.528 54.914
422916 91.261 -65.539 54.909
422917 91.304 -65.550 54.906
422918 91.347 -65.561 54.904
422919 91.390 -65.571 54.904
422920 91.433 -65.583 54.905
422921 91.475 -65.593 54.907
422922 91.518 -65.604 54.911
422923 91.562 -65.615 54.916
422924 91.604 -65.625 54.922
422925 91.646 -65.638 54.931
422926 91.686 -65.655 54.940
422927 91.726 -65.672 54.946
422928 91.760 -65.689 54.949
422929 91.789 -65.705 54.949
422930 91.823 -65.720 54.950
422931 91.862 -65.731 54.947
422932 91.911 -65.737 54.945
422933 91.959 -65.743 54.945
422934 92.008 -65.748 54.948
422935 92.057 -65.752 54.949
422936 92.108 -65.753 54.951
422937 92.158 -65.755 54.951
422938 92.209 -65.756 54.948
422939 92.260 -65.758 54.943
422940 92.310 -65.769 54.934
422941 92.362 -65.781 54.923
422942 92.413 -65.792 54.914
422943 92.463 -65.803 54.907
422944 92.512 -65.805 54.903
422945 92.561 -65.806 54.899
422946 92.610 -65.808 54.892
422947 92.659 -65.809 54.883
422948 92.708 -65.810 54.874
422949 92.757 -65.812 54.863
422950 92.805 -65.813 54.852
422951 92.854 -65.812 54.843
422952 92.902 -65.806 54.835
422953 92.949 -65.800 54.831
422954 92.997 -65.793 54.829
422955 93.046 -65.789 54.828
422956 93.095 -65.785 54.828
422957 93.144 -65.781 54.825
422958 93.193 -65.776 54.820
422959 93.241 -65.771 54.813
422960 93.290 -65.768 54.808
422961 93.339 -65.763 54.804
422962 93.386 -65.761 54.801
422963 93.429 -65.762 54.800
422964 93.473 -65.762 54.797
422965 93.517 -65.762 54.789
422966 93.560 -65.762 54.779
422967 93.604 -65.762 54.765
422968 93.648 -65.762 54.752
422969 93.694 -65.762 54.734
422970 93.739 -65.760 54.715
422971 93.784 -65.759 54.694
422972 93.829 -65.758 54.671
422973 93.874 -65.757 54.646
422974 93.920 -65.749 54.620
422975 93.966 -65.734 54.593
422976 94.012 -65.720 54.566
422977 94.058 -65.705 54.538
422978 94.101 -65.701 54.507
422979 94.140 -65.704 54.475
422980 94.180 -65.708 54.445
422981 94.210 -65.716 54.416
422982 94.228 -65.733 54.390
422983 94.247 -65.749 54.362
422984 94.265 -65.766 54.331
422985 94.284 -65.783 54.298
422986 94.302 -65.799 54.265
422987 94.333 -65.809 54.232
422988 94.368 -65.816 54.201
422989 94.404 -65.824 54.174
422990 94.439 -65.831 54.150
422991 94.474 -65.839 54.131
422992 94.511 -65.845 54.119
422993 94.551 -65.849 54.113
422994 94.590 -65.853 54.112
422995 94.630 -65.857 54.113
422996 94.670 -65.861 54.112
422997 94.709 -65.865 54.104
422998 94.749 -65.869 54.089
422999 94.789 -65.873 54.070
423000 94.829 -65.877 54.046
423001 94.869 -65.881 54.019
423002 94.912 -65.882 53.990
423003 94.955 -65.883 53.956
423004 94.998 -65.884 53.919
423005 95.039 -65.887 53.879
423006 95.079 -65.889 53.835
423007 95.120 -65.892 53.793
423008 95.161 -65.894 53.753
423009 95.202 -65.897 53.714
423010 95.244 -65.901 53.677
423011 95.288 -65.909 53.643
423012 95.327 -65.903 53.613
423013 95.372 -65.910 53.588
423014 95.416 -65.917 53.570
423015 95.460 -65.925 53.561
423016 95.505 -65.931 53.557
423017 95.548 -65.938 53.552
423018 95.590 -65.943 53.544
423019 95.632 -65.947 53.534
423020 95.673 -65.952 53.521
423021 95.714 -65.952 53.503
423022 95.755 -65.947 53.479
423023 95.796 -65.941 53.452
423024 95.837 -65.936 53.421
423025 95.878 -65.931 53.392
423026 95.918 -65.925 53.360
423027 95.959 -65.920 53.328
423028 96.000 -65.914 53.298
423029 96.040 -65.909 53.269
423030 96.081 -65.906 53.244
423031 96.123 -65.902 53.218
423032 96.165 -65.898 53.193
423033 96.206 -65.895 53.170
423034 96.248 -65.891 53.146
423035 96.290 -65.887 53.119
423036 96.331 -65.884 53.089
423037 96.372 -65.876 53.056
423038 96.413 -65.868 53.022
423039 96.454 -65.859 52.988
423040 96.495 -65.850 52.954
423041 96.536 -65.842 52.919
423042 96.577 -65.834 52.886
423043 96.612 -65.834 52.853
423044 96.643 -65.847 52.821
423045 96.673 -65.859 52.792
423046 96.705 -65.870 52.763
423047 96.745 -65.876 52.735
423048 96.786 -65.882 52.703
423049 96.826 -65.888 52.670
423050 96.866 -65.894 52.635
423051 96.906 -65.900 52.599
423052 96.946 -65.906 52.563
423053 96.985 -65.911 52.527
423054 97.023 -65.916 52.489
423055 97.061 -65.923 52.451
423056 97.097 -65.930 52.414
423057 97.134 -65.937 52.380
423058 97.171 -65.943 52.348
423059 97.208 -65.950 52.318
423060 97.245 -65.956 52.286
423061 97.282 -65.963 52.251
423062 97.318 -65.969 52.216
423063 97.354 -65.975 52.181
423064 97.390 -65.981 52.148
423065 97.426 -65.988 52.116
423066 97.462 -65.994 52.087
423067 97.498 -66.000 52.055
423068 97.534 -66.006 52.024
423069 97.570 -66.012 51.993
423070 97.606 -66.019 51.961
423071 97.644 -66.025 51.928
423072 97.682 -66.034 51.895
423073 97.720 -66.041 51.858
423074 97.753 -66.046 51.818
423075 97.786 -66.050 51.776
423076 97.819 -66.054 51.731
423077 97.852 -66.058 51.686
423078 97.885 -66.063 51.641
423079 97.918 -66.066 51.594
423080 97.951 -66.071 51.547
423081 97.984 -66.075 51.500
423082 98.017 -66.079 51.454
423083 98.052 -66.084 51.409
423084 98.086 -66.088 51.364
423085 98.121 -66.093 51.317
423086 98.155 -66.097 51.267
423087 98.190 -66.102 51.219
423088 98.240 -66.124 51.171
423089 98.277 -66.128 51.122
423090 98.312 -66.128 51.072
423091 98.346 -66.129 51.021
423092 98.380 -66.129 50.971
423093 98.415 -66.129 50.921
423094 98.449 -66.129 50.870
423095 98.484 -66.130 50.821
423096 98.519 -66.132 50.769
423097 98.560 -66.135 50.716
423098 98.603 -66.140 50.662
423099 98.646 -66.144 50.608
423100 98.689 -66.148 50.556
423101 98.731 -66.153 50.506
423102 98.774 -66.157 50.456
423103 98.817 -66.162 50.407
423104 98.860 -66.166 50.356
423105 98.900 -66.172 50.307
423106 98.939 -66.181 50.258
423107 98.979 -66.188 50.211
423108 99.017 -66.201 50.166
423109 99.055 -66.217 50.118
423110 99.092 -66.234 50.069
423111 99.130 -66.251 50.021
423112 99.167 -66.267 49.973
423113 99.205 -66.279 49.924
423114 99.244 -66.279 49.875
423115 99.284 -66.280 49.821
423116 99.324 -66.282 49.766
423117 99.363 -66.282 49.710
423118 99.403 -66.281 49.655
423119 99.445 -66.276 49.600
423120 99.487 -66.271 49.547
423121 99.519 -66.270 49.493
423122 99.547 -66.273 49.438
423123 99.585 -66.277 49.382
423124 99.624 -66.282 49.328
423125 99.663 -66.285 49.276
423126 99.702 -66.289 49.224
423127 99.740 -66.294 49.172
423128 99.781 -66.299 49.120
423129 99.821 -66.305 49.070
423130 99.861 -66.310 49.020
423131 99.902 -66.317 48.971
423132 99.944 -66.323 48.924
423133 99.988 -66.325 48.877
423134 100.037 -66.323 48.831
423135 100.085 -66.320 48.786
423136 100.133 -66.318 48.741
423137 100.181 -66.317 48.698
423138 100.229 -66.314 48.656
423139 100.278 -66.312 48.613
423140 100.326 -66.310 48.571
423141 100.374 -66.307 48.529
423142 100.422 -66.306 48.487
423143 100.471 -66.304 48.448
423554 88.388 -65.082 55.326
423555 88.422 -65.091 55.312
423556 88.457 -65.101 55.299
423557 88.491 -65.110 55.285
423558 88.525 -65.120 55.274
423559 88.559 -65.129 55.265
423560 88.594 -65.139 55.256
423561 88.628 -65.149 55.249
423562 88.662 -65.158 55.241
423563 88.696 -65.168 55.233
423564 88.731 -65.177 55.226
423565 88.774 -65.182 55.220
423566 88.821 -65.186 55.215
423567 88.867 -65.190 55.208
423568 88.913 -65.194 55.200
423569 88.959 -65.198 55.192
423570 89.005 -65.202 55.186
423571 89.051 -65.206 55.177
423572 89.098 -65.210 55.167
423573 89.144 -65.213 55.155
423574 89.190 -65.217 55.142
423575 89.236 -65.222 55.130
423576 89.282 -65.226 55.116
423577 89.328 -65.229 55.104
423578 89.375 -65.233 55.091
423579 89.421 -65.237 55.078
423580 89.469 -65.236 55.063
423581 89.526 -65.227 55.048
423582 89.581 -65.232 55.033
423583 89.637 -65.235 55.019
423584 89.686 -65.245 55.006
423585 89.734 -65.254 54.993
423586 89.783 -65.263 54.981
423587 89.831 -65.269 54.971
423588 89.880 -65.275 54.963
423597 90.316 -65.330 54.963
423598 90.364 -65.336 54.968
423599 90.412 -65.342 54.973
423600 90.461 -65.349 54.976
423601 90.510 -65.355 54.979
423602 90.558 -65.361 54.981
423603 90.606 -65.367 54.983
423604 90.655 -65.373 54.982
423605 90.704 -65.380 54.981
423606 90.752 -65.386 54.978
423607 90.800 -65.392 54.972
423608 90.849 -65.398 54.967
423609 90.897 -65.405 54.960
423610 90.944 -65.413 54.953
423611 90.990 -65.424 54.946
423612 91.036 -65.436 54.938
423613 91.082 -65.448 54.931
423614 91.127 -65.459 54.923
423615 91.171 -65.470 54.916
423616 91.214 -65.481 54.910
423617 91.257 -65.492 54.904
423618 91.300 -65.502 54.900
423619 91.343 -65.513 54.898
423620 91.386 -65.524 54.896
423621 91.429 -65.535 54.896
423622 91.472 -65.546 54.898
423623 91.515 -65.556 54.901
423624 91.558 -65.567 54.906
423625 91.599 -65.581 54.912
423626 91.639 -65.598 54.922
423627 91.678 -65.615 54.931
423628 91.718 -65.631 54.938
423629 91.758 -65.649 54.942
423630 91.788 -65.665 54.944
423631 91.815 -65.683 54.945
423632 91.852 -65.695 54.941
423633 91.901 -65.700 54.935
423634 91.952 -65.702 54.930
423635 92.002 -65.702 54.926
423636 92.053 -65.704 54.924
423637 92.103 -65.705 54.925
423638 92.154 -65.707 54.925
423639 92.205 -65.709 54.924
423640 92.255 -65.712 54.920
423641 92.306 -65.724 54.912
423642 92.357 -65.736 54.904
423643 92.408 -65.747 54.898
423644 92.459 -65.759 54.894
423645 92.510 -65.770 54.892
423646 92.561 -65.781 54.890
423647 92.610 -65.784 54.883
423648 92.659 -65.784 54.875
423649 92.708 -65.786 54.866
423650 92.757 -65.787 54.855
423651 92.805 -65.788 54.845
423652 92.853 -65.781 54.836
423653 92.901 -65.775 54.830
423654 92.949 -65.771 54.828
423655 92.998 -65.767 54.829
423656 93.047 -65.762 54.831
423657 93.096 -65.758 54.832
423658 93.145 -65.753 54.831
423659 93.194 -65.749 54.829
423660 93.242 -65.745 54.822
423661 93.291 -65.740 54.816
423662 93.340 -65.737 54.811
423663 93.384 -65.736 54.806
423664 93.428 -65.736 54.805
423665 93.472 -65.737 54.800
423666 93.516 -65.737 54.794
423667 93.560 -65.737 54.785
423668 93.604 -65.737 54.772
423669 93.647 -65.738 54.760
423670 93.692 -65.737 54.745
423671 93.737 -65.737 54.728
423672 93.782 -65.735 54.708
423673 93.827 -65.734 54.684
423674 93.872 -65.733 54.659
423675 93.918 -65.724 54.632
423676 93.964 -65.710 54.604
423677 94.010 -65.695 54.575
423678 94.056 -65.680 54.547
423679 94.103 -65.665 54.515
423680 94.144 -65.662 54.485
423681 94.183 -65.666 54.457
423682 94.201 -65.683 54.431
423683 94.220 -65.699 54.403
423684 94.238 -65.715 54.375
423685 94.257 -65.732 54.344
423686 94.275 -65.749 54.311
423687 94.293 -65.765 54.279
423688 94.327 -65.774 54.248
423689 94.362 -65.781 54.219
423690 94.398 -65.788 54.193
423691 94.434 -65.794 54.170
423692 94.474 -65.799 54.152
423693 94.514 -65.803 54.140
423694 94.554 -65.806 54.135
423695 94.593 -65.811 54.133
423696 94.633 -65.815 54.134
423697 94.673 -65.818 54.132
423698 94.712 -65.822 54.123
423699 94.752 -65.827 54.107
423700 94.792 -65.831 54.085
423701 94.832 -65.834 54.059
423702 94.873 -65.837 54.029
423703 94.916 -65.838 53.999
423704 94.959 -65.840 53.965
423705 95.002 -65.840 53.927
423706 95.044 -65.843 53.887
423707 95.084 -65.845 53.844
423708 95.125 -65.848 53.803
423709 95.166 -65.850 53.762
423710 95.206 -65.853 53.723
423711 95.251 -65.862 53.685
423712 95.295 -65.868 53.652
423713 95.334 -65.861 53.622
423714 95.377 -65.865 53.597
423715 95.421 -65.872 53.580
423716 95.465 -65.879 53.572
423717 95.510 -65.886 53.567
423718 95.554 -65.893 53.562
423719 95.598 -65.900 53.556
423720 95.642 -65.907 53.547
423721 95.684 -65.912 53.535
423722 95.726 -65.914 53.517
423723 95.766 -65.909 53.494
423724 95.807 -65.903 53.466
423725 95.848 -65.898 53.435
423726 95.887 -65.892 53.403
423727 95.926 -65.886 53.371
423728 95.965 -65.880 53.339
423729 96.004 -65.874 53.309
423730 96.044 -65.869 53.280
423731 96.085 -65.865 53.254
423732 96.127 -65.862 53.227
423733 96.169 -65.859 53.201
423734 96.210 -65.855 53.176
423735 96.252 -65.851 53.151
423736 96.294 -65.847 53.122
423737 96.337 -65.844 53.091
423738 96.382 -65.835 53.058
423739 96.425 -65.826 53.024
423740 96.465 -65.818 52.988
423741 96.506 -65.809 52.952
423742 96.547 -65.800 52.917
423743 96.588 -65.792 52.880
423744 96.624 -65.793 52.844
423745 96.654 -65.806 52.811
423746 96.685 -65.818 52.781
423747 96.715 -65.829 52.752
423748 96.748 -65.840 52.723
423749 96.788 -65.846 52.691
423750 96.828 -65.852 52.656
423751 96.869 -65.858 52.621
423752 96.909 -65.865 52.586
423753 96.948 -65.869 52.552
423754 96.987 -65.874 52.516
423755 97.026 -65.879 52.478
423756 97.063 -65.886 52.440
423757 97.100 -65.892 52.404
423758 97.137 -65.899 52.370
423759 97.174 -65.906 52.338
423760 97.211 -65.912 52.308
423761 97.247 -65.919 52.276
423762 97.282 -65.923 52.241
423763 97.319 -65.931 52.206
423764 97.355 -65.937 52.171
423765 97.391 -65.944 52.137
423766 97.427 -65.950 52.104
423767 97.463 -65.956 52.073
423768 97.499 -65.963 52.042
423769 97.535 -65.969 52.011
423770 97.571 -65.975 51.981
423771 97.607 -65.981 51.951
423772 97.643 -65.988 51.920
423773 97.680 -65.995 51.886
423774 97.717 -66.001 51.849
423775 97.750 -66.006 51.808
423776 97.783 -66.010 51.765
423777 97.816 -66.014 51.720
423778 97.849 -66.019 51.676
423779 97.882 -66.022 51.632
423780 97.917 -66.027 51.588
423781 97.951 -66.031 51.543
423782 97.986 -66.036 51.500
423783 98.020 -66.041 51.456
423784 98.055 -66.045 51.412
423785 98.089 -66.050 51.367
423786 98.124 -66.055 51.321
423787 98.158 -66.060 51.270
423788 98.196 -66.066 51.221
423789 98.258 -66.102 51.172
423790 98.293 -66.102 51.122
423791 98.327 -66.103 51.070
423792 98.360 -66.102 51.017
423793 98.393 -66.101 50.965
423794 98.426 -66.100 50.913
423795 98.459 -66.099 50.863
423796 98.492 -66.098 50.814
423797 98.525 -66.097 50.762
423798 98.565 -66.099 50.709
423799 98.608 -66.103 50.654
423800 98.651 -66.108 50.600
423801 98.694 -66.113 50.548
423802 98.736 -66.117 50.499
423803 98.779 -66.122 50.449
423804 98.822 -66.125 50.398
423805 98.865 -66.130 50.347
423806 98.906 -66.137 50.297
423807 98.945 -66.144 50.249
423808 98.984 -66.153 50.204
423809 99.023 -66.160 50.161
423810 99.061 -66.176 50.115
423811 99.098 -66.192 50.067
423812 99.136 -66.209 50.020
423813 99.173 -66.226 49.974
423814 99.213 -66.227 49.927
423815 99.252 -66.228 49.878
423816 99.291 -66.229 49.824
423817 99.331 -66.230 49.768
423818 99.370 -66.231 49.711
423819 99.410 -66.232 49.657
423820 99.449 -66.233 49.603
423821 99.490 -66.230 49.551
423822 99.524 -66.232 49.497
423823 99.563 -66.236 49.441
423824 99.602 -66.240 49.385
423825 99.640 -66.245 49.331
423826 99.679 -66.248 49.278
423827 99.718 -66.252 49.225
423828 99.757 -66.257 49.173
423829 99.796 -66.261 49.121
423830 99.836 -66.266 49.070
423831 99.876 -66.273 49.020
423832 99.917 -66.278 48.971
423833 99.956 -66.282 48.924
423834 100.003 -66.279 48.877
423835 100.051 -66.276 48.831
423836 100.099 -66.274 48.786
423837 100.147 -66.273 48.740
423838 100.196 -66.270 48.697
423839 100.244 -66.268 48.655
423840 100.292 -66.266 48.613
423841 100.340 -66.263 48.570
423842 100.388 -66.262 48.528
423843 100.437 -66.260 48.487
423844 100.485 -66.257 48.446
424255 88.380 -65.033 55.321
424256 88.414 -65.043 55.310
424257 88.449 -65.053 55.298
424258 88.483 -65.062 55.287
424259 88.517 -65.072 55.278
424260 88.551 -65.081 55.271
424261 88.586 -65.091 55.264
424262 88.620 -65.101 55.258
424263 88.654 -65.110 55.252
424264 88.688 -65.120 55.247
424265 88.731 -65.125 55.243
424266 88.777 -65.129 55.240
424267 88.823 -65.133 55.236
424268 88.869 -65.137 55.231
424269 88.915 -65.141 55.223
424270 88.962 -65.145 55.214
424271 89.008 -65.148 55.204
424272 89.054 -65.153 55.196
424273 89.100 -65.157 55.185
424274 89.146 -65.160 55.173
424275 89.192 -65.164 55.159
424276 89.239 -65.168 55.146
424277 89.284 -65.172 55.132
424278 89.331 -65.176 55.119
424279 89.377 -65.180 55.107
424280 89.423 -65.184 55.092
424281 89.477 -65.173 55.077
424282 89.533 -65.177 55.061
424283 89.588 -65.182 55.046
424284 89.636 -65.192 55.030
424285 89.685 -65.201 55.017
424286 89.733 -65.207 55.003
424287 89.781 -65.213 54.990
424288 89.830 -65.220 54.979
424289 89.879 -65.226 54.969
424299 90.363 -65.288 54.970
424300 90.411 -65.294 54.974
424301 90.460 -65.300 54.978
424302 90.508 -65.306 54.981
424303 90.557 -65.312 54.983
424304 90.605 -65.318 54.984
424305 90.654 -65.325 54.983
424306 90.702 -65.331 54.981
424307 90.750 -65.337 54.978
424308 90.799 -65.343 54.973
424309 90.848 -65.349 54.966
424310 90.896 -65.355 54.959
424311 90.942 -65.366 54.952
424312 90.988 -65.377 54.944
424313 91.034 -65.389 54.936
424314 91.079 -65.401 54.929
424315 91.124 -65.412 54.920
424316 91.167 -65.423 54.913
424317 91.210 -65.433 54.906
424318 91.253 -65.444 54.899
424319 91.296 -65.455 54.894
424320 91.339 -65.466 54.891
424321 91.382 -65.477 54.889
424322 91.425 -65.487 54.888
424323 91.468 -65.498 54.890
424324 91.511 -65.508 54.892
424325 91.552 -65.524 54.897
424326 91.591 -65.541 54.903
424327 91.631 -65.558 54.913
424328 91.671 -65.574 54.922
424329 91.711 -65.591 54.930
424330 91.750 -65.608 54.936
424331 91.788 -65.625 54.939
424332 91.815 -65.643 54.940
424333 91.844 -65.659 54.936
424334 91.896 -65.655 54.929
424335 91.947 -65.654 54.921
424336 91.998 -65.655 54.913
424337 92.049 -65.657 54.905
424338 92.099 -65.659 54.902
424339 92.149 -65.660 54.900
424340 92.200 -65.662 54.899
424341 92.250 -65.668 54.897
424342 92.302 -65.679 54.891
424343 92.353 -65.690 54.885
424344 92.404 -65.702 54.881
424345 92.454 -65.713 54.880
424346 92.506 -65.724 54.880
424347 92.557 -65.737 54.879
424348 92.607 -65.748 54.873
424349 92.658 -65.759 54.866
424350 92.708 -65.762 54.857
424351 92.757 -65.763 54.847
424352 92.804 -65.757 54.837
424353 92.853 -65.753 54.829
424354 92.902 -65.749 54.824
424355 92.951 -65.744 54.824
424356 93.000 -65.740 54.827
424357 93.049 -65.735 54.830
424358 93.097 -65.731 54.833
424359 93.146 -65.727 54.834
424360 93.195 -65.722 54.834
424361 93.243 -65.718 54.828
424362 93.292 -65.714 54.822
424363 93.340 -65.710 54.816
424364 93.384 -65.711 54.811
424365 93.428 -65.711 54.807
424366 93.472 -65.712 54.802
424367 93.515 -65.712 54.795
424368 93.559 -65.712 54.786
424369 93.603 -65.712 54.774
424370 93.647 -65.712 54.764
424371 93.690 -65.713 54.751
424372 93.735 -65.712 54.735
424373 93.780 -65.711 54.717
424374 93.826 -65.710 54.693
424375 93.870 -65.709 54.667
424376 93.916 -65.700 54.640
424377 93.963 -65.685 54.612
424378 94.008 -65.670 54.582
424379 94.054 -65.655 54.553
424380 94.101 -65.640 54.523
424381 94.146 -65.625 54.493
424382 94.174 -65.632 54.465
424383 94.193 -65.649 54.439
424384 94.211 -65.665 54.411
424385 94.229 -65.682 54.382
424386 94.248 -65.698 54.351
424387 94.266 -65.715 54.320
424388 94.285 -65.731 54.290
424389 94.321 -65.738 54.260
424390 94.358 -65.744 54.234
424391 94.398 -65.748 54.211
424392 94.437 -65.752 54.191
424393 94.477 -65.756 54.174
424394 94.517 -65.760 54.162
424395 94.557 -65.764 54.157
424396 94.596 -65.768 54.155
424397 94.636 -65.772 54.154
424398 94.676 -65.776 54.151
424399 94.715 -65.780 54.140
424400 94.755 -65.784 54.122
424401 94.795 -65.788 54.098
424402 94.835 -65.792 54.069
424403 94.877 -65.794 54.038
424404 94.920 -65.795 54.006
424405 94.963 -65.796 53.972
424406 95.006 -65.797 53.935
424407 95.048 -65.799 53.895
424408 95.089 -65.802 53.853
424409 95.130 -65.804 53.812
424410 95.171 -65.806 53.771
424411 95.213 -65.812 53.732
424412 95.258 -65.822 53.694
424413 95.301 -65.826 53.662
424414 95.340 -65.820 53.632
424415 95.382 -65.819 53.608
424416 95.426 -65.826 53.591
424417 95.470 -65.833 53.582
424418 95.515 -65.840 53.576
424419 95.559 -65.847 53.571
424420 95.603 -65.855 53.566
424421 95.647 -65.862 53.558
424422 95.691 -65.868 53.547
424423 95.735 -65.875 53.530
424424 95.774 -65.870 53.507
424425 95.813 -65.864 53.480
424426 95.852 -65.858 53.448
424427 95.891 -65.852 53.416
424428 95.930 -65.846 53.384
424429 95.969 -65.840 53.351
424430 96.008 -65.834 53.320
424431 96.048 -65.829 53.291
424432 96.090 -65.825 53.264
424433 96.131 -65.821 53.236
424434 96.173 -65.818 53.208
424435 96.215 -65.815 53.181
424436 96.256 -65.811 53.154
424437 96.299 -65.808 53.123
424438 96.343 -65.804 53.090
424439 96.387 -65.796 53.056
424440 96.433 -65.786 53.020
424441 96.477 -65.776 52.984
424442 96.518 -65.768 52.948
424443 96.559 -65.759 52.912
424444 96.600 -65.750 52.873
424445 96.636 -65.752 52.835
424446 96.666 -65.764 52.800
424447 96.696 -65.776 52.768
424448 96.727 -65.788 52.738
424449 96.757 -65.800 52.709
424450 96.791 -65.810 52.675
424451 96.831 -65.816 52.640
424452 96.871 -65.822 52.605
424453 96.911 -65.828 52.570
424454 96.950 -65.833 52.536
424455 96.989 -65.837 52.502
424456 97.028 -65.842 52.464
424457 97.066 -65.848 52.427
424458 97.102 -65.855 52.391
424459 97.139 -65.862 52.357
424460 97.176 -65.868 52.324
424461 97.213 -65.875 52.292
424462 97.249 -65.881 52.259
424463 97.280 -65.882 52.224
424464 97.321 -65.892 52.190
424465 97.357 -65.898 52.154
424466 97.393 -65.905 52.119
424467 97.429 -65.911 52.086
424468 97.464 -65.918 52.054
424469 97.500 -65.924 52.023
424470 97.536 -65.931 51.992
424471 97.572 -65.937 51.962
424472 97.607 -65.944 51.932
424473 97.643 -65.950 51.903
424474 97.679 -65.956 51.871
424475 97.714 -65.961 51.834
424476 97.747 -65.966 51.793
424477 97.782 -65.970 51.750
424478 97.816 -65.975 51.708
424479 97.851 -65.979 51.667
424480 97.885 -65.984 51.627
424481 97.920 -65.989 51.586
424482 97.954 -65.994 51.543
424483 97.989 -65.998 51.501
424484 98.023 -66.003 51.458
424485 98.058 -66.007 51.414
424486 98.092 -66.012 51.369
424487 98.127 -66.016 51.322
424488 98.161 -66.021 51.273
424489 98.214 -66.045 51.222
424490 98.263 -66.065 51.171
424491 98.300 -66.069 51.119
424492 98.333 -66.068 51.066
424493 98.366 -66.067 51.013
424494 98.398 -66.066 50.960
424495 98.432 -66.065 50.906
424496 98.464 -66.064 50.855
424497 98.498 -66.063 50.805
424498 98.530 -66.062 50.754
424499 98.568 -66.063 50.700
424500 98.613 -66.068 50.646
424501 98.656 -66.072 50.593
424502 98.699 -66.076 50.543
424503 98.742 -66.081 50.493
424504 98.785 -66.085 50.441
424505 98.828 -66.090 50.390
424506 98.870 -66.094 50.339
424507 98.911 -66.101 50.291
424508 98.950 -66.109 50.244
424509 98.989 -66.116 50.198
424510 99.028 -66.124 50.154
424511 99.067 -66.134 50.112
424512 99.104 -66.151 50.065
424513 99.142 -66.167 50.019
424514 99.181 -66.174 49.973
424515 99.220 -66.176 49.927
424516 99.260 -66.176 49.877
424517 99.299 -66.178 49.824
424518 99.339 -66.179 49.769
424519 99.378 -66.179 49.713
424520 99.417 -66.181 49.659
424521 99.457 -66.186 49.607
424522 99.497 -66.191 49.554
424523 99.539 -66.194 49.500
424524 99.578 -66.198 49.444
424525 99.617 -66.203 49.389
424526 99.656 -66.207 49.334
424527 99.695 -66.211 49.280
424528 99.734 -66.215 49.226
424529 99.773 -66.219 49.172
424530 99.811 -66.223 49.120
424531 99.851 -66.229 49.069
424532 99.892 -66.234 49.019
424533 99.930 -66.239 48.971
424534 99.970 -66.235 48.924
424535 100.017 -66.232 48.876
424536 100.065 -66.230 48.831
424537 100.114 -66.229 48.785
424538 100.162 -66.226 48.740
424539 100.210 -66.224 48.696
424540 100.258 -66.222 48.653
424541 100.307 -66.220 48.611
424542 100.355 -66.218 48.569
424543 100.403 -66.216 48.527
424544 100.451 -66.213 48.486
424545 100.499 -66.211 48.446
424956 88.372 -64.985 55.318
424957 88.406 -64.994 55.309
424958 88.440 -65.004 55.299
424959 88.475 -65.013 55.289
424960 88.509 -65.023 55.282
424961 88.543 -65.033 55.276
424962 88.577 -65.042 55.272
424963 88.612 -65.052 55.268
424964 88.646 -65.061 55.265
424965 88.687 -65.068 55.263
424966 88.733 -65.072 55.262
424967 88.779 -65.076 55.263
424968 88.825 -65.080 55.263
424969 88.872 -65.084 55.258
424970 88.917 -65.088 55.250
424971 88.964 -65.091 55.239
424972 89.010 -65.095 55.225
424973 89.056 -65.099 55.215
424974 89.102 -65.104 55.204
424975 89.148 -65.107 55.191
424976 89.195 -65.111 55.176
424977 89.241 -65.115 55.163
424978 89.287 -65.119 55.148
424979 89.333 -65.123 55.134
424980 89.379 -65.127 55.120
424981 89.428 -65.125 55.107
424982 89.484 -65.123 55.090
424983 89.538 -65.129 55.074
424984 89.587 -65.139 55.058
424985 89.635 -65.146 55.042
424986 89.683 -65.152 55.028
424987 89.732 -65.158 55.013
424988 89.780 -65.164 54.999
424989 89.829 -65.170 54.987
424990 89.877 -65.177 54.975
424991 89.926 -65.183 54.968
425000 90.362 -65.239 54.971
425001 90.410 -65.245 54.975
425002 90.459 -65.251 54.979
425003 90.507 -65.257 54.982
425004 90.555 -65.264 54.984
425005 90.604 -65.270 54.984
425006 90.653 -65.276 54.983
425007 90.701 -65.282 54.981
425008 90.749 -65.288 54.978
425009 90.798 -65.294 54.973
425010 90.846 -65.300 54.966
425011 90.894 -65.308 54.959
425012 90.940 -65.319 54.951
425013 90.986 -65.330 54.943
425014 91.031 -65.342 54.934
425015 91.077 -65.353 54.927
425016 91.120 -65.364 54.918
425017 91.163 -65.375 54.910
425018 91.207 -65.386 54.902
425019 91.249 -65.396 54.895
425020 91.292 -65.408 54.889
425021 91.335 -65.418 54.885
425022 91.378 -65.429 54.882
425023 91.421 -65.439 54.881
425024 91.464 -65.450 54.882
425025 91.504 -65.467 54.883
425026 91.544 -65.483 54.887
425027 91.584 -65.500 54.894
425028 91.624 -65.517 54.903
425029 91.663 -65.534 54.913
425030 91.703 -65.551 54.921
425031 91.743 -65.568 54.927
425032 91.783 -65.584 54.931
425033 91.816 -65.602 54.931
425034 91.843 -65.619 54.928
425035 91.888 -65.621 54.921
425036 91.940 -65.618 54.912
425037 91.992 -65.613 54.901
425038 92.044 -65.610 54.890
425039 92.094 -65.612 54.884
425040 92.145 -65.613 54.880
425041 92.196 -65.615 54.878
425042 92.246 -65.622 54.876
425043 92.297 -65.633 54.872
425044 92.348 -65.645 54.867
425045 92.399 -65.656 54.865
425046 92.450 -65.668 54.866
425047 92.501 -65.680 54.866
425048 92.552 -65.691 54.865
425049 92.603 -65.702 54.861
425050 92.654 -65.715 54.854
425051 92.706 -65.727 54.846
425052 92.757 -65.730 54.836
425053 92.806 -65.728 54.826
425054 92.855 -65.725 54.820
425055 92.903 -65.721 54.817
425056 92.952 -65.717 54.817
425057 93.001 -65.713 54.821
425058 93.050 -65.709 54.826
425059 93.098 -65.704 54.830
425060 93.147 -65.699 54.832
425061 93.196 -65.696 54.833
425062 93.245 -65.691 54.830
425063 93.294 -65.687 54.825
425064 93.340 -65.685 54.820
425065 93.383 -65.685 54.814
425066 93.427 -65.686 54.807
425067 93.471 -65.686 54.801
425068 93.515 -65.687 54.794
425069 93.558 -65.687 54.784
425070 93.602 -65.687 54.774
425071 93.646 -65.687 54.766
425072 93.690 -65.687 54.754
425073 93.733 -65.688 54.740
425074 93.778 -65.687 54.722
425075 93.824 -65.686 54.699
425076 93.869 -65.685 54.672
425077 93.914 -65.675 54.645
425078 93.961 -65.660 54.616
425079 94.006 -65.646 54.587
425080 94.053 -65.630 54.556
425081 94.099 -65.615 54.526
425082 94.120 -65.627 54.497
425083 94.142 -65.638 54.469
425084 94.163 -65.649 54.441
425085 94.185 -65.661 54.413
425086 94.206 -65.672 54.383
425087 94.227 -65.683 54.353
425088 94.249 -65.694 54.324
425089 94.282 -65.693 54.297
425090 94.321 -65.697 54.271
425091 94.361 -65.702 54.249
425092 94.401 -65.705 54.230
425093 94.440 -65.709 54.213
425094 94.480 -65.713 54.197
425095 94.520 -65.718 54.185
425096 94.560 -65.721 54.178
425097 94.599 -65.725 54.174
425098 94.639 -65.730 54.171
425099 94.679 -65.734 54.164
425100 94.718 -65.737 54.150
425101 94.758 -65.741 54.130
425102 94.798 -65.746 54.105
425103 94.838 -65.749 54.075
425104 94.881 -65.750 54.043
425105 94.924 -65.752 54.011
425106 94.967 -65.753 53.977
425107 95.011 -65.754 53.941
425108 95.053 -65.756 53.901
425109 95.094 -65.758 53.859
425110 95.134 -65.760 53.819
425111 95.176 -65.763 53.780
425112 95.220 -65.773 53.741
425113 95.265 -65.782 53.704
425114 95.308 -65.784 53.673
425115 95.347 -65.778 53.645
425116 95.387 -65.774 53.621
425117 95.431 -65.781 53.604
425118 95.475 -65.787 53.592
425119 95.520 -65.795 53.584
425120 95.564 -65.802 53.578
425121 95.608 -65.809 53.574
425122 95.652 -65.816 53.567
425123 95.696 -65.823 53.556
425124 95.737 -65.825 53.541
425125 95.777 -65.827 53.520
425126 95.817 -65.824 53.493
425127 95.856 -65.818 53.462
425128 95.895 -65.812 53.430
425129 95.934 -65.806 53.397
425130 95.973 -65.800 53.364
425131 96.012 -65.794 53.331
425132 96.052 -65.789 53.302
425133 96.094 -65.785 53.273
425134 96.135 -65.782 53.244
425135 96.177 -65.778 53.215
425136 96.219 -65.774 53.186
425137 96.261 -65.771 53.156
425138 96.305 -65.768 53.123
425139 96.348 -65.764 53.088
425140 96.393 -65.757 53.052
425141 96.438 -65.747 53.015
425142 96.483 -65.737 52.978
425143 96.529 -65.726 52.942
425144 96.571 -65.718 52.905
425145 96.612 -65.709 52.867
425146 96.647 -65.711 52.827
425147 96.677 -65.722 52.790
425148 96.708 -65.734 52.756
425149 96.738 -65.746 52.724
425150 96.768 -65.759 52.693
425151 96.799 -65.771 52.658
425152 96.833 -65.780 52.623
425153 96.874 -65.786 52.588
425154 96.913 -65.791 52.554
425155 96.952 -65.796 52.520
425156 96.991 -65.800 52.486
425157 97.030 -65.805 52.449
425158 97.068 -65.811 52.412
425159 97.105 -65.818 52.377
425160 97.142 -65.824 52.342
425161 97.179 -65.831 52.309
425162 97.216 -65.837 52.276
425163 97.244 -65.837 52.241
425164 97.278 -65.840 52.205
425165 97.320 -65.851 52.170
425166 97.359 -65.859 52.134
425167 97.394 -65.865 52.100
425168 97.430 -65.872 52.067
425169 97.466 -65.878 52.035
425170 97.502 -65.885 52.001
425171 97.537 -65.892 51.970
425172 97.573 -65.898 51.939
425173 97.609 -65.905 51.910
425174 97.645 -65.911 51.883
425175 97.681 -65.914 51.852
425176 97.717 -65.914 51.816
425177 97.751 -65.923 51.777
425178 97.785 -65.932 51.736
425179 97.819 -65.937 51.696
425180 97.854 -65.941 51.657
425181 97.888 -65.946 51.619
425182 97.923 -65.950 51.581
425183 97.957 -65.955 51.540
425184 97.992 -65.959 51.499
425185 98.026 -65.964 51.457
425186 98.061 -65.969 51.413
425187 98.095 -65.973 51.369
425188 98.130 -65.978 51.322
425189 98.169 -65.988 51.272
425190 98.222 -66.012 51.221
425191 98.262 -66.022 51.168
425192 98.300 -66.028 51.115
425193 98.337 -66.031 51.062
425194 98.371 -66.032 51.009
425195 98.404 -66.031 50.955
425196 98.437 -66.030 50.902
425197 98.470 -66.029 50.850
425198 98.503 -66.028 50.799
425199 98.536 -66.027 50.747
425200 98.571 -66.027 50.696
425201 98.616 -66.031 50.644
425202 98.660 -66.036 50.592
425203 98.704 -66.041 50.541
425204 98.747 -66.045 50.489
425205 98.790 -66.050 50.437
425206 98.833 -66.053 50.385
425207 98.875 -66.058 50.333
425208 98.916 -66.065 50.284
425209 98.955 -66.072 50.238
425210 98.994 -66.081 50.192
425211 99.033 -66.088 50.148
425212 99.073 -66.096 50.107
425213 99.111 -66.109 50.061
425214 99.149 -66.122 50.015
425215 99.188 -66.122 49.969
425216 99.228 -66.124 49.922
425217 99.267 -66.125 49.873
425218 99.307 -66.130 49.821
425219 99.347 -66.135 49.768
425220 99.386 -66.140 49.715
425221 99.426 -66.145 49.661
425222 99.466 -66.150 49.608
425223 99.506 -66.154 49.554
425224 99.549 -66.157 49.500
425225 99.591 -66.160 49.445
425226 99.631 -66.164 49.392
425227 99.670 -66.168 49.338
425228 99.709 -66.172 49.283
425229 99.749 -66.177 49.228
425230 99.788 -66.181 49.174
425231 99.827 -66.185 49.121
425232 99.866 -66.190 49.069
425233 99.905 -66.195 49.019
425234 99.944 -66.201 48.971
425235 99.983 -66.188 48.923
425236 100.031 -66.186 48.876
425237 100.080 -66.185 48.830
425238 100.128 -66.182 48.784
425239 100.176 -66.180 48.739
425240 100.224 -66.178 48.695
425241 100.273 -66.176 48.653
425242 100.321 -66.174 48.610
425243 100.369 -66.172 48.568
425244 100.417 -66.169 48.526
425245 100.466 -66.168 48.484
425246 100.514 -66.166 48.445
425656 88.330 -64.927 55.325
425657 88.364 -64.937 55.315
425658 88.399 -64.946 55.307
425659 88.433 -64.956 55.299
425660 88.467 -64.965 55.291
425661 88.501 -64.975 55.286
425662 88.536 -64.985 55.283
425663 88.570 -64.994 55.280
425664 88.604 -65.004 55.279
425665 88.643 -65.011 55.279
425666 88.689 -65.015 55.280
425667 88.735 -65.019 55.284
425668 88.781 -65.022 55.293
425669 88.828 -65.026 55.295
425670 88.874 -65.030 55.292
425671 88.920 -65.035 55.282
425672 88.966 -65.038 55.268
425673 89.012 -65.042 55.249
425674 89.058 -65.046 55.235
425675 89.105 -65.050 55.222
425676 89.151 -65.054 55.209
425677 89.197 -65.058 55.194
425678 89.243 -65.062 55.180
425679 89.290 -65.066 55.165
425680 89.335 -65.069 55.150
425681 89.382 -65.073 55.135
425682 89.436 -65.069 55.121
425683 89.488 -65.076 55.105
425684 89.537 -65.085 55.088
425685 89.586 -65.091 55.071
425686 89.634 -65.097 55.054
425687 89.682 -65.103 55.037
425688 89.731 -65.110 55.022
425689 89.779 -65.116 55.009
425690 89.828 -65.122 54.995
425691 89.876 -65.128 54.983
425692 89.924 -65.134 54.974
425702 90.409 -65.196 54.975
425703 90.457 -65.202 54.979
425704 90.506 -65.208 54.982
425705 90.554 -65.214 54.984
425706 90.603 -65.220 54.984
425707 90.651 -65.226 54.983
425708 90.699 -65.232 54.981
425709 90.748 -65.239 54.978
425710 90.797 -65.245 54.973
425711 90.845 -65.251 54.966
425712 90.892 -65.261 54.958
425713 90.938 -65.272 54.950
425714 90.983 -65.283 54.942
425715 91.029 -65.295 54.933
425716 91.074 -65.306 54.925
425717 91.116 -65.317 54.916
425718 91.160 -65.327 54.907
425719 91.203 -65.338 54.899
425720 91.245 -65.349 54.890
425721 91.288 -65.360 54.884
425722 91.331 -65.370 54.879
425723 91.374 -65.381 54.876
425724 91.417 -65.392 54.874
425725 91.456 -65.410 54.874
425726 91.496 -65.427 54.875
425727 91.536 -65.443 54.879
425728 91.576 -65.460 54.884
425729 91.615 -65.477 54.892
425730 91.655 -65.494 54.901
425731 91.695 -65.511 54.910
425732 91.735 -65.527 54.916
425733 91.775 -65.544 54.919
425734 91.814 -65.561 54.919
425735 91.844 -65.578 54.916
425736 91.880 -65.588 54.911
425737 91.932 -65.584 54.902
425738 91.984 -65.580 54.890
425739 92.036 -65.577 54.878
425740 92.088 -65.572 54.869
425741 92.140 -65.568 54.863
425742 92.191 -65.568 54.859
425743 92.242 -65.577 54.857
425744 92.293 -65.589 54.854
425745 92.345 -65.602 54.851
425746 92.397 -65.615 54.851
425747 92.448 -65.628 54.853
425748 92.500 -65.641 54.853
425749 92.552 -65.654 54.851
425750 92.603 -65.668 54.847
425751 92.655 -65.674 54.842
425752 92.706 -65.676 54.833
425753 92.757 -65.679 54.824
425754 92.808 -65.680 54.816
425755 92.858 -65.682 54.812
425756 92.909 -65.684 54.811
425757 92.957 -65.680 54.810
425758 93.005 -65.677 54.814
425759 93.054 -65.674 54.817
425760 93.102 -65.671 54.820
425761 93.150 -65.668 54.823
425762 93.199 -65.664 54.825
425763 93.248 -65.661 54.823
425764 93.296 -65.658 54.820
425765 93.339 -65.660 54.815
425766 93.382 -65.660 54.809
425767 93.426 -65.661 54.802
425768 93.470 -65.661 54.795
425769 93.514 -65.661 54.787
425770 93.558 -65.662 54.778
425771 93.602 -65.662 54.769
425772 93.645 -65.662 54.761
425773 93.689 -65.662 54.752
425774 93.733 -65.663 54.739
425775 93.777 -65.663 54.721
425776 93.822 -65.662 54.698
425777 93.867 -65.661 54.673
425778 93.913 -65.650 54.645
425779 93.959 -65.636 54.616
425780 94.005 -65.621 54.588
425781 94.041 -65.616 54.558
425782 94.063 -65.627 54.526
425783 94.084 -65.638 54.497
425784 94.106 -65.649 54.468
425785 94.127 -65.661 54.438
425786 94.152 -65.667 54.409
425787 94.185 -65.663 54.380
425788 94.218 -65.659 54.352
425789 94.252 -65.654 54.326
425790 94.286 -65.654 54.303
425791 94.325 -65.655 54.282
425792 94.364 -65.659 54.265
425793 94.404 -65.663 54.250
425794 94.443 -65.667 54.235
425795 94.483 -65.671 54.220
425796 94.523 -65.675 54.208
425797 94.563 -65.679 54.199
425798 94.602 -65.683 54.192
425799 94.642 -65.687 54.185
425800 94.682 -65.691 54.173
425801 94.721 -65.695 54.157
425802 94.761 -65.699 54.136
425803 94.801 -65.703 54.110
425804 94.842 -65.705 54.079
425805 94.885 -65.707 54.046
425806 94.928 -65.709 54.014
425807 94.971 -65.709 53.981
425808 95.015 -65.711 53.945
425809 95.057 -65.712 53.906
425810 95.099 -65.714 53.866
425811 95.139 -65.716 53.826
425812 95.183 -65.724 53.788
425813 95.228 -65.734 53.751
425814 95.272 -65.743 53.716
425815 95.314 -65.743 53.686
425816 95.353 -65.737 53.659
425817 95.393 -65.731 53.638
425818 95.436 -65.735 53.622
425819 95.480 -65.742 53.609
425820 95.524 -65.749 53.598
425821 95.568 -65.756 53.587
425822 95.613 -65.763 53.579
425823 95.657 -65.771 53.572
425824 95.699 -65.774 53.562
425825 95.740 -65.776 53.547
425826 95.780 -65.778 53.528
425827 95.820 -65.780 53.503
425828 95.860 -65.778 53.475
425829 95.899 -65.772 53.442
425830 95.938 -65.766 53.409
425831 95.977 -65.760 53.375
425832 96.016 -65.754 53.344
425833 96.056 -65.749 53.312
425834 96.098 -65.745 53.281
425835 96.139 -65.742 53.250
425836 96.181 -65.738 53.219
425837 96.224 -65.734 53.190
425838 96.267 -65.731 53.157
425839 96.311 -65.727 53.122
425840 96.355 -65.724 53.085
425841 96.398 -65.718 53.047
425842 96.444 -65.708 53.008
425843 96.489 -65.697 52.971
425844 96.534 -65.687 52.935
425845 96.580 -65.677 52.899
425846 96.623 -65.667 52.861
425847 96.659 -65.669 52.821
425848 96.689 -65.681 52.781
425849 96.720 -65.693 52.744
425850 96.750 -65.705 52.709
425851 96.780 -65.717 52.676
425852 96.810 -65.729 52.641
425853 96.840 -65.741 52.607
425854 96.876 -65.749 52.572
425855 96.915 -65.754 52.537
425856 96.954 -65.759 52.503
425857 96.993 -65.764 52.470
425858 97.032 -65.768 52.433
425859 97.071 -65.774 52.396
425860 97.107 -65.780 52.361
425861 97.144 -65.787 52.327
425862 97.181 -65.793 52.294
425863 97.217 -65.799 52.259
425864 97.240 -65.794 52.222
425865 97.276 -65.800 52.185
425866 97.318 -65.810 52.149
425867 97.361 -65.820 52.114
425868 97.396 -65.827 52.081
425869 97.432 -65.833 52.048
425870 97.468 -65.840 52.015
425871 97.504 -65.847 51.980
425872 97.539 -65.853 51.947
425873 97.575 -65.859 51.916
425874 97.611 -65.865 51.887
425875 97.647 -65.872 51.860
425876 97.683 -65.868 51.830
425877 97.720 -65.864 51.796
425878 97.755 -65.865 51.759
425879 97.790 -65.875 51.721
425880 97.823 -65.884 51.683
425881 97.858 -65.894 51.646
425882 97.892 -65.903 51.610
425883 97.926 -65.912 51.572
425884 97.960 -65.916 51.534
425885 97.995 -65.921 51.494
425886 98.029 -65.925 51.453
425887 98.064 -65.930 51.411
425888 98.098 -65.934 51.367
425889 98.133 -65.939 51.320
425890 98.181 -65.959 51.270
425891 98.221 -65.969 51.218
425892 98.261 -65.979 51.165
425893 98.299 -65.985 51.111
425894 98.337 -65.990 51.058
425895 98.374 -65.994 51.005
425896 98.410 -65.996 50.952
425897 98.442 -65.995 50.900
425898 98.475 -65.994 50.848
425899 98.508 -65.993 50.796
425900 98.541 -65.992 50.745
425901 98.574 -65.991 50.694
425902 98.619 -65.995 50.643
425903 98.663 -66.000 50.592
425904 98.708 -66.004 50.540
425905 98.752 -66.009 50.487
425906 98.795 -66.013 50.435
425907 98.838 -66.018 50.382
425908 98.880 -66.022 50.329
425909 98.921 -66.029 50.279
425910 98.960 -66.037 50.231
425911 98.999 -66.044 50.186
425912 99.039 -66.053 50.142
425913 99.078 -66.060 50.099
425914 99.117 -66.068 50.055
425915 99.157 -66.073 50.009
425916 99.196 -66.078 49.963
425917 99.236 -66.084 49.915
425918 99.276 -66.088 49.865
425919 99.315 -66.094 49.816
425920 99.355 -66.099 49.766
425921 99.395 -66.103 49.714
425922 99.435 -66.109 49.661
425923 99.475 -66.114 49.607
425924 99.515 -66.118 49.554
425925 99.558 -66.121 49.500
425926 99.600 -66.123 49.446
425927 99.643 -66.126 49.394
425928 99.684 -66.129 49.340
425929 99.723 -66.134 49.285
425930 99.762 -66.138 49.229
425931 99.801 -66.142 49.176
425932 99.840 -66.147 49.122
425933 99.879 -66.151 49.069
425934 99.919 -66.157 49.019
425935 99.958 -66.154 48.970
425936 99.998 -66.142 48.923
425937 100.046 -66.141 48.876
425938 100.094 -66.138 48.829
425939 100.142 -66.136 48.783
425940 100.191 -66.134 48.738
425941 100.239 -66.132 48.695
425942 100.287 -66.130 48.652
425943 100.335 -66.128 48.610
425944 100.384 -66.125 48.568
425945 100.432 -66.124 48.525
425946 100.480 -66.122 48.482
425947 100.528 -66.119 48.443
426357 88.322 -64.878 55.321
426358 88.356 -64.888 55.312
426359 88.391 -64.897 55.305
426360 88.425 -64.907 55.298
426361 88.459 -64.917 55.293
426362 88.493 -64.926 55.290
426363 88.528 -64.936 55.288
426364 88.562 -64.945 55.287
426365 88.599 -64.953 55.288
426366 88.645 -64.957 55.291
426367 88.691 -64.962 55.297
426368 88.738 -64.966 55.307
426369 88.784 -64.969 55.322
426370 88.830 -64.973 55.327
426371 88.876 -64.977 55.323
426372 88.922 -64.981 55.313
426373 88.968 -64.985 55.297
426374 89.015 -64.989 55.276
426375 89.061 -64.993 55.257
426376 89.107 -64.997 55.241
426377 89.153 -65.000 55.227
426378 89.199 -65.004 55.214
426379 89.245 -65.009 55.198
426380 89.292 -65.013 55.183
426381 89.338 -65.016 55.167
426382 89.387 -65.016 55.151
426383 89.439 -65.023 55.135
426384 89.487 -65.029 55.119
426385 89.536 -65.035 55.102
426386 89.584 -65.041 55.084
426387 89.633 -65.048 55.066
426388 89.681 -65.054 55.046
426389 89.730 -65.060 55.031
426390 89.778 -65.066 55.017
426391 89.826 -65.072 55.004
426392 89.875 -65.079 54.990
426393 89.923 -65.085 54.979
426394 89.972 -65.091 54.972
426404 90.456 -65.153 54.979
426405 90.505 -65.159 54.982
426406 90.553 -65.165 54.982
426407 90.602 -65.172 54.983
426408 90.650 -65.178 54.982
426409 90.698 -65.184 54.980
426410 90.747 -65.190 54.978
426411 90.796 -65.196 54.972
426412 90.844 -65.202 54.966
426413 90.890 -65.213 54.958
426414 90.936 -65.225 54.950
426415 90.981 -65.236 54.941
426416 91.027 -65.248 54.932
426417 91.070 -65.258 54.923
426418 91.113 -65.269 54.914
426419 91.156 -65.280 54.905
426420 91.199 -65.291 54.896
426421 91.241 -65.301 54.886
426422 91.285 -65.312 54.879
426423 91.327 -65.323 54.874
426424 91.369 -65.336 54.869
426425 91.409 -65.352 54.866
426426 91.449 -65.369 54.865
426427 91.489 -65.386 54.866
426428 91.528 -65.403 54.869
426429 91.568 -65.420 54.874
426430 91.608 -65.436 54.880
426431 91.648 -65.453 54.888
426432 91.687 -65.470 54.896
426433 91.727 -65.487 54.902
426434 91.767 -65.504 54.906
426435 91.807 -65.521 54.906
426436 91.844 -65.538 54.903
426437 91.871 -65.555 54.897
426438 91.923 -65.551 54.889
426439 91.976 -65.547 54.876
426440 92.027 -65.543 54.865
426441 92.079 -65.539 54.856
426442 92.132 -65.535 54.848
426443 92.187 -65.533 54.842
426444 92.241 -65.541 54.841
426445 92.293 -65.554 54.839
426446 92.345 -65.567 54.838
426447 92.397 -65.580 54.838
426448 92.448 -65.593 54.839
426449 92.500 -65.607 54.840
426450 92.552 -65.617 54.838
426451 92.603 -65.620 54.833
426452 92.654 -65.622 54.827
426453 92.706 -65.625 54.820
426454 92.757 -65.628 54.812
426455 92.808 -65.630 54.807
426456 92.858 -65.631 54.805
426457 92.909 -65.633 54.805
426458 92.959 -65.635 54.805
426459 93.010 -65.637 54.806
426460 93.059 -65.636 54.807
426461 93.108 -65.633 54.809
426462 93.156 -65.630 54.811
426463 93.204 -65.626 54.813
426464 93.253 -65.623 54.812
426465 93.300 -65.621 54.809
426466 93.343 -65.623 54.804
426467 93.386 -65.626 54.799
426468 93.429 -65.628 54.790
426469 93.472 -65.631 54.783
426470 93.514 -65.633 54.775
426471 93.557 -65.637 54.766
426472 93.601 -65.637 54.758
426473 93.645 -65.637 54.753
426474 93.688 -65.637 54.745
426475 93.732 -65.637 54.733
426476 93.776 -65.638 54.716
426477 93.820 -65.638 54.693
426478 93.865 -65.637 54.667
426479 93.911 -65.626 54.640
426480 93.957 -65.611 54.612
426481 93.984 -65.616 54.582
426482 94.005 -65.627 54.551
426483 94.027 -65.639 54.520
426484 94.057 -65.638 54.490
426485 94.090 -65.633 54.459
426486 94.123 -65.629 54.430
426487 94.156 -65.624 54.402
426488 94.189 -65.620 54.376
426489 94.222 -65.615 54.350
426490 94.256 -65.613 54.327
426491 94.290 -65.615 54.309
426492 94.328 -65.616 54.295
426493 94.367 -65.618 54.282
426494 94.407 -65.621 54.270
426495 94.447 -65.624 54.257
426496 94.486 -65.628 54.244
426497 94.526 -65.632 54.231
426498 94.566 -65.637 54.219
426499 94.606 -65.640 54.208
426500 94.645 -65.644 54.195
426501 94.685 -65.649 54.180
426502 94.724 -65.652 54.162
426503 94.764 -65.656 54.140
426504 94.804 -65.660 54.112
426505 94.846 -65.662 54.081
426506 94.889 -65.664 54.049
426507 94.932 -65.665 54.017
426508 94.975 -65.666 53.984
426509 95.019 -65.667 53.949
426510 95.062 -65.668 53.912
426511 95.103 -65.670 53.873
426512 95.145 -65.674 53.835
426513 95.190 -65.684 53.797
426514 95.235 -65.693 53.761
426515 95.279 -65.703 53.730
426516 95.321 -65.702 53.700
426517 95.360 -65.696 53.676
426518 95.399 -65.689 53.657
426519 95.441 -65.690 53.641
426520 95.485 -65.696 53.629
426521 95.529 -65.703 53.614
426522 95.573 -65.711 53.599
426523 95.618 -65.718 53.586
426524 95.662 -65.724 53.576
426525 95.702 -65.726 53.566
426526 95.742 -65.727 53.553
426527 95.782 -65.729 53.535
426528 95.823 -65.731 53.511
426529 95.863 -65.732 53.483
426530 95.903 -65.732 53.452
426531 95.942 -65.726 53.419
426532 95.981 -65.720 53.386
426533 96.020 -65.714 53.354
426534 96.060 -65.709 53.321
426535 96.102 -65.705 53.288
426536 96.144 -65.702 53.256
426537 96.186 -65.698 53.224
426538 96.230 -65.694 53.193
426539 96.273 -65.691 53.159
426540 96.317 -65.687 53.122
426541 96.360 -65.684 53.083
426542 96.404 -65.679 53.043
426543 96.449 -65.668 53.003
426544 96.495 -65.659 52.964
426545 96.540 -65.648 52.927
426546 96.585 -65.638 52.891
426547 96.630 -65.628 52.854
426548 96.671 -65.629 52.815
426549 96.701 -65.641 52.776
426550 96.732 -65.653 52.737
426551 96.762 -65.665 52.699
426552 96.793 -65.677 52.663
426553 96.823 -65.690 52.626
426554 96.853 -65.702 52.590
426555 96.884 -65.710 52.555
426556 96.917 -65.718 52.519
426557 96.956 -65.722 52.485
426558 96.995 -65.727 52.452
426559 97.034 -65.731 52.416
426560 97.073 -65.736 52.380
426561 97.110 -65.743 52.345
426562 97.147 -65.749 52.312
426563 97.184 -65.756 52.277
426564 97.212 -65.756 52.241
426565 97.235 -65.750 52.203
426566 97.274 -65.758 52.166
426567 97.316 -65.768 52.129
426568 97.359 -65.779 52.096
426569 97.398 -65.787 52.063
426570 97.434 -65.794 52.030
426571 97.469 -65.800 51.995
426572 97.505 -65.807 51.960
426573 97.541 -65.814 51.925
426574 97.577 -65.820 51.893
426575 97.612 -65.827 51.863
426576 97.649 -65.828 51.834
426577 97.686 -65.823 51.805
426578 97.723 -65.818 51.772
426579 97.759 -65.813 51.738
426580 97.794 -65.818 51.701
426581 97.828 -65.827 51.666
426582 97.862 -65.837 51.631
426583 97.896 -65.846 51.596
426584 97.930 -65.855 51.559
426585 97.964 -65.865 51.521
426586 97.999 -65.874 51.485
426587 98.032 -65.883 51.447
426588 98.067 -65.892 51.406
426589 98.101 -65.897 51.363
426590 98.140 -65.906 51.317
426591 98.180 -65.916 51.266
426592 98.220 -65.926 51.214
426593 98.260 -65.937 51.162
426594 98.298 -65.944 51.108
426595 98.336 -65.947 51.055
426596 98.373 -65.952 51.003
426597 98.411 -65.956 50.951
426598 98.448 -65.960 50.901
426599 98.481 -65.959 50.851
426600 98.513 -65.958 50.800
426601 98.547 -65.957 50.747
426602 98.579 -65.956 50.696
426603 98.622 -65.959 50.645
426604 98.666 -65.964 50.594
426605 98.711 -65.969 50.541
426606 98.756 -65.973 50.487
426607 98.800 -65.978 50.433
426608 98.843 -65.982 50.379
426609 98.885 -65.986 50.325
426610 98.926 -65.993 50.274
426611 98.965 -66.001 50.225
426612 99.004 -66.009 50.178
426613 99.044 -66.016 50.133
426614 99.082 -66.024 50.089
426615 99.121 -66.030 50.045
426616 99.163 -66.036 49.999
426617 99.205 -66.042 49.952
426618 99.245 -66.047 49.903
426619 99.285 -66.052 49.854
426620 99.325 -66.057 49.807
426621 99.364 -66.063 49.759
426622 99.404 -66.067 49.710
426623 99.443 -66.072 49.658
426624 99.483 -66.078 49.605
426625 99.524 -66.082 49.552
426626 99.567 -66.084 49.499
426627 99.610 -66.087 49.447
426628 99.652 -66.089 49.394
426629 99.695 -66.092 49.340
426630 99.736 -66.095 49.284
426631 99.776 -66.099 49.230
426632 99.815 -66.103 49.175
426633 99.854 -66.108 49.122
426634 99.893 -66.113 49.069
426635 99.932 -66.118 49.018
426636 99.971 -66.107 48.970
426637 100.012 -66.097 48.922
426638 100.060 -66.094 48.875
426639 100.109 -66.092 48.828
426640 100.157 -66.090 48.782
426641 100.205 -66.088 48.737
426642 100.253 -66.086 48.693
426643 100.301 -66.084 48.650
426644 100.350 -66.082 48.608
426645 100.398 -66.080 48.565
426646 100.446 -66.078 48.523
426647 100.494 -66.075 48.481
426648 100.542 -66.074 48.442
427058 88.314 -64.830 55.318
427059 88.348 -64.840 55.311
427060 88.383 -64.849 55.304
427061 88.417 -64.859 55.298
427062 88.451 -64.869 55.294
427063 88.485 -64.878 55.293
427064 88.520 -64.888 55.293
427065 88.555 -64.897 55.294
427066 88.601 -64.900 55.297
427067 88.647 -64.904 55.303
427068 88.694 -64.908 55.312
427069 88.740 -64.912 55.327
427070 88.786 -64.916 55.347
427071 88.832 -64.920 55.353
427072 88.878 -64.924 55.349
427073 88.924 -64.928 55.337
427074 88.971 -64.931 55.320
427075 89.017 -64.935 55.298
427076 89.063 -64.940 55.279
427077 89.109 -64.944 55.262
427078 89.156 -64.947 55.247
427079 89.201 -64.951 55.232
427080 89.248 -64.955 55.216
427081 89.294 -64.959 55.202
427082 89.340 -64.963 55.188
427083 89.389 -64.968 55.171
427084 89.438 -64.974 55.151
427085 89.486 -64.981 55.134
427086 89.535 -64.987 55.116
427087 89.583 -64.993 55.097
427088 89.631 -64.999 55.078
427089 89.680 -65.005 55.058
427090 89.729 -65.011 55.040
427091 89.777 -65.017 55.025
427092 89.825 -65.024 55.011
427093 89.874 -65.030 54.997
427094 89.922 -65.036 54.985
427095 89.971 -65.042 54.977
427107 90.552 -65.116 54.982
427108 90.600 -65.123 54.983
427109 90.649 -65.129 54.981
427110 90.697 -65.135 54.980
427111 90.746 -65.141 54.977
427112 90.794 -65.147 54.971
427113 90.842 -65.154 54.965
427114 90.888 -65.166 54.957
427115 90.934 -65.178 54.949
427116 90.979 -65.189 54.940
427117 91.023 -65.200 54.931
427118 91.066 -65.210 54.922
427119 91.109 -65.222 54.913
427120 91.152 -65.232 54.903
427121 91.194 -65.243 54.892
427122 91.238 -65.254 54.883
427123 91.281 -65.264 54.875
427124 91.322 -65.279 54.868
427125 91.362 -65.295 54.863
427126 91.401 -65.312 54.859
427127 91.441 -65.329 54.857
427128 91.481 -65.345 54.857
427129 91.521 -65.363 54.858
427130 91.561 -65.380 54.862
427131 91.600 -65.396 54.866
427132 91.640 -65.413 54.874
427133 91.680 -65.430 54.881
427134 91.720 -65.447 54.887
427135 91.759 -65.464 54.893
427136 91.799 -65.480 54.893
427137 91.839 -65.497 54.890
427138 91.871 -65.514 54.884
427139 91.915 -65.517 54.875
427140 91.967 -65.514 54.863
427141 92.019 -65.510 54.852
427142 92.071 -65.506 54.843
427143 92.126 -65.504 54.836
427144 92.184 -65.505 54.830
427145 92.242 -65.506 54.828
427146 92.294 -65.520 54.826
427147 92.345 -65.533 54.824
427148 92.397 -65.546 54.825
427149 92.449 -65.559 54.825
427150 92.500 -65.563 54.825
427151 92.552 -65.566 54.822
427152 92.603 -65.568 54.818
427153 92.654 -65.571 54.814
427154 92.706 -65.574 54.809
427155 92.757 -65.577 54.805
427156 92.808 -65.579 54.803
427157 92.858 -65.580 54.802
427158 92.909 -65.582 54.802
427159 92.959 -65.583 54.801
427160 93.010 -65.586 54.799
427161 93.060 -65.587 54.797
427162 93.111 -65.589 54.795
427163 93.161 -65.591 54.795
427164 93.210 -65.588 54.797
427165 93.258 -65.585 54.796
427166 93.304 -65.583 54.793
427167 93.347 -65.586 54.788
427168 93.390 -65.589 54.782
427169 93.433 -65.592 54.774
427170 93.476 -65.594 54.766
427171 93.518 -65.597 54.759
427172 93.561 -65.599 54.752
427173 93.604 -65.602 54.746
427174 93.647 -65.605 54.743
427175 93.690 -65.608 54.736
427176 93.732 -65.610 54.725
427177 93.775 -65.613 54.708
427178 93.819 -65.613 54.685
427179 93.863 -65.613 54.660
427180 93.905 -65.605 54.633
427181 93.929 -65.613 54.604
427182 93.962 -65.608 54.575
427183 93.995 -65.605 54.544
427184 94.028 -65.600 54.514
427185 94.061 -65.596 54.483
427186 94.094 -65.591 54.453
427187 94.127 -65.586 54.424
427188 94.160 -65.582 54.398
427189 94.193 -65.577 54.375
427190 94.226 -65.573 54.354
427191 94.260 -65.574 54.334
427192 94.295 -65.575 54.318
427193 94.331 -65.577 54.307
427194 94.370 -65.578 54.298
427195 94.410 -65.580 54.288
427196 94.450 -65.582 54.277
427197 94.489 -65.586 54.265
427198 94.529 -65.590 54.253
427199 94.569 -65.594 54.240
427200 94.609 -65.598 54.225
427201 94.648 -65.602 54.207
427202 94.688 -65.605 54.187
427203 94.728 -65.610 54.164
427204 94.767 -65.614 54.139
427205 94.807 -65.618 54.112
427206 94.850 -65.619 54.081
427207 94.893 -65.620 54.051
427208 94.936 -65.621 54.021
427209 94.979 -65.622 53.990
427210 95.023 -65.624 53.958
427211 95.066 -65.624 53.922
427212 95.108 -65.627 53.885
427213 95.153 -65.635 53.846
427214 95.197 -65.644 53.809
427215 95.242 -65.654 53.776
427216 95.286 -65.663 53.744
427217 95.327 -65.660 53.716
427218 95.367 -65.654 53.693
427219 95.406 -65.648 53.677
427220 95.446 -65.643 53.663
427221 95.490 -65.651 53.649
427222 95.534 -65.658 53.633
427223 95.579 -65.665 53.614
427224 95.623 -65.672 53.596
427225 95.664 -65.675 53.581
427226 95.704 -65.677 53.570
427227 95.745 -65.678 53.557
427228 95.785 -65.680 53.540
427229 95.825 -65.681 53.517
427230 95.866 -65.683 53.489
427231 95.906 -65.685 53.459
427232 95.946 -65.686 53.428
427233 95.985 -65.680 53.396
427234 96.024 -65.674 53.364
427235 96.064 -65.669 53.329
427236 96.106 -65.665 53.294
427237 96.148 -65.662 53.260
427238 96.192 -65.658 53.227
427239 96.235 -65.654 53.194
427240 96.279 -65.651 53.159
427241 96.322 -65.647 53.121
427242 96.366 -65.643 53.081
427243 96.410 -65.640 53.041
427244 96.455 -65.630 52.999
427245 96.500 -65.619 52.960
427246 96.545 -65.609 52.921
427247 96.590 -65.600 52.883
427248 96.632 -65.601 52.846
427249 96.674 -65.601 52.808
427250 96.715 -65.602 52.770
427251 96.745 -65.614 52.730
427252 96.775 -65.626 52.691
427253 96.806 -65.638 52.653
427254 96.836 -65.650 52.613
427255 96.866 -65.661 52.576
427256 96.897 -65.668 52.540
427257 96.928 -65.677 52.504
427258 96.958 -65.685 52.469
427259 96.997 -65.690 52.435
427260 97.036 -65.694 52.399
427261 97.075 -65.699 52.363
427262 97.112 -65.705 52.329
427263 97.150 -65.712 52.296
427264 97.184 -65.716 52.261
427265 97.207 -65.712 52.225
427266 97.230 -65.707 52.186
427267 97.272 -65.717 52.147
427268 97.314 -65.727 52.110
427269 97.357 -65.738 52.077
427270 97.399 -65.748 52.044
427271 97.436 -65.755 52.011
427272 97.471 -65.762 51.976
427273 97.507 -65.768 51.940
427274 97.543 -65.774 51.905
427275 97.579 -65.781 51.872
427276 97.614 -65.787 51.840
427277 97.651 -65.781 51.810
427278 97.688 -65.777 51.779
427279 97.725 -65.772 51.747
427280 97.761 -65.768 51.713
427281 97.799 -65.762 51.679
427282 97.833 -65.769 51.645
427283 97.867 -65.779 51.611
427284 97.901 -65.788 51.576
427285 97.935 -65.798 51.540
427286 97.969 -65.807 51.503
427287 98.003 -65.816 51.469
427288 98.037 -65.826 51.433
427289 98.071 -65.840 51.394
427290 98.103 -65.857 51.353
427291 98.139 -65.862 51.309
427292 98.179 -65.873 51.259
427293 98.219 -65.884 51.208
427294 98.259 -65.894 51.157
427295 98.298 -65.902 51.104
427296 98.335 -65.906 51.052
427297 98.372 -65.910 51.002
427298 98.410 -65.914 50.951
427299 98.447 -65.919 50.902
427300 98.485 -65.922 50.853
427301 98.519 -65.923 50.803
427302 98.552 -65.922 50.751
427303 98.585 -65.921 50.701
427304 98.625 -65.923 50.651
427305 98.669 -65.928 50.598
427306 98.714 -65.932 50.544
427307 98.759 -65.937 50.488
427308 98.803 -65.941 50.432
427309 98.848 -65.946 50.377
427310 98.890 -65.950 50.323
427311 98.931 -65.957 50.270
427312 98.970 -65.965 50.220
427313 99.009 -65.972 50.171
427314 99.047 -65.980 50.124
427315 99.084 -65.987 50.079
427316 99.124 -65.991 50.034
427317 99.166 -65.996 49.987
427318 99.211 -66.004 49.939
427319 99.254 -66.011 49.889
427320 99.293 -66.016 49.841
427321 99.333 -66.021 49.794
427322 99.373 -66.026 49.748
427323 99.413 -66.031 49.702
427324 99.452 -66.036 49.654
427325 99.492 -66.041 49.601
427326 99.534 -66.045 49.548
427327 99.576 -66.047 49.495
427328 99.619 -66.050 49.443
427329 99.661 -66.053 49.390
427330 99.704 -66.056 49.337
427331 99.747 -66.058 49.283
427332 99.789 -66.060 49.229
427333 99.828 -66.065 49.175
427334 99.867 -66.069 49.121
427335 99.907 -66.074 49.068
427336 99.945 -66.072 49.017
427337 99.984 -66.060 48.968
427338 100.027 -66.050 48.922
427339 100.075 -66.048 48.874
427340 100.123 -66.046 48.827
427341 100.171 -66.044 48.781
427342 100.219 -66.042 48.735
427343 100.268 -66.040 48.693
427344 100.316 -66.038 48.650
427345 100.364 -66.036 48.606
427346 100.412 -66.034 48.564
427347 100.461 -66.031 48.522
427348 100.509 -66.029 48.480
427759 88.307 -64.781 55.315
427760 88.340 -64.791 55.308
427761 88.375 -64.801 55.303
427762 88.409 -64.810 55.299
427763 88.444 -64.820 55.296
427764 88.478 -64.829 55.295
427765 88.512 -64.839 55.297
427766 88.557 -64.844 55.300
427767 88.604 -64.847 55.305
427768 88.650 -64.851 55.312
427769 88.696 -64.855 55.323
427770 88.742 -64.859 55.340
427771 88.788 -64.862 55.361
427772 88.834 -64.867 55.370
427773 88.881 -64.871 55.367
427774 88.927 -64.875 55.355
427775 88.973 -64.878 55.339
427776 89.019 -64.882 55.318
427777 89.065 -64.886 55.300
427778 89.112 -64.891 55.283
427779 89.158 -64.894 55.267
427780 89.204 -64.898 55.250
427781 89.250 -64.902 55.235
427782 89.296 -64.906 55.221
427783 89.344 -64.915 55.207
427784 89.389 -64.922 55.192
427785 89.437 -64.925 55.170
427786 89.485 -64.931 55.150
427787 89.533 -64.938 55.131
427788 89.582 -64.944 55.111
427789 89.630 -64.950 55.090
427790 89.679 -64.957 55.069
427791 89.727 -64.963 55.049
427792 89.776 -64.969 55.032
427793 89.824 -64.975 55.016
427794 89.873 -64.981 55.002
427795 89.921 -64.987 54.989
427796 89.969 -64.993 54.980
427797 90.018 -65.000 54.973
427810 90.648 -65.079 54.981
427811 90.696 -65.086 54.979
427812 90.745 -65.092 54.976
427813 90.793 -65.098 54.970
427814 90.840 -65.107 54.963
427815 90.886 -65.119 54.956
427816 90.932 -65.130 54.948
427817 90.976 -65.141 54.939
427818 91.019 -65.152 54.930
427819 91.062 -65.163 54.920
427820 91.105 -65.174 54.910
427821 91.148 -65.185 54.900
427822 91.191 -65.195 54.889
427823 91.234 -65.206 54.879
427824 91.275 -65.221 54.870
427825 91.314 -65.238 54.863
427826 91.354 -65.255 54.857
427827 91.394 -65.272 54.852
427828 91.434 -65.289 54.849
427829 91.473 -65.305 54.848
427830 91.513 -65.322 54.848
427831 91.553 -65.339 54.850
427832 91.593 -65.356 54.852
427833 91.633 -65.373 54.858
427834 91.672 -65.389 54.865
427835 91.712 -65.406 54.871
427836 91.752 -65.423 54.878
427837 91.792 -65.440 54.879
427838 91.831 -65.457 54.877
427839 91.871 -65.474 54.870
427840 91.907 -65.484 54.862
427841 91.959 -65.480 54.851
427842 92.011 -65.477 54.840
427843 92.066 -65.474 54.831
427844 92.124 -65.475 54.824
427845 92.182 -65.475 54.820
427846 92.240 -65.475 54.816
427847 92.294 -65.485 54.814
427848 92.346 -65.498 54.812
427849 92.398 -65.506 54.811
427850 92.449 -65.509 54.811
427851 92.500 -65.511 54.810
427852 92.552 -65.514 54.807
427853 92.603 -65.517 54.804
427854 92.654 -65.520 54.802
427855 92.706 -65.523 54.800
427856 92.757 -65.526 54.799
427857 92.808 -65.527 54.798
427858 92.858 -65.530 54.797
427859 92.909 -65.531 54.796
427860 92.959 -65.533 54.793
427861 93.010 -65.535 54.790
427862 93.060 -65.536 54.786
427863 93.111 -65.538 54.782
427864 93.161 -65.539 54.779
427865 93.212 -65.542 54.778
427866 93.262 -65.543 54.775
427867 93.308 -65.546 54.771
427868 93.351 -65.549 54.765
427869 93.394 -65.551 54.759
427870 93.436 -65.553 54.753
427871 93.479 -65.555 54.747
427872 93.521 -65.558 54.741
427873 93.564 -65.560 54.735
427874 93.607 -65.563 54.732
427875 93.649 -65.565 54.730
427876 93.692 -65.568 54.725
427877 93.734 -65.570 54.715
427878 93.777 -65.572 54.699
427879 93.820 -65.574 54.676
427880 93.863 -65.577 54.651
427881 93.900 -65.575 54.623
427882 93.932 -65.571 54.594
427883 93.966 -65.566 54.564
427884 93.998 -65.561 54.536
427885 94.032 -65.557 54.506
427886 94.064 -65.552 54.477
427887 94.098 -65.549 54.449
427888 94.131 -65.544 54.421
427889 94.164 -65.539 54.398
427890 94.197 -65.535 54.378
427891 94.230 -65.532 54.361
427892 94.265 -65.533 54.345
427893 94.299 -65.535 54.330
427894 94.335 -65.536 54.320
427895 94.374 -65.538 54.312
427896 94.413 -65.539 54.305
427897 94.453 -65.541 54.295
427898 94.492 -65.543 54.285
427899 94.532 -65.547 54.273
427900 94.572 -65.552 54.259
427901 94.612 -65.555 54.241
427902 94.651 -65.559 54.220
427903 94.691 -65.563 54.195
427904 94.731 -65.568 54.168
427905 94.770 -65.571 54.141
427906 94.811 -65.574 54.112
427907 94.854 -65.576 54.082
427908 94.897 -65.577 54.053
427909 94.941 -65.578 54.025
427910 94.983 -65.579 53.997
427911 95.027 -65.580 53.967
427912 95.070 -65.581 53.933
427913 95.114 -65.586 53.896
427914 95.160 -65.596 53.859
427915 95.204 -65.605 53.823
427916 95.249 -65.614 53.791
427917 95.294 -65.624 53.761
427918 95.334 -65.619 53.733
427919 95.373 -65.612 53.713
427920 95.412 -65.606 53.698
427921 95.452 -65.600 53.685
427922 95.495 -65.605 53.673
427923 95.539 -65.612 53.654
427924 95.584 -65.619 53.633
427925 95.626 -65.624 53.611
427926 95.667 -65.626 53.590
427927 95.707 -65.627 53.576
427928 95.747 -65.629 53.561
427929 95.788 -65.630 53.543
427930 95.828 -65.632 53.519
427931 95.868 -65.634 53.493
427932 95.908 -65.636 53.464
427933 95.949 -65.637 53.434
427934 95.989 -65.639 53.402
427935 96.028 -65.633 53.371
427936 96.068 -65.629 53.334
427937 96.110 -65.625 53.298
427938 96.154 -65.621 53.262
427939 96.197 -65.618 53.227
427940 96.241 -65.615 53.192
427941 96.285 -65.611 53.157
427942 96.328 -65.607 53.119
427943 96.372 -65.603 53.079
427944 96.415 -65.599 53.040
427945 96.461 -65.590 52.999
427946 96.506 -65.580 52.958
427947 96.550 -65.573 52.916
427948 96.592 -65.573 52.876
427949 96.633 -65.574 52.838
427950 96.675 -65.574 52.800
427951 96.717 -65.574 52.762
427952 96.758 -65.575 52.724
427953 96.788 -65.587 52.685
427954 96.819 -65.599 52.646
427955 96.849 -65.611 52.604
427956 96.880 -65.619 52.565
427957 96.910 -65.627 52.527
427958 96.941 -65.635 52.490
427959 96.971 -65.643 52.454
427960 97.002 -65.652 52.419
427961 97.038 -65.658 52.383
427962 97.077 -65.662 52.348
427963 97.115 -65.668 52.313
427964 97.152 -65.674 52.280
427965 97.179 -65.673 52.245
427966 97.202 -65.668 52.209
427967 97.227 -65.665 52.171
427968 97.270 -65.676 52.129
427969 97.312 -65.686 52.093
427970 97.355 -65.696 52.059
427971 97.397 -65.707 52.025
427972 97.437 -65.716 51.991
427973 97.473 -65.723 51.957
427974 97.509 -65.729 51.920
427975 97.544 -65.736 51.886
427976 97.580 -65.742 51.852
427977 97.617 -65.740 51.818
427978 97.654 -65.736 51.786
427979 97.690 -65.731 51.755
427980 97.727 -65.727 51.722
427981 97.764 -65.721 51.688
427982 97.801 -65.717 51.655
427983 97.838 -65.712 51.623
427984 97.872 -65.721 51.590
427985 97.906 -65.731 51.555
427986 97.940 -65.740 51.519
427987 97.972 -65.758 51.484
427988 98.005 -65.778 51.451
427989 98.037 -65.797 51.416
427990 98.071 -65.812 51.379
427991 98.106 -65.818 51.340
427992 98.141 -65.823 51.298
427993 98.178 -65.831 51.250
427994 98.218 -65.840 51.201
427995 98.258 -65.851 51.151
427996 98.297 -65.859 51.099
427997 98.335 -65.864 51.048
427998 98.372 -65.868 51.000
427999 98.409 -65.872 50.951
428000 98.447 -65.876 50.903
428001 98.484 -65.881 50.855
428002 98.521 -65.884 50.805
428003 98.557 -65.887 50.755
428004 98.590 -65.886 50.706
428005 98.628 -65.887 50.656
428006 98.672 -65.891 50.603
428007 98.717 -65.896 50.547
428008 98.762 -65.900 50.489
428009 98.806 -65.905 50.431
428010 98.851 -65.909 50.374
428011 98.896 -65.914 50.319
428012 98.936 -65.921 50.265
428013 98.974 -65.928 50.213
428014 99.011 -65.935 50.162
428015 99.049 -65.943 50.113
428016 99.087 -65.948 50.066
428017 99.128 -65.952 50.020
428018 99.168 -65.956 49.973
428019 99.213 -65.964 49.924
428020 99.258 -65.972 49.874
428021 99.302 -65.980 49.828
428022 99.342 -65.985 49.781
428023 99.381 -65.990 49.736
428024 99.421 -65.995 49.692
428025 99.461 -66.000 49.647
428026 99.501 -66.005 49.596
428027 99.543 -66.008 49.543
428028 99.585 -66.011 49.490
428029 99.628 -66.013 49.437
428030 99.671 -66.016 49.385
428031 99.713 -66.019 49.333
428032 99.756 -66.022 49.280
428033 99.798 -66.024 49.227
428034 99.841 -66.027 49.173
428035 99.881 -66.031 49.119
428036 99.920 -66.035 49.066
428037 99.959 -66.025 49.015
428038 99.998 -66.013 48.967
428039 100.041 -66.004 48.921
428040 100.089 -66.003 48.873
428041 100.137 -66.000 48.826
428042 100.186 -65.998 48.780
428043 100.234 -65.996 48.734
428044 100.282 -65.994 48.691
428045 100.330 -65.992 48.649
428046 100.379 -65.990 48.605
428047 100.427 -65.988 48.562
428048 100.475 -65.985 48.520
428049 100.524 -65.984 48.478
428460 88.298 -64.733 55.312
428461 88.333 -64.743 55.306
428462 88.367 -64.753 55.302
428463 88.401 -64.762 55.299
428464 88.435 -64.772 55.297
428465 88.470 -64.781 55.297
428466 88.514 -64.786 55.300
428467 88.560 -64.790 55.304
428468 88.606 -64.794 55.310
428469 88.652 -64.798 55.318
428470 88.698 -64.802 55.330
428471 88.745 -64.806 55.346
428472 88.790 -64.809 55.368
428473 88.837 -64.813 55.376
428474 88.883 -64.817 55.374
428475 88.929 -64.822 55.363
428476 88.975 -64.825 55.348
428477 89.022 -64.829 55.331
428478 89.067 -64.833 55.316
428479 89.114 -64.837 55.301
428480 89.160 -64.841 55.285
428481 89.206 -64.845 55.268
428482 89.259 -64.846 55.253
428483 89.312 -64.850 55.239
428484 89.349 -64.869 55.225
428485 89.390 -64.881 55.213
428486 89.438 -64.883 55.191
428487 89.485 -64.885 55.167
428488 89.532 -64.888 55.146
428489 89.581 -64.895 55.126
428490 89.629 -64.901 55.103
428491 89.678 -64.907 55.080
428492 89.726 -64.913 55.059
428493 89.774 -64.919 55.040
428494 89.823 -64.925 55.023
428495 89.872 -64.932 55.007
428496 89.920 -64.938 54.995
428497 89.968 -64.944 54.985
428498 90.017 -64.950 54.977
428509 90.549 -65.019 54.981
428510 90.598 -65.025 54.981
428511 90.647 -65.031 54.981
428512 90.695 -65.037 54.979
428513 90.743 -65.043 54.975
428514 90.792 -65.049 54.969
428515 90.838 -65.060 54.962
428516 90.884 -65.072 54.955
428517 90.929 -65.083 54.946
428518 90.972 -65.094 54.937
428519 91.015 -65.105 54.928
428520 91.058 -65.116 54.918
428521 91.101 -65.126 54.907
428522 91.144 -65.137 54.897
428523 91.187 -65.148 54.886
428524 91.227 -65.164 54.876
428525 91.267 -65.181 54.866
428526 91.307 -65.198 54.857
428527 91.347 -65.214 54.851
428528 91.386 -65.232 54.845
428529 91.426 -65.248 54.841
428530 91.466 -65.265 54.838
428531 91.506 -65.282 54.837
428532 91.545 -65.298 54.838
428533 91.585 -65.316 54.839
428534 91.625 -65.333 54.843
428535 91.665 -65.349 54.849
428536 91.705 -65.366 54.854
428537 91.744 -65.383 54.861
428538 91.784 -65.400 54.866
428539 91.824 -65.417 54.865
428540 91.864 -65.433 54.861
428541 91.899 -65.451 54.852
428542 91.950 -65.447 54.841
428543 92.005 -65.445 54.830
428544 92.064 -65.445 54.820
428545 92.122 -65.445 54.813
428546 92.180 -65.446 54.808
428547 92.238 -65.446 54.806
428548 92.294 -65.449 54.802
428549 92.346 -65.452 54.799
428550 92.397 -65.455 54.796
428551 92.448 -65.458 54.796
428552 92.500 -65.461 54.795
428553 92.551 -65.464 54.793
428554 92.602 -65.466 54.791
428555 92.654 -65.469 54.791
428556 92.705 -65.471 54.792
428557 92.757 -65.474 54.793
428558 92.808 -65.477 54.792
428559 92.858 -65.479 54.790
428560 92.908 -65.481 54.788
428561 92.958 -65.483 54.784
428562 93.008 -65.486 54.779
428563 93.058 -65.488 54.773
428564 93.109 -65.490 54.768
428565 93.158 -65.492 54.763
428566 93.209 -65.495 54.759
428567 93.259 -65.497 54.754
428568 93.307 -65.501 54.748
428569 93.351 -65.503 54.742
428570 93.393 -65.505 54.736
428571 93.436 -65.508 54.730
428572 93.479 -65.511 54.725
428573 93.521 -65.513 54.720
428574 93.564 -65.515 54.717
428575 93.607 -65.517 54.715
428576 93.649 -65.520 54.714
428577 93.692 -65.522 54.709
428578 93.734 -65.525 54.699
428579 93.777 -65.527 54.684
428580 93.820 -65.530 54.662
428581 93.864 -65.533 54.637
428582 93.903 -65.533 54.611
428583 93.936 -65.528 54.584
428584 93.969 -65.524 54.555
428585 94.002 -65.519 54.529
428586 94.035 -65.514 54.502
428587 94.068 -65.510 54.475
428588 94.101 -65.505 54.450
428589 94.134 -65.501 54.426
428590 94.168 -65.496 54.405
428591 94.200 -65.492 54.388
428592 94.235 -65.492 54.373
428593 94.269 -65.494 54.359
428594 94.304 -65.496 54.345
428595 94.339 -65.497 54.336
428596 94.378 -65.499 54.327
428597 94.417 -65.500 54.319
428598 94.456 -65.502 54.311
428599 94.496 -65.503 54.302
428600 94.535 -65.505 54.291
428601 94.575 -65.508 54.276
428602 94.615 -65.513 54.257
428603 94.654 -65.517 54.233
428604 94.694 -65.521 54.204
428605 94.734 -65.525 54.174
428606 94.773 -65.529 54.144
428607 94.815 -65.531 54.113
428608 94.858 -65.532 54.083
428609 94.901 -65.533 54.056
428610 94.945 -65.534 54.030
428611 94.987 -65.536 54.004
428612 95.031 -65.538 53.975
428613 95.075 -65.542 53.942
428614 95.119 -65.546 53.907
428615 95.166 -65.555 53.871
428616 95.211 -65.565 53.837
428617 95.256 -65.574 53.807
428618 95.301 -65.583 53.778
428619 95.340 -65.577 53.751
428620 95.380 -65.571 53.731
428621 95.419 -65.564 53.718
428622 95.458 -65.558 53.706
428623 95.500 -65.559 53.694
428624 95.544 -65.567 53.675
428625 95.588 -65.574 53.652
428626 95.628 -65.575 53.628
428627 95.669 -65.577 53.603
428628 95.709 -65.578 53.585
428629 95.750 -65.580 53.567
428630 95.790 -65.582 53.547
428631 95.830 -65.583 53.524
428632 95.871 -65.585 53.496
428633 95.911 -65.586 53.466
428634 95.951 -65.588 53.436
428635 95.991 -65.590 53.405
428636 96.032 -65.591 53.374
428637 96.073 -65.589 53.337
428638 96.116 -65.585 53.300
428639 96.160 -65.581 53.263
428640 96.203 -65.577 53.226
428641 96.247 -65.574 53.190
428642 96.291 -65.571 53.155
428643 96.334 -65.567 53.117
428644 96.378 -65.563 53.078
428645 96.421 -65.560 53.039
428646 96.466 -65.551 53.000
428647 96.510 -65.546 52.959
428648 96.552 -65.546 52.917
428649 96.594 -65.546 52.876
428650 96.636 -65.546 52.833
428651 96.677 -65.546 52.794
428652 96.719 -65.546 52.755
428653 96.761 -65.547 52.717
428654 96.801 -65.549 52.678
428655 96.831 -65.561 52.639
428656 96.862 -65.569 52.597
428657 96.893 -65.577 52.555
428658 96.924 -65.586 52.516
428659 96.954 -65.593 52.478
428660 96.985 -65.602 52.442
428661 97.016 -65.610 52.407
428662 97.046 -65.618 52.370
428663 97.079 -65.625 52.334
428664 97.117 -65.630 52.300
428665 97.151 -65.634 52.266
428666 97.174 -65.630 52.231
428667 97.197 -65.624 52.195
428668 97.226 -65.624 52.156
428669 97.268 -65.634 52.115
428670 97.310 -65.645 52.078
428671 97.353 -65.655 52.042
428672 97.395 -65.666 52.007
428673 97.438 -65.676 51.973
428674 97.475 -65.684 51.939
428675 97.511 -65.690 51.904
428676 97.546 -65.696 51.869
428677 97.583 -65.699 51.833
428678 97.619 -65.695 51.799
428679 97.656 -65.690 51.765
428680 97.693 -65.685 51.732
428681 97.730 -65.680 51.698
428682 97.767 -65.676 51.664
428683 97.803 -65.671 51.632
428684 97.840 -65.667 51.600
428685 97.875 -65.676 51.568
428686 97.907 -65.696 51.533
428687 97.939 -65.715 51.498
428688 97.971 -65.735 51.463
428689 98.004 -65.755 51.431
428690 98.038 -65.768 51.397
428691 98.073 -65.774 51.362
428692 98.108 -65.779 51.324
428693 98.144 -65.784 51.284
428694 98.179 -65.789 51.237
428695 98.217 -65.798 51.191
428696 98.257 -65.808 51.143
428697 98.297 -65.818 51.094
428698 98.334 -65.821 51.044
428699 98.371 -65.826 50.997
428700 98.409 -65.830 50.950
428701 98.446 -65.834 50.903
428702 98.483 -65.838 50.856
428703 98.521 -65.843 50.806
428704 98.558 -65.847 50.759
428705 98.595 -65.851 50.710
428706 98.630 -65.851 50.659
428707 98.675 -65.856 50.607
428708 98.720 -65.860 50.549
428709 98.765 -65.865 50.490
428710 98.809 -65.869 50.430
428711 98.853 -65.874 50.370
428712 98.897 -65.878 50.313
428713 98.938 -65.884 50.258
428714 98.976 -65.891 50.204
428715 99.013 -65.898 50.152
428716 99.051 -65.905 50.100
428717 99.091 -65.909 50.051
428718 99.131 -65.913 50.004
428719 99.171 -65.917 49.957
428720 99.215 -65.924 49.908
428721 99.260 -65.932 49.860
428722 99.305 -65.941 49.813
428723 99.350 -65.948 49.767
428724 99.390 -65.953 49.724
428725 99.430 -65.959 49.681
428726 99.470 -65.964 49.638
428727 99.509 -65.969 49.588
428728 99.552 -65.972 49.536
428729 99.594 -65.975 49.483
428730 99.637 -65.977 49.430
428731 99.680 -65.980 49.377
428732 99.722 -65.982 49.325
428733 99.765 -65.984 49.274
428734 99.807 -65.988 49.223
428735 99.850 -65.990 49.170
428736 99.893 -65.993 49.115
428737 99.934 -65.991 49.063
428738 99.973 -65.978 49.014
428739 100.011 -65.966 48.967
428740 100.055 -65.959 48.920
428741 100.104 -65.956 48.872
428742 100.152 -65.954 48.825
428743 100.200 -65.952 48.779
428744 100.248 -65.950 48.734
428745 100.296 -65.948 48.690
428746 100.345 -65.946 48.647
428747 100.393 -65.944 48.603
428748 100.441 -65.941 48.560
428749 100.489 -65.940 48.519
429161 88.290 -64.685 55.309
429162 88.325 -64.694 55.304
429163 88.359 -64.704 55.301
429164 88.393 -64.713 55.299
429165 88.427 -64.723 55.298
429166 88.470 -64.729 55.298
429167 88.516 -64.733 55.302
429168 88.562 -64.737 55.307
429169 88.608 -64.740 55.314
429170 88.654 -64.744 55.323
429171 88.700 -64.749 55.334
429172 88.747 -64.753 55.350
429173 88.793 -64.756 55.371
429174 88.839 -64.760 55.380
429175 88.885 -64.764 55.379
429176 88.931 -64.768 55.370
429177 88.978 -64.772 55.356
429178 89.024 -64.776 55.343
429179 89.070 -64.780 55.331
429180 89.116 -64.784 55.319
429181 89.162 -64.787 55.303
429182 89.222 -64.785 55.285
429183 89.280 -64.785 55.270
429184 89.317 -64.803 55.257
429185 89.354 -64.822 55.244
429186 89.391 -64.840 55.232
429187 89.439 -64.842 55.212
429188 89.486 -64.844 55.189
429189 89.533 -64.847 55.164
429190 89.581 -64.849 55.139
429191 89.628 -64.852 55.113
429192 89.676 -64.858 55.090
429193 89.725 -64.864 55.068
429194 89.773 -64.871 55.049
429195 89.822 -64.877 55.031
429196 89.870 -64.883 55.014
429197 89.918 -64.889 55.001
429198 89.967 -64.895 54.991
429199 90.016 -64.901 54.983
429200 90.064 -64.907 54.976
429205 90.306 -64.938 54.972
429206 90.355 -64.944 54.975
429207 90.403 -64.950 54.977
429208 90.452 -64.957 54.979
429209 90.500 -64.963 54.981
429210 90.548 -64.969 54.980
429211 90.597 -64.975 54.981
429212 90.646 -64.982 54.980
429213 90.694 -64.988 54.978
429214 90.742 -64.994 54.974
429215 90.790 -65.002 54.968
429216 90.836 -65.013 54.961
429217 90.881 -65.025 54.953
429218 90.926 -65.035 54.945
429219 90.968 -65.047 54.935
429220 91.011 -65.057 54.925
429221 91.054 -65.068 54.915
429222 91.097 -65.079 54.905
429223 91.140 -65.090 54.894
429224 91.179 -65.107 54.883
429225 91.219 -65.124 54.873
429226 91.259 -65.141 54.862
429227 91.299 -65.157 54.853
429228 91.338 -65.174 54.845
429229 91.378 -65.191 54.837
429230 91.418 -65.208 54.832
429231 91.458 -65.225 54.829
429232 91.498 -65.242 54.827
429233 91.537 -65.258 54.827
429234 91.577 -65.275 54.827
429235 91.617 -65.292 54.830
429236 91.657 -65.309 54.834
429237 91.697 -65.326 54.838
429238 91.736 -65.342 54.846
429239 91.776 -65.359 54.853
429240 91.816 -65.376 54.855
429241 91.856 -65.393 54.851
429242 91.895 -65.410 54.843
429243 91.945 -65.416 54.831
429244 92.003 -65.414 54.820
429245 92.061 -65.411 54.809
429246 92.119 -65.408 54.801
429247 92.176 -65.405 54.796
429248 92.235 -65.403 54.793
429249 92.292 -65.402 54.789
429250 92.349 -65.402 54.786
429251 92.401 -65.405 54.784
429252 92.452 -65.408 54.782
429253 92.503 -65.411 54.780
429254 92.555 -65.414 54.779
429255 92.606 -65.416 54.779
429256 92.657 -65.419 54.780
429257 92.709 -65.421 54.781
429258 92.760 -65.424 54.781
429259 92.809 -65.429 54.781
429260 92.857 -65.432 54.779
429261 92.905 -65.435 54.776
429262 92.955 -65.437 54.771
429263 93.005 -65.439 54.766
429264 93.055 -65.442 54.759
429265 93.105 -65.444 54.752
429266 93.155 -65.446 54.746
429267 93.205 -65.449 54.740
429268 93.256 -65.451 54.734
429269 93.303 -65.455 54.727
429270 93.350 -65.458 54.719
429271 93.393 -65.461 54.713
429272 93.435 -65.463 54.707
429273 93.478 -65.465 54.703
429274 93.521 -65.467 54.700
429275 93.564 -65.470 54.697
429276 93.606 -65.473 54.696
429277 93.649 -65.475 54.695
429278 93.692 -65.477 54.690
429279 93.734 -65.480 54.681
429280 93.777 -65.483 54.668
429281 93.822 -65.486 54.647
429282 93.866 -65.489 54.624
429283 93.907 -65.489 54.599
429284 93.940 -65.486 54.575
429285 93.973 -65.481 54.550
429286 94.006 -65.477 54.526
429287 94.039 -65.472 54.502
429288 94.072 -65.467 54.480
429289 94.105 -65.463 54.459
429290 94.138 -65.458 54.437
429291 94.171 -65.454 54.417
429292 94.205 -65.452 54.401
429293 94.240 -65.453 54.387
429294 94.274 -65.455 54.375
429295 94.309 -65.456 54.362
429296 94.343 -65.458 54.353
429297 94.381 -65.459 54.344
429298 94.420 -65.461 54.335
429299 94.460 -65.462 54.325
429300 94.499 -65.464 54.315
429301 94.539 -65.465 54.302
429302 94.578 -65.467 54.287
429303 94.618 -65.470 54.267
429304 94.657 -65.474 54.242
429305 94.697 -65.478 54.214
429306 94.737 -65.482 54.182
429307 94.776 -65.486 54.149
429308 94.819 -65.488 54.117
429309 94.862 -65.489 54.088
429310 94.905 -65.490 54.062
429311 94.949 -65.491 54.036
429312 94.992 -65.495 54.009
429313 95.036 -65.499 53.980
429314 95.080 -65.502 53.949
429315 95.124 -65.505 53.916
429316 95.171 -65.515 53.883
429317 95.217 -65.525 53.851
429318 95.263 -65.535 53.822
429319 95.308 -65.543 53.795
429320 95.347 -65.536 53.769
429321 95.386 -65.530 53.750
429322 95.425 -65.524 53.737
429323 95.464 -65.517 53.727
429324 95.505 -65.514 53.715
429325 95.549 -65.521 53.696
429326 95.591 -65.524 53.672
429327 95.631 -65.526 53.646
429328 95.671 -65.527 53.618
429329 95.711 -65.529 53.595
429330 95.752 -65.531 53.574
429331 95.793 -65.533 53.552
429332 95.833 -65.534 53.529
429333 95.873 -65.536 53.499
429334 95.913 -65.539 53.468
429335 95.954 -65.544 53.436
429336 95.993 -65.549 53.405
429337 96.034 -65.555 53.374
429338 96.076 -65.550 53.338
429339 96.120 -65.546 53.300
429340 96.164 -65.543 53.261
429341 96.208 -65.539 53.223
429342 96.252 -65.535 53.187
429343 96.296 -65.531 53.152
429344 96.339 -65.527 53.114
429345 96.383 -65.524 53.076
429346 96.427 -65.520 53.037
429347 96.470 -65.518 52.998
429348 96.511 -65.518 52.958
429349 96.553 -65.518 52.917
429350 96.595 -65.518 52.875
429351 96.636 -65.519 52.831
429352 96.678 -65.519 52.789
429353 96.720 -65.519 52.749
429354 96.761 -65.519 52.710
429355 96.803 -65.519 52.671
429356 96.845 -65.520 52.632
429357 96.879 -65.526 52.589
429358 96.909 -65.534 52.547
429359 96.939 -65.543 52.506
429360 96.969 -65.551 52.467
429361 97.000 -65.559 52.430
429362 97.030 -65.568 52.394
429363 97.060 -65.577 52.357
429364 97.090 -65.585 52.320
429365 97.121 -65.593 52.286
429366 97.148 -65.590 52.251
429367 97.175 -65.585 52.215
429368 97.203 -65.580 52.179
429369 97.226 -65.585 52.142
429370 97.266 -65.593 52.100
429371 97.308 -65.604 52.062
429372 97.351 -65.614 52.025
429373 97.393 -65.624 51.990
429374 97.436 -65.635 51.955
429375 97.476 -65.645 51.922
429376 97.513 -65.651 51.889
429377 97.548 -65.658 51.854
429378 97.585 -65.654 51.818
429379 97.622 -65.649 51.781
429380 97.659 -65.644 51.747
429381 97.696 -65.640 51.712
429382 97.732 -65.636 51.677
429383 97.768 -65.637 51.643
429384 97.804 -65.640 51.610
429385 97.840 -65.642 51.577
429386 97.874 -65.653 51.544
429387 97.906 -65.673 51.511
429388 97.938 -65.693 51.476
429389 97.971 -65.713 51.442
429390 98.005 -65.724 51.412
429391 98.040 -65.729 51.379
429392 98.075 -65.734 51.345
429393 98.110 -65.740 51.308
429394 98.146 -65.745 51.268
429395 98.181 -65.750 51.224
429396 98.216 -65.756 51.179
429397 98.256 -65.765 51.132
429398 98.296 -65.775 51.085
429399 98.334 -65.780 51.037
429400 98.371 -65.784 50.992
429401 98.408 -65.788 50.947
429402 98.445 -65.792 50.901
429403 98.483 -65.796 50.854
429404 98.520 -65.800 50.806
429405 98.558 -65.805 50.758
429406 98.595 -65.809 50.710
429407 98.633 -65.814 50.660
429408 98.676 -65.819 50.607
429409 98.720 -65.824 50.548
429410 98.764 -65.828 50.488
429411 98.808 -65.834 50.428
429412 98.852 -65.838 50.367
429413 98.896 -65.843 50.308
429414 98.940 -65.847 50.251
429415 98.978 -65.854 50.196
429416 99.015 -65.861 50.141
429417 99.055 -65.866 50.088
429418 99.095 -65.870 50.037
429419 99.135 -65.875 49.989
429420 99.175 -65.878 49.941
429421 99.218 -65.884 49.893
429422 99.263 -65.892 49.845
429423 99.307 -65.900 49.798
429424 99.352 -65.909 49.754
429425 99.397 -65.916 49.710
429426 99.439 -65.922 49.668
429427 99.478 -65.928 49.626
429428 99.518 -65.932 49.579
429429 99.561 -65.935 49.528
429430 99.604 -65.937 49.476
429431 99.646 -65.941 49.423
429432 99.689 -65.943 49.369
429433 99.731 -65.946 49.318
429434 99.774 -65.948 49.267
429435 99.822 -65.955 49.216
429436 99.870 -65.962 49.165
429437 99.919 -65.969 49.112
429438 99.947 -65.944 49.061
429439 99.986 -65.931 49.012
429440 100.025 -65.918 48.964
429441 100.070 -65.912 48.917
429442 100.118 -65.910 48.870
429443 100.167 -65.908 48.823
429444 100.214 -65.906 48.777
429445 100.263 -65.904 48.732
429446 100.311 -65.902 48.687
429447 100.359 -65.900 48.645
429448 100.407 -65.897 48.601
429449 100.456 -65.896 48.558
429450 100.504 -65.894 48.516
429863 88.317 -64.646 55.303
429864 88.351 -64.656 55.300
429865 88.385 -64.665 55.299
429866 88.426 -64.671 55.298
429867 88.472 -64.675 55.300
429868 88.518 -64.680 55.304
429869 88.564 -64.684 55.310
429870 88.611 -64.687 55.317
429871 88.657 -64.691 55.326
429872 88.703 -64.695 55.338
429873 88.749 -64.699 55.354
429874 88.795 -64.703 55.374
429875 88.841 -64.707 55.384
429876 88.888 -64.711 55.383
429877 88.934 -64.715 55.376
429878 88.980 -64.718 55.365
429879 89.026 -64.722 55.353
429880 89.072 -64.727 55.344
429881 89.121 -64.729 55.334
429882 89.185 -64.725 55.318
429883 89.248 -64.720 55.300
429884 89.285 -64.739 55.286
429885 89.322 -64.757 55.274
429886 89.358 -64.776 55.261
429887 89.396 -64.794 55.250
429888 89.440 -64.802 55.231
429889 89.487 -64.804 55.208
429890 89.534 -64.806 55.181
429891 89.582 -64.809 55.152
429892 89.629 -64.811 55.124
429893 89.676 -64.813 55.100
429894 89.724 -64.815 55.077
429895 89.772 -64.822 55.057
429896 89.821 -64.828 55.038
429897 89.869 -64.834 55.021
429898 89.917 -64.840 55.006
429899 89.966 -64.847 54.995
429900 90.014 -64.853 54.987
429901 90.063 -64.859 54.981
429902 90.111 -64.865 54.977
429903 90.160 -64.871 54.975
429904 90.208 -64.877 54.974
429905 90.257 -64.883 54.974
429906 90.305 -64.890 54.975
429907 90.354 -64.896 54.976
429908 90.402 -64.902 54.978
429909 90.450 -64.908 54.980
429910 90.499 -64.914 54.981
429911 90.547 -64.920 54.981
429912 90.596 -64.926 54.981
429913 90.644 -64.933 54.980
429914 90.692 -64.939 54.977
429915 90.741 -64.945 54.973
429916 90.788 -64.954 54.967
429917 90.833 -64.966 54.960
429918 90.879 -64.978 54.952
429919 90.922 -64.988 54.943
429920 90.964 -64.999 54.933
429921 91.007 -65.010 54.923
429922 91.050 -65.020 54.913
429923 91.092 -65.033 54.902
429924 91.132 -65.050 54.891
429925 91.172 -65.066 54.880
429926 91.212 -65.083 54.869
429927 91.251 -65.101 54.858
429928 91.291 -65.117 54.848
429929 91.331 -65.134 54.839
429930 91.371 -65.151 54.831
429931 91.411 -65.167 54.825
429932 91.450 -65.185 54.821
429933 91.490 -65.201 54.818
429934 91.530 -65.218 54.815
429935 91.570 -65.235 54.816
429936 91.609 -65.252 54.818
429937 91.649 -65.269 54.821
429938 91.689 -65.286 54.825
429939 91.729 -65.302 54.833
429940 91.769 -65.319 54.842
429941 91.810 -65.334 54.845
429942 91.855 -65.344 54.842
429943 91.900 -65.355 54.835
429944 91.941 -65.369 54.822
429945 91.999 -65.367 54.809
429946 92.057 -65.364 54.798
429947 92.115 -65.361 54.790
429948 92.173 -65.358 54.784
429949 92.231 -65.355 54.780
429950 92.288 -65.354 54.776
429951 92.345 -65.355 54.773
429952 92.402 -65.356 54.771
429953 92.456 -65.358 54.768
429954 92.507 -65.361 54.766
429955 92.559 -65.364 54.765
429956 92.610 -65.366 54.765
429957 92.661 -65.369 54.766
429958 92.713 -65.372 54.766
429959 92.763 -65.376 54.766
429960 92.812 -65.380 54.765
429961 92.860 -65.383 54.764
429962 92.907 -65.386 54.761
429963 92.955 -65.389 54.756
429964 93.003 -65.392 54.751
429965 93.052 -65.395 54.745
429966 93.102 -65.397 54.738
429967 93.152 -65.399 54.733
429968 93.202 -65.402 54.726
429969 93.252 -65.404 54.719
429970 93.300 -65.408 54.711
429971 93.348 -65.412 54.703
429972 93.393 -65.415 54.696
429973 93.435 -65.418 54.689
429974 93.478 -65.420 54.685
429975 93.520 -65.423 54.681
429976 93.563 -65.425 54.678
429977 93.606 -65.427 54.676
429978 93.648 -65.430 54.673
429979 93.691 -65.432 54.668
429980 93.734 -65.435 54.661
429981 93.778 -65.438 54.649
429982 93.823 -65.441 54.631
429983 93.867 -65.445 54.610
429984 93.911 -65.447 54.589
429985 93.943 -65.442 54.568
429986 93.977 -65.438 54.547
429987 94.010 -65.433 54.527
429988 94.043 -65.430 54.508
429989 94.076 -65.425 54.490
429990 94.109 -65.420 54.472
429991 94.142 -65.416 54.452
429992 94.175 -65.411 54.433
429993 94.209 -65.412 54.416
429994 94.244 -65.414 54.402
429995 94.278 -65.415 54.390
429996 94.313 -65.417 54.378
429997 94.348 -65.418 54.367
429998 94.384 -65.420 54.357
429999 94.424 -65.421 54.347
430000 94.463 -65.423 54.336
430001 94.502 -65.424 54.324
430002 94.542 -65.426 54.309
430003 94.581 -65.427 54.292
430004 94.621 -65.429 54.272
430005 94.660 -65.432 54.248
430006 94.700 -65.436 54.221
430007 94.740 -65.439 54.188
430008 94.780 -65.443 54.154
430009 94.823 -65.444 54.121
430010 94.866 -65.445 54.092
430011 94.910 -65.448 54.065
430012 94.954 -65.452 54.040
430013 94.997 -65.455 54.013
430014 95.042 -65.458 53.983
430015 95.086 -65.462 53.953
430016 95.129 -65.466 53.922
430017 95.176 -65.474 53.892
430018 95.222 -65.485 53.864
430019 95.269 -65.495 53.837
430020 95.314 -65.501 53.811
430021 95.353 -65.495 53.788
430022 95.392 -65.488 53.769
430023 95.431 -65.482 53.756
430024 95.471 -65.476 53.745
430025 95.510 -65.469 53.731
430026 95.553 -65.474 53.713
430027 95.593 -65.475 53.690
430028 95.633 -65.477 53.662
430029 95.674 -65.479 53.633
430030 95.714 -65.484 53.607
430031 95.754 -65.489 53.582
430032 95.795 -65.494 53.558
430033 95.835 -65.499 53.531
430034 95.875 -65.505 53.502
430035 95.915 -65.509 53.469
430036 95.955 -65.514 53.435
430037 95.995 -65.520 53.402
430038 96.037 -65.517 53.370
430039 96.079 -65.512 53.334
430040 96.123 -65.508 53.296
430041 96.167 -65.505 53.257
430042 96.210 -65.501 53.219
430043 96.255 -65.497 53.183
430044 96.299 -65.493 53.148
430045 96.342 -65.489 53.110
430046 96.386 -65.486 53.072
430047 96.430 -65.483 53.033
430048 96.468 -65.486 52.995
430049 96.507 -65.489 52.955
430050 96.548 -65.489 52.915
430051 96.590 -65.489 52.874
430052 96.632 -65.489 52.830
430053 96.673 -65.489 52.785
430054 96.715 -65.490 52.744
430055 96.757 -65.490 52.703
430056 96.798 -65.490 52.663
430057 96.840 -65.491 52.624
430058 96.883 -65.492 52.581
430059 96.926 -65.492 52.539
430060 96.967 -65.494 52.497
430061 96.997 -65.503 52.457
430062 97.027 -65.511 52.418
430063 97.058 -65.520 52.382
430064 97.088 -65.528 52.344
430065 97.118 -65.536 52.306
430066 97.148 -65.545 52.272
430067 97.176 -65.541 52.235
430068 97.203 -65.536 52.199
430069 97.205 -65.534 52.163
430070 97.231 -65.550 52.126
430071 97.272 -65.558 52.084
430072 97.311 -65.567 52.046
430073 97.351 -65.575 52.008
430074 97.391 -65.583 51.972
430075 97.434 -65.593 51.938
430076 97.476 -65.604 51.907
430077 97.514 -65.612 51.874
430078 97.550 -65.613 51.839
430079 97.587 -65.608 51.802
430080 97.624 -65.604 51.764
430081 97.660 -65.606 51.727
430082 97.696 -65.608 51.692
430083 97.732 -65.610 51.657
430084 97.768 -65.612 51.622
430085 97.803 -65.615 51.588
430086 97.840 -65.617 51.555
430087 97.873 -65.631 51.521
430088 97.905 -65.651 51.488
430089 97.938 -65.671 51.455
430090 97.972 -65.680 51.421
430091 98.007 -65.684 51.391
430092 98.042 -65.690 51.359
430093 98.078 -65.695 51.326
430094 98.113 -65.700 51.290
430095 98.148 -65.705 51.251
430096 98.183 -65.711 51.210
430097 98.219 -65.716 51.166
430098 98.255 -65.722 51.121
430099 98.295 -65.733 51.075
430100 98.333 -65.737 51.030
430101 98.370 -65.742 50.986
430102 98.408 -65.746 50.942
430103 98.445 -65.750 50.897
430104 98.482 -65.754 50.851
430105 98.519 -65.759 50.804
430106 98.557 -65.764 50.756
430107 98.595 -65.769 50.707
430108 98.636 -65.774 50.656
430109 98.678 -65.780 50.603
430110 98.721 -65.786 50.544
430111 98.764 -65.791 50.483
430112 98.807 -65.797 50.423
430113 98.851 -65.802 50.364
430114 98.895 -65.806 50.304
430115 98.939 -65.812 50.246
430116 98.980 -65.817 50.190
430117 99.018 -65.824 50.134
430118 99.058 -65.828 50.079
430119 99.098 -65.831 50.027
430120 99.139 -65.836 49.976
430121 99.179 -65.840 49.928
430122 99.220 -65.844 49.880
430123 99.265 -65.853 49.832
430124 99.309 -65.860 49.786
430125 99.354 -65.868 49.741
430126 99.399 -65.876 49.698
430127 99.444 -65.884 49.655
430128 99.487 -65.891 49.614
430129 99.527 -65.896 49.569
430130 99.570 -65.899 49.520
430131 99.613 -65.901 49.468
430132 99.657 -65.906 49.415
430133 99.706 -65.912 49.362
430134 99.755 -65.920 49.310
430135 99.804 -65.928 49.260
430136 99.853 -65.935 49.210
430137 99.902 -65.942 49.160
430138 99.934 -65.925 49.110
430139 99.961 -65.897 49.059
430140 99.999 -65.884 49.010
430141 100.038 -65.871 48.962
430142 100.084 -65.866 48.914
430143 100.132 -65.864 48.867
430144 100.181 -65.862 48.821
430145 100.229 -65.860 48.774
430146 100.277 -65.858 48.729
430147 100.325 -65.856 48.685
430148 100.373 -65.854 48.643
430149 100.422 -65.852 48.599
430150 100.470 -65.850 48.556
430564 88.309 -64.597 55.302
430565 88.343 -64.607 55.299
430566 88.382 -64.615 55.298
430567 88.428 -64.618 55.299
430568 88.474 -64.622 55.301
430569 88.521 -64.626 55.306
430570 88.567 -64.631 55.312
430571 88.613 -64.634 55.320
430572 88.659 -64.638 55.329
430573 88.705 -64.642 55.340
430574 88.751 -64.646 55.356
430575 88.798 -64.649 55.376
430576 88.844 -64.654 55.387
430577 88.890 -64.658 55.389
430578 88.936 -64.662 55.382
430579 88.982 -64.665 55.372
430580 89.028 -64.669 55.363
430581 89.084 -64.668 55.353
430582 89.147 -64.664 55.343
430583 89.210 -64.660 55.327
430584 89.253 -64.674 55.311
430585 89.290 -64.693 55.300
430586 89.327 -64.711 55.289
430587 89.364 -64.730 55.276
430588 89.400 -64.748 55.266
430589 89.441 -64.761 55.249
430590 89.488 -64.763 55.226
430591 89.536 -64.765 55.198
430592 89.583 -64.768 55.166
430593 89.630 -64.770 55.135
430594 89.678 -64.772 55.109
430595 89.725 -64.775 55.086
430596 89.772 -64.777 55.064
430597 89.819 -64.779 55.045
430598 89.868 -64.785 55.028
430599 89.916 -64.791 55.013
430600 89.965 -64.797 55.001
430601 90.013 -64.803 54.993
430602 90.061 -64.809 54.986
430603 90.110 -64.815 54.982
430604 90.159 -64.822 54.980
430605 90.207 -64.828 54.979
430606 90.255 -64.834 54.978
430607 90.304 -64.841 54.979
430608 90.353 -64.847 54.979
430609 90.401 -64.853 54.981
430610 90.449 -64.859 54.982
430611 90.498 -64.866 54.983
430612 90.546 -64.872 54.983
430613 90.595 -64.878 54.982
430614 90.643 -64.884 54.980
430615 90.691 -64.890 54.977
430616 90.740 -64.896 54.972
430617 90.786 -64.907 54.966
430618 90.831 -64.919 54.959
430619 90.875 -64.930 54.950
430620 90.918 -64.941 54.941
430621 90.961 -64.951 54.931
430622 91.004 -64.962 54.921
430623 91.045 -64.975 54.910
430624 91.085 -64.993 54.899
430625 91.124 -65.010 54.888
430626 91.164 -65.026 54.877
430627 91.204 -65.043 54.865
430628 91.244 -65.060 54.853
430629 91.284 -65.077 54.843
430630 91.323 -65.094 54.834
430631 91.363 -65.110 54.825
430632 91.403 -65.127 54.818
430633 91.443 -65.145 54.813
430634 91.483 -65.161 54.809
430635 91.522 -65.178 54.806
430636 91.562 -65.195 54.806
430637 91.602 -65.211 54.807
430638 91.642 -65.229 54.809
430639 91.684 -65.242 54.813
430640 91.729 -65.251 54.821
430641 91.775 -65.261 54.830
430642 91.820 -65.272 54.835
430643 91.865 -65.282 54.833
430644 91.910 -65.292 54.825
430645 91.955 -65.301 54.812
430646 91.995 -65.318 54.799
430647 92.053 -65.315 54.787
430648 92.111 -65.313 54.778
430649 92.169 -65.310 54.770
430650 92.227 -65.307 54.766
430651 92.284 -65.306 54.761
430652 92.341 -65.307 54.758
430653 92.398 -65.308 54.755
430654 92.455 -65.309 54.754
430655 92.511 -65.311 54.752
430656 92.563 -65.314 54.751
430657 92.614 -65.316 54.751
430658 92.665 -65.319 54.750
430659 92.716 -65.323 54.749
430660 92.766 -65.327 54.747
430661 92.815 -65.332 54.746
430662 92.863 -65.335 54.745
430663 92.910 -65.338 54.744
430664 92.958 -65.341 54.740
430665 93.005 -65.344 54.735
430666 93.053 -65.347 54.730
430667 93.100 -65.350 54.724
430668 93.149 -65.353 54.719
430669 93.199 -65.355 54.713
430670 93.249 -65.358 54.706
430671 93.296 -65.362 54.698
430672 93.344 -65.366 54.689
430673 93.392 -65.370 54.681
430674 93.435 -65.373 54.675
430675 93.478 -65.375 54.669
430676 93.520 -65.377 54.665
430677 93.563 -65.380 54.661
430678 93.606 -65.383 54.659
430679 93.648 -65.385 54.655
430680 93.691 -65.387 54.651
430681 93.735 -65.390 54.646
430682 93.780 -65.394 54.637
430683 93.824 -65.397 54.623
430684 93.868 -65.400 54.606
430685 93.913 -65.403 54.588
430686 93.947 -65.400 54.570
430687 93.980 -65.395 54.552
430688 94.013 -65.391 54.537
430689 94.047 -65.386 54.523
430690 94.079 -65.382 54.508
430691 94.113 -65.378 54.492
430692 94.145 -65.373 54.472
430693 94.179 -65.371 54.452
430694 94.214 -65.373 54.434
430695 94.249 -65.374 54.417
430696 94.283 -65.376 54.403
430697 94.318 -65.377 54.390
430698 94.352 -65.379 54.376
430699 94.388 -65.380 54.364
430700 94.427 -65.382 54.353
430701 94.467 -65.383 54.342
430702 94.506 -65.384 54.327
430703 94.545 -65.386 54.311
430704 94.584 -65.387 54.294
430705 94.624 -65.389 54.274
430706 94.664 -65.390 54.250
430707 94.703 -65.393 54.224
430708 94.743 -65.397 54.193
430709 94.784 -65.400 54.158
430710 94.827 -65.401 54.123
430711 94.871 -65.405 54.094
430712 94.915 -65.408 54.068
430713 94.959 -65.411 54.042
430714 95.003 -65.415 54.015
430715 95.047 -65.419 53.986
430716 95.091 -65.422 53.957
430717 95.135 -65.426 53.928
430718 95.181 -65.434 53.901
430719 95.228 -65.444 53.875
430720 95.274 -65.455 53.850
430721 95.316 -65.452 53.825
430722 95.355 -65.445 53.804
430723 95.395 -65.440 53.786
430724 95.435 -65.435 53.771
430725 95.474 -65.429 53.758
430726 95.514 -65.430 53.744
430727 95.555 -65.434 53.726
430728 95.595 -65.439 53.704
430729 95.635 -65.444 53.677
430730 95.676 -65.449 53.649
430731 95.716 -65.455 53.619
430732 95.756 -65.459 53.591
430733 95.797 -65.464 53.563
430734 95.837 -65.470 53.534
430735 95.877 -65.474 53.503
430736 95.917 -65.480 53.470
430737 95.957 -65.485 53.435
430738 95.999 -65.483 53.399
430739 96.040 -65.479 53.365
430740 96.082 -65.474 53.328
430741 96.126 -65.470 53.290
430742 96.170 -65.467 53.252
430743 96.214 -65.463 53.214
430744 96.257 -65.459 53.178
430745 96.301 -65.455 53.142
430746 96.345 -65.452 53.104
430747 96.389 -65.449 53.066
430748 96.432 -65.447 53.028
430749 96.470 -65.451 52.990
430750 96.507 -65.455 52.951
430751 96.545 -65.460 52.911
430752 96.586 -65.461 52.870
430753 96.627 -65.461 52.827
430754 96.669 -65.461 52.783
430755 96.711 -65.461 52.741
430756 96.752 -65.461 52.699
430757 96.794 -65.461 52.657
430758 96.836 -65.461 52.616
430759 96.879 -65.463 52.573
430760 96.922 -65.464 52.530
430761 96.964 -65.464 52.488
430762 97.007 -65.465 52.447
430763 97.049 -65.467 52.406
430764 97.085 -65.471 52.368
430765 97.115 -65.480 52.330
430766 97.146 -65.488 52.291
430767 97.176 -65.496 52.255
430768 97.204 -65.492 52.218
430769 97.196 -65.484 52.182
430770 97.211 -65.499 52.145
430771 97.237 -65.516 52.108
430772 97.277 -65.524 52.069
430773 97.317 -65.533 52.031
430774 97.357 -65.541 51.990
430775 97.397 -65.549 51.953
430776 97.437 -65.557 51.921
430777 97.476 -65.565 51.890
430778 97.516 -65.573 51.856
430779 97.552 -65.575 51.820
430780 97.588 -65.577 51.783
430781 97.624 -65.579 51.745
430782 97.660 -65.581 51.707
430783 97.696 -65.583 51.672
430784 97.731 -65.585 51.637
430785 97.768 -65.587 51.602
430786 97.803 -65.590 51.567
430787 97.839 -65.592 51.532
430788 97.872 -65.608 51.497
430789 97.904 -65.628 51.464
430790 97.939 -65.635 51.431
430791 97.974 -65.640 51.397
430792 98.010 -65.646 51.367
430793 98.045 -65.651 51.336
430794 98.080 -65.656 51.303
430795 98.115 -65.662 51.269
430796 98.151 -65.666 51.234
430797 98.186 -65.671 51.194
430798 98.221 -65.677 51.152
430799 98.256 -65.682 51.109
430800 98.294 -65.690 51.065
430801 98.332 -65.696 51.021
430802 98.369 -65.700 50.979
430803 98.407 -65.704 50.936
430804 98.444 -65.709 50.893
430805 98.482 -65.714 50.847
430806 98.519 -65.719 50.801
430807 98.557 -65.724 50.751
430808 98.595 -65.730 50.701
430809 98.638 -65.735 50.649
430810 98.681 -65.741 50.592
430811 98.724 -65.746 50.534
430812 98.767 -65.753 50.475
430813 98.809 -65.758 50.417
430814 98.853 -65.764 50.359
430815 98.896 -65.769 50.301
430816 98.938 -65.775 50.243
430817 98.982 -65.780 50.186
430818 99.021 -65.785 50.130
430819 99.062 -65.789 50.073
430820 99.102 -65.793 50.020
430821 99.142 -65.796 49.967
430822 99.183 -65.801 49.917
430823 99.222 -65.805 49.869
430824 99.267 -65.812 49.821
430825 99.312 -65.821 49.775
430826 99.357 -65.828 49.730
430827 99.402 -65.837 49.686
430828 99.446 -65.844 49.644
430829 99.491 -65.853 49.603
430830 99.540 -65.860 49.559
430831 99.591 -65.871 49.510
430832 99.640 -65.878 49.458
430833 99.689 -65.886 49.406
430834 99.737 -65.893 49.353
430835 99.787 -65.900 49.303
430836 99.836 -65.908 49.253
430837 99.885 -65.916 49.205
430838 99.927 -65.913 49.157
430839 99.947 -65.878 49.107
430840 99.974 -65.850 49.057
430841 100.013 -65.837 49.008
430842 100.052 -65.824 48.959
430843 100.099 -65.820 48.912
430844 100.147 -65.818 48.866
430845 100.195 -65.816 48.819
430846 100.244 -65.814 48.772
430847 100.291 -65.812 48.727
430848 100.340 -65.810 48.684
430849 100.388 -65.808 48.640
430850 100.436 -65.806 48.596
430851 100.484 -65.803 48.553
431265 88.301 -64.549 55.301
431266 88.338 -64.557 55.299
431267 88.384 -64.562 55.298
431268 88.430 -64.565 55.299
431269 88.477 -64.569 55.302
431270 88.523 -64.573 55.307
431271 88.569 -64.577 55.314
431272 88.615 -64.580 55.321
431273 88.662 -64.585 55.331
431274 88.707 -64.589 55.342
431275 88.754 -64.593 55.358
431276 88.800 -64.596 55.377
431277 88.846 -64.600 55.390
431278 88.892 -64.604 55.394
431279 88.939 -64.609 55.389
431280 88.984 -64.612 55.380
431281 89.047 -64.608 55.370
431282 89.110 -64.604 55.359
431283 89.173 -64.599 55.347
431284 89.221 -64.609 55.332
431285 89.257 -64.627 55.319
431286 89.295 -64.646 55.309
431287 89.331 -64.665 55.300
431288 89.368 -64.684 55.289
431289 89.405 -64.702 55.277
431290 89.442 -64.720 55.261
431291 89.489 -64.722 55.238
431292 89.537 -64.725 55.209
431293 89.584 -64.727 55.176
431294 89.631 -64.729 55.144
431295 89.679 -64.731 55.118
431296 89.726 -64.734 55.094
431297 89.773 -64.736 55.072
431298 89.821 -64.738 55.052
431299 89.868 -64.740 55.035
431300 89.915 -64.743 55.020
431301 89.964 -64.748 55.008
431302 90.012 -64.754 54.999
431303 90.060 -64.761 54.992
431304 90.109 -64.767 54.988
431305 90.157 -64.773 54.985
431306 90.206 -64.779 54.984
431307 90.254 -64.785 54.984
431308 90.303 -64.791 54.984
431309 90.351 -64.797 54.984
431310 90.399 -64.804 54.985
431311 90.448 -64.810 54.986
431312 90.497 -64.816 54.986
431313 90.545 -64.822 54.986
431314 90.593 -64.828 54.985
431315 90.642 -64.834 54.982
431316 90.690 -64.841 54.978
431317 90.738 -64.849 54.973
431318 90.784 -64.860 54.966
431319 90.828 -64.872 54.958
431320 90.871 -64.882 54.949
431321 90.914 -64.893 54.940
431322 90.957 -64.903 54.930
431323 90.998 -64.919 54.919
431324 91.037 -64.935 54.908
431325 91.077 -64.952 54.896
431326 91.117 -64.969 54.885
431327 91.157 -64.986 54.873
431328 91.196 -65.003 54.862
431329 91.236 -65.019 54.850
431330 91.276 -65.037 54.838
431331 91.316 -65.054 54.829
431332 91.356 -65.070 54.820
431333 91.395 -65.087 54.812
431334 91.435 -65.104 54.805
431335 91.475 -65.121 54.800
431336 91.515 -65.138 54.797
431337 91.559 -65.149 54.796
431338 91.604 -65.159 54.796
431339 91.649 -65.169 54.797
431340 91.694 -65.179 54.800
431341 91.739 -65.189 54.807
431342 91.785 -65.199 54.816
431343 91.829 -65.209 54.821
431344 91.875 -65.219 54.821
431345 91.920 -65.229 54.814
431346 91.965 -65.239 54.802
431347 92.002 -65.258 54.789
431348 92.049 -65.267 54.776
431349 92.107 -65.264 54.765
431350 92.165 -65.261 54.757
431351 92.223 -65.259 54.752
431352 92.280 -65.258 54.747
431353 92.337 -65.260 54.743
431354 92.394 -65.261 54.740
431355 92.451 -65.262 54.739
431356 92.508 -65.263 54.737
431357 92.565 -65.264 54.735
431358 92.618 -65.267 54.734
431359 92.669 -65.269 54.732
431360 92.719 -65.274 54.729
431361 92.768 -65.279 54.726
431362 92.818 -65.283 54.724
431363 92.866 -65.287 54.723
431364 92.913 -65.290 54.723
431365 92.961 -65.293 54.721
431366 93.008 -65.296 54.717
431367 93.056 -65.299 54.713
431368 93.103 -65.302 54.709
431369 93.151 -65.305 54.706
431370 93.199 -65.308 54.701
431371 93.246 -65.311 54.695
431372 93.293 -65.316 54.686
431373 93.341 -65.320 54.677
431374 93.389 -65.324 54.669
431375 93.435 -65.328 54.662
431376 93.478 -65.330 54.657
431377 93.520 -65.333 54.652
431378 93.563 -65.335 54.647
431379 93.606 -65.337 54.646
431380 93.648 -65.340 54.643
431381 93.693 -65.343 54.640
431382 93.737 -65.346 54.636
431383 93.781 -65.349 54.630
431384 93.826 -65.353 54.619
431385 93.870 -65.356 54.606
431386 93.914 -65.359 54.591
431387 93.951 -65.358 54.575
431388 93.984 -65.353 54.561
431389 94.017 -65.348 54.549
431390 94.050 -65.344 54.538
431391 94.083 -65.339 54.524
431392 94.116 -65.335 54.509
431393 94.149 -65.330 54.489
431394 94.184 -65.332 54.468
431395 94.218 -65.333 54.448
431396 94.253 -65.334 54.428
431397 94.288 -65.336 54.410
431398 94.323 -65.337 54.396
431399 94.357 -65.339 54.381
431400 94.392 -65.340 54.367
431401 94.430 -65.342 54.354
431402 94.470 -65.343 54.343
431403 94.509 -65.345 54.327
431404 94.549 -65.346 54.310
431405 94.588 -65.348 54.292
431406 94.627 -65.349 54.271
431407 94.667 -65.351 54.247
431408 94.706 -65.352 54.222
431409 94.746 -65.355 54.192
431410 94.788 -65.358 54.159
431411 94.832 -65.361 54.125
431412 94.877 -65.364 54.095
431413 94.920 -65.368 54.068
431414 94.964 -65.372 54.042
431415 95.008 -65.375 54.016
431416 95.052 -65.379 53.988
431417 95.096 -65.383 53.961
431418 95.140 -65.386 53.935
431419 95.186 -65.394 53.909
431420 95.233 -65.404 53.885
431421 95.275 -65.404 53.861
431422 95.315 -65.398 53.837
431423 95.354 -65.392 53.817
431424 95.394 -65.389 53.799
431425 95.433 -65.393 53.783
431426 95.472 -65.398 53.769
431427 95.511 -65.404 53.753
431428 95.557 -65.404 53.735
431429 95.597 -65.409 53.715
431430 95.637 -65.414 53.690
431431 95.678 -65.419 53.664
431432 95.718 -65.424 53.632
431433 95.758 -65.430 53.600
431434 95.798 -65.435 53.568
431435 95.838 -65.439 53.537
431436 95.879 -65.445 53.505
431437 95.919 -65.450 53.471
431438 95.960 -65.451 53.435
431439 96.002 -65.446 53.398
431440 96.044 -65.442 53.363
431441 96.086 -65.437 53.324
431442 96.129 -65.433 53.286
431443 96.173 -65.429 53.248
431444 96.217 -65.425 53.209
431445 96.260 -65.421 53.172
431446 96.304 -65.417 53.136
431447 96.348 -65.414 53.098
431448 96.391 -65.412 53.060
431449 96.433 -65.412 53.023
431450 96.471 -65.416 52.986
431451 96.508 -65.420 52.947
431452 96.546 -65.425 52.907
431453 96.583 -65.430 52.867
431454 96.623 -65.431 52.825
431455 96.665 -65.432 52.782
431456 96.706 -65.432 52.739
431457 96.748 -65.432 52.696
431458 96.790 -65.432 52.653
431459 96.832 -65.433 52.612
431460 96.875 -65.433 52.569
431461 96.917 -65.435 52.525
431462 96.960 -65.436 52.483
431463 97.002 -65.436 52.441
431464 97.045 -65.437 52.398
431465 97.087 -65.439 52.358
431466 97.130 -65.439 52.318
431467 97.173 -65.440 52.278
431468 97.204 -65.448 52.240
431469 97.197 -65.440 52.201
431470 97.196 -65.445 52.164
431471 97.217 -65.465 52.128
431472 97.243 -65.482 52.092
431473 97.283 -65.490 52.054
431474 97.323 -65.499 52.016
431475 97.363 -65.506 51.976
431476 97.402 -65.514 51.938
431477 97.442 -65.523 51.905
431478 97.483 -65.531 51.873
431479 97.522 -65.540 51.838
431480 97.559 -65.546 51.801
431481 97.593 -65.549 51.763
431482 97.627 -65.552 51.725
431483 97.660 -65.555 51.687
431484 97.695 -65.558 51.653
431485 97.731 -65.560 51.618
431486 97.767 -65.562 51.582
431487 97.803 -65.564 51.546
431488 97.839 -65.567 51.511
431489 97.871 -65.585 51.477
431490 97.907 -65.590 51.444
431491 97.942 -65.596 51.410
431492 97.977 -65.601 51.376
431493 98.012 -65.606 51.344
431494 98.048 -65.612 51.312
431495 98.083 -65.617 51.280
431496 98.118 -65.622 51.247
431497 98.153 -65.627 51.214
431498 98.189 -65.633 51.177
431499 98.224 -65.638 51.137
431500 98.259 -65.643 51.097
431501 98.294 -65.648 51.055
431502 98.332 -65.654 51.013
431503 98.369 -65.659 50.971
431504 98.407 -65.664 50.928
431505 98.444 -65.669 50.885
431506 98.482 -65.674 50.840
431507 98.519 -65.680 50.793
431508 98.557 -65.685 50.743
431509 98.596 -65.690 50.692
431510 98.640 -65.696 50.637
431511 98.684 -65.702 50.580
431512 98.726 -65.707 50.521
431513 98.769 -65.713 50.462
431514 98.812 -65.718 50.405
431515 98.855 -65.724 50.349
431516 98.898 -65.730 50.291
431517 98.941 -65.736 50.236
431518 98.983 -65.741 50.179
431519 99.026 -65.745 50.121
431520 99.066 -65.749 50.065
431521 99.106 -65.754 50.011
431522 99.146 -65.758 49.960
431523 99.186 -65.762 49.910
431524 99.226 -65.766 49.862
431525 99.269 -65.772 49.814
431526 99.314 -65.781 49.768
431527 99.359 -65.788 49.723
431528 99.407 -65.796 49.678
431529 99.457 -65.802 49.633
431530 99.507 -65.808 49.590
431531 99.558 -65.814 49.546
431532 99.609 -65.824 49.498
431533 99.659 -65.834 49.448
431534 99.709 -65.844 49.397
431535 99.760 -65.855 49.347
431536 99.810 -65.865 49.298
431537 99.860 -65.875 49.250
431538 99.909 -65.885 49.202
431539 99.940 -65.866 49.154
431540 99.961 -65.831 49.103
431541 99.987 -65.802 49.053
431542 100.026 -65.790 49.004
431543 100.065 -65.777 48.956
431544 100.113 -65.774 48.908
431545 100.161 -65.772 48.862
431546 100.209 -65.770 48.815
431547 100.258 -65.768 48.769
431548 100.306 -65.766 48.724
431549 100.354 -65.764 48.680
431550 100.403 -65.762 48.637
431551 100.451 -65.759 48.594
431552 100.496 -65.759 48.551
431967 88.340 -64.504 55.299
431968 88.387 -64.508 55.299
431969 88.433 -64.512 55.299
431970 88.479 -64.516 55.301
431971 88.525 -64.520 55.308
431972 88.571 -64.524 55.314
431973 88.617 -64.527 55.322
431974 88.664 -64.531 55.331
431975 88.710 -64.536 55.342
431976 88.756 -64.540 55.357
431977 88.802 -64.543 55.377
431978 88.848 -64.547 55.393
431979 88.895 -64.551 55.398
431980 88.947 -64.552 55.396
431981 89.010 -64.548 55.387
431982 89.073 -64.543 55.376
431983 89.136 -64.539 55.362
431984 89.188 -64.544 55.348
431985 89.225 -64.562 55.334
431986 89.262 -64.581 55.324
431987 89.299 -64.599 55.316
431988 89.336 -64.618 55.309
431989 89.373 -64.637 55.300
431990 89.410 -64.656 55.287
431991 89.447 -64.674 55.270
431992 89.490 -64.682 55.245
431993 89.538 -64.684 55.215
431994 89.585 -64.687 55.181
431995 89.632 -64.689 55.151
431996 89.680 -64.691 55.124
431997 89.727 -64.693 55.100
431998 89.774 -64.695 55.078
431999 89.822 -64.697 55.058
432000 89.869 -64.700 55.042
432001 89.916 -64.702 55.028
432002 89.964 -64.704 55.016
432003 90.011 -64.706 55.006
432004 90.059 -64.712 55.000
432005 90.108 -64.718 54.996
432006 90.156 -64.724 54.992
432007 90.204 -64.730 54.989
432008 90.253 -64.737 54.990
432009 90.302 -64.743 54.989
432010 90.350 -64.749 54.990
432011 90.398 -64.755 54.991
432012 90.447 -64.761 54.991
432013 90.496 -64.767 54.991
432014 90.544 -64.773 54.990
432015 90.592 -64.780 54.988
432016 90.641 -64.786 54.985
432017 90.689 -64.792 54.980
432018 90.736 -64.802 54.975
432019 90.781 -64.813 54.967
432020 90.824 -64.824 54.958
432021 90.867 -64.834 54.949
432022 90.910 -64.846 54.938
432023 90.950 -64.862 54.928
432024 90.990 -64.878 54.917
432025 91.030 -64.895 54.905
432026 91.070 -64.912 54.894
432027 91.109 -64.929 54.882
432028 91.149 -64.946 54.870
432029 91.189 -64.963 54.858
432030 91.229 -64.979 54.846
432031 91.269 -64.996 54.834
432032 91.308 -65.013 54.824
432033 91.348 -65.030 54.814
432034 91.388 -65.046 54.805
432035 91.433 -65.056 54.798
432036 91.478 -65.066 54.793
432037 91.523 -65.076 54.789
432038 91.569 -65.086 54.787
432039 91.614 -65.096 54.786
432040 91.659 -65.107 54.786
432041 91.704 -65.116 54.786
432042 91.749 -65.126 54.792
432043 91.794 -65.136 54.799
432044 91.840 -65.146 54.805
432045 91.884 -65.157 54.807
432046 91.930 -65.166 54.801
432047 91.975 -65.176 54.790
432048 92.020 -65.186 54.777
432049 92.045 -65.220 54.765
432050 92.103 -65.217 54.752
432051 92.161 -65.213 54.743
432052 92.219 -65.210 54.737
432053 92.276 -65.210 54.733
432054 92.333 -65.212 54.729
432055 92.390 -65.213 54.726
432056 92.447 -65.214 54.724
432057 92.504 -65.215 54.722
432058 92.561 -65.217 54.720
432059 92.618 -65.217 54.718
432060 92.672 -65.220 54.714
432061 92.722 -65.226 54.709
432062 92.771 -65.230 54.706
432063 92.820 -65.235 54.703
432064 92.868 -65.239 54.702
432065 92.916 -65.242 54.701
432066 92.964 -65.245 54.699
432067 93.011 -65.248 54.696
432068 93.059 -65.251 54.693
432069 93.107 -65.254 54.690
432070 93.154 -65.257 54.688
432071 93.200 -65.261 54.685
432072 93.245 -65.266 54.681
432073 93.291 -65.270 54.674
432074 93.337 -65.273 54.666
432075 93.385 -65.278 54.658
432076 93.433 -65.282 54.652
432077 93.477 -65.285 54.648
432078 93.520 -65.287 54.645
432079 93.563 -65.290 54.641
432080 93.606 -65.292 54.640
432081 93.649 -65.295 54.638
432082 93.694 -65.298 54.636
432083 93.738 -65.302 54.633
432084 93.782 -65.305 54.628
432085 93.827 -65.308 54.619
432086 93.871 -65.311 54.609
432087 93.915 -65.315 54.597
432088 93.955 -65.315 54.583
432089 93.988 -65.311 54.572
432090 94.021 -65.306 54.563
432091 94.054 -65.301 54.552
432092 94.087 -65.297 54.538
432093 94.120 -65.292 54.522
432094 94.153 -65.290 54.503
432095 94.188 -65.292 54.481
432096 94.223 -65.293 54.459
432097 94.258 -65.295 54.437
432098 94.292 -65.296 54.416
432099 94.327 -65.298 54.400
432100 94.361 -65.299 54.383
432101 94.396 -65.301 54.368
432102 94.434 -65.302 54.354
432103 94.473 -65.304 54.342
432104 94.512 -65.305 54.325
432105 94.552 -65.307 54.308
432106 94.592 -65.308 54.289
432107 94.631 -65.310 54.267
432108 94.670 -65.311 54.244
432109 94.710 -65.313 54.218
432110 94.751 -65.315 54.189
432111 94.795 -65.319 54.157
432112 94.838 -65.322 54.125
432113 94.882 -65.325 54.094
432114 94.925 -65.328 54.068
432115 94.970 -65.332 54.043
432116 95.014 -65.336 54.019
432117 95.057 -65.339 53.993
432118 95.101 -65.343 53.967
432119 95.145 -65.346 53.942
432120 95.191 -65.353 53.916
432121 95.234 -65.356 53.892
432122 95.274 -65.350 53.868
432123 95.313 -65.351 53.844
432124 95.352 -65.357 53.823
432125 95.391 -65.361 53.804
432126 95.430 -65.367 53.786
432127 95.470 -65.372 53.771
432128 95.511 -65.376 53.755
432129 95.558 -65.374 53.738
432130 95.599 -65.380 53.719
432131 95.639 -65.384 53.697
432132 95.680 -65.389 53.670
432133 95.720 -65.395 53.641
432134 95.760 -65.399 53.609
432135 95.800 -65.405 53.575
432136 95.840 -65.410 53.539
432137 95.881 -65.414 53.505
432138 95.921 -65.417 53.471
432139 95.963 -65.413 53.435
432140 96.005 -65.408 53.399
432141 96.047 -65.404 53.363
432142 96.089 -65.399 53.325
432143 96.131 -65.395 53.286
432144 96.175 -65.392 53.247
432145 96.220 -65.387 53.206
432146 96.263 -65.383 53.167
432147 96.307 -65.380 53.131
432148 96.350 -65.378 53.093
432149 96.393 -65.376 53.055
432150 96.434 -65.377 53.019
432151 96.472 -65.382 52.981
432152 96.509 -65.386 52.943
432153 96.547 -65.390 52.904
432154 96.584 -65.395 52.864
432155 96.622 -65.398 52.822
432156 96.660 -65.402 52.781
432157 96.702 -65.402 52.738
432158 96.743 -65.403 52.695
432159 96.785 -65.403 52.652
432160 96.827 -65.404 52.611
432161 96.870 -65.405 52.567
432162 96.913 -65.405 52.523
432163 96.955 -65.407 52.480
432164 96.998 -65.408 52.436
432165 97.040 -65.408 52.394
432166 97.083 -65.409 52.351
432167 97.125 -65.410 52.308
432168 97.167 -65.410 52.266
432169 97.191 -65.402 52.226
432170 97.181 -65.390 52.185
432171 97.202 -65.411 52.147
432172 97.223 -65.430 52.111
432173 97.248 -65.448 52.075
432174 97.289 -65.456 52.038
432175 97.328 -65.464 52.001
432176 97.368 -65.472 51.961
432177 97.408 -65.480 51.924
432178 97.448 -65.489 51.889
432179 97.488 -65.498 51.856
432180 97.528 -65.507 51.820
432181 97.568 -65.516 51.782
432182 97.602 -65.519 51.743
432183 97.636 -65.522 51.705
432184 97.669 -65.525 51.668
432185 97.704 -65.528 51.634
432186 97.737 -65.532 51.599
432187 97.771 -65.535 51.562
432188 97.805 -65.538 51.527
432189 97.838 -65.541 51.492
432190 97.873 -65.546 51.459
432191 97.909 -65.551 51.425
432192 97.943 -65.556 51.391
432193 97.979 -65.561 51.357
432194 98.013 -65.567 51.323
432195 98.049 -65.572 51.289
432196 98.084 -65.577 51.257
432197 98.119 -65.583 51.224
432198 98.154 -65.588 51.193
432199 98.189 -65.593 51.157
432200 98.224 -65.598 51.120
432201 98.259 -65.603 51.082
432202 98.294 -65.608 51.043
432203 98.332 -65.615 51.002
432204 98.369 -65.620 50.960
432205 98.407 -65.625 50.918
432206 98.444 -65.630 50.875
432207 98.482 -65.636 50.831
432208 98.519 -65.641 50.783
432209 98.557 -65.646 50.733
432210 98.598 -65.651 50.680
432211 98.643 -65.656 50.625
432212 98.686 -65.662 50.567
432213 98.729 -65.668 50.508
432214 98.772 -65.674 50.449
432215 98.814 -65.679 50.393
432216 98.857 -65.685 50.337
432217 98.900 -65.691 50.280
432218 98.942 -65.697 50.225
432219 98.986 -65.701 50.168
432220 99.030 -65.704 50.111
432221 99.070 -65.709 50.054
432222 99.110 -65.713 50.001
432223 99.150 -65.718 49.951
432224 99.190 -65.722 49.903
432225 99.230 -65.727 49.855
432226 99.273 -65.732 49.809
432227 99.323 -65.738 49.762
432228 99.373 -65.744 49.717
432229 99.424 -65.750 49.670
432230 99.475 -65.756 49.623
432231 99.525 -65.762 49.579
432232 99.575 -65.768 49.534
432233 99.626 -65.777 49.488
432234 99.677 -65.787 49.439
432235 99.727 -65.797 49.390
432236 99.777 -65.808 49.341
432237 99.826 -65.818 49.293
432238 99.870 -65.828 49.246
432239 99.919 -65.834 49.199
432240 99.954 -65.818 49.150
432241 99.974 -65.783 49.099
432242 100.001 -65.755 49.049
432243 100.040 -65.743 49.000
432244 100.079 -65.731 48.952
432245 100.127 -65.728 48.904
432246 100.176 -65.726 48.857
432247 100.224 -65.724 48.812
432248 100.272 -65.722 48.765
432249 100.321 -65.720 48.720
432250 100.368 -65.718 48.676
432251 100.417 -65.715 48.634
432252 100.465 -65.714 48.591
432253 100.506 -65.717 48.548
432668 88.343 -64.451 55.300
432669 88.389 -64.455 55.299
432670 88.435 -64.458 55.300
432671 88.481 -64.462 55.302
432672 88.528 -64.467 55.308
432673 88.573 -64.471 55.314
432674 88.620 -64.474 55.322
432675 88.666 -64.478 55.330
432676 88.712 -64.482 55.340
432677 88.758 -64.486 55.353
432678 88.805 -64.490 55.370
432679 88.850 -64.494 55.386
432680 88.909 -64.492 55.394
432681 88.973 -64.487 55.393
432682 89.036 -64.483 55.387
432683 89.099 -64.479 55.375
432684 89.156 -64.479 55.361
432685 89.193 -64.498 55.348
432686 89.230 -64.516 55.335
432687 89.267 -64.535 55.326
432688 89.304 -64.553 55.320
432689 89.341 -64.572 55.315
432690 89.378 -64.590 55.308
432691 89.414 -64.609 55.295
432692 89.452 -64.627 55.275
432693 89.492 -64.641 55.249
432694 89.539 -64.643 55.217
432695 89.587 -64.646 55.184
432696 89.634 -64.648 55.155
432697 89.681 -64.650 55.130
432698 89.729 -64.653 55.106
432699 89.776 -64.655 55.084
432700 89.823 -64.657 55.065
432701 89.871 -64.659 55.049
432702 89.917 -64.662 55.035
432703 89.965 -64.664 55.024
432704 90.012 -64.666 55.015
432705 90.059 -64.668 55.009
432706 90.107 -64.671 55.004
432707 90.155 -64.675 55.001
432708 90.203 -64.681 54.998
432709 90.252 -64.687 54.998
432710 90.300 -64.693 54.998
432711 90.349 -64.700 54.998
432712 90.397 -64.706 54.999
432713 90.446 -64.712 54.998
432714 90.494 -64.718 54.997
432715 90.542 -64.725 54.996
432716 90.591 -64.731 54.993
432717 90.640 -64.737 54.989
432718 90.688 -64.743 54.983
432719 90.734 -64.755 54.977
432720 90.777 -64.765 54.969
432721 90.820 -64.776 54.960
432722 90.863 -64.787 54.949
432723 90.902 -64.804 54.937
432724 90.942 -64.822 54.927
432725 90.982 -64.838 54.915
432726 91.022 -64.855 54.903
432727 91.062 -64.872 54.891
432728 91.101 -64.888 54.879
432729 91.141 -64.906 54.867
432730 91.181 -64.922 54.854
432731 91.221 -64.939 54.842
432732 91.262 -64.953 54.830
432733 91.308 -64.963 54.818
432734 91.353 -64.973 54.808
432735 91.398 -64.984 54.799
432736 91.443 -64.994 54.791
432737 91.488 -65.004 54.785
432738 91.533 -65.013 54.780
432739 91.579 -65.024 54.777
432740 91.624 -65.034 54.774
432741 91.669 -65.044 54.773
432742 91.714 -65.054 54.772
432743 91.759 -65.063 54.775
432744 91.804 -65.074 54.779
432745 91.849 -65.084 54.783
432746 91.894 -65.094 54.785
432747 91.940 -65.104 54.781
432748 91.985 -65.114 54.772
432749 92.030 -65.124 54.761
432750 92.064 -65.146 54.750
432751 92.099 -65.168 54.737
432752 92.157 -65.166 54.727
432753 92.215 -65.163 54.722
432754 92.272 -65.163 54.718
432755 92.329 -65.164 54.715
432756 92.386 -65.166 54.712
432757 92.443 -65.166 54.710
432758 92.500 -65.168 54.707
432759 92.562 -65.168 54.705
432760 92.620 -65.169 54.701
432761 92.672 -65.172 54.697
432762 92.724 -65.177 54.691
432763 92.774 -65.182 54.686
432764 92.823 -65.186 54.683
432765 92.871 -65.190 54.681
432766 92.919 -65.193 54.680
432767 92.967 -65.196 54.677
432768 93.014 -65.199 54.675
432769 93.062 -65.202 54.673
432770 93.109 -65.206 54.671
432771 93.154 -65.210 54.670
432772 93.199 -65.215 54.668
432773 93.245 -65.220 54.664
432774 93.290 -65.224 54.659
432775 93.337 -65.228 54.652
432776 93.382 -65.232 54.646
432777 93.429 -65.235 54.642
432778 93.477 -65.240 54.640
432779 93.519 -65.242 54.639
432780 93.562 -65.245 54.637
432781 93.607 -65.248 54.635
432782 93.651 -65.251 54.634
432783 93.695 -65.254 54.632
432784 93.740 -65.257 54.630
432785 93.784 -65.261 54.625
432786 93.828 -65.264 54.618
432787 93.872 -65.267 54.611
432788 93.917 -65.270 54.602
432789 93.959 -65.272 54.592
432790 93.992 -65.267 54.581
432791 94.025 -65.263 54.572
432792 94.058 -65.259 54.561
432793 94.090 -65.254 54.547
432794 94.124 -65.250 54.530
432795 94.158 -65.251 54.510
432796 94.193 -65.252 54.489
432797 94.227 -65.254 54.466
432798 94.262 -65.255 54.443
432799 94.297 -65.257 54.420
432800 94.332 -65.258 54.401
432801 94.366 -65.260 54.384
432802 94.401 -65.261 54.368
432803 94.437 -65.263 54.353
432804 94.477 -65.264 54.340
432805 94.516 -65.266 54.323
432806 94.555 -65.267 54.304
432807 94.595 -65.269 54.285
432808 94.634 -65.270 54.263
432809 94.674 -65.272 54.240
432810 94.714 -65.274 54.213
432811 94.759 -65.278 54.184
432812 94.803 -65.281 54.154
432813 94.846 -65.285 54.123
432814 94.890 -65.288 54.094
432815 94.933 -65.291 54.069
432816 94.977 -65.294 54.045
432817 95.020 -65.297 54.022
432818 95.063 -65.301 53.997
432819 95.107 -65.304 53.973
432820 95.150 -65.307 53.948
432821 95.193 -65.310 53.922
432822 95.233 -65.315 53.897
432823 95.271 -65.320 53.872
432824 95.311 -65.325 53.849
432825 95.350 -65.330 53.826
432826 95.389 -65.335 53.805
432827 95.428 -65.340 53.786
432828 95.467 -65.345 53.769
432829 95.510 -65.348 53.753
432830 95.558 -65.347 53.736
432831 95.601 -65.349 53.719
432832 95.641 -65.355 53.701
432833 95.681 -65.360 53.674
432834 95.722 -65.364 53.645
432835 95.762 -65.370 53.613
432836 95.802 -65.375 53.578
432837 95.842 -65.380 53.540
432838 95.883 -65.384 53.504
432839 95.924 -65.380 53.469
432840 95.966 -65.375 53.433
432841 96.008 -65.370 53.397
432842 96.050 -65.366 53.362
432843 96.092 -65.361 53.325
432844 96.134 -65.358 53.286
432845 96.178 -65.354 53.246
432846 96.222 -65.350 53.204
432847 96.266 -65.346 53.163
432848 96.309 -65.344 53.125
432849 96.352 -65.342 53.087
432850 96.395 -65.341 53.050
432851 96.435 -65.342 53.014
432852 96.473 -65.347 52.978
432853 96.511 -65.351 52.940
432854 96.548 -65.355 52.901
432855 96.586 -65.360 52.861
432856 96.623 -65.364 52.819
432857 96.661 -65.368 52.778
432858 96.698 -65.373 52.736
432859 96.739 -65.373 52.694
432860 96.781 -65.373 52.651
432861 96.823 -65.375 52.609
432862 96.866 -65.376 52.566
432863 96.908 -65.376 52.522
432864 96.951 -65.377 52.478
432865 96.993 -65.377 52.435
432866 97.034 -65.377 52.391
432867 97.076 -65.376 52.349
432868 97.118 -65.376 52.305
432869 97.160 -65.376 52.259
432870 97.162 -65.359 52.215
432871 97.184 -65.358 52.172
432872 97.208 -65.376 52.131
432873 97.229 -65.396 52.094
432874 97.254 -65.414 52.057
432875 97.294 -65.422 52.021
432876 97.334 -65.430 51.985
432877 97.374 -65.438 51.946
432878 97.414 -65.446 51.909
432879 97.454 -65.455 51.874
432880 97.494 -65.464 51.842
432881 97.534 -65.474 51.803
432882 97.574 -65.483 51.764
432883 97.611 -65.489 51.725
432884 97.645 -65.492 51.687
432885 97.678 -65.495 51.651
432886 97.712 -65.499 51.616
432887 97.746 -65.502 51.579
432888 97.780 -65.504 51.543
432889 97.812 -65.502 51.508
432890 97.845 -65.501 51.475
432891 97.873 -65.506 51.442
432892 97.907 -65.511 51.409
432893 97.942 -65.517 51.376
432894 97.977 -65.522 51.340
432895 98.012 -65.527 51.304
432896 98.047 -65.532 51.269
432897 98.082 -65.537 51.234
432898 98.117 -65.543 51.201
432899 98.152 -65.548 51.170
432900 98.187 -65.553 51.136
432901 98.222 -65.558 51.100
432902 98.257 -65.564 51.063
432903 98.295 -65.569 51.025
432904 98.332 -65.574 50.987
432905 98.369 -65.580 50.948
432906 98.407 -65.586 50.907
432907 98.444 -65.591 50.864
432908 98.482 -65.596 50.819
432909 98.519 -65.602 50.770
432910 98.557 -65.607 50.720
432911 98.601 -65.612 50.667
432912 98.645 -65.618 50.612
432913 98.688 -65.623 50.554
432914 98.731 -65.628 50.495
432915 98.774 -65.634 50.437
432916 98.817 -65.640 50.380
432917 98.859 -65.646 50.324
432918 98.902 -65.652 50.267
432919 98.943 -65.658 50.210
432920 98.988 -65.661 50.153
432921 99.034 -65.663 50.097
432922 99.074 -65.667 50.042
432923 99.115 -65.672 49.989
432924 99.154 -65.677 49.941
432925 99.194 -65.681 49.894
432926 99.234 -65.686 49.849
432927 99.278 -65.688 49.805
432928 99.332 -65.694 49.757
432929 99.386 -65.699 49.710
432930 99.440 -65.705 49.663
432931 99.492 -65.711 49.615
432932 99.543 -65.717 49.570
432933 99.593 -65.723 49.524
432934 99.643 -65.730 49.479
432935 99.694 -65.740 49.433
432936 99.743 -65.751 49.386
432937 99.787 -65.761 49.337
432938 99.831 -65.771 49.290
432939 99.874 -65.782 49.242
432940 99.937 -65.775 49.195
432941 99.967 -65.771 49.146
432942 99.988 -65.736 49.094
432943 100.014 -65.708 49.044
432944 100.053 -65.695 48.995
432945 100.093 -65.684 48.948
432946 100.142 -65.682 48.900
432947 100.190 -65.680 48.853
432948 100.239 -65.678 48.807
432949 100.286 -65.676 48.761
432950 100.335 -65.674 48.715
432951 100.383 -65.671 48.672
432952 100.431 -65.670 48.630
432953 100.477 -65.669 48.588
432954 100.516 -65.674 48.546
433369 88.345 -64.398 55.300
433370 88.391 -64.402 55.300
433371 88.437 -64.405 55.301
433372 88.484 -64.409 55.304
433373 88.530 -64.413 55.308
433374 88.576 -64.418 55.314
433375 88.622 -64.421 55.321
433376 88.668 -64.425 55.328
433377 88.714 -64.429 55.338
433378 88.761 -64.433 55.348
433379 88.809 -64.436 55.362
433380 88.872 -64.431 55.376
433381 88.935 -64.427 55.385
433382 88.998 -64.423 55.385
433383 89.061 -64.418 55.380
433384 89.124 -64.414 55.369
433385 89.161 -64.433 55.355
433386 89.198 -64.452 55.342
433387 89.235 -64.470 55.329
433388 89.272 -64.489 55.322
433389 89.309 -64.507 55.317
433390 89.345 -64.526 55.312
433391 89.382 -64.544 55.305
433392 89.419 -64.563 55.292
433393 89.456 -64.581 55.273
433394 89.493 -64.600 55.247
433395 89.540 -64.602 55.217
433396 89.588 -64.605 55.186
433397 89.635 -64.607 55.159
433398 89.682 -64.609 55.134
433399 89.730 -64.612 55.111
433400 89.777 -64.614 55.089
433401 89.824 -64.616 55.072
433402 89.872 -64.618 55.057
433403 89.919 -64.621 55.043
433404 89.966 -64.623 55.032
433405 90.014 -64.625 55.023
433406 90.061 -64.627 55.018
433407 90.108 -64.630 55.014
433408 90.156 -64.632 55.011
433409 90.203 -64.634 55.009
433410 90.251 -64.638 55.007
433411 90.299 -64.645 55.008
433412 90.347 -64.651 55.008
433413 90.396 -64.657 55.008
433414 90.445 -64.663 55.007
433415 90.493 -64.669 55.006
433416 90.541 -64.675 55.005
433417 90.590 -64.681 55.001
433418 90.638 -64.688 54.996
433419 90.686 -64.696 54.989
433420 90.730 -64.707 54.981
433421 90.773 -64.718 54.972
433422 90.815 -64.731 54.962
433423 90.855 -64.747 54.950
433424 90.895 -64.764 54.938
433425 90.935 -64.781 54.926
433426 90.974 -64.798 54.913
433427 91.014 -64.815 54.901
433428 91.054 -64.831 54.889
433429 91.094 -64.848 54.876
433430 91.137 -64.861 54.863
433431 91.182 -64.871 54.850
433432 91.227 -64.881 54.837
433433 91.273 -64.891 54.825
433434 91.317 -64.901 54.813
433435 91.363 -64.911 54.802
433436 91.408 -64.921 54.793
433437 91.453 -64.931 54.785
433438 91.498 -64.941 54.777
433439 91.543 -64.951 54.772
433440 91.588 -64.961 54.767
433441 91.634 -64.971 54.763
433442 91.678 -64.981 54.760
433443 91.724 -64.991 54.758
433444 91.769 -65.001 54.758
433445 91.814 -65.011 54.760
433446 91.859 -65.021 54.762
433447 91.905 -65.032 54.762
433448 91.949 -65.041 54.757
433449 91.995 -65.051 54.750
433450 92.040 -65.061 54.741
433451 92.085 -65.071 54.730
433452 92.104 -65.110 54.719
433453 92.153 -65.117 54.711
433454 92.211 -65.114 54.707
433455 92.268 -65.116 54.704
433456 92.325 -65.116 54.701
433457 92.382 -65.118 54.698
433458 92.441 -65.119 54.694
433459 92.507 -65.119 54.691
433460 92.569 -65.120 54.687
433461 92.618 -65.120 54.683
433462 92.669 -65.123 54.678
433463 92.724 -65.128 54.673
433464 92.777 -65.133 54.668
433465 92.826 -65.138 54.664
433466 92.874 -65.142 54.661
433467 92.922 -65.145 54.660
433468 92.970 -65.148 54.658
433469 93.017 -65.151 54.655
433470 93.063 -65.155 54.654
433471 93.108 -65.160 54.652
433472 93.153 -65.164 54.650
433473 93.199 -65.169 54.649
433474 93.245 -65.173 54.647
433475 93.290 -65.178 54.643
433476 93.336 -65.182 54.637
433477 93.382 -65.186 54.634
433478 93.428 -65.190 54.633
433479 93.474 -65.194 54.633
433480 93.519 -65.198 54.634
433481 93.564 -65.201 54.632
433482 93.608 -65.204 54.630
433483 93.652 -65.207 54.629
433484 93.697 -65.210 54.627
433485 93.741 -65.213 54.625
433486 93.785 -65.217 54.621
433487 93.830 -65.220 54.617
433488 93.874 -65.223 54.611
433489 93.918 -65.226 54.604
433490 93.963 -65.229 54.595
433491 93.995 -65.225 54.586
433492 94.029 -65.220 54.576
433493 94.061 -65.216 54.564
433494 94.095 -65.211 54.549
433495 94.128 -65.210 54.532
433496 94.163 -65.211 54.511
433497 94.197 -65.213 54.490
433498 94.232 -65.214 54.468
433499 94.267 -65.216 54.445
433500 94.301 -65.217 54.422
433501 94.336 -65.219 54.401
433502 94.371 -65.220 54.383
433503 94.405 -65.222 54.367
433504 94.441 -65.223 54.351
433505 94.480 -65.225 54.336
433506 94.520 -65.226 54.318
433507 94.559 -65.228 54.300
433508 94.598 -65.229 54.281
433509 94.638 -65.231 54.260
433510 94.678 -65.232 54.236
433511 94.722 -65.236 54.209
433512 94.766 -65.240 54.180
433513 94.811 -65.244 54.150
433514 94.854 -65.247 54.120
433515 94.898 -65.251 54.094
433516 94.941 -65.254 54.071
433517 94.984 -65.257 54.049
433518 95.028 -65.260 54.027
433519 95.071 -65.263 54.003
433520 95.115 -65.267 53.979
433521 95.158 -65.270 53.953
433522 95.195 -65.279 53.927
433523 95.230 -65.289 53.901
433524 95.269 -65.293 53.875
433525 95.308 -65.298 53.851
433526 95.347 -65.304 53.827
433527 95.387 -65.308 53.805
433528 95.426 -65.314 53.784
433529 95.465 -65.319 53.765
433530 95.510 -65.320 53.749
433531 95.557 -65.319 53.732
433532 95.603 -65.320 53.716
433533 95.643 -65.325 53.700
433534 95.683 -65.329 53.674
433535 95.724 -65.335 53.646
433536 95.764 -65.340 53.614
433537 95.804 -65.345 53.578
433538 95.844 -65.350 53.540
433539 95.886 -65.346 53.502
433540 95.928 -65.342 53.465
433541 95.970 -65.338 53.429
433542 96.012 -65.333 53.392
433543 96.053 -65.329 53.359
433544 96.095 -65.324 53.323
433545 96.137 -65.320 53.285
433546 96.181 -65.316 53.245
433547 96.225 -65.312 53.202
433548 96.268 -65.309 53.160
433549 96.311 -65.308 53.120
433550 96.355 -65.307 53.082
433551 96.397 -65.305 53.045
433552 96.437 -65.308 53.010
433553 96.474 -65.312 52.973
433554 96.512 -65.317 52.935
433555 96.549 -65.320 52.896
433556 96.587 -65.325 52.856
433557 96.624 -65.329 52.815
433558 96.662 -65.333 52.774
433559 96.699 -65.338 52.733
433560 96.737 -65.342 52.691
433561 96.777 -65.345 52.649
433562 96.818 -65.345 52.606
433563 96.860 -65.345 52.563
433564 96.902 -65.345 52.520
433565 96.944 -65.344 52.476
433566 96.985 -65.344 52.433
433567 97.027 -65.344 52.389
433568 97.069 -65.343 52.347
433569 97.111 -65.343 52.301
433570 97.138 -65.336 52.254
433571 97.145 -65.325 52.208
433572 97.180 -65.333 52.163
433573 97.215 -65.342 52.119
433574 97.235 -65.361 52.079
433575 97.260 -65.380 52.041
433576 97.300 -65.388 52.003
433577 97.340 -65.395 51.967
433578 97.380 -65.404 51.929
433579 97.420 -65.413 51.893
433580 97.460 -65.422 51.860
433581 97.500 -65.431 51.828
433582 97.540 -65.440 51.788
433583 97.580 -65.449 51.748
433584 97.620 -65.458 51.708
433585 97.654 -65.461 51.671
433586 97.687 -65.465 51.634
433587 97.721 -65.468 51.598
433588 97.754 -65.467 51.562
433589 97.787 -65.464 51.525
433590 97.819 -65.463 51.491
433591 97.852 -65.461 51.459
433592 97.880 -65.466 51.426
433593 97.909 -65.471 51.393
433594 97.941 -65.477 51.359
433595 97.975 -65.482 51.322
433596 98.011 -65.487 51.284
433597 98.045 -65.492 51.248
433598 98.081 -65.498 51.214
433599 98.116 -65.502 51.181
433600 98.151 -65.508 51.149
433601 98.186 -65.513 51.115
433602 98.221 -65.518 51.078
433603 98.258 -65.524 51.041
433604 98.295 -65.530 51.004
433605 98.333 -65.535 50.967
433606 98.370 -65.541 50.932
433607 98.407 -65.546 50.893
433608 98.444 -65.552 50.852
433609 98.482 -65.557 50.808
433610 98.519 -65.562 50.761
433611 98.559 -65.568 50.710
433612 98.603 -65.573 50.657
433613 98.647 -65.578 50.601
433614 98.691 -65.583 50.544
433615 98.733 -65.590 50.489
433616 98.776 -65.595 50.431
433617 98.818 -65.601 50.373
433618 98.861 -65.607 50.315
433619 98.903 -65.613 50.257
433620 98.945 -65.619 50.199
433621 98.991 -65.621 50.141
433622 99.037 -65.623 50.085
433623 99.079 -65.626 50.030
433624 99.119 -65.630 49.978
433625 99.159 -65.635 49.929
433626 99.199 -65.640 49.883
433627 99.237 -65.642 49.839
433628 99.282 -65.645 49.796
433629 99.336 -65.650 49.751
433630 99.390 -65.656 49.703
433631 99.444 -65.662 49.655
433632 99.498 -65.667 49.607
433633 99.552 -65.673 49.561
433634 99.606 -65.678 49.517
433635 99.659 -65.684 49.473
433636 99.704 -65.693 49.427
433637 99.748 -65.704 49.380
433638 99.791 -65.715 49.332
433639 99.836 -65.725 49.284
433640 99.879 -65.735 49.236
433641 99.957 -65.716 49.188
433642 99.981 -65.724 49.138
433643 100.001 -65.689 49.088
433644 100.028 -65.661 49.038
433645 100.067 -65.648 48.989
433646 100.108 -65.638 48.941
433647 100.156 -65.637 48.894
433648 100.204 -65.634 48.848
433649 100.253 -65.632 48.802
433650 100.301 -65.630 48.757
433651 100.349 -65.627 48.712
433652 100.398 -65.626 48.668
433653 100.445 -65.624 48.627
433654 100.487 -65.627 48.584
433655 100.526 -65.632 48.542
434071 88.394 -64.349 55.302
434072 88.439 -64.352 55.303
434073 88.486 -64.356 55.305
434074 88.532 -64.360 55.309
434075 88.578 -64.364 55.314
434076 88.624 -64.367 55.320
434077 88.671 -64.372 55.326
434078 88.717 -64.376 55.334
434079 88.772 -64.375 55.342
434080 88.835 -64.371 55.352
434081 88.898 -64.367 55.362
434082 88.961 -64.362 55.369
434083 89.024 -64.358 55.371
434084 89.088 -64.354 55.367
434085 89.129 -64.368 55.358
434086 89.166 -64.386 55.346
434087 89.202 -64.405 55.334
434088 89.240 -64.424 55.322
434089 89.276 -64.443 55.315
434090 89.313 -64.461 55.311
434091 89.350 -64.480 55.307
434092 89.387 -64.498 55.298
434093 89.424 -64.517 55.284
434094 89.461 -64.535 55.266
434095 89.498 -64.554 55.243
434096 89.541 -64.562 55.216
434097 89.589 -64.564 55.189
434098 89.636 -64.566 55.162
434099 89.683 -64.568 55.139
434100 89.731 -64.571 55.118
434101 89.778 -64.573 55.097
434102 89.825 -64.575 55.080
434103 89.873 -64.577 55.065
434104 89.920 -64.580 55.052
434105 89.967 -64.582 55.041
434106 90.015 -64.584 55.033
434107 90.062 -64.587 55.028
434108 90.109 -64.589 55.024
434109 90.157 -64.591 55.021
434110 90.204 -64.593 55.020
434111 90.251 -64.596 55.018
434112 90.299 -64.598 55.018
434113 90.346 -64.602 55.019
434114 90.395 -64.608 55.019
434115 90.443 -64.614 55.019
434116 90.492 -64.620 55.018
434117 90.540 -64.627 55.016
434118 90.589 -64.633 55.012
434119 90.637 -64.639 55.006
434120 90.683 -64.649 54.998
434121 90.727 -64.659 54.988
434122 90.768 -64.673 54.978
434123 90.808 -64.690 54.966
434124 90.848 -64.707 54.953
434125 90.887 -64.724 54.940
434126 90.927 -64.740 54.926
434127 90.967 -64.757 54.913
434128 91.011 -64.769 54.899
434129 91.057 -64.778 54.887
434130 91.102 -64.788 54.873
434131 91.147 -64.798 54.860
434132 91.192 -64.808 54.846
434133 91.237 -64.819 54.833
434134 91.282 -64.828 54.821
434135 91.327 -64.838 54.808
434136 91.373 -64.848 54.798
434137 91.418 -64.859 54.788
434138 91.463 -64.869 54.778
434139 91.508 -64.878 54.770
434140 91.553 -64.888 54.763
434141 91.598 -64.898 54.758
434142 91.643 -64.909 54.753
434143 91.688 -64.919 54.749
434144 91.734 -64.928 54.745
434145 91.779 -64.938 54.743
434146 91.824 -64.949 54.742
434147 91.869 -64.959 54.740
434148 91.914 -64.969 54.737
434149 91.959 -64.978 54.733
434150 92.005 -64.988 54.727
434151 92.050 -64.999 54.719
434152 92.095 -65.009 54.710
434153 92.126 -65.035 54.701
434154 92.149 -65.069 54.695
434155 92.207 -65.067 54.692
434156 92.264 -65.068 54.689
434157 92.321 -65.069 54.686
434158 92.386 -65.069 54.684
434159 92.452 -65.069 54.680
434160 92.517 -65.070 54.675
434161 92.567 -65.070 54.671
434162 92.616 -65.071 54.666
434163 92.667 -65.073 54.662
434164 92.721 -65.079 54.658
434165 92.776 -65.085 54.653
434166 92.829 -65.089 54.649
434167 92.877 -65.094 54.646
434168 92.925 -65.097 54.644
434169 92.972 -65.101 54.641
434170 93.017 -65.105 54.638
434171 93.062 -65.110 54.636
434172 93.108 -65.114 54.635
434173 93.153 -65.119 54.633
434174 93.199 -65.123 54.631
434175 93.244 -65.127 54.629
434176 93.289 -65.132 54.626
434177 93.335 -65.136 54.624
434178 93.381 -65.140 54.623
434179 93.427 -65.144 54.623
434180 93.474 -65.148 54.625
434181 93.519 -65.152 54.626
434182 93.565 -65.156 54.624
434183 93.610 -65.160 54.623
434184 93.653 -65.163 54.622
434185 93.698 -65.166 54.620
434186 93.742 -65.169 54.619
434187 93.786 -65.173 54.617
434188 93.831 -65.176 54.615
434189 93.875 -65.179 54.610
434190 93.919 -65.182 54.604
434191 93.964 -65.185 54.596
434192 93.999 -65.182 54.588
434193 94.032 -65.178 54.577
434194 94.065 -65.173 54.563
434195 94.098 -65.169 54.548
434196 94.133 -65.170 54.531
434197 94.168 -65.172 54.511
434198 94.202 -65.173 54.489
434199 94.237 -65.175 54.468
434200 94.271 -65.176 54.445
434201 94.306 -65.178 54.422
434202 94.341 -65.179 54.401
434203 94.376 -65.181 54.382
434204 94.410 -65.182 54.364
434205 94.445 -65.184 54.348
434206 94.484 -65.185 54.332
434207 94.523 -65.186 54.313
434208 94.562 -65.188 54.294
434209 94.602 -65.189 54.275
434210 94.641 -65.191 54.254
434211 94.686 -65.195 54.230
434212 94.730 -65.199 54.204
434213 94.774 -65.203 54.177
434214 94.818 -65.207 54.147
434215 94.862 -65.210 54.117
434216 94.905 -65.213 54.092
434217 94.949 -65.216 54.069
434218 94.992 -65.220 54.048
434219 95.036 -65.223 54.027
434220 95.079 -65.226 54.004
434221 95.122 -65.229 53.980
434222 95.162 -65.235 53.955
434223 95.202 -65.241 53.929
434224 95.229 -65.261 53.903
434225 95.267 -65.267 53.876
434226 95.306 -65.272 53.849
434227 95.345 -65.277 53.824
434228 95.384 -65.282 53.801
434229 95.423 -65.287 53.780
434230 95.462 -65.292 53.760
434231 95.509 -65.292 53.743
434232 95.557 -65.291 53.727
434233 95.605 -65.290 53.710
434234 95.645 -65.295 53.692
434235 95.685 -65.300 53.668
434236 95.726 -65.305 53.640
434237 95.766 -65.310 53.609
434238 95.806 -65.315 53.575
434239 95.847 -65.314 53.538
434240 95.889 -65.309 53.500
434241 95.931 -65.304 53.461
434242 95.973 -65.300 53.423
434243 96.015 -65.295 53.386
434244 96.056 -65.291 53.351
434245 96.099 -65.286 53.318
434246 96.141 -65.282 53.281
434247 96.184 -65.278 53.241
434248 96.228 -65.275 53.201
434249 96.270 -65.273 53.158
434250 96.314 -65.272 53.117
434251 96.357 -65.270 53.078
434252 96.399 -65.270 53.041
434253 96.438 -65.273 53.004
434254 96.475 -65.277 52.968
434255 96.513 -65.282 52.930
434256 96.550 -65.286 52.891
434257 96.588 -65.290 52.850
434258 96.625 -65.295 52.810
434259 96.663 -65.299 52.769
434260 96.701 -65.304 52.728
434261 96.741 -65.310 52.687
434262 96.781 -65.315 52.645
434263 96.811 -65.311 52.602
434264 96.853 -65.311 52.558
434265 96.895 -65.311 52.515
434266 96.937 -65.311 52.472
434267 96.978 -65.311 52.429
434268 97.020 -65.310 52.386
434269 97.062 -65.310 52.343
434270 97.104 -65.310 52.297
434271 97.108 -65.294 52.249
434272 97.141 -65.300 52.202
434273 97.176 -65.308 52.155
434274 97.212 -65.315 52.109
434275 97.242 -65.327 52.067
434276 97.265 -65.345 52.026
434277 97.306 -65.353 51.986
434278 97.346 -65.362 51.949
434279 97.385 -65.371 51.912
434280 97.426 -65.380 51.877
434281 97.465 -65.389 51.845
434282 97.506 -65.398 51.812
434283 97.545 -65.408 51.773
434284 97.586 -65.417 51.732
434285 97.626 -65.426 51.692
434286 97.663 -65.432 51.655
434287 97.696 -65.431 51.618
434288 97.728 -65.429 51.582
434289 97.760 -65.427 51.544
434290 97.793 -65.425 51.508
434291 97.825 -65.423 51.474
434292 97.858 -65.421 51.442
434293 97.887 -65.426 51.409
434294 97.916 -65.431 51.376
434295 97.944 -65.436 51.341
434296 97.974 -65.442 51.303
434297 98.009 -65.447 51.264
434298 98.044 -65.452 51.228
434299 98.079 -65.458 51.194
434300 98.114 -65.463 51.161
434301 98.149 -65.468 51.127
434302 98.184 -65.473 51.092
434303 98.221 -65.479 51.055
434304 98.258 -65.484 51.018
434305 98.296 -65.490 50.982
434306 98.333 -65.496 50.947
434307 98.370 -65.501 50.915
434308 98.407 -65.507 50.880
434309 98.444 -65.512 50.840
434310 98.482 -65.517 50.797
434311 98.519 -65.523 50.750
434312 98.561 -65.528 50.700
434313 98.605 -65.533 50.646
434314 98.650 -65.539 50.591
434315 98.693 -65.544 50.534
434316 98.735 -65.550 50.479
434317 98.778 -65.556 50.422
434318 98.820 -65.562 50.363
434319 98.862 -65.568 50.303
434320 98.904 -65.574 50.245
434321 98.947 -65.579 50.186
434322 98.993 -65.581 50.128
434323 99.039 -65.583 50.072
434324 99.083 -65.585 50.018
434325 99.123 -65.590 49.965
434326 99.163 -65.594 49.917
434327 99.202 -65.598 49.871
434328 99.239 -65.598 49.828
434329 99.286 -65.602 49.786
434330 99.340 -65.607 49.742
434331 99.394 -65.612 49.696
434332 99.447 -65.618 49.648
434333 99.501 -65.624 49.601
434334 99.555 -65.629 49.555
434335 99.610 -65.634 49.511
434336 99.664 -65.637 49.467
434337 99.709 -65.647 49.421
434338 99.753 -65.658 49.373
434339 99.796 -65.668 49.325
434340 99.840 -65.678 49.277
434341 99.893 -65.681 49.229
434342 99.975 -65.657 49.180
434343 99.994 -65.677 49.131
434344 100.014 -65.641 49.081
434345 100.041 -65.614 49.031
434346 100.080 -65.601 48.982
434347 100.123 -65.593 48.935
434348 100.171 -65.590 48.888
434349 100.219 -65.588 48.842
434350 100.267 -65.586 48.797
434351 100.316 -65.584 48.752
434352 100.363 -65.582 48.707
434353 100.412 -65.580 48.664
434354 100.459 -65.579 48.622
434355 100.497 -65.584 48.579
434772 88.396 -64.295 55.304
434773 88.442 -64.299 55.304
434774 88.488 -64.303 55.306
434775 88.534 -64.307 55.310
434776 88.580 -64.311 55.314
434777 88.627 -64.314 55.319
434778 88.673 -64.318 55.324
434779 88.735 -64.315 55.330
434780 88.798 -64.311 55.335
434781 88.861 -64.306 55.342
434782 88.924 -64.302 55.348
434783 88.987 -64.298 55.353
434784 89.050 -64.293 55.354
434785 89.097 -64.303 55.351
434786 89.133 -64.322 55.343
434787 89.171 -64.340 55.333
434788 89.207 -64.359 55.323
434789 89.244 -64.377 55.313
434790 89.281 -64.396 55.307
434791 89.318 -64.414 55.302
434792 89.355 -64.433 55.298
434793 89.392 -64.452 55.288
434794 89.428 -64.471 55.274
434795 89.466 -64.489 55.258
434796 89.503 -64.508 55.239
434797 89.543 -64.521 55.215
434798 89.590 -64.524 55.191
434799 89.637 -64.526 55.167
434800 89.685 -64.528 55.145
434801 89.732 -64.530 55.125
434802 89.779 -64.533 55.106
434803 89.827 -64.535 55.090
434804 89.874 -64.537 55.075
434805 89.921 -64.540 55.062
434806 89.969 -64.542 55.051
434807 90.016 -64.544 55.045
434808 90.063 -64.546 55.040
434809 90.110 -64.549 55.036
434810 90.158 -64.551 55.034
434811 90.205 -64.553 55.033
434812 90.252 -64.555 55.032
434813 90.300 -64.558 55.033
434814 90.347 -64.560 55.034
434815 90.394 -64.562 55.034
434816 90.442 -64.565 55.034
434817 90.490 -64.571 55.033
434818 90.539 -64.577 55.031
434819 90.588 -64.584 55.027
434820 90.636 -64.590 55.021
434821 90.680 -64.601 55.009
434822 90.721 -64.616 54.998
434823 90.760 -64.633 54.985
434824 90.800 -64.649 54.972
434825 90.841 -64.665 54.958
434826 90.886 -64.675 54.943
434827 90.931 -64.686 54.928
434828 90.976 -64.696 54.913
434829 91.021 -64.706 54.899
434830 91.067 -64.715 54.884
434831 91.111 -64.725 54.871
434832 91.157 -64.736 54.857
434833 91.202 -64.746 54.843
434834 91.247 -64.756 54.830
434835 91.292 -64.765 54.817
434836 91.338 -64.776 54.804
434837 91.382 -64.786 54.793
434838 91.428 -64.796 54.783
434839 91.472 -64.806 54.773
434840 91.518 -64.815 54.764
434841 91.563 -64.826 54.756
434842 91.608 -64.836 54.749
434843 91.653 -64.846 54.744
434844 91.699 -64.856 54.738
434845 91.743 -64.866 54.734
434846 91.789 -64.876 54.730
434847 91.834 -64.886 54.727
434848 91.879 -64.896 54.723
434849 91.924 -64.906 54.719
434850 91.969 -64.916 54.713
434851 92.014 -64.926 54.707
434852 92.060 -64.936 54.700
434853 92.104 -64.946 54.693
434854 92.148 -64.958 54.685
434855 92.166 -64.998 54.680
434856 92.203 -65.019 54.677
434857 92.266 -65.020 54.673
434858 92.332 -65.020 54.671
434859 92.398 -65.020 54.669
434860 92.463 -65.021 54.665
434861 92.515 -65.021 54.660
434862 92.564 -65.022 54.655
434863 92.614 -65.022 54.651
434864 92.664 -65.024 54.647
434865 92.719 -65.030 54.644
434866 92.774 -65.035 54.639
434867 92.829 -65.041 54.635
434868 92.880 -65.045 54.633
434869 92.926 -65.050 54.630
434870 92.971 -65.054 54.627
434871 93.016 -65.059 54.623
434872 93.062 -65.063 54.621
434873 93.108 -65.068 54.618
434874 93.153 -65.072 54.617
434875 93.198 -65.077 54.616
434876 93.243 -65.082 54.614
434877 93.289 -65.086 54.612
434878 93.335 -65.090 54.612
434879 93.380 -65.094 54.613
434880 93.427 -65.098 54.614
434881 93.473 -65.102 54.616
434882 93.518 -65.106 54.618
434883 93.564 -65.110 54.617
434884 93.609 -65.114 54.617
434885 93.655 -65.118 54.617
434886 93.699 -65.122 54.616
434887 93.744 -65.125 54.616
434888 93.788 -65.128 54.615
434889 93.832 -65.131 54.612
434890 93.876 -65.135 54.608
434891 93.921 -65.138 54.602
434892 93.965 -65.141 54.595
434893 94.003 -65.140 54.585
434894 94.036 -65.135 54.575
434895 94.069 -65.131 54.562
434896 94.103 -65.129 54.546
434897 94.137 -65.131 54.528
434898 94.172 -65.132 54.509
434899 94.207 -65.133 54.488
434900 94.242 -65.135 54.466
434901 94.276 -65.136 54.443
434902 94.311 -65.138 54.421
434903 94.345 -65.139 54.399
434904 94.380 -65.141 54.379
434905 94.415 -65.142 54.360
434906 94.450 -65.144 54.343
434907 94.487 -65.145 54.326
434908 94.526 -65.147 54.307
434909 94.566 -65.148 54.287
434910 94.605 -65.150 54.268
434911 94.649 -65.154 54.246
434912 94.693 -65.157 54.223
434913 94.737 -65.161 54.199
434914 94.781 -65.165 54.172
434915 94.826 -65.169 54.143
434916 94.870 -65.173 54.114
434917 94.913 -65.176 54.088
434918 94.957 -65.179 54.065
434919 95.000 -65.182 54.045
434920 95.043 -65.185 54.025
434921 95.086 -65.189 54.003
434922 95.126 -65.195 53.979
434923 95.166 -65.201 53.955
434924 95.206 -65.207 53.930
434925 95.236 -65.223 53.903
434926 95.264 -65.240 53.875
434927 95.304 -65.245 53.847
434928 95.343 -65.251 53.820
434929 95.382 -65.255 53.796
434930 95.421 -65.261 53.774
434931 95.460 -65.266 53.753
434932 95.508 -65.264 53.737
434933 95.556 -65.264 53.721
434934 95.604 -65.262 53.703
434935 95.647 -65.265 53.683
434936 95.687 -65.270 53.660
434937 95.728 -65.275 53.633
434938 95.768 -65.280 53.602
434939 95.809 -65.280 53.569
434940 95.850 -65.276 53.532
434941 95.892 -65.271 53.494
434942 95.935 -65.267 53.456
434943 95.976 -65.262 53.418
434944 96.018 -65.257 53.380
434945 96.060 -65.253 53.343
434946 96.102 -65.248 53.308
434947 96.144 -65.244 53.273
434948 96.187 -65.240 53.235
434949 96.230 -65.239 53.197
434950 96.273 -65.238 53.154
434951 96.316 -65.236 53.114
434952 96.359 -65.235 53.074
434953 96.401 -65.234 53.033
434954 96.439 -65.239 52.995
434955 96.476 -65.242 52.959
434956 96.514 -65.247 52.921
434957 96.551 -65.251 52.881
434958 96.589 -65.255 52.841
434959 96.629 -65.261 52.801
434960 96.669 -65.267 52.761
434961 96.709 -65.273 52.720
434962 96.749 -65.279 52.679
434963 96.789 -65.285 52.637
434964 96.814 -65.281 52.594
434965 96.846 -65.278 52.550
434966 96.888 -65.278 52.507
434967 96.929 -65.277 52.463
434968 96.971 -65.277 52.421
434969 97.013 -65.276 52.379
434970 97.054 -65.276 52.336
434971 97.084 -65.271 52.291
434972 97.102 -65.267 52.244
434973 97.137 -65.274 52.196
434974 97.172 -65.282 52.148
434975 97.208 -65.289 52.101
434976 97.243 -65.297 52.057
434977 97.272 -65.311 52.015
434978 97.311 -65.320 51.972
434979 97.352 -65.329 51.932
434980 97.391 -65.338 51.895
434981 97.432 -65.347 51.860
434982 97.471 -65.356 51.827
434983 97.510 -65.363 51.796
434984 97.549 -65.370 51.757
434985 97.588 -65.377 51.715
434986 97.627 -65.385 51.675
434987 97.667 -65.391 51.638
434988 97.702 -65.392 51.602
434989 97.735 -65.389 51.565
434990 97.767 -65.388 51.527
434991 97.800 -65.386 51.490
434992 97.832 -65.384 51.456
434993 97.865 -65.382 51.424
434994 97.894 -65.386 51.392
434995 97.923 -65.392 51.358
434996 97.951 -65.397 51.322
434997 97.980 -65.402 51.282
434998 98.008 -65.407 51.243
434999 98.043 -65.412 51.207
435000 98.078 -65.417 51.173
435001 98.113 -65.423 51.139
435002 98.148 -65.428 51.106
435003 98.184 -65.433 51.069
435004 98.222 -65.439 51.032
435005 98.259 -65.445 50.995
435006 98.296 -65.450 50.961
435007 98.334 -65.456 50.928
435008 98.370 -65.461 50.899
435009 98.408 -65.467 50.867
435010 98.445 -65.473 50.830
435011 98.482 -65.478 50.786
435012 98.519 -65.483 50.738
435013 98.564 -65.489 50.688
435014 98.608 -65.494 50.635
435015 98.652 -65.499 50.580
435016 98.695 -65.505 50.523
435017 98.737 -65.511 50.466
435018 98.779 -65.517 50.408
435019 98.821 -65.524 50.349
435020 98.863 -65.530 50.289
435021 98.906 -65.536 50.230
435022 98.950 -65.539 50.171
435023 98.996 -65.541 50.115
435024 99.042 -65.543 50.059
435025 99.088 -65.544 50.005
435026 99.128 -65.549 49.952
435027 99.167 -65.553 49.903
435028 99.205 -65.554 49.858
435029 99.242 -65.555 49.816
435030 99.289 -65.558 49.774
435031 99.343 -65.564 49.731
435032 99.397 -65.569 49.688
435033 99.451 -65.575 49.642
435034 99.505 -65.580 49.596
435035 99.559 -65.585 49.551
435036 99.614 -65.588 49.506
435037 99.668 -65.591 49.461
435038 99.713 -65.601 49.413
435039 99.757 -65.611 49.365
435040 99.801 -65.621 49.317
435041 99.845 -65.632 49.269
435042 99.911 -65.622 49.221
435043 99.994 -65.598 49.172
435044 100.007 -65.630 49.123
435045 100.028 -65.594 49.073
435046 100.055 -65.567 49.024
435047 100.093 -65.554 48.976
435048 100.137 -65.546 48.928
435049 100.185 -65.544 48.882
435050 100.233 -65.542 48.837
435051 100.281 -65.540 48.791
435052 100.330 -65.538 48.745
435053 100.378 -65.536 48.702
435054 100.426 -65.533 48.659
435055 100.469 -65.536 48.617
435056 100.507 -65.541 48.575
435473 88.398 -64.242 55.306
435474 88.445 -64.245 55.305
435475 88.490 -64.249 55.307
435476 88.537 -64.254 55.310
435477 88.583 -64.258 55.314
435478 88.634 -64.259 55.318
435479 88.697 -64.255 55.322
435480 88.760 -64.250 55.327
435481 88.824 -64.245 55.330
435482 88.887 -64.242 55.333
435483 88.950 -64.237 55.337
435484 89.013 -64.233 55.340
435485 89.064 -64.239 55.340
435486 89.101 -64.257 55.337
435487 89.138 -64.276 55.331
435488 89.175 -64.294 55.323
435489 89.212 -64.313 55.315
435490 89.249 -64.331 55.309
435491 89.286 -64.350 55.304
435492 89.323 -64.368 55.300
435493 89.359 -64.387 55.295
435494 89.397 -64.405 55.287
435495 89.434 -64.424 55.275
435496 89.470 -64.443 55.259
435497 89.507 -64.461 55.241
435498 89.544 -64.480 55.219
435499 89.591 -64.483 55.196
435500 89.638 -64.485 55.174
435501 89.686 -64.487 55.153
435502 89.733 -64.490 55.133
435503 89.780 -64.492 55.114
435504 89.828 -64.494 55.099
435505 89.875 -64.496 55.085
435506 89.922 -64.499 55.073
435507 89.970 -64.501 55.063
435508 90.017 -64.503 55.058
435509 90.064 -64.505 55.053
435510 90.112 -64.508 55.050
435511 90.159 -64.510 55.048
435512 90.206 -64.512 55.047
435513 90.254 -64.515 55.047
435514 90.301 -64.517 55.049
435515 90.348 -64.519 55.051
435516 90.396 -64.521 55.052
435517 90.443 -64.524 55.053
435518 90.490 -64.526 55.053
435519 90.538 -64.528 55.051
435520 90.586 -64.535 55.048
435521 90.633 -64.543 55.042
435522 90.673 -64.559 55.031
435523 90.715 -64.573 55.018
435524 90.760 -64.583 55.002
435525 90.805 -64.593 54.986
435526 90.851 -64.603 54.968
435527 90.896 -64.613 54.948
435528 90.941 -64.623 54.931
435529 90.986 -64.633 54.915
435530 91.031 -64.643 54.899
435531 91.076 -64.653 54.882
435532 91.121 -64.663 54.869
435533 91.167 -64.673 54.854
435534 91.212 -64.683 54.840
435535 91.257 -64.693 54.826
435536 91.302 -64.703 54.813
435537 91.347 -64.713 54.800
435538 91.392 -64.723 54.789
435539 91.437 -64.733 54.779
435540 91.483 -64.743 54.767
435541 91.528 -64.753 54.758
435542 91.573 -64.763 54.749
435543 91.618 -64.773 54.741
435544 91.663 -64.784 54.735
435545 91.708 -64.794 54.729
435546 91.753 -64.803 54.724
435547 91.799 -64.813 54.719
435548 91.844 -64.823 54.714
435549 91.889 -64.834 54.709
435550 91.934 -64.844 54.704
435551 91.979 -64.853 54.698
435552 92.024 -64.863 54.692
435553 92.069 -64.874 54.685
435554 92.115 -64.884 54.679
435555 92.160 -64.894 54.673
435556 92.188 -64.922 54.668
435557 92.217 -64.963 54.664
435558 92.277 -64.971 54.660
435559 92.343 -64.972 54.657
435560 92.409 -64.972 54.655
435561 92.464 -64.972 54.651
435562 92.513 -64.972 54.647
435563 92.562 -64.973 54.642
435564 92.611 -64.973 54.640
435565 92.662 -64.975 54.635
435566 92.717 -64.981 54.631
435567 92.771 -64.986 54.628
435568 92.826 -64.991 54.625
435569 92.876 -65.003 54.623
435570 92.924 -65.012 54.618
435571 92.970 -65.014 54.615
435572 93.016 -65.016 54.611
435573 93.061 -65.019 54.607
435574 93.107 -65.022 54.603
435575 93.152 -65.026 54.602
435576 93.198 -65.031 54.602
435577 93.243 -65.035 54.602
435578 93.289 -65.040 54.602
435579 93.334 -65.044 54.602
435580 93.380 -65.048 54.604
435581 93.426 -65.052 54.606
435582 93.471 -65.057 54.608
435583 93.517 -65.060 54.610
435584 93.562 -65.064 54.611
435585 93.608 -65.069 54.612
435586 93.653 -65.072 54.612
435587 93.699 -65.076 54.613
435588 93.745 -65.080 54.613
435589 93.789 -65.084 54.613
435590 93.834 -65.087 54.611
435591 93.878 -65.090 54.607
435592 93.922 -65.094 54.600
435593 93.967 -65.097 54.593
435594 94.006 -65.097 54.584
435595 94.040 -65.092 54.574
435596 94.072 -65.088 54.560
435597 94.107 -65.089 54.543
435598 94.142 -65.091 54.525
435599 94.177 -65.092 54.506
435600 94.211 -65.094 54.484
435601 94.246 -65.095 54.462
435602 94.280 -65.097 54.439
435603 94.315 -65.098 54.416
435604 94.350 -65.100 54.394
435605 94.385 -65.101 54.373
435606 94.419 -65.103 54.353
435607 94.454 -65.104 54.334
435608 94.490 -65.106 54.316
435609 94.530 -65.107 54.298
435610 94.569 -65.109 54.279
435611 94.612 -65.112 54.259
435612 94.657 -65.116 54.238
435613 94.700 -65.120 54.213
435614 94.745 -65.123 54.189
435615 94.789 -65.127 54.162
435616 94.833 -65.131 54.134
435617 94.877 -65.135 54.108
435618 94.921 -65.138 54.082
435619 94.964 -65.141 54.059
435620 95.007 -65.145 54.039
435621 95.051 -65.148 54.020
435622 95.091 -65.154 53.999
435623 95.130 -65.160 53.977
435624 95.170 -65.166 53.954
435625 95.210 -65.172 53.928
435626 95.243 -65.185 53.900
435627 95.269 -65.205 53.872
435628 95.301 -65.219 53.844
435629 95.340 -65.224 53.816
435630 95.380 -65.229 53.790
435631 95.418 -65.234 53.767
435632 95.460 -65.238 53.746
435633 95.508 -65.236 53.729
435634 95.555 -65.235 53.713
435635 95.603 -65.234 53.694
435636 95.649 -65.235 53.674
435637 95.689 -65.240 53.651
435638 95.730 -65.245 53.624
435639 95.770 -65.247 53.594
435640 95.812 -65.242 53.560
435641 95.854 -65.238 53.524
435642 95.896 -65.233 53.487
435643 95.938 -65.229 53.450
435644 95.979 -65.224 53.411
435645 96.021 -65.220 53.373
435646 96.063 -65.216 53.334
435647 96.105 -65.211 53.299
435648 96.148 -65.204 53.264
435649 96.189 -65.203 53.228
435650 96.232 -65.203 53.190
435651 96.275 -65.201 53.149
435652 96.318 -65.201 53.110
435653 96.361 -65.199 53.069
435654 96.402 -65.199 53.027
435655 96.440 -65.204 52.987
435656 96.477 -65.207 52.949
435657 96.516 -65.213 52.910
435658 96.556 -65.219 52.870
435659 96.597 -65.225 52.830
435660 96.637 -65.231 52.790
435661 96.677 -65.236 52.751
435662 96.717 -65.242 52.711
435663 96.757 -65.248 52.671
435664 96.797 -65.254 52.629
435665 96.827 -65.254 52.587
435666 96.847 -65.247 52.542
435667 96.880 -65.244 52.497
435668 96.922 -65.244 52.453
435669 96.964 -65.244 52.410
435670 97.006 -65.243 52.369
435671 97.047 -65.243 52.326
435672 97.063 -65.233 52.281
435673 97.098 -65.241 52.234
435674 97.133 -65.248 52.187
435675 97.168 -65.256 52.139
435676 97.204 -65.264 52.093
435677 97.239 -65.271 52.049
435678 97.277 -65.274 52.006
435679 97.314 -65.281 51.962
435680 97.353 -65.288 51.920
435681 97.391 -65.295 51.880
435682 97.430 -65.302 51.842
435683 97.469 -65.310 51.808
435684 97.508 -65.317 51.776
435685 97.547 -65.324 51.738
435686 97.586 -65.332 51.697
435687 97.625 -65.339 51.656
435688 97.665 -65.345 51.619
435689 97.706 -65.350 51.583
435690 97.741 -65.350 51.545
435691 97.774 -65.348 51.507
435692 97.806 -65.346 51.469
435693 97.839 -65.344 51.435
435694 97.871 -65.342 51.401
435695 97.901 -65.346 51.368
435696 97.929 -65.351 51.334
435697 97.958 -65.357 51.299
435698 97.987 -65.362 51.260
435699 98.015 -65.367 51.221
435700 98.043 -65.373 51.184
435701 98.077 -65.377 51.149
435702 98.111 -65.383 51.115
435703 98.148 -65.389 51.082
435704 98.185 -65.394 51.046
435705 98.222 -65.399 51.010
435706 98.260 -65.405 50.976
435707 98.297 -65.411 50.943
435708 98.334 -65.416 50.911
435709 98.371 -65.422 50.880
435710 98.408 -65.427 50.848
435711 98.446 -65.432 50.812
435712 98.483 -65.436 50.769
435713 98.522 -65.441 50.721
435714 98.566 -65.445 50.670
435715 98.610 -65.449 50.615
435716 98.654 -65.453 50.559
435717 98.697 -65.460 50.502
435718 98.738 -65.467 50.445
435719 98.781 -65.473 50.387
435720 98.823 -65.480 50.329
435721 98.865 -65.486 50.270
435722 98.904 -65.492 50.211
435723 98.949 -65.496 50.154
435724 98.998 -65.501 50.097
435725 99.044 -65.502 50.041
435726 99.090 -65.504 49.987
435727 99.132 -65.508 49.936
435728 99.170 -65.509 49.890
435729 99.207 -65.510 49.847
435730 99.244 -65.511 49.806
435731 99.293 -65.514 49.767
435732 99.347 -65.521 49.725
435733 99.401 -65.526 49.682
435734 99.455 -65.531 49.638
435735 99.509 -65.536 49.591
435736 99.563 -65.539 49.545
435737 99.618 -65.542 49.500
435738 99.672 -65.545 49.453
435739 99.718 -65.554 49.405
435740 99.762 -65.564 49.356
435741 99.806 -65.575 49.307
435742 99.850 -65.585 49.259
435743 99.930 -65.563 49.211
435744 100.013 -65.539 49.163
435745 100.020 -65.583 49.114
435746 100.041 -65.547 49.065
435747 100.068 -65.520 49.015
435748 100.107 -65.507 48.967
435749 100.151 -65.500 48.920
435750 100.200 -65.498 48.875
435751 100.248 -65.496 48.830
435752 100.296 -65.494 48.784
435753 100.344 -65.492 48.740
435754 100.393 -65.489 48.696
435755 100.440 -65.488 48.654
435756 100.479 -65.493 48.612
435757 100.517 -65.499 48.570
436175 88.447 -64.192 55.308
436176 88.493 -64.196 55.309
436177 88.539 -64.200 55.312
436178 88.597 -64.198 55.315
436179 88.660 -64.194 55.318
436180 88.723 -64.189 55.320
436181 88.786 -64.186 55.324
436182 88.849 -64.181 55.326
436183 88.913 -64.176 55.327
436184 88.972 -64.176 55.328
436185 89.021 -64.187 55.330
436186 89.069 -64.197 55.329
436187 89.107 -64.214 55.326
436188 89.143 -64.230 55.322
436189 89.180 -64.248 55.315
436190 89.217 -64.267 55.311
436191 89.254 -64.285 55.308
436192 89.291 -64.304 55.305
436193 89.327 -64.322 55.301
436194 89.365 -64.341 55.298
436195 89.401 -64.359 55.291
436196 89.438 -64.378 55.281
436197 89.475 -64.396 55.267
436198 89.512 -64.415 55.248
436199 89.549 -64.433 55.226
436200 89.592 -64.442 55.204
436201 89.639 -64.444 55.183
436202 89.687 -64.446 55.162
436203 89.734 -64.449 55.144
436204 89.781 -64.451 55.125
436205 89.829 -64.453 55.110
436206 89.876 -64.455 55.097
436207 89.923 -64.458 55.086
436208 89.971 -64.460 55.077
436209 90.018 -64.462 55.071
436210 90.065 -64.464 55.067
436211 90.113 -64.467 55.064
436212 90.160 -64.469 55.064
436213 90.207 -64.471 55.063
436214 90.255 -64.474 55.065
436215 90.302 -64.476 55.067
436216 90.350 -64.478 55.070
436217 90.397 -64.480 55.072
436218 90.444 -64.483 55.075
436219 90.492 -64.485 55.075
436220 90.539 -64.487 55.075
436221 90.588 -64.486 55.072
436222 90.635 -64.494 55.066
436223 90.679 -64.505 55.055
436224 90.725 -64.515 55.040
436225 90.770 -64.524 55.023
436226 90.815 -64.534 55.003
436227 90.860 -64.544 54.983
436228 90.905 -64.554 54.957
436229 90.950 -64.564 54.936
436230 90.996 -64.574 54.917
436231 91.041 -64.584 54.899
436232 91.086 -64.594 54.882
436233 91.131 -64.604 54.868
436234 91.176 -64.614 54.853
436235 91.221 -64.624 54.838
436236 91.266 -64.634 54.823
436237 91.311 -64.643 54.810
436238 91.357 -64.654 54.796
436239 91.402 -64.664 54.785
436240 91.447 -64.674 54.773
436241 91.492 -64.684 54.762
436242 91.537 -64.693 54.752
436243 91.582 -64.703 54.743
436244 91.628 -64.713 54.735
436245 91.673 -64.724 54.728
436246 91.718 -64.734 54.720
436247 91.763 -64.743 54.715
436248 91.808 -64.753 54.709
436249 91.853 -64.763 54.703
436250 91.898 -64.773 54.697
436251 91.944 -64.783 54.692
436252 91.989 -64.793 54.685
436253 92.034 -64.803 54.679
436254 92.079 -64.813 54.673
436255 92.124 -64.823 54.667
436256 92.169 -64.833 54.661
436257 92.215 -64.847 54.656
436258 92.249 -64.888 54.652
436259 92.288 -64.922 54.648
436260 92.354 -64.922 54.645
436261 92.413 -64.922 54.642
436262 92.462 -64.923 54.639
436263 92.511 -64.924 54.635
436264 92.560 -64.924 54.631
436265 92.609 -64.925 54.629
436266 92.659 -64.926 54.624
436267 92.714 -64.931 54.620
436268 92.769 -64.937 54.617
436269 92.821 -64.946 54.614
436270 92.870 -64.959 54.612
436271 92.920 -64.972 54.608
436272 92.968 -64.982 54.604
436273 93.013 -64.984 54.600
436274 93.059 -64.987 54.597
436275 93.105 -64.989 54.595
436276 93.150 -64.991 54.594
436277 93.196 -64.994 54.596
436278 93.242 -64.997 54.597
436279 93.288 -64.999 54.598
436280 93.333 -65.001 54.599
436281 93.379 -65.003 54.599
436282 93.425 -65.007 54.600
436283 93.470 -65.010 54.602
436284 93.516 -65.015 54.604
436285 93.561 -65.019 54.606
436286 93.607 -65.022 54.607
436287 93.652 -65.026 54.609
436288 93.698 -65.031 54.611
436289 93.744 -65.035 54.612
436290 93.789 -65.038 54.611
436291 93.835 -65.043 54.610
436292 93.879 -65.046 54.605
436293 93.923 -65.049 54.598
436294 93.968 -65.052 54.592
436295 94.010 -65.054 54.583
436296 94.043 -65.050 54.571
436297 94.077 -65.048 54.557
436298 94.112 -65.050 54.540
436299 94.146 -65.051 54.520
436300 94.181 -65.053 54.501
436301 94.216 -65.054 54.479
436302 94.251 -65.056 54.456
436303 94.285 -65.057 54.434
436304 94.320 -65.059 54.409
436305 94.354 -65.060 54.387
436306 94.389 -65.062 54.366
436307 94.424 -65.063 54.345
436308 94.459 -65.065 54.325
436309 94.494 -65.066 54.307
436310 94.533 -65.068 54.288
436311 94.575 -65.071 54.269
436312 94.620 -65.075 54.249
436313 94.664 -65.079 54.228
436314 94.708 -65.082 54.203
436315 94.752 -65.086 54.179
436316 94.796 -65.090 54.153
436317 94.841 -65.094 54.126
436318 94.885 -65.098 54.103
436319 94.928 -65.101 54.078
436320 94.972 -65.104 54.056
436321 95.015 -65.107 54.035
436322 95.055 -65.113 54.015
436323 95.095 -65.120 53.994
436324 95.135 -65.125 53.972
436325 95.175 -65.131 53.949
436326 95.214 -65.137 53.925
436327 95.250 -65.148 53.898
436328 95.276 -65.167 53.870
436329 95.303 -65.188 53.839
436330 95.338 -65.197 53.810
436331 95.377 -65.202 53.782
436332 95.416 -65.207 53.758
436333 95.459 -65.210 53.737
436334 95.507 -65.209 53.719
436335 95.555 -65.207 53.703
436336 95.603 -65.207 53.686
436337 95.651 -65.205 53.667
436338 95.691 -65.210 53.644
436339 95.732 -65.213 53.616
436340 95.773 -65.210 53.585
436341 95.815 -65.205 53.550
436342 95.857 -65.201 53.514
436343 95.899 -65.196 53.478
436344 95.941 -65.192 53.440
436345 95.983 -65.184 53.402
436346 96.026 -65.177 53.365
436347 96.069 -65.170 53.326
436348 96.111 -65.163 53.290
436349 96.151 -65.158 53.253
436350 96.192 -65.158 53.217
436351 96.235 -65.159 53.182
436352 96.277 -65.160 53.142
436353 96.320 -65.161 53.104
436354 96.363 -65.163 53.064
436355 96.404 -65.165 53.022
436356 96.444 -65.171 52.980
436357 96.484 -65.176 52.938
436358 96.524 -65.182 52.897
436359 96.565 -65.188 52.858
436360 96.605 -65.195 52.818
436361 96.645 -65.200 52.778
436362 96.685 -65.206 52.740
436363 96.725 -65.212 52.700
436364 96.765 -65.218 52.661
436365 96.805 -65.223 52.621
436366 96.841 -65.227 52.579
436367 96.860 -65.220 52.533
436368 96.879 -65.213 52.487
436369 96.915 -65.210 52.442
436370 96.957 -65.210 52.399
436371 96.999 -65.210 52.357
436372 97.031 -65.206 52.314
436373 97.058 -65.208 52.270
436374 97.094 -65.216 52.224
436375 97.129 -65.223 52.177
436376 97.164 -65.231 52.130
436377 97.203 -65.232 52.085
436378 97.242 -65.233 52.041
436379 97.282 -65.234 51.997
436380 97.317 -65.238 51.953
436381 97.350 -65.242 51.910
436382 97.389 -65.249 51.867
436383 97.428 -65.257 51.826
436384 97.467 -65.264 51.789
436385 97.506 -65.271 51.755
436386 97.544 -65.279 51.717
436387 97.584 -65.286 51.677
436388 97.623 -65.293 51.637
436389 97.663 -65.299 51.598
436390 97.704 -65.304 51.562
436391 97.745 -65.309 51.525
436392 97.780 -65.308 51.487
436393 97.813 -65.307 51.449
436394 97.846 -65.304 51.414
436395 97.878 -65.303 51.379
436396 97.908 -65.306 51.344
436397 97.936 -65.311 51.310
436398 97.965 -65.317 51.274
436399 97.993 -65.322 51.237
436400 98.022 -65.327 51.198
436401 98.051 -65.333 51.161
436402 98.079 -65.338 51.126
436403 98.111 -65.342 51.093
436404 98.149 -65.348 51.058
436405 98.186 -65.352 51.023
436406 98.224 -65.357 50.990
436407 98.261 -65.361 50.958
436408 98.298 -65.367 50.926
436409 98.336 -65.371 50.893
436410 98.373 -65.376 50.861
436411 98.411 -65.380 50.827
436412 98.448 -65.386 50.790
436413 98.486 -65.390 50.748
436414 98.525 -65.395 50.700
436415 98.569 -65.398 50.648
436416 98.614 -65.402 50.592
436417 98.656 -65.408 50.535
436418 98.698 -65.414 50.478
436419 98.740 -65.421 50.422
436420 98.782 -65.428 50.365
436421 98.824 -65.434 50.308
436422 98.861 -65.439 50.250
436423 98.899 -65.445 50.191
436424 98.943 -65.449 50.135
436425 98.992 -65.454 50.078
436426 99.042 -65.458 50.023
436427 99.090 -65.462 49.970
436428 99.135 -65.465 49.921
436429 99.172 -65.465 49.877
436430 99.210 -65.466 49.837
436431 99.247 -65.467 49.799
436432 99.296 -65.471 49.761
436433 99.351 -65.477 49.720
436434 99.405 -65.483 49.677
436435 99.458 -65.487 49.632
436436 99.513 -65.490 49.586
436437 99.567 -65.493 49.539
436438 99.622 -65.496 49.493
436439 99.677 -65.499 49.445
436440 99.723 -65.508 49.396
436441 99.767 -65.518 49.346
436442 99.811 -65.528 49.296
436443 99.866 -65.529 49.248
436444 99.949 -65.504 49.201
436445 100.032 -65.480 49.153
436446 100.034 -65.533 49.104
436447 100.054 -65.499 49.055
436448 100.081 -65.472 49.006
436449 100.120 -65.460 48.958
436450 100.165 -65.454 48.912
436451 100.214 -65.452 48.867
436452 100.262 -65.450 48.823
436453 100.311 -65.448 48.778
436454 100.358 -65.445 48.734
436455 100.407 -65.444 48.690
436456 100.450 -65.445 48.648
436457 100.489 -65.451 48.607
436458 100.527 -65.456 48.565
436876 88.449 -64.139 55.311
436877 88.497 -64.142 55.311
436878 88.560 -64.138 55.313
436879 88.623 -64.133 55.315
436880 88.686 -64.129 55.317
436881 88.749 -64.125 55.319
436882 88.812 -64.120 55.321
436883 88.876 -64.117 55.322
436884 88.924 -64.126 55.323
436885 88.973 -64.137 55.325
436886 89.022 -64.148 55.325
436887 89.071 -64.157 55.323
436888 89.113 -64.172 55.321
436889 89.149 -64.189 55.318
436890 89.185 -64.205 55.314
436891 89.222 -64.222 55.313
436892 89.258 -64.239 55.311
436893 89.296 -64.258 55.309
436894 89.332 -64.276 55.306
436895 89.369 -64.295 55.305
436896 89.406 -64.313 55.301
436897 89.443 -64.332 55.292
436898 89.480 -64.350 55.278
436899 89.517 -64.369 55.258
436900 89.553 -64.387 55.235
436901 89.594 -64.402 55.214
436902 89.641 -64.404 55.193
436903 89.688 -64.406 55.173
436904 89.736 -64.408 55.155
436905 89.783 -64.411 55.137
436906 89.830 -64.413 55.122
436907 89.878 -64.415 55.109
436908 89.925 -64.418 55.099
436909 89.972 -64.420 55.091
436910 90.020 -64.422 55.085
436911 90.067 -64.424 55.082
436912 90.114 -64.426 55.080
436913 90.162 -64.428 55.080
436914 90.209 -64.430 55.081
436915 90.256 -64.433 55.086
436916 90.303 -64.435 55.090
436917 90.351 -64.437 55.094
436918 90.399 -64.438 55.098
436919 90.449 -64.436 55.101
436920 90.498 -64.436 55.102
436921 90.548 -64.434 55.101
436922 90.595 -64.439 55.097
436923 90.641 -64.451 55.092
436924 90.687 -64.458 55.082
436925 90.733 -64.466 55.066
436926 90.778 -64.475 55.046
436927 90.823 -64.486 55.023
436928 90.869 -64.496 54.999
436929 90.913 -64.505 54.968
436930 90.959 -64.515 54.942
436931 91.004 -64.525 54.920
436932 91.049 -64.535 54.900
436933 91.094 -64.545 54.882
436934 91.139 -64.555 54.866
436935 91.184 -64.565 54.851
436936 91.230 -64.575 54.835
436937 91.275 -64.585 54.820
436938 91.320 -64.595 54.806
436939 91.365 -64.605 54.793
436940 91.410 -64.615 54.780
436941 91.455 -64.624 54.768
436942 91.501 -64.635 54.757
436943 91.545 -64.645 54.746
436944 91.591 -64.655 54.737
436945 91.636 -64.665 54.729
436946 91.681 -64.674 54.721
436947 91.726 -64.684 54.713
436948 91.772 -64.694 54.707
436949 91.816 -64.704 54.700
436950 91.862 -64.715 54.694
436951 91.907 -64.725 54.688
436952 91.952 -64.734 54.683
436953 91.997 -64.744 54.677
436954 92.042 -64.754 54.670
436955 92.087 -64.764 54.664
436956 92.133 -64.774 54.657
436957 92.178 -64.784 54.651
436958 92.224 -64.794 54.647
436959 92.269 -64.825 54.642
436960 92.313 -64.855 54.638
436961 92.362 -64.873 54.635
436962 92.411 -64.874 54.631
436963 92.460 -64.875 54.627
436964 92.509 -64.875 54.625
436965 92.558 -64.875 54.622
436966 92.607 -64.875 54.619
436967 92.657 -64.877 54.615
436968 92.712 -64.882 54.611
436969 92.766 -64.889 54.607
436970 92.815 -64.902 54.604
436971 92.864 -64.915 54.602
436972 92.914 -64.928 54.598
436973 92.963 -64.941 54.595
436974 93.011 -64.952 54.591
436975 93.057 -64.954 54.589
436976 93.103 -64.957 54.589
436977 93.148 -64.959 54.589
436978 93.194 -64.962 54.591
436979 93.240 -64.964 54.594
436980 93.285 -64.966 54.594
436981 93.333 -64.966 54.596
436982 93.382 -64.963 54.597
436983 93.427 -64.966 54.597
436984 93.472 -64.969 54.599
436985 93.517 -64.972 54.600
436986 93.562 -64.975 54.601
436987 93.607 -64.978 54.602
436988 93.652 -64.982 54.604
436989 93.697 -64.985 54.606
436990 93.743 -64.989 54.607
436991 93.788 -64.993 54.606
436992 93.834 -64.997 54.603
436993 93.880 -65.001 54.599
436994 93.925 -65.005 54.592
436995 93.969 -65.008 54.586
436996 94.013 -65.011 54.578
436997 94.047 -65.007 54.566
436998 94.082 -65.009 54.551
436999 94.116 -65.010 54.533
437000 94.151 -65.012 54.514
437001 94.186 -65.013 54.492
437002 94.220 -65.015 54.470
437003 94.255 -65.016 54.448
437004 94.290 -65.018 54.426
437005 94.325 -65.019 54.402
437006 94.359 -65.021 54.379
437007 94.394 -65.022 54.357
437008 94.428 -65.024 54.337
437009 94.463 -65.025 54.317
437010 94.498 -65.027 54.298
437011 94.539 -65.029 54.277
437012 94.583 -65.033 54.257
437013 94.627 -65.037 54.238
437014 94.671 -65.041 54.217
437015 94.715 -65.044 54.193
437016 94.760 -65.048 54.169
437017 94.804 -65.052 54.144
437018 94.848 -65.056 54.119
437019 94.892 -65.060 54.098
437020 94.936 -65.063 54.075
437021 94.979 -65.067 54.053
437022 95.019 -65.073 54.032
437023 95.059 -65.079 54.012
437024 95.099 -65.085 53.990
437025 95.139 -65.091 53.968
437026 95.179 -65.096 53.946
437027 95.219 -65.102 53.922
437028 95.257 -65.110 53.897
437029 95.283 -65.129 53.868
437030 95.310 -65.150 53.836
437031 95.336 -65.170 53.805
437032 95.375 -65.176 53.775
437033 95.414 -65.181 53.749
437034 95.458 -65.182 53.727
437035 95.506 -65.181 53.710
437036 95.554 -65.179 53.694
437037 95.602 -65.179 53.679
437038 95.650 -65.177 53.661
437039 95.693 -65.180 53.638
437040 95.735 -65.176 53.612
437041 95.776 -65.172 53.581
437042 95.819 -65.164 53.547
437043 95.862 -65.157 53.509
437044 95.905 -65.150 53.471
437045 95.948 -65.142 53.433
437046 95.990 -65.135 53.394
437047 96.033 -65.128 53.356
437048 96.075 -65.122 53.317
437049 96.114 -65.117 53.280
437050 96.154 -65.113 53.243
437051 96.195 -65.113 53.206
437052 96.238 -65.114 53.170
437053 96.281 -65.116 53.132
437054 96.323 -65.116 53.095
437055 96.366 -65.118 53.056
437056 96.406 -65.126 53.016
437057 96.448 -65.134 52.975
437058 96.488 -65.141 52.933
437059 96.530 -65.148 52.892
437060 96.571 -65.156 52.850
437061 96.612 -65.163 52.809
437062 96.653 -65.170 52.768
437063 96.693 -65.176 52.728
437064 96.733 -65.181 52.689
437065 96.773 -65.187 52.650
437066 96.813 -65.193 52.610
437067 96.853 -65.199 52.569
437068 96.874 -65.192 52.524
437069 96.893 -65.185 52.479
437070 96.912 -65.179 52.434
437071 96.949 -65.176 52.388
437072 96.991 -65.176 52.345
437073 97.019 -65.175 52.302
437074 97.054 -65.182 52.259
437075 97.090 -65.190 52.213
437076 97.129 -65.191 52.167
437077 97.168 -65.192 52.121
437078 97.208 -65.192 52.075
437079 97.247 -65.193 52.031
437080 97.287 -65.195 51.989
437081 97.323 -65.197 51.944
437082 97.357 -65.201 51.901
437083 97.390 -65.205 51.856
437084 97.426 -65.210 51.813
437085 97.465 -65.218 51.771
437086 97.504 -65.226 51.734
437087 97.542 -65.232 51.696
437088 97.581 -65.240 51.657
437089 97.620 -65.248 51.617
437090 97.660 -65.253 51.578
437091 97.702 -65.258 51.541
437092 97.743 -65.263 51.505
437093 97.785 -65.267 51.467
437094 97.819 -65.267 51.429
437095 97.852 -65.265 51.393
437096 97.884 -65.264 51.357
437097 97.915 -65.266 51.320
437098 97.943 -65.271 51.284
437099 97.971 -65.276 51.248
437100 98.000 -65.282 51.210
437101 98.028 -65.287 51.174
437102 98.057 -65.292 51.138
437103 98.081 -65.296 51.105
437104 98.112 -65.300 51.071
437105 98.151 -65.302 51.036
437106 98.189 -65.306 51.002
437107 98.226 -65.311 50.971
437108 98.264 -65.315 50.941
437109 98.301 -65.320 50.909
437110 98.339 -65.325 50.876
437111 98.376 -65.329 50.841
437112 98.414 -65.334 50.803
437113 98.451 -65.339 50.764
437114 98.488 -65.344 50.721
437115 98.527 -65.348 50.674
437116 98.572 -65.352 50.622
437117 98.616 -65.357 50.566
437118 98.657 -65.363 50.509
437119 98.700 -65.370 50.452
437120 98.742 -65.376 50.398
437121 98.781 -65.383 50.343
437122 98.818 -65.388 50.286
437123 98.856 -65.392 50.229
437124 98.893 -65.398 50.170
437125 98.938 -65.402 50.114
437126 98.987 -65.407 50.058
437127 99.035 -65.411 50.004
437128 99.085 -65.415 49.953
437129 99.132 -65.419 49.906
437130 99.175 -65.422 49.865
437131 99.212 -65.423 49.828
437132 99.250 -65.423 49.793
437133 99.300 -65.428 49.754
437134 99.354 -65.433 49.714
437135 99.408 -65.438 49.672
437136 99.463 -65.441 49.626
437137 99.517 -65.444 49.579
437138 99.572 -65.447 49.532
437139 99.626 -65.450 49.484
437140 99.681 -65.453 49.436
437141 99.728 -65.461 49.386
437142 99.772 -65.471 49.335
437143 99.816 -65.482 49.285
437144 99.885 -65.470 49.237
437145 99.968 -65.445 49.190
437146 100.051 -65.420 49.142
437147 100.051 -65.477 49.094
437148 100.068 -65.452 49.044
437149 100.095 -65.425 48.996
437150 100.134 -65.412 48.949
437151 100.180 -65.408 48.904
437152 100.228 -65.406 48.859
437153 100.277 -65.404 48.815
437154 100.325 -65.402 48.771
437155 100.373 -65.400 48.727
437156 100.421 -65.398 48.684
437157 100.460 -65.403 48.643
437158 100.498 -65.408 48.600
437159 100.537 -65.414 48.560
437577 88.460 -64.082 55.312
437578 88.523 -64.078 55.312
437579 88.586 -64.073 55.313
437580 88.649 -64.069 55.314
437581 88.712 -64.064 55.317
437582 88.775 -64.060 55.318
437583 88.828 -64.067 55.319
437584 88.877 -64.077 55.320
437585 88.926 -64.087 55.321
437586 88.975 -64.098 55.322
437587 89.024 -64.108 55.322
437588 89.073 -64.118 55.321
437589 89.118 -64.130 55.320
437590 89.155 -64.147 55.319
437591 89.191 -64.164 55.318
437592 89.228 -64.180 55.320
437593 89.264 -64.197 55.321
437594 89.301 -64.214 55.321
437595 89.337 -64.230 55.320
437596 89.374 -64.248 55.319
437597 89.411 -64.267 55.314
437598 89.448 -64.286 55.305
437599 89.484 -64.304 55.290
437600 89.522 -64.323 55.271
437601 89.558 -64.341 55.248
437602 89.595 -64.360 55.226
437603 89.642 -64.363 55.205
437604 89.689 -64.365 55.186
437605 89.737 -64.367 55.168
437606 89.784 -64.370 55.151
437607 89.831 -64.372 55.136
437608 89.879 -64.374 55.123
437609 89.926 -64.377 55.113
437610 89.973 -64.379 55.105
437611 90.021 -64.381 55.099
437612 90.068 -64.383 55.097
437613 90.115 -64.386 55.097
437614 90.163 -64.388 55.097
437615 90.211 -64.389 55.101
437616 90.260 -64.388 55.108
437617 90.309 -64.387 55.115
437618 90.359 -64.386 55.120
437619 90.408 -64.384 55.124
437620 90.458 -64.383 55.127
437621 90.508 -64.382 55.130
437622 90.557 -64.380 55.129
437623 90.602 -64.396 55.124
437624 90.647 -64.407 55.115
437625 90.693 -64.414 55.102
437626 90.739 -64.422 55.086
437627 90.786 -64.430 55.065
437628 90.831 -64.437 55.039
437629 90.877 -64.446 55.010
437630 90.922 -64.456 54.976
437631 90.967 -64.466 54.947
437632 91.012 -64.477 54.922
437633 91.058 -64.486 54.900
437634 91.102 -64.496 54.881
437635 91.148 -64.506 54.864
437636 91.193 -64.516 54.848
437637 91.238 -64.526 54.832
437638 91.283 -64.536 54.817
437639 91.328 -64.546 54.801
437640 91.373 -64.556 54.789
437641 91.419 -64.566 54.776
437642 91.464 -64.576 54.764
437643 91.509 -64.586 54.752
437644 91.554 -64.596 54.741
437645 91.599 -64.606 54.732
437646 91.644 -64.615 54.723
437647 91.689 -64.626 54.714
437648 91.735 -64.636 54.706
437649 91.780 -64.646 54.699
437650 91.825 -64.656 54.693
437651 91.870 -64.665 54.686
437652 91.915 -64.675 54.680
437653 91.960 -64.685 54.674
437654 92.006 -64.696 54.669
437655 92.051 -64.706 54.662
437656 92.096 -64.715 54.656
437657 92.141 -64.725 54.650
437658 92.186 -64.735 54.644
437659 92.233 -64.745 54.640
437660 92.278 -64.772 54.636
437661 92.322 -64.802 54.631
437662 92.367 -64.821 54.627
437663 92.411 -64.824 54.623
437664 92.458 -64.825 54.621
437665 92.507 -64.826 54.618
437666 92.556 -64.826 54.615
437667 92.604 -64.827 54.611
437668 92.654 -64.828 54.607
437669 92.709 -64.834 54.603
437670 92.760 -64.844 54.599
437671 92.809 -64.857 54.596
437672 92.858 -64.870 54.592
437673 92.908 -64.883 54.588
437674 92.956 -64.897 54.586
437675 93.006 -64.910 54.584
437676 93.055 -64.922 54.584
437677 93.100 -64.924 54.584
437678 93.146 -64.926 54.586
437679 93.192 -64.929 54.589
437680 93.237 -64.931 54.591
437681 93.283 -64.934 54.592
437682 93.335 -64.926 54.594
437683 93.384 -64.922 54.595
437684 93.429 -64.925 54.596
437685 93.474 -64.928 54.596
437686 93.519 -64.931 54.596
437687 93.564 -64.935 54.597
437688 93.609 -64.938 54.597
437689 93.654 -64.941 54.598
437690 93.699 -64.944 54.599
437691 93.744 -64.947 54.599
437692 93.789 -64.950 54.597
437693 93.834 -64.953 54.593
437694 93.879 -64.957 54.588
437695 93.924 -64.960 54.582
437696 93.970 -64.963 54.576
437697 94.015 -64.967 54.569
437698 94.052 -64.968 54.558
437699 94.086 -64.969 54.544
437700 94.121 -64.971 54.526
437701 94.155 -64.972 54.506
437702 94.190 -64.974 54.485
437703 94.224 -64.975 54.462
437704 94.258 -64.977 54.439
437705 94.292 -64.979 54.417
437706 94.326 -64.981 54.392
437707 94.359 -64.982 54.369
437708 94.394 -64.985 54.347
437709 94.427 -64.986 54.327
437710 94.461 -64.987 54.308
437711 94.494 -64.985 54.288
437712 94.552 -64.992 54.267
437713 94.600 -64.997 54.245
437714 94.643 -65.000 54.224
437715 94.686 -65.004 54.203
437716 94.730 -65.008 54.180
437717 94.773 -65.012 54.157
437718 94.817 -65.016 54.135
437719 94.860 -65.019 54.113
437720 94.903 -65.022 54.094
437721 94.944 -65.026 54.074
437722 94.984 -65.032 54.052
437723 95.024 -65.038 54.031
437724 95.063 -65.044 54.011
437725 95.103 -65.050 53.989
437726 95.143 -65.056 53.966
437727 95.183 -65.061 53.943
437728 95.223 -65.067 53.920
437729 95.263 -65.073 53.896
437730 95.290 -65.092 53.867
437731 95.317 -65.112 53.835
437732 95.343 -65.132 53.802
437733 95.372 -65.149 53.770
437734 95.411 -65.154 53.742
437735 95.458 -65.154 53.719
437736 95.506 -65.153 53.701
437737 95.553 -65.152 53.686
437738 95.602 -65.151 53.671
437739 95.650 -65.145 53.655
437740 95.698 -65.137 53.633
437741 95.741 -65.130 53.608
437742 95.783 -65.123 53.578
437743 95.826 -65.116 53.545
437744 95.869 -65.108 53.506
437745 95.912 -65.101 53.467
437746 95.955 -65.094 53.428
437747 95.997 -65.086 53.389
437748 96.038 -65.081 53.351
437749 96.078 -65.076 53.312
437750 96.117 -65.072 53.273
437751 96.156 -65.067 53.234
437752 96.199 -65.068 53.196
437753 96.241 -65.069 53.158
437754 96.284 -65.070 53.121
437755 96.327 -65.072 53.084
437756 96.368 -65.078 53.046
437757 96.408 -65.086 53.008
437758 96.449 -65.094 52.968
437759 96.490 -65.101 52.927
437760 96.531 -65.109 52.886
437761 96.573 -65.116 52.844
437762 96.614 -65.124 52.802
437763 96.655 -65.131 52.759
437764 96.696 -65.138 52.718
437765 96.737 -65.146 52.678
437766 96.779 -65.154 52.639
437767 96.820 -65.160 52.599
437768 96.861 -65.168 52.559
437769 96.887 -65.166 52.515
437770 96.906 -65.158 52.471
437771 96.926 -65.151 52.426
437772 96.945 -65.144 52.381
437773 96.979 -65.141 52.336
437774 97.016 -65.148 52.293
437775 97.055 -65.149 52.248
437776 97.095 -65.150 52.202
437777 97.134 -65.151 52.156
437778 97.173 -65.151 52.110
437779 97.213 -65.152 52.065
437780 97.252 -65.154 52.021
437781 97.292 -65.154 51.978
437782 97.330 -65.156 51.933
437783 97.363 -65.160 51.890
437784 97.397 -65.165 51.846
437785 97.431 -65.169 51.802
437786 97.464 -65.173 51.759
437787 97.501 -65.179 51.719
437788 97.540 -65.187 51.678
437789 97.579 -65.194 51.638
437790 97.618 -65.201 51.598
437791 97.658 -65.207 51.558
437792 97.700 -65.212 51.521
437793 97.741 -65.217 51.484
437794 97.783 -65.222 51.447
437795 97.824 -65.226 51.410
437796 97.859 -65.226 51.372
437797 97.891 -65.224 51.335
437798 97.921 -65.226 51.297
437799 97.950 -65.232 51.259
437800 97.979 -65.237 51.221
437801 98.007 -65.242 51.184
437802 98.033 -65.246 51.148
437803 98.057 -65.250 51.114
437804 98.080 -65.254 51.082
437805 98.114 -65.257 51.049
437806 98.152 -65.261 51.012
437807 98.190 -65.263 50.980
437808 98.228 -65.266 50.950
437809 98.266 -65.269 50.920
437810 98.304 -65.273 50.889
437811 98.341 -65.279 50.855
437812 98.379 -65.283 50.816
437813 98.416 -65.288 50.775
437814 98.453 -65.292 50.733
437815 98.491 -65.298 50.688
437816 98.530 -65.302 50.643
437817 98.574 -65.305 50.593
437818 98.617 -65.311 50.540
437819 98.659 -65.318 50.485
437820 98.701 -65.325 50.429
437821 98.739 -65.330 50.374
437822 98.776 -65.336 50.318
437823 98.813 -65.341 50.261
437824 98.850 -65.345 50.204
437825 98.887 -65.351 50.146
437826 98.932 -65.355 50.090
437827 98.981 -65.360 50.036
437828 99.030 -65.364 49.984
437829 99.078 -65.368 49.936
437830 99.125 -65.372 49.892
437831 99.171 -65.376 49.851
437832 99.215 -65.379 49.816
437833 99.252 -65.380 49.782
437834 99.304 -65.384 49.745
437835 99.358 -65.389 49.705
437836 99.413 -65.392 49.662
437837 99.467 -65.395 49.616
437838 99.521 -65.398 49.569
437839 99.576 -65.401 49.522
437840 99.630 -65.404 49.474
437841 99.685 -65.407 49.424
437842 99.732 -65.414 49.374
437843 99.777 -65.425 49.323
437844 99.821 -65.435 49.272
437845 99.904 -65.411 49.224
437846 99.987 -65.386 49.177
437847 100.070 -65.361 49.130
437848 100.066 -65.421 49.082
437849 100.081 -65.405 49.033
437850 100.109 -65.378 48.986
437851 100.147 -65.365 48.940
437852 100.195 -65.362 48.895
437853 100.243 -65.360 48.851
437854 100.291 -65.358 48.807
437855 100.339 -65.356 48.763
437856 100.388 -65.354 48.719
437857 100.431 -65.355 48.677
437858 100.470 -65.361 48.636
437859 100.508 -65.365 48.594
437860 100.547 -65.370 48.554
438279 88.548 -64.013 55.315
438280 88.612 -64.008 55.315
438281 88.675 -64.004 55.315
438282 88.732 -64.006 55.317
438283 88.780 -64.016 55.318
438284 88.830 -64.027 55.319
438285 88.879 -64.038 55.320
438286 88.927 -64.048 55.320
438287 88.976 -64.058 55.321
438288 89.025 -64.069 55.322
438289 89.074 -64.079 55.322
438290 89.123 -64.089 55.322
438291 89.161 -64.105 55.323
438292 89.197 -64.122 55.324
438293 89.234 -64.139 55.329
438294 89.270 -64.155 55.333
438295 89.307 -64.172 55.335
438296 89.343 -64.189 55.335
438297 89.379 -64.205 55.332
438298 89.416 -64.222 55.326
438299 89.452 -64.239 55.317
438300 89.489 -64.258 55.304
438301 89.526 -64.276 55.285
438302 89.563 -64.295 55.263
438303 89.600 -64.314 55.240
438304 89.643 -64.322 55.219
438305 89.690 -64.324 55.200
438306 89.738 -64.327 55.181
438307 89.785 -64.329 55.164
438308 89.832 -64.331 55.149
438309 89.880 -64.333 55.136
438310 89.927 -64.336 55.126
438311 89.974 -64.338 55.118
438312 90.022 -64.340 55.114
438313 90.071 -64.339 55.112
438314 90.120 -64.338 55.113
438315 90.170 -64.337 55.116
438316 90.220 -64.336 55.123
438317 90.269 -64.334 55.133
438318 90.319 -64.333 55.142
438319 90.368 -64.332 55.147
438320 90.417 -64.330 55.152
438321 90.467 -64.329 55.154
438322 90.517 -64.328 55.158
438323 90.564 -64.334 55.157
438324 90.608 -64.351 55.151
438325 90.653 -64.363 55.139
438326 90.699 -64.371 55.124
438327 90.745 -64.378 55.105
438328 90.792 -64.386 55.081
438329 90.838 -64.393 55.051
438330 90.884 -64.401 55.017
438331 90.930 -64.408 54.981
438332 90.975 -64.418 54.950
438333 91.021 -64.427 54.923
438334 91.066 -64.437 54.899
438335 91.111 -64.447 54.880
438336 91.156 -64.458 54.861
438337 91.201 -64.468 54.844
438338 91.246 -64.477 54.827
438339 91.292 -64.487 54.813
438340 91.336 -64.497 54.798
438341 91.382 -64.507 54.784
438342 91.427 -64.517 54.771
438343 91.472 -64.527 54.759
438344 91.517 -64.537 54.747
438345 91.563 -64.547 54.736
438346 91.607 -64.557 54.726
438347 91.653 -64.567 54.717
438348 91.698 -64.577 54.708
438349 91.743 -64.587 54.700
438350 91.788 -64.596 54.693
438351 91.834 -64.607 54.686
438352 91.878 -64.617 54.680
438353 91.924 -64.627 54.673
438354 91.969 -64.637 54.666
438355 92.014 -64.646 54.662
438356 92.059 -64.656 54.656
438357 92.104 -64.666 54.649
438358 92.149 -64.676 54.644
438359 92.195 -64.687 54.638
438360 92.242 -64.696 54.634
438361 92.288 -64.719 54.629
438362 92.333 -64.749 54.625
438363 92.376 -64.765 54.621
438364 92.421 -64.769 54.618
438365 92.464 -64.772 54.616
438366 92.508 -64.775 54.613
438367 92.554 -64.778 54.610
438368 92.602 -64.778 54.608
438369 92.652 -64.779 54.603
438370 92.704 -64.787 54.599
438371 92.754 -64.800 54.595
438372 92.803 -64.813 54.590
438373 92.852 -64.826 54.586
438374 92.901 -64.839 54.583
438375 92.950 -64.852 54.581
438376 93.000 -64.865 54.580
438377 93.049 -64.878 54.581
438378 93.098 -64.891 54.582
438379 93.144 -64.894 54.584
438380 93.190 -64.897 54.587
438381 93.235 -64.899 54.590
438382 93.285 -64.894 54.592
438383 93.337 -64.887 54.594
438384 93.387 -64.881 54.595
438385 93.432 -64.884 54.595
438386 93.477 -64.888 54.594
438387 93.522 -64.891 54.594
438388 93.567 -64.894 54.594
438389 93.612 -64.897 54.592
438390 93.657 -64.900 54.591
438391 93.702 -64.903 54.591
438392 93.747 -64.906 54.589
438393 93.792 -64.910 54.585
438394 93.837 -64.913 54.580
438395 93.882 -64.916 54.575
438396 93.927 -64.919 54.569
438397 93.969 -64.920 54.564
438398 94.010 -64.919 54.558
438399 94.046 -64.928 54.548
438400 94.081 -64.932 54.535
438401 94.115 -64.935 54.517
438402 94.149 -64.936 54.498
438403 94.183 -64.938 54.476
438404 94.217 -64.940 54.453
438405 94.251 -64.941 54.429
438406 94.285 -64.943 54.406
438407 94.319 -64.944 54.381
438408 94.352 -64.947 54.357
438409 94.386 -64.946 54.336
438410 94.418 -64.944 54.316
438411 94.451 -64.943 54.298
438412 94.484 -64.941 54.277
438413 94.545 -64.949 54.255
438414 94.607 -64.957 54.233
438415 94.668 -64.966 54.212
438416 94.711 -64.969 54.191
438417 94.754 -64.972 54.168
438418 94.797 -64.976 54.146
438419 94.841 -64.980 54.125
438420 94.884 -64.984 54.104
438421 94.919 -64.987 54.086
438422 94.949 -64.988 54.068
438423 94.985 -64.994 54.048
438424 95.028 -65.004 54.029
438425 95.068 -65.009 54.009
438426 95.108 -65.015 53.987
438427 95.147 -65.021 53.964
438428 95.188 -65.026 53.941
438429 95.228 -65.032 53.918
438430 95.267 -65.038 53.893
438431 95.298 -65.054 53.865
438432 95.324 -65.074 53.834
438433 95.350 -65.094 53.800
438434 95.376 -65.115 53.767
438435 95.409 -65.128 53.736
438436 95.457 -65.126 53.711
438437 95.506 -65.121 53.692
438438 95.554 -65.113 53.676
438439 95.604 -65.105 53.661
438440 95.653 -65.097 53.645
438441 95.701 -65.089 53.626
438442 95.748 -65.081 53.602
438443 95.791 -65.074 53.574
438444 95.833 -65.066 53.543
438445 95.876 -65.060 53.505
438446 95.919 -65.052 53.466
438447 95.962 -65.044 53.428
438448 96.002 -65.040 53.390
438449 96.041 -65.035 53.351
438450 96.081 -65.031 53.312
438451 96.120 -65.026 53.271
438452 96.159 -65.022 53.228
438453 96.202 -65.023 53.187
438454 96.245 -65.024 53.148
438455 96.287 -65.025 53.109
438456 96.329 -65.029 53.072
438457 96.369 -65.038 53.035
438458 96.409 -65.046 52.997
438459 96.450 -65.054 52.958
438460 96.491 -65.062 52.919
438461 96.532 -65.069 52.880
438462 96.574 -65.076 52.839
438463 96.615 -65.084 52.796
438464 96.656 -65.091 52.751
438465 96.697 -65.099 52.709
438466 96.739 -65.106 52.668
438467 96.780 -65.113 52.627
438468 96.821 -65.121 52.587
438469 96.854 -65.123 52.547
438470 96.876 -65.119 52.504
438471 96.898 -65.114 52.462
438472 96.921 -65.110 52.418
438473 96.944 -65.105 52.374
438474 96.982 -65.108 52.329
438475 97.021 -65.108 52.284
438476 97.060 -65.109 52.239
438477 97.100 -65.110 52.193
438478 97.139 -65.110 52.147
438479 97.178 -65.112 52.101
438480 97.218 -65.113 52.056
438481 97.257 -65.113 52.011
438482 97.297 -65.114 51.967
438483 97.336 -65.115 51.922
438484 97.370 -65.120 51.878
438485 97.403 -65.124 51.835
438486 97.437 -65.128 51.791
438487 97.470 -65.132 51.748
438488 97.504 -65.137 51.706
438489 97.538 -65.141 51.664
438490 97.577 -65.148 51.622
438491 97.615 -65.156 51.582
438492 97.656 -65.161 51.541
438493 97.698 -65.166 51.502
438494 97.739 -65.171 51.464
438495 97.781 -65.176 51.427
438496 97.822 -65.180 51.389
438497 97.864 -65.185 51.351
438498 97.897 -65.184 51.313
438499 97.928 -65.186 51.275
438500 97.957 -65.192 51.236
438501 97.985 -65.197 51.197
438502 98.008 -65.201 51.159
438503 98.032 -65.204 51.124
438504 98.056 -65.208 51.090
438505 98.080 -65.212 51.057
438506 98.115 -65.215 51.025
438507 98.153 -65.218 50.989
438508 98.191 -65.221 50.957
438509 98.229 -65.223 50.927
438510 98.268 -65.226 50.896
438511 98.305 -65.229 50.865
438512 98.344 -65.232 50.831
438513 98.381 -65.237 50.790
438514 98.419 -65.242 50.748
438515 98.456 -65.246 50.704
438516 98.494 -65.251 50.659
438517 98.533 -65.255 50.615
438518 98.576 -65.260 50.566
438519 98.619 -65.267 50.514
438520 98.659 -65.273 50.461
438521 98.696 -65.278 50.405
438522 98.733 -65.283 50.350
438523 98.771 -65.289 50.293
438524 98.808 -65.294 50.236
438525 98.845 -65.299 50.179
438526 98.882 -65.304 50.122
438527 98.926 -65.309 50.068
438528 98.975 -65.313 50.015
438529 99.024 -65.317 49.965
438530 99.072 -65.321 49.918
438531 99.118 -65.325 49.875
438532 99.165 -65.329 49.836
438533 99.211 -65.333 49.803
438534 99.254 -65.336 49.770
438535 99.308 -65.339 49.733
438536 99.362 -65.342 49.693
438537 99.417 -65.345 49.650
438538 99.471 -65.348 49.604
438539 99.525 -65.351 49.557
438540 99.580 -65.355 49.510
438541 99.635 -65.358 49.461
438542 99.689 -65.361 49.411
438543 99.737 -65.368 49.360
438544 99.781 -65.378 49.309
438545 99.840 -65.376 49.259
438546 99.923 -65.351 49.211
438547 100.006 -65.326 49.164
438548 100.089 -65.302 49.118
438549 100.082 -65.364 49.071
438550 100.094 -65.358 49.022
438551 100.122 -65.331 48.975
438552 100.160 -65.318 48.931
438553 100.209 -65.316 48.887
438554 100.257 -65.314 48.843
438555 100.305 -65.312 48.799
438556 100.354 -65.310 48.755
438557 100.402 -65.308 48.712
438558 100.441 -65.312 48.670
438559 100.480 -65.317 48.629
438560 100.519 -65.323 48.589
438561 100.557 -65.328 48.548
438981 88.635 -63.946 55.316
438982 88.684 -63.957 55.316
438983 88.733 -63.966 55.318
438984 88.782 -63.977 55.319
438985 88.831 -63.988 55.320
438986 88.880 -63.998 55.320
438987 88.929 -64.008 55.321
438988 88.978 -64.019 55.322
438989 89.027 -64.029 55.322
438990 89.076 -64.039 55.324
438991 89.125 -64.050 55.326
438992 89.166 -64.064 55.328
438993 89.203 -64.081 55.333
438994 89.239 -64.098 55.340
438995 89.276 -64.114 55.347
438996 89.312 -64.130 55.351
438997 89.349 -64.147 55.350
438998 89.385 -64.164 55.345
438999 89.421 -64.180 55.338
439000 89.458 -64.197 55.329
439001 89.494 -64.214 55.318
439002 89.531 -64.230 55.301
439003 89.568 -64.248 55.279
439004 89.605 -64.267 55.256
439005 89.645 -64.282 55.234
439006 89.692 -64.283 55.214
439007 89.739 -64.286 55.195
439008 89.787 -64.288 55.178
439009 89.834 -64.290 55.163
439010 89.882 -64.291 55.150
439011 89.932 -64.289 55.140
439012 89.981 -64.289 55.133
439013 90.031 -64.287 55.129
439014 90.080 -64.286 55.128
439015 90.129 -64.285 55.129
439016 90.179 -64.283 55.133
439017 90.229 -64.282 55.142
439018 90.278 -64.280 55.155
439019 90.328 -64.280 55.166
439020 90.377 -64.278 55.174
439021 90.427 -64.276 55.179
439022 90.476 -64.276 55.182
439023 90.526 -64.274 55.186
439024 90.570 -64.290 55.186
439025 90.614 -64.307 55.180
439026 90.660 -64.319 55.165
439027 90.706 -64.327 55.147
439028 90.752 -64.334 55.123
439029 90.798 -64.342 55.094
439030 90.844 -64.349 55.060
439031 90.890 -64.357 55.022
439032 90.936 -64.364 54.984
439033 90.982 -64.372 54.952
439034 91.029 -64.380 54.924
439035 91.074 -64.389 54.898
439036 91.119 -64.399 54.878
439037 91.164 -64.408 54.858
439038 91.210 -64.418 54.840
439039 91.255 -64.428 54.823
439040 91.300 -64.438 54.808
439041 91.345 -64.449 54.794
439042 91.390 -64.458 54.779
439043 91.435 -64.468 54.766
439044 91.481 -64.478 54.754
439045 91.526 -64.488 54.742
439046 91.571 -64.498 54.731
439047 91.616 -64.508 54.721
439048 91.661 -64.518 54.712
439049 91.706 -64.528 54.704
439050 91.751 -64.538 54.695
439051 91.797 -64.548 54.688
439052 91.842 -64.558 54.681
439053 91.887 -64.568 54.674
439054 91.932 -64.577 54.668
439055 91.977 -64.587 54.662
439056 92.022 -64.598 54.657
439057 92.067 -64.608 54.650
439058 92.112 -64.618 54.644
439059 92.158 -64.627 54.638
439060 92.203 -64.637 54.633
439061 92.251 -64.647 54.628
439062 92.298 -64.666 54.624
439063 92.342 -64.696 54.620
439064 92.386 -64.711 54.616
439065 92.430 -64.714 54.614
439066 92.474 -64.717 54.612
439067 92.518 -64.720 54.610
439068 92.562 -64.723 54.608
439069 92.606 -64.726 54.606
439070 92.649 -64.731 54.602
439071 92.698 -64.743 54.597
439072 92.747 -64.756 54.592
439073 92.797 -64.769 54.586
439074 92.846 -64.782 54.582
439075 92.895 -64.795 54.580
439076 92.944 -64.808 54.579
439077 92.994 -64.821 54.579
439078 93.043 -64.834 54.579
439079 93.092 -64.847 54.581
439080 93.141 -64.859 54.583
439081 93.188 -64.864 54.586
439082 93.236 -64.862 54.590
439083 93.287 -64.855 54.592
439084 93.339 -64.847 54.595
439085 93.389 -64.841 54.596
439086 93.434 -64.844 54.596
439087 93.479 -64.847 54.596
439088 93.524 -64.850 54.595
439089 93.569 -64.853 54.592
439090 93.614 -64.857 54.589
439091 93.659 -64.860 54.586
439092 93.704 -64.863 54.583
439093 93.749 -64.866 54.579
439094 93.794 -64.869 54.574
439095 93.838 -64.872 54.568
439096 93.879 -64.871 54.561
439097 93.919 -64.869 54.554
439098 93.960 -64.868 54.550
439099 93.999 -64.872 54.544
439100 94.035 -64.881 54.536
439101 94.072 -64.890 54.524
439102 94.108 -64.898 54.508
439103 94.142 -64.900 54.487
439104 94.176 -64.902 54.465
439105 94.210 -64.903 54.442
439106 94.244 -64.905 54.417
439107 94.277 -64.907 54.393
439108 94.310 -64.906 54.369
439109 94.343 -64.904 54.345
439110 94.376 -64.902 54.324
439111 94.408 -64.900 54.305
439112 94.441 -64.899 54.287
439113 94.476 -64.898 54.266
439114 94.539 -64.906 54.244
439115 94.601 -64.915 54.221
439116 94.663 -64.923 54.201
439117 94.725 -64.931 54.180
439118 94.779 -64.938 54.157
439119 94.822 -64.941 54.135
439120 94.865 -64.945 54.114
439121 94.895 -64.947 54.094
439122 94.924 -64.949 54.076
439123 94.954 -64.950 54.059
439124 94.985 -64.954 54.041
439125 95.028 -64.963 54.023
439126 95.070 -64.972 54.005
439127 95.112 -64.980 53.984
439128 95.152 -64.986 53.961
439129 95.192 -64.992 53.938
439130 95.232 -64.997 53.914
439131 95.272 -65.004 53.889
439132 95.305 -65.016 53.862
439133 95.331 -65.036 53.830
439134 95.357 -65.057 53.796
439135 95.384 -65.077 53.763
439136 95.417 -65.090 53.730
439137 95.460 -65.088 53.703
439138 95.508 -65.072 53.681
439139 95.557 -65.065 53.663
439140 95.606 -65.057 53.647
439141 95.655 -65.048 53.631
439142 95.704 -65.041 53.614
439143 95.753 -65.032 53.592
439144 95.798 -65.025 53.566
439145 95.840 -65.018 53.537
439146 95.883 -65.010 53.501
439147 95.925 -65.003 53.464
439148 95.965 -64.999 53.427
439149 96.005 -64.994 53.390
439150 96.044 -64.990 53.351
439151 96.083 -64.985 53.312
439152 96.123 -64.981 53.270
439153 96.163 -64.977 53.227
439154 96.205 -64.978 53.184
439155 96.248 -64.979 53.144
439156 96.291 -64.981 53.104
439157 96.330 -64.989 53.066
439158 96.371 -64.997 53.028
439159 96.411 -65.006 52.988
439160 96.451 -65.015 52.947
439161 96.493 -65.022 52.909
439162 96.534 -65.029 52.870
439163 96.575 -65.037 52.831
439164 96.616 -65.044 52.791
439165 96.657 -65.052 52.749
439166 96.698 -65.059 52.706
439167 96.740 -65.066 52.662
439168 96.781 -65.074 52.618
439169 96.817 -65.078 52.575
439170 96.839 -65.073 52.533
439171 96.862 -65.069 52.490
439172 96.884 -65.064 52.448
439173 96.913 -65.063 52.407
439174 96.951 -65.067 52.364
439175 96.989 -65.071 52.322
439176 97.027 -65.070 52.279
439177 97.065 -65.069 52.234
439178 97.104 -65.069 52.188
439179 97.144 -65.071 52.142
439180 97.183 -65.072 52.094
439181 97.223 -65.072 52.048
439182 97.262 -65.073 52.002
439183 97.302 -65.074 51.957
439184 97.341 -65.076 51.912
439185 97.376 -65.079 51.866
439186 97.410 -65.083 51.822
439187 97.444 -65.087 51.779
439188 97.477 -65.091 51.736
439189 97.511 -65.096 51.695
439190 97.544 -65.100 51.652
439191 97.578 -65.104 51.611
439192 97.613 -65.110 51.569
439193 97.654 -65.115 51.528
439194 97.696 -65.120 51.486
439195 97.737 -65.125 51.446
439196 97.779 -65.129 51.407
439197 97.820 -65.135 51.368
439198 97.862 -65.139 51.329
439199 97.903 -65.144 51.292
439200 97.935 -65.146 51.253
439201 97.960 -65.151 51.213
439202 97.984 -65.154 51.174
439203 98.008 -65.158 51.135
439204 98.031 -65.162 51.100
439205 98.055 -65.166 51.065
439206 98.079 -65.170 51.032
439207 98.116 -65.173 51.000
439208 98.154 -65.176 50.965
439209 98.192 -65.179 50.933
439210 98.230 -65.182 50.901
439211 98.269 -65.184 50.870
439212 98.306 -65.187 50.838
439213 98.345 -65.190 50.803
439214 98.383 -65.193 50.763
439215 98.421 -65.196 50.720
439216 98.459 -65.200 50.676
439217 98.496 -65.204 50.634
439218 98.535 -65.209 50.589
439219 98.578 -65.215 50.541
439220 98.616 -65.220 50.489
439221 98.653 -65.226 50.435
439222 98.691 -65.231 50.380
439223 98.728 -65.236 50.324
439224 98.765 -65.242 50.268
439225 98.802 -65.247 50.211
439226 98.840 -65.252 50.153
439227 98.877 -65.257 50.097
439228 98.920 -65.262 50.046
439229 98.970 -65.266 49.995
439230 99.018 -65.270 49.947
439231 99.065 -65.274 49.901
439232 99.112 -65.278 49.858
439233 99.158 -65.282 49.821
439234 99.205 -65.286 49.789
439235 99.256 -65.288 49.756
439236 99.312 -65.293 49.720
439237 99.366 -65.296 49.679
439238 99.421 -65.299 49.636
439239 99.476 -65.302 49.590
439240 99.530 -65.305 49.543
439241 99.584 -65.308 49.496
439242 99.639 -65.311 49.446
439243 99.693 -65.314 49.396
439244 99.742 -65.321 49.345
439245 99.786 -65.332 49.295
439246 99.859 -65.317 49.246
439247 99.941 -65.292 49.199
439248 100.025 -65.267 49.151
439249 100.108 -65.243 49.104
439250 100.099 -65.308 49.058
439251 100.108 -65.311 49.011
439252 100.135 -65.284 48.966
439253 100.175 -65.272 48.921
439254 100.223 -65.270 48.878
439255 100.272 -65.268 48.834
439256 100.320 -65.266 48.791
439257 100.368 -65.264 48.748
439258 100.412 -65.264 48.705
439259 100.451 -65.270 48.662
439260 100.490 -65.275 48.622
439261 100.528 -65.280 48.583
439262 100.567 -65.286 48.542
439682 88.637 -63.907 55.318
439683 88.686 -63.917 55.318
439684 88.735 -63.927 55.319
439685 88.784 -63.938 55.320
439686 88.833 -63.948 55.320
439687 88.882 -63.958 55.321
439688 88.930 -63.969 55.322
439689 88.979 -63.979 55.323
439690 89.029 -63.990 55.324
439691 89.077 -64.000 55.327
439692 89.126 -64.010 55.330
439693 89.172 -64.023 55.335
439694 89.208 -64.039 55.343
439695 89.245 -64.056 55.353
439696 89.281 -64.073 55.360
439697 89.318 -64.089 55.364
439698 89.354 -64.106 55.363
439699 89.391 -64.123 55.358
439700 89.427 -64.139 55.351
439701 89.464 -64.155 55.343
439702 89.500 -64.172 55.334
439703 89.537 -64.189 55.319
439704 89.574 -64.205 55.298
439705 89.610 -64.222 55.274
439706 89.647 -64.239 55.250
439707 89.693 -64.242 55.228
439708 89.743 -64.241 55.210
439709 89.792 -64.240 55.192
439710 89.842 -64.239 55.177
439711 89.891 -64.237 55.164
439712 89.941 -64.236 55.153
439713 89.990 -64.235 55.147
439714 90.040 -64.233 55.144
439715 90.089 -64.232 55.143
439716 90.139 -64.231 55.144
439717 90.188 -64.229 55.149
439718 90.238 -64.228 55.160
439719 90.288 -64.226 55.174
439720 90.337 -64.226 55.188
439721 90.386 -64.224 55.199
439722 90.436 -64.223 55.206
439723 90.485 -64.222 55.210
439724 90.532 -64.229 55.212
439725 90.577 -64.246 55.211
439726 90.621 -64.263 55.205
439727 90.666 -64.275 55.190
439728 90.712 -64.283 55.169
439729 90.758 -64.290 55.140
439730 90.804 -64.298 55.105
439731 90.851 -64.305 55.067
439732 90.896 -64.313 55.025
439733 90.943 -64.320 54.986
439734 90.989 -64.328 54.952
439735 91.035 -64.336 54.923
439736 91.081 -64.343 54.897
439737 91.128 -64.351 54.875
439738 91.173 -64.360 54.855
439739 91.218 -64.370 54.836
439740 91.263 -64.380 54.819
439741 91.308 -64.389 54.803
439742 91.354 -64.399 54.789
439743 91.398 -64.409 54.775
439744 91.444 -64.419 54.762
439745 91.489 -64.430 54.750
439746 91.534 -64.439 54.738
439747 91.579 -64.449 54.727
439748 91.625 -64.459 54.717
439749 91.669 -64.469 54.708
439750 91.715 -64.479 54.699
439751 91.760 -64.489 54.691
439752 91.805 -64.499 54.684
439753 91.850 -64.509 54.677
439754 91.895 -64.519 54.670
439755 91.940 -64.529 54.664
439756 91.986 -64.539 54.658
439757 92.030 -64.549 54.653
439758 92.076 -64.559 54.647
439759 92.121 -64.568 54.640
439760 92.166 -64.579 54.635
439761 92.211 -64.589 54.630
439762 92.260 -64.599 54.626
439763 92.307 -64.614 54.622
439764 92.352 -64.643 54.618
439765 92.396 -64.656 54.615
439766 92.440 -64.659 54.613
439767 92.484 -64.662 54.611
439768 92.527 -64.665 54.610
439769 92.571 -64.668 54.609
439770 92.613 -64.672 54.606
439771 92.642 -64.685 54.601
439772 92.691 -64.699 54.596
439773 92.740 -64.712 54.590
439774 92.790 -64.725 54.584
439775 92.839 -64.737 54.581
439776 92.888 -64.750 54.579
439777 92.938 -64.763 54.579
439778 92.987 -64.776 54.580
439779 93.037 -64.790 54.580
439780 93.086 -64.803 54.581
439781 93.135 -64.815 54.583
439782 93.188 -64.826 54.586
439783 93.238 -64.822 54.590
439784 93.289 -64.815 54.593
439785 93.340 -64.808 54.598
439786 93.392 -64.801 54.600
439787 93.437 -64.804 54.600
439788 93.482 -64.807 54.599
439789 93.526 -64.810 54.596
439790 93.572 -64.813 54.592
439791 93.617 -64.816 54.587
439792 93.661 -64.819 54.582
439793 93.707 -64.822 54.577
439794 93.748 -64.822 54.572
439795 93.788 -64.821 54.566
439796 93.829 -64.820 54.559
439797 93.870 -64.819 54.551
439798 93.910 -64.818 54.543
439799 93.951 -64.816 54.536
439800 93.988 -64.824 54.530
439801 94.025 -64.833 54.521
439802 94.061 -64.843 54.509
439803 94.097 -64.852 54.494
439804 94.134 -64.861 54.474
439805 94.169 -64.866 54.453
439806 94.202 -64.867 54.429
439807 94.235 -64.865 54.404
439808 94.268 -64.863 54.379
439809 94.300 -64.862 54.354
439810 94.333 -64.860 54.332
439811 94.365 -64.858 54.313
439812 94.398 -64.856 54.295
439813 94.431 -64.855 54.276
439814 94.470 -64.855 54.255
439815 94.532 -64.863 54.233
439816 94.594 -64.872 54.211
439817 94.657 -64.880 54.190
439818 94.718 -64.888 54.169
439819 94.781 -64.897 54.147
439820 94.839 -64.905 54.125
439821 94.870 -64.907 54.103
439822 94.899 -64.910 54.083
439823 94.929 -64.911 54.064
439824 94.958 -64.913 54.047
439825 94.988 -64.915 54.031
439826 95.028 -64.922 54.014
439827 95.070 -64.931 53.997
439828 95.113 -64.941 53.977
439829 95.155 -64.950 53.956
439830 95.196 -64.957 53.933
439831 95.236 -64.963 53.909
439832 95.276 -64.969 53.884
439833 95.312 -64.978 53.857
439834 95.338 -64.999 53.825
439835 95.365 -65.019 53.791
439836 95.399 -65.032 53.757
439837 95.437 -65.041 53.724
439838 95.476 -65.049 53.695
439839 95.514 -65.058 53.670
439840 95.562 -65.041 53.649
439841 95.610 -65.022 53.631
439842 95.658 -65.004 53.616
439843 95.706 -64.992 53.599
439844 95.755 -64.984 53.579
439845 95.804 -64.976 53.555
439846 95.847 -64.969 53.527
439847 95.889 -64.962 53.494
439848 95.928 -64.957 53.458
439849 95.968 -64.953 53.423
439850 96.008 -64.948 53.387
439851 96.047 -64.944 53.350
439852 96.086 -64.940 53.310
439853 96.125 -64.935 53.268
439854 96.166 -64.932 53.224
439855 96.208 -64.933 53.181
439856 96.251 -64.935 53.140
439857 96.292 -64.941 53.100
439858 96.332 -64.949 53.062
439859 96.373 -64.957 53.023
439860 96.412 -64.966 52.981
439861 96.453 -64.975 52.940
439862 96.493 -64.982 52.900
439863 96.535 -64.990 52.861
439864 96.576 -64.997 52.822
439865 96.617 -65.004 52.783
439866 96.659 -65.012 52.744
439867 96.700 -65.019 52.701
439868 96.741 -65.027 52.657
439869 96.780 -65.033 52.611
439870 96.803 -65.029 52.564
439871 96.825 -65.024 52.520
439872 96.847 -65.019 52.477
439873 96.882 -65.021 52.434
439874 96.920 -65.025 52.393
439875 96.958 -65.029 52.353
439876 96.996 -65.033 52.312
439877 97.034 -65.034 52.271
439878 97.072 -65.033 52.227
439879 97.110 -65.032 52.182
439880 97.149 -65.031 52.135
439881 97.188 -65.032 52.087
439882 97.228 -65.032 52.039
439883 97.267 -65.034 51.993
439884 97.307 -65.035 51.948
439885 97.346 -65.035 51.901
439886 97.383 -65.038 51.856
439887 97.417 -65.042 51.812
439888 97.450 -65.046 51.769
439889 97.484 -65.051 51.726
439890 97.517 -65.055 51.685
439891 97.551 -65.059 51.642
439892 97.585 -65.063 51.600
439893 97.616 -65.068 51.559
439894 97.646 -65.072 51.518
439895 97.677 -65.076 51.475
439896 97.719 -65.081 51.433
439897 97.760 -65.085 51.391
439898 97.803 -65.090 51.350
439899 97.845 -65.095 51.310
439900 97.887 -65.100 51.273
439901 97.930 -65.104 51.234
439902 97.959 -65.108 51.196
439903 97.984 -65.112 51.157
439904 98.007 -65.116 51.118
439905 98.031 -65.120 51.082
439906 98.055 -65.123 51.048
439907 98.079 -65.127 51.014
439908 98.117 -65.130 50.980
439909 98.155 -65.133 50.943
439910 98.193 -65.136 50.909
439911 98.231 -65.139 50.876
439912 98.270 -65.142 50.843
439913 98.308 -65.145 50.810
439914 98.346 -65.148 50.775
439915 98.384 -65.151 50.734
439916 98.422 -65.154 50.690
439917 98.460 -65.157 50.647
439918 98.498 -65.160 50.606
439919 98.536 -65.163 50.562
439920 98.573 -65.168 50.514
439921 98.610 -65.173 50.462
439922 98.648 -65.179 50.408
439923 98.685 -65.184 50.354
439924 98.722 -65.189 50.299
439925 98.760 -65.195 50.243
439926 98.797 -65.200 50.186
439927 98.834 -65.205 50.129
439928 98.871 -65.210 50.074
439929 98.915 -65.215 50.025
439930 98.963 -65.219 49.977
439931 99.011 -65.223 49.929
439932 99.058 -65.227 49.883
439933 99.104 -65.231 49.842
439934 99.151 -65.235 49.805
439935 99.204 -65.236 49.771
439936 99.257 -65.238 49.738
439937 99.310 -65.239 49.703
439938 99.368 -65.246 49.661
439939 99.425 -65.253 49.619
439940 99.480 -65.256 49.573
439941 99.534 -65.259 49.526
439942 99.588 -65.262 49.478
439943 99.643 -65.265 49.429
439944 99.698 -65.268 49.379
439945 99.747 -65.275 49.329
439946 99.794 -65.282 49.280
439947 99.877 -65.257 49.231
439948 99.961 -65.233 49.185
439949 100.044 -65.208 49.138
439950 100.124 -65.187 49.092
439951 100.115 -65.252 49.047
439952 100.121 -65.264 49.001
439953 100.149 -65.237 48.956
439954 100.190 -65.226 48.912
439955 100.238 -65.224 48.869
439956 100.286 -65.222 48.826
439957 100.334 -65.220 48.784
439958 100.382 -65.218 48.741
439959 100.422 -65.222 48.697
439960 100.461 -65.227 48.656
439961 100.500 -65.232 48.615
439962 100.539 -65.238 48.575
439963 100.577 -65.242 48.535
440384 88.687 -63.878 55.320
440385 88.736 -63.888 55.321
440386 88.785 -63.898 55.321
440387 88.834 -63.909 55.321
440388 88.883 -63.919 55.322
440389 88.932 -63.929 55.323
440390 88.981 -63.940 55.326
440391 89.030 -63.951 55.328
440392 89.079 -63.960 55.331
440393 89.128 -63.971 55.336
440394 89.177 -63.982 55.343
440395 89.214 -63.998 55.353
440396 89.251 -64.014 55.365
440397 89.287 -64.031 55.373
440398 89.324 -64.048 55.378
440399 89.361 -64.064 55.377
440400 89.397 -64.081 55.372
440401 89.433 -64.098 55.367
440402 89.470 -64.114 55.362
440403 89.506 -64.131 55.356
440404 89.543 -64.148 55.343
440405 89.579 -64.164 55.321
440406 89.615 -64.181 55.296
440407 89.656 -64.192 55.269
440408 89.703 -64.189 55.244
440409 89.752 -64.187 55.225
440410 89.802 -64.186 55.207
440411 89.851 -64.185 55.192
440412 89.900 -64.183 55.178
440413 89.950 -64.182 55.166
440414 90.000 -64.181 55.161
440415 90.049 -64.179 55.158
440416 90.099 -64.178 55.158
440417 90.148 -64.177 55.160
440418 90.197 -64.176 55.165
440419 90.247 -64.174 55.176
440420 90.297 -64.173 55.191
440421 90.346 -64.172 55.207
440422 90.396 -64.170 55.222
440423 90.445 -64.169 55.232
440424 90.495 -64.168 55.236
440425 90.539 -64.185 55.237
440426 90.583 -64.201 55.234
440427 90.627 -64.218 55.227
440428 90.672 -64.231 55.210
440429 90.718 -64.239 55.184
440430 90.764 -64.246 55.151
440431 90.811 -64.254 55.111
440432 90.857 -64.261 55.069
440433 90.903 -64.269 55.026
440434 90.949 -64.276 54.986
440435 90.995 -64.284 54.952
440436 91.041 -64.292 54.922
440437 91.087 -64.299 54.894
440438 91.134 -64.307 54.871
440439 91.180 -64.314 54.850
440440 91.226 -64.322 54.831
440441 91.272 -64.330 54.814
440442 91.317 -64.341 54.798
440443 91.362 -64.351 54.783
440444 91.407 -64.361 54.770
440445 91.452 -64.371 54.757
440446 91.497 -64.380 54.745
440447 91.542 -64.390 54.734
440448 91.588 -64.400 54.723
440449 91.633 -64.410 54.713
440450 91.678 -64.421 54.704
440451 91.723 -64.430 54.696
440452 91.768 -64.440 54.687
440453 91.813 -64.450 54.680
440454 91.858 -64.460 54.673
440455 91.904 -64.470 54.667
440456 91.949 -64.480 54.661
440457 91.994 -64.490 54.655
440458 92.039 -64.500 54.649
440459 92.084 -64.510 54.644
440460 92.129 -64.520 54.638
440461 92.174 -64.530 54.633
440462 92.220 -64.540 54.628
440463 92.269 -64.549 54.624
440464 92.317 -64.561 54.621
440465 92.361 -64.591 54.618
440466 92.406 -64.601 54.615
440467 92.449 -64.604 54.613
440468 92.493 -64.607 54.612
440469 92.537 -64.610 54.611
440470 92.578 -64.615 54.610
440471 92.607 -64.627 54.606
440472 92.637 -64.640 54.600
440473 92.684 -64.653 54.595
440474 92.733 -64.666 54.590
440475 92.783 -64.679 54.587
440476 92.832 -64.693 54.584
440477 92.881 -64.706 54.583
440478 92.930 -64.718 54.583
440479 92.980 -64.731 54.583
440480 93.029 -64.744 54.583
440481 93.078 -64.757 54.584
440482 93.129 -64.769 54.587
440483 93.194 -64.771 54.590
440484 93.242 -64.774 54.594
440485 93.291 -64.775 54.598
440486 93.342 -64.769 54.604
440487 93.394 -64.762 54.607
440488 93.439 -64.763 54.608
440489 93.484 -64.767 54.605
440490 93.529 -64.770 54.599
440491 93.574 -64.773 54.594
440492 93.617 -64.774 54.588
440493 93.658 -64.773 54.582
440494 93.698 -64.772 54.574
440495 93.739 -64.771 54.568
440496 93.779 -64.769 54.560
440497 93.820 -64.768 54.552
440498 93.860 -64.767 54.543
440499 93.901 -64.765 54.534
440500 93.940 -64.768 54.525
440501 93.977 -64.777 54.516
440502 94.013 -64.786 54.506
440503 94.050 -64.795 54.494
440504 94.086 -64.804 54.480
440505 94.123 -64.813 54.461
440506 94.161 -64.822 54.439
440507 94.192 -64.823 54.416
440508 94.225 -64.821 54.391
440509 94.258 -64.819 54.366
440510 94.290 -64.818 54.341
440511 94.323 -64.816 54.319
440512 94.356 -64.814 54.300
440513 94.389 -64.812 54.283
440514 94.421 -64.811 54.264
440515 94.463 -64.812 54.244
440516 94.526 -64.821 54.223
440517 94.588 -64.829 54.201
440518 94.650 -64.837 54.180
440519 94.712 -64.846 54.158
440520 94.774 -64.854 54.137
440521 94.828 -64.864 54.115
440522 94.874 -64.869 54.093
440523 94.904 -64.872 54.071
440524 94.933 -64.873 54.052
440525 94.963 -64.875 54.034
440526 94.992 -64.877 54.018
440527 95.028 -64.882 54.002
440528 95.070 -64.891 53.985
440529 95.113 -64.900 53.964
440530 95.155 -64.910 53.943
440531 95.197 -64.919 53.920
440532 95.240 -64.928 53.897
440533 95.280 -64.934 53.872
440534 95.320 -64.941 53.845
440535 95.346 -64.961 53.815
440536 95.380 -64.973 53.782
440537 95.418 -64.982 53.747
440538 95.457 -64.991 53.713
440539 95.495 -65.000 53.683
440540 95.534 -65.009 53.656
440541 95.572 -65.017 53.634
440542 95.616 -65.011 53.615
440543 95.664 -64.993 53.598
440544 95.712 -64.975 53.581
440545 95.760 -64.957 53.562
440546 95.808 -64.938 53.539
440547 95.852 -64.921 53.513
440548 95.892 -64.916 53.482
440549 95.932 -64.912 53.448
440550 95.971 -64.907 53.414
440551 96.010 -64.903 53.380
440552 96.050 -64.898 53.345
440553 96.089 -64.894 53.307
440554 96.128 -64.889 53.265
440555 96.169 -64.887 53.222
440556 96.212 -64.888 53.179
440557 96.253 -64.891 53.137
440558 96.294 -64.900 53.097
440559 96.334 -64.909 53.059
440560 96.374 -64.917 53.019
440561 96.414 -64.926 52.977
440562 96.455 -64.935 52.936
440563 96.495 -64.943 52.893
440564 96.536 -64.950 52.852
440565 96.578 -64.957 52.812
440566 96.619 -64.965 52.774
440567 96.660 -64.972 52.736
440568 96.701 -64.980 52.695
440569 96.742 -64.987 52.651
440570 96.765 -64.983 52.604
440571 96.788 -64.978 52.555
440572 96.813 -64.975 52.509
440573 96.851 -64.979 52.465
440574 96.889 -64.983 52.421
440575 96.927 -64.987 52.380
440576 96.965 -64.991 52.341
440577 97.003 -64.995 52.302
440578 97.041 -64.998 52.261
440579 97.079 -64.997 52.219
440580 97.117 -64.995 52.175
440581 97.156 -64.994 52.128
440582 97.194 -64.993 52.079
440583 97.233 -64.993 52.031
440584 97.272 -64.994 51.985
440585 97.311 -64.994 51.939
440586 97.351 -64.995 51.891
440587 97.389 -64.997 51.846
440588 97.423 -65.001 51.801
440589 97.457 -65.006 51.757
440590 97.490 -65.010 51.715
440591 97.524 -65.014 51.673
440592 97.557 -65.019 51.631
440593 97.588 -65.022 51.589
440594 97.619 -65.027 51.547
440595 97.650 -65.032 51.505
440596 97.678 -65.035 51.462
440597 97.706 -65.039 51.419
440598 97.733 -65.043 51.376
440599 97.773 -65.047 51.334
440600 97.815 -65.052 51.294
440601 97.857 -65.057 51.256
440602 97.899 -65.062 51.218
440603 97.946 -65.066 51.180
440604 97.983 -65.069 51.142
440605 98.007 -65.073 51.102
440606 98.030 -65.078 51.066
440607 98.054 -65.082 51.031
440608 98.080 -65.085 50.996
440609 98.118 -65.088 50.960
440610 98.156 -65.091 50.921
440611 98.194 -65.094 50.885
440612 98.232 -65.097 50.850
440613 98.271 -65.100 50.816
440614 98.309 -65.103 50.782
440615 98.347 -65.105 50.746
440616 98.385 -65.108 50.704
440617 98.423 -65.111 50.660
440618 98.461 -65.115 50.616
440619 98.499 -65.118 50.573
440620 98.535 -65.121 50.531
440621 98.570 -65.123 50.483
440622 98.605 -65.126 50.432
440623 98.642 -65.132 50.380
440624 98.679 -65.137 50.326
440625 98.717 -65.142 50.272
440626 98.754 -65.148 50.216
440627 98.791 -65.153 50.159
440628 98.829 -65.158 50.103
440629 98.866 -65.163 50.050
440630 98.909 -65.168 50.004
440631 98.958 -65.172 49.958
440632 99.005 -65.176 49.911
440633 99.051 -65.180 49.866
440634 99.098 -65.184 49.824
440635 99.151 -65.185 49.787
440636 99.205 -65.186 49.752
440637 99.258 -65.187 49.719
440638 99.311 -65.188 49.683
440639 99.366 -65.192 49.643
440640 99.424 -65.200 49.600
440641 99.482 -65.207 49.555
440642 99.539 -65.213 49.507
440643 99.593 -65.216 49.459
440644 99.647 -65.219 49.411
440645 99.702 -65.222 49.362
440646 99.752 -65.228 49.313
440647 99.814 -65.223 49.265
440648 99.897 -65.198 49.217
440649 99.980 -65.174 49.171
440650 100.062 -65.149 49.126
440651 100.139 -65.131 49.080
440652 100.131 -65.195 49.035
440653 100.135 -65.217 48.990
440654 100.162 -65.189 48.946
440655 100.204 -65.180 48.902
440656 100.252 -65.178 48.860
440657 100.300 -65.176 48.818
440658 100.349 -65.174 48.776
440659 100.394 -65.174 48.733
440660 100.432 -65.179 48.690
440661 100.471 -65.185 48.649
440662 100.510 -65.190 48.608
440663 100.549 -65.195 48.568
440664 100.587 -65.200 48.529
441085 88.689 -63.838 55.321
441086 88.738 -63.848 55.321
441087 88.787 -63.859 55.322
441088 88.836 -63.869 55.322
441089 88.885 -63.879 55.323
441090 88.933 -63.890 55.324
441091 88.983 -63.900 55.327
441092 89.032 -63.910 55.331
441093 89.081 -63.921 55.335
441094 89.129 -63.932 55.340
441095 89.179 -63.942 55.350
441096 89.220 -63.956 55.363
441097 89.257 -63.973 55.375
441098 89.293 -63.989 55.385
441099 89.330 -64.006 55.390
441100 89.366 -64.023 55.391
441101 89.402 -64.039 55.387
441102 89.439 -64.056 55.385
441103 89.475 -64.073 55.383
441104 89.512 -64.089 55.378
441105 89.548 -64.106 55.366
441106 89.585 -64.123 55.344
441107 89.624 -64.136 55.317
441108 89.669 -64.140 55.288
441109 89.716 -64.137 55.261
441110 89.762 -64.134 55.240
441111 89.811 -64.132 55.221
441112 89.860 -64.131 55.205
441113 89.910 -64.129 55.191
441114 89.959 -64.129 55.180
441115 90.009 -64.127 55.174
441116 90.058 -64.126 55.171
441117 90.108 -64.124 55.171
441118 90.157 -64.123 55.174
441119 90.207 -64.122 55.181
441120 90.256 -64.120 55.191
441121 90.306 -64.120 55.206
441122 90.356 -64.118 55.223
441123 90.405 -64.117 55.240
441124 90.454 -64.115 55.253
441125 90.501 -64.124 55.260
441126 90.545 -64.141 55.260
441127 90.589 -64.157 55.255
441128 90.634 -64.174 55.244
441129 90.678 -64.187 55.224
441130 90.725 -64.195 55.194
441131 90.770 -64.202 55.156
441132 90.817 -64.210 55.113
441133 90.863 -64.217 55.069
441134 90.909 -64.225 55.026
441135 90.955 -64.233 54.986
441136 91.002 -64.240 54.951
441137 91.047 -64.248 54.920
441138 91.094 -64.255 54.891
441139 91.140 -64.263 54.867
441140 91.186 -64.270 54.846
441141 91.232 -64.278 54.826
441142 91.278 -64.286 54.808
441143 91.324 -64.293 54.793
441144 91.370 -64.302 54.778
441145 91.416 -64.311 54.765
441146 91.460 -64.321 54.753
441147 91.506 -64.332 54.742
441148 91.551 -64.342 54.731
441149 91.596 -64.352 54.720
441150 91.641 -64.361 54.710
441151 91.686 -64.371 54.701
441152 91.731 -64.381 54.693
441153 91.777 -64.391 54.685
441154 91.821 -64.402 54.678
441155 91.867 -64.411 54.671
441156 91.912 -64.421 54.665
441157 91.957 -64.431 54.659
441158 92.002 -64.441 54.653
441159 92.048 -64.451 54.647
441160 92.092 -64.461 54.643
441161 92.138 -64.471 54.638
441162 92.183 -64.481 54.631
441163 92.229 -64.491 54.627
441164 92.278 -64.501 54.624
441165 92.326 -64.511 54.621
441166 92.371 -64.538 54.620
441167 92.415 -64.546 54.616
441168 92.459 -64.549 54.614
441169 92.503 -64.552 54.613
441170 92.544 -64.557 54.611
441171 92.573 -64.570 54.610
441172 92.602 -64.583 54.605
441173 92.631 -64.596 54.600
441174 92.676 -64.609 54.595
441175 92.726 -64.621 54.592
441176 92.775 -64.634 54.589
441177 92.825 -64.647 54.586
441178 92.874 -64.660 54.585
441179 92.923 -64.673 54.585
441180 92.973 -64.687 54.584
441181 93.022 -64.700 54.585
441182 93.071 -64.712 54.587
441183 93.145 -64.713 54.590
441184 93.200 -64.715 54.594
441185 93.249 -64.719 54.599
441186 93.297 -64.722 54.604
441187 93.345 -64.726 54.610
441188 93.396 -64.722 54.614
441189 93.441 -64.723 54.616
441190 93.486 -64.726 54.614
441191 93.527 -64.725 54.607
441192 93.567 -64.724 54.599
441193 93.608 -64.722 54.592
441194 93.648 -64.721 54.584
441195 93.689 -64.720 54.575
441196 93.729 -64.718 54.568
441197 93.770 -64.718 54.560
441198 93.811 -64.716 54.551
441199 93.851 -64.715 54.541
441200 93.892 -64.714 54.529
441201 93.929 -64.721 54.517
441202 93.966 -64.730 54.506
441203 94.002 -64.739 54.493
441204 94.039 -64.748 54.480
441205 94.075 -64.757 54.465
441206 94.115 -64.766 54.447
441207 94.153 -64.775 54.426
441208 94.182 -64.778 54.404
441209 94.215 -64.777 54.381
441210 94.248 -64.775 54.354
441211 94.280 -64.774 54.329
441212 94.313 -64.772 54.306
441213 94.346 -64.770 54.286
441214 94.379 -64.769 54.269
441215 94.411 -64.767 54.252
441216 94.457 -64.769 54.232
441217 94.519 -64.778 54.212
441218 94.581 -64.786 54.191
441219 94.643 -64.794 54.170
441220 94.705 -64.803 54.148
441221 94.764 -64.812 54.127
441222 94.815 -64.822 54.104
441223 94.862 -64.828 54.082
441224 94.908 -64.834 54.060
441225 94.938 -64.835 54.039
441226 94.968 -64.837 54.020
441227 94.997 -64.839 54.003
441228 95.028 -64.842 53.987
441229 95.070 -64.851 53.969
441230 95.113 -64.860 53.949
441231 95.155 -64.869 53.928
441232 95.198 -64.878 53.906
441233 95.240 -64.887 53.882
441234 95.282 -64.896 53.859
441235 95.325 -64.905 53.833
441236 95.361 -64.916 53.803
441237 95.400 -64.924 53.770
441238 95.438 -64.933 53.734
441239 95.477 -64.941 53.701
441240 95.515 -64.950 53.670
441241 95.554 -64.959 53.642
441242 95.592 -64.968 53.618
441243 95.631 -64.976 53.598
441244 95.671 -64.982 53.580
441245 95.719 -64.963 53.563
441246 95.766 -64.946 53.544
441247 95.814 -64.928 53.522
441248 95.858 -64.908 53.495
441249 95.896 -64.887 53.464
441250 95.935 -64.866 53.432
441251 95.974 -64.862 53.399
441252 96.013 -64.857 53.366
441253 96.052 -64.853 53.333
441254 96.092 -64.848 53.297
441255 96.131 -64.844 53.256
441256 96.172 -64.842 53.213
441257 96.215 -64.843 53.171
441258 96.255 -64.852 53.130
441259 96.295 -64.860 53.090
441260 96.335 -64.869 53.051
441261 96.376 -64.877 53.012
441262 96.416 -64.886 52.971
441263 96.456 -64.894 52.929
441264 96.497 -64.903 52.887
441265 96.537 -64.910 52.846
441266 96.579 -64.918 52.805
441267 96.620 -64.925 52.764
441268 96.661 -64.933 52.724
441269 96.702 -64.940 52.683
441270 96.729 -64.938 52.641
441271 96.751 -64.934 52.596
441272 96.782 -64.934 52.548
441273 96.820 -64.938 52.503
441274 96.858 -64.941 52.457
441275 96.896 -64.945 52.412
441276 96.934 -64.949 52.369
441277 96.972 -64.953 52.329
441278 97.010 -64.957 52.290
441279 97.048 -64.961 52.252
441280 97.086 -64.960 52.211
441281 97.124 -64.960 52.167
441282 97.163 -64.958 52.121
441283 97.201 -64.957 52.073
441284 97.239 -64.956 52.024
441285 97.278 -64.954 51.977
441286 97.316 -64.954 51.929
441287 97.356 -64.955 51.882
441288 97.395 -64.957 51.835
441289 97.430 -64.960 51.789
441290 97.463 -64.965 51.745
441291 97.497 -64.969 51.702
441292 97.529 -64.973 51.660
441293 97.560 -64.978 51.619
441294 97.592 -64.982 51.576
441295 97.623 -64.986 51.532
441296 97.654 -64.991 51.490
441297 97.682 -64.994 51.447
441298 97.710 -64.998 51.405
441299 97.737 -65.002 51.361
441300 97.765 -65.007 51.319
441301 97.793 -65.010 51.279
441302 97.828 -65.014 51.243
441303 97.870 -65.019 51.205
441304 97.914 -65.024 51.167
441305 97.962 -65.028 51.128
441306 98.006 -65.032 51.088
441307 98.030 -65.035 51.052
441308 98.054 -65.039 51.016
441309 98.081 -65.043 50.979
441310 98.119 -65.046 50.940
441311 98.157 -65.048 50.900
441312 98.196 -65.051 50.861
441313 98.233 -65.054 50.824
441314 98.272 -65.057 50.789
441315 98.310 -65.060 50.753
441316 98.348 -65.063 50.717
441317 98.386 -65.066 50.675
441318 98.424 -65.070 50.630
441319 98.461 -65.073 50.583
441320 98.499 -65.076 50.537
441321 98.534 -65.079 50.495
441322 98.569 -65.082 50.449
441323 98.604 -65.085 50.400
441324 98.639 -65.087 50.349
441325 98.674 -65.090 50.296
441326 98.711 -65.095 50.242
441327 98.748 -65.101 50.187
441328 98.786 -65.106 50.131
441329 98.823 -65.111 50.077
441330 98.860 -65.116 50.026
441331 98.903 -65.121 49.984
441332 98.951 -65.125 49.938
441333 98.998 -65.129 49.892
441334 99.046 -65.132 49.848
441335 99.099 -65.133 49.806
441336 99.152 -65.135 49.768
441337 99.206 -65.135 49.733
441338 99.259 -65.137 49.698
441339 99.312 -65.138 49.663
441340 99.365 -65.139 49.623
441341 99.422 -65.145 49.581
441342 99.480 -65.153 49.535
441343 99.538 -65.160 49.488
441344 99.596 -65.168 49.440
441345 99.651 -65.173 49.392
441346 99.706 -65.176 49.345
441347 99.757 -65.182 49.297
441348 99.832 -65.163 49.249
441349 99.915 -65.139 49.203
441350 99.998 -65.115 49.158
441351 100.081 -65.090 49.114
441352 100.155 -65.075 49.069
441353 100.146 -65.139 49.023
441354 100.148 -65.169 48.979
441355 100.176 -65.142 48.935
441356 100.218 -65.134 48.893
441357 100.267 -65.132 48.851
441358 100.315 -65.130 48.809
441359 100.363 -65.128 48.768
441360 100.404 -65.132 48.724
441361 100.442 -65.137 48.682
441362 100.481 -65.141 48.642
441363 100.520 -65.147 48.602
441364 100.559 -65.152 48.562
441787 88.740 -63.809 55.322
441788 88.788 -63.819 55.322
441789 88.837 -63.830 55.323
441790 88.886 -63.840 55.324
441791 88.935 -63.850 55.326
441792 88.984 -63.861 55.329
441793 89.033 -63.871 55.333
441794 89.082 -63.882 55.338
441795 89.131 -63.892 55.345
441796 89.180 -63.903 55.355
441797 89.226 -63.914 55.369
441798 89.262 -63.931 55.383
441799 89.299 -63.947 55.395
441800 89.335 -63.964 55.403
441801 89.372 -63.981 55.405
441802 89.408 -63.998 55.402
441803 89.445 -64.014 55.401
441804 89.481 -64.031 55.399
441805 89.518 -64.048 55.395
441806 89.554 -64.064 55.384
441807 89.592 -64.080 55.363
441808 89.636 -64.087 55.337
441809 89.682 -64.088 55.308
441810 89.728 -64.085 55.279
441811 89.775 -64.082 55.255
441812 89.821 -64.079 55.235
441813 89.870 -64.077 55.218
441814 89.919 -64.076 55.204
441815 89.968 -64.075 55.193
441816 90.018 -64.073 55.187
441817 90.068 -64.072 55.184
441818 90.117 -64.071 55.184
441819 90.167 -64.070 55.187
441820 90.216 -64.068 55.196
441821 90.265 -64.067 55.205
441822 90.315 -64.066 55.219
441823 90.365 -64.064 55.236
441824 90.414 -64.063 55.254
441825 90.463 -64.063 55.270
441826 90.507 -64.079 55.279
441827 90.551 -64.096 55.281
441828 90.596 -64.113 55.274
441829 90.640 -64.129 55.259
441830 90.685 -64.143 55.232
441831 90.731 -64.151 55.196
441832 90.777 -64.158 55.154
441833 90.823 -64.166 55.109
441834 90.869 -64.173 55.065
441835 90.916 -64.181 55.023
441836 90.962 -64.189 54.985
441837 91.008 -64.196 54.950
441838 91.054 -64.204 54.917
441839 91.100 -64.211 54.888
441840 91.146 -64.219 54.863
441841 91.192 -64.226 54.841
441842 91.238 -64.234 54.820
441843 91.285 -64.242 54.801
441844 91.331 -64.249 54.787
441845 91.377 -64.257 54.774
441846 91.423 -64.264 54.761
441847 91.469 -64.273 54.749
441848 91.514 -64.283 54.738
441849 91.559 -64.292 54.728
441850 91.604 -64.302 54.718
441851 91.650 -64.313 54.708
441852 91.695 -64.323 54.699
441853 91.740 -64.333 54.691
441854 91.785 -64.342 54.683
441855 91.830 -64.352 54.676
441856 91.875 -64.362 54.669
441857 91.920 -64.372 54.663
441858 91.965 -64.382 54.657
441859 92.011 -64.392 54.651
441860 92.056 -64.402 54.646
441861 92.101 -64.412 54.642
441862 92.146 -64.422 54.637
441863 92.191 -64.432 54.630
441864 92.238 -64.442 54.627
441865 92.287 -64.452 54.624
441866 92.335 -64.461 54.621
441867 92.381 -64.485 54.620
441868 92.425 -64.491 54.618
441869 92.469 -64.494 54.615
441870 92.509 -64.499 54.613
441871 92.538 -64.512 54.611
441872 92.567 -64.524 54.609
441873 92.596 -64.537 54.605
441874 92.626 -64.550 54.600
441875 92.669 -64.563 54.596
441876 92.719 -64.576 54.593
441877 92.768 -64.590 54.590
441878 92.817 -64.602 54.588
441879 92.867 -64.615 54.586
441880 92.916 -64.628 54.586
441881 92.966 -64.641 54.585
441882 93.015 -64.654 54.587
441883 93.084 -64.657 54.589
441884 93.158 -64.657 54.592
441885 93.206 -64.660 54.597
441886 93.255 -64.664 54.604
441887 93.303 -64.667 54.610
441888 93.351 -64.671 54.615
441889 93.400 -64.674 54.621
441890 93.436 -64.675 54.623
441891 93.477 -64.674 54.621
441892 93.517 -64.673 54.614
441893 93.558 -64.671 54.604
441894 93.599 -64.671 54.595
441895 93.639 -64.669 54.586
441896 93.680 -64.668 54.577
441897 93.720 -64.667 54.570
441898 93.761 -64.665 54.562
441899 93.801 -64.665 54.551
441900 93.842 -64.663 54.539
441901 93.882 -64.664 54.526
441902 93.918 -64.673 54.513
441903 93.955 -64.682 54.500
441904 93.991 -64.691 54.486
441905 94.028 -64.700 54.472
441906 94.066 -64.710 54.455
441907 94.112 -64.720 54.437
441908 94.145 -64.726 54.416
441909 94.174 -64.732 54.394
441910 94.205 -64.733 54.371
441911 94.238 -64.731 54.345
441912 94.270 -64.730 54.318
441913 94.303 -64.728 54.293
441914 94.336 -64.726 54.270
441915 94.368 -64.725 54.251
441916 94.401 -64.723 54.234
441917 94.450 -64.727 54.217
441918 94.512 -64.735 54.198
441919 94.574 -64.743 54.178
441920 94.637 -64.752 54.157
441921 94.699 -64.760 54.135
441922 94.751 -64.770 54.114
441923 94.802 -64.780 54.092
441924 94.850 -64.787 54.070
441925 94.896 -64.792 54.049
441926 94.942 -64.797 54.026
441927 94.973 -64.800 54.005
441928 95.002 -64.802 53.986
441929 95.032 -64.803 53.968
441930 95.070 -64.811 53.950
441931 95.113 -64.819 53.932
441932 95.156 -64.828 53.911
441933 95.198 -64.837 53.889
441934 95.240 -64.847 53.866
441935 95.282 -64.856 53.845
441936 95.328 -64.864 53.819
441937 95.375 -64.872 53.789
441938 95.415 -64.879 53.756
441939 95.455 -64.886 53.721
441940 95.495 -64.894 53.688
441941 95.535 -64.900 53.656
441942 95.573 -64.910 53.627
441943 95.612 -64.918 53.601
441944 95.651 -64.927 53.579
441945 95.689 -64.935 53.562
441946 95.728 -64.944 53.545
441947 95.773 -64.935 53.525
441948 95.821 -64.916 53.503
441949 95.863 -64.896 53.477
441950 95.901 -64.875 53.444
441951 95.939 -64.853 53.412
441952 95.978 -64.832 53.381
441953 96.016 -64.812 53.350
441954 96.055 -64.807 53.320
441955 96.095 -64.803 53.285
441956 96.134 -64.798 53.246
441957 96.175 -64.797 53.204
441958 96.217 -64.803 53.162
441959 96.257 -64.812 53.121
441960 96.297 -64.820 53.081
441961 96.337 -64.828 53.041
441962 96.377 -64.837 53.001
441963 96.417 -64.846 52.961
441964 96.458 -64.854 52.920
441965 96.498 -64.862 52.880
441966 96.539 -64.871 52.840
441967 96.580 -64.878 52.799
441968 96.621 -64.885 52.758
441969 96.662 -64.893 52.716
441970 96.692 -64.893 52.674
441971 96.714 -64.888 52.631
441972 96.751 -64.891 52.587
441973 96.789 -64.896 52.543
441974 96.827 -64.900 52.499
441975 96.865 -64.903 52.453
441976 96.903 -64.907 52.408
441977 96.941 -64.911 52.363
441978 96.979 -64.915 52.319
441979 97.017 -64.919 52.279
441980 97.055 -64.923 52.241
441981 97.093 -64.924 52.201
441982 97.131 -64.923 52.158
441983 97.169 -64.922 52.113
441984 97.208 -64.920 52.065
441985 97.246 -64.919 52.018
441986 97.284 -64.918 51.970
441987 97.322 -64.916 51.922
441988 97.361 -64.916 51.873
441989 97.400 -64.916 51.825
441990 97.436 -64.919 51.778
441991 97.469 -64.924 51.733
441992 97.501 -64.928 51.689
441993 97.532 -64.932 51.646
441994 97.564 -64.937 51.603
441995 97.595 -64.941 51.560
441996 97.626 -64.945 51.518
441997 97.657 -64.950 51.474
441998 97.686 -64.953 51.431
441999 97.714 -64.957 51.389
442000 97.742 -64.962 51.346
442001 97.770 -64.966 51.304
442002 97.797 -64.969 51.264
442003 97.824 -64.973 51.228
442004 97.852 -64.977 51.190
442005 97.882 -64.982 51.152
442006 97.930 -64.985 51.113
442007 97.978 -64.989 51.074
442008 98.025 -64.993 51.037
442009 98.053 -64.997 51.000
442010 98.082 -65.000 50.962
442011 98.120 -65.004 50.922
442012 98.158 -65.007 50.882
442013 98.197 -65.010 50.842
442014 98.234 -65.013 50.802
442015 98.273 -65.016 50.764
442016 98.310 -65.019 50.727
442017 98.348 -65.022 50.688
442018 98.386 -65.025 50.648
442019 98.424 -65.029 50.603
442020 98.462 -65.032 50.554
442021 98.498 -65.035 50.505
442022 98.533 -65.037 50.458
442023 98.568 -65.040 50.411
442024 98.603 -65.043 50.363
442025 98.638 -65.045 50.311
442026 98.673 -65.048 50.259
442027 98.708 -65.051 50.205
442028 98.743 -65.054 50.152
442029 98.780 -65.059 50.100
442030 98.817 -65.064 50.050
442031 98.855 -65.069 50.001
442032 98.898 -65.074 49.959
442033 98.945 -65.078 49.915
442034 98.994 -65.081 49.872
442035 99.047 -65.082 49.830
442036 99.100 -65.083 49.789
442037 99.153 -65.085 49.752
442038 99.207 -65.085 49.717
442039 99.260 -65.086 49.682
442040 99.313 -65.088 49.645
442041 99.366 -65.088 49.604
442042 99.420 -65.091 49.562
442043 99.478 -65.099 49.517
442044 99.536 -65.107 49.469
442045 99.594 -65.114 49.422
442046 99.652 -65.122 49.375
442047 99.710 -65.129 49.329
442048 99.768 -65.129 49.282
442049 99.851 -65.104 49.235
442050 99.934 -65.080 49.188
442051 100.017 -65.056 49.144
442052 100.101 -65.031 49.101
442053 100.171 -65.018 49.057
442054 100.162 -65.083 49.013
442055 100.161 -65.122 48.969
442056 100.189 -65.095 48.925
442057 100.233 -65.088 48.883
442058 100.281 -65.086 48.843
442059 100.329 -65.084 48.802
442060 100.375 -65.084 48.759
442061 100.414 -65.088 48.717
442062 100.453 -65.094 48.676
442063 100.491 -65.099 48.635
442064 100.530 -65.104 48.596
442065 100.568 -65.110 48.556
442457 87.353 -63.545 55.740
442458 87.387 -63.554 55.721
442459 87.421 -63.564 55.703
442460 87.456 -63.573 55.683
442461 87.490 -63.583 55.664
442462 87.524 -63.593 55.645
442463 87.558 -63.602 55.627
442464 87.592 -63.612 55.609
442465 87.627 -63.622 55.589
442466 87.661 -63.631 55.570
442488 88.741 -63.769 55.324
442489 88.790 -63.780 55.324
442490 88.839 -63.791 55.325
442491 88.888 -63.800 55.326
442492 88.937 -63.811 55.328
442493 88.986 -63.822 55.331
442494 89.035 -63.831 55.336
442495 89.084 -63.842 55.342
442496 89.132 -63.853 55.349
442497 89.182 -63.863 55.359
442498 89.231 -63.873 55.374
442499 89.268 -63.889 55.389
442500 89.305 -63.906 55.404
442501 89.341 -63.922 55.415
442502 89.378 -63.939 55.417
442503 89.414 -63.956 55.415
442504 89.451 -63.973 55.413
442505 89.487 -63.989 55.411
442506 89.524 -64.006 55.408
442507 89.560 -64.023 55.396
442508 89.604 -64.031 55.378
442509 89.649 -64.038 55.354
442510 89.694 -64.036 55.325
442511 89.741 -64.034 55.294
442512 89.788 -64.031 55.269
442513 89.834 -64.028 55.248
442514 89.881 -64.025 55.231
442515 89.928 -64.022 55.218
442516 89.978 -64.021 55.206
442517 90.027 -64.020 55.200
442518 90.077 -64.018 55.196
442519 90.126 -64.017 55.197
442520 90.176 -64.016 55.200
442521 90.225 -64.014 55.209
442522 90.275 -64.013 55.219
442523 90.324 -64.012 55.231
442524 90.374 -64.010 55.246
442525 90.424 -64.009 55.265
442526 90.469 -64.019 55.285
442527 90.514 -64.035 55.296
442528 90.558 -64.052 55.298
442529 90.602 -64.069 55.291
442530 90.646 -64.085 55.271
442531 90.691 -64.099 55.239
442532 90.737 -64.107 55.200
442533 90.783 -64.114 55.155
442534 90.829 -64.122 55.107
442535 90.876 -64.129 55.062
442536 90.922 -64.137 55.020
442537 90.968 -64.145 54.982
442538 91.014 -64.152 54.946
442539 91.060 -64.160 54.912
442540 91.106 -64.167 54.883
442541 91.153 -64.175 54.857
442542 91.199 -64.182 54.835
442543 91.245 -64.190 54.815
442544 91.291 -64.198 54.797
442545 91.337 -64.205 54.782
442546 91.383 -64.213 54.769
442547 91.429 -64.220 54.756
442548 91.475 -64.228 54.745
442549 91.522 -64.236 54.734
442550 91.568 -64.244 54.724
442551 91.612 -64.254 54.715
442552 91.658 -64.264 54.706
442553 91.703 -64.273 54.697
442554 91.748 -64.283 54.689
442555 91.793 -64.293 54.681
442556 91.839 -64.304 54.674
442557 91.883 -64.314 54.667
442558 91.929 -64.323 54.661
442559 91.974 -64.333 54.654
442560 92.019 -64.343 54.650
442561 92.064 -64.353 54.646
442562 92.109 -64.363 54.640
442563 92.154 -64.374 54.635
442564 92.200 -64.383 54.630
442565 92.247 -64.393 54.627
442566 92.295 -64.403 54.624
442567 92.344 -64.413 54.622
442568 92.390 -64.433 54.621
442569 92.434 -64.436 54.618
442570 92.475 -64.441 54.615
442571 92.504 -64.454 54.612
442572 92.532 -64.467 54.610
442573 92.562 -64.480 54.607
442574 92.591 -64.493 54.604
442575 92.620 -64.505 54.600
442576 92.662 -64.518 54.595
442577 92.712 -64.531 54.592
442578 92.761 -64.544 54.588
442579 92.810 -64.557 54.586
442580 92.860 -64.571 54.585
442581 92.909 -64.584 54.584
442582 92.958 -64.596 54.585
442583 93.023 -64.602 54.587
442584 93.100 -64.601 54.590
442585 93.164 -64.602 54.594
442586 93.213 -64.606 54.601
442587 93.261 -64.609 54.608
442588 93.309 -64.612 54.614
442589 93.358 -64.615 54.620
442590 93.401 -64.617 54.624
442591 93.427 -64.624 54.627
442592 93.468 -64.622 54.626
442593 93.508 -64.621 54.618
442594 93.549 -64.620 54.608
442595 93.589 -64.618 54.597
442596 93.630 -64.618 54.588
442597 93.671 -64.616 54.580
442598 93.711 -64.615 54.572
442599 93.752 -64.614 54.563
442600 93.792 -64.612 54.552
442601 93.833 -64.612 54.538
442602 93.871 -64.617 54.523
442603 93.907 -64.626 54.510
442604 93.943 -64.635 54.496
442605 93.980 -64.644 54.481
442606 94.017 -64.653 54.465
442607 94.063 -64.663 54.447
442608 94.108 -64.673 54.427
442609 94.137 -64.678 54.406
442610 94.166 -64.684 54.383
442611 94.195 -64.689 54.360
442612 94.227 -64.687 54.334
442613 94.260 -64.686 54.306
442614 94.293 -64.684 54.279
442615 94.326 -64.682 54.253
442616 94.358 -64.681 54.233
442617 94.391 -64.679 54.215
442618 94.443 -64.684 54.198
442619 94.506 -64.692 54.181
442620 94.568 -64.700 54.163
442621 94.630 -64.709 54.143
442622 94.687 -64.718 54.122
442623 94.739 -64.728 54.101
442624 94.789 -64.737 54.079
442625 94.838 -64.745 54.058
442626 94.884 -64.751 54.036
442627 94.930 -64.756 54.013
442628 94.976 -64.762 53.989
442629 95.007 -64.764 53.968
442630 95.036 -64.765 53.948
442631 95.070 -64.770 53.929
442632 95.113 -64.779 53.911
442633 95.156 -64.788 53.891
442634 95.198 -64.797 53.869
442635 95.240 -64.806 53.846
442636 95.285 -64.815 53.825
442637 95.334 -64.823 53.801
442638 95.383 -64.831 53.774
442639 95.425 -64.838 53.742
442640 95.466 -64.845 53.706
442641 95.506 -64.853 53.672
442642 95.546 -64.859 53.641
442643 95.586 -64.867 53.611
442644 95.626 -64.874 53.584
442645 95.667 -64.881 53.559
442646 95.706 -64.888 53.543
442647 95.747 -64.895 53.526
442648 95.786 -64.903 53.506
442649 95.827 -64.905 53.483
442650 95.868 -64.884 53.456
442651 95.906 -64.863 53.422
442652 95.945 -64.842 53.391
442653 95.983 -64.821 53.361
442654 96.021 -64.799 53.332
442655 96.059 -64.778 53.305
442656 96.098 -64.757 53.271
442657 96.137 -64.753 53.235
442658 96.178 -64.754 53.196
442659 96.218 -64.762 53.153
442660 96.258 -64.772 53.111
442661 96.299 -64.780 53.070
442662 96.339 -64.788 53.029
442663 96.379 -64.797 52.988
442664 96.419 -64.806 52.948
442665 96.459 -64.814 52.910
442666 96.500 -64.822 52.872
442667 96.540 -64.831 52.833
442668 96.581 -64.838 52.794
442669 96.622 -64.846 52.755
442670 96.655 -64.848 52.710
442671 96.682 -64.846 52.666
442672 96.720 -64.850 52.623
442673 96.758 -64.853 52.581
442674 96.796 -64.857 52.539
442675 96.834 -64.862 52.496
442676 96.872 -64.866 52.451
442677 96.910 -64.869 52.404
442678 96.948 -64.873 52.357
442679 96.986 -64.877 52.311
442680 97.024 -64.881 52.269
442681 97.062 -64.884 52.230
442682 97.100 -64.888 52.191
442683 97.138 -64.887 52.149
442684 97.176 -64.885 52.104
442685 97.215 -64.884 52.058
442686 97.253 -64.883 52.011
442687 97.291 -64.881 51.963
442688 97.329 -64.881 51.915
442689 97.368 -64.879 51.865
442690 97.406 -64.878 51.815
442691 97.442 -64.878 51.767
442692 97.473 -64.883 51.721
442693 97.504 -64.887 51.676
442694 97.535 -64.891 51.633
442695 97.567 -64.896 51.589
442696 97.598 -64.900 51.546
442697 97.629 -64.904 51.503
442698 97.660 -64.909 51.460
442699 97.690 -64.913 51.416
442700 97.718 -64.917 51.373
442701 97.746 -64.921 51.331
442702 97.774 -64.925 51.289
442703 97.801 -64.928 51.248
442704 97.829 -64.932 51.212
442705 97.857 -64.937 51.174
442706 97.887 -64.940 51.136
442707 97.924 -64.941 51.097
442708 97.963 -64.946 51.059
442709 98.004 -64.950 51.021
442710 98.044 -64.954 50.983
442711 98.083 -64.958 50.945
442712 98.121 -64.961 50.905
442713 98.159 -64.965 50.865
442714 98.197 -64.968 50.823
442715 98.235 -64.971 50.781
442716 98.273 -64.974 50.740
442717 98.311 -64.977 50.701
442718 98.349 -64.981 50.661
442719 98.387 -64.984 50.620
442720 98.425 -64.987 50.575
442721 98.462 -64.990 50.527
442722 98.498 -64.993 50.475
442723 98.532 -64.995 50.425
442724 98.568 -64.998 50.377
442725 98.602 -65.001 50.327
442726 98.638 -65.004 50.276
442727 98.672 -65.007 50.224
442728 98.707 -65.009 50.172
442729 98.742 -65.012 50.122
442730 98.777 -65.015 50.073
442731 98.812 -65.017 50.027
442732 98.849 -65.022 49.981
442733 98.891 -65.027 49.940
442734 98.942 -65.029 49.897
442735 98.995 -65.031 49.855
442736 99.048 -65.032 49.814
442737 99.101 -65.033 49.774
442738 99.154 -65.034 49.736
442739 99.208 -65.035 49.701
442740 99.262 -65.035 49.666
442741 99.315 -65.036 49.628
442742 99.369 -65.036 49.586
442743 99.421 -65.039 49.543
442744 99.477 -65.047 49.497
442745 99.535 -65.055 49.451
442746 99.593 -65.063 49.404
442747 99.651 -65.071 49.358
442748 99.709 -65.079 49.312
442749 99.784 -65.070 49.267
442750 99.870 -65.045 49.221
442751 99.953 -65.021 49.175
442752 100.036 -64.997 49.132
442753 100.119 -64.972 49.089
442754 100.188 -64.962 49.046
442755 100.179 -65.026 49.002
442756 100.174 -65.075 48.959
442757 100.203 -65.048 48.916
442758 100.247 -65.042 48.874
442759 100.295 -65.040 48.834
442760 100.344 -65.038 48.793
442761 100.385 -65.041 48.751
442762 100.424 -65.046 48.710
442763 100.463 -65.051 48.668
442764 100.501 -65.057 48.628
442765 100.540 -65.062 48.589
442766 100.578 -65.067 48.550
443157 87.311 -63.487 55.764
443158 87.345 -63.496 55.747
443159 87.379 -63.506 55.729
443160 87.414 -63.515 55.711
443161 87.448 -63.525 55.693
443162 87.482 -63.534 55.674
443163 87.516 -63.544 55.656
443164 87.551 -63.553 55.639
443165 87.585 -63.563 55.621
443166 87.619 -63.573 55.602
443167 87.653 -63.582 55.584
443168 87.688 -63.592 55.566
443169 87.731 -63.597 55.551
443189 88.743 -63.730 55.325
443190 88.791 -63.741 55.325
443191 88.840 -63.751 55.326
443192 88.889 -63.761 55.327
443193 88.939 -63.772 55.329
443194 88.987 -63.782 55.333
443195 89.036 -63.792 55.338
443196 89.085 -63.803 55.344
443197 89.134 -63.813 55.352
443198 89.183 -63.823 55.363
443199 89.232 -63.834 55.377
443200 89.274 -63.847 55.393
443201 89.311 -63.864 55.411
443202 89.347 -63.881 55.424
443203 89.384 -63.897 55.427
443204 89.420 -63.914 55.426
443205 89.456 -63.931 55.424
443206 89.493 -63.947 55.421
443207 89.529 -63.964 55.418
443208 89.571 -63.974 55.407
443209 89.616 -63.982 55.390
443210 89.661 -63.988 55.367
443211 89.707 -63.985 55.340
443212 89.754 -63.982 55.308
443213 89.801 -63.979 55.282
443214 89.847 -63.976 55.261
443215 89.893 -63.973 55.243
443216 89.940 -63.969 55.229
443217 89.987 -63.967 55.218
443218 90.036 -63.966 55.211
443219 90.086 -63.964 55.207
443220 90.135 -63.963 55.207
443221 90.185 -63.962 55.210
443222 90.234 -63.960 55.220
443223 90.284 -63.960 55.229
443224 90.333 -63.958 55.241
443225 90.383 -63.957 55.256
443226 90.432 -63.958 55.275
443227 90.476 -63.975 55.298
443228 90.520 -63.991 55.310
443229 90.564 -64.008 55.311
443230 90.608 -64.025 55.304
443231 90.653 -64.041 55.281
443232 90.697 -64.055 55.246
443233 90.743 -64.063 55.204
443234 90.790 -64.070 55.158
443235 90.836 -64.078 55.108
443236 90.882 -64.085 55.060
443237 90.928 -64.093 55.018
443238 90.974 -64.101 54.978
443239 91.020 -64.108 54.942
443240 91.067 -64.116 54.907
443241 91.112 -64.123 54.877
443242 91.159 -64.131 54.851
443243 91.205 -64.139 54.829
443244 91.251 -64.146 54.809
443245 91.297 -64.154 54.791
443246 91.344 -64.161 54.777
443247 91.389 -64.169 54.763
443248 91.436 -64.176 54.751
443249 91.482 -64.184 54.740
443250 91.528 -64.192 54.730
443251 91.574 -64.199 54.721
443252 91.620 -64.207 54.712
443253 91.666 -64.215 54.703
443254 91.711 -64.225 54.694
443255 91.756 -64.235 54.686
443256 91.802 -64.245 54.679
443257 91.847 -64.255 54.672
443258 91.892 -64.264 54.665
443259 91.937 -64.274 54.658
443260 91.982 -64.285 54.653
443261 92.027 -64.295 54.648
443262 92.073 -64.305 54.644
443263 92.118 -64.314 54.639
443264 92.163 -64.324 54.634
443265 92.208 -64.334 54.630
443266 92.256 -64.344 54.627
443267 92.304 -64.354 54.625
443268 92.353 -64.364 54.623
443269 92.400 -64.377 54.622
443270 92.440 -64.383 54.618
443271 92.469 -64.396 54.615
443272 92.498 -64.409 54.611
443273 92.527 -64.422 54.608
443274 92.556 -64.435 54.605
443275 92.585 -64.448 54.603
443276 92.615 -64.461 54.599
443277 92.655 -64.474 54.594
443278 92.704 -64.486 54.590
443279 92.754 -64.499 54.586
443280 92.803 -64.512 54.584
443281 92.852 -64.525 54.582
443282 92.902 -64.538 54.582
443283 92.962 -64.546 54.584
443284 93.039 -64.545 54.587
443285 93.116 -64.544 54.591
443286 93.170 -64.547 54.597
443287 93.219 -64.550 54.604
443288 93.267 -64.554 54.611
443289 93.315 -64.557 54.617
443290 93.362 -64.559 54.622
443291 93.400 -64.561 54.625
443292 93.424 -64.568 54.627
443293 93.459 -64.571 54.625
443294 93.499 -64.569 54.618
443295 93.540 -64.568 54.608
443296 93.580 -64.567 54.597
443297 93.621 -64.565 54.588
443298 93.661 -64.565 54.580
443299 93.702 -64.563 54.570
443300 93.743 -64.562 54.559
443301 93.783 -64.561 54.547
443302 93.823 -64.560 54.534
443303 93.860 -64.569 54.520
443304 93.896 -64.578 54.507
443305 93.932 -64.588 54.492
443306 93.969 -64.597 54.477
443307 94.014 -64.607 54.460
443308 94.059 -64.616 54.439
443309 94.100 -64.625 54.417
443310 94.129 -64.631 54.394
443311 94.158 -64.636 54.371
443312 94.187 -64.642 54.347
443313 94.218 -64.643 54.321
443314 94.251 -64.642 54.293
443315 94.283 -64.640 54.264
443316 94.316 -64.638 54.236
443317 94.348 -64.637 54.214
443318 94.381 -64.635 54.196
443319 94.437 -64.641 54.179
443320 94.499 -64.649 54.164
443321 94.561 -64.658 54.148
443322 94.624 -64.666 54.129
443323 94.675 -64.676 54.109
443324 94.726 -64.686 54.087
443325 94.777 -64.696 54.066
443326 94.826 -64.704 54.045
443327 94.873 -64.709 54.024
443328 94.918 -64.715 53.999
443329 94.965 -64.721 53.975
443330 95.011 -64.726 53.951
443331 95.041 -64.728 53.928
443332 95.071 -64.730 53.909
443333 95.113 -64.739 53.889
443334 95.156 -64.748 53.869
443335 95.198 -64.757 53.847
443336 95.242 -64.765 53.825
443337 95.291 -64.774 53.803
443338 95.340 -64.781 53.781
443339 95.389 -64.790 53.755
443340 95.437 -64.797 53.723
443341 95.476 -64.804 53.689
443342 95.517 -64.812 53.654
443343 95.557 -64.819 53.622
443344 95.597 -64.826 53.593
443345 95.637 -64.833 53.565
443346 95.677 -64.840 53.541
443347 95.718 -64.847 53.524
443348 95.758 -64.854 53.506
443349 95.798 -64.862 53.486
443350 95.837 -64.868 53.463
443351 95.873 -64.869 53.436
443352 95.912 -64.859 53.401
443353 95.951 -64.839 53.369
443354 95.989 -64.817 53.339
443355 96.028 -64.795 53.312
443356 96.065 -64.773 53.287
443357 96.104 -64.751 53.257
443358 96.142 -64.729 53.222
443359 96.181 -64.726 53.184
443360 96.221 -64.728 53.142
443361 96.260 -64.731 53.097
443362 96.300 -64.740 53.054
443363 96.340 -64.748 53.013
443364 96.381 -64.756 52.972
443365 96.421 -64.765 52.932
443366 96.461 -64.774 52.895
443367 96.501 -64.782 52.859
443368 96.541 -64.791 52.824
443369 96.583 -64.799 52.787
443370 96.618 -64.803 52.751
443371 96.651 -64.804 52.708
443372 96.689 -64.808 52.664
443373 96.727 -64.812 52.621
443374 96.765 -64.815 52.579
443375 96.803 -64.819 52.537
443376 96.841 -64.823 52.493
443377 96.879 -64.828 52.446
443378 96.917 -64.831 52.399
443379 96.955 -64.835 52.350
443380 96.993 -64.839 52.303
443381 97.031 -64.843 52.259
443382 97.069 -64.847 52.219
443383 97.107 -64.850 52.180
443384 97.145 -64.850 52.139
443385 97.183 -64.850 52.096
443386 97.222 -64.848 52.051
443387 97.260 -64.847 52.005
443388 97.298 -64.846 51.958
443389 97.336 -64.844 51.909
443390 97.375 -64.843 51.857
443391 97.413 -64.841 51.806
443392 97.441 -64.837 51.757
443393 97.476 -64.842 51.710
443394 97.507 -64.847 51.665
443395 97.538 -64.850 51.623
443396 97.570 -64.855 51.577
443397 97.601 -64.859 51.532
443398 97.632 -64.863 51.489
443399 97.664 -64.868 51.446
443400 97.695 -64.872 51.402
443401 97.723 -64.876 51.359
443402 97.750 -64.880 51.316
443403 97.778 -64.884 51.274
443404 97.806 -64.888 51.232
443405 97.833 -64.892 51.194
443406 97.861 -64.896 51.156
443407 97.895 -64.898 51.119
443408 97.932 -64.899 51.081
443409 97.968 -64.901 51.042
443410 98.008 -64.906 51.004
443411 98.049 -64.910 50.966
443412 98.087 -64.914 50.927
443413 98.123 -64.919 50.888
443414 98.160 -64.923 50.847
443415 98.198 -64.926 50.804
443416 98.236 -64.929 50.760
443417 98.274 -64.932 50.716
443418 98.312 -64.936 50.675
443419 98.350 -64.939 50.635
443420 98.387 -64.942 50.593
443421 98.426 -64.945 50.548
443422 98.462 -64.948 50.499
443423 98.497 -64.951 50.448
443424 98.532 -64.953 50.397
443425 98.567 -64.957 50.345
443426 98.602 -64.960 50.294
443427 98.637 -64.962 50.243
443428 98.671 -64.965 50.192
443429 98.707 -64.967 50.143
443430 98.741 -64.970 50.096
443431 98.777 -64.972 50.050
443432 98.811 -64.975 50.006
443433 98.849 -64.977 49.963
443434 98.891 -64.976 49.924
443435 98.946 -64.976 49.882
443436 99.000 -64.977 49.840
443437 99.054 -64.977 49.799
443438 99.108 -64.978 49.761
443439 99.161 -64.978 49.722
443440 99.215 -64.978 49.686
443441 99.270 -64.978 49.649
443442 99.324 -64.979 49.611
443443 99.375 -64.983 49.569
443444 99.423 -64.991 49.524
443445 99.476 -64.999 49.478
443446 99.534 -65.007 49.432
443447 99.592 -65.014 49.386
443448 99.650 -65.022 49.340
443449 99.708 -65.030 49.296
443450 99.793 -65.013 49.251
443451 99.888 -64.987 49.207
443452 99.972 -64.962 49.162
443453 100.055 -64.938 49.119
443454 100.138 -64.913 49.077
443455 100.204 -64.906 49.034
443456 100.195 -64.970 48.991
443457 100.188 -65.027 48.949
443458 100.216 -65.001 48.907
443459 100.262 -64.996 48.866
443460 100.310 -64.994 48.825
443461 100.356 -64.993 48.785
443462 100.395 -64.998 48.743
443463 100.434 -65.004 48.701
443464 100.472 -65.009 48.661
443465 100.511 -65.014 48.621
443466 100.550 -65.019 48.581
443467 100.588 -65.024 48.543
443857 87.269 -63.429 55.786
443858 87.303 -63.438 55.770
443859 87.337 -63.448 55.754
443860 87.371 -63.457 55.736
443861 87.406 -63.467 55.719
443862 87.440 -63.477 55.702
443863 87.474 -63.486 55.684
443864 87.508 -63.496 55.667
443865 87.543 -63.506 55.650
443866 87.577 -63.515 55.632
443867 87.611 -63.524 55.614
443868 87.645 -63.534 55.597
443869 87.687 -63.540 55.580
443870 87.733 -63.544 55.564
443871 87.779 -63.548 55.548
443872 87.826 -63.552 55.531
443890 88.744 -63.690 55.327
443891 88.793 -63.701 55.327
443892 88.842 -63.712 55.327
443893 88.891 -63.722 55.328
443894 88.940 -63.732 55.330
443895 88.989 -63.743 55.334
443896 89.038 -63.753 55.339
443897 89.087 -63.763 55.345
443898 89.136 -63.774 55.353
443899 89.185 -63.784 55.364
443900 89.234 -63.794 55.378
443901 89.280 -63.806 55.395
443902 89.316 -63.823 55.413
443903 89.353 -63.840 55.428
443904 89.389 -63.857 55.434
443905 89.426 -63.873 55.435
443906 89.462 -63.889 55.433
443907 89.498 -63.906 55.431
443908 89.539 -63.918 55.427
443909 89.584 -63.925 55.416
443910 89.628 -63.932 55.398
443911 89.674 -63.936 55.375
443912 89.720 -63.933 55.350
443913 89.767 -63.930 55.320
443914 89.813 -63.927 55.294
443915 89.860 -63.924 55.272
443916 89.906 -63.921 55.255
443917 89.953 -63.918 55.239
443918 89.999 -63.915 55.228
443919 90.046 -63.912 55.221
443920 90.095 -63.910 55.218
443921 90.145 -63.910 55.217
443922 90.194 -63.908 55.221
443923 90.244 -63.907 55.230
443924 90.293 -63.906 55.240
443925 90.342 -63.904 55.252
443926 90.392 -63.903 55.266
443927 90.438 -63.913 55.284
443928 90.482 -63.930 55.305
443929 90.526 -63.947 55.315
443930 90.571 -63.963 55.317
443931 90.615 -63.980 55.308
443932 90.659 -63.997 55.283
443933 90.704 -64.011 55.245
443934 90.750 -64.019 55.203
443935 90.796 -64.026 55.158
443936 90.842 -64.034 55.110
443937 90.888 -64.041 55.064
443938 90.934 -64.049 55.017
443939 90.980 -64.057 54.975
443940 91.027 -64.064 54.937
443941 91.073 -64.072 54.902
443942 91.119 -64.079 54.871
443943 91.165 -64.087 54.846
443944 91.211 -64.095 54.823
443945 91.257 -64.102 54.803
443946 91.303 -64.110 54.785
443947 91.350 -64.117 54.771
443948 91.396 -64.125 54.757
443949 91.442 -64.132 54.746
443950 91.488 -64.140 54.736
443951 91.534 -64.148 54.726
443952 91.580 -64.155 54.717
443953 91.627 -64.163 54.708
443954 91.673 -64.170 54.700
443955 91.719 -64.178 54.692
443956 91.765 -64.186 54.684
443957 91.810 -64.196 54.677
443958 91.855 -64.206 54.670
443959 91.900 -64.216 54.664
443960 91.945 -64.226 54.658
443961 91.991 -64.236 54.653
443962 92.036 -64.245 54.647
443963 92.081 -64.255 54.643
443964 92.126 -64.265 54.638
443965 92.171 -64.276 54.633
443966 92.216 -64.286 54.630
443967 92.265 -64.295 54.627
443968 92.313 -64.305 54.625
443969 92.362 -64.315 54.624
443970 92.405 -64.326 54.623
443971 92.434 -64.339 54.618
443972 92.463 -64.352 54.614
443973 92.493 -64.364 54.611
443974 92.522 -64.377 54.607
443975 92.551 -64.389 54.604
443976 92.580 -64.402 54.601
443977 92.609 -64.415 54.597
443978 92.648 -64.428 54.592
443979 92.697 -64.441 54.587
443980 92.746 -64.454 54.585
443981 92.796 -64.468 54.581
443982 92.845 -64.480 54.580
443983 92.901 -64.490 54.581
443984 92.978 -64.490 54.582
443985 93.055 -64.489 54.587
443986 93.129 -64.488 54.593
443987 93.177 -64.492 54.600
443988 93.225 -64.495 54.606
443989 93.274 -64.499 54.613
443990 93.321 -64.500 54.618
443991 93.366 -64.500 54.622
443992 93.398 -64.505 54.624
443993 93.423 -64.512 54.624
443994 93.449 -64.518 54.621
443995 93.490 -64.518 54.614
443996 93.531 -64.516 54.604
443997 93.571 -64.515 54.594
443998 93.612 -64.514 54.584
443999 93.652 -64.512 54.575
444000 93.693 -64.511 54.564
444001 93.733 -64.510 54.552
444002 93.774 -64.508 54.540
444003 93.812 -64.513 54.527
444004 93.849 -64.522 54.514
444005 93.885 -64.531 54.500
444006 93.922 -64.540 54.485
444007 93.965 -64.550 54.469
444008 94.010 -64.560 54.450
444009 94.056 -64.570 54.428
444010 94.092 -64.577 54.404
444011 94.121 -64.583 54.379
444012 94.150 -64.588 54.354
444013 94.179 -64.594 54.328
444014 94.208 -64.599 54.301
444015 94.241 -64.597 54.274
444016 94.273 -64.596 54.245
444017 94.306 -64.594 54.218
444018 94.338 -64.593 54.195
444019 94.371 -64.590 54.175
444020 94.430 -64.598 54.159
444021 94.493 -64.606 54.144
444022 94.555 -64.615 54.130
444023 94.611 -64.624 54.112
444024 94.662 -64.634 54.094
444025 94.713 -64.643 54.074
444026 94.764 -64.653 54.053
444027 94.815 -64.662 54.032
444028 94.860 -64.668 54.011
444029 94.907 -64.674 53.986
444030 94.953 -64.679 53.961
444031 94.999 -64.684 53.936
444032 95.045 -64.690 53.910
444033 95.075 -64.692 53.888
444034 95.113 -64.699 53.868
444035 95.156 -64.708 53.847
444036 95.199 -64.716 53.825
444037 95.248 -64.725 53.803
444038 95.297 -64.732 53.781
444039 95.345 -64.740 53.759
444040 95.395 -64.748 53.733
444041 95.444 -64.756 53.703
444042 95.487 -64.763 53.669
444043 95.528 -64.771 53.634
444044 95.567 -64.778 53.603
444045 95.608 -64.785 53.573
444046 95.648 -64.792 53.545
444047 95.688 -64.800 53.522
444048 95.729 -64.806 53.504
444049 95.768 -64.813 53.484
444050 95.809 -64.821 53.463
444051 95.845 -64.825 53.441
444052 95.879 -64.828 53.414
444053 95.915 -64.827 53.380
444054 95.955 -64.818 53.347
444055 95.995 -64.809 53.317
444056 96.035 -64.800 53.291
444057 96.074 -64.778 53.267
444058 96.112 -64.756 53.239
444059 96.150 -64.734 53.206
444060 96.189 -64.734 53.169
444061 96.229 -64.736 53.128
444062 96.268 -64.737 53.083
444063 96.307 -64.739 53.039
444064 96.346 -64.741 52.997
444065 96.386 -64.743 52.955
444066 96.425 -64.744 52.915
444067 96.464 -64.747 52.879
444068 96.504 -64.748 52.845
444069 96.543 -64.751 52.812
444070 96.582 -64.758 52.779
444071 96.620 -64.762 52.745
444072 96.658 -64.765 52.704
444073 96.696 -64.770 52.661
444074 96.734 -64.774 52.617
444075 96.772 -64.778 52.574
444076 96.810 -64.781 52.529
444077 96.848 -64.785 52.484
444078 96.886 -64.789 52.437
444079 96.924 -64.794 52.389
444080 96.962 -64.797 52.340
444081 97.000 -64.801 52.295
444082 97.038 -64.805 52.250
444083 97.076 -64.809 52.209
444084 97.114 -64.812 52.168
444085 97.152 -64.814 52.126
444086 97.190 -64.813 52.083
444087 97.228 -64.812 52.039
444088 97.266 -64.810 51.994
444089 97.305 -64.809 51.949
444090 97.343 -64.808 51.903
444091 97.381 -64.806 51.851
444092 97.410 -64.799 51.800
444093 97.439 -64.794 51.753
444094 97.474 -64.800 51.707
444095 97.510 -64.805 51.660
444096 97.541 -64.809 51.616
444097 97.573 -64.814 51.570
444098 97.604 -64.819 51.525
444099 97.636 -64.823 51.481
444100 97.667 -64.827 51.437
444101 97.698 -64.831 51.392
444102 97.727 -64.835 51.347
444103 97.754 -64.839 51.303
444104 97.782 -64.844 51.259
444105 97.810 -64.847 51.217
444106 97.838 -64.851 51.177
444107 97.868 -64.854 51.138
444108 97.904 -64.856 51.101
444109 97.940 -64.857 51.064
444110 97.977 -64.859 51.025
444111 98.013 -64.861 50.987
444112 98.054 -64.865 50.948
444113 98.093 -64.869 50.908
444114 98.129 -64.874 50.867
444115 98.165 -64.878 50.826
444116 98.201 -64.883 50.783
444117 98.236 -64.888 50.738
444118 98.274 -64.891 50.693
444119 98.312 -64.894 50.650
444120 98.350 -64.897 50.608
444121 98.388 -64.900 50.564
444122 98.426 -64.904 50.518
444123 98.461 -64.906 50.472
444124 98.496 -64.910 50.424
444125 98.531 -64.912 50.376
444126 98.566 -64.915 50.324
444127 98.601 -64.917 50.271
444128 98.636 -64.920 50.218
444129 98.671 -64.923 50.168
444130 98.711 -64.924 50.121
444131 98.750 -64.925 50.075
444132 98.790 -64.925 50.030
444133 98.830 -64.926 49.988
444134 98.870 -64.927 49.947
444135 98.912 -64.924 49.909
444136 98.955 -64.920 49.868
444137 99.008 -64.920 49.827
444138 99.062 -64.920 49.786
444139 99.116 -64.921 49.748
444140 99.169 -64.921 49.710
444141 99.224 -64.922 49.673
444142 99.278 -64.922 49.634
444143 99.329 -64.926 49.593
444144 99.376 -64.934 49.550
444145 99.424 -64.942 49.503
444146 99.476 -64.950 49.457
444147 99.534 -64.958 49.411
444148 99.591 -64.966 49.367
444149 99.650 -64.973 49.324
444150 99.709 -64.980 49.281
444151 99.803 -64.954 49.237
444152 99.897 -64.928 49.193
444153 99.991 -64.903 49.149
444154 100.074 -64.878 49.106
444155 100.157 -64.853 49.064
444156 100.219 -64.849 49.022
444157 100.211 -64.914 48.981
444158 100.202 -64.978 48.939
444159 100.229 -64.954 48.897
444160 100.276 -64.950 48.856
444161 100.324 -64.947 48.816
444162 100.366 -64.950 48.776
444163 100.405 -64.956 48.736
444164 100.443 -64.961 48.694
444165 100.482 -64.966 48.653
444166 100.521 -64.971 48.614
444167 100.560 -64.976 48.575
444168 100.599 -64.982 48.536
444556 87.192 -63.361 55.823
444557 87.226 -63.371 55.807
444558 87.261 -63.380 55.792
444559 87.295 -63.390 55.776
444560 87.329 -63.399 55.761
444561 87.363 -63.409 55.744
444562 87.398 -63.418 55.728
444563 87.432 -63.428 55.711
444564 87.466 -63.437 55.695
444565 87.500 -63.447 55.678
444566 87.535 -63.457 55.661
444567 87.569 -63.466 55.643
444568 87.603 -63.476 55.626
444569 87.643 -63.483 55.610
444570 87.689 -63.487 55.593
444571 87.735 -63.490 55.577
444572 87.782 -63.495 55.561
444573 87.835 -63.495 55.545
444574 87.898 -63.491 55.527
444575 87.961 -63.487 55.508
444591 88.746 -63.651 55.330
444592 88.794 -63.662 55.329
444593 88.843 -63.672 55.328
444594 88.893 -63.682 55.330
444595 88.942 -63.693 55.332
444596 88.990 -63.703 55.335
444597 89.039 -63.713 55.340
444598 89.089 -63.724 55.347
444599 89.137 -63.734 55.354
444600 89.186 -63.744 55.365
444601 89.235 -63.755 55.378
444602 89.284 -63.766 55.395
444603 89.322 -63.781 55.414
444604 89.358 -63.798 55.430
444605 89.395 -63.815 55.438
444606 89.431 -63.831 55.441
444607 89.468 -63.848 55.441
444608 89.507 -63.862 55.440
444609 89.551 -63.869 55.436
444610 89.596 -63.875 55.425
444611 89.641 -63.882 55.406
444612 89.686 -63.884 55.383
444613 89.733 -63.882 55.359
444614 89.779 -63.878 55.332
444615 89.826 -63.875 55.305
444616 89.873 -63.872 55.283
444617 89.919 -63.869 55.264
444618 89.965 -63.866 55.249
444619 90.012 -63.863 55.238
444620 90.058 -63.860 55.231
444621 90.105 -63.857 55.228
444622 90.154 -63.856 55.229
444623 90.203 -63.854 55.232
444624 90.253 -63.853 55.240
444625 90.302 -63.852 55.250
444626 90.352 -63.850 55.264
444627 90.400 -63.853 55.278
444628 90.445 -63.869 55.293
444629 90.488 -63.886 55.308
444630 90.533 -63.903 55.317
444631 90.577 -63.919 55.318
444632 90.621 -63.936 55.308
444633 90.665 -63.953 55.281
444634 90.710 -63.967 55.241
444635 90.756 -63.975 55.198
444636 90.802 -63.982 55.153
444637 90.849 -63.990 55.106
444638 90.894 -63.998 55.062
444639 90.941 -64.005 55.015
444640 90.987 -64.013 54.972
444641 91.033 -64.020 54.932
444642 91.079 -64.028 54.896
444643 91.125 -64.035 54.866
444644 91.171 -64.043 54.840
444645 91.218 -64.051 54.817
444646 91.263 -64.058 54.798
444647 91.310 -64.066 54.780
444648 91.356 -64.073 54.766
444649 91.402 -64.081 54.753
444650 91.448 -64.088 54.742
444651 91.495 -64.096 54.732
444652 91.540 -64.104 54.723
444653 91.587 -64.111 54.714
444654 91.633 -64.119 54.706
444655 91.679 -64.126 54.698
444656 91.725 -64.134 54.690
444657 91.771 -64.142 54.682
444658 91.817 -64.149 54.676
444659 91.864 -64.157 54.670
444660 91.909 -64.167 54.663
444661 91.954 -64.176 54.658
444662 91.999 -64.187 54.652
444663 92.044 -64.197 54.647
444664 92.089 -64.207 54.642
444665 92.135 -64.217 54.638
444666 92.179 -64.226 54.633
444667 92.225 -64.236 54.629
444668 92.274 -64.246 54.626
444669 92.322 -64.256 54.624
444670 92.368 -64.266 54.623
444671 92.406 -64.276 54.621
444672 92.430 -64.292 54.617
444673 92.458 -64.306 54.612
444674 92.487 -64.319 54.607
444675 92.516 -64.332 54.604
444676 92.545 -64.345 54.600
444677 92.574 -64.358 54.597
444678 92.603 -64.371 54.593
444679 92.641 -64.383 54.588
444680 92.690 -64.396 54.583
444681 92.739 -64.409 54.581
444682 92.789 -64.422 54.579
444683 92.840 -64.434 54.578
444684 92.917 -64.433 54.579
444685 92.994 -64.433 54.581
444686 93.070 -64.432 54.586
444687 93.135 -64.433 54.593
444688 93.183 -64.436 54.600
444689 93.231 -64.440 54.606
444690 93.279 -64.443 54.612
444691 93.325 -64.443 54.617
444692 93.371 -64.443 54.619
444693 93.396 -64.449 54.619
444694 93.421 -64.456 54.617
444695 93.446 -64.464 54.613
444696 93.481 -64.465 54.606
444697 93.521 -64.464 54.597
444698 93.562 -64.463 54.587
444699 93.603 -64.461 54.576
444700 93.643 -64.461 54.566
444701 93.684 -64.459 54.554
444702 93.724 -64.458 54.542
444703 93.765 -64.457 54.530
444704 93.801 -64.466 54.519
444705 93.838 -64.475 54.506
444706 93.874 -64.484 54.492
444707 93.916 -64.493 54.476
444708 93.962 -64.503 54.458
444709 94.007 -64.513 54.438
444710 94.052 -64.523 54.414
444711 94.084 -64.530 54.389
444712 94.113 -64.535 54.362
444713 94.142 -64.540 54.334
444714 94.171 -64.546 54.308
444715 94.200 -64.552 54.281
444716 94.230 -64.553 54.254
444717 94.263 -64.552 54.226
444718 94.296 -64.550 54.200
444719 94.329 -64.549 54.177
444720 94.361 -64.547 54.156
444721 94.424 -64.555 54.139
444722 94.486 -64.564 54.124
444723 94.547 -64.572 54.110
444724 94.599 -64.582 54.094
444725 94.649 -64.592 54.077
444726 94.700 -64.602 54.058
444727 94.752 -64.612 54.039
444728 94.803 -64.621 54.020
444729 94.849 -64.627 54.000
444730 94.895 -64.632 53.977
444731 94.941 -64.637 53.952
444732 94.987 -64.643 53.925
444733 95.033 -64.649 53.896
444734 95.079 -64.654 53.870
444735 95.113 -64.659 53.846
444736 95.156 -64.667 53.823
444737 95.205 -64.675 53.801
444738 95.254 -64.683 53.778
444739 95.303 -64.691 53.757
444740 95.351 -64.699 53.735
444741 95.400 -64.707 53.709
444742 95.449 -64.715 53.679
444743 95.498 -64.723 53.647
444744 95.538 -64.730 53.615
444745 95.579 -64.737 53.584
444746 95.619 -64.744 53.553
444747 95.659 -64.751 53.526
444748 95.699 -64.759 53.500
444749 95.739 -64.765 53.480
444750 95.779 -64.773 53.460
444751 95.819 -64.779 53.438
444752 95.852 -64.782 53.416
444753 95.887 -64.785 53.390
444754 95.920 -64.788 53.358
444755 95.958 -64.785 53.325
444756 95.998 -64.776 53.294
444757 96.038 -64.767 53.269
444758 96.078 -64.758 53.245
444759 96.118 -64.748 53.219
444760 96.158 -64.740 53.188
444761 96.197 -64.737 53.152
444762 96.237 -64.736 53.112
444763 96.275 -64.734 53.069
444764 96.315 -64.731 53.024
444765 96.354 -64.729 52.981
444766 96.393 -64.727 52.938
444767 96.432 -64.725 52.898
444768 96.471 -64.722 52.862
444769 96.511 -64.720 52.830
444770 96.550 -64.718 52.800
444771 96.586 -64.718 52.769
444772 96.626 -64.723 52.736
444773 96.665 -64.728 52.697
444774 96.703 -64.731 52.655
444775 96.741 -64.736 52.611
444776 96.779 -64.740 52.567
444777 96.817 -64.743 52.520
444778 96.855 -64.747 52.473
444779 96.893 -64.751 52.425
444780 96.931 -64.755 52.378
444781 96.969 -64.759 52.331
444782 97.007 -64.763 52.288
444783 97.045 -64.767 52.243
444784 97.083 -64.771 52.200
444785 97.121 -64.775 52.157
444786 97.159 -64.778 52.114
444787 97.197 -64.777 52.069
444788 97.235 -64.775 52.026
444789 97.274 -64.775 51.982
444790 97.312 -64.773 51.939
444791 97.350 -64.772 51.893
444792 97.384 -64.768 51.843
444793 97.408 -64.756 51.794
444794 97.437 -64.753 51.748
444795 97.472 -64.758 51.703
444796 97.508 -64.763 51.657
444797 97.543 -64.769 51.612
444798 97.576 -64.773 51.565
444799 97.607 -64.778 51.519
444800 97.639 -64.782 51.474
444801 97.670 -64.786 51.429
444802 97.701 -64.790 51.383
444803 97.731 -64.795 51.337
444804 97.759 -64.799 51.291
444805 97.787 -64.803 51.246
444806 97.814 -64.806 51.202
444807 97.842 -64.810 51.161
444808 97.876 -64.812 51.121
444809 97.913 -64.814 51.082
444810 97.949 -64.815 51.044
444811 97.985 -64.817 51.006
444812 98.021 -64.819 50.968
444813 98.058 -64.820 50.928
444814 98.098 -64.825 50.887
444815 98.135 -64.829 50.846
444816 98.170 -64.834 50.804
444817 98.206 -64.838 50.760
444818 98.242 -64.843 50.715
444819 98.278 -64.848 50.669
444820 98.313 -64.853 50.626
444821 98.351 -64.856 50.583
444822 98.389 -64.859 50.539
444823 98.426 -64.862 50.494
444824 98.460 -64.865 50.448
444825 98.495 -64.868 50.402
444826 98.532 -64.869 50.353
444827 98.572 -64.870 50.303
444828 98.612 -64.871 50.250
444829 98.652 -64.872 50.197
444830 98.692 -64.872 50.148
444831 98.731 -64.873 50.101
444832 98.771 -64.874 50.056
444833 98.811 -64.875 50.013
444834 98.851 -64.876 49.972
444835 98.891 -64.877 49.932
444836 98.933 -64.873 49.894
444837 98.977 -64.869 49.854
444838 99.019 -64.865 49.813
444839 99.070 -64.864 49.772
444840 99.124 -64.864 49.734
444841 99.178 -64.865 49.696
444842 99.232 -64.865 49.658
444843 99.282 -64.869 49.618
444844 99.330 -64.878 49.575
444845 99.378 -64.885 49.530
444846 99.426 -64.894 49.483
444847 99.475 -64.901 49.437
444848 99.533 -64.909 49.393
444849 99.591 -64.917 49.350
444850 99.649 -64.925 49.308
444851 99.718 -64.922 49.265
444852 99.812 -64.897 49.222
444853 99.906 -64.871 49.178
444854 100.000 -64.845 49.135
444855 100.093 -64.819 49.093
444856 100.176 -64.794 49.051
444857 100.235 -64.793 49.010
444858 100.227 -64.858 48.970
444859 100.218 -64.922 48.929
444860 100.243 -64.906 48.887
444861 100.290 -64.904 48.847
444862 100.338 -64.903 48.807
444863 100.376 -64.908 48.768
444864 100.415 -64.913 48.728
444865 100.454 -64.918 48.688
444866 100.492 -64.923 48.647
444867 100.531 -64.928 48.607
444868 100.570 -64.934 48.569
444869 100.609 -64.939 48.530
445256 87.150 -63.303 55.841
445257 87.184 -63.313 55.826
445258 87.218 -63.322 55.811
445259 87.253 -63.332 55.797
445260 87.287 -63.341 55.782
445261 87.321 -63.351 55.767
445262 87.355 -63.361 55.751
445263 87.390 -63.370 55.736
445264 87.424 -63.380 55.720
445265 87.458 -63.390 55.704
445266 87.493 -63.399 55.688
445267 87.527 -63.408 55.672
445268 87.561 -63.418 55.655
445269 87.600 -63.426 55.639
445270 87.645 -63.430 55.623
445271 87.692 -63.433 55.606
445272 87.738 -63.437 55.590
445273 87.797 -63.435 55.574
445274 87.860 -63.430 55.558
445275 87.915 -63.435 55.540
445276 87.964 -63.446 55.522
445277 88.013 -63.456 55.504
445278 88.062 -63.466 55.486
445292 88.747 -63.612 55.332
445293 88.796 -63.622 55.331
445294 88.845 -63.633 55.330
445295 88.894 -63.643 55.331
445296 88.943 -63.653 55.333
445297 88.992 -63.664 55.336
445298 89.041 -63.674 55.341
445299 89.090 -63.684 55.348
445300 89.139 -63.695 55.355
445301 89.188 -63.705 55.364
445302 89.237 -63.715 55.377
445303 89.286 -63.726 55.394
445304 89.328 -63.740 55.414
445305 89.365 -63.756 55.429
445306 89.401 -63.773 55.439
445307 89.438 -63.790 55.444
445308 89.474 -63.806 55.446
445309 89.519 -63.813 55.446
445310 89.564 -63.819 55.444
445311 89.608 -63.826 55.433
445312 89.653 -63.833 55.415
445313 89.699 -63.832 55.392
445314 89.746 -63.829 55.368
445315 89.792 -63.826 55.342
445316 89.839 -63.823 55.315
445317 89.885 -63.820 55.292
445318 89.932 -63.818 55.274
445319 89.978 -63.815 55.259
445320 90.025 -63.812 55.248
445321 90.072 -63.809 55.241
445322 90.118 -63.806 55.238
445323 90.164 -63.803 55.238
445324 90.213 -63.800 55.241
445325 90.262 -63.799 55.252
445326 90.312 -63.798 55.265
445327 90.361 -63.797 55.279
445328 90.407 -63.809 55.292
445329 90.451 -63.825 55.303
445330 90.495 -63.842 55.311
445331 90.539 -63.859 55.315
445332 90.583 -63.875 55.314
445333 90.627 -63.892 55.304
445334 90.672 -63.909 55.275
445335 90.716 -63.923 55.234
445336 90.762 -63.931 55.189
445337 90.808 -63.938 55.144
445338 90.855 -63.946 55.099
445339 90.901 -63.954 55.056
445340 90.947 -63.961 55.011
445341 90.993 -63.969 54.967
445342 91.039 -63.976 54.926
445343 91.085 -63.984 54.891
445344 91.132 -63.991 54.861
445345 91.178 -63.999 54.834
445346 91.224 -64.007 54.812
445347 91.270 -64.014 54.793
445348 91.316 -64.022 54.776
445349 91.362 -64.029 54.762
445350 91.409 -64.037 54.749
445351 91.454 -64.045 54.739
445352 91.501 -64.052 54.729
445353 91.547 -64.060 54.720
445354 91.593 -64.067 54.711
445355 91.639 -64.075 54.703
445356 91.685 -64.082 54.695
445357 91.731 -64.090 54.688
445358 91.778 -64.098 54.681
445359 91.823 -64.105 54.675
445360 91.870 -64.113 54.669
445361 91.916 -64.120 54.662
445362 91.962 -64.128 54.657
445363 92.007 -64.138 54.651
445364 92.053 -64.148 54.646
445365 92.098 -64.157 54.641
445366 92.143 -64.168 54.637
445367 92.188 -64.178 54.633
445368 92.234 -64.188 54.629
445369 92.283 -64.198 54.625
445370 92.329 -64.208 54.622
445371 92.369 -64.217 54.620
445372 92.406 -64.228 54.617
445373 92.431 -64.244 54.613
445374 92.456 -64.259 54.608
445375 92.482 -64.274 54.603
445376 92.511 -64.287 54.600
445377 92.540 -64.300 54.595
445378 92.569 -64.313 54.591
445379 92.598 -64.325 54.588
445380 92.634 -64.338 54.584
445381 92.683 -64.352 54.580
445382 92.732 -64.364 54.578
445383 92.782 -64.377 54.576
445384 92.856 -64.377 54.576
445385 92.933 -64.377 54.578
445386 93.009 -64.377 54.580
445387 93.086 -64.376 54.584
445388 93.141 -64.378 54.590
445389 93.190 -64.381 54.596
445390 93.237 -64.384 54.602
445391 93.283 -64.384 54.607
445392 93.329 -64.385 54.611
445393 93.370 -64.386 54.611
445394 93.395 -64.393 54.609
445395 93.420 -64.401 54.606
445396 93.444 -64.408 54.601
445397 93.472 -64.414 54.594
445398 93.512 -64.412 54.585
445399 93.553 -64.411 54.575
445400 93.593 -64.410 54.563
445401 93.634 -64.408 54.551
445402 93.674 -64.408 54.540
445403 93.715 -64.406 54.529
445404 93.754 -64.409 54.519
445405 93.790 -64.418 54.508
445406 93.827 -64.427 54.494
445407 93.867 -64.437 54.479
445408 93.913 -64.447 54.462
445409 93.958 -64.457 54.442
445410 94.003 -64.467 54.420
445411 94.047 -64.476 54.396
445412 94.076 -64.481 54.370
445413 94.105 -64.487 54.342
445414 94.134 -64.493 54.313
445415 94.163 -64.498 54.286
445416 94.192 -64.504 54.260
445417 94.221 -64.509 54.234
445418 94.253 -64.508 54.207
445419 94.286 -64.506 54.181
445420 94.319 -64.505 54.158
445421 94.355 -64.504 54.137
445422 94.417 -64.512 54.119
445423 94.480 -64.521 54.103
445424 94.535 -64.530 54.089
445425 94.586 -64.540 54.074
445426 94.637 -64.549 54.058
445427 94.688 -64.559 54.041
445428 94.739 -64.569 54.023
445429 94.790 -64.579 54.005
445430 94.837 -64.585 53.986
445431 94.883 -64.591 53.964
445432 94.929 -64.596 53.939
445433 94.975 -64.602 53.912
445434 95.021 -64.607 53.881
445435 95.067 -64.613 53.852
445436 95.113 -64.618 53.825
445437 95.162 -64.626 53.801
445438 95.210 -64.634 53.778
445439 95.259 -64.642 53.755
445440 95.308 -64.649 53.734
445441 95.357 -64.658 53.711
445442 95.406 -64.665 53.686
445443 95.455 -64.674 53.657
445444 95.503 -64.681 53.627
445445 95.549 -64.689 53.596
445446 95.589 -64.696 53.565
445447 95.629 -64.703 53.536
445448 95.670 -64.710 53.508
445449 95.710 -64.718 53.481
445450 95.750 -64.725 53.457
445451 95.790 -64.732 53.434
445452 95.826 -64.737 53.410
445453 95.860 -64.740 53.387
445454 95.894 -64.743 53.362
445455 95.928 -64.746 53.332
445456 95.962 -64.749 53.299
445457 96.001 -64.743 53.269
445458 96.041 -64.734 53.243
445459 96.081 -64.725 53.219
445460 96.121 -64.716 53.193
445461 96.161 -64.706 53.162
445462 96.203 -64.696 53.128
445463 96.243 -64.695 53.091
445464 96.282 -64.693 53.049
445465 96.321 -64.690 53.008
445466 96.360 -64.688 52.966
445467 96.399 -64.686 52.922
445468 96.439 -64.684 52.881
445469 96.478 -64.681 52.844
445470 96.517 -64.679 52.814
445471 96.555 -64.678 52.784
445472 96.590 -64.679 52.754
445473 96.629 -64.684 52.721
445474 96.668 -64.688 52.684
445475 96.708 -64.693 52.643
445476 96.747 -64.697 52.600
445477 96.786 -64.702 52.554
445478 96.824 -64.706 52.507
445479 96.862 -64.709 52.460
445480 96.900 -64.713 52.414
445481 96.938 -64.717 52.369
445482 96.976 -64.721 52.326
445483 97.014 -64.725 52.284
445484 97.052 -64.729 52.238
445485 97.090 -64.733 52.193
445486 97.128 -64.737 52.148
445487 97.166 -64.740 52.102
445488 97.204 -64.740 52.056
445489 97.242 -64.740 52.012
445490 97.281 -64.738 51.969
445491 97.319 -64.737 51.925
445492 97.357 -64.736 51.880
445493 97.382 -64.725 51.834
445494 97.405 -64.714 51.788
445495 97.435 -64.710 51.743
445496 97.470 -64.715 51.699
445497 97.506 -64.721 51.654
445498 97.541 -64.726 51.608
445499 97.577 -64.731 51.561
445500 97.610 -64.737 51.515
445501 97.642 -64.741 51.469
445502 97.673 -64.745 51.423
445503 97.704 -64.750 51.376
445504 97.736 -64.754 51.328
445505 97.764 -64.758 51.281
445506 97.791 -64.762 51.234
445507 97.818 -64.765 51.188
445508 97.848 -64.769 51.145
445509 97.884 -64.771 51.104
445510 97.921 -64.772 51.064
445511 97.957 -64.774 51.024
445512 97.994 -64.775 50.985
445513 98.030 -64.777 50.947
445514 98.066 -64.778 50.907
445515 98.103 -64.780 50.866
445516 98.140 -64.784 50.823
445517 98.176 -64.789 50.780
445518 98.212 -64.794 50.736
445519 98.247 -64.798 50.691
445520 98.283 -64.803 50.646
445521 98.319 -64.808 50.601
445522 98.355 -64.813 50.559
445523 98.391 -64.819 50.516
445524 98.433 -64.817 50.472
445525 98.473 -64.818 50.427
445526 98.513 -64.819 50.380
445527 98.553 -64.819 50.330
445528 98.593 -64.820 50.280
445529 98.633 -64.821 50.229
445530 98.672 -64.822 50.177
445531 98.712 -64.822 50.129
445532 98.752 -64.823 50.084
445533 98.792 -64.824 50.039
445534 98.832 -64.825 49.997
445535 98.872 -64.825 49.957
445536 98.912 -64.826 49.918
445537 98.955 -64.822 49.879
445538 98.998 -64.818 49.839
445539 99.041 -64.813 49.799
445540 99.084 -64.809 49.758
445541 99.132 -64.807 49.719
445542 99.186 -64.808 49.681
445543 99.236 -64.812 49.642
445544 99.284 -64.821 49.600
445545 99.332 -64.828 49.556
445546 99.379 -64.837 49.510
445547 99.427 -64.844 49.464
445548 99.475 -64.853 49.419
445549 99.533 -64.860 49.375
445550 99.590 -64.869 49.332
445551 99.648 -64.876 49.290
445552 99.727 -64.864 49.249
445553 99.822 -64.838 49.206
445554 99.916 -64.812 49.164
445555 100.009 -64.787 49.121
445556 100.104 -64.761 49.079
445557 100.195 -64.736 49.039
445558 100.252 -64.737 48.999
445559 100.243 -64.801 48.959
445560 100.234 -64.866 48.919
445561 100.257 -64.860 48.877
445562 100.305 -64.858 48.839
445563 100.348 -64.860 48.799
445564 100.386 -64.865 48.759
445565 100.425 -64.870 48.720
445566 100.464 -64.875 48.681
445567 100.502 -64.881 48.640
445568 100.541 -64.886 48.601
445569 100.580 -64.891 48.563
445570 100.619 -64.897 48.525
445956 87.108 -63.245 55.859
445957 87.142 -63.255 55.845
445958 87.177 -63.264 55.831
445959 87.211 -63.274 55.817
445960 87.245 -63.283 55.803
445961 87.279 -63.293 55.788
445962 87.314 -63.302 55.773
445963 87.348 -63.312 55.758
445964 87.382 -63.321 55.744
445965 87.416 -63.331 55.728
445966 87.450 -63.341 55.713
445967 87.485 -63.350 55.698
445968 87.519 -63.360 55.683
445969 87.555 -63.368 55.666
445970 87.602 -63.372 55.650
445971 87.648 -63.377 55.635
445972 87.697 -63.379 55.618
445973 87.760 -63.374 55.603
445974 87.819 -63.375 55.587
445975 87.867 -63.386 55.571
445976 87.916 -63.396 55.553
445977 87.965 -63.406 55.535
445978 88.014 -63.417 55.517
445979 88.063 -63.427 55.498
445980 88.112 -63.437 55.480
445981 88.161 -63.448 55.464
445993 88.749 -63.572 55.334
445994 88.798 -63.583 55.333
445995 88.847 -63.593 55.332
445996 88.896 -63.603 55.333
445997 88.945 -63.614 55.334
445998 88.993 -63.625 55.338
445999 89.042 -63.634 55.342
446000 89.092 -63.645 55.349
446001 89.140 -63.656 55.355
446002 89.189 -63.665 55.365
446003 89.238 -63.676 55.377
446004 89.287 -63.687 55.391
446005 89.334 -63.698 55.409
446006 89.370 -63.715 55.425
446007 89.407 -63.731 55.437
446008 89.443 -63.748 55.444
446009 89.487 -63.756 55.447
446010 89.531 -63.763 55.449
446011 89.576 -63.770 55.448
446012 89.621 -63.777 55.440
446013 89.666 -63.784 55.424
446014 89.712 -63.781 55.402
446015 89.758 -63.778 55.378
446016 89.805 -63.775 55.351
446017 89.851 -63.772 55.325
446018 89.898 -63.769 55.302
446019 89.945 -63.766 55.283
446020 89.991 -63.763 55.269
446021 90.038 -63.759 55.258
446022 90.084 -63.756 55.251
446023 90.130 -63.753 55.248
446024 90.177 -63.751 55.248
446025 90.224 -63.748 55.253
446026 90.271 -63.745 55.264
446027 90.321 -63.744 55.278
446028 90.369 -63.747 55.292
446029 90.413 -63.764 55.304
446030 90.457 -63.781 55.313
446031 90.502 -63.797 55.317
446032 90.545 -63.814 55.317
446033 90.590 -63.831 55.312
446034 90.634 -63.847 55.299
446035 90.678 -63.864 55.269
446036 90.723 -63.879 55.227
446037 90.768 -63.887 55.181
446038 90.815 -63.894 55.134
446039 90.861 -63.902 55.088
446040 90.907 -63.910 55.045
446041 90.953 -63.917 55.001
446042 91.000 -63.925 54.958
446043 91.045 -63.932 54.920
446044 91.092 -63.940 54.886
446045 91.138 -63.947 54.856
446046 91.184 -63.955 54.829
446047 91.230 -63.963 54.807
446048 91.276 -63.970 54.788
446049 91.322 -63.978 54.772
446050 91.369 -63.985 54.758
446051 91.415 -63.993 54.746
446052 91.461 -64.001 54.735
446053 91.507 -64.008 54.726
446054 91.553 -64.016 54.717
446055 91.599 -64.023 54.708
446056 91.645 -64.031 54.700
446057 91.692 -64.038 54.693
446058 91.738 -64.046 54.687
446059 91.784 -64.054 54.680
446060 91.830 -64.061 54.674
446061 91.876 -64.069 54.668
446062 91.922 -64.076 54.662
446063 91.968 -64.084 54.657
446064 92.015 -64.092 54.651
446065 92.061 -64.099 54.646
446066 92.106 -64.109 54.642
446067 92.151 -64.119 54.637
446068 92.197 -64.129 54.633
446069 92.243 -64.139 54.628
446070 92.289 -64.148 54.624
446071 92.330 -64.158 54.621
446072 92.370 -64.168 54.619
446073 92.406 -64.179 54.615
446074 92.431 -64.195 54.610
446075 92.456 -64.211 54.605
446076 92.481 -64.226 54.600
446077 92.506 -64.242 54.596
446078 92.534 -64.255 54.591
446079 92.563 -64.267 54.587
446080 92.592 -64.280 54.583
446081 92.626 -64.293 54.581
446082 92.675 -64.306 54.577
446083 92.725 -64.319 54.576
446084 92.795 -64.322 54.575
446085 92.872 -64.321 54.575
446086 92.948 -64.320 54.576
446087 93.025 -64.320 54.578
446088 93.099 -64.320 54.581
446089 93.147 -64.323 54.586
446090 93.196 -64.327 54.591
446091 93.241 -64.327 54.596
446092 93.287 -64.327 54.600
446093 93.334 -64.327 54.602
446094 93.368 -64.330 54.601
446095 93.393 -64.337 54.598
446096 93.418 -64.345 54.593
446097 93.443 -64.352 54.588
446098 93.468 -64.358 54.581
446099 93.503 -64.361 54.572
446100 93.543 -64.359 54.561
446101 93.584 -64.358 54.548
446102 93.625 -64.357 54.535
446103 93.665 -64.355 54.525
446104 93.706 -64.355 54.515
446105 93.743 -64.362 54.505
446106 93.779 -64.371 54.494
446107 93.819 -64.380 54.478
446108 93.864 -64.390 54.461
446109 93.909 -64.400 54.442
446110 93.954 -64.410 54.421
446111 93.999 -64.420 54.399
446112 94.040 -64.428 54.373
446113 94.068 -64.433 54.347
446114 94.097 -64.439 54.319
446115 94.126 -64.445 54.292
446116 94.155 -64.450 54.264
446117 94.184 -64.455 54.238
446118 94.213 -64.461 54.212
446119 94.243 -64.464 54.187
446120 94.276 -64.462 54.162
446121 94.309 -64.461 54.140
446122 94.348 -64.461 54.119
446123 94.411 -64.470 54.100
446124 94.471 -64.478 54.082
446125 94.522 -64.488 54.067
446126 94.573 -64.498 54.051
446127 94.624 -64.508 54.037
446128 94.675 -64.518 54.021
446129 94.726 -64.527 54.003
446130 94.777 -64.537 53.985
446131 94.825 -64.544 53.966
446132 94.871 -64.549 53.945
446133 94.917 -64.555 53.922
446134 94.963 -64.561 53.894
446135 95.009 -64.566 53.865
446136 95.055 -64.571 53.835
446137 95.108 -64.578 53.807
446138 95.162 -64.586 53.781
446139 95.216 -64.593 53.757
446140 95.265 -64.600 53.734
446141 95.314 -64.609 53.711
446142 95.363 -64.616 53.688
446143 95.412 -64.624 53.664
446144 95.461 -64.632 53.637
446145 95.510 -64.640 53.608
446146 95.558 -64.648 53.577
446147 95.600 -64.656 53.548
446148 95.640 -64.662 53.520
446149 95.680 -64.670 53.492
446150 95.721 -64.677 53.466
446151 95.761 -64.684 53.438
446152 95.800 -64.690 53.412
446153 95.834 -64.693 53.386
446154 95.868 -64.696 53.360
446155 95.902 -64.700 53.335
446156 95.936 -64.703 53.306
446157 95.970 -64.706 53.275
446158 96.004 -64.709 53.243
446159 96.043 -64.702 53.215
446160 96.083 -64.693 53.189
446161 96.123 -64.684 53.163
446162 96.164 -64.674 53.134
446163 96.208 -64.660 53.100
446164 96.249 -64.653 53.066
446165 96.288 -64.652 53.028
446166 96.327 -64.649 52.989
446167 96.367 -64.647 52.948
446168 96.406 -64.645 52.906
446169 96.445 -64.643 52.864
446170 96.484 -64.640 52.827
446171 96.523 -64.638 52.797
446172 96.558 -64.639 52.768
446173 96.594 -64.640 52.737
446174 96.632 -64.644 52.704
446175 96.672 -64.649 52.667
446176 96.711 -64.653 52.627
446177 96.751 -64.658 52.584
446178 96.790 -64.662 52.538
446179 96.829 -64.667 52.492
446180 96.869 -64.671 52.447
446181 96.907 -64.675 52.402
446182 96.945 -64.679 52.359
446183 96.983 -64.683 52.318
446184 97.021 -64.687 52.276
446185 97.059 -64.691 52.230
446186 97.097 -64.695 52.184
446187 97.135 -64.699 52.140
446188 97.173 -64.703 52.094
446189 97.211 -64.704 52.048
446190 97.249 -64.703 52.005
446191 97.288 -64.702 51.963
446192 97.326 -64.700 51.920
446193 97.356 -64.694 51.877
446194 97.380 -64.682 51.829
446195 97.403 -64.671 51.783
446196 97.433 -64.668 51.738
446197 97.468 -64.674 51.694
446198 97.504 -64.679 51.649
446199 97.539 -64.684 51.603
446200 97.575 -64.690 51.556
446201 97.610 -64.695 51.509
446202 97.645 -64.700 51.462
446203 97.676 -64.705 51.414
446204 97.708 -64.709 51.367
446205 97.739 -64.713 51.320
446206 97.768 -64.717 51.272
446207 97.795 -64.722 51.224
446208 97.823 -64.725 51.177
446209 97.857 -64.728 51.132
446210 97.893 -64.729 51.089
446211 97.929 -64.730 51.047
446212 97.965 -64.731 51.005
446213 98.002 -64.733 50.964
446214 98.038 -64.734 50.925
446215 98.075 -64.736 50.885
446216 98.112 -64.736 50.843
446217 98.146 -64.744 50.800
446218 98.182 -64.750 50.756
446219 98.218 -64.756 50.711
446220 98.254 -64.762 50.666
446221 98.290 -64.768 50.621
446222 98.326 -64.773 50.577
446223 98.362 -64.779 50.534
446224 98.398 -64.784 50.492
446225 98.445 -64.777 50.450
446226 98.494 -64.768 50.405
446227 98.533 -64.769 50.358
446228 98.574 -64.769 50.309
446229 98.614 -64.770 50.258
446230 98.653 -64.771 50.207
446231 98.693 -64.772 50.159
446232 98.733 -64.772 50.113
446233 98.773 -64.773 50.068
446234 98.812 -64.774 50.025
446235 98.852 -64.775 49.984
446236 98.892 -64.775 49.944
446237 98.933 -64.775 49.904
446238 98.976 -64.771 49.865
446239 99.019 -64.766 49.824
446240 99.062 -64.762 49.783
446241 99.105 -64.758 49.743
446242 99.148 -64.754 49.705
446243 99.190 -64.756 49.666
446244 99.238 -64.764 49.625
446245 99.285 -64.772 49.582
446246 99.333 -64.780 49.537
446247 99.381 -64.788 49.491
446248 99.429 -64.796 49.444
446249 99.477 -64.804 49.399
446250 99.532 -64.812 49.356
446251 99.590 -64.819 49.313
446252 99.648 -64.828 49.272
446253 99.737 -64.806 49.232
446254 99.831 -64.781 49.191
446255 99.925 -64.755 49.149
446256 100.019 -64.729 49.107
446257 100.113 -64.703 49.067
446258 100.207 -64.678 49.027
446259 100.268 -64.681 48.988
446260 100.259 -64.745 48.949
446261 100.250 -64.810 48.909
446262 100.271 -64.814 48.869
446263 100.319 -64.812 48.830
446264 100.357 -64.817 48.792
446265 100.396 -64.822 48.752
446266 100.435 -64.828 48.713
446267 100.474 -64.833 48.673
446268 100.512 -64.838 48.633
446269 100.551 -64.844 48.594
446270 100.590 -64.848 48.556
446271 100.628 -64.853 48.518
446656 87.066 -63.187 55.877
446657 87.100 -63.197 55.864
446658 87.134 -63.206 55.850
446659 87.168 -63.216 55.836
446660 87.203 -63.225 55.822
446661 87.237 -63.235 55.809
446662 87.271 -63.245 55.795
446663 87.305 -63.254 55.781
446664 87.340 -63.264 55.765
446665 87.374 -63.274 55.751
446666 87.408 -63.283 55.737
446667 87.442 -63.292 55.723
446668 87.477 -63.302 55.707
446669 87.511 -63.311 55.693
446670 87.558 -63.315 55.677
446671 87.604 -63.319 55.662
446672 87.660 -63.318 55.647
446673 87.722 -63.315 55.630
446674 87.771 -63.325 55.615
446675 87.820 -63.336 55.600
446676 87.869 -63.346 55.584
446677 87.918 -63.356 55.566
446678 87.967 -63.367 55.547
446679 88.016 -63.377 55.529
446680 88.065 -63.387 55.511
446681 88.114 -63.398 55.492
446682 88.163 -63.408 55.474
446683 88.212 -63.418 55.458
446684 88.261 -63.429 55.441
446694 88.750 -63.533 55.337
446695 88.799 -63.543 55.335
446696 88.848 -63.553 55.334
446697 88.897 -63.564 55.335
446698 88.946 -63.574 55.336
446699 88.995 -63.585 55.340
446700 89.044 -63.595 55.344
446701 89.093 -63.606 55.350
446702 89.142 -63.616 55.356
446703 89.191 -63.626 55.365
446704 89.240 -63.637 55.377
446705 89.289 -63.647 55.390
446706 89.338 -63.657 55.405
446707 89.376 -63.673 55.420
446708 89.412 -63.690 55.432
446709 89.454 -63.700 55.439
446710 89.499 -63.707 55.444
446711 89.544 -63.714 55.448
446712 89.589 -63.721 55.449
446713 89.633 -63.728 55.443
446714 89.678 -63.732 55.430
446715 89.725 -63.729 55.409
446716 89.771 -63.726 55.384
446717 89.818 -63.723 55.356
446718 89.865 -63.720 55.331
446719 89.911 -63.717 55.309
446720 89.957 -63.714 55.291
446721 90.004 -63.711 55.277
446722 90.050 -63.708 55.267
446723 90.097 -63.705 55.261
446724 90.144 -63.702 55.259
446725 90.190 -63.699 55.259
446726 90.237 -63.696 55.266
446727 90.283 -63.693 55.277
446728 90.330 -63.690 55.290
446729 90.375 -63.703 55.303
446730 90.419 -63.720 55.314
446731 90.464 -63.737 55.323
446732 90.508 -63.753 55.325
446733 90.552 -63.770 55.322
446734 90.596 -63.787 55.314
446735 90.640 -63.803 55.296
446736 90.684 -63.820 55.264
446737 90.729 -63.835 55.223
446738 90.775 -63.843 55.177
446739 90.821 -63.850 55.128
446740 90.867 -63.858 55.082
446741 90.913 -63.866 55.037
446742 90.959 -63.873 54.994
446743 91.006 -63.881 54.953
446744 91.052 -63.888 54.914
446745 91.098 -63.896 54.880
446746 91.144 -63.904 54.851
446747 91.190 -63.911 54.826
446748 91.236 -63.919 54.804
446749 91.283 -63.926 54.784
446750 91.329 -63.934 54.768
446751 91.375 -63.941 54.754
446752 91.421 -63.949 54.743
446753 91.467 -63.957 54.732
446754 91.513 -63.964 54.722
446755 91.560 -63.972 54.714
446756 91.605 -63.979 54.706
446757 91.652 -63.987 54.699
446758 91.698 -63.994 54.692
446759 91.744 -64.002 54.686
446760 91.790 -64.010 54.680
446761 91.837 -64.017 54.674
446762 91.882 -64.025 54.668
446763 91.929 -64.032 54.663
446764 91.975 -64.040 54.658
446765 92.021 -64.048 54.651
446766 92.067 -64.055 54.647
446767 92.113 -64.063 54.642
446768 92.159 -64.070 54.638
446769 92.205 -64.080 54.633
446770 92.250 -64.090 54.628
446771 92.290 -64.100 54.624
446772 92.331 -64.110 54.621
446773 92.371 -64.119 54.617
446774 92.406 -64.131 54.613
446775 92.431 -64.146 54.608
446776 92.456 -64.162 54.603
446777 92.481 -64.177 54.599
446778 92.506 -64.192 54.594
446779 92.531 -64.208 54.590
446780 92.558 -64.223 54.586
446781 92.587 -64.236 54.582
446782 92.619 -64.248 54.579
446783 92.668 -64.261 54.576
446784 92.734 -64.266 54.575
446785 92.811 -64.265 54.574
446786 92.887 -64.264 54.574
446787 92.965 -64.264 54.573
446788 93.041 -64.264 54.575
446789 93.106 -64.264 54.577
446790 93.154 -64.268 54.580
446791 93.200 -64.269 54.584
446792 93.246 -64.269 54.587
446793 93.292 -64.269 54.588
446794 93.338 -64.269 54.589
446795 93.367 -64.274 54.587
446796 93.392 -64.282 54.584
446797 93.416 -64.289 54.580
446798 93.441 -64.295 54.575
446799 93.466 -64.303 54.567
446800 93.494 -64.308 54.557
446801 93.534 -64.308 54.544
446802 93.575 -64.306 54.531
446803 93.615 -64.305 54.519
446804 93.656 -64.304 54.508
446805 93.695 -64.305 54.499
446806 93.731 -64.314 54.489
446807 93.770 -64.324 54.476
446808 93.815 -64.334 54.459
446809 93.860 -64.344 54.441
446810 93.905 -64.353 54.420
446811 93.951 -64.363 54.398
446812 93.996 -64.373 54.376
446813 94.032 -64.380 54.350
446814 94.060 -64.386 54.323
446815 94.089 -64.391 54.297
446816 94.118 -64.397 54.270
446817 94.147 -64.402 54.243
446818 94.176 -64.408 54.217
446819 94.205 -64.413 54.191
446820 94.234 -64.419 54.167
446821 94.266 -64.418 54.144
446822 94.298 -64.417 54.122
446823 94.342 -64.418 54.101
446824 94.404 -64.427 54.081
446825 94.459 -64.436 54.062
446826 94.509 -64.446 54.045
446827 94.560 -64.455 54.028
446828 94.612 -64.465 54.013
446829 94.663 -64.475 53.997
446830 94.713 -64.485 53.979
446831 94.765 -64.495 53.960
446832 94.813 -64.502 53.940
446833 94.859 -64.508 53.918
446834 94.905 -64.514 53.894
446835 94.951 -64.519 53.869
446836 94.997 -64.524 53.841
446837 95.050 -64.531 53.815
446838 95.104 -64.539 53.788
446839 95.158 -64.546 53.763
446840 95.211 -64.553 53.739
446841 95.266 -64.560 53.715
446842 95.320 -64.567 53.691
446843 95.369 -64.575 53.667
446844 95.417 -64.583 53.642
446845 95.466 -64.591 53.616
446846 95.515 -64.599 53.590
446847 95.564 -64.607 53.561
446848 95.611 -64.615 53.534
446849 95.651 -64.621 53.507
446850 95.691 -64.629 53.482
446851 95.732 -64.636 53.456
446852 95.772 -64.643 53.428
446853 95.808 -64.647 53.399
446854 95.841 -64.650 53.369
446855 95.876 -64.654 53.338
446856 95.909 -64.657 53.310
446857 95.944 -64.660 53.281
446858 95.978 -64.663 53.251
446859 96.012 -64.666 53.218
446860 96.046 -64.669 53.185
446861 96.086 -64.660 53.156
446862 96.126 -64.651 53.131
446863 96.166 -64.642 53.103
446864 96.210 -64.627 53.071
446865 96.255 -64.613 53.037
446866 96.295 -64.611 53.004
446867 96.334 -64.609 52.968
446868 96.373 -64.606 52.929
446869 96.412 -64.604 52.889
446870 96.451 -64.602 52.848
446871 96.490 -64.599 52.812
446872 96.528 -64.599 52.782
446873 96.562 -64.599 52.752
446874 96.597 -64.600 52.720
446875 96.636 -64.604 52.687
446876 96.675 -64.609 52.649
446877 96.715 -64.613 52.610
446878 96.754 -64.618 52.569
446879 96.794 -64.622 52.524
446880 96.833 -64.627 52.480
446881 96.872 -64.631 52.437
446882 96.912 -64.636 52.393
446883 96.951 -64.641 52.351
446884 96.990 -64.645 52.310
446885 97.028 -64.649 52.267
446886 97.066 -64.653 52.222
446887 97.104 -64.657 52.178
446888 97.142 -64.661 52.134
446889 97.180 -64.665 52.089
446890 97.218 -64.668 52.044
446891 97.256 -64.667 52.001
446892 97.294 -64.665 51.959
446893 97.330 -64.662 51.917
446894 97.354 -64.651 51.873
446895 97.377 -64.640 51.824
446896 97.401 -64.628 51.777
446897 97.431 -64.626 51.732
446898 97.466 -64.631 51.688
446899 97.502 -64.637 51.642
446900 97.537 -64.642 51.596
446901 97.573 -64.647 51.549
446902 97.608 -64.653 51.502
446903 97.643 -64.658 51.455
446904 97.679 -64.663 51.407
446905 97.711 -64.668 51.359
446906 97.742 -64.672 51.312
446907 97.772 -64.677 51.264
446908 97.800 -64.681 51.215
446909 97.829 -64.684 51.167
446910 97.865 -64.685 51.121
446911 97.901 -64.687 51.076
446912 97.938 -64.688 51.032
446913 97.974 -64.690 50.988
446914 98.011 -64.691 50.946
446915 98.052 -64.688 50.904
446916 98.088 -64.688 50.863
446917 98.114 -64.695 50.821
446918 98.145 -64.703 50.778
446919 98.185 -64.713 50.733
446920 98.225 -64.722 50.687
446921 98.261 -64.728 50.642
446922 98.297 -64.733 50.598
446923 98.333 -64.739 50.553
446924 98.368 -64.744 50.510
446925 98.405 -64.750 50.468
446926 98.452 -64.742 50.425
446927 98.501 -64.733 50.381
446928 98.550 -64.724 50.335
446929 98.594 -64.719 50.287
446930 98.634 -64.720 50.237
446931 98.674 -64.721 50.188
446932 98.714 -64.722 50.141
446933 98.754 -64.722 50.097
446934 98.793 -64.723 50.054
446935 98.833 -64.724 50.013
446936 98.873 -64.725 49.972
446937 98.913 -64.725 49.931
446938 98.954 -64.723 49.890
446939 98.997 -64.719 49.850
446940 99.041 -64.715 49.809
446941 99.083 -64.711 49.768
446942 99.127 -64.707 49.728
446943 99.165 -64.709 49.690
446944 99.203 -64.712 49.650
446945 99.240 -64.715 49.606
446946 99.287 -64.723 49.562
446947 99.335 -64.731 49.516
446948 99.383 -64.739 49.471
446949 99.430 -64.747 49.425
446950 99.478 -64.755 49.381
446951 99.532 -64.763 49.338
446952 99.589 -64.771 49.297
446953 99.652 -64.774 49.255
446954 99.747 -64.748 49.215
446955 99.840 -64.722 49.175
446956 99.934 -64.696 49.134
446957 100.028 -64.671 49.093
446958 100.122 -64.645 49.055
446959 100.191 -64.643 49.016
446960 100.246 -64.656 48.977
446961 100.275 -64.689 48.939
446962 100.266 -64.753 48.899
446963 100.285 -64.768 48.860
446964 100.329 -64.769 48.821
446965 100.367 -64.775 48.783
446966 100.406 -64.780 48.745
446967 100.445 -64.785 48.706
446968 100.484 -64.790 48.666
446969 100.522 -64.796 48.626
446970 100.561 -64.800 48.588
446971 100.600 -64.806 48.550
446972 100.638 -64.811 48.512
447357 87.070 -63.145 55.884
447358 87.098 -63.151 55.870
447359 87.127 -63.158 55.855
447360 87.160 -63.167 55.842
447361 87.195 -63.177 55.828
447362 87.229 -63.186 55.814
447363 87.263 -63.196 55.801
447364 87.297 -63.205 55.788
447365 87.332 -63.215 55.773
447366 87.366 -63.225 55.759
447367 87.400 -63.234 55.745
447368 87.434 -63.244 55.732
447369 87.469 -63.254 55.717
447370 87.514 -63.258 55.702
447371 87.560 -63.262 55.687
447372 87.623 -63.258 55.673
447373 87.675 -63.265 55.658
447374 87.723 -63.275 55.643
447375 87.773 -63.286 55.627
447376 87.822 -63.296 55.612
447377 87.870 -63.307 55.595
447378 87.919 -63.317 55.579
447379 87.969 -63.327 55.561
447380 88.017 -63.338 55.542
447381 88.066 -63.348 55.523
447382 88.115 -63.358 55.504
447383 88.165 -63.369 55.485
447384 88.213 -63.379 55.467
447385 88.262 -63.390 55.450
447386 88.311 -63.400 55.434
447394 88.703 -63.483 55.344
447395 88.752 -63.493 55.341
447396 88.801 -63.504 55.339
447397 88.850 -63.514 55.337
447398 88.899 -63.524 55.338
447399 88.948 -63.535 55.339
447400 88.996 -63.545 55.342
447401 89.046 -63.556 55.346
447402 89.095 -63.566 55.351
447403 89.143 -63.577 55.358
447404 89.192 -63.587 55.366
447405 89.242 -63.597 55.376
447406 89.291 -63.608 55.388
447407 89.339 -63.618 55.402
447408 89.382 -63.631 55.415
447409 89.422 -63.644 55.425
447410 89.467 -63.651 55.433
447411 89.512 -63.658 55.440
447412 89.556 -63.665 55.445
447413 89.601 -63.672 55.448
447414 89.646 -63.678 55.444
447415 89.691 -63.680 55.433
447416 89.738 -63.677 55.414
447417 89.784 -63.674 55.388
447418 89.831 -63.671 55.361
447419 89.877 -63.669 55.337
447420 89.923 -63.665 55.317
447421 89.970 -63.662 55.298
447422 90.017 -63.659 55.285
447423 90.063 -63.656 55.276
447424 90.110 -63.653 55.271
447425 90.156 -63.650 55.269
447426 90.203 -63.647 55.271
447427 90.249 -63.644 55.282
447428 90.295 -63.646 55.292
447429 90.335 -63.650 55.302
447430 90.378 -63.662 55.314
447431 90.426 -63.676 55.321
447432 90.470 -63.693 55.328
447433 90.514 -63.709 55.332
447434 90.558 -63.726 55.331
447435 90.602 -63.743 55.321
447436 90.647 -63.759 55.297
447437 90.691 -63.776 55.261
447438 90.735 -63.792 55.221
447439 90.781 -63.800 55.177
447440 90.827 -63.807 55.128
447441 90.874 -63.815 55.079
447442 90.920 -63.822 55.033
447443 90.966 -63.830 54.989
447444 91.012 -63.838 54.948
447445 91.058 -63.845 54.909
447446 91.104 -63.853 54.875
447447 91.150 -63.860 54.846
447448 91.196 -63.868 54.821
447449 91.243 -63.875 54.799
447450 91.289 -63.883 54.779
447451 91.335 -63.891 54.764
447452 91.381 -63.898 54.751
447453 91.427 -63.906 54.739
447454 91.473 -63.913 54.729
447455 91.520 -63.921 54.719
447456 91.566 -63.929 54.712
447457 91.612 -63.936 54.704
447458 91.658 -63.944 54.697
447459 91.704 -63.951 54.691
447460 91.750 -63.959 54.685
447461 91.796 -63.966 54.679
447462 91.843 -63.974 54.674
447463 91.889 -63.982 54.668
447464 91.935 -63.989 54.663
447465 91.981 -63.997 54.658
447466 92.027 -64.004 54.653
447467 92.074 -64.013 54.648
447468 92.121 -64.021 54.643
447469 92.168 -64.030 54.638
447470 92.214 -64.034 54.633
447471 92.251 -64.041 54.628
447472 92.291 -64.051 54.624
447473 92.332 -64.060 54.620
447474 92.372 -64.070 54.616
447475 92.406 -64.082 54.611
447476 92.431 -64.098 54.606
447477 92.456 -64.114 54.602
447478 92.481 -64.129 54.598
447479 92.506 -64.144 54.593
447480 92.531 -64.160 54.590
447481 92.557 -64.175 54.587
447482 92.582 -64.190 54.583
447483 92.612 -64.203 54.580
447484 92.673 -64.211 54.577
447485 92.749 -64.210 54.576
447486 92.827 -64.209 54.575
447487 92.903 -64.208 54.575
447488 92.980 -64.208 54.575
447489 93.057 -64.207 54.576
447490 93.112 -64.209 54.578
447491 93.158 -64.211 54.580
447492 93.204 -64.211 54.581
447493 93.250 -64.211 54.583
447494 93.296 -64.211 54.583
447495 93.340 -64.211 54.581
447496 93.365 -64.219 54.577
447497 93.390 -64.226 54.573
447498 93.415 -64.233 54.569
447499 93.440 -64.240 54.563
447500 93.465 -64.247 54.555
447501 93.490 -64.254 54.543
447502 93.525 -64.255 54.530
447503 93.565 -64.255 54.515
447504 93.606 -64.253 54.502
447505 93.646 -64.251 54.490
447506 93.684 -64.258 54.479
447507 93.721 -64.267 54.466
447508 93.766 -64.277 54.452
447509 93.812 -64.287 54.435
447510 93.856 -64.297 54.416
447511 93.902 -64.307 54.395
447512 93.947 -64.317 54.373
447513 93.992 -64.326 54.350
447514 94.024 -64.333 54.325
447515 94.053 -64.338 54.300
447516 94.081 -64.343 54.274
447517 94.110 -64.349 54.247
447518 94.139 -64.355 54.222
447519 94.168 -64.360 54.196
447520 94.197 -64.365 54.171
447521 94.226 -64.371 54.147
447522 94.256 -64.374 54.125
447523 94.288 -64.372 54.104
447524 94.335 -64.376 54.084
447525 94.395 -64.384 54.063
447526 94.446 -64.394 54.042
447527 94.497 -64.404 54.023
447528 94.548 -64.414 54.006
447529 94.599 -64.424 53.989
447530 94.650 -64.433 53.972
447531 94.701 -64.443 53.955
447532 94.752 -64.453 53.936
447533 94.801 -64.461 53.914
447534 94.847 -64.467 53.892
447535 94.893 -64.472 53.868
447536 94.939 -64.478 53.844
447537 94.991 -64.484 53.819
447538 95.045 -64.492 53.795
447539 95.099 -64.499 53.771
447540 95.153 -64.506 53.747
447541 95.207 -64.513 53.724
447542 95.261 -64.520 53.699
447543 95.315 -64.527 53.674
447544 95.370 -64.534 53.650
447545 95.423 -64.542 53.625
447546 95.472 -64.549 53.600
447547 95.521 -64.558 53.575
447548 95.570 -64.565 53.549
447549 95.619 -64.574 53.524
447550 95.662 -64.580 53.500
447551 95.702 -64.588 53.476
447552 95.742 -64.595 53.450
447553 95.781 -64.602 53.421
447554 95.815 -64.605 53.390
447555 95.849 -64.608 53.358
447556 95.883 -64.611 53.324
447557 95.917 -64.614 53.293
447558 95.951 -64.617 53.262
447559 95.985 -64.620 53.230
447560 96.019 -64.623 53.198
447561 96.053 -64.626 53.164
447562 96.088 -64.627 53.134
447563 96.128 -64.618 53.106
447564 96.169 -64.609 53.076
447565 96.213 -64.596 53.044
447566 96.257 -64.581 53.012
447567 96.301 -64.570 52.980
447568 96.340 -64.568 52.945
447569 96.379 -64.565 52.910
447570 96.418 -64.563 52.874
447571 96.457 -64.561 52.838
447572 96.497 -64.559 52.805
447573 96.532 -64.559 52.774
447574 96.567 -64.560 52.742
447575 96.601 -64.562 52.708
447576 96.639 -64.565 52.673
447577 96.678 -64.569 52.636
447578 96.718 -64.574 52.598
447579 96.757 -64.578 52.559
447580 96.797 -64.583 52.519
447581 96.836 -64.587 52.476
447582 96.876 -64.592 52.433
447583 96.915 -64.596 52.389
447584 96.954 -64.601 52.345
447585 96.994 -64.606 52.302
447586 97.033 -64.610 52.259
447587 97.073 -64.615 52.216
447588 97.111 -64.618 52.174
447589 97.149 -64.623 52.131
447590 97.187 -64.627 52.087
447591 97.225 -64.631 52.042
447592 97.263 -64.631 51.999
447593 97.301 -64.630 51.957
447594 97.328 -64.620 51.913
447595 97.352 -64.609 51.868
447596 97.376 -64.597 51.818
447597 97.405 -64.587 51.771
447598 97.434 -64.590 51.725
447599 97.475 -64.586 51.680
447600 97.508 -64.590 51.635
447601 97.537 -64.599 51.588
447602 97.571 -64.606 51.541
447603 97.606 -64.611 51.494
447604 97.641 -64.616 51.447
447605 97.677 -64.621 51.400
447606 97.712 -64.627 51.352
447607 97.745 -64.631 51.304
447608 97.776 -64.636 51.256
447609 97.804 -64.640 51.208
447610 97.837 -64.642 51.159
447611 97.873 -64.643 51.111
447612 97.910 -64.645 51.064
447613 97.948 -64.645 51.019
447614 97.989 -64.642 50.974
447615 98.030 -64.639 50.930
447616 98.064 -64.640 50.886
447617 98.090 -64.647 50.842
447618 98.116 -64.653 50.799
447619 98.144 -64.660 50.756
447620 98.184 -64.671 50.711
447621 98.225 -64.681 50.665
447622 98.266 -64.691 50.620
447623 98.303 -64.699 50.575
447624 98.339 -64.705 50.531
447625 98.375 -64.710 50.487
447626 98.411 -64.715 50.444
447627 98.459 -64.707 50.400
447628 98.508 -64.698 50.357
447629 98.557 -64.689 50.312
447630 98.605 -64.680 50.265
447631 98.654 -64.671 50.217
447632 98.695 -64.670 50.170
447633 98.734 -64.671 50.125
447634 98.774 -64.671 50.082
447635 98.814 -64.672 50.041
447636 98.854 -64.673 50.000
447637 98.894 -64.674 49.958
447638 98.934 -64.674 49.916
447639 98.976 -64.672 49.876
447640 99.019 -64.668 49.834
447641 99.062 -64.664 49.794
447642 99.105 -64.659 49.753
447643 99.143 -64.662 49.713
447644 99.181 -64.666 49.674
447645 99.218 -64.670 49.632
447646 99.256 -64.673 49.587
447647 99.293 -64.677 49.541
447648 99.337 -64.683 49.496
447649 99.384 -64.690 49.451
447650 99.432 -64.699 49.407
447651 99.480 -64.706 49.364
447652 99.531 -64.715 49.322
447653 99.589 -64.722 49.281
447654 99.662 -64.716 49.239
447655 99.756 -64.690 49.200
447656 99.850 -64.665 49.160
447657 99.944 -64.639 49.120
447658 100.035 -64.615 49.080
447659 100.090 -64.627 49.042
447660 100.145 -64.639 49.004
447661 100.200 -64.650 48.966
447662 100.255 -64.662 48.928
447663 100.282 -64.697 48.890
447664 100.300 -64.722 48.850
447665 100.339 -64.727 48.812
447666 100.378 -64.732 48.774
447667 100.416 -64.737 48.736
447668 100.455 -64.743 48.698
447669 100.494 -64.747 48.659
447670 100.533 -64.753 48.619
447671 100.571 -64.758 48.581
447672 100.610 -64.763 48.544
447673 100.648 -64.769 48.505
448057 87.046 -63.096 55.904
448058 87.074 -63.102 55.890
448059 87.103 -63.109 55.876
448060 87.130 -63.115 55.863
448061 87.158 -63.122 55.848
448062 87.187 -63.129 55.834
448063 87.221 -63.138 55.820
448064 87.256 -63.148 55.807
448065 87.290 -63.158 55.794
448066 87.324 -63.167 55.781
448067 87.358 -63.177 55.767
448068 87.393 -63.186 55.754
448069 87.427 -63.195 55.741
448070 87.470 -63.201 55.727
448071 87.522 -63.202 55.712
448072 87.578 -63.205 55.698
448073 87.627 -63.215 55.683
448074 87.676 -63.226 55.669
448075 87.725 -63.236 55.654
448076 87.774 -63.246 55.639
448077 87.823 -63.257 55.623
448078 87.872 -63.267 55.607
448079 87.921 -63.277 55.590
448080 87.970 -63.288 55.572
448081 88.019 -63.299 55.553
448082 88.068 -63.308 55.534
448083 88.117 -63.319 55.515
448084 88.166 -63.330 55.496
448085 88.215 -63.339 55.478
448086 88.264 -63.350 55.459
448087 88.313 -63.361 55.442
448088 88.362 -63.371 55.427
448095 88.704 -63.443 55.348
448096 88.753 -63.454 55.345
448097 88.803 -63.465 55.343
448098 88.851 -63.474 55.341
448099 88.900 -63.485 55.342
448100 88.949 -63.496 55.343
448101 88.998 -63.506 55.345
448102 89.047 -63.516 55.349
448103 89.096 -63.527 55.353
448104 89.145 -63.537 55.359
448105 89.194 -63.547 55.367
448106 89.243 -63.558 55.375
448107 89.292 -63.568 55.386
448108 89.341 -63.578 55.401
448109 89.390 -63.587 55.413
448110 89.435 -63.594 55.424
448111 89.479 -63.602 55.432
448112 89.524 -63.609 55.438
448113 89.568 -63.615 55.442
448114 89.613 -63.622 55.445
448115 89.658 -63.629 55.442
448116 89.704 -63.628 55.433
448117 89.750 -63.625 55.416
448118 89.797 -63.622 55.390
448119 89.843 -63.619 55.365
448120 89.890 -63.616 55.343
448121 89.937 -63.613 55.323
448122 89.983 -63.610 55.304
448123 90.029 -63.607 55.292
448124 90.076 -63.604 55.284
448125 90.122 -63.602 55.279
448126 90.168 -63.602 55.279
448127 90.214 -63.603 55.285
448128 90.259 -63.604 55.297
448129 90.305 -63.606 55.307
448130 90.334 -63.615 55.315
448131 90.373 -63.628 55.323
448132 90.423 -63.640 55.327
448133 90.472 -63.652 55.330
448134 90.521 -63.665 55.335
448135 90.565 -63.681 55.336
448136 90.609 -63.698 55.328
448137 90.653 -63.715 55.302
448138 90.697 -63.731 55.266
448139 90.741 -63.748 55.228
448140 90.788 -63.756 55.186
448141 90.834 -63.763 55.137
448142 90.880 -63.771 55.085
448143 90.926 -63.778 55.035
448144 90.972 -63.786 54.988
448145 91.018 -63.794 54.945
448146 91.065 -63.801 54.905
448147 91.110 -63.809 54.871
448148 91.157 -63.816 54.841
448149 91.203 -63.824 54.816
448150 91.249 -63.831 54.794
448151 91.295 -63.839 54.775
448152 91.342 -63.847 54.761
448153 91.387 -63.854 54.747
448154 91.434 -63.862 54.736
448155 91.480 -63.869 54.726
448156 91.526 -63.877 54.717
448157 91.572 -63.885 54.710
448158 91.618 -63.892 54.702
448159 91.664 -63.900 54.696
448160 91.711 -63.908 54.691
448161 91.758 -63.916 54.684
448162 91.805 -63.925 54.679
448163 91.853 -63.934 54.674
448164 91.899 -63.942 54.669
448165 91.947 -63.951 54.663
448166 91.994 -63.960 54.658
448167 92.041 -63.968 54.654
448168 92.088 -63.977 54.649
448169 92.135 -63.985 54.644
448170 92.178 -63.986 54.638
448171 92.223 -63.988 54.633
448172 92.254 -63.992 54.628
448173 92.292 -64.002 54.624
448174 92.333 -64.012 54.619
448175 92.373 -64.021 54.615
448176 92.406 -64.034 54.611
448177 92.431 -64.049 54.607
448178 92.456 -64.064 54.603
448179 92.481 -64.080 54.598
448180 92.506 -64.095 54.594
448181 92.531 -64.110 54.591
448182 92.557 -64.126 54.588
448183 92.582 -64.142 54.585
448184 92.621 -64.154 54.582
448185 92.689 -64.154 54.580
448186 92.766 -64.153 54.578
448187 92.842 -64.152 54.579
448188 92.919 -64.151 54.580
448189 92.996 -64.151 54.580
448190 93.070 -64.151 54.581
448191 93.117 -64.153 54.582
448192 93.163 -64.153 54.583
448193 93.209 -64.153 54.583
448194 93.255 -64.153 54.582
448195 93.300 -64.153 54.581
448196 93.339 -64.156 54.577
448197 93.363 -64.163 54.571
448198 93.389 -64.170 54.566
448199 93.413 -64.177 54.561
448200 93.438 -64.184 54.554
448201 93.463 -64.191 54.545
448202 93.488 -64.198 54.532
448203 93.516 -64.204 54.517
448204 93.556 -64.202 54.500
448205 93.597 -64.201 54.485
448206 93.637 -64.202 54.471
448207 93.673 -64.211 54.457
448208 93.717 -64.220 54.443
448209 93.763 -64.230 54.427
448210 93.807 -64.240 54.410
448211 93.853 -64.250 54.390
448212 93.898 -64.260 54.369
448213 93.943 -64.270 54.347
448214 93.987 -64.279 54.324
448215 94.015 -64.285 54.299
448216 94.045 -64.290 54.275
448217 94.073 -64.295 54.250
448218 94.102 -64.301 54.225
448219 94.131 -64.307 54.202
448220 94.160 -64.312 54.178
448221 94.189 -64.317 54.152
448222 94.218 -64.323 54.128
448223 94.247 -64.329 54.107
448224 94.279 -64.328 54.086
448225 94.329 -64.333 54.065
448226 94.382 -64.342 54.045
448227 94.433 -64.352 54.024
448228 94.484 -64.362 54.003
448229 94.535 -64.372 53.983
448230 94.586 -64.382 53.965
448231 94.637 -64.392 53.948
448232 94.688 -64.402 53.930
448233 94.739 -64.411 53.912
448234 94.789 -64.420 53.890
448235 94.835 -64.426 53.867
448236 94.882 -64.431 53.844
448237 94.932 -64.437 53.820
448238 94.987 -64.445 53.797
448239 95.041 -64.452 53.776
448240 95.095 -64.459 53.754
448241 95.149 -64.466 53.733
448242 95.203 -64.474 53.711
448243 95.257 -64.480 53.685
448244 95.311 -64.487 53.659
448245 95.365 -64.495 53.635
448246 95.419 -64.502 53.610
448247 95.473 -64.509 53.586
448248 95.527 -64.516 53.563
448249 95.576 -64.524 53.539
448250 95.624 -64.532 53.516
448251 95.673 -64.540 53.494
448252 95.713 -64.547 53.472
448253 95.753 -64.554 53.446
448254 95.789 -64.559 53.416
448255 95.823 -64.562 53.383
448256 95.857 -64.565 53.349
448257 95.891 -64.568 53.314
448258 95.925 -64.571 53.279
448259 95.959 -64.574 53.245
448260 95.993 -64.577 53.212
448261 96.027 -64.580 53.179
448262 96.061 -64.584 53.146
448263 96.095 -64.587 53.115
448264 96.131 -64.586 53.085
448265 96.171 -64.577 53.054
448266 96.215 -64.564 53.022
448267 96.259 -64.549 52.990
448268 96.305 -64.534 52.960
448269 96.346 -64.527 52.928
448270 96.385 -64.524 52.896
448271 96.425 -64.522 52.866
448272 96.464 -64.520 52.835
448273 96.501 -64.519 52.804
448274 96.535 -64.520 52.773
448275 96.571 -64.521 52.741
448276 96.605 -64.522 52.707
448277 96.643 -64.524 52.673
448278 96.682 -64.530 52.635
448279 96.721 -64.534 52.597
448280 96.761 -64.539 52.558
448281 96.800 -64.543 52.517
448282 96.839 -64.548 52.475
448283 96.879 -64.552 52.431
448284 96.919 -64.557 52.386
448285 96.958 -64.562 52.341
448286 96.997 -64.566 52.297
448287 97.037 -64.571 52.253
448288 97.076 -64.575 52.212
448289 97.115 -64.580 52.171
448290 97.155 -64.584 52.129
448291 97.194 -64.589 52.084
448292 97.232 -64.593 52.039
448293 97.270 -64.594 51.996
448294 97.302 -64.589 51.952
448295 97.326 -64.577 51.907
448296 97.355 -64.567 51.859
448297 97.384 -64.556 51.810
448298 97.413 -64.546 51.763
448299 97.441 -64.553 51.717
448300 97.475 -64.555 51.671
448301 97.516 -64.551 51.626
448302 97.557 -64.547 51.581
448303 97.589 -64.552 51.535
448304 97.619 -64.561 51.488
448305 97.648 -64.569 51.440
448306 97.678 -64.577 51.394
448307 97.710 -64.584 51.347
448308 97.745 -64.590 51.300
448309 97.780 -64.595 51.252
448310 97.809 -64.599 51.204
448311 97.846 -64.600 51.155
448312 97.886 -64.598 51.106
448313 97.927 -64.596 51.058
448314 97.967 -64.593 51.010
448315 98.009 -64.590 50.963
448316 98.040 -64.593 50.917
448317 98.067 -64.599 50.871
448318 98.092 -64.606 50.825
448319 98.119 -64.612 50.781
448320 98.144 -64.618 50.737
448321 98.184 -64.629 50.692
448322 98.224 -64.639 50.646
448323 98.265 -64.649 50.600
448324 98.305 -64.659 50.554
448325 98.346 -64.670 50.510
448326 98.382 -64.676 50.467
448327 98.418 -64.681 50.424
448328 98.466 -64.672 50.380
448329 98.515 -64.663 50.336
448330 98.564 -64.654 50.291
448331 98.612 -64.645 50.246
448332 98.661 -64.636 50.200
448333 98.709 -64.627 50.155
448334 98.755 -64.621 50.112
448335 98.795 -64.621 50.070
448336 98.835 -64.622 50.029
448337 98.875 -64.623 49.987
448338 98.915 -64.624 49.944
448339 98.954 -64.624 49.901
448340 98.997 -64.621 49.859
448341 99.040 -64.617 49.818
448342 99.083 -64.613 49.777
448343 99.120 -64.617 49.737
448344 99.158 -64.620 49.697
448345 99.196 -64.624 49.656
448346 99.233 -64.627 49.613
448347 99.271 -64.631 49.568
448348 99.308 -64.634 49.522
448349 99.346 -64.637 49.476
448350 99.386 -64.642 49.432
448351 99.434 -64.649 49.389
448352 99.482 -64.658 49.347
448353 99.530 -64.665 49.305
448354 99.588 -64.674 49.266
448355 99.672 -64.658 49.225
448356 99.765 -64.632 49.186
448357 99.859 -64.606 49.146
448358 99.934 -64.599 49.106
448359 99.989 -64.611 49.068
448360 100.044 -64.622 49.031
448361 100.099 -64.634 48.993
448362 100.154 -64.646 48.955
448363 100.209 -64.657 48.918
448364 100.264 -64.669 48.879
448365 100.310 -64.679 48.840
448366 100.349 -64.684 48.802
448367 100.388 -64.690 48.765
448368 100.426 -64.694 48.728
448369 100.465 -64.700 48.689
448370 100.504 -64.705 48.651
448371 100.542 -64.710 48.612
448372 100.581 -64.715 48.574
448373 100.620 -64.721 48.537
448374 100.658 -64.726 48.500
448758 87.050 -63.054 55.910
448759 87.078 -63.060 55.896
448760 87.107 -63.067 55.883
448761 87.134 -63.073 55.869
448762 87.162 -63.079 55.855
448763 87.191 -63.086 55.841
448764 87.219 -63.092 55.826
448765 87.248 -63.099 55.813
448766 87.282 -63.109 55.800
448767 87.316 -63.118 55.788
448768 87.350 -63.128 55.775
448769 87.385 -63.138 55.762
448770 87.426 -63.144 55.751
448771 87.482 -63.145 55.738
448772 87.531 -63.155 55.723
448773 87.580 -63.165 55.709
448774 87.629 -63.176 55.694
448775 87.678 -63.186 55.679
448776 87.727 -63.196 55.665
448777 87.776 -63.207 55.650
448778 87.825 -63.217 55.634
448779 87.873 -63.227 55.618
448780 87.923 -63.238 55.601
448781 87.972 -63.249 55.583
448782 88.021 -63.259 55.565
448783 88.069 -63.269 55.546
448784 88.118 -63.280 55.527
448785 88.168 -63.290 55.508
448786 88.216 -63.300 55.489
448787 88.265 -63.311 55.470
448788 88.314 -63.321 55.452
448789 88.363 -63.331 55.436
448790 88.412 -63.342 55.419
448791 88.461 -63.352 55.403
448795 88.660 -63.391 55.359
448796 88.710 -63.400 55.353
448797 88.761 -63.409 55.349
448798 88.811 -63.418 55.346
448799 88.861 -63.427 55.345
448800 88.912 -63.436 55.346
448801 88.963 -63.445 55.346
448802 89.013 -63.454 55.348
448803 89.063 -63.463 55.352
448804 89.114 -63.471 55.355
448805 89.165 -63.480 55.361
448806 89.215 -63.490 55.367
448807 89.265 -63.499 55.374
448808 89.309 -63.514 55.383
448809 89.351 -63.529 55.397
448810 89.400 -63.533 55.410
448811 89.446 -63.544 55.422
448812 89.491 -63.552 55.431
448813 89.536 -63.559 55.437
448814 89.581 -63.566 55.441
448815 89.625 -63.573 55.442
448816 89.670 -63.580 55.439
448817 89.717 -63.577 55.430
448818 89.763 -63.574 55.414
448819 89.810 -63.571 55.390
448820 89.856 -63.568 55.367
448821 89.903 -63.565 55.347
448822 89.949 -63.562 55.328
448823 89.996 -63.559 55.309
448824 90.042 -63.558 55.296
448825 90.087 -63.559 55.289
448826 90.132 -63.560 55.286
448827 90.178 -63.562 55.288
448828 90.224 -63.563 55.300
448829 90.269 -63.565 55.312
448830 90.314 -63.565 55.322
448831 90.334 -63.581 55.328
448832 90.368 -63.595 55.332
448833 90.417 -63.606 55.333
448834 90.467 -63.618 55.332
448835 90.517 -63.630 55.335
448836 90.567 -63.641 55.338
448837 90.615 -63.654 55.333
448838 90.659 -63.671 55.306
448839 90.704 -63.687 55.273
448840 90.748 -63.704 55.238
448841 90.794 -63.712 55.197
448842 90.840 -63.719 55.148
448843 90.886 -63.727 55.093
448844 90.932 -63.734 55.038
448845 90.978 -63.742 54.987
448846 91.025 -63.750 54.943
448847 91.071 -63.757 54.903
448848 91.117 -63.765 54.868
448849 91.163 -63.772 54.838
448850 91.209 -63.780 54.812
448851 91.255 -63.788 54.790
448852 91.301 -63.795 54.771
448853 91.348 -63.803 54.757
448854 91.395 -63.812 54.743
448855 91.442 -63.820 54.732
448856 91.490 -63.828 54.723
448857 91.536 -63.838 54.715
448858 91.584 -63.846 54.707
448859 91.631 -63.855 54.700
448860 91.678 -63.863 54.694
448861 91.725 -63.872 54.690
448862 91.772 -63.881 54.684
448863 91.819 -63.889 54.679
448864 91.866 -63.898 54.674
448865 91.914 -63.907 54.670
448866 91.960 -63.915 54.665
448867 92.008 -63.924 54.660
448868 92.055 -63.932 54.655
448869 92.100 -63.938 54.650
448870 92.143 -63.938 54.645
448871 92.187 -63.938 54.639
448872 92.232 -63.941 54.634
448873 92.267 -63.946 54.629
448874 92.293 -63.953 54.624
448875 92.333 -63.963 54.620
448876 92.374 -63.973 54.616
448877 92.406 -63.985 54.612
448878 92.431 -64.001 54.609
448879 92.456 -64.016 54.604
448880 92.482 -64.032 54.600
448881 92.506 -64.047 54.596
448882 92.531 -64.062 54.595
448883 92.557 -64.078 54.592
448884 92.605 -64.089 54.589
448885 92.666 -64.093 54.587
448886 92.733 -64.094 54.585
448887 92.800 -64.095 54.585
448888 92.867 -64.095 54.585
448889 92.935 -64.095 54.586
448890 93.012 -64.095 54.588
448891 93.075 -64.095 54.589
448892 93.121 -64.095 54.589
448893 93.167 -64.095 54.590
448894 93.213 -64.095 54.590
448895 93.259 -64.095 54.588
448896 93.304 -64.095 54.584
448897 93.337 -64.100 54.579
448898 93.362 -64.107 54.571
448899 93.387 -64.114 54.564
448900 93.412 -64.121 54.558
448901 93.437 -64.128 54.548
448902 93.462 -64.135 54.537
448903 93.486 -64.142 54.523
448904 93.511 -64.149 54.505
448905 93.547 -64.151 54.487
448906 93.587 -64.149 54.468
448907 93.626 -64.154 54.452
448908 93.669 -64.164 54.436
448909 93.714 -64.174 54.419
448910 93.759 -64.184 54.401
448911 93.804 -64.194 54.384
448912 93.849 -64.203 54.364
448913 93.895 -64.213 54.343
448914 93.939 -64.223 54.321
448915 93.979 -64.231 54.297
448916 94.008 -64.236 54.273
448917 94.037 -64.242 54.249
448918 94.065 -64.248 54.226
448919 94.095 -64.253 54.202
448920 94.123 -64.258 54.180
448921 94.152 -64.264 54.156
448922 94.181 -64.270 54.131
448923 94.210 -64.275 54.108
448924 94.239 -64.281 54.088
448925 94.269 -64.284 54.068
448926 94.319 -64.290 54.047
448927 94.369 -64.300 54.026
448928 94.420 -64.310 54.004
448929 94.472 -64.320 53.982
448930 94.523 -64.330 53.961
448931 94.573 -64.339 53.943
448932 94.625 -64.349 53.925
448933 94.676 -64.359 53.908
448934 94.726 -64.369 53.891
448935 94.777 -64.379 53.869
448936 94.824 -64.384 53.847
448937 94.874 -64.391 53.825
448938 94.928 -64.398 53.801
448939 94.982 -64.405 53.778
448940 95.036 -64.412 53.758
448941 95.090 -64.419 53.738
448942 95.144 -64.427 53.717
448943 95.198 -64.433 53.695
448944 95.252 -64.441 53.672
448945 95.307 -64.448 53.647
448946 95.360 -64.455 53.624
448947 95.414 -64.462 53.600
448948 95.469 -64.469 53.576
448949 95.523 -64.477 53.552
448950 95.577 -64.483 53.530
448951 95.630 -64.491 53.508
448952 95.679 -64.499 53.487
448953 95.724 -64.506 53.466
448954 95.763 -64.512 53.441
448955 95.797 -64.515 53.410
448956 95.831 -64.519 53.378
448957 95.865 -64.522 53.344
448958 95.899 -64.525 53.308
448959 95.933 -64.528 53.270
448960 95.967 -64.531 53.233
448961 96.001 -64.534 53.197
448962 96.035 -64.537 53.163
448963 96.068 -64.540 53.131
448964 96.103 -64.543 53.099
448965 96.136 -64.546 53.069
448966 96.174 -64.544 53.038
448967 96.217 -64.531 53.004
448968 96.262 -64.517 52.974
448969 96.307 -64.502 52.945
448970 96.352 -64.488 52.916
448971 96.392 -64.483 52.889
448972 96.431 -64.481 52.862
448973 96.470 -64.479 52.834
448974 96.505 -64.480 52.805
448975 96.539 -64.481 52.775
448976 96.575 -64.482 52.744
448977 96.609 -64.483 52.712
448978 96.646 -64.485 52.678
448979 96.685 -64.490 52.640
448980 96.725 -64.494 52.601
448981 96.764 -64.499 52.561
448982 96.803 -64.503 52.518
448983 96.843 -64.508 52.475
448984 96.882 -64.512 52.430
448985 96.922 -64.517 52.384
448986 96.961 -64.521 52.337
448987 97.001 -64.526 52.292
448988 97.040 -64.530 52.249
448989 97.080 -64.536 52.209
448990 97.119 -64.540 52.169
448991 97.158 -64.545 52.126
448992 97.198 -64.549 52.081
448993 97.237 -64.554 52.035
448994 97.277 -64.558 51.991
448995 97.305 -64.547 51.946
448996 97.334 -64.537 51.900
448997 97.363 -64.525 51.852
448998 97.391 -64.515 51.803
448999 97.420 -64.510 51.756
449000 97.448 -64.517 51.710
449001 97.476 -64.524 51.665
449002 97.516 -64.521 51.619
449003 97.557 -64.517 51.575
449004 97.598 -64.512 51.529
449005 97.639 -64.508 51.482
449006 97.671 -64.514 51.435
449007 97.701 -64.522 51.388
449008 97.730 -64.530 51.343
449009 97.760 -64.540 51.296
449010 97.789 -64.548 51.248
449011 97.823 -64.552 51.200
449012 97.864 -64.549 51.151
449013 97.905 -64.546 51.101
449014 97.946 -64.543 51.051
449015 97.987 -64.541 51.003
449016 98.017 -64.545 50.955
449017 98.043 -64.552 50.907
449018 98.069 -64.558 50.860
449019 98.095 -64.565 50.813
449020 98.121 -64.571 50.767
449021 98.147 -64.577 50.722
449022 98.183 -64.587 50.676
449023 98.223 -64.597 50.630
449024 98.264 -64.607 50.583
449025 98.304 -64.618 50.537
449026 98.345 -64.627 50.492
449027 98.385 -64.638 50.449
449028 98.425 -64.646 50.407
449029 98.474 -64.637 50.363
449030 98.522 -64.628 50.319
449031 98.571 -64.619 50.274
449032 98.619 -64.610 50.229
449033 98.668 -64.601 50.184
449034 98.717 -64.592 50.140
449035 98.765 -64.583 50.098
449036 98.814 -64.574 50.057
449037 98.855 -64.572 50.016
449038 98.896 -64.573 49.972
449039 98.935 -64.574 49.928
449040 98.976 -64.574 49.885
449041 99.018 -64.570 49.842
449042 99.060 -64.567 49.801
449043 99.098 -64.571 49.759
449044 99.136 -64.574 49.718
449045 99.173 -64.577 49.678
449046 99.211 -64.580 49.636
449047 99.249 -64.584 49.592
449048 99.286 -64.588 49.548
449049 99.324 -64.591 49.502
449050 99.361 -64.595 49.458
449051 99.399 -64.598 49.414
449052 99.436 -64.602 49.372
449053 99.483 -64.609 49.330
449054 99.531 -64.617 49.290
449055 99.587 -64.624 49.251
449056 99.681 -64.599 49.212
449057 99.775 -64.574 49.172
449058 99.833 -64.582 49.132
449059 99.888 -64.594 49.093
449060 99.943 -64.606 49.055
449061 99.998 -64.618 49.018
449062 100.058 -64.627 48.981
449063 100.129 -64.633 48.944
449064 100.200 -64.637 48.906
449065 100.267 -64.639 48.869
449066 100.310 -64.643 48.830
449067 100.346 -64.650 48.793
449068 100.382 -64.658 48.756
449069 100.419 -64.664 48.718
449070 100.459 -64.668 48.681
449071 100.498 -64.673 48.643
449072 100.537 -64.678 48.604
449073 100.577 -64.683 48.567
449074 100.616 -64.687 48.531
449075 100.655 -64.692 48.494
449459 87.054 -63.011 55.916
449460 87.082 -63.017 55.903
449461 87.110 -63.024 55.890
449462 87.138 -63.030 55.876
449463 87.166 -63.037 55.863
449464 87.195 -63.043 55.849
449465 87.223 -63.050 55.833
449466 87.251 -63.057 55.820
449467 87.279 -63.063 55.807
449468 87.308 -63.070 55.796
449469 87.342 -63.079 55.784
449470 87.385 -63.086 55.773
449471 87.434 -63.095 55.764
449472 87.484 -63.105 55.751
449473 87.533 -63.116 55.737
449474 87.581 -63.126 55.721
449475 87.630 -63.136 55.704
449476 87.679 -63.147 55.689
449477 87.728 -63.157 55.674
449478 87.777 -63.167 55.659
449479 87.827 -63.177 55.644
449480 87.877 -63.186 55.628
449481 87.928 -63.195 55.612
449482 87.978 -63.205 55.594
449483 88.029 -63.213 55.576
449484 88.079 -63.222 55.558
449485 88.130 -63.231 55.539
449486 88.180 -63.240 55.519
449487 88.231 -63.249 55.500
449488 88.281 -63.258 55.481
449489 88.332 -63.267 55.462
449490 88.382 -63.276 55.445
449491 88.433 -63.285 55.427
449492 88.483 -63.293 55.412
449493 88.534 -63.302 55.398
449496 88.685 -63.330 55.365
449497 88.736 -63.338 55.358
449498 88.786 -63.347 55.354
449499 88.837 -63.356 55.351
449500 88.887 -63.365 55.349
449501 88.937 -63.374 55.349
449502 88.988 -63.383 55.350
449503 89.039 -63.392 55.352
449504 89.089 -63.401 55.355
449505 89.139 -63.410 55.358
449506 89.186 -63.422 55.362
449507 89.229 -63.437 55.366
449508 89.271 -63.452 55.372
449509 89.314 -63.468 55.380
449510 89.356 -63.484 55.391
449511 89.409 -63.477 55.405
449512 89.456 -63.489 55.419
449513 89.502 -63.500 55.429
449514 89.548 -63.509 55.437
449515 89.593 -63.517 55.440
449516 89.638 -63.524 55.441
449517 89.683 -63.528 55.437
449518 89.730 -63.524 55.427
449519 89.776 -63.521 55.411
449520 89.822 -63.519 55.388
449521 89.869 -63.516 55.367
449522 89.915 -63.514 55.347
449523 89.961 -63.515 55.328
449524 90.006 -63.516 55.310
449525 90.051 -63.518 55.298
449526 90.097 -63.519 55.291
449527 90.143 -63.521 55.290
449528 90.188 -63.521 55.294
449529 90.233 -63.523 55.310
449530 90.279 -63.524 55.324
449531 90.320 -63.528 55.335
449532 90.333 -63.546 55.341
449533 90.363 -63.561 55.342
449534 90.413 -63.573 55.339
449535 90.462 -63.584 55.334
449536 90.512 -63.596 55.333
449537 90.562 -63.607 55.336
449538 90.611 -63.619 55.335
449539 90.661 -63.631 55.310
449540 90.710 -63.643 55.283
449541 90.754 -63.659 55.250
449542 90.800 -63.668 55.210
449543 90.847 -63.675 55.159
449544 90.892 -63.683 55.099
449545 90.939 -63.690 55.040
449546 90.985 -63.698 54.986
449547 91.032 -63.707 54.940
449548 91.079 -63.715 54.900
449549 91.127 -63.724 54.865
449550 91.173 -63.733 54.835
449551 91.221 -63.741 54.808
449552 91.267 -63.750 54.786
449553 91.315 -63.759 54.768
449554 91.362 -63.767 54.754
449555 91.409 -63.776 54.741
449556 91.456 -63.784 54.729
449557 91.503 -63.793 54.720
449558 91.551 -63.802 54.713
449559 91.597 -63.810 54.707
449560 91.645 -63.819 54.700
449561 91.692 -63.828 54.695
449562 91.739 -63.836 54.690
449563 91.786 -63.845 54.685
449564 91.833 -63.853 54.681
449565 91.880 -63.862 54.676
449566 91.927 -63.871 54.672
449567 91.975 -63.879 54.667
449568 92.021 -63.888 54.662
449569 92.066 -63.890 54.657
449570 92.108 -63.890 54.652
449571 92.151 -63.889 54.647
449572 92.195 -63.889 54.642
449573 92.241 -63.895 54.636
449574 92.280 -63.900 54.631
449575 92.299 -63.905 54.626
449576 92.334 -63.914 54.622
449577 92.374 -63.923 54.619
449578 92.406 -63.937 54.615
449579 92.431 -63.952 54.611
449580 92.456 -63.967 54.607
449581 92.482 -63.983 54.604
449582 92.507 -63.998 54.600
449583 92.543 -64.012 54.598
449584 92.593 -64.023 54.597
449585 92.645 -64.032 54.594
449586 92.712 -64.033 54.592
449587 92.779 -64.033 54.591
449588 92.846 -64.034 54.591
449589 92.913 -64.035 54.591
449590 92.980 -64.035 54.592
449591 93.039 -64.038 54.595
449592 93.080 -64.038 54.595
449593 93.126 -64.037 54.596
449594 93.171 -64.037 54.596
449595 93.217 -64.037 54.595
449596 93.263 -64.037 54.591
449597 93.309 -64.038 54.587
449598 93.336 -64.044 54.580
449599 93.360 -64.051 54.572
449600 93.386 -64.058 54.563
449601 93.410 -64.065 54.556
449602 93.435 -64.072 54.544
449603 93.460 -64.079 54.531
449604 93.485 -64.086 54.515
449605 93.510 -64.093 54.496
449606 93.538 -64.098 54.476
449607 93.578 -64.098 54.454
449608 93.620 -64.107 54.434
449609 93.665 -64.117 54.414
449610 93.710 -64.127 54.395
449611 93.755 -64.137 54.376
449612 93.800 -64.147 54.357
449613 93.846 -64.157 54.337
449614 93.891 -64.167 54.316
449615 93.936 -64.176 54.294
449616 93.971 -64.183 54.271
449617 94.000 -64.189 54.247
449618 94.029 -64.194 54.224
449619 94.058 -64.200 54.201
449620 94.086 -64.205 54.178
449621 94.115 -64.211 54.155
449622 94.144 -64.217 54.132
449623 94.173 -64.222 54.109
449624 94.202 -64.227 54.087
449625 94.231 -64.233 54.070
449626 94.262 -64.240 54.049
449627 94.309 -64.249 54.029
449628 94.358 -64.258 54.007
449629 94.408 -64.268 53.984
449630 94.459 -64.278 53.961
449631 94.510 -64.288 53.940
449632 94.561 -64.298 53.921
449633 94.612 -64.308 53.903
449634 94.663 -64.317 53.886
449635 94.714 -64.327 53.870
449636 94.765 -64.337 53.849
449637 94.816 -64.344 53.828
449638 94.870 -64.351 53.806
449639 94.924 -64.358 53.783
449640 94.978 -64.365 53.760
449641 95.032 -64.372 53.740
449642 95.086 -64.380 53.720
449643 95.140 -64.386 53.700
449644 95.194 -64.394 53.679
449645 95.248 -64.401 53.657
449646 95.302 -64.408 53.634
449647 95.356 -64.415 53.612
449648 95.410 -64.422 53.589
449649 95.464 -64.430 53.565
449650 95.518 -64.436 53.543
449651 95.572 -64.444 53.521
449652 95.626 -64.451 53.500
449653 95.680 -64.458 53.479
449654 95.734 -64.465 53.459
449655 95.770 -64.470 53.433
449656 95.805 -64.473 53.402
449657 95.838 -64.476 53.370
449658 95.873 -64.479 53.336
449659 95.906 -64.482 53.300
449660 95.941 -64.485 53.263
449661 95.974 -64.488 53.225
449662 96.009 -64.491 53.190
449663 96.042 -64.495 53.156
449664 96.077 -64.498 53.124
449665 96.110 -64.501 53.093
449666 96.145 -64.504 53.062
449667 96.178 -64.507 53.031
449668 96.219 -64.499 53.000
449669 96.264 -64.485 52.972
449670 96.309 -64.470 52.945
449671 96.354 -64.455 52.918
449672 96.398 -64.443 52.892
449673 96.437 -64.440 52.867
449674 96.474 -64.439 52.838
449675 96.509 -64.441 52.807
449676 96.543 -64.442 52.778
449677 96.578 -64.443 52.749
449678 96.613 -64.443 52.719
449679 96.649 -64.446 52.685
449680 96.688 -64.450 52.647
449681 96.728 -64.455 52.607
449682 96.767 -64.459 52.565
449683 96.807 -64.464 52.520
449684 96.846 -64.468 52.475
449685 96.886 -64.473 52.429
449686 96.925 -64.477 52.381
449687 96.964 -64.482 52.334
449688 97.004 -64.486 52.289
449689 97.043 -64.491 52.247
449690 97.083 -64.496 52.207
449691 97.123 -64.499 52.167
449692 97.163 -64.503 52.124
449693 97.203 -64.507 52.077
449694 97.242 -64.511 52.032
449695 97.283 -64.515 51.987
449696 97.312 -64.505 51.941
449697 97.341 -64.495 51.893
449698 97.370 -64.484 51.845
449699 97.398 -64.474 51.798
449700 97.427 -64.474 51.752
449701 97.455 -64.480 51.706
449702 97.484 -64.487 51.660
449703 97.517 -64.490 51.614
449704 97.558 -64.486 51.570
449705 97.598 -64.482 51.523
449706 97.639 -64.478 51.476
449707 97.680 -64.474 51.429
449708 97.721 -64.470 51.383
449709 97.753 -64.475 51.337
449710 97.783 -64.483 51.291
449711 97.812 -64.493 51.244
449712 97.843 -64.500 51.196
449713 97.883 -64.497 51.146
449714 97.925 -64.494 51.095
449715 97.965 -64.492 51.045
449716 97.993 -64.497 50.995
449717 98.019 -64.504 50.947
449718 98.045 -64.510 50.899
449719 98.071 -64.517 50.850
449720 98.097 -64.523 50.802
449721 98.123 -64.530 50.755
449722 98.149 -64.536 50.709
449723 98.182 -64.544 50.662
449724 98.223 -64.555 50.615
449725 98.263 -64.565 50.569
449726 98.303 -64.575 50.523
449727 98.344 -64.586 50.477
449728 98.384 -64.596 50.433
449729 98.430 -64.597 50.391
449730 98.479 -64.591 50.348
449731 98.528 -64.585 50.304
449732 98.577 -64.578 50.258
449733 98.626 -64.572 50.213
449734 98.675 -64.566 50.168
449735 98.724 -64.557 50.125
449736 98.772 -64.548 50.083
449737 98.821 -64.539 50.042
449738 98.869 -64.530 50.000
449739 98.916 -64.523 49.956
449740 98.956 -64.524 49.912
449741 98.997 -64.522 49.868
449742 99.038 -64.521 49.825
449743 99.075 -64.524 49.782
449744 99.113 -64.528 49.740
449745 99.151 -64.531 49.698
449746 99.188 -64.535 49.657
449747 99.226 -64.538 49.615
449748 99.264 -64.542 49.572
449749 99.301 -64.545 49.527
449750 99.339 -64.549 49.483
449751 99.376 -64.552 49.440
449752 99.414 -64.555 49.397
449753 99.451 -64.559 49.355
449754 99.489 -64.562 49.314
449755 99.533 -64.568 49.274
449756 99.596 -64.568 49.237
449757 99.678 -64.554 49.198
449758 99.732 -64.566 49.158
449759 99.791 -64.576 49.119
449760 99.863 -64.581 49.081
449761 99.934 -64.587 49.043
449762 100.005 -64.592 49.006
449763 100.076 -64.596 48.968
449764 100.147 -64.602 48.932
449765 100.212 -64.602 48.895
449766 100.274 -64.600 48.857
449767 100.311 -64.606 48.819
449768 100.346 -64.614 48.783
449769 100.382 -64.621 48.746
449770 100.418 -64.629 48.709
449771 100.454 -64.637 48.672
449772 100.489 -64.644 48.635
449773 100.528 -64.649 48.597
449774 100.567 -64.654 48.560
449775 100.606 -64.659 48.525
449776 100.645 -64.664 48.487
450160 87.058 -62.968 55.925
450161 87.086 -62.975 55.912
450162 87.114 -62.982 55.898
450163 87.142 -62.988 55.884
450164 87.170 -62.995 55.871
450165 87.199 -63.001 55.857
450166 87.227 -63.007 55.843
450167 87.255 -63.014 55.827
450168 87.283 -63.020 55.814
450169 87.311 -63.027 55.803
450170 87.348 -63.029 55.794
450171 87.398 -63.035 55.785
450172 87.448 -63.044 55.776
450173 87.499 -63.053 55.764
450174 87.549 -63.062 55.750
450175 87.600 -63.070 55.734
450176 87.650 -63.079 55.716
450177 87.701 -63.089 55.699
450178 87.751 -63.098 55.683
450179 87.802 -63.107 55.668
450180 87.852 -63.115 55.653
450181 87.903 -63.124 55.637
450182 87.954 -63.133 55.621
450183 88.004 -63.142 55.604
450184 88.054 -63.151 55.587
450185 88.105 -63.160 55.568
450186 88.155 -63.169 55.550
450187 88.206 -63.178 55.531
450188 88.256 -63.187 55.512
450189 88.307 -63.195 55.492
450190 88.357 -63.205 55.472
450191 88.408 -63.214 55.453
450192 88.458 -63.223 55.436
450193 88.509 -63.232 55.420
450194 88.559 -63.240 55.406
450195 88.610 -63.249 55.392
450197 88.711 -63.267 55.371
450198 88.761 -63.277 55.364
450199 88.812 -63.285 55.359
450200 88.862 -63.294 55.355
450201 88.913 -63.303 55.354
450202 88.963 -63.312 55.353
450203 89.014 -63.321 55.354
450204 89.063 -63.331 55.355
450205 89.106 -63.346 55.358
450206 89.148 -63.361 55.360
450207 89.191 -63.377 55.363
450208 89.233 -63.392 55.367
450209 89.276 -63.407 55.371
450210 89.318 -63.422 55.376
450211 89.365 -63.431 55.386
450212 89.419 -63.423 55.400
450213 89.466 -63.434 55.415
450214 89.512 -63.445 55.427
450215 89.559 -63.456 55.435
450216 89.605 -63.467 55.438
450217 89.650 -63.474 55.439
450218 89.696 -63.476 55.435
450219 89.742 -63.473 55.424
450220 89.789 -63.470 55.406
450221 89.834 -63.471 55.385
450222 89.880 -63.472 55.364
450223 89.925 -63.474 55.344
450224 89.970 -63.475 55.326
450225 90.016 -63.477 55.310
450226 90.061 -63.477 55.299
450227 90.107 -63.479 55.292
450228 90.152 -63.480 55.292
450229 90.197 -63.482 55.297
450230 90.243 -63.483 55.313
450231 90.289 -63.484 55.329
450232 90.319 -63.493 55.342
450233 90.333 -63.512 55.349
450234 90.358 -63.528 55.350
450235 90.408 -63.539 55.345
450236 90.457 -63.550 55.337
450237 90.507 -63.562 55.333
450238 90.557 -63.574 55.333
450239 90.606 -63.585 55.330
450240 90.656 -63.596 55.312
450241 90.706 -63.609 55.286
450242 90.758 -63.619 55.254
450243 90.810 -63.628 55.216
450244 90.858 -63.637 55.162
450245 90.904 -63.645 55.098
450246 90.952 -63.654 55.038
450247 90.999 -63.662 54.984
450248 91.046 -63.671 54.938
450249 91.093 -63.680 54.897
450250 91.140 -63.688 54.862
450251 91.187 -63.697 54.831
450252 91.234 -63.706 54.805
450253 91.282 -63.714 54.783
450254 91.328 -63.723 54.766
450255 91.376 -63.731 54.751
450256 91.423 -63.740 54.739
450257 91.470 -63.749 54.728
450258 91.517 -63.757 54.719
450259 91.564 -63.766 54.713
450260 91.611 -63.775 54.707
450261 91.658 -63.783 54.701
450262 91.706 -63.792 54.695
450263 91.752 -63.800 54.691
450264 91.800 -63.809 54.687
450265 91.847 -63.818 54.683
450266 91.894 -63.826 54.679
450267 91.941 -63.835 54.675
450268 91.988 -63.842 54.671
450269 92.030 -63.841 54.667
450270 92.073 -63.841 54.662
450271 92.117 -63.841 54.656
450272 92.159 -63.841 54.650
450273 92.204 -63.843 54.644
450274 92.250 -63.848 54.638
450275 92.292 -63.854 54.634
450276 92.312 -63.859 54.630
450277 92.335 -63.865 54.626
450278 92.375 -63.875 54.623
450279 92.407 -63.888 54.619
450280 92.431 -63.904 54.615
450281 92.456 -63.919 54.611
450282 92.482 -63.934 54.608
450283 92.530 -63.945 54.605
450284 92.580 -63.957 54.602
450285 92.630 -63.968 54.601
450286 92.691 -63.973 54.601
450287 92.758 -63.973 54.598
450288 92.824 -63.973 54.597
450289 92.891 -63.974 54.598
450290 92.958 -63.974 54.598
450291 93.009 -63.980 54.598
450292 93.048 -63.986 54.600
450293 93.087 -63.986 54.601
450294 93.132 -63.984 54.601
450295 93.176 -63.981 54.600
450296 93.221 -63.979 54.598
450297 93.267 -63.979 54.594
450298 93.309 -63.981 54.589
450299 93.334 -63.988 54.582
450300 93.359 -63.995 54.573
450301 93.383 -64.002 54.563
450302 93.409 -64.009 54.553
450303 93.433 -64.016 54.540
450304 93.459 -64.023 54.525
450305 93.483 -64.030 54.507
450306 93.508 -64.038 54.487
450307 93.533 -64.045 54.465
450308 93.571 -64.051 54.442
450309 93.616 -64.061 54.419
450310 93.661 -64.071 54.396
450311 93.706 -64.080 54.375
450312 93.752 -64.090 54.353
450313 93.797 -64.100 54.332
450314 93.842 -64.110 54.310
450315 93.887 -64.120 54.289
450316 93.932 -64.129 54.266
450317 93.963 -64.135 54.244
450318 93.992 -64.141 54.220
450319 94.020 -64.146 54.197
450320 94.050 -64.152 54.175
450321 94.078 -64.157 54.152
450322 94.107 -64.163 54.130
450323 94.136 -64.168 54.107
450324 94.165 -64.174 54.085
450325 94.194 -64.179 54.065
450326 94.247 -64.203 54.049
450327 94.294 -64.220 54.029
450328 94.322 -64.214 54.008
450329 94.372 -64.223 53.986
450330 94.422 -64.233 53.964
450331 94.472 -64.242 53.941
450332 94.521 -64.251 53.921
450333 94.571 -64.261 53.902
450334 94.621 -64.270 53.883
450335 94.671 -64.280 53.865
450336 94.720 -64.290 53.848
450337 94.769 -64.296 53.830
450338 94.821 -64.301 53.810
450339 94.875 -64.308 53.789
450340 94.929 -64.315 53.765
450341 94.984 -64.322 53.742
450342 95.038 -64.330 53.722
450343 95.092 -64.336 53.702
450344 95.146 -64.343 53.683
450345 95.200 -64.351 53.663
450346 95.255 -64.358 53.642
450347 95.309 -64.365 53.620
450348 95.363 -64.372 53.599
450349 95.417 -64.379 53.577
450350 95.471 -64.386 53.554
450351 95.526 -64.393 53.534
450352 95.580 -64.400 53.512
450353 95.634 -64.408 53.490
450354 95.688 -64.414 53.470
450355 95.733 -64.421 53.450
450356 95.765 -64.419 53.423
450357 95.799 -64.422 53.392
450358 95.833 -64.425 53.359
450359 95.867 -64.428 53.326
450360 95.901 -64.431 53.291
450361 95.935 -64.434 53.256
450362 95.969 -64.437 53.220
450363 96.003 -64.440 53.185
450364 96.037 -64.444 53.152
450365 96.071 -64.447 53.120
450366 96.105 -64.450 53.088
450367 96.139 -64.453 53.057
450368 96.176 -64.449 53.027
450369 96.221 -64.441 52.999
450370 96.265 -64.433 52.974
450371 96.311 -64.425 52.949
450372 96.356 -64.418 52.924
450373 96.401 -64.409 52.898
450374 96.443 -64.399 52.871
450375 96.477 -64.401 52.840
450376 96.513 -64.402 52.809
450377 96.547 -64.402 52.781
450378 96.582 -64.403 52.752
450379 96.617 -64.405 52.722
450380 96.652 -64.405 52.689
450381 96.692 -64.410 52.650
450382 96.731 -64.414 52.609
450383 96.770 -64.419 52.566
450384 96.810 -64.424 52.521
450385 96.850 -64.429 52.474
450386 96.889 -64.433 52.426
450387 96.929 -64.437 52.377
450388 96.969 -64.441 52.329
450389 97.009 -64.445 52.285
450390 97.048 -64.449 52.244
450391 97.089 -64.452 52.204
450392 97.129 -64.456 52.162
450393 97.168 -64.460 52.118
450394 97.208 -64.464 52.071
450395 97.248 -64.468 52.025
450396 97.288 -64.471 51.980
450397 97.320 -64.464 51.933
450398 97.349 -64.454 51.885
450399 97.377 -64.443 51.838
450400 97.405 -64.432 51.793
450401 97.434 -64.437 51.748
450402 97.462 -64.444 51.702
450403 97.491 -64.451 51.654
450404 97.519 -64.458 51.608
450405 97.558 -64.456 51.562
450406 97.599 -64.452 51.515
450407 97.640 -64.448 51.468
450408 97.680 -64.443 51.421
450409 97.721 -64.439 51.375
450410 97.762 -64.436 51.329
450411 97.803 -64.431 51.283
450412 97.835 -64.437 51.237
450413 97.865 -64.446 51.190
450414 97.903 -64.446 51.141
450415 97.943 -64.443 51.090
450416 97.969 -64.449 51.040
450417 97.995 -64.456 50.989
450418 98.021 -64.462 50.939
450419 98.047 -64.469 50.890
450420 98.073 -64.475 50.841
450421 98.099 -64.482 50.792
450422 98.125 -64.488 50.744
450423 98.151 -64.495 50.697
450424 98.181 -64.502 50.650
450425 98.222 -64.513 50.604
450426 98.263 -64.523 50.557
450427 98.303 -64.533 50.511
450428 98.344 -64.543 50.464
450429 98.388 -64.546 50.418
450430 98.435 -64.545 50.375
450431 98.484 -64.539 50.331
450432 98.533 -64.533 50.287
450433 98.582 -64.526 50.242
450434 98.631 -64.520 50.196
450435 98.680 -64.514 50.152
450436 98.729 -64.508 50.109
450437 98.778 -64.501 50.067
450438 98.827 -64.495 50.024
450439 98.876 -64.489 49.981
450440 98.925 -64.483 49.938
450441 98.974 -64.476 49.894
450442 99.015 -64.475 49.850
450443 99.053 -64.478 49.806
450444 99.091 -64.482 49.762
450445 99.128 -64.485 49.719
450446 99.166 -64.489 49.677
450447 99.203 -64.493 49.636
450448 99.240 -64.496 49.594
450449 99.278 -64.499 49.551
450450 99.315 -64.503 49.508
450451 99.352 -64.507 49.464
450452 99.390 -64.511 49.422
450453 99.427 -64.514 49.381
450454 99.464 -64.518 49.340
450455 99.501 -64.521 49.300
450456 99.539 -64.525 49.260
450457 99.596 -64.530 49.222
450458 99.667 -64.535 49.183
450459 99.739 -64.540 49.144
450460 99.810 -64.546 49.106
450461 99.881 -64.551 49.068
450462 99.952 -64.556 49.030
450463 100.023 -64.561 48.993
450464 100.095 -64.566 48.956
450465 100.158 -64.565 48.919
450466 100.221 -64.563 48.882
450467 100.276 -64.562 48.845
450468 100.312 -64.570 48.808
450469 100.347 -64.577 48.772
450470 100.383 -64.584 48.736
450471 100.419 -64.592 48.699
450472 100.455 -64.599 48.663
450473 100.490 -64.607 48.626
450474 100.525 -64.615 48.589
450475 100.561 -64.622 48.552
450476 100.597 -64.630 48.517
450477 100.635 -64.635 48.482
450857 86.949 -62.900 55.991
450858 86.978 -62.907 55.978
450859 87.006 -62.913 55.963
450860 87.034 -62.919 55.949
450861 87.062 -62.926 55.935
450862 87.090 -62.932 55.921
450863 87.118 -62.939 55.907
450864 87.146 -62.945 55.894
450865 87.175 -62.952 55.880
450866 87.203 -62.958 55.866
450867 87.230 -62.965 55.851
450868 87.259 -62.972 55.837
450869 87.287 -62.978 55.823
450870 87.325 -62.979 55.811
450871 87.374 -62.974 55.801
450872 87.423 -62.973 55.794
450873 87.474 -62.982 55.786
450874 87.525 -62.991 55.775
450875 87.575 -63.000 55.762
450876 87.625 -63.009 55.747
450877 87.676 -63.017 55.729
450878 87.726 -63.026 55.710
450879 87.777 -63.035 55.693
450880 87.827 -63.045 55.676
450881 87.878 -63.054 55.661
450882 87.928 -63.062 55.646
450883 87.979 -63.071 55.630
450884 88.030 -63.080 55.614
450885 88.080 -63.089 55.597
450886 88.130 -63.098 55.579
450887 88.181 -63.107 55.562
450888 88.232 -63.116 55.542
450889 88.282 -63.125 55.523
450890 88.332 -63.134 55.503
450891 88.383 -63.142 55.483
450892 88.433 -63.151 55.463
450893 88.484 -63.161 55.445
450894 88.534 -63.170 55.429
450895 88.585 -63.179 55.413
450896 88.635 -63.187 55.400
450897 88.686 -63.196 55.387
450898 88.736 -63.205 55.377
450899 88.787 -63.214 55.369
450900 88.837 -63.224 55.363
450901 88.888 -63.232 55.360
450902 88.939 -63.241 55.359
450903 88.983 -63.255 55.358
450904 89.026 -63.270 55.357
450905 89.068 -63.285 55.358
450906 89.111 -63.301 55.360
450907 89.153 -63.316 55.362
450908 89.196 -63.331 55.365
450909 89.238 -63.346 55.367
450910 89.281 -63.361 55.370
450911 89.323 -63.377 55.374
450912 89.374 -63.376 55.381
450913 89.428 -63.368 55.394
450914 89.475 -63.379 55.409
450915 89.522 -63.390 55.422
450916 89.568 -63.401 55.431
450917 89.615 -63.412 55.436
450918 89.662 -63.424 55.436
450919 89.708 -63.427 55.431
450920 89.753 -63.428 55.420
450921 89.799 -63.430 55.402
450922 89.844 -63.431 55.380
450923 89.889 -63.433 55.359
450924 89.935 -63.433 55.340
450925 89.980 -63.435 55.323
450926 90.026 -63.437 55.308
450927 90.071 -63.438 55.298
450928 90.116 -63.439 55.292
450929 90.162 -63.440 55.291
450930 90.207 -63.442 55.296
450931 90.253 -63.443 55.312
450932 90.298 -63.444 55.330
450933 90.319 -63.459 55.344
450934 90.332 -63.477 55.353
450935 90.353 -63.493 55.356
450936 90.403 -63.505 55.349
450937 90.453 -63.517 55.340
450938 90.502 -63.528 55.333
450939 90.552 -63.540 55.330
450940 90.602 -63.551 55.326
450941 90.651 -63.563 55.311
450942 90.703 -63.574 55.285
450943 90.761 -63.583 55.251
450944 90.820 -63.592 55.210
450945 90.871 -63.601 55.156
450946 90.919 -63.609 55.093
450947 90.965 -63.618 55.033
450948 91.013 -63.627 54.980
450949 91.060 -63.635 54.934
450950 91.107 -63.644 54.894
450951 91.154 -63.653 54.859
450952 91.201 -63.661 54.829
450953 91.248 -63.670 54.803
450954 91.295 -63.678 54.782
450955 91.343 -63.687 54.764
450956 91.389 -63.696 54.750
450957 91.437 -63.704 54.738
450958 91.484 -63.713 54.729
450959 91.531 -63.722 54.720
450960 91.578 -63.730 54.713
450961 91.625 -63.739 54.708
450962 91.672 -63.747 54.702
450963 91.719 -63.756 54.697
450964 91.767 -63.765 54.693
450965 91.813 -63.773 54.689
450966 91.861 -63.782 54.686
450967 91.908 -63.791 54.682
450968 91.952 -63.794 54.680
450969 91.995 -63.794 54.677
450970 92.038 -63.794 54.674
450971 92.081 -63.793 54.668
450972 92.124 -63.793 54.662
450973 92.167 -63.793 54.655
450974 92.213 -63.797 54.648
450975 92.260 -63.802 54.643
450976 92.305 -63.808 54.638
450977 92.324 -63.813 54.634
450978 92.344 -63.818 54.631
450979 92.376 -63.825 54.627
450980 92.407 -63.839 54.624
450981 92.432 -63.855 54.620
450982 92.467 -63.868 54.616
450983 92.517 -63.879 54.613
450984 92.567 -63.891 54.610
450985 92.617 -63.901 54.608
450986 92.669 -63.911 54.607
450987 92.736 -63.912 54.606
450988 92.803 -63.913 54.605
450989 92.870 -63.913 54.604
450990 92.937 -63.913 54.604
450991 92.980 -63.922 54.603
450992 93.019 -63.931 54.604
450993 93.057 -63.933 54.604
450994 93.095 -63.933 54.605
450995 93.138 -63.932 54.605
450996 93.183 -63.929 54.604
450997 93.227 -63.926 54.601
450998 93.272 -63.923 54.596
450999 93.308 -63.925 54.590
451000 93.334 -63.932 54.583
451001 93.360 -63.938 54.574
451002 93.386 -63.945 54.562
451003 93.411 -63.952 54.551
451004 93.437 -63.959 54.536
451005 93.463 -63.966 54.519
451006 93.488 -63.973 54.500
451007 93.514 -63.979 54.478
451008 93.540 -63.986 54.455
451009 93.572 -64.001 54.431
451010 93.612 -64.014 54.406
451011 93.657 -64.024 54.381
451012 93.703 -64.034 54.357
451013 93.748 -64.044 54.332
451014 93.793 -64.054 54.308
451015 93.838 -64.063 54.285
451016 93.884 -64.073 54.262
451017 93.926 -64.082 54.240
451018 93.955 -64.088 54.217
451019 93.984 -64.093 54.194
451020 94.013 -64.098 54.171
451021 94.042 -64.104 54.148
451022 94.070 -64.110 54.126
451023 94.100 -64.115 54.104
451024 94.128 -64.120 54.082
451025 94.176 -64.140 54.062
451026 94.231 -64.166 54.042
451027 94.287 -64.192 54.026
451028 94.323 -64.195 54.006
451029 94.347 -64.186 53.986
451030 94.386 -64.188 53.965
451031 94.435 -64.198 53.943
451032 94.485 -64.207 53.922
451033 94.535 -64.217 53.903
451034 94.584 -64.226 53.884
451035 94.634 -64.236 53.865
451036 94.684 -64.245 53.846
451037 94.734 -64.255 53.828
451038 94.781 -64.254 53.809
451039 94.831 -64.257 53.790
451040 94.886 -64.264 53.768
451041 94.939 -64.271 53.745
451042 94.994 -64.278 53.724
451043 95.048 -64.285 53.705
451044 95.102 -64.292 53.686
451045 95.157 -64.299 53.666
451046 95.210 -64.306 53.647
451047 95.265 -64.314 53.626
451048 95.319 -64.320 53.606
451049 95.373 -64.328 53.585
451050 95.427 -64.335 53.565
451051 95.481 -64.342 53.543
451052 95.536 -64.349 53.522
451053 95.590 -64.356 53.501
451054 95.642 -64.363 53.479
451055 95.686 -64.370 53.457
451056 95.731 -64.377 53.436
451057 95.758 -64.367 53.408
451058 95.792 -64.370 53.378
451059 95.826 -64.373 53.346
451060 95.860 -64.377 53.313
451061 95.894 -64.380 53.279
451062 95.927 -64.383 53.248
451063 95.962 -64.386 53.216
451064 95.995 -64.389 53.182
451065 96.030 -64.392 53.149
451066 96.063 -64.395 53.118
451067 96.098 -64.398 53.086
451068 96.133 -64.396 53.055
451069 96.177 -64.383 53.027
451070 96.222 -64.374 53.001
451071 96.267 -64.367 52.978
451072 96.312 -64.359 52.955
451073 96.358 -64.351 52.931
451074 96.404 -64.347 52.904
451075 96.452 -64.355 52.875
451076 96.486 -64.357 52.842
451077 96.520 -64.359 52.811
451078 96.553 -64.361 52.782
451079 96.587 -64.363 52.753
451080 96.621 -64.365 52.723
451081 96.656 -64.366 52.691
451082 96.695 -64.371 52.651
451083 96.735 -64.374 52.608
451084 96.775 -64.378 52.564
451085 96.815 -64.382 52.518
451086 96.855 -64.386 52.470
451087 96.895 -64.389 52.421
451088 96.935 -64.393 52.371
451089 96.974 -64.397 52.322
451090 97.014 -64.401 52.280
451091 97.054 -64.405 52.240
451092 97.094 -64.408 52.199
451093 97.134 -64.412 52.156
451094 97.174 -64.416 52.112
451095 97.214 -64.420 52.064
451096 97.254 -64.424 52.018
451097 97.294 -64.428 51.972
451098 97.327 -64.423 51.924
451099 97.356 -64.412 51.876
451100 97.384 -64.402 51.830
451101 97.413 -64.393 51.786
451102 97.441 -64.400 51.741
451103 97.469 -64.408 51.696
451104 97.498 -64.414 51.649
451105 97.526 -64.422 51.602
451106 97.558 -64.426 51.555
451107 97.599 -64.421 51.508
451108 97.640 -64.418 51.461
451109 97.681 -64.413 51.414
451110 97.722 -64.409 51.368
451111 97.762 -64.405 51.322
451112 97.803 -64.401 51.276
451113 97.845 -64.397 51.230
451114 97.885 -64.392 51.184
451115 97.920 -64.395 51.136
451116 97.946 -64.402 51.085
451117 97.971 -64.408 51.034
451118 97.998 -64.414 50.983
451119 98.023 -64.421 50.932
451120 98.050 -64.427 50.881
451121 98.075 -64.434 50.831
451122 98.101 -64.440 50.782
451123 98.128 -64.447 50.733
451124 98.153 -64.453 50.686
451125 98.180 -64.460 50.639
451126 98.221 -64.471 50.592
451127 98.262 -64.481 50.545
451128 98.303 -64.490 50.498
451129 98.348 -64.493 50.451
451130 98.392 -64.496 50.404
451131 98.440 -64.493 50.359
451132 98.489 -64.486 50.315
451133 98.538 -64.480 50.270
451134 98.587 -64.474 50.226
451135 98.636 -64.468 50.181
451136 98.685 -64.461 50.137
451137 98.734 -64.455 50.094
451138 98.783 -64.449 50.051
451139 98.832 -64.443 50.007
451140 98.881 -64.436 49.964
451141 98.928 -64.436 49.920
451142 98.974 -64.438 49.876
451143 99.014 -64.442 49.832
451144 99.052 -64.445 49.787
451145 99.089 -64.449 49.743
451146 99.126 -64.452 49.700
451147 99.163 -64.456 49.657
451148 99.201 -64.459 49.615
451149 99.238 -64.463 49.573
451150 99.275 -64.467 49.531
451151 99.312 -64.471 49.489
451152 99.350 -64.474 49.447
451153 99.387 -64.477 49.405
451154 99.424 -64.481 49.365
451155 99.462 -64.485 49.326
451156 99.499 -64.488 49.286
451157 99.544 -64.494 49.246
451158 99.615 -64.499 49.207
451159 99.686 -64.505 49.169
451160 99.757 -64.510 49.131
451161 99.829 -64.515 49.092
451162 99.900 -64.521 49.054
451163 99.971 -64.525 49.016
451164 100.041 -64.530 48.980
451165 100.104 -64.527 48.943
451166 100.167 -64.526 48.907
451167 100.229 -64.524 48.871
451168 100.277 -64.525 48.833
451169 100.313 -64.533 48.797
451170 100.348 -64.540 48.761
451171 100.384 -64.548 48.726
451172 100.420 -64.555 48.690
451173 100.455 -64.563 48.653
451174 100.491 -64.571 48.617
451175 100.527 -64.578 48.581
451176 100.562 -64.585 48.545
451177 100.598 -64.593 48.511
451555 86.869 -62.838 56.042
451556 86.898 -62.844 56.028
451557 86.925 -62.851 56.015
451558 86.953 -62.857 56.001
451559 86.982 -62.864 55.987
451560 87.010 -62.870 55.974
451561 87.038 -62.877 55.959
451562 87.066 -62.883 55.945
451563 87.094 -62.890 55.932
451564 87.122 -62.897 55.916
451565 87.150 -62.903 55.903
451566 87.179 -62.910 55.889
451567 87.206 -62.916 55.874
451568 87.234 -62.923 55.859
451569 87.263 -62.929 55.846
451570 87.303 -62.929 55.832
451571 87.352 -62.924 55.819
451572 87.401 -62.919 55.809
451573 87.449 -62.915 55.801
451574 87.499 -62.919 55.793
451575 87.550 -62.929 55.782
451576 87.601 -62.938 55.769
451577 87.651 -62.947 55.754
451578 87.701 -62.956 55.738
451579 87.752 -62.964 55.719
451580 87.802 -62.973 55.701
451581 87.853 -62.982 55.685
451582 87.904 -62.992 55.670
451583 87.954 -63.001 55.655
451584 88.004 -63.009 55.640
451585 88.055 -63.018 55.625
451586 88.106 -63.027 55.608
451587 88.156 -63.036 55.589
451588 88.206 -63.045 55.573
451589 88.257 -63.054 55.554
451590 88.308 -63.063 55.534
451591 88.358 -63.072 55.514
451592 88.408 -63.081 55.494
451593 88.459 -63.089 55.473
451594 88.509 -63.098 55.455
451595 88.560 -63.108 55.438
451596 88.610 -63.117 55.421
451597 88.661 -63.126 55.407
451598 88.711 -63.134 55.393
451599 88.762 -63.143 55.383
451600 88.813 -63.152 55.374
451601 88.860 -63.164 55.369
451602 88.903 -63.179 55.364
451603 88.945 -63.194 55.362
451604 88.988 -63.209 55.361
451605 89.030 -63.224 55.361
451606 89.073 -63.239 55.360
451607 89.115 -63.255 55.362
451608 89.158 -63.270 55.364
451609 89.200 -63.286 55.365
451610 89.243 -63.301 55.367
451611 89.285 -63.316 55.368
451612 89.328 -63.331 55.371
451613 89.384 -63.321 55.377
451614 89.439 -63.312 55.389
451615 89.485 -63.324 55.403
451616 89.532 -63.335 55.416
451617 89.579 -63.346 55.426
451618 89.625 -63.358 55.432
451619 89.671 -63.374 55.432
451620 89.717 -63.386 55.426
451621 89.763 -63.389 55.414
451622 89.809 -63.390 55.397
451623 89.854 -63.391 55.376
451624 89.899 -63.393 55.356
451625 89.945 -63.394 55.338
451626 89.990 -63.395 55.320
451627 90.036 -63.396 55.305
451628 90.081 -63.398 55.295
451629 90.126 -63.399 55.289
451630 90.172 -63.400 55.287
451631 90.217 -63.402 55.290
451632 90.263 -63.403 55.304
451633 90.305 -63.406 55.321
451634 90.318 -63.424 55.338
451635 90.332 -63.442 55.350
451636 90.348 -63.459 55.354
451637 90.398 -63.471 55.350
451638 90.448 -63.483 55.342
451639 90.497 -63.494 55.334
451640 90.547 -63.506 55.328
451641 90.597 -63.518 55.324
451642 90.647 -63.529 55.307
451643 90.706 -63.538 55.280
451644 90.764 -63.547 55.243
451645 90.823 -63.556 55.196
451646 90.882 -63.565 55.145
451647 90.932 -63.574 55.083
451648 90.979 -63.582 55.025
451649 91.026 -63.591 54.974
451650 91.074 -63.600 54.930
451651 91.121 -63.608 54.891
451652 91.168 -63.617 54.857
451653 91.215 -63.625 54.827
451654 91.262 -63.634 54.802
451655 91.309 -63.643 54.781
451656 91.356 -63.651 54.764
451657 91.403 -63.660 54.750
451658 91.450 -63.669 54.739
451659 91.498 -63.677 54.730
451660 91.545 -63.686 54.721
451661 91.592 -63.694 54.715
451662 91.639 -63.703 54.710
451663 91.686 -63.712 54.705
451664 91.733 -63.720 54.700
451665 91.780 -63.729 54.696
451666 91.827 -63.737 54.693
451667 91.874 -63.746 54.690
451668 91.918 -63.746 54.689
451669 91.960 -63.745 54.687
451670 92.003 -63.745 54.684
451671 92.046 -63.745 54.680
451672 92.089 -63.744 54.675
451673 92.132 -63.744 54.669
451674 92.176 -63.745 54.662
451675 92.223 -63.750 54.655
451676 92.269 -63.756 54.650
451677 92.316 -63.761 54.645
451678 92.337 -63.766 54.640
451679 92.357 -63.772 54.636
451680 92.377 -63.777 54.632
451681 92.407 -63.791 54.629
451682 92.454 -63.802 54.625
451683 92.504 -63.813 54.621
451684 92.555 -63.824 54.618
451685 92.604 -63.835 54.616
451686 92.654 -63.847 54.614
451687 92.714 -63.851 54.613
451688 92.781 -63.852 54.613
451689 92.848 -63.852 54.612
451690 92.911 -63.854 54.611
451691 92.950 -63.863 54.610
451692 92.989 -63.872 54.610
451693 93.028 -63.880 54.609
451694 93.066 -63.880 54.609
451695 93.105 -63.880 54.608
451696 93.145 -63.879 54.607
451697 93.190 -63.877 54.605
451698 93.234 -63.874 54.601
451699 93.279 -63.871 54.596
451700 93.316 -63.869 54.591
451701 93.344 -63.872 54.583
451702 93.369 -63.879 54.574
451703 93.395 -63.886 54.562
451704 93.421 -63.893 54.549
451705 93.446 -63.900 54.533
451706 93.473 -63.907 54.515
451707 93.498 -63.913 54.494
451708 93.524 -63.920 54.471
451709 93.550 -63.929 54.447
451710 93.583 -63.945 54.421
451711 93.617 -63.963 54.395
451712 93.654 -63.977 54.369
451713 93.699 -63.987 54.343
451714 93.744 -63.997 54.314
451715 93.789 -64.007 54.287
451716 93.835 -64.016 54.262
451717 93.880 -64.026 54.237
451718 93.918 -64.034 54.215
451719 93.947 -64.039 54.192
451720 93.976 -64.045 54.168
451721 94.005 -64.051 54.145
451722 94.034 -64.056 54.122
451723 94.063 -64.062 54.099
451724 94.105 -64.077 54.077
451725 94.160 -64.103 54.057
451726 94.216 -64.129 54.038
451727 94.272 -64.154 54.019
451728 94.316 -64.165 54.002
451729 94.351 -64.170 53.982
451730 94.376 -64.161 53.962
451731 94.400 -64.153 53.943
451732 94.449 -64.162 53.924
451733 94.499 -64.172 53.903
451734 94.549 -64.181 53.885
451735 94.598 -64.191 53.867
451736 94.648 -64.201 53.848
451737 94.698 -64.210 53.829
451738 94.746 -64.212 53.810
451739 94.793 -64.212 53.791
451740 94.841 -64.212 53.770
451741 94.896 -64.220 53.749
451742 94.950 -64.226 53.727
451743 95.004 -64.234 53.707
451744 95.058 -64.241 53.688
451745 95.112 -64.248 53.669
451746 95.167 -64.255 53.650
451747 95.220 -64.262 53.631
451748 95.275 -64.269 53.610
451749 95.329 -64.276 53.591
451750 95.383 -64.283 53.572
451751 95.438 -64.291 53.552
451752 95.491 -64.298 53.531
451753 95.546 -64.305 53.510
451754 95.596 -64.311 53.487
451755 95.640 -64.318 53.464
451756 95.684 -64.325 53.441
451757 95.727 -64.331 53.418
451758 95.753 -64.320 53.390
451759 95.784 -64.318 53.360
451760 95.818 -64.321 53.330
451761 95.852 -64.324 53.301
451762 95.886 -64.327 53.272
451763 95.920 -64.330 53.243
451764 95.954 -64.333 53.212
451765 95.988 -64.336 53.179
451766 96.022 -64.339 53.146
451767 96.056 -64.342 53.113
451768 96.092 -64.342 53.081
451769 96.133 -64.323 53.052
451770 96.179 -64.316 53.026
451771 96.224 -64.308 53.003
451772 96.269 -64.300 52.981
451773 96.314 -64.292 52.960
451774 96.361 -64.291 52.935
451775 96.413 -64.300 52.907
451776 96.463 -64.309 52.877
451777 96.497 -64.311 52.844
451778 96.531 -64.313 52.812
451779 96.565 -64.315 52.781
451780 96.598 -64.317 52.751
451781 96.632 -64.320 52.721
451782 96.666 -64.322 52.688
451783 96.703 -64.327 52.649
451784 96.741 -64.331 52.605
451785 96.781 -64.335 52.559
451786 96.820 -64.339 52.510
451787 96.861 -64.342 52.462
451788 96.900 -64.346 52.413
451789 96.940 -64.350 52.362
451790 96.980 -64.354 52.313
451791 97.020 -64.358 52.273
451792 97.060 -64.361 52.232
451793 97.100 -64.365 52.192
451794 97.140 -64.369 52.150
451795 97.180 -64.373 52.104
451796 97.220 -64.377 52.058
451797 97.259 -64.380 52.010
451798 97.299 -64.384 51.963
451799 97.334 -64.382 51.915
451800 97.363 -64.371 51.867
451801 97.391 -64.360 51.821
451802 97.420 -64.357 51.777
451803 97.448 -64.364 51.734
451804 97.476 -64.371 51.689
451805 97.505 -64.378 51.643
451806 97.533 -64.385 51.597
451807 97.562 -64.392 51.549
451808 97.599 -64.391 51.502
451809 97.640 -64.387 51.454
451810 97.681 -64.383 51.407
451811 97.722 -64.379 51.361
451812 97.763 -64.375 51.315
451813 97.804 -64.371 51.269
451814 97.845 -64.367 51.223
451815 97.885 -64.362 51.177
451816 97.920 -64.361 51.130
451817 97.946 -64.366 51.080
451818 97.973 -64.371 51.028
451819 97.999 -64.375 50.976
451820 98.026 -64.380 50.924
451821 98.052 -64.386 50.872
451822 98.078 -64.392 50.822
451823 98.103 -64.399 50.772
451824 98.130 -64.405 50.723
451825 98.156 -64.412 50.675
451826 98.181 -64.418 50.628
451827 98.220 -64.428 50.581
451828 98.262 -64.436 50.534
451829 98.307 -64.439 50.486
451830 98.351 -64.443 50.439
451831 98.396 -64.446 50.391
451832 98.445 -64.440 50.345
451833 98.494 -64.434 50.299
451834 98.543 -64.428 50.254
451835 98.592 -64.421 50.210
451836 98.641 -64.415 50.167
451837 98.690 -64.409 50.122
451838 98.739 -64.403 50.079
451839 98.788 -64.396 50.035
451840 98.834 -64.397 49.990
451841 98.880 -64.400 49.947
451842 98.926 -64.403 49.902
451843 98.973 -64.405 49.858
451844 99.012 -64.408 49.814
451845 99.049 -64.412 49.768
451846 99.086 -64.416 49.725
451847 99.124 -64.419 49.681
451848 99.161 -64.423 49.637
451849 99.198 -64.427 49.595
451850 99.235 -64.430 49.553
451851 99.273 -64.433 49.511
451852 99.310 -64.437 49.471
451853 99.347 -64.441 49.430
451854 99.384 -64.445 49.389
451855 99.422 -64.449 49.349
451856 99.454 -64.452 49.311
451857 99.491 -64.458 49.271
451858 99.562 -64.464 49.232
451859 99.633 -64.469 49.193
451860 99.704 -64.474 49.155
451861 99.776 -64.480 49.117
451862 99.847 -64.485 49.079
451863 99.918 -64.490 49.041
451864 99.986 -64.493 49.003
451865 100.049 -64.490 48.966
451866 100.112 -64.489 48.930
451867 100.175 -64.487 48.894
451868 100.238 -64.485 48.859
451869 100.278 -64.489 48.822
451870 100.314 -64.496 48.786
451871 100.349 -64.504 48.750
451872 100.385 -64.511 48.715
451873 100.420 -64.518 48.680
451874 100.456 -64.526 48.644
451875 100.492 -64.533 48.609
451876 100.528 -64.541 48.573
451877 100.563 -64.549 48.538
451878 100.599 -64.556 48.505
452256 86.873 -62.795 56.053
452257 86.902 -62.802 56.040
452258 86.929 -62.809 56.026
452259 86.957 -62.815 56.013
452260 86.986 -62.822 56.000
452261 87.014 -62.828 55.986
452262 87.042 -62.835 55.972
452263 87.070 -62.841 55.958
452264 87.098 -62.847 55.943
452265 87.126 -62.854 55.928
452266 87.154 -62.860 55.913
452267 87.183 -62.867 55.899
452268 87.210 -62.873 55.883
452269 87.238 -62.880 55.868
452270 87.280 -62.879 55.854
452271 87.329 -62.875 55.840
452272 87.378 -62.869 55.828
452273 87.427 -62.865 55.817
452274 87.476 -62.860 55.806
452275 87.525 -62.858 55.798
452276 87.575 -62.866 55.788
452277 87.626 -62.876 55.776
452278 87.677 -62.885 55.760
452279 87.727 -62.894 55.744
452280 87.777 -62.903 55.726
452281 87.828 -62.911 55.709
452282 87.879 -62.920 55.693
452283 87.929 -62.929 55.679
452284 87.980 -62.938 55.664
452285 88.030 -62.948 55.650
452286 88.080 -62.956 55.635
452287 88.131 -62.965 55.619
452288 88.182 -62.974 55.601
452289 88.232 -62.983 55.583
452290 88.282 -62.992 55.565
452291 88.333 -63.001 55.545
452292 88.384 -63.010 55.525
452293 88.434 -63.019 55.505
452294 88.484 -63.028 55.483
452295 88.535 -63.036 55.465
452296 88.586 -63.045 55.447
452297 88.636 -63.054 55.429
452298 88.686 -63.064 55.414
452299 88.737 -63.073 55.401
452300 88.780 -63.087 55.389
452301 88.823 -63.102 55.380
452302 88.865 -63.118 55.374
452303 88.908 -63.133 55.368
452304 88.950 -63.148 55.366
452305 88.993 -63.164 55.365
452306 89.035 -63.179 55.364
452307 89.078 -63.194 55.363
452308 89.120 -63.209 55.364
452309 89.163 -63.225 55.366
452310 89.205 -63.240 55.366
452311 89.248 -63.255 55.366
452312 89.291 -63.270 55.366
452313 89.337 -63.278 55.369
452314 89.393 -63.266 55.373
452315 89.448 -63.257 55.381
452316 89.494 -63.268 55.393
452317 89.541 -63.280 55.407
452318 89.588 -63.291 55.418
452319 89.634 -63.305 55.425
452320 89.679 -63.326 55.425
452321 89.726 -63.337 55.419
452322 89.772 -63.346 55.407
452323 89.818 -63.350 55.391
452324 89.864 -63.351 55.372
452325 89.909 -63.352 55.353
452326 89.955 -63.354 55.336
452327 90.000 -63.355 55.319
452328 90.045 -63.356 55.301
452329 90.091 -63.358 55.291
452330 90.136 -63.359 55.284
452331 90.182 -63.360 55.281
452332 90.227 -63.361 55.283
452333 90.272 -63.363 55.293
452334 90.304 -63.371 55.309
452335 90.318 -63.390 55.327
452336 90.331 -63.408 55.341
452337 90.344 -63.425 55.346
452338 90.393 -63.437 55.343
452339 90.443 -63.449 55.337
452340 90.493 -63.461 55.330
452341 90.542 -63.472 55.322
452342 90.592 -63.484 55.312
452343 90.650 -63.493 55.292
452344 90.709 -63.502 55.263
452345 90.767 -63.511 55.225
452346 90.826 -63.520 55.179
452347 90.885 -63.529 55.126
452348 90.943 -63.538 55.069
452349 90.993 -63.546 55.016
452350 91.040 -63.555 54.967
452351 91.087 -63.564 54.926
452352 91.135 -63.572 54.888
452353 91.182 -63.581 54.854
452354 91.229 -63.590 54.826
452355 91.276 -63.598 54.802
452356 91.323 -63.607 54.781
452357 91.370 -63.615 54.765
452358 91.417 -63.624 54.751
452359 91.464 -63.633 54.740
452360 91.512 -63.641 54.731
452361 91.559 -63.650 54.723
452362 91.606 -63.659 54.717
452363 91.653 -63.667 54.712
452364 91.700 -63.676 54.708
452365 91.747 -63.684 54.704
452366 91.794 -63.694 54.700
452367 91.840 -63.697 54.698
452368 91.882 -63.697 54.697
452369 91.925 -63.697 54.696
452370 91.968 -63.697 54.696
452371 92.011 -63.697 54.693
452372 92.054 -63.697 54.688
452373 92.097 -63.697 54.682
452374 92.140 -63.696 54.675
452375 92.185 -63.699 54.669
452376 92.232 -63.704 54.661
452377 92.279 -63.709 54.656
452378 92.325 -63.715 54.651
452379 92.350 -63.720 54.646
452380 92.369 -63.725 54.643
452381 92.389 -63.731 54.639
452382 92.442 -63.736 54.635
452383 92.492 -63.747 54.631
452384 92.542 -63.758 54.626
452385 92.591 -63.769 54.624
452386 92.641 -63.780 54.621
452387 92.693 -63.791 54.619
452388 92.760 -63.791 54.619
452389 92.827 -63.791 54.619
452390 92.882 -63.796 54.618
452391 92.921 -63.805 54.618
452392 92.959 -63.814 54.617
452393 92.998 -63.823 54.616
452394 93.037 -63.827 54.615
452395 93.075 -63.827 54.613
452396 93.114 -63.827 54.611
452397 93.152 -63.827 54.609
452398 93.196 -63.825 54.606
452399 93.241 -63.822 54.601
452400 93.286 -63.819 54.596
452401 93.320 -63.816 54.590
452402 93.353 -63.813 54.582
452403 93.379 -63.820 54.572
452404 93.405 -63.827 54.559
452405 93.430 -63.834 54.545
452406 93.457 -63.841 54.529
452407 93.482 -63.847 54.510
452408 93.508 -63.854 54.488
452409 93.534 -63.861 54.464
452410 93.562 -63.872 54.439
452411 93.595 -63.889 54.412
452412 93.628 -63.907 54.385
452413 93.661 -63.923 54.357
452414 93.695 -63.940 54.329
452415 93.741 -63.950 54.298
452416 93.786 -63.960 54.270
452417 93.831 -63.969 54.243
452418 93.876 -63.979 54.216
452419 93.910 -63.986 54.191
452420 93.939 -63.991 54.168
452421 93.968 -63.998 54.144
452422 93.997 -64.003 54.119
452423 94.034 -64.014 54.096
452424 94.089 -64.040 54.073
452425 94.145 -64.066 54.053
452426 94.193 -64.081 54.032
452427 94.235 -64.091 54.013
452428 94.278 -64.101 53.994
452429 94.320 -64.111 53.977
452430 94.366 -64.129 53.958
452431 94.404 -64.136 53.939
452432 94.428 -64.128 53.922
452433 94.463 -64.127 53.906
452434 94.512 -64.136 53.888
452435 94.562 -64.146 53.870
452436 94.612 -64.155 53.852
452437 94.662 -64.165 53.834
452438 94.710 -64.170 53.814
452439 94.758 -64.170 53.794
452440 94.805 -64.170 53.774
452441 94.852 -64.169 53.753
452442 94.906 -64.176 53.731
452443 94.960 -64.182 53.711
452444 95.014 -64.189 53.691
452445 95.068 -64.197 53.672
452446 95.122 -64.204 53.653
452447 95.177 -64.211 53.633
452448 95.231 -64.218 53.614
452449 95.285 -64.225 53.595
452450 95.339 -64.232 53.576
452451 95.393 -64.239 53.558
452452 95.448 -64.246 53.539
452453 95.502 -64.254 53.519
452454 95.549 -64.261 53.498
452455 95.593 -64.267 53.474
452456 95.637 -64.273 53.450
452457 95.681 -64.280 53.425
452458 95.723 -64.284 53.400
452459 95.749 -64.274 53.372
452460 95.777 -64.266 53.344
452461 95.811 -64.269 53.316
452462 95.845 -64.272 53.289
452463 95.879 -64.275 53.264
452464 95.913 -64.278 53.237
452465 95.947 -64.281 53.208
452466 95.981 -64.284 53.176
452467 96.015 -64.287 53.143
452468 96.049 -64.289 53.109
452469 96.092 -64.270 53.077
452470 96.135 -64.258 53.049
452471 96.180 -64.249 53.025
452472 96.225 -64.242 53.003
452473 96.270 -64.234 52.983
452474 96.319 -64.235 52.961
452475 96.370 -64.244 52.937
452476 96.421 -64.254 52.909
452477 96.473 -64.263 52.879
452478 96.508 -64.265 52.847
452479 96.542 -64.267 52.814
452480 96.576 -64.270 52.781
452481 96.609 -64.271 52.749
452482 96.643 -64.273 52.717
452483 96.678 -64.276 52.681
452484 96.715 -64.281 52.641
452485 96.752 -64.286 52.597
452486 96.789 -64.290 52.549
452487 96.826 -64.295 52.499
452488 96.866 -64.299 52.450
452489 96.906 -64.303 52.401
452490 96.946 -64.307 52.351
452491 96.985 -64.311 52.303
452492 97.026 -64.314 52.261
452493 97.066 -64.318 52.222
452494 97.105 -64.322 52.183
452495 97.145 -64.326 52.141
452496 97.185 -64.330 52.096
452497 97.225 -64.333 52.049
452498 97.265 -64.337 52.001
452499 97.305 -64.341 51.953
452500 97.342 -64.340 51.906
452501 97.370 -64.330 51.859
452502 97.399 -64.319 51.814
452503 97.427 -64.320 51.771
452504 97.455 -64.327 51.728
452505 97.484 -64.335 51.684
452506 97.512 -64.342 51.639
452507 97.540 -64.349 51.592
452508 97.569 -64.356 51.546
452509 97.600 -64.361 51.498
452510 97.641 -64.357 51.450
452511 97.681 -64.352 51.403
452512 97.722 -64.349 51.357
452513 97.763 -64.344 51.311
452514 97.804 -64.340 51.265
452515 97.845 -64.336 51.218
452516 97.886 -64.331 51.171
452517 97.920 -64.330 51.123
452518 97.946 -64.334 51.073
452519 97.972 -64.339 51.022
452520 97.999 -64.343 50.970
452521 98.025 -64.349 50.918
452522 98.052 -64.353 50.865
452523 98.078 -64.358 50.814
452524 98.105 -64.362 50.763
452525 98.131 -64.367 50.714
452526 98.158 -64.371 50.665
452527 98.184 -64.377 50.618
452528 98.221 -64.383 50.570
452529 98.266 -64.386 50.522
452530 98.310 -64.389 50.475
452531 98.355 -64.393 50.427
452532 98.401 -64.394 50.379
452533 98.450 -64.388 50.332
452534 98.499 -64.382 50.286
452535 98.548 -64.376 50.241
452536 98.597 -64.369 50.196
452537 98.646 -64.363 50.151
452538 98.695 -64.357 50.107
452539 98.741 -64.358 50.063
452540 98.786 -64.361 50.020
452541 98.832 -64.364 49.975
452542 98.878 -64.367 49.931
452543 98.925 -64.370 49.886
452544 98.972 -64.372 49.840
452545 99.009 -64.376 49.795
452546 99.047 -64.380 49.751
452547 99.084 -64.383 49.707
452548 99.121 -64.386 49.663
452549 99.158 -64.390 49.619
452550 99.196 -64.394 49.577
452551 99.233 -64.397 49.536
452552 99.270 -64.401 49.495
452553 99.307 -64.405 49.454
452554 99.345 -64.408 49.413
452555 99.382 -64.411 49.374
452556 99.410 -64.417 49.334
452557 99.438 -64.423 49.296
452558 99.509 -64.428 49.256
452559 99.580 -64.433 49.216
452560 99.652 -64.439 49.179
452561 99.723 -64.444 49.141
452562 99.794 -64.449 49.103
452563 99.865 -64.454 49.065
452564 99.932 -64.455 49.028
452565 99.995 -64.454 48.990
452566 100.058 -64.452 48.954
452567 100.121 -64.450 48.918
452568 100.184 -64.448 48.882
452569 100.240 -64.448 48.846
452570 100.279 -64.452 48.810
452571 100.315 -64.459 48.774
452572 100.350 -64.467 48.739
452573 100.385 -64.474 48.705
452574 100.421 -64.482 48.670
452575 100.457 -64.490 48.635
452576 100.493 -64.497 48.601
452577 100.528 -64.505 48.565
452578 100.564 -64.512 48.531
452956 86.849 -62.747 56.079
452957 86.877 -62.753 56.066
452958 86.905 -62.760 56.052
452959 86.933 -62.766 56.038
452960 86.962 -62.772 56.025
452961 86.990 -62.779 56.013
452962 87.018 -62.785 55.999
452963 87.046 -62.792 55.985
452964 87.074 -62.798 55.970
452965 87.102 -62.805 55.954
452966 87.130 -62.812 55.939
452967 87.158 -62.818 55.924
452968 87.186 -62.825 55.909
452969 87.214 -62.831 55.893
452970 87.258 -62.829 55.877
452971 87.307 -62.825 55.862
452972 87.356 -62.819 55.847
452973 87.405 -62.815 55.833
452974 87.454 -62.810 55.821
452975 87.502 -62.806 55.812
452976 87.551 -62.801 55.801
452977 87.601 -62.805 55.792
452978 87.651 -62.813 55.780
452979 87.702 -62.822 55.765
452980 87.753 -62.832 55.748
452981 87.803 -62.841 55.731
452982 87.854 -62.850 55.715
452983 87.904 -62.858 55.701
452984 87.955 -62.867 55.689
452985 88.005 -62.876 55.675
452986 88.056 -62.885 55.662
452987 88.106 -62.894 55.647
452988 88.156 -62.903 55.631
452989 88.207 -62.912 55.613
452990 88.258 -62.921 55.594
452991 88.308 -62.930 55.576
452992 88.358 -62.939 55.556
452993 88.409 -62.948 55.536
452994 88.460 -62.957 55.515
452995 88.510 -62.966 55.493
452996 88.560 -62.975 55.475
452997 88.611 -62.983 55.457
452998 88.657 -62.996 55.439
452999 88.700 -63.011 55.422
453000 88.742 -63.026 55.408
453001 88.785 -63.042 55.396
453002 88.827 -63.057 55.386
453003 88.870 -63.073 55.379
453004 88.912 -63.088 55.373
453005 88.955 -63.103 55.369
453006 88.997 -63.118 55.368
453007 89.040 -63.133 55.367
453008 89.083 -63.148 55.365
453009 89.125 -63.164 55.366
453010 89.168 -63.180 55.367
453011 89.210 -63.195 55.367
453012 89.253 -63.210 55.366
453013 89.296 -63.225 55.364
453014 89.346 -63.224 55.365
453015 89.403 -63.211 55.368
453016 89.458 -63.202 55.373
453017 89.505 -63.213 55.383
453018 89.551 -63.224 55.396
453019 89.597 -63.238 55.407
453020 89.642 -63.258 55.414
453021 89.687 -63.278 55.415
453022 89.734 -63.289 55.410
453023 89.780 -63.299 55.399
453024 89.828 -63.308 55.384
453025 89.874 -63.311 55.368
453026 89.919 -63.312 55.352
453027 89.964 -63.314 55.335
453028 90.010 -63.315 55.319
453029 90.055 -63.316 55.300
453030 90.101 -63.318 55.287
453031 90.146 -63.319 55.279
453032 90.191 -63.321 55.274
453033 90.237 -63.321 55.274
453034 90.283 -63.323 55.281
453035 90.304 -63.337 55.294
453036 90.317 -63.355 55.311
453037 90.330 -63.373 55.326
453038 90.344 -63.391 55.332
453039 90.388 -63.403 55.330
453040 90.438 -63.415 55.326
453041 90.487 -63.427 55.320
453042 90.537 -63.438 55.311
453043 90.595 -63.448 55.296
453044 90.653 -63.456 55.272
453045 90.712 -63.465 55.242
453046 90.770 -63.474 55.205
453047 90.829 -63.484 55.158
453048 90.888 -63.493 55.106
453049 90.946 -63.502 55.054
453050 91.005 -63.511 55.005
453051 91.054 -63.519 54.960
453052 91.101 -63.528 54.920
453053 91.149 -63.537 54.885
453054 91.195 -63.545 54.853
453055 91.243 -63.554 54.825
453056 91.290 -63.562 54.802
453057 91.337 -63.571 54.782
453058 91.384 -63.580 54.766
453059 91.431 -63.588 54.753
453060 91.478 -63.597 54.742
453061 91.525 -63.606 54.734
453062 91.573 -63.614 54.726
453063 91.619 -63.623 54.721
453064 91.667 -63.631 54.716
453065 91.714 -63.640 54.711
453066 91.761 -63.649 54.708
453067 91.804 -63.650 54.705
453068 91.847 -63.650 54.705
453069 91.890 -63.649 54.706
453070 91.933 -63.649 54.706
453071 91.976 -63.649 54.705
453072 92.019 -63.648 54.702
453073 92.062 -63.648 54.697
453074 92.105 -63.648 54.690
453075 92.148 -63.647 54.682
453076 92.195 -63.653 54.675
453077 92.241 -63.658 54.668
453078 92.288 -63.663 54.662
453079 92.335 -63.669 54.657
453080 92.363 -63.674 54.653
453081 92.382 -63.679 54.649
453082 92.422 -63.683 54.644
453083 92.476 -63.687 54.640
453084 92.529 -63.692 54.636
453085 92.579 -63.703 54.631
453086 92.629 -63.714 54.629
453087 92.678 -63.725 54.626
453088 92.738 -63.730 54.624
453089 92.805 -63.731 54.624
453090 92.852 -63.737 54.623
453091 92.891 -63.747 54.623
453092 92.930 -63.756 54.622
453093 92.969 -63.765 54.621
453094 93.007 -63.774 54.620
453095 93.046 -63.774 54.618
453096 93.084 -63.774 54.617
453097 93.123 -63.774 54.614
453098 93.161 -63.774 54.610
453099 93.203 -63.772 54.606
453100 93.248 -63.770 54.601
453101 93.291 -63.767 54.596
453102 93.325 -63.764 54.589
453103 93.357 -63.762 54.579
453104 93.389 -63.761 54.567
453105 93.415 -63.768 54.554
453106 93.440 -63.775 54.540
453107 93.466 -63.781 54.523
453108 93.492 -63.788 54.503
453109 93.517 -63.795 54.481
453110 93.543 -63.802 54.456
453111 93.574 -63.816 54.430
453112 93.607 -63.833 54.402
453113 93.640 -63.850 54.375
453114 93.673 -63.867 54.346
453115 93.706 -63.885 54.315
453116 93.740 -63.902 54.284
453117 93.782 -63.913 54.255
453118 93.827 -63.922 54.226
453119 93.872 -63.932 54.198
453120 93.902 -63.938 54.169
453121 93.931 -63.944 54.144
453122 93.963 -63.951 54.119
453123 94.018 -63.977 54.095
453124 94.069 -63.997 54.072
453125 94.112 -64.007 54.050
453126 94.154 -64.016 54.030
453127 94.197 -64.027 54.009
453128 94.240 -64.037 53.989
453129 94.282 -64.047 53.970
453130 94.325 -64.058 53.952
453131 94.371 -64.076 53.936
453132 94.417 -64.093 53.921
453133 94.456 -64.103 53.906
453134 94.480 -64.095 53.892
453135 94.526 -64.101 53.875
453136 94.576 -64.110 53.858
453137 94.626 -64.120 53.840
453138 94.675 -64.128 53.821
453139 94.722 -64.128 53.802
453140 94.770 -64.127 53.782
453141 94.817 -64.127 53.761
453142 94.864 -64.126 53.740
453143 94.916 -64.131 53.720
453144 94.970 -64.139 53.700
453145 95.024 -64.145 53.680
453146 95.078 -64.152 53.660
453147 95.133 -64.160 53.641
453148 95.187 -64.167 53.621
453149 95.241 -64.173 53.601
453150 95.295 -64.181 53.581
453151 95.349 -64.188 53.562
453152 95.404 -64.195 53.544
453153 95.458 -64.202 53.527
453154 95.502 -64.209 53.508
453155 95.546 -64.216 53.488
453156 95.590 -64.222 53.463
453157 95.634 -64.229 53.436
453158 95.678 -64.236 53.410
453159 95.718 -64.238 53.383
453160 95.745 -64.228 53.357
453161 95.771 -64.217 53.331
453162 95.804 -64.217 53.304
453163 95.837 -64.220 53.279
453164 95.872 -64.223 53.256
453165 95.905 -64.226 53.231
453166 95.939 -64.229 53.203
453167 95.973 -64.233 53.173
453168 96.008 -64.235 53.140
453169 96.049 -64.217 53.104
453170 96.092 -64.199 53.073
453171 96.136 -64.191 53.046
453172 96.182 -64.183 53.023
453173 96.227 -64.176 53.003
453174 96.276 -64.179 52.983
453175 96.327 -64.188 52.961
453176 96.379 -64.197 52.937
453177 96.431 -64.207 52.910
453178 96.482 -64.216 52.881
453179 96.519 -64.220 52.850
453180 96.553 -64.221 52.816
453181 96.587 -64.223 52.781
453182 96.620 -64.226 52.746
453183 96.654 -64.227 52.710
453184 96.689 -64.231 52.673
453185 96.727 -64.236 52.631
453186 96.763 -64.240 52.586
453187 96.801 -64.245 52.537
453188 96.838 -64.250 52.487
453189 96.875 -64.255 52.438
453190 96.912 -64.259 52.389
453191 96.951 -64.263 52.341
453192 96.992 -64.267 52.294
453193 97.031 -64.270 52.251
453194 97.071 -64.274 52.212
453195 97.111 -64.279 52.173
453196 97.151 -64.283 52.131
453197 97.191 -64.286 52.087
453198 97.231 -64.290 52.039
453199 97.270 -64.294 51.991
453200 97.311 -64.298 51.943
453201 97.349 -64.299 51.897
453202 97.377 -64.289 51.851
453203 97.406 -64.277 51.807
453204 97.435 -64.284 51.766
453205 97.463 -64.291 51.724
453206 97.491 -64.298 51.680
453207 97.519 -64.305 51.635
453208 97.548 -64.312 51.588
453209 97.576 -64.319 51.542
453210 97.604 -64.327 51.494
453211 97.641 -64.327 51.446
453212 97.682 -64.322 51.399
453213 97.723 -64.318 51.353
453214 97.764 -64.314 51.307
453215 97.804 -64.310 51.261
453216 97.845 -64.306 51.214
453217 97.887 -64.298 51.166
453218 97.920 -64.298 51.117
453219 97.946 -64.303 51.067
453220 97.972 -64.308 51.016
453221 97.999 -64.312 50.964
453222 98.025 -64.317 50.912
453223 98.052 -64.321 50.859
453224 98.078 -64.326 50.807
453225 98.104 -64.330 50.755
453226 98.131 -64.336 50.705
453227 98.157 -64.340 50.656
453228 98.189 -64.342 50.607
453229 98.227 -64.337 50.559
453230 98.270 -64.336 50.511
453231 98.314 -64.339 50.463
453232 98.359 -64.343 50.415
453233 98.406 -64.342 50.367
453234 98.454 -64.336 50.320
453235 98.504 -64.330 50.273
453236 98.553 -64.323 50.227
453237 98.601 -64.317 50.182
453238 98.647 -64.320 50.137
453239 98.693 -64.323 50.092
453240 98.738 -64.326 50.049
453241 98.784 -64.329 50.005
453242 98.830 -64.332 49.960
453243 98.876 -64.334 49.916
453244 98.923 -64.336 49.871
453245 98.970 -64.339 49.825
453246 99.007 -64.342 49.779
453247 99.044 -64.346 49.734
453248 99.081 -64.350 49.691
453249 99.119 -64.354 49.647
453250 99.156 -64.357 49.603
453251 99.193 -64.361 49.561
453252 99.230 -64.364 49.519
453253 99.268 -64.368 49.478
453254 99.305 -64.372 49.437
453255 99.343 -64.375 49.397
453256 99.365 -64.381 49.358
453257 99.385 -64.387 49.319
453258 99.456 -64.392 49.280
453259 99.528 -64.398 49.241
453260 99.599 -64.403 49.202
453261 99.670 -64.408 49.164
453262 99.742 -64.413 49.127
453263 99.812 -64.418 49.089
453264 99.877 -64.418 49.051
453265 99.940 -64.417 49.014
453266 100.003 -64.414 48.976
453267 100.066 -64.413 48.940
453268 100.129 -64.411 48.904
453269 100.192 -64.409 48.869
453270 100.238 -64.411 48.834
453271 100.279 -64.415 48.797
453272 100.315 -64.423 48.762
453273 100.351 -64.430 48.728
453274 100.387 -64.438 48.694
453275 100.422 -64.446 48.660
453276 100.458 -64.452 48.626
453277 100.494 -64.460 48.592
453278 100.529 -64.468 48.558
453279 100.565 -64.475 48.524
453657 86.853 -62.704 56.091
453658 86.881 -62.710 56.078
453659 86.909 -62.717 56.064
453660 86.937 -62.723 56.051
453661 86.966 -62.730 56.038
453662 86.994 -62.737 56.026
453663 87.022 -62.743 56.012
453664 87.050 -62.750 55.998
453665 87.078 -62.756 55.983
453666 87.106 -62.763 55.966
453667 87.134 -62.769 55.950
453668 87.162 -62.775 55.935
453669 87.190 -62.782 55.919
453670 87.235 -62.779 55.903
453671 87.284 -62.775 55.886
453672 87.333 -62.770 55.870
453673 87.382 -62.765 55.854
453674 87.431 -62.760 55.840
453675 87.480 -62.756 55.826
453676 87.529 -62.751 55.815
453677 87.578 -62.747 55.804
453678 87.627 -62.743 55.793
453679 87.677 -62.752 55.780
453680 87.727 -62.760 55.765
453681 87.778 -62.769 55.750
453682 87.829 -62.778 55.735
453683 87.879 -62.788 55.722
453684 87.930 -62.797 55.711
453685 87.980 -62.805 55.702
453686 88.031 -62.814 55.689
453687 88.081 -62.823 55.675
453688 88.132 -62.832 55.661
453689 88.182 -62.841 55.644
453690 88.233 -62.850 55.626
453691 88.283 -62.859 55.606
453692 88.334 -62.868 55.587
453693 88.384 -62.877 55.567
453694 88.434 -62.886 55.547
453695 88.485 -62.894 55.525
453696 88.535 -62.904 55.504
453697 88.577 -62.919 55.484
453698 88.620 -62.935 55.466
453699 88.662 -62.951 55.447
453700 88.705 -62.966 55.430
453701 88.747 -62.981 55.415
453702 88.790 -62.996 55.402
453703 88.832 -63.011 55.391
453704 88.875 -63.026 55.383
453705 88.917 -63.042 55.377
453706 88.960 -63.057 55.373
453707 89.002 -63.073 55.370
453708 89.045 -63.088 55.368
453709 89.088 -63.103 55.367
453710 89.130 -63.118 55.368
453711 89.173 -63.134 55.367
453712 89.215 -63.149 55.367
453713 89.258 -63.164 55.366
453714 89.301 -63.180 55.363
453715 89.356 -63.169 55.361
453716 89.412 -63.156 55.362
453717 89.467 -63.147 55.366
453718 89.514 -63.158 55.373
453719 89.560 -63.170 55.385
453720 89.605 -63.190 55.395
453721 89.650 -63.210 55.402
453722 89.695 -63.230 55.404
453723 89.742 -63.242 55.399
453724 89.789 -63.251 55.391
453725 89.836 -63.260 55.378
453726 89.883 -63.269 55.364
453727 89.929 -63.272 55.349
453728 89.974 -63.274 55.335
453729 90.020 -63.275 55.317
453730 90.065 -63.277 55.297
453731 90.110 -63.277 55.282
453732 90.156 -63.279 55.271
453733 90.201 -63.280 55.264
453734 90.247 -63.282 55.262
453735 90.290 -63.284 55.266
453736 90.303 -63.302 55.276
453737 90.317 -63.321 55.290
453738 90.330 -63.338 55.303
453739 90.343 -63.356 55.307
453740 90.383 -63.370 55.307
453741 90.433 -63.381 55.306
453742 90.482 -63.393 55.302
453743 90.539 -63.402 55.290
453744 90.598 -63.412 55.274
453745 90.656 -63.421 55.250
453746 90.715 -63.430 55.219
453747 90.773 -63.439 55.181
453748 90.832 -63.447 55.135
453749 90.891 -63.456 55.086
453750 90.950 -63.465 55.038
453751 91.008 -63.474 54.993
453752 91.067 -63.484 54.953
453753 91.115 -63.492 54.915
453754 91.162 -63.501 54.881
453755 91.210 -63.509 54.851
453756 91.256 -63.518 54.825
453757 91.304 -63.527 54.802
453758 91.351 -63.535 54.784
453759 91.398 -63.544 54.769
453760 91.445 -63.553 54.757
453761 91.492 -63.562 54.746
453762 91.539 -63.570 54.738
453763 91.586 -63.578 54.731
453764 91.634 -63.587 54.726
453765 91.680 -63.596 54.721
453766 91.726 -63.602 54.717
453767 91.770 -63.601 54.715
453768 91.812 -63.601 54.712
453769 91.855 -63.601 54.714
453770 91.898 -63.600 54.716
453771 91.941 -63.600 54.717
453772 91.984 -63.600 54.716
453773 92.027 -63.600 54.710
453774 92.070 -63.600 54.705
453775 92.113 -63.600 54.698
453776 92.157 -63.601 54.690
453777 92.204 -63.606 54.682
453778 92.250 -63.612 54.675
453779 92.297 -63.617 54.668
453780 92.344 -63.622 54.663
453781 92.375 -63.628 54.659
453782 92.402 -63.633 54.655
453783 92.455 -63.636 54.650
453784 92.509 -63.640 54.645
453785 92.563 -63.643 54.641
453786 92.616 -63.648 54.637
453787 92.665 -63.659 54.634
453788 92.717 -63.669 54.631
453789 92.784 -63.670 54.630
453790 92.823 -63.679 54.628
453791 92.861 -63.688 54.627
453792 92.900 -63.697 54.627
453793 92.939 -63.706 54.626
453794 92.978 -63.715 54.625
453795 93.016 -63.721 54.623
453796 93.055 -63.721 54.621
453797 93.093 -63.721 54.618
453798 93.132 -63.721 54.614
453799 93.170 -63.721 54.609
453800 93.209 -63.721 54.604
453801 93.254 -63.718 54.599
453802 93.295 -63.715 54.591
453803 93.329 -63.712 54.582
453804 93.362 -63.709 54.571
453805 93.395 -63.706 54.558
453806 93.424 -63.709 54.545
453807 93.450 -63.715 54.529
453808 93.476 -63.722 54.511
453809 93.501 -63.729 54.491
453810 93.527 -63.736 54.469
453811 93.553 -63.743 54.445
453812 93.585 -63.759 54.418
453813 93.619 -63.777 54.391
453814 93.652 -63.794 54.362
453815 93.685 -63.811 54.333
453816 93.718 -63.828 54.302
453817 93.751 -63.846 54.272
453818 93.784 -63.863 54.241
453819 93.824 -63.876 54.210
453820 93.865 -63.885 54.180
453821 93.894 -63.891 54.150
453822 93.946 -63.913 54.122
453823 93.988 -63.922 54.096
453824 94.031 -63.932 54.071
453825 94.073 -63.943 54.049
453826 94.116 -63.953 54.028
453827 94.158 -63.963 54.007
453828 94.201 -63.973 53.987
453829 94.244 -63.983 53.966
453830 94.286 -63.993 53.947
453831 94.330 -64.005 53.929
453832 94.376 -64.023 53.918
453833 94.422 -64.040 53.905
453834 94.468 -64.057 53.893
453835 94.509 -64.070 53.880
453836 94.540 -64.066 53.864
453837 94.590 -64.076 53.848
453838 94.639 -64.085 53.830
453839 94.687 -64.086 53.811
453840 94.734 -64.085 53.791
453841 94.781 -64.085 53.771
453842 94.829 -64.085 53.750
453843 94.876 -64.084 53.729
453844 94.926 -64.087 53.709
453845 94.980 -64.094 53.689
453846 95.035 -64.101 53.669
453847 95.089 -64.108 53.650
453848 95.143 -64.115 53.631
453849 95.197 -64.123 53.610
453850 95.251 -64.129 53.590
453851 95.306 -64.136 53.570
453852 95.359 -64.144 53.551
453853 95.411 -64.151 53.532
453854 95.456 -64.157 53.514
453855 95.499 -64.164 53.495
453856 95.543 -64.170 53.474
453857 95.588 -64.177 53.450
453858 95.631 -64.184 53.424
453859 95.675 -64.191 53.397
453860 95.713 -64.192 53.370
453861 95.740 -64.181 53.344
453862 95.767 -64.171 53.318
453863 95.796 -64.165 53.292
453864 95.830 -64.168 53.268
453865 95.864 -64.171 53.246
453866 95.898 -64.174 53.222
453867 95.932 -64.177 53.196
453868 95.966 -64.180 53.167
453869 96.008 -64.163 53.136
453870 96.049 -64.145 53.101
453871 96.093 -64.132 53.069
453872 96.138 -64.125 53.041
453873 96.183 -64.117 53.018
453874 96.234 -64.123 52.998
453875 96.285 -64.132 52.979
453876 96.336 -64.141 52.958
453877 96.388 -64.151 52.935
453878 96.439 -64.160 52.910
453879 96.491 -64.169 52.881
453880 96.530 -64.173 52.849
453881 96.564 -64.176 52.814
453882 96.598 -64.177 52.778
453883 96.631 -64.179 52.741
453884 96.665 -64.182 52.702
453885 96.701 -64.186 52.662
453886 96.738 -64.190 52.618
453887 96.776 -64.195 52.572
453888 96.812 -64.199 52.523
453889 96.850 -64.204 52.474
453890 96.887 -64.209 52.426
453891 96.924 -64.214 52.378
453892 96.961 -64.218 52.331
453893 96.998 -64.223 52.285
453894 97.037 -64.227 52.242
453895 97.077 -64.231 52.201
453896 97.117 -64.235 52.162
453897 97.157 -64.239 52.121
453898 97.196 -64.242 52.077
453899 97.236 -64.246 52.029
453900 97.277 -64.250 51.981
453901 97.316 -64.254 51.933
453902 97.356 -64.258 51.888
453903 97.385 -64.247 51.844
453904 97.414 -64.240 51.802
453905 97.442 -64.248 51.761
453906 97.470 -64.255 51.719
453907 97.498 -64.261 51.675
453908 97.527 -64.269 51.630
453909 97.555 -64.276 51.585
453910 97.583 -64.283 51.538
453911 97.611 -64.290 51.490
453912 97.641 -64.296 51.442
453913 97.682 -64.292 51.395
453914 97.723 -64.288 51.349
453915 97.764 -64.283 51.303
453916 97.805 -64.280 51.257
453917 97.846 -64.274 51.210
453918 97.888 -64.266 51.161
453919 97.919 -64.267 51.112
453920 97.945 -64.271 51.062
453921 97.972 -64.276 51.011
453922 97.998 -64.280 50.958
453923 98.025 -64.286 50.905
453924 98.051 -64.290 50.852
453925 98.078 -64.295 50.799
453926 98.104 -64.299 50.747
453927 98.131 -64.304 50.696
453928 98.164 -64.305 50.647
453929 98.200 -64.306 50.596
453930 98.237 -64.301 50.548
453931 98.277 -64.292 50.500
453932 98.318 -64.289 50.452
453933 98.363 -64.293 50.404
453934 98.411 -64.289 50.356
453935 98.459 -64.283 50.308
453936 98.508 -64.279 50.261
453937 98.554 -64.282 50.214
453938 98.599 -64.285 50.168
453939 98.645 -64.288 50.123
453940 98.691 -64.290 50.079
453941 98.736 -64.293 50.034
453942 98.782 -64.296 49.990
453943 98.829 -64.299 49.945
453944 98.875 -64.302 49.902
453945 98.922 -64.304 49.856
453946 98.967 -64.306 49.810
453947 99.004 -64.310 49.764
453948 99.042 -64.314 49.719
453949 99.079 -64.317 49.675
453950 99.116 -64.320 49.632
453951 99.153 -64.324 49.588
453952 99.191 -64.328 49.545
453953 99.228 -64.332 49.503
453954 99.266 -64.335 49.462
453955 99.300 -64.339 49.422
453956 99.320 -64.345 49.382
453957 99.339 -64.352 49.343
453958 99.404 -64.357 49.305
453959 99.475 -64.362 49.265
453960 99.546 -64.367 49.226
453961 99.618 -64.373 49.188
453962 99.689 -64.377 49.150
453963 99.760 -64.383 49.113
453964 99.823 -64.381 49.075
453965 99.886 -64.380 49.037
453966 99.949 -64.378 48.999
453967 100.012 -64.376 48.962
453968 100.075 -64.374 48.927
453969 100.138 -64.372 48.891
453970 100.201 -64.371 48.856
453971 100.243 -64.372 48.821
453972 100.280 -64.379 48.785
453973 100.316 -64.386 48.750
453974 100.352 -64.393 48.717
453975 100.388 -64.401 48.684
453976 100.423 -64.408 48.651
453977 100.459 -64.416 48.616
453978 100.494 -64.424 48.583
453979 100.530 -64.431 48.550
454357 86.793 -62.675 56.115
454358 86.836 -62.674 56.102
454359 86.878 -62.672 56.089
454360 86.913 -62.675 56.077
454361 86.941 -62.681 56.064
454362 86.970 -62.688 56.051
454363 86.998 -62.694 56.039
454364 87.025 -62.700 56.025
454365 87.054 -62.707 56.010
454366 87.082 -62.713 55.994
454367 87.110 -62.720 55.979
454368 87.138 -62.726 55.963
454369 87.166 -62.733 55.947
454370 87.213 -62.729 55.931
454371 87.262 -62.725 55.914
454372 87.311 -62.720 55.896
454373 87.360 -62.716 55.878
454374 87.409 -62.710 55.862
454375 87.458 -62.706 55.846
454376 87.506 -62.701 55.831
454377 87.555 -62.697 55.818
454378 87.604 -62.692 55.806
454379 87.653 -62.687 55.793
454380 87.703 -62.690 55.779
454381 87.753 -62.699 55.765
454382 87.804 -62.707 55.752
454383 87.854 -62.716 55.740
454384 87.905 -62.725 55.730
454385 87.955 -62.735 55.722
454386 88.006 -62.744 55.716
454387 88.056 -62.752 55.705
454388 88.107 -62.761 55.693
454389 88.157 -62.770 55.678
454390 88.208 -62.779 55.661
454391 88.258 -62.788 55.640
454392 88.309 -62.797 55.620
454393 88.359 -62.806 55.600
454394 88.410 -62.815 55.579
454395 88.454 -62.829 55.558
454396 88.497 -62.844 55.536
454397 88.539 -62.859 55.514
454398 88.582 -62.874 55.494
454399 88.624 -62.889 55.474
454400 88.667 -62.905 55.456
454401 88.709 -62.920 55.438
454402 88.752 -62.935 55.422
454403 88.794 -62.951 55.408
454404 88.837 -62.966 55.395
454405 88.880 -62.981 55.386
454406 88.922 -62.997 55.380
454407 88.965 -63.012 55.376
454408 89.008 -63.027 55.372
454409 89.050 -63.042 55.370
454410 89.093 -63.057 55.368
454411 89.135 -63.073 55.368
454412 89.178 -63.088 55.368
454413 89.220 -63.104 55.367
454414 89.263 -63.119 55.364
454415 89.309 -63.126 55.360
454416 89.366 -63.114 55.357
454417 89.422 -63.101 55.356
454418 89.477 -63.092 55.358
454419 89.524 -63.103 55.362
454420 89.569 -63.122 55.372
454421 89.614 -63.142 55.383
454422 89.659 -63.162 55.390
454423 89.703 -63.183 55.391
454424 89.750 -63.193 55.388
454425 89.797 -63.203 55.382
454426 89.844 -63.212 55.371
454427 89.891 -63.221 55.359
454428 89.938 -63.230 55.346
454429 89.984 -63.233 55.334
454430 90.029 -63.235 55.315
454431 90.075 -63.236 55.294
454432 90.120 -63.238 55.276
454433 90.166 -63.239 55.263
454434 90.211 -63.240 55.254
454435 90.256 -63.242 55.250
454436 90.289 -63.250 55.250
454437 90.303 -63.267 55.256
454438 90.316 -63.286 55.265
454439 90.329 -63.304 55.274
454440 90.343 -63.321 55.278
454441 90.378 -63.336 55.280
454442 90.428 -63.347 55.280
454443 90.483 -63.357 55.274
454444 90.542 -63.366 55.263
454445 90.601 -63.375 55.246
454446 90.659 -63.384 55.224
454447 90.718 -63.393 55.195
454448 90.777 -63.402 55.159
454449 90.835 -63.412 55.114
454450 90.894 -63.421 55.068
454451 90.953 -63.430 55.024
454452 91.011 -63.438 54.983
454453 91.070 -63.447 54.945
454454 91.129 -63.456 54.909
454455 91.176 -63.465 54.877
454456 91.223 -63.474 54.848
454457 91.271 -63.482 54.824
454458 91.317 -63.491 54.804
454459 91.365 -63.499 54.788
454460 91.412 -63.509 54.773
454461 91.459 -63.517 54.761
454462 91.506 -63.525 54.751
454463 91.553 -63.534 54.743
454464 91.600 -63.543 54.736
454465 91.647 -63.551 54.731
454466 91.692 -63.553 54.727
454467 91.734 -63.553 54.724
454468 91.777 -63.553 54.721
454469 91.820 -63.553 54.720
454470 91.863 -63.553 54.723
454471 91.906 -63.553 54.725
454472 91.949 -63.552 54.726
454473 91.992 -63.552 54.725
454474 92.035 -63.552 54.718
454475 92.078 -63.551 54.711
454476 92.121 -63.551 54.704
454477 92.166 -63.555 54.696
454478 92.213 -63.560 54.687
454479 92.260 -63.565 54.680
454480 92.306 -63.571 54.674
454481 92.353 -63.576 54.669
454482 92.388 -63.581 54.665
454483 92.435 -63.585 54.661
454484 92.488 -63.589 54.656
454485 92.542 -63.593 54.651
454486 92.596 -63.596 54.646
454487 92.649 -63.600 54.641
454488 92.703 -63.604 54.638
454489 92.754 -63.612 54.636
454490 92.793 -63.621 54.634
454491 92.832 -63.630 54.632
454492 92.870 -63.639 54.631
454493 92.910 -63.648 54.630
454494 92.948 -63.658 54.629
454495 92.987 -63.667 54.627
454496 93.025 -63.668 54.625
454497 93.064 -63.668 54.622
454498 93.102 -63.668 54.617
454499 93.141 -63.668 54.612
454500 93.179 -63.668 54.607
454501 93.217 -63.668 54.601
454502 93.261 -63.665 54.594
454503 93.300 -63.663 54.585
454504 93.333 -63.660 54.574
454505 93.366 -63.657 54.561
454506 93.399 -63.654 54.547
454507 93.432 -63.652 54.531
454508 93.460 -63.656 54.514
454509 93.485 -63.663 54.496
454510 93.511 -63.670 54.475
454511 93.537 -63.677 54.454
454512 93.564 -63.687 54.429
454513 93.597 -63.703 54.403
454514 93.630 -63.721 54.376
454515 93.663 -63.737 54.347
454516 93.697 -63.755 54.318
454517 93.730 -63.772 54.288
454518 93.763 -63.789 54.257
454519 93.796 -63.806 54.226
454520 93.829 -63.824 54.195
454521 93.866 -63.839 54.164
454522 93.908 -63.848 54.134
454523 93.951 -63.858 54.104
454524 93.993 -63.869 54.076
454525 94.036 -63.878 54.051
454526 94.078 -63.888 54.028
454527 94.121 -63.899 54.006
454528 94.163 -63.909 53.986
454529 94.206 -63.919 53.966
454530 94.248 -63.929 53.947
454531 94.291 -63.939 53.928
454532 94.335 -63.952 53.913
454533 94.381 -63.969 53.903
454534 94.426 -63.987 53.893
454535 94.472 -64.004 53.882
454536 94.518 -64.022 53.870
454537 94.561 -64.036 53.857
454538 94.603 -64.041 53.841
454539 94.651 -64.044 53.823
454540 94.699 -64.044 53.803
454541 94.746 -64.043 53.783
454542 94.793 -64.043 53.762
454543 94.841 -64.042 53.741
454544 94.888 -64.042 53.720
454545 94.936 -64.042 53.699
454546 94.990 -64.050 53.680
454547 95.045 -64.057 53.660
454548 95.099 -64.064 53.642
454549 95.153 -64.071 53.623
454550 95.207 -64.078 53.601
454551 95.261 -64.085 53.581
454552 95.316 -64.092 53.561
454553 95.365 -64.099 53.542
454554 95.409 -64.106 53.522
454555 95.453 -64.113 53.503
454556 95.496 -64.120 53.483
454557 95.541 -64.126 53.461
454558 95.585 -64.132 53.438
454559 95.628 -64.139 53.412
454560 95.673 -64.146 53.386
454561 95.709 -64.145 53.359
454562 95.736 -64.135 53.332
454563 95.763 -64.124 53.306
454564 95.790 -64.114 53.280
454565 95.823 -64.116 53.256
454566 95.856 -64.119 53.234
454567 95.890 -64.122 53.212
454568 95.924 -64.125 53.188
454569 95.965 -64.110 53.160
454570 96.008 -64.092 53.130
454571 96.050 -64.074 53.096
454572 96.095 -64.066 53.063
454573 96.140 -64.058 53.035
454574 96.191 -64.067 53.011
454575 96.242 -64.076 52.991
454576 96.294 -64.085 52.974
454577 96.345 -64.095 52.953
454578 96.397 -64.104 52.930
454579 96.448 -64.113 52.905
454580 96.500 -64.122 52.876
454581 96.541 -64.127 52.842
454582 96.575 -64.129 52.806
454583 96.609 -64.132 52.768
454584 96.643 -64.134 52.730
454585 96.676 -64.135 52.690
454586 96.713 -64.140 52.650
454587 96.750 -64.145 52.605
454588 96.787 -64.149 52.558
454589 96.824 -64.154 52.510
454590 96.861 -64.159 52.461
454591 96.898 -64.164 52.413
454592 96.936 -64.168 52.366
454593 96.972 -64.173 52.319
454594 97.010 -64.178 52.275
454595 97.047 -64.182 52.233
454596 97.084 -64.187 52.192
454597 97.122 -64.192 52.151
454598 97.162 -64.195 52.110
454599 97.202 -64.199 52.065
454600 97.242 -64.203 52.017
454601 97.282 -64.207 51.969
454602 97.322 -64.211 51.923
454603 97.362 -64.214 51.879
454604 97.392 -64.206 51.837
454605 97.421 -64.204 51.797
454606 97.449 -64.211 51.756
454607 97.477 -64.218 51.714
454608 97.506 -64.225 51.670
454609 97.534 -64.232 51.626
454610 97.562 -64.239 51.580
454611 97.590 -64.246 51.534
454612 97.619 -64.253 51.487
454613 97.647 -64.261 51.439
454614 97.682 -64.261 51.391
454615 97.723 -64.258 51.345
454616 97.764 -64.253 51.299
454617 97.805 -64.249 51.252
454618 97.847 -64.242 51.205
454619 97.889 -64.233 51.157
454620 97.919 -64.235 51.109
454621 97.945 -64.240 51.059
454622 97.971 -64.245 51.006
454623 97.998 -64.249 50.953
454624 98.024 -64.254 50.898
454625 98.051 -64.258 50.845
454626 98.077 -64.263 50.792
454627 98.104 -64.267 50.740
454628 98.139 -64.269 50.689
454629 98.174 -64.270 50.638
454630 98.210 -64.270 50.587
454631 98.248 -64.265 50.538
454632 98.287 -64.256 50.489
454633 98.326 -64.248 50.441
454634 98.367 -64.243 50.394
454635 98.414 -64.241 50.346
454636 98.460 -64.243 50.298
454637 98.506 -64.246 50.250
454638 98.552 -64.249 50.202
454639 98.597 -64.252 50.155
454640 98.643 -64.255 50.111
454641 98.689 -64.258 50.066
454642 98.734 -64.261 50.021
454643 98.780 -64.264 49.976
454644 98.827 -64.266 49.931
454645 98.874 -64.268 49.888
454646 98.921 -64.270 49.842
454647 98.964 -64.273 49.796
454648 99.002 -64.277 49.750
454649 99.039 -64.280 49.706
454650 99.076 -64.284 49.662
454651 99.114 -64.288 49.618
454652 99.151 -64.292 49.574
454653 99.189 -64.295 49.531
454654 99.226 -64.298 49.488
454655 99.256 -64.303 49.447
454656 99.275 -64.309 49.407
454657 99.294 -64.316 49.366
454658 99.351 -64.321 49.327
454659 99.422 -64.327 49.290
454660 99.494 -64.332 49.251
454661 99.565 -64.337 49.211
454662 99.636 -64.342 49.173
454663 99.706 -64.346 49.135
454664 99.769 -64.344 49.098
454665 99.832 -64.342 49.060
454666 99.895 -64.341 49.023
454667 99.958 -64.339 48.985
454668 100.020 -64.337 48.948
454669 100.083 -64.335 48.913
454670 100.146 -64.333 48.878
454671 100.206 -64.332 48.843
454672 100.246 -64.334 48.808
454673 100.281 -64.342 48.773
454674 100.317 -64.349 48.739
454675 100.353 -64.357 48.706
454676 100.388 -64.364 48.674
454677 100.424 -64.372 48.641
455058 86.773 -62.647 56.126
455059 86.815 -62.644 56.112
455060 86.858 -62.643 56.100
455061 86.901 -62.641 56.088
455062 86.944 -62.639 56.076
455063 86.974 -62.645 56.064
455064 87.001 -62.652 56.051
455065 87.029 -62.658 56.037
455066 87.058 -62.665 56.022
455067 87.086 -62.671 56.006
455068 87.114 -62.678 55.991
455069 87.142 -62.684 55.976
455070 87.191 -62.679 55.960
455071 87.239 -62.675 55.942
455072 87.288 -62.670 55.924
455073 87.337 -62.666 55.906
455074 87.386 -62.661 55.888
455075 87.435 -62.656 55.870
455076 87.484 -62.651 55.854
455077 87.533 -62.647 55.837
455078 87.582 -62.642 55.821
455079 87.631 -62.637 55.806
455080 87.680 -62.632 55.793
455081 87.728 -62.628 55.780
455082 87.779 -62.637 55.766
455083 87.829 -62.646 55.755
455084 87.880 -62.654 55.746
455085 87.930 -62.663 55.738
455086 87.981 -62.672 55.732
455087 88.031 -62.681 55.728
455088 88.082 -62.691 55.721
455089 88.132 -62.699 55.710
455090 88.183 -62.708 55.696
455091 88.234 -62.717 55.678
455092 88.284 -62.726 55.655
455093 88.332 -62.737 55.634
455094 88.374 -62.753 55.612
455095 88.417 -62.768 55.589
455096 88.459 -62.783 55.568
455097 88.502 -62.798 55.545
455098 88.544 -62.813 55.523
455099 88.587 -62.829 55.502
455100 88.629 -62.844 55.482
455101 88.672 -62.860 55.463
455102 88.714 -62.875 55.444
455103 88.757 -62.890 55.428
455104 88.800 -62.905 55.414
455105 88.842 -62.920 55.401
455106 88.885 -62.935 55.390
455107 88.927 -62.951 55.383
455108 88.970 -62.967 55.378
455109 89.013 -62.982 55.374
455110 89.055 -62.997 55.371
455111 89.098 -63.012 55.368
455112 89.140 -63.027 55.368
455113 89.183 -63.042 55.368
455114 89.225 -63.058 55.366
455115 89.268 -63.073 55.363
455116 89.319 -63.071 55.357
455117 89.375 -63.059 55.353
455118 89.431 -63.046 55.350
455119 89.487 -63.036 55.349
455120 89.532 -63.054 55.352
455121 89.577 -63.074 55.358
455122 89.622 -63.095 55.368
455123 89.667 -63.114 55.375
455124 89.711 -63.135 55.378
455125 89.758 -63.145 55.376
455126 89.805 -63.155 55.371
455127 89.852 -63.164 55.362
455128 89.899 -63.173 55.351
455129 89.946 -63.183 55.341
455130 89.993 -63.192 55.329
455131 90.039 -63.195 55.312
455132 90.085 -63.196 55.290
455133 90.130 -63.198 55.271
455134 90.175 -63.199 55.255
455135 90.221 -63.200 55.245
455136 90.266 -63.202 55.238
455137 90.289 -63.215 55.235
455138 90.302 -63.233 55.235
455139 90.315 -63.251 55.239
455140 90.329 -63.269 55.244
455141 90.342 -63.287 55.247
455142 90.373 -63.302 55.251
455143 90.428 -63.312 55.251
455144 90.486 -63.321 55.245
455145 90.545 -63.330 55.234
455146 90.604 -63.339 55.219
455147 90.662 -63.348 55.198
455148 90.721 -63.357 55.170
455149 90.780 -63.366 55.135
455150 90.838 -63.375 55.092
455151 90.897 -63.384 55.049
455152 90.956 -63.393 55.009
455153 91.014 -63.402 54.972
455154 91.073 -63.412 54.937
455155 91.132 -63.421 54.904
455156 91.190 -63.430 54.874
455157 91.237 -63.438 54.848
455158 91.284 -63.446 54.826
455159 91.331 -63.455 54.806
455160 91.378 -63.464 54.792
455161 91.426 -63.472 54.778
455162 91.472 -63.481 54.767
455163 91.520 -63.490 54.757
455164 91.567 -63.499 54.749
455165 91.613 -63.506 54.743
455166 91.656 -63.506 54.737
455167 91.699 -63.505 54.733
455168 91.742 -63.505 54.731
455169 91.785 -63.505 54.728
455170 91.828 -63.504 54.729
455171 91.871 -63.504 54.732
455172 91.914 -63.504 54.735
455173 91.957 -63.504 54.735
455174 92.000 -63.503 54.733
455175 92.043 -63.503 54.726
455176 92.086 -63.503 54.718
455177 92.129 -63.503 54.710
455178 92.176 -63.509 54.702
455179 92.222 -63.514 54.692
455180 92.269 -63.519 54.685
455181 92.316 -63.524 54.680
455182 92.362 -63.530 54.675
455183 92.410 -63.535 54.670
455184 92.461 -63.539 54.666
455185 92.513 -63.543 54.660
455186 92.564 -63.546 54.656
455187 92.616 -63.551 54.651
455188 92.667 -63.555 54.647
455189 92.718 -63.559 54.643
455190 92.762 -63.563 54.640
455191 92.802 -63.571 54.638
455192 92.841 -63.581 54.636
455193 92.880 -63.590 54.633
455194 92.919 -63.600 54.632
455195 92.957 -63.609 54.630
455196 92.996 -63.614 54.628
455197 93.035 -63.614 54.625
455198 93.073 -63.615 54.621
455199 93.112 -63.615 54.615
455200 93.150 -63.615 54.610
455201 93.188 -63.615 54.603
455202 93.226 -63.615 54.595
455203 93.268 -63.614 54.587
455204 93.304 -63.611 54.576
455205 93.337 -63.608 54.563
455206 93.370 -63.606 54.549
455207 93.404 -63.603 54.532
455208 93.437 -63.600 54.514
455209 93.469 -63.597 54.495
455210 93.495 -63.604 54.476
455211 93.520 -63.611 54.456
455212 93.546 -63.618 54.435
455213 93.576 -63.630 54.410
455214 93.609 -63.647 54.385
455215 93.642 -63.665 54.359
455216 93.675 -63.681 54.331
455217 93.708 -63.699 54.302
455218 93.742 -63.716 54.272
455219 93.775 -63.733 54.241
455220 93.820 -63.768 54.210
455221 93.852 -63.788 54.179
455222 93.916 -63.797 54.147
455223 93.928 -63.807 54.116
455224 93.968 -63.815 54.086
455225 94.007 -63.822 54.057
455226 94.047 -63.831 54.030
455227 94.087 -63.838 54.008
455228 94.127 -63.846 53.986
455229 94.168 -63.855 53.965
455230 94.210 -63.865 53.945
455231 94.253 -63.875 53.927
455232 94.295 -63.885 53.911
455233 94.339 -63.899 53.899
455234 94.385 -63.916 53.890
455235 94.431 -63.934 53.881
455236 94.477 -63.951 53.872
455237 94.523 -63.969 53.862
455238 94.569 -63.988 53.849
455239 94.616 -64.002 53.834
455240 94.664 -64.001 53.816
455241 94.711 -64.001 53.797
455242 94.758 -64.001 53.775
455243 94.806 -64.001 53.754
455244 94.853 -64.000 53.733
455245 94.900 -64.000 53.712
455246 94.947 -63.999 53.692
455247 95.001 -64.005 53.671
455248 95.055 -64.013 53.653
455249 95.109 -64.020 53.634
455250 95.163 -64.027 53.615
455251 95.217 -64.034 53.595
455252 95.272 -64.041 53.574
455253 95.318 -64.048 53.554
455254 95.362 -64.054 53.535
455255 95.406 -64.061 53.516
455256 95.450 -64.068 53.496
455257 95.494 -64.074 53.475
455258 95.538 -64.081 53.451
455259 95.582 -64.088 53.426
455260 95.626 -64.095 53.401
455261 95.670 -64.101 53.376
455262 95.705 -64.099 53.351
455263 95.732 -64.088 53.322
455264 95.758 -64.078 53.295
455265 95.785 -64.068 53.269
455266 95.815 -64.063 53.243
455267 95.849 -64.067 53.222
455268 95.883 -64.070 53.201
455269 95.923 -64.056 53.178
455270 95.965 -64.038 53.151
455271 96.008 -64.020 53.121
455272 96.051 -64.007 53.089
455273 96.097 -64.001 53.057
455274 96.148 -64.010 53.027
455275 96.200 -64.020 53.003
455276 96.251 -64.029 52.983
455277 96.303 -64.038 52.966
455278 96.354 -64.048 52.944
455279 96.405 -64.057 52.921
455280 96.457 -64.066 52.895
455281 96.509 -64.076 52.865
455282 96.552 -64.082 52.832
455283 96.586 -64.084 52.794
455284 96.620 -64.085 52.755
455285 96.654 -64.088 52.715
455286 96.687 -64.090 52.675
455287 96.725 -64.095 52.635
455288 96.761 -64.099 52.590
455289 96.799 -64.104 52.543
455290 96.836 -64.109 52.495
455291 96.873 -64.114 52.446
455292 96.910 -64.118 52.398
455293 96.947 -64.123 52.352
455294 96.984 -64.128 52.307
455295 97.021 -64.132 52.264
455296 97.058 -64.137 52.223
455297 97.096 -64.142 52.182
455298 97.133 -64.147 52.141
455299 97.170 -64.151 52.098
455300 97.208 -64.156 52.053
455301 97.248 -64.160 52.005
455302 97.288 -64.164 51.958
455303 97.327 -64.167 51.914
455304 97.368 -64.171 51.872
455305 97.399 -64.164 51.832
455306 97.428 -64.167 51.794
455307 97.456 -64.174 51.753
455308 97.485 -64.182 51.710
455309 97.513 -64.189 51.666
455310 97.541 -64.195 51.621
455311 97.570 -64.203 51.576
455312 97.598 -64.210 51.530
455313 97.626 -64.217 51.483
455314 97.654 -64.224 51.435
455315 97.683 -64.231 51.387
455316 97.724 -64.227 51.340
455317 97.764 -64.223 51.295
455318 97.806 -64.217 51.248
455319 97.848 -64.209 51.201
455320 97.889 -64.201 51.152
455321 97.919 -64.204 51.104
455322 97.945 -64.208 51.053
455323 97.971 -64.213 51.001
455324 97.998 -64.217 50.946
455325 98.024 -64.222 50.891
455326 98.051 -64.227 50.837
455327 98.079 -64.231 50.783
455328 98.114 -64.232 50.731
455329 98.149 -64.233 50.680
455330 98.185 -64.233 50.628
455331 98.220 -64.234 50.578
455332 98.258 -64.229 50.528
455333 98.297 -64.220 50.479
455334 98.336 -64.211 50.430
455335 98.374 -64.205 50.383
455336 98.412 -64.208 50.335
455337 98.458 -64.211 50.287
455338 98.504 -64.214 50.239
455339 98.550 -64.217 50.191
455340 98.595 -64.220 50.144
455341 98.641 -64.222 50.098
455342 98.687 -64.225 50.053
455343 98.732 -64.228 50.008
455344 98.779 -64.230 49.963
455345 98.826 -64.233 49.918
455346 98.873 -64.235 49.874
455347 98.920 -64.238 49.829
455348 98.962 -64.241 49.784
455349 98.999 -64.244 49.738
455350 99.037 -64.248 49.694
455351 99.074 -64.251 49.650
455352 99.112 -64.255 49.604
455353 99.149 -64.258 49.559
455354 99.186 -64.262 49.515
455355 99.211 -64.267 49.473
455356 99.230 -64.273 49.432
455357 99.250 -64.280 49.391
455358 99.298 -64.286 49.352
455359 99.369 -64.291 49.313
455360 99.441 -64.296 49.274
455361 99.512 -64.302 49.235
455362 99.583 -64.306 49.196
455363 99.651 -64.309 49.158
455364 99.714 -64.308 49.121
455365 99.777 -64.305 49.083
455366 99.840 -64.304 49.046
455367 99.903 -64.302 49.009
455368 99.966 -64.300 48.972
455369 100.029 -64.298 48.935
455370 100.092 -64.296 48.900
455371 100.155 -64.295 48.865
455372 100.205 -64.295 48.831
455373 100.247 -64.298 48.796
455374 100.282 -64.305 48.762
455759 86.753 -62.618 56.135
455760 86.796 -62.616 56.122
455761 86.839 -62.614 56.111
455762 86.881 -62.612 56.099
455763 86.924 -62.610 56.087
455764 86.967 -62.609 56.075
455765 87.005 -62.609 56.063
455766 87.034 -62.616 56.050
455767 87.062 -62.622 56.035
455768 87.090 -62.628 56.020
455769 87.119 -62.634 56.004
455770 87.168 -62.630 55.988
455771 87.217 -62.625 55.972
455772 87.266 -62.620 55.954
455773 87.315 -62.616 55.935
455774 87.364 -62.611 55.917
455775 87.413 -62.606 55.898
455776 87.462 -62.601 55.880
455777 87.510 -62.597 55.862
455778 87.559 -62.592 55.844
455779 87.608 -62.587 55.826
455780 87.657 -62.583 55.811
455781 87.706 -62.578 55.797
455782 87.755 -62.573 55.784
455783 87.804 -62.575 55.772
455784 87.855 -62.584 55.762
455785 87.905 -62.593 55.754
455786 87.956 -62.601 55.747
455787 88.006 -62.610 55.742
455788 88.057 -62.619 55.739
455789 88.107 -62.628 55.735
455790 88.158 -62.637 55.727
455791 88.208 -62.646 55.714
455792 88.252 -62.661 55.695
455793 88.294 -62.676 55.670
455794 88.337 -62.691 55.647
455795 88.379 -62.706 55.623
455796 88.421 -62.722 55.600
455797 88.464 -62.738 55.577
455798 88.506 -62.753 55.554
455799 88.549 -62.768 55.531
455800 88.592 -62.783 55.509
455801 88.634 -62.798 55.489
455802 88.677 -62.814 55.470
455803 88.719 -62.829 55.450
455804 88.762 -62.844 55.433
455805 88.805 -62.860 55.418
455806 88.847 -62.875 55.404
455807 88.890 -62.890 55.394
455808 88.932 -62.905 55.386
455809 88.975 -62.921 55.380
455810 89.017 -62.936 55.375
455811 89.060 -62.951 55.371
455812 89.102 -62.967 55.368
455813 89.145 -62.982 55.367
455814 89.187 -62.997 55.366
455815 89.230 -63.012 55.364
455816 89.272 -63.028 55.361
455817 89.328 -63.016 55.353
455818 89.385 -63.004 55.348
455819 89.438 -63.001 55.344
455820 89.490 -63.001 55.341
455821 89.540 -63.006 55.341
455822 89.585 -63.026 55.345
455823 89.630 -63.046 55.353
455824 89.675 -63.067 55.360
455825 89.720 -63.087 55.364
455826 89.766 -63.098 55.364
455827 89.813 -63.107 55.359
455828 89.861 -63.116 55.351
455829 89.907 -63.125 55.343
455830 89.955 -63.135 55.334
455831 90.002 -63.144 55.323
455832 90.049 -63.153 55.306
455833 90.095 -63.156 55.285
455834 90.140 -63.158 55.264
455835 90.185 -63.159 55.247
455836 90.231 -63.161 55.234
455837 90.274 -63.163 55.226
455838 90.288 -63.180 55.220
455839 90.301 -63.198 55.216
455840 90.315 -63.217 55.215
455841 90.328 -63.234 55.215
455842 90.342 -63.252 55.217
455843 90.373 -63.267 55.219
455844 90.431 -63.276 55.219
455845 90.489 -63.285 55.214
455846 90.548 -63.294 55.205
455847 90.607 -63.303 55.193
455848 90.665 -63.312 55.173
455849 90.724 -63.321 55.144
455850 90.783 -63.330 55.109
455851 90.841 -63.339 55.069
455852 90.900 -63.348 55.031
455853 90.959 -63.357 54.994
455854 91.017 -63.366 54.960
455855 91.076 -63.375 54.929
455856 91.135 -63.384 54.899
455857 91.193 -63.393 54.872
455858 91.251 -63.402 54.848
455859 91.298 -63.411 54.828
455860 91.345 -63.419 54.810
455861 91.392 -63.428 54.797
455862 91.439 -63.437 54.785
455863 91.487 -63.446 54.773
455864 91.533 -63.454 54.763
455865 91.578 -63.457 54.756
455866 91.622 -63.457 54.750
455867 91.664 -63.457 54.744
455868 91.707 -63.457 54.741
455869 91.750 -63.456 54.739
455870 91.793 -63.456 54.736
455871 91.836 -63.456 54.737
455872 91.879 -63.455 54.739
455873 91.922 -63.455 54.742
455874 91.965 -63.455 54.744
455875 92.008 -63.455 54.743
455876 92.051 -63.455 54.737
455877 92.094 -63.455 54.728
455878 92.139 -63.457 54.718
455879 92.185 -63.462 54.708
455880 92.232 -63.468 54.698
455881 92.279 -63.473 54.691
455882 92.325 -63.479 54.685
455883 92.371 -63.485 54.680
455884 92.417 -63.490 54.675
455885 92.466 -63.496 54.669
455886 92.517 -63.499 54.664
455887 92.568 -63.503 54.660
455888 92.620 -63.507 54.655
455889 92.671 -63.511 54.651
455890 92.721 -63.515 54.647
455891 92.764 -63.521 54.643
455892 92.808 -63.525 54.641
455893 92.850 -63.532 54.639
455894 92.889 -63.541 54.636
455895 92.928 -63.550 54.634
455896 92.967 -63.559 54.630
455897 93.005 -63.561 54.627
455898 93.044 -63.561 54.623
455899 93.082 -63.562 54.617
455900 93.121 -63.562 54.611
455901 93.159 -63.562 54.604
455902 93.197 -63.562 54.596
455903 93.235 -63.562 54.586
455904 93.274 -63.562 54.576
455905 93.308 -63.559 54.563
455906 93.342 -63.556 54.549
455907 93.375 -63.553 54.532
455908 93.408 -63.550 54.513
455909 93.441 -63.547 54.493
455910 93.474 -63.545 54.474
455911 93.504 -63.546 54.454
455912 93.530 -63.553 54.434
455913 93.556 -63.559 54.413
455914 93.587 -63.574 54.390
455915 93.621 -63.591 54.364
455916 93.654 -63.609 54.339
455917 93.687 -63.625 54.312
455918 93.720 -63.643 54.284
455919 93.767 -63.681 54.256
455920 93.814 -63.718 54.224
455921 93.843 -63.734 54.192
455922 93.871 -63.750 54.160
455923 93.943 -63.757 54.127
455924 93.979 -63.765 54.097
455925 93.990 -63.775 54.066
455926 94.030 -63.782 54.037
455927 94.069 -63.791 54.010
455928 94.109 -63.798 53.987
455929 94.149 -63.806 53.965
455930 94.189 -63.814 53.945
455931 94.228 -63.822 53.925
455932 94.268 -63.830 53.909
455933 94.308 -63.838 53.895
455934 94.351 -63.854 53.886
455935 94.399 -63.874 53.878
455936 94.447 -63.894 53.870
455937 94.494 -63.913 53.862
455938 94.542 -63.933 53.853
455939 94.582 -63.949 53.840
455940 94.619 -63.958 53.826
455941 94.671 -63.958 53.809
455942 94.722 -63.959 53.789
455943 94.770 -63.959 53.769
455944 94.817 -63.958 53.747
455945 94.864 -63.957 53.726
455946 94.912 -63.957 53.705
455947 94.959 -63.957 53.685
455948 95.011 -63.961 53.665
455949 95.065 -63.968 53.646
455950 95.119 -63.976 53.627
455951 95.174 -63.982 53.608
455952 95.228 -63.990 53.588
455953 95.271 -63.996 53.568
455954 95.315 -64.003 53.550
455955 95.359 -64.010 53.532
455956 95.403 -64.016 53.513
455957 95.447 -64.023 53.494
455958 95.491 -64.029 53.471
455959 95.535 -64.036 53.447
455960 95.579 -64.043 53.422
455961 95.623 -64.050 53.396
455962 95.667 -64.057 53.371
455963 95.700 -64.053 53.344
455964 95.727 -64.042 53.315
455965 95.754 -64.032 53.286
455966 95.780 -64.021 53.257
455967 95.808 -64.012 53.230
455968 95.841 -64.015 53.207
455969 95.881 -64.003 53.186
455970 95.923 -63.985 53.163
455971 95.965 -63.966 53.137
455972 96.008 -63.949 53.108
455973 96.054 -63.945 53.077
455974 96.106 -63.954 53.045
455975 96.157 -63.963 53.016
455976 96.208 -63.973 52.991
455977 96.260 -63.982 52.971
455978 96.312 -63.991 52.952
455979 96.363 -64.001 52.929
455980 96.414 -64.010 52.904
455981 96.466 -64.020 52.878
455982 96.518 -64.029 52.848
455983 96.564 -64.035 52.814
455984 96.597 -64.038 52.776
455985 96.631 -64.040 52.736
455986 96.665 -64.041 52.695
455987 96.699 -64.045 52.655
455988 96.736 -64.049 52.615
455989 96.773 -64.054 52.572
455990 96.810 -64.058 52.526
455991 96.848 -64.063 52.478
455992 96.885 -64.068 52.429
455993 96.922 -64.073 52.382
455994 96.959 -64.077 52.337
455995 96.996 -64.082 52.293
455996 97.033 -64.087 52.252
455997 97.070 -64.092 52.212
455998 97.107 -64.096 52.171
455999 97.144 -64.101 52.129
456000 97.181 -64.106 52.086
456001 97.219 -64.110 52.042
456002 97.256 -64.115 51.994
456003 97.293 -64.120 51.947
456004 97.333 -64.124 51.905
456005 97.373 -64.128 51.866
456006 97.406 -64.123 51.829
456007 97.435 -64.131 51.791
456008 97.463 -64.138 51.751
456009 97.492 -64.145 51.708
456010 97.520 -64.152 51.663
456011 97.548 -64.159 51.618
456012 97.577 -64.166 51.572
456013 97.605 -64.173 51.526
456014 97.633 -64.180 51.479
456015 97.662 -64.187 51.431
456016 97.690 -64.195 51.383
456017 97.724 -64.197 51.337
456018 97.764 -64.192 51.290
456019 97.806 -64.185 51.244
456020 97.849 -64.177 51.196
456021 97.890 -64.169 51.147
456022 97.918 -64.172 51.098
456023 97.945 -64.177 51.047
456024 97.971 -64.182 50.994
456025 97.998 -64.186 50.939
456026 98.024 -64.191 50.884
456027 98.054 -64.194 50.829
456028 98.089 -64.195 50.774
456029 98.125 -64.195 50.722
456030 98.160 -64.197 50.670
456031 98.195 -64.198 50.618
456032 98.231 -64.198 50.568
456033 98.269 -64.193 50.518
456034 98.307 -64.184 50.468
456035 98.347 -64.176 50.419
456036 98.382 -64.173 50.371
456037 98.417 -64.176 50.324
456038 98.456 -64.178 50.276
456039 98.502 -64.181 50.228
456040 98.548 -64.184 50.180
456041 98.593 -64.187 50.133
456042 98.639 -64.189 50.086
456043 98.685 -64.192 50.040
456044 98.731 -64.195 49.995
456045 98.778 -64.198 49.950
456046 98.825 -64.200 49.905
456047 98.872 -64.202 49.861
456048 98.919 -64.204 49.817
456049 98.960 -64.208 49.772
456050 98.997 -64.211 49.726
456051 99.034 -64.215 49.682
456052 99.072 -64.218 49.637
456053 99.109 -64.222 49.590
456054 99.146 -64.226 49.543
456055 99.166 -64.232 49.500
456056 99.186 -64.238 49.458
456057 99.205 -64.244 49.417
456058 99.245 -64.250 49.376
456059 99.317 -64.255 49.337
456060 99.388 -64.261 49.298
456061 99.460 -64.265 49.259
456062 99.531 -64.270 49.220
456063 99.597 -64.272 49.181
456064 99.660 -64.270 49.144
456065 99.723 -64.268 49.107
456066 99.786 -64.267 49.070
456067 99.849 -64.265 49.032
456068 99.912 -64.263 48.996
456069 99.975 -64.261 48.959
456070 100.037 -64.259 48.922
456071 100.101 -64.258 48.888
456072 100.163 -64.256 48.854
456073 100.203 -64.258 48.820
456459 86.690 -62.590 56.158
456460 86.733 -62.589 56.144
456461 86.776 -62.587 56.131
456462 86.818 -62.585 56.120
456463 86.861 -62.584 56.109
456464 86.904 -62.581 56.097
456465 86.947 -62.580 56.086
456466 86.989 -62.578 56.074
456467 87.032 -62.576 56.061
456468 87.066 -62.580 56.047
456469 87.097 -62.584 56.033
456470 87.146 -62.580 56.018
456471 87.195 -62.575 56.002
456472 87.244 -62.570 55.984
456473 87.292 -62.565 55.966
456474 87.341 -62.561 55.947
456475 87.390 -62.556 55.928
456476 87.439 -62.552 55.909
456477 87.488 -62.547 55.889
456478 87.537 -62.542 55.870
456479 87.585 -62.537 55.852
456480 87.635 -62.533 55.834
456481 87.684 -62.528 55.817
456482 87.732 -62.523 55.802
456483 87.781 -62.518 55.789
456484 87.830 -62.514 55.778
456485 87.881 -62.522 55.770
456486 87.931 -62.531 55.763
456487 87.981 -62.540 55.758
456488 88.032 -62.548 55.754
456489 88.082 -62.557 55.752
456490 88.129 -62.570 55.748
456491 88.171 -62.585 55.741
456492 88.214 -62.600 55.728
456493 88.256 -62.616 55.708
456494 88.299 -62.631 55.683
456495 88.341 -62.646 55.660
456496 88.384 -62.661 55.636
456497 88.426 -62.677 55.609
456498 88.469 -62.692 55.585
456499 88.511 -62.707 55.563
456500 88.554 -62.722 55.539
456501 88.597 -62.738 55.516
456502 88.639 -62.753 55.496
456503 88.682 -62.768 55.476
456504 88.725 -62.784 55.457
456505 88.767 -62.799 55.439
456506 88.810 -62.814 55.423
456507 88.852 -62.829 55.408
456508 88.895 -62.844 55.397
456509 88.937 -62.860 55.388
456510 88.980 -62.876 55.381
456511 89.022 -62.891 55.376
456512 89.065 -62.906 55.371
456513 89.107 -62.921 55.368
456514 89.150 -62.936 55.366
456515 89.192 -62.951 55.363
456516 89.235 -62.967 55.361
456517 89.280 -62.979 55.358
456518 89.332 -62.979 55.350
456519 89.384 -62.979 55.344
456520 89.436 -62.980 55.338
456521 89.488 -62.981 55.333
456522 89.540 -62.982 55.331
456523 89.592 -62.982 55.332
456524 89.638 -62.998 55.338
456525 89.683 -63.019 55.344
456526 89.728 -63.039 55.349
456527 89.774 -63.050 55.350
456528 89.821 -63.059 55.347
456529 89.869 -63.068 55.341
456530 89.915 -63.077 55.334
456531 89.963 -63.086 55.326
456532 90.010 -63.096 55.315
456533 90.057 -63.105 55.299
456534 90.104 -63.114 55.279
456535 90.150 -63.117 55.258
456536 90.195 -63.119 55.239
456537 90.241 -63.120 55.225
456538 90.274 -63.128 55.215
456539 90.287 -63.146 55.207
456540 90.301 -63.164 55.200
456541 90.314 -63.182 55.196
456542 90.335 -63.197 55.192
456543 90.365 -63.208 55.191
456544 90.395 -63.220 55.190
456545 90.447 -63.230 55.188
456546 90.503 -63.241 55.185
456547 90.559 -63.252 55.177
456548 90.616 -63.263 55.162
456549 90.672 -63.274 55.141
456550 90.728 -63.284 55.114
456551 90.786 -63.294 55.081
456552 90.845 -63.303 55.046
456553 90.903 -63.312 55.013
456554 90.962 -63.321 54.981
456555 91.020 -63.330 54.951
456556 91.079 -63.339 54.921
456557 91.138 -63.348 54.894
456558 91.196 -63.357 54.870
456559 91.255 -63.366 54.849
456560 91.312 -63.375 54.832
456561 91.359 -63.383 54.815
456562 91.406 -63.393 54.802
456563 91.453 -63.401 54.791
456564 91.500 -63.409 54.781
456565 91.543 -63.409 54.771
456566 91.586 -63.409 54.764
456567 91.629 -63.408 54.757
456568 91.672 -63.408 54.752
456569 91.715 -63.408 54.749
456570 91.758 -63.408 54.747
456571 91.801 -63.408 54.745
456572 91.844 -63.408 54.745
456573 91.887 -63.407 54.749
456574 91.930 -63.407 54.753
456575 91.973 -63.407 54.756
456576 92.016 -63.407 54.755
456577 92.059 -63.406 54.750
456578 92.102 -63.406 54.741
456579 92.148 -63.411 54.729
456580 92.195 -63.416 54.717
456581 92.241 -63.422 54.706
456582 92.287 -63.427 54.697
456583 92.333 -63.433 54.690
456584 92.379 -63.440 54.685
456585 92.425 -63.446 54.679
456586 92.471 -63.452 54.673
456587 92.521 -63.455 54.668
456588 92.573 -63.460 54.664
456589 92.624 -63.464 54.659
456590 92.675 -63.468 54.653
456591 92.723 -63.472 54.649
456592 92.767 -63.477 54.646
456593 92.810 -63.483 54.644
456594 92.854 -63.487 54.641
456595 92.897 -63.493 54.638
456596 92.937 -63.501 54.635
456597 92.976 -63.508 54.630
456598 93.014 -63.508 54.625
456599 93.053 -63.508 54.620
456600 93.091 -63.508 54.612
456601 93.130 -63.509 54.605
456602 93.169 -63.515 54.596
456603 93.208 -63.519 54.586
456604 93.248 -63.524 54.575
456605 93.288 -63.526 54.563
456606 93.323 -63.516 54.549
456607 93.350 -63.509 54.533
456608 93.379 -63.501 54.514
456609 93.412 -63.498 54.493
456610 93.445 -63.496 54.472
456611 93.478 -63.493 54.452
456612 93.511 -63.490 54.432
456613 93.540 -63.493 54.413
456614 93.566 -63.501 54.392
456615 93.600 -63.518 54.368
456616 93.632 -63.535 54.344
456617 93.668 -63.556 54.318
456618 93.714 -63.593 54.292
456619 93.761 -63.631 54.265
456620 93.805 -63.665 54.236
456621 93.833 -63.680 54.204
456622 93.862 -63.696 54.171
456623 93.890 -63.712 54.139
456624 93.971 -63.717 54.105
456625 94.053 -63.722 54.074
456626 94.036 -63.733 54.044
456627 94.051 -63.742 54.016
456628 94.091 -63.750 53.989
456629 94.131 -63.758 53.966
456630 94.171 -63.766 53.945
456631 94.210 -63.774 53.925
456632 94.250 -63.781 53.907
456633 94.290 -63.790 53.893
456634 94.323 -63.800 53.882
456635 94.370 -63.819 53.874
456636 94.418 -63.839 53.865
456637 94.466 -63.859 53.858
456638 94.513 -63.878 53.850
456639 94.553 -63.894 53.841
456640 94.585 -63.906 53.831
456641 94.623 -63.914 53.817
456642 94.674 -63.915 53.800
456643 94.725 -63.915 53.782
456644 94.777 -63.916 53.762
456645 94.829 -63.916 53.741
456646 94.877 -63.916 53.720
456647 94.924 -63.915 53.699
456648 94.971 -63.914 53.679
456649 95.021 -63.917 53.659
456650 95.075 -63.924 53.639
456651 95.129 -63.931 53.620
456652 95.181 -63.938 53.601
456653 95.225 -63.944 53.582
456654 95.268 -63.951 53.563
456655 95.313 -63.958 53.547
456656 95.356 -63.965 53.529
456657 95.400 -63.972 53.511
456658 95.444 -63.978 53.492
456659 95.488 -63.985 53.469
456660 95.532 -63.991 53.444
456661 95.576 -63.998 53.419
456662 95.620 -64.005 53.393
456663 95.664 -64.011 53.367
456664 95.696 -64.007 53.339
456665 95.723 -63.996 53.309
456666 95.749 -63.985 53.278
456667 95.776 -63.975 53.247
456668 95.803 -63.965 53.219
456669 95.839 -63.949 53.194
456670 95.881 -63.931 53.170
456671 95.923 -63.913 53.146
456672 95.965 -63.894 53.120
456673 96.012 -63.888 53.092
456674 96.063 -63.898 53.062
456675 96.114 -63.907 53.031
456676 96.166 -63.916 53.002
456677 96.218 -63.926 52.976
456678 96.269 -63.935 52.953
456679 96.320 -63.944 52.931
456680 96.372 -63.954 52.907
456681 96.424 -63.963 52.880
456682 96.475 -63.973 52.852
456683 96.526 -63.982 52.823
456684 96.575 -63.990 52.790
456685 96.608 -63.991 52.753
456686 96.642 -63.994 52.713
456687 96.676 -63.996 52.672
456688 96.711 -63.999 52.632
456689 96.748 -64.004 52.592
456690 96.785 -64.008 52.550
456691 96.822 -64.013 52.506
456692 96.859 -64.018 52.459
456693 96.896 -64.023 52.412
456694 96.934 -64.027 52.366
456695 96.970 -64.032 52.323
456696 97.008 -64.037 52.281
456697 97.045 -64.041 52.241
456698 97.082 -64.046 52.201
456699 97.119 -64.051 52.158
456700 97.156 -64.056 52.116
456701 97.193 -64.060 52.073
456702 97.230 -64.065 52.029
456703 97.267 -64.070 51.981
456704 97.304 -64.075 51.938
456705 97.342 -64.079 51.898
456706 97.379 -64.084 51.861
456707 97.414 -64.087 51.825
456708 97.442 -64.095 51.789
456709 97.470 -64.101 51.749
456710 97.499 -64.108 51.707
456711 97.527 -64.116 51.663
456712 97.556 -64.123 51.617
456713 97.584 -64.129 51.571
456714 97.612 -64.137 51.524
456715 97.641 -64.144 51.476
456716 97.669 -64.151 51.428
456717 97.697 -64.158 51.380
456718 97.725 -64.165 51.332
456719 97.766 -64.161 51.284
456720 97.807 -64.153 51.237
456721 97.849 -64.145 51.189
456722 97.891 -64.136 51.140
456723 97.918 -64.141 51.091
456724 97.944 -64.145 51.040
456725 97.971 -64.150 50.987
456726 97.997 -64.154 50.931
456727 98.028 -64.157 50.875
456728 98.064 -64.158 50.820
456729 98.099 -64.159 50.765
456730 98.135 -64.160 50.712
456731 98.170 -64.161 50.660
456732 98.206 -64.161 50.608
456733 98.241 -64.162 50.558
456734 98.279 -64.157 50.507
456735 98.318 -64.148 50.457
456736 98.357 -64.139 50.408
456737 98.391 -64.142 50.360
456738 98.425 -64.144 50.312
456739 98.459 -64.146 50.265
456740 98.500 -64.148 50.217
456741 98.545 -64.151 50.170
456742 98.591 -64.154 50.122
456743 98.637 -64.157 50.076
456744 98.683 -64.160 50.029
456745 98.730 -64.162 49.983
456746 98.777 -64.164 49.938
456747 98.823 -64.167 49.893
456748 98.870 -64.169 49.848
456749 98.917 -64.171 49.804
456750 98.957 -64.175 49.760
456751 98.995 -64.179 49.714
456752 99.032 -64.182 49.668
456753 99.069 -64.186 49.622
456754 99.102 -64.190 49.575
456755 99.122 -64.196 49.529
456756 99.141 -64.202 49.486
456757 99.160 -64.208 49.443
456758 99.193 -64.214 49.402
456759 99.264 -64.220 49.362
456760 99.335 -64.225 49.322
456761 99.407 -64.229 49.282
456762 99.478 -64.235 49.244
456763 99.543 -64.235 49.206
456764 99.606 -64.233 49.168
456765 99.668 -64.232 49.130
456766 99.731 -64.229 49.093
456767 99.794 -64.228 49.057
456768 99.857 -64.226 49.020
456769 99.920 -64.224 48.983
456770 99.983 -64.223 48.947
456771 100.046 -64.220 48.910
456772 100.109 -64.219 48.876
456773 100.172 -64.217 48.843
457160 86.690 -62.551 56.167
457161 86.732 -62.550 56.153
457162 86.775 -62.548 56.140
457163 86.817 -62.546 56.129
457164 86.860 -62.544 56.118
457165 86.902 -62.543 56.107
457166 86.944 -62.541 56.096
457167 86.987 -62.540 56.084
457168 87.029 -62.538 56.072
457169 87.072 -62.537 56.059
457170 87.115 -62.535 56.047
457171 87.165 -62.530 56.033
457172 87.215 -62.525 56.016
457173 87.266 -62.518 55.998
457174 87.316 -62.513 55.980
457175 87.366 -62.507 55.961
457176 87.417 -62.502 55.940
457177 87.466 -62.497 55.920
457178 87.514 -62.492 55.901
457179 87.563 -62.487 55.881
457180 87.612 -62.483 55.861
457181 87.661 -62.478 55.841
457182 87.710 -62.474 55.824
457183 87.759 -62.468 55.809
457184 87.808 -62.464 55.795
457185 87.856 -62.459 55.785
457186 87.906 -62.459 55.778
457187 87.957 -62.468 55.773
457188 88.006 -62.478 55.769
457189 88.049 -62.493 55.766
457190 88.091 -62.509 55.764
457191 88.134 -62.524 55.760
457192 88.176 -62.540 55.753
457193 88.219 -62.555 55.741
457194 88.261 -62.570 55.722
457195 88.304 -62.585 55.695
457196 88.346 -62.600 55.671
457197 88.389 -62.616 55.646
457198 88.431 -62.631 55.619
457199 88.474 -62.647 55.592
457200 88.517 -62.662 55.569
457201 88.560 -62.676 55.545
457202 88.604 -62.691 55.523
457203 88.646 -62.706 55.503
457204 88.690 -62.722 55.483
457205 88.733 -62.737 55.464
457206 88.776 -62.752 55.445
457207 88.819 -62.767 55.427
457208 88.862 -62.782 55.414
457209 88.905 -62.797 55.400
457210 88.949 -62.813 55.389
457211 88.992 -62.828 55.382
457212 89.035 -62.843 55.376
457213 89.078 -62.857 55.371
457214 89.121 -62.873 55.367
457215 89.164 -62.888 55.363
457216 89.207 -62.903 55.360
457217 89.251 -62.915 55.357
457218 89.295 -62.924 55.354
457219 89.339 -62.932 55.347
457220 89.388 -62.941 55.339
457221 89.438 -62.949 55.332
457222 89.487 -62.958 55.326
457223 89.538 -62.962 55.322
457224 89.590 -62.963 55.321
457225 89.641 -62.963 55.324
457226 89.691 -62.971 55.328
457227 89.736 -62.991 55.332
457228 89.782 -63.001 55.334
457229 89.830 -63.010 55.332
457230 89.877 -63.020 55.328
457231 89.924 -63.029 55.323
457232 89.971 -63.039 55.317
457233 90.018 -63.048 55.306
457234 90.065 -63.057 55.291
457235 90.112 -63.067 55.273
457236 90.159 -63.076 55.253
457237 90.205 -63.079 55.233
457238 90.250 -63.080 55.217
457239 90.273 -63.093 55.204
457240 90.287 -63.111 55.194
457241 90.305 -63.127 55.187
457242 90.335 -63.139 55.178
457243 90.365 -63.150 55.173
457244 90.395 -63.161 55.169
457245 90.425 -63.173 55.164
457246 90.465 -63.184 55.160
457247 90.522 -63.195 55.156
457248 90.578 -63.205 55.147
457249 90.634 -63.216 55.131
457250 90.690 -63.227 55.110
457251 90.746 -63.237 55.086
457252 90.803 -63.248 55.057
457253 90.859 -63.258 55.027
457254 90.915 -63.270 54.998
457255 90.971 -63.280 54.970
457256 91.028 -63.291 54.943
457257 91.084 -63.302 54.916
457258 91.141 -63.312 54.892
457259 91.199 -63.321 54.870
457260 91.258 -63.330 54.851
457261 91.316 -63.339 54.835
457262 91.373 -63.348 54.822
457263 91.420 -63.357 54.810
457264 91.465 -63.361 54.799
457265 91.508 -63.361 54.789
457266 91.551 -63.361 54.780
457267 91.594 -63.361 54.773
457268 91.637 -63.361 54.767
457269 91.680 -63.360 54.760
457270 91.723 -63.360 54.756
457271 91.766 -63.360 54.753
457272 91.809 -63.359 54.752
457273 91.852 -63.359 54.754
457274 91.895 -63.359 54.759
457275 91.938 -63.358 54.765
457276 91.981 -63.358 54.769
457277 92.024 -63.358 54.769
457278 92.067 -63.358 54.763
457279 92.110 -63.359 54.754
457280 92.157 -63.365 54.743
457281 92.203 -63.371 54.728
457282 92.249 -63.377 54.715
457283 92.295 -63.383 54.704
457284 92.341 -63.389 54.695
457285 92.387 -63.395 54.688
457286 92.433 -63.401 54.682
457287 92.479 -63.406 54.676
457288 92.526 -63.412 54.671
457289 92.577 -63.416 54.667
457290 92.629 -63.420 54.661
457291 92.680 -63.424 54.655
457292 92.725 -63.429 54.649
457293 92.769 -63.434 54.646
457294 92.812 -63.440 54.642
457295 92.856 -63.445 54.639
457296 92.900 -63.449 54.637
457297 92.943 -63.455 54.632
457298 92.986 -63.463 54.627
457299 93.026 -63.469 54.621
457300 93.065 -63.474 54.614
457301 93.105 -63.478 54.606
457302 93.144 -63.483 54.597
457303 93.183 -63.488 54.585
457304 93.222 -63.493 54.573
457305 93.262 -63.497 54.560
457306 93.303 -63.494 54.548
457307 93.346 -63.484 54.532
457308 93.374 -63.475 54.515
457309 93.402 -63.468 54.494
457310 93.429 -63.459 54.471
457311 93.458 -63.451 54.450
457312 93.485 -63.443 54.429
457313 93.516 -63.437 54.410
457314 93.549 -63.435 54.391
457315 93.578 -63.445 54.371
457316 93.616 -63.468 54.347
457317 93.662 -63.506 54.323
457318 93.708 -63.543 54.297
457319 93.755 -63.580 54.271
457320 93.795 -63.610 54.244
457321 93.824 -63.626 54.214
457322 93.852 -63.641 54.183
457323 93.881 -63.657 54.150
457324 93.909 -63.673 54.117
457325 93.998 -63.678 54.084
457326 94.087 -63.681 54.052
457327 94.109 -63.690 54.023
457328 94.092 -63.701 53.995
457329 94.113 -63.710 53.969
457330 94.153 -63.718 53.946
457331 94.193 -63.726 53.925
457332 94.232 -63.734 53.907
457333 94.272 -63.741 53.892
457334 94.305 -63.751 53.879
457335 94.342 -63.765 53.870
457336 94.390 -63.784 53.861
457337 94.437 -63.804 53.853
457338 94.485 -63.824 53.846
457339 94.524 -63.839 53.840
457340 94.556 -63.850 53.832
457341 94.589 -63.862 53.822
457342 94.626 -63.870 53.809
457343 94.678 -63.871 53.792
457344 94.729 -63.871 53.774
457345 94.781 -63.872 53.755
457346 94.832 -63.872 53.735
457347 94.884 -63.872 53.714
457348 94.935 -63.872 53.694
457349 94.983 -63.872 53.674
457350 95.031 -63.872 53.653
457351 95.086 -63.880 53.633
457352 95.134 -63.887 53.614
457353 95.178 -63.894 53.595
457354 95.222 -63.900 53.577
457355 95.266 -63.907 53.558
457356 95.310 -63.913 53.542
457357 95.353 -63.920 53.525
457358 95.398 -63.927 53.508
457359 95.442 -63.933 53.488
457360 95.485 -63.940 53.466
457361 95.529 -63.947 53.441
457362 95.573 -63.954 53.417
457363 95.617 -63.960 53.392
457364 95.661 -63.966 53.364
457365 95.692 -63.960 53.336
457366 95.719 -63.950 53.304
457367 95.745 -63.939 53.271
457368 95.772 -63.929 53.238
457369 95.806 -63.922 53.209
457370 95.848 -63.913 53.181
457371 95.889 -63.889 53.154
457372 95.930 -63.866 53.129
457373 95.970 -63.841 53.103
457374 96.021 -63.842 53.074
457375 96.072 -63.851 53.045
457376 96.123 -63.860 53.015
457377 96.175 -63.869 52.986
457378 96.226 -63.879 52.958
457379 96.278 -63.888 52.933
457380 96.329 -63.897 52.908
457381 96.381 -63.907 52.881
457382 96.432 -63.916 52.853
457383 96.484 -63.925 52.824
457384 96.535 -63.935 52.795
457385 96.586 -63.944 52.763
457386 96.619 -63.946 52.728
457387 96.653 -63.948 52.689
457388 96.687 -63.950 52.648
457389 96.723 -63.954 52.609
457390 96.759 -63.958 52.569
457391 96.797 -63.963 52.527
457392 96.834 -63.968 52.484
457393 96.871 -63.973 52.440
457394 96.908 -63.977 52.395
457395 96.945 -63.982 52.351
457396 96.982 -63.987 52.309
457397 97.019 -63.991 52.269
457398 97.056 -63.996 52.229
457399 97.094 -64.001 52.189
457400 97.131 -64.006 52.146
457401 97.168 -64.010 52.103
457402 97.205 -64.015 52.060
457403 97.242 -64.020 52.016
457404 97.279 -64.025 51.970
457405 97.316 -64.029 51.929
457406 97.353 -64.034 51.892
457407 97.387 -64.043 51.857
457408 97.414 -64.056 51.822
457409 97.442 -64.063 51.786
457410 97.471 -64.070 51.747
457411 97.500 -64.076 51.705
457412 97.528 -64.083 51.662
457413 97.557 -64.090 51.617
457414 97.585 -64.098 51.571
457415 97.614 -64.104 51.523
457416 97.643 -64.111 51.475
457417 97.671 -64.118 51.426
457418 97.700 -64.125 51.377
457419 97.728 -64.132 51.328
457420 97.766 -64.129 51.279
457421 97.808 -64.120 51.231
457422 97.850 -64.112 51.183
457423 97.891 -64.104 51.133
457424 97.918 -64.109 51.083
457425 97.944 -64.114 51.032
457426 97.970 -64.119 50.979
457427 98.004 -64.120 50.924
457428 98.039 -64.121 50.867
457429 98.074 -64.122 50.811
457430 98.109 -64.123 50.756
457431 98.145 -64.123 50.702
457432 98.180 -64.125 50.650
457433 98.216 -64.126 50.598
457434 98.251 -64.126 50.547
457435 98.289 -64.120 50.496
457436 98.329 -64.112 50.446
457437 98.365 -64.107 50.397
457438 98.400 -64.110 50.349
457439 98.434 -64.112 50.301
457440 98.467 -64.114 50.254
457441 98.502 -64.116 50.207
457442 98.543 -64.119 50.159
457443 98.589 -64.121 50.112
457444 98.635 -64.124 50.065
457445 98.682 -64.126 50.019
457446 98.729 -64.129 49.972
457447 98.776 -64.131 49.926
457448 98.822 -64.134 49.881
457449 98.869 -64.136 49.835
457450 98.916 -64.139 49.790
457451 98.955 -64.142 49.745
457452 98.992 -64.145 49.699
457453 99.029 -64.149 49.652
457454 99.057 -64.154 49.606
457455 99.077 -64.161 49.559
457456 99.096 -64.167 49.514
457457 99.116 -64.173 49.471
457458 99.140 -64.179 49.428
457459 99.211 -64.184 49.387
457460 99.283 -64.189 49.347
457461 99.354 -64.194 49.307
457462 99.425 -64.199 49.268
457463 99.489 -64.198 49.229
457464 99.552 -64.196 49.192
457465 99.615 -64.195 49.154
457466 99.678 -64.192 49.116
457467 99.740 -64.191 49.080
457468 99.803 -64.189 49.043
457469 99.866 -64.187 49.007
457470 99.929 -64.186 48.971
457471 99.992 -64.183 48.934
457472 100.055 -64.182 48.898
457473 100.118 -64.180 48.865
457861 86.707 -62.503 56.174
457862 86.749 -62.500 56.161
457863 86.791 -62.499 56.149
457864 86.834 -62.497 56.138
457865 86.876 -62.496 56.127
457866 86.919 -62.494 56.116
457867 86.961 -62.493 56.106
457868 87.003 -62.491 56.095
457869 87.046 -62.490 56.083
457870 87.088 -62.488 56.072
457871 87.125 -62.490 56.060
457872 87.173 -62.487 56.047
457873 87.221 -62.483 56.032
457874 87.271 -62.478 56.014
457875 87.321 -62.472 55.995
457876 87.371 -62.467 55.975
457877 87.422 -62.461 55.954
457878 87.472 -62.456 55.932
457879 87.522 -62.449 55.912
457880 87.573 -62.444 55.891
457881 87.623 -62.438 55.871
457882 87.674 -62.433 55.851
457883 87.724 -62.427 55.833
457884 87.774 -62.421 55.816
457885 87.825 -62.415 55.802
457886 87.875 -62.410 55.793
457887 87.925 -62.404 55.786
457888 87.972 -62.416 55.781
457889 88.015 -62.431 55.777
457890 88.058 -62.446 55.775
457891 88.102 -62.462 55.773
457892 88.144 -62.477 55.770
457893 88.188 -62.492 55.763
457894 88.231 -62.507 55.752
457895 88.274 -62.522 55.734
457896 88.317 -62.537 55.707
457897 88.360 -62.553 55.681
457898 88.403 -62.567 55.655
457899 88.447 -62.582 55.628
457900 88.490 -62.597 55.600
457901 88.533 -62.612 55.574
457902 88.576 -62.628 55.550
457903 88.619 -62.643 55.528
457904 88.663 -62.658 55.509
457905 88.705 -62.673 55.489
457906 88.749 -62.688 55.469
457907 88.792 -62.703 55.450
457908 88.835 -62.719 55.432
457909 88.878 -62.733 55.417
457910 88.921 -62.748 55.403
457911 88.964 -62.763 55.392
457912 89.008 -62.778 55.383
457913 89.051 -62.794 55.376
457914 89.094 -62.809 55.370
457915 89.137 -62.824 55.365
457916 89.180 -62.839 55.361
457917 89.224 -62.851 55.357
457918 89.267 -62.860 55.353
457919 89.311 -62.868 55.349
457920 89.355 -62.876 55.342
457921 89.399 -62.885 55.334
457922 89.448 -62.894 55.327
457923 89.497 -62.902 55.320
457924 89.546 -62.910 55.315
457925 89.596 -62.919 55.311
457926 89.645 -62.928 55.311
457927 89.694 -62.936 55.313
457928 89.743 -62.945 55.317
457929 89.791 -62.954 55.319
457930 89.838 -62.963 55.319
457931 89.885 -62.972 55.316
457932 89.932 -62.982 55.312
457933 89.979 -62.991 55.307
457934 90.026 -63.000 55.298
457935 90.073 -63.009 55.286
457936 90.120 -63.018 55.270
457937 90.167 -63.028 55.251
457938 90.215 -63.037 55.230
457939 90.259 -63.041 55.211
457940 90.276 -63.057 55.197
457941 90.306 -63.069 55.185
457942 90.336 -63.080 55.174
457943 90.366 -63.092 55.165
457944 90.396 -63.103 55.156
457945 90.426 -63.114 55.149
457946 90.456 -63.126 55.142
457947 90.485 -63.137 55.136
457948 90.540 -63.148 55.130
457949 90.596 -63.158 55.118
457950 90.652 -63.169 55.101
457951 90.709 -63.180 55.080
457952 90.764 -63.191 55.059
457953 90.821 -63.202 55.036
457954 90.877 -63.212 55.011
457955 90.933 -63.223 54.985
457956 90.990 -63.233 54.960
457957 91.046 -63.244 54.936
457958 91.102 -63.255 54.912
457959 91.158 -63.265 54.890
457960 91.215 -63.277 54.871
457961 91.271 -63.287 54.855
457962 91.327 -63.298 54.840
457963 91.383 -63.308 54.829
457964 91.439 -63.315 54.819
457965 91.483 -63.315 54.807
457966 91.527 -63.315 54.797
457967 91.572 -63.314 54.787
457968 91.616 -63.314 54.780
457969 91.660 -63.314 54.773
457970 91.704 -63.314 54.766
457971 91.748 -63.314 54.761
457972 91.792 -63.314 54.758
457973 91.837 -63.314 54.757
457974 91.881 -63.314 54.759
457975 91.925 -63.314 54.766
457976 91.969 -63.314 54.775
457977 92.013 -63.314 54.781
457978 92.058 -63.314 54.781
457979 92.093 -63.317 54.777
457980 92.127 -63.320 54.769
457981 92.164 -63.323 54.757
457982 92.211 -63.327 54.743
457983 92.257 -63.332 54.727
457984 92.303 -63.338 54.713
457985 92.349 -63.343 54.701
457986 92.395 -63.349 54.692
457987 92.441 -63.355 54.684
457988 92.487 -63.361 54.678
457989 92.533 -63.368 54.673
457990 92.581 -63.373 54.669
457991 92.633 -63.377 54.663
457992 92.684 -63.380 54.656
457993 92.727 -63.386 54.650
457994 92.771 -63.391 54.645
457995 92.814 -63.396 54.640
457996 92.858 -63.402 54.636
457997 92.902 -63.409 54.633
457998 92.946 -63.417 54.628
457999 92.990 -63.425 54.621
458000 93.031 -63.432 54.614
458001 93.073 -63.440 54.606
458002 93.115 -63.447 54.597
458003 93.156 -63.455 54.585
458004 93.198 -63.461 54.571
458005 93.237 -63.466 54.557
458006 93.276 -63.471 54.542
458007 93.319 -63.462 54.527
458008 93.362 -63.452 54.511
458009 93.397 -63.443 54.493
458010 93.425 -63.434 54.472
458011 93.453 -63.426 54.449
458012 93.481 -63.418 54.428
458013 93.509 -63.410 54.408
458014 93.537 -63.402 54.388
458015 93.561 -63.397 54.370
458016 93.602 -63.418 54.350
458017 93.656 -63.455 54.326
458018 93.702 -63.493 54.302
458019 93.749 -63.530 54.275
458020 93.786 -63.556 54.248
458021 93.815 -63.571 54.222
458022 93.843 -63.587 54.192
458023 93.871 -63.603 54.162
458024 93.900 -63.618 54.129
458025 93.936 -63.633 54.095
458026 94.025 -63.637 54.061
458027 94.114 -63.641 54.031
458028 94.183 -63.647 54.002
458029 94.166 -63.658 53.975
458030 94.148 -63.669 53.949
458031 94.175 -63.678 53.926
458032 94.214 -63.686 53.906
458033 94.254 -63.694 53.891
458034 94.288 -63.703 53.878
458035 94.318 -63.712 53.867
458036 94.362 -63.730 53.857
458037 94.410 -63.750 53.848
458038 94.457 -63.770 53.841
458039 94.495 -63.784 53.834
458040 94.527 -63.796 53.828
458041 94.559 -63.807 53.822
458042 94.591 -63.819 53.812
458043 94.629 -63.826 53.799
458044 94.681 -63.827 53.784
458045 94.733 -63.827 53.767
458046 94.784 -63.828 53.748
458047 94.836 -63.828 53.729
458048 94.887 -63.828 53.709
458049 94.939 -63.828 53.688
458050 94.990 -63.829 53.667
458051 95.042 -63.829 53.647
458052 95.087 -63.835 53.627
458053 95.131 -63.842 53.607
458054 95.175 -63.848 53.589
458055 95.219 -63.855 53.570
458056 95.263 -63.862 53.552
458057 95.307 -63.869 53.535
458058 95.351 -63.875 53.519
458059 95.395 -63.882 53.502
458060 95.439 -63.888 53.484
458061 95.483 -63.895 53.462
458062 95.527 -63.902 53.440
458063 95.570 -63.909 53.416
458064 95.614 -63.915 53.390
458065 95.659 -63.922 53.363
458066 95.687 -63.913 53.334
458067 95.714 -63.904 53.301
458068 95.741 -63.893 53.266
458069 95.770 -63.884 53.232
458070 95.815 -63.883 53.199
458071 95.859 -63.880 53.169
458072 95.899 -63.856 53.140
458073 95.940 -63.832 53.112
458074 95.981 -63.810 53.084
458075 96.031 -63.805 53.056
458076 96.081 -63.804 53.026
458077 96.132 -63.813 52.996
458078 96.184 -63.823 52.967
458079 96.235 -63.832 52.938
458080 96.287 -63.841 52.910
458081 96.338 -63.851 52.882
458082 96.390 -63.860 52.852
458083 96.441 -63.869 52.822
458084 96.493 -63.878 52.793
458085 96.544 -63.888 52.763
458086 96.596 -63.897 52.733
458087 96.631 -63.900 52.700
458088 96.664 -63.902 52.664
458089 96.698 -63.904 52.626
458090 96.734 -63.908 52.586
458091 96.771 -63.913 52.547
458092 96.808 -63.917 52.505
458093 96.846 -63.922 52.462
458094 96.883 -63.927 52.419
458095 96.920 -63.932 52.378
458096 96.957 -63.936 52.336
458097 96.994 -63.941 52.296
458098 97.031 -63.946 52.257
458099 97.068 -63.951 52.217
458100 97.105 -63.955 52.176
458101 97.143 -63.960 52.134
458102 97.179 -63.965 52.091
458103 97.217 -63.969 52.047
458104 97.254 -63.974 52.003
458105 97.291 -63.979 51.960
458106 97.326 -63.986 51.922
458107 97.357 -63.999 51.887
458108 97.388 -64.011 51.854
458109 97.412 -64.026 51.819
458110 97.439 -64.034 51.783
458111 97.468 -64.041 51.745
458112 97.496 -64.048 51.703
458113 97.525 -64.054 51.660
458114 97.554 -64.061 51.617
458115 97.582 -64.068 51.571
458116 97.611 -64.075 51.524
458117 97.639 -64.082 51.475
458118 97.668 -64.089 51.425
458119 97.697 -64.096 51.375
458120 97.732 -64.095 51.324
458121 97.770 -64.092 51.275
458122 97.809 -64.088 51.226
458123 97.851 -64.079 51.177
458124 97.891 -64.073 51.125
458125 97.917 -64.078 51.074
458126 97.944 -64.082 51.023
458127 97.979 -64.083 50.971
458128 98.014 -64.085 50.916
458129 98.049 -64.085 50.860
458130 98.085 -64.086 50.802
458131 98.120 -64.087 50.747
458132 98.155 -64.088 50.693
458133 98.191 -64.088 50.640
458134 98.226 -64.089 50.588
458135 98.262 -64.090 50.537
458136 98.299 -64.085 50.485
458137 98.339 -64.076 50.435
458138 98.374 -64.076 50.386
458139 98.408 -64.078 50.338
458140 98.442 -64.079 50.289
458141 98.476 -64.082 50.243
458142 98.510 -64.084 50.196
458143 98.545 -64.086 50.148
458144 98.587 -64.088 50.101
458145 98.634 -64.091 50.055
458146 98.680 -64.094 50.009
458147 98.727 -64.096 49.963
458148 98.774 -64.098 49.916
458149 98.821 -64.101 49.870
458150 98.868 -64.103 49.825
458151 98.915 -64.105 49.779
458152 98.952 -64.109 49.732
458153 98.990 -64.113 49.686
458154 99.012 -64.118 49.638
458155 99.032 -64.124 49.591
458156 99.051 -64.130 49.545
458157 99.071 -64.137 49.500
458158 99.090 -64.143 49.457
458159 99.159 -64.148 49.414
458160 99.230 -64.154 49.373
458161 99.301 -64.158 49.333
458162 99.371 -64.163 49.293
458163 99.434 -64.161 49.253
458164 99.497 -64.159 49.214
458165 99.560 -64.157 49.177
458166 99.623 -64.156 49.140
458167 99.686 -64.154 49.103
458168 99.749 -64.152 49.066
458169 99.812 -64.150 49.030
458170 99.875 -64.148 48.993
458171 99.938 -64.147 48.958
458562 86.723 -62.453 56.180
458563 86.765 -62.452 56.168
458564 86.808 -62.450 56.156
458565 86.850 -62.449 56.145
458566 86.893 -62.447 56.135
458567 86.935 -62.446 56.124
458568 86.978 -62.444 56.114
458569 87.020 -62.443 56.104
458570 87.062 -62.440 56.093
458571 87.103 -62.440 56.083
458572 87.140 -62.442 56.073
458573 87.183 -62.441 56.061
458574 87.230 -62.437 56.047
458575 87.279 -62.434 56.031
458576 87.327 -62.431 56.011
458577 87.376 -62.427 55.989
458578 87.427 -62.421 55.967
458579 87.477 -62.415 55.945
458580 87.528 -62.409 55.923
458581 87.578 -62.404 55.901
458582 87.628 -62.398 55.881
458583 87.679 -62.393 55.861
458584 87.729 -62.387 55.842
458585 87.779 -62.381 55.825
458586 87.830 -62.375 55.811
458587 87.880 -62.370 55.801
458588 87.930 -62.364 55.793
458589 87.980 -62.366 55.787
458590 88.029 -62.382 55.784
458591 88.074 -62.398 55.781
458592 88.117 -62.413 55.780
458593 88.160 -62.428 55.777
458594 88.203 -62.443 55.772
458595 88.247 -62.458 55.762
458596 88.290 -62.473 55.744
458597 88.333 -62.488 55.719
458598 88.376 -62.503 55.691
458599 88.419 -62.518 55.663
458600 88.462 -62.534 55.636
458601 88.505 -62.549 55.607
458602 88.549 -62.564 55.579
458603 88.592 -62.579 55.554
458604 88.635 -62.594 55.531
458605 88.678 -62.609 55.513
458606 88.721 -62.624 55.493
458607 88.764 -62.639 55.473
458608 88.808 -62.654 55.454
458609 88.851 -62.669 55.436
458610 88.894 -62.684 55.420
458611 88.937 -62.700 55.406
458612 88.980 -62.715 55.394
458613 89.023 -62.730 55.384
458614 89.066 -62.745 55.376
458615 89.110 -62.760 55.369
458616 89.153 -62.775 55.363
458617 89.196 -62.787 55.358
458618 89.240 -62.795 55.353
458619 89.284 -62.804 55.349
458620 89.328 -62.812 55.344
458621 89.372 -62.820 55.337
458622 89.415 -62.829 55.329
458623 89.459 -62.838 55.321
458624 89.507 -62.846 55.314
458625 89.556 -62.854 55.307
458626 89.605 -62.863 55.302
458627 89.655 -62.872 55.299
458628 89.704 -62.880 55.300
458629 89.753 -62.888 55.304
458630 89.801 -62.898 55.306
458631 89.847 -62.908 55.306
458632 89.894 -62.917 55.304
458633 89.942 -62.927 55.300
458634 89.988 -62.937 55.297
458635 90.035 -62.946 55.289
458636 90.083 -62.956 55.280
458637 90.129 -62.966 55.266
458638 90.176 -62.976 55.249
458639 90.224 -62.985 55.228
458640 90.272 -63.000 55.208
458641 90.306 -63.010 55.191
458642 90.336 -63.022 55.176
458643 90.366 -63.033 55.164
458644 90.396 -63.045 55.153
458645 90.426 -63.056 55.143
458646 90.456 -63.067 55.134
458647 90.486 -63.079 55.126
458648 90.516 -63.090 55.117
458649 90.558 -63.101 55.108
458650 90.614 -63.111 55.095
458651 90.670 -63.123 55.079
458652 90.727 -63.133 55.061
458653 90.783 -63.144 55.042
458654 90.839 -63.155 55.020
458655 90.895 -63.165 54.999
458656 90.952 -63.176 54.976
458657 91.008 -63.186 54.953
458658 91.064 -63.197 54.931
458659 91.120 -63.208 54.910
458660 91.176 -63.219 54.891
458661 91.233 -63.230 54.873
458662 91.289 -63.240 54.859
458663 91.345 -63.251 54.846
458664 91.399 -63.262 54.836
458665 91.455 -63.269 54.826
458666 91.515 -63.271 54.815
458667 91.568 -63.273 54.803
458668 91.612 -63.273 54.793
458669 91.656 -63.273 54.785
458670 91.700 -63.273 54.778
458671 91.744 -63.273 54.772
458672 91.789 -63.273 54.767
458673 91.832 -63.273 54.764
458674 91.877 -63.273 54.762
458675 91.921 -63.273 54.765
458676 91.965 -63.273 54.773
458677 92.009 -63.273 54.784
458678 92.047 -63.274 54.792
458679 92.080 -63.277 54.794
458680 92.114 -63.280 54.791
458681 92.148 -63.282 54.783
458682 92.181 -63.285 54.770
458683 92.218 -63.288 54.756
458684 92.264 -63.292 54.740
458685 92.310 -63.296 54.724
458686 92.357 -63.301 54.710
458687 92.403 -63.305 54.697
458688 92.449 -63.311 54.688
458689 92.496 -63.317 54.681
458690 92.542 -63.322 54.676
458691 92.588 -63.328 54.671
458692 92.637 -63.333 54.665
458693 92.686 -63.337 54.658
458694 92.730 -63.343 54.651
458695 92.773 -63.348 54.644
458696 92.817 -63.355 54.638
458697 92.862 -63.364 54.631
458698 92.906 -63.371 54.626
458699 92.950 -63.379 54.620
458700 92.994 -63.387 54.614
458701 93.036 -63.394 54.606
458702 93.077 -63.402 54.596
458703 93.119 -63.409 54.584
458704 93.160 -63.416 54.570
458705 93.202 -63.424 54.555
458706 93.244 -63.431 54.539
458707 93.285 -63.436 54.523
458708 93.335 -63.431 54.506
458709 93.378 -63.420 54.489
458710 93.421 -63.409 54.469
458711 93.448 -63.402 54.449
458712 93.477 -63.393 54.427
458713 93.504 -63.385 54.406
458714 93.532 -63.377 54.386
458715 93.549 -63.382 54.367
458716 93.565 -63.386 54.348
458717 93.592 -63.406 54.328
458718 93.666 -63.443 54.305
458719 93.742 -63.479 54.280
458720 93.777 -63.502 54.253
458721 93.805 -63.518 54.226
458722 93.834 -63.533 54.199
458723 93.862 -63.549 54.170
458724 93.891 -63.565 54.138
458725 93.919 -63.581 54.105
458726 93.963 -63.593 54.071
458727 94.052 -63.597 54.039
458728 94.141 -63.601 54.009
458729 94.231 -63.606 53.981
458730 94.239 -63.615 53.955
458731 94.222 -63.626 53.931
458732 94.205 -63.637 53.910
458733 94.236 -63.646 53.894
458734 94.271 -63.655 53.879
458735 94.301 -63.665 53.867
458736 94.334 -63.676 53.857
458737 94.382 -63.696 53.846
458738 94.429 -63.715 53.836
458739 94.466 -63.729 53.828
458740 94.498 -63.741 53.822
458741 94.530 -63.752 53.816
458742 94.562 -63.763 53.811
458743 94.594 -63.775 53.800
458744 94.633 -63.782 53.788
458745 94.685 -63.783 53.775
458746 94.736 -63.783 53.759
458747 94.787 -63.784 53.741
458748 94.839 -63.784 53.723
458749 94.891 -63.784 53.703
458750 94.943 -63.784 53.682
458751 94.994 -63.785 53.661
458752 95.052 -63.786 53.640
458753 95.091 -63.791 53.620
458754 95.131 -63.797 53.600
458755 95.172 -63.803 53.581
458756 95.216 -63.810 53.562
458757 95.260 -63.817 53.544
458758 95.304 -63.824 53.527
458759 95.348 -63.831 53.511
458760 95.392 -63.837 53.495
458761 95.436 -63.844 53.478
458762 95.480 -63.850 53.457
458763 95.524 -63.857 53.436
458764 95.568 -63.864 53.413
458765 95.612 -63.870 53.388
458766 95.656 -63.877 53.361
458767 95.683 -63.867 53.332
458768 95.709 -63.857 53.297
458769 95.736 -63.847 53.263
458770 95.779 -63.845 53.228
458771 95.824 -63.844 53.192
458772 95.868 -63.843 53.159
458773 95.909 -63.823 53.128
458774 95.950 -63.800 53.097
458775 95.992 -63.779 53.067
458776 96.042 -63.774 53.036
458777 96.091 -63.769 53.006
458778 96.141 -63.767 52.976
458779 96.193 -63.776 52.947
458780 96.244 -63.785 52.917
458781 96.296 -63.794 52.887
458782 96.347 -63.804 52.856
458783 96.398 -63.813 52.825
458784 96.450 -63.822 52.793
458785 96.502 -63.832 52.763
458786 96.553 -63.841 52.732
458787 96.604 -63.850 52.702
458788 96.642 -63.854 52.670
458789 96.676 -63.856 52.636
458790 96.709 -63.858 52.600
458791 96.746 -63.863 52.563
458792 96.783 -63.867 52.524
458793 96.820 -63.872 52.484
458794 96.857 -63.877 52.442
458795 96.894 -63.882 52.400
458796 96.932 -63.886 52.360
458797 96.968 -63.891 52.321
458798 97.006 -63.896 52.283
458799 97.043 -63.900 52.245
458800 97.080 -63.905 52.206
458801 97.117 -63.910 52.165
458802 97.154 -63.915 52.122
458803 97.191 -63.919 52.079
458804 97.228 -63.924 52.037
458805 97.265 -63.929 51.994
458806 97.296 -63.941 51.954
458807 97.328 -63.954 51.917
458808 97.359 -63.966 51.883
458809 97.387 -63.980 51.850
458810 97.410 -63.994 51.816
458811 97.436 -64.005 51.780
458812 97.464 -64.012 51.741
458813 97.493 -64.019 51.701
458814 97.522 -64.026 51.659
458815 97.550 -64.032 51.616
458816 97.579 -64.039 51.571
458817 97.607 -64.046 51.524
458818 97.636 -64.053 51.475
458819 97.665 -64.060 51.424
458820 97.699 -64.061 51.374
458821 97.737 -64.058 51.322
458822 97.775 -64.055 51.272
458823 97.813 -64.051 51.222
458824 97.852 -64.047 51.171
458825 97.890 -64.041 51.119
458826 97.918 -64.046 51.068
458827 97.953 -64.047 51.017
458828 97.989 -64.048 50.965
458829 98.024 -64.048 50.911
458830 98.060 -64.049 50.855
458831 98.095 -64.050 50.797
458832 98.131 -64.051 50.740
458833 98.166 -64.051 50.686
458834 98.201 -64.053 50.633
458835 98.236 -64.054 50.579
458836 98.272 -64.054 50.527
458837 98.310 -64.048 50.475
458838 98.348 -64.041 50.424
458839 98.382 -64.044 50.374
458840 98.417 -64.045 50.325
458841 98.450 -64.048 50.277
458842 98.485 -64.050 50.231
458843 98.519 -64.052 50.184
458844 98.551 -64.054 50.138
458845 98.577 -64.057 50.091
458846 98.628 -64.059 50.046
458847 98.678 -64.060 50.001
458848 98.726 -64.063 49.955
458849 98.773 -64.065 49.909
458850 98.820 -64.067 49.863
458851 98.867 -64.070 49.818
458852 98.913 -64.073 49.771
458853 98.948 -64.076 49.723
458854 98.968 -64.082 49.674
458855 98.987 -64.088 49.626
458856 99.006 -64.095 49.578
458857 99.026 -64.101 49.532
458858 99.045 -64.107 49.488
458859 99.106 -64.113 49.444
458860 99.177 -64.117 49.401
458861 99.249 -64.123 49.359
458862 99.317 -64.126 49.319
458863 99.380 -64.124 49.279
458864 99.443 -64.123 49.239
458865 99.506 -64.120 49.200
458866 99.569 -64.119 49.163
458867 99.632 -64.117 49.126
458868 99.695 -64.115 49.090
458869 99.758 -64.113 49.053
459263 86.739 -62.405 56.186
459264 86.782 -62.403 56.174
459265 86.824 -62.402 56.162
459266 86.867 -62.400 56.151
459267 86.909 -62.399 56.141
459268 86.951 -62.397 56.131
459269 86.994 -62.396 56.121
459270 87.036 -62.393 56.112
459271 87.079 -62.392 56.102
459272 87.118 -62.392 56.093
459273 87.155 -62.394 56.084
459274 87.192 -62.396 56.074
459275 87.240 -62.393 56.062
459276 87.288 -62.389 56.046
459277 87.337 -62.386 56.028
459278 87.385 -62.382 56.006
459279 87.433 -62.379 55.983
459280 87.482 -62.375 55.959
459281 87.533 -62.369 55.936
459282 87.583 -62.364 55.913
459283 87.633 -62.358 55.892
459284 87.684 -62.352 55.871
459285 87.734 -62.346 55.852
459286 87.784 -62.341 55.834
459287 87.835 -62.335 55.820
459288 87.885 -62.329 55.809
459289 87.935 -62.324 55.799
459290 87.986 -62.318 55.792
459291 88.035 -62.331 55.788
459292 88.083 -62.347 55.785
459293 88.132 -62.364 55.785
459294 88.176 -62.379 55.784
459295 88.219 -62.394 55.780
459296 88.262 -62.409 55.771
459297 88.306 -62.424 55.753
459298 88.349 -62.440 55.728
459299 88.392 -62.455 55.700
459300 88.435 -62.470 55.672
459301 88.478 -62.485 55.644
459302 88.521 -62.500 55.612
459303 88.564 -62.515 55.583
459304 88.608 -62.530 55.557
459305 88.650 -62.545 55.533
459306 88.694 -62.560 55.515
459307 88.737 -62.575 55.495
459308 88.780 -62.590 55.475
459309 88.823 -62.606 55.456
459310 88.867 -62.621 55.438
459311 88.909 -62.636 55.422
459312 88.953 -62.651 55.409
459313 88.996 -62.666 55.397
459314 89.039 -62.681 55.385
459315 89.082 -62.696 55.377
459316 89.125 -62.711 55.368
459317 89.169 -62.722 55.361
459318 89.213 -62.731 55.355
459319 89.257 -62.739 55.350
459320 89.301 -62.747 55.344
459321 89.344 -62.756 55.340
459322 89.388 -62.765 55.333
459323 89.432 -62.773 55.324
459324 89.476 -62.782 55.316
459325 89.520 -62.790 55.307
459326 89.566 -62.798 55.300
459327 89.615 -62.807 55.293
459328 89.665 -62.816 55.289
459329 89.714 -62.824 55.287
459330 89.763 -62.833 55.288
459331 89.811 -62.842 55.290
459332 89.858 -62.852 55.291
459333 89.904 -62.862 55.290
459334 89.951 -62.871 55.287
459335 89.999 -62.881 55.285
459336 90.045 -62.891 55.280
459337 90.092 -62.900 55.272
459338 90.140 -62.911 55.261
459339 90.188 -62.925 55.246
459340 90.236 -62.938 55.227
459341 90.285 -62.953 55.205
459342 90.329 -62.964 55.185
459343 90.367 -62.975 55.169
459344 90.397 -62.986 55.155
459345 90.427 -62.998 55.142
459346 90.457 -63.009 55.131
459347 90.486 -63.020 55.121
459348 90.517 -63.032 55.111
459349 90.546 -63.043 55.101
459350 90.577 -63.054 55.089
459351 90.633 -63.065 55.076
459352 90.688 -63.076 55.061
459353 90.745 -63.086 55.045
459354 90.801 -63.097 55.027
459355 90.857 -63.108 55.008
459356 90.913 -63.118 54.989
459357 90.970 -63.129 54.969
459358 91.026 -63.140 54.949
459359 91.082 -63.151 54.929
459360 91.139 -63.161 54.911
459361 91.194 -63.172 54.893
459362 91.251 -63.183 54.877
459363 91.307 -63.193 54.863
459364 91.362 -63.205 54.851
459365 91.415 -63.216 54.842
459366 91.470 -63.224 54.832
459367 91.530 -63.226 54.821
459368 91.589 -63.228 54.809
459369 91.649 -63.230 54.799
459370 91.697 -63.230 54.791
459371 91.740 -63.230 54.784
459372 91.785 -63.230 54.778
459373 91.829 -63.230 54.774
459374 91.873 -63.230 54.769
459375 91.917 -63.230 54.768
459376 91.961 -63.230 54.771
459377 92.000 -63.232 54.779
459378 92.034 -63.234 54.792
459379 92.068 -63.237 54.803
459380 92.101 -63.239 54.807
459381 92.135 -63.242 54.805
459382 92.169 -63.245 54.796
459383 92.203 -63.247 54.784
459384 92.236 -63.250 54.770
459385 92.272 -63.252 54.754
459386 92.318 -63.257 54.736
459387 92.364 -63.261 54.719
459388 92.411 -63.266 54.704
459389 92.457 -63.270 54.692
459390 92.503 -63.274 54.683
459391 92.550 -63.279 54.676
459392 92.596 -63.283 54.671
459393 92.642 -63.289 54.666
459394 92.689 -63.294 54.659
459395 92.732 -63.302 54.651
459396 92.777 -63.310 54.643
459397 92.821 -63.318 54.634
459398 92.865 -63.325 54.625
459399 92.909 -63.333 54.618
459400 92.953 -63.341 54.610
459401 92.998 -63.349 54.601
459402 93.040 -63.356 54.592
459403 93.081 -63.364 54.581
459404 93.123 -63.371 54.567
459405 93.165 -63.378 54.552
459406 93.206 -63.386 54.536
459407 93.243 -63.392 54.519
459408 93.289 -63.392 54.502
459409 93.340 -63.389 54.484
459410 93.391 -63.385 54.465
459411 93.437 -63.378 54.445
459412 93.472 -63.368 54.425
459413 93.500 -63.361 54.404
459414 93.519 -63.361 54.384
459415 93.536 -63.366 54.365
459416 93.552 -63.371 54.346
459417 93.568 -63.376 54.326
459418 93.597 -63.398 54.306
459419 93.654 -63.430 54.283
459420 93.727 -63.461 54.258
459421 93.763 -63.475 54.231
459422 93.797 -63.496 54.204
459423 93.845 -63.499 54.176
459424 93.882 -63.510 54.146
459425 93.910 -63.526 54.114
459426 93.938 -63.542 54.080
459427 93.990 -63.553 54.048
459428 94.079 -63.557 54.017
459429 94.169 -63.561 53.988
459430 94.258 -63.565 53.962
459431 94.313 -63.571 53.938
459432 94.295 -63.583 53.916
459433 94.278 -63.594 53.898
459434 94.254 -63.606 53.883
459435 94.284 -63.616 53.869
459436 94.314 -63.626 53.858
459437 94.353 -63.641 53.847
459438 94.401 -63.661 53.836
459439 94.437 -63.675 53.825
459440 94.469 -63.686 53.817
459441 94.501 -63.697 53.810
459442 94.533 -63.709 53.803
459443 94.565 -63.720 53.797
459444 94.597 -63.731 53.788
459445 94.636 -63.738 53.777
459446 94.688 -63.739 53.764
459447 94.740 -63.739 53.750
459448 94.791 -63.740 53.735
459449 94.843 -63.740 53.717
459450 94.894 -63.741 53.697
459451 94.946 -63.741 53.676
459452 95.001 -63.741 53.655
459453 95.062 -63.743 53.635
459454 95.101 -63.747 53.613
459455 95.139 -63.753 53.592
459456 95.179 -63.759 53.573
459457 95.218 -63.766 53.554
459458 95.257 -63.772 53.536
459459 95.301 -63.779 53.518
459460 95.345 -63.785 53.501
459461 95.389 -63.792 53.485
459462 95.433 -63.799 53.470
459463 95.477 -63.806 53.449
459464 95.521 -63.813 53.429
459465 95.565 -63.819 53.408
459466 95.609 -63.825 53.385
459467 95.652 -63.831 53.359
459468 95.679 -63.821 53.329
459469 95.705 -63.810 53.295
459470 95.744 -63.806 53.260
459471 95.789 -63.806 53.225
459472 95.833 -63.805 53.187
459473 95.877 -63.804 53.150
459474 95.919 -63.791 53.117
459475 95.960 -63.766 53.083
459476 96.004 -63.748 53.049
459477 96.053 -63.744 53.016
459478 96.102 -63.738 52.986
459479 96.152 -63.733 52.956
459480 96.201 -63.729 52.925
459481 96.253 -63.738 52.895
459482 96.305 -63.748 52.864
459483 96.356 -63.757 52.831
459484 96.407 -63.766 52.798
459485 96.459 -63.775 52.766
459486 96.510 -63.785 52.733
459487 96.562 -63.794 52.702
459488 96.613 -63.803 52.671
459489 96.653 -63.808 52.639
459490 96.687 -63.810 52.607
459491 96.721 -63.813 52.573
459492 96.758 -63.817 52.537
459493 96.795 -63.822 52.500
459494 96.832 -63.826 52.460
459495 96.869 -63.831 52.420
459496 96.906 -63.836 52.380
459497 96.943 -63.841 52.342
459498 96.980 -63.845 52.305
459499 97.017 -63.850 52.268
459500 97.054 -63.855 52.231
459501 97.092 -63.860 52.192
459502 97.129 -63.865 52.152
459503 97.166 -63.869 52.110
459504 97.203 -63.874 52.068
459505 97.236 -63.885 52.027
459506 97.267 -63.897 51.986
459507 97.298 -63.910 51.948
459508 97.329 -63.922 51.912
459509 97.361 -63.935 51.878
459510 97.386 -63.948 51.845
459511 97.408 -63.963 51.812
459512 97.433 -63.976 51.776
459513 97.461 -63.983 51.738
459514 97.490 -63.990 51.698
459515 97.518 -63.997 51.656
459516 97.547 -64.004 51.614
459517 97.576 -64.010 51.569
459518 97.604 -64.017 51.522
459519 97.633 -64.024 51.473
459520 97.665 -64.027 51.422
459521 97.703 -64.024 51.371
459522 97.741 -64.021 51.318
459523 97.779 -64.017 51.266
459524 97.817 -64.014 51.215
459525 97.856 -64.011 51.163
459526 97.893 -64.009 51.110
459527 97.928 -64.010 51.059
459528 97.964 -64.010 51.008
459529 97.999 -64.011 50.957
459530 98.034 -64.013 50.904
459531 98.070 -64.013 50.849
459532 98.105 -64.014 50.791
459533 98.141 -64.015 50.734
459534 98.176 -64.016 50.679
459535 98.212 -64.016 50.625
459536 98.247 -64.017 50.570
459537 98.282 -64.018 50.517
459538 98.320 -64.013 50.464
459539 98.357 -64.010 50.413
459540 98.391 -64.011 50.363
459541 98.425 -64.013 50.313
459542 98.459 -64.016 50.266
459543 98.493 -64.018 50.219
459544 98.518 -64.020 50.173
459545 98.544 -64.023 50.128
459546 98.570 -64.026 50.082
459547 98.609 -64.028 50.038
459548 98.659 -64.030 49.993
459549 98.710 -64.032 49.949
459550 98.761 -64.034 49.904
459551 98.811 -64.035 49.858
459552 98.861 -64.038 49.814
459553 98.904 -64.041 49.766
459554 98.923 -64.047 49.717
459555 98.942 -64.053 49.665
459556 98.961 -64.059 49.614
459557 98.981 -64.065 49.566
459558 99.000 -64.071 49.520
459559 99.053 -64.077 49.475
459560 99.125 -64.082 49.431
459561 99.196 -64.087 49.388
459562 99.263 -64.088 49.347
459563 99.326 -64.087 49.306
459564 99.389 -64.085 49.266
459565 99.451 -64.083 49.227
459566 99.514 -64.082 49.188
459567 99.577 -64.079 49.151
459964 86.756 -62.356 56.191
459965 86.798 -62.355 56.179
459966 86.841 -62.353 56.167
459967 86.883 -62.352 56.156
459968 86.926 -62.350 56.146
459969 86.968 -62.348 56.137
459970 87.010 -62.346 56.128
459971 87.053 -62.345 56.119
459972 87.095 -62.343 56.110
459973 87.133 -62.344 56.101
459974 87.170 -62.346 56.093
459975 87.207 -62.348 56.085
459976 87.250 -62.347 56.075
459977 87.298 -62.343 56.061
459978 87.346 -62.340 56.044
459979 87.395 -62.337 56.022
459980 87.442 -62.333 55.998
459981 87.491 -62.330 55.974
459982 87.539 -62.327 55.949
459983 87.587 -62.323 55.926
459984 87.638 -62.318 55.904
459985 87.689 -62.312 55.883
459986 87.739 -62.306 55.862
459987 87.789 -62.300 55.844
459988 87.840 -62.295 55.829
459989 87.890 -62.289 55.817
459990 87.940 -62.283 55.806
459991 87.991 -62.277 55.796
459992 88.041 -62.280 55.791
459993 88.089 -62.296 55.788
459994 88.138 -62.313 55.788
459995 88.187 -62.329 55.789
459996 88.235 -62.346 55.787
459997 88.278 -62.361 55.778
459998 88.321 -62.376 55.760
459999 88.364 -62.390 55.735
460000 88.408 -62.406 55.708
460001 88.451 -62.421 55.680
460002 88.494 -62.436 55.649
460003 88.537 -62.451 55.615
460004 88.580 -62.466 55.585
460005 88.623 -62.481 55.558
460006 88.667 -62.496 55.534
460007 88.709 -62.512 55.516
460008 88.753 -62.527 55.496
460009 88.796 -62.542 55.477
460010 88.839 -62.556 55.458
460011 88.882 -62.572 55.440
460012 88.925 -62.587 55.425
460013 88.968 -62.602 55.411
460014 89.012 -62.617 55.397
460015 89.055 -62.632 55.385
460016 89.098 -62.647 55.375
460017 89.141 -62.658 55.367
460018 89.185 -62.666 55.360
460019 89.229 -62.675 55.353
460020 89.273 -62.683 55.347
460021 89.317 -62.691 55.339
460022 89.361 -62.700 55.336
460023 89.405 -62.709 55.328
460024 89.449 -62.717 55.319
460025 89.492 -62.725 55.310
460026 89.536 -62.734 55.301
460027 89.580 -62.743 55.293
460028 89.625 -62.751 55.285
460029 89.675 -62.760 55.279
460030 89.724 -62.768 55.275
460031 89.773 -62.777 55.273
460032 89.820 -62.786 55.274
460033 89.868 -62.796 55.275
460034 89.914 -62.806 55.274
460035 89.961 -62.815 55.273
460036 90.008 -62.825 55.271
460037 90.056 -62.836 55.268
460038 90.104 -62.850 55.262
460039 90.152 -62.864 55.253
460040 90.200 -62.878 55.240
460041 90.249 -62.892 55.223
460042 90.297 -62.906 55.203
460043 90.344 -62.919 55.184
460044 90.387 -62.929 55.166
460045 90.427 -62.939 55.150
460046 90.457 -62.951 55.135
460047 90.487 -62.962 55.123
460048 90.517 -62.973 55.112
460049 90.547 -62.985 55.099
460050 90.577 -62.996 55.087
460051 90.607 -63.007 55.075
460052 90.651 -63.018 55.062
460053 90.707 -63.029 55.048
460054 90.763 -63.039 55.032
460055 90.819 -63.051 55.016
460056 90.876 -63.061 54.998
460057 90.932 -63.072 54.981
460058 90.988 -63.083 54.964
460059 91.044 -63.093 54.946
460060 91.100 -63.104 54.929
460061 91.157 -63.114 54.912
460062 91.213 -63.125 54.896
460063 91.269 -63.136 54.881
460064 91.324 -63.147 54.868
460065 91.377 -63.158 54.857
460066 91.430 -63.170 54.848
460067 91.486 -63.177 54.837
460068 91.545 -63.180 54.826
460069 91.604 -63.182 54.814
460070 91.664 -63.184 54.803
460071 91.723 -63.186 54.796
460072 91.781 -63.189 54.790
460073 91.825 -63.189 54.785
460074 91.869 -63.189 54.781
460075 91.914 -63.189 54.776
460076 91.954 -63.189 54.774
460077 91.988 -63.192 54.776
460078 92.021 -63.195 54.784
460079 92.055 -63.197 54.798
460080 92.089 -63.199 54.810
460081 92.123 -63.202 54.816
460082 92.156 -63.205 54.815
460083 92.190 -63.208 54.808
460084 92.224 -63.210 54.798
460085 92.258 -63.212 54.785
460086 92.291 -63.215 54.768
460087 92.325 -63.217 54.749
460088 92.371 -63.222 54.730
460089 92.418 -63.227 54.713
460090 92.464 -63.231 54.698
460091 92.510 -63.235 54.686
460092 92.557 -63.239 54.677
460093 92.603 -63.244 54.670
460094 92.649 -63.249 54.664
460095 92.691 -63.257 54.657
460096 92.735 -63.264 54.648
460097 92.780 -63.271 54.639
460098 92.825 -63.280 54.630
460099 92.868 -63.287 54.619
460100 92.913 -63.295 54.610
460101 92.957 -63.303 54.600
460102 93.001 -63.311 54.588
460103 93.044 -63.318 54.576
460104 93.086 -63.326 54.563
460105 93.127 -63.333 54.548
460106 93.166 -63.339 54.532
460107 93.204 -63.346 54.515
460108 93.242 -63.352 54.496
460109 93.292 -63.348 54.478
460110 93.343 -63.345 54.459
460111 93.394 -63.341 54.438
460112 93.445 -63.338 54.418
460113 93.490 -63.339 54.399
460114 93.507 -63.346 54.380
460115 93.523 -63.351 54.360
460116 93.540 -63.355 54.342
460117 93.556 -63.360 54.323
460118 93.572 -63.365 54.304
460119 93.604 -63.390 54.285
460120 93.642 -63.417 54.261
460121 93.706 -63.430 54.236
460122 93.744 -63.440 54.209
460123 93.774 -63.466 54.181
460124 93.804 -63.492 54.154
460125 93.842 -63.509 54.121
460126 93.890 -63.512 54.088
460127 93.939 -63.514 54.057
460128 94.010 -63.515 54.025
460129 94.107 -63.517 53.995
460130 94.196 -63.521 53.969
460131 94.285 -63.525 53.945
460132 94.374 -63.530 53.923
460133 94.369 -63.540 53.904
460134 94.338 -63.553 53.888
460135 94.305 -63.566 53.873
460136 94.297 -63.578 53.861
460137 94.327 -63.587 53.849
460138 94.373 -63.606 53.837
460139 94.408 -63.619 53.827
460140 94.439 -63.631 53.816
460141 94.472 -63.642 53.807
460142 94.504 -63.653 53.799
460143 94.536 -63.665 53.791
460144 94.568 -63.676 53.785
460145 94.600 -63.687 53.776
460146 94.640 -63.694 53.765
460147 94.691 -63.695 53.753
460148 94.743 -63.695 53.739
460149 94.794 -63.696 53.726
460150 94.846 -63.696 53.709
460151 94.898 -63.697 53.690
460152 94.950 -63.697 53.669
460153 95.013 -63.698 53.649
460154 95.072 -63.700 53.629
460155 95.110 -63.704 53.606
460156 95.148 -63.709 53.585
460157 95.187 -63.714 53.564
460158 95.227 -63.721 53.544
460159 95.266 -63.728 53.525
460160 95.305 -63.734 53.507
460161 95.345 -63.741 53.490
460162 95.386 -63.747 53.473
460163 95.430 -63.754 53.457
460164 95.474 -63.761 53.439
460165 95.518 -63.768 53.420
460166 95.562 -63.774 53.399
460167 95.606 -63.781 53.377
460168 95.648 -63.785 53.351
460169 95.674 -63.775 53.322
460170 95.709 -63.769 53.290
460171 95.754 -63.768 53.256
460172 95.798 -63.767 53.220
460173 95.842 -63.766 53.181
460174 95.887 -63.765 53.142
460175 95.930 -63.757 53.105
460176 95.970 -63.734 53.069
460177 96.015 -63.717 53.033
460178 96.064 -63.712 52.998
460179 96.113 -63.707 52.965
460180 96.163 -63.702 52.934
460181 96.213 -63.697 52.903
460182 96.262 -63.692 52.872
460183 96.313 -63.701 52.839
460184 96.365 -63.710 52.806
460185 96.416 -63.719 52.773
460186 96.468 -63.729 52.739
460187 96.519 -63.738 52.706
460188 96.571 -63.747 52.673
460189 96.622 -63.756 52.641
460190 96.664 -63.763 52.608
460191 96.698 -63.764 52.576
460192 96.732 -63.767 52.543
460193 96.769 -63.772 52.508
460194 96.806 -63.776 52.474
460195 96.843 -63.781 52.435
460196 96.881 -63.786 52.396
460197 96.918 -63.791 52.359
460198 96.955 -63.795 52.322
460199 96.992 -63.800 52.286
460200 97.029 -63.805 52.251
460201 97.066 -63.810 52.215
460202 97.103 -63.814 52.177
460203 97.141 -63.819 52.139
460204 97.175 -63.828 52.098
460205 97.206 -63.840 52.057
460206 97.237 -63.853 52.017
460207 97.268 -63.865 51.978
460208 97.300 -63.877 51.941
460209 97.331 -63.890 51.906
460210 97.362 -63.903 51.872
460211 97.384 -63.917 51.839
460212 97.407 -63.932 51.806
460213 97.430 -63.946 51.772
460214 97.458 -63.954 51.734
460215 97.487 -63.961 51.694
460216 97.515 -63.968 51.652
460217 97.544 -63.975 51.610
460218 97.572 -63.982 51.565
460219 97.601 -63.988 51.519
460220 97.632 -63.993 51.469
460221 97.669 -63.990 51.418
460222 97.708 -63.987 51.366
460223 97.746 -63.984 51.312
460224 97.784 -63.980 51.259
460225 97.822 -63.977 51.206
460226 97.861 -63.975 51.153
460227 97.900 -63.973 51.100
460228 97.935 -63.975 51.048
460229 97.971 -63.975 50.998
460230 98.009 -63.976 50.948
460231 98.045 -63.976 50.895
460232 98.080 -63.977 50.840
460233 98.115 -63.978 50.784
460234 98.151 -63.979 50.726
460235 98.186 -63.979 50.670
460236 98.222 -63.981 50.615
460237 98.257 -63.982 50.559
460238 98.293 -63.982 50.506
460239 98.331 -63.976 50.453
460240 98.365 -63.977 50.402
460241 98.400 -63.979 50.352
460242 98.433 -63.982 50.302
460243 98.460 -63.984 50.255
460244 98.486 -63.987 50.209
460245 98.512 -63.990 50.164
460246 98.537 -63.992 50.118
460247 98.563 -63.995 50.075
460248 98.591 -63.998 50.031
460249 98.641 -64.000 49.987
460250 98.692 -64.001 49.943
460251 98.742 -64.004 49.899
460252 98.792 -64.005 49.854
460253 98.843 -64.007 49.809
460254 98.878 -64.011 49.760
460255 98.898 -64.017 49.709
460256 98.917 -64.023 49.655
460257 98.936 -64.029 49.602
460258 98.955 -64.035 49.554
460259 99.000 -64.041 49.507
460260 99.072 -64.046 49.462
460261 99.143 -64.051 49.418
460262 99.208 -64.052 49.375
460263 99.271 -64.050 49.334
460264 99.334 -64.048 49.294
460265 99.397 -64.046 49.254
460266 99.460 -64.045 49.215
460267 99.523 -64.043 49.177
460664 86.730 -62.309 56.207
460665 86.773 -62.308 56.195
460666 86.815 -62.306 56.182
460667 86.857 -62.305 56.171
460668 86.900 -62.303 56.160
460669 86.942 -62.301 56.150
460670 86.985 -62.299 56.141
460671 87.027 -62.298 56.132
460672 87.069 -62.296 56.123
460673 87.112 -62.295 56.115
460674 87.149 -62.296 56.106
460675 87.186 -62.298 56.100
460676 87.222 -62.300 56.094
460677 87.259 -62.302 56.086
460678 87.307 -62.299 56.073
460679 87.356 -62.295 56.058
460680 87.404 -62.292 56.037
460681 87.452 -62.288 56.013
460682 87.500 -62.285 55.989
460683 87.549 -62.281 55.963
460684 87.597 -62.278 55.939
460685 87.645 -62.275 55.916
460686 87.693 -62.271 55.894
460687 87.744 -62.266 55.874
460688 87.794 -62.260 55.854
460689 87.844 -62.255 55.838
460690 87.895 -62.249 55.825
460691 87.945 -62.243 55.813
460692 87.996 -62.237 55.802
460693 88.046 -62.232 55.794
460694 88.095 -62.245 55.790
460695 88.144 -62.261 55.788
460696 88.193 -62.278 55.788
460697 88.241 -62.294 55.786
460698 88.290 -62.311 55.778
460699 88.337 -62.327 55.761
460700 88.380 -62.342 55.739
460701 88.423 -62.357 55.713
460702 88.467 -62.372 55.683
460703 88.509 -62.387 55.650
460704 88.553 -62.402 55.615
460705 88.596 -62.418 55.585
460706 88.639 -62.432 55.558
460707 88.682 -62.447 55.536
460708 88.725 -62.462 55.516
460709 88.768 -62.478 55.497
460710 88.812 -62.493 55.478
460711 88.855 -62.508 55.460
460712 88.898 -62.523 55.443
460713 88.941 -62.538 55.427
460714 88.984 -62.553 55.413
460715 89.027 -62.569 55.399
460716 89.070 -62.584 55.386
460717 89.114 -62.594 55.375
460718 89.158 -62.602 55.365
460719 89.202 -62.610 55.357
460720 89.246 -62.619 55.350
460721 89.290 -62.627 55.343
460722 89.333 -62.636 55.334
460723 89.377 -62.644 55.330
460724 89.421 -62.653 55.322
460725 89.465 -62.661 55.313
460726 89.509 -62.669 55.304
460727 89.553 -62.678 55.295
460728 89.597 -62.687 55.286
460729 89.640 -62.695 55.277
460730 89.685 -62.703 55.270
460731 89.734 -62.713 55.264
460732 89.783 -62.721 55.259
460733 89.830 -62.731 55.258
460734 89.877 -62.740 55.258
460735 89.924 -62.750 55.259
460736 89.972 -62.761 55.258
460737 90.020 -62.775 55.257
460738 90.069 -62.789 55.255
460739 90.116 -62.804 55.251
460740 90.165 -62.817 55.243
460741 90.213 -62.831 55.232
460742 90.261 -62.845 55.217
460743 90.310 -62.859 55.199
460744 90.358 -62.873 55.182
460745 90.401 -62.884 55.165
460746 90.444 -62.894 55.150
460747 90.486 -62.904 55.134
460748 90.518 -62.915 55.119
460749 90.547 -62.926 55.105
460750 90.578 -62.938 55.091
460751 90.607 -62.949 55.078
460752 90.638 -62.960 55.065
460753 90.669 -62.971 55.052
460754 90.725 -62.982 55.038
460755 90.781 -62.993 55.023
460756 90.837 -63.004 55.008
460757 90.894 -63.014 54.992
460758 90.950 -63.025 54.977
460759 91.006 -63.035 54.961
460760 91.063 -63.046 54.945
460761 91.118 -63.057 54.929
460762 91.175 -63.068 54.914
460763 91.231 -63.079 54.900
460764 91.286 -63.089 54.887
460765 91.339 -63.101 54.874
460766 91.392 -63.112 54.864
460767 91.445 -63.123 54.854
460768 91.501 -63.132 54.842
460769 91.560 -63.134 54.831
460770 91.619 -63.136 54.819
460771 91.679 -63.139 54.807
460772 91.738 -63.141 54.800
460773 91.797 -63.143 54.795
460774 91.857 -63.145 54.791
460775 91.908 -63.147 54.787
460776 91.942 -63.149 54.782
460777 91.975 -63.152 54.781
460778 92.009 -63.155 54.782
460779 92.042 -63.157 54.790
460780 92.076 -63.160 54.804
460781 92.110 -63.162 54.818
460782 92.144 -63.164 54.825
460783 92.177 -63.167 54.826
460784 92.211 -63.170 54.822
460785 92.245 -63.173 54.814
460786 92.279 -63.175 54.802
460787 92.312 -63.177 54.785
460788 92.346 -63.180 54.766
460789 92.379 -63.183 54.745
460790 92.425 -63.187 54.725
460791 92.472 -63.192 54.705
460792 92.518 -63.195 54.689
460793 92.564 -63.200 54.677
460794 92.611 -63.205 54.667
460795 92.657 -63.209 54.659
460796 92.693 -63.221 54.651
460797 92.735 -63.230 54.642
460798 92.781 -63.236 54.633
460799 92.826 -63.243 54.623
460800 92.871 -63.250 54.613
460801 92.916 -63.258 54.602
460802 92.960 -63.264 54.589
460803 93.004 -63.273 54.575
460804 93.048 -63.280 54.560
460805 93.089 -63.288 54.543
460806 93.127 -63.294 54.528
460807 93.164 -63.300 54.510
460808 93.202 -63.306 54.492
460809 93.246 -63.308 54.472
460810 93.296 -63.305 54.452
460811 93.347 -63.301 54.433
460812 93.396 -63.299 54.412
460813 93.442 -63.297 54.392
460814 93.481 -63.304 54.374
460815 93.505 -63.324 54.355
460816 93.527 -63.339 54.336
460817 93.543 -63.344 54.318
460818 93.560 -63.349 54.301
460819 93.577 -63.355 54.283
460820 93.609 -63.381 54.264
460821 93.641 -63.408 54.240
460822 93.686 -63.398 54.214
460823 93.724 -63.406 54.187
460824 93.755 -63.432 54.158
460825 93.785 -63.458 54.128
460826 93.826 -63.477 54.097
460827 93.879 -63.465 54.066
460828 93.937 -63.462 54.035
460829 93.993 -63.481 54.006
460830 94.048 -63.500 53.978
460831 94.139 -63.506 53.954
460832 94.243 -63.506 53.932
460833 94.347 -63.506 53.913
460834 94.366 -63.510 53.895
460835 94.376 -63.515 53.880
460836 94.356 -63.526 53.865
460837 94.324 -63.539 53.852
460838 94.345 -63.552 53.841
460839 94.379 -63.565 53.829
460840 94.410 -63.576 53.818
460841 94.442 -63.587 53.807
460842 94.475 -63.599 53.797
460843 94.506 -63.610 53.789
460844 94.539 -63.622 53.781
460845 94.571 -63.633 53.775
460846 94.603 -63.644 53.764
460847 94.643 -63.650 53.753
460848 94.695 -63.651 53.741
460849 94.747 -63.651 53.728
460850 94.798 -63.652 53.715
460851 94.849 -63.652 53.700
460852 94.901 -63.653 53.682
460853 94.961 -63.653 53.662
460854 95.024 -63.655 53.643
460855 95.082 -63.657 53.622
460856 95.120 -63.661 53.599
460857 95.158 -63.665 53.576
460858 95.196 -63.670 53.555
460859 95.235 -63.676 53.533
460860 95.274 -63.683 53.514
460861 95.314 -63.689 53.495
460862 95.353 -63.696 53.477
460863 95.393 -63.703 53.459
460864 95.432 -63.709 53.443
460865 95.471 -63.716 53.426
460866 95.516 -63.722 53.407
460867 95.559 -63.729 53.387
460868 95.603 -63.736 53.366
460869 95.643 -63.739 53.340
460870 95.674 -63.731 53.313
460871 95.719 -63.730 53.283
460872 95.763 -63.728 53.250
460873 95.807 -63.728 53.213
460874 95.851 -63.727 53.174
460875 95.896 -63.726 53.134
460876 95.940 -63.725 53.094
460877 95.980 -63.700 53.055
460878 96.026 -63.686 53.016
460879 96.075 -63.681 52.980
460880 96.124 -63.676 52.944
460881 96.174 -63.671 52.911
460882 96.224 -63.666 52.879
460883 96.273 -63.661 52.847
460884 96.322 -63.656 52.814
460885 96.374 -63.663 52.779
460886 96.425 -63.673 52.745
460887 96.476 -63.682 52.711
460888 96.528 -63.691 52.676
460889 96.580 -63.700 52.641
460890 96.631 -63.710 52.607
460891 96.675 -63.716 52.574
460892 96.709 -63.719 52.542
460893 96.744 -63.722 52.509
460894 96.781 -63.726 52.474
460895 96.818 -63.731 52.441
460896 96.855 -63.736 52.406
460897 96.892 -63.741 52.370
460898 96.930 -63.745 52.335
460899 96.966 -63.750 52.300
460900 97.004 -63.755 52.265
460901 97.041 -63.759 52.230
460902 97.078 -63.764 52.195
460903 97.114 -63.770 52.158
460904 97.145 -63.783 52.122
460905 97.176 -63.795 52.084
460906 97.208 -63.808 52.045
460907 97.239 -63.820 52.007
460908 97.270 -63.833 51.970
460909 97.301 -63.845 51.932
460910 97.332 -63.858 51.897
460911 97.360 -63.871 51.864
460912 97.383 -63.885 51.830
460913 97.406 -63.900 51.797
460914 97.429 -63.915 51.764
460915 97.455 -63.925 51.728
460916 97.484 -63.932 51.688
460917 97.512 -63.939 51.646
460918 97.540 -63.946 51.604
460919 97.569 -63.953 51.559
460920 97.598 -63.960 51.511
460921 97.636 -63.956 51.461
460922 97.674 -63.953 51.409
460923 97.712 -63.950 51.356
460924 97.750 -63.946 51.302
460925 97.789 -63.944 51.248
460926 97.828 -63.941 51.194
460927 97.868 -63.940 51.140
460928 97.907 -63.938 51.087
460929 97.942 -63.938 51.035
460930 97.977 -63.940 50.986
460931 98.012 -63.941 50.935
460932 98.050 -63.941 50.883
460933 98.087 -63.942 50.828
460934 98.124 -63.942 50.771
460935 98.161 -63.943 50.714
460936 98.197 -63.944 50.657
460937 98.232 -63.944 50.602
460938 98.268 -63.945 50.547
460939 98.303 -63.946 50.494
460940 98.340 -63.943 50.442
460941 98.374 -63.945 50.391
460942 98.403 -63.948 50.341
460943 98.428 -63.951 50.293
460944 98.454 -63.954 50.247
460945 98.480 -63.956 50.202
460946 98.505 -63.959 50.156
460947 98.531 -63.961 50.111
460948 98.557 -63.964 50.067
460949 98.582 -63.966 50.024
460950 98.623 -63.969 49.982
460951 98.673 -63.971 49.939
460952 98.723 -63.973 49.895
460953 98.774 -63.975 49.851
460954 98.824 -63.976 49.802
460955 98.853 -63.982 49.751
460956 98.872 -63.988 49.697
460957 98.891 -63.994 49.642
460958 98.911 -64.000 49.590
460959 98.948 -64.006 49.541
460960 99.019 -64.010 49.495
460961 99.090 -64.016 49.450
460962 99.154 -64.015 49.405
460963 99.217 -64.013 49.363
460964 99.280 -64.011 49.321
460965 99.343 -64.010 49.282
460966 99.406 -64.007 49.242
460967 99.469 -64.006 49.204
460968 99.532 -64.004 49.166
461366 86.789 -62.259 56.197
461367 86.832 -62.258 56.185
461368 86.874 -62.255 56.174
461369 86.916 -62.254 56.163
461370 86.958 -62.252 56.153
461371 87.001 -62.251 56.144
461372 87.044 -62.249 56.135
461373 87.086 -62.248 56.127
461374 87.127 -62.247 56.119
461375 87.164 -62.249 56.111
461376 87.201 -62.250 56.105
461377 87.237 -62.252 56.099
461378 87.274 -62.254 56.092
461379 87.317 -62.253 56.082
461380 87.365 -62.249 56.067
461381 87.413 -62.246 56.048
461382 87.462 -62.243 56.025
461383 87.510 -62.239 56.001
461384 87.558 -62.236 55.975
461385 87.607 -62.233 55.950
461386 87.654 -62.230 55.928
461387 87.703 -62.226 55.906
461388 87.751 -62.223 55.885
461389 87.799 -62.219 55.865
461390 87.849 -62.214 55.848
461391 87.900 -62.208 55.833
461392 87.950 -62.203 55.819
461393 88.000 -62.197 55.808
461394 88.051 -62.192 55.798
461395 88.101 -62.194 55.791
461396 88.149 -62.211 55.787
461397 88.198 -62.227 55.785
461398 88.247 -62.243 55.782
461399 88.296 -62.260 55.772
461400 88.344 -62.276 55.757
461401 88.393 -62.293 55.737
461402 88.439 -62.308 55.711
461403 88.482 -62.323 55.680
461404 88.525 -62.338 55.645
461405 88.568 -62.353 55.612
461406 88.612 -62.368 55.582
461407 88.654 -62.384 55.557
461408 88.698 -62.399 55.535
461409 88.741 -62.414 55.515
461410 88.784 -62.429 55.497
461411 88.827 -62.444 55.480
461412 88.871 -62.459 55.463
461413 88.914 -62.474 55.444
461414 88.957 -62.489 55.429
461415 89.000 -62.504 55.415
461416 89.043 -62.519 55.401
461417 89.087 -62.529 55.388
461418 89.131 -62.537 55.375
461419 89.175 -62.546 55.364
461420 89.218 -62.554 55.354
461421 89.262 -62.563 55.346
461422 89.306 -62.572 55.339
461423 89.350 -62.580 55.329
461424 89.394 -62.588 55.324
461425 89.438 -62.597 55.316
461426 89.481 -62.606 55.307
461427 89.525 -62.614 55.297
461428 89.569 -62.622 55.288
461429 89.613 -62.631 55.279
461430 89.657 -62.639 55.270
461431 89.701 -62.647 55.262
461432 89.745 -62.656 55.255
461433 89.793 -62.665 55.250
461434 89.840 -62.675 55.247
461435 89.888 -62.687 55.246
461436 89.936 -62.700 55.245
461437 89.984 -62.714 55.245
461438 90.033 -62.728 55.244
461439 90.081 -62.742 55.242
461440 90.129 -62.757 55.238
461441 90.177 -62.770 55.232
461442 90.226 -62.785 55.222
461443 90.274 -62.798 55.210
461444 90.322 -62.812 55.195
461445 90.370 -62.826 55.180
461446 90.416 -62.838 55.166
461447 90.459 -62.848 55.152
461448 90.502 -62.858 55.135
461449 90.544 -62.868 55.118
461450 90.578 -62.879 55.102
461451 90.608 -62.890 55.086
461452 90.638 -62.901 55.072
461453 90.668 -62.913 55.057
461454 90.698 -62.924 55.044
461455 90.743 -62.935 55.030
461456 90.800 -62.946 55.016
461457 90.856 -62.957 55.002
461458 90.912 -62.967 54.988
461459 90.968 -62.978 54.974
461460 91.024 -62.989 54.960
461461 91.081 -63.000 54.947
461462 91.137 -63.010 54.933
461463 91.193 -63.021 54.919
461464 91.248 -63.032 54.907
461465 91.301 -63.043 54.895
461466 91.355 -63.054 54.883
461467 91.407 -63.066 54.873
461468 91.461 -63.078 54.862
461469 91.516 -63.086 54.850
461470 91.575 -63.088 54.837
461471 91.635 -63.090 54.824
461472 91.694 -63.092 54.813
461473 91.753 -63.095 54.804
461474 91.812 -63.097 54.798
461475 91.870 -63.101 54.795
461476 91.913 -63.110 54.792
461477 91.952 -63.110 54.788
461478 91.994 -63.113 54.787
461479 92.029 -63.117 54.789
461480 92.064 -63.120 54.797
461481 92.097 -63.122 54.812
461482 92.131 -63.125 54.826
461483 92.165 -63.127 54.835
461484 92.199 -63.130 54.838
461485 92.232 -63.133 54.837
461486 92.266 -63.135 54.831
461487 92.300 -63.138 54.820
461488 92.333 -63.140 54.803
461489 92.367 -63.142 54.783
461490 92.401 -63.145 54.761
461491 92.434 -63.148 54.738
461492 92.479 -63.152 54.715
461493 92.525 -63.156 54.694
461494 92.571 -63.161 54.676
461495 92.618 -63.165 54.663
461496 92.664 -63.170 54.652
461497 92.698 -63.183 54.643
461498 92.735 -63.195 54.633
461499 92.781 -63.202 54.623
461500 92.826 -63.208 54.613
461501 92.871 -63.215 54.602
461502 92.916 -63.223 54.590
461503 92.961 -63.230 54.576
461504 93.006 -63.236 54.561
461505 93.059 -63.239 54.543
461506 93.088 -63.248 54.525
461507 93.125 -63.254 54.506
461508 93.163 -63.260 54.486
461509 93.200 -63.266 54.466
461510 93.249 -63.264 54.445
461511 93.298 -63.261 54.425
461512 93.345 -63.260 54.406
461513 93.392 -63.258 54.386
461514 93.438 -63.256 54.367
461515 93.472 -63.268 54.348
461516 93.497 -63.289 54.331
461517 93.521 -63.310 54.312
461518 93.546 -63.331 54.294
461519 93.563 -63.338 54.278
461520 93.582 -63.347 54.262
461521 93.614 -63.371 54.242
461522 93.642 -63.368 54.219
461523 93.669 -63.364 54.193
461524 93.704 -63.372 54.165
461525 93.745 -63.392 54.134
461526 93.786 -63.412 54.104
461527 93.825 -63.418 54.074
461528 93.878 -63.404 54.044
461529 93.934 -63.423 54.016
461530 93.989 -63.442 53.989
461531 94.045 -63.462 53.963
461532 94.101 -63.480 53.942
461533 94.164 -63.497 53.922
461534 94.229 -63.499 53.904
461535 94.239 -63.504 53.888
461536 94.256 -63.507 53.873
461537 94.285 -63.506 53.859
461538 94.314 -63.505 53.847
461539 94.349 -63.510 53.836
461540 94.381 -63.521 53.824
461541 94.413 -63.532 53.813
461542 94.445 -63.543 53.802
461543 94.477 -63.555 53.791
461544 94.509 -63.566 53.782
461545 94.542 -63.578 53.772
461546 94.573 -63.589 53.765
461547 94.606 -63.600 53.754
461548 94.647 -63.606 53.742
461549 94.698 -63.607 53.729
461550 94.750 -63.607 53.715
461551 94.802 -63.608 53.702
461552 94.853 -63.608 53.688
461553 94.910 -63.609 53.671
461554 94.973 -63.610 53.652
461555 95.036 -63.612 53.633
461556 95.091 -63.614 53.611
461557 95.129 -63.618 53.588
461558 95.167 -63.622 53.566
461559 95.205 -63.627 53.543
461560 95.244 -63.631 53.521
461561 95.283 -63.638 53.500
461562 95.322 -63.644 53.481
461563 95.361 -63.651 53.462
461564 95.401 -63.658 53.443
461565 95.441 -63.665 53.426
461566 95.480 -63.672 53.408
461567 95.519 -63.678 53.391
461568 95.559 -63.684 53.372
461569 95.601 -63.691 53.351
461570 95.639 -63.693 53.328
461571 95.683 -63.691 53.302
461572 95.728 -63.690 53.274
461573 95.772 -63.690 53.242
461574 95.816 -63.689 53.206
461575 95.861 -63.688 53.167
461576 95.905 -63.687 53.125
461577 95.949 -63.686 53.084
461578 95.991 -63.668 53.042
461579 96.037 -63.655 53.000
461580 96.086 -63.650 52.961
461581 96.135 -63.645 52.923
461582 96.185 -63.640 52.887
461583 96.235 -63.635 52.854
461584 96.284 -63.630 52.821
461585 96.333 -63.625 52.788
461586 96.383 -63.620 52.752
461587 96.434 -63.626 52.716
461588 96.485 -63.635 52.680
461589 96.537 -63.644 52.644
461590 96.589 -63.654 52.608
461591 96.640 -63.663 52.573
461592 96.686 -63.670 52.539
461593 96.720 -63.672 52.507
461594 96.756 -63.676 52.474
461595 96.793 -63.681 52.441
461596 96.830 -63.685 52.408
461597 96.867 -63.690 52.375
461598 96.904 -63.695 52.341
461599 96.941 -63.700 52.308
461600 96.978 -63.704 52.274
461601 97.015 -63.709 52.240
461602 97.052 -63.714 52.206
461603 97.084 -63.726 52.171
461604 97.115 -63.738 52.136
461605 97.147 -63.750 52.101
461606 97.178 -63.763 52.066
461607 97.209 -63.775 52.030
461608 97.240 -63.788 51.994
461609 97.272 -63.800 51.958
461610 97.303 -63.813 51.921
461611 97.334 -63.825 51.887
461612 97.359 -63.840 51.853
461613 97.382 -63.854 51.820
461614 97.404 -63.869 51.788
461615 97.427 -63.883 51.755
461616 97.451 -63.896 51.719
461617 97.480 -63.903 51.679
461618 97.509 -63.910 51.637
461619 97.537 -63.917 51.593
461620 97.566 -63.924 51.548
461621 97.602 -63.922 51.499
461622 97.641 -63.919 51.449
461623 97.679 -63.916 51.397
461624 97.717 -63.913 51.344
461625 97.756 -63.910 51.290
461626 97.796 -63.909 51.236
461627 97.835 -63.907 51.181
461628 97.874 -63.904 51.127
461629 97.914 -63.902 51.072
461630 97.949 -63.903 51.021
461631 97.984 -63.904 50.971
461632 98.018 -63.905 50.920
461633 98.053 -63.907 50.867
461634 98.090 -63.907 50.813
461635 98.128 -63.907 50.757
461636 98.165 -63.908 50.700
461637 98.202 -63.909 50.645
461638 98.239 -63.909 50.591
461639 98.277 -63.910 50.538
461640 98.313 -63.910 50.485
461641 98.344 -63.912 50.433
461642 98.370 -63.914 50.382
461643 98.395 -63.917 50.333
461644 98.421 -63.919 50.284
461645 98.447 -63.922 50.238
461646 98.473 -63.925 50.193
461647 98.498 -63.928 50.147
461648 98.524 -63.930 50.102
461649 98.550 -63.933 50.059
461650 98.575 -63.935 50.016
461651 98.604 -63.938 49.974
461652 98.654 -63.940 49.932
461653 98.705 -63.942 49.888
461654 98.755 -63.944 49.842
461655 98.805 -63.946 49.792
461656 98.828 -63.951 49.738
461657 98.847 -63.958 49.683
461658 98.866 -63.964 49.628
461659 98.895 -63.970 49.577
461660 98.966 -63.975 49.529
461661 99.037 -63.979 49.483
461662 99.099 -63.978 49.438
461663 99.162 -63.976 49.394
461664 99.225 -63.974 49.352
461665 99.288 -63.973 49.310
461666 99.351 -63.970 49.271
461667 99.414 -63.969 49.232
461668 99.477 -63.967 49.193
462067 86.805 -62.211 56.201
462068 86.848 -62.208 56.189
462069 86.890 -62.207 56.177
462070 86.933 -62.205 56.166
462071 86.975 -62.204 56.155
462072 87.017 -62.202 56.147
462073 87.060 -62.201 56.138
462074 87.103 -62.199 56.129
462075 87.142 -62.199 56.122
462076 87.179 -62.201 56.115
462077 87.216 -62.202 56.108
462078 87.253 -62.204 56.102
462079 87.289 -62.206 56.096
462080 87.326 -62.208 56.087
462081 87.374 -62.205 56.074
462082 87.423 -62.201 56.057
462083 87.471 -62.198 56.036
462084 87.519 -62.194 56.013
462085 87.568 -62.191 55.988
462086 87.616 -62.187 55.963
462087 87.664 -62.184 55.939
462088 87.712 -62.181 55.916
462089 87.761 -62.177 55.895
462090 87.809 -62.174 55.875
462091 87.857 -62.171 55.857
462092 87.905 -62.167 55.841
462093 87.955 -62.162 55.827
462094 88.005 -62.157 55.814
462095 88.056 -62.151 55.802
462096 88.106 -62.145 55.792
462097 88.155 -62.159 55.785
462098 88.204 -62.176 55.780
462099 88.253 -62.192 55.773
462100 88.302 -62.208 55.764
462101 88.350 -62.225 55.751
462102 88.399 -62.241 55.732
462103 88.448 -62.258 55.708
462104 88.496 -62.274 55.677
462105 88.541 -62.290 55.642
462106 88.584 -62.305 55.609
462107 88.627 -62.320 55.580
462108 88.671 -62.335 55.555
462109 88.713 -62.350 55.534
462110 88.757 -62.365 55.515
462111 88.800 -62.380 55.498
462112 88.843 -62.395 55.481
462113 88.886 -62.410 55.464
462114 88.929 -62.425 55.447
462115 88.972 -62.440 55.431
462116 89.016 -62.456 55.416
462117 89.059 -62.465 55.402
462118 89.103 -62.473 55.389
462119 89.147 -62.481 55.375
462120 89.191 -62.490 55.362
462121 89.235 -62.499 55.352
462122 89.279 -62.507 55.344
462123 89.323 -62.515 55.334
462124 89.367 -62.524 55.325
462125 89.410 -62.532 55.318
462126 89.454 -62.541 55.311
462127 89.498 -62.550 55.302
462128 89.542 -62.558 55.291
462129 89.586 -62.566 55.282
462130 89.629 -62.575 55.273
462131 89.674 -62.584 55.264
462132 89.718 -62.592 55.256
462133 89.760 -62.601 55.249
462134 89.803 -62.612 55.243
462135 89.852 -62.625 55.238
462136 89.900 -62.640 55.235
462137 89.949 -62.653 55.232
462138 89.997 -62.668 55.232
462139 90.045 -62.681 55.231
462140 90.094 -62.696 55.229
462141 90.142 -62.710 55.225
462142 90.190 -62.723 55.220
462143 90.238 -62.738 55.212
462144 90.287 -62.751 55.202
462145 90.334 -62.766 55.191
462146 90.383 -62.779 55.180
462147 90.431 -62.793 55.169
462148 90.474 -62.803 55.156
462149 90.516 -62.813 55.140
462150 90.559 -62.822 55.122
462151 90.602 -62.832 55.103
462152 90.639 -62.843 55.085
462153 90.668 -62.854 55.067
462154 90.698 -62.866 55.052
462155 90.728 -62.877 55.038
462156 90.761 -62.888 55.024
462157 90.818 -62.899 55.011
462158 90.874 -62.910 54.997
462159 90.930 -62.921 54.986
462160 90.987 -62.932 54.973
462161 91.042 -62.942 54.960
462162 91.099 -62.953 54.949
462163 91.155 -62.963 54.937
462164 91.211 -62.974 54.926
462165 91.263 -62.986 54.915
462166 91.317 -62.998 54.904
462167 91.370 -63.009 54.892
462168 91.423 -63.020 54.882
462169 91.476 -63.032 54.870
462170 91.531 -63.040 54.857
462171 91.590 -63.042 54.842
462172 91.650 -63.045 54.828
462173 91.709 -63.047 54.816
462174 91.769 -63.049 54.806
462175 91.826 -63.052 54.800
462176 91.872 -63.063 54.796
462177 91.917 -63.074 54.794
462178 91.951 -63.070 54.792
462179 91.991 -63.072 54.791
462180 92.033 -63.075 54.795
462181 92.076 -63.079 54.805
462182 92.114 -63.082 54.822
462183 92.149 -63.086 54.837
462184 92.185 -63.089 54.848
462185 92.220 -63.092 54.853
462186 92.253 -63.095 54.854
462187 92.287 -63.098 54.849
462188 92.320 -63.100 54.838
462189 92.354 -63.103 54.821
462190 92.388 -63.105 54.799
462191 92.422 -63.108 54.776
462192 92.455 -63.111 54.750
462193 92.489 -63.113 54.725
462194 92.532 -63.117 54.699
462195 92.579 -63.121 54.677
462196 92.625 -63.126 54.659
462197 92.670 -63.131 54.645
462198 92.704 -63.144 54.633
462199 92.738 -63.158 54.622
462200 92.781 -63.167 54.612
462201 92.826 -63.173 54.601
462202 92.871 -63.180 54.587
462203 92.916 -63.187 54.574
462204 92.961 -63.195 54.560
462205 93.012 -63.198 54.543
462206 93.080 -63.193 54.525
462207 93.122 -63.198 54.506
462208 93.147 -63.208 54.485
462209 93.171 -63.217 54.463
462210 93.200 -63.224 54.440
462211 93.247 -63.223 54.418
462212 93.294 -63.221 54.398
462213 93.341 -63.220 54.379
462214 93.388 -63.218 54.359
462215 93.434 -63.216 54.340
462216 93.464 -63.233 54.322
462217 93.488 -63.253 54.304
462218 93.512 -63.274 54.286
462219 93.537 -63.296 54.270
462220 93.562 -63.316 54.254
462221 93.587 -63.330 54.237
462222 93.617 -63.330 54.217
462223 93.647 -63.329 54.194
462224 93.677 -63.328 54.168
462225 93.709 -63.331 54.141
462226 93.747 -63.346 54.112
462227 93.789 -63.365 54.082
462228 93.824 -63.357 54.053
462229 93.875 -63.365 54.025
462230 93.931 -63.383 53.998
462231 93.987 -63.399 53.974
462232 94.044 -63.416 53.951
462233 94.101 -63.433 53.932
462234 94.157 -63.450 53.914
462235 94.191 -63.465 53.898
462236 94.220 -63.465 53.882
462237 94.249 -63.463 53.868
462238 94.278 -63.462 53.856
462239 94.312 -63.466 53.844
462240 94.347 -63.472 53.832
462241 94.383 -63.478 53.821
462242 94.416 -63.488 53.810
462243 94.449 -63.499 53.799
462244 94.480 -63.511 53.788
462245 94.512 -63.522 53.776
462246 94.544 -63.534 53.766
462247 94.576 -63.545 53.756
462248 94.609 -63.556 53.744
462249 94.650 -63.562 53.732
462250 94.702 -63.563 53.718
462251 94.753 -63.563 53.704
462252 94.805 -63.564 53.689
462253 94.859 -63.565 53.674
462254 94.922 -63.566 53.657
462255 94.984 -63.567 53.638
462256 95.047 -63.568 53.618
462257 95.101 -63.571 53.597
462258 95.139 -63.575 53.575
462259 95.177 -63.580 53.552
462260 95.215 -63.584 53.529
462261 95.253 -63.588 53.506
462262 95.291 -63.593 53.484
462263 95.331 -63.600 53.463
462264 95.370 -63.606 53.443
462265 95.410 -63.613 53.424
462266 95.449 -63.620 53.405
462267 95.488 -63.626 53.387
462268 95.528 -63.633 53.368
462269 95.567 -63.640 53.349
462270 95.607 -63.647 53.329
462271 95.646 -63.648 53.308
462272 95.692 -63.651 53.284
462273 95.737 -63.652 53.256
462274 95.781 -63.650 53.226
462275 95.825 -63.650 53.193
462276 95.870 -63.649 53.156
462277 95.914 -63.648 53.115
462278 95.959 -63.647 53.073
462279 96.001 -63.634 53.029
462280 96.048 -63.625 52.985
462281 96.097 -63.619 52.943
462282 96.147 -63.614 52.902
462283 96.196 -63.609 52.863
462284 96.246 -63.604 52.827
462285 96.295 -63.599 52.793
462286 96.344 -63.594 52.758
462287 96.394 -63.589 52.723
462288 96.444 -63.584 52.686
462289 96.495 -63.588 52.648
462290 96.546 -63.598 52.610
462291 96.597 -63.607 52.574
462292 96.649 -63.616 52.538
462293 96.697 -63.625 52.504
462294 96.731 -63.627 52.471
462295 96.767 -63.631 52.439
462296 96.805 -63.635 52.407
462297 96.841 -63.640 52.376
462298 96.879 -63.645 52.343
462299 96.916 -63.650 52.310
462300 96.953 -63.654 52.279
462301 96.990 -63.659 52.246
462302 97.023 -63.669 52.213
462303 97.054 -63.681 52.179
462304 97.086 -63.694 52.145
462305 97.117 -63.706 52.111
462306 97.148 -63.719 52.078
462307 97.179 -63.731 52.045
462308 97.211 -63.744 52.012
462309 97.242 -63.756 51.978
462310 97.274 -63.769 51.944
462311 97.305 -63.781 51.908
462312 97.334 -63.794 51.874
462313 97.357 -63.808 51.841
462314 97.380 -63.822 51.809
462315 97.403 -63.838 51.777
462316 97.426 -63.852 51.744
462317 97.449 -63.866 51.708
462318 97.477 -63.874 51.668
462319 97.505 -63.881 51.626
462320 97.534 -63.888 51.581
462321 97.569 -63.888 51.535
462322 97.607 -63.885 51.486
462323 97.645 -63.882 51.436
462324 97.684 -63.879 51.384
462325 97.724 -63.878 51.332
462326 97.763 -63.875 51.277
462327 97.803 -63.873 51.223
462328 97.842 -63.871 51.168
462329 97.881 -63.869 51.113
462330 97.921 -63.867 51.058
462331 97.956 -63.867 51.006
462332 97.991 -63.869 50.955
462333 98.025 -63.869 50.904
462334 98.060 -63.871 50.852
462335 98.095 -63.872 50.798
462336 98.131 -63.872 50.744
462337 98.168 -63.873 50.688
462338 98.206 -63.874 50.634
462339 98.242 -63.875 50.581
462340 98.276 -63.878 50.530
462341 98.309 -63.882 50.478
462342 98.340 -63.885 50.425
462343 98.364 -63.885 50.374
462344 98.389 -63.886 50.324
462345 98.415 -63.888 50.275
462346 98.440 -63.891 50.228
462347 98.466 -63.894 50.183
462348 98.492 -63.897 50.137
462349 98.517 -63.900 50.091
462350 98.543 -63.902 50.048
462351 98.569 -63.905 50.006
462352 98.594 -63.907 49.964
462353 98.636 -63.910 49.921
462354 98.686 -63.911 49.877
462355 98.736 -63.913 49.831
462356 98.783 -63.916 49.779
462357 98.802 -63.922 49.723
462358 98.821 -63.928 49.668
462359 98.842 -63.934 49.614
462360 98.914 -63.939 49.564
462361 98.982 -63.943 49.517
462362 99.045 -63.941 49.471
462363 99.108 -63.939 49.427
462364 99.171 -63.937 49.383
462365 99.234 -63.935 49.341
462366 99.297 -63.934 49.299
462367 99.360 -63.932 49.260
462368 99.423 -63.930 49.221
462369 99.486 -63.928 49.183
462769 86.864 -62.160 56.192
462770 86.907 -62.158 56.179
462771 86.949 -62.157 56.168
462772 86.992 -62.155 56.158
462773 87.034 -62.154 56.148
462774 87.076 -62.152 56.139
462775 87.119 -62.151 56.131
462776 87.158 -62.151 56.123
462777 87.194 -62.153 56.116
462778 87.231 -62.155 56.108
462779 87.268 -62.156 56.101
462780 87.304 -62.158 56.094
462781 87.341 -62.160 56.088
462782 87.384 -62.159 56.077
462783 87.432 -62.155 56.063
462784 87.481 -62.152 56.046
462785 87.529 -62.149 56.024
462786 87.577 -62.145 56.001
462787 87.625 -62.142 55.975
462788 87.674 -62.139 55.950
462789 87.722 -62.136 55.926
462790 87.770 -62.132 55.904
462791 87.819 -62.129 55.883
462792 87.867 -62.125 55.865
462793 87.915 -62.122 55.848
462794 87.963 -62.119 55.833
462795 88.011 -62.115 55.819
462796 88.061 -62.111 55.806
462797 88.111 -62.105 55.796
462798 88.161 -62.108 55.787
462799 88.210 -62.124 55.778
462800 88.258 -62.141 55.769
462801 88.307 -62.158 55.759
462802 88.356 -62.174 55.746
462803 88.405 -62.190 55.729
462804 88.454 -62.207 55.706
462805 88.502 -62.223 55.676
462806 88.551 -62.239 55.642
462807 88.600 -62.255 55.608
462808 88.643 -62.271 55.578
462809 88.686 -62.286 55.553
462810 88.730 -62.301 55.533
462811 88.772 -62.316 55.515
462812 88.816 -62.331 55.499
462813 88.859 -62.346 55.482
462814 88.902 -62.362 55.465
462815 88.945 -62.377 55.448
462816 88.988 -62.392 55.432
462817 89.032 -62.400 55.417
462818 89.076 -62.409 55.403
462819 89.120 -62.417 55.389
462820 89.164 -62.425 55.375
462821 89.207 -62.434 55.362
462822 89.251 -62.443 55.351
462823 89.295 -62.451 55.340
462824 89.339 -62.459 55.330
462825 89.383 -62.468 55.320
462826 89.427 -62.477 55.313
462827 89.471 -62.485 55.305
462828 89.515 -62.493 55.296
462829 89.557 -62.503 55.286
462830 89.600 -62.512 55.277
462831 89.643 -62.522 55.267
462832 89.685 -62.531 55.259
462833 89.728 -62.541 55.251
462834 89.770 -62.550 55.244
462835 89.814 -62.563 55.238
462836 89.863 -62.580 55.231
462837 89.913 -62.593 55.225
462838 89.961 -62.606 55.221
462839 90.010 -62.621 55.220
462840 90.058 -62.634 55.218
462841 90.106 -62.649 55.216
462842 90.154 -62.663 55.213
462843 90.202 -62.677 55.209
462844 90.251 -62.691 55.204
462845 90.299 -62.704 55.196
462846 90.347 -62.719 55.190
462847 90.395 -62.732 55.183
462848 90.444 -62.747 55.174
462849 90.488 -62.758 55.162
462850 90.531 -62.768 55.146
462851 90.574 -62.778 55.127
462852 90.616 -62.788 55.107
462853 90.659 -62.797 55.086
462854 90.699 -62.807 55.066
462855 90.729 -62.819 55.049
462856 90.759 -62.830 55.034
462857 90.789 -62.841 55.020
462858 90.836 -62.853 55.008
462859 90.892 -62.863 54.995
462860 90.948 -62.874 54.984
462861 91.005 -62.885 54.972
462862 91.061 -62.895 54.961
462863 91.117 -62.906 54.952
462864 91.173 -62.917 54.941
462865 91.226 -62.929 54.932
462866 91.279 -62.940 54.923
462867 91.332 -62.951 54.913
462868 91.385 -62.963 54.901
462869 91.438 -62.974 54.890
462870 91.492 -62.985 54.877
462871 91.546 -62.994 54.862
462872 91.606 -62.996 54.846
462873 91.665 -62.998 54.831
462874 91.724 -63.001 54.818
462875 91.784 -63.003 54.809
462876 91.828 -63.014 54.802
462877 91.874 -63.025 54.797
462878 91.919 -63.036 54.795
462879 91.955 -63.036 54.795
462880 91.990 -63.032 54.796
462881 92.030 -63.034 54.802
462882 92.073 -63.037 54.813
462883 92.116 -63.041 54.831
462884 92.158 -63.045 54.847
462885 92.198 -63.048 54.859
462886 92.234 -63.051 54.866
462887 92.270 -63.055 54.868
462888 92.305 -63.059 54.863
462889 92.341 -63.063 54.851
462890 92.375 -63.065 54.834
462891 92.409 -63.068 54.813
462892 92.443 -63.070 54.788
462893 92.477 -63.073 54.762
462894 92.510 -63.076 54.734
462895 92.544 -63.078 54.706
462896 92.586 -63.082 54.680
462897 92.632 -63.086 54.657
462898 92.676 -63.092 54.638
462899 92.710 -63.106 54.623
462900 92.744 -63.120 54.609
462901 92.781 -63.132 54.597
462902 92.826 -63.139 54.585
462903 92.871 -63.145 54.570
462904 92.916 -63.152 54.556
462905 92.966 -63.157 54.541
462906 93.034 -63.152 54.523
462907 93.083 -63.158 54.505
462908 93.118 -63.167 54.486
462909 93.142 -63.177 54.464
462910 93.166 -63.188 54.440
462911 93.211 -63.185 54.416
462912 93.243 -63.183 54.393
462913 93.290 -63.181 54.373
462914 93.337 -63.180 54.353
462915 93.383 -63.177 54.334
462916 93.430 -63.176 54.316
462917 93.455 -63.197 54.298
462918 93.479 -63.217 54.279
462919 93.504 -63.239 54.262
462920 93.529 -63.260 54.245
462921 93.555 -63.279 54.228
462922 93.594 -63.290 54.210
462923 93.631 -63.299 54.190
462924 93.661 -63.299 54.166
462925 93.691 -63.298 54.141
462926 93.721 -63.297 54.115
462927 93.758 -63.310 54.088
462928 93.796 -63.324 54.061
462929 93.833 -63.313 54.035
462930 93.878 -63.308 54.009
462931 93.934 -63.326 53.985
462932 93.991 -63.343 53.962
462933 94.047 -63.359 53.943
462934 94.104 -63.376 53.925
462935 94.151 -63.402 53.909
462936 94.184 -63.422 53.893
462937 94.213 -63.421 53.879
462938 94.242 -63.420 53.865
462939 94.274 -63.423 53.854
462940 94.310 -63.429 53.843
462941 94.346 -63.435 53.831
462942 94.382 -63.441 53.821
462943 94.418 -63.447 53.810
462944 94.451 -63.456 53.798
462945 94.483 -63.468 53.786
462946 94.516 -63.479 53.774
462947 94.547 -63.490 53.761
462948 94.579 -63.502 53.749
462949 94.611 -63.513 53.735
462950 94.653 -63.518 53.721
462951 94.705 -63.519 53.708
462952 94.757 -63.519 53.693
462953 94.809 -63.520 53.676
462954 94.871 -63.521 53.659
462955 94.933 -63.523 53.642
462956 94.996 -63.524 53.622
462957 95.059 -63.525 53.602
462958 95.110 -63.528 53.581
462959 95.148 -63.533 53.558
462960 95.186 -63.537 53.535
462961 95.225 -63.541 53.512
462962 95.262 -63.545 53.489
462963 95.301 -63.549 53.466
462964 95.339 -63.555 53.444
462965 95.379 -63.562 53.424
462966 95.418 -63.568 53.404
462967 95.458 -63.575 53.384
462968 95.497 -63.581 53.364
462969 95.536 -63.588 53.344
462970 95.576 -63.595 53.324
462971 95.609 -63.595 53.305
462972 95.650 -63.596 53.284
462973 95.696 -63.600 53.262
462974 95.742 -63.603 53.235
462975 95.788 -63.606 53.205
462976 95.834 -63.609 53.173
462977 95.879 -63.610 53.139
462978 95.923 -63.609 53.101
462979 95.968 -63.608 53.061
462980 96.011 -63.602 53.017
462981 96.059 -63.593 52.972
462982 96.108 -63.588 52.926
462983 96.158 -63.583 52.883
462984 96.207 -63.578 52.841
462985 96.257 -63.573 52.802
462986 96.306 -63.568 52.764
462987 96.356 -63.563 52.728
462988 96.405 -63.558 52.691
462989 96.455 -63.553 52.653
462990 96.504 -63.548 52.615
462991 96.555 -63.551 52.576
462992 96.606 -63.560 52.539
462993 96.658 -63.569 52.502
462994 96.709 -63.578 52.468
462995 96.742 -63.581 52.435
462996 96.779 -63.585 52.404
462997 96.816 -63.590 52.372
462998 96.853 -63.595 52.341
462999 96.890 -63.600 52.309
463000 96.928 -63.604 52.278
463001 96.962 -63.612 52.245
463002 96.994 -63.624 52.213
463003 97.025 -63.637 52.181
463004 97.056 -63.649 52.148
463005 97.087 -63.662 52.114
463006 97.119 -63.674 52.081
463007 97.150 -63.687 52.049
463008 97.181 -63.699 52.018
463009 97.213 -63.712 51.988
463010 97.244 -63.724 51.958
463011 97.275 -63.736 51.926
463012 97.306 -63.749 51.891
463013 97.333 -63.763 51.859
463014 97.356 -63.777 51.827
463015 97.379 -63.791 51.795
463016 97.402 -63.806 51.763
463017 97.425 -63.820 51.729
463018 97.447 -63.835 51.693
463019 97.473 -63.845 51.654
463020 97.502 -63.852 51.611
463021 97.535 -63.854 51.566
463022 97.574 -63.850 51.521
463023 97.613 -63.848 51.472
463024 97.652 -63.847 51.422
463025 97.691 -63.844 51.371
463026 97.731 -63.842 51.319
463027 97.770 -63.840 51.266
463028 97.809 -63.838 51.211
463029 97.849 -63.836 51.155
463030 97.888 -63.834 51.099
463031 97.927 -63.831 51.044
463032 97.963 -63.831 50.989
463033 97.998 -63.833 50.938
463034 98.033 -63.834 50.888
463035 98.067 -63.835 50.837
463036 98.102 -63.836 50.786
463037 98.137 -63.838 50.734
463038 98.168 -63.842 50.679
463039 98.201 -63.846 50.625
463040 98.234 -63.849 50.573
463041 98.268 -63.853 50.522
463042 98.301 -63.857 50.470
463043 98.335 -63.860 50.417
463044 98.362 -63.861 50.366
463045 98.385 -63.860 50.316
463046 98.409 -63.860 50.265
463047 98.434 -63.860 50.217
463048 98.459 -63.863 50.170
463049 98.485 -63.866 50.124
463050 98.511 -63.869 50.079
463051 98.536 -63.871 50.034
463052 98.562 -63.874 49.992
463053 98.588 -63.876 49.948
463054 98.617 -63.879 49.906
463055 98.667 -63.881 49.863
463056 98.718 -63.883 49.817
463057 98.757 -63.886 49.763
463058 98.777 -63.892 49.707
463059 98.796 -63.898 49.652
463060 98.861 -63.904 49.600
463061 98.928 -63.906 49.551
463062 98.991 -63.904 49.504
463063 99.054 -63.902 49.459
463064 99.117 -63.900 49.415
463065 99.180 -63.898 49.372
463066 99.242 -63.897 49.330
463067 99.305 -63.894 49.289
463068 99.368 -63.893 49.249
463069 99.431 -63.891 49.211
463470 86.881 -62.111 56.194
463471 86.923 -62.110 56.182
463472 86.966 -62.108 56.171
463473 87.008 -62.107 56.160
463474 87.051 -62.105 56.149
463475 87.093 -62.103 56.140
463476 87.135 -62.102 56.132
463477 87.173 -62.103 56.124
463478 87.210 -62.105 56.115
463479 87.247 -62.107 56.107
463480 87.283 -62.108 56.099
463481 87.320 -62.110 56.092
463482 87.356 -62.112 56.086
463483 87.394 -62.114 56.078
463484 87.442 -62.111 56.067
463485 87.490 -62.107 56.053
463486 87.539 -62.104 56.033
463487 87.586 -62.100 56.011
463488 87.635 -62.097 55.986
463489 87.683 -62.093 55.960
463490 87.731 -62.090 55.935
463491 87.780 -62.087 55.911
463492 87.828 -62.083 55.890
463493 87.876 -62.080 55.871
463494 87.924 -62.077 55.853
463495 87.973 -62.073 55.838
463496 88.021 -62.070 55.823
463497 88.069 -62.067 55.810
463498 88.118 -62.064 55.800
463499 88.167 -62.059 55.789
463500 88.215 -62.073 55.779
463501 88.264 -62.090 55.767
463502 88.313 -62.106 55.755
463503 88.361 -62.123 55.741
463504 88.410 -62.139 55.723
463505 88.459 -62.155 55.699
463506 88.508 -62.172 55.668
463507 88.557 -62.188 55.634
463508 88.606 -62.205 55.603
463509 88.654 -62.221 55.574
463510 88.702 -62.237 55.551
463511 88.745 -62.252 55.532
463512 88.788 -62.268 55.515
463513 88.831 -62.283 55.499
463514 88.875 -62.298 55.482
463515 88.918 -62.312 55.466
463516 88.961 -62.327 55.449
463517 89.005 -62.336 55.433
463518 89.049 -62.344 55.418
463519 89.093 -62.352 55.403
463520 89.136 -62.361 55.389
463521 89.180 -62.370 55.374
463522 89.224 -62.378 55.361
463523 89.268 -62.387 55.349
463524 89.312 -62.395 55.337
463525 89.355 -62.404 55.326
463526 89.397 -62.414 55.316
463527 89.440 -62.424 55.309
463528 89.482 -62.433 55.301
463529 89.525 -62.443 55.291
463530 89.567 -62.452 55.281
463531 89.610 -62.462 55.272
463532 89.653 -62.471 55.264
463533 89.695 -62.481 55.256
463534 89.738 -62.490 55.249
463535 89.780 -62.500 55.242
463536 89.824 -62.515 55.235
463537 89.871 -62.534 55.229
463538 89.922 -62.547 55.223
463539 89.974 -62.560 55.217
463540 90.022 -62.574 55.213
463541 90.071 -62.587 55.207
463542 90.118 -62.602 55.203
463543 90.167 -62.616 55.201
463544 90.215 -62.630 55.198
463545 90.263 -62.644 55.195
463546 90.312 -62.658 55.194
463547 90.360 -62.672 55.192
463548 90.408 -62.685 55.188
463549 90.456 -62.700 55.180
463550 90.504 -62.713 55.167
463551 90.546 -62.722 55.150
463552 90.589 -62.732 55.132
463553 90.632 -62.742 55.110
463554 90.674 -62.752 55.086
463555 90.717 -62.762 55.064
463556 90.759 -62.772 55.046
463557 90.789 -62.783 55.030
463558 90.819 -62.794 55.017
463559 90.854 -62.806 55.005
463560 90.910 -62.816 54.993
463561 90.966 -62.827 54.983
463562 91.023 -62.838 54.973
463563 91.079 -62.849 54.963
463564 91.135 -62.860 54.954
463565 91.188 -62.871 54.945
463566 91.241 -62.882 54.937
463567 91.294 -62.894 54.929
463568 91.348 -62.905 54.921
463569 91.400 -62.916 54.909
463570 91.454 -62.928 54.896
463571 91.507 -62.939 54.883
463572 91.562 -62.948 54.867
463573 91.621 -62.951 54.849
463574 91.680 -62.953 54.834
463575 91.739 -62.955 54.822
463576 91.786 -62.966 54.812
463577 91.830 -62.976 54.804
463578 91.876 -62.988 54.799
463579 91.921 -62.998 54.796
463580 91.959 -63.001 54.797
463581 91.994 -62.997 54.801
463582 92.027 -62.992 54.808
463583 92.070 -62.996 54.820
463584 92.112 -62.999 54.838
463585 92.155 -63.003 54.856
463586 92.198 -63.007 54.868
463587 92.240 -63.010 54.876
463588 92.283 -63.014 54.879
463589 92.318 -63.017 54.875
463590 92.354 -63.021 54.862
463591 92.390 -63.024 54.844
463592 92.426 -63.028 54.822
463593 92.461 -63.032 54.795
463594 92.497 -63.035 54.767
463595 92.531 -63.039 54.738
463596 92.565 -63.041 54.709
463597 92.599 -63.043 54.682
463598 92.640 -63.047 54.656
463599 92.683 -63.054 54.634
463600 92.716 -63.067 54.615
463601 92.751 -63.081 54.596
463602 92.784 -63.095 54.579
463603 92.826 -63.104 54.564
463604 92.871 -63.111 54.548
463605 92.920 -63.116 54.533
463606 92.980 -63.115 54.517
463607 93.029 -63.122 54.499
463608 93.078 -63.129 54.481
463609 93.113 -63.137 54.460
463610 93.137 -63.148 54.438
463611 93.174 -63.148 54.415
463612 93.221 -63.142 54.392
463613 93.238 -63.142 54.371
463614 93.285 -63.141 54.349
463615 93.332 -63.139 54.330
463616 93.379 -63.137 54.311
463617 93.421 -63.140 54.293
463618 93.446 -63.161 54.275
463619 93.471 -63.182 54.256
463620 93.495 -63.203 54.238
463621 93.520 -63.224 54.220
463622 93.559 -63.235 54.202
463623 93.597 -63.245 54.183
463624 93.636 -63.257 54.163
463625 93.675 -63.267 54.140
463626 93.705 -63.267 54.116
463627 93.735 -63.267 54.091
463628 93.770 -63.274 54.066
463629 93.807 -63.286 54.042
463630 93.847 -63.277 54.018
463631 93.892 -63.272 53.995
463632 93.937 -63.268 53.974
463633 93.994 -63.286 53.953
463634 94.050 -63.302 53.935
463635 94.107 -63.319 53.919
463636 94.144 -63.355 53.903
463637 94.177 -63.379 53.888
463638 94.206 -63.377 53.875
463639 94.237 -63.380 53.863
463640 94.273 -63.386 53.852
463641 94.309 -63.392 53.842
463642 94.345 -63.398 53.831
463643 94.381 -63.403 53.821
463644 94.416 -63.409 53.810
463645 94.452 -63.415 53.798
463646 94.486 -63.424 53.786
463647 94.518 -63.435 53.773
463648 94.550 -63.446 53.759
463649 94.582 -63.458 53.744
463650 94.614 -63.469 53.730
463651 94.657 -63.474 53.715
463652 94.709 -63.475 53.699
463653 94.760 -63.475 53.683
463654 94.820 -63.477 53.665
463655 94.882 -63.478 53.646
463656 94.945 -63.480 53.627
463657 95.007 -63.481 53.605
463658 95.070 -63.483 53.583
463659 95.120 -63.485 53.561
463660 95.158 -63.490 53.538
463661 95.196 -63.494 53.515
463662 95.234 -63.498 53.492
463663 95.272 -63.502 53.469
463664 95.310 -63.506 53.446
463665 95.348 -63.511 53.425
463666 95.387 -63.517 53.403
463667 95.426 -63.524 53.381
463668 95.466 -63.531 53.362
463669 95.506 -63.537 53.340
463670 95.545 -63.543 53.319
463671 95.580 -63.545 53.298
463672 95.611 -63.543 53.278
463673 95.654 -63.544 53.256
463674 95.700 -63.548 53.234
463675 95.746 -63.551 53.209
463676 95.792 -63.554 53.180
463677 95.838 -63.557 53.149
463678 95.884 -63.560 53.118
463679 95.930 -63.563 53.083
463680 95.976 -63.566 53.046
463681 96.021 -63.568 53.004
463682 96.070 -63.562 52.958
463683 96.119 -63.557 52.910
463684 96.169 -63.552 52.865
463685 96.219 -63.547 52.820
463686 96.268 -63.542 52.777
463687 96.317 -63.537 52.736
463688 96.367 -63.532 52.697
463689 96.416 -63.527 52.658
463690 96.466 -63.522 52.619
463691 96.515 -63.517 52.581
463692 96.565 -63.512 52.542
463693 96.615 -63.513 52.504
463694 96.667 -63.523 52.466
463695 96.718 -63.532 52.432
463696 96.754 -63.535 52.400
463697 96.791 -63.540 52.370
463698 96.828 -63.544 52.339
463699 96.865 -63.549 52.308
463700 96.901 -63.555 52.277
463701 96.933 -63.567 52.245
463702 96.964 -63.580 52.213
463703 96.996 -63.592 52.181
463704 97.027 -63.605 52.148
463705 97.058 -63.617 52.115
463706 97.089 -63.629 52.082
463707 97.120 -63.642 52.049
463708 97.152 -63.654 52.017
463709 97.183 -63.667 51.987
463710 97.214 -63.679 51.959
463711 97.245 -63.692 51.931
463712 97.277 -63.704 51.902
463713 97.308 -63.716 51.870
463714 97.332 -63.731 51.839
463715 97.355 -63.745 51.808
463716 97.377 -63.760 51.776
463717 97.400 -63.775 51.744
463718 97.423 -63.789 51.710
463719 97.446 -63.803 51.675
463720 97.470 -63.816 51.636
463721 97.502 -63.820 51.594
463722 97.541 -63.817 51.550
463723 97.580 -63.816 51.504
463724 97.619 -63.813 51.456
463725 97.659 -63.811 51.406
463726 97.698 -63.809 51.355
463727 97.737 -63.807 51.304
463728 97.777 -63.805 51.252
463729 97.816 -63.803 51.196
463730 97.855 -63.800 51.140
463731 97.894 -63.798 51.084
463732 97.934 -63.797 51.028
463733 97.971 -63.796 50.975
463734 98.005 -63.797 50.923
463735 98.040 -63.798 50.871
463736 98.075 -63.800 50.821
463737 98.109 -63.800 50.770
463738 98.144 -63.802 50.718
463739 98.174 -63.806 50.665
463740 98.194 -63.820 50.613
463741 98.227 -63.823 50.562
463742 98.261 -63.827 50.513
463743 98.294 -63.831 50.462
463744 98.327 -63.835 50.411
463745 98.360 -63.838 50.359
463746 98.384 -63.837 50.308
463747 98.408 -63.836 50.257
463748 98.431 -63.836 50.208
463749 98.454 -63.835 50.160
463750 98.479 -63.835 50.113
463751 98.504 -63.838 50.066
463752 98.530 -63.840 50.020
463753 98.556 -63.843 49.976
463754 98.581 -63.845 49.931
463755 98.607 -63.848 49.888
463756 98.649 -63.850 49.844
463757 98.700 -63.852 49.797
463758 98.732 -63.857 49.743
463759 98.751 -63.863 49.689
463760 98.808 -63.868 49.636
463761 98.873 -63.869 49.586
463762 98.936 -63.866 49.537
463763 98.999 -63.865 49.492
463764 99.062 -63.863 49.447
463765 99.125 -63.861 49.404
463766 99.188 -63.860 49.361
463767 99.251 -63.857 49.319
463768 99.314 -63.856 49.278
463769 99.377 -63.854 49.239
463770 99.440 -63.852 49.201
464172 86.940 -62.061 56.185
464173 86.982 -62.060 56.173
464174 87.024 -62.058 56.161
464175 87.067 -62.056 56.150
464176 87.110 -62.055 56.141
464177 87.151 -62.053 56.132
464178 87.188 -62.055 56.124
464179 87.225 -62.057 56.115
464180 87.262 -62.059 56.105
464181 87.298 -62.061 56.097
464182 87.335 -62.062 56.090
464183 87.371 -62.064 56.083
464184 87.408 -62.066 56.077
464185 87.451 -62.065 56.069
464186 87.500 -62.061 56.056
464187 87.548 -62.058 56.038
464188 87.596 -62.055 56.018
464189 87.644 -62.051 55.993
464190 87.693 -62.048 55.968
464191 87.741 -62.045 55.943
464192 87.789 -62.042 55.919
464193 87.837 -62.038 55.897
464194 87.886 -62.035 55.877
464195 87.934 -62.031 55.858
464196 87.982 -62.028 55.841
464197 88.031 -62.025 55.826
464198 88.079 -62.021 55.814
464199 88.127 -62.018 55.803
464200 88.175 -62.014 55.792
464201 88.229 -62.024 55.781
464202 88.277 -62.040 55.766
464203 88.324 -62.056 55.752
464204 88.370 -62.072 55.736
464205 88.418 -62.088 55.716
464206 88.465 -62.105 55.690
464207 88.514 -62.120 55.658
464208 88.562 -62.137 55.626
464209 88.611 -62.153 55.597
464210 88.660 -62.170 55.571
464211 88.709 -62.186 55.549
464212 88.757 -62.202 55.531
464213 88.804 -62.218 55.515
464214 88.847 -62.233 55.500
464215 88.890 -62.249 55.483
464216 88.933 -62.263 55.467
464217 88.977 -62.271 55.450
464218 89.021 -62.280 55.434
464219 89.065 -62.288 55.419
464220 89.109 -62.297 55.404
464221 89.152 -62.306 55.390
464222 89.194 -62.315 55.375
464223 89.237 -62.325 55.360
464224 89.279 -62.335 55.346
464225 89.322 -62.344 55.334
464226 89.365 -62.354 55.323
464227 89.407 -62.363 55.312
464228 89.450 -62.373 55.304
464229 89.492 -62.383 55.296
464230 89.535 -62.392 55.287
464231 89.578 -62.402 55.277
464232 89.620 -62.412 55.269
464233 89.663 -62.421 55.261
464234 89.705 -62.431 55.254
464235 89.748 -62.440 55.250
464236 89.791 -62.449 55.244
464237 89.835 -62.466 55.236
464238 89.880 -62.487 55.230
464239 89.930 -62.502 55.225
464240 89.982 -62.515 55.218
464241 90.034 -62.528 55.211
464242 90.083 -62.541 55.203
464243 90.131 -62.555 55.197
464244 90.179 -62.569 55.192
464245 90.228 -62.583 55.189
464246 90.276 -62.597 55.189
464247 90.324 -62.611 55.192
464248 90.372 -62.625 55.193
464249 90.421 -62.639 55.191
464250 90.469 -62.653 55.184
464251 90.517 -62.666 55.171
464252 90.561 -62.677 55.154
464253 90.603 -62.688 55.134
464254 90.646 -62.697 55.111
464255 90.689 -62.707 55.084
464256 90.731 -62.717 55.061
464257 90.774 -62.727 55.043
464258 90.817 -62.737 55.027
464259 90.850 -62.747 55.014
464260 90.880 -62.759 55.002
464261 90.929 -62.769 54.991
464262 90.985 -62.781 54.982
464263 91.041 -62.791 54.973
464264 91.097 -62.802 54.964
464265 91.150 -62.813 54.955
464266 91.204 -62.825 54.948
464267 91.256 -62.836 54.941
464268 91.310 -62.847 54.935
464269 91.363 -62.859 54.927
464270 91.416 -62.870 54.916
464271 91.469 -62.882 54.901
464272 91.522 -62.893 54.887
464273 91.577 -62.902 54.870
464274 91.636 -62.904 54.852
464275 91.696 -62.907 54.837
464276 91.742 -62.916 54.824
464277 91.788 -62.928 54.814
464278 91.832 -62.938 54.806
464279 91.877 -62.950 54.800
464280 91.923 -62.961 54.798
464281 91.964 -62.967 54.800
464282 91.998 -62.962 54.805
464283 92.031 -62.958 54.813
464284 92.067 -62.954 54.825
464285 92.109 -62.958 54.843
464286 92.152 -62.961 54.862
464287 92.194 -62.965 54.877
464288 92.237 -62.969 54.885
464289 92.280 -62.972 54.887
464290 92.322 -62.976 54.882
464291 92.365 -62.979 54.870
464292 92.403 -62.983 54.851
464293 92.439 -62.986 54.826
464294 92.475 -62.990 54.798
464295 92.510 -62.994 54.769
464296 92.546 -62.998 54.739
464297 92.581 -63.001 54.710
464298 92.617 -63.004 54.681
464299 92.653 -63.008 54.654
464300 92.689 -63.016 54.631
464301 92.722 -63.029 54.607
464302 92.756 -63.042 54.584
464303 92.790 -63.056 54.563
464304 92.826 -63.069 54.543
464305 92.873 -63.074 54.525
464306 92.926 -63.079 54.509
464307 92.976 -63.086 54.491
464308 93.024 -63.092 54.473
464309 93.073 -63.098 54.454
464310 93.108 -63.108 54.433
464311 93.134 -63.117 54.411
464312 93.184 -63.107 54.390
464313 93.232 -63.100 54.369
464314 93.249 -63.104 54.349
464315 93.281 -63.101 54.328
464316 93.328 -63.098 54.309
464317 93.375 -63.097 54.290
464318 93.413 -63.104 54.272
464319 93.437 -63.126 54.254
464320 93.462 -63.146 54.235
464321 93.486 -63.167 54.217
464322 93.523 -63.180 54.198
464323 93.562 -63.190 54.179
464324 93.601 -63.202 54.159
464325 93.640 -63.212 54.139
464326 93.678 -63.224 54.116
464327 93.717 -63.234 54.094
464328 93.750 -63.236 54.071
464329 93.781 -63.239 54.048
464330 93.818 -63.246 54.025
464331 93.861 -63.240 54.004
464332 93.906 -63.236 53.983
464333 93.951 -63.231 53.964
464334 93.997 -63.228 53.945
464335 94.054 -63.245 53.928
464336 94.102 -63.270 53.912
464337 94.136 -63.308 53.897
464338 94.170 -63.336 53.883
464339 94.200 -63.336 53.871
464340 94.236 -63.342 53.860
464341 94.272 -63.348 53.850
464342 94.308 -63.354 53.840
464343 94.343 -63.360 53.830
464344 94.379 -63.366 53.820
464345 94.415 -63.372 53.809
464346 94.451 -63.378 53.798
464347 94.486 -63.384 53.785
464348 94.521 -63.392 53.772
464349 94.553 -63.403 53.757
464350 94.585 -63.415 53.741
464351 94.617 -63.426 53.724
464352 94.661 -63.430 53.708
464353 94.712 -63.431 53.690
464354 94.768 -63.432 53.673
464355 94.831 -63.433 53.654
464356 94.894 -63.435 53.634
464357 94.957 -63.437 53.612
464358 95.019 -63.438 53.589
464359 95.082 -63.440 53.566
464360 95.129 -63.443 53.542
464361 95.167 -63.446 53.518
464362 95.205 -63.451 53.494
464363 95.244 -63.455 53.470
464364 95.281 -63.459 53.446
464365 95.320 -63.464 53.424
464366 95.357 -63.468 53.401
464367 95.396 -63.472 53.379
464368 95.435 -63.479 53.357
464369 95.475 -63.485 53.337
464370 95.514 -63.492 53.315
464371 95.550 -63.495 53.293
464372 95.581 -63.492 53.271
464373 95.612 -63.490 53.248
464374 95.658 -63.493 53.225
464375 95.704 -63.496 53.201
464376 95.750 -63.499 53.176
464377 95.796 -63.502 53.148
464378 95.842 -63.506 53.119
464379 95.888 -63.509 53.090
464380 95.934 -63.512 53.058
464381 95.980 -63.517 53.023
464382 96.027 -63.523 52.985
464383 96.063 -63.528 52.943
464384 96.134 -63.525 52.895
464385 96.177 -63.524 52.847
464386 96.224 -63.519 52.800
464387 96.279 -63.511 52.753
464388 96.328 -63.506 52.708
464389 96.378 -63.501 52.665
464390 96.428 -63.496 52.625
464391 96.477 -63.491 52.585
464392 96.526 -63.486 52.545
464393 96.576 -63.480 52.506
464394 96.625 -63.476 52.469
464395 96.675 -63.476 52.431
464396 96.727 -63.485 52.396
464397 96.765 -63.490 52.365
464398 96.803 -63.494 52.337
464399 96.839 -63.499 52.308
464400 96.872 -63.510 52.277
464401 96.903 -63.522 52.245
464402 96.935 -63.535 52.213
464403 96.966 -63.547 52.181
464404 96.997 -63.560 52.149
464405 97.028 -63.572 52.115
464406 97.060 -63.585 52.081
464407 97.091 -63.597 52.047
464408 97.122 -63.609 52.014
464409 97.153 -63.622 51.983
464410 97.184 -63.634 51.955
464411 97.216 -63.647 51.927
464412 97.247 -63.659 51.900
464413 97.278 -63.672 51.874
464414 97.307 -63.685 51.845
464415 97.330 -63.700 51.816
464416 97.353 -63.714 51.785
464417 97.376 -63.728 51.752
464418 97.399 -63.743 51.720
464419 97.422 -63.758 51.688
464420 97.445 -63.772 51.653
464421 97.469 -63.784 51.615
464422 97.508 -63.784 51.574
464423 97.547 -63.782 51.530
464424 97.587 -63.780 51.485
464425 97.626 -63.778 51.437
464426 97.665 -63.776 51.388
464427 97.705 -63.774 51.337
464428 97.744 -63.772 51.287
464429 97.783 -63.769 51.235
464430 97.822 -63.767 51.180
464431 97.862 -63.766 51.124
464432 97.901 -63.763 51.067
464433 97.940 -63.761 51.012
464434 97.978 -63.760 50.961
464435 98.013 -63.761 50.908
464436 98.047 -63.763 50.856
464437 98.082 -63.763 50.804
464438 98.117 -63.765 50.754
464439 98.154 -63.764 50.702
464440 98.185 -63.768 50.650
464441 98.204 -63.781 50.599
464442 98.223 -63.796 50.549
464443 98.253 -63.801 50.501
464444 98.287 -63.805 50.452
464445 98.320 -63.809 50.401
464446 98.354 -63.813 50.350
464447 98.382 -63.814 50.299
464448 98.406 -63.813 50.248
464449 98.429 -63.813 50.198
464450 98.453 -63.812 50.148
464451 98.477 -63.811 50.100
464452 98.500 -63.810 50.053
464453 98.523 -63.810 50.006
464454 98.549 -63.812 49.960
464455 98.574 -63.815 49.915
464456 98.600 -63.817 49.870
464457 98.631 -63.819 49.825
464458 98.681 -63.822 49.778
464459 98.706 -63.827 49.724
464460 98.755 -63.832 49.671
464461 98.819 -63.831 49.621
464462 98.882 -63.830 49.572
464463 98.945 -63.828 49.524
464464 99.008 -63.826 49.480
464465 99.071 -63.824 49.436
464466 99.134 -63.822 49.392
464467 99.197 -63.821 49.350
464468 99.260 -63.819 49.309
464469 99.323 -63.817 49.268
464470 99.385 -63.815 49.229
464471 99.448 -63.813 49.191
464873 86.956 -62.013 56.186
464874 86.999 -62.011 56.173
464875 87.041 -62.009 56.161
464876 87.083 -62.008 56.150
464877 87.126 -62.006 56.140
464878 87.167 -62.005 56.131
464879 87.203 -62.008 56.122
464880 87.240 -62.009 56.112
464881 87.277 -62.011 56.103
464882 87.314 -62.013 56.095
464883 87.350 -62.014 56.087
464884 87.387 -62.016 56.081
464885 87.424 -62.018 56.075
464886 87.461 -62.020 56.067
464887 87.509 -62.016 56.056
464888 87.557 -62.013 56.040
464889 87.606 -62.010 56.022
464890 87.654 -62.006 55.999
464891 87.702 -62.003 55.974
464892 87.751 -61.999 55.949
464893 87.798 -61.996 55.926
464894 87.847 -61.992 55.904
464895 87.895 -61.989 55.884
464896 87.943 -61.986 55.864
464897 87.992 -61.983 55.847
464898 88.040 -61.979 55.831
464899 88.088 -61.976 55.817
464900 88.136 -61.973 55.805
464901 88.185 -61.969 55.793
464902 88.246 -61.977 55.782
464903 88.294 -61.992 55.766
464904 88.342 -62.008 55.750
464905 88.390 -62.023 55.731
464906 88.436 -62.039 55.709
464907 88.484 -62.056 55.679
464908 88.531 -62.072 55.647
464909 88.578 -62.088 55.617
464910 88.625 -62.104 55.591
464911 88.672 -62.120 55.568
464912 88.719 -62.136 55.548
464913 88.766 -62.152 55.532
464914 88.813 -62.168 55.517
464915 88.860 -62.184 55.501
464916 88.906 -62.199 55.485
464917 88.949 -62.208 55.468
464918 88.992 -62.218 55.451
464919 89.034 -62.227 55.435
464920 89.077 -62.236 55.419
464921 89.119 -62.246 55.404
464922 89.162 -62.255 55.389
464923 89.204 -62.265 55.374
464924 89.247 -62.274 55.358
464925 89.290 -62.284 55.344
464926 89.332 -62.294 55.332
464927 89.375 -62.303 55.320
464928 89.417 -62.313 55.309
464929 89.460 -62.322 55.300
464930 89.502 -62.332 55.292
464931 89.545 -62.342 55.283
464932 89.588 -62.351 55.273
464933 89.630 -62.361 55.266
464934 89.673 -62.371 55.260
464935 89.716 -62.380 55.256
464936 89.758 -62.390 55.254
464937 89.801 -62.399 55.250
464938 89.845 -62.418 55.241
464939 89.890 -62.439 55.234
464940 89.938 -62.457 55.229
464941 89.990 -62.470 55.222
464942 90.042 -62.482 55.214
464943 90.094 -62.495 55.204
464944 90.144 -62.508 55.195
464945 90.192 -62.522 55.187
464946 90.240 -62.536 55.182
464947 90.289 -62.550 55.183
464948 90.336 -62.564 55.187
464949 90.385 -62.578 55.191
464950 90.433 -62.592 55.190
464951 90.481 -62.606 55.184
464952 90.530 -62.620 55.171
464953 90.576 -62.632 55.154
464954 90.618 -62.642 55.132
464955 90.661 -62.652 55.106
464956 90.704 -62.662 55.078
464957 90.746 -62.672 55.057
464958 90.789 -62.681 55.040
464959 90.831 -62.691 55.025
464960 90.874 -62.701 55.013
464961 90.910 -62.712 55.001
464962 90.947 -62.723 54.991
464963 91.003 -62.734 54.981
464964 91.059 -62.744 54.972
464965 91.112 -62.756 54.964
464966 91.166 -62.767 54.957
464967 91.219 -62.778 54.951
464968 91.272 -62.790 54.945
464969 91.325 -62.801 54.939
464970 91.378 -62.813 54.931
464971 91.431 -62.824 54.920
464972 91.485 -62.835 54.905
464973 91.537 -62.847 54.890
464974 91.592 -62.857 54.874
464975 91.651 -62.859 54.856
464976 91.700 -62.868 54.841
464977 91.744 -62.879 54.827
464978 91.789 -62.890 54.816
464979 91.835 -62.901 54.808
464980 91.879 -62.912 54.802
464981 91.925 -62.923 54.800
464982 91.968 -62.932 54.802
464983 92.002 -62.927 54.807
464984 92.035 -62.923 54.817
464985 92.070 -62.919 54.829
464986 92.106 -62.916 54.846
464987 92.148 -62.920 54.864
464988 92.191 -62.924 54.878
464989 92.233 -62.927 54.887
464990 92.276 -62.931 54.888
464991 92.319 -62.935 54.882
464992 92.361 -62.938 54.869
464993 92.404 -62.941 54.849
464994 92.446 -62.945 54.824
464995 92.488 -62.948 54.795
464996 92.523 -62.952 54.767
464997 92.559 -62.956 54.736
464998 92.594 -62.960 54.707
464999 92.630 -62.963 54.679
465000 92.666 -62.967 54.650
465001 92.701 -62.974 54.626
465002 92.733 -62.988 54.600
465003 92.767 -63.002 54.574
465004 92.799 -63.017 54.548
465005 92.831 -63.029 54.525
465006 92.873 -63.042 54.503
465007 92.922 -63.048 54.485
465008 92.971 -63.055 54.466
465009 93.019 -63.062 54.446
465010 93.068 -63.069 54.426
465011 93.103 -63.078 54.406
465012 93.144 -63.075 54.386
465013 93.194 -63.066 54.366
465014 93.243 -63.057 54.346
465015 93.278 -63.067 54.327
465016 93.277 -63.061 54.308
465017 93.324 -63.058 54.290
465018 93.371 -63.057 54.271
465019 93.404 -63.069 54.253
465020 93.429 -63.090 54.235
465021 93.453 -63.111 54.217
465022 93.489 -63.124 54.198
465023 93.528 -63.135 54.180
465024 93.566 -63.146 54.160
465025 93.605 -63.157 54.140
465026 93.643 -63.168 54.120
465027 93.682 -63.179 54.099
465028 93.720 -63.189 54.078
465029 93.759 -63.201 54.057
465030 93.793 -63.206 54.035
465031 93.831 -63.208 54.014
465032 93.876 -63.204 53.994
465033 93.921 -63.198 53.975
465034 93.966 -63.194 53.957
465035 94.010 -63.189 53.938
465036 94.057 -63.188 53.922
465037 94.095 -63.224 53.907
465038 94.129 -63.261 53.893
465039 94.163 -63.292 53.879
465040 94.198 -63.299 53.867
465041 94.235 -63.305 53.857
465042 94.270 -63.311 53.848
465043 94.306 -63.317 53.838
465044 94.342 -63.323 53.828
465045 94.378 -63.329 53.819
465046 94.413 -63.335 53.808
465047 94.449 -63.341 53.797
465048 94.485 -63.347 53.784
465049 94.521 -63.353 53.770
465050 94.556 -63.359 53.754
465051 94.588 -63.371 53.737
465052 94.620 -63.382 53.718
465053 94.664 -63.386 53.700
465054 94.717 -63.387 53.681
465055 94.780 -63.389 53.662
465056 94.843 -63.390 53.642
465057 94.905 -63.392 53.621
465058 94.968 -63.393 53.598
465059 95.031 -63.395 53.574
465060 95.094 -63.396 53.549
465061 95.139 -63.399 53.523
465062 95.177 -63.404 53.498
465063 95.215 -63.408 53.473
465064 95.253 -63.412 53.448
465065 95.291 -63.416 53.423
465066 95.329 -63.421 53.400
465067 95.367 -63.424 53.377
465068 95.405 -63.429 53.354
465069 95.444 -63.434 53.333
465070 95.483 -63.440 53.311
465071 95.520 -63.444 53.289
465072 95.551 -63.442 53.266
465073 95.582 -63.440 53.243
465074 95.616 -63.438 53.219
465075 95.662 -63.441 53.194
465076 95.708 -63.444 53.169
465077 95.754 -63.447 53.142
465078 95.800 -63.450 53.114
465079 95.846 -63.453 53.086
465080 95.892 -63.458 53.057
465081 95.939 -63.465 53.028
465082 95.985 -63.471 52.996
465083 96.031 -63.477 52.961
465084 96.067 -63.479 52.920
465085 96.105 -63.487 52.875
465086 96.160 -63.491 52.827
465087 96.241 -63.484 52.779
465088 96.294 -63.480 52.730
465089 96.335 -63.479 52.683
465090 96.377 -63.477 52.637
465091 96.424 -63.474 52.594
465092 96.478 -63.465 52.552
465093 96.533 -63.458 52.511
465094 96.587 -63.449 52.471
465095 96.637 -63.445 52.433
465096 96.686 -63.440 52.396
465097 96.736 -63.438 52.363
465098 96.777 -63.444 52.333
465099 96.811 -63.453 52.306
465100 96.842 -63.465 52.278
465101 96.874 -63.478 52.246
465102 96.905 -63.490 52.213
465103 96.936 -63.502 52.181
465104 96.967 -63.515 52.149
465105 96.999 -63.528 52.116
465106 97.030 -63.540 52.082
465107 97.061 -63.553 52.048
465108 97.092 -63.565 52.014
465109 97.123 -63.578 51.981
465110 97.155 -63.590 51.949
465111 97.186 -63.603 51.920
465112 97.217 -63.615 51.892
465113 97.248 -63.628 51.865
465114 97.280 -63.640 51.840
465115 97.306 -63.653 51.814
465116 97.329 -63.668 51.785
465117 97.352 -63.683 51.753
465118 97.375 -63.697 51.721
465119 97.397 -63.712 51.688
465120 97.421 -63.723 51.659
465121 97.443 -63.734 51.625
465122 97.479 -63.731 51.589
465123 97.508 -63.737 51.550
465124 97.554 -63.747 51.507
465125 97.593 -63.745 51.462
465126 97.633 -63.743 51.415
465127 97.672 -63.741 51.365
465128 97.711 -63.738 51.315
465129 97.750 -63.736 51.265
465130 97.790 -63.734 51.214
465131 97.829 -63.732 51.160
465132 97.868 -63.730 51.106
465133 97.908 -63.728 51.052
465134 97.947 -63.725 50.998
465135 97.985 -63.724 50.946
465136 98.020 -63.725 50.894
465137 98.055 -63.727 50.841
465138 98.089 -63.728 50.788
465139 98.124 -63.729 50.736
465140 98.164 -63.726 50.684
465141 98.197 -63.729 50.633
465142 98.215 -63.744 50.584
465143 98.234 -63.757 50.534
465144 98.252 -63.771 50.484
465145 98.279 -63.779 50.436
465146 98.312 -63.783 50.388
465147 98.346 -63.787 50.339
465148 98.379 -63.790 50.288
465149 98.404 -63.790 50.236
465150 98.427 -63.789 50.185
465151 98.451 -63.788 50.135
465152 98.475 -63.788 50.086
465153 98.498 -63.788 50.038
465154 98.522 -63.787 49.992
465155 98.545 -63.786 49.946
465156 98.569 -63.785 49.900
465157 98.593 -63.786 49.853
465158 98.619 -63.789 49.805
465159 98.661 -63.791 49.758
465160 98.702 -63.797 49.705
465161 98.765 -63.794 49.654
465162 98.828 -63.793 49.605
465163 98.890 -63.791 49.558
465164 98.953 -63.789 49.512
465165 99.016 -63.788 49.468
465166 99.079 -63.785 49.424
465167 99.142 -63.784 49.381
465168 99.205 -63.781 49.339
465169 99.268 -63.780 49.299
465170 99.331 -63.778 49.258
465171 99.394 -63.776 49.219
465575 87.015 -61.962 56.173
465576 87.058 -61.961 56.161
465577 87.100 -61.959 56.149
465578 87.142 -61.957 56.139
465579 87.182 -61.957 56.129
465580 87.219 -61.959 56.119
465581 87.256 -61.961 56.110
465582 87.292 -61.963 56.101
465583 87.329 -61.965 56.092
465584 87.365 -61.967 56.085
465585 87.402 -61.968 56.077
465586 87.439 -61.970 56.070
465587 87.476 -61.972 56.061
465588 87.518 -61.971 56.051
465589 87.567 -61.967 56.038
465590 87.615 -61.964 56.022
465591 87.663 -61.961 56.002
465592 87.712 -61.957 55.978
465593 87.760 -61.954 55.955
465594 87.808 -61.951 55.932
465595 87.856 -61.948 55.911
465596 87.905 -61.944 55.891
465597 87.953 -61.941 55.871
465598 88.001 -61.937 55.853
465599 88.049 -61.934 55.837
465600 88.098 -61.931 55.821
465601 88.146 -61.927 55.806
465602 88.194 -61.924 55.793
465603 88.261 -61.930 55.782
465604 88.311 -61.945 55.765
465605 88.359 -61.961 55.748
465606 88.407 -61.976 55.728
465607 88.456 -61.991 55.705
465608 88.503 -62.007 55.674
465609 88.550 -62.023 55.643
465610 88.597 -62.039 55.615
465611 88.644 -62.055 55.591
465612 88.691 -62.071 55.571
465613 88.738 -62.087 55.553
465614 88.785 -62.104 55.538
465615 88.832 -62.120 55.521
465616 88.876 -62.137 55.505
465617 88.919 -62.151 55.488
465618 88.959 -62.157 55.468
465619 89.001 -62.167 55.453
465620 89.044 -62.177 55.436
465621 89.087 -62.186 55.420
465622 89.129 -62.196 55.404
465623 89.172 -62.205 55.388
465624 89.214 -62.214 55.373
465625 89.257 -62.224 55.357
465626 89.300 -62.233 55.341
465627 89.342 -62.243 55.329
465628 89.385 -62.253 55.317
465629 89.427 -62.262 55.305
465630 89.470 -62.272 55.296
465631 89.513 -62.282 55.288
465632 89.555 -62.291 55.278
465633 89.598 -62.301 55.270
465634 89.640 -62.310 55.264
465635 89.683 -62.320 55.258
465636 89.725 -62.330 55.255
465637 89.768 -62.339 55.256
465638 89.810 -62.349 55.254
465639 89.856 -62.369 55.248
465640 89.901 -62.390 55.243
465641 89.946 -62.412 55.237
465642 89.998 -62.424 55.232
465643 90.050 -62.437 55.222
465644 90.102 -62.449 55.210
465645 90.154 -62.462 55.198
465646 90.204 -62.475 55.187
465647 90.253 -62.489 55.177
465648 90.301 -62.503 55.177
465649 90.349 -62.517 55.181
465650 90.397 -62.531 55.185
465651 90.446 -62.545 55.185
465652 90.494 -62.559 55.180
465653 90.542 -62.573 55.169
465654 90.590 -62.587 55.150
465655 90.633 -62.597 55.126
465656 90.676 -62.606 55.099
465657 90.719 -62.616 55.072
465658 90.761 -62.626 55.053
465659 90.804 -62.636 55.037
465660 90.847 -62.646 55.023
465661 90.889 -62.656 55.012
465662 90.932 -62.666 55.001
465663 90.970 -62.676 54.991
465664 91.021 -62.687 54.983
465665 91.075 -62.698 54.974
465666 91.128 -62.710 54.967
465667 91.181 -62.721 54.960
465668 91.234 -62.732 54.954
465669 91.287 -62.744 54.949
465670 91.341 -62.755 54.944
465671 91.393 -62.766 54.938
465672 91.447 -62.778 54.926
465673 91.500 -62.790 54.911
465674 91.553 -62.801 54.895
465675 91.607 -62.811 54.879
465676 91.656 -62.819 54.862
465677 91.701 -62.830 54.845
465678 91.746 -62.841 54.831
465679 91.791 -62.852 54.819
465680 91.837 -62.863 54.810
465681 91.881 -62.875 54.804
465682 91.927 -62.885 54.801
465683 91.971 -62.897 54.803
465684 92.006 -62.893 54.809
465685 92.040 -62.888 54.819
465686 92.074 -62.884 54.832
465687 92.107 -62.880 54.847
465688 92.145 -62.879 54.863
465689 92.188 -62.882 54.877
465690 92.230 -62.886 54.885
465691 92.273 -62.889 54.886
465692 92.315 -62.893 54.878
465693 92.358 -62.897 54.863
465694 92.401 -62.900 54.842
465695 92.443 -62.904 54.816
465696 92.486 -62.907 54.788
465697 92.528 -62.910 54.759
465698 92.571 -62.914 54.728
465699 92.607 -62.918 54.698
465700 92.643 -62.921 54.670
465701 92.679 -62.925 54.641
465702 92.714 -62.933 54.614
465703 92.746 -62.947 54.587
465704 92.780 -62.960 54.560
465705 92.807 -62.973 54.534
465706 92.836 -62.984 54.509
465707 92.881 -62.996 54.486
465708 92.925 -63.008 54.464
465709 92.970 -63.020 54.443
465710 93.014 -63.032 54.422
465711 93.064 -63.039 54.402
465712 93.105 -63.043 54.383
465713 93.155 -63.033 54.364
465714 93.205 -63.024 54.345
465715 93.254 -63.016 54.327
465716 93.292 -63.026 54.308
465717 93.279 -63.020 54.290
465718 93.320 -63.018 54.271
465719 93.366 -63.017 54.254
465720 93.396 -63.033 54.236
465721 93.420 -63.054 54.218
465722 93.453 -63.069 54.200
465723 93.492 -63.079 54.181
465724 93.531 -63.091 54.163
465725 93.569 -63.101 54.144
465726 93.608 -63.113 54.124
465727 93.647 -63.123 54.105
465728 93.686 -63.135 54.086
465729 93.724 -63.145 54.065
465730 93.763 -63.156 54.045
465731 93.801 -63.168 54.026
465732 93.846 -63.179 54.006
465733 93.892 -63.177 53.987
465734 93.937 -63.177 53.969
465735 93.982 -63.175 53.952
465736 94.028 -63.173 53.934
465737 94.073 -63.172 53.918
465738 94.111 -63.199 53.904
465739 94.146 -63.238 53.890
465740 94.180 -63.257 53.877
465741 94.212 -63.263 53.865
465742 94.245 -63.268 53.855
465743 94.277 -63.274 53.845
465744 94.310 -63.280 53.836
465745 94.342 -63.286 53.826
465746 94.376 -63.291 53.817
465747 94.412 -63.297 53.806
465748 94.448 -63.303 53.794
465749 94.483 -63.309 53.780
465750 94.519 -63.315 53.765
465751 94.555 -63.321 53.748
465752 94.591 -63.327 53.730
465753 94.623 -63.339 53.711
465754 94.668 -63.343 53.690
465755 94.729 -63.344 53.670
465756 94.791 -63.346 53.649
465757 94.854 -63.347 53.628
465758 94.917 -63.349 53.606
465759 94.980 -63.350 53.583
465760 95.042 -63.352 53.558
465761 95.105 -63.353 53.532
465762 95.148 -63.357 53.506
465763 95.186 -63.361 53.479
465764 95.225 -63.365 53.453
465765 95.262 -63.369 53.427
465766 95.301 -63.374 53.400
465767 95.339 -63.377 53.376
465768 95.377 -63.382 53.352
465769 95.415 -63.386 53.329
465770 95.453 -63.390 53.307
465771 95.490 -63.394 53.284
465772 95.521 -63.392 53.261
465773 95.552 -63.389 53.238
465774 95.584 -63.386 53.214
465775 95.620 -63.386 53.189
465776 95.666 -63.390 53.162
465777 95.712 -63.393 53.136
465778 95.758 -63.396 53.108
465779 95.804 -63.399 53.079
465780 95.850 -63.406 53.051
465781 95.897 -63.412 53.023
465782 95.944 -63.419 52.995
465783 95.990 -63.426 52.965
465784 96.035 -63.432 52.932
465785 96.068 -63.436 52.894
465786 96.110 -63.433 52.852
465787 96.146 -63.446 52.806
465788 96.184 -63.458 52.758
465789 96.265 -63.450 52.708
465790 96.346 -63.443 52.659
465791 96.410 -63.438 52.611
465792 96.452 -63.437 52.566
465793 96.494 -63.435 52.521
465794 96.536 -63.433 52.478
465795 96.578 -63.431 52.438
465796 96.623 -63.428 52.399
465797 96.678 -63.420 52.362
465798 96.733 -63.412 52.330
465799 96.784 -63.410 52.301
465800 96.814 -63.422 52.275
465801 96.845 -63.433 52.247
465802 96.875 -63.446 52.215
465803 96.906 -63.459 52.182
465804 96.938 -63.471 52.149
465805 96.969 -63.483 52.117
465806 97.000 -63.496 52.084
465807 97.031 -63.508 52.050
465808 97.063 -63.521 52.016
465809 97.094 -63.533 51.982
465810 97.125 -63.546 51.948
465811 97.156 -63.558 51.916
465812 97.187 -63.571 51.886
465813 97.219 -63.583 51.857
465814 97.250 -63.595 51.830
465815 97.282 -63.608 51.803
465816 97.305 -63.622 51.777
465817 97.327 -63.637 51.748
465818 97.351 -63.650 51.717
465819 97.373 -63.661 51.685
465820 97.396 -63.672 51.653
465821 97.419 -63.683 51.624
465822 97.442 -63.694 51.593
465823 97.489 -63.680 51.559
465824 97.517 -63.684 51.522
465825 97.551 -63.693 51.481
465826 97.599 -63.708 51.437
465827 97.639 -63.707 51.390
465828 97.678 -63.705 51.341
465829 97.718 -63.703 51.291
465830 97.757 -63.701 51.241
465831 97.796 -63.699 51.191
465832 97.836 -63.697 51.140
465833 97.875 -63.694 51.087
465834 97.914 -63.693 51.035
465835 97.953 -63.690 50.982
465836 97.993 -63.688 50.930
465837 98.027 -63.690 50.877
465838 98.062 -63.690 50.824
465839 98.096 -63.692 50.771
465840 98.132 -63.693 50.718
465841 98.174 -63.688 50.666
465842 98.208 -63.691 50.615
465843 98.226 -63.705 50.564
465844 98.244 -63.719 50.514
465845 98.263 -63.733 50.464
465846 98.282 -63.747 50.417
465847 98.305 -63.757 50.368
465848 98.339 -63.761 50.320
465849 98.372 -63.765 50.271
465850 98.402 -63.766 50.222
465851 98.425 -63.766 50.173
465852 98.449 -63.766 50.124
465853 98.473 -63.765 50.075
465854 98.496 -63.764 50.027
465855 98.520 -63.763 49.980
465856 98.544 -63.763 49.933
465857 98.567 -63.762 49.885
465858 98.590 -63.761 49.836
465859 98.614 -63.761 49.786
465860 98.648 -63.759 49.739
465861 98.711 -63.757 49.687
465862 98.774 -63.756 49.638
465863 98.837 -63.754 49.591
465864 98.900 -63.752 49.544
465865 98.962 -63.750 49.499
465866 99.025 -63.748 49.456
465867 99.088 -63.747 49.413
465868 99.151 -63.745 49.370
465869 99.214 -63.743 49.328
465870 99.277 -63.741 49.289
465871 99.340 -63.739 49.249
465872 99.403 -63.737 49.211
466276 87.032 -61.914 56.173
466277 87.074 -61.912 56.160
466278 87.117 -61.910 56.148
466279 87.159 -61.909 56.138
466280 87.197 -61.910 56.128
466281 87.234 -61.911 56.118
466282 87.271 -61.914 56.108
466283 87.307 -61.915 56.099
466284 87.344 -61.917 56.089
466285 87.380 -61.919 56.082
466286 87.417 -61.920 56.074
466287 87.454 -61.923 56.065
466288 87.491 -61.924 56.056
466289 87.528 -61.926 56.045
466290 87.576 -61.922 56.033
466291 87.625 -61.919 56.019
466292 87.673 -61.916 56.001
466293 87.721 -61.912 55.980
466294 87.769 -61.909 55.957
466295 87.818 -61.905 55.935
466296 87.866 -61.902 55.915
466297 87.914 -61.898 55.897
466298 87.963 -61.895 55.879
466299 88.010 -61.892 55.861
466300 88.059 -61.888 55.844
466301 88.107 -61.885 55.826
466302 88.155 -61.882 55.809
466303 88.205 -61.879 55.794
466304 88.272 -61.885 55.779
466305 88.328 -61.898 55.764
466306 88.376 -61.914 55.746
466307 88.424 -61.929 55.726
466308 88.472 -61.944 55.701
466309 88.521 -61.959 55.670
466310 88.569 -61.975 55.640
466311 88.616 -61.991 55.614
466312 88.663 -62.007 55.593
466313 88.710 -62.023 55.576
466314 88.757 -62.039 55.560
466315 88.802 -62.056 55.545
466316 88.846 -62.073 55.528
466317 88.892 -62.096 55.509
466318 88.933 -62.108 55.490
466319 88.971 -62.110 55.470
466320 89.012 -62.116 55.452
466321 89.054 -62.126 55.434
466322 89.097 -62.136 55.417
466323 89.139 -62.145 55.401
466324 89.182 -62.155 55.386
466325 89.225 -62.164 55.371
466326 89.267 -62.174 55.356
466327 89.310 -62.183 55.341
466328 89.352 -62.192 55.325
466329 89.395 -62.202 55.314
466330 89.437 -62.212 55.302
466331 89.480 -62.221 55.292
466332 89.523 -62.231 55.285
466333 89.565 -62.241 55.274
466334 89.608 -62.250 55.267
466335 89.650 -62.260 55.261
466336 89.693 -62.269 55.256
466337 89.735 -62.279 55.255
466338 89.778 -62.289 55.256
466339 89.821 -62.299 55.256
466340 89.866 -62.321 55.254
466341 89.911 -62.342 55.250
466342 89.957 -62.363 55.245
466343 90.006 -62.380 55.240
466344 90.058 -62.392 55.231
466345 90.110 -62.405 55.219
466346 90.162 -62.417 55.204
466347 90.214 -62.430 55.189
466348 90.265 -62.442 55.176
466349 90.314 -62.456 55.171
466350 90.362 -62.470 55.171
466351 90.410 -62.484 55.173
466352 90.458 -62.498 55.174
466353 90.507 -62.512 55.172
466354 90.554 -62.526 55.159
466355 90.603 -62.540 55.140
466356 90.648 -62.552 55.116
466357 90.690 -62.562 55.089
466358 90.733 -62.572 55.066
466359 90.776 -62.581 55.049
466360 90.818 -62.591 55.034
466361 90.861 -62.601 55.021
466362 90.904 -62.611 55.011
466363 90.946 -62.621 55.002
466364 90.989 -62.631 54.993
466365 91.034 -62.641 54.985
466366 91.088 -62.652 54.977
466367 91.141 -62.663 54.970
466368 91.193 -62.675 54.964
466369 91.247 -62.686 54.958
466370 91.300 -62.697 54.953
466371 91.353 -62.709 54.949
466372 91.406 -62.720 54.944
466373 91.459 -62.732 54.932
466374 91.512 -62.743 54.916
466375 91.565 -62.754 54.900
466376 91.613 -62.769 54.884
466377 91.658 -62.782 54.867
466378 91.703 -62.793 54.851
466379 91.748 -62.804 54.837
466380 91.793 -62.815 54.824
466381 91.839 -62.825 54.814
466382 91.883 -62.837 54.807
466383 91.929 -62.847 54.803
466384 91.974 -62.859 54.804
466385 92.010 -62.858 54.809
466386 92.044 -62.854 54.818
466387 92.078 -62.849 54.831
466388 92.111 -62.845 54.843
466389 92.146 -62.841 54.858
466390 92.184 -62.841 54.872
466391 92.227 -62.844 54.880
466392 92.270 -62.848 54.880
466393 92.312 -62.851 54.869
466394 92.355 -62.855 54.852
466395 92.398 -62.859 54.831
466396 92.440 -62.862 54.805
466397 92.483 -62.866 54.776
466398 92.525 -62.869 54.747
466399 92.568 -62.873 54.717
466400 92.611 -62.876 54.687
466401 92.653 -62.880 54.657
466402 92.692 -62.883 54.628
466403 92.727 -62.891 54.600
466404 92.756 -62.904 54.573
466405 92.783 -62.915 54.546
466406 92.810 -62.926 54.519
466407 92.844 -62.938 54.493
466408 92.889 -62.951 54.468
466409 92.933 -62.963 54.445
466410 92.978 -62.975 54.423
466411 93.022 -62.987 54.402
466412 93.066 -62.998 54.383
466413 93.111 -62.995 54.364
466414 93.161 -62.987 54.345
466415 93.212 -62.979 54.327
466416 93.263 -62.973 54.310
466417 93.303 -62.984 54.291
466418 93.309 -62.985 54.274
466419 93.315 -62.978 54.256
466420 93.362 -62.977 54.238
466421 93.387 -62.998 54.220
466422 93.419 -63.014 54.204
466423 93.457 -63.024 54.186
466424 93.496 -63.035 54.168
466425 93.535 -63.046 54.149
466426 93.573 -63.057 54.131
466427 93.612 -63.068 54.113
466428 93.650 -63.079 54.095
466429 93.690 -63.092 54.076
466430 93.730 -63.105 54.057
466431 93.770 -63.119 54.038
466432 93.810 -63.132 54.020
466433 93.857 -63.136 54.001
466434 93.906 -63.149 53.984
466435 93.955 -63.165 53.966
466436 94.001 -63.167 53.948
466437 94.046 -63.167 53.931
466438 94.091 -63.165 53.916
466439 94.133 -63.180 53.901
466440 94.168 -63.216 53.888
466441 94.200 -63.222 53.876
466442 94.232 -63.227 53.864
466443 94.265 -63.233 53.853
466444 94.298 -63.239 53.843
466445 94.331 -63.245 53.833
466446 94.363 -63.250 53.824
466447 94.396 -63.256 53.815
466448 94.428 -63.261 53.803
466449 94.461 -63.267 53.790
466450 94.493 -63.273 53.775
466451 94.526 -63.279 53.759
466452 94.559 -63.284 53.741
466453 94.591 -63.290 53.722
466454 94.625 -63.296 53.702
466455 94.678 -63.299 53.680
466456 94.741 -63.301 53.657
466457 94.803 -63.302 53.635
466458 94.866 -63.304 53.613
466459 94.928 -63.305 53.591
466460 94.992 -63.309 53.567
466461 95.056 -63.314 53.541
466462 95.120 -63.321 53.515
466463 95.156 -63.315 53.488
466464 95.193 -63.320 53.461
466465 95.231 -63.325 53.435
466466 95.267 -63.330 53.408
466467 95.305 -63.336 53.380
466468 95.341 -63.340 53.354
466469 95.379 -63.346 53.328
466470 95.415 -63.351 53.304
466471 95.453 -63.355 53.280
466472 95.500 -63.326 53.257
466473 95.523 -63.339 53.232
466474 95.554 -63.336 53.209
466475 95.585 -63.334 53.184
466476 95.624 -63.334 53.158
466477 95.670 -63.337 53.130
466478 95.716 -63.341 53.103
466479 95.763 -63.348 53.074
466480 95.809 -63.354 53.045
466481 95.856 -63.361 53.018
466482 95.902 -63.368 52.989
466483 95.949 -63.374 52.959
466484 95.996 -63.380 52.930
466485 96.039 -63.386 52.899
466486 96.073 -63.390 52.865
466487 96.111 -63.390 52.826
466488 96.151 -63.392 52.784
466489 96.187 -63.405 52.738
466490 96.223 -63.417 52.688
466491 96.291 -63.416 52.636
466492 96.372 -63.408 52.586
466493 96.453 -63.402 52.539
466494 96.527 -63.395 52.492
466495 96.569 -63.393 52.447
466496 96.611 -63.392 52.407
466497 96.653 -63.390 52.368
466498 96.695 -63.389 52.331
466499 96.763 -63.386 52.297
466500 96.795 -63.385 52.268
466501 96.823 -63.394 52.243
466502 96.854 -63.406 52.216
466503 96.884 -63.418 52.185
466504 96.915 -63.430 52.151
466505 96.945 -63.442 52.117
466506 96.976 -63.454 52.084
466507 97.006 -63.466 52.053
466508 97.037 -63.478 52.019
466509 97.067 -63.490 51.984
466510 97.097 -63.502 51.950
466511 97.128 -63.514 51.915
466512 97.158 -63.526 51.883
466513 97.189 -63.538 51.852
466514 97.221 -63.551 51.823
466515 97.252 -63.563 51.795
466516 97.281 -63.576 51.766
466517 97.303 -63.588 51.738
466518 97.326 -63.599 51.709
466519 97.349 -63.610 51.679
466520 97.372 -63.621 51.648
466521 97.394 -63.631 51.617
466522 97.418 -63.642 51.587
466523 97.449 -63.644 51.558
466524 97.499 -63.628 51.527
466525 97.527 -63.633 51.492
466526 97.554 -63.637 51.453
466527 97.593 -63.649 51.411
466528 97.641 -63.663 51.365
466529 97.685 -63.670 51.317
466530 97.724 -63.668 51.268
466531 97.764 -63.665 51.219
466532 97.803 -63.663 51.170
466533 97.842 -63.662 51.120
466534 97.881 -63.659 51.069
466535 97.921 -63.657 51.017
466536 97.960 -63.655 50.966
466537 97.999 -63.653 50.913
466538 98.034 -63.654 50.860
466539 98.069 -63.655 50.807
466540 98.103 -63.656 50.754
466541 98.142 -63.655 50.700
466542 98.184 -63.651 50.648
466543 98.219 -63.653 50.596
466544 98.237 -63.666 50.545
466545 98.255 -63.681 50.494
466546 98.274 -63.694 50.445
466547 98.293 -63.708 50.396
466548 98.311 -63.722 50.348
466549 98.331 -63.735 50.301
466550 98.365 -63.739 50.254
466551 98.398 -63.742 50.207
466552 98.424 -63.743 50.159
466553 98.447 -63.742 50.111
466554 98.471 -63.741 50.062
466555 98.494 -63.741 50.014
466556 98.518 -63.740 49.965
466557 98.541 -63.739 49.917
466558 98.565 -63.739 49.867
466559 98.589 -63.738 49.817
466560 98.615 -63.725 49.767
466561 98.656 -63.721 49.719
466562 98.719 -63.719 49.669
466563 98.782 -63.717 49.622
466564 98.845 -63.715 49.575
466565 98.908 -63.713 49.530
466566 98.971 -63.712 49.485
466567 99.034 -63.709 49.443
466568 99.097 -63.708 49.401
466569 99.160 -63.706 49.359
466570 99.223 -63.704 49.318
466571 99.286 -63.703 49.279
466572 99.349 -63.700 49.241
466977 87.048 -61.865 56.172
466978 87.090 -61.863 56.159
466979 87.133 -61.862 56.147
466980 87.176 -61.860 56.136
466981 87.212 -61.862 56.126
466982 87.249 -61.863 56.115
466983 87.286 -61.865 56.105
466984 87.323 -61.867 56.095
466985 87.359 -61.869 56.086
466986 87.396 -61.871 56.079
466987 87.433 -61.873 56.070
466988 87.469 -61.874 56.060
466989 87.506 -61.876 56.051
466990 87.543 -61.878 56.039
466991 87.586 -61.877 56.028
466992 87.634 -61.873 56.015
466993 87.682 -61.870 55.999
466994 87.730 -61.867 55.980
466995 87.779 -61.863 55.958
466996 87.827 -61.860 55.938
466997 87.875 -61.857 55.919
466998 87.924 -61.854 55.902
466999 87.972 -61.850 55.885
467000 88.020 -61.847 55.868
467001 88.068 -61.843 55.850
467002 88.117 -61.840 55.832
467003 88.165 -61.837 55.814
467004 88.216 -61.835 55.797
467005 88.283 -61.841 55.780
467006 88.344 -61.851 55.763
467007 88.393 -61.866 55.743
467008 88.441 -61.881 55.721
467009 88.489 -61.897 55.696
467010 88.537 -61.912 55.665
467011 88.586 -61.927 55.638
467012 88.634 -61.942 55.616
467013 88.682 -61.958 55.599
467014 88.727 -61.975 55.584
467015 88.771 -61.992 55.570
467016 88.816 -62.010 55.555
467017 88.860 -62.027 55.536
467018 88.909 -62.058 55.515
467019 88.949 -62.065 55.493
467020 88.986 -62.067 55.471
467021 89.023 -62.068 55.451
467022 89.064 -62.076 55.432
467023 89.107 -62.085 55.415
467024 89.150 -62.095 55.399
467025 89.192 -62.104 55.383
467026 89.235 -62.114 55.368
467027 89.277 -62.123 55.353
467028 89.320 -62.133 55.338
467029 89.362 -62.142 55.322
467030 89.405 -62.152 55.310
467031 89.447 -62.161 55.298
467032 89.490 -62.171 55.288
467033 89.532 -62.180 55.281
467034 89.575 -62.190 55.271
467035 89.617 -62.200 55.264
467036 89.660 -62.209 55.258
467037 89.702 -62.219 55.254
467038 89.745 -62.229 55.253
467039 89.788 -62.238 55.255
467040 89.831 -62.251 55.257
467041 89.877 -62.272 55.258
467042 89.922 -62.293 55.256
467043 89.967 -62.315 55.253
467044 90.014 -62.334 55.248
467045 90.065 -62.347 55.240
467046 90.117 -62.359 55.227
467047 90.169 -62.372 55.211
467048 90.221 -62.384 55.194
467049 90.273 -62.397 55.178
467050 90.325 -62.410 55.168
467051 90.374 -62.423 55.163
467052 90.423 -62.437 55.162
467053 90.471 -62.451 55.163
467054 90.519 -62.465 55.159
467055 90.567 -62.479 55.146
467056 90.615 -62.493 55.126
467057 90.663 -62.506 55.102
467058 90.706 -62.516 55.079
467059 90.748 -62.526 55.060
467060 90.791 -62.536 55.045
467061 90.833 -62.546 55.033
467062 90.876 -62.556 55.022
467063 90.919 -62.565 55.012
467064 90.961 -62.575 55.005
467065 91.003 -62.585 54.998
467066 91.044 -62.594 54.989
467067 91.096 -62.606 54.980
467068 91.149 -62.617 54.973
467069 91.202 -62.628 54.968
467070 91.255 -62.640 54.962
467071 91.308 -62.651 54.957
467072 91.361 -62.663 54.953
467073 91.415 -62.674 54.949
467074 91.467 -62.685 54.937
467075 91.520 -62.697 54.921
467076 91.571 -62.710 54.905
467077 91.616 -62.725 54.889
467078 91.660 -62.742 54.873
467079 91.705 -62.755 54.858
467080 91.750 -62.766 54.843
467081 91.795 -62.777 54.830
467082 91.841 -62.788 54.819
467083 91.885 -62.799 54.810
467084 91.931 -62.810 54.806
467085 91.976 -62.821 54.805
467086 92.014 -62.823 54.808
467087 92.048 -62.819 54.816
467088 92.082 -62.815 54.828
467089 92.116 -62.810 54.839
467090 92.150 -62.806 54.850
467091 92.183 -62.802 54.860
467092 92.224 -62.803 54.866
467093 92.267 -62.807 54.865
467094 92.309 -62.810 54.855
467095 92.352 -62.813 54.837
467096 92.394 -62.817 54.816
467097 92.437 -62.821 54.790
467098 92.480 -62.824 54.761
467099 92.522 -62.828 54.731
467100 92.565 -62.832 54.700
467101 92.607 -62.835 54.670
467102 92.650 -62.838 54.640
467103 92.692 -62.842 54.612
467104 92.732 -62.847 54.585
467105 92.759 -62.858 54.558
467106 92.786 -62.869 54.530
467107 92.813 -62.881 54.503
467108 92.852 -62.892 54.476
467109 92.897 -62.905 54.451
467110 92.941 -62.917 54.428
467111 92.986 -62.929 54.406
467112 93.029 -62.940 54.385
467113 93.071 -62.939 54.365
467114 93.112 -62.939 54.347
467115 93.160 -62.932 54.328
467116 93.211 -62.924 54.310
467117 93.261 -62.924 54.293
467118 93.310 -62.940 54.276
467119 93.339 -62.948 54.258
467120 93.311 -62.938 54.241
467121 93.353 -62.941 54.223
467122 93.383 -62.958 54.207
467123 93.422 -62.969 54.191
467124 93.461 -62.980 54.174
467125 93.499 -62.991 54.157
467126 93.539 -63.004 54.140
467127 93.579 -63.017 54.123
467128 93.619 -63.030 54.106
467129 93.659 -63.043 54.089
467130 93.699 -63.056 54.071
467131 93.739 -63.070 54.052
467132 93.778 -63.082 54.034
467133 93.818 -63.095 54.016
467134 93.866 -63.099 53.998
467135 93.917 -63.103 53.980
467136 93.966 -63.119 53.962
467137 94.015 -63.135 53.945
467138 94.063 -63.150 53.927
467139 94.111 -63.158 53.912
467140 94.154 -63.161 53.899
467141 94.188 -63.181 53.886
467142 94.221 -63.186 53.874
467143 94.253 -63.192 53.863
467144 94.286 -63.198 53.852
467145 94.319 -63.204 53.840
467146 94.351 -63.209 53.830
467147 94.384 -63.214 53.820
467148 94.416 -63.219 53.812
467149 94.449 -63.224 53.799
467150 94.481 -63.227 53.785
467151 94.513 -63.232 53.770
467152 94.546 -63.236 53.752
467153 94.578 -63.240 53.732
467154 94.610 -63.245 53.713
467155 94.638 -63.257 53.692
467156 94.674 -63.258 53.669
467157 94.714 -63.255 53.645
467158 94.773 -63.261 53.622
467159 94.839 -63.270 53.599
467160 94.905 -63.280 53.575
467161 94.971 -63.290 53.551
467162 95.037 -63.300 53.525
467163 95.103 -63.310 53.499
467164 95.146 -63.286 53.471
467165 95.178 -63.270 53.444
467166 95.216 -63.277 53.418
467167 95.254 -63.285 53.391
467168 95.293 -63.292 53.362
467169 95.331 -63.299 53.334
467170 95.369 -63.305 53.306
467171 95.408 -63.316 53.280
467172 95.468 -63.279 53.254
467173 95.518 -63.244 53.230
467174 95.539 -63.259 53.204
467175 95.560 -63.275 53.180
467176 95.586 -63.281 53.154
467177 95.628 -63.283 53.127
467178 95.675 -63.289 53.099
467179 95.721 -63.296 53.070
467180 95.768 -63.302 53.041
467181 95.814 -63.308 53.011
467182 95.861 -63.315 52.982
467183 95.908 -63.322 52.953
467184 95.954 -63.329 52.924
467185 96.001 -63.335 52.895
467186 96.043 -63.341 52.866
467187 96.077 -63.345 52.833
467188 96.113 -63.346 52.797
467189 96.154 -63.343 52.758
467190 96.192 -63.350 52.714
467191 96.228 -63.363 52.665
467192 96.264 -63.376 52.613
467193 96.316 -63.382 52.562
467194 96.397 -63.375 52.513
467195 96.478 -63.368 52.466
467196 96.559 -63.360 52.420
467197 96.640 -63.353 52.378
467198 96.686 -63.351 52.338
467199 96.732 -63.355 52.300
467200 96.788 -63.361 52.266
467201 96.810 -63.363 52.237
467202 96.833 -63.367 52.210
467203 96.863 -63.378 52.185
467204 96.894 -63.390 52.154
467205 96.924 -63.402 52.121
467206 96.954 -63.415 52.087
467207 96.985 -63.427 52.053
467208 97.015 -63.439 52.021
467209 97.046 -63.450 51.988
467210 97.076 -63.462 51.954
467211 97.107 -63.474 51.919
467212 97.137 -63.487 51.884
467213 97.168 -63.499 51.851
467214 97.198 -63.510 51.820
467215 97.229 -63.522 51.790
467216 97.253 -63.531 51.761
467217 97.279 -63.537 51.732
467218 97.302 -63.548 51.701
467219 97.324 -63.559 51.671
467220 97.348 -63.569 51.641
467221 97.370 -63.581 51.610
467222 97.393 -63.591 51.579
467223 97.416 -63.602 51.549
467224 97.459 -63.593 51.521
467225 97.509 -63.577 51.492
467226 97.537 -63.581 51.460
467227 97.564 -63.586 51.425
467228 97.590 -63.591 51.383
467229 97.636 -63.605 51.338
467230 97.683 -63.619 51.292
467231 97.731 -63.632 51.245
467232 97.770 -63.631 51.199
467233 97.809 -63.628 51.150
467234 97.849 -63.626 51.101
467235 97.888 -63.624 51.050
467236 97.927 -63.622 50.998
467237 97.966 -63.620 50.947
467238 98.006 -63.618 50.895
467239 98.041 -63.618 50.842
467240 98.076 -63.619 50.790
467241 98.111 -63.621 50.736
467242 98.152 -63.617 50.682
467243 98.195 -63.613 50.629
467244 98.230 -63.614 50.577
467245 98.248 -63.628 50.526
467246 98.267 -63.642 50.476
467247 98.285 -63.656 50.426
467248 98.304 -63.670 50.376
467249 98.322 -63.684 50.328
467250 98.341 -63.697 50.283
467251 98.359 -63.712 50.238
467252 98.390 -63.716 50.193
467253 98.422 -63.719 50.145
467254 98.445 -63.719 50.096
467255 98.469 -63.718 50.048
467256 98.493 -63.717 49.999
467257 98.516 -63.717 49.949
467258 98.539 -63.716 49.898
467259 98.565 -63.706 49.847
467260 98.591 -63.694 49.797
467261 98.617 -63.681 49.747
467262 98.665 -63.681 49.698
467263 98.728 -63.680 49.651
467264 98.791 -63.678 49.605
467265 98.854 -63.676 49.560
467266 98.917 -63.675 49.516
467267 98.980 -63.672 49.472
467268 99.043 -63.671 49.430
467269 99.106 -63.669 49.389
467270 99.168 -63.667 49.349
467271 99.231 -63.665 49.308
467272 99.294 -63.663 49.269
467273 99.357 -63.662 49.232
467678 87.065 -61.816 56.171
467679 87.107 -61.815 56.158
467680 87.149 -61.813 56.146
467681 87.191 -61.812 56.135
467682 87.228 -61.813 56.124
467683 87.264 -61.816 56.113
467684 87.301 -61.817 56.103
467685 87.338 -61.820 56.093
467686 87.374 -61.821 56.083
467687 87.411 -61.823 56.075
467688 87.448 -61.825 56.066
467689 87.485 -61.826 56.056
467690 87.521 -61.829 56.048
467691 87.558 -61.830 56.037
467692 87.596 -61.832 56.024
467693 87.643 -61.828 56.012
467694 87.692 -61.825 55.998
467695 87.740 -61.822 55.980
467696 87.788 -61.818 55.959
467697 87.837 -61.815 55.940
467698 87.885 -61.811 55.922
467699 87.933 -61.808 55.905
467700 87.981 -61.804 55.889
467701 88.030 -61.801 55.874
467702 88.078 -61.798 55.856
467703 88.126 -61.794 55.838
467704 88.175 -61.791 55.820
467705 88.227 -61.790 55.802
467706 88.294 -61.797 55.783
467707 88.361 -61.804 55.763
467708 88.409 -61.819 55.742
467709 88.458 -61.834 55.719
467710 88.506 -61.849 55.694
467711 88.554 -61.865 55.664
467712 88.603 -61.880 55.639
467713 88.651 -61.895 55.620
467714 88.698 -61.911 55.605
467715 88.741 -61.929 55.592
467716 88.785 -61.947 55.580
467717 88.830 -61.964 55.565
467718 88.876 -61.988 55.546
467719 88.926 -62.020 55.522
467720 88.963 -62.023 55.496
467721 89.000 -62.024 55.471
467722 89.038 -62.026 55.449
467723 89.075 -62.027 55.430
467724 89.117 -62.035 55.412
467725 89.159 -62.044 55.396
467726 89.202 -62.054 55.381
467727 89.244 -62.063 55.365
467728 89.287 -62.073 55.350
467729 89.330 -62.083 55.335
467730 89.372 -62.092 55.319
467731 89.415 -62.102 55.307
467732 89.457 -62.111 55.295
467733 89.500 -62.120 55.285
467734 89.542 -62.130 55.277
467735 89.585 -62.139 55.267
467736 89.627 -62.149 55.260
467737 89.670 -62.159 55.254
467738 89.712 -62.168 55.251
467739 89.755 -62.178 55.250
467740 89.798 -62.188 55.255
467741 89.841 -62.202 55.259
467742 89.887 -62.224 55.261
467743 89.932 -62.245 55.261
467744 89.977 -62.266 55.261
467745 90.023 -62.287 55.257
467746 90.074 -62.302 55.248
467747 90.125 -62.315 55.235
467748 90.177 -62.327 55.218
467749 90.229 -62.340 55.200
467750 90.281 -62.352 55.183
467751 90.333 -62.365 55.168
467752 90.385 -62.377 55.158
467753 90.435 -62.390 55.153
467754 90.483 -62.404 55.151
467755 90.532 -62.418 55.144
467756 90.580 -62.432 55.131
467757 90.628 -62.446 55.111
467758 90.676 -62.460 55.089
467759 90.720 -62.471 55.070
467760 90.763 -62.481 55.055
467761 90.805 -62.490 55.043
467762 90.848 -62.500 55.032
467763 90.891 -62.510 55.023
467764 90.933 -62.520 55.015
467765 90.976 -62.530 55.011
467766 91.017 -62.540 55.004
467767 91.057 -62.549 54.995
467768 91.105 -62.559 54.986
467769 91.158 -62.571 54.980
467770 91.211 -62.582 54.977
467771 91.264 -62.594 54.971
467772 91.317 -62.605 54.966
467773 91.370 -62.616 54.961
467774 91.423 -62.628 54.953
467775 91.474 -62.640 54.942
467776 91.525 -62.652 54.927
467777 91.576 -62.665 54.911
467778 91.619 -62.681 54.895
467779 91.663 -62.699 54.882
467780 91.707 -62.716 54.867
467781 91.752 -62.728 54.852
467782 91.797 -62.739 54.838
467783 91.843 -62.750 54.826
467784 91.887 -62.761 54.815
467785 91.932 -62.772 54.809
467786 91.978 -62.783 54.806
467787 92.018 -62.788 54.808
467788 92.052 -62.784 54.814
467789 92.086 -62.780 54.825
467790 92.120 -62.775 54.834
467791 92.154 -62.772 54.841
467792 92.188 -62.767 54.848
467793 92.222 -62.763 54.852
467794 92.263 -62.765 54.849
467795 92.306 -62.769 54.839
467796 92.348 -62.772 54.821
467797 92.391 -62.775 54.799
467798 92.433 -62.779 54.774
467799 92.476 -62.783 54.745
467800 92.519 -62.786 54.713
467801 92.561 -62.790 54.682
467802 92.604 -62.794 54.651
467803 92.646 -62.797 54.621
467804 92.689 -62.800 54.594
467805 92.733 -62.802 54.567
467806 92.763 -62.812 54.541
467807 92.790 -62.823 54.515
467808 92.816 -62.835 54.489
467809 92.860 -62.847 54.463
467810 92.905 -62.859 54.437
467811 92.949 -62.872 54.414
467812 92.993 -62.881 54.392
467813 93.034 -62.880 54.371
467814 93.075 -62.880 54.351
467815 93.116 -62.879 54.332
467816 93.159 -62.878 54.313
467817 93.210 -62.869 54.295
467818 93.260 -62.875 54.278
467819 93.309 -62.891 54.261
467820 93.335 -62.900 54.244
467821 93.340 -62.901 54.227
467822 93.349 -62.903 54.211
467823 93.388 -62.915 54.195
467824 93.428 -62.928 54.181
467825 93.468 -62.941 54.165
467826 93.507 -62.954 54.149
467827 93.548 -62.967 54.133
467828 93.587 -62.981 54.117
467829 93.627 -62.994 54.101
467830 93.667 -63.007 54.084
467831 93.707 -63.020 54.067
467832 93.747 -63.033 54.049
467833 93.787 -63.046 54.030
467834 93.827 -63.060 54.012
467835 93.876 -63.062 53.995
467836 93.927 -63.061 53.976
467837 93.977 -63.072 53.958
467838 94.026 -63.088 53.942
467839 94.074 -63.104 53.925
467840 94.123 -63.120 53.910
467841 94.173 -63.138 53.896
467842 94.208 -63.140 53.884
467843 94.241 -63.145 53.873
467844 94.273 -63.148 53.862
467845 94.305 -63.153 53.850
467846 94.338 -63.158 53.839
467847 94.370 -63.161 53.828
467848 94.402 -63.166 53.818
467849 94.435 -63.170 53.806
467850 94.467 -63.174 53.792
467851 94.499 -63.179 53.777
467852 94.532 -63.183 53.761
467853 94.564 -63.187 53.743
467854 94.597 -63.192 53.723
467855 94.624 -63.205 53.703
467856 94.649 -63.220 53.682
467857 94.686 -63.220 53.658
467858 94.726 -63.217 53.635
467859 94.766 -63.214 53.610
467860 94.807 -63.211 53.586
467861 94.859 -63.214 53.562
467862 94.925 -63.224 53.537
467863 94.991 -63.233 53.511
467864 95.057 -63.243 53.484
467865 95.116 -63.246 53.456
467866 95.167 -63.238 53.429
467867 95.198 -63.226 53.402
467868 95.237 -63.233 53.376
467869 95.275 -63.239 53.345
467870 95.315 -63.256 53.316
467871 95.359 -63.289 53.287
467872 95.405 -63.314 53.258
467873 95.478 -63.222 53.230
467874 95.536 -63.163 53.205
467875 95.551 -63.187 53.177
467876 95.565 -63.212 53.150
467877 95.580 -63.237 53.123
467878 95.625 -63.243 53.095
467879 95.679 -63.244 53.067
467880 95.727 -63.250 53.038
467881 95.773 -63.257 53.007
467882 95.820 -63.264 52.976
467883 95.866 -63.270 52.946
467884 95.913 -63.277 52.916
467885 95.960 -63.283 52.887
467886 96.006 -63.289 52.859
467887 96.047 -63.296 52.830
467888 96.081 -63.299 52.799
467889 96.114 -63.303 52.765
467890 96.156 -63.299 52.727
467891 96.197 -63.296 52.685
467892 96.233 -63.309 52.639
467893 96.269 -63.322 52.588
467894 96.306 -63.335 52.537
467895 96.342 -63.348 52.487
467896 96.422 -63.341 52.440
467897 96.503 -63.333 52.394
467898 96.584 -63.326 52.350
467899 96.679 -63.317 52.309
467900 96.712 -63.330 52.270
467901 96.801 -63.339 52.235
467902 96.824 -63.342 52.205
467903 96.846 -63.343 52.178
467904 96.872 -63.351 52.153
467905 96.903 -63.363 52.124
467906 96.933 -63.375 52.092
467907 96.964 -63.386 52.058
467908 96.994 -63.399 52.023
467909 97.025 -63.411 51.991
467910 97.055 -63.423 51.958
467911 97.085 -63.435 51.924
467912 97.116 -63.447 51.890
467913 97.146 -63.459 51.855
467914 97.177 -63.471 51.821
467915 97.201 -63.480 51.790
467916 97.223 -63.486 51.759
467917 97.256 -63.492 51.729
467918 97.283 -63.498 51.698
467919 97.300 -63.508 51.666
467920 97.323 -63.518 51.635
467921 97.346 -63.529 51.604
467922 97.369 -63.540 51.573
467923 97.392 -63.550 51.543
467924 97.420 -63.556 51.514
467925 97.469 -63.540 51.486
467926 97.519 -63.524 51.456
467927 97.547 -63.529 51.426
467928 97.574 -63.534 51.393
467929 97.600 -63.540 51.353
467930 97.630 -63.546 51.310
467931 97.678 -63.560 51.266
467932 97.726 -63.575 51.221
467933 97.774 -63.590 51.175
467934 97.816 -63.593 51.126
467935 97.855 -63.591 51.078
467936 97.894 -63.589 51.028
467937 97.934 -63.587 50.978
467938 97.973 -63.584 50.927
467939 98.012 -63.582 50.875
467940 98.049 -63.582 50.823
467941 98.083 -63.584 50.770
467942 98.120 -63.584 50.715
467943 98.162 -63.580 50.661
467944 98.205 -63.575 50.607
467945 98.241 -63.576 50.555
467946 98.260 -63.590 50.503
467947 98.278 -63.603 50.452
467948 98.296 -63.618 50.403
467949 98.315 -63.631 50.355
467950 98.334 -63.645 50.309
467951 98.352 -63.659 50.264
467952 98.370 -63.673 50.219
467953 98.389 -63.687 50.174
467954 98.417 -63.694 50.127
467955 98.443 -63.695 50.078
467956 98.467 -63.694 50.029
467957 98.491 -63.694 49.979
467958 98.515 -63.688 49.927
467959 98.541 -63.675 49.877
467960 98.567 -63.663 49.825
467961 98.593 -63.650 49.775
467962 98.624 -63.644 49.726
467963 98.673 -63.643 49.678
467964 98.736 -63.641 49.633
467965 98.799 -63.639 49.588
467966 98.862 -63.637 49.543
467967 98.925 -63.635 49.501
467968 98.988 -63.634 49.458
467969 99.051 -63.632 49.417
467970 99.114 -63.630 49.377
467971 99.177 -63.628 49.337
467972 99.240 -63.626 49.298
467973 99.303 -63.625 49.260
468379 87.081 -61.768 56.171
468380 87.124 -61.766 56.158
468381 87.166 -61.764 56.145
468382 87.206 -61.764 56.133
468383 87.243 -61.766 56.123
468384 87.280 -61.768 56.112
468385 87.317 -61.769 56.101
468386 87.353 -61.771 56.090
468387 87.390 -61.773 56.081
468388 87.426 -61.775 56.071
468389 87.463 -61.777 56.062
468390 87.500 -61.779 56.053
468391 87.537 -61.780 56.045
468392 87.573 -61.782 56.034
468393 87.610 -61.784 56.022
468394 87.653 -61.783 56.010
468395 87.701 -61.779 55.997
468396 87.750 -61.776 55.980
468397 87.798 -61.773 55.959
468398 87.846 -61.769 55.940
468399 87.894 -61.766 55.922
468400 87.942 -61.763 55.907
468401 87.991 -61.760 55.893
468402 88.039 -61.756 55.877
468403 88.087 -61.753 55.861
468404 88.136 -61.749 55.845
468405 88.184 -61.746 55.828
468406 88.238 -61.745 55.808
468407 88.306 -61.752 55.787
468408 88.373 -61.758 55.765
468409 88.426 -61.771 55.742
468410 88.474 -61.787 55.718
468411 88.523 -61.802 55.693
468412 88.571 -61.817 55.664
468413 88.619 -61.832 55.640
468414 88.668 -61.848 55.622
468415 88.713 -61.865 55.609
468416 88.756 -61.882 55.598
468417 88.800 -61.901 55.588
468418 88.843 -61.918 55.574
468419 88.893 -61.950 55.554
468420 88.941 -61.978 55.529
468421 88.978 -61.979 55.499
468422 89.016 -61.981 55.472
468423 89.053 -61.983 55.448
468424 89.091 -61.984 55.428
468425 89.128 -61.986 55.409
468426 89.169 -61.994 55.393
468427 89.212 -62.003 55.377
468428 89.254 -62.013 55.361
468429 89.297 -62.022 55.346
468430 89.339 -62.032 55.332
468431 89.382 -62.042 55.318
468432 89.424 -62.051 55.306
468433 89.467 -62.061 55.293
468434 89.510 -62.070 55.282
468435 89.552 -62.080 55.272
468436 89.595 -62.089 55.263
468437 89.637 -62.098 55.256
468438 89.680 -62.108 55.250
468439 89.723 -62.118 55.247
468440 89.765 -62.127 55.247
468441 89.808 -62.137 55.253
468442 89.852 -62.154 55.259
468443 89.897 -62.175 55.263
468444 89.943 -62.196 55.265
468445 89.988 -62.218 55.267
468446 90.033 -62.239 55.264
468447 90.082 -62.257 55.257
468448 90.133 -62.269 55.244
468449 90.185 -62.282 55.227
468450 90.237 -62.294 55.210
468451 90.289 -62.307 55.190
468452 90.341 -62.320 55.173
468453 90.393 -62.332 55.159
468454 90.445 -62.345 55.148
468455 90.496 -62.358 55.139
468456 90.544 -62.371 55.129
468457 90.592 -62.385 55.114
468458 90.641 -62.399 55.095
468459 90.689 -62.413 55.077
468460 90.735 -62.426 55.062
468461 90.778 -62.436 55.050
468462 90.820 -62.446 55.041
468463 90.863 -62.456 55.034
468464 90.905 -62.465 55.026
468465 90.948 -62.475 55.019
468466 90.990 -62.484 55.017
468467 91.030 -62.493 55.010
468468 91.071 -62.503 55.001
468469 91.113 -62.512 54.993
468470 91.167 -62.524 54.989
468471 91.220 -62.535 54.987
468472 91.273 -62.547 54.983
468473 91.325 -62.558 54.978
468474 91.378 -62.570 54.972
468475 91.429 -62.582 54.960
468476 91.480 -62.594 54.947
468477 91.531 -62.606 54.932
468478 91.579 -62.621 54.918
468479 91.623 -62.638 54.904
468480 91.666 -62.656 54.893
468481 91.710 -62.673 54.880
468482 91.754 -62.691 54.865
468483 91.799 -62.701 54.849
468484 91.844 -62.713 54.834
468485 91.889 -62.724 54.821
468486 91.934 -62.735 54.813
468487 91.980 -62.746 54.807
468488 92.022 -62.753 54.807
468489 92.056 -62.750 54.814
468490 92.090 -62.745 54.822
468491 92.124 -62.741 54.828
468492 92.158 -62.737 54.833
468493 92.192 -62.732 54.836
468494 92.226 -62.728 54.838
468495 92.260 -62.724 54.833
468496 92.302 -62.727 54.821
468497 92.345 -62.731 54.804
468498 92.387 -62.735 54.782
468499 92.430 -62.738 54.756
468500 92.473 -62.741 54.726
468501 92.515 -62.745 54.695
468502 92.558 -62.748 54.664
468503 92.600 -62.752 54.633
468504 92.643 -62.756 54.603
468505 92.687 -62.758 54.577
468506 92.730 -62.760 54.552
468507 92.766 -62.766 54.526
468508 92.793 -62.778 54.501
468509 92.824 -62.789 54.476
468510 92.868 -62.801 54.450
468511 92.913 -62.813 54.425
468512 92.956 -62.822 54.401
468513 92.997 -62.822 54.380
468514 93.038 -62.821 54.359
468515 93.079 -62.821 54.339
468516 93.120 -62.820 54.320
468517 93.161 -62.820 54.302
468518 93.209 -62.814 54.284
468519 93.259 -62.826 54.267
468520 93.274 -62.832 54.250
468521 93.290 -62.841 54.234
468522 93.320 -62.860 54.217
468523 93.349 -62.860 54.202
468524 93.396 -62.879 54.188
468525 93.436 -62.892 54.173
468526 93.476 -62.905 54.159
468527 93.516 -62.918 54.144
468528 93.556 -62.932 54.128
468529 93.595 -62.945 54.112
468530 93.636 -62.957 54.098
468531 93.676 -62.971 54.080
468532 93.715 -62.984 54.063
468533 93.755 -62.997 54.046
468534 93.795 -63.010 54.027
468535 93.835 -63.023 54.009
468536 93.885 -63.024 53.991
468537 93.937 -63.024 53.974
468538 93.988 -63.025 53.956
468539 94.037 -63.041 53.938
468540 94.085 -63.057 53.922
468541 94.134 -63.073 53.907
468542 94.189 -63.098 53.893
468543 94.226 -63.092 53.881
468544 94.259 -63.095 53.871
468545 94.291 -63.100 53.861
468546 94.324 -63.104 53.848
468547 94.356 -63.108 53.837
468548 94.389 -63.113 53.826
468549 94.421 -63.117 53.814
468550 94.454 -63.121 53.800
468551 94.486 -63.126 53.785
468552 94.518 -63.130 53.769
468553 94.551 -63.135 53.753
468554 94.583 -63.139 53.734
468555 94.609 -63.153 53.714
468556 94.635 -63.168 53.694
468557 94.661 -63.184 53.672
468558 94.698 -63.183 53.649
468559 94.738 -63.180 53.625
468560 94.778 -63.177 53.599
468561 94.818 -63.173 53.574
468562 94.858 -63.170 53.550
468563 94.899 -63.166 53.524
468564 94.946 -63.167 53.496
468565 95.012 -63.177 53.470
468566 95.074 -63.183 53.442
468567 95.127 -63.180 53.415
468568 95.180 -63.177 53.389
468569 95.222 -63.195 53.362
468570 95.268 -63.221 53.331
468571 95.314 -63.242 53.301
468572 95.360 -63.262 53.270
468573 95.407 -63.280 53.240
468574 95.476 -63.198 53.211
468575 95.529 -63.143 53.182
468576 95.538 -63.166 53.152
468577 95.548 -63.189 53.122
468578 95.559 -63.212 53.093
468579 95.605 -63.218 53.063
468580 95.659 -63.219 53.034
468581 95.713 -63.220 53.004
468582 95.767 -63.220 52.973
468583 95.821 -63.221 52.942
468584 95.872 -63.225 52.910
468585 95.918 -63.231 52.879
468586 95.965 -63.238 52.849
468587 96.012 -63.245 52.820
468588 96.051 -63.250 52.790
468589 96.085 -63.254 52.759
468590 96.118 -63.258 52.726
468591 96.157 -63.257 52.689
468592 96.198 -63.253 52.649
468593 96.238 -63.255 52.605
468594 96.274 -63.267 52.559
468595 96.311 -63.280 52.510
468596 96.347 -63.293 52.461
468597 96.383 -63.306 52.414
468598 96.469 -63.302 52.368
468599 96.592 -63.286 52.322
468600 96.643 -63.292 52.281
468601 96.721 -63.304 52.241
468602 96.810 -63.314 52.205
468603 96.836 -63.319 52.174
468604 96.860 -63.322 52.146
468605 96.882 -63.324 52.120
468606 96.912 -63.335 52.093
468607 96.942 -63.347 52.062
468608 96.973 -63.359 52.029
468609 97.003 -63.371 51.994
468610 97.034 -63.383 51.962
468611 97.064 -63.395 51.929
468612 97.095 -63.407 51.895
468613 97.125 -63.419 51.861
468614 97.151 -63.429 51.826
468615 97.171 -63.435 51.792
468616 97.192 -63.442 51.761
468617 97.226 -63.447 51.729
468618 97.259 -63.453 51.697
468619 97.288 -63.459 51.665
468620 97.299 -63.467 51.632
468621 97.322 -63.478 51.600
468622 97.345 -63.489 51.569
468623 97.367 -63.499 51.538
468624 97.390 -63.510 51.508
468625 97.430 -63.505 51.480
468626 97.479 -63.489 51.451
468627 97.529 -63.473 51.422
468628 97.557 -63.477 51.391
468629 97.584 -63.482 51.360
468630 97.610 -63.487 51.322
468631 97.638 -63.493 51.282
468632 97.672 -63.501 51.240
468633 97.720 -63.516 51.196
468634 97.768 -63.531 51.150
468635 97.816 -63.546 51.101
468636 97.862 -63.556 51.053
468637 97.901 -63.553 51.004
468638 97.940 -63.551 50.955
468639 97.980 -63.549 50.904
468640 98.019 -63.547 50.853
468641 98.056 -63.546 50.800
468642 98.091 -63.548 50.746
468643 98.130 -63.546 50.690
468644 98.172 -63.542 50.635
468645 98.215 -63.537 50.582
468646 98.252 -63.537 50.530
468647 98.271 -63.551 50.479
468648 98.289 -63.565 50.428
468649 98.307 -63.579 50.378
468650 98.326 -63.593 50.332
468651 98.345 -63.606 50.287
468652 98.363 -63.621 50.243
468653 98.381 -63.634 50.198
468654 98.400 -63.648 50.152
468655 98.419 -63.662 50.105
468656 98.442 -63.672 50.058
468657 98.465 -63.669 50.008
468658 98.492 -63.657 49.958
468659 98.518 -63.644 49.905
468660 98.544 -63.632 49.854
468661 98.570 -63.619 49.802
468662 98.596 -63.607 49.752
468663 98.630 -63.607 49.704
468664 98.682 -63.604 49.658
468665 98.745 -63.602 49.614
468666 98.808 -63.600 49.570
468667 98.871 -63.599 49.527
468668 98.934 -63.596 49.486
468669 98.997 -63.595 49.444
468670 99.060 -63.593 49.404
468671 99.123 -63.591 49.364
468672 99.186 -63.590 49.326
468673 99.249 -63.587 49.288
469080 87.097 -61.719 56.170
469081 87.140 -61.717 56.156
469082 87.183 -61.716 56.142
469083 87.221 -61.716 56.131
469084 87.258 -61.718 56.120
469085 87.295 -61.720 56.109
469086 87.332 -61.722 56.098
469087 87.368 -61.723 56.088
469088 87.405 -61.726 56.077
469089 87.441 -61.727 56.068
469090 87.478 -61.729 56.059
469091 87.515 -61.731 56.051
469092 87.552 -61.732 56.043
469093 87.588 -61.735 56.032
469094 87.625 -61.736 56.020
469095 87.662 -61.738 56.007
469096 87.711 -61.734 55.995
469097 87.759 -61.731 55.978
469098 87.808 -61.728 55.958
469099 87.855 -61.724 55.939
469100 87.904 -61.721 55.922
469101 87.952 -61.717 55.907
469102 88.000 -61.714 55.892
469103 88.049 -61.710 55.878
469104 88.097 -61.707 55.863
469105 88.145 -61.704 55.848
469106 88.193 -61.700 55.832
469107 88.249 -61.701 55.813
469108 88.317 -61.707 55.791
469109 88.384 -61.713 55.767
469110 88.442 -61.724 55.742
469111 88.491 -61.739 55.717
469112 88.539 -61.755 55.690
469113 88.588 -61.770 55.662
469114 88.636 -61.785 55.641
469115 88.684 -61.801 55.625
469116 88.729 -61.818 55.612
469117 88.772 -61.836 55.603
469118 88.815 -61.854 55.594
469119 88.860 -61.879 55.580
469120 88.910 -61.912 55.560
469121 88.955 -61.935 55.534
469122 88.993 -61.936 55.501
469123 89.030 -61.938 55.472
469124 89.068 -61.940 55.446
469125 89.105 -61.942 55.425
469126 89.142 -61.943 55.405
469127 89.180 -61.945 55.389
469128 89.222 -61.953 55.373
469129 89.264 -61.962 55.357
469130 89.307 -61.972 55.342
469131 89.349 -61.982 55.327
469132 89.392 -61.991 55.313
469133 89.435 -62.001 55.301
469134 89.477 -62.010 55.289
469135 89.520 -62.020 55.278
469136 89.562 -62.029 55.267
469137 89.605 -62.039 55.259
469138 89.648 -62.048 55.252
469139 89.690 -62.058 55.246
469140 89.733 -62.067 55.243
469141 89.775 -62.077 55.244
469142 89.818 -62.086 55.250
469143 89.863 -62.105 55.258
469144 89.908 -62.127 55.264
469145 89.953 -62.148 55.268
469146 89.999 -62.169 55.269
469147 90.044 -62.190 55.268
469148 90.089 -62.211 55.263
469149 90.141 -62.224 55.251
469150 90.193 -62.236 55.235
469151 90.245 -62.249 55.218
469152 90.297 -62.262 55.197
469153 90.349 -62.274 55.179
469154 90.401 -62.287 55.162
469155 90.453 -62.299 55.145
469156 90.505 -62.312 55.132
469157 90.556 -62.324 55.117
469158 90.605 -62.339 55.101
469159 90.653 -62.352 55.084
469160 90.701 -62.366 55.069
469161 90.750 -62.381 55.057
469162 90.793 -62.390 55.048
469163 90.835 -62.400 55.042
469164 90.878 -62.410 55.037
469165 90.921 -62.420 55.031
469166 90.962 -62.430 55.025
469167 91.003 -62.439 55.023
469168 91.043 -62.448 55.015
469169 91.084 -62.457 55.007
469170 91.125 -62.467 55.001
469171 91.175 -62.478 55.000
469172 91.228 -62.489 54.998
469173 91.281 -62.500 54.995
469174 91.333 -62.512 54.992
469175 91.384 -62.525 54.984
469176 91.435 -62.537 54.969
469177 91.486 -62.549 54.952
469178 91.537 -62.561 54.938
469179 91.582 -62.578 54.926
469180 91.626 -62.595 54.914
469181 91.669 -62.612 54.906
469182 91.713 -62.630 54.897
469183 91.757 -62.647 54.882
469184 91.801 -62.664 54.864
469185 91.846 -62.675 54.845
469186 91.891 -62.686 54.829
469187 91.936 -62.697 54.817
469188 91.982 -62.708 54.809
469189 92.026 -62.719 54.806
469190 92.060 -62.715 54.809
469191 92.094 -62.710 54.815
469192 92.128 -62.706 54.820
469193 92.162 -62.702 54.822
469194 92.196 -62.697 54.823
469195 92.230 -62.694 54.822
469196 92.264 -62.689 54.816
469197 92.299 -62.686 54.803
469198 92.342 -62.690 54.785
469199 92.384 -62.693 54.763
469200 92.427 -62.697 54.736
469201 92.470 -62.700 54.707
469202 92.512 -62.703 54.676
469203 92.555 -62.707 54.645
469204 92.597 -62.711 54.615
469205 92.640 -62.714 54.587
469206 92.684 -62.716 54.562
469207 92.727 -62.717 54.539
469208 92.769 -62.720 54.515
469209 92.796 -62.732 54.490
469210 92.832 -62.744 54.465
469211 92.876 -62.756 54.439
469212 92.919 -62.763 54.414
469213 92.960 -62.763 54.391
469214 93.001 -62.763 54.369
469215 93.043 -62.762 54.349
469216 93.083 -62.762 54.329
469217 93.125 -62.761 54.310
469218 93.165 -62.761 54.293
469219 93.208 -62.761 54.275
469220 93.235 -62.779 54.258
469221 93.266 -62.799 54.241
469222 93.295 -62.819 54.225
469223 93.325 -62.832 54.210
469224 93.347 -62.817 54.196
469225 93.390 -62.831 54.182
469226 93.437 -62.850 54.167
469227 93.484 -62.869 54.153
469228 93.524 -62.882 54.139
469229 93.564 -62.895 54.124
469230 93.604 -62.908 54.107
469231 93.644 -62.922 54.093
469232 93.684 -62.935 54.077
469233 93.724 -62.948 54.060
469234 93.764 -62.961 54.042
469235 93.803 -62.974 54.024
469236 93.843 -62.987 54.006
469237 93.895 -62.987 53.988
469238 93.946 -62.987 53.970
469239 93.998 -62.987 53.953
469240 94.048 -62.994 53.936
469241 94.097 -63.010 53.919
469242 94.151 -63.037 53.904
469243 94.205 -63.057 53.891
469244 94.246 -63.042 53.879
469245 94.278 -63.047 53.868
469246 94.310 -63.051 53.858
469247 94.342 -63.055 53.845
469248 94.375 -63.060 53.834
469249 94.407 -63.064 53.822
469250 94.439 -63.068 53.808
469251 94.472 -63.073 53.794
469252 94.504 -63.077 53.779
469253 94.537 -63.082 53.762
469254 94.569 -63.086 53.745
469255 94.595 -63.101 53.726
469256 94.620 -63.117 53.706
469257 94.646 -63.133 53.685
469258 94.672 -63.148 53.663
469259 94.710 -63.146 53.640
469260 94.750 -63.142 53.615
469261 94.790 -63.139 53.589
469262 94.830 -63.136 53.563
469263 94.871 -63.133 53.537
469264 94.910 -63.129 53.511
469265 94.951 -63.125 53.482
469266 94.990 -63.122 53.456
469267 95.032 -63.120 53.429
469268 95.102 -63.128 53.403
469269 95.173 -63.139 53.377
469270 95.240 -63.151 53.349
469271 95.295 -63.160 53.318
469272 95.338 -63.148 53.286
469273 95.382 -63.138 53.255
469274 95.426 -63.142 53.223
469275 95.473 -63.164 53.192
469276 95.524 -63.126 53.160
469277 95.533 -63.148 53.127
469278 95.558 -63.176 53.095
469279 95.575 -63.194 53.064
469280 95.608 -63.180 53.033
469281 95.658 -63.173 53.003
469282 95.709 -63.177 52.972
469283 95.760 -63.181 52.940
469284 95.811 -63.186 52.907
469285 95.862 -63.189 52.875
469286 95.912 -63.194 52.842
469287 95.964 -63.198 52.811
469288 96.017 -63.199 52.780
469289 96.056 -63.204 52.751
469290 96.095 -63.205 52.719
469291 96.127 -63.211 52.685
469292 96.150 -63.224 52.649
469293 96.195 -63.216 52.611
469294 96.240 -63.208 52.570
469295 96.285 -63.214 52.528
469296 96.330 -63.227 52.482
469297 96.376 -63.240 52.434
469298 96.422 -63.254 52.387
469299 96.486 -63.260 52.342
469300 96.575 -63.253 52.296
469301 96.653 -63.265 52.254
469302 96.732 -63.278 52.214
469303 96.819 -63.289 52.177
469304 96.846 -63.294 52.144
469305 96.872 -63.299 52.115
469306 96.896 -63.302 52.088
469307 96.921 -63.308 52.062
469308 96.952 -63.320 52.033
469309 96.982 -63.332 52.001
469310 97.012 -63.343 51.968
469311 97.043 -63.355 51.934
469312 97.073 -63.368 51.900
469313 97.100 -63.377 51.867
469314 97.120 -63.384 51.833
469315 97.140 -63.391 51.798
469316 97.162 -63.398 51.764
469317 97.195 -63.403 51.732
469318 97.229 -63.408 51.699
469319 97.262 -63.415 51.666
469320 97.294 -63.421 51.633
469321 97.302 -63.428 51.600
469322 97.320 -63.437 51.567
469323 97.343 -63.448 51.536
469324 97.366 -63.459 51.505
469325 97.389 -63.469 51.477
469326 97.439 -63.453 51.447
469327 97.489 -63.437 51.417
469328 97.539 -63.421 51.387
469329 97.567 -63.425 51.356
469330 97.594 -63.430 51.325
469331 97.620 -63.436 51.290
469332 97.648 -63.441 51.253
469333 97.674 -63.446 51.212
469334 97.715 -63.457 51.168
469335 97.762 -63.472 51.123
469336 97.811 -63.487 51.074
469337 97.859 -63.502 51.027
469338 97.907 -63.516 50.979
469339 97.947 -63.516 50.930
469340 97.987 -63.514 50.880
469341 98.026 -63.512 50.829
469342 98.064 -63.511 50.774
469343 98.098 -63.512 50.717
469344 98.140 -63.509 50.660
469345 98.182 -63.504 50.605
469346 98.225 -63.499 50.553
469347 98.263 -63.499 50.502
469348 98.281 -63.512 50.451
469349 98.300 -63.527 50.401
469350 98.318 -63.540 50.353
469351 98.337 -63.554 50.308
469352 98.356 -63.568 50.264
469353 98.374 -63.582 50.219
469354 98.392 -63.596 50.174
469355 98.411 -63.610 50.127
469356 98.427 -63.624 50.081
469357 98.442 -63.638 50.033
469358 98.468 -63.625 49.984
469359 98.494 -63.613 49.934
469360 98.520 -63.600 49.883
469361 98.546 -63.588 49.830
469362 98.572 -63.575 49.779
469363 98.599 -63.565 49.729
469364 98.636 -63.570 49.682
469365 98.691 -63.565 49.638
469366 98.754 -63.563 49.595
469367 98.816 -63.562 49.553
469368 98.879 -63.559 49.511
469369 98.942 -63.558 49.470
469370 99.005 -63.556 49.429
469371 99.068 -63.554 49.390
469372 99.131 -63.553 49.352
469373 99.194 -63.550 49.314
469374 99.257 -63.549 49.276
469782 87.156 -61.669 56.154
469783 87.199 -61.667 56.140
469784 87.236 -61.669 56.129
469785 87.273 -61.670 56.117
469786 87.310 -61.672 56.106
469787 87.347 -61.674 56.095
469788 87.383 -61.675 56.085
469789 87.420 -61.677 56.074
469790 87.457 -61.679 56.064
469791 87.494 -61.681 56.055
469792 87.530 -61.683 56.047
469793 87.567 -61.685 56.039
469794 87.604 -61.686 56.028
469795 87.640 -61.688 56.015
469796 87.677 -61.690 56.003
469797 87.720 -61.689 55.990
469798 87.769 -61.685 55.974
469799 87.817 -61.682 55.954
469800 87.865 -61.678 55.937
469801 87.913 -61.675 55.921
469802 87.962 -61.672 55.905
469803 88.010 -61.669 55.890
469804 88.058 -61.666 55.877
469805 88.107 -61.662 55.864
469806 88.154 -61.659 55.850
469807 88.203 -61.655 55.835
469808 88.260 -61.657 55.816
469809 88.328 -61.663 55.794
469810 88.395 -61.669 55.768
469811 88.460 -61.677 55.741
469812 88.508 -61.692 55.715
469813 88.556 -61.707 55.689
469814 88.605 -61.723 55.661
469815 88.652 -61.738 55.638
469816 88.701 -61.754 55.622
469817 88.744 -61.772 55.610
469818 88.787 -61.789 55.603
469819 88.831 -61.810 55.594
469820 88.877 -61.841 55.580
469821 88.926 -61.874 55.560
469822 88.970 -61.892 55.535
469823 89.008 -61.894 55.502
469824 89.045 -61.895 55.471
469825 89.083 -61.897 55.443
469826 89.120 -61.898 55.421
469827 89.158 -61.900 55.400
469828 89.195 -61.901 55.383
469829 89.233 -61.903 55.368
469830 89.274 -61.912 55.352
469831 89.317 -61.921 55.336
469832 89.359 -61.931 55.321
469833 89.402 -61.941 55.307
469834 89.445 -61.950 55.295
469835 89.487 -61.960 55.284
469836 89.530 -61.969 55.273
469837 89.573 -61.979 55.263
469838 89.615 -61.989 55.254
469839 89.658 -61.998 55.247
469840 89.700 -62.008 55.241
469841 89.743 -62.017 55.237
469842 89.785 -62.026 55.237
469843 89.828 -62.036 55.242
469844 89.873 -62.057 55.250
469845 89.918 -62.078 55.257
469846 89.964 -62.099 55.262
469847 90.009 -62.120 55.265
469848 90.054 -62.142 55.266
469849 90.100 -62.163 55.262
469850 90.149 -62.179 55.254
469851 90.201 -62.192 55.242
469852 90.253 -62.204 55.223
469853 90.305 -62.217 55.203
469854 90.357 -62.230 55.185
469855 90.408 -62.242 55.166
469856 90.460 -62.255 55.145
469857 90.512 -62.267 55.127
469858 90.565 -62.280 55.108
469859 90.616 -62.292 55.091
469860 90.666 -62.305 55.075
469861 90.714 -62.320 55.063
469862 90.762 -62.333 55.054
469863 90.807 -62.345 55.047
469864 90.850 -62.355 55.043
469865 90.892 -62.365 55.040
469866 90.935 -62.374 55.036
469867 90.976 -62.384 55.030
469868 91.017 -62.393 55.026
469869 91.057 -62.402 55.019
469870 91.098 -62.412 55.013
469871 91.138 -62.421 55.010
469872 91.184 -62.431 55.011
469873 91.237 -62.443 55.009
469874 91.288 -62.455 55.007
469875 91.338 -62.467 55.002
469876 91.390 -62.479 54.993
469877 91.441 -62.491 54.977
469878 91.492 -62.503 54.960
469879 91.540 -62.517 54.947
469880 91.585 -62.534 54.937
469881 91.629 -62.552 54.928
469882 91.672 -62.569 54.922
469883 91.716 -62.586 54.917
469884 91.760 -62.603 54.903
469885 91.804 -62.621 54.882
469886 91.848 -62.637 54.858
469887 91.893 -62.648 54.838
469888 91.938 -62.659 54.822
469889 91.984 -62.670 54.811
469890 92.028 -62.681 54.805
469891 92.065 -62.680 54.803
469892 92.098 -62.675 54.807
469893 92.132 -62.672 54.810
469894 92.166 -62.667 54.810
469895 92.200 -62.663 54.809
469896 92.234 -62.659 54.805
469897 92.268 -62.654 54.797
469898 92.302 -62.650 54.784
469899 92.339 -62.648 54.765
469900 92.381 -62.652 54.743
469901 92.424 -62.655 54.717
469902 92.466 -62.659 54.687
469903 92.509 -62.663 54.658
469904 92.552 -62.666 54.629
469905 92.594 -62.669 54.600
469906 92.638 -62.672 54.574
469907 92.681 -62.673 54.551
469908 92.725 -62.675 54.528
469909 92.769 -62.677 54.505
469910 92.799 -62.686 54.481
469911 92.840 -62.698 54.455
469912 92.882 -62.704 54.428
469913 92.924 -62.703 54.404
469914 92.965 -62.703 54.382
469915 93.006 -62.703 54.361
469916 93.047 -62.703 54.341
469917 93.089 -62.706 54.323
469918 93.133 -62.711 54.304
469919 93.176 -62.716 54.286
469920 93.212 -62.735 54.269
469921 93.241 -62.757 54.252
469922 93.271 -62.776 54.235
469923 93.301 -62.796 54.219
469924 93.327 -62.797 54.204
469925 93.349 -62.782 54.191
469926 93.383 -62.782 54.176
469927 93.431 -62.802 54.163
469928 93.479 -62.822 54.149
469929 93.526 -62.841 54.134
469930 93.572 -62.859 54.119
469931 93.613 -62.873 54.103
469932 93.652 -62.885 54.088
469933 93.692 -62.898 54.073
469934 93.732 -62.912 54.055
469935 93.772 -62.925 54.038
469936 93.812 -62.938 54.020
469937 93.852 -62.950 54.002
469938 93.904 -62.950 53.984
469939 93.956 -62.950 53.966
469940 94.007 -62.950 53.949
469941 94.059 -62.949 53.933
469942 94.112 -62.972 53.916
469943 94.168 -63.001 53.902
469944 94.220 -63.015 53.888
469945 94.264 -62.994 53.876
469946 94.296 -62.998 53.864
469947 94.329 -63.002 53.852
469948 94.361 -63.007 53.840
469949 94.394 -63.011 53.828
469950 94.426 -63.016 53.815
469951 94.459 -63.020 53.801
469952 94.491 -63.024 53.787
469953 94.523 -63.029 53.770
469954 94.555 -63.034 53.754
469955 94.580 -63.049 53.736
469956 94.606 -63.065 53.717
469957 94.631 -63.080 53.695
469958 94.657 -63.096 53.674
469959 94.683 -63.111 53.652
469960 94.722 -63.108 53.628
469961 94.762 -63.105 53.603
469962 94.802 -63.101 53.577
469963 94.842 -63.098 53.550
469964 94.882 -63.095 53.523
469965 94.922 -63.092 53.497
469966 94.963 -63.088 53.469
469967 95.002 -63.085 53.442
469968 95.046 -63.094 53.417
469969 95.102 -63.095 53.393
469970 95.173 -63.106 53.368
469971 95.244 -63.116 53.338
469972 95.306 -63.131 53.307
469973 95.365 -63.148 53.273
469974 95.414 -63.148 53.240
469975 95.458 -63.137 53.207
469976 95.501 -63.126 53.173
469977 95.545 -63.116 53.139
469978 95.570 -63.144 53.104
469979 95.597 -63.171 53.070
469980 95.590 -63.163 53.037
469981 95.621 -63.149 53.004
469982 95.671 -63.140 52.973
469983 95.721 -63.132 52.941
469984 95.771 -63.123 52.907
469985 95.822 -63.126 52.873
469986 95.873 -63.130 52.840
469987 95.923 -63.135 52.806
469988 95.974 -63.139 52.773
469989 96.025 -63.142 52.742
469990 96.074 -63.151 52.711
469991 96.113 -63.153 52.678
469992 96.133 -63.167 52.645
469993 96.155 -63.182 52.609
469994 96.185 -63.187 52.573
469995 96.271 -63.228 52.535
469996 96.315 -63.180 52.497
469997 96.363 -63.181 52.453
469998 96.412 -63.192 52.406
469999 96.462 -63.203 52.360
470000 96.511 -63.214 52.315
470001 96.586 -63.227 52.271
470002 96.664 -63.239 52.228
470003 96.742 -63.252 52.188
470004 96.828 -63.264 52.150
470005 96.855 -63.268 52.115
470006 96.881 -63.274 52.085
470007 96.908 -63.279 52.058
470008 96.932 -63.283 52.032
470009 96.961 -63.292 52.004
470010 96.991 -63.304 51.974
470011 97.022 -63.316 51.942
470012 97.049 -63.327 51.907
470013 97.070 -63.333 51.874
470014 97.090 -63.340 51.841
470015 97.109 -63.346 51.807
470016 97.132 -63.353 51.773
470017 97.165 -63.359 51.738
470018 97.198 -63.365 51.704
470019 97.232 -63.370 51.670
470020 97.265 -63.376 51.636
470021 97.298 -63.381 51.602
470022 97.308 -63.389 51.569
470023 97.319 -63.397 51.536
470024 97.342 -63.408 51.504
470025 97.364 -63.418 51.474
470026 97.399 -63.417 51.444
470027 97.449 -63.401 51.414
470028 97.499 -63.386 51.383
470029 97.549 -63.370 51.352
470030 97.577 -63.374 51.320
470031 97.604 -63.379 51.288
470032 97.631 -63.383 51.254
470033 97.658 -63.389 51.219
470034 97.684 -63.394 51.180
470035 97.711 -63.399 51.138
470036 97.757 -63.413 51.093
470037 97.805 -63.427 51.046
470038 97.853 -63.443 50.998
470039 97.901 -63.457 50.950
470040 97.949 -63.472 50.901
470041 97.993 -63.479 50.848
470042 98.032 -63.477 50.795
470043 98.071 -63.475 50.740
470044 98.107 -63.474 50.683
470045 98.150 -63.471 50.626
470046 98.193 -63.466 50.571
470047 98.235 -63.462 50.518
470048 98.274 -63.460 50.466
470049 98.292 -63.474 50.415
470050 98.311 -63.488 50.366
470051 98.329 -63.502 50.321
470052 98.348 -63.516 50.274
470053 98.367 -63.530 50.230
470054 98.385 -63.543 50.185
470055 98.398 -63.559 50.141
470056 98.409 -63.574 50.096
470057 98.420 -63.588 50.049
470058 98.444 -63.594 50.000
470059 98.471 -63.582 49.950
470060 98.496 -63.569 49.900
470061 98.522 -63.557 49.850
470062 98.549 -63.544 49.801
470063 98.574 -63.532 49.753
470064 98.605 -63.528 49.706
470065 98.642 -63.532 49.661
470066 98.699 -63.526 49.619
470067 98.762 -63.524 49.576
470068 98.825 -63.523 49.534
470069 98.888 -63.521 49.494
470070 98.951 -63.519 49.455
470071 99.014 -63.517 49.415
470072 99.077 -63.515 49.376
470073 99.140 -63.514 49.339
470074 99.203 -63.512 49.302
470483 87.173 -61.620 56.151
470484 87.215 -61.619 56.137
470485 87.252 -61.620 56.126
470486 87.289 -61.622 56.114
470487 87.325 -61.624 56.102
470488 87.362 -61.626 56.092
470489 87.399 -61.628 56.081
470490 87.435 -61.629 56.070
470491 87.472 -61.632 56.060
470492 87.509 -61.633 56.050
470493 87.546 -61.635 56.042
470494 87.582 -61.637 56.033
470495 87.619 -61.638 56.022
470496 87.655 -61.641 56.009
470497 87.692 -61.642 55.996
470498 87.730 -61.644 55.984
470499 87.778 -61.640 55.968
470500 87.826 -61.637 55.950
470501 87.874 -61.634 55.932
470502 87.923 -61.630 55.915
470503 87.971 -61.627 55.899
470504 88.020 -61.623 55.885
470505 88.068 -61.620 55.873
470506 88.116 -61.616 55.861
470507 88.164 -61.613 55.849
470508 88.212 -61.610 55.836
470509 88.271 -61.612 55.819
470510 88.338 -61.618 55.796
470511 88.406 -61.625 55.769
470512 88.473 -61.631 55.740
470513 88.525 -61.645 55.713
470514 88.573 -61.660 55.686
470515 88.621 -61.675 55.658
470516 88.670 -61.691 55.634
470517 88.716 -61.707 55.618
470518 88.759 -61.725 55.607
470519 88.803 -61.743 55.600
470520 88.848 -61.772 55.590
470521 88.894 -61.804 55.577
470522 88.943 -61.836 55.557
470523 88.985 -61.849 55.533
470524 89.023 -61.851 55.502
470525 89.060 -61.852 55.471
470526 89.097 -61.854 55.442
470527 89.135 -61.855 55.417
470528 89.172 -61.857 55.396
470529 89.210 -61.859 55.378
470530 89.247 -61.860 55.361
470531 89.285 -61.862 55.345
470532 89.327 -61.871 55.330
470533 89.370 -61.880 55.314
470534 89.412 -61.890 55.301
470535 89.455 -61.900 55.289
470536 89.497 -61.909 55.278
470537 89.540 -61.919 55.267
470538 89.582 -61.929 55.258
470539 89.625 -61.938 55.249
470540 89.667 -61.948 55.242
470541 89.710 -61.957 55.236
470542 89.753 -61.967 55.232
470543 89.795 -61.976 55.230
470544 89.838 -61.987 55.233
470545 89.883 -62.008 55.240
470546 89.929 -62.029 55.246
470547 89.974 -62.051 55.252
470548 90.019 -62.072 55.255
470549 90.065 -62.093 55.258
470550 90.110 -62.114 55.257
470551 90.157 -62.134 55.253
470552 90.209 -62.146 55.243
470553 90.261 -62.159 55.227
470554 90.313 -62.172 55.208
470555 90.365 -62.184 55.190
470556 90.416 -62.197 55.171
470557 90.468 -62.209 55.149
470558 90.520 -62.222 55.125
470559 90.572 -62.234 55.103
470560 90.624 -62.247 55.085
470561 90.676 -62.260 55.070
470562 90.727 -62.273 55.059
470563 90.775 -62.287 55.052
470564 90.822 -62.300 55.048
470565 90.865 -62.310 55.046
470566 90.907 -62.320 55.045
470567 90.949 -62.329 55.041
470568 90.990 -62.339 55.035
470569 91.030 -62.348 55.029
470570 91.071 -62.357 55.023
470571 91.111 -62.366 55.019
470572 91.152 -62.375 55.018
470573 91.192 -62.385 55.020
470574 91.242 -62.397 55.018
470575 91.293 -62.409 55.015
470576 91.345 -62.421 55.009
470577 91.395 -62.434 54.998
470578 91.446 -62.446 54.983
470579 91.498 -62.458 54.969
470580 91.543 -62.474 54.957
470581 91.587 -62.491 54.948
470582 91.632 -62.508 54.944
470583 91.675 -62.525 54.939
470584 91.719 -62.543 54.936
470585 91.763 -62.560 54.924
470586 91.807 -62.578 54.903
470587 91.851 -62.595 54.876
470588 91.895 -62.610 54.850
470589 91.940 -62.622 54.830
470590 91.986 -62.633 54.814
470591 92.030 -62.644 54.803
470592 92.069 -62.645 54.798
470593 92.102 -62.641 54.798
470594 92.136 -62.637 54.798
470595 92.170 -62.632 54.797
470596 92.204 -62.628 54.793
470597 92.238 -62.624 54.788
470598 92.272 -62.619 54.778
470599 92.306 -62.616 54.763
470600 92.340 -62.611 54.744
470601 92.378 -62.610 54.722
470602 92.420 -62.614 54.697
470603 92.463 -62.617 54.669
470604 92.506 -62.621 54.641
470605 92.548 -62.625 54.614
470606 92.591 -62.627 54.587
470607 92.635 -62.629 54.561
470608 92.678 -62.631 54.539
470609 92.722 -62.632 54.517
470610 92.766 -62.634 54.494
470611 92.803 -62.640 54.470
470612 92.846 -62.645 54.445
470613 92.887 -62.645 54.419
470614 92.930 -62.649 54.396
470615 92.973 -62.653 54.375
470616 93.016 -62.658 54.355
470617 93.060 -62.663 54.336
470618 93.103 -62.668 54.319
470619 93.146 -62.672 54.301
470620 93.190 -62.677 54.283
470621 93.219 -62.709 54.267
470622 93.247 -62.735 54.251
470623 93.277 -62.754 54.234
470624 93.307 -62.774 54.219
470625 93.330 -62.761 54.204
470626 93.352 -62.746 54.189
470627 93.377 -62.734 54.174
470628 93.425 -62.753 54.160
470629 93.473 -62.773 54.145
470630 93.520 -62.793 54.130
470631 93.568 -62.812 54.114
470632 93.615 -62.832 54.099
470633 93.660 -62.849 54.083
470634 93.701 -62.863 54.067
470635 93.741 -62.876 54.050
470636 93.780 -62.888 54.032
470637 93.820 -62.902 54.016
470638 93.862 -62.913 53.998
470639 93.913 -62.913 53.980
470640 93.965 -62.913 53.963
470641 94.016 -62.912 53.946
470642 94.073 -62.920 53.929
470643 94.129 -62.935 53.913
470644 94.184 -62.964 53.898
470645 94.236 -62.973 53.884
470646 94.281 -62.948 53.871
470647 94.315 -62.950 53.858
470648 94.347 -62.954 53.846
470649 94.380 -62.958 53.833
470650 94.412 -62.963 53.821
470651 94.445 -62.967 53.808
470652 94.477 -62.971 53.793
470653 94.509 -62.976 53.778
470654 94.541 -62.982 53.761
470655 94.566 -62.998 53.743
470656 94.592 -63.013 53.725
470657 94.617 -63.029 53.705
470658 94.642 -63.044 53.683
470659 94.668 -63.060 53.662
470660 94.694 -63.074 53.639
470661 94.734 -63.071 53.614
470662 94.774 -63.067 53.589
470663 94.814 -63.064 53.563
470664 94.854 -63.061 53.536
470665 94.894 -63.057 53.509
470666 94.934 -63.054 53.481
470667 94.974 -63.051 53.454
470668 95.011 -63.067 53.428
470669 95.062 -63.070 53.404
470670 95.115 -63.070 53.380
470671 95.173 -63.073 53.356
470672 95.244 -63.083 53.326
470673 95.313 -63.095 53.293
470674 95.372 -63.111 53.259
470675 95.431 -63.129 53.223
470676 95.489 -63.146 53.188
470677 95.534 -63.136 53.153
470678 95.578 -63.125 53.118
470679 95.609 -63.139 53.081
470680 95.628 -63.159 53.046
470681 95.605 -63.133 53.012
470682 95.633 -63.118 52.977
470683 95.683 -63.109 52.943
470684 95.734 -63.100 52.911
470685 95.783 -63.092 52.877
470686 95.834 -63.082 52.841
470687 95.884 -63.073 52.806
470688 95.935 -63.076 52.772
470689 95.985 -63.079 52.739
470690 96.036 -63.084 52.706
470691 96.092 -63.098 52.674
470692 96.119 -63.108 52.640
470693 96.140 -63.123 52.606
470694 96.161 -63.138 52.572
470695 96.183 -63.155 52.537
470696 96.302 -63.235 52.501
470697 96.346 -63.187 52.464
470698 96.390 -63.139 52.422
470699 96.439 -63.148 52.376
470700 96.488 -63.160 52.331
470701 96.538 -63.172 52.287
470702 96.594 -63.187 52.245
470703 96.674 -63.209 52.202
470704 96.753 -63.226 52.162
470705 96.834 -63.238 52.124
470706 96.864 -63.243 52.087
470707 96.890 -63.249 52.056
470708 96.917 -63.254 52.029
470709 96.944 -63.259 52.004
470710 96.970 -63.264 51.977
470711 96.999 -63.276 51.948
470712 97.019 -63.283 51.916
470713 97.039 -63.289 51.882
470714 97.059 -63.296 51.849
470715 97.079 -63.302 51.816
470716 97.101 -63.309 51.783
470717 97.135 -63.314 51.749
470718 97.168 -63.321 51.714
470719 97.201 -63.326 51.679
470720 97.235 -63.331 51.643
470721 97.268 -63.337 51.608
470722 97.301 -63.343 51.574
470723 97.313 -63.350 51.539
470724 97.322 -63.358 51.506
470725 97.340 -63.368 51.474
470726 97.363 -63.378 51.442
470727 97.409 -63.365 51.411
470728 97.459 -63.349 51.380
470729 97.509 -63.333 51.349
470730 97.559 -63.318 51.317
470731 97.587 -63.322 51.283
470732 97.614 -63.327 51.250
470733 97.641 -63.332 51.217
470734 97.668 -63.337 51.182
470735 97.695 -63.343 51.145
470736 97.721 -63.347 51.105
470737 97.751 -63.354 51.061
470738 97.799 -63.368 51.015
470739 97.847 -63.383 50.967
470740 97.895 -63.398 50.918
470741 97.943 -63.413 50.868
470742 97.992 -63.427 50.815
470743 98.039 -63.441 50.760
470744 98.078 -63.440 50.706
470745 98.118 -63.437 50.649
470746 98.160 -63.433 50.592
470747 98.203 -63.428 50.536
470748 98.245 -63.424 50.482
470749 98.285 -63.422 50.429
470750 98.303 -63.436 50.378
470751 98.322 -63.449 50.330
470752 98.341 -63.464 50.284
470753 98.358 -63.477 50.237
470754 98.368 -63.493 50.191
470755 98.379 -63.508 50.147
470756 98.390 -63.523 50.103
470757 98.402 -63.538 50.059
470758 98.421 -63.553 50.012
470759 98.447 -63.551 49.964
470760 98.473 -63.538 49.916
470761 98.499 -63.526 49.868
470762 98.525 -63.513 49.820
470763 98.551 -63.501 49.774
470764 98.577 -63.488 49.728
470765 98.612 -63.490 49.684
470766 98.648 -63.491 49.640
470767 98.708 -63.487 49.599
470768 98.771 -63.486 49.557
470769 98.834 -63.484 49.517
470770 98.897 -63.482 49.478
470771 98.959 -63.480 49.440
470772 99.022 -63.478 49.400
470773 99.085 -63.477 49.363
470774 99.148 -63.474 49.326
471185 87.230 -61.571 56.134
471186 87.267 -61.572 56.123
471187 87.304 -61.575 56.111
471188 87.341 -61.576 56.099
471189 87.377 -61.578 56.088
471190 87.414 -61.580 56.076
471191 87.450 -61.581 56.065
471192 87.487 -61.583 56.055
471193 87.524 -61.585 56.045
471194 87.561 -61.587 56.036
471195 87.598 -61.589 56.026
471196 87.634 -61.591 56.014
471197 87.671 -61.592 56.001
471198 87.707 -61.594 55.988
471199 87.744 -61.596 55.976
471200 87.787 -61.595 55.962
471201 87.836 -61.591 55.945
471202 87.884 -61.588 55.927
471203 87.932 -61.584 55.909
471204 87.981 -61.581 55.892
471205 88.029 -61.578 55.878
471206 88.077 -61.575 55.866
471207 88.125 -61.572 55.855
471208 88.174 -61.568 55.844
471209 88.222 -61.565 55.833
471210 88.282 -61.567 55.816
471211 88.349 -61.574 55.794
471212 88.417 -61.580 55.768
471213 88.484 -61.586 55.739
471214 88.541 -61.597 55.708
471215 88.590 -61.613 55.680
471216 88.638 -61.628 55.652
471217 88.686 -61.644 55.629
471218 88.732 -61.660 55.613
471219 88.775 -61.678 55.603
471220 88.819 -61.701 55.595
471221 88.865 -61.733 55.586
471222 88.911 -61.766 55.573
471223 88.960 -61.798 55.554
471224 89.000 -61.806 55.530
471225 89.037 -61.807 55.501
471226 89.074 -61.809 55.470
471227 89.112 -61.811 55.441
471228 89.150 -61.813 55.414
471229 89.187 -61.814 55.392
471230 89.225 -61.816 55.373
471231 89.262 -61.817 55.355
471232 89.300 -61.819 55.339
471233 89.337 -61.820 55.323
471234 89.380 -61.830 55.308
471235 89.422 -61.839 55.295
471236 89.465 -61.849 55.283
471237 89.507 -61.859 55.271
471238 89.550 -61.868 55.260
471239 89.592 -61.878 55.251
471240 89.635 -61.888 55.244
471241 89.677 -61.897 55.237
471242 89.720 -61.907 55.230
471243 89.762 -61.916 55.225
471244 89.805 -61.926 55.224
471245 89.848 -61.939 55.224
471246 89.894 -61.960 55.228
471247 89.939 -61.981 55.233
471248 89.984 -62.002 55.240
471249 90.030 -62.023 55.243
471250 90.075 -62.045 55.248
471251 90.120 -62.066 55.250
471252 90.166 -62.088 55.249
471253 90.217 -62.102 55.243
471254 90.268 -62.114 55.230
471255 90.321 -62.127 55.213
471256 90.373 -62.139 55.195
471257 90.425 -62.152 55.174
471258 90.476 -62.164 55.150
471259 90.528 -62.177 55.123
471260 90.580 -62.189 55.100
471261 90.632 -62.202 55.082
471262 90.684 -62.214 55.068
471263 90.736 -62.227 55.058
471264 90.787 -62.239 55.053
471265 90.835 -62.254 55.053
471266 90.880 -62.265 55.054
471267 90.922 -62.274 55.053
471268 90.962 -62.283 55.048
471269 91.003 -62.293 55.042
471270 91.043 -62.302 55.035
471271 91.082 -62.314 55.028
471272 91.119 -62.326 55.025
471273 91.157 -62.337 55.026
471274 91.204 -62.342 55.028
471275 91.248 -62.352 55.026
471276 91.299 -62.364 55.021
471277 91.350 -62.376 55.013
471278 91.401 -62.388 55.002
471279 91.452 -62.400 54.990
471280 91.503 -62.413 54.979
471281 91.546 -62.431 54.968
471282 91.590 -62.447 54.961
471283 91.635 -62.465 54.960
471284 91.678 -62.482 54.957
471285 91.722 -62.500 54.954
471286 91.766 -62.517 54.944
471287 91.810 -62.534 54.923
471288 91.854 -62.551 54.894
471289 91.898 -62.569 54.862
471290 91.942 -62.584 54.837
471291 91.987 -62.595 54.817
471292 92.032 -62.606 54.802
471293 92.073 -62.610 54.792
471294 92.106 -62.606 54.787
471295 92.141 -62.602 54.785
471296 92.174 -62.597 54.782
471297 92.208 -62.594 54.777
471298 92.242 -62.589 54.769
471299 92.276 -62.584 54.758
471300 92.310 -62.581 54.743
471301 92.344 -62.576 54.724
471302 92.378 -62.572 54.703
471303 92.417 -62.572 54.678
471304 92.460 -62.576 54.651
471305 92.502 -62.579 54.625
471306 92.545 -62.583 54.599
471307 92.588 -62.585 54.573
471308 92.632 -62.587 54.547
471309 92.676 -62.588 54.526
471310 92.719 -62.590 54.504
471311 92.763 -62.593 54.482
471312 92.806 -62.597 54.458
471313 92.857 -62.600 54.434
471314 92.900 -62.606 54.411
471315 92.943 -62.610 54.389
471316 92.987 -62.615 54.369
471317 93.030 -62.619 54.351
471318 93.073 -62.625 54.334
471319 93.117 -62.629 54.317
471320 93.160 -62.634 54.300
471321 93.200 -62.644 54.284
471322 93.226 -62.682 54.269
471323 93.253 -62.713 54.254
471324 93.283 -62.732 54.237
471325 93.311 -62.741 54.221
471326 93.333 -62.725 54.206
471327 93.355 -62.710 54.190
471328 93.377 -62.694 54.174
471329 93.419 -62.705 54.158
471330 93.466 -62.725 54.143
471331 93.514 -62.744 54.127
471332 93.561 -62.764 54.112
471333 93.609 -62.783 54.096
471334 93.656 -62.803 54.081
471335 93.704 -62.822 54.064
471336 93.749 -62.839 54.047
471337 93.789 -62.853 54.029
471338 93.829 -62.866 54.013
471339 93.871 -62.876 53.994
471340 93.923 -62.876 53.977
471341 93.976 -62.877 53.960
471342 94.032 -62.885 53.943
471343 94.088 -62.894 53.926
471344 94.145 -62.901 53.910
471345 94.201 -62.928 53.894
471346 94.251 -62.931 53.879
471347 94.297 -62.907 53.865
471348 94.334 -62.901 53.851
471349 94.366 -62.905 53.838
471350 94.399 -62.910 53.825
471351 94.431 -62.913 53.812
471352 94.464 -62.918 53.799
471353 94.496 -62.923 53.784
471354 94.526 -62.930 53.767
471355 94.552 -62.946 53.750
471356 94.577 -62.961 53.732
471357 94.603 -62.977 53.712
471358 94.628 -62.992 53.690
471359 94.653 -63.008 53.670
471360 94.679 -63.023 53.649
471361 94.706 -63.037 53.625
471362 94.746 -63.034 53.599
471363 94.786 -63.030 53.573
471364 94.826 -63.026 53.548
471365 94.866 -63.023 53.521
471366 94.906 -63.020 53.494
471367 94.947 -63.017 53.466
471368 94.983 -63.033 53.438
471369 95.025 -63.045 53.413
471370 95.077 -63.045 53.389
471371 95.130 -63.045 53.366
471372 95.183 -63.045 53.342
471373 95.245 -63.051 53.312
471374 95.316 -63.061 53.278
471375 95.379 -63.075 53.243
471376 95.438 -63.092 53.207
471377 95.497 -63.109 53.170
471378 95.555 -63.126 53.134
471379 95.610 -63.135 53.097
471380 95.648 -63.136 53.059
471381 95.643 -63.129 53.022
471382 95.619 -63.101 52.987
471383 95.646 -63.087 52.951
471384 95.696 -63.078 52.915
471385 95.746 -63.069 52.882
471386 95.797 -63.061 52.848
471387 95.846 -63.051 52.812
471388 95.897 -63.042 52.776
471389 95.947 -63.033 52.741
471390 95.997 -63.024 52.708
471391 96.047 -63.025 52.675
471392 96.109 -63.056 52.642
471393 96.152 -63.101 52.607
471394 96.194 -63.146 52.574
471395 96.236 -63.191 52.539
471396 96.279 -63.236 52.504
471397 96.331 -63.236 52.470
471398 96.377 -63.194 52.434
471399 96.421 -63.146 52.393
471400 96.466 -63.105 52.350
471401 96.515 -63.117 52.306
471402 96.564 -63.129 52.262
471403 96.612 -63.142 52.219
471404 96.684 -63.162 52.178
471405 96.773 -63.186 52.137
471406 96.835 -63.205 52.098
471407 96.869 -63.213 52.062
471408 96.895 -63.217 52.030
471409 96.921 -63.222 52.002
471410 96.947 -63.226 51.976
471411 96.973 -63.230 51.950
471412 96.989 -63.238 51.922
471413 97.009 -63.245 51.892
471414 97.028 -63.252 51.859
471415 97.048 -63.258 51.826
471416 97.071 -63.264 51.794
471417 97.104 -63.270 51.761
471418 97.138 -63.276 51.728
471419 97.171 -63.281 51.693
471420 97.204 -63.287 51.657
471421 97.237 -63.292 51.620
471422 97.271 -63.299 51.583
471423 97.304 -63.304 51.548
471424 97.318 -63.311 51.512
471425 97.327 -63.319 51.478
471426 97.338 -63.327 51.445
471427 97.370 -63.330 51.412
471428 97.420 -63.314 51.379
471429 97.469 -63.298 51.347
471430 97.519 -63.282 51.315
471431 97.569 -63.266 51.282
471432 97.597 -63.270 51.248
471433 97.624 -63.275 51.213
471434 97.651 -63.280 51.179
471435 97.678 -63.286 51.144
471436 97.705 -63.290 51.107
471437 97.731 -63.296 51.068
471438 97.758 -63.301 51.026
471439 97.793 -63.310 50.980
471440 97.842 -63.324 50.931
471441 97.889 -63.339 50.881
471442 97.938 -63.354 50.831
471443 97.986 -63.369 50.780
471444 98.034 -63.383 50.727
471445 98.082 -63.397 50.673
471446 98.127 -63.398 50.616
471447 98.170 -63.395 50.558
471448 98.213 -63.390 50.501
471449 98.255 -63.386 50.447
471450 98.296 -63.383 50.394
471451 98.314 -63.397 50.342
471452 98.328 -63.412 50.291
471453 98.339 -63.427 50.242
471454 98.350 -63.442 50.196
471455 98.361 -63.457 50.150
471456 98.372 -63.472 50.105
471457 98.383 -63.487 50.060
471458 98.399 -63.502 50.017
471459 98.423 -63.518 49.973
471460 98.449 -63.507 49.927
471461 98.475 -63.494 49.881
471462 98.501 -63.482 49.836
471463 98.527 -63.470 49.792
471464 98.553 -63.457 49.748
471465 98.582 -63.449 49.704
471466 98.618 -63.453 49.661
471467 98.653 -63.450 49.619
471468 98.716 -63.449 49.579
471469 98.779 -63.447 49.540
471470 98.842 -63.445 49.500
471471 98.905 -63.443 49.462
471472 98.968 -63.441 49.424
471473 99.031 -63.440 49.386
471474 99.094 -63.438 49.350
471475 99.157 -63.436 49.315
471886 87.246 -61.523 56.131
471887 87.282 -61.525 56.119
471888 87.319 -61.526 56.107
471889 87.356 -61.528 56.096
471890 87.393 -61.530 56.084
471891 87.429 -61.532 56.072
471892 87.466 -61.534 56.061
471893 87.502 -61.535 56.051
471894 87.539 -61.538 56.040
471895 87.576 -61.539 56.030
471896 87.613 -61.541 56.019
471897 87.649 -61.543 56.006
471898 87.686 -61.544 55.993
471899 87.723 -61.547 55.980
471900 87.760 -61.548 55.967
471901 87.797 -61.550 55.954
471902 87.845 -61.546 55.938
471903 87.894 -61.543 55.920
471904 87.942 -61.540 55.902
471905 87.990 -61.536 55.885
471906 88.038 -61.533 55.870
471907 88.086 -61.529 55.858
471908 88.135 -61.526 55.847
471909 88.183 -61.522 55.837
471910 88.232 -61.519 55.826
471911 88.293 -61.523 55.810
471912 88.360 -61.529 55.789
471913 88.428 -61.535 55.765
471914 88.495 -61.542 55.735
471915 88.558 -61.550 55.703
471916 88.606 -61.566 55.673
471917 88.654 -61.581 55.645
471918 88.703 -61.597 55.622
471919 88.747 -61.614 55.605
471920 88.790 -61.632 55.596
471921 88.836 -61.663 55.589
471922 88.882 -61.695 55.581
471923 88.928 -61.728 55.570
471924 88.977 -61.760 55.552
471925 89.015 -61.763 55.528
471926 89.052 -61.765 55.499
471927 89.090 -61.766 55.469
471928 89.127 -61.768 55.439
471929 89.165 -61.769 55.412
471930 89.202 -61.771 55.388
471931 89.239 -61.773 55.368
471932 89.277 -61.774 55.349
471933 89.314 -61.776 55.331
471934 89.352 -61.778 55.316
471935 89.389 -61.779 55.301
471936 89.432 -61.789 55.287
471937 89.474 -61.799 55.275
471938 89.517 -61.808 55.264
471939 89.560 -61.818 55.254
471940 89.602 -61.827 55.245
471941 89.645 -61.837 55.238
471942 89.687 -61.847 55.232
471943 89.730 -61.856 55.226
471944 89.772 -61.866 55.221
471945 89.815 -61.875 55.219
471946 89.859 -61.890 55.217
471947 89.904 -61.911 55.218
471948 89.950 -61.932 55.222
471949 89.995 -61.954 55.229
471950 90.040 -61.975 55.232
471951 90.086 -61.996 55.237
471952 90.131 -62.018 55.242
471953 90.176 -62.039 55.244
471954 90.225 -62.057 55.243
471955 90.276 -62.069 55.234
471956 90.328 -62.082 55.218
471957 90.380 -62.094 55.198
471958 90.432 -62.107 55.176
471959 90.484 -62.119 55.151
471960 90.536 -62.132 55.122
471961 90.588 -62.144 55.098
471962 90.640 -62.157 55.079
471963 90.692 -62.170 55.066
471964 90.744 -62.182 55.057
471965 90.796 -62.195 55.054
471966 90.848 -62.207 55.057
471967 90.894 -62.220 55.060
471968 90.935 -62.229 55.060
471969 90.974 -62.239 55.057
471970 91.011 -62.252 55.051
471971 91.048 -62.264 55.045
471972 91.085 -62.276 55.038
471973 91.122 -62.288 55.033
471974 91.170 -62.293 55.032
471975 91.216 -62.299 55.033
471976 91.259 -62.308 55.029
471977 91.305 -62.318 55.024
471978 91.356 -62.330 55.017
471979 91.407 -62.343 55.012
471980 91.458 -62.355 55.007
471981 91.506 -62.370 54.997
471982 91.549 -62.387 54.984
471983 91.593 -62.404 54.975
471984 91.638 -62.421 54.979
471985 91.681 -62.439 54.976
471986 91.725 -62.456 54.971
471987 91.769 -62.474 54.960
471988 91.813 -62.490 54.942
471989 91.857 -62.508 54.911
471990 91.901 -62.525 54.874
471991 91.945 -62.543 54.844
471992 91.989 -62.557 54.820
471993 92.034 -62.569 54.801
471994 92.077 -62.575 54.786
471995 92.110 -62.572 54.779
471996 92.145 -62.567 54.773
471997 92.178 -62.562 54.768
471998 92.213 -62.559 54.762
471999 92.246 -62.554 54.753
472000 92.280 -62.550 54.740
472001 92.314 -62.546 54.724
472002 92.348 -62.541 54.706
472003 92.382 -62.537 54.685
472004 92.416 -62.533 54.661
472005 92.456 -62.534 54.635
472006 92.499 -62.538 54.609
472007 92.542 -62.540 54.583
472008 92.586 -62.543 54.557
472009 92.629 -62.544 54.532
472010 92.672 -62.547 54.509
472011 92.715 -62.553 54.488
472012 92.758 -62.558 54.466
472013 92.800 -62.562 54.443
472014 92.851 -62.566 54.421
472015 92.905 -62.569 54.400
472016 92.957 -62.572 54.381
472017 93.000 -62.576 54.364
472018 93.044 -62.581 54.348
472019 93.087 -62.586 54.333
472020 93.130 -62.590 54.318
472021 93.174 -62.595 54.303
472022 93.207 -62.618 54.289
472023 93.233 -62.656 54.274
472024 93.259 -62.691 54.259
472025 93.289 -62.710 54.244
472026 93.313 -62.706 54.227
472027 93.336 -62.690 54.210
472028 93.358 -62.675 54.193
472029 93.380 -62.659 54.175
472030 93.412 -62.657 54.159
472031 93.460 -62.677 54.142
472032 93.507 -62.696 54.127
472033 93.555 -62.716 54.111
472034 93.603 -62.735 54.095
472035 93.650 -62.755 54.080
472036 93.698 -62.774 54.063
472037 93.746 -62.794 54.045
472038 93.793 -62.813 54.027
472039 93.837 -62.829 54.011
472040 93.880 -62.838 53.992
472041 93.935 -62.843 53.975
472042 93.992 -62.851 53.958
472043 94.048 -62.859 53.941
472044 94.105 -62.867 53.924
472045 94.161 -62.876 53.907
472046 94.217 -62.891 53.891
472047 94.267 -62.889 53.875
472048 94.313 -62.864 53.859
472049 94.352 -62.852 53.845
472050 94.385 -62.857 53.831
472051 94.417 -62.860 53.817
472052 94.450 -62.865 53.803
472053 94.482 -62.869 53.789
472054 94.511 -62.879 53.773
472055 94.537 -62.894 53.756
472056 94.563 -62.910 53.738
472057 94.588 -62.925 53.718
472058 94.614 -62.941 53.698
472059 94.639 -62.956 53.677
472060 94.665 -62.972 53.655
472061 94.690 -62.987 53.633
472062 94.718 -63.000 53.609
472063 94.758 -62.996 53.584
472064 94.798 -62.993 53.559
472065 94.838 -62.989 53.533
472066 94.878 -62.986 53.506
472067 94.918 -62.982 53.478
472068 94.955 -63.000 53.450
472069 94.991 -63.017 53.421
472070 95.040 -63.022 53.396
472071 95.093 -63.022 53.373
472072 95.146 -63.022 53.350
472073 95.199 -63.022 53.325
472074 95.252 -63.021 53.295
472075 95.316 -63.028 53.261
472076 95.387 -63.038 53.226
472077 95.445 -63.055 53.189
472078 95.504 -63.073 53.153
472079 95.563 -63.089 53.115
472080 95.621 -63.107 53.076
472081 95.677 -63.122 53.037
472082 95.663 -63.098 52.999
472083 95.634 -63.071 52.963
472084 95.659 -63.056 52.926
472085 95.708 -63.044 52.890
472086 95.757 -63.031 52.855
472087 95.806 -63.019 52.820
472088 95.854 -63.007 52.784
472089 95.903 -62.994 52.748
472090 95.952 -62.982 52.713
472091 96.001 -62.969 52.679
472092 96.051 -62.967 52.646
472093 96.106 -62.999 52.612
472094 96.148 -63.044 52.579
472095 96.190 -63.089 52.545
472096 96.237 -63.130 52.510
472097 96.293 -63.164 52.474
472098 96.329 -63.170 52.440
472099 96.390 -63.151 52.405
472100 96.445 -63.132 52.365
472101 96.497 -63.105 52.325
472102 96.542 -63.073 52.282
472103 96.591 -63.086 52.238
472104 96.671 -63.108 52.195
472105 96.731 -63.127 52.154
472106 96.774 -63.140 52.113
472107 96.811 -63.151 52.074
472108 96.849 -63.161 52.038
472109 96.886 -63.172 52.005
472110 96.919 -63.181 51.976
472111 96.945 -63.186 51.950
472112 96.970 -63.190 51.923
472113 96.997 -63.194 51.895
472114 96.998 -63.207 51.866
472115 97.018 -63.214 51.834
472116 97.040 -63.220 51.802
472117 97.074 -63.226 51.771
472118 97.107 -63.232 51.739
472119 97.140 -63.237 51.706
472120 97.174 -63.242 51.672
472121 97.207 -63.249 51.637
472122 97.240 -63.254 51.600
472123 97.274 -63.260 51.562
472124 97.307 -63.265 51.526
472125 97.323 -63.273 51.489
472126 97.332 -63.280 51.453
472127 97.342 -63.287 51.419
472128 97.380 -63.278 51.385
472129 97.430 -63.262 51.350
472130 97.479 -63.246 51.317
472131 97.529 -63.230 51.283
472132 97.579 -63.214 51.249
472133 97.607 -63.218 51.214
472134 97.634 -63.224 51.179
472135 97.661 -63.228 51.143
472136 97.688 -63.233 51.107
472137 97.715 -63.239 51.069
472138 97.741 -63.244 51.029
472139 97.769 -63.249 50.986
472140 97.795 -63.254 50.939
472141 97.836 -63.265 50.891
472142 97.884 -63.280 50.842
472143 97.932 -63.295 50.792
472144 97.980 -63.310 50.741
472145 98.028 -63.324 50.688
472146 98.074 -63.325 50.635
472147 98.119 -63.326 50.579
472148 98.164 -63.327 50.522
472149 98.210 -63.328 50.465
472150 98.257 -63.332 50.411
472151 98.299 -63.346 50.358
472152 98.310 -63.361 50.307
472153 98.321 -63.376 50.258
472154 98.332 -63.391 50.210
472155 98.343 -63.406 50.164
472156 98.354 -63.421 50.119
472157 98.364 -63.437 50.075
472158 98.376 -63.452 50.030
472159 98.401 -63.467 49.987
472160 98.425 -63.476 49.942
472161 98.451 -63.463 49.898
472162 98.478 -63.451 49.854
472163 98.503 -63.438 49.810
472164 98.529 -63.426 49.767
472165 98.556 -63.413 49.725
472166 98.588 -63.412 49.683
472167 98.624 -63.416 49.641
472168 98.662 -63.412 49.601
472169 98.725 -63.410 49.561
472170 98.788 -63.408 49.523
472171 98.851 -63.406 49.485
472172 98.914 -63.404 49.447
472173 98.977 -63.402 49.410
472174 99.040 -63.401 49.373
472175 99.102 -63.399 49.338
472588 87.298 -61.477 56.114
472589 87.334 -61.478 56.103
472590 87.371 -61.481 56.092
472591 87.408 -61.482 56.080
472592 87.444 -61.484 56.068
472593 87.481 -61.486 56.057
472594 87.518 -61.487 56.046
472595 87.555 -61.490 56.035
472596 87.591 -61.491 56.024
472597 87.628 -61.493 56.012
472598 87.664 -61.495 55.999
472599 87.701 -61.497 55.985
472600 87.738 -61.498 55.972
472601 87.775 -61.500 55.959
472602 87.812 -61.502 55.945
472603 87.855 -61.501 55.930
472604 87.903 -61.497 55.913
472605 87.952 -61.494 55.895
472606 87.999 -61.491 55.877
472607 88.048 -61.487 55.861
472608 88.096 -61.484 55.848
472609 88.144 -61.481 55.836
472610 88.193 -61.478 55.826
472611 88.241 -61.474 55.816
472612 88.304 -61.478 55.801
472613 88.371 -61.484 55.783
472614 88.439 -61.491 55.759
472615 88.506 -61.497 55.730
472616 88.574 -61.503 55.697
472617 88.623 -61.519 55.665
472618 88.672 -61.534 55.636
472619 88.719 -61.550 55.613
472620 88.762 -61.567 55.596
472621 88.807 -61.593 55.587
472622 88.853 -61.625 55.581
472623 88.900 -61.657 55.574
472624 88.946 -61.690 55.564
472625 88.992 -61.719 55.548
472626 89.029 -61.720 55.523
472627 89.067 -61.722 55.496
472628 89.104 -61.723 55.466
472629 89.142 -61.725 55.437
472630 89.179 -61.726 55.410
472631 89.216 -61.729 55.385
472632 89.254 -61.730 55.363
472633 89.292 -61.732 55.342
472634 89.329 -61.733 55.324
472635 89.367 -61.735 55.307
472636 89.404 -61.736 55.292
472637 89.442 -61.738 55.279
472638 89.484 -61.748 55.268
472639 89.527 -61.758 55.257
472640 89.569 -61.767 55.247
472641 89.612 -61.777 55.239
472642 89.655 -61.786 55.232
472643 89.697 -61.796 55.226
472644 89.740 -61.806 55.221
472645 89.782 -61.815 55.217
472646 89.825 -61.825 55.215
472647 89.869 -61.841 55.212
472648 89.915 -61.863 55.211
472649 89.960 -61.884 55.213
472650 90.006 -61.905 55.218
472651 90.051 -61.926 55.223
472652 90.096 -61.948 55.228
472653 90.142 -61.970 55.233
472654 90.187 -61.991 55.236
472655 90.233 -62.011 55.237
472656 90.285 -62.024 55.231
472657 90.336 -62.036 55.219
472658 90.388 -62.049 55.201
472659 90.440 -62.061 55.178
472660 90.492 -62.074 55.151
472661 90.544 -62.086 55.121
472662 90.596 -62.099 55.096
472663 90.648 -62.112 55.076
472664 90.699 -62.124 55.063
472665 90.751 -62.137 55.055
472666 90.803 -62.149 55.055
472667 90.856 -62.162 55.060
472668 90.903 -62.177 55.065
472669 90.941 -62.190 55.067
472670 90.978 -62.202 55.065
472671 91.015 -62.214 55.061
472672 91.052 -62.227 55.057
472673 91.089 -62.239 55.052
472674 91.133 -62.246 55.045
472675 91.183 -62.248 55.039
472676 91.228 -62.255 55.037
472677 91.271 -62.265 55.033
472678 91.314 -62.274 55.028
472679 91.362 -62.285 55.026
472680 91.413 -62.297 55.028
472681 91.464 -62.309 55.026
472682 91.509 -62.326 55.013
472683 91.553 -62.343 54.997
472684 91.596 -62.361 54.989
472685 91.641 -62.378 54.995
472686 91.684 -62.396 54.995
472687 91.728 -62.412 54.987
472688 91.772 -62.430 54.975
472689 91.816 -62.447 54.956
472690 91.860 -62.465 54.926
472691 91.904 -62.482 54.886
472692 91.948 -62.500 54.851
472693 91.992 -62.516 54.823
472694 92.036 -62.531 54.801
472695 92.081 -62.541 54.784
472696 92.115 -62.537 54.773
472697 92.149 -62.532 54.764
472698 92.182 -62.528 54.757
472699 92.217 -62.524 54.750
472700 92.250 -62.519 54.739
472701 92.284 -62.515 54.725
472702 92.318 -62.511 54.707
472703 92.352 -62.506 54.688
472704 92.386 -62.503 54.667
472705 92.420 -62.498 54.643
472706 92.454 -62.493 54.617
472707 92.496 -62.496 54.591
472708 92.540 -62.499 54.565
472709 92.582 -62.503 54.539
472710 92.625 -62.508 54.515
472711 92.667 -62.512 54.492
472712 92.710 -62.518 54.470
472713 92.753 -62.523 54.449
472714 92.795 -62.528 54.428
472715 92.842 -62.532 54.408
472716 92.896 -62.534 54.390
472717 92.949 -62.537 54.376
472718 93.003 -62.540 54.363
472719 93.057 -62.543 54.351
472720 93.100 -62.547 54.340
472721 93.144 -62.552 54.328
472722 93.187 -62.556 54.316
472723 93.214 -62.591 54.303
472724 93.240 -62.629 54.288
472725 93.265 -62.666 54.271
472726 93.294 -62.685 54.254
472727 93.317 -62.669 54.235
472728 93.339 -62.654 54.217
472729 93.361 -62.639 54.199
472730 93.383 -62.623 54.180
472731 93.406 -62.609 54.163
472732 93.453 -62.628 54.146
472733 93.501 -62.648 54.130
472734 93.549 -62.667 54.113
472735 93.596 -62.687 54.098
472736 93.644 -62.706 54.081
472737 93.692 -62.726 54.064
472738 93.739 -62.745 54.046
472739 93.787 -62.765 54.027
472740 93.843 -62.788 54.010
472741 93.893 -62.806 53.992
472742 93.950 -62.815 53.974
472743 94.007 -62.824 53.957
472744 94.064 -62.832 53.940
472745 94.121 -62.841 53.923
472746 94.177 -62.849 53.905
472747 94.234 -62.857 53.888
472748 94.282 -62.847 53.872
472749 94.328 -62.822 53.855
472750 94.371 -62.804 53.840
472751 94.404 -62.808 53.825
472752 94.436 -62.812 53.809
472753 94.468 -62.816 53.793
472754 94.497 -62.826 53.779
472755 94.523 -62.842 53.763
472756 94.548 -62.857 53.745
472757 94.574 -62.873 53.726
472758 94.599 -62.888 53.705
472759 94.625 -62.904 53.684
472760 94.650 -62.919 53.663
472761 94.676 -62.935 53.641
472762 94.701 -62.951 53.618
472763 94.730 -62.962 53.594
472764 94.770 -62.959 53.569
472765 94.810 -62.955 53.543
472766 94.850 -62.952 53.517
472767 94.890 -62.948 53.490
472768 94.927 -62.966 53.462
472769 94.964 -62.983 53.433
472770 95.003 -62.998 53.403
472771 95.056 -62.998 53.376
472772 95.109 -62.998 53.352
472773 95.162 -62.998 53.328
472774 95.214 -62.998 53.302
472775 95.267 -62.998 53.273
472776 95.320 -62.998 53.239
472777 95.387 -63.005 53.204
472778 95.453 -63.019 53.168
472779 95.512 -63.035 53.131
472780 95.570 -63.053 53.093
472781 95.629 -63.070 53.056
472782 95.673 -63.076 53.018
472783 95.694 -63.061 52.980
472784 95.697 -63.043 52.942
472785 95.670 -63.016 52.905
472786 95.716 -62.997 52.868
472787 95.763 -62.978 52.832
472788 95.809 -62.959 52.795
472789 95.855 -62.940 52.758
472790 95.902 -62.923 52.722
472791 95.951 -62.910 52.688
472792 96.000 -62.898 52.653
472793 96.051 -62.909 52.620
472794 96.103 -62.942 52.588
472795 96.145 -62.987 52.555
472796 96.195 -63.026 52.520
472797 96.251 -63.059 52.484
472798 96.300 -63.089 52.448
472799 96.325 -63.097 52.413
472800 96.394 -63.085 52.378
472801 96.449 -63.066 52.339
472802 96.504 -63.047 52.300
472803 96.558 -63.029 52.260
472804 96.618 -63.048 52.216
472805 96.675 -63.065 52.171
472806 96.713 -63.076 52.131
472807 96.750 -63.086 52.090
472808 96.788 -63.097 52.052
472809 96.825 -63.108 52.016
472810 96.862 -63.118 51.982
472811 96.899 -63.128 51.952
472812 96.937 -63.139 51.924
472813 96.968 -63.150 51.896
472814 96.994 -63.154 51.867
472815 97.001 -63.165 51.839
472816 97.010 -63.176 51.808
472817 97.043 -63.182 51.778
472818 97.077 -63.187 51.747
472819 97.110 -63.193 51.716
472820 97.143 -63.198 51.684
472821 97.177 -63.204 51.650
472822 97.210 -63.210 51.616
472823 97.243 -63.215 51.580
472824 97.277 -63.221 51.543
472825 97.310 -63.227 51.506
472826 97.329 -63.233 51.468
472827 97.338 -63.241 51.431
472828 97.366 -63.253 51.395
472829 97.409 -63.245 51.359
472830 97.450 -63.220 51.323
472831 97.491 -63.195 51.288
472832 97.539 -63.179 51.253
472833 97.589 -63.163 51.217
472834 97.617 -63.167 51.182
472835 97.644 -63.171 51.144
472836 97.671 -63.177 51.107
472837 97.698 -63.182 51.070
472838 97.725 -63.187 51.031
472839 97.751 -63.192 50.990
472840 97.779 -63.197 50.946
472841 97.805 -63.202 50.900
472842 97.832 -63.208 50.853
472843 97.878 -63.221 50.804
472844 97.926 -63.236 50.754
472845 97.974 -63.251 50.703
472846 98.020 -63.252 50.650
472847 98.065 -63.253 50.597
472848 98.111 -63.254 50.542
472849 98.164 -63.267 50.486
472850 98.217 -63.281 50.431
472851 98.271 -63.296 50.377
472852 98.291 -63.311 50.325
472853 98.302 -63.326 50.276
472854 98.313 -63.340 50.227
472855 98.324 -63.355 50.180
472856 98.335 -63.371 50.135
472857 98.346 -63.386 50.091
472858 98.357 -63.401 50.047
472859 98.378 -63.416 50.003
472860 98.403 -63.432 49.960
472861 98.428 -63.432 49.915
472862 98.454 -63.420 49.872
472863 98.480 -63.407 49.829
472864 98.506 -63.395 49.787
472865 98.532 -63.382 49.745
472866 98.558 -63.370 49.704
472867 98.594 -63.374 49.664
472868 98.631 -63.378 49.624
472869 98.670 -63.373 49.584
472870 98.733 -63.371 49.544
472871 98.796 -63.369 49.507
472872 98.859 -63.368 49.470
472873 98.922 -63.365 49.433
472874 98.985 -63.364 49.397
472875 99.048 -63.361 49.361
473290 87.350 -61.431 56.099
473291 87.387 -61.432 56.087
473292 87.423 -61.434 56.076
473293 87.460 -61.436 56.064
473294 87.496 -61.438 56.053
473295 87.533 -61.440 56.041
473296 87.570 -61.441 56.030
473297 87.607 -61.443 56.018
473298 87.643 -61.445 56.005
473299 87.680 -61.447 55.992
473300 87.716 -61.449 55.979
473301 87.753 -61.450 55.965
473302 87.790 -61.453 55.951
473303 87.827 -61.454 55.937
473304 87.864 -61.456 55.921
473305 87.913 -61.452 55.905
473306 87.961 -61.449 55.888
473307 88.009 -61.445 55.869
473308 88.057 -61.442 55.852
473309 88.106 -61.439 55.838
473310 88.154 -61.435 55.824
473311 88.202 -61.432 55.812
473312 88.250 -61.428 55.802
473313 88.315 -61.434 55.790
473314 88.383 -61.440 55.773
473315 88.450 -61.447 55.751
473316 88.518 -61.453 55.724
473317 88.585 -61.459 55.691
473318 88.640 -61.472 55.656
473319 88.688 -61.487 55.626
473320 88.735 -61.503 55.602
473321 88.778 -61.522 55.586
473322 88.824 -61.555 55.576
473323 88.871 -61.588 55.571
473324 88.917 -61.619 55.565
473325 88.963 -61.652 55.556
473326 89.006 -61.675 55.541
473327 89.044 -61.677 55.518
473328 89.082 -61.678 55.492
473329 89.119 -61.680 55.464
473330 89.157 -61.682 55.436
473331 89.194 -61.684 55.408
473332 89.232 -61.685 55.382
473333 89.269 -61.687 55.358
473334 89.307 -61.688 55.335
473335 89.344 -61.690 55.316
473336 89.382 -61.691 55.298
473337 89.419 -61.694 55.283
473338 89.456 -61.695 55.270
473339 89.494 -61.697 55.259
473340 89.537 -61.707 55.249
473341 89.580 -61.717 55.239
473342 89.622 -61.726 55.233
473343 89.665 -61.736 55.225
473344 89.707 -61.745 55.219
473345 89.750 -61.755 55.215
473346 89.793 -61.765 55.212
473347 89.835 -61.774 55.211
473348 89.880 -61.793 55.209
473349 89.925 -61.814 55.207
473350 89.970 -61.835 55.206
473351 90.016 -61.857 55.209
473352 90.061 -61.878 55.214
473353 90.107 -61.900 55.218
473354 90.152 -61.921 55.222
473355 90.197 -61.942 55.227
473356 90.243 -61.964 55.229
473357 90.292 -61.979 55.226
473358 90.344 -61.992 55.216
473359 90.396 -62.004 55.200
473360 90.448 -62.017 55.177
473361 90.500 -62.029 55.149
473362 90.552 -62.042 55.118
473363 90.604 -62.055 55.093
473364 90.656 -62.067 55.074
473365 90.708 -62.080 55.061
473366 90.759 -62.092 55.054
473367 90.812 -62.106 55.053
473368 90.865 -62.120 55.058
473369 90.908 -62.137 55.064
473370 90.945 -62.153 55.070
473371 90.982 -62.165 55.073
473372 91.019 -62.177 55.074
473373 91.056 -62.189 55.074
473374 91.096 -62.199 55.070
473375 91.146 -62.202 55.061
473376 91.196 -62.204 55.047
473377 91.239 -62.212 55.040
473378 91.283 -62.222 55.036
473379 91.325 -62.231 55.035
473380 91.369 -62.240 55.038
473381 91.419 -62.252 55.048
473382 91.468 -62.265 55.043
473383 91.512 -62.283 55.027
473384 91.556 -62.300 55.010
473385 91.599 -62.318 55.001
473386 91.643 -62.335 55.008
473387 91.687 -62.352 55.009
473388 91.731 -62.369 55.002
473389 91.775 -62.387 54.988
473390 91.819 -62.404 54.968
473391 91.863 -62.421 54.936
473392 91.907 -62.439 54.895
473393 91.951 -62.456 54.858
473394 91.995 -62.473 54.827
473395 92.038 -62.490 54.802
473396 92.083 -62.504 54.783
473397 92.119 -62.502 54.770
473398 92.153 -62.497 54.758
473399 92.187 -62.493 54.747
473400 92.221 -62.489 54.735
473401 92.254 -62.485 54.723
473402 92.289 -62.481 54.706
473403 92.322 -62.476 54.687
473404 92.356 -62.472 54.666
473405 92.390 -62.468 54.643
473406 92.424 -62.460 54.619
473407 92.458 -62.453 54.593
473408 92.499 -62.458 54.568
473409 92.542 -62.462 54.543
473410 92.584 -62.468 54.518
473411 92.628 -62.473 54.494
473412 92.670 -62.478 54.472
473413 92.713 -62.483 54.452
473414 92.756 -62.488 54.432
473415 92.799 -62.493 54.413
473416 92.840 -62.498 54.396
473417 92.893 -62.501 54.383
473418 92.944 -62.503 54.372
473419 92.997 -62.506 54.364
473420 93.049 -62.509 54.358
473421 93.102 -62.511 54.350
473422 93.156 -62.514 54.341
473423 93.196 -62.527 54.330
473424 93.221 -62.565 54.317
473425 93.247 -62.602 54.301
473426 93.272 -62.640 54.284
473427 93.297 -62.650 54.265
473428 93.319 -62.634 54.245
473429 93.341 -62.619 54.225
473430 93.363 -62.603 54.205
473431 93.386 -62.587 54.186
473432 93.408 -62.572 54.170
473433 93.447 -62.580 54.154
473434 93.495 -62.600 54.138
473435 93.543 -62.619 54.122
473436 93.590 -62.638 54.106
473437 93.641 -62.659 54.089
473438 93.699 -62.684 54.072
473439 93.757 -62.708 54.053
473440 93.815 -62.732 54.034
473441 93.855 -62.757 54.015
473442 93.905 -62.771 53.996
473443 93.962 -62.780 53.978
473444 94.019 -62.789 53.960
473445 94.076 -62.798 53.942
473446 94.133 -62.807 53.924
473447 94.190 -62.816 53.905
473448 94.246 -62.823 53.887
473449 94.297 -62.805 53.870
473450 94.344 -62.781 53.853
473451 94.389 -62.756 53.837
473452 94.422 -62.759 53.820
473453 94.455 -62.763 53.802
473454 94.483 -62.775 53.785
473455 94.508 -62.791 53.768
473456 94.534 -62.806 53.753
473457 94.559 -62.822 53.735
473458 94.585 -62.837 53.715
473459 94.610 -62.853 53.693
473460 94.636 -62.868 53.671
473461 94.662 -62.884 53.650
473462 94.687 -62.899 53.627
473463 94.712 -62.915 53.603
473464 94.742 -62.925 53.579
473465 94.782 -62.921 53.554
473466 94.822 -62.918 53.527
473467 94.862 -62.915 53.501
473468 94.899 -62.932 53.473
473469 94.935 -62.949 53.444
473470 94.972 -62.967 53.414
473471 95.019 -62.973 53.383
473472 95.071 -62.973 53.354
473473 95.124 -62.973 53.328
473474 95.177 -62.973 53.303
473475 95.230 -62.973 53.276
473476 95.283 -62.973 53.247
473477 95.336 -62.973 53.214
473478 95.389 -62.973 53.180
473479 95.459 -62.982 53.144
473480 95.519 -62.999 53.108
473481 95.578 -63.017 53.071
473482 95.636 -63.033 53.034
473483 95.669 -63.029 52.997
473484 95.690 -63.015 52.959
473485 95.710 -63.001 52.921
473486 95.732 -62.988 52.884
473487 95.731 -62.963 52.848
473488 95.777 -62.945 52.811
473489 95.823 -62.926 52.774
473490 95.870 -62.907 52.736
473491 95.915 -62.888 52.699
473492 95.960 -62.873 52.664
473493 96.005 -62.857 52.630
473494 96.052 -62.851 52.596
473495 96.100 -62.885 52.563
473496 96.154 -62.920 52.530
473497 96.209 -62.954 52.495
473498 96.264 -62.988 52.459
473499 96.307 -63.011 52.422
473500 96.331 -63.020 52.386
473501 96.385 -63.017 52.350
473502 96.453 -63.000 52.313
473503 96.508 -62.981 52.274
473504 96.566 -62.986 52.233
473505 96.614 -63.001 52.191
473506 96.651 -63.011 52.148
473507 96.688 -63.022 52.108
473508 96.726 -63.032 52.069
473509 96.763 -63.043 52.031
473510 96.801 -63.054 51.995
473511 96.838 -63.064 51.961
473512 96.875 -63.074 51.930
473513 96.912 -63.085 51.900
473514 96.950 -63.095 51.870
473515 96.987 -63.106 51.841
473516 97.018 -63.117 51.812
473517 97.033 -63.128 51.782
473518 97.058 -63.137 51.752
473519 97.084 -63.147 51.722
473520 97.113 -63.155 51.690
473521 97.146 -63.160 51.660
473522 97.180 -63.165 51.628
473523 97.213 -63.171 51.595
473524 97.246 -63.177 51.560
473525 97.280 -63.183 51.525
473526 97.313 -63.188 51.488
473527 97.334 -63.195 51.450
473528 97.353 -63.205 51.411
473529 97.399 -63.220 51.372
473530 97.444 -63.218 51.334
473531 97.485 -63.193 51.297
473532 97.525 -63.168 51.260
473533 97.566 -63.143 51.223
473534 97.600 -63.120 51.187
473535 97.626 -63.120 51.149
473536 97.653 -63.122 51.110
473537 97.681 -63.125 51.071
473538 97.708 -63.130 51.032
473539 97.735 -63.135 50.993
473540 97.761 -63.140 50.952
473541 97.789 -63.145 50.906
473542 97.815 -63.151 50.862
473543 97.842 -63.155 50.816
473544 97.872 -63.162 50.769
473545 97.921 -63.177 50.718
473546 97.966 -63.179 50.666
473547 98.017 -63.189 50.613
473548 98.071 -63.203 50.559
473549 98.124 -63.217 50.505
473550 98.177 -63.231 50.450
473551 98.231 -63.245 50.397
473552 98.273 -63.260 50.345
473553 98.284 -63.275 50.295
473554 98.295 -63.290 50.247
473555 98.306 -63.305 50.200
473556 98.317 -63.320 50.154
473557 98.327 -63.335 50.109
473558 98.339 -63.350 50.065
473559 98.356 -63.365 50.022
473560 98.380 -63.381 49.978
473561 98.404 -63.396 49.935
473562 98.430 -63.389 49.891
473563 98.456 -63.376 49.849
473564 98.482 -63.364 49.808
473565 98.508 -63.351 49.766
473566 98.534 -63.339 49.725
473567 98.564 -63.333 49.684
473568 98.600 -63.336 49.646
473569 98.637 -63.341 49.607
473570 98.679 -63.334 49.568
473571 98.742 -63.332 49.529
473572 98.805 -63.330 49.493
473573 98.868 -63.328 49.456
473574 98.931 -63.327 49.420
473575 98.994 -63.325 49.385
473991 87.365 -61.383 56.095
473992 87.402 -61.384 56.083
473993 87.438 -61.387 56.071
473994 87.475 -61.388 56.060
473995 87.511 -61.390 56.048
473996 87.548 -61.392 56.036
473997 87.585 -61.393 56.025
473998 87.622 -61.396 56.012
473999 87.658 -61.397 56.000
474000 87.695 -61.399 55.986
474001 87.732 -61.401 55.972
474002 87.768 -61.403 55.957
474003 87.805 -61.404 55.943
474004 87.842 -61.406 55.927
474005 87.879 -61.408 55.912
474006 87.922 -61.407 55.896
474007 87.970 -61.403 55.879
474008 88.019 -61.400 55.860
474009 88.067 -61.396 55.843
474010 88.115 -61.393 55.827
474011 88.164 -61.390 55.812
474012 88.211 -61.387 55.798
474013 88.260 -61.384 55.787
474014 88.326 -61.389 55.775
474015 88.394 -61.396 55.760
474016 88.461 -61.402 55.740
474017 88.528 -61.408 55.715
474018 88.596 -61.415 55.683
474019 88.657 -61.424 55.648
474020 88.705 -61.439 55.615
474021 88.750 -61.456 55.590
474022 88.795 -61.484 55.571
474023 88.841 -61.517 55.562
474024 88.888 -61.549 55.556
474025 88.933 -61.581 55.550
474026 88.978 -61.606 55.543
474027 89.022 -61.627 55.531
474028 89.062 -61.638 55.512
474029 89.098 -61.637 55.489
474030 89.134 -61.638 55.463
474031 89.171 -61.639 55.436
474032 89.209 -61.641 55.406
474033 89.246 -61.642 55.379
474034 89.284 -61.644 55.353
474035 89.321 -61.646 55.329
474036 89.359 -61.647 55.308
474037 89.396 -61.649 55.289
474038 89.434 -61.650 55.274
474039 89.471 -61.652 55.261
474040 89.509 -61.653 55.250
474041 89.547 -61.657 55.240
474042 89.590 -61.666 55.231
474043 89.632 -61.676 55.225
474044 89.675 -61.685 55.218
474045 89.718 -61.695 55.213
474046 89.760 -61.705 55.210
474047 89.803 -61.714 55.209
474048 89.845 -61.724 55.211
474049 89.890 -61.744 55.210
474050 89.936 -61.766 55.208
474051 89.981 -61.787 55.204
474052 90.026 -61.808 55.201
474053 90.072 -61.829 55.202
474054 90.117 -61.851 55.206
474055 90.162 -61.873 55.210
474056 90.207 -61.894 55.215
474057 90.253 -61.915 55.219
474058 90.300 -61.934 55.217
474059 90.352 -61.946 55.210
474060 90.404 -61.959 55.196
474061 90.456 -61.971 55.173
474062 90.508 -61.984 55.146
474063 90.559 -61.997 55.114
474064 90.612 -62.009 55.089
474065 90.664 -62.022 55.070
474066 90.716 -62.035 55.058
474067 90.769 -62.050 55.052
474068 90.822 -62.065 55.051
474069 90.875 -62.080 55.056
474070 90.912 -62.096 55.064
474071 90.949 -62.114 55.072
474072 90.986 -62.127 55.079
474073 91.023 -62.139 55.084
474074 91.060 -62.152 55.087
474075 91.109 -62.155 55.085
474076 91.159 -62.157 55.076
474077 91.208 -62.160 55.060
474078 91.251 -62.169 55.050
474079 91.294 -62.179 55.047
474080 91.338 -62.188 55.050
474081 91.380 -62.197 55.057
474082 91.424 -62.207 55.063
474083 91.470 -62.222 55.056
474084 91.515 -62.239 55.038
474085 91.559 -62.257 55.020
474086 91.602 -62.274 55.013
474087 91.646 -62.291 55.017
474088 91.690 -62.308 55.020
474089 91.734 -62.326 55.014
474090 91.778 -62.343 55.000
474091 91.822 -62.361 54.976
474092 91.866 -62.378 54.942
474093 91.910 -62.395 54.902
474094 91.954 -62.412 54.864
474095 91.998 -62.430 54.830
474096 92.041 -62.447 54.803
474097 92.086 -62.465 54.782
474098 92.123 -62.467 54.767
474099 92.157 -62.459 54.752
474100 92.191 -62.451 54.736
474101 92.225 -62.442 54.721
474102 92.260 -62.434 54.705
474103 92.293 -62.425 54.686
474104 92.328 -62.417 54.665
474105 92.362 -62.409 54.643
474106 92.393 -62.406 54.620
474107 92.423 -62.409 54.596
474108 92.462 -62.414 54.572
474109 92.511 -62.421 54.549
474110 92.559 -62.428 54.526
474111 92.602 -62.433 54.503
474112 92.645 -62.438 54.482
474113 92.688 -62.443 54.462
474114 92.730 -62.448 54.443
474115 92.774 -62.453 54.425
474116 92.816 -62.459 54.407
474117 92.850 -62.465 54.391
474118 92.902 -62.468 54.380
474119 92.954 -62.470 54.374
474120 93.006 -62.473 54.370
474121 93.058 -62.475 54.369
474122 93.111 -62.478 54.363
474123 93.159 -62.483 54.355
474124 93.199 -62.494 54.344
474125 93.225 -62.531 54.331
474126 93.251 -62.569 54.316
474127 93.276 -62.607 54.298
474128 93.299 -62.607 54.277
474129 93.322 -62.593 54.256
474130 93.344 -62.579 54.234
474131 93.366 -62.565 54.212
474132 93.389 -62.551 54.193
474133 93.411 -62.537 54.178
474134 93.441 -62.532 54.163
474135 93.497 -62.555 54.148
474136 93.555 -62.579 54.133
474137 93.613 -62.603 54.118
474138 93.671 -62.628 54.101
474139 93.729 -62.652 54.082
474140 93.787 -62.676 54.063
474141 93.826 -62.700 54.043
474142 93.863 -62.725 54.022
474143 93.916 -62.737 54.003
474144 93.974 -62.746 53.984
474145 94.031 -62.755 53.965
474146 94.087 -62.764 53.946
474147 94.144 -62.773 53.927
474148 94.201 -62.782 53.907
474149 94.254 -62.772 53.888
474150 94.306 -62.753 53.870
474151 94.357 -62.736 53.852
474152 94.405 -62.714 53.834
474153 94.441 -62.710 53.814
474154 94.468 -62.723 53.794
474155 94.494 -62.738 53.775
474156 94.520 -62.754 53.756
474157 94.545 -62.769 53.738
474158 94.570 -62.785 53.720
474159 94.596 -62.800 53.700
474160 94.621 -62.816 53.678
474161 94.647 -62.832 53.656
474162 94.673 -62.847 53.634
474163 94.698 -62.863 53.610
474164 94.723 -62.879 53.586
474165 94.754 -62.888 53.561
474166 94.794 -62.884 53.535
474167 94.834 -62.881 53.508
474168 94.871 -62.898 53.481
474169 94.907 -62.916 53.452
474170 94.944 -62.933 53.422
474171 94.981 -62.949 53.392
474172 95.034 -62.949 53.359
474173 95.087 -62.949 53.330
474174 95.140 -62.949 53.302
474175 95.193 -62.949 53.275
474176 95.246 -62.949 53.248
474177 95.299 -62.949 53.219
474178 95.352 -62.949 53.186
474179 95.405 -62.949 53.152
474180 95.459 -62.950 53.117
474181 95.526 -62.962 53.082
474182 95.585 -62.979 53.047
474183 95.643 -62.997 53.011
474184 95.664 -62.983 52.975
474185 95.685 -62.969 52.938
474186 95.706 -62.955 52.901
474187 95.727 -62.941 52.865
474188 95.748 -62.929 52.828
474189 95.790 -62.915 52.791
474190 95.835 -62.899 52.754
474191 95.880 -62.883 52.715
474192 95.924 -62.867 52.677
474193 95.969 -62.851 52.642
474194 96.014 -62.835 52.607
474195 96.059 -62.820 52.573
474196 96.112 -62.819 52.539
474197 96.167 -62.849 52.506
474198 96.222 -62.883 52.470
474199 96.277 -62.916 52.434
474200 96.313 -62.934 52.397
474201 96.338 -62.943 52.361
474202 96.372 -62.948 52.324
474203 96.453 -62.933 52.287
474204 96.514 -62.925 52.249
474205 96.552 -62.936 52.208
474206 96.590 -62.947 52.167
474207 96.627 -62.957 52.126
474208 96.664 -62.968 52.087
474209 96.702 -62.979 52.049
474210 96.739 -62.989 52.011
474211 96.777 -63.000 51.975
474212 96.816 -63.013 51.941
474213 96.855 -63.025 51.909
474214 96.894 -63.037 51.877
474215 96.933 -63.049 51.847
474216 96.972 -63.062 51.817
474217 97.011 -63.074 51.787
474218 97.043 -63.092 51.757
474219 97.075 -63.096 51.727
474220 97.105 -63.101 51.697
474221 97.136 -63.105 51.666
474222 97.167 -63.110 51.636
474223 97.197 -63.119 51.605
474224 97.226 -63.128 51.572
474225 97.251 -63.137 51.539
474226 97.283 -63.144 51.506
474227 97.316 -63.149 51.471
474228 97.339 -63.156 51.434
474229 97.386 -63.172 51.394
474230 97.434 -63.189 51.353
474231 97.478 -63.191 51.313
474232 97.519 -63.166 51.274
474233 97.560 -63.141 51.235
474234 97.597 -63.117 51.197
474235 97.604 -63.100 51.159
474236 97.626 -63.097 51.121
474237 97.654 -63.099 51.080
474238 97.681 -63.101 51.038
474239 97.709 -63.104 50.997
474240 97.737 -63.106 50.955
474241 97.765 -63.108 50.913
474242 97.793 -63.111 50.869
474243 97.820 -63.113 50.825
474244 97.848 -63.116 50.780
474245 97.876 -63.117 50.733
474246 97.924 -63.124 50.684
474247 97.977 -63.138 50.632
474248 98.031 -63.152 50.578
474249 98.084 -63.167 50.525
474250 98.138 -63.181 50.472
474251 98.192 -63.195 50.419
474252 98.245 -63.209 50.368
474253 98.266 -63.224 50.318
474254 98.277 -63.239 50.270
474255 98.287 -63.254 50.223
474256 98.298 -63.269 50.176
474257 98.309 -63.284 50.130
474258 98.320 -63.299 50.087
474259 98.334 -63.314 50.044
474260 98.358 -63.330 50.000
474261 98.382 -63.346 49.957
474262 98.407 -63.357 49.915
474263 98.432 -63.345 49.872
474264 98.458 -63.332 49.830
474265 98.485 -63.320 49.789
474266 98.510 -63.307 49.748
474267 98.536 -63.295 49.708
474268 98.570 -63.295 49.668
474269 98.606 -63.299 49.629
474270 98.643 -63.304 49.592
474271 98.688 -63.295 49.554
474272 98.750 -63.293 49.517
474273 98.813 -63.292 49.481
474274 98.876 -63.289 49.445
474275 98.939 -63.288 49.410
474276 99.002 -63.286 49.375
474692 87.380 -61.335 56.089
474693 87.417 -61.337 56.078
474694 87.454 -61.338 56.066
474695 87.490 -61.340 56.055
474696 87.527 -61.342 56.043
474697 87.563 -61.344 56.030
474698 87.600 -61.346 56.019
474699 87.637 -61.347 56.007
474700 87.674 -61.349 55.994
474701 87.710 -61.351 55.980
474702 87.747 -61.353 55.965
474703 87.784 -61.355 55.950
474704 87.821 -61.356 55.935
474705 87.857 -61.359 55.920
474706 87.894 -61.360 55.904
474707 87.931 -61.362 55.888
474708 87.980 -61.358 55.870
474709 88.028 -61.355 55.851
474710 88.076 -61.351 55.834
474711 88.125 -61.348 55.817
474712 88.173 -61.345 55.801
474713 88.221 -61.341 55.785
474714 88.270 -61.338 55.771
474715 88.337 -61.344 55.759
474716 88.405 -61.351 55.744
474717 88.472 -61.357 55.725
474718 88.539 -61.364 55.701
474719 88.607 -61.370 55.672
474720 88.673 -61.377 55.638
474721 88.721 -61.392 55.604
474722 88.766 -61.415 55.577
474723 88.812 -61.447 55.556
474724 88.858 -61.479 55.544
474725 88.903 -61.505 55.536
474726 88.948 -61.531 55.530
474727 88.992 -61.556 55.522
474728 89.036 -61.574 55.510
474729 89.079 -61.592 55.492
474730 89.119 -61.603 55.471
474731 89.154 -61.601 55.448
474732 89.189 -61.600 55.425
474733 89.224 -61.598 55.399
474734 89.261 -61.600 55.372
474735 89.299 -61.601 55.346
474736 89.336 -61.603 55.322
474737 89.374 -61.604 55.300
474738 89.411 -61.606 55.280
474739 89.449 -61.607 55.265
474740 89.486 -61.609 55.252
474741 89.524 -61.611 55.240
474742 89.561 -61.613 55.230
474743 89.600 -61.616 55.222
474744 89.642 -61.625 55.216
474745 89.685 -61.635 55.211
474746 89.727 -61.644 55.207
474747 89.770 -61.654 55.204
474748 89.812 -61.664 55.204
474749 89.856 -61.675 55.208
474750 89.901 -61.696 55.211
474751 89.946 -61.717 55.210
474752 89.991 -61.738 55.206
474753 90.037 -61.760 55.200
474754 90.082 -61.781 55.198
474755 90.127 -61.803 55.197
474756 90.173 -61.824 55.198
474757 90.218 -61.845 55.201
474758 90.263 -61.867 55.204
474759 90.309 -61.888 55.204
474760 90.360 -61.901 55.198
474761 90.412 -61.914 55.186
474762 90.464 -61.926 55.165
474763 90.516 -61.939 55.138
474764 90.568 -61.951 55.108
474765 90.620 -61.964 55.083
474766 90.673 -61.979 55.066
474767 90.726 -61.994 55.054
474768 90.779 -62.009 55.050
474769 90.831 -62.023 55.050
474770 90.880 -62.039 55.055
474771 90.917 -62.056 55.063
474772 90.953 -62.073 55.073
474773 90.990 -62.090 55.084
474774 91.027 -62.102 55.092
474775 91.072 -62.108 55.097
474776 91.122 -62.110 55.097
474777 91.172 -62.112 55.089
474778 91.220 -62.117 55.072
474779 91.263 -62.126 55.061
474780 91.306 -62.136 55.058
474781 91.349 -62.145 55.063
474782 91.392 -62.154 55.071
474783 91.435 -62.164 55.073
474784 91.477 -62.177 55.062
474785 91.521 -62.193 55.044
474786 91.566 -62.211 55.029
474787 91.609 -62.228 55.027
474788 91.654 -62.246 55.030
474789 91.698 -62.262 55.029
474790 91.742 -62.280 55.021
474791 91.786 -62.297 55.005
474792 91.830 -62.315 54.980
474793 91.874 -62.332 54.945
474794 91.918 -62.349 54.905
474795 91.962 -62.366 54.866
474796 92.006 -62.384 54.832
474797 92.047 -62.401 54.804
474798 92.085 -62.420 54.779
474799 92.120 -62.418 54.760
474800 92.156 -62.406 54.744
474801 92.193 -62.392 54.726
474802 92.229 -62.377 54.707
474803 92.265 -62.365 54.688
474804 92.299 -62.357 54.668
474805 92.329 -62.359 54.645
474806 92.358 -62.362 54.622
474807 92.387 -62.365 54.599
474808 92.425 -62.370 54.576
474809 92.474 -62.377 54.555
474810 92.523 -62.384 54.534
474811 92.572 -62.392 54.513
474812 92.619 -62.399 54.492
474813 92.662 -62.403 54.474
474814 92.705 -62.409 54.456
474815 92.747 -62.414 54.438
474816 92.791 -62.418 54.420
474817 92.823 -62.426 54.402
474818 92.860 -62.432 54.389
474819 92.912 -62.434 54.380
474820 92.964 -62.437 54.376
474821 93.016 -62.440 54.375
474822 93.068 -62.442 54.375
474823 93.114 -62.449 54.370
474824 93.154 -62.461 54.362
474825 93.194 -62.472 54.350
474826 93.227 -62.496 54.337
474827 93.253 -62.534 54.321
474828 93.278 -62.572 54.304
474829 93.300 -62.560 54.284
474830 93.323 -62.547 54.262
474831 93.345 -62.532 54.240
474832 93.367 -62.518 54.219
474833 93.390 -62.504 54.203
474834 93.426 -62.498 54.188
474835 93.469 -62.499 54.174
474836 93.527 -62.523 54.160
474837 93.585 -62.547 54.147
474838 93.643 -62.572 54.132
474839 93.701 -62.596 54.114
474840 93.759 -62.620 54.095
474841 93.797 -62.644 54.076
474842 93.834 -62.669 54.055
474843 93.871 -62.693 54.034
474844 93.928 -62.702 54.014
474845 93.985 -62.711 53.994
474846 94.042 -62.720 53.973
474847 94.099 -62.729 53.953
474848 94.156 -62.738 53.931
474849 94.211 -62.738 53.910
474850 94.262 -62.720 53.890
474851 94.314 -62.703 53.870
474852 94.365 -62.684 53.850
474853 94.416 -62.666 53.830
474854 94.454 -62.667 53.808
474855 94.479 -62.685 53.786
474856 94.505 -62.702 53.764
474857 94.531 -62.718 53.743
474858 94.556 -62.733 53.723
474859 94.581 -62.749 53.704
474860 94.607 -62.764 53.683
474861 94.632 -62.780 53.661
474862 94.658 -62.795 53.637
474863 94.683 -62.811 53.613
474864 94.709 -62.826 53.589
474865 94.735 -62.842 53.564
474866 94.766 -62.850 53.538
474867 94.806 -62.847 53.511
474868 94.842 -62.864 53.483
474869 94.879 -62.882 53.455
474870 94.916 -62.899 53.426
474871 94.953 -62.916 53.395
474872 94.997 -62.925 53.365
474873 95.050 -62.925 53.333
474874 95.103 -62.925 53.302
474875 95.156 -62.925 53.272
474876 95.209 -62.925 53.244
474877 95.262 -62.925 53.217
474878 95.315 -62.925 53.189
474879 95.368 -62.925 53.157
474880 95.420 -62.925 53.124
474881 95.473 -62.925 53.090
474882 95.530 -62.927 53.055
474883 95.592 -62.943 53.020
474884 95.639 -62.951 52.987
474885 95.660 -62.937 52.952
474886 95.681 -62.923 52.916
474887 95.703 -62.912 52.880
474888 95.731 -62.910 52.845
474889 95.759 -62.908 52.808
474890 95.801 -62.895 52.771
474891 95.846 -62.879 52.733
474892 95.891 -62.863 52.695
474893 95.936 -62.847 52.659
474894 95.980 -62.832 52.622
474895 96.025 -62.816 52.586
474896 96.078 -62.814 52.551
474897 96.130 -62.813 52.517
474898 96.183 -62.813 52.481
474899 96.236 -62.812 52.446
474900 96.291 -62.845 52.410
474901 96.319 -62.857 52.373
474902 96.344 -62.866 52.337
474903 96.369 -62.874 52.300
474904 96.452 -62.879 52.262
474905 96.508 -62.890 52.224
474906 96.546 -62.903 52.185
474907 96.585 -62.915 52.144
474908 96.624 -62.927 52.105
474909 96.663 -62.939 52.067
474910 96.702 -62.952 52.029
474911 96.741 -62.964 51.992
474912 96.780 -62.976 51.956
474913 96.819 -62.988 51.921
474914 96.858 -63.001 51.888
474915 96.897 -63.014 51.856
474916 96.936 -63.026 51.824
474917 96.972 -63.040 51.793
474918 97.005 -63.057 51.763
474919 97.038 -63.076 51.733
474920 97.070 -63.092 51.703
474921 97.101 -63.097 51.672
474922 97.132 -63.101 51.642
474923 97.163 -63.106 51.612
474924 97.193 -63.111 51.581
474925 97.225 -63.115 51.549
474926 97.255 -63.120 51.517
474927 97.286 -63.124 51.485
474928 97.320 -63.128 51.451
474929 97.372 -63.128 51.415
474930 97.420 -63.140 51.375
474931 97.467 -63.156 51.334
474932 97.513 -63.164 51.293
474933 97.554 -63.139 51.252
474934 97.594 -63.114 51.211
474935 97.601 -63.097 51.172
474936 97.608 -63.080 51.133
474937 97.626 -63.073 51.093
474938 97.654 -63.076 51.051
474939 97.681 -63.079 51.007
474940 97.710 -63.081 50.964
474941 97.737 -63.083 50.921
474942 97.765 -63.086 50.878
474943 97.793 -63.088 50.834
474944 97.820 -63.090 50.790
474945 97.851 -63.086 50.745
474946 97.884 -63.074 50.698
474947 97.938 -63.088 50.650
474948 97.991 -63.102 50.599
474949 98.044 -63.117 50.546
474950 98.098 -63.130 50.494
474951 98.152 -63.145 50.443
474952 98.206 -63.159 50.391
474953 98.247 -63.173 50.342
474954 98.258 -63.189 50.294
474955 98.269 -63.204 50.247
474956 98.280 -63.218 50.201
474957 98.291 -63.233 50.155
474958 98.302 -63.249 50.110
474959 98.313 -63.264 50.067
474960 98.336 -63.279 50.024
474961 98.360 -63.295 49.982
474962 98.384 -63.311 49.939
474963 98.409 -63.314 49.896
474964 98.435 -63.301 49.854
474965 98.461 -63.289 49.813
474966 98.487 -63.276 49.772
474967 98.513 -63.264 49.732
474968 98.540 -63.253 49.693
474969 98.576 -63.258 49.654
474970 98.613 -63.262 49.616
474971 98.649 -63.266 49.579
474972 98.696 -63.256 49.542
474973 98.759 -63.255 49.506
474974 98.822 -63.252 49.470
474975 98.885 -63.251 49.435
474976 98.948 -63.249 49.400
475394 87.432 -61.289 56.071
475395 87.469 -61.290 56.061
475396 87.505 -61.293 56.049
475397 87.542 -61.294 56.037
475398 87.579 -61.296 56.024
475399 87.616 -61.298 56.012
475400 87.652 -61.299 56.000
475401 87.689 -61.302 55.987
475402 87.725 -61.303 55.973
475403 87.762 -61.305 55.958
475404 87.799 -61.307 55.942
475405 87.836 -61.309 55.928
475406 87.872 -61.310 55.912
475407 87.909 -61.312 55.896
475408 87.945 -61.314 55.879
475409 87.989 -61.313 55.861
475410 88.038 -61.309 55.842
475411 88.086 -61.306 55.824
475412 88.134 -61.302 55.807
475413 88.182 -61.299 55.789
475414 88.231 -61.296 55.771
475415 88.281 -61.293 55.755
475416 88.348 -61.300 55.739
475417 88.415 -61.306 55.722
475418 88.483 -61.312 55.704
475419 88.550 -61.319 55.683
475420 88.618 -61.325 55.657
475421 88.685 -61.332 55.625
475422 88.737 -61.345 55.592
475423 88.783 -61.376 55.563
475424 88.828 -61.403 55.540
475425 88.873 -61.429 55.524
475426 88.917 -61.454 55.514
475427 88.962 -61.480 55.506
475428 89.006 -61.503 55.498
475429 89.049 -61.521 55.486
475430 89.092 -61.539 55.468
475431 89.135 -61.557 55.449
475432 89.175 -61.567 55.429
475433 89.210 -61.566 55.408
475434 89.245 -61.563 55.383
475435 89.280 -61.562 55.359
475436 89.315 -61.560 55.335
475437 89.351 -61.559 55.312
475438 89.388 -61.561 55.290
475439 89.426 -61.563 55.271
475440 89.463 -61.565 55.255
475441 89.501 -61.566 55.242
475442 89.538 -61.568 55.231
475443 89.576 -61.569 55.222
475444 89.613 -61.571 55.213
475445 89.652 -61.575 55.207
475446 89.695 -61.584 55.203
475447 89.737 -61.594 55.200
475448 89.780 -61.603 55.198
475449 89.822 -61.613 55.198
475450 89.866 -61.626 55.204
475451 89.911 -61.647 55.209
475452 89.957 -61.669 55.212
475453 90.002 -61.690 55.210
475454 90.047 -61.711 55.203
475455 90.093 -61.733 55.197
475456 90.138 -61.754 55.192
475457 90.183 -61.776 55.189
475458 90.229 -61.797 55.191
475459 90.274 -61.818 55.192
475460 90.319 -61.839 55.191
475461 90.368 -61.856 55.185
475462 90.420 -61.869 55.172
475463 90.472 -61.881 55.152
475464 90.524 -61.894 55.127
475465 90.577 -61.908 55.099
475466 90.629 -61.923 55.076
475467 90.682 -61.938 55.061
475468 90.735 -61.953 55.050
475469 90.788 -61.967 55.046
475470 90.841 -61.983 55.046
475471 90.885 -61.998 55.051
475472 90.921 -62.016 55.060
475473 90.957 -62.033 55.072
475474 90.994 -62.050 55.085
475475 91.035 -62.061 55.096
475476 91.085 -62.064 55.103
475477 91.135 -62.066 55.104
475478 91.185 -62.067 55.099
475479 91.232 -62.073 55.081
475480 91.275 -62.083 55.071
475481 91.318 -62.092 55.070
475482 91.361 -62.102 55.076
475483 91.403 -62.111 55.083
475484 91.445 -62.120 55.079
475485 91.487 -62.130 55.066
475486 91.530 -62.147 55.049
475487 91.575 -62.164 55.038
475488 91.618 -62.181 55.041
475489 91.663 -62.199 55.041
475490 91.707 -62.216 55.036
475491 91.750 -62.233 55.024
475492 91.795 -62.250 55.006
475493 91.839 -62.268 54.981
475494 91.883 -62.285 54.946
475495 91.927 -62.302 54.907
475496 91.968 -62.320 54.869
475497 92.006 -62.338 54.833
475498 92.044 -62.356 54.803
475499 92.081 -62.369 54.778
475500 92.116 -62.365 54.756
475501 92.150 -62.362 54.736
475502 92.186 -62.352 54.717
475503 92.223 -62.337 54.696
475504 92.260 -62.323 54.675
475505 92.293 -62.316 54.653
475506 92.322 -62.319 54.631
475507 92.351 -62.321 54.608
475508 92.388 -62.327 54.586
475509 92.437 -62.333 54.564
475510 92.486 -62.341 54.544
475511 92.535 -62.348 54.525
475512 92.584 -62.355 54.505
475513 92.633 -62.362 54.487
475514 92.679 -62.369 54.470
475515 92.722 -62.374 54.454
475516 92.765 -62.379 54.437
475517 92.808 -62.384 54.419
475518 92.826 -62.394 54.400
475519 92.869 -62.399 54.387
475520 92.922 -62.401 54.381
475521 92.974 -62.404 54.377
475522 93.026 -62.406 54.377
475523 93.068 -62.416 54.377
475524 93.108 -62.428 54.372
475525 93.148 -62.438 54.363
475526 93.189 -62.449 54.351
475527 93.228 -62.461 54.337
475528 93.255 -62.498 54.322
475529 93.279 -62.528 54.304
475530 93.302 -62.514 54.285
475531 93.324 -62.500 54.264
475532 93.347 -62.486 54.243
475533 93.377 -62.476 54.224
475534 93.417 -62.472 54.210
475535 93.458 -62.468 54.196
475536 93.499 -62.467 54.182
475537 93.557 -62.491 54.170
475538 93.615 -62.515 54.157
475539 93.673 -62.540 54.142
475540 93.731 -62.564 54.125
475541 93.768 -62.588 54.109
475542 93.805 -62.612 54.092
475543 93.843 -62.637 54.074
475544 93.883 -62.659 54.055
475545 93.940 -62.668 54.036
475546 93.997 -62.677 54.014
475547 94.054 -62.686 53.991
475548 94.111 -62.695 53.967
475549 94.168 -62.704 53.942
475550 94.219 -62.687 53.917
475551 94.271 -62.669 53.893
475552 94.322 -62.651 53.869
475553 94.373 -62.633 53.846
475554 94.418 -62.633 53.824
475555 94.468 -62.639 53.800
475556 94.517 -62.646 53.777
475557 94.518 -62.646 53.753
475558 94.543 -62.665 53.729
475559 94.568 -62.684 53.709
475560 94.594 -62.702 53.687
475561 94.619 -62.721 53.664
475562 94.644 -62.739 53.640
475563 94.669 -62.758 53.616
475564 94.694 -62.775 53.590
475565 94.720 -62.791 53.565
475566 94.746 -62.806 53.539
475567 94.778 -62.813 53.512
475568 94.814 -62.831 53.484
475569 94.851 -62.848 53.455
475570 94.888 -62.865 53.426
475571 94.924 -62.882 53.396
475572 94.961 -62.900 53.366
475573 95.013 -62.901 53.335
475574 95.066 -62.901 53.304
475575 95.119 -62.901 53.273
475576 95.172 -62.901 53.243
475577 95.225 -62.901 53.215
475578 95.277 -62.901 53.187
475579 95.330 -62.901 53.160
475580 95.383 -62.901 53.130
475581 95.436 -62.901 53.097
475582 95.489 -62.901 53.063
475583 95.542 -62.901 53.028
475584 95.599 -62.906 52.995
475585 95.634 -62.904 52.962
475586 95.658 -62.895 52.928
475587 95.686 -62.894 52.893
475588 95.714 -62.892 52.858
475589 95.742 -62.890 52.823
475590 95.770 -62.888 52.786
475591 95.812 -62.875 52.749
475592 95.857 -62.859 52.712
475593 95.902 -62.843 52.675
475594 95.947 -62.827 52.639
475595 95.991 -62.811 52.602
475596 96.044 -62.810 52.565
475597 96.096 -62.809 52.529
475598 96.149 -62.808 52.494
475599 96.201 -62.807 52.458
475600 96.254 -62.807 52.423
475601 96.300 -62.797 52.387
475602 96.324 -62.813 52.350
475603 96.349 -62.829 52.313
475604 96.376 -62.844 52.277
475605 96.471 -62.854 52.240
475606 96.510 -62.866 52.202
475607 96.549 -62.879 52.164
475608 96.588 -62.891 52.125
475609 96.627 -62.904 52.086
475610 96.666 -62.916 52.048
475611 96.705 -62.928 52.010
475612 96.744 -62.940 51.972
475613 96.783 -62.953 51.936
475614 96.822 -62.965 51.901
475615 96.861 -62.977 51.868
475616 96.899 -62.989 51.836
475617 96.934 -63.006 51.804
475618 96.966 -63.023 51.772
475619 96.999 -63.042 51.742
475620 97.032 -63.059 51.711
475621 97.065 -63.077 51.680
475622 97.097 -63.093 51.649
475623 97.128 -63.098 51.618
475624 97.159 -63.102 51.587
475625 97.189 -63.107 51.557
475626 97.220 -63.111 51.525
475627 97.251 -63.116 51.493
475628 97.298 -63.117 51.460
475629 97.349 -63.117 51.427
475630 97.400 -63.117 51.391
475631 97.452 -63.117 51.353
475632 97.501 -63.123 51.313
475633 97.548 -63.136 51.272
475634 97.588 -63.111 51.230
475635 97.598 -63.094 51.188
475636 97.605 -63.077 51.148
475637 97.611 -63.061 51.107
475638 97.626 -63.051 51.065
475639 97.654 -63.053 51.022
475640 97.682 -63.055 50.978
475641 97.710 -63.057 50.934
475642 97.737 -63.061 50.890
475643 97.765 -63.063 50.845
475644 97.793 -63.065 50.801
475645 97.826 -63.054 50.758
475646 97.859 -63.042 50.712
475647 97.898 -63.038 50.664
475648 97.951 -63.051 50.616
475649 98.005 -63.066 50.567
475650 98.059 -63.080 50.516
475651 98.112 -63.095 50.466
475652 98.166 -63.109 50.417
475653 98.220 -63.123 50.367
475654 98.240 -63.138 50.319
475655 98.250 -63.153 50.273
475656 98.262 -63.167 50.227
475657 98.273 -63.183 50.181
475658 98.283 -63.198 50.136
475659 98.294 -63.213 50.092
475660 98.313 -63.228 50.049
475661 98.337 -63.244 50.007
475662 98.361 -63.259 49.965
475663 98.385 -63.275 49.922
475664 98.411 -63.270 49.879
475665 98.437 -63.258 49.839
475666 98.463 -63.245 49.797
475667 98.489 -63.233 49.757
475668 98.515 -63.220 49.718
475669 98.547 -63.216 49.679
475670 98.583 -63.220 49.641
475671 98.619 -63.224 49.604
475672 98.655 -63.228 49.567
475673 98.705 -63.217 49.531
475674 98.768 -63.216 49.495
475675 98.831 -63.214 49.461
475676 98.894 -63.212 49.426
476096 87.484 -61.243 56.054
476097 87.520 -61.244 56.042
476098 87.557 -61.246 56.030
476099 87.594 -61.248 56.017
476100 87.631 -61.250 56.005
476101 87.668 -61.252 55.993
476102 87.704 -61.253 55.980
476103 87.741 -61.255 55.965
476104 87.777 -61.257 55.951
476105 87.814 -61.259 55.934
476106 87.851 -61.261 55.920
476107 87.888 -61.262 55.905
476108 87.924 -61.265 55.888
476109 87.961 -61.266 55.870
476110 87.999 -61.268 55.853
476111 88.047 -61.264 55.834
476112 88.095 -61.261 55.815
476113 88.143 -61.257 55.797
476114 88.192 -61.254 55.778
476115 88.240 -61.251 55.759
476116 88.292 -61.249 55.740
476117 88.359 -61.255 55.721
476118 88.426 -61.262 55.702
476119 88.494 -61.268 55.682
476120 88.561 -61.274 55.661
476121 88.629 -61.281 55.636
476122 88.696 -61.287 55.607
476123 88.753 -61.302 55.575
476124 88.798 -61.328 55.547
476125 88.842 -61.353 55.523
476126 88.887 -61.378 55.504
476127 88.931 -61.404 55.490
476128 88.976 -61.430 55.480
476129 89.020 -61.450 55.472
476130 89.063 -61.468 55.458
476131 89.106 -61.486 55.441
476132 89.149 -61.504 55.423
476133 89.192 -61.522 55.405
476134 89.232 -61.531 55.384
476135 89.266 -61.529 55.361
476136 89.302 -61.528 55.340
476137 89.337 -61.526 55.319
476138 89.372 -61.525 55.300
476139 89.407 -61.523 55.281
476140 89.442 -61.522 55.261
476141 89.478 -61.522 55.245
476142 89.516 -61.523 55.232
476143 89.553 -61.525 55.220
476144 89.591 -61.526 55.212
476145 89.628 -61.528 55.204
476146 89.666 -61.530 55.198
476147 89.704 -61.534 55.194
476148 89.747 -61.544 55.191
476149 89.790 -61.553 55.189
476150 89.832 -61.563 55.190
476151 89.876 -61.578 55.194
476152 89.921 -61.599 55.200
476153 89.967 -61.620 55.208
476154 90.012 -61.641 55.211
476155 90.057 -61.663 55.208
476156 90.103 -61.685 55.202
476157 90.149 -61.706 55.194
476158 90.194 -61.727 55.186
476159 90.239 -61.748 55.184
476160 90.285 -61.769 55.183
476161 90.330 -61.791 55.179
476162 90.376 -61.811 55.170
476163 90.428 -61.823 55.156
476164 90.480 -61.838 55.138
476165 90.533 -61.852 55.114
476166 90.586 -61.867 55.089
476167 90.639 -61.882 55.069
476168 90.692 -61.897 55.056
476169 90.745 -61.912 55.046
476170 90.798 -61.926 55.043
476171 90.851 -61.942 55.043
476172 90.889 -61.958 55.049
476173 90.926 -61.975 55.058
476174 90.962 -61.992 55.070
476175 91.002 -62.006 55.083
476176 91.050 -62.013 55.096
476177 91.098 -62.019 55.105
476178 91.148 -62.021 55.108
476179 91.198 -62.023 55.104
476180 91.243 -62.030 55.085
476181 91.287 -62.039 55.076
476182 91.329 -62.049 55.075
476183 91.371 -62.058 55.081
476184 91.413 -62.068 55.087
476185 91.455 -62.078 55.080
476186 91.497 -62.087 55.065
476187 91.539 -62.100 55.053
476188 91.583 -62.117 55.048
476189 91.628 -62.134 55.053
476190 91.671 -62.152 55.051
476191 91.716 -62.169 55.040
476192 91.759 -62.186 55.023
476193 91.804 -62.203 55.001
476194 91.848 -62.221 54.978
476195 91.889 -62.238 54.945
476196 91.927 -62.257 54.907
476197 91.964 -62.275 54.870
476198 92.002 -62.293 54.834
476199 92.040 -62.312 54.804
476200 92.076 -62.317 54.778
476201 92.111 -62.314 54.754
476202 92.146 -62.311 54.731
476203 92.180 -62.308 54.711
476204 92.216 -62.297 54.690
476205 92.253 -62.283 54.667
476206 92.287 -62.275 54.645
476207 92.316 -62.278 54.622
476208 92.351 -62.283 54.600
476209 92.400 -62.290 54.579
476210 92.449 -62.297 54.558
476211 92.498 -62.304 54.538
476212 92.547 -62.312 54.520
476213 92.596 -62.318 54.503
476214 92.645 -62.326 54.488
476215 92.694 -62.333 54.472
476216 92.739 -62.339 54.457
476217 92.782 -62.344 54.440
476218 92.813 -62.352 54.419
476219 92.829 -62.362 54.398
476220 92.879 -62.365 54.386
476221 92.931 -62.368 54.380
476222 92.983 -62.371 54.377
476223 93.023 -62.383 54.375
476224 93.063 -62.393 54.374
476225 93.103 -62.405 54.367
476226 93.143 -62.416 54.359
476227 93.183 -62.428 54.347
476228 93.223 -62.439 54.332
476229 93.257 -62.462 54.317
476230 93.281 -62.481 54.300
476231 93.303 -62.467 54.281
476232 93.328 -62.454 54.261
476233 93.368 -62.450 54.243
476234 93.409 -62.446 54.227
476235 93.449 -62.443 54.214
476236 93.489 -62.440 54.202
476237 93.530 -62.436 54.191
476238 93.587 -62.459 54.180
476239 93.645 -62.484 54.168
476240 93.702 -62.508 54.154
476241 93.739 -62.532 54.139
476242 93.776 -62.556 54.125
476243 93.814 -62.581 54.112
476244 93.851 -62.605 54.098
476245 93.895 -62.624 54.082
476246 93.952 -62.633 54.063
476247 94.009 -62.642 54.040
476248 94.066 -62.651 54.014
476249 94.123 -62.660 53.988
476250 94.176 -62.653 53.960
476251 94.227 -62.635 53.932
476252 94.279 -62.618 53.903
476253 94.330 -62.600 53.876
476254 94.376 -62.599 53.849
476255 94.420 -62.603 53.823
476256 94.469 -62.609 53.796
476257 94.523 -62.617 53.769
476258 94.577 -62.625 53.742
476259 94.580 -62.625 53.717
476260 94.581 -62.625 53.693
476261 94.607 -62.644 53.668
476262 94.632 -62.663 53.642
476263 94.657 -62.681 53.617
476264 94.682 -62.700 53.592
476265 94.707 -62.719 53.566
476266 94.733 -62.738 53.539
476267 94.772 -62.765 53.511
476268 94.802 -62.805 53.482
476269 94.829 -62.819 53.452
476270 94.859 -62.832 53.423
476271 94.896 -62.849 53.393
476272 94.933 -62.866 53.362
476273 94.976 -62.877 53.332
476274 95.029 -62.877 53.303
476275 95.082 -62.877 53.273
476276 95.134 -62.877 53.243
476277 95.187 -62.877 53.214
476278 95.240 -62.877 53.187
476279 95.293 -62.877 53.161
476280 95.346 -62.877 53.134
476281 95.398 -62.876 53.103
476282 95.449 -62.876 53.071
476283 95.499 -62.875 53.037
476284 95.549 -62.873 53.002
476285 95.600 -62.873 52.970
476286 95.641 -62.877 52.937
476287 95.669 -62.876 52.903
476288 95.697 -62.873 52.869
476289 95.725 -62.872 52.835
476290 95.753 -62.870 52.800
476291 95.781 -62.869 52.764
476292 95.823 -62.855 52.727
476293 95.868 -62.839 52.691
476294 95.913 -62.823 52.655
476295 95.958 -62.807 52.618
476296 96.010 -62.804 52.582
476297 96.062 -62.804 52.545
476298 96.115 -62.803 52.508
476299 96.167 -62.802 52.473
476300 96.213 -62.791 52.438
476301 96.256 -62.775 52.402
476302 96.304 -62.774 52.366
476303 96.329 -62.790 52.330
476304 96.354 -62.806 52.292
476305 96.400 -62.819 52.256
476306 96.474 -62.830 52.219
476307 96.513 -62.842 52.182
476308 96.552 -62.855 52.144
476309 96.591 -62.867 52.106
476310 96.630 -62.879 52.067
476311 96.669 -62.891 52.029
476312 96.708 -62.904 51.992
476313 96.746 -62.916 51.954
476314 96.786 -62.929 51.918
476315 96.824 -62.941 51.883
476316 96.862 -62.954 51.849
476317 96.895 -62.972 51.816
476318 96.928 -62.990 51.784
476319 96.960 -63.007 51.752
476320 96.993 -63.025 51.721
476321 97.026 -63.043 51.690
476322 97.058 -63.061 51.659
476323 97.091 -63.078 51.628
476324 97.123 -63.094 51.597
476325 97.154 -63.098 51.566
476326 97.185 -63.103 51.534
476327 97.223 -63.106 51.503
476328 97.275 -63.106 51.470
476329 97.326 -63.105 51.437
476330 97.377 -63.105 51.403
476331 97.429 -63.105 51.367
476332 97.480 -63.105 51.330
476333 97.531 -63.105 51.291
476334 97.582 -63.108 51.251
476335 97.595 -63.091 51.209
476336 97.602 -63.074 51.167
476337 97.609 -63.057 51.126
476338 97.615 -63.041 51.084
476339 97.627 -63.028 51.042
476340 97.654 -63.030 50.998
476341 97.682 -63.032 50.953
476342 97.710 -63.035 50.908
476343 97.738 -63.037 50.864
476344 97.768 -63.034 50.820
476345 97.801 -63.022 50.776
476346 97.833 -63.010 50.731
476347 97.867 -62.998 50.685
476348 97.912 -63.001 50.638
476349 97.965 -63.016 50.590
476350 98.019 -63.030 50.542
476351 98.073 -63.044 50.493
476352 98.126 -63.058 50.444
476353 98.180 -63.073 50.395
476354 98.221 -63.087 50.348
476355 98.232 -63.102 50.301
476356 98.243 -63.117 50.255
476357 98.254 -63.132 50.209
476358 98.265 -63.147 50.165
476359 98.276 -63.162 50.121
476360 98.291 -63.177 50.078
476361 98.315 -63.193 50.034
476362 98.339 -63.208 49.993
476363 98.363 -63.224 49.951
476364 98.387 -63.239 49.909
476365 98.414 -63.226 49.867
476366 98.439 -63.214 49.826
476367 98.465 -63.201 49.786
476368 98.492 -63.189 49.746
476369 98.517 -63.177 49.707
476370 98.553 -63.179 49.669
476371 98.589 -63.183 49.631
476372 98.625 -63.187 49.594
476373 98.661 -63.188 49.558
476374 98.714 -63.179 49.522
476375 98.777 -63.177 49.487
476376 98.840 -63.175 49.453
476798 87.536 -61.196 56.034
476799 87.572 -61.199 56.022
476800 87.609 -61.200 56.010
476801 87.646 -61.202 55.997
476802 87.683 -61.204 55.985
476803 87.719 -61.205 55.972
476804 87.756 -61.208 55.957
476805 87.793 -61.209 55.942
476806 87.829 -61.211 55.926
476807 87.866 -61.213 55.911
476808 87.903 -61.215 55.896
476809 87.939 -61.216 55.879
476810 87.976 -61.218 55.862
476811 88.013 -61.220 55.844
476812 88.056 -61.218 55.824
476813 88.105 -61.215 55.805
476814 88.153 -61.212 55.786
476815 88.201 -61.208 55.768
476816 88.250 -61.205 55.748
476817 88.303 -61.205 55.726
476818 88.370 -61.211 55.705
476819 88.437 -61.217 55.683
476820 88.505 -61.224 55.662
476821 88.572 -61.230 55.639
476822 88.640 -61.237 55.615
476823 88.709 -61.240 55.588
476824 88.767 -61.252 55.559
476825 88.812 -61.277 55.531
476826 88.856 -61.302 55.507
476827 88.901 -61.328 55.486
476828 88.946 -61.353 55.468
476829 88.990 -61.378 55.453
476830 89.033 -61.396 55.442
476831 89.076 -61.414 55.427
476832 89.119 -61.433 55.409
476833 89.163 -61.451 55.391
476834 89.206 -61.469 55.374
476835 89.249 -61.487 55.355
476836 89.288 -61.495 55.334
476837 89.323 -61.494 55.315
476838 89.358 -61.492 55.299
476839 89.393 -61.491 55.285
476840 89.428 -61.489 55.271
476841 89.463 -61.487 55.251
476842 89.498 -61.486 55.234
476843 89.533 -61.484 55.220
476844 89.568 -61.483 55.209
476845 89.605 -61.484 55.201
476846 89.643 -61.485 55.194
476847 89.680 -61.487 55.189
476848 89.718 -61.488 55.184
476849 89.757 -61.494 55.181
476850 89.800 -61.503 55.180
476851 89.842 -61.512 55.180
476852 89.887 -61.529 55.183
476853 89.932 -61.550 55.189
476854 89.977 -61.572 55.198
476855 90.023 -61.593 55.206
476856 90.068 -61.615 55.208
476857 90.113 -61.636 55.205
476858 90.159 -61.657 55.198
476859 90.204 -61.678 55.190
476860 90.249 -61.700 55.185
476861 90.295 -61.721 55.181
476862 90.341 -61.744 55.174
476863 90.389 -61.768 55.161
476864 90.437 -61.782 55.144
476865 90.490 -61.797 55.123
476866 90.543 -61.811 55.100
476867 90.596 -61.826 55.078
476868 90.649 -61.841 55.061
476869 90.702 -61.856 55.050
476870 90.755 -61.870 55.043
476871 90.808 -61.885 55.040
476872 90.857 -61.901 55.041
476873 90.893 -61.918 55.046
476874 90.930 -61.935 55.055
476875 90.969 -61.949 55.067
476876 91.018 -61.955 55.082
476877 91.066 -61.962 55.095
476878 91.114 -61.968 55.105
476879 91.162 -61.975 55.110
476880 91.211 -61.979 55.106
476881 91.255 -61.987 55.088
476882 91.297 -61.997 55.078
476883 91.339 -62.006 55.077
476884 91.381 -62.016 55.082
476885 91.423 -62.026 55.086
476886 91.465 -62.035 55.077
476887 91.507 -62.045 55.067
476888 91.548 -62.055 55.062
476889 91.592 -62.070 55.062
476890 91.637 -62.087 55.062
476891 91.680 -62.105 55.053
476892 91.725 -62.122 55.037
476893 91.769 -62.139 55.015
476894 91.809 -62.157 54.990
476895 91.848 -62.175 54.966
476896 91.885 -62.194 54.935
476897 91.923 -62.212 54.900
476898 91.961 -62.230 54.866
476899 91.999 -62.249 54.834
476900 92.037 -62.268 54.805
476901 92.072 -62.265 54.779
476902 92.106 -62.262 54.754
476903 92.141 -62.258 54.730
476904 92.176 -62.255 54.708
476905 92.211 -62.252 54.688
476906 92.246 -62.243 54.665
476907 92.280 -62.235 54.641
476908 92.314 -62.239 54.618
476909 92.363 -62.246 54.596
476910 92.412 -62.253 54.575
476911 92.461 -62.261 54.554
476912 92.510 -62.268 54.535
476913 92.559 -62.275 54.518
476914 92.608 -62.282 54.503
476915 92.657 -62.290 54.490
476916 92.706 -62.296 54.475
476917 92.755 -62.304 54.460
476918 92.799 -62.309 54.442
476919 92.817 -62.320 54.420
476920 92.837 -62.330 54.397
476921 92.889 -62.333 54.384
476922 92.937 -62.338 54.377
476923 92.977 -62.349 54.372
476924 93.017 -62.360 54.369
476925 93.058 -62.371 54.365
476926 93.097 -62.383 54.357
476927 93.138 -62.394 54.347
476928 93.178 -62.406 54.335
476929 93.218 -62.417 54.322
476930 93.258 -62.428 54.308
476931 93.282 -62.434 54.292
476932 93.320 -62.429 54.273
476933 93.360 -62.425 54.255
476934 93.400 -62.421 54.239
476935 93.440 -62.418 54.226
476936 93.481 -62.414 54.216
476937 93.521 -62.410 54.207
476938 93.561 -62.406 54.201
476939 93.618 -62.428 54.194
476940 93.673 -62.452 54.183
476941 93.710 -62.476 54.170
476942 93.748 -62.500 54.157
476943 93.785 -62.525 54.146
476944 93.822 -62.549 54.137
476945 93.859 -62.573 54.127
476946 93.906 -62.590 54.113
476947 93.964 -62.599 54.093
476948 94.020 -62.608 54.068
476949 94.077 -62.617 54.041
476950 94.133 -62.620 54.011
476951 94.184 -62.602 53.981
476952 94.236 -62.584 53.950
476953 94.287 -62.566 53.918
476954 94.333 -62.565 53.886
476955 94.377 -62.569 53.856
476956 94.421 -62.573 53.826
476957 94.469 -62.578 53.796
476958 94.524 -62.587 53.765
476959 94.578 -62.595 53.736
476960 94.632 -62.603 53.707
476961 94.643 -62.604 53.679
476962 94.643 -62.603 53.651
476963 94.645 -62.604 53.623
476964 94.671 -62.623 53.595
476965 94.705 -62.648 53.568
476966 94.745 -62.676 53.540
476967 94.785 -62.704 53.512
476968 94.825 -62.732 53.483
476969 94.859 -62.766 53.452
476970 94.872 -62.824 53.421
476971 94.900 -62.837 53.391
476972 94.928 -62.848 53.361
476973 94.953 -62.851 53.331
476974 94.978 -62.847 53.301
476975 95.028 -62.846 53.272
476976 95.078 -62.845 53.243
476977 95.129 -62.844 53.214
476978 95.179 -62.843 53.187
476979 95.230 -62.842 53.161
476980 95.280 -62.841 53.136
476981 95.331 -62.840 53.109
476982 95.381 -62.839 53.079
476983 95.431 -62.838 53.047
476984 95.482 -62.837 53.013
476985 95.532 -62.836 52.979
476986 95.583 -62.835 52.945
476987 95.629 -62.847 52.912
476988 95.680 -62.855 52.879
476989 95.708 -62.854 52.845
476990 95.736 -62.852 52.810
476991 95.764 -62.851 52.777
476992 95.792 -62.849 52.742
476993 95.834 -62.835 52.706
476994 95.879 -62.819 52.670
476995 95.924 -62.803 52.634
476996 95.975 -62.799 52.599
476997 96.028 -62.799 52.562
476998 96.081 -62.798 52.526
476999 96.127 -62.788 52.491
477000 96.170 -62.773 52.457
477001 96.214 -62.759 52.420
477002 96.257 -62.744 52.384
477003 96.308 -62.750 52.348
477004 96.334 -62.766 52.312
477005 96.359 -62.782 52.275
477006 96.424 -62.794 52.238
477007 96.477 -62.807 52.201
477008 96.516 -62.819 52.164
477009 96.554 -62.831 52.126
477010 96.594 -62.843 52.088
477011 96.632 -62.856 52.050
477012 96.671 -62.868 52.013
477013 96.711 -62.880 51.975
477014 96.749 -62.892 51.937
477015 96.788 -62.905 51.900
477016 96.823 -62.920 51.865
477017 96.856 -62.938 51.831
477018 96.889 -62.956 51.798
477019 96.922 -62.973 51.765
477020 96.954 -62.992 51.732
477021 96.987 -63.009 51.701
477022 97.020 -63.026 51.669
477023 97.052 -63.045 51.638
477024 97.085 -63.062 51.608
477025 97.118 -63.079 51.578
477026 97.150 -63.095 51.545
477027 97.200 -63.095 51.514
477028 97.247 -63.092 51.482
477029 97.295 -63.091 51.449
477030 97.342 -63.089 51.415
477031 97.389 -63.086 51.380
477032 97.436 -63.085 51.345
477033 97.483 -63.082 51.309
477034 97.530 -63.080 51.271
477035 97.577 -63.079 51.230
477036 97.597 -63.069 51.190
477037 97.606 -63.054 51.148
477038 97.612 -63.038 51.107
477039 97.619 -63.021 51.064
477040 97.627 -63.004 51.021
477041 97.654 -63.007 50.978
477042 97.682 -63.010 50.932
477043 97.710 -63.012 50.886
477044 97.742 -63.002 50.841
477045 97.776 -62.990 50.797
477046 97.808 -62.978 50.753
477047 97.842 -62.967 50.707
477048 97.875 -62.954 50.662
477049 97.926 -62.966 50.615
477050 97.980 -62.979 50.567
477051 98.033 -62.994 50.520
477052 98.087 -63.008 50.472
477053 98.140 -63.023 50.424
477054 98.194 -63.036 50.377
477055 98.214 -63.051 50.330
477056 98.225 -63.067 50.285
477057 98.236 -63.082 50.239
477058 98.246 -63.096 50.195
477059 98.257 -63.111 50.151
477060 98.269 -63.126 50.108
477061 98.292 -63.142 50.065
477062 98.316 -63.158 50.022
477063 98.341 -63.173 49.980
477064 98.365 -63.189 49.939
477065 98.389 -63.195 49.897
477066 98.416 -63.183 49.857
477067 98.442 -63.170 49.816
477068 98.468 -63.158 49.776
477069 98.494 -63.145 49.736
477070 98.522 -63.137 49.698
477071 98.559 -63.141 49.661
477072 98.595 -63.145 49.623
477073 98.631 -63.150 49.587
477074 98.667 -63.147 49.551
477501 87.625 -61.152 56.002
477502 87.661 -61.154 55.990
477503 87.698 -61.156 55.978
477504 87.734 -61.158 55.963
477505 87.771 -61.159 55.948
477506 87.808 -61.161 55.934
477507 87.845 -61.163 55.918
477508 87.882 -61.165 55.902
477509 87.918 -61.167 55.887
477510 87.955 -61.168 55.870
477511 87.991 -61.171 55.853
477512 88.028 -61.172 55.835
477513 88.066 -61.174 55.815
477514 88.114 -61.170 55.796
477515 88.163 -61.167 55.777
477516 88.211 -61.163 55.758
477517 88.259 -61.160 55.737
477518 88.314 -61.160 55.716
477519 88.381 -61.166 55.693
477520 88.449 -61.173 55.670
477521 88.516 -61.179 55.646
477522 88.583 -61.185 55.623
477523 88.651 -61.192 55.598
477524 88.727 -61.185 55.571
477525 88.781 -61.201 55.543
477526 88.826 -61.227 55.516
477527 88.871 -61.252 55.492
477528 88.915 -61.277 55.469
477529 88.960 -61.303 55.448
477530 89.003 -61.324 55.428
477531 89.047 -61.343 55.413
477532 89.090 -61.361 55.396
477533 89.133 -61.379 55.378
477534 89.176 -61.397 55.361
477535 89.219 -61.415 55.345
477536 89.263 -61.434 55.327
477537 89.306 -61.453 55.310
477538 89.344 -61.459 55.293
477539 89.379 -61.457 55.280
477540 89.414 -61.456 55.269
477541 89.450 -61.454 55.258
477542 89.484 -61.453 55.241
477543 89.520 -61.451 55.224
477544 89.554 -61.450 55.209
477545 89.590 -61.448 55.198
477546 89.625 -61.447 55.190
477547 89.660 -61.445 55.183
477548 89.695 -61.443 55.178
477549 89.733 -61.446 55.174
477550 89.770 -61.447 55.170
477551 89.810 -61.453 55.170
477552 89.852 -61.462 55.169
477553 89.897 -61.481 55.171
477554 89.943 -61.502 55.176
477555 89.988 -61.523 55.185
477556 90.033 -61.544 55.196
477557 90.079 -61.566 55.203
477558 90.124 -61.588 55.204
477559 90.172 -61.613 55.201
477560 90.222 -61.642 55.194
477561 90.271 -61.671 55.187
477562 90.321 -61.700 55.180
477563 90.367 -61.719 55.170
477564 90.409 -61.729 55.154
477565 90.455 -61.742 55.135
477566 90.500 -61.755 55.113
477567 90.553 -61.770 55.090
477568 90.606 -61.785 55.071
477569 90.659 -61.800 55.056
477570 90.712 -61.814 55.046
477571 90.764 -61.829 55.039
477572 90.817 -61.844 55.037
477573 90.862 -61.860 55.039
477574 90.898 -61.877 55.045
477575 90.937 -61.892 55.055
477576 90.985 -61.898 55.067
477577 91.033 -61.905 55.082
477578 91.081 -61.911 55.094
477579 91.130 -61.918 55.104
477580 91.177 -61.925 55.109
477581 91.223 -61.935 55.105
477582 91.265 -61.945 55.090
477583 91.307 -61.954 55.078
477584 91.349 -61.964 55.075
477585 91.391 -61.973 55.080
477586 91.433 -61.983 55.082
477587 91.475 -61.992 55.073
477588 91.517 -62.001 55.069
477589 91.559 -62.011 55.070
477590 91.601 -62.023 55.073
477591 91.645 -62.040 55.066
477592 91.689 -62.058 55.052
477593 91.730 -62.076 55.033
477594 91.768 -62.094 55.009
477595 91.806 -62.112 54.980
477596 91.844 -62.131 54.954
477597 91.882 -62.149 54.926
477598 91.920 -62.167 54.895
477599 91.958 -62.186 54.865
477600 91.996 -62.205 54.837
477601 92.032 -62.217 54.809
477602 92.067 -62.213 54.782
477603 92.102 -62.210 54.757
477604 92.137 -62.207 54.734
477605 92.171 -62.204 54.714
477606 92.206 -62.201 54.694
477607 92.241 -62.197 54.669
477608 92.277 -62.195 54.644
477609 92.326 -62.202 54.620
477610 92.375 -62.209 54.597
477611 92.424 -62.217 54.574
477612 92.473 -62.224 54.554
477613 92.521 -62.231 54.536
477614 92.570 -62.238 54.519
477615 92.619 -62.246 54.505
477616 92.667 -62.252 54.495
477617 92.716 -62.259 54.480
477618 92.765 -62.267 54.464
477619 92.804 -62.276 54.444
477620 92.820 -62.288 54.419
477621 92.847 -62.296 54.395
477622 92.892 -62.304 54.380
477623 92.932 -62.315 54.371
477624 92.972 -62.327 54.365
477625 93.012 -62.338 54.359
477626 93.052 -62.349 54.352
477627 93.092 -62.361 54.343
477628 93.135 -62.371 54.332
477629 93.177 -62.381 54.320
477630 93.219 -62.392 54.309
477631 93.262 -62.402 54.295
477632 93.303 -62.406 54.280
477633 93.346 -62.404 54.263
477634 93.389 -62.398 54.247
477635 93.432 -62.392 54.233
477636 93.472 -62.388 54.225
477637 93.512 -62.384 54.221
477638 93.553 -62.381 54.219
477639 93.593 -62.377 54.217
477640 93.644 -62.396 54.212
477641 93.681 -62.420 54.202
477642 93.718 -62.444 54.191
477643 93.756 -62.468 54.180
477644 93.793 -62.493 54.169
477645 93.830 -62.517 54.158
477646 93.867 -62.541 54.146
477647 93.918 -62.555 54.130
477648 93.975 -62.564 54.109
477649 94.032 -62.573 54.085
477650 94.089 -62.582 54.057
477651 94.141 -62.569 54.028
477652 94.193 -62.551 53.997
477653 94.244 -62.533 53.966
477654 94.289 -62.531 53.933
477655 94.334 -62.535 53.899
477656 94.378 -62.539 53.867
477657 94.422 -62.543 53.834
477658 94.470 -62.549 53.801
477659 94.525 -62.557 53.767
477660 94.579 -62.565 53.734
477661 94.633 -62.573 53.700
477662 94.688 -62.581 53.668
477663 94.706 -62.584 53.637
477664 94.699 -62.589 53.606
477665 94.687 -62.598 53.575
477666 94.736 -62.619 53.545
477667 94.797 -62.643 53.515
477668 94.838 -62.671 53.486
477669 94.878 -62.699 53.455
477670 94.918 -62.727 53.423
477671 94.930 -62.785 53.392
477672 94.943 -62.843 53.360
477673 94.958 -62.825 53.331
477674 94.946 -62.799 53.301
477675 94.983 -62.802 53.272
477676 95.021 -62.805 53.243
477677 95.061 -62.808 53.214
477678 95.112 -62.807 53.187
477679 95.163 -62.806 53.161
477680 95.213 -62.804 53.135
477681 95.263 -62.804 53.110
477682 95.314 -62.803 53.083
477683 95.364 -62.801 53.054
477684 95.414 -62.800 53.022
477685 95.465 -62.800 52.990
477686 95.515 -62.798 52.957
477687 95.565 -62.797 52.923
477688 95.616 -62.797 52.891
477689 95.657 -62.832 52.859
477690 95.713 -62.835 52.826
477691 95.747 -62.832 52.791
477692 95.775 -62.831 52.759
477693 95.803 -62.829 52.725
477694 95.845 -62.815 52.691
477695 95.891 -62.799 52.655
477696 95.941 -62.794 52.620
477697 95.994 -62.794 52.584
477698 96.041 -62.785 52.547
477699 96.085 -62.770 52.512
477700 96.128 -62.756 52.477
477701 96.171 -62.741 52.443
477702 96.215 -62.727 52.406
477703 96.258 -62.712 52.369
477704 96.311 -62.728 52.333
477705 96.338 -62.744 52.297
477706 96.363 -62.759 52.260
477707 96.441 -62.770 52.223
477708 96.479 -62.782 52.186
477709 96.519 -62.794 52.149
477710 96.557 -62.807 52.111
477711 96.596 -62.819 52.073
477712 96.636 -62.832 52.035
477713 96.674 -62.844 51.998
477714 96.713 -62.857 51.960
477715 96.752 -62.869 51.922
477716 96.785 -62.886 51.885
477717 96.817 -62.904 51.848
477718 96.850 -62.922 51.814
477719 96.883 -62.940 51.780
477720 96.915 -62.957 51.746
477721 96.948 -62.975 51.714
477722 96.981 -62.993 51.683
477723 97.014 -63.010 51.651
477724 97.046 -63.028 51.620
477725 97.081 -63.040 51.589
477726 97.118 -63.047 51.559
477727 97.157 -63.052 51.527
477728 97.196 -63.057 51.495
477729 97.236 -63.062 51.462
477730 97.282 -63.061 51.429
477731 97.329 -63.059 51.395
477732 97.376 -63.057 51.360
477733 97.423 -63.054 51.325
477734 97.470 -63.053 51.289
477735 97.517 -63.051 51.251
477736 97.565 -63.048 51.211
477737 97.596 -63.042 51.173
477738 97.605 -63.029 51.132
477739 97.615 -63.016 51.090
477740 97.623 -63.001 51.047
477741 97.630 -62.985 51.004
477742 97.655 -62.984 50.960
477743 97.684 -62.982 50.914
477744 97.717 -62.970 50.868
477745 97.750 -62.958 50.822
477746 97.783 -62.946 50.777
477747 97.816 -62.935 50.732
477748 97.850 -62.923 50.686
477749 97.886 -62.915 50.641
477750 97.940 -62.929 50.595
477751 97.993 -62.944 50.547
477752 98.047 -62.957 50.501
477753 98.100 -62.972 50.454
477754 98.154 -62.986 50.407
477755 98.196 -63.001 50.361
477756 98.206 -63.016 50.315
477757 98.217 -63.031 50.270
477758 98.228 -63.045 50.226
477759 98.239 -63.061 50.182
477760 98.250 -63.076 50.139
477761 98.270 -63.091 50.096
477762 98.294 -63.107 50.053
477763 98.318 -63.123 50.011
477764 98.343 -63.138 49.970
477765 98.367 -63.154 49.928
477766 98.392 -63.151 49.887
477767 98.418 -63.139 49.847
477768 98.444 -63.126 49.807
478203 87.677 -61.106 55.981
478204 87.713 -61.108 55.968
478205 87.750 -61.110 55.954
478206 87.786 -61.111 55.940
478207 87.823 -61.114 55.925
478208 87.860 -61.115 55.909
478209 87.897 -61.117 55.894
478210 87.933 -61.119 55.878
478211 87.970 -61.121 55.860
478212 88.006 -61.122 55.844
478213 88.043 -61.124 55.826
478214 88.080 -61.126 55.806
478215 88.124 -61.124 55.787
478216 88.172 -61.121 55.767
478217 88.220 -61.118 55.748
478218 88.269 -61.114 55.727
478219 88.325 -61.115 55.705
478220 88.392 -61.121 55.682
478221 88.460 -61.128 55.658
478222 88.527 -61.134 55.633
478223 88.595 -61.141 55.610
478224 88.668 -61.138 55.584
478225 88.743 -61.133 55.557
478226 88.796 -61.151 55.530
478227 88.840 -61.176 55.502
478228 88.885 -61.202 55.477
478229 88.929 -61.227 55.454
478230 88.974 -61.252 55.430
478231 89.018 -61.271 55.408
478232 89.060 -61.290 55.388
478233 89.104 -61.308 55.369
478234 89.147 -61.326 55.351
478235 89.190 -61.344 55.334
478236 89.233 -61.362 55.318
478237 89.276 -61.381 55.302
478238 89.319 -61.399 55.287
478239 89.363 -61.417 55.274
478240 89.401 -61.423 55.262
478241 89.436 -61.422 55.253
478242 89.471 -61.420 55.243
478243 89.506 -61.418 55.228
478244 89.541 -61.417 55.213
478245 89.576 -61.415 55.198
478246 89.611 -61.414 55.187
478247 89.646 -61.412 55.179
478248 89.681 -61.411 55.173
478249 89.716 -61.409 55.167
478250 89.751 -61.408 55.163
478251 89.787 -61.406 55.160
478252 89.822 -61.406 55.159
478253 89.863 -61.412 55.159
478254 89.907 -61.432 55.160
478255 89.953 -61.453 55.162
478256 90.003 -61.482 55.168
478257 90.052 -61.511 55.180
478258 90.102 -61.541 55.191
478259 90.152 -61.569 55.197
478260 90.201 -61.598 55.198
478261 90.251 -61.628 55.196
478262 90.301 -61.657 55.189
478263 90.344 -61.671 55.180
478264 90.385 -61.678 55.167
478265 90.431 -61.691 55.149
478266 90.476 -61.704 55.128
478267 90.521 -61.717 55.106
478268 90.566 -61.730 55.083
478269 90.615 -61.744 55.065
478270 90.668 -61.759 55.052
478271 90.721 -61.773 55.041
478272 90.775 -61.788 55.035
478273 90.827 -61.803 55.033
478274 90.866 -61.820 55.036
478275 90.904 -61.835 55.043
478276 90.953 -61.841 55.053
478277 91.001 -61.848 55.066
478278 91.049 -61.854 55.082
478279 91.097 -61.862 55.094
478280 91.145 -61.870 55.103
478281 91.190 -61.880 55.107
478282 91.232 -61.892 55.100
478283 91.275 -61.903 55.087
478284 91.316 -61.912 55.078
478285 91.359 -61.922 55.074
478286 91.400 -61.931 55.076
478287 91.443 -61.941 55.077
478288 91.485 -61.950 55.069
478289 91.527 -61.959 55.069
478290 91.569 -61.969 55.073
478291 91.611 -61.978 55.076
478292 91.651 -61.994 55.066
478293 91.689 -62.013 55.050
478294 91.727 -62.031 55.029
478295 91.765 -62.049 55.003
478296 91.803 -62.068 54.973
478297 91.841 -62.086 54.945
478298 91.878 -62.105 54.918
478299 91.917 -62.123 54.892
478300 91.954 -62.142 54.868
478301 91.992 -62.160 54.844
478302 92.028 -62.164 54.815
478303 92.063 -62.161 54.786
478304 92.101 -62.161 54.760
478305 92.139 -62.160 54.739
478306 92.177 -62.159 54.721
478307 92.216 -62.158 54.701
478308 92.254 -62.158 54.675
478309 92.290 -62.158 54.649
478310 92.335 -62.164 54.624
478311 92.383 -62.171 54.600
478312 92.432 -62.178 54.578
478313 92.481 -62.186 54.558
478314 92.529 -62.192 54.540
478315 92.578 -62.200 54.524
478316 92.627 -62.207 54.512
478317 92.675 -62.214 54.499
478318 92.724 -62.221 54.483
478319 92.773 -62.228 54.463
478320 92.806 -62.239 54.439
478321 92.822 -62.253 54.413
478322 92.854 -62.268 54.389
478323 92.896 -62.278 54.373
478324 92.938 -62.289 54.362
478325 92.981 -62.299 54.354
478326 93.022 -62.310 54.346
478327 93.065 -62.321 54.336
478328 93.107 -62.330 54.325
478329 93.149 -62.341 54.314
478330 93.191 -62.352 54.303
478331 93.234 -62.361 54.291
478332 93.278 -62.367 54.280
478333 93.319 -62.371 54.265
478334 93.360 -62.375 54.251
478335 93.401 -62.380 54.237
478336 93.442 -62.381 54.226
478337 93.485 -62.374 54.223
478338 93.529 -62.368 54.228
478339 93.571 -62.363 54.231
478340 93.615 -62.357 54.231
478341 93.652 -62.364 54.227
478342 93.689 -62.388 54.219
478343 93.726 -62.412 54.208
478344 93.764 -62.437 54.198
478345 93.801 -62.461 54.186
478346 93.839 -62.485 54.172
478347 93.876 -62.509 54.156
478348 93.930 -62.521 54.138
478349 93.987 -62.530 54.116
478350 94.044 -62.539 54.091
478351 94.098 -62.535 54.063
478352 94.149 -62.518 54.034
478353 94.201 -62.500 54.004
478354 94.247 -62.498 53.973
478355 94.291 -62.502 53.941
478356 94.335 -62.506 53.907
478357 94.379 -62.509 53.873
478358 94.423 -62.513 53.840
478359 94.471 -62.518 53.805
478360 94.526 -62.527 53.768
478361 94.580 -62.535 53.732
478362 94.634 -62.543 53.695
478363 94.688 -62.551 53.660
478364 94.720 -62.566 53.626
478365 94.740 -62.584 53.592
478366 94.731 -62.594 53.559
478367 94.719 -62.603 53.527
478368 94.773 -62.625 53.495
478369 94.834 -62.649 53.463
478370 94.896 -62.673 53.431
478371 94.957 -62.697 53.399
478372 94.924 -62.778 53.366
478373 94.962 -62.797 53.335
478374 94.950 -62.772 53.305
478375 94.951 -62.753 53.275
478376 94.989 -62.757 53.246
478377 95.027 -62.760 53.218
478378 95.065 -62.763 53.188
478379 95.103 -62.766 53.160
478380 95.145 -62.768 53.134
478381 95.196 -62.767 53.108
478382 95.246 -62.766 53.082
478383 95.297 -62.765 53.057
478384 95.347 -62.764 53.028
478385 95.397 -62.763 52.997
478386 95.448 -62.762 52.966
478387 95.498 -62.761 52.935
478388 95.549 -62.760 52.903
478389 95.599 -62.759 52.872
478390 95.643 -62.783 52.841
478391 95.684 -62.818 52.809
478392 95.743 -62.815 52.776
478393 95.785 -62.811 52.744
478394 95.814 -62.809 52.711
478395 95.857 -62.794 52.678
478396 95.907 -62.789 52.643
478397 95.955 -62.782 52.607
478398 95.999 -62.767 52.571
478399 96.042 -62.753 52.534
478400 96.086 -62.738 52.499
478401 96.129 -62.724 52.465
478402 96.172 -62.710 52.430
478403 96.216 -62.695 52.393
478404 96.261 -62.685 52.357
478405 96.315 -62.704 52.320
478406 96.343 -62.720 52.284
478407 96.372 -62.735 52.247
478408 96.444 -62.746 52.210
478409 96.482 -62.759 52.172
478410 96.521 -62.771 52.134
478411 96.560 -62.783 52.097
478412 96.599 -62.795 52.059
478413 96.638 -62.808 52.021
478414 96.677 -62.820 51.983
478415 96.713 -62.835 51.944
478416 96.746 -62.853 51.906
478417 96.779 -62.870 51.869
478418 96.811 -62.888 51.833
478419 96.844 -62.906 51.798
478420 96.877 -62.923 51.763
478421 96.909 -62.941 51.729
478422 96.942 -62.959 51.696
478423 96.975 -62.975 51.664
478424 97.013 -62.981 51.633
478425 97.050 -62.987 51.601
478426 97.088 -62.993 51.570
478427 97.125 -62.999 51.539
478428 97.164 -63.004 51.508
478429 97.204 -63.009 51.476
478430 97.243 -63.014 51.444
478431 97.283 -63.020 51.411
478432 97.322 -63.024 51.376
478433 97.364 -63.027 51.342
478434 97.411 -63.025 51.307
478435 97.458 -63.023 51.271
478436 97.505 -63.021 51.235
478437 97.552 -63.019 51.196
478438 97.594 -63.016 51.157
478439 97.604 -63.002 51.117
478440 97.613 -62.989 51.075
478441 97.623 -62.976 51.033
478442 97.633 -62.963 50.990
478443 97.659 -62.950 50.944
478444 97.692 -62.938 50.897
478445 97.725 -62.926 50.851
478446 97.758 -62.914 50.805
478447 97.791 -62.903 50.760
478448 97.824 -62.891 50.714
478449 97.857 -62.879 50.669
478450 97.900 -62.879 50.623
478451 97.954 -62.893 50.578
478452 98.007 -62.907 50.531
478453 98.061 -62.922 50.484
478454 98.114 -62.936 50.439
478455 98.168 -62.950 50.393
478456 98.188 -62.965 50.347
478457 98.199 -62.980 50.302
478458 98.210 -62.995 50.257
478459 98.221 -63.010 50.214
478460 98.232 -63.025 50.171
478461 98.247 -63.040 50.128
478462 98.272 -63.056 50.086
478463 98.296 -63.072 50.043
478464 98.320 -63.087 50.001
478465 98.344 -63.103 49.960
478905 87.728 -61.060 55.959
478906 87.765 -61.062 55.945
478907 87.802 -61.064 55.931
478908 87.838 -61.065 55.916
478909 87.875 -61.067 55.900
478910 87.912 -61.069 55.884
478911 87.948 -61.071 55.868
478912 87.985 -61.073 55.852
478913 88.022 -61.074 55.834
478914 88.059 -61.077 55.817
478915 88.095 -61.078 55.797
478916 88.133 -61.080 55.777
478917 88.182 -61.076 55.759
478918 88.230 -61.073 55.738
478919 88.278 -61.069 55.717
478920 88.336 -61.071 55.695
478921 88.403 -61.077 55.671
478922 88.471 -61.083 55.647
478923 88.538 -61.089 55.623
478924 88.609 -61.090 55.598
478925 88.685 -61.084 55.572
478926 88.760 -61.080 55.545
478927 88.810 -61.100 55.518
478928 88.854 -61.126 55.490
478929 88.899 -61.151 55.464
478930 88.944 -61.177 55.440
478931 88.988 -61.200 55.415
478932 89.031 -61.218 55.392
478933 89.074 -61.237 55.371
478934 89.117 -61.255 55.351
478935 89.161 -61.273 55.335
478936 89.203 -61.291 55.320
478937 89.247 -61.309 55.305
478938 89.290 -61.327 55.291
478939 89.333 -61.346 55.275
478940 89.376 -61.364 55.260
478941 89.419 -61.382 55.246
478942 89.457 -61.387 55.235
478943 89.492 -61.386 55.224
478944 89.527 -61.384 55.213
478945 89.562 -61.383 55.200
478946 89.597 -61.381 55.188
478947 89.632 -61.379 55.177
478948 89.667 -61.378 55.168
478949 89.702 -61.376 55.162
478950 89.738 -61.375 55.156
478951 89.772 -61.373 55.152
478952 89.808 -61.371 55.150
478953 89.843 -61.370 55.148
478954 89.881 -61.375 55.148
478955 89.933 -61.409 55.148
478956 89.982 -61.438 55.149
478957 90.032 -61.467 55.152
478958 90.082 -61.497 55.161
478959 90.131 -61.525 55.173
478960 90.181 -61.554 55.183
478961 90.231 -61.584 55.189
478962 90.281 -61.613 55.189
478963 90.322 -61.622 55.184
478964 90.363 -61.629 55.176
478965 90.406 -61.639 55.163
478966 90.451 -61.652 55.146
478967 90.497 -61.665 55.126
478968 90.542 -61.678 55.105
478969 90.587 -61.691 55.081
478970 90.633 -61.704 55.061
478971 90.678 -61.717 55.047
478972 90.731 -61.732 55.037
478973 90.784 -61.747 55.031
478974 90.834 -61.762 55.029
478975 90.872 -61.778 55.032
478976 90.920 -61.784 55.040
478977 90.968 -61.791 55.050
478978 91.016 -61.798 55.064
478979 91.065 -61.807 55.081
478980 91.112 -61.815 55.094
478981 91.157 -61.826 55.102
478982 91.200 -61.838 55.104
478983 91.242 -61.850 55.095
478984 91.284 -61.861 55.086
478985 91.326 -61.870 55.081
478986 91.368 -61.880 55.079
478987 91.410 -61.889 55.078
478988 91.452 -61.898 55.077
478989 91.494 -61.908 55.069
478990 91.536 -61.917 55.068
478991 91.578 -61.927 55.072
478992 91.620 -61.937 55.073
478993 91.660 -61.951 55.065
478994 91.698 -61.970 55.053
478995 91.735 -61.988 55.034
478996 91.773 -62.007 55.008
478997 91.811 -62.025 54.978
478998 91.848 -62.043 54.944
478999 91.886 -62.061 54.917
479000 91.924 -62.080 54.898
479001 91.961 -62.098 54.881
479002 92.002 -62.116 54.857
479003 92.045 -62.127 54.823
479004 92.084 -62.127 54.791
479005 92.122 -62.127 54.765
479006 92.160 -62.126 54.744
479007 92.199 -62.125 54.729
479008 92.237 -62.124 54.708
479009 92.274 -62.124 54.682
479010 92.309 -62.126 54.655
479011 92.344 -62.127 54.630
479012 92.390 -62.133 54.607
479013 92.439 -62.140 54.585
479014 92.488 -62.147 54.565
479015 92.536 -62.155 54.546
479016 92.585 -62.161 54.531
479017 92.634 -62.169 54.517
479018 92.683 -62.176 54.501
479019 92.731 -62.183 54.482
479020 92.780 -62.190 54.458
479021 92.807 -62.202 54.432
479022 92.824 -62.216 54.406
479023 92.866 -62.232 54.383
479024 92.910 -62.248 54.366
479025 92.952 -62.259 54.353
479026 92.995 -62.270 54.341
479027 93.037 -62.280 54.330
479028 93.079 -62.290 54.318
479029 93.121 -62.301 54.306
479030 93.163 -62.312 54.295
479031 93.206 -62.321 54.283
479032 93.250 -62.326 54.271
479033 93.294 -62.332 54.260
479034 93.335 -62.337 54.246
479035 93.375 -62.340 54.233
479036 93.416 -62.345 54.223
479037 93.457 -62.349 54.218
479038 93.498 -62.354 54.224
479039 93.539 -62.357 54.231
479040 93.582 -62.351 54.236
479041 93.625 -62.345 54.237
479042 93.661 -62.340 54.233
479043 93.698 -62.356 54.227
479044 93.735 -62.381 54.218
479045 93.772 -62.405 54.207
479046 93.809 -62.429 54.194
479047 93.847 -62.453 54.177
479048 93.885 -62.477 54.157
479049 93.941 -62.486 54.137
479050 93.999 -62.495 54.115
479051 94.055 -62.502 54.090
479052 94.106 -62.484 54.063
479053 94.157 -62.466 54.034
479054 94.204 -62.464 54.005
479055 94.248 -62.468 53.975
479056 94.292 -62.471 53.943
479057 94.336 -62.476 53.909
479058 94.381 -62.480 53.875
479059 94.425 -62.484 53.842
479060 94.472 -62.489 53.806
479061 94.526 -62.496 53.767
479062 94.580 -62.505 53.728
479063 94.635 -62.513 53.688
479064 94.679 -62.524 53.650
479065 94.699 -62.542 53.614
479066 94.718 -62.560 53.580
479067 94.737 -62.578 53.546
479068 94.756 -62.597 53.512
479069 94.751 -62.607 53.477
479070 94.810 -62.631 53.444
479071 94.834 -62.678 53.412
479072 94.878 -62.704 53.379
479073 94.923 -62.731 53.345
479074 94.955 -62.745 53.313
479075 94.943 -62.719 53.283
479076 94.957 -62.709 53.252
479077 94.994 -62.712 53.223
479078 95.032 -62.715 53.196
479079 95.070 -62.718 53.166
479080 95.108 -62.722 53.139
479081 95.146 -62.725 53.112
479082 95.184 -62.728 53.085
479083 95.229 -62.728 53.060
479084 95.280 -62.728 53.035
479085 95.330 -62.726 53.006
479086 95.381 -62.725 52.977
479087 95.431 -62.725 52.948
479088 95.481 -62.723 52.918
479089 95.532 -62.722 52.887
479090 95.582 -62.722 52.857
479091 95.629 -62.734 52.827
479092 95.670 -62.769 52.796
479093 95.711 -62.803 52.765
479094 95.772 -62.795 52.733
479095 95.825 -62.789 52.701
479096 95.870 -62.779 52.668
479097 95.913 -62.765 52.632
479098 95.956 -62.750 52.595
479099 96.000 -62.735 52.560
479100 96.043 -62.721 52.524
479101 96.087 -62.706 52.489
479102 96.130 -62.692 52.453
479103 96.173 -62.678 52.419
479104 96.217 -62.663 52.382
479105 96.265 -62.662 52.346
479106 96.319 -62.681 52.309
479107 96.347 -62.697 52.272
479108 96.396 -62.710 52.235
479109 96.446 -62.722 52.198
479110 96.485 -62.735 52.160
479111 96.524 -62.747 52.122
479112 96.563 -62.760 52.083
479113 96.602 -62.772 52.045
479114 96.641 -62.784 52.007
479115 96.674 -62.801 51.968
479116 96.707 -62.819 51.929
479117 96.740 -62.836 51.890
479118 96.772 -62.854 51.852
479119 96.805 -62.872 51.817
479120 96.838 -62.889 51.781
479121 96.871 -62.907 51.746
479122 96.907 -62.915 51.712
479123 96.945 -62.921 51.678
479124 96.982 -62.927 51.646
479125 97.019 -62.933 51.614
479126 97.056 -62.939 51.582
479127 97.094 -62.945 51.551
479128 97.132 -62.951 51.520
479129 97.171 -62.957 51.488
479130 97.211 -62.961 51.457
479131 97.250 -62.967 51.425
479132 97.290 -62.972 51.393
479133 97.329 -62.976 51.358
479134 97.369 -62.982 51.325
479135 97.408 -62.986 51.289
479136 97.448 -62.992 51.254
479137 97.493 -62.992 51.218
479138 97.540 -62.989 51.180
479139 97.587 -62.987 51.142
479140 97.602 -62.976 51.101
479141 97.612 -62.963 51.059
479142 97.624 -62.945 51.017
479143 97.639 -62.922 50.974
479144 97.667 -62.907 50.928
479145 97.700 -62.894 50.881
479146 97.733 -62.882 50.834
479147 97.766 -62.870 50.789
479148 97.799 -62.859 50.744
479149 97.832 -62.847 50.698
479150 97.865 -62.835 50.653
479151 97.914 -62.843 50.607
479152 97.967 -62.857 50.561
479153 98.021 -62.871 50.516
479154 98.075 -62.885 50.469
479155 98.129 -62.900 50.424
479156 98.169 -62.914 50.379
479157 98.180 -62.929 50.334
479158 98.192 -62.945 50.289
479159 98.202 -62.960 50.245
479160 98.213 -62.974 50.202
479161 98.225 -62.989 50.160
479162 98.249 -63.005 50.118
479163 98.274 -63.020 50.076
479164 98.298 -63.036 50.034
479607 87.780 -61.014 55.937
479608 87.817 -61.016 55.922
479609 87.854 -61.017 55.907
479610 87.890 -61.020 55.890
479611 87.927 -61.021 55.874
479612 87.964 -61.023 55.858
479613 88.000 -61.025 55.843
479614 88.037 -61.027 55.825
479615 88.074 -61.028 55.807
479616 88.111 -61.030 55.788
479617 88.147 -61.032 55.768
479618 88.191 -61.030 55.750
479619 88.239 -61.027 55.729
479620 88.287 -61.024 55.708
479621 88.347 -61.026 55.686
479622 88.414 -61.033 55.662
479623 88.482 -61.039 55.637
479624 88.550 -61.043 55.613
479625 88.626 -61.036 55.587
479626 88.702 -61.030 55.561
479627 88.777 -61.027 55.534
479628 88.824 -61.050 55.507
479629 88.869 -61.075 55.479
479630 88.913 -61.101 55.453
479631 88.958 -61.127 55.427
479632 89.001 -61.147 55.402
479633 89.045 -61.165 55.378
479634 89.088 -61.183 55.356
479635 89.131 -61.202 55.337
479636 89.174 -61.220 55.321
479637 89.217 -61.238 55.308
479638 89.260 -61.256 55.295
479639 89.304 -61.274 55.281
479640 89.346 -61.293 55.268
479641 89.390 -61.311 55.254
479642 89.433 -61.329 55.240
479643 89.476 -61.347 55.227
479644 89.513 -61.351 55.214
479645 89.548 -61.349 55.203
479646 89.584 -61.348 55.193
479647 89.618 -61.346 55.182
479648 89.654 -61.345 55.168
479649 89.689 -61.343 55.159
479650 89.724 -61.342 55.151
479651 89.759 -61.340 55.145
479652 89.794 -61.339 55.141
479653 89.829 -61.337 55.138
479654 89.864 -61.336 55.136
479655 89.905 -61.348 55.136
479656 89.962 -61.394 55.136
479657 90.012 -61.424 55.136
479658 90.061 -61.453 55.137
479659 90.111 -61.481 55.143
479660 90.161 -61.511 55.154
479661 90.211 -61.540 55.165
479662 90.260 -61.567 55.174
479663 90.300 -61.574 55.177
479664 90.341 -61.581 55.175
479665 90.382 -61.588 55.168
479666 90.427 -61.600 55.157
479667 90.472 -61.613 55.143
479668 90.517 -61.626 55.125
479669 90.563 -61.640 55.103
479670 90.608 -61.653 55.079
479671 90.653 -61.666 55.059
479672 90.698 -61.678 55.045
479673 90.744 -61.691 55.036
479674 90.794 -61.706 55.030
479675 90.839 -61.721 55.028
479676 90.888 -61.728 55.029
479677 90.936 -61.735 55.034
479678 90.984 -61.744 55.045
479679 91.032 -61.752 55.061
479680 91.080 -61.760 55.078
479681 91.124 -61.771 55.092
479682 91.167 -61.783 55.100
479683 91.209 -61.795 55.101
479684 91.251 -61.807 55.091
479685 91.294 -61.819 55.086
479686 91.336 -61.829 55.086
479687 91.378 -61.838 55.087
479688 91.420 -61.848 55.084
479689 91.462 -61.857 55.080
479690 91.504 -61.867 55.071
479691 91.546 -61.876 55.068
479692 91.588 -61.885 55.071
479693 91.630 -61.896 55.070
479694 91.670 -61.908 55.065
479695 91.708 -61.927 55.057
479696 91.745 -61.945 55.039
479697 91.783 -61.964 55.013
479698 91.821 -61.982 54.981
479699 91.859 -62.001 54.947
479700 91.896 -62.019 54.924
479701 91.934 -62.037 54.908
479702 91.974 -62.055 54.892
479703 92.018 -62.070 54.866
479704 92.063 -62.086 54.827
479705 92.104 -62.092 54.794
479706 92.143 -62.092 54.768
479707 92.181 -62.091 54.748
479708 92.220 -62.090 54.733
479709 92.258 -62.090 54.712
479710 92.294 -62.091 54.685
479711 92.329 -62.092 54.659
479712 92.364 -62.093 54.637
479713 92.399 -62.095 54.618
479714 92.447 -62.102 54.598
479715 92.496 -62.109 54.577
479716 92.545 -62.116 54.556
479717 92.593 -62.123 54.538
479718 92.642 -62.130 54.519
479719 92.691 -62.138 54.500
479720 92.739 -62.145 54.476
479721 92.788 -62.152 54.450
479722 92.809 -62.165 54.424
479723 92.834 -62.180 54.398
479724 92.878 -62.196 54.378
479725 92.922 -62.212 54.361
479726 92.967 -62.228 54.344
479727 93.009 -62.239 54.329
479728 93.051 -62.250 54.316
479729 93.093 -62.261 54.301
479730 93.136 -62.271 54.289
479731 93.179 -62.280 54.277
479732 93.222 -62.286 54.264
479733 93.266 -62.291 54.251
479734 93.310 -62.297 54.240
479735 93.351 -62.302 54.228
479736 93.392 -62.305 54.216
479737 93.432 -62.310 54.209
479738 93.473 -62.315 54.209
479739 93.514 -62.319 54.218
479740 93.555 -62.323 54.225
479741 93.595 -62.327 54.229
479742 93.635 -62.333 54.230
479743 93.671 -62.329 54.227
479744 93.706 -62.324 54.219
479745 93.743 -62.349 54.211
479746 93.780 -62.373 54.199
479747 93.818 -62.397 54.184
479748 93.855 -62.421 54.166
479749 93.896 -62.443 54.146
479750 93.954 -62.452 54.125
479751 94.010 -62.461 54.103
479752 94.063 -62.450 54.080
479753 94.114 -62.433 54.053
479754 94.160 -62.431 54.026
479755 94.205 -62.434 53.997
479756 94.249 -62.438 53.968
479757 94.293 -62.442 53.937
479758 94.338 -62.446 53.904
479759 94.382 -62.449 53.872
479760 94.426 -62.454 53.838
479761 94.473 -62.459 53.801
479762 94.527 -62.467 53.762
479763 94.581 -62.475 53.722
479764 94.636 -62.483 53.681
479765 94.658 -62.500 53.640
479766 94.677 -62.518 53.603
479767 94.696 -62.537 53.568
479768 94.715 -62.555 53.533
479769 94.735 -62.573 53.498
479770 94.754 -62.590 53.462
479771 94.787 -62.605 53.428
479772 94.832 -62.631 53.394
479773 94.877 -62.658 53.360
479774 94.922 -62.684 53.326
479775 94.948 -62.692 53.293
479776 94.935 -62.666 53.262
479777 94.962 -62.663 53.232
479778 95.000 -62.666 53.202
479779 95.038 -62.669 53.176
479780 95.076 -62.673 53.147
479781 95.114 -62.676 53.119
479782 95.151 -62.679 53.092
479783 95.189 -62.683 53.066
479784 95.228 -62.686 53.041
479785 95.265 -62.689 53.015
479786 95.313 -62.689 52.988
479787 95.364 -62.688 52.960
479788 95.414 -62.687 52.932
479789 95.464 -62.686 52.903
479790 95.515 -62.685 52.875
479791 95.565 -62.684 52.846
479792 95.615 -62.684 52.816
479793 95.656 -62.719 52.786
479794 95.697 -62.753 52.756
479795 95.741 -62.783 52.723
479796 95.802 -62.775 52.691
479797 95.871 -62.747 52.658
479798 95.914 -62.733 52.624
479799 95.957 -62.719 52.588
479800 96.001 -62.703 52.552
479801 96.044 -62.689 52.517
479802 96.088 -62.675 52.481
479803 96.131 -62.660 52.446
479804 96.174 -62.646 52.410
479805 96.218 -62.631 52.373
479806 96.268 -62.639 52.336
479807 96.323 -62.657 52.299
479808 96.352 -62.674 52.262
479809 96.410 -62.686 52.224
479810 96.449 -62.698 52.186
479811 96.488 -62.711 52.148
479812 96.527 -62.723 52.110
479813 96.566 -62.735 52.071
479814 96.603 -62.749 52.032
479815 96.636 -62.767 51.993
479816 96.668 -62.785 51.954
479817 96.701 -62.803 51.915
479818 96.734 -62.820 51.877
479819 96.766 -62.838 51.837
479820 96.802 -62.850 51.801
479821 96.839 -62.856 51.764
479822 96.876 -62.862 51.728
479823 96.914 -62.868 51.694
479824 96.951 -62.874 51.661
479825 96.989 -62.880 51.627
479826 97.026 -62.886 51.595
479827 97.063 -62.892 51.564
479828 97.100 -62.898 51.532
479829 97.139 -62.904 51.500
479830 97.178 -62.909 51.469
479831 97.218 -62.913 51.438
479832 97.257 -62.919 51.406
479833 97.297 -62.924 51.373
479834 97.336 -62.929 51.339
479835 97.376 -62.934 51.306
479836 97.416 -62.938 51.271
479837 97.455 -62.944 51.235
479838 97.494 -62.949 51.200
479839 97.534 -62.954 51.163
479840 97.575 -62.957 51.125
479841 97.601 -62.949 51.085
479842 97.616 -62.926 51.043
479843 97.631 -62.901 51.000
479844 97.646 -62.877 50.957
479845 97.674 -62.863 50.912
479846 97.708 -62.851 50.865
479847 97.741 -62.838 50.818
479848 97.774 -62.827 50.772
479849 97.807 -62.815 50.727
479850 97.840 -62.803 50.682
479851 97.874 -62.793 50.637
479852 97.928 -62.807 50.591
479853 97.982 -62.821 50.546
479854 98.035 -62.835 50.501
479855 98.089 -62.850 50.456
479856 98.143 -62.863 50.411
479857 98.162 -62.879 50.366
479858 98.173 -62.894 50.321
479859 98.184 -62.909 50.277
479860 98.195 -62.923 50.234
479861 98.206 -62.938 50.192
479862 98.227 -62.954 50.150
479863 98.251 -62.970 50.109
480310 87.869 -60.970 55.897
480311 87.906 -60.971 55.881
480312 87.942 -60.973 55.865
480313 87.979 -60.975 55.850
480314 88.015 -60.977 55.832
480315 88.052 -60.979 55.815
480316 88.089 -60.980 55.797
480317 88.126 -60.983 55.779
480318 88.163 -60.984 55.759
480319 88.200 -60.986 55.740
480320 88.249 -60.982 55.720
480321 88.297 -60.979 55.699
480322 88.358 -60.982 55.677
480323 88.425 -60.988 55.653
480324 88.492 -60.994 55.628
480325 88.567 -60.989 55.603
480326 88.643 -60.983 55.577
480327 88.719 -60.976 55.550
480328 88.793 -60.974 55.523
480329 88.838 -60.999 55.496
480330 88.883 -61.025 55.468
480331 88.927 -61.050 55.441
480332 88.972 -61.076 55.415
480333 89.015 -61.094 55.390
480334 89.058 -61.112 55.366
480335 89.101 -61.130 55.345
480336 89.144 -61.149 55.325
480337 89.188 -61.167 55.309
480338 89.231 -61.185 55.298
480339 89.274 -61.203 55.285
480340 89.317 -61.221 55.273
480341 89.360 -61.240 55.262
480342 89.403 -61.258 55.251
480343 89.447 -61.276 55.238
480344 89.489 -61.294 55.223
480345 89.533 -61.312 55.209
480346 89.569 -61.315 55.197
480347 89.605 -61.314 55.186
480348 89.640 -61.312 55.175
480349 89.675 -61.311 55.161
480350 89.710 -61.309 55.150
480351 89.745 -61.308 55.141
480352 89.780 -61.306 55.135
480353 89.815 -61.305 55.130
480354 89.850 -61.303 55.126
480355 89.885 -61.302 55.124
480356 89.930 -61.321 55.123
480357 89.987 -61.370 55.124
480358 90.041 -61.409 55.125
480359 90.091 -61.438 55.125
480360 90.141 -61.467 55.127
480361 90.190 -61.496 55.134
480362 90.238 -61.519 55.144
480363 90.279 -61.525 55.153
480364 90.319 -61.532 55.158
480365 90.360 -61.540 55.160
480366 90.402 -61.549 55.157
480367 90.447 -61.562 55.151
480368 90.493 -61.575 55.140
480369 90.538 -61.588 55.123
480370 90.583 -61.601 55.101
480371 90.628 -61.614 55.078
480372 90.674 -61.627 55.057
480373 90.719 -61.640 55.044
480374 90.764 -61.653 55.036
480375 90.809 -61.664 55.030
480376 90.855 -61.672 55.027
480377 90.903 -61.681 55.027
480378 90.951 -61.688 55.030
480379 91.000 -61.697 55.039
480380 91.047 -61.705 55.055
480381 91.091 -61.716 55.073
480382 91.134 -61.729 55.088
480383 91.176 -61.741 55.099
480384 91.219 -61.753 55.101
480385 91.261 -61.765 55.091
480386 91.303 -61.777 55.087
480387 91.346 -61.787 55.088
480388 91.387 -61.796 55.091
480389 91.430 -61.806 55.090
480390 91.471 -61.815 55.086
480391 91.514 -61.824 55.078
480392 91.556 -61.834 55.073
480393 91.597 -61.845 55.074
480394 91.639 -61.855 55.070
480395 91.680 -61.867 55.070
480396 91.718 -61.884 55.061
480397 91.756 -61.902 55.043
480398 91.793 -61.921 55.017
480399 91.831 -61.939 54.983
480400 91.869 -61.957 54.952
480401 91.907 -61.976 54.934
480402 91.945 -61.994 54.920
480403 91.990 -62.010 54.901
480404 92.035 -62.025 54.870
480405 92.080 -62.040 54.829
480406 92.125 -62.056 54.795
480407 92.164 -62.058 54.770
480408 92.203 -62.057 54.751
480409 92.241 -62.056 54.735
480410 92.278 -62.056 54.712
480411 92.313 -62.058 54.688
480412 92.348 -62.059 54.666
480413 92.383 -62.061 54.650
480414 92.418 -62.061 54.635
480415 92.454 -62.064 54.616
480416 92.503 -62.071 54.592
480417 92.552 -62.078 54.568
480418 92.600 -62.085 54.545
480419 92.649 -62.092 54.521
480420 92.698 -62.099 54.496
480421 92.746 -62.107 54.469
480422 92.794 -62.114 54.442
480423 92.810 -62.128 54.416
480424 92.846 -62.144 54.392
480425 92.890 -62.160 54.374
480426 92.934 -62.176 54.357
480427 92.979 -62.192 54.339
480428 93.022 -62.208 54.321
480429 93.066 -62.221 54.306
480430 93.108 -62.231 54.288
480431 93.151 -62.239 54.273
480432 93.195 -62.245 54.260
480433 93.238 -62.251 54.247
480434 93.282 -62.257 54.234
480435 93.326 -62.262 54.221
480436 93.366 -62.267 54.210
480437 93.407 -62.271 54.200
480438 93.448 -62.275 54.195
480439 93.489 -62.280 54.197
480440 93.530 -62.283 54.206
480441 93.567 -62.287 54.212
480442 93.604 -62.293 54.215
480443 93.641 -62.305 54.215
480444 93.680 -62.316 54.212
480445 93.715 -62.312 54.204
480446 93.752 -62.317 54.195
480447 93.789 -62.341 54.182
480448 93.826 -62.365 54.167
480449 93.863 -62.390 54.149
480450 93.908 -62.408 54.129
480451 93.965 -62.417 54.110
480452 94.020 -62.417 54.090
480453 94.071 -62.399 54.068
480454 94.118 -62.396 54.044
480455 94.162 -62.400 54.019
480456 94.206 -62.404 53.992
480457 94.250 -62.409 53.963
480458 94.294 -62.412 53.932
480459 94.339 -62.416 53.899
480460 94.383 -62.420 53.865
480461 94.427 -62.424 53.830
480462 94.473 -62.428 53.792
480463 94.528 -62.437 53.753
480464 94.582 -62.445 53.713
480465 94.617 -62.459 53.671
480466 94.636 -62.477 53.631
480467 94.655 -62.495 53.594
480468 94.675 -62.513 53.558
480469 94.694 -62.531 53.521
480470 94.736 -62.532 53.482
480471 94.771 -62.537 53.445
480472 94.786 -62.558 53.411
480473 94.831 -62.584 53.377
480474 94.876 -62.611 53.342
480475 94.921 -62.637 53.309
480476 94.941 -62.640 53.276
480477 94.930 -62.615 53.244
480478 94.968 -62.618 53.214
480479 95.005 -62.622 53.186
480480 95.044 -62.625 53.159
480481 95.082 -62.628 53.130
480482 95.119 -62.631 53.102
480483 95.157 -62.634 53.075
480484 95.195 -62.637 53.049
480485 95.233 -62.641 53.025
480486 95.271 -62.644 52.999
480487 95.309 -62.647 52.972
480488 95.346 -62.650 52.945
480489 95.397 -62.650 52.918
480490 95.447 -62.648 52.891
480491 95.498 -62.647 52.864
480492 95.548 -62.647 52.837
480493 95.599 -62.645 52.808
480494 95.642 -62.669 52.778
480495 95.683 -62.704 52.748
480496 95.725 -62.739 52.715
480497 95.771 -62.760 52.683
480498 95.841 -62.732 52.651
480499 95.910 -62.703 52.617
480500 95.958 -62.687 52.582
480501 96.002 -62.672 52.546
480502 96.045 -62.657 52.511
480503 96.089 -62.643 52.476
480504 96.132 -62.628 52.440
480505 96.176 -62.614 52.404
480506 96.219 -62.599 52.366
480507 96.272 -62.616 52.328
480508 96.327 -62.634 52.291
480509 96.357 -62.650 52.253
480510 96.413 -62.663 52.215
480511 96.452 -62.675 52.177
480512 96.490 -62.687 52.138
480513 96.530 -62.699 52.099
480514 96.564 -62.716 52.059
480515 96.597 -62.733 52.019
480516 96.629 -62.750 51.980
480517 96.662 -62.769 51.940
480518 96.696 -62.784 51.901
480519 96.733 -62.790 51.862
480520 96.771 -62.796 51.822
480521 96.808 -62.802 51.784
480522 96.846 -62.808 51.747
480523 96.883 -62.814 51.711
480524 96.920 -62.820 51.677
480525 96.957 -62.826 51.642
480526 96.995 -62.832 51.609
480527 97.032 -62.838 51.576
480528 97.070 -62.844 51.545
480529 97.107 -62.851 51.513
480530 97.146 -62.856 51.481
480531 97.186 -62.861 51.449
480532 97.225 -62.866 51.418
480533 97.264 -62.871 51.385
480534 97.304 -62.876 51.352
480535 97.344 -62.881 51.319
480536 97.383 -62.886 51.285
480537 97.423 -62.891 51.250
480538 97.462 -62.896 51.214
480539 97.502 -62.901 51.179
480540 97.541 -62.906 51.143
480541 97.577 -62.912 51.105
480542 97.608 -62.905 51.065
480543 97.624 -62.881 51.024
480544 97.638 -62.857 50.982
480545 97.653 -62.833 50.940
480546 97.682 -62.819 50.895
480547 97.716 -62.807 50.849
480548 97.749 -62.795 50.803
480549 97.782 -62.783 50.757
480550 97.815 -62.771 50.712
480551 97.848 -62.759 50.667
480552 97.888 -62.757 50.622
480553 97.942 -62.771 50.577
480554 97.996 -62.785 50.532
480555 98.049 -62.799 50.487
480556 98.103 -62.813 50.442
480557 98.144 -62.828 50.398
480558 98.155 -62.843 50.354
480559 98.165 -62.858 50.310
480560 98.176 -62.873 50.266
480561 98.187 -62.888 50.224
480562 98.205 -62.904 50.182
480563 98.229 -62.919 50.141
481011 87.884 -60.922 55.887
481012 87.921 -60.923 55.871
481013 87.958 -60.926 55.855
481014 87.994 -60.927 55.839
481015 88.031 -60.929 55.823
481016 88.067 -60.931 55.806
481017 88.104 -60.933 55.788
481018 88.141 -60.934 55.769
481019 88.178 -60.936 55.750
481020 88.214 -60.938 55.731
481021 88.258 -60.936 55.712
481022 88.307 -60.933 55.690
481023 88.369 -60.937 55.667
481024 88.436 -60.943 55.643
481025 88.509 -60.942 55.619
481026 88.585 -60.936 55.593
481027 88.661 -60.929 55.567
481028 88.737 -60.923 55.540
481029 88.808 -60.923 55.513
481030 88.852 -60.949 55.485
481031 88.897 -60.974 55.457
481032 88.942 -61.000 55.431
481033 88.985 -61.022 55.405
481034 89.029 -61.040 55.379
481035 89.072 -61.058 55.355
481036 89.115 -61.077 55.333
481037 89.158 -61.095 55.313
481038 89.201 -61.114 55.297
481039 89.244 -61.132 55.284
481040 89.287 -61.150 55.274
481041 89.331 -61.168 55.264
481042 89.374 -61.187 55.256
481043 89.417 -61.205 55.248
481044 89.460 -61.223 55.237
481045 89.503 -61.241 55.224
481046 89.546 -61.259 55.210
481047 89.590 -61.277 55.194
481048 89.626 -61.279 55.181
481049 89.661 -61.277 55.169
481050 89.696 -61.276 55.155
481051 89.731 -61.274 55.143
481052 89.766 -61.273 55.132
481053 89.801 -61.271 55.125
481054 89.836 -61.270 55.119
481055 89.872 -61.268 55.114
481056 89.906 -61.267 55.112
481057 89.954 -61.293 55.111
481058 90.012 -61.343 55.111
481059 90.071 -61.393 55.112
481060 90.120 -61.423 55.112
481061 90.170 -61.453 55.113
481062 90.216 -61.470 55.116
481063 90.256 -61.477 55.121
481064 90.297 -61.484 55.128
481065 90.337 -61.491 55.135
481066 90.378 -61.498 55.141
481067 90.423 -61.510 55.142
481068 90.468 -61.523 55.140
481069 90.513 -61.536 55.132
481070 90.558 -61.550 55.117
481071 90.604 -61.563 55.097
481072 90.649 -61.575 55.074
481073 90.694 -61.588 55.055
481074 90.739 -61.600 55.043
481075 90.782 -61.609 55.038
481076 90.825 -61.617 55.032
481077 90.871 -61.625 55.028
481078 90.919 -61.634 55.025
481079 90.967 -61.642 55.027
481080 91.015 -61.650 55.034
481081 91.059 -61.661 55.049
481082 91.101 -61.674 55.067
481083 91.143 -61.686 55.084
481084 91.186 -61.698 55.097
481085 91.228 -61.710 55.101
481086 91.271 -61.722 55.092
481087 91.313 -61.735 55.087
481088 91.355 -61.745 55.087
481089 91.397 -61.754 55.091
481090 91.439 -61.763 55.093
481091 91.481 -61.773 55.090
481092 91.523 -61.782 55.083
481093 91.565 -61.793 55.079
481094 91.606 -61.804 55.079
481095 91.648 -61.815 55.073
481096 91.689 -61.826 55.070
481097 91.728 -61.841 55.061
481098 91.766 -61.860 55.043
481099 91.804 -61.878 55.017
481100 91.842 -61.897 54.987
481101 91.879 -61.915 54.959
481102 91.917 -61.933 54.943
481103 91.962 -61.948 54.928
481104 92.007 -61.964 54.907
481105 92.052 -61.979 54.871
481106 92.096 -61.995 54.830
481107 92.141 -62.011 54.797
481108 92.185 -62.023 54.772
481109 92.223 -62.023 54.753
481110 92.262 -62.022 54.736
481111 92.297 -62.023 54.712
481112 92.333 -62.024 54.690
481113 92.367 -62.026 54.674
481114 92.403 -62.027 54.664
481115 92.437 -62.028 54.653
481116 92.473 -62.029 54.633
481117 92.511 -62.033 54.608
481118 92.560 -62.040 54.581
481119 92.609 -62.047 54.556
481120 92.657 -62.054 54.527
481121 92.706 -62.061 54.497
481122 92.755 -62.068 54.467
481123 92.795 -62.077 54.438
481124 92.814 -62.092 54.411
481125 92.859 -62.108 54.387
481126 92.903 -62.123 54.369
481127 92.947 -62.139 54.351
481128 92.991 -62.156 54.333
481129 93.035 -62.172 54.313
481130 93.079 -62.188 54.295
481131 93.123 -62.199 54.278
481132 93.167 -62.205 54.262
481133 93.211 -62.210 54.247
481134 93.255 -62.216 54.233
481135 93.298 -62.222 54.219
481136 93.341 -62.227 54.206
481137 93.382 -62.232 54.194
481138 93.423 -62.236 54.186
481139 93.464 -62.240 54.181
481140 93.501 -62.243 54.181
481141 93.535 -62.244 54.186
481142 93.572 -62.253 54.188
481143 93.610 -62.265 54.190
481144 93.648 -62.276 54.189
481145 93.687 -62.287 54.186
481146 93.724 -62.299 54.178
481147 93.761 -62.296 54.169
481148 93.797 -62.309 54.157
481149 93.834 -62.333 54.142
481150 93.871 -62.358 54.125
481151 93.920 -62.374 54.109
481152 93.977 -62.383 54.093
481153 94.028 -62.366 54.075
481154 94.075 -62.363 54.055
481155 94.119 -62.367 54.034
481156 94.163 -62.371 54.012
481157 94.207 -62.374 53.986
481158 94.252 -62.378 53.958
481159 94.296 -62.382 53.927
481160 94.340 -62.387 53.893
481161 94.384 -62.390 53.858
481162 94.428 -62.394 53.821
481163 94.474 -62.399 53.782
481164 94.529 -62.407 53.743
481165 94.576 -62.417 53.703
481166 94.596 -62.435 53.661
481167 94.615 -62.453 53.621
481168 94.634 -62.471 53.583
481169 94.672 -62.475 53.546
481170 94.722 -62.469 53.507
481171 94.762 -62.472 53.468
481172 94.778 -62.493 53.430
481173 94.793 -62.512 53.394
481174 94.830 -62.537 53.360
481175 94.876 -62.564 53.327
481176 94.920 -62.590 53.293
481177 94.933 -62.587 53.260
481178 94.935 -62.569 53.228
481179 94.973 -62.573 53.199
481180 95.011 -62.576 53.170
481181 95.049 -62.579 53.143
481182 95.087 -62.583 53.116
481183 95.125 -62.586 53.089
481184 95.163 -62.589 53.063
481185 95.200 -62.592 53.038
481186 95.239 -62.596 53.013
481187 95.276 -62.599 52.987
481188 95.314 -62.602 52.962
481189 95.352 -62.606 52.936
481190 95.390 -62.609 52.910
481191 95.430 -62.611 52.885
481192 95.481 -62.610 52.858
481193 95.531 -62.609 52.830
481194 95.582 -62.608 52.802
481195 95.628 -62.620 52.772
481196 95.670 -62.655 52.741
481197 95.710 -62.690 52.710
481198 95.777 -62.716 52.678
481199 95.822 -62.710 52.645
481200 95.880 -62.688 52.613
481201 95.949 -62.660 52.579
481202 96.003 -62.640 52.543
481203 96.046 -62.625 52.508
481204 96.090 -62.611 52.473
481205 96.133 -62.597 52.437
481206 96.177 -62.582 52.401
481207 96.222 -62.574 52.362
481208 96.276 -62.592 52.323
481209 96.331 -62.611 52.284
481210 96.368 -62.627 52.246
481211 96.415 -62.638 52.208
481212 96.455 -62.650 52.170
481213 96.493 -62.664 52.130
481214 96.525 -62.681 52.089
481215 96.558 -62.699 52.048
481216 96.591 -62.717 52.007
481217 96.628 -62.724 51.967
481218 96.665 -62.730 51.927
481219 96.702 -62.736 51.887
481220 96.740 -62.743 51.847
481221 96.777 -62.749 51.807
481222 96.814 -62.755 51.768
481223 96.852 -62.761 51.731
481224 96.889 -62.767 51.695
481225 96.927 -62.773 51.659
481226 96.964 -62.779 51.625
481227 97.001 -62.785 51.591
481228 97.038 -62.791 51.559
481229 97.076 -62.797 51.526
481230 97.114 -62.803 51.494
481231 97.154 -62.808 51.462
481232 97.193 -62.813 51.431
481233 97.232 -62.818 51.398
481234 97.272 -62.823 51.365
481235 97.311 -62.828 51.332
481236 97.351 -62.833 51.300
481237 97.390 -62.838 51.265
481238 97.430 -62.843 51.230
481239 97.469 -62.848 51.195
481240 97.509 -62.853 51.159
481241 97.544 -62.860 51.123
481242 97.572 -62.869 51.085
481243 97.616 -62.861 51.045
481244 97.631 -62.837 51.005
481245 97.646 -62.813 50.964
481246 97.660 -62.789 50.922
481247 97.690 -62.775 50.877
481248 97.724 -62.763 50.833
481249 97.756 -62.751 50.788
481250 97.790 -62.739 50.743
481251 97.823 -62.727 50.698
481252 97.856 -62.716 50.654
481253 97.902 -62.720 50.609
481254 97.956 -62.735 50.564
481255 98.010 -62.749 50.519
481256 98.063 -62.763 50.474
481257 98.117 -62.777 50.430
481258 98.136 -62.792 50.387
481259 98.147 -62.807 50.344
481260 98.158 -62.822 50.300
481261 98.169 -62.837 50.257
481262 98.182 -62.852 50.214
481712 87.899 -60.874 55.878
481713 87.936 -60.876 55.862
481714 87.973 -60.877 55.845
481715 88.009 -60.879 55.830
481716 88.046 -60.881 55.814
481717 88.083 -60.883 55.796
481718 88.120 -60.885 55.778
481719 88.156 -60.886 55.760
481720 88.193 -60.889 55.741
481721 88.229 -60.890 55.721
481722 88.268 -60.892 55.702
481723 88.316 -60.888 55.681
481724 88.380 -60.892 55.659
481725 88.450 -60.895 55.635
481726 88.526 -60.888 55.610
481727 88.602 -60.881 55.584
481728 88.678 -60.875 55.558
481729 88.753 -60.870 55.531
481730 88.822 -60.873 55.504
481731 88.867 -60.898 55.475
481732 88.911 -60.924 55.447
481733 88.956 -60.949 55.421
481734 88.999 -60.969 55.395
481735 89.042 -60.987 55.369
481736 89.086 -61.005 55.345
481737 89.128 -61.024 55.322
481738 89.172 -61.042 55.302
481739 89.215 -61.060 55.284
481740 89.258 -61.078 55.270
481741 89.301 -61.096 55.261
481742 89.344 -61.114 55.254
481743 89.387 -61.133 55.249
481744 89.430 -61.152 55.244
481745 89.474 -61.170 55.237
481746 89.517 -61.188 55.225
481747 89.560 -61.206 55.211
481748 89.603 -61.224 55.195
481749 89.647 -61.243 55.179
481750 89.682 -61.243 55.163
481751 89.718 -61.242 55.149
481752 89.753 -61.240 55.136
481753 89.788 -61.239 55.125
481754 89.823 -61.237 55.116
481755 89.858 -61.236 55.108
481756 89.893 -61.234 55.104
481757 89.928 -61.233 55.100
481758 89.979 -61.267 55.098
481759 90.037 -61.316 55.098
481760 90.095 -61.365 55.098
481761 90.150 -61.409 55.099
481762 90.193 -61.422 55.099
481763 90.234 -61.428 55.100
481764 90.274 -61.436 55.102
481765 90.315 -61.443 55.106
481766 90.356 -61.450 55.114
481767 90.398 -61.459 55.122
481768 90.444 -61.472 55.127
481769 90.489 -61.484 55.127
481770 90.534 -61.497 55.123
481771 90.580 -61.511 55.110
481772 90.625 -61.524 55.092
481773 90.669 -61.536 55.072
481774 90.712 -61.544 55.054
481775 90.755 -61.553 55.044
481776 90.798 -61.562 55.040
481777 90.840 -61.571 55.035
481778 90.886 -61.579 55.030
481779 90.935 -61.587 55.027
481780 90.982 -61.595 55.026
481781 91.026 -61.606 55.030
481782 91.068 -61.619 55.042
481783 91.110 -61.632 55.059
481784 91.153 -61.644 55.078
481785 91.195 -61.656 55.094
481786 91.238 -61.668 55.100
481787 91.280 -61.680 55.094
481788 91.322 -61.692 55.088
481789 91.365 -61.703 55.085
481790 91.406 -61.713 55.089
481791 91.449 -61.722 55.093
481792 91.491 -61.732 55.092
481793 91.532 -61.741 55.088
481794 91.574 -61.752 55.085
481795 91.615 -61.763 55.084
481796 91.657 -61.774 55.077
481797 91.699 -61.785 55.070
481798 91.739 -61.798 55.058
481799 91.776 -61.817 55.042
481800 91.814 -61.835 55.021
481801 91.852 -61.854 54.995
481802 91.889 -61.872 54.968
481803 91.933 -61.888 54.952
481804 91.979 -61.904 54.936
481805 92.023 -61.919 54.911
481806 92.068 -61.934 54.873
481807 92.113 -61.950 54.831
481808 92.158 -61.965 54.797
481809 92.203 -61.981 54.772
481810 92.244 -61.988 54.755
481811 92.282 -61.988 54.737
481812 92.317 -61.989 54.713
481813 92.352 -61.991 54.695
481814 92.387 -61.992 54.685
481815 92.422 -61.994 54.679
481816 92.457 -61.995 54.668
481817 92.492 -61.996 54.648
481818 92.526 -61.998 54.623
481819 92.567 -62.001 54.597
481820 92.616 -62.009 54.572
481821 92.664 -62.016 54.540
481822 92.713 -62.023 54.505
481823 92.762 -62.030 54.470
481824 92.797 -62.040 54.438
481825 92.827 -62.055 54.409
481826 92.871 -62.071 54.384
481827 92.915 -62.087 54.364
481828 92.959 -62.104 54.345
481829 93.003 -62.120 54.326
481830 93.047 -62.136 54.305
481831 93.086 -62.147 54.286
481832 93.131 -62.158 54.269
481833 93.179 -62.167 54.252
481834 93.227 -62.176 54.236
481835 93.271 -62.181 54.222
481836 93.314 -62.187 54.207
481837 93.357 -62.192 54.194
481838 93.398 -62.197 54.182
481839 93.435 -62.199 54.174
481840 93.469 -62.201 54.168
481841 93.503 -62.202 54.164
481842 93.541 -62.213 54.163
481843 93.579 -62.224 54.163
481844 93.617 -62.236 54.161
481845 93.655 -62.247 54.157
481846 93.693 -62.258 54.152
481847 93.731 -62.270 54.144
481848 93.769 -62.281 54.135
481849 93.805 -62.279 54.125
481850 93.843 -62.302 54.113
481851 93.880 -62.326 54.103
481852 93.931 -62.339 54.092
481853 93.985 -62.332 54.078
481854 94.032 -62.329 54.062
481855 94.076 -62.333 54.043
481856 94.120 -62.337 54.023
481857 94.165 -62.341 54.004
481858 94.209 -62.345 53.980
481859 94.253 -62.349 53.952
481860 94.297 -62.352 53.921
481861 94.341 -62.356 53.887
481862 94.386 -62.360 53.849
481863 94.430 -62.365 53.812
481864 94.475 -62.368 53.773
481865 94.530 -62.377 53.733
481866 94.555 -62.393 53.691
481867 94.574 -62.411 53.650
481868 94.608 -62.418 53.609
481869 94.659 -62.412 53.571
481870 94.709 -62.407 53.533
481871 94.753 -62.406 53.492
481872 94.769 -62.427 53.454
481873 94.784 -62.447 53.415
481874 94.800 -62.468 53.379
481875 94.829 -62.490 53.345
481876 94.875 -62.517 53.312
481877 94.919 -62.543 53.278
481878 94.926 -62.534 53.246
481879 94.941 -62.525 53.215
481880 94.979 -62.528 53.185
481881 95.017 -62.531 53.157
481882 95.055 -62.534 53.132
481883 95.093 -62.537 53.105
481884 95.130 -62.540 53.079
481885 95.168 -62.543 53.055
481886 95.206 -62.547 53.030
481887 95.244 -62.550 53.005
481888 95.282 -62.553 52.980
481889 95.320 -62.557 52.955
481890 95.357 -62.560 52.930
481891 95.396 -62.563 52.905
481892 95.433 -62.566 52.880
481893 95.471 -62.570 52.854
481894 95.514 -62.572 52.827
481895 95.564 -62.570 52.799
481896 95.615 -62.571 52.769
481897 95.656 -62.606 52.738
481898 95.716 -62.633 52.706
481899 95.791 -62.658 52.674
481900 95.839 -62.669 52.642
481901 95.881 -62.656 52.610
481902 95.923 -62.643 52.577
481903 95.988 -62.617 52.542
481904 96.047 -62.594 52.508
481905 96.091 -62.579 52.473
481906 96.134 -62.565 52.438
481907 96.178 -62.550 52.401
481908 96.226 -62.550 52.364
481909 96.281 -62.569 52.326
481910 96.335 -62.587 52.288
481911 96.380 -62.603 52.248
481912 96.418 -62.615 52.209
481913 96.454 -62.630 52.169
481914 96.486 -62.647 52.126
481915 96.522 -62.659 52.083
481916 96.559 -62.665 52.040
481917 96.597 -62.671 51.998
481918 96.634 -62.677 51.957
481919 96.671 -62.683 51.916
481920 96.709 -62.689 51.875
481921 96.746 -62.695 51.835
481922 96.784 -62.701 51.794
481923 96.821 -62.707 51.755
481924 96.858 -62.713 51.717
481925 96.895 -62.719 51.680
481926 96.933 -62.725 51.644
481927 96.970 -62.732 51.609
481928 97.008 -62.738 51.575
481929 97.045 -62.744 51.541
481930 97.082 -62.750 51.509
481931 97.121 -62.755 51.477
481932 97.161 -62.760 51.445
481933 97.200 -62.766 51.413
481934 97.240 -62.770 51.380
481935 97.279 -62.775 51.347
481936 97.318 -62.780 51.315
481937 97.358 -62.785 51.281
481938 97.397 -62.791 51.246
481939 97.437 -62.795 51.212
481940 97.476 -62.800 51.176
481941 97.510 -62.807 51.139
481942 97.538 -62.817 51.103
481943 97.567 -62.826 51.064
481944 97.623 -62.817 51.025
481945 97.638 -62.793 50.985
481946 97.653 -62.769 50.944
481947 97.668 -62.744 50.904
481948 97.699 -62.731 50.860
481949 97.731 -62.719 50.817
481950 97.764 -62.707 50.774
481951 97.798 -62.695 50.730
481952 97.830 -62.684 50.686
481953 97.864 -62.672 50.642
481954 97.916 -62.684 50.597
481955 97.970 -62.699 50.552
481956 98.023 -62.713 50.508
481957 98.077 -62.727 50.464
481958 98.118 -62.741 50.420
481959 98.129 -62.757 50.378
481960 98.140 -62.772 50.334
481961 98.151 -62.787 50.291
482413 87.915 -60.826 55.868
482414 87.952 -60.828 55.852
482415 87.988 -60.829 55.836
482416 88.025 -60.832 55.821
482417 88.061 -60.833 55.804
482418 88.098 -60.835 55.787
482419 88.135 -60.837 55.769
482420 88.172 -60.839 55.750
482421 88.208 -60.841 55.731
482422 88.245 -60.842 55.712
482423 88.281 -60.844 55.692
482424 88.326 -60.842 55.672
482425 88.391 -60.847 55.651
482426 88.467 -60.841 55.626
482427 88.543 -60.834 55.601
482428 88.619 -60.827 55.575
482429 88.695 -60.820 55.549
482430 88.770 -60.817 55.522
482431 88.836 -60.823 55.493
482432 88.881 -60.848 55.466
482433 88.925 -60.873 55.438
482434 88.970 -60.898 55.411
482435 89.013 -60.916 55.384
482436 89.056 -60.934 55.358
482437 89.099 -60.952 55.334
482438 89.142 -60.970 55.311
482439 89.185 -60.989 55.290
482440 89.229 -61.007 55.271
482441 89.271 -61.025 55.258
482442 89.315 -61.043 55.248
482443 89.358 -61.061 55.242
482444 89.401 -61.080 55.240
482445 89.444 -61.098 55.238
482446 89.487 -61.116 55.234
482447 89.531 -61.134 55.225
482448 89.574 -61.152 55.212
482449 89.617 -61.171 55.196
482450 89.660 -61.190 55.178
482451 89.703 -61.208 55.162
482452 89.739 -61.208 55.146
482453 89.774 -61.206 55.131
482454 89.809 -61.205 55.118
482455 89.844 -61.203 55.108
482456 89.879 -61.202 55.099
482457 89.914 -61.200 55.094
482458 89.949 -61.198 55.090
482459 90.004 -61.240 55.087
482460 90.061 -61.289 55.085
482461 90.119 -61.339 55.084
482462 90.170 -61.369 55.086
482463 90.212 -61.381 55.087
482464 90.252 -61.387 55.087
482465 90.293 -61.394 55.087
482466 90.334 -61.401 55.089
482467 90.374 -61.408 55.094
482468 90.419 -61.420 55.103
482469 90.464 -61.433 55.111
482470 90.510 -61.446 55.114
482471 90.555 -61.459 55.112
482472 90.600 -61.472 55.102
482473 90.643 -61.480 55.086
482474 90.685 -61.489 55.069
482475 90.728 -61.497 55.053
482476 90.771 -61.506 55.045
482477 90.814 -61.515 55.043
482478 90.857 -61.524 55.039
482479 90.902 -61.532 55.035
482480 90.950 -61.540 55.030
482481 90.993 -61.552 55.026
482482 91.035 -61.564 55.027
482483 91.078 -61.576 55.036
482484 91.120 -61.589 55.052
482485 91.162 -61.601 55.072
482486 91.205 -61.613 55.089
482487 91.247 -61.625 55.098
482488 91.290 -61.638 55.097
482489 91.332 -61.650 55.091
482490 91.374 -61.661 55.085
482491 91.416 -61.671 55.085
482492 91.458 -61.680 55.090
482493 91.500 -61.690 55.091
482494 91.542 -61.700 55.091
482495 91.583 -61.711 55.091
482496 91.625 -61.722 55.090
482497 91.666 -61.733 55.082
482498 91.708 -61.744 55.068
482499 91.749 -61.756 55.055
482500 91.787 -61.774 55.043
482501 91.824 -61.792 55.027
482502 91.862 -61.811 55.004
482503 91.905 -61.827 54.978
482504 91.950 -61.842 54.960
482505 91.995 -61.858 54.942
482506 92.040 -61.873 54.913
482507 92.085 -61.889 54.873
482508 92.130 -61.904 54.832
482509 92.174 -61.920 54.799
482510 92.220 -61.936 54.775
482511 92.265 -61.951 54.757
482512 92.301 -61.954 54.737
482513 92.336 -61.956 54.715
482514 92.371 -61.957 54.701
482515 92.406 -61.959 54.694
482516 92.441 -61.961 54.688
482517 92.476 -61.962 54.677
482518 92.511 -61.963 54.658
482519 92.546 -61.964 54.638
482520 92.581 -61.966 54.616
482521 92.624 -61.970 54.593
482522 92.672 -61.978 54.559
482523 92.721 -61.985 54.520
482524 92.770 -61.992 54.480
482525 92.798 -62.003 54.442
482526 92.839 -62.019 54.411
482527 92.883 -62.036 54.383
482528 92.927 -62.051 54.359
482529 92.971 -62.067 54.338
482530 93.013 -62.081 54.317
482531 93.053 -62.092 54.296
482532 93.092 -62.105 54.277
482533 93.137 -62.116 54.260
482534 93.183 -62.127 54.243
482535 93.231 -62.136 54.227
482536 93.280 -62.144 54.213
482537 93.329 -62.152 54.198
482538 93.368 -62.155 54.186
482539 93.403 -62.157 54.173
482540 93.437 -62.158 54.163
482541 93.472 -62.161 54.155
482542 93.510 -62.173 54.148
482543 93.548 -62.184 54.142
482544 93.586 -62.196 54.138
482545 93.624 -62.207 54.134
482546 93.662 -62.218 54.127
482547 93.700 -62.230 54.120
482548 93.738 -62.241 54.112
482549 93.776 -62.252 54.104
482550 93.815 -62.264 54.096
482551 93.851 -62.270 54.090
482552 93.888 -62.294 54.087
482553 93.942 -62.299 54.079
482554 93.989 -62.296 54.067
482555 94.033 -62.299 54.052
482556 94.077 -62.303 54.034
482557 94.121 -62.307 54.015
482558 94.166 -62.311 53.996
482559 94.210 -62.315 53.972
482560 94.254 -62.319 53.944
482561 94.298 -62.323 53.912
482562 94.342 -62.327 53.878
482563 94.387 -62.330 53.839
482564 94.431 -62.334 53.800
482565 94.476 -62.339 53.760
482566 94.514 -62.351 53.718
482567 94.545 -62.361 53.677
482568 94.595 -62.355 53.636
482569 94.646 -62.349 53.596
482570 94.696 -62.344 53.557
482571 94.745 -62.340 53.519
482572 94.760 -62.361 53.480
482573 94.775 -62.381 53.441
482574 94.790 -62.402 53.404
482575 94.806 -62.422 53.368
482576 94.829 -62.444 53.334
482577 94.874 -62.470 53.300
482578 94.919 -62.496 53.266
482579 94.918 -62.481 53.234
482580 94.947 -62.479 53.204
482581 94.984 -62.482 53.175
482582 95.022 -62.486 53.149
482583 95.060 -62.489 53.124
482584 95.098 -62.492 53.099
482585 95.136 -62.496 53.074
482586 95.174 -62.499 53.050
482587 95.211 -62.502 53.025
482588 95.250 -62.505 53.001
482589 95.287 -62.509 52.976
482590 95.325 -62.512 52.951
482591 95.363 -62.515 52.927
482592 95.401 -62.518 52.903
482593 95.439 -62.522 52.878
482594 95.477 -62.525 52.852
482595 95.515 -62.528 52.826
482596 95.552 -62.531 52.798
482597 95.598 -62.532 52.768
482598 95.647 -62.555 52.738
482599 95.734 -62.572 52.706
482600 95.804 -62.599 52.673
482601 95.856 -62.626 52.641
482602 95.898 -62.613 52.609
482603 95.941 -62.600 52.575
482604 95.983 -62.587 52.541
482605 96.027 -62.574 52.507
482606 96.092 -62.547 52.473
482607 96.135 -62.533 52.439
482608 96.179 -62.518 52.402
482609 96.230 -62.527 52.365
482610 96.284 -62.546 52.328
482611 96.338 -62.564 52.290
482612 96.382 -62.578 52.250
482613 96.416 -62.593 52.210
482614 96.454 -62.599 52.168
482615 96.491 -62.605 52.124
482616 96.528 -62.611 52.079
482617 96.566 -62.617 52.034
482618 96.603 -62.623 51.990
482619 96.641 -62.629 51.948
482620 96.678 -62.635 51.906
482621 96.715 -62.641 51.865
482622 96.752 -62.647 51.823
482623 96.790 -62.654 51.782
482624 96.827 -62.660 51.743
482625 96.865 -62.666 51.705
482626 96.902 -62.672 51.667
482627 96.939 -62.678 51.631
482628 96.976 -62.684 51.595
482629 97.014 -62.691 51.561
482630 97.051 -62.697 51.527
482631 97.089 -62.703 51.494
482632 97.129 -62.707 51.462
482633 97.168 -62.713 51.429
482634 97.208 -62.718 51.397
482635 97.247 -62.722 51.365
482636 97.286 -62.728 51.332
482637 97.326 -62.732 51.299
482638 97.365 -62.738 51.265
482639 97.404 -62.743 51.230
482640 97.444 -62.747 51.195
482641 97.477 -62.755 51.160
482642 97.505 -62.765 51.123
482643 97.533 -62.774 51.086
482644 97.565 -62.782 51.048
482645 97.631 -62.772 51.009
482646 97.645 -62.749 50.970
482647 97.660 -62.725 50.929
482648 97.675 -62.700 50.889
482649 97.706 -62.688 50.847
482650 97.739 -62.675 50.804
482651 97.773 -62.663 50.761
482652 97.805 -62.652 50.718
482653 97.839 -62.640 50.674
482654 97.877 -62.634 50.631
482655 97.930 -62.648 50.587
482656 97.984 -62.663 50.543
482657 98.037 -62.677 50.499
482658 98.091 -62.691 50.455
482659 98.110 -62.706 50.413
482660 98.121 -62.721 50.370
482661 98.132 -62.736 50.327
483114 87.930 -60.778 55.858
483115 87.967 -60.780 55.841
483116 88.003 -60.782 55.825
483117 88.040 -60.783 55.809
483118 88.076 -60.785 55.793
483119 88.113 -60.787 55.777
483120 88.150 -60.789 55.759
483121 88.187 -60.791 55.740
483122 88.223 -60.792 55.722
483123 88.260 -60.795 55.702
483124 88.297 -60.796 55.683
483125 88.335 -60.798 55.663
483126 88.409 -60.793 55.641
483127 88.485 -60.786 55.618
483128 88.561 -60.780 55.593
483129 88.637 -60.773 55.567
483130 88.713 -60.767 55.540
483131 88.786 -60.764 55.513
483132 88.850 -60.772 55.484
483133 88.895 -60.798 55.457
483134 88.940 -60.823 55.429
483135 88.983 -60.845 55.401
483136 89.026 -60.863 55.374
483137 89.069 -60.881 55.348
483138 89.113 -60.899 55.324
483139 89.156 -60.917 55.301
483140 89.199 -60.936 55.280
483141 89.242 -60.954 55.261
483142 89.285 -60.972 55.246
483143 89.328 -60.990 55.236
483144 89.372 -61.008 55.231
483145 89.415 -61.027 55.229
483146 89.458 -61.045 55.229
483147 89.501 -61.063 55.228
483148 89.544 -61.081 55.221
483149 89.587 -61.099 55.211
483150 89.630 -61.118 55.196
483151 89.674 -61.136 55.179
483152 89.717 -61.154 55.162
483153 89.760 -61.172 55.144
483154 89.795 -61.171 55.127
483155 89.830 -61.170 55.113
483156 89.865 -61.168 55.100
483157 89.900 -61.167 55.091
483158 89.936 -61.165 55.085
483159 89.970 -61.164 55.080
483160 90.028 -61.212 55.076
483161 90.086 -61.262 55.074
483162 90.141 -61.305 55.073
483163 90.185 -61.320 55.074
483164 90.229 -61.335 55.076
483165 90.271 -61.346 55.077
483166 90.312 -61.353 55.078
483167 90.352 -61.359 55.077
483168 90.394 -61.368 55.082
483169 90.440 -61.381 55.089
483170 90.485 -61.394 55.096
483171 90.530 -61.407 55.100
483172 90.573 -61.416 55.100
483173 90.616 -61.425 55.090
483174 90.659 -61.433 55.077
483175 90.701 -61.442 55.063
483176 90.744 -61.450 55.051
483177 90.787 -61.459 55.046
483178 90.830 -61.468 55.045
483179 90.873 -61.477 55.044
483180 90.918 -61.485 55.040
483181 90.960 -61.497 55.034
483182 91.002 -61.509 55.026
483183 91.045 -61.522 55.025
483184 91.087 -61.534 55.030
483185 91.130 -61.547 55.045
483186 91.172 -61.559 55.064
483187 91.214 -61.571 55.080
483188 91.256 -61.583 55.093
483189 91.299 -61.595 55.097
483190 91.342 -61.607 55.094
483191 91.384 -61.619 55.089
483192 91.426 -61.628 55.083
483193 91.468 -61.638 55.085
483194 91.509 -61.649 55.088
483195 91.551 -61.660 55.090
483196 91.592 -61.671 55.092
483197 91.634 -61.682 55.093
483198 91.675 -61.693 55.085
483199 91.717 -61.704 55.073
483200 91.758 -61.714 55.062
483201 91.797 -61.732 55.052
483202 91.835 -61.750 55.034
483203 91.877 -61.766 55.009
483204 91.922 -61.782 54.983
483205 91.966 -61.798 54.963
483206 92.012 -61.813 54.943
483207 92.057 -61.828 54.911
483208 92.101 -61.844 54.871
483209 92.144 -61.860 54.834
483210 92.187 -61.876 54.803
483211 92.229 -61.892 54.779
483212 92.274 -61.907 54.761
483213 92.317 -61.916 54.737
483214 92.354 -61.920 54.717
483215 92.390 -61.923 54.707
483216 92.426 -61.926 54.701
483217 92.460 -61.927 54.695
483218 92.496 -61.929 54.685
483219 92.530 -61.930 54.670
483220 92.565 -61.931 54.654
483221 92.600 -61.932 54.636
483222 92.635 -61.934 54.612
483223 92.680 -61.939 54.578
483224 92.728 -61.947 54.536
483225 92.777 -61.954 54.491
483226 92.807 -61.967 54.449
483227 92.851 -61.983 54.414
483228 92.895 -61.999 54.384
483229 92.939 -62.014 54.358
483230 92.979 -62.026 54.335
483231 93.018 -62.038 54.313
483232 93.058 -62.050 54.293
483233 93.098 -62.061 54.274
483234 93.144 -62.073 54.257
483235 93.190 -62.084 54.241
483236 93.235 -62.095 54.226
483237 93.284 -62.104 54.211
483238 93.333 -62.111 54.197
483239 93.370 -62.114 54.183
483240 93.405 -62.116 54.170
483241 93.440 -62.121 54.157
483242 93.479 -62.133 54.145
483243 93.516 -62.144 54.135
483244 93.555 -62.155 54.126
483245 93.593 -62.167 54.118
483246 93.631 -62.178 54.110
483247 93.669 -62.189 54.102
483248 93.707 -62.201 54.093
483249 93.745 -62.212 54.084
483250 93.783 -62.224 54.076
483251 93.821 -62.236 54.072
483252 93.859 -62.247 54.072
483253 93.898 -62.261 54.075
483254 93.945 -62.261 54.068
483255 93.990 -62.265 54.056
483256 94.034 -62.269 54.042
483257 94.078 -62.274 54.024
483258 94.123 -62.277 54.006
483259 94.167 -62.281 53.986
483260 94.211 -62.285 53.962
483261 94.255 -62.289 53.933
483262 94.299 -62.293 53.901
483263 94.344 -62.297 53.866
483264 94.389 -62.300 53.827
483265 94.435 -62.302 53.787
483266 94.481 -62.303 53.745
483267 94.532 -62.298 53.703
483268 94.582 -62.293 53.662
483269 94.632 -62.287 53.621
483270 94.683 -62.281 53.581
483271 94.734 -62.276 53.543
483272 94.751 -62.295 53.505
483273 94.766 -62.315 53.467
483274 94.782 -62.336 53.430
483275 94.797 -62.356 53.394
483276 94.813 -62.377 53.359
483277 94.828 -62.397 53.324
483278 94.873 -62.424 53.291
483279 94.918 -62.450 53.258
483280 94.914 -62.431 53.226
483281 94.952 -62.434 53.196
483282 94.990 -62.437 53.168
483283 95.028 -62.440 53.144
483284 95.066 -62.443 53.120
483285 95.104 -62.447 53.096
483286 95.141 -62.450 53.071
483287 95.179 -62.453 53.047
483288 95.217 -62.457 53.022
483289 95.255 -62.460 52.998
483290 95.293 -62.463 52.975
483291 95.331 -62.466 52.951
483292 95.369 -62.470 52.928
483293 95.407 -62.473 52.905
483294 95.445 -62.476 52.880
483295 95.482 -62.480 52.855
483296 95.520 -62.483 52.830
483297 95.558 -62.486 52.802
483298 95.598 -62.490 52.771
483299 95.655 -62.503 52.741
483300 95.738 -62.518 52.709
483301 95.818 -62.540 52.676
483302 95.872 -62.578 52.643
483303 95.915 -62.571 52.610
483304 95.958 -62.558 52.575
483305 96.000 -62.545 52.540
483306 96.042 -62.532 52.506
483307 96.085 -62.519 52.473
483308 96.135 -62.502 52.438
483309 96.180 -62.486 52.403
483310 96.234 -62.503 52.365
483311 96.288 -62.521 52.327
483312 96.340 -62.528 52.289
483313 96.385 -62.540 52.250
483314 96.423 -62.546 52.210
483315 96.460 -62.552 52.167
483316 96.498 -62.558 52.122
483317 96.535 -62.564 52.076
483318 96.572 -62.570 52.029
483319 96.609 -62.576 51.983
483320 96.647 -62.582 51.940
483321 96.684 -62.588 51.897
483322 96.722 -62.594 51.855
483323 96.759 -62.600 51.812
483324 96.796 -62.606 51.771
483325 96.833 -62.612 51.731
483326 96.871 -62.619 51.692
483327 96.908 -62.625 51.655
483328 96.946 -62.631 51.618
483329 96.983 -62.637 51.582
483330 97.020 -62.643 51.548
483331 97.058 -62.649 51.513
483332 97.096 -62.655 51.480
483333 97.136 -62.659 51.448
483334 97.175 -62.665 51.415
483335 97.215 -62.669 51.382
483336 97.254 -62.675 51.350
483337 97.294 -62.680 51.317
483338 97.333 -62.684 51.284
483339 97.372 -62.690 51.250
483340 97.412 -62.694 51.215
483341 97.443 -62.703 51.179
483342 97.471 -62.713 51.144
483343 97.499 -62.722 51.108
483344 97.527 -62.732 51.071
483345 97.567 -62.735 51.033
483346 97.634 -62.727 50.994
483347 97.653 -62.704 50.955
483348 97.668 -62.681 50.915
483349 97.682 -62.656 50.876
483350 97.714 -62.644 50.834
483351 97.747 -62.631 50.792
483352 97.780 -62.619 50.749
483353 97.813 -62.608 50.707
483354 97.847 -62.596 50.664
483355 97.890 -62.598 50.621
483356 97.944 -62.612 50.577
483357 97.998 -62.626 50.534
483358 98.052 -62.641 50.490
483359 98.092 -62.655 50.447
483360 98.103 -62.670 50.405
483361 98.114 -62.685 50.362
483362 98.125 -62.700 50.320
483815 87.945 -60.730 55.848
483816 87.982 -60.732 55.831
483817 88.019 -60.734 55.814
483818 88.055 -60.735 55.798
483819 88.092 -60.738 55.782
483820 88.129 -60.739 55.765
483821 88.165 -60.741 55.748
483822 88.202 -60.743 55.729
483823 88.239 -60.745 55.710
483824 88.275 -60.747 55.692
483825 88.312 -60.748 55.673
483826 88.358 -60.750 55.653
483827 88.426 -60.739 55.632
483828 88.502 -60.732 55.610
483829 88.578 -60.726 55.585
483830 88.654 -60.720 55.559
483831 88.730 -60.713 55.532
483832 88.803 -60.712 55.505
483833 88.864 -60.722 55.476
483834 88.909 -60.747 55.448
483835 88.954 -60.773 55.420
483836 88.997 -60.792 55.392
483837 89.040 -60.810 55.364
483838 89.083 -60.828 55.339
483839 89.126 -60.846 55.315
483840 89.169 -60.864 55.292
483841 89.212 -60.882 55.270
483842 89.256 -60.901 55.251
483843 89.299 -60.919 55.235
483844 89.342 -60.937 55.224
483845 89.385 -60.955 55.217
483846 89.428 -60.973 55.214
483847 89.471 -60.992 55.214
483848 89.515 -61.010 55.213
483849 89.558 -61.028 55.209
483850 89.601 -61.046 55.202
483851 89.644 -61.064 55.191
483852 89.687 -61.083 55.175
483853 89.730 -61.101 55.159
483854 89.773 -61.119 55.141
483855 89.816 -61.137 55.124
483856 89.851 -61.136 55.108
483857 89.887 -61.134 55.095
483858 89.921 -61.133 55.085
483859 89.957 -61.131 55.076
483860 89.994 -61.136 55.070
483861 90.052 -61.185 55.065
483862 90.111 -61.235 55.063
483863 90.157 -61.255 55.062
483864 90.200 -61.271 55.063
483865 90.244 -61.286 55.065
483866 90.289 -61.301 55.068
483867 90.330 -61.311 55.070
483868 90.371 -61.318 55.067
483869 90.415 -61.330 55.071
483870 90.460 -61.343 55.076
483871 90.504 -61.352 55.081
483872 90.546 -61.360 55.085
483873 90.589 -61.369 55.085
483874 90.632 -61.378 55.077
483875 90.674 -61.387 55.066
483876 90.717 -61.395 55.055
483877 90.759 -61.405 55.048
483878 90.801 -61.415 55.046
483879 90.843 -61.425 55.046
483880 90.886 -61.436 55.046
483881 90.928 -61.447 55.044
483882 90.969 -61.459 55.037
483883 91.012 -61.469 55.028
483884 91.054 -61.481 55.024
483885 91.097 -61.491 55.026
483886 91.139 -61.503 55.035
483887 91.181 -61.516 55.050
483888 91.224 -61.528 55.067
483889 91.266 -61.541 55.084
483890 91.309 -61.553 55.094
483891 91.351 -61.565 55.097
483892 91.393 -61.577 55.094
483893 91.436 -61.587 55.085
483894 91.477 -61.597 55.080
483895 91.519 -61.608 55.083
483896 91.560 -61.619 55.087
483897 91.602 -61.630 55.091
483898 91.643 -61.641 55.093
483899 91.684 -61.652 55.088
483900 91.726 -61.663 55.078
483901 91.768 -61.674 55.068
483902 91.807 -61.688 55.057
483903 91.849 -61.706 55.037
483904 91.893 -61.721 55.008
483905 91.938 -61.736 54.982
483906 91.982 -61.753 54.961
483907 92.024 -61.769 54.940
483908 92.066 -61.785 54.908
483909 92.109 -61.802 54.875
483910 92.151 -61.818 54.843
483911 92.194 -61.835 54.811
483912 92.237 -61.851 54.782
483913 92.285 -61.863 54.764
483914 92.332 -61.875 54.740
483915 92.368 -61.879 54.722
483916 92.405 -61.882 54.712
483917 92.441 -61.886 54.707
483918 92.478 -61.890 54.701
483919 92.514 -61.893 54.694
483920 92.550 -61.897 54.685
483921 92.585 -61.898 54.674
483922 92.620 -61.899 54.657
483923 92.655 -61.901 54.631
483924 92.690 -61.902 54.595
483925 92.736 -61.908 54.551
483926 92.785 -61.916 54.504
483927 92.819 -61.931 54.457
483928 92.863 -61.947 54.418
483929 92.905 -61.960 54.386
483930 92.944 -61.972 54.360
483931 92.984 -61.983 54.335
483932 93.024 -61.995 54.313
483933 93.064 -62.007 54.293
483934 93.104 -62.018 54.275
483935 93.150 -62.029 54.258
483936 93.196 -62.041 54.243
483937 93.242 -62.052 54.228
483938 93.288 -62.063 54.214
483939 93.337 -62.071 54.200
483940 93.372 -62.073 54.185
483941 93.409 -62.080 54.170
483942 93.447 -62.092 54.155
483943 93.485 -62.104 54.140
483944 93.523 -62.115 54.127
483945 93.562 -62.127 54.116
483946 93.600 -62.138 54.106
483947 93.638 -62.149 54.097
483948 93.676 -62.161 54.088
483949 93.714 -62.172 54.078
483950 93.752 -62.183 54.068
483951 93.790 -62.193 54.059
483952 93.828 -62.203 54.053
483953 93.866 -62.213 54.052
483954 93.906 -62.223 54.053
483955 93.948 -62.230 54.047
483956 94.006 -62.220 54.037
483957 94.060 -62.212 54.024
483958 94.106 -62.214 54.009
483959 94.152 -62.216 53.993
483960 94.198 -62.218 53.972
483961 94.245 -62.219 53.949
483962 94.291 -62.221 53.921
483963 94.337 -62.223 53.888
483964 94.383 -62.224 53.851
483965 94.429 -62.227 53.812
483966 94.476 -62.228 53.771
483967 94.522 -62.230 53.730
483968 94.569 -62.230 53.688
483969 94.619 -62.224 53.645
483970 94.670 -62.219 53.605
483971 94.720 -62.214 53.567
483972 94.742 -62.230 53.529
483973 94.757 -62.249 53.492
483974 94.773 -62.270 53.455
483975 94.788 -62.290 53.419
483976 94.804 -62.311 53.384
483977 94.819 -62.331 53.350
483978 94.834 -62.352 53.317
483979 94.872 -62.377 53.283
483980 94.916 -62.402 53.251
483981 94.920 -62.386 53.220
483982 94.958 -62.389 53.191
483983 94.995 -62.392 53.163
483984 95.033 -62.395 53.140
483985 95.071 -62.399 53.117
483986 95.109 -62.402 53.093
483987 95.147 -62.405 53.069
483988 95.185 -62.409 53.044
483989 95.223 -62.412 53.020
483990 95.261 -62.415 52.997
483991 95.299 -62.418 52.974
483992 95.336 -62.421 52.952
483993 95.374 -62.424 52.930
483994 95.412 -62.428 52.907
483995 95.450 -62.431 52.884
483996 95.491 -62.435 52.860
483997 95.531 -62.440 52.834
483998 95.572 -62.444 52.807
483999 95.613 -62.448 52.776
484000 95.674 -62.462 52.744
484001 95.742 -62.483 52.712
484002 95.810 -62.503 52.679
484003 95.878 -62.525 52.645
484004 95.933 -62.528 52.612
484005 95.975 -62.515 52.576
484006 96.016 -62.502 52.540
484007 96.057 -62.487 52.505
484008 96.098 -62.473 52.470
484009 96.139 -62.459 52.435
484010 96.182 -62.450 52.400
484011 96.234 -62.456 52.362
484012 96.287 -62.463 52.322
484013 96.339 -62.471 52.283
484014 96.390 -62.479 52.243
484015 96.429 -62.498 52.202
484016 96.466 -62.504 52.160
484017 96.504 -62.510 52.115
484018 96.541 -62.516 52.069
484019 96.579 -62.522 52.022
484020 96.616 -62.528 51.976
484021 96.653 -62.534 51.931
484022 96.690 -62.540 51.887
484023 96.728 -62.547 51.843
484024 96.765 -62.553 51.800
484025 96.803 -62.559 51.759
484026 96.840 -62.565 51.719
484027 96.877 -62.572 51.680
484028 96.914 -62.578 51.642
484029 96.952 -62.584 51.605
484030 96.990 -62.590 51.569
484031 97.027 -62.596 51.535
484032 97.064 -62.602 51.501
484033 97.103 -62.607 51.467
484034 97.143 -62.612 51.434
484035 97.182 -62.617 51.402
484036 97.222 -62.622 51.369
484037 97.261 -62.627 51.336
484038 97.301 -62.632 51.303
484039 97.340 -62.637 51.270
484040 97.380 -62.642 51.235
484041 97.409 -62.650 51.200
484042 97.438 -62.660 51.165
484043 97.466 -62.669 51.129
484044 97.494 -62.679 51.093
484045 97.522 -62.688 51.055
484046 97.568 -62.690 51.018
484047 97.635 -62.681 50.980
484048 97.660 -62.660 50.941
484049 97.675 -62.636 50.902
484050 97.690 -62.612 50.863
484051 97.722 -62.600 50.822
484052 97.755 -62.587 50.780
484053 97.788 -62.576 50.738
484054 97.821 -62.564 50.696
484055 97.854 -62.552 50.654
484056 97.904 -62.562 50.611
484057 97.958 -62.576 50.568
484058 98.012 -62.590 50.525
484059 98.065 -62.605 50.482
484060 98.084 -62.619 50.438
484061 98.095 -62.634 50.396
484062 98.106 -62.650 50.355
484516 87.960 -60.682 55.837
484517 87.997 -60.684 55.821
484518 88.034 -60.686 55.803
484519 88.070 -60.688 55.787
484520 88.107 -60.689 55.770
484521 88.144 -60.691 55.753
484522 88.181 -60.693 55.736
484523 88.217 -60.695 55.718
484524 88.254 -60.697 55.699
484525 88.290 -60.698 55.681
484526 88.332 -60.702 55.662
484527 88.386 -60.702 55.643
484528 88.445 -60.686 55.623
484529 88.519 -60.679 55.602
484530 88.595 -60.672 55.578
484531 88.671 -60.665 55.552
484532 88.747 -60.660 55.524
484533 88.820 -60.659 55.497
484534 88.879 -60.671 55.469
484535 88.923 -60.697 55.441
484536 88.967 -60.720 55.412
484537 89.011 -60.738 55.384
484538 89.053 -60.756 55.356
484539 89.097 -60.775 55.330
484540 89.140 -60.793 55.306
484541 89.183 -60.811 55.283
484542 89.226 -60.829 55.261
484543 89.269 -60.848 55.241
484544 89.313 -60.866 55.224
484545 89.355 -60.884 55.211
484546 89.399 -60.902 55.202
484547 89.442 -60.920 55.197
484548 89.485 -60.939 55.195
484549 89.528 -60.957 55.194
484550 89.571 -60.975 55.194
484551 89.614 -60.993 55.191
484552 89.658 -61.011 55.183
484553 89.701 -61.030 55.169
484554 89.744 -61.048 55.153
484555 89.787 -61.066 55.135
484556 89.830 -61.084 55.118
484557 89.873 -61.101 55.104
484558 89.908 -61.099 55.090
484559 89.943 -61.098 55.078
484560 89.978 -61.096 55.068
484561 90.019 -61.108 55.061
484562 90.077 -61.158 55.056
484563 90.128 -61.190 55.053
484564 90.172 -61.206 55.052
484565 90.216 -61.221 55.052
484566 90.260 -61.237 55.055
484567 90.304 -61.252 55.059
484568 90.347 -61.267 55.063
484569 90.391 -61.278 55.061
484570 90.434 -61.287 55.062
484571 90.476 -61.296 55.064
484572 90.518 -61.307 55.067
484573 90.559 -61.318 55.069
484574 90.601 -61.328 55.068
484575 90.643 -61.338 55.063
484576 90.684 -61.348 55.055
484577 90.726 -61.359 55.048
484578 90.768 -61.369 55.045
484579 90.810 -61.379 55.046
484580 90.855 -61.390 55.048
484581 90.902 -61.401 55.048
484582 90.941 -61.413 55.046
484583 90.979 -61.425 55.040
484584 91.021 -61.436 55.031
484585 91.064 -61.447 55.025
484586 91.106 -61.458 55.025
484587 91.148 -61.469 55.029
484588 91.191 -61.480 55.038
484589 91.233 -61.491 55.053
484590 91.276 -61.501 55.068
484591 91.318 -61.512 55.081
484592 91.361 -61.523 55.089
484593 91.403 -61.534 55.093
484594 91.445 -61.546 55.088
484595 91.487 -61.556 55.078
484596 91.528 -61.568 55.077
484597 91.569 -61.578 55.082
484598 91.611 -61.589 55.088
484599 91.652 -61.600 55.090
484600 91.694 -61.612 55.089
484601 91.735 -61.622 55.082
484602 91.777 -61.634 55.073
484603 91.818 -61.645 55.060
484604 91.861 -61.662 55.036
484605 91.904 -61.678 55.004
484606 91.946 -61.694 54.978
484607 91.989 -61.711 54.957
484608 92.031 -61.728 54.936
484609 92.073 -61.744 54.908
484610 92.116 -61.760 54.883
484611 92.158 -61.777 54.854
484612 92.201 -61.793 54.820
484613 92.247 -61.807 54.786
484614 92.295 -61.819 54.765
484615 92.343 -61.831 54.743
484616 92.383 -61.838 54.728
484617 92.420 -61.841 54.720
484618 92.456 -61.845 54.717
484619 92.492 -61.849 54.713
484620 92.528 -61.853 54.709
484621 92.565 -61.856 54.701
484622 92.601 -61.860 54.688
484623 92.638 -61.863 54.669
484624 92.674 -61.867 54.641
484625 92.709 -61.869 54.606
484626 92.744 -61.870 54.563
484627 92.787 -61.879 54.515
484628 92.831 -61.894 54.466
484629 92.870 -61.906 54.424
484630 92.910 -61.917 54.392
484631 92.950 -61.929 54.364
484632 92.990 -61.941 54.340
484633 93.029 -61.952 54.318
484634 93.069 -61.964 54.299
484635 93.111 -61.976 54.283
484636 93.156 -61.987 54.266
484637 93.203 -61.998 54.250
484638 93.249 -62.009 54.235
484639 93.294 -62.020 54.221
484640 93.341 -62.031 54.206
484641 93.378 -62.040 54.189
484642 93.416 -62.051 54.172
484643 93.454 -62.061 54.154
484644 93.492 -62.071 54.136
484645 93.530 -62.081 54.121
484646 93.568 -62.091 54.107
484647 93.607 -62.101 54.096
484648 93.644 -62.111 54.085
484649 93.683 -62.120 54.075
484650 93.720 -62.130 54.064
484651 93.759 -62.139 54.053
484652 93.796 -62.149 54.043
484653 93.835 -62.159 54.035
484654 93.873 -62.170 54.030
484655 93.915 -62.181 54.027
484656 93.957 -62.189 54.020
484657 94.000 -62.197 54.010
484658 94.043 -62.204 53.999
484659 94.090 -62.205 53.986
484660 94.148 -62.194 53.972
484661 94.206 -62.182 53.953
484662 94.264 -62.170 53.929
484663 94.322 -62.158 53.901
484664 94.378 -62.149 53.868
484665 94.424 -62.151 53.831
484666 94.470 -62.152 53.792
484667 94.516 -62.155 53.751
484668 94.562 -62.156 53.710
484669 94.609 -62.158 53.668
484670 94.657 -62.156 53.628
484671 94.707 -62.151 53.590
484672 94.733 -62.164 53.553
484673 94.749 -62.184 53.517
484674 94.764 -62.205 53.480
484675 94.779 -62.225 53.444
484676 94.794 -62.245 53.409
484677 94.810 -62.265 53.375
484678 94.825 -62.286 53.341
484679 94.841 -62.306 53.309
484680 94.871 -62.330 53.277
484681 94.908 -62.349 53.246
484682 94.925 -62.340 53.215
484683 94.963 -62.343 53.187
484684 95.001 -62.347 53.161
484685 95.039 -62.350 53.137
484686 95.077 -62.353 53.116
484687 95.115 -62.356 53.091
484688 95.153 -62.360 53.066
484689 95.190 -62.363 53.042
484690 95.229 -62.366 53.019
484691 95.266 -62.370 52.996
484692 95.304 -62.373 52.975
484693 95.343 -62.376 52.954
484694 95.383 -62.381 52.932
484695 95.424 -62.385 52.909
484696 95.464 -62.389 52.887
484697 95.505 -62.393 52.864
484698 95.545 -62.398 52.839
484699 95.586 -62.402 52.811
484700 95.631 -62.404 52.781
484701 95.694 -62.421 52.748
484702 95.762 -62.441 52.715
484703 95.829 -62.462 52.681
484704 95.897 -62.484 52.646
484705 95.939 -62.472 52.612
484706 95.977 -62.457 52.576
484707 96.015 -62.442 52.539
484708 96.056 -62.428 52.503
484709 96.097 -62.413 52.467
484710 96.141 -62.406 52.431
484711 96.187 -62.402 52.396
484712 96.233 -62.399 52.357
484713 96.285 -62.406 52.317
484714 96.337 -62.412 52.277
484715 96.390 -62.420 52.236
484716 96.433 -62.436 52.195
484717 96.473 -62.456 52.153
484718 96.510 -62.463 52.108
484719 96.547 -62.469 52.062
484720 96.585 -62.475 52.015
484721 96.622 -62.481 51.968
484722 96.660 -62.487 51.923
484723 96.697 -62.493 51.878
484724 96.734 -62.500 51.834
484725 96.771 -62.506 51.790
484726 96.809 -62.512 51.748
484727 96.847 -62.518 51.708
484728 96.884 -62.524 51.669
484729 96.921 -62.530 51.631
484730 96.958 -62.536 51.594
484731 96.996 -62.542 51.557
484732 97.033 -62.548 51.522
484733 97.071 -62.554 51.488
484734 97.110 -62.559 51.455
484735 97.150 -62.564 51.422
484736 97.189 -62.569 51.389
484737 97.229 -62.574 51.356
484738 97.268 -62.579 51.323
484739 97.308 -62.584 51.289
484740 97.347 -62.589 51.256
484741 97.376 -62.598 51.222
484742 97.404 -62.608 51.186
484743 97.432 -62.617 51.151
484744 97.460 -62.627 51.116
484745 97.489 -62.636 51.079
484746 97.516 -62.646 51.042
484747 97.569 -62.644 51.004
484748 97.636 -62.635 50.966
484749 97.667 -62.616 50.927
484750 97.682 -62.592 50.889
484751 97.697 -62.568 50.849
484752 97.730 -62.556 50.809
484753 97.763 -62.544 50.768
484754 97.796 -62.532 50.726
484755 97.829 -62.520 50.685
484756 97.865 -62.512 50.645
484757 97.919 -62.526 50.602
484758 97.972 -62.540 50.560
484759 98.026 -62.554 50.517
484760 98.066 -62.569 50.473
484761 98.077 -62.584 50.431
484762 98.088 -62.599 50.389
484763 98.099 -62.614 50.347
484764 98.119 -62.629 50.306
485217 87.976 -60.635 55.826
485218 88.012 -60.636 55.810
485219 88.049 -60.638 55.791
485220 88.085 -60.640 55.775
485221 88.122 -60.641 55.759
485222 88.159 -60.644 55.742
485223 88.196 -60.645 55.725
485224 88.233 -60.647 55.706
485225 88.269 -60.649 55.688
485226 88.307 -60.651 55.670
485227 88.358 -60.658 55.651
485228 88.414 -60.654 55.633
485229 88.473 -60.638 55.613
485230 88.537 -60.625 55.594
485231 88.613 -60.618 55.570
485232 88.689 -60.611 55.544
485233 88.763 -60.607 55.517
485234 88.837 -60.606 55.490
485235 88.893 -60.621 55.462
485236 88.937 -60.646 55.434
485237 88.981 -60.666 55.405
485238 89.024 -60.685 55.376
485239 89.067 -60.703 55.349
485240 89.110 -60.721 55.323
485241 89.154 -60.739 55.298
485242 89.197 -60.757 55.274
485243 89.240 -60.776 55.251
485244 89.283 -60.795 55.230
485245 89.326 -60.813 55.213
485246 89.369 -60.831 55.199
485247 89.412 -60.849 55.187
485248 89.456 -60.867 55.179
485249 89.498 -60.886 55.174
485250 89.542 -60.904 55.172
485251 89.585 -60.922 55.175
485252 89.628 -60.940 55.176
485253 89.671 -60.958 55.172
485254 89.715 -60.977 55.161
485255 89.757 -60.995 55.144
485256 89.801 -61.013 55.126
485257 89.844 -61.031 55.111
485258 89.887 -61.049 55.099
485259 89.929 -61.065 55.086
485260 89.964 -61.064 55.072
485261 90.000 -61.062 55.061
485262 90.044 -61.082 55.053
485263 90.100 -61.126 55.047
485264 90.144 -61.141 55.044
485265 90.187 -61.156 55.042
485266 90.231 -61.172 55.041
485267 90.275 -61.187 55.043
485268 90.318 -61.204 55.047
485269 90.360 -61.224 55.053
485270 90.401 -61.240 55.055
485271 90.443 -61.250 55.055
485272 90.484 -61.261 55.054
485273 90.526 -61.271 55.053
485274 90.568 -61.281 55.052
485275 90.610 -61.292 55.050
485276 90.651 -61.302 55.048
485277 90.693 -61.312 55.045
485278 90.735 -61.323 55.042
485279 90.777 -61.333 55.043
485280 90.823 -61.344 55.046
485281 90.872 -61.355 55.050
485282 90.916 -61.366 55.050
485283 90.954 -61.378 55.047
485284 90.993 -61.390 55.042
485285 91.031 -61.403 55.035
485286 91.073 -61.413 55.029
485287 91.115 -61.425 55.027
485288 91.158 -61.435 55.027
485289 91.200 -61.446 55.030
485290 91.242 -61.457 55.039
485291 91.285 -61.468 55.052
485292 91.327 -61.478 55.065
485293 91.370 -61.490 55.077
485294 91.412 -61.500 55.086
485295 91.454 -61.511 55.086
485296 91.496 -61.522 55.079
485297 91.537 -61.534 55.077
485298 91.579 -61.544 55.080
485299 91.621 -61.556 55.085
485300 91.662 -61.567 55.087
485301 91.704 -61.578 55.085
485302 91.745 -61.589 55.080
485303 91.787 -61.600 55.070
485304 91.826 -61.608 55.054
485305 91.868 -61.620 55.032
485306 91.911 -61.637 54.999
485307 91.953 -61.653 54.973
485308 91.996 -61.669 54.952
485309 92.038 -61.686 54.934
485310 92.080 -61.703 54.914
485311 92.123 -61.719 54.895
485312 92.165 -61.735 54.867
485313 92.210 -61.750 54.830
485314 92.258 -61.763 54.791
485315 92.305 -61.775 54.764
485316 92.353 -61.787 54.746
485317 92.398 -61.798 54.735
485318 92.434 -61.801 54.729
485319 92.471 -61.804 54.729
485320 92.507 -61.808 54.727
485321 92.544 -61.812 54.723
485322 92.580 -61.816 54.713
485323 92.616 -61.819 54.697
485324 92.652 -61.823 54.675
485325 92.689 -61.826 54.645
485326 92.725 -61.830 54.609
485327 92.762 -61.834 54.568
485328 92.797 -61.839 54.520
485329 92.837 -61.851 54.471
485330 92.876 -61.863 54.429
485331 92.916 -61.874 54.397
485332 92.955 -61.886 54.371
485333 92.995 -61.898 54.348
485334 93.035 -61.910 54.328
485335 93.075 -61.921 54.313
485336 93.117 -61.932 54.298
485337 93.163 -61.944 54.282
485338 93.209 -61.955 54.266
485339 93.255 -61.967 54.250
485340 93.301 -61.978 54.235
485341 93.344 -61.988 54.216
485342 93.384 -61.998 54.196
485343 93.423 -62.008 54.175
485344 93.461 -62.018 54.154
485345 93.499 -62.028 54.133
485346 93.537 -62.038 54.114
485347 93.575 -62.047 54.098
485348 93.613 -62.057 54.084
485349 93.651 -62.067 54.072
485350 93.689 -62.077 54.061
485351 93.727 -62.086 54.049
485352 93.765 -62.096 54.038
485353 93.803 -62.106 54.026
485354 93.841 -62.116 54.016
485355 93.882 -62.127 54.006
485356 93.924 -62.139 53.998
485357 93.966 -62.148 53.989
485358 94.008 -62.155 53.979
485359 94.051 -62.163 53.970
485360 94.094 -62.171 53.960
485361 94.136 -62.178 53.946
485362 94.179 -62.186 53.929
485363 94.232 -62.180 53.905
485364 94.290 -62.168 53.876
485365 94.348 -62.156 53.844
485366 94.406 -62.144 53.808
485367 94.465 -62.132 53.769
485368 94.523 -62.120 53.728
485369 94.580 -62.108 53.688
485370 94.639 -62.096 53.649
485371 94.695 -62.086 53.611
485372 94.724 -62.098 53.575
485373 94.740 -62.118 53.539
485374 94.755 -62.139 53.504
485375 94.770 -62.159 53.470
485376 94.785 -62.180 53.435
485377 94.801 -62.200 53.401
485378 94.817 -62.221 53.367
485379 94.832 -62.240 53.335
485380 94.847 -62.261 53.305
485381 94.870 -62.283 53.274
485382 94.901 -62.296 53.243
485383 94.931 -62.295 53.214
485384 94.969 -62.299 53.187
485385 95.006 -62.302 53.161
485386 95.044 -62.305 53.137
485387 95.083 -62.308 53.115
485388 95.120 -62.312 53.091
485389 95.158 -62.315 53.066
485390 95.196 -62.318 53.042
485391 95.235 -62.321 53.019
485392 95.276 -62.326 52.998
485393 95.316 -62.330 52.977
485394 95.357 -62.334 52.956
485395 95.398 -62.339 52.934
485396 95.438 -62.343 52.913
485397 95.479 -62.347 52.892
485398 95.519 -62.352 52.869
485399 95.560 -62.356 52.844
485400 95.601 -62.360 52.817
485401 95.648 -62.361 52.784
485402 95.714 -62.379 52.752
485403 95.781 -62.400 52.718
485404 95.849 -62.421 52.682
485405 95.908 -62.431 52.646
485406 95.946 -62.415 52.610
485407 95.984 -62.400 52.574
485408 96.022 -62.385 52.536
485409 96.060 -62.370 52.499
485410 96.098 -62.368 52.462
485411 96.146 -62.359 52.426
485412 96.191 -62.355 52.390
485413 96.237 -62.351 52.352
485414 96.283 -62.347 52.312
485415 96.336 -62.355 52.271
485416 96.388 -62.362 52.230
485417 96.437 -62.373 52.189
485418 96.476 -62.393 52.147
485419 96.516 -62.413 52.102
485420 96.554 -62.421 52.055
485421 96.591 -62.428 52.009
485422 96.628 -62.434 51.961
485423 96.666 -62.440 51.915
485424 96.703 -62.446 51.869
485425 96.741 -62.452 51.825
485426 96.778 -62.458 51.781
485427 96.815 -62.465 51.738
485428 96.853 -62.471 51.697
485429 96.890 -62.477 51.658
485430 96.928 -62.483 51.620
485431 96.965 -62.489 51.583
485432 97.002 -62.495 51.546
485433 97.039 -62.501 51.510
485434 97.078 -62.506 51.477
485435 97.118 -62.511 51.443
485436 97.157 -62.516 51.410
485437 97.196 -62.522 51.377
485438 97.236 -62.526 51.344
485439 97.276 -62.531 51.311
485440 97.315 -62.537 51.277
485441 97.343 -62.546 51.243
485442 97.371 -62.556 51.209
485443 97.399 -62.565 51.173
485444 97.427 -62.575 51.138
485445 97.455 -62.584 51.103
485446 97.483 -62.594 51.066
485447 97.511 -62.603 51.029
485448 97.571 -62.597 50.991
485449 97.638 -62.589 50.953
485450 97.675 -62.572 50.914
485451 97.689 -62.548 50.876
485452 97.705 -62.524 50.837
485453 97.738 -62.512 50.797
485454 97.771 -62.500 50.757
485455 97.804 -62.488 50.716
485456 97.837 -62.476 50.675
485457 97.879 -62.475 50.635
485458 97.933 -62.490 50.592
485459 97.986 -62.504 50.551
485460 98.039 -62.518 50.509
485461 98.059 -62.533 50.466
485462 98.070 -62.548 50.423
485463 98.080 -62.563 50.382
485464 98.096 -62.578 50.341
485465 98.121 -62.594 50.300
485466 98.145 -62.609 50.258
485918 87.991 -60.587 55.815
485919 88.028 -60.588 55.798
485920 88.064 -60.590 55.779
485921 88.101 -60.592 55.763
485922 88.137 -60.594 55.746
485923 88.174 -60.595 55.729
485924 88.211 -60.597 55.712
485925 88.248 -60.599 55.694
485926 88.284 -60.601 55.676
485927 88.333 -60.607 55.658
485928 88.385 -60.613 55.640
485929 88.442 -60.605 55.622
485930 88.501 -60.590 55.604
485931 88.560 -60.574 55.585
485932 88.630 -60.564 55.563
485933 88.706 -60.557 55.538
485934 88.780 -60.555 55.510
485935 88.853 -60.553 55.483
485936 88.907 -60.570 55.456
485937 88.952 -60.595 55.427
485938 88.994 -60.613 55.398
485939 89.038 -60.632 55.370
485940 89.081 -60.650 55.342
485941 89.124 -60.668 55.316
485942 89.167 -60.686 55.290
485943 89.210 -60.704 55.266
485944 89.253 -60.723 55.242
485945 89.297 -60.741 55.221
485946 89.340 -60.759 55.203
485947 89.383 -60.777 55.187
485948 89.426 -60.795 55.174
485949 89.469 -60.814 55.163
485950 89.512 -60.832 55.155
485951 89.555 -60.851 55.151
485952 89.599 -60.869 55.151
485953 89.641 -60.887 55.152
485954 89.685 -60.905 55.150
485955 89.728 -60.923 55.142
485956 89.771 -60.942 55.129
485957 89.814 -60.960 55.117
485958 89.858 -60.978 55.106
485959 89.900 -60.996 55.095
485960 89.944 -61.014 55.082
485961 89.985 -61.030 55.068
485962 90.021 -61.028 55.056
485963 90.069 -61.055 55.046
485964 90.115 -61.077 55.039
485965 90.159 -61.092 55.035
485966 90.203 -61.107 55.031
485967 90.244 -61.127 55.031
485968 90.286 -61.148 55.032
485969 90.327 -61.168 55.035
485970 90.368 -61.189 55.042
485971 90.410 -61.205 55.046
485972 90.451 -61.215 55.046
485973 90.493 -61.225 55.043
485974 90.535 -61.235 55.040
485975 90.577 -61.246 55.037
485976 90.618 -61.256 55.036
485977 90.660 -61.266 55.036
485978 90.701 -61.277 55.036
485979 90.743 -61.287 55.037
485980 90.792 -61.298 55.040
485981 90.841 -61.309 55.046
485982 90.890 -61.320 55.051
485983 90.929 -61.332 55.052
485984 90.968 -61.344 55.050
485985 91.006 -61.356 55.045
485986 91.045 -61.368 55.040
485987 91.083 -61.380 55.036
485988 91.124 -61.390 55.033
485989 91.167 -61.402 55.030
485990 91.209 -61.412 55.025
485991 91.252 -61.423 55.027
485992 91.294 -61.434 55.036
485993 91.336 -61.445 55.049
485994 91.379 -61.456 55.063
485995 91.422 -61.467 55.075
485996 91.464 -61.478 55.080
485997 91.506 -61.489 55.078
485998 91.547 -61.500 55.077
485999 91.589 -61.511 55.080
486000 91.630 -61.522 55.081
486001 91.671 -61.533 55.082
486002 91.713 -61.544 55.079
486003 91.754 -61.555 55.074
486004 91.796 -61.566 55.063
486005 91.836 -61.573 55.046
486006 91.875 -61.579 55.025
486007 91.918 -61.595 54.995
486008 91.960 -61.612 54.969
486009 92.003 -61.628 54.950
486010 92.045 -61.644 54.937
486011 92.087 -61.661 54.920
486012 92.130 -61.677 54.898
486013 92.173 -61.694 54.868
486014 92.221 -61.706 54.830
486015 92.268 -61.718 54.791
486016 92.316 -61.731 54.764
486017 92.364 -61.743 54.748
486018 92.411 -61.755 54.741
486019 92.449 -61.760 54.739
486020 92.486 -61.764 54.741
486021 92.522 -61.767 54.739
486022 92.558 -61.771 54.732
486023 92.594 -61.775 54.718
486024 92.631 -61.779 54.699
486025 92.667 -61.782 54.674
486026 92.704 -61.785 54.643
486027 92.740 -61.789 54.606
486028 92.775 -61.796 54.566
486029 92.810 -61.804 54.520
486030 92.845 -61.810 54.472
486031 92.882 -61.820 54.434
486032 92.922 -61.832 54.404
486033 92.961 -61.843 54.378
486034 93.001 -61.855 54.358
486035 93.041 -61.867 54.342
486036 93.080 -61.879 54.331
486037 93.124 -61.890 54.317
486038 93.169 -61.901 54.301
486039 93.216 -61.912 54.283
486040 93.261 -61.923 54.266
486041 93.304 -61.934 54.247
486042 93.348 -61.945 54.224
486043 93.391 -61.954 54.200
486044 93.429 -61.964 54.175
486045 93.468 -61.974 54.151
486046 93.505 -61.984 54.127
486047 93.544 -61.994 54.105
486048 93.581 -62.004 54.087
486049 93.620 -62.014 54.071
486050 93.658 -62.023 54.057
486051 93.696 -62.033 54.044
486052 93.734 -62.043 54.032
486053 93.772 -62.053 54.019
486054 93.810 -62.063 54.007
486055 93.849 -62.073 53.995
486056 93.891 -62.085 53.984
486057 93.932 -62.096 53.973
486058 93.974 -62.107 53.960
486059 94.017 -62.114 53.947
486060 94.060 -62.122 53.935
486061 94.102 -62.129 53.925
486062 94.145 -62.136 53.912
486063 94.188 -62.144 53.895
486064 94.230 -62.152 53.873
486065 94.273 -62.159 53.846
486066 94.316 -62.166 53.815
486067 94.374 -62.155 53.779
486068 94.432 -62.142 53.741
486069 94.490 -62.130 53.703
486070 94.548 -62.118 53.664
486071 94.607 -62.107 53.629
486072 94.665 -62.095 53.595
486073 94.709 -62.089 53.561
486074 94.733 -62.097 53.527
486075 94.756 -62.104 53.493
486076 94.777 -62.114 53.461
486077 94.792 -62.134 53.427
486078 94.808 -62.155 53.395
486079 94.823 -62.175 53.363
486080 94.838 -62.196 53.332
486081 94.853 -62.216 53.302
486082 94.869 -62.236 53.273
486083 94.899 -62.246 53.244
486084 94.936 -62.250 53.216
486085 94.974 -62.253 53.189
486086 95.012 -62.256 53.164
486087 95.050 -62.260 53.140
486088 95.088 -62.263 53.116
486089 95.128 -62.267 53.093
486090 95.168 -62.271 53.068
486091 95.209 -62.275 53.044
486092 95.250 -62.280 53.021
486093 95.290 -62.284 53.000
486094 95.331 -62.288 52.979
486095 95.371 -62.293 52.958
486096 95.412 -62.297 52.937
486097 95.453 -62.301 52.917
486098 95.493 -62.305 52.897
486099 95.534 -62.310 52.874
486100 95.574 -62.314 52.850
486101 95.618 -62.317 52.822
486102 95.666 -62.317 52.789
486103 95.734 -62.338 52.756
486104 95.801 -62.359 52.721
486105 95.869 -62.380 52.685
486106 95.915 -62.374 52.647
486107 95.953 -62.359 52.609
486108 95.991 -62.343 52.571
486109 96.029 -62.328 52.533
486110 96.065 -62.324 52.495
486111 96.101 -62.329 52.457
486112 96.145 -62.330 52.419
486113 96.196 -62.307 52.383
486114 96.242 -62.303 52.344
486115 96.288 -62.299 52.305
486116 96.334 -62.297 52.266
486117 96.387 -62.304 52.225
486118 96.440 -62.312 52.184
486119 96.480 -62.330 52.142
486120 96.520 -62.349 52.098
486121 96.559 -62.370 52.051
486122 96.598 -62.381 52.003
486123 96.635 -62.387 51.955
486124 96.672 -62.393 51.908
486125 96.710 -62.399 51.862
486126 96.747 -62.405 51.816
486127 96.785 -62.411 51.773
486128 96.822 -62.417 51.729
486129 96.859 -62.423 51.688
486130 96.896 -62.429 51.648
486131 96.934 -62.435 51.609
486132 96.971 -62.441 51.572
486133 97.009 -62.447 51.536
486134 97.046 -62.453 51.501
486135 97.086 -62.459 51.467
486136 97.125 -62.463 51.433
486137 97.164 -62.468 51.400
486138 97.204 -62.474 51.366
486139 97.243 -62.478 51.333
486140 97.281 -62.484 51.299
486141 97.309 -62.493 51.266
486142 97.337 -62.503 51.232
486143 97.365 -62.512 51.196
486144 97.393 -62.522 51.161
486145 97.421 -62.531 51.126
486146 97.449 -62.541 51.090
486147 97.477 -62.550 51.054
486148 97.505 -62.560 51.016
486149 97.572 -62.552 50.979
486150 97.639 -62.543 50.941
486151 97.682 -62.528 50.902
486152 97.697 -62.503 50.864
486153 97.713 -62.480 50.826
486154 97.746 -62.468 50.787
486155 97.779 -62.456 50.746
486156 97.812 -62.444 50.706
486157 97.845 -62.433 50.665
486158 97.893 -62.440 50.625
486159 97.946 -62.453 50.583
486160 98.000 -62.468 50.542
486161 98.040 -62.482 50.501
486162 98.051 -62.497 50.459
486163 98.062 -62.512 50.417
486164 98.074 -62.528 50.376
486165 98.098 -62.543 50.335
486166 98.122 -62.559 50.294
486167 98.146 -62.575 50.252
486168 98.170 -62.590 50.211
486619 88.006 -60.538 55.804
486620 88.043 -60.541 55.786
486621 88.079 -60.542 55.767
486622 88.116 -60.544 55.751
486623 88.153 -60.546 55.734
486624 88.190 -60.547 55.716
486625 88.226 -60.550 55.698
486626 88.263 -60.551 55.681
486627 88.308 -60.556 55.663
486628 88.360 -60.563 55.645
486629 88.412 -60.569 55.628
486630 88.470 -60.557 55.611
486631 88.530 -60.541 55.594
486632 88.589 -60.525 55.576
486633 88.648 -60.510 55.555
486634 88.723 -60.503 55.530
486635 88.797 -60.502 55.503
486636 88.870 -60.500 55.476
486637 88.921 -60.520 55.448
486638 88.965 -60.542 55.420
486639 89.008 -60.560 55.391
486640 89.051 -60.579 55.363
486641 89.095 -60.597 55.335
486642 89.137 -60.615 55.309
486643 89.181 -60.633 55.283
486644 89.224 -60.651 55.258
486645 89.267 -60.669 55.235
486646 89.310 -60.688 55.213
486647 89.353 -60.706 55.194
486648 89.396 -60.724 55.177
486649 89.440 -60.742 55.162
486650 89.483 -60.760 55.149
486651 89.526 -60.779 55.138
486652 89.569 -60.797 55.132
486653 89.612 -60.815 55.128
486654 89.655 -60.833 55.127
486655 89.698 -60.852 55.126
486656 89.742 -60.870 55.121
486657 89.785 -60.889 55.112
486658 89.828 -60.907 55.105
486659 89.871 -60.925 55.098
486660 89.914 -60.943 55.090
486661 89.957 -60.961 55.076
486662 90.001 -60.980 55.063
486663 90.042 -60.993 55.051
486664 90.087 -61.012 55.040
486665 90.129 -61.030 55.032
486666 90.170 -61.050 55.026
486667 90.212 -61.071 55.023
486668 90.253 -61.092 55.022
486669 90.294 -61.112 55.022
486670 90.335 -61.133 55.024
486671 90.377 -61.153 55.028
486672 90.418 -61.168 55.033
486673 90.460 -61.179 55.034
486674 90.502 -61.190 55.032
486675 90.543 -61.199 55.026
486676 90.585 -61.210 55.024
486677 90.627 -61.220 55.024
486678 90.668 -61.230 55.026
486679 90.712 -61.241 55.029
486680 90.761 -61.252 55.033
486681 90.810 -61.263 55.037
486682 90.859 -61.274 55.045
486683 90.904 -61.286 55.051
486684 90.943 -61.297 55.053
486685 90.981 -61.309 55.052
486686 91.020 -61.321 55.048
486687 91.058 -61.333 55.045
486688 91.097 -61.345 55.043
486689 91.135 -61.357 55.041
486690 91.176 -61.368 55.037
486691 91.219 -61.379 55.029
486692 91.261 -61.390 55.024
486693 91.303 -61.401 55.026
486694 91.346 -61.412 55.034
486695 91.388 -61.422 55.045
486696 91.431 -61.434 55.059
486697 91.473 -61.444 55.068
486698 91.515 -61.455 55.071
486699 91.557 -61.466 55.075
486700 91.598 -61.478 55.078
486701 91.640 -61.488 55.077
486702 91.681 -61.500 55.076
486703 91.723 -61.511 55.072
486704 91.764 -61.522 55.064
486705 91.805 -61.532 55.051
486706 91.845 -61.538 55.034
486707 91.884 -61.544 55.017
486708 91.925 -61.553 54.990
486709 91.967 -61.570 54.966
486710 92.010 -61.587 54.950
486711 92.052 -61.603 54.941
486712 92.094 -61.619 54.925
486713 92.137 -61.636 54.897
486714 92.183 -61.649 54.863
486715 92.231 -61.662 54.826
486716 92.279 -61.674 54.792
486717 92.326 -61.686 54.766
486718 92.374 -61.699 54.750
486719 92.422 -61.711 54.743
486720 92.464 -61.719 54.743
486721 92.500 -61.723 54.746
486722 92.536 -61.727 54.743
486723 92.573 -61.730 54.732
486724 92.609 -61.734 54.715
486725 92.646 -61.738 54.693
486726 92.682 -61.741 54.667
486727 92.718 -61.746 54.634
486728 92.753 -61.753 54.600
486729 92.788 -61.760 54.561
486730 92.823 -61.767 54.518
486731 92.858 -61.775 54.475
486732 92.893 -61.782 54.439
486733 92.928 -61.789 54.412
486734 92.967 -61.801 54.390
486735 93.007 -61.812 54.373
486736 93.047 -61.824 54.362
486737 93.086 -61.835 54.352
486738 93.130 -61.847 54.339
486739 93.176 -61.858 54.323
486740 93.221 -61.869 54.304
486741 93.265 -61.879 54.281
486742 93.308 -61.890 54.257
486743 93.351 -61.901 54.230
486744 93.395 -61.911 54.202
486745 93.436 -61.921 54.173
486746 93.474 -61.931 54.146
486747 93.512 -61.941 54.121
486748 93.550 -61.951 54.096
486749 93.588 -61.961 54.075
486750 93.626 -61.970 54.057
486751 93.664 -61.980 54.042
486752 93.702 -61.990 54.027
486753 93.741 -62.000 54.013
486754 93.779 -62.009 53.999
486755 93.817 -62.020 53.985
486756 93.858 -62.031 53.973
486757 93.900 -62.042 53.961
486758 93.941 -62.054 53.947
486759 93.983 -62.065 53.932
486760 94.026 -62.073 53.917
486761 94.068 -62.080 53.902
486762 94.111 -62.088 53.888
486763 94.153 -62.095 53.873
486764 94.196 -62.103 53.854
486765 94.239 -62.111 53.833
486766 94.281 -62.117 53.807
486767 94.324 -62.125 53.778
486768 94.367 -62.133 53.747
486769 94.410 -62.140 53.713
486770 94.458 -62.140 53.679
486771 94.516 -62.129 53.645
486772 94.574 -62.117 53.612
486773 94.632 -62.105 53.580
486774 94.688 -62.094 53.549
486775 94.711 -62.101 53.518
486776 94.734 -62.108 53.486
486777 94.757 -62.115 53.455
486778 94.780 -62.122 53.423
486779 94.803 -62.129 53.392
486780 94.826 -62.136 53.361
486781 94.845 -62.150 53.332
486782 94.860 -62.171 53.303
486783 94.876 -62.191 53.277
486784 94.904 -62.202 53.249
486785 94.942 -62.205 53.221
486786 94.980 -62.208 53.195
486787 95.020 -62.212 53.170
486788 95.061 -62.217 53.145
486789 95.102 -62.221 53.120
486790 95.142 -62.225 53.096
486791 95.183 -62.230 53.072
486792 95.223 -62.233 53.047
486793 95.264 -62.238 53.024
486794 95.305 -62.243 53.002
486795 95.345 -62.246 52.981
486796 95.386 -62.251 52.960
486797 95.426 -62.255 52.940
486798 95.467 -62.259 52.921
486799 95.508 -62.264 52.901
486800 95.548 -62.268 52.880
486801 95.589 -62.272 52.855
486802 95.635 -62.273 52.827
486803 95.686 -62.276 52.795
486804 95.753 -62.296 52.760
486805 95.821 -62.318 52.725
486806 95.884 -62.333 52.687
486807 95.922 -62.318 52.649
486808 95.960 -62.302 52.610
486809 95.999 -62.287 52.569
486810 96.035 -62.279 52.530
486811 96.071 -62.284 52.491
486812 96.105 -62.289 52.452
486813 96.148 -62.291 52.414
486814 96.192 -62.290 52.377
486815 96.244 -62.263 52.338
486816 96.292 -62.251 52.298
486817 96.338 -62.247 52.259
486818 96.385 -62.246 52.220
486819 96.438 -62.254 52.180
486820 96.484 -62.267 52.138
486821 96.524 -62.287 52.094
486822 96.564 -62.306 52.048
486823 96.603 -62.327 52.001
486824 96.642 -62.339 51.953
486825 96.679 -62.345 51.906
486826 96.716 -62.351 51.858
486827 96.753 -62.357 51.812
486828 96.791 -62.363 51.767
486829 96.828 -62.370 51.723
486830 96.866 -62.376 51.681
486831 96.903 -62.382 51.641
486832 96.940 -62.388 51.602
486833 96.977 -62.394 51.565
486834 97.015 -62.400 51.529
486835 97.053 -62.406 51.494
486836 97.093 -62.411 51.459
486837 97.132 -62.415 51.425
486838 97.172 -62.421 51.392
486839 97.211 -62.426 51.358
486840 97.247 -62.432 51.325
486841 97.276 -62.441 51.291
486842 97.304 -62.451 51.258
486843 97.331 -62.460 51.223
486844 97.360 -62.470 51.189
486845 97.388 -62.479 51.153
486846 97.416 -62.489 51.117
486847 97.444 -62.498 51.081
486848 97.472 -62.508 51.044
486849 97.506 -62.514 51.007
486850 97.573 -62.506 50.969
486851 97.640 -62.497 50.931
486852 97.689 -62.484 50.894
486853 97.704 -62.459 50.855
486854 97.721 -62.437 50.817
486855 97.753 -62.424 50.778
486856 97.787 -62.412 50.738
486857 97.820 -62.401 50.698
486858 97.853 -62.389 50.658
486859 97.907 -62.403 50.617
486860 97.960 -62.418 50.576
486861 98.014 -62.432 50.535
486862 98.032 -62.446 50.495
486863 98.043 -62.462 50.454
486864 98.055 -62.477 50.412
486865 98.076 -62.492 50.371
486866 98.100 -62.508 50.329
486867 98.124 -62.524 50.288
486868 98.148 -62.539 50.247
486869 98.172 -62.555 50.206
486870 98.197 -62.570 50.166
487320 88.021 -60.491 55.793
487321 88.058 -60.493 55.775
487322 88.095 -60.494 55.756
487323 88.131 -60.496 55.739
487324 88.168 -60.498 55.721
487325 88.205 -60.500 55.703
487326 88.242 -60.501 55.685
487327 88.283 -60.505 55.667
487328 88.335 -60.512 55.650
487329 88.387 -60.519 55.632
487330 88.439 -60.525 55.615
487331 88.498 -60.509 55.599
487332 88.557 -60.493 55.584
487333 88.617 -60.478 55.567
487334 88.676 -60.462 55.547
487335 88.740 -60.450 55.523
487336 88.813 -60.449 55.497
487337 88.887 -60.447 55.469
487338 88.935 -60.469 55.442
487339 88.979 -60.489 55.414
487340 89.022 -60.507 55.385
487341 89.065 -60.525 55.356
487342 89.108 -60.544 55.330
487343 89.151 -60.562 55.303
487344 89.194 -60.580 55.277
487345 89.238 -60.598 55.252
487346 89.280 -60.616 55.228
487347 89.324 -60.635 55.205
487348 89.367 -60.653 55.186
487349 89.410 -60.671 55.168
487350 89.453 -60.689 55.152
487351 89.496 -60.707 55.138
487352 89.539 -60.726 55.125
487353 89.583 -60.744 55.115
487354 89.626 -60.762 55.107
487355 89.669 -60.780 55.102
487356 89.712 -60.798 55.100
487357 89.755 -60.817 55.098
487358 89.799 -60.835 55.095
487359 89.841 -60.853 55.091
487360 89.885 -60.871 55.087
487361 89.928 -60.891 55.080
487362 89.970 -60.912 55.069
487363 90.013 -60.934 55.058
487364 90.055 -60.955 55.047
487365 90.096 -60.973 55.036
487366 90.138 -60.994 55.025
487367 90.179 -61.014 55.019
487368 90.220 -61.035 55.015
487369 90.261 -61.056 55.013
487370 90.303 -61.077 55.013
487371 90.344 -61.097 55.013
487372 90.385 -61.118 55.016
487373 90.427 -61.133 55.019
487374 90.468 -61.143 55.020
487375 90.510 -61.153 55.019
487376 90.552 -61.164 55.017
487377 90.594 -61.174 55.017
487378 90.636 -61.184 55.018
487379 90.680 -61.195 55.020
487380 90.729 -61.206 55.025
487381 90.778 -61.217 55.031
487382 90.827 -61.228 55.036
487383 90.876 -61.240 55.043
487384 90.918 -61.251 55.050
487385 90.956 -61.263 55.054
487386 90.995 -61.274 55.055
487387 91.033 -61.287 55.051
487388 91.072 -61.299 55.050
487389 91.110 -61.310 55.049
487390 91.149 -61.322 55.048
487391 91.187 -61.334 55.044
487392 91.228 -61.346 55.034
487393 91.271 -61.356 55.025
487394 91.313 -61.367 55.021
487395 91.355 -61.378 55.022
487396 91.397 -61.389 55.029
487397 91.440 -61.400 55.041
487398 91.483 -61.411 55.051
487399 91.525 -61.422 55.058
487400 91.566 -61.433 55.064
487401 91.608 -61.443 55.069
487402 91.649 -61.455 55.068
487403 91.691 -61.466 55.066
487404 91.732 -61.477 55.061
487405 91.774 -61.488 55.051
487406 91.814 -61.497 55.037
487407 91.854 -61.503 55.020
487408 91.893 -61.509 55.005
487409 91.932 -61.515 54.983
487410 91.974 -61.528 54.965
487411 92.017 -61.545 54.954
487412 92.059 -61.562 54.946
487413 92.101 -61.578 54.927
487414 92.146 -61.593 54.892
487415 92.194 -61.605 54.853
487416 92.241 -61.617 54.817
487417 92.289 -61.630 54.792
487418 92.337 -61.642 54.771
487419 92.384 -61.654 54.753
487420 92.432 -61.667 54.744
487421 92.479 -61.678 54.743
487422 92.515 -61.682 54.747
487423 92.552 -61.686 54.743
487424 92.588 -61.690 54.730
487425 92.624 -61.693 54.711
487426 92.660 -61.697 54.689
487427 92.696 -61.702 54.663
487428 92.731 -61.710 54.631
487429 92.766 -61.717 54.599
487430 92.801 -61.724 54.561
487431 92.836 -61.732 54.520
487432 92.871 -61.738 54.477
487433 92.906 -61.746 54.444
487434 92.941 -61.754 54.421
487435 92.976 -61.760 54.405
487436 93.012 -61.769 54.394
487437 93.052 -61.781 54.387
487438 93.092 -61.792 54.376
487439 93.137 -61.804 54.360
487440 93.182 -61.815 54.341
487441 93.225 -61.826 54.318
487442 93.268 -61.835 54.292
487443 93.311 -61.846 54.261
487444 93.355 -61.857 54.229
487445 93.398 -61.867 54.197
487446 93.441 -61.877 54.167
487447 93.481 -61.888 54.139
487448 93.519 -61.898 54.113
487449 93.557 -61.907 54.087
487450 93.595 -61.917 54.064
487451 93.633 -61.926 54.044
487452 93.671 -61.936 54.026
487453 93.709 -61.946 54.009
487454 93.747 -61.956 53.993
487455 93.785 -61.966 53.977
487456 93.826 -61.977 53.962
487457 93.867 -61.989 53.948
487458 93.909 -62.000 53.936
487459 93.951 -62.011 53.922
487460 93.992 -62.023 53.906
487461 94.034 -62.032 53.889
487462 94.077 -62.039 53.870
487463 94.119 -62.046 53.853
487464 94.162 -62.054 53.835
487465 94.205 -62.061 53.816
487466 94.248 -62.069 53.795
487467 94.290 -62.077 53.771
487468 94.333 -62.084 53.745
487469 94.376 -62.092 53.717
487470 94.418 -62.098 53.687
487471 94.461 -62.106 53.658
487472 94.503 -62.114 53.628
487473 94.546 -62.121 53.598
487474 94.600 -62.114 53.569
487475 94.658 -62.103 53.541
487476 94.689 -62.105 53.512
487477 94.712 -62.112 53.483
487478 94.735 -62.119 53.455
487479 94.758 -62.127 53.425
487480 94.781 -62.133 53.395
487481 94.804 -62.140 53.367
487482 94.827 -62.147 53.340
487483 94.851 -62.152 53.313
487484 94.880 -62.154 53.286
487485 94.913 -62.158 53.259
487486 94.954 -62.161 53.232
487487 94.994 -62.166 53.205
487488 95.035 -62.171 53.180
487489 95.075 -62.174 53.155
487490 95.116 -62.179 53.131
487491 95.157 -62.183 53.106
487492 95.197 -62.187 53.080
487493 95.238 -62.192 53.053
487494 95.278 -62.196 53.029
487495 95.319 -62.200 53.007
487496 95.359 -62.205 52.985
487497 95.400 -62.209 52.964
487498 95.441 -62.213 52.944
487499 95.481 -62.218 52.924
487500 95.522 -62.222 52.905
487501 95.562 -62.226 52.883
487502 95.605 -62.229 52.859
487503 95.653 -62.230 52.831
487504 95.705 -62.234 52.800
487505 95.773 -62.255 52.765
487506 95.841 -62.276 52.728
487507 95.891 -62.276 52.690
487508 95.930 -62.261 52.651
487509 95.967 -62.246 52.611
487510 96.005 -62.234 52.569
487511 96.040 -62.239 52.527
487512 96.075 -62.244 52.486
487513 96.110 -62.249 52.448
487514 96.151 -62.252 52.410
487515 96.194 -62.253 52.371
487516 96.239 -62.251 52.332
487517 96.291 -62.224 52.292
487518 96.342 -62.199 52.253
487519 96.388 -62.196 52.214
487520 96.437 -62.196 52.175
487521 96.488 -62.204 52.133
487522 96.528 -62.224 52.090
487523 96.568 -62.243 52.045
487524 96.607 -62.263 51.999
487525 96.647 -62.283 51.952
487526 96.685 -62.298 51.904
487527 96.723 -62.304 51.856
487528 96.760 -62.310 51.809
487529 96.797 -62.316 51.763
487530 96.834 -62.322 51.719
487531 96.872 -62.328 51.677
487532 96.909 -62.334 51.636
487533 96.947 -62.340 51.597
487534 96.984 -62.346 51.560
487535 97.021 -62.352 51.523
487536 97.061 -62.358 51.488
487537 97.100 -62.363 51.453
487538 97.139 -62.368 51.419
487539 97.179 -62.373 51.385
487540 97.214 -62.380 51.352
487541 97.242 -62.389 51.319
487542 97.270 -62.399 51.285
487543 97.298 -62.408 51.251
487544 97.326 -62.418 51.217
487545 97.354 -62.427 51.183
487546 97.382 -62.437 51.146
487547 97.410 -62.446 51.110
487548 97.438 -62.456 51.073
487549 97.466 -62.465 51.036
487550 97.508 -62.468 50.999
487551 97.575 -62.460 50.962
487552 97.642 -62.451 50.924
487553 97.697 -62.440 50.886
487554 97.712 -62.415 50.848
487555 97.728 -62.393 50.810
487556 97.761 -62.381 50.771
487557 97.795 -62.369 50.732
487558 97.828 -62.357 50.692
487559 97.867 -62.353 50.652
487560 97.921 -62.367 50.612
487561 97.974 -62.381 50.572
487562 98.014 -62.396 50.531
487563 98.025 -62.411 50.490
487564 98.036 -62.426 50.449
487565 98.053 -62.441 50.407
487566 98.077 -62.457 50.366
487567 98.101 -62.472 50.326
487568 98.126 -62.488 50.285
487569 98.150 -62.504 50.244
487570 98.174 -62.519 50.203
487571 98.198 -62.535 50.162
488021 88.037 -60.443 55.783
488022 88.073 -60.444 55.763
488023 88.110 -60.447 55.744
488024 88.146 -60.448 55.726
488025 88.183 -60.450 55.708
488026 88.220 -60.452 55.690
488027 88.257 -60.454 55.672
488028 88.310 -60.461 55.653
488029 88.361 -60.468 55.636
488030 88.413 -60.475 55.619
488031 88.467 -60.476 55.602
488032 88.527 -60.460 55.586
488033 88.586 -60.445 55.572
488034 88.645 -60.429 55.555
488035 88.704 -60.414 55.537
488036 88.761 -60.401 55.515
488037 88.830 -60.397 55.493
488038 88.903 -60.395 55.464
488039 88.949 -60.418 55.436
488040 88.992 -60.436 55.408
488041 89.035 -60.454 55.380
488042 89.079 -60.472 55.351
488043 89.122 -60.491 55.325
488044 89.165 -60.509 55.298
488045 89.208 -60.527 55.271
488046 89.251 -60.545 55.246
488047 89.294 -60.563 55.223
488048 89.337 -60.582 55.199
488049 89.381 -60.600 55.179
488050 89.423 -60.618 55.161
488051 89.467 -60.636 55.144
488052 89.510 -60.654 55.128
488053 89.553 -60.673 55.115
488054 89.596 -60.691 55.104
488055 89.639 -60.709 55.094
488056 89.682 -60.731 55.086
488057 89.725 -60.752 55.080
488058 89.767 -60.774 55.076
488059 89.810 -60.795 55.073
488060 89.852 -60.817 55.072
488061 89.895 -60.839 55.070
488062 89.938 -60.860 55.064
488063 89.981 -60.882 55.059
488064 90.023 -60.903 55.052
488065 90.066 -60.925 55.043
488066 90.107 -60.943 55.032
488067 90.147 -60.961 55.020
488068 90.187 -60.979 55.012
488069 90.229 -60.999 55.007
488070 90.270 -61.020 55.005
488071 90.311 -61.041 55.004
488072 90.353 -61.061 55.003
488073 90.394 -61.082 55.005
488074 90.435 -61.097 55.007
488075 90.477 -61.107 55.009
488076 90.519 -61.118 55.010
488077 90.561 -61.128 55.011
488078 90.602 -61.138 55.013
488079 90.649 -61.149 55.015
488080 90.698 -61.160 55.019
488081 90.747 -61.171 55.025
488082 90.796 -61.182 55.032
488083 90.845 -61.193 55.038
488084 90.892 -61.205 55.044
488085 90.931 -61.216 55.051
488086 90.969 -61.228 55.056
488087 91.008 -61.240 55.059
488088 91.046 -61.252 55.058
488089 91.085 -61.264 55.056
488090 91.123 -61.276 55.054
488091 91.162 -61.287 55.052
488092 91.201 -61.299 55.049
488093 91.239 -61.312 55.040
488094 91.280 -61.323 55.029
488095 91.322 -61.334 55.020
488096 91.365 -61.344 55.015
488097 91.407 -61.356 55.016
488098 91.449 -61.366 55.023
488099 91.492 -61.377 55.032
488100 91.534 -61.388 55.040
488101 91.576 -61.399 55.048
488102 91.617 -61.410 55.055
488103 91.659 -61.422 55.056
488104 91.701 -61.432 55.054
488105 91.742 -61.443 55.048
488106 91.784 -61.454 55.037
488107 91.823 -61.462 55.021
488108 91.863 -61.469 55.004
488109 91.902 -61.474 54.991
488110 91.941 -61.480 54.975
488111 91.981 -61.487 54.964
488112 92.024 -61.503 54.959
488113 92.066 -61.520 54.950
488114 92.109 -61.536 54.926
488115 92.157 -61.549 54.890
488116 92.204 -61.561 54.851
488117 92.252 -61.573 54.817
488118 92.299 -61.586 54.799
488119 92.347 -61.598 54.779
488120 92.395 -61.610 54.758
488121 92.442 -61.623 54.743
488122 92.490 -61.635 54.738
488123 92.530 -61.641 54.742
488124 92.566 -61.645 54.738
488125 92.603 -61.649 54.726
488126 92.639 -61.653 54.709
488127 92.674 -61.659 54.688
488128 92.709 -61.666 54.661
488129 92.744 -61.674 54.633
488130 92.779 -61.681 54.602
488131 92.814 -61.688 54.566
488132 92.849 -61.695 54.524
488133 92.884 -61.703 54.479
488134 92.919 -61.710 54.450
488135 92.954 -61.717 54.432
488136 92.989 -61.725 54.422
488137 93.024 -61.732 54.417
488138 93.059 -61.739 54.411
488139 93.097 -61.750 54.398
488140 93.142 -61.760 54.379
488141 93.185 -61.771 54.356
488142 93.228 -61.782 54.329
488143 93.272 -61.792 54.297
488144 93.315 -61.802 54.262
488145 93.358 -61.813 54.228
488146 93.402 -61.823 54.194
488147 93.445 -61.834 54.163
488148 93.487 -61.844 54.135
488149 93.525 -61.854 54.107
488150 93.564 -61.863 54.081
488151 93.602 -61.873 54.057
488152 93.640 -61.883 54.035
488153 93.678 -61.893 54.016
488154 93.716 -61.903 53.997
488155 93.754 -61.913 53.979
488156 93.793 -61.923 53.962
488157 93.835 -61.935 53.945
488158 93.876 -61.946 53.929
488159 93.918 -61.957 53.914
488160 93.960 -61.969 53.898
488161 94.001 -61.980 53.881
488162 94.043 -61.990 53.863
488163 94.085 -61.998 53.844
488164 94.128 -62.005 53.825
488165 94.171 -62.013 53.806
488166 94.213 -62.020 53.786
488167 94.256 -62.028 53.765
488168 94.299 -62.035 53.742
488169 94.341 -62.042 53.718
488170 94.384 -62.050 53.693
488171 94.427 -62.058 53.666
488172 94.469 -62.065 53.641
488173 94.512 -62.073 53.615
488174 94.555 -62.080 53.588
488175 94.598 -62.087 53.561
488176 94.640 -62.095 53.536
488177 94.678 -62.099 53.510
488178 94.705 -62.101 53.485
488179 94.734 -62.102 53.460
488180 94.762 -62.104 53.432
488181 94.789 -62.105 53.404
488182 94.818 -62.106 53.378
488183 94.846 -62.108 53.352
488184 94.874 -62.109 53.326
488185 94.902 -62.111 53.300
488186 94.931 -62.114 53.273
488187 94.969 -62.120 53.246
488188 95.008 -62.124 53.220
488189 95.049 -62.129 53.194
488190 95.090 -62.133 53.168
488191 95.130 -62.137 53.144
488192 95.171 -62.142 53.118
488193 95.211 -62.145 53.091
488194 95.252 -62.150 53.063
488195 95.293 -62.155 53.038
488196 95.333 -62.158 53.014
488197 95.374 -62.163 52.992
488198 95.414 -62.167 52.970
488199 95.455 -62.171 52.949
488200 95.495 -62.176 52.929
488201 95.536 -62.180 52.908
488202 95.577 -62.184 52.886
488203 95.623 -62.186 52.861
488204 95.670 -62.186 52.834
488205 95.725 -62.193 52.802
488206 95.793 -62.214 52.768
488207 95.861 -62.235 52.731
488208 95.898 -62.220 52.692
488209 95.937 -62.204 52.652
488210 95.974 -62.189 52.611
488211 96.010 -62.194 52.568
488212 96.044 -62.199 52.527
488213 96.080 -62.204 52.486
488214 96.115 -62.208 52.447
488215 96.154 -62.212 52.408
488216 96.197 -62.214 52.369
488217 96.240 -62.215 52.328
488218 96.286 -62.211 52.288
488219 96.338 -62.183 52.249
488220 96.390 -62.156 52.210
488221 96.438 -62.143 52.170
488222 96.487 -62.145 52.129
488223 96.532 -62.161 52.086
488224 96.571 -62.180 52.042
488225 96.611 -62.200 51.998
488226 96.650 -62.220 51.951
488227 96.690 -62.239 51.904
488228 96.729 -62.256 51.857
488229 96.766 -62.262 51.810
488230 96.804 -62.268 51.762
488231 96.841 -62.275 51.716
488232 96.878 -62.281 51.675
488233 96.915 -62.287 51.634
488234 96.953 -62.293 51.594
488235 96.990 -62.299 51.556
488236 97.028 -62.305 51.519
488237 97.068 -62.310 51.484
488238 97.107 -62.315 51.449
488239 97.147 -62.320 51.415
488240 97.180 -62.327 51.381
488241 97.209 -62.337 51.348
488242 97.237 -62.346 51.314
488243 97.264 -62.355 51.281
488244 97.293 -62.365 51.247
488245 97.321 -62.374 51.213
488246 97.349 -62.384 51.178
488247 97.377 -62.393 51.141
488248 97.405 -62.403 51.105
488249 97.433 -62.412 51.068
488250 97.461 -62.422 51.031
488251 97.509 -62.422 50.994
488252 97.576 -62.414 50.957
488253 97.643 -62.406 50.918
488254 97.704 -62.395 50.880
488255 97.719 -62.371 50.842
488256 97.736 -62.349 50.805
488257 97.770 -62.337 50.766
488258 97.803 -62.325 50.726
488259 97.836 -62.313 50.686
488260 97.881 -62.317 50.646
488261 97.935 -62.331 50.608
488262 97.988 -62.346 50.568
488263 98.007 -62.360 50.527
488264 98.018 -62.375 50.486
488265 98.031 -62.390 50.445
488266 98.055 -62.406 50.403
488267 98.079 -62.421 50.362
488268 98.103 -62.437 50.322
488269 98.128 -62.453 50.282
488270 98.152 -62.468 50.241
488271 98.176 -62.484 50.201
488272 98.200 -62.500 50.160
488273 98.224 -62.515 50.121
488723 88.088 -60.397 55.752
488724 88.125 -60.399 55.732
488725 88.162 -60.400 55.714
488726 88.198 -60.402 55.695
488727 88.235 -60.404 55.676
488728 88.284 -60.410 55.658
488729 88.336 -60.417 55.640
488730 88.388 -60.424 55.622
488731 88.440 -60.431 55.605
488732 88.495 -60.428 55.588
488733 88.555 -60.412 55.573
488734 88.614 -60.397 55.559
488735 88.673 -60.381 55.543
488736 88.732 -60.365 55.526
488737 88.788 -60.354 55.506
488738 88.846 -60.343 55.484
488739 88.919 -60.343 55.458
488740 88.963 -60.364 55.430
488741 89.006 -60.382 55.402
488742 89.049 -60.400 55.374
488743 89.092 -60.419 55.346
488744 89.135 -60.437 55.319
488745 89.178 -60.456 55.293
488746 89.222 -60.474 55.266
488747 89.265 -60.492 55.241
488748 89.308 -60.510 55.217
488749 89.351 -60.528 55.194
488750 89.394 -60.549 55.173
488751 89.437 -60.571 55.154
488752 89.479 -60.592 55.136
488753 89.522 -60.614 55.120
488754 89.564 -60.635 55.106
488755 89.607 -60.657 55.094
488756 89.650 -60.679 55.083
488757 89.692 -60.700 55.074
488758 89.735 -60.722 55.066
488759 89.777 -60.743 55.059
488760 89.820 -60.765 55.055
488761 89.863 -60.786 55.054
488762 89.906 -60.807 55.052
488763 89.948 -60.829 55.049
488764 89.991 -60.851 55.048
488765 90.034 -60.873 55.043
488766 90.076 -60.894 55.037
488767 90.117 -60.913 55.029
488768 90.157 -60.930 55.018
488769 90.197 -60.947 55.008
488770 90.237 -60.964 55.001
488771 90.279 -60.984 54.997
488772 90.320 -61.005 54.996
488773 90.361 -61.026 54.995
488774 90.402 -61.046 54.995
488775 90.444 -61.061 54.998
488776 90.485 -61.071 55.001
488777 90.527 -61.082 55.004
488778 90.569 -61.092 55.008
488779 90.618 -61.103 55.011
488780 90.667 -61.114 55.014
488781 90.716 -61.125 55.020
488782 90.764 -61.136 55.027
488783 90.813 -61.147 55.035
488784 90.862 -61.158 55.042
488785 90.905 -61.170 55.047
488786 90.944 -61.182 55.055
488787 90.982 -61.193 55.062
488788 91.021 -61.205 55.067
488789 91.060 -61.218 55.067
488790 91.098 -61.229 55.064
488791 91.137 -61.241 55.061
488792 91.175 -61.253 55.059
488793 91.214 -61.265 55.055
488794 91.252 -61.277 55.047
488795 91.291 -61.289 55.035
488796 91.331 -61.300 55.023
488797 91.374 -61.311 55.014
488798 91.416 -61.321 55.009
488799 91.458 -61.333 55.011
488800 91.501 -61.343 55.015
488801 91.543 -61.355 55.020
488802 91.585 -61.365 55.026
488803 91.627 -61.377 55.033
488804 91.668 -61.387 55.038
488805 91.710 -61.399 55.040
488806 91.751 -61.409 55.035
488807 91.793 -61.421 55.022
488808 91.832 -61.428 55.006
488809 91.872 -61.434 54.988
488810 91.911 -61.439 54.977
488811 91.950 -61.445 54.969
488812 91.990 -61.451 54.965
488813 92.031 -61.462 54.963
488814 92.073 -61.478 54.952
488815 92.120 -61.492 54.926
488816 92.167 -61.504 54.890
488817 92.215 -61.517 54.850
488818 92.262 -61.529 54.816
488819 92.310 -61.541 54.801
488820 92.358 -61.554 54.786
488821 92.405 -61.566 54.766
488822 92.453 -61.578 54.746
488823 92.501 -61.591 54.732
488824 92.545 -61.601 54.731
488825 92.581 -61.604 54.729
488826 92.618 -61.609 54.719
488827 92.652 -61.616 54.704
488828 92.688 -61.623 54.682
488829 92.722 -61.630 54.655
488830 92.757 -61.638 54.629
488831 92.792 -61.645 54.598
488832 92.827 -61.652 54.561
488833 92.862 -61.660 54.520
488834 92.897 -61.666 54.481
488835 92.932 -61.674 54.458
488836 92.967 -61.681 54.447
488837 93.002 -61.688 54.443
488838 93.037 -61.695 54.440
488839 93.072 -61.702 54.432
488840 93.107 -61.709 54.417
488841 93.146 -61.717 54.395
488842 93.189 -61.727 54.367
488843 93.232 -61.738 54.334
488844 93.275 -61.748 54.296
488845 93.319 -61.758 54.260
488846 93.362 -61.769 54.226
488847 93.405 -61.779 54.193
488848 93.448 -61.790 54.162
488849 93.491 -61.801 54.133
488850 93.532 -61.810 54.104
488851 93.570 -61.820 54.077
488852 93.608 -61.830 54.053
488853 93.646 -61.840 54.030
488854 93.684 -61.850 54.009
488855 93.722 -61.860 53.989
488856 93.761 -61.870 53.969
488857 93.802 -61.881 53.951
488858 93.844 -61.892 53.933
488859 93.885 -61.904 53.915
488860 93.927 -61.915 53.898
488861 93.968 -61.926 53.881
488862 94.010 -61.938 53.863
488863 94.051 -61.948 53.844
488864 94.094 -61.957 53.826
488865 94.137 -61.964 53.806
488866 94.179 -61.971 53.788
488867 94.222 -61.979 53.769
488868 94.265 -61.986 53.748
488869 94.308 -61.994 53.726
488870 94.350 -62.001 53.703
488871 94.393 -62.009 53.680
488872 94.435 -62.017 53.656
488873 94.478 -62.023 53.633
488874 94.521 -62.031 53.609
488875 94.563 -62.039 53.585
488876 94.606 -62.046 53.560
488877 94.648 -62.053 53.538
488878 94.685 -62.055 53.516
488879 94.720 -62.057 53.495
488880 94.755 -62.059 53.473
488881 94.784 -62.060 53.447
488882 94.812 -62.061 53.422
488883 94.840 -62.063 53.395
488884 94.868 -62.064 53.369
488885 94.896 -62.066 53.345
488886 94.924 -62.067 53.319
488887 94.952 -62.069 53.292
488888 94.988 -62.075 53.265
488889 95.025 -62.081 53.239
488890 95.063 -62.086 53.211
488891 95.104 -62.091 53.186
488892 95.144 -62.095 53.160
488893 95.185 -62.099 53.133
488894 95.226 -62.104 53.104
488895 95.266 -62.108 53.075
488896 95.307 -62.112 53.049
488897 95.347 -62.117 53.024
488898 95.388 -62.121 53.000
488899 95.428 -62.125 52.978
488900 95.469 -62.130 52.955
488901 95.510 -62.134 52.932
488902 95.550 -62.138 52.910
488903 95.592 -62.142 52.886
488904 95.640 -62.142 52.861
488905 95.688 -62.142 52.833
488906 95.745 -62.152 52.801
488907 95.812 -62.173 52.767
488908 95.868 -62.178 52.730
488909 95.905 -62.163 52.690
488910 95.944 -62.148 52.650
488911 95.979 -62.149 52.609
488912 96.014 -62.154 52.567
488913 96.049 -62.159 52.528
488914 96.084 -62.164 52.489
488915 96.119 -62.168 52.448
488916 96.157 -62.172 52.408
488917 96.200 -62.174 52.368
488918 96.243 -62.176 52.327
488919 96.287 -62.177 52.286
488920 96.333 -62.171 52.246
488921 96.385 -62.144 52.208
488922 96.437 -62.117 52.167
488923 96.488 -62.092 52.125
488924 96.535 -62.098 52.082
488925 96.575 -62.117 52.038
488926 96.614 -62.137 51.993
488927 96.654 -62.157 51.947
488928 96.694 -62.177 51.902
488929 96.733 -62.196 51.856
488930 96.772 -62.215 51.810
488931 96.810 -62.221 51.762
488932 96.847 -62.227 51.717
488933 96.885 -62.233 51.674
488934 96.922 -62.239 51.632
488935 96.959 -62.246 51.592
488936 96.997 -62.252 51.554
488937 97.035 -62.257 51.517
488938 97.075 -62.262 51.482
488939 97.114 -62.268 51.447
488940 97.147 -62.275 51.413
488941 97.175 -62.284 51.379
488942 97.203 -62.294 51.345
488943 97.231 -62.303 51.311
488944 97.259 -62.313 51.278
488945 97.287 -62.322 51.244
488946 97.315 -62.332 51.210
488947 97.343 -62.341 51.174
488948 97.371 -62.351 51.138
488949 97.399 -62.360 51.102
488950 97.427 -62.370 51.065
488951 97.455 -62.379 51.028
488952 97.510 -62.376 50.990
488953 97.577 -62.368 50.952
488954 97.644 -62.359 50.914
488955 97.711 -62.351 50.876
488956 97.726 -62.327 50.838
488957 97.744 -62.305 50.800
488958 97.777 -62.293 50.761
488959 97.810 -62.281 50.722
488960 97.844 -62.269 50.682
488961 97.895 -62.281 50.643
488962 97.949 -62.295 50.603
488963 97.989 -62.309 50.564
488964 97.999 -62.324 50.523
488965 98.010 -62.340 50.482
488966 98.033 -62.355 50.441
488967 98.057 -62.371 50.400
488968 98.081 -62.387 50.359
488969 98.105 -62.402 50.318
488970 98.130 -62.418 50.278
488971 98.154 -62.434 50.238
488972 98.178 -62.449 50.197
488973 98.202 -62.465 50.157
488974 98.226 -62.480 50.119
489424 88.104 -60.349 55.741
489425 88.140 -60.350 55.721
489426 88.177 -60.353 55.701
489427 88.214 -60.354 55.682
489428 88.259 -60.359 55.663
489429 88.311 -60.365 55.644
489430 88.363 -60.373 55.626
489431 88.415 -60.380 55.608
489432 88.467 -60.387 55.591
489433 88.524 -60.380 55.575
489434 88.582 -60.364 55.559
489435 88.642 -60.349 55.545
489436 88.701 -60.333 55.530
489437 88.759 -60.318 55.513
489438 88.816 -60.307 55.494
489439 88.872 -60.296 55.474
489440 88.933 -60.293 55.451
489441 88.976 -60.311 55.424
489442 89.020 -60.329 55.396
489443 89.062 -60.347 55.369
489444 89.106 -60.368 55.341
489445 89.148 -60.389 55.315
489446 89.191 -60.411 55.288
489447 89.234 -60.432 55.262
489448 89.276 -60.454 55.237
489449 89.319 -60.475 55.213
489450 89.362 -60.497 55.189
489451 89.404 -60.519 55.168
489452 89.447 -60.540 55.148
489453 89.489 -60.562 55.131
489454 89.532 -60.583 55.114
489455 89.575 -60.604 55.099
489456 89.617 -60.626 55.086
489457 89.660 -60.647 55.074
489458 89.703 -60.669 55.064
489459 89.745 -60.691 55.055
489460 89.788 -60.713 55.046
489461 89.831 -60.734 55.040
489462 89.873 -60.755 55.037
489463 89.916 -60.777 55.035
489464 89.959 -60.798 55.033
489465 90.002 -60.820 55.035
489466 90.044 -60.842 55.034
489467 90.087 -60.864 55.031
489468 90.127 -60.882 55.025
489469 90.168 -60.899 55.015
489470 90.207 -60.916 55.003
489471 90.248 -60.933 54.995
489472 90.288 -60.951 54.991
489473 90.328 -60.969 54.989
489474 90.370 -60.990 54.989
489475 90.411 -61.011 54.990
489476 90.452 -61.025 54.992
489477 90.494 -61.036 54.996
489478 90.537 -61.046 55.002
489479 90.587 -61.057 55.010
489480 90.636 -61.068 55.014
489481 90.684 -61.079 55.017
489482 90.733 -61.090 55.022
489483 90.782 -61.102 55.031
489484 90.831 -61.112 55.041
489485 90.880 -61.124 55.047
489486 90.919 -61.135 55.053
489487 90.957 -61.147 55.061
489488 90.996 -61.159 55.071
489489 91.034 -61.171 55.077
489490 91.073 -61.183 55.077
489491 91.111 -61.195 55.074
489492 91.150 -61.206 55.070
489493 91.188 -61.218 55.067
489494 91.227 -61.230 55.062
489495 91.265 -61.242 55.054
489496 91.304 -61.254 55.043
489497 91.343 -61.266 55.031
489498 91.383 -61.277 55.019
489499 91.426 -61.288 55.009
489500 91.468 -61.299 55.004
489501 91.510 -61.310 55.003
489502 91.553 -61.321 55.004
489503 91.595 -61.332 55.006
489504 91.637 -61.343 55.011
489505 91.678 -61.354 55.018
489506 91.720 -61.365 55.021
489507 91.761 -61.376 55.018
489508 91.802 -61.387 55.008
489509 91.842 -61.393 54.992
489510 91.881 -61.398 54.975
489511 91.920 -61.404 54.966
489512 91.959 -61.410 54.966
489513 91.999 -61.416 54.969
489514 92.038 -61.422 54.966
489515 92.082 -61.435 54.951
489516 92.130 -61.448 54.924
489517 92.177 -61.460 54.890
489518 92.225 -61.472 54.852
489519 92.273 -61.485 54.816
489520 92.320 -61.497 54.800
489521 92.368 -61.509 54.792
489522 92.416 -61.522 54.774
489523 92.463 -61.534 54.750
489524 92.511 -61.547 54.727
489525 92.559 -61.559 54.719
489526 92.595 -61.566 54.717
489527 92.631 -61.572 54.709
489528 92.665 -61.580 54.694
489529 92.700 -61.587 54.672
489530 92.735 -61.594 54.645
489531 92.770 -61.600 54.618
489532 92.805 -61.608 54.588
489533 92.840 -61.615 54.553
489534 92.875 -61.622 54.515
489535 92.910 -61.628 54.484
489536 92.944 -61.636 54.468
489537 92.980 -61.643 54.462
489538 93.014 -61.650 54.459
489539 93.049 -61.657 54.454
489540 93.084 -61.664 54.442
489541 93.119 -61.671 54.422
489542 93.155 -61.678 54.395
489543 93.196 -61.686 54.363
489544 93.236 -61.694 54.328
489545 93.279 -61.704 54.293
489546 93.322 -61.715 54.259
489547 93.365 -61.725 54.227
489548 93.408 -61.735 54.195
489549 93.451 -61.746 54.163
489550 93.495 -61.757 54.134
489551 93.538 -61.766 54.106
489552 93.577 -61.777 54.078
489553 93.615 -61.786 54.052
489554 93.653 -61.796 54.028
489555 93.691 -61.806 54.005
489556 93.729 -61.816 53.985
489557 93.770 -61.827 53.963
489558 93.811 -61.838 53.943
489559 93.853 -61.850 53.925
489560 93.894 -61.861 53.905
489561 93.936 -61.873 53.887
489562 93.977 -61.884 53.869
489563 94.018 -61.895 53.850
489564 94.060 -61.907 53.831
489565 94.103 -61.915 53.812
489566 94.145 -61.923 53.794
489567 94.188 -61.930 53.775
489568 94.230 -61.938 53.756
489569 94.273 -61.945 53.736
489570 94.316 -61.953 53.715
489571 94.358 -61.960 53.693
489572 94.401 -61.967 53.671
489573 94.444 -61.975 53.649
489574 94.487 -61.983 53.628
489575 94.529 -61.990 53.607
489576 94.572 -61.998 53.585
489577 94.615 -62.005 53.563
489578 94.654 -62.009 53.544
489579 94.691 -62.012 53.527
489580 94.726 -62.014 53.512
489581 94.761 -62.015 53.495
489582 94.796 -62.017 53.473
489583 94.832 -62.018 53.450
489584 94.862 -62.020 53.426
489585 94.890 -62.021 53.403
489586 94.918 -62.023 53.380
489587 94.946 -62.024 53.356
489588 94.974 -62.026 53.330
489589 95.007 -62.030 53.303
489590 95.045 -62.036 53.275
489591 95.082 -62.042 53.245
489592 95.119 -62.048 53.215
489593 95.159 -62.054 53.185
489594 95.199 -62.058 53.153
489595 95.240 -62.062 53.121
489596 95.280 -62.067 53.089
489597 95.321 -62.070 53.062
489598 95.361 -62.075 53.036
489599 95.402 -62.080 53.011
489600 95.443 -62.083 52.985
489601 95.483 -62.088 52.959
489602 95.524 -62.092 52.933
489603 95.564 -62.096 52.907
489604 95.610 -62.098 52.881
489605 95.658 -62.098 52.855
489606 95.705 -62.098 52.827
489607 95.764 -62.111 52.794
489608 95.832 -62.131 52.760
489609 95.875 -62.121 52.722
489610 95.913 -62.106 52.683
489611 95.949 -62.105 52.644
489612 95.984 -62.109 52.606
489613 96.019 -62.114 52.568
489614 96.054 -62.119 52.532
489615 96.089 -62.123 52.493
489616 96.124 -62.128 52.450
489617 96.160 -62.133 52.408
489618 96.203 -62.134 52.368
489619 96.246 -62.136 52.326
489620 96.290 -62.138 52.285
489621 96.333 -62.139 52.245
489622 96.380 -62.131 52.206
489623 96.432 -62.104 52.164
489624 96.484 -62.077 52.122
489625 96.536 -62.049 52.079
489626 96.579 -62.055 52.035
489627 96.618 -62.074 51.990
489628 96.658 -62.094 51.944
489629 96.697 -62.114 51.900
489630 96.737 -62.133 51.855
489631 96.777 -62.153 51.810
489632 96.816 -62.174 51.764
489633 96.854 -62.180 51.718
489634 96.891 -62.186 51.674
489635 96.928 -62.192 51.632
489636 96.966 -62.199 51.592
489637 97.003 -62.205 51.553
489638 97.042 -62.209 51.516
489639 97.082 -62.214 51.480
489640 97.113 -62.223 51.445
489641 97.142 -62.232 51.411
489642 97.169 -62.242 51.377
489643 97.197 -62.251 51.343
489644 97.226 -62.261 51.310
489645 97.254 -62.270 51.276
489646 97.282 -62.280 51.243
489647 97.310 -62.289 51.208
489648 97.338 -62.299 51.172
489649 97.366 -62.308 51.136
489650 97.394 -62.318 51.099
489651 97.422 -62.327 51.063
489652 97.450 -62.337 51.025
489653 97.512 -62.330 50.987
489654 97.579 -62.322 50.949
489655 97.646 -62.313 50.911
489656 97.713 -62.305 50.873
489657 97.733 -62.283 50.834
489658 97.752 -62.261 50.797
489659 97.785 -62.249 50.757
489660 97.818 -62.237 50.718
489661 97.855 -62.230 50.679
489662 97.909 -62.245 50.639
489663 97.962 -62.259 50.599
489664 97.981 -62.274 50.560
489665 97.992 -62.289 50.519
489666 98.010 -62.304 50.478
489667 98.034 -62.320 50.437
489668 98.059 -62.336 50.397
489669 98.083 -62.351 50.356
489670 98.107 -62.367 50.315
489671 98.131 -62.382 50.275
489672 98.155 -62.398 50.235
489673 98.179 -62.414 50.195
489674 98.204 -62.429 50.155
489675 98.228 -62.445 50.116
490126 88.155 -60.303 55.710
490127 88.192 -60.305 55.689
490128 88.234 -60.308 55.669
490129 88.286 -60.315 55.651
490130 88.338 -60.322 55.632
490131 88.390 -60.328 55.613
490132 88.442 -60.336 55.594
490133 88.494 -60.343 55.577
490134 88.552 -60.331 55.562
490135 88.611 -60.316 55.546
490136 88.670 -60.300 55.531
490137 88.729 -60.284 55.515
490138 88.786 -60.271 55.499
490139 88.843 -60.260 55.481
490140 88.899 -60.249 55.463
490141 88.947 -60.249 55.442
490142 88.989 -60.269 55.418
490143 89.032 -60.290 55.391
490144 89.075 -60.312 55.363
490145 89.118 -60.334 55.336
490146 89.160 -60.355 55.310
490147 89.203 -60.377 55.284
490148 89.246 -60.398 55.258
490149 89.288 -60.419 55.233
490150 89.331 -60.441 55.209
490151 89.374 -60.463 55.186
490152 89.416 -60.484 55.164
490153 89.459 -60.506 55.144
490154 89.502 -60.527 55.127
490155 89.545 -60.548 55.110
490156 89.587 -60.570 55.095
490157 89.630 -60.591 55.081
490158 89.673 -60.613 55.068
490159 89.716 -60.634 55.056
490160 89.758 -60.656 55.047
490161 89.801 -60.677 55.039
490162 89.843 -60.698 55.031
490163 89.886 -60.720 55.025
490164 89.929 -60.742 55.021
490165 89.972 -60.763 55.019
490166 90.015 -60.785 55.020
490167 90.057 -60.806 55.024
490168 90.100 -60.827 55.024
490169 90.144 -60.847 55.019
490170 90.182 -60.865 55.009
490171 90.220 -60.884 54.997
490172 90.259 -60.902 54.989
490173 90.298 -60.920 54.985
490174 90.338 -60.937 54.983
490175 90.378 -60.955 54.984
490176 90.419 -60.975 54.985
490177 90.461 -60.989 54.987
490178 90.506 -61.000 54.991
490179 90.555 -61.011 54.999
490180 90.604 -61.022 55.009
490181 90.653 -61.033 55.015
490182 90.702 -61.044 55.019
490183 90.751 -61.055 55.025
490184 90.800 -61.067 55.033
490185 90.849 -61.077 55.045
490186 90.894 -61.089 55.051
490187 90.932 -61.101 55.057
490188 90.971 -61.112 55.066
490189 91.009 -61.124 55.078
490190 91.048 -61.136 55.086
490191 91.086 -61.148 55.088
490192 91.125 -61.160 55.085
490193 91.163 -61.172 55.081
490194 91.202 -61.183 55.075
490195 91.240 -61.196 55.070
490196 91.279 -61.208 55.063
490197 91.317 -61.219 55.053
490198 91.356 -61.231 55.042
490199 91.394 -61.243 55.029
490200 91.435 -61.255 55.014
490201 91.477 -61.265 55.001
490202 91.520 -61.277 54.995
490203 91.562 -61.287 54.991
490204 91.604 -61.298 54.989
490205 91.646 -61.309 54.990
490206 91.687 -61.320 54.996
490207 91.729 -61.331 54.999
490208 91.771 -61.343 54.997
490209 91.811 -61.352 54.988
490210 91.851 -61.358 54.975
490211 91.890 -61.363 54.963
490212 91.929 -61.369 54.959
490213 91.968 -61.375 54.962
490214 92.008 -61.381 54.965
490215 92.047 -61.387 54.961
490216 92.088 -61.397 54.946
490217 92.136 -61.409 54.918
490218 92.183 -61.422 54.885
490219 92.231 -61.434 54.849
490220 92.279 -61.447 54.816
490221 92.326 -61.459 54.801
490222 92.374 -61.472 54.796
490223 92.422 -61.484 54.781
490224 92.470 -61.497 54.755
490225 92.517 -61.509 54.725
490226 92.565 -61.522 54.710
490227 92.604 -61.528 54.702
490228 92.643 -61.534 54.692
490229 92.678 -61.541 54.677
490230 92.713 -61.548 54.657
490231 92.747 -61.556 54.630
490232 92.783 -61.563 54.602
490233 92.817 -61.569 54.574
490234 92.852 -61.576 54.543
490235 92.887 -61.583 54.510
490236 92.922 -61.591 54.488
490237 92.957 -61.597 54.480
490238 92.992 -61.604 54.476
490239 93.027 -61.611 54.472
490240 93.062 -61.619 54.463
490241 93.096 -61.625 54.446
490242 93.132 -61.632 54.420
490243 93.166 -61.639 54.390
490244 93.205 -61.647 54.356
490245 93.246 -61.656 54.321
490246 93.286 -61.664 54.288
490247 93.327 -61.672 54.256
490248 93.368 -61.682 54.226
490249 93.412 -61.691 54.195
490250 93.455 -61.702 54.166
490251 93.498 -61.713 54.137
490252 93.542 -61.723 54.110
490253 93.583 -61.733 54.082
490254 93.622 -61.743 54.056
490255 93.659 -61.753 54.031
490256 93.698 -61.763 54.007
490257 93.737 -61.773 53.986
490258 93.778 -61.785 53.965
490259 93.820 -61.796 53.944
490260 93.861 -61.807 53.924
490261 93.903 -61.819 53.903
490262 93.944 -61.830 53.883
490263 93.986 -61.841 53.863
490264 94.028 -61.853 53.844
490265 94.069 -61.864 53.824
490266 94.111 -61.874 53.805
490267 94.154 -61.882 53.787
490268 94.196 -61.889 53.769
490269 94.239 -61.896 53.750
490270 94.282 -61.904 53.730
490271 94.325 -61.911 53.708
490272 94.367 -61.919 53.687
490273 94.410 -61.926 53.666
490274 94.453 -61.934 53.646
490275 94.495 -61.942 53.627
490276 94.538 -61.948 53.608
490277 94.580 -61.956 53.589
490278 94.621 -61.962 53.569
490279 94.660 -61.966 53.553
490280 94.696 -61.968 53.540
490281 94.732 -61.970 53.530
490282 94.767 -61.972 53.517
490283 94.803 -61.973 53.499
490284 94.838 -61.975 53.479
490285 94.874 -61.976 53.459
490286 94.909 -61.978 53.439
490287 94.940 -61.979 53.419
490288 94.968 -61.981 53.395
490289 94.996 -61.983 53.370
490290 95.027 -61.986 53.344
490291 95.064 -61.992 53.316
490292 95.101 -61.998 53.286
490293 95.138 -62.004 53.253
490294 95.176 -62.011 53.220
490295 95.213 -62.016 53.185
490296 95.254 -62.020 53.148
490297 95.295 -62.024 53.112
490298 95.335 -62.029 53.080
490299 95.376 -62.033 53.051
490300 95.416 -62.037 53.023
490301 95.457 -62.042 52.994
490302 95.497 -62.046 52.966
490303 95.538 -62.050 52.937
490304 95.580 -62.054 52.909
490305 95.627 -62.055 52.881
490306 95.675 -62.055 52.851
490307 95.723 -62.055 52.820
490308 95.784 -62.069 52.786
490309 95.844 -62.080 52.751
490310 95.882 -62.065 52.714
490311 95.918 -62.059 52.676
490312 95.953 -62.064 52.638
490313 95.988 -62.069 52.600
490314 96.024 -62.073 52.564
490315 96.058 -62.079 52.527
490316 96.094 -62.083 52.489
490317 96.128 -62.088 52.450
490318 96.164 -62.093 52.408
490319 96.206 -62.095 52.367
490320 96.249 -62.097 52.324
490321 96.293 -62.098 52.283
490322 96.336 -62.100 52.244
490323 96.380 -62.102 52.204
490324 96.427 -62.092 52.163
490325 96.479 -62.064 52.122
490326 96.531 -62.037 52.079
490327 96.577 -62.036 52.035
490328 96.619 -62.040 51.990
490329 96.660 -62.044 51.944
490330 96.701 -62.051 51.899
490331 96.741 -62.070 51.855
490332 96.781 -62.090 51.811
490333 96.820 -62.110 51.767
490334 96.860 -62.130 51.721
490335 96.897 -62.139 51.677
490336 96.935 -62.145 51.634
490337 96.972 -62.151 51.593
490338 97.010 -62.157 51.553
490339 97.050 -62.161 51.516
490340 97.080 -62.171 51.480
490341 97.108 -62.180 51.445
490342 97.136 -62.189 51.411
490343 97.164 -62.199 51.376
490344 97.192 -62.208 51.342
490345 97.220 -62.218 51.309
490346 97.248 -62.227 51.276
490347 97.276 -62.236 51.242
490348 97.304 -62.246 51.208
490349 97.332 -62.255 51.172
490350 97.360 -62.265 51.135
490351 97.388 -62.274 51.098
490352 97.417 -62.284 51.061
490353 97.446 -62.293 51.023
490354 97.513 -62.284 50.985
490355 97.580 -62.276 50.947
490356 97.647 -62.268 50.908
490357 97.714 -62.259 50.870
490358 97.741 -62.239 50.832
490359 97.760 -62.218 50.794
490360 97.793 -62.205 50.754
490361 97.826 -62.193 50.714
490362 97.869 -62.195 50.675
490363 97.923 -62.208 50.636
490364 97.962 -62.223 50.596
490365 97.973 -62.238 50.557
490366 97.988 -62.253 50.516
490367 98.012 -62.269 50.475
490368 98.036 -62.284 50.434
490369 98.061 -62.300 50.393
490370 98.085 -62.316 50.353
490371 98.109 -62.331 50.312
490372 98.133 -62.347 50.272
490373 98.157 -62.363 50.232
490374 98.181 -62.378 50.192
490375 98.206 -62.394 50.152
490376 98.230 -62.409 50.114
490828 88.208 -60.257 55.678
490829 88.260 -60.264 55.657
490830 88.313 -60.271 55.638
490831 88.364 -60.278 55.620
490832 88.416 -60.284 55.601
490833 88.469 -60.291 55.582
490834 88.521 -60.299 55.564
490835 88.579 -60.283 55.549
490836 88.639 -60.268 55.533
490837 88.698 -60.252 55.517
490838 88.757 -60.237 55.501
490839 88.814 -60.225 55.485
490840 88.870 -60.214 55.468
490841 88.926 -60.202 55.451
490842 88.967 -60.217 55.432
490843 89.003 -60.233 55.409
490844 89.046 -60.254 55.384
490845 89.089 -60.275 55.358
490846 89.131 -60.297 55.331
490847 89.174 -60.318 55.305
490848 89.216 -60.340 55.279
490849 89.259 -60.362 55.254
490850 89.302 -60.383 55.231
490851 89.345 -60.404 55.206
490852 89.387 -60.426 55.183
490853 89.430 -60.447 55.161
490854 89.473 -60.469 55.142
490855 89.516 -60.491 55.124
490856 89.558 -60.512 55.106
490857 89.601 -60.533 55.091
490858 89.644 -60.554 55.077
490859 89.686 -60.576 55.064
490860 89.729 -60.597 55.052
490861 89.772 -60.619 55.042
490862 89.815 -60.641 55.033
490863 89.857 -60.662 55.024
490864 89.900 -60.683 55.018
490865 89.943 -60.705 55.013
490866 89.985 -60.726 55.010
490867 90.028 -60.748 55.009
490868 90.071 -60.770 55.010
490869 90.115 -60.790 55.010
490870 90.161 -60.810 55.007
490871 90.199 -60.829 55.000
490872 90.238 -60.847 54.990
490873 90.276 -60.866 54.984
490874 90.315 -60.884 54.980
490875 90.354 -60.902 54.977
490876 90.392 -60.921 54.978
490877 90.431 -60.939 54.980
490878 90.475 -60.954 54.982
490879 90.524 -60.965 54.986
490880 90.573 -60.976 54.994
490881 90.621 -60.987 55.006
490882 90.671 -60.998 55.015
490883 90.720 -61.009 55.021
490884 90.768 -61.020 55.027
490885 90.817 -61.031 55.034
490886 90.866 -61.042 55.045
490887 90.907 -61.054 55.052
490888 90.946 -61.066 55.058
490889 90.984 -61.078 55.067
490890 91.023 -61.089 55.079
490891 91.061 -61.102 55.090
490892 91.100 -61.114 55.094
490893 91.138 -61.125 55.094
490894 91.177 -61.137 55.091
490895 91.215 -61.149 55.086
490896 91.254 -61.161 55.080
490897 91.292 -61.173 55.074
490898 91.331 -61.185 55.065
490899 91.369 -61.196 55.055
490900 91.408 -61.208 55.042
490901 91.446 -61.221 55.028
490902 91.487 -61.232 55.006
490903 91.529 -61.243 54.992
490904 91.571 -61.254 54.983
490905 91.613 -61.265 54.976
490906 91.656 -61.276 54.970
490907 91.698 -61.287 54.970
490908 91.739 -61.298 54.972
490909 91.780 -61.309 54.971
490910 91.820 -61.317 54.966
490911 91.860 -61.322 54.956
490912 91.899 -61.328 54.951
490913 91.938 -61.334 54.953
490914 91.978 -61.340 54.957
490915 92.017 -61.347 54.958
490916 92.056 -61.356 54.953
490917 92.095 -61.363 54.938
490918 92.140 -61.375 54.910
490919 92.188 -61.387 54.877
490920 92.235 -61.400 54.843
490921 92.283 -61.412 54.813
490922 92.331 -61.425 54.804
490923 92.378 -61.437 54.800
490924 92.426 -61.450 54.787
490925 92.473 -61.462 54.762
490926 92.521 -61.475 54.728
490927 92.568 -61.487 54.703
490928 92.607 -61.493 54.690
490929 92.646 -61.499 54.678
490930 92.684 -61.505 54.664
490931 92.722 -61.511 54.648
490932 92.760 -61.517 54.627
490933 92.795 -61.524 54.604
490934 92.830 -61.531 54.577
490935 92.865 -61.538 54.546
490936 92.900 -61.545 54.513
490937 92.935 -61.552 54.496
490938 92.970 -61.559 54.492
490939 93.004 -61.566 54.489
490940 93.040 -61.573 54.481
490941 93.074 -61.580 54.467
490942 93.109 -61.587 54.443
490943 93.144 -61.594 54.411
490944 93.179 -61.601 54.378
490945 93.214 -61.608 54.344
490946 93.255 -61.616 54.312
490947 93.295 -61.625 54.281
490948 93.336 -61.633 54.252
490949 93.377 -61.641 54.224
490950 93.418 -61.650 54.197
490951 93.459 -61.658 54.169
490952 93.502 -61.669 54.142
490953 93.545 -61.679 54.115
490954 93.588 -61.690 54.088
490955 93.628 -61.700 54.063
490956 93.666 -61.710 54.038
490957 93.704 -61.719 54.013
490958 93.746 -61.731 53.990
490959 93.787 -61.742 53.968
490960 93.829 -61.754 53.946
490961 93.870 -61.765 53.924
490962 93.912 -61.776 53.903
490963 93.954 -61.788 53.882
490964 93.995 -61.799 53.862
490965 94.037 -61.810 53.842
490966 94.078 -61.822 53.823
490967 94.120 -61.832 53.804
490968 94.163 -61.840 53.786
490969 94.205 -61.848 53.767
490970 94.248 -61.855 53.749
490971 94.290 -61.863 53.729
490972 94.333 -61.870 53.708
490973 94.376 -61.877 53.687
490974 94.418 -61.885 53.667
490975 94.461 -61.892 53.648
490976 94.504 -61.900 53.631
490977 94.546 -61.907 53.614
490978 94.589 -61.914 53.594
490979 94.627 -61.918 53.577
490980 94.666 -61.923 53.563
490981 94.702 -61.925 53.552
490982 94.738 -61.926 53.543
490983 94.773 -61.928 53.533
490984 94.809 -61.929 53.519
490985 94.844 -61.931 53.505
490986 94.880 -61.933 53.490
490987 94.915 -61.935 53.474
490988 94.950 -61.936 53.455
490989 94.986 -61.938 53.435
490990 95.019 -61.939 53.411
490991 95.047 -61.941 53.386
490992 95.084 -61.947 53.360
490993 95.121 -61.953 53.332
490994 95.158 -61.960 53.299
490995 95.195 -61.966 53.264
490996 95.232 -61.972 53.224
490997 95.269 -61.978 53.182
490998 95.309 -61.983 53.141
490999 95.349 -61.987 53.104
491000 95.390 -61.992 53.069
491001 95.431 -61.995 53.036
491002 95.471 -62.000 53.005
491003 95.512 -62.004 52.974
491004 95.552 -62.008 52.943
491005 95.597 -62.011 52.912
491006 95.645 -62.011 52.881
491007 95.692 -62.011 52.849
491008 95.740 -62.011 52.816
491009 95.804 -62.028 52.780
491010 95.851 -62.023 52.744
491011 95.888 -62.014 52.707
491012 95.923 -62.019 52.670
491013 95.958 -62.024 52.633
491014 95.993 -62.029 52.596
491015 96.028 -62.033 52.558
491016 96.063 -62.039 52.522
491017 96.098 -62.043 52.485
491018 96.133 -62.048 52.447
491019 96.168 -62.053 52.406
491020 96.209 -62.055 52.367
491021 96.252 -62.057 52.326
491022 96.296 -62.059 52.286
491023 96.339 -62.061 52.248
491024 96.383 -62.062 52.208
491025 96.426 -62.064 52.166
491026 96.473 -62.051 52.124
491027 96.526 -62.026 52.081
491028 96.575 -62.022 52.037
491029 96.616 -62.026 51.992
491030 96.658 -62.030 51.947
491031 96.699 -62.034 51.901
491032 96.741 -62.038 51.857
491033 96.783 -62.042 51.813
491034 96.824 -62.047 51.768
491035 96.864 -62.067 51.723
491036 96.903 -62.087 51.679
491037 96.941 -62.098 51.637
491038 96.978 -62.104 51.595
491039 97.017 -62.109 51.555
491040 97.046 -62.118 51.517
491041 97.075 -62.127 51.481
491042 97.102 -62.137 51.446
491043 97.131 -62.146 51.411
491044 97.159 -62.156 51.377
491045 97.186 -62.165 51.343
491046 97.215 -62.175 51.310
491047 97.243 -62.184 51.276
491048 97.271 -62.194 51.242
491049 97.299 -62.203 51.208
491050 97.327 -62.213 51.172
491051 97.355 -62.222 51.135
491052 97.383 -62.232 51.098
491053 97.411 -62.241 51.061
491054 97.447 -62.247 51.022
491055 97.514 -62.238 50.984
491056 97.581 -62.230 50.946
491057 97.648 -62.221 50.907
491058 97.715 -62.213 50.868
491059 97.748 -62.195 50.829
491060 97.768 -62.174 50.791
491061 97.801 -62.161 50.752
491062 97.834 -62.150 50.712
491063 97.883 -62.158 50.672
491064 97.937 -62.173 50.632
491065 97.955 -62.187 50.593
491066 97.966 -62.202 50.553
491067 97.990 -62.218 50.513
491068 98.014 -62.233 50.472
491069 98.038 -62.249 50.431
491070 98.062 -62.265 50.390
491071 98.086 -62.280 50.349
491072 98.110 -62.296 50.310
491073 98.135 -62.312 50.269
491074 98.159 -62.327 50.229
491075 98.183 -62.343 50.188
491076 98.207 -62.359 50.149
491077 98.231 -62.374 50.111
491529 88.235 -60.213 55.669
491530 88.287 -60.220 55.648
491531 88.339 -60.227 55.628
491532 88.391 -60.234 55.608
491533 88.444 -60.240 55.589
491534 88.495 -60.247 55.571
491535 88.549 -60.250 55.552
491536 88.608 -60.235 55.537
491537 88.667 -60.219 55.520
491538 88.726 -60.204 55.504
491539 88.784 -60.190 55.488
491540 88.841 -60.178 55.473
491541 88.897 -60.167 55.455
491542 88.948 -60.165 55.439
491543 88.988 -60.186 55.421
491544 89.018 -60.198 55.400
491545 89.059 -60.218 55.377
491546 89.102 -60.239 55.351
491547 89.145 -60.260 55.325
491548 89.187 -60.282 55.300
491549 89.230 -60.303 55.275
491550 89.273 -60.325 55.251
491551 89.316 -60.347 55.227
491552 89.358 -60.368 55.203
491553 89.401 -60.389 55.181
491554 89.444 -60.410 55.159
491555 89.486 -60.432 55.139
491556 89.529 -60.453 55.122
491557 89.572 -60.475 55.105
491558 89.615 -60.497 55.089
491559 89.657 -60.518 55.075
491560 89.700 -60.539 55.061
491561 89.743 -60.561 55.050
491562 89.786 -60.582 55.038
491563 89.828 -60.604 55.028
491564 89.871 -60.626 55.020
491565 89.913 -60.647 55.013
491566 89.956 -60.668 55.007
491567 89.999 -60.690 55.003
491568 90.042 -60.711 55.000
491569 90.085 -60.732 54.999
491570 90.130 -60.752 54.998
491571 90.176 -60.773 54.996
491572 90.218 -60.792 54.992
491573 90.256 -60.810 54.984
491574 90.295 -60.829 54.979
491575 90.333 -60.848 54.975
491576 90.371 -60.866 54.973
491577 90.410 -60.884 54.974
491578 90.448 -60.903 54.975
491579 90.493 -60.919 54.978
491580 90.541 -60.930 54.982
491581 90.590 -60.941 54.989
491582 90.639 -60.952 54.998
491583 90.688 -60.964 55.011
491584 90.737 -60.974 55.020
491585 90.786 -60.986 55.026
491586 90.835 -60.996 55.033
491587 90.882 -61.008 55.043
491588 90.921 -61.020 55.050
491589 90.959 -61.031 55.057
491590 90.998 -61.043 55.065
491591 91.036 -61.055 55.077
491592 91.075 -61.067 55.089
491593 91.113 -61.079 55.097
491594 91.152 -61.091 55.101
491595 91.190 -61.103 55.101
491596 91.229 -61.114 55.097
491597 91.267 -61.127 55.091
491598 91.306 -61.139 55.085
491599 91.344 -61.150 55.077
491600 91.383 -61.162 55.066
491601 91.421 -61.174 55.052
491602 91.460 -61.186 55.037
491603 91.498 -61.198 55.015
491604 91.538 -61.209 54.997
491605 91.581 -61.220 54.984
491606 91.623 -61.231 54.974
491607 91.665 -61.242 54.962
491608 91.707 -61.253 54.955
491609 91.748 -61.264 54.950
491610 91.790 -61.275 54.947
491611 91.829 -61.282 54.944
491612 91.869 -61.287 54.938
491613 91.908 -61.293 54.941
491614 91.947 -61.299 54.945
491615 91.987 -61.306 54.947
491616 92.025 -61.315 54.947
491617 92.065 -61.323 54.943
491618 92.103 -61.331 54.928
491619 92.144 -61.340 54.899
491620 92.192 -61.353 54.865
491621 92.239 -61.365 54.832
491622 92.287 -61.377 54.809
491623 92.334 -61.390 54.808
491624 92.382 -61.402 54.805
491625 92.429 -61.415 54.792
491626 92.477 -61.427 54.769
491627 92.525 -61.440 54.734
491628 92.572 -61.451 54.699
491629 92.610 -61.457 54.681
491630 92.648 -61.463 54.668
491631 92.687 -61.469 54.656
491632 92.725 -61.475 54.642
491633 92.764 -61.481 54.623
491634 92.802 -61.487 54.602
491635 92.841 -61.493 54.574
491636 92.877 -61.500 54.544
491637 92.912 -61.506 54.519
491638 92.947 -61.513 54.512
491639 92.982 -61.521 54.509
491640 93.017 -61.528 54.501
491641 93.052 -61.534 54.487
491642 93.087 -61.541 54.466
491643 93.122 -61.549 54.438
491644 93.156 -61.556 54.403
491645 93.191 -61.563 54.370
491646 93.226 -61.569 54.339
491647 93.264 -61.577 54.309
491648 93.304 -61.585 54.280
491649 93.345 -61.594 54.254
491650 93.386 -61.602 54.228
491651 93.427 -61.610 54.202
491652 93.468 -61.619 54.176
491653 93.508 -61.628 54.150
491654 93.549 -61.636 54.123
491655 93.591 -61.646 54.098
491656 93.635 -61.656 54.074
491657 93.673 -61.666 54.049
491658 93.713 -61.677 54.023
491659 93.755 -61.688 53.998
491660 93.796 -61.700 53.973
491661 93.838 -61.711 53.949
491662 93.880 -61.722 53.927
491663 93.921 -61.734 53.906
491664 93.963 -61.745 53.885
491665 94.004 -61.757 53.865
491666 94.046 -61.768 53.845
491667 94.087 -61.779 53.825
491668 94.129 -61.790 53.807
491669 94.171 -61.799 53.788
491670 94.213 -61.807 53.769
491671 94.256 -61.813 53.751
491672 94.299 -61.821 53.731
491673 94.342 -61.829 53.710
491674 94.384 -61.836 53.689
491675 94.427 -61.844 53.669
491676 94.470 -61.851 53.651
491677 94.512 -61.859 53.634
491678 94.555 -61.866 53.619
491679 94.595 -61.871 53.600
491680 94.633 -61.875 53.583
491681 94.672 -61.879 53.570
491682 94.708 -61.881 53.560
491683 94.744 -61.882 53.552
491684 94.779 -61.885 53.543
491685 94.815 -61.886 53.533
491686 94.850 -61.888 53.522
491687 94.886 -61.889 53.510
491688 94.921 -61.891 53.496
491689 94.956 -61.892 53.479
491690 94.991 -61.895 53.460
491691 95.027 -61.896 53.438
491692 95.062 -61.898 53.414
491693 95.102 -61.903 53.388
491694 95.140 -61.909 53.360
491695 95.178 -61.915 53.328
491696 95.214 -61.921 53.293
491697 95.252 -61.927 53.255
491698 95.289 -61.933 53.214
491699 95.326 -61.939 53.173
491700 95.364 -61.945 53.131
491701 95.404 -61.949 53.091
491702 95.445 -61.954 53.053
491703 95.486 -61.958 53.018
491704 95.526 -61.962 52.982
491705 95.567 -61.967 52.949
491706 95.614 -61.967 52.916
491707 95.662 -61.967 52.882
491708 95.710 -61.967 52.848
491709 95.758 -61.967 52.814
491710 95.820 -61.982 52.777
491711 95.858 -61.970 52.740
491712 95.893 -61.974 52.703
491713 95.927 -61.979 52.666
491714 95.963 -61.984 52.629
491715 95.998 -61.989 52.592
491716 96.033 -61.993 52.553
491717 96.068 -61.998 52.516
491718 96.103 -62.003 52.479
491719 96.138 -62.008 52.442
491720 96.173 -62.013 52.405
491721 96.212 -62.016 52.369
491722 96.255 -62.017 52.331
491723 96.299 -62.020 52.293
491724 96.342 -62.021 52.255
491725 96.386 -62.023 52.214
491726 96.429 -62.024 52.172
491727 96.472 -62.026 52.128
491728 96.521 -62.015 52.084
491729 96.573 -62.008 52.040
491730 96.614 -62.012 51.995
491731 96.656 -62.016 51.950
491732 96.697 -62.020 51.904
491733 96.739 -62.024 51.860
491734 96.780 -62.028 51.816
491735 96.822 -62.032 51.771
491736 96.863 -62.036 51.727
491737 96.905 -62.040 51.684
491738 96.946 -62.044 51.641
491739 96.985 -62.056 51.599
491740 97.013 -62.066 51.559
491741 97.041 -62.075 51.520
491742 97.069 -62.085 51.484
491743 97.097 -62.094 51.449
491744 97.125 -62.104 51.414
491745 97.153 -62.113 51.380
491746 97.181 -62.123 51.345
491747 97.209 -62.132 51.312
491748 97.237 -62.142 51.279
491749 97.265 -62.151 51.245
491750 97.293 -62.161 51.210
491751 97.321 -62.170 51.175
491752 97.349 -62.180 51.138
491753 97.377 -62.189 51.101
491754 97.405 -62.199 51.063
491755 97.448 -62.201 51.024
491756 97.516 -62.192 50.986
491757 97.583 -62.184 50.947
491758 97.650 -62.176 50.908
491759 97.717 -62.167 50.869
491760 97.755 -62.150 50.829
491761 97.776 -62.130 50.790
491762 97.809 -62.118 50.751
491763 97.844 -62.108 50.711
491764 97.897 -62.122 50.670
491765 97.937 -62.136 50.630
491766 97.947 -62.152 50.592
491767 97.967 -62.167 50.551
491768 97.992 -62.183 50.510
491769 98.016 -62.199 50.470
491770 98.040 -62.214 50.429
491771 98.064 -62.230 50.388
491772 98.088 -62.246 50.348
491773 98.112 -62.261 50.308
491774 98.137 -62.277 50.268
491775 98.161 -62.292 50.227
491776 98.185 -62.308 50.187
491777 98.209 -62.324 50.147
492231 88.314 -60.176 55.639
492232 88.366 -60.183 55.619
492233 88.418 -60.190 55.599
492234 88.470 -60.196 55.580
492235 88.522 -60.203 55.561
492236 88.577 -60.202 55.542
492237 88.636 -60.187 55.526
492238 88.695 -60.171 55.510
492239 88.754 -60.155 55.494
492240 88.812 -60.143 55.477
492241 88.868 -60.131 55.461
492242 88.924 -60.120 55.443
492243 88.968 -60.134 55.428
492244 89.008 -60.153 55.410
492245 89.032 -60.165 55.390
492246 89.073 -60.181 55.368
492247 89.116 -60.202 55.344
492248 89.159 -60.224 55.319
492249 89.201 -60.245 55.295
492250 89.244 -60.267 55.271
492251 89.286 -60.288 55.247
492252 89.329 -60.309 55.223
492253 89.372 -60.331 55.201
492254 89.415 -60.353 55.179
492255 89.457 -60.374 55.158
492256 89.500 -60.395 55.138
492257 89.543 -60.417 55.121
492258 89.586 -60.438 55.104
492259 89.628 -60.459 55.090
492260 89.671 -60.481 55.075
492261 89.713 -60.503 55.061
492262 89.756 -60.524 55.049
492263 89.799 -60.546 55.039
492264 89.842 -60.567 55.028
492265 89.885 -60.588 55.019
492266 89.927 -60.610 55.010
492267 89.970 -60.632 55.004
492268 90.013 -60.653 55.000
492269 90.055 -60.674 54.995
492270 90.100 -60.695 54.991
492271 90.146 -60.715 54.990
492272 90.191 -60.735 54.988
492273 90.236 -60.755 54.986
492274 90.274 -60.773 54.980
492275 90.312 -60.792 54.975
492276 90.351 -60.810 54.971
492277 90.389 -60.829 54.970
492278 90.428 -60.848 54.970
492279 90.466 -60.866 54.972
492280 90.511 -60.882 54.975
492281 90.559 -60.895 54.979
492282 90.608 -60.906 54.984
492283 90.657 -60.917 54.992
492284 90.706 -60.929 55.004
492285 90.755 -60.939 55.016
492286 90.804 -60.951 55.025
492287 90.853 -60.961 55.033
492288 90.895 -60.973 55.041
492289 90.934 -60.985 55.046
492290 90.972 -60.997 55.052
492291 91.011 -61.009 55.060
492292 91.049 -61.020 55.073
492293 91.088 -61.033 55.086
492294 91.127 -61.045 55.096
492295 91.165 -61.056 55.103
492296 91.204 -61.068 55.107
492297 91.242 -61.080 55.107
492298 91.281 -61.092 55.103
492299 91.319 -61.104 55.096
492300 91.358 -61.116 55.087
492301 91.396 -61.127 55.075
492302 91.435 -61.139 55.060
492303 91.473 -61.152 55.044
492304 91.512 -61.163 55.024
492305 91.550 -61.175 55.005
492306 91.590 -61.187 54.989
492307 91.632 -61.198 54.976
492308 91.675 -61.208 54.960
492309 91.717 -61.219 54.946
492310 91.758 -61.230 54.935
492311 91.799 -61.241 54.929
492312 91.839 -61.246 54.927
492313 91.878 -61.252 54.926
492314 91.917 -61.259 54.931
492315 91.956 -61.266 54.934
492316 91.995 -61.274 54.935
492317 92.034 -61.282 54.935
492318 92.073 -61.290 54.932
492319 92.112 -61.299 54.920
492320 92.151 -61.306 54.896
492321 92.195 -61.318 54.865
492322 92.243 -61.330 54.833
492323 92.291 -61.342 54.808
492324 92.338 -61.355 54.807
492325 92.386 -61.367 54.804
492326 92.433 -61.380 54.793
492327 92.481 -61.392 54.771
492328 92.528 -61.405 54.739
492329 92.574 -61.416 54.702
492330 92.613 -61.422 54.677
492331 92.651 -61.428 54.663
492332 92.690 -61.434 54.653
492333 92.728 -61.440 54.641
492334 92.767 -61.446 54.619
492335 92.805 -61.452 54.596
492336 92.844 -61.458 54.566
492337 92.882 -61.463 54.540
492338 92.920 -61.469 54.526
492339 92.958 -61.475 54.529
492340 92.994 -61.482 54.524
492341 93.029 -61.489 54.511
492342 93.064 -61.497 54.490
492343 93.099 -61.503 54.464
492344 93.134 -61.510 54.431
492345 93.169 -61.517 54.398
492346 93.204 -61.524 54.367
492347 93.239 -61.531 54.338
492348 93.274 -61.538 54.310
492349 93.313 -61.547 54.284
492350 93.354 -61.555 54.261
492351 93.395 -61.563 54.238
492352 93.436 -61.572 54.216
492353 93.477 -61.580 54.194
492354 93.517 -61.588 54.170
492355 93.558 -61.597 54.143
492356 93.599 -61.605 54.119
492357 93.640 -61.613 54.096
492358 93.681 -61.623 54.070
492359 93.722 -61.635 54.043
492360 93.764 -61.646 54.016
492361 93.805 -61.657 53.989
492362 93.847 -61.669 53.962
492363 93.888 -61.680 53.938
492364 93.930 -61.691 53.916
492365 93.971 -61.703 53.895
492366 94.013 -61.714 53.874
492367 94.054 -61.726 53.854
492368 94.096 -61.737 53.834
492369 94.137 -61.747 53.814
492370 94.180 -61.757 53.795
492371 94.222 -61.765 53.775
492372 94.265 -61.773 53.755
492373 94.308 -61.780 53.734
492374 94.350 -61.788 53.712
492375 94.393 -61.795 53.692
492376 94.435 -61.802 53.673
492377 94.478 -61.810 53.655
492378 94.521 -61.817 53.639
492379 94.562 -61.823 53.624
492380 94.601 -61.828 53.605
492381 94.639 -61.832 53.589
492382 94.678 -61.835 53.576
492383 94.714 -61.838 53.566
492384 94.750 -61.839 53.559
492385 94.785 -61.841 53.551
492386 94.821 -61.842 53.543
492387 94.856 -61.844 53.534
492388 94.891 -61.846 53.524
492389 94.927 -61.848 53.512
492390 94.962 -61.849 53.496
492391 94.997 -61.851 53.478
492392 95.033 -61.853 53.456
492393 95.069 -61.860 53.432
492394 95.107 -61.866 53.406
492395 95.148 -61.870 53.378
492396 95.189 -61.874 53.346
492397 95.230 -61.879 53.312
492398 95.270 -61.883 53.275
492399 95.309 -61.888 53.236
492400 95.345 -61.895 53.196
492401 95.383 -61.901 53.153
492402 95.420 -61.907 53.111
492403 95.459 -61.912 53.071
492404 95.500 -61.916 53.033
492405 95.541 -61.920 52.995
492406 95.584 -61.923 52.959
492407 95.632 -61.923 52.923
492408 95.680 -61.923 52.887
492409 95.728 -61.923 52.851
492410 95.775 -61.924 52.815
492411 95.827 -61.926 52.777
492412 95.863 -61.929 52.739
492413 95.897 -61.934 52.701
492414 95.933 -61.939 52.664
492415 95.967 -61.944 52.626
492416 96.003 -61.948 52.590
492417 96.037 -61.953 52.552
492418 96.073 -61.958 52.515
492419 96.108 -61.963 52.478
492420 96.143 -61.967 52.443
492421 96.178 -61.973 52.408
492422 96.215 -61.976 52.374
492423 96.258 -61.978 52.337
492424 96.302 -61.979 52.299
492425 96.345 -61.982 52.260
492426 96.389 -61.983 52.219
492427 96.432 -61.985 52.176
492428 96.475 -61.987 52.133
492429 96.519 -61.984 52.088
492430 96.568 -61.978 52.043
492431 96.609 -61.981 51.997
492432 96.651 -61.984 51.952
492433 96.692 -61.988 51.908
492434 96.734 -61.992 51.864
492435 96.775 -61.996 51.821
492436 96.816 -62.001 51.777
492437 96.857 -62.007 51.735
492438 96.897 -62.011 51.691
492439 96.939 -62.017 51.647
492440 96.974 -62.022 51.606
492441 97.005 -62.027 51.566
492442 97.035 -62.033 51.527
492443 97.064 -62.042 51.489
492444 97.092 -62.051 51.454
492445 97.119 -62.061 51.419
492446 97.148 -62.070 51.384
492447 97.176 -62.080 51.350
492448 97.204 -62.089 51.317
492449 97.232 -62.098 51.284
492450 97.260 -62.108 51.250
492451 97.288 -62.117 51.216
492452 97.316 -62.127 51.180
492453 97.344 -62.136 51.143
492454 97.372 -62.146 51.105
492455 97.400 -62.155 51.067
492456 97.450 -62.155 51.028
492457 97.517 -62.146 50.989
492458 97.584 -62.138 50.950
492459 97.651 -62.130 50.910
492460 97.718 -62.121 50.870
492461 97.763 -62.106 50.831
492462 97.784 -62.086 50.791
492463 97.817 -62.074 50.752
492464 97.858 -62.072 50.711
492465 97.911 -62.086 50.670
492466 97.929 -62.101 50.630
492467 97.945 -62.117 50.591
492468 97.969 -62.132 50.550
492469 97.993 -62.148 50.509
492470 98.017 -62.163 50.469
492471 98.041 -62.179 50.427
492472 98.066 -62.195 50.386
492473 98.090 -62.210 50.347
492474 98.114 -62.226 50.308
492475 98.138 -62.241 50.267
492476 98.162 -62.257 50.227
492477 98.186 -62.273 50.186
492933 88.393 -60.139 55.611
492934 88.445 -60.146 55.591
492935 88.497 -60.152 55.571
492936 88.549 -60.159 55.552
492937 88.605 -60.154 55.533
492938 88.664 -60.139 55.516
492939 88.723 -60.123 55.500
492940 88.782 -60.108 55.484
492941 88.839 -60.096 55.467
492942 88.895 -60.085 55.450
492943 88.947 -60.081 55.433
492944 88.987 -60.102 55.417
492945 89.027 -60.121 55.399
492946 89.047 -60.133 55.380
492947 89.088 -60.147 55.359
492948 89.129 -60.165 55.335
492949 89.172 -60.187 55.312
492950 89.215 -60.209 55.289
492951 89.257 -60.230 55.266
492952 89.300 -60.251 55.243
492953 89.343 -60.273 55.220
492954 89.386 -60.294 55.199
492955 89.428 -60.315 55.177
492956 89.471 -60.337 55.156
492957 89.514 -60.359 55.137
492958 89.556 -60.380 55.121
492959 89.599 -60.402 55.105
492960 89.642 -60.423 55.090
492961 89.685 -60.444 55.076
492962 89.727 -60.466 55.063
492963 89.770 -60.488 55.051
492964 89.813 -60.509 55.040
492965 89.856 -60.530 55.030
492966 89.898 -60.552 55.020
492967 89.941 -60.573 55.012
492968 89.983 -60.594 55.005
492969 90.026 -60.616 54.999
492970 90.070 -60.637 54.994
492971 90.115 -60.657 54.992
492972 90.161 -60.678 54.991
492973 90.206 -60.698 54.990
492974 90.252 -60.718 54.987
492975 90.292 -60.737 54.980
492976 90.330 -60.755 54.974
492977 90.369 -60.774 54.971
492978 90.407 -60.792 54.968
492979 90.446 -60.810 54.967
492980 90.485 -60.829 54.969
492981 90.530 -60.845 54.973
492982 90.577 -60.861 54.977
492983 90.625 -60.871 54.981
492984 90.674 -60.882 54.988
492985 90.723 -60.893 54.998
492986 90.772 -60.904 55.011
492987 90.821 -60.916 55.023
492988 90.870 -60.926 55.033
492989 90.908 -60.939 55.040
492990 90.947 -60.951 55.044
492991 90.986 -60.962 55.049
492992 91.024 -60.974 55.057
492993 91.063 -60.986 55.068
492994 91.101 -60.998 55.080
492995 91.140 -61.010 55.091
492996 91.178 -61.022 55.100
492997 91.217 -61.033 55.107
492998 91.255 -61.045 55.109
492999 91.294 -61.058 55.107
493000 91.332 -61.069 55.100
493001 91.371 -61.081 55.092
493002 91.410 -61.093 55.082
493003 91.448 -61.105 55.069
493004 91.487 -61.117 55.053
493005 91.525 -61.129 55.034
493006 91.564 -61.140 55.014
493007 91.602 -61.152 54.996
493008 91.642 -61.164 54.980
493009 91.684 -61.175 54.964
493010 91.726 -61.186 54.944
493011 91.768 -61.197 54.927
493012 91.808 -61.206 54.915
493013 91.848 -61.212 54.912
493014 91.887 -61.218 54.917
493015 91.926 -61.225 54.923
493016 91.965 -61.233 54.925
493017 92.004 -61.242 54.925
493018 92.043 -61.250 54.924
493019 92.082 -61.258 54.922
493020 92.121 -61.266 54.912
493021 92.160 -61.274 54.892
493022 92.199 -61.283 54.864
493023 92.247 -61.295 54.833
493024 92.294 -61.307 54.811
493025 92.342 -61.320 54.810
493026 92.389 -61.332 54.806
493027 92.437 -61.345 54.795
493028 92.485 -61.357 54.775
493029 92.532 -61.370 54.746
493030 92.577 -61.381 54.711
493031 92.616 -61.386 54.677
493032 92.654 -61.392 54.653
493033 92.693 -61.398 54.640
493034 92.731 -61.404 54.627
493035 92.770 -61.410 54.608
493036 92.808 -61.416 54.582
493037 92.846 -61.422 54.555
493038 92.885 -61.428 54.538
493039 92.923 -61.434 54.538
493040 92.961 -61.440 54.539
493041 93.000 -61.446 54.531
493042 93.039 -61.452 54.514
493043 93.077 -61.458 54.489
493044 93.112 -61.465 54.459
493045 93.146 -61.472 54.425
493046 93.182 -61.479 54.395
493047 93.216 -61.486 54.367
493048 93.251 -61.493 54.341
493049 93.286 -61.500 54.315
493050 93.322 -61.507 54.293
493051 93.363 -61.516 54.275
493052 93.404 -61.524 54.256
493053 93.444 -61.532 54.236
493054 93.485 -61.541 54.216
493055 93.526 -61.549 54.192
493056 93.567 -61.557 54.164
493057 93.608 -61.566 54.139
493058 93.648 -61.576 54.115
493059 93.688 -61.586 54.089
493060 93.731 -61.592 54.062
493061 93.773 -61.603 54.035
493062 93.814 -61.615 54.008
493063 93.856 -61.626 53.981
493064 93.897 -61.638 53.955
493065 93.939 -61.649 53.930
493066 93.980 -61.660 53.909
493067 94.022 -61.672 53.887
493068 94.063 -61.683 53.866
493069 94.105 -61.694 53.844
493070 94.146 -61.706 53.823
493071 94.188 -61.716 53.802
493072 94.231 -61.724 53.782
493073 94.273 -61.732 53.761
493074 94.316 -61.738 53.740
493075 94.359 -61.746 53.721
493076 94.402 -61.754 53.701
493077 94.444 -61.761 53.682
493078 94.487 -61.769 53.664
493079 94.530 -61.776 53.647
493080 94.568 -61.780 53.630
493081 94.607 -61.784 53.613
493082 94.645 -61.788 53.594
493083 94.684 -61.792 53.580
493084 94.720 -61.794 53.570
493085 94.756 -61.796 53.562
493086 94.791 -61.798 53.554
493087 94.827 -61.799 53.546
493088 94.862 -61.801 53.538
493089 94.897 -61.802 53.529
493090 94.933 -61.804 53.518
493091 94.968 -61.805 53.504
493092 95.003 -61.807 53.486
493093 95.040 -61.815 53.464
493094 95.076 -61.822 53.439
493095 95.114 -61.828 53.413
493096 95.154 -61.833 53.385
493097 95.194 -61.837 53.354
493098 95.235 -61.841 53.322
493099 95.276 -61.846 53.288
493100 95.317 -61.850 53.250
493101 95.357 -61.854 53.212
493102 95.398 -61.858 53.170
493103 95.439 -61.863 53.129
493104 95.476 -61.869 53.088
493105 95.514 -61.874 53.048
493106 95.555 -61.879 53.008
493107 95.602 -61.879 52.971
493108 95.650 -61.879 52.932
493109 95.697 -61.880 52.894
493110 95.745 -61.880 52.855
493111 95.795 -61.883 52.818
493112 95.832 -61.884 52.778
493113 95.867 -61.889 52.739
493114 95.902 -61.894 52.701
493115 95.937 -61.898 52.663
493116 95.972 -61.904 52.626
493117 96.007 -61.908 52.590
493118 96.042 -61.913 52.552
493119 96.077 -61.918 52.515
493120 96.112 -61.923 52.479
493121 96.147 -61.927 52.446
493122 96.182 -61.932 52.416
493123 96.218 -61.937 52.384
493124 96.261 -61.939 52.348
493125 96.305 -61.940 52.308
493126 96.348 -61.942 52.267
493127 96.392 -61.944 52.225
493128 96.435 -61.945 52.182
493129 96.478 -61.942 52.138
493130 96.522 -61.938 52.094
493131 96.564 -61.939 52.049
493132 96.602 -61.945 52.003
493133 96.643 -61.951 51.958
493134 96.683 -61.956 51.915
493135 96.724 -61.961 51.872
493136 96.765 -61.966 51.830
493137 96.806 -61.971 51.788
493138 96.847 -61.976 51.745
493139 96.887 -61.982 51.703
493140 96.928 -61.987 51.661
493141 96.962 -61.992 51.619
493142 96.993 -61.997 51.578
493143 97.023 -62.002 51.538
493144 97.053 -62.007 51.501
493145 97.084 -62.011 51.464
493146 97.114 -62.018 51.429
493147 97.142 -62.027 51.394
493148 97.170 -62.037 51.361
493149 97.198 -62.046 51.327
493150 97.226 -62.056 51.293
493151 97.254 -62.065 51.259
493152 97.282 -62.075 51.224
493153 97.310 -62.084 51.187
493154 97.338 -62.094 51.149
493155 97.366 -62.103 51.111
493156 97.394 -62.113 51.073
493157 97.451 -62.109 51.034
493158 97.518 -62.101 50.994
493159 97.585 -62.092 50.954
493160 97.652 -62.083 50.914
493161 97.719 -62.075 50.873
493162 97.770 -62.062 50.834
493163 97.792 -62.042 50.794
493164 97.825 -62.030 50.753
493165 97.872 -62.036 50.713
493166 97.911 -62.051 50.672
493167 97.923 -62.065 50.631
493168 97.947 -62.081 50.591
493169 97.971 -62.097 50.551
493170 97.995 -62.112 50.510
493171 98.019 -62.128 50.469
493172 98.043 -62.143 50.428
493173 98.068 -62.159 50.387
493174 98.092 -62.175 50.347
493175 98.116 -62.190 50.307
493176 98.140 -62.206 50.268
493177 98.164 -62.221 50.228
493178 98.189 -62.237 50.188
493635 88.472 -60.102 55.585
493636 88.524 -60.108 55.565
493637 88.575 -60.115 55.544
493638 88.633 -60.106 55.525
493639 88.692 -60.090 55.507
493640 88.751 -60.074 55.491
493641 88.810 -60.061 55.474
493642 88.866 -60.049 55.457
493643 88.922 -60.038 55.440
493644 88.967 -60.049 55.423
493645 89.006 -60.069 55.407
493646 89.047 -60.090 55.389
493647 89.064 -60.100 55.371
493648 89.103 -60.115 55.351
493649 89.143 -60.129 55.328
493650 89.186 -60.150 55.305
493651 89.229 -60.171 55.283
493652 89.271 -60.193 55.261
493653 89.314 -60.215 55.240
493654 89.356 -60.236 55.218
493655 89.399 -60.258 55.196
493656 89.442 -60.279 55.175
493657 89.485 -60.300 55.156
493658 89.527 -60.322 55.138
493659 89.570 -60.343 55.121
493660 89.613 -60.365 55.106
493661 89.656 -60.387 55.091
493662 89.698 -60.408 55.077
493663 89.741 -60.429 55.065
493664 89.783 -60.450 55.053
493665 89.826 -60.472 55.042
493666 89.869 -60.494 55.032
493667 89.912 -60.515 55.023
493668 89.954 -60.537 55.015
493669 89.997 -60.558 55.007
493670 90.040 -60.579 55.000
493671 90.085 -60.600 54.996
493672 90.130 -60.620 54.994
493673 90.176 -60.640 54.995
493674 90.222 -60.660 54.997
493675 90.267 -60.680 54.994
493676 90.310 -60.700 54.986
493677 90.348 -60.719 54.978
493678 90.387 -60.737 54.972
493679 90.425 -60.755 54.968
493680 90.463 -60.774 54.968
493681 90.504 -60.792 54.969
493682 90.549 -60.808 54.971
493683 90.595 -60.824 54.976
493684 90.643 -60.836 54.981
493685 90.692 -60.848 54.987
493686 90.741 -60.858 54.995
493687 90.790 -60.870 55.006
493688 90.838 -60.881 55.020
493689 90.883 -60.892 55.033
493690 90.922 -60.904 55.040
493691 90.960 -60.916 55.044
493692 90.999 -60.928 55.048
493693 91.037 -60.939 55.054
493694 91.076 -60.951 55.062
493695 91.114 -60.964 55.072
493696 91.153 -60.975 55.084
493697 91.191 -60.987 55.094
493698 91.230 -60.999 55.102
493699 91.269 -61.011 55.106
493700 91.307 -61.023 55.105
493701 91.346 -61.035 55.100
493702 91.384 -61.046 55.093
493703 91.423 -61.058 55.085
493704 91.461 -61.071 55.075
493705 91.500 -61.082 55.061
493706 91.538 -61.094 55.045
493707 91.577 -61.106 55.025
493708 91.615 -61.118 55.005
493709 91.654 -61.130 54.984
493710 91.693 -61.141 54.963
493711 91.736 -61.152 54.942
493712 91.777 -61.163 54.922
493713 91.817 -61.171 54.908
493714 91.857 -61.177 54.903
493715 91.895 -61.185 54.910
493716 91.935 -61.193 54.917
493717 91.974 -61.201 54.921
493718 92.013 -61.209 54.920
493719 92.052 -61.218 54.917
493720 92.090 -61.225 54.914
493721 92.130 -61.233 54.904
493722 92.168 -61.242 54.885
493723 92.208 -61.250 54.860
493724 92.250 -61.260 54.832
493725 92.298 -61.273 54.816
493726 92.346 -61.285 54.815
493727 92.393 -61.297 54.809
493728 92.441 -61.310 54.796
493729 92.489 -61.322 54.777
493730 92.536 -61.335 54.750
493731 92.580 -61.345 54.717
493732 92.619 -61.351 54.680
493733 92.657 -61.356 54.649
493734 92.696 -61.362 54.628
493735 92.734 -61.368 54.610
493736 92.772 -61.375 54.590
493737 92.811 -61.381 54.564
493738 92.849 -61.387 54.542
493739 92.887 -61.393 54.533
493740 92.926 -61.399 54.539
493741 92.965 -61.404 54.536
493742 93.003 -61.410 54.525
493743 93.041 -61.416 54.504
493744 93.080 -61.422 54.477
493745 93.118 -61.428 54.449
493746 93.156 -61.434 54.423
493747 93.194 -61.440 54.397
493748 93.229 -61.447 54.373
493749 93.264 -61.454 54.352
493750 93.298 -61.462 54.336
493751 93.334 -61.469 54.320
493752 93.372 -61.476 54.305
493753 93.413 -61.484 54.287
493754 93.453 -61.493 54.265
493755 93.494 -61.501 54.240
493756 93.535 -61.510 54.211
493757 93.575 -61.519 54.182
493758 93.616 -61.530 54.154
493759 93.655 -61.541 54.128
493760 93.696 -61.551 54.103
493761 93.738 -61.557 54.079
493762 93.781 -61.564 54.056
493763 93.823 -61.572 54.031
493764 93.864 -61.584 54.004
493765 93.906 -61.595 53.976
493766 93.947 -61.606 53.948
493767 93.989 -61.618 53.925
493768 94.031 -61.629 53.901
493769 94.072 -61.641 53.877
493770 94.114 -61.652 53.854
493771 94.155 -61.663 53.833
493772 94.197 -61.674 53.811
493773 94.240 -61.682 53.790
493774 94.282 -61.690 53.770
493775 94.325 -61.697 53.750
493776 94.367 -61.705 53.732
493777 94.410 -61.713 53.713
493778 94.453 -61.720 53.695
493779 94.495 -61.727 53.676
493780 94.536 -61.732 53.658
493781 94.574 -61.737 53.640
493782 94.613 -61.741 53.620
493783 94.652 -61.744 53.601
493784 94.690 -61.748 53.584
493785 94.726 -61.751 53.572
493786 94.762 -61.752 53.561
493787 94.797 -61.754 53.552
493788 94.833 -61.755 53.544
493789 94.868 -61.757 53.534
493790 94.903 -61.759 53.524
493791 94.938 -61.760 53.513
493792 94.974 -61.763 53.499
493793 95.011 -61.769 53.481
493794 95.046 -61.777 53.459
493795 95.083 -61.784 53.434
493796 95.121 -61.789 53.408
493797 95.159 -61.795 53.380
493798 95.200 -61.800 53.351
493799 95.241 -61.804 53.320
493800 95.281 -61.809 53.287
493801 95.322 -61.813 53.252
493802 95.363 -61.817 53.215
493803 95.403 -61.821 53.177
493804 95.444 -61.826 53.137
493805 95.485 -61.829 53.098
493806 95.526 -61.834 53.061
493807 95.565 -61.838 53.023
493808 95.613 -61.838 52.984
493809 95.660 -61.839 52.943
493810 95.707 -61.839 52.901
493811 95.754 -61.840 52.861
493812 95.796 -61.842 52.822
493813 95.834 -61.845 52.781
493814 95.872 -61.849 52.741
493815 95.907 -61.854 52.703
493816 95.942 -61.858 52.665
493817 95.977 -61.863 52.627
493818 96.012 -61.868 52.590
493819 96.047 -61.873 52.553
493820 96.082 -61.878 52.516
493821 96.117 -61.882 52.482
493822 96.152 -61.887 52.452
493823 96.187 -61.892 52.425
493824 96.222 -61.897 52.395
493825 96.264 -61.899 52.359
493826 96.308 -61.901 52.318
493827 96.352 -61.902 52.275
493828 96.395 -61.903 52.232
493829 96.438 -61.899 52.189
493830 96.480 -61.899 52.145
493831 96.521 -61.903 52.100
493832 96.562 -61.907 52.054
493833 96.599 -61.913 52.008
493834 96.636 -61.920 51.963
493835 96.673 -61.926 51.921
493836 96.714 -61.931 51.879
493837 96.755 -61.936 51.838
493838 96.796 -61.942 51.797
493839 96.836 -61.947 51.756
493840 96.877 -61.952 51.715
493841 96.918 -61.957 51.674
493842 96.949 -61.962 51.632
493843 96.980 -61.967 51.591
493844 97.011 -61.972 51.551
493845 97.041 -61.977 51.513
493846 97.072 -61.982 51.476
493847 97.102 -61.986 51.442
493848 97.133 -61.992 51.408
493849 97.163 -61.996 51.375
493850 97.192 -62.004 51.341
493851 97.221 -62.013 51.306
493852 97.249 -62.023 51.270
493853 97.277 -62.032 51.234
493854 97.305 -62.042 51.196
493855 97.333 -62.051 51.158
493856 97.361 -62.061 51.119
493857 97.389 -62.070 51.081
493858 97.452 -62.063 51.041
493859 97.519 -62.055 51.000
493860 97.586 -62.046 50.960
493861 97.653 -62.038 50.919
493862 97.720 -62.029 50.878
493863 97.778 -62.017 50.838
493864 97.800 -61.998 50.798
493865 97.832 -61.986 50.756
493866 97.885 -62.000 50.715
493867 97.903 -62.014 50.674
493868 97.924 -62.030 50.632
493869 97.948 -62.045 50.592
493870 97.972 -62.061 50.551
493871 97.997 -62.077 50.511
493872 98.021 -62.092 50.470
493873 98.045 -62.108 50.428
493874 98.069 -62.124 50.387
493875 98.093 -62.139 50.347
493876 98.118 -62.155 50.308
493877 98.142 -62.171 50.268
493878 98.166 -62.186 50.228
494336 88.498 -60.058 55.583
494337 88.550 -60.065 55.561
494338 88.603 -60.071 55.540
494339 88.661 -60.058 55.520
494340 88.720 -60.042 55.502
494341 88.779 -60.027 55.485
494342 88.837 -60.014 55.467
494343 88.893 -60.002 55.450
494344 88.946 -59.997 55.432
494345 88.986 -60.017 55.415
494346 89.026 -60.037 55.398
494347 89.066 -60.057 55.381
494348 89.086 -60.069 55.362
494349 89.118 -60.082 55.343
494350 89.158 -60.096 55.322
494351 89.199 -60.114 55.300
494352 89.242 -60.135 55.278
494353 89.285 -60.156 55.257
494354 89.327 -60.178 55.237
494355 89.370 -60.199 55.217
494356 89.413 -60.221 55.195
494357 89.456 -60.243 55.175
494358 89.498 -60.264 55.156
494359 89.541 -60.285 55.140
494360 89.584 -60.306 55.124
494361 89.626 -60.328 55.109
494362 89.669 -60.350 55.094
494363 89.712 -60.371 55.081
494364 89.754 -60.393 55.069
494365 89.797 -60.414 55.057
494366 89.840 -60.435 55.047
494367 89.883 -60.457 55.037
494368 89.925 -60.478 55.027
494369 89.968 -60.500 55.019
494370 90.011 -60.522 55.011
494371 90.054 -60.542 55.004
494372 90.100 -60.563 54.999
494373 90.146 -60.582 54.998
494374 90.191 -60.603 55.001
494375 90.237 -60.622 55.006
494376 90.282 -60.643 55.001
494377 90.327 -60.663 54.992
494378 90.366 -60.682 54.982
494379 90.404 -60.700 54.974
494380 90.443 -60.719 54.968
494381 90.481 -60.737 54.968
494382 90.523 -60.754 54.969
494383 90.569 -60.771 54.970
494384 90.614 -60.787 54.975
494385 90.661 -60.801 54.981
494386 90.710 -60.813 54.986
494387 90.758 -60.823 54.993
494388 90.807 -60.835 55.002
494389 90.856 -60.845 55.015
494390 90.897 -60.857 55.030
494391 90.935 -60.870 55.040
494392 90.974 -60.881 55.047
494393 91.012 -60.893 55.051
494394 91.051 -60.905 55.053
494395 91.089 -60.917 55.056
494396 91.128 -60.929 55.063
494397 91.166 -60.941 55.073
494398 91.205 -60.952 55.084
494399 91.243 -60.964 55.093
494400 91.282 -60.977 55.098
494401 91.320 -60.988 55.098
494402 91.359 -61.000 55.096
494403 91.397 -61.012 55.091
494404 91.436 -61.024 55.085
494405 91.474 -61.036 55.079
494406 91.513 -61.048 55.068
494407 91.552 -61.059 55.054
494408 91.590 -61.071 55.036
494409 91.629 -61.083 55.014
494410 91.667 -61.095 54.990
494411 91.706 -61.107 54.964
494412 91.745 -61.118 54.940
494413 91.787 -61.130 54.918
494414 91.826 -61.136 54.900
494415 91.865 -61.144 54.892
494416 91.905 -61.152 54.897
494417 91.943 -61.161 54.904
494418 91.982 -61.168 54.909
494419 92.021 -61.177 54.909
494420 92.060 -61.185 54.906
494421 92.099 -61.193 54.902
494422 92.138 -61.201 54.891
494423 92.177 -61.209 54.873
494424 92.216 -61.218 54.851
494425 92.255 -61.226 54.828
494426 92.302 -61.238 54.820
494427 92.350 -61.250 54.819
494428 92.398 -61.262 54.812
494429 92.445 -61.275 54.798
494430 92.493 -61.287 54.774
494431 92.540 -61.300 54.745
494432 92.583 -61.309 54.716
494433 92.622 -61.315 54.686
494434 92.660 -61.321 54.654
494435 92.699 -61.327 54.619
494436 92.737 -61.333 54.592
494437 92.775 -61.339 54.569
494438 92.813 -61.345 54.546
494439 92.852 -61.351 54.529
494440 92.890 -61.357 54.525
494441 92.929 -61.363 54.532
494442 92.968 -61.368 54.527
494443 93.006 -61.375 54.513
494444 93.044 -61.381 54.492
494445 93.082 -61.387 54.466
494446 93.121 -61.393 54.443
494447 93.159 -61.399 54.422
494448 93.198 -61.405 54.399
494449 93.236 -61.411 54.379
494450 93.275 -61.416 54.365
494451 93.311 -61.423 54.356
494452 93.346 -61.430 54.342
494453 93.381 -61.437 54.325
494454 93.422 -61.446 54.304
494455 93.463 -61.454 54.278
494456 93.503 -61.463 54.249
494457 93.543 -61.473 54.219
494458 93.583 -61.484 54.190
494459 93.623 -61.494 54.162
494460 93.663 -61.505 54.136
494461 93.703 -61.515 54.111
494462 93.746 -61.522 54.087
494463 93.788 -61.528 54.065
494464 93.831 -61.535 54.043
494465 93.873 -61.541 54.019
494466 93.915 -61.553 53.995
494467 93.957 -61.564 53.966
494468 93.998 -61.575 53.939
494469 94.040 -61.587 53.913
494470 94.081 -61.598 53.887
494471 94.123 -61.610 53.862
494472 94.164 -61.621 53.840
494473 94.206 -61.632 53.820
494474 94.248 -61.641 53.801
494475 94.290 -61.649 53.783
494476 94.333 -61.656 53.765
494477 94.376 -61.663 53.747
494478 94.419 -61.671 53.730
494479 94.461 -61.678 53.711
494480 94.503 -61.685 53.691
494481 94.542 -61.689 53.671
494482 94.580 -61.693 53.651
494483 94.619 -61.697 53.629
494484 94.658 -61.701 53.608
494485 94.697 -61.705 53.589
494486 94.733 -61.707 53.573
494487 94.768 -61.709 53.560
494488 94.803 -61.710 53.548
494489 94.838 -61.712 53.538
494490 94.874 -61.713 53.527
494491 94.909 -61.715 53.515
494492 94.945 -61.717 53.503
494493 94.981 -61.725 53.488
494494 95.017 -61.732 53.471
494495 95.053 -61.739 53.449
494496 95.090 -61.746 53.425
494497 95.128 -61.751 53.399
494498 95.166 -61.757 53.373
494499 95.204 -61.763 53.344
494500 95.246 -61.767 53.314
494501 95.287 -61.771 53.282
494502 95.328 -61.776 53.249
494503 95.368 -61.780 53.214
494504 95.409 -61.784 53.178
494505 95.450 -61.788 53.141
494506 95.490 -61.792 53.105
494507 95.529 -61.797 53.070
494508 95.567 -61.801 53.032
494509 95.612 -61.803 52.993
494510 95.659 -61.803 52.951
494511 95.706 -61.804 52.909
494512 95.748 -61.806 52.868
494513 95.788 -61.809 52.829
494514 95.827 -61.812 52.786
494515 95.865 -61.815 52.745
494516 95.903 -61.818 52.706
494517 95.941 -61.821 52.667
494518 95.979 -61.824 52.628
494519 96.017 -61.828 52.592
494520 96.051 -61.832 52.554
494521 96.087 -61.838 52.518
494522 96.121 -61.842 52.485
494523 96.157 -61.847 52.456
494524 96.191 -61.852 52.430
494525 96.227 -61.857 52.400
494526 96.267 -61.859 52.364
494527 96.311 -61.861 52.323
494528 96.355 -61.860 52.280
494529 96.397 -61.859 52.238
494530 96.438 -61.863 52.195
494531 96.479 -61.867 52.151
494532 96.520 -61.871 52.105
494533 96.560 -61.875 52.058
494534 96.598 -61.880 52.010
494535 96.635 -61.887 51.965
494536 96.672 -61.893 51.924
494537 96.708 -61.900 51.883
494538 96.745 -61.907 51.843
494539 96.786 -61.912 51.805
494540 96.826 -61.917 51.767
494541 96.867 -61.923 51.728
494542 96.906 -61.927 51.687
494543 96.937 -61.932 51.646
494544 96.967 -61.937 51.604
494545 96.998 -61.942 51.564
494546 97.028 -61.947 51.526
494547 97.059 -61.952 51.490
494548 97.090 -61.957 51.457
494549 97.120 -61.962 51.426
494550 97.151 -61.967 51.391
494551 97.181 -61.971 51.356
494552 97.212 -61.976 51.321
494553 97.242 -61.981 51.283
494554 97.272 -61.989 51.245
494555 97.299 -61.998 51.206
494556 97.327 -62.008 51.167
494557 97.356 -62.017 51.128
494558 97.387 -62.026 51.089
494559 97.454 -62.017 51.049
494560 97.521 -62.008 51.008
494561 97.588 -62.000 50.967
494562 97.655 -61.992 50.926
494563 97.722 -61.983 50.884
494564 97.785 -61.973 50.843
494565 97.807 -61.954 50.801
494566 97.846 -61.949 50.759
494567 97.885 -61.964 50.718
494568 97.902 -61.979 50.676
494569 97.926 -61.995 50.633
494570 97.950 -62.011 50.593
494571 97.974 -62.026 50.552
494572 97.999 -62.042 50.511
494573 98.023 -62.058 50.470
494574 98.047 -62.073 50.429
494575 98.071 -62.089 50.387
494576 98.095 -62.105 50.347
494577 98.120 -62.120 50.308
494578 98.144 -62.136 50.268
494579 98.168 -62.151 50.228
495038 88.577 -60.021 55.559
495039 88.630 -60.025 55.538
495040 88.689 -60.009 55.518
495041 88.748 -59.994 55.499
495042 88.808 -59.978 55.481
495043 88.863 -59.967 55.462
495044 88.920 -59.955 55.444
495045 88.966 -59.965 55.426
495046 89.005 -59.985 55.409
495047 89.046 -60.005 55.392
495048 89.086 -60.025 55.374
495049 89.107 -60.037 55.355
495050 89.132 -60.049 55.337
495051 89.173 -60.064 55.316
495052 89.213 -60.079 55.295
495053 89.256 -60.099 55.274
495054 89.299 -60.120 55.254
495055 89.341 -60.141 55.234
495056 89.384 -60.163 55.215
495057 89.426 -60.184 55.195
495058 89.469 -60.206 55.175
495059 89.512 -60.227 55.157
495060 89.554 -60.249 55.142
495061 89.597 -60.270 55.127
495062 89.640 -60.291 55.112
495063 89.683 -60.313 55.098
495064 89.726 -60.334 55.085
495065 89.768 -60.356 55.074
495066 89.811 -60.378 55.063
495067 89.853 -60.399 55.053
495068 89.896 -60.420 55.043
495069 89.939 -60.442 55.034
495070 89.982 -60.463 55.026
495071 90.024 -60.485 55.018
495072 90.070 -60.505 55.010
495073 90.115 -60.525 55.004
495074 90.161 -60.545 55.001
495075 90.206 -60.566 55.000
495076 90.252 -60.585 55.004
495077 90.297 -60.606 55.001
495078 90.342 -60.626 54.993
495079 90.384 -60.645 54.983
495080 90.423 -60.663 54.975
495081 90.461 -60.682 54.970
495082 90.499 -60.701 54.968
495083 90.542 -60.717 54.968
495084 90.588 -60.734 54.970
495085 90.633 -60.750 54.974
495086 90.679 -60.766 54.982
495087 90.727 -60.778 54.987
495088 90.776 -60.788 54.992
495089 90.825 -60.800 55.000
495090 90.872 -60.811 55.012
495091 90.910 -60.823 55.027
495092 90.949 -60.835 55.041
495093 90.987 -60.847 55.050
495094 91.026 -60.858 55.055
495095 91.064 -60.870 55.058
495096 91.103 -60.882 55.057
495097 91.141 -60.894 55.061
495098 91.180 -60.906 55.066
495099 91.218 -60.918 55.072
495100 91.257 -60.930 55.077
495101 91.295 -60.942 55.083
495102 91.334 -60.954 55.084
495103 91.372 -60.965 55.084
495104 91.411 -60.977 55.082
495105 91.449 -60.989 55.081
495106 91.488 -61.001 55.079
495107 91.526 -61.013 55.072
495108 91.565 -61.025 55.060
495109 91.603 -61.036 55.043
495110 91.642 -61.049 55.023
495111 91.680 -61.061 54.998
495112 91.719 -61.072 54.967
495113 91.757 -61.084 54.938
495114 91.796 -61.096 54.914
495115 91.835 -61.104 54.894
495116 91.874 -61.112 54.881
495117 91.913 -61.120 54.880
495118 91.952 -61.128 54.885
495119 91.991 -61.136 54.891
495120 92.030 -61.144 54.893
495121 92.069 -61.152 54.891
495122 92.108 -61.161 54.887
495123 92.147 -61.169 54.877
495124 92.186 -61.177 54.861
495125 92.225 -61.185 54.843
495126 92.264 -61.193 54.825
495127 92.306 -61.203 54.819
495128 92.354 -61.215 54.815
495129 92.401 -61.227 54.808
495130 92.449 -61.240 54.794
495131 92.496 -61.252 54.772
495132 92.544 -61.265 54.753
495133 92.586 -61.274 54.730
495134 92.625 -61.280 54.704
495135 92.663 -61.286 54.672
495136 92.701 -61.292 54.632
495137 92.740 -61.297 54.595
495138 92.778 -61.303 54.569
495139 92.816 -61.309 54.543
495140 92.855 -61.315 54.521
495141 92.894 -61.321 54.511
495142 92.932 -61.328 54.517
495143 92.970 -61.334 54.513
495144 93.009 -61.339 54.500
495145 93.047 -61.345 54.481
495146 93.085 -61.351 54.459
495147 93.124 -61.357 54.442
495148 93.162 -61.363 54.421
495149 93.201 -61.369 54.398
495150 93.239 -61.375 54.384
495151 93.278 -61.381 54.378
495152 93.316 -61.387 54.372
495153 93.354 -61.393 54.357
495154 93.393 -61.399 54.338
495155 93.431 -61.406 54.313
495156 93.471 -61.417 54.283
495157 93.511 -61.428 54.251
495158 93.551 -61.437 54.222
495159 93.591 -61.448 54.195
495160 93.631 -61.459 54.170
495161 93.671 -61.469 54.145
495162 93.711 -61.479 54.121
495163 93.753 -61.487 54.099
495164 93.795 -61.494 54.076
495165 93.838 -61.500 54.053
495166 93.880 -61.506 54.029
495167 93.923 -61.513 54.005
495168 93.966 -61.522 53.978
495169 94.007 -61.533 53.952
495170 94.049 -61.544 53.924
495171 94.090 -61.556 53.898
495172 94.132 -61.567 53.873
495173 94.173 -61.578 53.853
495174 94.215 -61.589 53.835
495175 94.257 -61.600 53.818
495176 94.299 -61.607 53.802
495177 94.342 -61.615 53.785
495178 94.385 -61.622 53.767
495179 94.427 -61.630 53.749
495180 94.470 -61.638 53.729
495181 94.509 -61.641 53.708
495182 94.548 -61.646 53.686
495183 94.586 -61.650 53.665
495184 94.625 -61.653 53.640
495185 94.664 -61.657 53.617
495186 94.703 -61.662 53.595
495187 94.738 -61.663 53.575
495188 94.774 -61.665 53.556
495189 94.809 -61.666 53.542
495190 94.844 -61.668 53.529
495191 94.880 -61.670 53.515
495192 94.915 -61.672 53.502
495193 94.952 -61.679 53.488
495194 94.988 -61.687 53.473
495195 95.024 -61.694 53.456
495196 95.060 -61.701 53.436
495197 95.097 -61.708 53.413
495198 95.135 -61.713 53.388
495199 95.173 -61.719 53.362
495200 95.210 -61.724 53.334
495201 95.251 -61.729 53.305
495202 95.293 -61.734 53.275
495203 95.333 -61.738 53.243
495204 95.374 -61.743 53.210
495205 95.414 -61.747 53.177
495206 95.455 -61.751 53.144
495207 95.494 -61.755 53.110
495208 95.532 -61.760 53.075
495209 95.569 -61.763 53.038
495210 95.611 -61.766 53.000
495211 95.658 -61.767 52.959
495212 95.701 -61.769 52.917
495213 95.740 -61.773 52.876
495214 95.779 -61.776 52.836
495215 95.819 -61.779 52.792
495216 95.858 -61.782 52.750
495217 95.896 -61.785 52.709
495218 95.934 -61.788 52.670
495219 95.972 -61.791 52.631
495220 96.010 -61.794 52.594
495221 96.048 -61.798 52.556
495222 96.086 -61.801 52.519
495223 96.124 -61.804 52.486
495224 96.161 -61.807 52.456
495225 96.196 -61.812 52.428
495226 96.231 -61.816 52.398
495227 96.270 -61.820 52.363
495228 96.314 -61.820 52.322
495229 96.355 -61.823 52.280
495230 96.395 -61.827 52.238
495231 96.436 -61.831 52.194
495232 96.477 -61.835 52.149
495233 96.518 -61.838 52.102
495234 96.558 -61.842 52.054
495235 96.597 -61.848 52.006
495236 96.633 -61.854 51.962
495237 96.670 -61.861 51.922
495238 96.707 -61.867 51.883
495239 96.744 -61.874 51.845
495240 96.781 -61.880 51.808
495241 96.817 -61.887 51.773
495242 96.857 -61.892 51.735
495243 96.894 -61.898 51.696
495244 96.924 -61.902 51.656
495245 96.955 -61.907 51.616
495246 96.985 -61.912 51.576
495247 97.016 -61.917 51.538
495248 97.046 -61.922 51.502
495249 97.077 -61.927 51.469
495250 97.107 -61.932 51.439
495251 97.138 -61.937 51.405
495252 97.169 -61.942 51.370
495253 97.199 -61.946 51.334
495254 97.230 -61.951 51.296
495255 97.260 -61.956 51.258
495256 97.291 -61.961 51.218
495257 97.321 -61.966 51.178
495258 97.350 -61.975 51.138
495259 97.388 -61.979 51.098
495260 97.455 -61.971 51.057
495261 97.522 -61.963 51.016
495262 97.589 -61.954 50.974
495263 97.656 -61.945 50.931
495264 97.723 -61.937 50.889
495265 97.790 -61.929 50.848
495266 97.815 -61.910 50.805
495267 97.860 -61.914 50.762
495268 97.880 -61.929 50.719
495269 97.904 -61.944 50.677
495270 97.928 -61.960 50.634
495271 97.952 -61.975 50.594
495272 97.977 -61.991 50.552
495273 98.001 -62.007 50.512
495274 98.025 -62.022 50.471
495275 98.049 -62.038 50.430
495276 98.073 -62.053 50.388
495277 98.097 -62.069 50.347
495278 98.122 -62.085 50.308
495279 98.146 -62.100 50.268
495280 98.170 -62.116 50.228
495740 88.658 -59.977 55.537
495741 88.717 -59.961 55.516
495742 88.776 -59.946 55.497
495743 88.835 -59.932 55.478
495744 88.891 -59.920 55.459
495745 88.945 -59.913 55.441
495746 88.985 -59.933 55.422
495747 89.025 -59.953 55.404
495748 89.065 -59.973 55.386
495749 89.105 -59.993 55.368
495750 89.128 -60.006 55.350
495751 89.147 -60.016 55.332
495752 89.188 -60.031 55.312
495753 89.229 -60.046 55.292
495754 89.269 -60.061 55.272
495755 89.312 -60.083 55.253
495756 89.354 -60.105 55.233
495757 89.397 -60.126 55.214
495758 89.440 -60.147 55.195
495759 89.483 -60.169 55.177
495760 89.526 -60.190 55.160
495761 89.568 -60.212 55.145
495762 89.611 -60.234 55.129
495763 89.654 -60.255 55.115
495764 89.696 -60.276 55.102
495765 89.739 -60.298 55.091
495766 89.782 -60.319 55.080
495767 89.824 -60.340 55.070
495768 89.867 -60.362 55.060
495769 89.910 -60.384 55.050
495770 89.953 -60.405 55.042
495771 89.996 -60.426 55.034
495772 90.039 -60.447 55.026
495773 90.085 -60.468 55.018
495774 90.130 -60.488 55.011
495775 90.176 -60.508 55.005
495776 90.222 -60.528 55.001
495777 90.267 -60.548 55.001
495778 90.312 -60.568 54.997
495779 90.358 -60.588 54.991
495780 90.402 -60.608 54.983
495781 90.440 -60.626 54.977
495782 90.479 -60.645 54.973
495783 90.517 -60.663 54.971
495784 90.561 -60.680 54.971
495785 90.607 -60.697 54.971
495786 90.652 -60.713 54.973
495787 90.698 -60.729 54.979
495788 90.745 -60.743 54.986
495789 90.794 -60.754 54.993
495790 90.842 -60.765 55.001
495791 90.885 -60.776 55.010
495792 90.924 -60.788 55.027
495793 90.962 -60.800 55.043
495794 91.001 -60.812 55.054
495795 91.039 -60.824 55.062
495796 91.078 -60.836 55.065
495797 91.116 -60.848 55.063
495798 91.155 -60.860 55.063
495799 91.193 -60.871 55.064
495800 91.232 -60.883 55.065
495801 91.270 -60.895 55.065
495802 91.309 -60.907 55.068
495803 91.347 -60.919 55.068
495804 91.386 -60.931 55.069
495805 91.424 -60.942 55.072
495806 91.463 -60.955 55.076
495807 91.501 -60.967 55.077
495808 91.540 -60.978 55.072
495809 91.579 -60.989 55.060
495810 91.618 -60.998 55.044
495811 91.658 -61.008 55.023
495812 91.698 -61.017 55.000
495813 91.737 -61.027 54.971
495814 91.777 -61.036 54.944
495815 91.803 -61.049 54.918
495816 91.843 -61.064 54.893
495817 91.883 -61.080 54.872
495818 91.922 -61.087 54.863
495819 91.961 -61.096 54.862
495820 92.000 -61.104 54.865
495821 92.038 -61.112 54.869
495822 92.078 -61.120 54.870
495823 92.117 -61.128 54.869
495824 92.156 -61.136 54.862
495825 92.195 -61.144 54.849
495826 92.233 -61.152 54.836
495827 92.273 -61.161 54.825
495828 92.311 -61.169 54.816
495829 92.358 -61.180 54.810
495830 92.405 -61.193 54.802
495831 92.453 -61.205 54.789
495832 92.500 -61.218 54.771
495833 92.548 -61.230 54.760
495834 92.589 -61.238 54.743
495835 92.627 -61.244 54.719
495836 92.666 -61.250 54.686
495837 92.704 -61.256 54.645
495838 92.742 -61.262 54.601
495839 92.781 -61.268 54.569
495840 92.819 -61.274 54.539
495841 92.858 -61.280 54.515
495842 92.896 -61.286 54.502
495843 92.935 -61.292 54.505
495844 92.973 -61.298 54.501
495845 93.011 -61.304 54.490
495846 93.050 -61.309 54.473
495847 93.088 -61.315 54.454
495848 93.127 -61.321 54.434
495849 93.165 -61.328 54.411
495850 93.204 -61.334 54.395
495851 93.242 -61.340 54.386
495852 93.280 -61.346 54.380
495853 93.319 -61.351 54.371
495854 93.359 -61.357 54.358
495855 93.399 -61.362 54.339
495856 93.439 -61.370 54.314
495857 93.479 -61.381 54.283
495858 93.518 -61.392 54.249
495859 93.559 -61.402 54.223
495860 93.599 -61.412 54.200
495861 93.639 -61.423 54.178
495862 93.679 -61.433 54.156
495863 93.719 -61.443 54.136
495864 93.760 -61.453 54.114
495865 93.802 -61.459 54.090
495866 93.845 -61.465 54.064
495867 93.888 -61.472 54.037
495868 93.930 -61.478 54.012
495869 93.973 -61.484 53.988
495870 94.016 -61.491 53.963
495871 94.058 -61.502 53.936
495872 94.099 -61.513 53.910
495873 94.140 -61.525 53.887
495874 94.182 -61.536 53.869
495875 94.223 -61.547 53.853
495876 94.265 -61.558 53.838
495877 94.308 -61.566 53.822
495878 94.350 -61.574 53.806
495879 94.393 -61.581 53.788
495880 94.436 -61.588 53.768
495881 94.477 -61.594 53.747
495882 94.516 -61.598 53.725
495883 94.554 -61.602 53.702
495884 94.593 -61.606 53.679
495885 94.631 -61.610 53.653
495886 94.670 -61.614 53.627
495887 94.709 -61.618 53.603
495888 94.744 -61.619 53.579
495889 94.780 -61.622 53.556
495890 94.815 -61.623 53.538
495891 94.850 -61.625 53.521
495892 94.886 -61.627 53.504
495893 94.922 -61.635 53.487
495894 94.958 -61.641 53.471
495895 94.994 -61.649 53.454
495896 95.031 -61.656 53.436
495897 95.066 -61.663 53.416
495898 95.104 -61.669 53.393
495899 95.141 -61.675 53.370
495900 95.179 -61.681 53.345
495901 95.217 -61.686 53.320
495902 95.255 -61.691 53.294
495903 95.297 -61.697 53.266
495904 95.338 -61.701 53.238
495905 95.379 -61.706 53.208
495906 95.420 -61.710 53.177
495907 95.458 -61.714 53.145
495908 95.496 -61.718 53.112
495909 95.534 -61.722 53.077
495910 95.572 -61.726 53.042
495911 95.610 -61.731 53.003
495912 95.653 -61.732 52.964
495913 95.692 -61.735 52.924
495914 95.732 -61.738 52.884
495915 95.771 -61.742 52.844
495916 95.811 -61.745 52.800
495917 95.850 -61.748 52.756
495918 95.889 -61.751 52.714
495919 95.926 -61.754 52.675
495920 95.965 -61.757 52.635
495921 96.003 -61.760 52.596
495922 96.041 -61.763 52.557
495923 96.079 -61.767 52.520
495924 96.117 -61.770 52.485
495925 96.155 -61.773 52.453
495926 96.194 -61.776 52.422
495927 96.235 -61.779 52.392
495928 96.271 -61.783 52.357
495929 96.312 -61.787 52.317
495930 96.353 -61.791 52.274
495931 96.394 -61.794 52.231
495932 96.435 -61.798 52.187
495933 96.475 -61.803 52.141
495934 96.516 -61.807 52.094
495935 96.557 -61.810 52.045
495936 96.596 -61.816 51.999
495937 96.632 -61.822 51.956
495938 96.669 -61.829 51.917
495939 96.706 -61.835 51.880
495940 96.742 -61.841 51.844
495941 96.779 -61.848 51.809
495942 96.816 -61.854 51.775
495943 96.853 -61.861 51.740
495944 96.891 -61.861 51.703
495945 96.922 -61.867 51.664
495946 96.953 -61.871 51.625
495947 96.983 -61.876 51.586
495948 97.014 -61.881 51.549
495949 97.045 -61.885 51.513
495950 97.076 -61.891 51.479
495951 97.106 -61.895 51.449
495952 97.137 -61.900 51.414
495953 97.167 -61.905 51.379
495954 97.198 -61.910 51.343
495955 97.229 -61.915 51.306
495956 97.259 -61.920 51.268
495957 97.290 -61.924 51.229
495958 97.320 -61.929 51.189
495959 97.351 -61.934 51.148
495960 97.400 -61.931 51.107
495961 97.456 -61.925 51.064
495962 97.523 -61.917 51.023
495963 97.590 -61.908 50.980
495964 97.657 -61.900 50.936
495965 97.724 -61.892 50.893
495966 97.791 -61.882 50.851
495967 97.823 -61.867 50.809
495968 97.859 -61.878 50.766
495969 97.881 -61.893 50.722
495970 97.906 -61.909 50.678
495971 97.930 -61.924 50.635
495972 97.954 -61.940 50.595
495973 97.978 -61.956 50.553
495974 98.002 -61.971 50.512
495975 98.026 -61.987 50.471
495976 98.051 -62.002 50.429
495977 98.075 -62.018 50.388
495978 98.099 -62.034 50.347
495979 98.123 -62.049 50.308
495980 98.147 -62.065 50.269
495981 98.171 -62.080 50.229
496438 88.527 -59.918 55.609
496439 88.579 -59.925 55.585
496440 88.631 -59.932 55.562
496441 88.686 -59.929 55.540
496442 88.746 -59.913 55.518
496443 88.805 -59.897 55.498
496444 88.861 -59.885 55.478
496445 88.918 -59.873 55.459
496446 88.965 -59.881 55.440
496447 89.004 -59.901 55.420
496448 89.045 -59.920 55.400
496449 89.085 -59.941 55.382
496450 89.125 -59.961 55.364
496451 89.150 -59.974 55.346
496452 89.163 -59.983 55.328
496453 89.203 -59.999 55.309
496454 89.243 -60.013 55.289
496455 89.284 -60.028 55.271
496456 89.326 -60.046 55.252
496457 89.369 -60.068 55.233
496458 89.411 -60.090 55.215
496459 89.454 -60.111 55.197
496460 89.496 -60.132 55.180
496461 89.539 -60.154 55.164
496462 89.582 -60.175 55.149
496463 89.624 -60.196 55.134
496464 89.667 -60.218 55.121
496465 89.710 -60.240 55.110
496466 89.753 -60.261 55.098
496467 89.796 -60.283 55.087
496468 89.838 -60.304 55.078
496469 89.881 -60.325 55.069
496470 89.923 -60.347 55.061
496471 89.966 -60.369 55.052
496472 90.009 -60.390 55.043
496473 90.054 -60.410 55.035
496474 90.100 -60.430 55.026
496475 90.146 -60.450 55.018
496476 90.191 -60.470 55.011
496477 90.237 -60.491 55.005
496478 90.282 -60.510 54.999
496479 90.327 -60.531 54.994
496480 90.373 -60.551 54.989
496481 90.418 -60.571 54.983
496482 90.458 -60.590 54.980
496483 90.497 -60.608 54.977
496484 90.535 -60.627 54.975
496485 90.580 -60.643 54.974
496486 90.625 -60.660 54.974
496487 90.671 -60.676 54.974
496488 90.717 -60.691 54.979
496489 90.762 -60.707 54.986
496490 90.811 -60.719 54.994
496491 90.860 -60.730 55.002
496492 90.898 -60.742 55.012
496493 90.937 -60.754 55.027
496494 90.975 -60.766 55.042
496495 91.014 -60.777 55.055
496496 91.052 -60.789 55.064
496497 91.091 -60.801 55.070
496498 91.130 -60.813 55.071
496499 91.168 -60.825 55.070
496500 91.207 -60.837 55.069
496501 91.245 -60.848 55.067
496502 91.284 -60.861 55.061
496503 91.322 -60.873 55.055
496504 91.361 -60.883 55.052
496505 91.400 -60.893 55.054
496506 91.440 -60.902 55.060
496507 91.480 -60.911 55.069
496508 91.519 -60.921 55.073
496509 91.559 -60.930 55.069
496510 91.599 -60.939 55.058
496511 91.638 -60.949 55.042
496512 91.678 -60.958 55.023
496513 91.718 -60.968 55.001
496514 91.757 -60.977 54.975
496515 91.797 -60.986 54.948
496516 91.809 -61.002 54.921
496517 91.850 -61.017 54.892
496518 91.889 -61.033 54.867
496519 91.929 -61.048 54.852
496520 91.969 -61.063 54.844
496521 92.008 -61.071 54.841
496522 92.048 -61.080 54.840
496523 92.086 -61.088 54.839
496524 92.125 -61.096 54.838
496525 92.164 -61.104 54.835
496526 92.203 -61.112 54.833
496527 92.242 -61.120 54.831
496528 92.281 -61.128 54.827
496529 92.320 -61.136 54.821
496530 92.361 -61.146 54.812
496531 92.409 -61.158 54.801
496532 92.456 -61.171 54.788
496533 92.504 -61.183 54.777
496534 92.552 -61.196 54.766
496535 92.592 -61.202 54.749
496536 92.630 -61.208 54.725
496537 92.668 -61.215 54.695
496538 92.707 -61.221 54.657
496539 92.745 -61.227 54.610
496540 92.784 -61.232 54.569
496541 92.823 -61.238 54.536
496542 92.861 -61.244 54.510
496543 92.899 -61.250 54.494
496544 92.937 -61.256 54.496
496545 92.976 -61.262 54.490
496546 93.014 -61.268 54.479
496547 93.053 -61.274 54.464
496548 93.091 -61.280 54.447
496549 93.130 -61.286 54.425
496550 93.168 -61.292 54.404
496551 93.206 -61.298 54.390
496552 93.246 -61.304 54.381
496553 93.286 -61.309 54.374
496554 93.327 -61.314 54.364
496555 93.368 -61.319 54.348
496556 93.409 -61.324 54.330
496557 93.448 -61.333 54.308
496558 93.488 -61.343 54.281
496559 93.528 -61.355 54.257
496560 93.567 -61.365 54.238
496561 93.607 -61.377 54.218
496562 93.646 -61.387 54.200
496563 93.686 -61.397 54.183
496564 93.726 -61.408 54.169
496565 93.767 -61.418 54.149
496566 93.809 -61.424 54.123
496567 93.852 -61.431 54.094
496568 93.895 -61.437 54.062
496569 93.937 -61.443 54.030
496570 93.980 -61.450 54.001
496571 94.023 -61.456 53.975
496572 94.066 -61.462 53.950
496573 94.108 -61.471 53.927
496574 94.149 -61.482 53.905
496575 94.191 -61.494 53.888
496576 94.233 -61.505 53.873
496577 94.274 -61.516 53.859
496578 94.317 -61.525 53.843
496579 94.359 -61.532 53.827
496580 94.402 -61.540 53.807
496581 94.444 -61.547 53.787
496582 94.483 -61.551 53.766
496583 94.522 -61.555 53.743
496584 94.560 -61.559 53.720
496585 94.599 -61.563 53.695
496586 94.637 -61.566 53.668
496587 94.676 -61.571 53.640
496588 94.715 -61.575 53.612
496589 94.750 -61.576 53.585
496590 94.785 -61.578 53.558
496591 94.821 -61.579 53.535
496592 94.856 -61.581 53.514
496593 94.893 -61.589 53.493
496594 94.929 -61.597 53.473
496595 94.965 -61.603 53.454
496596 95.001 -61.611 53.435
496597 95.037 -61.618 53.415
496598 95.073 -61.625 53.393
496599 95.111 -61.632 53.371
496600 95.148 -61.637 53.347
496601 95.186 -61.643 53.324
496602 95.224 -61.648 53.301
496603 95.262 -61.653 53.278
496604 95.302 -61.659 53.254
496605 95.343 -61.663 53.228
496606 95.385 -61.669 53.200
496607 95.422 -61.672 53.170
496608 95.461 -61.676 53.139
496609 95.499 -61.681 53.107
496610 95.537 -61.685 53.074
496611 95.576 -61.688 53.038
496612 95.614 -61.693 53.001
496613 95.652 -61.697 52.963
496614 95.690 -61.700 52.924
496615 95.728 -61.704 52.885
496616 95.766 -61.707 52.846
496617 95.805 -61.711 52.803
496618 95.843 -61.714 52.761
496619 95.881 -61.718 52.719
496620 95.920 -61.721 52.679
496621 95.962 -61.723 52.638
496622 96.003 -61.726 52.599
496623 96.044 -61.729 52.559
496624 96.085 -61.732 52.518
496625 96.126 -61.734 52.480
496626 96.168 -61.736 52.445
496627 96.205 -61.739 52.411
496628 96.242 -61.741 52.378
496629 96.277 -61.744 52.342
496630 96.310 -61.755 52.302
496631 96.351 -61.759 52.259
496632 96.392 -61.763 52.214
496633 96.433 -61.766 52.168
496634 96.474 -61.770 52.122
496635 96.515 -61.774 52.077
496636 96.555 -61.778 52.030
496637 96.595 -61.783 51.987
496638 96.631 -61.789 51.946
496639 96.668 -61.796 51.908
496640 96.704 -61.802 51.872
496641 96.741 -61.809 51.840
496642 96.778 -61.816 51.806
496643 96.815 -61.822 51.775
496644 96.857 -61.825 51.741
496645 96.897 -61.823 51.706
496646 96.934 -61.822 51.668
496647 96.965 -61.827 51.631
496648 96.996 -61.832 51.594
496649 97.026 -61.836 51.557
496650 97.057 -61.841 51.522
496651 97.088 -61.846 51.487
496652 97.118 -61.851 51.454
496653 97.149 -61.856 51.421
496654 97.179 -61.860 51.386
496655 97.210 -61.866 51.349
496656 97.241 -61.870 51.313
496657 97.272 -61.875 51.276
496658 97.302 -61.880 51.238
496659 97.333 -61.885 51.198
496660 97.378 -61.883 51.157
496661 97.432 -61.878 51.115
496662 97.486 -61.873 51.071
496663 97.539 -61.867 51.028
496664 97.593 -61.862 50.985
496665 97.659 -61.854 50.941
496666 97.726 -61.845 50.898
496667 97.793 -61.837 50.855
496668 97.834 -61.827 50.813
496669 97.859 -61.842 50.769
496670 97.883 -61.858 50.724
496671 97.908 -61.873 50.680
496672 97.932 -61.889 50.637
496673 97.956 -61.904 50.595
496674 97.980 -61.920 50.554
496675 98.004 -61.936 50.513
496676 98.028 -61.951 50.471
496677 98.053 -61.967 50.429
496678 98.077 -61.983 50.388
496679 98.101 -61.998 50.348
496680 98.125 -62.014 50.309
496681 98.149 -62.029 50.270
496682 98.173 -62.045 50.229
497135 88.345 -59.847 55.718
497136 88.397 -59.854 55.690
497137 88.450 -59.861 55.663
497138 88.501 -59.867 55.638
497139 88.553 -59.874 55.614
497140 88.606 -59.881 55.590
497141 88.658 -59.888 55.567
497142 88.714 -59.880 55.545
497143 88.773 -59.864 55.522
497144 88.833 -59.849 55.500
497145 88.889 -59.838 55.481
497146 88.944 -59.829 55.461
497147 88.984 -59.848 55.441
497148 89.024 -59.869 55.420
497149 89.064 -59.889 55.400
497150 89.104 -59.908 55.382
497151 89.144 -59.929 55.363
497152 89.171 -59.943 55.345
497153 89.183 -59.952 55.327
497154 89.218 -59.966 55.308
497155 89.258 -59.980 55.289
497156 89.299 -59.995 55.270
497157 89.339 -60.010 55.253
497158 89.382 -60.031 55.235
497159 89.424 -60.052 55.217
497160 89.467 -60.074 55.200
497161 89.510 -60.096 55.183
497162 89.553 -60.117 55.168
497163 89.596 -60.139 55.153
497164 89.638 -60.160 55.140
497165 89.681 -60.181 55.128
497166 89.724 -60.203 55.117
497167 89.766 -60.224 55.106
497168 89.809 -60.246 55.096
497169 89.852 -60.267 55.087
497170 89.894 -60.289 55.079
497171 89.937 -60.310 55.071
497172 89.980 -60.331 55.063
497173 90.024 -60.353 55.054
497174 90.070 -60.372 55.046
497175 90.115 -60.393 55.037
497176 90.161 -60.412 55.028
497177 90.206 -60.433 55.020
497178 90.252 -60.453 55.012
497179 90.297 -60.473 55.004
497180 90.342 -60.494 54.999
497181 90.388 -60.513 54.993
497182 90.434 -60.534 54.988
497183 90.476 -60.553 54.984
497184 90.515 -60.572 54.981
497185 90.553 -60.590 54.979
497186 90.599 -60.606 54.978
497187 90.645 -60.622 54.978
497188 90.690 -60.638 54.977
497189 90.736 -60.654 54.979
497190 90.782 -60.670 54.986
497191 90.829 -60.684 54.996
497192 90.873 -60.695 55.006
497193 90.911 -60.707 55.017
497194 90.950 -60.719 55.030
497195 90.989 -60.731 55.044
497196 91.027 -60.743 55.056
497197 91.066 -60.754 55.066
497198 91.104 -60.767 55.074
497199 91.143 -60.778 55.077
497200 91.183 -60.787 55.076
497201 91.222 -60.796 55.074
497202 91.262 -60.806 55.072
497203 91.302 -60.815 55.065
497204 91.341 -60.824 55.056
497205 91.381 -60.834 55.046
497206 91.421 -60.843 55.042
497207 91.460 -60.853 55.047
497208 91.500 -60.862 55.056
497209 91.539 -60.871 55.062
497210 91.579 -60.881 55.061
497211 91.618 -60.890 55.054
497212 91.658 -60.900 55.041
497213 91.698 -60.909 55.025
497214 91.737 -60.918 55.003
497215 91.777 -60.928 54.979
497216 91.810 -60.939 54.953
497217 91.816 -60.955 54.923
497218 91.856 -60.970 54.892
497219 91.896 -60.985 54.863
497220 91.936 -61.001 54.844
497221 91.976 -61.016 54.832
497222 92.016 -61.032 54.823
497223 92.056 -61.047 54.817
497224 92.095 -61.055 54.813
497225 92.134 -61.063 54.810
497226 92.173 -61.071 54.810
497227 92.212 -61.080 54.812
497228 92.251 -61.088 54.817
497229 92.290 -61.096 54.816
497230 92.329 -61.104 54.810
497231 92.368 -61.112 54.803
497232 92.413 -61.123 54.792
497233 92.460 -61.136 54.781
497234 92.508 -61.148 54.773
497235 92.556 -61.161 54.761
497236 92.595 -61.167 54.744
497237 92.633 -61.173 54.719
497238 92.671 -61.179 54.686
497239 92.710 -61.185 54.650
497240 92.748 -61.191 54.604
497241 92.787 -61.197 54.564
497242 92.825 -61.202 54.530
497243 92.864 -61.208 54.503
497244 92.902 -61.215 54.488
497245 92.940 -61.221 54.489
497246 92.979 -61.227 54.480
497247 93.017 -61.233 54.467
497248 93.056 -61.239 54.452
497249 93.094 -61.244 54.436
497250 93.133 -61.250 54.417
497251 93.174 -61.255 54.398
497252 93.214 -61.261 54.383
497253 93.255 -61.266 54.372
497254 93.295 -61.271 54.362
497255 93.336 -61.277 54.352
497256 93.377 -61.282 54.335
497257 93.418 -61.287 54.318
497258 93.458 -61.295 54.300
497259 93.497 -61.306 54.282
497260 93.537 -61.318 54.268
497261 93.576 -61.328 54.255
497262 93.616 -61.340 54.237
497263 93.655 -61.350 54.220
497264 93.694 -61.362 54.208
497265 93.734 -61.372 54.198
497266 93.774 -61.383 54.179
497267 93.817 -61.389 54.153
497268 93.859 -61.396 54.120
497269 93.902 -61.402 54.085
497270 93.945 -61.409 54.050
497271 93.987 -61.415 54.022
497272 94.030 -61.422 53.998
497273 94.073 -61.428 53.976
497274 94.116 -61.434 53.953
497275 94.158 -61.440 53.931
497276 94.200 -61.451 53.913
497277 94.242 -61.462 53.897
497278 94.283 -61.473 53.882
497279 94.325 -61.484 53.867
497280 94.367 -61.491 53.851
497281 94.410 -61.498 53.831
497282 94.450 -61.503 53.811
497283 94.489 -61.507 53.789
497284 94.528 -61.511 53.765
497285 94.566 -61.516 53.739
497286 94.605 -61.519 53.714
497287 94.643 -61.523 53.684
497288 94.682 -61.527 53.653
497289 94.721 -61.531 53.623
497290 94.756 -61.533 53.593
497291 94.791 -61.534 53.562
497292 94.827 -61.537 53.535
497293 94.863 -61.544 53.509
497294 94.899 -61.551 53.484
497295 94.935 -61.559 53.459
497296 94.972 -61.566 53.438
497297 95.008 -61.573 53.416
497298 95.044 -61.580 53.394
497299 95.080 -61.588 53.370
497300 95.118 -61.594 53.346
497301 95.156 -61.599 53.322
497302 95.193 -61.604 53.300
497303 95.231 -61.610 53.278
497304 95.269 -61.615 53.259
497305 95.307 -61.620 53.238
497306 95.347 -61.625 53.215
497307 95.388 -61.630 53.187
497308 95.427 -61.634 53.158
497309 95.466 -61.638 53.129
497310 95.504 -61.642 53.099
497311 95.542 -61.646 53.067
497312 95.580 -61.650 53.031
497313 95.618 -61.654 52.994
497314 95.656 -61.659 52.957
497315 95.694 -61.663 52.920
497316 95.733 -61.666 52.881
497317 95.771 -61.669 52.843
497318 95.809 -61.673 52.802
497319 95.847 -61.676 52.761
497320 95.885 -61.680 52.720
497321 95.927 -61.683 52.679
497322 95.971 -61.686 52.639
497323 96.014 -61.688 52.599
497324 96.058 -61.691 52.557
497325 96.100 -61.694 52.515
497326 96.139 -61.697 52.475
497327 96.176 -61.699 52.437
497328 96.212 -61.702 52.400
497329 96.248 -61.704 52.364
497330 96.284 -61.707 52.326
497331 96.316 -61.717 52.285
497332 96.349 -61.726 52.242
497333 96.390 -61.730 52.198
497334 96.431 -61.734 52.152
497335 96.472 -61.738 52.108
497336 96.513 -61.742 52.064
497337 96.553 -61.746 52.020
497338 96.593 -61.751 51.978
497339 96.630 -61.757 51.938
497340 96.667 -61.763 51.900
497341 96.703 -61.770 51.865
497342 96.740 -61.776 51.832
497343 96.777 -61.783 51.799
497344 96.815 -61.788 51.768
497345 96.862 -61.787 51.736
497346 96.901 -61.785 51.702
497347 96.941 -61.782 51.666
497348 96.977 -61.782 51.631
497349 97.008 -61.788 51.596
497350 97.038 -61.792 51.561
497351 97.069 -61.797 51.526
497352 97.100 -61.802 51.492
497353 97.131 -61.807 51.458
497354 97.161 -61.811 51.424
497355 97.191 -61.816 51.390
497356 97.222 -61.821 51.354
497357 97.253 -61.826 51.318
497358 97.284 -61.831 51.281
497359 97.314 -61.835 51.242
497360 97.357 -61.835 51.202
497361 97.410 -61.830 51.161
497362 97.464 -61.825 51.119
497363 97.518 -61.820 51.076
497364 97.572 -61.814 51.033
497365 97.626 -61.808 50.990
497366 97.679 -61.803 50.946
497367 97.733 -61.798 50.902
497368 97.794 -61.791 50.859
497369 97.837 -61.791 50.817
497370 97.861 -61.807 50.773
497371 97.885 -61.823 50.729
497372 97.909 -61.838 50.683
497373 97.933 -61.854 50.639
497374 97.957 -61.870 50.597
497375 97.982 -61.885 50.556
497376 98.006 -61.901 50.515
497377 98.030 -61.917 50.473
497378 98.054 -61.932 50.431
497379 98.078 -61.948 50.389
497380 98.102 -61.964 50.350
497381 98.127 -61.979 50.311
497382 98.151 -61.995 50.271
497383 98.175 -62.010 50.231
497832 88.164 -59.775 55.843
497833 88.216 -59.782 55.814
497834 88.268 -59.789 55.785
497835 88.320 -59.795 55.756
497836 88.372 -59.803 55.727
497837 88.424 -59.810 55.699
497838 88.476 -59.817 55.671
497839 88.528 -59.823 55.646
497840 88.580 -59.830 55.622
497841 88.632 -59.837 55.598
497842 88.684 -59.844 55.574
497843 88.743 -59.832 55.551
497844 88.802 -59.817 55.527
497845 88.859 -59.803 55.505
497846 88.916 -59.792 55.485
497847 88.964 -59.797 55.465
497848 89.003 -59.817 55.444
497849 89.044 -59.836 55.423
497850 89.084 -59.857 55.403
497851 89.124 -59.877 55.383
497852 89.164 -59.897 55.365
497853 89.192 -59.912 55.346
497854 89.204 -59.920 55.328
497855 89.233 -59.933 55.310
497856 89.273 -59.948 55.292
497857 89.314 -59.962 55.273
497858 89.354 -59.977 55.257
497859 89.396 -59.995 55.240
497860 89.438 -60.016 55.223
497861 89.481 -60.037 55.206
497862 89.524 -60.059 55.190
497863 89.566 -60.080 55.175
497864 89.609 -60.102 55.160
497865 89.652 -60.123 55.148
497866 89.694 -60.145 55.136
497867 89.737 -60.166 55.126
497868 89.780 -60.187 55.115
497869 89.823 -60.209 55.106
497870 89.866 -60.231 55.097
497871 89.908 -60.252 55.090
497872 89.951 -60.274 55.082
497873 89.994 -60.295 55.074
497874 90.039 -60.315 55.066
497875 90.085 -60.335 55.058
497876 90.130 -60.355 55.048
497877 90.176 -60.375 55.039
497878 90.221 -60.396 55.030
497879 90.267 -60.416 55.021
497880 90.312 -60.436 55.013
497881 90.358 -60.456 55.006
497882 90.403 -60.476 55.000
497883 90.449 -60.496 54.994
497884 90.494 -60.516 54.991
497885 90.532 -60.535 54.988
497886 90.573 -60.553 54.985
497887 90.618 -60.569 54.982
497888 90.664 -60.585 54.981
497889 90.710 -60.601 54.980
497890 90.755 -60.617 54.982
497891 90.801 -60.633 54.989
497892 90.847 -60.649 55.001
497893 90.886 -60.660 55.014
497894 90.925 -60.672 55.026
497895 90.965 -60.681 55.037
497896 91.005 -60.691 55.047
497897 91.044 -60.700 55.057
497898 91.084 -60.710 55.066
497899 91.123 -60.719 55.075
497900 91.163 -60.728 55.080
497901 91.203 -60.738 55.080
497902 91.242 -60.747 55.078
497903 91.282 -60.756 55.075
497904 91.321 -60.766 55.068
497905 91.361 -60.775 55.057
497906 91.401 -60.785 55.042
497907 91.440 -60.794 55.033
497908 91.480 -60.803 55.033
497909 91.520 -60.813 55.040
497910 91.559 -60.822 55.047
497911 91.599 -60.832 55.049
497912 91.639 -60.841 55.045
497913 91.678 -60.850 55.036
497914 91.718 -60.860 55.024
497915 91.757 -60.869 55.005
497916 91.797 -60.879 54.982
497917 91.819 -60.891 54.955
497918 91.824 -60.907 54.924
497919 91.863 -60.923 54.890
497920 91.903 -60.938 54.861
497921 91.943 -60.954 54.841
497922 91.983 -60.969 54.828
497923 92.023 -60.984 54.818
497924 92.063 -61.000 54.808
497925 92.102 -61.015 54.797
497926 92.143 -61.030 54.789
497927 92.181 -61.039 54.786
497928 92.221 -61.047 54.788
497929 92.260 -61.055 54.795
497930 92.298 -61.064 54.796
497931 92.338 -61.071 54.792
497932 92.376 -61.080 54.788
497933 92.417 -61.089 54.781
497934 92.464 -61.101 54.772
497935 92.512 -61.113 54.764
497936 92.559 -61.125 54.753
497937 92.597 -61.131 54.736
497938 92.636 -61.137 54.711
497939 92.674 -61.143 54.679
497940 92.713 -61.149 54.645
497941 92.751 -61.155 54.603
497942 92.790 -61.161 54.565
497943 92.828 -61.168 54.529
497944 92.866 -61.173 54.500
497945 92.905 -61.179 54.487
497946 92.943 -61.185 54.483
497947 92.982 -61.191 54.474
497948 93.020 -61.197 54.461
497949 93.060 -61.202 54.446
497950 93.101 -61.208 54.429
497951 93.142 -61.213 54.410
497952 93.183 -61.218 54.391
497953 93.223 -61.224 54.375
497954 93.264 -61.229 54.361
497955 93.304 -61.234 54.350
497956 93.345 -61.240 54.339
497957 93.386 -61.245 54.324
497958 93.427 -61.250 54.311
497959 93.467 -61.258 54.299
497960 93.506 -61.269 54.291
497961 93.546 -61.280 54.286
497962 93.585 -61.291 54.272
497963 93.625 -61.302 54.252
497964 93.664 -61.313 54.236
497965 93.703 -61.324 54.229
497966 93.743 -61.336 54.221
497967 93.782 -61.346 54.205
497968 93.824 -61.354 54.178
497969 93.867 -61.360 54.143
497970 93.909 -61.367 54.105
497971 93.952 -61.373 54.074
497972 93.995 -61.380 54.048
497973 94.037 -61.386 54.027
497974 94.080 -61.393 54.005
497975 94.123 -61.399 53.982
497976 94.166 -61.406 53.960
497977 94.208 -61.412 53.941
497978 94.251 -61.420 53.924
497979 94.292 -61.431 53.907
497980 94.334 -61.442 53.892
497981 94.376 -61.450 53.875
497982 94.417 -61.456 53.855
497983 94.456 -61.460 53.835
497984 94.495 -61.464 53.812
497985 94.534 -61.468 53.786
497986 94.572 -61.472 53.760
497987 94.611 -61.476 53.732
497988 94.649 -61.480 53.700
497989 94.688 -61.484 53.667
497990 94.726 -61.487 53.635
497991 94.762 -61.489 53.603
497992 94.797 -61.491 53.569
497993 94.834 -61.499 53.540
497994 94.870 -61.506 53.510
497995 94.906 -61.513 53.480
497996 94.942 -61.520 53.453
497997 94.978 -61.528 53.428
497998 95.015 -61.535 53.404
497999 95.051 -61.542 53.379
498000 95.087 -61.550 53.354
498001 95.124 -61.555 53.329
498002 95.162 -61.561 53.303
498003 95.200 -61.566 53.278
498004 95.238 -61.572 53.256
498005 95.276 -61.576 53.235
498006 95.314 -61.581 53.215
498007 95.352 -61.587 53.192
498008 95.392 -61.591 53.165
498009 95.432 -61.596 53.140
498010 95.470 -61.600 53.114
498011 95.509 -61.604 53.086
498012 95.547 -61.608 53.055
498013 95.585 -61.612 53.019
498014 95.623 -61.616 52.982
498015 95.661 -61.620 52.946
498016 95.699 -61.625 52.909
498017 95.737 -61.628 52.872
498018 95.775 -61.632 52.834
498019 95.813 -61.635 52.795
498020 95.851 -61.639 52.756
498021 95.890 -61.642 52.717
498022 95.932 -61.645 52.677
498023 95.976 -61.648 52.638
498024 96.019 -61.650 52.596
498025 96.063 -61.653 52.554
498026 96.105 -61.657 52.512
498027 96.146 -61.660 52.471
498028 96.182 -61.662 52.430
498029 96.218 -61.665 52.391
498030 96.254 -61.667 52.351
498031 96.290 -61.669 52.311
498032 96.323 -61.678 52.269
498033 96.354 -61.689 52.227
498034 96.389 -61.698 52.184
498035 96.430 -61.702 52.140
498036 96.470 -61.706 52.097
498037 96.511 -61.710 52.055
498038 96.552 -61.713 52.012
498039 96.592 -61.718 51.973
498040 96.629 -61.724 51.933
498041 96.665 -61.731 51.895
498042 96.702 -61.738 51.859
498043 96.739 -61.744 51.825
498044 96.776 -61.751 51.791
498045 96.821 -61.751 51.760
498046 96.866 -61.749 51.729
498047 96.905 -61.747 51.697
498048 96.945 -61.744 51.664
498049 96.985 -61.741 51.629
498050 97.020 -61.743 51.595
498051 97.050 -61.747 51.560
498052 97.081 -61.753 51.527
498053 97.112 -61.757 51.493
498054 97.143 -61.763 51.459
498055 97.173 -61.767 51.425
498056 97.204 -61.772 51.391
498057 97.234 -61.777 51.356
498058 97.265 -61.782 51.319
498059 97.296 -61.786 51.281
498060 97.335 -61.787 51.242
498061 97.389 -61.782 51.202
498062 97.443 -61.776 51.161
498063 97.497 -61.771 51.120
498064 97.550 -61.766 51.077
498065 97.604 -61.760 51.035
498066 97.658 -61.755 50.991
498067 97.712 -61.750 50.948
498068 97.766 -61.744 50.904
498069 97.806 -61.748 50.862
498070 97.839 -61.756 50.818
498071 97.863 -61.772 50.775
498072 97.887 -61.788 50.731
498073 97.911 -61.803 50.686
498074 97.935 -61.819 50.641
498075 97.959 -61.834 50.599
498076 97.984 -61.850 50.557
498077 98.008 -61.866 50.516
498078 98.032 -61.881 50.475
498079 98.056 -61.897 50.433
498080 98.080 -61.912 50.392
498081 98.104 -61.928 50.352
498082 98.129 -61.944 50.313
498083 98.153 -61.959 50.274
498084 98.177 -61.975 50.235
498531 88.087 -59.717 55.914
498532 88.139 -59.724 55.883
498533 88.191 -59.731 55.852
498534 88.243 -59.738 55.823
498535 88.295 -59.745 55.794
498536 88.347 -59.751 55.765
498537 88.399 -59.758 55.736
498538 88.451 -59.766 55.708
498539 88.503 -59.773 55.680
498540 88.555 -59.779 55.654
498541 88.607 -59.786 55.630
498542 88.659 -59.793 55.607
498543 88.711 -59.799 55.582
498544 88.770 -59.784 55.558
498545 88.830 -59.768 55.534
498546 88.887 -59.756 55.512
498547 88.943 -59.745 55.491
498548 88.983 -59.764 55.471
498549 89.023 -59.785 55.449
498550 89.063 -59.804 55.428
498551 89.103 -59.824 55.407
498552 89.143 -59.845 55.387
498553 89.184 -59.864 55.368
498554 89.214 -59.880 55.349
498555 89.226 -59.889 55.331
498556 89.248 -59.900 55.314
498557 89.288 -59.915 55.296
498558 89.329 -59.930 55.278
498559 89.369 -59.945 55.263
498560 89.410 -59.959 55.246
498561 89.452 -59.980 55.230
498562 89.494 -60.001 55.215
498563 89.537 -60.022 55.198
498564 89.580 -60.043 55.184
498565 89.623 -60.065 55.170
498566 89.666 -60.086 55.158
498567 89.708 -60.108 55.147
498568 89.751 -60.130 55.136
498569 89.794 -60.151 55.127
498570 89.836 -60.172 55.118
498571 89.879 -60.194 55.110
498572 89.921 -60.215 55.102
498573 89.964 -60.237 55.095
498574 90.009 -60.257 55.087
498575 90.054 -60.278 55.079
498576 90.100 -60.298 55.070
498577 90.146 -60.318 55.061
498578 90.191 -60.338 55.051
498579 90.236 -60.358 55.042
498580 90.282 -60.378 55.032
498581 90.327 -60.398 55.023
498582 90.373 -60.419 55.015
498583 90.418 -60.439 55.009
498584 90.464 -60.459 55.002
498585 90.509 -60.479 54.999
498586 90.550 -60.498 54.996
498587 90.592 -60.516 54.992
498588 90.637 -60.532 54.988
498589 90.683 -60.548 54.985
498590 90.728 -60.564 54.984
498591 90.774 -60.580 54.986
498592 90.818 -60.591 54.992
498593 90.861 -60.603 55.002
498594 90.901 -60.616 55.014
498595 90.944 -60.623 55.027
498596 90.984 -60.632 55.038
498597 91.024 -60.641 55.047
498598 91.064 -60.651 55.056
498599 91.104 -60.660 55.063
498600 91.143 -60.669 55.072
498601 91.183 -60.679 55.078
498602 91.223 -60.688 55.081
498603 91.262 -60.698 55.079
498604 91.302 -60.707 55.075
498605 91.342 -60.716 55.067
498606 91.381 -60.726 55.055
498607 91.421 -60.735 55.040
498608 91.460 -60.745 55.027
498609 91.500 -60.754 55.020
498610 91.539 -60.763 55.024
498611 91.579 -60.773 55.031
498612 91.619 -60.782 55.034
498613 91.658 -60.792 55.034
498614 91.698 -60.801 55.029
498615 91.738 -60.810 55.021
498616 91.777 -60.820 55.004
498617 91.817 -60.829 54.981
498618 91.827 -60.844 54.954
498619 91.833 -60.860 54.923
498620 91.869 -60.875 54.889
498621 91.910 -60.891 54.862
498622 91.949 -60.906 54.844
498623 91.990 -60.922 54.830
498624 92.029 -60.937 54.818
498625 92.069 -60.952 54.813
498626 92.109 -60.968 54.802
498627 92.149 -60.983 54.790
498628 92.189 -60.999 54.778
498629 92.229 -61.014 54.770
498630 92.268 -61.023 54.770
498631 92.307 -61.031 54.777
498632 92.346 -61.039 54.780
498633 92.385 -61.047 54.780
498634 92.424 -61.055 54.774
498635 92.468 -61.066 54.764
498636 92.516 -61.078 54.750
498637 92.562 -61.090 54.740
498638 92.600 -61.096 54.724
498639 92.639 -61.102 54.702
498640 92.677 -61.108 54.675
498641 92.716 -61.114 54.641
498642 92.754 -61.120 54.605
498643 92.793 -61.126 54.568
498644 92.831 -61.132 54.532
498645 92.869 -61.137 54.501
498646 92.908 -61.143 54.490
498647 92.947 -61.149 54.482
498648 92.988 -61.155 54.474
498649 93.029 -61.160 54.463
498650 93.070 -61.165 54.448
498651 93.110 -61.171 54.430
498652 93.151 -61.176 54.409
498653 93.192 -61.180 54.392
498654 93.232 -61.186 54.378
498655 93.273 -61.191 54.367
498656 93.314 -61.196 54.355
498657 93.354 -61.202 54.342
498658 93.395 -61.207 54.330
498659 93.436 -61.212 54.321
498660 93.476 -61.221 54.313
498661 93.515 -61.232 54.306
498662 93.555 -61.243 54.297
498663 93.594 -61.254 54.284
498664 93.634 -61.265 54.268
498665 93.673 -61.276 54.255
498666 93.712 -61.287 54.248
498667 93.752 -61.298 54.239
498668 93.791 -61.308 54.222
498669 93.835 -61.318 54.194
498670 93.877 -61.324 54.161
498671 93.919 -61.331 54.128
498672 93.961 -61.338 54.105
498673 94.003 -61.344 54.081
498674 94.046 -61.351 54.056
498675 94.087 -61.358 54.032
498676 94.130 -61.364 54.009
498677 94.173 -61.371 53.989
498678 94.215 -61.377 53.969
498679 94.258 -61.384 53.951
498680 94.300 -61.390 53.934
498681 94.342 -61.400 53.918
498682 94.385 -61.409 53.900
498683 94.423 -61.412 53.879
498684 94.462 -61.416 53.857
498685 94.501 -61.421 53.833
498686 94.540 -61.425 53.807
498687 94.578 -61.428 53.779
498688 94.617 -61.432 53.749
498689 94.656 -61.436 53.715
498690 94.694 -61.440 53.681
498691 94.733 -61.443 53.647
498692 94.768 -61.447 53.612
498693 94.805 -61.453 53.577
498694 94.840 -61.461 53.545
498695 94.877 -61.468 53.512
498696 94.913 -61.475 53.480
498697 94.949 -61.482 53.450
498698 94.985 -61.490 53.422
498699 95.021 -61.497 53.395
498700 95.057 -61.504 53.369
498701 95.094 -61.512 53.343
498702 95.131 -61.517 53.316
498703 95.169 -61.522 53.288
498704 95.207 -61.528 53.261
498705 95.245 -61.532 53.236
498706 95.283 -61.538 53.212
498707 95.321 -61.543 53.190
498708 95.358 -61.547 53.166
498709 95.396 -61.553 53.142
498710 95.437 -61.557 53.118
498711 95.475 -61.562 53.094
498712 95.513 -61.566 53.067
498713 95.551 -61.570 53.037
498714 95.589 -61.574 53.002
498715 95.627 -61.578 52.966
498716 95.665 -61.582 52.931
498717 95.703 -61.586 52.895
498718 95.741 -61.591 52.858
498719 95.779 -61.594 52.820
498720 95.818 -61.597 52.782
498721 95.856 -61.601 52.745
498722 95.894 -61.604 52.707
498723 95.937 -61.607 52.668
498724 95.981 -61.610 52.630
498725 96.024 -61.613 52.590
498726 96.066 -61.616 52.549
498727 96.107 -61.619 52.508
498728 96.147 -61.622 52.468
498729 96.187 -61.625 52.428
498730 96.224 -61.628 52.387
498731 96.260 -61.630 52.346
498732 96.296 -61.632 52.304
498733 96.330 -61.638 52.262
498734 96.362 -61.650 52.219
498735 96.392 -61.662 52.177
498736 96.428 -61.670 52.134
498737 96.468 -61.674 52.092
498738 96.509 -61.678 52.051
498739 96.550 -61.682 52.009
498740 96.591 -61.685 51.970
498741 96.627 -61.692 51.931
498742 96.664 -61.698 51.893
498743 96.701 -61.705 51.856
498744 96.738 -61.711 51.820
498745 96.780 -61.713 51.785
498746 96.827 -61.713 51.753
498747 96.870 -61.711 51.723
498748 96.909 -61.709 51.692
498749 96.949 -61.706 51.660
498750 96.989 -61.704 51.626
498751 97.029 -61.701 51.591
498752 97.063 -61.704 51.557
498753 97.093 -61.708 51.524
498754 97.124 -61.713 51.492
498755 97.155 -61.718 51.458
498756 97.185 -61.722 51.424
498757 97.216 -61.728 51.390
498758 97.247 -61.732 51.355
498759 97.277 -61.738 51.319
498760 97.314 -61.739 51.279
498761 97.368 -61.734 51.241
498762 97.422 -61.729 51.200
498763 97.475 -61.723 51.159
498764 97.529 -61.718 51.118
498765 97.583 -61.713 51.076
498766 97.637 -61.707 51.034
498767 97.686 -61.705 50.991
498768 97.726 -61.711 50.947
498769 97.765 -61.716 50.904
498770 97.810 -61.712 50.862
498771 97.840 -61.721 50.818
498772 97.864 -61.736 50.775
498773 97.888 -61.752 50.731
498774 97.913 -61.768 50.686
498775 97.937 -61.783 50.643
498776 97.961 -61.799 50.601
498777 97.985 -61.814 50.559
498778 98.009 -61.830 50.518
498779 98.033 -61.846 50.477
498780 98.058 -61.861 50.436
498781 98.082 -61.877 50.394
498782 98.106 -61.893 50.355
498783 98.130 -61.908 50.316
498784 98.154 -61.924 50.277
498785 98.178 -61.939 50.238
499231 88.061 -59.666 55.954
499232 88.114 -59.673 55.923
499233 88.166 -59.680 55.892
499234 88.217 -59.687 55.862
499235 88.270 -59.694 55.833
499236 88.322 -59.701 55.804
499237 88.374 -59.707 55.775
499238 88.426 -59.714 55.746
499239 88.478 -59.721 55.719
499240 88.530 -59.729 55.690
499241 88.582 -59.736 55.664
499242 88.634 -59.742 55.640
499243 88.686 -59.749 55.616
499244 88.740 -59.751 55.592
499245 88.799 -59.736 55.567
499246 88.857 -59.720 55.543
499247 88.914 -59.709 55.521
499248 88.963 -59.712 55.499
499249 89.002 -59.732 55.478
499250 89.043 -59.752 55.457
499251 89.083 -59.773 55.435
499252 89.123 -59.792 55.414
499253 89.163 -59.812 55.394
499254 89.203 -59.833 55.376
499255 89.235 -59.849 55.356
499256 89.247 -59.858 55.338
499257 89.263 -59.867 55.320
499258 89.303 -59.882 55.303
499259 89.344 -59.897 55.286
499260 89.384 -59.911 55.271
499261 89.424 -59.927 55.255
499262 89.466 -59.942 55.239
499263 89.508 -59.964 55.225
499264 89.551 -59.986 55.209
499265 89.594 -60.007 55.195
499266 89.636 -60.028 55.183
499267 89.679 -60.050 55.171
499268 89.722 -60.071 55.160
499269 89.764 -60.093 55.149
499270 89.807 -60.115 55.140
499271 89.850 -60.136 55.132
499272 89.893 -60.157 55.124
499273 89.936 -60.179 55.117
499274 89.979 -60.199 55.109
499275 90.024 -60.220 55.102
499276 90.070 -60.240 55.094
499277 90.115 -60.260 55.085
499278 90.161 -60.281 55.075
499279 90.206 -60.300 55.066
499280 90.251 -60.321 55.056
499281 90.297 -60.340 55.045
499282 90.342 -60.361 55.034
499283 90.388 -60.381 55.026
499284 90.434 -60.401 55.018
499285 90.479 -60.422 55.012
499286 90.524 -60.441 55.007
499287 90.568 -60.461 55.004
499288 90.611 -60.478 55.001
499289 90.656 -60.494 54.996
499290 90.702 -60.510 54.991
499291 90.745 -60.520 54.990
499292 90.788 -60.530 54.992
499293 90.831 -60.539 54.996
499294 90.874 -60.553 55.003
499295 90.912 -60.575 55.011
499296 90.955 -60.582 55.023
499297 90.997 -60.588 55.033
499298 91.039 -60.596 55.043
499299 91.082 -60.603 55.051
499300 91.123 -60.610 55.058
499301 91.163 -60.619 55.066
499302 91.203 -60.629 55.072
499303 91.242 -60.638 55.075
499304 91.282 -60.648 55.075
499305 91.322 -60.657 55.071
499306 91.361 -60.666 55.061
499307 91.401 -60.676 55.049
499308 91.441 -60.685 55.034
499309 91.480 -60.695 55.021
499310 91.520 -60.704 55.012
499311 91.560 -60.713 55.010
499312 91.599 -60.723 55.015
499313 91.639 -60.732 55.020
499314 91.678 -60.742 55.022
499315 91.718 -60.751 55.020
499316 91.757 -60.760 55.014
499317 91.797 -60.770 54.999
499318 91.830 -60.781 54.978
499319 91.836 -60.797 54.951
499320 91.842 -60.813 54.921
499321 91.876 -60.828 54.888
499322 91.916 -60.844 54.863
499323 91.956 -60.859 54.848
499324 91.996 -60.874 54.833
499325 92.036 -60.890 54.819
499326 92.076 -60.905 54.816
499327 92.116 -60.920 54.806
499328 92.156 -60.936 54.794
499329 92.196 -60.951 54.779
499330 92.236 -60.967 54.766
499331 92.276 -60.983 54.757
499332 92.316 -60.998 54.760
499333 92.355 -61.007 54.767
499334 92.394 -61.014 54.769
499335 92.433 -61.023 54.764
499336 92.472 -61.031 54.754
499337 92.520 -61.043 54.739
499338 92.565 -61.055 54.723
499339 92.603 -61.061 54.706
499340 92.642 -61.066 54.687
499341 92.680 -61.072 54.665
499342 92.719 -61.078 54.636
499343 92.757 -61.084 54.608
499344 92.795 -61.090 54.575
499345 92.835 -61.096 54.539
499346 92.875 -61.101 54.506
499347 92.916 -61.106 54.495
499348 92.956 -61.111 54.491
499349 92.997 -61.117 54.486
499350 93.038 -61.122 54.477
499351 93.079 -61.127 54.463
499352 93.120 -61.133 54.443
499353 93.160 -61.138 54.417
499354 93.201 -61.143 54.401
499355 93.241 -61.149 54.392
499356 93.282 -61.154 54.384
499357 93.323 -61.159 54.372
499358 93.364 -61.165 54.358
499359 93.405 -61.170 54.345
499360 93.445 -61.175 54.339
499361 93.485 -61.183 54.333
499362 93.524 -61.195 54.324
499363 93.564 -61.205 54.311
499364 93.604 -61.217 54.298
499365 93.643 -61.227 54.284
499366 93.682 -61.239 54.272
499367 93.722 -61.249 54.265
499368 93.761 -61.259 54.254
499369 93.801 -61.270 54.235
499370 93.845 -61.280 54.208
499371 93.888 -61.289 54.177
499372 93.929 -61.296 54.148
499373 93.972 -61.302 54.131
499374 94.014 -61.309 54.109
499375 94.056 -61.315 54.085
499376 94.098 -61.321 54.059
499377 94.140 -61.328 54.035
499378 94.183 -61.335 54.014
499379 94.225 -61.341 53.994
499380 94.267 -61.348 53.975
499381 94.309 -61.355 53.958
499382 94.351 -61.361 53.940
499383 94.393 -61.368 53.921
499384 94.431 -61.371 53.898
499385 94.470 -61.375 53.874
499386 94.508 -61.378 53.848
499387 94.546 -61.382 53.820
499388 94.585 -61.385 53.790
499389 94.623 -61.389 53.759
499390 94.662 -61.393 53.725
499391 94.701 -61.396 53.690
499392 94.739 -61.401 53.656
499393 94.775 -61.408 53.619
499394 94.811 -61.415 53.584
499395 94.847 -61.423 53.550
499396 94.883 -61.430 53.516
499397 94.919 -61.437 53.483
499398 94.956 -61.444 53.451
499399 94.992 -61.452 53.419
499400 95.028 -61.459 53.391
499401 95.064 -61.466 53.363
499402 95.101 -61.473 53.335
499403 95.138 -61.478 53.306
499404 95.176 -61.484 53.275
499405 95.214 -61.488 53.246
499406 95.252 -61.494 53.217
499407 95.290 -61.499 53.190
499408 95.328 -61.503 53.165
499409 95.366 -61.509 53.139
499410 95.403 -61.514 53.117
499411 95.441 -61.519 53.094
499412 95.480 -61.523 53.070
499413 95.518 -61.528 53.044
499414 95.556 -61.531 53.015
499415 95.594 -61.535 52.980
499416 95.632 -61.540 52.946
499417 95.670 -61.544 52.912
499418 95.708 -61.548 52.878
499419 95.746 -61.552 52.841
499420 95.784 -61.556 52.805
499421 95.822 -61.559 52.768
499422 95.860 -61.563 52.731
499423 95.899 -61.566 52.695
499424 95.942 -61.569 52.659
499425 95.986 -61.572 52.621
499426 96.028 -61.575 52.583
499427 96.068 -61.578 52.542
499428 96.109 -61.581 52.502
499429 96.149 -61.584 52.463
499430 96.189 -61.587 52.423
499431 96.230 -61.590 52.384
499432 96.266 -61.593 52.344
499433 96.303 -61.595 52.303
499434 96.338 -61.599 52.260
499435 96.369 -61.611 52.217
499436 96.400 -61.622 52.173
499437 96.431 -61.634 52.130
499438 96.467 -61.641 52.089
499439 96.508 -61.645 52.049
499440 96.548 -61.649 52.007
499441 96.589 -61.653 51.968
499442 96.626 -61.660 51.929
499443 96.663 -61.666 51.891
499444 96.699 -61.672 51.854
499445 96.739 -61.677 51.818
499446 96.786 -61.676 51.782
499447 96.833 -61.675 51.749
499448 96.874 -61.673 51.717
499449 96.913 -61.671 51.686
499450 96.953 -61.668 51.655
499451 96.993 -61.666 51.622
499452 97.033 -61.663 51.587
499453 97.073 -61.660 51.553
499454 97.105 -61.664 51.520
499455 97.136 -61.669 51.487
499456 97.167 -61.674 51.454
499457 97.197 -61.678 51.420
499458 97.228 -61.683 51.386
499459 97.259 -61.688 51.351
499460 97.293 -61.691 51.313
499461 97.347 -61.686 51.275
499462 97.400 -61.681 51.236
499463 97.454 -61.675 51.195
499464 97.508 -61.670 51.154
499465 97.562 -61.665 51.113
499466 97.606 -61.666 51.072
499467 97.646 -61.672 51.030
499468 97.685 -61.678 50.987
499469 97.724 -61.684 50.945
499470 97.768 -61.677 50.902
499471 97.814 -61.677 50.859
499472 97.842 -61.685 50.815
499473 97.866 -61.701 50.772
499474 97.890 -61.717 50.728
499475 97.915 -61.732 50.685
499476 97.939 -61.748 50.642
499477 97.963 -61.763 50.600
499478 97.987 -61.779 50.560
499479 98.011 -61.795 50.520
499480 98.035 -61.810 50.479
499481 98.060 -61.826 50.438
499482 98.084 -61.841 50.397
499483 98.108 -61.857 50.358
499484 98.132 -61.873 50.319
499485 98.156 -61.888 50.281
499486 98.180 -61.904 50.242
499931 88.039 -59.616 55.994
499932 88.088 -59.622 55.964
499933 88.140 -59.629 55.934
499934 88.192 -59.636 55.903
499935 88.245 -59.643 55.873
499936 88.296 -59.650 55.844
499937 88.348 -59.657 55.816
499938 88.401 -59.663 55.787
499939 88.453 -59.670 55.759
499940 88.504 -59.677 55.730
499941 88.557 -59.684 55.702
499942 88.609 -59.692 55.676
499943 88.661 -59.698 55.651
499944 88.713 -59.705 55.626
499945 88.768 -59.703 55.602
499946 88.827 -59.687 55.577
499947 88.885 -59.673 55.554
499948 88.941 -59.662 55.531
499949 88.982 -59.680 55.508
499950 89.022 -59.701 55.487
499951 89.062 -59.720 55.466
499952 89.102 -59.740 55.445
499953 89.142 -59.761 55.424
499954 89.183 -59.780 55.403
499955 89.223 -59.800 55.385
499956 89.257 -59.818 55.366
499957 89.268 -59.826 55.348
499958 89.280 -59.836 55.330
499959 89.318 -59.849 55.314
499960 89.358 -59.864 55.297
499961 89.399 -59.879 55.283
499962 89.440 -59.894 55.269
499963 89.480 -59.908 55.254
499964 89.522 -59.927 55.238
499965 89.564 -59.949 55.224
499966 89.607 -59.971 55.209
499967 89.650 -59.992 55.197
499968 89.693 -60.013 55.186
499969 89.736 -60.035 55.176
499970 89.778 -60.056 55.165
499971 89.821 -60.077 55.156
499972 89.864 -60.099 55.148
499973 89.906 -60.121 55.141
499974 89.949 -60.142 55.133
499975 89.994 -60.162 55.126
499976 90.039 -60.183 55.118
499977 90.085 -60.202 55.110
499978 90.130 -60.223 55.101
499979 90.176 -60.243 55.091
499980 90.221 -60.263 55.081
499981 90.267 -60.284 55.071
499982 90.312 -60.303 55.060
499983 90.358 -60.324 55.048
499984 90.403 -60.343 55.039
499985 90.449 -60.364 55.031
499986 90.494 -60.384 55.023
499987 90.539 -60.404 55.017
499988 90.585 -60.425 55.013
499989 90.629 -60.439 55.009
499990 90.672 -60.449 55.004
499991 90.716 -60.458 55.000
499992 90.759 -60.468 55.000
499993 90.802 -60.478 55.000
499994 90.845 -60.487 55.001
499995 90.886 -60.503 55.004
499996 90.923 -60.533 55.007
499997 90.965 -60.541 55.013
499998 91.008 -60.547 55.022
499999 91.050 -60.554 55.034
500000 91.093 -60.561 55.045
500001 91.135 -60.568 55.052
500002 91.178 -60.575 55.059
500003 91.220 -60.582 55.065
500004 91.262 -60.589 55.068
500005 91.302 -60.598 55.068
500006 91.342 -60.608 55.064
500007 91.381 -60.617 55.054
500008 91.421 -60.627 55.042
500009 91.460 -60.636 55.029
500010 91.500 -60.645 55.016
500011 91.540 -60.655 55.004
500012 91.579 -60.664 54.999
500013 91.619 -60.674 55.000
500014 91.659 -60.683 55.005
500015 91.698 -60.692 55.009
500016 91.738 -60.702 55.010
500017 91.778 -60.711 55.005
500018 91.817 -60.720 54.994
500019 91.839 -60.733 54.975
500020 91.844 -60.749 54.951
500021 91.850 -60.765 54.921
500022 91.883 -60.781 54.888
500023 91.923 -60.796 54.864
500024 91.963 -60.812 54.848
500025 92.003 -60.827 54.832
500026 92.042 -60.842 54.817
500027 92.083 -60.858 54.811
500028 92.123 -60.873 54.804
500029 92.163 -60.889 54.797
500030 92.203 -60.904 54.788
500031 92.242 -60.920 54.774
500032 92.283 -60.936 54.756
500033 92.322 -60.951 54.747
500034 92.362 -60.966 54.752
500035 92.403 -60.982 54.755
500036 92.441 -60.990 54.751
500037 92.481 -60.998 54.742
500038 92.523 -61.008 54.726
500039 92.568 -61.019 54.706
500040 92.606 -61.025 54.687
500041 92.645 -61.030 54.669
500042 92.683 -61.036 54.652
500043 92.721 -61.042 54.629
500044 92.762 -61.048 54.605
500045 92.803 -61.053 54.576
500046 92.844 -61.058 54.541
500047 92.884 -61.064 54.507
500048 92.925 -61.069 54.501
500049 92.966 -61.074 54.500
500050 93.006 -61.080 54.494
500051 93.047 -61.085 54.482
500052 93.088 -61.090 54.465
500053 93.129 -61.096 54.443
500054 93.169 -61.101 54.428
500055 93.210 -61.106 54.419
500056 93.251 -61.111 54.412
500057 93.291 -61.117 54.405
500058 93.332 -61.122 54.395
500059 93.373 -61.127 54.380
500060 93.414 -61.132 54.370
500061 93.454 -61.137 54.362
500062 93.494 -61.146 54.353
500063 93.534 -61.157 54.342
500064 93.573 -61.168 54.331
500065 93.613 -61.180 54.318
500066 93.652 -61.190 54.303
500067 93.692 -61.201 54.288
500068 93.731 -61.211 54.280
500069 93.771 -61.221 54.265
500070 93.812 -61.231 54.244
500071 93.854 -61.241 54.219
500072 93.898 -61.252 54.191
500073 93.940 -61.259 54.166
500074 93.982 -61.266 54.154
500075 94.025 -61.273 54.135
500076 94.067 -61.279 54.111
500077 94.109 -61.286 54.085
500078 94.151 -61.293 54.060
500079 94.193 -61.299 54.037
500080 94.236 -61.306 54.016
500081 94.277 -61.312 53.995
500082 94.320 -61.318 53.976
500083 94.362 -61.327 53.956
500084 94.404 -61.335 53.934
500085 94.445 -61.342 53.910
500086 94.485 -61.346 53.884
500087 94.523 -61.349 53.856
500088 94.562 -61.353 53.826
500089 94.600 -61.356 53.794
500090 94.638 -61.360 53.763
500091 94.677 -61.364 53.727
500092 94.715 -61.368 53.692
500093 94.748 -61.368 53.657
500094 94.782 -61.374 53.622
500095 94.818 -61.381 53.588
500096 94.854 -61.388 53.555
500097 94.891 -61.396 53.520
500098 94.927 -61.403 53.485
500099 94.963 -61.410 53.451
500100 94.999 -61.418 53.418
500101 95.035 -61.425 53.387
500102 95.071 -61.431 53.357
500103 95.109 -61.434 53.327
500104 95.146 -61.439 53.295
500105 95.184 -61.444 53.263
500106 95.222 -61.449 53.231
500107 95.260 -61.454 53.200
500108 95.298 -61.459 53.170
500109 95.336 -61.465 53.142
500110 95.374 -61.469 53.114
500111 95.411 -61.475 53.090
500112 95.449 -61.482 53.067
500113 95.485 -61.487 53.043
500114 95.523 -61.490 53.016
500115 95.560 -61.494 52.986
500116 95.598 -61.497 52.953
500117 95.636 -61.501 52.920
500118 95.674 -61.506 52.889
500119 95.712 -61.509 52.858
500120 95.750 -61.514 52.823
500121 95.788 -61.518 52.787
500122 95.826 -61.522 52.751
500123 95.865 -61.525 52.716
500124 95.904 -61.529 52.682
500125 95.947 -61.531 52.648
500126 95.990 -61.534 52.611
500127 96.030 -61.538 52.573
500128 96.071 -61.541 52.533
500129 96.111 -61.544 52.494
500130 96.151 -61.547 52.455
500131 96.191 -61.550 52.417
500132 96.232 -61.553 52.380
500133 96.272 -61.556 52.342
500134 96.309 -61.558 52.302
500135 96.345 -61.560 52.260
500136 96.376 -61.572 52.216
500137 96.407 -61.583 52.171
500138 96.438 -61.594 52.129
500139 96.469 -61.606 52.089
500140 96.506 -61.613 52.048
500141 96.547 -61.617 52.007
500142 96.588 -61.621 51.967
500143 96.625 -61.627 51.928
500144 96.662 -61.633 51.890
500145 96.698 -61.640 51.852
500146 96.745 -61.639 51.816
500147 96.792 -61.638 51.781
500148 96.838 -61.638 51.746
500149 96.878 -61.635 51.713
500150 96.918 -61.633 51.680
500151 96.958 -61.630 51.649
500152 96.998 -61.628 51.616
500153 97.037 -61.625 51.582
500154 97.077 -61.622 51.548
500155 97.117 -61.620 51.514
500156 97.148 -61.625 51.481
500157 97.179 -61.629 51.448
500158 97.210 -61.634 51.415
500159 97.240 -61.639 51.380
500160 97.272 -61.644 51.345
500161 97.325 -61.638 51.308
500162 97.379 -61.633 51.269
500163 97.433 -61.628 51.229
500164 97.487 -61.622 51.188
500165 97.527 -61.628 51.147
500166 97.566 -61.633 51.107
500167 97.605 -61.639 51.066
500168 97.644 -61.645 51.024
500169 97.683 -61.651 50.982
500170 97.726 -61.645 50.939
500171 97.770 -61.638 50.896
500172 97.819 -61.643 50.853
500173 97.844 -61.650 50.811
500174 97.868 -61.666 50.767
500175 97.892 -61.682 50.724
500176 97.916 -61.697 50.682
500177 97.940 -61.713 50.640
500178 97.964 -61.729 50.600
500179 97.989 -61.744 50.560
500180 98.013 -61.760 50.519
500181 98.037 -61.776 50.479
500182 98.061 -61.791 50.439
500183 98.085 -61.807 50.398
500184 98.109 -61.822 50.360
500185 98.134 -61.838 50.321
500186 98.158 -61.854 50.283
500187 98.182 -61.869 50.244
500631 88.017 -59.566 56.034
500632 88.063 -59.571 56.004
500633 88.115 -59.578 55.974
500634 88.167 -59.585 55.944
500635 88.219 -59.592 55.915
500636 88.271 -59.599 55.885
500637 88.323 -59.606 55.857
500638 88.375 -59.613 55.829
500639 88.427 -59.620 55.801
500640 88.479 -59.626 55.772
500641 88.531 -59.633 55.743
500642 88.583 -59.640 55.715
500643 88.635 -59.647 55.689
500644 88.687 -59.654 55.664
500645 88.740 -59.661 55.638
500646 88.796 -59.654 55.613
500647 88.855 -59.639 55.590
500648 88.912 -59.627 55.566
500649 88.962 -59.628 55.543
500650 89.001 -59.648 55.520
500651 89.042 -59.668 55.499
500652 89.082 -59.689 55.478
500653 89.122 -59.708 55.457
500654 89.162 -59.728 55.435
500655 89.202 -59.748 55.415
500656 89.242 -59.768 55.395
500657 89.278 -59.786 55.377
500658 89.290 -59.795 55.360
500659 89.302 -59.804 55.343
500660 89.333 -59.817 55.326
500661 89.374 -59.832 55.311
500662 89.414 -59.846 55.297
500663 89.455 -59.861 55.284
500664 89.495 -59.876 55.269
500665 89.536 -59.891 55.255
500666 89.578 -59.912 55.241
500667 89.621 -59.933 55.228
500668 89.664 -59.955 55.216
500669 89.706 -59.977 55.205
500670 89.749 -59.998 55.195
500671 89.792 -60.020 55.185
500672 89.834 -60.041 55.177
500673 89.877 -60.062 55.169
500674 89.920 -60.083 55.161
500675 89.963 -60.105 55.153
500676 90.009 -60.125 55.145
500677 90.054 -60.145 55.136
500678 90.100 -60.165 55.127
500679 90.146 -60.185 55.118
500680 90.191 -60.206 55.108
500681 90.236 -60.226 55.098
500682 90.282 -60.246 55.087
500683 90.327 -60.266 55.076
500684 90.373 -60.286 55.064
500685 90.418 -60.306 55.053
500686 90.464 -60.326 55.044
500687 90.509 -60.347 55.034
500688 90.554 -60.363 55.027
500689 90.599 -60.375 55.022
500690 90.643 -60.387 55.017
500691 90.686 -60.397 55.012
500692 90.729 -60.406 55.009
500693 90.773 -60.416 55.011
500694 90.815 -60.425 55.012
500695 90.859 -60.434 55.013
500696 90.899 -60.453 55.014
500697 90.935 -60.487 55.015
500698 90.976 -60.499 55.016
500699 91.019 -60.506 55.022
500700 91.061 -60.513 55.031
500701 91.103 -60.519 55.040
500702 91.146 -60.527 55.049
500703 91.188 -60.534 55.057
500704 91.231 -60.541 55.061
500705 91.273 -60.547 55.061
500706 91.316 -60.554 55.059
500707 91.358 -60.561 55.055
500708 91.400 -60.569 55.045
500709 91.441 -60.577 55.035
500710 91.481 -60.587 55.023
500711 91.520 -60.596 55.011
500712 91.560 -60.605 54.998
500713 91.599 -60.615 54.989
500714 91.639 -60.624 54.987
500715 91.678 -60.634 54.990
500716 91.718 -60.643 54.995
500717 91.758 -60.652 54.998
500718 91.797 -60.662 54.996
500719 91.837 -60.671 54.988
500720 91.847 -60.686 54.973
500721 91.853 -60.702 54.951
500722 91.858 -60.718 54.922
500723 91.889 -60.734 54.889
500724 91.929 -60.749 54.867
500725 91.969 -60.764 54.851
500726 92.009 -60.780 54.837
500727 92.050 -60.795 54.820
500728 92.089 -60.810 54.806
500729 92.129 -60.826 54.801
500730 92.169 -60.842 54.797
500731 92.209 -60.857 54.791
500732 92.249 -60.873 54.781
500733 92.289 -60.888 54.765
500734 92.329 -60.904 54.746
500735 92.369 -60.919 54.737
500736 92.409 -60.935 54.734
500737 92.449 -60.950 54.732
500738 92.489 -60.965 54.722
500739 92.528 -60.974 54.708
500740 92.571 -60.983 54.689
500741 92.609 -60.989 54.669
500742 92.649 -60.995 54.649
500743 92.690 -61.000 54.634
500744 92.730 -61.005 54.619
500745 92.771 -61.011 54.597
500746 92.812 -61.016 54.570
500747 92.853 -61.021 54.537
500748 92.894 -61.027 54.505
500749 92.934 -61.032 54.503
500750 92.975 -61.037 54.504
500751 93.015 -61.042 54.498
500752 93.056 -61.048 54.484
500753 93.097 -61.052 54.467
500754 93.138 -61.058 54.451
500755 93.179 -61.063 54.444
500756 93.219 -61.068 54.437
500757 93.260 -61.074 54.430
500758 93.300 -61.079 54.423
500759 93.341 -61.084 54.414
500760 93.382 -61.089 54.401
500761 93.423 -61.095 54.391
500762 93.464 -61.100 54.382
500763 93.503 -61.109 54.372
500764 93.543 -61.120 54.359
500765 93.582 -61.131 54.345
500766 93.622 -61.142 54.328
500767 93.661 -61.152 54.312
500768 93.701 -61.162 54.298
500769 93.741 -61.172 54.285
500770 93.780 -61.183 54.265
500771 93.822 -61.193 54.241
500772 93.864 -61.203 54.216
500773 93.908 -61.213 54.194
500774 93.951 -61.223 54.179
500775 93.993 -61.230 54.171
500776 94.035 -61.237 54.156
500777 94.077 -61.243 54.135
500778 94.119 -61.250 54.109
500779 94.161 -61.256 54.082
500780 94.204 -61.263 54.057
500781 94.246 -61.270 54.033
500782 94.288 -61.276 54.010
500783 94.330 -61.284 53.987
500784 94.372 -61.293 53.965
500785 94.415 -61.302 53.940
500786 94.457 -61.309 53.915
500787 94.498 -61.315 53.888
500788 94.539 -61.321 53.859
500789 94.577 -61.324 53.827
500790 94.615 -61.328 53.794
500791 94.654 -61.331 53.763
500792 94.692 -61.335 53.726
500793 94.731 -61.339 53.690
500794 94.759 -61.337 53.656
500795 94.789 -61.340 53.622
500796 94.826 -61.346 53.590
500797 94.862 -61.354 53.557
500798 94.898 -61.361 53.521
500799 94.934 -61.368 53.484
500800 94.970 -61.376 53.448
500801 95.006 -61.383 53.413
500802 95.043 -61.390 53.381
500803 95.080 -61.393 53.349
500804 95.117 -61.395 53.316
500805 95.155 -61.400 53.282
500806 95.193 -61.405 53.248
500807 95.231 -61.409 53.215
500808 95.268 -61.415 53.183
500809 95.306 -61.421 53.152
500810 95.344 -61.427 53.123
500811 95.382 -61.434 53.092
500812 95.419 -61.440 53.065
500813 95.456 -61.446 53.039
500814 95.491 -61.450 53.011
500815 95.529 -61.454 52.983
500816 95.566 -61.458 52.954
500817 95.604 -61.461 52.922
500818 95.641 -61.465 52.892
500819 95.679 -61.468 52.862
500820 95.717 -61.472 52.833
500821 95.755 -61.475 52.802
500822 95.793 -61.480 52.768
500823 95.831 -61.484 52.733
500824 95.869 -61.487 52.700
500825 95.909 -61.491 52.667
500826 95.952 -61.494 52.633
500827 95.992 -61.497 52.597
500828 96.032 -61.500 52.559
500829 96.073 -61.503 52.518
500830 96.113 -61.506 52.478
500831 96.153 -61.509 52.441
500832 96.193 -61.512 52.405
500833 96.234 -61.515 52.372
500834 96.274 -61.518 52.336
500835 96.314 -61.521 52.298
500836 96.351 -61.523 52.257
500837 96.383 -61.532 52.214
500838 96.414 -61.544 52.171
500839 96.445 -61.556 52.130
500840 96.476 -61.567 52.089
500841 96.508 -61.578 52.049
500842 96.545 -61.584 52.009
500843 96.586 -61.589 51.967
500844 96.624 -61.594 51.928
500845 96.661 -61.601 51.889
500846 96.704 -61.603 51.851
500847 96.751 -61.602 51.814
500848 96.798 -61.601 51.780
500849 96.842 -61.600 51.745
500850 96.882 -61.597 51.711
500851 96.922 -61.595 51.676
500852 96.962 -61.592 51.642
500853 97.002 -61.590 51.610
500854 97.041 -61.588 51.577
500855 97.081 -61.584 51.542
500856 97.121 -61.582 51.508
500857 97.161 -61.580 51.473
500858 97.191 -61.585 51.441
500859 97.222 -61.590 51.408
500860 97.252 -61.594 51.373
500861 97.304 -61.591 51.338
500862 97.358 -61.585 51.301
500863 97.407 -61.583 51.261
500864 97.447 -61.588 51.220
500865 97.486 -61.594 51.179
500866 97.525 -61.600 51.139
500867 97.565 -61.606 51.099
500868 97.603 -61.613 51.060
500869 97.643 -61.619 51.018
500870 97.685 -61.613 50.975
500871 97.729 -61.606 50.932
500872 97.775 -61.604 50.890
500873 97.823 -61.609 50.847
500874 97.847 -61.616 50.805
500875 97.870 -61.631 50.762
500876 97.894 -61.646 50.720
500877 97.918 -61.662 50.679
500878 97.942 -61.678 50.639
500879 97.966 -61.693 50.599
500880 97.991 -61.709 50.559
500881 98.015 -61.725 50.519
500882 98.039 -61.740 50.480
500883 98.063 -61.756 50.440
500884 98.087 -61.771 50.399
500885 98.111 -61.787 50.361
500886 98.136 -61.803 50.323
500887 98.160 -61.818 50.285
500888 98.184 -61.834 50.247
501331 87.996 -59.516 56.071
501332 88.038 -59.520 56.042
501333 88.090 -59.527 56.013
501334 88.142 -59.534 55.985
501335 88.194 -59.541 55.956
501336 88.246 -59.547 55.928
501337 88.298 -59.555 55.898
501338 88.350 -59.562 55.871
501339 88.402 -59.569 55.843
501340 88.454 -59.576 55.814
501341 88.506 -59.582 55.786
501342 88.558 -59.589 55.757
501343 88.610 -59.596 55.729
501344 88.662 -59.603 55.702
501345 88.714 -59.610 55.677
501346 88.766 -59.617 55.651
501347 88.824 -59.607 55.626
501348 88.883 -59.591 55.602
501349 88.939 -59.580 55.578
501350 88.981 -59.596 55.555
501351 89.021 -59.616 55.533
501352 89.061 -59.636 55.512
501353 89.101 -59.656 55.491
501354 89.141 -59.676 55.470
501355 89.182 -59.696 55.449
501356 89.222 -59.716 55.429
501357 89.262 -59.736 55.410
501358 89.300 -59.755 55.392
501359 89.311 -59.764 55.375
501360 89.323 -59.773 55.358
501361 89.348 -59.784 55.343
501362 89.389 -59.798 55.327
501363 89.429 -59.814 55.313
501364 89.469 -59.828 55.300
501365 89.510 -59.843 55.286
501366 89.550 -59.858 55.273
501367 89.592 -59.876 55.260
501368 89.634 -59.897 55.248
501369 89.677 -59.918 55.237
501370 89.720 -59.939 55.227
501371 89.763 -59.961 55.217
501372 89.805 -59.983 55.208
501373 89.848 -60.004 55.200
501374 89.891 -60.026 55.193
501375 89.934 -60.047 55.184
501376 89.978 -60.068 55.176
501377 90.024 -60.087 55.168
501378 90.070 -60.108 55.159
501379 90.115 -60.128 55.150
501380 90.161 -60.148 55.140
501381 90.206 -60.168 55.129
501382 90.251 -60.188 55.118
501383 90.297 -60.209 55.106
501384 90.342 -60.228 55.094
501385 90.388 -60.249 55.082
501386 90.434 -60.269 55.069
501387 90.478 -60.285 55.058
501388 90.523 -60.297 55.047
501389 90.567 -60.309 55.038
501390 90.611 -60.322 55.031
501391 90.656 -60.334 55.026
501392 90.699 -60.344 55.020
501393 90.743 -60.353 55.017
501394 90.786 -60.363 55.021
501395 90.829 -60.372 55.023
501396 90.872 -60.382 55.025
501397 90.912 -60.403 55.027
501398 90.948 -60.437 55.026
501399 90.987 -60.457 55.024
501400 91.029 -60.464 55.026
501401 91.072 -60.472 55.032
501402 91.114 -60.478 55.040
501403 91.157 -60.485 55.048
501404 91.199 -60.492 55.054
501405 91.242 -60.499 55.058
501406 91.284 -60.506 55.058
501407 91.326 -60.513 55.055
501408 91.369 -60.520 55.048
501409 91.412 -60.527 55.038
501410 91.454 -60.534 55.028
501411 91.496 -60.541 55.018
501412 91.538 -60.547 55.007
501413 91.580 -60.556 54.996
501414 91.619 -60.566 54.985
501415 91.659 -60.575 54.977
501416 91.699 -60.584 54.976
501417 91.738 -60.594 54.981
501418 91.778 -60.603 54.986
501419 91.817 -60.613 54.987
501420 91.850 -60.623 54.982
501421 91.855 -60.639 54.970
501422 91.861 -60.655 54.950
501423 91.867 -60.671 54.922
501424 91.896 -60.686 54.890
501425 91.936 -60.702 54.868
501426 91.976 -60.717 54.854
501427 92.016 -60.732 54.842
501428 92.056 -60.748 54.825
501429 92.096 -60.763 54.805
501430 92.136 -60.779 54.799
501431 92.176 -60.795 54.796
501432 92.216 -60.810 54.792
501433 92.256 -60.826 54.786
501434 92.296 -60.841 54.774
501435 92.336 -60.856 54.753
501436 92.376 -60.872 54.733
501437 92.416 -60.887 54.719
501438 92.456 -60.903 54.711
501439 92.496 -60.918 54.706
501440 92.534 -60.935 54.691
501441 92.576 -60.945 54.674
501442 92.617 -60.951 54.654
501443 92.658 -60.956 54.633
501444 92.699 -60.962 54.614
501445 92.740 -60.967 54.600
501446 92.780 -60.973 54.579
501447 92.821 -60.978 54.550
501448 92.862 -60.983 54.520
501449 92.903 -60.989 54.495
501450 92.943 -60.994 54.500
501451 92.984 -60.999 54.500
501452 93.025 -61.005 54.493
501453 93.065 -61.010 54.481
501454 93.106 -61.015 54.472
501455 93.147 -61.020 54.470
501456 93.188 -61.026 54.464
501457 93.228 -61.031 54.456
501458 93.269 -61.036 54.447
501459 93.310 -61.042 54.439
501460 93.350 -61.047 54.430
501461 93.391 -61.052 54.417
501462 93.432 -61.058 54.408
501463 93.473 -61.063 54.398
501464 93.512 -61.071 54.386
501465 93.552 -61.083 54.371
501466 93.591 -61.093 54.353
501467 93.631 -61.103 54.334
501468 93.671 -61.114 54.317
501469 93.710 -61.124 54.301
501470 93.750 -61.134 54.283
501471 93.789 -61.144 54.259
501472 93.831 -61.154 54.235
501473 93.874 -61.165 54.215
501474 93.917 -61.174 54.201
501475 93.961 -61.185 54.194
501476 94.003 -61.194 54.187
501477 94.046 -61.201 54.175
501478 94.088 -61.208 54.155
501479 94.130 -61.214 54.130
501480 94.172 -61.221 54.104
501481 94.214 -61.227 54.079
501482 94.257 -61.233 54.051
501483 94.298 -61.242 54.022
501484 94.341 -61.250 53.994
501485 94.384 -61.259 53.966
501486 94.426 -61.268 53.939
501487 94.468 -61.276 53.912
501488 94.509 -61.282 53.884
501489 94.551 -61.289 53.855
501490 94.592 -61.296 53.825
501491 94.631 -61.299 53.794
501492 94.669 -61.302 53.761
501493 94.707 -61.306 53.723
501494 94.744 -61.309 53.687
501495 94.770 -61.307 53.653
501496 94.797 -61.306 53.621
501497 94.833 -61.312 53.588
501498 94.869 -61.319 53.554
501499 94.905 -61.327 53.516
501500 94.942 -61.334 53.478
501501 94.978 -61.341 53.441
501502 95.014 -61.349 53.406
501503 95.051 -61.351 53.372
501504 95.089 -61.354 53.338
501505 95.126 -61.356 53.304
501506 95.164 -61.360 53.268
501507 95.201 -61.365 53.233
501508 95.239 -61.372 53.200
501509 95.276 -61.378 53.167
501510 95.314 -61.385 53.136
501511 95.352 -61.392 53.106
501512 95.389 -61.398 53.074
501513 95.427 -61.405 53.043
501514 95.463 -61.409 53.013
501515 95.498 -61.414 52.983
501516 95.534 -61.418 52.954
501517 95.572 -61.422 52.925
501518 95.610 -61.425 52.894
501519 95.647 -61.428 52.865
501520 95.685 -61.432 52.836
501521 95.723 -61.435 52.807
501522 95.760 -61.439 52.779
501523 95.798 -61.443 52.748
501524 95.835 -61.446 52.715
501525 95.873 -61.450 52.683
501526 95.913 -61.453 52.651
501527 95.953 -61.456 52.616
501528 95.994 -61.459 52.579
501529 96.034 -61.462 52.539
501530 96.075 -61.465 52.499
501531 96.115 -61.468 52.459
501532 96.155 -61.471 52.422
501533 96.195 -61.474 52.389
501534 96.236 -61.477 52.357
501535 96.276 -61.480 52.324
501536 96.316 -61.483 52.288
501537 96.357 -61.486 52.249
501538 96.391 -61.493 52.208
501539 96.421 -61.505 52.166
501540 96.453 -61.516 52.126
501541 96.484 -61.528 52.087
501542 96.515 -61.539 52.047
501543 96.546 -61.550 52.008
501544 96.584 -61.556 51.969
501545 96.623 -61.562 51.930
501546 96.663 -61.566 51.891
501547 96.710 -61.565 51.852
501548 96.757 -61.564 51.814
501549 96.804 -61.563 51.779
501550 96.847 -61.562 51.743
501551 96.886 -61.559 51.708
501552 96.926 -61.557 51.673
501553 96.966 -61.554 51.637
501554 97.006 -61.552 51.603
501555 97.045 -61.550 51.568
501556 97.085 -61.547 51.533
501557 97.125 -61.544 51.499
501558 97.165 -61.542 51.466
501559 97.204 -61.541 51.434
501560 97.234 -61.545 51.400
501561 97.283 -61.542 51.364
501562 97.327 -61.544 51.327
501563 97.367 -61.550 51.289
501564 97.406 -61.556 51.249
501565 97.445 -61.562 51.207
501566 97.485 -61.568 51.167
501567 97.524 -61.574 51.128
501568 97.563 -61.580 51.089
501569 97.602 -61.585 51.051
501570 97.644 -61.581 51.010
501571 97.687 -61.574 50.968
501572 97.731 -61.566 50.925
501573 97.779 -61.569 50.882
501574 97.828 -61.575 50.840
501575 97.851 -61.581 50.799
501576 97.871 -61.595 50.758
501577 97.895 -61.611 50.717
501578 97.920 -61.627 50.677
501579 97.944 -61.642 50.636
501580 97.968 -61.658 50.597
501581 97.992 -61.673 50.558
501582 98.016 -61.689 50.519
501583 98.040 -61.705 50.480
501584 98.065 -61.720 50.441
501585 98.089 -61.736 50.401
501586 98.113 -61.751 50.363
501587 98.137 -61.767 50.325
501588 98.161 -61.783 50.288
502032 88.012 -59.469 56.079
502033 88.065 -59.476 56.051
502034 88.117 -59.483 56.023
502035 88.169 -59.490 55.995
502036 88.221 -59.497 55.968
502037 88.273 -59.504 55.940
502038 88.325 -59.510 55.911
502039 88.377 -59.518 55.884
502040 88.429 -59.525 55.857
502041 88.481 -59.532 55.830
502042 88.533 -59.538 55.801
502043 88.585 -59.545 55.773
502044 88.637 -59.552 55.744
502045 88.689 -59.559 55.716
502046 88.741 -59.566 55.690
502047 88.793 -59.573 55.664
502048 88.852 -59.558 55.639
502049 88.910 -59.544 55.614
502050 88.961 -59.544 55.591
502051 89.000 -59.564 55.568
502052 89.041 -59.584 55.546
502053 89.081 -59.604 55.526
502054 89.121 -59.624 55.505
502055 89.161 -59.644 55.485
502056 89.201 -59.664 55.465
502057 89.241 -59.684 55.445
502058 89.281 -59.704 55.426
502059 89.321 -59.723 55.408
502060 89.333 -59.732 55.392
502061 89.344 -59.741 55.376
502062 89.363 -59.751 55.361
502063 89.403 -59.766 55.346
502064 89.444 -59.781 55.333
502065 89.484 -59.795 55.321
502066 89.525 -59.811 55.308
502067 89.565 -59.825 55.296
502068 89.606 -59.840 55.283
502069 89.648 -59.860 55.272
502070 89.691 -59.882 55.261
502071 89.734 -59.903 55.252
502072 89.776 -59.924 55.243
502073 89.819 -59.946 55.234
502074 89.862 -59.967 55.226
502075 89.904 -59.989 55.219
502076 89.948 -60.010 55.211
502077 89.993 -60.030 55.202
502078 90.039 -60.050 55.193
502079 90.085 -60.071 55.184
502080 90.130 -60.090 55.175
502081 90.176 -60.111 55.164
502082 90.221 -60.130 55.153
502083 90.266 -60.151 55.140
502084 90.312 -60.171 55.127
502085 90.358 -60.191 55.114
502086 90.402 -60.207 55.100
502087 90.447 -60.219 55.087
502088 90.491 -60.231 55.073
502089 90.535 -60.243 55.062
502090 90.580 -60.256 55.051
502091 90.624 -60.267 55.042
502092 90.669 -60.279 55.035
502093 90.713 -60.291 55.030
502094 90.756 -60.301 55.025
502095 90.799 -60.310 55.026
502096 90.842 -60.320 55.029
502097 90.886 -60.330 55.034
502098 90.925 -60.354 55.038
502099 90.961 -60.387 55.039
502100 90.998 -60.416 55.036
502101 91.040 -60.423 55.035
502102 91.083 -60.430 55.037
502103 91.125 -60.437 55.044
502104 91.167 -60.444 55.049
502105 91.210 -60.450 55.052
502106 91.252 -60.458 55.055
502107 91.295 -60.465 55.055
502108 91.337 -60.472 55.052
502109 91.380 -60.478 55.045
502110 91.422 -60.485 55.036
502111 91.464 -60.492 55.027
502112 91.507 -60.499 55.017
502113 91.549 -60.506 55.006
502114 91.592 -60.513 54.996
502115 91.634 -60.520 54.982
502116 91.677 -60.527 54.970
502117 91.718 -60.535 54.964
502118 91.758 -60.544 54.965
502119 91.798 -60.553 54.971
502120 91.837 -60.563 54.974
502121 91.858 -60.576 54.971
502122 91.864 -60.591 54.962
502123 91.869 -60.607 54.946
502124 91.875 -60.623 54.922
502125 91.903 -60.639 54.891
502126 91.943 -60.654 54.868
502127 91.983 -60.670 54.854
502128 92.023 -60.685 54.843
502129 92.063 -60.701 54.829
502130 92.102 -60.716 54.810
502131 92.143 -60.732 54.797
502132 92.182 -60.748 54.793
502133 92.223 -60.763 54.791
502134 92.263 -60.778 54.787
502135 92.302 -60.794 54.780
502136 92.343 -60.809 54.762
502137 92.382 -60.825 54.736
502138 92.423 -60.840 54.711
502139 92.458 -60.858 54.694
502140 92.493 -60.877 54.688
502141 92.536 -60.889 54.674
502142 92.581 -60.898 54.659
502143 92.624 -60.905 54.641
502144 92.664 -60.911 54.621
502145 92.705 -60.917 54.600
502146 92.746 -60.923 54.582
502147 92.787 -60.929 54.559
502148 92.828 -60.935 54.532
502149 92.869 -60.940 54.505
502150 92.910 -60.946 54.489
502151 92.951 -60.952 54.497
502152 92.992 -60.958 54.499
502153 93.032 -60.964 54.495
502154 93.074 -60.970 54.489
502155 93.115 -60.976 54.485
502156 93.155 -60.981 54.479
502157 93.196 -60.987 54.472
502158 93.237 -60.993 54.466
502159 93.278 -60.998 54.459
502160 93.319 -61.004 54.449
502161 93.360 -61.009 54.436
502162 93.400 -61.014 54.424
502163 93.441 -61.020 54.415
502164 93.482 -61.025 54.405
502165 93.521 -61.034 54.390
502166 93.561 -61.045 54.372
502167 93.601 -61.055 54.353
502168 93.640 -61.064 54.334
502169 93.680 -61.075 54.316
502170 93.720 -61.085 54.297
502171 93.759 -61.096 54.273
502172 93.799 -61.105 54.248
502173 93.841 -61.116 54.227
502174 93.884 -61.126 54.215
502175 93.927 -61.136 54.210
502176 93.970 -61.146 54.207
502177 94.013 -61.156 54.201
502178 94.056 -61.165 54.190
502179 94.098 -61.171 54.172
502180 94.140 -61.178 54.149
502181 94.183 -61.185 54.124
502182 94.225 -61.191 54.097
502183 94.267 -61.199 54.067
502184 94.310 -61.208 54.035
502185 94.352 -61.217 54.002
502186 94.394 -61.226 53.969
502187 94.436 -61.235 53.939
502188 94.479 -61.243 53.908
502189 94.520 -61.249 53.879
502190 94.561 -61.255 53.850
502191 94.603 -61.262 53.823
502192 94.644 -61.269 53.793
502193 94.684 -61.274 53.758
502194 94.723 -61.277 53.721
502195 94.755 -61.278 53.684
502196 94.782 -61.276 53.649
502197 94.809 -61.274 53.615
502198 94.840 -61.277 53.579
502199 94.877 -61.285 53.544
502200 94.913 -61.293 53.506
502201 94.949 -61.299 53.468
502202 94.985 -61.307 53.431
502203 95.022 -61.309 53.396
502204 95.060 -61.312 53.361
502205 95.097 -61.315 53.327
502206 95.134 -61.318 53.292
502207 95.171 -61.324 53.255
502208 95.209 -61.330 53.220
502209 95.247 -61.337 53.186
502210 95.284 -61.343 53.153
502211 95.322 -61.349 53.122
502212 95.359 -61.356 53.091
502213 95.397 -61.363 53.058
502214 95.434 -61.368 53.026
502215 95.469 -61.373 52.994
502216 95.505 -61.378 52.964
502217 95.540 -61.382 52.933
502218 95.578 -61.386 52.903
502219 95.615 -61.389 52.871
502220 95.653 -61.393 52.841
502221 95.691 -61.396 52.810
502222 95.728 -61.400 52.781
502223 95.766 -61.403 52.753
502224 95.804 -61.406 52.726
502225 95.841 -61.410 52.696
502226 95.879 -61.414 52.665
502227 95.917 -61.418 52.631
502228 95.957 -61.421 52.596
502229 95.997 -61.424 52.557
502230 96.038 -61.427 52.517
502231 96.078 -61.430 52.476
502232 96.118 -61.432 52.437
502233 96.159 -61.435 52.401
502234 96.199 -61.438 52.369
502235 96.239 -61.441 52.338
502236 96.279 -61.444 52.306
502237 96.320 -61.447 52.274
502238 96.360 -61.450 52.238
502239 96.400 -61.453 52.199
502240 96.431 -61.463 52.159
502241 96.461 -61.476 52.120
502242 96.491 -61.488 52.081
502243 96.522 -61.500 52.043
502244 96.553 -61.512 52.005
502245 96.584 -61.523 51.968
502246 96.622 -61.529 51.930
502247 96.669 -61.528 51.891
502248 96.716 -61.528 51.853
502249 96.762 -61.527 51.816
502250 96.810 -61.526 51.780
502251 96.851 -61.524 51.744
502252 96.890 -61.522 51.708
502253 96.930 -61.519 51.672
502254 96.970 -61.516 51.636
502255 97.010 -61.514 51.600
502256 97.049 -61.512 51.564
502257 97.089 -61.509 51.528
502258 97.129 -61.506 51.493
502259 97.169 -61.505 51.460
502260 97.208 -61.503 51.427
502261 97.247 -61.505 51.392
502262 97.287 -61.511 51.355
502263 97.326 -61.517 51.317
502264 97.365 -61.523 51.278
502265 97.404 -61.529 51.237
502266 97.444 -61.535 51.196
502267 97.483 -61.541 51.155
502268 97.522 -61.547 51.116
502269 97.562 -61.553 51.078
502270 97.603 -61.550 51.039
502271 97.647 -61.542 50.999
502272 97.690 -61.534 50.958
502273 97.735 -61.531 50.917
502274 97.784 -61.535 50.875
502275 97.832 -61.540 50.835
502276 97.855 -61.547 50.795
502277 97.873 -61.560 50.755
502278 97.897 -61.575 50.715
502279 97.922 -61.591 50.675
502280 97.946 -61.607 50.635
502281 97.970 -61.622 50.596
502282 97.994 -61.638 50.558
502283 98.018 -61.653 50.520
502284 98.042 -61.669 50.481
502285 98.067 -61.685 50.442
502286 98.091 -61.700 50.403
502287 98.115 -61.716 50.367
502288 98.139 -61.732 50.330
502733 88.039 -59.425 56.086
502734 88.092 -59.432 56.060
502735 88.143 -59.439 56.033
502736 88.195 -59.446 56.006
502737 88.248 -59.453 55.979
502738 88.299 -59.460 55.951
502739 88.351 -59.466 55.925
502740 88.404 -59.473 55.898
502741 88.456 -59.481 55.872
502742 88.507 -59.488 55.846
502743 88.560 -59.494 55.817
502744 88.612 -59.501 55.789
502745 88.664 -59.508 55.759
502746 88.716 -59.515 55.731
502747 88.768 -59.522 55.704
502748 88.821 -59.526 55.678
502749 88.880 -59.510 55.653
502750 88.937 -59.498 55.629
502751 88.980 -59.512 55.605
502752 89.020 -59.532 55.583
502753 89.060 -59.552 55.562
502754 89.100 -59.572 55.541
502755 89.140 -59.591 55.521
502756 89.181 -59.612 55.501
502757 89.221 -59.632 55.482
502758 89.261 -59.652 55.462
502759 89.301 -59.672 55.444
502760 89.341 -59.692 55.427
502761 89.354 -59.701 55.410
502762 89.366 -59.710 55.396
502763 89.378 -59.719 55.381
502764 89.418 -59.733 55.368
502765 89.459 -59.748 55.356
502766 89.499 -59.763 55.345
502767 89.540 -59.777 55.334
502768 89.581 -59.792 55.323
502769 89.621 -59.807 55.311
502770 89.662 -59.823 55.300
502771 89.704 -59.845 55.290
502772 89.747 -59.867 55.281
502773 89.790 -59.888 55.273
502774 89.833 -59.909 55.264
502775 89.875 -59.931 55.257
502776 89.918 -59.952 55.249
502777 89.963 -59.973 55.241
502778 90.009 -59.992 55.232
502779 90.054 -60.013 55.222
502780 90.100 -60.033 55.213
502781 90.145 -60.053 55.202
502782 90.191 -60.073 55.190
502783 90.236 -60.093 55.178
502784 90.282 -60.114 55.165
502785 90.327 -60.129 55.150
502786 90.371 -60.141 55.136
502787 90.415 -60.153 55.121
502788 90.460 -60.165 55.105
502789 90.504 -60.177 55.091
502790 90.548 -60.190 55.077
502791 90.593 -60.201 55.064
502792 90.637 -60.213 55.054
502793 90.681 -60.225 55.046
502794 90.726 -60.237 55.038
502795 90.770 -60.249 55.033
502796 90.813 -60.258 55.032
502797 90.856 -60.268 55.033
502798 90.899 -60.277 55.037
502799 90.937 -60.304 55.041
502800 90.973 -60.337 55.042
502801 91.010 -60.370 55.041
502802 91.051 -60.381 55.039
502803 91.093 -60.388 55.039
502804 91.136 -60.395 55.042
502805 91.178 -60.403 55.044
502806 91.221 -60.409 55.047
502807 91.263 -60.416 55.049
502808 91.306 -60.423 55.049
502809 91.348 -60.430 55.048
502810 91.390 -60.437 55.045
502811 91.433 -60.444 55.038
502812 91.475 -60.451 55.030
502813 91.518 -60.458 55.019
502814 91.560 -60.465 55.007
502815 91.602 -60.472 54.995
502816 91.645 -60.478 54.980
502817 91.687 -60.485 54.965
502818 91.730 -60.493 54.953
502819 91.773 -60.500 54.949
502820 91.815 -60.506 54.956
502821 91.857 -60.513 54.960
502822 91.866 -60.528 54.958
502823 91.872 -60.544 54.950
502824 91.878 -60.560 54.936
502825 91.883 -60.576 54.915
502826 91.910 -60.592 54.887
502827 91.949 -60.607 54.863
502828 91.990 -60.622 54.847
502829 92.029 -60.638 54.837
502830 92.069 -60.654 54.826
502831 92.109 -60.669 54.810
502832 92.149 -60.685 54.795
502833 92.189 -60.700 54.786
502834 92.229 -60.716 54.782
502835 92.269 -60.731 54.780
502836 92.309 -60.746 54.774
502837 92.348 -60.763 54.762
502838 92.382 -60.782 54.742
502839 92.418 -60.801 54.719
502840 92.453 -60.820 54.693
502841 92.496 -60.832 54.669
502842 92.541 -60.841 54.656
502843 92.585 -60.851 54.644
502844 92.630 -60.860 54.629
502845 92.671 -60.866 54.611
502846 92.712 -60.872 54.590
502847 92.753 -60.878 54.567
502848 92.794 -60.883 54.541
502849 92.835 -60.889 54.515
502850 92.875 -60.895 54.493
502851 92.916 -60.901 54.486
502852 92.957 -60.907 54.495
502853 92.998 -60.913 54.498
502854 93.039 -60.919 54.498
502855 93.080 -60.924 54.496
502856 93.121 -60.930 54.493
502857 93.162 -60.936 54.484
502858 93.203 -60.942 54.474
502859 93.243 -60.948 54.466
502860 93.285 -60.954 54.457
502861 93.326 -60.960 54.447
502862 93.366 -60.965 54.435
502863 93.407 -60.971 54.423
502864 93.448 -60.977 54.414
502865 93.489 -60.983 54.403
502866 93.531 -60.995 54.389
502867 93.571 -61.006 54.370
502868 93.610 -61.016 54.347
502869 93.650 -61.027 54.326
502870 93.690 -61.036 54.307
502871 93.729 -61.047 54.287
502872 93.769 -61.057 54.265
502873 93.808 -61.067 54.244
502874 93.851 -61.077 54.226
502875 93.894 -61.087 54.216
502876 93.937 -61.098 54.214
502877 93.980 -61.108 54.212
502878 94.024 -61.118 54.208
502879 94.066 -61.128 54.199
502880 94.109 -61.136 54.185
502881 94.151 -61.143 54.165
502882 94.193 -61.149 54.140
502883 94.236 -61.156 54.111
502884 94.278 -61.165 54.080
502885 94.320 -61.174 54.047
502886 94.362 -61.183 54.012
502887 94.405 -61.193 53.975
502888 94.448 -61.202 53.941
502889 94.489 -61.209 53.906
502890 94.531 -61.216 53.874
502891 94.572 -61.223 53.846
502892 94.614 -61.229 53.819
502893 94.655 -61.236 53.789
502894 94.697 -61.243 53.754
502895 94.738 -61.249 53.717
502896 94.766 -61.247 53.679
502897 94.793 -61.246 53.643
502898 94.820 -61.244 53.609
502899 94.848 -61.243 53.571
502900 94.884 -61.251 53.535
502901 94.920 -61.258 53.498
502902 94.956 -61.265 53.460
502903 94.993 -61.268 53.424
502904 95.031 -61.271 53.387
502905 95.067 -61.274 53.351
502906 95.104 -61.279 53.317
502907 95.140 -61.283 53.282
502908 95.178 -61.288 53.245
502909 95.217 -61.294 53.211
502910 95.254 -61.301 53.177
502911 95.292 -61.308 53.144
502912 95.330 -61.314 53.113
502913 95.368 -61.321 53.081
502914 95.405 -61.328 53.048
502915 95.441 -61.332 53.015
502916 95.476 -61.337 52.983
502917 95.512 -61.341 52.951
502918 95.547 -61.346 52.920
502919 95.584 -61.349 52.889
502920 95.621 -61.353 52.857
502921 95.659 -61.357 52.825
502922 95.696 -61.360 52.794
502923 95.734 -61.364 52.763
502924 95.772 -61.367 52.734
502925 95.810 -61.371 52.706
502926 95.847 -61.375 52.675
502927 95.884 -61.380 52.642
502928 95.922 -61.384 52.606
502929 95.962 -61.387 52.569
502930 96.003 -61.390 52.530
502931 96.043 -61.393 52.492
502932 96.083 -61.395 52.453
502933 96.123 -61.398 52.415
502934 96.164 -61.401 52.379
502935 96.204 -61.404 52.345
502936 96.244 -61.407 52.314
502937 96.285 -61.410 52.284
502938 96.325 -61.414 52.254
502939 96.365 -61.418 52.222
502940 96.409 -61.416 52.186
502941 96.445 -61.420 52.148
502942 96.477 -61.428 52.110
502943 96.508 -61.440 52.072
502944 96.538 -61.453 52.034
502945 96.568 -61.465 51.998
502946 96.597 -61.478 51.964
502947 96.629 -61.489 51.929
502948 96.676 -61.488 51.892
502949 96.723 -61.488 51.854
502950 96.770 -61.487 51.819
502951 96.814 -61.488 51.783
502952 96.852 -61.493 51.747
502953 96.891 -61.491 51.710
502954 96.931 -61.490 51.674
502955 96.970 -61.488 51.637
502956 97.010 -61.487 51.602
502957 97.049 -61.485 51.564
502958 97.088 -61.484 51.526
502959 97.127 -61.482 51.490
502960 97.167 -61.481 51.454
502961 97.207 -61.479 51.421
502962 97.246 -61.478 51.384
502963 97.285 -61.484 51.346
502964 97.324 -61.491 51.307
502965 97.364 -61.497 51.267
502966 97.403 -61.502 51.226
502967 97.442 -61.508 51.185
502968 97.481 -61.514 51.145
502969 97.521 -61.520 51.106
502970 97.562 -61.518 51.067
502971 97.605 -61.510 51.029
502972 97.649 -61.502 50.990
502973 97.692 -61.494 50.950
502974 97.739 -61.497 50.910
502975 97.788 -61.501 50.870
502976 97.837 -61.506 50.831
502977 97.860 -61.512 50.793
502978 97.875 -61.525 50.754
502979 97.899 -61.541 50.714
502980 97.924 -61.556 50.674
502981 97.948 -61.572 50.635
502982 97.972 -61.588 50.596
502983 97.996 -61.603 50.559
502984 98.020 -61.619 50.521
502985 98.044 -61.635 50.483
502986 98.069 -61.650 50.445
502987 98.093 -61.666 50.406
502988 98.117 -61.681 50.370
503434 88.066 -59.381 56.092
503435 88.118 -59.388 56.067
503436 88.170 -59.395 56.042
503437 88.222 -59.402 56.016
503438 88.274 -59.409 55.991
503439 88.326 -59.416 55.964
503440 88.379 -59.422 55.938
503441 88.430 -59.429 55.913
503442 88.482 -59.436 55.887
503443 88.534 -59.444 55.861
503444 88.587 -59.450 55.833
503445 88.638 -59.457 55.804
503446 88.690 -59.464 55.775
503447 88.743 -59.471 55.746
503448 88.794 -59.478 55.718
503449 88.849 -59.477 55.692
503450 88.908 -59.463 55.667
503451 88.960 -59.460 55.643
503452 88.999 -59.480 55.621
503453 89.040 -59.500 55.599
503454 89.080 -59.519 55.578
503455 89.120 -59.540 55.558
503456 89.160 -59.560 55.537
503457 89.200 -59.580 55.518
503458 89.240 -59.600 55.499
503459 89.280 -59.620 55.480
503460 89.320 -59.640 55.463
503461 89.361 -59.660 55.447
503462 89.376 -59.670 55.432
503463 89.387 -59.679 55.418
503464 89.399 -59.687 55.404
503465 89.434 -59.701 55.392
503466 89.474 -59.715 55.381
503467 89.515 -59.730 55.372
503468 89.555 -59.745 55.363
503469 89.595 -59.760 55.353
503470 89.636 -59.774 55.342
503471 89.676 -59.789 55.331
503472 89.718 -59.808 55.322
503473 89.761 -59.830 55.313
503474 89.804 -59.851 55.306
503475 89.846 -59.873 55.298
503476 89.889 -59.894 55.292
503477 89.933 -59.915 55.284
503478 89.978 -59.935 55.277
503479 90.024 -59.955 55.268
503480 90.070 -59.975 55.258
503481 90.115 -59.996 55.248
503482 90.161 -60.016 55.236
503483 90.206 -60.036 55.224
503484 90.251 -60.051 55.210
503485 90.295 -60.063 55.194
503486 90.339 -60.075 55.178
503487 90.384 -60.087 55.161
503488 90.428 -60.099 55.143
503489 90.472 -60.112 55.126
503490 90.517 -60.123 55.110
503491 90.561 -60.135 55.094
503492 90.605 -60.147 55.080
503493 90.650 -60.159 55.068
503494 90.694 -60.171 55.057
503495 90.739 -60.184 55.047
503496 90.783 -60.195 55.041
503497 90.826 -60.206 55.037
503498 90.870 -60.215 55.037
503499 90.913 -60.224 55.039
503500 90.950 -60.254 55.043
503501 90.986 -60.287 55.043
503502 91.022 -60.321 55.042
503503 91.062 -60.340 55.040
503504 91.104 -60.347 55.039
503505 91.147 -60.354 55.040
503506 91.189 -60.361 55.042
503507 91.231 -60.368 55.045
503508 91.274 -60.375 55.047
503509 91.316 -60.381 55.047
503510 91.359 -60.388 55.045
503511 91.401 -60.396 55.043
503512 91.444 -60.403 55.039
503513 91.486 -60.409 55.033
503514 91.528 -60.416 55.023
503515 91.571 -60.423 55.009
503516 91.613 -60.430 54.993
503517 91.656 -60.437 54.976
503518 91.698 -60.444 54.959
503519 91.741 -60.451 54.944
503520 91.783 -60.458 54.936
503521 91.820 -60.466 54.941
503522 91.847 -60.474 54.944
503523 91.873 -60.482 54.941
503524 91.880 -60.497 54.934
503525 91.886 -60.513 54.921
503526 91.892 -60.529 54.903
503527 91.916 -60.544 54.878
503528 91.956 -60.560 54.852
503529 91.996 -60.575 54.835
503530 92.036 -60.591 54.827
503531 92.076 -60.607 54.819
503532 92.116 -60.622 54.807
503533 92.156 -60.638 54.792
503534 92.196 -60.653 54.781
503535 92.236 -60.668 54.775
503536 92.272 -60.686 54.774
503537 92.307 -60.706 54.771
503538 92.342 -60.725 54.762
503539 92.377 -60.744 54.745
503540 92.412 -60.763 54.721
503541 92.455 -60.774 54.691
503542 92.500 -60.784 54.662
503543 92.545 -60.794 54.644
503544 92.589 -60.803 54.627
503545 92.634 -60.813 54.611
503546 92.677 -60.821 54.594
503547 92.718 -60.826 54.577
503548 92.759 -60.832 54.556
503549 92.800 -60.839 54.533
503550 92.841 -60.844 54.508
503551 92.882 -60.850 54.485
503552 92.923 -60.856 54.484
503553 92.964 -60.862 54.491
503554 93.005 -60.867 54.498
503555 93.046 -60.873 54.503
503556 93.086 -60.879 54.502
503557 93.128 -60.885 54.495
503558 93.168 -60.891 54.484
503559 93.209 -60.897 54.472
503560 93.250 -60.903 54.459
503561 93.291 -60.908 54.448
503562 93.332 -60.914 54.440
503563 93.373 -60.920 54.428
503564 93.414 -60.926 54.417
503565 93.455 -60.932 54.407
503566 93.496 -60.943 54.397
503567 93.539 -60.955 54.384
503568 93.580 -60.967 54.365
503569 93.620 -60.978 54.342
503570 93.659 -60.988 54.321
503571 93.699 -60.998 54.300
503572 93.739 -61.008 54.279
503573 93.778 -61.019 54.260
503574 93.818 -61.029 54.241
503575 93.861 -61.039 54.224
503576 93.904 -61.049 54.215
503577 93.947 -61.059 54.214
503578 93.990 -61.069 54.211
503579 94.033 -61.080 54.208
503580 94.076 -61.089 54.201
503581 94.119 -61.099 54.189
503582 94.161 -61.107 54.172
503583 94.204 -61.114 54.149
503584 94.246 -61.123 54.120
503585 94.288 -61.132 54.089
503586 94.331 -61.141 54.056
503587 94.373 -61.150 54.022
503588 94.416 -61.158 53.986
503589 94.458 -61.168 53.951
503590 94.500 -61.177 53.911
503591 94.542 -61.183 53.874
503592 94.583 -61.190 53.842
503593 94.625 -61.196 53.811
503594 94.667 -61.202 53.781
503595 94.707 -61.209 53.746
503596 94.745 -61.216 53.709
503597 94.776 -61.217 53.672
503598 94.804 -61.215 53.635
503599 94.831 -61.213 53.601
503600 94.857 -61.212 53.566
503601 94.891 -61.216 53.531
503602 94.927 -61.224 53.494
503603 94.965 -61.227 53.455
503604 95.001 -61.230 53.418
503605 95.038 -61.234 53.380
503606 95.074 -61.239 53.343
503607 95.110 -61.243 53.309
503608 95.147 -61.248 53.275
503609 95.185 -61.253 53.238
503610 95.225 -61.259 53.204
503611 95.262 -61.265 53.170
503612 95.300 -61.272 53.138
503613 95.337 -61.279 53.107
503614 95.375 -61.285 53.075
503615 95.412 -61.291 53.042
503616 95.448 -61.296 53.009
503617 95.483 -61.300 52.976
503618 95.519 -61.305 52.944
503619 95.554 -61.309 52.912
503620 95.590 -61.314 52.881
503621 95.627 -61.318 52.849
503622 95.665 -61.321 52.815
503623 95.702 -61.324 52.782
503624 95.740 -61.328 52.750
503625 95.777 -61.331 52.719
503626 95.815 -61.336 52.689
503627 95.852 -61.340 52.657
503628 95.889 -61.345 52.623
503629 95.927 -61.349 52.587
503630 95.967 -61.353 52.548
503631 96.008 -61.355 52.511
503632 96.048 -61.358 52.475
503633 96.088 -61.361 52.438
503634 96.128 -61.364 52.400
503635 96.169 -61.367 52.364
503636 96.209 -61.370 52.330
503637 96.249 -61.374 52.297
503638 96.290 -61.378 52.266
503639 96.330 -61.381 52.235
503640 96.370 -61.385 52.202
503641 96.415 -61.381 52.168
503642 96.458 -61.380 52.132
503643 96.490 -61.388 52.096
503644 96.522 -61.396 52.061
503645 96.554 -61.405 52.027
503646 96.585 -61.417 51.992
503647 96.614 -61.430 51.957
503648 96.644 -61.443 51.922
503649 96.685 -61.447 51.887
503650 96.732 -61.446 51.851
503651 96.779 -61.445 51.817
503652 96.816 -61.451 51.782
503653 96.854 -61.457 51.747
503654 96.891 -61.463 51.711
503655 96.929 -61.465 51.675
503656 96.968 -61.464 51.639
503657 97.008 -61.462 51.604
503658 97.047 -61.461 51.566
503659 97.087 -61.459 51.527
503660 97.126 -61.458 51.489
503661 97.166 -61.456 51.450
503662 97.205 -61.454 51.412
503663 97.244 -61.452 51.375
503664 97.284 -61.457 51.337
503665 97.323 -61.463 51.298
503666 97.362 -61.469 51.259
503667 97.401 -61.475 51.218
503668 97.441 -61.481 51.177
503669 97.480 -61.487 51.138
503670 97.521 -61.486 51.099
503671 97.565 -61.478 51.061
503672 97.608 -61.470 51.023
503673 97.651 -61.462 50.986
503674 97.696 -61.457 50.946
503675 97.744 -61.462 50.906
503676 97.793 -61.467 50.868
503677 97.842 -61.472 50.831
503678 97.864 -61.478 50.792
503679 97.877 -61.490 50.753
503680 97.901 -61.505 50.714
503681 97.925 -61.521 50.675
503682 97.949 -61.537 50.636
503683 97.973 -61.552 50.598
503684 97.998 -61.568 50.561
503685 98.022 -61.583 50.523
503686 98.046 -61.599 50.485
503687 98.070 -61.615 50.449
504135 88.093 -59.337 56.098
504136 88.145 -59.344 56.074
504137 88.197 -59.351 56.049
504138 88.249 -59.358 56.025
504139 88.301 -59.365 56.001
504140 88.353 -59.372 55.977
504141 88.405 -59.378 55.951
504142 88.457 -59.385 55.925
504143 88.509 -59.392 55.901
504144 88.561 -59.399 55.875
504145 88.613 -59.406 55.848
504146 88.665 -59.413 55.819
504147 88.717 -59.420 55.789
504148 88.769 -59.427 55.760
504149 88.821 -59.434 55.733
504150 88.877 -59.429 55.706
504151 88.934 -59.416 55.681
504152 88.979 -59.428 55.658
504153 89.019 -59.447 55.637
504154 89.059 -59.468 55.615
504155 89.099 -59.488 55.594
504156 89.139 -59.507 55.574
504157 89.180 -59.528 55.554
504158 89.219 -59.547 55.535
504159 89.260 -59.568 55.517
504160 89.300 -59.588 55.500
504161 89.340 -59.607 55.485
504162 89.380 -59.628 55.471
504163 89.397 -59.638 55.458
504164 89.409 -59.647 55.446
504165 89.420 -59.656 55.434
504166 89.449 -59.668 55.423
504167 89.489 -59.682 55.414
504168 89.529 -59.698 55.407
504169 89.570 -59.712 55.400
504170 89.610 -59.727 55.391
504171 89.651 -59.742 55.379
504172 89.691 -59.757 55.368
504173 89.732 -59.772 55.358
504174 89.774 -59.793 55.349
504175 89.817 -59.814 55.342
504176 89.860 -59.836 55.336
504177 89.903 -59.858 55.330
504178 89.948 -59.877 55.323
504179 89.993 -59.898 55.315
504180 90.039 -59.917 55.308
504181 90.085 -59.938 55.300
504182 90.130 -59.958 55.290
504183 90.175 -59.973 55.277
504184 90.219 -59.985 55.262
504185 90.263 -59.997 55.246
504186 90.308 -60.009 55.227
504187 90.352 -60.021 55.209
504188 90.397 -60.033 55.190
504189 90.441 -60.045 55.170
504190 90.485 -60.057 55.150
504191 90.530 -60.069 55.131
504192 90.574 -60.081 55.112
504193 90.618 -60.093 55.095
504194 90.663 -60.105 55.081
504195 90.707 -60.117 55.069
504196 90.751 -60.129 55.057
504197 90.796 -60.141 55.047
504198 90.840 -60.153 55.041
504199 90.883 -60.163 55.038
504200 90.927 -60.172 55.040
504201 90.963 -60.205 55.042
504202 90.999 -60.237 55.041
504203 91.035 -60.271 55.039
504204 91.073 -60.299 55.037
504205 91.115 -60.306 55.036
504206 91.157 -60.312 55.037
504207 91.200 -60.319 55.040
504208 91.242 -60.326 55.043
504209 91.285 -60.333 55.045
504210 91.327 -60.340 55.044
504211 91.370 -60.347 55.040
504212 91.412 -60.354 55.038
504213 91.454 -60.361 55.036
504214 91.497 -60.368 55.031
504215 91.539 -60.375 55.023
504216 91.582 -60.382 55.009
504217 91.624 -60.389 54.992
504218 91.666 -60.396 54.974
504219 91.709 -60.403 54.955
504220 91.751 -60.409 54.936
504221 91.781 -60.418 54.922
504222 91.808 -60.426 54.919
504223 91.834 -60.434 54.918
504224 91.860 -60.443 54.915
504225 91.886 -60.451 54.907
504226 91.894 -60.466 54.894
504227 91.900 -60.481 54.877
504228 91.923 -60.497 54.855
504229 91.963 -60.513 54.834
504230 92.003 -60.528 54.818
504231 92.042 -60.544 54.811
504232 92.083 -60.559 54.807
504233 92.123 -60.575 54.800
504234 92.161 -60.591 54.790
504235 92.196 -60.610 54.779
504236 92.231 -60.629 54.770
504237 92.266 -60.648 54.770
504238 92.301 -60.667 54.768
504239 92.337 -60.686 54.761
504240 92.371 -60.705 54.745
504241 92.414 -60.717 54.721
504242 92.459 -60.727 54.690
504243 92.504 -60.736 54.658
504244 92.549 -60.746 54.636
504245 92.593 -60.756 54.618
504246 92.638 -60.765 54.601
504247 92.684 -60.775 54.586
504248 92.725 -60.782 54.570
504249 92.766 -60.787 54.551
504250 92.807 -60.793 54.527
504251 92.848 -60.799 54.500
504252 92.888 -60.805 54.478
504253 92.929 -60.810 54.482
504254 92.971 -60.817 54.489
504255 93.011 -60.823 54.498
504256 93.052 -60.828 54.504
504257 93.093 -60.834 54.505
504258 93.134 -60.840 54.503
504259 93.175 -60.846 54.497
504260 93.216 -60.851 54.485
504261 93.257 -60.857 54.471
504262 93.298 -60.864 54.457
504263 93.339 -60.869 54.440
504264 93.379 -60.875 54.419
504265 93.420 -60.881 54.404
504266 93.462 -60.890 54.394
504267 93.504 -60.902 54.386
504268 93.546 -60.915 54.375
504269 93.587 -60.927 54.357
504270 93.629 -60.939 54.339
504271 93.669 -60.949 54.319
504272 93.708 -60.960 54.297
504273 93.748 -60.970 54.272
504274 93.787 -60.980 54.256
504275 93.828 -60.990 54.238
504276 93.871 -61.001 54.222
504277 93.914 -61.011 54.213
504278 93.957 -61.020 54.211
504279 94.000 -61.031 54.206
504280 94.043 -61.041 54.203
504281 94.086 -61.051 54.197
504282 94.129 -61.061 54.188
504283 94.172 -61.071 54.174
504284 94.214 -61.080 54.153
504285 94.257 -61.089 54.126
504286 94.299 -61.099 54.095
504287 94.342 -61.107 54.062
504288 94.384 -61.116 54.029
504289 94.426 -61.125 53.995
504290 94.469 -61.134 53.960
504291 94.511 -61.143 53.919
504292 94.553 -61.150 53.880
504293 94.595 -61.156 53.843
504294 94.636 -61.163 53.807
504295 94.675 -61.170 53.772
504296 94.712 -61.177 53.736
504297 94.750 -61.183 53.700
504298 94.783 -61.187 53.665
504299 94.812 -61.184 53.631
504300 94.840 -61.183 53.600
504301 94.868 -61.181 53.567
504302 94.899 -61.182 53.532
504303 94.935 -61.186 53.493
504304 94.972 -61.190 53.452
504305 95.008 -61.195 53.412
504306 95.044 -61.199 53.374
504307 95.080 -61.204 53.338
504308 95.117 -61.208 53.304
504309 95.154 -61.213 53.270
504310 95.192 -61.218 53.235
504311 95.232 -61.224 53.202
504312 95.270 -61.230 53.169
504313 95.308 -61.237 53.138
504314 95.345 -61.243 53.108
504315 95.383 -61.250 53.075
504316 95.418 -61.255 53.041
504317 95.454 -61.259 53.009
504318 95.490 -61.264 52.976
504319 95.525 -61.268 52.943
504320 95.561 -61.273 52.912
504321 95.596 -61.277 52.880
504322 95.633 -61.281 52.846
504323 95.670 -61.285 52.811
504324 95.708 -61.288 52.775
504325 95.746 -61.292 52.741
504326 95.783 -61.296 52.707
504327 95.820 -61.301 52.675
504328 95.857 -61.306 52.642
504329 95.894 -61.311 52.607
504330 95.932 -61.315 52.572
504331 95.972 -61.318 52.534
504332 96.013 -61.321 52.499
504333 96.052 -61.324 52.464
504334 96.093 -61.327 52.428
504335 96.133 -61.331 52.392
504336 96.173 -61.334 52.355
504337 96.214 -61.337 52.319
504338 96.254 -61.341 52.285
504339 96.294 -61.345 52.251
504340 96.334 -61.349 52.218
504341 96.375 -61.353 52.184
504342 96.421 -61.347 52.150
504343 96.467 -61.343 52.117
504344 96.503 -61.348 52.082
504345 96.535 -61.356 52.048
504346 96.567 -61.365 52.018
504347 96.599 -61.373 51.984
504348 96.631 -61.381 51.950
504349 96.662 -61.394 51.917
504350 96.695 -61.404 51.883
504351 96.742 -61.403 51.849
504352 96.782 -61.408 51.815
504353 96.819 -61.414 51.780
504354 96.856 -61.420 51.745
504355 96.893 -61.426 51.710
504356 96.930 -61.432 51.675
504357 96.967 -61.438 51.640
504358 97.007 -61.438 51.604
504359 97.046 -61.437 51.567
504360 97.085 -61.435 51.528
504361 97.125 -61.434 51.489
504362 97.164 -61.432 51.448
504363 97.204 -61.428 51.408
504364 97.243 -61.425 51.368
504365 97.282 -61.431 51.330
504366 97.321 -61.437 51.291
504367 97.361 -61.443 51.251
504368 97.400 -61.449 51.212
504369 97.439 -61.454 51.171
504370 97.480 -61.454 51.132
504371 97.523 -61.447 51.094
504372 97.567 -61.438 51.057
504373 97.610 -61.431 51.019
504374 97.653 -61.423 50.982
504375 97.700 -61.423 50.943
504376 97.749 -61.428 50.904
504377 97.797 -61.433 50.866
504378 97.846 -61.437 50.830
504379 97.869 -61.444 50.791
504380 97.879 -61.454 50.753
504381 97.903 -61.470 50.714
504382 97.927 -61.486 50.676
504383 97.951 -61.501 50.637
504384 97.975 -61.517 50.599
504385 98.000 -61.532 50.562
504386 98.024 -61.548 50.525
504387 98.048 -61.564 50.489
504836 88.120 -59.293 56.104
504837 88.172 -59.300 56.080
504838 88.224 -59.307 56.056
504839 88.276 -59.314 56.033
504840 88.328 -59.321 56.011
504841 88.380 -59.328 55.988
504842 88.432 -59.334 55.962
504843 88.484 -59.341 55.937
504844 88.536 -59.348 55.914
504845 88.588 -59.355 55.888
504846 88.640 -59.362 55.862
504847 88.692 -59.369 55.834
504848 88.744 -59.376 55.804
504849 88.796 -59.383 55.775
504850 88.848 -59.390 55.747
504851 88.905 -59.381 55.721
504852 88.959 -59.375 55.696
504853 88.998 -59.395 55.673
504854 89.039 -59.416 55.651
504855 89.079 -59.435 55.629
504856 89.119 -59.456 55.608
504857 89.159 -59.475 55.589
504858 89.199 -59.495 55.571
504859 89.239 -59.516 55.554
504860 89.279 -59.535 55.537
504861 89.319 -59.555 55.521
504862 89.353 -59.573 55.508
504863 89.388 -59.589 55.496
504864 89.409 -59.602 55.486
504865 89.428 -59.614 55.477
504866 89.442 -59.624 55.468
504867 89.463 -59.635 55.458
504868 89.504 -59.650 55.450
504869 89.544 -59.665 55.444
504870 89.585 -59.679 55.440
504871 89.625 -59.694 55.434
504872 89.666 -59.709 55.424
504873 89.706 -59.723 55.412
504874 89.747 -59.739 55.401
504875 89.788 -59.756 55.392
504876 89.831 -59.778 55.386
504877 89.874 -59.799 55.380
504878 89.918 -59.820 55.373
504879 89.963 -59.840 55.366
504880 90.009 -59.861 55.359
504881 90.054 -59.880 55.352
504882 90.099 -59.895 55.348
504883 90.144 -59.907 55.340
504884 90.188 -59.919 55.325
504885 90.232 -59.931 55.306
504886 90.276 -59.943 55.287
504887 90.321 -59.955 55.265
504888 90.365 -59.967 55.243
504889 90.409 -59.979 55.221
504890 90.454 -59.991 55.198
504891 90.498 -60.003 55.175
504892 90.542 -60.015 55.153
504893 90.587 -60.027 55.132
504894 90.631 -60.039 55.112
504895 90.675 -60.051 55.094
504896 90.720 -60.063 55.078
504897 90.764 -60.075 55.064
504898 90.809 -60.087 55.052
504899 90.853 -60.099 55.044
504900 90.897 -60.110 55.039
504901 90.940 -60.121 55.038
504902 90.975 -60.155 55.038
504903 91.012 -60.188 55.035
504904 91.048 -60.221 55.032
504905 91.084 -60.254 55.030
504906 91.126 -60.264 55.029
504907 91.168 -60.271 55.032
504908 91.211 -60.278 55.035
504909 91.253 -60.285 55.036
504910 91.295 -60.292 55.035
504911 91.338 -60.299 55.034
504912 91.380 -60.306 55.029
504913 91.423 -60.312 55.029
504914 91.465 -60.319 55.030
504915 91.508 -60.327 55.026
504916 91.550 -60.334 55.018
504917 91.592 -60.340 55.006
504918 91.635 -60.347 54.990
504919 91.677 -60.354 54.972
504920 91.716 -60.362 54.951
504921 91.742 -60.370 54.928
504922 91.769 -60.378 54.907
504923 91.795 -60.387 54.896
504924 91.821 -60.395 54.890
504925 91.848 -60.403 54.883
504926 91.874 -60.411 54.874
504927 91.900 -60.419 54.862
504928 91.909 -60.434 54.846
504929 91.930 -60.450 54.828
504930 91.969 -60.466 54.810
504931 92.009 -60.481 54.797
504932 92.050 -60.497 54.789
504933 92.086 -60.515 54.783
504934 92.121 -60.534 54.778
504935 92.156 -60.553 54.771
504936 92.191 -60.572 54.763
504937 92.226 -60.591 54.757
504938 92.261 -60.610 54.757
504939 92.296 -60.629 54.757
504940 92.331 -60.647 54.751
504941 92.373 -60.660 54.739
504942 92.418 -60.669 54.718
504943 92.463 -60.679 54.690
504944 92.508 -60.689 54.656
504945 92.553 -60.698 54.634
504946 92.597 -60.708 54.617
504947 92.642 -60.718 54.602
504948 92.688 -60.728 54.586
504949 92.731 -60.736 54.568
504950 92.772 -60.742 54.549
504951 92.813 -60.748 54.522
504952 92.854 -60.754 54.493
504953 92.895 -60.760 54.470
504954 92.936 -60.766 54.475
504955 92.977 -60.771 54.483
504956 93.018 -60.777 54.492
504957 93.059 -60.783 54.498
504958 93.099 -60.789 54.502
504959 93.140 -60.795 54.505
504960 93.182 -60.801 54.502
504961 93.222 -60.807 54.493
504962 93.263 -60.812 54.482
504963 93.304 -60.818 54.469
504964 93.345 -60.824 54.451
504965 93.386 -60.830 54.431
504966 93.427 -60.837 54.413
504967 93.469 -60.849 54.398
504968 93.511 -60.862 54.383
504969 93.553 -60.874 54.365
504970 93.594 -60.887 54.345
504971 93.637 -60.899 54.327
504972 93.678 -60.911 54.310
504973 93.718 -60.921 54.290
504974 93.757 -60.932 54.268
504975 93.797 -60.942 54.249
504976 93.838 -60.952 54.230
504977 93.881 -60.962 54.217
504978 93.924 -60.972 54.210
504979 93.967 -60.983 54.205
504980 94.010 -60.992 54.195
504981 94.053 -61.002 54.192
504982 94.096 -61.013 54.187
504983 94.137 -61.024 54.181
504984 94.179 -61.036 54.169
504985 94.222 -61.045 54.149
504986 94.266 -61.055 54.124
504987 94.310 -61.064 54.095
504988 94.352 -61.074 54.064
504989 94.395 -61.083 54.033
504990 94.437 -61.092 54.002
504991 94.480 -61.101 53.967
504992 94.522 -61.110 53.928
504993 94.564 -61.117 53.888
504994 94.605 -61.124 53.847
504995 94.643 -61.130 53.805
504996 94.681 -61.136 53.765
504997 94.718 -61.143 53.728
504998 94.755 -61.150 53.692
504999 94.791 -61.156 53.659
505000 94.820 -61.154 53.627
505001 94.848 -61.152 53.598
505002 94.877 -61.150 53.566
505003 94.908 -61.149 53.531
505004 94.942 -61.151 53.493
505005 94.978 -61.155 53.452
505006 95.015 -61.160 53.411
505007 95.051 -61.165 53.373
505008 95.087 -61.169 53.337
505009 95.123 -61.174 53.303
505010 95.160 -61.178 53.270
505011 95.198 -61.183 53.236
505012 95.239 -61.189 53.202
505013 95.278 -61.195 53.170
505014 95.316 -61.202 53.139
505015 95.353 -61.208 53.108
505016 95.390 -61.214 53.076
505017 95.425 -61.218 53.043
505018 95.461 -61.223 53.010
505019 95.496 -61.227 52.977
505020 95.532 -61.232 52.945
505021 95.567 -61.237 52.913
505022 95.603 -61.241 52.882
505023 95.638 -61.246 52.848
505024 95.676 -61.249 52.812
505025 95.714 -61.252 52.777
505026 95.751 -61.257 52.742
505027 95.788 -61.262 52.705
505028 95.825 -61.267 52.670
505029 95.863 -61.271 52.634
505030 95.899 -61.276 52.598
505031 95.937 -61.281 52.563
505032 95.977 -61.284 52.527
505033 96.017 -61.287 52.493
505034 96.057 -61.290 52.460
505035 96.098 -61.294 52.426
505036 96.138 -61.298 52.390
505037 96.178 -61.302 52.352
505038 96.219 -61.306 52.315
505039 96.259 -61.309 52.277
505040 96.299 -61.313 52.242
505041 96.339 -61.317 52.208
505042 96.381 -61.318 52.171
505043 96.427 -61.313 52.137
505044 96.473 -61.309 52.102
505045 96.515 -61.308 52.067
505046 96.547 -61.316 52.035
505047 96.580 -61.324 52.006
505048 96.612 -61.333 51.975
505049 96.644 -61.341 51.944
505050 96.676 -61.349 51.913
505051 96.709 -61.359 51.881
505052 96.747 -61.364 51.849
505053 96.784 -61.370 51.814
505054 96.821 -61.376 51.779
505055 96.859 -61.382 51.744
505056 96.895 -61.388 51.709
505057 96.933 -61.394 51.673
505058 96.970 -61.400 51.639
505059 97.007 -61.406 51.604
505060 97.044 -61.412 51.568
505061 97.084 -61.410 51.529
505062 97.123 -61.409 51.489
505063 97.163 -61.407 51.447
505064 97.202 -61.403 51.405
505065 97.241 -61.398 51.363
505066 97.281 -61.404 51.324
505067 97.320 -61.410 51.285
505068 97.359 -61.415 51.244
505069 97.398 -61.422 51.206
505070 97.439 -61.422 51.166
505071 97.483 -61.415 51.127
505072 97.526 -61.406 51.090
505073 97.569 -61.399 51.053
505074 97.612 -61.391 51.016
505075 97.656 -61.384 50.979
505076 97.705 -61.389 50.941
505077 97.753 -61.393 50.902
505078 97.802 -61.398 50.865
505079 97.851 -61.403 50.828
505080 97.873 -61.409 50.791
505081 97.880 -61.419 50.753
505082 97.904 -61.434 50.715
505083 97.929 -61.450 50.677
505084 97.953 -61.466 50.638
505085 97.977 -61.481 50.601
505086 98.001 -61.497 50.564
505087 98.025 -61.512 50.527
505088 98.050 -61.528 50.491
505537 88.146 -59.249 56.109
505538 88.198 -59.256 56.085
505539 88.251 -59.262 56.062
505540 88.303 -59.270 56.040
505541 88.354 -59.277 56.017
505542 88.407 -59.284 55.996
505543 88.459 -59.291 55.972
505544 88.510 -59.297 55.949
505545 88.563 -59.304 55.926
505546 88.615 -59.311 55.901
505547 88.667 -59.318 55.875
505548 88.719 -59.325 55.848
505549 88.771 -59.332 55.819
505550 88.823 -59.339 55.790
505551 88.875 -59.346 55.761
505552 88.932 -59.334 55.734
505553 88.978 -59.344 55.709
505554 89.018 -59.363 55.685
505555 89.058 -59.384 55.662
505556 89.098 -59.403 55.640
505557 89.138 -59.423 55.620
505558 89.179 -59.444 55.602
505559 89.218 -59.463 55.584
505560 89.257 -59.482 55.568
505561 89.292 -59.500 55.554
505562 89.326 -59.516 55.539
505563 89.361 -59.533 55.529
505564 89.395 -59.550 55.521
505565 89.415 -59.563 55.516
505566 89.434 -59.576 55.511
505567 89.453 -59.588 55.505
505568 89.478 -59.601 55.497
505569 89.519 -59.617 55.490
505570 89.559 -59.632 55.484
505571 89.600 -59.647 55.480
505572 89.640 -59.661 55.477
505573 89.681 -59.676 55.470
505574 89.721 -59.691 55.461
505575 89.761 -59.706 55.450
505576 89.802 -59.720 55.440
505577 89.844 -59.741 55.438
505578 89.887 -59.763 55.432
505579 89.933 -59.783 55.424
505580 89.978 -59.803 55.415
505581 90.023 -59.817 55.407
505582 90.068 -59.829 55.402
505583 90.112 -59.841 55.402
505584 90.156 -59.853 55.397
505585 90.200 -59.865 55.385
505586 90.245 -59.877 55.365
505587 90.289 -59.889 55.341
505588 90.333 -59.901 55.311
505589 90.378 -59.913 55.284
505590 90.422 -59.925 55.257
505591 90.466 -59.937 55.229
505592 90.511 -59.949 55.202
505593 90.555 -59.961 55.177
505594 90.600 -59.973 55.153
505595 90.644 -59.985 55.129
505596 90.688 -59.997 55.107
505597 90.733 -60.009 55.087
505598 90.777 -60.021 55.070
505599 90.821 -60.033 55.057
505600 90.866 -60.045 55.046
505601 90.910 -60.057 55.039
505602 90.952 -60.071 55.035
505603 90.989 -60.105 55.032
505604 91.025 -60.138 55.028
505605 91.061 -60.171 55.024
505606 91.097 -60.204 55.024
505607 91.137 -60.222 55.025
505608 91.179 -60.230 55.027
505609 91.221 -60.237 55.026
505610 91.264 -60.243 55.025
505611 91.306 -60.250 55.022
505612 91.349 -60.257 55.020
505613 91.391 -60.264 55.017
505614 91.434 -60.271 55.017
505615 91.476 -60.278 55.017
505616 91.518 -60.285 55.015
505617 91.561 -60.292 55.010
505618 91.603 -60.299 55.000
505619 91.646 -60.306 54.985
505620 91.677 -60.314 54.967
505621 91.703 -60.322 54.945
505622 91.729 -60.331 54.920
505623 91.756 -60.339 54.892
505624 91.782 -60.347 54.872
505625 91.808 -60.355 54.858
505626 91.835 -60.363 54.848
505627 91.861 -60.372 54.837
505628 91.887 -60.380 54.826
505629 91.914 -60.388 54.813
505630 91.936 -60.403 54.797
505631 91.975 -60.419 54.783
505632 92.010 -60.438 54.772
505633 92.045 -60.457 54.763
505634 92.080 -60.476 54.758
505635 92.115 -60.495 54.754
505636 92.150 -60.514 54.750
505637 92.185 -60.533 54.746
505638 92.220 -60.553 54.743
505639 92.255 -60.572 54.740
505640 92.290 -60.591 54.738
505641 92.332 -60.603 54.731
505642 92.377 -60.613 54.720
505643 92.422 -60.622 54.704
505644 92.467 -60.632 54.681
505645 92.512 -60.641 54.652
505646 92.557 -60.651 54.626
505647 92.601 -60.660 54.605
505648 92.646 -60.670 54.589
505649 92.692 -60.680 54.575
505650 92.736 -60.690 54.560
505651 92.779 -60.697 54.539
505652 92.819 -60.703 54.514
505653 92.861 -60.709 54.487
505654 92.902 -60.714 54.465
505655 92.942 -60.720 54.459
505656 92.983 -60.726 54.466
505657 93.024 -60.732 54.475
505658 93.065 -60.738 54.484
505659 93.106 -60.744 54.492
505660 93.147 -60.750 54.499
505661 93.188 -60.755 54.498
505662 93.229 -60.761 54.496
505663 93.270 -60.767 54.491
505664 93.310 -60.773 54.483
505665 93.352 -60.779 54.465
505666 93.393 -60.785 54.451
505667 93.434 -60.796 54.432
505668 93.476 -60.809 54.410
505669 93.518 -60.821 54.388
505670 93.560 -60.834 54.364
505671 93.602 -60.846 54.344
505672 93.644 -60.858 54.327
505673 93.686 -60.871 54.310
505674 93.727 -60.883 54.291
505675 93.767 -60.893 54.269
505676 93.806 -60.903 54.248
505677 93.848 -60.914 54.230
505678 93.891 -60.923 54.215
505679 93.934 -60.934 54.204
505680 93.977 -60.944 54.196
505681 94.020 -60.954 54.184
505682 94.061 -60.965 54.178
505683 94.103 -60.977 54.173
505684 94.144 -60.989 54.165
505685 94.186 -60.999 54.153
505686 94.229 -61.009 54.135
505687 94.273 -61.019 54.113
505688 94.317 -61.028 54.087
505689 94.360 -61.038 54.058
505690 94.403 -61.048 54.028
505691 94.447 -61.057 53.997
505692 94.490 -61.067 53.965
505693 94.533 -61.076 53.929
505694 94.574 -61.084 53.891
505695 94.611 -61.090 53.849
505696 94.648 -61.097 53.805
505697 94.686 -61.104 53.762
505698 94.723 -61.111 53.724
505699 94.760 -61.118 53.688
505700 94.797 -61.124 53.654
505701 94.827 -61.124 53.623
505702 94.856 -61.122 53.594
505703 94.885 -61.120 53.562
505704 94.918 -61.121 53.528
505705 94.951 -61.121 53.492
505706 94.985 -61.121 53.453
505707 95.021 -61.125 53.413
505708 95.057 -61.130 53.376
505709 95.094 -61.134 53.340
505710 95.130 -61.138 53.306
505711 95.166 -61.143 53.273
505712 95.205 -61.148 53.239
505713 95.245 -61.154 53.205
505714 95.285 -61.159 53.173
505715 95.323 -61.166 53.142
505716 95.361 -61.173 53.110
505717 95.397 -61.177 53.078
505718 95.432 -61.182 53.045
505719 95.468 -61.187 53.010
505720 95.503 -61.191 52.978
505721 95.539 -61.196 52.946
505722 95.574 -61.200 52.914
505723 95.610 -61.205 52.882
505724 95.645 -61.209 52.848
505725 95.682 -61.214 52.813
505726 95.719 -61.218 52.777
505727 95.756 -61.223 52.741
505728 95.793 -61.227 52.704
505729 95.830 -61.233 52.668
505730 95.868 -61.237 52.632
505731 95.905 -61.242 52.596
505732 95.942 -61.246 52.560
505733 95.982 -61.250 52.525
505734 96.022 -61.254 52.490
505735 96.062 -61.258 52.457
505736 96.103 -61.262 52.424
505737 96.143 -61.265 52.388
505738 96.183 -61.269 52.350
505739 96.224 -61.273 52.312
505740 96.264 -61.277 52.275
505741 96.304 -61.281 52.239
505742 96.344 -61.284 52.203
505743 96.387 -61.284 52.167
505744 96.434 -61.279 52.130
505745 96.479 -61.274 52.095
505746 96.525 -61.270 52.059
505747 96.560 -61.276 52.025
505748 96.592 -61.284 51.994
505749 96.624 -61.293 51.962
505750 96.657 -61.301 51.932
505751 96.689 -61.309 51.904
505752 96.721 -61.318 51.875
505753 96.756 -61.330 51.845
505754 96.792 -61.335 51.812
505755 96.827 -61.340 51.776
505756 96.864 -61.346 51.741
505757 96.899 -61.352 51.705
505758 96.936 -61.357 51.670
505759 96.972 -61.363 51.636
505760 97.010 -61.369 51.600
505761 97.046 -61.375 51.564
505762 97.084 -61.381 51.527
505763 97.122 -61.384 51.487
505764 97.161 -61.383 51.445
505765 97.201 -61.378 51.403
505766 97.240 -61.371 51.361
505767 97.279 -61.377 51.321
505768 97.318 -61.383 51.282
505769 97.358 -61.389 51.241
505770 97.398 -61.390 51.203
505771 97.442 -61.383 51.164
505772 97.485 -61.375 51.126
505773 97.528 -61.367 51.088
505774 97.571 -61.359 51.050
505775 97.614 -61.352 51.013
505776 97.660 -61.350 50.976
505777 97.709 -61.355 50.939
505778 97.758 -61.359 50.901
505779 97.806 -61.364 50.864
505780 97.855 -61.368 50.827
505781 97.878 -61.375 50.791
505782 97.882 -61.384 50.753
505783 97.907 -61.400 50.715
505784 97.931 -61.415 50.678
505785 97.955 -61.431 50.638
505786 97.979 -61.447 50.603
505787 98.003 -61.462 50.566
505788 98.027 -61.478 50.529
506238 88.173 -59.205 56.113
506239 88.225 -59.212 56.089
506240 88.277 -59.218 56.067
506241 88.329 -59.225 56.045
506242 88.382 -59.233 56.023
506243 88.433 -59.240 56.002
506244 88.485 -59.247 55.980
506245 88.538 -59.253 55.957
506246 88.590 -59.260 55.936
506247 88.641 -59.267 55.912
506248 88.694 -59.274 55.887
506249 88.746 -59.281 55.861
506250 88.798 -59.287 55.833
506251 88.850 -59.295 55.802
506252 88.902 -59.300 55.775
506253 88.958 -59.291 55.746
506254 88.997 -59.311 55.720
506255 89.038 -59.331 55.696
506256 89.078 -59.351 55.672
506257 89.118 -59.372 55.650
506258 89.158 -59.391 55.631
506259 89.195 -59.410 55.612
506260 89.230 -59.427 55.595
506261 89.264 -59.444 55.580
506262 89.299 -59.460 55.567
506263 89.333 -59.477 55.555
506264 89.367 -59.494 55.548
506265 89.401 -59.511 55.543
506266 89.421 -59.523 55.540
506267 89.441 -59.536 55.537
506268 89.460 -59.548 55.532
506269 89.493 -59.563 55.525
506270 89.533 -59.579 55.518
506271 89.574 -59.595 55.513
506272 89.614 -59.610 55.512
506273 89.655 -59.626 55.515
506274 89.695 -59.642 55.514
506275 89.736 -59.658 55.511
506276 89.776 -59.673 55.504
506277 89.817 -59.688 55.495
506278 89.858 -59.704 55.497
506279 89.902 -59.725 55.491
506280 89.947 -59.739 55.480
506281 89.991 -59.751 55.467
506282 90.036 -59.763 55.457
506283 90.080 -59.775 55.453
506284 90.124 -59.787 55.453
506285 90.169 -59.799 55.449
506286 90.213 -59.811 55.440
506287 90.258 -59.823 55.423
506288 90.302 -59.835 55.397
506289 90.346 -59.847 55.364
506290 90.391 -59.859 55.332
506291 90.435 -59.871 55.299
506292 90.479 -59.883 55.265
506293 90.524 -59.895 55.233
506294 90.568 -59.908 55.202
506295 90.612 -59.920 55.174
506296 90.657 -59.933 55.146
506297 90.701 -59.946 55.121
506298 90.746 -59.958 55.098
506299 90.790 -59.971 55.079
506300 90.835 -59.983 55.062
506301 90.879 -59.996 55.048
506302 90.924 -60.008 55.038
506303 90.965 -60.024 55.033
506304 91.001 -60.055 55.026
506305 91.037 -60.088 55.019
506306 91.074 -60.121 55.016
506307 91.109 -60.155 55.018
506308 91.147 -60.181 55.021
506309 91.189 -60.188 55.020
506310 91.232 -60.195 55.017
506311 91.275 -60.202 55.014
506312 91.317 -60.209 55.011
506313 91.359 -60.216 55.008
506314 91.402 -60.223 55.005
506315 91.444 -60.230 55.004
506316 91.487 -60.237 55.003
506317 91.529 -60.243 55.001
506318 91.572 -60.250 54.997
506319 91.611 -60.258 54.989
506320 91.638 -60.266 54.975
506321 91.664 -60.275 54.957
506322 91.690 -60.283 54.935
506323 91.717 -60.291 54.909
506324 91.743 -60.300 54.882
506325 91.769 -60.307 54.858
506326 91.796 -60.315 54.838
506327 91.822 -60.324 54.820
506328 91.848 -60.332 54.805
506329 91.875 -60.340 54.789
506330 91.910 -60.350 54.777
506331 91.934 -60.362 54.764
506332 91.969 -60.381 54.751
506333 92.004 -60.400 54.741
506334 92.039 -60.419 54.735
506335 92.074 -60.438 54.731
506336 92.109 -60.457 54.730
506337 92.144 -60.476 54.729
506338 92.179 -60.495 54.727
506339 92.214 -60.514 54.726
506340 92.249 -60.533 54.722
506341 92.291 -60.546 54.716
506342 92.336 -60.555 54.709
506343 92.381 -60.565 54.698
506344 92.426 -60.575 54.685
506345 92.471 -60.585 54.668
506346 92.516 -60.594 54.644
506347 92.561 -60.604 54.619
506348 92.605 -60.613 54.597
506349 92.650 -60.622 54.581
506350 92.696 -60.632 54.566
506351 92.740 -60.642 54.552
506352 92.785 -60.652 54.535
506353 92.826 -60.657 54.515
506354 92.867 -60.663 54.494
506355 92.908 -60.669 54.472
506356 92.949 -60.676 54.458
506357 92.990 -60.681 54.454
506358 93.030 -60.687 54.457
506359 93.072 -60.693 54.464
506360 93.113 -60.698 54.475
506361 93.153 -60.704 54.482
506362 93.194 -60.710 54.487
506363 93.235 -60.716 54.490
506364 93.276 -60.722 54.491
506365 93.317 -60.728 54.489
506366 93.358 -60.734 54.477
506367 93.400 -60.744 54.466
506368 93.441 -60.756 54.449
506369 93.483 -60.768 54.428
506370 93.525 -60.781 54.405
506371 93.567 -60.793 54.380
506372 93.609 -60.806 54.362
506373 93.651 -60.818 54.342
506374 93.693 -60.830 54.321
506375 93.735 -60.843 54.299
506376 93.776 -60.854 54.276
506377 93.816 -60.865 54.254
506378 93.858 -60.875 54.236
506379 93.901 -60.885 54.217
506380 93.943 -60.895 54.199
506381 93.985 -60.907 54.185
506382 94.027 -60.918 54.172
506383 94.068 -60.930 54.163
506384 94.110 -60.941 54.155
506385 94.151 -60.952 54.146
506386 94.193 -60.964 54.132
506387 94.237 -60.973 54.116
506388 94.280 -60.983 54.097
506389 94.324 -60.992 54.075
506390 94.367 -61.002 54.049
506391 94.410 -61.011 54.021
506392 94.454 -61.021 53.991
506393 94.497 -61.031 53.959
506394 94.540 -61.040 53.925
506395 94.578 -61.045 53.889
506396 94.616 -61.055 53.848
506397 94.654 -61.064 53.803
506398 94.691 -61.071 53.762
506399 94.728 -61.078 53.725
506400 94.765 -61.084 53.691
506401 94.803 -61.091 53.657
506402 94.835 -61.093 53.626
506403 94.863 -61.092 53.596
506404 94.895 -61.091 53.564
506405 94.928 -61.091 53.531
506406 94.962 -61.091 53.496
506407 94.995 -61.091 53.459
506408 95.029 -61.092 53.422
506409 95.064 -61.094 53.386
506410 95.100 -61.099 53.352
506411 95.136 -61.103 53.318
506412 95.173 -61.108 53.284
506413 95.211 -61.113 53.247
506414 95.252 -61.119 53.212
506415 95.292 -61.124 53.178
506416 95.331 -61.130 53.145
506417 95.368 -61.136 53.113
506418 95.403 -61.141 53.080
506419 95.439 -61.146 53.045
506420 95.474 -61.150 53.010
506421 95.510 -61.155 52.976
506422 95.544 -61.161 52.944
506423 95.577 -61.167 52.912
506424 95.611 -61.173 52.880
506425 95.649 -61.176 52.846
506426 95.692 -61.181 52.811
506427 95.751 -61.193 52.776
506428 95.788 -61.199 52.740
506429 95.826 -61.203 52.703
506430 95.864 -61.208 52.668
506431 95.899 -61.215 52.633
506432 95.927 -61.228 52.597
506433 95.958 -61.226 52.560
506434 95.995 -61.207 52.524
506435 96.027 -61.221 52.488
506436 96.067 -61.225 52.455
506437 96.108 -61.229 52.422
506438 96.148 -61.233 52.387
506439 96.188 -61.237 52.349
506440 96.229 -61.240 52.312
506441 96.269 -61.244 52.276
506442 96.309 -61.248 52.239
506443 96.349 -61.252 52.202
506444 96.394 -61.249 52.165
506445 96.440 -61.245 52.127
506446 96.485 -61.240 52.091
506447 96.531 -61.236 52.054
506448 96.573 -61.236 52.018
506449 96.605 -61.244 51.985
506450 96.637 -61.252 51.952
506451 96.669 -61.261 51.921
506452 96.701 -61.269 51.893
506453 96.735 -61.280 51.865
506454 96.768 -61.294 51.837
506455 96.804 -61.302 51.806
506456 96.840 -61.308 51.772
506457 96.876 -61.313 51.738
506458 96.912 -61.319 51.703
506459 96.948 -61.324 51.667
506460 96.984 -61.330 51.631
506461 97.020 -61.335 51.595
506462 97.055 -61.341 51.559
506463 97.092 -61.346 51.522
506464 97.127 -61.352 51.485
506465 97.161 -61.356 51.444
506466 97.199 -61.352 51.401
506467 97.238 -61.346 51.360
506468 97.278 -61.350 51.319
506469 97.317 -61.356 51.280
506470 97.357 -61.359 51.240
506471 97.400 -61.351 51.201
506472 97.444 -61.343 51.162
506473 97.487 -61.335 51.124
506474 97.530 -61.328 51.086
506475 97.573 -61.320 51.048
506476 97.616 -61.312 51.011
506477 97.665 -61.315 50.975
506478 97.714 -61.321 50.937
506479 97.762 -61.325 50.900
506480 97.811 -61.330 50.863
506481 97.860 -61.334 50.827
506482 97.882 -61.341 50.790
506483 97.886 -61.349 50.753
506484 97.908 -61.364 50.715
506485 97.932 -61.380 50.678
506486 97.956 -61.396 50.640
506487 97.981 -61.411 50.604
506488 98.005 -61.427 50.567
506939 88.200 -59.161 56.114
506940 88.252 -59.168 56.093
506941 88.304 -59.175 56.071
506942 88.356 -59.181 56.049
506943 88.408 -59.188 56.028
506944 88.460 -59.196 56.007
506945 88.512 -59.203 55.986
506946 88.564 -59.209 55.964
506947 88.616 -59.216 55.943
506948 88.668 -59.223 55.920
506949 88.720 -59.230 55.896
506950 88.772 -59.237 55.871
506951 88.824 -59.244 55.844
506952 88.877 -59.250 55.815
506953 88.930 -59.252 55.786
506954 88.977 -59.259 55.759
506955 89.017 -59.279 55.732
506956 89.057 -59.299 55.707
506957 89.097 -59.319 55.684
506958 89.134 -59.338 55.660
506959 89.168 -59.354 55.639
506960 89.202 -59.371 55.621
506961 89.237 -59.388 55.604
506962 89.271 -59.404 55.588
506963 89.305 -59.421 55.578
506964 89.339 -59.438 55.568
506965 89.374 -59.455 55.563
506966 89.408 -59.472 55.561
506967 89.428 -59.485 55.558
506968 89.447 -59.497 55.556
506969 89.470 -59.510 55.552
506970 89.509 -59.525 55.545
506971 89.547 -59.540 55.540
506972 89.588 -59.556 55.538
506973 89.628 -59.572 55.541
506974 89.669 -59.588 55.548
506975 89.709 -59.604 55.552
506976 89.750 -59.620 55.553
506977 89.791 -59.635 55.553
506978 89.831 -59.651 55.554
506979 89.872 -59.667 55.559
506980 89.916 -59.681 55.556
506981 89.961 -59.694 55.544
506982 90.006 -59.706 55.526
506983 90.050 -59.719 55.510
506984 90.094 -59.732 55.501
506985 90.139 -59.745 55.497
506986 90.183 -59.757 55.493
506987 90.228 -59.770 55.487
506988 90.272 -59.783 55.478
506989 90.316 -59.795 55.453
506990 90.361 -59.808 55.418
506991 90.405 -59.820 55.382
506992 90.450 -59.833 55.345
506993 90.494 -59.845 55.304
506994 90.538 -59.858 55.265
506995 90.583 -59.871 55.229
506996 90.627 -59.884 55.196
506997 90.672 -59.896 55.164
506998 90.716 -59.909 55.135
506999 90.761 -59.922 55.109
507000 90.805 -59.934 55.087
507001 90.850 -59.947 55.067
507002 90.894 -59.960 55.051
507003 90.938 -59.973 55.038
507004 90.977 -59.989 55.029
507005 91.014 -60.008 55.020
507006 91.050 -60.039 55.013
507007 91.086 -60.071 55.008
507008 91.122 -60.105 55.007
507009 91.158 -60.138 55.009
507010 91.201 -60.146 55.010
507011 91.243 -60.153 55.009
507012 91.285 -60.161 55.006
507013 91.328 -60.168 55.003
507014 91.370 -60.174 55.000
507015 91.413 -60.181 54.996
507016 91.455 -60.188 54.991
507017 91.498 -60.195 54.988
507018 91.540 -60.202 54.985
507019 91.572 -60.210 54.982
507020 91.599 -60.218 54.975
507021 91.625 -60.227 54.962
507022 91.651 -60.235 54.944
507023 91.677 -60.243 54.923
507024 91.704 -60.251 54.897
507025 91.730 -60.259 54.872
507026 91.756 -60.268 54.846
507027 91.784 -60.277 54.822
507028 91.819 -60.286 54.801
507029 91.854 -60.295 54.783
507030 91.889 -60.305 54.765
507031 91.917 -60.315 54.749
507032 91.936 -60.327 54.735
507033 91.963 -60.343 54.722
507034 91.999 -60.362 54.712
507035 92.033 -60.381 54.706
507036 92.069 -60.400 54.703
507037 92.103 -60.419 54.703
507038 92.139 -60.437 54.704
507039 92.174 -60.456 54.704
507040 92.209 -60.475 54.703
507041 92.250 -60.488 54.701
507042 92.295 -60.498 54.697
507043 92.340 -60.508 54.690
507044 92.385 -60.517 54.679
507045 92.430 -60.527 54.666
507046 92.475 -60.537 54.650
507047 92.520 -60.547 54.631
507048 92.565 -60.556 54.612
507049 92.609 -60.566 54.594
507050 92.654 -60.575 54.577
507051 92.700 -60.585 54.561
507052 92.744 -60.594 54.547
507053 92.789 -60.604 54.535
507054 92.833 -60.613 54.521
507055 92.873 -60.619 54.503
507056 92.915 -60.624 54.483
507057 92.955 -60.630 54.463
507058 92.996 -60.636 54.450
507059 93.037 -60.641 54.445
507060 93.078 -60.647 54.447
507061 93.119 -60.654 54.455
507062 93.160 -60.660 54.463
507063 93.201 -60.665 54.472
507064 93.242 -60.671 54.479
507065 93.283 -60.677 54.484
507066 93.324 -60.682 54.487
507067 93.365 -60.691 54.480
507068 93.407 -60.703 54.473
507069 93.448 -60.715 54.459
507070 93.491 -60.728 54.439
507071 93.533 -60.740 54.416
507072 93.574 -60.753 54.393
507073 93.616 -60.765 54.376
507074 93.658 -60.777 54.357
507075 93.700 -60.790 54.338
507076 93.742 -60.802 54.318
507077 93.784 -60.815 54.295
507078 93.825 -60.826 54.278
507079 93.867 -60.836 54.259
507080 93.909 -60.848 54.234
507081 93.951 -60.860 54.204
507082 93.992 -60.871 54.178
507083 94.034 -60.882 54.159
507084 94.075 -60.894 54.143
507085 94.117 -60.905 54.133
507086 94.158 -60.917 54.122
507087 94.200 -60.928 54.107
507088 94.244 -60.937 54.094
507089 94.287 -60.947 54.078
507090 94.331 -60.957 54.059
507091 94.374 -60.966 54.036
507092 94.417 -60.976 54.011
507093 94.461 -60.986 53.983
507094 94.504 -60.995 53.950
507095 94.544 -61.001 53.916
507096 94.581 -61.005 53.883
507097 94.618 -61.011 53.845
507098 94.656 -61.019 53.802
507099 94.695 -61.033 53.764
507100 94.734 -61.045 53.732
507101 94.771 -61.052 53.700
507102 94.808 -61.058 53.668
507103 94.843 -61.063 53.637
507104 94.872 -61.061 53.606
507105 94.905 -61.061 53.575
507106 94.938 -61.061 53.543
507107 94.972 -61.061 53.508
507108 95.005 -61.062 53.472
507109 95.039 -61.062 53.436
507110 95.072 -61.062 53.402
507111 95.107 -61.064 53.369
507112 95.143 -61.068 53.334
507113 95.179 -61.073 53.298
507114 95.218 -61.078 53.261
507115 95.258 -61.083 53.225
507116 95.299 -61.089 53.190
507117 95.339 -61.096 53.155
507118 95.373 -61.101 53.120
507119 95.407 -61.108 53.085
507120 95.440 -61.114 53.048
507121 95.473 -61.120 53.012
507122 95.507 -61.127 52.977
507123 95.540 -61.133 52.942
507124 95.575 -61.138 52.909
507125 95.614 -61.139 52.876
507126 95.653 -61.142 52.841
507127 95.692 -61.144 52.806
507128 95.752 -61.157 52.771
507129 95.811 -61.170 52.734
507130 95.857 -61.183 52.698
507131 95.885 -61.197 52.664
507132 95.912 -61.211 52.629
507133 95.940 -61.224 52.592
507134 95.976 -61.202 52.555
507135 96.019 -61.152 52.519
507136 96.049 -61.168 52.484
507137 96.079 -61.184 52.452
507138 96.113 -61.196 52.420
507139 96.153 -61.200 52.387
507140 96.193 -61.204 52.353
507141 96.234 -61.208 52.318
507142 96.274 -61.212 52.282
507143 96.314 -61.215 52.245
507144 96.355 -61.219 52.205
507145 96.400 -61.215 52.165
507146 96.446 -61.211 52.127
507147 96.491 -61.206 52.090
507148 96.537 -61.202 52.052
507149 96.583 -61.197 52.015
507150 96.617 -61.204 51.979
507151 96.650 -61.212 51.945
507152 96.682 -61.221 51.912
507153 96.714 -61.229 51.880
507154 96.748 -61.243 51.851
507155 96.782 -61.257 51.824
507156 96.817 -61.269 51.797
507157 96.853 -61.275 51.766
507158 96.888 -61.281 51.734
507159 96.925 -61.286 51.700
507160 96.960 -61.291 51.664
507161 96.996 -61.297 51.627
507162 97.032 -61.302 51.589
507163 97.068 -61.308 51.553
507164 97.104 -61.314 51.517
507165 97.140 -61.316 51.481
507166 97.171 -61.316 51.442
507167 97.198 -61.326 51.399
507168 97.237 -61.321 51.359
507169 97.276 -61.323 51.318
507170 97.316 -61.327 51.278
507171 97.360 -61.319 51.239
507172 97.402 -61.312 51.200
507173 97.446 -61.303 51.161
507174 97.489 -61.296 51.123
507175 97.532 -61.288 51.085
507176 97.576 -61.281 51.048
507177 97.621 -61.277 51.010
507178 97.669 -61.281 50.974
507179 97.718 -61.286 50.937
507180 97.767 -61.291 50.900
507181 97.815 -61.296 50.863
507182 97.864 -61.300 50.827
507183 97.887 -61.307 50.790
507184 97.890 -61.315 50.753
507185 97.910 -61.329 50.716
507186 97.934 -61.344 50.679
507187 97.958 -61.360 50.641
507188 97.983 -61.376 50.605
507640 88.227 -59.117 56.116
507641 88.279 -59.124 56.094
507642 88.331 -59.131 56.072
507643 88.383 -59.137 56.052
507644 88.435 -59.144 56.031
507645 88.487 -59.151 56.011
507646 88.539 -59.159 55.991
507647 88.591 -59.165 55.969
507648 88.643 -59.172 55.948
507649 88.695 -59.179 55.926
507650 88.747 -59.186 55.903
507651 88.799 -59.193 55.879
507652 88.851 -59.200 55.853
507653 88.903 -59.206 55.826
507654 88.957 -59.207 55.797
507655 88.996 -59.227 55.770
507656 89.037 -59.247 55.743
507657 89.072 -59.265 55.718
507658 89.106 -59.281 55.693
507659 89.140 -59.298 55.669
507660 89.175 -59.315 55.649
507661 89.209 -59.331 55.630
507662 89.243 -59.348 55.613
507663 89.277 -59.366 55.598
507664 89.312 -59.382 55.586
507665 89.346 -59.399 55.577
507666 89.381 -59.416 55.571
507667 89.415 -59.432 55.569
507668 89.434 -59.445 55.565
507669 89.453 -59.457 55.564
507670 89.486 -59.472 55.561
507671 89.524 -59.487 55.558
507672 89.563 -59.502 55.556
507673 89.602 -59.517 55.559
507674 89.642 -59.533 55.567
507675 89.683 -59.549 55.576
507676 89.724 -59.565 55.584
507677 89.764 -59.581 55.590
507678 89.805 -59.597 55.596
507679 89.846 -59.613 55.606
507680 89.887 -59.628 55.614
507681 89.932 -59.642 55.614
507682 89.976 -59.657 55.604
507683 90.020 -59.670 55.586
507684 90.064 -59.682 55.567
507685 90.109 -59.695 55.552
507686 90.154 -59.708 55.541
507687 90.198 -59.720 55.532
507688 90.242 -59.733 55.525
507689 90.287 -59.746 55.513
507690 90.331 -59.759 55.489
507691 90.376 -59.771 55.459
507692 90.420 -59.784 55.425
507693 90.464 -59.797 55.389
507694 90.509 -59.809 55.347
507695 90.553 -59.822 55.300
507696 90.598 -59.835 55.257
507697 90.642 -59.848 55.218
507698 90.687 -59.860 55.183
507699 90.731 -59.873 55.150
507700 90.776 -59.886 55.122
507701 90.820 -59.898 55.095
507702 90.864 -59.911 55.073
507703 90.909 -59.924 55.053
507704 90.953 -59.936 55.038
507705 90.990 -59.955 55.025
507706 91.026 -59.974 55.015
507707 91.063 -59.992 55.008
507708 91.099 -60.022 55.003
507709 91.135 -60.055 54.999
507710 91.171 -60.088 54.995
507711 91.211 -60.105 54.994
507712 91.253 -60.112 54.992
507713 91.296 -60.119 54.989
507714 91.338 -60.126 54.986
507715 91.381 -60.133 54.984
507716 91.423 -60.140 54.980
507717 91.466 -60.147 54.977
507718 91.507 -60.154 54.975
507719 91.533 -60.162 54.972
507720 91.560 -60.171 54.968
507721 91.586 -60.179 54.959
507722 91.612 -60.187 54.945
507723 91.639 -60.196 54.927
507724 91.665 -60.203 54.907
507725 91.693 -60.212 54.884
507726 91.728 -60.221 54.861
507727 91.763 -60.231 54.835
507728 91.798 -60.240 54.810
507729 91.833 -60.250 54.787
507730 91.868 -60.259 54.768
507731 91.901 -60.269 54.748
507732 91.920 -60.281 54.729
507733 91.939 -60.293 54.711
507734 91.958 -60.304 54.697
507735 91.993 -60.323 54.686
507736 92.028 -60.342 54.679
507737 92.063 -60.361 54.675
507738 92.098 -60.380 54.674
507739 92.133 -60.400 54.674
507740 92.168 -60.419 54.674
507741 92.210 -60.431 54.674
507742 92.254 -60.441 54.673
507743 92.299 -60.450 54.670
507744 92.344 -60.460 54.665
507745 92.389 -60.470 54.656
507746 92.434 -60.479 54.646
507747 92.479 -60.489 54.635
507748 92.524 -60.499 54.622
507749 92.569 -60.509 54.606
507750 92.614 -60.518 54.591
507751 92.658 -60.528 54.578
507752 92.703 -60.538 54.565
507753 92.748 -60.547 54.554
507754 92.793 -60.557 54.543
507755 92.838 -60.566 54.528
507756 92.880 -60.573 54.509
507757 92.921 -60.579 54.493
507758 92.962 -60.585 54.474
507759 93.003 -60.591 54.456
507760 93.044 -60.597 54.443
507761 93.085 -60.603 54.434
507762 93.126 -60.608 54.432
507763 93.166 -60.614 54.442
507764 93.207 -60.620 54.454
507765 93.249 -60.626 54.464
507766 93.289 -60.632 54.472
507767 93.330 -60.638 54.479
507768 93.372 -60.650 54.476
507769 93.414 -60.663 54.472
507770 93.456 -60.675 54.460
507771 93.498 -60.687 54.442
507772 93.540 -60.700 54.419
507773 93.581 -60.712 54.398
507774 93.624 -60.725 54.381
507775 93.665 -60.737 54.365
507776 93.707 -60.749 54.348
507777 93.749 -60.762 54.330
507778 93.791 -60.774 54.311
507779 93.832 -60.789 54.294
507780 93.874 -60.801 54.272
507781 93.916 -60.812 54.244
507782 93.958 -60.823 54.212
507783 93.999 -60.835 54.182
507784 94.041 -60.846 54.155
507785 94.082 -60.857 54.133
507786 94.124 -60.870 54.117
507787 94.166 -60.881 54.102
507788 94.207 -60.892 54.084
507789 94.251 -60.901 54.069
507790 94.294 -60.911 54.052
507791 94.338 -60.921 54.033
507792 94.381 -60.930 54.014
507793 94.424 -60.940 53.992
507794 94.468 -60.950 53.966
507795 94.509 -60.957 53.936
507796 94.548 -60.961 53.904
507797 94.585 -60.966 53.872
507798 94.622 -60.971 53.839
507799 94.659 -60.976 53.804
507800 94.696 -60.983 53.774
507801 94.735 -60.996 53.747
507802 94.774 -61.010 53.718
507803 94.813 -61.024 53.689
507804 94.850 -61.032 53.661
507805 94.883 -61.032 53.628
507806 94.916 -61.032 53.597
507807 94.949 -61.032 53.564
507808 94.982 -61.033 53.530
507809 95.016 -61.033 53.493
507810 95.049 -61.033 53.457
507811 95.083 -61.033 53.423
507812 95.116 -61.033 53.389
507813 95.149 -61.033 53.353
507814 95.186 -61.038 53.316
507815 95.225 -61.043 53.279
507816 95.265 -61.049 53.242
507817 95.303 -61.054 53.206
507818 95.339 -61.060 53.169
507819 95.373 -61.068 53.132
507820 95.403 -61.079 53.094
507821 95.436 -61.086 53.055
507822 95.469 -61.092 53.017
507823 95.502 -61.099 52.980
507824 95.541 -61.102 52.943
507825 95.580 -61.104 52.909
507826 95.619 -61.105 52.874
507827 95.658 -61.108 52.838
507828 95.697 -61.110 52.802
507829 95.751 -61.121 52.766
507830 95.811 -61.133 52.729
507831 95.850 -61.139 52.693
507832 95.878 -61.142 52.659
507833 95.920 -61.149 52.624
507834 95.950 -61.180 52.587
507835 95.988 -61.162 52.550
507836 96.035 -61.102 52.514
507837 96.072 -61.112 52.482
507838 96.103 -61.129 52.452
507839 96.133 -61.145 52.424
507840 96.163 -61.161 52.393
507841 96.198 -61.172 52.361
507842 96.239 -61.176 52.327
507843 96.279 -61.180 52.292
507844 96.319 -61.183 52.255
507845 96.360 -61.186 52.216
507846 96.406 -61.181 52.173
507847 96.452 -61.177 52.131
507848 96.498 -61.172 52.091
507849 96.543 -61.168 52.053
507850 96.589 -61.163 52.015
507851 96.629 -61.164 51.979
507852 96.662 -61.172 51.943
507853 96.694 -61.180 51.908
507854 96.727 -61.191 51.875
507855 96.761 -61.205 51.844
507856 96.796 -61.220 51.814
507857 96.829 -61.234 51.785
507858 96.865 -61.242 51.755
507859 96.901 -61.248 51.725
507860 96.937 -61.253 51.693
507861 96.973 -61.259 51.660
507862 97.009 -61.264 51.626
507863 97.044 -61.270 51.590
507864 97.081 -61.275 51.553
507865 97.116 -61.281 51.515
507866 97.151 -61.272 51.476
507867 97.180 -61.277 51.437
507868 97.208 -61.286 51.397
507869 97.235 -61.295 51.357
507870 97.275 -61.295 51.318
507871 97.318 -61.287 51.279
507872 97.362 -61.280 51.240
507873 97.405 -61.271 51.200
507874 97.448 -61.264 51.161
507875 97.492 -61.256 51.123
507876 97.534 -61.249 51.085
507877 97.578 -61.240 51.048
507878 97.626 -61.243 51.012
507879 97.674 -61.247 50.975
507880 97.723 -61.252 50.938
507881 97.772 -61.256 50.901
507882 97.820 -61.261 50.864
507883 97.869 -61.266 50.827
507884 97.891 -61.272 50.792
507885 97.895 -61.281 50.755
507886 97.912 -61.293 50.717
507887 97.936 -61.309 50.680
507888 97.960 -61.325 50.643
507889 97.984 -61.340 50.607
508342 88.306 -59.080 56.094
508343 88.358 -59.087 56.072
508344 88.410 -59.093 56.052
508345 88.462 -59.100 56.032
508346 88.514 -59.107 56.012
508347 88.566 -59.114 55.992
508348 88.618 -59.121 55.971
508349 88.670 -59.128 55.951
508350 88.722 -59.135 55.930
508351 88.774 -59.142 55.908
508352 88.826 -59.149 55.885
508353 88.878 -59.156 55.861
508354 88.930 -59.162 55.834
508355 88.976 -59.175 55.806
508356 89.010 -59.192 55.780
508357 89.044 -59.209 55.754
508358 89.079 -59.225 55.728
508359 89.113 -59.242 55.702
508360 89.147 -59.259 55.678
508361 89.182 -59.275 55.657
508362 89.216 -59.293 55.639
508363 89.250 -59.309 55.622
508364 89.284 -59.326 55.607
508365 89.319 -59.343 55.595
508366 89.353 -59.359 55.586
508367 89.387 -59.377 55.581
508368 89.421 -59.394 55.577
508369 89.441 -59.406 55.573
508370 89.463 -59.419 55.572
508371 89.502 -59.434 55.571
508372 89.540 -59.449 55.569
508373 89.578 -59.463 55.570
508374 89.617 -59.479 55.575
508375 89.657 -59.494 55.584
508376 89.697 -59.510 55.596
508377 89.738 -59.526 55.607
508378 89.779 -59.542 55.619
508379 89.819 -59.558 55.632
508380 89.860 -59.574 55.647
508381 89.902 -59.589 55.661
508382 89.947 -59.604 55.665
508383 89.991 -59.618 55.659
508384 90.035 -59.632 55.645
508385 90.080 -59.646 55.627
508386 90.124 -59.659 55.608
508387 90.168 -59.672 55.590
508388 90.213 -59.685 55.577
508389 90.257 -59.697 55.567
508390 90.302 -59.710 55.553
508391 90.346 -59.723 55.525
508392 90.390 -59.735 55.491
508393 90.435 -59.748 55.456
508394 90.479 -59.761 55.419
508395 90.524 -59.773 55.379
508396 90.568 -59.786 55.333
508397 90.612 -59.798 55.285
508398 90.657 -59.811 55.241
508399 90.701 -59.823 55.201
508400 90.746 -59.836 55.165
508401 90.790 -59.849 55.133
508402 90.835 -59.862 55.104
508403 90.879 -59.874 55.078
508404 90.924 -59.887 55.056
508405 90.965 -59.902 55.038
508406 91.002 -59.920 55.024
508407 91.038 -59.939 55.013
508408 91.075 -59.958 55.006
508409 91.111 -59.977 55.000
508410 91.148 -60.005 54.992
508411 91.184 -60.038 54.983
508412 91.222 -60.064 54.978
508413 91.264 -60.071 54.974
508414 91.307 -60.077 54.971
508415 91.349 -60.084 54.970
508416 91.392 -60.092 54.968
508417 91.434 -60.099 54.966
508418 91.468 -60.106 54.964
508419 91.494 -60.115 54.963
508420 91.520 -60.123 54.960
508421 91.547 -60.131 54.955
508422 91.573 -60.140 54.945
508423 91.602 -60.148 54.931
508424 91.637 -60.157 54.914
508425 91.672 -60.167 54.893
508426 91.707 -60.176 54.870
508427 91.742 -60.186 54.846
508428 91.777 -60.195 54.823
508429 91.812 -60.205 54.800
508430 91.847 -60.214 54.778
508431 91.882 -60.224 54.758
508432 91.905 -60.234 54.737
508433 91.923 -60.246 54.718
508434 91.942 -60.258 54.699
508435 91.961 -60.270 54.682
508436 91.987 -60.285 54.667
508437 92.022 -60.304 54.654
508438 92.057 -60.323 54.647
508439 92.092 -60.342 54.642
508440 92.128 -60.361 54.639
508441 92.168 -60.374 54.638
508442 92.213 -60.384 54.640
508443 92.259 -60.394 54.639
508444 92.303 -60.403 54.637
508445 92.348 -60.412 54.634
508446 92.393 -60.422 54.629
508447 92.438 -60.432 54.621
508448 92.483 -60.441 54.614
508449 92.528 -60.451 54.607
508450 92.573 -60.461 54.597
508451 92.618 -60.471 54.586
508452 92.662 -60.480 54.577
508453 92.707 -60.490 54.567
508454 92.753 -60.500 54.560
508455 92.797 -60.510 54.548
508456 92.842 -60.519 54.532
508457 92.886 -60.528 54.513
508458 92.927 -60.534 54.496
508459 92.969 -60.540 54.478
508460 93.009 -60.546 54.459
508461 93.050 -60.551 54.443
508462 93.091 -60.557 54.432
508463 93.132 -60.563 54.424
508464 93.173 -60.569 54.424
508465 93.214 -60.575 54.430
508466 93.255 -60.581 54.441
508467 93.296 -60.587 54.453
508468 93.337 -60.597 54.462
508469 93.379 -60.610 54.463
508470 93.421 -60.622 54.459
508471 93.463 -60.635 54.449
508472 93.505 -60.647 54.433
508473 93.547 -60.659 54.412
508474 93.589 -60.672 54.392
508475 93.631 -60.684 54.377
508476 93.673 -60.696 54.364
508477 93.714 -60.709 54.350
508478 93.756 -60.723 54.336
508479 93.796 -60.738 54.323
508480 93.837 -60.753 54.303
508481 93.880 -60.765 54.279
508482 93.923 -60.776 54.250
508483 93.965 -60.788 54.221
508484 94.006 -60.799 54.191
508485 94.048 -60.810 54.159
508486 94.089 -60.822 54.134
508487 94.131 -60.833 54.113
508488 94.173 -60.845 54.095
508489 94.214 -60.856 54.078
508490 94.258 -60.866 54.062
508491 94.301 -60.876 54.044
508492 94.345 -60.885 54.025
508493 94.388 -60.895 54.006
508494 94.431 -60.904 53.986
508495 94.474 -60.913 53.963
508496 94.512 -60.917 53.939
508497 94.551 -60.922 53.913
508498 94.589 -60.926 53.884
508499 94.625 -60.932 53.854
508500 94.662 -60.936 53.824
508501 94.699 -60.941 53.796
508502 94.736 -60.947 53.768
508503 94.775 -60.961 53.742
508504 94.814 -60.974 53.716
508505 94.853 -60.988 53.690
508506 94.887 -60.990 53.662
508507 94.921 -60.992 53.632
508508 94.955 -60.993 53.601
508509 94.989 -60.995 53.565
508510 95.023 -60.996 53.528
508511 95.057 -60.998 53.488
508512 95.091 -60.999 53.450
508513 95.125 -61.001 53.413
508514 95.159 -61.002 53.377
508515 95.193 -61.004 53.343
508516 95.229 -61.008 53.305
508517 95.266 -61.013 53.267
508518 95.303 -61.019 53.229
508519 95.339 -61.024 53.190
508520 95.375 -61.030 53.150
508521 95.405 -61.041 53.110
508522 95.436 -61.052 53.068
508523 95.468 -61.062 53.027
508524 95.507 -61.065 52.987
508525 95.545 -61.067 52.948
508526 95.585 -61.070 52.910
508527 95.623 -61.071 52.874
508528 95.663 -61.074 52.835
508529 95.701 -61.076 52.798
508530 95.751 -61.083 52.761
508531 95.811 -61.096 52.725
508532 95.842 -61.100 52.689
508533 95.870 -61.102 52.655
508534 95.898 -61.105 52.619
508535 95.954 -61.077 52.582
508536 95.989 -61.078 52.546
508537 96.022 -61.085 52.511
508538 96.081 -61.081 52.480
508539 96.116 -61.090 52.453
508540 96.151 -61.099 52.427
508541 96.185 -61.108 52.400
508542 96.217 -61.122 52.370
508543 96.247 -61.138 52.336
508544 96.284 -61.147 52.302
508545 96.324 -61.151 52.266
508546 96.366 -61.152 52.228
508547 96.412 -61.147 52.183
508548 96.458 -61.143 52.138
508549 96.504 -61.138 52.096
508550 96.549 -61.133 52.056
508551 96.595 -61.129 52.017
508552 96.641 -61.124 51.980
508553 96.674 -61.131 51.943
508554 96.707 -61.139 51.908
508555 96.741 -61.153 51.874
508556 96.774 -61.168 51.841
508557 96.809 -61.183 51.810
508558 96.843 -61.197 51.778
508559 96.877 -61.209 51.747
508560 96.913 -61.215 51.717
508561 96.949 -61.221 51.686
508562 96.985 -61.226 51.654
508563 97.021 -61.231 51.622
508564 97.057 -61.237 51.587
508565 97.093 -61.243 51.551
508566 97.128 -61.238 51.513
508567 97.163 -61.228 51.474
508568 97.190 -61.237 51.437
508569 97.218 -61.246 51.397
508570 97.245 -61.255 51.359
508571 97.278 -61.255 51.320
508572 97.321 -61.248 51.281
508573 97.364 -61.240 51.242
508574 97.407 -61.232 51.203
508575 97.450 -61.224 51.164
508576 97.494 -61.217 51.126
508577 97.537 -61.208 51.088
508578 97.581 -61.203 51.051
508579 97.630 -61.208 51.014
508580 97.678 -61.213 50.977
508581 97.727 -61.218 50.941
508582 97.776 -61.222 50.903
508583 97.824 -61.227 50.866
508584 97.873 -61.231 50.831
508585 97.896 -61.238 50.795
508586 97.899 -61.246 50.758
508587 97.914 -61.259 50.721
508588 97.938 -61.274 50.684
508589 97.962 -61.290 50.646
509043 88.332 -59.036 56.093
509044 88.385 -59.043 56.072
509045 88.436 -59.049 56.052
509046 88.488 -59.056 56.032
509047 88.541 -59.063 56.012
509048 88.593 -59.070 55.992
509049 88.644 -59.077 55.972
509050 88.697 -59.084 55.952
509051 88.749 -59.091 55.932
509052 88.801 -59.098 55.910
509053 88.853 -59.105 55.888
509054 88.904 -59.112 55.864
509055 88.949 -59.124 55.839
509056 88.985 -59.137 55.813
509057 89.018 -59.153 55.787
509058 89.051 -59.169 55.761
509059 89.086 -59.186 55.735
509060 89.120 -59.203 55.710
509061 89.154 -59.220 55.686
509062 89.188 -59.237 55.665
509063 89.223 -59.253 55.646
509064 89.257 -59.270 55.630
509065 89.291 -59.287 55.617
509066 89.325 -59.304 55.605
509067 89.360 -59.321 55.597
509068 89.394 -59.338 55.592
509069 89.427 -59.354 55.589
509070 89.447 -59.366 55.586
509071 89.478 -59.381 55.585
509072 89.517 -59.396 55.585
509073 89.555 -59.410 55.585
509074 89.594 -59.425 55.588
509075 89.632 -59.441 55.592
509076 89.671 -59.456 55.601
509077 89.712 -59.472 55.612
509078 89.752 -59.488 55.625
509079 89.793 -59.504 55.640
509080 89.833 -59.519 55.658
509081 89.874 -59.535 55.677
509082 89.917 -59.550 55.695
509083 89.962 -59.564 55.707
509084 90.006 -59.579 55.709
509085 90.050 -59.593 55.699
509086 90.095 -59.607 55.687
509087 90.139 -59.622 55.674
509088 90.183 -59.635 55.651
509089 90.228 -59.648 55.628
509090 90.272 -59.660 55.615
509091 90.316 -59.673 55.598
509092 90.361 -59.686 55.561
509093 90.405 -59.698 55.521
509094 90.450 -59.711 55.481
509095 90.494 -59.724 55.442
509096 90.538 -59.737 55.403
509097 90.583 -59.749 55.359
509098 90.627 -59.762 55.309
509099 90.672 -59.775 55.261
509100 90.716 -59.787 55.218
509101 90.761 -59.800 55.179
509102 90.805 -59.813 55.144
509103 90.850 -59.826 55.112
509104 90.894 -59.838 55.085
509105 90.938 -59.851 55.060
509106 90.977 -59.868 55.040
509107 91.014 -59.886 55.025
509108 91.050 -59.905 55.013
509109 91.087 -59.924 55.005
509110 91.123 -59.942 54.998
509111 91.160 -59.961 54.991
509112 91.196 -59.989 54.980
509113 91.233 -60.021 54.970
509114 91.275 -60.029 54.961
509115 91.317 -60.036 54.956
509116 91.360 -60.043 54.955
509117 91.402 -60.050 54.957
509118 91.429 -60.058 54.954
509119 91.455 -60.067 54.952
509120 91.482 -60.075 54.951
509121 91.511 -60.083 54.948
509122 91.546 -60.093 54.943
509123 91.581 -60.102 54.932
509124 91.616 -60.112 54.918
509125 91.651 -60.121 54.900
509126 91.686 -60.130 54.879
509127 91.721 -60.140 54.856
509128 91.756 -60.149 54.832
509129 91.791 -60.159 54.811
509130 91.826 -60.168 54.789
509131 91.861 -60.178 54.768
509132 91.888 -60.189 54.750
509133 91.908 -60.200 54.732
509134 91.926 -60.212 54.713
509135 91.945 -60.224 54.692
509136 91.964 -60.235 54.672
509137 91.983 -60.247 54.655
509138 92.017 -60.265 54.640
509139 92.052 -60.284 54.629
509140 92.087 -60.303 54.619
509141 92.128 -60.317 54.612
509142 92.172 -60.327 54.606
509143 92.217 -60.336 54.601
509144 92.263 -60.346 54.599
509145 92.307 -60.356 54.597
509146 92.352 -60.365 54.596
509147 92.397 -60.375 54.593
509148 92.442 -60.384 54.587
509149 92.487 -60.394 54.582
509150 92.532 -60.404 54.582
509151 92.577 -60.413 54.583
509152 92.622 -60.423 54.580
509153 92.666 -60.433 54.572
509154 92.711 -60.442 54.567
509155 92.757 -60.452 54.562
509156 92.801 -60.462 54.550
509157 92.846 -60.472 54.533
509158 92.891 -60.481 54.514
509159 92.934 -60.489 54.495
509160 92.975 -60.494 54.477
509161 93.016 -60.500 54.458
509162 93.057 -60.506 54.441
509163 93.097 -60.512 54.428
509164 93.139 -60.518 54.416
509165 93.180 -60.524 54.410
509166 93.220 -60.530 54.411
509167 93.261 -60.535 54.418
509168 93.302 -60.544 54.427
509169 93.345 -60.557 54.436
509170 93.386 -60.569 54.442
509171 93.428 -60.582 54.441
509172 93.470 -60.594 54.431
509173 93.512 -60.606 54.415
509174 93.554 -60.619 54.398
509175 93.596 -60.631 54.380
509176 93.638 -60.644 54.365
509177 93.679 -60.658 54.352
509178 93.719 -60.673 54.340
509179 93.760 -60.688 54.330
509180 93.801 -60.703 54.318
509181 93.843 -60.717 54.301
509182 93.886 -60.729 54.280
509183 93.929 -60.740 54.256
509184 93.972 -60.752 54.229
509185 94.013 -60.763 54.197
509186 94.055 -60.775 54.174
509187 94.097 -60.786 54.153
509188 94.138 -60.798 54.129
509189 94.180 -60.809 54.107
509190 94.221 -60.820 54.093
509191 94.265 -60.830 54.077
509192 94.308 -60.839 54.059
509193 94.352 -60.849 54.036
509194 94.395 -60.859 54.012
509195 94.438 -60.868 53.992
509196 94.478 -60.873 53.972
509197 94.517 -60.878 53.954
509198 94.555 -60.882 53.932
509199 94.593 -60.887 53.906
509200 94.629 -60.892 53.876
509201 94.666 -60.897 53.848
509202 94.702 -60.901 53.819
509203 94.739 -60.906 53.792
509204 94.776 -60.911 53.768
509205 94.815 -60.924 53.746
509206 94.854 -60.938 53.722
509207 94.888 -60.941 53.697
509208 94.922 -60.942 53.668
509209 94.956 -60.944 53.637
509210 94.990 -60.945 53.602
509211 95.024 -60.946 53.564
509212 95.058 -60.948 53.524
509213 95.093 -60.951 53.483
509214 95.128 -60.956 53.443
509215 95.164 -60.961 53.405
509216 95.199 -60.967 53.369
509217 95.235 -60.973 53.330
509218 95.271 -60.978 53.290
509219 95.308 -60.984 53.250
509220 95.344 -60.989 53.209
509221 95.380 -60.995 53.168
509222 95.411 -61.007 53.127
509223 95.442 -61.018 53.084
509224 95.477 -61.025 53.041
509225 95.514 -61.029 53.000
509226 95.551 -61.033 52.959
509227 95.589 -61.035 52.919
509228 95.628 -61.037 52.881
509229 95.667 -61.039 52.841
509230 95.706 -61.042 52.802
509231 95.751 -61.047 52.764
509232 95.808 -61.059 52.725
509233 95.835 -61.061 52.689
509234 95.863 -61.063 52.653
509235 95.891 -61.065 52.616
509236 95.918 -61.067 52.579
509237 95.951 -61.071 52.543
509238 96.009 -61.067 52.510
509239 96.068 -61.064 52.480
509240 96.124 -61.062 52.453
509241 96.158 -61.071 52.427
509242 96.193 -61.080 52.401
509243 96.228 -61.089 52.374
509244 96.262 -61.098 52.343
509245 96.297 -61.106 52.311
509246 96.331 -61.115 52.276
509247 96.372 -61.118 52.239
509248 96.418 -61.113 52.195
509249 96.464 -61.108 52.148
509250 96.510 -61.104 52.104
509251 96.555 -61.099 52.062
509252 96.602 -61.095 52.022
509253 96.648 -61.089 51.985
509254 96.687 -61.091 51.946
509255 96.720 -61.102 51.911
509256 96.754 -61.116 51.876
509257 96.788 -61.130 51.842
509258 96.822 -61.146 51.809
509259 96.856 -61.160 51.777
509260 96.890 -61.174 51.744
509261 96.926 -61.182 51.712
509262 96.962 -61.187 51.680
509263 96.998 -61.193 51.648
509264 97.034 -61.199 51.617
509265 97.070 -61.204 51.582
509266 97.105 -61.204 51.547
509267 97.140 -61.194 51.511
509268 97.173 -61.188 51.474
509269 97.200 -61.197 51.438
509270 97.228 -61.207 51.400
509271 97.255 -61.224 51.362
509272 97.282 -61.216 51.323
509273 97.323 -61.208 51.284
509274 97.366 -61.200 51.245
509275 97.409 -61.193 51.206
509276 97.453 -61.185 51.167
509277 97.496 -61.177 51.129
509278 97.539 -61.169 51.092
509279 97.586 -61.169 51.055
509280 97.635 -61.174 51.018
509281 97.683 -61.179 50.980
509282 97.732 -61.183 50.944
509283 97.781 -61.188 50.907
509284 97.829 -61.193 50.869
509285 97.878 -61.197 50.834
509286 97.900 -61.204 50.798
509287 97.904 -61.212 50.761
509288 97.916 -61.223 50.724
509289 97.940 -61.239 50.687
509290 97.964 -61.255 50.649
509744 88.359 -58.992 56.092
509745 88.411 -58.999 56.071
509746 88.463 -59.005 56.051
509747 88.516 -59.012 56.031
509748 88.567 -59.019 56.011
509749 88.619 -59.026 55.991
509750 88.672 -59.033 55.969
509751 88.723 -59.040 55.951
509752 88.775 -59.047 55.931
509753 88.825 -59.055 55.910
509754 88.871 -59.067 55.889
509755 88.916 -59.079 55.866
509756 88.960 -59.091 55.842
509757 88.995 -59.102 55.817
509758 89.029 -59.118 55.792
509759 89.062 -59.134 55.767
509760 89.095 -59.149 55.742
509761 89.128 -59.165 55.717
509762 89.161 -59.181 55.694
509763 89.195 -59.197 55.673
509764 89.230 -59.214 55.655
509765 89.264 -59.231 55.640
509766 89.298 -59.248 55.627
509767 89.332 -59.265 55.617
509768 89.367 -59.281 55.613
509769 89.401 -59.298 55.612
509770 89.434 -59.315 55.613
509771 89.455 -59.328 55.609
509772 89.493 -59.343 55.607
509773 89.532 -59.357 55.607
509774 89.570 -59.372 55.607
509775 89.609 -59.388 55.609
509776 89.648 -59.403 55.612
509777 89.686 -59.417 55.620
509778 89.726 -59.433 55.630
509779 89.766 -59.449 55.643
509780 89.807 -59.465 55.660
509781 89.848 -59.481 55.681
509782 89.889 -59.497 55.700
509783 89.933 -59.511 55.720
509784 89.977 -59.526 55.737
509785 90.021 -59.540 55.746
509786 90.065 -59.554 55.742
509787 90.110 -59.569 55.735
509788 90.154 -59.582 55.729
509789 90.198 -59.597 55.710
509790 90.242 -59.611 55.684
509791 90.287 -59.624 55.666
509792 90.331 -59.637 55.648
509793 90.376 -59.650 55.607
509794 90.420 -59.663 55.558
509795 90.464 -59.675 55.507
509796 90.509 -59.688 55.459
509797 90.553 -59.701 55.418
509798 90.598 -59.713 55.375
509799 90.642 -59.726 55.327
509800 90.687 -59.739 55.279
509801 90.731 -59.751 55.233
509802 90.776 -59.764 55.191
509803 90.820 -59.776 55.154
509804 90.864 -59.789 55.121
509805 90.909 -59.801 55.091
509806 90.952 -59.815 55.064
509807 90.989 -59.834 55.042
509808 91.026 -59.852 55.026
509809 91.063 -59.870 55.013
509810 91.099 -59.889 55.002
509811 91.136 -59.908 54.993
509812 91.172 -59.927 54.986
509813 91.209 -59.945 54.975
509814 91.245 -59.972 54.963
509815 91.286 -59.988 54.952
509816 91.328 -59.995 54.944
509817 91.363 -60.002 54.941
509818 91.390 -60.011 54.941
509819 91.420 -60.019 54.938
509820 91.455 -60.029 54.936
509821 91.490 -60.038 54.934
509822 91.525 -60.048 54.931
509823 91.560 -60.057 54.926
509824 91.595 -60.067 54.916
509825 91.630 -60.076 54.902
509826 91.665 -60.086 54.885
509827 91.700 -60.095 54.864
509828 91.735 -60.105 54.841
509829 91.771 -60.114 54.818
509830 91.805 -60.123 54.797
509831 91.841 -60.133 54.777
509832 91.873 -60.143 54.760
509833 91.891 -60.154 54.744
509834 91.911 -60.166 54.730
509835 91.929 -60.177 54.712
509836 91.948 -60.189 54.693
509837 91.967 -60.201 54.674
509838 91.986 -60.212 54.656
509839 92.011 -60.227 54.638
509840 92.046 -60.246 54.619
509841 92.087 -60.260 54.603
509842 92.132 -60.269 54.590
509843 92.176 -60.279 54.578
509844 92.221 -60.289 54.569
509845 92.267 -60.298 54.562
509846 92.311 -60.308 54.558
509847 92.356 -60.318 54.554
509848 92.401 -60.328 54.551
509849 92.446 -60.337 54.547
509850 92.491 -60.347 54.544
509851 92.536 -60.356 54.548
509852 92.581 -60.366 54.555
509853 92.626 -60.375 54.561
509854 92.670 -60.385 54.559
509855 92.715 -60.395 54.553
509856 92.761 -60.404 54.545
509857 92.805 -60.414 54.535
509858 92.850 -60.424 54.524
509859 92.895 -60.434 54.506
509860 92.940 -60.443 54.487
509861 92.982 -60.450 54.469
509862 93.022 -60.455 54.453
509863 93.063 -60.461 54.438
509864 93.104 -60.467 54.421
509865 93.145 -60.473 54.406
509866 93.186 -60.478 54.398
509867 93.227 -60.485 54.398
509868 93.268 -60.491 54.400
509869 93.310 -60.503 54.403
509870 93.352 -60.516 54.410
509871 93.394 -60.528 54.418
509872 93.435 -60.541 54.418
509873 93.478 -60.553 54.409
509874 93.519 -60.566 54.394
509875 93.561 -60.578 54.381
509876 93.603 -60.592 54.365
509877 93.643 -60.607 54.349
509878 93.684 -60.622 54.335
509879 93.724 -60.638 54.325
509880 93.765 -60.653 54.318
509881 93.806 -60.668 54.307
509882 93.848 -60.681 54.292
509883 93.891 -60.693 54.274
509884 93.935 -60.704 54.253
509885 93.978 -60.716 54.230
509886 94.020 -60.727 54.207
509887 94.062 -60.738 54.188
509888 94.104 -60.750 54.167
509889 94.145 -60.762 54.144
509890 94.187 -60.773 54.121
509891 94.228 -60.785 54.107
509892 94.272 -60.794 54.093
509893 94.315 -60.804 54.075
509894 94.358 -60.814 54.053
509895 94.402 -60.823 54.030
509896 94.442 -60.829 54.010
509897 94.481 -60.834 53.993
509898 94.520 -60.839 53.975
509899 94.559 -60.843 53.952
509900 94.596 -60.848 53.926
509901 94.633 -60.853 53.897
509902 94.670 -60.857 53.865
509903 94.706 -60.862 53.838
509904 94.743 -60.867 53.815
509905 94.779 -60.871 53.796
509906 94.816 -60.876 53.778
509907 94.855 -60.889 53.757
509908 94.889 -60.891 53.733
509909 94.923 -60.892 53.704
509910 94.957 -60.894 53.673
509911 94.992 -60.899 53.640
509912 95.028 -60.904 53.601
509913 95.063 -60.910 53.562
509914 95.099 -60.915 53.520
509915 95.134 -60.920 53.478
509916 95.170 -60.926 53.436
509917 95.205 -60.931 53.396
509918 95.242 -60.938 53.354
509919 95.278 -60.944 53.312
509920 95.315 -60.949 53.270
509921 95.351 -60.955 53.228
509922 95.386 -60.961 53.185
509923 95.417 -60.973 53.143
509924 95.449 -60.984 53.099
509925 95.485 -60.988 53.055
509926 95.523 -60.992 53.012
509927 95.559 -60.995 52.970
509928 95.597 -60.999 52.929
509929 95.633 -61.003 52.889
509930 95.672 -61.005 52.848
509931 95.711 -61.008 52.808
509932 95.751 -61.011 52.768
509933 95.800 -61.020 52.727
509934 95.828 -61.022 52.690
509935 95.855 -61.024 52.654
509936 95.888 -61.027 52.616
509937 95.929 -61.033 52.579
509938 95.970 -61.037 52.542
509939 96.011 -61.042 52.510
509940 96.055 -61.046 52.482
509941 96.114 -61.042 52.455
509942 96.166 -61.042 52.428
509943 96.200 -61.052 52.402
509944 96.235 -61.061 52.376
509945 96.270 -61.069 52.347
509946 96.304 -61.078 52.317
509947 96.340 -61.081 52.284
509948 96.377 -61.083 52.249
509949 96.419 -61.080 52.206
509950 96.466 -61.075 52.160
509951 96.513 -61.071 52.115
509952 96.560 -61.065 52.072
509953 96.607 -61.061 52.031
509954 96.654 -61.055 51.993
509955 96.699 -61.051 51.954
509956 96.733 -61.064 51.917
509957 96.767 -61.079 51.881
509958 96.802 -61.093 51.845
509959 96.835 -61.108 51.811
509960 96.870 -61.122 51.779
509961 96.904 -61.137 51.745
509962 96.938 -61.149 51.712
509963 96.974 -61.155 51.679
509964 97.010 -61.160 51.646
509965 97.046 -61.166 51.613
509966 97.082 -61.169 51.578
509967 97.116 -61.160 51.544
509968 97.151 -61.150 51.511
509969 97.182 -61.149 51.476
509970 97.210 -61.158 51.441
509971 97.238 -61.174 51.403
509972 97.266 -61.191 51.365
509973 97.292 -61.179 51.326
509974 97.325 -61.168 51.286
509975 97.368 -61.161 51.248
509976 97.412 -61.153 51.209
509977 97.455 -61.145 51.171
509978 97.498 -61.137 51.133
509979 97.541 -61.130 51.095
509980 97.590 -61.135 51.058
509981 97.639 -61.139 51.021
509982 97.687 -61.144 50.984
509983 97.736 -61.149 50.948
509984 97.785 -61.154 50.910
509985 97.833 -61.158 50.873
509986 97.882 -61.163 50.837
509987 97.905 -61.170 50.802
509988 97.909 -61.178 50.765
509989 97.917 -61.188 50.728
509990 97.941 -61.203 50.690
509991 97.965 -61.219 50.653
509992 97.990 -61.235 50.617
510446 88.438 -58.955 56.069
510447 88.490 -58.961 56.049
510448 88.542 -58.968 56.029
510449 88.594 -58.975 56.009
510450 88.646 -58.982 55.989
510451 88.698 -58.989 55.968
510452 88.747 -58.998 55.949
510453 88.792 -59.010 55.929
510454 88.838 -59.022 55.909
510455 88.884 -59.034 55.888
510456 88.929 -59.046 55.866
510457 88.970 -59.057 55.843
510458 89.005 -59.068 55.819
510459 89.040 -59.081 55.795
510460 89.072 -59.097 55.772
510461 89.106 -59.113 55.748
510462 89.139 -59.129 55.725
510463 89.172 -59.145 55.702
510464 89.205 -59.161 55.681
510465 89.238 -59.177 55.664
510466 89.271 -59.193 55.650
510467 89.305 -59.209 55.638
510468 89.339 -59.225 55.631
510469 89.373 -59.243 55.631
510470 89.407 -59.259 55.635
510471 89.440 -59.275 55.640
510472 89.471 -59.290 55.640
510473 89.509 -59.304 55.639
510474 89.547 -59.319 55.640
510475 89.586 -59.334 55.642
510476 89.624 -59.350 55.642
510477 89.663 -59.364 55.640
510478 89.701 -59.379 55.645
510479 89.740 -59.394 55.654
510480 89.781 -59.410 55.665
510481 89.821 -59.426 55.680
510482 89.862 -59.442 55.699
510483 89.904 -59.458 55.718
510484 89.948 -59.472 55.739
510485 89.992 -59.487 55.758
510486 90.036 -59.501 55.773
510487 90.081 -59.515 55.777
510488 90.125 -59.529 55.775
510489 90.169 -59.544 55.773
510490 90.214 -59.558 55.757
510491 90.257 -59.573 55.731
510492 90.302 -59.587 55.709
510493 90.346 -59.601 55.686
510494 90.390 -59.613 55.643
510495 90.435 -59.626 55.591
510496 90.479 -59.638 55.535
510497 90.524 -59.651 55.479
510498 90.568 -59.664 55.431
510499 90.613 -59.676 55.385
510500 90.657 -59.689 55.339
510501 90.701 -59.702 55.294
510502 90.746 -59.715 55.248
510503 90.790 -59.727 55.202
510504 90.835 -59.740 55.162
510505 90.879 -59.753 55.127
510506 90.924 -59.765 55.095
510507 90.964 -59.781 55.067
510508 91.001 -59.799 55.045
510509 91.038 -59.818 55.028
510510 91.075 -59.836 55.014
510511 91.111 -59.855 54.999
510512 91.148 -59.873 54.986
510513 91.184 -59.892 54.977
510514 91.221 -59.911 54.967
510515 91.258 -59.929 54.954
510516 91.297 -59.946 54.943
510517 91.329 -59.955 54.935
510518 91.364 -59.964 54.928
510519 91.399 -59.974 54.924
510520 91.434 -59.983 54.920
510521 91.469 -59.992 54.917
510522 91.504 -60.002 54.913
510523 91.539 -60.011 54.909
510524 91.574 -60.021 54.901
510525 91.609 -60.030 54.890
510526 91.644 -60.040 54.877
510527 91.679 -60.049 54.862
510528 91.715 -60.059 54.843
510529 91.749 -60.068 54.820
510530 91.785 -60.078 54.799
510531 91.819 -60.087 54.781
510532 91.855 -60.097 54.768
510533 91.876 -60.108 54.757
510534 91.894 -60.120 54.744
510535 91.914 -60.131 54.731
510536 91.932 -60.143 54.720
510537 91.951 -60.155 54.710
510538 91.970 -60.166 54.697
510539 91.989 -60.178 54.676
510540 92.007 -60.190 54.650
510541 92.046 -60.202 54.619
510542 92.091 -60.212 54.592
510543 92.136 -60.222 54.573
510544 92.180 -60.231 54.557
510545 92.225 -60.241 54.543
510546 92.271 -60.251 54.533
510547 92.315 -60.261 54.525
510548 92.360 -60.270 54.518
510549 92.405 -60.280 54.514
510550 92.450 -60.290 54.513
510551 92.495 -60.299 54.515
510552 92.540 -60.309 54.520
510553 92.585 -60.318 54.527
510554 92.630 -60.328 54.534
510555 92.674 -60.337 54.538
510556 92.719 -60.347 54.531
510557 92.765 -60.357 54.521
510558 92.809 -60.367 54.511
510559 92.854 -60.376 54.499
510560 92.899 -60.386 54.485
510561 92.944 -60.396 54.466
510562 92.988 -60.404 54.450
510563 93.029 -60.410 54.436
510564 93.070 -60.416 54.423
510565 93.111 -60.422 54.409
510566 93.151 -60.428 54.399
510567 93.193 -60.434 54.400
510568 93.233 -60.439 54.405
510569 93.275 -60.450 54.408
510570 93.317 -60.463 54.406
510571 93.359 -60.475 54.401
510572 93.401 -60.488 54.396
510573 93.443 -60.500 54.389
510574 93.485 -60.513 54.380
510575 93.525 -60.527 54.369
510576 93.566 -60.542 54.362
510577 93.607 -60.557 54.346
510578 93.648 -60.572 54.328
510579 93.689 -60.587 54.314
510580 93.729 -60.602 54.305
510581 93.770 -60.617 54.300
510582 93.811 -60.632 54.290
510583 93.853 -60.645 54.279
510584 93.897 -60.657 54.265
510585 93.940 -60.668 54.247
510586 93.983 -60.680 54.228
510587 94.027 -60.691 54.214
510588 94.069 -60.703 54.198
510589 94.111 -60.714 54.178
510590 94.152 -60.726 54.156
510591 94.194 -60.737 54.133
510592 94.236 -60.748 54.118
510593 94.279 -60.758 54.104
510594 94.322 -60.768 54.086
510595 94.365 -60.777 54.066
510596 94.407 -60.785 54.045
510597 94.446 -60.790 54.027
510598 94.485 -60.795 54.010
510599 94.524 -60.799 53.991
510600 94.563 -60.804 53.968
510601 94.600 -60.808 53.940
510602 94.636 -60.814 53.912
510603 94.673 -60.818 53.884
510604 94.710 -60.823 53.863
510605 94.747 -60.827 53.845
510606 94.783 -60.832 53.828
510607 94.820 -60.836 53.810
510608 94.857 -60.842 53.787
510609 94.893 -60.848 53.762
510610 94.928 -60.853 53.733
510611 94.964 -60.858 53.704
510612 94.999 -60.864 53.673
510613 95.035 -60.869 53.636
510614 95.070 -60.874 53.597
510615 95.106 -60.879 53.555
510616 95.141 -60.885 53.512
510617 95.176 -60.890 53.468
510618 95.212 -60.896 53.424
510619 95.249 -60.903 53.378
510620 95.285 -60.909 53.333
510621 95.321 -60.914 53.289
510622 95.357 -60.920 53.244
510623 95.392 -60.927 53.199
510624 95.423 -60.939 53.157
510625 95.457 -60.946 53.112
510626 95.494 -60.950 53.068
510627 95.531 -60.954 53.024
510628 95.568 -60.958 52.980
510629 95.605 -60.961 52.939
510630 95.642 -60.965 52.898
510631 95.679 -60.969 52.856
510632 95.716 -60.973 52.814
510633 95.755 -60.976 52.772
510634 95.793 -60.980 52.731
510635 95.827 -60.984 52.693
510636 95.868 -60.989 52.656
510637 95.908 -60.994 52.618
510638 95.950 -60.999 52.580
510639 95.990 -61.005 52.543
510640 96.031 -61.009 52.512
510641 96.072 -61.014 52.484
510642 96.113 -61.019 52.458
510643 96.159 -61.021 52.431
510644 96.208 -61.024 52.403
510645 96.243 -61.032 52.377
510646 96.278 -61.039 52.349
510647 96.315 -61.040 52.320
510648 96.352 -61.042 52.290
510649 96.388 -61.043 52.257
510650 96.425 -61.045 52.218
510651 96.466 -61.043 52.173
510652 96.513 -61.039 52.129
510653 96.559 -61.033 52.085
510654 96.606 -61.029 52.043
510655 96.653 -61.024 52.003
510656 96.700 -61.019 51.963
510657 96.741 -61.027 51.924
510658 96.780 -61.042 51.887
510659 96.815 -61.056 51.851
510660 96.849 -61.071 51.816
510661 96.883 -61.085 51.783
510662 96.917 -61.099 51.748
510663 96.951 -61.114 51.715
510664 96.986 -61.122 51.681
510665 97.023 -61.127 51.647
510666 97.058 -61.133 51.614
510667 97.093 -61.125 51.579
510668 97.128 -61.116 51.545
510669 97.163 -61.106 51.510
510670 97.192 -61.109 51.476
510671 97.220 -61.124 51.441
510672 97.248 -61.141 51.406
510673 97.276 -61.153 51.369
510674 97.302 -61.141 51.330
510675 97.328 -61.129 51.290
510676 97.371 -61.121 51.251
510677 97.414 -61.114 51.213
510678 97.457 -61.105 51.175
510679 97.500 -61.098 51.137
510680 97.546 -61.096 51.099
510681 97.595 -61.101 51.062
510682 97.644 -61.105 51.025
510683 97.692 -61.110 50.988
510684 97.741 -61.114 50.951
510685 97.790 -61.119 50.914
510686 97.838 -61.124 50.877
510687 97.887 -61.129 50.841
510688 97.910 -61.136 50.805
510689 97.913 -61.143 50.768
510690 97.919 -61.152 50.731
510691 97.943 -61.168 50.693
510692 97.967 -61.183 50.656
510693 97.992 -61.199 50.620
510694 98.016 -61.215 50.584
511147 88.465 -58.911 56.067
511148 88.517 -58.918 56.046
511149 88.569 -58.924 56.026
511150 88.621 -58.931 56.006
511151 88.669 -58.941 55.986
511152 88.714 -58.952 55.966
511153 88.760 -58.965 55.946
511154 88.806 -58.977 55.927
511155 88.851 -58.989 55.907
511156 88.896 -59.000 55.887
511157 88.942 -59.012 55.866
511158 88.979 -59.024 55.843
511159 89.015 -59.035 55.821
511160 89.050 -59.046 55.798
511161 89.084 -59.062 55.777
511162 89.117 -59.077 55.754
511163 89.150 -59.093 55.731
511164 89.183 -59.109 55.710
511165 89.216 -59.125 55.690
511166 89.249 -59.141 55.673
511167 89.282 -59.157 55.661
511168 89.315 -59.173 55.651
511169 89.348 -59.188 55.646
511170 89.382 -59.204 55.648
511171 89.414 -59.220 55.657
511172 89.448 -59.237 55.666
511173 89.486 -59.251 55.669
511174 89.525 -59.266 55.670
511175 89.563 -59.281 55.673
511176 89.601 -59.297 55.675
511177 89.640 -59.311 55.673
511178 89.678 -59.326 55.671
511179 89.717 -59.341 55.677
511180 89.755 -59.356 55.686
511181 89.795 -59.372 55.696
511182 89.836 -59.388 55.710
511183 89.876 -59.403 55.725
511184 89.919 -59.419 55.739
511185 89.963 -59.433 55.756
511186 90.007 -59.447 55.773
511187 90.051 -59.462 55.788
511188 90.096 -59.476 55.798
511189 90.140 -59.491 55.806
511190 90.184 -59.505 55.800
511191 90.229 -59.519 55.783
511192 90.272 -59.533 55.763
511193 90.317 -59.547 55.742
511194 90.361 -59.562 55.715
511195 90.405 -59.576 55.671
511196 90.450 -59.590 55.619
511197 90.494 -59.602 55.561
511198 90.539 -59.615 55.499
511199 90.583 -59.628 55.446
511200 90.627 -59.641 55.396
511201 90.672 -59.653 55.351
511202 90.716 -59.666 55.308
511203 90.761 -59.679 55.260
511204 90.805 -59.691 55.212
511205 90.850 -59.704 55.169
511206 90.894 -59.717 55.132
511207 90.939 -59.729 55.098
511208 90.976 -59.747 55.069
511209 91.013 -59.765 55.045
511210 91.050 -59.784 55.027
511211 91.087 -59.802 55.010
511212 91.123 -59.820 54.993
511213 91.160 -59.839 54.977
511214 91.197 -59.858 54.966
511215 91.233 -59.877 54.954
511216 91.271 -59.893 54.943
511217 91.308 -59.908 54.934
511218 91.343 -59.919 54.926
511219 91.378 -59.929 54.915
511220 91.413 -59.938 54.906
511221 91.448 -59.948 54.901
511222 91.483 -59.957 54.896
511223 91.518 -59.967 54.890
511224 91.554 -59.976 54.884
511225 91.588 -59.985 54.875
511226 91.624 -59.995 54.863
511227 91.658 -60.004 54.849
511228 91.694 -60.014 54.835
511229 91.729 -60.023 54.818
511230 91.764 -60.033 54.798
511231 91.799 -60.042 54.780
511232 91.834 -60.052 54.766
511233 91.860 -60.062 54.756
511234 91.879 -60.074 54.750
511235 91.897 -60.085 54.746
511236 91.917 -60.097 54.741
511237 91.935 -60.108 54.734
511238 91.954 -60.121 54.725
511239 91.973 -60.132 54.712
511240 91.992 -60.143 54.691
511241 92.018 -60.150 54.662
511242 92.058 -60.155 54.629
511243 92.100 -60.165 54.594
511244 92.143 -60.174 54.565
511245 92.186 -60.184 54.549
511246 92.229 -60.193 54.538
511247 92.275 -60.203 54.525
511248 92.319 -60.213 54.513
511249 92.364 -60.223 54.502
511250 92.409 -60.232 54.491
511251 92.454 -60.242 54.492
511252 92.499 -60.252 54.500
511253 92.544 -60.261 54.500
511254 92.589 -60.271 54.498
511255 92.634 -60.281 54.502
511256 92.678 -60.290 54.511
511257 92.723 -60.300 54.505
511258 92.768 -60.309 54.496
511259 92.813 -60.319 54.486
511260 92.858 -60.329 54.473
511261 92.903 -60.338 54.461
511262 92.948 -60.348 54.445
511263 92.993 -60.358 54.430
511264 93.036 -60.365 54.418
511265 93.076 -60.371 54.409
511266 93.117 -60.377 54.402
511267 93.158 -60.383 54.400
511268 93.199 -60.388 54.406
511269 93.240 -60.397 54.412
511270 93.282 -60.410 54.415
511271 93.324 -60.422 54.413
511272 93.366 -60.434 54.405
511273 93.408 -60.447 54.397
511274 93.449 -60.461 54.388
511275 93.490 -60.476 54.376
511276 93.531 -60.491 54.365
511277 93.571 -60.506 54.353
511278 93.612 -60.522 54.336
511279 93.653 -60.537 54.318
511280 93.693 -60.552 54.301
511281 93.734 -60.566 54.287
511282 93.775 -60.582 54.277
511283 93.816 -60.597 54.269
511284 93.858 -60.609 54.260
511285 93.902 -60.620 54.249
511286 93.945 -60.632 54.237
511287 93.989 -60.644 54.228
511288 94.032 -60.655 54.217
511289 94.075 -60.667 54.202
511290 94.118 -60.679 54.184
511291 94.159 -60.690 54.163
511292 94.201 -60.701 54.141
511293 94.243 -60.713 54.125
511294 94.286 -60.723 54.110
511295 94.329 -60.732 54.092
511296 94.372 -60.741 54.074
511297 94.411 -60.746 54.059
511298 94.450 -60.751 54.043
511299 94.489 -60.755 54.024
511300 94.528 -60.760 54.004
511301 94.567 -60.764 53.980
511302 94.603 -60.769 53.954
511303 94.640 -60.774 53.927
511304 94.677 -60.779 53.905
511305 94.713 -60.783 53.889
511306 94.750 -60.788 53.875
511307 94.787 -60.792 53.858
511308 94.826 -60.798 53.838
511309 94.863 -60.803 53.814
511310 94.899 -60.810 53.787
511311 94.934 -60.817 53.759
511312 94.970 -60.823 53.730
511313 95.005 -60.828 53.700
511314 95.041 -60.833 53.664
511315 95.076 -60.839 53.624
511316 95.112 -60.844 53.583
511317 95.147 -60.849 53.540
511318 95.183 -60.854 53.495
511319 95.219 -60.861 53.449
511320 95.255 -60.867 53.401
511321 95.292 -60.874 53.353
511322 95.328 -60.879 53.306
511323 95.365 -60.885 53.260
511324 95.398 -60.894 53.215
511325 95.429 -60.905 53.170
511326 95.466 -60.909 53.124
511327 95.503 -60.913 53.079
511328 95.540 -60.917 53.034
511329 95.577 -60.920 52.990
511330 95.614 -60.924 52.946
511331 95.651 -60.928 52.904
511332 95.688 -60.932 52.861
511333 95.725 -60.936 52.818
511334 95.761 -60.941 52.777
511335 95.801 -60.945 52.737
511336 95.842 -60.950 52.697
511337 95.883 -60.955 52.660
511338 95.924 -60.960 52.622
511339 95.966 -60.965 52.583
511340 96.007 -60.970 52.547
511341 96.048 -60.975 52.515
511342 96.090 -60.980 52.486
511343 96.131 -60.986 52.460
511344 96.172 -60.990 52.434
511345 96.214 -60.995 52.406
511346 96.252 -60.998 52.378
511347 96.289 -60.999 52.349
511348 96.325 -61.001 52.321
511349 96.362 -61.002 52.293
511350 96.398 -61.005 52.264
511351 96.435 -61.006 52.229
511352 96.472 -61.008 52.188
511353 96.512 -61.007 52.145
511354 96.559 -61.002 52.101
511355 96.605 -60.997 52.057
511356 96.652 -60.992 52.015
511357 96.699 -60.987 51.974
511358 96.742 -60.991 51.934
511359 96.781 -61.005 51.896
511360 96.819 -61.020 51.859
511361 96.858 -61.033 51.824
511362 96.896 -61.048 51.789
511363 96.931 -61.062 51.753
511364 96.964 -61.077 51.719
511365 96.999 -61.089 51.685
511366 97.035 -61.095 51.651
511367 97.070 -61.091 51.618
511368 97.105 -61.081 51.582
511369 97.140 -61.072 51.547
511370 97.174 -61.062 51.512
511371 97.203 -61.074 51.476
511372 97.231 -61.091 51.443
511373 97.258 -61.108 51.408
511374 97.286 -61.116 51.372
511375 97.312 -61.104 51.334
511376 97.338 -61.092 51.294
511377 97.373 -61.082 51.255
511378 97.416 -61.074 51.217
511379 97.459 -61.066 51.180
511380 97.503 -61.058 51.142
511381 97.550 -61.061 51.103
511382 97.599 -61.067 51.067
511383 97.648 -61.071 51.030
511384 97.697 -61.076 50.993
511385 97.745 -61.080 50.955
511386 97.794 -61.085 50.918
511387 97.843 -61.089 50.881
511388 97.891 -61.094 50.845
511389 97.914 -61.101 50.808
511390 97.918 -61.109 50.772
511391 97.921 -61.118 50.735
511392 97.945 -61.133 50.697
511393 97.969 -61.149 50.659
511394 97.993 -61.165 50.623
511395 98.017 -61.180 50.587
511396 98.041 -61.196 50.551
511823 87.341 -58.780 56.862
511824 87.383 -58.779 56.828
511825 87.425 -58.777 56.793
511826 87.468 -58.776 56.757
511827 87.510 -58.774 56.720
511828 87.553 -58.772 56.683
511829 87.595 -58.770 56.644
511830 87.637 -58.769 56.606
511831 87.674 -58.771 56.568
511832 87.711 -58.773 56.529
511833 87.747 -58.775 56.490
511848 88.491 -58.866 56.065
511849 88.544 -58.874 56.044
511850 88.591 -58.883 56.023
511851 88.636 -58.896 56.003
511852 88.682 -58.908 55.983
511853 88.728 -58.919 55.963
511854 88.773 -58.931 55.943
511855 88.818 -58.943 55.924
511856 88.863 -58.955 55.905
511857 88.909 -58.967 55.886
511858 88.954 -58.979 55.866
511859 88.989 -58.990 55.844
511860 89.025 -59.001 55.822
511861 89.060 -59.012 55.801
511862 89.094 -59.026 55.780
511863 89.127 -59.041 55.759
511864 89.161 -59.057 55.738
511865 89.194 -59.073 55.718
511866 89.227 -59.089 55.699
511867 89.260 -59.105 55.683
511868 89.293 -59.121 55.670
511869 89.326 -59.137 55.661
511870 89.359 -59.153 55.656
511871 89.392 -59.168 55.658
511872 89.425 -59.184 55.668
511873 89.460 -59.196 55.681
511874 89.500 -59.212 55.690
511875 89.540 -59.228 55.697
511876 89.579 -59.244 55.703
511877 89.617 -59.258 55.706
511878 89.655 -59.273 55.703
511879 89.693 -59.288 55.704
511880 89.732 -59.303 55.712
511881 89.770 -59.318 55.722
511882 89.809 -59.333 55.734
511883 89.850 -59.349 55.749
511884 89.891 -59.365 55.760
511885 89.934 -59.380 55.770
511886 89.978 -59.394 55.782
511887 90.022 -59.409 55.795
511888 90.067 -59.423 55.811
511889 90.111 -59.437 55.824
511890 90.155 -59.451 55.831
511891 90.199 -59.466 55.825
511892 90.244 -59.480 55.808
511893 90.288 -59.494 55.784
511894 90.332 -59.509 55.755
511895 90.376 -59.523 55.726
511896 90.421 -59.538 55.684
511897 90.465 -59.551 55.632
511898 90.509 -59.566 55.574
511899 90.553 -59.579 55.516
511900 90.598 -59.591 55.461
511901 90.642 -59.604 55.409
511902 90.687 -59.616 55.362
511903 90.731 -59.629 55.319
511904 90.776 -59.642 55.271
511905 90.820 -59.654 55.219
511906 90.865 -59.667 55.175
511907 90.909 -59.680 55.135
511908 90.952 -59.694 55.100
511909 90.989 -59.712 55.069
511910 91.025 -59.731 55.043
511911 91.062 -59.749 55.022
511912 91.099 -59.768 55.003
511913 91.136 -59.786 54.984
511914 91.172 -59.805 54.968
511915 91.209 -59.823 54.955
511916 91.246 -59.838 54.942
511917 91.283 -59.854 54.932
511918 91.320 -59.869 54.923
511919 91.357 -59.883 54.915
511920 91.392 -59.892 54.904
511921 91.427 -59.902 54.894
511922 91.462 -59.911 54.887
511923 91.497 -59.921 54.882
511924 91.532 -59.930 54.877
511925 91.568 -59.940 54.869
511926 91.602 -59.949 54.858
511927 91.638 -59.959 54.842
511928 91.672 -59.968 54.824
511929 91.708 -59.977 54.806
511930 91.743 -59.987 54.789
511931 91.778 -59.996 54.773
511932 91.813 -60.006 54.759
511933 91.844 -60.016 54.750
511934 91.863 -60.027 54.744
511935 91.882 -60.040 54.741
511936 91.900 -60.051 54.750
511937 91.920 -60.062 54.752
511938 91.938 -60.074 54.748
511939 91.957 -60.086 54.738
511940 91.977 -60.096 54.722
511941 92.005 -60.102 54.699
511942 92.032 -60.108 54.669
511943 92.075 -60.108 54.632
511944 92.118 -60.118 54.594
511945 92.160 -60.127 54.560
511946 92.203 -60.137 54.542
511947 92.245 -60.146 54.535
511948 92.287 -60.155 54.525
511949 92.330 -60.165 54.514
511950 92.372 -60.175 54.501
511951 92.415 -60.185 54.486
511952 92.458 -60.194 54.476
511953 92.503 -60.204 54.475
511954 92.548 -60.214 54.478
511955 92.593 -60.224 54.480
511956 92.638 -60.233 54.476
511957 92.683 -60.243 54.479
511958 92.727 -60.253 54.484
511959 92.772 -60.262 54.482
511960 92.817 -60.271 54.472
511961 92.862 -60.281 54.457
511962 92.907 -60.291 54.443
511963 92.952 -60.300 54.429
511964 92.997 -60.310 54.413
511965 93.042 -60.320 54.401
511966 93.083 -60.326 54.398
511967 93.124 -60.331 54.402
511968 93.164 -60.337 54.409
511969 93.206 -60.344 54.414
511970 93.248 -60.357 54.419
511971 93.289 -60.369 54.422
511972 93.332 -60.382 54.422
511973 93.372 -60.396 54.415
511974 93.413 -60.411 54.408
511975 93.454 -60.426 54.397
511976 93.495 -60.441 54.384
511977 93.536 -60.456 54.371
511978 93.576 -60.471 54.357
511979 93.617 -60.486 54.340
511980 93.657 -60.501 54.321
511981 93.698 -60.516 54.302
511982 93.739 -60.532 54.283
511983 93.780 -60.546 54.268
511984 93.821 -60.561 54.255
511985 93.864 -60.573 54.244
511986 93.907 -60.585 54.232
511987 93.951 -60.596 54.222
511988 93.994 -60.607 54.216
511989 94.038 -60.619 54.207
511990 94.081 -60.631 54.195
511991 94.124 -60.642 54.181
511992 94.167 -60.654 54.164
511993 94.208 -60.666 54.145
511994 94.250 -60.677 54.127
511995 94.293 -60.686 54.112
511996 94.336 -60.696 54.097
511997 94.376 -60.701 54.082
511998 94.415 -60.707 54.068
511999 94.454 -60.711 54.051
512000 94.493 -60.716 54.034
512001 94.532 -60.720 54.013
512002 94.570 -60.725 53.990
512003 94.607 -60.729 53.968
512004 94.644 -60.735 53.949
512005 94.680 -60.739 53.931
512006 94.717 -60.744 53.916
512007 94.755 -60.748 53.901
512008 94.794 -60.754 53.883
512009 94.832 -60.759 53.861
512010 94.870 -60.764 53.837
512011 94.905 -60.771 53.809
512012 94.941 -60.779 53.780
512013 94.976 -60.785 53.751
512014 95.012 -60.792 53.722
512015 95.047 -60.798 53.685
512016 95.083 -60.803 53.646
512017 95.118 -60.808 53.605
512018 95.154 -60.814 53.563
512019 95.189 -60.819 53.518
512020 95.226 -60.826 53.471
512021 95.262 -60.832 53.422
512022 95.299 -60.839 53.373
512023 95.335 -60.845 53.325
512024 95.371 -60.851 53.278
512025 95.404 -60.860 53.233
512026 95.438 -60.868 53.187
512027 95.475 -60.872 53.139
512028 95.512 -60.876 53.092
512029 95.548 -60.879 53.046
512030 95.586 -60.883 53.000
512031 95.622 -60.887 52.955
512032 95.660 -60.891 52.912
512033 95.696 -60.897 52.868
512034 95.732 -60.903 52.825
512035 95.768 -60.909 52.782
512036 95.808 -60.908 52.741
512037 95.849 -60.913 52.702
512038 95.891 -60.918 52.663
512039 95.932 -60.923 52.624
512040 95.973 -60.928 52.586
512041 96.014 -60.933 52.550
512042 96.055 -60.939 52.517
512043 96.097 -60.944 52.489
512044 96.138 -60.948 52.463
512045 96.179 -60.954 52.436
512046 96.219 -60.958 52.408
512047 96.257 -60.960 52.379
512048 96.297 -60.961 52.351
512049 96.336 -60.962 52.324
512050 96.373 -60.964 52.297
512051 96.409 -60.965 52.269
512052 96.446 -60.967 52.237
512053 96.482 -60.968 52.201
512054 96.519 -60.970 52.161
512055 96.558 -60.970 52.119
512056 96.605 -60.964 52.076
512057 96.652 -60.960 52.032
512058 96.698 -60.955 51.989
512059 96.743 -60.955 51.948
512060 96.782 -60.969 51.909
512061 96.821 -60.983 51.871
512062 96.859 -60.997 51.834
512063 96.898 -61.011 51.799
512064 96.937 -61.026 51.763
512065 96.975 -61.039 51.728
512066 97.012 -61.054 51.693
512067 97.047 -61.057 51.658
512068 97.082 -61.047 51.625
512069 97.118 -61.033 51.590
512070 97.156 -61.011 51.554
512071 97.187 -61.014 51.517
512072 97.213 -61.041 51.481
512073 97.241 -61.058 51.446
512074 97.269 -61.077 51.410
512075 97.296 -61.078 51.374
512076 97.322 -61.066 51.337
512077 97.349 -61.054 51.298
512078 97.375 -61.042 51.260
512079 97.418 -61.034 51.222
512080 97.461 -61.027 51.185
512081 97.507 -61.023 51.147
512082 97.555 -61.027 51.108
512083 97.604 -61.032 51.071
512084 97.653 -61.036 51.036
512085 97.701 -61.042 50.998
512086 97.750 -61.046 50.960
512087 97.799 -61.051 50.922
512088 97.847 -61.055 50.885
512089 97.896 -61.060 50.849
512090 97.918 -61.067 50.811
512091 97.922 -61.075 50.775
512092 97.926 -61.083 50.738
512093 97.947 -61.098 50.700
512094 97.971 -61.113 50.662
512095 97.995 -61.129 50.626
512096 98.019 -61.145 50.589
512097 98.043 -61.160 50.553
512098 98.068 -61.176 50.517
512522 87.272 -58.735 56.937
512523 87.315 -58.733 56.905
512524 87.357 -58.732 56.872
512525 87.400 -58.730 56.838
512526 87.442 -58.729 56.804
512527 87.484 -58.726 56.768
512528 87.527 -58.725 56.731
512529 87.569 -58.723 56.693
512530 87.612 -58.722 56.655
512531 87.652 -58.721 56.615
512532 87.689 -58.723 56.577
512533 87.726 -58.725 56.537
512534 87.763 -58.726 56.498
512535 87.799 -58.729 56.460
512536 87.842 -58.733 56.423
512549 88.513 -58.827 56.063
512550 88.558 -58.838 56.042
512551 88.604 -58.850 56.021
512552 88.649 -58.862 56.001
512553 88.695 -58.874 55.981
512554 88.740 -58.886 55.960
512555 88.785 -58.898 55.940
512556 88.831 -58.910 55.921
512557 88.877 -58.921 55.903
512558 88.922 -58.933 55.884
512559 88.963 -58.945 55.864
512560 88.999 -58.956 55.843
512561 89.034 -58.968 55.823
512562 89.070 -58.979 55.803
512563 89.105 -58.990 55.783
512564 89.138 -59.005 55.764
512565 89.172 -59.021 55.744
512566 89.204 -59.037 55.725
512567 89.238 -59.053 55.707
512568 89.271 -59.069 55.692
512569 89.304 -59.085 55.680
512570 89.337 -59.101 55.670
512571 89.370 -59.117 55.665
512572 89.403 -59.133 55.665
512573 89.437 -59.149 55.671
512574 89.472 -59.155 55.683
512575 89.512 -59.171 55.697
512576 89.552 -59.187 55.708
512577 89.592 -59.203 55.717
512578 89.631 -59.219 55.724
512579 89.671 -59.235 55.726
512580 89.709 -59.250 55.734
512581 89.747 -59.265 55.748
512582 89.786 -59.280 55.763
512583 89.824 -59.295 55.780
512584 89.864 -59.311 55.799
512585 89.905 -59.327 55.806
512586 89.949 -59.341 55.812
512587 89.993 -59.355 55.817
512588 90.037 -59.369 55.823
512589 90.082 -59.384 55.839
512590 90.126 -59.398 55.849
512591 90.170 -59.413 55.851
512592 90.215 -59.427 55.845
512593 90.259 -59.441 55.829
512594 90.303 -59.456 55.802
512595 90.347 -59.469 55.768
512596 90.391 -59.484 55.733
512597 90.436 -59.498 55.691
512598 90.480 -59.513 55.644
512599 90.524 -59.527 55.590
512600 90.569 -59.541 55.529
512601 90.613 -59.555 55.472
512602 90.657 -59.568 55.417
512603 90.701 -59.580 55.367
512604 90.746 -59.593 55.320
512605 90.791 -59.606 55.271
512606 90.835 -59.619 55.222
512607 90.879 -59.631 55.176
512608 90.924 -59.644 55.135
512609 90.964 -59.660 55.099
512610 91.001 -59.678 55.067
512611 91.037 -59.697 55.039
512612 91.074 -59.715 55.015
512613 91.111 -59.734 54.993
512614 91.148 -59.752 54.975
512615 91.185 -59.767 54.959
512616 91.222 -59.783 54.944
512617 91.259 -59.798 54.931
512618 91.296 -59.814 54.920
512619 91.333 -59.830 54.912
512620 91.371 -59.845 54.904
512621 91.406 -59.857 54.893
512622 91.441 -59.866 54.882
512623 91.476 -59.876 54.873
512624 91.512 -59.885 54.866
512625 91.546 -59.895 54.859
512626 91.582 -59.904 54.850
512627 91.616 -59.914 54.838
512628 91.652 -59.923 54.821
512629 91.687 -59.933 54.803
512630 91.722 -59.942 54.785
512631 91.757 -59.952 54.768
512632 91.792 -59.961 54.752
512633 91.827 -59.970 54.740
512634 91.847 -59.982 54.734
512635 91.866 -59.993 54.735
512636 91.885 -60.005 54.744
512637 91.904 -60.017 54.753
512638 91.923 -60.028 54.756
512639 91.941 -60.040 54.753
512640 91.964 -60.049 54.743
512641 91.992 -60.054 54.727
512642 92.019 -60.059 54.700
512643 92.047 -60.065 54.668
512644 92.089 -60.066 54.629
512645 92.135 -60.070 54.591
512646 92.177 -60.080 54.559
512647 92.219 -60.089 54.538
512648 92.262 -60.099 54.532
512649 92.304 -60.108 54.526
512650 92.347 -60.118 54.517
512651 92.389 -60.127 54.503
512652 92.432 -60.137 54.487
512653 92.474 -60.147 54.469
512654 92.516 -60.156 54.460
512655 92.559 -60.166 54.458
512656 92.601 -60.176 54.458
512657 92.644 -60.186 54.456
512658 92.687 -60.195 54.455
512659 92.731 -60.205 54.457
512660 92.776 -60.215 54.456
512661 92.821 -60.224 54.451
512662 92.866 -60.234 54.442
512663 92.911 -60.243 54.429
512664 92.956 -60.253 54.417
512665 93.001 -60.262 54.408
512666 93.046 -60.272 54.403
512667 93.089 -60.281 54.404
512668 93.130 -60.287 54.408
512669 93.171 -60.292 54.421
512670 93.213 -60.304 54.425
512671 93.255 -60.316 54.427
512672 93.296 -60.331 54.430
512673 93.337 -60.345 54.434
512674 93.377 -60.360 54.431
512675 93.418 -60.375 54.424
512676 93.459 -60.391 54.415
512677 93.500 -60.406 54.403
512678 93.540 -60.421 54.390
512679 93.581 -60.435 54.376
512680 93.622 -60.450 54.359
512681 93.662 -60.466 54.338
512682 93.703 -60.481 54.315
512683 93.744 -60.496 54.292
512684 93.784 -60.511 54.271
512685 93.826 -60.525 54.250
512686 93.869 -60.537 54.233
512687 93.913 -60.548 54.218
512688 93.956 -60.560 54.206
512689 93.999 -60.572 54.198
512690 94.043 -60.583 54.192
512691 94.086 -60.594 54.183
512692 94.130 -60.607 54.172
512693 94.173 -60.618 54.159
512694 94.215 -60.629 54.144
512695 94.257 -60.641 54.127
512696 94.300 -60.651 54.111
512697 94.341 -60.657 54.098
512698 94.380 -60.662 54.085
512699 94.419 -60.666 54.072
512700 94.458 -60.672 54.056
512701 94.496 -60.676 54.039
512702 94.536 -60.681 54.021
512703 94.574 -60.685 54.003
512704 94.611 -60.690 53.986
512705 94.647 -60.695 53.970
512706 94.685 -60.700 53.952
512707 94.724 -60.705 53.936
512708 94.762 -60.710 53.918
512709 94.801 -60.715 53.897
512710 94.839 -60.720 53.875
512711 94.876 -60.726 53.848
512712 94.912 -60.733 53.820
512713 94.947 -60.740 53.791
512714 94.983 -60.747 53.761
512715 95.018 -60.754 53.730
512716 95.054 -60.761 53.692
512717 95.089 -60.767 53.654
512718 95.125 -60.773 53.614
512719 95.160 -60.778 53.573
512720 95.196 -60.783 53.530
512721 95.232 -60.790 53.486
512722 95.269 -60.797 53.439
512723 95.305 -60.804 53.393
512724 95.342 -60.810 53.348
512725 95.378 -60.816 53.303
512726 95.410 -60.826 53.257
512727 95.446 -60.831 53.209
512728 95.483 -60.835 53.159
512729 95.520 -60.839 53.109
512730 95.557 -60.842 53.060
512731 95.594 -60.846 53.012
512732 95.630 -60.852 52.967
512733 95.667 -60.858 52.922
512734 95.703 -60.865 52.877
512735 95.740 -60.871 52.832
512736 95.776 -60.875 52.789
512737 95.815 -60.871 52.746
512738 95.856 -60.876 52.706
512739 95.898 -60.882 52.666
512740 95.939 -60.886 52.627
512741 95.980 -60.892 52.588
512742 96.022 -60.897 52.552
512743 96.063 -60.901 52.521
512744 96.104 -60.907 52.492
512745 96.145 -60.912 52.466
512746 96.186 -60.917 52.439
512747 96.223 -60.920 52.411
512748 96.261 -60.922 52.382
512749 96.301 -60.923 52.354
512750 96.340 -60.924 52.327
512751 96.379 -60.925 52.301
512752 96.419 -60.926 52.274
512753 96.456 -60.927 52.245
512754 96.493 -60.929 52.212
512755 96.530 -60.930 52.175
512756 96.566 -60.932 52.135
512757 96.604 -60.933 52.093
512758 96.651 -60.928 52.050
512759 96.698 -60.923 52.007
512760 96.744 -60.918 51.965
512761 96.783 -60.933 51.925
512762 96.822 -60.947 51.886
512763 96.861 -60.961 51.848
512764 96.899 -60.975 51.811
512765 96.938 -60.989 51.775
512766 96.976 -61.003 51.739
512767 97.022 -61.011 51.704
512768 97.071 -61.016 51.670
512769 97.111 -60.980 51.637
512770 97.150 -60.944 51.602
512771 97.185 -60.922 51.565
512772 97.206 -60.957 51.528
512773 97.227 -60.997 51.491
512774 97.251 -61.027 51.454
512775 97.280 -61.044 51.417
512776 97.306 -61.041 51.380
512777 97.332 -61.029 51.343
512778 97.359 -61.017 51.305
512779 97.385 -61.005 51.268
512780 97.421 -60.995 51.230
512781 97.464 -60.987 51.192
512782 97.511 -60.989 51.154
512783 97.560 -60.993 51.116
512784 97.608 -60.998 51.079
512785 97.657 -61.002 51.042
512786 97.706 -61.007 51.005
512787 97.754 -61.012 50.968
512788 97.803 -61.017 50.930
512789 97.852 -61.021 50.892
512790 97.900 -61.026 50.855
512791 97.923 -61.033 50.818
512792 97.927 -61.041 50.780
512793 97.930 -61.049 50.743
512794 97.948 -61.062 50.704
512795 97.972 -61.078 50.666
512796 97.997 -61.094 50.629
512797 98.021 -61.109 50.593
512798 98.045 -61.125 50.556
512799 98.069 -61.140 50.519
512800 98.093 -61.156 50.482
512801 98.119 -61.154 50.445
513222 87.247 -58.688 56.976
513223 87.289 -58.686 56.947
513224 87.331 -58.685 56.916
513225 87.374 -58.683 56.883
513226 87.416 -58.682 56.850
513227 87.459 -58.679 56.816
513228 87.501 -58.678 56.780
513229 87.543 -58.676 56.743
513230 87.586 -58.675 56.705
513231 87.628 -58.673 56.666
513232 87.668 -58.673 56.626
513233 87.704 -58.675 56.587
513234 87.741 -58.677 56.547
513235 87.778 -58.679 56.507
513236 87.817 -58.682 56.468
513237 87.869 -58.689 56.430
513238 87.921 -58.695 56.394
513250 88.526 -58.793 56.061
513251 88.571 -58.805 56.039
513252 88.617 -58.817 56.019
513253 88.662 -58.829 55.999
513254 88.707 -58.840 55.978
513255 88.753 -58.852 55.957
513256 88.798 -58.864 55.937
513257 88.844 -58.877 55.918
513258 88.889 -58.888 55.900
513259 88.934 -58.900 55.881
513260 88.973 -58.911 55.862
513261 89.009 -58.923 55.842
513262 89.044 -58.934 55.822
513263 89.080 -58.945 55.804
513264 89.115 -58.956 55.786
513265 89.149 -58.970 55.767
513266 89.182 -58.986 55.748
513267 89.215 -59.002 55.730
513268 89.249 -59.018 55.713
513269 89.281 -59.034 55.700
513270 89.315 -59.049 55.689
513271 89.348 -59.065 55.680
513272 89.381 -59.081 55.673
513273 89.414 -59.096 55.670
513274 89.448 -59.105 55.672
513275 89.484 -59.114 55.682
513276 89.524 -59.130 55.696
513277 89.564 -59.146 55.711
513278 89.604 -59.162 55.724
513279 89.643 -59.178 55.736
513280 89.683 -59.194 55.744
513281 89.723 -59.210 55.759
513282 89.762 -59.227 55.776
513283 89.801 -59.242 55.796
513284 89.840 -59.257 55.817
513285 89.878 -59.272 55.840
513286 89.920 -59.287 55.850
513287 89.964 -59.302 55.853
513288 90.009 -59.316 55.854
513289 90.052 -59.331 55.859
513290 90.097 -59.345 55.873
513291 90.141 -59.359 55.868
513292 90.185 -59.373 55.861
513293 90.230 -59.388 55.853
513294 90.274 -59.402 55.843
513295 90.318 -59.416 55.818
513296 90.362 -59.431 55.781
513297 90.406 -59.445 55.737
513298 90.451 -59.460 55.696
513299 90.495 -59.474 55.656
513300 90.539 -59.488 55.608
513301 90.584 -59.502 55.544
513302 90.628 -59.516 55.484
513303 90.672 -59.531 55.426
513304 90.717 -59.544 55.372
513305 90.761 -59.557 55.320
513306 90.805 -59.569 55.270
513307 90.850 -59.582 55.221
513308 90.894 -59.595 55.175
513309 90.939 -59.607 55.133
513310 90.976 -59.626 55.096
513311 91.013 -59.644 55.063
513312 91.049 -59.662 55.034
513313 91.086 -59.681 55.008
513314 91.123 -59.697 54.985
513315 91.160 -59.713 54.966
513316 91.198 -59.728 54.949
513317 91.235 -59.744 54.933
513318 91.272 -59.759 54.921
513319 91.309 -59.774 54.911
513320 91.346 -59.790 54.902
513321 91.383 -59.805 54.893
513322 91.421 -59.821 54.882
513323 91.455 -59.830 54.870
513324 91.491 -59.839 54.858
513325 91.526 -59.849 54.847
513326 91.561 -59.858 54.839
513327 91.596 -59.868 54.830
513328 91.631 -59.877 54.817
513329 91.666 -59.887 54.801
513330 91.701 -59.896 54.784
513331 91.736 -59.906 54.766
513332 91.771 -59.915 54.749
513333 91.806 -59.925 54.734
513334 91.831 -59.936 54.724
513335 91.850 -59.947 54.719
513336 91.869 -59.959 54.723
513337 91.888 -59.971 54.736
513338 91.907 -59.982 54.744
513339 91.926 -59.994 54.745
513340 91.951 -60.000 54.740
513341 91.979 -60.006 54.728
513342 92.007 -60.011 54.711
513343 92.034 -60.017 54.688
513344 92.062 -60.022 54.662
513345 92.104 -60.023 54.632
513346 92.151 -60.022 54.601
513347 92.194 -60.032 54.571
513348 92.236 -60.042 54.544
513349 92.279 -60.051 54.523
513350 92.321 -60.061 54.516
513351 92.364 -60.071 54.512
513352 92.406 -60.080 54.504
513353 92.448 -60.090 54.493
513354 92.491 -60.099 54.475
513355 92.533 -60.109 54.459
513356 92.576 -60.118 54.443
513357 92.619 -60.128 54.433
513358 92.661 -60.138 54.439
513359 92.703 -60.148 54.437
513360 92.745 -60.157 54.431
513361 92.788 -60.167 54.430
513362 92.831 -60.177 54.431
513363 92.873 -60.187 54.429
513364 92.916 -60.196 54.419
513365 92.960 -60.206 54.412
513366 93.005 -60.215 54.407
513367 93.050 -60.224 54.407
513368 93.095 -60.234 54.412
513369 93.137 -60.241 54.416
513370 93.178 -60.251 54.429
513371 93.219 -60.265 54.434
513372 93.260 -60.280 54.436
513373 93.301 -60.295 54.438
513374 93.342 -60.310 54.439
513375 93.382 -60.325 54.437
513376 93.423 -60.340 54.433
513377 93.464 -60.355 54.427
513378 93.504 -60.370 54.419
513379 93.545 -60.385 54.408
513380 93.586 -60.400 54.394
513381 93.627 -60.416 54.377
513382 93.667 -60.430 54.357
513383 93.708 -60.445 54.336
513384 93.749 -60.460 54.313
513385 93.789 -60.475 54.290
513386 93.831 -60.489 54.266
513387 93.874 -60.501 54.242
513388 93.918 -60.513 54.218
513389 93.961 -60.524 54.196
513390 94.005 -60.535 54.182
513391 94.048 -60.547 54.172
513392 94.091 -60.559 54.164
513393 94.135 -60.570 54.156
513394 94.178 -60.582 54.147
513395 94.222 -60.594 54.137
513396 94.264 -60.605 54.122
513397 94.306 -60.613 54.108
513398 94.345 -60.618 54.096
513399 94.384 -60.622 54.084
513400 94.422 -60.627 54.072
513401 94.462 -60.632 54.058
513402 94.500 -60.637 54.043
513403 94.539 -60.641 54.029
513404 94.577 -60.646 54.015
513405 94.615 -60.651 54.002
513406 94.654 -60.656 53.988
513407 94.692 -60.661 53.968
513408 94.731 -60.666 53.949
513409 94.769 -60.671 53.929
513410 94.808 -60.676 53.907
513411 94.846 -60.681 53.882
513412 94.882 -60.688 53.854
513413 94.918 -60.695 53.825
513414 94.954 -60.701 53.795
513415 94.989 -60.708 53.764
513416 95.025 -60.716 53.730
513417 95.060 -60.723 53.692
513418 95.096 -60.729 53.654
513419 95.131 -60.736 53.616
513420 95.167 -60.742 53.576
513421 95.202 -60.748 53.534
513422 95.239 -60.755 53.493
513423 95.275 -60.762 53.450
513424 95.312 -60.769 53.409
513425 95.348 -60.776 53.368
513426 95.385 -60.781 53.325
513427 95.418 -60.789 53.280
513428 95.455 -60.793 53.233
513429 95.492 -60.797 53.181
513430 95.529 -60.801 53.128
513431 95.565 -60.807 53.078
513432 95.602 -60.814 53.027
513433 95.638 -60.820 52.979
513434 95.674 -60.826 52.933
513435 95.710 -60.833 52.886
513436 95.747 -60.839 52.840
513437 95.784 -60.841 52.795
513438 95.823 -60.835 52.752
513439 95.864 -60.839 52.710
513440 95.905 -60.845 52.670
513441 95.946 -60.850 52.631
513442 95.987 -60.855 52.592
513443 96.029 -60.860 52.556
513444 96.070 -60.865 52.525
513445 96.111 -60.870 52.497
513446 96.153 -60.875 52.469
513447 96.191 -60.879 52.442
513448 96.228 -60.882 52.413
513449 96.265 -60.884 52.385
513450 96.305 -60.885 52.358
513451 96.344 -60.886 52.331
513452 96.384 -60.887 52.305
513453 96.423 -60.888 52.279
513454 96.462 -60.889 52.251
513455 96.502 -60.890 52.220
513456 96.540 -60.892 52.186
513457 96.577 -60.893 52.148
513458 96.613 -60.895 52.108
513459 96.650 -60.896 52.068
513460 96.697 -60.891 52.026
513461 96.744 -60.886 51.983
513462 96.785 -60.896 51.942
513463 96.823 -60.911 51.903
513464 96.862 -60.924 51.864
513465 96.900 -60.939 51.826
513466 96.941 -60.951 51.789
513467 96.991 -60.955 51.752
513468 97.039 -60.960 51.718
513469 97.089 -60.964 51.685
513470 97.137 -60.965 51.651
513471 97.176 -60.929 51.616
513472 97.205 -60.923 51.579
513473 97.225 -60.942 51.541
513474 97.246 -60.961 51.503
513475 97.266 -60.980 51.464
513476 97.290 -61.011 51.426
513477 97.316 -61.003 51.389
513478 97.343 -60.991 51.351
513479 97.369 -60.979 51.314
513480 97.395 -60.967 51.278
513481 97.423 -60.955 51.240
513482 97.467 -60.949 51.202
513483 97.516 -60.955 51.163
513484 97.564 -60.959 51.125
513485 97.613 -60.964 51.089
513486 97.662 -60.968 51.050
513487 97.710 -60.973 51.013
513488 97.759 -60.977 50.976
513489 97.808 -60.982 50.938
513490 97.857 -60.987 50.899
513491 97.905 -60.992 50.862
513492 97.927 -60.998 50.825
513493 97.931 -61.007 50.787
513494 97.935 -61.014 50.748
513495 97.950 -61.027 50.710
513496 97.974 -61.042 50.671
513497 97.999 -61.058 50.634
513498 98.023 -61.074 50.597
513499 98.047 -61.089 50.560
513500 98.071 -61.105 50.523
513501 98.095 -61.121 50.485
513502 98.121 -61.110 50.448
513503 98.155 -61.111 50.412
513921 87.178 -58.642 57.041
513922 87.221 -58.641 57.014
513923 87.263 -58.639 56.986
513924 87.305 -58.638 56.958
513925 87.348 -58.636 56.928
513926 87.390 -58.635 56.896
513927 87.433 -58.632 56.864
513928 87.475 -58.631 56.830
513929 87.517 -58.629 56.793
513930 87.560 -58.628 56.756
513931 87.602 -58.626 56.718
513932 87.645 -58.625 56.679
513933 87.683 -58.626 56.638
513934 87.720 -58.627 56.599
513935 87.756 -58.629 56.557
513936 87.793 -58.631 56.517
513937 87.843 -58.637 56.477
513938 87.896 -58.645 56.438
513939 87.947 -58.651 56.401
513940 87.999 -58.658 56.364
513951 88.539 -58.759 56.059
513952 88.584 -58.771 56.038
513953 88.629 -58.783 56.017
513954 88.675 -58.795 55.996
513955 88.720 -58.807 55.974
513956 88.766 -58.819 55.954
513957 88.811 -58.831 55.935
513958 88.856 -58.842 55.916
513959 88.902 -58.855 55.897
513960 88.947 -58.867 55.878
513961 88.983 -58.878 55.859
513962 89.019 -58.889 55.841
513963 89.054 -58.900 55.822
513964 89.089 -58.911 55.804
513965 89.125 -58.923 55.787
513966 89.160 -58.933 55.770
513967 89.193 -58.949 55.752
513968 89.227 -58.965 55.735
513969 89.259 -58.981 55.720
513970 89.293 -58.997 55.707
513971 89.326 -59.013 55.696
513972 89.359 -59.029 55.687
513973 89.392 -59.045 55.679
513974 89.425 -59.061 55.675
513975 89.459 -59.062 55.675
513976 89.496 -59.072 55.680
513977 89.536 -59.089 55.689
513978 89.576 -59.105 55.704
513979 89.616 -59.121 55.723
513980 89.655 -59.137 55.742
513981 89.695 -59.153 55.758
513982 89.735 -59.169 55.777
513983 89.774 -59.185 55.797
513984 89.814 -59.202 55.819
513985 89.853 -59.218 55.843
513986 89.893 -59.234 55.867
513987 89.935 -59.249 55.880
513988 89.979 -59.263 55.882
513989 90.024 -59.277 55.879
513990 90.068 -59.291 55.876
513991 90.112 -59.306 55.882
513992 90.156 -59.320 55.875
513993 90.200 -59.334 55.859
513994 90.245 -59.349 55.846
513995 90.289 -59.363 55.838
513996 90.333 -59.378 55.822
513997 90.377 -59.392 55.790
513998 90.422 -59.406 55.745
513999 90.466 -59.420 55.708
514000 90.510 -59.435 55.676
514001 90.554 -59.449 55.630
514002 90.599 -59.463 55.565
514003 90.643 -59.478 55.500
514004 90.687 -59.492 55.437
514005 90.731 -59.507 55.377
514006 90.776 -59.520 55.320
514007 90.820 -59.533 55.268
514008 90.865 -59.546 55.217
514009 90.909 -59.558 55.171
514010 90.952 -59.573 55.129
514011 90.988 -59.591 55.092
514012 91.025 -59.610 55.059
514013 91.062 -59.626 55.028
514014 91.099 -59.642 55.002
514015 91.136 -59.657 54.978
514016 91.173 -59.673 54.958
514017 91.211 -59.689 54.940
514018 91.248 -59.704 54.924
514019 91.285 -59.720 54.911
514020 91.322 -59.735 54.899
514021 91.359 -59.751 54.890
514022 91.396 -59.766 54.881
514023 91.434 -59.782 54.869
514024 91.469 -59.795 54.856
514025 91.505 -59.804 54.844
514026 91.540 -59.814 54.833
514027 91.575 -59.823 54.822
514028 91.610 -59.832 54.811
514029 91.645 -59.842 54.798
514030 91.680 -59.851 54.783
514031 91.715 -59.861 54.767
514032 91.750 -59.870 54.750
514033 91.785 -59.880 54.735
514034 91.816 -59.889 54.721
514035 91.835 -59.901 54.709
514036 91.853 -59.913 54.705
514037 91.872 -59.924 54.708
514038 91.891 -59.936 54.720
514039 91.911 -59.947 54.726
514040 91.938 -59.952 54.724
514041 91.966 -59.958 54.716
514042 91.994 -59.963 54.704
514043 92.021 -59.969 54.691
514044 92.049 -59.974 54.673
514045 92.077 -59.980 54.654
514046 92.119 -59.980 54.631
514047 92.166 -59.980 54.607
514048 92.211 -59.984 54.580
514049 92.253 -59.994 54.549
514050 92.296 -60.004 54.522
514051 92.338 -60.013 54.506
514052 92.380 -60.023 54.497
514053 92.423 -60.033 54.491
514054 92.465 -60.043 54.474
514055 92.508 -60.052 54.460
514056 92.551 -60.061 54.445
514057 92.593 -60.071 54.429
514058 92.635 -60.081 54.419
514059 92.677 -60.090 54.422
514060 92.720 -60.100 54.432
514061 92.763 -60.110 54.432
514062 92.805 -60.119 54.431
514063 92.848 -60.129 54.428
514064 92.890 -60.139 54.418
514065 92.932 -60.149 54.418
514066 92.975 -60.158 54.418
514067 93.017 -60.168 54.412
514068 93.060 -60.177 54.409
514069 93.102 -60.187 54.420
514070 93.143 -60.199 54.426
514071 93.184 -60.214 54.433
514072 93.224 -60.229 54.439
514073 93.265 -60.244 54.443
514074 93.306 -60.259 54.443
514075 93.346 -60.275 54.438
514076 93.387 -60.290 54.439
514077 93.428 -60.305 54.437
514078 93.469 -60.319 54.434
514079 93.509 -60.334 54.427
514080 93.550 -60.350 54.419
514081 93.590 -60.365 54.405
514082 93.631 -60.380 54.390
514083 93.672 -60.395 54.371
514084 93.713 -60.409 54.350
514085 93.754 -60.425 54.327
514086 93.794 -60.440 54.303
514087 93.836 -60.453 54.278
514088 93.880 -60.465 54.252
514089 93.923 -60.476 54.226
514090 93.967 -60.488 54.199
514091 94.010 -60.500 54.175
514092 94.053 -60.511 54.158
514093 94.097 -60.522 54.147
514094 94.140 -60.535 54.139
514095 94.184 -60.546 54.131
514096 94.227 -60.557 54.125
514097 94.270 -60.569 54.114
514098 94.309 -60.573 54.101
514099 94.348 -60.578 54.088
514100 94.387 -60.582 54.075
514101 94.426 -60.587 54.064
514102 94.465 -60.592 54.054
514103 94.503 -60.597 54.045
514104 94.542 -60.601 54.035
514105 94.583 -60.607 54.025
514106 94.622 -60.612 54.013
514107 94.661 -60.617 53.997
514108 94.699 -60.622 53.976
514109 94.738 -60.627 53.954
514110 94.776 -60.632 53.933
514111 94.815 -60.637 53.910
514112 94.853 -60.642 53.883
514113 94.889 -60.649 53.854
514114 94.924 -60.656 53.824
514115 94.960 -60.663 53.793
514116 94.995 -60.669 53.761
514117 95.031 -60.677 53.727
514118 95.066 -60.684 53.688
514119 95.102 -60.691 53.652
514120 95.137 -60.698 53.615
514121 95.173 -60.704 53.577
514122 95.209 -60.713 53.537
514123 95.246 -60.720 53.497
514124 95.282 -60.726 53.459
514125 95.319 -60.733 53.422
514126 95.355 -60.740 53.384
514127 95.391 -60.747 53.343
514128 95.426 -60.752 53.301
514129 95.464 -60.757 53.254
514130 95.500 -60.763 53.201
514131 95.536 -60.770 53.147
514132 95.572 -60.776 53.094
514133 95.609 -60.782 53.042
514134 95.645 -60.788 52.992
514135 95.682 -60.795 52.944
514136 95.718 -60.801 52.895
514137 95.754 -60.808 52.847
514138 95.792 -60.806 52.800
514139 95.830 -60.801 52.756
514140 95.871 -60.803 52.714
514141 95.912 -60.808 52.674
514142 95.953 -60.814 52.635
514143 95.994 -60.818 52.597
514144 96.036 -60.823 52.562
514145 96.077 -60.829 52.529
514146 96.118 -60.833 52.500
514147 96.159 -60.839 52.472
514148 96.195 -60.842 52.445
514149 96.232 -60.845 52.417
514150 96.270 -60.846 52.389
514151 96.309 -60.847 52.361
514152 96.348 -60.848 52.334
514153 96.388 -60.849 52.307
514154 96.427 -60.850 52.281
514155 96.466 -60.851 52.253
514156 96.506 -60.852 52.225
514157 96.545 -60.854 52.193
514158 96.585 -60.854 52.158
514159 96.624 -60.855 52.120
514160 96.661 -60.857 52.082
514161 96.697 -60.858 52.042
514162 96.743 -60.854 52.003
514163 96.786 -60.860 51.964
514164 96.824 -60.874 51.924
514165 96.863 -60.888 51.883
514166 96.909 -60.895 51.844
514167 96.959 -60.899 51.806
514168 97.008 -60.904 51.769
514169 97.057 -60.909 51.733
514170 97.106 -60.914 51.700
514171 97.155 -60.918 51.664
514172 97.202 -60.914 51.629
514173 97.225 -60.925 51.592
514174 97.244 -60.941 51.555
514175 97.260 -60.950 51.516
514176 97.277 -60.960 51.477
514177 97.300 -60.969 51.439
514178 97.331 -60.960 51.401
514179 97.357 -60.949 51.364
514180 97.382 -60.939 51.327
514181 97.406 -60.927 51.290
514182 97.432 -60.917 51.253
514183 97.471 -60.915 51.214
514184 97.520 -60.920 51.174
514185 97.569 -60.924 51.136
514186 97.617 -60.930 51.099
514187 97.666 -60.934 51.060
514188 97.715 -60.939 51.023
514189 97.764 -60.943 50.985
514190 97.812 -60.948 50.946
514191 97.861 -60.952 50.907
514192 97.910 -60.958 50.870
514193 97.932 -60.964 50.832
514194 97.936 -60.972 50.794
514195 97.939 -60.980 50.755
514196 97.952 -60.992 50.717
514197 97.976 -61.008 50.678
514198 98.000 -61.024 50.640
514199 98.024 -61.039 50.602
514200 98.049 -61.055 50.564
514201 98.073 -61.070 50.527
514202 98.097 -61.079 50.490
514203 98.125 -61.069 50.451
514204 98.161 -61.074 50.414
514205 98.198 -61.077 50.378
514206 98.234 -61.082 50.342
514621 87.154 -58.587 57.074
514622 87.195 -58.594 57.049
514623 87.237 -58.592 57.023
514624 87.280 -58.591 56.997
514625 87.322 -58.589 56.968
514626 87.364 -58.587 56.940
514627 87.407 -58.585 56.909
514628 87.449 -58.584 56.877
514629 87.492 -58.582 56.843
514630 87.534 -58.581 56.807
514631 87.576 -58.579 56.770
514632 87.619 -58.578 56.732
514633 87.661 -58.576 56.692
514634 87.698 -58.578 56.650
514635 87.735 -58.579 56.611
514636 87.772 -58.581 56.569
514637 87.818 -58.586 56.528
514638 87.870 -58.593 56.487
514639 87.922 -58.600 56.447
514640 87.974 -58.607 56.409
514641 88.027 -58.614 56.372
514642 88.078 -58.621 56.336
514643 88.130 -58.628 56.302
514652 88.551 -58.726 56.058
514653 88.597 -58.738 56.036
514654 88.642 -58.750 56.015
514655 88.687 -58.761 55.993
514656 88.733 -58.773 55.971
514657 88.778 -58.786 55.951
514658 88.824 -58.798 55.931
514659 88.869 -58.809 55.913
514660 88.914 -58.821 55.893
514661 88.957 -58.833 55.876
514662 88.993 -58.844 55.857
514663 89.028 -58.855 55.839
514664 89.063 -58.867 55.822
514665 89.099 -58.878 55.805
514666 89.134 -58.889 55.788
514667 89.170 -58.900 55.772
514668 89.204 -58.914 55.756
514669 89.237 -58.930 55.740
514670 89.270 -58.946 55.726
514671 89.304 -58.961 55.713
514672 89.336 -58.977 55.702
514673 89.370 -58.993 55.693
514674 89.403 -59.009 55.686
514675 89.436 -59.018 55.681
514676 89.470 -59.018 55.679
514677 89.509 -59.031 55.681
514678 89.548 -59.047 55.688
514679 89.588 -59.063 55.701
514680 89.627 -59.080 55.719
514681 89.667 -59.096 55.743
514682 89.707 -59.112 55.764
514683 89.747 -59.128 55.782
514684 89.787 -59.144 55.802
514685 89.826 -59.160 55.825
514686 89.866 -59.176 55.853
514687 89.907 -59.195 55.875
514688 89.947 -59.209 55.891
514689 89.986 -59.214 55.897
514690 90.032 -59.229 55.892
514691 90.077 -59.245 55.882
514692 90.122 -59.260 55.878
514693 90.167 -59.275 55.871
514694 90.212 -59.291 55.852
514695 90.257 -59.306 55.835
514696 90.302 -59.322 55.827
514697 90.347 -59.337 55.818
514698 90.392 -59.353 55.789
514699 90.437 -59.367 55.747
514700 90.481 -59.381 55.712
514701 90.525 -59.396 55.685
514702 90.570 -59.410 55.644
514703 90.614 -59.424 55.582
514704 90.658 -59.438 55.516
514705 90.702 -59.453 55.450
514706 90.746 -59.467 55.384
514707 90.791 -59.482 55.321
514708 90.835 -59.496 55.265
514709 90.879 -59.510 55.214
514710 90.924 -59.522 55.167
514711 90.964 -59.538 55.125
514712 91.001 -59.556 55.087
514713 91.038 -59.572 55.053
514714 91.075 -59.587 55.023
514715 91.112 -59.602 54.996
514716 91.149 -59.618 54.972
514717 91.186 -59.633 54.950
514718 91.223 -59.649 54.932
514719 91.260 -59.664 54.915
514720 91.298 -59.680 54.901
514721 91.335 -59.695 54.888
514722 91.372 -59.711 54.877
514723 91.409 -59.726 54.867
514724 91.446 -59.742 54.855
514725 91.484 -59.757 54.842
514726 91.519 -59.768 54.830
514727 91.554 -59.777 54.819
514728 91.589 -59.787 54.808
514729 91.624 -59.796 54.796
514730 91.659 -59.806 54.784
514731 91.694 -59.815 54.770
514732 91.729 -59.824 54.756
514733 91.764 -59.834 54.740
514734 91.799 -59.843 54.729
514735 91.819 -59.855 54.717
514736 91.838 -59.867 54.707
514737 91.856 -59.878 54.699
514738 91.875 -59.890 54.690
514739 91.898 -59.898 54.690
514740 91.926 -59.904 54.692
514741 91.953 -59.910 54.686
514742 91.981 -59.915 54.676
514743 92.008 -59.920 54.669
514744 92.036 -59.926 54.666
514745 92.064 -59.931 54.657
514746 92.091 -59.937 54.642
514747 92.133 -59.938 54.626
514748 92.180 -59.937 54.608
514749 92.228 -59.937 54.585
514750 92.270 -59.946 54.553
514751 92.312 -59.956 54.525
514752 92.355 -59.966 54.500
514753 92.398 -59.976 54.484
514754 92.440 -59.985 54.477
514755 92.483 -59.995 54.463
514756 92.525 -60.005 54.450
514757 92.567 -60.014 54.436
514758 92.609 -60.024 54.421
514759 92.652 -60.033 54.409
514760 92.695 -60.043 54.407
514761 92.737 -60.052 54.418
514762 92.780 -60.062 54.423
514763 92.822 -60.072 54.425
514764 92.864 -60.082 54.422
514765 92.907 -60.091 54.413
514766 92.949 -60.101 54.415
514767 92.992 -60.111 54.415
514768 93.034 -60.120 54.413
514769 93.077 -60.130 54.415
514770 93.119 -60.140 54.422
514771 93.158 -60.150 54.429
514772 93.195 -60.171 54.435
514773 93.230 -60.192 54.438
514774 93.270 -60.209 54.439
514775 93.310 -60.224 54.436
514776 93.351 -60.239 54.429
514777 93.392 -60.254 54.433
514778 93.433 -60.269 54.435
514779 93.474 -60.284 54.432
514780 93.514 -60.299 54.427
514781 93.555 -60.314 54.419
514782 93.595 -60.329 54.408
514783 93.636 -60.344 54.396
514784 93.677 -60.359 54.379
514785 93.718 -60.375 54.359
514786 93.758 -60.390 54.335
514787 93.799 -60.404 54.310
514788 93.842 -60.417 54.286
514789 93.885 -60.428 54.260
514790 93.928 -60.441 54.232
514791 93.972 -60.452 54.203
514792 94.015 -60.463 54.173
514793 94.059 -60.475 54.151
514794 94.102 -60.487 54.136
514795 94.145 -60.498 54.125
514796 94.189 -60.510 54.115
514797 94.230 -60.519 54.109
514798 94.272 -60.527 54.100
514799 94.311 -60.532 54.089
514800 94.350 -60.537 54.076
514801 94.389 -60.541 54.062
514802 94.428 -60.546 54.052
514803 94.467 -60.550 54.045
514804 94.505 -60.555 54.040
514805 94.546 -60.560 54.034
514806 94.590 -60.566 54.026
514807 94.629 -60.572 54.013
514808 94.668 -60.578 53.995
514809 94.706 -60.583 53.974
514810 94.745 -60.588 53.952
514811 94.783 -60.593 53.930
514812 94.821 -60.598 53.908
514813 94.859 -60.604 53.880
514814 94.895 -60.610 53.850
514815 94.930 -60.617 53.820
514816 94.966 -60.624 53.789
514817 95.001 -60.632 53.757
514818 95.037 -60.638 53.728
514819 95.072 -60.645 53.692
514820 95.109 -60.651 53.657
514821 95.146 -60.657 53.621
514822 95.181 -60.666 53.582
514823 95.216 -60.675 53.541
514824 95.253 -60.682 53.501
514825 95.289 -60.690 53.464
514826 95.325 -60.697 53.428
514827 95.362 -60.704 53.392
514828 95.398 -60.712 53.356
514829 95.435 -60.718 53.315
514830 95.471 -60.725 53.270
514831 95.507 -60.731 53.219
514832 95.543 -60.738 53.165
514833 95.580 -60.744 53.109
514834 95.616 -60.751 53.056
514835 95.653 -60.757 53.003
514836 95.689 -60.763 52.953
514837 95.725 -60.770 52.903
514838 95.761 -60.776 52.853
514839 95.800 -60.772 52.805
514840 95.838 -60.766 52.760
514841 95.878 -60.767 52.719
514842 95.919 -60.771 52.679
514843 95.961 -60.776 52.640
514844 96.002 -60.782 52.603
514845 96.043 -60.787 52.568
514846 96.084 -60.792 52.534
514847 96.125 -60.797 52.503
514848 96.163 -60.801 52.475
514849 96.199 -60.804 52.449
514850 96.236 -60.807 52.421
514851 96.274 -60.808 52.392
514852 96.313 -60.809 52.363
514853 96.353 -60.810 52.335
514854 96.392 -60.811 52.308
514855 96.431 -60.812 52.281
514856 96.471 -60.814 52.253
514857 96.510 -60.814 52.226
514858 96.549 -60.816 52.197
514859 96.589 -60.817 52.165
514860 96.628 -60.817 52.131
514861 96.667 -60.819 52.094
514862 96.707 -60.820 52.057
514863 96.745 -60.821 52.020
514864 96.787 -60.823 51.982
514865 96.829 -60.834 51.943
514866 96.878 -60.839 51.901
514867 96.927 -60.844 51.862
514868 96.976 -60.848 51.824
514869 97.023 -60.856 51.787
514870 97.069 -60.863 51.750
514871 97.116 -60.870 51.715
514872 97.163 -60.877 51.679
514873 97.210 -60.885 51.642
514874 97.227 -60.888 51.605
514875 97.242 -60.892 51.567
514876 97.259 -60.901 51.528
514877 97.276 -60.911 51.490
514878 97.316 -60.919 51.452
514879 97.391 -60.926 51.414
514880 97.384 -60.902 51.377
514881 97.400 -60.894 51.340
514882 97.422 -60.889 51.305
514883 97.444 -60.884 51.267
514884 97.478 -60.882 51.228
514885 97.525 -60.886 51.189
514886 97.573 -60.890 51.149
514887 97.622 -60.895 51.112
514888 97.671 -60.900 51.072
514889 97.719 -60.904 51.034
514890 97.768 -60.909 50.996
514891 97.817 -60.914 50.957
514892 97.866 -60.918 50.918
514893 97.914 -60.923 50.879
514894 97.936 -60.930 50.841
514895 97.940 -60.938 50.802
514896 97.944 -60.946 50.763
514897 97.954 -60.957 50.724
514898 97.978 -60.972 50.683
514899 98.002 -60.988 50.646
514900 98.026 -61.004 50.608
514901 98.051 -61.019 50.570
514902 98.075 -61.035 50.531
514903 98.100 -61.036 50.493
514904 98.131 -61.032 50.454
514905 98.167 -61.036 50.417
514906 98.204 -61.040 50.380
514907 98.240 -61.045 50.343
514908 98.277 -61.049 50.306
515321 87.133 -58.507 57.104
515322 87.174 -58.518 57.080
515323 87.214 -58.529 57.056
515324 87.254 -58.540 57.031
515325 87.296 -58.542 57.006
515326 87.338 -58.540 56.979
515327 87.381 -58.538 56.951
515328 87.423 -58.537 56.922
515329 87.466 -58.535 56.889
515330 87.508 -58.534 56.856
515331 87.550 -58.532 56.821
515332 87.593 -58.531 56.784
515333 87.635 -58.529 56.745
515334 87.677 -58.528 56.706
515335 87.713 -58.530 56.664
515336 87.750 -58.532 56.624
515337 87.793 -58.535 56.582
515338 87.845 -58.542 56.540
515339 87.897 -58.549 56.499
515340 87.949 -58.556 56.459
515341 88.001 -58.563 56.420
515342 88.053 -58.570 56.381
515343 88.105 -58.577 56.344
515344 88.155 -58.585 56.309
515345 88.201 -58.598 56.275
515353 88.564 -58.692 56.058
515354 88.609 -58.705 56.035
515355 88.655 -58.717 56.014
515356 88.700 -58.728 55.992
515357 88.746 -58.740 55.970
515358 88.791 -58.752 55.949
515359 88.836 -58.764 55.930
515360 88.882 -58.776 55.911
515361 88.927 -58.788 55.892
515362 88.967 -58.799 55.874
515363 89.002 -58.811 55.857
515364 89.038 -58.822 55.839
515365 89.073 -58.833 55.821
515366 89.109 -58.844 55.805
515367 89.144 -58.855 55.790
515368 89.180 -58.867 55.774
515369 89.215 -58.878 55.759
515370 89.248 -58.894 55.745
515371 89.281 -58.910 55.730
515372 89.314 -58.926 55.719
515373 89.347 -58.942 55.709
515374 89.381 -58.958 55.700
515375 89.413 -58.973 55.692
515376 89.447 -58.974 55.687
515377 89.481 -58.974 55.683
515378 89.521 -58.990 55.683
515379 89.560 -59.006 55.689
515380 89.600 -59.022 55.699
515381 89.639 -59.038 55.715
515382 89.679 -59.055 55.740
515383 89.719 -59.071 55.760
515384 89.759 -59.087 55.776
515385 89.799 -59.103 55.792
515386 89.839 -59.121 55.814
515387 89.880 -59.139 55.841
515388 89.921 -59.158 55.866
515389 89.962 -59.175 55.882
515390 90.000 -59.178 55.890
515391 90.038 -59.181 55.888
515392 90.081 -59.193 55.875
515393 90.126 -59.208 55.862
515394 90.171 -59.223 55.850
515395 90.217 -59.238 55.834
515396 90.261 -59.254 55.820
515397 90.307 -59.269 55.812
515398 90.352 -59.284 55.807
515399 90.397 -59.300 55.779
515400 90.442 -59.316 55.743
515401 90.487 -59.331 55.711
515402 90.532 -59.347 55.686
515403 90.577 -59.362 55.648
515404 90.622 -59.377 55.590
515405 90.667 -59.392 55.525
515406 90.713 -59.408 55.456
515407 90.757 -59.423 55.387
515408 90.803 -59.438 55.318
515409 90.848 -59.454 55.259
515410 90.893 -59.469 55.207
515411 90.938 -59.485 55.160
515412 90.976 -59.501 55.119
515413 91.013 -59.516 55.080
515414 91.050 -59.532 55.046
515415 91.088 -59.547 55.016
515416 91.125 -59.563 54.989
515417 91.162 -59.579 54.964
515418 91.199 -59.594 54.943
515419 91.236 -59.610 54.924
515420 91.274 -59.625 54.907
515421 91.311 -59.641 54.892
515422 91.348 -59.656 54.878
515423 91.385 -59.671 54.864
515424 91.422 -59.687 54.853
515425 91.459 -59.702 54.840
515426 91.496 -59.718 54.829
515427 91.533 -59.732 54.818
515428 91.568 -59.742 54.807
515429 91.603 -59.751 54.797
515430 91.638 -59.761 54.785
515431 91.673 -59.770 54.774
515432 91.708 -59.779 54.762
515433 91.743 -59.789 54.751
515434 91.778 -59.798 54.738
515435 91.803 -59.809 54.729
515436 91.822 -59.820 54.721
515437 91.841 -59.833 54.714
515438 91.859 -59.844 54.703
515439 91.885 -59.851 54.686
515440 91.913 -59.856 54.671
515441 91.940 -59.861 54.663
515442 91.968 -59.867 54.654
515443 91.996 -59.873 54.645
515444 92.023 -59.878 54.638
515445 92.051 -59.883 54.633
515446 92.078 -59.889 54.625
515447 92.106 -59.894 54.613
515448 92.148 -59.895 54.599
515449 92.195 -59.895 54.582
515450 92.242 -59.894 54.567
515451 92.287 -59.899 54.550
515452 92.330 -59.908 54.527
515453 92.372 -59.918 54.503
515454 92.415 -59.928 54.482
515455 92.457 -59.938 54.468
515456 92.499 -59.947 54.467
515457 92.542 -59.957 54.451
515458 92.584 -59.967 54.434
515459 92.627 -59.976 54.419
515460 92.669 -59.986 54.405
515461 92.712 -59.996 54.394
515462 92.754 -60.005 54.399
515463 92.796 -60.014 54.408
515464 92.839 -60.024 54.414
515465 92.881 -60.034 54.414
515466 92.924 -60.044 54.408
515467 92.966 -60.053 54.409
515468 93.009 -60.063 54.406
515469 93.051 -60.073 54.407
515470 93.093 -60.082 54.413
515471 93.135 -60.092 54.416
515472 93.174 -60.102 54.427
515473 93.210 -60.123 54.436
515474 93.246 -60.143 54.439
515475 93.282 -60.165 54.438
515476 93.318 -60.185 54.436
515477 93.356 -60.203 54.426
515478 93.397 -60.218 54.422
515479 93.437 -60.234 54.417
515480 93.478 -60.249 54.412
515481 93.519 -60.264 54.408
515482 93.560 -60.279 54.404
515483 93.601 -60.293 54.395
515484 93.641 -60.309 54.384
515485 93.682 -60.324 54.368
515486 93.722 -60.339 54.349
515487 93.763 -60.354 54.329
515488 93.804 -60.369 54.306
515489 93.847 -60.381 54.279
515490 93.890 -60.393 54.251
515491 93.934 -60.404 54.222
515492 93.977 -60.416 54.198
515493 94.020 -60.428 54.175
515494 94.064 -60.439 54.154
515495 94.108 -60.450 54.134
515496 94.149 -60.460 54.116
515497 94.191 -60.469 54.101
515498 94.232 -60.477 54.090
515499 94.274 -60.485 54.082
515500 94.314 -60.491 54.072
515501 94.352 -60.496 54.060
515502 94.391 -60.500 54.046
515503 94.430 -60.505 54.037
515504 94.469 -60.510 54.032
515505 94.509 -60.514 54.031
515506 94.552 -60.520 54.028
515507 94.595 -60.525 54.021
515508 94.636 -60.532 54.008
515509 94.674 -60.538 53.991
515510 94.713 -60.544 53.971
515511 94.751 -60.549 53.951
515512 94.790 -60.554 53.930
515513 94.828 -60.559 53.908
515514 94.865 -60.565 53.882
515515 94.901 -60.572 53.851
515516 94.936 -60.579 53.821
515517 94.972 -60.586 53.790
515518 95.008 -60.593 53.758
515519 95.045 -60.599 53.728
515520 95.082 -60.604 53.694
515521 95.118 -60.610 53.658
515522 95.154 -60.619 53.620
515523 95.188 -60.629 53.580
515524 95.224 -60.638 53.538
515525 95.260 -60.645 53.498
515526 95.296 -60.653 53.461
515527 95.333 -60.660 53.426
515528 95.369 -60.667 53.392
515529 95.405 -60.676 53.357
515530 95.442 -60.684 53.318
515531 95.478 -60.693 53.275
515532 95.515 -60.700 53.226
515533 95.551 -60.706 53.173
515534 95.587 -60.713 53.119
515535 95.623 -60.719 53.067
515536 95.660 -60.725 53.015
515537 95.696 -60.732 52.963
515538 95.733 -60.738 52.911
515539 95.769 -60.743 52.860
515540 95.808 -60.737 52.810
515541 95.846 -60.732 52.765
515542 95.885 -60.729 52.724
515543 95.926 -60.735 52.685
515544 95.968 -60.740 52.647
515545 96.009 -60.745 52.609
515546 96.050 -60.750 52.573
515547 96.092 -60.755 52.540
515548 96.131 -60.760 52.507
515549 96.167 -60.763 52.478
515550 96.203 -60.766 52.452
515551 96.240 -60.769 52.424
515552 96.278 -60.770 52.394
515553 96.318 -60.771 52.365
515554 96.357 -60.773 52.336
515555 96.396 -60.773 52.308
515556 96.436 -60.774 52.281
515557 96.475 -60.776 52.252
515558 96.514 -60.776 52.225
515559 96.553 -60.777 52.199
515560 96.593 -60.779 52.170
515561 96.632 -60.779 52.138
515562 96.671 -60.781 52.105
515563 96.711 -60.782 52.071
515564 96.750 -60.782 52.035
515565 96.791 -60.787 51.998
515566 96.836 -60.795 51.959
515567 96.882 -60.802 51.918
515568 96.929 -60.809 51.879
515569 96.976 -60.817 51.841
515570 97.023 -60.824 51.803
515571 97.069 -60.831 51.766
515572 97.116 -60.839 51.729
515573 97.163 -60.845 51.693
515574 97.210 -60.853 51.655
515575 97.227 -60.856 51.617
515576 97.243 -60.859 51.579
515577 97.258 -60.861 51.541
515578 97.274 -60.864 51.504
515579 97.304 -60.868 51.466
515580 97.341 -60.871 51.429
515581 97.377 -60.875 51.391
515582 97.411 -60.863 51.355
515583 97.433 -60.858 51.319
515584 97.455 -60.854 51.282
515585 97.487 -60.851 51.243
515586 97.533 -60.854 51.204
515587 97.579 -60.857 51.164
515588 97.627 -60.861 51.126
515589 97.675 -60.865 51.086
515590 97.724 -60.870 51.047
515591 97.773 -60.875 51.008
515592 97.821 -60.879 50.969
515593 97.870 -60.884 50.929
515594 97.919 -60.889 50.890
515595 97.941 -60.895 50.851
515596 97.944 -60.904 50.811
515597 97.948 -60.912 50.771
515598 97.955 -60.921 50.731
515599 97.980 -60.937 50.691
515600 98.004 -60.952 50.653
515601 98.028 -60.968 50.614
515602 98.052 -60.984 50.575
515603 98.076 -60.999 50.536
515604 98.102 -60.992 50.497
515605 98.138 -60.994 50.457
515606 98.174 -60.998 50.420
515607 98.210 -61.003 50.382
515608 98.246 -61.008 50.344
515609 98.283 -61.011 50.306
515610 98.319 -61.016 50.270
515611 98.355 -61.020 50.232
516021 87.113 -58.427 57.131
516022 87.153 -58.438 57.110
516023 87.194 -58.449 57.087
516024 87.234 -58.460 57.064
516025 87.274 -58.471 57.040
516026 87.315 -58.482 57.015
516027 87.355 -58.491 56.989
516028 87.397 -58.490 56.962
516029 87.440 -58.488 56.933
516030 87.482 -58.487 56.903
516031 87.525 -58.485 56.870
516032 87.567 -58.484 56.835
516033 87.609 -58.482 56.798
516034 87.652 -58.480 56.760
516035 87.692 -58.480 56.720
516036 87.728 -58.482 56.679
516037 87.768 -58.485 56.637
516038 87.820 -58.491 56.596
516039 87.872 -58.498 56.555
516040 87.924 -58.505 56.513
516041 87.976 -58.512 56.472
516042 88.028 -58.519 56.432
516043 88.077 -58.528 56.392
516044 88.123 -58.540 56.355
516045 88.168 -58.552 56.319
516046 88.213 -58.563 56.284
516047 88.259 -58.576 56.251
516052 88.486 -58.636 56.109
516053 88.531 -58.647 56.083
516054 88.576 -58.659 56.058
516055 88.622 -58.671 56.035
516056 88.668 -58.683 56.014
516057 88.713 -58.695 55.991
516058 88.758 -58.707 55.969
516059 88.804 -58.719 55.949
516060 88.849 -58.730 55.929
516061 88.895 -58.742 55.910
516062 88.940 -58.755 55.891
516063 88.977 -58.766 55.873
516064 89.012 -58.777 55.856
516065 89.048 -58.788 55.839
516066 89.083 -58.799 55.821
516067 89.119 -58.811 55.805
516068 89.154 -58.822 55.791
516069 89.189 -58.833 55.777
516070 89.225 -58.844 55.763
516071 89.259 -58.858 55.749
516072 89.292 -58.874 55.735
516073 89.325 -58.889 55.725
516074 89.358 -58.905 55.715
516075 89.391 -58.921 55.707
516076 89.425 -58.931 55.699
516077 89.458 -58.931 55.693
516078 89.493 -58.933 55.688
516079 89.533 -58.949 55.689
516080 89.573 -58.965 55.692
516081 89.612 -58.981 55.698
516082 89.652 -58.997 55.711
516083 89.691 -59.013 55.733
516084 89.731 -59.030 55.749
516085 89.771 -59.046 55.762
516086 89.812 -59.065 55.777
516087 89.853 -59.084 55.796
516088 89.894 -59.102 55.821
516089 89.936 -59.121 55.848
516090 89.976 -59.139 55.867
516091 90.015 -59.145 55.876
516092 90.053 -59.148 55.875
516093 90.092 -59.150 55.861
516094 90.131 -59.156 55.841
516095 90.176 -59.171 55.826
516096 90.221 -59.186 55.814
516097 90.266 -59.202 55.804
516098 90.311 -59.217 55.795
516099 90.356 -59.232 55.782
516100 90.401 -59.248 55.757
516101 90.447 -59.263 55.730
516102 90.492 -59.278 55.705
516103 90.537 -59.294 55.680
516104 90.582 -59.309 55.641
516105 90.627 -59.325 55.585
516106 90.672 -59.340 55.521
516107 90.717 -59.356 55.453
516108 90.762 -59.371 55.385
516109 90.807 -59.386 55.313
516110 90.853 -59.402 55.252
516111 90.897 -59.418 55.198
516112 90.942 -59.443 55.152
516113 90.983 -59.462 55.111
516114 91.023 -59.477 55.072
516115 91.063 -59.492 55.039
516116 91.100 -59.508 55.010
516117 91.138 -59.523 54.983
516118 91.175 -59.539 54.958
516119 91.212 -59.554 54.936
516120 91.249 -59.570 54.917
516121 91.286 -59.585 54.900
516122 91.323 -59.601 54.884
516123 91.361 -59.616 54.869
516124 91.398 -59.632 54.855
516125 91.435 -59.648 54.843
516126 91.472 -59.663 54.832
516127 91.509 -59.679 54.820
516128 91.546 -59.694 54.810
516129 91.582 -59.705 54.800
516130 91.617 -59.715 54.790
516131 91.652 -59.724 54.780
516132 91.687 -59.734 54.771
516133 91.722 -59.743 54.763
516134 91.757 -59.753 54.754
516135 91.787 -59.763 54.746
516136 91.806 -59.774 54.737
516137 91.824 -59.786 54.734
516138 91.845 -59.797 54.728
516139 91.872 -59.802 54.714
516140 91.900 -59.808 54.691
516141 91.928 -59.814 54.661
516142 91.955 -59.819 54.640
516143 91.983 -59.824 54.627
516144 92.010 -59.830 54.617
516145 92.038 -59.835 54.608
516146 92.066 -59.841 54.600
516147 92.093 -59.846 54.590
516148 92.121 -59.851 54.579
516149 92.162 -59.852 54.564
516150 92.210 -59.851 54.548
516151 92.257 -59.851 54.536
516152 92.304 -59.851 54.530
516153 92.347 -59.861 54.517
516154 92.389 -59.870 54.500
516155 92.431 -59.880 54.482
516156 92.474 -59.890 54.465
516157 92.516 -59.900 54.459
516158 92.559 -59.909 54.449
516159 92.601 -59.919 54.438
516160 92.644 -59.929 54.424
516161 92.686 -59.939 54.405
516162 92.728 -59.948 54.385
516163 92.771 -59.958 54.379
516164 92.813 -59.967 54.385
516165 92.856 -59.977 54.394
516166 92.898 -59.986 54.399
516167 92.941 -59.996 54.402
516168 92.983 -60.006 54.399
516169 93.025 -60.015 54.389
516170 93.068 -60.025 54.390
516171 93.111 -60.035 54.402
516172 93.150 -60.043 54.402
516173 93.189 -60.054 54.421
516174 93.224 -60.074 54.435
516175 93.261 -60.096 54.440
516176 93.297 -60.116 54.439
516177 93.333 -60.137 54.439
516178 93.369 -60.158 54.427
516179 93.406 -60.178 54.416
516180 93.442 -60.198 54.406
516181 93.483 -60.213 54.398
516182 93.524 -60.228 54.392
516183 93.564 -60.243 54.386
516184 93.605 -60.259 54.377
516185 93.646 -60.273 54.366
516186 93.687 -60.288 54.352
516187 93.727 -60.303 54.335
516188 93.768 -60.318 54.317
516189 93.809 -60.334 54.295
516190 93.852 -60.345 54.272
516191 93.896 -60.356 54.252
516192 93.939 -60.369 54.234
516193 93.982 -60.380 54.218
516194 94.026 -60.391 54.200
516195 94.069 -60.402 54.180
516196 94.110 -60.410 54.158
516197 94.151 -60.419 54.134
516198 94.193 -60.428 54.111
516199 94.235 -60.436 54.089
516200 94.276 -60.444 54.071
516201 94.316 -60.450 54.055
516202 94.354 -60.454 54.040
516203 94.393 -60.459 54.027
516204 94.432 -60.463 54.017
516205 94.471 -60.468 54.016
516206 94.514 -60.474 54.016
516207 94.557 -60.479 54.014
516208 94.600 -60.485 54.007
516209 94.642 -60.491 53.997
516210 94.681 -60.497 53.984
516211 94.719 -60.503 53.967
516212 94.758 -60.509 53.951
516213 94.796 -60.515 53.934
516214 94.835 -60.520 53.914
516215 94.872 -60.526 53.887
516216 94.907 -60.533 53.858
516217 94.944 -60.540 53.827
516218 94.980 -60.546 53.795
516219 95.017 -60.552 53.762
516220 95.054 -60.557 53.728
516221 95.091 -60.563 53.693
516222 95.126 -60.572 53.654
516223 95.161 -60.581 53.614
516224 95.196 -60.591 53.573
516225 95.231 -60.600 53.532
516226 95.267 -60.608 53.492
516227 95.303 -60.615 53.455
516228 95.340 -60.622 53.420
516229 95.376 -60.630 53.387
516230 95.412 -60.639 53.353
516231 95.449 -60.647 53.316
516232 95.485 -60.657 53.274
516233 95.522 -60.665 53.227
516234 95.558 -60.674 53.177
516235 95.595 -60.681 53.125
516236 95.631 -60.687 53.076
516237 95.667 -60.694 53.024
516238 95.703 -60.700 52.973
516239 95.740 -60.707 52.920
516240 95.777 -60.708 52.867
516241 95.815 -60.703 52.817
516242 95.853 -60.698 52.772
516243 95.892 -60.693 52.729
516244 95.934 -60.698 52.690
516245 95.975 -60.703 52.652
516246 96.016 -60.708 52.615
516247 96.057 -60.713 52.579
516248 96.099 -60.719 52.544
516249 96.135 -60.722 52.511
516250 96.171 -60.724 52.480
516251 96.208 -60.727 52.453
516252 96.244 -60.730 52.424
516253 96.283 -60.732 52.394
516254 96.322 -60.733 52.365
516255 96.361 -60.735 52.336
516256 96.400 -60.735 52.306
516257 96.440 -60.736 52.278
516258 96.479 -60.738 52.249
516259 96.518 -60.738 52.222
516260 96.557 -60.739 52.196
516261 96.597 -60.741 52.169
516262 96.637 -60.742 52.141
516263 96.676 -60.743 52.112
516264 96.717 -60.741 52.080
516265 96.758 -60.741 52.045
516266 96.800 -60.751 52.008
516267 96.836 -60.763 51.970
516268 96.882 -60.770 51.931
516269 96.929 -60.777 51.892
516270 96.976 -60.785 51.854
516271 97.023 -60.792 51.816
516272 97.069 -60.799 51.778
516273 97.116 -60.807 51.741
516274 97.163 -60.814 51.705
516275 97.210 -60.821 51.668
516276 97.228 -60.824 51.631
516277 97.243 -60.827 51.594
516278 97.259 -60.830 51.557
516279 97.275 -60.832 51.519
516280 97.311 -60.834 51.481
516281 97.348 -60.838 51.444
516282 97.384 -60.841 51.407
516283 97.422 -60.832 51.369
516284 97.444 -60.828 51.334
516285 97.466 -60.823 51.297
516286 97.495 -60.820 51.259
516287 97.541 -60.823 51.219
516288 97.587 -60.826 51.180
516289 97.633 -60.828 51.141
516290 97.680 -60.831 51.102
516291 97.729 -60.836 51.063
516292 97.777 -60.840 51.022
516293 97.826 -60.845 50.982
516294 97.875 -60.850 50.942
516295 97.923 -60.854 50.902
516296 97.945 -60.861 50.861
516297 97.949 -60.870 50.821
516298 97.953 -60.877 50.780
516299 97.957 -60.886 50.739
516300 97.982 -60.901 50.699
516301 98.006 -60.917 50.659
516302 98.030 -60.933 50.619
516303 98.054 -60.948 50.580
516304 98.078 -60.961 50.541
516305 98.107 -60.952 50.501
516306 98.144 -60.957 50.462
516307 98.180 -60.961 50.423
516308 98.216 -60.966 50.384
516309 98.252 -60.970 50.346
516310 98.289 -60.974 50.307
516311 98.325 -60.979 50.270
516312 98.361 -60.983 50.231
516722 87.133 -58.358 57.136
516723 87.174 -58.369 57.116
516724 87.214 -58.380 57.095
516725 87.254 -58.391 57.072
516726 87.294 -58.402 57.048
516727 87.335 -58.413 57.024
516728 87.375 -58.424 56.999
516729 87.415 -58.435 56.973
516730 87.456 -58.440 56.945
516731 87.499 -58.438 56.916
516732 87.541 -58.437 56.883
516733 87.583 -58.435 56.848
516734 87.626 -58.433 56.812
516735 87.668 -58.432 56.775
516736 87.707 -58.432 56.735
516737 87.744 -58.434 56.695
516738 87.794 -58.441 56.654
516739 87.847 -58.448 56.613
516740 87.899 -58.454 56.571
516741 87.951 -58.461 56.529
516742 87.999 -58.471 56.487
516743 88.044 -58.483 56.445
516744 88.090 -58.495 56.405
516745 88.135 -58.507 56.367
516746 88.181 -58.519 56.330
516747 88.226 -58.530 56.295
516748 88.271 -58.542 56.260
516749 88.317 -58.554 56.228
516750 88.362 -58.567 56.198
516751 88.408 -58.578 56.169
516752 88.453 -58.590 56.140
516753 88.498 -58.602 56.113
516754 88.544 -58.614 56.086
516755 88.590 -58.626 56.060
516756 88.635 -58.638 56.036
516757 88.680 -58.649 56.014
516758 88.725 -58.661 55.991
516759 88.771 -58.673 55.969
516760 88.817 -58.686 55.948
516761 88.862 -58.697 55.928
516762 88.907 -58.709 55.909
516763 88.951 -58.721 55.892
516764 88.986 -58.732 55.873
516765 89.022 -58.743 55.856
516766 89.057 -58.755 55.839
516767 89.093 -58.766 55.822
516768 89.128 -58.777 55.807
516769 89.164 -58.788 55.793
516770 89.199 -58.799 55.779
516771 89.235 -58.811 55.766
516772 89.270 -58.822 55.754
516773 89.303 -58.838 55.741
516774 89.336 -58.854 55.731
516775 89.369 -58.870 55.722
516776 89.402 -58.886 55.714
516777 89.436 -58.888 55.706
516778 89.470 -58.887 55.700
516779 89.505 -58.892 55.695
516780 89.545 -58.908 55.694
516781 89.584 -58.924 55.695
516782 89.624 -58.940 55.700
516783 89.664 -58.956 55.708
516784 89.703 -58.972 55.722
516785 89.745 -58.991 55.733
516786 89.786 -59.009 55.745
516787 89.827 -59.028 55.757
516788 89.868 -59.046 55.774
516789 89.909 -59.065 55.797
516790 89.950 -59.084 55.822
516791 89.991 -59.102 55.844
516792 90.030 -59.112 55.857
516793 90.069 -59.115 55.855
516794 90.107 -59.117 55.841
516795 90.145 -59.120 55.819
516796 90.183 -59.123 55.805
516797 90.226 -59.134 55.796
516798 90.270 -59.150 55.787
516799 90.316 -59.165 55.775
516800 90.361 -59.180 55.757
516801 90.406 -59.195 55.737
516802 90.451 -59.211 55.716
516803 90.496 -59.226 55.693
516804 90.541 -59.241 55.664
516805 90.586 -59.257 55.620
516806 90.632 -59.272 55.563
516807 90.676 -59.287 55.497
516808 90.722 -59.303 55.429
516809 90.766 -59.319 55.363
516810 90.812 -59.334 55.297
516811 90.857 -59.352 55.238
516812 90.901 -59.377 55.186
516813 90.945 -59.401 55.141
516814 90.988 -59.423 55.101
516815 91.028 -59.438 55.064
516816 91.068 -59.453 55.032
516817 91.107 -59.469 55.003
516818 91.147 -59.484 54.977
516819 91.186 -59.500 54.952
516820 91.225 -59.515 54.930
516821 91.262 -59.530 54.911
516822 91.299 -59.546 54.893
516823 91.336 -59.561 54.877
516824 91.374 -59.577 54.862
516825 91.411 -59.592 54.848
516826 91.448 -59.608 54.836
516827 91.485 -59.623 54.825
516828 91.522 -59.638 54.815
516829 91.559 -59.654 54.807
516830 91.596 -59.670 54.799
516831 91.632 -59.679 54.790
516832 91.666 -59.689 54.782
516833 91.702 -59.698 54.777
516834 91.736 -59.707 54.773
516835 91.771 -59.717 54.769
516836 91.790 -59.729 54.762
516837 91.809 -59.740 54.756
516838 91.832 -59.749 54.752
516839 91.859 -59.754 54.744
516840 91.887 -59.760 54.726
516841 91.915 -59.765 54.699
516842 91.942 -59.770 54.661
516843 91.970 -59.776 54.629
516844 91.997 -59.782 54.606
516845 92.025 -59.787 54.591
516846 92.053 -59.792 54.581
516847 92.080 -59.798 54.567
516848 92.108 -59.804 54.554
516849 92.135 -59.809 54.540
516850 92.177 -59.810 54.526
516851 92.224 -59.809 54.513
516852 92.272 -59.808 54.505
516853 92.319 -59.808 54.505
516854 92.363 -59.814 54.503
516855 92.406 -59.823 54.495
516856 92.448 -59.833 54.482
516857 92.491 -59.842 54.465
516858 92.533 -59.852 54.453
516859 92.576 -59.862 54.446
516860 92.618 -59.871 54.438
516861 92.660 -59.881 54.426
516862 92.703 -59.891 54.407
516863 92.745 -59.901 54.388
516864 92.788 -59.910 54.378
516865 92.830 -59.920 54.373
516866 92.873 -59.930 54.374
516867 92.915 -59.939 54.382
516868 92.957 -59.949 54.387
516869 93.000 -59.958 54.383
516870 93.043 -59.968 54.378
516871 93.085 -59.977 54.382
516872 93.127 -59.987 54.397
516873 93.165 -59.996 54.411
516874 93.204 -60.005 54.415
516875 93.239 -60.027 54.427
516876 93.276 -60.047 54.437
516877 93.312 -60.068 54.441
516878 93.348 -60.089 54.446
516879 93.384 -60.109 54.434
516880 93.421 -60.130 54.424
516881 93.457 -60.151 54.413
516882 93.493 -60.171 54.400
516883 93.529 -60.192 54.385
516884 93.569 -60.208 54.367
516885 93.610 -60.223 54.354
516886 93.651 -60.238 54.343
516887 93.692 -60.253 54.332
516888 93.732 -60.268 54.318
516889 93.773 -60.283 54.303
516890 93.814 -60.297 54.284
516891 93.857 -60.309 54.269
516892 93.901 -60.321 54.259
516893 93.944 -60.332 54.253
516894 93.988 -60.343 54.244
516895 94.029 -60.353 54.230
516896 94.071 -60.361 54.211
516897 94.112 -60.369 54.187
516898 94.153 -60.378 54.160
516899 94.195 -60.386 54.131
516900 94.237 -60.394 54.102
516901 94.278 -60.403 54.077
516902 94.318 -60.409 54.055
516903 94.356 -60.413 54.034
516904 94.395 -60.418 54.016
516905 94.434 -60.422 54.001
516906 94.477 -60.428 53.997
516907 94.520 -60.434 53.997
516908 94.562 -60.439 53.997
516909 94.606 -60.445 53.992
516910 94.648 -60.450 53.983
516911 94.688 -60.456 53.972
516912 94.726 -60.463 53.959
516913 94.765 -60.469 53.947
516914 94.804 -60.475 53.932
516915 94.842 -60.481 53.912
516916 94.880 -60.487 53.886
516917 94.916 -60.493 53.857
516918 94.953 -60.499 53.826
516919 94.990 -60.505 53.794
516920 95.026 -60.510 53.760
516921 95.063 -60.516 53.724
516922 95.099 -60.524 53.685
516923 95.134 -60.534 53.644
516924 95.169 -60.544 53.601
516925 95.203 -60.553 53.559
516926 95.238 -60.563 53.520
516927 95.274 -60.570 53.483
516928 95.311 -60.578 53.445
516929 95.347 -60.585 53.409
516930 95.383 -60.594 53.376
516931 95.419 -60.602 53.343
516932 95.456 -60.611 53.306
516933 95.492 -60.619 53.267
516934 95.529 -60.629 53.223
516935 95.565 -60.637 53.176
516936 95.602 -60.646 53.129
516937 95.638 -60.654 53.083
516938 95.675 -60.662 53.033
516939 95.711 -60.668 52.983
516940 95.747 -60.675 52.930
516941 95.784 -60.674 52.876
516942 95.823 -60.669 52.826
516943 95.862 -60.663 52.779
516944 95.900 -60.657 52.735
516945 95.941 -60.661 52.696
516946 95.982 -60.666 52.658
516947 96.023 -60.672 52.620
516948 96.064 -60.676 52.584
516949 96.103 -60.680 52.548
516950 96.139 -60.683 52.515
516951 96.176 -60.686 52.483
516952 96.212 -60.689 52.453
516953 96.248 -60.692 52.423
516954 96.287 -60.695 52.393
516955 96.326 -60.695 52.364
516956 96.365 -60.697 52.334
516957 96.404 -60.698 52.305
516958 96.444 -60.698 52.276
516959 96.483 -60.700 52.246
516960 96.523 -60.701 52.219
516961 96.562 -60.701 52.192
516962 96.601 -60.703 52.166
516963 96.642 -60.701 52.141
516964 96.682 -60.698 52.115
516965 96.724 -60.696 52.086
516966 96.768 -60.704 52.053
516967 96.810 -60.715 52.017
516968 96.841 -60.729 51.980
516969 96.882 -60.738 51.942
516970 96.929 -60.745 51.904
516971 96.976 -60.753 51.866
516972 97.023 -60.760 51.829
516973 97.069 -60.767 51.792
516974 97.116 -60.775 51.755
516975 97.163 -60.782 51.718
516976 97.210 -60.789 51.681
516977 97.229 -60.792 51.644
516978 97.244 -60.795 51.606
516979 97.259 -60.798 51.569
516980 97.281 -60.801 51.531
516981 97.317 -60.801 51.494
516982 97.354 -60.804 51.457
516983 97.390 -60.807 51.421
516984 97.427 -60.810 51.384
516985 97.455 -60.798 51.348
516986 97.477 -60.793 51.312
516987 97.504 -60.789 51.274
516988 97.550 -60.792 51.236
516989 97.596 -60.795 51.197
516990 97.642 -60.797 51.157
516991 97.688 -60.799 51.118
516992 97.734 -60.802 51.079
516993 97.782 -60.806 51.038
516994 97.830 -60.810 50.998
516995 97.879 -60.816 50.956
516996 97.928 -60.820 50.915
516997 97.950 -60.827 50.873
516998 97.953 -60.835 50.832
516999 97.957 -60.843 50.789
517000 97.961 -60.851 50.748
517001 97.984 -60.867 50.707
517002 98.008 -60.882 50.667
517003 98.032 -60.898 50.627
517004 98.056 -60.914 50.586
517005 98.081 -60.917 50.546
517006 98.113 -60.915 50.506
517007 98.150 -60.920 50.465
517008 98.186 -60.924 50.426
517009 98.222 -60.928 50.387
517010 98.258 -60.933 50.348
517011 98.295 -60.937 50.309
517012 98.332 -60.941 50.270
517013 98.368 -60.945 50.231
517014 98.404 -60.950 50.193
517423 87.153 -58.289 57.140
517424 87.194 -58.300 57.121
517425 87.234 -58.311 57.101
517426 87.274 -58.322 57.079
517427 87.314 -58.333 57.056
517428 87.354 -58.344 57.033
517429 87.395 -58.355 57.008
517430 87.435 -58.366 56.983
517431 87.475 -58.376 56.957
517432 87.515 -58.388 56.927
517433 87.558 -58.388 56.895
517434 87.600 -58.386 56.861
517435 87.642 -58.385 56.827
517436 87.685 -58.383 56.791
517437 87.722 -58.385 56.751
517438 87.769 -58.389 56.712
517439 87.821 -58.397 56.671
517440 87.873 -58.404 56.630
517441 87.921 -58.413 56.588
517442 87.966 -58.426 56.545
517443 88.012 -58.438 56.503
517444 88.057 -58.449 56.461
517445 88.103 -58.461 56.421
517446 88.148 -58.473 56.382
517447 88.193 -58.485 56.344
517448 88.239 -58.497 56.306
517449 88.284 -58.509 56.270
517450 88.330 -58.521 56.236
517451 88.375 -58.532 56.205
517452 88.420 -58.545 56.175
517453 88.466 -58.557 56.146
517454 88.511 -58.569 56.117
517455 88.557 -58.580 56.090
517456 88.602 -58.592 56.064
517457 88.647 -58.604 56.038
517458 88.693 -58.616 56.015
517459 88.739 -58.628 55.992
517460 88.784 -58.640 55.970
517461 88.829 -58.652 55.949
517462 88.875 -58.664 55.929
517463 88.920 -58.676 55.910
517464 88.961 -58.687 55.892
517465 88.996 -58.698 55.874
517466 89.032 -58.710 55.856
517467 89.067 -58.721 55.839
517468 89.103 -58.732 55.823
517469 89.138 -58.743 55.809
517470 89.174 -58.755 55.795
517471 89.209 -58.766 55.782
517472 89.244 -58.777 55.770
517473 89.280 -58.788 55.759
517474 89.314 -58.802 55.747
517475 89.347 -58.818 55.738
517476 89.380 -58.834 55.729
517477 89.413 -58.845 55.721
517478 89.447 -58.844 55.714
517479 89.481 -58.844 55.709
517480 89.517 -58.850 55.704
517481 89.557 -58.866 55.701
517482 89.596 -58.883 55.701
517483 89.636 -58.899 55.703
517484 89.677 -58.917 55.708
517485 89.718 -58.935 55.717
517486 89.759 -58.954 55.726
517487 89.800 -58.972 55.735
517488 89.841 -58.991 55.747
517489 89.882 -59.009 55.760
517490 89.923 -59.027 55.777
517491 89.964 -59.046 55.796
517492 90.005 -59.065 55.814
517493 90.045 -59.078 55.826
517494 90.084 -59.081 55.825
517495 90.121 -59.084 55.815
517496 90.160 -59.087 55.801
517497 90.198 -59.090 55.791
517498 90.236 -59.093 55.782
517499 90.275 -59.096 55.769
517500 90.320 -59.112 55.750
517501 90.365 -59.128 55.731
517502 90.410 -59.143 55.717
517503 90.456 -59.159 55.701
517504 90.501 -59.174 55.677
517505 90.546 -59.189 55.640
517506 90.591 -59.204 55.591
517507 90.636 -59.220 55.532
517508 90.681 -59.235 55.467
517509 90.726 -59.250 55.400
517510 90.771 -59.266 55.338
517511 90.816 -59.285 55.278
517512 90.860 -59.310 55.224
517513 90.904 -59.335 55.176
517514 90.949 -59.359 55.132
517515 90.993 -59.384 55.093
517516 91.033 -59.399 55.057
517517 91.072 -59.415 55.025
517518 91.112 -59.430 54.996
517519 91.152 -59.445 54.971
517520 91.191 -59.460 54.948
517521 91.231 -59.475 54.926
517522 91.271 -59.491 54.906
517523 91.310 -59.507 54.887
517524 91.349 -59.522 54.871
517525 91.386 -59.538 54.856
517526 91.424 -59.553 54.842
517527 91.461 -59.568 54.831
517528 91.498 -59.584 54.820
517529 91.535 -59.599 54.812
517530 91.572 -59.615 54.807
517531 91.609 -59.630 54.802
517532 91.646 -59.643 54.794
517533 91.680 -59.652 54.787
517534 91.716 -59.662 54.784
517535 91.750 -59.671 54.783
517536 91.774 -59.682 54.781
517537 91.793 -59.694 54.775
517538 91.819 -59.701 54.769
517539 91.847 -59.706 54.764
517540 91.874 -59.712 54.753
517541 91.901 -59.717 54.733
517542 91.929 -59.723 54.704
517543 91.957 -59.728 54.663
517544 91.985 -59.733 54.624
517545 92.012 -59.739 54.595
517546 92.039 -59.745 54.576
517547 92.067 -59.750 54.566
517548 92.095 -59.755 54.560
517549 92.123 -59.761 54.554
517550 92.150 -59.766 54.548
517551 92.192 -59.767 54.539
517552 92.239 -59.767 54.527
517553 92.287 -59.766 54.513
517554 92.334 -59.765 54.494
517555 92.380 -59.766 54.489
517556 92.423 -59.776 54.488
517557 92.465 -59.786 54.481
517558 92.508 -59.795 54.466
517559 92.550 -59.804 54.451
517560 92.592 -59.814 54.442
517561 92.635 -59.824 54.434
517562 92.677 -59.834 54.423
517563 92.720 -59.843 54.412
517564 92.762 -59.853 54.398
517565 92.805 -59.863 54.387
517566 92.847 -59.872 54.371
517567 92.889 -59.882 54.361
517568 92.932 -59.892 54.367
517569 92.975 -59.902 54.373
517570 93.017 -59.911 54.374
517571 93.060 -59.920 54.376
517572 93.102 -59.930 54.384
517573 93.142 -59.939 54.397
517574 93.181 -59.948 54.412
517575 93.219 -59.958 54.413
517576 93.255 -59.978 54.422
517577 93.291 -59.999 54.432
517578 93.327 -60.020 54.439
517579 93.363 -60.040 54.444
517580 93.400 -60.061 54.437
517581 93.436 -60.082 54.430
517582 93.472 -60.102 54.421
517583 93.508 -60.123 54.409
517584 93.544 -60.144 54.392
517585 93.580 -60.165 54.373
517586 93.617 -60.185 54.361
517587 93.655 -60.202 54.348
517588 93.696 -60.218 54.334
517589 93.737 -60.233 54.320
517590 93.778 -60.247 54.305
517591 93.820 -60.262 54.293
517592 93.863 -60.273 54.284
517593 93.906 -60.284 54.279
517594 93.948 -60.294 54.277
517595 93.990 -60.303 54.271
517596 94.031 -60.311 54.258
517597 94.073 -60.319 54.241
517598 94.114 -60.328 54.219
517599 94.155 -60.336 54.191
517600 94.197 -60.344 54.159
517601 94.239 -60.353 54.128
517602 94.280 -60.362 54.098
517603 94.320 -60.367 54.070
517604 94.358 -60.372 54.041
517605 94.397 -60.376 54.013
517606 94.439 -60.381 53.991
517607 94.482 -60.387 53.980
517608 94.525 -60.393 53.978
517609 94.568 -60.399 53.978
517610 94.611 -60.404 53.974
517611 94.654 -60.410 53.966
517612 94.695 -60.416 53.957
517613 94.733 -60.422 53.948
517614 94.772 -60.428 53.937
517615 94.810 -60.434 53.923
517616 94.850 -60.440 53.903
517617 94.889 -60.446 53.878
517618 94.925 -60.452 53.849
517619 94.962 -60.458 53.819
517620 94.999 -60.463 53.787
517621 95.035 -60.469 53.752
517622 95.071 -60.477 53.716
517623 95.106 -60.486 53.675
517624 95.141 -60.496 53.632
517625 95.176 -60.506 53.591
517626 95.210 -60.516 53.551
517627 95.245 -60.525 53.515
517628 95.281 -60.533 53.479
517629 95.318 -60.541 53.441
517630 95.354 -60.548 53.402
517631 95.390 -60.557 53.365
517632 95.427 -60.566 53.330
517633 95.463 -60.575 53.293
517634 95.499 -60.583 53.255
517635 95.536 -60.592 53.213
517636 95.572 -60.600 53.170
517637 95.609 -60.610 53.127
517638 95.645 -60.618 53.084
517639 95.682 -60.627 53.038
517640 95.718 -60.635 52.990
517641 95.754 -60.643 52.939
517642 95.793 -60.640 52.886
517643 95.831 -60.634 52.835
517644 95.869 -60.629 52.788
517645 95.908 -60.622 52.744
517646 95.948 -60.625 52.703
517647 95.989 -60.630 52.664
517648 96.031 -60.635 52.625
517649 96.071 -60.639 52.587
517650 96.107 -60.642 52.551
517651 96.144 -60.645 52.517
517652 96.180 -60.648 52.485
517653 96.217 -60.651 52.453
517654 96.253 -60.654 52.422
517655 96.291 -60.657 52.391
517656 96.330 -60.657 52.361
517657 96.370 -60.659 52.332
517658 96.409 -60.660 52.303
517659 96.448 -60.660 52.274
517660 96.487 -60.662 52.244
517661 96.527 -60.663 52.215
517662 96.567 -60.662 52.189
517663 96.608 -60.660 52.163
517664 96.649 -60.657 52.139
517665 96.690 -60.654 52.116
517666 96.733 -60.659 52.090
517667 96.778 -60.668 52.059
517668 96.819 -60.679 52.025
517669 96.851 -60.692 51.989
517670 96.883 -60.706 51.953
517671 96.929 -60.713 51.916
517672 96.976 -60.721 51.879
517673 97.022 -60.729 51.842
517674 97.069 -60.735 51.806
517675 97.116 -60.743 51.770
517676 97.163 -60.750 51.733
517677 97.210 -60.757 51.695
517678 97.229 -60.760 51.658
517679 97.245 -60.763 51.619
517680 97.260 -60.767 51.581
517681 97.287 -60.768 51.544
517682 97.324 -60.769 51.507
517683 97.361 -60.770 51.471
517684 97.397 -60.773 51.435
517685 97.434 -60.777 51.399
517686 97.466 -60.767 51.363
517687 97.489 -60.762 51.327
517688 97.512 -60.758 51.290
517689 97.559 -60.760 51.253
517690 97.604 -60.763 51.214
517691 97.650 -60.766 51.175
517692 97.697 -60.768 51.135
517693 97.742 -60.771 51.096
517694 97.788 -60.773 51.055
517695 97.835 -60.776 51.013
517696 97.884 -60.781 50.971
517697 97.932 -60.785 50.929
517698 97.954 -60.792 50.886
517699 97.958 -60.801 50.843
517700 97.962 -60.809 50.801
517701 97.965 -60.817 50.758
517702 97.985 -60.831 50.716
517703 98.009 -60.847 50.675
517704 98.033 -60.863 50.634
517705 98.058 -60.878 50.594
517706 98.083 -60.873 50.553
517707 98.120 -60.878 50.512
517708 98.156 -60.882 50.470
517709 98.193 -60.886 50.430
517710 98.229 -60.891 50.390
517711 98.265 -60.895 50.350
517712 98.301 -60.899 50.310
517713 98.338 -60.904 50.271
517714 98.374 -60.908 50.231
517715 98.410 -60.912 50.192
518124 87.173 -58.220 57.145
518125 87.213 -58.231 57.126
518126 87.254 -58.242 57.106
518127 87.294 -58.253 57.085
518128 87.334 -58.264 57.064
518129 87.374 -58.275 57.042
518130 87.415 -58.286 57.018
518131 87.455 -58.297 56.993
518132 87.495 -58.307 56.967
518133 87.536 -58.319 56.939
518134 87.576 -58.329 56.907
518135 87.617 -58.338 56.875
518136 87.659 -58.336 56.842
518137 87.701 -58.335 56.808
518138 87.744 -58.338 56.768
518139 87.796 -58.345 56.728
518140 87.843 -58.357 56.688
518141 87.888 -58.368 56.647
518142 87.933 -58.380 56.605
518143 87.979 -58.392 56.563
518144 88.025 -58.404 56.520
518145 88.070 -58.416 56.479
518146 88.115 -58.428 56.438
518147 88.160 -58.440 56.398
518148 88.206 -58.451 56.358
518149 88.252 -58.463 56.320
518150 88.297 -58.476 56.283
518151 88.342 -58.488 56.248
518152 88.388 -58.499 56.215
518153 88.433 -58.511 56.184
518154 88.479 -58.523 56.152
518155 88.524 -58.535 56.123
518156 88.569 -58.547 56.094
518157 88.615 -58.559 56.067
518158 88.660 -58.571 56.042
518159 88.706 -58.582 56.018
518160 88.751 -58.595 55.994
518161 88.797 -58.607 55.971
518162 88.842 -58.618 55.950
518163 88.888 -58.629 55.929
518164 88.933 -58.641 55.911
518165 88.971 -58.654 55.892
518166 89.006 -58.665 55.874
518167 89.042 -58.676 55.857
518168 89.077 -58.687 55.840
518169 89.112 -58.698 55.824
518170 89.148 -58.710 55.811
518171 89.183 -58.720 55.798
518172 89.218 -58.732 55.785
518173 89.254 -58.743 55.774
518174 89.290 -58.755 55.763
518175 89.325 -58.766 55.753
518176 89.358 -58.782 55.744
518177 89.391 -58.798 55.737
518178 89.424 -58.801 55.730
518179 89.458 -58.801 55.725
518180 89.492 -58.800 55.720
518181 89.529 -58.809 55.715
518182 89.568 -58.825 55.712
518183 89.609 -58.842 55.710
518184 89.650 -58.861 55.709
518185 89.691 -58.880 55.711
518186 89.732 -58.898 55.717
518187 89.773 -58.917 55.725
518188 89.814 -58.935 55.733
518189 89.856 -58.954 55.742
518190 89.896 -58.972 55.752
518191 89.938 -58.990 55.763
518192 89.978 -59.009 55.775
518193 90.020 -59.027 55.787
518194 90.060 -59.045 55.795
518195 90.098 -59.048 55.796
518196 90.136 -59.051 55.792
518197 90.175 -59.053 55.785
518198 90.213 -59.056 55.779
518199 90.251 -59.059 55.770
518200 90.290 -59.062 55.750
518201 90.327 -59.065 55.726
518202 90.370 -59.075 55.714
518203 90.415 -59.090 55.700
518204 90.460 -59.106 55.678
518205 90.505 -59.121 55.646
518206 90.550 -59.137 55.604
518207 90.595 -59.152 55.551
518208 90.641 -59.168 55.493
518209 90.685 -59.183 55.431
518210 90.731 -59.198 55.368
518211 90.776 -59.219 55.309
518212 90.820 -59.244 55.257
518213 90.864 -59.269 55.210
518214 90.908 -59.294 55.165
518215 90.953 -59.318 55.124
518216 90.997 -59.343 55.085
518217 91.038 -59.360 55.050
518218 91.077 -59.375 55.018
518219 91.117 -59.391 54.990
518220 91.157 -59.406 54.965
518221 91.196 -59.422 54.943
518222 91.236 -59.437 54.922
518223 91.276 -59.452 54.901
518224 91.315 -59.468 54.883
518225 91.355 -59.483 54.866
518226 91.394 -59.498 54.851
518227 91.434 -59.513 54.837
518228 91.473 -59.529 54.827
518229 91.511 -59.544 54.816
518230 91.548 -59.560 54.808
518231 91.585 -59.576 54.805
518232 91.622 -59.591 54.804
518233 91.659 -59.607 54.798
518234 91.695 -59.616 54.791
518235 91.730 -59.626 54.787
518236 91.758 -59.636 54.786
518237 91.779 -59.647 54.786
518238 91.806 -59.653 54.781
518239 91.834 -59.658 54.777
518240 91.861 -59.663 54.770
518241 91.889 -59.669 54.757
518242 91.917 -59.674 54.737
518243 91.944 -59.680 54.706
518244 91.971 -59.685 54.666
518245 91.999 -59.691 54.624
518246 92.027 -59.696 54.590
518247 92.055 -59.701 54.567
518248 92.082 -59.707 54.554
518249 92.109 -59.713 54.554
518250 92.137 -59.718 54.552
518251 92.165 -59.723 54.548
518252 92.206 -59.725 54.543
518253 92.254 -59.724 54.534
518254 92.301 -59.723 54.520
518255 92.349 -59.722 54.499
518256 92.396 -59.721 54.486
518257 92.440 -59.728 54.479
518258 92.482 -59.738 54.472
518259 92.524 -59.748 54.464
518260 92.567 -59.757 54.452
518261 92.609 -59.767 54.441
518262 92.652 -59.776 54.432
518263 92.694 -59.786 54.422
518264 92.737 -59.796 54.412
518265 92.779 -59.805 54.404
518266 92.821 -59.815 54.393
518267 92.864 -59.825 54.381
518268 92.907 -59.834 54.369
518269 92.949 -59.844 54.360
518270 92.992 -59.854 54.365
518271 93.034 -59.864 54.376
518272 93.076 -59.873 54.388
518273 93.119 -59.883 54.397
518274 93.157 -59.891 54.404
518275 93.195 -59.899 54.408
518276 93.233 -59.909 54.413
518277 93.270 -59.930 54.418
518278 93.306 -59.951 54.425
518279 93.342 -59.971 54.433
518280 93.378 -59.992 54.437
518281 93.415 -60.013 54.436
518282 93.451 -60.033 54.433
518283 93.487 -60.054 54.425
518284 93.523 -60.075 54.413
518285 93.559 -60.096 54.398
518286 93.595 -60.116 54.380
518287 93.632 -60.137 54.368
518288 93.668 -60.158 54.356
518289 93.704 -60.178 54.344
518290 93.742 -60.197 54.332
518291 93.783 -60.212 54.319
518292 93.825 -60.225 54.309
518293 93.867 -60.236 54.300
518294 93.909 -60.244 54.294
518295 93.951 -60.253 54.291
518296 93.992 -60.261 54.284
518297 94.033 -60.269 54.275
518298 94.075 -60.278 54.259
518299 94.116 -60.287 54.240
518300 94.157 -60.295 54.216
518301 94.199 -60.303 54.188
518302 94.241 -60.312 54.161
518303 94.282 -60.320 54.131
518304 94.322 -60.326 54.099
518305 94.360 -60.331 54.064
518306 94.402 -60.336 54.028
518307 94.445 -60.341 53.997
518308 94.487 -60.347 53.975
518309 94.531 -60.353 53.963
518310 94.573 -60.358 53.957
518311 94.616 -60.364 53.951
518312 94.659 -60.369 53.944
518313 94.701 -60.375 53.939
518314 94.740 -60.381 53.932
518315 94.779 -60.387 53.922
518316 94.818 -60.393 53.907
518317 94.858 -60.399 53.889
518318 94.898 -60.405 53.864
518319 94.934 -60.411 53.835
518320 94.971 -60.416 53.805
518321 95.008 -60.422 53.773
518322 95.044 -60.429 53.738
518323 95.079 -60.439 53.703
518324 95.114 -60.449 53.662
518325 95.148 -60.459 53.621
518326 95.183 -60.469 53.582
518327 95.218 -60.478 53.546
518328 95.252 -60.488 53.513
518329 95.288 -60.496 53.478
518330 95.325 -60.503 53.439
518331 95.361 -60.512 53.398
518332 95.398 -60.520 53.357
518333 95.434 -60.529 53.319
518334 95.470 -60.538 53.282
518335 95.507 -60.547 53.243
518336 95.543 -60.555 53.203
518337 95.580 -60.564 53.162
518338 95.616 -60.572 53.122
518339 95.653 -60.582 53.081
518340 95.689 -60.590 53.038
518341 95.726 -60.597 52.993
518342 95.762 -60.604 52.945
518343 95.800 -60.605 52.895
518344 95.839 -60.600 52.845
518345 95.877 -60.594 52.796
518346 95.915 -60.588 52.751
518347 95.955 -60.588 52.709
518348 95.996 -60.593 52.669
518349 96.038 -60.598 52.628
518350 96.075 -60.601 52.590
518351 96.111 -60.604 52.553
518352 96.148 -60.607 52.519
518353 96.184 -60.610 52.487
518354 96.221 -60.613 52.455
518355 96.257 -60.616 52.423
518356 96.295 -60.619 52.392
518357 96.334 -60.619 52.361
518358 96.374 -60.621 52.331
518359 96.413 -60.622 52.302
518360 96.452 -60.622 52.273
518361 96.492 -60.623 52.244
518362 96.533 -60.620 52.215
518363 96.574 -60.617 52.187
518364 96.615 -60.615 52.160
518365 96.656 -60.612 52.136
518366 96.698 -60.613 52.113
518367 96.743 -60.623 52.089
518368 96.788 -60.632 52.061
518369 96.829 -60.643 52.030
518370 96.861 -60.657 51.996
518371 96.893 -60.670 51.961
518372 96.929 -60.682 51.926
518373 96.976 -60.689 51.891
518374 97.022 -60.697 51.857
518375 97.069 -60.704 51.823
518376 97.116 -60.711 51.787
518377 97.163 -60.718 51.750
518378 97.209 -60.726 51.712
518379 97.230 -60.729 51.673
518380 97.245 -60.732 51.634
518381 97.261 -60.735 51.595
518382 97.294 -60.735 51.557
518383 97.330 -60.736 51.520
518384 97.368 -60.737 51.483
518385 97.404 -60.739 51.448
518386 97.440 -60.743 51.413
518387 97.477 -60.737 51.377
518388 97.500 -60.732 51.342
518389 97.522 -60.727 51.306
518390 97.567 -60.729 51.270
518391 97.613 -60.732 51.232
518392 97.659 -60.735 51.194
518393 97.705 -60.737 51.154
518394 97.751 -60.740 51.114
518395 97.797 -60.742 51.072
518396 97.843 -60.745 51.029
518397 97.889 -60.748 50.987
518398 97.937 -60.751 50.944
518399 97.959 -60.758 50.900
518400 97.962 -60.767 50.856
518401 97.966 -60.775 50.813
518402 97.970 -60.783 50.770
518403 97.987 -60.796 50.726
518404 98.011 -60.811 50.684
518405 98.035 -60.827 50.643
518406 98.060 -60.842 50.602
518407 98.090 -60.836 50.561
518408 98.126 -60.841 50.518
518409 98.162 -60.845 50.475
518410 98.199 -60.849 50.434
518411 98.235 -60.854 50.394
518412 98.271 -60.857 50.353
518413 98.307 -60.862 50.312
518414 98.344 -60.867 50.272
518415 98.380 -60.870 50.231
518416 98.416 -60.875 50.192
518417 98.452 -60.879 50.154
518825 87.193 -58.151 57.150
518826 87.233 -58.162 57.131
518827 87.274 -58.173 57.112
518828 87.314 -58.184 57.091
518829 87.354 -58.195 57.072
518830 87.395 -58.206 57.050
518831 87.435 -58.217 57.027
518832 87.475 -58.228 57.003
518833 87.515 -58.238 56.977
518834 87.556 -58.250 56.950
518835 87.596 -58.260 56.919
518836 87.636 -58.271 56.888
518837 87.676 -58.282 56.856
518838 87.719 -58.288 56.823
518839 87.765 -58.299 56.784
518840 87.810 -58.311 56.744
518841 87.855 -58.323 56.704
518842 87.901 -58.335 56.664
518843 87.946 -58.346 56.623
518844 87.992 -58.358 56.582
518845 88.038 -58.369 56.540
518846 88.083 -58.381 56.498
518847 88.128 -58.392 56.456
518848 88.174 -58.404 56.414
518849 88.219 -58.416 56.374
518850 88.265 -58.427 56.335
518851 88.311 -58.438 56.297
518852 88.356 -58.450 56.260
518853 88.402 -58.462 56.226
518854 88.447 -58.473 56.193
518855 88.492 -58.485 56.160
518856 88.538 -58.496 56.130
518857 88.583 -58.508 56.100
518858 88.629 -58.520 56.072
518859 88.675 -58.531 56.046
518860 88.720 -58.542 56.021
518861 88.766 -58.554 55.997
518862 88.811 -58.566 55.974
518863 88.856 -58.577 55.952
518864 88.898 -58.588 55.931
518865 88.939 -58.601 55.912
518866 88.982 -58.623 55.893
518867 89.018 -58.633 55.876
518868 89.053 -58.644 55.859
518869 89.088 -58.654 55.842
518870 89.122 -58.665 55.826
518871 89.158 -58.676 55.813
518872 89.193 -58.687 55.800
518873 89.229 -58.698 55.789
518874 89.264 -58.710 55.777
518875 89.300 -58.720 55.767
518876 89.335 -58.732 55.758
518877 89.369 -58.746 55.751
518878 89.402 -58.758 55.745
518879 89.436 -58.758 55.741
518880 89.469 -58.757 55.739
518881 89.503 -58.757 55.735
518882 89.541 -58.768 55.730
518883 89.583 -58.787 55.724
518884 89.623 -58.805 55.719
518885 89.665 -58.824 55.717
518886 89.705 -58.842 55.719
518887 89.747 -58.861 55.726
518888 89.788 -58.880 55.734
518889 89.829 -58.898 55.741
518890 89.870 -58.917 55.746
518891 89.911 -58.935 55.752
518892 89.952 -58.953 55.756
518893 89.993 -58.972 55.759
518894 90.034 -58.990 55.763
518895 90.075 -59.008 55.767
518896 90.113 -59.015 55.769
518897 90.152 -59.018 55.771
518898 90.190 -59.020 55.769
518899 90.228 -59.023 55.764
518900 90.266 -59.026 55.753
518901 90.304 -59.029 55.729
518902 90.342 -59.031 55.704
518903 90.381 -59.034 55.692
518904 90.419 -59.038 55.674
518905 90.465 -59.053 55.646
518906 90.510 -59.069 55.607
518907 90.555 -59.084 55.561
518908 90.600 -59.099 55.508
518909 90.645 -59.115 55.452
518910 90.690 -59.131 55.396
518911 90.735 -59.153 55.341
518912 90.779 -59.178 55.290
518913 90.823 -59.203 55.243
518914 90.868 -59.227 55.199
518915 90.912 -59.252 55.158
518916 90.956 -59.276 55.118
518917 91.001 -59.301 55.078
518918 91.043 -59.322 55.045
518919 91.082 -59.337 55.012
518920 91.122 -59.352 54.984
518921 91.162 -59.367 54.959
518922 91.201 -59.383 54.938
518923 91.241 -59.398 54.917
518924 91.281 -59.413 54.897
518925 91.320 -59.428 54.879
518926 91.360 -59.444 54.862
518927 91.399 -59.460 54.846
518928 91.439 -59.475 54.832
518929 91.478 -59.490 54.822
518930 91.518 -59.505 54.811
518931 91.558 -59.520 54.803
518932 91.597 -59.536 54.798
518933 91.635 -59.551 54.797
518934 91.672 -59.566 54.791
518935 91.709 -59.581 54.786
518936 91.743 -59.590 54.783
518937 91.766 -59.599 54.783
518938 91.793 -59.604 54.784
518939 91.821 -59.610 54.787
518940 91.849 -59.616 54.786
518941 91.876 -59.621 54.780
518942 91.904 -59.626 54.766
518943 91.931 -59.632 54.744
518944 91.959 -59.637 54.713
518945 91.987 -59.643 54.672
518946 92.014 -59.648 54.629
518947 92.041 -59.654 54.592
518948 92.069 -59.659 54.564
518949 92.097 -59.664 54.547
518950 92.125 -59.670 54.548
518951 92.152 -59.676 54.546
518952 92.179 -59.681 54.543
518953 92.221 -59.682 54.540
518954 92.269 -59.681 54.537
518955 92.316 -59.680 54.528
518956 92.363 -59.679 54.511
518957 92.411 -59.679 54.493
518958 92.456 -59.681 54.477
518959 92.499 -59.691 54.466
518960 92.542 -59.700 54.462
518961 92.584 -59.710 54.454
518962 92.626 -59.720 54.444
518963 92.669 -59.729 54.432
518964 92.711 -59.739 54.421
518965 92.754 -59.748 54.409
518966 92.796 -59.758 54.399
518967 92.839 -59.767 54.389
518968 92.881 -59.777 54.379
518969 92.924 -59.787 54.369
518970 92.966 -59.797 54.359
518971 93.008 -59.806 54.364
518972 93.051 -59.816 54.378
518973 93.093 -59.826 54.392
518974 93.134 -59.835 54.402
518975 93.172 -59.843 54.408
518976 93.210 -59.851 54.406
518977 93.249 -59.861 54.405
518978 93.285 -59.882 54.406
518979 93.321 -59.902 54.410
518980 93.357 -59.923 54.416
518981 93.394 -59.944 54.424
518982 93.430 -59.964 54.429
518983 93.466 -59.985 54.428
518984 93.502 -60.006 54.421
518985 93.538 -60.027 54.409
518986 93.574 -60.047 54.397
518987 93.611 -60.068 54.385
518988 93.647 -60.089 54.375
518989 93.683 -60.109 54.366
518990 93.719 -60.130 54.357
518991 93.755 -60.151 54.347
518992 93.791 -60.171 54.335
518993 93.829 -60.186 54.325
518994 93.869 -60.194 54.314
518995 93.911 -60.202 54.303
518996 93.953 -60.212 54.297
518997 93.994 -60.220 54.291
518998 94.035 -60.228 54.283
518999 94.077 -60.237 54.270
519000 94.118 -60.245 54.253
519001 94.159 -60.253 54.233
519002 94.201 -60.262 54.212
519003 94.243 -60.270 54.190
519004 94.284 -60.278 54.162
519005 94.324 -60.284 54.129
519006 94.364 -60.290 54.092
519007 94.407 -60.295 54.052
519008 94.450 -60.301 54.013
519009 94.493 -60.306 53.980
519010 94.536 -60.312 53.956
519011 94.579 -60.318 53.942
519012 94.622 -60.324 53.936
519013 94.665 -60.329 53.930
519014 94.707 -60.335 53.923
519015 94.747 -60.340 53.913
519016 94.787 -60.347 53.900
519017 94.826 -60.353 53.884
519018 94.866 -60.358 53.864
519019 94.905 -60.364 53.840
519020 94.944 -60.370 53.812
519021 94.980 -60.375 53.782
519022 95.017 -60.382 53.750
519023 95.051 -60.392 53.716
519024 95.086 -60.401 53.682
519025 95.121 -60.411 53.646
519026 95.156 -60.421 53.611
519027 95.190 -60.431 53.578
519028 95.225 -60.441 53.546
519029 95.260 -60.450 53.514
519030 95.296 -60.459 53.478
519031 95.332 -60.466 53.438
519032 95.368 -60.475 53.396
519033 95.405 -60.484 53.353
519034 95.441 -60.493 53.313
519035 95.477 -60.501 53.274
519036 95.514 -60.510 53.233
519037 95.551 -60.518 53.193
519038 95.587 -60.526 53.153
519039 95.624 -60.534 53.114
519040 95.661 -60.542 53.074
519041 95.697 -60.550 53.034
519042 95.734 -60.557 52.993
519043 95.769 -60.563 52.948
519044 95.807 -60.564 52.901
519045 95.846 -60.563 52.853
519046 95.885 -60.560 52.803
519047 95.923 -60.553 52.757
519048 95.962 -60.551 52.715
519049 96.004 -60.557 52.674
519050 96.043 -60.560 52.633
519051 96.079 -60.563 52.594
519052 96.115 -60.566 52.557
519053 96.152 -60.569 52.523
519054 96.188 -60.572 52.490
519055 96.225 -60.575 52.458
519056 96.261 -60.579 52.425
519057 96.299 -60.581 52.394
519058 96.338 -60.582 52.363
519059 96.378 -60.583 52.333
519060 96.417 -60.584 52.303
519061 96.458 -60.581 52.274
519062 96.499 -60.579 52.244
519063 96.540 -60.575 52.215
519064 96.581 -60.572 52.186
519065 96.622 -60.569 52.158
519066 96.664 -60.569 52.133
519067 96.709 -60.578 52.110
519068 96.753 -60.587 52.085
519069 96.798 -60.596 52.059
519070 96.838 -60.607 52.030
519071 96.870 -60.620 51.999
519072 96.902 -60.634 51.966
519073 96.934 -60.647 51.934
519074 96.976 -60.657 51.902
519075 97.022 -60.665 51.870
519076 97.069 -60.672 51.838
519077 97.116 -60.679 51.804
519078 97.163 -60.686 51.768
519079 97.209 -60.694 51.729
519080 97.231 -60.698 51.689
519081 97.246 -60.700 51.649
519082 97.263 -60.703 51.609
519083 97.300 -60.704 51.570
519084 97.337 -60.704 51.533
519085 97.374 -60.705 51.497
519086 97.411 -60.706 51.461
519087 97.447 -60.709 51.426
519088 97.484 -60.713 51.391
519089 97.511 -60.701 51.355
519090 97.533 -60.697 51.319
519091 97.576 -60.698 51.283
519092 97.621 -60.701 51.246
519093 97.667 -60.704 51.207
519094 97.714 -60.706 51.168
519095 97.759 -60.708 51.128
519096 97.805 -60.711 51.087
519097 97.851 -60.713 51.045
519098 97.897 -60.716 51.002
519099 97.943 -60.719 50.959
519100 97.963 -60.724 50.915
519101 97.967 -60.732 50.870
519102 97.970 -60.741 50.826
519103 97.974 -60.748 50.782
519104 97.989 -60.760 50.738
519105 98.013 -60.776 50.696
519106 98.037 -60.792 50.654
519107 98.062 -60.798 50.611
519108 98.096 -60.798 50.569
519109 98.132 -60.803 50.526
519110 98.168 -60.807 50.482
519111 98.205 -60.812 50.440
519112 98.241 -60.816 50.399
519113 98.277 -60.820 50.358
519114 98.313 -60.825 50.316
519115 98.350 -60.829 50.276
519116 98.386 -60.833 50.234
519117 98.423 -60.838 50.194
519118 98.459 -60.842 50.155
519526 87.213 -58.082 57.155
519527 87.254 -58.095 57.136
519528 87.294 -58.109 57.117
519529 87.335 -58.122 57.096
519530 87.375 -58.137 57.077
519531 87.416 -58.150 57.056
519532 87.457 -58.164 57.034
519533 87.497 -58.178 57.010
519534 87.538 -58.191 56.986
519535 87.579 -58.206 56.959
519536 87.619 -58.219 56.931
519537 87.660 -58.233 56.900
519538 87.700 -58.247 56.868
519539 87.743 -58.254 56.836
519540 87.779 -58.259 56.799
519541 87.825 -58.270 56.760
519542 87.870 -58.282 56.721
519543 87.916 -58.294 56.681
519544 87.961 -58.305 56.641
519545 88.006 -58.316 56.600
519546 88.052 -58.328 56.558
519547 88.098 -58.340 56.516
519548 88.143 -58.351 56.474
519549 88.189 -58.363 56.433
519550 88.234 -58.375 56.393
519551 88.280 -58.386 56.352
519552 88.325 -58.397 56.313
519553 88.370 -58.409 56.275
519554 88.416 -58.421 56.239
519555 88.462 -58.432 56.204
519556 88.507 -58.444 56.170
519557 88.553 -58.455 56.138
519558 88.598 -58.466 56.107
519559 88.644 -58.479 56.077
519560 88.689 -58.490 56.051
519561 88.735 -58.501 56.025
519562 88.780 -58.513 56.001
519563 88.823 -58.524 55.977
519564 88.863 -58.535 55.954
519565 88.903 -58.545 55.932
519566 88.946 -58.563 55.914
519567 88.989 -58.585 55.895
519568 89.030 -58.603 55.877
519569 89.065 -58.614 55.860
519570 89.100 -58.624 55.844
519571 89.134 -58.635 55.827
519572 89.170 -58.645 55.814
519573 89.204 -58.655 55.802
519574 89.239 -58.666 55.791
519575 89.274 -58.676 55.780
519576 89.309 -58.687 55.772
519577 89.344 -58.698 55.763
519578 89.380 -58.711 55.758
519579 89.413 -58.714 55.755
519580 89.448 -58.715 55.752
519581 89.484 -58.719 55.753
519582 89.521 -58.723 55.749
519583 89.558 -58.733 55.742
519584 89.597 -58.750 55.734
519585 89.638 -58.768 55.727
519586 89.679 -58.787 55.726
519587 89.720 -58.805 55.730
519588 89.761 -58.824 55.739
519589 89.802 -58.842 55.748
519590 89.843 -58.861 55.755
519591 89.884 -58.879 55.758
519592 89.925 -58.898 55.762
519593 89.966 -58.916 55.761
519594 90.007 -58.935 55.758
519595 90.048 -58.953 55.755
519596 90.089 -58.971 55.752
519597 90.128 -58.981 55.751
519598 90.167 -58.984 55.753
519599 90.205 -58.987 55.749
519600 90.243 -58.990 55.742
519601 90.282 -58.993 55.728
519602 90.319 -58.996 55.703
519603 90.358 -58.998 55.682
519604 90.396 -59.001 55.667
519605 90.434 -59.004 55.643
519606 90.472 -59.007 55.608
519607 90.514 -59.017 55.566
519608 90.559 -59.032 55.517
519609 90.604 -59.047 55.467
519610 90.650 -59.062 55.416
519611 90.694 -59.087 55.367
519612 90.738 -59.111 55.320
519613 90.783 -59.136 55.276
519614 90.827 -59.161 55.234
519615 90.871 -59.185 55.196
519616 90.916 -59.210 55.158
519617 90.960 -59.234 55.119
519618 91.004 -59.259 55.079
519619 91.048 -59.282 55.043
519620 91.087 -59.298 55.010
519621 91.127 -59.313 54.980
519622 91.167 -59.328 54.954
519623 91.206 -59.344 54.931
519624 91.246 -59.359 54.910
519625 91.286 -59.375 54.893
519626 91.325 -59.390 54.875
519627 91.365 -59.405 54.858
519628 91.404 -59.420 54.841
519629 91.444 -59.435 54.828
519630 91.484 -59.451 54.816
519631 91.523 -59.466 54.806
519632 91.563 -59.482 54.796
519633 91.602 -59.497 54.789
519634 91.642 -59.512 54.784
519635 91.681 -59.528 54.780
519636 91.721 -59.543 54.776
519637 91.753 -59.551 54.774
519638 91.780 -59.557 54.774
519639 91.808 -59.562 54.779
519640 91.836 -59.567 54.787
519641 91.863 -59.573 54.790
519642 91.891 -59.578 54.785
519643 91.918 -59.584 54.773
519644 91.946 -59.589 54.751
519645 91.974 -59.595 54.718
519646 92.001 -59.600 54.680
519647 92.029 -59.605 54.641
519648 92.056 -59.611 54.606
519649 92.084 -59.616 54.576
519650 92.111 -59.622 54.552
519651 92.139 -59.627 54.547
519652 92.167 -59.632 54.550
519653 92.194 -59.638 54.550
519654 92.236 -59.639 54.544
519655 92.283 -59.638 54.539
519656 92.331 -59.638 54.537
519657 92.378 -59.637 54.528
519658 92.425 -59.636 54.514
519659 92.473 -59.635 54.496
519660 92.516 -59.643 54.474
519661 92.558 -59.653 54.461
519662 92.601 -59.662 54.457
519663 92.643 -59.672 54.449
519664 92.686 -59.682 54.435
519665 92.728 -59.692 54.420
519666 92.771 -59.701 54.404
519667 92.813 -59.710 54.391
519668 92.856 -59.720 54.381
519669 92.898 -59.729 54.373
519670 92.940 -59.739 54.365
519671 92.983 -59.749 54.358
519672 93.025 -59.759 54.363
519673 93.068 -59.768 54.379
519674 93.110 -59.778 54.396
519675 93.148 -59.786 54.408
519676 93.187 -59.795 54.411
519677 93.225 -59.804 54.411
519678 93.264 -59.813 54.411
519679 93.300 -59.833 54.412
519680 93.336 -59.854 54.412
519681 93.372 -59.875 54.411
519682 93.409 -59.895 54.410
519683 93.445 -59.916 54.413
519684 93.481 -59.937 54.413
519685 93.517 -59.958 54.407
519686 93.553 -59.978 54.396
519687 93.589 -59.999 54.384
519688 93.626 -60.020 54.375
519689 93.662 -60.040 54.363
519690 93.698 -60.061 54.354
519691 93.734 -60.082 54.347
519692 93.770 -60.102 54.343
519693 93.806 -60.120 54.337
519694 93.843 -60.135 54.327
519695 93.880 -60.149 54.314
519696 93.921 -60.159 54.303
519697 93.961 -60.167 54.293
519698 94.001 -60.176 54.288
519699 94.041 -60.185 54.280
519700 94.081 -60.194 54.269
519701 94.122 -60.203 54.257
519702 94.161 -60.212 54.243
519703 94.203 -60.220 54.230
519704 94.245 -60.229 54.215
519705 94.286 -60.237 54.190
519706 94.327 -60.243 54.158
519707 94.369 -60.249 54.121
519708 94.412 -60.255 54.079
519709 94.455 -60.261 54.033
519710 94.498 -60.266 53.991
519711 94.541 -60.272 53.956
519712 94.584 -60.278 53.933
519713 94.628 -60.284 53.922
519714 94.671 -60.289 53.915
519715 94.714 -60.295 53.905
519716 94.755 -60.300 53.892
519717 94.794 -60.306 53.876
519718 94.834 -60.312 53.859
519719 94.874 -60.317 53.839
519720 94.913 -60.323 53.816
519721 94.953 -60.328 53.790
519722 94.989 -60.334 53.761
519723 95.024 -60.344 53.731
519724 95.059 -60.354 53.700
519725 95.094 -60.364 53.669
519726 95.128 -60.374 53.636
519727 95.163 -60.384 53.605
519728 95.198 -60.393 53.575
519729 95.233 -60.403 53.545
519730 95.267 -60.412 53.513
519731 95.303 -60.422 53.476
519732 95.339 -60.430 53.435
519733 95.376 -60.438 53.393
519734 95.412 -60.447 53.351
519735 95.449 -60.454 53.310
519736 95.485 -60.463 53.271
519737 95.522 -60.471 53.231
519738 95.559 -60.478 53.190
519739 95.596 -60.487 53.149
519740 95.632 -60.494 53.108
519741 95.669 -60.503 53.067
519742 95.705 -60.510 53.027
519743 95.741 -60.516 52.988
519744 95.777 -60.520 52.947
519745 95.815 -60.523 52.903
519746 95.853 -60.522 52.854
519747 95.892 -60.522 52.806
519748 95.931 -60.519 52.762
519749 95.970 -60.515 52.722
519750 96.011 -60.519 52.682
519751 96.047 -60.522 52.640
519752 96.083 -60.525 52.600
519753 96.120 -60.528 52.563
519754 96.156 -60.532 52.528
519755 96.192 -60.535 52.495
519756 96.229 -60.538 52.463
519757 96.265 -60.541 52.429
519758 96.303 -60.543 52.398
519759 96.342 -60.544 52.367
519760 96.383 -60.542 52.337
519761 96.425 -60.539 52.307
519762 96.465 -60.536 52.277
519763 96.507 -60.533 52.246
519764 96.547 -60.531 52.216
519765 96.589 -60.528 52.187
519766 96.629 -60.525 52.158
519767 96.674 -60.532 52.130
519768 96.719 -60.541 52.105
519769 96.763 -60.550 52.079
519770 96.808 -60.560 52.053
519771 96.848 -60.570 52.026
519772 96.880 -60.584 51.998
519773 96.912 -60.598 51.969
519774 96.944 -60.612 51.939
519775 96.976 -60.626 51.910
519776 97.022 -60.633 51.881
519777 97.069 -60.640 51.850
519778 97.116 -60.647 51.817
519779 97.163 -60.654 51.780
519780 97.209 -60.662 51.743
519781 97.231 -60.666 51.703
519782 97.246 -60.668 51.662
519783 97.270 -60.670 51.621
519784 97.307 -60.671 51.581
519785 97.344 -60.672 51.543
519786 97.380 -60.673 51.506
519787 97.418 -60.673 51.471
519788 97.454 -60.675 51.436
519789 97.491 -60.679 51.401
519790 97.522 -60.671 51.365
519791 97.544 -60.666 51.330
519792 97.584 -60.667 51.294
519793 97.630 -60.669 51.257
519794 97.676 -60.673 51.219
519795 97.722 -60.675 51.180
519796 97.768 -60.677 51.140
519797 97.814 -60.680 51.099
519798 97.860 -60.682 51.058
519799 97.906 -60.685 51.015
519800 97.938 -60.688 50.973
519801 97.963 -60.691 50.928
519802 97.971 -60.698 50.883
519803 97.975 -60.706 50.839
519804 97.979 -60.714 50.794
519805 97.991 -60.726 50.750
519806 98.015 -60.741 50.707
519807 98.039 -60.757 50.664
519808 98.066 -60.757 50.622
519809 98.102 -60.761 50.578
519810 98.138 -60.766 50.535
519811 98.174 -60.770 50.490
519812 98.211 -60.774 50.447
519813 98.247 -60.779 50.405
519814 98.284 -60.783 50.362
519815 98.320 -60.787 50.320
519816 98.356 -60.792 50.278
519817 98.392 -60.796 50.236
519818 98.429 -60.801 50.195
519819 98.465 -60.804 50.156
520227 87.238 -58.056 57.159
520228 87.279 -58.069 57.140
520229 87.320 -58.083 57.121
520230 87.360 -58.096 57.101
520231 87.401 -58.111 57.081
520232 87.441 -58.125 57.060
520233 87.482 -58.138 57.039
520234 87.523 -58.152 57.017
520235 87.563 -58.165 56.993
520236 87.604 -58.180 56.967
520237 87.645 -58.194 56.939
520238 87.685 -58.207 56.910
520239 87.726 -58.221 56.879
520240 87.768 -58.225 56.848
520241 87.811 -58.227 56.812
520242 87.846 -58.232 56.775
520243 87.885 -58.241 56.737
520244 87.930 -58.253 56.698
520245 87.976 -58.264 56.659
520246 88.022 -58.275 56.618
520247 88.067 -58.288 56.576
520248 88.112 -58.299 56.535
520249 88.158 -58.310 56.493
520250 88.203 -58.322 56.451
520251 88.249 -58.334 56.410
520252 88.294 -58.345 56.369
520253 88.340 -58.357 56.329
520254 88.385 -58.368 56.290
520255 88.431 -58.379 56.253
520256 88.476 -58.391 56.216
520257 88.522 -58.403 56.182
520258 88.567 -58.414 56.149
520259 88.613 -58.426 56.116
520260 88.659 -58.438 56.085
520261 88.704 -58.449 56.057
520262 88.747 -58.460 56.031
520263 88.787 -58.470 56.005
520264 88.828 -58.481 55.980
520265 88.868 -58.491 55.958
520266 88.909 -58.505 55.935
520267 88.952 -58.526 55.916
520268 88.995 -58.547 55.897
520269 89.038 -58.568 55.879
520270 89.077 -58.583 55.862
520271 89.112 -58.594 55.846
520272 89.146 -58.604 55.829
520273 89.182 -58.614 55.816
520274 89.216 -58.625 55.805
520275 89.251 -58.636 55.794
520276 89.286 -58.646 55.784
520277 89.321 -58.657 55.776
520278 89.356 -58.667 55.769
520279 89.391 -58.677 55.764
520280 89.433 -58.682 55.761
520281 89.469 -58.686 55.757
520282 89.506 -58.689 55.756
520283 89.542 -58.693 55.752
520284 89.579 -58.699 55.745
520285 89.618 -58.717 55.739
520286 89.657 -58.733 55.735
520287 89.695 -58.751 55.736
520288 89.734 -58.768 55.744
520289 89.775 -58.786 55.755
520290 89.816 -58.805 55.764
520291 89.858 -58.824 55.770
520292 89.898 -58.842 55.772
520293 89.939 -58.861 55.772
520294 89.980 -58.879 55.766
520295 90.021 -58.898 55.759
520296 90.062 -58.916 55.750
520297 90.103 -58.934 55.743
520298 90.144 -58.948 55.735
520299 90.182 -58.951 55.731
520300 90.220 -58.954 55.723
520301 90.258 -58.956 55.711
520302 90.296 -58.959 55.695
520303 90.334 -58.962 55.671
520304 90.373 -58.965 55.651
520305 90.411 -58.968 55.631
520306 90.449 -58.971 55.604
520307 90.487 -58.974 55.566
520308 90.525 -58.976 55.523
520309 90.564 -58.980 55.479
520310 90.609 -58.996 55.434
520311 90.653 -59.020 55.388
520312 90.697 -59.045 55.344
520313 90.742 -59.070 55.304
520314 90.786 -59.094 55.266
520315 90.830 -59.119 55.231
520316 90.875 -59.143 55.199
520317 90.920 -59.168 55.165
520318 90.964 -59.193 55.127
520319 91.008 -59.218 55.086
520320 91.052 -59.243 55.047
520321 91.092 -59.259 55.011
520322 91.132 -59.274 54.978
520323 91.172 -59.290 54.950
520324 91.211 -59.305 54.926
520325 91.251 -59.320 54.905
520326 91.291 -59.335 54.887
520327 91.330 -59.350 54.870
520328 91.370 -59.366 54.854
520329 91.410 -59.381 54.837
520330 91.449 -59.397 54.823
520331 91.489 -59.412 54.811
520332 91.528 -59.427 54.799
520333 91.568 -59.443 54.790
520334 91.607 -59.458 54.782
520335 91.647 -59.473 54.774
520336 91.687 -59.488 54.768
520337 91.731 -59.501 54.763
520338 91.767 -59.509 54.760
520339 91.795 -59.513 54.761
520340 91.822 -59.519 54.769
520341 91.850 -59.525 54.781
520342 91.878 -59.530 54.788
520343 91.906 -59.535 54.787
520344 91.933 -59.541 54.776
520345 91.960 -59.547 54.757
520346 91.988 -59.552 54.726
520347 92.016 -59.557 54.691
520348 92.044 -59.563 54.657
520349 92.071 -59.568 54.626
520350 92.098 -59.574 54.596
520351 92.126 -59.579 54.567
520352 92.154 -59.585 54.553
520353 92.181 -59.590 54.554
520354 92.209 -59.595 54.557
520355 92.250 -59.597 54.554
520356 92.298 -59.596 54.547
520357 92.345 -59.595 54.548
520358 92.392 -59.595 54.535
520359 92.440 -59.594 54.517
520360 92.487 -59.593 54.497
520361 92.533 -59.595 54.477
520362 92.575 -59.605 54.464
520363 92.618 -59.615 54.456
520364 92.660 -59.624 54.444
520365 92.703 -59.634 54.430
520366 92.745 -59.644 54.418
520367 92.788 -59.654 54.401
520368 92.830 -59.663 54.389
520369 92.872 -59.673 54.376
520370 92.915 -59.682 54.365
520371 92.957 -59.692 54.358
520372 93.000 -59.701 54.356
520373 93.042 -59.711 54.361
520374 93.085 -59.721 54.379
520375 93.125 -59.730 54.400
520376 93.163 -59.739 54.416
520377 93.202 -59.747 54.414
520378 93.240 -59.755 54.419
520379 93.279 -59.764 54.423
520380 93.315 -59.785 54.423
520381 93.351 -59.806 54.418
520382 93.388 -59.826 54.410
520383 93.424 -59.847 54.399
520384 93.460 -59.868 54.398
520385 93.496 -59.889 54.396
520386 93.532 -59.909 54.391
520387 93.568 -59.930 54.382
520388 93.605 -59.951 54.371
520389 93.641 -59.971 54.362
520390 93.677 -59.992 54.351
520391 93.713 -60.013 54.341
520392 93.749 -60.033 54.334
520393 93.786 -60.052 54.327
520394 93.822 -60.067 54.320
520395 93.858 -60.083 54.312
520396 93.895 -60.098 54.303
520397 93.932 -60.113 54.292
520398 93.972 -60.122 54.281
520399 94.012 -60.131 54.273
520400 94.052 -60.140 54.265
520401 94.092 -60.149 54.255
520402 94.133 -60.158 54.247
520403 94.173 -60.167 54.240
520404 94.213 -60.176 54.230
520405 94.253 -60.184 54.221
520406 94.293 -60.193 54.201
520407 94.334 -60.202 54.174
520408 94.376 -60.209 54.141
520409 94.419 -60.215 54.102
520410 94.462 -60.221 54.054
520411 94.505 -60.226 54.009
520412 94.548 -60.232 53.966
520413 94.592 -60.238 53.929
520414 94.634 -60.243 53.907
520415 94.678 -60.249 53.897
520416 94.720 -60.255 53.883
520417 94.762 -60.261 53.868
520418 94.802 -60.266 53.850
520419 94.841 -60.271 53.831
520420 94.881 -60.277 53.813
520421 94.921 -60.282 53.792
520422 94.961 -60.288 53.769
520423 94.996 -60.297 53.743
520424 95.031 -60.307 53.715
520425 95.066 -60.317 53.687
520426 95.101 -60.327 53.658
520427 95.135 -60.337 53.630
520428 95.170 -60.347 53.601
520429 95.205 -60.356 53.572
520430 95.240 -60.366 53.542
520431 95.274 -60.375 53.510
520432 95.310 -60.383 53.472
520433 95.347 -60.391 53.431
520434 95.384 -60.399 53.389
520435 95.420 -60.407 53.348
520436 95.457 -60.416 53.308
520437 95.494 -60.423 53.269
520438 95.530 -60.431 53.230
520439 95.567 -60.439 53.190
520440 95.604 -60.447 53.149
520441 95.640 -60.455 53.107
520442 95.677 -60.463 53.066
520443 95.713 -60.469 53.024
520444 95.749 -60.474 52.983
520445 95.785 -60.478 52.942
520446 95.822 -60.481 52.900
520447 95.861 -60.481 52.857
520448 95.899 -60.481 52.816
520449 95.938 -60.481 52.775
520450 95.977 -60.479 52.735
520451 96.015 -60.481 52.695
520452 96.051 -60.485 52.652
520453 96.088 -60.488 52.612
520454 96.124 -60.491 52.575
520455 96.160 -60.494 52.539
520456 96.197 -60.497 52.505
520457 96.233 -60.500 52.473
520458 96.270 -60.503 52.440
520459 96.309 -60.503 52.408
520460 96.349 -60.500 52.376
520461 96.391 -60.497 52.345
520462 96.432 -60.494 52.314
520463 96.473 -60.491 52.284
520464 96.514 -60.488 52.252
520465 96.555 -60.486 52.220
520466 96.596 -60.483 52.189
520467 96.639 -60.487 52.160
520468 96.683 -60.496 52.131
520469 96.728 -60.505 52.104
520470 96.773 -60.514 52.075
520471 96.817 -60.523 52.047
520472 96.858 -60.535 52.020
520473 96.889 -60.548 51.993
520474 96.922 -60.562 51.968
520475 96.953 -60.575 51.942
520476 96.985 -60.589 51.915
520477 97.022 -60.601 51.885
520478 97.069 -60.608 51.854
520479 97.116 -60.616 51.821
520480 97.163 -60.622 51.784
520481 97.209 -60.630 51.748
520482 97.232 -60.634 51.710
520483 97.247 -60.637 51.671
520484 97.277 -60.638 51.630
520485 97.313 -60.638 51.589
520486 97.350 -60.639 51.549
520487 97.387 -60.640 51.511
520488 97.424 -60.641 51.475
520489 97.461 -60.641 51.440
520490 97.497 -60.644 51.407
520491 97.533 -60.641 51.373
520492 97.555 -60.636 51.338
520493 97.593 -60.636 51.302
520494 97.639 -60.638 51.266
520495 97.684 -60.641 51.229
520496 97.731 -60.644 51.191
520497 97.777 -60.646 51.151
520498 97.822 -60.649 51.110
520499 97.868 -60.651 51.069
520500 97.903 -60.654 51.026
520501 97.929 -60.657 50.984
520502 97.954 -60.660 50.940
520503 97.976 -60.663 50.895
520504 97.980 -60.672 50.851
520505 97.984 -60.680 50.805
520506 97.992 -60.690 50.761
520507 98.016 -60.706 50.718
520508 98.040 -60.721 50.675
520509 98.072 -60.720 50.632
520510 98.108 -60.724 50.588
520511 98.145 -60.729 50.544
520512 98.181 -60.732 50.497
520513 98.217 -60.737 50.454
520514 98.253 -60.742 50.411
520515 98.290 -60.745 50.367
520516 98.326 -60.750 50.324
520517 98.362 -60.754 50.281
520518 98.398 -60.759 50.238
520519 98.435 -60.763 50.197
520520 98.471 -60.767 50.157
520521 98.507 -60.772 50.116
520927 87.223 -58.015 57.180
520928 87.264 -58.030 57.161
520929 87.304 -58.043 57.143
520930 87.345 -58.057 57.125
520931 87.386 -58.071 57.105
520932 87.426 -58.084 57.085
520933 87.467 -58.099 57.064
520934 87.508 -58.112 57.043
520935 87.548 -58.126 57.022
520936 87.589 -58.140 56.998
520937 87.630 -58.153 56.973
520938 87.670 -58.168 56.947
520939 87.711 -58.181 56.919
520940 87.752 -58.193 56.888
520941 87.794 -58.195 56.857
520942 87.837 -58.197 56.823
520943 87.880 -58.200 56.787
520944 87.914 -58.205 56.750
520945 87.945 -58.212 56.712
520946 87.990 -58.223 56.673
520947 88.036 -58.234 56.634
520948 88.081 -58.247 56.594
520949 88.127 -58.258 56.553
520950 88.173 -58.269 56.512
520951 88.218 -58.281 56.470
520952 88.263 -58.292 56.428
520953 88.309 -58.304 56.386
520954 88.354 -58.316 56.345
520955 88.400 -58.327 56.306
520956 88.446 -58.338 56.267
520957 88.491 -58.350 56.230
520958 88.537 -58.362 56.194
520959 88.582 -58.373 56.160
520960 88.627 -58.385 56.127
520961 88.672 -58.396 56.095
520962 88.712 -58.407 56.065
520963 88.753 -58.417 56.038
520964 88.793 -58.428 56.011
520965 88.833 -58.438 55.986
520966 88.874 -58.448 55.962
520967 88.916 -58.468 55.939
520968 88.959 -58.488 55.919
520969 89.002 -58.510 55.900
520970 89.045 -58.530 55.882
520971 89.088 -58.551 55.865
520972 89.124 -58.563 55.847
520973 89.159 -58.573 55.833
520974 89.194 -58.584 55.819
520975 89.229 -58.595 55.808
520976 89.263 -58.605 55.797
520977 89.299 -58.615 55.788
520978 89.333 -58.626 55.781
520979 89.368 -58.636 55.774
520980 89.405 -58.646 55.770
520981 89.452 -58.652 55.766
520982 89.490 -58.657 55.761
520983 89.527 -58.661 55.759
520984 89.563 -58.664 55.755
520985 89.600 -58.667 55.750
520986 89.638 -58.683 55.747
520987 89.677 -58.700 55.748
520988 89.716 -58.717 55.754
520989 89.755 -58.734 55.763
520990 89.794 -58.751 55.772
520991 89.832 -58.769 55.778
520992 89.872 -58.786 55.781
520993 89.912 -58.805 55.781
520994 89.954 -58.824 55.777
520995 89.994 -58.842 55.770
520996 90.036 -58.861 55.760
520997 90.077 -58.879 55.749
520998 90.118 -58.897 55.736
520999 90.159 -58.914 55.722
521000 90.197 -58.918 55.709
521001 90.235 -58.921 55.695
521002 90.273 -58.923 55.679
521003 90.311 -58.926 55.662
521004 90.350 -58.929 55.639
521005 90.388 -58.932 55.619
521006 90.426 -58.934 55.596
521007 90.464 -58.937 55.566
521008 90.502 -58.940 55.530
521009 90.540 -58.944 55.491
521010 90.579 -58.949 55.451
521011 90.617 -58.955 55.411
521012 90.657 -58.978 55.371
521013 90.701 -59.003 55.333
521014 90.746 -59.028 55.299
521015 90.790 -59.052 55.270
521016 90.834 -59.077 55.240
521017 90.879 -59.102 55.211
521018 90.923 -59.127 55.180
521019 90.967 -59.152 55.144
521020 91.012 -59.176 55.102
521021 91.056 -59.201 55.058
521022 91.097 -59.220 55.016
521023 91.137 -59.235 54.980
521024 91.177 -59.250 54.950
521025 91.216 -59.265 54.924
521026 91.256 -59.281 54.901
521027 91.296 -59.297 54.882
521028 91.335 -59.312 54.865
521029 91.375 -59.327 54.850
521030 91.414 -59.343 54.834
521031 91.454 -59.358 54.820
521032 91.494 -59.373 54.806
521033 91.533 -59.388 54.794
521034 91.573 -59.403 54.784
521035 91.612 -59.419 54.776
521036 91.655 -59.433 54.768
521037 91.700 -59.446 54.763
521038 91.739 -59.460 54.757
521039 91.775 -59.470 54.753
521040 91.810 -59.471 54.752
521041 91.838 -59.476 54.757
521042 91.865 -59.482 54.766
521043 91.892 -59.488 54.772
521044 91.920 -59.493 54.775
521045 91.948 -59.498 54.773
521046 91.976 -59.504 54.764
521047 92.003 -59.509 54.742
521048 92.030 -59.515 54.712
521049 92.058 -59.520 54.681
521050 92.086 -59.526 54.651
521051 92.113 -59.531 54.622
521052 92.141 -59.536 54.592
521053 92.168 -59.542 54.566
521054 92.196 -59.547 54.559
521055 92.224 -59.553 54.564
521056 92.265 -59.554 54.565
521057 92.312 -59.554 54.560
521058 92.360 -59.552 54.559
521059 92.407 -59.551 54.541
521060 92.454 -59.551 54.519
521061 92.502 -59.550 54.501
521062 92.549 -59.549 54.487
521063 92.592 -59.557 54.473
521064 92.635 -59.567 54.462
521065 92.677 -59.577 54.449
521066 92.720 -59.586 54.436
521067 92.762 -59.596 54.423
521068 92.804 -59.606 54.408
521069 92.847 -59.616 54.393
521070 92.889 -59.625 54.379
521071 92.932 -59.635 54.367
521072 92.974 -59.645 54.357
521073 93.017 -59.654 54.352
521074 93.059 -59.663 54.356
521075 93.101 -59.673 54.369
521076 93.140 -59.682 54.388
521077 93.179 -59.690 54.406
521078 93.217 -59.699 54.416
521079 93.256 -59.707 54.420
521080 93.294 -59.716 54.419
521081 93.330 -59.737 54.414
521082 93.366 -59.757 54.408
521083 93.403 -59.778 54.404
521084 93.439 -59.799 54.393
521085 93.475 -59.820 54.388
521086 93.511 -59.840 54.383
521087 93.547 -59.861 54.376
521088 93.583 -59.882 54.367
521089 93.620 -59.902 54.359
521090 93.656 -59.924 54.349
521091 93.692 -59.944 54.339
521092 93.728 -59.964 54.330
521093 93.765 -59.983 54.320
521094 93.801 -59.999 54.307
521095 93.837 -60.014 54.297
521096 93.874 -60.030 54.291
521097 93.910 -60.046 54.284
521098 93.947 -60.061 54.275
521099 93.983 -60.077 54.264
521100 94.024 -60.086 54.253
521101 94.063 -60.095 54.245
521102 94.104 -60.104 54.237
521103 94.144 -60.113 54.232
521104 94.184 -60.121 54.230
521105 94.224 -60.130 54.222
521106 94.264 -60.140 54.216
521107 94.305 -60.149 54.201
521108 94.345 -60.157 54.178
521109 94.387 -60.165 54.150
521110 94.428 -60.172 54.114
521111 94.470 -60.180 54.069
521112 94.512 -60.187 54.023
521113 94.555 -60.193 53.976
521114 94.598 -60.198 53.930
521115 94.641 -60.204 53.894
521116 94.684 -60.209 53.873
521117 94.728 -60.215 53.854
521118 94.769 -60.221 53.835
521119 94.809 -60.226 53.817
521120 94.848 -60.232 53.798
521121 94.889 -60.237 53.783
521122 94.928 -60.244 53.765
521123 94.967 -60.253 53.746
521124 95.003 -60.262 53.723
521125 95.038 -60.271 53.699
521126 95.072 -60.281 53.675
521127 95.107 -60.291 53.649
521128 95.142 -60.301 53.622
521129 95.177 -60.310 53.595
521130 95.211 -60.320 53.566
521131 95.246 -60.330 53.535
521132 95.284 -60.339 53.502
521133 95.320 -60.346 53.466
521134 95.355 -60.352 53.426
521135 95.392 -60.359 53.385
521136 95.428 -60.368 53.345
521137 95.465 -60.376 53.306
521138 95.502 -60.384 53.268
521139 95.539 -60.392 53.230
521140 95.576 -60.400 53.192
521141 95.612 -60.408 53.152
521142 95.649 -60.416 53.111
521143 95.685 -60.422 53.069
521144 95.721 -60.427 53.023
521145 95.756 -60.431 52.980
521146 95.793 -60.437 52.938
521147 95.829 -60.441 52.897
521148 95.868 -60.441 52.862
521149 95.906 -60.440 52.826
521150 95.946 -60.439 52.789
521151 95.984 -60.439 52.750
521152 96.022 -60.439 52.709
521153 96.058 -60.442 52.667
521154 96.095 -60.445 52.627
521155 96.131 -60.447 52.590
521156 96.168 -60.450 52.554
521157 96.204 -60.453 52.519
521158 96.241 -60.456 52.487
521159 96.277 -60.459 52.453
521160 96.316 -60.458 52.420
521161 96.357 -60.455 52.387
521162 96.398 -60.452 52.355
521163 96.439 -60.450 52.324
521164 96.480 -60.447 52.293
521165 96.521 -60.444 52.260
521166 96.562 -60.441 52.227
521167 96.604 -60.441 52.194
521168 96.649 -60.450 52.163
521169 96.693 -60.459 52.133
521170 96.738 -60.469 52.105
521171 96.783 -60.478 52.074
521172 96.827 -60.488 52.044
521173 96.867 -60.498 52.016
521174 96.899 -60.512 51.988
521175 96.931 -60.525 51.962
521176 96.963 -60.540 51.938
521177 96.995 -60.553 51.912
521178 97.027 -60.567 51.884
521179 97.069 -60.576 51.852
521180 97.116 -60.584 51.820
521181 97.163 -60.591 51.783
521182 97.209 -60.598 51.746
521183 97.232 -60.602 51.709
521184 97.248 -60.605 51.671
521185 97.283 -60.606 51.633
521186 97.320 -60.607 51.592
521187 97.357 -60.607 51.552
521188 97.393 -60.608 51.513
521189 97.430 -60.609 51.475
521190 97.467 -60.610 51.440
521191 97.504 -60.611 51.407
521192 97.540 -60.614 51.374
521193 97.566 -60.606 51.340
521194 97.601 -60.605 51.305
521195 97.647 -60.607 51.270
521196 97.693 -60.610 51.235
521197 97.739 -60.613 51.198
521198 97.785 -60.615 51.159
521199 97.831 -60.617 51.119
521200 97.868 -60.620 51.079
521201 97.893 -60.623 51.037
521202 97.919 -60.626 50.994
521203 97.944 -60.629 50.950
521204 97.969 -60.632 50.906
521205 97.984 -60.638 50.861
521206 97.988 -60.646 50.816
521207 97.994 -60.655 50.772
521208 98.018 -60.670 50.729
521209 98.043 -60.679 50.687
521210 98.078 -60.682 50.644
521211 98.114 -60.687 50.598
521212 98.151 -60.691 50.551
521213 98.187 -60.695 50.505
521214 98.223 -60.700 50.461
521215 98.260 -60.704 50.417
521216 98.296 -60.708 50.372
521217 98.332 -60.713 50.328
521218 98.368 -60.717 50.284
521219 98.405 -60.721 50.240
521220 98.441 -60.726 50.198
521221 98.478 -60.730 50.158
521222 98.514 -60.734 50.117
521628 87.249 -57.990 57.181
521629 87.289 -58.004 57.163
521630 87.330 -58.018 57.145
521631 87.371 -58.031 57.127
521632 87.411 -58.045 57.107
521633 87.452 -58.059 57.087
521634 87.493 -58.072 57.067
521635 87.533 -58.087 57.047
521636 87.574 -58.100 57.025
521637 87.614 -58.114 57.003
521638 87.655 -58.128 56.979
521639 87.696 -58.141 56.953
521640 87.736 -58.156 56.926
521641 87.778 -58.163 56.897
521642 87.821 -58.165 56.866
521643 87.863 -58.168 56.832
521644 87.906 -58.170 56.798
521645 87.949 -58.172 56.763
521646 87.981 -58.178 56.726
521647 88.012 -58.185 56.688
521648 88.051 -58.194 56.650
521649 88.096 -58.205 56.611
521650 88.141 -58.217 56.571
521651 88.187 -58.228 56.531
521652 88.233 -58.240 56.489
521653 88.278 -58.251 56.446
521654 88.324 -58.263 56.404
521655 88.369 -58.275 56.363
521656 88.415 -58.286 56.322
521657 88.460 -58.297 56.283
521658 88.505 -58.310 56.245
521659 88.551 -58.321 56.207
521660 88.597 -58.332 56.172
521661 88.637 -58.343 56.138
521662 88.677 -58.354 56.105
521663 88.717 -58.363 56.075
521664 88.758 -58.374 56.046
521665 88.798 -58.385 56.019
521666 88.838 -58.395 55.993
521667 88.880 -58.409 55.968
521668 88.922 -58.430 55.945
521669 88.966 -58.451 55.923
521670 89.009 -58.472 55.904
521671 89.051 -58.493 55.885
521672 89.095 -58.513 55.867
521673 89.136 -58.532 55.851
521674 89.171 -58.543 55.836
521675 89.206 -58.554 55.823
521676 89.241 -58.563 55.812
521677 89.275 -58.574 55.803
521678 89.311 -58.585 55.794
521679 89.345 -58.595 55.786
521680 89.380 -58.606 55.780
521681 89.418 -58.615 55.775
521682 89.465 -58.621 55.770
521683 89.512 -58.627 55.766
521684 89.548 -58.631 55.763
521685 89.585 -58.635 55.760
521686 89.621 -58.639 55.758
521687 89.659 -58.648 55.759
521688 89.698 -58.666 55.765
521689 89.737 -58.683 55.774
521690 89.775 -58.701 55.780
521691 89.814 -58.718 55.785
521692 89.853 -58.735 55.787
521693 89.892 -58.752 55.784
521694 89.931 -58.770 55.780
521695 89.969 -58.787 55.772
521696 90.009 -58.805 55.763
521697 90.050 -58.824 55.752
521698 90.091 -58.842 55.739
521699 90.132 -58.860 55.724
521700 90.173 -58.879 55.708
521701 90.212 -58.884 55.689
521702 90.250 -58.887 55.670
521703 90.288 -58.890 55.650
521704 90.326 -58.892 55.631
521705 90.365 -58.896 55.608
521706 90.403 -58.899 55.587
521707 90.441 -58.902 55.561
521708 90.479 -58.908 55.531
521709 90.518 -58.912 55.499
521710 90.556 -58.918 55.465
521711 90.595 -58.922 55.429
521712 90.633 -58.927 55.393
521713 90.669 -58.941 55.358
521714 90.705 -58.961 55.326
521715 90.749 -58.987 55.298
521716 90.794 -59.011 55.273
521717 90.838 -59.036 55.247
521718 90.882 -59.061 55.218
521719 90.927 -59.085 55.187
521720 90.971 -59.110 55.153
521721 91.015 -59.134 55.114
521722 91.060 -59.159 55.072
521723 91.102 -59.181 55.030
521724 91.142 -59.196 54.990
521725 91.182 -59.212 54.954
521726 91.221 -59.227 54.925
521727 91.261 -59.242 54.900
521728 91.301 -59.258 54.879
521729 91.340 -59.273 54.861
521730 91.380 -59.288 54.846
521731 91.419 -59.303 54.832
521732 91.459 -59.319 54.817
521733 91.499 -59.334 54.802
521734 91.538 -59.350 54.790
521735 91.578 -59.365 54.781
521736 91.622 -59.378 54.773
521737 91.667 -59.391 54.766
521738 91.711 -59.403 54.761
521739 91.747 -59.418 54.755
521740 91.784 -59.432 54.750
521741 91.819 -59.432 54.747
521742 91.852 -59.434 54.748
521743 91.880 -59.439 54.753
521744 91.908 -59.444 54.758
521745 91.935 -59.450 54.762
521746 91.962 -59.456 54.766
521747 91.990 -59.461 54.763
521748 92.018 -59.466 54.748
521749 92.046 -59.472 54.724
521750 92.073 -59.478 54.697
521751 92.100 -59.483 54.672
521752 92.128 -59.488 54.648
521753 92.156 -59.494 54.622
521754 92.183 -59.499 54.592
521755 92.211 -59.505 54.571
521756 92.238 -59.510 54.565
521757 92.280 -59.511 54.573
521758 92.327 -59.510 54.571
521759 92.374 -59.510 54.567
521760 92.422 -59.509 54.552
521761 92.469 -59.508 54.534
521762 92.516 -59.507 54.523
521763 92.564 -59.507 54.511
521764 92.609 -59.510 54.490
521765 92.652 -59.519 54.477
521766 92.694 -59.529 54.465
521767 92.736 -59.539 54.451
521768 92.779 -59.549 54.436
521769 92.821 -59.558 54.422
521770 92.864 -59.568 54.406
521771 92.906 -59.578 54.390
521772 92.949 -59.587 54.376
521773 92.991 -59.597 54.363
521774 93.034 -59.607 54.354
521775 93.076 -59.616 54.353
521776 93.117 -59.626 54.360
521777 93.155 -59.634 54.373
521778 93.194 -59.642 54.389
521779 93.232 -59.651 54.407
521780 93.271 -59.659 54.415
521781 93.309 -59.668 54.416
521782 93.345 -59.689 54.412
521783 93.381 -59.709 54.405
521784 93.418 -59.729 54.402
521785 93.454 -59.751 54.394
521786 93.490 -59.771 54.389
521787 93.526 -59.792 54.385
521788 93.562 -59.813 54.379
521789 93.599 -59.833 54.372
521790 93.635 -59.855 54.364
521791 93.671 -59.875 54.356
521792 93.707 -59.895 54.347
521793 93.744 -59.914 54.337
521794 93.780 -59.930 54.324
521795 93.817 -59.946 54.307
521796 93.853 -59.961 54.289
521797 93.889 -59.977 54.271
521798 93.926 -59.992 54.258
521799 93.962 -60.008 54.250
521800 93.999 -60.024 54.242
521801 94.035 -60.040 54.230
521802 94.074 -60.050 54.222
521803 94.115 -60.058 54.215
521804 94.155 -60.068 54.212
521805 94.195 -60.077 54.211
521806 94.235 -60.086 54.206
521807 94.275 -60.094 54.202
521808 94.316 -60.103 54.192
521809 94.356 -60.112 54.174
521810 94.398 -60.119 54.150
521811 94.439 -60.127 54.119
521812 94.481 -60.135 54.078
521813 94.522 -60.143 54.034
521814 94.563 -60.151 53.985
521815 94.605 -60.159 53.933
521816 94.648 -60.164 53.886
521817 94.691 -60.170 53.854
521818 94.735 -60.175 53.828
521819 94.776 -60.181 53.805
521820 94.816 -60.187 53.784
521821 94.856 -60.192 53.766
521822 94.895 -60.200 53.753
521823 94.934 -60.209 53.738
521824 94.974 -60.217 53.721
521825 95.010 -60.226 53.701
521826 95.045 -60.236 53.679
521827 95.079 -60.246 53.657
521828 95.114 -60.255 53.633
521829 95.149 -60.265 53.608
521830 95.183 -60.275 53.581
521831 95.218 -60.284 53.553
521832 95.256 -60.293 53.522
521833 95.295 -60.302 53.489
521834 95.330 -60.308 53.454
521835 95.365 -60.314 53.417
521836 95.400 -60.320 53.380
521837 95.437 -60.328 53.343
521838 95.473 -60.337 53.306
521839 95.510 -60.344 53.269
521840 95.547 -60.353 53.233
521841 95.584 -60.360 53.196
521842 95.620 -60.369 53.157
521843 95.657 -60.375 53.116
521844 95.692 -60.380 53.073
521845 95.728 -60.385 53.026
521846 95.764 -60.390 52.984
521847 95.800 -60.395 52.945
521848 95.836 -60.400 52.904
521849 95.875 -60.400 52.868
521850 95.913 -60.399 52.835
521851 95.953 -60.398 52.799
521852 95.990 -60.398 52.763
521853 96.028 -60.398 52.725
521854 96.065 -60.398 52.684
521855 96.102 -60.401 52.645
521856 96.138 -60.404 52.608
521857 96.175 -60.407 52.572
521858 96.212 -60.409 52.538
521859 96.248 -60.412 52.503
521860 96.285 -60.415 52.468
521861 96.324 -60.412 52.434
521862 96.364 -60.410 52.400
521863 96.405 -60.407 52.367
521864 96.446 -60.404 52.335
521865 96.487 -60.402 52.304
521866 96.528 -60.399 52.269
521867 96.570 -60.396 52.235
521868 96.614 -60.405 52.202
521869 96.659 -60.414 52.169
521870 96.703 -60.424 52.137
521871 96.748 -60.433 52.108
521872 96.793 -60.442 52.076
521873 96.837 -60.451 52.045
521874 96.877 -60.463 52.014
521875 96.908 -60.476 51.985
521876 96.941 -60.490 51.957
521877 96.972 -60.503 51.932
521878 97.005 -60.517 51.907
521879 97.037 -60.531 51.879
521880 97.069 -60.544 51.848
521881 97.116 -60.552 51.815
521882 97.163 -60.559 51.779
521883 97.209 -60.566 51.743
521884 97.233 -60.570 51.706
521885 97.252 -60.572 51.669
521886 97.290 -60.573 51.633
521887 97.326 -60.574 51.593
521888 97.363 -60.575 51.553
521889 97.400 -60.575 51.513
521890 97.437 -60.576 51.474
521891 97.474 -60.577 51.439
521892 97.511 -60.578 51.407
521893 97.547 -60.581 51.374
521894 97.577 -60.575 51.341
521895 97.610 -60.574 51.307
521896 97.656 -60.576 51.273
521897 97.702 -60.579 51.239
521898 97.748 -60.582 51.203
521899 97.794 -60.584 51.165
521900 97.833 -60.587 51.127
521901 97.858 -60.589 51.087
521902 97.883 -60.592 51.045
521903 97.909 -60.595 51.002
521904 97.934 -60.598 50.958
521905 97.959 -60.601 50.914
521906 97.984 -60.604 50.869
521907 97.993 -60.612 50.824
521908 97.996 -60.619 50.781
521909 98.020 -60.635 50.739
521910 98.048 -60.641 50.697
521911 98.084 -60.645 50.655
521912 98.121 -60.649 50.609
521913 98.157 -60.654 50.560
521914 98.193 -60.658 50.513
521915 98.229 -60.662 50.467
521916 98.266 -60.666 50.422
521917 98.302 -60.671 50.378
521918 98.339 -60.676 50.333
521919 98.375 -60.679 50.288
521920 98.411 -60.684 50.243
521921 98.447 -60.688 50.201
521922 98.484 -60.692 50.159
521923 98.520 -60.697 50.118
521924 98.556 -60.701 50.077
522329 87.274 -57.964 57.181
522330 87.315 -57.977 57.163
522331 87.355 -57.992 57.145
522332 87.396 -58.006 57.127
522333 87.437 -58.019 57.109
522334 87.477 -58.033 57.089
522335 87.518 -58.046 57.069
522336 87.559 -58.061 57.049
522337 87.599 -58.075 57.029
522338 87.640 -58.088 57.006
522339 87.681 -58.102 56.983
522340 87.721 -58.115 56.958
522341 87.762 -58.129 56.933
522342 87.804 -58.134 56.904
522343 87.847 -58.136 56.873
522344 87.890 -58.138 56.841
522345 87.932 -58.140 56.808
522346 87.975 -58.143 56.774
522347 88.017 -58.145 56.738
522348 88.049 -58.152 56.702
522349 88.080 -58.158 56.665
522350 88.112 -58.165 56.627
522351 88.156 -58.176 56.588
522352 88.202 -58.187 56.548
522353 88.247 -58.199 56.506
522354 88.293 -58.210 56.464
522355 88.338 -58.222 56.422
522356 88.384 -58.234 56.381
522357 88.429 -58.245 56.339
522358 88.475 -58.256 56.299
522359 88.520 -58.268 56.260
522360 88.561 -58.279 56.222
522361 88.602 -58.289 56.185
522362 88.642 -58.300 56.150
522363 88.682 -58.310 56.117
522364 88.722 -58.321 56.084
522365 88.763 -58.332 56.055
522366 88.803 -58.341 56.027
522367 88.843 -58.352 56.000
522368 88.886 -58.372 55.976
522369 88.929 -58.392 55.951
522370 88.972 -58.413 55.929
522371 89.015 -58.435 55.908
522372 89.058 -58.455 55.889
522373 89.101 -58.476 55.871
522374 89.144 -58.497 55.855
522375 89.183 -58.512 55.840
522376 89.218 -58.523 55.828
522377 89.253 -58.533 55.817
522378 89.287 -58.544 55.808
522379 89.323 -58.554 55.800
522380 89.357 -58.565 55.794
522381 89.392 -58.576 55.789
522382 89.431 -58.584 55.786
522383 89.478 -58.590 55.782
522384 89.525 -58.596 55.778
522385 89.569 -58.601 55.775
522386 89.606 -58.605 55.772
522387 89.642 -58.609 55.772
522388 89.679 -58.615 55.776
522389 89.718 -58.632 55.784
522390 89.757 -58.649 55.794
522391 89.796 -58.667 55.796
522392 89.835 -58.684 55.795
522393 89.874 -58.701 55.791
522394 89.912 -58.719 55.783
522395 89.951 -58.736 55.775
522396 89.990 -58.753 55.765
522397 90.029 -58.770 55.754
522398 90.068 -58.788 55.742
522399 90.106 -58.805 55.728
522400 90.147 -58.823 55.711
522401 90.187 -58.842 55.692
522402 90.227 -58.851 55.670
522403 90.265 -58.854 55.649
522404 90.303 -58.857 55.628
522405 90.341 -58.861 55.607
522406 90.380 -58.866 55.586
522407 90.418 -58.871 55.563
522408 90.457 -58.876 55.539
522409 90.495 -58.881 55.511
522410 90.533 -58.886 55.481
522411 90.572 -58.891 55.449
522412 90.610 -58.896 55.416
522413 90.649 -58.901 55.383
522414 90.687 -58.905 55.351
522415 90.722 -58.926 55.323
522416 90.756 -58.946 55.300
522417 90.797 -58.969 55.276
522418 90.841 -58.994 55.249
522419 90.886 -59.019 55.219
522420 90.930 -59.043 55.188
522421 90.974 -59.068 55.155
522422 91.019 -59.093 55.121
522423 91.064 -59.118 55.082
522424 91.107 -59.142 55.041
522425 91.147 -59.157 55.000
522426 91.187 -59.173 54.962
522427 91.226 -59.188 54.930
522428 91.266 -59.203 54.903
522429 91.305 -59.218 54.880
522430 91.345 -59.234 54.860
522431 91.385 -59.250 54.843
522432 91.424 -59.265 54.828
522433 91.464 -59.280 54.814
522434 91.504 -59.295 54.800
522435 91.544 -59.310 54.788
522436 91.589 -59.322 54.779
522437 91.634 -59.335 54.772
522438 91.679 -59.348 54.768
522439 91.720 -59.362 54.763
522440 91.756 -59.376 54.757
522441 91.792 -59.391 54.751
522442 91.827 -59.394 54.746
522443 91.863 -59.394 54.745
522444 91.894 -59.397 54.744
522445 91.922 -59.402 54.745
522446 91.950 -59.407 54.749
522447 91.977 -59.413 54.755
522448 92.005 -59.419 54.757
522449 92.032 -59.424 54.750
522450 92.060 -59.429 54.734
522451 92.088 -59.435 54.712
522452 92.115 -59.440 54.689
522453 92.143 -59.446 54.669
522454 92.170 -59.451 54.645
522455 92.198 -59.457 54.615
522456 92.226 -59.462 54.585
522457 92.253 -59.467 54.566
522458 92.294 -59.469 54.565
522459 92.342 -59.468 54.566
522460 92.389 -59.467 54.565
522461 92.436 -59.466 54.557
522462 92.484 -59.466 54.550
522463 92.531 -59.465 54.549
522464 92.578 -59.464 54.535
522465 92.626 -59.463 54.518
522466 92.668 -59.472 54.503
522467 92.711 -59.482 54.492
522468 92.754 -59.491 54.480
522469 92.796 -59.501 54.467
522470 92.838 -59.511 54.456
522471 92.881 -59.520 54.437
522472 92.923 -59.530 54.416
522473 92.966 -59.540 54.398
522474 93.008 -59.550 54.381
522475 93.051 -59.559 54.366
522476 93.093 -59.569 54.356
522477 93.132 -59.577 54.351
522478 93.170 -59.585 54.354
522479 93.209 -59.595 54.367
522480 93.247 -59.603 54.390
522481 93.286 -59.611 54.405
522482 93.324 -59.620 54.411
522483 93.360 -59.640 54.411
522484 93.397 -59.660 54.405
522485 93.433 -59.682 54.400
522486 93.469 -59.702 54.395
522487 93.505 -59.723 54.391
522488 93.541 -59.744 54.387
522489 93.577 -59.764 54.381
522490 93.614 -59.785 54.375
522491 93.650 -59.806 54.366
522492 93.686 -59.826 54.356
522493 93.722 -59.846 54.348
522494 93.759 -59.862 54.338
522495 93.795 -59.877 54.324
522496 93.832 -59.893 54.305
522497 93.868 -59.908 54.281
522498 93.905 -59.924 54.255
522499 93.941 -59.939 54.233
522500 93.978 -59.955 54.219
522501 94.014 -59.971 54.211
522502 94.051 -59.986 54.201
522503 94.087 -60.002 54.194
522504 94.126 -60.014 54.189
522505 94.166 -60.023 54.188
522506 94.206 -60.031 54.186
522507 94.247 -60.040 54.183
522508 94.286 -60.049 54.179
522509 94.327 -60.058 54.170
522510 94.367 -60.067 54.157
522511 94.409 -60.074 54.139
522512 94.451 -60.082 54.114
522513 94.492 -60.090 54.077
522514 94.533 -60.098 54.036
522515 94.574 -60.105 53.991
522516 94.616 -60.113 53.942
522517 94.657 -60.121 53.888
522518 94.699 -60.129 53.845
522519 94.741 -60.136 53.810
522520 94.783 -60.141 53.779
522521 94.823 -60.148 53.753
522522 94.862 -60.156 53.733
522523 94.901 -60.165 53.722
522524 94.941 -60.173 53.710
522525 94.980 -60.181 53.695
522526 95.017 -60.190 53.677
522527 95.051 -60.199 53.658
522528 95.086 -60.209 53.636
522529 95.121 -60.219 53.614
522530 95.156 -60.229 53.590
522531 95.190 -60.239 53.564
522532 95.229 -60.247 53.536
522533 95.267 -60.256 53.507
522534 95.305 -60.264 53.475
522535 95.340 -60.270 53.442
522536 95.375 -60.276 53.409
522537 95.410 -60.282 53.375
522538 95.445 -60.289 53.340
522539 95.482 -60.297 53.305
522540 95.518 -60.305 53.270
522541 95.555 -60.313 53.236
522542 95.592 -60.321 53.200
522543 95.628 -60.328 53.160
522544 95.664 -60.333 53.119
522545 95.700 -60.338 53.074
522546 95.736 -60.343 53.029
522547 95.771 -60.348 52.989
522548 95.808 -60.353 52.955
522549 95.843 -60.358 52.916
522550 95.882 -60.358 52.880
522551 95.921 -60.358 52.846
522552 95.959 -60.357 52.812
522553 95.997 -60.357 52.777
522554 96.035 -60.357 52.742
522555 96.073 -60.357 52.703
522556 96.110 -60.358 52.664
522557 96.146 -60.360 52.627
522558 96.183 -60.363 52.592
522559 96.219 -60.366 52.558
522560 96.256 -60.369 52.524
522561 96.294 -60.370 52.489
522562 96.333 -60.368 52.455
522563 96.372 -60.365 52.420
522564 96.412 -60.362 52.385
522565 96.454 -60.359 52.352
522566 96.495 -60.357 52.318
522567 96.536 -60.354 52.283
522568 96.579 -60.360 52.248
522569 96.624 -60.369 52.213
522570 96.669 -60.378 52.179
522571 96.713 -60.387 52.145
522572 96.758 -60.397 52.114
522573 96.803 -60.406 52.082
522574 96.847 -60.415 52.049
522575 96.886 -60.426 52.017
522576 96.918 -60.440 51.986
522577 96.950 -60.453 51.956
522578 96.982 -60.467 51.928
522579 97.014 -60.481 51.900
522580 97.046 -60.495 51.872
522581 97.078 -60.509 51.842
522582 97.116 -60.520 51.809
522583 97.162 -60.527 51.774
522584 97.209 -60.535 51.739
522585 97.234 -60.538 51.703
522586 97.259 -60.541 51.667
522587 97.296 -60.541 51.630
522588 97.333 -60.541 51.592
522589 97.370 -60.542 51.551
522590 97.406 -60.543 51.511
522591 97.443 -60.544 51.474
522592 97.480 -60.544 51.440
522593 97.517 -60.545 51.407
522594 97.554 -60.547 51.375
522595 97.588 -60.545 51.342
522596 97.618 -60.542 51.308
522597 97.664 -60.545 51.275
522598 97.710 -60.547 51.241
522599 97.756 -60.550 51.207
522600 97.798 -60.553 51.170
522601 97.823 -60.556 51.133
522602 97.848 -60.559 51.093
522603 97.873 -60.562 51.050
522604 97.899 -60.565 51.008
522605 97.924 -60.568 50.965
522606 97.949 -60.571 50.920
522607 97.974 -60.574 50.876
522608 97.997 -60.577 50.831
522609 98.001 -60.585 50.788
522610 98.022 -60.600 50.747
522611 98.054 -60.604 50.706
522612 98.090 -60.607 50.664
522613 98.127 -60.612 50.619
522614 98.163 -60.616 50.570
522615 98.200 -60.620 50.521
522616 98.236 -60.625 50.474
522617 98.272 -60.629 50.428
522618 98.308 -60.633 50.383
522619 98.345 -60.638 50.337
522620 98.381 -60.642 50.293
522621 98.417 -60.647 50.247
522622 98.453 -60.651 50.204
522623 98.490 -60.655 50.162
522624 98.526 -60.660 50.119
522625 98.563 -60.663 50.078
523030 87.300 -57.938 57.180
523031 87.340 -57.952 57.162
523032 87.381 -57.965 57.145
523033 87.422 -57.980 57.126
523034 87.462 -57.993 57.108
523035 87.503 -58.007 57.090
523036 87.544 -58.021 57.070
523037 87.584 -58.034 57.051
523038 87.625 -58.049 57.031
523039 87.666 -58.062 57.008
523040 87.706 -58.076 56.985
523041 87.747 -58.090 56.961
523042 87.787 -58.102 56.937
523043 87.830 -58.104 56.909
523044 87.873 -58.106 56.879
523045 87.916 -58.109 56.847
523046 87.959 -58.111 56.815
523047 88.001 -58.113 56.782
523048 88.044 -58.115 56.747
523049 88.084 -58.118 56.713
523050 88.116 -58.125 56.677
523051 88.147 -58.131 56.640
523052 88.179 -58.137 56.602
523053 88.216 -58.147 56.563
523054 88.262 -58.158 56.523
523055 88.308 -58.169 56.481
523056 88.353 -58.181 56.440
523057 88.398 -58.193 56.398
523058 88.444 -58.204 56.356
523059 88.486 -58.215 56.316
523060 88.527 -58.225 56.276
523061 88.567 -58.236 56.237
523062 88.607 -58.247 56.199
523063 88.647 -58.256 56.163
523064 88.688 -58.267 56.129
523065 88.728 -58.278 56.096
523066 88.768 -58.288 56.066
523067 88.809 -58.299 56.038
523068 88.850 -58.313 56.010
523069 88.893 -58.334 55.983
523070 88.935 -58.355 55.958
523071 88.979 -58.375 55.936
523072 89.022 -58.397 55.914
523073 89.065 -58.418 55.894
523074 89.108 -58.438 55.876
523075 89.151 -58.460 55.860
523076 89.194 -58.480 55.845
523077 89.230 -58.492 55.833
523078 89.265 -58.503 55.823
523079 89.300 -58.513 55.814
523080 89.335 -58.524 55.806
523081 89.370 -58.534 55.803
523082 89.404 -58.545 55.801
523083 89.445 -58.554 55.799
523084 89.491 -58.559 55.798
523085 89.538 -58.565 55.795
523086 89.585 -58.571 55.792
523087 89.627 -58.576 55.790
523088 89.664 -58.580 55.789
523089 89.700 -58.584 55.791
523090 89.739 -58.598 55.797
523091 89.777 -58.616 55.801
523092 89.816 -58.633 55.799
523093 89.855 -58.651 55.795
523094 89.894 -58.667 55.788
523095 89.933 -58.685 55.777
523096 89.972 -58.702 55.766
523097 90.011 -58.720 55.755
523098 90.049 -58.736 55.743
523099 90.088 -58.754 55.731
523100 90.127 -58.771 55.715
523101 90.166 -58.789 55.697
523102 90.204 -58.806 55.676
523103 90.242 -58.820 55.653
523104 90.281 -58.825 55.630
523105 90.319 -58.830 55.609
523106 90.357 -58.835 55.588
523107 90.396 -58.839 55.568
523108 90.434 -58.845 55.547
523109 90.472 -58.849 55.524
523110 90.511 -58.855 55.499
523111 90.549 -58.859 55.473
523112 90.588 -58.864 55.444
523113 90.626 -58.869 55.415
523114 90.664 -58.874 55.383
523115 90.702 -58.879 55.350
523116 90.740 -58.890 55.316
523117 90.774 -58.911 55.294
523118 90.809 -58.931 55.269
523119 90.845 -58.952 55.240
523120 90.890 -58.977 55.210
523121 90.934 -59.002 55.178
523122 90.978 -59.027 55.148
523123 91.023 -59.052 55.117
523124 91.067 -59.076 55.084
523125 91.111 -59.101 55.048
523126 91.152 -59.118 55.011
523127 91.192 -59.134 54.974
523128 91.231 -59.149 54.937
523129 91.271 -59.165 54.908
523130 91.310 -59.180 54.882
523131 91.350 -59.195 54.860
523132 91.390 -59.210 54.842
523133 91.429 -59.225 54.826
523134 91.469 -59.241 54.811
523135 91.512 -59.255 54.798
523136 91.557 -59.268 54.787
523137 91.602 -59.281 54.778
523138 91.647 -59.294 54.771
523139 91.692 -59.306 54.769
523140 91.728 -59.320 54.765
523141 91.764 -59.334 54.761
523142 91.800 -59.349 54.757
523143 91.836 -59.355 54.755
523144 91.871 -59.355 54.758
523145 91.907 -59.356 54.760
523146 91.937 -59.359 54.758
523147 91.964 -59.365 54.752
523148 91.992 -59.370 54.746
523149 92.019 -59.375 54.749
523150 92.047 -59.381 54.750
523151 92.075 -59.387 54.743
523152 92.102 -59.392 54.727
523153 92.130 -59.397 54.704
523154 92.157 -59.403 54.682
523155 92.185 -59.409 54.660
523156 92.213 -59.414 54.634
523157 92.240 -59.419 54.601
523158 92.268 -59.425 54.568
523159 92.309 -59.426 54.551
523160 92.356 -59.425 54.555
523161 92.404 -59.425 54.558
523162 92.451 -59.424 54.559
523163 92.498 -59.423 54.561
523164 92.546 -59.422 54.564
523165 92.593 -59.422 54.553
523166 92.640 -59.421 54.540
523167 92.686 -59.425 54.529
523168 92.728 -59.435 54.519
523169 92.770 -59.444 54.508
523170 92.813 -59.453 54.497
523171 92.855 -59.463 54.486
523172 92.898 -59.473 54.467
523173 92.940 -59.482 54.446
523174 92.983 -59.492 54.423
523175 93.025 -59.502 54.399
523176 93.067 -59.512 54.378
523177 93.109 -59.521 54.364
523178 93.147 -59.529 54.356
523179 93.185 -59.538 54.354
523180 93.224 -59.546 54.355
523181 93.262 -59.554 54.366
523182 93.300 -59.563 54.379
523183 93.339 -59.572 54.391
523184 93.375 -59.591 54.398
523185 93.412 -59.613 54.397
523186 93.448 -59.633 54.392
523187 93.484 -59.654 54.391
523188 93.520 -59.675 54.389
523189 93.556 -59.695 54.386
523190 93.592 -59.716 54.381
523191 93.629 -59.737 54.373
523192 93.665 -59.757 54.362
523193 93.701 -59.778 54.350
523194 93.737 -59.794 54.341
523195 93.774 -59.809 54.333
523196 93.811 -59.825 54.319
523197 93.847 -59.840 54.299
523198 93.884 -59.856 54.272
523199 93.920 -59.871 54.241
523200 93.957 -59.887 54.213
523201 93.993 -59.902 54.193
523202 94.030 -59.918 54.182
523203 94.066 -59.934 54.172
523204 94.103 -59.949 54.164
523205 94.139 -59.965 54.161
523206 94.177 -59.977 54.161
523207 94.217 -59.986 54.158
523208 94.258 -59.996 54.155
523209 94.297 -60.005 54.149
523210 94.338 -60.013 54.143
523211 94.379 -60.021 54.133
523212 94.420 -60.029 54.118
523213 94.462 -60.037 54.096
523214 94.503 -60.045 54.065
523215 94.544 -60.052 54.029
523216 94.586 -60.061 53.987
523217 94.627 -60.068 53.941
523218 94.669 -60.076 53.893
523219 94.710 -60.083 53.846
523220 94.751 -60.093 53.809
523221 94.790 -60.103 53.778
523222 94.829 -60.112 53.749
523223 94.868 -60.121 53.723
523224 94.908 -60.129 53.699
523225 94.947 -60.137 53.682
523226 94.986 -60.146 53.667
523227 95.024 -60.154 53.653
523228 95.058 -60.164 53.636
523229 95.093 -60.174 53.614
523230 95.127 -60.184 53.592
523231 95.162 -60.193 53.567
523232 95.200 -60.202 53.542
523233 95.239 -60.211 53.517
523234 95.278 -60.219 53.489
523235 95.314 -60.227 53.461
523236 95.349 -60.233 53.432
523237 95.384 -60.239 53.401
523238 95.419 -60.245 53.369
523239 95.454 -60.251 53.337
523240 95.490 -60.258 53.304
523241 95.527 -60.265 53.271
523242 95.563 -60.274 53.237
523243 95.600 -60.281 53.202
523244 95.635 -60.286 53.162
523245 95.672 -60.291 53.120
523246 95.707 -60.297 53.074
523247 95.743 -60.301 53.031
523248 95.779 -60.306 52.995
523249 95.815 -60.311 52.965
523250 95.851 -60.316 52.930
523251 95.889 -60.317 52.895
523252 95.928 -60.317 52.861
523253 95.966 -60.316 52.827
523254 96.004 -60.316 52.794
523255 96.041 -60.316 52.762
523256 96.079 -60.316 52.725
523257 96.116 -60.316 52.687
523258 96.154 -60.317 52.651
523259 96.190 -60.320 52.616
523260 96.227 -60.322 52.582
523261 96.263 -60.325 52.548
523262 96.302 -60.325 52.514
523263 96.341 -60.322 52.479
523264 96.380 -60.320 52.443
523265 96.420 -60.318 52.407
523266 96.461 -60.315 52.373
523267 96.502 -60.312 52.337
523268 96.544 -60.315 52.301
523269 96.589 -60.324 52.264
523270 96.634 -60.333 52.228
523271 96.678 -60.342 52.192
523272 96.723 -60.351 52.157
523273 96.768 -60.360 52.124
523274 96.812 -60.369 52.089
523275 96.857 -60.378 52.055
523276 96.896 -60.390 52.021
523277 96.928 -60.403 51.988
523278 96.960 -60.418 51.957
523279 96.992 -60.431 51.927
523280 97.024 -60.445 51.896
523281 97.056 -60.459 51.866
523282 97.088 -60.472 51.836
523283 97.120 -60.486 51.802
523284 97.162 -60.495 51.768
523285 97.209 -60.503 51.733
523286 97.234 -60.506 51.697
523287 97.265 -60.508 51.661
523288 97.302 -60.509 51.624
523289 97.339 -60.510 51.585
523290 97.376 -60.510 51.547
523291 97.413 -60.511 51.509
523292 97.450 -60.512 51.474
523293 97.487 -60.513 51.443
523294 97.524 -60.513 51.410
523295 97.561 -60.514 51.377
523296 97.597 -60.516 51.343
523297 97.627 -60.511 51.308
523298 97.673 -60.514 51.273
523299 97.719 -60.516 51.241
523300 97.763 -60.519 51.206
523301 97.788 -60.522 51.171
523302 97.813 -60.525 51.134
523303 97.839 -60.528 51.095
523304 97.864 -60.531 51.054
523305 97.889 -60.534 51.011
523306 97.914 -60.537 50.968
523307 97.939 -60.540 50.924
523308 97.964 -60.543 50.880
523309 97.990 -60.546 50.835
523310 98.005 -60.551 50.793
523311 98.024 -60.562 50.753
523312 98.061 -60.566 50.712
523313 98.097 -60.570 50.670
523314 98.133 -60.575 50.626
523315 98.169 -60.579 50.579
523316 98.206 -60.583 50.529
523317 98.242 -60.588 50.481
523318 98.278 -60.591 50.434
523319 98.314 -60.596 50.388
523320 98.351 -60.600 50.342
523321 98.387 -60.605 50.297
523322 98.424 -60.609 50.251
523323 98.460 -60.613 50.207
523324 98.496 -60.618 50.164
523325 98.532 -60.622 50.122
523326 98.569 -60.626 50.080
523327 98.605 -60.631 50.038
523731 87.322 -57.905 57.177
523732 87.362 -57.918 57.159
523733 87.403 -57.932 57.142
523734 87.443 -57.946 57.124
523735 87.483 -57.959 57.106
523736 87.524 -57.972 57.088
523737 87.564 -57.985 57.069
523738 87.605 -57.999 57.051
523739 87.645 -58.012 57.031
523740 87.685 -58.025 57.009
523741 87.726 -58.039 56.987
523742 87.766 -58.052 56.963
523743 87.810 -58.062 56.939
523744 87.854 -58.068 56.912
523745 87.899 -58.076 56.882
523746 87.942 -58.079 56.851
523747 87.985 -58.081 56.820
523748 88.028 -58.084 56.788
523749 88.070 -58.086 56.755
523750 88.113 -58.088 56.721
523751 88.152 -58.092 56.687
523752 88.184 -58.098 56.651
523753 88.215 -58.105 56.613
523754 88.247 -58.111 56.576
523755 88.278 -58.118 56.537
523756 88.322 -58.128 56.497
523757 88.367 -58.140 56.456
523758 88.411 -58.151 56.415
523759 88.451 -58.162 56.373
523760 88.491 -58.172 56.332
523761 88.532 -58.182 56.292
523762 88.572 -58.193 56.253
523763 88.612 -58.203 56.215
523764 88.652 -58.214 56.177
523765 88.693 -58.225 56.142
523766 88.733 -58.234 56.109
523767 88.773 -58.245 56.078
523768 88.814 -58.256 56.049
523769 88.856 -58.275 56.020
523770 88.899 -58.297 55.992
523771 88.942 -58.317 55.966
523772 88.985 -58.338 55.943
523773 89.028 -58.359 55.921
523774 89.071 -58.380 55.901
523775 89.114 -58.400 55.882
523776 89.158 -58.422 55.866
523777 89.200 -58.443 55.850
523778 89.242 -58.462 55.838
523779 89.277 -58.473 55.828
523780 89.312 -58.482 55.819
523781 89.347 -58.493 55.814
523782 89.382 -58.504 55.813
523783 89.416 -58.514 55.814
523784 89.458 -58.523 55.815
523785 89.505 -58.529 55.816
523786 89.551 -58.535 55.813
523787 89.599 -58.540 55.811
523788 89.646 -58.546 55.809
523789 89.685 -58.551 55.806
523790 89.722 -58.554 55.805
523791 89.759 -58.565 55.806
523792 89.798 -58.582 55.803
523793 89.837 -58.599 55.798
523794 89.876 -58.617 55.791
523795 89.914 -58.634 55.784
523796 89.953 -58.651 55.772
523797 89.992 -58.668 55.762
523798 90.031 -58.686 55.751
523799 90.070 -58.703 55.740
523800 90.108 -58.720 55.728
523801 90.148 -58.737 55.714
523802 90.183 -58.754 55.691
523803 90.216 -58.767 55.666
523804 90.254 -58.775 55.641
523805 90.294 -58.789 55.617
523806 90.334 -58.803 55.595
523807 90.373 -58.808 55.575
523808 90.411 -58.813 55.557
523809 90.450 -58.818 55.538
523810 90.488 -58.823 55.517
523811 90.526 -58.828 55.494
523812 90.565 -58.833 55.471
523813 90.603 -58.838 55.446
523814 90.642 -58.842 55.418
523815 90.680 -58.848 55.386
523816 90.718 -58.852 55.351
523817 90.756 -58.858 55.313
523818 90.792 -58.875 55.287
523819 90.826 -58.896 55.260
523820 90.861 -58.916 55.231
523821 90.896 -58.936 55.202
523822 90.938 -58.960 55.173
523823 90.982 -58.985 55.145
523824 91.026 -59.010 55.118
523825 91.071 -59.034 55.090
523826 91.115 -59.059 55.058
523827 91.157 -59.080 55.024
523828 91.196 -59.095 54.988
523829 91.236 -59.110 54.950
523830 91.276 -59.125 54.916
523831 91.315 -59.140 54.887
523832 91.355 -59.156 54.863
523833 91.395 -59.171 54.843
523834 91.435 -59.187 54.825
523835 91.480 -59.200 54.809
523836 91.524 -59.212 54.797
523837 91.569 -59.225 54.787
523838 91.614 -59.238 54.777
523839 91.659 -59.251 54.771
523840 91.701 -59.264 54.770
523841 91.736 -59.278 54.768
523842 91.773 -59.293 54.766
523843 91.809 -59.308 54.767
523844 91.845 -59.316 54.769
523845 91.880 -59.316 54.772
523846 91.916 -59.317 54.775
523847 91.951 -59.317 54.773
523848 91.979 -59.322 54.764
523849 92.007 -59.328 54.750
523850 92.034 -59.333 54.744
523851 92.062 -59.338 54.743
523852 92.089 -59.344 54.741
523853 92.117 -59.350 54.733
523854 92.145 -59.355 54.716
523855 92.172 -59.360 54.692
523856 92.200 -59.366 54.667
523857 92.227 -59.371 54.640
523858 92.255 -59.377 54.609
523859 92.283 -59.382 54.580
523860 92.323 -59.384 54.559
523861 92.371 -59.383 54.548
523862 92.418 -59.381 54.551
523863 92.465 -59.381 54.559
523864 92.513 -59.380 54.563
523865 92.560 -59.379 54.554
523866 92.607 -59.378 54.559
523867 92.655 -59.378 54.554
523868 92.702 -59.377 54.551
523869 92.745 -59.387 54.545
523870 92.787 -59.397 54.533
523871 92.830 -59.406 54.523
523872 92.872 -59.416 54.509
523873 92.915 -59.425 54.491
523874 92.957 -59.435 54.468
523875 92.999 -59.444 54.441
523876 93.042 -59.454 54.413
523877 93.084 -59.464 54.389
523878 93.123 -59.472 54.373
523879 93.162 -59.482 54.363
523880 93.200 -59.490 54.357
523881 93.239 -59.498 54.352
523882 93.277 -59.507 54.354
523883 93.315 -59.515 54.360
523884 93.354 -59.523 54.368
523885 93.391 -59.544 54.375
523886 93.427 -59.564 54.378
523887 93.463 -59.585 54.374
523888 93.499 -59.606 54.374
523889 93.535 -59.626 54.373
523890 93.571 -59.647 54.370
523891 93.608 -59.668 54.366
523892 93.644 -59.689 54.358
523893 93.680 -59.709 54.347
523894 93.716 -59.725 54.336
523895 93.753 -59.741 54.325
523896 93.789 -59.757 54.313
523897 93.826 -59.772 54.300
523898 93.862 -59.788 54.279
523899 93.899 -59.803 54.255
523900 93.935 -59.819 54.231
523901 93.972 -59.834 54.205
523902 94.008 -59.850 54.181
523903 94.045 -59.865 54.161
523904 94.081 -59.881 54.144
523905 94.118 -59.897 54.133
523906 94.154 -59.912 54.130
523907 94.191 -59.928 54.131
523908 94.228 -59.941 54.128
523909 94.269 -59.950 54.122
523910 94.309 -59.959 54.116
523911 94.349 -59.968 54.110
523912 94.390 -59.977 54.103
523913 94.431 -59.984 54.091
523914 94.473 -59.992 54.071
523915 94.514 -60.000 54.045
523916 94.555 -60.008 54.013
523917 94.597 -60.015 53.975
523918 94.638 -60.023 53.934
523919 94.678 -60.033 53.892
523920 94.717 -60.043 53.846
523921 94.757 -60.055 53.807
523922 94.796 -60.065 53.774
523923 94.835 -60.075 53.743
523924 94.875 -60.085 53.713
523925 94.914 -60.093 53.686
523926 94.953 -60.102 53.663
523927 94.992 -60.110 53.643
523928 95.030 -60.118 53.626
523929 95.065 -60.128 53.610
523930 95.100 -60.138 53.590
523931 95.134 -60.148 53.568
523932 95.173 -60.156 53.543
523933 95.211 -60.165 53.518
523934 95.250 -60.174 53.494
523935 95.288 -60.182 53.468
523936 95.324 -60.189 53.442
523937 95.359 -60.195 53.416
523938 95.394 -60.201 53.388
523939 95.429 -60.207 53.359
523940 95.464 -60.213 53.329
523941 95.499 -60.219 53.298
523942 95.535 -60.226 53.267
523943 95.571 -60.234 53.235
523944 95.607 -60.240 53.199
523945 95.643 -60.244 53.160
523946 95.679 -60.249 53.117
523947 95.715 -60.254 53.071
523948 95.751 -60.259 53.031
523949 95.787 -60.265 52.999
523950 95.823 -60.269 52.974
523951 95.859 -60.275 52.944
523952 95.896 -60.276 52.915
523953 95.935 -60.276 52.882
523954 95.972 -60.276 52.850
523955 96.010 -60.275 52.819
523956 96.047 -60.275 52.787
523957 96.085 -60.275 52.751
523958 96.123 -60.275 52.717
523959 96.161 -60.275 52.682
523960 96.198 -60.276 52.646
523961 96.234 -60.279 52.611
523962 96.271 -60.282 52.577
523963 96.310 -60.280 52.543
523964 96.349 -60.278 52.506
523965 96.389 -60.275 52.469
523966 96.428 -60.272 52.433
523967 96.468 -60.270 52.398
523968 96.510 -60.269 52.361
523969 96.554 -60.278 52.323
523970 96.599 -60.287 52.284
523971 96.644 -60.297 52.246
523972 96.688 -60.306 52.209
523973 96.733 -60.315 52.172
523974 96.778 -60.324 52.136
523975 96.822 -60.333 52.100
523976 96.867 -60.343 52.063
523977 96.905 -60.354 52.028
523978 96.937 -60.368 51.993
523979 96.969 -60.381 51.961
523980 97.001 -60.395 51.928
523981 97.033 -60.409 51.895
523982 97.066 -60.422 51.863
523983 97.097 -60.437 51.831
523984 97.130 -60.450 51.796
523985 97.162 -60.463 51.762
523986 97.209 -60.471 51.727
523987 97.235 -60.475 51.691
523988 97.272 -60.475 51.655
523989 97.309 -60.476 51.618
523990 97.346 -60.477 51.580
523991 97.383 -60.478 51.544
523992 97.420 -60.478 51.509
523993 97.456 -60.479 51.475
523994 97.493 -60.480 51.444
523995 97.530 -60.481 51.411
523996 97.567 -60.481 51.377
523997 97.604 -60.482 51.342
523998 97.636 -60.480 51.308
523999 97.681 -60.483 51.273
524000 97.727 -60.485 51.240
524001 97.753 -60.488 51.206
524002 97.778 -60.491 51.171
524003 97.803 -60.494 51.134
524004 97.828 -60.497 51.094
524005 97.854 -60.500 51.054
524006 97.879 -60.503 51.012
524007 97.904 -60.506 50.969
524008 97.929 -60.509 50.926
524009 97.954 -60.512 50.883
524010 97.980 -60.515 50.840
524011 98.005 -60.518 50.799
524012 98.030 -60.524 50.757
524013 98.067 -60.528 50.716
524014 98.103 -60.533 50.674
524015 98.139 -60.537 50.630
524016 98.175 -60.541 50.585
524017 98.212 -60.546 50.536
524018 98.248 -60.550 50.488
524019 98.284 -60.554 50.440
524020 98.321 -60.559 50.393
524021 98.357 -60.563 50.347
524022 98.393 -60.567 50.302
524023 98.430 -60.572 50.256
524024 98.466 -60.576 50.211
524025 98.502 -60.580 50.168
524026 98.538 -60.585 50.126
524027 98.575 -60.589 50.083
524028 98.611 -60.594 50.040
524432 87.339 -57.862 57.173
524433 87.379 -57.875 57.155
524434 87.419 -57.889 57.138
524435 87.460 -57.902 57.121
524436 87.500 -57.915 57.103
524437 87.541 -57.929 57.085
524438 87.581 -57.942 57.066
524439 87.621 -57.955 57.047
524440 87.662 -57.969 57.028
524441 87.702 -57.982 57.007
524442 87.743 -57.996 56.985
524443 87.785 -58.007 56.962
524444 87.830 -58.015 56.939
524445 87.874 -58.021 56.911
524446 87.919 -58.029 56.882
524447 87.964 -58.036 56.852
524448 88.008 -58.043 56.822
524449 88.053 -58.051 56.791
524450 88.097 -58.056 56.761
524451 88.139 -58.059 56.728
524452 88.182 -58.061 56.695
524453 88.219 -58.065 56.660
524454 88.251 -58.071 56.624
524455 88.282 -58.078 56.587
524456 88.314 -58.084 56.549
524457 88.344 -58.090 56.510
524458 88.376 -58.098 56.470
524459 88.416 -58.108 56.429
524460 88.456 -58.118 56.389
524461 88.496 -58.129 56.348
524462 88.537 -58.140 56.308
524463 88.577 -58.150 56.269
524464 88.617 -58.160 56.230
524465 88.658 -58.171 56.193
524466 88.698 -58.181 56.157
524467 88.738 -58.192 56.122
524468 88.778 -58.203 56.090
524469 88.820 -58.217 56.060
524470 88.863 -58.238 56.031
524471 88.906 -58.259 56.002
524472 88.949 -58.280 55.976
524473 88.992 -58.300 55.952
524474 89.035 -58.322 55.929
524475 89.078 -58.342 55.908
524476 89.121 -58.363 55.889
524477 89.164 -58.384 55.872
524478 89.207 -58.405 55.857
524479 89.250 -58.426 55.844
524480 89.289 -58.441 55.834
524481 89.324 -58.452 55.825
524482 89.359 -58.463 55.821
524483 89.394 -58.473 55.820
524484 89.428 -58.484 55.822
524485 89.471 -58.491 55.824
524486 89.518 -58.498 55.825
524487 89.565 -58.504 55.826
524488 89.612 -58.510 55.827
524489 89.659 -58.515 55.825
524490 89.705 -58.521 55.821
524491 89.743 -58.525 55.815
524492 89.779 -58.531 55.811
524493 89.818 -58.548 55.802
524494 89.858 -58.566 55.793
524495 89.896 -58.582 55.787
524496 89.935 -58.600 55.780
524497 89.974 -58.617 55.770
524498 90.013 -58.635 55.761
524499 90.051 -58.652 55.752
524500 90.090 -58.669 55.741
524501 90.127 -58.686 55.728
524502 90.160 -58.701 55.711
524503 90.193 -58.715 55.685
524504 90.229 -58.725 55.657
524505 90.267 -58.733 55.630
524506 90.305 -58.739 55.606
524507 90.346 -58.758 55.587
524508 90.387 -58.777 55.571
524509 90.427 -58.786 55.557
524510 90.465 -58.792 55.540
524511 90.504 -58.796 55.520
524512 90.542 -58.802 55.501
524513 90.580 -58.806 55.477
524514 90.619 -58.811 55.452
524515 90.657 -58.816 55.423
524516 90.695 -58.821 55.391
524517 90.734 -58.826 55.354
524518 90.773 -58.831 55.313
524519 90.810 -58.839 55.280
524520 90.845 -58.860 55.251
524521 90.879 -58.880 55.224
524522 90.913 -58.901 55.197
524523 90.948 -58.922 55.173
524524 90.986 -58.943 55.146
524525 91.030 -58.968 55.122
524526 91.074 -58.993 55.096
524527 91.118 -59.018 55.068
524528 91.162 -59.040 55.037
524529 91.202 -59.055 55.003
524530 91.241 -59.071 54.967
524531 91.281 -59.087 54.932
524532 91.320 -59.102 54.900
524533 91.360 -59.117 54.872
524534 91.402 -59.131 54.848
524535 91.447 -59.144 54.827
524536 91.492 -59.157 54.811
524537 91.536 -59.170 54.797
524538 91.581 -59.183 54.785
524539 91.626 -59.196 54.776
524540 91.671 -59.209 54.770
524541 91.709 -59.223 54.768
524542 91.745 -59.237 54.770
524543 91.781 -59.252 54.775
524544 91.817 -59.266 54.784
524545 91.853 -59.277 54.788
524546 91.888 -59.278 54.786
524547 91.924 -59.278 54.787
524548 91.959 -59.278 54.787
524549 91.994 -59.279 54.781
524550 92.021 -59.285 54.765
524551 92.049 -59.291 54.746
524552 92.077 -59.296 54.737
524553 92.104 -59.301 54.735
524554 92.132 -59.306 54.733
524555 92.159 -59.312 54.722
524556 92.187 -59.318 54.701
524557 92.215 -59.323 54.677
524558 92.242 -59.328 54.653
524559 92.270 -59.334 54.627
524560 92.297 -59.340 54.600
524561 92.338 -59.341 54.577
524562 92.385 -59.340 54.560
524563 92.433 -59.339 54.554
524564 92.480 -59.338 54.556
524565 92.527 -59.338 54.561
524566 92.575 -59.337 54.556
524567 92.622 -59.336 54.559
524568 92.669 -59.335 54.560
524569 92.717 -59.334 54.557
524570 92.762 -59.339 54.550
524571 92.804 -59.349 54.541
524572 92.847 -59.359 54.526
524573 92.889 -59.369 54.507
524574 92.931 -59.378 54.488
524575 92.974 -59.388 54.466
524576 93.016 -59.397 54.441
524577 93.059 -59.407 54.417
524578 93.100 -59.416 54.399
524579 93.138 -59.425 54.384
524580 93.177 -59.433 54.372
524581 93.215 -59.441 54.363
524582 93.254 -59.450 54.356
524583 93.292 -59.459 54.350
524584 93.331 -59.467 54.347
524585 93.369 -59.475 54.347
524586 93.406 -59.495 54.351
524587 93.442 -59.516 54.354
524588 93.478 -59.537 54.353
524589 93.514 -59.557 54.353
524590 93.550 -59.578 54.352
524591 93.586 -59.599 54.350
524592 93.623 -59.620 54.347
524593 93.659 -59.640 54.342
524594 93.695 -59.657 54.334
524595 93.732 -59.673 54.325
524596 93.768 -59.688 54.315
524597 93.805 -59.704 54.302
524598 93.841 -59.720 54.289
524599 93.878 -59.735 54.270
524600 93.914 -59.751 54.251
524601 93.951 -59.766 54.230
524602 93.987 -59.782 54.208
524603 94.024 -59.797 54.186
524604 94.060 -59.813 54.165
524605 94.097 -59.828 54.144
524606 94.133 -59.844 54.128
524607 94.170 -59.860 54.115
524608 94.206 -59.875 54.103
524609 94.243 -59.891 54.095
524610 94.280 -59.905 54.089
524611 94.320 -59.914 54.081
524612 94.360 -59.923 54.072
524613 94.401 -59.931 54.063
524614 94.442 -59.939 54.053
524615 94.484 -59.947 54.039
524616 94.525 -59.955 54.015
524617 94.567 -59.962 53.985
524618 94.606 -59.973 53.952
524619 94.645 -59.983 53.917
524620 94.684 -59.994 53.881
524621 94.723 -60.005 53.840
524622 94.763 -60.016 53.801
524623 94.802 -60.027 53.765
524624 94.841 -60.036 53.734
524625 94.881 -60.047 53.704
524626 94.920 -60.058 53.681
524627 94.960 -60.066 53.651
524628 94.999 -60.074 53.622
524629 95.037 -60.083 53.600
524630 95.071 -60.093 53.584
524631 95.106 -60.102 53.564
524632 95.144 -60.111 53.542
524633 95.183 -60.119 53.518
524634 95.222 -60.128 53.493
524635 95.260 -60.137 53.468
524636 95.299 -60.145 53.444
524637 95.334 -60.151 53.420
524638 95.369 -60.157 53.397
524639 95.404 -60.163 53.373
524640 95.439 -60.169 53.346
524641 95.474 -60.175 53.318
524642 95.509 -60.182 53.289
524643 95.544 -60.188 53.259
524644 95.579 -60.193 53.228
524645 95.615 -60.198 53.192
524646 95.651 -60.202 53.153
524647 95.687 -60.208 53.109
524648 95.723 -60.212 53.065
524649 95.758 -60.218 53.030
524650 95.794 -60.222 53.004
524651 95.830 -60.228 52.982
524652 95.866 -60.233 52.958
524653 95.903 -60.235 52.932
524654 95.941 -60.235 52.903
524655 95.978 -60.235 52.873
524656 96.016 -60.234 52.844
524657 96.054 -60.234 52.814
524658 96.092 -60.234 52.781
524659 96.129 -60.234 52.747
524660 96.167 -60.234 52.712
524661 96.205 -60.234 52.677
524662 96.242 -60.235 52.642
524663 96.279 -60.237 52.608
524664 96.318 -60.234 52.572
524665 96.358 -60.232 52.534
524666 96.397 -60.230 52.497
524667 96.437 -60.228 52.460
524668 96.476 -60.225 52.427
524669 96.520 -60.233 52.390
524670 96.565 -60.242 52.351
524671 96.609 -60.251 52.310
524672 96.654 -60.260 52.269
524673 96.698 -60.269 52.229
524674 96.743 -60.279 52.190
524675 96.788 -60.288 52.153
524676 96.832 -60.297 52.114
524677 96.877 -60.306 52.075
524678 96.915 -60.318 52.038
524679 96.947 -60.331 52.001
524680 96.979 -60.345 51.966
524681 97.011 -60.359 51.932
524682 97.043 -60.373 51.897
524683 97.075 -60.387 51.862
524684 97.107 -60.400 51.828
524685 97.139 -60.414 51.792
524686 97.171 -60.428 51.757
524687 97.209 -60.439 51.722
524688 97.242 -60.443 51.685
524689 97.279 -60.444 51.648
524690 97.315 -60.444 51.612
524691 97.352 -60.445 51.577
524692 97.389 -60.445 51.544
524693 97.426 -60.446 51.510
524694 97.463 -60.447 51.476
524695 97.500 -60.447 51.443
524696 97.536 -60.448 51.410
524697 97.574 -60.449 51.376
524698 97.610 -60.450 51.342
524699 97.644 -60.449 51.308
524700 97.690 -60.452 51.274
524701 97.718 -60.454 51.239
524702 97.743 -60.457 51.205
524703 97.768 -60.460 51.170
524704 97.793 -60.463 51.133
524705 97.818 -60.466 51.093
524706 97.844 -60.469 51.053
524707 97.869 -60.472 51.012
524708 97.894 -60.475 50.970
524709 97.921 -60.478 50.928
524710 97.948 -60.481 50.885
524711 97.974 -60.484 50.844
524712 98.001 -60.486 50.803
524713 98.036 -60.487 50.761
524714 98.073 -60.491 50.719
524715 98.109 -60.495 50.676
524716 98.145 -60.500 50.633
524717 98.182 -60.504 50.589
524718 98.218 -60.508 50.541
524719 98.254 -60.513 50.494
524720 98.291 -60.517 50.446
524721 98.327 -60.522 50.398
524722 98.363 -60.525 50.352
524723 98.400 -60.530 50.307
524724 98.436 -60.535 50.261
524725 98.472 -60.538 50.216
524726 98.508 -60.543 50.173
524727 98.545 -60.547 50.129
524728 98.581 -60.552 50.086
524729 98.618 -60.556 50.043
525133 87.355 -57.819 57.168
525134 87.396 -57.832 57.151
525135 87.436 -57.846 57.133
525136 87.477 -57.858 57.117
525137 87.517 -57.872 57.100
525138 87.557 -57.886 57.081
525139 87.598 -57.899 57.062
525140 87.638 -57.912 57.043
525141 87.679 -57.925 57.024
525142 87.719 -57.939 57.003
525143 87.759 -57.952 56.982
525144 87.805 -57.960 56.959
525145 87.850 -57.967 56.935
525146 87.895 -57.974 56.908
525147 87.939 -57.982 56.880
525148 87.984 -57.989 56.851
525149 88.028 -57.996 56.823
525150 88.073 -58.004 56.793
525151 88.117 -58.011 56.763
525152 88.162 -58.018 56.732
525153 88.206 -58.025 56.700
525154 88.251 -58.033 56.667
525155 88.286 -58.038 56.632
525156 88.314 -58.043 56.597
525157 88.341 -58.050 56.560
525158 88.369 -58.055 56.521
525159 88.397 -58.061 56.482
525160 88.424 -58.066 56.442
525161 88.462 -58.075 56.403
525162 88.502 -58.086 56.363
525163 88.542 -58.096 56.324
525164 88.582 -58.107 56.285
525165 88.623 -58.118 56.246
525166 88.663 -58.128 56.208
525167 88.703 -58.138 56.172
525168 88.744 -58.149 56.137
525169 88.783 -58.159 56.104
525170 88.826 -58.179 56.073
525171 88.870 -58.200 56.043
525172 88.912 -58.221 56.013
525173 88.956 -58.242 55.986
525174 88.998 -58.263 55.961
525175 89.042 -58.284 55.938
525176 89.085 -58.305 55.917
525177 89.127 -58.325 55.897
525178 89.171 -58.347 55.879
525179 89.213 -58.367 55.864
525180 89.257 -58.388 55.850
525181 89.300 -58.410 55.839
525182 89.336 -58.422 55.831
525183 89.371 -58.432 55.826
525184 89.406 -58.443 55.825
525185 89.441 -58.453 55.826
525186 89.484 -58.460 55.828
525187 89.531 -58.466 55.831
525188 89.578 -58.473 55.835
525189 89.625 -58.479 55.836
525190 89.672 -58.485 55.833
525191 89.719 -58.491 55.827
525192 89.764 -58.496 55.817
525193 89.801 -58.500 55.808
525194 89.839 -58.514 55.797
525195 89.878 -58.532 55.790
525196 89.916 -58.549 55.786
525197 89.955 -58.567 55.779
525198 89.994 -58.583 55.770
525199 90.033 -58.601 55.764
525200 90.071 -58.618 55.757
525201 90.104 -58.632 55.746
525202 90.137 -58.648 55.730
525203 90.170 -58.663 55.706
525204 90.204 -58.676 55.677
525205 90.242 -58.683 55.647
525206 90.280 -58.690 55.620
525207 90.318 -58.697 55.599
525208 90.357 -58.708 55.581
525209 90.398 -58.726 55.571
525210 90.439 -58.745 55.559
525211 90.481 -58.764 55.544
525212 90.519 -58.770 55.526
525213 90.557 -58.775 55.507
525214 90.596 -58.780 55.483
525215 90.635 -58.785 55.456
525216 90.673 -58.789 55.426
525217 90.711 -58.795 55.392
525218 90.749 -58.799 55.353
525219 90.788 -58.805 55.313
525220 90.826 -58.809 55.278
525221 90.862 -58.824 55.249
525222 90.897 -58.845 55.223
525223 90.932 -58.866 55.199
525224 90.966 -58.886 55.176
525225 91.001 -58.907 55.152
525226 91.035 -58.927 55.128
525227 91.078 -58.951 55.105
525228 91.122 -58.976 55.080
525229 91.167 -59.001 55.051
525230 91.207 -59.017 55.018
525231 91.246 -59.032 54.985
525232 91.286 -59.048 54.952
525233 91.325 -59.063 54.918
525234 91.369 -59.076 54.885
525235 91.414 -59.089 54.856
525236 91.459 -59.102 54.832
525237 91.504 -59.115 54.813
525238 91.548 -59.128 54.798
525239 91.593 -59.140 54.785
525240 91.638 -59.153 54.776
525241 91.681 -59.167 54.770
525242 91.717 -59.181 54.767
525243 91.753 -59.196 54.771
525244 91.789 -59.210 54.779
525245 91.825 -59.225 54.789
525246 91.862 -59.238 54.795
525247 91.897 -59.239 54.796
525248 91.932 -59.239 54.795
525249 91.968 -59.240 54.793
525250 92.003 -59.240 54.790
525251 92.036 -59.242 54.781
525252 92.064 -59.248 54.761
525253 92.091 -59.253 54.743
525254 92.119 -59.259 54.730
525255 92.147 -59.264 54.723
525256 92.174 -59.269 54.717
525257 92.202 -59.275 54.707
525258 92.229 -59.281 54.692
525259 92.257 -59.286 54.674
525260 92.285 -59.291 54.653
525261 92.312 -59.297 54.628
525262 92.353 -59.298 54.602
525263 92.400 -59.297 54.582
525264 92.447 -59.297 54.565
525265 92.495 -59.296 54.559
525266 92.542 -59.295 54.566
525267 92.589 -59.294 54.567
525268 92.637 -59.294 54.565
525269 92.684 -59.293 54.565
525270 92.731 -59.292 54.561
525271 92.779 -59.292 54.553
525272 92.821 -59.302 54.547
525273 92.863 -59.311 54.535
525274 92.906 -59.321 54.517
525275 92.948 -59.331 54.497
525276 92.991 -59.340 54.474
525277 93.034 -59.350 54.449
525278 93.076 -59.359 54.424
525279 93.115 -59.368 54.403
525280 93.153 -59.377 54.386
525281 93.192 -59.385 54.373
525282 93.230 -59.394 54.362
525283 93.269 -59.402 54.355
525284 93.307 -59.410 54.348
525285 93.346 -59.419 54.343
525286 93.384 -59.428 54.339
525287 93.421 -59.447 54.337
525288 93.457 -59.468 54.335
525289 93.493 -59.488 54.331
525290 93.529 -59.509 54.329
525291 93.565 -59.530 54.327
525292 93.602 -59.551 54.326
525293 93.638 -59.571 54.327
525294 93.674 -59.588 54.325
525295 93.711 -59.604 54.321
525296 93.747 -59.620 54.315
525297 93.784 -59.635 54.308
525298 93.820 -59.651 54.297
525299 93.857 -59.667 54.283
525300 93.893 -59.682 54.266
525301 93.929 -59.698 54.250
525302 93.966 -59.714 54.231
525303 94.002 -59.729 54.211
525304 94.039 -59.745 54.190
525305 94.075 -59.760 54.169
525306 94.112 -59.776 54.147
525307 94.148 -59.791 54.125
525308 94.185 -59.807 54.106
525309 94.221 -59.822 54.092
525310 94.258 -59.838 54.083
525311 94.294 -59.854 54.078
525312 94.331 -59.869 54.071
525313 94.371 -59.878 54.062
525314 94.412 -59.886 54.054
525315 94.454 -59.894 54.043
525316 94.495 -59.902 54.027
525317 94.534 -59.913 54.007
525318 94.573 -59.924 53.984
525319 94.612 -59.934 53.958
525320 94.651 -59.945 53.929
525321 94.690 -59.956 53.896
525322 94.730 -59.967 53.857
525323 94.769 -59.977 53.817
525324 94.808 -59.988 53.778
525325 94.847 -59.998 53.742
525326 94.887 -60.008 53.706
525327 94.926 -60.019 53.673
525328 94.966 -60.029 53.640
525329 95.005 -60.038 53.609
525330 95.044 -60.047 53.582
525331 95.078 -60.057 53.558
525332 95.117 -60.065 53.536
525333 95.156 -60.074 53.513
525334 95.194 -60.083 53.488
525335 95.233 -60.091 53.463
525336 95.271 -60.099 53.440
525337 95.309 -60.108 53.417
525338 95.344 -60.114 53.397
525339 95.379 -60.120 53.377
525340 95.414 -60.126 53.355
525341 95.449 -60.132 53.330
525342 95.484 -60.138 53.304
525343 95.518 -60.144 53.277
525344 95.553 -60.151 53.247
525345 95.588 -60.156 53.216
525346 95.624 -60.161 53.182
525347 95.660 -60.165 53.142
525348 95.695 -60.170 53.100
525349 95.731 -60.174 53.059
525350 95.767 -60.179 53.029
525351 95.803 -60.184 53.008
525352 95.838 -60.188 52.991
525353 95.874 -60.193 52.972
525354 95.910 -60.199 52.948
525355 95.948 -60.199 52.922
525356 95.985 -60.198 52.895
525357 96.023 -60.197 52.868
525358 96.060 -60.196 52.840
525359 96.098 -60.196 52.808
525360 96.136 -60.195 52.775
525361 96.174 -60.194 52.741
525362 96.211 -60.193 52.707
525363 96.249 -60.193 52.672
525364 96.288 -60.193 52.637
525365 96.327 -60.194 52.599
525366 96.366 -60.195 52.561
525367 96.405 -60.196 52.524
525368 96.444 -60.196 52.487
525369 96.481 -60.197 52.454
525370 96.527 -60.203 52.417
525371 96.573 -60.210 52.377
525372 96.618 -60.216 52.336
525373 96.664 -60.224 52.293
525374 96.708 -60.234 52.252
525375 96.753 -60.243 52.211
525376 96.797 -60.252 52.171
525377 96.842 -60.261 52.131
525378 96.887 -60.270 52.091
525379 96.925 -60.281 52.052
525380 96.956 -60.296 52.014
525381 96.989 -60.309 51.978
525382 97.020 -60.323 51.941
525383 97.052 -60.337 51.905
525384 97.085 -60.350 51.868
525385 97.116 -60.364 51.831
525386 97.149 -60.378 51.795
525387 97.180 -60.392 51.758
525388 97.213 -60.406 51.722
525389 97.248 -60.410 51.685
525390 97.285 -60.411 51.648
525391 97.322 -60.412 51.612
525392 97.359 -60.412 51.577
525393 97.396 -60.413 51.543
525394 97.433 -60.414 51.510
525395 97.469 -60.415 51.478
525396 97.506 -60.416 51.445
525397 97.543 -60.416 51.411
525398 97.580 -60.417 51.378
525399 97.617 -60.418 51.345
525400 97.654 -60.419 51.311
525401 97.682 -60.421 51.276
525402 97.709 -60.424 51.241
525403 97.735 -60.426 51.204
525404 97.762 -60.429 51.168
525405 97.789 -60.431 51.130
525406 97.815 -60.434 51.091
525407 97.842 -60.437 51.052
525408 97.868 -60.440 51.012
525409 97.895 -60.443 50.971
525410 97.921 -60.446 50.930
525411 97.948 -60.448 50.888
525412 97.974 -60.451 50.847
525413 98.001 -60.453 50.806
525414 98.037 -60.460 50.764
525415 98.079 -60.453 50.722
525416 98.115 -60.458 50.679
525417 98.152 -60.463 50.635
525418 98.188 -60.466 50.591
525419 98.224 -60.471 50.545
525420 98.261 -60.475 50.498
525421 98.297 -60.480 50.451
525422 98.333 -60.484 50.405
525423 98.369 -60.488 50.357
525424 98.406 -60.493 50.312
525425 98.442 -60.497 50.266
525426 98.479 -60.501 50.221
525427 98.515 -60.506 50.178
525428 98.551 -60.510 50.134
525429 98.587 -60.514 50.091
525430 98.624 -60.519 50.047
525431 98.660 -60.523 50.004
525834 87.372 -57.775 57.160
525835 87.413 -57.789 57.143
525836 87.453 -57.802 57.126
525837 87.493 -57.815 57.110
525838 87.534 -57.829 57.093
525839 87.574 -57.842 57.075
525840 87.615 -57.855 57.058
525841 87.655 -57.869 57.039
525842 87.695 -57.882 57.019
525843 87.736 -57.896 56.998
525844 87.781 -57.905 56.977
525845 87.826 -57.913 56.953
525846 87.870 -57.920 56.929
525847 87.915 -57.927 56.904
525848 87.959 -57.935 56.877
525849 88.004 -57.942 56.849
525850 88.048 -57.949 56.821
525851 88.093 -57.956 56.792
525852 88.137 -57.964 56.762
525853 88.182 -57.971 56.731
525854 88.226 -57.978 56.700
525855 88.269 -57.987 56.668
525856 88.303 -57.998 56.635
525857 88.336 -58.009 56.601
525858 88.366 -58.014 56.566
525859 88.395 -58.020 56.530
525860 88.422 -58.025 56.492
525861 88.450 -58.031 56.454
525862 88.477 -58.037 56.415
525863 88.507 -58.043 56.376
525864 88.547 -58.053 56.338
525865 88.588 -58.064 56.300
525866 88.628 -58.075 56.261
525867 88.668 -58.085 56.223
525868 88.708 -58.096 56.186
525869 88.749 -58.106 56.152
525870 88.790 -58.121 56.118
525871 88.833 -58.142 56.086
525872 88.876 -58.162 56.055
525873 88.919 -58.184 56.026
525874 88.962 -58.204 55.997
525875 89.005 -58.225 55.971
525876 89.048 -58.247 55.948
525877 89.091 -58.267 55.925
525878 89.134 -58.288 55.905
525879 89.177 -58.309 55.886
525880 89.220 -58.330 55.870
525881 89.263 -58.350 55.856
525882 89.306 -58.372 55.844
525883 89.348 -58.391 55.836
525884 89.383 -58.401 55.830
525885 89.418 -58.412 55.826
525886 89.453 -58.422 55.825
525887 89.497 -58.429 55.828
525888 89.544 -58.435 55.834
525889 89.592 -58.441 55.839
525890 89.638 -58.448 55.840
525891 89.685 -58.454 55.836
525892 89.732 -58.460 55.828
525893 89.779 -58.466 55.816
525894 89.822 -58.470 55.804
525895 89.860 -58.481 55.794
525896 89.898 -58.498 55.789
525897 89.937 -58.515 55.786
525898 89.976 -58.532 55.780
525899 90.015 -58.550 55.773
525900 90.048 -58.565 55.768
525901 90.081 -58.579 55.760
525902 90.114 -58.595 55.747
525903 90.147 -58.610 55.727
525904 90.179 -58.624 55.700
525905 90.217 -58.632 55.668
525906 90.255 -58.640 55.638
525907 90.293 -58.648 55.613
525908 90.330 -58.654 55.594
525909 90.368 -58.662 55.581
525910 90.408 -58.677 55.574
525911 90.450 -58.695 55.564
525912 90.492 -58.714 55.550
525913 90.533 -58.733 55.533
525914 90.573 -58.748 55.513
525915 90.612 -58.753 55.488
525916 90.650 -58.758 55.458
525917 90.688 -58.763 55.426
525918 90.727 -58.768 55.390
525919 90.765 -58.773 55.350
525920 90.804 -58.778 55.311
525921 90.842 -58.783 55.277
525922 90.880 -58.789 55.248
525923 90.915 -58.810 55.223
525924 90.949 -58.830 55.199
525925 90.984 -58.851 55.176
525926 91.019 -58.871 55.154
525927 91.053 -58.892 55.133
525928 91.088 -58.912 55.112
525929 91.126 -58.934 55.090
525930 91.170 -58.959 55.066
525931 91.212 -58.978 55.038
525932 91.251 -58.993 55.010
525933 91.292 -59.008 54.978
525934 91.336 -59.021 54.942
525935 91.381 -59.034 54.902
525936 91.426 -59.047 54.865
525937 91.471 -59.060 54.839
525938 91.516 -59.072 54.817
525939 91.561 -59.085 54.800
525940 91.605 -59.098 54.786
525941 91.650 -59.111 54.775
525942 91.689 -59.125 54.770
525943 91.726 -59.140 54.767
525944 91.761 -59.154 54.770
525945 91.798 -59.168 54.779
525946 91.834 -59.183 54.790
525947 91.870 -59.197 54.798
525948 91.906 -59.200 54.801
525949 91.941 -59.200 54.802
525950 91.977 -59.201 54.801
525951 92.012 -59.202 54.800
525952 92.048 -59.202 54.796
525953 92.078 -59.205 54.780
525954 92.106 -59.210 54.761
525955 92.134 -59.216 54.743
525956 92.161 -59.222 54.732
525957 92.189 -59.227 54.733
525958 92.216 -59.232 54.731
525959 92.244 -59.237 54.723
525960 92.272 -59.244 54.709
525961 92.299 -59.249 54.689
525962 92.327 -59.254 54.663
525963 92.367 -59.256 54.636
525964 92.415 -59.255 54.613
525965 92.462 -59.254 54.591
525966 92.509 -59.253 54.578
525967 92.557 -59.253 54.583
525968 92.604 -59.252 54.588
525969 92.651 -59.251 54.580
525970 92.699 -59.250 54.571
525971 92.746 -59.250 54.563
525972 92.794 -59.249 54.556
525973 92.838 -59.254 54.552
525974 92.880 -59.264 54.546
525975 92.923 -59.273 54.533
525976 92.966 -59.283 54.512
525977 93.008 -59.293 54.487
525978 93.050 -59.302 54.461
525979 93.091 -59.312 54.434
525980 93.130 -59.320 54.410
525981 93.168 -59.328 54.391
525982 93.207 -59.337 54.375
525983 93.246 -59.346 54.363
525984 93.284 -59.354 54.355
525985 93.323 -59.363 54.347
525986 93.361 -59.371 54.341
525987 93.399 -59.379 54.338
525988 93.436 -59.399 54.333
525989 93.472 -59.419 54.327
525990 93.508 -59.440 54.321
525991 93.544 -59.461 54.314
525992 93.580 -59.482 54.308
525993 93.617 -59.502 54.305
525994 93.653 -59.520 54.304
525995 93.690 -59.536 54.303
525996 93.726 -59.551 54.301
525997 93.763 -59.567 54.298
525998 93.799 -59.582 54.293
525999 93.836 -59.598 54.285
526000 93.872 -59.613 54.275
526001 93.909 -59.629 54.260
526002 93.945 -59.645 54.245
526003 93.981 -59.660 54.228
526004 94.018 -59.676 54.207
526005 94.054 -59.692 54.188
526006 94.091 -59.707 54.168
526007 94.127 -59.723 54.146
526008 94.164 -59.739 54.122
526009 94.200 -59.754 54.103
526010 94.237 -59.770 54.096
526011 94.273 -59.785 54.088
526012 94.310 -59.801 54.081
526013 94.346 -59.817 54.074
526014 94.383 -59.832 54.067
526015 94.423 -59.842 54.059
526016 94.462 -59.852 54.046
526017 94.501 -59.864 54.026
526018 94.540 -59.874 54.008
526019 94.579 -59.885 53.990
526020 94.618 -59.895 53.969
526021 94.657 -59.907 53.942
526022 94.696 -59.917 53.911
526023 94.736 -59.928 53.872
526024 94.775 -59.939 53.830
526025 94.814 -59.949 53.789
526026 94.853 -59.960 53.749
526027 94.893 -59.970 53.710
526028 94.932 -59.980 53.674
526029 94.972 -59.991 53.642
526030 95.012 -60.001 53.613
526031 95.050 -60.011 53.585
526032 95.089 -60.020 53.557
526033 95.127 -60.028 53.530
526034 95.166 -60.036 53.504
526035 95.204 -60.046 53.477
526036 95.243 -60.054 53.448
526037 95.281 -60.062 53.420
526038 95.319 -60.070 53.396
526039 95.353 -60.076 53.375
526040 95.389 -60.082 53.356
526041 95.423 -60.088 53.336
526042 95.458 -60.095 53.313
526043 95.492 -60.101 53.287
526044 95.527 -60.108 53.259
526045 95.561 -60.114 53.231
526046 95.597 -60.119 53.201
526047 95.632 -60.124 53.168
526048 95.668 -60.129 53.131
526049 95.704 -60.134 53.091
526050 95.740 -60.138 53.055
526051 95.775 -60.143 53.030
526052 95.811 -60.147 53.014
526053 95.846 -60.152 53.000
526054 95.881 -60.159 52.983
526055 95.917 -60.165 52.961
526056 95.955 -60.170 52.938
526057 95.992 -60.170 52.914
526058 96.030 -60.169 52.888
526059 96.067 -60.168 52.861
526060 96.105 -60.168 52.829
526061 96.143 -60.167 52.798
526062 96.181 -60.166 52.766
526063 96.218 -60.165 52.733
526064 96.256 -60.165 52.699
526065 96.295 -60.165 52.665
526066 96.334 -60.166 52.626
526067 96.373 -60.167 52.588
526068 96.412 -60.168 52.550
526069 96.450 -60.168 52.513
526070 96.487 -60.170 52.479
526071 96.533 -60.176 52.443
526072 96.579 -60.182 52.403
526073 96.625 -60.189 52.361
526074 96.671 -60.195 52.318
526075 96.717 -60.201 52.276
526076 96.762 -60.208 52.233
526077 96.807 -60.215 52.191
526078 96.852 -60.224 52.150
526079 96.896 -60.234 52.108
526080 96.934 -60.246 52.068
526081 96.966 -60.259 52.029
526082 96.998 -60.273 51.992
526083 97.030 -60.287 51.954
526084 97.062 -60.300 51.916
526085 97.094 -60.314 51.878
526086 97.124 -60.327 51.839
526087 97.154 -60.340 51.801
526088 97.184 -60.353 51.763
526089 97.215 -60.365 51.726
526090 97.248 -60.372 51.689
526091 97.285 -60.373 51.651
526092 97.321 -60.374 51.614
526093 97.358 -60.375 51.579
526094 97.395 -60.375 51.544
526095 97.432 -60.376 51.510
526096 97.469 -60.377 51.478
526097 97.506 -60.378 51.445
526098 97.542 -60.378 51.412
526099 97.579 -60.379 51.379
526100 97.616 -60.380 51.346
526101 97.654 -60.382 51.313
526102 97.692 -60.392 51.279
526103 97.715 -60.394 51.243
526104 97.737 -60.394 51.205
526105 97.762 -60.397 51.167
526106 97.789 -60.400 51.130
526107 97.815 -60.402 51.091
526108 97.842 -60.405 51.052
526109 97.868 -60.407 51.012
526110 97.895 -60.410 50.973
526111 97.922 -60.413 50.932
526112 97.948 -60.416 50.892
526113 97.974 -60.419 50.851
526114 98.001 -60.422 50.811
526115 98.038 -60.428 50.770
526116 98.081 -60.424 50.728
526117 98.122 -60.421 50.685
526118 98.158 -60.425 50.640
526119 98.194 -60.429 50.594
526120 98.230 -60.434 50.548
526121 98.267 -60.438 50.502
526122 98.303 -60.442 50.456
526123 98.339 -60.447 50.409
526124 98.376 -60.451 50.363
526125 98.412 -60.455 50.317
526126 98.448 -60.459 50.271
526127 98.485 -60.464 50.226
526128 98.521 -60.468 50.182
526129 98.557 -60.472 50.138
526130 98.593 -60.477 50.094
526131 98.630 -60.481 50.051
526132 98.666 -60.485 50.008
526535 87.389 -57.732 57.151
526536 87.429 -57.746 57.135
526537 87.470 -57.759 57.118
526538 87.510 -57.772 57.102
526539 87.551 -57.786 57.085
526540 87.591 -57.799 57.067
526541 87.631 -57.812 57.050
526542 87.672 -57.826 57.031
526543 87.712 -57.839 57.011
526544 87.753 -57.852 56.991
526545 87.801 -57.858 56.970
526546 87.846 -57.866 56.947
526547 87.890 -57.873 56.922
526548 87.934 -57.880 56.897
526549 87.979 -57.888 56.871
526550 88.024 -57.895 56.845
526551 88.068 -57.902 56.817
526552 88.113 -57.909 56.788
526553 88.157 -57.917 56.759
526554 88.202 -57.924 56.729
526555 88.247 -57.931 56.699
526556 88.282 -57.943 56.668
526557 88.315 -57.954 56.636
526558 88.349 -57.965 56.603
526559 88.383 -57.977 56.570
526560 88.414 -57.984 56.535
526561 88.444 -57.990 56.498
526562 88.474 -57.996 56.462
526563 88.502 -58.001 56.424
526564 88.530 -58.007 56.387
526565 88.558 -58.012 56.350
526566 88.593 -58.021 56.313
526567 88.633 -58.031 56.275
526568 88.673 -58.042 56.237
526569 88.713 -58.053 56.201
526570 88.754 -58.063 56.167
526571 88.797 -58.084 56.134
526572 88.840 -58.104 56.101
526573 88.883 -58.125 56.068
526574 88.926 -58.146 56.038
526575 88.969 -58.167 56.009
526576 89.012 -58.187 55.982
526577 89.055 -58.209 55.958
526578 89.098 -58.230 55.934
526579 89.141 -58.250 55.914
526580 89.184 -58.272 55.894
526581 89.227 -58.292 55.877
526582 89.270 -58.313 55.862
526583 89.313 -58.334 55.850
526584 89.356 -58.355 55.840
526585 89.395 -58.371 55.833
526586 89.430 -58.382 55.827
526587 89.465 -58.392 55.825
526588 89.511 -58.398 55.826
526589 89.558 -58.404 55.832
526590 89.605 -58.410 55.835
526591 89.652 -58.416 55.835
526592 89.698 -58.422 55.831
526593 89.745 -58.429 55.823
526594 89.793 -58.435 55.809
526595 89.839 -58.441 55.799
526596 89.880 -58.447 55.793
526597 89.919 -58.464 55.790
526598 89.958 -58.482 55.786
526599 89.992 -58.497 55.781
526600 90.025 -58.512 55.776
526601 90.058 -58.527 55.770
526602 90.091 -58.542 55.759
526603 90.124 -58.557 55.743
526604 90.157 -58.572 55.720
526605 90.191 -58.583 55.692
526606 90.230 -58.590 55.660
526607 90.267 -58.598 55.631
526608 90.305 -58.605 55.610
526609 90.343 -58.612 55.594
526610 90.381 -58.620 55.584
526611 90.419 -58.627 55.576
526612 90.460 -58.645 55.566
526613 90.502 -58.664 55.554
526614 90.543 -58.683 55.538
526615 90.585 -58.702 55.518
526616 90.626 -58.720 55.490
526617 90.666 -58.732 55.458
526618 90.704 -58.736 55.422
526619 90.742 -58.742 55.383
526620 90.781 -58.746 55.344
526621 90.819 -58.751 55.307
526622 90.858 -58.756 55.277
526623 90.896 -58.761 55.249
526624 90.933 -58.774 55.224
526625 90.967 -58.795 55.202
526626 91.002 -58.815 55.178
526627 91.036 -58.836 55.158
526628 91.071 -58.856 55.139
526629 91.105 -58.877 55.120
526630 91.140 -58.897 55.102
526631 91.175 -58.918 55.081
526632 91.217 -58.939 55.054
526633 91.259 -58.953 55.026
526634 91.304 -58.966 54.995
526635 91.349 -58.979 54.959
526636 91.393 -58.992 54.919
526637 91.438 -59.005 54.879
526638 91.483 -59.018 54.848
526639 91.528 -59.030 54.823
526640 91.573 -59.043 54.804
526641 91.618 -59.056 54.790
526642 91.662 -59.069 54.777
526643 91.698 -59.084 54.770
526644 91.734 -59.098 54.765
526645 91.770 -59.112 54.766
526646 91.806 -59.127 54.774
526647 91.842 -59.141 54.789
526648 91.878 -59.156 54.796
526649 91.914 -59.162 54.801
526650 91.949 -59.162 54.805
526651 91.985 -59.162 54.808
526652 92.020 -59.162 54.810
526653 92.056 -59.163 54.808
526654 92.091 -59.164 54.799
526655 92.121 -59.168 54.784
526656 92.148 -59.173 54.765
526657 92.176 -59.179 54.749
526658 92.204 -59.184 54.746
526659 92.231 -59.190 54.745
526660 92.259 -59.195 54.739
526661 92.286 -59.200 54.729
526662 92.314 -59.206 54.714
526663 92.342 -59.212 54.694
526664 92.382 -59.213 54.669
526665 92.429 -59.212 54.645
526666 92.477 -59.211 54.623
526667 92.524 -59.210 54.605
526668 92.571 -59.209 54.591
526669 92.619 -59.209 54.590
526670 92.666 -59.208 54.581
526671 92.714 -59.207 54.569
526672 92.761 -59.206 54.562
526673 92.808 -59.206 54.563
526674 92.855 -59.206 54.566
526675 92.898 -59.216 54.561
526676 92.940 -59.226 54.548
526677 92.982 -59.235 54.529
526678 93.025 -59.245 54.506
526679 93.067 -59.255 54.479
526680 93.107 -59.263 54.449
526681 93.145 -59.272 54.423
526682 93.184 -59.281 54.400
526683 93.222 -59.289 54.381
526684 93.261 -59.297 54.366
526685 93.299 -59.306 54.356
526686 93.337 -59.315 54.347
526687 93.376 -59.323 54.341
526688 93.414 -59.331 54.338
526689 93.451 -59.350 54.332
526690 93.487 -59.371 54.325
526691 93.523 -59.392 54.315
526692 93.559 -59.413 54.303
526693 93.595 -59.433 54.294
526694 93.632 -59.452 54.288
526695 93.669 -59.468 54.284
526696 93.705 -59.483 54.281
526697 93.742 -59.499 54.280
526698 93.778 -59.514 54.277
526699 93.815 -59.530 54.273
526700 93.851 -59.545 54.267
526701 93.888 -59.561 54.258
526702 93.924 -59.576 54.245
526703 93.960 -59.592 54.230
526704 93.997 -59.607 54.212
526705 94.033 -59.623 54.192
526706 94.070 -59.639 54.172
526707 94.106 -59.654 54.154
526708 94.143 -59.670 54.135
526709 94.179 -59.685 54.115
526710 94.216 -59.701 54.099
526711 94.252 -59.717 54.089
526712 94.289 -59.732 54.083
526713 94.324 -59.751 54.077
526714 94.359 -59.772 54.074
526715 94.394 -59.792 54.069
526716 94.429 -59.804 54.061
526717 94.468 -59.814 54.050
526718 94.507 -59.825 54.033
526719 94.546 -59.836 54.016
526720 94.585 -59.847 53.998
526721 94.624 -59.858 53.976
526722 94.664 -59.868 53.950
526723 94.702 -59.879 53.919
526724 94.742 -59.890 53.881
526725 94.781 -59.901 53.841
526726 94.820 -59.911 53.798
526727 94.860 -59.921 53.755
526728 94.899 -59.931 53.715
526729 94.938 -59.942 53.680
526730 94.978 -59.952 53.651
526731 95.019 -59.962 53.624
526732 95.059 -59.973 53.596
526733 95.098 -59.982 53.565
526734 95.137 -59.991 53.535
526735 95.177 -60.000 53.506
526736 95.215 -60.008 53.476
526737 95.254 -60.017 53.443
526738 95.293 -60.026 53.411
526739 95.328 -60.032 53.383
526740 95.363 -60.038 53.360
526741 95.398 -60.045 53.339
526742 95.432 -60.052 53.319
526743 95.467 -60.058 53.296
526744 95.501 -60.065 53.270
526745 95.536 -60.071 53.241
526746 95.570 -60.077 53.212
526747 95.605 -60.083 53.183
526748 95.641 -60.087 53.153
526749 95.677 -60.092 53.118
526750 95.712 -60.096 53.082
526751 95.749 -60.101 53.054
526752 95.784 -60.105 53.035
526753 95.819 -60.112 53.022
526754 95.853 -60.118 53.012
526755 95.888 -60.125 52.995
526756 95.924 -60.131 52.975
526757 95.962 -60.136 52.953
526758 96.000 -60.141 52.930
526759 96.037 -60.141 52.905
526760 96.075 -60.140 52.878
526761 96.112 -60.140 52.848
526762 96.150 -60.139 52.818
526763 96.188 -60.138 52.785
526764 96.226 -60.137 52.752
526765 96.263 -60.137 52.718
526766 96.303 -60.137 52.684
526767 96.341 -60.139 52.647
526768 96.381 -60.140 52.610
526769 96.418 -60.140 52.573
526770 96.454 -60.140 52.538
526771 96.494 -60.142 52.503
526772 96.540 -60.149 52.466
526773 96.586 -60.155 52.426
526774 96.631 -60.161 52.385
526775 96.677 -60.168 52.344
526776 96.723 -60.174 52.302
526777 96.768 -60.181 52.259
526778 96.814 -60.187 52.216
526779 96.860 -60.193 52.172
526780 96.906 -60.199 52.129
526781 96.942 -60.209 52.088
526782 96.972 -60.221 52.048
526783 97.002 -60.234 52.009
526784 97.032 -60.247 51.970
526785 97.063 -60.260 51.931
526786 97.093 -60.272 51.891
526787 97.123 -60.285 51.852
526788 97.153 -60.298 51.812
526789 97.183 -60.311 51.773
526790 97.214 -60.324 51.734
526791 97.246 -60.334 51.696
526792 97.283 -60.334 51.657
526793 97.319 -60.335 51.619
526794 97.356 -60.336 51.582
526795 97.393 -60.337 51.546
526796 97.430 -60.337 51.511
526797 97.466 -60.337 51.478
526798 97.503 -60.338 51.445
526799 97.540 -60.339 51.412
526800 97.577 -60.340 51.379
526801 97.614 -60.340 51.346
526802 97.653 -60.347 51.314
526803 97.694 -60.358 51.282
526804 97.731 -60.368 51.245
526805 97.754 -60.369 51.206
526806 97.776 -60.369 51.168
526807 97.799 -60.371 51.130
526808 97.821 -60.372 51.092
526809 97.844 -60.373 51.053
526810 97.868 -60.375 51.014
526811 97.895 -60.378 50.975
526812 97.922 -60.381 50.935
526813 97.948 -60.383 50.896
526814 97.974 -60.386 50.856
526815 98.001 -60.389 50.818
526816 98.040 -60.397 50.777
526817 98.080 -60.402 50.735
526818 98.127 -60.385 50.692
526819 98.164 -60.387 50.646
526820 98.200 -60.392 50.599
526821 98.237 -60.397 50.551
526822 98.273 -60.400 50.506
526823 98.309 -60.405 50.460
526824 98.346 -60.409 50.414
526825 98.382 -60.413 50.368
526826 98.418 -60.418 50.322
526827 98.454 -60.422 50.276
526828 98.491 -60.426 50.231
526829 98.527 -60.431 50.187
526830 98.563 -60.435 50.143
526831 98.599 -60.440 50.099
526832 98.636 -60.444 50.055
526833 98.672 -60.448 50.012
526834 98.709 -60.453 49.969
527237 87.446 -57.702 57.126
527238 87.487 -57.715 57.109
527239 87.527 -57.729 57.093
527240 87.567 -57.742 57.077
527241 87.608 -57.755 57.058
527242 87.648 -57.769 57.041
527243 87.689 -57.782 57.022
527244 87.729 -57.796 57.003
527245 87.776 -57.804 56.982
527246 87.821 -57.811 56.961
527247 87.865 -57.819 56.938
527248 87.910 -57.826 56.915
527249 87.955 -57.833 56.890
527250 87.999 -57.840 56.864
527251 88.044 -57.848 56.837
527252 88.088 -57.855 56.811
527253 88.133 -57.862 56.783
527254 88.178 -57.870 56.755
527255 88.222 -57.877 56.726
527256 88.260 -57.887 56.697
527257 88.294 -57.899 56.667
527258 88.328 -57.910 56.635
527259 88.361 -57.921 56.603
527260 88.395 -57.933 56.571
527261 88.428 -57.944 56.537
527262 88.462 -57.955 56.502
527263 88.492 -57.961 56.468
527264 88.522 -57.966 56.431
527265 88.552 -57.971 56.396
527266 88.582 -57.977 56.360
527267 88.611 -57.983 56.324
527268 88.638 -57.989 56.288
527269 88.678 -57.999 56.251
527270 88.718 -58.009 56.215
527271 88.760 -58.024 56.181
527272 88.803 -58.046 56.148
527273 88.846 -58.067 56.114
527274 88.889 -58.087 56.082
527275 88.932 -58.109 56.051
527276 88.975 -58.129 56.021
527277 89.018 -58.150 55.993
527278 89.061 -58.171 55.967
527279 89.104 -58.192 55.944
527280 89.147 -58.213 55.923
527281 89.190 -58.234 55.902
527282 89.234 -58.255 55.884
527283 89.276 -58.275 55.868
527284 89.319 -58.297 55.856
527285 89.363 -58.317 55.845
527286 89.405 -58.338 55.836
527287 89.442 -58.351 55.829
527288 89.477 -58.361 55.826
527289 89.524 -58.367 55.827
527290 89.571 -58.373 55.829
527291 89.618 -58.379 55.829
527292 89.665 -58.385 55.827
527293 89.712 -58.391 55.821
527294 89.759 -58.397 55.813
527295 89.806 -58.404 55.803
527296 89.852 -58.410 55.796
527297 89.899 -58.416 55.793
527298 89.937 -58.429 55.790
527299 89.969 -58.444 55.785
527300 90.002 -58.459 55.782
527301 90.035 -58.474 55.776
527302 90.068 -58.488 55.767
527303 90.101 -58.504 55.753
527304 90.133 -58.519 55.734
527305 90.167 -58.533 55.710
527306 90.204 -58.541 55.682
527307 90.242 -58.548 55.652
527308 90.280 -58.555 55.626
527309 90.318 -58.563 55.607
527310 90.356 -58.570 55.595
527311 90.394 -58.577 55.589
527312 90.432 -58.585 55.580
527313 90.471 -58.596 55.570
527314 90.512 -58.614 55.556
527315 90.554 -58.633 55.538
527316 90.595 -58.652 55.516
527317 90.637 -58.670 55.486
527318 90.678 -58.689 55.450
527319 90.719 -58.708 55.411
527320 90.758 -58.715 55.374
527321 90.797 -58.720 55.341
527322 90.835 -58.725 55.313
527323 90.873 -58.730 55.287
527324 90.911 -58.735 55.261
527325 90.950 -58.739 55.238
527326 90.985 -58.759 55.216
527327 91.019 -58.780 55.190
527328 91.054 -58.800 55.165
527329 91.089 -58.820 55.144
527330 91.123 -58.841 55.129
527331 91.158 -58.861 55.115
527332 91.192 -58.882 55.095
527333 91.229 -58.898 55.068
527334 91.271 -58.911 55.038
527335 91.316 -58.924 55.004
527336 91.361 -58.936 54.968
527337 91.406 -58.949 54.930
527338 91.451 -58.962 54.889
527339 91.496 -58.975 54.857
527340 91.540 -58.988 54.832
527341 91.585 -59.001 54.811
527342 91.630 -59.014 54.795
527343 91.670 -59.027 54.780
527344 91.706 -59.042 54.770
527345 91.742 -59.056 54.764
527346 91.779 -59.071 54.762
527347 91.814 -59.085 54.765
527348 91.851 -59.099 54.776
527349 91.887 -59.114 54.786
527350 91.923 -59.123 54.793
527351 91.958 -59.123 54.796
527352 91.994 -59.124 54.798
527353 92.029 -59.124 54.803
527354 92.064 -59.124 54.808
527355 92.100 -59.125 54.811
527356 92.135 -59.125 54.806
527357 92.163 -59.131 54.792
527358 92.191 -59.136 54.772
527359 92.218 -59.141 54.754
527360 92.246 -59.147 54.746
527361 92.274 -59.153 54.742
527362 92.301 -59.158 54.737
527363 92.329 -59.163 54.730
527364 92.356 -59.168 54.720
527365 92.397 -59.170 54.701
527366 92.444 -59.169 54.677
527367 92.491 -59.168 54.656
527368 92.538 -59.168 54.635
527369 92.586 -59.167 54.609
527370 92.633 -59.166 54.603
527371 92.681 -59.165 54.593
527372 92.728 -59.165 54.581
527373 92.776 -59.164 54.572
527374 92.823 -59.163 54.567
527375 92.870 -59.162 54.564
527376 92.914 -59.168 54.562
527377 92.957 -59.178 54.553
527378 92.999 -59.188 54.536
527379 93.042 -59.197 54.513
527380 93.083 -59.207 54.486
527381 93.122 -59.215 54.455
527382 93.160 -59.224 54.429
527383 93.199 -59.232 54.406
527384 93.237 -59.241 54.387
527385 93.276 -59.250 54.372
527386 93.314 -59.258 54.359
527387 93.352 -59.266 54.349
527388 93.391 -59.275 54.341
527389 93.429 -59.284 54.334
527390 93.466 -59.302 54.328
527391 93.502 -59.323 54.323
527392 93.538 -59.344 54.312
527393 93.574 -59.364 54.298
527394 93.611 -59.384 54.285
527395 93.647 -59.399 54.275
527396 93.684 -59.415 54.265
527397 93.720 -59.430 54.261
527398 93.757 -59.446 54.258
527399 93.793 -59.462 54.254
527400 93.830 -59.477 54.250
527401 93.866 -59.493 54.244
527402 93.903 -59.508 54.234
527403 93.939 -59.524 54.223
527404 93.976 -59.539 54.208
527405 94.012 -59.555 54.191
527406 94.049 -59.570 54.172
527407 94.085 -59.586 54.154
527408 94.122 -59.602 54.136
527409 94.158 -59.617 54.118
527410 94.194 -59.636 54.101
527411 94.228 -59.656 54.087
527412 94.264 -59.676 54.076
527413 94.298 -59.697 54.070
527414 94.334 -59.717 54.065
527415 94.368 -59.737 54.063
527416 94.404 -59.757 54.060
527417 94.437 -59.766 54.053
527418 94.475 -59.776 54.045
527419 94.513 -59.786 54.030
527420 94.552 -59.798 54.014
527421 94.591 -59.808 53.996
527422 94.630 -59.819 53.973
527423 94.670 -59.830 53.946
527424 94.708 -59.841 53.914
527425 94.748 -59.851 53.876
527426 94.787 -59.862 53.836
527427 94.826 -59.872 53.795
527428 94.866 -59.883 53.754
527429 94.905 -59.893 53.715
527430 94.946 -59.903 53.684
527431 94.986 -59.914 53.657
527432 95.027 -59.924 53.630
527433 95.067 -59.935 53.603
527434 95.107 -59.944 53.576
527435 95.146 -59.953 53.549
527436 95.185 -59.962 53.520
527437 95.225 -59.971 53.488
527438 95.264 -59.980 53.454
527439 95.303 -59.989 53.419
527440 95.338 -59.995 53.383
527441 95.372 -60.002 53.351
527442 95.406 -60.008 53.325
527443 95.441 -60.014 53.302
527444 95.475 -60.021 53.279
527445 95.510 -60.027 53.252
527446 95.544 -60.034 53.224
527447 95.579 -60.040 53.194
527448 95.614 -60.046 53.165
527449 95.650 -60.050 53.137
527450 95.686 -60.055 53.105
527451 95.722 -60.059 53.076
527452 95.757 -60.065 53.055
527453 95.792 -60.071 53.042
527454 95.826 -60.078 53.032
527455 95.861 -60.085 53.024
527456 95.895 -60.091 53.008
527457 95.931 -60.097 52.990
527458 95.969 -60.102 52.969
527459 96.007 -60.107 52.946
527460 96.044 -60.112 52.921
527461 96.082 -60.112 52.894
527462 96.119 -60.112 52.865
527463 96.157 -60.111 52.834
527464 96.195 -60.110 52.800
527465 96.233 -60.109 52.766
527466 96.271 -60.109 52.734
527467 96.310 -60.110 52.702
527468 96.349 -60.111 52.667
527469 96.387 -60.112 52.632
527470 96.423 -60.112 52.598
527471 96.459 -60.112 52.563
527472 96.501 -60.115 52.528
527473 96.546 -60.121 52.491
527474 96.592 -60.127 52.452
527475 96.638 -60.134 52.411
527476 96.683 -60.140 52.370
527477 96.729 -60.146 52.327
527478 96.775 -60.152 52.283
527479 96.821 -60.159 52.239
527480 96.867 -60.165 52.196
527481 96.907 -60.168 52.153
527482 96.940 -60.169 52.111
527483 96.971 -60.180 52.070
527484 97.002 -60.193 52.031
527485 97.032 -60.206 51.992
527486 97.062 -60.218 51.952
527487 97.092 -60.231 51.911
527488 97.122 -60.243 51.870
527489 97.153 -60.256 51.829
527490 97.183 -60.269 51.788
527491 97.213 -60.282 51.747
527492 97.243 -60.295 51.707
527493 97.280 -60.295 51.667
527494 97.317 -60.296 51.628
527495 97.354 -60.297 51.590
527496 97.391 -60.297 51.553
527497 97.428 -60.298 51.517
527498 97.464 -60.299 51.482
527499 97.501 -60.300 51.449
527500 97.538 -60.300 51.415
527501 97.575 -60.301 51.381
527502 97.611 -60.302 51.348
527503 97.652 -60.312 51.315
527504 97.693 -60.323 51.281
527505 97.734 -60.334 51.245
527506 97.771 -60.343 51.208
527507 97.793 -60.343 51.171
527508 97.815 -60.345 51.134
527509 97.838 -60.346 51.096
527510 97.860 -60.347 51.058
527511 97.882 -60.348 51.018
527512 97.905 -60.349 50.979
527513 97.927 -60.350 50.939
527514 97.950 -60.351 50.901
527515 97.975 -60.353 50.862
527516 98.002 -60.356 50.823
527517 98.041 -60.365 50.783
527518 98.081 -60.373 50.742
527519 98.126 -60.363 50.699
527520 98.170 -60.350 50.654
527521 98.207 -60.354 50.607
527522 98.243 -60.359 50.558
527523 98.279 -60.363 50.511
527524 98.315 -60.368 50.466
527525 98.352 -60.372 50.420
527526 98.388 -60.376 50.373
527527 98.424 -60.381 50.328
527528 98.460 -60.384 50.282
527529 98.497 -60.389 50.236
527530 98.533 -60.394 50.192
527531 98.570 -60.398 50.148
527532 98.606 -60.402 50.104
527533 98.642 -60.406 50.060
527534 98.678 -60.411 50.016
527535 98.715 -60.415 49.973
527940 87.544 -57.686 57.083
527941 87.584 -57.699 57.067
527942 87.625 -57.712 57.049
527943 87.665 -57.726 57.031
527944 87.706 -57.739 57.013
527945 87.748 -57.752 56.993
527946 87.796 -57.757 56.972
527947 87.841 -57.764 56.951
527948 87.886 -57.772 56.928
527949 87.930 -57.779 56.905
527950 87.975 -57.786 56.881
527951 88.020 -57.793 56.855
527952 88.064 -57.801 56.829
527953 88.109 -57.808 56.803
527954 88.153 -57.815 56.775
527955 88.198 -57.823 56.747
527956 88.239 -57.831 56.719
527957 88.272 -57.843 56.690
527958 88.306 -57.855 56.660
527959 88.340 -57.866 56.630
527960 88.374 -57.877 56.600
527961 88.407 -57.889 56.569
527962 88.441 -57.900 56.537
527963 88.475 -57.912 56.505
527964 88.508 -57.924 56.471
527965 88.540 -57.931 56.437
527966 88.570 -57.937 56.403
527967 88.600 -57.943 56.368
527968 88.630 -57.948 56.333
527969 88.661 -57.953 56.299
527970 88.690 -57.959 56.263
527971 88.724 -57.967 56.228
527972 88.767 -57.987 56.194
527973 88.810 -58.008 56.161
527974 88.853 -58.029 56.128
527975 88.896 -58.050 56.096
527976 88.939 -58.071 56.064
527977 88.982 -58.092 56.033
527978 89.025 -58.112 56.004
527979 89.068 -58.134 55.977
527980 89.111 -58.154 55.953
527981 89.154 -58.175 55.931
527982 89.197 -58.196 55.910
527983 89.240 -58.217 55.891
527984 89.283 -58.238 55.873
527985 89.326 -58.259 55.861
527986 89.369 -58.280 55.850
527987 89.412 -58.300 55.840
527988 89.454 -58.320 55.833
527989 89.490 -58.331 55.829
527990 89.537 -58.337 55.829
527991 89.585 -58.342 55.827
527992 89.631 -58.348 55.823
527993 89.678 -58.354 55.818
527994 89.725 -58.360 55.812
527995 89.771 -58.366 55.805
527996 89.816 -58.370 55.798
527997 89.862 -58.375 55.794
527998 89.907 -58.379 55.791
527999 89.947 -58.389 55.789
528000 89.979 -58.406 55.785
528001 90.012 -58.421 55.781
528002 90.045 -58.436 55.772
528003 90.078 -58.451 55.760
528004 90.110 -58.466 55.743
528005 90.144 -58.481 55.721
528006 90.179 -58.491 55.696
528007 90.217 -58.498 55.669
528008 90.255 -58.505 55.643
528009 90.293 -58.513 55.621
528010 90.331 -58.520 55.606
528011 90.369 -58.527 55.598
528012 90.406 -58.535 55.594
528013 90.445 -58.542 55.585
528014 90.482 -58.549 55.573
528015 90.523 -58.565 55.557
528016 90.564 -58.583 55.536
528017 90.606 -58.602 55.510
528018 90.647 -58.621 55.478
528019 90.688 -58.639 55.440
528020 90.730 -58.658 55.402
528021 90.771 -58.677 55.371
528022 90.812 -58.693 55.346
528023 90.851 -58.698 55.322
528024 90.889 -58.703 55.298
528025 90.927 -58.708 55.273
528026 90.965 -58.713 55.248
528027 91.003 -58.723 55.226
528028 91.037 -58.744 55.202
528029 91.072 -58.764 55.175
528030 91.106 -58.785 55.152
528031 91.141 -58.805 55.134
528032 91.176 -58.826 55.120
528033 91.212 -58.843 55.100
528034 91.249 -58.856 55.074
528035 91.287 -58.869 55.039
528036 91.328 -58.881 55.002
528037 91.373 -58.894 54.964
528038 91.418 -58.907 54.928
528039 91.463 -58.920 54.892
528040 91.508 -58.933 54.865
528041 91.553 -58.946 54.841
528042 91.597 -58.958 54.819
528043 91.642 -58.971 54.801
528044 91.678 -58.986 54.785
528045 91.715 -59.000 54.773
528046 91.750 -59.015 54.764
528047 91.787 -59.029 54.758
528048 91.823 -59.043 54.757
528049 91.859 -59.058 54.762
528050 91.895 -59.072 54.772
528051 91.931 -59.084 54.778
528052 91.966 -59.084 54.783
528053 92.002 -59.085 54.789
528054 92.037 -59.085 54.798
528055 92.073 -59.086 54.807
528056 92.108 -59.086 54.812
528057 92.144 -59.087 54.811
528058 92.178 -59.088 54.804
528059 92.206 -59.093 54.791
528060 92.233 -59.099 54.775
528061 92.261 -59.104 54.766
528062 92.288 -59.110 54.760
528063 92.316 -59.115 54.754
528064 92.344 -59.121 54.748
528065 92.371 -59.126 54.739
528066 92.411 -59.128 54.725
528067 92.458 -59.127 54.706
528068 92.506 -59.126 54.685
528069 92.553 -59.125 54.664
528070 92.600 -59.124 54.647
528071 92.648 -59.124 54.634
528072 92.696 -59.123 54.621
528073 92.743 -59.122 54.608
528074 92.790 -59.121 54.593
528075 92.838 -59.121 54.571
528076 92.885 -59.120 54.558
528077 92.931 -59.121 54.557
528078 92.974 -59.131 54.550
528079 93.016 -59.140 54.534
528080 93.059 -59.150 54.512
528081 93.098 -59.159 54.484
528082 93.137 -59.168 54.456
528083 93.175 -59.176 54.432
528084 93.214 -59.184 54.411
528085 93.252 -59.193 54.392
528086 93.290 -59.201 54.377
528087 93.329 -59.210 54.364
528088 93.367 -59.218 54.353
528089 93.406 -59.227 54.342
528090 93.444 -59.235 54.333
528091 93.481 -59.254 54.325
528092 93.517 -59.275 54.317
528093 93.553 -59.295 54.306
528094 93.589 -59.316 54.292
528095 93.626 -59.331 54.277
528096 93.662 -59.347 54.265
528097 93.699 -59.362 54.253
528098 93.735 -59.378 54.244
528099 93.772 -59.393 54.237
528100 93.808 -59.409 54.229
528101 93.845 -59.425 54.222
528102 93.882 -59.440 54.215
528103 93.918 -59.456 54.206
528104 93.955 -59.471 54.193
528105 93.991 -59.487 54.178
528106 94.028 -59.502 54.163
528107 94.063 -59.520 54.150
528108 94.098 -59.541 54.132
528109 94.133 -59.561 54.113
528110 94.168 -59.581 54.096
528111 94.203 -59.601 54.082
528112 94.238 -59.622 54.069
528113 94.273 -59.642 54.057
528114 94.308 -59.662 54.050
528115 94.343 -59.682 54.047
528116 94.378 -59.703 54.045
528117 94.413 -59.723 54.043
528118 94.447 -59.729 54.037
528119 94.483 -59.738 54.031
528120 94.521 -59.748 54.019
528121 94.559 -59.759 54.004
528122 94.598 -59.770 53.985
528123 94.636 -59.781 53.961
528124 94.676 -59.792 53.934
528125 94.714 -59.802 53.900
528126 94.754 -59.813 53.863
528127 94.793 -59.823 53.825
528128 94.833 -59.834 53.786
528129 94.873 -59.844 53.747
528130 94.913 -59.855 53.710
528131 94.954 -59.865 53.682
528132 94.994 -59.875 53.657
528133 95.035 -59.886 53.632
528134 95.075 -59.896 53.607
528135 95.115 -59.906 53.582
528136 95.155 -59.914 53.555
528137 95.194 -59.924 53.527
528138 95.233 -59.933 53.496
528139 95.273 -59.942 53.462
528140 95.313 -59.951 53.427
528141 95.349 -59.958 53.390
528142 95.382 -59.964 53.355
528143 95.415 -59.971 53.323
528144 95.450 -59.978 53.293
528145 95.484 -59.984 53.265
528146 95.519 -59.991 53.236
528147 95.553 -59.997 53.208
528148 95.588 -60.004 53.179
528149 95.623 -60.009 53.152
528150 95.659 -60.014 53.127
528151 95.694 -60.018 53.101
528152 95.729 -60.024 53.078
528153 95.764 -60.031 53.062
528154 95.798 -60.037 53.052
528155 95.833 -60.044 53.045
528156 95.867 -60.051 53.037
528157 95.902 -60.057 53.023
528158 95.938 -60.063 53.006
528159 95.976 -60.068 52.985
528160 96.014 -60.073 52.962
528161 96.051 -60.078 52.938
528162 96.089 -60.083 52.909
528163 96.126 -60.083 52.880
528164 96.164 -60.082 52.846
528165 96.202 -60.082 52.811
528166 96.240 -60.081 52.777
528167 96.278 -60.081 52.747
528168 96.318 -60.082 52.717
528169 96.355 -60.083 52.686
528170 96.391 -60.083 52.655
528171 96.428 -60.083 52.622
528172 96.464 -60.083 52.589
528173 96.507 -60.087 52.554
528174 96.553 -60.093 52.518
528175 96.599 -60.100 52.479
528176 96.644 -60.106 52.439
528177 96.690 -60.112 52.397
528178 96.736 -60.119 52.352
528179 96.782 -60.125 52.307
528180 96.827 -60.131 52.264
528181 96.869 -60.134 52.220
528182 96.903 -60.134 52.178
528183 96.937 -60.136 52.136
528184 96.970 -60.138 52.094
528185 97.001 -60.151 52.054
528186 97.031 -60.164 52.015
528187 97.061 -60.177 51.974
528188 97.091 -60.190 51.932
528189 97.121 -60.202 51.890
528190 97.152 -60.215 51.847
528191 97.182 -60.228 51.804
528192 97.212 -60.240 51.762
528193 97.242 -60.253 51.721
528194 97.278 -60.256 51.680
528195 97.315 -60.257 51.640
528196 97.352 -60.258 51.600
528197 97.388 -60.259 51.562
528198 97.425 -60.259 51.524
528199 97.462 -60.260 51.489
528200 97.499 -60.261 51.454
528201 97.536 -60.262 51.420
528202 97.573 -60.262 51.386
528203 97.610 -60.265 51.354
528204 97.651 -60.276 51.321
528205 97.692 -60.287 51.286
528206 97.733 -60.299 51.250
528207 97.774 -60.309 51.213
528208 97.809 -60.318 51.176
528209 97.831 -60.318 51.139
528210 97.854 -60.320 51.100
528211 97.876 -60.321 51.062
528212 97.899 -60.322 51.022
528213 97.921 -60.323 50.982
528214 97.944 -60.325 50.943
528215 97.966 -60.325 50.904
528216 97.989 -60.326 50.865
528217 98.010 -60.327 50.827
528218 98.043 -60.333 50.788
528219 98.083 -60.341 50.747
528220 98.125 -60.341 50.705
528221 98.171 -60.325 50.662
528222 98.213 -60.317 50.616
528223 98.249 -60.322 50.567
528224 98.285 -60.326 50.519
528225 98.321 -60.330 50.472
528226 98.358 -60.334 50.426
528227 98.394 -60.339 50.379
528228 98.431 -60.343 50.333
528229 98.467 -60.347 50.288
528230 98.503 -60.352 50.242
528231 98.539 -60.356 50.198
528232 98.576 -60.360 50.153
528233 98.612 -60.365 50.109
528234 98.648 -60.369 50.065
528235 98.685 -60.373 50.021
528236 98.721 -60.378 49.978
528642 87.601 -57.655 57.055
528643 87.642 -57.669 57.038
528644 87.682 -57.683 57.019
528645 87.722 -57.695 57.000
528646 87.772 -57.703 56.981
528647 87.817 -57.710 56.961
528648 87.861 -57.717 56.940
528649 87.906 -57.724 56.917
528650 87.950 -57.732 56.894
528651 87.995 -57.739 56.869
528652 88.039 -57.746 56.844
528653 88.084 -57.754 56.819
528654 88.128 -57.761 56.794
528655 88.173 -57.768 56.766
528656 88.217 -57.776 56.738
528657 88.251 -57.787 56.711
528658 88.284 -57.799 56.682
528659 88.318 -57.811 56.653
528660 88.352 -57.822 56.624
528661 88.386 -57.833 56.594
528662 88.419 -57.845 56.565
528663 88.453 -57.856 56.534
528664 88.487 -57.868 56.504
528665 88.521 -57.880 56.472
528666 88.554 -57.891 56.439
528667 88.587 -57.902 56.406
528668 88.617 -57.907 56.373
528669 88.647 -57.913 56.340
528670 88.678 -57.918 56.307
528671 88.708 -57.924 56.274
528672 88.738 -57.930 56.240
528673 88.768 -57.951 56.207
528674 88.812 -57.971 56.174
528675 88.855 -57.993 56.143
528676 88.899 -58.013 56.109
528677 88.942 -58.034 56.077
528678 88.985 -58.055 56.045
528679 89.029 -58.075 56.015
528680 89.072 -58.096 55.988
528681 89.116 -58.117 55.963
528682 89.159 -58.138 55.940
528683 89.203 -58.159 55.918
528684 89.246 -58.180 55.897
528685 89.290 -58.200 55.879
528686 89.333 -58.222 55.866
528687 89.376 -58.242 55.855
528688 89.419 -58.263 55.844
528689 89.462 -58.284 55.835
528690 89.504 -58.300 55.829
528691 89.551 -58.306 55.826
528692 89.596 -58.310 55.822
528693 89.641 -58.315 55.817
528694 89.686 -58.319 55.812
528695 89.732 -58.324 55.807
528696 89.777 -58.328 55.800
528697 89.822 -58.333 55.796
528698 89.868 -58.338 55.793
528699 89.913 -58.342 55.790
528700 89.957 -58.347 55.787
528701 89.989 -58.365 55.785
528702 90.022 -58.382 55.779
528703 90.054 -58.398 55.767
528704 90.088 -58.413 55.751
528705 90.120 -58.428 55.732
528706 90.154 -58.441 55.709
528707 90.192 -58.448 55.683
528708 90.230 -58.456 55.659
528709 90.268 -58.463 55.636
528710 90.306 -58.470 55.618
528711 90.343 -58.477 55.606
528712 90.382 -58.485 55.600
528713 90.419 -58.492 55.593
528714 90.457 -58.499 55.584
528715 90.495 -58.507 55.569
528716 90.533 -58.515 55.549
528717 90.575 -58.533 55.524
528718 90.616 -58.552 55.496
528719 90.658 -58.571 55.465
528720 90.699 -58.590 55.430
528721 90.740 -58.608 55.399
528722 90.782 -58.627 55.376
528723 90.823 -58.646 55.359
528724 90.865 -58.664 55.335
528725 90.904 -58.676 55.309
528726 90.943 -58.682 55.283
528727 90.981 -58.686 55.257
528728 91.020 -58.692 55.234
528729 91.055 -58.708 55.211
528730 91.090 -58.729 55.184
528731 91.124 -58.749 55.159
528732 91.159 -58.770 55.138
528733 91.194 -58.789 55.122
528734 91.232 -58.802 55.103
528735 91.270 -58.814 55.076
528736 91.308 -58.827 55.043
528737 91.346 -58.839 55.007
528738 91.385 -58.852 54.972
528739 91.430 -58.864 54.945
528740 91.475 -58.877 54.918
528741 91.520 -58.890 54.893
528742 91.565 -58.903 54.866
528743 91.609 -58.916 54.838
528744 91.651 -58.930 54.813
528745 91.687 -58.944 54.795
528746 91.723 -58.958 54.778
528747 91.759 -58.973 54.766
528748 91.795 -58.987 54.757
528749 91.831 -59.002 54.752
528750 91.867 -59.016 54.749
528751 91.904 -59.030 54.752
528752 91.939 -59.045 54.758
528753 91.975 -59.046 54.768
528754 92.010 -59.046 54.779
528755 92.046 -59.046 54.793
528756 92.081 -59.047 54.803
528757 92.117 -59.047 54.809
528758 92.152 -59.048 54.811
528759 92.188 -59.049 54.809
528760 92.220 -59.051 54.802
528761 92.248 -59.056 54.788
528762 92.275 -59.062 54.778
528763 92.303 -59.067 54.769
528764 92.331 -59.072 54.761
528765 92.358 -59.078 54.752
528766 92.386 -59.084 54.744
528767 92.426 -59.085 54.736
528768 92.473 -59.084 54.723
528769 92.520 -59.084 54.706
528770 92.568 -59.083 54.691
528771 92.616 -59.082 54.678
528772 92.663 -59.081 54.663
528773 92.710 -59.081 54.648
528774 92.758 -59.080 54.633
528775 92.805 -59.079 54.617
528776 92.852 -59.078 54.600
528777 92.900 -59.077 54.578
528778 92.947 -59.076 54.558
528779 92.991 -59.084 54.539
528780 93.033 -59.093 54.520
528781 93.075 -59.102 54.499
528782 93.113 -59.111 54.474
528783 93.152 -59.119 54.452
528784 93.190 -59.128 54.432
528785 93.228 -59.137 54.414
528786 93.267 -59.145 54.397
528787 93.305 -59.153 54.384
528788 93.344 -59.162 54.370
528789 93.382 -59.170 54.358
528790 93.421 -59.179 54.346
528791 93.460 -59.187 54.335
528792 93.496 -59.206 54.323
528793 93.532 -59.226 54.313
528794 93.568 -59.247 54.301
528795 93.605 -59.262 54.287
528796 93.641 -59.278 54.272
528797 93.678 -59.294 54.258
528798 93.714 -59.309 54.244
528799 93.751 -59.325 54.231
528800 93.787 -59.341 54.220
528801 93.824 -59.356 54.209
528802 93.860 -59.372 54.198
528803 93.897 -59.388 54.187
528804 93.933 -59.405 54.177
528805 93.968 -59.425 54.165
528806 94.003 -59.445 54.151
528807 94.038 -59.466 54.136
528808 94.073 -59.486 54.120
528809 94.108 -59.506 54.101
528810 94.142 -59.526 54.082
528811 94.178 -59.547 54.065
528812 94.212 -59.567 54.051
528813 94.248 -59.587 54.039
528814 94.282 -59.607 54.030
528815 94.318 -59.628 54.024
528816 94.352 -59.648 54.021
528817 94.388 -59.668 54.018
528818 94.422 -59.685 54.015
528819 94.456 -59.693 54.010
528820 94.490 -59.700 54.003
528821 94.529 -59.710 53.995
528822 94.567 -59.721 53.980
528823 94.605 -59.732 53.961
528824 94.643 -59.742 53.937
528825 94.682 -59.753 53.910
528826 94.720 -59.764 53.876
528827 94.760 -59.774 53.841
528828 94.800 -59.785 53.806
528829 94.840 -59.795 53.772
528830 94.881 -59.806 53.735
528831 94.921 -59.816 53.702
528832 94.962 -59.826 53.675
528833 95.002 -59.837 53.652
528834 95.043 -59.848 53.630
528835 95.084 -59.858 53.608
528836 95.124 -59.867 53.583
528837 95.164 -59.877 53.557
528838 95.203 -59.886 53.530
528839 95.244 -59.895 53.500
528840 95.283 -59.904 53.467
528841 95.324 -59.913 53.433
528842 95.364 -59.922 53.396
528843 95.396 -59.929 53.361
528844 95.428 -59.935 53.327
528845 95.461 -59.941 53.293
528846 95.493 -59.948 53.261
528847 95.527 -59.954 53.231
528848 95.562 -59.961 53.204
528849 95.596 -59.967 53.177
528850 95.631 -59.972 53.153
528851 95.667 -59.977 53.129
528852 95.701 -59.984 53.104
528853 95.736 -59.990 53.085
528854 95.770 -59.997 53.073
528855 95.805 -60.004 53.065
528856 95.840 -60.010 53.057
528857 95.874 -60.017 53.050
528858 95.909 -60.024 53.036
528859 95.945 -60.029 53.019
528860 95.983 -60.034 52.999
528861 96.021 -60.039 52.976
528862 96.058 -60.044 52.949
528863 96.096 -60.049 52.920
528864 96.133 -60.054 52.888
528865 96.171 -60.054 52.854
528866 96.209 -60.053 52.821
528867 96.247 -60.052 52.791
528868 96.286 -60.054 52.763
528869 96.323 -60.054 52.735
528870 96.360 -60.054 52.706
528871 96.396 -60.054 52.676
528872 96.433 -60.055 52.646
528873 96.469 -60.055 52.614
528874 96.514 -60.060 52.580
528875 96.559 -60.066 52.544
528876 96.605 -60.072 52.505
528877 96.651 -60.079 52.466
528878 96.696 -60.085 52.423
528879 96.742 -60.092 52.378
528880 96.788 -60.098 52.333
528881 96.829 -60.101 52.290
528882 96.868 -60.102 52.247
528883 96.900 -60.100 52.204
528884 96.934 -60.102 52.162
528885 96.967 -60.105 52.119
528886 97.000 -60.109 52.078
528887 97.030 -60.122 52.038
528888 97.061 -60.135 51.996
528889 97.091 -60.148 51.954
528890 97.121 -60.161 51.910
528891 97.151 -60.173 51.867
528892 97.181 -60.186 51.824
528893 97.212 -60.199 51.781
528894 97.242 -60.212 51.738
528895 97.276 -60.218 51.695
528896 97.313 -60.218 51.653
528897 97.350 -60.219 51.612
528898 97.386 -60.219 51.572
528899 97.423 -60.220 51.533
528900 97.460 -60.221 51.497
528901 97.497 -60.221 51.461
528902 97.533 -60.222 51.427
528903 97.570 -60.223 51.393
528904 97.609 -60.230 51.361
528905 97.651 -60.241 51.329
528906 97.691 -60.252 51.293
528907 97.732 -60.263 51.257
528908 97.773 -60.275 51.219
528909 97.814 -60.286 51.181
528910 97.848 -60.293 51.143
528911 97.870 -60.294 51.104
528912 97.893 -60.295 51.065
528913 97.915 -60.297 51.025
528914 97.938 -60.297 50.985
528915 97.960 -60.298 50.945
528916 97.983 -60.300 50.906
528917 98.005 -60.300 50.866
528918 98.025 -60.301 50.828
528919 98.045 -60.301 50.791
528920 98.084 -60.309 50.751
528921 98.124 -60.317 50.710
528922 98.170 -60.302 50.668
528923 98.216 -60.285 50.623
528924 98.255 -60.284 50.575
528925 98.292 -60.288 50.527
528926 98.328 -60.293 50.479
528927 98.364 -60.297 50.432
528928 98.401 -60.301 50.385
528929 98.437 -60.306 50.339
528930 98.473 -60.310 50.294
528931 98.509 -60.315 50.248
528932 98.546 -60.318 50.204
528933 98.582 -60.323 50.159
528934 98.618 -60.328 50.115
528935 98.654 -60.331 50.071
528936 98.691 -60.336 50.027
528937 98.727 -60.340 49.984
528938 98.764 -60.345 49.942
529345 87.699 -57.639 57.007
529346 87.743 -57.650 56.987
529347 87.792 -57.656 56.968
529348 87.837 -57.663 56.948
529349 87.881 -57.670 56.927
529350 87.926 -57.677 56.904
529351 87.970 -57.685 56.880
529352 88.015 -57.692 56.856
529353 88.059 -57.699 56.832
529354 88.104 -57.707 56.808
529355 88.148 -57.714 56.781
529356 88.193 -57.721 56.755
529357 88.229 -57.732 56.728
529358 88.263 -57.743 56.701
529359 88.297 -57.755 56.673
529360 88.331 -57.767 56.645
529361 88.364 -57.778 56.616
529362 88.398 -57.789 56.587
529363 88.431 -57.801 56.558
529364 88.465 -57.812 56.529
529365 88.499 -57.824 56.501
529366 88.533 -57.836 56.471
529367 88.566 -57.847 56.439
529368 88.600 -57.858 56.406
529369 88.634 -57.870 56.375
529370 88.665 -57.877 56.344
529371 88.695 -57.883 56.313
529372 88.725 -57.889 56.282
529373 88.756 -57.895 56.250
529374 88.782 -57.912 56.218
529375 88.811 -57.936 56.187
529376 88.855 -57.956 56.155
529377 88.898 -57.977 56.122
529378 88.942 -57.998 56.089
529379 88.985 -58.019 56.057
529380 89.029 -58.040 56.026
529381 89.072 -58.061 55.999
529382 89.115 -58.081 55.974
529383 89.159 -58.102 55.950
529384 89.202 -58.123 55.926
529385 89.246 -58.144 55.905
529386 89.289 -58.165 55.886
529387 89.333 -58.185 55.872
529388 89.376 -58.206 55.859
529389 89.419 -58.227 55.847
529390 89.468 -58.244 55.837
529391 89.516 -58.260 55.829
529392 89.556 -58.268 55.823
529393 89.602 -58.272 55.818
529394 89.647 -58.277 55.812
529395 89.692 -58.281 55.807
529396 89.738 -58.286 55.802
529397 89.782 -58.291 55.797
529398 89.828 -58.295 55.793
529399 89.873 -58.300 55.790
529400 89.918 -58.305 55.787
529401 89.964 -58.310 55.785
529402 90.000 -58.322 55.784
529403 90.033 -58.340 55.774
529404 90.065 -58.357 55.760
529405 90.097 -58.375 55.741
529406 90.130 -58.390 55.720
529407 90.167 -58.398 55.695
529408 90.205 -58.406 55.672
529409 90.243 -58.413 55.650
529410 90.281 -58.420 55.630
529411 90.319 -58.428 55.615
529412 90.357 -58.435 55.606
529413 90.394 -58.442 55.601
529414 90.432 -58.450 55.592
529415 90.470 -58.457 55.580
529416 90.508 -58.464 55.561
529417 90.546 -58.472 55.536
529418 90.585 -58.484 55.508
529419 90.626 -58.502 55.480
529420 90.668 -58.521 55.450
529421 90.710 -58.540 55.420
529422 90.751 -58.558 55.394
529423 90.792 -58.577 55.377
529424 90.834 -58.596 55.360
529425 90.875 -58.615 55.340
529426 90.917 -58.633 55.318
529427 90.958 -58.652 55.293
529428 90.997 -58.660 55.264
529429 91.035 -58.665 55.239
529430 91.073 -58.673 55.217
529431 91.107 -58.693 55.192
529432 91.142 -58.714 55.166
529433 91.177 -58.735 55.141
529434 91.215 -58.747 55.122
529435 91.252 -58.760 55.102
529436 91.290 -58.772 55.077
529437 91.328 -58.785 55.048
529438 91.366 -58.797 55.016
529439 91.403 -58.810 54.985
529440 91.442 -58.823 54.966
529441 91.487 -58.835 54.944
529442 91.532 -58.848 54.918
529443 91.577 -58.861 54.890
529444 91.622 -58.874 54.862
529445 91.659 -58.888 54.831
529446 91.695 -58.902 54.807
529447 91.731 -58.917 54.787
529448 91.768 -58.931 54.770
529449 91.803 -58.946 54.758
529450 91.840 -58.960 54.749
529451 91.876 -58.974 54.745
529452 91.912 -58.989 54.743
529453 91.948 -59.003 54.746
529454 91.984 -59.007 54.752
529455 92.019 -59.008 54.761
529456 92.055 -59.008 54.775
529457 92.090 -59.008 54.788
529458 92.126 -59.008 54.801
529459 92.161 -59.009 54.809
529460 92.196 -59.009 54.810
529461 92.232 -59.010 54.804
529462 92.263 -59.014 54.791
529463 92.290 -59.019 54.781
529464 92.317 -59.024 54.771
529465 92.345 -59.030 54.760
529466 92.373 -59.035 54.748
529467 92.401 -59.040 54.740
529468 92.440 -59.043 54.737
529469 92.488 -59.041 54.730
529470 92.535 -59.040 54.721
529471 92.583 -59.040 54.711
529472 92.630 -59.039 54.700
529473 92.677 -59.038 54.683
529474 92.725 -59.037 54.667
529475 92.772 -59.037 54.650
529476 92.819 -59.036 54.632
529477 92.867 -59.035 54.614
529478 92.914 -59.034 54.591
529479 92.961 -59.034 54.563
529480 93.007 -59.036 54.537
529481 93.050 -59.046 54.514
529482 93.090 -59.054 54.496
529483 93.128 -59.063 54.475
529484 93.167 -59.071 54.455
529485 93.205 -59.080 54.438
529486 93.243 -59.088 54.422
529487 93.282 -59.096 54.407
529488 93.321 -59.106 54.394
529489 93.359 -59.114 54.381
529490 93.398 -59.122 54.369
529491 93.436 -59.131 54.355
529492 93.475 -59.139 54.342
529493 93.511 -59.157 54.327
529494 93.547 -59.178 54.314
529495 93.584 -59.194 54.300
529496 93.620 -59.210 54.286
529497 93.657 -59.225 54.270
529498 93.693 -59.241 54.254
529499 93.730 -59.256 54.238
529500 93.766 -59.272 54.221
529501 93.802 -59.289 54.207
529502 93.837 -59.309 54.194
529503 93.872 -59.330 54.179
529504 93.907 -59.350 54.164
529505 93.942 -59.370 54.152
529506 93.977 -59.391 54.140
529507 94.012 -59.411 54.127
529508 94.047 -59.431 54.111
529509 94.082 -59.451 54.091
529510 94.117 -59.472 54.072
529511 94.152 -59.492 54.054
529512 94.187 -59.512 54.038
529513 94.222 -59.532 54.024
529514 94.257 -59.553 54.012
529515 94.292 -59.573 54.003
529516 94.327 -59.593 53.997
529517 94.362 -59.613 53.994
529518 94.397 -59.634 53.991
529519 94.431 -59.649 53.988
529520 94.466 -59.657 53.983
529521 94.499 -59.663 53.975
529522 94.536 -59.673 53.963
529523 94.574 -59.683 53.948
529524 94.612 -59.694 53.931
529525 94.650 -59.704 53.910
529526 94.689 -59.715 53.886
529527 94.727 -59.726 53.858
529528 94.767 -59.736 53.826
529529 94.808 -59.746 53.792
529530 94.848 -59.757 53.756
529531 94.889 -59.767 53.718
529532 94.930 -59.777 53.687
529533 94.970 -59.788 53.662
529534 95.011 -59.798 53.640
529535 95.052 -59.808 53.620
529536 95.092 -59.819 53.602
529537 95.132 -59.830 53.580
529538 95.174 -59.839 53.556
529539 95.214 -59.848 53.530
529540 95.254 -59.857 53.500
529541 95.294 -59.867 53.469
529542 95.334 -59.876 53.437
529543 95.374 -59.885 53.402
529544 95.410 -59.892 53.370
529545 95.443 -59.898 53.337
529546 95.475 -59.905 53.303
529547 95.508 -59.911 53.266
529548 95.540 -59.917 53.236
529549 95.573 -59.924 53.209
529550 95.606 -59.930 53.185
529551 95.639 -59.937 53.161
529552 95.674 -59.943 53.137
529553 95.708 -59.950 53.110
529554 95.743 -59.957 53.094
529555 95.777 -59.963 53.085
529556 95.812 -59.970 53.077
529557 95.846 -59.977 53.069
529558 95.881 -59.983 53.061
529559 95.915 -59.989 53.047
529560 95.952 -59.995 53.030
529561 95.990 -60.000 53.009
529562 96.028 -60.005 52.985
529563 96.065 -60.010 52.957
529564 96.103 -60.015 52.927
529565 96.141 -60.020 52.893
529566 96.178 -60.025 52.859
529567 96.216 -60.025 52.829
529568 96.254 -60.025 52.802
529569 96.292 -60.026 52.776
529570 96.328 -60.026 52.749
529571 96.365 -60.026 52.722
529572 96.401 -60.026 52.694
529573 96.438 -60.026 52.666
529574 96.474 -60.026 52.635
529575 96.520 -60.032 52.602
529576 96.566 -60.039 52.567
529577 96.611 -60.045 52.528
529578 96.657 -60.052 52.488
529579 96.703 -60.058 52.446
529580 96.749 -60.064 52.402
529581 96.791 -60.068 52.360
529582 96.829 -60.068 52.317
529583 96.865 -60.068 52.274
529584 96.896 -60.067 52.232
529585 96.930 -60.068 52.189
529586 96.964 -60.071 52.146
529587 96.998 -60.073 52.105
529588 97.029 -60.080 52.063
529589 97.060 -60.093 52.019
529590 97.090 -60.106 51.976
529591 97.120 -60.119 51.930
529592 97.150 -60.131 51.887
529593 97.180 -60.144 51.843
529594 97.211 -60.157 51.799
529595 97.241 -60.170 51.756
529596 97.274 -60.178 51.713
529597 97.310 -60.179 51.669
529598 97.347 -60.180 51.626
529599 97.384 -60.181 51.585
529600 97.421 -60.181 51.545
529601 97.458 -60.182 51.506
529602 97.495 -60.183 51.469
529603 97.531 -60.184 51.435
529604 97.568 -60.184 51.401
529605 97.609 -60.194 51.368
529606 97.650 -60.206 51.334
529607 97.690 -60.217 51.300
529608 97.732 -60.228 51.263
529609 97.773 -60.239 51.225
529610 97.813 -60.250 51.186
529611 97.854 -60.262 51.147
529612 97.887 -60.268 51.108
529613 97.910 -60.269 51.068
529614 97.932 -60.270 51.029
529615 97.954 -60.271 50.988
529616 97.977 -60.272 50.947
529617 97.999 -60.274 50.906
529618 98.021 -60.275 50.866
529619 98.040 -60.275 50.827
529620 98.057 -60.273 50.791
529621 98.086 -60.277 50.753
529622 98.126 -60.285 50.713
529623 98.169 -60.280 50.670
529624 98.215 -60.263 50.626
529625 98.262 -60.246 50.580
529626 98.298 -60.251 50.534
529627 98.334 -60.256 50.487
529628 98.370 -60.259 50.439
529629 98.407 -60.264 50.391
529630 98.443 -60.268 50.345
529631 98.479 -60.273 50.300
529632 98.515 -60.277 50.255
529633 98.552 -60.281 50.210
529634 98.588 -60.286 50.167
529635 98.625 -60.290 50.122
529636 98.661 -60.294 50.077
529637 98.697 -60.299 50.033
529638 98.733 -60.303 49.990
529639 98.770 -60.307 49.949
530047 87.768 -57.601 56.974
530048 87.812 -57.609 56.954
530049 87.857 -57.616 56.934
530050 87.901 -57.623 56.913
530051 87.946 -57.630 56.890
530052 87.990 -57.638 56.867
530053 88.035 -57.645 56.844
530054 88.079 -57.652 56.820
530055 88.124 -57.660 56.796
530056 88.169 -57.667 56.770
530057 88.208 -57.677 56.744
530058 88.242 -57.688 56.718
530059 88.275 -57.699 56.691
530060 88.309 -57.711 56.662
530061 88.343 -57.723 56.634
530062 88.377 -57.734 56.606
530063 88.410 -57.746 56.578
530064 88.444 -57.757 56.550
530065 88.478 -57.768 56.522
530066 88.511 -57.780 56.495
530067 88.545 -57.792 56.466
530068 88.578 -57.803 56.435
530069 88.612 -57.814 56.405
530070 88.646 -57.826 56.376
530071 88.680 -57.838 56.345
530072 88.713 -57.848 56.316
530073 88.743 -57.854 56.287
530074 88.773 -57.859 56.258
530075 88.801 -57.874 56.227
530076 88.825 -57.898 56.197
530077 88.854 -57.921 56.166
530078 88.898 -57.942 56.133
530079 88.941 -57.962 56.100
530080 88.985 -57.983 56.068
530081 89.028 -58.004 56.036
530082 89.071 -58.024 56.009
530083 89.115 -58.046 55.983
530084 89.159 -58.066 55.958
530085 89.202 -58.087 55.934
530086 89.245 -58.108 55.911
530087 89.289 -58.128 55.893
530088 89.334 -58.148 55.877
530089 89.383 -58.165 55.863
530090 89.432 -58.182 55.850
530091 89.481 -58.199 55.839
530092 89.530 -58.216 55.830
530093 89.568 -58.227 55.821
530094 89.607 -58.235 55.815
530095 89.653 -58.240 55.809
530096 89.698 -58.244 55.804
530097 89.743 -58.249 55.798
530098 89.789 -58.253 55.795
530099 89.834 -58.258 55.790
530100 89.879 -58.263 55.787
530101 89.924 -58.267 55.785
530102 89.969 -58.272 55.782
530103 90.011 -58.280 55.780
530104 90.043 -58.297 55.769
530105 90.076 -58.315 55.753
530106 90.108 -58.332 55.733
530107 90.141 -58.349 55.711
530108 90.180 -58.356 55.686
530109 90.218 -58.363 55.663
530110 90.256 -58.370 55.643
530111 90.294 -58.378 55.626
530112 90.331 -58.385 55.614
530113 90.369 -58.392 55.607
530114 90.407 -58.400 55.601
530115 90.445 -58.407 55.590
530116 90.483 -58.414 55.573
530117 90.521 -58.422 55.549
530118 90.558 -58.429 55.520
530119 90.597 -58.436 55.490
530120 90.637 -58.453 55.463
530121 90.679 -58.471 55.437
530122 90.720 -58.490 55.413
530123 90.761 -58.509 55.393
530124 90.803 -58.527 55.378
530125 90.844 -58.546 55.363
530126 90.886 -58.565 55.346
530127 90.927 -58.583 55.326
530128 90.968 -58.602 55.302
530129 91.010 -58.621 55.274
530130 91.051 -58.639 55.248
530131 91.089 -58.643 55.222
530132 91.126 -58.658 55.195
530133 91.160 -58.679 55.167
530134 91.197 -58.692 55.141
530135 91.235 -58.705 55.119
530136 91.273 -58.717 55.098
530137 91.310 -58.730 55.074
530138 91.348 -58.742 55.051
530139 91.386 -58.755 55.028
530140 91.424 -58.768 55.007
530141 91.461 -58.780 54.988
530142 91.500 -58.793 54.966
530143 91.544 -58.806 54.939
530144 91.589 -58.819 54.913
530145 91.632 -58.832 54.887
530146 91.667 -58.846 54.854
530147 91.704 -58.861 54.824
530148 91.740 -58.875 54.798
530149 91.776 -58.889 54.777
530150 91.812 -58.904 54.761
530151 91.848 -58.918 54.750
530152 91.884 -58.933 54.743
530153 91.920 -58.947 54.738
530154 91.956 -58.961 54.738
530155 91.992 -58.968 54.742
530156 92.027 -58.969 54.750
530157 92.063 -58.969 54.763
530158 92.098 -58.970 54.782
530159 92.134 -58.970 54.800
530160 92.169 -58.971 54.812
530161 92.205 -58.971 54.815
530162 92.240 -58.971 54.811
530163 92.276 -58.972 54.798
530164 92.305 -58.976 54.782
530165 92.333 -58.982 54.767
530166 92.360 -58.987 54.755
530167 92.387 -58.993 54.743
530168 92.415 -58.998 54.731
530169 92.455 -58.999 54.724
530170 92.502 -58.999 54.723
530171 92.550 -58.998 54.725
530172 92.597 -58.997 54.722
530173 92.645 -58.996 54.710
530174 92.692 -58.996 54.694
530175 92.739 -58.995 54.677
530176 92.787 -58.994 54.660
530177 92.834 -58.993 54.642
530178 92.881 -58.993 54.619
530179 92.929 -58.992 54.599
530180 92.976 -58.991 54.571
530181 93.023 -58.990 54.543
530182 93.066 -58.998 54.520
530183 93.105 -59.006 54.502
530184 93.143 -59.015 54.484
530185 93.182 -59.023 54.466
530186 93.220 -59.032 54.449
530187 93.259 -59.040 54.434
530188 93.297 -59.049 54.421
530189 93.336 -59.057 54.409
530190 93.374 -59.065 54.397
530191 93.413 -59.074 54.385
530192 93.451 -59.083 54.374
530193 93.489 -59.091 54.360
530194 93.526 -59.109 54.345
530195 93.563 -59.126 54.327
530196 93.599 -59.141 54.308
530197 93.636 -59.157 54.289
530198 93.672 -59.174 54.272
530199 93.707 -59.194 54.255
530200 93.742 -59.214 54.237
530201 93.777 -59.234 54.219
530202 93.812 -59.255 54.203
530203 93.847 -59.275 54.188
530204 93.882 -59.295 54.172
530205 93.917 -59.316 54.157
530206 93.952 -59.336 54.136
530207 93.987 -59.356 54.118
530208 94.022 -59.376 54.103
530209 94.057 -59.397 54.086
530210 94.091 -59.417 54.067
530211 94.126 -59.438 54.049
530212 94.161 -59.457 54.032
530213 94.196 -59.478 54.016
530214 94.231 -59.498 54.001
530215 94.266 -59.519 53.988
530216 94.301 -59.538 53.976
530217 94.336 -59.559 53.969
530218 94.371 -59.579 53.965
530219 94.406 -59.599 53.963
530220 94.441 -59.613 53.960
530221 94.475 -59.620 53.956
530222 94.509 -59.627 53.945
530223 94.543 -59.635 53.930
530224 94.581 -59.645 53.915
530225 94.620 -59.656 53.898
530226 94.658 -59.667 53.879
530227 94.696 -59.677 53.857
530228 94.735 -59.687 53.831
530229 94.776 -59.697 53.801
530230 94.817 -59.707 53.767
530231 94.857 -59.717 53.731
530232 94.898 -59.727 53.694
530233 94.938 -59.738 53.664
530234 94.979 -59.748 53.639
530235 95.020 -59.758 53.619
530236 95.060 -59.769 53.601
530237 95.101 -59.779 53.584
530238 95.141 -59.792 53.564
530239 95.181 -59.802 53.542
530240 95.222 -59.811 53.519
530241 95.263 -59.820 53.493
530242 95.304 -59.829 53.464
530243 95.344 -59.838 53.435
530244 95.384 -59.847 53.404
530245 95.424 -59.856 53.373
530246 95.457 -59.862 53.342
530247 95.490 -59.869 53.309
530248 95.522 -59.875 53.277
530249 95.555 -59.881 53.247
530250 95.585 -59.889 53.219
530251 95.614 -59.896 53.192
530252 95.647 -59.903 53.164
530253 95.681 -59.910 53.138
530254 95.715 -59.917 53.118
530255 95.750 -59.923 53.107
530256 95.784 -59.930 53.099
530257 95.819 -59.936 53.090
530258 95.853 -59.942 53.080
530259 95.888 -59.949 53.069
530260 95.922 -59.955 53.056
530261 95.959 -59.961 53.037
530262 95.997 -59.966 53.014
530263 96.035 -59.971 52.989
530264 96.073 -59.976 52.961
530265 96.110 -59.981 52.929
530266 96.148 -59.986 52.893
530267 96.185 -59.991 52.862
530268 96.223 -59.996 52.836
530269 96.260 -59.997 52.813
530270 96.297 -59.997 52.786
530271 96.333 -59.997 52.761
530272 96.370 -59.997 52.735
530273 96.406 -59.997 52.710
530274 96.443 -59.997 52.683
530275 96.481 -59.999 52.655
530276 96.527 -60.005 52.623
530277 96.572 -60.011 52.588
530278 96.618 -60.018 52.549
530279 96.664 -60.024 52.508
530280 96.710 -60.030 52.467
530281 96.751 -60.033 52.425
530282 96.790 -60.034 52.385
530283 96.828 -60.035 52.343
530284 96.861 -60.034 52.301
530285 96.893 -60.033 52.259
530286 96.927 -60.035 52.217
530287 96.960 -60.037 52.175
530288 96.994 -60.040 52.133
530289 97.028 -60.042 52.089
530290 97.059 -60.052 52.044
530291 97.089 -60.065 51.999
530292 97.119 -60.077 51.953
530293 97.150 -60.090 51.909
530294 97.180 -60.102 51.864
530295 97.210 -60.115 51.819
530296 97.240 -60.128 51.774
530297 97.272 -60.140 51.730
530298 97.308 -60.140 51.685
530299 97.345 -60.141 51.641
530300 97.382 -60.142 51.599
530301 97.419 -60.143 51.557
530302 97.455 -60.143 51.517
530303 97.492 -60.143 51.479
530304 97.529 -60.144 51.442
530305 97.567 -60.148 51.406
530306 97.608 -60.159 51.371
530307 97.649 -60.170 51.336
530308 97.690 -60.181 51.302
530309 97.731 -60.193 51.264
530310 97.772 -60.204 51.226
530311 97.813 -60.215 51.187
530312 97.854 -60.226 51.148
530313 97.894 -60.237 51.110
530314 97.926 -60.243 51.070
530315 97.948 -60.244 51.031
530316 97.970 -60.245 50.990
530317 97.993 -60.246 50.948
530318 98.015 -60.247 50.907
530319 98.038 -60.249 50.866
530320 98.056 -60.248 50.825
530321 98.072 -60.246 50.787
530322 98.088 -60.246 50.749
530323 98.127 -60.253 50.709
530324 98.167 -60.258 50.667
530325 98.214 -60.241 50.624
530326 98.261 -60.224 50.580
530327 98.304 -60.214 50.535
530328 98.340 -60.218 50.490
530329 98.376 -60.222 50.444
530330 98.413 -60.227 50.397
530331 98.449 -60.231 50.351
530332 98.486 -60.235 50.305
530333 98.522 -60.240 50.261
530334 98.558 -60.244 50.217
530335 98.594 -60.248 50.173
530336 98.631 -60.253 50.129
530337 98.667 -60.257 50.084
530338 98.703 -60.261 50.040
530339 98.739 -60.265 49.997
530340 98.776 -60.270 49.956
530749 87.832 -57.561 56.941
530750 87.877 -57.569 56.921
530751 87.921 -57.576 56.899
530752 87.966 -57.583 56.877
530753 88.010 -57.591 56.855
530754 88.055 -57.598 56.832
530755 88.100 -57.605 56.809
530756 88.144 -57.613 56.783
530757 88.187 -57.620 56.759
530758 88.220 -57.633 56.733
530759 88.254 -57.644 56.707
530760 88.287 -57.655 56.680
530761 88.321 -57.667 56.651
530762 88.355 -57.679 56.622
530763 88.389 -57.690 56.595
530764 88.422 -57.702 56.568
530765 88.456 -57.713 56.540
530766 88.490 -57.724 56.513
530767 88.524 -57.736 56.486
530768 88.557 -57.748 56.458
530769 88.591 -57.759 56.429
530770 88.624 -57.771 56.400
530771 88.658 -57.782 56.372
530772 88.692 -57.794 56.343
530773 88.725 -57.805 56.316
530774 88.759 -57.817 56.288
530775 88.790 -57.824 56.261
530776 88.820 -57.834 56.233
530777 88.844 -57.858 56.206
530778 88.868 -57.883 56.175
530779 88.897 -57.905 56.143
530780 88.941 -57.927 56.110
530781 88.984 -57.947 56.077
530782 89.028 -57.968 56.046
530783 89.071 -57.989 56.018
530784 89.115 -58.009 55.990
530785 89.158 -58.031 55.965
530786 89.201 -58.051 55.941
530787 89.249 -58.069 55.917
530788 89.298 -58.087 55.899
530789 89.347 -58.103 55.882
530790 89.396 -58.120 55.867
530791 89.445 -58.137 55.854
530792 89.494 -58.154 55.842
530793 89.543 -58.171 55.831
530794 89.584 -58.184 55.822
530795 89.620 -58.194 55.815
530796 89.659 -58.203 55.808
530797 89.704 -58.207 55.802
530798 89.749 -58.212 55.797
530799 89.794 -58.216 55.793
530800 89.839 -58.221 55.788
530801 89.885 -58.225 55.784
530802 89.930 -58.230 55.782
530803 89.975 -58.234 55.779
530804 90.020 -58.239 55.774
530805 90.054 -58.256 55.763
530806 90.086 -58.272 55.745
530807 90.118 -58.290 55.725
530808 90.151 -58.307 55.702
530809 90.186 -58.318 55.678
530810 90.228 -58.322 55.657
530811 90.268 -58.328 55.637
530812 90.306 -58.335 55.621
530813 90.344 -58.343 55.609
530814 90.382 -58.350 55.601
530815 90.420 -58.357 55.591
530816 90.458 -58.365 55.578
530817 90.496 -58.372 55.558
530818 90.534 -58.379 55.532
530819 90.572 -58.387 55.502
530820 90.609 -58.394 55.472
530821 90.648 -58.403 55.450
530822 90.689 -58.421 55.429
530823 90.731 -58.440 55.411
530824 90.772 -58.459 55.398
530825 90.813 -58.478 55.382
530826 90.855 -58.496 55.368
530827 90.896 -58.515 55.351
530828 90.938 -58.534 55.331
530829 90.979 -58.552 55.309
530830 91.020 -58.571 55.282
530831 91.062 -58.590 55.255
530832 91.103 -58.608 55.227
530833 91.143 -58.623 55.199
530834 91.179 -58.638 55.171
530835 91.218 -58.651 55.145
530836 91.255 -58.663 55.120
530837 91.293 -58.676 55.099
530838 91.331 -58.689 55.078
530839 91.369 -58.701 55.059
530840 91.406 -58.714 55.042
530841 91.444 -58.726 55.025
530842 91.482 -58.739 55.007
530843 91.520 -58.751 54.987
530844 91.558 -58.764 54.962
530845 91.601 -58.777 54.935
530846 91.640 -58.790 54.910
530847 91.676 -58.805 54.879
530848 91.712 -58.819 54.846
530849 91.748 -58.833 54.815
530850 91.784 -58.848 54.789
530851 91.820 -58.862 54.768
530852 91.856 -58.877 54.754
530853 91.892 -58.891 54.743
530854 91.929 -58.905 54.737
530855 91.964 -58.920 54.735
530856 92.001 -58.930 54.737
530857 92.036 -58.930 54.744
530858 92.071 -58.930 54.755
530859 92.107 -58.931 54.777
530860 92.142 -58.931 54.796
530861 92.178 -58.932 54.809
530862 92.213 -58.932 54.813
530863 92.249 -58.933 54.808
530864 92.284 -58.933 54.794
530865 92.319 -58.933 54.775
530866 92.347 -58.939 54.756
530867 92.375 -58.945 54.742
530868 92.403 -58.950 54.732
530869 92.430 -58.955 54.718
530870 92.470 -58.957 54.708
530871 92.517 -58.956 54.703
530872 92.565 -58.955 54.704
530873 92.612 -58.955 54.702
530874 92.659 -58.954 54.693
530875 92.707 -58.953 54.681
530876 92.754 -58.952 54.669
530877 92.801 -58.952 54.656
530878 92.849 -58.951 54.641
530879 92.896 -58.950 54.622
530880 92.943 -58.949 54.604
530881 92.991 -58.949 54.587
530882 93.038 -58.948 54.570
530883 93.080 -58.955 54.548
530884 93.119 -58.963 54.525
530885 93.157 -58.971 54.504
530886 93.196 -58.978 54.486
530887 93.234 -58.986 54.468
530888 93.273 -58.993 54.451
530889 93.312 -59.001 54.439
530890 93.351 -59.009 54.426
530891 93.389 -59.018 54.415
530892 93.428 -59.026 54.405
530893 93.466 -59.034 54.395
530894 93.504 -59.043 54.383
530895 93.541 -59.058 54.365
530896 93.576 -59.078 54.345
530897 93.611 -59.099 54.321
530898 93.646 -59.119 54.298
530899 93.681 -59.139 54.277
530900 93.716 -59.159 54.259
530901 93.751 -59.180 54.239
530902 93.786 -59.200 54.220
530903 93.821 -59.220 54.201
530904 93.856 -59.240 54.183
530905 93.891 -59.261 54.165
530906 93.926 -59.281 54.147
530907 93.961 -59.301 54.125
530908 93.996 -59.322 54.105
530909 94.031 -59.342 54.087
530910 94.066 -59.363 54.071
530911 94.101 -59.382 54.052
530912 94.136 -59.403 54.032
530913 94.171 -59.423 54.014
530914 94.206 -59.444 53.996
530915 94.241 -59.463 53.980
530916 94.276 -59.484 53.967
530917 94.311 -59.504 53.953
530918 94.346 -59.525 53.944
530919 94.381 -59.544 53.936
530920 94.416 -59.565 53.931
530921 94.450 -59.576 53.928
530922 94.484 -59.583 53.925
530923 94.518 -59.591 53.914
530924 94.552 -59.598 53.898
530925 94.589 -59.607 53.881
530926 94.627 -59.618 53.862
530927 94.666 -59.629 53.842
530928 94.704 -59.639 53.821
530929 94.744 -59.649 53.796
530930 94.784 -59.658 53.767
530931 94.825 -59.667 53.735
530932 94.866 -59.677 53.701
530933 94.906 -59.688 53.666
530934 94.947 -59.698 53.636
530935 94.988 -59.708 53.612
530936 95.028 -59.719 53.592
530937 95.069 -59.729 53.575
530938 95.109 -59.740 53.558
530939 95.149 -59.752 53.541
530940 95.189 -59.765 53.522
530941 95.230 -59.774 53.502
530942 95.271 -59.783 53.480
530943 95.312 -59.792 53.456
530944 95.353 -59.801 53.429
530945 95.394 -59.809 53.402
530946 95.435 -59.818 53.373
530947 95.472 -59.826 53.344
530948 95.504 -59.833 53.313
530949 95.534 -59.840 53.283
530950 95.563 -59.848 53.252
530951 95.592 -59.855 53.222
530952 95.621 -59.863 53.193
530953 95.654 -59.870 53.166
530954 95.688 -59.877 53.145
530955 95.723 -59.883 53.131
530956 95.757 -59.889 53.122
530957 95.792 -59.896 53.113
530958 95.826 -59.902 53.102
530959 95.861 -59.909 53.091
530960 95.895 -59.915 53.079
530961 95.930 -59.922 53.063
530962 95.966 -59.927 53.042
530963 96.004 -59.933 53.018
530964 96.041 -59.938 52.990
530965 96.079 -59.942 52.961
530966 96.116 -59.948 52.929
530967 96.154 -59.953 52.897
530968 96.191 -59.958 52.870
530969 96.229 -59.963 52.846
530970 96.265 -59.968 52.823
530971 96.302 -59.968 52.798
530972 96.338 -59.968 52.772
530973 96.375 -59.969 52.747
530974 96.411 -59.969 52.723
530975 96.448 -59.969 52.698
530976 96.487 -59.971 52.670
530977 96.533 -59.977 52.639
530978 96.579 -59.983 52.605
530979 96.624 -59.990 52.568
530980 96.670 -59.996 52.527
530981 96.713 -60.000 52.486
530982 96.751 -60.001 52.446
530983 96.789 -60.002 52.407
530984 96.825 -60.002 52.368
530985 96.857 -60.000 52.328
530986 96.889 -59.999 52.287
530987 96.923 -60.001 52.245
530988 96.957 -60.004 52.206
530989 96.991 -60.006 52.163
530990 97.024 -60.008 52.118
530991 97.058 -60.011 52.071
530992 97.088 -60.023 52.025
530993 97.118 -60.036 51.979
530994 97.149 -60.049 51.933
530995 97.179 -60.061 51.886
530996 97.210 -60.074 51.839
530997 97.240 -60.086 51.793
530998 97.270 -60.099 51.747
530999 97.306 -60.101 51.701
531000 97.343 -60.102 51.657
531001 97.380 -60.102 51.613
531002 97.417 -60.103 51.571
531003 97.453 -60.104 51.529
531004 97.490 -60.105 51.489
531005 97.527 -60.105 51.449
531006 97.566 -60.112 51.411
531007 97.607 -60.124 51.373
531008 97.648 -60.135 51.337
531009 97.689 -60.146 51.301
531010 97.730 -60.157 51.264
531011 97.771 -60.168 51.226
531012 97.812 -60.180 51.187
531013 97.853 -60.190 51.149
531014 97.894 -60.202 51.111
531015 97.935 -60.213 51.072
531016 97.965 -60.218 51.033
531017 97.987 -60.219 50.992
531018 98.010 -60.221 50.950
531019 98.032 -60.221 50.908
531020 98.055 -60.223 50.866
531021 98.072 -60.221 50.824
531022 98.087 -60.221 50.784
531023 98.103 -60.219 50.745
531024 98.129 -60.221 50.705
531025 98.168 -60.230 50.664
531026 98.213 -60.219 50.621
531027 98.259 -60.202 50.578
531028 98.306 -60.186 50.534
531029 98.347 -60.181 50.491
531030 98.383 -60.185 50.447
531031 98.419 -60.189 50.402
531032 98.455 -60.193 50.356
531033 98.492 -60.198 50.310
531034 98.528 -60.202 50.267
531035 98.564 -60.206 50.224
531036 98.600 -60.211 50.180
531037 98.637 -60.215 50.135
531038 98.673 -60.219 50.091
531039 98.709 -60.224 50.047
531040 98.745 -60.228 50.004
531041 98.782 -60.232 49.963
531452 87.941 -57.529 56.883
531453 87.986 -57.536 56.861
531454 88.031 -57.544 56.839
531455 88.075 -57.551 56.815
531456 88.120 -57.558 56.793
531457 88.164 -57.566 56.768
531458 88.199 -57.576 56.743
531459 88.233 -57.589 56.717
531460 88.266 -57.600 56.692
531461 88.300 -57.611 56.667
531462 88.333 -57.623 56.640
531463 88.367 -57.634 56.611
531464 88.401 -57.646 56.583
531465 88.434 -57.658 56.556
531466 88.468 -57.669 56.529
531467 88.502 -57.680 56.503
531468 88.536 -57.692 56.476
531469 88.569 -57.704 56.449
531470 88.603 -57.715 56.421
531471 88.637 -57.727 56.393
531472 88.671 -57.738 56.365
531473 88.704 -57.750 56.338
531474 88.738 -57.761 56.311
531475 88.771 -57.773 56.285
531476 88.805 -57.784 56.259
531477 88.838 -57.795 56.234
531478 88.862 -57.819 56.208
531479 88.887 -57.843 56.179
531480 88.911 -57.868 56.149
531481 88.941 -57.890 56.117
531482 88.984 -57.912 56.086
531483 89.027 -57.932 56.055
531484 89.071 -57.953 56.026
531485 89.115 -57.974 55.997
531486 89.164 -57.990 55.971
531487 89.212 -58.008 55.946
531488 89.262 -58.024 55.923
531489 89.311 -58.041 55.905
531490 89.359 -58.059 55.887
531491 89.409 -58.075 55.871
531492 89.458 -58.092 55.857
531493 89.507 -58.109 55.845
531494 89.556 -58.126 55.834
531495 89.600 -58.140 55.824
531496 89.636 -58.150 55.816
531497 89.672 -58.160 55.809
531498 89.709 -58.169 55.804
531499 89.755 -58.174 55.798
531500 89.800 -58.178 55.793
531501 89.845 -58.183 55.787
531502 89.890 -58.187 55.782
531503 89.936 -58.193 55.778
531504 89.981 -58.197 55.773
531505 90.026 -58.202 55.766
531506 90.064 -58.213 55.753
531507 90.097 -58.231 55.734
531508 90.129 -58.248 55.714
531509 90.162 -58.265 55.693
531510 90.195 -58.277 55.672
531511 90.233 -58.285 55.651
531512 90.275 -58.289 55.632
531513 90.318 -58.294 55.616
531514 90.357 -58.300 55.602
531515 90.395 -58.307 55.591
531516 90.433 -58.315 55.577
531517 90.471 -58.322 55.561
531518 90.509 -58.329 55.541
531519 90.546 -58.337 55.515
531520 90.584 -58.344 55.487
531521 90.622 -58.351 55.461
531522 90.660 -58.359 55.441
531523 90.699 -58.372 55.424
531524 90.741 -58.390 55.408
531525 90.783 -58.409 55.393
531526 90.824 -58.428 55.377
531527 90.865 -58.446 55.361
531528 90.906 -58.465 55.343
531529 90.948 -58.484 55.324
531530 90.990 -58.503 55.304
531531 91.031 -58.521 55.280
531532 91.072 -58.540 55.255
531533 91.113 -58.559 55.231
531534 91.154 -58.579 55.205
531535 91.194 -58.601 55.178
531536 91.232 -58.614 55.151
531537 91.270 -58.626 55.124
531538 91.308 -58.639 55.102
531539 91.346 -58.651 55.083
531540 91.384 -58.663 55.066
531541 91.423 -58.675 55.051
531542 91.460 -58.687 55.036
531543 91.499 -58.700 55.019
531544 91.537 -58.712 55.002
531545 91.575 -58.724 54.981
531546 91.612 -58.736 54.957
531547 91.648 -58.748 54.933
531548 91.684 -58.763 54.905
531549 91.720 -58.777 54.870
531550 91.756 -58.792 54.836
531551 91.793 -58.806 54.806
531552 91.828 -58.820 54.782
531553 91.865 -58.835 54.763
531554 91.901 -58.849 54.749
531555 91.937 -58.864 54.738
531556 91.973 -58.879 54.732
531557 92.009 -58.891 54.731
531558 92.044 -58.892 54.735
531559 92.080 -58.892 54.743
531560 92.116 -58.892 54.763
531561 92.151 -58.892 54.782
531562 92.187 -58.893 54.794
531563 92.222 -58.893 54.794
531564 92.258 -58.894 54.789
531565 92.293 -58.894 54.779
531566 92.328 -58.895 54.761
531567 92.362 -58.896 54.739
531568 92.389 -58.902 54.725
531569 92.417 -58.907 54.717
531570 92.445 -58.913 54.705
531571 92.485 -58.914 54.695
531572 92.532 -58.914 54.692
531573 92.579 -58.913 54.694
531574 92.627 -58.912 54.695
531575 92.674 -58.911 54.690
531576 92.721 -58.911 54.684
531577 92.769 -58.910 54.677
531578 92.816 -58.909 54.667
531579 92.863 -58.908 54.653
531580 92.911 -58.907 54.635
531581 92.957 -58.908 54.616
531582 93.003 -58.914 54.600
531583 93.049 -58.921 54.585
531584 93.087 -58.930 54.568
531585 93.126 -58.937 54.548
531586 93.165 -58.945 54.528
531587 93.204 -58.952 54.509
531588 93.242 -58.960 54.492
531589 93.281 -58.968 54.474
531590 93.320 -58.975 54.458
531591 93.359 -58.983 54.443
531592 93.397 -58.991 54.433
531593 93.436 -58.999 54.424
531594 93.475 -59.006 54.415
531595 93.513 -59.014 54.403
531596 93.550 -59.027 54.386
531597 93.586 -59.044 54.366
531598 93.621 -59.064 54.340
531599 93.655 -59.084 54.312
531600 93.691 -59.105 54.287
531601 93.725 -59.125 54.264
531602 93.761 -59.145 54.244
531603 93.795 -59.165 54.222
531604 93.831 -59.186 54.202
531605 93.865 -59.206 54.182
531606 93.901 -59.226 54.161
531607 93.935 -59.247 54.140
531608 93.971 -59.267 54.119
531609 94.005 -59.287 54.099
531610 94.041 -59.307 54.080
531611 94.075 -59.328 54.062
531612 94.110 -59.348 54.042
531613 94.145 -59.369 54.023
531614 94.180 -59.388 54.004
531615 94.215 -59.409 53.986
531616 94.250 -59.429 53.969
531617 94.285 -59.450 53.953
531618 94.320 -59.469 53.937
531619 94.355 -59.490 53.925
531620 94.390 -59.510 53.913
531621 94.425 -59.531 53.904
531622 94.459 -59.540 53.897
531623 94.493 -59.547 53.890
531624 94.527 -59.554 53.876
531625 94.561 -59.561 53.861
531626 94.597 -59.569 53.843
531627 94.635 -59.579 53.822
531628 94.673 -59.590 53.801
531629 94.711 -59.601 53.783
531630 94.752 -59.610 53.759
531631 94.793 -59.620 53.732
531632 94.834 -59.629 53.703
531633 94.875 -59.638 53.672
531634 94.915 -59.648 53.640
531635 94.956 -59.658 53.610
531636 94.996 -59.669 53.583
531637 95.037 -59.679 53.559
531638 95.077 -59.689 53.539
531639 95.118 -59.701 53.525
531640 95.158 -59.714 53.510
531641 95.197 -59.726 53.494
531642 95.238 -59.737 53.478
531643 95.278 -59.746 53.461
531644 95.320 -59.754 53.441
531645 95.361 -59.764 53.418
531646 95.402 -59.773 53.394
531647 95.443 -59.781 53.370
531648 95.482 -59.790 53.343
531649 95.513 -59.798 53.314
531650 95.542 -59.806 53.284
531651 95.570 -59.814 53.252
531652 95.600 -59.822 53.219
531653 95.629 -59.830 53.191
531654 95.661 -59.836 53.169
531655 95.695 -59.843 53.156
531656 95.730 -59.849 53.146
531657 95.764 -59.856 53.137
531658 95.799 -59.862 53.127
531659 95.833 -59.869 53.115
531660 95.868 -59.876 53.102
531661 95.902 -59.882 53.088
531662 95.936 -59.889 53.070
531663 95.971 -59.895 53.047
531664 96.009 -59.900 53.021
531665 96.046 -59.905 52.993
531666 96.084 -59.910 52.963
531667 96.121 -59.915 52.933
531668 96.159 -59.920 52.903
531669 96.196 -59.926 52.878
531670 96.233 -59.931 52.855
531671 96.270 -59.936 52.832
531672 96.307 -59.940 52.806
531673 96.343 -59.940 52.783
531674 96.380 -59.940 52.759
531675 96.416 -59.940 52.735
531676 96.453 -59.940 52.710
531677 96.493 -59.943 52.683
531678 96.539 -59.950 52.652
531679 96.585 -59.956 52.619
531680 96.631 -59.963 52.583
531681 96.673 -59.966 52.543
531682 96.712 -59.967 52.503
531683 96.750 -59.967 52.465
531684 96.789 -59.968 52.428
531685 96.822 -59.967 52.392
531686 96.854 -59.967 52.354
531687 96.886 -59.965 52.313
531688 96.920 -59.967 52.270
531689 96.953 -59.970 52.232
531690 96.987 -59.972 52.190
531691 97.021 -59.974 52.146
531692 97.054 -59.977 52.100
531693 97.088 -59.981 52.052
531694 97.118 -59.994 52.003
531695 97.148 -60.007 51.955
531696 97.178 -60.020 51.907
531697 97.209 -60.032 51.860
531698 97.239 -60.045 51.813
531699 97.269 -60.058 51.765
531700 97.304 -60.062 51.718
531701 97.341 -60.063 51.674
531702 97.377 -60.064 51.630
531703 97.414 -60.065 51.587
531704 97.451 -60.065 51.544
531705 97.488 -60.066 51.502
531706 97.525 -60.067 51.460
531707 97.566 -60.077 51.420
531708 97.606 -60.088 51.380
531709 97.647 -60.099 51.341
531710 97.688 -60.111 51.302
531711 97.729 -60.122 51.263
531712 97.770 -60.133 51.225
531713 97.811 -60.144 51.187
531714 97.852 -60.155 51.149
531715 97.893 -60.167 51.111
531716 97.934 -60.177 51.073
531717 97.975 -60.189 51.034
531718 98.004 -60.193 50.994
531719 98.026 -60.194 50.953
531720 98.049 -60.196 50.911
531721 98.071 -60.196 50.868
531722 98.087 -60.196 50.825
531723 98.103 -60.194 50.783
531724 98.119 -60.193 50.743
531725 98.135 -60.191 50.703
531726 98.169 -60.198 50.661
531727 98.212 -60.197 50.618
531728 98.258 -60.181 50.575
531729 98.304 -60.164 50.532
531730 98.351 -60.146 50.490
531731 98.389 -60.147 50.449
531732 98.425 -60.152 50.405
531733 98.461 -60.156 50.360
531734 98.498 -60.161 50.316
531735 98.534 -60.165 50.273
531736 98.570 -60.169 50.230
531737 98.606 -60.174 50.187
531738 98.643 -60.177 50.142
531739 98.679 -60.182 50.098
531740 98.716 -60.187 50.054
531741 98.752 -60.191 50.011
531742 98.788 -60.191 49.969
531743 98.820 -60.168 49.927
532163 88.346 -57.579 56.626
532164 88.380 -57.590 56.600
532165 88.413 -57.602 56.572
532166 88.447 -57.614 56.543
532167 88.480 -57.625 56.517
532168 88.514 -57.636 56.491
532169 88.548 -57.648 56.464
532170 88.581 -57.660 56.438
532171 88.615 -57.671 56.411
532172 88.649 -57.683 56.384
532173 88.683 -57.694 56.357
532174 88.716 -57.706 56.331
532175 88.750 -57.717 56.306
532176 88.784 -57.729 56.280
532177 88.818 -57.740 56.256
532178 88.856 -57.757 56.232
532179 88.881 -57.780 56.207
532180 88.905 -57.804 56.180
532181 88.930 -57.828 56.152
532182 88.954 -57.852 56.122
532183 88.983 -57.876 56.092
532184 89.029 -57.895 56.062
532185 89.079 -57.912 56.033
532186 89.127 -57.929 56.005
532187 89.177 -57.946 55.978
532188 89.226 -57.962 55.953
532189 89.274 -57.980 55.929
532190 89.324 -57.996 55.909
532191 89.373 -58.013 55.891
532192 89.421 -58.031 55.875
532193 89.471 -58.047 55.861
532194 89.520 -58.065 55.848
532195 89.569 -58.081 55.837
532196 89.615 -58.096 55.827
532197 89.652 -58.106 55.818
532198 89.688 -58.117 55.812
532199 89.724 -58.127 55.806
532200 89.761 -58.137 55.801
532201 89.806 -58.141 55.794
532202 89.851 -58.146 55.788
532203 89.896 -58.150 55.781
532204 89.942 -58.155 55.775
532205 89.986 -58.159 55.768
532206 90.032 -58.164 55.759
532207 90.075 -58.171 55.745
532208 90.107 -58.188 55.726
532209 90.140 -58.206 55.707
532210 90.172 -58.223 55.688
532211 90.206 -58.234 55.668
532212 90.240 -58.246 55.648
532213 90.281 -58.252 55.628
532214 90.323 -58.256 55.609
532215 90.365 -58.261 55.590
532216 90.407 -58.266 55.575
532217 90.446 -58.272 55.559
532218 90.483 -58.280 55.541
532219 90.521 -58.287 55.521
532220 90.559 -58.294 55.498
532221 90.597 -58.301 55.474
532222 90.635 -58.309 55.452
532223 90.673 -58.316 55.434
532224 90.711 -58.323 55.417
532225 90.751 -58.341 55.400
532226 90.793 -58.359 55.382
532227 90.834 -58.378 55.365
532228 90.876 -58.397 55.348
532229 90.917 -58.415 55.330
532230 90.959 -58.434 55.313
532231 91.000 -58.453 55.294
532232 91.041 -58.471 55.270
532233 91.083 -58.490 55.243
532234 91.124 -58.510 55.219
532235 91.164 -58.533 55.196
532236 91.204 -58.557 55.171
532237 91.243 -58.580 55.143
532238 91.281 -58.592 55.117
532239 91.319 -58.604 55.096
532240 91.357 -58.617 55.078
532241 91.395 -58.629 55.062
532242 91.433 -58.642 55.048
532243 91.471 -58.654 55.035
532244 91.510 -58.666 55.022
532245 91.547 -58.678 55.010
532246 91.586 -58.690 54.995
532247 91.622 -58.702 54.975
532248 91.657 -58.712 54.954
532249 91.693 -58.721 54.927
532250 91.729 -58.736 54.893
532251 91.765 -58.750 54.857
532252 91.801 -58.764 54.824
532253 91.837 -58.780 54.797
532254 91.873 -58.794 54.774
532255 91.909 -58.808 54.756
532256 91.945 -58.823 54.742
532257 91.982 -58.837 54.732
532258 92.017 -58.852 54.727
532259 92.053 -58.853 54.727
532260 92.088 -58.853 54.730
532261 92.124 -58.854 54.741
532262 92.159 -58.854 54.756
532263 92.195 -58.855 54.767
532264 92.230 -58.855 54.768
532265 92.266 -58.855 54.759
532266 92.301 -58.855 54.746
532267 92.337 -58.856 54.729
532268 92.372 -58.856 54.713
532269 92.404 -58.859 54.703
532270 92.432 -58.864 54.697
532271 92.459 -58.870 54.692
532272 92.499 -58.872 54.698
532273 92.547 -58.871 54.709
532274 92.594 -58.870 54.716
532275 92.641 -58.869 54.717
532276 92.689 -58.868 54.712
532277 92.736 -58.867 54.706
532278 92.783 -58.867 54.698
532279 92.831 -58.866 54.687
532280 92.877 -58.869 54.671
532281 92.923 -58.875 54.650
532282 92.968 -58.881 54.630
532283 93.014 -58.886 54.611
532284 93.058 -58.893 54.594
532285 93.095 -58.903 54.581
532286 93.134 -58.911 54.564
532287 93.173 -58.919 54.544
532288 93.212 -58.927 54.525
532289 93.251 -58.934 54.507
532290 93.289 -58.942 54.489
532291 93.328 -58.949 54.471
532292 93.366 -58.958 54.456
532293 93.405 -58.965 54.444
532294 93.444 -58.973 54.434
532295 93.483 -58.980 54.424
532296 93.521 -58.989 54.412
532297 93.558 -59.002 54.397
532298 93.594 -59.016 54.378
532299 93.630 -59.030 54.355
532300 93.665 -59.050 54.330
532301 93.700 -59.070 54.305
532302 93.735 -59.090 54.279
532303 93.770 -59.111 54.258
532304 93.805 -59.131 54.236
532305 93.840 -59.151 54.214
532306 93.875 -59.171 54.194
532307 93.910 -59.192 54.173
532308 93.945 -59.212 54.149
532309 93.980 -59.232 54.128
532310 94.015 -59.253 54.106
532311 94.050 -59.273 54.084
532312 94.085 -59.294 54.061
532313 94.120 -59.313 54.038
532314 94.155 -59.334 54.021
532315 94.190 -59.354 54.003
532316 94.224 -59.375 53.985
532317 94.260 -59.394 53.967
532318 94.294 -59.415 53.947
532319 94.330 -59.435 53.929
532320 94.364 -59.456 53.914
532321 94.400 -59.475 53.899
532322 94.434 -59.496 53.885
532323 94.469 -59.504 53.873
532324 94.502 -59.510 53.862
532325 94.537 -59.517 53.848
532326 94.570 -59.525 53.833
532327 94.605 -59.532 53.816
532328 94.642 -59.541 53.795
532329 94.680 -59.552 53.773
532330 94.720 -59.563 53.753
532331 94.761 -59.572 53.729
532332 94.802 -59.581 53.702
532333 94.842 -59.590 53.672
532334 94.883 -59.599 53.642
532335 94.924 -59.609 53.611
532336 94.964 -59.619 53.582
532337 95.005 -59.629 53.555
532338 95.046 -59.639 53.529
532339 95.086 -59.650 53.508
532340 95.126 -59.662 53.490
532341 95.166 -59.675 53.475
532342 95.205 -59.687 53.459
532343 95.245 -59.700 53.444
532344 95.286 -59.709 53.429
532345 95.327 -59.718 53.412
532346 95.369 -59.726 53.392
532347 95.409 -59.736 53.373
532348 95.451 -59.744 53.354
532349 95.486 -59.754 53.332
532350 95.519 -59.764 53.306
532351 95.549 -59.773 53.275
532352 95.578 -59.780 53.243
532353 95.607 -59.788 53.216
532354 95.636 -59.796 53.195
532355 95.668 -59.803 53.179
532356 95.702 -59.810 53.169
532357 95.737 -59.816 53.160
532358 95.771 -59.823 53.151
532359 95.806 -59.829 53.141
532360 95.840 -59.836 53.128
532361 95.874 -59.842 53.112
532362 95.908 -59.849 53.097
532363 95.942 -59.855 53.076
532364 95.977 -59.862 53.051
532365 96.014 -59.867 53.024
532366 96.052 -59.872 52.996
532367 96.089 -59.877 52.969
532368 96.127 -59.883 52.938
532369 96.164 -59.888 52.911
532370 96.201 -59.893 52.887
532371 96.238 -59.898 52.864
532372 96.274 -59.904 52.840
532373 96.311 -59.909 52.814
532374 96.348 -59.911 52.792
532375 96.385 -59.911 52.769
532376 96.421 -59.911 52.746
532377 96.457 -59.911 52.722
532378 96.500 -59.916 52.695
532379 96.546 -59.923 52.664
532380 96.592 -59.929 52.631
532381 96.635 -59.933 52.595
532382 96.673 -59.933 52.557
532383 96.711 -59.934 52.518
532384 96.749 -59.935 52.481
532385 96.786 -59.935 52.448
532386 96.818 -59.934 52.414
532387 96.850 -59.933 52.378
532388 96.882 -59.932 52.338
532389 96.916 -59.934 52.296
532390 96.950 -59.936 52.258
532391 96.983 -59.939 52.217
532392 97.017 -59.941 52.175
532393 97.051 -59.943 52.129
532394 97.085 -59.946 52.080
532395 97.117 -59.952 52.029
532396 97.148 -59.965 51.978
532397 97.178 -59.978 51.929
532398 97.208 -59.991 51.881
532399 97.238 -60.003 51.834
532400 97.268 -60.016 51.785
532401 97.302 -60.024 51.738
532402 97.338 -60.024 51.694
532403 97.375 -60.024 51.650
532404 97.412 -60.025 51.606
532405 97.449 -60.026 51.562
532406 97.486 -60.027 51.519
532407 97.524 -60.030 51.476
532408 97.565 -60.042 51.434
532409 97.606 -60.053 51.392
532410 97.647 -60.065 51.350
532411 97.687 -60.075 51.309
532412 97.728 -60.086 51.268
532413 97.770 -60.098 51.228
532414 97.810 -60.108 51.189
532415 97.851 -60.120 51.151
532416 97.892 -60.131 51.112
532417 97.933 -60.143 51.073
532418 97.974 -60.153 51.034
532419 98.015 -60.165 50.995
532420 98.042 -60.168 50.955
532421 98.065 -60.170 50.913
532422 98.087 -60.171 50.870
532423 98.102 -60.169 50.828
532424 98.119 -60.168 50.786
532425 98.134 -60.167 50.746
532426 98.150 -60.165 50.705
532427 98.171 -60.166 50.663
532428 98.211 -60.174 50.621
532429 98.257 -60.159 50.578
532430 98.303 -60.141 50.536
532431 98.350 -60.124 50.494
532432 98.395 -60.110 50.452
532433 98.431 -60.115 50.409
532434 98.467 -60.119 50.366
532435 98.504 -60.123 50.323
532436 98.541 -60.127 50.279
532437 98.577 -60.132 50.237
532438 98.613 -60.136 50.195
532439 98.649 -60.140 50.151
532440 98.686 -60.145 50.107
532441 98.722 -60.149 50.063
532442 98.758 -60.153 50.020
532443 98.793 -60.150 49.976
532444 98.827 -60.127 49.935
532445 98.859 -60.105 49.892
532866 88.425 -57.558 56.559
532867 88.459 -57.570 56.531
532868 88.493 -57.581 56.504
532869 88.527 -57.592 56.478
532870 88.560 -57.604 56.452
532871 88.594 -57.616 56.426
532872 88.627 -57.627 56.400
532873 88.661 -57.639 56.375
532874 88.695 -57.650 56.349
532875 88.729 -57.662 56.325
532876 88.762 -57.673 56.300
532877 88.797 -57.686 56.275
532878 88.835 -57.702 56.251
532879 88.874 -57.718 56.226
532880 88.900 -57.741 56.203
532881 88.924 -57.765 56.178
532882 88.949 -57.789 56.152
532883 88.973 -57.814 56.124
532884 88.998 -57.836 56.095
532885 89.042 -57.850 56.067
532886 89.091 -57.867 56.039
532887 89.140 -57.883 56.011
532888 89.189 -57.901 55.983
532889 89.238 -57.918 55.957
532890 89.287 -57.934 55.933
532891 89.336 -57.952 55.913
532892 89.386 -57.968 55.895
532893 89.435 -57.986 55.878
532894 89.483 -58.003 55.864
532895 89.533 -58.019 55.851
532896 89.582 -58.037 55.840
532897 89.631 -58.053 55.830
532898 89.667 -58.063 55.822
532899 89.703 -58.073 55.815
532900 89.740 -58.083 55.809
532901 89.776 -58.093 55.804
532902 89.813 -58.103 55.797
532903 89.857 -58.109 55.790
532904 89.902 -58.113 55.783
532905 89.947 -58.118 55.775
532906 89.992 -58.122 55.766
532907 90.038 -58.127 55.755
532908 90.083 -58.131 55.740
532909 90.118 -58.146 55.722
532910 90.150 -58.163 55.703
532911 90.183 -58.181 55.686
532912 90.217 -58.193 55.666
532913 90.250 -58.204 55.646
532914 90.286 -58.214 55.627
532915 90.328 -58.219 55.607
532916 90.370 -58.223 55.586
532917 90.412 -58.228 55.568
532918 90.454 -58.232 55.550
532919 90.496 -58.237 55.531
532920 90.534 -58.244 55.511
532921 90.572 -58.252 55.491
532922 90.610 -58.259 55.467
532923 90.648 -58.266 55.447
532924 90.686 -58.274 55.426
532925 90.724 -58.281 55.403
532926 90.762 -58.291 55.383
532927 90.804 -58.309 55.363
532928 90.845 -58.328 55.346
532929 90.886 -58.347 55.329
532930 90.928 -58.366 55.314
532931 90.969 -58.384 55.299
532932 91.011 -58.403 55.282
532933 91.052 -58.422 55.258
532934 91.093 -58.440 55.228
532935 91.134 -58.463 55.202
532936 91.173 -58.488 55.179
532937 91.213 -58.511 55.155
532938 91.252 -58.535 55.129
532939 91.292 -58.558 55.104
532940 91.330 -58.571 55.083
532941 91.368 -58.583 55.066
532942 91.406 -58.595 55.050
532943 91.444 -58.607 55.035
532944 91.483 -58.620 55.023
532945 91.521 -58.632 55.012
532946 91.559 -58.645 55.001
532947 91.597 -58.657 54.988
532948 91.631 -58.668 54.973
532949 91.666 -58.678 54.954
532950 91.702 -58.687 54.931
532951 91.737 -58.697 54.903
532952 91.773 -58.709 54.872
532953 91.809 -58.723 54.841
532954 91.846 -58.738 54.811
532955 91.881 -58.752 54.784
532956 91.918 -58.767 54.763
532957 91.954 -58.781 54.748
532958 91.990 -58.795 54.736
532959 92.026 -58.810 54.725
532960 92.062 -58.814 54.720
532961 92.097 -58.814 54.718
532962 92.133 -58.815 54.721
532963 92.168 -58.815 54.729
532964 92.203 -58.816 54.739
532965 92.239 -58.816 54.740
532966 92.274 -58.817 54.732
532967 92.310 -58.817 54.722
532968 92.345 -58.817 54.711
532969 92.381 -58.817 54.699
532970 92.416 -58.818 54.690
532971 92.446 -58.822 54.686
532972 92.474 -58.827 54.687
532973 92.514 -58.829 54.697
532974 92.561 -58.828 54.709
532975 92.609 -58.827 54.715
532976 92.656 -58.827 54.715
532977 92.703 -58.826 54.713
532978 92.751 -58.825 54.707
532979 92.796 -58.830 54.696
532980 92.842 -58.836 54.684
532981 92.888 -58.841 54.669
532982 92.933 -58.847 54.651
532983 92.979 -58.853 54.633
532984 93.024 -58.859 54.616
532985 93.066 -58.867 54.599
532986 93.103 -58.877 54.583
532987 93.142 -58.886 54.567
532988 93.181 -58.893 54.550
532989 93.219 -58.901 54.531
532990 93.258 -58.908 54.513
532991 93.297 -58.916 54.496
532992 93.336 -58.924 54.480
532993 93.374 -58.932 54.465
532994 93.413 -58.940 54.453
532995 93.452 -58.947 54.441
532996 93.490 -58.955 54.429
532997 93.529 -58.964 54.417
532998 93.565 -58.977 54.403
532999 93.602 -58.991 54.385
533000 93.638 -59.005 54.365
533001 93.674 -59.018 54.344
533002 93.709 -59.036 54.320
533003 93.745 -59.056 54.295
533004 93.779 -59.077 54.273
533005 93.815 -59.096 54.252
533006 93.849 -59.117 54.231
533007 93.885 -59.137 54.210
533008 93.919 -59.158 54.189
533009 93.955 -59.178 54.166
533010 93.989 -59.198 54.143
533011 94.024 -59.218 54.120
533012 94.059 -59.239 54.098
533013 94.094 -59.259 54.077
533014 94.129 -59.279 54.057
533015 94.164 -59.300 54.037
533016 94.199 -59.319 54.018
533017 94.234 -59.340 54.000
533018 94.269 -59.360 53.981
533019 94.304 -59.381 53.961
533020 94.339 -59.400 53.941
533021 94.374 -59.421 53.921
533022 94.409 -59.441 53.902
533023 94.444 -59.460 53.884
533024 94.478 -59.466 53.869
533025 94.512 -59.474 53.857
533026 94.546 -59.481 53.843
533027 94.580 -59.488 53.828
533028 94.614 -59.495 53.810
533029 94.649 -59.504 53.787
533030 94.688 -59.514 53.762
533031 94.729 -59.524 53.735
533032 94.769 -59.533 53.707
533033 94.810 -59.542 53.677
533034 94.851 -59.551 53.645
533035 94.892 -59.561 53.613
533036 94.932 -59.570 53.585
533037 94.973 -59.579 53.558
533038 95.014 -59.589 53.532
533039 95.054 -59.600 53.507
533040 95.095 -59.610 53.484
533041 95.134 -59.623 53.464
533042 95.174 -59.635 53.447
533043 95.214 -59.648 53.430
533044 95.254 -59.661 53.415
533045 95.294 -59.673 53.400
533046 95.335 -59.681 53.386
533047 95.376 -59.690 53.369
533048 95.417 -59.698 53.352
533049 95.457 -59.707 53.334
533050 95.489 -59.718 53.315
533051 95.522 -59.728 53.293
533052 95.555 -59.738 53.265
533053 95.586 -59.747 53.240
533054 95.615 -59.754 53.222
533055 95.644 -59.762 53.207
533056 95.675 -59.770 53.195
533057 95.709 -59.776 53.186
533058 95.744 -59.783 53.177
533059 95.778 -59.789 53.166
533060 95.813 -59.795 53.155
533061 95.846 -59.802 53.142
533062 95.881 -59.809 53.126
533063 95.914 -59.816 53.108
533064 95.948 -59.823 53.084
533065 95.982 -59.830 53.058
533066 96.020 -59.835 53.031
533067 96.057 -59.839 53.003
533068 96.095 -59.845 52.973
533069 96.131 -59.850 52.945
533070 96.168 -59.855 52.919
533071 96.205 -59.861 52.896
533072 96.242 -59.866 52.873
533073 96.279 -59.871 52.848
533074 96.316 -59.877 52.822
533075 96.353 -59.882 52.801
533076 96.389 -59.883 52.780
533077 96.426 -59.883 52.757
533078 96.462 -59.883 52.733
533079 96.507 -59.889 52.704
533080 96.552 -59.895 52.672
533081 96.595 -59.899 52.640
533082 96.633 -59.900 52.604
533083 96.672 -59.901 52.566
533084 96.711 -59.902 52.529
533085 96.749 -59.902 52.496
533086 96.783 -59.902 52.464
533087 96.815 -59.900 52.433
533088 96.847 -59.898 52.399
533089 96.879 -59.898 52.362
533090 96.912 -59.901 52.323
533091 96.946 -59.903 52.285
533092 96.980 -59.905 52.246
533093 97.014 -59.908 52.205
533094 97.047 -59.910 52.160
533095 97.082 -59.912 52.109
533096 97.115 -59.914 52.055
533097 97.147 -59.924 52.003
533098 97.177 -59.936 51.952
533099 97.207 -59.949 51.904
533100 97.237 -59.962 51.856
533101 97.267 -59.974 51.807
533102 97.299 -59.984 51.761
533103 97.336 -59.985 51.717
533104 97.373 -59.986 51.673
533105 97.410 -59.986 51.628
533106 97.447 -59.987 51.584
533107 97.484 -59.988 51.539
533108 97.523 -59.996 51.495
533109 97.564 -60.007 51.450
533110 97.605 -60.018 51.406
533111 97.646 -60.029 51.362
533112 97.687 -60.040 51.319
533113 97.728 -60.051 51.277
533114 97.769 -60.062 51.235
533115 97.809 -60.074 51.194
533116 97.851 -60.085 51.155
533117 97.891 -60.096 51.115
533118 97.932 -60.107 51.075
533119 97.973 -60.118 51.036
533120 98.014 -60.129 50.997
533121 98.055 -60.140 50.957
533122 98.081 -60.143 50.915
533123 98.102 -60.144 50.874
533124 98.118 -60.143 50.832
533125 98.134 -60.142 50.792
533126 98.150 -60.140 50.751
533127 98.166 -60.139 50.710
533128 98.181 -60.137 50.668
533129 98.212 -60.142 50.627
533130 98.256 -60.136 50.584
533131 98.302 -60.119 50.541
533132 98.349 -60.102 50.500
533133 98.395 -60.086 50.458
533134 98.438 -60.077 50.415
533135 98.474 -60.081 50.373
533136 98.510 -60.086 50.330
533137 98.547 -60.090 50.287
533138 98.583 -60.094 50.245
533139 98.619 -60.099 50.203
533140 98.655 -60.103 50.159
533141 98.692 -60.108 50.115
533142 98.728 -60.112 50.072
533143 98.764 -60.116 50.028
533144 98.799 -60.109 49.985
533145 98.832 -60.086 49.943
533146 98.865 -60.065 49.900
533147 98.906 -60.047 49.858
533568 88.471 -57.526 56.519
533569 88.505 -57.537 56.491
533570 88.539 -57.548 56.465
533571 88.572 -57.560 56.440
533572 88.606 -57.572 56.415
533573 88.640 -57.583 56.390
533574 88.674 -57.595 56.366
533575 88.707 -57.606 56.341
533576 88.741 -57.617 56.316
533577 88.777 -57.632 56.291
533578 88.815 -57.648 56.267
533579 88.853 -57.664 56.243
533580 88.892 -57.680 56.219
533581 88.918 -57.702 56.197
533582 88.943 -57.726 56.173
533583 88.967 -57.750 56.148
533584 88.991 -57.774 56.122
533585 89.018 -57.797 56.097
533586 89.059 -57.808 56.069
533587 89.104 -57.822 56.042
533588 89.153 -57.839 56.015
533589 89.202 -57.855 55.987
533590 89.251 -57.873 55.961
533591 89.300 -57.890 55.937
533592 89.349 -57.907 55.917
533593 89.398 -57.924 55.898
533594 89.448 -57.940 55.881
533595 89.496 -57.958 55.866
533596 89.545 -57.974 55.854
533597 89.595 -57.991 55.843
533598 89.644 -58.009 55.833
533599 89.683 -58.020 55.826
533600 89.719 -58.030 55.820
533601 89.756 -58.040 55.814
533602 89.792 -58.050 55.809
533603 89.828 -58.059 55.803
533604 89.865 -58.070 55.796
533605 89.908 -58.075 55.788
533606 89.953 -58.081 55.780
533607 89.998 -58.085 55.769
533608 90.043 -58.090 55.757
533609 90.089 -58.094 55.742
533610 90.128 -58.104 55.724
533611 90.161 -58.122 55.705
533612 90.193 -58.139 55.686
533613 90.227 -58.150 55.667
533614 90.261 -58.162 55.648
533615 90.295 -58.173 55.629
533616 90.333 -58.181 55.609
533617 90.376 -58.186 55.587
533618 90.417 -58.191 55.567
533619 90.460 -58.195 55.548
533620 90.502 -58.200 55.528
533621 90.544 -58.204 55.509
533622 90.585 -58.209 55.491
533623 90.622 -58.216 55.470
533624 90.661 -58.224 55.451
533625 90.698 -58.231 55.430
533626 90.736 -58.238 55.407
533627 90.774 -58.246 55.384
533628 90.814 -58.260 55.358
533629 90.856 -58.278 55.336
533630 90.897 -58.297 55.319
533631 90.938 -58.316 55.304
533632 90.979 -58.334 55.288
533633 91.021 -58.353 55.269
533634 91.063 -58.372 55.244
533635 91.103 -58.393 55.211
533636 91.143 -58.417 55.181
533637 91.182 -58.441 55.153
533638 91.222 -58.466 55.129
533639 91.262 -58.489 55.108
533640 91.301 -58.513 55.089
533641 91.341 -58.537 55.070
533642 91.379 -58.549 55.053
533643 91.417 -58.561 55.037
533644 91.455 -58.574 55.023
533645 91.494 -58.586 55.013
533646 91.531 -58.598 55.003
533647 91.570 -58.610 54.993
533648 91.608 -58.623 54.982
533649 91.641 -58.633 54.972
533650 91.674 -58.644 54.958
533651 91.710 -58.653 54.939
533652 91.746 -58.663 54.918
533653 91.782 -58.672 54.893
533654 91.818 -58.682 54.865
533655 91.854 -58.696 54.833
533656 91.890 -58.711 54.810
533657 91.926 -58.725 54.785
533658 91.962 -58.739 54.761
533659 91.998 -58.754 54.742
533660 92.034 -58.768 54.730
533661 92.070 -58.776 54.721
533662 92.105 -58.776 54.715
533663 92.141 -58.777 54.712
533664 92.176 -58.777 54.711
533665 92.212 -58.777 54.712
533666 92.247 -58.777 54.711
533667 92.283 -58.778 54.708
533668 92.318 -58.778 54.707
533669 92.354 -58.779 54.705
533670 92.389 -58.779 54.698
533671 92.425 -58.780 54.688
533672 92.460 -58.780 54.684
533673 92.489 -58.785 54.688
533674 92.528 -58.786 54.695
533675 92.576 -58.786 54.701
533676 92.623 -58.785 54.702
533677 92.670 -58.784 54.703
533678 92.716 -58.790 54.704
533679 92.762 -58.796 54.701
533680 92.807 -58.802 54.692
533681 92.853 -58.808 54.678
533682 92.899 -58.814 54.662
533683 92.944 -58.820 54.649
533684 92.990 -58.826 54.635
533685 93.036 -58.832 54.621
533686 93.074 -58.841 54.607
533687 93.112 -58.851 54.591
533688 93.150 -58.860 54.576
533689 93.189 -58.867 54.560
533690 93.227 -58.875 54.542
533691 93.266 -58.883 54.525
533692 93.305 -58.890 54.509
533693 93.344 -58.899 54.493
533694 93.382 -58.906 54.477
533695 93.421 -58.914 54.463
533696 93.460 -58.921 54.449
533697 93.498 -58.929 54.435
533698 93.536 -58.939 54.420
533699 93.572 -58.952 54.405
533700 93.609 -58.966 54.388
533701 93.645 -58.980 54.370
533702 93.682 -58.993 54.350
533703 93.718 -59.007 54.330
533704 93.754 -59.021 54.311
533705 93.789 -59.042 54.291
533706 93.824 -59.062 54.272
533707 93.859 -59.083 54.251
533708 93.894 -59.102 54.230
533709 93.929 -59.123 54.207
533710 93.964 -59.143 54.185
533711 93.999 -59.164 54.161
533712 94.034 -59.184 54.138
533713 94.069 -59.204 54.116
533714 94.104 -59.225 54.096
533715 94.139 -59.245 54.078
533716 94.174 -59.265 54.058
533717 94.208 -59.285 54.037
533718 94.244 -59.306 54.017
533719 94.278 -59.326 53.998
533720 94.314 -59.346 53.977
533721 94.348 -59.366 53.955
533722 94.384 -59.387 53.932
533723 94.418 -59.407 53.912
533724 94.453 -59.423 53.893
533725 94.487 -59.430 53.877
533726 94.521 -59.438 53.865
533727 94.555 -59.444 53.853
533728 94.589 -59.451 53.839
533729 94.623 -59.459 53.821
533730 94.657 -59.466 53.799
533731 94.696 -59.475 53.771
533732 94.737 -59.485 53.738
533733 94.778 -59.494 53.704
533734 94.818 -59.504 53.670
533735 94.859 -59.513 53.636
533736 94.900 -59.522 53.604
533737 94.941 -59.532 53.578
533738 94.981 -59.541 53.554
533739 95.022 -59.550 53.531
533740 95.063 -59.560 53.507
533741 95.103 -59.571 53.481
533742 95.143 -59.584 53.458
533743 95.183 -59.597 53.437
533744 95.222 -59.610 53.418
533745 95.262 -59.622 53.400
533746 95.302 -59.635 53.381
533747 95.342 -59.645 53.365
533748 95.384 -59.653 53.349
533749 95.424 -59.662 53.331
533750 95.460 -59.671 53.312
533751 95.493 -59.682 53.294
533752 95.526 -59.692 53.277
533753 95.558 -59.702 53.256
533754 95.591 -59.712 53.243
533755 95.622 -59.721 53.234
533756 95.652 -59.729 53.225
533757 95.682 -59.736 53.216
533758 95.716 -59.742 53.207
533759 95.751 -59.749 53.195
533760 95.785 -59.756 53.182
533761 95.819 -59.763 53.169
533762 95.853 -59.769 53.155
533763 95.887 -59.776 53.139
533764 95.920 -59.783 53.120
533765 95.955 -59.789 53.094
533766 95.988 -59.796 53.067
533767 96.025 -59.802 53.039
533768 96.062 -59.807 53.010
533769 96.099 -59.812 52.980
533770 96.136 -59.817 52.954
533771 96.173 -59.823 52.930
533772 96.210 -59.828 52.907
533773 96.247 -59.833 52.885
533774 96.284 -59.839 52.860
533775 96.321 -59.844 52.834
533776 96.358 -59.849 52.812
533777 96.394 -59.855 52.790
533778 96.431 -59.855 52.767
533779 96.467 -59.855 52.742
533780 96.513 -59.861 52.712
533781 96.556 -59.865 52.680
533782 96.595 -59.866 52.647
533783 96.633 -59.867 52.612
533784 96.671 -59.867 52.576
533785 96.710 -59.868 52.542
533786 96.747 -59.869 52.511
533787 96.779 -59.867 52.479
533788 96.811 -59.866 52.448
533789 96.843 -59.865 52.415
533790 96.875 -59.864 52.379
533791 96.909 -59.867 52.342
533792 96.943 -59.869 52.306
533793 96.976 -59.871 52.268
533794 97.011 -59.873 52.228
533795 97.044 -59.876 52.182
533796 97.078 -59.879 52.132
533797 97.112 -59.881 52.078
533798 97.146 -59.883 52.026
533799 97.176 -59.895 51.977
533800 97.207 -59.908 51.930
533801 97.237 -59.920 51.883
533802 97.267 -59.933 51.836
533803 97.297 -59.946 51.792
533804 97.334 -59.946 51.747
533805 97.371 -59.947 51.701
533806 97.407 -59.948 51.655
533807 97.444 -59.948 51.609
533808 97.481 -59.949 51.562
533809 97.522 -59.960 51.516
533810 97.563 -59.971 51.469
533811 97.604 -59.983 51.423
533812 97.645 -59.993 51.378
533813 97.686 -60.005 51.334
533814 97.727 -60.016 51.290
533815 97.768 -60.027 51.247
533816 97.809 -60.038 51.204
533817 97.850 -60.049 51.163
533818 97.890 -60.061 51.123
533819 97.932 -60.071 51.081
533820 97.972 -60.083 51.040
533821 98.013 -60.094 51.000
533822 98.054 -60.105 50.960
533823 98.095 -60.116 50.919
533824 98.118 -60.118 50.879
533825 98.134 -60.117 50.840
533826 98.149 -60.115 50.800
533827 98.165 -60.114 50.760
533828 98.181 -60.112 50.719
533829 98.197 -60.112 50.678
533830 98.214 -60.110 50.635
533831 98.254 -60.114 50.591
533832 98.301 -60.097 50.549
533833 98.348 -60.080 50.507
533834 98.394 -60.064 50.465
533835 98.440 -60.047 50.422
533836 98.480 -60.044 50.380
533837 98.516 -60.049 50.338
533838 98.553 -60.052 50.296
533839 98.589 -60.057 50.254
533840 98.625 -60.061 50.211
533841 98.661 -60.066 50.168
533842 98.698 -60.070 50.124
533843 98.734 -60.074 50.081
533844 98.771 -60.079 50.037
533845 98.805 -60.068 49.994
533846 98.838 -60.046 49.951
533847 98.871 -60.024 49.908
533848 98.915 -60.008 49.866
533849 98.978 -60.006 49.823
534270 88.517 -57.493 56.478
534271 88.551 -57.504 56.455
534272 88.585 -57.516 56.430
534273 88.618 -57.528 56.405
534274 88.652 -57.539 56.381
534275 88.686 -57.551 56.358
534276 88.719 -57.562 56.333
534277 88.756 -57.577 56.307
534278 88.794 -57.593 56.282
534279 88.833 -57.609 56.259
534280 88.871 -57.625 56.236
534281 88.909 -57.641 56.212
534282 88.937 -57.663 56.190
534283 88.961 -57.687 56.167
534284 88.986 -57.711 56.143
534285 89.010 -57.736 56.118
534286 89.037 -57.758 56.093
534287 89.079 -57.768 56.067
534288 89.120 -57.779 56.042
534289 89.166 -57.794 56.016
534290 89.215 -57.811 55.990
534291 89.264 -57.828 55.965
534292 89.313 -57.845 55.943
534293 89.362 -57.861 55.921
534294 89.411 -57.879 55.901
534295 89.460 -57.896 55.884
534296 89.510 -57.912 55.869
534297 89.558 -57.930 55.857
534298 89.607 -57.946 55.846
534299 89.657 -57.963 55.837
534300 89.698 -57.977 55.829
534301 89.735 -57.987 55.824
534302 89.771 -57.996 55.819
534303 89.808 -58.006 55.814
534304 89.844 -58.016 55.809
534305 89.880 -58.026 55.803
534306 89.917 -58.036 55.796
534307 89.959 -58.043 55.790
534308 90.004 -58.047 55.779
534309 90.049 -58.052 55.767
534310 90.094 -58.056 55.750
534311 90.139 -58.062 55.731
534312 90.172 -58.079 55.712
534313 90.204 -58.096 55.693
534314 90.238 -58.109 55.675
534315 90.271 -58.120 55.657
534316 90.305 -58.131 55.640
534317 90.339 -58.143 55.621
534318 90.381 -58.148 55.596
534319 90.423 -58.153 55.575
534320 90.465 -58.157 55.553
534321 90.507 -58.162 55.532
534322 90.549 -58.166 55.513
534323 90.591 -58.171 55.497
534324 90.634 -58.175 55.478
534325 90.673 -58.181 55.460
534326 90.711 -58.189 55.440
534327 90.749 -58.196 55.418
534328 90.787 -58.203 55.393
534329 90.825 -58.210 55.362
534330 90.866 -58.228 55.332
534331 90.907 -58.247 55.309
534332 90.949 -58.266 55.290
534333 90.990 -58.285 55.272
534334 91.031 -58.303 55.251
534335 91.073 -58.323 55.225
534336 91.112 -58.347 55.196
534337 91.152 -58.372 55.170
534338 91.191 -58.395 55.148
534339 91.231 -58.419 55.130
534340 91.271 -58.444 55.114
534341 91.311 -58.468 55.094
534342 91.351 -58.491 55.073
534343 91.390 -58.516 55.052
534344 91.428 -58.528 55.033
534345 91.466 -58.540 55.019
534346 91.504 -58.552 55.009
534347 91.542 -58.564 54.996
534348 91.581 -58.577 54.985
534349 91.617 -58.589 54.978
534350 91.650 -58.599 54.973
534351 91.683 -58.610 54.965
534352 91.719 -58.619 54.952
534353 91.755 -58.629 54.935
534354 91.791 -58.638 54.914
534355 91.826 -58.648 54.889
534356 91.862 -58.657 54.857
534357 91.898 -58.669 54.834
534358 91.934 -58.683 54.808
534359 91.970 -58.698 54.781
534360 92.007 -58.712 54.758
534361 92.042 -58.726 54.738
534362 92.079 -58.737 54.726
534363 92.114 -58.737 54.717
534364 92.149 -58.738 54.712
534365 92.185 -58.738 54.709
534366 92.220 -58.739 54.711
534367 92.256 -58.739 54.715
534368 92.291 -58.739 54.720
534369 92.327 -58.739 54.724
534370 92.362 -58.740 54.726
534371 92.398 -58.740 54.724
534372 92.433 -58.741 54.715
534373 92.469 -58.741 54.707
534374 92.504 -58.742 54.699
534375 92.543 -58.744 54.690
534376 92.590 -58.745 54.683
534377 92.636 -58.751 54.678
534378 92.681 -58.757 54.679
534379 92.727 -58.762 54.685
534380 92.772 -58.768 54.688
534381 92.818 -58.774 54.686
534382 92.863 -58.780 54.673
534383 92.909 -58.786 54.656
534384 92.955 -58.792 54.650
534385 93.000 -58.798 54.641
534386 93.044 -58.805 54.632
534387 93.082 -58.814 54.622
534388 93.120 -58.824 54.607
534389 93.158 -58.834 54.591
534390 93.197 -58.842 54.575
534391 93.235 -58.849 54.559
534392 93.274 -58.857 54.544
534393 93.312 -58.864 54.529
534394 93.351 -58.873 54.513
534395 93.390 -58.880 54.498
534396 93.429 -58.888 54.482
534397 93.468 -58.896 54.465
534398 93.506 -58.903 54.449
534399 93.544 -58.914 54.432
534400 93.580 -58.927 54.416
534401 93.616 -58.942 54.400
534402 93.652 -58.955 54.383
534403 93.689 -58.969 54.366
534404 93.725 -58.983 54.349
534405 93.762 -58.996 54.332
534406 93.798 -59.010 54.317
534407 93.833 -59.027 54.297
534408 93.868 -59.048 54.275
534409 93.903 -59.068 54.251
534410 93.938 -59.089 54.228
534411 93.973 -59.109 54.207
534412 94.008 -59.129 54.183
534413 94.043 -59.150 54.159
534414 94.078 -59.170 54.137
534415 94.113 -59.190 54.118
534416 94.148 -59.210 54.100
534417 94.183 -59.231 54.080
534418 94.218 -59.251 54.059
534419 94.253 -59.271 54.038
534420 94.288 -59.291 54.017
534421 94.323 -59.312 53.995
534422 94.358 -59.332 53.970
534423 94.393 -59.352 53.947
534424 94.428 -59.372 53.926
534425 94.462 -59.387 53.907
534426 94.496 -59.394 53.890
534427 94.530 -59.400 53.876
534428 94.564 -59.408 53.865
534429 94.598 -59.415 53.851
534430 94.632 -59.422 53.833
534431 94.666 -59.429 53.811
534432 94.705 -59.438 53.780
534433 94.746 -59.447 53.743
534434 94.787 -59.457 53.706
534435 94.828 -59.466 53.669
534436 94.868 -59.475 53.635
534437 94.909 -59.484 53.604
534438 94.950 -59.494 53.579
534439 94.991 -59.503 53.555
534440 95.031 -59.512 53.532
534441 95.072 -59.521 53.508
534442 95.112 -59.532 53.482
534443 95.151 -59.545 53.456
534444 95.191 -59.557 53.434
534445 95.231 -59.570 53.414
534446 95.270 -59.583 53.396
534447 95.310 -59.596 53.378
534448 95.350 -59.607 53.359
534449 95.391 -59.616 53.341
534450 95.431 -59.625 53.322
534451 95.464 -59.635 53.304
534452 95.497 -59.645 53.289
534453 95.530 -59.656 53.275
534454 95.562 -59.666 53.259
534455 95.595 -59.676 53.252
534456 95.628 -59.686 53.248
534457 95.659 -59.695 53.244
534458 95.689 -59.703 53.236
534459 95.723 -59.709 53.229
534460 95.757 -59.716 53.214
534461 95.792 -59.723 53.199
534462 95.825 -59.729 53.183
534463 95.859 -59.736 53.167
534464 95.893 -59.743 53.151
534465 95.927 -59.750 53.131
534466 95.961 -59.757 53.106
534467 95.994 -59.764 53.077
534468 96.030 -59.770 53.047
534469 96.067 -59.774 53.017
534470 96.104 -59.779 52.989
534471 96.141 -59.785 52.965
534472 96.178 -59.790 52.942
534473 96.215 -59.795 52.921
534474 96.251 -59.801 52.899
534475 96.288 -59.806 52.874
534476 96.325 -59.811 52.848
534477 96.362 -59.817 52.824
534478 96.399 -59.822 52.800
534479 96.436 -59.826 52.776
534480 96.474 -59.827 52.750
534481 96.517 -59.832 52.720
534482 96.555 -59.833 52.687
534483 96.594 -59.833 52.654
534484 96.632 -59.834 52.620
534485 96.671 -59.835 52.586
534486 96.709 -59.836 52.554
534487 96.743 -59.835 52.525
534488 96.776 -59.834 52.494
534489 96.807 -59.833 52.463
534490 96.839 -59.831 52.430
534491 96.872 -59.831 52.394
534492 96.906 -59.833 52.359
534493 96.940 -59.836 52.323
534494 96.973 -59.838 52.285
534495 97.007 -59.840 52.244
534496 97.041 -59.842 52.198
534497 97.075 -59.845 52.149
534498 97.108 -59.847 52.097
534499 97.142 -59.849 52.047
534500 97.175 -59.853 52.000
534501 97.206 -59.866 51.954
534502 97.236 -59.879 51.908
534503 97.266 -59.892 51.864
534504 97.296 -59.904 51.821
534505 97.332 -59.907 51.776
534506 97.369 -59.908 51.730
534507 97.405 -59.908 51.683
534508 97.442 -59.909 51.635
534509 97.480 -59.914 51.588
534510 97.521 -59.925 51.540
534511 97.563 -59.936 51.491
534512 97.603 -59.947 51.443
534513 97.644 -59.958 51.396
534514 97.685 -59.969 51.349
534515 97.726 -59.980 51.304
534516 97.767 -59.992 51.259
534517 97.808 -60.003 51.215
534518 97.849 -60.014 51.172
534519 97.890 -60.025 51.130
534520 97.931 -60.036 51.088
534521 97.971 -60.047 51.046
534522 98.013 -60.058 51.005
534523 98.054 -60.070 50.965
534524 98.094 -60.081 50.924
534525 98.129 -60.085 50.884
534526 98.149 -60.090 50.846
534527 98.165 -60.089 50.806
534528 98.181 -60.087 50.767
534529 98.197 -60.086 50.726
534530 98.213 -60.085 50.686
534531 98.228 -60.083 50.644
534532 98.255 -60.086 50.601
534533 98.300 -60.075 50.559
534534 98.346 -60.058 50.516
534535 98.393 -60.042 50.474
534536 98.439 -60.025 50.432
534537 98.486 -60.008 50.390
534538 98.522 -60.011 50.347
534539 98.559 -60.015 50.305
534540 98.595 -60.020 50.263
534541 98.632 -60.024 50.220
534542 98.668 -60.028 50.177
534543 98.704 -60.033 50.134
534544 98.740 -60.037 50.090
534545 98.777 -60.041 50.047
534546 98.811 -60.028 50.004
534547 98.844 -60.005 49.961
534548 98.877 -59.983 49.918
534549 98.923 -59.970 49.875
534550 98.986 -59.967 49.832
534551 99.049 -59.966 49.790
534972 88.563 -57.460 56.441
534973 88.597 -57.472 56.417
534974 88.630 -57.484 56.393
534975 88.664 -57.495 56.369
534976 88.698 -57.507 56.346
534977 88.736 -57.523 56.322
534978 88.774 -57.539 56.299
534979 88.813 -57.554 56.276
534980 88.851 -57.570 56.253
534981 88.889 -57.586 56.230
534982 88.927 -57.602 56.205
534983 88.956 -57.624 56.182
534984 88.980 -57.648 56.160
534985 89.004 -57.672 56.137
534986 89.029 -57.696 56.112
534987 89.056 -57.717 56.089
534988 89.098 -57.729 56.064
534989 89.139 -57.739 56.040
534990 89.181 -57.750 56.015
534991 89.228 -57.766 55.990
534992 89.277 -57.783 55.967
534993 89.326 -57.800 55.944
534994 89.375 -57.817 55.922
534995 89.424 -57.833 55.903
534996 89.473 -57.851 55.886
534997 89.522 -57.868 55.872
534998 89.571 -57.884 55.859
534999 89.620 -57.902 55.849
535000 89.669 -57.918 55.840
535001 89.714 -57.933 55.833
535002 89.751 -57.943 55.828
535003 89.787 -57.953 55.823
535004 89.823 -57.963 55.819
535005 89.860 -57.973 55.815
535006 89.896 -57.983 55.810
535007 89.933 -57.993 55.806
535008 89.969 -58.003 55.803
535009 90.010 -58.010 55.794
535010 90.055 -58.015 55.781
535011 90.100 -58.019 55.763
535012 90.146 -58.024 55.741
535013 90.182 -58.037 55.722
535014 90.215 -58.054 55.704
535015 90.248 -58.066 55.687
535016 90.282 -58.078 55.671
535017 90.316 -58.090 55.656
535018 90.350 -58.101 55.638
535019 90.386 -58.110 55.611
535020 90.428 -58.115 55.590
535021 90.470 -58.119 55.567
535022 90.512 -58.124 55.545
535023 90.554 -58.128 55.524
535024 90.597 -58.133 55.511
535025 90.639 -58.137 55.494
535026 90.681 -58.142 55.477
535027 90.723 -58.147 55.458
535028 90.762 -58.153 55.435
535029 90.800 -58.161 55.407
535030 90.837 -58.168 55.373
535031 90.876 -58.178 55.335
535032 90.918 -58.197 55.301
535033 90.959 -58.216 55.275
535034 91.001 -58.234 55.254
535035 91.042 -58.253 55.234
535036 91.082 -58.277 55.208
535037 91.121 -58.301 55.186
535038 91.161 -58.325 55.167
535039 91.201 -58.350 55.152
535040 91.241 -58.373 55.139
535041 91.280 -58.397 55.127
535042 91.320 -58.422 55.109
535043 91.360 -58.446 55.087
535044 91.399 -58.469 55.063
535045 91.439 -58.494 55.042
535046 91.477 -58.507 55.024
535047 91.515 -58.519 55.015
535048 91.554 -58.531 55.007
535049 91.592 -58.543 55.000
535050 91.627 -58.554 54.995
535051 91.659 -58.565 54.988
535052 91.692 -58.576 54.981
535053 91.728 -58.585 54.972
535054 91.764 -58.595 54.957
535055 91.799 -58.604 54.935
535056 91.836 -58.614 54.907
535057 91.871 -58.623 54.882
535058 91.907 -58.632 54.855
535059 91.943 -58.642 54.830
535060 91.979 -58.656 54.807
535061 92.015 -58.670 54.781
535062 92.051 -58.685 54.751
535063 92.087 -58.698 54.734
535064 92.123 -58.698 54.723
535065 92.158 -58.699 54.716
535066 92.194 -58.699 54.712
535067 92.229 -58.700 54.714
535068 92.265 -58.700 54.721
535069 92.300 -58.701 54.727
535070 92.335 -58.701 54.732
535071 92.371 -58.701 54.736
535072 92.406 -58.701 54.738
535073 92.442 -58.702 54.733
535074 92.477 -58.703 54.728
535075 92.516 -58.706 54.721
535076 92.555 -58.711 54.715
535077 92.600 -58.717 54.711
535078 92.646 -58.723 54.706
535079 92.692 -58.729 54.701
535080 92.737 -58.735 54.694
535081 92.783 -58.741 54.688
535082 92.829 -58.747 54.684
535083 92.874 -58.753 54.679
535084 92.920 -58.758 54.674
535085 92.966 -58.764 54.673
535086 93.011 -58.770 54.669
535087 93.052 -58.779 54.660
535088 93.090 -58.789 54.649
535089 93.128 -58.798 54.634
535090 93.166 -58.808 54.615
535091 93.204 -58.816 54.598
535092 93.243 -58.824 54.581
535093 93.282 -58.831 54.567
535094 93.321 -58.839 54.555
535095 93.359 -58.847 54.540
535096 93.398 -58.855 54.524
535097 93.437 -58.862 54.507
535098 93.476 -58.870 54.488
535099 93.514 -58.877 54.469
535100 93.551 -58.889 54.449
535101 93.587 -58.903 54.432
535102 93.624 -58.917 54.416
535103 93.660 -58.930 54.401
535104 93.696 -58.944 54.386
535105 93.732 -58.958 54.372
535106 93.769 -58.971 54.358
535107 93.805 -58.985 54.345
535108 93.842 -58.999 54.327
535109 93.877 -59.014 54.304
535110 93.913 -59.034 54.279
535111 93.947 -59.054 54.251
535112 93.983 -59.074 54.227
535113 94.017 -59.095 54.203
535114 94.053 -59.115 54.179
535115 94.087 -59.135 54.156
535116 94.123 -59.156 54.138
535117 94.157 -59.176 54.121
535118 94.192 -59.196 54.102
535119 94.227 -59.216 54.081
535120 94.262 -59.237 54.060
535121 94.297 -59.257 54.038
535122 94.332 -59.277 54.015
535123 94.367 -59.297 53.992
535124 94.402 -59.318 53.969
535125 94.437 -59.338 53.946
535126 94.472 -59.350 53.924
535127 94.505 -59.357 53.902
535128 94.540 -59.364 53.884
535129 94.573 -59.372 53.873
535130 94.608 -59.378 53.862
535131 94.641 -59.385 53.845
535132 94.678 -59.394 53.822
535133 94.719 -59.403 53.790
535134 94.760 -59.413 53.751
535135 94.801 -59.422 53.711
535136 94.841 -59.432 53.673
535137 94.882 -59.441 53.639
535138 94.923 -59.450 53.608
535139 94.964 -59.459 53.582
535140 95.004 -59.468 53.556
535141 95.045 -59.478 53.530
535142 95.085 -59.488 53.505
535143 95.122 -59.498 53.479
535144 95.161 -59.510 53.454
535145 95.200 -59.520 53.430
535146 95.239 -59.532 53.409
535147 95.279 -59.544 53.390
535148 95.318 -59.557 53.372
535149 95.358 -59.569 53.353
535150 95.399 -59.579 53.334
535151 95.435 -59.589 53.316
535152 95.468 -59.599 53.299
535153 95.500 -59.609 53.284
535154 95.533 -59.620 53.270
535155 95.566 -59.629 53.256
535156 95.599 -59.640 53.255
535157 95.631 -59.650 53.256
535158 95.664 -59.660 53.256
535159 95.696 -59.670 53.253
535160 95.730 -59.676 53.246
535161 95.764 -59.682 53.231
535162 95.798 -59.689 53.213
535163 95.831 -59.697 53.193
535164 95.865 -59.704 53.173
535165 95.898 -59.710 53.154
535166 95.932 -59.717 53.134
535167 95.966 -59.724 53.109
535168 96.000 -59.731 53.081
535169 96.034 -59.737 53.052
535170 96.071 -59.742 53.024
535171 96.108 -59.748 53.000
535172 96.145 -59.753 52.979
535173 96.182 -59.758 52.957
535174 96.219 -59.764 52.935
535175 96.256 -59.768 52.911
535176 96.293 -59.773 52.888
535177 96.330 -59.779 52.862
535178 96.367 -59.784 52.836
535179 96.403 -59.789 52.812
535180 96.441 -59.795 52.787
535181 96.478 -59.798 52.760
535182 96.517 -59.799 52.729
535183 96.555 -59.800 52.694
535184 96.593 -59.801 52.663
535185 96.631 -59.801 52.631
535186 96.670 -59.802 52.598
535187 96.708 -59.802 52.567
535188 96.740 -59.801 52.538
535189 96.772 -59.800 52.507
535190 96.804 -59.798 52.476
535191 96.836 -59.798 52.444
535192 96.869 -59.797 52.408
535193 96.902 -59.800 52.373
535194 96.936 -59.802 52.336
535195 96.970 -59.804 52.297
535196 97.004 -59.807 52.255
535197 97.037 -59.809 52.208
535198 97.071 -59.811 52.160
535199 97.105 -59.814 52.111
535200 97.139 -59.816 52.065
535201 97.172 -59.818 52.020
535202 97.205 -59.824 51.975
535203 97.235 -59.837 51.932
535204 97.265 -59.850 51.891
535205 97.296 -59.862 51.849
535206 97.329 -59.868 51.806
535207 97.366 -59.869 51.760
535208 97.403 -59.870 51.712
535209 97.440 -59.870 51.662
535210 97.480 -59.878 51.615
535211 97.521 -59.889 51.565
535212 97.562 -59.901 51.515
535213 97.602 -59.911 51.465
535214 97.644 -59.923 51.415
535215 97.684 -59.934 51.367
535216 97.725 -59.946 51.318
535217 97.766 -59.956 51.271
535218 97.807 -59.967 51.225
535219 97.848 -59.979 51.180
535220 97.889 -59.989 51.137
535221 97.930 -60.001 51.095
535222 97.971 -60.012 51.052
535223 98.011 -60.024 51.010
535224 98.051 -60.034 50.969
535225 98.090 -60.046 50.929
535226 98.130 -60.046 50.889
535227 98.153 -60.048 50.850
535228 98.174 -60.054 50.811
535229 98.196 -60.061 50.772
535230 98.212 -60.060 50.732
535231 98.228 -60.058 50.692
535232 98.244 -60.057 50.651
535233 98.260 -60.056 50.610
535234 98.299 -60.053 50.568
535235 98.345 -60.036 50.526
535236 98.391 -60.020 50.484
535237 98.438 -60.002 50.443
535238 98.485 -59.986 50.401
535239 98.529 -59.974 50.358
535240 98.565 -59.978 50.314
535241 98.601 -59.982 50.273
535242 98.638 -59.986 50.231
535243 98.674 -59.991 50.188
535244 98.710 -59.996 50.145
535245 98.746 -59.999 50.101
535246 98.783 -60.004 50.058
535247 98.816 -59.987 50.014
535248 98.850 -59.964 49.971
535249 98.882 -59.942 49.928
535250 98.932 -59.930 49.885
535251 98.995 -59.929 49.842
535252 99.055 -59.927 49.799
535674 88.609 -57.428 56.402
535675 88.642 -57.440 56.379
535676 88.677 -57.452 56.356
535677 88.715 -57.468 56.333
535678 88.754 -57.484 56.311
535679 88.792 -57.500 56.289
535680 88.830 -57.516 56.267
535681 88.869 -57.532 56.244
535682 88.907 -57.548 56.221
535683 88.945 -57.564 56.197
535684 88.974 -57.585 56.173
535685 88.998 -57.609 56.150
535686 89.023 -57.633 56.128
535687 89.047 -57.658 56.107
535688 89.076 -57.678 56.085
535689 89.117 -57.689 56.060
535690 89.159 -57.700 56.036
535691 89.200 -57.711 56.013
535692 89.242 -57.721 55.988
535693 89.290 -57.738 55.966
535694 89.339 -57.755 55.944
535695 89.388 -57.772 55.923
535696 89.437 -57.789 55.904
535697 89.486 -57.805 55.887
535698 89.535 -57.823 55.873
535699 89.584 -57.840 55.861
535700 89.633 -57.856 55.851
535701 89.682 -57.874 55.842
535702 89.730 -57.890 55.835
535703 89.766 -57.899 55.830
535704 89.803 -57.909 55.828
535705 89.839 -57.920 55.824
535706 89.875 -57.930 55.822
535707 89.912 -57.940 55.820
535708 89.948 -57.949 55.818
535709 89.984 -57.959 55.817
535710 90.021 -57.969 55.811
535711 90.060 -57.977 55.799
535712 90.106 -57.982 55.780
535713 90.151 -57.987 55.756
535714 90.193 -57.994 55.736
535715 90.225 -58.012 55.718
535716 90.259 -58.024 55.700
535717 90.293 -58.036 55.686
535718 90.326 -58.047 55.672
535719 90.361 -58.059 55.653
535720 90.394 -58.070 55.625
535721 90.434 -58.078 55.602
535722 90.475 -58.082 55.578
535723 90.518 -58.087 55.555
535724 90.559 -58.091 55.536
535725 90.602 -58.096 55.523
535726 90.644 -58.100 55.506
535727 90.686 -58.105 55.488
535728 90.728 -58.109 55.467
535729 90.770 -58.114 55.440
535730 90.812 -58.118 55.408
535731 90.851 -58.125 55.373
535732 90.888 -58.133 55.336
535733 90.929 -58.147 55.300
535734 90.970 -58.165 55.268
535735 91.011 -58.184 55.244
535736 91.051 -58.207 55.220
535737 91.091 -58.231 55.200
535738 91.131 -58.255 55.185
535739 91.170 -58.279 55.175
535740 91.210 -58.303 55.166
535741 91.250 -58.328 55.157
535742 91.290 -58.351 55.146
535743 91.329 -58.375 55.129
535744 91.369 -58.400 55.106
535745 91.409 -58.424 55.081
535746 91.448 -58.448 55.056
535747 91.488 -58.472 55.035
535748 91.526 -58.485 55.024
535749 91.565 -58.497 55.017
535750 91.602 -58.510 55.012
535751 91.637 -58.520 55.008
535752 91.669 -58.531 55.003
535753 91.701 -58.542 54.997
535754 91.737 -58.551 54.987
535755 91.773 -58.560 54.971
535756 91.808 -58.570 54.950
535757 91.844 -58.579 54.922
535758 91.880 -58.589 54.895
535759 91.916 -58.598 54.864
535760 91.951 -58.608 54.836
535761 91.988 -58.617 54.812
535762 92.023 -58.629 54.788
535763 92.060 -58.643 54.768
535764 92.095 -58.657 54.750
535765 92.131 -58.660 54.736
535766 92.166 -58.661 54.725
535767 92.202 -58.661 54.718
535768 92.237 -58.661 54.719
535769 92.273 -58.661 54.726
535770 92.308 -58.662 54.730
535771 92.344 -58.662 54.733
535772 92.381 -58.664 54.737
535773 92.420 -58.667 54.744
535774 92.459 -58.671 54.744
535775 92.498 -58.674 54.743
535776 92.533 -58.678 54.742
535777 92.566 -58.683 54.740
535778 92.612 -58.689 54.739
535779 92.657 -58.695 54.733
535780 92.703 -58.701 54.725
535781 92.748 -58.708 54.716
535782 92.794 -58.713 54.711
535783 92.839 -58.719 54.708
535784 92.885 -58.725 54.706
535785 92.931 -58.731 54.705
535786 92.976 -58.737 54.702
535787 93.022 -58.743 54.695
535788 93.060 -58.752 54.685
535789 93.098 -58.762 54.674
535790 93.136 -58.771 54.662
535791 93.174 -58.781 54.644
535792 93.212 -58.790 54.625
535793 93.251 -58.798 54.608
535794 93.290 -58.805 54.593
535795 93.329 -58.813 54.584
535796 93.367 -58.821 54.573
535797 93.406 -58.829 54.558
535798 93.444 -58.836 54.540
535799 93.483 -58.844 54.519
535800 93.522 -58.852 54.496
535801 93.559 -58.864 54.472
535802 93.595 -58.878 54.452
535803 93.631 -58.892 54.436
535804 93.667 -58.905 54.422
535805 93.704 -58.919 54.410
535806 93.740 -58.933 54.399
535807 93.776 -58.946 54.388
535808 93.813 -58.960 54.377
535809 93.849 -58.974 54.359
535810 93.886 -58.988 54.336
535811 93.922 -59.002 54.309
535812 93.957 -59.020 54.276
535813 93.992 -59.040 54.248
535814 94.027 -59.060 54.223
535815 94.062 -59.081 54.199
535816 94.097 -59.101 54.177
535817 94.132 -59.121 54.159
535818 94.167 -59.141 54.144
535819 94.202 -59.162 54.126
535820 94.237 -59.182 54.106
535821 94.272 -59.202 54.084
535822 94.307 -59.222 54.063
535823 94.342 -59.243 54.039
535824 94.377 -59.263 54.016
535825 94.412 -59.283 53.992
535826 94.447 -59.303 53.968
535827 94.481 -59.313 53.943
535828 94.515 -59.321 53.918
535829 94.549 -59.328 53.894
535830 94.583 -59.334 53.877
535831 94.617 -59.342 53.861
535832 94.652 -59.350 53.842
535833 94.691 -59.358 53.820
535834 94.732 -59.368 53.790
535835 94.773 -59.378 53.754
535836 94.814 -59.388 53.715
535837 94.855 -59.397 53.676
535838 94.896 -59.406 53.640
535839 94.936 -59.415 53.608
535840 94.977 -59.425 53.576
535841 95.018 -59.434 53.545
535842 95.059 -59.443 53.517
535843 95.096 -59.454 53.490
535844 95.133 -59.465 53.465
535845 95.172 -59.476 53.445
535846 95.211 -59.487 53.424
535847 95.250 -59.498 53.401
535848 95.289 -59.510 53.380
535849 95.328 -59.520 53.362
535850 95.367 -59.532 53.343
535851 95.406 -59.542 53.326
535852 95.439 -59.553 53.311
535853 95.471 -59.563 53.295
535854 95.505 -59.574 53.279
535855 95.537 -59.585 53.262
535856 95.570 -59.595 53.253
535857 95.603 -59.605 53.255
535858 95.636 -59.616 53.261
535859 95.669 -59.626 53.265
535860 95.702 -59.637 53.266
535861 95.734 -59.645 53.260
535862 95.768 -59.651 53.245
535863 95.802 -59.658 53.225
535864 95.835 -59.665 53.203
535865 95.869 -59.672 53.180
535866 95.902 -59.679 53.159
535867 95.936 -59.686 53.138
535868 95.970 -59.693 53.113
535869 96.004 -59.699 53.086
535870 96.039 -59.705 53.059
535871 96.075 -59.710 53.033
535872 96.112 -59.716 53.012
535873 96.149 -59.721 52.993
535874 96.186 -59.726 52.972
535875 96.223 -59.732 52.949
535876 96.260 -59.737 52.926
535877 96.297 -59.742 52.901
535878 96.334 -59.747 52.874
535879 96.371 -59.752 52.847
535880 96.407 -59.757 52.822
535881 96.443 -59.761 52.796
535882 96.479 -59.764 52.768
535883 96.516 -59.766 52.736
535884 96.554 -59.767 52.703
535885 96.593 -59.767 52.673
535886 96.631 -59.768 52.643
535887 96.669 -59.769 52.612
535888 96.704 -59.769 52.581
535889 96.736 -59.767 52.550
535890 96.768 -59.766 52.519
535891 96.800 -59.765 52.489
535892 96.832 -59.764 52.457
535893 96.865 -59.764 52.422
535894 96.899 -59.766 52.386
535895 96.933 -59.768 52.347
535896 96.966 -59.770 52.306
535897 97.000 -59.773 52.262
535898 97.034 -59.775 52.217
535899 97.068 -59.778 52.169
535900 97.101 -59.780 52.123
535901 97.135 -59.783 52.081
535902 97.169 -59.785 52.038
535903 97.203 -59.787 51.995
535904 97.234 -59.795 51.953
535905 97.264 -59.808 51.914
535906 97.295 -59.821 51.875
535907 97.327 -59.830 51.835
535908 97.364 -59.830 51.791
535909 97.401 -59.830 51.744
535910 97.438 -59.832 51.694
535911 97.479 -59.843 51.644
535912 97.520 -59.854 51.592
535913 97.561 -59.865 51.542
535914 97.602 -59.877 51.491
535915 97.643 -59.888 51.440
535916 97.683 -59.898 51.388
535917 97.725 -59.910 51.337
535918 97.766 -59.921 51.287
535919 97.806 -59.932 51.238
535920 97.847 -59.943 51.190
535921 97.888 -59.955 51.144
535922 97.928 -59.965 51.100
535923 97.967 -59.977 51.056
535924 98.008 -59.988 51.013
535925 98.048 -59.999 50.972
535926 98.087 -60.008 50.931
535927 98.129 -60.011 50.893
535928 98.155 -60.004 50.854
535929 98.177 -60.010 50.815
535930 98.199 -60.017 50.776
535931 98.220 -60.023 50.737
535932 98.241 -60.030 50.696
535933 98.260 -60.031 50.656
535934 98.275 -60.030 50.616
535935 98.298 -60.031 50.575
535936 98.344 -60.014 50.536
535937 98.390 -59.997 50.497
535938 98.437 -59.980 50.457
535939 98.484 -59.964 50.414
535940 98.530 -59.947 50.370
535941 98.571 -59.940 50.326
535942 98.607 -59.945 50.284
535943 98.644 -59.949 50.242
535944 98.680 -59.954 50.200
535945 98.716 -59.958 50.157
535946 98.753 -59.962 50.112
535947 98.789 -59.967 50.069
535948 98.822 -59.946 50.025
535949 98.855 -59.924 49.981
535950 98.888 -59.901 49.938
535951 98.940 -59.892 49.895
535952 99.003 -59.890 49.851
535953 99.054 -59.890 49.808
535954 99.082 -59.894 49.766
536376 88.657 -57.398 56.364
536377 88.695 -57.413 56.342
536378 88.733 -57.429 56.319
536379 88.771 -57.445 56.298
536380 88.810 -57.461 56.277
536381 88.848 -57.477 56.255
536382 88.886 -57.493 56.234
536383 88.924 -57.510 56.211
536384 88.963 -57.526 56.189
536385 88.993 -57.546 56.166
536386 89.017 -57.570 56.142
536387 89.042 -57.594 56.120
536388 89.066 -57.618 56.099
536389 89.095 -57.639 56.077
536390 89.137 -57.649 56.053
536391 89.178 -57.660 56.030
536392 89.219 -57.671 56.007
536393 89.261 -57.682 55.984
536394 89.303 -57.693 55.964
536395 89.352 -57.710 55.943
536396 89.401 -57.727 55.922
536397 89.450 -57.744 55.904
536398 89.499 -57.761 55.887
536399 89.548 -57.777 55.873
536400 89.597 -57.795 55.862
536401 89.646 -57.811 55.852
536402 89.695 -57.829 55.843
536403 89.744 -57.846 55.837
536404 89.782 -57.856 55.833
536405 89.818 -57.866 55.831
536406 89.855 -57.876 55.828
536407 89.891 -57.886 55.827
536408 89.928 -57.896 55.826
536409 89.964 -57.906 55.825
536410 90.000 -57.916 55.824
536411 90.036 -57.926 55.818
536412 90.073 -57.936 55.807
536413 90.112 -57.944 55.791
536414 90.157 -57.949 55.772
536415 90.202 -57.953 55.757
536416 90.236 -57.970 55.739
536417 90.269 -57.982 55.717
536418 90.303 -57.993 55.698
536419 90.337 -58.006 55.680
536420 90.371 -58.017 55.659
536421 90.405 -58.028 55.632
536422 90.439 -58.040 55.608
536423 90.481 -58.044 55.585
536424 90.523 -58.049 55.563
536425 90.565 -58.053 55.547
536426 90.607 -58.058 55.532
536427 90.649 -58.062 55.515
536428 90.691 -58.067 55.494
536429 90.733 -58.071 55.469
536430 90.776 -58.076 55.439
536431 90.818 -58.081 55.404
536432 90.860 -58.085 55.369
536433 90.901 -58.090 55.333
536434 90.939 -58.098 55.297
536435 90.980 -58.116 55.264
536436 91.021 -58.137 55.240
536437 91.061 -58.161 55.219
536438 91.100 -58.185 55.202
536439 91.140 -58.209 55.191
536440 91.180 -58.234 55.185
536441 91.220 -58.257 55.181
536442 91.259 -58.281 55.176
536443 91.299 -58.306 55.164
536444 91.338 -58.330 55.147
536445 91.378 -58.354 55.125
536446 91.418 -58.378 55.102
536447 91.458 -58.402 55.074
536448 91.498 -58.426 55.051
536449 91.537 -58.450 55.034
536450 91.575 -58.463 55.023
536451 91.613 -58.476 55.018
536452 91.646 -58.486 55.018
536453 91.678 -58.497 55.017
536454 91.711 -58.507 55.011
536455 91.745 -58.517 54.998
536456 91.781 -58.526 54.980
536457 91.817 -58.536 54.958
536458 91.853 -58.545 54.932
536459 91.889 -58.555 54.903
536460 91.925 -58.564 54.872
536461 91.960 -58.574 54.847
536462 91.996 -58.583 54.828
536463 92.032 -58.593 54.811
536464 92.068 -58.602 54.799
536465 92.104 -58.616 54.781
536466 92.140 -58.621 54.759
536467 92.175 -58.622 54.739
536468 92.211 -58.622 54.726
536469 92.246 -58.623 54.725
536470 92.285 -58.626 54.733
536471 92.324 -58.629 54.741
536472 92.363 -58.632 54.745
536473 92.402 -58.636 54.745
536474 92.441 -58.639 54.745
536475 92.480 -58.642 54.749
536476 92.518 -58.645 54.748
536477 92.548 -58.651 54.747
536478 92.577 -58.656 54.747
536479 92.622 -58.662 54.748
536480 92.667 -58.668 54.747
536481 92.713 -58.673 54.748
536482 92.759 -58.679 54.750
536483 92.804 -58.686 54.752
536484 92.850 -58.692 54.752
536485 92.896 -58.698 54.747
536486 92.941 -58.704 54.743
536487 92.987 -58.709 54.735
536488 93.030 -58.717 54.724
536489 93.068 -58.726 54.711
536490 93.106 -58.736 54.699
536491 93.144 -58.745 54.687
536492 93.182 -58.755 54.669
536493 93.220 -58.764 54.651
536494 93.259 -58.772 54.632
536495 93.297 -58.780 54.617
536496 93.336 -58.787 54.606
536497 93.375 -58.795 54.597
536498 93.414 -58.803 54.583
536499 93.452 -58.811 54.566
536500 93.491 -58.818 54.544
536501 93.530 -58.826 54.519
536502 93.566 -58.839 54.495
536503 93.603 -58.853 54.474
536504 93.639 -58.867 54.458
536505 93.675 -58.881 54.445
536506 93.711 -58.895 54.435
536507 93.748 -58.908 54.425
536508 93.784 -58.922 54.417
536509 93.820 -58.936 54.404
536510 93.856 -58.949 54.387
536511 93.893 -58.963 54.365
536512 93.929 -58.977 54.338
536513 93.966 -58.990 54.303
536514 94.001 -59.005 54.271
536515 94.037 -59.026 54.243
536516 94.071 -59.046 54.220
536517 94.106 -59.066 54.200
536518 94.141 -59.087 54.182
536519 94.176 -59.107 54.168
536520 94.211 -59.128 54.152
536521 94.246 -59.147 54.133
536522 94.281 -59.168 54.112
536523 94.316 -59.188 54.091
536524 94.351 -59.209 54.067
536525 94.386 -59.228 54.041
536526 94.421 -59.249 54.016
536527 94.456 -59.269 53.990
536528 94.490 -59.277 53.962
536529 94.524 -59.284 53.935
536530 94.558 -59.291 53.907
536531 94.592 -59.298 53.883
536532 94.626 -59.306 53.860
536533 94.665 -59.314 53.837
536534 94.704 -59.323 53.813
536535 94.745 -59.333 53.784
536536 94.786 -59.343 53.750
536537 94.828 -59.353 53.713
536538 94.868 -59.363 53.674
536539 94.909 -59.372 53.637
536540 94.950 -59.381 53.604
536541 94.991 -59.390 53.573
536542 95.032 -59.399 53.543
536543 95.070 -59.410 53.515
536544 95.107 -59.421 53.487
536545 95.144 -59.432 53.461
536546 95.182 -59.443 53.435
536547 95.222 -59.453 53.411
536548 95.260 -59.465 53.389
536549 95.300 -59.475 53.368
536550 95.338 -59.487 53.349
536551 95.378 -59.498 53.332
536552 95.411 -59.510 53.317
536553 95.445 -59.521 53.301
536554 95.477 -59.532 53.284
536555 95.511 -59.541 53.269
536556 95.543 -59.552 53.261
536557 95.576 -59.563 53.260
536558 95.609 -59.573 53.262
536559 95.642 -59.584 53.266
536560 95.675 -59.595 53.271
536561 95.707 -59.604 53.272
536562 95.739 -59.613 53.265
536563 95.772 -59.620 53.252
536564 95.806 -59.627 53.233
536565 95.839 -59.634 53.213
536566 95.873 -59.641 53.191
536567 95.906 -59.648 53.168
536568 95.940 -59.654 53.144
536569 95.974 -59.661 53.119
536570 96.009 -59.667 53.094
536571 96.044 -59.673 53.069
536572 96.080 -59.679 53.046
536573 96.116 -59.684 53.026
536574 96.153 -59.689 53.009
536575 96.190 -59.695 52.988
536576 96.227 -59.700 52.965
536577 96.264 -59.705 52.941
536578 96.301 -59.710 52.915
536579 96.338 -59.715 52.885
536580 96.375 -59.720 52.857
536581 96.410 -59.723 52.830
536582 96.445 -59.726 52.803
536583 96.480 -59.729 52.776
536584 96.516 -59.732 52.744
536585 96.553 -59.734 52.713
536586 96.592 -59.735 52.684
536587 96.630 -59.736 52.655
536588 96.669 -59.736 52.625
536589 96.700 -59.735 52.595
536590 96.733 -59.733 52.563
536591 96.764 -59.732 52.533
536592 96.797 -59.731 52.502
536593 96.829 -59.729 52.469
536594 96.862 -59.730 52.435
536595 96.895 -59.732 52.398
536596 96.929 -59.735 52.357
536597 96.963 -59.737 52.315
536598 96.997 -59.739 52.271
536599 97.030 -59.742 52.226
536600 97.064 -59.744 52.181
536601 97.098 -59.746 52.138
536602 97.132 -59.748 52.097
536603 97.165 -59.751 52.055
536604 97.199 -59.753 52.013
536605 97.233 -59.756 51.974
536606 97.264 -59.767 51.936
536607 97.294 -59.779 51.899
536608 97.325 -59.790 51.860
536609 97.362 -59.791 51.818
536610 97.399 -59.792 51.772
536611 97.437 -59.796 51.723
536612 97.478 -59.808 51.674
536613 97.519 -59.819 51.622
536614 97.560 -59.830 51.570
536615 97.601 -59.841 51.519
536616 97.642 -59.852 51.466
536617 97.683 -59.864 51.412
536618 97.724 -59.874 51.358
536619 97.764 -59.886 51.306
536620 97.805 -59.897 51.255
536621 97.845 -59.908 51.205
536622 97.885 -59.919 51.157
536623 97.925 -59.930 51.109
536624 97.964 -59.941 51.062
536625 98.004 -59.952 51.017
536626 98.044 -59.964 50.974
536627 98.085 -59.971 50.932
536628 98.127 -59.974 50.894
536629 98.157 -59.962 50.855
536630 98.179 -59.966 50.816
536631 98.201 -59.973 50.777
536632 98.223 -59.979 50.738
536633 98.244 -59.986 50.698
536634 98.266 -59.992 50.658
536635 98.287 -59.999 50.619
536636 98.307 -60.002 50.583
536637 98.343 -59.992 50.547
536638 98.389 -59.975 50.511
536639 98.436 -59.958 50.472
536640 98.482 -59.942 50.429
536641 98.528 -59.925 50.385
536642 98.575 -59.908 50.341
536643 98.614 -59.908 50.297
536644 98.650 -59.912 50.255
536645 98.687 -59.916 50.212
536646 98.723 -59.920 50.168
536647 98.759 -59.925 50.124
536648 98.795 -59.928 50.080
536649 98.828 -59.905 50.036
536650 98.861 -59.883 49.992
536651 98.894 -59.861 49.948
536652 98.949 -59.853 49.904
536653 99.012 -59.851 49.861
536654 99.052 -59.853 49.818
536655 99.081 -59.857 49.776
536656 99.110 -59.861 49.734
537078 88.713 -57.375 56.327
537079 88.751 -57.391 56.305
537080 88.789 -57.407 56.284
537081 88.828 -57.423 56.263
537082 88.866 -57.439 56.242
537083 88.904 -57.455 56.222
537084 88.943 -57.471 56.200
537085 88.981 -57.487 56.179
537086 89.012 -57.507 56.158
537087 89.036 -57.531 56.136
537088 89.060 -57.555 56.112
537089 89.085 -57.579 56.091
537090 89.115 -57.598 56.069
537091 89.156 -57.610 56.047
537092 89.198 -57.620 56.025
537093 89.239 -57.631 56.003
537094 89.280 -57.642 55.980
537095 89.322 -57.653 55.961
537096 89.365 -57.665 55.940
537097 89.414 -57.682 55.920
537098 89.463 -57.699 55.902
537099 89.512 -57.716 55.886
537100 89.561 -57.733 55.873
537101 89.610 -57.750 55.862
537102 89.659 -57.767 55.853
537103 89.708 -57.783 55.845
537104 89.757 -57.801 55.838
537105 89.798 -57.813 55.835
537106 89.834 -57.823 55.832
537107 89.870 -57.833 55.830
537108 89.906 -57.843 55.830
537109 89.943 -57.852 55.829
537110 89.979 -57.862 55.829
537111 90.016 -57.873 55.825
537112 90.052 -57.883 55.820
537113 90.089 -57.893 55.810
537114 90.125 -57.902 55.797
537115 90.163 -57.912 55.785
537116 90.208 -57.916 55.773
537117 90.246 -57.927 55.755
537118 90.280 -57.940 55.731
537119 90.314 -57.952 55.706
537120 90.347 -57.963 55.682
537121 90.381 -57.974 55.657
537122 90.415 -57.986 55.630
537123 90.449 -57.998 55.606
537124 90.486 -58.007 55.584
537125 90.528 -58.012 55.565
537126 90.570 -58.016 55.550
537127 90.612 -58.021 55.534
537128 90.655 -58.025 55.514
537129 90.696 -58.030 55.489
537130 90.739 -58.034 55.461
537131 90.781 -58.039 55.429
537132 90.823 -58.043 55.394
537133 90.865 -58.048 55.359
537134 90.907 -58.053 55.324
537135 90.949 -58.057 55.291
537136 90.991 -58.067 55.262
537137 91.030 -58.091 55.239
537138 91.070 -58.115 55.220
537139 91.109 -58.139 55.205
537140 91.149 -58.163 55.195
537141 91.189 -58.187 55.191
537142 91.229 -58.212 55.190
537143 91.269 -58.235 55.187
537144 91.308 -58.260 55.175
537145 91.348 -58.284 55.158
537146 91.387 -58.308 55.137
537147 91.427 -58.332 55.115
537148 91.467 -58.356 55.090
537149 91.507 -58.380 55.069
537150 91.546 -58.404 55.052
537151 91.586 -58.428 55.040
537152 91.623 -58.441 55.032
537153 91.655 -58.452 55.026
537154 91.688 -58.463 55.022
537155 91.720 -58.473 55.014
537156 91.754 -58.483 55.002
537157 91.790 -58.492 54.984
537158 91.826 -58.502 54.959
537159 91.862 -58.511 54.932
537160 91.897 -58.521 54.905
537161 91.933 -58.530 54.880
537162 91.969 -58.540 54.862
537163 92.005 -58.549 54.851
537164 92.041 -58.559 54.842
537165 92.077 -58.568 54.834
537166 92.112 -58.578 54.814
537167 92.151 -58.585 54.785
537168 92.190 -58.588 54.756
537169 92.228 -58.592 54.735
537170 92.267 -58.595 54.728
537171 92.306 -58.598 54.733
537172 92.345 -58.601 54.743
537173 92.384 -58.604 54.749
537174 92.423 -58.607 54.750
537175 92.461 -58.610 54.745
537176 92.500 -58.614 54.749
537177 92.533 -58.618 54.753
537178 92.563 -58.623 54.758
537179 92.592 -58.629 54.764
537180 92.633 -58.634 54.770
537181 92.678 -58.640 54.774
537182 92.724 -58.646 54.777
537183 92.770 -58.652 54.778
537184 92.815 -58.658 54.781
537185 92.861 -58.664 54.784
537186 92.907 -58.670 54.780
537187 92.952 -58.676 54.780
537188 92.998 -58.682 54.772
537189 93.039 -58.690 54.761
537190 93.076 -58.700 54.746
537191 93.115 -58.709 54.725
537192 93.152 -58.719 54.705
537193 93.190 -58.729 54.691
537194 93.228 -58.738 54.673
537195 93.267 -58.746 54.655
537196 93.305 -58.754 54.639
537197 93.344 -58.762 54.623
537198 93.383 -58.770 54.613
537199 93.422 -58.777 54.601
537200 93.461 -58.785 54.584
537201 93.499 -58.792 54.564
537202 93.538 -58.802 54.539
537203 93.574 -58.815 54.515
537204 93.610 -58.829 54.495
537205 93.646 -58.842 54.480
537206 93.683 -58.856 54.470
537207 93.719 -58.870 54.462
537208 93.755 -58.883 54.453
537209 93.791 -58.897 54.445
537210 93.828 -58.911 54.431
537211 93.864 -58.924 54.412
537212 93.900 -58.938 54.388
537213 93.936 -58.952 54.361
537214 93.973 -58.965 54.326
537215 94.009 -58.979 54.295
537216 94.046 -58.993 54.267
537217 94.081 -59.012 54.244
537218 94.116 -59.032 54.227
537219 94.151 -59.052 54.210
537220 94.186 -59.072 54.193
537221 94.221 -59.093 54.179
537222 94.256 -59.113 54.163
537223 94.290 -59.134 54.144
537224 94.326 -59.153 54.122
537225 94.360 -59.174 54.100
537226 94.396 -59.194 54.071
537227 94.430 -59.215 54.041
537228 94.466 -59.234 54.011
537229 94.499 -59.240 53.983
537230 94.534 -59.247 53.953
537231 94.567 -59.255 53.923
537232 94.602 -59.262 53.892
537233 94.639 -59.270 53.861
537234 94.677 -59.278 53.831
537235 94.717 -59.287 53.802
537236 94.758 -59.297 53.773
537237 94.800 -59.308 53.741
537238 94.841 -59.318 53.707
537239 94.882 -59.328 53.671
537240 94.923 -59.337 53.635
537241 94.964 -59.347 53.601
537242 95.004 -59.356 53.573
537243 95.044 -59.366 53.545
537244 95.082 -59.376 53.517
537245 95.119 -59.388 53.489
537246 95.156 -59.398 53.460
537247 95.193 -59.409 53.428
537248 95.232 -59.420 53.403
537249 95.271 -59.432 53.381
537250 95.310 -59.442 53.360
537251 95.349 -59.453 53.340
537252 95.383 -59.465 53.323
537253 95.417 -59.476 53.308
537254 95.450 -59.488 53.292
537255 95.483 -59.499 53.278
537256 95.517 -59.510 53.269
537257 95.549 -59.520 53.266
537258 95.582 -59.531 53.266
537259 95.615 -59.541 53.267
537260 95.648 -59.552 53.268
537261 95.681 -59.563 53.270
537262 95.713 -59.572 53.270
537263 95.744 -59.580 53.265
537264 95.776 -59.588 53.254
537265 95.810 -59.595 53.239
537266 95.843 -59.602 53.221
537267 95.877 -59.609 53.201
537268 95.910 -59.616 53.180
537269 95.944 -59.623 53.159
537270 95.979 -59.629 53.135
537271 96.014 -59.635 53.111
537272 96.049 -59.641 53.087
537273 96.084 -59.647 53.065
537274 96.120 -59.652 53.046
537275 96.157 -59.657 53.026
537276 96.194 -59.663 53.005
537277 96.231 -59.668 52.981
537278 96.268 -59.673 52.956
537279 96.305 -59.679 52.929
537280 96.341 -59.682 52.899
537281 96.376 -59.685 52.868
537282 96.411 -59.689 52.839
537283 96.447 -59.692 52.811
537284 96.482 -59.695 52.784
537285 96.518 -59.698 52.753
537286 96.553 -59.701 52.724
537287 96.591 -59.701 52.697
537288 96.629 -59.702 52.669
537289 96.665 -59.702 52.639
537290 96.697 -59.701 52.609
537291 96.729 -59.700 52.578
537292 96.761 -59.698 52.547
537293 96.793 -59.697 52.516
537294 96.825 -59.696 52.482
537295 96.858 -59.697 52.448
537296 96.892 -59.699 52.409
537297 96.926 -59.701 52.369
537298 96.959 -59.704 52.326
537299 96.993 -59.706 52.281
537300 97.027 -59.708 52.236
537301 97.061 -59.710 52.194
537302 97.094 -59.713 52.153
537303 97.128 -59.715 52.112
537304 97.162 -59.717 52.071
537305 97.195 -59.720 52.031
537306 97.229 -59.722 51.995
537307 97.263 -59.725 51.957
537308 97.293 -59.738 51.921
537309 97.323 -59.751 51.884
537310 97.360 -59.752 51.842
537311 97.396 -59.753 51.797
537312 97.437 -59.761 51.750
537313 97.477 -59.772 51.702
537314 97.518 -59.783 51.650
537315 97.559 -59.795 51.598
537316 97.600 -59.806 51.546
537317 97.641 -59.817 51.492
537318 97.682 -59.828 51.436
537319 97.722 -59.839 51.380
537320 97.762 -59.850 51.325
537321 97.802 -59.861 51.272
537322 97.842 -59.873 51.221
537323 97.881 -59.883 51.170
537324 97.922 -59.895 51.118
537325 97.961 -59.906 51.069
537326 98.001 -59.917 51.022
537327 98.041 -59.928 50.976
537328 98.082 -59.933 50.934
537329 98.124 -59.936 50.894
537330 98.157 -59.927 50.854
537331 98.182 -59.923 50.816
537332 98.204 -59.929 50.777
537333 98.225 -59.935 50.738
537334 98.247 -59.942 50.698
537335 98.269 -59.948 50.660
537336 98.290 -59.955 50.624
537337 98.311 -59.961 50.590
537338 98.344 -59.962 50.557
537339 98.388 -59.953 50.523
537340 98.435 -59.936 50.486
537341 98.481 -59.920 50.444
537342 98.527 -59.903 50.401
537343 98.574 -59.886 50.357
537344 98.620 -59.870 50.313
537345 98.656 -59.874 50.269
537346 98.693 -59.879 50.225
537347 98.729 -59.883 50.181
537348 98.765 -59.887 50.136
537349 98.801 -59.887 50.091
537350 98.834 -59.864 50.048
537351 98.867 -59.842 50.003
537352 98.900 -59.820 49.959
537353 98.958 -59.814 49.915
537354 99.021 -59.812 49.871
537355 99.051 -59.816 49.828
537356 99.079 -59.820 49.786
537357 99.108 -59.824 49.744
537358 99.137 -59.828 49.702
537780 88.769 -57.352 56.290
537781 88.807 -57.369 56.269
537782 88.845 -57.385 56.249
537783 88.884 -57.401 56.229
537784 88.922 -57.416 56.209
537785 88.960 -57.432 56.188
537786 88.998 -57.448 56.167
537787 89.030 -57.468 56.147
537788 89.054 -57.492 56.127
537789 89.079 -57.517 56.106
537790 89.103 -57.540 56.085
537791 89.134 -57.559 56.064
537792 89.176 -57.570 56.042
537793 89.217 -57.581 56.021
537794 89.258 -57.592 55.999
537795 89.300 -57.602 55.976
537796 89.341 -57.614 55.957
537797 89.383 -57.624 55.937
537798 89.426 -57.637 55.917
537799 89.476 -57.654 55.899
537800 89.525 -57.671 55.883
537801 89.574 -57.688 55.871
537802 89.623 -57.705 55.860
537803 89.672 -57.722 55.851
537804 89.721 -57.739 55.844
537805 89.770 -57.755 55.837
537806 89.813 -57.770 55.834
537807 89.849 -57.780 55.831
537808 89.886 -57.789 55.829
537809 89.922 -57.799 55.827
537810 89.959 -57.809 55.826
537811 89.995 -57.819 55.825
537812 90.031 -57.829 55.821
537813 90.068 -57.839 55.816
537814 90.104 -57.849 55.809
537815 90.141 -57.859 55.800
537816 90.177 -57.869 55.792
537817 90.214 -57.879 55.784
537818 90.257 -57.885 55.767
537819 90.291 -57.898 55.742
537820 90.324 -57.910 55.711
537821 90.358 -57.921 55.679
537822 90.392 -57.933 55.652
537823 90.426 -57.944 55.625
537824 90.460 -57.955 55.602
537825 90.494 -57.967 55.581
537826 90.533 -57.974 55.563
537827 90.576 -57.978 55.546
537828 90.617 -57.983 55.526
537829 90.660 -57.987 55.501
537830 90.702 -57.992 55.473
537831 90.744 -57.996 55.442
537832 90.786 -58.001 55.409
537833 90.828 -58.006 55.375
537834 90.871 -58.011 55.342
537835 90.916 -58.017 55.311
537836 90.961 -58.024 55.282
537837 91.004 -58.036 55.257
537838 91.043 -58.059 55.236
537839 91.083 -58.083 55.219
537840 91.123 -58.106 55.204
537841 91.162 -58.131 55.195
537842 91.202 -58.154 55.192
537843 91.242 -58.178 55.194
537844 91.281 -58.201 55.191
537845 91.321 -58.225 55.180
537846 91.360 -58.249 55.163
537847 91.400 -58.272 55.142
537848 91.440 -58.296 55.120
537849 91.479 -58.320 55.099
537850 91.519 -58.344 55.081
537851 91.559 -58.367 55.066
537852 91.598 -58.391 55.056
537853 91.638 -58.407 55.049
537854 91.665 -58.418 55.041
537855 91.697 -58.429 55.036
537856 91.730 -58.439 55.027
537857 91.763 -58.449 55.014
537858 91.799 -58.459 54.998
537859 91.835 -58.468 54.975
537860 91.870 -58.478 54.949
537861 91.907 -58.487 54.923
537862 91.942 -58.497 54.900
537863 91.980 -58.507 54.884
537864 92.017 -58.517 54.871
537865 92.055 -58.527 54.863
537866 92.092 -58.538 54.853
537867 92.130 -58.548 54.835
537868 92.171 -58.557 54.810
537869 92.210 -58.560 54.779
537870 92.249 -58.563 54.747
537871 92.288 -58.566 54.729
537872 92.326 -58.569 54.726
537873 92.366 -58.573 54.735
537874 92.405 -58.576 54.746
537875 92.443 -58.579 54.751
537876 92.482 -58.582 54.745
537877 92.519 -58.585 54.746
537878 92.549 -58.591 54.756
537879 92.578 -58.596 54.767
537880 92.608 -58.601 54.778
537881 92.643 -58.607 54.785
537882 92.689 -58.613 54.790
537883 92.735 -58.619 54.794
537884 92.780 -58.625 54.796
537885 92.826 -58.630 54.801
537886 92.871 -58.636 54.805
537887 92.917 -58.642 54.802
537888 92.963 -58.648 54.801
537889 93.008 -58.654 54.795
537890 93.047 -58.664 54.784
537891 93.084 -58.673 54.768
537892 93.123 -58.683 54.748
537893 93.160 -58.692 54.726
537894 93.198 -58.702 54.707
537895 93.236 -58.712 54.687
537896 93.275 -58.720 54.669
537897 93.313 -58.728 54.652
537898 93.352 -58.736 54.636
537899 93.391 -58.744 54.622
537900 93.429 -58.751 54.608
537901 93.468 -58.759 54.592
537902 93.507 -58.767 54.574
537903 93.545 -58.777 54.552
537904 93.581 -58.790 54.530
537905 93.618 -58.804 54.516
537906 93.654 -58.817 54.505
537907 93.690 -58.831 54.497
537908 93.726 -58.845 54.490
537909 93.763 -58.858 54.482
537910 93.799 -58.872 54.473
537911 93.835 -58.886 54.457
537912 93.871 -58.899 54.435
537913 93.908 -58.914 54.409
537914 93.944 -58.927 54.382
537915 93.980 -58.941 54.350
537916 94.016 -58.955 54.320
537917 94.053 -58.968 54.292
537918 94.089 -58.982 54.268
537919 94.125 -58.997 54.251
537920 94.160 -59.018 54.238
537921 94.195 -59.038 54.219
537922 94.230 -59.059 54.205
537923 94.265 -59.078 54.190
537924 94.300 -59.099 54.171
537925 94.335 -59.119 54.149
537926 94.370 -59.140 54.124
537927 94.405 -59.159 54.094
537928 94.440 -59.180 54.062
537929 94.475 -59.197 54.030
537930 94.508 -59.204 53.998
537931 94.543 -59.211 53.967
537932 94.576 -59.218 53.936
537933 94.613 -59.226 53.905
537934 94.651 -59.234 53.873
537935 94.690 -59.244 53.842
537936 94.731 -59.253 53.812
537937 94.772 -59.262 53.781
537938 94.813 -59.272 53.746
537939 94.854 -59.282 53.712
537940 94.895 -59.292 53.678
537941 94.936 -59.302 53.645
537942 94.977 -59.312 53.611
537943 95.018 -59.321 53.581
537944 95.056 -59.332 53.553
537945 95.093 -59.343 53.523
537946 95.130 -59.354 53.490
537947 95.167 -59.365 53.458
537948 95.204 -59.375 53.429
537949 95.243 -59.387 53.403
537950 95.282 -59.398 53.379
537951 95.321 -59.409 53.356
537952 95.355 -59.420 53.334
537953 95.389 -59.432 53.316
537954 95.422 -59.443 53.298
537955 95.456 -59.454 53.285
537956 95.489 -59.466 53.279
537957 95.522 -59.477 53.276
537958 95.555 -59.488 53.276
537959 95.588 -59.499 53.273
537960 95.621 -59.510 53.272
537961 95.654 -59.519 53.269
537962 95.687 -59.530 53.268
537963 95.719 -59.539 53.268
537964 95.750 -59.547 53.264
537965 95.781 -59.556 53.256
537966 95.814 -59.564 53.243
537967 95.847 -59.571 53.229
537968 95.881 -59.578 53.212
537969 95.914 -59.585 53.193
537970 95.949 -59.591 53.173
537971 95.984 -59.597 53.150
537972 96.019 -59.603 53.126
537973 96.054 -59.609 53.103
537974 96.089 -59.615 53.083
537975 96.124 -59.620 53.064
537976 96.161 -59.626 53.045
537977 96.198 -59.631 53.023
537978 96.235 -59.636 52.998
537979 96.272 -59.641 52.971
537980 96.308 -59.645 52.941
537981 96.343 -59.648 52.909
537982 96.378 -59.651 52.878
537983 96.413 -59.654 52.849
537984 96.449 -59.657 52.821
537985 96.484 -59.660 52.794
537986 96.520 -59.663 52.764
537987 96.555 -59.666 52.737
537988 96.591 -59.669 52.712
537989 96.629 -59.670 52.685
537990 96.661 -59.669 52.656
537991 96.693 -59.667 52.626
537992 96.726 -59.666 52.594
537993 96.757 -59.664 52.562
537994 96.790 -59.663 52.529
537995 96.821 -59.662 52.495
537996 96.855 -59.663 52.457
537997 96.888 -59.665 52.416
537998 96.922 -59.667 52.374
537999 96.956 -59.670 52.330
538000 96.990 -59.672 52.288
538001 97.023 -59.675 52.246
538002 97.057 -59.677 52.206
538003 97.091 -59.679 52.167
538004 97.124 -59.682 52.125
538005 97.158 -59.684 52.085
538006 97.192 -59.686 52.049
538007 97.226 -59.689 52.015
538008 97.260 -59.691 51.979
538009 97.293 -59.696 51.943
538010 97.323 -59.709 51.905
538011 97.358 -59.713 51.864
538012 97.395 -59.714 51.820
538013 97.436 -59.726 51.773
538014 97.476 -59.737 51.727
538015 97.518 -59.748 51.677
538016 97.559 -59.759 51.626
538017 97.599 -59.770 51.572
538018 97.640 -59.782 51.516
538019 97.679 -59.792 51.458
538020 97.719 -59.804 51.399
538021 97.759 -59.815 51.342
538022 97.799 -59.826 51.288
538023 97.839 -59.837 51.235
538024 97.878 -59.848 51.182
538025 97.918 -59.859 51.128
538026 97.958 -59.870 51.076
538027 97.998 -59.881 51.027
538028 98.037 -59.892 50.980
538029 98.080 -59.895 50.936
538030 98.122 -59.898 50.894
538031 98.156 -59.892 50.853
538032 98.185 -59.879 50.814
538033 98.207 -59.885 50.775
538034 98.228 -59.892 50.736
538035 98.249 -59.898 50.699
538036 98.271 -59.905 50.662
538037 98.293 -59.911 50.630
538038 98.314 -59.917 50.598
538039 98.352 -59.925 50.567
538040 98.390 -59.910 50.534
538041 98.433 -59.909 50.499
538042 98.480 -59.898 50.458
538043 98.526 -59.881 50.416
538044 98.573 -59.864 50.373
538045 98.619 -59.847 50.330
538046 98.662 -59.837 50.285
538047 98.699 -59.842 50.239
538048 98.735 -59.845 50.193
538049 98.771 -59.850 50.148
538050 98.807 -59.846 50.103
538051 98.840 -59.824 50.059
538052 98.873 -59.801 50.015
538053 98.906 -59.779 49.971
538054 98.967 -59.775 49.927
538055 99.020 -59.775 49.882
538056 99.049 -59.779 49.838
538057 99.078 -59.783 49.796
538058 99.106 -59.787 49.754
538059 99.135 -59.791 49.712
538060 99.164 -59.795 49.671
538482 88.825 -57.330 56.254
538483 88.863 -57.346 56.236
538484 88.901 -57.362 56.216
538485 88.940 -57.378 56.196
538486 88.978 -57.394 56.175
538487 89.016 -57.410 56.155
538488 89.049 -57.429 56.136
538489 89.073 -57.453 56.116
538490 89.098 -57.477 56.096
538491 89.122 -57.501 56.076
538492 89.154 -57.520 56.055
538493 89.195 -57.530 56.034
538494 89.237 -57.541 56.013
538495 89.278 -57.552 55.992
538496 89.319 -57.563 55.971
538497 89.361 -57.573 55.951
538498 89.402 -57.585 55.932
538499 89.444 -57.595 55.914
538500 89.488 -57.609 55.895
538501 89.538 -57.626 55.880
538502 89.587 -57.643 55.868
538503 89.635 -57.660 55.858
538504 89.685 -57.677 55.849
538505 89.734 -57.694 55.842
538506 89.782 -57.711 55.835
538507 89.829 -57.726 55.831
538508 89.865 -57.736 55.828
538509 89.901 -57.746 55.826
538510 89.938 -57.756 55.823
538511 89.974 -57.766 55.821
538512 90.011 -57.776 55.819
538513 90.047 -57.786 55.814
538514 90.084 -57.796 55.809
538515 90.120 -57.805 55.802
538516 90.156 -57.815 55.793
538517 90.192 -57.826 55.786
538518 90.229 -57.836 55.778
538519 90.265 -57.846 55.764
538520 90.301 -57.856 55.740
538521 90.335 -57.868 55.708
538522 90.369 -57.879 55.671
538523 90.402 -57.890 55.644
538524 90.437 -57.902 55.618
538525 90.470 -57.914 55.594
538526 90.504 -57.925 55.574
538527 90.539 -57.936 55.555
538528 90.581 -57.940 55.536
538529 90.623 -57.945 55.511
538530 90.665 -57.949 55.482
538531 90.707 -57.954 55.452
538532 90.750 -57.959 55.420
538533 90.795 -57.966 55.387
538534 90.840 -57.973 55.355
538535 90.885 -57.979 55.325
538536 90.930 -57.986 55.298
538537 90.975 -57.992 55.272
538538 91.018 -58.006 55.249
538539 91.058 -58.029 55.230
538540 91.097 -58.053 55.213
538541 91.137 -58.077 55.198
538542 91.176 -58.100 55.187
538543 91.216 -58.124 55.180
538544 91.255 -58.148 55.175
538545 91.295 -58.172 55.170
538546 91.335 -58.195 55.162
538547 91.374 -58.219 55.148
538548 91.414 -58.243 55.131
538549 91.453 -58.266 55.111
538550 91.493 -58.290 55.093
538551 91.533 -58.313 55.080
538552 91.572 -58.338 55.073
538553 91.614 -58.355 55.071
538554 91.658 -58.366 55.068
538555 91.688 -58.388 55.061
538556 91.716 -58.397 55.053
538557 91.744 -58.407 55.043
538558 91.773 -58.416 55.029
538559 91.810 -58.426 55.012
538560 91.847 -58.436 54.989
538561 91.885 -58.446 54.962
538562 91.922 -58.457 54.934
538563 91.959 -58.466 54.911
538564 91.997 -58.477 54.892
538565 92.034 -58.487 54.876
538566 92.072 -58.498 54.865
538567 92.109 -58.507 54.851
538568 92.150 -58.518 54.835
538569 92.192 -58.528 54.813
538570 92.231 -58.531 54.786
538571 92.270 -58.534 54.753
538572 92.308 -58.537 54.732
538573 92.347 -58.541 54.722
538574 92.386 -58.544 54.725
538575 92.425 -58.547 54.733
538576 92.464 -58.550 54.741
538577 92.503 -58.553 54.745
538578 92.534 -58.558 54.745
538579 92.564 -58.563 54.753
538580 92.593 -58.569 54.767
538581 92.623 -58.574 54.780
538582 92.654 -58.579 54.784
538583 92.700 -58.585 54.789
538584 92.745 -58.591 54.796
538585 92.791 -58.597 54.804
538586 92.837 -58.603 54.812
538587 92.882 -58.609 54.816
538588 92.928 -58.615 54.814
538589 92.974 -58.621 54.813
538590 93.016 -58.628 54.808
538591 93.055 -58.638 54.798
538592 93.092 -58.647 54.784
538593 93.131 -58.657 54.766
538594 93.168 -58.667 54.742
538595 93.206 -58.676 54.720
538596 93.245 -58.686 54.700
538597 93.282 -58.695 54.680
538598 93.321 -58.702 54.661
538599 93.360 -58.711 54.644
538600 93.399 -58.718 54.628
538601 93.437 -58.726 54.613
538602 93.476 -58.733 54.599
538603 93.515 -58.741 54.582
538604 93.552 -58.751 54.566
538605 93.588 -58.765 54.550
538606 93.625 -58.779 54.534
538607 93.661 -58.792 54.521
538608 93.698 -58.806 54.512
538609 93.734 -58.820 54.507
538610 93.770 -58.834 54.501
538611 93.806 -58.848 54.489
538612 93.843 -58.861 54.474
538613 93.879 -58.875 54.456
538614 93.915 -58.889 54.433
538615 93.952 -58.902 54.406
538616 93.988 -58.916 54.376
538617 94.024 -58.930 54.347
538618 94.060 -58.943 54.317
538619 94.097 -58.957 54.290
538620 94.133 -58.971 54.272
538621 94.170 -58.984 54.264
538622 94.205 -59.003 54.246
538623 94.240 -59.024 54.230
538624 94.274 -59.044 54.214
538625 94.310 -59.065 54.194
538626 94.344 -59.084 54.172
538627 94.380 -59.105 54.144
538628 94.414 -59.125 54.114
538629 94.450 -59.146 54.082
538630 94.484 -59.160 54.049
538631 94.518 -59.168 54.014
538632 94.552 -59.175 53.983
538633 94.587 -59.182 53.952
538634 94.626 -59.190 53.920
538635 94.664 -59.200 53.888
538636 94.702 -59.208 53.855
538637 94.744 -59.218 53.823
538638 94.785 -59.228 53.789
538639 94.826 -59.237 53.753
538640 94.867 -59.247 53.719
538641 94.908 -59.257 53.686
538642 94.950 -59.267 53.653
538643 94.991 -59.277 53.621
538644 95.030 -59.287 53.590
538645 95.067 -59.299 53.559
538646 95.104 -59.309 53.527
538647 95.141 -59.320 53.495
538648 95.179 -59.331 53.466
538649 95.216 -59.342 53.441
538650 95.254 -59.353 53.418
538651 95.293 -59.364 53.395
538652 95.328 -59.375 53.374
538653 95.361 -59.387 53.355
538654 95.394 -59.398 53.334
538655 95.428 -59.410 53.313
538656 95.461 -59.421 53.299
538657 95.494 -59.432 53.290
538658 95.528 -59.444 53.285
538659 95.561 -59.455 53.282
538660 95.594 -59.466 53.280
538661 95.627 -59.477 53.278
538662 95.660 -59.488 53.273
538663 95.693 -59.498 53.269
538664 95.725 -59.507 53.269
538665 95.756 -59.515 53.265
538666 95.787 -59.524 53.259
538667 95.818 -59.532 53.250
538668 95.851 -59.539 53.237
538669 95.885 -59.547 53.223
538670 95.919 -59.554 53.205
538671 95.954 -59.559 53.186
538672 95.989 -59.565 53.163
538673 96.024 -59.571 53.140
538674 96.059 -59.577 53.118
538675 96.094 -59.582 53.099
538676 96.129 -59.588 53.082
538677 96.166 -59.594 53.063
538678 96.202 -59.599 53.041
538679 96.239 -59.604 53.014
538680 96.274 -59.607 52.984
538681 96.309 -59.610 52.951
538682 96.344 -59.613 52.917
538683 96.380 -59.616 52.887
538684 96.415 -59.619 52.860
538685 96.451 -59.622 52.833
538686 96.486 -59.625 52.806
538687 96.521 -59.628 52.775
538688 96.556 -59.631 52.751
538689 96.592 -59.634 52.727
538690 96.626 -59.636 52.701
538691 96.658 -59.635 52.672
538692 96.690 -59.633 52.642
538693 96.722 -59.632 52.610
538694 96.754 -59.631 52.577
538695 96.786 -59.629 52.542
538696 96.818 -59.629 52.505
538697 96.851 -59.629 52.465
538698 96.885 -59.632 52.422
538699 96.919 -59.634 52.380
538700 96.952 -59.636 52.337
538701 96.986 -59.638 52.297
538702 97.020 -59.641 52.257
538703 97.053 -59.643 52.217
538704 97.087 -59.645 52.176
538705 97.121 -59.648 52.135
538706 97.155 -59.650 52.098
538707 97.189 -59.653 52.065
538708 97.223 -59.655 52.034
538709 97.256 -59.657 51.999
538710 97.290 -59.660 51.963
538711 97.322 -59.667 51.926
538712 97.356 -59.674 51.885
538713 97.394 -59.679 51.840
538714 97.435 -59.690 51.794
538715 97.476 -59.701 51.747
538716 97.517 -59.713 51.697
538717 97.557 -59.724 51.645
538718 97.596 -59.735 51.591
538719 97.636 -59.746 51.535
538720 97.676 -59.757 51.476
538721 97.716 -59.768 51.415
538722 97.755 -59.779 51.356
538723 97.795 -59.790 51.299
538724 97.836 -59.801 51.244
538725 97.875 -59.813 51.191
538726 97.915 -59.823 51.136
538727 97.955 -59.835 51.082
538728 97.995 -59.846 51.031
538729 98.035 -59.855 50.984
538730 98.077 -59.858 50.938
538731 98.119 -59.861 50.895
538732 98.155 -59.857 50.854
538733 98.184 -59.844 50.815
538734 98.209 -59.841 50.776
538735 98.231 -59.848 50.738
538736 98.252 -59.854 50.701
538737 98.274 -59.861 50.666
538738 98.295 -59.867 50.633
538739 98.317 -59.873 50.603
538740 98.358 -59.881 50.574
538741 98.398 -59.873 50.543
538742 98.437 -59.858 50.509
538743 98.479 -59.852 50.470
538744 98.525 -59.858 50.429
538745 98.572 -59.842 50.388
538746 98.618 -59.825 50.345
538747 98.664 -59.808 50.302
538748 98.705 -59.804 50.256
538749 98.741 -59.808 50.208
538750 98.778 -59.813 50.160
538751 98.812 -59.806 50.115
538752 98.846 -59.783 50.071
538753 98.878 -59.761 50.026
538754 98.912 -59.739 49.981
538755 98.975 -59.736 49.937
538756 99.019 -59.738 49.892
538757 99.048 -59.742 49.847
538758 99.076 -59.745 49.806
538759 99.105 -59.750 49.764
538760 99.134 -59.754 49.722
538761 99.162 -59.757 49.681
538762 99.191 -59.761 49.641
539185 88.919 -57.323 56.202
539186 88.958 -57.339 56.182
539187 88.996 -57.355 56.162
539188 89.034 -57.371 56.143
539189 89.067 -57.390 56.124
539190 89.092 -57.414 56.105
539191 89.116 -57.438 56.085
539192 89.140 -57.462 56.066
539193 89.173 -57.480 56.045
539194 89.214 -57.491 56.025
539195 89.256 -57.501 56.005
539196 89.298 -57.512 55.986
539197 89.339 -57.523 55.965
539198 89.380 -57.534 55.946
539199 89.421 -57.545 55.928
539200 89.463 -57.556 55.910
539201 89.505 -57.567 55.892
539202 89.550 -57.581 55.878
539203 89.599 -57.598 55.866
539204 89.649 -57.615 55.855
539205 89.697 -57.632 55.846
539206 89.747 -57.648 55.838
539207 89.796 -57.666 55.830
539208 89.844 -57.683 55.827
539209 89.881 -57.692 55.824
539210 89.917 -57.702 55.821
539211 89.954 -57.712 55.817
539212 89.990 -57.723 55.814
539213 90.026 -57.733 55.811
539214 90.062 -57.742 55.805
539215 90.099 -57.752 55.799
539216 90.135 -57.762 55.791
539217 90.172 -57.772 55.783
539218 90.208 -57.782 55.774
539219 90.245 -57.792 55.766
539220 90.280 -57.805 55.750
539221 90.314 -57.816 55.725
539222 90.347 -57.827 55.693
539223 90.381 -57.839 55.657
539224 90.414 -57.849 55.629
539225 90.448 -57.861 55.603
539226 90.481 -57.872 55.579
539227 90.515 -57.883 55.559
539228 90.548 -57.894 55.541
539229 90.586 -57.903 55.520
539230 90.630 -57.909 55.490
539231 90.675 -57.915 55.459
539232 90.720 -57.921 55.428
539233 90.765 -57.928 55.396
539234 90.810 -57.935 55.365
539235 90.855 -57.941 55.336
539236 90.900 -57.948 55.310
539237 90.945 -57.954 55.286
539238 90.990 -57.961 55.263
539239 91.032 -57.976 55.242
539240 91.072 -57.999 55.224
539241 91.111 -58.023 55.208
539242 91.151 -58.046 55.194
539243 91.190 -58.071 55.183
539244 91.230 -58.094 55.174
539245 91.270 -58.118 55.167
539246 91.309 -58.141 55.159
539247 91.349 -58.165 55.149
539248 91.388 -58.189 55.138
539249 91.428 -58.212 55.125
539250 91.467 -58.236 55.111
539251 91.507 -58.260 55.099
539252 91.546 -58.284 55.089
539253 91.587 -58.305 55.081
539254 91.631 -58.316 55.075
539255 91.674 -58.326 55.073
539256 91.709 -58.345 55.070
539257 91.739 -58.367 55.066
539258 91.767 -58.378 55.055
539259 91.795 -58.386 55.039
539260 91.827 -58.396 55.019
539261 91.864 -58.406 54.992
539262 91.901 -58.416 54.965
539263 91.939 -58.426 54.937
539264 91.977 -58.437 54.911
539265 92.014 -58.447 54.890
539266 92.052 -58.457 54.874
539267 92.089 -58.467 54.860
539268 92.127 -58.478 54.844
539269 92.170 -58.488 54.830
539270 92.212 -58.498 54.811
539271 92.251 -58.502 54.786
539272 92.290 -58.505 54.757
539273 92.329 -58.509 54.735
539274 92.368 -58.512 54.722
539275 92.407 -58.515 54.718
539276 92.446 -58.518 54.721
539277 92.485 -58.521 54.727
539278 92.520 -58.526 54.734
539279 92.549 -58.531 54.736
539280 92.579 -58.536 54.742
539281 92.609 -58.542 54.753
539282 92.638 -58.546 54.767
539283 92.667 -58.551 54.774
539284 92.711 -58.557 54.772
539285 92.756 -58.563 54.776
539286 92.802 -58.570 54.786
539287 92.847 -58.576 54.799
539288 92.893 -58.581 54.809
539289 92.938 -58.587 54.809
539290 92.984 -58.593 54.808
539291 93.025 -58.601 54.806
539292 93.063 -58.611 54.801
539293 93.100 -58.621 54.790
539294 93.139 -58.630 54.774
539295 93.177 -58.640 54.752
539296 93.215 -58.650 54.730
539297 93.253 -58.659 54.710
539298 93.290 -58.669 54.690
539299 93.329 -58.676 54.669
539300 93.368 -58.685 54.650
539301 93.407 -58.692 54.633
539302 93.445 -58.700 54.620
539303 93.484 -58.708 54.607
539304 93.522 -58.715 54.594
539305 93.560 -58.726 54.583
539306 93.596 -58.741 54.572
539307 93.632 -58.755 54.557
539308 93.669 -58.768 54.542
539309 93.705 -58.782 54.530
539310 93.742 -58.795 54.521
539311 93.778 -58.809 54.513
539312 93.814 -58.823 54.503
539313 93.850 -58.836 54.491
539314 93.887 -58.850 54.476
539315 93.923 -58.864 54.457
539316 93.959 -58.877 54.431
539317 93.995 -58.891 54.405
539318 94.032 -58.905 54.377
539319 94.068 -58.918 54.348
539320 94.104 -58.932 54.321
539321 94.140 -58.946 54.299
539322 94.177 -58.960 54.285
539323 94.213 -58.974 54.269
539324 94.249 -58.990 54.252
539325 94.284 -59.010 54.234
539326 94.319 -59.030 54.215
539327 94.354 -59.050 54.190
539328 94.389 -59.071 54.163
539329 94.424 -59.091 54.134
539330 94.459 -59.111 54.102
539331 94.493 -59.124 54.069
539332 94.527 -59.131 54.034
539333 94.561 -59.138 54.004
539334 94.600 -59.146 53.972
539335 94.638 -59.156 53.939
539336 94.677 -59.164 53.905
539337 94.715 -59.172 53.869
539338 94.757 -59.182 53.834
539339 94.798 -59.193 53.797
539340 94.839 -59.203 53.762
539341 94.880 -59.213 53.728
539342 94.921 -59.224 53.693
539343 94.962 -59.235 53.660
539344 95.002 -59.246 53.629
539345 95.043 -59.256 53.597
539346 95.079 -59.267 53.564
539347 95.116 -59.278 53.532
539348 95.153 -59.287 53.503
539349 95.190 -59.298 53.478
539350 95.227 -59.309 53.456
539351 95.264 -59.320 53.434
539352 95.300 -59.331 53.412
539353 95.333 -59.343 53.391
539354 95.367 -59.354 53.372
539355 95.400 -59.366 53.351
539356 95.433 -59.377 53.331
539357 95.467 -59.388 53.315
539358 95.500 -59.400 53.305
539359 95.533 -59.411 53.299
539360 95.566 -59.422 53.293
539361 95.600 -59.434 53.291
539362 95.633 -59.445 53.287
539363 95.666 -59.456 53.282
539364 95.699 -59.466 53.277
539365 95.730 -59.474 53.277
539366 95.761 -59.482 53.273
539367 95.793 -59.491 53.266
539368 95.824 -59.500 53.257
539369 95.855 -59.508 53.245
539370 95.889 -59.515 53.230
539371 95.923 -59.521 53.214
539372 95.959 -59.527 53.194
539373 95.994 -59.533 53.172
539374 96.029 -59.539 53.151
539375 96.064 -59.544 53.133
539376 96.099 -59.551 53.115
539377 96.134 -59.557 53.098
539378 96.170 -59.562 53.077
539379 96.205 -59.566 53.053
539380 96.240 -59.569 53.024
539381 96.276 -59.573 52.992
539382 96.311 -59.576 52.958
539383 96.346 -59.580 52.923
539384 96.382 -59.583 52.895
539385 96.417 -59.587 52.868
539386 96.453 -59.591 52.843
539387 96.488 -59.595 52.816
539388 96.524 -59.598 52.788
539389 96.559 -59.602 52.764
539390 96.595 -59.606 52.741
539391 96.625 -59.604 52.715
539392 96.654 -59.601 52.687
539393 96.686 -59.600 52.657
539394 96.718 -59.598 52.625
539395 96.750 -59.597 52.590
539396 96.783 -59.596 52.553
539397 96.814 -59.595 52.514
539398 96.848 -59.596 52.471
539399 96.881 -59.598 52.428
539400 96.915 -59.601 52.387
539401 96.949 -59.603 52.347
539402 96.982 -59.605 52.308
539403 97.016 -59.607 52.269
539404 97.050 -59.610 52.226
539405 97.084 -59.612 52.183
539406 97.118 -59.614 52.143
539407 97.152 -59.616 52.109
539408 97.185 -59.619 52.080
539409 97.219 -59.621 52.052
539410 97.253 -59.623 52.017
539411 97.287 -59.626 51.982
539412 97.320 -59.628 51.945
539413 97.353 -59.635 51.904
539414 97.393 -59.644 51.859
539415 97.434 -59.655 51.812
539416 97.474 -59.666 51.763
539417 97.514 -59.677 51.712
539418 97.554 -59.689 51.659
539419 97.593 -59.699 51.605
539420 97.633 -59.710 51.548
539421 97.673 -59.722 51.489
539422 97.713 -59.732 51.428
539423 97.752 -59.744 51.368
539424 97.792 -59.755 51.310
539425 97.832 -59.766 51.253
539426 97.872 -59.777 51.199
539427 97.912 -59.789 51.143
539428 97.951 -59.799 51.089
539429 97.991 -59.811 51.037
539430 98.032 -59.817 50.988
539431 98.074 -59.820 50.941
539432 98.117 -59.823 50.898
539433 98.155 -59.822 50.856
539434 98.184 -59.808 50.817
539435 98.212 -59.797 50.778
539436 98.233 -59.804 50.740
539437 98.255 -59.810 50.703
539438 98.277 -59.817 50.669
539439 98.298 -59.823 50.636
539440 98.319 -59.830 50.606
539441 98.363 -59.838 50.578
539442 98.407 -59.836 50.549
539443 98.445 -59.821 50.518
539444 98.484 -59.806 50.480
539445 98.523 -59.796 50.441
539446 98.570 -59.805 50.400
539447 98.617 -59.803 50.358
539448 98.663 -59.786 50.316
539449 98.710 -59.770 50.270
539450 98.747 -59.771 50.221
539451 98.784 -59.775 50.173
539452 98.818 -59.765 50.126
539453 98.851 -59.742 50.082
539454 98.884 -59.720 50.037
539455 98.921 -59.699 49.992
539456 98.984 -59.698 49.947
539457 99.017 -59.701 49.902
539458 99.046 -59.704 49.858
539459 99.075 -59.708 49.816
539460 99.103 -59.712 49.774
539461 99.132 -59.717 49.733
539462 99.161 -59.720 49.692
539463 99.190 -59.724 49.651
539464 99.218 -59.728 49.612
539887 88.975 -57.300 56.167
539888 89.014 -57.316 56.149
539889 89.052 -57.333 56.130
539890 89.086 -57.351 56.111
539891 89.110 -57.375 56.093
539892 89.135 -57.399 56.074
539893 89.159 -57.423 56.055
539894 89.193 -57.440 56.036
539895 89.234 -57.451 56.016
539896 89.275 -57.462 55.996
539897 89.317 -57.473 55.979
539898 89.358 -57.483 55.959
539899 89.400 -57.495 55.941
539900 89.441 -57.505 55.923
539901 89.482 -57.516 55.906
539902 89.524 -57.527 55.888
539903 89.565 -57.538 55.875
539904 89.612 -57.553 55.863
539905 89.661 -57.570 55.852
539906 89.710 -57.587 55.843
539907 89.759 -57.604 55.834
539908 89.809 -57.621 55.826
539909 89.858 -57.638 55.822
539910 89.896 -57.649 55.818
539911 89.933 -57.659 55.814
539912 89.969 -57.669 55.810
539913 90.006 -57.679 55.805
539914 90.042 -57.689 55.799
539915 90.078 -57.699 55.793
539916 90.115 -57.709 55.785
539917 90.151 -57.719 55.776
539918 90.187 -57.729 55.767
539919 90.224 -57.739 55.759
539920 90.259 -57.751 55.747
539921 90.294 -57.766 55.728
539922 90.327 -57.777 55.702
539923 90.361 -57.789 55.671
539924 90.395 -57.799 55.639
539925 90.428 -57.811 55.611
539926 90.462 -57.822 55.584
539927 90.495 -57.833 55.561
539928 90.529 -57.844 55.541
539929 90.562 -57.855 55.522
539930 90.599 -57.871 55.499
539931 90.644 -57.877 55.467
539932 90.689 -57.884 55.436
539933 90.734 -57.890 55.406
539934 90.779 -57.897 55.378
539935 90.824 -57.903 55.351
539936 90.869 -57.910 55.325
539937 90.914 -57.916 55.303
539938 90.959 -57.923 55.282
539939 91.004 -57.929 55.262
539940 91.046 -57.946 55.242
539941 91.085 -57.969 55.225
539942 91.125 -57.993 55.210
539943 91.165 -58.017 55.197
539944 91.204 -58.040 55.186
539945 91.244 -58.064 55.177
539946 91.284 -58.088 55.169
539947 91.323 -58.112 55.156
539948 91.363 -58.135 55.143
539949 91.402 -58.159 55.132
539950 91.442 -58.183 55.124
539951 91.482 -58.206 55.117
539952 91.521 -58.230 55.111
539953 91.561 -58.253 55.104
539954 91.604 -58.266 55.095
539955 91.648 -58.277 55.087
539956 91.692 -58.287 55.081
539957 91.729 -58.303 55.079
539958 91.759 -58.325 55.075
539959 91.790 -58.347 55.066
539960 91.818 -58.357 55.049
539961 91.846 -58.366 55.026
539962 91.881 -58.376 54.995
539963 91.919 -58.386 54.968
539964 91.956 -58.397 54.946
539965 91.994 -58.407 54.926
539966 92.031 -58.417 54.910
539967 92.068 -58.428 54.896
539968 92.106 -58.438 54.880
539969 92.147 -58.448 54.862
539970 92.190 -58.458 54.840
539971 92.232 -58.468 54.812
539972 92.272 -58.474 54.782
539973 92.311 -58.477 54.758
539974 92.350 -58.480 54.740
539975 92.388 -58.483 54.725
539976 92.428 -58.486 54.715
539977 92.466 -58.490 54.710
539978 92.505 -58.493 54.713
539979 92.535 -58.498 54.719
539980 92.564 -58.504 54.724
539981 92.594 -58.508 54.731
539982 92.624 -58.513 54.742
539983 92.653 -58.519 54.755
539984 92.683 -58.524 54.763
539985 92.721 -58.530 54.764
539986 92.767 -58.536 54.773
539987 92.812 -58.542 54.789
539988 92.858 -58.548 54.806
539989 92.904 -58.554 54.817
539990 92.949 -58.560 54.821
539991 92.995 -58.566 54.821
539992 93.033 -58.575 54.819
539993 93.071 -58.585 54.812
539994 93.109 -58.595 54.799
539995 93.147 -58.604 54.785
539996 93.185 -58.614 54.767
539997 93.223 -58.623 54.749
539998 93.261 -58.632 54.731
539999 93.298 -58.642 54.712
540000 93.337 -58.651 54.692
540001 93.376 -58.659 54.668
540002 93.414 -58.667 54.647
540003 93.453 -58.674 54.631
540004 93.492 -58.682 54.621
540005 93.531 -58.689 54.615
540006 93.567 -58.702 54.605
540007 93.604 -58.716 54.598
540008 93.640 -58.730 54.587
540009 93.676 -58.743 54.572
540010 93.712 -58.757 54.553
540011 93.749 -58.770 54.532
540012 93.785 -58.784 54.521
540013 93.822 -58.798 54.513
540014 93.858 -58.811 54.506
540015 93.894 -58.825 54.495
540016 93.930 -58.839 54.480
540017 93.967 -58.852 54.457
540018 94.003 -58.867 54.434
540019 94.039 -58.880 54.408
540020 94.075 -58.894 54.381
540021 94.112 -58.908 54.356
540022 94.148 -58.921 54.331
540023 94.184 -58.935 54.307
540024 94.220 -58.949 54.288
540025 94.257 -58.962 54.270
540026 94.293 -58.976 54.250
540027 94.328 -58.996 54.231
540028 94.363 -59.016 54.207
540029 94.398 -59.036 54.183
540030 94.433 -59.056 54.155
540031 94.468 -59.077 54.124
540032 94.502 -59.087 54.090
540033 94.537 -59.094 54.055
540034 94.574 -59.102 54.025
540035 94.613 -59.112 53.991
540036 94.651 -59.120 53.958
540037 94.689 -59.128 53.924
540038 94.729 -59.139 53.889
540039 94.769 -59.150 53.851
540040 94.810 -59.161 53.812
540041 94.850 -59.171 53.774
540042 94.891 -59.182 53.735
540043 94.932 -59.193 53.697
540044 94.973 -59.204 53.661
540045 95.014 -59.215 53.628
540046 95.054 -59.225 53.597
540047 95.092 -59.236 53.567
540048 95.129 -59.247 53.539
540049 95.166 -59.257 53.517
540050 95.203 -59.268 53.494
540051 95.240 -59.278 53.471
540052 95.276 -59.289 53.450
540053 95.309 -59.302 53.428
540054 95.342 -59.312 53.405
540055 95.375 -59.324 53.385
540056 95.408 -59.335 53.368
540057 95.441 -59.346 53.350
540058 95.474 -59.357 53.334
540059 95.508 -59.369 53.321
540060 95.540 -59.379 53.315
540061 95.573 -59.391 53.307
540062 95.607 -59.402 53.303
540063 95.639 -59.413 53.299
540064 95.673 -59.424 53.294
540065 95.704 -59.433 53.289
540066 95.736 -59.441 53.287
540067 95.767 -59.450 53.281
540068 95.798 -59.459 53.273
540069 95.829 -59.467 53.264
540070 95.861 -59.475 53.252
540071 95.894 -59.483 53.236
540072 95.929 -59.489 53.220
540073 95.964 -59.495 53.200
540074 95.999 -59.501 53.180
540075 96.034 -59.507 53.161
540076 96.070 -59.513 53.146
540077 96.105 -59.518 53.129
540078 96.140 -59.524 53.110
540079 96.174 -59.530 53.088
540080 96.208 -59.536 53.060
540081 96.243 -59.540 53.028
540082 96.278 -59.544 52.995
540083 96.314 -59.547 52.963
540084 96.349 -59.551 52.932
540085 96.385 -59.555 52.905
540086 96.420 -59.559 52.880
540087 96.456 -59.563 52.854
540088 96.491 -59.566 52.828
540089 96.527 -59.569 52.802
540090 96.562 -59.573 52.777
540091 96.595 -59.575 52.753
540092 96.624 -59.572 52.728
540093 96.654 -59.569 52.700
540094 96.683 -59.566 52.670
540095 96.715 -59.564 52.637
540096 96.747 -59.563 52.600
540097 96.779 -59.562 52.561
540098 96.811 -59.560 52.522
540099 96.844 -59.562 52.480
540100 96.878 -59.564 52.439
540101 96.911 -59.566 52.399
540102 96.945 -59.569 52.358
540103 96.979 -59.571 52.318
540104 97.013 -59.574 52.278
540105 97.047 -59.576 52.233
540106 97.081 -59.579 52.189
540107 97.114 -59.581 52.148
540108 97.148 -59.583 52.115
540109 97.182 -59.585 52.089
540110 97.216 -59.588 52.063
540111 97.249 -59.590 52.032
540112 97.283 -59.592 51.998
540113 97.317 -59.595 51.960
540114 97.352 -59.598 51.918
540115 97.391 -59.608 51.873
540116 97.431 -59.620 51.823
540117 97.470 -59.631 51.773
540118 97.510 -59.641 51.719
540119 97.550 -59.653 51.664
540120 97.590 -59.664 51.609
540121 97.630 -59.675 51.552
540122 97.669 -59.686 51.495
540123 97.709 -59.698 51.436
540124 97.749 -59.708 51.379
540125 97.789 -59.720 51.323
540126 97.829 -59.731 51.265
540127 97.869 -59.742 51.208
540128 97.909 -59.753 51.152
540129 97.948 -59.764 51.096
540130 97.988 -59.775 51.043
540131 98.030 -59.779 50.992
540132 98.072 -59.783 50.945
540133 98.113 -59.786 50.901
540134 98.154 -59.787 50.859
540135 98.183 -59.773 50.819
540136 98.212 -59.760 50.780
540137 98.236 -59.760 50.741
540138 98.258 -59.766 50.705
540139 98.279 -59.773 50.671
540140 98.301 -59.779 50.638
540141 98.322 -59.786 50.608
540142 98.368 -59.794 50.580
540143 98.416 -59.799 50.554
540144 98.454 -59.784 50.524
540145 98.492 -59.769 50.488
540146 98.530 -59.754 50.450
540147 98.568 -59.740 50.409
540148 98.615 -59.749 50.368
540149 98.662 -59.758 50.327
540150 98.709 -59.748 50.281
540151 98.754 -59.733 50.234
540152 98.790 -59.738 50.186
540153 98.824 -59.724 50.138
540154 98.857 -59.701 50.092
540155 98.890 -59.679 50.047
540156 98.929 -59.660 50.002
540157 98.987 -59.660 49.957
540158 99.016 -59.663 49.912
540159 99.045 -59.667 49.868
540160 99.073 -59.671 49.827
540161 99.102 -59.675 49.785
540162 99.131 -59.679 49.744
540163 99.159 -59.683 49.702
540164 99.188 -59.687 49.661
540165 99.217 -59.691 49.622
540589 89.031 -57.279 56.135
540590 89.069 -57.294 56.117
540591 89.105 -57.312 56.099
540592 89.129 -57.336 56.081
540593 89.154 -57.360 56.063
540594 89.178 -57.385 56.045
540595 89.212 -57.401 56.027
540596 89.254 -57.411 56.008
540597 89.295 -57.423 55.989
540598 89.336 -57.433 55.972
540599 89.378 -57.444 55.953
540600 89.419 -57.454 55.935
540601 89.461 -57.466 55.919
540602 89.502 -57.476 55.902
540603 89.543 -57.487 55.884
540604 89.585 -57.498 55.871
540605 89.626 -57.509 55.859
540606 89.674 -57.525 55.849
540607 89.723 -57.542 55.839
540608 89.772 -57.559 55.830
540609 89.821 -57.576 55.822
540610 89.871 -57.593 55.816
540611 89.912 -57.606 55.811
540612 89.949 -57.616 55.806
540613 89.985 -57.626 55.801
540614 90.021 -57.636 55.795
540615 90.057 -57.645 55.787
540616 90.094 -57.655 55.779
540617 90.130 -57.665 55.770
540618 90.167 -57.675 55.759
540619 90.203 -57.686 55.748
540620 90.239 -57.697 55.735
540621 90.273 -57.713 55.718
540622 90.308 -57.727 55.697
540623 90.341 -57.739 55.672
540624 90.375 -57.749 55.643
540625 90.408 -57.761 55.615
540626 90.442 -57.772 55.588
540627 90.475 -57.783 55.562
540628 90.509 -57.794 55.539
540629 90.542 -57.805 55.520
540630 90.579 -57.820 55.500
540631 90.616 -57.836 55.474
540632 90.659 -57.846 55.443
540633 90.704 -57.852 55.414
540634 90.749 -57.859 55.387
540635 90.794 -57.865 55.363
540636 90.838 -57.872 55.339
540637 90.883 -57.878 55.316
540638 90.929 -57.885 55.298
540639 90.973 -57.892 55.281
540640 91.018 -57.898 55.264
540641 91.060 -57.916 55.247
540642 91.099 -57.940 55.232
540643 91.139 -57.963 55.219
540644 91.178 -57.987 55.207
540645 91.218 -58.011 55.199
540646 91.258 -58.034 55.194
540647 91.297 -58.058 55.189
540648 91.337 -58.081 55.177
540649 91.377 -58.106 55.161
540650 91.416 -58.129 55.144
540651 91.456 -58.153 55.133
540652 91.495 -58.176 55.133
540653 91.535 -58.200 55.131
540654 91.577 -58.216 55.127
540655 91.621 -58.227 55.119
540656 91.664 -58.238 55.108
540657 91.708 -58.248 55.094
540658 91.749 -58.261 55.089
540659 91.780 -58.283 55.084
540660 91.810 -58.305 55.074
540661 91.841 -58.328 55.058
540662 91.869 -58.338 55.034
540663 91.898 -58.346 55.004
540664 91.936 -58.357 54.979
540665 91.973 -58.366 54.958
540666 92.011 -58.377 54.939
540667 92.048 -58.388 54.922
540668 92.085 -58.397 54.905
540669 92.125 -58.408 54.886
540670 92.167 -58.418 54.866
540671 92.210 -58.428 54.842
540672 92.252 -58.438 54.818
540673 92.293 -58.445 54.798
540674 92.332 -58.448 54.783
540675 92.370 -58.451 54.768
540676 92.409 -58.454 54.750
540677 92.448 -58.458 54.730
540678 92.487 -58.461 54.714
540679 92.520 -58.466 54.708
540680 92.550 -58.471 54.711
540681 92.580 -58.476 54.715
540682 92.609 -58.481 54.726
540683 92.639 -58.486 54.740
540684 92.668 -58.491 54.751
540685 92.698 -58.497 54.755
540686 92.732 -58.502 54.768
540687 92.778 -58.508 54.788
540688 92.823 -58.514 54.810
540689 92.869 -58.520 54.827
540690 92.914 -58.526 54.835
540691 92.960 -58.532 54.839
540692 93.003 -58.539 54.840
540693 93.041 -58.549 54.837
540694 93.079 -58.558 54.828
540695 93.117 -58.568 54.814
540696 93.155 -58.578 54.801
540697 93.193 -58.587 54.787
540698 93.231 -58.597 54.772
540699 93.269 -58.607 54.755
540700 93.307 -58.616 54.736
540701 93.345 -58.625 54.716
540702 93.383 -58.633 54.693
540703 93.422 -58.641 54.670
540704 93.461 -58.648 54.651
540705 93.500 -58.656 54.637
540706 93.539 -58.664 54.633
540707 93.575 -58.677 54.627
540708 93.611 -58.691 54.619
540709 93.647 -58.705 54.608
540710 93.684 -58.718 54.593
540711 93.720 -58.732 54.573
540712 93.756 -58.745 54.551
540713 93.792 -58.759 54.536
540714 93.829 -58.773 54.525
540715 93.865 -58.787 54.515
540716 93.902 -58.801 54.506
540717 93.938 -58.814 54.497
540718 93.974 -58.828 54.482
540719 94.010 -58.842 54.461
540720 94.047 -58.855 54.439
540721 94.083 -58.869 54.418
540722 94.119 -58.883 54.395
540723 94.155 -58.896 54.368
540724 94.192 -58.910 54.330
540725 94.228 -58.924 54.304
540726 94.264 -58.937 54.284
540727 94.300 -58.951 54.265
540728 94.337 -58.965 54.243
540729 94.372 -58.981 54.223
540730 94.408 -59.002 54.202
540731 94.442 -59.022 54.176
540732 94.478 -59.042 54.145
540733 94.512 -59.051 54.111
540734 94.548 -59.059 54.075
540735 94.586 -59.068 54.043
540736 94.625 -59.076 54.009
540737 94.664 -59.087 53.975
540738 94.703 -59.097 53.940
540739 94.742 -59.108 53.903
540740 94.781 -59.119 53.863
540741 94.821 -59.130 53.824
540742 94.861 -59.140 53.784
540743 94.902 -59.151 53.743
540744 94.943 -59.162 53.704
540745 94.984 -59.173 53.668
540746 95.025 -59.184 53.636
540747 95.065 -59.194 53.605
540748 95.105 -59.205 53.578
540749 95.142 -59.215 53.554
540750 95.179 -59.226 53.532
540751 95.215 -59.237 53.509
540752 95.253 -59.247 53.486
540753 95.287 -59.259 53.464
540754 95.319 -59.275 53.443
540755 95.352 -59.285 53.423
540756 95.386 -59.297 53.404
540757 95.418 -59.308 53.387
540758 95.452 -59.319 53.369
540759 95.485 -59.330 53.351
540760 95.518 -59.341 53.336
540761 95.551 -59.352 53.326
540762 95.584 -59.363 53.319
540763 95.617 -59.374 53.316
540764 95.650 -59.385 53.313
540765 95.680 -59.394 53.308
540766 95.709 -59.403 53.301
540767 95.738 -59.411 53.297
540768 95.770 -59.419 53.286
540769 95.801 -59.428 53.278
540770 95.832 -59.436 53.269
540771 95.865 -59.444 53.258
540772 95.900 -59.450 53.242
540773 95.935 -59.457 53.223
540774 95.970 -59.463 53.204
540775 96.006 -59.468 53.186
540776 96.041 -59.474 53.170
540777 96.076 -59.480 53.157
540778 96.111 -59.485 53.141
540779 96.146 -59.491 53.120
540780 96.179 -59.498 53.095
540781 96.213 -59.505 53.065
540782 96.246 -59.512 53.030
540783 96.282 -59.516 52.996
540784 96.317 -59.519 52.969
540785 96.352 -59.522 52.942
540786 96.388 -59.526 52.918
540787 96.423 -59.530 52.893
540788 96.458 -59.534 52.867
540789 96.494 -59.538 52.841
540790 96.529 -59.541 52.815
540791 96.565 -59.545 52.789
540792 96.595 -59.544 52.764
540793 96.624 -59.540 52.739
540794 96.653 -59.537 52.712
540795 96.682 -59.534 52.682
540796 96.712 -59.531 52.648
540797 96.743 -59.529 52.609
540798 96.775 -59.528 52.571
540799 96.807 -59.527 52.532
540800 96.840 -59.529 52.490
540801 96.875 -59.531 52.450
540802 96.908 -59.533 52.410
540803 96.942 -59.535 52.369
540804 96.976 -59.538 52.328
540805 97.010 -59.540 52.286
540806 97.043 -59.542 52.240
540807 97.077 -59.544 52.195
540808 97.111 -59.547 52.155
540809 97.145 -59.549 52.123
540810 97.178 -59.552 52.097
540811 97.215 -59.554 52.072
540812 97.251 -59.557 52.042
540813 97.288 -59.560 52.008
540814 97.322 -59.564 51.970
540815 97.355 -59.567 51.926
540816 97.388 -59.573 51.879
540817 97.428 -59.584 51.828
540818 97.467 -59.595 51.775
540819 97.507 -59.607 51.721
540820 97.547 -59.617 51.666
540821 97.587 -59.629 51.611
540822 97.627 -59.640 51.555
540823 97.666 -59.651 51.500
540824 97.706 -59.662 51.446
540825 97.746 -59.673 51.393
540826 97.786 -59.684 51.339
540827 97.825 -59.695 51.281
540828 97.865 -59.706 51.223
540829 97.905 -59.717 51.164
540830 97.945 -59.729 51.107
540831 97.985 -59.738 51.052
540832 98.027 -59.741 50.999
540833 98.069 -59.745 50.951
540834 98.111 -59.748 50.905
540835 98.153 -59.751 50.862
540836 98.182 -59.739 50.821
540837 98.212 -59.725 50.782
540838 98.239 -59.716 50.744
540839 98.261 -59.723 50.708
540840 98.282 -59.729 50.672
540841 98.303 -59.735 50.639
540842 98.325 -59.742 50.608
540843 98.374 -59.750 50.579
540844 98.423 -59.759 50.550
540845 98.462 -59.748 50.521
540846 98.500 -59.732 50.489
540847 98.538 -59.717 50.454
540848 98.576 -59.701 50.415
540849 98.615 -59.686 50.374
540850 98.660 -59.692 50.333
540851 98.707 -59.702 50.290
540852 98.754 -59.708 50.245
540853 98.796 -59.701 50.198
540854 98.830 -59.683 50.151
540855 98.863 -59.660 50.103
540856 98.896 -59.638 50.057
540857 98.938 -59.622 50.011
540858 98.986 -59.623 49.967
540859 99.014 -59.626 49.922
540860 99.043 -59.630 49.879
540861 99.072 -59.634 49.836
540862 99.100 -59.638 49.795
540863 99.129 -59.642 49.754
540864 99.158 -59.646 49.712
540865 99.187 -59.650 49.672
540866 99.215 -59.654 49.633
540867 99.244 -59.657 49.596
541292 89.123 -57.273 56.088
541293 89.147 -57.297 56.070
541294 89.172 -57.321 56.053
541295 89.196 -57.345 56.035
541296 89.232 -57.361 56.018
541297 89.273 -57.372 56.000
541298 89.314 -57.382 55.981
541299 89.356 -57.393 55.965
541300 89.397 -57.404 55.947
541301 89.439 -57.415 55.931
541302 89.480 -57.426 55.914
541303 89.522 -57.437 55.898
541304 89.563 -57.448 55.881
541305 89.604 -57.458 55.867
541306 89.646 -57.470 55.855
541307 89.687 -57.480 55.845
541308 89.736 -57.497 55.835
541309 89.785 -57.514 55.826
541310 89.834 -57.531 55.817
541311 89.883 -57.548 55.810
541312 89.928 -57.562 55.804
541313 89.964 -57.572 55.797
541314 90.001 -57.583 55.790
541315 90.037 -57.592 55.783
541316 90.073 -57.602 55.775
541317 90.110 -57.612 55.765
541318 90.146 -57.622 55.754
541319 90.182 -57.632 55.741
541320 90.219 -57.643 55.728
541321 90.253 -57.659 55.711
541322 90.288 -57.675 55.692
541323 90.322 -57.688 55.669
541324 90.355 -57.699 55.645
541325 90.389 -57.711 55.617
541326 90.422 -57.721 55.590
541327 90.456 -57.733 55.564
541328 90.489 -57.744 55.538
541329 90.523 -57.755 55.514
541330 90.558 -57.769 55.493
541331 90.596 -57.785 55.473
541332 90.633 -57.802 55.448
541333 90.673 -57.814 55.421
541334 90.718 -57.821 55.395
541335 90.763 -57.827 55.372
541336 90.808 -57.834 55.351
541337 90.853 -57.840 55.331
541338 90.898 -57.847 55.310
541339 90.943 -57.854 55.295
541340 90.988 -57.860 55.282
541341 91.033 -57.867 55.268
541342 91.074 -57.886 55.252
541343 91.113 -57.909 55.239
541344 91.153 -57.934 55.228
541345 91.192 -57.957 55.218
541346 91.232 -57.981 55.212
541347 91.272 -58.004 55.209
541348 91.311 -58.028 55.206
541349 91.351 -58.052 55.197
541350 91.390 -58.075 55.182
541351 91.430 -58.099 55.164
541352 91.470 -58.123 55.146
541353 91.509 -58.147 55.138
541354 91.550 -58.167 55.132
541355 91.594 -58.178 55.130
541356 91.638 -58.187 55.128
541357 91.681 -58.198 55.124
541358 91.725 -58.209 55.116
541359 91.769 -58.219 55.106
541360 91.800 -58.241 55.094
541361 91.831 -58.263 55.079
541362 91.861 -58.285 55.062
541363 91.892 -58.307 55.043
541364 91.920 -58.317 55.022
541365 91.953 -58.327 54.999
541366 91.990 -58.337 54.975
541367 92.028 -58.347 54.950
541368 92.065 -58.357 54.927
541369 92.102 -58.368 54.905
541370 92.145 -58.378 54.886
541371 92.187 -58.388 54.866
541372 92.229 -58.397 54.846
541373 92.272 -58.408 54.830
541374 92.313 -58.416 54.819
541375 92.352 -58.419 54.807
541376 92.391 -58.422 54.791
541377 92.430 -58.426 54.769
541378 92.469 -58.429 54.745
541379 92.506 -58.433 54.723
541380 92.535 -58.438 54.706
541381 92.565 -58.443 54.705
541382 92.595 -58.448 54.710
541383 92.624 -58.454 54.720
541384 92.654 -58.459 54.732
541385 92.684 -58.464 54.742
541386 92.713 -58.469 54.749
541387 92.742 -58.475 54.774
541388 92.788 -58.481 54.802
541389 92.834 -58.487 54.826
541390 92.879 -58.493 54.842
541391 92.925 -58.498 54.854
541392 92.971 -58.504 54.855
541393 93.011 -58.513 54.857
541394 93.049 -58.523 54.857
541395 93.087 -58.532 54.850
541396 93.125 -58.542 54.837
541397 93.163 -58.551 54.824
541398 93.201 -58.561 54.812
541399 93.239 -58.570 54.797
541400 93.277 -58.580 54.780
541401 93.315 -58.590 54.761
541402 93.353 -58.599 54.739
541403 93.392 -58.607 54.719
541404 93.430 -58.615 54.698
541405 93.469 -58.623 54.675
541406 93.507 -58.630 54.655
541407 93.546 -58.639 54.648
541408 93.582 -58.652 54.645
541409 93.619 -58.666 54.637
541410 93.655 -58.680 54.625
541411 93.691 -58.694 54.610
541412 93.727 -58.708 54.590
541413 93.764 -58.721 54.568
541414 93.800 -58.735 54.551
541415 93.836 -58.748 54.538
541416 93.872 -58.762 54.529
541417 93.909 -58.776 54.521
541418 93.945 -58.789 54.513
541419 93.982 -58.803 54.503
541420 94.018 -58.817 54.487
541421 94.054 -58.830 54.469
541422 94.090 -58.844 54.449
541423 94.127 -58.858 54.425
541424 94.163 -58.871 54.394
541425 94.199 -58.885 54.354
541426 94.236 -58.899 54.320
541427 94.272 -58.913 54.292
541428 94.308 -58.927 54.271
541429 94.344 -58.940 54.253
541430 94.381 -58.954 54.235
541431 94.417 -58.968 54.216
541432 94.452 -58.987 54.191
541433 94.487 -59.007 54.162
541434 94.523 -59.015 54.131
541435 94.561 -59.024 54.095
541436 94.600 -59.034 54.056
541437 94.639 -59.045 54.022
541438 94.678 -59.055 53.988
541439 94.717 -59.066 53.951
541440 94.756 -59.077 53.911
541441 94.795 -59.088 53.871
541442 94.834 -59.099 53.833
541443 94.873 -59.109 53.793
541444 94.913 -59.121 53.753
541445 94.954 -59.131 53.715
541446 94.995 -59.142 53.679
541447 95.036 -59.153 53.648
541448 95.076 -59.163 53.618
541449 95.117 -59.175 53.592
541450 95.155 -59.185 53.569
541451 95.192 -59.195 53.546
541452 95.229 -59.206 53.524
541453 95.265 -59.216 53.501
541454 95.298 -59.231 53.480
541455 95.330 -59.245 53.460
541456 95.363 -59.258 53.442
541457 95.396 -59.269 53.424
541458 95.429 -59.280 53.406
541459 95.462 -59.291 53.388
541460 95.495 -59.303 53.371
541461 95.528 -59.313 53.354
541462 95.561 -59.325 53.340
541463 95.594 -59.336 53.330
541464 95.627 -59.347 53.326
541465 95.657 -59.356 53.325
541466 95.686 -59.364 53.322
541467 95.714 -59.372 53.314
541468 95.743 -59.380 53.306
541469 95.772 -59.388 53.295
541470 95.804 -59.397 53.283
541471 95.835 -59.406 53.271
541472 95.871 -59.412 53.257
541473 95.906 -59.418 53.238
541474 95.941 -59.424 53.222
541475 95.976 -59.430 53.207
541476 96.012 -59.435 53.193
541477 96.047 -59.441 53.179
541478 96.082 -59.447 53.165
541479 96.117 -59.453 53.148
541480 96.152 -59.460 53.124
541481 96.185 -59.466 53.097
541482 96.218 -59.473 53.067
541483 96.251 -59.480 53.035
541484 96.285 -59.487 53.006
541485 96.320 -59.491 52.981
541486 96.355 -59.494 52.956
541487 96.390 -59.498 52.931
541488 96.426 -59.502 52.907
541489 96.461 -59.506 52.882
541490 96.497 -59.509 52.856
541491 96.532 -59.513 52.828
541492 96.566 -59.515 52.800
541493 96.595 -59.512 52.774
541494 96.624 -59.509 52.750
541495 96.653 -59.505 52.724
541496 96.682 -59.502 52.691
541497 96.711 -59.499 52.658
541498 96.740 -59.496 52.621
541499 96.771 -59.494 52.582
541500 96.804 -59.493 52.544
541501 96.837 -59.495 52.503
541502 96.871 -59.497 52.461
541503 96.905 -59.500 52.420
541504 96.939 -59.502 52.379
541505 96.972 -59.504 52.337
541506 97.006 -59.507 52.293
541507 97.040 -59.509 52.246
541508 97.077 -59.512 52.202
541509 97.113 -59.515 52.164
541510 97.150 -59.518 52.132
541511 97.186 -59.521 52.106
541512 97.223 -59.524 52.080
541513 97.259 -59.527 52.049
541514 97.294 -59.530 52.015
541515 97.326 -59.534 51.975
541516 97.358 -59.537 51.930
541517 97.390 -59.542 51.881
541518 97.424 -59.549 51.828
541519 97.464 -59.560 51.774
541520 97.504 -59.571 51.719
541521 97.543 -59.582 51.665
541522 97.583 -59.593 51.611
541523 97.623 -59.604 51.557
541524 97.663 -59.615 51.505
541525 97.703 -59.626 51.456
541526 97.743 -59.638 51.406
541527 97.783 -59.648 51.354
541528 97.822 -59.660 51.297
541529 97.862 -59.671 51.237
541530 97.902 -59.682 51.177
541531 97.942 -59.693 51.117
541532 97.983 -59.700 51.060
541533 98.024 -59.704 51.007
541534 98.067 -59.707 50.957
541535 98.108 -59.710 50.911
541536 98.150 -59.714 50.867
541537 98.182 -59.703 50.825
541538 98.211 -59.690 50.785
541539 98.240 -59.676 50.747
541540 98.263 -59.679 50.711
541541 98.285 -59.685 50.674
541542 98.306 -59.692 50.639
541543 98.331 -59.698 50.607
541544 98.379 -59.707 50.576
541545 98.428 -59.715 50.546
541546 98.471 -59.710 50.518
541547 98.509 -59.695 50.488
541548 98.547 -59.680 50.455
541549 98.585 -59.665 50.418
541550 98.623 -59.649 50.379
541551 98.659 -59.633 50.339
541552 98.704 -59.635 50.297
541553 98.751 -59.645 50.254
541554 98.798 -59.654 50.208
541555 98.836 -59.642 50.160
541556 98.868 -59.620 50.112
541557 98.902 -59.598 50.066
541558 98.946 -59.583 50.021
541559 98.984 -59.585 49.976
541560 99.013 -59.589 49.932
541561 99.042 -59.593 49.888
541562 99.070 -59.597 49.846
541563 99.099 -59.601 49.805
541564 99.128 -59.605 49.764
541565 99.156 -59.609 49.723
541566 99.185 -59.613 49.683
541567 99.214 -59.616 49.645
541568 99.242 -59.620 49.607
541569 99.271 -59.624 49.570
541995 89.191 -57.282 56.042
541996 89.215 -57.307 56.025
541997 89.251 -57.321 56.009
541998 89.293 -57.332 55.992
541999 89.334 -57.343 55.974
542000 89.375 -57.354 55.958
542001 89.417 -57.364 55.942
542002 89.458 -57.376 55.926
542003 89.499 -57.386 55.910
542004 89.541 -57.397 55.894
542005 89.582 -57.408 55.877
542006 89.624 -57.419 55.864
542007 89.665 -57.429 55.852
542008 89.706 -57.441 55.840
542009 89.749 -57.452 55.830
542010 89.798 -57.469 55.820
542011 89.847 -57.486 55.811
542012 89.896 -57.503 55.803
542013 89.943 -57.519 55.795
542014 89.980 -57.529 55.788
542015 90.016 -57.539 55.780
542016 90.052 -57.548 55.771
542017 90.089 -57.559 55.761
542018 90.125 -57.569 55.750
542019 90.162 -57.579 55.737
542020 90.198 -57.589 55.723
542021 90.233 -57.605 55.707
542022 90.267 -57.621 55.688
542023 90.302 -57.638 55.666
542024 90.335 -57.649 55.644
542025 90.369 -57.661 55.620
542026 90.402 -57.671 55.593
542027 90.436 -57.683 55.566
542028 90.469 -57.694 55.540
542029 90.503 -57.705 55.514
542030 90.538 -57.717 55.490
542031 90.575 -57.734 55.468
542032 90.612 -57.751 55.448
542033 90.650 -57.767 55.425
542034 90.687 -57.783 55.402
542035 90.732 -57.789 55.379
542036 90.778 -57.796 55.359
542037 90.822 -57.802 55.340
542038 90.867 -57.809 55.323
542039 90.912 -57.816 55.305
542040 90.957 -57.822 55.290
542041 91.002 -57.829 55.277
542042 91.047 -57.835 55.265
542043 91.088 -57.856 55.253
542044 91.127 -57.880 55.243
542045 91.167 -57.903 55.236
542046 91.207 -57.927 55.229
542047 91.246 -57.951 55.224
542048 91.286 -57.974 55.220
542049 91.325 -57.998 55.220
542050 91.365 -58.021 55.215
542051 91.404 -58.046 55.204
542052 91.444 -58.069 55.187
542053 91.484 -58.093 55.165
542054 91.523 -58.116 55.148
542055 91.567 -58.128 55.136
542056 91.610 -58.138 55.131
542057 91.654 -58.149 55.130
542058 91.698 -58.159 55.130
542059 91.742 -58.169 55.126
542060 91.786 -58.180 55.119
542061 91.820 -58.199 55.107
542062 91.851 -58.221 55.092
542063 91.882 -58.243 55.074
542064 91.912 -58.265 55.053
542065 91.943 -58.288 55.027
542066 91.970 -58.297 55.000
542067 92.007 -58.307 54.973
542068 92.044 -58.317 54.948
542069 92.082 -58.328 54.924
542070 92.122 -58.338 54.904
542071 92.164 -58.347 54.889
542072 92.207 -58.358 54.876
542073 92.249 -58.368 54.865
542074 92.292 -58.378 54.856
542075 92.334 -58.388 54.845
542076 92.373 -58.391 54.829
542077 92.412 -58.394 54.807
542078 92.450 -58.397 54.783
542079 92.489 -58.400 54.759
542080 92.521 -58.405 54.734
542081 92.551 -58.410 54.710
542082 92.580 -58.416 54.703
542083 92.610 -58.421 54.706
542084 92.639 -58.426 54.713
542085 92.669 -58.432 54.722
542086 92.699 -58.437 54.731
542087 92.728 -58.442 54.746
542088 92.758 -58.448 54.776
542089 92.799 -58.453 54.807
542090 92.845 -58.459 54.830
542091 92.890 -58.465 54.846
542092 92.936 -58.471 54.860
542093 92.981 -58.477 54.862
542094 93.019 -58.487 54.863
542095 93.057 -58.496 54.863
542096 93.095 -58.506 54.858
542097 93.133 -58.516 54.850
542098 93.171 -58.525 54.841
542099 93.209 -58.535 54.831
542100 93.247 -58.545 54.817
542101 93.285 -58.554 54.799
542102 93.323 -58.563 54.780
542103 93.361 -58.573 54.760
542104 93.399 -58.582 54.743
542105 93.438 -58.589 54.723
542106 93.477 -58.597 54.699
542107 93.515 -58.604 54.675
542108 93.554 -58.614 54.658
542109 93.590 -58.628 54.657
542110 93.626 -58.642 54.651
542111 93.662 -58.655 54.639
542112 93.699 -58.669 54.621
542113 93.735 -58.683 54.603
542114 93.771 -58.696 54.583
542115 93.807 -58.710 54.564
542116 93.844 -58.723 54.553
542117 93.880 -58.737 54.545
542118 93.916 -58.751 54.538
542119 93.953 -58.764 54.530
542120 93.989 -58.778 54.521
542121 94.026 -58.792 54.509
542122 94.062 -58.805 54.494
542123 94.098 -58.820 54.473
542124 94.134 -58.833 54.447
542125 94.171 -58.847 54.414
542126 94.207 -58.861 54.373
542127 94.243 -58.874 54.333
542128 94.279 -58.888 54.299
542129 94.316 -58.902 54.274
542130 94.352 -58.915 54.259
542131 94.388 -58.929 54.244
542132 94.424 -58.943 54.225
542133 94.461 -58.956 54.203
542134 94.497 -58.971 54.177
542135 94.536 -58.981 54.144
542136 94.574 -58.992 54.105
542137 94.614 -59.003 54.066
542138 94.653 -59.014 54.030
542139 94.692 -59.024 53.995
542140 94.731 -59.035 53.960
542141 94.770 -59.046 53.924
542142 94.809 -59.057 53.888
542143 94.848 -59.068 53.850
542144 94.887 -59.078 53.810
542145 94.926 -59.090 53.770
542146 94.965 -59.100 53.732
542147 95.006 -59.111 53.696
542148 95.047 -59.121 53.665
542149 95.088 -59.133 53.636
542150 95.128 -59.143 53.608
542151 95.168 -59.154 53.583
542152 95.204 -59.165 53.561
542153 95.241 -59.175 53.541
542154 95.277 -59.187 53.518
542155 95.309 -59.201 53.497
542156 95.341 -59.216 53.479
542157 95.373 -59.231 53.462
542158 95.406 -59.241 53.443
542159 95.440 -59.253 53.425
542160 95.472 -59.264 53.407
542161 95.506 -59.275 53.390
542162 95.539 -59.286 53.374
542163 95.572 -59.297 53.357
542164 95.605 -59.308 53.342
542165 95.634 -59.317 53.336
542166 95.663 -59.325 53.335
542167 95.692 -59.334 53.333
542168 95.720 -59.341 53.326
542169 95.749 -59.350 53.315
542170 95.777 -59.358 53.305
542171 95.807 -59.366 53.288
542172 95.841 -59.372 53.272
542173 95.877 -59.379 53.257
542174 95.912 -59.385 53.237
542175 95.947 -59.391 53.223
542176 95.982 -59.397 53.211
542177 96.018 -59.403 53.198
542178 96.053 -59.409 53.185
542179 96.088 -59.415 53.171
542180 96.123 -59.421 53.152
542181 96.157 -59.427 53.127
542182 96.190 -59.434 53.101
542183 96.224 -59.441 53.072
542184 96.257 -59.447 53.041
542185 96.290 -59.454 53.014
542186 96.323 -59.461 52.990
542187 96.358 -59.466 52.966
542188 96.393 -59.469 52.942
542189 96.429 -59.473 52.919
542190 96.464 -59.477 52.893
542191 96.500 -59.481 52.866
542192 96.535 -59.485 52.838
542193 96.566 -59.483 52.811
542194 96.595 -59.480 52.785
542195 96.624 -59.477 52.760
542196 96.653 -59.474 52.734
542197 96.682 -59.470 52.701
542198 96.711 -59.467 52.667
542199 96.740 -59.464 52.630
542200 96.769 -59.461 52.591
542201 96.800 -59.459 52.553
542202 96.834 -59.461 52.513
542203 96.868 -59.463 52.469
542204 96.902 -59.466 52.427
542205 96.939 -59.469 52.383
542206 96.975 -59.472 52.337
542207 97.012 -59.475 52.292
542208 97.048 -59.478 52.249
542209 97.085 -59.481 52.212
542210 97.121 -59.484 52.179
542211 97.158 -59.487 52.148
542212 97.194 -59.490 52.119
542213 97.231 -59.493 52.089
542214 97.264 -59.497 52.055
542215 97.297 -59.500 52.018
542216 97.329 -59.504 51.975
542217 97.361 -59.507 51.927
542218 97.393 -59.512 51.877
542219 97.425 -59.516 51.823
542220 97.461 -59.524 51.768
542221 97.501 -59.535 51.713
542222 97.540 -59.547 51.659
542223 97.580 -59.557 51.607
542224 97.620 -59.569 51.556
542225 97.660 -59.580 51.509
542226 97.700 -59.591 51.464
542227 97.739 -59.602 51.417
542228 97.779 -59.613 51.366
542229 97.819 -59.624 51.309
542230 97.859 -59.635 51.248
542231 97.898 -59.646 51.188
542232 97.938 -59.657 51.127
542233 97.980 -59.663 51.068
542234 98.022 -59.666 51.015
542235 98.064 -59.669 50.965
542236 98.106 -59.673 50.918
542237 98.148 -59.676 50.873
542238 98.181 -59.668 50.830
542239 98.211 -59.654 50.789
542240 98.239 -59.641 50.750
542241 98.266 -59.635 50.712
542242 98.287 -59.641 50.676
542243 98.309 -59.648 50.640
542244 98.336 -59.654 50.606
542245 98.385 -59.663 50.573
542246 98.433 -59.672 50.542
542247 98.474 -59.669 50.513
542248 98.511 -59.653 50.484
542249 98.548 -59.636 50.453
542250 98.585 -59.620 50.419
542251 98.622 -59.604 50.382
542252 98.660 -59.590 50.342
542253 98.703 -59.585 50.302
542254 98.745 -59.579 50.260
542255 98.791 -59.584 50.214
542256 98.838 -59.594 50.167
542257 98.874 -59.579 50.120
542258 98.907 -59.557 50.076
542259 98.954 -59.544 50.030
542260 98.983 -59.548 49.985
542261 99.011 -59.552 49.941
542262 99.040 -59.556 49.898
542263 99.069 -59.560 49.856
542264 99.097 -59.563 49.814
542265 99.126 -59.568 49.774
542266 99.155 -59.572 49.734
542267 99.184 -59.576 49.695
542268 99.212 -59.579 49.656
542269 99.241 -59.583 49.618
542270 99.270 -59.587 49.582
542625 86.443 -56.295 57.556
542626 86.484 -56.309 57.538
542627 86.524 -56.323 57.516
542628 86.564 -56.335 57.491
542629 86.605 -56.349 57.466
542630 86.645 -56.362 57.438
542631 86.686 -56.376 57.411
542632 86.726 -56.389 57.382
542633 86.766 -56.402 57.351
542634 86.807 -56.416 57.317
542635 86.847 -56.429 57.283
542636 86.887 -56.442 57.249
542637 86.928 -56.456 57.216
542638 86.969 -56.469 57.183
542698 89.270 -57.282 56.000
542699 89.312 -57.292 55.983
542700 89.353 -57.304 55.966
542701 89.395 -57.314 55.951
542702 89.436 -57.325 55.936
542703 89.477 -57.335 55.920
542704 89.519 -57.347 55.905
542705 89.560 -57.357 55.891
542706 89.602 -57.368 55.875
542707 89.643 -57.379 55.861
542708 89.685 -57.390 55.848
542709 89.726 -57.401 55.836
542710 89.767 -57.411 55.825
542711 89.811 -57.424 55.815
542712 89.860 -57.441 55.806
542713 89.909 -57.458 55.796
542714 89.958 -57.475 55.788
542715 89.996 -57.485 55.779
542716 90.032 -57.495 55.770
542717 90.068 -57.505 55.759
542718 90.105 -57.515 55.747
542719 90.141 -57.525 55.734
542720 90.177 -57.536 55.720
542721 90.212 -57.551 55.704
542722 90.247 -57.567 55.687
542723 90.281 -57.584 55.667
542724 90.316 -57.599 55.645
542725 90.349 -57.611 55.621
542726 90.383 -57.621 55.598
542727 90.416 -57.633 55.572
542728 90.450 -57.644 55.545
542729 90.483 -57.655 55.519
542730 90.517 -57.667 55.493
542731 90.554 -57.683 55.469
542732 90.592 -57.699 55.446
542733 90.629 -57.716 55.426
542734 90.667 -57.732 55.405
542735 90.705 -57.749 55.384
542736 90.747 -57.758 55.365
542737 90.792 -57.764 55.346
542738 90.836 -57.771 55.330
542739 90.882 -57.778 55.317
542740 90.927 -57.784 55.302
542741 90.971 -57.791 55.288
542742 91.017 -57.797 55.276
542743 91.062 -57.804 55.268
542744 91.101 -57.826 55.263
542745 91.141 -57.849 55.260
542746 91.181 -57.874 55.258
542747 91.220 -57.897 55.255
542748 91.260 -57.921 55.250
542749 91.300 -57.944 55.242
542750 91.339 -57.968 55.233
542751 91.379 -57.992 55.226
542752 91.419 -58.015 55.219
542753 91.458 -58.039 55.209
542754 91.498 -58.063 55.194
542755 91.540 -58.078 55.174
542756 91.584 -58.088 55.153
542757 91.628 -58.099 55.135
542758 91.671 -58.109 55.126
542759 91.715 -58.120 55.127
542760 91.759 -58.131 55.128
542761 91.803 -58.141 55.128
542762 91.841 -58.156 55.120
542763 91.871 -58.178 55.105
542764 91.902 -58.201 55.085
542765 91.933 -58.223 55.062
542766 91.963 -58.245 55.033
542767 91.994 -58.267 55.005
542768 92.024 -58.277 54.978
542769 92.062 -58.288 54.952
542770 92.100 -58.297 54.927
542771 92.142 -58.308 54.905
542772 92.184 -58.318 54.888
542773 92.227 -58.328 54.876
542774 92.269 -58.338 54.866
542775 92.311 -58.348 54.858
542776 92.354 -58.358 54.848
542777 92.393 -58.363 54.834
542778 92.432 -58.366 54.814
542779 92.471 -58.369 54.792
542780 92.507 -58.372 54.767
542781 92.536 -58.378 54.744
542782 92.566 -58.383 54.727
542783 92.595 -58.388 54.710
542784 92.625 -58.394 54.702
542785 92.654 -58.399 54.704
542786 92.684 -58.404 54.712
542787 92.714 -58.410 54.719
542788 92.743 -58.415 54.744
542789 92.773 -58.420 54.774
542790 92.810 -58.426 54.802
542791 92.855 -58.432 54.824
542792 92.901 -58.438 54.841
542793 92.946 -58.444 54.857
542794 92.989 -58.451 54.860
542795 93.027 -58.460 54.861
542796 93.065 -58.470 54.860
542797 93.103 -58.479 54.856
542798 93.141 -58.489 54.853
542799 93.179 -58.499 54.848
542800 93.217 -58.508 54.841
542801 93.255 -58.518 54.828
542802 93.293 -58.528 54.811
542803 93.331 -58.537 54.792
542804 93.369 -58.547 54.772
542805 93.407 -58.556 54.757
542806 93.446 -58.563 54.739
542807 93.485 -58.571 54.717
542808 93.523 -58.579 54.691
542809 93.561 -58.589 54.668
542810 93.597 -58.603 54.666
542811 93.634 -58.617 54.661
542812 93.670 -58.630 54.649
542813 93.706 -58.644 54.631
542814 93.743 -58.657 54.609
542815 93.779 -58.671 54.585
542816 93.815 -58.685 54.569
542817 93.851 -58.698 54.563
542818 93.888 -58.712 54.561
542819 93.924 -58.726 54.556
542820 93.960 -58.740 54.550
542821 93.996 -58.754 54.540
542822 94.033 -58.767 54.526
542823 94.069 -58.781 54.510
542824 94.106 -58.795 54.490
542825 94.142 -58.808 54.461
542826 94.178 -58.822 54.427
542827 94.214 -58.836 54.387
542828 94.251 -58.849 54.344
542829 94.287 -58.863 54.305
542830 94.323 -58.877 54.276
542831 94.359 -58.890 54.263
542832 94.396 -58.904 54.250
542833 94.431 -58.916 54.232
542834 94.465 -58.927 54.212
542835 94.498 -58.940 54.186
542836 94.540 -58.950 54.153
542837 94.582 -58.961 54.113
542838 94.625 -58.972 54.076
542839 94.667 -58.983 54.039
542840 94.706 -58.993 54.006
542841 94.745 -59.005 53.975
542842 94.784 -59.015 53.942
542843 94.823 -59.026 53.906
542844 94.862 -59.037 53.868
542845 94.901 -59.048 53.828
542846 94.940 -59.059 53.788
542847 94.979 -59.069 53.750
542848 95.018 -59.080 53.715
542849 95.058 -59.090 53.684
542850 95.099 -59.102 53.655
542851 95.139 -59.112 53.627
542852 95.180 -59.123 53.603
542853 95.217 -59.134 53.580
542854 95.254 -59.144 53.560
542855 95.288 -59.157 53.537
542856 95.320 -59.172 53.515
542857 95.352 -59.187 53.496
542858 95.384 -59.202 53.479
542859 95.417 -59.214 53.462
542860 95.450 -59.225 53.445
542861 95.483 -59.236 53.427
542862 95.516 -59.247 53.411
542863 95.549 -59.259 53.396
542864 95.582 -59.269 53.381
542865 95.612 -59.278 53.366
542866 95.640 -59.287 53.354
542867 95.669 -59.295 53.344
542868 95.697 -59.303 53.338
542869 95.726 -59.311 53.336
542870 95.754 -59.319 53.324
542871 95.783 -59.327 53.312
542872 95.812 -59.334 53.292
542873 95.847 -59.340 53.273
542874 95.883 -59.346 53.259
542875 95.918 -59.352 53.239
542876 95.954 -59.358 53.226
542877 95.989 -59.365 53.214
542878 96.024 -59.371 53.203
542879 96.059 -59.376 53.190
542880 96.095 -59.382 53.174
542881 96.129 -59.388 53.153
542882 96.163 -59.395 53.129
542883 96.196 -59.402 53.105
542884 96.229 -59.409 53.078
542885 96.262 -59.416 53.048
542886 96.295 -59.422 53.021
542887 96.328 -59.429 52.996
542888 96.362 -59.436 52.974
542889 96.396 -59.441 52.952
542890 96.431 -59.444 52.929
542891 96.467 -59.448 52.904
542892 96.502 -59.452 52.875
542893 96.536 -59.455 52.847
542894 96.566 -59.451 52.822
542895 96.595 -59.448 52.796
542896 96.623 -59.445 52.769
542897 96.653 -59.442 52.742
542898 96.682 -59.439 52.708
542899 96.711 -59.435 52.674
542900 96.740 -59.432 52.637
542901 96.769 -59.429 52.597
542902 96.800 -59.427 52.557
542903 96.836 -59.429 52.516
542904 96.873 -59.432 52.471
542905 96.910 -59.435 52.426
542906 96.947 -59.438 52.381
542907 96.983 -59.441 52.338
542908 97.020 -59.444 52.298
542909 97.056 -59.447 52.261
542910 97.093 -59.450 52.227
542911 97.130 -59.453 52.194
542912 97.166 -59.457 52.162
542913 97.203 -59.460 52.130
542914 97.236 -59.463 52.095
542915 97.268 -59.466 52.057
542916 97.300 -59.469 52.015
542917 97.332 -59.473 51.969
542918 97.365 -59.477 51.920
542919 97.396 -59.482 51.870
542920 97.428 -59.486 51.817
542921 97.460 -59.491 51.762
542922 97.497 -59.500 51.706
542923 97.537 -59.511 51.652
542924 97.577 -59.522 51.602
542925 97.617 -59.533 51.556
542926 97.657 -59.544 51.512
542927 97.697 -59.555 51.467
542928 97.736 -59.566 51.420
542929 97.776 -59.578 51.368
542930 97.816 -59.588 51.311
542931 97.856 -59.600 51.251
542932 97.895 -59.611 51.191
542933 97.935 -59.621 51.132
542934 97.978 -59.625 51.078
542935 98.019 -59.628 51.026
542936 98.061 -59.632 50.977
542937 98.103 -59.635 50.928
542938 98.145 -59.638 50.882
542939 98.181 -59.633 50.837
542940 98.210 -59.620 50.794
542941 98.239 -59.606 50.754
542942 98.268 -59.592 50.715
542943 98.290 -59.597 50.679
542944 98.312 -59.604 50.642
542945 98.342 -59.610 50.607
542946 98.390 -59.620 50.573
542947 98.436 -59.626 50.541
542948 98.473 -59.623 50.509
542949 98.510 -59.607 50.478
542950 98.547 -59.591 50.447
542951 98.584 -59.575 50.413
542952 98.621 -59.559 50.377
542953 98.661 -59.547 50.339
542954 98.704 -59.542 50.300
542955 98.746 -59.536 50.259
542956 98.789 -59.530 50.215
542957 98.831 -59.525 50.171
542958 98.878 -59.533 50.129
542959 98.913 -59.516 50.087
542960 98.952 -59.507 50.040
542961 98.981 -59.511 49.994
542962 99.010 -59.515 49.950
542963 99.039 -59.519 49.906
542964 99.067 -59.522 49.865
542965 99.096 -59.526 49.824
542966 99.125 -59.531 49.784
542967 99.153 -59.535 49.745
542968 99.182 -59.538 49.706
542969 99.211 -59.542 49.668
542970 99.239 -59.546 49.631
542971 99.278 -59.547 49.594
542972 99.350 -59.542 49.558
543323 86.339 -56.212 57.610
543324 86.379 -56.226 57.596
543325 86.420 -56.239 57.581
543326 86.460 -56.252 57.565
543327 86.500 -56.266 57.547
543328 86.541 -56.279 57.526
543329 86.581 -56.292 57.503
543330 86.622 -56.306 57.477
543331 86.662 -56.319 57.451
543332 86.702 -56.332 57.424
543333 86.743 -56.345 57.395
543334 86.783 -56.359 57.364
543335 86.824 -56.373 57.329
543336 86.864 -56.385 57.295
543337 86.905 -56.399 57.261
543338 86.945 -56.412 57.227
543339 86.985 -56.426 57.194
543340 87.025 -56.439 57.160
543402 89.414 -57.275 55.945
543403 89.456 -57.285 55.931
543404 89.497 -57.296 55.916
543405 89.538 -57.307 55.902
543406 89.580 -57.318 55.887
543407 89.621 -57.329 55.871
543408 89.663 -57.339 55.858
543409 89.704 -57.351 55.845
543410 89.745 -57.361 55.834
543411 89.787 -57.372 55.822
543412 89.828 -57.382 55.811
543413 89.873 -57.396 55.801
543414 89.922 -57.413 55.791
543415 89.971 -57.430 55.781
543416 90.011 -57.442 55.770
543417 90.047 -57.452 55.759
543418 90.084 -57.462 55.747
543419 90.120 -57.472 55.734
543420 90.157 -57.482 55.719
543421 90.191 -57.498 55.703
543422 90.226 -57.514 55.687
543423 90.261 -57.529 55.669
543424 90.295 -57.546 55.648
543425 90.329 -57.561 55.625
543426 90.363 -57.571 55.601
543427 90.396 -57.583 55.577
543428 90.430 -57.594 55.552
543429 90.463 -57.605 55.526
543430 90.497 -57.616 55.501
543431 90.534 -57.632 55.476
543432 90.572 -57.648 55.451
543433 90.609 -57.664 55.429
543434 90.647 -57.681 55.409
543435 90.684 -57.697 55.389
543436 90.721 -57.714 55.370
543437 90.761 -57.727 55.352
543438 90.806 -57.733 55.335
543439 90.851 -57.740 55.323
543440 90.896 -57.746 55.314
543441 90.941 -57.753 55.301
543442 90.986 -57.759 55.288
543443 91.031 -57.766 55.278
543444 91.076 -57.772 55.273
543445 91.115 -57.796 55.275
543446 91.155 -57.820 55.277
543447 91.194 -57.843 55.277
543448 91.234 -57.867 55.276
543449 91.274 -57.891 55.272
543450 91.313 -57.915 55.263
543451 91.353 -57.938 55.249
543452 91.393 -57.962 55.240
543453 91.432 -57.986 55.236
543454 91.472 -58.009 55.232
543455 91.513 -58.028 55.220
543456 91.557 -58.039 55.201
543457 91.600 -58.050 55.178
543458 91.644 -58.060 55.155
543459 91.688 -58.070 55.135
543460 91.732 -58.081 55.128
543461 91.776 -58.091 55.134
543462 91.819 -58.102 55.134
543463 91.861 -58.115 55.126
543464 91.891 -58.137 55.109
543465 91.922 -58.159 55.088
543466 91.953 -58.181 55.066
543467 91.984 -58.203 55.044
543468 92.014 -58.225 55.020
543469 92.044 -58.247 54.991
543470 92.079 -58.257 54.961
543471 92.120 -58.268 54.934
543472 92.162 -58.278 54.908
543473 92.204 -58.288 54.886
543474 92.247 -58.297 54.870
543475 92.289 -58.308 54.859
543476 92.332 -58.318 54.849
543477 92.374 -58.328 54.840
543478 92.414 -58.334 54.827
543479 92.453 -58.337 54.812
543480 92.492 -58.340 54.792
543481 92.522 -58.345 54.771
543482 92.552 -58.350 54.749
543483 92.581 -58.356 54.735
543484 92.611 -58.361 54.717
543485 92.640 -58.366 54.703
543486 92.670 -58.372 54.696
543487 92.699 -58.377 54.698
543488 92.729 -58.382 54.704
543489 92.759 -58.388 54.725
543490 92.788 -58.393 54.749
543491 92.820 -58.398 54.774
543492 92.866 -58.404 54.798
543493 92.912 -58.410 54.820
543494 92.957 -58.416 54.837
543495 92.997 -58.425 54.844
543496 93.036 -58.434 54.847
543497 93.073 -58.444 54.845
543498 93.111 -58.454 54.842
543499 93.149 -58.463 54.843
543500 93.187 -58.473 54.844
543501 93.225 -58.482 54.840
543502 93.263 -58.491 54.831
543503 93.301 -58.501 54.816
543504 93.339 -58.511 54.797
543505 93.377 -58.520 54.776
543506 93.415 -58.530 54.762
543507 93.454 -58.538 54.746
543508 93.492 -58.545 54.726
543509 93.531 -58.553 54.700
543510 93.568 -58.564 54.675
543511 93.605 -58.578 54.671
543512 93.641 -58.592 54.666
543513 93.677 -58.605 54.655
543514 93.714 -58.619 54.637
543515 93.750 -58.633 54.614
543516 93.786 -58.647 54.588
543517 93.823 -58.661 54.571
543518 93.859 -58.674 54.565
543519 93.895 -58.688 54.563
543520 93.931 -58.701 54.561
543521 93.968 -58.715 54.558
543522 94.004 -58.729 54.550
543523 94.040 -58.742 54.536
543524 94.076 -58.756 54.518
543525 94.113 -58.770 54.495
543526 94.149 -58.783 54.464
543527 94.186 -58.797 54.428
543528 94.222 -58.811 54.388
543529 94.258 -58.824 54.348
543530 94.294 -58.838 54.310
543531 94.330 -58.851 54.280
543532 94.364 -58.861 54.262
543533 94.398 -58.873 54.251
543534 94.432 -58.884 54.236
543535 94.467 -58.896 54.215
543536 94.499 -58.908 54.187
543537 94.536 -58.920 54.156
543538 94.578 -58.930 54.121
543539 94.621 -58.941 54.087
543540 94.664 -58.952 54.053
543541 94.706 -58.963 54.022
543542 94.749 -58.974 53.995
543543 94.791 -58.984 53.962
543544 94.834 -58.995 53.926
543545 94.876 -59.005 53.886
543546 94.915 -59.017 53.846
543547 94.954 -59.027 53.807
543548 94.993 -59.038 53.769
543549 95.032 -59.049 53.734
543550 95.071 -59.060 53.703
543551 95.110 -59.071 53.674
543552 95.150 -59.081 53.647
543553 95.191 -59.092 53.623
543554 95.230 -59.102 53.600
543555 95.267 -59.113 53.578
543556 95.299 -59.128 53.556
543557 95.331 -59.143 53.534
543558 95.363 -59.158 53.515
543559 95.395 -59.172 53.498
543560 95.427 -59.187 53.481
543561 95.460 -59.198 53.466
543562 95.493 -59.209 53.450
543563 95.527 -59.220 53.434
543564 95.559 -59.231 53.419
543565 95.589 -59.240 53.405
543566 95.617 -59.248 53.387
543567 95.646 -59.256 53.371
543568 95.675 -59.264 53.356
543569 95.703 -59.272 53.343
543570 95.732 -59.280 53.334
543571 95.760 -59.287 53.323
543572 95.789 -59.295 53.310
543573 95.818 -59.302 53.289
543574 95.853 -59.308 53.270
543575 95.889 -59.314 53.255
543576 95.924 -59.320 53.237
543577 95.960 -59.325 53.226
543578 95.995 -59.331 53.215
543579 96.030 -59.338 53.204
543580 96.065 -59.344 53.190
543581 96.101 -59.350 53.173
543582 96.135 -59.356 53.152
543583 96.168 -59.363 53.131
543584 96.201 -59.369 53.107
543585 96.234 -59.376 53.080
543586 96.267 -59.383 53.052
543587 96.301 -59.390 53.025
543588 96.334 -59.397 53.000
543589 96.367 -59.403 52.977
543590 96.400 -59.410 52.958
543591 96.434 -59.416 52.937
543592 96.469 -59.420 52.912
543593 96.505 -59.424 52.885
543594 96.536 -59.423 52.857
543595 96.566 -59.420 52.832
543596 96.594 -59.416 52.806
543597 96.623 -59.413 52.777
543598 96.652 -59.410 52.749
543599 96.682 -59.407 52.714
543600 96.713 -59.405 52.677
543601 96.743 -59.403 52.640
543602 96.773 -59.400 52.600
543603 96.805 -59.398 52.556
543604 96.839 -59.400 52.512
543605 96.876 -59.403 52.467
543606 96.913 -59.406 52.420
543607 96.950 -59.409 52.377
543608 96.987 -59.411 52.341
543609 97.025 -59.414 52.309
543610 97.062 -59.417 52.276
543611 97.099 -59.420 52.242
543612 97.136 -59.423 52.208
543613 97.173 -59.425 52.173
543614 97.207 -59.429 52.136
543615 97.239 -59.432 52.098
543616 97.271 -59.436 52.055
543617 97.303 -59.440 52.009
543618 97.336 -59.443 51.961
543619 97.368 -59.447 51.911
543620 97.400 -59.451 51.860
543621 97.432 -59.456 51.808
543622 97.463 -59.460 51.755
543623 97.495 -59.465 51.699
543624 97.534 -59.475 51.645
543625 97.574 -59.487 51.595
543626 97.613 -59.497 51.550
543627 97.653 -59.509 51.507
543628 97.693 -59.520 51.463
543629 97.733 -59.531 51.416
543630 97.773 -59.542 51.364
543631 97.812 -59.554 51.308
543632 97.852 -59.564 51.249
543633 97.892 -59.576 51.191
543634 97.933 -59.584 51.136
543635 97.974 -59.587 51.085
543636 98.017 -59.590 51.037
543637 98.059 -59.594 50.989
543638 98.100 -59.597 50.939
543639 98.143 -59.601 50.891
543640 98.180 -59.598 50.845
543641 98.209 -59.585 50.801
543642 98.238 -59.571 50.759
543643 98.268 -59.557 50.720
543644 98.293 -59.553 50.682
543645 98.314 -59.560 50.645
543646 98.347 -59.567 50.609
543647 98.395 -59.576 50.574
543648 98.435 -59.576 50.540
543649 98.473 -59.573 50.505
543650 98.510 -59.562 50.473
543651 98.547 -59.546 50.441
543652 98.584 -59.529 50.407
543653 98.621 -59.513 50.371
543654 98.662 -59.505 50.335
543655 98.705 -59.500 50.298
543656 98.747 -59.494 50.259
543657 98.790 -59.488 50.218
543658 98.832 -59.482 50.177
543659 98.875 -59.476 50.137
543660 98.918 -59.472 50.096
543661 98.950 -59.470 50.050
543662 98.980 -59.474 50.003
543663 99.008 -59.478 49.958
543664 99.037 -59.482 49.915
543665 99.066 -59.485 49.874
543666 99.094 -59.489 49.833
543667 99.123 -59.494 49.794
543668 99.152 -59.497 49.755
543669 99.181 -59.501 49.716
543670 99.209 -59.505 49.678
543671 99.238 -59.509 49.642
543672 99.286 -59.509 49.606
543673 99.358 -59.503 49.570
544022 86.275 -56.142 57.640
544023 86.315 -56.156 57.630
544024 86.356 -56.169 57.617
544025 86.396 -56.182 57.604
544026 86.436 -56.195 57.589
544027 86.477 -56.209 57.574
544028 86.517 -56.222 57.556
544029 86.558 -56.235 57.536
544030 86.598 -56.249 57.514
544031 86.638 -56.262 57.490
544032 86.679 -56.276 57.464
544033 86.719 -56.289 57.437
544034 86.760 -56.302 57.409
544035 86.800 -56.316 57.378
544036 86.841 -56.329 57.344
544037 86.881 -56.342 57.310
544038 86.921 -56.356 57.275
544039 86.962 -56.369 57.241
544040 87.002 -56.382 57.206
544041 87.043 -56.395 57.171
544042 87.083 -56.409 57.136
544107 89.599 -57.278 55.884
544108 89.640 -57.289 55.869
544109 89.682 -57.300 55.855
544110 89.724 -57.310 55.843
544111 89.765 -57.322 55.832
544112 89.806 -57.332 55.820
544113 89.848 -57.343 55.808
544114 89.889 -57.354 55.796
544115 89.935 -57.368 55.785
544116 89.984 -57.385 55.775
544117 90.027 -57.398 55.763
544118 90.063 -57.409 55.751
544119 90.099 -57.419 55.737
544120 90.136 -57.429 55.722
544121 90.171 -57.444 55.705
544122 90.205 -57.460 55.688
544123 90.240 -57.476 55.671
544124 90.274 -57.492 55.651
544125 90.309 -57.508 55.630
544126 90.343 -57.521 55.607
544127 90.377 -57.532 55.583
544128 90.410 -57.543 55.559
544129 90.444 -57.554 55.534
544130 90.477 -57.566 55.508
544131 90.514 -57.581 55.484
544132 90.551 -57.597 55.459
544133 90.589 -57.614 55.435
544134 90.626 -57.630 55.414
544135 90.663 -57.646 55.395
544136 90.701 -57.663 55.375
544137 90.738 -57.679 55.357
544138 90.776 -57.695 55.341
544139 90.820 -57.702 55.324
544140 90.866 -57.708 55.312
544141 90.910 -57.715 55.304
544142 90.955 -57.721 55.296
544143 91.000 -57.728 55.288
544144 91.045 -57.735 55.282
544145 91.090 -57.742 55.279
544146 91.130 -57.766 55.281
544147 91.169 -57.789 55.284
544148 91.209 -57.814 55.287
544149 91.248 -57.837 55.288
544150 91.288 -57.861 55.286
544151 91.327 -57.884 55.280
544152 91.367 -57.909 55.266
544153 91.406 -57.932 55.257
544154 91.446 -57.955 55.253
544155 91.486 -57.979 55.250
544156 91.530 -57.989 55.240
544157 91.574 -57.999 55.222
544158 91.617 -58.010 55.200
544159 91.661 -58.021 55.175
544160 91.705 -58.031 55.152
544161 91.749 -58.042 55.139
544162 91.793 -58.052 55.136
544163 91.837 -58.062 55.133
544164 91.880 -58.073 55.126
544165 91.912 -58.094 55.113
544166 91.942 -58.116 55.093
544167 91.973 -58.139 55.070
544168 92.004 -58.161 55.048
544169 92.034 -58.183 55.024
544170 92.065 -58.205 54.997
544171 92.097 -58.226 54.969
544172 92.140 -58.238 54.937
544173 92.182 -58.247 54.906
544174 92.224 -58.258 54.880
544175 92.267 -58.268 54.860
544176 92.309 -58.278 54.844
544177 92.351 -58.288 54.831
544178 92.394 -58.298 54.820
544179 92.435 -58.305 54.809
544180 92.474 -58.308 54.797
544181 92.507 -58.313 54.785
544182 92.537 -58.318 54.770
544183 92.567 -58.323 54.752
544184 92.596 -58.328 54.736
544185 92.626 -58.334 54.721
544186 92.655 -58.339 54.704
544187 92.685 -58.344 54.691
544188 92.714 -58.350 54.688
544189 92.744 -58.355 54.692
544190 92.774 -58.360 54.706
544191 92.803 -58.365 54.724
544192 92.833 -58.370 54.745
544193 92.877 -58.376 54.767
544194 92.922 -58.382 54.789
544195 92.968 -58.388 54.806
544196 93.005 -58.398 54.816
544197 93.044 -58.408 54.820
544198 93.081 -58.417 54.818
544199 93.119 -58.427 54.814
544200 93.157 -58.436 54.815
544201 93.195 -58.446 54.817
544202 93.233 -58.456 54.818
544203 93.271 -58.465 54.814
544204 93.309 -58.475 54.804
544205 93.347 -58.485 54.788
544206 93.385 -58.494 54.769
544207 93.423 -58.504 54.750
544208 93.462 -58.512 54.735
544209 93.500 -58.520 54.720
544210 93.539 -58.527 54.700
544211 93.576 -58.540 54.676
544212 93.612 -58.554 54.670
544213 93.648 -58.567 54.663
544214 93.685 -58.581 54.653
544215 93.721 -58.595 54.637
544216 93.758 -58.608 54.616
544217 93.794 -58.622 54.590
544218 93.830 -58.636 54.571
544219 93.866 -58.649 54.560
544220 93.903 -58.663 54.558
544221 93.939 -58.676 54.559
544222 93.975 -58.690 54.558
544223 94.011 -58.704 54.552
544224 94.048 -58.717 54.540
544225 94.084 -58.731 54.522
544226 94.120 -58.745 54.498
544227 94.156 -58.759 54.467
544228 94.193 -58.773 54.430
544229 94.228 -58.785 54.391
544230 94.262 -58.796 54.351
544231 94.297 -58.807 54.315
544232 94.331 -58.818 54.284
544233 94.365 -58.830 54.261
544234 94.400 -58.841 54.250
544235 94.434 -58.852 54.235
544236 94.468 -58.863 54.216
544237 94.499 -58.876 54.190
544238 94.533 -58.889 54.162
544239 94.575 -58.899 54.130
544240 94.618 -58.911 54.097
544241 94.661 -58.921 54.066
544242 94.703 -58.932 54.038
544243 94.746 -58.943 54.011
544244 94.788 -58.953 53.980
544245 94.831 -58.965 53.944
544246 94.873 -58.975 53.905
544247 94.916 -58.986 53.865
544248 94.959 -58.996 53.826
544249 95.001 -59.007 53.787
544250 95.044 -59.018 53.755
544251 95.085 -59.029 53.721
544252 95.124 -59.040 53.691
544253 95.163 -59.050 53.666
544254 95.202 -59.061 53.644
544255 95.243 -59.072 53.619
544256 95.278 -59.084 53.597
544257 95.310 -59.099 53.575
544258 95.342 -59.113 53.553
544259 95.374 -59.128 53.534
544260 95.406 -59.143 53.519
544261 95.438 -59.158 53.500
544262 95.471 -59.170 53.486
544263 95.504 -59.181 53.473
544264 95.537 -59.193 53.458
544265 95.566 -59.201 53.441
544266 95.595 -59.209 53.426
544267 95.623 -59.217 53.406
544268 95.652 -59.225 53.388
544269 95.680 -59.232 53.369
544270 95.708 -59.240 53.349
544271 95.737 -59.247 53.330
544272 95.765 -59.255 53.317
544273 95.794 -59.262 53.302
544274 95.824 -59.270 53.282
544275 95.860 -59.275 53.263
544276 95.895 -59.281 53.247
544277 95.931 -59.287 53.230
544278 95.966 -59.294 53.221
544279 96.001 -59.300 53.210
544280 96.036 -59.306 53.199
544281 96.072 -59.311 53.186
544282 96.107 -59.317 53.168
544283 96.140 -59.324 53.149
544284 96.173 -59.331 53.127
544285 96.206 -59.338 53.104
544286 96.240 -59.344 53.079
544287 96.273 -59.351 53.052
544288 96.306 -59.358 53.026
544289 96.339 -59.365 53.001
544290 96.372 -59.372 52.977
544291 96.405 -59.378 52.958
544292 96.439 -59.385 52.939
544293 96.472 -59.391 52.914
544294 96.507 -59.394 52.889
544295 96.536 -59.391 52.862
544296 96.566 -59.388 52.835
544297 96.596 -59.386 52.808
544298 96.626 -59.383 52.778
544299 96.656 -59.381 52.746
544300 96.687 -59.378 52.710
544301 96.717 -59.375 52.671
544302 96.747 -59.373 52.628
544303 96.778 -59.371 52.583
544304 96.810 -59.370 52.538
544305 96.843 -59.371 52.497
544306 96.879 -59.373 52.455
544307 96.916 -59.375 52.417
544308 96.953 -59.378 52.383
544309 96.991 -59.381 52.351
544310 97.028 -59.385 52.320
544311 97.065 -59.388 52.284
544312 97.102 -59.391 52.248
544313 97.139 -59.393 52.213
544314 97.176 -59.396 52.175
544315 97.209 -59.399 52.134
544316 97.242 -59.403 52.092
544317 97.275 -59.406 52.045
544318 97.307 -59.410 51.997
544319 97.339 -59.413 51.947
544320 97.371 -59.416 51.897
544321 97.403 -59.421 51.846
544322 97.435 -59.425 51.794
544323 97.466 -59.431 51.742
544324 97.498 -59.435 51.689
544325 97.531 -59.440 51.635
544326 97.571 -59.451 51.584
544327 97.610 -59.463 51.536
544328 97.650 -59.473 51.493
544329 97.690 -59.485 51.448
544330 97.730 -59.496 51.403
544331 97.770 -59.507 51.353
544332 97.809 -59.518 51.299
544333 97.849 -59.529 51.243
544334 97.889 -59.540 51.187
544335 97.930 -59.546 51.138
544336 97.972 -59.549 51.092
544337 98.014 -59.553 51.046
544338 98.056 -59.556 51.001
544339 98.098 -59.560 50.951
544340 98.140 -59.563 50.901
544341 98.179 -59.563 50.854
544342 98.209 -59.550 50.808
544343 98.238 -59.536 50.765
544344 98.267 -59.522 50.725
544345 98.296 -59.510 50.686
544346 98.317 -59.516 50.649
544347 98.353 -59.523 50.612
544348 98.397 -59.529 50.575
544349 98.435 -59.526 50.539
544350 98.472 -59.523 50.503
544351 98.509 -59.517 50.469
544352 98.546 -59.501 50.435
544353 98.583 -59.485 50.400
544354 98.620 -59.469 50.365
544355 98.663 -59.463 50.332
544356 98.705 -59.457 50.296
544357 98.748 -59.451 50.260
544358 98.790 -59.445 50.221
544359 98.833 -59.440 50.182
544360 98.875 -59.434 50.145
544361 98.913 -59.428 50.104
544362 98.949 -59.433 50.059
544363 98.978 -59.437 50.013
544364 99.007 -59.441 49.967
544365 99.035 -59.444 49.923
544366 99.064 -59.448 49.882
544367 99.093 -59.452 49.842
544368 99.122 -59.457 49.803
544369 99.150 -59.460 49.765
544370 99.179 -59.464 49.727
544371 99.208 -59.468 49.688
544372 99.236 -59.472 49.653
544373 99.293 -59.469 49.618
544374 99.365 -59.463 49.583
544723 86.292 -56.099 57.645
544724 86.332 -56.112 57.635
544725 86.372 -56.125 57.623
544726 86.413 -56.139 57.611
544727 86.453 -56.152 57.598
544728 86.494 -56.166 57.582
544729 86.534 -56.179 57.566
544730 86.574 -56.192 57.547
544731 86.615 -56.206 57.525
544732 86.655 -56.219 57.502
544733 86.696 -56.232 57.478
544734 86.736 -56.246 57.451
544735 86.777 -56.259 57.422
544736 86.817 -56.273 57.392
544737 86.857 -56.285 57.358
544738 86.898 -56.299 57.324
544739 86.938 -56.313 57.289
544740 86.979 -56.326 57.254
544741 87.019 -56.339 57.219
544742 87.059 -56.352 57.185
544743 87.099 -56.366 57.149
544744 87.140 -56.379 57.114
544808 89.619 -57.238 55.882
544809 89.660 -57.249 55.867
544810 89.701 -57.260 55.854
544811 89.743 -57.271 55.842
544812 89.785 -57.282 55.830
544813 89.826 -57.293 55.818
544814 89.867 -57.304 55.806
544815 89.908 -57.314 55.793
544816 89.950 -57.325 55.781
544817 89.996 -57.340 55.770
544818 90.042 -57.355 55.757
544819 90.079 -57.365 55.743
544820 90.115 -57.375 55.728
544821 90.150 -57.390 55.712
544822 90.185 -57.406 55.694
544823 90.220 -57.422 55.676
544824 90.254 -57.438 55.656
544825 90.289 -57.454 55.636
544826 90.323 -57.470 55.613
544827 90.357 -57.482 55.590
544828 90.391 -57.493 55.566
544829 90.424 -57.504 55.542
544830 90.458 -57.516 55.518
544831 90.493 -57.529 55.492
544832 90.531 -57.546 55.467
544833 90.568 -57.562 55.444
544834 90.605 -57.579 55.420
544835 90.643 -57.595 55.400
544836 90.680 -57.611 55.381
544837 90.718 -57.628 55.362
544838 90.755 -57.644 55.345
544839 90.793 -57.661 55.330
544840 90.835 -57.670 55.314
544841 90.880 -57.677 55.302
544842 90.925 -57.683 55.294
544843 90.970 -57.690 55.287
544844 91.015 -57.697 55.281
544845 91.060 -57.703 55.276
544846 91.104 -57.712 55.276
544847 91.144 -57.736 55.277
544848 91.183 -57.760 55.280
544849 91.223 -57.783 55.283
544850 91.262 -57.807 55.285
544851 91.302 -57.831 55.287
544852 91.342 -57.855 55.287
544853 91.381 -57.878 55.281
544854 91.421 -57.902 55.273
544855 91.460 -57.926 55.265
544856 91.503 -57.940 55.257
544857 91.546 -57.950 55.249
544858 91.590 -57.961 55.234
544859 91.634 -57.971 55.213
544860 91.678 -57.981 55.192
544861 91.722 -57.992 55.174
544862 91.766 -58.003 55.160
544863 91.809 -58.013 55.140
544864 91.853 -58.024 55.130
544865 91.897 -58.034 55.122
544866 91.932 -58.052 55.112
544867 91.963 -58.075 55.095
544868 91.994 -58.096 55.072
544869 92.024 -58.118 55.047
544870 92.055 -58.140 55.022
544871 92.085 -58.163 54.996
544872 92.121 -58.181 54.971
544873 92.162 -58.196 54.941
544874 92.204 -58.210 54.908
544875 92.245 -58.225 54.879
544876 92.287 -58.238 54.855
544877 92.329 -58.248 54.838
544878 92.371 -58.258 54.821
544879 92.414 -58.268 54.807
544880 92.455 -58.277 54.794
544881 92.493 -58.280 54.782
544882 92.522 -58.285 54.769
544883 92.552 -58.291 54.757
544884 92.582 -58.296 54.744
544885 92.611 -58.301 54.728
544886 92.641 -58.306 54.713
544887 92.670 -58.312 54.700
544888 92.700 -58.317 54.690
544889 92.729 -58.322 54.684
544890 92.759 -58.328 54.685
544891 92.789 -58.332 54.689
544892 92.818 -58.338 54.699
544893 92.848 -58.343 54.715
544894 92.887 -58.349 54.734
544895 92.933 -58.355 54.753
544896 92.976 -58.362 54.768
544897 93.013 -58.372 54.779
544898 93.052 -58.382 54.784
544899 93.089 -58.391 54.782
544900 93.128 -58.400 54.778
544901 93.165 -58.410 54.779
544902 93.203 -58.419 54.784
544903 93.241 -58.429 54.789
544904 93.279 -58.439 54.791
544905 93.318 -58.448 54.785
544906 93.355 -58.458 54.773
544907 93.393 -58.468 54.757
544908 93.431 -58.477 54.740
544909 93.470 -58.486 54.724
544910 93.508 -58.494 54.706
544911 93.547 -58.501 54.686
544912 93.583 -58.515 54.664
544913 93.620 -58.529 54.653
544914 93.656 -58.542 54.644
544915 93.692 -58.556 54.635
544916 93.728 -58.570 54.621
544917 93.765 -58.583 54.603
544918 93.801 -58.597 54.580
544919 93.838 -58.610 54.562
544920 93.874 -58.624 54.551
544921 93.910 -58.638 54.548
544922 93.946 -58.651 54.548
544923 93.983 -58.666 54.548
544924 94.019 -58.679 54.542
544925 94.055 -58.693 54.533
544926 94.091 -58.707 54.520
544927 94.127 -58.719 54.500
544928 94.161 -58.730 54.472
544929 94.195 -58.742 54.435
544930 94.229 -58.752 54.394
544931 94.264 -58.764 54.355
544932 94.298 -58.775 54.319
544933 94.333 -58.786 54.287
544934 94.367 -58.798 54.260
544935 94.401 -58.808 54.246
544936 94.435 -58.820 54.232
544937 94.470 -58.831 54.214
544938 94.500 -58.845 54.191
544939 94.531 -58.858 54.165
544940 94.572 -58.869 54.135
544941 94.614 -58.880 54.105
544942 94.657 -58.890 54.078
544943 94.699 -58.901 54.053
544944 94.742 -58.912 54.026
544945 94.784 -58.923 53.997
544946 94.827 -58.933 53.962
544947 94.870 -58.944 53.923
544948 94.912 -58.955 53.883
544949 94.955 -58.965 53.844
544950 94.997 -58.977 53.805
544951 95.040 -58.987 53.771
544952 95.083 -58.998 53.738
544953 95.125 -59.008 53.709
544954 95.168 -59.019 53.685
544955 95.210 -59.030 53.663
544956 95.253 -59.041 53.640
544957 95.289 -59.055 53.618
544958 95.321 -59.069 53.595
544959 95.353 -59.084 53.573
544960 95.385 -59.099 53.554
544961 95.417 -59.114 53.535
544962 95.449 -59.129 53.518
544963 95.481 -59.143 53.502
544964 95.514 -59.154 53.488
544965 95.543 -59.162 53.473
544966 95.571 -59.170 53.456
544967 95.600 -59.178 53.440
544968 95.628 -59.185 53.419
544969 95.657 -59.193 53.400
544970 95.686 -59.200 53.379
544971 95.714 -59.208 53.355
544972 95.743 -59.215 53.329
544973 95.771 -59.223 53.306
544974 95.800 -59.231 53.287
544975 95.831 -59.237 53.270
544976 95.866 -59.244 53.252
544977 95.901 -59.250 53.235
544978 95.937 -59.256 53.220
544979 95.972 -59.261 53.211
544980 96.007 -59.267 53.200
544981 96.042 -59.273 53.189
544982 96.078 -59.279 53.177
544983 96.111 -59.285 53.160
544984 96.145 -59.292 53.142
544985 96.178 -59.299 53.120
544986 96.212 -59.306 53.097
544987 96.245 -59.312 53.074
544988 96.278 -59.319 53.049
544989 96.311 -59.326 53.024
544990 96.344 -59.333 52.999
544991 96.377 -59.340 52.975
544992 96.410 -59.347 52.956
544993 96.444 -59.353 52.937
544994 96.477 -59.359 52.913
544995 96.509 -59.364 52.889
544996 96.539 -59.361 52.864
544997 96.570 -59.359 52.835
544998 96.600 -59.356 52.806
544999 96.630 -59.354 52.776
545000 96.661 -59.351 52.740
545001 96.691 -59.349 52.702
545002 96.722 -59.347 52.661
545003 96.752 -59.344 52.616
545004 96.782 -59.341 52.570
545005 96.815 -59.341 52.526
545006 96.849 -59.342 52.485
545007 96.882 -59.343 52.447
545008 96.920 -59.346 52.414
545009 96.957 -59.349 52.384
545010 96.994 -59.352 52.354
545011 97.031 -59.355 52.321
545012 97.068 -59.357 52.284
545013 97.105 -59.360 52.246
545014 97.142 -59.363 52.207
545015 97.178 -59.366 52.166
545016 97.211 -59.369 52.122
545017 97.244 -59.373 52.076
545018 97.277 -59.376 52.027
545019 97.310 -59.379 51.976
545020 97.343 -59.383 51.924
545021 97.375 -59.386 51.871
545022 97.406 -59.391 51.820
545023 97.438 -59.396 51.769
545024 97.470 -59.400 51.717
545025 97.502 -59.405 51.665
545026 97.533 -59.410 51.613
545027 97.567 -59.416 51.562
545028 97.607 -59.427 51.511
545029 97.647 -59.438 51.466
545030 97.686 -59.449 51.420
545031 97.726 -59.460 51.375
545032 97.766 -59.471 51.330
545033 97.806 -59.482 51.281
545034 97.846 -59.494 51.229
545035 97.886 -59.504 51.179
545036 97.928 -59.508 51.136
545037 97.969 -59.512 51.095
545038 98.011 -59.515 51.053
545039 98.054 -59.518 51.008
545040 98.095 -59.522 50.960
545041 98.137 -59.525 50.911
545042 98.179 -59.528 50.864
545043 98.208 -59.514 50.818
545044 98.237 -59.501 50.775
545045 98.266 -59.488 50.733
545046 98.295 -59.474 50.695
545047 98.320 -59.472 50.656
545048 98.358 -59.479 50.617
545049 98.397 -59.479 50.577
545050 98.434 -59.476 50.540
545051 98.472 -59.474 50.503
545052 98.508 -59.472 50.467
545053 98.546 -59.456 50.432
545054 98.582 -59.439 50.398
545055 98.621 -59.426 50.365
545056 98.664 -59.420 50.333
545057 98.706 -59.415 50.298
545058 98.749 -59.409 50.263
545059 98.791 -59.403 50.225
545060 98.833 -59.397 50.187
545061 98.869 -59.392 50.151
545062 98.905 -59.388 50.111
545063 98.947 -59.396 50.067
545064 98.976 -59.400 50.021
545065 99.005 -59.403 49.976
545066 99.033 -59.407 49.931
545067 99.062 -59.411 49.890
545068 99.091 -59.415 49.851
545069 99.120 -59.419 49.812
545070 99.149 -59.423 49.774
545071 99.178 -59.427 49.736
545072 99.206 -59.431 49.699
545073 99.235 -59.435 49.664
545074 99.301 -59.430 49.630
545075 99.373 -59.424 49.595
545424 86.308 -56.056 57.650
545425 86.349 -56.069 57.641
545426 86.389 -56.082 57.629
545427 86.430 -56.096 57.617
545428 86.470 -56.109 57.604
545429 86.510 -56.122 57.590
545430 86.551 -56.135 57.575
545431 86.591 -56.149 57.557
545432 86.632 -56.163 57.536
545433 86.672 -56.175 57.514
545434 86.713 -56.189 57.490
545435 86.753 -56.202 57.464
545436 86.793 -56.216 57.436
545437 86.834 -56.229 57.407
545438 86.874 -56.242 57.373
545439 86.915 -56.256 57.338
545440 86.955 -56.269 57.303
545441 86.995 -56.282 57.268
545442 87.036 -56.296 57.233
545443 87.076 -56.309 57.199
545444 87.117 -56.323 57.162
545445 87.157 -56.335 57.126
545509 89.638 -57.199 55.883
545510 89.680 -57.210 55.868
545511 89.721 -57.220 55.855
545512 89.762 -57.232 55.843
545513 89.804 -57.242 55.830
545514 89.845 -57.253 55.817
545515 89.887 -57.264 55.805
545516 89.928 -57.275 55.791
545517 89.969 -57.285 55.778
545518 90.011 -57.296 55.765
545519 90.058 -57.312 55.752
545520 90.094 -57.322 55.737
545521 90.129 -57.336 55.721
545522 90.164 -57.352 55.703
545523 90.199 -57.368 55.684
545524 90.234 -57.385 55.665
545525 90.268 -57.401 55.644
545526 90.303 -57.416 55.622
545527 90.337 -57.432 55.599
545528 90.371 -57.443 55.575
545529 90.404 -57.454 55.550
545530 90.438 -57.466 55.526
545531 90.473 -57.479 55.502
545532 90.510 -57.495 55.478
545533 90.547 -57.511 55.454
545534 90.585 -57.527 55.431
545535 90.622 -57.544 55.408
545536 90.660 -57.561 55.388
545537 90.697 -57.576 55.369
545538 90.735 -57.593 55.351
545539 90.772 -57.610 55.334
545540 90.810 -57.626 55.319
545541 90.849 -57.639 55.305
545542 90.894 -57.645 55.292
545543 90.939 -57.652 55.283
545544 90.984 -57.659 55.276
545545 91.029 -57.665 55.271
545546 91.074 -57.672 55.265
545547 91.118 -57.683 55.265
545548 91.157 -57.706 55.266
545549 91.197 -57.730 55.268
545550 91.236 -57.754 55.271
545551 91.276 -57.777 55.274
545552 91.316 -57.801 55.280
545553 91.355 -57.824 55.286
545554 91.395 -57.849 55.289
545555 91.435 -57.872 55.286
545556 91.476 -57.890 55.281
545557 91.520 -57.900 55.274
545558 91.564 -57.911 55.265
545559 91.607 -57.921 55.249
545560 91.651 -57.932 55.229
545561 91.695 -57.943 55.210
545562 91.739 -57.952 55.198
545563 91.783 -57.963 55.186
545564 91.826 -57.974 55.164
545565 91.870 -57.984 55.139
545566 91.914 -57.995 55.116
545567 91.952 -58.010 55.099
545568 91.983 -58.032 55.087
545569 92.014 -58.054 55.067
545570 92.044 -58.077 55.041
545571 92.075 -58.099 55.013
545572 92.105 -58.121 54.988
545573 92.145 -58.137 54.968
545574 92.186 -58.151 54.944
545575 92.227 -58.165 54.912
545576 92.269 -58.180 54.881
545577 92.310 -58.194 54.855
545578 92.351 -58.209 54.836
545579 92.392 -58.223 54.818
545580 92.434 -58.238 54.803
545581 92.476 -58.248 54.789
545582 92.508 -58.253 54.777
545583 92.537 -58.258 54.765
545584 92.567 -58.263 54.754
545585 92.597 -58.269 54.744
545586 92.626 -58.274 54.732
545587 92.656 -58.279 54.721
545588 92.686 -58.285 54.712
545589 92.715 -58.290 54.705
545590 92.744 -58.294 54.697
545591 92.774 -58.300 54.690
545592 92.804 -58.305 54.687
545593 92.834 -58.310 54.689
545594 92.863 -58.316 54.695
545595 92.898 -58.321 54.706
545596 92.944 -58.327 54.716
545597 92.984 -58.336 54.724
545598 93.021 -58.345 54.730
545599 93.060 -58.355 54.733
545600 93.097 -58.365 54.734
545601 93.136 -58.374 54.734
545602 93.174 -58.384 54.736
545603 93.211 -58.394 54.744
545604 93.250 -58.403 54.754
545605 93.287 -58.413 54.760
545606 93.326 -58.422 54.759
545607 93.363 -58.432 54.754
545608 93.401 -58.441 54.743
545609 93.439 -58.451 54.729
545610 93.478 -58.460 54.711
545611 93.516 -58.468 54.689
545612 93.555 -58.476 54.670
545613 93.591 -58.490 54.651
545614 93.627 -58.504 54.635
545615 93.663 -58.517 54.624
545616 93.700 -58.531 54.614
545617 93.736 -58.545 54.603
545618 93.772 -58.558 54.586
545619 93.808 -58.573 54.569
545620 93.845 -58.586 54.553
545621 93.881 -58.600 54.543
545622 93.918 -58.614 54.538
545623 93.954 -58.627 54.537
545624 93.990 -58.641 54.539
545625 94.026 -58.653 54.536
545626 94.060 -58.664 54.528
545627 94.094 -58.676 54.513
545628 94.128 -58.687 54.493
545629 94.163 -58.698 54.467
545630 94.197 -58.709 54.434
545631 94.231 -58.720 54.397
545632 94.266 -58.732 54.359
545633 94.300 -58.743 54.323
545634 94.334 -58.754 54.290
545635 94.368 -58.765 54.263
545636 94.403 -58.777 54.242
545637 94.437 -58.788 54.227
545638 94.470 -58.799 54.211
545639 94.501 -58.813 54.187
545640 94.532 -58.826 54.162
545641 94.568 -58.838 54.135
545642 94.611 -58.849 54.111
545643 94.653 -58.859 54.089
545644 94.696 -58.871 54.066
545645 94.739 -58.881 54.041
545646 94.781 -58.892 54.012
545647 94.824 -58.902 53.978
545648 94.866 -58.913 53.940
545649 94.909 -58.924 53.900
545650 94.952 -58.935 53.860
545651 94.994 -58.946 53.821
545652 95.037 -58.956 53.786
545653 95.079 -58.967 53.754
545654 95.122 -58.978 53.727
545655 95.164 -58.989 53.703
545656 95.207 -58.999 53.680
545657 95.251 -59.011 53.659
545658 95.295 -59.024 53.637
545659 95.327 -59.039 53.613
545660 95.360 -59.054 53.591
545661 95.393 -59.069 53.571
545662 95.425 -59.084 53.549
545663 95.458 -59.099 53.533
545664 95.491 -59.114 53.515
545665 95.520 -59.123 53.499
545666 95.549 -59.131 53.482
545667 95.577 -59.138 53.464
545668 95.606 -59.146 53.445
545669 95.634 -59.153 53.423
545670 95.663 -59.161 53.402
545671 95.691 -59.168 53.380
545672 95.720 -59.176 53.354
545673 95.748 -59.184 53.326
545674 95.776 -59.191 53.300
545675 95.805 -59.199 53.276
545676 95.837 -59.206 53.255
545677 95.872 -59.211 53.235
545678 95.907 -59.217 53.217
545679 95.943 -59.223 53.204
545680 95.978 -59.229 53.194
545681 96.013 -59.235 53.184
545682 96.048 -59.240 53.174
545683 96.083 -59.247 53.161
545684 96.114 -59.254 53.144
545685 96.149 -59.260 53.127
545686 96.183 -59.267 53.107
545687 96.216 -59.274 53.086
545688 96.249 -59.281 53.064
545689 96.283 -59.287 53.041
545690 96.316 -59.294 53.018
545691 96.349 -59.300 52.993
545692 96.382 -59.307 52.969
545693 96.415 -59.314 52.950
545694 96.449 -59.321 52.931
545695 96.482 -59.328 52.908
545696 96.513 -59.331 52.885
545697 96.544 -59.332 52.860
545698 96.574 -59.329 52.830
545699 96.604 -59.327 52.800
545700 96.635 -59.325 52.769
545701 96.665 -59.322 52.730
545702 96.695 -59.319 52.690
545703 96.726 -59.317 52.647
545704 96.756 -59.314 52.603
545705 96.787 -59.312 52.560
545706 96.820 -59.312 52.517
545707 96.854 -59.313 52.477
545708 96.887 -59.315 52.441
545709 96.923 -59.316 52.410
545710 96.960 -59.319 52.381
545711 96.997 -59.322 52.351
545712 97.034 -59.325 52.317
545713 97.071 -59.328 52.277
545714 97.108 -59.331 52.237
545715 97.145 -59.334 52.195
545716 97.180 -59.337 52.150
545717 97.213 -59.340 52.103
545718 97.246 -59.343 52.054
545719 97.279 -59.347 52.004
545720 97.312 -59.350 51.951
545721 97.346 -59.353 51.897
545722 97.378 -59.356 51.842
545723 97.410 -59.361 51.790
545724 97.442 -59.366 51.740
545725 97.473 -59.370 51.688
545726 97.505 -59.375 51.636
545727 97.537 -59.379 51.585
545728 97.569 -59.384 51.534
545729 97.604 -59.391 51.483
545730 97.644 -59.403 51.437
545731 97.683 -59.413 51.391
545732 97.723 -59.425 51.347
545733 97.763 -59.436 51.303
545734 97.803 -59.447 51.259
545735 97.843 -59.458 51.213
545736 97.883 -59.467 51.170
545737 97.925 -59.470 51.132
545738 97.967 -59.474 51.095
545739 98.009 -59.477 51.056
545740 98.051 -59.481 51.013
545741 98.093 -59.484 50.966
545742 98.135 -59.487 50.919
545743 98.176 -59.491 50.872
545744 98.208 -59.479 50.826
545745 98.236 -59.466 50.783
545746 98.266 -59.452 50.741
545747 98.295 -59.438 50.704
545748 98.322 -59.428 50.665
545749 98.359 -59.432 50.625
545750 98.396 -59.429 50.584
545751 98.433 -59.427 50.544
545752 98.471 -59.425 50.505
545753 98.508 -59.422 50.468
545754 98.545 -59.410 50.433
545755 98.582 -59.394 50.398
545756 98.622 -59.384 50.365
545757 98.664 -59.378 50.333
545758 98.707 -59.372 50.300
545759 98.749 -59.366 50.265
545760 98.789 -59.361 50.229
545761 98.825 -59.356 50.192
545762 98.861 -59.351 50.155
545763 98.897 -59.346 50.116
545764 98.934 -59.344 50.072
545765 98.975 -59.363 50.028
545766 99.003 -59.366 49.983
545767 99.032 -59.370 49.939
545768 99.061 -59.374 49.898
545769 99.089 -59.378 49.859
545770 99.119 -59.382 49.821
545771 99.147 -59.386 49.783
545772 99.176 -59.390 49.746
545773 99.205 -59.394 49.709
545774 99.237 -59.397 49.675
545775 99.309 -59.391 49.641
545776 99.381 -59.385 49.606
546125 86.326 -56.013 57.655
546126 86.366 -56.025 57.646
546127 86.406 -56.039 57.634
546128 86.446 -56.052 57.623
546129 86.487 -56.066 57.610
546130 86.527 -56.079 57.597
546131 86.568 -56.092 57.583
546132 86.608 -56.106 57.567
546133 86.649 -56.119 57.547
546134 86.689 -56.132 57.527
546135 86.729 -56.146 57.505
546136 86.770 -56.159 57.480
546137 86.810 -56.172 57.451
546138 86.851 -56.185 57.423
546139 86.891 -56.199 57.389
546140 86.931 -56.213 57.354
546141 86.972 -56.226 57.318
546142 87.012 -56.239 57.283
546143 87.053 -56.252 57.248
546144 87.093 -56.266 57.213
546145 87.133 -56.279 57.177
546146 87.174 -56.292 57.140
546147 87.214 -56.306 57.104
546210 89.658 -57.159 55.884
546211 89.699 -57.170 55.870
546212 89.741 -57.181 55.857
546213 89.782 -57.191 55.845
546214 89.823 -57.203 55.831
546215 89.865 -57.213 55.818
546216 89.906 -57.224 55.805
546217 89.948 -57.235 55.791
546218 89.989 -57.246 55.777
546219 90.030 -57.257 55.764
546220 90.072 -57.267 55.749
546221 90.109 -57.282 55.733
546222 90.144 -57.298 55.715
546223 90.178 -57.314 55.696
546224 90.213 -57.330 55.675
546225 90.248 -57.347 55.655
546226 90.283 -57.363 55.633
546227 90.317 -57.379 55.610
546228 90.351 -57.393 55.586
546229 90.385 -57.404 55.562
546230 90.418 -57.416 55.536
546231 90.452 -57.427 55.512
546232 90.489 -57.444 55.488
546233 90.527 -57.460 55.465
546234 90.565 -57.476 55.441
546235 90.602 -57.493 55.419
546236 90.640 -57.509 55.396
546237 90.677 -57.526 55.377
546238 90.714 -57.542 55.359
546239 90.752 -57.558 55.342
546240 90.789 -57.575 55.325
546241 90.826 -57.591 55.309
546242 90.864 -57.608 55.296
546243 90.908 -57.614 55.283
546244 90.954 -57.621 55.273
546245 90.999 -57.627 55.265
546246 91.043 -57.634 55.257
546247 91.088 -57.640 55.250
546248 91.132 -57.652 55.247
546249 91.171 -57.677 55.246
546250 91.211 -57.700 55.247
546251 91.250 -57.724 55.249
546252 91.290 -57.747 55.253
546253 91.329 -57.771 55.264
546254 91.369 -57.795 55.280
546255 91.409 -57.818 55.290
546256 91.449 -57.840 55.294
546257 91.493 -57.851 55.294
546258 91.536 -57.861 55.290
546259 91.580 -57.871 55.279
546260 91.624 -57.882 55.263
546261 91.668 -57.893 55.244
546262 91.712 -57.903 55.227
546263 91.755 -57.914 55.217
546264 91.799 -57.924 55.205
546265 91.843 -57.934 55.184
546266 91.887 -57.945 55.157
546267 91.931 -57.955 55.127
546268 91.973 -57.968 55.101
546269 92.003 -57.990 55.081
546270 92.034 -58.012 55.065
546271 92.065 -58.034 55.040
546272 92.095 -58.056 55.010
546273 92.127 -58.078 54.980
546274 92.168 -58.092 54.958
546275 92.209 -58.107 54.935
546276 92.250 -58.122 54.910
546277 92.292 -58.136 54.884
546278 92.333 -58.150 54.860
546279 92.374 -58.165 54.840
546280 92.416 -58.179 54.825
546281 92.457 -58.194 54.810
546282 92.485 -58.209 54.796
546283 92.511 -58.222 54.782
546284 92.544 -58.228 54.771
546285 92.577 -58.234 54.761
546286 92.611 -58.241 54.753
546287 92.642 -58.247 54.747
546288 92.671 -58.252 54.740
546289 92.701 -58.256 54.732
546290 92.730 -58.262 54.725
546291 92.760 -58.267 54.715
546292 92.789 -58.272 54.702
546293 92.819 -58.278 54.692
546294 92.849 -58.283 54.687
546295 92.878 -58.288 54.686
546296 92.909 -58.294 54.689
546297 92.954 -58.300 54.692
546298 92.992 -58.310 54.697
546299 93.029 -58.319 54.700
546300 93.068 -58.328 54.701
546301 93.106 -58.338 54.702
546302 93.144 -58.348 54.703
546303 93.182 -58.357 54.704
546304 93.220 -58.367 54.709
546305 93.258 -58.377 54.717
546306 93.295 -58.386 54.725
546307 93.334 -58.396 54.729
546308 93.371 -58.406 54.730
546309 93.410 -58.415 54.725
546310 93.447 -58.425 54.714
546311 93.485 -58.435 54.695
546312 93.524 -58.442 54.672
546313 93.562 -58.451 54.654
546314 93.599 -58.465 54.636
546315 93.635 -58.479 54.619
546316 93.671 -58.493 54.604
546317 93.707 -58.507 54.592
546318 93.744 -58.520 54.581
546319 93.780 -58.534 54.569
546320 93.816 -58.548 54.556
546321 93.852 -58.561 54.545
546322 93.889 -58.575 54.535
546323 93.924 -58.588 54.529
546324 93.959 -58.598 54.527
546325 93.993 -58.610 54.528
546326 94.027 -58.621 54.528
546327 94.061 -58.632 54.521
546328 94.096 -58.643 54.508
546329 94.130 -58.654 54.488
546330 94.164 -58.666 54.464
546331 94.199 -58.677 54.433
546332 94.233 -58.689 54.398
546333 94.267 -58.699 54.361
546334 94.301 -58.711 54.325
546335 94.336 -58.722 54.293
546336 94.370 -58.733 54.265
546337 94.404 -58.744 54.240
546338 94.438 -58.755 54.223
546339 94.471 -58.767 54.207
546340 94.502 -58.781 54.185
546341 94.533 -58.794 54.158
546342 94.565 -58.807 54.130
546343 94.607 -58.818 54.108
546344 94.650 -58.829 54.089
546345 94.692 -58.839 54.068
546346 94.735 -58.850 54.044
546347 94.777 -58.861 54.018
546348 94.820 -58.872 53.985
546349 94.863 -58.883 53.949
546350 94.905 -58.893 53.910
546351 94.948 -58.904 53.871
546352 94.990 -58.914 53.834
546353 95.033 -58.926 53.799
546354 95.075 -58.936 53.769
546355 95.118 -58.947 53.743
546356 95.161 -58.958 53.718
546357 95.204 -58.969 53.695
546358 95.251 -58.981 53.675
546359 95.298 -58.993 53.653
546360 95.333 -59.008 53.629
546361 95.366 -59.024 53.606
546362 95.399 -59.038 53.584
546363 95.431 -59.053 53.561
546364 95.464 -59.068 53.544
546365 95.496 -59.082 53.525
546366 95.526 -59.090 53.506
546367 95.554 -59.098 53.487
546368 95.583 -59.106 53.467
546369 95.611 -59.113 53.445
546370 95.640 -59.121 53.422
546371 95.668 -59.128 53.399
546372 95.697 -59.136 53.375
546373 95.725 -59.143 53.348
546374 95.754 -59.151 53.320
546375 95.782 -59.159 53.293
546376 95.811 -59.166 53.265
546377 95.843 -59.173 53.239
546378 95.878 -59.179 53.218
546379 95.913 -59.185 53.200
546380 95.949 -59.190 53.185
546381 95.984 -59.196 53.175
546382 96.020 -59.203 53.166
546383 96.055 -59.209 53.156
546384 96.086 -59.215 53.143
546385 96.118 -59.222 53.126
546386 96.153 -59.228 53.110
546387 96.188 -59.234 53.092
546388 96.221 -59.241 53.070
546389 96.254 -59.248 53.046
546390 96.287 -59.255 53.024
546391 96.320 -59.262 53.002
546392 96.354 -59.269 52.978
546393 96.387 -59.275 52.954
546394 96.420 -59.282 52.933
546395 96.453 -59.289 52.916
546396 96.486 -59.296 52.894
546397 96.517 -59.298 52.871
546398 96.548 -59.300 52.845
546399 96.579 -59.300 52.817
546400 96.609 -59.297 52.788
546401 96.639 -59.295 52.755
546402 96.670 -59.293 52.714
546403 96.700 -59.290 52.673
546404 96.730 -59.287 52.630
546405 96.761 -59.285 52.587
546406 96.792 -59.283 52.545
546407 96.825 -59.284 52.502
546408 96.859 -59.285 52.461
546409 96.892 -59.286 52.428
546410 96.926 -59.287 52.398
546411 96.963 -59.289 52.369
546412 97.000 -59.292 52.337
546413 97.037 -59.295 52.302
546414 97.074 -59.298 52.260
546415 97.111 -59.301 52.217
546416 97.149 -59.303 52.172
546417 97.182 -59.307 52.126
546418 97.215 -59.310 52.077
546419 97.248 -59.313 52.027
546420 97.281 -59.316 51.975
546421 97.314 -59.320 51.921
546422 97.348 -59.323 51.868
546423 97.382 -59.326 51.815
546424 97.413 -59.331 51.760
546425 97.445 -59.335 51.707
546426 97.476 -59.340 51.653
546427 97.508 -59.344 51.600
546428 97.540 -59.349 51.550
546429 97.572 -59.353 51.501
546430 97.603 -59.358 51.453
546431 97.640 -59.367 51.407
546432 97.680 -59.378 51.363
546433 97.720 -59.389 51.319
546434 97.760 -59.400 51.275
546435 97.800 -59.412 51.233
546436 97.840 -59.422 51.192
546437 97.880 -59.429 51.158
546438 97.922 -59.433 51.126
546439 97.964 -59.436 51.093
546440 98.006 -59.440 51.057
546441 98.048 -59.443 51.014
546442 98.090 -59.446 50.969
546443 98.132 -59.450 50.924
546444 98.174 -59.453 50.878
546445 98.207 -59.444 50.833
546446 98.236 -59.431 50.789
546447 98.265 -59.417 50.748
546448 98.294 -59.403 50.712
546449 98.323 -59.390 50.674
546450 98.359 -59.382 50.634
546451 98.395 -59.380 50.592
546452 98.433 -59.377 50.549
546453 98.470 -59.375 50.508
546454 98.507 -59.372 50.470
546455 98.545 -59.365 50.435
546456 98.581 -59.349 50.400
546457 98.623 -59.341 50.366
546458 98.665 -59.335 50.333
546459 98.708 -59.330 50.301
546460 98.744 -59.325 50.266
546461 98.780 -59.320 50.231
546462 98.816 -59.315 50.196
546463 98.852 -59.310 50.160
546464 98.888 -59.305 50.119
546465 98.924 -59.300 50.077
546466 98.973 -59.325 50.034
546467 99.002 -59.329 49.990
546468 99.030 -59.333 49.946
546469 99.059 -59.337 49.905
546470 99.088 -59.341 49.867
546471 99.117 -59.345 49.830
546472 99.146 -59.349 49.792
546473 99.174 -59.353 49.756
546474 99.203 -59.356 49.720
546475 99.244 -59.357 49.685
546476 99.316 -59.352 49.651
546477 99.389 -59.346 49.617
546826 86.342 -55.969 57.658
546827 86.382 -55.982 57.651
546828 86.423 -55.996 57.640
546829 86.463 -56.009 57.629
546830 86.504 -56.022 57.617
546831 86.544 -56.036 57.604
546832 86.585 -56.049 57.590
546833 86.625 -56.063 57.577
546834 86.665 -56.075 57.557
546835 86.706 -56.089 57.539
546836 86.746 -56.103 57.519
546837 86.787 -56.116 57.495
546838 86.827 -56.129 57.466
546839 86.867 -56.142 57.438
546840 86.908 -56.156 57.404
546841 86.948 -56.169 57.370
546842 86.989 -56.182 57.334
546843 87.029 -56.196 57.298
546844 87.069 -56.209 57.263
546845 87.110 -56.222 57.228
546846 87.150 -56.236 57.191
546847 87.191 -56.249 57.154
546848 87.231 -56.263 57.117
546849 87.271 -56.276 57.082
546911 89.677 -57.119 55.884
546912 89.719 -57.130 55.871
546913 89.760 -57.141 55.859
546914 89.802 -57.152 55.847
546915 89.843 -57.163 55.834
546916 89.884 -57.174 55.820
546917 89.925 -57.185 55.806
546918 89.967 -57.195 55.792
546919 90.009 -57.206 55.777
546920 90.050 -57.217 55.763
546921 90.086 -57.230 55.748
546922 90.123 -57.244 55.731
546923 90.158 -57.260 55.712
546924 90.192 -57.276 55.691
546925 90.227 -57.293 55.669
546926 90.262 -57.309 55.648
546927 90.296 -57.325 55.625
546928 90.331 -57.341 55.601
546929 90.365 -57.354 55.576
546930 90.398 -57.365 55.551
546931 90.432 -57.376 55.525
546932 90.469 -57.392 55.500
546933 90.507 -57.409 55.476
546934 90.544 -57.426 55.453
546935 90.582 -57.442 55.430
546936 90.619 -57.458 55.408
546937 90.656 -57.474 55.386
546938 90.694 -57.491 55.367
546939 90.731 -57.507 55.350
546940 90.768 -57.523 55.333
546941 90.806 -57.540 55.317
546942 90.844 -57.557 55.302
546943 90.881 -57.573 55.288
546944 90.923 -57.583 55.276
546945 90.968 -57.589 55.266
546946 91.013 -57.596 55.256
546947 91.058 -57.602 55.247
546948 91.103 -57.609 55.241
546949 91.146 -57.623 55.236
546950 91.185 -57.646 55.234
546951 91.225 -57.670 55.233
546952 91.264 -57.694 55.233
546953 91.304 -57.717 55.237
546954 91.344 -57.741 55.244
546955 91.383 -57.764 55.257
546956 91.423 -57.788 55.271
546957 91.466 -57.801 55.282
546958 91.510 -57.811 55.288
546959 91.554 -57.822 55.287
546960 91.597 -57.833 55.283
546961 91.641 -57.843 55.273
546962 91.685 -57.853 55.259
546963 91.729 -57.864 55.244
546964 91.773 -57.874 55.233
546965 91.816 -57.885 55.219
546966 91.860 -57.896 55.200
546967 91.904 -57.906 55.175
546968 91.948 -57.916 55.145
546969 91.992 -57.927 55.111
546970 92.024 -57.948 55.083
546971 92.054 -57.970 55.065
546972 92.085 -57.992 55.041
546973 92.116 -58.014 55.013
546974 92.150 -58.034 54.985
546975 92.192 -58.048 54.964
546976 92.233 -58.062 54.943
546977 92.274 -58.077 54.924
546978 92.315 -58.091 54.905
546979 92.357 -58.106 54.888
546980 92.398 -58.120 54.873
546981 92.439 -58.134 54.855
546982 92.463 -58.150 54.839
546983 92.486 -58.165 54.824
546984 92.509 -58.181 54.812
546985 92.536 -58.193 54.804
546986 92.569 -58.200 54.797
546987 92.602 -58.206 54.789
546988 92.636 -58.212 54.780
546989 92.669 -58.219 54.770
546990 92.702 -58.225 54.760
546991 92.735 -58.231 54.750
546992 92.769 -58.238 54.739
546993 92.802 -58.244 54.722
546994 92.834 -58.250 54.705
546995 92.864 -58.256 54.693
546996 92.893 -58.261 54.686
546997 92.923 -58.266 54.681
546998 92.962 -58.274 54.679
546999 93.000 -58.283 54.680
547000 93.038 -58.293 54.680
547001 93.076 -58.303 54.680
547002 93.114 -58.312 54.679
547003 93.152 -58.322 54.680
547004 93.190 -58.332 54.681
547005 93.228 -58.341 54.683
547006 93.266 -58.350 54.688
547007 93.303 -58.360 54.695
547008 93.342 -58.369 54.700
547009 93.379 -58.379 54.702
547010 93.418 -58.389 54.701
547011 93.455 -58.398 54.692
547012 93.493 -58.408 54.676
547013 93.532 -58.416 54.657
547014 93.569 -58.427 54.641
547015 93.606 -58.441 54.623
547016 93.642 -58.454 54.608
547017 93.679 -58.468 54.596
547018 93.715 -58.482 54.585
547019 93.751 -58.495 54.576
547020 93.787 -58.509 54.563
547021 93.823 -58.522 54.552
547022 93.857 -58.533 54.541
547023 93.892 -58.544 54.530
547024 93.926 -58.555 54.520
547025 93.960 -58.567 54.514
547026 93.994 -58.578 54.512
547027 94.029 -58.589 54.513
547028 94.063 -58.600 54.511
547029 94.097 -58.611 54.503
547030 94.131 -58.623 54.487
547031 94.166 -58.634 54.462
547032 94.200 -58.645 54.430
547033 94.235 -58.656 54.397
547034 94.269 -58.667 54.362
547035 94.303 -58.679 54.327
547036 94.337 -58.689 54.297
547037 94.371 -58.701 54.269
547038 94.406 -58.712 54.242
547039 94.440 -58.723 54.220
547040 94.472 -58.736 54.201
547041 94.502 -58.749 54.181
547042 94.533 -58.762 54.153
547043 94.564 -58.776 54.124
547044 94.604 -58.787 54.102
547045 94.646 -58.798 54.084
547046 94.689 -58.808 54.065
547047 94.732 -58.820 54.044
547048 94.774 -58.830 54.018
547049 94.817 -58.841 53.988
547050 94.859 -58.852 53.954
547051 94.902 -58.862 53.917
547052 94.945 -58.874 53.879
547053 94.987 -58.884 53.846
547054 95.030 -58.895 53.813
547055 95.072 -58.905 53.784
547056 95.115 -58.916 53.757
547057 95.158 -58.927 53.730
547058 95.204 -58.940 53.708
547059 95.251 -58.952 53.685
547060 95.298 -58.964 53.662
547061 95.339 -58.977 53.638
547062 95.372 -58.993 53.613
547063 95.404 -59.008 53.589
547064 95.437 -59.023 53.567
547065 95.470 -59.037 53.549
547066 95.500 -59.049 53.529
547067 95.530 -59.058 53.509
547068 95.560 -59.066 53.487
547069 95.588 -59.074 53.464
547070 95.617 -59.081 53.441
547071 95.645 -59.089 53.417
547072 95.674 -59.096 53.390
547073 95.702 -59.104 53.363
547074 95.731 -59.112 53.336
547075 95.759 -59.119 53.308
547076 95.788 -59.127 53.280
547077 95.816 -59.134 53.251
547078 95.849 -59.140 53.224
547079 95.884 -59.146 53.201
547080 95.919 -59.153 53.182
547081 95.955 -59.159 53.165
547082 95.990 -59.165 53.154
547083 96.026 -59.171 53.145
547084 96.059 -59.177 53.135
547085 96.089 -59.184 53.123
547086 96.121 -59.190 53.107
547087 96.157 -59.196 53.092
547088 96.191 -59.203 53.074
547089 96.226 -59.209 53.051
547090 96.259 -59.216 53.026
547091 96.292 -59.223 53.002
547092 96.325 -59.230 52.981
547093 96.359 -59.237 52.958
547094 96.392 -59.244 52.934
547095 96.425 -59.250 52.913
547096 96.458 -59.257 52.894
547097 96.490 -59.262 52.873
547098 96.521 -59.265 52.851
547099 96.552 -59.267 52.825
547100 96.583 -59.269 52.799
547101 96.613 -59.268 52.771
547102 96.644 -59.265 52.736
547103 96.674 -59.263 52.693
547104 96.704 -59.260 52.652
547105 96.735 -59.258 52.609
547106 96.765 -59.256 52.568
547107 96.797 -59.254 52.527
547108 96.830 -59.255 52.485
547109 96.864 -59.256 52.444
547110 96.898 -59.257 52.412
547111 96.932 -59.258 52.381
547112 96.966 -59.259 52.350
547113 97.003 -59.262 52.318
547114 97.040 -59.265 52.282
547115 97.077 -59.269 52.240
547116 97.114 -59.271 52.195
547117 97.151 -59.274 52.149
547118 97.184 -59.278 52.100
547119 97.217 -59.281 52.048
547120 97.250 -59.284 51.997
547121 97.283 -59.287 51.942
547122 97.316 -59.291 51.887
547123 97.351 -59.293 51.834
547124 97.385 -59.296 51.782
547125 97.418 -59.300 51.729
547126 97.449 -59.305 51.675
547127 97.480 -59.309 51.623
547128 97.512 -59.314 51.572
547129 97.543 -59.319 51.523
547130 97.575 -59.324 51.476
547131 97.607 -59.328 51.429
547132 97.639 -59.333 51.385
547133 97.677 -59.343 51.342
547134 97.717 -59.353 51.300
547135 97.756 -59.365 51.258
547136 97.796 -59.376 51.217
547137 97.836 -59.387 51.179
547138 97.878 -59.392 51.146
547139 97.920 -59.395 51.113
547140 97.961 -59.398 51.080
547141 98.004 -59.402 51.044
547142 98.045 -59.405 51.005
547143 98.087 -59.409 50.963
547144 98.130 -59.412 50.921
547145 98.171 -59.415 50.878
547146 98.206 -59.410 50.834
547147 98.235 -59.396 50.791
547148 98.265 -59.382 50.749
547149 98.294 -59.369 50.711
547150 98.324 -59.357 50.673
547151 98.359 -59.347 50.633
547152 98.395 -59.337 50.591
547153 98.432 -59.328 50.549
547154 98.469 -59.325 50.510
547155 98.507 -59.323 50.473
547156 98.544 -59.320 50.438
547157 98.581 -59.305 50.403
547158 98.624 -59.299 50.368
547159 98.664 -59.294 50.334
547160 98.700 -59.288 50.302
547161 98.736 -59.284 50.267
547162 98.772 -59.278 50.233
547163 98.808 -59.274 50.200
547164 98.844 -59.269 50.164
547165 98.880 -59.264 50.124
547166 98.916 -59.259 50.082
547167 98.968 -59.284 50.041
547168 99.000 -59.292 49.998
547169 99.029 -59.296 49.954
547170 99.058 -59.300 49.913
547171 99.086 -59.303 49.875
547172 99.115 -59.308 49.837
547173 99.144 -59.312 49.801
547174 99.172 -59.316 49.765
547175 99.202 -59.319 49.729
547176 99.253 -59.319 49.695
547177 99.324 -59.312 49.662
547178 99.396 -59.306 49.628
547179 99.468 -59.300 49.594
547527 86.359 -55.925 57.662
547528 86.400 -55.939 57.654
547529 86.440 -55.953 57.644
547530 86.480 -55.965 57.633
547531 86.521 -55.979 57.622
547532 86.561 -55.992 57.609
547533 86.601 -56.006 57.597
547534 86.642 -56.019 57.584
547535 86.682 -56.032 57.567
547536 86.723 -56.046 57.550
547537 86.763 -56.059 57.531
547538 86.803 -56.072 57.509
547539 86.844 -56.086 57.482
547540 86.884 -56.099 57.452
547541 86.925 -56.113 57.420
547542 86.965 -56.125 57.386
547543 87.005 -56.139 57.351
547544 87.046 -56.153 57.314
547545 87.086 -56.166 57.279
547546 87.127 -56.179 57.243
547547 87.167 -56.192 57.206
547548 87.207 -56.206 57.169
547549 87.248 -56.219 57.132
547550 87.288 -56.232 57.095
547551 87.329 -56.246 57.061
547612 89.697 -57.080 55.887
547613 89.738 -57.091 55.875
547614 89.779 -57.101 55.862
547615 89.821 -57.113 55.850
547616 89.862 -57.123 55.836
547617 89.904 -57.134 55.823
547618 89.945 -57.145 55.809
547619 89.986 -57.156 55.794
547620 90.028 -57.166 55.778
547621 90.067 -57.178 55.762
547622 90.097 -57.194 55.747
547623 90.134 -57.209 55.729
547624 90.171 -57.223 55.709
547625 90.206 -57.239 55.687
547626 90.241 -57.255 55.664
547627 90.276 -57.271 55.642
547628 90.310 -57.287 55.618
547629 90.345 -57.304 55.593
547630 90.379 -57.315 55.568
547631 90.412 -57.326 55.543
547632 90.449 -57.341 55.516
547633 90.486 -57.357 55.491
547634 90.524 -57.374 55.467
547635 90.561 -57.391 55.443
547636 90.598 -57.407 55.420
547637 90.636 -57.423 55.399
547638 90.673 -57.440 55.377
547639 90.711 -57.456 55.359
547640 90.748 -57.473 55.341
547641 90.786 -57.489 55.325
547642 90.823 -57.505 55.310
547643 90.860 -57.522 55.296
547644 90.898 -57.538 55.281
547645 90.937 -57.551 55.270
547646 90.982 -57.558 55.260
547647 91.027 -57.564 55.249
547648 91.072 -57.571 55.241
547649 91.117 -57.578 55.234
547650 91.160 -57.592 55.229
547651 91.199 -57.617 55.226
547652 91.239 -57.640 55.224
547653 91.278 -57.664 55.224
547654 91.318 -57.687 55.229
547655 91.358 -57.711 55.233
547656 91.397 -57.735 55.241
547657 91.439 -57.752 55.252
547658 91.483 -57.762 55.264
547659 91.526 -57.772 55.271
547660 91.570 -57.783 55.272
547661 91.614 -57.793 55.271
547662 91.658 -57.804 55.265
547663 91.702 -57.814 55.255
547664 91.745 -57.825 55.241
547665 91.789 -57.835 55.228
547666 91.833 -57.846 55.217
547667 91.877 -57.856 55.201
547668 91.921 -57.867 55.181
547669 91.964 -57.877 55.157
547670 92.008 -57.888 55.130
547671 92.044 -57.905 55.096
547672 92.075 -57.927 55.068
547673 92.105 -57.950 55.043
547674 92.136 -57.972 55.021
547675 92.174 -57.989 55.000
547676 92.215 -58.003 54.982
547677 92.257 -58.018 54.963
547678 92.298 -58.032 54.947
547679 92.339 -58.046 54.934
547680 92.380 -58.061 54.924
547681 92.418 -58.076 54.912
547682 92.441 -58.091 54.892
547683 92.464 -58.106 54.874
547684 92.488 -58.122 54.858
547685 92.511 -58.137 54.848
547686 92.533 -58.153 54.842
547687 92.561 -58.165 54.837
547688 92.594 -58.171 54.828
547689 92.627 -58.178 54.818
547690 92.660 -58.184 54.806
547691 92.694 -58.191 54.792
547692 92.727 -58.197 54.779
547693 92.761 -58.203 54.764
547694 92.794 -58.209 54.745
547695 92.828 -58.216 54.725
547696 92.864 -58.222 54.705
547697 92.901 -58.229 54.691
547698 92.937 -58.235 54.685
547699 92.974 -58.245 54.680
547700 93.010 -58.256 54.674
547701 93.047 -58.266 54.670
547702 93.084 -58.276 54.668
547703 93.122 -58.286 54.668
547704 93.160 -58.295 54.669
547705 93.198 -58.305 54.668
547706 93.236 -58.315 54.667
547707 93.274 -58.324 54.668
547708 93.311 -58.334 54.671
547709 93.350 -58.344 54.674
547710 93.388 -58.353 54.676
547711 93.426 -58.363 54.676
547712 93.464 -58.372 54.671
547713 93.502 -58.382 54.661
547714 93.540 -58.391 54.647
547715 93.577 -58.402 54.635
547716 93.613 -58.416 54.619
547717 93.650 -58.429 54.606
547718 93.686 -58.443 54.596
547719 93.721 -58.456 54.586
547720 93.756 -58.467 54.575
547721 93.790 -58.479 54.563
547722 93.825 -58.489 54.552
547723 93.859 -58.501 54.540
547724 93.893 -58.512 54.528
547725 93.927 -58.523 54.516
547726 93.962 -58.534 54.507
547727 93.996 -58.545 54.500
547728 94.030 -58.557 54.497
547729 94.064 -58.568 54.496
547730 94.099 -58.579 54.493
547731 94.133 -58.590 54.481
547732 94.168 -58.601 54.460
547733 94.202 -58.613 54.433
547734 94.236 -58.624 54.399
547735 94.270 -58.635 54.364
547736 94.305 -58.646 54.329
547737 94.339 -58.657 54.300
547738 94.373 -58.669 54.276
547739 94.408 -58.680 54.250
547740 94.442 -58.691 54.221
547741 94.473 -58.705 54.194
547742 94.503 -58.717 54.173
547743 94.534 -58.731 54.149
547744 94.565 -58.744 54.118
547745 94.600 -58.756 54.093
547746 94.643 -58.767 54.075
547747 94.685 -58.778 54.059
547748 94.728 -58.789 54.040
547749 94.771 -58.799 54.016
547750 94.813 -58.810 53.988
547751 94.856 -58.821 53.956
547752 94.898 -58.832 53.921
547753 94.941 -58.842 53.886
547754 94.983 -58.853 53.854
547755 95.026 -58.864 53.824
547756 95.068 -58.875 53.796
547757 95.111 -58.886 53.768
547758 95.157 -58.897 53.740
547759 95.204 -58.910 53.717
547760 95.251 -58.922 53.691
547761 95.298 -58.934 53.667
547762 95.344 -58.947 53.642
547763 95.377 -58.961 53.617
547764 95.410 -58.977 53.592
547765 95.443 -58.992 53.569
547766 95.475 -59.007 53.548
547767 95.505 -59.015 53.528
547768 95.535 -59.024 53.506
547769 95.565 -59.034 53.481
547770 95.594 -59.042 53.458
547771 95.622 -59.049 53.434
547772 95.651 -59.057 53.409
547773 95.679 -59.065 53.381
547774 95.708 -59.072 53.353
547775 95.736 -59.080 53.325
547776 95.765 -59.087 53.297
547777 95.794 -59.094 53.267
547778 95.822 -59.102 53.240
547779 95.855 -59.109 53.212
547780 95.890 -59.115 53.187
547781 95.926 -59.121 53.167
547782 95.961 -59.126 53.149
547783 95.996 -59.132 53.136
547784 96.032 -59.138 53.125
547785 96.062 -59.145 53.112
547786 96.093 -59.152 53.099
547787 96.125 -59.159 53.083
547788 96.160 -59.165 53.068
547789 96.195 -59.171 53.050
547790 96.230 -59.178 53.029
547791 96.263 -59.184 53.005
547792 96.297 -59.191 52.979
547793 96.330 -59.198 52.954
547794 96.363 -59.205 52.933
547795 96.396 -59.212 52.910
547796 96.430 -59.218 52.887
547797 96.463 -59.225 52.865
547798 96.495 -59.229 52.844
547799 96.525 -59.231 52.825
547800 96.556 -59.234 52.800
547801 96.587 -59.236 52.776
547802 96.617 -59.238 52.750
547803 96.648 -59.236 52.713
547804 96.678 -59.234 52.669
547805 96.709 -59.231 52.628
547806 96.739 -59.228 52.586
547807 96.769 -59.226 52.545
547808 96.802 -59.225 52.505
547809 96.835 -59.227 52.467
547810 96.869 -59.228 52.427
547811 96.903 -59.228 52.394
547812 96.937 -59.229 52.360
547813 96.970 -59.231 52.327
547814 97.006 -59.233 52.295
547815 97.043 -59.235 52.258
547816 97.081 -59.238 52.216
547817 97.117 -59.241 52.171
547818 97.153 -59.244 52.123
547819 97.186 -59.247 52.073
547820 97.219 -59.251 52.019
547821 97.252 -59.254 51.966
547822 97.286 -59.257 51.912
547823 97.320 -59.259 51.857
547824 97.355 -59.262 51.804
547825 97.389 -59.266 51.753
547826 97.422 -59.270 51.701
547827 97.454 -59.275 51.648
547828 97.485 -59.279 51.598
547829 97.516 -59.284 51.548
547830 97.546 -59.289 51.500
547831 97.578 -59.294 51.454
547832 97.610 -59.298 51.409
547833 97.642 -59.303 51.368
547834 97.674 -59.307 51.326
547835 97.714 -59.319 51.284
547836 97.753 -59.329 51.243
547837 97.793 -59.341 51.204
547838 97.833 -59.350 51.167
547839 97.875 -59.354 51.135
547840 97.917 -59.357 51.101
547841 97.959 -59.361 51.065
547842 98.001 -59.364 51.028
547843 98.043 -59.367 50.990
547844 98.085 -59.371 50.950
547845 98.127 -59.374 50.911
547846 98.169 -59.378 50.871
547847 98.206 -59.374 50.830
547848 98.235 -59.360 50.790
547849 98.264 -59.348 50.749
547850 98.294 -59.336 50.711
547851 98.324 -59.325 50.672
547852 98.359 -59.314 50.632
547853 98.396 -59.304 50.591
547854 98.433 -59.294 50.549
547855 98.469 -59.284 50.512
547856 98.506 -59.275 50.475
547857 98.543 -59.271 50.440
547858 98.582 -59.262 50.406
547859 98.620 -59.257 50.371
547860 98.656 -59.252 50.337
547861 98.692 -59.247 50.303
547862 98.728 -59.242 50.269
547863 98.764 -59.237 50.235
547864 98.799 -59.232 50.203
547865 98.836 -59.228 50.168
547866 98.871 -59.222 50.128
547867 98.907 -59.218 50.088
547868 98.953 -59.231 50.048
547869 98.999 -59.255 50.006
547870 99.027 -59.259 49.962
547871 99.056 -59.262 49.921
547872 99.085 -59.266 49.882
547873 99.114 -59.271 49.845
547874 99.142 -59.275 49.809
547875 99.171 -59.278 49.773
547876 99.200 -59.282 49.738
547877 99.260 -59.279 49.704
547878 99.332 -59.273 49.671
547879 99.404 -59.267 49.638
547880 99.471 -59.263 49.604
548228 86.376 -55.882 57.665
548229 86.416 -55.896 57.658
548230 86.456 -55.909 57.647
548231 86.497 -55.922 57.637
548232 86.538 -55.936 57.626
548233 86.578 -55.949 57.615
548234 86.618 -55.962 57.603
548235 86.659 -55.975 57.591
548236 86.699 -55.989 57.576
548237 86.739 -56.003 57.560
548238 86.780 -56.016 57.544
548239 86.820 -56.029 57.525
548240 86.861 -56.042 57.496
548241 86.901 -56.056 57.467
548242 86.941 -56.069 57.435
548243 86.982 -56.082 57.402
548244 87.022 -56.096 57.367
548245 87.063 -56.109 57.331
548246 87.103 -56.122 57.295
548247 87.143 -56.136 57.260
548248 87.184 -56.149 57.222
548249 87.224 -56.163 57.184
548250 87.265 -56.176 57.148
548251 87.305 -56.189 57.110
548252 87.345 -56.203 57.076
548253 87.386 -56.216 57.042
548313 89.716 -57.040 55.889
548314 89.757 -57.051 55.878
548315 89.799 -57.062 55.866
548316 89.840 -57.072 55.854
548317 89.882 -57.084 55.841
548318 89.923 -57.094 55.827
548319 89.965 -57.105 55.813
548320 90.006 -57.116 55.798
548321 90.047 -57.127 55.781
548322 90.074 -57.144 55.765
548323 90.108 -57.159 55.748
548324 90.146 -57.173 55.729
548325 90.183 -57.187 55.708
548326 90.220 -57.201 55.685
548327 90.255 -57.217 55.661
548328 90.290 -57.233 55.638
548329 90.324 -57.250 55.614
548330 90.359 -57.265 55.588
548331 90.393 -57.276 55.563
548332 90.428 -57.290 55.537
548333 90.466 -57.307 55.509
548334 90.503 -57.323 55.484
548335 90.540 -57.339 55.460
548336 90.578 -57.356 55.436
548337 90.615 -57.373 55.412
548338 90.653 -57.388 55.391
548339 90.690 -57.405 55.371
548340 90.728 -57.421 55.352
548341 90.765 -57.438 55.336
548342 90.802 -57.454 55.320
548343 90.840 -57.470 55.306
548344 90.877 -57.487 55.293
548345 90.915 -57.504 55.278
548346 90.952 -57.520 55.267
548347 90.997 -57.526 55.257
548348 91.042 -57.533 55.247
548349 91.087 -57.540 55.238
548350 91.132 -57.546 55.231
548351 91.173 -57.563 55.226
548352 91.213 -57.586 55.223
548353 91.253 -57.610 55.223
548354 91.292 -57.634 55.225
548355 91.332 -57.658 55.230
548356 91.371 -57.681 55.233
548357 91.412 -57.702 55.232
548358 91.456 -57.712 55.236
548359 91.500 -57.723 55.245
548360 91.543 -57.733 55.251
548361 91.587 -57.744 55.253
548362 91.631 -57.754 55.254
548363 91.675 -57.764 55.251
548364 91.719 -57.775 55.244
548365 91.763 -57.786 55.233
548366 91.806 -57.796 55.223
548367 91.850 -57.807 55.214
548368 91.894 -57.817 55.203
548369 91.938 -57.827 55.188
548370 91.982 -57.838 55.169
548371 92.025 -57.849 55.146
548372 92.064 -57.864 55.119
548373 92.095 -57.886 55.096
548374 92.126 -57.908 55.074
548375 92.156 -57.930 55.054
548376 92.197 -57.944 55.035
548377 92.238 -57.959 55.016
548378 92.280 -57.974 54.997
548379 92.321 -57.988 54.980
548380 92.362 -58.003 54.970
548381 92.397 -58.017 54.965
548382 92.420 -58.032 54.952
548383 92.442 -58.048 54.933
548384 92.465 -58.063 54.912
548385 92.489 -58.078 54.897
548386 92.512 -58.094 54.888
548387 92.535 -58.109 54.883
548388 92.558 -58.125 54.878
548389 92.592 -58.137 54.867
548390 92.629 -58.143 54.855
548391 92.665 -58.150 54.840
548392 92.702 -58.156 54.823
548393 92.738 -58.162 54.806
548394 92.775 -58.169 54.785
548395 92.811 -58.175 54.764
548396 92.848 -58.181 54.741
548397 92.884 -58.188 54.718
548398 92.921 -58.194 54.698
548399 92.957 -58.203 54.689
548400 92.994 -58.212 54.682
548401 93.030 -58.223 54.675
548402 93.067 -58.234 54.668
548403 93.103 -58.244 54.664
548404 93.140 -58.254 54.662
548405 93.177 -58.264 54.662
548406 93.213 -58.275 54.662
548407 93.250 -58.285 54.662
548408 93.286 -58.295 54.663
548409 93.323 -58.306 54.665
548410 93.359 -58.316 54.666
548411 93.396 -58.326 54.666
548412 93.434 -58.336 54.666
548413 93.472 -58.346 54.663
548414 93.510 -58.355 54.656
548415 93.548 -58.365 54.648
548416 93.584 -58.377 54.639
548417 93.620 -58.390 54.627
548418 93.654 -58.401 54.616
548419 93.689 -58.413 54.605
548420 93.723 -58.424 54.594
548421 93.757 -58.435 54.581
548422 93.792 -58.446 54.569
548423 93.826 -58.457 54.556
548424 93.860 -58.469 54.543
548425 93.895 -58.479 54.529
548426 93.929 -58.491 54.514
548427 93.963 -58.502 54.503
548428 93.997 -58.513 54.493
548429 94.032 -58.525 54.486
548430 94.066 -58.535 54.483
548431 94.101 -58.547 54.480
548432 94.135 -58.558 54.472
548433 94.169 -58.570 54.454
548434 94.203 -58.580 54.430
548435 94.238 -58.592 54.398
548436 94.272 -58.603 54.363
548437 94.306 -58.614 54.328
548438 94.340 -58.626 54.299
548439 94.375 -58.636 54.277
548440 94.409 -58.648 54.252
548441 94.443 -58.659 54.224
548442 94.474 -58.673 54.193
548443 94.504 -58.686 54.163
548444 94.535 -58.699 54.137
548445 94.566 -58.712 54.108
548446 94.597 -58.726 54.081
548447 94.639 -58.736 54.060
548448 94.682 -58.747 54.044
548449 94.724 -58.758 54.029
548450 94.767 -58.768 54.008
548451 94.810 -58.780 53.984
548452 94.852 -58.790 53.956
548453 94.895 -58.801 53.923
548454 94.937 -58.811 53.887
548455 94.980 -58.822 53.855
548456 95.023 -58.833 53.827
548457 95.065 -58.844 53.802
548458 95.110 -58.855 53.775
548459 95.157 -58.867 53.747
548460 95.203 -58.880 53.721
548461 95.250 -58.892 53.693
548462 95.297 -58.905 53.668
548463 95.344 -58.918 53.643
548464 95.383 -58.931 53.618
548465 95.415 -58.946 53.594
548466 95.448 -58.961 53.569
548467 95.480 -58.974 53.546
548468 95.510 -58.983 53.524
548469 95.540 -58.992 53.500
548470 95.569 -59.000 53.475
548471 95.599 -59.009 53.452
548472 95.628 -59.017 53.426
548473 95.657 -59.024 53.399
548474 95.685 -59.032 53.373
548475 95.713 -59.040 53.346
548476 95.742 -59.047 53.316
548477 95.770 -59.055 53.287
548478 95.799 -59.062 53.258
548479 95.827 -59.070 53.231
548480 95.862 -59.076 53.203
548481 95.897 -59.082 53.175
548482 95.932 -59.088 53.152
548483 95.967 -59.094 53.133
548484 96.003 -59.100 53.117
548485 96.035 -59.106 53.104
548486 96.065 -59.113 53.090
548487 96.096 -59.121 53.075
548488 96.129 -59.127 53.060
548489 96.164 -59.134 53.042
548490 96.198 -59.140 53.021
548491 96.234 -59.146 53.001
548492 96.268 -59.153 52.980
548493 96.302 -59.159 52.953
548494 96.335 -59.165 52.927
548495 96.368 -59.172 52.904
548496 96.401 -59.179 52.881
548497 96.434 -59.186 52.858
548498 96.467 -59.193 52.835
548499 96.499 -59.196 52.814
548500 96.529 -59.198 52.794
548501 96.560 -59.200 52.769
548502 96.591 -59.203 52.744
548503 96.621 -59.205 52.716
548504 96.653 -59.207 52.679
548505 96.683 -59.204 52.640
548506 96.713 -59.202 52.604
548507 96.743 -59.200 52.563
548508 96.774 -59.196 52.521
548509 96.807 -59.197 52.479
548510 96.840 -59.198 52.440
548511 96.874 -59.199 52.403
548512 96.908 -59.200 52.369
548513 96.942 -59.201 52.334
548514 96.975 -59.202 52.299
548515 97.009 -59.203 52.266
548516 97.046 -59.206 52.231
548517 97.084 -59.209 52.190
548518 97.121 -59.212 52.144
548519 97.155 -59.215 52.095
548520 97.188 -59.218 52.044
548521 97.221 -59.221 51.991
548522 97.254 -59.224 51.941
548523 97.290 -59.227 51.890
548524 97.324 -59.229 51.838
548525 97.359 -59.232 51.785
548526 97.392 -59.236 51.731
548527 97.426 -59.240 51.677
548528 97.459 -59.244 51.627
548529 97.490 -59.249 51.577
548530 97.521 -59.253 51.528
548531 97.551 -59.258 51.480
548532 97.583 -59.263 51.436
548533 97.613 -59.268 51.396
548534 97.645 -59.272 51.357
548535 97.677 -59.277 51.315
548536 97.710 -59.283 51.273
548537 97.750 -59.294 51.232
548538 97.790 -59.305 51.193
548539 97.830 -59.313 51.158
548540 97.872 -59.316 51.125
548541 97.915 -59.320 51.088
548542 97.956 -59.323 51.048
548543 97.998 -59.326 51.010
548544 98.040 -59.330 50.970
548545 98.082 -59.333 50.933
548546 98.124 -59.336 50.895
548547 98.166 -59.340 50.858
548548 98.205 -59.339 50.822
548549 98.235 -59.327 50.785
548550 98.265 -59.315 50.745
548551 98.295 -59.303 50.709
548552 98.325 -59.292 50.671
548553 98.360 -59.281 50.631
548554 98.396 -59.272 50.590
548555 98.433 -59.262 50.550
548556 98.469 -59.252 50.513
548557 98.506 -59.242 50.477
548558 98.543 -59.231 50.442
548559 98.577 -59.221 50.407
548560 98.613 -59.215 50.373
548561 98.648 -59.211 50.339
548562 98.684 -59.206 50.305
548563 98.719 -59.201 50.271
548564 98.756 -59.196 50.238
548565 98.791 -59.191 50.205
548566 98.827 -59.186 50.171
548567 98.863 -59.181 50.132
548568 98.899 -59.176 50.094
548569 98.939 -59.178 50.055
548570 98.997 -59.218 50.013
548571 99.026 -59.222 49.969
548572 99.055 -59.225 49.927
548573 99.083 -59.229 49.889
548574 99.112 -59.234 49.852
548575 99.141 -59.237 49.817
548576 99.169 -59.241 49.781
548577 99.198 -59.245 49.746
548578 99.268 -59.240 49.713
548579 99.340 -59.234 49.681
548580 99.412 -59.228 49.647
548581 99.472 -59.227 49.614
548929 86.392 -55.839 57.667
548930 86.433 -55.853 57.660
548931 86.474 -55.865 57.651
548932 86.514 -55.879 57.641
548933 86.554 -55.893 57.630
548934 86.595 -55.906 57.620
548935 86.635 -55.919 57.608
548936 86.675 -55.932 57.597
548937 86.716 -55.946 57.583
548938 86.756 -55.959 57.569
548939 86.797 -55.972 57.554
548940 86.837 -55.986 57.537
548941 86.877 -55.999 57.510
548942 86.918 -56.013 57.481
548943 86.958 -56.026 57.450
548944 86.999 -56.039 57.417
548945 87.039 -56.053 57.384
548946 87.079 -56.066 57.349
548947 87.120 -56.079 57.313
548948 87.160 -56.093 57.278
548949 87.201 -56.106 57.241
548950 87.241 -56.119 57.202
548951 87.281 -56.132 57.165
548952 87.322 -56.146 57.127
548953 87.362 -56.160 57.092
548954 87.403 -56.172 57.057
548955 87.443 -56.186 57.021
549014 89.736 -57.000 55.892
549015 89.777 -57.012 55.881
549016 89.818 -57.022 55.869
549017 89.860 -57.033 55.859
549018 89.901 -57.044 55.847
549019 89.943 -57.055 55.833
549020 89.984 -57.065 55.819
549021 90.025 -57.076 55.803
549022 90.056 -57.091 55.786
549023 90.083 -57.108 55.770
549024 90.120 -57.122 55.752
549025 90.157 -57.137 55.732
549026 90.194 -57.151 55.710
549027 90.231 -57.166 55.687
549028 90.268 -57.180 55.663
549029 90.304 -57.195 55.640
549030 90.338 -57.212 55.615
549031 90.373 -57.226 55.589
549032 90.408 -57.239 55.561
549033 90.445 -57.256 55.534
549034 90.482 -57.272 55.506
549035 90.520 -57.288 55.479
549036 90.557 -57.304 55.454
549037 90.595 -57.321 55.430
549038 90.633 -57.338 55.406
549039 90.670 -57.354 55.386
549040 90.707 -57.370 55.365
549041 90.744 -57.387 55.348
549042 90.782 -57.403 55.331
549043 90.819 -57.420 55.316
549044 90.857 -57.435 55.303
549045 90.894 -57.452 55.290
549046 90.932 -57.469 55.277
549047 90.969 -57.485 55.267
549048 91.011 -57.495 55.257
549049 91.056 -57.502 55.248
549050 91.101 -57.508 55.239
549051 91.146 -57.515 55.233
549052 91.187 -57.532 55.228
549053 91.227 -57.557 55.225
549054 91.266 -57.580 55.224
549055 91.306 -57.604 55.227
549056 91.346 -57.627 55.232
549057 91.385 -57.651 55.234
549058 91.429 -57.663 55.233
549059 91.472 -57.673 55.232
549060 91.516 -57.683 55.234
549061 91.560 -57.694 55.235
549062 91.604 -57.705 55.236
549063 91.648 -57.715 55.236
549064 91.692 -57.726 55.233
549065 91.735 -57.736 55.227
549066 91.779 -57.746 55.220
549067 91.823 -57.757 55.218
549068 91.867 -57.767 55.214
549069 91.911 -57.778 55.208
549070 91.954 -57.789 55.198
549071 91.998 -57.799 55.185
549072 92.042 -57.809 55.167
549073 92.085 -57.821 55.148
549074 92.115 -57.843 55.128
549075 92.146 -57.865 55.107
549076 92.179 -57.886 55.089
549077 92.221 -57.900 55.073
549078 92.262 -57.915 55.054
549079 92.303 -57.929 55.035
549080 92.345 -57.943 55.016
549081 92.374 -57.959 55.001
549082 92.398 -57.974 54.991
549083 92.421 -57.989 54.983
549084 92.444 -58.005 54.964
549085 92.467 -58.020 54.942
549086 92.490 -58.035 54.926
549087 92.516 -58.053 54.916
549088 92.546 -58.074 54.912
549089 92.576 -58.094 54.904
549090 92.612 -58.103 54.893
549091 92.648 -58.109 54.881
549092 92.685 -58.115 54.866
549093 92.721 -58.122 54.847
549094 92.758 -58.128 54.827
549095 92.795 -58.134 54.802
549096 92.831 -58.141 54.778
549097 92.868 -58.147 54.754
549098 92.904 -58.154 54.730
549099 92.941 -58.160 54.710
549100 92.977 -58.170 54.695
549101 93.014 -58.180 54.687
549102 93.050 -58.191 54.678
549103 93.087 -58.201 54.670
549104 93.124 -58.211 54.664
549105 93.160 -58.222 54.662
549106 93.197 -58.232 54.664
549107 93.233 -58.242 54.666
549108 93.270 -58.253 54.668
549109 93.306 -58.263 54.670
549110 93.343 -58.273 54.671
549111 93.379 -58.284 54.670
549112 93.416 -58.294 54.668
549113 93.452 -58.304 54.667
549114 93.488 -58.313 54.666
549115 93.522 -58.322 54.666
549116 93.556 -58.330 54.664
549117 93.591 -58.347 54.657
549118 93.622 -58.358 54.647
549119 93.656 -58.369 54.636
549120 93.690 -58.380 54.627
549121 93.724 -58.391 54.619
549122 93.759 -58.403 54.608
549123 93.793 -58.414 54.593
549124 93.828 -58.425 54.576
549125 93.862 -58.436 54.557
549126 93.896 -58.448 54.536
549127 93.930 -58.459 54.517
549128 93.965 -58.470 54.502
549129 93.999 -58.481 54.493
549130 94.034 -58.492 54.483
549131 94.068 -58.504 54.473
549132 94.102 -58.515 54.465
549133 94.136 -58.526 54.460
549134 94.171 -58.537 54.445
549135 94.205 -58.548 54.422
549136 94.239 -58.560 54.393
549137 94.273 -58.571 54.360
549138 94.308 -58.582 54.324
549139 94.342 -58.593 54.295
549140 94.377 -58.604 54.273
549141 94.411 -58.616 54.250
549142 94.443 -58.628 54.224
549143 94.474 -58.641 54.192
549144 94.505 -58.654 54.155
549145 94.536 -58.667 54.126
549146 94.566 -58.680 54.096
549147 94.597 -58.694 54.067
549148 94.636 -58.705 54.044
549149 94.679 -58.716 54.026
549150 94.721 -58.727 54.013
549151 94.764 -58.738 53.995
549152 94.806 -58.748 53.974
549153 94.849 -58.759 53.949
549154 94.891 -58.770 53.919
549155 94.934 -58.781 53.885
549156 94.976 -58.792 53.854
549157 95.019 -58.802 53.824
549158 95.062 -58.814 53.797
549159 95.109 -58.826 53.771
549160 95.156 -58.838 53.743
549161 95.203 -58.851 53.715
549162 95.250 -58.863 53.691
549163 95.297 -58.875 53.666
549164 95.344 -58.888 53.642
549165 95.389 -58.900 53.618
549166 95.421 -58.915 53.593
549167 95.454 -58.930 53.569
549168 95.484 -58.940 53.543
549169 95.514 -58.949 53.520
549170 95.544 -58.958 53.497
549171 95.574 -58.968 53.471
549172 95.604 -58.977 53.446
549173 95.633 -58.985 53.419
549174 95.662 -58.993 53.392
549175 95.690 -59.000 53.367
549176 95.719 -59.008 53.341
549177 95.747 -59.015 53.309
549178 95.776 -59.023 53.280
549179 95.805 -59.030 53.251
549180 95.833 -59.038 53.225
549181 95.868 -59.044 53.196
549182 95.903 -59.050 53.164
549183 95.938 -59.055 53.136
549184 95.973 -59.062 53.116
549185 96.008 -59.068 53.097
549186 96.038 -59.074 53.082
549187 96.068 -59.082 53.069
549188 96.099 -59.089 53.052
549189 96.132 -59.095 53.036
549190 96.167 -59.102 53.015
549191 96.202 -59.108 52.991
549192 96.237 -59.114 52.971
549193 96.272 -59.121 52.952
549194 96.306 -59.127 52.927
549195 96.339 -59.134 52.900
549196 96.373 -59.140 52.874
549197 96.406 -59.147 52.850
549198 96.439 -59.154 52.827
549199 96.472 -59.160 52.803
549200 96.503 -59.162 52.780
549201 96.533 -59.165 52.760
549202 96.564 -59.167 52.734
549203 96.595 -59.169 52.708
549204 96.626 -59.171 52.679
549205 96.657 -59.174 52.644
549206 96.687 -59.175 52.609
549207 96.718 -59.172 52.575
549208 96.748 -59.170 52.535
549209 96.779 -59.168 52.493
549210 96.812 -59.168 52.451
549211 96.846 -59.169 52.411
549212 96.879 -59.170 52.372
549213 96.913 -59.171 52.337
549214 96.947 -59.172 52.302
549215 96.980 -59.173 52.267
549216 97.014 -59.174 52.234
549217 97.049 -59.176 52.199
549218 97.087 -59.179 52.158
549219 97.123 -59.182 52.114
549220 97.157 -59.185 52.067
549221 97.190 -59.188 52.017
549222 97.224 -59.191 51.967
549223 97.258 -59.193 51.916
549224 97.293 -59.196 51.865
549225 97.328 -59.199 51.814
549226 97.363 -59.202 51.763
549227 97.396 -59.206 51.711
549228 97.429 -59.210 51.659
549229 97.462 -59.214 51.611
549230 97.495 -59.218 51.562
549231 97.526 -59.223 51.513
549232 97.557 -59.228 51.465
549233 97.588 -59.233 51.419
549234 97.618 -59.237 51.378
549235 97.649 -59.242 51.342
549236 97.680 -59.247 51.303
549237 97.712 -59.251 51.263
549238 97.747 -59.259 51.222
549239 97.787 -59.269 51.182
549240 97.828 -59.275 51.144
549241 97.870 -59.278 51.108
549242 97.912 -59.282 51.069
549243 97.954 -59.285 51.028
549244 97.996 -59.289 50.987
549245 98.037 -59.292 50.946
549246 98.080 -59.295 50.909
549247 98.122 -59.298 50.874
549248 98.164 -59.300 50.838
549249 98.205 -59.303 50.802
549250 98.235 -59.294 50.767
549251 98.266 -59.282 50.731
549252 98.296 -59.271 50.698
549253 98.326 -59.259 50.662
549254 98.360 -59.249 50.625
549255 98.397 -59.239 50.585
549256 98.433 -59.229 50.546
549257 98.470 -59.219 50.511
549258 98.507 -59.209 50.476
549259 98.541 -59.198 50.442
549260 98.575 -59.187 50.408
549261 98.608 -59.177 50.376
549262 98.643 -59.169 50.341
549263 98.678 -59.165 50.306
549264 98.714 -59.159 50.273
549265 98.749 -59.155 50.240
549266 98.785 -59.150 50.206
549267 98.820 -59.145 50.171
549268 98.856 -59.140 50.135
549269 98.891 -59.135 50.099
549270 98.927 -59.131 50.059
549271 98.987 -59.171 50.017
549272 99.024 -59.184 49.975
549273 99.053 -59.188 49.933
549274 99.082 -59.192 49.895
549275 99.111 -59.196 49.859
549276 99.139 -59.200 49.823
549277 99.168 -59.204 49.788
549278 99.204 -59.206 49.754
549279 99.275 -59.200 49.721
549280 99.347 -59.195 49.689
549281 99.419 -59.189 49.657
549282 99.473 -59.190 49.624
549630 86.410 -55.796 57.668
549631 86.450 -55.809 57.662
549632 86.490 -55.822 57.653
549633 86.531 -55.836 57.643
549634 86.571 -55.849 57.634
549635 86.612 -55.862 57.624
549636 86.652 -55.876 57.613
549637 86.692 -55.889 57.602
549638 86.733 -55.903 57.589
549639 86.773 -55.915 57.576
549640 86.813 -55.929 57.561
549641 86.854 -55.943 57.544
549642 86.895 -55.956 57.523
549643 86.935 -55.969 57.494
549644 86.975 -55.982 57.464
549645 87.015 -55.996 57.432
549646 87.056 -56.009 57.400
549647 87.096 -56.022 57.366
549648 87.137 -56.036 57.331
549649 87.177 -56.049 57.296
549650 87.217 -56.063 57.259
549651 87.258 -56.076 57.221
549652 87.298 -56.089 57.183
549653 87.339 -56.103 57.146
549654 87.379 -56.116 57.109
549655 87.419 -56.129 57.073
549656 87.460 -56.143 57.037
549657 87.500 -56.156 57.000
549715 89.755 -56.961 55.896
549716 89.797 -56.971 55.888
549717 89.838 -56.982 55.878
549718 89.879 -56.993 55.867
549719 89.920 -57.004 55.854
549720 89.962 -57.015 55.841
549721 90.004 -57.026 55.826
549722 90.038 -57.040 55.810
549723 90.063 -57.056 55.792
549724 90.094 -57.072 55.776
549725 90.131 -57.087 55.757
549726 90.168 -57.101 55.737
549727 90.205 -57.116 55.715
549728 90.243 -57.130 55.691
549729 90.280 -57.144 55.667
549730 90.317 -57.159 55.647
549731 90.353 -57.174 55.623
549732 90.387 -57.188 55.595
549733 90.425 -57.204 55.565
549734 90.462 -57.221 55.534
549735 90.500 -57.237 55.504
549736 90.537 -57.254 55.477
549737 90.575 -57.270 55.451
549738 90.612 -57.286 55.427
549739 90.649 -57.303 55.403
549740 90.686 -57.319 55.382
549741 90.724 -57.335 55.362
549742 90.761 -57.352 55.344
549743 90.799 -57.368 55.329
549744 90.836 -57.385 55.314
549745 90.874 -57.401 55.301
549746 90.911 -57.417 55.289
549747 90.949 -57.434 55.278
549748 90.986 -57.450 55.269
549749 91.025 -57.464 55.260
549750 91.071 -57.470 55.250
549751 91.115 -57.477 55.243
549752 91.160 -57.483 55.237
549753 91.202 -57.503 55.233
549754 91.241 -57.526 55.229
549755 91.281 -57.550 55.228
549756 91.320 -57.573 55.230
549757 91.360 -57.598 55.235
549758 91.402 -57.613 55.238
549759 91.446 -57.623 55.238
549760 91.490 -57.634 55.239
549761 91.533 -57.645 55.243
549762 91.577 -57.655 55.244
549763 91.621 -57.665 55.244
549764 91.665 -57.676 55.243
549765 91.709 -57.686 55.241
549766 91.752 -57.697 55.239
549767 91.796 -57.708 55.237
549768 91.840 -57.717 55.237
549769 91.884 -57.728 55.234
549770 91.928 -57.739 55.228
549771 91.971 -57.749 55.220
549772 92.015 -57.760 55.210
549773 92.059 -57.771 55.199
549774 92.103 -57.781 55.184
549775 92.136 -57.801 55.163
549776 92.166 -57.824 55.142
549777 92.203 -57.841 55.125
549778 92.244 -57.855 55.113
549779 92.286 -57.870 55.093
549780 92.327 -57.884 55.074
549781 92.353 -57.899 55.053
549782 92.376 -57.915 55.032
549783 92.399 -57.930 55.016
549784 92.422 -57.946 55.008
549785 92.445 -57.961 54.990
549786 92.474 -57.981 54.972
549787 92.504 -58.002 54.955
549788 92.533 -58.023 54.941
549789 92.563 -58.043 54.931
549790 92.596 -58.059 54.919
549791 92.632 -58.069 54.903
549792 92.668 -58.075 54.887
549793 92.705 -58.081 54.870
549794 92.741 -58.087 54.851
549795 92.778 -58.093 54.830
549796 92.814 -58.100 54.806
549797 92.851 -58.106 54.786
549798 92.887 -58.113 54.769
549799 92.924 -58.119 54.753
549800 92.960 -58.127 54.736
549801 92.997 -58.137 54.714
549802 93.034 -58.147 54.695
549803 93.070 -58.158 54.682
549804 93.107 -58.169 54.674
549805 93.143 -58.178 54.669
549806 93.180 -58.189 54.669
549807 93.216 -58.200 54.675
549808 93.253 -58.209 54.680
549809 93.289 -58.219 54.686
549810 93.324 -58.228 54.688
549811 93.358 -58.236 54.686
549812 93.393 -58.244 54.682
549813 93.427 -58.253 54.678
549814 93.462 -58.261 54.675
549815 93.496 -58.269 54.675
549816 93.531 -58.278 54.677
549817 93.565 -58.286 54.679
549818 93.603 -58.306 54.674
549819 93.639 -58.325 54.667
549820 93.666 -58.336 54.660
549821 93.693 -58.348 54.654
549822 93.726 -58.360 54.649
549823 93.761 -58.370 54.639
549824 93.795 -58.382 54.625
549825 93.829 -58.393 54.605
549826 93.863 -58.404 54.582
549827 93.898 -58.416 54.558
549828 93.932 -58.426 54.534
549829 93.966 -58.438 54.511
549830 94.001 -58.449 54.494
549831 94.035 -58.460 54.480
549832 94.069 -58.471 54.466
549833 94.104 -58.482 54.451
549834 94.138 -58.494 54.440
549835 94.172 -58.505 54.428
549836 94.206 -58.516 54.408
549837 94.241 -58.527 54.381
549838 94.275 -58.538 54.350
549839 94.310 -58.550 54.316
549840 94.344 -58.561 54.286
549841 94.378 -58.573 54.261
549842 94.412 -58.583 54.241
549843 94.445 -58.596 54.218
549844 94.475 -58.609 54.189
549845 94.506 -58.623 54.153
549846 94.537 -58.636 54.116
549847 94.567 -58.649 54.081
549848 94.598 -58.662 54.051
549849 94.632 -58.675 54.025
549850 94.675 -58.686 54.004
549851 94.717 -58.696 53.992
549852 94.760 -58.707 53.977
549853 94.803 -58.717 53.959
549854 94.845 -58.729 53.937
549855 94.888 -58.739 53.910
549856 94.930 -58.750 53.881
549857 94.973 -58.761 53.851
549858 95.016 -58.771 53.820
549859 95.062 -58.784 53.790
549860 95.109 -58.796 53.763
549861 95.156 -58.808 53.737
549862 95.203 -58.821 53.708
549863 95.250 -58.833 53.686
549864 95.297 -58.846 53.664
549865 95.343 -58.858 53.640
549866 95.390 -58.871 53.617
549867 95.427 -58.885 53.589
549868 95.459 -58.899 53.564
549869 95.489 -58.908 53.538
549870 95.519 -58.917 53.513
549871 95.549 -58.925 53.487
549872 95.579 -58.934 53.462
549873 95.608 -58.943 53.436
549874 95.638 -58.952 53.409
549875 95.668 -58.961 53.382
549876 95.696 -58.968 53.357
549877 95.725 -58.976 53.332
549878 95.753 -58.983 53.301
549879 95.781 -58.990 53.272
549880 95.810 -58.998 53.242
549881 95.838 -59.005 53.211
549882 95.874 -59.012 53.180
549883 95.909 -59.018 53.149
549884 95.944 -59.024 53.119
549885 95.979 -59.030 53.097
549886 96.011 -59.037 53.077
549887 96.041 -59.043 53.061
549888 96.072 -59.050 53.046
549889 96.102 -59.057 53.029
549890 96.136 -59.064 53.011
549891 96.171 -59.070 52.992
549892 96.206 -59.076 52.967
549893 96.241 -59.082 52.943
549894 96.276 -59.089 52.922
549895 96.311 -59.095 52.901
549896 96.344 -59.102 52.873
549897 96.377 -59.109 52.844
549898 96.410 -59.115 52.817
549899 96.444 -59.122 52.793
549900 96.476 -59.127 52.768
549901 96.507 -59.129 52.741
549902 96.537 -59.131 52.720
549903 96.568 -59.134 52.695
549904 96.599 -59.136 52.668
549905 96.630 -59.138 52.639
549906 96.661 -59.140 52.607
549907 96.691 -59.143 52.577
549908 96.722 -59.143 52.542
549909 96.752 -59.140 52.503
549910 96.784 -59.139 52.463
549911 96.817 -59.140 52.422
549912 96.851 -59.140 52.382
549913 96.885 -59.142 52.338
549914 96.918 -59.143 52.301
549915 96.952 -59.143 52.267
549916 96.985 -59.144 52.233
549917 97.019 -59.146 52.199
549918 97.053 -59.146 52.165
549919 97.090 -59.149 52.125
549920 97.125 -59.152 52.083
549921 97.159 -59.156 52.038
549922 97.193 -59.158 51.990
549923 97.228 -59.161 51.941
549924 97.262 -59.163 51.890
549925 97.297 -59.166 51.839
549926 97.332 -59.168 51.789
549927 97.366 -59.171 51.740
549928 97.399 -59.176 51.691
549929 97.433 -59.180 51.643
549930 97.466 -59.184 51.595
549931 97.499 -59.188 51.547
549932 97.531 -59.193 51.498
549933 97.562 -59.197 51.450
549934 97.593 -59.202 51.403
549935 97.624 -59.207 51.361
549936 97.654 -59.212 51.326
549937 97.685 -59.216 51.289
549938 97.716 -59.222 51.251
549939 97.747 -59.226 51.211
549940 97.783 -59.234 51.171
549941 97.825 -59.237 51.132
549942 97.867 -59.241 51.092
549943 97.910 -59.244 51.050
549944 97.952 -59.246 51.007
549945 97.994 -59.248 50.965
549946 98.036 -59.250 50.924
549947 98.079 -59.252 50.886
549948 98.121 -59.254 50.849
549949 98.162 -59.257 50.813
549950 98.202 -59.262 50.778
549951 98.236 -59.260 50.745
549952 98.266 -59.250 50.712
549953 98.296 -59.238 50.680
549954 98.326 -59.227 50.647
549955 98.361 -59.216 50.612
549956 98.397 -59.206 50.576
549957 98.434 -59.196 50.541
549958 98.471 -59.186 50.508
549959 98.505 -59.175 50.475
549960 98.538 -59.165 50.441
549961 98.572 -59.154 50.407
549962 98.606 -59.143 50.374
549963 98.640 -59.132 50.339
549964 98.673 -59.123 50.305
549965 98.709 -59.118 50.273
549966 98.744 -59.113 50.240
549967 98.780 -59.109 50.206
549968 98.815 -59.103 50.172
549969 98.851 -59.099 50.136
549970 98.886 -59.093 50.100
549971 98.924 -59.096 50.061
549972 98.973 -59.118 50.020
549973 99.023 -59.147 49.979
549974 99.052 -59.151 49.939
549975 99.080 -59.155 49.902
549976 99.109 -59.159 49.866
549977 99.138 -59.163 49.830
549978 99.166 -59.167 49.796
549979 99.211 -59.167 49.762
549980 99.283 -59.162 49.729
549981 99.355 -59.156 49.696
549982 99.427 -59.150 49.664
549983 99.474 -59.153 49.632
550332 86.467 -55.765 57.662
550333 86.507 -55.779 57.655
550334 86.548 -55.793 57.646
550335 86.588 -55.806 57.636
550336 86.628 -55.819 57.627
550337 86.669 -55.832 57.617
550338 86.709 -55.846 57.607
550339 86.750 -55.859 57.595
550340 86.790 -55.872 57.581
550341 86.831 -55.886 57.566
550342 86.871 -55.899 57.550
550343 86.911 -55.912 57.529
550344 86.951 -55.926 57.504
550345 86.992 -55.939 57.476
550346 87.033 -55.953 57.446
550347 87.073 -55.965 57.415
550348 87.113 -55.979 57.382
550349 87.153 -55.993 57.349
550350 87.194 -56.006 57.314
550351 87.234 -56.019 57.278
550352 87.275 -56.033 57.240
550353 87.315 -56.046 57.202
550354 87.355 -56.059 57.165
550355 87.396 -56.072 57.128
550356 87.436 -56.086 57.090
550357 87.477 -56.100 57.053
550358 87.517 -56.113 57.015
550359 87.564 -56.122 56.978
550416 89.774 -56.921 55.902
550417 89.816 -56.932 55.896
550418 89.858 -56.943 55.888
550419 89.899 -56.953 55.878
550420 89.940 -56.965 55.866
550421 89.981 -56.975 55.853
550422 90.019 -56.987 55.838
550423 90.045 -57.005 55.822
550424 90.071 -57.022 55.803
550425 90.105 -57.037 55.786
550426 90.143 -57.051 55.766
550427 90.180 -57.065 55.745
550428 90.217 -57.080 55.723
550429 90.254 -57.094 55.699
550430 90.291 -57.109 55.677
550431 90.328 -57.122 55.659
550432 90.366 -57.137 55.636
550433 90.404 -57.153 55.607
550434 90.442 -57.169 55.572
550435 90.479 -57.186 55.536
550436 90.516 -57.202 55.504
550437 90.554 -57.219 55.475
550438 90.591 -57.235 55.449
550439 90.628 -57.251 55.424
550440 90.666 -57.268 55.401
550441 90.704 -57.285 55.379
550442 90.741 -57.300 55.359
550443 90.779 -57.317 55.342
550444 90.816 -57.334 55.327
550445 90.853 -57.350 55.313
550446 90.891 -57.366 55.300
550447 90.928 -57.382 55.288
550448 90.965 -57.399 55.279
550449 91.003 -57.416 55.271
550450 91.040 -57.432 55.263
550451 91.085 -57.439 55.255
550452 91.130 -57.445 55.251
550453 91.175 -57.452 55.246
550454 91.215 -57.473 55.242
550455 91.255 -57.496 55.239
550456 91.295 -57.520 55.238
550457 91.334 -57.544 55.238
550458 91.375 -57.564 55.240
550459 91.419 -57.573 55.243
550460 91.463 -57.584 55.248
550461 91.507 -57.595 55.253
550462 91.551 -57.605 55.259
550463 91.594 -57.616 55.260
550464 91.638 -57.626 55.257
550465 91.682 -57.636 55.253
550466 91.726 -57.647 55.249
550467 91.770 -57.658 55.247
550468 91.813 -57.668 55.245
550469 91.857 -57.679 55.245
550470 91.901 -57.689 55.242
550471 91.945 -57.700 55.237
550472 91.989 -57.710 55.230
550473 92.032 -57.720 55.225
550474 92.076 -57.731 55.220
550475 92.120 -57.742 55.206
550476 92.156 -57.759 55.186
550477 92.187 -57.781 55.166
550478 92.226 -57.796 55.152
550479 92.268 -57.811 55.146
550480 92.308 -57.825 55.128
550481 92.331 -57.841 55.108
550482 92.354 -57.856 55.089
550483 92.377 -57.871 55.068
550484 92.403 -57.889 55.046
550485 92.432 -57.909 55.026
550486 92.461 -57.930 55.012
550487 92.491 -57.951 55.000
550488 92.521 -57.972 54.985
550489 92.551 -57.993 54.964
550490 92.580 -58.014 54.946
550491 92.616 -58.025 54.930
550492 92.652 -58.036 54.910
550493 92.689 -58.042 54.891
550494 92.725 -58.048 54.872
550495 92.762 -58.054 54.854
550496 92.798 -58.060 54.834
550497 92.835 -58.066 54.816
550498 92.871 -58.072 54.803
550499 92.908 -58.078 54.792
550500 92.944 -58.085 54.780
550501 92.981 -58.094 54.763
550502 93.017 -58.105 54.738
550503 93.054 -58.115 54.714
550504 93.090 -58.125 54.694
550505 93.125 -58.134 54.681
550506 93.159 -58.142 54.677
550507 93.194 -58.150 54.680
550508 93.228 -58.159 54.691
550509 93.263 -58.167 54.700
550510 93.297 -58.175 54.706
550511 93.332 -58.184 54.708
550512 93.366 -58.192 54.707
550513 93.401 -58.200 54.700
550514 93.435 -58.209 54.697
550515 93.470 -58.217 54.692
550516 93.504 -58.225 54.686
550517 93.539 -58.234 54.685
550518 93.574 -58.244 54.685
550519 93.612 -58.264 54.686
550520 93.650 -58.284 54.687
550521 93.688 -58.303 54.687
550522 93.714 -58.314 54.685
550523 93.741 -58.326 54.682
550524 93.768 -58.338 54.673
550525 93.796 -58.350 54.660
550526 93.831 -58.361 54.638
550527 93.865 -58.372 54.613
550528 93.899 -58.383 54.587
550529 93.933 -58.394 54.559
550530 93.968 -58.406 54.529
550531 94.002 -58.417 54.502
550532 94.037 -58.428 54.481
550533 94.071 -58.439 54.461
550534 94.105 -58.451 54.440
550535 94.139 -58.462 54.425
550536 94.174 -58.473 54.411
550537 94.208 -58.484 54.391
550538 94.242 -58.495 54.365
550539 94.277 -58.507 54.334
550540 94.311 -58.518 54.302
550541 94.345 -58.529 54.273
550542 94.380 -58.540 54.249
550543 94.414 -58.551 54.228
550544 94.445 -58.564 54.206
550545 94.476 -58.577 54.180
550546 94.506 -58.591 54.147
550547 94.537 -58.604 54.109
550548 94.568 -58.617 54.070
550549 94.599 -58.630 54.036
550550 94.629 -58.644 54.008
550551 94.672 -58.654 53.985
550552 94.714 -58.665 53.968
550553 94.757 -58.676 53.952
550554 94.799 -58.687 53.935
550555 94.842 -58.698 53.917
550556 94.884 -58.708 53.897
550557 94.927 -58.719 53.874
550558 94.970 -58.730 53.846
550559 95.015 -58.742 53.814
550560 95.062 -58.754 53.782
550561 95.109 -58.767 53.754
550562 95.156 -58.779 53.728
550563 95.202 -58.791 53.701
550564 95.249 -58.804 53.680
550565 95.296 -58.816 53.659
550566 95.343 -58.828 53.638
550567 95.390 -58.841 53.613
550568 95.432 -58.854 53.582
550569 95.464 -58.865 53.557
550570 95.493 -58.874 53.530
550571 95.524 -58.883 53.502
550572 95.553 -58.892 53.474
550573 95.583 -58.902 53.448
550574 95.613 -58.910 53.423
550575 95.643 -58.919 53.396
550576 95.673 -58.928 53.369
550577 95.701 -58.936 53.343
550578 95.730 -58.943 53.317
550579 95.758 -58.951 53.287
550580 95.787 -58.958 53.258
550581 95.816 -58.966 53.227
550582 95.844 -58.974 53.194
550583 95.880 -58.980 53.161
550584 95.915 -58.986 53.130
550585 95.950 -58.991 53.099
550586 95.984 -58.998 53.076
550587 96.014 -59.005 53.055
550588 96.044 -59.012 53.037
550589 96.075 -59.018 53.020
550590 96.105 -59.025 53.000
550591 96.139 -59.032 52.982
550592 96.175 -59.038 52.961
550593 96.209 -59.044 52.937
550594 96.244 -59.051 52.914
550595 96.279 -59.057 52.891
550596 96.314 -59.063 52.866
550597 96.349 -59.070 52.840
550598 96.382 -59.077 52.811
550599 96.415 -59.084 52.783
550600 96.448 -59.090 52.758
550601 96.480 -59.093 52.734
550602 96.511 -59.096 52.704
550603 96.541 -59.098 52.678
550604 96.572 -59.100 52.652
550605 96.603 -59.102 52.626
550606 96.634 -59.105 52.602
550607 96.665 -59.107 52.576
550608 96.695 -59.109 52.544
550609 96.726 -59.112 52.510
550610 96.757 -59.111 52.472
550611 96.789 -59.110 52.433
550612 96.822 -59.111 52.394
550613 96.856 -59.112 52.355
550614 96.890 -59.113 52.308
550615 96.924 -59.114 52.267
550616 96.957 -59.115 52.231
550617 96.991 -59.115 52.197
550618 97.024 -59.117 52.164
550619 97.058 -59.118 52.130
550620 97.093 -59.119 52.090
550621 97.127 -59.122 52.049
550622 97.162 -59.125 52.006
550623 97.197 -59.128 51.960
550624 97.232 -59.131 51.913
550625 97.266 -59.133 51.864
550626 97.301 -59.135 51.813
550627 97.336 -59.138 51.764
550628 97.370 -59.142 51.718
550629 97.403 -59.146 51.672
550630 97.436 -59.150 51.628
550631 97.469 -59.154 51.580
550632 97.503 -59.158 51.532
550633 97.536 -59.162 51.483
550634 97.567 -59.167 51.434
550635 97.598 -59.171 51.388
550636 97.629 -59.176 51.346
550637 97.659 -59.181 51.309
550638 97.690 -59.186 51.273
550639 97.721 -59.190 51.237
550640 97.752 -59.196 51.199
550641 97.786 -59.198 51.160
550642 97.825 -59.196 51.119
550643 97.866 -59.197 51.076
550644 97.909 -59.200 51.031
550645 97.951 -59.202 50.987
550646 97.994 -59.204 50.944
550647 98.036 -59.206 50.902
550648 98.078 -59.208 50.862
550649 98.120 -59.211 50.824
550650 98.160 -59.215 50.786
550651 98.200 -59.220 50.751
550652 98.236 -59.222 50.719
550653 98.267 -59.217 50.689
550654 98.297 -59.206 50.660
550655 98.327 -59.194 50.629
550656 98.361 -59.184 50.596
550657 98.397 -59.174 50.565
550658 98.434 -59.164 50.534
550659 98.469 -59.153 50.505
550660 98.503 -59.142 50.472
550661 98.536 -59.131 50.439
550662 98.570 -59.121 50.405
550663 98.603 -59.110 50.370
550664 98.637 -59.099 50.336
550665 98.671 -59.088 50.304
550666 98.705 -59.077 50.271
550667 98.740 -59.072 50.239
550668 98.775 -59.067 50.206
550669 98.811 -59.062 50.171
550670 98.846 -59.057 50.136
550671 98.882 -59.054 50.099
550672 98.921 -59.062 50.061
550673 98.960 -59.068 50.023
550674 99.021 -59.110 49.983
550675 99.050 -59.114 49.944
550676 99.079 -59.118 49.908
550677 99.108 -59.122 49.873
550678 99.136 -59.126 49.838
550679 99.165 -59.130 49.803
550680 99.219 -59.128 49.769
550681 99.291 -59.122 49.736
550682 99.363 -59.116 49.704
550683 99.435 -59.110 49.672
550684 99.474 -59.116 49.640
551033 86.484 -55.722 57.663
551034 86.524 -55.736 57.656
551035 86.564 -55.749 57.647
551036 86.605 -55.762 57.638
551037 86.645 -55.776 57.630
551038 86.686 -55.789 57.621
551039 86.726 -55.803 57.611
551040 86.767 -55.816 57.599
551041 86.807 -55.829 57.586
551042 86.847 -55.843 57.571
551043 86.887 -55.856 57.554
551044 86.928 -55.869 57.535
551045 86.969 -55.883 57.513
551046 87.009 -55.896 57.489
551047 87.049 -55.909 57.462
551048 87.089 -55.922 57.431
551049 87.130 -55.936 57.399
551050 87.170 -55.950 57.366
551051 87.211 -55.962 57.332
551052 87.251 -55.976 57.297
551053 87.291 -55.989 57.260
551054 87.332 -56.003 57.223
551055 87.372 -56.016 57.184
551056 87.413 -56.029 57.147
551057 87.453 -56.043 57.109
551058 87.493 -56.056 57.070
551059 87.535 -56.069 57.031
551060 87.589 -56.073 56.995
551061 87.645 -56.077 56.958
551101 89.184 -56.652 55.992
551102 89.222 -56.668 55.983
551103 89.260 -56.684 55.974
551104 89.299 -56.700 55.967
551116 89.753 -56.871 55.914
551117 89.794 -56.881 55.910
551118 89.835 -56.893 55.906
551119 89.877 -56.903 55.899
551120 89.918 -56.914 55.891
551121 89.960 -56.924 55.880
551122 90.001 -56.936 55.867
551123 90.027 -56.953 55.852
551124 90.052 -56.970 55.836
551125 90.080 -56.987 55.817
551126 90.117 -57.001 55.798
551127 90.154 -57.015 55.777
551128 90.191 -57.030 55.755
551129 90.228 -57.044 55.733
551130 90.265 -57.058 55.708
551131 90.303 -57.072 55.690
551132 90.339 -57.087 55.671
551133 90.385 -57.103 55.647
551134 90.428 -57.119 55.615
551135 90.462 -57.135 55.579
551136 90.498 -57.151 55.541
551137 90.533 -57.168 55.507
551138 90.571 -57.184 55.477
551139 90.608 -57.200 55.449
551140 90.646 -57.217 55.423
551141 90.683 -57.233 55.399
551142 90.721 -57.250 55.378
551143 90.758 -57.266 55.358
551144 90.795 -57.282 55.340
551145 90.833 -57.299 55.325
551146 90.870 -57.315 55.312
551147 90.907 -57.332 55.299
551148 90.945 -57.348 55.288
551149 90.982 -57.364 55.279
551150 91.020 -57.381 55.271
551151 91.058 -57.397 55.264
551152 91.099 -57.407 55.258
551153 91.144 -57.414 55.256
551154 91.189 -57.420 55.255
551155 91.229 -57.443 55.255
551156 91.269 -57.467 55.257
551157 91.308 -57.490 55.259
551158 91.348 -57.514 55.262
551159 91.392 -57.524 55.265
551160 91.436 -57.535 55.268
551161 91.480 -57.545 55.271
551162 91.523 -57.555 55.271
551163 91.567 -57.566 55.271
551164 91.611 -57.576 55.269
551165 91.655 -57.587 55.266
551166 91.699 -57.598 55.262
551167 91.742 -57.608 55.255
551168 91.786 -57.619 55.249
551169 91.830 -57.629 55.247
551170 91.874 -57.639 55.246
551171 91.918 -57.650 55.245
551172 91.961 -57.661 55.243
551173 92.005 -57.671 55.240
551174 92.049 -57.682 55.240
551175 92.093 -57.692 55.239
551176 92.137 -57.702 55.228
551177 92.176 -57.717 55.211
551178 92.209 -57.738 55.193
551179 92.250 -57.752 55.179
551180 92.286 -57.767 55.173
551181 92.309 -57.782 55.157
551182 92.333 -57.798 55.139
551183 92.360 -57.817 55.119
551184 92.390 -57.838 55.098
551185 92.420 -57.858 55.076
551186 92.449 -57.880 55.059
551187 92.479 -57.900 55.041
551188 92.509 -57.921 55.024
551189 92.538 -57.942 55.007
551190 92.568 -57.963 54.990
551191 92.599 -57.981 54.972
551192 92.636 -57.991 54.951
551193 92.672 -58.002 54.930
551194 92.708 -58.009 54.907
551195 92.745 -58.015 54.882
551196 92.781 -58.021 54.859
551197 92.818 -58.027 54.843
551198 92.854 -58.033 54.836
551199 92.891 -58.039 54.830
551200 92.927 -58.045 54.821
551201 92.962 -58.050 54.808
551202 92.996 -58.057 54.788
551203 93.030 -58.065 54.762
551204 93.065 -58.074 54.734
551205 93.099 -58.081 54.708
551206 93.134 -58.090 54.689
551207 93.168 -58.098 54.683
551208 93.202 -58.106 54.688
551209 93.237 -58.115 54.700
551210 93.271 -58.123 54.712
551211 93.306 -58.131 54.721
551212 93.340 -58.140 54.725
551213 93.375 -58.148 54.725
551214 93.409 -58.156 54.718
551215 93.443 -58.165 54.714
551216 93.478 -58.173 54.708
551217 93.512 -58.181 54.703
551218 93.547 -58.189 54.702
551219 93.583 -58.201 54.703
551220 93.621 -58.221 54.704
551221 93.659 -58.241 54.708
551222 93.698 -58.261 54.711
551223 93.735 -58.281 54.710
551224 93.762 -58.292 54.706
551225 93.789 -58.304 54.698
551226 93.816 -58.316 54.686
551227 93.843 -58.328 54.669
551228 93.869 -58.340 54.647
551229 93.901 -58.351 54.623
551230 93.935 -58.363 54.594
551231 93.970 -58.373 54.559
551232 94.004 -58.385 54.522
551233 94.038 -58.396 54.487
551234 94.072 -58.407 54.457
551235 94.107 -58.418 54.432
551236 94.141 -58.429 54.415
551237 94.175 -58.441 54.395
551238 94.210 -58.452 54.372
551239 94.244 -58.463 54.346
551240 94.278 -58.474 54.316
551241 94.313 -58.485 54.285
551242 94.347 -58.497 54.258
551243 94.381 -58.508 54.235
551244 94.415 -58.520 54.213
551245 94.446 -58.532 54.190
551246 94.477 -58.546 54.166
551247 94.507 -58.559 54.136
551248 94.538 -58.573 54.099
551249 94.569 -58.585 54.058
551250 94.600 -58.598 54.021
551251 94.630 -58.612 53.991
551252 94.668 -58.623 53.966
551253 94.710 -58.635 53.946
551254 94.753 -58.645 53.927
551255 94.795 -58.656 53.911
551256 94.838 -58.667 53.895
551257 94.881 -58.677 53.878
551258 94.923 -58.689 53.856
551259 94.968 -58.700 53.833
551260 95.015 -58.712 53.803
551261 95.061 -58.724 53.772
551262 95.108 -58.737 53.742
551263 95.155 -58.749 53.717
551264 95.202 -58.761 53.692
551265 95.249 -58.774 53.670
551266 95.296 -58.786 53.648
551267 95.343 -58.799 53.625
551268 95.390 -58.811 53.599
551269 95.438 -58.823 53.569
551270 95.468 -58.833 53.540
551271 95.498 -58.841 53.512
551272 95.528 -58.850 53.483
551273 95.558 -58.859 53.453
551274 95.588 -58.868 53.427
551275 95.618 -58.877 53.401
551276 95.648 -58.886 53.376
551277 95.677 -58.895 53.349
551278 95.707 -58.904 53.322
551279 95.736 -58.911 53.294
551280 95.764 -58.919 53.265
551281 95.793 -58.927 53.236
551282 95.821 -58.934 53.207
551283 95.850 -58.941 53.175
551284 95.886 -58.947 53.141
551285 95.921 -58.953 53.109
551286 95.957 -58.959 53.078
551287 95.987 -58.966 53.053
551288 96.017 -58.973 53.031
551289 96.047 -58.980 53.011
551290 96.078 -58.987 52.992
551291 96.108 -58.994 52.970
551292 96.143 -59.000 52.950
551293 96.178 -59.006 52.926
551294 96.213 -59.012 52.903
551295 96.248 -59.019 52.881
551296 96.283 -59.025 52.857
551297 96.318 -59.031 52.830
551298 96.353 -59.038 52.807
551299 96.387 -59.045 52.779
551300 96.420 -59.051 52.750
551301 96.453 -59.058 52.723
551302 96.484 -59.060 52.699
551303 96.515 -59.062 52.668
551304 96.545 -59.065 52.639
551305 96.576 -59.067 52.612
551306 96.607 -59.069 52.588
551307 96.638 -59.071 52.565
551308 96.669 -59.074 52.541
551309 96.699 -59.077 52.508
551310 96.730 -59.079 52.476
551311 96.761 -59.081 52.438
551312 96.794 -59.081 52.401
551313 96.827 -59.082 52.364
551314 96.861 -59.084 52.324
551315 96.895 -59.084 52.274
551316 96.929 -59.085 52.232
551317 96.962 -59.086 52.194
551318 96.996 -59.087 52.158
551319 97.029 -59.087 52.126
551320 97.063 -59.088 52.094
551321 97.097 -59.089 52.056
551322 97.132 -59.092 52.017
551323 97.166 -59.095 51.975
551324 97.201 -59.097 51.931
551325 97.236 -59.099 51.886
551326 97.271 -59.102 51.839
551327 97.305 -59.105 51.790
551328 97.340 -59.108 51.743
551329 97.373 -59.112 51.697
551330 97.406 -59.115 51.651
551331 97.440 -59.120 51.608
551332 97.473 -59.124 51.562
551333 97.506 -59.128 51.515
551334 97.539 -59.132 51.467
551335 97.572 -59.137 51.420
551336 97.603 -59.141 51.375
551337 97.634 -59.146 51.335
551338 97.664 -59.150 51.300
551339 97.696 -59.156 51.260
551340 97.727 -59.160 51.219
551341 97.758 -59.165 51.180
551342 97.794 -59.165 51.141
551343 97.833 -59.164 51.100
551344 97.873 -59.162 51.058
551345 97.913 -59.161 51.015
551346 97.952 -59.159 50.971
551347 97.993 -59.160 50.928
551348 98.035 -59.162 50.885
551349 98.077 -59.165 50.842
551350 98.117 -59.169 50.799
551351 98.157 -59.174 50.759
551352 98.197 -59.179 50.724
551353 98.236 -59.183 50.693
551354 98.267 -59.178 50.667
551355 98.298 -59.173 50.640
551356 98.328 -59.162 50.612
551357 98.361 -59.151 50.583
551358 98.398 -59.141 50.554
551359 98.433 -59.131 50.527
551360 98.466 -59.119 50.497
551361 98.500 -59.109 50.467
551362 98.534 -59.098 50.434
551363 98.567 -59.087 50.402
551364 98.601 -59.076 50.369
551365 98.635 -59.065 50.336
551366 98.668 -59.055 50.303
551367 98.702 -59.044 50.270
551368 98.736 -59.033 50.236
551369 98.770 -59.026 50.203
551370 98.806 -59.021 50.170
551371 98.841 -59.016 50.136
551372 98.879 -59.020 50.098
551373 98.918 -59.027 50.061
551374 98.957 -59.034 50.024
551375 99.007 -59.057 49.986
551376 99.049 -59.077 49.949
551377 99.077 -59.081 49.915
551378 99.106 -59.085 49.880
551379 99.135 -59.089 49.845
551380 99.163 -59.093 49.810
551381 99.226 -59.089 49.776
551382 99.298 -59.083 49.745
551383 99.370 -59.077 49.713
551384 99.439 -59.072 49.681
551385 99.475 -59.080 49.648
551386 99.511 -59.087 49.616
551734 86.500 -55.679 57.663
551735 86.541 -55.693 57.657
551736 86.581 -55.705 57.649
551737 86.622 -55.719 57.640
551738 86.662 -55.733 57.632
551739 86.702 -55.746 57.623
551740 86.743 -55.759 57.614
551741 86.783 -55.772 57.602
551742 86.824 -55.786 57.590
551743 86.864 -55.799 57.575
551744 86.905 -55.812 57.559
551745 86.945 -55.826 57.541
551746 86.985 -55.839 57.522
551747 87.025 -55.853 57.501
551748 87.066 -55.866 57.477
551749 87.107 -55.879 57.446
551750 87.147 -55.893 57.415
551751 87.187 -55.906 57.383
551752 87.227 -55.919 57.350
551753 87.268 -55.933 57.317
551754 87.308 -55.946 57.281
551755 87.349 -55.959 57.243
551756 87.389 -55.972 57.205
551757 87.429 -55.986 57.167
551758 87.470 -56.000 57.129
551759 87.510 -56.013 57.090
551760 87.560 -56.020 57.050
551761 87.615 -56.024 57.012
551762 87.670 -56.028 56.975
551763 87.721 -56.034 56.937
551799 89.087 -56.566 56.025
551800 89.125 -56.582 56.011
551801 89.163 -56.598 55.997
551802 89.201 -56.614 55.985
551803 89.240 -56.630 55.976
551804 89.278 -56.646 55.968
551805 89.316 -56.662 55.962
551806 89.354 -56.678 55.955
551807 89.393 -56.694 55.949
551816 89.731 -56.820 55.925
551817 89.772 -56.831 55.924
551818 89.814 -56.842 55.923
551819 89.855 -56.852 55.920
551820 89.896 -56.864 55.914
551821 89.938 -56.874 55.906
551822 89.979 -56.885 55.896
551823 90.009 -56.901 55.883
551824 90.034 -56.918 55.868
551825 90.059 -56.935 55.852
551826 90.091 -56.951 55.832
551827 90.128 -56.965 55.811
551828 90.165 -56.979 55.790
551829 90.202 -56.993 55.768
551830 90.240 -57.008 55.744
551831 90.276 -57.022 55.718
551832 90.314 -57.037 55.700
551833 90.358 -57.053 55.680
551834 90.404 -57.069 55.653
551835 90.451 -57.084 55.620
551836 90.486 -57.101 55.585
551837 90.521 -57.117 55.548
551838 90.556 -57.133 55.511
551839 90.591 -57.150 55.478
551840 90.626 -57.166 55.450
551841 90.663 -57.182 55.424
551842 90.700 -57.198 55.400
551843 90.737 -57.215 55.378
551844 90.775 -57.232 55.358
551845 90.812 -57.247 55.340
551846 90.850 -57.264 55.325
551847 90.887 -57.280 55.312
551848 90.925 -57.297 55.299
551849 90.962 -57.313 55.288
551850 91.000 -57.329 55.279
551851 91.037 -57.346 55.271
551852 91.074 -57.363 55.265
551853 91.113 -57.376 55.260
551854 91.159 -57.383 55.259
551855 91.204 -57.389 55.263
551856 91.243 -57.413 55.268
551857 91.283 -57.436 55.275
551858 91.322 -57.460 55.281
551859 91.365 -57.474 55.285
551860 91.409 -57.485 55.288
551861 91.453 -57.495 55.291
551862 91.497 -57.506 55.291
551863 91.540 -57.517 55.287
551864 91.584 -57.527 55.284
551865 91.628 -57.538 55.276
551866 91.672 -57.548 55.269
551867 91.716 -57.558 55.263
551868 91.759 -57.569 55.260
551869 91.803 -57.579 55.265
551870 91.847 -57.590 55.268
551871 91.891 -57.601 55.268
551872 91.935 -57.611 55.266
551873 91.979 -57.621 55.265
551874 92.022 -57.632 55.265
551875 92.066 -57.642 55.266
551876 92.110 -57.653 55.260
551877 92.154 -57.664 55.251
551878 92.196 -57.675 55.238
551879 92.232 -57.693 55.223
551880 92.264 -57.708 55.207
551881 92.289 -57.724 55.195
551882 92.318 -57.746 55.181
551883 92.348 -57.766 55.164
551884 92.378 -57.787 55.144
551885 92.407 -57.808 55.122
551886 92.437 -57.829 55.101
551887 92.466 -57.849 55.087
551888 92.496 -57.871 55.067
551889 92.526 -57.891 55.046
551890 92.556 -57.912 55.026
551891 92.585 -57.933 55.005
551892 92.620 -57.946 54.985
551893 92.656 -57.957 54.961
551894 92.692 -57.968 54.935
551895 92.728 -57.974 54.908
551896 92.765 -57.981 54.883
551897 92.801 -57.987 54.862
551898 92.838 -57.993 54.850
551899 92.874 -57.999 54.844
551900 92.911 -58.006 54.837
551901 92.947 -58.012 54.827
551902 92.974 -58.017 54.814
551903 93.005 -58.016 54.796
551904 93.039 -58.021 54.774
551905 93.073 -58.029 54.748
551906 93.108 -58.037 54.720
551907 93.142 -58.046 54.697
551908 93.177 -58.054 54.687
551909 93.211 -58.062 54.691
551910 93.245 -58.071 54.702
551911 93.280 -58.079 54.715
551912 93.314 -58.087 54.729
551913 93.349 -58.096 54.738
551914 93.383 -58.104 54.740
551915 93.418 -58.112 54.735
551916 93.452 -58.120 54.730
551917 93.486 -58.128 54.723
551918 93.521 -58.137 54.719
551919 93.555 -58.145 54.720
551920 93.591 -58.159 54.722
551921 93.630 -58.178 54.723
551922 93.668 -58.198 54.727
551923 93.706 -58.219 54.730
551924 93.745 -58.238 54.728
551925 93.783 -58.258 54.723
551926 93.811 -58.270 54.715
551927 93.837 -58.282 54.705
551928 93.864 -58.294 54.690
551929 93.891 -58.306 54.672
551930 93.918 -58.318 54.650
551931 93.944 -58.329 54.621
551932 93.971 -58.341 54.587
551933 94.005 -58.353 54.547
551934 94.040 -58.363 54.505
551935 94.074 -58.375 54.465
551936 94.108 -58.386 54.430
551937 94.142 -58.397 54.406
551938 94.177 -58.409 54.379
551939 94.211 -58.419 54.353
551940 94.246 -58.431 54.325
551941 94.280 -58.442 54.297
551942 94.314 -58.454 54.267
551943 94.348 -58.464 54.239
551944 94.383 -58.476 54.216
551945 94.416 -58.488 54.193
551946 94.447 -58.501 54.169
551947 94.477 -58.514 54.145
551948 94.508 -58.527 54.117
551949 94.539 -58.541 54.083
551950 94.569 -58.554 54.044
551951 94.600 -58.567 54.007
551952 94.631 -58.580 53.976
551953 94.665 -58.593 53.948
551954 94.707 -58.604 53.925
551955 94.750 -58.614 53.905
551956 94.792 -58.625 53.887
551957 94.835 -58.636 53.872
551958 94.877 -58.647 53.857
551959 94.920 -58.657 53.837
551960 94.967 -58.670 53.817
551961 95.014 -58.683 53.791
551962 95.061 -58.695 53.760
551963 95.108 -58.708 53.729
551964 95.155 -58.720 53.703
551965 95.202 -58.732 53.678
551966 95.249 -58.745 53.656
551967 95.296 -58.757 53.632
551968 95.346 -58.768 53.607
551969 95.399 -58.777 53.581
551970 95.440 -58.783 53.553
551971 95.471 -58.793 53.524
551972 95.502 -58.807 53.496
551973 95.533 -58.817 53.465
551974 95.562 -58.826 53.435
551975 95.592 -58.835 53.408
551976 95.622 -58.844 53.380
551977 95.652 -58.853 53.353
551978 95.682 -58.862 53.326
551979 95.712 -58.871 53.298
551980 95.741 -58.879 53.268
551981 95.770 -58.886 53.238
551982 95.798 -58.894 53.209
551983 95.827 -58.902 53.181
551984 95.856 -58.909 53.151
551985 95.892 -58.915 53.119
551986 95.927 -58.921 53.087
551987 95.960 -58.927 53.055
551988 95.990 -58.934 53.029
551989 96.021 -58.941 53.008
551990 96.051 -58.949 52.989
551991 96.081 -58.955 52.968
551992 96.112 -58.962 52.944
551993 96.147 -58.968 52.919
551994 96.182 -58.974 52.892
551995 96.217 -58.981 52.868
551996 96.252 -58.987 52.845
551997 96.287 -58.993 52.822
551998 96.322 -59.000 52.795
551999 96.357 -59.006 52.776
552000 96.391 -59.012 52.749
552001 96.425 -59.019 52.719
552002 96.457 -59.024 52.689
552003 96.488 -59.027 52.663
552004 96.519 -59.029 52.634
552005 96.549 -59.031 52.602
552006 96.580 -59.034 52.577
552007 96.611 -59.036 52.552
552008 96.642 -59.038 52.527
552009 96.673 -59.040 52.503
552010 96.704 -59.041 52.471
552011 96.735 -59.042 52.439
552012 96.766 -59.043 52.403
552013 96.799 -59.047 52.367
552014 96.832 -59.048 52.332
552015 96.866 -59.048 52.290
552016 96.899 -59.049 52.239
552017 96.933 -59.049 52.196
552018 96.967 -59.050 52.157
552019 97.001 -59.051 52.120
552020 97.034 -59.052 52.089
552021 97.068 -59.052 52.058
552022 97.101 -59.054 52.022
552023 97.135 -59.058 51.984
552024 97.170 -59.061 51.943
552025 97.205 -59.065 51.901
552026 97.240 -59.068 51.860
552027 97.275 -59.071 51.814
552028 97.309 -59.074 51.769
552029 97.344 -59.077 51.723
552030 97.377 -59.081 51.677
552031 97.410 -59.086 51.630
552032 97.443 -59.090 51.589
552033 97.476 -59.093 51.542
552034 97.510 -59.098 51.496
552035 97.543 -59.102 51.451
552036 97.576 -59.106 51.406
552037 97.608 -59.111 51.361
552038 97.639 -59.115 51.321
552039 97.670 -59.120 51.285
552040 97.701 -59.124 51.243
552041 97.732 -59.129 51.200
552042 97.764 -59.134 51.158
552043 97.802 -59.133 51.118
552044 97.842 -59.131 51.076
552045 97.881 -59.130 51.034
552046 97.921 -59.128 50.993
552047 97.960 -59.127 50.950
552048 98.000 -59.125 50.907
552049 98.039 -59.124 50.864
552050 98.075 -59.123 50.820
552051 98.114 -59.128 50.776
552052 98.154 -59.132 50.735
552053 98.194 -59.137 50.698
552054 98.233 -59.142 50.667
552055 98.267 -59.140 50.642
552056 98.298 -59.135 50.618
552057 98.329 -59.129 50.593
552058 98.362 -59.118 50.567
552059 98.397 -59.108 50.540
552060 98.430 -59.097 50.512
552061 98.464 -59.086 50.483
552062 98.498 -59.075 50.454
552063 98.531 -59.065 50.423
552064 98.565 -59.054 50.392
552065 98.599 -59.043 50.361
552066 98.632 -59.032 50.331
552067 98.666 -59.021 50.299
552068 98.700 -59.011 50.267
552069 98.733 -58.999 50.235
552070 98.767 -58.989 50.202
552071 98.801 -58.979 50.170
552072 98.838 -58.978 50.137
552073 98.876 -58.986 50.100
552074 98.915 -58.993 50.064
552075 98.954 -58.999 50.029
552076 98.993 -59.006 49.993
552077 99.047 -59.040 49.956
552078 99.076 -59.043 49.922
552079 99.104 -59.048 49.888
552080 99.133 -59.052 49.853
552081 99.162 -59.055 49.817
552082 99.234 -59.049 49.783
552083 99.306 -59.043 49.753
552084 99.378 -59.038 49.720
552085 99.440 -59.036 49.688
552086 99.476 -59.043 49.657
552087 99.512 -59.050 49.624
552088 99.548 -59.058 49.592
552089 99.583 -59.065 49.560
552436 86.558 -55.649 57.657
552437 86.598 -55.662 57.649
552438 86.638 -55.676 57.641
552439 86.679 -55.689 57.633
552440 86.719 -55.702 57.625
552441 86.760 -55.716 57.616
552442 86.800 -55.729 57.606
552443 86.841 -55.743 57.593
552444 86.881 -55.756 57.579
552445 86.921 -55.769 57.564
552446 86.962 -55.783 57.547
552447 87.002 -55.796 57.530
552448 87.043 -55.809 57.510
552449 87.083 -55.822 57.486
552450 87.123 -55.836 57.458
552451 87.163 -55.850 57.429
552452 87.204 -55.862 57.399
552453 87.245 -55.876 57.368
552454 87.285 -55.890 57.336
552455 87.325 -55.903 57.301
552456 87.365 -55.916 57.263
552457 87.406 -55.929 57.226
552458 87.446 -55.943 57.189
552459 87.487 -55.956 57.150
552460 87.530 -55.968 57.111
552461 87.585 -55.972 57.071
552462 87.640 -55.975 57.032
552463 87.695 -55.980 56.993
552464 87.741 -55.986 56.955
552465 87.786 -55.994 56.917
552498 89.028 -56.495 56.050
552499 89.066 -56.511 56.034
552500 89.105 -56.527 56.020
552501 89.143 -56.543 56.005
552502 89.181 -56.559 55.992
552503 89.219 -56.575 55.981
552504 89.258 -56.592 55.971
552505 89.296 -56.608 55.963
552506 89.334 -56.624 55.957
552507 89.373 -56.639 55.951
552508 89.410 -56.655 55.945
552509 89.446 -56.673 55.939
552510 89.470 -56.697 55.934
552516 89.709 -56.770 55.931
552517 89.750 -56.780 55.933
552518 89.792 -56.791 55.935
552519 89.833 -56.802 55.936
552520 89.874 -56.813 55.934
552521 89.916 -56.824 55.930
552522 89.957 -56.835 55.924
552523 89.990 -56.849 55.914
552524 90.016 -56.866 55.902
552525 90.041 -56.883 55.888
552526 90.067 -56.900 55.870
552527 90.102 -56.915 55.850
552528 90.140 -56.929 55.828
552529 90.177 -56.943 55.807
552530 90.214 -56.958 55.783
552531 90.251 -56.972 55.756
552532 90.288 -56.987 55.729
552533 90.331 -57.002 55.706
552534 90.378 -57.018 55.682
552535 90.424 -57.034 55.654
552536 90.470 -57.050 55.621
552537 90.509 -57.066 55.590
552538 90.544 -57.083 55.554
552539 90.579 -57.099 55.515
552540 90.614 -57.116 55.480
552541 90.649 -57.131 55.451
552542 90.684 -57.147 55.424
552543 90.719 -57.164 55.400
552544 90.754 -57.180 55.378
552545 90.792 -57.197 55.358
552546 90.829 -57.213 55.341
552547 90.867 -57.229 55.325
552548 90.904 -57.246 55.312
552549 90.942 -57.262 55.301
552550 90.979 -57.279 55.290
552551 91.016 -57.295 55.280
552552 91.054 -57.311 55.273
552553 91.091 -57.328 55.268
552554 91.129 -57.344 55.265
552555 91.173 -57.351 55.266
552556 91.218 -57.359 55.272
552557 91.257 -57.383 55.282
552558 91.297 -57.407 55.291
552559 91.338 -57.425 55.300
552560 91.382 -57.435 55.305
552561 91.426 -57.446 55.310
552562 91.469 -57.457 55.310
552563 91.513 -57.467 55.306
552564 91.557 -57.477 55.300
552565 91.601 -57.488 55.293
552566 91.645 -57.498 55.280
552567 91.688 -57.509 55.269
552568 91.732 -57.520 55.262
552569 91.776 -57.529 55.262
552570 91.820 -57.540 55.273
552571 91.864 -57.551 55.278
552572 91.908 -57.561 55.278
552573 91.951 -57.572 55.277
552574 91.995 -57.583 55.276
552575 92.039 -57.592 55.276
552576 92.083 -57.603 55.278
552577 92.127 -57.614 55.275
552578 92.170 -57.624 55.269
552579 92.214 -57.635 55.259
552580 92.246 -57.653 55.245
552581 92.276 -57.673 55.230
552582 92.306 -57.695 55.217
552583 92.336 -57.715 55.200
552584 92.365 -57.736 55.179
552585 92.395 -57.757 55.156
552586 92.425 -57.778 55.135
552587 92.454 -57.799 55.118
552588 92.484 -57.820 55.106
552589 92.513 -57.840 55.088
552590 92.543 -57.861 55.066
552591 92.573 -57.882 55.041
552592 92.603 -57.901 55.014
552593 92.640 -57.912 54.992
552594 92.676 -57.922 54.966
552595 92.712 -57.933 54.937
552596 92.748 -57.941 54.910
552597 92.785 -57.947 54.887
552598 92.821 -57.953 54.867
552599 92.858 -57.959 54.855
552600 92.895 -57.965 54.844
552601 92.931 -57.971 54.832
552602 92.960 -57.977 54.819
552603 92.986 -57.983 54.805
552604 93.017 -57.982 54.790
552605 93.049 -57.981 54.772
552606 93.081 -57.985 54.749
552607 93.116 -57.993 54.726
552608 93.150 -58.002 54.707
552609 93.185 -58.010 54.696
552610 93.219 -58.018 54.696
552611 93.254 -58.027 54.704
552612 93.288 -58.035 54.716
552613 93.323 -58.043 54.728
552614 93.357 -58.051 54.735
552615 93.392 -58.059 54.745
552616 93.426 -58.068 54.747
552617 93.460 -58.076 54.745
552618 93.495 -58.084 54.739
552619 93.529 -58.093 54.733
552620 93.564 -58.101 54.736
552621 93.601 -58.116 54.740
552622 93.639 -58.136 54.741
552623 93.677 -58.156 54.743
552624 93.715 -58.176 54.743
552625 93.754 -58.196 54.741
552626 93.792 -58.216 54.734
552627 93.830 -58.235 54.726
552628 93.859 -58.248 54.717
552629 93.886 -58.260 54.704
552630 93.912 -58.272 54.688
552631 93.939 -58.284 54.668
552632 93.966 -58.296 54.641
552633 93.992 -58.307 54.608
552634 94.019 -58.319 54.567
552635 94.046 -58.331 54.522
552636 94.075 -58.343 54.476
552637 94.110 -58.354 54.434
552638 94.144 -58.365 54.402
552639 94.179 -58.376 54.370
552640 94.213 -58.388 54.339
552641 94.247 -58.399 54.310
552642 94.281 -58.410 54.282
552643 94.316 -58.421 54.253
552644 94.350 -58.432 54.225
552645 94.384 -58.444 54.200
552646 94.417 -58.456 54.176
552647 94.448 -58.469 54.152
552648 94.478 -58.482 54.129
552649 94.509 -58.495 54.099
552650 94.540 -58.509 54.067
552651 94.570 -58.522 54.032
552652 94.601 -58.535 53.996
552653 94.632 -58.548 53.962
552654 94.663 -58.562 53.932
552655 94.703 -58.573 53.909
552656 94.746 -58.584 53.889
552657 94.788 -58.595 53.871
552658 94.831 -58.605 53.854
552659 94.874 -58.616 53.840
552660 94.920 -58.628 53.823
552661 94.967 -58.641 53.801
552662 95.014 -58.653 53.776
552663 95.061 -58.665 53.747
552664 95.108 -58.678 53.715
552665 95.155 -58.690 53.683
552666 95.201 -58.702 53.659
552667 95.254 -58.713 53.636
552668 95.307 -58.722 53.610
552669 95.360 -58.732 53.583
552670 95.413 -58.742 53.558
552671 95.442 -58.742 53.534
552672 95.472 -58.751 53.507
552673 95.503 -58.764 53.479
552674 95.535 -58.778 53.449
552675 95.566 -58.791 53.418
552676 95.597 -58.802 53.390
552677 95.627 -58.811 53.359
552678 95.657 -58.820 53.331
552679 95.687 -58.829 53.303
552680 95.716 -58.838 53.275
552681 95.746 -58.847 53.246
552682 95.775 -58.855 53.215
552683 95.804 -58.862 53.186
552684 95.832 -58.870 53.159
552685 95.863 -58.877 53.129
552686 95.898 -58.883 53.099
552687 95.933 -58.889 53.067
552688 95.963 -58.896 53.035
552689 95.993 -58.902 53.007
552690 96.024 -58.910 52.983
552691 96.054 -58.917 52.962
552692 96.084 -58.924 52.941
552693 96.115 -58.930 52.918
552694 96.151 -58.936 52.894
552695 96.185 -58.943 52.867
552696 96.220 -58.949 52.840
552697 96.255 -58.955 52.815
552698 96.290 -58.961 52.791
552699 96.325 -58.968 52.766
552700 96.360 -58.974 52.746
552701 96.395 -58.980 52.718
552702 96.430 -58.987 52.686
552703 96.461 -58.991 52.655
552704 96.492 -58.993 52.628
552705 96.523 -58.996 52.597
552706 96.553 -58.997 52.564
552707 96.585 -58.998 52.537
552708 96.616 -58.999 52.513
552709 96.647 -58.999 52.486
552710 96.678 -59.000 52.460
552711 96.709 -59.002 52.426
552712 96.740 -59.002 52.395
552713 96.771 -59.005 52.363
552714 96.803 -59.011 52.328
552715 96.837 -59.012 52.292
552716 96.871 -59.012 52.250
552717 96.904 -59.013 52.206
552718 96.938 -59.014 52.164
552719 96.971 -59.015 52.123
552720 97.005 -59.015 52.084
552721 97.039 -59.016 52.050
552722 97.073 -59.017 52.020
552723 97.106 -59.019 51.986
552724 97.140 -59.024 51.950
552725 97.174 -59.027 51.910
552726 97.209 -59.030 51.868
552727 97.243 -59.034 51.831
552728 97.278 -59.037 51.791
552729 97.313 -59.040 51.749
552730 97.347 -59.043 51.704
552731 97.380 -59.048 51.656
552732 97.413 -59.052 51.610
552733 97.446 -59.056 51.571
552734 97.479 -59.060 51.523
552735 97.513 -59.065 51.476
552736 97.546 -59.068 51.433
552737 97.579 -59.073 51.392
552738 97.613 -59.080 51.346
552739 97.645 -59.084 51.304
552740 97.676 -59.090 51.265
552741 97.707 -59.094 51.222
552742 97.738 -59.099 51.177
552743 97.771 -59.102 51.134
552744 97.811 -59.100 51.092
552745 97.850 -59.099 51.049
552746 97.890 -59.097 51.007
552747 97.929 -59.096 50.966
552748 97.968 -59.094 50.926
552749 98.008 -59.093 50.883
552750 98.049 -59.087 50.840
552751 98.085 -59.085 50.796
552752 98.118 -59.087 50.753
552753 98.151 -59.090 50.711
552754 98.191 -59.096 50.672
552755 98.231 -59.100 50.640
552756 98.267 -59.101 50.616
552757 98.298 -59.096 50.595
552758 98.329 -59.092 50.573
552759 98.361 -59.085 50.549
552760 98.395 -59.074 50.523
552761 98.429 -59.064 50.493
552762 98.463 -59.053 50.465
552763 98.497 -59.042 50.437
552764 98.531 -59.031 50.407
552765 98.565 -59.021 50.378
552766 98.599 -59.010 50.350
552767 98.633 -58.999 50.321
552768 98.667 -58.988 50.291
552769 98.701 -58.977 50.261
552770 98.735 -58.967 50.231
552771 98.769 -58.956 50.201
552772 98.802 -58.946 50.169
552773 98.835 -58.945 50.138
552774 98.873 -58.951 50.104
552775 98.913 -58.958 50.069
552776 98.951 -58.965 50.036
552777 98.990 -58.972 50.001
552778 99.041 -58.996 49.965
552779 99.074 -59.006 49.929
552780 99.103 -59.011 49.895
552781 99.132 -59.015 49.860
552782 99.170 -59.016 49.824
552783 99.242 -59.010 49.791
552784 99.314 -59.005 49.760
552785 99.386 -58.999 49.728
552786 99.441 -58.999 49.696
552787 99.477 -59.006 49.664
552788 99.513 -59.014 49.632
552789 99.548 -59.021 49.600
552790 99.584 -59.029 49.569
552791 99.620 -59.037 49.536
552792 99.655 -59.044 49.503
552793 99.691 -59.051 49.470
553138 86.615 -55.619 57.649
553139 86.655 -55.633 57.642
553140 86.696 -55.646 57.634
553141 86.736 -55.659 57.626
553142 86.777 -55.673 57.619
553143 86.817 -55.686 57.609
553144 86.857 -55.699 57.596
553145 86.898 -55.712 57.583
553146 86.938 -55.726 57.568
553147 86.979 -55.740 57.554
553148 87.019 -55.752 57.537
553149 87.059 -55.766 57.519
553150 87.099 -55.779 57.496
553151 87.140 -55.793 57.471
553152 87.181 -55.806 57.444
553153 87.221 -55.819 57.416
553154 87.261 -55.833 57.387
553155 87.301 -55.846 57.356
553156 87.342 -55.859 57.322
553157 87.382 -55.873 57.285
553158 87.423 -55.886 57.249
553159 87.463 -55.900 57.212
553160 87.503 -55.912 57.173
553161 87.555 -55.919 57.134
553162 87.610 -55.923 57.093
553163 87.666 -55.927 57.053
553164 87.717 -55.932 57.013
553165 87.761 -55.939 56.973
553166 87.802 -55.948 56.934
553167 87.836 -55.959 56.898
553199 89.046 -56.457 56.046
553200 89.084 -56.473 56.030
553201 89.122 -56.489 56.015
553202 89.161 -56.505 56.001
553203 89.199 -56.521 55.987
553204 89.237 -56.537 55.976
553205 89.275 -56.553 55.966
553206 89.314 -56.569 55.959
553207 89.352 -56.585 55.952
553208 89.390 -56.601 55.946
553209 89.428 -56.617 55.940
553210 89.464 -56.634 55.934
553211 89.488 -56.658 55.931
553212 89.521 -56.676 55.929
553213 89.562 -56.686 55.929
553215 89.646 -56.708 55.934
553216 89.687 -56.719 55.934
553217 89.728 -56.730 55.939
553218 89.769 -56.741 55.943
553219 89.811 -56.752 55.946
553220 89.852 -56.762 55.948
553221 89.894 -56.774 55.949
553222 89.935 -56.784 55.947
553223 89.972 -56.797 55.943
553224 89.998 -56.814 55.935
553225 90.023 -56.831 55.924
553226 90.049 -56.848 55.910
553227 90.077 -56.865 55.891
553228 90.114 -56.879 55.871
553229 90.151 -56.893 55.849
553230 90.188 -56.908 55.825
553231 90.225 -56.922 55.799
553232 90.262 -56.937 55.769
553233 90.305 -56.952 55.739
553234 90.351 -56.968 55.710
553235 90.397 -56.984 55.681
553236 90.444 -57.000 55.651
553237 90.490 -57.016 55.618
553238 90.533 -57.032 55.591
553239 90.568 -57.049 55.557
553240 90.603 -57.065 55.518
553241 90.638 -57.081 55.482
553242 90.673 -57.097 55.452
553243 90.708 -57.113 55.425
553244 90.743 -57.130 55.401
553245 90.778 -57.146 55.379
553246 90.813 -57.162 55.359
553247 90.848 -57.178 55.342
553248 90.884 -57.194 55.327
553249 90.921 -57.211 55.313
553250 90.958 -57.227 55.302
553251 90.996 -57.244 55.292
553252 91.033 -57.260 55.282
553253 91.071 -57.276 55.275
553254 91.108 -57.293 55.270
553255 91.146 -57.310 55.268
553256 91.187 -57.320 55.268
553257 91.231 -57.329 55.274
553258 91.271 -57.353 55.284
553259 91.311 -57.376 55.296
553260 91.355 -57.385 55.305
553261 91.399 -57.396 55.309
553262 91.443 -57.407 55.311
553263 91.487 -57.417 55.308
553264 91.530 -57.428 55.303
553265 91.574 -57.438 55.296
553266 91.618 -57.448 55.289
553267 91.662 -57.459 55.278
553268 91.706 -57.470 55.266
553269 91.749 -57.480 55.259
553270 91.793 -57.491 55.263
553271 91.837 -57.501 55.272
553272 91.881 -57.511 55.278
553273 91.925 -57.522 55.280
553274 91.968 -57.532 55.280
553275 92.012 -57.543 55.279
553276 92.055 -57.553 55.280
553277 92.098 -57.562 55.283
553278 92.141 -57.572 55.282
553279 92.184 -57.582 55.278
553280 92.226 -57.597 55.270
553281 92.264 -57.623 55.257
553282 92.294 -57.644 55.243
553283 92.323 -57.664 55.226
553284 92.353 -57.686 55.207
553285 92.382 -57.706 55.184
553286 92.412 -57.727 55.160
553287 92.442 -57.748 55.138
553288 92.472 -57.768 55.120
553289 92.501 -57.789 55.103
553290 92.531 -57.810 55.088
553291 92.561 -57.831 55.071
553292 92.590 -57.852 55.050
553293 92.624 -57.867 55.025
553294 92.660 -57.877 55.000
553295 92.696 -57.888 54.974
553296 92.732 -57.899 54.948
553297 92.768 -57.908 54.924
553298 92.805 -57.914 54.902
553299 92.841 -57.920 54.879
553300 92.878 -57.926 54.857
553301 92.914 -57.932 54.836
553302 92.948 -57.938 54.817
553303 92.973 -57.944 54.800
553304 92.998 -57.949 54.789
553305 93.029 -57.949 54.779
553306 93.060 -57.948 54.767
553307 93.091 -57.947 54.752
553308 93.124 -57.949 54.736
553309 93.158 -57.958 54.724
553310 93.193 -57.966 54.716
553311 93.227 -57.974 54.710
553312 93.262 -57.982 54.711
553313 93.296 -57.990 54.720
553314 93.331 -57.999 54.731
553315 93.365 -58.007 54.738
553316 93.400 -58.015 54.746
553317 93.434 -58.024 54.752
553318 93.469 -58.032 54.754
553319 93.503 -58.040 54.751
553320 93.538 -58.049 54.745
553321 93.572 -58.057 54.742
553322 93.609 -58.074 54.746
553323 93.647 -58.093 54.750
553324 93.686 -58.113 54.751
553325 93.724 -58.133 54.748
553326 93.762 -58.153 54.743
553327 93.800 -58.173 54.734
553328 93.839 -58.193 54.724
553329 93.877 -58.212 54.714
553330 93.907 -58.227 54.705
553331 93.934 -58.238 54.693
553332 93.961 -58.250 54.675
553333 93.987 -58.262 54.650
553334 94.014 -58.274 54.618
553335 94.041 -58.285 54.581
553336 94.067 -58.297 54.536
553337 94.094 -58.310 54.488
553338 94.121 -58.321 54.444
553339 94.148 -58.333 54.405
553340 94.180 -58.344 54.367
553341 94.214 -58.355 54.331
553342 94.249 -58.366 54.299
553343 94.283 -58.378 54.270
553344 94.317 -58.389 54.241
553345 94.351 -58.400 54.212
553346 94.386 -58.411 54.185
553347 94.417 -58.424 54.160
553348 94.448 -58.438 54.136
553349 94.479 -58.451 54.113
553350 94.509 -58.464 54.083
553351 94.540 -58.477 54.051
553352 94.571 -58.491 54.018
553353 94.602 -58.504 53.985
553354 94.632 -58.516 53.951
553355 94.663 -58.530 53.920
553356 94.700 -58.542 53.895
553357 94.743 -58.553 53.875
553358 94.785 -58.563 53.857
553359 94.828 -58.574 53.839
553360 94.873 -58.586 53.822
553361 94.920 -58.598 53.805
553362 94.967 -58.611 53.782
553363 95.014 -58.623 53.756
553364 95.060 -58.636 53.726
553365 95.108 -58.648 53.694
553366 95.162 -58.657 53.663
553367 95.215 -58.667 53.634
553368 95.268 -58.676 53.607
553369 95.321 -58.686 53.582
553370 95.375 -58.696 53.557
553371 95.416 -58.701 53.533
553372 95.444 -58.701 53.511
553373 95.473 -58.708 53.486
553374 95.505 -58.722 53.460
553375 95.536 -58.736 53.433
553376 95.567 -58.749 53.403
553377 95.599 -58.762 53.371
553378 95.631 -58.776 53.341
553379 95.661 -58.787 53.311
553380 95.691 -58.795 53.284
553381 95.721 -58.805 53.257
553382 95.751 -58.814 53.229
553383 95.781 -58.823 53.200
553384 95.809 -58.830 53.170
553385 95.838 -58.838 53.142
553386 95.869 -58.845 53.112
553387 95.904 -58.851 53.080
553388 95.936 -58.857 53.049
553389 95.966 -58.864 53.018
553390 95.996 -58.871 52.988
553391 96.027 -58.878 52.958
553392 96.057 -58.885 52.934
553393 96.087 -58.892 52.913
553394 96.119 -58.899 52.892
553395 96.154 -58.905 52.867
553396 96.189 -58.911 52.842
553397 96.224 -58.918 52.814
553398 96.259 -58.924 52.785
553399 96.294 -58.930 52.759
553400 96.329 -58.936 52.734
553401 96.364 -58.943 52.713
553402 96.399 -58.949 52.687
553403 96.434 -58.953 52.657
553404 96.465 -58.955 52.625
553405 96.497 -58.955 52.596
553406 96.528 -58.956 52.563
553407 96.559 -58.957 52.529
553408 96.590 -58.958 52.501
553409 96.621 -58.959 52.475
553410 96.652 -58.960 52.445
553411 96.683 -58.961 52.417
553412 96.714 -58.961 52.384
553413 96.745 -58.962 52.353
553414 96.777 -58.968 52.321
553415 96.809 -58.973 52.286
553416 96.841 -58.976 52.248
553417 96.875 -58.977 52.206
553418 96.909 -58.977 52.166
553419 96.943 -58.978 52.126
553420 96.976 -58.979 52.086
553421 97.010 -58.980 52.047
553422 97.044 -58.980 52.011
553423 97.077 -58.981 51.979
553424 97.111 -58.985 51.947
553425 97.144 -58.989 51.913
553426 97.177 -58.993 51.875
553427 97.212 -58.996 51.834
553428 97.247 -58.999 51.796
553429 97.282 -59.002 51.755
553430 97.317 -59.005 51.715
553431 97.350 -59.009 51.673
553432 97.383 -59.014 51.628
553433 97.416 -59.018 51.581
553434 97.449 -59.022 51.542
553435 97.483 -59.026 51.499
553436 97.516 -59.030 51.453
553437 97.549 -59.034 51.410
553438 97.582 -59.039 51.368
553439 97.617 -59.048 51.323
553440 97.651 -59.054 51.280
553441 97.682 -59.059 51.239
553442 97.713 -59.063 51.195
553443 97.744 -59.068 51.151
553444 97.780 -59.070 51.110
553445 97.819 -59.068 51.068
553446 97.859 -59.066 51.025
553447 97.898 -59.065 50.983
553448 97.938 -59.063 50.940
553449 97.976 -59.063 50.896
553450 98.015 -59.062 50.853
553451 98.060 -59.049 50.810
553452 98.096 -59.046 50.767
553453 98.128 -59.049 50.724
553454 98.160 -59.052 50.684
553455 98.193 -59.055 50.646
553456 98.228 -59.059 50.615
553457 98.267 -59.062 50.590
553458 98.298 -59.058 50.570
553459 98.332 -59.055 50.551
553460 98.368 -59.053 50.527
553461 98.402 -59.043 50.500
553462 98.436 -59.031 50.472
553463 98.470 -59.021 50.444
553464 98.504 -59.010 50.416
553465 98.538 -58.999 50.388
553466 98.572 -58.989 50.360
553467 98.606 -58.978 50.332
553468 98.640 -58.967 50.306
553469 98.674 -58.956 50.278
553470 98.708 -58.946 50.249
553471 98.742 -58.935 50.222
553472 98.775 -58.925 50.195
553473 98.805 -58.922 50.165
553474 98.839 -58.919 50.137
553475 98.872 -58.918 50.106
553476 98.910 -58.924 50.075
553477 98.948 -58.930 50.044
553478 98.987 -58.937 50.010
553479 99.026 -58.944 49.974
553480 99.073 -58.969 49.937
553481 99.101 -58.974 49.901
553482 99.130 -58.977 49.866
553483 99.178 -58.977 49.831
553484 99.250 -58.971 49.798
553485 99.322 -58.965 49.766
553486 99.394 -58.959 49.735
553487 99.442 -58.962 49.704
553488 99.478 -58.970 49.672
553489 99.513 -58.977 49.639
553490 99.549 -58.984 49.608
553491 99.585 -58.992 49.576
553492 99.621 -58.999 49.544
553493 99.656 -59.007 49.512
553494 99.692 -59.015 49.478
553495 99.727 -59.022 49.443
553496 99.763 -59.030 49.407
553497 99.799 -59.037 49.372
553839 86.632 -55.576 57.648
553840 86.672 -55.589 57.643
553841 86.713 -55.602 57.635
553842 86.753 -55.616 57.628
553843 86.793 -55.629 57.621
553844 86.834 -55.643 57.612
553845 86.874 -55.656 57.599
553846 86.915 -55.669 57.586
553847 86.955 -55.683 57.574
553848 86.995 -55.696 57.559
553849 87.036 -55.709 57.544
553850 87.076 -55.723 57.528
553851 87.117 -55.736 57.506
553852 87.157 -55.749 57.483
553853 87.197 -55.762 57.458
553854 87.237 -55.776 57.433
553855 87.278 -55.790 57.407
553856 87.319 -55.803 57.376
553857 87.359 -55.816 57.342
553858 87.399 -55.829 57.307
553859 87.439 -55.843 57.271
553860 87.480 -55.856 57.234
553861 87.526 -55.866 57.196
553862 87.580 -55.870 57.158
553863 87.635 -55.875 57.117
553864 87.691 -55.878 57.075
553865 87.737 -55.885 57.034
553866 87.781 -55.893 56.994
553867 87.815 -55.904 56.954
553868 87.848 -55.915 56.916
553869 87.882 -55.927 56.878
553899 89.026 -56.402 56.059
553900 89.064 -56.418 56.041
553901 89.102 -56.434 56.025
553902 89.140 -56.451 56.010
553903 89.178 -56.467 55.995
553904 89.216 -56.483 55.982
553905 89.255 -56.498 55.970
553906 89.293 -56.514 55.961
553907 89.331 -56.530 55.954
553908 89.370 -56.546 55.948
553909 89.408 -56.562 55.942
553910 89.446 -56.578 55.935
553911 89.483 -56.595 55.929
553912 89.507 -56.619 55.928
553913 89.541 -56.636 55.927
553914 89.582 -56.647 55.929
553915 89.623 -56.658 55.932
553916 89.665 -56.669 55.936
553917 89.706 -56.680 55.940
553918 89.748 -56.690 55.946
553919 89.789 -56.701 55.952
553920 89.830 -56.712 55.957
553921 89.872 -56.723 55.962
553922 89.913 -56.733 55.964
553923 89.954 -56.745 55.964
553924 89.979 -56.762 55.963
553925 90.005 -56.779 55.958
553926 90.030 -56.796 55.949
553927 90.056 -56.813 55.934
553928 90.088 -56.829 55.917
553929 90.125 -56.843 55.895
553930 90.162 -56.858 55.872
553931 90.199 -56.872 55.846
553932 90.237 -56.887 55.816
553933 90.278 -56.902 55.785
553934 90.324 -56.918 55.751
553935 90.371 -56.934 55.716
553936 90.417 -56.950 55.683
553937 90.463 -56.966 55.650
553938 90.510 -56.982 55.618
553939 90.556 -56.998 55.591
553940 90.591 -57.015 55.557
553941 90.626 -57.031 55.519
553942 90.661 -57.047 55.482
553943 90.696 -57.063 55.452
553944 90.731 -57.079 55.426
553945 90.766 -57.095 55.404
553946 90.801 -57.112 55.382
553947 90.836 -57.128 55.362
553948 90.871 -57.144 55.344
553949 90.906 -57.160 55.328
553950 90.941 -57.176 55.315
553951 90.976 -57.193 55.304
553952 91.013 -57.209 55.294
553953 91.050 -57.225 55.285
553954 91.088 -57.241 55.277
553955 91.125 -57.258 55.271
553956 91.163 -57.275 55.269
553957 91.202 -57.288 55.270
553958 91.245 -57.299 55.274
553959 91.285 -57.323 55.282
553960 91.328 -57.336 55.293
553961 91.372 -57.347 55.300
553962 91.416 -57.357 55.302
553963 91.459 -57.367 55.301
553964 91.503 -57.378 55.296
553965 91.547 -57.388 55.289
553966 91.591 -57.399 55.282
553967 91.635 -57.410 55.276
553968 91.678 -57.420 55.268
553969 91.722 -57.430 55.259
553970 91.766 -57.441 55.254
553971 91.809 -57.451 55.259
553972 91.852 -57.460 55.264
553973 91.895 -57.470 55.269
553974 91.938 -57.480 55.272
553975 91.981 -57.489 55.272
553976 92.024 -57.499 55.271
553977 92.067 -57.509 55.276
553978 92.110 -57.518 55.278
553979 92.153 -57.528 55.278
553980 92.196 -57.539 55.276
553981 92.236 -57.563 55.270
553982 92.276 -57.588 55.262
553983 92.311 -57.614 55.247
553984 92.341 -57.634 55.227
553985 92.370 -57.655 55.207
553986 92.400 -57.676 55.183
553987 92.430 -57.697 55.158
553988 92.459 -57.717 55.136
553989 92.489 -57.739 55.117
553990 92.518 -57.759 55.099
553991 92.548 -57.780 55.083
553992 92.578 -57.801 55.068
553993 92.607 -57.822 55.049
553994 92.644 -57.833 55.027
553995 92.680 -57.843 55.004
553996 92.716 -57.854 54.981
553997 92.752 -57.865 54.956
553998 92.788 -57.874 54.930
553999 92.825 -57.880 54.902
554000 92.861 -57.887 54.879
554001 92.898 -57.893 54.859
554002 92.934 -57.899 54.839
554003 92.959 -57.905 54.819
554004 92.984 -57.911 54.800
554005 93.009 -57.915 54.786
554006 93.041 -57.915 54.779
554007 93.072 -57.914 54.773
554008 93.103 -57.913 54.767
554009 93.135 -57.912 54.761
554010 93.167 -57.913 54.756
554011 93.201 -57.921 54.750
554012 93.236 -57.930 54.736
554013 93.270 -57.938 54.725
554014 93.304 -57.946 54.726
554015 93.339 -57.955 54.736
554016 93.373 -57.963 54.745
554017 93.408 -57.971 54.745
554018 93.442 -57.980 54.754
554019 93.477 -57.988 54.761
554020 93.511 -57.996 54.762
554021 93.546 -58.005 54.757
554022 93.580 -58.013 54.749
554023 93.618 -58.031 54.749
554024 93.656 -58.051 54.751
554025 93.695 -58.071 54.751
554026 93.733 -58.090 54.748
554027 93.771 -58.111 54.740
554028 93.809 -58.131 54.729
554029 93.848 -58.150 54.718
554030 93.886 -58.170 54.708
554031 93.924 -58.190 54.699
554032 93.956 -58.205 54.687
554033 93.982 -58.216 54.671
554034 94.009 -58.228 54.648
554035 94.036 -58.240 54.620
554036 94.062 -58.252 54.587
554037 94.089 -58.263 54.549
554038 94.116 -58.275 54.508
554039 94.142 -58.288 54.464
554040 94.169 -58.299 54.421
554041 94.196 -58.311 54.378
554042 94.222 -58.322 54.338
554043 94.250 -58.335 54.299
554044 94.284 -58.346 54.264
554045 94.319 -58.357 54.233
554046 94.353 -58.368 54.201
554047 94.388 -58.379 54.173
554048 94.418 -58.392 54.147
554049 94.449 -58.406 54.121
554050 94.480 -58.419 54.095
554051 94.510 -58.432 54.068
554052 94.541 -58.445 54.035
554053 94.572 -58.459 54.003
554054 94.603 -58.472 53.972
554055 94.633 -58.485 53.941
554056 94.664 -58.498 53.911
554057 94.696 -58.511 53.884
554058 94.739 -58.522 53.862
554059 94.781 -58.533 53.843
554060 94.826 -58.544 53.825
554061 94.873 -58.557 53.804
554062 94.919 -58.569 53.784
554063 94.967 -58.581 53.761
554064 95.016 -58.592 53.733
554065 95.069 -58.602 53.702
554066 95.123 -58.612 53.671
554067 95.176 -58.621 53.641
554068 95.229 -58.631 53.612
554069 95.282 -58.641 53.585
554070 95.336 -58.650 53.560
554071 95.389 -58.660 53.536
554072 95.418 -58.661 53.512
554073 95.446 -58.661 53.488
554074 95.474 -58.666 53.465
554075 95.506 -58.679 53.441
554076 95.537 -58.693 53.413
554077 95.569 -58.707 53.381
554078 95.600 -58.720 53.351
554079 95.632 -58.733 53.325
554080 95.664 -58.747 53.298
554081 95.695 -58.761 53.272
554082 95.726 -58.772 53.246
554083 95.755 -58.780 53.221
554084 95.785 -58.789 53.194
554085 95.815 -58.798 53.164
554086 95.843 -58.805 53.132
554087 95.875 -58.812 53.100
554088 95.909 -58.818 53.068
554089 95.939 -58.826 53.039
554090 95.969 -58.833 53.008
554091 96.000 -58.839 52.972
554092 96.030 -58.846 52.940
554093 96.060 -58.853 52.912
554094 96.090 -58.860 52.887
554095 96.123 -58.867 52.859
554096 96.158 -58.874 52.834
554097 96.192 -58.880 52.810
554098 96.228 -58.886 52.782
554099 96.262 -58.892 52.752
554100 96.298 -58.899 52.723
554101 96.332 -58.903 52.697
554102 96.367 -58.906 52.677
554103 96.402 -58.909 52.656
554104 96.437 -58.912 52.630
554105 96.471 -58.914 52.601
554106 96.502 -58.915 52.570
554107 96.533 -58.917 52.533
554108 96.564 -58.918 52.499
554109 96.595 -58.918 52.469
554110 96.626 -58.919 52.438
554111 96.657 -58.920 52.405
554112 96.688 -58.921 52.375
554113 96.719 -58.922 52.343
554114 96.750 -58.924 52.312
554115 96.783 -58.930 52.278
554116 96.814 -58.935 52.241
554117 96.847 -58.940 52.202
554118 96.880 -58.941 52.161
554119 96.914 -58.942 52.124
554120 96.947 -58.943 52.086
554121 96.981 -58.943 52.047
554122 97.015 -58.944 52.009
554123 97.048 -58.944 51.973
554124 97.082 -58.946 51.938
554125 97.115 -58.950 51.907
554126 97.149 -58.954 51.874
554127 97.182 -58.958 51.838
554128 97.216 -58.961 51.800
554129 97.251 -58.965 51.761
554130 97.286 -58.968 51.720
554131 97.320 -58.971 51.680
554132 97.353 -58.975 51.639
554133 97.386 -58.980 51.596
554134 97.419 -58.984 51.551
554135 97.452 -58.988 51.510
554136 97.485 -58.993 51.470
554137 97.518 -58.996 51.426
554138 97.551 -59.001 51.383
554139 97.586 -59.006 51.342
554140 97.621 -59.015 51.297
554141 97.656 -59.023 51.252
554142 97.687 -59.028 51.209
554143 97.719 -59.033 51.164
554144 97.750 -59.037 51.120
554145 97.788 -59.037 51.078
554146 97.827 -59.035 51.037
554147 97.867 -59.034 50.994
554148 97.906 -59.033 50.952
554149 97.944 -59.032 50.909
554150 97.983 -59.032 50.865
554151 98.025 -59.024 50.821
554152 98.070 -59.012 50.777
554153 98.107 -59.008 50.734
554154 98.139 -59.012 50.693
554155 98.171 -59.014 50.654
554156 98.203 -59.017 50.618
554157 98.235 -59.020 50.588
554158 98.268 -59.022 50.564
554159 98.302 -59.021 50.543
554160 98.338 -59.020 50.521
554161 98.374 -59.018 50.497
554162 98.409 -59.010 50.471
554163 98.443 -58.999 50.444
554164 98.477 -58.989 50.417
554165 98.511 -58.978 50.392
554166 98.545 -58.968 50.367
554167 98.579 -58.956 50.342
554168 98.613 -58.946 50.316
554169 98.647 -58.935 50.290
554170 98.680 -58.924 50.264
554171 98.715 -58.914 50.238
554172 98.748 -58.903 50.213
554173 98.777 -58.901 50.188
554174 98.808 -58.898 50.162
554175 98.843 -58.894 50.136
554176 98.876 -58.893 50.108
554177 98.910 -58.892 50.079
554178 98.945 -58.896 50.048
554179 98.984 -58.903 50.016
554180 99.023 -58.910 49.980
554181 99.071 -58.932 49.943
554182 99.100 -58.936 49.907
554183 99.129 -58.940 49.872
554184 99.186 -58.938 49.837
554185 99.258 -58.932 49.805
554186 99.329 -58.926 49.773
554187 99.401 -58.920 49.742
554188 99.443 -58.925 49.710
554189 99.479 -58.933 49.679
554190 99.514 -58.940 49.647
554191 99.550 -58.948 49.615
554192 99.586 -58.955 49.584
554193 99.622 -58.963 49.553
554194 99.657 -58.971 49.521
554195 99.693 -58.978 49.488
554196 99.728 -58.985 49.453
554197 99.764 -58.993 49.417
554198 99.800 -59.000 49.382
554541 86.690 -55.546 57.643
554542 86.729 -55.559 57.636
554543 86.770 -55.573 57.629
554544 86.810 -55.586 57.622
554545 86.851 -55.599 57.613
554546 86.891 -55.612 57.602
554547 86.931 -55.626 57.591
554548 86.972 -55.640 57.578
554549 87.012 -55.652 57.564
554550 87.053 -55.666 57.551
554551 87.093 -55.679 57.536
554552 87.133 -55.693 57.515
554553 87.174 -55.706 57.494
554554 87.214 -55.719 57.472
554555 87.255 -55.733 57.450
554556 87.295 -55.746 57.427
554557 87.335 -55.759 57.396
554558 87.375 -55.773 57.362
554559 87.416 -55.786 57.328
554560 87.457 -55.799 57.293
554561 87.497 -55.813 57.257
554562 87.551 -55.818 57.220
554563 87.606 -55.821 57.181
554564 87.660 -55.825 57.140
554565 87.712 -55.831 57.099
554566 87.757 -55.838 57.057
554567 87.793 -55.849 57.016
554568 87.827 -55.860 56.976
554569 87.860 -55.872 56.937
554570 87.896 -55.884 56.898
554571 87.934 -55.900 56.858
554600 89.043 -56.364 56.055
554601 89.082 -56.380 56.037
554602 89.120 -56.396 56.021
554603 89.158 -56.412 56.005
554604 89.196 -56.428 55.991
554605 89.235 -56.444 55.978
554606 89.273 -56.460 55.966
554607 89.311 -56.476 55.956
554608 89.349 -56.492 55.948
554609 89.388 -56.508 55.942
554610 89.426 -56.523 55.936
554611 89.464 -56.539 55.930
554612 89.502 -56.556 55.925
554613 89.526 -56.580 55.926
554614 89.560 -56.596 55.926
554615 89.602 -56.608 55.928
554616 89.643 -56.618 55.932
554617 89.684 -56.629 55.938
554618 89.726 -56.640 55.944
554619 89.767 -56.651 55.952
554620 89.809 -56.661 55.960
554621 89.850 -56.672 55.968
554622 89.891 -56.683 55.975
554623 89.933 -56.694 55.980
554624 89.961 -56.710 55.983
554625 89.986 -56.727 55.985
554626 90.012 -56.744 55.983
554627 90.038 -56.762 55.977
554628 90.063 -56.778 55.965
554629 90.099 -56.793 55.955
554630 90.136 -56.808 55.934
554631 90.174 -56.822 55.904
554632 90.211 -56.836 55.870
554633 90.251 -56.851 55.837
554634 90.298 -56.868 55.802
554635 90.344 -56.884 55.764
554636 90.390 -56.899 55.726
554637 90.437 -56.915 55.689
554638 90.483 -56.932 55.653
554639 90.529 -56.948 55.621
554640 90.573 -56.960 55.592
554641 90.613 -56.975 55.557
554642 90.648 -56.991 55.519
554643 90.683 -57.009 55.483
554644 90.719 -57.026 55.454
554645 90.754 -57.044 55.429
554646 90.790 -57.061 55.407
554647 90.825 -57.078 55.386
554648 90.860 -57.094 55.365
554649 90.895 -57.109 55.347
554650 90.930 -57.126 55.332
554651 90.965 -57.142 55.319
554652 91.000 -57.159 55.308
554653 91.034 -57.175 55.298
554654 91.070 -57.191 55.289
554655 91.105 -57.207 55.281
554656 91.142 -57.223 55.276
554657 91.179 -57.240 55.273
554658 91.217 -57.257 55.273
554659 91.259 -57.269 55.274
554660 91.301 -57.286 55.276
554661 91.345 -57.297 55.279
554662 91.389 -57.307 55.282
554663 91.433 -57.318 55.284
554664 91.476 -57.329 55.281
554665 91.520 -57.339 55.276
554666 91.564 -57.349 55.271
554667 91.607 -57.359 55.267
554668 91.650 -57.369 55.266
554669 91.693 -57.378 55.261
554670 91.735 -57.388 55.256
554671 91.779 -57.398 55.253
554672 91.821 -57.407 55.258
554673 91.864 -57.416 55.262
554674 91.907 -57.426 55.267
554675 91.950 -57.435 55.270
554676 91.993 -57.445 55.271
554677 92.036 -57.455 55.271
554678 92.079 -57.464 55.273
554679 92.122 -57.474 55.279
554680 92.165 -57.484 55.284
554681 92.206 -57.505 55.286
554682 92.247 -57.528 55.281
554683 92.287 -57.553 55.270
554684 92.323 -57.579 55.251
554685 92.358 -57.605 55.227
554686 92.387 -57.625 55.201
554687 92.417 -57.646 55.175
554688 92.447 -57.667 55.150
554689 92.477 -57.688 55.127
554690 92.506 -57.708 55.109
554691 92.536 -57.730 55.094
554692 92.565 -57.750 55.076
554693 92.595 -57.771 55.057
554694 92.628 -57.788 55.040
554695 92.663 -57.799 55.022
554696 92.700 -57.809 55.003
554697 92.736 -57.820 54.982
554698 92.772 -57.830 54.958
554699 92.808 -57.840 54.929
554700 92.845 -57.846 54.899
554701 92.881 -57.852 54.878
554702 92.918 -57.858 54.859
554703 92.947 -57.865 54.842
554704 92.972 -57.871 54.825
554705 92.996 -57.877 54.810
554706 93.021 -57.881 54.799
554707 93.053 -57.880 54.797
554708 93.084 -57.880 54.797
554709 93.116 -57.880 54.797
554710 93.147 -57.879 54.796
554711 93.178 -57.878 54.795
554712 93.210 -57.877 54.788
554713 93.244 -57.886 54.776
554714 93.279 -57.894 54.758
554715 93.313 -57.902 54.743
554716 93.347 -57.911 54.738
554717 93.382 -57.919 54.745
554718 93.416 -57.927 54.753
554719 93.451 -57.936 54.758
554720 93.485 -57.944 54.765
554721 93.520 -57.952 54.771
554722 93.554 -57.960 54.770
554723 93.589 -57.969 54.761
554724 93.627 -57.989 54.749
554725 93.665 -58.009 54.745
554726 93.704 -58.028 54.745
554727 93.742 -58.048 54.741
554728 93.780 -58.068 54.734
554729 93.819 -58.088 54.720
554730 93.857 -58.108 54.710
554731 93.895 -58.128 54.700
554732 93.933 -58.147 54.689
554733 93.972 -58.167 54.676
554734 94.003 -58.183 54.661
554735 94.030 -58.194 54.640
554736 94.057 -58.206 54.617
554737 94.084 -58.218 54.588
554738 94.111 -58.230 54.556
554739 94.137 -58.242 54.521
554740 94.164 -58.253 54.480
554741 94.191 -58.266 54.437
554742 94.217 -58.277 54.393
554743 94.244 -58.289 54.348
554744 94.271 -58.300 54.306
554745 94.297 -58.313 54.267
554746 94.324 -58.325 54.234
554747 94.355 -58.336 54.203
554748 94.388 -58.347 54.173
554749 94.419 -58.360 54.143
554750 94.450 -58.374 54.113
554751 94.480 -58.387 54.087
554752 94.511 -58.400 54.060
554753 94.542 -58.413 54.030
554754 94.572 -58.427 53.996
554755 94.603 -58.440 53.962
554756 94.634 -58.454 53.932
554757 94.665 -58.466 53.905
554758 94.695 -58.480 53.877
554759 94.736 -58.491 53.851
554760 94.779 -58.502 53.829
554761 94.826 -58.514 53.809
554762 94.873 -58.527 53.786
554763 94.924 -58.537 53.763
554764 94.977 -58.547 53.735
554765 95.031 -58.557 53.707
554766 95.084 -58.567 53.677
554767 95.137 -58.576 53.647
554768 95.190 -58.585 53.617
554769 95.244 -58.595 53.592
554770 95.297 -58.604 53.568
554771 95.350 -58.614 53.544
554772 95.392 -58.620 53.519
554773 95.420 -58.620 53.494
554774 95.447 -58.620 53.466
554775 95.476 -58.624 53.443
554776 95.507 -58.637 53.420
554777 95.539 -58.651 53.392
554778 95.570 -58.664 53.360
554779 95.602 -58.678 53.334
554780 95.633 -58.691 53.311
554781 95.665 -58.705 53.287
554782 95.696 -58.718 53.262
554783 95.728 -58.732 53.239
554784 95.759 -58.745 53.215
554785 95.790 -58.757 53.188
554786 95.820 -58.765 53.158
554787 95.849 -58.773 53.125
554788 95.881 -58.780 53.092
554789 95.912 -58.787 53.060
554790 95.942 -58.794 53.032
554791 95.972 -58.801 52.998
554792 96.003 -58.808 52.962
554793 96.033 -58.814 52.930
554794 96.063 -58.822 52.901
554795 96.093 -58.829 52.872
554796 96.126 -58.835 52.843
554797 96.161 -58.842 52.818
554798 96.196 -58.848 52.789
554799 96.231 -58.852 52.760
554800 96.266 -58.855 52.732
554801 96.301 -58.858 52.707
554802 96.335 -58.861 52.679
554803 96.370 -58.864 52.651
554804 96.404 -58.867 52.624
554805 96.439 -58.871 52.596
554806 96.474 -58.874 52.568
554807 96.507 -58.876 52.537
554808 96.538 -58.877 52.504
554809 96.569 -58.877 52.471
554810 96.600 -58.879 52.436
554811 96.631 -58.880 52.399
554812 96.662 -58.880 52.363
554813 96.693 -58.881 52.331
554814 96.724 -58.882 52.302
554815 96.756 -58.886 52.269
554816 96.788 -58.892 52.233
554817 96.820 -58.897 52.197
554818 96.852 -58.902 52.159
554819 96.885 -58.905 52.121
554820 96.919 -58.906 52.083
554821 96.952 -58.907 52.045
554822 96.986 -58.908 52.007
554823 97.020 -58.908 51.971
554824 97.053 -58.908 51.934
554825 97.087 -58.911 51.898
554826 97.120 -58.915 51.864
554827 97.153 -58.920 51.832
554828 97.186 -58.924 51.800
554829 97.220 -58.927 51.766
554830 97.254 -58.930 51.728
554831 97.289 -58.934 51.687
554832 97.324 -58.937 51.644
554833 97.357 -58.942 51.602
554834 97.389 -58.946 51.561
554835 97.422 -58.950 51.519
554836 97.455 -58.954 51.474
554837 97.488 -58.958 51.434
554838 97.521 -58.962 51.394
554839 97.555 -58.967 51.354
554840 97.589 -58.974 51.312
554841 97.624 -58.982 51.268
554842 97.659 -58.990 51.221
554843 97.693 -58.997 51.175
554844 97.725 -59.002 51.129
554845 97.757 -59.006 51.084
554846 97.796 -59.004 51.042
554847 97.836 -59.002 51.002
554848 97.875 -59.002 50.958
554849 97.913 -59.002 50.918
554850 97.951 -59.002 50.877
554851 97.992 -58.998 50.833
554852 98.036 -58.986 50.789
554853 98.080 -58.974 50.743
554854 98.118 -58.971 50.701
554855 98.150 -58.974 50.661
554856 98.182 -58.976 50.624
554857 98.214 -58.979 50.590
554858 98.246 -58.981 50.561
554859 98.277 -58.984 50.537
554860 98.308 -58.987 50.514
554861 98.344 -58.985 50.490
554862 98.380 -58.983 50.465
554863 98.416 -58.978 50.440
554864 98.450 -58.968 50.415
554865 98.484 -58.957 50.390
554866 98.518 -58.946 50.367
554867 98.552 -58.935 50.344
554868 98.585 -58.924 50.322
554869 98.620 -58.914 50.299
554870 98.653 -58.903 50.274
554871 98.688 -58.892 50.250
554872 98.721 -58.882 50.227
554873 98.750 -58.879 50.204
554874 98.779 -58.877 50.181
554875 98.812 -58.874 50.158
554876 98.847 -58.870 50.135
554877 98.880 -58.867 50.109
554878 98.914 -58.867 50.080
554879 98.946 -58.867 50.050
554880 98.982 -58.868 50.018
554881 99.020 -58.875 49.985
554882 99.060 -58.883 49.948
554883 99.098 -58.899 49.912
554884 99.127 -58.903 49.877
554885 99.193 -58.899 49.843
554886 99.265 -58.892 49.810
554887 99.337 -58.886 49.779
554888 99.408 -58.881 49.748
554889 99.444 -58.889 49.717
554890 99.480 -58.896 49.686
554891 99.515 -58.904 49.655
554892 99.551 -58.911 49.623
554893 99.586 -58.918 49.592
554894 99.622 -58.926 49.561
554895 99.658 -58.933 49.529
554896 99.694 -58.941 49.497
554897 99.729 -58.949 49.462
554898 99.765 -58.956 49.427
554899 99.800 -58.964 49.393
555242 86.720 -55.495 57.643
555243 86.755 -55.511 57.636
555244 86.790 -55.527 57.630
555245 86.827 -55.542 57.623
555246 86.867 -55.556 57.615
555247 86.908 -55.569 57.605
555248 86.948 -55.583 57.594
555249 86.989 -55.596 57.582
555250 87.029 -55.609 57.570
555251 87.069 -55.623 57.557
555252 87.110 -55.636 57.543
555253 87.150 -55.649 57.524
555254 87.191 -55.663 57.505
555255 87.231 -55.676 57.484
555256 87.271 -55.690 57.463
555257 87.311 -55.702 57.441
555258 87.352 -55.716 57.414
555259 87.393 -55.730 57.383
555260 87.433 -55.743 57.349
555261 87.473 -55.756 57.315
555262 87.521 -55.765 57.280
555263 87.576 -55.769 57.243
555264 87.631 -55.773 57.205
555265 87.686 -55.777 57.164
555266 87.732 -55.784 57.123
555267 87.772 -55.793 57.082
555268 87.805 -55.804 57.039
555269 87.839 -55.816 57.000
555270 87.875 -55.830 56.959
555271 87.913 -55.846 56.919
555272 87.952 -55.862 56.878
555273 87.990 -55.878 56.839
555301 89.061 -56.326 56.052
555302 89.099 -56.342 56.033
555303 89.137 -56.357 56.016
555304 89.176 -56.373 56.001
555305 89.214 -56.389 55.987
555306 89.252 -56.405 55.973
555307 89.291 -56.421 55.962
555308 89.329 -56.437 55.952
555309 89.367 -56.453 55.944
555310 89.405 -56.469 55.937
555311 89.444 -56.485 55.932
555312 89.482 -56.501 55.927
555313 89.520 -56.517 55.923
555314 89.544 -56.541 55.923
555315 89.580 -56.557 55.925
555316 89.621 -56.567 55.928
555317 89.663 -56.579 55.933
555318 89.704 -56.589 55.940
555319 89.745 -56.600 55.948
555320 89.787 -56.611 55.957
555321 89.828 -56.622 55.967
555322 89.870 -56.633 55.977
555323 89.911 -56.643 55.987
555324 89.943 -56.658 55.996
555325 89.968 -56.675 56.002
555326 89.994 -56.693 56.007
555327 90.020 -56.709 56.010
555328 90.045 -56.727 56.009
555329 90.074 -56.743 56.005
555330 90.111 -56.758 56.001
555331 90.148 -56.771 55.978
555332 90.185 -56.786 55.942
555333 90.225 -56.801 55.901
555334 90.271 -56.817 55.860
555335 90.317 -56.833 55.820
555336 90.364 -56.849 55.778
555337 90.410 -56.865 55.737
555338 90.456 -56.881 55.697
555339 90.501 -56.895 55.659
555340 90.544 -56.907 55.623
555341 90.587 -56.919 55.588
555342 90.630 -56.931 55.553
555343 90.669 -56.947 55.518
555344 90.704 -56.965 55.485
555345 90.739 -56.982 55.457
555346 90.775 -56.999 55.434
555347 90.810 -57.016 55.412
555348 90.845 -57.034 55.390
555349 90.881 -57.051 55.370
555350 90.916 -57.069 55.352
555351 90.951 -57.086 55.337
555352 90.987 -57.103 55.325
555353 91.022 -57.120 55.313
555354 91.057 -57.138 55.303
555355 91.093 -57.155 55.294
555356 91.128 -57.172 55.286
555357 91.163 -57.189 55.281
555358 91.198 -57.205 55.277
555359 91.234 -57.222 55.276
555360 91.274 -57.237 55.276
555361 91.318 -57.247 55.275
555362 91.361 -57.257 55.275
555363 91.404 -57.267 55.275
555364 91.447 -57.276 55.276
555365 91.490 -57.286 55.275
555366 91.533 -57.296 55.273
555367 91.576 -57.305 55.272
555368 91.618 -57.315 55.272
555369 91.662 -57.325 55.272
555370 91.705 -57.334 55.269
555371 91.747 -57.344 55.263
555372 91.790 -57.354 55.260
555373 91.834 -57.363 55.261
555374 91.876 -57.373 55.267
555375 91.919 -57.382 55.271
555376 91.962 -57.391 55.275
555377 92.005 -57.401 55.278
555378 92.048 -57.411 55.279
555379 92.091 -57.420 55.280
555380 92.134 -57.430 55.285
555381 92.176 -57.447 55.292
555382 92.217 -57.470 55.295
555383 92.258 -57.494 55.292
555384 92.298 -57.517 55.281
555385 92.335 -57.545 55.261
555386 92.371 -57.571 55.238
555387 92.405 -57.595 55.211
555388 92.434 -57.616 55.186
555389 92.464 -57.637 55.162
555390 92.494 -57.658 55.137
555391 92.523 -57.679 55.115
555392 92.553 -57.699 55.096
555393 92.583 -57.720 55.075
555394 92.613 -57.741 55.054
555395 92.647 -57.753 55.033
555396 92.684 -57.764 55.012
555397 92.720 -57.774 54.990
555398 92.756 -57.785 54.967
555399 92.792 -57.796 54.943
555400 92.828 -57.806 54.918
555401 92.864 -57.813 54.894
555402 92.901 -57.819 54.875
555403 92.934 -57.825 54.858
555404 92.959 -57.831 54.844
555405 92.983 -57.837 54.834
555406 93.007 -57.843 54.827
555407 93.034 -57.848 54.824
555408 93.065 -57.847 54.826
555409 93.096 -57.846 54.829
555410 93.127 -57.846 54.830
555411 93.159 -57.845 54.829
555412 93.190 -57.844 54.828
555413 93.221 -57.843 54.823
555414 93.253 -57.843 54.811
555415 93.287 -57.850 54.790
555416 93.322 -57.858 54.767
555417 93.356 -57.867 54.749
555418 93.390 -57.875 54.744
555419 93.425 -57.883 54.754
555420 93.459 -57.891 54.760
555421 93.494 -57.899 54.765
555422 93.528 -57.908 54.769
555423 93.562 -57.916 54.771
555424 93.597 -57.927 54.764
555425 93.636 -57.946 54.752
555426 93.674 -57.966 54.741
555427 93.712 -57.986 54.734
555428 93.751 -58.006 54.728
555429 93.789 -58.025 54.718
555430 93.827 -58.045 54.707
555431 93.865 -58.065 54.697
555432 93.904 -58.085 54.686
555433 93.942 -58.105 54.673
555434 93.980 -58.125 54.659
555435 94.018 -58.144 54.644
555436 94.052 -58.161 54.626
555437 94.078 -58.172 54.606
555438 94.105 -58.184 54.582
555439 94.132 -58.196 54.556
555440 94.158 -58.208 54.528
555441 94.186 -58.220 54.492
555442 94.212 -58.231 54.452
555443 94.239 -58.244 54.407
555444 94.266 -58.255 54.361
555445 94.292 -58.267 54.315
555446 94.319 -58.278 54.272
555447 94.346 -58.291 54.238
555448 94.372 -58.303 54.207
555449 94.400 -58.315 54.176
555450 94.429 -58.328 54.143
555451 94.458 -58.342 54.110
555452 94.486 -58.355 54.080
555453 94.515 -58.369 54.053
555454 94.544 -58.382 54.024
555455 94.573 -58.395 53.991
555456 94.604 -58.409 53.955
555457 94.635 -58.422 53.924
555458 94.666 -58.435 53.897
555459 94.696 -58.448 53.869
555460 94.732 -58.460 53.840
555461 94.779 -58.473 53.812
555462 94.832 -58.482 53.790
555463 94.885 -58.491 53.766
555464 94.938 -58.501 53.740
555465 94.992 -58.511 53.713
555466 95.045 -58.520 53.685
555467 95.098 -58.530 53.656
555468 95.151 -58.540 53.624
555469 95.205 -58.549 53.599
555470 95.258 -58.559 53.575
555471 95.312 -58.569 53.551
555472 95.365 -58.579 53.527
555473 95.394 -58.579 53.502
555474 95.421 -58.580 53.478
555475 95.449 -58.580 53.451
555476 95.477 -58.582 53.422
555477 95.509 -58.595 53.396
555478 95.540 -58.608 53.372
555479 95.571 -58.622 53.349
555480 95.603 -58.636 53.326
555481 95.634 -58.649 53.300
555482 95.666 -58.662 53.277
555483 95.697 -58.676 53.256
555484 95.729 -58.689 53.234
555485 95.760 -58.703 53.212
555486 95.792 -58.716 53.182
555487 95.823 -58.730 53.153
555488 95.854 -58.741 53.121
555489 95.885 -58.748 53.088
555490 95.915 -58.755 53.056
555491 95.945 -58.762 53.029
555492 95.975 -58.769 52.990
555493 96.006 -58.776 52.955
555494 96.036 -58.783 52.924
555495 96.066 -58.790 52.894
555496 96.096 -58.797 52.862
555497 96.130 -58.802 52.834
555498 96.165 -58.805 52.805
555499 96.199 -58.808 52.773
555500 96.234 -58.811 52.743
555501 96.268 -58.814 52.718
555502 96.303 -58.817 52.693
555503 96.338 -58.820 52.663
555504 96.373 -58.824 52.630
555505 96.407 -58.827 52.598
555506 96.442 -58.830 52.568
555507 96.477 -58.833 52.539
555508 96.512 -58.836 52.508
555509 96.543 -58.837 52.476
555510 96.574 -58.838 52.443
555511 96.605 -58.839 52.406
555512 96.637 -58.839 52.367
555513 96.667 -58.840 52.331
555514 96.698 -58.842 52.297
555515 96.730 -58.842 52.266
555516 96.761 -58.848 52.231
555517 96.794 -58.854 52.195
555518 96.825 -58.859 52.159
555519 96.858 -58.864 52.123
555520 96.890 -58.870 52.087
555521 96.923 -58.871 52.049
555522 96.957 -58.871 52.009
555523 96.991 -58.871 51.971
555524 97.024 -58.872 51.935
555525 97.058 -58.874 51.901
555526 97.091 -58.877 51.863
555527 97.124 -58.881 51.827
555528 97.158 -58.885 51.793
555529 97.191 -58.889 51.761
555530 97.225 -58.892 51.728
555531 97.258 -58.896 51.692
555532 97.293 -58.900 51.654
555533 97.327 -58.903 51.612
555534 97.360 -58.908 51.567
555535 97.392 -58.911 51.523
555536 97.425 -58.916 51.483
555537 97.457 -58.920 51.442
555538 97.491 -58.924 51.400
555539 97.524 -58.928 51.359
555540 97.557 -58.933 51.318
555541 97.592 -58.940 51.276
555542 97.628 -58.949 51.236
555543 97.663 -58.957 51.193
555544 97.698 -58.966 51.150
555545 97.730 -58.971 51.105
555546 97.765 -58.974 51.060
555547 97.805 -58.971 51.016
555548 97.843 -58.971 50.972
555549 97.881 -58.971 50.929
555550 97.920 -58.971 50.885
555551 97.958 -58.971 50.841
555552 98.002 -58.960 50.798
555553 98.046 -58.948 50.753
555554 98.091 -58.936 50.710
555555 98.128 -58.933 50.671
555556 98.160 -58.935 50.633
555557 98.193 -58.938 50.598
555558 98.224 -58.940 50.565
555559 98.255 -58.943 50.534
555560 98.287 -58.946 50.509
555561 98.318 -58.948 50.485
555562 98.350 -58.949 50.461
555563 98.386 -58.948 50.436
555564 98.422 -58.946 50.412
555565 98.456 -58.936 50.390
555566 98.490 -58.924 50.366
555567 98.524 -58.914 50.341
555568 98.558 -58.903 50.319
555569 98.592 -58.892 50.299
555570 98.626 -58.882 50.279
555571 98.660 -58.871 50.257
555572 98.694 -58.860 50.238
555573 98.723 -58.858 50.218
555574 98.751 -58.855 50.196
555575 98.781 -58.852 50.175
555576 98.816 -58.849 50.155
555577 98.850 -58.845 50.131
555578 98.884 -58.842 50.106
555579 98.918 -58.842 50.078
555580 98.950 -58.841 50.048
555581 98.984 -58.841 50.017
555582 99.017 -58.841 49.985
555583 99.056 -58.848 49.951
555584 99.097 -58.862 49.916
555585 99.129 -58.865 49.881
555586 99.201 -58.859 49.847
555587 99.273 -58.853 49.815
555588 99.345 -58.848 49.784
555589 99.409 -58.845 49.753
555590 99.445 -58.852 49.723
555591 99.481 -58.859 49.692
555592 99.516 -58.867 49.661
555593 99.552 -58.874 49.630
555594 99.587 -58.882 49.600
555595 99.623 -58.889 49.569
555596 99.659 -58.897 49.538
555597 99.695 -58.905 49.505
555598 99.730 -58.912 49.472
555599 99.766 -58.919 49.436
555600 99.801 -58.927 49.402
555944 86.785 -55.461 57.636
555945 86.820 -55.477 57.630
555946 86.855 -55.493 57.624
555947 86.890 -55.510 57.618
555948 86.925 -55.526 57.609
555949 86.965 -55.539 57.598
555950 87.005 -55.552 57.587
555951 87.046 -55.566 57.575
555952 87.086 -55.580 57.563
555953 87.127 -55.593 57.550
555954 87.167 -55.606 57.533
555955 87.207 -55.619 57.515
555956 87.248 -55.633 57.496
555957 87.288 -55.646 57.476
555958 87.329 -55.659 57.455
555959 87.369 -55.673 57.430
555960 87.409 -55.686 57.400
555961 87.450 -55.699 57.369
555962 87.491 -55.712 57.336
555963 87.546 -55.716 57.303
555964 87.601 -55.720 57.266
555965 87.656 -55.724 57.229
555966 87.708 -55.729 57.189
555967 87.750 -55.737 57.148
555968 87.784 -55.749 57.107
555969 87.818 -55.760 57.064
555970 87.855 -55.775 57.023
555971 87.893 -55.791 56.982
555972 87.931 -55.807 56.940
555973 87.970 -55.824 56.899
555974 88.008 -55.840 56.859
556001 89.041 -56.271 56.068
556002 89.079 -56.287 56.049
556003 89.117 -56.303 56.031
556004 89.156 -56.319 56.013
556005 89.194 -56.335 55.996
556006 89.232 -56.351 55.982
556007 89.270 -56.367 55.969
556008 89.308 -56.382 55.957
556009 89.346 -56.398 55.948
556010 89.385 -56.415 55.940
556011 89.423 -56.431 55.934
556012 89.461 -56.447 55.929
556013 89.499 -56.463 55.925
556014 89.538 -56.479 55.921
556015 89.563 -56.502 55.921
556016 89.599 -56.517 55.923
556017 89.640 -56.528 55.926
556018 89.682 -56.539 55.932
556019 89.723 -56.550 55.941
556020 89.765 -56.561 55.950
556021 89.806 -56.571 55.961
556022 89.847 -56.583 55.973
556023 89.889 -56.593 55.986
556024 89.924 -56.606 56.000
556025 89.950 -56.624 56.011
556026 89.976 -56.640 56.021
556027 90.002 -56.658 56.029
556028 90.027 -56.674 56.039
556029 90.052 -56.692 56.050
556030 90.085 -56.707 56.055
556031 90.122 -56.721 56.056
556032 90.159 -56.736 56.030
556033 90.198 -56.751 55.986
556034 90.244 -56.767 55.937
556035 90.291 -56.783 55.885
556036 90.337 -56.799 55.838
556037 90.383 -56.815 55.792
556038 90.429 -56.830 55.748
556039 90.472 -56.842 55.705
556040 90.515 -56.854 55.665
556041 90.558 -56.866 55.626
556042 90.602 -56.878 55.589
556043 90.645 -56.891 55.554
556044 90.688 -56.903 55.520
556045 90.724 -56.920 55.488
556046 90.759 -56.937 55.461
556047 90.795 -56.955 55.439
556048 90.830 -56.972 55.417
556049 90.866 -56.990 55.395
556050 90.901 -57.006 55.376
556051 90.936 -57.024 55.358
556052 90.971 -57.041 55.344
556053 91.007 -57.059 55.331
556054 91.042 -57.076 55.320
556055 91.078 -57.094 55.310
556056 91.113 -57.110 55.300
556057 91.148 -57.128 55.291
556058 91.183 -57.145 55.286
556059 91.219 -57.163 55.282
556060 91.254 -57.180 55.280
556061 91.288 -57.194 55.279
556062 91.330 -57.203 55.277
556063 91.373 -57.213 55.277
556064 91.416 -57.222 55.276
556065 91.459 -57.232 55.278
556066 91.502 -57.242 55.280
556067 91.545 -57.252 55.281
556068 91.588 -57.261 55.283
556069 91.631 -57.271 55.284
556070 91.673 -57.281 55.283
556071 91.717 -57.290 55.278
556072 91.759 -57.300 55.273
556073 91.802 -57.310 55.271
556074 91.845 -57.319 55.276
556075 91.888 -57.329 55.284
556076 91.931 -57.338 55.292
556077 91.974 -57.348 55.299
556078 92.017 -57.357 55.305
556079 92.060 -57.367 55.308
556080 92.103 -57.376 55.306
556081 92.145 -57.389 55.304
556082 92.187 -57.413 55.304
556083 92.227 -57.436 55.303
556084 92.269 -57.459 55.300
556085 92.309 -57.482 55.292
556086 92.346 -57.510 55.276
556087 92.383 -57.536 55.253
556088 92.420 -57.563 55.228
556089 92.452 -57.586 55.204
556090 92.482 -57.607 55.181
556091 92.511 -57.628 55.156
556092 92.541 -57.648 55.135
556093 92.570 -57.670 55.114
556094 92.600 -57.690 55.093
556095 92.631 -57.708 55.070
556096 92.667 -57.719 55.046
556097 92.704 -57.730 55.023
556098 92.739 -57.740 55.001
556099 92.776 -57.751 54.978
556100 92.812 -57.761 54.956
556101 92.848 -57.772 54.933
556102 92.884 -57.780 54.917
556103 92.921 -57.786 54.902
556104 92.946 -57.792 54.890
556105 92.971 -57.798 54.879
556106 92.995 -57.804 54.870
556107 93.019 -57.810 54.865
556108 93.045 -57.814 54.864
556109 93.076 -57.813 54.863
556110 93.108 -57.812 54.863
556111 93.139 -57.811 54.860
556112 93.170 -57.811 54.855
556113 93.202 -57.810 54.851
556114 93.233 -57.809 54.851
556115 93.265 -57.809 54.841
556116 93.296 -57.808 54.821
556117 93.330 -57.814 54.795
556118 93.364 -57.822 54.766
556119 93.399 -57.830 54.746
556120 93.433 -57.839 54.751
556121 93.468 -57.847 54.758
556122 93.502 -57.855 54.763
556123 93.537 -57.864 54.764
556124 93.571 -57.872 54.767
556125 93.607 -57.884 54.763
556126 93.645 -57.904 54.753
556127 93.683 -57.924 54.739
556128 93.721 -57.943 54.728
556129 93.760 -57.963 54.721
556130 93.798 -57.983 54.712
556131 93.836 -58.003 54.701
556132 93.874 -58.023 54.688
556133 93.913 -58.043 54.674
556134 93.951 -58.062 54.658
556135 93.989 -58.082 54.642
556136 94.028 -58.102 54.625
556137 94.066 -58.122 54.606
556138 94.100 -58.139 54.587
556139 94.127 -58.150 54.567
556140 94.153 -58.162 54.544
556141 94.180 -58.175 54.517
556142 94.207 -58.186 54.486
556143 94.234 -58.198 54.450
556144 94.260 -58.209 54.412
556145 94.287 -58.222 54.370
556146 94.314 -58.233 54.324
556147 94.341 -58.245 54.276
556148 94.367 -58.256 54.240
556149 94.394 -58.269 54.208
556150 94.423 -58.282 54.178
556151 94.452 -58.295 54.146
556152 94.480 -58.309 54.112
556153 94.509 -58.322 54.076
556154 94.537 -58.335 54.048
556155 94.566 -58.349 54.018
556156 94.595 -58.363 53.986
556157 94.623 -58.375 53.951
556158 94.652 -58.389 53.916
556159 94.681 -58.403 53.889
556160 94.708 -58.416 53.860
556161 94.735 -58.428 53.828
556162 94.791 -58.437 53.797
556163 94.846 -58.446 53.771
556164 94.900 -58.456 53.745
556165 94.953 -58.466 53.718
556166 95.006 -58.475 53.691
556167 95.059 -58.485 53.664
556168 95.113 -58.495 53.633
556169 95.166 -58.504 53.602
556170 95.219 -58.513 53.577
556171 95.273 -58.523 53.554
556172 95.326 -58.532 53.530
556173 95.368 -58.538 53.505
556174 95.396 -58.539 53.479
556175 95.423 -58.539 53.455
556176 95.451 -58.540 53.429
556177 95.478 -58.540 53.400
556178 95.510 -58.553 53.372
556179 95.541 -58.567 53.349
556180 95.572 -58.579 53.331
556181 95.604 -58.593 53.314
556182 95.635 -58.607 53.292
556183 95.667 -58.620 53.271
556184 95.699 -58.633 53.250
556185 95.732 -58.647 53.229
556186 95.764 -58.661 53.206
556187 95.796 -58.674 53.176
556188 95.829 -58.688 53.149
556189 95.861 -58.701 53.124
556190 95.891 -58.708 53.095
556191 95.921 -58.715 53.060
556192 95.952 -58.722 53.026
556193 95.982 -58.729 52.983
556194 96.013 -58.735 52.946
556195 96.043 -58.742 52.915
556196 96.073 -58.748 52.886
556197 96.105 -58.757 52.855
556198 96.137 -58.767 52.825
556199 96.169 -58.766 52.789
556200 96.202 -58.767 52.758
556201 96.237 -58.770 52.732
556202 96.271 -58.773 52.707
556203 96.306 -58.777 52.678
556204 96.341 -58.780 52.646
556205 96.376 -58.783 52.614
556206 96.410 -58.786 52.581
556207 96.445 -58.789 52.548
556208 96.480 -58.792 52.516
556209 96.515 -58.795 52.484
556210 96.548 -58.797 52.450
556211 96.579 -58.798 52.416
556212 96.610 -58.799 52.378
556213 96.642 -58.800 52.342
556214 96.672 -58.801 52.308
556215 96.703 -58.802 52.272
556216 96.735 -58.805 52.233
556217 96.767 -58.810 52.197
556218 96.799 -58.816 52.161
556219 96.831 -58.821 52.126
556220 96.863 -58.827 52.091
556221 96.895 -58.832 52.057
556222 96.928 -58.834 52.018
556223 96.962 -58.835 51.978
556224 96.996 -58.836 51.938
556225 97.029 -58.836 51.902
556226 97.063 -58.839 51.870
556227 97.096 -58.842 51.830
556228 97.129 -58.846 51.792
556229 97.163 -58.850 51.757
556230 97.196 -58.855 51.725
556231 97.229 -58.858 51.693
556232 97.262 -58.862 51.658
556233 97.297 -58.866 51.622
556234 97.330 -58.869 51.581
556235 97.363 -58.874 51.537
556236 97.395 -58.878 51.492
556237 97.428 -58.882 51.450
556238 97.460 -58.886 51.410
556239 97.494 -58.890 51.367
556240 97.527 -58.895 51.326
556241 97.561 -58.899 51.286
556242 97.596 -58.908 51.245
556243 97.631 -58.916 51.207
556244 97.666 -58.924 51.167
556245 97.702 -58.933 51.125
556246 97.736 -58.941 51.081
556247 97.773 -58.941 51.036
556248 97.812 -58.941 50.992
556249 97.850 -58.940 50.946
556250 97.888 -58.940 50.904
556251 97.927 -58.940 50.861
556252 97.968 -58.934 50.816
556253 98.012 -58.922 50.772
556254 98.057 -58.911 50.728
556255 98.101 -58.899 50.687
556256 98.139 -58.894 50.649
556257 98.171 -58.897 50.613
556258 98.203 -58.899 50.579
556259 98.234 -58.902 50.547
556260 98.265 -58.905 50.516
556261 98.296 -58.907 50.490
556262 98.327 -58.909 50.463
556263 98.359 -58.912 50.438
556264 98.392 -58.912 50.414
556265 98.428 -58.911 50.390
556266 98.463 -58.903 50.369
556267 98.497 -58.892 50.344
556268 98.531 -58.882 50.319
556269 98.565 -58.871 50.296
556270 98.599 -58.861 50.276
556271 98.633 -58.849 50.257
556272 98.667 -58.839 50.240
556273 98.696 -58.836 50.224
556274 98.725 -58.833 50.205
556275 98.754 -58.831 50.186
556276 98.785 -58.828 50.166
556277 98.819 -58.824 50.147
556278 98.854 -58.820 50.124
556279 98.888 -58.817 50.099
556280 98.922 -58.816 50.072
556281 98.954 -58.816 50.044
556282 98.988 -58.815 50.015
556283 99.021 -58.814 49.985
556284 99.054 -58.814 49.952
556285 99.094 -58.823 49.918
556286 99.137 -58.826 49.883
556287 99.209 -58.820 49.850
556288 99.280 -58.814 49.818
556289 99.352 -58.808 49.787
556290 99.410 -58.808 49.757
556291 99.446 -58.815 49.728
556292 99.481 -58.823 49.698
556293 99.517 -58.830 49.668
556294 99.553 -58.838 49.637
556295 99.588 -58.846 49.607
556296 99.624 -58.852 49.577
556297 99.660 -58.860 49.546
556298 99.695 -58.867 49.513
556299 99.731 -58.875 49.479
556300 99.767 -58.883 49.445
556301 99.802 -58.890 49.411
556646 86.850 -55.427 57.631
556647 86.885 -55.443 57.624
556648 86.920 -55.460 57.619
556649 86.955 -55.476 57.611
556650 86.990 -55.492 57.601
556651 87.025 -55.508 57.591
556652 87.063 -55.523 57.580
556653 87.103 -55.536 57.569
556654 87.143 -55.549 57.556
556655 87.184 -55.563 57.541
556656 87.224 -55.576 57.524
556657 87.265 -55.589 57.507
556658 87.305 -55.603 57.489
556659 87.345 -55.616 57.468
556660 87.386 -55.630 57.444
556661 87.426 -55.643 57.417
556662 87.467 -55.656 57.388
556663 87.516 -55.664 57.357
556664 87.571 -55.668 57.325
556665 87.626 -55.671 57.289
556666 87.681 -55.675 57.253
556667 87.728 -55.682 57.214
556668 87.762 -55.693 57.173
556669 87.796 -55.705 57.132
556670 87.834 -55.721 57.090
556671 87.872 -55.737 57.048
556672 87.911 -55.753 57.006
556673 87.949 -55.769 56.964
556674 87.987 -55.785 56.922
556675 88.026 -55.801 56.880
556676 88.064 -55.817 56.840
556702 89.058 -56.232 56.065
556703 89.097 -56.248 56.047
556704 89.135 -56.264 56.028
556705 89.173 -56.280 56.010
556706 89.211 -56.296 55.994
556707 89.250 -56.312 55.979
556708 89.288 -56.328 55.966
556709 89.326 -56.345 55.954
556710 89.365 -56.360 55.945
556711 89.403 -56.376 55.937
556712 89.441 -56.392 55.930
556713 89.479 -56.408 55.925
556714 89.518 -56.424 55.921
556715 89.556 -56.440 55.919
556716 89.582 -56.463 55.919
556717 89.618 -56.477 55.920
556718 89.660 -56.489 55.924
556719 89.701 -56.499 55.931
556720 89.743 -56.510 55.940
556721 89.784 -56.521 55.951
556722 89.826 -56.532 55.964
556723 89.867 -56.542 55.978
556724 89.906 -56.555 55.993
556725 89.932 -56.571 56.009
556726 89.957 -56.589 56.023
556727 89.983 -56.605 56.036
556728 90.009 -56.623 56.048
556729 90.034 -56.639 56.062
556730 90.060 -56.657 56.080
556731 90.096 -56.671 56.089
556732 90.133 -56.686 56.090
556733 90.171 -56.700 56.068
556734 90.218 -56.716 56.028
556735 90.264 -56.733 55.978
556736 90.310 -56.749 55.914
556737 90.357 -56.765 55.857
556738 90.400 -56.777 55.806
556739 90.443 -56.789 55.758
556740 90.486 -56.801 55.712
556741 90.530 -56.813 55.670
556742 90.573 -56.826 55.630
556743 90.616 -56.838 55.592
556744 90.660 -56.850 55.557
556745 90.703 -56.862 55.524
556746 90.744 -56.876 55.493
556747 90.780 -56.893 55.467
556748 90.815 -56.910 55.445
556749 90.851 -56.928 55.423
556750 90.886 -56.945 55.402
556751 90.921 -56.962 55.383
556752 90.956 -56.980 55.366
556753 90.992 -56.997 55.352
556754 91.027 -57.014 55.339
556755 91.063 -57.031 55.328
556756 91.098 -57.049 55.317
556757 91.133 -57.066 55.308
556758 91.168 -57.084 55.299
556759 91.204 -57.101 55.292
556760 91.239 -57.118 55.288
556761 91.275 -57.135 55.286
556762 91.307 -57.148 55.284
556763 91.342 -57.160 55.284
556764 91.385 -57.169 55.284
556765 91.428 -57.178 55.285
556766 91.471 -57.188 55.290
556767 91.514 -57.198 55.295
556768 91.557 -57.207 55.298
556769 91.600 -57.217 55.300
556770 91.643 -57.227 55.297
556771 91.685 -57.236 55.292
556772 91.728 -57.246 55.286
556773 91.772 -57.256 55.282
556774 91.814 -57.265 55.282
556775 91.857 -57.275 55.289
556776 91.900 -57.285 55.295
556777 91.943 -57.294 55.300
556778 91.986 -57.304 55.305
556779 92.029 -57.313 55.310
556780 92.072 -57.323 55.314
556781 92.115 -57.332 55.316
556782 92.156 -57.354 55.315
556783 92.197 -57.378 55.316
556784 92.238 -57.401 55.316
556785 92.279 -57.425 55.312
556786 92.320 -57.448 55.303
556787 92.358 -57.474 55.284
556788 92.394 -57.501 55.263
556789 92.431 -57.528 55.243
556790 92.467 -57.554 55.224
556791 92.499 -57.576 55.204
556792 92.528 -57.598 55.184
556793 92.558 -57.618 55.169
556794 92.588 -57.639 55.151
556795 92.618 -57.660 55.128
556796 92.651 -57.674 55.101
556797 92.688 -57.685 55.075
556798 92.723 -57.695 55.049
556799 92.760 -57.706 55.025
556800 92.796 -57.717 55.002
556801 92.832 -57.727 54.979
556802 92.868 -57.738 54.959
556803 92.905 -57.746 54.946
556804 92.933 -57.752 54.933
556805 92.958 -57.758 54.921
556806 92.982 -57.764 54.911
556807 93.006 -57.770 54.902
556808 93.030 -57.776 54.896
556809 93.057 -57.780 54.894
556810 93.088 -57.780 54.891
556811 93.120 -57.779 54.887
556812 93.151 -57.778 54.881
556813 93.183 -57.777 54.874
556814 93.214 -57.777 54.866
556815 93.245 -57.776 54.862
556816 93.277 -57.775 54.852
556817 93.308 -57.774 54.838
556818 93.339 -57.774 54.817
556819 93.372 -57.778 54.791
556820 93.407 -57.786 54.760
556821 93.441 -57.795 54.746
556822 93.476 -57.803 54.750
556823 93.510 -57.811 54.757
556824 93.545 -57.820 54.758
556825 93.579 -57.828 54.758
556826 93.615 -57.841 54.755
556827 93.653 -57.861 54.749
556828 93.692 -57.881 54.739
556829 93.730 -57.901 54.728
556830 93.768 -57.921 54.721
556831 93.806 -57.940 54.714
556832 93.845 -57.960 54.701
556833 93.883 -57.980 54.684
556834 93.921 -58.000 54.666
556835 93.960 -58.020 54.647
556836 93.998 -58.040 54.629
556837 94.036 -58.059 54.610
556838 94.074 -58.079 54.590
556839 94.113 -58.099 54.571
556840 94.148 -58.117 54.551
556841 94.175 -58.128 54.529
556842 94.202 -58.140 54.504
556843 94.228 -58.153 54.475
556844 94.255 -58.164 54.442
556845 94.282 -58.176 54.405
556846 94.309 -58.187 54.366
556847 94.335 -58.200 54.324
556848 94.362 -58.211 54.278
556849 94.389 -58.223 54.241
556850 94.417 -58.236 54.208
556851 94.445 -58.249 54.177
556852 94.474 -58.263 54.145
556853 94.502 -58.276 54.113
556854 94.531 -58.289 54.077
556855 94.560 -58.303 54.045
556856 94.589 -58.316 54.013
556857 94.617 -58.329 53.979
556858 94.646 -58.343 53.942
556859 94.674 -58.356 53.907
556860 94.699 -58.368 53.876
556861 94.724 -58.380 53.846
556862 94.749 -58.392 53.815
556863 94.795 -58.402 53.784
556864 94.851 -58.412 53.753
556865 94.907 -58.421 53.726
556866 94.963 -58.430 53.697
556867 95.019 -58.439 53.669
556868 95.074 -58.448 53.640
556869 95.127 -58.458 53.610
556870 95.181 -58.468 53.580
556871 95.234 -58.478 53.553
556872 95.287 -58.487 53.530
556873 95.340 -58.497 53.506
556874 95.370 -58.498 53.480
556875 95.398 -58.498 53.453
556876 95.425 -58.499 53.429
556877 95.454 -58.500 53.404
556878 95.483 -58.501 53.376
556879 95.515 -58.511 53.349
556880 95.547 -58.525 53.326
556881 95.579 -58.538 53.311
556882 95.611 -58.551 53.298
556883 95.643 -58.565 53.280
556884 95.676 -58.579 53.262
556885 95.708 -58.592 53.243
556886 95.741 -58.606 53.222
556887 95.773 -58.620 53.200
556888 95.805 -58.633 53.170
556889 95.837 -58.647 53.145
556890 95.869 -58.658 53.119
556891 95.899 -58.664 53.087
556892 95.930 -58.671 53.050
556893 95.960 -58.678 53.018
556894 95.990 -58.685 52.974
556895 96.021 -58.691 52.939
556896 96.051 -58.698 52.907
556897 96.082 -58.706 52.876
556898 96.115 -58.717 52.844
556899 96.148 -58.728 52.817
556900 96.180 -58.736 52.783
556901 96.211 -58.734 52.754
556902 96.242 -58.733 52.728
556903 96.274 -58.733 52.703
556904 96.309 -58.736 52.673
556905 96.344 -58.739 52.644
556906 96.379 -58.742 52.614
556907 96.413 -58.745 52.583
556908 96.448 -58.748 52.551
556909 96.483 -58.751 52.517
556910 96.518 -58.754 52.479
556911 96.552 -58.757 52.441
556912 96.585 -58.758 52.405
556913 96.615 -58.759 52.370
556914 96.647 -58.760 52.334
556915 96.678 -58.761 52.296
556916 96.709 -58.762 52.256
556917 96.740 -58.767 52.213
556918 96.772 -58.772 52.173
556919 96.805 -58.777 52.136
556920 96.836 -58.783 52.101
556921 96.869 -58.789 52.065
556922 96.900 -58.794 52.028
556923 96.933 -58.799 51.991
556924 96.966 -58.799 51.953
556925 97.000 -58.800 51.914
556926 97.034 -58.801 51.877
556927 97.067 -58.804 51.839
556928 97.100 -58.808 51.799
556929 97.134 -58.812 51.760
556930 97.167 -58.816 51.723
556931 97.201 -58.820 51.692
556932 97.234 -58.824 51.663
556933 97.267 -58.827 51.629
556934 97.301 -58.831 51.591
556935 97.333 -58.836 51.553
556936 97.366 -58.839 51.514
556937 97.399 -58.844 51.470
556938 97.431 -58.848 51.421
556939 97.464 -58.852 51.377
556940 97.497 -58.856 51.337
556941 97.530 -58.861 51.296
556942 97.564 -58.867 51.255
556943 97.599 -58.875 51.217
556944 97.634 -58.883 51.180
556945 97.669 -58.892 51.142
556946 97.705 -58.899 51.102
556947 97.741 -58.905 51.059
556948 97.780 -58.911 51.014
556949 97.818 -58.910 50.969
556950 97.857 -58.910 50.923
556951 97.895 -58.909 50.882
556952 97.934 -58.908 50.840
556953 97.979 -58.897 50.796
556954 98.023 -58.885 50.752
556955 98.067 -58.873 50.710
556956 98.112 -58.861 50.671
556957 98.150 -58.856 50.633
556958 98.181 -58.858 50.599
556959 98.212 -58.861 50.566
556960 98.243 -58.864 50.534
556961 98.275 -58.866 50.504
556962 98.306 -58.869 50.477
556963 98.337 -58.871 50.446
556964 98.368 -58.874 50.419
556965 98.400 -58.876 50.394
556966 98.434 -58.875 50.370
556967 98.470 -58.871 50.348
556968 98.504 -58.861 50.323
556969 98.538 -58.850 50.298
556970 98.572 -58.839 50.275
556971 98.606 -58.828 50.254
556972 98.640 -58.817 50.235
556973 98.669 -58.814 50.221
556974 98.698 -58.812 50.207
556975 98.726 -58.809 50.190
556976 98.755 -58.807 50.173
556977 98.789 -58.803 50.155
556978 98.823 -58.799 50.136
556979 98.857 -58.796 50.114
556980 98.892 -58.792 50.089
556981 98.926 -58.791 50.064
556982 98.958 -58.790 50.038
556983 98.992 -58.789 50.012
556984 99.025 -58.789 49.984
556985 99.058 -58.789 49.952
556986 99.091 -58.789 49.919
556987 99.144 -58.786 49.886
556988 99.216 -58.781 49.853
556989 99.288 -58.775 49.821
556990 99.360 -58.769 49.791
556991 99.411 -58.771 49.761
556992 99.446 -58.779 49.732
556993 99.482 -58.786 49.703
556994 99.518 -58.793 49.673
556995 99.554 -58.801 49.643
556996 99.589 -58.808 49.614
556997 99.625 -58.816 49.585
556998 99.660 -58.824 49.553
556999 99.696 -58.831 49.521
557000 99.732 -58.839 49.487
557001 99.768 -58.846 49.454
557002 99.803 -58.853 49.420
557347 86.880 -55.377 57.631
557348 86.915 -55.393 57.625
557349 86.950 -55.410 57.620
557350 86.985 -55.426 57.612
557351 87.020 -55.442 57.604
557352 87.055 -55.458 57.594
557353 87.090 -55.474 57.584
557354 87.125 -55.490 57.573
557355 87.160 -55.506 57.562
557356 87.201 -55.520 57.548
557357 87.241 -55.533 57.534
557358 87.281 -55.546 57.519
557359 87.322 -55.559 57.504
557360 87.362 -55.573 57.482
557361 87.403 -55.586 57.459
557362 87.443 -55.599 57.434
557363 87.487 -55.611 57.406
557364 87.542 -55.614 57.376
557365 87.597 -55.619 57.346
557366 87.651 -55.623 57.312
557367 87.703 -55.628 57.277
557368 87.741 -55.638 57.239
557369 87.776 -55.650 57.199
557370 87.814 -55.666 57.158
557371 87.852 -55.683 57.116
557372 87.891 -55.699 57.074
557373 87.929 -55.715 57.032
557374 87.967 -55.730 56.989
557375 88.005 -55.746 56.946
557376 88.043 -55.762 56.904
557377 88.081 -55.778 56.863
557378 88.120 -55.794 56.823
557403 89.076 -56.194 56.064
557404 89.114 -56.210 56.045
557405 89.153 -56.226 56.026
557406 89.191 -56.241 56.008
557407 89.229 -56.257 55.992
557408 89.267 -56.274 55.976
557409 89.306 -56.290 55.963
557410 89.344 -56.306 55.951
557411 89.382 -56.322 55.941
557412 89.420 -56.338 55.933
557413 89.459 -56.354 55.926
557414 89.497 -56.370 55.920
557415 89.535 -56.385 55.917
557416 89.574 -56.401 55.914
557417 89.600 -56.424 55.915
557418 89.638 -56.438 55.917
557419 89.679 -56.448 55.921
557420 89.721 -56.460 55.928
557421 89.762 -56.470 55.939
557422 89.804 -56.481 55.950
557423 89.845 -56.492 55.964
557424 89.886 -56.503 55.981
557425 89.913 -56.520 55.998
557426 89.939 -56.536 56.015
557427 89.965 -56.554 56.032
557428 89.990 -56.570 56.048
557429 90.016 -56.588 56.064
557430 90.042 -56.605 56.082
557431 90.071 -56.621 56.102
557432 90.107 -56.636 56.110
557433 90.145 -56.650 56.110
557434 90.191 -56.666 56.091
557435 90.237 -56.682 56.055
557436 90.284 -56.699 56.006
557437 90.328 -56.711 55.937
557438 90.371 -56.724 55.872
557439 90.414 -56.736 55.816
557440 90.458 -56.748 55.766
557441 90.501 -56.761 55.718
557442 90.544 -56.773 55.675
557443 90.588 -56.785 55.635
557444 90.631 -56.797 55.597
557445 90.674 -56.809 55.562
557446 90.718 -56.821 55.530
557447 90.761 -56.834 55.499
557448 90.800 -56.849 55.474
557449 90.835 -56.866 55.452
557450 90.871 -56.884 55.431
557451 90.906 -56.900 55.410
557452 90.942 -56.918 55.391
557453 90.977 -56.935 55.374
557454 91.012 -56.953 55.361
557455 91.047 -56.970 55.348
557456 91.083 -56.987 55.336
557457 91.118 -57.004 55.326
557458 91.154 -57.022 55.316
557459 91.189 -57.039 55.307
557460 91.224 -57.056 55.300
557461 91.259 -57.074 55.295
557462 91.294 -57.090 55.293
557463 91.326 -57.102 55.292
557464 91.359 -57.114 55.291
557465 91.397 -57.125 55.293
557466 91.440 -57.134 55.297
557467 91.483 -57.144 55.306
557468 91.526 -57.154 55.313
557469 91.569 -57.163 55.316
557470 91.611 -57.173 55.312
557471 91.655 -57.183 55.303
557472 91.698 -57.192 55.294
557473 91.740 -57.202 55.289
557474 91.783 -57.212 55.290
557475 91.826 -57.221 55.294
557476 91.869 -57.231 55.300
557477 91.912 -57.241 55.303
557478 91.955 -57.250 55.304
557479 91.998 -57.260 55.305
557480 92.041 -57.269 55.309
557481 92.084 -57.279 55.313
557482 92.126 -57.297 55.317
557483 92.166 -57.320 55.320
557484 92.208 -57.344 55.324
557485 92.248 -57.367 55.327
557486 92.290 -57.390 55.325
557487 92.331 -57.413 55.317
557488 92.369 -57.439 55.303
557489 92.406 -57.466 55.287
557490 92.442 -57.493 55.272
557491 92.479 -57.520 55.256
557492 92.515 -57.546 55.236
557493 92.546 -57.567 55.217
557494 92.575 -57.589 55.198
557495 92.605 -57.609 55.179
557496 92.635 -57.630 55.157
557497 92.671 -57.640 55.133
557498 92.707 -57.651 55.109
557499 92.743 -57.661 55.083
557500 92.780 -57.672 55.057
557501 92.816 -57.683 55.032
557502 92.852 -57.693 55.010
557503 92.888 -57.704 54.994
557504 92.921 -57.712 54.980
557505 92.945 -57.718 54.965
557506 92.970 -57.724 54.952
557507 92.994 -57.730 54.940
557508 93.018 -57.736 54.930
557509 93.042 -57.742 54.922
557510 93.069 -57.746 54.917
557511 93.100 -57.746 54.912
557512 93.132 -57.745 54.905
557513 93.163 -57.744 54.897
557514 93.194 -57.743 54.888
557515 93.226 -57.742 54.877
557516 93.257 -57.742 54.868
557517 93.288 -57.741 54.860
557518 93.320 -57.740 54.849
557519 93.351 -57.739 54.833
557520 93.382 -57.739 54.811
557521 93.415 -57.742 54.781
557522 93.449 -57.751 54.756
557523 93.484 -57.759 54.744
557524 93.518 -57.767 54.744
557525 93.553 -57.776 54.750
557526 93.587 -57.784 54.754
557527 93.624 -57.799 54.753
557528 93.662 -57.818 54.750
557529 93.701 -57.839 54.744
557530 93.739 -57.858 54.737
557531 93.777 -57.878 54.731
557532 93.816 -57.898 54.723
557533 93.854 -57.918 54.708
557534 93.892 -57.937 54.690
557535 93.930 -57.958 54.669
557536 93.969 -57.977 54.647
557537 94.007 -57.997 54.626
557538 94.045 -58.017 54.605
557539 94.083 -58.037 54.581
557540 94.122 -58.056 54.558
557541 94.160 -58.076 54.537
557542 94.196 -58.095 54.514
557543 94.223 -58.106 54.489
557544 94.250 -58.118 54.460
557545 94.277 -58.131 54.429
557546 94.303 -58.142 54.394
557547 94.330 -58.154 54.357
557548 94.357 -58.165 54.322
557549 94.384 -58.178 54.281
557550 94.410 -58.190 54.246
557551 94.439 -58.203 54.214
557552 94.468 -58.216 54.183
557553 94.496 -58.229 54.152
557554 94.525 -58.243 54.121
557555 94.554 -58.256 54.088
557556 94.582 -58.269 54.055
557557 94.611 -58.283 54.020
557558 94.638 -58.296 53.984
557559 94.664 -58.308 53.945
557560 94.689 -58.320 53.909
557561 94.714 -58.332 53.875
557562 94.740 -58.344 53.841
557563 94.765 -58.357 53.806
557564 94.799 -58.368 53.773
557565 94.854 -58.377 53.740
557566 94.910 -58.386 53.709
557567 94.966 -58.395 53.678
557568 95.022 -58.405 53.649
557569 95.078 -58.414 53.619
557570 95.134 -58.423 53.590
557571 95.190 -58.432 53.560
557572 95.246 -58.442 53.534
557573 95.302 -58.451 53.510
557574 95.344 -58.457 53.484
557575 95.374 -58.459 53.456
557576 95.404 -58.460 53.430
557577 95.433 -58.462 53.405
557578 95.463 -58.463 53.378
557579 95.493 -58.465 53.350
557580 95.523 -58.469 53.324
557581 95.556 -58.483 53.303
557582 95.588 -58.497 53.291
557583 95.620 -58.510 53.278
557584 95.653 -58.524 53.263
557585 95.685 -58.538 53.248
557586 95.717 -58.551 53.232
557587 95.749 -58.565 53.213
557588 95.781 -58.578 53.192
557589 95.814 -58.592 53.165
557590 95.846 -58.605 53.140
557591 95.877 -58.614 53.106
557592 95.907 -58.621 53.068
557593 95.938 -58.627 53.034
557594 95.968 -58.634 53.008
557595 95.998 -58.641 52.967
557596 96.028 -58.648 52.935
557597 96.059 -58.654 52.903
557598 96.092 -58.666 52.868
557599 96.125 -58.676 52.836
557600 96.158 -58.688 52.811
557601 96.190 -58.698 52.779
557602 96.223 -58.705 52.752
557603 96.253 -58.703 52.727
557604 96.285 -58.701 52.699
557605 96.316 -58.699 52.669
557606 96.347 -58.698 52.644
557607 96.382 -58.701 52.616
557608 96.416 -58.704 52.586
557609 96.451 -58.707 52.553
557610 96.486 -58.710 52.518
557611 96.521 -58.713 52.478
557612 96.555 -58.716 52.438
557613 96.590 -58.719 52.402
557614 96.620 -58.720 52.367
557615 96.652 -58.720 52.329
557616 96.683 -58.721 52.288
557617 96.714 -58.723 52.245
557618 96.746 -58.729 52.202
557619 96.778 -58.734 52.161
557620 96.810 -58.739 52.121
557621 96.842 -58.745 52.082
557622 96.874 -58.751 52.044
557623 96.906 -58.756 52.004
557624 96.938 -58.762 51.967
557625 96.971 -58.764 51.929
557626 97.005 -58.764 51.892
557627 97.038 -58.765 51.854
557628 97.072 -58.770 51.816
557629 97.105 -58.773 51.776
557630 97.139 -58.777 51.736
557631 97.172 -58.781 51.697
557632 97.205 -58.785 51.662
557633 97.238 -58.789 51.631
557634 97.272 -58.793 51.603
557635 97.304 -58.797 51.567
557636 97.337 -58.802 51.529
557637 97.369 -58.805 51.489
557638 97.402 -58.810 51.448
557639 97.435 -58.814 51.404
557640 97.467 -58.818 51.360
557641 97.500 -58.822 51.317
557642 97.532 -58.827 51.273
557643 97.567 -58.833 51.231
557644 97.602 -58.842 51.193
557645 97.638 -58.849 51.158
557646 97.673 -58.855 51.121
557647 97.708 -58.861 51.080
557648 97.746 -58.867 51.038
557649 97.786 -58.872 50.994
557650 97.825 -58.877 50.951
557651 97.864 -58.879 50.905
557652 97.902 -58.879 50.861
557653 97.945 -58.871 50.819
557654 97.989 -58.859 50.778
557655 98.033 -58.847 50.737
557656 98.076 -58.835 50.699
557657 98.118 -58.824 50.662
557658 98.154 -58.820 50.627
557659 98.189 -58.821 50.593
557660 98.222 -58.823 50.560
557661 98.253 -58.825 50.529
557662 98.284 -58.828 50.498
557663 98.315 -58.830 50.467
557664 98.347 -58.833 50.435
557665 98.378 -58.835 50.404
557666 98.409 -58.838 50.375
557667 98.440 -58.840 50.350
557668 98.477 -58.838 50.327
557669 98.511 -58.828 50.302
557670 98.545 -58.817 50.279
557671 98.579 -58.807 50.256
557672 98.613 -58.796 50.233
557673 98.642 -58.793 50.214
557674 98.670 -58.790 50.198
557675 98.700 -58.788 50.185
557676 98.728 -58.786 50.172
557677 98.758 -58.783 50.158
557678 98.792 -58.779 50.140
557679 98.827 -58.775 50.122
557680 98.861 -58.771 50.101
557681 98.896 -58.767 50.078
557682 98.930 -58.765 50.056
557683 98.962 -58.764 50.032
557684 98.996 -58.764 50.006
557685 99.029 -58.764 49.980
557686 99.068 -58.758 49.951
557687 99.109 -58.753 49.919
557688 99.152 -58.748 49.887
557689 99.224 -58.742 49.855
557690 99.296 -58.736 49.823
557691 99.368 -58.730 49.794
557692 99.412 -58.734 49.765
557693 99.447 -58.742 49.737
557694 99.483 -58.749 49.708
557695 99.519 -58.757 49.678
557696 99.555 -58.764 49.649
557697 99.590 -58.772 49.621
557698 99.626 -58.780 49.591
557699 99.661 -58.786 49.560
557700 99.697 -58.794 49.529
557701 99.733 -58.802 49.496
557702 99.769 -58.809 49.462
557703 99.804 -58.817 49.430
558048 86.910 -55.327 57.631
558049 86.945 -55.344 57.626
558050 86.980 -55.360 57.620
558051 87.015 -55.376 57.614
558052 87.050 -55.392 57.605
558053 87.085 -55.408 57.596
558054 87.120 -55.424 57.588
558055 87.155 -55.440 57.578
558056 87.190 -55.456 57.567
558057 87.225 -55.473 57.555
558058 87.260 -55.489 57.542
558059 87.298 -55.503 57.530
558060 87.339 -55.516 57.515
558061 87.379 -55.530 57.496
558062 87.419 -55.542 57.474
558063 87.460 -55.556 57.451
558064 87.512 -55.562 57.425
558065 87.567 -55.566 57.397
558066 87.622 -55.570 57.367
558067 87.677 -55.574 57.335
558068 87.719 -55.583 57.301
558069 87.755 -55.596 57.265
558070 87.793 -55.612 57.225
558071 87.832 -55.628 57.185
558072 87.870 -55.644 57.142
558073 87.908 -55.660 57.101
558074 87.946 -55.676 57.059
558075 87.985 -55.692 57.016
558076 88.023 -55.708 56.972
558077 88.061 -55.724 56.928
558078 88.100 -55.740 56.887
558079 88.138 -55.756 56.846
558104 89.094 -56.155 56.064
558105 89.132 -56.171 56.043
558106 89.171 -56.187 56.024
558107 89.209 -56.204 56.007
558108 89.247 -56.219 55.990
558109 89.285 -56.235 55.974
558110 89.324 -56.251 55.960
558111 89.362 -56.267 55.948
558112 89.400 -56.283 55.937
558113 89.438 -56.299 55.928
558114 89.476 -56.315 55.921
558115 89.515 -56.331 55.915
558116 89.553 -56.347 55.912
558117 89.591 -56.363 55.909
558118 89.619 -56.385 55.909
558119 89.658 -56.398 55.912
558120 89.699 -56.409 55.917
558121 89.740 -56.420 55.924
558122 89.781 -56.431 55.935
558123 89.823 -56.442 55.947
558124 89.865 -56.452 55.961
558125 89.895 -56.467 55.979
558126 89.921 -56.485 55.996
558127 89.947 -56.501 56.016
558128 89.972 -56.519 56.037
558129 89.998 -56.536 56.059
558130 90.023 -56.553 56.081
558131 90.049 -56.570 56.099
558132 90.082 -56.586 56.113
558133 90.119 -56.600 56.117
558134 90.164 -56.616 56.115
558135 90.211 -56.632 56.101
558136 90.255 -56.646 56.070
558137 90.299 -56.658 56.022
558138 90.342 -56.671 55.951
558139 90.386 -56.683 55.883
558140 90.429 -56.696 55.823
558141 90.472 -56.708 55.770
558142 90.516 -56.720 55.722
558143 90.559 -56.732 55.679
558144 90.602 -56.744 55.639
558145 90.646 -56.756 55.601
558146 90.689 -56.768 55.567
558147 90.732 -56.780 55.535
558148 90.776 -56.793 55.505
558149 90.819 -56.805 55.481
558150 90.856 -56.821 55.459
558151 90.891 -56.839 55.438
558152 90.927 -56.856 55.419
558153 90.962 -56.874 55.401
558154 90.997 -56.890 55.385
558155 91.032 -56.908 55.371
558156 91.068 -56.925 55.358
558157 91.103 -56.943 55.346
558158 91.139 -56.960 55.335
558159 91.174 -56.978 55.325
558160 91.209 -56.995 55.315
558161 91.244 -57.012 55.308
558162 91.280 -57.029 55.303
558163 91.313 -57.044 55.300
558164 91.346 -57.056 55.298
558165 91.378 -57.069 55.298
558166 91.410 -57.081 55.301
558167 91.452 -57.091 55.308
558168 91.495 -57.100 55.318
558169 91.538 -57.110 55.326
558170 91.581 -57.119 55.327
558171 91.624 -57.129 55.318
558172 91.666 -57.139 55.306
558173 91.710 -57.148 55.300
558174 91.752 -57.158 55.303
558175 91.795 -57.168 55.315
558176 91.838 -57.177 55.329
558177 91.881 -57.187 55.341
558178 91.924 -57.197 55.344
558179 91.967 -57.206 55.341
558180 92.010 -57.216 55.334
558181 92.053 -57.225 55.324
558182 92.095 -57.238 55.316
558183 92.136 -57.262 55.313
558184 92.177 -57.285 55.319
558185 92.218 -57.309 55.328
558186 92.260 -57.332 55.336
558187 92.300 -57.356 55.339
558188 92.342 -57.379 55.333
558189 92.380 -57.404 55.326
558190 92.417 -57.431 55.317
558191 92.454 -57.457 55.303
558192 92.490 -57.485 55.286
558193 92.527 -57.511 55.265
558194 92.563 -57.538 55.242
558195 92.593 -57.558 55.218
558196 92.622 -57.579 55.196
558197 92.655 -57.595 55.173
558198 92.691 -57.605 55.151
558199 92.727 -57.616 55.129
558200 92.764 -57.626 55.104
558201 92.800 -57.637 55.076
558202 92.836 -57.648 55.050
558203 92.868 -57.659 55.028
558204 92.902 -57.670 55.009
558205 92.931 -57.679 54.994
558206 92.957 -57.685 54.977
558207 92.981 -57.691 54.963
558208 93.005 -57.697 54.952
558209 93.029 -57.703 54.942
558210 93.054 -57.709 54.932
558211 93.081 -57.712 54.925
558212 93.112 -57.711 54.918
558213 93.144 -57.711 54.912
558214 93.175 -57.710 54.904
558215 93.206 -57.710 54.895
558216 93.237 -57.709 54.883
558217 93.269 -57.708 54.872
558218 93.300 -57.708 54.865
558219 93.332 -57.707 54.856
558220 93.363 -57.706 54.843
558221 93.395 -57.705 54.826
558222 93.426 -57.705 54.800
558223 93.458 -57.707 54.769
558224 93.492 -57.715 54.742
558225 93.527 -57.724 54.731
558226 93.561 -57.731 54.737
558227 93.596 -57.739 54.745
558228 93.633 -57.756 54.748
558229 93.672 -57.776 54.748
558230 93.710 -57.796 54.746
558231 93.748 -57.816 54.742
558232 93.786 -57.836 54.736
558233 93.825 -57.855 54.728
558234 93.863 -57.875 54.713
558235 93.901 -57.895 54.696
558236 93.939 -57.915 54.677
558237 93.978 -57.935 54.657
558238 94.016 -57.955 54.637
558239 94.054 -57.974 54.613
558240 94.092 -57.994 54.587
558241 94.131 -58.014 54.562
558242 94.169 -58.034 54.536
558243 94.207 -58.053 54.508
558244 94.245 -58.073 54.481
558245 94.271 -58.085 54.451
558246 94.298 -58.096 54.420
558247 94.325 -58.109 54.387
558248 94.352 -58.120 54.355
558249 94.379 -58.132 54.323
558250 94.405 -58.144 54.291
558251 94.433 -58.156 54.262
558252 94.462 -58.170 54.231
558253 94.490 -58.183 54.200
558254 94.519 -58.197 54.170
558255 94.547 -58.210 54.141
558256 94.576 -58.223 54.109
558257 94.604 -58.236 54.075
558258 94.629 -58.248 54.038
558259 94.654 -58.260 53.999
558260 94.680 -58.272 53.959
558261 94.705 -58.285 53.920
558262 94.730 -58.297 53.882
558263 94.755 -58.309 53.844
558264 94.780 -58.321 53.805
558265 94.806 -58.333 53.766
558266 94.858 -58.343 53.730
558267 94.914 -58.352 53.696
558268 94.970 -58.361 53.662
558269 95.026 -58.370 53.630
558270 95.082 -58.379 53.599
558271 95.138 -58.389 53.570
558272 95.194 -58.398 53.542
558273 95.250 -58.407 53.514
558274 95.306 -58.416 53.487
558275 95.352 -58.421 53.459
558276 95.384 -58.422 53.432
558277 95.413 -58.423 53.406
558278 95.443 -58.425 53.379
558279 95.473 -58.427 53.351
558280 95.502 -58.429 53.324
558281 95.532 -58.430 53.298
558282 95.564 -58.442 53.278
558283 95.597 -58.456 53.266
558284 95.629 -58.469 53.257
558285 95.661 -58.483 53.245
558286 95.693 -58.496 53.230
558287 95.726 -58.510 53.216
558288 95.758 -58.523 53.203
558289 95.790 -58.537 53.181
558290 95.822 -58.551 53.153
558291 95.854 -58.563 53.128
558292 95.885 -58.570 53.097
558293 95.915 -58.577 53.064
558294 95.945 -58.583 53.034
558295 95.976 -58.590 53.012
558296 96.006 -58.597 52.973
558297 96.036 -58.604 52.939
558298 96.069 -58.614 52.905
558299 96.102 -58.626 52.871
558300 96.135 -58.636 52.843
558301 96.168 -58.648 52.814
558302 96.200 -58.658 52.779
558303 96.233 -58.670 52.752
558304 96.265 -58.673 52.726
558305 96.296 -58.671 52.695
558306 96.327 -58.670 52.663
558307 96.359 -58.668 52.642
558308 96.390 -58.667 52.618
558309 96.420 -58.664 52.587
558310 96.454 -58.666 52.552
558311 96.488 -58.669 52.518
558312 96.523 -58.672 52.480
558313 96.558 -58.675 52.441
558314 96.593 -58.678 52.404
558315 96.626 -58.679 52.367
558316 96.657 -58.680 52.327
558317 96.688 -58.682 52.284
558318 96.720 -58.685 52.240
558319 96.751 -58.691 52.198
558320 96.784 -58.696 52.156
558321 96.815 -58.701 52.112
558322 96.848 -58.708 52.070
558323 96.879 -58.713 52.030
558324 96.911 -58.718 51.987
558325 96.943 -58.724 51.948
558326 96.976 -58.728 51.911
558327 97.010 -58.729 51.872
558328 97.043 -58.731 51.834
558329 97.077 -58.735 51.796
558330 97.110 -58.739 51.756
558331 97.143 -58.742 51.717
558332 97.176 -58.747 51.677
558333 97.209 -58.751 51.639
558334 97.242 -58.755 51.605
558335 97.275 -58.759 51.578
558336 97.307 -58.764 51.546
558337 97.340 -58.767 51.507
558338 97.373 -58.772 51.466
558339 97.405 -58.776 51.425
558340 97.438 -58.780 51.384
558341 97.470 -58.784 51.343
558342 97.503 -58.789 51.300
558343 97.535 -58.792 51.256
558344 97.571 -58.799 51.213
558345 97.605 -58.805 51.173
558346 97.641 -58.811 51.137
558347 97.676 -58.817 51.099
558348 97.712 -58.823 51.060
558349 97.751 -58.828 51.019
558350 97.791 -58.833 50.975
558351 97.830 -58.838 50.930
558352 97.870 -58.843 50.886
558353 97.909 -58.846 50.843
558354 97.951 -58.834 50.802
558355 97.993 -58.822 50.762
558356 98.035 -58.811 50.722
558357 98.077 -58.799 50.685
558358 98.119 -58.787 50.649
558359 98.152 -58.787 50.614
558360 98.185 -58.788 50.580
558361 98.221 -58.789 50.547
558362 98.256 -58.789 50.515
558363 98.291 -58.790 50.484
558364 98.325 -58.792 50.451
558365 98.356 -58.794 50.419
558366 98.387 -58.797 50.388
558367 98.419 -58.799 50.359
558368 98.450 -58.802 50.332
558369 98.483 -58.803 50.307
558370 98.517 -58.796 50.282
558371 98.552 -58.786 50.258
558372 98.585 -58.775 50.235
558373 98.615 -58.771 50.213
558374 98.644 -58.769 50.193
558375 98.672 -58.766 50.178
558376 98.701 -58.764 50.164
558377 98.730 -58.761 50.152
558378 98.762 -58.758 50.140
558379 98.796 -58.755 50.124
558380 98.830 -58.751 50.107
558381 98.865 -58.747 50.087
558382 98.899 -58.743 50.066
558383 98.934 -58.739 50.045
558384 98.971 -58.736 50.023
558385 99.011 -58.730 49.999
558386 99.051 -58.725 49.974
558387 99.091 -58.720 49.947
558388 99.132 -58.714 49.917
558389 99.172 -58.708 49.886
558390 99.231 -58.702 49.855
558391 99.303 -58.696 49.823
558392 99.375 -58.691 49.794
558393 99.413 -58.698 49.766
558394 99.448 -58.705 49.738
558395 99.484 -58.713 49.711
558396 99.519 -58.720 49.682
558397 99.555 -58.727 49.654
558398 99.591 -58.735 49.625
558399 99.627 -58.742 49.596
558400 99.662 -58.750 49.566
558401 99.698 -58.758 49.535
558402 99.734 -58.765 49.503
558403 99.769 -58.773 49.469
558404 99.805 -58.780 49.437
558749 86.940 -55.277 57.631
558750 86.975 -55.294 57.626
558751 87.010 -55.310 57.621
558752 87.045 -55.326 57.614
558753 87.080 -55.342 57.607
558754 87.115 -55.358 57.599
558755 87.150 -55.374 57.590
558756 87.185 -55.390 57.581
558757 87.220 -55.406 57.572
558758 87.255 -55.423 57.562
558759 87.290 -55.439 57.550
558760 87.325 -55.455 57.539
558761 87.360 -55.470 57.525
558762 87.396 -55.486 57.509
558763 87.436 -55.499 57.490
558764 87.482 -55.509 57.468
558765 87.537 -55.514 57.445
558766 87.592 -55.517 57.419
558767 87.647 -55.521 57.390
558768 87.698 -55.527 57.358
558769 87.734 -55.542 57.325
558770 87.773 -55.558 57.290
558771 87.811 -55.574 57.252
558772 87.849 -55.589 57.213
558773 87.888 -55.605 57.172
558774 87.926 -55.621 57.130
558775 87.964 -55.637 57.087
558776 88.002 -55.653 57.043
558777 88.041 -55.669 56.999
558778 88.079 -55.685 56.954
558779 88.117 -55.701 56.911
558780 88.155 -55.717 56.870
558781 88.194 -55.734 56.829
558805 89.112 -56.117 56.063
558806 89.150 -56.133 56.042
558807 89.188 -56.149 56.022
558808 89.227 -56.165 56.004
558809 89.265 -56.181 55.987
558810 89.303 -56.197 55.972
558811 89.341 -56.213 55.956
558812 89.380 -56.229 55.944
558813 89.418 -56.244 55.933
558814 89.456 -56.260 55.924
558815 89.494 -56.276 55.916
558816 89.533 -56.292 55.910
558817 89.571 -56.308 55.906
558818 89.609 -56.325 55.903
558819 89.637 -56.346 55.904
558820 89.677 -56.359 55.906
558821 89.719 -56.370 55.911
558822 89.760 -56.380 55.918
558823 89.801 -56.391 55.928
558824 89.842 -56.402 55.941
558825 89.877 -56.416 55.955
558826 89.903 -56.432 55.973
558827 89.929 -56.450 55.992
558828 89.954 -56.467 56.014
558829 89.979 -56.484 56.041
558830 90.005 -56.501 56.070
558831 90.031 -56.518 56.096
558832 90.056 -56.536 56.114
558833 90.093 -56.550 56.120
558834 90.138 -56.565 56.116
558835 90.183 -56.581 56.110
558836 90.227 -56.593 56.094
558837 90.270 -56.605 56.061
558838 90.313 -56.617 56.013
558839 90.357 -56.630 55.949
558840 90.400 -56.642 55.882
558841 90.443 -56.655 55.823
558842 90.487 -56.667 55.771
558843 90.530 -56.679 55.723
558844 90.574 -56.691 55.681
558845 90.617 -56.703 55.642
558846 90.660 -56.715 55.606
558847 90.704 -56.728 55.572
558848 90.747 -56.740 55.541
558849 90.790 -56.752 55.512
558850 90.833 -56.765 55.489
558851 90.876 -56.777 55.467
558852 90.911 -56.794 55.448
558853 90.947 -56.812 55.429
558854 90.982 -56.829 55.413
558855 91.017 -56.846 55.396
558856 91.053 -56.864 55.382
558857 91.088 -56.881 55.370
558858 91.123 -56.898 55.358
558859 91.159 -56.915 55.346
558860 91.194 -56.933 55.336
558861 91.229 -56.950 55.326
558862 91.264 -56.968 55.317
558863 91.300 -56.985 55.311
558864 91.333 -56.998 55.306
558865 91.365 -57.010 55.303
558866 91.397 -57.022 55.304
558867 91.430 -57.035 55.308
558868 91.464 -57.047 55.316
558869 91.507 -57.056 55.326
558870 91.549 -57.066 55.333
558871 91.593 -57.075 55.331
558872 91.636 -57.085 55.320
558873 91.678 -57.095 55.308
558874 91.721 -57.104 55.307
558875 91.764 -57.114 55.318
558876 91.807 -57.124 55.338
558877 91.850 -57.133 55.359
558878 91.893 -57.143 55.370
558879 91.936 -57.153 55.375
558880 91.979 -57.162 55.373
558881 92.022 -57.172 55.366
558882 92.065 -57.181 55.356
558883 92.106 -57.204 55.347
558884 92.147 -57.228 55.336
558885 92.188 -57.251 55.332
558886 92.229 -57.275 55.336
558887 92.270 -57.298 55.346
558888 92.311 -57.321 55.354
558889 92.352 -57.344 55.352
558890 92.392 -57.369 55.346
558891 92.428 -57.396 55.335
558892 92.465 -57.423 55.321
558893 92.502 -57.449 55.304
558894 92.538 -57.476 55.282
558895 92.575 -57.503 55.254
558896 92.610 -57.529 55.227
558897 92.640 -57.549 55.201
558898 92.675 -57.561 55.178
558899 92.711 -57.571 55.159
558900 92.747 -57.582 55.138
558901 92.780 -57.593 55.114
558902 92.813 -57.604 55.087
558903 92.846 -57.615 55.061
558904 92.879 -57.626 55.038
558905 92.912 -57.637 55.017
558906 92.940 -57.644 55.001
558907 92.969 -57.651 54.984
558908 92.993 -57.658 54.969
558909 93.017 -57.664 54.956
558910 93.041 -57.669 54.946
558911 93.066 -57.675 54.934
558912 93.093 -57.679 54.923
558913 93.124 -57.678 54.913
558914 93.155 -57.677 54.905
558915 93.187 -57.677 54.896
558916 93.218 -57.676 54.887
558917 93.250 -57.675 54.874
558918 93.281 -57.674 54.862
558919 93.312 -57.673 54.856
558920 93.344 -57.673 54.850
558921 93.375 -57.672 54.839
558922 93.406 -57.671 54.823
558923 93.437 -57.670 54.798
558924 93.469 -57.670 54.770
558925 93.501 -57.670 54.743
558926 93.535 -57.679 54.722
558927 93.570 -57.687 54.716
558928 93.604 -57.695 54.728
558929 93.642 -57.714 54.737
558930 93.680 -57.733 54.742
558931 93.718 -57.753 54.744
558932 93.757 -57.774 54.743
558933 93.795 -57.793 54.736
558934 93.833 -57.813 54.728
558935 93.871 -57.833 54.715
558936 93.910 -57.852 54.700
558937 93.948 -57.872 54.685
558938 93.986 -57.892 54.669
558939 94.025 -57.912 54.649
558940 94.063 -57.932 54.625
558941 94.101 -57.952 54.599
558942 94.139 -57.971 54.571
558943 94.178 -57.991 54.543
558944 94.216 -58.011 54.514
558945 94.254 -58.031 54.487
558946 94.292 -58.051 54.458
558947 94.320 -58.063 54.428
558948 94.346 -58.075 54.397
558949 94.373 -58.087 54.366
558950 94.400 -58.098 54.335
558951 94.426 -58.110 54.307
558952 94.455 -58.124 54.283
558953 94.484 -58.137 54.256
558954 94.512 -58.150 54.227
558955 94.541 -58.163 54.198
558956 94.569 -58.177 54.168
558957 94.594 -58.189 54.141
558958 94.619 -58.201 54.108
558959 94.644 -58.212 54.072
558960 94.670 -58.225 54.033
558961 94.695 -58.237 53.989
558962 94.720 -58.249 53.943
558963 94.746 -58.261 53.900
558964 94.771 -58.273 53.861
558965 94.796 -58.285 53.818
558966 94.821 -58.297 53.772
558967 94.862 -58.308 53.726
558968 94.918 -58.317 53.688
558969 94.974 -58.327 53.650
558970 95.030 -58.336 53.613
558971 95.086 -58.345 53.579
558972 95.141 -58.354 53.550
558973 95.197 -58.363 53.523
558974 95.253 -58.373 53.494
558975 95.309 -58.382 53.463
558976 95.355 -58.387 53.433
558977 95.393 -58.386 53.407
558978 95.423 -58.387 53.381
558979 95.453 -58.388 53.352
558980 95.482 -58.390 53.322
558981 95.512 -58.391 53.296
558982 95.542 -58.393 53.272
558983 95.573 -58.400 53.251
558984 95.605 -58.414 53.238
558985 95.637 -58.428 53.233
558986 95.670 -58.441 53.226
558987 95.702 -58.455 53.212
558988 95.734 -58.469 53.198
558989 95.766 -58.482 53.190
558990 95.799 -58.496 53.171
558991 95.831 -58.510 53.141
558992 95.863 -58.520 53.117
558993 95.893 -58.526 53.092
558994 95.923 -58.533 53.065
558995 95.953 -58.540 53.039
558996 95.983 -58.546 53.010
558997 96.014 -58.553 52.971
558998 96.046 -58.563 52.941
558999 96.079 -58.574 52.909
559000 96.112 -58.585 52.876
559001 96.145 -58.596 52.852
559002 96.178 -58.607 52.820
559003 96.210 -58.618 52.787
559004 96.243 -58.629 52.762
559005 96.276 -58.640 52.736
559006 96.308 -58.642 52.703
559007 96.338 -58.640 52.670
559008 96.370 -58.639 52.649
559009 96.401 -58.636 52.625
559010 96.432 -58.635 52.591
559011 96.463 -58.633 52.555
559012 96.495 -58.631 52.523
559013 96.526 -58.630 52.487
559014 96.561 -58.633 52.449
559015 96.596 -58.636 52.411
559016 96.630 -58.639 52.372
559017 96.662 -58.641 52.331
559018 96.693 -58.642 52.290
559019 96.725 -58.647 52.248
559020 96.757 -58.653 52.205
559021 96.789 -58.658 52.160
559022 96.821 -58.664 52.115
559023 96.853 -58.670 52.071
559024 96.885 -58.675 52.029
559025 96.917 -58.680 51.987
559026 96.949 -58.686 51.946
559027 96.981 -58.692 51.902
559028 97.014 -58.693 51.859
559029 97.048 -58.696 51.818
559030 97.081 -58.700 51.778
559031 97.114 -58.705 51.740
559032 97.147 -58.708 51.704
559033 97.180 -58.713 51.665
559034 97.213 -58.717 51.624
559035 97.245 -58.721 51.588
559036 97.278 -58.725 51.556
559037 97.311 -58.730 51.526
559038 97.343 -58.733 51.489
559039 97.376 -58.738 51.445
559040 97.408 -58.742 51.401
559041 97.441 -58.746 51.362
559042 97.473 -58.750 51.325
559043 97.506 -58.755 51.286
559044 97.542 -58.755 51.244
559045 97.574 -58.761 51.199
559046 97.608 -58.767 51.155
559047 97.644 -58.773 51.117
559048 97.679 -58.779 51.078
559049 97.717 -58.784 51.040
559050 97.756 -58.789 51.000
559051 97.796 -58.795 50.955
559052 97.835 -58.799 50.909
559053 97.873 -58.802 50.867
559054 97.911 -58.805 50.827
559055 97.952 -58.797 50.786
559056 97.994 -58.786 50.747
559057 98.035 -58.774 50.708
559058 98.077 -58.762 50.672
559059 98.118 -58.753 50.636
559060 98.150 -58.754 50.600
559061 98.183 -58.755 50.567
559062 98.217 -58.756 50.534
559063 98.252 -58.757 50.500
559064 98.288 -58.758 50.469
559065 98.323 -58.758 50.436
559066 98.358 -58.759 50.404
559067 98.393 -58.760 50.375
559068 98.428 -58.761 50.345
559069 98.459 -58.763 50.316
559070 98.491 -58.766 50.289
559071 98.524 -58.764 50.263
559072 98.558 -58.754 50.239
559073 98.588 -58.750 50.215
559074 98.617 -58.747 50.194
559075 98.645 -58.745 50.174
559076 98.674 -58.742 50.158
559077 98.703 -58.739 50.144
559078 98.732 -58.737 50.132
559079 98.765 -58.733 50.120
559080 98.799 -58.730 50.107
559081 98.834 -58.726 50.090
559082 98.868 -58.723 50.072
559083 98.903 -58.714 50.053
559084 98.941 -58.699 50.034
559085 98.988 -58.695 50.013
559086 99.034 -58.691 49.991
559087 99.074 -58.686 49.968
559088 99.115 -58.680 49.941
559089 99.154 -58.674 49.914
559090 99.195 -58.669 49.885
559091 99.239 -58.663 49.854
559092 99.311 -58.657 49.823
559093 99.378 -58.653 49.794
559094 99.414 -58.661 49.766
559095 99.449 -58.668 49.739
559096 99.485 -58.676 49.712
559097 99.520 -58.683 49.684
559098 99.556 -58.691 49.657
559099 99.592 -58.698 49.629
559100 99.628 -58.706 49.601
559101 99.663 -58.714 49.571
559102 99.699 -58.720 49.541
559103 99.734 -58.728 49.509
559104 99.770 -58.736 49.476
559105 99.806 -58.743 49.444
559450 86.970 -55.227 57.630
559451 87.005 -55.244 57.626
559452 87.040 -55.260 57.622
559453 87.075 -55.276 57.615
559454 87.110 -55.292 57.608
559455 87.145 -55.308 57.601
559456 87.180 -55.324 57.593
559457 87.215 -55.340 57.585
559458 87.250 -55.356 57.577
559459 87.285 -55.373 57.568
559460 87.320 -55.389 57.557
559461 87.355 -55.405 57.548
559462 87.390 -55.420 57.542
559463 87.425 -55.437 57.527
559464 87.457 -55.454 57.508
559465 87.500 -55.461 57.486
559466 87.556 -55.464 57.466
559467 87.612 -55.469 57.444
559468 87.668 -55.473 57.413
559469 87.715 -55.486 57.382
559470 87.753 -55.503 57.349
559471 87.791 -55.519 57.315
559472 87.829 -55.535 57.279
559473 87.867 -55.551 57.241
559474 87.906 -55.567 57.200
559475 87.944 -55.583 57.159
559476 87.982 -55.599 57.116
559477 88.021 -55.614 57.072
559478 88.059 -55.630 57.027
559479 88.097 -55.646 56.982
559480 88.135 -55.663 56.937
559481 88.173 -55.679 56.894
559482 88.211 -55.695 56.853
559483 88.250 -55.711 56.812
559506 89.130 -56.078 56.062
559507 89.168 -56.094 56.040
559508 89.206 -56.110 56.020
559509 89.244 -56.126 56.002
559510 89.282 -56.142 55.985
559511 89.321 -56.158 55.968
559512 89.359 -56.174 55.953
559513 89.397 -56.190 55.940
559514 89.436 -56.206 55.929
559515 89.474 -56.222 55.919
559516 89.512 -56.238 55.911
559517 89.550 -56.254 55.904
559518 89.589 -56.270 55.899
559519 89.627 -56.286 55.896
559520 89.656 -56.307 55.896
559521 89.696 -56.319 55.898
559522 89.738 -56.330 55.902
559523 89.779 -56.341 55.909
559524 89.821 -56.351 55.918
559525 89.859 -56.364 55.929
559526 89.884 -56.381 55.944
559527 89.910 -56.398 55.964
559528 89.936 -56.415 55.984
559529 89.961 -56.432 56.009
559530 89.987 -56.449 56.042
559531 90.013 -56.466 56.077
559532 90.038 -56.483 56.104
559533 90.068 -56.499 56.121
559534 90.111 -56.515 56.120
559535 90.155 -56.528 56.110
559536 90.198 -56.540 56.097
559537 90.241 -56.552 56.077
559538 90.285 -56.565 56.041
559539 90.328 -56.577 55.993
559540 90.371 -56.589 55.934
559541 90.414 -56.602 55.874
559542 90.458 -56.614 55.818
559543 90.501 -56.626 55.769
559544 90.544 -56.638 55.723
559545 90.588 -56.650 55.683
559546 90.632 -56.663 55.646
559547 90.675 -56.675 55.611
559548 90.718 -56.687 55.579
559549 90.761 -56.699 55.549
559550 90.805 -56.711 55.521
559551 90.848 -56.724 55.498
559552 90.891 -56.736 55.478
559553 90.932 -56.750 55.459
559554 90.967 -56.768 55.441
559555 91.003 -56.784 55.425
559556 91.038 -56.802 55.409
559557 91.073 -56.819 55.395
559558 91.108 -56.837 55.383
559559 91.144 -56.854 55.371
559560 91.179 -56.871 55.359
559561 91.214 -56.889 55.348
559562 91.250 -56.906 55.337
559563 91.285 -56.923 55.327
559564 91.320 -56.940 55.320
559565 91.352 -56.953 55.313
559566 91.384 -56.965 55.309
559567 91.417 -56.977 55.308
559568 91.449 -56.989 55.310
559569 91.482 -57.001 55.316
559570 91.519 -57.012 55.323
559571 91.562 -57.022 55.328
559572 91.604 -57.031 55.327
559573 91.647 -57.041 55.317
559574 91.690 -57.050 55.311
559575 91.733 -57.060 55.317
559576 91.776 -57.070 55.333
559577 91.819 -57.080 55.358
559578 91.862 -57.089 55.378
559579 91.905 -57.099 55.390
559580 91.948 -57.109 55.396
559581 91.991 -57.118 55.397
559582 92.034 -57.128 55.394
559583 92.075 -57.147 55.390
559584 92.117 -57.169 55.382
559585 92.157 -57.193 55.366
559586 92.199 -57.216 55.354
559587 92.239 -57.240 55.350
559588 92.281 -57.263 55.355
559589 92.321 -57.287 55.364
559590 92.363 -57.310 55.362
559591 92.403 -57.334 55.354
559592 92.440 -57.361 55.342
559593 92.477 -57.388 55.327
559594 92.513 -57.414 55.311
559595 92.550 -57.442 55.293
559596 92.586 -57.468 55.272
559597 92.623 -57.495 55.248
559598 92.658 -57.516 55.223
559599 92.692 -57.527 55.195
559600 92.724 -57.538 55.165
559601 92.758 -57.549 55.139
559602 92.790 -57.561 55.113
559603 92.824 -57.571 55.086
559604 92.856 -57.583 55.061
559605 92.889 -57.593 55.040
559606 92.921 -57.603 55.020
559607 92.949 -57.610 55.002
559608 92.978 -57.617 54.985
559609 93.004 -57.623 54.969
559610 93.028 -57.630 54.955
559611 93.053 -57.636 54.944
559612 93.077 -57.642 54.931
559613 93.105 -57.645 54.916
559614 93.136 -57.644 54.903
559615 93.167 -57.643 54.893
559616 93.199 -57.642 54.883
559617 93.230 -57.642 54.872
559618 93.262 -57.641 54.859
559619 93.293 -57.640 54.848
559620 93.324 -57.639 54.841
559621 93.355 -57.639 54.835
559622 93.387 -57.639 54.826
559623 93.418 -57.638 54.813
559624 93.449 -57.637 54.790
559625 93.481 -57.636 54.765
559626 93.512 -57.636 54.739
559627 93.544 -57.635 54.717
559628 93.578 -57.643 54.703
559629 93.613 -57.651 54.707
559630 93.651 -57.671 54.715
559631 93.689 -57.691 54.723
559632 93.727 -57.711 54.727
559633 93.766 -57.730 54.727
559634 93.804 -57.751 54.725
559635 93.842 -57.771 54.718
559636 93.880 -57.790 54.709
559637 93.919 -57.810 54.699
559638 93.957 -57.830 54.689
559639 93.995 -57.849 54.677
559640 94.034 -57.870 54.660
559641 94.072 -57.890 54.638
559642 94.110 -57.909 54.614
559643 94.148 -57.929 54.587
559644 94.187 -57.949 54.558
559645 94.225 -57.968 54.530
559646 94.263 -57.988 54.503
559647 94.301 -58.009 54.476
559648 94.340 -58.028 54.446
559649 94.368 -58.041 54.415
559650 94.395 -58.053 54.384
559651 94.421 -58.065 54.351
559652 94.449 -58.077 54.326
559653 94.478 -58.090 54.305
559654 94.506 -58.104 54.282
559655 94.534 -58.117 54.255
559656 94.559 -58.129 54.226
559657 94.584 -58.141 54.195
559658 94.610 -58.153 54.168
559659 94.635 -58.165 54.136
559660 94.660 -58.178 54.100
559661 94.685 -58.189 54.060
559662 94.710 -58.201 54.014
559663 94.736 -58.213 53.964
559664 94.761 -58.225 53.917
559665 94.786 -58.238 53.874
559666 94.812 -58.250 53.831
559667 94.837 -58.262 53.785
559668 94.865 -58.274 53.734
559669 94.921 -58.283 53.686
559670 94.977 -58.292 53.643
559671 95.033 -58.301 53.603
559672 95.089 -58.311 53.566
559673 95.145 -58.320 53.533
559674 95.201 -58.329 53.504
559675 95.257 -58.338 53.472
559676 95.313 -58.347 53.440
559677 95.359 -58.352 53.408
559678 95.396 -58.351 53.378
559679 95.432 -58.350 53.351
559680 95.462 -58.352 53.320
559681 95.492 -58.354 53.291
559682 95.522 -58.355 53.266
559683 95.552 -58.357 53.244
559684 95.582 -58.360 53.224
559685 95.614 -58.373 53.209
559686 95.646 -58.387 53.207
559687 95.678 -58.400 53.205
559688 95.710 -58.414 53.193
559689 95.743 -58.428 53.180
559690 95.775 -58.441 53.174
559691 95.808 -58.454 53.161
559692 95.840 -58.468 53.128
559693 95.870 -58.476 53.106
559694 95.901 -58.482 53.088
559695 95.931 -58.489 53.069
559696 95.961 -58.496 53.045
559697 95.991 -58.502 53.006
559698 96.024 -58.513 52.966
559699 96.056 -58.523 52.941
559700 96.089 -58.534 52.915
559701 96.122 -58.545 52.883
559702 96.155 -58.556 52.862
559703 96.188 -58.567 52.828
559704 96.221 -58.578 52.796
559705 96.253 -58.589 52.772
559706 96.286 -58.600 52.747
559707 96.319 -58.611 52.715
559708 96.350 -58.610 52.681
559709 96.381 -58.608 52.659
559710 96.412 -58.607 52.632
559711 96.444 -58.605 52.597
559712 96.475 -58.603 52.560
559713 96.506 -58.601 52.529
559714 96.537 -58.600 52.496
559715 96.568 -58.598 52.459
559716 96.599 -58.596 52.420
559717 96.633 -58.598 52.379
559718 96.667 -58.601 52.336
559719 96.698 -58.604 52.295
559720 96.731 -58.609 52.255
559721 96.762 -58.615 52.210
559722 96.795 -58.620 52.165
559723 96.826 -58.626 52.119
559724 96.859 -58.632 52.074
559725 96.890 -58.637 52.033
559726 96.923 -58.642 51.994
559727 96.954 -58.648 51.954
559728 96.986 -58.654 51.911
559729 97.019 -58.657 51.868
559730 97.052 -58.662 51.828
559731 97.085 -58.666 51.788
559732 97.117 -58.670 51.748
559733 97.150 -58.675 51.708
559734 97.183 -58.679 51.667
559735 97.216 -58.683 51.626
559736 97.248 -58.687 51.589
559737 97.281 -58.692 51.553
559738 97.314 -58.695 51.515
559739 97.347 -58.700 51.475
559740 97.379 -58.704 51.432
559741 97.412 -58.708 51.388
559742 97.444 -58.712 51.348
559743 97.477 -58.716 51.310
559744 97.513 -58.716 51.271
559745 97.548 -58.717 51.230
559746 97.580 -58.723 51.186
559747 97.612 -58.729 51.144
559748 97.647 -58.735 51.102
559749 97.682 -58.741 51.061
559750 97.722 -58.745 51.020
559751 97.761 -58.751 50.979
559752 97.800 -58.753 50.936
559753 97.838 -58.756 50.892
559754 97.876 -58.759 50.850
559755 97.914 -58.761 50.810
559756 97.953 -58.761 50.773
559757 97.995 -58.749 50.736
559758 98.036 -58.738 50.699
559759 98.078 -58.726 50.663
559760 98.116 -58.720 50.626
559761 98.149 -58.721 50.593
559762 98.181 -58.722 50.559
559763 98.214 -58.723 50.525
559764 98.249 -58.723 50.492
559765 98.285 -58.724 50.459
559766 98.319 -58.725 50.427
559767 98.355 -58.726 50.396
559768 98.390 -58.727 50.365
559769 98.425 -58.728 50.334
559770 98.460 -58.729 50.304
559771 98.496 -58.730 50.277
559772 98.531 -58.730 50.250
559773 98.561 -58.728 50.223
559774 98.589 -58.726 50.199
559775 98.618 -58.723 50.178
559776 98.647 -58.720 50.158
559777 98.676 -58.718 50.139
559778 98.705 -58.715 50.125
559779 98.734 -58.713 50.112
559780 98.769 -58.709 50.100
559781 98.803 -58.705 50.087
559782 98.838 -58.700 50.072
559783 98.873 -58.683 50.056
559784 98.909 -58.667 50.038
559785 98.949 -58.655 50.020
559786 98.996 -58.651 50.001
559787 99.044 -58.648 49.980
559788 99.090 -58.645 49.959
559789 99.137 -58.641 49.935
559790 99.178 -58.635 49.909
559791 99.218 -58.629 49.881
559792 99.258 -58.623 49.852
559793 99.319 -58.618 49.823
559794 99.379 -58.617 49.792
559795 99.414 -58.624 49.765
559796 99.450 -58.632 49.739
559797 99.486 -58.639 49.712
559798 99.521 -58.647 49.685
559799 99.557 -58.654 49.659
559800 99.593 -58.661 49.632
559801 99.629 -58.669 49.604
559802 99.664 -58.676 49.575
559803 99.700 -58.684 49.545
559804 99.735 -58.692 49.515
559805 99.771 -58.699 49.483
559806 99.807 -58.707 49.451
560151 87.000 -55.177 57.629
560152 87.035 -55.194 57.625
560153 87.070 -55.210 57.621
560154 87.105 -55.226 57.616
560155 87.140 -55.242 57.609
560156 87.175 -55.258 57.603
560157 87.210 -55.274 57.595
560158 87.245 -55.290 57.589
560159 87.280 -55.306 57.582
560160 87.315 -55.323 57.573
560161 87.350 -55.339 57.563
560162 87.385 -55.354 57.556
560163 87.416 -55.371 57.551
560164 87.446 -55.389 57.538
560165 87.477 -55.406 57.522
560166 87.511 -55.419 57.506
560167 87.561 -55.416 57.489
560168 87.616 -55.420 57.471
560169 87.677 -55.429 57.440
560170 87.734 -55.443 57.406
560171 87.775 -55.460 57.373
560172 87.813 -55.477 57.341
560173 87.850 -55.493 57.305
560174 87.888 -55.510 57.268
560175 87.925 -55.527 57.229
560176 87.963 -55.542 57.187
560177 88.001 -55.559 57.144
560178 88.038 -55.576 57.101
560179 88.076 -55.593 57.055
560180 88.115 -55.608 57.010
560181 88.153 -55.624 56.964
560182 88.191 -55.640 56.920
560183 88.229 -55.656 56.877
560184 88.268 -55.672 56.836
560185 88.306 -55.688 56.794
560208 89.186 -56.056 56.040
560209 89.224 -56.072 56.019
560210 89.262 -56.088 56.000
560211 89.301 -56.103 55.982
560212 89.339 -56.119 55.965
560213 89.377 -56.135 55.949
560214 89.415 -56.151 55.936
560215 89.454 -56.168 55.925
560216 89.492 -56.184 55.914
560217 89.530 -56.200 55.905
560218 89.568 -56.216 55.898
560219 89.606 -56.232 55.892
560220 89.645 -56.248 55.889
560221 89.675 -56.268 55.888
560222 89.716 -56.279 55.890
560223 89.757 -56.290 55.894
560224 89.799 -56.301 55.899
560225 89.840 -56.312 55.907
560226 89.866 -56.329 55.917
560227 89.892 -56.346 55.930
560228 89.917 -56.363 55.949
560229 89.943 -56.380 55.969
560230 89.969 -56.397 55.997
560231 89.994 -56.414 56.034
560232 90.020 -56.431 56.068
560233 90.045 -56.448 56.093
560234 90.083 -56.463 56.111
560235 90.126 -56.475 56.108
560236 90.169 -56.487 56.095
560237 90.213 -56.500 56.077
560238 90.256 -56.512 56.052
560239 90.299 -56.524 56.014
560240 90.342 -56.536 55.966
560241 90.386 -56.548 55.915
560242 90.429 -56.561 55.862
560243 90.472 -56.573 55.811
560244 90.516 -56.585 55.765
560245 90.559 -56.598 55.722
560246 90.602 -56.610 55.684
560247 90.646 -56.622 55.649
560248 90.689 -56.634 55.616
560249 90.733 -56.646 55.585
560250 90.776 -56.658 55.558
560251 90.819 -56.671 55.531
560252 90.863 -56.683 55.509
560253 90.906 -56.696 55.489
560254 90.949 -56.708 55.471
560255 90.988 -56.723 55.454
560256 91.023 -56.740 55.439
560257 91.058 -56.758 55.425
560258 91.093 -56.775 55.411
560259 91.129 -56.792 55.398
560260 91.164 -56.809 55.386
560261 91.200 -56.827 55.374
560262 91.235 -56.844 55.361
560263 91.270 -56.862 55.348
560264 91.305 -56.879 55.338
560265 91.340 -56.894 55.329
560266 91.372 -56.906 55.321
560267 91.404 -56.918 55.315
560268 91.436 -56.931 55.312
560269 91.468 -56.943 55.311
560270 91.501 -56.956 55.314
560271 91.533 -56.968 55.318
560272 91.574 -56.978 55.321
560273 91.616 -56.987 55.320
560274 91.659 -56.997 55.316
560275 91.702 -57.006 55.315
560276 91.745 -57.016 55.322
560277 91.788 -57.026 55.337
560278 91.831 -57.035 55.357
560279 91.874 -57.045 55.374
560280 91.917 -57.055 55.383
560281 91.960 -57.064 55.391
560282 92.003 -57.074 55.398
560283 92.045 -57.088 55.405
560284 92.086 -57.112 55.410
560285 92.127 -57.135 55.408
560286 92.168 -57.159 55.399
560287 92.209 -57.182 55.385
560288 92.250 -57.206 55.371
560289 92.291 -57.229 55.362
560290 92.333 -57.252 55.369
560291 92.373 -57.275 55.366
560292 92.415 -57.299 55.355
560293 92.451 -57.326 55.342
560294 92.488 -57.353 55.327
560295 92.524 -57.379 55.312
560296 92.561 -57.406 55.298
560297 92.598 -57.433 55.282
560298 92.634 -57.459 55.262
560299 92.668 -57.479 55.237
560300 92.702 -57.494 55.210
560301 92.735 -57.505 55.181
560302 92.768 -57.517 55.154
560303 92.801 -57.527 55.128
560304 92.834 -57.539 55.101
560305 92.867 -57.549 55.073
560306 92.900 -57.561 55.048
560307 92.929 -57.568 55.023
560308 92.958 -57.576 55.001
560309 92.986 -57.583 54.983
560310 93.015 -57.590 54.967
560311 93.041 -57.596 54.952
560312 93.065 -57.602 54.937
560313 93.089 -57.608 54.924
560314 93.117 -57.611 54.907
560315 93.148 -57.611 54.891
560316 93.179 -57.610 54.877
560317 93.211 -57.609 54.865
560318 93.242 -57.608 54.851
560319 93.273 -57.608 54.837
560320 93.304 -57.607 54.828
560321 93.336 -57.606 54.819
560322 93.367 -57.605 54.811
560323 93.399 -57.605 54.806
560324 93.430 -57.604 54.798
560325 93.462 -57.603 54.779
560326 93.493 -57.602 54.756
560327 93.524 -57.601 54.733
560328 93.555 -57.601 54.711
560329 93.587 -57.600 54.693
560330 93.621 -57.609 54.688
560331 93.659 -57.629 54.693
560332 93.698 -57.648 54.699
560333 93.736 -57.668 54.704
560334 93.774 -57.688 54.705
560335 93.813 -57.708 54.706
560336 93.851 -57.728 54.702
560337 93.889 -57.748 54.695
560338 93.927 -57.767 54.688
560339 93.966 -57.787 54.680
560340 94.004 -57.807 54.671
560341 94.042 -57.827 54.659
560342 94.080 -57.847 54.642
560343 94.119 -57.867 54.623
560344 94.157 -57.887 54.600
560345 94.195 -57.906 54.575
560346 94.234 -57.926 54.549
560347 94.272 -57.946 54.521
560348 94.310 -57.966 54.494
560349 94.348 -57.986 54.464
560350 94.387 -58.006 54.434
560351 94.416 -58.019 54.404
560352 94.443 -58.031 54.375
560353 94.472 -58.044 54.350
560354 94.499 -58.057 54.328
560355 94.524 -58.069 54.306
560356 94.549 -58.081 54.282
560357 94.575 -58.093 54.253
560358 94.600 -58.106 54.221
560359 94.625 -58.118 54.191
560360 94.650 -58.130 54.159
560361 94.676 -58.142 54.124
560362 94.701 -58.154 54.082
560363 94.726 -58.166 54.035
560364 94.752 -58.178 53.985
560365 94.777 -58.190 53.934
560366 94.802 -58.202 53.886
560367 94.827 -58.214 53.841
560368 94.852 -58.226 53.795
560369 94.878 -58.238 53.742
560370 94.925 -58.248 53.688
560371 94.981 -58.258 53.640
560372 95.037 -58.267 53.597
560373 95.093 -58.276 53.557
560374 95.149 -58.285 53.519
560375 95.205 -58.295 53.485
560376 95.261 -58.304 53.452
560377 95.317 -58.313 53.418
560378 95.363 -58.317 53.384
560379 95.400 -58.317 53.351
560380 95.437 -58.316 53.323
560381 95.472 -58.316 53.293
560382 95.502 -58.317 53.265
560383 95.532 -58.319 53.239
560384 95.561 -58.320 53.217
560385 95.591 -58.322 53.200
560386 95.622 -58.332 53.189
560387 95.655 -58.345 53.186
560388 95.687 -58.359 53.179
560389 95.720 -58.372 53.167
560390 95.752 -58.386 53.155
560391 95.784 -58.400 53.148
560392 95.816 -58.413 53.142
560393 95.848 -58.426 53.114
560394 95.878 -58.432 53.091
560395 95.908 -58.438 53.072
560396 95.939 -58.445 53.054
560397 95.969 -58.452 53.033
560398 96.001 -58.461 53.002
560399 96.034 -58.473 52.968
560400 96.066 -58.483 52.945
560401 96.099 -58.494 52.922
560402 96.132 -58.505 52.893
560403 96.165 -58.516 52.868
560404 96.198 -58.527 52.833
560405 96.231 -58.538 52.804
560406 96.263 -58.549 52.779
560407 96.296 -58.560 52.755
560408 96.329 -58.571 52.728
560409 96.362 -58.580 52.694
560410 96.393 -58.579 52.671
560411 96.424 -58.577 52.641
560412 96.455 -58.576 52.604
560413 96.486 -58.573 52.567
560414 96.517 -58.572 52.535
560415 96.548 -58.570 52.504
560416 96.580 -58.568 52.468
560417 96.611 -58.567 52.429
560418 96.642 -58.565 52.386
560419 96.673 -58.563 52.341
560420 96.705 -58.567 52.300
560421 96.737 -58.573 52.260
560422 96.769 -58.579 52.214
560423 96.802 -58.585 52.169
560424 96.834 -58.592 52.123
560425 96.867 -58.598 52.078
560426 96.899 -58.604 52.039
560427 96.932 -58.610 52.001
560428 96.964 -58.616 51.963
560429 96.996 -58.622 51.922
560430 97.025 -58.625 51.881
560431 97.055 -58.628 51.842
560432 97.088 -58.632 51.800
560433 97.120 -58.636 51.759
560434 97.153 -58.641 51.717
560435 97.186 -58.645 51.675
560436 97.219 -58.649 51.636
560437 97.251 -58.653 51.599
560438 97.284 -58.657 51.559
560439 97.317 -58.662 51.516
560440 97.350 -58.666 51.473
560441 97.382 -58.670 51.429
560442 97.415 -58.674 51.385
560443 97.448 -58.678 51.343
560444 97.484 -58.678 51.303
560445 97.519 -58.677 51.263
560446 97.555 -58.679 51.221
560447 97.586 -58.685 51.178
560448 97.617 -58.691 51.136
560449 97.650 -58.697 51.092
560450 97.687 -58.702 51.050
560451 97.726 -58.705 51.009
560452 97.764 -58.708 50.966
560453 97.802 -58.711 50.923
560454 97.840 -58.713 50.882
560455 97.878 -58.716 50.841
560456 97.916 -58.718 50.802
560457 97.954 -58.721 50.765
560458 97.996 -58.713 50.728
560459 98.037 -58.701 50.692
560460 98.079 -58.689 50.656
560461 98.114 -58.687 50.622
560462 98.147 -58.688 50.589
560463 98.180 -58.689 50.555
560464 98.213 -58.690 50.521
560465 98.246 -58.691 50.487
560466 98.281 -58.692 50.454
560467 98.316 -58.692 50.422
560468 98.352 -58.693 50.390
560469 98.387 -58.694 50.358
560470 98.422 -58.695 50.327
560471 98.457 -58.695 50.296
560472 98.492 -58.696 50.267
560473 98.526 -58.698 50.240
560474 98.557 -58.698 50.213
560475 98.589 -58.699 50.188
560476 98.620 -58.699 50.166
560477 98.649 -58.696 50.145
560478 98.677 -58.694 50.126
560479 98.706 -58.692 50.110
560480 98.738 -58.688 50.095
560481 98.772 -58.685 50.081
560482 98.808 -58.670 50.069
560483 98.843 -58.654 50.054
560484 98.879 -58.637 50.039
560485 98.915 -58.620 50.023
560486 98.957 -58.612 50.006
560487 99.005 -58.608 49.988
560488 99.052 -58.604 49.969
560489 99.098 -58.601 49.949
560490 99.146 -58.597 49.926
560491 99.193 -58.593 49.902
560492 99.239 -58.590 49.876
560493 99.281 -58.585 49.849
560494 99.327 -58.579 49.821
560495 99.379 -58.580 49.790
560496 99.415 -58.588 49.763
560497 99.451 -58.595 49.736
560498 99.487 -58.602 49.710
560499 99.522 -58.610 49.685
560500 99.558 -58.617 49.658
560501 99.593 -58.625 49.632
560502 99.628 -58.633 49.605
560503 99.663 -58.642 49.577
560504 99.697 -58.651 49.548
560505 99.732 -58.660 49.518
560506 99.766 -58.668 49.487
560507 99.801 -58.677 49.456
560852 87.030 -55.127 57.626
560853 87.065 -55.144 57.624
560854 87.100 -55.160 57.620
560855 87.135 -55.176 57.616
560856 87.170 -55.191 57.609
560857 87.205 -55.208 57.603
560858 87.240 -55.224 57.597
560859 87.275 -55.240 57.591
560860 87.310 -55.256 57.585
560861 87.345 -55.273 57.577
560862 87.375 -55.289 57.568
560863 87.406 -55.307 57.562
560864 87.436 -55.324 57.554
560865 87.466 -55.341 57.543
560866 87.496 -55.358 57.532
560867 87.527 -55.376 57.522
560868 87.573 -55.376 57.509
560869 87.625 -55.376 57.492
560870 87.687 -55.386 57.463
560871 87.745 -55.401 57.429
560872 87.798 -55.417 57.396
560873 87.836 -55.433 57.365
560874 87.873 -55.450 57.331
560875 87.911 -55.467 57.294
560876 87.948 -55.483 57.255
560877 87.986 -55.500 57.215
560878 88.024 -55.516 57.172
560879 88.061 -55.533 57.129
560880 88.099 -55.549 57.083
560881 88.136 -55.566 57.037
560882 88.174 -55.583 56.992
560883 88.211 -55.599 56.946
560884 88.249 -55.616 56.902
560885 88.287 -55.633 56.860
560886 88.324 -55.649 56.817
560909 89.203 -56.017 56.039
560910 89.242 -56.033 56.018
560911 89.280 -56.049 55.998
560912 89.318 -56.065 55.980
560913 89.356 -56.081 55.962
560914 89.395 -56.097 55.947
560915 89.433 -56.113 55.933
560916 89.471 -56.129 55.920
560917 89.510 -56.145 55.909
560918 89.548 -56.161 55.900
560919 89.586 -56.177 55.892
560920 89.624 -56.193 55.887
560921 89.663 -56.209 55.883
560922 89.694 -56.229 55.881
560923 89.735 -56.240 55.882
560924 89.777 -56.251 55.885
560925 89.818 -56.261 55.888
560926 89.848 -56.277 55.895
560927 89.874 -56.294 55.904
560928 89.899 -56.311 55.916
560929 89.924 -56.328 55.933
560930 89.950 -56.345 55.953
560931 89.976 -56.362 55.980
560932 90.002 -56.379 56.013
560933 90.029 -56.398 56.043
560934 90.061 -56.413 56.064
560935 90.096 -56.425 56.077
560936 90.140 -56.436 56.076
560937 90.184 -56.447 56.063
560938 90.227 -56.459 56.043
560939 90.270 -56.471 56.019
560940 90.314 -56.483 55.983
560941 90.357 -56.495 55.939
560942 90.400 -56.508 55.894
560943 90.444 -56.520 55.847
560944 90.487 -56.533 55.802
560945 90.530 -56.545 55.759
560946 90.574 -56.557 55.721
560947 90.617 -56.569 55.685
560948 90.660 -56.581 55.652
560949 90.704 -56.593 55.621
560950 90.747 -56.605 55.593
560951 90.791 -56.617 55.566
560952 90.834 -56.630 55.541
560953 90.877 -56.642 55.520
560954 90.921 -56.655 55.502
560955 90.964 -56.667 55.485
560956 91.007 -56.679 55.469
560957 91.043 -56.696 55.456
560958 91.078 -56.713 55.444
560959 91.113 -56.730 55.431
560960 91.149 -56.748 55.418
560961 91.184 -56.765 55.404
560962 91.220 -56.783 55.390
560963 91.255 -56.799 55.376
560964 91.290 -56.817 55.362
560965 91.325 -56.834 55.350
560966 91.359 -56.848 55.340
560967 91.391 -56.860 55.330
560968 91.423 -56.873 55.323
560969 91.456 -56.885 55.316
560970 91.488 -56.897 55.312
560971 91.520 -56.909 55.311
560972 91.553 -56.921 55.311
560973 91.585 -56.934 55.313
560974 91.629 -56.943 55.315
560975 91.671 -56.953 55.317
560976 91.714 -56.962 55.322
560977 91.757 -56.972 55.331
560978 91.800 -56.982 55.344
560979 91.843 -56.991 55.364
560980 91.886 -57.001 55.380
560981 91.929 -57.011 55.388
560982 91.972 -57.020 55.393
560983 92.015 -57.031 55.401
560984 92.056 -57.054 55.411
560985 92.097 -57.078 55.419
560986 92.138 -57.100 55.421
560987 92.179 -57.124 55.415
560988 92.220 -57.147 55.404
560989 92.261 -57.171 55.389
560990 92.302 -57.194 55.375
560991 92.343 -57.218 55.373
560992 92.384 -57.241 55.365
560993 92.425 -57.265 55.352
560994 92.462 -57.291 55.336
560995 92.499 -57.317 55.320
560996 92.536 -57.344 55.305
560997 92.573 -57.371 55.294
560998 92.609 -57.398 55.282
560999 92.643 -57.419 55.265
561000 92.678 -57.440 55.244
561001 92.712 -57.461 55.222
561002 92.745 -57.473 55.200
561003 92.778 -57.483 55.175
561004 92.811 -57.495 55.148
561005 92.844 -57.505 55.119
561006 92.877 -57.517 55.089
561007 92.910 -57.527 55.059
561008 92.938 -57.534 55.031
561009 92.967 -57.542 55.007
561010 92.995 -57.548 54.987
561011 93.023 -57.555 54.968
561012 93.052 -57.563 54.951
561013 93.076 -57.569 54.933
561014 93.100 -57.574 54.917
561015 93.128 -57.577 54.898
561016 93.160 -57.576 54.880
561017 93.191 -57.576 54.864
561018 93.222 -57.575 54.850
561019 93.254 -57.574 54.837
561020 93.285 -57.573 54.822
561021 93.317 -57.573 54.811
561022 93.348 -57.572 54.802
561023 93.379 -57.571 54.793
561024 93.411 -57.570 54.787
561025 93.442 -57.570 54.778
561026 93.473 -57.569 54.762
561027 93.504 -57.568 54.742
561028 93.536 -57.568 54.720
561029 93.567 -57.567 54.700
561030 93.600 -57.570 54.684
561031 93.632 -57.572 54.673
561032 93.670 -57.589 54.670
561033 93.708 -57.608 54.673
561034 93.746 -57.628 54.677
561035 93.784 -57.648 54.678
561036 93.823 -57.667 54.681
561037 93.861 -57.687 54.680
561038 93.899 -57.707 54.676
561039 93.937 -57.727 54.671
561040 93.975 -57.746 54.666
561041 94.013 -57.766 54.660
561042 94.052 -57.786 54.653
561043 94.090 -57.805 54.641
561044 94.128 -57.825 54.627
561045 94.166 -57.845 54.608
561046 94.204 -57.864 54.586
561047 94.243 -57.883 54.562
561048 94.281 -57.903 54.535
561049 94.319 -57.923 54.507
561050 94.357 -57.943 54.479
561051 94.396 -57.963 54.450
561052 94.434 -57.983 54.423
561053 94.464 -57.997 54.396
561054 94.489 -58.009 54.371
561055 94.514 -58.021 54.348
561056 94.540 -58.034 54.325
561057 94.565 -58.046 54.302
561058 94.591 -58.058 54.274
561059 94.616 -58.070 54.243
561060 94.641 -58.082 54.211
561061 94.666 -58.094 54.179
561062 94.691 -58.106 54.143
561063 94.716 -58.118 54.102
561064 94.742 -58.131 54.056
561065 94.767 -58.143 54.004
561066 94.792 -58.155 53.952
561067 94.818 -58.166 53.900
561068 94.843 -58.178 53.848
561069 94.868 -58.191 53.797
561070 94.893 -58.203 53.747
561071 94.929 -58.214 53.693
561072 94.985 -58.223 53.643
561073 95.041 -58.232 53.595
561074 95.097 -58.242 53.549
561075 95.153 -58.251 53.506
561076 95.208 -58.260 53.469
561077 95.264 -58.269 53.434
561078 95.320 -58.279 53.397
561079 95.367 -58.283 53.361
561080 95.403 -58.282 53.327
561081 95.440 -58.281 53.297
561082 95.477 -58.281 53.268
561083 95.512 -58.281 53.241
561084 95.541 -58.282 53.214
561085 95.571 -58.284 53.191
561086 95.601 -58.285 53.175
561087 95.631 -58.291 53.168
561088 95.664 -58.304 53.161
561089 95.696 -58.318 53.151
561090 95.728 -58.332 53.141
561091 95.760 -58.345 53.131
561092 95.793 -58.359 53.121
561093 95.825 -58.372 53.118
561094 95.856 -58.382 53.097
561095 95.886 -58.388 53.078
561096 95.916 -58.394 53.059
561097 95.947 -58.401 53.040
561098 95.978 -58.410 53.024
561099 96.011 -58.421 52.995
561100 96.044 -58.432 52.968
561101 96.077 -58.443 52.947
561102 96.109 -58.454 52.924
561103 96.142 -58.465 52.892
561104 96.175 -58.476 52.865
561105 96.208 -58.487 52.833
561106 96.241 -58.498 52.805
561107 96.273 -58.509 52.782
561108 96.306 -58.520 52.760
561109 96.339 -58.531 52.734
561110 96.372 -58.542 52.700
561111 96.404 -58.549 52.675
561112 96.435 -58.548 52.648
561113 96.466 -58.545 52.612
561114 96.498 -58.544 52.572
561115 96.529 -58.542 52.538
561116 96.560 -58.540 52.499
561117 96.591 -58.538 52.463
561118 96.622 -58.537 52.425
561119 96.653 -58.535 52.386
561120 96.684 -58.533 52.345
561121 96.717 -58.538 52.304
561122 96.749 -58.544 52.262
561123 96.781 -58.550 52.217
561124 96.813 -58.556 52.172
561125 96.846 -58.562 52.128
561126 96.878 -58.568 52.083
561127 96.910 -58.574 52.040
561128 96.943 -58.580 52.002
561129 96.975 -58.586 51.966
561130 97.005 -58.590 51.929
561131 97.034 -58.592 51.891
561132 97.063 -58.595 51.851
561133 97.091 -58.598 51.809
561134 97.124 -58.603 51.769
561135 97.157 -58.607 51.729
561136 97.189 -58.611 51.691
561137 97.222 -58.615 51.654
561138 97.255 -58.620 51.618
561139 97.288 -58.623 51.575
561140 97.320 -58.628 51.529
561141 97.353 -58.632 51.483
561142 97.385 -58.636 51.437
561143 97.419 -58.639 51.392
561144 97.454 -58.639 51.346
561145 97.490 -58.639 51.303
561146 97.526 -58.639 51.261
561147 97.561 -58.641 51.218
561148 97.593 -58.647 51.174
561149 97.624 -58.653 51.131
561150 97.655 -58.659 51.088
561151 97.690 -58.661 51.046
561152 97.728 -58.662 51.003
561153 97.767 -58.665 50.960
561154 97.805 -58.667 50.917
561155 97.843 -58.670 50.876
561156 97.881 -58.673 50.836
561157 97.919 -58.676 50.798
561158 97.957 -58.678 50.761
561159 97.996 -58.676 50.724
561160 98.038 -58.664 50.688
561161 98.080 -58.653 50.652
561162 98.112 -58.654 50.620
561163 98.145 -58.655 50.586
561164 98.178 -58.656 50.553
561165 98.211 -58.657 50.519
561166 98.244 -58.657 50.485
561167 98.278 -58.658 50.452
561168 98.313 -58.659 50.420
561169 98.348 -58.661 50.388
561170 98.383 -58.661 50.355
561171 98.419 -58.662 50.323
561172 98.454 -58.663 50.292
561173 98.488 -58.664 50.261
561174 98.520 -58.665 50.232
561175 98.552 -58.666 50.204
561176 98.583 -58.667 50.179
561177 98.615 -58.667 50.156
561178 98.646 -58.667 50.135
561179 98.678 -58.668 50.116
561180 98.708 -58.667 50.099
561181 98.742 -58.657 50.082
561182 98.778 -58.641 50.066
561183 98.813 -58.624 50.053
561184 98.849 -58.607 50.039
561185 98.885 -58.591 50.024
561186 98.920 -58.574 50.009
561187 98.966 -58.569 49.992
561188 99.013 -58.565 49.975
561189 99.060 -58.561 49.957
561190 99.107 -58.557 49.938
561191 99.154 -58.554 49.917
561192 99.201 -58.550 49.894
561193 99.248 -58.546 49.870
561194 99.295 -58.542 49.844
561195 99.342 -58.538 49.817
561196 99.379 -58.545 49.787
561197 99.414 -58.553 49.760
561198 99.448 -58.562 49.733
561199 99.483 -58.570 49.708
561200 99.517 -58.579 49.683
561201 99.552 -58.588 49.657
561202 99.586 -58.597 49.632
561203 99.621 -58.606 49.606
561204 99.655 -58.614 49.579
561205 99.690 -58.623 49.551
561206 99.724 -58.632 49.521
561207 99.759 -58.641 49.491
561208 99.793 -58.649 49.460
561554 87.095 -55.094 57.622
561555 87.130 -55.110 57.620
561556 87.165 -55.126 57.616
561557 87.200 -55.141 57.609
561558 87.235 -55.158 57.604
561559 87.270 -55.174 57.599
561560 87.304 -55.190 57.593
561561 87.335 -55.207 57.587
561562 87.365 -55.225 57.580
561563 87.395 -55.242 57.572
561564 87.425 -55.259 57.564
561565 87.456 -55.276 57.557
561566 87.486 -55.293 57.548
561567 87.516 -55.311 57.539
561568 87.547 -55.328 57.532
561569 87.586 -55.336 57.522
561570 87.635 -55.333 57.507
561571 87.697 -55.344 57.482
561572 87.755 -55.358 57.451
561573 87.809 -55.373 57.417
561574 87.858 -55.390 57.387
561575 87.896 -55.407 57.354
561576 87.933 -55.423 57.319
561577 87.971 -55.440 57.281
561578 88.008 -55.457 57.242
561579 88.046 -55.473 57.201
561580 88.084 -55.489 57.158
561581 88.121 -55.506 57.112
561582 88.159 -55.523 57.066
561583 88.197 -55.539 57.019
561584 88.234 -55.556 56.973
561585 88.272 -55.573 56.929
561586 88.310 -55.589 56.885
561587 88.347 -55.606 56.842
561588 88.385 -55.622 56.799
561610 89.222 -55.978 56.038
561611 89.260 -55.994 56.017
561612 89.298 -56.010 55.996
561613 89.336 -56.027 55.977
561614 89.374 -56.043 55.960
561615 89.412 -56.059 55.944
561616 89.451 -56.075 55.929
561617 89.489 -56.091 55.916
561618 89.527 -56.107 55.904
561619 89.565 -56.122 55.895
561620 89.604 -56.138 55.887
561621 89.642 -56.154 55.881
561622 89.680 -56.170 55.876
561623 89.713 -56.189 55.874
561624 89.755 -56.200 55.874
561625 89.796 -56.211 55.876
561626 89.830 -56.225 55.880
561627 89.856 -56.242 55.885
561628 89.881 -56.259 55.893
561629 89.906 -56.276 55.903
561630 89.932 -56.293 55.918
561631 89.958 -56.310 55.936
561632 89.986 -56.329 55.959
561633 90.015 -56.347 55.984
561634 90.046 -56.363 56.008
561635 90.081 -56.375 56.027
561636 90.116 -56.388 56.037
561637 90.152 -56.400 56.038
561638 90.196 -56.410 56.026
561639 90.240 -56.420 56.007
561640 90.285 -56.431 55.987
561641 90.328 -56.442 55.955
561642 90.372 -56.454 55.916
561643 90.415 -56.467 55.876
561644 90.458 -56.479 55.835
561645 90.502 -56.492 55.793
561646 90.545 -56.504 55.755
561647 90.588 -56.516 55.719
561648 90.632 -56.528 55.687
561649 90.675 -56.540 55.657
561650 90.718 -56.552 55.628
561651 90.761 -56.565 55.600
561652 90.805 -56.577 55.574
561653 90.849 -56.589 55.551
561654 90.892 -56.602 55.532
561655 90.935 -56.614 55.515
561656 90.978 -56.626 55.500
561657 91.022 -56.638 55.485
561658 91.063 -56.652 55.473
561659 91.099 -56.669 55.464
561660 91.134 -56.686 55.454
561661 91.169 -56.703 55.441
561662 91.205 -56.721 55.426
561663 91.240 -56.738 55.410
561664 91.275 -56.755 55.393
561665 91.311 -56.773 55.377
561666 91.346 -56.790 55.363
561667 91.378 -56.802 55.350
561668 91.411 -56.815 55.340
561669 91.443 -56.827 55.331
561670 91.475 -56.839 55.324
561671 91.507 -56.851 55.319
561672 91.539 -56.864 55.316
561673 91.572 -56.876 55.316
561674 91.604 -56.888 55.319
561675 91.640 -56.899 55.327
561676 91.683 -56.909 55.331
561677 91.726 -56.918 55.336
561678 91.769 -56.928 55.345
561679 91.812 -56.938 55.361
561680 91.855 -56.947 55.383
561681 91.898 -56.957 55.398
561682 91.941 -56.967 55.403
561683 91.984 -56.976 55.404
561684 92.025 -56.996 55.408
561685 92.066 -57.019 55.417
561686 92.107 -57.043 55.424
561687 92.148 -57.066 55.426
561688 92.190 -57.090 55.423
561689 92.230 -57.113 55.415
561690 92.272 -57.137 55.402
561691 92.312 -57.160 55.388
561692 92.354 -57.183 55.381
561693 92.394 -57.206 55.371
561694 92.436 -57.230 55.359
561695 92.474 -57.256 55.345
561696 92.511 -57.282 55.329
561697 92.547 -57.309 55.311
561698 92.584 -57.336 55.294
561699 92.619 -57.357 55.277
561700 92.653 -57.379 55.259
561701 92.687 -57.400 55.240
561702 92.722 -57.421 55.222
561703 92.756 -57.439 55.206
561704 92.789 -57.451 55.185
561705 92.821 -57.462 55.160
561706 92.855 -57.473 55.132
561707 92.888 -57.484 55.102
561708 92.918 -57.493 55.073
561709 92.947 -57.500 55.048
561710 92.975 -57.507 55.022
561711 93.004 -57.514 54.996
561712 93.032 -57.521 54.973
561713 93.061 -57.528 54.953
561714 93.088 -57.535 54.934
561715 93.112 -57.541 54.912
561716 93.140 -57.543 54.892
561717 93.171 -57.542 54.874
561718 93.203 -57.542 54.857
561719 93.234 -57.541 54.841
561720 93.266 -57.540 54.827
561721 93.297 -57.540 54.812
561722 93.329 -57.539 54.798
561723 93.360 -57.539 54.788
561724 93.391 -57.538 54.778
561725 93.422 -57.537 54.769
561726 93.454 -57.537 54.760
561727 93.487 -57.539 54.745
561728 93.519 -57.542 54.727
561729 93.552 -57.544 54.706
561730 93.584 -57.547 54.685
561731 93.617 -57.549 54.669
561732 93.649 -57.551 54.655
561733 93.682 -57.554 54.646
561734 93.720 -57.573 54.645
561735 93.758 -57.592 54.648
561736 93.796 -57.612 54.651
561737 93.834 -57.632 54.654
561738 93.872 -57.651 54.656
561739 93.911 -57.671 54.653
561740 93.949 -57.691 54.649
561741 93.987 -57.711 54.645
561742 94.026 -57.730 54.644
561743 94.063 -57.750 54.639
561744 94.102 -57.770 54.635
561745 94.140 -57.789 54.625
561746 94.178 -57.809 54.610
561747 94.216 -57.828 54.591
561748 94.255 -57.848 54.569
561749 94.292 -57.868 54.544
561750 94.331 -57.887 54.517
561751 94.369 -57.907 54.490
561752 94.407 -57.927 54.463
561753 94.445 -57.950 54.437
561754 94.477 -57.963 54.413
561755 94.505 -57.974 54.388
561756 94.530 -57.986 54.363
561757 94.555 -57.998 54.340
561758 94.580 -58.010 54.317
561759 94.606 -58.022 54.291
561760 94.631 -58.034 54.261
561761 94.657 -58.046 54.229
561762 94.682 -58.059 54.197
561763 94.707 -58.071 54.161
561764 94.732 -58.083 54.120
561765 94.757 -58.095 54.076
561766 94.782 -58.107 54.026
561767 94.808 -58.119 53.974
561768 94.833 -58.131 53.920
561769 94.858 -58.143 53.867
561770 94.884 -58.155 53.814
561771 94.909 -58.167 53.761
561772 94.934 -58.179 53.707
561773 94.988 -58.189 53.654
561774 95.044 -58.198 53.603
561775 95.100 -58.207 53.554
561776 95.156 -58.216 53.509
561777 95.212 -58.226 53.466
561778 95.268 -58.235 53.423
561779 95.324 -58.244 53.381
561780 95.370 -58.248 53.342
561781 95.407 -58.247 53.307
561782 95.444 -58.247 53.275
561783 95.480 -58.247 53.247
561784 95.517 -58.246 53.219
561785 95.551 -58.246 53.191
561786 95.581 -58.247 53.166
561787 95.610 -58.249 53.149
561788 95.640 -58.250 53.140
561789 95.672 -58.263 53.128
561790 95.704 -58.277 53.121
561791 95.737 -58.291 53.115
561792 95.769 -58.303 53.108
561793 95.801 -58.317 53.094
561794 95.833 -58.331 53.087
561795 95.864 -58.338 53.077
561796 95.894 -58.344 53.063
561797 95.924 -58.350 53.046
561798 95.956 -58.359 53.027
561799 95.988 -58.370 53.013
561800 96.021 -58.381 52.987
561801 96.054 -58.392 52.965
561802 96.087 -58.403 52.946
561803 96.119 -58.414 52.922
561804 96.152 -58.425 52.888
561805 96.185 -58.435 52.860
561806 96.218 -58.447 52.832
561807 96.251 -58.457 52.805
561808 96.284 -58.469 52.782
561809 96.316 -58.479 52.761
561810 96.349 -58.491 52.736
561811 96.382 -58.501 52.704
561812 96.414 -58.513 52.677
561813 96.447 -58.517 52.650
561814 96.478 -58.516 52.618
561815 96.509 -58.514 52.579
561816 96.540 -58.512 52.539
561817 96.571 -58.510 52.499
561818 96.602 -58.509 52.462
561819 96.633 -58.507 52.423
561820 96.665 -58.505 52.383
561821 96.696 -58.504 52.345
561822 96.728 -58.508 52.304
561823 96.760 -58.514 52.262
561824 96.793 -58.520 52.217
561825 96.825 -58.526 52.173
561826 96.857 -58.532 52.129
561827 96.890 -58.538 52.086
561828 96.922 -58.545 52.044
561829 96.954 -58.551 52.006
561830 96.984 -58.554 51.967
561831 97.013 -58.557 51.930
561832 97.042 -58.560 51.894
561833 97.071 -58.563 51.856
561834 97.100 -58.567 51.818
561835 97.129 -58.570 51.782
561836 97.160 -58.573 51.748
561837 97.192 -58.577 51.714
561838 97.225 -58.581 51.678
561839 97.258 -58.585 51.642
561840 97.291 -58.590 51.601
561841 97.323 -58.594 51.557
561842 97.356 -58.598 51.510
561843 97.389 -58.601 51.462
561844 97.425 -58.601 51.413
561845 97.461 -58.601 51.361
561846 97.497 -58.601 51.313
561847 97.533 -58.601 51.268
561848 97.568 -58.603 51.224
561849 97.599 -58.608 51.181
561850 97.631 -58.614 51.138
561851 97.658 -58.624 51.095
561852 97.694 -58.626 51.051
561853 97.732 -58.626 51.007
561854 97.770 -58.626 50.963
561855 97.807 -58.626 50.920
561856 97.845 -58.627 50.877
561857 97.883 -58.630 50.839
561858 97.921 -58.632 50.800
561859 97.959 -58.636 50.762
561860 97.998 -58.638 50.725
561861 98.039 -58.628 50.688
561862 98.078 -58.620 50.654
561863 98.111 -58.621 50.621
561864 98.144 -58.622 50.587
561865 98.176 -58.623 50.553
561866 98.209 -58.624 50.519
561867 98.242 -58.625 50.486
561868 98.275 -58.626 50.455
561869 98.310 -58.626 50.421
561870 98.345 -58.627 50.390
561871 98.380 -58.628 50.359
561872 98.415 -58.629 50.327
561873 98.449 -58.630 50.293
561874 98.482 -58.632 50.258
561875 98.514 -58.632 50.227
561876 98.546 -58.633 50.198
561877 98.578 -58.634 50.172
561878 98.609 -58.635 50.148
561879 98.641 -58.635 50.126
561880 98.679 -58.634 50.107
561881 98.717 -58.632 50.090
561882 98.756 -58.620 50.073
561883 98.791 -58.601 50.057
561884 98.825 -58.582 50.042
561885 98.858 -58.563 50.027
561886 98.891 -58.545 50.012
561887 98.927 -58.529 49.995
561888 98.974 -58.526 49.979
561889 99.017 -58.525 49.963
561890 99.061 -58.524 49.945
561891 99.105 -58.523 49.926
561892 99.149 -58.523 49.907
561893 99.193 -58.523 49.886
561894 99.236 -58.522 49.863
561895 99.280 -58.521 49.838
561896 99.324 -58.513 49.813
561897 99.372 -58.516 49.784
561898 99.406 -58.526 49.756
561899 99.441 -58.534 49.730
561900 99.475 -58.543 49.705
561901 99.510 -58.551 49.681
561902 99.544 -58.560 49.655
561903 99.579 -58.569 49.631
561904 99.613 -58.578 49.606
561905 99.648 -58.587 49.579
561906 99.682 -58.595 49.552
561907 99.717 -58.604 49.523
561908 99.751 -58.613 49.493
562255 87.125 -55.044 57.620
562256 87.160 -55.060 57.618
562257 87.195 -55.075 57.615
562258 87.230 -55.092 57.610
562259 87.263 -55.108 57.605
562260 87.293 -55.126 57.600
562261 87.324 -55.143 57.595
562262 87.354 -55.160 57.589
562263 87.385 -55.177 57.582
562264 87.415 -55.195 57.575
562265 87.445 -55.211 57.566
562266 87.475 -55.229 57.559
562267 87.506 -55.246 57.551
562268 87.536 -55.263 57.543
562269 87.566 -55.280 57.537
562270 87.599 -55.295 57.528
562271 87.648 -55.293 57.514
562272 87.707 -55.301 57.494
562273 87.765 -55.315 57.467
562274 87.819 -55.331 57.436
562275 87.873 -55.347 57.407
562276 87.919 -55.364 57.375
562277 87.956 -55.380 57.342
562278 87.994 -55.397 57.307
562279 88.032 -55.414 57.269
562280 88.069 -55.430 57.229
562281 88.107 -55.446 57.186
562282 88.144 -55.463 57.141
562283 88.182 -55.480 57.094
562284 88.219 -55.496 57.048
562285 88.257 -55.513 57.001
562286 88.294 -55.530 56.956
562287 88.332 -55.546 56.912
562288 88.370 -55.563 56.868
562289 88.407 -55.579 56.825
562290 88.445 -55.596 56.783
562311 89.239 -55.940 56.039
562312 89.277 -55.956 56.018
562313 89.316 -55.972 55.996
562314 89.354 -55.988 55.977
562315 89.392 -56.004 55.958
562316 89.430 -56.020 55.942
562317 89.469 -56.036 55.926
562318 89.507 -56.052 55.911
562319 89.545 -56.068 55.900
562320 89.584 -56.084 55.891
562321 89.622 -56.100 55.882
562322 89.660 -56.116 55.875
562323 89.698 -56.132 55.870
562324 89.733 -56.150 55.866
562325 89.774 -56.160 55.866
562326 89.812 -56.173 55.867
562327 89.837 -56.190 55.869
562328 89.863 -56.207 55.873
562329 89.888 -56.224 55.879
562330 89.914 -56.241 55.887
562331 89.943 -56.260 55.900
562332 89.972 -56.278 55.913
562333 90.000 -56.297 55.929
562334 90.031 -56.312 55.946
562335 90.066 -56.325 55.965
562336 90.101 -56.338 55.981
562337 90.136 -56.350 55.992
562338 90.171 -56.363 55.994
562339 90.208 -56.374 55.985
562340 90.252 -56.385 55.970
562341 90.297 -56.395 55.955
562342 90.341 -56.406 55.928
562343 90.385 -56.417 55.894
562344 90.429 -56.427 55.859
562345 90.473 -56.439 55.822
562346 90.516 -56.451 55.784
562347 90.559 -56.463 55.749
562348 90.603 -56.475 55.717
562349 90.646 -56.487 55.688
562350 90.689 -56.500 55.661
562351 90.733 -56.512 55.635
562352 90.776 -56.524 55.609
562353 90.819 -56.536 55.585
562354 90.863 -56.548 55.563
562355 90.906 -56.561 55.544
562356 90.950 -56.573 55.528
562357 90.993 -56.585 55.515
562358 91.036 -56.597 55.502
562359 91.080 -56.610 55.495
562360 91.119 -56.624 55.488
562361 91.154 -56.642 55.480
562362 91.189 -56.659 55.469
562363 91.225 -56.677 55.454
562364 91.260 -56.693 55.433
562365 91.296 -56.711 55.412
562366 91.331 -56.728 55.393
562367 91.365 -56.744 55.377
562368 91.397 -56.756 55.362
562369 91.430 -56.768 55.351
562370 91.462 -56.781 55.340
562371 91.495 -56.793 55.332
562372 91.527 -56.805 55.326
562373 91.559 -56.818 55.324
562374 91.591 -56.830 55.326
562375 91.624 -56.842 55.331
562376 91.656 -56.855 55.342
562377 91.695 -56.865 55.351
562378 91.738 -56.874 55.359
562379 91.781 -56.884 55.367
562380 91.824 -56.894 55.377
562381 91.867 -56.903 55.397
562382 91.910 -56.913 55.412
562383 91.953 -56.923 55.419
562384 91.995 -56.938 55.423
562385 92.036 -56.962 55.425
562386 92.077 -56.985 55.428
562387 92.118 -57.009 55.426
562388 92.159 -57.031 55.425
562389 92.200 -57.055 55.422
562390 92.241 -57.078 55.417
562391 92.282 -57.102 55.409
562392 92.323 -57.125 55.398
562393 92.364 -57.149 55.389
562394 92.406 -57.172 55.380
562395 92.446 -57.196 55.369
562396 92.486 -57.221 55.355
562397 92.522 -57.247 55.340
562398 92.559 -57.274 55.321
562399 92.593 -57.297 55.300
562400 92.628 -57.318 55.281
562401 92.662 -57.338 55.262
562402 92.697 -57.360 55.243
562403 92.731 -57.381 55.224
562404 92.766 -57.402 55.208
562405 92.799 -57.418 55.189
562406 92.832 -57.429 55.166
562407 92.865 -57.440 55.142
562408 92.898 -57.451 55.117
562409 92.927 -57.458 55.091
562410 92.955 -57.466 55.065
562411 92.984 -57.473 55.038
562412 93.012 -57.479 55.012
562413 93.041 -57.487 54.987
562414 93.069 -57.494 54.965
562415 93.098 -57.501 54.941
562416 93.124 -57.507 54.919
562417 93.152 -57.510 54.899
562418 93.184 -57.509 54.879
562419 93.215 -57.508 54.859
562420 93.246 -57.507 54.840
562421 93.278 -57.507 54.823
562422 93.309 -57.506 54.805
562423 93.341 -57.507 54.790
562424 93.373 -57.509 54.777
562425 93.406 -57.511 54.767
562426 93.438 -57.514 54.755
562427 93.471 -57.517 54.745
562428 93.503 -57.519 54.731
562429 93.536 -57.521 54.714
562430 93.568 -57.523 54.694
562431 93.601 -57.526 54.674
562432 93.633 -57.529 54.658
562433 93.666 -57.531 54.644
562434 93.699 -57.533 54.631
562435 93.731 -57.537 54.624
562436 93.770 -57.557 54.623
562437 93.808 -57.576 54.625
562438 93.846 -57.596 54.628
562439 93.885 -57.616 54.632
562440 93.922 -57.636 54.631
562441 93.961 -57.655 54.627
562442 93.999 -57.675 54.626
562443 94.037 -57.695 54.625
562444 94.075 -57.714 54.624
562445 94.114 -57.734 54.621
562446 94.152 -57.754 54.614
562447 94.190 -57.774 54.603
562448 94.228 -57.793 54.586
562449 94.266 -57.812 54.566
562450 94.305 -57.832 54.543
562451 94.342 -57.852 54.519
562452 94.380 -57.875 54.493
562453 94.417 -57.899 54.468
562454 94.452 -57.916 54.444
562455 94.484 -57.928 54.421
562456 94.516 -57.940 54.397
562457 94.546 -57.950 54.373
562458 94.571 -57.962 54.349
562459 94.596 -57.974 54.327
562460 94.621 -57.987 54.303
562461 94.647 -57.999 54.276
562462 94.672 -58.011 54.245
562463 94.697 -58.023 54.213
562464 94.722 -58.035 54.177
562465 94.748 -58.047 54.137
562466 94.773 -58.059 54.095
562467 94.798 -58.071 54.047
562468 94.824 -58.084 53.996
562469 94.849 -58.096 53.943
562470 94.874 -58.108 53.889
562471 94.899 -58.120 53.835
562472 94.924 -58.131 53.780
562473 94.950 -58.144 53.724
562474 94.992 -58.154 53.669
562475 95.048 -58.163 53.615
562476 95.104 -58.173 53.564
562477 95.160 -58.182 53.517
562478 95.216 -58.191 53.469
562479 95.272 -58.200 53.421
562480 95.328 -58.210 53.375
562481 95.374 -58.214 53.332
562482 95.410 -58.213 53.292
562483 95.447 -58.212 53.257
562484 95.484 -58.212 53.226
562485 95.521 -58.211 53.196
562486 95.558 -58.210 53.168
562487 95.590 -58.211 53.141
562488 95.620 -58.212 53.122
562489 95.650 -58.214 53.114
562490 95.681 -58.222 53.104
562491 95.713 -58.235 53.098
562492 95.745 -58.249 53.089
562493 95.777 -58.263 53.075
562494 95.810 -58.276 53.060
562495 95.841 -58.287 53.051
562496 95.872 -58.294 53.042
562497 95.902 -58.300 53.033
562498 95.933 -58.308 53.022
562499 95.966 -58.319 53.010
562500 95.999 -58.330 52.995
562501 96.031 -58.341 52.973
562502 96.064 -58.352 52.953
562503 96.097 -58.363 52.933
562504 96.129 -58.374 52.912
562505 96.162 -58.385 52.886
562506 96.195 -58.395 52.857
562507 96.228 -58.407 52.833
562508 96.261 -58.417 52.807
562509 96.294 -58.429 52.782
562510 96.326 -58.439 52.759
562511 96.359 -58.451 52.734
562512 96.392 -58.461 52.705
562513 96.425 -58.473 52.677
562514 96.457 -58.483 52.650
562515 96.489 -58.486 52.621
562516 96.520 -58.485 52.586
562517 96.551 -58.482 52.541
562518 96.583 -58.481 52.501
562519 96.614 -58.479 52.463
562520 96.645 -58.477 52.422
562521 96.676 -58.476 52.381
562522 96.707 -58.474 52.343
562523 96.739 -58.479 52.302
562524 96.771 -58.485 52.260
562525 96.804 -58.491 52.217
562526 96.836 -58.497 52.174
562527 96.869 -58.503 52.131
562528 96.901 -58.509 52.090
562529 96.934 -58.515 52.050
562530 96.964 -58.520 52.013
562531 96.993 -58.523 51.974
562532 97.022 -58.526 51.937
562533 97.050 -58.529 51.903
562534 97.079 -58.532 51.868
562535 97.108 -58.534 51.834
562536 97.137 -58.537 51.801
562537 97.166 -58.540 51.770
562538 97.195 -58.543 51.739
562539 97.228 -58.548 51.705
562540 97.261 -58.551 51.668
562541 97.294 -58.556 51.628
562542 97.326 -58.560 51.584
562543 97.360 -58.563 51.537
562544 97.396 -58.563 51.487
562545 97.432 -58.563 51.434
562546 97.468 -58.563 51.378
562547 97.504 -58.563 51.327
562548 97.540 -58.562 51.280
562549 97.574 -58.564 51.235
562550 97.605 -58.570 51.191
562551 97.634 -58.579 51.146
562552 97.662 -58.589 51.102
562553 97.698 -58.590 51.057
562554 97.735 -58.590 51.013
562555 97.773 -58.590 50.970
562556 97.811 -58.590 50.928
562557 97.849 -58.589 50.887
562558 97.886 -58.589 50.846
562559 97.924 -58.590 50.806
562560 97.962 -58.592 50.768
562561 98.000 -58.595 50.731
562562 98.040 -58.592 50.694
562563 98.077 -58.587 50.659
562564 98.109 -58.588 50.624
562565 98.142 -58.589 50.590
562566 98.175 -58.590 50.556
562567 98.208 -58.591 50.524
562568 98.240 -58.592 50.493
562569 98.273 -58.592 50.464
562570 98.306 -58.594 50.433
562571 98.342 -58.595 50.401
562572 98.377 -58.595 50.369
562573 98.411 -58.596 50.334
562574 98.444 -58.598 50.297
562575 98.477 -58.599 50.261
562576 98.509 -58.600 50.227
562577 98.540 -58.601 50.195
562578 98.578 -58.599 50.167
562579 98.616 -58.598 50.143
562580 98.654 -58.597 50.120
562581 98.692 -58.595 50.101
562582 98.729 -58.594 50.084
562583 98.768 -58.587 50.067
562584 98.807 -58.570 50.051
562585 98.846 -58.551 50.036
562586 98.881 -58.533 50.022
562587 98.914 -58.514 50.007
562588 98.932 -58.505 49.990
562589 98.955 -58.507 49.973
562590 98.999 -58.507 49.955
562591 99.043 -58.506 49.936
562592 99.087 -58.505 49.916
562593 99.131 -58.504 49.895
562594 99.174 -58.504 49.875
562595 99.218 -58.503 49.852
562596 99.262 -58.502 49.829
562597 99.305 -58.479 49.804
562598 99.362 -58.489 49.779
562599 99.399 -58.498 49.753
562600 99.433 -58.507 49.727
562601 99.468 -58.515 49.701
562602 99.502 -58.524 49.677
562603 99.537 -58.532 49.653
562604 99.571 -58.542 49.629
562605 99.606 -58.550 49.604
562606 99.640 -58.559 49.578
562607 99.675 -58.567 49.551
562608 99.709 -58.576 49.523
562609 99.744 -58.585 49.493
562957 87.190 -55.010 57.616
562958 87.222 -55.026 57.614
562959 87.253 -55.044 57.610
562960 87.283 -55.060 57.605
562961 87.314 -55.078 57.600
562962 87.344 -55.095 57.595
562963 87.374 -55.112 57.590
562964 87.404 -55.129 57.584
562965 87.434 -55.147 57.576
562966 87.465 -55.163 57.568
562967 87.495 -55.181 57.561
562968 87.526 -55.198 57.553
562969 87.556 -55.215 57.546
562970 87.586 -55.232 57.541
562971 87.616 -55.250 57.532
562972 87.660 -55.253 57.521
562973 87.718 -55.258 57.504
562974 87.775 -55.272 57.480
562975 87.830 -55.288 57.451
562976 87.884 -55.304 57.423
562977 87.938 -55.320 57.394
562978 87.979 -55.337 57.363
562979 88.016 -55.354 57.330
562980 88.054 -55.370 57.295
562981 88.092 -55.387 57.255
562982 88.129 -55.403 57.214
562983 88.167 -55.420 57.169
562984 88.204 -55.436 57.123
562985 88.242 -55.453 57.077
562986 88.280 -55.470 57.031
562987 88.317 -55.486 56.984
562988 88.355 -55.503 56.940
562989 88.393 -55.520 56.895
562990 88.430 -55.536 56.851
562991 88.468 -55.552 56.808
562992 88.505 -55.569 56.766
563013 89.295 -55.918 56.017
563014 89.333 -55.934 55.996
563015 89.372 -55.950 55.976
563016 89.410 -55.965 55.957
563017 89.448 -55.981 55.939
563018 89.486 -55.997 55.923
563019 89.525 -56.013 55.908
563020 89.563 -56.029 55.895
563021 89.601 -56.045 55.885
563022 89.639 -56.061 55.876
563023 89.678 -56.078 55.869
563024 89.716 -56.094 55.863
563025 89.752 -56.110 55.858
563026 89.793 -56.121 55.858
563027 89.819 -56.138 55.858
563028 89.844 -56.155 55.859
563029 89.872 -56.173 55.861
563030 89.900 -56.191 55.865
563031 89.929 -56.210 55.871
563032 89.957 -56.228 55.881
563033 89.986 -56.246 55.891
563034 90.017 -56.262 55.901
563035 90.051 -56.275 55.914
563036 90.086 -56.287 55.927
563037 90.121 -56.300 55.938
563038 90.156 -56.312 55.949
563039 90.191 -56.325 55.952
563040 90.226 -56.337 55.946
563041 90.264 -56.349 55.934
563042 90.309 -56.360 55.921
563043 90.353 -56.370 55.896
563044 90.397 -56.380 55.867
563045 90.441 -56.391 55.836
563046 90.485 -56.401 55.806
563047 90.530 -56.412 55.774
563048 90.574 -56.423 55.744
563049 90.617 -56.435 55.716
563050 90.661 -56.447 55.690
563051 90.704 -56.459 55.665
563052 90.747 -56.471 55.641
563053 90.791 -56.483 55.618
563054 90.834 -56.495 55.596
563055 90.877 -56.508 55.575
563056 90.921 -56.520 55.557
563057 90.964 -56.533 55.542
563058 91.008 -56.545 55.530
563059 91.051 -56.557 55.518
563060 91.094 -56.569 55.514
563061 91.138 -56.581 55.508
563062 91.174 -56.597 55.502
563063 91.210 -56.614 55.493
563064 91.245 -56.632 55.479
563065 91.281 -56.649 55.458
563066 91.316 -56.667 55.435
563067 91.351 -56.683 55.412
563068 91.385 -56.698 55.392
563069 91.417 -56.711 55.376
563070 91.449 -56.723 55.361
563071 91.482 -56.735 55.350
563072 91.514 -56.747 55.341
563073 91.546 -56.759 55.335
563074 91.578 -56.772 55.334
563075 91.610 -56.784 55.338
563076 91.643 -56.796 55.346
563077 91.675 -56.809 55.359
563078 91.707 -56.821 55.372
563079 91.750 -56.830 55.380
563080 91.793 -56.840 55.385
563081 91.836 -56.849 55.389
563082 91.879 -56.859 55.403
563083 91.922 -56.869 55.415
563084 91.964 -56.880 55.425
563085 92.006 -56.903 55.432
563086 92.047 -56.927 55.437
563087 92.088 -56.950 55.439
563088 92.129 -56.974 55.433
563089 92.170 -56.997 55.424
563090 92.211 -57.021 55.413
563091 92.252 -57.044 55.400
563092 92.293 -57.068 55.389
563093 92.334 -57.091 55.384
563094 92.375 -57.114 55.380
563095 92.416 -57.138 55.376
563096 92.457 -57.161 55.368
563097 92.497 -57.185 55.356
563098 92.533 -57.213 55.343
563099 92.569 -57.235 55.329
563100 92.603 -57.257 55.313
563101 92.637 -57.278 55.295
563102 92.672 -57.299 55.274
563103 92.706 -57.320 55.252
563104 92.740 -57.341 55.229
563105 92.775 -57.362 55.212
563106 92.809 -57.383 55.194
563107 92.843 -57.396 55.176
563108 92.876 -57.407 55.157
563109 92.908 -57.417 55.139
563110 92.936 -57.424 55.118
563111 92.965 -57.431 55.092
563112 92.993 -57.438 55.066
563113 93.021 -57.445 55.039
563114 93.050 -57.452 55.013
563115 93.078 -57.460 54.989
563116 93.107 -57.467 54.966
563117 93.135 -57.473 54.942
563118 93.164 -57.476 54.920
563119 93.196 -57.475 54.898
563120 93.227 -57.476 54.877
563121 93.260 -57.479 54.856
563122 93.293 -57.481 54.837
563123 93.325 -57.484 54.819
563124 93.358 -57.486 54.802
563125 93.390 -57.489 54.786
563126 93.423 -57.491 54.772
563127 93.455 -57.493 54.757
563128 93.488 -57.496 54.742
563129 93.520 -57.498 54.725
563130 93.553 -57.501 54.707
563131 93.585 -57.503 54.688
563132 93.618 -57.505 54.671
563133 93.650 -57.508 54.654
563134 93.683 -57.510 54.641
563135 93.715 -57.513 54.625
563136 93.748 -57.515 54.610
563137 93.782 -57.521 54.601
563138 93.820 -57.541 54.602
563139 93.858 -57.561 54.605
563140 93.896 -57.580 54.609
563141 93.934 -57.600 54.608
563142 93.973 -57.620 54.607
563143 94.011 -57.639 54.606
563144 94.049 -57.659 54.607
563145 94.087 -57.679 54.609
563146 94.125 -57.699 54.608
563147 94.164 -57.718 54.603
563148 94.202 -57.738 54.594
563149 94.240 -57.757 54.579
563150 94.278 -57.777 54.559
563151 94.315 -57.800 54.538
563152 94.352 -57.823 54.515
563153 94.390 -57.846 54.491
563154 94.427 -57.869 54.467
563155 94.460 -57.882 54.443
563156 94.492 -57.894 54.420
563157 94.525 -57.906 54.396
563158 94.555 -57.918 54.374
563159 94.585 -57.927 54.351
563160 94.612 -57.939 54.328
563161 94.637 -57.951 54.306
563162 94.663 -57.963 54.281
563163 94.688 -57.975 54.254
563164 94.713 -57.987 54.224
563165 94.738 -57.999 54.190
563166 94.763 -58.012 54.153
563167 94.788 -58.024 54.112
563168 94.814 -58.036 54.065
563169 94.839 -58.048 54.017
563170 94.864 -58.060 53.967
563171 94.890 -58.072 53.915
563172 94.915 -58.084 53.860
563173 94.940 -58.096 53.804
563174 94.965 -58.109 53.746
563175 94.996 -58.120 53.689
563176 95.052 -58.129 53.633
563177 95.108 -58.138 53.579
563178 95.164 -58.147 53.530
563179 95.219 -58.157 53.478
563180 95.275 -58.166 53.425
563181 95.331 -58.175 53.375
563182 95.377 -58.179 53.328
563183 95.414 -58.178 53.282
563184 95.451 -58.178 53.242
563185 95.487 -58.177 53.207
563186 95.525 -58.176 53.175
563187 95.561 -58.175 53.145
563188 95.598 -58.175 53.118
563189 95.630 -58.176 53.095
563190 95.660 -58.178 53.085
563191 95.689 -58.181 53.078
563192 95.722 -58.194 53.071
563193 95.754 -58.208 53.059
563194 95.787 -58.222 53.041
563195 95.819 -58.235 53.025
563196 95.849 -58.243 53.015
563197 95.880 -58.250 53.006
563198 95.910 -58.256 52.999
563199 95.943 -58.268 52.992
563200 95.976 -58.278 52.984
563201 96.009 -58.290 52.970
563202 96.041 -58.300 52.956
563203 96.074 -58.312 52.937
563204 96.107 -58.322 52.915
563205 96.139 -58.334 52.893
563206 96.172 -58.344 52.873
563207 96.205 -58.356 52.843
563208 96.238 -58.366 52.820
563209 96.271 -58.377 52.793
563210 96.304 -58.388 52.768
563211 96.336 -58.399 52.746
563212 96.369 -58.410 52.722
563213 96.402 -58.421 52.698
563214 96.435 -58.432 52.672
563215 96.467 -58.443 52.644
563216 96.500 -58.454 52.616
563217 96.532 -58.454 52.584
563218 96.563 -58.453 52.541
563219 96.594 -58.451 52.505
563220 96.625 -58.449 52.464
563221 96.656 -58.448 52.423
563222 96.687 -58.446 52.383
563223 96.719 -58.444 52.342
563224 96.751 -58.450 52.299
563225 96.783 -58.456 52.258
563226 96.815 -58.462 52.218
563227 96.848 -58.468 52.179
563228 96.880 -58.474 52.141
563229 96.912 -58.480 52.103
563230 96.944 -58.485 52.063
563231 96.972 -58.488 52.028
563232 97.001 -58.491 51.994
563233 97.030 -58.494 51.958
563234 97.059 -58.496 51.924
563235 97.088 -58.499 51.892
563236 97.117 -58.502 51.859
563237 97.146 -58.505 51.828
563238 97.174 -58.508 51.798
563239 97.203 -58.510 51.768
563240 97.232 -58.513 51.735
563241 97.264 -58.518 51.696
563242 97.297 -58.522 51.656
563243 97.330 -58.525 51.613
563244 97.367 -58.525 51.565
563245 97.402 -58.525 51.512
563246 97.439 -58.524 51.454
563247 97.474 -58.524 51.396
563248 97.511 -58.524 51.345
563249 97.546 -58.524 51.295
563250 97.580 -58.526 51.249
563251 97.610 -58.534 51.203
563252 97.638 -58.544 51.156
563253 97.665 -58.554 51.109
563254 97.702 -58.554 51.065
563255 97.739 -58.554 51.022
563256 97.777 -58.554 50.980
563257 97.814 -58.554 50.939
563258 97.852 -58.554 50.900
563259 97.890 -58.554 50.856
563260 97.928 -58.554 50.815
563261 97.965 -58.554 50.777
563262 98.003 -58.554 50.740
563263 98.040 -58.555 50.703
563264 98.075 -58.554 50.666
563265 98.108 -58.555 50.630
563266 98.141 -58.556 50.595
563267 98.173 -58.557 50.562
563268 98.206 -58.557 50.532
563269 98.239 -58.559 50.501
563270 98.272 -58.560 50.474
563271 98.304 -58.560 50.444
563272 98.339 -58.561 50.412
563273 98.372 -58.563 50.378
563274 98.406 -58.563 50.342
563275 98.439 -58.565 50.302
563276 98.473 -58.564 50.266
563277 98.515 -58.563 50.229
563278 98.553 -58.563 50.195
563279 98.590 -58.561 50.165
563280 98.629 -58.560 50.139
563281 98.666 -58.558 50.115
563282 98.704 -58.557 50.096
563283 98.742 -58.556 50.079
563284 98.780 -58.554 50.063
563285 98.819 -58.536 50.048
563286 98.858 -58.519 50.033
563287 98.897 -58.501 50.019
563288 98.936 -58.483 50.004
563289 98.942 -58.483 49.988
563290 98.939 -58.488 49.970
563291 98.981 -58.488 49.950
563292 99.025 -58.488 49.930
563293 99.069 -58.487 49.909
563294 99.113 -58.486 49.887
563295 99.156 -58.485 49.865
563296 99.200 -58.485 49.842
563297 99.244 -58.480 49.819
563298 99.286 -58.443 49.796
563299 99.341 -58.460 49.774
563300 99.391 -58.469 49.748
563301 99.426 -58.479 49.723
563302 99.460 -58.488 49.698
563303 99.495 -58.496 49.674
563304 99.529 -58.505 49.650
563305 99.564 -58.513 49.626
563306 99.598 -58.523 49.601
563307 99.633 -58.531 49.576
563308 99.667 -58.540 49.549
563309 99.702 -58.548 49.522
563310 99.736 -58.557 49.493
563659 87.242 -54.978 57.612
563660 87.273 -54.996 57.609
563661 87.303 -55.013 57.604
563662 87.333 -55.030 57.600
563663 87.363 -55.047 57.595
563664 87.394 -55.065 57.591
563665 87.424 -55.082 57.586
563666 87.454 -55.099 57.578
563667 87.485 -55.116 57.571
563668 87.515 -55.133 57.564
563669 87.545 -55.151 57.557
563670 87.575 -55.168 57.550
563671 87.606 -55.185 57.544
563672 87.636 -55.202 57.537
563673 87.673 -55.213 57.527
563674 87.728 -55.215 57.512
563675 87.785 -55.229 57.491
563676 87.840 -55.245 57.464
563677 87.894 -55.261 57.439
563678 87.948 -55.278 57.412
563679 88.002 -55.294 57.383
563680 88.039 -55.311 57.353
563681 88.077 -55.327 57.320
563682 88.115 -55.344 57.282
563683 88.152 -55.360 57.242
563684 88.190 -55.376 57.198
563685 88.227 -55.393 57.152
563686 88.265 -55.410 57.106
563687 88.303 -55.426 57.060
563688 88.340 -55.443 57.013
563689 88.378 -55.460 56.968
563690 88.415 -55.477 56.923
563691 88.453 -55.493 56.878
563692 88.490 -55.509 56.834
563693 88.528 -55.526 56.791
563694 88.566 -55.542 56.750
563714 89.318 -55.875 56.017
563715 89.355 -55.891 55.995
563716 89.393 -55.908 55.974
563717 89.430 -55.925 55.955
563718 89.468 -55.941 55.936
563719 89.506 -55.957 55.919
563720 89.543 -55.974 55.903
563721 89.581 -55.991 55.890
563722 89.619 -56.007 55.879
563723 89.657 -56.023 55.869
563724 89.695 -56.039 55.862
563725 89.734 -56.055 55.856
563726 89.772 -56.070 55.851
563727 89.801 -56.086 55.849
563728 89.829 -56.104 55.848
563729 89.858 -56.122 55.848
563730 89.886 -56.141 55.849
563731 89.914 -56.159 55.850
563732 89.943 -56.177 55.855
563733 89.971 -56.196 55.862
563734 90.002 -56.212 55.869
563735 90.037 -56.225 55.876
563736 90.072 -56.237 55.884
563737 90.106 -56.250 55.893
563738 90.142 -56.262 55.900
563739 90.176 -56.275 55.908
563740 90.211 -56.287 55.911
563741 90.246 -56.300 55.907
563742 90.281 -56.312 55.898
563743 90.321 -56.323 55.886
563744 90.365 -56.334 55.864
563745 90.409 -56.345 55.841
563746 90.453 -56.355 55.817
563747 90.498 -56.366 55.792
563748 90.542 -56.376 55.766
563749 90.586 -56.387 55.740
563750 90.630 -56.397 55.716
563751 90.674 -56.407 55.692
563752 90.719 -56.418 55.670
563753 90.762 -56.430 55.648
563754 90.805 -56.442 55.627
563755 90.849 -56.454 55.607
563756 90.892 -56.467 55.587
563757 90.935 -56.479 55.571
563758 90.978 -56.492 55.557
563759 91.022 -56.504 55.546
563760 91.065 -56.516 55.535
563761 91.109 -56.528 55.531
563762 91.152 -56.540 55.527
563763 91.195 -56.553 55.521
563764 91.230 -56.570 55.513
563765 91.265 -56.587 55.500
563766 91.301 -56.605 55.480
563767 91.336 -56.622 55.456
563768 91.372 -56.639 55.431
563769 91.404 -56.652 55.408
563770 91.436 -56.664 55.389
563771 91.469 -56.677 55.374
563772 91.501 -56.689 55.361
563773 91.533 -56.702 55.350
563774 91.566 -56.714 55.344
563775 91.598 -56.726 55.343
563776 91.630 -56.738 55.349
563777 91.662 -56.750 55.360
563778 91.695 -56.763 55.373
563779 91.727 -56.775 55.385
563780 91.762 -56.787 55.392
563781 91.805 -56.796 55.392
563782 91.848 -56.805 55.392
563783 91.891 -56.815 55.400
563784 91.934 -56.825 55.410
563785 91.975 -56.846 55.423
563786 92.016 -56.869 55.435
563787 92.057 -56.893 55.444
563788 92.098 -56.916 55.445
563789 92.139 -56.940 55.435
563790 92.180 -56.962 55.420
563791 92.221 -56.986 55.400
563792 92.263 -57.009 55.382
563793 92.303 -57.033 55.374
563794 92.345 -57.056 55.372
563795 92.385 -57.080 55.370
563796 92.427 -57.103 55.368
563797 92.467 -57.127 55.365
563798 92.508 -57.150 55.359
563799 92.544 -57.174 55.352
563800 92.578 -57.195 55.341
563801 92.613 -57.216 55.324
563802 92.647 -57.238 55.305
563803 92.682 -57.259 55.286
563804 92.716 -57.280 55.268
563805 92.750 -57.301 55.248
563806 92.785 -57.322 55.230
563807 92.819 -57.344 55.215
563808 92.853 -57.363 55.203
563809 92.886 -57.374 55.190
563810 92.916 -57.382 55.175
563811 92.945 -57.390 55.157
563812 92.973 -57.397 55.133
563813 93.002 -57.404 55.106
563814 93.030 -57.411 55.077
563815 93.059 -57.418 55.048
563816 93.087 -57.425 55.020
563817 93.116 -57.432 54.996
563818 93.145 -57.438 54.970
563819 93.177 -57.448 54.944
563820 93.210 -57.451 54.919
563821 93.244 -57.455 54.897
563822 93.277 -57.458 54.875
563823 93.309 -57.460 54.857
563824 93.342 -57.464 54.839
563825 93.374 -57.466 54.822
563826 93.407 -57.468 54.806
563827 93.439 -57.470 54.791
563828 93.472 -57.473 54.774
563829 93.504 -57.476 54.757
563830 93.537 -57.478 54.738
563831 93.570 -57.480 54.720
563832 93.602 -57.482 54.701
563833 93.635 -57.485 54.681
563834 93.667 -57.488 54.660
563835 93.700 -57.490 54.642
563836 93.732 -57.492 54.624
563837 93.765 -57.495 54.606
563838 93.797 -57.497 54.591
563839 93.832 -57.505 54.579
563840 93.870 -57.525 54.582
563841 93.908 -57.545 54.585
563842 93.946 -57.564 54.587
563843 93.984 -57.584 54.586
563844 94.023 -57.604 54.586
563845 94.061 -57.623 54.591
563846 94.099 -57.643 54.596
563847 94.137 -57.663 54.598
563848 94.176 -57.683 54.594
563849 94.213 -57.702 54.583
563850 94.251 -57.725 54.568
563851 94.288 -57.748 54.549
563852 94.325 -57.771 54.530
563853 94.363 -57.794 54.509
563854 94.400 -57.817 54.486
563855 94.435 -57.835 54.463
563856 94.467 -57.847 54.438
563857 94.500 -57.859 54.416
563858 94.532 -57.872 54.392
563859 94.564 -57.884 54.371
563860 94.595 -57.895 54.350
563861 94.625 -57.905 54.327
563862 94.652 -57.915 54.307
563863 94.678 -57.927 54.284
563864 94.703 -57.940 54.259
563865 94.729 -57.952 54.230
563866 94.754 -57.964 54.198
563867 94.779 -57.976 54.164
563868 94.804 -57.988 54.124
563869 94.829 -58.000 54.080
563870 94.854 -58.012 54.034
563871 94.880 -58.024 53.986
563872 94.905 -58.037 53.936
563873 94.930 -58.049 53.882
563874 94.956 -58.061 53.826
563875 94.981 -58.073 53.770
563876 95.006 -58.085 53.715
563877 95.055 -58.095 53.660
563878 95.111 -58.104 53.604
563879 95.168 -58.113 53.548
563880 95.224 -58.122 53.491
563881 95.279 -58.131 53.438
563882 95.335 -58.141 53.385
563883 95.381 -58.145 53.333
563884 95.417 -58.144 53.282
563885 95.454 -58.144 53.236
563886 95.491 -58.143 53.194
563887 95.528 -58.142 53.158
563888 95.565 -58.141 53.126
563889 95.602 -58.140 53.097
563890 95.638 -58.140 53.070
563891 95.669 -58.141 53.053
563892 95.699 -58.143 53.047
563893 95.731 -58.153 53.038
563894 95.763 -58.166 53.025
563895 95.795 -58.180 53.008
563896 95.827 -58.193 52.990
563897 95.857 -58.199 52.980
563898 95.888 -58.206 52.969
563899 95.920 -58.216 52.962
563900 95.953 -58.228 52.958
563901 95.986 -58.238 52.952
563902 96.019 -58.250 52.941
563903 96.051 -58.260 52.934
563904 96.084 -58.272 52.918
563905 96.117 -58.282 52.894
563906 96.150 -58.294 52.871
563907 96.182 -58.304 52.856
563908 96.215 -58.316 52.827
563909 96.248 -58.326 52.802
563910 96.281 -58.337 52.777
563911 96.314 -58.348 52.752
563912 96.346 -58.359 52.729
563913 96.379 -58.370 52.707
563914 96.412 -58.381 52.685
563915 96.445 -58.392 52.659
563916 96.477 -58.403 52.631
563917 96.510 -58.414 52.603
563918 96.543 -58.425 52.574
563919 96.574 -58.423 52.535
563920 96.605 -58.421 52.503
563921 96.637 -58.419 52.462
563922 96.668 -58.418 52.422
563923 96.699 -58.416 52.384
563924 96.730 -58.414 52.345
563925 96.762 -58.420 52.305
563926 96.795 -58.426 52.270
563927 96.827 -58.432 52.233
563928 96.859 -58.438 52.196
563929 96.892 -58.444 52.159
563930 96.923 -58.450 52.121
563931 96.952 -58.453 52.084
563932 96.981 -58.455 52.050
563933 97.010 -58.458 52.018
563934 97.038 -58.461 51.987
563935 97.068 -58.464 51.955
563936 97.096 -58.467 51.922
563937 97.125 -58.470 51.891
563938 97.154 -58.473 51.863
563939 97.183 -58.476 51.834
563940 97.212 -58.479 51.805
563941 97.241 -58.482 51.774
563942 97.270 -58.485 51.739
563943 97.301 -58.486 51.698
563944 97.337 -58.486 51.650
563945 97.373 -58.486 51.597
563946 97.409 -58.486 51.539
563947 97.445 -58.486 51.478
563948 97.481 -58.485 51.421
563949 97.517 -58.485 51.369
563950 97.553 -58.485 51.320
563951 97.586 -58.489 51.270
563952 97.613 -58.499 51.220
563953 97.641 -58.508 51.172
563954 97.668 -58.518 51.124
563955 97.705 -58.519 51.079
563956 97.743 -58.519 51.038
563957 97.781 -58.519 50.998
563958 97.818 -58.519 50.956
563959 97.856 -58.519 50.914
563960 97.893 -58.519 50.870
563961 97.931 -58.518 50.827
563962 97.969 -58.518 50.788
563963 98.006 -58.519 50.751
563964 98.041 -58.520 50.714
563965 98.075 -58.522 50.675
563966 98.107 -58.523 50.639
563967 98.140 -58.523 50.604
563968 98.173 -58.524 50.572
563969 98.206 -58.525 50.540
563970 98.238 -58.526 50.510
563971 98.271 -58.527 50.483
563972 98.303 -58.528 50.453
563973 98.337 -58.527 50.419
563974 98.370 -58.526 50.383
563975 98.405 -58.525 50.347
563976 98.439 -58.525 50.309
563977 98.473 -58.524 50.273
563978 98.509 -58.524 50.235
563979 98.560 -58.523 50.198
563980 98.603 -58.523 50.166
563981 98.641 -58.521 50.138
563982 98.679 -58.520 50.115
563983 98.717 -58.519 50.094
563984 98.755 -58.517 50.078
563985 98.793 -58.516 50.062
563986 98.832 -58.503 50.047
563987 98.870 -58.485 50.033
563988 98.909 -58.468 50.018
563989 98.943 -58.454 50.004
563990 98.952 -58.460 49.987
563991 98.949 -58.466 49.969
563992 98.963 -58.469 49.950
563993 99.007 -58.469 49.929
563994 99.051 -58.468 49.905
563995 99.094 -58.467 49.882
563996 99.138 -58.466 49.857
563997 99.182 -58.466 49.832
563998 99.225 -58.445 49.809
563999 99.268 -58.408 49.788
564000 99.320 -58.431 49.767
564001 99.383 -58.442 49.744
564002 99.418 -58.451 49.718
564003 99.452 -58.460 49.694
564004 99.487 -58.469 49.671
564005 99.521 -58.477 49.646
564006 99.556 -58.486 49.623
564007 99.590 -58.495 49.598
564008 99.625 -58.504 49.573
564009 99.659 -58.512 49.547
564010 99.694 -58.521 49.519
564011 99.728 -58.529 49.491
564360 87.262 -54.931 57.611
564361 87.292 -54.948 57.607
564362 87.323 -54.966 57.603
564363 87.353 -54.982 57.599
564364 87.383 -55.000 57.595
564365 87.414 -55.017 57.592
564366 87.444 -55.034 57.588
564367 87.474 -55.051 57.580
564368 87.504 -55.069 57.573
564369 87.535 -55.085 57.567
564370 87.565 -55.103 57.561
564371 87.596 -55.120 57.554
564372 87.626 -55.137 57.548
564373 87.656 -55.154 57.540
564374 87.686 -55.172 57.529
564375 87.738 -55.173 57.516
564376 87.796 -55.186 57.498
564377 87.850 -55.202 57.476
564378 87.904 -55.219 57.454
564379 87.959 -55.235 57.429
564380 88.012 -55.251 57.402
564381 88.062 -55.267 57.374
564382 88.100 -55.284 57.343
564383 88.137 -55.301 57.308
564384 88.175 -55.317 57.269
564385 88.212 -55.333 57.227
564386 88.250 -55.350 57.182
564387 88.288 -55.367 57.136
564388 88.325 -55.383 57.090
564389 88.363 -55.400 57.044
564390 88.401 -55.417 56.998
564391 88.438 -55.433 56.952
564392 88.476 -55.450 56.906
564393 88.514 -55.466 56.861
564394 88.551 -55.483 56.817
564395 88.589 -55.499 56.774
564396 88.626 -55.516 56.733
564416 89.378 -55.848 55.994
564417 89.416 -55.865 55.972
564418 89.453 -55.881 55.952
564419 89.491 -55.898 55.934
564420 89.529 -55.914 55.915
564421 89.566 -55.931 55.898
564422 89.604 -55.947 55.886
564423 89.641 -55.964 55.873
564424 89.679 -55.981 55.863
564425 89.716 -55.995 55.855
564426 89.753 -56.013 55.848
564427 89.787 -56.034 55.844
564428 89.814 -56.054 55.842
564429 89.843 -56.072 55.840
564430 89.872 -56.091 55.838
564431 89.900 -56.109 55.837
564432 89.929 -56.127 55.837
564433 89.957 -56.145 55.840
564434 89.987 -56.162 55.844
564435 90.022 -56.175 55.848
564436 90.057 -56.187 55.853
564437 90.092 -56.200 55.858
564438 90.126 -56.212 55.864
564439 90.162 -56.225 55.866
564440 90.196 -56.237 55.870
564441 90.231 -56.250 55.871
564442 90.266 -56.262 55.868
564443 90.301 -56.275 55.861
564444 90.336 -56.287 55.851
564445 90.377 -56.298 55.835
564446 90.421 -56.309 55.817
564447 90.465 -56.320 55.799
564448 90.510 -56.330 55.780
564449 90.554 -56.340 55.760
564450 90.598 -56.351 55.738
564451 90.642 -56.361 55.716
564452 90.686 -56.372 55.696
564453 90.731 -56.382 55.676
564454 90.775 -56.393 55.656
564455 90.819 -56.404 55.637
564456 90.863 -56.414 55.619
564457 90.906 -56.426 55.601
564458 90.950 -56.439 55.586
564459 90.993 -56.451 55.573
564460 91.036 -56.463 55.562
564461 91.080 -56.475 55.551
564462 91.123 -56.487 55.545
564463 91.166 -56.499 55.540
564464 91.210 -56.512 55.535
564465 91.250 -56.526 55.527
564466 91.286 -56.543 55.514
564467 91.321 -56.561 55.498
564468 91.357 -56.577 55.477
564469 91.391 -56.594 55.451
564470 91.424 -56.606 55.425
564471 91.456 -56.619 55.403
564472 91.488 -56.631 55.387
564473 91.520 -56.643 55.372
564474 91.553 -56.655 55.360
564475 91.585 -56.668 55.353
564476 91.617 -56.680 55.350
564477 91.649 -56.693 55.354
564478 91.681 -56.705 55.364
564479 91.714 -56.717 55.375
564480 91.746 -56.729 55.386
564481 91.779 -56.741 55.393
564482 91.817 -56.752 55.394
564483 91.860 -56.762 55.401
564484 91.903 -56.771 55.411
564485 91.945 -56.788 55.419
564486 91.986 -56.811 55.429
564487 92.027 -56.834 55.439
564488 92.068 -56.858 55.444
564489 92.109 -56.881 55.440
564490 92.150 -56.905 55.425
564491 92.191 -56.928 55.405
564492 92.232 -56.952 55.382
564493 92.273 -56.975 55.364
564494 92.314 -56.999 55.364
564495 92.355 -57.022 55.364
564496 92.396 -57.045 55.361
564497 92.437 -57.067 55.361
564498 92.478 -57.087 55.362
564499 92.517 -57.115 55.362
564500 92.552 -57.135 55.360
564501 92.586 -57.156 55.348
564502 92.621 -57.177 55.332
564503 92.656 -57.198 55.317
564504 92.691 -57.219 55.304
564505 92.725 -57.240 55.292
564506 92.760 -57.261 55.275
564507 92.794 -57.282 55.257
564508 92.829 -57.304 55.244
564509 92.863 -57.325 55.234
564510 92.897 -57.341 55.223
564511 92.925 -57.348 55.210
564512 92.953 -57.355 55.191
564513 92.983 -57.362 55.166
564514 93.011 -57.369 55.136
564515 93.041 -57.375 55.104
564516 93.069 -57.382 55.074
564517 93.098 -57.388 55.045
564518 93.127 -57.395 55.018
564519 93.156 -57.401 54.991
564520 93.188 -57.412 54.963
564521 93.221 -57.423 54.936
564522 93.254 -57.427 54.913
564523 93.288 -57.431 54.896
564524 93.322 -57.435 54.881
564525 93.355 -57.438 54.865
564526 93.389 -57.442 54.849
564527 93.423 -57.446 54.833
564528 93.457 -57.450 54.817
564529 93.489 -57.452 54.800
564530 93.521 -57.455 54.781
564531 93.554 -57.457 54.761
564532 93.586 -57.460 54.740
564533 93.619 -57.462 54.719
564534 93.651 -57.464 54.697
564535 93.684 -57.467 54.671
564536 93.716 -57.470 54.648
564537 93.749 -57.472 54.625
564538 93.781 -57.474 54.604
564539 93.814 -57.476 54.583
564540 93.847 -57.479 54.563
564541 93.882 -57.489 54.561
564542 93.920 -57.509 54.562
564543 93.958 -57.529 54.564
564544 93.996 -57.548 54.564
564545 94.035 -57.568 54.564
564546 94.073 -57.588 54.571
564547 94.111 -57.608 54.577
564548 94.149 -57.627 54.580
564549 94.186 -57.650 54.578
564550 94.223 -57.673 54.570
564551 94.261 -57.695 54.556
564552 94.298 -57.718 54.539
564553 94.335 -57.742 54.521
564554 94.373 -57.764 54.500
564555 94.410 -57.787 54.478
564556 94.443 -57.800 54.455
564557 94.475 -57.813 54.430
564558 94.507 -57.825 54.408
564559 94.540 -57.837 54.386
564560 94.572 -57.850 54.365
564561 94.604 -57.861 54.345
564562 94.634 -57.871 54.325
564563 94.664 -57.882 54.305
564564 94.693 -57.892 54.284
564565 94.719 -57.904 54.260
564566 94.744 -57.916 54.233
564567 94.769 -57.928 54.202
564568 94.794 -57.940 54.169
564569 94.820 -57.952 54.133
564570 94.845 -57.965 54.092
564571 94.870 -57.977 54.048
564572 94.896 -57.989 54.001
564573 94.921 -58.001 53.953
564574 94.946 -58.013 53.901
564575 94.971 -58.025 53.847
564576 94.996 -58.037 53.791
564577 95.022 -58.050 53.736
564578 95.059 -58.060 53.680
564579 95.115 -58.069 53.623
564580 95.171 -58.079 53.562
564581 95.227 -58.088 53.504
564582 95.283 -58.097 53.448
564583 95.339 -58.106 53.393
564584 95.384 -58.110 53.338
564585 95.421 -58.109 53.285
564586 95.458 -58.109 53.234
564587 95.495 -58.108 53.186
564588 95.532 -58.107 53.145
564589 95.568 -58.106 53.109
564590 95.605 -58.106 53.077
564591 95.642 -58.105 53.047
564592 95.679 -58.105 53.027
564593 95.708 -58.106 53.013
564594 95.739 -58.112 53.003
564595 95.771 -58.125 52.993
564596 95.804 -58.139 52.981
564597 95.835 -58.149 52.963
564598 95.865 -58.156 52.949
564599 95.897 -58.165 52.937
564600 95.930 -58.176 52.928
564601 95.963 -58.187 52.921
564602 95.996 -58.198 52.913
564603 96.029 -58.209 52.909
564604 96.061 -58.220 52.904
564605 96.094 -58.231 52.891
564606 96.127 -58.242 52.871
564607 96.160 -58.253 52.850
564608 96.192 -58.264 52.837
564609 96.225 -58.275 52.811
564610 96.258 -58.286 52.783
564611 96.291 -58.297 52.760
564612 96.324 -58.308 52.736
564613 96.357 -58.319 52.709
564614 96.389 -58.330 52.688
564615 96.422 -58.341 52.664
564616 96.455 -58.352 52.638
564617 96.487 -58.363 52.611
564618 96.520 -58.374 52.583
564619 96.553 -58.385 52.556
564620 96.586 -58.394 52.524
564621 96.617 -58.391 52.496
564622 96.648 -58.390 52.458
564623 96.679 -58.388 52.420
564624 96.710 -58.386 52.385
564625 96.741 -58.385 52.348
564626 96.773 -58.391 52.313
564627 96.806 -58.397 52.284
564628 96.838 -58.403 52.251
564629 96.871 -58.409 52.215
564630 96.903 -58.415 52.178
564631 96.932 -58.417 52.142
564632 96.960 -58.420 52.106
564633 96.990 -58.423 52.074
564634 97.018 -58.426 52.044
564635 97.047 -58.429 52.015
564636 97.076 -58.432 51.985
564637 97.105 -58.435 51.954
564638 97.134 -58.438 51.927
564639 97.162 -58.441 51.900
564640 97.191 -58.444 51.872
564641 97.220 -58.447 51.842
564642 97.245 -58.451 51.811
564643 97.272 -58.455 51.778
564644 97.304 -58.454 51.736
564645 97.340 -58.453 51.686
564646 97.376 -58.452 51.629
564647 97.413 -58.452 51.567
564648 97.449 -58.451 51.505
564649 97.485 -58.451 51.449
564650 97.521 -58.450 51.396
564651 97.558 -58.450 51.346
564652 97.590 -58.454 51.295
564653 97.617 -58.463 51.243
564654 97.644 -58.473 51.193
564655 97.672 -58.482 51.145
564656 97.709 -58.483 51.100
564657 97.746 -58.483 51.059
564658 97.784 -58.483 51.020
564659 97.822 -58.483 50.979
564660 97.859 -58.483 50.934
564661 97.897 -58.483 50.889
564662 97.935 -58.483 50.845
564663 97.970 -58.485 50.804
564664 98.006 -58.486 50.765
564665 98.041 -58.488 50.726
564666 98.076 -58.489 50.686
564667 98.108 -58.490 50.650
564668 98.141 -58.491 50.614
564669 98.173 -58.491 50.580
564670 98.206 -58.493 50.547
564671 98.239 -58.494 50.514
564672 98.272 -58.495 50.483
564673 98.307 -58.495 50.451
564674 98.341 -58.493 50.416
564675 98.372 -58.487 50.380
564676 98.405 -58.485 50.342
564677 98.439 -58.485 50.303
564678 98.474 -58.484 50.265
564679 98.508 -58.484 50.228
564680 98.552 -58.483 50.194
564681 98.602 -58.483 50.164
564682 98.653 -58.483 50.137
564683 98.692 -58.482 50.114
564684 98.729 -58.480 50.094
564685 98.768 -58.479 50.076
564686 98.805 -58.478 50.060
564687 98.844 -58.469 50.046
564688 98.882 -58.452 50.032
564689 98.922 -58.435 50.018
564690 98.946 -58.431 50.003
564691 98.959 -58.435 49.987
564692 98.960 -58.444 49.970
564693 98.956 -58.450 49.951
564694 98.989 -58.451 49.928
564695 99.032 -58.450 49.904
564696 99.076 -58.449 49.878
564697 99.120 -58.448 49.852
564698 99.164 -58.447 49.827
564699 99.207 -58.410 49.803
564700 99.250 -58.372 49.779
564701 99.299 -58.401 49.759
564702 99.369 -58.415 49.737
564703 99.411 -58.423 49.712
564704 99.445 -58.432 49.688
564705 99.480 -58.441 49.665
564706 99.514 -58.449 49.641
564707 99.549 -58.458 49.618
564708 99.583 -58.467 49.594
564709 99.618 -58.476 49.568
564710 99.652 -58.485 49.542
564711 99.687 -58.493 49.516
564712 99.721 -58.502 49.488
565062 87.312 -54.900 57.606
565063 87.342 -54.918 57.603
565064 87.372 -54.935 57.600
565065 87.403 -54.952 57.596
565066 87.433 -54.969 57.592
565067 87.464 -54.987 57.588
565068 87.494 -55.003 57.582
565069 87.524 -55.021 57.576
565070 87.555 -55.038 57.570
565071 87.585 -55.055 57.564
565072 87.615 -55.072 57.558
565073 87.645 -55.090 57.551
565074 87.676 -55.107 57.543
565075 87.706 -55.124 57.533
565076 87.748 -55.130 57.520
565077 87.806 -55.144 57.505
565078 87.860 -55.160 57.486
565079 87.914 -55.176 57.466
565080 87.969 -55.191 57.444
565081 88.023 -55.208 57.419
565082 88.077 -55.224 57.393
565083 88.122 -55.241 57.363
565084 88.160 -55.257 57.331
565085 88.198 -55.274 57.294
565086 88.235 -55.290 57.254
565087 88.273 -55.307 57.212
565088 88.311 -55.323 57.168
565089 88.348 -55.340 57.121
565090 88.386 -55.357 57.074
565091 88.423 -55.373 57.028
565092 88.461 -55.390 56.982
565093 88.498 -55.407 56.935
565094 88.536 -55.423 56.889
565095 88.573 -55.439 56.844
565096 88.611 -55.456 56.800
565097 88.649 -55.473 56.757
565117 89.401 -55.805 55.993
565118 89.439 -55.821 55.970
565119 89.476 -55.838 55.949
565120 89.513 -55.853 55.929
565121 89.549 -55.866 55.911
565122 89.585 -55.881 55.895
565123 89.621 -55.895 55.881
565124 89.658 -55.912 55.868
565125 89.697 -55.934 55.857
565126 89.736 -55.958 55.848
565127 89.774 -55.981 55.841
565128 89.806 -56.001 55.838
565129 89.832 -56.019 55.836
565130 89.859 -56.038 55.832
565131 89.886 -56.057 55.829
565132 89.914 -56.077 55.827
565133 89.943 -56.095 55.826
565134 89.972 -56.112 55.826
565135 90.007 -56.125 55.827
565136 90.042 -56.137 55.829
565137 90.077 -56.150 55.832
565138 90.112 -56.162 55.835
565139 90.147 -56.175 55.838
565140 90.182 -56.187 55.840
565141 90.217 -56.200 55.842
565142 90.251 -56.212 55.843
565143 90.286 -56.225 55.840
565144 90.321 -56.237 55.835
565145 90.356 -56.250 55.826
565146 90.391 -56.262 55.815
565147 90.433 -56.273 55.801
565148 90.477 -56.283 55.787
565149 90.522 -56.294 55.772
565150 90.566 -56.304 55.756
565151 90.610 -56.315 55.737
565152 90.654 -56.326 55.718
565153 90.698 -56.336 55.700
565154 90.743 -56.347 55.683
565155 90.787 -56.357 55.666
565156 90.831 -56.367 55.649
565157 90.875 -56.378 55.632
565158 90.920 -56.389 55.616
565159 90.964 -56.399 55.602
565160 91.008 -56.410 55.589
565161 91.051 -56.422 55.579
565162 91.094 -56.434 55.568
565163 91.138 -56.447 55.559
565164 91.181 -56.459 55.552
565165 91.224 -56.471 55.546
565166 91.267 -56.483 55.539
565167 91.306 -56.498 55.526
565168 91.342 -56.516 55.510
565169 91.377 -56.533 55.491
565170 91.411 -56.548 55.468
565171 91.443 -56.561 55.443
565172 91.475 -56.573 55.420
565173 91.508 -56.585 55.400
565174 91.540 -56.597 55.384
565175 91.572 -56.610 55.371
565176 91.604 -56.622 55.361
565177 91.637 -56.634 55.355
565178 91.669 -56.646 55.354
565179 91.701 -56.658 55.361
565180 91.733 -56.671 55.371
565181 91.766 -56.683 55.381
565182 91.798 -56.696 55.389
565183 91.830 -56.708 55.394
565184 91.872 -56.718 55.408
565185 91.914 -56.730 55.419
565186 91.955 -56.753 55.426
565187 91.996 -56.777 55.430
565188 92.037 -56.800 55.434
565189 92.078 -56.824 55.434
565190 92.120 -56.847 55.423
565191 92.161 -56.871 55.402
565192 92.202 -56.892 55.379
565193 92.242 -56.912 55.359
565194 92.283 -56.934 55.348
565195 92.323 -56.954 55.353
565196 92.364 -56.975 55.356
565197 92.405 -56.996 55.360
565198 92.446 -57.016 55.363
565199 92.485 -57.048 55.365
565200 92.522 -57.078 55.365
565201 92.558 -57.099 55.359
565202 92.593 -57.120 55.347
565203 92.627 -57.141 55.336
565204 92.662 -57.162 55.330
565205 92.697 -57.182 55.328
565206 92.731 -57.203 55.321
565207 92.766 -57.224 55.306
565208 92.801 -57.245 55.287
565209 92.836 -57.266 55.273
565210 92.870 -57.287 55.263
565211 92.907 -57.305 55.252
565212 92.936 -57.312 55.238
565213 92.965 -57.319 55.216
565214 92.994 -57.325 55.189
565215 93.022 -57.332 55.158
565216 93.052 -57.338 55.125
565217 93.080 -57.344 55.095
565218 93.110 -57.351 55.067
565219 93.138 -57.357 55.040
565220 93.167 -57.364 55.013
565221 93.200 -57.375 54.986
565222 93.232 -57.387 54.961
565223 93.265 -57.399 54.938
565224 93.298 -57.404 54.916
565225 93.332 -57.407 54.898
565226 93.366 -57.410 54.883
565227 93.399 -57.414 54.869
565228 93.433 -57.418 54.856
565229 93.467 -57.422 54.842
565230 93.500 -57.426 54.827
565231 93.534 -57.429 54.809
565232 93.568 -57.433 54.789
565233 93.602 -57.437 54.767
565234 93.635 -57.441 54.743
565235 93.669 -57.444 54.717
565236 93.701 -57.446 54.688
565237 93.733 -57.449 54.659
565238 93.766 -57.451 54.630
565239 93.798 -57.454 54.602
565240 93.831 -57.456 54.578
565241 93.863 -57.458 54.554
565242 93.896 -57.461 54.544
565243 93.932 -57.473 54.541
565244 93.970 -57.493 54.540
565245 94.008 -57.513 54.540
565246 94.046 -57.532 54.542
565247 94.084 -57.552 54.548
565248 94.122 -57.575 54.554
565249 94.159 -57.598 54.557
565250 94.196 -57.620 54.557
565251 94.234 -57.643 54.552
565252 94.271 -57.666 54.541
565253 94.308 -57.689 54.525
565254 94.346 -57.712 54.507
565255 94.383 -57.735 54.488
565256 94.418 -57.753 54.467
565257 94.451 -57.766 54.445
565258 94.483 -57.778 54.421
565259 94.515 -57.790 54.396
565260 94.547 -57.803 54.374
565261 94.580 -57.815 54.354
565262 94.612 -57.828 54.335
565263 94.643 -57.839 54.318
565264 94.673 -57.849 54.300
565265 94.703 -57.859 54.280
565266 94.734 -57.869 54.256
565267 94.760 -57.880 54.230
565268 94.785 -57.893 54.200
565269 94.810 -57.905 54.169
565270 94.835 -57.917 54.134
565271 94.860 -57.929 54.096
565272 94.886 -57.941 54.056
565273 94.911 -57.953 54.012
565274 94.936 -57.965 53.966
565275 94.962 -57.977 53.916
565276 94.987 -57.990 53.863
565277 95.012 -58.002 53.808
565278 95.037 -58.014 53.751
565279 95.063 -58.026 53.694
565280 95.119 -58.035 53.634
565281 95.175 -58.044 53.575
565282 95.231 -58.053 53.517
565283 95.286 -58.063 53.457
565284 95.342 -58.072 53.399
565285 95.388 -58.076 53.343
565286 95.425 -58.075 53.288
565287 95.462 -58.075 53.234
565288 95.498 -58.074 53.182
565289 95.535 -58.073 53.136
565290 95.572 -58.072 53.095
565291 95.609 -58.071 53.059
565292 95.645 -58.071 53.026
565293 95.682 -58.070 53.002
565294 95.718 -58.069 52.983
565295 95.748 -58.071 52.969
565296 95.780 -58.084 52.959
565297 95.812 -58.097 52.949
565298 95.843 -58.105 52.933
565299 95.875 -58.115 52.915
565300 95.907 -58.125 52.902
565301 95.940 -58.136 52.891
565302 95.973 -58.147 52.882
565303 96.006 -58.158 52.875
565304 96.039 -58.169 52.871
565305 96.072 -58.180 52.865
565306 96.104 -58.191 52.857
565307 96.137 -58.202 52.843
565308 96.170 -58.213 52.827
565309 96.202 -58.224 52.814
565310 96.235 -58.235 52.792
565311 96.268 -58.246 52.766
565312 96.301 -58.257 52.744
565313 96.334 -58.268 52.720
565314 96.367 -58.278 52.692
565315 96.399 -58.290 52.667
565316 96.432 -58.300 52.642
565317 96.465 -58.312 52.615
565318 96.498 -58.322 52.588
565319 96.530 -58.334 52.560
565320 96.563 -58.344 52.533
565321 96.596 -58.356 52.507
565322 96.628 -58.362 52.481
565323 96.659 -58.360 52.450
565324 96.690 -58.358 52.418
565325 96.722 -58.357 52.385
565326 96.753 -58.355 52.351
565327 96.785 -58.361 52.322
565328 96.817 -58.367 52.296
565329 96.850 -58.373 52.265
565330 96.882 -58.379 52.231
565331 96.911 -58.382 52.196
565332 96.940 -58.385 52.162
565333 96.969 -58.388 52.128
565334 96.998 -58.391 52.097
565335 97.027 -58.394 52.068
565336 97.055 -58.397 52.040
565337 97.084 -58.400 52.012
565338 97.113 -58.403 51.986
565339 97.142 -58.406 51.965
565340 97.169 -58.410 51.939
565341 97.194 -58.415 51.911
565342 97.219 -58.420 51.880
565343 97.244 -58.425 51.846
565344 97.273 -58.423 51.812
565345 97.305 -58.421 51.769
565346 97.342 -58.421 51.719
565347 97.378 -58.420 51.661
565348 97.414 -58.419 51.599
565349 97.450 -58.419 51.536
565350 97.487 -58.419 51.481
565351 97.523 -58.418 51.426
565352 97.558 -58.419 51.374
565353 97.590 -58.422 51.322
565354 97.620 -58.428 51.269
565355 97.648 -58.438 51.218
565356 97.675 -58.448 51.170
565357 97.712 -58.448 51.125
565358 97.750 -58.448 51.083
565359 97.788 -58.448 51.044
565360 97.825 -58.448 51.003
565361 97.863 -58.448 50.957
565362 97.899 -58.448 50.911
565363 97.935 -58.451 50.866
565364 97.970 -58.452 50.823
565365 98.006 -58.454 50.780
565366 98.041 -58.455 50.738
565367 98.076 -58.457 50.698
565368 98.109 -58.458 50.660
565369 98.142 -58.459 50.623
565370 98.174 -58.460 50.587
565371 98.207 -58.460 50.552
565372 98.241 -58.461 50.516
565373 98.276 -58.462 50.481
565374 98.311 -58.462 50.446
565375 98.344 -58.459 50.411
565376 98.375 -58.453 50.374
565377 98.406 -58.446 50.335
565378 98.440 -58.444 50.296
565379 98.474 -58.444 50.258
565380 98.508 -58.444 50.222
565381 98.544 -58.443 50.190
565382 98.594 -58.443 50.162
565383 98.645 -58.443 50.136
565384 98.695 -58.442 50.114
565385 98.742 -58.442 50.093
565386 98.780 -58.441 50.075
565387 98.818 -58.439 50.059
565388 98.856 -58.436 50.045
565389 98.895 -58.419 50.032
565390 98.930 -58.405 50.018
565391 98.949 -58.407 50.002
565392 98.966 -58.409 49.987
565393 98.967 -58.419 49.969
565394 98.968 -58.428 49.950
565395 98.971 -58.432 49.927
565396 99.014 -58.432 49.902
565397 99.058 -58.431 49.875
565398 99.102 -58.430 49.848
565399 99.145 -58.412 49.822
565400 99.188 -58.375 49.797
565401 99.231 -58.338 49.772
565402 99.278 -58.372 49.751
565403 99.348 -58.385 49.729
565404 99.403 -58.395 49.706
565405 99.437 -58.404 49.682
565406 99.472 -58.413 49.659
565407 99.506 -58.422 49.635
565408 99.541 -58.430 49.612
565409 99.575 -58.439 49.588
565410 99.610 -58.448 49.563
565411 99.644 -58.457 49.537
565412 99.679 -58.466 49.512
565413 99.713 -58.474 49.484
565763 87.332 -54.853 57.605
565764 87.362 -54.870 57.602
565765 87.393 -54.887 57.600
565766 87.423 -54.904 57.596
565767 87.453 -54.922 57.593
565768 87.483 -54.939 57.589
565769 87.513 -54.956 57.584
565770 87.544 -54.973 57.578
565771 87.574 -54.991 57.573
565772 87.605 -55.007 57.567
565773 87.635 -55.025 57.561
565774 87.665 -55.042 57.555
565775 87.695 -55.059 57.547
565776 87.726 -55.076 57.537
565777 87.760 -55.089 57.525
565778 87.817 -55.101 57.513
565779 87.870 -55.116 57.496
565780 87.925 -55.133 57.478
565781 87.979 -55.149 57.458
565782 88.033 -55.165 57.436
565783 88.087 -55.182 57.410
565784 88.141 -55.198 57.382
565785 88.183 -55.214 57.352
565786 88.220 -55.231 57.319
565787 88.258 -55.247 57.281
565788 88.295 -55.263 57.240
565789 88.333 -55.280 57.196
565790 88.371 -55.297 57.150
565791 88.408 -55.314 57.104
565792 88.446 -55.330 57.058
565793 88.484 -55.347 57.012
565794 88.521 -55.364 56.964
565795 88.559 -55.380 56.917
565796 88.597 -55.396 56.871
565797 88.634 -55.413 56.826
565798 88.672 -55.430 56.782
565799 88.709 -55.446 56.738
565819 89.455 -55.766 55.968
565820 89.491 -55.781 55.946
565821 89.527 -55.794 55.926
565822 89.564 -55.809 55.908
565823 89.602 -55.833 55.891
565824 89.641 -55.856 55.876
565825 89.680 -55.880 55.863
565826 89.718 -55.903 55.851
565827 89.757 -55.926 55.841
565828 89.796 -55.950 55.832
565829 89.825 -55.968 55.827
565830 89.849 -55.984 55.823
565831 89.877 -56.003 55.819
565832 89.904 -56.022 55.817
565833 89.932 -56.041 55.816
565834 89.959 -56.061 55.816
565835 89.992 -56.075 55.814
565836 90.027 -56.087 55.812
565837 90.062 -56.099 55.811
565838 90.097 -56.112 55.812
565839 90.132 -56.125 55.815
565840 90.167 -56.137 55.817
565841 90.202 -56.150 55.818
565842 90.237 -56.162 55.820
565843 90.271 -56.175 55.820
565844 90.306 -56.187 55.817
565845 90.341 -56.200 55.815
565846 90.376 -56.212 55.808
565847 90.411 -56.225 55.800
565848 90.446 -56.237 55.791
565849 90.489 -56.248 55.780
565850 90.534 -56.258 55.769
565851 90.578 -56.269 55.755
565852 90.622 -56.279 55.739
565853 90.666 -56.290 55.724
565854 90.711 -56.300 55.708
565855 90.755 -56.310 55.693
565856 90.799 -56.321 55.678
565857 90.843 -56.332 55.663
565858 90.887 -56.342 55.647
565859 90.932 -56.353 55.632
565860 90.976 -56.364 55.618
565861 91.020 -56.373 55.606
565862 91.064 -56.384 55.596
565863 91.108 -56.395 55.585
565864 91.152 -56.406 55.574
565865 91.195 -56.418 55.565
565866 91.239 -56.430 55.558
565867 91.282 -56.442 55.553
565868 91.325 -56.454 55.541
565869 91.362 -56.471 55.527
565870 91.397 -56.489 55.509
565871 91.430 -56.502 55.488
565872 91.462 -56.514 55.463
565873 91.495 -56.527 55.439
565874 91.527 -56.539 55.416
565875 91.559 -56.552 55.397
565876 91.591 -56.564 55.382
565877 91.624 -56.576 55.369
565878 91.656 -56.588 55.359
565879 91.688 -56.601 55.354
565880 91.720 -56.613 55.355
565881 91.753 -56.625 55.360
565882 91.785 -56.637 55.368
565883 91.817 -56.649 55.379
565884 91.850 -56.662 55.393
565885 91.884 -56.674 55.413
565886 91.925 -56.696 55.425
565887 91.966 -56.717 55.429
565888 92.007 -56.738 55.428
565889 92.048 -56.758 55.424
565890 92.088 -56.779 55.418
565891 92.129 -56.800 55.400
565892 92.169 -56.821 55.377
565893 92.210 -56.842 55.356
565894 92.251 -56.862 55.343
565895 92.292 -56.883 55.340
565896 92.333 -56.904 55.345
565897 92.373 -56.924 55.348
565898 92.414 -56.946 55.349
565899 92.452 -56.979 55.350
565900 92.489 -57.015 55.350
565901 92.527 -57.042 55.351
565902 92.564 -57.062 55.349
565903 92.599 -57.083 55.341
565904 92.634 -57.104 55.341
565905 92.668 -57.125 55.346
565906 92.703 -57.146 55.348
565907 92.737 -57.166 55.343
565908 92.772 -57.188 55.329
565909 92.807 -57.208 55.310
565910 92.842 -57.229 55.292
565911 92.876 -57.250 55.279
565912 92.916 -57.267 55.269
565913 92.947 -57.275 55.253
565914 92.976 -57.282 55.230
565915 93.005 -57.288 55.202
565916 93.034 -57.294 55.172
565917 93.063 -57.301 55.143
565918 93.091 -57.307 55.115
565919 93.121 -57.314 55.088
565920 93.149 -57.321 55.062
565921 93.179 -57.327 55.037
565922 93.211 -57.338 55.014
565923 93.243 -57.351 54.990
565924 93.276 -57.363 54.965
565925 93.309 -57.374 54.938
565926 93.342 -57.379 54.917
565927 93.376 -57.383 54.899
565928 93.410 -57.387 54.884
565929 93.443 -57.390 54.872
565930 93.477 -57.394 54.861
565931 93.511 -57.398 54.847
565932 93.544 -57.401 54.831
565933 93.578 -57.405 54.812
565934 93.612 -57.409 54.789
565935 93.645 -57.413 54.764
565936 93.679 -57.416 54.737
565937 93.713 -57.420 54.708
565938 93.746 -57.424 54.677
565939 93.780 -57.428 54.645
565940 93.814 -57.432 54.613
565941 93.847 -57.435 54.583
565942 93.880 -57.438 54.553
565943 93.913 -57.441 54.533
565944 93.945 -57.443 54.522
565945 93.982 -57.457 54.516
565946 94.020 -57.477 54.514
565947 94.057 -57.500 54.518
565948 94.095 -57.523 54.523
565949 94.132 -57.545 54.526
565950 94.169 -57.568 54.529
565951 94.206 -57.591 54.532
565952 94.244 -57.614 54.530
565953 94.281 -57.637 54.522
565954 94.319 -57.660 54.507
565955 94.356 -57.683 54.491
565956 94.393 -57.705 54.474
565957 94.426 -57.719 54.455
565958 94.458 -57.731 54.433
565959 94.490 -57.744 54.409
565960 94.523 -57.756 54.384
565961 94.555 -57.768 54.361
565962 94.588 -57.781 54.340
565963 94.620 -57.793 54.324
565964 94.652 -57.806 54.309
565965 94.682 -57.816 54.292
565966 94.712 -57.826 54.273
565967 94.743 -57.836 54.250
565968 94.773 -57.846 54.225
565969 94.801 -57.857 54.196
565970 94.826 -57.869 54.165
565971 94.851 -57.881 54.130
565972 94.876 -57.893 54.094
565973 94.901 -57.905 54.056
565974 94.926 -57.918 54.014
565975 94.952 -57.930 53.970
565976 94.977 -57.942 53.922
565977 95.002 -57.954 53.871
565978 95.028 -57.966 53.817
565979 95.053 -57.978 53.762
565980 95.078 -57.990 53.705
565981 95.123 -58.000 53.647
565982 95.179 -58.010 53.587
565983 95.235 -58.019 53.529
565984 95.290 -58.028 53.468
565985 95.346 -58.037 53.407
565986 95.392 -58.041 53.348
565987 95.428 -58.040 53.289
565988 95.465 -58.040 53.236
565989 95.502 -58.039 53.184
565990 95.539 -58.038 53.133
565991 95.576 -58.037 53.086
565992 95.612 -58.037 53.045
565993 95.649 -58.036 53.008
565994 95.686 -58.035 52.981
565995 95.723 -58.034 52.958
565996 95.758 -58.034 52.938
565997 95.789 -58.043 52.923
565998 95.820 -58.054 52.913
565999 95.852 -58.063 52.899
566000 95.885 -58.075 52.880
566001 95.917 -58.085 52.864
566002 95.950 -58.096 52.852
566003 95.983 -58.107 52.843
566004 96.016 -58.118 52.837
566005 96.049 -58.129 52.830
566006 96.082 -58.140 52.824
566007 96.114 -58.151 52.818
566008 96.147 -58.162 52.811
566009 96.180 -58.173 52.800
566010 96.213 -58.184 52.786
566011 96.245 -58.195 52.769
566012 96.278 -58.206 52.746
566013 96.311 -58.217 52.725
566014 96.344 -58.228 52.702
566015 96.377 -58.238 52.676
566016 96.409 -58.250 52.648
566017 96.442 -58.260 52.623
566018 96.475 -58.272 52.597
566019 96.508 -58.282 52.570
566020 96.540 -58.294 52.544
566021 96.573 -58.304 52.517
566022 96.606 -58.316 52.491
566023 96.639 -58.326 52.466
566024 96.671 -58.331 52.438
566025 96.702 -58.328 52.408
566026 96.733 -58.327 52.379
566027 96.764 -58.325 52.351
566028 96.797 -58.332 52.325
566029 96.829 -58.338 52.300
566030 96.861 -58.344 52.271
566031 96.891 -58.347 52.240
566032 96.920 -58.350 52.207
566033 96.949 -58.354 52.175
566034 96.977 -58.357 52.142
566035 97.006 -58.359 52.112
566036 97.035 -58.362 52.083
566037 97.064 -58.365 52.057
566038 97.092 -58.369 52.034
566039 97.117 -58.373 52.012
566040 97.142 -58.379 51.992
566041 97.167 -58.384 51.968
566042 97.192 -58.388 51.941
566043 97.217 -58.394 51.909
566044 97.245 -58.394 51.875
566045 97.275 -58.391 51.838
566046 97.307 -58.389 51.795
566047 97.343 -58.388 51.748
566048 97.379 -58.388 51.694
566049 97.416 -58.387 51.636
566050 97.452 -58.387 51.575
566051 97.488 -58.386 51.518
566052 97.523 -58.387 51.462
566053 97.556 -58.389 51.407
566054 97.589 -58.392 51.354
566055 97.622 -58.395 51.302
566056 97.651 -58.403 51.251
566057 97.678 -58.412 51.203
566058 97.716 -58.412 51.156
566059 97.754 -58.412 51.112
566060 97.791 -58.412 51.069
566061 97.828 -58.413 51.025
566062 97.864 -58.414 50.980
566063 97.899 -58.416 50.934
566064 97.935 -58.418 50.887
566065 97.970 -58.419 50.842
566066 98.006 -58.421 50.798
566067 98.041 -58.422 50.753
566068 98.077 -58.425 50.710
566069 98.109 -58.426 50.668
566070 98.142 -58.426 50.629
566071 98.175 -58.427 50.591
566072 98.210 -58.428 50.553
566073 98.245 -58.429 50.516
566074 98.281 -58.429 50.476
566075 98.316 -58.429 50.439
566076 98.348 -58.425 50.401
566077 98.379 -58.419 50.364
566078 98.410 -58.412 50.326
566079 98.441 -58.406 50.288
566080 98.474 -58.404 50.252
566081 98.508 -58.404 50.219
566082 98.543 -58.403 50.189
566083 98.586 -58.403 50.161
566084 98.637 -58.403 50.135
566085 98.687 -58.402 50.112
566086 98.737 -58.402 50.092
566087 98.788 -58.402 50.075
566088 98.831 -58.401 50.060
566089 98.868 -58.400 50.045
566090 98.907 -58.385 50.031
566091 98.932 -58.381 50.018
566092 98.951 -58.382 50.002
566093 98.971 -58.384 49.985
566094 98.974 -58.393 49.967
566095 98.975 -58.402 49.947
566096 98.975 -58.410 49.924
566097 98.997 -58.413 49.899
566098 99.041 -58.412 49.872
566099 99.084 -58.411 49.845
566100 99.127 -58.376 49.818
566101 99.169 -58.339 49.792
566102 99.212 -58.303 49.767
566103 99.258 -58.329 49.744
566104 99.327 -58.357 49.721
566105 99.392 -58.368 49.699
566106 99.430 -58.376 49.676
566107 99.465 -58.385 49.652
566108 99.499 -58.394 49.629
566109 99.534 -58.403 49.606
566110 99.568 -58.411 49.582
566111 99.603 -58.420 49.557
566112 99.637 -58.429 49.532
566113 99.671 -58.438 49.507
566114 99.706 -58.447 49.479
566465 87.382 -54.822 57.602
566466 87.412 -54.840 57.600
566467 87.442 -54.857 57.596
566468 87.473 -54.874 57.594
566469 87.503 -54.891 57.591
566470 87.534 -54.909 57.586
566471 87.564 -54.925 57.581
566472 87.594 -54.943 57.576
566473 87.624 -54.960 57.571
566474 87.654 -54.977 57.566
566475 87.685 -54.994 57.560
566476 87.715 -55.012 57.552
566477 87.746 -55.028 57.543
566478 87.776 -55.046 57.532
566479 87.827 -55.057 57.521
566480 87.881 -55.074 57.506
566481 87.935 -55.090 57.491
566482 87.989 -55.106 57.473
566483 88.044 -55.123 57.453
566484 88.098 -55.138 57.428
566485 88.152 -55.154 57.401
566486 88.205 -55.171 57.373
566487 88.243 -55.188 57.342
566488 88.281 -55.204 57.308
566489 88.318 -55.220 57.268
566490 88.356 -55.237 57.224
566491 88.394 -55.254 57.180
566492 88.431 -55.270 57.134
566493 88.469 -55.287 57.088
566494 88.506 -55.304 57.041
566495 88.544 -55.320 56.992
566496 88.581 -55.337 56.945
566497 88.619 -55.353 56.898
566498 88.657 -55.370 56.852
566499 88.694 -55.386 56.806
566500 88.732 -55.403 56.762
566501 88.769 -55.420 56.718
566520 89.469 -55.709 55.967
566521 89.508 -55.730 55.945
566522 89.546 -55.754 55.924
566523 89.585 -55.777 55.905
566524 89.624 -55.801 55.888
566525 89.662 -55.824 55.871
566526 89.701 -55.847 55.857
566527 89.739 -55.871 55.845
566528 89.778 -55.894 55.834
566529 89.817 -55.918 55.824
566530 89.844 -55.934 55.817
566531 89.867 -55.948 55.811
566532 89.894 -55.968 55.807
566533 89.921 -55.987 55.804
566534 89.949 -56.006 55.802
566535 89.976 -56.025 55.800
566536 90.004 -56.041 55.798
566537 90.043 -56.051 55.796
566538 90.083 -56.062 55.796
566539 90.117 -56.074 55.797
566540 90.152 -56.087 55.799
566541 90.187 -56.099 55.800
566542 90.222 -56.112 55.801
566543 90.257 -56.124 55.802
566544 90.292 -56.137 55.802
566545 90.326 -56.149 55.800
566546 90.362 -56.162 55.799
566547 90.396 -56.175 55.795
566548 90.431 -56.187 55.789
566549 90.466 -56.200 55.783
566550 90.502 -56.212 55.776
566551 90.546 -56.222 55.767
566552 90.590 -56.233 55.756
566553 90.634 -56.243 55.743
566554 90.678 -56.254 55.731
566555 90.723 -56.264 55.718
566556 90.767 -56.275 55.705
566557 90.811 -56.285 55.691
566558 90.855 -56.296 55.678
566559 90.899 -56.307 55.664
566560 90.944 -56.317 55.649
566561 90.988 -56.327 55.636
566562 91.032 -56.338 55.625
566563 91.076 -56.348 55.615
566564 91.120 -56.359 55.603
566565 91.165 -56.370 55.591
566566 91.209 -56.379 55.581
566567 91.253 -56.390 55.574
566568 91.297 -56.401 55.570
566569 91.340 -56.414 55.560
566570 91.382 -56.427 55.547
566571 91.417 -56.444 55.530
566572 91.449 -56.457 55.509
566573 91.482 -56.469 55.484
566574 91.514 -56.481 55.459
566575 91.546 -56.493 55.433
566576 91.579 -56.505 55.411
566577 91.611 -56.518 55.392
566578 91.643 -56.530 55.377
566579 91.675 -56.542 55.364
566580 91.707 -56.555 55.356
566581 91.737 -56.567 55.352
566582 91.768 -56.580 55.353
566583 91.798 -56.593 55.358
566584 91.829 -56.606 55.368
566585 91.859 -56.619 55.385
566586 91.894 -56.627 55.400
566587 91.934 -56.646 55.410
566588 91.975 -56.667 55.415
566589 92.015 -56.687 55.415
566590 92.056 -56.708 55.408
566591 92.097 -56.729 55.397
566592 92.138 -56.749 55.374
566593 92.178 -56.771 55.354
566594 92.219 -56.791 55.342
566595 92.260 -56.812 55.339
566596 92.301 -56.833 55.342
566597 92.342 -56.853 55.343
566598 92.382 -56.874 55.342
566599 92.420 -56.910 55.339
566600 92.457 -56.946 55.336
566601 92.494 -56.984 55.336
566602 92.533 -57.006 55.340
566603 92.570 -57.025 55.344
566604 92.605 -57.047 55.344
566605 92.640 -57.067 55.352
566606 92.674 -57.088 55.361
566607 92.709 -57.109 55.365
566608 92.743 -57.130 55.360
566609 92.778 -57.151 55.346
566610 92.813 -57.172 55.327
566611 92.848 -57.193 55.310
566612 92.882 -57.213 55.297
566613 92.924 -57.228 55.287
566614 92.958 -57.238 55.273
566615 92.987 -57.244 55.253
566616 93.016 -57.251 55.228
566617 93.045 -57.257 55.201
566618 93.074 -57.264 55.173
566619 93.103 -57.270 55.145
566620 93.132 -57.277 55.117
566621 93.160 -57.284 55.093
566622 93.190 -57.290 55.071
566623 93.222 -57.302 55.048
566624 93.255 -57.314 55.022
566625 93.288 -57.326 54.993
566626 93.321 -57.338 54.963
566627 93.353 -57.350 54.936
566628 93.387 -57.355 54.913
566629 93.420 -57.359 54.894
566630 93.454 -57.363 54.882
566631 93.487 -57.366 54.873
566632 93.521 -57.370 54.861
566633 93.555 -57.373 54.847
566634 93.588 -57.377 54.828
566635 93.622 -57.381 54.806
566636 93.656 -57.385 54.781
566637 93.689 -57.388 54.754
566638 93.723 -57.392 54.725
566639 93.757 -57.396 54.694
566640 93.790 -57.400 54.660
566641 93.824 -57.404 54.625
566642 93.858 -57.407 54.592
566643 93.891 -57.411 54.557
566644 93.925 -57.415 54.529
566645 93.959 -57.418 54.509
566646 93.993 -57.425 54.497
566647 94.030 -57.448 54.491
566648 94.067 -57.470 54.493
566649 94.105 -57.493 54.495
566650 94.142 -57.516 54.497
566651 94.179 -57.539 54.500
566652 94.216 -57.562 54.503
566653 94.254 -57.585 54.502
566654 94.291 -57.608 54.496
566655 94.329 -57.630 54.482
566656 94.366 -57.653 54.468
566657 94.401 -57.672 54.453
566658 94.433 -57.684 54.436
566659 94.466 -57.697 54.417
566660 94.498 -57.709 54.394
566661 94.531 -57.722 54.370
566662 94.563 -57.734 54.346
566663 94.595 -57.746 54.326
566664 94.627 -57.759 54.310
566665 94.660 -57.771 54.296
566666 94.692 -57.783 54.282
566667 94.721 -57.793 54.264
566668 94.752 -57.803 54.242
566669 94.782 -57.813 54.218
566670 94.812 -57.824 54.189
566671 94.841 -57.833 54.158
566672 94.866 -57.846 54.124
566673 94.892 -57.858 54.089
566674 94.917 -57.870 54.052
566675 94.942 -57.882 54.011
566676 94.968 -57.894 53.969
566677 94.993 -57.906 53.924
566678 95.018 -57.918 53.873
566679 95.043 -57.930 53.820
566680 95.068 -57.943 53.766
566681 95.094 -57.955 53.711
566682 95.126 -57.966 53.653
566683 95.182 -57.975 53.595
566684 95.238 -57.984 53.535
566685 95.294 -57.994 53.475
566686 95.350 -58.003 53.415
566687 95.395 -58.007 53.354
566688 95.432 -58.006 53.292
566689 95.469 -58.006 53.237
566690 95.506 -58.005 53.184
566691 95.542 -58.004 53.131
566692 95.579 -58.003 53.079
566693 95.616 -58.003 53.034
566694 95.653 -58.002 52.996
566695 95.690 -58.001 52.964
566696 95.727 -58.000 52.935
566697 95.763 -57.999 52.911
566698 95.798 -58.002 52.892
566699 95.829 -58.012 52.874
566700 95.862 -58.023 52.856
566701 95.895 -58.034 52.839
566702 95.927 -58.045 52.824
566703 95.960 -58.056 52.812
566704 95.993 -58.067 52.802
566705 96.026 -58.078 52.793
566706 96.059 -58.089 52.788
566707 96.092 -58.100 52.781
566708 96.124 -58.111 52.776
566709 96.157 -58.122 52.771
566710 96.190 -58.133 52.762
566711 96.223 -58.144 52.752
566712 96.255 -58.155 52.740
566713 96.288 -58.165 52.721
566714 96.321 -58.177 52.702
566715 96.354 -58.187 52.681
566716 96.387 -58.198 52.658
566717 96.419 -58.209 52.633
566718 96.452 -58.220 52.608
566719 96.485 -58.231 52.584
566720 96.518 -58.242 52.558
566721 96.550 -58.253 52.532
566722 96.583 -58.264 52.506
566723 96.616 -58.275 52.478
566724 96.649 -58.286 52.452
566725 96.682 -58.297 52.425
566726 96.713 -58.299 52.398
566727 96.744 -58.297 52.372
566728 96.776 -58.296 52.350
566729 96.808 -58.302 52.326
566730 96.840 -58.308 52.301
566731 96.871 -58.313 52.274
566732 96.899 -58.316 52.245
566733 96.928 -58.318 52.215
566734 96.957 -58.321 52.185
566735 96.986 -58.324 52.154
566736 97.015 -58.327 52.125
566737 97.040 -58.332 52.098
566738 97.065 -58.337 52.074
566739 97.090 -58.342 52.055
566740 97.115 -58.347 52.035
566741 97.141 -58.352 52.015
566742 97.165 -58.357 51.992
566743 97.190 -58.363 51.965
566744 97.218 -58.364 51.935
566745 97.247 -58.361 51.901
566746 97.277 -58.359 51.863
566747 97.308 -58.357 51.821
566748 97.345 -58.356 51.776
566749 97.381 -58.356 51.727
566750 97.417 -58.355 51.673
566751 97.453 -58.354 51.614
566752 97.490 -58.354 51.557
566753 97.522 -58.357 51.500
566754 97.555 -58.360 51.444
566755 97.588 -58.363 51.391
566756 97.621 -58.365 51.339
566757 97.654 -58.368 51.288
566758 97.682 -58.376 51.239
566759 97.720 -58.376 51.191
566760 97.757 -58.377 51.143
566761 97.792 -58.379 51.096
566762 97.828 -58.380 51.047
566763 97.863 -58.382 50.998
566764 97.899 -58.384 50.951
566765 97.935 -58.385 50.903
566766 97.970 -58.387 50.857
566767 98.006 -58.388 50.812
566768 98.041 -58.391 50.767
566769 98.077 -58.392 50.723
566770 98.110 -58.393 50.679
566771 98.144 -58.394 50.638
566772 98.179 -58.394 50.598
566773 98.214 -58.395 50.557
566774 98.249 -58.395 50.517
566775 98.285 -58.396 50.475
566776 98.320 -58.396 50.435
566777 98.352 -58.391 50.396
566778 98.382 -58.385 50.358
566779 98.414 -58.378 50.321
566780 98.444 -58.372 50.286
566781 98.475 -58.365 50.253
566782 98.509 -58.363 50.220
566783 98.543 -58.363 50.190
566784 98.578 -58.363 50.163
566785 98.629 -58.363 50.138
566786 98.679 -58.362 50.114
566787 98.729 -58.362 50.094
566788 98.780 -58.362 50.077
566789 98.830 -58.362 50.061
566790 98.880 -58.361 50.046
566791 98.916 -58.356 50.032
566792 98.935 -58.357 50.018
566793 98.954 -58.359 50.001
566794 98.974 -58.360 49.984
566795 98.980 -58.367 49.965
566796 98.981 -58.376 49.944
566797 98.982 -58.386 49.921
566798 98.982 -58.394 49.896
566799 99.022 -58.394 49.869
566800 99.066 -58.378 49.843
566801 99.109 -58.341 49.816
566802 99.151 -58.304 49.789
566803 99.212 -58.265 49.763
566804 99.238 -58.266 49.740
566805 99.306 -58.327 49.717
566806 99.376 -58.341 49.694
566807 99.422 -58.349 49.670
566808 99.457 -58.357 49.647
566809 99.491 -58.366 49.623
566810 99.526 -58.375 49.599
566811 99.560 -58.384 49.576
566812 99.595 -58.392 49.551
566813 99.629 -58.401 49.526
566814 99.664 -58.410 49.500
566815 99.698 -58.419 49.473
567167 87.432 -54.792 57.599
567168 87.462 -54.809 57.596
567169 87.493 -54.826 57.594
567170 87.523 -54.844 57.592
567171 87.553 -54.861 57.588
567172 87.583 -54.878 57.584
567173 87.614 -54.895 57.579
567174 87.644 -54.913 57.576
567175 87.675 -54.929 57.571
567176 87.705 -54.947 57.565
567177 87.735 -54.964 57.558
567178 87.765 -54.981 57.549
567179 87.795 -54.998 57.539
567180 87.837 -55.015 57.529
567181 87.891 -55.031 57.516
567182 87.945 -55.047 57.502
567183 87.999 -55.063 57.486
567184 88.054 -55.079 57.467
567185 88.108 -55.096 57.444
567186 88.162 -55.112 57.420
567187 88.216 -55.128 57.393
567188 88.266 -55.144 57.364
567189 88.304 -55.161 57.333
567190 88.341 -55.177 57.294
567191 88.379 -55.194 57.251
567192 88.416 -55.210 57.208
567193 88.454 -55.227 57.163
567194 88.491 -55.244 57.117
567195 88.529 -55.260 57.071
567196 88.567 -55.277 57.021
567197 88.604 -55.294 56.973
567198 88.642 -55.311 56.925
567199 88.680 -55.326 56.878
567200 88.717 -55.343 56.831
567201 88.755 -55.360 56.785
567202 88.792 -55.376 56.740
567203 88.830 -55.393 56.697
567222 89.529 -55.699 55.944
567223 89.567 -55.722 55.923
567224 89.606 -55.746 55.902
567225 89.645 -55.769 55.884
567226 89.683 -55.793 55.867
567227 89.722 -55.815 55.852
567228 89.761 -55.839 55.839
567229 89.799 -55.862 55.827
567230 89.838 -55.886 55.817
567231 89.863 -55.901 55.808
567232 89.886 -55.915 55.801
567233 89.911 -55.932 55.796
567234 89.939 -55.952 55.791
567235 89.966 -55.971 55.787
567236 89.993 -55.991 55.785
567237 90.021 -56.008 55.782
567238 90.055 -56.018 55.781
567239 90.095 -56.028 55.781
567240 90.134 -56.038 55.783
567241 90.172 -56.049 55.785
567242 90.207 -56.062 55.786
567243 90.242 -56.074 55.787
567244 90.277 -56.087 55.789
567245 90.312 -56.099 55.790
567246 90.346 -56.112 55.790
567247 90.382 -56.124 55.789
567248 90.416 -56.137 55.787
567249 90.451 -56.149 55.783
567250 90.486 -56.162 55.778
567251 90.521 -56.174 55.774
567252 90.557 -56.186 55.768
567253 90.602 -56.197 55.760
567254 90.646 -56.207 55.750
567255 90.690 -56.218 55.739
567256 90.735 -56.229 55.729
567257 90.779 -56.239 55.718
567258 90.823 -56.250 55.707
567259 90.867 -56.260 55.695
567260 90.911 -56.270 55.682
567261 90.956 -56.281 55.669
567262 91.000 -56.291 55.655
567263 91.044 -56.302 55.644
567264 91.088 -56.313 55.633
567265 91.133 -56.323 55.621
567266 91.177 -56.333 55.609
567267 91.221 -56.344 55.597
567268 91.265 -56.354 55.587
567269 91.309 -56.365 55.582
567270 91.354 -56.376 55.575
567271 91.398 -56.386 55.564
567272 91.437 -56.398 55.548
567273 91.468 -56.410 55.528
567274 91.499 -56.423 55.504
567275 91.529 -56.436 55.478
567276 91.560 -56.449 55.450
567277 91.590 -56.462 55.425
567278 91.621 -56.475 55.404
567279 91.651 -56.487 55.385
567280 91.681 -56.500 55.370
567281 91.712 -56.513 55.359
567282 91.742 -56.526 55.351
567283 91.773 -56.539 55.347
567284 91.803 -56.552 55.349
567285 91.834 -56.564 55.356
567286 91.868 -56.573 55.370
567287 91.904 -56.580 55.380
567288 91.943 -56.595 55.387
567289 91.984 -56.616 55.392
567290 92.024 -56.637 55.392
567291 92.065 -56.658 55.384
567292 92.106 -56.678 55.372
567293 92.147 -56.699 55.354
567294 92.188 -56.720 55.342
567295 92.228 -56.741 55.341
567296 92.269 -56.762 55.347
567297 92.309 -56.782 55.354
567298 92.350 -56.804 55.359
567299 92.387 -56.841 55.360
567300 92.425 -56.877 55.359
567301 92.462 -56.915 55.355
567302 92.499 -56.950 55.349
567303 92.538 -56.969 55.346
567304 92.576 -56.989 55.355
567305 92.611 -57.010 55.364
567306 92.646 -57.031 55.371
567307 92.681 -57.052 55.375
567308 92.715 -57.072 55.377
567309 92.749 -57.094 55.372
567310 92.784 -57.114 55.359
567311 92.819 -57.135 55.341
567312 92.854 -57.156 55.327
567313 92.889 -57.177 55.316
567314 92.932 -57.190 55.305
567315 92.969 -57.201 55.290
567316 92.998 -57.207 55.272
567317 93.027 -57.214 55.249
567318 93.056 -57.220 55.223
567319 93.085 -57.227 55.195
567320 93.114 -57.234 55.166
567321 93.143 -57.240 55.137
567322 93.172 -57.247 55.112
567323 93.201 -57.254 55.087
567324 93.234 -57.266 55.061
567325 93.267 -57.278 55.035
567326 93.299 -57.289 55.010
567327 93.332 -57.301 54.982
567328 93.364 -57.313 54.954
567329 93.397 -57.325 54.927
567330 93.430 -57.331 54.902
567331 93.464 -57.335 54.882
567332 93.498 -57.338 54.871
567333 93.532 -57.342 54.862
567334 93.565 -57.346 54.851
567335 93.599 -57.350 54.836
567336 93.632 -57.353 54.815
567337 93.666 -57.357 54.792
567338 93.700 -57.360 54.766
567339 93.733 -57.364 54.737
567340 93.767 -57.368 54.707
567341 93.801 -57.372 54.675
567342 93.834 -57.376 54.639
567343 93.868 -57.379 54.603
567344 93.902 -57.383 54.565
567345 93.935 -57.387 54.530
567346 93.969 -57.391 54.504
567347 94.003 -57.395 54.484
567348 94.040 -57.418 54.470
567349 94.077 -57.441 54.467
567350 94.115 -57.464 54.468
567351 94.152 -57.486 54.468
567352 94.189 -57.510 54.469
567353 94.226 -57.532 54.471
567354 94.264 -57.555 54.470
567355 94.301 -57.578 54.464
567356 94.339 -57.601 54.453
567357 94.376 -57.623 54.439
567358 94.409 -57.638 54.425
567359 94.441 -57.650 54.410
567360 94.474 -57.662 54.395
567361 94.506 -57.675 54.374
567362 94.538 -57.687 54.352
567363 94.570 -57.699 54.332
567364 94.603 -57.712 54.314
567365 94.635 -57.724 54.300
567366 94.668 -57.737 54.286
567367 94.700 -57.749 54.270
567368 94.731 -57.760 54.251
567369 94.761 -57.771 54.230
567370 94.791 -57.780 54.205
567371 94.821 -57.790 54.178
567372 94.851 -57.800 54.147
567373 94.882 -57.811 54.114
567374 94.907 -57.822 54.079
567375 94.932 -57.834 54.043
567376 94.958 -57.846 54.004
567377 94.983 -57.858 53.963
567378 95.008 -57.871 53.919
567379 95.034 -57.883 53.871
567380 95.059 -57.895 53.819
567381 95.084 -57.907 53.766
567382 95.109 -57.919 53.711
567383 95.134 -57.931 53.655
567384 95.186 -57.941 53.597
567385 95.242 -57.950 53.537
567386 95.298 -57.959 53.478
567387 95.353 -57.968 53.420
567388 95.399 -57.972 53.359
567389 95.436 -57.971 53.296
567390 95.472 -57.971 53.239
567391 95.509 -57.970 53.185
567392 95.546 -57.969 53.130
567393 95.583 -57.968 53.076
567394 95.619 -57.968 53.027
567395 95.657 -57.967 52.984
567396 95.693 -57.966 52.948
567397 95.730 -57.965 52.916
567398 95.767 -57.965 52.889
567399 95.806 -57.966 52.866
567400 95.835 -57.974 52.843
567401 95.869 -57.985 52.823
567402 95.903 -57.995 52.806
567403 95.937 -58.006 52.791
567404 95.970 -58.016 52.778
567405 96.003 -58.027 52.767
567406 96.036 -58.037 52.756
567407 96.069 -58.049 52.748
567408 96.102 -58.059 52.740
567409 96.134 -58.071 52.735
567410 96.167 -58.081 52.732
567411 96.200 -58.093 52.726
567412 96.233 -58.103 52.719
567413 96.265 -58.115 52.709
567414 96.298 -58.125 52.695
567415 96.331 -58.137 52.678
567416 96.364 -58.147 52.659
567417 96.397 -58.159 52.637
567418 96.430 -58.169 52.616
567419 96.462 -58.180 52.594
567420 96.495 -58.191 52.573
567421 96.528 -58.202 52.550
567422 96.560 -58.213 52.525
567423 96.593 -58.224 52.500
567424 96.626 -58.235 52.474
567425 96.659 -58.246 52.447
567426 96.691 -58.257 52.421
567427 96.725 -58.268 52.394
567428 96.756 -58.267 52.368
567429 96.787 -58.267 52.345
567430 96.819 -58.273 52.323
567431 96.850 -58.278 52.299
567432 96.879 -58.280 52.274
567433 96.908 -58.283 52.248
567434 96.937 -58.286 52.221
567435 96.963 -58.291 52.193
567436 96.988 -58.296 52.165
567437 97.013 -58.300 52.138
567438 97.038 -58.306 52.113
567439 97.064 -58.311 52.092
567440 97.089 -58.316 52.073
567441 97.114 -58.321 52.056
567442 97.139 -58.326 52.035
567443 97.164 -58.331 52.011
567444 97.190 -58.335 51.986
567445 97.220 -58.332 51.957
567446 97.249 -58.329 51.924
567447 97.279 -58.326 51.888
567448 97.310 -58.325 51.847
567449 97.346 -58.324 51.804
567450 97.382 -58.323 51.758
567451 97.419 -58.322 51.708
567452 97.455 -58.322 51.652
567453 97.489 -58.325 51.595
567454 97.521 -58.327 51.538
567455 97.554 -58.330 51.483
567456 97.587 -58.333 51.430
567457 97.620 -58.335 51.377
567458 97.653 -58.338 51.325
567459 97.685 -58.341 51.276
567460 97.721 -58.343 51.226
567461 97.757 -58.344 51.174
567462 97.793 -58.346 51.121
567463 97.828 -58.347 51.068
567464 97.864 -58.350 51.015
567465 97.899 -58.351 50.965
567466 97.935 -58.353 50.916
567467 97.971 -58.354 50.869
567468 98.006 -58.357 50.824
567469 98.042 -58.358 50.778
567470 98.078 -58.360 50.734
567471 98.113 -58.360 50.691
567472 98.148 -58.361 50.648
567473 98.183 -58.362 50.606
567474 98.219 -58.362 50.563
567475 98.254 -58.363 50.519
567476 98.289 -58.363 50.475
567477 98.324 -58.363 50.432
567478 98.355 -58.357 50.392
567479 98.386 -58.350 50.354
567480 98.417 -58.344 50.318
567481 98.448 -58.338 50.284
567482 98.479 -58.331 50.252
567483 98.510 -58.325 50.220
567484 98.544 -58.323 50.191
567485 98.578 -58.322 50.165
567486 98.621 -58.322 50.140
567487 98.671 -58.322 50.117
567488 98.721 -58.322 50.097
567489 98.772 -58.322 50.079
567490 98.822 -58.322 50.063
567491 98.873 -58.321 50.048
567492 98.903 -58.324 50.032
567493 98.928 -58.328 50.016
567494 98.952 -58.332 49.999
567495 98.976 -58.336 49.982
567496 98.987 -58.341 49.962
567497 98.988 -58.350 49.940
567498 98.989 -58.360 49.917
567499 98.990 -58.369 49.892
567500 99.004 -58.375 49.865
567501 99.047 -58.343 49.839
567502 99.126 -58.300 49.813
567503 99.172 -58.285 49.787
567504 99.199 -58.286 49.762
567505 99.226 -58.288 49.738
567506 99.285 -58.297 49.715
567507 99.355 -58.311 49.691
567508 99.414 -58.321 49.667
567509 99.449 -58.330 49.642
567510 99.484 -58.338 49.618
567511 99.518 -58.347 49.594
567512 99.553 -58.356 49.570
567513 99.587 -58.365 49.546
567514 99.622 -58.373 49.521
567515 99.656 -58.382 49.495
567516 99.691 -58.391 49.468
567868 87.451 -54.744 57.598
567869 87.482 -54.762 57.597
567870 87.512 -54.779 57.595
567871 87.543 -54.796 57.594
567872 87.573 -54.813 57.591
567873 87.603 -54.831 57.586
567874 87.634 -54.847 57.583
567875 87.664 -54.865 57.579
567876 87.694 -54.882 57.574
567877 87.724 -54.899 57.570
567878 87.755 -54.918 57.564
567879 87.786 -54.941 57.556
567880 87.818 -54.964 57.547
567881 87.857 -54.969 57.537
567882 87.911 -54.981 57.525
567883 87.962 -55.000 57.512
567884 88.012 -55.019 57.498
567885 88.064 -55.037 57.480
567886 88.118 -55.053 57.459
567887 88.173 -55.069 57.436
567888 88.226 -55.085 57.411
567889 88.281 -55.101 57.385
567890 88.326 -55.118 57.356
567891 88.364 -55.134 57.321
567892 88.401 -55.151 57.278
567893 88.439 -55.167 57.235
567894 88.477 -55.184 57.191
567895 88.514 -55.201 57.146
567896 88.552 -55.217 57.099
567897 88.590 -55.234 57.050
567898 88.627 -55.251 57.001
567899 88.664 -55.265 56.953
567900 88.700 -55.279 56.905
567901 88.737 -55.294 56.857
567902 88.773 -55.308 56.809
567903 88.809 -55.322 56.763
567904 88.845 -55.336 56.719
567924 89.589 -55.690 55.921
567925 89.627 -55.714 55.899
567926 89.666 -55.737 55.880
567927 89.705 -55.760 55.863
567928 89.743 -55.784 55.847
567929 89.782 -55.807 55.832
567930 89.821 -55.831 55.820
567931 89.859 -55.853 55.809
567932 89.882 -55.868 55.799
567933 89.905 -55.882 55.791
567934 89.929 -55.897 55.785
567935 89.957 -55.916 55.780
567936 89.984 -55.936 55.776
567937 90.011 -55.955 55.772
567938 90.038 -55.974 55.769
567939 90.067 -55.984 55.767
567940 90.107 -55.994 55.768
567941 90.146 -56.005 55.770
567942 90.186 -56.015 55.772
567943 90.226 -56.025 55.775
567944 90.262 -56.037 55.776
567945 90.297 -56.049 55.778
567946 90.332 -56.062 55.781
567947 90.367 -56.074 55.782
567948 90.402 -56.087 55.783
567949 90.437 -56.099 55.782
567950 90.471 -56.112 55.780
567951 90.506 -56.124 55.778
567952 90.541 -56.137 55.776
567953 90.576 -56.149 55.773
567954 90.614 -56.161 55.768
567955 90.658 -56.172 55.760
567956 90.702 -56.182 55.752
567957 90.747 -56.193 55.743
567958 90.791 -56.203 55.734
567959 90.835 -56.213 55.724
567960 90.879 -56.224 55.713
567961 90.924 -56.235 55.701
567962 90.968 -56.245 55.689
567963 91.012 -56.256 55.676
567964 91.056 -56.266 55.665
567965 91.100 -56.276 55.653
567966 91.145 -56.287 55.641
567967 91.189 -56.298 55.627
567968 91.233 -56.308 55.614
567969 91.277 -56.319 55.602
567970 91.321 -56.329 55.595
567971 91.366 -56.339 55.588
567972 91.410 -56.350 55.578
567973 91.442 -56.356 55.563
567974 91.473 -56.369 55.543
567975 91.503 -56.382 55.519
567976 91.534 -56.395 55.492
567977 91.564 -56.407 55.465
567978 91.595 -56.420 55.439
567979 91.625 -56.432 55.416
567980 91.656 -56.445 55.395
567981 91.686 -56.458 55.375
567982 91.717 -56.471 55.361
567983 91.747 -56.484 55.350
567984 91.777 -56.497 55.344
567985 91.808 -56.510 55.340
567986 91.842 -56.519 55.340
567987 91.878 -56.526 55.346
567988 91.914 -56.533 55.350
567989 91.952 -56.545 55.356
567990 91.993 -56.566 55.362
567991 92.033 -56.586 55.363
567992 92.074 -56.608 55.357
567993 92.115 -56.628 55.349
567994 92.155 -56.649 55.340
567995 92.196 -56.670 55.338
567996 92.237 -56.690 55.344
567997 92.278 -56.711 55.355
567998 92.318 -56.735 55.365
567999 92.355 -56.772 55.372
568000 92.392 -56.809 55.373
568001 92.429 -56.846 55.371
568002 92.466 -56.883 55.367
568003 92.505 -56.914 55.363
568004 92.544 -56.933 55.362
568005 92.582 -56.953 55.369
568006 92.617 -56.973 55.378
568007 92.652 -56.994 55.384
568008 92.687 -57.015 55.385
568009 92.721 -57.036 55.384
568010 92.756 -57.056 55.378
568011 92.790 -57.078 55.369
568012 92.825 -57.099 55.357
568013 92.860 -57.119 55.346
568014 92.897 -57.138 55.335
568015 92.940 -57.152 55.321
568016 92.980 -57.164 55.304
568017 93.009 -57.170 55.286
568018 93.038 -57.177 55.263
568019 93.067 -57.184 55.237
568020 93.096 -57.190 55.211
568021 93.125 -57.197 55.180
568022 93.154 -57.203 55.151
568023 93.183 -57.210 55.126
568024 93.212 -57.217 55.100
568025 93.245 -57.229 55.072
568026 93.278 -57.241 55.046
568027 93.310 -57.253 55.023
568028 93.343 -57.265 54.996
568029 93.376 -57.276 54.967
568030 93.409 -57.288 54.937
568031 93.441 -57.301 54.908
568032 93.475 -57.307 54.884
568033 93.508 -57.310 54.867
568034 93.542 -57.314 54.855
568035 93.575 -57.318 54.844
568036 93.609 -57.322 54.828
568037 93.643 -57.326 54.809
568038 93.677 -57.329 54.787
568039 93.710 -57.333 54.763
568040 93.744 -57.336 54.736
568041 93.777 -57.340 54.709
568042 93.811 -57.344 54.679
568043 93.845 -57.348 54.646
568044 93.878 -57.351 54.611
568045 93.912 -57.355 54.576
568046 93.946 -57.359 54.541
568047 93.979 -57.363 54.509
568048 94.013 -57.366 54.480
568049 94.050 -57.388 54.455
568050 94.087 -57.411 54.442
568051 94.125 -57.434 54.440
568052 94.162 -57.457 54.439
568053 94.199 -57.480 54.439
568054 94.237 -57.503 54.438
568055 94.274 -57.526 54.434
568056 94.312 -57.548 54.428
568057 94.349 -57.571 54.419
568058 94.384 -57.591 54.405
568059 94.417 -57.603 54.392
568060 94.449 -57.615 54.381
568061 94.481 -57.628 54.368
568062 94.513 -57.640 54.351
568063 94.546 -57.653 54.331
568064 94.578 -57.665 54.313
568065 94.611 -57.677 54.299
568066 94.643 -57.690 54.288
568067 94.675 -57.702 54.274
568068 94.707 -57.714 54.257
568069 94.740 -57.727 54.238
568070 94.770 -57.737 54.216
568071 94.801 -57.748 54.192
568072 94.830 -57.758 54.164
568073 94.860 -57.767 54.133
568074 94.891 -57.777 54.101
568075 94.921 -57.788 54.067
568076 94.948 -57.799 54.031
568077 94.973 -57.811 53.993
568078 94.998 -57.823 53.952
568079 95.024 -57.835 53.910
568080 95.049 -57.847 53.863
568081 95.074 -57.859 53.812
568082 95.100 -57.871 53.759
568083 95.125 -57.883 53.704
568084 95.150 -57.896 53.648
568085 95.190 -57.906 53.592
568086 95.246 -57.915 53.532
568087 95.302 -57.925 53.474
568088 95.357 -57.934 53.417
568089 95.402 -57.938 53.359
568090 95.439 -57.937 53.302
568091 95.476 -57.937 53.244
568092 95.513 -57.936 53.187
568093 95.549 -57.935 53.132
568094 95.586 -57.934 53.078
568095 95.623 -57.934 53.025
568096 95.660 -57.933 52.975
568097 95.697 -57.932 52.935
568098 95.734 -57.931 52.901
568099 95.773 -57.933 52.871
568100 95.811 -57.937 52.843
568101 95.835 -57.944 52.818
568102 95.866 -57.952 52.798
568103 95.899 -57.962 52.778
568104 95.933 -57.973 52.762
568105 95.967 -57.984 52.748
568106 96.001 -57.993 52.736
568107 96.035 -58.004 52.724
568108 96.069 -58.014 52.714
568109 96.103 -58.024 52.705
568110 96.136 -58.035 52.699
568111 96.170 -58.045 52.696
568112 96.204 -58.056 52.693
568113 96.238 -58.066 52.687
568114 96.272 -58.076 52.678
568115 96.306 -58.087 52.667
568116 96.340 -58.097 52.653
568117 96.374 -58.107 52.636
568118 96.407 -58.118 52.616
568119 96.440 -58.129 52.598
568120 96.472 -58.140 52.579
568121 96.505 -58.151 52.559
568122 96.538 -58.162 52.537
568123 96.571 -58.173 52.514
568124 96.603 -58.184 52.492
568125 96.636 -58.195 52.467
568126 96.669 -58.206 52.441
568127 96.701 -58.217 52.416
568128 96.735 -58.228 52.391
568129 96.767 -58.238 52.367
568130 96.799 -58.238 52.346
568131 96.830 -58.242 52.325
568132 96.859 -58.245 52.303
568133 96.886 -58.249 52.279
568134 96.911 -58.254 52.256
568135 96.936 -58.260 52.233
568136 96.961 -58.264 52.208
568137 96.986 -58.269 52.184
568138 97.012 -58.275 52.161
568139 97.037 -58.279 52.139
568140 97.062 -58.285 52.118
568141 97.087 -58.290 52.097
568142 97.112 -58.294 52.076
568143 97.137 -58.300 52.054
568144 97.162 -58.304 52.031
568145 97.192 -58.302 52.005
568146 97.222 -58.300 51.978
568147 97.251 -58.297 51.950
568148 97.281 -58.294 51.915
568149 97.311 -58.292 51.875
568150 97.348 -58.291 51.832
568151 97.384 -58.291 51.786
568152 97.420 -58.291 51.737
568153 97.454 -58.292 51.683
568154 97.488 -58.294 51.628
568155 97.520 -58.297 51.571
568156 97.553 -58.300 51.517
568157 97.586 -58.303 51.465
568158 97.619 -58.306 51.415
568159 97.653 -58.306 51.362
568160 97.688 -58.306 51.309
568161 97.726 -58.310 51.255
568162 97.762 -58.312 51.199
568163 97.798 -58.313 51.142
568164 97.833 -58.316 51.086
568165 97.869 -58.317 51.030
568166 97.905 -58.319 50.978
568167 97.940 -58.320 50.928
568168 97.976 -58.322 50.880
568169 98.012 -58.324 50.832
568170 98.048 -58.325 50.785
568171 98.083 -58.327 50.744
568172 98.118 -58.328 50.703
568173 98.153 -58.328 50.661
568174 98.188 -58.328 50.617
568175 98.223 -58.329 50.570
568176 98.258 -58.329 50.524
568177 98.294 -58.330 50.477
568178 98.328 -58.328 50.431
568179 98.359 -58.322 50.389
568180 98.390 -58.316 50.350
568181 98.421 -58.310 50.315
568182 98.451 -58.303 50.282
568183 98.483 -58.297 50.248
568184 98.513 -58.291 50.218
568185 98.544 -58.284 50.191
568186 98.578 -58.282 50.166
568187 98.613 -58.282 50.142
568188 98.663 -58.281 50.122
568189 98.714 -58.281 50.100
568190 98.764 -58.281 50.082
568191 98.814 -58.281 50.066
568192 98.853 -58.283 50.049
568193 98.878 -58.287 50.031
568194 98.903 -58.291 50.013
568195 98.927 -58.295 49.997
568196 98.954 -58.309 49.980
568197 98.962 -58.319 49.960
568198 98.966 -58.328 49.937
568199 98.970 -58.338 49.913
568200 98.973 -58.347 49.886
568201 99.024 -58.331 49.860
568202 99.107 -58.299 49.835
568203 99.138 -58.290 49.810
568204 99.169 -58.280 49.786
568205 99.200 -58.271 49.762
568206 99.231 -58.262 49.737
568207 99.265 -58.269 49.714
568208 99.334 -58.282 49.689
568209 99.400 -58.294 49.664
568210 99.442 -58.302 49.640
568211 99.476 -58.310 49.615
568212 99.511 -58.319 49.590
568213 99.545 -58.328 49.566
568214 99.580 -58.337 49.541
568215 99.614 -58.346 49.516
568216 99.648 -58.354 49.490
568217 99.683 -58.363 49.463
568570 87.502 -54.714 57.597
568571 87.532 -54.731 57.596
568572 87.562 -54.748 57.595
568573 87.592 -54.765 57.593
568574 87.623 -54.783 57.589
568575 87.653 -54.800 57.586
568576 87.684 -54.818 57.582
568577 87.715 -54.842 57.578
568578 87.746 -54.866 57.574
568579 87.777 -54.888 57.569
568580 87.808 -54.912 57.562
568581 87.842 -54.929 57.554
568582 87.877 -54.944 57.545
568583 87.933 -54.934 57.534
568584 87.986 -54.947 57.522
568585 88.037 -54.966 57.509
568586 88.088 -54.984 57.493
568587 88.139 -55.002 57.473
568588 88.190 -55.021 57.451
568589 88.241 -55.039 57.428
568590 88.292 -55.058 57.403
568591 88.345 -55.075 57.377
568592 88.387 -55.091 57.344
568593 88.424 -55.107 57.303
568594 88.461 -55.123 57.261
568595 88.497 -55.136 57.218
568596 88.533 -55.151 57.174
568597 88.570 -55.165 57.129
568598 88.606 -55.179 57.079
568599 88.642 -55.193 57.029
568600 88.679 -55.207 56.980
568601 88.715 -55.222 56.931
568602 88.751 -55.236 56.882
568603 88.787 -55.250 56.835
568604 88.824 -55.264 56.788
568605 88.860 -55.279 56.742
568625 89.610 -55.658 55.918
568626 89.649 -55.682 55.897
568627 89.687 -55.705 55.877
568628 89.726 -55.728 55.858
568629 89.765 -55.752 55.842
568630 89.803 -55.775 55.827
568631 89.842 -55.799 55.813
568632 89.878 -55.821 55.801
568633 89.901 -55.834 55.791
568634 89.924 -55.849 55.783
568635 89.947 -55.862 55.776
568636 89.974 -55.881 55.771
568637 90.002 -55.900 55.766
568638 90.029 -55.920 55.762
568639 90.056 -55.940 55.759
568640 90.082 -55.951 55.758
568641 90.118 -55.961 55.759
568642 90.158 -55.972 55.760
568643 90.198 -55.981 55.763
568644 90.237 -55.991 55.766
568645 90.277 -56.002 55.767
568646 90.317 -56.012 55.769
568647 90.352 -56.024 55.773
568648 90.387 -56.037 55.776
568649 90.422 -56.049 55.779
568650 90.457 -56.062 55.780
568651 90.492 -56.074 55.780
568652 90.526 -56.087 55.780
568653 90.561 -56.099 55.780
568654 90.596 -56.112 55.780
568655 90.631 -56.124 55.777
568656 90.670 -56.136 55.773
568657 90.715 -56.146 55.767
568658 90.758 -56.157 55.760
568659 90.803 -56.167 55.752
568660 90.847 -56.178 55.744
568661 90.891 -56.188 55.734
568662 90.936 -56.199 55.724
568663 90.980 -56.210 55.712
568664 91.024 -56.219 55.699
568665 91.068 -56.230 55.688
568666 91.112 -56.241 55.676
568667 91.157 -56.251 55.662
568668 91.201 -56.262 55.647
568669 91.245 -56.273 55.633
568670 91.289 -56.282 55.619
568671 91.333 -56.293 55.611
568672 91.378 -56.304 55.602
568673 91.414 -56.311 55.592
568674 91.446 -56.317 55.576
568675 91.477 -56.327 55.555
568676 91.508 -56.340 55.530
568677 91.538 -56.353 55.504
568678 91.569 -56.366 55.478
568679 91.599 -56.378 55.452
568680 91.630 -56.391 55.427
568681 91.660 -56.404 55.404
568682 91.690 -56.417 55.381
568683 91.721 -56.429 55.364
568684 91.751 -56.442 55.351
568685 91.782 -56.455 55.341
568686 91.815 -56.465 55.334
568687 91.852 -56.472 55.330
568688 91.887 -56.479 55.330
568689 91.924 -56.486 55.333
568690 91.960 -56.495 55.335
568691 92.001 -56.515 55.338
568692 92.042 -56.536 55.340
568693 92.083 -56.557 55.339
568694 92.124 -56.578 55.338
568695 92.164 -56.599 55.339
568696 92.205 -56.619 55.344
568697 92.246 -56.640 55.352
568698 92.285 -56.666 55.363
568699 92.322 -56.703 55.373
568700 92.360 -56.740 55.382
568701 92.397 -56.777 55.382
568702 92.434 -56.814 55.379
568703 92.472 -56.850 55.377
568704 92.510 -56.877 55.378
568705 92.549 -56.897 55.381
568706 92.588 -56.916 55.385
568707 92.623 -56.937 55.391
568708 92.658 -56.958 55.394
568709 92.693 -56.978 55.393
568710 92.727 -57.000 55.390
568711 92.762 -57.020 55.384
568712 92.797 -57.041 55.376
568713 92.831 -57.062 55.366
568714 92.866 -57.083 55.354
568715 92.905 -57.100 55.339
568716 92.948 -57.113 55.322
568717 92.991 -57.127 55.302
568718 93.020 -57.134 55.283
568719 93.049 -57.140 55.262
568720 93.078 -57.147 55.239
568721 93.107 -57.153 55.213
568722 93.136 -57.160 55.186
568723 93.165 -57.166 55.161
568724 93.194 -57.172 55.139
568725 93.224 -57.181 55.116
568726 93.257 -57.193 55.091
568727 93.289 -57.204 55.061
568728 93.322 -57.216 55.034
568729 93.355 -57.228 55.003
568730 93.388 -57.240 54.974
568731 93.420 -57.252 54.945
568732 93.452 -57.264 54.915
568733 93.485 -57.275 54.888
568734 93.518 -57.282 54.865
568735 93.552 -57.286 54.848
568736 93.586 -57.290 54.833
568737 93.620 -57.294 54.816
568738 93.653 -57.297 54.797
568739 93.687 -57.301 54.776
568740 93.720 -57.305 54.754
568741 93.754 -57.309 54.730
568742 93.788 -57.313 54.704
568743 93.822 -57.316 54.676
568744 93.855 -57.319 54.647
568745 93.889 -57.323 54.616
568746 93.922 -57.327 54.585
568747 93.956 -57.331 54.553
568748 93.990 -57.335 54.521
568749 94.024 -57.338 54.489
568750 94.060 -57.359 54.457
568751 94.098 -57.382 54.433
568752 94.135 -57.405 54.420
568753 94.172 -57.428 54.416
568754 94.209 -57.451 54.414
568755 94.247 -57.473 54.409
568756 94.284 -57.496 54.401
568757 94.322 -57.519 54.390
568758 94.359 -57.542 54.378
568759 94.392 -57.556 54.365
568760 94.424 -57.569 54.355
568761 94.457 -57.581 54.346
568762 94.489 -57.593 54.336
568763 94.522 -57.606 54.322
568764 94.554 -57.618 54.304
568765 94.586 -57.630 54.289
568766 94.618 -57.643 54.279
568767 94.650 -57.655 54.271
568768 94.683 -57.668 54.259
568769 94.715 -57.680 54.243
568770 94.748 -57.692 54.223
568771 94.779 -57.705 54.201
568772 94.810 -57.714 54.176
568773 94.840 -57.724 54.149
568774 94.870 -57.735 54.118
568775 94.900 -57.745 54.085
568776 94.930 -57.755 54.052
568777 94.960 -57.764 54.017
568778 94.989 -57.775 53.980
568779 95.014 -57.787 53.939
568780 95.039 -57.799 53.899
568781 95.065 -57.811 53.853
568782 95.090 -57.824 53.805
568783 95.115 -57.836 53.752
568784 95.140 -57.848 53.697
568785 95.166 -57.860 53.640
568786 95.193 -57.872 53.584
568787 95.249 -57.881 53.526
568788 95.305 -57.890 53.468
568789 95.361 -57.899 53.410
568790 95.406 -57.903 53.353
568791 95.443 -57.902 53.298
568792 95.479 -57.902 53.243
568793 95.516 -57.901 53.189
568794 95.553 -57.900 53.136
568795 95.590 -57.899 53.084
568796 95.627 -57.899 53.030
568797 95.664 -57.898 52.980
568798 95.700 -57.897 52.934
568799 95.740 -57.899 52.893
568800 95.784 -57.904 52.856
568801 95.816 -57.909 52.825
568802 95.837 -57.915 52.797
568803 95.867 -57.922 52.774
568804 95.896 -57.930 52.755
568805 95.930 -57.940 52.739
568806 95.964 -57.950 52.725
568807 95.997 -57.961 52.713
568808 96.032 -57.971 52.701
568809 96.065 -57.981 52.690
568810 96.099 -57.992 52.681
568811 96.133 -58.003 52.674
568812 96.167 -58.012 52.669
568813 96.201 -58.023 52.664
568814 96.235 -58.034 52.653
568815 96.269 -58.043 52.645
568816 96.303 -58.054 52.637
568817 96.336 -58.065 52.626
568818 96.371 -58.075 52.612
568819 96.404 -58.085 52.597
568820 96.438 -58.095 52.579
568821 96.472 -58.106 52.562
568822 96.506 -58.116 52.543
568823 96.540 -58.126 52.522
568824 96.574 -58.137 52.502
568825 96.608 -58.147 52.481
568826 96.642 -58.157 52.458
568827 96.675 -58.168 52.434
568828 96.710 -58.178 52.410
568829 96.743 -58.188 52.387
568830 96.777 -58.199 52.367
568831 96.809 -58.208 52.348
568832 96.834 -58.212 52.329
568833 96.859 -58.218 52.309
568834 96.884 -58.223 52.288
568835 96.909 -58.228 52.268
568836 96.935 -58.233 52.247
568837 96.960 -58.238 52.226
568838 96.985 -58.243 52.206
568839 97.010 -58.248 52.186
568840 97.035 -58.253 52.166
568841 97.060 -58.258 52.146
568842 97.085 -58.263 52.123
568843 97.110 -58.269 52.100
568844 97.136 -58.273 52.077
568845 97.164 -58.272 52.053
568846 97.194 -58.270 52.027
568847 97.224 -58.267 52.001
568848 97.253 -58.265 51.974
568849 97.283 -58.262 51.940
568850 97.313 -58.259 51.901
568851 97.349 -58.259 51.858
568852 97.385 -58.259 51.812
568853 97.421 -58.259 51.763
568854 97.453 -58.262 51.710
568855 97.486 -58.265 51.655
568856 97.519 -58.268 51.598
568857 97.552 -58.270 51.544
568858 97.587 -58.271 51.492
568859 97.621 -58.271 51.442
568860 97.656 -58.272 51.389
568861 97.690 -58.272 51.334
568862 97.728 -58.275 51.276
568863 97.768 -58.280 51.218
568864 97.803 -58.281 51.159
568865 97.839 -58.283 51.101
568866 97.874 -58.285 51.044
568867 97.910 -58.286 50.990
568868 97.946 -58.288 50.938
568869 97.982 -58.290 50.887
568870 98.017 -58.291 50.838
568871 98.053 -58.293 50.791
568872 98.088 -58.294 50.749
568873 98.123 -58.295 50.708
568874 98.157 -58.295 50.665
568875 98.193 -58.296 50.621
568876 98.228 -58.296 50.575
568877 98.263 -58.297 50.526
568878 98.298 -58.297 50.477
568879 98.332 -58.294 50.431
568880 98.363 -58.288 50.388
568881 98.393 -58.281 50.349
568882 98.424 -58.275 50.313
568883 98.455 -58.269 50.280
568884 98.486 -58.263 50.250
568885 98.517 -58.256 50.220
568886 98.548 -58.250 50.193
568887 98.579 -58.243 50.168
568888 98.613 -58.242 50.144
568889 98.655 -58.241 50.124
568890 98.706 -58.241 50.103
568891 98.756 -58.241 50.085
568892 98.803 -58.241 50.067
568893 98.828 -58.246 50.048
568894 98.853 -58.250 50.029
568895 98.878 -58.259 50.010
568896 98.906 -58.275 49.994
568897 98.928 -58.289 49.977
568898 98.932 -58.298 49.956
568899 98.935 -58.307 49.933
568900 98.938 -58.316 49.909
568901 98.976 -58.308 49.883
568902 99.060 -58.275 49.858
568903 99.102 -58.268 49.834
568904 99.144 -58.261 49.812
568905 99.175 -58.252 49.789
568906 99.207 -58.243 49.764
568907 99.237 -58.233 49.739
568908 99.251 -58.239 49.714
568909 99.313 -58.253 49.688
568910 99.382 -58.266 49.663
568911 99.434 -58.274 49.638
568912 99.469 -58.283 49.612
568913 99.503 -58.291 49.587
568914 99.538 -58.300 49.562
568915 99.572 -58.310 49.537
568916 99.607 -58.318 49.512
568917 99.641 -58.327 49.486
568918 99.676 -58.335 49.459
569272 87.552 -54.684 57.596
569273 87.582 -54.701 57.596
569274 87.613 -54.718 57.595
569275 87.643 -54.742 57.593
569276 87.675 -54.765 57.588
569277 87.705 -54.789 57.585
569278 87.737 -54.812 57.582
569279 87.767 -54.835 57.578
569280 87.798 -54.859 57.575
569281 87.831 -54.878 57.568
569282 87.867 -54.894 57.561
569283 87.902 -54.909 57.552
569284 87.953 -54.894 57.542
569285 88.009 -54.898 57.532
569286 88.062 -54.912 57.520
569287 88.113 -54.931 57.504
569288 88.164 -54.949 57.486
569289 88.215 -54.968 57.466
569290 88.266 -54.986 57.444
569291 88.317 -55.005 57.420
569292 88.363 -55.022 57.394
569293 88.403 -55.036 57.363
569294 88.439 -55.050 57.326
569295 88.476 -55.064 57.287
569296 88.512 -55.079 57.245
569297 88.548 -55.093 57.201
569298 88.585 -55.107 57.156
569299 88.621 -55.121 57.106
569300 88.657 -55.135 57.057
569301 88.693 -55.150 57.007
569302 88.730 -55.164 56.957
569303 88.766 -55.178 56.908
569304 88.803 -55.192 56.859
569305 88.839 -55.207 56.811
569306 88.875 -55.221 56.764
569307 88.911 -55.235 56.718
569327 89.670 -55.650 55.894
569328 89.709 -55.673 55.874
569329 89.747 -55.696 55.855
569330 89.786 -55.720 55.837
569331 89.825 -55.743 55.822
569332 89.863 -55.767 55.808
569333 89.897 -55.787 55.795
569334 89.920 -55.801 55.785
569335 89.943 -55.815 55.776
569336 89.966 -55.830 55.769
569337 89.991 -55.846 55.763
569338 90.019 -55.865 55.758
569339 90.046 -55.885 55.754
569340 90.074 -55.904 55.751
569341 90.100 -55.918 55.750
569342 90.130 -55.928 55.750
569343 90.170 -55.937 55.752
569344 90.210 -55.948 55.756
569345 90.249 -55.958 55.759
569346 90.289 -55.969 55.762
569347 90.328 -55.978 55.765
569348 90.368 -55.988 55.770
569349 90.407 -55.999 55.774
569350 90.442 -56.012 55.778
569351 90.477 -56.024 55.781
569352 90.512 -56.037 55.783
569353 90.546 -56.049 55.785
569354 90.581 -56.062 55.787
569355 90.616 -56.074 55.790
569356 90.651 -56.087 55.789
569357 90.686 -56.099 55.787
569358 90.727 -56.110 55.784
569359 90.770 -56.121 55.778
569360 90.815 -56.132 55.773
569361 90.859 -56.142 55.766
569362 90.903 -56.153 55.758
569363 90.948 -56.163 55.748
569364 90.992 -56.173 55.737
569365 91.036 -56.184 55.726
569366 91.080 -56.194 55.715
569367 91.124 -56.205 55.700
569368 91.169 -56.216 55.685
569369 91.213 -56.226 55.669
569370 91.257 -56.236 55.652
569371 91.301 -56.247 55.635
569372 91.346 -56.257 55.622
569373 91.385 -56.266 55.610
569374 91.418 -56.272 55.598
569375 91.450 -56.277 55.582
569376 91.483 -56.285 55.561
569377 91.513 -56.298 55.538
569378 91.543 -56.311 55.516
569379 91.574 -56.323 55.491
569380 91.604 -56.336 55.465
569381 91.635 -56.349 55.438
569382 91.665 -56.362 55.411
569383 91.696 -56.375 55.387
569384 91.726 -56.388 55.367
569385 91.756 -56.400 55.352
569386 91.789 -56.410 55.340
569387 91.825 -56.417 55.330
569388 91.861 -56.424 55.324
569389 91.897 -56.432 55.321
569390 91.934 -56.439 55.322
569391 91.969 -56.445 55.326
569392 92.010 -56.465 55.331
569393 92.051 -56.486 55.336
569394 92.092 -56.507 55.340
569395 92.132 -56.527 55.347
569396 92.173 -56.548 55.358
569397 92.214 -56.569 55.369
569398 92.253 -56.597 55.379
569399 92.290 -56.634 55.387
569400 92.328 -56.671 55.392
569401 92.365 -56.708 55.395
569402 92.402 -56.745 55.395
569403 92.439 -56.781 55.395
569404 92.477 -56.818 55.396
569405 92.515 -56.841 55.398
569406 92.554 -56.861 55.400
569407 92.593 -56.880 55.402
569408 92.630 -56.900 55.404
569409 92.664 -56.921 55.403
569410 92.699 -56.942 55.398
569411 92.733 -56.962 55.392
569412 92.768 -56.984 55.387
569413 92.803 -57.004 55.379
569414 92.838 -57.025 55.370
569415 92.872 -57.046 55.357
569416 92.914 -57.061 55.341
569417 92.956 -57.075 55.324
569418 93.000 -57.088 55.306
569419 93.031 -57.097 55.288
569420 93.060 -57.103 55.268
569421 93.089 -57.109 55.248
569422 93.118 -57.116 55.226
569423 93.147 -57.122 55.206
569424 93.176 -57.129 55.189
569425 93.205 -57.136 55.171
569426 93.235 -57.144 55.151
569427 93.268 -57.156 55.128
569428 93.301 -57.168 55.099
569429 93.334 -57.180 55.069
569430 93.366 -57.191 55.037
569431 93.399 -57.203 55.003
569432 93.431 -57.216 54.968
569433 93.464 -57.227 54.933
569434 93.497 -57.239 54.902
569435 93.530 -57.251 54.873
569436 93.563 -57.258 54.847
569437 93.596 -57.262 54.824
569438 93.630 -57.266 54.802
569439 93.663 -57.269 54.780
569440 93.697 -57.273 54.761
569441 93.731 -57.277 54.739
569442 93.765 -57.281 54.717
569443 93.798 -57.285 54.693
569444 93.832 -57.288 54.669
569445 93.865 -57.292 54.644
569446 93.899 -57.296 54.618
569447 93.933 -57.299 54.593
569448 93.967 -57.303 54.564
569449 94.000 -57.307 54.534
569450 94.034 -57.310 54.502
569451 94.070 -57.330 54.466
569452 94.108 -57.353 54.433
569453 94.145 -57.376 54.408
569454 94.182 -57.398 54.395
569455 94.219 -57.421 54.389
569456 94.257 -57.444 54.383
569457 94.294 -57.467 54.370
569458 94.332 -57.489 54.356
569459 94.367 -57.509 54.341
569460 94.400 -57.522 54.328
569461 94.432 -57.534 54.317
569462 94.464 -57.546 54.309
569463 94.497 -57.559 54.298
569464 94.529 -57.571 54.284
569465 94.561 -57.584 54.269
569466 94.594 -57.596 54.258
569467 94.626 -57.608 54.252
569468 94.659 -57.621 54.247
569469 94.691 -57.633 54.237
569470 94.723 -57.646 54.223
569471 94.755 -57.658 54.205
569472 94.787 -57.670 54.184
569473 94.819 -57.682 54.159
569474 94.849 -57.692 54.131
569475 94.879 -57.702 54.101
569476 94.909 -57.712 54.068
569477 94.939 -57.722 54.034
569478 94.969 -57.732 54.000
569479 94.999 -57.742 53.964
569480 95.030 -57.752 53.926
569481 95.055 -57.764 53.886
569482 95.080 -57.776 53.842
569483 95.106 -57.788 53.796
569484 95.131 -57.800 53.746
569485 95.156 -57.812 53.691
569486 95.181 -57.824 53.632
569487 95.206 -57.836 53.575
569488 95.253 -57.847 53.518
569489 95.309 -57.856 53.460
569490 95.365 -57.865 53.401
569491 95.409 -57.868 53.343
569492 95.446 -57.868 53.289
569493 95.483 -57.867 53.237
569494 95.520 -57.867 53.187
569495 95.557 -57.866 53.139
569496 95.594 -57.865 53.089
569497 95.630 -57.865 53.036
569498 95.667 -57.864 52.985
569499 95.707 -57.865 52.936
569500 95.751 -57.870 52.891
569501 95.795 -57.875 52.850
569502 95.822 -57.880 52.813
569503 95.842 -57.886 52.781
569504 95.869 -57.893 52.755
569505 95.898 -57.900 52.735
569506 95.927 -57.908 52.719
569507 95.961 -57.918 52.706
569508 95.994 -57.928 52.694
569509 96.028 -57.939 52.683
569510 96.062 -57.949 52.675
569511 96.096 -57.959 52.668
569512 96.130 -57.970 52.659
569513 96.164 -57.980 52.648
569514 96.198 -57.990 52.634
569515 96.232 -58.001 52.620
569516 96.265 -58.011 52.611
569517 96.299 -58.021 52.605
569518 96.333 -58.031 52.599
569519 96.367 -58.042 52.590
569520 96.401 -58.053 52.578
569521 96.435 -58.062 52.562
569522 96.469 -58.073 52.546
569523 96.503 -58.084 52.529
569524 96.537 -58.093 52.511
569525 96.571 -58.104 52.491
569526 96.604 -58.115 52.468
569527 96.638 -58.125 52.445
569528 96.672 -58.135 52.423
569529 96.706 -58.146 52.401
569530 96.740 -58.156 52.381
569531 96.774 -58.166 52.366
569532 96.805 -58.177 52.351
569533 96.832 -58.187 52.334
569534 96.858 -58.191 52.318
569535 96.883 -58.197 52.301
569536 96.908 -58.202 52.284
569537 96.933 -58.206 52.265
569538 96.958 -58.212 52.249
569539 96.983 -58.217 52.230
569540 97.009 -58.222 52.212
569541 97.033 -58.227 52.194
569542 97.058 -58.232 52.174
569543 97.084 -58.237 52.152
569544 97.109 -58.242 52.128
569545 97.137 -58.243 52.102
569546 97.166 -58.241 52.078
569547 97.196 -58.238 52.051
569548 97.225 -58.235 52.023
569549 97.255 -58.232 51.997
569550 97.285 -58.230 51.963
569551 97.314 -58.228 51.925
569552 97.351 -58.227 51.882
569553 97.387 -58.227 51.836
569554 97.420 -58.229 51.787
569555 97.452 -58.232 51.734
569556 97.486 -58.234 51.679
569557 97.520 -58.235 51.621
569558 97.555 -58.236 51.566
569559 97.589 -58.236 51.514
569560 97.624 -58.237 51.462
569561 97.658 -58.238 51.409
569562 97.692 -58.238 51.352
569563 97.729 -58.240 51.293
569564 97.770 -58.245 51.233
569565 97.808 -58.249 51.174
569566 97.844 -58.250 51.115
569567 97.880 -58.252 51.057
569568 97.916 -58.254 51.000
569569 97.951 -58.256 50.945
569570 97.987 -58.257 50.893
569571 98.022 -58.259 50.843
569572 98.058 -58.261 50.795
569573 98.093 -58.262 50.751
569574 98.128 -58.262 50.708
569575 98.162 -58.263 50.666
569576 98.197 -58.263 50.622
569577 98.232 -58.263 50.576
569578 98.268 -58.263 50.526
569579 98.303 -58.264 50.477
569580 98.336 -58.260 50.432
569581 98.366 -58.254 50.389
569582 98.397 -58.247 50.349
569583 98.428 -58.241 50.313
569584 98.459 -58.234 50.280
569585 98.490 -58.228 50.250
569586 98.520 -58.222 50.222
569587 98.552 -58.216 50.195
569588 98.582 -58.209 50.170
569589 98.613 -58.203 50.147
569590 98.647 -58.201 50.125
569591 98.698 -58.201 50.105
569592 98.748 -58.201 50.085
569593 98.778 -58.204 50.066
569594 98.802 -58.209 50.046
569595 98.830 -58.225 50.027
569596 98.858 -58.241 50.008
569597 98.885 -58.256 49.992
569598 98.898 -58.268 49.974
569599 98.901 -58.277 49.953
569600 98.904 -58.286 49.930
569601 98.928 -58.285 49.905
569602 99.013 -58.252 49.881
569603 99.055 -58.244 49.856
569604 99.096 -58.238 49.834
569605 99.138 -58.231 49.813
569606 99.181 -58.223 49.791
569607 99.213 -58.215 49.766
569608 99.244 -58.205 49.740
569609 99.259 -58.209 49.714
569610 99.292 -58.224 49.688
569611 99.362 -58.237 49.662
569612 99.422 -58.247 49.637
569613 99.461 -58.255 49.610
569614 99.495 -58.264 49.585
569615 99.530 -58.272 49.559
569616 99.564 -58.281 49.534
569617 99.599 -58.290 49.508
569618 99.633 -58.299 49.482
569619 99.668 -58.308 49.455
569973 87.572 -54.643 57.596
569974 87.603 -54.666 57.596
569975 87.634 -54.689 57.596
569976 87.666 -54.712 57.594
569977 87.696 -54.736 57.592
569978 87.727 -54.759 57.589
569979 87.758 -54.783 57.586
569980 87.789 -54.806 57.583
569981 87.821 -54.828 57.579
569982 87.856 -54.844 57.574
569983 87.892 -54.859 57.567
569984 87.927 -54.875 57.559
569985 87.971 -54.862 57.550
569986 88.029 -54.859 57.540
569987 88.085 -54.862 57.530
569988 88.138 -54.878 57.515
569989 88.189 -54.896 57.499
569990 88.240 -54.915 57.480
569991 88.290 -54.933 57.459
569992 88.335 -54.948 57.436
569993 88.379 -54.963 57.412
569994 88.418 -54.978 57.382
569995 88.454 -54.992 57.349
569996 88.490 -55.007 57.312
569997 88.527 -55.021 57.271
569998 88.563 -55.035 57.226
569999 88.600 -55.049 57.181
570000 88.636 -55.063 57.133
570001 88.672 -55.078 57.083
570002 88.708 -55.092 57.033
570003 88.745 -55.107 56.982
570004 88.781 -55.120 56.932
570005 88.817 -55.135 56.883
570006 88.853 -55.149 56.833
570007 88.890 -55.163 56.785
570008 88.926 -55.177 56.739
570029 89.730 -55.641 55.871
570030 89.769 -55.665 55.852
570031 89.807 -55.688 55.834
570032 89.846 -55.712 55.817
570033 89.885 -55.735 55.803
570034 89.915 -55.754 55.791
570035 89.939 -55.768 55.779
570036 89.962 -55.782 55.769
570037 89.985 -55.796 55.762
570038 90.009 -55.811 55.756
570039 90.036 -55.831 55.751
570040 90.063 -55.850 55.746
570041 90.091 -55.869 55.743
570042 90.117 -55.884 55.742
570043 90.143 -55.894 55.743
570044 90.182 -55.904 55.745
570045 90.222 -55.915 55.750
570046 90.261 -55.925 55.754
570047 90.301 -55.934 55.757
570048 90.340 -55.945 55.762
570049 90.380 -55.955 55.768
570050 90.419 -55.965 55.774
570051 90.459 -55.975 55.780
570052 90.497 -55.987 55.784
570053 90.532 -55.999 55.789
570054 90.567 -56.012 55.794
570055 90.601 -56.024 55.798
570056 90.637 -56.037 55.803
570057 90.671 -56.049 55.806
570058 90.706 -56.062 55.807
570059 90.741 -56.074 55.806
570060 90.783 -56.085 55.801
570061 90.827 -56.096 55.797
570062 90.871 -56.106 55.792
570063 90.916 -56.116 55.784
570064 90.959 -56.127 55.775
570065 91.004 -56.138 55.765
570066 91.048 -56.148 55.755
570067 91.092 -56.159 55.745
570068 91.137 -56.169 55.728
570069 91.181 -56.179 55.710
570070 91.225 -56.190 55.692
570071 91.269 -56.201 55.673
570072 91.313 -56.211 55.653
570073 91.357 -56.221 55.636
570074 91.389 -56.227 55.620
570075 91.422 -56.232 55.605
570076 91.454 -56.238 55.588
570077 91.487 -56.244 55.568
570078 91.517 -56.257 55.548
570079 91.548 -56.269 55.527
570080 91.578 -56.282 55.503
570081 91.609 -56.295 55.475
570082 91.639 -56.307 55.447
570083 91.670 -56.320 55.419
570084 91.700 -56.333 55.391
570085 91.731 -56.345 55.371
570086 91.763 -56.357 55.354
570087 91.799 -56.364 55.340
570088 91.835 -56.370 55.329
570089 91.871 -56.377 55.321
570090 91.908 -56.384 55.316
570091 91.943 -56.391 55.318
570092 91.980 -56.398 55.324
570093 92.019 -56.414 55.334
570094 92.060 -56.436 55.345
570095 92.100 -56.456 55.352
570096 92.141 -56.477 55.364
570097 92.182 -56.498 55.378
570098 92.221 -56.528 55.390
570099 92.258 -56.565 55.398
570100 92.295 -56.602 55.401
570101 92.333 -56.639 55.402
570102 92.370 -56.676 55.401
570103 92.407 -56.712 55.400
570104 92.444 -56.749 55.401
570105 92.482 -56.786 55.405
570106 92.520 -56.805 55.408
570107 92.560 -56.824 55.408
570108 92.599 -56.843 55.408
570109 92.636 -56.863 55.404
570110 92.670 -56.884 55.396
570111 92.705 -56.905 55.388
570112 92.739 -56.926 55.381
570113 92.774 -56.947 55.376
570114 92.809 -56.968 55.372
570115 92.844 -56.989 55.365
570116 92.878 -57.009 55.354
570117 92.922 -57.023 55.341
570118 92.965 -57.037 55.332
570119 93.008 -57.050 55.318
570120 93.043 -57.059 55.303
570121 93.072 -57.066 55.283
570122 93.100 -57.072 55.263
570123 93.129 -57.079 55.246
570124 93.158 -57.086 55.234
570125 93.187 -57.092 55.224
570126 93.216 -57.099 55.209
570127 93.247 -57.107 55.189
570128 93.279 -57.119 55.165
570129 93.312 -57.131 55.136
570130 93.345 -57.143 55.103
570131 93.377 -57.155 55.068
570132 93.410 -57.167 55.032
570133 93.443 -57.178 54.993
570134 93.476 -57.191 54.954
570135 93.508 -57.203 54.920
570136 93.541 -57.215 54.888
570137 93.573 -57.226 54.859
570138 93.607 -57.234 54.832
570139 93.640 -57.238 54.806
570140 93.674 -57.241 54.779
570141 93.708 -57.245 54.754
570142 93.742 -57.249 54.729
570143 93.775 -57.253 54.705
570144 93.808 -57.257 54.682
570145 93.842 -57.260 54.658
570146 93.876 -57.264 54.635
570147 93.910 -57.268 54.615
570148 93.943 -57.272 54.595
570149 93.977 -57.275 54.572
570150 94.010 -57.279 54.545
570151 94.044 -57.282 54.515
570152 94.080 -57.301 54.480
570153 94.118 -57.323 54.442
570154 94.155 -57.346 54.405
570155 94.192 -57.369 54.378
570156 94.230 -57.391 54.364
570157 94.267 -57.414 54.358
570158 94.305 -57.437 54.342
570159 94.342 -57.460 54.323
570160 94.375 -57.475 54.307
570161 94.407 -57.487 54.292
570162 94.439 -57.500 54.280
570163 94.472 -57.512 54.271
570164 94.504 -57.524 54.261
570165 94.537 -57.537 54.249
570166 94.569 -57.549 54.238
570167 94.601 -57.561 54.230
570168 94.634 -57.574 54.225
570169 94.666 -57.586 54.220
570170 94.698 -57.599 54.211
570171 94.731 -57.611 54.200
570172 94.763 -57.623 54.184
570173 94.795 -57.636 54.164
570174 94.828 -57.648 54.141
570175 94.858 -57.659 54.113
570176 94.888 -57.669 54.084
570177 94.918 -57.679 54.052
570178 94.949 -57.689 54.018
570179 94.978 -57.699 53.982
570180 95.008 -57.709 53.946
570181 95.039 -57.719 53.910
570182 95.069 -57.729 53.870
570183 95.096 -57.741 53.828
570184 95.121 -57.752 53.784
570185 95.146 -57.764 53.736
570186 95.172 -57.777 53.683
570187 95.197 -57.789 53.625
570188 95.222 -57.801 53.568
570189 95.257 -57.812 53.508
570190 95.313 -57.821 53.448
570191 95.369 -57.830 53.389
570192 95.413 -57.834 53.331
570193 95.450 -57.833 53.279
570194 95.487 -57.833 53.228
570195 95.524 -57.832 53.181
570196 95.560 -57.831 53.137
570197 95.597 -57.830 53.092
570198 95.634 -57.830 53.042
570199 95.674 -57.832 52.989
570200 95.718 -57.836 52.941
570201 95.762 -57.841 52.895
570202 95.806 -57.846 52.850
570203 95.827 -57.852 52.807
570204 95.848 -57.857 52.767
570205 95.871 -57.863 52.738
570206 95.900 -57.871 52.717
570207 95.929 -57.878 52.701
570208 95.958 -57.886 52.689
570209 95.991 -57.896 52.677
570210 96.025 -57.906 52.668
570211 96.059 -57.916 52.662
570212 96.093 -57.927 52.654
570213 96.126 -57.937 52.641
570214 96.161 -57.947 52.625
570215 96.194 -57.958 52.603
570216 96.228 -57.968 52.588
570217 96.262 -57.978 52.578
570218 96.296 -57.989 52.572
570219 96.330 -57.999 52.568
570220 96.364 -58.009 52.563
570221 96.398 -58.020 52.553
570222 96.432 -58.030 52.541
570223 96.465 -58.040 52.527
570224 96.500 -58.051 52.512
570225 96.533 -58.061 52.495
570226 96.567 -58.071 52.474
570227 96.601 -58.082 52.452
570228 96.635 -58.092 52.429
570229 96.669 -58.103 52.407
570230 96.702 -58.112 52.388
570231 96.737 -58.123 52.372
570232 96.770 -58.134 52.358
570233 96.800 -58.144 52.344
570234 96.828 -58.155 52.331
570235 96.856 -58.165 52.318
570236 96.881 -58.170 52.306
570237 96.906 -58.175 52.293
570238 96.932 -58.180 52.279
570239 96.957 -58.185 52.264
570240 96.982 -58.191 52.248
570241 97.007 -58.196 52.231
570242 97.032 -58.200 52.213
570243 97.057 -58.206 52.192
570244 97.082 -58.211 52.171
570245 97.109 -58.213 52.150
570246 97.139 -58.210 52.126
570247 97.168 -58.208 52.103
570248 97.197 -58.206 52.077
570249 97.227 -58.203 52.049
570250 97.257 -58.200 52.022
570251 97.287 -58.197 51.987
570252 97.316 -58.195 51.949
570253 97.352 -58.194 51.905
570254 97.385 -58.197 51.858
570255 97.419 -58.200 51.809
570256 97.453 -58.200 51.756
570257 97.488 -58.200 51.699
570258 97.522 -58.201 51.640
570259 97.557 -58.201 51.583
570260 97.591 -58.202 51.528
570261 97.626 -58.202 51.475
570262 97.660 -58.203 51.421
570263 97.695 -58.203 51.364
570264 97.730 -58.205 51.306
570265 97.771 -58.209 51.247
570266 97.811 -58.215 51.187
570267 97.850 -58.218 51.126
570268 97.885 -58.220 51.065
570269 97.921 -58.222 51.006
570270 97.956 -58.223 50.949
570271 97.992 -58.225 50.896
570272 98.028 -58.227 50.846
570273 98.064 -58.228 50.796
570274 98.098 -58.229 50.750
570275 98.133 -58.229 50.705
570276 98.167 -58.229 50.662
570277 98.202 -58.230 50.618
570278 98.236 -58.230 50.572
570279 98.272 -58.231 50.524
570280 98.307 -58.231 50.478
570281 98.339 -58.226 50.433
570282 98.370 -58.220 50.389
570283 98.401 -58.213 50.349
570284 98.432 -58.207 50.314
570285 98.462 -58.200 50.280
570286 98.493 -58.194 50.249
570287 98.524 -58.187 50.223
570288 98.555 -58.181 50.197
570289 98.586 -58.175 50.172
570290 98.617 -58.169 50.149
570291 98.648 -58.162 50.126
570292 98.690 -58.161 50.104
570293 98.728 -58.162 50.083
570294 98.755 -58.175 50.063
570295 98.782 -58.191 50.043
570296 98.809 -58.206 50.025
570297 98.837 -58.222 50.007
570298 98.863 -58.238 49.991
570299 98.867 -58.247 49.972
570300 98.870 -58.256 49.951
570301 98.879 -58.262 49.927
570302 98.966 -58.228 49.902
570303 99.007 -58.221 49.878
570304 99.050 -58.214 49.856
570305 99.091 -58.207 49.835
570306 99.133 -58.200 49.814
570307 99.175 -58.193 49.792
570308 99.217 -58.186 49.768
570309 99.251 -58.177 49.742
570310 99.267 -58.180 49.715
570311 99.271 -58.194 49.688
570312 99.341 -58.208 49.662
570313 99.407 -58.220 49.635
570314 99.453 -58.228 49.609
570315 99.488 -58.236 49.584
570316 99.522 -58.245 49.557
570317 99.557 -58.253 49.530
570318 99.591 -58.263 49.505
570319 99.626 -58.271 49.478
570320 99.660 -58.280 49.452
570675 87.625 -54.636 57.596
570676 87.656 -54.659 57.597
570677 87.687 -54.683 57.596
570678 87.718 -54.706 57.594
570679 87.749 -54.730 57.591
570680 87.780 -54.753 57.589
570681 87.811 -54.776 57.586
570682 87.845 -54.793 57.584
570683 87.881 -54.809 57.578
570684 87.916 -54.824 57.572
570685 87.952 -54.839 57.564
570686 87.987 -54.855 57.557
570687 88.048 -54.819 57.549
570688 88.105 -54.822 57.538
570689 88.162 -54.826 57.525
570690 88.213 -54.843 57.510
570691 88.262 -54.860 57.494
570692 88.306 -54.875 57.475
570693 88.350 -54.891 57.455
570694 88.395 -54.906 57.429
570695 88.432 -54.920 57.401
570696 88.469 -54.934 57.370
570697 88.505 -54.949 57.334
570698 88.542 -54.963 57.294
570699 88.578 -54.977 57.250
570700 88.614 -54.991 57.205
570701 88.650 -55.006 57.157
570702 88.687 -55.020 57.107
570703 88.723 -55.035 57.056
570704 88.760 -55.048 57.005
570705 88.796 -55.063 56.955
570706 88.832 -55.077 56.905
570707 88.869 -55.091 56.855
570708 88.905 -55.105 56.806
570709 88.941 -55.119 56.759
570732 89.829 -55.656 55.831
570733 89.867 -55.680 55.815
570734 89.906 -55.703 55.800
570735 89.935 -55.721 55.787
570736 89.958 -55.735 55.775
570737 89.981 -55.749 55.764
570738 90.005 -55.763 55.757
570739 90.028 -55.777 55.751
570740 90.054 -55.795 55.745
570741 90.081 -55.815 55.741
570742 90.108 -55.834 55.739
570743 90.135 -55.850 55.738
570744 90.161 -55.861 55.739
570745 90.193 -55.871 55.742
570746 90.233 -55.881 55.747
570747 90.273 -55.891 55.751
570748 90.312 -55.901 55.755
570749 90.352 -55.912 55.760
570750 90.392 -55.922 55.768
570751 90.432 -55.931 55.776
570752 90.471 -55.942 55.783
570753 90.511 -55.952 55.790
570754 90.550 -55.962 55.797
570755 90.587 -55.974 55.804
570756 90.621 -55.987 55.812
570757 90.657 -55.999 55.819
570758 90.691 -56.012 55.825
570759 90.726 -56.024 55.829
570760 90.761 -56.037 55.831
570761 90.796 -56.049 55.829
570762 90.839 -56.059 55.826
570763 90.883 -56.070 55.821
570764 90.928 -56.081 55.814
570765 90.971 -56.091 55.806
570766 91.016 -56.102 55.798
570767 91.060 -56.113 55.791
570768 91.104 -56.122 55.780
570769 91.149 -56.133 55.761
570770 91.192 -56.144 55.740
570771 91.237 -56.154 55.717
570772 91.281 -56.165 55.695
570773 91.325 -56.175 55.673
570774 91.361 -56.182 55.653
570775 91.393 -56.188 55.633
570776 91.426 -56.194 55.613
570777 91.458 -56.199 55.594
570778 91.491 -56.205 55.575
570779 91.522 -56.214 55.557
570780 91.553 -56.227 55.536
570781 91.583 -56.240 55.511
570782 91.613 -56.253 55.483
570783 91.644 -56.266 55.454
570784 91.674 -56.279 55.426
570785 91.705 -56.291 55.398
570786 91.737 -56.302 55.375
570787 91.773 -56.310 55.357
570788 91.809 -56.317 55.343
570789 91.845 -56.323 55.330
570790 91.881 -56.330 55.320
570791 91.917 -56.337 55.316
570792 91.953 -56.344 55.318
570793 91.990 -56.351 55.328
570794 92.028 -56.364 55.343
570795 92.069 -56.385 55.357
570796 92.109 -56.406 55.368
570797 92.150 -56.426 55.381
570798 92.189 -56.459 55.393
570799 92.226 -56.496 55.401
570800 92.263 -56.533 55.404
570801 92.300 -56.570 55.402
570802 92.337 -56.607 55.400
570803 92.374 -56.643 55.399
570804 92.412 -56.680 55.398
570805 92.449 -56.718 55.400
570806 92.487 -56.749 55.405
570807 92.526 -56.769 55.409
570808 92.565 -56.788 55.409
570809 92.604 -56.807 55.405
570810 92.642 -56.827 55.396
570811 92.676 -56.848 55.384
570812 92.711 -56.868 55.373
570813 92.745 -56.890 55.371
570814 92.780 -56.910 55.370
570815 92.815 -56.931 55.370
570816 92.850 -56.952 55.367
570817 92.887 -56.971 55.362
570818 92.930 -56.984 55.357
570819 92.973 -56.998 55.352
570820 93.016 -57.012 55.342
570821 93.054 -57.022 55.328
570822 93.083 -57.029 55.311
570823 93.112 -57.036 55.294
570824 93.141 -57.042 55.279
570825 93.169 -57.049 55.268
570826 93.199 -57.055 55.259
570827 93.227 -57.062 55.244
570828 93.258 -57.071 55.223
570829 93.291 -57.083 55.196
570830 93.324 -57.095 55.167
570831 93.356 -57.106 55.132
570832 93.389 -57.119 55.096
570833 93.422 -57.131 55.059
570834 93.454 -57.142 55.019
570835 93.487 -57.154 54.977
570836 93.519 -57.166 54.941
570837 93.552 -57.178 54.908
570838 93.585 -57.190 54.878
570839 93.618 -57.202 54.848
570840 93.651 -57.210 54.818
570841 93.685 -57.214 54.787
570842 93.718 -57.217 54.759
570843 93.752 -57.221 54.732
570844 93.785 -57.225 54.707
570845 93.819 -57.228 54.682
570846 93.853 -57.232 54.657
570847 93.887 -57.236 54.631
570848 93.920 -57.240 54.609
570849 93.954 -57.244 54.588
570850 93.987 -57.247 54.567
570851 94.021 -57.251 54.545
570852 94.055 -57.255 54.518
570853 94.090 -57.271 54.486
570854 94.128 -57.294 54.448
570855 94.165 -57.316 54.409
570856 94.202 -57.339 54.372
570857 94.240 -57.362 54.343
570858 94.277 -57.385 54.326
570859 94.315 -57.407 54.312
570860 94.350 -57.428 54.293
570861 94.383 -57.440 54.276
570862 94.415 -57.453 54.260
570863 94.448 -57.465 54.247
570864 94.480 -57.477 54.238
570865 94.512 -57.490 54.226
570866 94.544 -57.502 54.219
570867 94.577 -57.515 54.212
570868 94.609 -57.527 54.205
570869 94.641 -57.539 54.198
570870 94.674 -57.552 54.190
570871 94.706 -57.564 54.183
570872 94.739 -57.577 54.173
570873 94.771 -57.589 54.159
570874 94.803 -57.601 54.142
570875 94.835 -57.614 54.121
570876 94.867 -57.626 54.095
570877 94.897 -57.636 54.067
570878 94.927 -57.646 54.035
570879 94.958 -57.656 54.001
570880 94.988 -57.666 53.965
570881 95.018 -57.677 53.928
570882 95.048 -57.686 53.892
570883 95.078 -57.696 53.853
570884 95.108 -57.706 53.813
570885 95.137 -57.717 53.770
570886 95.162 -57.729 53.723
570887 95.187 -57.741 53.673
570888 95.212 -57.753 53.617
570889 95.238 -57.765 53.561
570890 95.263 -57.777 53.501
570891 95.316 -57.787 53.441
570892 95.372 -57.796 53.380
570893 95.416 -57.799 53.322
570894 95.454 -57.799 53.267
570895 95.490 -57.798 53.218
570896 95.527 -57.797 53.171
570897 95.564 -57.796 53.125
570898 95.601 -57.796 53.081
570899 95.641 -57.798 53.034
570900 95.685 -57.803 52.984
570901 95.729 -57.808 52.937
570902 95.772 -57.812 52.889
570903 95.811 -57.818 52.841
570904 95.832 -57.823 52.794
570905 95.853 -57.828 52.752
570906 95.874 -57.834 52.721
570907 95.902 -57.841 52.700
570908 95.932 -57.849 52.685
570909 95.961 -57.856 52.672
570910 95.989 -57.864 52.662
570911 96.022 -57.874 52.657
570912 96.055 -57.883 52.645
570913 96.090 -57.894 52.629
570914 96.123 -57.905 52.611
570915 96.157 -57.915 52.592
570916 96.191 -57.925 52.571
570917 96.225 -57.935 52.557
570918 96.259 -57.946 52.545
570919 96.293 -57.956 52.537
570920 96.327 -57.966 52.533
570921 96.361 -57.977 52.529
570922 96.394 -57.987 52.523
570923 96.429 -57.997 52.515
570924 96.462 -58.008 52.506
570925 96.496 -58.018 52.493
570926 96.530 -58.028 52.474
570927 96.564 -58.039 52.454
570928 96.598 -58.049 52.433
570929 96.631 -58.059 52.411
570930 96.666 -58.070 52.389
570931 96.699 -58.080 52.372
570932 96.733 -58.090 52.359
570933 96.767 -58.101 52.345
570934 96.795 -58.112 52.334
570935 96.824 -58.123 52.324
570936 96.852 -58.134 52.314
570937 96.880 -58.144 52.304
570938 96.905 -58.149 52.295
570939 96.930 -58.154 52.286
570940 96.955 -58.159 52.274
570941 96.980 -58.164 52.261
570942 97.005 -58.169 52.246
570943 97.030 -58.174 52.229
570944 97.055 -58.179 52.209
570945 97.081 -58.184 52.189
570946 97.111 -58.181 52.169
570947 97.141 -58.178 52.148
570948 97.170 -58.176 52.125
570949 97.199 -58.173 52.100
570950 97.229 -58.171 52.074
570951 97.259 -58.168 52.048
570952 97.288 -58.165 52.013
570953 97.318 -58.163 51.973
570954 97.352 -58.164 51.928
570955 97.386 -58.165 51.880
570956 97.421 -58.165 51.830
570957 97.455 -58.165 51.775
570958 97.490 -58.165 51.718
570959 97.524 -58.166 51.659
570960 97.559 -58.167 51.601
570961 97.593 -58.167 51.544
570962 97.628 -58.168 51.486
570963 97.662 -58.169 51.430
570964 97.697 -58.169 51.374
570965 97.731 -58.169 51.318
570966 97.772 -58.175 51.259
570967 97.812 -58.179 51.198
570968 97.853 -58.184 51.135
570969 97.890 -58.187 51.073
570970 97.926 -58.189 51.011
570971 97.962 -58.191 50.954
570972 97.998 -58.193 50.900
570973 98.033 -58.194 50.849
570974 98.069 -58.196 50.797
570975 98.103 -58.197 50.748
570976 98.138 -58.197 50.702
570977 98.172 -58.197 50.658
570978 98.207 -58.197 50.615
570979 98.241 -58.197 50.572
570980 98.276 -58.198 50.526
570981 98.311 -58.198 50.480
570982 98.343 -58.192 50.434
570983 98.373 -58.186 50.391
570984 98.405 -58.179 50.351
570985 98.435 -58.173 50.315
570986 98.466 -58.166 50.280
570987 98.497 -58.160 50.250
570988 98.528 -58.153 50.222
570989 98.559 -58.147 50.195
570990 98.589 -58.140 50.171
570991 98.621 -58.134 50.148
570992 98.651 -58.128 50.124
570993 98.679 -58.125 50.102
570994 98.706 -58.140 50.081
570995 98.733 -58.156 50.061
570996 98.761 -58.172 50.042
570997 98.789 -58.188 50.023
570998 98.816 -58.204 50.006
570999 98.833 -58.216 49.989
571000 98.836 -58.226 49.972
571001 98.839 -58.235 49.951
571002 98.919 -58.205 49.927
571003 98.960 -58.197 49.902
571004 99.002 -58.191 49.878
571005 99.044 -58.184 49.857
571006 99.086 -58.176 49.837
571007 99.128 -58.169 49.816
571008 99.170 -58.162 49.794
571009 99.212 -58.156 49.770
571010 99.254 -58.148 49.744
571011 99.276 -58.150 49.716
571012 99.278 -58.165 49.689
571013 99.320 -58.178 49.662
571014 99.390 -58.192 49.635
571015 99.444 -58.200 49.608
571016 99.480 -58.209 49.582
571017 99.515 -58.217 49.555
571018 99.549 -58.226 49.528
571019 99.584 -58.234 49.502
571020 99.618 -58.244 49.475
571021 99.653 -58.252 49.448
571377 87.678 -54.630 57.598
571378 87.709 -54.653 57.596
571379 87.740 -54.677 57.595
571380 87.771 -54.700 57.594
571381 87.801 -54.724 57.592
571382 87.835 -54.743 57.590
571383 87.870 -54.759 57.587
571384 87.905 -54.774 57.583
571385 87.941 -54.789 57.576
571386 87.976 -54.804 57.570
571387 88.011 -54.819 57.563
571388 88.068 -54.780 57.556
571389 88.124 -54.783 57.545
571390 88.181 -54.787 57.533
571391 88.233 -54.787 57.520
571392 88.277 -54.802 57.508
571393 88.322 -54.818 57.495
571394 88.366 -54.833 57.475
571395 88.410 -54.848 57.447
571396 88.448 -54.862 57.420
571397 88.484 -54.877 57.390
571398 88.520 -54.891 57.355
571399 88.556 -54.905 57.314
571400 88.593 -54.919 57.271
571401 88.629 -54.934 57.226
571402 88.666 -54.948 57.178
571403 88.702 -54.963 57.128
571404 88.738 -54.976 57.077
571405 88.774 -54.991 57.026
571406 88.811 -55.005 56.976
571407 88.847 -55.019 56.925
571408 88.885 -55.039 56.875
571409 88.923 -55.063 56.825
571410 88.962 -55.086 56.777
571434 89.888 -55.648 55.813
571435 89.927 -55.671 55.798
571436 89.954 -55.687 55.784
571437 89.977 -55.702 55.772
571438 90.000 -55.715 55.761
571439 90.023 -55.730 55.752
571440 90.047 -55.744 55.746
571441 90.072 -55.760 55.741
571442 90.099 -55.779 55.737
571443 90.126 -55.799 55.735
571444 90.153 -55.817 55.735
571445 90.179 -55.827 55.736
571446 90.205 -55.837 55.740
571447 90.245 -55.847 55.745
571448 90.285 -55.858 55.750
571449 90.324 -55.868 55.756
571450 90.364 -55.878 55.762
571451 90.403 -55.888 55.771
571452 90.443 -55.898 55.780
571453 90.483 -55.909 55.789
571454 90.523 -55.919 55.798
571455 90.562 -55.928 55.807
571456 90.602 -55.939 55.816
571457 90.642 -55.949 55.827
571458 90.677 -55.961 55.838
571459 90.712 -55.974 55.847
571460 90.746 -55.986 55.855
571461 90.781 -55.999 55.859
571462 90.816 -56.011 55.860
571463 90.851 -56.024 55.858
571464 90.895 -56.034 55.853
571465 90.939 -56.045 55.847
571466 90.983 -56.056 55.839
571467 91.028 -56.066 55.831
571468 91.072 -56.076 55.826
571469 91.116 -56.087 55.817
571470 91.160 -56.097 55.797
571471 91.205 -56.108 55.770
571472 91.249 -56.119 55.743
571473 91.293 -56.128 55.719
571474 91.332 -56.138 55.695
571475 91.365 -56.143 55.674
571476 91.397 -56.149 55.652
571477 91.430 -56.154 55.628
571478 91.462 -56.160 55.603
571479 91.495 -56.166 55.582
571480 91.527 -56.172 55.562
571481 91.558 -56.185 55.540
571482 91.588 -56.198 55.515
571483 91.618 -56.211 55.487
571484 91.649 -56.224 55.459
571485 91.679 -56.236 55.432
571486 91.711 -56.248 55.405
571487 91.746 -56.255 55.382
571488 91.783 -56.262 55.364
571489 91.819 -56.269 55.348
571490 91.855 -56.276 55.333
571491 91.891 -56.283 55.321
571492 91.927 -56.290 55.315
571493 91.963 -56.297 55.317
571494 92.000 -56.304 55.326
571495 92.037 -56.313 55.341
571496 92.077 -56.335 55.355
571497 92.118 -56.355 55.368
571498 92.156 -56.390 55.380
571499 92.193 -56.426 55.388
571500 92.230 -56.464 55.391
571501 92.268 -56.501 55.388
571502 92.305 -56.538 55.380
571503 92.342 -56.574 55.380
571504 92.379 -56.611 55.382
571505 92.417 -56.649 55.383
571506 92.454 -56.685 55.385
571507 92.492 -56.713 55.390
571508 92.531 -56.733 55.396
571509 92.570 -56.752 55.397
571510 92.609 -56.771 55.390
571511 92.648 -56.790 55.379
571512 92.683 -56.811 55.367
571513 92.717 -56.832 55.355
571514 92.752 -56.852 55.363
571515 92.786 -56.874 55.368
571516 92.821 -56.894 55.371
571517 92.856 -56.915 55.372
571518 92.895 -56.932 55.372
571519 92.938 -56.946 55.372
571520 92.981 -56.959 55.369
571521 93.024 -56.973 55.361
571522 93.065 -56.985 55.349
571523 93.094 -56.992 55.334
571524 93.123 -56.999 55.320
571525 93.152 -57.005 55.305
571526 93.181 -57.012 55.295
571527 93.210 -57.018 55.285
571528 93.238 -57.025 55.269
571529 93.270 -57.034 55.247
571530 93.302 -57.047 55.219
571531 93.335 -57.058 55.189
571532 93.368 -57.070 55.153
571533 93.401 -57.082 55.115
571534 93.433 -57.094 55.076
571535 93.466 -57.106 55.039
571536 93.498 -57.118 55.002
571537 93.531 -57.129 54.968
571538 93.564 -57.141 54.934
571539 93.596 -57.153 54.902
571540 93.629 -57.165 54.871
571541 93.662 -57.177 54.840
571542 93.695 -57.186 54.807
571543 93.728 -57.190 54.778
571544 93.762 -57.194 54.749
571545 93.796 -57.197 54.721
571546 93.830 -57.200 54.693
571547 93.863 -57.204 54.664
571548 93.897 -57.208 54.631
571549 93.930 -57.212 54.605
571550 93.964 -57.216 54.581
571551 93.998 -57.219 54.560
571552 94.032 -57.223 54.539
571553 94.065 -57.227 54.514
571554 94.101 -57.241 54.486
571555 94.138 -57.264 54.450
571556 94.175 -57.287 54.412
571557 94.213 -57.310 54.371
571558 94.250 -57.332 54.333
571559 94.287 -57.356 54.304
571560 94.325 -57.379 54.286
571561 94.358 -57.393 54.268
571562 94.391 -57.406 54.250
571563 94.423 -57.418 54.233
571564 94.455 -57.431 54.217
571565 94.487 -57.443 54.206
571566 94.520 -57.455 54.195
571567 94.552 -57.468 54.188
571568 94.584 -57.480 54.180
571569 94.617 -57.493 54.172
571570 94.649 -57.505 54.166
571571 94.681 -57.517 54.160
571572 94.714 -57.530 54.153
571573 94.746 -57.542 54.145
571574 94.778 -57.554 54.133
571575 94.811 -57.567 54.118
571576 94.843 -57.579 54.100
571577 94.876 -57.592 54.077
571578 94.906 -57.603 54.050
571579 94.936 -57.613 54.019
571580 94.967 -57.623 53.985
571581 94.997 -57.633 53.948
571582 95.027 -57.643 53.910
571583 95.057 -57.653 53.871
571584 95.087 -57.664 53.835
571585 95.117 -57.673 53.797
571586 95.147 -57.683 53.755
571587 95.178 -57.694 53.709
571588 95.203 -57.706 53.659
571589 95.228 -57.717 53.608
571590 95.253 -57.730 53.552
571591 95.278 -57.742 53.494
571592 95.320 -57.752 53.434
571593 95.376 -57.762 53.373
571594 95.420 -57.765 53.314
571595 95.457 -57.764 53.256
571596 95.494 -57.764 53.207
571597 95.531 -57.763 53.158
571598 95.567 -57.762 53.111
571599 95.608 -57.764 53.066
571600 95.652 -57.769 53.021
571601 95.696 -57.774 52.974
571602 95.740 -57.779 52.929
571603 95.783 -57.783 52.881
571604 95.817 -57.789 52.831
571605 95.838 -57.794 52.782
571606 95.859 -57.799 52.739
571607 95.880 -57.805 52.705
571608 95.905 -57.811 52.680
571609 95.934 -57.819 52.661
571610 95.963 -57.827 52.645
571611 95.992 -57.834 52.632
571612 96.021 -57.842 52.623
571613 96.052 -57.851 52.610
571614 96.086 -57.861 52.594
571615 96.120 -57.872 52.577
571616 96.154 -57.882 52.560
571617 96.188 -57.893 52.542
571618 96.222 -57.902 52.527
571619 96.256 -57.913 52.515
571620 96.290 -57.924 52.507
571621 96.323 -57.934 52.501
571622 96.357 -57.944 52.495
571623 96.391 -57.955 52.491
571624 96.425 -57.965 52.484
571625 96.459 -57.975 52.473
571626 96.493 -57.986 52.461
571627 96.527 -57.996 52.448
571628 96.560 -58.006 52.434
571629 96.594 -58.016 52.414
571630 96.628 -58.027 52.391
571631 96.662 -58.037 52.370
571632 96.696 -58.047 52.352
571633 96.730 -58.058 52.339
571634 96.762 -58.068 52.329
571635 96.791 -58.079 52.319
571636 96.819 -58.090 52.311
571637 96.847 -58.101 52.301
571638 96.875 -58.112 52.294
571639 96.903 -58.122 52.289
571640 96.928 -58.128 52.285
571641 96.953 -58.133 52.278
571642 96.978 -58.137 52.269
571643 97.004 -58.143 52.257
571644 97.029 -58.148 52.243
571645 97.054 -58.153 52.225
571646 97.083 -58.151 52.205
571647 97.113 -58.149 52.185
571648 97.142 -58.146 52.166
571649 97.172 -58.144 52.144
571650 97.202 -58.141 52.121
571651 97.231 -58.138 52.100
571652 97.260 -58.136 52.074
571653 97.290 -58.132 52.037
571654 97.320 -58.128 51.997
571655 97.354 -58.131 51.951
571656 97.389 -58.130 51.902
571657 97.423 -58.131 51.851
571658 97.458 -58.131 51.794
571659 97.492 -58.131 51.737
571660 97.527 -58.132 51.679
571661 97.561 -58.132 51.620
571662 97.596 -58.133 51.560
571663 97.630 -58.134 51.498
571664 97.665 -58.134 51.439
571665 97.699 -58.134 51.384
571666 97.734 -58.135 51.329
571667 97.773 -58.139 51.271
571668 97.813 -58.144 51.208
571669 97.854 -58.149 51.142
571670 97.894 -58.154 51.079
571671 97.932 -58.156 51.016
571672 97.967 -58.159 50.958
571673 98.003 -58.160 50.904
571674 98.038 -58.162 50.851
571675 98.074 -58.163 50.797
571676 98.108 -58.163 50.747
571677 98.143 -58.164 50.700
571678 98.177 -58.164 50.656
571679 98.211 -58.164 50.614
571680 98.246 -58.165 50.574
571681 98.281 -58.165 50.528
571682 98.315 -58.164 50.482
571683 98.346 -58.158 50.436
571684 98.377 -58.152 50.394
571685 98.408 -58.145 50.354
571686 98.439 -58.139 50.317
571687 98.469 -58.132 50.281
571688 98.501 -58.126 50.251
571689 98.531 -58.119 50.223
571690 98.563 -58.117 50.196
571691 98.593 -58.116 50.172
571692 98.625 -58.115 50.149
571693 98.656 -58.115 50.124
571694 98.672 -58.115 50.102
571695 98.686 -58.122 50.080
571696 98.713 -58.138 50.060
571697 98.740 -58.154 50.041
571698 98.768 -58.170 50.023
571699 98.795 -58.186 50.005
571700 98.801 -58.196 49.988
571701 98.805 -58.205 49.970
571702 98.871 -58.181 49.949
571703 98.913 -58.174 49.926
571704 98.955 -58.167 49.903
571705 98.997 -58.160 49.880
571706 99.039 -58.153 49.860
571707 99.081 -58.146 49.840
571708 99.123 -58.139 49.819
571709 99.165 -58.131 49.797
571710 99.206 -58.125 49.773
571711 99.246 -58.121 49.747
571712 99.284 -58.120 49.718
571713 99.286 -58.134 49.691
571714 99.299 -58.150 49.663
571715 99.368 -58.162 49.635
571716 99.430 -58.173 49.608
571717 99.473 -58.181 49.581
571718 99.507 -58.190 49.554
571719 99.542 -58.198 49.526
571720 99.576 -58.207 49.500
571721 99.611 -58.216 49.472
571722 99.645 -58.225 49.444
571723 99.680 -58.233 49.418
572078 87.699 -54.600 57.598
572079 87.730 -54.624 57.597
572080 87.761 -54.647 57.597
572081 87.792 -54.671 57.596
572082 87.824 -54.693 57.594
572083 87.859 -54.708 57.592
572084 87.895 -54.723 57.589
572085 87.930 -54.739 57.586
572086 87.965 -54.754 57.581
572087 88.001 -54.769 57.574
572088 88.036 -54.784 57.568
572089 88.084 -54.762 57.561
572090 88.144 -54.743 57.552
572091 88.200 -54.747 57.540
572092 88.255 -54.750 57.529
572093 88.292 -54.744 57.518
572094 88.337 -54.760 57.506
572095 88.382 -54.775 57.488
572096 88.426 -54.790 57.461
572097 88.462 -54.805 57.434
572098 88.499 -54.819 57.403
572099 88.535 -54.833 57.369
572100 88.571 -54.847 57.330
572101 88.608 -54.862 57.289
572102 88.644 -54.876 57.245
572103 88.680 -54.891 57.198
572104 88.716 -54.904 57.148
572105 88.753 -54.919 57.097
572106 88.790 -54.937 57.046
572107 88.829 -54.960 56.995
572108 88.868 -54.984 56.945
572109 88.906 -55.007 56.894
572110 88.945 -55.031 56.844
572111 88.983 -55.054 56.795
572137 89.973 -55.654 55.784
572138 89.996 -55.668 55.772
572139 90.019 -55.682 55.761
572140 90.042 -55.696 55.752
572141 90.065 -55.711 55.745
572142 90.089 -55.724 55.740
572143 90.116 -55.744 55.736
572144 90.144 -55.763 55.733
572145 90.171 -55.783 55.732
572146 90.196 -55.793 55.735
572147 90.222 -55.804 55.741
572148 90.257 -55.814 55.745
572149 90.297 -55.824 55.751
572150 90.336 -55.834 55.757
572151 90.376 -55.844 55.765
572152 90.415 -55.855 55.776
572153 90.455 -55.865 55.786
572154 90.495 -55.875 55.797
572155 90.534 -55.885 55.809
572156 90.574 -55.895 55.821
572157 90.614 -55.906 55.834
572158 90.653 -55.915 55.850
572159 90.693 -55.925 55.864
572160 90.732 -55.936 55.875
572161 90.766 -55.949 55.884
572162 90.801 -55.961 55.892
572163 90.836 -55.974 55.895
572164 90.871 -55.986 55.895
572165 90.907 -55.999 55.889
572166 90.951 -56.009 55.882
572167 90.996 -56.019 55.874
572168 91.040 -56.030 55.865
572169 91.084 -56.041 55.861
572170 91.129 -56.051 55.854
572171 91.172 -56.062 55.832
572172 91.217 -56.072 55.802
572173 91.261 -56.082 55.771
572174 91.304 -56.093 55.744
572175 91.336 -56.098 55.718
572176 91.369 -56.103 55.697
572177 91.401 -56.110 55.675
572178 91.434 -56.115 55.649
572179 91.466 -56.121 55.620
572180 91.499 -56.126 55.596
572181 91.531 -56.132 55.569
572182 91.562 -56.144 55.542
572183 91.592 -56.157 55.515
572184 91.623 -56.169 55.487
572185 91.653 -56.182 55.460
572186 91.684 -56.194 55.434
572187 91.720 -56.201 55.411
572188 91.756 -56.208 55.394
572189 91.793 -56.215 55.378
572190 91.829 -56.222 55.358
572191 91.865 -56.229 55.339
572192 91.901 -56.235 55.325
572193 91.937 -56.242 55.315
572194 91.974 -56.250 55.313
572195 92.009 -56.257 55.319
572196 92.046 -56.263 55.331
572197 92.086 -56.284 55.343
572198 92.124 -56.321 55.356
572199 92.161 -56.357 55.366
572200 92.198 -56.395 55.371
572201 92.235 -56.432 55.370
572202 92.273 -56.469 55.365
572203 92.310 -56.505 55.357
572204 92.347 -56.542 55.358
572205 92.384 -56.580 55.362
572206 92.422 -56.616 55.364
572207 92.459 -56.653 55.366
572208 92.497 -56.677 55.372
572209 92.536 -56.696 55.375
572210 92.576 -56.715 55.373
572211 92.615 -56.734 55.368
572212 92.654 -56.753 55.360
572213 92.689 -56.774 55.353
572214 92.723 -56.795 55.359
572215 92.758 -56.816 55.369
572216 92.793 -56.837 55.375
572217 92.827 -56.858 55.380
572218 92.862 -56.879 55.381
572219 92.903 -56.894 55.381
572220 92.946 -56.908 55.383
572221 92.989 -56.921 55.380
572222 93.032 -56.934 55.373
572223 93.075 -56.948 55.363
572224 93.105 -56.955 55.352
572225 93.134 -56.962 55.341
572226 93.163 -56.968 55.327
572227 93.192 -56.975 55.317
572228 93.221 -56.981 55.304
572229 93.250 -56.987 55.286
572230 93.281 -56.998 55.264
572231 93.314 -57.010 55.236
572232 93.346 -57.022 55.206
572233 93.379 -57.034 55.171
572234 93.412 -57.046 55.134
572235 93.444 -57.057 55.097
572236 93.477 -57.069 55.062
572237 93.510 -57.081 55.027
572238 93.543 -57.093 54.993
572239 93.575 -57.105 54.958
572240 93.608 -57.117 54.926
572241 93.640 -57.128 54.895
572242 93.673 -57.141 54.864
572243 93.706 -57.153 54.831
572244 93.739 -57.162 54.799
572245 93.773 -57.166 54.768
572246 93.806 -57.169 54.740
572247 93.840 -57.173 54.712
572248 93.873 -57.177 54.683
572249 93.907 -57.180 54.650
572250 93.941 -57.184 54.618
572251 93.975 -57.188 54.587
572252 94.008 -57.191 54.556
572253 94.042 -57.195 54.527
572254 94.075 -57.199 54.500
572255 94.111 -57.212 54.477
572256 94.148 -57.235 54.448
572257 94.186 -57.257 54.414
572258 94.223 -57.280 54.375
572259 94.260 -57.304 54.333
572260 94.297 -57.326 54.290
572261 94.334 -57.347 54.264
572262 94.366 -57.359 54.246
572263 94.398 -57.371 54.227
572264 94.430 -57.384 54.208
572265 94.463 -57.396 54.190
572266 94.495 -57.409 54.175
572267 94.528 -57.421 54.164
572268 94.560 -57.433 54.155
572269 94.592 -57.446 54.146
572270 94.624 -57.458 54.138
572271 94.657 -57.470 54.132
572272 94.689 -57.483 54.127
572273 94.721 -57.495 54.120
572274 94.754 -57.508 54.114
572275 94.786 -57.520 54.104
572276 94.819 -57.532 54.090
572277 94.851 -57.545 54.074
572278 94.883 -57.558 54.054
572279 94.915 -57.570 54.029
572280 94.946 -57.580 54.000
572281 94.976 -57.590 53.969
572282 95.006 -57.600 53.934
572283 95.036 -57.611 53.896
572284 95.066 -57.620 53.856
572285 95.097 -57.630 53.816
572286 95.127 -57.641 53.776
572287 95.157 -57.651 53.734
572288 95.187 -57.661 53.690
572289 95.217 -57.670 53.643
572290 95.244 -57.683 53.592
572291 95.269 -57.695 53.538
572292 95.294 -57.706 53.482
572293 95.324 -57.718 53.425
572294 95.380 -57.727 53.366
572295 95.424 -57.730 53.306
572296 95.461 -57.730 53.247
572297 95.497 -57.729 53.194
572298 95.534 -57.728 53.144
572299 95.576 -57.731 53.095
572300 95.619 -57.736 53.048
572301 95.663 -57.740 53.004
572302 95.706 -57.746 52.961
572303 95.750 -57.750 52.917
572304 95.794 -57.755 52.870
572305 95.822 -57.760 52.821
572306 95.843 -57.765 52.772
572307 95.864 -57.771 52.728
572308 95.885 -57.777 52.690
572309 95.907 -57.783 52.660
572310 95.936 -57.790 52.636
572311 95.965 -57.798 52.615
572312 95.994 -57.805 52.599
572313 96.023 -57.813 52.585
572314 96.052 -57.821 52.572
572315 96.083 -57.829 52.557
572316 96.117 -57.839 52.542
572317 96.151 -57.849 52.527
572318 96.184 -57.860 52.511
572319 96.219 -57.870 52.496
572320 96.252 -57.880 52.484
572321 96.286 -57.891 52.475
572322 96.320 -57.901 52.468
572323 96.354 -57.912 52.463
572324 96.388 -57.922 52.458
572325 96.421 -57.932 52.451
572326 96.456 -57.943 52.442
572327 96.489 -57.952 52.433
572328 96.523 -57.963 52.423
572329 96.557 -57.974 52.409
572330 96.591 -57.984 52.391
572331 96.625 -57.994 52.370
572332 96.659 -58.005 52.350
572333 96.693 -58.015 52.332
572334 96.727 -58.025 52.318
572335 96.757 -58.036 52.309
572336 96.786 -58.046 52.300
572337 96.814 -58.058 52.291
572338 96.842 -58.068 52.285
572339 96.871 -58.080 52.281
572340 96.899 -58.090 52.279
572341 96.927 -58.101 52.277
572342 96.952 -58.106 52.275
572343 96.977 -58.112 52.268
572344 97.002 -58.116 52.259
572345 97.027 -58.122 52.246
572346 97.055 -58.122 52.231
572347 97.085 -58.119 52.215
572348 97.114 -58.116 52.197
572349 97.144 -58.114 52.179
572350 97.174 -58.112 52.159
572351 97.204 -58.108 52.139
572352 97.233 -58.104 52.119
572353 97.262 -58.100 52.092
572354 97.292 -58.096 52.054
572355 97.322 -58.092 52.014
572356 97.354 -58.095 51.970
572357 97.389 -58.099 51.923
572358 97.425 -58.098 51.873
572359 97.460 -58.096 51.818
572360 97.495 -58.097 51.762
572361 97.529 -58.097 51.702
572362 97.564 -58.098 51.640
572363 97.598 -58.099 51.577
572364 97.633 -58.099 51.513
572365 97.667 -58.100 51.452
572366 97.702 -58.100 51.397
572367 97.736 -58.100 51.341
572368 97.774 -58.103 51.282
572369 97.814 -58.109 51.217
572370 97.855 -58.114 51.148
572371 97.895 -58.118 51.083
572372 97.936 -58.124 51.020
572373 97.972 -58.126 50.961
572374 98.008 -58.128 50.905
572375 98.044 -58.129 50.851
572376 98.079 -58.131 50.797
572377 98.113 -58.131 50.746
572378 98.148 -58.131 50.700
572379 98.182 -58.131 50.657
572380 98.216 -58.131 50.617
572381 98.251 -58.131 50.577
572382 98.285 -58.132 50.531
572383 98.319 -58.130 50.484
572384 98.350 -58.124 50.440
572385 98.381 -58.117 50.397
572386 98.412 -58.111 50.357
572387 98.443 -58.109 50.320
572388 98.474 -58.109 50.285
572389 98.505 -58.108 50.254
572390 98.536 -58.107 50.225
572391 98.567 -58.106 50.200
572392 98.598 -58.106 50.175
572393 98.629 -58.106 50.151
572394 98.660 -58.105 50.126
572395 98.666 -58.104 50.102
572396 98.664 -58.104 50.081
572397 98.692 -58.120 50.060
572398 98.720 -58.136 50.041
572399 98.747 -58.152 50.023
572400 98.767 -58.165 50.005
572401 98.771 -58.175 49.987
572402 98.824 -58.157 49.968
572403 98.866 -58.150 49.947
572404 98.908 -58.144 49.925
572405 98.950 -58.137 49.904
572406 98.992 -58.129 49.882
572407 99.034 -58.122 49.863
572408 99.076 -58.115 49.844
572409 99.116 -58.110 49.824
572410 99.156 -58.105 49.801
572411 99.196 -58.100 49.776
572412 99.235 -58.096 49.751
572413 99.275 -58.093 49.722
572414 99.295 -58.105 49.693
572415 99.297 -58.120 49.665
572416 99.348 -58.134 49.637
572417 99.415 -58.147 49.609
572418 99.465 -58.153 49.580
572419 99.500 -58.162 49.552
572420 99.534 -58.171 49.525
572421 99.569 -58.179 49.497
572422 99.603 -58.188 49.469
572423 99.638 -58.197 49.441
572424 99.672 -58.206 49.415
572779 87.721 -54.571 57.598
572780 87.752 -54.594 57.598
572781 87.783 -54.618 57.598
572782 87.814 -54.641 57.597
572783 87.849 -54.658 57.595
572784 87.884 -54.673 57.595
572785 87.919 -54.688 57.592
572786 87.955 -54.703 57.589
572787 87.990 -54.719 57.584
572788 88.026 -54.734 57.578
572789 88.061 -54.750 57.572
572790 88.100 -54.756 57.565
572791 88.163 -54.704 57.557
572792 88.220 -54.708 57.547
572793 88.276 -54.712 57.536
572794 88.314 -54.703 57.525
572795 88.353 -54.703 57.513
572796 88.397 -54.718 57.495
572797 88.441 -54.733 57.471
572798 88.477 -54.747 57.445
572799 88.514 -54.761 57.415
572800 88.550 -54.775 57.381
572801 88.586 -54.790 57.344
572802 88.623 -54.804 57.304
572803 88.659 -54.818 57.261
572804 88.696 -54.834 57.214
572805 88.735 -54.858 57.165
572806 88.773 -54.881 57.115
572807 88.812 -54.905 57.064
572808 88.850 -54.928 57.012
572809 88.889 -54.952 56.962
572810 88.927 -54.975 56.912
572811 88.966 -54.999 56.862
572812 89.004 -55.022 56.813
572843 90.108 -55.691 55.740
572844 90.133 -55.709 55.736
572845 90.161 -55.728 55.734
572846 90.188 -55.747 55.734
572847 90.214 -55.760 55.737
572848 90.240 -55.771 55.742
572849 90.268 -55.781 55.747
572850 90.308 -55.790 55.753
572851 90.348 -55.801 55.760
572852 90.388 -55.811 55.770
572853 90.427 -55.821 55.782
572854 90.467 -55.831 55.794
572855 90.507 -55.841 55.807
572856 90.546 -55.852 55.821
572857 90.586 -55.862 55.837
572858 90.625 -55.872 55.854
572859 90.665 -55.882 55.875
572860 90.705 -55.892 55.894
572861 90.744 -55.903 55.910
572862 90.784 -55.912 55.923
572863 90.821 -55.924 55.936
572864 90.857 -55.936 55.946
572865 90.891 -55.949 55.950
572866 90.926 -55.961 55.939
572867 90.963 -55.973 55.924
572868 91.008 -55.984 55.914
572869 91.052 -55.994 55.902
572870 91.096 -56.005 55.898
572871 91.140 -56.016 55.887
572872 91.184 -56.025 55.864
572873 91.229 -56.036 55.833
572874 91.273 -56.047 55.800
572875 91.308 -56.053 55.770
572876 91.341 -56.059 55.741
572877 91.373 -56.065 55.723
572878 91.405 -56.070 55.702
572879 91.438 -56.076 55.675
572880 91.470 -56.081 55.642
572881 91.503 -56.088 55.614
572882 91.535 -56.093 55.580
572883 91.567 -56.102 55.549
572884 91.597 -56.115 55.519
572885 91.628 -56.127 55.490
572886 91.658 -56.140 55.462
572887 91.694 -56.147 55.438
572888 91.730 -56.154 55.420
572889 91.767 -56.161 55.408
572890 91.803 -56.168 55.395
572891 91.839 -56.175 55.377
572892 91.875 -56.182 55.355
572893 91.911 -56.188 55.336
572894 91.947 -56.195 55.319
572895 91.983 -56.202 55.308
572896 92.019 -56.209 55.307
572897 92.056 -56.220 55.313
572898 92.091 -56.252 55.323
572899 92.129 -56.288 55.330
572900 92.166 -56.326 55.337
572901 92.203 -56.363 55.340
572902 92.240 -56.400 55.339
572903 92.278 -56.436 55.335
572904 92.315 -56.473 55.333
572905 92.352 -56.511 55.334
572906 92.389 -56.547 55.339
572907 92.427 -56.584 55.342
572908 92.464 -56.621 55.346
572909 92.503 -56.641 55.353
572910 92.542 -56.660 55.354
572911 92.581 -56.679 55.350
572912 92.620 -56.698 55.345
572913 92.659 -56.718 55.342
572914 92.695 -56.738 55.342
572915 92.729 -56.758 55.360
572916 92.764 -56.780 55.372
572917 92.799 -56.800 55.378
572918 92.834 -56.821 55.378
572919 92.868 -56.842 55.376
572920 92.912 -56.855 55.377
572921 92.954 -56.869 55.380
572922 92.998 -56.883 55.380
572923 93.041 -56.896 55.376
572924 93.084 -56.910 55.370
572925 93.116 -56.918 55.362
572926 93.145 -56.924 55.353
572927 93.174 -56.931 55.340
572928 93.203 -56.937 55.327
572929 93.232 -56.944 55.312
572930 93.261 -56.951 55.293
572931 93.292 -56.962 55.272
572932 93.325 -56.973 55.246
572933 93.358 -56.985 55.219
572934 93.391 -56.997 55.187
572935 93.423 -57.009 55.154
572936 93.456 -57.021 55.121
572937 93.489 -57.033 55.087
572938 93.521 -57.044 55.051
572939 93.554 -57.056 55.016
572940 93.586 -57.069 54.981
572941 93.619 -57.080 54.948
572942 93.652 -57.092 54.917
572943 93.685 -57.104 54.886
572944 93.717 -57.116 54.853
572945 93.750 -57.128 54.820
572946 93.783 -57.138 54.790
572947 93.817 -57.141 54.761
572948 93.850 -57.145 54.733
572949 93.884 -57.149 54.704
572950 93.918 -57.153 54.670
572951 93.951 -57.156 54.636
572952 93.985 -57.160 54.601
572953 94.018 -57.163 54.567
572954 94.052 -57.167 54.534
572955 94.086 -57.171 54.502
572956 94.121 -57.182 54.470
572957 94.158 -57.205 54.438
572958 94.196 -57.228 54.406
572959 94.233 -57.251 54.368
572960 94.270 -57.274 54.326
572961 94.308 -57.297 54.286
572962 94.341 -57.312 54.252
572963 94.373 -57.325 54.227
572964 94.406 -57.337 54.205
572965 94.438 -57.349 54.185
572966 94.471 -57.362 54.166
572967 94.503 -57.374 54.146
572968 94.535 -57.386 54.131
572969 94.567 -57.399 54.120
572970 94.600 -57.411 54.111
572971 94.632 -57.424 54.103
572972 94.665 -57.436 54.097
572973 94.697 -57.448 54.092
572974 94.729 -57.461 54.086
572975 94.761 -57.473 54.080
572976 94.794 -57.485 54.072
572977 94.826 -57.498 54.060
572978 94.858 -57.510 54.046
572979 94.891 -57.523 54.028
572980 94.923 -57.536 54.006
572981 94.955 -57.548 53.980
572982 94.985 -57.558 53.951
572983 95.015 -57.567 53.919
572984 95.045 -57.577 53.883
572985 95.075 -57.588 53.841
572986 95.106 -57.598 53.800
572987 95.136 -57.608 53.757
572988 95.166 -57.617 53.713
572989 95.196 -57.628 53.670
572990 95.226 -57.638 53.622
572991 95.256 -57.648 53.572
572992 95.284 -57.659 53.521
572993 95.310 -57.671 53.469
572994 95.335 -57.683 53.414
572995 95.383 -57.693 53.358
572996 95.427 -57.696 53.299
572997 95.464 -57.695 53.239
572998 95.501 -57.695 53.182
572999 95.542 -57.697 53.129
573000 95.586 -57.702 53.079
573001 95.630 -57.707 53.032
573002 95.674 -57.711 52.988
573003 95.718 -57.716 52.947
573004 95.761 -57.721 52.904
573005 95.805 -57.726 52.859
573006 95.827 -57.731 52.812
573007 95.848 -57.736 52.764
573008 95.870 -57.742 52.717
573009 95.890 -57.748 52.678
573010 95.911 -57.753 52.645
573011 95.939 -57.761 52.616
573012 95.967 -57.768 52.592
573013 95.996 -57.776 52.571
573014 96.026 -57.783 52.552
573015 96.054 -57.791 52.536
573016 96.084 -57.799 52.521
573017 96.113 -57.806 52.507
573018 96.148 -57.817 52.492
573019 96.181 -57.827 52.477
573020 96.215 -57.837 52.464
573021 96.249 -57.848 52.452
573022 96.283 -57.858 52.443
573023 96.317 -57.868 52.437
573024 96.350 -57.879 52.431
573025 96.385 -57.890 52.426
573026 96.418 -57.899 52.418
573027 96.452 -57.910 52.413
573028 96.486 -57.920 52.407
573029 96.520 -57.930 52.397
573030 96.554 -57.941 52.383
573031 96.588 -57.951 52.365
573032 96.622 -57.962 52.347
573033 96.656 -57.972 52.328
573034 96.689 -57.982 52.311
573035 96.723 -57.993 52.297
573036 96.753 -58.003 52.287
573037 96.781 -58.015 52.279
573038 96.809 -58.025 52.271
573039 96.837 -58.036 52.267
573040 96.866 -58.047 52.266
573041 96.894 -58.058 52.266
573042 96.922 -58.069 52.266
573043 96.950 -58.080 52.267
573044 96.975 -58.085 52.264
573045 97.001 -58.090 52.257
573046 97.028 -58.092 52.247
573047 97.058 -58.090 52.235
573048 97.087 -58.087 52.223
573049 97.116 -58.084 52.208
573050 97.146 -58.080 52.192
573051 97.176 -58.076 52.174
573052 97.206 -58.071 52.155
573053 97.235 -58.068 52.134
573054 97.264 -58.064 52.107
573055 97.294 -58.060 52.070
573056 97.324 -58.056 52.034
573057 97.354 -58.053 51.993
573058 97.388 -58.068 51.946
573059 97.424 -58.068 51.895
573060 97.460 -58.066 51.841
573061 97.495 -58.065 51.785
573062 97.531 -58.064 51.723
573063 97.566 -58.063 51.658
573064 97.600 -58.064 51.591
573065 97.635 -58.065 51.526
573066 97.669 -58.065 51.465
573067 97.704 -58.065 51.409
573068 97.738 -58.066 51.352
573069 97.775 -58.068 51.290
573070 97.815 -58.073 51.223
573071 97.856 -58.078 51.155
573072 97.896 -58.084 51.089
573073 97.937 -58.088 51.026
573074 97.978 -58.093 50.967
573075 98.013 -58.095 50.911
573076 98.049 -58.097 50.856
573077 98.084 -58.097 50.801
573078 98.118 -58.098 50.750
573079 98.153 -58.098 50.703
573080 98.187 -58.098 50.661
573081 98.221 -58.099 50.622
573082 98.255 -58.099 50.583
573083 98.290 -58.099 50.541
573084 98.323 -58.100 50.495
573085 98.354 -58.101 50.448
573086 98.385 -58.100 50.404
573087 98.416 -58.100 50.363
573088 98.447 -58.099 50.327
573089 98.478 -58.098 50.291
573090 98.509 -58.097 50.259
573091 98.541 -58.096 50.230
573092 98.571 -58.096 50.204
573093 98.602 -58.096 50.179
573094 98.634 -58.095 50.155
573095 98.660 -58.094 50.131
573096 98.659 -58.094 50.107
573097 98.658 -58.093 50.085
573098 98.671 -58.102 50.064
573099 98.699 -58.118 50.045
573100 98.726 -58.133 50.026
573101 98.736 -58.144 50.006
573102 98.777 -58.134 49.989
573103 98.819 -58.127 49.970
573104 98.861 -58.120 49.951
573105 98.903 -58.113 49.931
573106 98.945 -58.106 49.909
573107 98.986 -58.100 49.888
573108 99.026 -58.094 49.869
573109 99.066 -58.090 49.850
573110 99.106 -58.085 49.829
573111 99.145 -58.081 49.806
573112 99.185 -58.076 49.780
573113 99.225 -58.071 49.754
573114 99.265 -58.067 49.726
573115 99.302 -58.075 49.696
573116 99.305 -58.090 49.668
573117 99.327 -58.104 49.639
573118 99.397 -58.118 49.611
573119 99.452 -58.126 49.582
573120 99.492 -58.134 49.553
573121 99.526 -58.143 49.525
573122 99.561 -58.151 49.496
573123 99.595 -58.160 49.468
573124 99.630 -58.169 49.439
573125 99.664 -58.178 49.411
573126 99.699 -58.187 49.383
573481 87.774 -54.565 57.599
573482 87.805 -54.588 57.599
573483 87.838 -54.607 57.598
573484 87.873 -54.623 57.596
573485 87.909 -54.638 57.597
573486 87.944 -54.653 57.594
573487 87.979 -54.668 57.592
573488 88.014 -54.684 57.587
573489 88.050 -54.699 57.582
573490 88.085 -54.715 57.576
573491 88.121 -54.730 57.570
573492 88.182 -54.668 57.562
573493 88.239 -54.668 57.552
573494 88.296 -54.672 57.541
573495 88.336 -54.665 57.530
573496 88.372 -54.656 57.517
573497 88.413 -54.660 57.499
573498 88.456 -54.675 57.478
573499 88.492 -54.689 57.453
573500 88.529 -54.703 57.425
573501 88.565 -54.718 57.393
573502 88.601 -54.733 57.357
573503 88.640 -54.756 57.318
573504 88.679 -54.779 57.276
573505 88.717 -54.803 57.230
573506 88.756 -54.826 57.181
573507 88.794 -54.850 57.132
573508 88.833 -54.873 57.081
573509 88.872 -54.897 57.030
573510 88.910 -54.920 56.980
573511 88.949 -54.944 56.930
573512 88.987 -54.967 56.879
573513 89.026 -54.990 56.830
573547 90.206 -55.712 55.735
573548 90.232 -55.727 55.739
573549 90.257 -55.737 55.744
573550 90.283 -55.747 55.750
573551 90.320 -55.757 55.757
573552 90.360 -55.768 55.766
573553 90.399 -55.777 55.776
573554 90.439 -55.787 55.790
573555 90.479 -55.798 55.805
573556 90.518 -55.808 55.820
573557 90.558 -55.818 55.837
573558 90.598 -55.828 55.855
573559 90.637 -55.838 55.877
573560 90.677 -55.849 55.903
573561 90.717 -55.859 55.928
573562 90.756 -55.868 55.950
573563 90.796 -55.879 55.970
573564 90.835 -55.889 55.989
573565 90.875 -55.900 56.006
573566 90.911 -55.911 56.009
573567 90.946 -55.924 55.991
573568 90.981 -55.936 55.969
573569 91.020 -55.948 55.959
573570 91.064 -55.959 55.941
573571 91.108 -55.969 55.936
573572 91.152 -55.979 55.924
573573 91.196 -55.990 55.900
573574 91.241 -56.000 55.866
573575 91.279 -56.009 55.834
573576 91.312 -56.014 55.797
573577 91.344 -56.020 55.767
573578 91.377 -56.025 55.746
573579 91.410 -56.031 55.726
573580 91.442 -56.037 55.699
573581 91.474 -56.042 55.668
573582 91.507 -56.048 55.637
573583 91.539 -56.053 55.601
573584 91.572 -56.060 55.565
573585 91.602 -56.072 55.531
573586 91.633 -56.085 55.498
573587 91.668 -56.093 55.467
573588 91.704 -56.100 55.443
573589 91.740 -56.107 55.429
573590 91.777 -56.113 55.420
573591 91.812 -56.121 55.411
573592 91.849 -56.128 55.395
573593 91.885 -56.137 55.373
573594 91.922 -56.147 55.350
573595 91.958 -56.158 55.327
573596 91.995 -56.169 55.308
573597 92.032 -56.179 55.297
573598 92.065 -56.200 55.294
573599 92.097 -56.223 55.297
573600 92.134 -56.257 55.303
573601 92.171 -56.294 55.308
573602 92.208 -56.331 55.313
573603 92.245 -56.367 55.317
573604 92.282 -56.404 55.318
573605 92.319 -56.442 55.319
573606 92.357 -56.478 55.318
573607 92.394 -56.515 55.318
573608 92.431 -56.552 55.319
573609 92.469 -56.586 55.326
573610 92.508 -56.605 55.338
573611 92.547 -56.624 55.339
573612 92.586 -56.642 55.332
573613 92.625 -56.661 55.325
573614 92.664 -56.681 55.324
573615 92.701 -56.701 55.333
573616 92.735 -56.722 55.357
573617 92.770 -56.743 55.372
573618 92.805 -56.764 55.376
573619 92.840 -56.784 55.373
573620 92.876 -56.803 55.371
573621 92.920 -56.817 55.377
573622 92.963 -56.830 55.382
573623 93.006 -56.844 55.382
573624 93.049 -56.858 55.379
573625 93.092 -56.871 55.374
573626 93.127 -56.881 55.367
573627 93.156 -56.887 55.359
573628 93.185 -56.894 55.347
573629 93.214 -56.901 55.332
573630 93.243 -56.907 55.315
573631 93.272 -56.914 55.296
573632 93.304 -56.924 55.275
573633 93.337 -56.937 55.251
573634 93.369 -56.949 55.227
573635 93.402 -56.960 55.200
573636 93.435 -56.972 55.171
573637 93.468 -56.984 55.140
573638 93.500 -56.997 55.106
573639 93.533 -57.008 55.070
573640 93.565 -57.020 55.035
573641 93.598 -57.032 55.001
573642 93.631 -57.044 54.968
573643 93.663 -57.056 54.936
573644 93.696 -57.068 54.905
573645 93.729 -57.079 54.872
573646 93.761 -57.091 54.841
573647 93.794 -57.103 54.812
573648 93.827 -57.113 54.784
573649 93.860 -57.117 54.756
573650 93.894 -57.121 54.727
573651 93.928 -57.125 54.690
573652 93.962 -57.128 54.654
573653 93.995 -57.132 54.618
573654 94.029 -57.136 54.582
573655 94.063 -57.140 54.547
573656 94.096 -57.143 54.511
573657 94.131 -57.153 54.473
573658 94.169 -57.175 54.439
573659 94.206 -57.199 54.405
573660 94.243 -57.222 54.369
573661 94.280 -57.244 54.330
573662 94.317 -57.265 54.293
573663 94.349 -57.278 54.254
573664 94.381 -57.290 54.219
573665 94.414 -57.302 54.190
573666 94.446 -57.315 54.167
573667 94.478 -57.327 54.146
573668 94.510 -57.340 54.124
573669 94.543 -57.352 54.106
573670 94.575 -57.364 54.090
573671 94.608 -57.377 54.079
573672 94.640 -57.389 54.070
573673 94.672 -57.401 54.063
573674 94.704 -57.414 54.055
573675 94.737 -57.426 54.049
573676 94.769 -57.439 54.044
573677 94.802 -57.451 54.036
573678 94.834 -57.464 54.026
573679 94.866 -57.476 54.015
573680 94.899 -57.489 53.998
573681 94.931 -57.501 53.981
573682 94.963 -57.514 53.959
573683 94.994 -57.524 53.933
573684 95.024 -57.535 53.903
573685 95.054 -57.545 53.868
573686 95.085 -57.554 53.827
573687 95.115 -57.564 53.785
573688 95.145 -57.575 53.740
573689 95.175 -57.585 53.694
573690 95.205 -57.595 53.647
573691 95.235 -57.605 53.600
573692 95.265 -57.615 53.551
573693 95.296 -57.625 53.504
573694 95.325 -57.636 53.455
573695 95.350 -57.648 53.404
573696 95.387 -57.658 53.349
573697 95.431 -57.661 53.292
573698 95.468 -57.661 53.232
573699 95.510 -57.664 53.173
573700 95.553 -57.668 53.117
573701 95.597 -57.673 53.065
573702 95.640 -57.678 53.017
573703 95.684 -57.683 52.973
573704 95.728 -57.687 52.931
573705 95.772 -57.692 52.888
573706 95.812 -57.697 52.844
573707 95.833 -57.702 52.799
573708 95.854 -57.708 52.754
573709 95.875 -57.714 52.708
573710 95.896 -57.719 52.667
573711 95.916 -57.725 52.630
573712 95.941 -57.731 52.599
573713 95.970 -57.739 52.573
573714 95.999 -57.746 52.549
573715 96.028 -57.754 52.528
573716 96.057 -57.761 52.509
573717 96.086 -57.769 52.493
573718 96.115 -57.777 52.478
573719 96.144 -57.784 52.462
573720 96.178 -57.795 52.448
573721 96.212 -57.805 52.436
573722 96.246 -57.815 52.424
573723 96.279 -57.826 52.415
573724 96.313 -57.836 52.409
573725 96.347 -57.846 52.404
573726 96.381 -57.856 52.401
573727 96.415 -57.867 52.394
573728 96.449 -57.877 52.388
573729 96.483 -57.887 52.380
573730 96.517 -57.898 52.369
573731 96.550 -57.909 52.354
573732 96.585 -57.918 52.340
573733 96.618 -57.929 52.325
573734 96.652 -57.940 52.306
573735 96.686 -57.949 52.288
573736 96.720 -57.960 52.274
573737 96.748 -57.971 52.265
573738 96.777 -57.982 52.256
573739 96.805 -57.993 52.251
573740 96.833 -58.004 52.248
573741 96.861 -58.015 52.248
573742 96.889 -58.026 52.250
573743 96.918 -58.037 52.252
573744 96.946 -58.047 52.255
573745 96.974 -58.059 52.255
573746 97.000 -58.062 52.252
573747 97.030 -58.059 52.247
573748 97.059 -58.056 52.238
573749 97.089 -58.052 52.229
573750 97.118 -58.047 52.218
573751 97.148 -58.043 52.204
573752 97.178 -58.040 52.187
573753 97.208 -58.036 52.168
573754 97.237 -58.032 52.145
573755 97.267 -58.028 52.120
573756 97.296 -58.024 52.086
573757 97.326 -58.020 52.054
573758 97.356 -58.016 52.016
573759 97.389 -58.025 51.969
573760 97.424 -58.037 51.917
573761 97.459 -58.037 51.863
573762 97.495 -58.035 51.806
573763 97.530 -58.034 51.742
573764 97.566 -58.032 51.674
573765 97.602 -58.031 51.605
573766 97.637 -58.030 51.538
573767 97.671 -58.031 51.478
573768 97.706 -58.031 51.421
573769 97.740 -58.031 51.362
573770 97.776 -58.033 51.300
573771 97.816 -58.037 51.233
573772 97.857 -58.043 51.166
573773 97.897 -58.048 51.100
573774 97.938 -58.053 51.037
573775 97.979 -58.058 50.977
573776 98.019 -58.063 50.920
573777 98.054 -58.065 50.864
573778 98.089 -58.065 50.808
573779 98.123 -58.065 50.757
573780 98.157 -58.065 50.710
573781 98.192 -58.066 50.667
573782 98.225 -58.068 50.627
573783 98.260 -58.068 50.588
573784 98.294 -58.070 50.547
573785 98.327 -58.071 50.502
573786 98.359 -58.073 50.455
573787 98.390 -58.075 50.411
573788 98.421 -58.078 50.371
573789 98.452 -58.080 50.336
573790 98.483 -58.082 50.300
573791 98.514 -58.084 50.267
573792 98.545 -58.086 50.238
573793 98.576 -58.086 50.211
573794 98.607 -58.085 50.186
573795 98.638 -58.084 50.162
573796 98.654 -58.084 50.137
573797 98.652 -58.084 50.113
573798 98.651 -58.083 50.091
573799 98.651 -58.084 50.071
573800 98.678 -58.099 50.052
573801 98.702 -58.114 50.032
573802 98.729 -58.111 50.012
573803 98.772 -58.103 49.995
573804 98.814 -58.096 49.977
573805 98.856 -58.089 49.958
573806 98.896 -58.084 49.937
573807 98.936 -58.079 49.915
573808 98.976 -58.075 49.895
573809 99.015 -58.070 49.876
573810 99.055 -58.065 49.856
573811 99.095 -58.061 49.835
573812 99.135 -58.056 49.811
573813 99.174 -58.051 49.785
573814 99.214 -58.046 49.759
573815 99.254 -58.042 49.731
573816 99.292 -58.048 49.700
573817 99.314 -58.060 49.670
573818 99.316 -58.075 49.641
573819 99.375 -58.089 49.612
573820 99.437 -58.099 49.583
573821 99.485 -58.106 49.554
573822 99.519 -58.115 49.525
573823 99.554 -58.124 49.496
573824 99.588 -58.132 49.466
573825 99.623 -58.141 49.437
573826 99.657 -58.150 49.408
573827 99.692 -58.159 49.380
573828 99.726 -58.168 49.351
574182 87.795 -54.535 57.598
574183 87.827 -54.557 57.598
574184 87.862 -54.572 57.599
574185 87.898 -54.588 57.598
574186 87.933 -54.603 57.598
574187 87.969 -54.618 57.596
574188 88.004 -54.634 57.593
574189 88.039 -54.649 57.590
574190 88.075 -54.665 57.585
574191 88.110 -54.680 57.580
574192 88.145 -54.695 57.573
574193 88.197 -54.661 57.566
574194 88.259 -54.629 57.557
574195 88.316 -54.633 57.546
574196 88.359 -54.627 57.534
574197 88.395 -54.618 57.520
574198 88.431 -54.609 57.503
574199 88.471 -54.617 57.485
574200 88.507 -54.631 57.464
574201 88.545 -54.654 57.440
574202 88.584 -54.678 57.407
574203 88.623 -54.701 57.369
574204 88.661 -54.724 57.331
574205 88.700 -54.747 57.290
574206 88.739 -54.771 57.244
574207 88.777 -54.794 57.196
574208 88.816 -54.818 57.148
574209 88.854 -54.841 57.097
574210 88.893 -54.865 57.046
574211 88.931 -54.888 56.996
574212 88.970 -54.912 56.945
574213 89.009 -54.935 56.894
574214 89.047 -54.958 56.845
574247 90.196 -55.658 55.738
574248 90.223 -55.677 55.737
574249 90.250 -55.693 55.742
574250 90.275 -55.703 55.747
574251 90.301 -55.714 55.753
574252 90.332 -55.724 55.762
574253 90.372 -55.734 55.773
574254 90.411 -55.744 55.784
574255 90.451 -55.754 55.800
574256 90.490 -55.765 55.818
574257 90.530 -55.774 55.838
574258 90.570 -55.784 55.860
574259 90.609 -55.795 55.879
574260 90.649 -55.805 55.900
574261 90.689 -55.815 55.928
574262 90.728 -55.825 55.959
574263 90.768 -55.835 55.994
574264 90.808 -55.846 56.029
574265 90.848 -55.856 56.064
574266 90.887 -55.865 56.077
574267 90.927 -55.876 56.068
574268 90.966 -55.886 56.041
574269 91.001 -55.899 56.016
574270 91.036 -55.911 56.008
574271 91.076 -55.922 55.983
574272 91.120 -55.933 55.975
574273 91.164 -55.944 55.965
574274 91.209 -55.954 55.939
574275 91.251 -55.964 55.900
574276 91.283 -55.969 55.867
574277 91.316 -55.975 55.825
574278 91.348 -55.981 55.792
574279 91.381 -55.986 55.768
574280 91.413 -55.992 55.745
574281 91.446 -55.997 55.717
574282 91.478 -56.003 55.685
574283 91.511 -56.009 55.654
574284 91.543 -56.015 55.619
574285 91.576 -56.020 55.581
574286 91.607 -56.031 55.544
574287 91.642 -56.039 55.507
574288 91.678 -56.046 55.474
574289 91.714 -56.053 55.449
574290 91.751 -56.064 55.433
574291 91.788 -56.075 55.424
574292 91.824 -56.085 55.417
574293 91.861 -56.096 55.405
574294 91.897 -56.107 55.387
574295 91.934 -56.118 55.365
574296 91.971 -56.128 55.339
574297 92.008 -56.139 55.312
574298 92.042 -56.157 55.290
574299 92.074 -56.180 55.276
574300 92.106 -56.204 55.273
574301 92.139 -56.227 55.276
574302 92.175 -56.261 55.282
574303 92.213 -56.298 55.289
574304 92.250 -56.335 55.296
574305 92.287 -56.373 55.301
574306 92.324 -56.409 55.303
574307 92.362 -56.446 55.302
574308 92.399 -56.483 55.300
574309 92.436 -56.520 55.302
574310 92.475 -56.549 55.309
574311 92.513 -56.568 55.318
574312 92.553 -56.587 55.316
574313 92.591 -56.606 55.309
574314 92.631 -56.626 55.308
574315 92.670 -56.645 55.319
574316 92.707 -56.664 55.336
574317 92.741 -56.686 55.353
574318 92.776 -56.706 55.361
574319 92.811 -56.727 55.363
574320 92.846 -56.748 55.369
574321 92.885 -56.765 55.382
574322 92.928 -56.778 55.391
574323 92.971 -56.792 55.395
574324 93.014 -56.805 55.391
574325 93.057 -56.819 55.383
574326 93.100 -56.833 55.375
574327 93.138 -56.844 55.368
574328 93.167 -56.851 55.361
574329 93.196 -56.857 55.350
574330 93.225 -56.864 55.334
574331 93.254 -56.870 55.315
574332 93.283 -56.877 55.296
574333 93.315 -56.888 55.276
574334 93.348 -56.900 55.254
574335 93.381 -56.912 55.232
574336 93.413 -56.924 55.206
574337 93.446 -56.936 55.178
574338 93.479 -56.948 55.150
574339 93.511 -56.959 55.117
574340 93.544 -56.971 55.082
574341 93.577 -56.984 55.045
574342 93.610 -56.995 55.010
574343 93.642 -57.007 54.977
574344 93.675 -57.019 54.948
574345 93.707 -57.031 54.918
574346 93.740 -57.043 54.886
574347 93.773 -57.055 54.855
574348 93.805 -57.066 54.827
574349 93.838 -57.078 54.799
574350 93.871 -57.089 54.770
574351 93.905 -57.093 54.742
574352 93.938 -57.097 54.706
574353 93.972 -57.100 54.671
574354 94.005 -57.104 54.635
574355 94.039 -57.108 54.599
574356 94.073 -57.112 54.563
574357 94.107 -57.116 54.528
574358 94.141 -57.123 54.490
574359 94.179 -57.147 54.452
574360 94.216 -57.169 54.414
574361 94.253 -57.192 54.377
574362 94.290 -57.215 54.342
574363 94.324 -57.231 54.306
574364 94.356 -57.243 54.264
574365 94.389 -57.256 54.218
574366 94.421 -57.268 54.179
574367 94.454 -57.280 54.151
574368 94.486 -57.293 54.128
574369 94.518 -57.305 54.105
574370 94.551 -57.317 54.083
574371 94.583 -57.330 54.065
574372 94.615 -57.342 54.050
574373 94.647 -57.355 54.038
574374 94.680 -57.367 54.028
574375 94.712 -57.379 54.019
574376 94.745 -57.392 54.012
574377 94.777 -57.404 54.007
574378 94.809 -57.417 54.001
574379 94.841 -57.429 53.993
574380 94.874 -57.442 53.983
574381 94.906 -57.454 53.970
574382 94.938 -57.467 53.954
574383 94.971 -57.479 53.935
574384 95.003 -57.492 53.911
574385 95.034 -57.501 53.883
574386 95.064 -57.512 53.852
574387 95.094 -57.522 53.813
574388 95.124 -57.532 53.771
574389 95.154 -57.542 53.727
574390 95.184 -57.552 53.680
574391 95.214 -57.562 53.632
574392 95.245 -57.572 53.581
574393 95.275 -57.582 53.531
574394 95.305 -57.592 53.487
574395 95.335 -57.602 53.441
574396 95.365 -57.612 53.392
574397 95.391 -57.624 53.340
574398 95.435 -57.627 53.284
574399 95.476 -57.630 53.225
574400 95.520 -57.635 53.166
574401 95.564 -57.639 53.108
574402 95.608 -57.645 53.054
574403 95.652 -57.649 53.003
574404 95.695 -57.654 52.959
574405 95.739 -57.658 52.913
574406 95.783 -57.664 52.869
574407 95.817 -57.668 52.826
574408 95.838 -57.674 52.784
574409 95.859 -57.680 52.742
574410 95.880 -57.685 52.700
574411 95.901 -57.690 52.655
574412 95.922 -57.696 52.617
574413 95.943 -57.702 52.584
574414 95.972 -57.709 52.555
574415 96.001 -57.717 52.529
574416 96.030 -57.724 52.506
574417 96.059 -57.732 52.485
574418 96.088 -57.739 52.468
574419 96.117 -57.747 52.452
574420 96.147 -57.755 52.435
574421 96.175 -57.762 52.422
574422 96.208 -57.772 52.410
574423 96.242 -57.783 52.399
574424 96.276 -57.793 52.391
574425 96.310 -57.803 52.386
574426 96.344 -57.814 52.382
574427 96.378 -57.824 52.380
574428 96.412 -57.834 52.374
574429 96.446 -57.845 52.366
574430 96.479 -57.855 52.355
574431 96.514 -57.865 52.343
574432 96.547 -57.876 52.328
574433 96.581 -57.886 52.313
574434 96.615 -57.896 52.297
574435 96.649 -57.907 52.278
574436 96.683 -57.917 52.262
574437 96.715 -57.927 52.248
574438 96.743 -57.938 52.240
574439 96.771 -57.949 52.233
574440 96.800 -57.960 52.230
574441 96.828 -57.971 52.229
574442 96.856 -57.982 52.228
574443 96.884 -57.993 52.228
574444 96.912 -58.004 52.228
574445 96.941 -58.015 52.232
574446 96.970 -58.025 52.236
574447 97.004 -58.026 52.239
574448 97.032 -58.023 52.240
574449 97.062 -58.019 52.237
574450 97.091 -58.015 52.231
574451 97.120 -58.012 52.223
574452 97.150 -58.008 52.210
574453 97.180 -58.003 52.193
574454 97.210 -57.999 52.174
574455 97.239 -57.996 52.154
574456 97.269 -57.992 52.132
574457 97.298 -57.988 52.103
574458 97.328 -57.984 52.073
574459 97.358 -57.980 52.036
574460 97.389 -57.983 51.991
574461 97.424 -57.998 51.939
574462 97.459 -58.006 51.884
574463 97.495 -58.005 51.824
574464 97.530 -58.003 51.758
574465 97.566 -58.003 51.689
574466 97.601 -58.001 51.618
574467 97.637 -57.999 51.551
574468 97.672 -57.998 51.493
574469 97.708 -57.997 51.433
574470 97.743 -57.997 51.374
574471 97.777 -57.997 51.313
574472 97.817 -58.003 51.248
574473 97.858 -58.007 51.184
574474 97.898 -58.012 51.120
574475 97.939 -58.018 51.055
574476 97.980 -58.022 50.992
574477 98.018 -58.028 50.933
574478 98.057 -58.032 50.877
574479 98.093 -58.034 50.820
574480 98.127 -58.035 50.768
574481 98.161 -58.036 50.720
574482 98.195 -58.037 50.675
574483 98.229 -58.038 50.632
574484 98.263 -58.040 50.592
574485 98.297 -58.040 50.549
574486 98.331 -58.042 50.505
574487 98.364 -58.043 50.460
574488 98.394 -58.046 50.418
574489 98.426 -58.047 50.380
574490 98.456 -58.050 50.345
574491 98.488 -58.052 50.310
574492 98.519 -58.054 50.278
574493 98.550 -58.056 50.249
574494 98.581 -58.059 50.221
574495 98.612 -58.061 50.195
574496 98.642 -58.059 50.171
574497 98.657 -58.062 50.146
574498 98.646 -58.073 50.122
574499 98.645 -58.073 50.100
574500 98.643 -58.072 50.080
574501 98.657 -58.081 50.061
574502 98.681 -58.088 50.041
574503 98.725 -58.080 50.021
574504 98.766 -58.074 50.004
574505 98.806 -58.068 49.986
574506 98.846 -58.064 49.966
574507 98.885 -58.059 49.944
574508 98.925 -58.055 49.923
574509 98.965 -58.050 49.903
574510 99.005 -58.046 49.883
574511 99.045 -58.040 49.863
574512 99.084 -58.036 49.842
574513 99.124 -58.031 49.817
574514 99.164 -58.027 49.792
574515 99.204 -58.022 49.765
574516 99.243 -58.018 49.736
574517 99.282 -58.021 49.705
574518 99.319 -58.031 49.674
574519 99.325 -58.045 49.643
574520 99.355 -58.059 49.614
574521 99.423 -58.072 49.584
574522 99.474 -58.079 49.554
574523 99.511 -58.087 49.525
574524 99.546 -58.096 49.495
574525 99.580 -58.105 49.465
574526 99.615 -58.113 49.435
574527 99.649 -58.122 49.406
574528 99.684 -58.131 49.376
574529 99.718 -58.140 49.347
574530 99.753 -58.149 49.317
574884 87.852 -54.522 57.598
574885 87.887 -54.537 57.598
574886 87.922 -54.552 57.599
574887 87.958 -54.568 57.598
574888 87.993 -54.583 57.597
574889 88.029 -54.599 57.594
574890 88.064 -54.614 57.591
574891 88.100 -54.629 57.588
574892 88.135 -54.645 57.582
574893 88.170 -54.660 57.576
574894 88.212 -54.655 57.569
574895 88.278 -54.590 57.561
574896 88.335 -54.593 57.551
574897 88.381 -54.590 57.538
574898 88.417 -54.581 57.524
574899 88.453 -54.571 57.509
574900 88.489 -54.577 57.493
574901 88.528 -54.599 57.476
574902 88.566 -54.622 57.456
574903 88.605 -54.646 57.422
574904 88.644 -54.669 57.383
574905 88.682 -54.692 57.344
574906 88.721 -54.715 57.303
574907 88.760 -54.739 57.258
574908 88.798 -54.762 57.210
574909 88.837 -54.786 57.161
574910 88.876 -54.809 57.111
574911 88.914 -54.833 57.061
574912 88.953 -54.856 57.010
574913 88.991 -54.880 56.960
574914 89.030 -54.903 56.909
574915 89.068 -54.926 56.860
574947 90.188 -55.605 55.744
574948 90.214 -55.622 55.742
574949 90.241 -55.642 55.743
574950 90.268 -55.659 55.746
574951 90.293 -55.670 55.751
574952 90.319 -55.680 55.759
574953 90.344 -55.690 55.768
574954 90.383 -55.700 55.779
574955 90.423 -55.711 55.792
574956 90.463 -55.721 55.811
574957 90.502 -55.731 55.833
574958 90.542 -55.741 55.858
574959 90.582 -55.751 55.886
574960 90.621 -55.762 55.906
574961 90.661 -55.771 55.925
574962 90.700 -55.781 55.953
574963 90.740 -55.792 55.992
574964 90.780 -55.802 56.037
574965 90.819 -55.812 56.084
574966 90.859 -55.822 56.117
574967 90.899 -55.832 56.120
574968 90.939 -55.843 56.105
574969 90.978 -55.853 56.082
574970 91.018 -55.862 56.060
574971 91.056 -55.874 56.051
574972 91.091 -55.886 56.023
574973 91.132 -55.897 56.021
574974 91.176 -55.908 56.011
574975 91.221 -55.919 55.982
574976 91.255 -55.925 55.938
574977 91.287 -55.930 55.897
574978 91.320 -55.936 55.852
574979 91.352 -55.941 55.817
574980 91.385 -55.947 55.790
574981 91.417 -55.953 55.763
574982 91.450 -55.959 55.730
574983 91.482 -55.964 55.696
574984 91.515 -55.969 55.666
574985 91.548 -55.975 55.631
574986 91.584 -55.983 55.592
574987 91.616 -55.992 55.553
574988 91.653 -56.003 55.513
574989 91.690 -56.013 55.479
574990 91.727 -56.024 55.453
574991 91.764 -56.034 55.435
574992 91.800 -56.045 55.423
574993 91.837 -56.056 55.414
574994 91.873 -56.066 55.403
574995 91.910 -56.077 55.387
574996 91.947 -56.088 55.366
574997 91.983 -56.098 55.342
574998 92.019 -56.113 55.315
574999 92.051 -56.137 55.288
575000 92.083 -56.160 55.267
575001 92.116 -56.183 55.258
575002 92.148 -56.207 55.257
575003 92.180 -56.230 55.261
575004 92.218 -56.266 55.268
575005 92.255 -56.304 55.276
575006 92.292 -56.340 55.284
575007 92.330 -56.377 55.287
575008 92.367 -56.414 55.286
575009 92.404 -56.451 55.287
575010 92.441 -56.488 55.290
575011 92.480 -56.513 55.296
575012 92.519 -56.532 55.297
575013 92.558 -56.551 55.292
575014 92.597 -56.570 55.287
575015 92.636 -56.589 55.293
575016 92.675 -56.608 55.311
575017 92.713 -56.628 55.331
575018 92.747 -56.649 55.345
575019 92.782 -56.670 55.352
575020 92.817 -56.690 55.362
575021 92.852 -56.711 55.377
575022 92.893 -56.727 55.393
575023 92.936 -56.740 55.401
575024 92.979 -56.754 55.403
575025 93.022 -56.768 55.399
575026 93.065 -56.781 55.390
575027 93.109 -56.795 55.380
575028 93.149 -56.807 55.370
575029 93.179 -56.814 55.361
575030 93.207 -56.820 55.349
575031 93.236 -56.827 55.332
575032 93.265 -56.834 55.312
575033 93.294 -56.840 55.291
575034 93.327 -56.852 55.273
575035 93.359 -56.864 55.254
575036 93.392 -56.875 55.232
575037 93.425 -56.887 55.206
575038 93.458 -56.899 55.179
575039 93.490 -56.911 55.152
575040 93.523 -56.923 55.122
575041 93.556 -56.935 55.087
575042 93.588 -56.946 55.049
575043 93.621 -56.959 55.013
575044 93.653 -56.971 54.980
575045 93.686 -56.982 54.952
575046 93.719 -56.994 54.924
575047 93.752 -57.006 54.893
575048 93.784 -57.018 54.863
575049 93.817 -57.030 54.835
575050 93.850 -57.042 54.807
575051 93.883 -57.054 54.778
575052 93.915 -57.065 54.750
575053 93.948 -57.069 54.716
575054 93.982 -57.072 54.682
575055 94.016 -57.076 54.647
575056 94.050 -57.080 54.613
575057 94.083 -57.084 54.579
575058 94.117 -57.087 54.546
575059 94.151 -57.094 54.509
575060 94.189 -57.117 54.473
575061 94.226 -57.140 54.435
575062 94.263 -57.163 54.397
575063 94.299 -57.184 54.359
575064 94.332 -57.196 54.319
575065 94.364 -57.209 54.275
575066 94.397 -57.221 54.227
575067 94.429 -57.233 54.180
575068 94.461 -57.246 54.140
575069 94.494 -57.258 54.112
575070 94.526 -57.271 54.088
575071 94.558 -57.283 54.064
575072 94.591 -57.295 54.043
575073 94.623 -57.308 54.024
575074 94.655 -57.320 54.008
575075 94.688 -57.333 53.994
575076 94.720 -57.345 53.984
575077 94.752 -57.357 53.976
575078 94.784 -57.370 53.971
575079 94.817 -57.382 53.967
575080 94.849 -57.395 53.961
575081 94.882 -57.407 53.950
575082 94.914 -57.420 53.939
575083 94.946 -57.432 53.924
575084 94.979 -57.445 53.907
575085 95.011 -57.457 53.887
575086 95.043 -57.469 53.861
575087 95.073 -57.479 53.832
575088 95.103 -57.489 53.796
575089 95.133 -57.499 53.755
575090 95.163 -57.509 53.711
575091 95.193 -57.519 53.664
575092 95.224 -57.529 53.615
575093 95.254 -57.539 53.564
575094 95.284 -57.549 53.514
575095 95.314 -57.559 53.469
575096 95.344 -57.570 53.425
575097 95.374 -57.579 53.379
575098 95.404 -57.589 53.328
575099 95.444 -57.597 53.274
575100 95.487 -57.601 53.217
575101 95.531 -57.606 53.159
575102 95.574 -57.611 53.102
575103 95.618 -57.616 53.045
575104 95.662 -57.620 52.989
575105 95.706 -57.625 52.939
575106 95.750 -57.630 52.892
575107 95.794 -57.634 52.848
575108 95.823 -57.639 52.807
575109 95.843 -57.645 52.765
575110 95.865 -57.651 52.725
575111 95.886 -57.656 52.684
575112 95.906 -57.662 52.643
575113 95.927 -57.667 52.605
575114 95.948 -57.673 52.571
575115 95.974 -57.680 52.538
575116 96.004 -57.687 52.511
575117 96.033 -57.695 52.487
575118 96.061 -57.702 52.465
575119 96.091 -57.710 52.446
575120 96.120 -57.717 52.428
575121 96.149 -57.725 52.412
575122 96.178 -57.733 52.398
575123 96.207 -57.740 52.386
575124 96.239 -57.750 52.377
575125 96.273 -57.760 52.370
575126 96.307 -57.771 52.365
575127 96.341 -57.781 52.363
575128 96.375 -57.791 52.362
575129 96.408 -57.802 52.356
575130 96.442 -57.812 52.345
575131 96.476 -57.822 52.331
575132 96.510 -57.833 52.318
575133 96.544 -57.843 52.303
575134 96.578 -57.853 52.285
575135 96.612 -57.864 52.268
575136 96.646 -57.874 52.251
575137 96.679 -57.884 52.235
575138 96.711 -57.895 52.223
575139 96.739 -57.906 52.214
575140 96.767 -57.917 52.209
575141 96.795 -57.928 52.207
575142 96.823 -57.939 52.206
575143 96.852 -57.949 52.204
575144 96.880 -57.961 52.202
575145 96.908 -57.971 52.200
575146 96.938 -57.981 52.205
575147 96.975 -57.986 52.213
575148 97.009 -57.989 52.222
575149 97.037 -57.985 52.228
575150 97.066 -57.982 52.231
575151 97.094 -57.978 52.228
575152 97.123 -57.975 52.223
575153 97.152 -57.971 52.213
575154 97.182 -57.968 52.198
575155 97.212 -57.964 52.180
575156 97.241 -57.959 52.161
575157 97.271 -57.955 52.141
575158 97.300 -57.952 52.115
575159 97.330 -57.948 52.085
575160 97.360 -57.944 52.050
575161 97.389 -57.940 52.006
575162 97.424 -57.955 51.956
575163 97.459 -57.971 51.901
575164 97.494 -57.974 51.839
575165 97.530 -57.974 51.772
575166 97.566 -57.972 51.701
575167 97.601 -57.971 51.630
575168 97.637 -57.969 51.565
575169 97.672 -57.968 51.507
575170 97.708 -57.967 51.451
575171 97.743 -57.965 51.394
575172 97.779 -57.964 51.334
575173 97.815 -57.967 51.271
575174 97.853 -57.972 51.205
575175 97.891 -57.977 51.140
575176 97.929 -57.983 51.074
575177 97.967 -57.988 51.011
575178 98.005 -57.993 50.951
575179 98.043 -57.999 50.894
575180 98.083 -58.001 50.839
575181 98.124 -58.004 50.786
575182 98.163 -58.006 50.736
575183 98.199 -58.008 50.688
575184 98.232 -58.009 50.642
575185 98.267 -58.010 50.598
575186 98.300 -58.011 50.555
575187 98.335 -58.012 50.513
575188 98.368 -58.013 50.473
575189 98.400 -58.015 50.434
575190 98.431 -58.018 50.397
575191 98.461 -58.020 50.361
575192 98.493 -58.022 50.325
575193 98.523 -58.024 50.291
575194 98.555 -58.027 50.260
575195 98.584 -58.025 50.231
575196 98.614 -58.023 50.205
575197 98.644 -58.021 50.181
575198 98.673 -58.018 50.157
575199 98.671 -58.027 50.132
575200 98.658 -58.040 50.111
575201 98.646 -58.053 50.091
575202 98.637 -58.062 50.071
575203 98.676 -58.059 50.052
575204 98.716 -58.053 50.032
575205 98.756 -58.049 50.014
575206 98.795 -58.044 49.996
575207 98.835 -58.040 49.975
575208 98.875 -58.035 49.953
575209 98.915 -58.031 49.931
575210 98.954 -58.025 49.910
575211 98.994 -58.021 49.890
575212 99.034 -58.016 49.870
575213 99.074 -58.012 49.847
575214 99.114 -58.007 49.822
575215 99.153 -58.003 49.796
575216 99.193 -57.997 49.769
575217 99.233 -57.993 49.740
575218 99.272 -57.993 49.709
575219 99.308 -58.003 49.677
575220 99.333 -58.015 49.645
575221 99.336 -58.030 49.615
575222 99.403 -58.043 49.585
575223 99.460 -58.052 49.554
575224 99.504 -58.059 49.524
575225 99.539 -58.068 49.494
575226 99.573 -58.077 49.463
575227 99.608 -58.086 49.431
575228 99.642 -58.094 49.402
575229 99.677 -58.103 49.373
575230 99.711 -58.112 49.342
575231 99.746 -58.121 49.312
575584 87.841 -54.472 57.596
575585 87.876 -54.487 57.597
575586 87.912 -54.502 57.598
575587 87.947 -54.518 57.599
575588 87.983 -54.533 57.598
575589 88.018 -54.549 57.597
575590 88.053 -54.564 57.595
575591 88.088 -54.579 57.592
575592 88.124 -54.594 57.589
575593 88.160 -54.609 57.584
575594 88.195 -54.625 57.579
575595 88.230 -54.640 57.572
575596 88.295 -54.567 57.564
575597 88.354 -54.554 57.555
575598 88.404 -54.552 57.543
575599 88.440 -54.543 57.530
575600 88.476 -54.534 57.515
575601 88.511 -54.554 57.501
575602 88.550 -54.571 57.488
575603 88.588 -54.590 57.471
575604 88.626 -54.614 57.437
575605 88.665 -54.637 57.396
575606 88.704 -54.660 57.356
575607 88.742 -54.684 57.315
575608 88.781 -54.707 57.270
575609 88.820 -54.731 57.222
575610 88.858 -54.754 57.173
575611 88.897 -54.778 57.123
575612 88.935 -54.801 57.073
575613 88.974 -54.825 57.023
575614 89.013 -54.848 56.973
575615 89.051 -54.871 56.922
575616 89.090 -54.894 56.873
575648 90.207 -55.573 55.748
575649 90.231 -55.587 55.746
575650 90.258 -55.607 55.748
575651 90.286 -55.626 55.751
575652 90.311 -55.636 55.757
575653 90.336 -55.646 55.765
575654 90.362 -55.657 55.775
575655 90.395 -55.667 55.787
575656 90.435 -55.677 55.802
575657 90.474 -55.687 55.823
575658 90.514 -55.697 55.850
575659 90.554 -55.708 55.881
575660 90.593 -55.718 55.913
575661 90.633 -55.728 55.936
575662 90.673 -55.738 55.952
575663 90.713 -55.748 55.981
575664 90.752 -55.759 56.025
575665 90.792 -55.768 56.080
575666 90.831 -55.778 56.135
575667 90.871 -55.789 56.156
575668 90.910 -55.799 56.151
575669 90.950 -55.809 56.134
575670 90.990 -55.819 56.115
575671 91.029 -55.829 56.098
575672 91.069 -55.840 56.084
575673 91.109 -55.850 56.056
575674 91.146 -55.861 56.054
575675 91.188 -55.872 56.045
575676 91.226 -55.880 56.014
575677 91.259 -55.885 55.963
575678 91.291 -55.891 55.917
575679 91.324 -55.897 55.873
575680 91.357 -55.903 55.843
575681 91.393 -55.909 55.816
575682 91.429 -55.917 55.782
575683 91.464 -55.924 55.740
575684 91.500 -55.931 55.701
575685 91.536 -55.937 55.669
575686 91.572 -55.945 55.633
575687 91.604 -55.953 55.594
575688 91.635 -55.963 55.556
575689 91.666 -55.972 55.516
575690 91.703 -55.984 55.481
575691 91.739 -55.994 55.455
575692 91.776 -56.005 55.435
575693 91.812 -56.016 55.419
575694 91.849 -56.026 55.408
575695 91.886 -56.037 55.396
575696 91.922 -56.047 55.382
575697 91.959 -56.058 55.363
575698 91.995 -56.069 55.341
575699 92.028 -56.093 55.314
575700 92.060 -56.116 55.286
575701 92.093 -56.140 55.262
575702 92.125 -56.163 55.250
575703 92.157 -56.187 55.246
575704 92.190 -56.210 55.249
575705 92.223 -56.235 55.257
575706 92.260 -56.271 55.267
575707 92.297 -56.308 55.274
575708 92.335 -56.345 55.281
575709 92.372 -56.382 55.285
575710 92.409 -56.419 55.288
575711 92.446 -56.456 55.289
575712 92.485 -56.476 55.286
575713 92.524 -56.495 55.277
575714 92.563 -56.514 55.271
575715 92.602 -56.534 55.270
575716 92.641 -56.553 55.279
575717 92.681 -56.572 55.299
575718 92.719 -56.591 55.321
575719 92.754 -56.612 55.332
575720 92.788 -56.633 55.344
575721 92.823 -56.654 55.364
575722 92.858 -56.674 55.385
575723 92.901 -56.688 55.400
575724 92.944 -56.702 55.408
575725 92.987 -56.715 55.410
575726 93.030 -56.729 55.408
575727 93.073 -56.743 55.401
575728 93.117 -56.756 55.392
575729 93.160 -56.770 55.382
575730 93.190 -56.777 55.368
575731 93.218 -56.783 55.353
575732 93.248 -56.790 55.334
575733 93.276 -56.796 55.312
575734 93.305 -56.803 55.290
575735 93.338 -56.815 55.270
575736 93.371 -56.827 55.250
575737 93.404 -56.839 55.228
575738 93.436 -56.851 55.202
575739 93.469 -56.863 55.175
575740 93.502 -56.874 55.146
575741 93.535 -56.887 55.117
575742 93.567 -56.899 55.082
575743 93.600 -56.910 55.046
575744 93.632 -56.922 55.010
575745 93.665 -56.934 54.976
575746 93.698 -56.946 54.947
575747 93.730 -56.958 54.921
575748 93.763 -56.970 54.892
575749 93.796 -56.981 54.864
575750 93.828 -56.993 54.836
575751 93.861 -57.006 54.808
575752 93.894 -57.017 54.780
575753 93.926 -57.029 54.752
575754 93.959 -57.041 54.720
575755 93.993 -57.044 54.688
575756 94.027 -57.048 54.656
575757 94.060 -57.052 54.625
575758 94.094 -57.056 54.593
575759 94.127 -57.059 54.560
575760 94.161 -57.065 54.525
575761 94.199 -57.087 54.489
575762 94.236 -57.110 54.451
575763 94.273 -57.133 54.412
575764 94.308 -57.149 54.373
575765 94.340 -57.162 54.333
575766 94.372 -57.174 54.288
575767 94.404 -57.187 54.241
575768 94.436 -57.199 54.192
575769 94.469 -57.211 54.143
575770 94.501 -57.224 54.100
575771 94.534 -57.236 54.070
575772 94.566 -57.249 54.044
575773 94.598 -57.261 54.022
575774 94.631 -57.273 54.001
575775 94.663 -57.286 53.983
575776 94.695 -57.298 53.966
575777 94.728 -57.310 53.953
575778 94.760 -57.323 53.943
575779 94.792 -57.335 53.937
575780 94.825 -57.348 53.933
575781 94.857 -57.360 53.926
575782 94.889 -57.373 53.913
575783 94.921 -57.385 53.901
575784 94.954 -57.398 53.889
575785 94.986 -57.410 53.875
575786 95.019 -57.423 53.859
575787 95.051 -57.435 53.838
575788 95.082 -57.446 53.810
575789 95.112 -57.456 53.777
575790 95.142 -57.466 53.738
575791 95.172 -57.476 53.694
575792 95.202 -57.486 53.646
575793 95.233 -57.496 53.596
575794 95.263 -57.506 53.547
575795 95.293 -57.517 53.497
575796 95.323 -57.526 53.450
575797 95.353 -57.536 53.406
575798 95.384 -57.546 53.363
575799 95.419 -57.557 53.314
575800 95.456 -57.567 53.262
575801 95.498 -57.573 53.206
575802 95.542 -57.577 53.151
575803 95.586 -57.582 53.095
575804 95.629 -57.586 53.037
575805 95.673 -57.592 52.978
575806 95.717 -57.596 52.925
575807 95.761 -57.601 52.874
575808 95.805 -57.605 52.830
575809 95.828 -57.611 52.789
575810 95.849 -57.617 52.749
575811 95.870 -57.622 52.710
575812 95.891 -57.627 52.669
575813 95.912 -57.633 52.630
575814 95.933 -57.639 52.594
575815 95.954 -57.644 52.561
575816 95.977 -57.650 52.529
575817 96.006 -57.658 52.500
575818 96.035 -57.665 52.471
575819 96.064 -57.673 52.447
575820 96.093 -57.680 52.426
575821 96.122 -57.688 52.408
575822 96.151 -57.696 52.392
575823 96.180 -57.704 52.377
575824 96.209 -57.711 52.365
575825 96.238 -57.719 52.356
575826 96.269 -57.727 52.348
575827 96.304 -57.738 52.343
575828 96.337 -57.749 52.339
575829 96.371 -57.758 52.336
575830 96.405 -57.769 52.331
575831 96.439 -57.780 52.321
575832 96.473 -57.789 52.308
575833 96.507 -57.800 52.294
575834 96.541 -57.811 52.277
575835 96.575 -57.821 52.260
575836 96.608 -57.831 52.244
575837 96.643 -57.841 52.228
575838 96.676 -57.852 52.213
575839 96.705 -57.862 52.201
575840 96.734 -57.874 52.192
575841 96.762 -57.884 52.185
575842 96.790 -57.896 52.182
575843 96.818 -57.906 52.179
575844 96.847 -57.918 52.175
575845 96.875 -57.928 52.171
575846 96.906 -57.937 52.169
575847 96.943 -57.942 52.174
575848 96.979 -57.947 52.185
575849 97.014 -57.951 52.199
575850 97.043 -57.947 52.211
575851 97.071 -57.944 52.219
575852 97.100 -57.941 52.223
575853 97.128 -57.937 52.220
575854 97.157 -57.934 52.212
575855 97.185 -57.930 52.199
575856 97.214 -57.927 52.184
575857 97.243 -57.924 52.167
575858 97.273 -57.920 52.146
575859 97.303 -57.915 52.121
575860 97.332 -57.912 52.092
575861 97.362 -57.908 52.057
575862 97.391 -57.904 52.016
575863 97.425 -57.913 51.970
575864 97.459 -57.928 51.914
575865 97.494 -57.943 51.853
575866 97.530 -57.943 51.784
575867 97.565 -57.942 51.712
575868 97.601 -57.940 51.642
575869 97.636 -57.940 51.578
575870 97.672 -57.938 51.520
575871 97.708 -57.937 51.467
575872 97.742 -57.936 51.412
575873 97.774 -57.936 51.354
575874 97.805 -57.936 51.291
575875 97.840 -57.938 51.225
575876 97.878 -57.943 51.159
575877 97.916 -57.949 51.093
575878 97.954 -57.954 51.031
575879 97.992 -57.959 50.971
575880 98.030 -57.964 50.913
575881 98.070 -57.968 50.858
575882 98.110 -57.970 50.805
575883 98.150 -57.973 50.754
575884 98.190 -57.975 50.705
575885 98.230 -57.978 50.658
575886 98.270 -57.981 50.612
575887 98.304 -57.982 50.568
575888 98.338 -57.983 50.527
575889 98.372 -57.984 50.488
575890 98.405 -57.986 50.450
575891 98.435 -57.988 50.413
575892 98.466 -57.990 50.376
575893 98.497 -57.992 50.339
575894 98.527 -57.990 50.306
575895 98.557 -57.988 50.274
575896 98.586 -57.986 50.245
575897 98.616 -57.984 50.218
575898 98.645 -57.981 50.193
575899 98.675 -57.979 50.167
575900 98.698 -57.979 50.143
575901 98.685 -57.992 50.122
575902 98.672 -58.004 50.102
575903 98.660 -58.017 50.083
575904 98.680 -58.024 50.063
575905 98.705 -58.029 50.043
575906 98.745 -58.024 50.024
575907 98.785 -58.020 50.005
575908 98.825 -58.015 49.984
575909 98.864 -58.010 49.962
575910 98.904 -58.006 49.940
575911 98.944 -58.001 49.918
575912 98.984 -57.996 49.896
575913 99.023 -57.992 49.874
575914 99.063 -57.987 49.850
575915 99.103 -57.982 49.824
575916 99.143 -57.977 49.798
575917 99.183 -57.973 49.772
575918 99.222 -57.968 49.742
575919 99.262 -57.966 49.711
575920 99.298 -57.976 49.679
575921 99.336 -57.987 49.646
575922 99.344 -58.000 49.615
575923 99.382 -58.015 49.584
575924 99.445 -58.025 49.553
575925 99.496 -58.031 49.522
575926 99.531 -58.040 49.491
575927 99.565 -58.050 49.460
575928 99.600 -58.058 49.428
575929 99.634 -58.067 49.398
575930 99.669 -58.075 49.367
575931 99.703 -58.084 49.337
575932 99.738 -58.093 49.306
575933 99.772 -58.102 49.276
576285 87.866 -54.436 57.596
576286 87.901 -54.452 57.596
576287 87.936 -54.467 57.597
576288 87.972 -54.483 57.598
576289 88.007 -54.498 57.598
576290 88.043 -54.513 57.598
576291 88.078 -54.529 57.596
576292 88.113 -54.544 57.593
576293 88.149 -54.559 57.590
576294 88.184 -54.574 57.585
576295 88.219 -54.590 57.581
576296 88.255 -54.605 57.575
576297 88.311 -54.561 57.568
576298 88.374 -54.515 57.560
576299 88.426 -54.515 57.550
576300 88.462 -54.506 57.539
576301 88.498 -54.497 57.526
576302 88.534 -54.530 57.512
576303 88.571 -54.548 57.498
576304 88.610 -54.565 57.480
576305 88.648 -54.582 57.446
576306 88.686 -54.605 57.407
576307 88.725 -54.628 57.367
576308 88.764 -54.652 57.325
576309 88.802 -54.675 57.281
576310 88.841 -54.699 57.233
576311 88.880 -54.722 57.184
576312 88.918 -54.746 57.135
576313 88.957 -54.769 57.085
576314 88.995 -54.793 57.035
576315 89.034 -54.815 56.985
576316 89.072 -54.839 56.936
576317 89.111 -54.862 56.887
576348 90.203 -55.525 55.758
576349 90.226 -55.539 55.753
576350 90.250 -55.553 55.751
576351 90.276 -55.571 55.752
576352 90.303 -55.591 55.757
576353 90.329 -55.603 55.763
576354 90.355 -55.613 55.771
576355 90.380 -55.624 55.782
576356 90.407 -55.633 55.795
576357 90.447 -55.644 55.812
576358 90.486 -55.654 55.835
576359 90.526 -55.664 55.864
576360 90.566 -55.674 55.897
576361 90.605 -55.684 55.931
576362 90.645 -55.694 55.957
576363 90.684 -55.705 55.975
576364 90.724 -55.715 56.005
576365 90.764 -55.725 56.051
576366 90.804 -55.735 56.107
576367 90.843 -55.745 56.162
576368 90.883 -55.756 56.175
576369 90.923 -55.765 56.171
576370 90.962 -55.775 56.157
576371 91.002 -55.786 56.141
576372 91.041 -55.796 56.126
576373 91.081 -55.806 56.106
576374 91.120 -55.816 56.082
576375 91.160 -55.826 56.078
576376 91.202 -55.837 56.069
576377 91.238 -55.843 56.038
576378 91.274 -55.850 55.984
576379 91.309 -55.858 55.933
576380 91.345 -55.865 55.895
576381 91.381 -55.872 55.869
576382 91.417 -55.878 55.843
576383 91.452 -55.886 55.804
576384 91.488 -55.893 55.754
576385 91.524 -55.900 55.706
576386 91.560 -55.906 55.668
576387 91.592 -55.915 55.628
576388 91.623 -55.925 55.587
576389 91.653 -55.934 55.547
576390 91.684 -55.944 55.508
576391 91.715 -55.953 55.475
576392 91.751 -55.964 55.449
576393 91.788 -55.975 55.427
576394 91.825 -55.986 55.409
576395 91.861 -55.997 55.396
576396 91.898 -56.007 55.384
576397 91.935 -56.018 55.372
576398 91.971 -56.028 55.356
576399 92.005 -56.050 55.335
576400 92.037 -56.073 55.310
576401 92.069 -56.097 55.283
576402 92.102 -56.119 55.257
576403 92.134 -56.143 55.244
576404 92.167 -56.166 55.239
576405 92.199 -56.190 55.241
576406 92.231 -56.213 55.248
576407 92.265 -56.239 55.258
576408 92.302 -56.276 55.267
576409 92.339 -56.313 55.276
576410 92.376 -56.350 55.283
576411 92.414 -56.387 55.285
576412 92.451 -56.421 55.281
576413 92.490 -56.440 55.273
576414 92.529 -56.459 55.268
576415 92.569 -56.478 55.271
576416 92.607 -56.498 55.280
576417 92.647 -56.517 55.290
576418 92.686 -56.536 55.301
576419 92.725 -56.555 55.309
576420 92.760 -56.576 55.321
576421 92.795 -56.596 55.342
576422 92.829 -56.617 55.365
576423 92.866 -56.636 55.385
576424 92.910 -56.649 55.400
576425 92.952 -56.663 55.410
576426 92.996 -56.677 55.415
576427 93.039 -56.690 55.418
576428 93.082 -56.704 55.417
576429 93.125 -56.718 55.411
576430 93.168 -56.731 55.401
576431 93.201 -56.740 55.382
576432 93.229 -56.746 55.363
576433 93.259 -56.752 55.342
576434 93.287 -56.759 55.321
576435 93.317 -56.767 55.298
576436 93.350 -56.779 55.275
576437 93.382 -56.790 55.251
576438 93.415 -56.802 55.226
576439 93.448 -56.815 55.198
576440 93.480 -56.826 55.169
576441 93.513 -56.838 55.139
576442 93.546 -56.850 55.110
576443 93.578 -56.862 55.075
576444 93.611 -56.874 55.040
576445 93.644 -56.886 55.006
576446 93.677 -56.897 54.973
576447 93.709 -56.909 54.942
576448 93.742 -56.921 54.913
576449 93.774 -56.933 54.882
576450 93.807 -56.945 54.854
576451 93.840 -56.957 54.826
576452 93.872 -56.968 54.800
576453 93.905 -56.981 54.775
576454 93.938 -56.993 54.748
576455 93.971 -57.005 54.719
576456 94.003 -57.016 54.689
576457 94.037 -57.021 54.660
576458 94.070 -57.024 54.631
576459 94.104 -57.028 54.602
576460 94.138 -57.031 54.571
576461 94.172 -57.035 54.536
576462 94.209 -57.058 54.499
576463 94.246 -57.081 54.462
576464 94.283 -57.103 54.424
576465 94.315 -57.115 54.385
576466 94.347 -57.127 54.345
576467 94.380 -57.140 54.302
576468 94.412 -57.152 54.255
576469 94.445 -57.165 54.205
576470 94.477 -57.177 54.151
576471 94.509 -57.189 54.095
576472 94.541 -57.202 54.056
576473 94.574 -57.214 54.026
576474 94.606 -57.226 54.001
576475 94.638 -57.239 53.979
576476 94.671 -57.251 53.959
576477 94.703 -57.264 53.940
576478 94.736 -57.276 53.925
576479 94.768 -57.288 53.912
576480 94.800 -57.301 53.903
576481 94.832 -57.313 53.896
576482 94.864 -57.326 53.886
576483 94.897 -57.338 53.874
576484 94.929 -57.351 53.864
576485 94.962 -57.363 53.853
576486 94.994 -57.376 53.842
576487 95.026 -57.388 53.827
576488 95.059 -57.401 53.807
576489 95.091 -57.413 53.783
576490 95.121 -57.423 53.750
576491 95.151 -57.433 53.713
576492 95.182 -57.443 53.670
576493 95.212 -57.453 53.622
576494 95.242 -57.464 53.571
576495 95.272 -57.473 53.520
576496 95.302 -57.483 53.472
576497 95.332 -57.494 53.426
576498 95.363 -57.504 53.382
576499 95.399 -57.514 53.340
576500 95.436 -57.523 53.295
576501 95.472 -57.534 53.244
576502 95.509 -57.544 53.192
576503 95.552 -57.548 53.139
576504 95.596 -57.553 53.085
576505 95.640 -57.558 53.028
576506 95.684 -57.563 52.971
576507 95.728 -57.567 52.916
576508 95.771 -57.572 52.862
576509 95.812 -57.576 52.815
576510 95.833 -57.583 52.773
576511 95.854 -57.588 52.734
576512 95.875 -57.593 52.695
576513 95.896 -57.599 52.655
576514 95.917 -57.605 52.617
576515 95.938 -57.610 52.582
576516 95.959 -57.616 52.551
576517 95.980 -57.621 52.521
576518 96.008 -57.629 52.489
576519 96.037 -57.636 52.459
576520 96.066 -57.644 52.432
576521 96.095 -57.651 52.408
576522 96.124 -57.659 52.389
576523 96.154 -57.667 52.372
576524 96.182 -57.674 52.358
576525 96.212 -57.682 52.347
576526 96.241 -57.689 52.338
576527 96.269 -57.697 52.330
576528 96.300 -57.705 52.324
576529 96.334 -57.716 52.320
576530 96.368 -57.726 52.316
576531 96.402 -57.736 52.311
576532 96.436 -57.747 52.302
576533 96.470 -57.757 52.288
576534 96.504 -57.767 52.271
576535 96.537 -57.777 52.252
576536 96.571 -57.788 52.235
576537 96.605 -57.799 52.218
576538 96.639 -57.808 52.202
576539 96.672 -57.819 52.189
576540 96.701 -57.830 52.178
576541 96.729 -57.841 52.169
576542 96.757 -57.852 52.162
576543 96.786 -57.863 52.156
576544 96.814 -57.874 52.150
576545 96.842 -57.885 52.145
576546 96.874 -57.893 52.140
576547 96.910 -57.898 52.140
576548 96.947 -57.903 52.145
576549 96.983 -57.908 52.156
576550 97.020 -57.913 52.171
576551 97.048 -57.909 52.186
576552 97.077 -57.906 52.198
576553 97.105 -57.903 52.205
576554 97.134 -57.899 52.207
576555 97.162 -57.896 52.202
576556 97.190 -57.893 52.194
576557 97.219 -57.890 52.182
576558 97.247 -57.887 52.167
576559 97.276 -57.883 52.145
576560 97.305 -57.880 52.118
576561 97.334 -57.876 52.087
576562 97.364 -57.872 52.054
576563 97.393 -57.868 52.017
576564 97.425 -57.871 51.977
576565 97.460 -57.886 51.925
576566 97.495 -57.900 51.863
576567 97.529 -57.913 51.795
576568 97.565 -57.912 51.723
576569 97.600 -57.911 51.653
576570 97.636 -57.909 51.589
576571 97.672 -57.908 51.531
576572 97.705 -57.907 51.479
576573 97.737 -57.907 51.426
576574 97.769 -57.907 51.369
576575 97.800 -57.908 51.307
576576 97.831 -57.908 51.240
576577 97.865 -57.909 51.177
576578 97.903 -57.914 51.113
576579 97.941 -57.919 51.051
576580 97.979 -57.924 50.992
576581 98.017 -57.930 50.935
576582 98.056 -57.934 50.879
576583 98.096 -57.937 50.825
576584 98.136 -57.939 50.773
576585 98.176 -57.942 50.725
576586 98.216 -57.944 50.679
576587 98.256 -57.947 50.634
576588 98.296 -57.949 50.587
576589 98.337 -57.952 50.545
576590 98.375 -57.955 50.506
576591 98.409 -57.955 50.468
576592 98.440 -57.958 50.430
576593 98.470 -57.956 50.392
576594 98.500 -57.954 50.353
576595 98.529 -57.951 50.321
576596 98.559 -57.949 50.290
576597 98.588 -57.946 50.260
576598 98.618 -57.944 50.233
576599 98.647 -57.942 50.206
576600 98.677 -57.940 50.178
576601 98.707 -57.937 50.154
576602 98.712 -57.943 50.134
576603 98.699 -57.956 50.115
576604 98.687 -57.968 50.095
576605 98.704 -57.976 50.075
576606 98.727 -57.982 50.056
576607 98.750 -57.989 50.035
576608 98.774 -57.995 50.014
576609 98.814 -57.990 49.992
576610 98.854 -57.986 49.971
576611 98.894 -57.981 49.949
576612 98.933 -57.977 49.925
576613 98.973 -57.971 49.901
576614 99.013 -57.967 49.876
576615 99.053 -57.962 49.851
576616 99.092 -57.958 49.825
576617 99.132 -57.953 49.799
576618 99.172 -57.949 49.773
576619 99.212 -57.943 49.743
576620 99.252 -57.939 49.712
576621 99.288 -57.949 49.680
576622 99.326 -57.959 49.646
576623 99.352 -57.971 49.614
576624 99.361 -57.985 49.583
576625 99.430 -57.998 49.551
576626 99.482 -58.005 49.520
576627 99.523 -58.012 49.488
576628 99.558 -58.021 49.457
576629 99.592 -58.031 49.424
576630 99.627 -58.039 49.393
576631 99.661 -58.048 49.362
576632 99.696 -58.056 49.331
576633 99.730 -58.065 49.300
576634 99.765 -58.074 49.269
576635 99.799 -58.083 49.239
576985 87.855 -54.386 57.592
576986 87.891 -54.402 57.594
576987 87.926 -54.417 57.595
576988 87.961 -54.433 57.596
576989 87.996 -54.448 57.597
576990 88.032 -54.463 57.597
576991 88.067 -54.478 57.597
576992 88.103 -54.494 57.597
576993 88.138 -54.509 57.594
576994 88.173 -54.524 57.590
576995 88.209 -54.540 57.586
576996 88.244 -54.555 57.582
576997 88.279 -54.571 57.577
576998 88.326 -54.554 57.571
576999 88.393 -54.475 57.564
577000 88.449 -54.478 57.556
577001 88.485 -54.468 57.548
577002 88.521 -54.471 57.537
577003 88.555 -54.507 57.523
577004 88.594 -54.524 57.508
577005 88.631 -54.541 57.489
577006 88.670 -54.558 57.454
577007 88.708 -54.574 57.415
577008 88.746 -54.596 57.375
577009 88.785 -54.620 57.334
577010 88.824 -54.643 57.291
577011 88.862 -54.667 57.243
577012 88.901 -54.690 57.194
577013 88.940 -54.714 57.145
577014 88.978 -54.737 57.095
577015 89.017 -54.761 57.045
577016 89.055 -54.784 56.997
577017 89.094 -54.807 56.948
577018 89.132 -54.831 56.899
577048 90.199 -55.478 55.770
577049 90.222 -55.492 55.764
577050 90.245 -55.506 55.761
577051 90.268 -55.520 55.759
577052 90.293 -55.536 55.760
577053 90.321 -55.555 55.764
577054 90.347 -55.569 55.770
577055 90.372 -55.580 55.779
577056 90.398 -55.589 55.790
577057 90.424 -55.600 55.805
577058 90.458 -55.610 55.821
577059 90.498 -55.621 55.845
577060 90.538 -55.630 55.875
577061 90.578 -55.641 55.908
577062 90.617 -55.651 55.942
577063 90.657 -55.661 55.971
577064 90.696 -55.671 55.994
577065 90.736 -55.681 56.024
577066 90.776 -55.691 56.068
577067 90.815 -55.702 56.123
577068 90.855 -55.712 56.170
577069 90.894 -55.722 56.187
577070 90.934 -55.732 56.187
577071 90.974 -55.742 56.175
577072 91.014 -55.752 56.158
577073 91.053 -55.762 56.133
577074 91.093 -55.772 56.112
577075 91.133 -55.783 56.097
577076 91.178 -55.793 56.089
577077 91.224 -55.805 56.077
577078 91.261 -55.812 56.048
577079 91.297 -55.820 56.000
577080 91.333 -55.827 55.947
577081 91.369 -55.834 55.916
577082 91.404 -55.840 55.895
577083 91.440 -55.848 55.869
577084 91.476 -55.855 55.828
577085 91.512 -55.862 55.770
577086 91.547 -55.868 55.712
577087 91.581 -55.877 55.666
577088 91.611 -55.887 55.621
577089 91.642 -55.897 55.577
577090 91.672 -55.906 55.537
577091 91.703 -55.915 55.499
577092 91.733 -55.925 55.467
577093 91.764 -55.934 55.439
577094 91.800 -55.945 55.416
577095 91.837 -55.956 55.396
577096 91.874 -55.966 55.381
577097 91.911 -55.977 55.371
577098 91.947 -55.987 55.361
577099 91.981 -56.006 55.349
577100 92.014 -56.029 55.332
577101 92.046 -56.053 55.309
577102 92.079 -56.076 55.281
577103 92.111 -56.100 55.252
577104 92.143 -56.123 55.239
577105 92.176 -56.147 55.234
577106 92.208 -56.170 55.235
577107 92.241 -56.193 55.241
577108 92.273 -56.216 55.249
577109 92.307 -56.244 55.261
577110 92.344 -56.281 55.271
577111 92.381 -56.318 55.278
577112 92.419 -56.354 55.278
577113 92.456 -56.385 55.270
577114 92.496 -56.404 55.262
577115 92.535 -56.423 55.266
577116 92.574 -56.442 55.280
577117 92.613 -56.461 55.296
577118 92.652 -56.480 55.309
577119 92.691 -56.499 55.317
577120 92.730 -56.518 55.320
577121 92.766 -56.539 55.328
577122 92.801 -56.560 55.345
577123 92.836 -56.580 55.365
577124 92.874 -56.598 55.385
577125 92.918 -56.611 55.400
577126 92.961 -56.625 55.411
577127 93.004 -56.639 55.420
577128 93.047 -56.652 55.427
577129 93.090 -56.665 55.429
577130 93.133 -56.679 55.427
577131 93.176 -56.693 55.421
577132 93.212 -56.702 55.403
577133 93.240 -56.709 55.386
577134 93.270 -56.716 55.367
577135 93.298 -56.722 55.346
577136 93.329 -56.730 55.322
577137 93.361 -56.742 55.296
577138 93.394 -56.754 55.261
577139 93.426 -56.766 55.227
577140 93.459 -56.777 55.195
577141 93.492 -56.790 55.164
577142 93.524 -56.802 55.133
577143 93.557 -56.814 55.104
577144 93.590 -56.825 55.070
577145 93.623 -56.837 55.036
577146 93.655 -56.849 55.003
577147 93.688 -56.861 54.971
577148 93.720 -56.873 54.939
577149 93.753 -56.885 54.907
577150 93.786 -56.896 54.874
577151 93.819 -56.909 54.843
577152 93.851 -56.921 54.814
577153 93.884 -56.932 54.787
577154 93.917 -56.944 54.763
577155 93.949 -56.956 54.738
577156 93.982 -56.968 54.712
577157 94.014 -56.980 54.684
577158 94.047 -56.992 54.657
577159 94.081 -56.997 54.630
577160 94.115 -57.000 54.602
577161 94.148 -57.004 54.572
577162 94.182 -57.007 54.539
577163 94.219 -57.028 54.503
577164 94.256 -57.051 54.465
577165 94.290 -57.068 54.427
577166 94.323 -57.081 54.390
577167 94.355 -57.093 54.350
577168 94.388 -57.105 54.308
577169 94.420 -57.118 54.263
577170 94.452 -57.130 54.213
577171 94.484 -57.142 54.159
577172 94.517 -57.155 54.101
577173 94.549 -57.167 54.049
577174 94.581 -57.180 54.010
577175 94.614 -57.192 53.981
577176 94.646 -57.204 53.957
577177 94.678 -57.217 53.936
577178 94.711 -57.229 53.916
577179 94.743 -57.241 53.899
577180 94.775 -57.254 53.883
577181 94.808 -57.266 53.871
577182 94.840 -57.279 53.859
577183 94.873 -57.291 53.846
577184 94.905 -57.304 53.834
577185 94.937 -57.316 53.825
577186 94.969 -57.329 53.815
577187 95.001 -57.341 53.805
577188 95.034 -57.354 53.791
577189 95.066 -57.366 53.772
577190 95.099 -57.379 53.750
577191 95.130 -57.390 53.721
577192 95.161 -57.401 53.685
577193 95.191 -57.410 53.644
577194 95.221 -57.420 53.598
577195 95.251 -57.430 53.549
577196 95.281 -57.441 53.498
577197 95.311 -57.451 53.448
577198 95.343 -57.460 53.401
577199 95.379 -57.470 53.356
577200 95.415 -57.481 53.314
577201 95.452 -57.491 53.271
577202 95.488 -57.501 53.223
577203 95.525 -57.511 53.174
577204 95.563 -57.520 53.124
577205 95.607 -57.524 53.073
577206 95.651 -57.529 53.019
577207 95.695 -57.534 52.965
577208 95.739 -57.539 52.910
577209 95.782 -57.543 52.859
577210 95.818 -57.548 52.811
577211 95.839 -57.554 52.766
577212 95.860 -57.559 52.722
577213 95.881 -57.565 52.682
577214 95.901 -57.570 52.642
577215 95.922 -57.576 52.603
577216 95.944 -57.581 52.568
577217 95.964 -57.587 52.537
577218 95.985 -57.592 52.508
577219 96.011 -57.599 52.478
577220 96.039 -57.607 52.449
577221 96.068 -57.614 52.420
577222 96.098 -57.622 52.393
577223 96.126 -57.630 52.371
577224 96.156 -57.637 52.354
577225 96.185 -57.645 52.340
577226 96.214 -57.652 52.332
577227 96.243 -57.660 52.324
577228 96.272 -57.667 52.314
577229 96.301 -57.675 52.306
577230 96.331 -57.683 52.301
577231 96.365 -57.693 52.298
577232 96.398 -57.704 52.294
577233 96.433 -57.714 52.283
577234 96.466 -57.724 52.268
577235 96.500 -57.735 52.250
577236 96.534 -57.745 52.229
577237 96.568 -57.755 52.212
577238 96.602 -57.766 52.194
577239 96.636 -57.776 52.178
577240 96.668 -57.786 52.166
577241 96.696 -57.798 52.157
577242 96.724 -57.808 52.147
577243 96.752 -57.820 52.140
577244 96.781 -57.830 52.132
577245 96.809 -57.841 52.124
577246 96.841 -57.849 52.118
577247 96.878 -57.854 52.112
577248 96.914 -57.859 52.113
577249 96.951 -57.864 52.118
577250 96.987 -57.869 52.128
577251 97.024 -57.874 52.143
577252 97.054 -57.872 52.160
577253 97.082 -57.868 52.175
577254 97.111 -57.865 52.184
577255 97.139 -57.861 52.188
577256 97.168 -57.858 52.187
577257 97.196 -57.855 52.182
577258 97.225 -57.852 52.173
577259 97.253 -57.849 52.158
577260 97.282 -57.845 52.135
577261 97.310 -57.842 52.107
577262 97.338 -57.839 52.078
577263 97.367 -57.835 52.046
577264 97.395 -57.832 52.013
577265 97.425 -57.828 51.977
577266 97.460 -57.843 51.928
577267 97.495 -57.858 51.869
577268 97.529 -57.873 51.804
577269 97.565 -57.882 51.735
577270 97.600 -57.880 51.665
577271 97.636 -57.879 51.598
577272 97.668 -57.879 51.539
577273 97.700 -57.879 51.487
577274 97.731 -57.879 51.434
577275 97.763 -57.879 51.379
577276 97.795 -57.879 51.320
577277 97.826 -57.879 51.257
577278 97.858 -57.879 51.194
577279 97.890 -57.880 51.132
577280 97.928 -57.885 51.071
577281 97.966 -57.890 51.013
577282 98.004 -57.896 50.957
577283 98.043 -57.900 50.901
577284 98.083 -57.903 50.846
577285 98.123 -57.905 50.794
577286 98.163 -57.909 50.746
577287 98.203 -57.911 50.701
577288 98.243 -57.913 50.658
577289 98.283 -57.916 50.613
577290 98.323 -57.918 50.573
577291 98.363 -57.921 50.532
577292 98.407 -57.924 50.491
577293 98.442 -57.921 50.450
577294 98.472 -57.917 50.410
577295 98.501 -57.915 50.372
577296 98.531 -57.912 50.339
577297 98.561 -57.910 50.308
577298 98.590 -57.907 50.278
577299 98.620 -57.905 50.249
577300 98.649 -57.902 50.221
577301 98.679 -57.900 50.194
577302 98.709 -57.898 50.171
577303 98.738 -57.896 50.151
577304 98.726 -57.908 50.131
577305 98.713 -57.921 50.111
577306 98.728 -57.928 50.090
577307 98.751 -57.935 50.070
577308 98.774 -57.941 50.047
577309 98.798 -57.948 50.024
577310 98.821 -57.954 50.001
577311 98.844 -57.961 49.978
577312 98.882 -57.956 49.955
577313 98.922 -57.952 49.930
577314 98.962 -57.947 49.904
577315 99.002 -57.943 49.878
577316 99.042 -57.938 49.853
577317 99.082 -57.934 49.826
577318 99.122 -57.928 49.800
577319 99.161 -57.924 49.772
577320 99.201 -57.919 49.743
577321 99.241 -57.915 49.712
577322 99.278 -57.921 49.679
577323 99.315 -57.931 49.647
577324 99.352 -57.942 49.613
577325 99.363 -57.955 49.582
577326 99.410 -57.969 49.550
577327 99.467 -57.978 49.517
577328 99.516 -57.985 49.485
577329 99.550 -57.993 49.453
577330 99.585 -58.003 49.421
577331 99.619 -58.011 49.390
577332 99.654 -58.020 49.358
577333 99.688 -58.029 49.326
577334 99.723 -58.037 49.294
577335 99.757 -58.046 49.263
577336 99.791 -58.055 49.232
577337 99.826 -58.064 49.201
577686 87.880 -54.352 57.590
577687 87.915 -54.367 57.592
577688 87.951 -54.382 57.593
577689 87.986 -54.397 57.594
577690 88.021 -54.413 57.595
577691 88.057 -54.428 57.596
577692 88.092 -54.443 57.596
577693 88.127 -54.458 57.596
577694 88.163 -54.474 57.594
577695 88.198 -54.489 57.591
577696 88.234 -54.505 57.587
577697 88.269 -54.520 57.583
577698 88.304 -54.535 57.579
577699 88.341 -54.547 57.574
577700 88.408 -54.466 57.568
577701 88.469 -54.439 57.562
577702 88.507 -54.431 57.555
577703 88.543 -54.446 57.545
577704 88.577 -54.482 57.534
577705 88.615 -54.501 57.519
577706 88.653 -54.518 57.496
577707 88.692 -54.534 57.460
577708 88.730 -54.552 57.421
577709 88.768 -54.568 57.382
577710 88.806 -54.588 57.342
577711 88.845 -54.612 57.300
577712 88.884 -54.635 57.253
577713 88.922 -54.659 57.204
577714 88.961 -54.682 57.155
577715 88.999 -54.706 57.106
577716 89.038 -54.728 57.056
577717 89.076 -54.752 57.008
577718 89.115 -54.775 56.959
577719 89.154 -54.799 56.910
577748 90.194 -55.430 55.785
577749 90.218 -55.445 55.777
577750 90.241 -55.458 55.771
577751 90.264 -55.473 55.768
577752 90.288 -55.487 55.767
577753 90.311 -55.501 55.768
577754 90.338 -55.520 55.772
577755 90.365 -55.536 55.778
577756 90.390 -55.546 55.787
577757 90.415 -55.556 55.799
577758 90.441 -55.567 55.815
577759 90.470 -55.577 55.833
577760 90.510 -55.587 55.855
577761 90.549 -55.597 55.881
577762 90.589 -55.607 55.911
577763 90.629 -55.618 55.946
577764 90.669 -55.627 55.978
577765 90.708 -55.638 56.007
577766 90.748 -55.648 56.037
577767 90.788 -55.658 56.076
577768 90.827 -55.668 56.128
577769 90.867 -55.678 56.165
577770 90.906 -55.688 56.192
577771 90.946 -55.699 56.197
577772 90.986 -55.709 56.188
577773 91.025 -55.719 56.167
577774 91.065 -55.729 56.132
577775 91.109 -55.740 56.116
577776 91.154 -55.751 56.104
577777 91.200 -55.762 56.090
577778 91.245 -55.773 56.071
577779 91.285 -55.781 56.044
577780 91.321 -55.789 56.008
577781 91.357 -55.796 55.956
577782 91.392 -55.803 55.926
577783 91.428 -55.809 55.901
577784 91.464 -55.817 55.869
577785 91.500 -55.824 55.830
577786 91.535 -55.831 55.775
577787 91.569 -55.839 55.714
577788 91.599 -55.848 55.661
577789 91.630 -55.858 55.612
577790 91.660 -55.868 55.567
577791 91.691 -55.877 55.526
577792 91.722 -55.887 55.488
577793 91.752 -55.897 55.455
577794 91.783 -55.906 55.426
577795 91.813 -55.915 55.401
577796 91.850 -55.926 55.380
577797 91.886 -55.937 55.364
577798 91.923 -55.947 55.353
577799 91.958 -55.962 55.343
577800 91.991 -55.986 55.332
577801 92.023 -56.009 55.317
577802 92.055 -56.033 55.297
577803 92.088 -56.056 55.272
577804 92.120 -56.079 55.243
577805 92.153 -56.103 55.226
577806 92.185 -56.126 55.217
577807 92.217 -56.150 55.216
577808 92.250 -56.173 55.221
577809 92.282 -56.197 55.227
577810 92.315 -56.220 55.242
577811 92.349 -56.249 55.255
577812 92.386 -56.285 55.261
577813 92.424 -56.323 55.260
577814 92.462 -56.348 55.255
577815 92.501 -56.367 55.257
577816 92.540 -56.386 55.276
577817 92.579 -56.406 55.299
577818 92.618 -56.425 55.319
577819 92.657 -56.444 55.334
577820 92.697 -56.463 55.343
577821 92.735 -56.482 55.344
577822 92.772 -56.502 55.346
577823 92.807 -56.523 55.352
577824 92.842 -56.544 55.365
577825 92.883 -56.559 55.385
577826 92.926 -56.573 55.399
577827 92.969 -56.586 55.410
577828 93.012 -56.600 55.421
577829 93.055 -56.614 55.430
577830 93.098 -56.627 55.435
577831 93.141 -56.641 55.436
577832 93.184 -56.655 55.431
577833 93.223 -56.666 55.416
577834 93.252 -56.672 55.401
577835 93.281 -56.679 55.384
577836 93.309 -56.685 55.365
577837 93.340 -56.694 55.345
577838 93.372 -56.705 55.321
577839 93.405 -56.718 55.288
577840 93.438 -56.730 55.254
577841 93.471 -56.741 55.218
577842 93.503 -56.753 55.182
577843 93.536 -56.765 55.144
577844 93.569 -56.777 55.108
577845 93.602 -56.789 55.070
577846 93.634 -56.801 55.033
577847 93.666 -56.812 55.000
577848 93.699 -56.824 54.968
577849 93.732 -56.837 54.937
577850 93.765 -56.848 54.903
577851 93.797 -56.860 54.867
577852 93.830 -56.872 54.832
577853 93.863 -56.884 54.800
577854 93.896 -56.896 54.772
577855 93.928 -56.908 54.748
577856 93.961 -56.919 54.726
577857 93.993 -56.931 54.701
577858 94.026 -56.943 54.676
577859 94.059 -56.955 54.650
577860 94.091 -56.967 54.624
577861 94.125 -56.972 54.598
577862 94.158 -56.976 54.569
577863 94.192 -56.980 54.536
577864 94.229 -56.999 54.501
577865 94.266 -57.021 54.463
577866 94.298 -57.034 54.424
577867 94.330 -57.046 54.386
577868 94.363 -57.058 54.348
577869 94.395 -57.071 54.306
577870 94.427 -57.083 54.261
577871 94.460 -57.096 54.212
577872 94.492 -57.108 54.160
577873 94.525 -57.120 54.102
577874 94.557 -57.133 54.047
577875 94.589 -57.145 53.999
577876 94.621 -57.157 53.962
577877 94.654 -57.170 53.935
577878 94.686 -57.182 53.912
577879 94.718 -57.195 53.892
577880 94.751 -57.207 53.874
577881 94.783 -57.219 53.857
577882 94.816 -57.232 53.841
577883 94.848 -57.244 53.826
577884 94.880 -57.257 53.812
577885 94.912 -57.269 53.796
577886 94.945 -57.282 53.784
577887 94.977 -57.294 53.774
577888 95.010 -57.307 53.763
577889 95.042 -57.319 53.750
577890 95.074 -57.332 53.733
577891 95.106 -57.344 53.714
577892 95.139 -57.357 53.687
577893 95.170 -57.367 53.655
577894 95.200 -57.377 53.616
577895 95.230 -57.388 53.573
577896 95.260 -57.398 53.526
577897 95.290 -57.407 53.477
577898 95.323 -57.417 53.425
577899 95.359 -57.428 53.376
577900 95.395 -57.438 53.330
577901 95.431 -57.448 53.287
577902 95.468 -57.458 53.244
577903 95.505 -57.468 53.199
577904 95.541 -57.479 53.153
577905 95.577 -57.489 53.107
577906 95.618 -57.495 53.059
577907 95.662 -57.500 53.009
577908 95.705 -57.504 52.958
577909 95.749 -57.510 52.904
577910 95.793 -57.514 52.854
577911 95.823 -57.520 52.807
577912 95.844 -57.525 52.761
577913 95.865 -57.530 52.716
577914 95.886 -57.536 52.674
577915 95.907 -57.542 52.634
577916 95.928 -57.547 52.594
577917 95.949 -57.553 52.558
577918 95.970 -57.558 52.526
577919 95.991 -57.564 52.494
577920 96.013 -57.570 52.463
577921 96.042 -57.577 52.434
577922 96.071 -57.585 52.406
577923 96.100 -57.592 52.381
577924 96.129 -57.600 52.357
577925 96.158 -57.608 52.337
577926 96.187 -57.615 52.321
577927 96.216 -57.623 52.309
577928 96.245 -57.630 52.298
577929 96.274 -57.638 52.289
577930 96.303 -57.645 52.281
577931 96.332 -57.653 52.274
577932 96.362 -57.661 52.270
577933 96.395 -57.671 52.267
577934 96.429 -57.681 52.261
577935 96.463 -57.692 52.251
577936 96.497 -57.702 52.233
577937 96.531 -57.712 52.212
577938 96.565 -57.723 52.191
577939 96.599 -57.733 52.173
577940 96.632 -57.743 52.157
577941 96.663 -57.754 52.145
577942 96.691 -57.765 52.137
577943 96.720 -57.776 52.129
577944 96.748 -57.787 52.122
577945 96.776 -57.798 52.110
577946 96.809 -57.805 52.100
577947 96.846 -57.810 52.092
577948 96.882 -57.815 52.088
577949 96.919 -57.820 52.088
577950 96.955 -57.825 52.093
577951 96.992 -57.830 52.102
577952 97.028 -57.835 52.116
577953 97.059 -57.834 52.134
577954 97.088 -57.830 52.150
577955 97.116 -57.827 52.161
577956 97.145 -57.824 52.166
577957 97.173 -57.821 52.168
577958 97.202 -57.818 52.166
577959 97.230 -57.814 52.159
577960 97.258 -57.811 52.143
577961 97.287 -57.808 52.120
577962 97.315 -57.804 52.092
577963 97.344 -57.801 52.064
577964 97.372 -57.797 52.035
577965 97.400 -57.794 52.005
577966 97.428 -57.791 51.972
577967 97.461 -57.802 51.927
577968 97.496 -57.817 51.872
577969 97.530 -57.830 51.810
577970 97.565 -57.846 51.744
577971 97.599 -57.851 51.676
577972 97.631 -57.851 51.608
577973 97.663 -57.851 51.547
577974 97.694 -57.851 51.492
577975 97.726 -57.851 51.439
577976 97.757 -57.851 51.385
577977 97.789 -57.851 51.328
577978 97.820 -57.851 51.267
577979 97.852 -57.851 51.205
577980 97.884 -57.851 51.145
577981 97.916 -57.851 51.086
577982 97.953 -57.856 51.030
577983 97.992 -57.861 50.974
577984 98.029 -57.866 50.919
577985 98.069 -57.869 50.864
577986 98.109 -57.872 50.812
577987 98.149 -57.874 50.763
577988 98.190 -57.877 50.717
577989 98.229 -57.880 50.675
577990 98.269 -57.883 50.631
577991 98.311 -57.885 50.591
577992 98.356 -57.888 50.551
577993 98.401 -57.890 50.510
577994 98.442 -57.891 50.469
577995 98.473 -57.886 50.429
577996 98.503 -57.880 50.392
577997 98.533 -57.874 50.359
577998 98.563 -57.871 50.327
577999 98.592 -57.868 50.296
578000 98.622 -57.866 50.266
578001 98.651 -57.864 50.237
578002 98.681 -57.861 50.211
578003 98.710 -57.858 50.189
578004 98.740 -57.856 50.168
578005 98.752 -57.860 50.147
578006 98.740 -57.873 50.126
578007 98.752 -57.881 50.105
578008 98.775 -57.887 50.082
578009 98.798 -57.894 50.058
578010 98.821 -57.900 50.034
578011 98.845 -57.906 50.009
578012 98.868 -57.913 49.984
578013 98.891 -57.919 49.959
578014 98.914 -57.926 49.933
578015 98.951 -57.923 49.906
578016 98.991 -57.918 49.880
578017 99.031 -57.913 49.854
578018 99.071 -57.909 49.826
578019 99.111 -57.904 49.800
578020 99.150 -57.899 49.771
578021 99.190 -57.895 49.742
578022 99.230 -57.890 49.710
578023 99.268 -57.894 49.678
578024 99.305 -57.904 49.646
578025 99.342 -57.915 49.613
578026 99.371 -57.926 49.580
578027 99.389 -57.940 49.548
578028 99.452 -57.951 49.515
578029 99.504 -57.958 49.482
578030 99.543 -57.966 49.450
578031 99.577 -57.974 49.417
578032 99.612 -57.984 49.385
578033 99.646 -57.992 49.353
578034 99.681 -58.001 49.321
578035 99.715 -58.010 49.289
578036 99.750 -58.018 49.257
578037 99.784 -58.028 49.226
578038 99.819 -58.036 49.195
578039 99.853 -58.045 49.165
578387 87.904 -54.317 57.588
578388 87.940 -54.332 57.591
578389 87.975 -54.347 57.592
578390 88.010 -54.362 57.593
578391 88.046 -54.378 57.594
578392 88.081 -54.393 57.595
578393 88.117 -54.408 57.595
578394 88.152 -54.424 57.595
578395 88.187 -54.439 57.593
578396 88.222 -54.455 57.590
578397 88.258 -54.470 57.587
578398 88.293 -54.485 57.583
578399 88.329 -54.500 57.580
578400 88.364 -54.515 57.575
578401 88.423 -54.460 57.571
578402 88.488 -54.399 57.566
578403 88.530 -54.393 57.560
578404 88.565 -54.421 57.553
578405 88.600 -54.457 57.542
578406 88.637 -54.477 57.523
578407 88.675 -54.494 57.499
578408 88.713 -54.511 57.464
578409 88.752 -54.528 57.427
578410 88.790 -54.545 57.389
578411 88.828 -54.562 57.350
578412 88.866 -54.580 57.308
578413 88.905 -54.603 57.262
578414 88.944 -54.627 57.214
578415 88.982 -54.650 57.165
578416 89.021 -54.674 57.116
578417 89.059 -54.696 57.067
578418 89.098 -54.720 57.018
578419 89.136 -54.743 56.969
578420 89.175 -54.767 56.921
578449 90.214 -55.397 55.794
578450 90.237 -55.411 55.786
578451 90.260 -55.426 55.779
578452 90.284 -55.439 55.775
578453 90.307 -55.454 55.775
578454 90.330 -55.467 55.776
578455 90.356 -55.485 55.780
578456 90.383 -55.502 55.786
578457 90.408 -55.513 55.796
578458 90.434 -55.523 55.809
578459 90.459 -55.533 55.827
578460 90.484 -55.543 55.848
578461 90.522 -55.553 55.870
578462 90.561 -55.564 55.892
578463 90.601 -55.574 55.918
578464 90.641 -55.584 55.947
578465 90.680 -55.594 55.981
578466 90.720 -55.604 56.012
578467 90.759 -55.614 56.041
578468 90.799 -55.624 56.073
578469 90.839 -55.635 56.114
578470 90.879 -55.645 56.151
578471 90.918 -55.655 56.181
578472 90.958 -55.665 56.197
578473 90.998 -55.675 56.192
578474 91.039 -55.686 56.164
578475 91.085 -55.696 56.126
578476 91.130 -55.708 56.117
578477 91.175 -55.718 56.105
578478 91.221 -55.730 56.085
578479 91.266 -55.740 56.058
578480 91.309 -55.751 56.031
578481 91.345 -55.758 56.005
578482 91.380 -55.765 55.958
578483 91.416 -55.771 55.925
578484 91.452 -55.779 55.895
578485 91.488 -55.786 55.859
578486 91.523 -55.793 55.818
578487 91.557 -55.800 55.765
578488 91.588 -55.810 55.707
578489 91.618 -55.820 55.653
578490 91.649 -55.829 55.604
578491 91.679 -55.839 55.557
578492 91.710 -55.849 55.516
578493 91.740 -55.858 55.478
578494 91.771 -55.868 55.443
578495 91.801 -55.878 55.412
578496 91.832 -55.887 55.384
578497 91.862 -55.897 55.361
578498 91.898 -55.907 55.343
578499 91.935 -55.919 55.329
578500 91.967 -55.943 55.315
578501 92.000 -55.965 55.303
578502 92.032 -55.989 55.290
578503 92.065 -56.013 55.275
578504 92.097 -56.036 55.255
578505 92.129 -56.059 55.229
578506 92.162 -56.083 55.210
578507 92.194 -56.107 55.198
578508 92.227 -56.130 55.195
578509 92.259 -56.153 55.199
578510 92.291 -56.176 55.205
578511 92.324 -56.200 55.219
578512 92.356 -56.223 55.232
578513 92.391 -56.254 55.240
578514 92.428 -56.291 55.243
578515 92.467 -56.312 55.248
578516 92.506 -56.331 55.264
578517 92.546 -56.350 55.285
578518 92.584 -56.370 55.306
578519 92.624 -56.389 55.323
578520 92.663 -56.407 55.337
578521 92.702 -56.426 55.346
578522 92.741 -56.446 55.351
578523 92.778 -56.466 55.355
578524 92.813 -56.486 55.359
578525 92.848 -56.507 55.368
578526 92.891 -56.520 55.382
578527 92.934 -56.534 55.395
578528 92.977 -56.548 55.404
578529 93.020 -56.561 55.416
578530 93.063 -56.575 55.427
578531 93.107 -56.589 55.435
578532 93.149 -56.602 55.438
578533 93.193 -56.616 55.435
578534 93.234 -56.629 55.422
578535 93.263 -56.635 55.409
578536 93.292 -56.642 55.395
578537 93.321 -56.649 55.379
578538 93.351 -56.657 55.362
578539 93.384 -56.669 55.341
578540 93.417 -56.681 55.311
578541 93.449 -56.693 55.277
578542 93.482 -56.705 55.240
578543 93.515 -56.717 55.202
578544 93.547 -56.728 55.161
578545 93.580 -56.740 55.123
578546 93.613 -56.752 55.082
578547 93.645 -56.764 55.043
578548 93.678 -56.776 55.006
578549 93.711 -56.788 54.970
578550 93.744 -56.800 54.936
578551 93.776 -56.812 54.903
578552 93.809 -56.824 54.864
578553 93.841 -56.836 54.825
578554 93.874 -56.847 54.789
578555 93.907 -56.859 54.759
578556 93.939 -56.871 54.735
578557 93.972 -56.883 54.713
578558 94.005 -56.895 54.688
578559 94.038 -56.907 54.663
578560 94.070 -56.918 54.638
578561 94.103 -56.931 54.613
578562 94.135 -56.943 54.588
578563 94.169 -56.948 54.561
578564 94.203 -56.952 54.529
578565 94.239 -56.969 54.494
578566 94.273 -56.987 54.457
578567 94.306 -56.999 54.417
578568 94.338 -57.012 54.378
578569 94.370 -57.024 54.340
578570 94.403 -57.036 54.299
578571 94.435 -57.049 54.255
578572 94.468 -57.061 54.207
578573 94.500 -57.073 54.156
578574 94.532 -57.086 54.100
578575 94.564 -57.098 54.045
578576 94.597 -57.111 53.992
578577 94.629 -57.123 53.948
578578 94.662 -57.135 53.914
578579 94.694 -57.148 53.890
578580 94.726 -57.160 53.870
578581 94.758 -57.173 53.850
578582 94.791 -57.185 53.831
578583 94.823 -57.197 53.813
578584 94.855 -57.210 53.796
578585 94.888 -57.222 53.779
578586 94.920 -57.235 53.760
578587 94.953 -57.247 53.744
578588 94.985 -57.260 53.731
578589 95.017 -57.272 53.718
578590 95.049 -57.285 53.706
578591 95.082 -57.297 53.691
578592 95.114 -57.310 53.672
578593 95.146 -57.322 53.648
578594 95.179 -57.334 53.619
578595 95.209 -57.344 53.584
578596 95.239 -57.354 53.543
578597 95.269 -57.365 53.499
578598 95.303 -57.375 53.451
578599 95.339 -57.385 53.401
578600 95.375 -57.395 53.352
578601 95.411 -57.405 53.304
578602 95.448 -57.415 53.257
578603 95.484 -57.426 53.215
578604 95.521 -57.435 53.171
578605 95.557 -57.445 53.130
578606 95.594 -57.456 53.087
578607 95.630 -57.466 53.042
578608 95.673 -57.471 52.998
578609 95.716 -57.476 52.951
578610 95.760 -57.481 52.898
578611 95.804 -57.485 52.849
578612 95.828 -57.491 52.802
578613 95.849 -57.496 52.757
578614 95.871 -57.502 52.713
578615 95.891 -57.507 52.670
578616 95.912 -57.513 52.629
578617 95.933 -57.518 52.589
578618 95.954 -57.524 52.551
578619 95.975 -57.529 52.516
578620 95.996 -57.535 52.483
578621 96.017 -57.541 52.450
578622 96.044 -57.548 52.420
578623 96.073 -57.555 52.393
578624 96.102 -57.563 52.368
578625 96.131 -57.570 52.344
578626 96.160 -57.578 52.322
578627 96.189 -57.586 52.303
578628 96.219 -57.593 52.287
578629 96.247 -57.601 52.273
578630 96.276 -57.608 52.263
578631 96.306 -57.616 52.254
578632 96.334 -57.623 52.246
578633 96.364 -57.631 52.240
578634 96.393 -57.639 52.236
578635 96.426 -57.648 52.233
578636 96.460 -57.659 52.227
578637 96.493 -57.670 52.213
578638 96.528 -57.680 52.194
578639 96.561 -57.690 52.170
578640 96.595 -57.701 52.150
578641 96.629 -57.711 52.135
578642 96.658 -57.722 52.123
578643 96.686 -57.733 52.114
578644 96.715 -57.743 52.108
578645 96.743 -57.755 52.099
578646 96.777 -57.761 52.087
578647 96.814 -57.766 52.075
578648 96.850 -57.771 52.067
578649 96.887 -57.776 52.063
578650 96.923 -57.781 52.064
578651 96.960 -57.786 52.072
578652 96.996 -57.791 52.083
578653 97.033 -57.796 52.096
578654 97.065 -57.796 52.110
578655 97.093 -57.793 52.126
578656 97.121 -57.789 52.137
578657 97.150 -57.786 52.143
578658 97.178 -57.783 52.145
578659 97.207 -57.780 52.143
578660 97.235 -57.776 52.137
578661 97.264 -57.773 52.122
578662 97.292 -57.770 52.099
578663 97.321 -57.766 52.073
578664 97.349 -57.763 52.047
578665 97.376 -57.760 52.020
578666 97.403 -57.756 51.992
578667 97.431 -57.753 51.961
578668 97.463 -57.763 51.920
578669 97.498 -57.777 51.868
578670 97.532 -57.792 51.811
578671 97.567 -57.806 51.750
578672 97.599 -57.817 51.686
578673 97.629 -57.819 51.617
578674 97.659 -57.821 51.553
578675 97.689 -57.822 51.495
578676 97.720 -57.822 51.441
578677 97.752 -57.822 51.386
578678 97.784 -57.822 51.329
578679 97.815 -57.822 51.269
578680 97.847 -57.823 51.210
578681 97.878 -57.823 51.151
578682 97.910 -57.823 51.095
578683 97.942 -57.823 51.041
578684 97.979 -57.827 50.988
578685 98.016 -57.832 50.933
578686 98.056 -57.836 50.879
578687 98.096 -57.838 50.827
578688 98.136 -57.841 50.776
578689 98.176 -57.843 50.728
578690 98.216 -57.846 50.685
578691 98.261 -57.849 50.642
578692 98.306 -57.852 50.602
578693 98.351 -57.855 50.563
578694 98.396 -57.858 50.524
578695 98.441 -57.860 50.486
578696 98.473 -57.855 50.448
578697 98.503 -57.850 50.412
578698 98.533 -57.844 50.379
578699 98.563 -57.838 50.346
578700 98.593 -57.833 50.315
578701 98.624 -57.827 50.285
578702 98.653 -57.824 50.256
578703 98.683 -57.822 50.230
578704 98.712 -57.820 50.207
578705 98.742 -57.817 50.185
578706 98.772 -57.814 50.163
578707 98.767 -57.824 50.142
578708 98.776 -57.833 50.119
578709 98.799 -57.840 50.094
578710 98.822 -57.846 50.069
578711 98.845 -57.852 50.043
578712 98.868 -57.859 50.016
578713 98.891 -57.865 49.989
578714 98.915 -57.872 49.963
578715 98.938 -57.878 49.936
578716 98.961 -57.884 49.909
578717 98.984 -57.891 49.882
578718 99.020 -57.889 49.855
578719 99.060 -57.884 49.827
578720 99.100 -57.880 49.799
578721 99.140 -57.874 49.769
578722 99.180 -57.870 49.740
578723 99.219 -57.865 49.709
578724 99.258 -57.867 49.677
578725 99.295 -57.877 49.645
578726 99.332 -57.887 49.612
578727 99.369 -57.897 49.579
578728 99.382 -57.911 49.545
578729 99.438 -57.924 49.513
578730 99.490 -57.931 49.480
578731 99.535 -57.938 49.446
578732 99.570 -57.947 49.412
578733 99.604 -57.955 49.381
578734 99.639 -57.965 49.349
578735 99.673 -57.973 49.317
578736 99.707 -57.982 49.285
578737 99.742 -57.990 49.252
578738 99.776 -57.999 49.221
578739 99.811 -58.009 49.190
578740 99.845 -58.017 49.160
579088 87.929 -54.282 57.586
579089 87.965 -54.297 57.589
579090 88.000 -54.312 57.590
579091 88.035 -54.327 57.591
579092 88.070 -54.342 57.592
579093 88.106 -54.358 57.593
579094 88.141 -54.373 57.593
579095 88.177 -54.389 57.594
579096 88.212 -54.404 57.592
579097 88.247 -54.419 57.589
579098 88.283 -54.435 57.587
579099 88.318 -54.450 57.584
579100 88.353 -54.465 57.580
579101 88.389 -54.480 57.577
579102 88.438 -54.453 57.574
579103 88.506 -54.372 57.570
579104 88.552 -54.361 57.564
579105 88.587 -54.396 57.557
579106 88.622 -54.431 57.545
579107 88.659 -54.454 57.525
579108 88.697 -54.471 57.500
579109 88.736 -54.488 57.468
579110 88.773 -54.505 57.433
579111 88.812 -54.521 57.395
579112 88.850 -54.538 57.357
579113 88.888 -54.555 57.316
579114 88.926 -54.572 57.271
579115 88.965 -54.595 57.224
579116 89.003 -54.618 57.175
579117 89.042 -54.641 57.125
579118 89.081 -54.665 57.076
579119 89.119 -54.688 57.028
579120 89.158 -54.712 56.979
579121 89.196 -54.735 56.931
579149 90.210 -55.350 55.813
579150 90.233 -55.364 55.802
579151 90.256 -55.378 55.794
579152 90.279 -55.392 55.787
579153 90.302 -55.406 55.784
579154 90.326 -55.420 55.782
579155 90.349 -55.435 55.783
579156 90.373 -55.450 55.787
579157 90.400 -55.469 55.794
579158 90.426 -55.479 55.806
579159 90.451 -55.489 55.820
579160 90.477 -55.499 55.842
579161 90.502 -55.510 55.866
579162 90.534 -55.520 55.888
579163 90.573 -55.530 55.905
579164 90.613 -55.540 55.926
579165 90.653 -55.550 55.949
579166 90.692 -55.561 55.981
579167 90.732 -55.571 56.012
579168 90.771 -55.581 56.039
579169 90.811 -55.591 56.066
579170 90.851 -55.601 56.098
579171 90.890 -55.611 56.132
579172 90.930 -55.621 56.161
579173 90.970 -55.632 56.175
579174 91.015 -55.643 56.170
579175 91.061 -55.654 56.147
579176 91.106 -55.665 56.113
579177 91.151 -55.676 56.103
579178 91.196 -55.687 56.094
579179 91.242 -55.698 56.072
579180 91.287 -55.709 56.037
579181 91.332 -55.720 56.007
579182 91.368 -55.727 55.985
579183 91.404 -55.734 55.947
579184 91.440 -55.740 55.912
579185 91.475 -55.748 55.879
579186 91.511 -55.755 55.841
579187 91.545 -55.762 55.798
579188 91.576 -55.772 55.748
579189 91.606 -55.781 55.695
579190 91.637 -55.791 55.643
579191 91.668 -55.801 55.594
579192 91.698 -55.810 55.547
579193 91.729 -55.820 55.506
579194 91.759 -55.830 55.468
579195 91.790 -55.839 55.431
579196 91.820 -55.849 55.398
579197 91.851 -55.858 55.368
579198 91.881 -55.868 55.342
579199 91.912 -55.878 55.322
579200 91.944 -55.899 55.304
579201 91.977 -55.922 55.287
579202 92.009 -55.946 55.271
579203 92.041 -55.969 55.255
579204 92.074 -55.993 55.239
579205 92.106 -56.016 55.222
579206 92.139 -56.039 55.205
579207 92.171 -56.063 55.191
579208 92.203 -56.086 55.182
579209 92.236 -56.110 55.177
579210 92.268 -56.133 55.178
579211 92.301 -56.157 55.182
579212 92.333 -56.180 55.192
579213 92.365 -56.204 55.205
579214 92.398 -56.226 55.217
579215 92.433 -56.257 55.227
579216 92.473 -56.276 55.243
579217 92.512 -56.295 55.269
579218 92.551 -56.314 55.290
579219 92.590 -56.333 55.308
579220 92.629 -56.352 55.323
579221 92.668 -56.371 55.335
579222 92.707 -56.390 55.343
579223 92.746 -56.410 55.348
579224 92.784 -56.429 55.352
579225 92.819 -56.450 55.356
579226 92.856 -56.469 55.362
579227 92.899 -56.483 55.372
579228 92.942 -56.496 55.383
579229 92.985 -56.510 55.391
579230 93.028 -56.523 55.402
579231 93.072 -56.536 55.414
579232 93.115 -56.550 55.424
579233 93.158 -56.564 55.429
579234 93.201 -56.577 55.427
579235 93.244 -56.591 55.415
579236 93.274 -56.599 55.404
579237 93.303 -56.605 55.395
579238 93.332 -56.611 55.385
579239 93.363 -56.620 55.369
579240 93.396 -56.633 55.354
579241 93.428 -56.645 55.327
579242 93.461 -56.656 55.295
579243 93.493 -56.668 55.260
579244 93.526 -56.680 55.222
579245 93.559 -56.692 55.182
579246 93.591 -56.704 55.142
579247 93.624 -56.716 55.099
579248 93.657 -56.727 55.058
579249 93.690 -56.740 55.018
579250 93.722 -56.752 54.977
579251 93.755 -56.763 54.939
579252 93.787 -56.775 54.904
579253 93.820 -56.787 54.865
579254 93.853 -56.799 54.825
579255 93.886 -56.811 54.786
579256 93.918 -56.823 54.751
579257 93.951 -56.834 54.722
579258 93.984 -56.846 54.700
579259 94.016 -56.859 54.674
579260 94.049 -56.870 54.648
579261 94.081 -56.882 54.623
579262 94.114 -56.894 54.597
579263 94.147 -56.906 54.573
579264 94.180 -56.918 54.547
579265 94.213 -56.924 54.516
579266 94.249 -56.940 54.482
579267 94.281 -56.952 54.446
579268 94.314 -56.965 54.408
579269 94.346 -56.977 54.368
579270 94.378 -56.989 54.328
579271 94.410 -57.002 54.286
579272 94.443 -57.014 54.242
579273 94.475 -57.027 54.197
579274 94.507 -57.039 54.147
579275 94.540 -57.051 54.094
579276 94.572 -57.064 54.042
579277 94.605 -57.076 53.989
579278 94.637 -57.089 53.938
579279 94.669 -57.101 53.896
579280 94.701 -57.113 53.867
579281 94.734 -57.126 53.848
579282 94.766 -57.138 53.826
579283 94.799 -57.150 53.807
579284 94.831 -57.163 53.787
579285 94.863 -57.175 53.767
579286 94.896 -57.188 53.748
579287 94.928 -57.200 53.727
579288 94.960 -57.213 53.708
579289 94.992 -57.225 53.692
579290 95.025 -57.238 53.677
579291 95.057 -57.250 53.662
579292 95.090 -57.263 53.647
579293 95.122 -57.275 53.628
579294 95.154 -57.288 53.606
579295 95.186 -57.300 53.580
579296 95.218 -57.312 53.549
579297 95.248 -57.322 53.511
579298 95.282 -57.332 53.469
579299 95.319 -57.342 53.422
579300 95.355 -57.352 53.374
579301 95.391 -57.362 53.328
579302 95.428 -57.373 53.282
579303 95.464 -57.382 53.237
579304 95.500 -57.392 53.194
579305 95.537 -57.403 53.149
579306 95.573 -57.413 53.105
579307 95.610 -57.423 53.063
579308 95.646 -57.433 53.022
579309 95.683 -57.442 52.982
579310 95.727 -57.447 52.940
579311 95.771 -57.452 52.894
579312 95.813 -57.457 52.849
579313 95.834 -57.462 52.803
579314 95.855 -57.467 52.756
579315 95.876 -57.473 52.711
579316 95.897 -57.479 52.669
579317 95.917 -57.484 52.628
579318 95.939 -57.490 52.587
579319 95.960 -57.495 52.547
579320 95.980 -57.501 52.510
579321 96.002 -57.507 52.475
579322 96.022 -57.512 52.440
579323 96.046 -57.518 52.408
579324 96.076 -57.526 52.381
579325 96.105 -57.533 52.355
579326 96.133 -57.541 52.330
579327 96.163 -57.548 52.307
579328 96.192 -57.556 52.285
579329 96.221 -57.564 52.266
579330 96.250 -57.571 52.250
579331 96.279 -57.579 52.236
579332 96.308 -57.586 52.224
579333 96.337 -57.594 52.216
579334 96.366 -57.601 52.209
579335 96.395 -57.609 52.204
579336 96.424 -57.617 52.201
579337 96.456 -57.626 52.199
579338 96.490 -57.637 52.191
579339 96.524 -57.647 52.174
579340 96.558 -57.658 52.150
579341 96.592 -57.668 52.128
579342 96.625 -57.678 52.111
579343 96.654 -57.689 52.100
579344 96.682 -57.700 52.092
579345 96.710 -57.711 52.085
579346 96.745 -57.717 52.075
579347 96.782 -57.722 52.063
579348 96.818 -57.727 52.051
579349 96.855 -57.732 52.043
579350 96.891 -57.737 52.041
579351 96.928 -57.742 52.045
579352 96.964 -57.747 52.054
579353 97.001 -57.752 52.066
579354 97.037 -57.757 52.078
579355 97.070 -57.758 52.091
579356 97.099 -57.755 52.104
579357 97.127 -57.752 52.115
579358 97.156 -57.749 52.120
579359 97.184 -57.745 52.121
579360 97.213 -57.742 52.118
579361 97.241 -57.739 52.112
579362 97.270 -57.735 52.100
579363 97.297 -57.732 52.076
579364 97.324 -57.729 52.051
579365 97.352 -57.725 52.026
579366 97.379 -57.722 52.001
579367 97.406 -57.719 51.976
579368 97.433 -57.716 51.948
579369 97.465 -57.724 51.908
579370 97.500 -57.739 51.860
579371 97.534 -57.753 51.807
579372 97.569 -57.767 51.749
579373 97.601 -57.777 51.687
579374 97.631 -57.783 51.620
579375 97.661 -57.785 51.557
579376 97.691 -57.786 51.496
579377 97.721 -57.789 51.437
579378 97.751 -57.790 51.381
579379 97.781 -57.792 51.326
579380 97.810 -57.793 51.269
579381 97.842 -57.794 51.212
579382 97.873 -57.794 51.155
579383 97.904 -57.794 51.102
579384 97.936 -57.794 51.050
579385 97.968 -57.794 50.999
579386 98.004 -57.798 50.946
579387 98.042 -57.802 50.893
579388 98.082 -57.805 50.842
579389 98.122 -57.807 50.792
579390 98.165 -57.810 50.744
579391 98.210 -57.813 50.699
579392 98.255 -57.815 50.654
579393 98.300 -57.818 50.613
579394 98.345 -57.821 50.574
579395 98.390 -57.824 50.536
579396 98.436 -57.827 50.500
579397 98.474 -57.826 50.466
579398 98.504 -57.820 50.433
579399 98.534 -57.814 50.400
579400 98.564 -57.808 50.369
579401 98.594 -57.802 50.337
579402 98.624 -57.797 50.307
579403 98.654 -57.791 50.277
579404 98.684 -57.786 50.249
579405 98.714 -57.780 50.223
579406 98.744 -57.778 50.199
579407 98.773 -57.776 50.177
579408 98.793 -57.777 50.155
579409 98.799 -57.786 50.131
579410 98.822 -57.793 50.106
579411 98.846 -57.799 50.079
579412 98.869 -57.805 50.051
579413 98.892 -57.811 50.023
579414 98.916 -57.818 49.995
579415 98.939 -57.824 49.968
579416 98.962 -57.830 49.940
579417 98.985 -57.837 49.912
579418 99.008 -57.843 49.884
579419 99.031 -57.850 49.857
579420 99.055 -57.856 49.829
579421 99.089 -57.855 49.800
579422 99.129 -57.850 49.770
579423 99.169 -57.846 49.740
579424 99.209 -57.841 49.709
579425 99.248 -57.840 49.677
579426 99.285 -57.849 49.644
579427 99.322 -57.860 49.611
579428 99.359 -57.870 49.577
579429 99.391 -57.880 49.544
579430 99.417 -57.895 49.511
579431 99.475 -57.904 49.478
579432 99.526 -57.911 49.444
579433 99.562 -57.919 49.410
579434 99.596 -57.928 49.377
579435 99.631 -57.937 49.345
579436 99.665 -57.946 49.314
579437 99.700 -57.954 49.281
579438 99.734 -57.963 49.248
579439 99.769 -57.971 49.216
579440 99.803 -57.981 49.185
579441 99.838 -57.990 49.155
579442 99.872 -57.998 49.125
579789 87.954 -54.246 57.583
579790 87.989 -54.262 57.587
579791 88.025 -54.277 57.587
579792 88.060 -54.292 57.588
579793 88.095 -54.308 57.589
579794 88.130 -54.323 57.591
579795 88.166 -54.339 57.592
579796 88.201 -54.354 57.592
579797 88.237 -54.369 57.590
579798 88.272 -54.384 57.588
579799 88.308 -54.400 57.586
579800 88.343 -54.415 57.584
579801 88.378 -54.430 57.581
579802 88.413 -54.446 57.578
579803 88.454 -54.446 57.576
579804 88.522 -54.366 57.573
579805 88.574 -54.336 57.567
579806 88.609 -54.371 57.556
579807 88.644 -54.406 57.544
579808 88.681 -54.431 57.525
579809 88.719 -54.448 57.502
579810 88.757 -54.464 57.473
579811 88.796 -54.481 57.439
579812 88.834 -54.498 57.402
579813 88.872 -54.515 57.364
579814 88.910 -54.532 57.324
579815 88.948 -54.549 57.279
579816 88.986 -54.565 57.232
579817 89.025 -54.587 57.184
579818 89.063 -54.609 57.135
579819 89.102 -54.633 57.085
579820 89.140 -54.656 57.037
579821 89.179 -54.680 56.989
579822 89.217 -54.703 56.941
579849 90.205 -55.302 55.836
579850 90.229 -55.317 55.822
579851 90.252 -55.331 55.810
579852 90.275 -55.345 55.802
579853 90.298 -55.359 55.795
579854 90.321 -55.373 55.792
579855 90.344 -55.387 55.790
579856 90.368 -55.401 55.790
579857 90.391 -55.415 55.794
579858 90.418 -55.434 55.802
579859 90.444 -55.445 55.814
579860 90.469 -55.456 55.832
579861 90.495 -55.466 55.858
579862 90.520 -55.477 55.883
579863 90.546 -55.486 55.902
579864 90.585 -55.497 55.917
579865 90.625 -55.507 55.934
579866 90.664 -55.517 55.952
579867 90.704 -55.527 55.978
579868 90.744 -55.537 56.007
579869 90.783 -55.547 56.032
579870 90.823 -55.558 56.056
579871 90.863 -55.568 56.081
579872 90.902 -55.578 56.112
579873 90.946 -55.589 56.136
579874 90.991 -55.599 56.146
579875 91.036 -55.611 56.139
579876 91.082 -55.621 56.121
579877 91.127 -55.633 56.093
579878 91.172 -55.643 56.085
579879 91.218 -55.655 56.078
579880 91.263 -55.665 56.057
579881 91.308 -55.677 56.018
579882 91.353 -55.687 55.982
579883 91.392 -55.696 55.957
579884 91.428 -55.702 55.923
579885 91.463 -55.709 55.889
579886 91.499 -55.717 55.854
579887 91.534 -55.724 55.815
579888 91.564 -55.734 55.775
579889 91.595 -55.743 55.731
579890 91.626 -55.753 55.681
579891 91.656 -55.762 55.630
579892 91.686 -55.772 55.582
579893 91.717 -55.781 55.536
579894 91.747 -55.791 55.495
579895 91.778 -55.801 55.458
579896 91.808 -55.810 55.421
579897 91.839 -55.821 55.385
579898 91.869 -55.833 55.354
579899 91.899 -55.844 55.325
579900 91.930 -55.856 55.303
579901 91.961 -55.875 55.283
579902 91.993 -55.900 55.264
579903 92.025 -55.923 55.246
579904 92.057 -55.947 55.229
579905 92.089 -55.970 55.215
579906 92.122 -55.994 55.201
579907 92.154 -56.017 55.189
579908 92.186 -56.041 55.179
579909 92.218 -56.064 55.172
579910 92.250 -56.088 55.169
579911 92.283 -56.111 55.169
579912 92.315 -56.135 55.175
579913 92.347 -56.158 55.183
579914 92.379 -56.182 55.191
579915 92.412 -56.205 55.201
579916 92.444 -56.226 55.215
579917 92.479 -56.240 55.239
579918 92.517 -56.258 55.266
579919 92.556 -56.278 55.287
579920 92.595 -56.297 55.304
579921 92.634 -56.316 55.319
579922 92.673 -56.335 55.328
579923 92.712 -56.354 55.333
579924 92.752 -56.373 55.338
579925 92.791 -56.392 55.342
579926 92.825 -56.414 55.346
579927 92.864 -56.430 55.351
579928 92.908 -56.444 55.359
579929 92.950 -56.458 55.367
579930 92.994 -56.471 55.376
579931 93.037 -56.485 55.388
579932 93.080 -56.498 55.400
579933 93.123 -56.512 55.410
579934 93.166 -56.526 55.415
579935 93.209 -56.539 55.414
579936 93.252 -56.553 55.404
579937 93.285 -56.561 55.395
579938 93.314 -56.567 55.388
579939 93.343 -56.574 55.380
579940 93.374 -56.584 55.367
579941 93.407 -56.596 55.353
579942 93.439 -56.608 55.330
579943 93.472 -56.620 55.303
579944 93.505 -56.632 55.271
579945 93.538 -56.643 55.235
579946 93.570 -56.655 55.199
579947 93.603 -56.668 55.161
579948 93.636 -56.679 55.119
579949 93.669 -56.691 55.077
579950 93.701 -56.703 55.035
579951 93.733 -56.714 54.992
579952 93.766 -56.727 54.949
579953 93.799 -56.739 54.908
579954 93.832 -56.750 54.869
579955 93.864 -56.762 54.829
579956 93.897 -56.774 54.788
579957 93.930 -56.787 54.747
579958 93.963 -56.798 54.712
579959 93.995 -56.810 54.687
579960 94.028 -56.822 54.661
579961 94.060 -56.834 54.635
579962 94.093 -56.846 54.608
579963 94.126 -56.858 54.581
579964 94.158 -56.869 54.555
579965 94.191 -56.881 54.530
579966 94.224 -56.893 54.500
579967 94.257 -56.905 54.467
579968 94.289 -56.918 54.431
579969 94.321 -56.930 54.394
579970 94.353 -56.943 54.356
579971 94.386 -56.955 54.316
579972 94.418 -56.967 54.273
579973 94.451 -56.980 54.229
579974 94.483 -56.992 54.183
579975 94.515 -57.005 54.133
579976 94.548 -57.017 54.082
579977 94.580 -57.029 54.032
579978 94.612 -57.042 53.981
579979 94.644 -57.054 53.930
579980 94.677 -57.066 53.882
579981 94.709 -57.079 53.847
579982 94.742 -57.091 53.829
579983 94.774 -57.104 53.809
579984 94.806 -57.116 53.787
579985 94.838 -57.129 53.764
579986 94.871 -57.141 53.742
579987 94.903 -57.154 53.720
579988 94.935 -57.166 53.699
579989 94.967 -57.179 53.681
579990 95.000 -57.191 53.661
579991 95.032 -57.204 53.640
579992 95.064 -57.216 53.619
579993 95.097 -57.229 53.602
579994 95.129 -57.241 53.582
579995 95.161 -57.254 53.561
579996 95.193 -57.266 53.538
579997 95.226 -57.279 53.510
579998 95.262 -57.289 53.476
579999 95.299 -57.299 53.436
580000 95.335 -57.309 53.392
580001 95.372 -57.319 53.348
580002 95.408 -57.329 53.305
580003 95.444 -57.339 53.262
580004 95.480 -57.350 53.218
580005 95.517 -57.360 53.175
580006 95.553 -57.369 53.128
580007 95.590 -57.380 53.082
580008 95.626 -57.390 53.040
580009 95.662 -57.400 53.001
580010 95.699 -57.410 52.964
580011 95.738 -57.418 52.926
580012 95.781 -57.423 52.887
580013 95.818 -57.428 52.846
580014 95.839 -57.433 52.803
580015 95.860 -57.439 52.757
580016 95.881 -57.445 52.711
580017 95.902 -57.450 52.668
580018 95.923 -57.456 52.626
580019 95.944 -57.461 52.583
580020 95.965 -57.467 52.542
580021 95.986 -57.472 52.503
580022 96.007 -57.478 52.468
580023 96.028 -57.483 52.433
580024 96.049 -57.489 52.400
580025 96.078 -57.496 52.371
580026 96.107 -57.504 52.345
580027 96.136 -57.511 52.322
580028 96.165 -57.519 52.298
580029 96.194 -57.526 52.274
580030 96.223 -57.534 52.249
580031 96.252 -57.542 52.227
580032 96.281 -57.550 52.210
580033 96.310 -57.558 52.196
580034 96.339 -57.565 52.186
580035 96.368 -57.573 52.177
580036 96.397 -57.580 52.171
580037 96.427 -57.588 52.167
580038 96.455 -57.595 52.165
580039 96.487 -57.605 52.163
580040 96.521 -57.614 52.152
580041 96.555 -57.625 52.129
580042 96.589 -57.636 52.106
580043 96.620 -57.645 52.087
580044 96.649 -57.657 52.075
580045 96.677 -57.667 52.070
580046 96.713 -57.673 52.062
580047 96.749 -57.678 52.050
580048 96.786 -57.683 52.037
580049 96.822 -57.688 52.026
580050 96.859 -57.693 52.021
580051 96.895 -57.698 52.023
580052 96.932 -57.703 52.030
580053 96.968 -57.708 52.038
580054 97.005 -57.713 52.050
580055 97.041 -57.717 52.063
580056 97.076 -57.720 52.074
580057 97.104 -57.717 52.085
580058 97.133 -57.714 52.093
580059 97.161 -57.711 52.098
580060 97.189 -57.707 52.096
580061 97.218 -57.704 52.092
580062 97.245 -57.701 52.086
580063 97.272 -57.698 52.075
580064 97.300 -57.694 52.053
580065 97.327 -57.691 52.028
580066 97.354 -57.688 52.003
580067 97.381 -57.685 51.980
580068 97.409 -57.681 51.958
580069 97.436 -57.678 51.932
580070 97.468 -57.686 51.894
580071 97.502 -57.700 51.848
580072 97.536 -57.714 51.799
580073 97.571 -57.728 51.744
580074 97.602 -57.737 51.684
580075 97.633 -57.745 51.619
580076 97.663 -57.749 51.557
580077 97.693 -57.752 51.494
580078 97.723 -57.753 51.433
580079 97.753 -57.755 51.376
580080 97.783 -57.756 51.322
580081 97.813 -57.758 51.267
580082 97.843 -57.759 51.212
580083 97.873 -57.761 51.159
580084 97.902 -57.763 51.107
580085 97.932 -57.764 51.057
580086 97.962 -57.766 51.008
580087 97.994 -57.766 50.957
580088 98.029 -57.768 50.906
580089 98.069 -57.771 50.856
580090 98.114 -57.774 50.807
580091 98.160 -57.777 50.759
580092 98.205 -57.780 50.713
580093 98.250 -57.782 50.668
580094 98.295 -57.785 50.628
580095 98.340 -57.788 50.589
580096 98.385 -57.791 50.552
580097 98.430 -57.793 50.518
580098 98.475 -57.796 50.486
580099 98.504 -57.790 50.454
580100 98.534 -57.784 50.421
580101 98.565 -57.778 50.390
580102 98.594 -57.773 50.360
580103 98.625 -57.767 50.329
580104 98.655 -57.761 50.298
580105 98.685 -57.755 50.268
580106 98.715 -57.749 50.240
580107 98.745 -57.744 50.214
580108 98.775 -57.738 50.190
580109 98.805 -57.734 50.166
580110 98.823 -57.739 50.141
580111 98.847 -57.745 50.115
580112 98.870 -57.751 50.087
580113 98.893 -57.758 50.059
580114 98.916 -57.764 50.030
580115 98.939 -57.771 50.001
580116 98.962 -57.777 49.973
580117 98.986 -57.783 49.944
580118 99.009 -57.789 49.916
580119 99.032 -57.796 49.887
580120 99.055 -57.802 49.858
580121 99.078 -57.808 49.830
580122 99.101 -57.815 49.800
580123 99.125 -57.821 49.770
580124 99.158 -57.821 49.739
580125 99.198 -57.816 49.707
580126 99.238 -57.812 49.675
580127 99.275 -57.822 49.642
580128 99.312 -57.833 49.609
580129 99.349 -57.843 49.576
580130 99.385 -57.852 49.543
580131 99.402 -57.866 49.509
580132 99.460 -57.877 49.475
580133 99.512 -57.883 49.442
580134 99.554 -57.892 49.407
580135 99.589 -57.900 49.374
580136 99.623 -57.909 49.342
580137 99.658 -57.918 49.310
580138 99.692 -57.927 49.277
580139 99.727 -57.935 49.243
580140 99.761 -57.944 49.211
580141 99.796 -57.952 49.180
580142 99.830 -57.962 49.149
580143 99.865 -57.970 49.120
580144 99.899 -57.979 49.090
580490 87.978 -54.211 57.580
580491 88.014 -54.226 57.584
580492 88.049 -54.242 57.585
580493 88.084 -54.258 57.587
580494 88.120 -54.273 57.588
580495 88.155 -54.288 57.588
580496 88.191 -54.303 57.589
580497 88.226 -54.319 57.590
580498 88.261 -54.334 57.588
580499 88.296 -54.349 57.586
580500 88.332 -54.364 57.585
580501 88.367 -54.380 57.583
580502 88.403 -54.395 57.581
580503 88.438 -54.411 57.578
580504 88.473 -54.426 57.577
580505 88.537 -54.359 57.574
580506 88.597 -54.311 57.567
580507 88.631 -54.346 57.558
580508 88.667 -54.381 57.546
580509 88.703 -54.408 57.528
580510 88.741 -54.424 57.510
580511 88.779 -54.441 57.483
580512 88.817 -54.458 57.448
580513 88.855 -54.475 57.410
580514 88.894 -54.492 57.372
580515 88.932 -54.508 57.332
580516 88.970 -54.525 57.287
580517 89.008 -54.542 57.240
580518 89.046 -54.559 57.193
580519 89.085 -54.577 57.144
580520 89.123 -54.601 57.095
580521 89.162 -54.624 57.047
580522 89.200 -54.648 56.999
580523 89.239 -54.671 56.951
580549 90.201 -55.255 55.863
580550 90.224 -55.269 55.846
580551 90.247 -55.283 55.831
580552 90.271 -55.298 55.819
580553 90.294 -55.311 55.809
580554 90.317 -55.326 55.803
580555 90.340 -55.340 55.799
580556 90.364 -55.354 55.797
580557 90.387 -55.368 55.797
580558 90.410 -55.382 55.801
580559 90.436 -55.398 55.810
580560 90.462 -55.412 55.822
580561 90.487 -55.422 55.843
580562 90.513 -55.433 55.870
580563 90.538 -55.442 55.894
580564 90.564 -55.453 55.912
580565 90.597 -55.463 55.925
580566 90.637 -55.474 55.938
580567 90.676 -55.483 55.951
580568 90.716 -55.494 55.972
580569 90.755 -55.504 55.995
580570 90.795 -55.514 56.018
580571 90.835 -55.524 56.040
580572 90.877 -55.535 56.065
580573 90.922 -55.546 56.091
580574 90.967 -55.557 56.109
580575 91.012 -55.568 56.113
580576 91.058 -55.579 56.103
580577 91.103 -55.589 56.085
580578 91.148 -55.601 56.067
580579 91.193 -55.611 56.062
580580 91.239 -55.623 56.059
580581 91.284 -55.633 56.041
580582 91.329 -55.645 56.000
580583 91.374 -55.655 55.961
580584 91.416 -55.665 55.931
580585 91.451 -55.671 55.898
580586 91.487 -55.679 55.865
580587 91.522 -55.686 55.830
580588 91.553 -55.696 55.792
580589 91.583 -55.705 55.759
580590 91.613 -55.715 55.719
580591 91.644 -55.724 55.671
580592 91.675 -55.734 55.620
580593 91.705 -55.743 55.571
580594 91.736 -55.753 55.525
580595 91.766 -55.765 55.484
580596 91.796 -55.776 55.447
580597 91.826 -55.788 55.410
580598 91.857 -55.799 55.375
580599 91.887 -55.812 55.343
580600 91.918 -55.823 55.314
580601 91.948 -55.835 55.288
580602 91.980 -55.850 55.266
580603 92.011 -55.875 55.246
580604 92.042 -55.900 55.228
580605 92.074 -55.924 55.212
580606 92.106 -55.947 55.200
580607 92.138 -55.971 55.190
580608 92.170 -55.994 55.180
580609 92.203 -56.018 55.173
580610 92.235 -56.041 55.168
580611 92.267 -56.065 55.167
580612 92.299 -56.088 55.170
580613 92.332 -56.112 55.176
580614 92.364 -56.135 55.183
580615 92.396 -56.159 55.190
580616 92.428 -56.182 55.200
580617 92.461 -56.201 55.212
580618 92.496 -56.215 55.229
580619 92.530 -56.229 55.250
580620 92.564 -56.244 55.269
580621 92.600 -56.260 55.284
580622 92.640 -56.279 55.296
580623 92.678 -56.298 55.307
580624 92.718 -56.318 55.313
580625 92.757 -56.337 55.316
580626 92.796 -56.356 55.317
580627 92.832 -56.376 55.322
580628 92.872 -56.392 55.332
580629 92.916 -56.405 55.342
580630 92.959 -56.419 55.348
580631 93.002 -56.432 55.359
580632 93.045 -56.446 55.372
580633 93.088 -56.460 55.383
580634 93.131 -56.473 55.393
580635 93.174 -56.487 55.397
580636 93.217 -56.501 55.398
580637 93.260 -56.514 55.391
580638 93.296 -56.524 55.384
580639 93.325 -56.531 55.377
580640 93.354 -56.537 55.370
580641 93.386 -56.548 55.360
580642 93.418 -56.559 55.348
580643 93.451 -56.571 55.327
580644 93.484 -56.583 55.303
580645 93.516 -56.595 55.275
580646 93.549 -56.607 55.241
580647 93.582 -56.619 55.209
580648 93.614 -56.631 55.172
580649 93.647 -56.642 55.133
580650 93.680 -56.655 55.093
580651 93.712 -56.667 55.052
580652 93.745 -56.678 55.010
580653 93.778 -56.690 54.967
580654 93.811 -56.702 54.925
580655 93.843 -56.714 54.881
580656 93.876 -56.726 54.836
580657 93.908 -56.738 54.793
580658 93.941 -56.749 54.751
580659 93.974 -56.762 54.712
580660 94.006 -56.774 54.680
580661 94.039 -56.785 54.651
580662 94.072 -56.797 54.623
580663 94.105 -56.809 54.594
580664 94.137 -56.821 54.565
580665 94.170 -56.833 54.536
580666 94.202 -56.849 54.510
580667 94.233 -56.866 54.480
580668 94.263 -56.877 54.447
580669 94.294 -56.890 54.413
580670 94.324 -56.905 54.377
580671 94.353 -56.920 54.340
580672 94.385 -56.933 54.302
580673 94.417 -56.946 54.260
580674 94.450 -56.958 54.215
580675 94.482 -56.971 54.168
580676 94.514 -56.983 54.119
580677 94.546 -56.996 54.070
580678 94.578 -57.008 54.019
580679 94.611 -57.021 53.969
580680 94.643 -57.033 53.919
580681 94.675 -57.046 53.871
580682 94.707 -57.058 53.831
580683 94.740 -57.071 53.806
580684 94.772 -57.083 53.787
580685 94.805 -57.096 53.767
580686 94.837 -57.108 53.743
580687 94.869 -57.121 53.719
580688 94.901 -57.133 53.696
580689 94.933 -57.145 53.674
580690 94.966 -57.158 53.653
580691 94.998 -57.170 53.631
580692 95.030 -57.183 53.609
580693 95.063 -57.195 53.586
580694 95.095 -57.208 53.564
580695 95.127 -57.220 53.542
580696 95.160 -57.233 53.520
580697 95.192 -57.245 53.498
580698 95.226 -57.255 53.473
580699 95.265 -57.260 53.442
580700 95.303 -57.270 53.404
580701 95.341 -57.279 53.364
580702 95.380 -57.289 53.324
580703 95.418 -57.298 53.283
580704 95.457 -57.307 53.241
580705 95.495 -57.317 53.199
580706 95.533 -57.327 53.156
580707 95.569 -57.337 53.109
580708 95.606 -57.347 53.062
580709 95.642 -57.357 53.017
580710 95.679 -57.367 52.979
580711 95.715 -57.377 52.945
580712 95.751 -57.388 52.910
580713 95.793 -57.394 52.876
580714 95.824 -57.399 52.840
580715 95.844 -57.405 52.801
580716 95.866 -57.410 52.759
580717 95.887 -57.416 52.712
580718 95.907 -57.421 52.667
580719 95.928 -57.427 52.624
580720 95.949 -57.432 52.579
580721 95.970 -57.438 52.536
580722 95.991 -57.444 52.497
580723 96.012 -57.449 52.461
580724 96.033 -57.454 52.426
580725 96.054 -57.460 52.392
580726 96.080 -57.467 52.361
580727 96.109 -57.475 52.335
580728 96.138 -57.482 52.313
580729 96.167 -57.490 52.288
580730 96.196 -57.498 52.260
580731 96.225 -57.505 52.232
580732 96.254 -57.513 52.208
580733 96.284 -57.520 52.187
580734 96.313 -57.528 52.171
580735 96.341 -57.536 52.158
580736 96.371 -57.543 52.147
580737 96.400 -57.551 52.139
580738 96.429 -57.558 52.133
580739 96.458 -57.566 52.130
580740 96.487 -57.573 52.131
580741 96.518 -57.582 52.124
580742 96.551 -57.592 52.107
580743 96.585 -57.602 52.086
580744 96.616 -57.614 52.065
580745 96.644 -57.624 52.049
580746 96.681 -57.629 52.041
580747 96.717 -57.634 52.034
580748 96.754 -57.639 52.023
580749 96.790 -57.644 52.009
580750 96.827 -57.649 51.999
580751 96.863 -57.654 51.997
580752 96.900 -57.659 52.002
580753 96.936 -57.664 52.010
580754 96.973 -57.669 52.019
580755 97.009 -57.673 52.031
580756 97.046 -57.679 52.043
580757 97.081 -57.683 52.053
580758 97.109 -57.680 52.062
580759 97.138 -57.676 52.069
580760 97.166 -57.673 52.073
580761 97.193 -57.670 52.071
580762 97.220 -57.667 52.064
580763 97.247 -57.663 52.056
580764 97.275 -57.660 52.047
580765 97.302 -57.657 52.029
580766 97.329 -57.654 52.006
580767 97.357 -57.650 51.982
580768 97.384 -57.647 51.959
580769 97.412 -57.644 51.938
580770 97.439 -57.641 51.914
580771 97.470 -57.647 51.878
580772 97.505 -57.661 51.834
580773 97.539 -57.675 51.786
580774 97.573 -57.689 51.734
580775 97.604 -57.698 51.677
580776 97.635 -57.705 51.614
580777 97.665 -57.712 51.552
580778 97.696 -57.716 51.489
580779 97.725 -57.717 51.428
580780 97.755 -57.719 51.370
580781 97.785 -57.720 51.315
580782 97.815 -57.722 51.262
580783 97.845 -57.724 51.210
580784 97.875 -57.726 51.160
580785 97.904 -57.727 51.112
580786 97.935 -57.729 51.063
580787 97.964 -57.730 51.015
580788 97.994 -57.732 50.967
580789 98.025 -57.734 50.919
580790 98.064 -57.738 50.869
580791 98.109 -57.740 50.820
580792 98.154 -57.743 50.772
580793 98.199 -57.746 50.727
580794 98.244 -57.749 50.685
580795 98.290 -57.752 50.645
580796 98.335 -57.755 50.607
580797 98.380 -57.758 50.570
580798 98.425 -57.760 50.537
580799 98.470 -57.763 50.506
580800 98.505 -57.760 50.474
580801 98.535 -57.755 50.442
580802 98.565 -57.749 50.411
580803 98.595 -57.742 50.381
580804 98.625 -57.737 50.350
580805 98.655 -57.731 50.318
580806 98.686 -57.726 50.287
580807 98.715 -57.720 50.256
580808 98.745 -57.714 50.229
580809 98.776 -57.708 50.202
580810 98.805 -57.702 50.176
580811 98.838 -57.704 50.150
580812 98.864 -57.710 50.122
580813 98.887 -57.714 50.094
580814 98.912 -57.720 50.065
580815 98.935 -57.724 50.037
580816 98.959 -57.730 50.007
580817 98.983 -57.735 49.978
580818 99.007 -57.739 49.948
580819 99.031 -57.745 49.919
580820 99.055 -57.750 49.889
580821 99.079 -57.755 49.860
580822 99.102 -57.761 49.831
580823 99.126 -57.767 49.800
580824 99.149 -57.774 49.770
580825 99.172 -57.780 49.738
580826 99.195 -57.786 49.706
580827 99.227 -57.787 49.673
580828 99.265 -57.795 49.639
580829 99.302 -57.805 49.606
580830 99.339 -57.815 49.575
580831 99.375 -57.825 49.541
580832 99.410 -57.836 49.507
580833 99.445 -57.850 49.473
580834 99.497 -57.857 49.439
580835 99.547 -57.864 49.405
580836 99.581 -57.872 49.371
580837 99.616 -57.881 49.338
580838 99.650 -57.890 49.305
580839 99.685 -57.899 49.272
580840 99.719 -57.908 49.238
580841 99.754 -57.916 49.205
580842 99.788 -57.925 49.174
580843 99.823 -57.934 49.143
580844 99.857 -57.943 49.113
580845 99.892 -57.951 49.084
580846 99.926 -57.960 49.056
581192 88.039 -54.192 57.581
581193 88.074 -54.207 57.582
581194 88.109 -54.223 57.584
581195 88.144 -54.238 57.585
581196 88.180 -54.253 57.586
581197 88.215 -54.268 57.586
581198 88.251 -54.284 57.588
581199 88.286 -54.299 57.586
581200 88.321 -54.314 57.585
581201 88.356 -54.330 57.583
581202 88.392 -54.345 57.582
581203 88.427 -54.361 57.581
581204 88.463 -54.376 57.579
581205 88.498 -54.391 57.577
581206 88.552 -54.352 57.573
581207 88.616 -54.292 57.568
581208 88.654 -54.321 57.560
581209 88.689 -54.356 57.552
581210 88.725 -54.384 57.540
581211 88.763 -54.401 57.526
581212 88.801 -54.417 57.499
581213 88.839 -54.435 57.460
581214 88.878 -54.452 57.419
581215 88.915 -54.468 57.381
581216 88.954 -54.485 57.340
581217 88.992 -54.502 57.295
581218 89.030 -54.519 57.248
581219 89.068 -54.536 57.201
581220 89.106 -54.552 57.153
581221 89.144 -54.569 57.104
581222 89.183 -54.593 57.056
581223 89.222 -54.616 57.008
581224 89.260 -54.640 56.960
581225 89.299 -54.663 56.913
581249 90.197 -55.207 55.891
581250 90.220 -55.222 55.872
581251 90.243 -55.236 55.856
581252 90.266 -55.250 55.841
581253 90.290 -55.264 55.828
581254 90.313 -55.278 55.818
581255 90.336 -55.292 55.811
581256 90.360 -55.307 55.806
581257 90.383 -55.320 55.804
581258 90.406 -55.335 55.805
581259 90.429 -55.349 55.809
581260 90.453 -55.364 55.817
581261 90.479 -55.379 55.831
581262 90.505 -55.389 55.854
581263 90.530 -55.399 55.879
581264 90.556 -55.409 55.900
581265 90.581 -55.420 55.917
581266 90.609 -55.430 55.929
581267 90.648 -55.440 55.939
581268 90.688 -55.450 55.949
581269 90.728 -55.460 55.964
581270 90.767 -55.470 55.982
581271 90.807 -55.480 56.001
581272 90.853 -55.492 56.021
581273 90.898 -55.502 56.045
581274 90.943 -55.514 56.066
581275 90.988 -55.524 56.078
581276 91.033 -55.536 56.078
581277 91.079 -55.546 56.066
581278 91.124 -55.558 56.051
581279 91.169 -55.568 56.038
581280 91.215 -55.580 56.037
581281 91.260 -55.590 56.037
581282 91.305 -55.602 56.022
581283 91.350 -55.612 55.981
581284 91.395 -55.624 55.943
581285 91.439 -55.633 55.909
581286 91.475 -55.640 55.873
581287 91.510 -55.648 55.839
581288 91.541 -55.657 55.807
581289 91.572 -55.667 55.771
581290 91.602 -55.677 55.747
581291 91.633 -55.686 55.709
581292 91.663 -55.696 55.663
581293 91.693 -55.708 55.612
581294 91.723 -55.720 55.561
581295 91.754 -55.732 55.514
581296 91.784 -55.743 55.473
581297 91.814 -55.756 55.436
581298 91.845 -55.767 55.400
581299 91.875 -55.779 55.366
581300 91.905 -55.790 55.333
581301 91.935 -55.803 55.304
581302 91.966 -55.814 55.279
581303 91.999 -55.827 55.256
581304 92.029 -55.850 55.236
581305 92.060 -55.875 55.220
581306 92.091 -55.901 55.207
581307 92.123 -55.925 55.198
581308 92.155 -55.948 55.191
581309 92.188 -55.972 55.184
581310 92.220 -55.995 55.176
581311 92.252 -56.019 55.172
581312 92.284 -56.042 55.176
581313 92.316 -56.065 55.180
581314 92.348 -56.088 55.185
581315 92.380 -56.112 55.190
581316 92.413 -56.135 55.198
581317 92.445 -56.159 55.207
581318 92.479 -56.176 55.214
581319 92.513 -56.191 55.222
581320 92.547 -56.205 55.235
581321 92.581 -56.219 55.250
581322 92.616 -56.234 55.264
581323 92.650 -56.248 55.275
581324 92.685 -56.263 55.285
581325 92.723 -56.282 55.292
581326 92.762 -56.301 55.300
581327 92.801 -56.320 55.309
581328 92.838 -56.340 55.320
581329 92.881 -56.354 55.334
581330 92.924 -56.367 55.345
581331 92.967 -56.380 55.351
581332 93.010 -56.394 55.356
581333 93.053 -56.407 55.360
581334 93.096 -56.421 55.367
581335 93.139 -56.435 55.376
581336 93.183 -56.448 55.378
581337 93.225 -56.462 55.378
581338 93.269 -56.476 55.374
581339 93.307 -56.487 55.367
581340 93.336 -56.494 55.360
581341 93.365 -56.501 55.355
581342 93.397 -56.511 55.348
581343 93.430 -56.523 55.338
581344 93.463 -56.535 55.319
581345 93.495 -56.547 55.296
581346 93.528 -56.558 55.271
581347 93.560 -56.570 55.240
581348 93.593 -56.583 55.211
581349 93.626 -56.594 55.177
581350 93.658 -56.606 55.141
581351 93.691 -56.618 55.103
581352 93.724 -56.630 55.064
581353 93.757 -56.642 55.023
581354 93.789 -56.654 54.981
581355 93.822 -56.665 54.939
581356 93.854 -56.677 54.894
581357 93.887 -56.689 54.847
581358 93.920 -56.701 54.802
581359 93.953 -56.713 54.757
581360 93.985 -56.725 54.714
581361 94.018 -56.736 54.676
581362 94.051 -56.749 54.642
581363 94.083 -56.761 54.612
581364 94.116 -56.775 54.581
581365 94.147 -56.793 54.550
581366 94.180 -56.809 54.519
581367 94.211 -56.825 54.489
581368 94.241 -56.841 54.457
581369 94.269 -56.851 54.423
581370 94.304 -56.861 54.387
581371 94.335 -56.874 54.353
581372 94.364 -56.889 54.317
581373 94.394 -56.904 54.280
581374 94.423 -56.918 54.241
581375 94.452 -56.934 54.199
581376 94.482 -56.948 54.154
581377 94.512 -56.962 54.107
581378 94.545 -56.975 54.057
581379 94.577 -56.987 54.004
581380 94.609 -57.000 53.953
581381 94.641 -57.012 53.906
581382 94.674 -57.025 53.862
581383 94.706 -57.037 53.817
581384 94.738 -57.050 53.784
581385 94.770 -57.062 53.763
581386 94.803 -57.075 53.745
581387 94.835 -57.087 53.723
581388 94.867 -57.100 53.698
581389 94.900 -57.112 53.673
581390 94.932 -57.125 53.650
581391 94.964 -57.137 53.627
581392 94.996 -57.150 53.604
581393 95.029 -57.162 53.581
581394 95.061 -57.175 53.556
581395 95.093 -57.187 53.531
581396 95.126 -57.200 53.506
581397 95.158 -57.212 53.481
581398 95.190 -57.225 53.459
581399 95.229 -57.228 53.435
581400 95.267 -57.232 53.406
581401 95.306 -57.242 53.373
581402 95.344 -57.251 53.336
581403 95.382 -57.261 53.298
581404 95.421 -57.270 53.259
581405 95.459 -57.279 53.218
581406 95.498 -57.289 53.174
581407 95.536 -57.298 53.130
581408 95.574 -57.308 53.083
581409 95.613 -57.317 53.036
581410 95.652 -57.326 52.993
581411 95.690 -57.336 52.955
581412 95.729 -57.345 52.922
581413 95.767 -57.355 52.892
581414 95.804 -57.365 52.862
581415 95.829 -57.370 52.827
581416 95.850 -57.376 52.790
581417 95.871 -57.382 52.753
581418 95.892 -57.387 52.711
581419 95.913 -57.393 52.667
581420 95.934 -57.398 52.621
581421 95.955 -57.404 52.575
581422 95.976 -57.409 52.531
581423 95.996 -57.415 52.490
581424 96.018 -57.420 52.454
581425 96.038 -57.426 52.418
581426 96.059 -57.432 52.383
581427 96.083 -57.438 52.350
581428 96.111 -57.445 52.323
581429 96.141 -57.453 52.302
581430 96.170 -57.460 52.276
581431 96.198 -57.468 52.245
581432 96.228 -57.476 52.216
581433 96.257 -57.483 52.191
581434 96.286 -57.491 52.168
581435 96.315 -57.498 52.149
581436 96.344 -57.506 52.132
581437 96.373 -57.514 52.119
581438 96.402 -57.521 52.109
581439 96.431 -57.529 52.102
581440 96.460 -57.536 52.096
581441 96.489 -57.544 52.097
581442 96.518 -57.551 52.094
581443 96.548 -57.560 52.084
581444 96.582 -57.570 52.066
581445 96.612 -57.580 52.044
581446 96.649 -57.585 52.024
581447 96.685 -57.590 52.012
581448 96.722 -57.595 52.003
581449 96.758 -57.600 51.991
581450 96.795 -57.605 51.978
581451 96.831 -57.610 51.971
581452 96.868 -57.615 51.970
581453 96.904 -57.620 51.976
581454 96.941 -57.625 51.984
581455 96.977 -57.630 51.994
581456 97.014 -57.635 52.006
581457 97.050 -57.640 52.018
581458 97.086 -57.645 52.027
581459 97.114 -57.642 52.034
581460 97.141 -57.639 52.038
581461 97.168 -57.635 52.040
581462 97.195 -57.632 52.038
581463 97.223 -57.629 52.032
581464 97.250 -57.626 52.022
581465 97.278 -57.623 52.013
581466 97.305 -57.619 51.997
581467 97.332 -57.616 51.976
581468 97.360 -57.613 51.955
581469 97.387 -57.610 51.934
581470 97.415 -57.606 51.914
581471 97.442 -57.603 51.891
581472 97.472 -57.608 51.856
581473 97.507 -57.622 51.814
581474 97.541 -57.636 51.767
581475 97.576 -57.651 51.717
581476 97.606 -57.658 51.661
581477 97.637 -57.665 51.603
581478 97.667 -57.673 51.544
581479 97.698 -57.680 51.484
581480 97.728 -57.682 51.423
581481 97.757 -57.683 51.362
581482 97.788 -57.685 51.306
581483 97.817 -57.687 51.254
581484 97.847 -57.689 51.204
581485 97.877 -57.690 51.157
581486 97.907 -57.692 51.111
581487 97.937 -57.693 51.064
581488 97.967 -57.695 51.018
581489 97.997 -57.697 50.971
581490 98.029 -57.699 50.924
581491 98.065 -57.704 50.877
581492 98.104 -57.707 50.829
581493 98.149 -57.710 50.783
581494 98.194 -57.713 50.740
581495 98.239 -57.716 50.697
581496 98.284 -57.718 50.659
581497 98.329 -57.721 50.622
581498 98.374 -57.724 50.587
581499 98.419 -57.727 50.555
581500 98.464 -57.730 50.524
581501 98.506 -57.730 50.492
581502 98.535 -57.724 50.459
581503 98.566 -57.719 50.429
581504 98.596 -57.713 50.399
581505 98.626 -57.707 50.367
581506 98.656 -57.702 50.334
581507 98.686 -57.695 50.302
581508 98.716 -57.690 50.271
581509 98.746 -57.684 50.242
581510 98.776 -57.678 50.214
581511 98.806 -57.673 50.186
581512 98.842 -57.676 50.158
581513 98.872 -57.683 50.129
581514 98.899 -57.690 50.100
581515 98.923 -57.695 50.071
581516 98.946 -57.700 50.042
581517 98.971 -57.705 50.012
581518 98.995 -57.710 49.982
581519 99.018 -57.715 49.951
581520 99.043 -57.720 49.922
581521 99.066 -57.725 49.892
581522 99.090 -57.730 49.862
581523 99.114 -57.735 49.831
581524 99.138 -57.740 49.800
581525 99.162 -57.746 49.768
581526 99.186 -57.750 49.736
581527 99.210 -57.755 49.703
581528 99.230 -57.761 49.670
581529 99.256 -57.768 49.636
581530 99.292 -57.778 49.604
581531 99.329 -57.788 49.572
581532 99.365 -57.798 49.540
581533 99.403 -57.808 49.506
581534 99.424 -57.821 49.471
581535 99.483 -57.830 49.435
581536 99.534 -57.836 49.401
581537 99.574 -57.845 49.367
581538 99.608 -57.853 49.334
581539 99.643 -57.862 49.300
581540 99.677 -57.871 49.266
581541 99.712 -57.880 49.233
581542 99.746 -57.889 49.199
581543 99.781 -57.897 49.168
581544 99.815 -57.906 49.138
581545 99.850 -57.915 49.107
581546 99.884 -57.924 49.078
581547 99.919 -57.932 49.050
581548 99.953 -57.941 49.023
581893 88.063 -54.157 57.578
581894 88.099 -54.172 57.580
581895 88.134 -54.188 57.581
581896 88.169 -54.203 57.582
581897 88.204 -54.218 57.583
581898 88.240 -54.233 57.584
581899 88.275 -54.248 57.585
581900 88.311 -54.264 57.585
581901 88.346 -54.279 57.583
581902 88.381 -54.295 57.582
581903 88.417 -54.310 57.581
581904 88.452 -54.325 57.581
581905 88.487 -54.341 57.579
581906 88.523 -54.356 57.576
581907 88.567 -54.345 57.572
581908 88.632 -54.276 57.567
581909 88.676 -54.296 57.565
581910 88.711 -54.331 57.560
581911 88.747 -54.361 57.557
581912 88.785 -54.378 57.544
581913 88.823 -54.395 57.516
581914 88.861 -54.411 57.474
581915 88.899 -54.428 57.430
581916 88.937 -54.445 57.391
581917 88.976 -54.461 57.349
581918 89.014 -54.479 57.304
581919 89.052 -54.496 57.257
581920 89.090 -54.512 57.210
581921 89.128 -54.529 57.161
581922 89.166 -54.546 57.113
581923 89.204 -54.562 57.065
581924 89.243 -54.584 57.017
581925 89.281 -54.608 56.968
581926 89.320 -54.631 56.921
581948 90.169 -55.146 55.945
581949 90.192 -55.160 55.923
581950 90.216 -55.175 55.901
581951 90.239 -55.188 55.882
581952 90.262 -55.203 55.866
581953 90.286 -55.217 55.850
581954 90.309 -55.231 55.837
581955 90.332 -55.245 55.827
581956 90.355 -55.259 55.820
581957 90.379 -55.273 55.814
581958 90.402 -55.287 55.811
581959 90.425 -55.301 55.812
581960 90.448 -55.316 55.815
581961 90.471 -55.329 55.823
581962 90.497 -55.345 55.837
581963 90.523 -55.355 55.857
581964 90.548 -55.366 55.878
581965 90.574 -55.376 55.898
581966 90.599 -55.386 55.914
581967 90.624 -55.396 55.925
581968 90.660 -55.407 55.933
581969 90.700 -55.417 55.942
581970 90.739 -55.426 55.954
581971 90.783 -55.437 55.968
581972 90.828 -55.448 55.982
581973 90.874 -55.459 56.001
581974 90.919 -55.470 56.022
581975 90.964 -55.481 56.038
581976 91.009 -55.492 56.046
581977 91.054 -55.503 56.043
581978 91.100 -55.514 56.031
581979 91.145 -55.525 56.019
581980 91.190 -55.536 56.010
581981 91.236 -55.547 56.009
581982 91.281 -55.558 56.006
581983 91.326 -55.569 55.992
581984 91.371 -55.580 55.956
581985 91.417 -55.591 55.920
581986 91.462 -55.602 55.888
581987 91.499 -55.610 55.851
581988 91.529 -55.619 55.815
581989 91.560 -55.629 55.784
581990 91.590 -55.640 55.756
581991 91.621 -55.652 55.732
581992 91.651 -55.664 55.696
581993 91.681 -55.675 55.652
581994 91.711 -55.687 55.602
581995 91.741 -55.699 55.550
581996 91.772 -55.711 55.502
581997 91.802 -55.722 55.462
581998 91.832 -55.734 55.424
581999 91.863 -55.746 55.389
582000 91.893 -55.758 55.356
582001 91.923 -55.769 55.325
582002 91.953 -55.781 55.296
582003 91.985 -55.793 55.271
582004 92.017 -55.806 55.248
582005 92.049 -55.824 55.228
582006 92.079 -55.850 55.212
582007 92.109 -55.875 55.200
582008 92.140 -55.901 55.193
582009 92.172 -55.925 55.188
582010 92.204 -55.948 55.183
582011 92.236 -55.972 55.179
582012 92.269 -55.995 55.177
582013 92.301 -56.019 55.182
582014 92.333 -56.042 55.187
582015 92.365 -56.066 55.190
582016 92.398 -56.089 55.193
582017 92.430 -56.113 55.199
582018 92.462 -56.136 55.209
582019 92.496 -56.151 55.219
582020 92.530 -56.166 55.224
582021 92.565 -56.180 55.226
582022 92.599 -56.194 55.234
582023 92.633 -56.209 55.246
582024 92.667 -56.223 55.256
582025 92.702 -56.238 55.263
582026 92.736 -56.252 55.271
582027 92.770 -56.267 55.290
582028 92.806 -56.283 55.310
582029 92.846 -56.301 55.326
582030 92.889 -56.315 55.341
582031 92.932 -56.329 55.351
582032 92.975 -56.342 55.357
582033 93.018 -56.356 55.358
582034 93.061 -56.370 55.359
582035 93.105 -56.383 55.361
582036 93.147 -56.397 55.366
582037 93.191 -56.410 55.369
582038 93.233 -56.423 55.369
582039 93.277 -56.437 55.364
582040 93.319 -56.450 55.355
582041 93.347 -56.457 55.345
582042 93.376 -56.464 55.337
582043 93.409 -56.474 55.328
582044 93.441 -56.486 55.317
582045 93.474 -56.498 55.299
582046 93.506 -56.510 55.277
582047 93.539 -56.522 55.252
582048 93.572 -56.534 55.226
582049 93.605 -56.545 55.199
582050 93.637 -56.558 55.171
582051 93.670 -56.570 55.138
582052 93.703 -56.581 55.102
582053 93.735 -56.593 55.066
582054 93.768 -56.605 55.028
582055 93.800 -56.617 54.989
582056 93.833 -56.629 54.949
582057 93.866 -56.641 54.906
582058 93.899 -56.653 54.860
582059 93.931 -56.664 54.813
582060 93.964 -56.677 54.766
582061 93.997 -56.689 54.719
582062 94.029 -56.702 54.675
582063 94.061 -56.718 54.637
582064 94.094 -56.735 54.602
582065 94.125 -56.752 54.569
582066 94.157 -56.769 54.536
582067 94.188 -56.785 54.503
582068 94.218 -56.801 54.470
582069 94.249 -56.817 54.436
582070 94.277 -56.826 54.400
582071 94.309 -56.835 54.363
582072 94.344 -56.844 54.328
582073 94.375 -56.859 54.291
582074 94.404 -56.873 54.254
582075 94.433 -56.888 54.216
582076 94.463 -56.903 54.175
582077 94.492 -56.918 54.131
582078 94.522 -56.932 54.085
582079 94.551 -56.947 54.038
582080 94.580 -56.962 53.989
582081 94.610 -56.977 53.939
582082 94.640 -56.991 53.890
582083 94.672 -57.003 53.844
582084 94.704 -57.016 53.802
582085 94.737 -57.028 53.766
582086 94.769 -57.041 53.739
582087 94.801 -57.053 53.717
582088 94.833 -57.066 53.699
582089 94.865 -57.078 53.678
582090 94.898 -57.091 53.653
582091 94.930 -57.103 53.628
582092 94.963 -57.116 53.604
582093 94.995 -57.128 53.581
582094 95.027 -57.141 53.556
582095 95.059 -57.153 53.530
582096 95.092 -57.166 53.502
582097 95.124 -57.178 53.473
582098 95.156 -57.191 53.445
582099 95.192 -57.198 53.420
582100 95.231 -57.202 53.397
582101 95.269 -57.205 53.369
582102 95.308 -57.214 53.341
582103 95.346 -57.223 53.307
582104 95.385 -57.233 53.270
582105 95.423 -57.242 53.232
582106 95.462 -57.251 53.191
582107 95.500 -57.261 53.147
582108 95.539 -57.270 53.101
582109 95.577 -57.280 53.054
582110 95.615 -57.289 53.008
582111 95.654 -57.298 52.966
582112 95.692 -57.308 52.930
582113 95.731 -57.316 52.898
582114 95.770 -57.325 52.871
582115 95.809 -57.333 52.843
582116 95.833 -57.340 52.810
582117 95.855 -57.347 52.775
582118 95.876 -57.353 52.740
582119 95.897 -57.358 52.702
582120 95.918 -57.364 52.660
582121 95.939 -57.369 52.615
582122 95.960 -57.375 52.569
582123 95.981 -57.381 52.525
582124 96.002 -57.386 52.483
582125 96.023 -57.391 52.447
582126 96.044 -57.398 52.412
582127 96.065 -57.403 52.376
582128 96.086 -57.408 52.342
582129 96.114 -57.416 52.312
582130 96.143 -57.423 52.286
582131 96.172 -57.431 52.260
582132 96.201 -57.438 52.233
582133 96.230 -57.446 52.207
582134 96.259 -57.454 52.179
582135 96.288 -57.461 52.152
582136 96.317 -57.469 52.131
582137 96.346 -57.476 52.113
582138 96.375 -57.484 52.096
582139 96.404 -57.492 52.084
582140 96.434 -57.499 52.075
582141 96.462 -57.507 52.068
582142 96.491 -57.514 52.064
582143 96.521 -57.522 52.063
582144 96.549 -57.529 52.058
582145 96.579 -57.536 52.046
582146 96.615 -57.542 52.026
582147 96.651 -57.547 52.002
582148 96.686 -57.552 51.982
582149 96.722 -57.558 51.968
582150 96.758 -57.563 51.955
582151 96.794 -57.568 51.945
582152 96.829 -57.573 51.940
582153 96.865 -57.578 51.941
582154 96.901 -57.583 51.945
582155 96.937 -57.589 51.952
582156 96.973 -57.594 51.964
582157 97.009 -57.599 51.976
582158 97.044 -57.605 51.987
582159 97.081 -57.606 51.995
582160 97.116 -57.604 52.000
582161 97.144 -57.601 52.003
582162 97.171 -57.598 52.004
582163 97.198 -57.595 52.002
582164 97.226 -57.591 51.996
582165 97.253 -57.588 51.986
582166 97.281 -57.585 51.975
582167 97.308 -57.582 51.961
582168 97.335 -57.579 51.943
582169 97.363 -57.575 51.924
582170 97.390 -57.572 51.905
582171 97.417 -57.569 51.887
582172 97.445 -57.566 51.863
582173 97.474 -57.569 51.831
582174 97.509 -57.583 51.791
582175 97.543 -57.598 51.745
582176 97.578 -57.611 51.696
582177 97.608 -57.618 51.643
582178 97.639 -57.626 51.588
582179 97.669 -57.633 51.532
582180 97.700 -57.640 51.475
582181 97.730 -57.646 51.415
582182 97.760 -57.648 51.354
582183 97.790 -57.650 51.296
582184 97.820 -57.651 51.244
582185 97.850 -57.653 51.198
582186 97.879 -57.655 51.153
582187 97.909 -57.656 51.109
582188 97.939 -57.658 51.064
582189 97.969 -57.660 51.020
582190 97.999 -57.661 50.975
582191 98.033 -57.664 50.929
582192 98.069 -57.669 50.883
582193 98.104 -57.673 50.837
582194 98.143 -57.677 50.793
582195 98.189 -57.680 50.750
582196 98.234 -57.683 50.709
582197 98.279 -57.685 50.671
582198 98.324 -57.688 50.635
582199 98.369 -57.691 50.602
582200 98.414 -57.693 50.571
582201 98.459 -57.696 50.540
582202 98.504 -57.699 50.507
582203 98.536 -57.695 50.475
582204 98.566 -57.689 50.444
582205 98.596 -57.683 50.414
582206 98.627 -57.677 50.382
582207 98.656 -57.671 50.348
582208 98.687 -57.666 50.315
582209 98.717 -57.660 50.284
582210 98.746 -57.654 50.255
582211 98.777 -57.648 50.226
582212 98.808 -57.645 50.196
582213 98.846 -57.648 50.166
582214 98.876 -57.655 50.136
582215 98.906 -57.664 50.106
582216 98.934 -57.670 50.077
582217 98.958 -57.676 50.048
582218 98.982 -57.680 50.018
582219 99.006 -57.686 49.988
582220 99.029 -57.691 49.957
582221 99.054 -57.695 49.926
582222 99.077 -57.701 49.895
582223 99.101 -57.705 49.864
582224 99.125 -57.711 49.832
582225 99.149 -57.716 49.800
582226 99.173 -57.720 49.767
582227 99.197 -57.726 49.734
582228 99.219 -57.731 49.701
582229 99.231 -57.738 49.668
582230 99.257 -57.745 49.635
582231 99.286 -57.752 49.601
582232 99.319 -57.761 49.570
582233 99.355 -57.771 49.538
582234 99.393 -57.781 49.505
582235 99.429 -57.791 49.470
582236 99.468 -57.803 49.434
582237 99.519 -57.810 49.399
582238 99.566 -57.817 49.364
582239 99.600 -57.826 49.330
582240 99.635 -57.834 49.296
582241 99.669 -57.843 49.261
582242 99.704 -57.852 49.227
582243 99.738 -57.861 49.194
582244 99.773 -57.870 49.163
582245 99.807 -57.878 49.132
582246 99.842 -57.887 49.102
582247 99.876 -57.896 49.073
582248 99.911 -57.905 49.045
582249 99.945 -57.913 49.018
582250 99.980 -57.922 48.993
582594 88.088 -54.122 57.574
582595 88.123 -54.137 57.576
582596 88.158 -54.152 57.578
582597 88.194 -54.168 57.580
582598 88.229 -54.183 57.581
582599 88.265 -54.198 57.583
582600 88.300 -54.214 57.583
582601 88.335 -54.229 57.583
582602 88.370 -54.245 57.582
582603 88.406 -54.260 57.582
582604 88.441 -54.275 57.581
582605 88.477 -54.290 57.580
582606 88.512 -54.306 57.579
582607 88.547 -54.321 57.576
582608 88.582 -54.336 57.573
582609 88.649 -54.261 57.570
582610 88.698 -54.270 57.570
582611 88.733 -54.306 57.572
582612 88.769 -54.338 57.574
582613 88.807 -54.355 57.560
582614 88.845 -54.371 57.529
582615 88.883 -54.388 57.493
582616 88.921 -54.405 57.445
582617 88.959 -54.421 57.402
582618 88.997 -54.439 57.358
582619 89.036 -54.455 57.313
582620 89.074 -54.472 57.267
582621 89.112 -54.489 57.218
582622 89.150 -54.505 57.170
582623 89.188 -54.523 57.121
582624 89.226 -54.540 57.073
582625 89.264 -54.556 57.025
582626 89.303 -54.576 56.976
582627 89.341 -54.599 56.929
582628 89.380 -54.623 56.885
582648 90.152 -55.091 55.980
582649 90.188 -55.113 55.956
582650 90.212 -55.127 55.933
582651 90.235 -55.141 55.911
582652 90.258 -55.155 55.893
582653 90.282 -55.169 55.876
582654 90.305 -55.183 55.860
582655 90.328 -55.198 55.847
582656 90.351 -55.212 55.836
582657 90.374 -55.226 55.829
582658 90.397 -55.240 55.823
582659 90.421 -55.254 55.819
582660 90.444 -55.268 55.818
582661 90.467 -55.282 55.821
582662 90.490 -55.296 55.829
582663 90.515 -55.311 55.843
582664 90.540 -55.322 55.859
582665 90.566 -55.332 55.877
582666 90.592 -55.342 55.895
582667 90.617 -55.352 55.909
582668 90.643 -55.363 55.918
582669 90.672 -55.373 55.924
582670 90.714 -55.383 55.932
582671 90.759 -55.395 55.941
582672 90.804 -55.405 55.953
582673 90.850 -55.417 55.966
582674 90.895 -55.427 55.981
582675 90.940 -55.439 55.997
582676 90.986 -55.449 56.009
582677 91.030 -55.461 56.012
582678 91.076 -55.471 56.009
582679 91.121 -55.483 56.001
582680 91.166 -55.493 55.992
582681 91.212 -55.505 55.984
582682 91.257 -55.515 55.979
582683 91.302 -55.527 55.972
582684 91.347 -55.538 55.957
582685 91.392 -55.549 55.926
582686 91.437 -55.561 55.893
582687 91.482 -55.571 55.862
582688 91.517 -55.583 55.827
582689 91.547 -55.596 55.790
582690 91.578 -55.607 55.760
582691 91.608 -55.619 55.735
582692 91.638 -55.630 55.708
582693 91.668 -55.643 55.673
582694 91.699 -55.655 55.630
582695 91.729 -55.666 55.582
582696 91.759 -55.678 55.532
582697 91.790 -55.690 55.487
582698 91.820 -55.702 55.448
582699 91.850 -55.713 55.413
582700 91.880 -55.725 55.379
582701 91.911 -55.737 55.347
582702 91.941 -55.749 55.316
582703 91.971 -55.760 55.289
582704 92.003 -55.772 55.265
582705 92.035 -55.785 55.242
582706 92.068 -55.798 55.223
582707 92.098 -55.824 55.206
582708 92.129 -55.850 55.193
582709 92.159 -55.875 55.186
582710 92.190 -55.901 55.182
582711 92.221 -55.925 55.179
582712 92.253 -55.949 55.177
582713 92.285 -55.972 55.178
582714 92.317 -55.996 55.183
582715 92.350 -56.019 55.188
582716 92.382 -56.043 55.192
582717 92.414 -56.066 55.194
582718 92.446 -56.090 55.198
582719 92.479 -56.112 55.206
582720 92.513 -56.126 55.215
582721 92.548 -56.141 55.221
582722 92.582 -56.155 55.225
582723 92.616 -56.169 55.228
582724 92.650 -56.185 55.234
582725 92.685 -56.199 55.242
582726 92.719 -56.213 55.252
582727 92.753 -56.228 55.272
582728 92.787 -56.242 55.296
582729 92.823 -56.257 55.317
582730 92.859 -56.269 55.335
582731 92.897 -56.276 55.348
582732 92.940 -56.290 55.358
582733 92.984 -56.304 55.363
582734 93.026 -56.317 55.366
582735 93.070 -56.331 55.367
582736 93.113 -56.345 55.365
582737 93.156 -56.358 55.363
582738 93.199 -56.372 55.366
582739 93.242 -56.385 55.366
582740 93.285 -56.399 55.359
582741 93.328 -56.413 55.349
582742 93.358 -56.420 55.337
582743 93.388 -56.426 55.327
582744 93.420 -56.438 55.316
582745 93.452 -56.450 55.303
582746 93.485 -56.462 55.286
582747 93.518 -56.473 55.263
582748 93.551 -56.486 55.239
582749 93.583 -56.498 55.214
582750 93.616 -56.509 55.186
582751 93.649 -56.521 55.156
582752 93.681 -56.533 55.123
582753 93.714 -56.545 55.088
582754 93.747 -56.557 55.052
582755 93.779 -56.569 55.018
582756 93.812 -56.580 54.983
582757 93.845 -56.592 54.948
582758 93.877 -56.605 54.909
582759 93.910 -56.616 54.867
582760 93.943 -56.628 54.824
582761 93.975 -56.644 54.778
582762 94.007 -56.661 54.730
582763 94.039 -56.678 54.683
582764 94.071 -56.695 54.636
582765 94.103 -56.712 54.592
582766 94.135 -56.729 54.557
582767 94.166 -56.745 54.523
582768 94.196 -56.761 54.489
582769 94.226 -56.776 54.454
582770 94.256 -56.791 54.417
582771 94.284 -56.800 54.379
582772 94.313 -56.809 54.342
582773 94.349 -56.818 54.306
582774 94.385 -56.828 54.269
582775 94.415 -56.842 54.232
582776 94.445 -56.857 54.194
582777 94.474 -56.872 54.154
582778 94.503 -56.887 54.111
582779 94.533 -56.902 54.067
582780 94.562 -56.916 54.023
582781 94.591 -56.931 53.979
582782 94.621 -56.946 53.932
582783 94.650 -56.961 53.886
582784 94.679 -56.975 53.841
582785 94.708 -56.990 53.801
582786 94.738 -57.005 53.761
582787 94.767 -57.020 53.725
582788 94.800 -57.033 53.694
582789 94.832 -57.045 53.670
582790 94.864 -57.058 53.650
582791 94.896 -57.070 53.627
582792 94.928 -57.082 53.601
582793 94.961 -57.095 53.576
582794 94.993 -57.107 53.552
582795 95.026 -57.120 53.528
582796 95.058 -57.132 53.503
582797 95.090 -57.145 53.476
582798 95.122 -57.157 53.444
582799 95.157 -57.168 53.413
582800 95.195 -57.172 53.385
582801 95.233 -57.175 53.359
582802 95.272 -57.178 53.331
582803 95.310 -57.186 53.305
582804 95.348 -57.195 53.273
582805 95.387 -57.205 53.238
582806 95.425 -57.214 53.200
582807 95.464 -57.223 53.160
582808 95.502 -57.233 53.116
582809 95.541 -57.242 53.070
582810 95.580 -57.251 53.023
582811 95.618 -57.260 52.977
582812 95.657 -57.268 52.936
582813 95.696 -57.276 52.902
582814 95.735 -57.285 52.873
582815 95.773 -57.293 52.847
582816 95.812 -57.301 52.820
582817 95.837 -57.308 52.789
582818 95.859 -57.315 52.757
582819 95.880 -57.322 52.723
582820 95.902 -57.329 52.687
582821 95.923 -57.335 52.647
582822 95.945 -57.341 52.604
582823 95.965 -57.347 52.560
582824 95.986 -57.352 52.516
582825 96.007 -57.357 52.474
582826 96.028 -57.363 52.437
582827 96.049 -57.369 52.402
582828 96.070 -57.374 52.366
582829 96.091 -57.379 52.332
582830 96.116 -57.386 52.301
582831 96.145 -57.394 52.273
582832 96.174 -57.401 52.246
582833 96.203 -57.409 52.222
582834 96.232 -57.416 52.195
582835 96.261 -57.424 52.166
582836 96.291 -57.432 52.138
582837 96.319 -57.439 52.115
582838 96.348 -57.447 52.095
582839 96.378 -57.454 52.076
582840 96.406 -57.462 52.061
582841 96.436 -57.470 52.050
582842 96.465 -57.477 52.041
582843 96.493 -57.484 52.032
582844 96.522 -57.491 52.029
582845 96.550 -57.497 52.026
582846 96.581 -57.502 52.018
582847 96.613 -57.507 52.003
582848 96.645 -57.511 51.981
582849 96.681 -57.517 51.959
582850 96.717 -57.522 51.940
582851 96.752 -57.527 51.927
582852 96.788 -57.532 51.918
582853 96.824 -57.538 51.912
582854 96.860 -57.543 51.910
582855 96.895 -57.548 51.911
582856 96.932 -57.553 51.917
582857 96.967 -57.558 51.925
582858 97.003 -57.564 51.935
582859 97.039 -57.569 51.944
582860 97.076 -57.568 51.951
582861 97.113 -57.566 51.957
582862 97.147 -57.564 51.961
582863 97.174 -57.560 51.961
582864 97.201 -57.557 51.959
582865 97.229 -57.554 51.955
582866 97.256 -57.551 51.946
582867 97.283 -57.548 51.935
582868 97.311 -57.544 51.921
582869 97.338 -57.541 51.906
582870 97.365 -57.538 51.890
582871 97.392 -57.535 51.873
582872 97.420 -57.531 51.854
582873 97.447 -57.528 51.833
582874 97.477 -57.530 51.801
582875 97.511 -57.545 51.762
582876 97.545 -57.558 51.719
582877 97.579 -57.571 51.672
582878 97.610 -57.579 51.622
582879 97.640 -57.586 51.568
582880 97.671 -57.593 51.514
582881 97.702 -57.601 51.460
582882 97.732 -57.608 51.404
582883 97.762 -57.613 51.345
582884 97.792 -57.614 51.285
582885 97.822 -57.616 51.234
582886 97.852 -57.617 51.192
582887 97.882 -57.619 51.150
582888 97.912 -57.620 51.107
582889 97.941 -57.622 51.064
582890 97.971 -57.624 51.021
582891 98.002 -57.626 50.978
582892 98.037 -57.630 50.934
582893 98.073 -57.634 50.888
582894 98.108 -57.638 50.844
582895 98.144 -57.642 50.801
582896 98.183 -57.646 50.759
582897 98.228 -57.649 50.719
582898 98.273 -57.651 50.683
582899 98.318 -57.655 50.648
582900 98.363 -57.658 50.615
582901 98.409 -57.660 50.584
582902 98.454 -57.663 50.553
582903 98.499 -57.666 50.519
582904 98.537 -57.664 50.488
582905 98.567 -57.659 50.457
582906 98.597 -57.653 50.426
582907 98.627 -57.647 50.394
582908 98.657 -57.642 50.361
582909 98.687 -57.636 50.327
582910 98.717 -57.630 50.297
582911 98.747 -57.624 50.267
582912 98.777 -57.618 50.236
582913 98.811 -57.617 50.205
582914 98.849 -57.620 50.174
582915 98.880 -57.627 50.142
582916 98.910 -57.636 50.112
582917 98.940 -57.643 50.083
582918 98.969 -57.651 50.053
582919 98.993 -57.656 50.024
582920 99.017 -57.661 49.994
582921 99.041 -57.666 49.963
582922 99.065 -57.671 49.931
582923 99.088 -57.677 49.899
582924 99.113 -57.681 49.866
582925 99.137 -57.686 49.833
582926 99.160 -57.691 49.802
582927 99.185 -57.696 49.768
582928 99.208 -57.702 49.734
582929 99.220 -57.708 49.700
582930 99.232 -57.714 49.667
582931 99.259 -57.721 49.634
582932 99.287 -57.729 49.600
582933 99.316 -57.736 49.569
582934 99.345 -57.743 49.537
582935 99.382 -57.754 49.504
582936 99.419 -57.764 49.470
582937 99.451 -57.776 49.434
582938 99.505 -57.783 49.397
582939 99.556 -57.789 49.361
582940 99.593 -57.798 49.326
582941 99.628 -57.807 49.291
582942 99.662 -57.815 49.257
582943 99.697 -57.824 49.222
582944 99.731 -57.833 49.189
582945 99.766 -57.842 49.158
582946 99.800 -57.851 49.126
582947 99.835 -57.859 49.097
582948 99.869 -57.868 49.069
582949 99.904 -57.877 49.041
582950 99.938 -57.886 49.014
582951 99.973 -57.894 48.989
582952 100.007 -57.903 48.966
583295 88.112 -54.087 57.571
583296 88.148 -54.102 57.574
583297 88.183 -54.117 57.576
583298 88.218 -54.132 57.577
583299 88.254 -54.148 57.579
583300 88.289 -54.164 57.580
583301 88.325 -54.179 57.582
583302 88.360 -54.194 57.582
583303 88.395 -54.209 57.582
583304 88.430 -54.225 57.581
583305 88.466 -54.240 57.581
583306 88.501 -54.255 57.579
583307 88.537 -54.270 57.578
583308 88.572 -54.286 57.576
583309 88.607 -54.301 57.574
583310 88.665 -54.251 57.572
583311 88.717 -54.251 57.574
583312 88.755 -54.281 57.579
583313 88.790 -54.314 57.578
583314 88.829 -54.331 57.566
583315 88.867 -54.348 57.541
583316 88.905 -54.364 57.505
583317 88.943 -54.382 57.458
583318 88.981 -54.399 57.413
583319 89.020 -54.415 57.368
583320 89.057 -54.432 57.323
583321 89.096 -54.449 57.275
583322 89.134 -54.465 57.227
583323 89.172 -54.483 57.178
583324 89.210 -54.499 57.130
583325 89.248 -54.516 57.081
583326 89.286 -54.533 57.033
583327 89.324 -54.549 56.984
583328 89.363 -54.568 56.937
583329 89.401 -54.590 56.892
583347 90.096 -55.012 56.042
583348 90.134 -55.035 56.015
583349 90.173 -55.059 55.990
583350 90.207 -55.080 55.966
583351 90.231 -55.094 55.943
583352 90.254 -55.108 55.922
583353 90.277 -55.122 55.903
583354 90.300 -55.136 55.886
583355 90.324 -55.151 55.870
583356 90.347 -55.164 55.856
583357 90.370 -55.179 55.845
583358 90.393 -55.192 55.838
583359 90.416 -55.207 55.831
583360 90.440 -55.221 55.827
583361 90.463 -55.235 55.825
583362 90.486 -55.249 55.828
583363 90.510 -55.263 55.835
583364 90.533 -55.277 55.848
583365 90.558 -55.289 55.862
583366 90.584 -55.298 55.876
583367 90.610 -55.309 55.890
583368 90.636 -55.319 55.903
583369 90.662 -55.329 55.909
583370 90.692 -55.342 55.914
583371 90.730 -55.357 55.920
583372 90.775 -55.368 55.929
583373 90.820 -55.379 55.940
583374 90.865 -55.390 55.954
583375 90.910 -55.401 55.969
583376 90.955 -55.413 55.981
583377 91.000 -55.424 55.988
583378 91.045 -55.436 55.987
583379 91.090 -55.447 55.982
583380 91.135 -55.458 55.975
583381 91.180 -55.469 55.970
583382 91.225 -55.480 55.965
583383 91.270 -55.492 55.955
583384 91.315 -55.503 55.938
583385 91.360 -55.514 55.917
583386 91.404 -55.526 55.888
583387 91.450 -55.537 55.858
583388 91.493 -55.549 55.830
583389 91.528 -55.561 55.799
583390 91.563 -55.574 55.765
583391 91.596 -55.586 55.734
583392 91.626 -55.598 55.709
583393 91.656 -55.610 55.678
583394 91.686 -55.621 55.643
583395 91.717 -55.633 55.603
583396 91.747 -55.645 55.559
583397 91.778 -55.657 55.513
583398 91.808 -55.668 55.472
583399 91.838 -55.680 55.435
583400 91.868 -55.692 55.401
583401 91.898 -55.704 55.370
583402 91.929 -55.716 55.339
583403 91.959 -55.727 55.310
583404 91.989 -55.740 55.284
583405 92.021 -55.752 55.260
583406 92.053 -55.765 55.240
583407 92.086 -55.777 55.221
583408 92.117 -55.797 55.204
583409 92.148 -55.824 55.190
583410 92.178 -55.850 55.180
583411 92.209 -55.875 55.174
583412 92.239 -55.901 55.170
583413 92.270 -55.926 55.170
583414 92.302 -55.950 55.172
583415 92.335 -55.973 55.177
583416 92.367 -55.997 55.184
583417 92.399 -56.020 55.190
583418 92.431 -56.044 55.193
583419 92.463 -56.067 55.194
583420 92.496 -56.088 55.198
583421 92.530 -56.102 55.206
583422 92.565 -56.116 55.216
583423 92.599 -56.131 55.223
583424 92.633 -56.145 55.227
583425 92.667 -56.160 55.230
583426 92.702 -56.174 55.237
583427 92.736 -56.188 55.248
583428 92.771 -56.204 55.271
583429 92.807 -56.219 55.295
583430 92.844 -56.235 55.315
583431 92.880 -56.246 55.331
583432 92.916 -56.251 55.343
583433 92.951 -56.255 55.351
583434 92.992 -56.265 55.359
583435 93.035 -56.279 55.364
583436 93.078 -56.292 55.367
583437 93.121 -56.306 55.368
583438 93.164 -56.320 55.366
583439 93.207 -56.333 55.367
583440 93.250 -56.347 55.367
583441 93.293 -56.360 55.362
583442 93.336 -56.374 55.354
583443 93.370 -56.383 55.344
583444 93.399 -56.389 55.333
583445 93.431 -56.401 55.318
583446 93.464 -56.414 55.300
583447 93.497 -56.425 55.279
583448 93.530 -56.437 55.256
583449 93.562 -56.449 55.233
583450 93.595 -56.461 55.207
583451 93.627 -56.473 55.176
583452 93.660 -56.485 55.144
583453 93.693 -56.496 55.111
583454 93.725 -56.508 55.077
583455 93.758 -56.520 55.041
583456 93.791 -56.532 55.008
583457 93.824 -56.544 54.974
583458 93.856 -56.556 54.941
583459 93.889 -56.570 54.906
583460 93.921 -56.587 54.867
583461 93.953 -56.604 54.827
583462 93.985 -56.621 54.784
583463 94.017 -56.638 54.738
583464 94.049 -56.655 54.690
583465 94.081 -56.672 54.640
583466 94.113 -56.689 54.592
583467 94.143 -56.705 54.551
583468 94.174 -56.720 54.514
583469 94.204 -56.736 54.480
583470 94.235 -56.752 54.446
583471 94.264 -56.765 54.409
583472 94.292 -56.774 54.372
583473 94.320 -56.783 54.336
583474 94.353 -56.793 54.301
583475 94.389 -56.802 54.265
583476 94.425 -56.812 54.227
583477 94.455 -56.826 54.187
583478 94.484 -56.841 54.145
583479 94.514 -56.855 54.103
583480 94.543 -56.871 54.059
583481 94.572 -56.886 54.017
583482 94.602 -56.900 53.975
583483 94.631 -56.915 53.933
583484 94.661 -56.930 53.891
583485 94.690 -56.945 53.849
583486 94.719 -56.959 53.810
583487 94.749 -56.975 53.764
583488 94.778 -56.989 53.718
583489 94.808 -57.004 53.677
583490 94.837 -57.018 53.644
583491 94.866 -57.034 53.620
583492 94.896 -57.049 53.600
583493 94.927 -57.062 53.574
583494 94.959 -57.074 53.548
583495 94.991 -57.087 53.523
583496 95.024 -57.099 53.498
583497 95.056 -57.112 53.473
583498 95.089 -57.124 53.445
583499 95.121 -57.137 53.413
583500 95.159 -57.141 53.380
583501 95.197 -57.144 53.350
583502 95.236 -57.148 53.324
583503 95.274 -57.152 53.298
583504 95.313 -57.158 53.271
583505 95.351 -57.167 53.240
583506 95.389 -57.177 53.205
583507 95.428 -57.185 53.167
583508 95.467 -57.194 53.128
583509 95.506 -57.203 53.083
583510 95.544 -57.211 53.036
583511 95.583 -57.219 52.990
583512 95.622 -57.228 52.945
583513 95.661 -57.236 52.905
583514 95.699 -57.244 52.873
583515 95.739 -57.253 52.845
583516 95.777 -57.261 52.818
583517 95.815 -57.269 52.791
583518 95.840 -57.276 52.764
583519 95.862 -57.283 52.734
583520 95.884 -57.290 52.701
583521 95.905 -57.297 52.665
583522 95.927 -57.304 52.624
583523 95.949 -57.310 52.583
583524 95.970 -57.317 52.542
583525 95.992 -57.323 52.501
583526 96.013 -57.329 52.461
583527 96.034 -57.335 52.423
583528 96.054 -57.340 52.387
583529 96.076 -57.345 52.352
583530 96.096 -57.351 52.318
583531 96.118 -57.357 52.288
583532 96.148 -57.364 52.261
583533 96.177 -57.372 52.234
583534 96.205 -57.379 52.209
583535 96.235 -57.387 52.182
583536 96.264 -57.395 52.152
583537 96.293 -57.402 52.124
583538 96.322 -57.410 52.100
583539 96.351 -57.417 52.079
583540 96.380 -57.425 52.058
583541 96.408 -57.432 52.040
583542 96.437 -57.438 52.025
583543 96.465 -57.445 52.014
583544 96.493 -57.451 52.001
583545 96.523 -57.457 51.995
583546 96.552 -57.463 51.992
583547 96.582 -57.468 51.986
583548 96.614 -57.473 51.974
583549 96.647 -57.477 51.955
583550 96.679 -57.482 51.933
583551 96.712 -57.486 51.912
583552 96.747 -57.492 51.900
583553 96.783 -57.497 51.893
583554 96.818 -57.502 51.887
583555 96.854 -57.507 51.882
583556 96.890 -57.513 51.881
583557 96.926 -57.518 51.883
583558 96.962 -57.523 51.887
583559 96.998 -57.528 51.893
583560 97.033 -57.532 51.898
583561 97.071 -57.529 51.903
583562 97.107 -57.527 51.909
583563 97.145 -57.525 51.913
583564 97.177 -57.523 51.913
583565 97.204 -57.520 51.912
583566 97.231 -57.516 51.909
583567 97.259 -57.513 51.903
583568 97.286 -57.510 51.892
583569 97.313 -57.507 51.880
583570 97.341 -57.504 51.866
583571 97.368 -57.500 51.851
583572 97.395 -57.497 51.834
583573 97.423 -57.494 51.815
583574 97.450 -57.491 51.794
583575 97.479 -57.492 51.766
583576 97.513 -57.505 51.729
583577 97.548 -57.520 51.687
583578 97.581 -57.532 51.641
583579 97.611 -57.539 51.593
583580 97.642 -57.546 51.541
583581 97.673 -57.554 51.489
583582 97.703 -57.561 51.436
583583 97.734 -57.568 51.382
583584 97.764 -57.576 51.325
583585 97.794 -57.579 51.270
583586 97.824 -57.580 51.224
583587 97.854 -57.582 51.183
583588 97.884 -57.583 51.143
583589 97.914 -57.585 51.102
583590 97.944 -57.587 51.061
583591 97.973 -57.589 51.020
583592 98.006 -57.591 50.978
583593 98.041 -57.595 50.935
583594 98.077 -57.599 50.891
583595 98.113 -57.604 50.847
583596 98.149 -57.608 50.805
583597 98.184 -57.612 50.764
583598 98.223 -57.616 50.726
583599 98.268 -57.618 50.689
583600 98.313 -57.621 50.655
583601 98.358 -57.624 50.622
583602 98.403 -57.626 50.591
583603 98.448 -57.630 50.560
583604 98.493 -57.633 50.526
583605 98.537 -57.635 50.494
583606 98.567 -57.629 50.463
583607 98.597 -57.623 50.432
583608 98.628 -57.617 50.402
583609 98.658 -57.611 50.372
583610 98.688 -57.606 50.340
583611 98.718 -57.600 50.308
583612 98.748 -57.595 50.277
583613 98.778 -57.589 50.245
583614 98.814 -57.590 50.213
583615 98.852 -57.592 50.181
583616 98.883 -57.600 50.149
583617 98.914 -57.608 50.118
583618 98.944 -57.615 50.088
583619 98.974 -57.623 50.058
583620 99.004 -57.631 50.029
583621 99.028 -57.636 49.998
583622 99.052 -57.642 49.967
583623 99.076 -57.647 49.937
583624 99.100 -57.651 49.904
583625 99.124 -57.657 49.871
583626 99.148 -57.661 49.838
583627 99.171 -57.667 49.806
583628 99.196 -57.672 49.772
583629 99.210 -57.678 49.738
583630 99.221 -57.685 49.703
583631 99.233 -57.691 49.669
583632 99.261 -57.699 49.635
583633 99.289 -57.705 49.601
583634 99.317 -57.713 49.568
583635 99.346 -57.720 49.536
583636 99.375 -57.727 49.503
583637 99.409 -57.736 49.470
583638 99.446 -57.746 49.435
583639 99.490 -57.756 49.397
583640 99.542 -57.763 49.360
583641 99.585 -57.771 49.324
583642 99.620 -57.779 49.288
583643 99.654 -57.788 49.253
583644 99.689 -57.796 49.218
583645 99.723 -57.805 49.184
583646 99.758 -57.814 49.152
583647 99.792 -57.823 49.121
583648 99.827 -57.831 49.092
583649 99.861 -57.840 49.063
583650 99.896 -57.849 49.037
583651 99.930 -57.858 49.010
583652 99.965 -57.867 48.986
583653 99.999 -57.875 48.964
583654 100.034 -57.884 48.941
583997 88.173 -54.067 57.572
583998 88.208 -54.082 57.574
583999 88.243 -54.098 57.575
584000 88.278 -54.113 57.577
584001 88.314 -54.129 57.578
584002 88.349 -54.144 57.580
584003 88.385 -54.159 57.580
584004 88.420 -54.174 57.580
584005 88.455 -54.190 57.580
584006 88.491 -54.205 57.578
584007 88.526 -54.220 57.578
584008 88.561 -54.236 57.576
584009 88.597 -54.251 57.575
584010 88.632 -54.267 57.574
584011 88.680 -54.245 57.573
584012 88.735 -54.235 57.576
584013 88.778 -54.255 57.580
584014 88.813 -54.291 57.577
584015 88.851 -54.308 57.567
584016 88.889 -54.324 57.552
584017 88.927 -54.342 57.517
584018 88.965 -54.358 57.472
584019 89.003 -54.375 57.424
584020 89.041 -54.392 57.378
584021 89.080 -54.408 57.332
584022 89.118 -54.425 57.283
584023 89.156 -54.442 57.234
584024 89.194 -54.459 57.186
584025 89.232 -54.476 57.137
584026 89.270 -54.493 57.088
584027 89.308 -54.509 57.040
584028 89.346 -54.527 56.991
584029 89.384 -54.543 56.944
584030 89.422 -54.560 56.899
584047 90.079 -54.957 56.079
584048 90.117 -54.980 56.050
584049 90.156 -55.003 56.024
584050 90.194 -55.027 55.999
584051 90.227 -55.047 55.976
584052 90.250 -55.060 55.953
584053 90.273 -55.075 55.932
584054 90.296 -55.089 55.913
584055 90.319 -55.103 55.895
584056 90.342 -55.117 55.880
584057 90.366 -55.131 55.866
584058 90.389 -55.145 55.855
584059 90.412 -55.160 55.846
584060 90.436 -55.173 55.839
584061 90.459 -55.188 55.834
584062 90.482 -55.201 55.831
584063 90.505 -55.216 55.832
584064 90.529 -55.230 55.840
584065 90.553 -55.244 55.850
584066 90.579 -55.255 55.861
584067 90.605 -55.265 55.872
584068 90.632 -55.276 55.882
584069 90.658 -55.285 55.891
584070 90.684 -55.296 55.897
584071 90.719 -55.312 55.901
584072 90.754 -55.328 55.908
584073 90.790 -55.343 55.919
584074 90.833 -55.356 55.932
584075 90.878 -55.367 55.947
584076 90.923 -55.378 55.961
584077 90.968 -55.389 55.970
584078 91.013 -55.401 55.972
584079 91.058 -55.412 55.967
584080 91.103 -55.423 55.959
584081 91.148 -55.435 55.952
584082 91.192 -55.446 55.949
584083 91.238 -55.457 55.945
584084 91.283 -55.468 55.933
584085 91.327 -55.480 55.912
584086 91.373 -55.491 55.884
584087 91.418 -55.502 55.852
584088 91.462 -55.514 55.823
584089 91.502 -55.526 55.795
584090 91.536 -55.539 55.766
584091 91.572 -55.551 55.736
584092 91.607 -55.564 55.705
584093 91.642 -55.577 55.676
584094 91.674 -55.589 55.644
584095 91.705 -55.601 55.608
584096 91.735 -55.612 55.571
584097 91.765 -55.624 55.532
584098 91.795 -55.636 55.493
584099 91.825 -55.648 55.456
584100 91.856 -55.659 55.422
584101 91.886 -55.671 55.390
584102 91.917 -55.683 55.360
584103 91.947 -55.695 55.332
584104 91.977 -55.706 55.305
584105 92.007 -55.718 55.280
584106 92.039 -55.730 55.258
584107 92.072 -55.743 55.238
584108 92.104 -55.756 55.220
584109 92.136 -55.771 55.204
584110 92.166 -55.797 55.189
584111 92.197 -55.823 55.177
584112 92.227 -55.850 55.169
584113 92.258 -55.875 55.164
584114 92.289 -55.901 55.163
584115 92.319 -55.927 55.164
584116 92.351 -55.950 55.169
584117 92.383 -55.974 55.176
584118 92.416 -55.997 55.180
584119 92.448 -56.021 55.181
584120 92.480 -56.044 55.181
584121 92.513 -56.063 55.184
584122 92.548 -56.077 55.195
584123 92.582 -56.091 55.208
584124 92.616 -56.106 55.220
584125 92.650 -56.120 55.229
584126 92.685 -56.135 55.235
584127 92.720 -56.150 55.243
584128 92.756 -56.165 55.249
584129 92.792 -56.181 55.267
584130 92.829 -56.196 55.287
584131 92.865 -56.212 55.305
584132 92.901 -56.224 55.320
584133 92.936 -56.229 55.330
584134 92.973 -56.233 55.338
584135 93.008 -56.238 55.349
584136 93.044 -56.241 55.358
584137 93.086 -56.254 55.362
584138 93.129 -56.267 55.365
584139 93.172 -56.281 55.365
584140 93.215 -56.295 55.364
584141 93.258 -56.308 55.364
584142 93.301 -56.322 55.361
584143 93.344 -56.335 55.355
584144 93.381 -56.346 55.346
584145 93.410 -56.353 55.339
584146 93.443 -56.365 55.325
584147 93.476 -56.376 55.309
584148 93.508 -56.389 55.291
584149 93.541 -56.401 55.268
584150 93.573 -56.413 55.243
584151 93.606 -56.424 55.217
584152 93.639 -56.436 55.185
584153 93.672 -56.448 55.152
584154 93.704 -56.460 55.117
584155 93.737 -56.472 55.079
584156 93.770 -56.484 55.041
584157 93.802 -56.496 55.002
584158 93.834 -56.513 54.964
584159 93.866 -56.530 54.930
584160 93.899 -56.547 54.897
584161 93.930 -56.564 54.862
584162 93.963 -56.581 54.824
584163 93.995 -56.598 54.785
584164 94.027 -56.614 54.742
584165 94.059 -56.631 54.696
584166 94.091 -56.649 54.648
584167 94.121 -56.664 54.599
584168 94.151 -56.680 54.554
584169 94.182 -56.696 54.513
584170 94.212 -56.711 54.476
584171 94.242 -56.727 54.441
584172 94.272 -56.740 54.406
584173 94.300 -56.749 54.369
584174 94.328 -56.758 54.334
584175 94.358 -56.767 54.299
584176 94.394 -56.777 54.264
584177 94.430 -56.786 54.226
584178 94.466 -56.796 54.186
584179 94.495 -56.810 54.143
584180 94.525 -56.825 54.101
584181 94.554 -56.840 54.059
584182 94.583 -56.855 54.017
584183 94.613 -56.869 53.975
584184 94.642 -56.884 53.934
584185 94.671 -56.899 53.894
584186 94.701 -56.914 53.854
584187 94.730 -56.929 53.814
584188 94.759 -56.943 53.770
584189 94.789 -56.959 53.720
584190 94.818 -56.973 53.670
584191 94.847 -56.988 53.627
584192 94.877 -57.003 53.595
584193 94.906 -57.018 53.572
584194 94.935 -57.032 53.548
584195 94.965 -57.047 53.522
584196 94.994 -57.062 53.495
584197 95.024 -57.077 53.467
584198 95.054 -57.091 53.440
584199 95.087 -57.103 53.410
584200 95.122 -57.111 53.379
584201 95.161 -57.114 53.347
584202 95.199 -57.118 53.316
584203 95.238 -57.122 53.289
584204 95.277 -57.125 53.266
584205 95.315 -57.128 53.238
584206 95.354 -57.137 53.205
584207 95.393 -57.145 53.170
584208 95.431 -57.154 53.133
584209 95.470 -57.163 53.094
584210 95.509 -57.171 53.049
584211 95.548 -57.179 53.002
584212 95.587 -57.188 52.956
584213 95.626 -57.196 52.913
584214 95.665 -57.204 52.876
584215 95.703 -57.213 52.845
584216 95.742 -57.222 52.817
584217 95.781 -57.230 52.790
584218 95.818 -57.238 52.764
584219 95.844 -57.244 52.738
584220 95.866 -57.251 52.709
584221 95.887 -57.258 52.677
584222 95.909 -57.265 52.638
584223 95.931 -57.272 52.596
584224 95.952 -57.278 52.555
584225 95.974 -57.285 52.515
584226 95.995 -57.291 52.475
584227 96.017 -57.298 52.437
584228 96.039 -57.305 52.400
584229 96.060 -57.311 52.364
584230 96.081 -57.316 52.331
584231 96.102 -57.322 52.300
584232 96.123 -57.328 52.270
584233 96.150 -57.335 52.243
584234 96.179 -57.342 52.217
584235 96.208 -57.350 52.192
584236 96.237 -57.357 52.166
584237 96.266 -57.365 52.137
584238 96.295 -57.373 52.110
584239 96.324 -57.379 52.087
584240 96.352 -57.386 52.064
584241 96.380 -57.392 52.043
584242 96.409 -57.398 52.024
584243 96.437 -57.405 52.004
584244 96.465 -57.411 51.986
584245 96.495 -57.417 51.971
584246 96.524 -57.423 51.960
584247 96.554 -57.428 51.955
584248 96.584 -57.434 51.948
584249 96.616 -57.438 51.938
584250 96.649 -57.443 51.924
584251 96.681 -57.448 51.904
584252 96.713 -57.452 51.884
584253 96.745 -57.457 51.875
584254 96.778 -57.461 51.870
584255 96.813 -57.467 51.865
584256 96.849 -57.472 51.856
584257 96.884 -57.477 51.853
584258 96.920 -57.482 51.852
584259 96.956 -57.488 51.851
584260 96.992 -57.493 51.851
584261 97.028 -57.494 51.852
584262 97.066 -57.492 51.854
584263 97.102 -57.489 51.859
584264 97.140 -57.487 51.864
584265 97.176 -57.485 51.863
584266 97.207 -57.482 51.862
584267 97.234 -57.479 51.860
584268 97.261 -57.476 51.857
584269 97.289 -57.472 51.848
584270 97.316 -57.469 51.838
584271 97.344 -57.466 51.825
584272 97.371 -57.463 51.811
584273 97.398 -57.460 51.794
584274 97.426 -57.456 51.774
584275 97.453 -57.453 51.753
584276 97.481 -57.452 51.728
584277 97.516 -57.467 51.694
584278 97.550 -57.481 51.653
584279 97.583 -57.492 51.609
584280 97.613 -57.499 51.562
584281 97.644 -57.507 51.512
584282 97.674 -57.514 51.461
584283 97.705 -57.521 51.409
584284 97.735 -57.529 51.356
584285 97.766 -57.536 51.303
584286 97.797 -57.543 51.251
584287 97.826 -57.545 51.208
584288 97.856 -57.546 51.170
584289 97.886 -57.548 51.132
584290 97.916 -57.550 51.094
584291 97.946 -57.551 51.055
584292 97.976 -57.553 51.016
584293 98.010 -57.556 50.975
584294 98.045 -57.561 50.935
584295 98.081 -57.564 50.893
584296 98.117 -57.569 50.850
584297 98.153 -57.573 50.808
584298 98.189 -57.577 50.769
584299 98.224 -57.581 50.731
584300 98.263 -57.585 50.695
584301 98.308 -57.588 50.660
584302 98.353 -57.591 50.626
584303 98.398 -57.593 50.594
584304 98.443 -57.596 50.563
584305 98.488 -57.599 50.529
584306 98.533 -57.602 50.497
584307 98.568 -57.599 50.467
584308 98.598 -57.593 50.437
584309 98.628 -57.587 50.408
584310 98.658 -57.582 50.379
584311 98.688 -57.576 50.346
584312 98.718 -57.570 50.314
584313 98.748 -57.564 50.282
584314 98.780 -57.560 50.250
584315 98.817 -57.563 50.218
584316 98.855 -57.565 50.186
584317 98.887 -57.572 50.155
584318 98.918 -57.579 50.124
584319 98.947 -57.588 50.093
584320 98.978 -57.595 50.064
584321 99.008 -57.604 50.034
584322 99.038 -57.611 50.004
584323 99.063 -57.617 49.973
584324 99.087 -57.622 49.941
584325 99.111 -57.627 49.909
584326 99.135 -57.632 49.876
584327 99.159 -57.637 49.843
584328 99.183 -57.642 49.811
584329 99.199 -57.648 49.777
584330 99.211 -57.655 49.742
584331 99.222 -57.661 49.707
584332 99.233 -57.668 49.672
584333 99.262 -57.675 49.637
584334 99.291 -57.683 49.602
584335 99.320 -57.689 49.568
584336 99.348 -57.697 49.536
584337 99.376 -57.704 49.503
584338 99.405 -57.711 49.470
584339 99.436 -57.719 49.435
584340 99.476 -57.730 49.397
584341 99.527 -57.736 49.360
584342 99.578 -57.742 49.322
584343 99.613 -57.752 49.285
584344 99.647 -57.760 49.248
584345 99.682 -57.769 49.213
584346 99.716 -57.777 49.179
584347 99.751 -57.786 49.147
584348 99.785 -57.795 49.117
584349 99.820 -57.804 49.088
584350 99.854 -57.812 49.060
584351 99.889 -57.821 49.033
584352 99.923 -57.830 49.006
584353 99.958 -57.839 48.984
584354 99.992 -57.848 48.962
584355 100.027 -57.856 48.940
584356 100.061 -57.865 48.918
584698 88.197 -54.032 57.569
584699 88.233 -54.048 57.571
584700 88.268 -54.063 57.572
584701 88.303 -54.078 57.574
584702 88.338 -54.094 57.576
584703 88.374 -54.109 57.577
584704 88.409 -54.124 57.577
584705 88.445 -54.139 57.577
584706 88.480 -54.154 57.578
584707 88.516 -54.170 57.576
584708 88.551 -54.185 57.575
584709 88.586 -54.201 57.574
584710 88.621 -54.216 57.572
584711 88.657 -54.231 57.571
584712 88.695 -54.239 57.570
584713 88.751 -54.220 57.572
584714 88.800 -54.230 57.573
584715 88.835 -54.266 57.572
584716 88.873 -54.284 57.568
584717 88.911 -54.301 57.555
584718 88.949 -54.318 57.532
584719 88.987 -54.335 57.487
584720 89.025 -54.352 57.435
584721 89.063 -54.368 57.386
584722 89.101 -54.385 57.339
584723 89.139 -54.402 57.290
584724 89.178 -54.419 57.241
584725 89.215 -54.436 57.192
584726 89.254 -54.452 57.143
584727 89.292 -54.469 57.094
584728 89.330 -54.486 57.046
584729 89.368 -54.503 56.996
584730 89.406 -54.520 56.949
584731 89.445 -54.536 56.903
584732 89.482 -54.553 56.859
584747 90.061 -54.902 56.117
584748 90.100 -54.925 56.087
584749 90.139 -54.948 56.060
584750 90.177 -54.972 56.034
584751 90.216 -54.995 56.009
584752 90.245 -55.013 55.986
584753 90.268 -55.027 55.964
584754 90.292 -55.041 55.943
584755 90.315 -55.056 55.924
584756 90.338 -55.069 55.907
584757 90.362 -55.084 55.892
584758 90.385 -55.098 55.877
584759 90.408 -55.112 55.864
584760 90.432 -55.126 55.855
584761 90.455 -55.140 55.847
584762 90.478 -55.154 55.841
584763 90.501 -55.169 55.837
584764 90.524 -55.182 55.836
584765 90.549 -55.197 55.842
584766 90.575 -55.210 55.849
584767 90.601 -55.222 55.857
584768 90.627 -55.232 55.864
584769 90.653 -55.242 55.871
584770 90.679 -55.253 55.878
584771 90.711 -55.267 55.884
584772 90.746 -55.282 55.890
584773 90.782 -55.298 55.900
584774 90.817 -55.314 55.914
584775 90.853 -55.329 55.928
584776 90.891 -55.344 55.942
584777 90.936 -55.354 55.954
584778 90.980 -55.366 55.960
584779 91.026 -55.377 55.960
584780 91.071 -55.389 55.954
584781 91.115 -55.400 55.942
584782 91.161 -55.411 55.932
584783 91.206 -55.423 55.929
584784 91.250 -55.433 55.924
584785 91.296 -55.445 55.912
584786 91.341 -55.456 55.890
584787 91.385 -55.467 55.854
584788 91.431 -55.479 55.818
584789 91.475 -55.490 55.790
584790 91.510 -55.503 55.764
584791 91.545 -55.515 55.738
584792 91.580 -55.528 55.711
584793 91.615 -55.541 55.680
584794 91.651 -55.554 55.647
584795 91.685 -55.567 55.614
584796 91.721 -55.579 55.580
584797 91.753 -55.591 55.545
584798 91.783 -55.603 55.510
584799 91.813 -55.615 55.474
584800 91.844 -55.627 55.440
584801 91.874 -55.639 55.408
584802 91.904 -55.650 55.380
584803 91.934 -55.662 55.352
584804 91.965 -55.674 55.326
584805 91.995 -55.686 55.302
584806 92.025 -55.697 55.279
584807 92.057 -55.710 55.258
584808 92.090 -55.722 55.239
584809 92.122 -55.735 55.221
584810 92.155 -55.748 55.206
584811 92.186 -55.771 55.191
584812 92.216 -55.797 55.178
584813 92.246 -55.823 55.168
584814 92.277 -55.849 55.161
584815 92.307 -55.875 55.158
584816 92.338 -55.901 55.156
584817 92.368 -55.927 55.159
584818 92.400 -55.950 55.164
584819 92.432 -55.974 55.167
584820 92.464 -55.997 55.167
584821 92.497 -56.021 55.169
584822 92.530 -56.038 55.176
584823 92.565 -56.052 55.189
584824 92.599 -56.066 55.203
584825 92.634 -56.081 55.217
584826 92.668 -56.096 55.227
584827 92.704 -56.111 55.234
584828 92.740 -56.127 55.239
584829 92.777 -56.142 55.242
584830 92.813 -56.158 55.250
584831 92.849 -56.173 55.264
584832 92.885 -56.188 55.281
584833 92.922 -56.201 55.294
584834 92.957 -56.206 55.306
584835 92.993 -56.210 55.319
584836 93.029 -56.215 55.334
584837 93.065 -56.219 55.346
584838 93.101 -56.224 55.352
584839 93.137 -56.229 55.355
584840 93.181 -56.243 55.358
584841 93.223 -56.257 55.358
584842 93.267 -56.270 55.357
584843 93.309 -56.284 55.355
584844 93.353 -56.298 55.350
584845 93.392 -56.309 55.343
584846 93.422 -56.317 55.340
584847 93.454 -56.329 55.327
584848 93.487 -56.340 55.313
584849 93.519 -56.352 55.297
584850 93.552 -56.364 55.275
584851 93.585 -56.376 55.248
584852 93.618 -56.388 55.225
584853 93.650 -56.400 55.192
584854 93.683 -56.411 55.160
584855 93.716 -56.423 55.126
584856 93.748 -56.439 55.088
584857 93.780 -56.456 55.050
584858 93.812 -56.473 55.012
584859 93.844 -56.489 54.973
584860 93.876 -56.507 54.934
584861 93.908 -56.523 54.894
584862 93.940 -56.540 54.854
584863 93.973 -56.558 54.816
584864 94.004 -56.574 54.778
584865 94.037 -56.591 54.739
584866 94.068 -56.608 54.699
584867 94.099 -56.624 54.658
584868 94.129 -56.639 54.614
584869 94.159 -56.655 54.569
584870 94.190 -56.671 54.522
584871 94.220 -56.686 54.476
584872 94.250 -56.702 54.439
584873 94.279 -56.714 54.405
584874 94.308 -56.723 54.370
584875 94.336 -56.732 54.335
584876 94.364 -56.741 54.300
584877 94.399 -56.751 54.263
584878 94.434 -56.760 54.226
584879 94.470 -56.770 54.186
584880 94.506 -56.780 54.142
584881 94.535 -56.794 54.099
584882 94.564 -56.809 54.057
584883 94.594 -56.824 54.016
584884 94.623 -56.839 53.974
584885 94.652 -56.853 53.933
584886 94.682 -56.868 53.892
584887 94.711 -56.883 53.853
584888 94.741 -56.898 53.812
584889 94.770 -56.913 53.769
584890 94.800 -56.928 53.721
584891 94.829 -56.943 53.668
584892 94.858 -56.957 53.616
584893 94.888 -56.972 53.577
584894 94.917 -56.987 53.547
584895 94.947 -57.002 53.521
584896 94.976 -57.016 53.496
584897 95.005 -57.031 53.469
584898 95.035 -57.046 53.441
584899 95.064 -57.061 53.412
584900 95.093 -57.075 53.382
584901 95.128 -57.081 53.349
584902 95.167 -57.084 53.315
584903 95.205 -57.087 53.282
584904 95.245 -57.091 53.253
584905 95.283 -57.094 53.228
584906 95.323 -57.097 53.200
584907 95.357 -57.106 53.168
584908 95.396 -57.114 53.132
584909 95.435 -57.122 53.094
584910 95.474 -57.131 53.055
584911 95.513 -57.139 53.012
584912 95.552 -57.147 52.966
584913 95.591 -57.156 52.922
584914 95.629 -57.164 52.882
584915 95.668 -57.172 52.848
584916 95.707 -57.181 52.817
584917 95.746 -57.190 52.790
584918 95.784 -57.198 52.764
584919 95.821 -57.206 52.738
584920 95.848 -57.213 52.711
584921 95.870 -57.219 52.682
584922 95.891 -57.226 52.649
584923 95.913 -57.232 52.610
584924 95.934 -57.239 52.567
584925 95.956 -57.246 52.525
584926 95.977 -57.253 52.485
584927 95.999 -57.260 52.446
584928 96.021 -57.266 52.410
584929 96.042 -57.273 52.375
584930 96.064 -57.280 52.340
584931 96.086 -57.287 52.308
584932 96.107 -57.294 52.279
584933 96.128 -57.299 52.252
584934 96.152 -57.305 52.224
584935 96.181 -57.313 52.199
584936 96.210 -57.321 52.174
584937 96.239 -57.327 52.149
584938 96.267 -57.333 52.123
584939 96.296 -57.340 52.096
584940 96.324 -57.346 52.074
584941 96.352 -57.352 52.052
584942 96.381 -57.359 52.033
584943 96.409 -57.365 52.012
584944 96.437 -57.371 51.990
584945 96.467 -57.377 51.967
584946 96.496 -57.382 51.947
584947 96.526 -57.388 51.932
584948 96.555 -57.394 51.922
584949 96.585 -57.399 51.913
584950 96.618 -57.404 51.903
584951 96.650 -57.409 51.891
584952 96.682 -57.413 51.873
584953 96.715 -57.418 51.855
584954 96.747 -57.423 51.843
584955 96.780 -57.427 51.839
584956 96.812 -57.432 51.837
584957 96.844 -57.436 51.833
584958 96.879 -57.442 51.829
584959 96.914 -57.447 51.823
584960 96.951 -57.452 51.818
584961 96.986 -57.457 51.813
584962 97.023 -57.455 51.809
584963 97.061 -57.453 51.807
584964 97.097 -57.451 51.808
584965 97.135 -57.448 51.813
584966 97.171 -57.446 51.814
584967 97.209 -57.444 51.813
584968 97.237 -57.441 51.811
584969 97.264 -57.438 51.807
584970 97.292 -57.435 51.800
584971 97.319 -57.432 51.793
584972 97.346 -57.428 51.784
584973 97.374 -57.425 51.772
584974 97.401 -57.422 51.755
584975 97.428 -57.419 51.731
584976 97.456 -57.416 51.710
584977 97.484 -57.413 51.687
584978 97.518 -57.428 51.657
584979 97.552 -57.442 51.619
584980 97.585 -57.452 51.576
584981 97.615 -57.460 51.529
584982 97.646 -57.467 51.479
584983 97.676 -57.474 51.429
584984 97.707 -57.482 51.378
584985 97.737 -57.489 51.328
584986 97.768 -57.496 51.278
584987 97.798 -57.503 51.229
584988 97.829 -57.509 51.187
584989 97.859 -57.510 51.151
584990 97.888 -57.513 51.116
584991 97.919 -57.514 51.081
584992 97.948 -57.516 51.045
584993 97.979 -57.517 51.008
584994 98.014 -57.522 50.970
584995 98.050 -57.526 50.932
584996 98.085 -57.530 50.892
584997 98.121 -57.534 50.850
584998 98.157 -57.539 50.810
584999 98.193 -57.542 50.772
585000 98.228 -57.547 50.736
585001 98.264 -57.551 50.699
585002 98.302 -57.554 50.662
585003 98.347 -57.558 50.628
585004 98.392 -57.561 50.595
585005 98.437 -57.563 50.564
585006 98.483 -57.566 50.529
585007 98.528 -57.569 50.498
585008 98.569 -57.569 50.468
585009 98.598 -57.564 50.440
585010 98.629 -57.558 50.411
585011 98.659 -57.551 50.382
585012 98.689 -57.546 50.349
585013 98.719 -57.540 50.317
585014 98.749 -57.535 50.284
585015 98.783 -57.532 50.252
585016 98.820 -57.536 50.222
585017 98.858 -57.538 50.190
585018 98.891 -57.544 50.160
585019 98.922 -57.551 50.129
585020 98.951 -57.560 50.098
585021 98.982 -57.567 50.069
585022 99.012 -57.576 50.039
585023 99.042 -57.583 50.009
585024 99.072 -57.592 49.978
585025 99.098 -57.598 49.945
585026 99.122 -57.603 49.913
585027 99.146 -57.608 49.882
585028 99.170 -57.613 49.849
585029 99.189 -57.618 49.817
585030 99.200 -57.625 49.783
585031 99.211 -57.632 49.747
585032 99.223 -57.639 49.711
585033 99.235 -57.645 49.675
585034 99.264 -57.652 49.640
585035 99.292 -57.659 49.603
585036 99.321 -57.667 49.569
585037 99.350 -57.673 49.536
585038 99.378 -57.681 49.503
585039 99.407 -57.688 49.469
585040 99.435 -57.695 49.434
585041 99.464 -57.702 49.397
585042 99.512 -57.709 49.360
585043 99.564 -57.716 49.320
585044 99.605 -57.724 49.281
585045 99.639 -57.733 49.244
585046 99.674 -57.741 49.208
585047 99.708 -57.750 49.174
585048 99.743 -57.758 49.142
585049 99.777 -57.767 49.113
585050 99.812 -57.776 49.084
585051 99.846 -57.785 49.056
585052 99.881 -57.793 49.029
585053 99.915 -57.802 49.003
585054 99.950 -57.811 48.981
585055 99.984 -57.820 48.960
585056 100.019 -57.829 48.938
585057 100.053 -57.837 48.916
585400 88.257 -54.013 57.567
585401 88.292 -54.028 57.568
585402 88.328 -54.043 57.570
585403 88.363 -54.058 57.572
585404 88.399 -54.074 57.574
585405 88.434 -54.089 57.575
585406 88.469 -54.104 57.574
585407 88.504 -54.120 57.574
585408 88.540 -54.135 57.573
585409 88.575 -54.151 57.573
585410 88.611 -54.166 57.571
585411 88.646 -54.181 57.569
585412 88.681 -54.196 57.566
585413 88.717 -54.212 57.565
585414 88.768 -54.204 57.565
585415 88.820 -54.210 57.564
585416 88.857 -54.241 57.563
585417 88.895 -54.261 57.559
585418 88.932 -54.278 57.544
585419 88.971 -54.295 57.520
585420 89.009 -54.311 57.485
585421 89.047 -54.328 57.439
585422 89.085 -54.345 57.390
585423 89.123 -54.362 57.344
585424 89.162 -54.379 57.295
585425 89.199 -54.395 57.246
585426 89.238 -54.412 57.198
585427 89.276 -54.429 57.148
585428 89.314 -54.446 57.100
585429 89.352 -54.463 57.051
585430 89.390 -54.480 57.002
585431 89.428 -54.496 56.954
585432 89.466 -54.513 56.908
585433 89.505 -54.530 56.864
585447 90.044 -54.847 56.157
585448 90.083 -54.870 56.126
585449 90.121 -54.893 56.097
585450 90.160 -54.916 56.069
585451 90.198 -54.940 56.044
585452 90.237 -54.963 56.019
585453 90.264 -54.980 55.997
585454 90.288 -54.994 55.975
585455 90.311 -55.008 55.955
585456 90.334 -55.022 55.936
585457 90.358 -55.036 55.919
585458 90.381 -55.050 55.903
585459 90.404 -55.065 55.888
585460 90.427 -55.079 55.876
585461 90.450 -55.093 55.865
585462 90.474 -55.107 55.855
585463 90.497 -55.121 55.848
585464 90.520 -55.135 55.841
585465 90.545 -55.149 55.839
585466 90.571 -55.163 55.841
585467 90.597 -55.176 55.844
585468 90.622 -55.188 55.848
585469 90.649 -55.199 55.852
585470 90.675 -55.209 55.858
585471 90.703 -55.220 55.864
585472 90.738 -55.236 55.872
585473 90.773 -55.252 55.880
585474 90.809 -55.268 55.893
585475 90.844 -55.283 55.908
585476 90.880 -55.299 55.923
585477 90.915 -55.315 55.934
585478 90.950 -55.331 55.942
585479 90.994 -55.342 55.946
585480 91.038 -55.354 55.944
585481 91.084 -55.365 55.936
585482 91.129 -55.376 55.923
585483 91.173 -55.388 55.910
585484 91.219 -55.399 55.903
585485 91.263 -55.411 55.896
585486 91.308 -55.421 55.883
585487 91.354 -55.433 55.860
585488 91.398 -55.444 55.825
585489 91.443 -55.455 55.790
585490 91.484 -55.467 55.762
585491 91.519 -55.480 55.738
585492 91.554 -55.492 55.714
585493 91.589 -55.505 55.690
585494 91.624 -55.518 55.659
585495 91.659 -55.531 55.623
585496 91.695 -55.544 55.589
585497 91.729 -55.556 55.556
585498 91.765 -55.569 55.522
585499 91.799 -55.582 55.489
585500 91.831 -55.594 55.457
585501 91.862 -55.605 55.425
585502 91.892 -55.618 55.396
585503 91.922 -55.629 55.370
585504 91.952 -55.641 55.345
585505 91.983 -55.652 55.321
585506 92.013 -55.665 55.299
585507 92.043 -55.677 55.279
585508 92.076 -55.689 55.261
585509 92.108 -55.702 55.246
585510 92.140 -55.715 55.230
585511 92.173 -55.727 55.215
585512 92.205 -55.746 55.198
585513 92.235 -55.771 55.185
585514 92.266 -55.797 55.173
585515 92.296 -55.823 55.163
585516 92.326 -55.849 55.156
585517 92.357 -55.875 55.151
585518 92.387 -55.901 55.148
585519 92.418 -55.927 55.148
585520 92.449 -55.951 55.150
585521 92.482 -55.975 55.154
585522 92.514 -55.998 55.160
585523 92.548 -56.013 55.169
585524 92.582 -56.028 55.182
585525 92.617 -56.042 55.198
585526 92.653 -56.057 55.212
585527 92.689 -56.073 55.224
585528 92.725 -56.088 55.233
585529 92.762 -56.104 55.237
585530 92.798 -56.119 55.239
585531 92.834 -56.135 55.243
585532 92.870 -56.150 55.251
585533 92.907 -56.166 55.262
585534 92.943 -56.179 55.273
585535 92.979 -56.184 55.285
585536 93.014 -56.188 55.300
585537 93.050 -56.193 55.315
585538 93.086 -56.197 55.327
585539 93.122 -56.201 55.334
585540 93.158 -56.206 55.338
585541 93.194 -56.210 55.340
585542 93.231 -56.218 55.342
585543 93.275 -56.232 55.342
585544 93.318 -56.245 55.341
585545 93.361 -56.259 55.338
585546 93.403 -56.272 55.334
585547 93.433 -56.280 55.332
585548 93.466 -56.291 55.322
585549 93.498 -56.304 55.310
585550 93.531 -56.316 55.296
585551 93.564 -56.327 55.278
585552 93.596 -56.339 55.249
585553 93.629 -56.351 55.227
585554 93.661 -56.365 55.195
585555 93.694 -56.382 55.163
585556 93.726 -56.398 55.130
585557 93.758 -56.416 55.094
585558 93.790 -56.432 55.055
585559 93.822 -56.449 55.018
585560 93.854 -56.467 54.979
585561 93.886 -56.483 54.940
585562 93.918 -56.500 54.900
585563 93.950 -56.517 54.859
585564 93.982 -56.534 54.822
585565 94.014 -56.551 54.784
585566 94.046 -56.567 54.747
585567 94.076 -56.583 54.710
585568 94.107 -56.599 54.672
585569 94.137 -56.615 54.630
585570 94.167 -56.630 54.585
585571 94.198 -56.646 54.537
585572 94.228 -56.662 54.487
585573 94.258 -56.677 54.441
585574 94.287 -56.689 54.402
585575 94.315 -56.698 54.367
585576 94.343 -56.707 54.333
585577 94.371 -56.716 54.298
585578 94.403 -56.725 54.261
585579 94.439 -56.734 54.224
585580 94.475 -56.744 54.184
585581 94.510 -56.754 54.141
585582 94.546 -56.764 54.096
585583 94.575 -56.778 54.054
585584 94.605 -56.793 54.012
585585 94.634 -56.808 53.971
585586 94.664 -56.823 53.928
585587 94.693 -56.837 53.886
585588 94.722 -56.852 53.847
585589 94.752 -56.867 53.806
585590 94.781 -56.882 53.764
585591 94.810 -56.896 53.718
585592 94.840 -56.912 53.667
585593 94.869 -56.927 53.610
585594 94.898 -56.941 53.564
585595 94.928 -56.956 53.526
585596 94.957 -56.971 53.496
585597 94.986 -56.986 53.470
585598 95.016 -57.000 53.443
585599 95.045 -57.015 53.414
585600 95.073 -57.028 53.384
585601 95.102 -57.041 53.352
585602 95.134 -57.050 53.318
585603 95.173 -57.053 53.283
585604 95.212 -57.056 53.248
585605 95.251 -57.059 53.215
585606 95.296 -57.060 53.186
585607 95.332 -57.064 53.157
585608 95.361 -57.074 53.123
585609 95.400 -57.082 53.087
585610 95.439 -57.091 53.049
585611 95.478 -57.099 53.009
585612 95.517 -57.107 52.968
585613 95.555 -57.116 52.927
585614 95.594 -57.124 52.890
585615 95.633 -57.133 52.857
585616 95.672 -57.141 52.825
585617 95.710 -57.150 52.794
585618 95.749 -57.158 52.764
585619 95.788 -57.166 52.736
585620 95.825 -57.173 52.711
585621 95.851 -57.180 52.683
585622 95.873 -57.187 52.651
585623 95.895 -57.194 52.617
585624 95.916 -57.200 52.579
585625 95.938 -57.207 52.538
585626 95.960 -57.214 52.495
585627 95.981 -57.221 52.453
585628 96.003 -57.228 52.414
585629 96.024 -57.235 52.379
585630 96.046 -57.241 52.348
585631 96.067 -57.248 52.316
585632 96.089 -57.255 52.285
585633 96.111 -57.262 52.257
585634 96.132 -57.269 52.232
585635 96.155 -57.276 52.206
585636 96.182 -57.282 52.180
585637 96.210 -57.288 52.155
585638 96.239 -57.294 52.132
585639 96.267 -57.300 52.109
585640 96.296 -57.307 52.084
585641 96.324 -57.313 52.060
585642 96.353 -57.319 52.041
585643 96.381 -57.326 52.023
585644 96.409 -57.332 52.002
585645 96.439 -57.337 51.977
585646 96.468 -57.343 51.951
585647 96.498 -57.348 51.927
585648 96.528 -57.354 51.906
585649 96.557 -57.360 51.891
585650 96.587 -57.365 51.878
585651 96.619 -57.369 51.867
585652 96.652 -57.375 51.855
585653 96.684 -57.379 51.840
585654 96.717 -57.384 51.823
585655 96.749 -57.388 51.810
585656 96.781 -57.393 51.804
585657 96.813 -57.398 51.804
585658 96.846 -57.402 51.803
585659 96.878 -57.407 51.801
585660 96.910 -57.412 51.795
585661 96.945 -57.416 51.789
585662 96.981 -57.419 51.781
585663 97.018 -57.416 51.772
585664 97.055 -57.414 51.763
585665 97.092 -57.413 51.760
585666 97.130 -57.410 51.763
585667 97.166 -57.408 51.764
585668 97.204 -57.406 51.762
585669 97.240 -57.404 51.758
585670 97.267 -57.401 51.753
585671 97.294 -57.397 51.750
585672 97.322 -57.394 51.746
585673 97.349 -57.391 51.739
585674 97.376 -57.388 51.728
585675 97.404 -57.385 51.711
585676 97.431 -57.381 51.688
585677 97.458 -57.378 51.665
585678 97.486 -57.375 51.640
585679 97.520 -57.388 51.613
585680 97.555 -57.403 51.580
585681 97.586 -57.413 51.540
585682 97.617 -57.420 51.494
585683 97.648 -57.427 51.445
585684 97.678 -57.435 51.396
585685 97.709 -57.442 51.346
585686 97.739 -57.449 51.297
585687 97.770 -57.457 51.249
585688 97.800 -57.464 51.202
585689 97.831 -57.471 51.161
585690 97.861 -57.476 51.126
585691 97.891 -57.477 51.092
585692 97.921 -57.479 51.059
585693 97.950 -57.480 51.024
585694 97.983 -57.482 50.989
585695 98.018 -57.487 50.952
585696 98.054 -57.491 50.917
585697 98.090 -57.495 50.880
585698 98.126 -57.499 50.842
585699 98.161 -57.504 50.805
585700 98.197 -57.507 50.770
585701 98.232 -57.512 50.735
585702 98.268 -57.516 50.698
585703 98.304 -57.520 50.661
585704 98.342 -57.524 50.626
585705 98.387 -57.527 50.593
585706 98.432 -57.529 50.561
585707 98.477 -57.532 50.526
585708 98.522 -57.536 50.495
585709 98.567 -57.538 50.466
585710 98.599 -57.533 50.437
585711 98.629 -57.528 50.407
585712 98.659 -57.522 50.378
585713 98.690 -57.516 50.345
585714 98.719 -57.510 50.313
585715 98.749 -57.504 50.282
585716 98.786 -57.505 50.251
585717 98.823 -57.508 50.223
585718 98.861 -57.510 50.192
585719 98.896 -57.516 50.163
585720 98.926 -57.523 50.132
585721 98.955 -57.532 50.102
585722 98.986 -57.539 50.073
585723 99.015 -57.548 50.043
585724 99.046 -57.555 50.013
585725 99.076 -57.564 49.982
585726 99.106 -57.571 49.950
585727 99.133 -57.578 49.918
585728 99.157 -57.583 49.886
585729 99.178 -57.589 49.853
585730 99.189 -57.595 49.822
585731 99.201 -57.601 49.788
585732 99.212 -57.608 49.752
585733 99.224 -57.615 49.715
585734 99.237 -57.622 49.678
585735 99.266 -57.629 49.642
585736 99.294 -57.636 49.605
585737 99.323 -57.643 49.570
585738 99.351 -57.651 49.536
585739 99.380 -57.658 49.501
585740 99.409 -57.665 49.467
585741 99.438 -57.670 49.431
585742 99.467 -57.677 49.394
585743 99.498 -57.682 49.356
585744 99.550 -57.689 49.317
585745 99.597 -57.695 49.277
585746 99.632 -57.705 49.240
585747 99.666 -57.713 49.204
585748 99.701 -57.722 49.170
585749 99.735 -57.731 49.136
585750 99.770 -57.739 49.106
585751 99.804 -57.749 49.078
585752 99.839 -57.757 49.050
585753 99.873 -57.766 49.025
585754 99.908 -57.774 48.999
585755 99.942 -57.783 48.978
585756 99.977 -57.792 48.956
585757 100.011 -57.801 48.935
585758 100.046 -57.810 48.914
586102 88.317 -53.993 57.564
586103 88.352 -54.008 57.566
586104 88.388 -54.023 57.568
586105 88.423 -54.038 57.570
586106 88.459 -54.054 57.571
586107 88.494 -54.070 57.571
586108 88.529 -54.085 57.570
586109 88.564 -54.100 57.569
586110 88.600 -54.115 57.568
586111 88.635 -54.131 57.567
586112 88.671 -54.146 57.564
586113 88.706 -54.161 57.561
586114 88.742 -54.176 57.559
586115 88.784 -54.189 57.557
586116 88.836 -54.194 57.554
586117 88.879 -54.215 57.551
586118 88.916 -54.238 57.545
586119 88.955 -54.254 57.529
586120 88.993 -54.271 57.506
586121 89.031 -54.288 57.476
586122 89.069 -54.305 57.436
586123 89.107 -54.322 57.390
586124 89.145 -54.339 57.344
586125 89.183 -54.355 57.297
586126 89.222 -54.372 57.250
586127 89.259 -54.389 57.201
586128 89.298 -54.406 57.153
586129 89.336 -54.423 57.105
586130 89.374 -54.439 57.056
586131 89.412 -54.456 57.007
586132 89.450 -54.473 56.960
586133 89.488 -54.490 56.914
586134 89.526 -54.507 56.869
586146 89.988 -54.768 56.227
586147 90.027 -54.791 56.195
586148 90.065 -54.815 56.165
586149 90.104 -54.837 56.135
586150 90.143 -54.861 56.106
586151 90.181 -54.884 56.079
586152 90.220 -54.908 56.054
586153 90.258 -54.931 56.030
586154 90.284 -54.947 56.008
586155 90.307 -54.961 55.986
586156 90.330 -54.975 55.967
586157 90.353 -54.989 55.948
586158 90.377 -55.003 55.931
586159 90.400 -55.017 55.914
586160 90.423 -55.032 55.901
586161 90.446 -55.045 55.889
586162 90.469 -55.060 55.876
586163 90.493 -55.074 55.864
586164 90.516 -55.088 55.855
586165 90.541 -55.101 55.848
586166 90.567 -55.116 55.843
586167 90.593 -55.129 55.841
586168 90.618 -55.143 55.841
586169 90.644 -55.155 55.842
586170 90.670 -55.166 55.845
586171 90.696 -55.176 55.849
586172 90.730 -55.191 55.855
586173 90.765 -55.206 55.863
586174 90.801 -55.222 55.875
586175 90.836 -55.238 55.888
586176 90.871 -55.254 55.901
586177 90.907 -55.269 55.915
586178 90.942 -55.285 55.925
586179 90.977 -55.301 55.928
586180 91.013 -55.317 55.929
586181 91.051 -55.330 55.925
586182 91.096 -55.342 55.916
586183 91.142 -55.353 55.903
586184 91.186 -55.364 55.887
586185 91.231 -55.376 55.876
586186 91.277 -55.387 55.867
586187 91.321 -55.398 55.852
586188 91.366 -55.409 55.828
586189 91.412 -55.420 55.794
586190 91.456 -55.432 55.762
586191 91.493 -55.445 55.736
586192 91.528 -55.457 55.715
586193 91.563 -55.470 55.694
586194 91.598 -55.483 55.674
586195 91.633 -55.495 55.645
586196 91.668 -55.508 55.608
586197 91.703 -55.520 55.571
586198 91.738 -55.533 55.537
586199 91.773 -55.546 55.503
586200 91.808 -55.559 55.471
586201 91.843 -55.572 55.441
586202 91.878 -55.584 55.411
586203 91.910 -55.596 55.384
586204 91.940 -55.608 55.360
586205 91.970 -55.620 55.338
586206 92.001 -55.632 55.317
586207 92.031 -55.643 55.297
586208 92.061 -55.655 55.280
586209 92.094 -55.668 55.268
586210 92.126 -55.680 55.258
586211 92.159 -55.693 55.247
586212 92.191 -55.706 55.232
586213 92.223 -55.720 55.214
586214 92.254 -55.746 55.197
586215 92.284 -55.771 55.183
586216 92.315 -55.797 55.172
586217 92.346 -55.823 55.162
586218 92.376 -55.849 55.155
586219 92.407 -55.875 55.151
586220 92.437 -55.901 55.150
586221 92.467 -55.927 55.151
586222 92.498 -55.952 55.152
586223 92.531 -55.974 55.155
586224 92.565 -55.988 55.158
586225 92.601 -56.003 55.170
586226 92.638 -56.019 55.188
586227 92.674 -56.034 55.207
586228 92.710 -56.050 55.221
586229 92.746 -56.066 55.230
586230 92.783 -56.081 55.236
586231 92.819 -56.097 55.238
586232 92.855 -56.112 55.240
586233 92.891 -56.128 55.244
586234 92.927 -56.143 55.249
586235 92.964 -56.157 55.257
586236 92.999 -56.161 55.268
586237 93.036 -56.166 55.283
586238 93.071 -56.170 55.297
586239 93.107 -56.175 55.309
586240 93.143 -56.179 55.318
586241 93.179 -56.184 55.322
586242 93.214 -56.188 55.323
586243 93.251 -56.193 55.324
586244 93.286 -56.197 55.324
586245 93.326 -56.207 55.322
586246 93.369 -56.220 55.319
586247 93.412 -56.234 55.314
586248 93.444 -56.244 55.311
586249 93.477 -56.255 55.304
586250 93.510 -56.267 55.294
586251 93.543 -56.279 55.282
586252 93.575 -56.291 55.265
586253 93.607 -56.307 55.239
586254 93.639 -56.325 55.215
586255 93.672 -56.342 55.188
586256 93.703 -56.358 55.157
586257 93.735 -56.376 55.125
586258 93.768 -56.392 55.091
586259 93.800 -56.409 55.054
586260 93.832 -56.426 55.018
586261 93.864 -56.443 54.982
586262 93.896 -56.460 54.945
586263 93.928 -56.477 54.909
586264 93.960 -56.494 54.871
586265 93.992 -56.511 54.833
586266 94.024 -56.527 54.795
586267 94.054 -56.543 54.758
586268 94.084 -56.559 54.722
586269 94.115 -56.574 54.685
586270 94.145 -56.590 54.642
586271 94.175 -56.606 54.597
586272 94.205 -56.621 54.549
586273 94.236 -56.637 54.497
586274 94.266 -56.653 54.446
586275 94.294 -56.663 54.400
586276 94.323 -56.672 54.361
586277 94.351 -56.681 54.326
586278 94.379 -56.690 54.291
586279 94.408 -56.699 54.254
586280 94.444 -56.708 54.216
586281 94.479 -56.718 54.176
586282 94.515 -56.728 54.132
586283 94.551 -56.737 54.087
586284 94.586 -56.747 54.044
586285 94.615 -56.762 54.001
586286 94.645 -56.777 53.958
586287 94.674 -56.792 53.916
586288 94.703 -56.807 53.875
586289 94.733 -56.821 53.835
586290 94.762 -56.837 53.794
586291 94.791 -56.851 53.753
586292 94.821 -56.866 53.710
586293 94.850 -56.881 53.662
586294 94.880 -56.896 53.609
586295 94.909 -56.910 53.555
586296 94.938 -56.925 53.511
586297 94.968 -56.940 53.475
586298 94.997 -56.955 53.444
586299 95.027 -56.969 53.416
586300 95.054 -56.982 53.386
586301 95.083 -56.995 53.355
586302 95.111 -57.008 53.322
586303 95.140 -57.019 53.287
586304 95.180 -57.022 53.251
586305 95.218 -57.026 53.215
586306 95.258 -57.028 53.178
586307 95.308 -57.025 53.145
586308 95.341 -57.032 53.113
586309 95.369 -57.041 53.079
586310 95.404 -57.050 53.044
586311 95.443 -57.059 53.007
586312 95.481 -57.067 52.967
586313 95.520 -57.075 52.929
586314 95.559 -57.084 52.892
586315 95.598 -57.093 52.860
586316 95.636 -57.101 52.830
586317 95.675 -57.109 52.800
586318 95.714 -57.118 52.770
586319 95.753 -57.126 52.738
586320 95.792 -57.134 52.708
586321 95.828 -57.141 52.680
586322 95.855 -57.148 52.651
586323 95.877 -57.155 52.618
586324 95.898 -57.162 52.583
586325 95.920 -57.169 52.545
586326 95.942 -57.175 52.506
586327 95.963 -57.182 52.465
586328 95.985 -57.189 52.423
586329 96.007 -57.196 52.384
586330 96.028 -57.203 52.351
586331 96.049 -57.210 52.320
586332 96.071 -57.216 52.289
586333 96.093 -57.223 52.259
586334 96.114 -57.230 52.232
586335 96.136 -57.237 52.208
586336 96.159 -57.247 52.186
586337 96.186 -57.254 52.160
586338 96.214 -57.259 52.136
586339 96.241 -57.264 52.115
586340 96.269 -57.269 52.094
586341 96.297 -57.275 52.072
586342 96.325 -57.280 52.048
586343 96.353 -57.285 52.030
586344 96.381 -57.291 52.012
586345 96.410 -57.297 51.991
586346 96.440 -57.303 51.966
586347 96.470 -57.309 51.939
586348 96.500 -57.314 51.910
586349 96.529 -57.319 51.883
586350 96.559 -57.326 51.861
586351 96.589 -57.331 51.844
586352 96.621 -57.335 51.829
586353 96.653 -57.340 51.816
586354 96.686 -57.344 51.804
586355 96.718 -57.350 51.786
586356 96.750 -57.354 51.773
586357 96.783 -57.359 51.768
586358 96.815 -57.363 51.768
586359 96.848 -57.368 51.769
586360 96.880 -57.373 51.769
586361 96.912 -57.377 51.767
586362 96.944 -57.383 51.762
586363 96.977 -57.382 51.752
586364 97.013 -57.379 51.737
586365 97.050 -57.376 51.721
586366 97.087 -57.374 51.714
586367 97.124 -57.372 51.714
586368 97.161 -57.369 51.713
586369 97.198 -57.367 51.710
586370 97.235 -57.365 51.705
586371 97.270 -57.363 51.700
586372 97.297 -57.360 51.698
586373 97.324 -57.357 51.697
586374 97.352 -57.353 51.692
586375 97.379 -57.350 51.681
586376 97.406 -57.347 51.665
586377 97.434 -57.344 51.644
586378 97.461 -57.341 51.618
586379 97.489 -57.337 51.592
586380 97.522 -57.350 51.567
586381 97.557 -57.364 51.538
586382 97.588 -57.373 51.502
586383 97.619 -57.380 51.458
586384 97.649 -57.388 51.411
586385 97.680 -57.395 51.364
586386 97.711 -57.402 51.315
586387 97.741 -57.410 51.267
586388 97.772 -57.416 51.221
586389 97.802 -57.424 51.175
586390 97.832 -57.431 51.135
586391 97.863 -57.438 51.100
586392 97.893 -57.442 51.067
586393 97.923 -57.443 51.034
586394 97.953 -57.445 51.001
586395 97.987 -57.448 50.968
586396 98.022 -57.452 50.933
586397 98.058 -57.457 50.900
586398 98.094 -57.460 50.866
586399 98.130 -57.465 50.832
586400 98.165 -57.469 50.797
586401 98.201 -57.473 50.763
586402 98.237 -57.477 50.728
586403 98.273 -57.482 50.691
586404 98.308 -57.485 50.655
586405 98.344 -57.489 50.620
586406 98.382 -57.494 50.587
586407 98.427 -57.496 50.555
586408 98.472 -57.499 50.521
586409 98.517 -57.502 50.489
586410 98.562 -57.504 50.459
586411 98.600 -57.504 50.429
586412 98.630 -57.498 50.399
586413 98.660 -57.492 50.369
586414 98.690 -57.486 50.337
586415 98.720 -57.480 50.306
586416 98.751 -57.476 50.276
586417 98.789 -57.478 50.248
586418 98.827 -57.481 50.221
586419 98.864 -57.483 50.192
586420 98.900 -57.488 50.163
586421 98.930 -57.495 50.134
586422 98.959 -57.504 50.104
586423 98.990 -57.511 50.075
586424 99.019 -57.520 50.045
586425 99.050 -57.527 50.015
586426 99.079 -57.536 49.984
586427 99.110 -57.543 49.954
586428 99.140 -57.551 49.923
586429 99.167 -57.559 49.891
586430 99.179 -57.565 49.857
586431 99.190 -57.572 49.824
586432 99.202 -57.579 49.790
586433 99.213 -57.586 49.755
586434 99.225 -57.592 49.718
586435 99.238 -57.598 49.681
586436 99.267 -57.606 49.643
586437 99.296 -57.613 49.604
586438 99.325 -57.620 49.568
586439 99.353 -57.627 49.533
586440 99.382 -57.633 49.498
586441 99.412 -57.639 49.462
586442 99.441 -57.645 49.426
586443 99.470 -57.651 49.389
586444 99.498 -57.656 49.351
586445 99.535 -57.662 49.312
586446 99.586 -57.668 49.272
586447 99.624 -57.677 49.234
586448 99.659 -57.686 49.198
586449 99.693 -57.694 49.164
586450 99.728 -57.703 49.131
586451 99.762 -57.712 49.101
586452 99.797 -57.720 49.072
586453 99.831 -57.730 49.045
586454 99.866 -57.738 49.020
586455 99.900 -57.747 48.995
586456 99.934 -57.755 48.974
586457 99.969 -57.764 48.952
586458 100.003 -57.773 48.931
586459 100.038 -57.782 48.912
586460 100.072 -57.790 48.892
586803 88.342 -53.958 57.561
586804 88.377 -53.973 57.563
586805 88.412 -53.988 57.564
586806 88.448 -54.004 57.566
586807 88.483 -54.019 57.568
586808 88.519 -54.035 57.567
586809 88.554 -54.050 57.566
586810 88.589 -54.065 57.565
586811 88.625 -54.080 57.563
586812 88.660 -54.096 57.561
586813 88.695 -54.111 57.559
586814 88.731 -54.126 57.555
586815 88.768 -54.145 57.551
586816 88.808 -54.166 57.547
586817 88.853 -54.179 57.543
586818 88.901 -54.190 57.536
586819 88.939 -54.214 57.527
586820 88.976 -54.231 57.513
586821 89.015 -54.248 57.492
586822 89.053 -54.265 57.462
586823 89.091 -54.282 57.426
586824 89.129 -54.298 57.385
586825 89.167 -54.315 57.343
586826 89.205 -54.332 57.298
586827 89.243 -54.348 57.251
586828 89.281 -54.366 57.203
586829 89.320 -54.383 57.155
586830 89.357 -54.399 57.108
586831 89.396 -54.416 57.060
586832 89.434 -54.433 57.011
586833 89.472 -54.450 56.964
586834 89.510 -54.467 56.918
586846 89.971 -54.712 56.266
586847 90.010 -54.736 56.233
586848 90.048 -54.759 56.203
586849 90.087 -54.783 56.173
586850 90.125 -54.806 56.144
586851 90.164 -54.829 56.115
586852 90.202 -54.853 56.088
586853 90.241 -54.876 56.064
586854 90.279 -54.900 56.041
586855 90.303 -54.913 56.019
586856 90.326 -54.928 55.999
586857 90.349 -54.941 55.980
586858 90.372 -54.956 55.962
586859 90.395 -54.970 55.945
586860 90.418 -54.984 55.929
586861 90.442 -54.998 55.917
586862 90.465 -55.013 55.904
586863 90.488 -55.026 55.889
586864 90.512 -55.041 55.874
586865 90.537 -55.054 55.862
586866 90.563 -55.068 55.854
586867 90.589 -55.082 55.847
586868 90.614 -55.095 55.843
586869 90.640 -55.109 55.840
586870 90.666 -55.122 55.839
586871 90.692 -55.132 55.840
586872 90.722 -55.144 55.843
586873 90.757 -55.160 55.848
586874 90.793 -55.176 55.859
586875 90.828 -55.192 55.871
586876 90.863 -55.207 55.883
586877 90.898 -55.223 55.895
586878 90.934 -55.239 55.906
586879 90.969 -55.255 55.909
586880 91.005 -55.270 55.909
586881 91.040 -55.286 55.908
586882 91.075 -55.302 55.903
586883 91.110 -55.318 55.894
586884 91.154 -55.329 55.880
586885 91.200 -55.341 55.862
586886 91.244 -55.352 55.849
586887 91.289 -55.364 55.837
586888 91.334 -55.375 55.820
586889 91.379 -55.386 55.794
586890 91.424 -55.397 55.763
586891 91.467 -55.409 55.734
586892 91.502 -55.422 55.711
586893 91.537 -55.434 55.694
586894 91.572 -55.447 55.677
586895 91.607 -55.460 55.659
586896 91.642 -55.473 55.632
586897 91.677 -55.485 55.595
586898 91.712 -55.498 55.557
586899 91.747 -55.511 55.520
586900 91.782 -55.524 55.487
586901 91.817 -55.536 55.456
586902 91.852 -55.549 55.427
586903 91.887 -55.561 55.398
586904 91.922 -55.574 55.373
586905 91.957 -55.587 55.351
586906 91.988 -55.599 55.331
586907 92.019 -55.611 55.313
586908 92.049 -55.623 55.298
586909 92.080 -55.634 55.285
586910 92.112 -55.647 55.279
586911 92.144 -55.660 55.277
586912 92.177 -55.672 55.271
586913 92.209 -55.685 55.256
586914 92.242 -55.698 55.235
586915 92.273 -55.720 55.214
586916 92.303 -55.746 55.197
586917 92.334 -55.771 55.184
586918 92.364 -55.797 55.173
586919 92.395 -55.823 55.165
586920 92.425 -55.849 55.160
586921 92.456 -55.875 55.158
586922 92.486 -55.900 55.158
586923 92.517 -55.927 55.159
586924 92.550 -55.947 55.161
586925 92.586 -55.965 55.161
586926 92.622 -55.981 55.165
586927 92.658 -55.997 55.176
586928 92.695 -56.012 55.193
586929 92.731 -56.028 55.209
586930 92.767 -56.043 55.220
586931 92.803 -56.059 55.226
586932 92.840 -56.074 55.230
586933 92.876 -56.089 55.234
586934 92.912 -56.105 55.240
586935 92.948 -56.120 55.247
586936 92.985 -56.135 55.253
586937 93.020 -56.139 55.258
586938 93.056 -56.144 55.269
586939 93.092 -56.148 55.281
586940 93.128 -56.153 55.292
586941 93.164 -56.157 55.302
586942 93.200 -56.162 55.307
586943 93.235 -56.166 55.308
586944 93.271 -56.170 55.308
586945 93.307 -56.175 55.306
586946 93.343 -56.179 55.303
586947 93.379 -56.184 55.300
586948 93.420 -56.195 55.294
586949 93.456 -56.207 55.288
586950 93.489 -56.219 55.282
586951 93.521 -56.234 55.273
586952 93.553 -56.251 55.260
586953 93.585 -56.267 55.242
586954 93.617 -56.285 55.219
586955 93.649 -56.301 55.194
586956 93.681 -56.318 55.168
586957 93.713 -56.335 55.139
586958 93.746 -56.352 55.108
586959 93.777 -56.369 55.074
586960 93.809 -56.386 55.039
586961 93.842 -56.403 55.005
586962 93.874 -56.420 54.972
586963 93.906 -56.437 54.940
586964 93.938 -56.454 54.907
586965 93.970 -56.470 54.873
586966 94.001 -56.487 54.839
586967 94.032 -56.503 54.803
586968 94.062 -56.518 54.765
586969 94.092 -56.534 54.729
586970 94.123 -56.550 54.691
586971 94.153 -56.566 54.648
586972 94.183 -56.581 54.604
586973 94.213 -56.597 54.557
586974 94.244 -56.613 54.506
586975 94.274 -56.628 54.453
586976 94.302 -56.637 54.401
586977 94.330 -56.646 54.353
586978 94.358 -56.655 54.317
586979 94.387 -56.664 54.281
586980 94.415 -56.674 54.244
586981 94.449 -56.683 54.205
586982 94.484 -56.693 54.165
586983 94.520 -56.702 54.121
586984 94.556 -56.711 54.077
586985 94.592 -56.721 54.033
586986 94.626 -56.731 53.988
586987 94.655 -56.746 53.945
586988 94.685 -56.761 53.902
586989 94.714 -56.776 53.860
586990 94.744 -56.791 53.820
586991 94.773 -56.805 53.781
586992 94.802 -56.821 53.742
586993 94.832 -56.835 53.702
586994 94.861 -56.850 53.659
586995 94.890 -56.865 53.612
586996 94.920 -56.880 53.555
586997 94.949 -56.894 53.502
586998 94.978 -56.909 53.457
586999 95.007 -56.924 53.421
587000 95.036 -56.936 53.389
587001 95.064 -56.949 53.359
587002 95.092 -56.962 53.326
587003 95.120 -56.975 53.292
587004 95.148 -56.987 53.257
587005 95.186 -56.992 53.220
587006 95.225 -56.995 53.182
587007 95.271 -56.994 53.144
587008 95.321 -56.991 53.108
587009 95.350 -57.000 53.073
587010 95.377 -57.009 53.040
587011 95.407 -57.018 53.007
587012 95.446 -57.027 52.970
587013 95.485 -57.035 52.932
587014 95.524 -57.044 52.895
587015 95.562 -57.053 52.862
587016 95.601 -57.061 52.831
587017 95.640 -57.069 52.802
587018 95.679 -57.078 52.775
587019 95.718 -57.086 52.746
587020 95.757 -57.094 52.714
587021 95.795 -57.103 52.680
587022 95.831 -57.109 52.648
587023 95.859 -57.116 52.617
587024 95.880 -57.123 52.584
587025 95.902 -57.130 52.549
587026 95.923 -57.137 52.511
587027 95.945 -57.144 52.474
587028 95.967 -57.150 52.434
587029 95.988 -57.157 52.394
587030 96.010 -57.164 52.355
587031 96.032 -57.171 52.322
587032 96.053 -57.178 52.292
587033 96.075 -57.185 52.263
587034 96.096 -57.191 52.234
587035 96.118 -57.197 52.207
587036 96.141 -57.207 52.182
587037 96.164 -57.217 52.162
587038 96.189 -57.225 52.137
587039 96.217 -57.231 52.114
587040 96.245 -57.236 52.093
587041 96.272 -57.241 52.074
587042 96.300 -57.247 52.053
587043 96.328 -57.252 52.034
587044 96.356 -57.257 52.016
587045 96.384 -57.261 51.997
587046 96.412 -57.264 51.975
587047 96.442 -57.269 51.948
587048 96.471 -57.274 51.918
587049 96.501 -57.280 51.891
587050 96.531 -57.285 51.866
587051 96.560 -57.291 51.844
587052 96.590 -57.297 51.822
587053 96.622 -57.301 51.800
587054 96.655 -57.306 51.782
587055 96.687 -57.310 51.765
587056 96.720 -57.315 51.751
587057 96.752 -57.320 51.740
587058 96.784 -57.325 51.733
587059 96.817 -57.329 51.731
587060 96.849 -57.334 51.730
587061 96.881 -57.340 51.734
587062 96.913 -57.347 51.731
587063 96.946 -57.354 51.723
587064 96.978 -57.353 51.709
587065 97.011 -57.344 51.693
587066 97.045 -57.338 51.679
587067 97.082 -57.335 51.669
587068 97.119 -57.333 51.664
587069 97.156 -57.332 51.662
587070 97.193 -57.329 51.658
587071 97.230 -57.327 51.651
587072 97.267 -57.325 51.647
587073 97.300 -57.322 51.645
587074 97.327 -57.319 51.645
587075 97.355 -57.316 51.641
587076 97.382 -57.313 51.631
587077 97.409 -57.309 51.615
587078 97.437 -57.306 51.596
587079 97.464 -57.303 51.571
587080 97.491 -57.300 51.544
587081 97.525 -57.311 51.520
587082 97.559 -57.326 51.494
587083 97.590 -57.333 51.461
587084 97.620 -57.341 51.421
587085 97.651 -57.348 51.378
587086 97.681 -57.355 51.333
587087 97.712 -57.363 51.286
587088 97.743 -57.369 51.238
587089 97.773 -57.377 51.192
587090 97.804 -57.385 51.147
587091 97.835 -57.391 51.108
587092 97.865 -57.399 51.074
587093 97.895 -57.406 51.040
587094 97.925 -57.407 51.007
587095 97.955 -57.409 50.976
587096 97.991 -57.413 50.945
587097 98.027 -57.417 50.914
587098 98.063 -57.422 50.882
587099 98.098 -57.426 50.851
587100 98.134 -57.430 50.820
587101 98.169 -57.434 50.787
587102 98.205 -57.438 50.753
587103 98.241 -57.442 50.718
587104 98.277 -57.447 50.682
587105 98.312 -57.451 50.646
587106 98.348 -57.455 50.612
587107 98.384 -57.459 50.579
587108 98.421 -57.463 50.547
587109 98.466 -57.466 50.514
587110 98.511 -57.469 50.481
587111 98.557 -57.471 50.450
587112 98.600 -57.473 50.420
587113 98.631 -57.468 50.390
587114 98.660 -57.462 50.359
587115 98.691 -57.456 50.328
587116 98.721 -57.451 50.298
587117 98.754 -57.448 50.271
587118 98.792 -57.451 50.244
587119 98.830 -57.454 50.219
587120 98.867 -57.456 50.192
587121 98.904 -57.460 50.163
587122 98.933 -57.467 50.134
587123 98.963 -57.476 50.105
587124 98.994 -57.483 50.076
587125 99.023 -57.492 50.046
587126 99.053 -57.499 50.018
587127 99.080 -57.507 49.986
587128 99.106 -57.515 49.958
587129 99.134 -57.523 49.927
587130 99.160 -57.535 49.894
587131 99.176 -57.542 49.860
587132 99.190 -57.549 49.825
587133 99.203 -57.555 49.791
587134 99.214 -57.562 49.756
587135 99.226 -57.569 49.720
587136 99.240 -57.576 49.682
587137 99.269 -57.583 49.643
587138 99.298 -57.589 49.604
587139 99.327 -57.595 49.566
587140 99.356 -57.601 49.530
587141 99.386 -57.607 49.494
587142 99.415 -57.613 49.458
587143 99.444 -57.619 49.421
587144 99.473 -57.625 49.384
587145 99.501 -57.630 49.346
587146 99.528 -57.636 49.307
587147 99.572 -57.642 49.268
587148 99.617 -57.649 49.230
587149 99.651 -57.658 49.193
587150 99.686 -57.667 49.158
587151 99.720 -57.675 49.126
587152 99.755 -57.684 49.096
587153 99.789 -57.692 49.068
587154 99.824 -57.702 49.040
587155 99.858 -57.711 49.015
587156 99.893 -57.719 48.990
587157 99.927 -57.728 48.969
587158 99.962 -57.736 48.948
587159 99.996 -57.746 48.928
587160 100.031 -57.754 48.908
587161 100.065 -57.763 48.889
587504 88.366 -53.922 57.556
587505 88.402 -53.938 57.558
587506 88.437 -53.954 57.560
587507 88.473 -53.969 57.562
587508 88.508 -53.984 57.563
587509 88.543 -54.000 57.562
587510 88.578 -54.015 57.561
587511 88.614 -54.030 57.560
587512 88.649 -54.045 57.558
587513 88.685 -54.060 57.555
587514 88.720 -54.076 57.553
587515 88.757 -54.095 57.549
587516 88.796 -54.113 57.543
587517 88.839 -54.134 57.538
587518 88.875 -54.158 57.533
587519 88.921 -54.169 57.525
587520 88.960 -54.191 57.514
587521 88.998 -54.207 57.497
587522 89.037 -54.225 57.475
587523 89.074 -54.242 57.448
587524 89.113 -54.258 57.415
587525 89.151 -54.275 57.379
587526 89.189 -54.292 57.340
587527 89.227 -54.309 57.296
587528 89.265 -54.326 57.251
587529 89.303 -54.342 57.205
587530 89.341 -54.359 57.157
587531 89.380 -54.376 57.110
587532 89.418 -54.392 57.063
587533 89.456 -54.410 57.015
587534 89.494 -54.427 56.967
587535 89.532 -54.443 56.921
587546 89.954 -54.657 56.307
587547 89.992 -54.681 56.273
587548 90.031 -54.704 56.240
587549 90.070 -54.728 56.210
587550 90.108 -54.750 56.181
587551 90.147 -54.774 56.153
587552 90.185 -54.797 56.125
587553 90.224 -54.821 56.098
587554 90.262 -54.844 56.073
587555 90.298 -54.866 56.052
587556 90.321 -54.880 56.031
587557 90.345 -54.894 56.011
587558 90.368 -54.909 55.994
587559 90.391 -54.922 55.977
587560 90.414 -54.937 55.961
587561 90.438 -54.950 55.948
587562 90.461 -54.965 55.938
587563 90.484 -54.979 55.924
587564 90.508 -54.993 55.905
587565 90.533 -55.007 55.886
587566 90.559 -55.020 55.870
587567 90.585 -55.035 55.860
587568 90.610 -55.048 55.852
587569 90.636 -55.062 55.846
587570 90.662 -55.075 55.841
587571 90.688 -55.088 55.839
587572 90.714 -55.099 55.838
587573 90.749 -55.115 55.840
587574 90.785 -55.131 55.845
587575 90.820 -55.146 55.855
587576 90.855 -55.162 55.866
587577 90.890 -55.178 55.876
587578 90.926 -55.194 55.886
587579 90.961 -55.209 55.893
587580 90.997 -55.225 55.892
587581 91.032 -55.241 55.889
587582 91.067 -55.256 55.885
587583 91.102 -55.272 55.880
587584 91.138 -55.288 55.872
587585 91.173 -55.304 55.858
587586 91.212 -55.317 55.839
587587 91.257 -55.329 55.823
587588 91.302 -55.340 55.808
587589 91.347 -55.351 55.790
587590 91.392 -55.363 55.763
587591 91.437 -55.373 55.734
587592 91.475 -55.386 55.708
587593 91.511 -55.398 55.688
587594 91.545 -55.411 55.674
587595 91.581 -55.424 55.663
587596 91.615 -55.437 55.644
587597 91.651 -55.450 55.619
587598 91.685 -55.462 55.583
587599 91.721 -55.475 55.542
587600 91.756 -55.488 55.504
587601 91.791 -55.501 55.471
587602 91.826 -55.514 55.442
587603 91.861 -55.526 55.413
587604 91.896 -55.539 55.387
587605 91.931 -55.552 55.364
587606 91.966 -55.564 55.344
587607 92.001 -55.577 55.325
587608 92.036 -55.589 55.311
587609 92.067 -55.602 55.298
587610 92.098 -55.614 55.290
587611 92.130 -55.627 55.287
587612 92.163 -55.639 55.288
587613 92.195 -55.652 55.287
587614 92.228 -55.665 55.277
587615 92.260 -55.677 55.261
587616 92.292 -55.694 55.241
587617 92.322 -55.720 55.221
587618 92.353 -55.746 55.203
587619 92.383 -55.771 55.190
587620 92.414 -55.797 55.179
587621 92.444 -55.823 55.176
587622 92.475 -55.849 55.173
587623 92.506 -55.871 55.171
587624 92.538 -55.890 55.171
587625 92.573 -55.909 55.174
587626 92.609 -55.928 55.170
587627 92.645 -55.948 55.166
587628 92.681 -55.967 55.169
587629 92.716 -55.987 55.179
587630 92.752 -56.005 55.194
587631 92.788 -56.020 55.205
587632 92.825 -56.036 55.211
587633 92.861 -56.051 55.217
587634 92.897 -56.066 55.226
587635 92.933 -56.082 55.236
587636 92.970 -56.097 55.246
587637 93.006 -56.113 55.253
587638 93.042 -56.117 55.257
587639 93.077 -56.121 55.267
587640 93.113 -56.125 55.281
587641 93.149 -56.130 55.295
587642 93.185 -56.135 55.305
587643 93.221 -56.139 55.309
587644 93.257 -56.144 55.307
587645 93.292 -56.148 55.302
587646 93.328 -56.153 55.296
587647 93.364 -56.157 55.290
587648 93.400 -56.161 55.282
587649 93.434 -56.164 55.276
587650 93.467 -56.176 55.266
587651 93.499 -56.194 55.257
587652 93.531 -56.210 55.246
587653 93.563 -56.227 55.232
587654 93.595 -56.244 55.213
587655 93.627 -56.261 55.193
587656 93.659 -56.278 55.169
587657 93.691 -56.295 55.143
587658 93.723 -56.312 55.116
587659 93.755 -56.329 55.087
587660 93.787 -56.345 55.054
587661 93.820 -56.363 55.020
587662 93.851 -56.379 54.989
587663 93.884 -56.396 54.958
587664 93.916 -56.414 54.928
587665 93.948 -56.430 54.898
587666 93.979 -56.447 54.864
587667 94.009 -56.462 54.831
587668 94.040 -56.478 54.797
587669 94.070 -56.494 54.762
587670 94.100 -56.510 54.727
587671 94.130 -56.525 54.688
587672 94.160 -56.541 54.644
587673 94.191 -56.557 54.600
587674 94.221 -56.572 54.554
587675 94.252 -56.588 54.505
587676 94.281 -56.602 54.454
587677 94.310 -56.611 54.400
587678 94.338 -56.620 54.347
587679 94.366 -56.630 54.309
587680 94.394 -56.639 54.272
587681 94.422 -56.648 54.234
587682 94.453 -56.657 54.194
587683 94.489 -56.667 54.152
587684 94.525 -56.677 54.109
587685 94.560 -56.686 54.066
587686 94.596 -56.696 54.021
587687 94.632 -56.705 53.975
587688 94.666 -56.715 53.929
587689 94.695 -56.730 53.886
587690 94.725 -56.745 53.842
587691 94.754 -56.760 53.804
587692 94.783 -56.774 53.767
587693 94.813 -56.790 53.730
587694 94.842 -56.805 53.693
587695 94.872 -56.819 53.655
587696 94.901 -56.834 53.611
587697 94.930 -56.849 53.555
587698 94.960 -56.864 53.496
587699 94.988 -56.877 53.443
587700 95.017 -56.890 53.400
587701 95.045 -56.903 53.363
587702 95.073 -56.915 53.330
587703 95.101 -56.928 53.297
587704 95.129 -56.941 53.262
587705 95.158 -56.954 53.227
587706 95.192 -56.961 53.190
587707 95.233 -56.963 53.152
587708 95.283 -56.960 53.114
587709 95.331 -56.958 53.075
587710 95.358 -56.968 53.037
587711 95.386 -56.978 53.002
587712 95.413 -56.987 52.969
587713 95.450 -56.996 52.934
587714 95.488 -57.004 52.897
587715 95.527 -57.012 52.863
587716 95.566 -57.021 52.833
587717 95.605 -57.029 52.804
587718 95.644 -57.037 52.776
587719 95.683 -57.046 52.747
587720 95.722 -57.054 52.719
587721 95.760 -57.062 52.689
587722 95.798 -57.071 52.656
587723 95.834 -57.078 52.618
587724 95.862 -57.084 52.580
587725 95.884 -57.091 52.548
587726 95.905 -57.098 52.515
587727 95.927 -57.105 52.480
587728 95.949 -57.112 52.442
587729 95.970 -57.119 52.402
587730 95.992 -57.125 52.363
587731 96.014 -57.132 52.326
587732 96.035 -57.139 52.295
587733 96.057 -57.146 52.265
587734 96.079 -57.152 52.237
587735 96.100 -57.159 52.210
587736 96.122 -57.168 52.183
587737 96.146 -57.178 52.155
587738 96.169 -57.188 52.134
587739 96.192 -57.197 52.111
587740 96.220 -57.203 52.089
587741 96.248 -57.208 52.069
587742 96.275 -57.213 52.050
587743 96.304 -57.219 52.031
587744 96.331 -57.224 52.016
587745 96.359 -57.227 51.999
587746 96.387 -57.231 51.980
587747 96.415 -57.234 51.957
587748 96.444 -57.237 51.929
587749 96.473 -57.240 51.899
587750 96.503 -57.246 51.876
587751 96.533 -57.251 51.852
587752 96.562 -57.257 51.828
587753 96.592 -57.263 51.802
587754 96.624 -57.267 51.776
587755 96.657 -57.272 51.753
587756 96.689 -57.276 51.733
587757 96.721 -57.281 51.718
587758 96.753 -57.285 51.706
587759 96.786 -57.290 51.698
587760 96.818 -57.297 51.692
587761 96.851 -57.304 51.689
587762 96.883 -57.310 51.692
587763 96.915 -57.316 51.687
587764 96.948 -57.323 51.676
587765 96.980 -57.325 51.661
587766 97.013 -57.316 51.645
587767 97.045 -57.307 51.634
587768 97.078 -57.298 51.626
587769 97.114 -57.295 51.617
587770 97.151 -57.293 51.611
587771 97.188 -57.291 51.605
587772 97.225 -57.288 51.599
587773 97.262 -57.286 51.594
587774 97.299 -57.284 51.592
587775 97.330 -57.282 51.592
587776 97.358 -57.278 51.589
587777 97.385 -57.275 51.579
587778 97.412 -57.272 51.564
587779 97.439 -57.269 51.545
587780 97.467 -57.265 51.521
587781 97.494 -57.262 51.495
587782 97.527 -57.272 51.470
587783 97.561 -57.286 51.444
587784 97.592 -57.294 51.415
587785 97.623 -57.301 51.380
587786 97.653 -57.308 51.339
587787 97.683 -57.316 51.298
587788 97.714 -57.322 51.255
587789 97.744 -57.330 51.210
587790 97.775 -57.338 51.165
587791 97.806 -57.344 51.120
587792 97.836 -57.352 51.081
587793 97.867 -57.359 51.046
587794 97.897 -57.366 51.012
587795 97.928 -57.372 50.980
587796 97.959 -57.375 50.949
587797 97.995 -57.379 50.920
587798 98.031 -57.383 50.889
587799 98.067 -57.387 50.860
587800 98.102 -57.391 50.832
587801 98.138 -57.395 50.803
587802 98.174 -57.400 50.773
587803 98.209 -57.404 50.740
587804 98.245 -57.407 50.706
587805 98.281 -57.412 50.671
587806 98.316 -57.416 50.636
587807 98.352 -57.420 50.602
587808 98.388 -57.424 50.570
587809 98.424 -57.429 50.539
587810 98.461 -57.432 50.506
587811 98.506 -57.435 50.472
587812 98.551 -57.438 50.440
587813 98.596 -57.441 50.409
587814 98.631 -57.438 50.378
587815 98.661 -57.432 50.349
587816 98.691 -57.426 50.319
587817 98.721 -57.420 50.292
587818 98.757 -57.421 50.266
587819 98.795 -57.423 50.241
587820 98.833 -57.426 50.215
587821 98.870 -57.429 50.189
587822 98.908 -57.432 50.161
587823 98.936 -57.439 50.132
587824 98.962 -57.447 50.104
587825 98.990 -57.455 50.077
587826 99.016 -57.463 50.049
587827 99.044 -57.470 50.020
587828 99.070 -57.478 49.990
587829 99.097 -57.485 49.960
587830 99.124 -57.497 49.929
587831 99.151 -57.509 49.896
587832 99.171 -57.519 49.862
587833 99.185 -57.526 49.827
587834 99.198 -57.532 49.791
587835 99.211 -57.539 49.757
587836 99.224 -57.545 49.721
587837 99.242 -57.551 49.683
587838 99.272 -57.558 49.644
587839 99.301 -57.564 49.603
587840 99.330 -57.570 49.566
587841 99.360 -57.575 49.528
587842 99.389 -57.581 49.490
587843 99.418 -57.587 49.453
587844 99.448 -57.593 49.417
587845 99.476 -57.598 49.379
587846 99.504 -57.604 49.342
587847 99.531 -57.609 49.303
587848 99.559 -57.615 49.265
587849 99.609 -57.621 49.228
587850 99.644 -57.630 49.190
587851 99.678 -57.639 49.155
587852 99.713 -57.648 49.122
587853 99.747 -57.656 49.092
587854 99.782 -57.665 49.063
587855 99.816 -57.673 49.036
587856 99.850 -57.683 49.010
587857 99.885 -57.692 48.986
587858 99.919 -57.700 48.965
587859 99.954 -57.709 48.944
587860 99.988 -57.717 48.924
587861 100.023 -57.727 48.904
587862 100.055 -57.739 48.885
588205 88.391 -53.888 57.552
588206 88.426 -53.903 57.554
588207 88.462 -53.919 57.555
588208 88.497 -53.934 57.556
588209 88.533 -53.949 57.558
588210 88.568 -53.964 57.557
588211 88.603 -53.980 57.556
588212 88.638 -53.995 57.555
588213 88.674 -54.010 57.552
588214 88.709 -54.026 57.550
588215 88.747 -54.044 57.546
588216 88.785 -54.063 57.541
588217 88.823 -54.083 57.536
588218 88.871 -54.103 57.530
588219 88.905 -54.127 57.524
588220 88.941 -54.151 57.516
588221 88.982 -54.167 57.502
588222 89.020 -54.185 57.483
588223 89.058 -54.201 57.461
588224 89.097 -54.218 57.435
588225 89.134 -54.235 57.405
588226 89.173 -54.251 57.373
588227 89.211 -54.269 57.336
588228 89.249 -54.286 57.294
588229 89.287 -54.302 57.251
588230 89.325 -54.319 57.205
588231 89.364 -54.336 57.159
588232 89.401 -54.352 57.113
588233 89.440 -54.370 57.065
588234 89.478 -54.386 57.017
588235 89.516 -54.403 56.969
588246 89.936 -54.602 56.348
588247 89.975 -54.625 56.312
588248 90.014 -54.649 56.278
588249 90.052 -54.672 56.246
588250 90.091 -54.695 56.216
588251 90.129 -54.718 56.188
588252 90.168 -54.742 56.160
588253 90.206 -54.765 56.133
588254 90.245 -54.789 56.108
588255 90.284 -54.812 56.084
588256 90.317 -54.833 56.064
588257 90.340 -54.847 56.043
588258 90.364 -54.861 56.025
588259 90.387 -54.875 56.008
588260 90.410 -54.889 55.993
588261 90.434 -54.903 55.979
588262 90.457 -54.918 55.969
588263 90.480 -54.931 55.959
588264 90.504 -54.946 55.942
588265 90.530 -54.960 55.920
588266 90.555 -54.973 55.897
588267 90.581 -54.987 55.878
588268 90.607 -55.000 55.866
588269 90.633 -55.014 55.856
588270 90.658 -55.028 55.848
588271 90.684 -55.041 55.843
588272 90.710 -55.056 55.840
588273 90.741 -55.069 55.840
588274 90.777 -55.085 55.842
588275 90.812 -55.101 55.845
588276 90.847 -55.116 55.850
588277 90.882 -55.132 55.858
588278 90.918 -55.148 55.867
588279 90.953 -55.163 55.873
588280 90.988 -55.179 55.875
588281 91.024 -55.195 55.872
588282 91.059 -55.210 55.867
588283 91.094 -55.226 55.863
588284 91.130 -55.242 55.858
588285 91.165 -55.257 55.852
588286 91.200 -55.273 55.838
588287 91.235 -55.289 55.818
588288 91.271 -55.305 55.798
588289 91.315 -55.317 55.782
588290 91.360 -55.328 55.763
588291 91.405 -55.339 55.738
588292 91.449 -55.351 55.710
588293 91.485 -55.363 55.686
588294 91.519 -55.376 55.667
588295 91.555 -55.389 55.654
588296 91.589 -55.401 55.645
588297 91.625 -55.414 55.629
588298 91.659 -55.426 55.602
588299 91.695 -55.439 55.564
588300 91.729 -55.452 55.523
588301 91.765 -55.465 55.487
588302 91.800 -55.478 55.456
588303 91.835 -55.490 55.428
588304 91.870 -55.503 55.401
588305 91.905 -55.516 55.377
588306 91.940 -55.529 55.355
588307 91.975 -55.542 55.336
588308 92.010 -55.554 55.320
588309 92.045 -55.566 55.308
588310 92.080 -55.577 55.299
588311 92.115 -55.591 55.294
588312 92.148 -55.605 55.294
588313 92.181 -55.618 55.299
588314 92.213 -55.630 55.303
588315 92.246 -55.643 55.298
588316 92.278 -55.656 55.287
588317 92.311 -55.668 55.272
588318 92.341 -55.694 55.253
588319 92.372 -55.720 55.233
588320 92.402 -55.746 55.214
588321 92.433 -55.771 55.199
588322 92.463 -55.794 55.189
588323 92.495 -55.813 55.183
588324 92.526 -55.833 55.181
588325 92.561 -55.852 55.183
588326 92.596 -55.872 55.188
588327 92.632 -55.891 55.185
588328 92.668 -55.910 55.177
588329 92.704 -55.930 55.171
588330 92.739 -55.949 55.169
588331 92.775 -55.969 55.175
588332 92.810 -55.987 55.185
588333 92.846 -56.007 55.193
588334 92.882 -56.026 55.204
588335 92.918 -56.044 55.218
588336 92.954 -56.059 55.233
588337 92.990 -56.075 55.246
588338 93.026 -56.090 55.254
588339 93.062 -56.094 55.259
588340 93.098 -56.099 55.267
588341 93.134 -56.103 55.280
588342 93.170 -56.108 55.294
588343 93.206 -56.113 55.304
588344 93.240 -56.115 55.309
588345 93.275 -56.118 55.308
588346 93.309 -56.120 55.304
588347 93.343 -56.123 55.298
588348 93.377 -56.125 55.289
588349 93.412 -56.128 55.276
588350 93.446 -56.131 55.263
588351 93.479 -56.143 55.249
588352 93.509 -56.165 55.236
588353 93.541 -56.187 55.223
588354 93.573 -56.204 55.209
588355 93.605 -56.221 55.190
588356 93.637 -56.238 55.170
588357 93.669 -56.254 55.146
588358 93.701 -56.272 55.120
588359 93.733 -56.288 55.092
588360 93.765 -56.305 55.064
588361 93.797 -56.323 55.032
588362 93.829 -56.339 55.000
588363 93.861 -56.356 54.970
588364 93.894 -56.373 54.941
588365 93.926 -56.390 54.914
588366 93.957 -56.406 54.887
588367 93.987 -56.422 54.855
588368 94.017 -56.438 54.823
588369 94.048 -56.454 54.791
588370 94.078 -56.469 54.759
588371 94.108 -56.485 54.723
588372 94.138 -56.501 54.683
588373 94.169 -56.517 54.638
588374 94.199 -56.532 54.594
588375 94.229 -56.548 54.547
588376 94.259 -56.564 54.499
588377 94.289 -56.577 54.448
588378 94.317 -56.586 54.397
588379 94.345 -56.595 54.343
588380 94.373 -56.604 54.300
588381 94.402 -56.613 54.262
588382 94.430 -56.622 54.222
588383 94.458 -56.631 54.181
588384 94.493 -56.641 54.138
588385 94.529 -56.650 54.092
588386 94.565 -56.660 54.049
588387 94.601 -56.670 54.004
588388 94.636 -56.679 53.958
588389 94.672 -56.689 53.912
588390 94.706 -56.699 53.868
588391 94.736 -56.714 53.824
588392 94.765 -56.729 53.787
588393 94.794 -56.744 53.751
588394 94.824 -56.758 53.716
588395 94.853 -56.774 53.680
588396 94.883 -56.788 53.643
588397 94.912 -56.803 53.601
588398 94.941 -56.818 53.551
588399 94.970 -56.831 53.492
588400 94.997 -56.844 53.434
588401 95.026 -56.857 53.381
588402 95.054 -56.869 53.338
588403 95.082 -56.882 53.302
588404 95.110 -56.895 53.268
588405 95.138 -56.908 53.232
588406 95.167 -56.921 53.196
588407 95.199 -56.930 53.160
588408 95.246 -56.929 53.123
588409 95.297 -56.925 53.084
588410 95.340 -56.926 53.045
588411 95.368 -56.936 53.005
588412 95.395 -56.945 52.967
588413 95.422 -56.955 52.932
588414 95.454 -56.964 52.898
588415 95.492 -56.972 52.863
588416 95.531 -56.981 52.832
588417 95.570 -56.989 52.804
588418 95.609 -56.997 52.777
588419 95.648 -57.006 52.749
588420 95.686 -57.014 52.720
588421 95.725 -57.023 52.693
588422 95.764 -57.031 52.664
588423 95.801 -57.038 52.630
588424 95.837 -57.046 52.592
588425 95.866 -57.053 52.551
588426 95.887 -57.059 52.514
588427 95.909 -57.066 52.481
588428 95.931 -57.073 52.446
588429 95.952 -57.080 52.409
588430 95.974 -57.087 52.372
588431 95.995 -57.094 52.333
588432 96.017 -57.100 52.295
588433 96.039 -57.106 52.264
588434 96.060 -57.113 52.235
588435 96.082 -57.120 52.209
588436 96.104 -57.128 52.184
588437 96.127 -57.138 52.157
588438 96.150 -57.148 52.130
588439 96.173 -57.158 52.106
588440 96.196 -57.168 52.083
588441 96.223 -57.174 52.063
588442 96.251 -57.179 52.044
588443 96.279 -57.185 52.024
588444 96.307 -57.190 52.004
588445 96.335 -57.194 51.990
588446 96.363 -57.197 51.975
588447 96.390 -57.200 51.957
588448 96.418 -57.204 51.934
588449 96.447 -57.207 51.910
588450 96.476 -57.210 51.887
588451 96.505 -57.213 51.866
588452 96.534 -57.217 51.840
588453 96.564 -57.222 51.811
588454 96.594 -57.228 51.783
588455 96.626 -57.233 51.756
588456 96.658 -57.238 51.730
588457 96.690 -57.242 51.708
588458 96.723 -57.247 51.687
588459 96.755 -57.254 51.672
588460 96.788 -57.260 51.662
588461 96.820 -57.266 51.653
588462 96.852 -57.273 51.647
588463 96.885 -57.280 51.645
588464 96.917 -57.286 51.637
588465 96.949 -57.293 51.625
588466 96.981 -57.296 51.610
588467 97.014 -57.287 51.596
588468 97.046 -57.279 51.587
588469 97.079 -57.269 51.583
588470 97.111 -57.260 51.572
588471 97.146 -57.254 51.561
588472 97.183 -57.252 51.553
588473 97.220 -57.250 51.547
588474 97.257 -57.248 51.541
588475 97.294 -57.246 51.539
588476 97.331 -57.244 51.539
588477 97.360 -57.241 51.536
588478 97.387 -57.238 51.527
588479 97.415 -57.234 51.513
588480 97.442 -57.231 51.495
588481 97.469 -57.228 51.472
588482 97.497 -57.225 51.447
588483 97.529 -57.233 51.421
588484 97.563 -57.247 51.395
588485 97.594 -57.254 51.368
588486 97.624 -57.261 51.335
588487 97.655 -57.269 51.297
588488 97.685 -57.275 51.259
588489 97.716 -57.283 51.219
588490 97.746 -57.291 51.177
588491 97.777 -57.297 51.134
588492 97.807 -57.305 51.091
588493 97.838 -57.312 51.053
588494 97.868 -57.319 51.018
588495 97.899 -57.327 50.984
588496 97.930 -57.334 50.953
588497 97.963 -57.340 50.922
588498 97.999 -57.344 50.893
588499 98.035 -57.348 50.864
588500 98.071 -57.352 50.836
588501 98.106 -57.357 50.810
588502 98.142 -57.360 50.784
588503 98.178 -57.365 50.755
588504 98.214 -57.369 50.725
588505 98.249 -57.373 50.692
588506 98.285 -57.377 50.658
588507 98.321 -57.382 50.624
588508 98.356 -57.385 50.590
588509 98.392 -57.390 50.558
588510 98.428 -57.394 50.527
588511 98.463 -57.398 50.496
588512 98.501 -57.402 50.463
588513 98.546 -57.405 50.431
588514 98.591 -57.407 50.399
588515 98.632 -57.408 50.369
588516 98.662 -57.402 50.339
588517 98.692 -57.396 50.310
588518 98.722 -57.391 50.285
588519 98.760 -57.394 50.261
588520 98.798 -57.396 50.236
588521 98.836 -57.399 50.210
588522 98.873 -57.401 50.184
588523 98.911 -57.406 50.157
588524 98.926 -57.410 50.129
588525 98.953 -57.418 50.103
588526 98.980 -57.426 50.077
588527 99.007 -57.433 50.050
588528 99.034 -57.442 50.022
588529 99.061 -57.449 49.993
588530 99.088 -57.459 49.963
588531 99.115 -57.471 49.931
588532 99.141 -57.484 49.897
588533 99.166 -57.496 49.861
588534 99.180 -57.502 49.825
588535 99.193 -57.509 49.789
588536 99.206 -57.516 49.754
588537 99.220 -57.522 49.718
588538 99.243 -57.528 49.680
588539 99.274 -57.532 49.642
588540 99.304 -57.538 49.601
588541 99.334 -57.543 49.563
588542 99.363 -57.549 49.525
588543 99.393 -57.555 49.488
588544 99.422 -57.561 49.451
588545 99.451 -57.567 49.414
588546 99.479 -57.573 49.376
588547 99.507 -57.578 49.339
588548 99.534 -57.583 49.302
588549 99.562 -57.589 49.264
588550 99.594 -57.595 49.226
588551 99.636 -57.602 49.189
588552 99.671 -57.611 49.153
588553 99.705 -57.620 49.120
588554 99.739 -57.629 49.090
588555 99.774 -57.637 49.061
588556 99.808 -57.646 49.033
588557 99.843 -57.655 49.007
588558 99.877 -57.664 48.982
588559 99.911 -57.673 48.962
588560 99.940 -57.690 48.941
588561 99.969 -57.708 48.921
588562 99.997 -57.725 48.901
588563 100.026 -57.742 48.882
588906 88.416 -53.853 57.547
588907 88.451 -53.868 57.549
588908 88.486 -53.884 57.550
588909 88.522 -53.899 57.550
588910 88.557 -53.914 57.553
588911 88.593 -53.929 57.552
588912 88.628 -53.944 57.550
588913 88.663 -53.960 57.548
588914 88.699 -53.976 57.546
588915 88.737 -53.994 57.543
588916 88.774 -54.013 57.540
588917 88.813 -54.033 57.534
588918 88.850 -54.052 57.528
588919 88.900 -54.073 57.522
588920 88.936 -54.096 57.513
588921 88.972 -54.120 57.503
588922 89.013 -54.136 57.491
588923 89.045 -54.159 57.472
588924 89.081 -54.178 57.449
588925 89.118 -54.195 57.423
588926 89.157 -54.211 57.397
588927 89.195 -54.229 57.366
588928 89.233 -54.245 57.332
588929 89.271 -54.262 57.292
588930 89.309 -54.279 57.249
588931 89.347 -54.295 57.205
588932 89.385 -54.312 57.159
588933 89.423 -54.330 57.114
588934 89.462 -54.346 57.067
588935 89.499 -54.363 57.019
588936 89.538 -54.380 56.972
588946 89.919 -54.548 56.389
588947 89.958 -54.570 56.353
588948 89.996 -54.593 56.317
588949 90.035 -54.617 56.283
588950 90.074 -54.640 56.251
588951 90.112 -54.663 56.222
588952 90.151 -54.687 56.195
588953 90.189 -54.710 56.168
588954 90.228 -54.734 56.142
588955 90.266 -54.757 56.117
588956 90.305 -54.781 56.094
588957 90.336 -54.800 56.075
588958 90.360 -54.814 56.057
588959 90.383 -54.828 56.041
588960 90.406 -54.842 56.026
588961 90.429 -54.856 56.013
588962 90.453 -54.870 56.001
588963 90.476 -54.884 55.991
588964 90.500 -54.898 55.978
588965 90.526 -54.912 55.959
588966 90.551 -54.925 55.934
588967 90.577 -54.939 55.906
588968 90.603 -54.953 55.886
588969 90.628 -54.967 55.871
588970 90.654 -54.981 55.860
588971 90.680 -54.994 55.851
588972 90.706 -55.008 55.845
588973 90.724 -55.025 55.843
588974 90.761 -55.041 55.843
588975 90.798 -55.057 55.843
588976 90.834 -55.072 55.845
588977 90.871 -55.087 55.847
588978 90.908 -55.102 55.851
588979 90.945 -55.118 55.856
588980 90.980 -55.133 55.859
588981 91.015 -55.149 55.856
588982 91.051 -55.165 55.851
588983 91.086 -55.181 55.846
588984 91.121 -55.196 55.842
588985 91.157 -55.212 55.839
588986 91.192 -55.228 55.836
588987 91.227 -55.244 55.821
588988 91.262 -55.259 55.798
588989 91.298 -55.275 55.775
588990 91.333 -55.291 55.757
588991 91.373 -55.304 55.737
588992 91.418 -55.316 55.717
588993 91.458 -55.328 55.690
588994 91.493 -55.340 55.666
588995 91.528 -55.353 55.647
588996 91.563 -55.366 55.634
588997 91.598 -55.379 55.626
588998 91.633 -55.392 55.613
588999 91.668 -55.404 55.585
589000 91.703 -55.417 55.546
589001 91.738 -55.430 55.505
589002 91.774 -55.440 55.472
589003 91.809 -55.451 55.443
589004 91.845 -55.462 55.416
589005 91.880 -55.473 55.391
589006 91.916 -55.483 55.368
589007 91.951 -55.494 55.347
589008 91.987 -55.505 55.330
589009 92.022 -55.516 55.317
589010 92.058 -55.527 55.306
589011 92.092 -55.541 55.299
589012 92.126 -55.558 55.298
589013 92.161 -55.574 55.302
589014 92.196 -55.591 55.312
589015 92.230 -55.608 55.318
589016 92.264 -55.623 55.318
589017 92.296 -55.635 55.314
589018 92.329 -55.648 55.303
589019 92.360 -55.668 55.287
589020 92.391 -55.694 55.267
589021 92.422 -55.718 55.243
589022 92.453 -55.737 55.223
589023 92.484 -55.756 55.207
589024 92.515 -55.776 55.197
589025 92.549 -55.795 55.194
589026 92.584 -55.815 55.194
589027 92.620 -55.834 55.197
589028 92.655 -55.853 55.196
589029 92.691 -55.872 55.190
589030 92.727 -55.892 55.182
589031 92.763 -55.911 55.175
589032 92.798 -55.931 55.173
589033 92.834 -55.950 55.180
589034 92.869 -55.969 55.193
589035 92.905 -55.988 55.207
589036 92.941 -56.008 55.222
589037 92.977 -56.027 55.236
589038 93.012 -56.047 55.250
589039 93.046 -56.066 55.258
589040 93.081 -56.069 55.262
589041 93.115 -56.072 55.267
589042 93.149 -56.074 55.276
589043 93.184 -56.077 55.289
589044 93.218 -56.079 55.301
589045 93.252 -56.082 55.308
589046 93.286 -56.085 55.311
589047 93.321 -56.088 55.310
589048 93.355 -56.090 55.304
589049 93.389 -56.092 55.294
589050 93.423 -56.095 55.277
589051 93.458 -56.097 55.257
589052 93.490 -56.110 55.239
589053 93.521 -56.132 55.221
589054 93.553 -56.154 55.205
589055 93.583 -56.176 55.190
589056 93.615 -56.197 55.172
589057 93.647 -56.214 55.154
589058 93.679 -56.232 55.133
589059 93.711 -56.248 55.110
589060 93.743 -56.265 55.085
589061 93.775 -56.282 55.058
589062 93.807 -56.299 55.032
589063 93.839 -56.316 55.005
589064 93.871 -56.333 54.979
589065 93.903 -56.350 54.954
589066 93.934 -56.366 54.928
589067 93.965 -56.382 54.903
589068 93.995 -56.398 54.872
589069 94.025 -56.413 54.838
589070 94.055 -56.429 54.801
589071 94.085 -56.445 54.763
589072 94.116 -56.461 54.722
589073 94.146 -56.476 54.680
589074 94.177 -56.492 54.633
589075 94.207 -56.508 54.585
589076 94.237 -56.523 54.536
589077 94.267 -56.539 54.488
589078 94.296 -56.551 54.440
589079 94.325 -56.560 54.392
589080 94.353 -56.569 54.340
589081 94.381 -56.578 54.292
589082 94.409 -56.587 54.251
589083 94.437 -56.596 54.210
589084 94.466 -56.605 54.167
589085 94.498 -56.615 54.123
589086 94.534 -56.624 54.075
589087 94.570 -56.634 54.031
589088 94.606 -56.644 53.986
589089 94.641 -56.653 53.939
589090 94.677 -56.663 53.892
589091 94.712 -56.672 53.848
589092 94.746 -56.683 53.805
589093 94.776 -56.699 53.767
589094 94.805 -56.713 53.733
589095 94.834 -56.728 53.698
589096 94.864 -56.743 53.663
589097 94.893 -56.758 53.627
589098 94.922 -56.772 53.586
589099 94.951 -56.785 53.539
589100 94.979 -56.798 53.484
589101 95.007 -56.811 53.425
589102 95.035 -56.824 53.367
589103 95.063 -56.836 53.318
589104 95.092 -56.849 53.274
589105 95.120 -56.862 53.237
589106 95.147 -56.874 53.200
589107 95.176 -56.887 53.164
589108 95.208 -56.897 53.127
589109 95.259 -56.894 53.090
589110 95.309 -56.891 53.053
589111 95.349 -56.894 53.015
589112 95.377 -56.903 52.975
589113 95.404 -56.913 52.936
589114 95.431 -56.922 52.898
589115 95.459 -56.932 52.863
589116 95.496 -56.940 52.832
589117 95.535 -56.949 52.803
589118 95.573 -56.957 52.774
589119 95.612 -56.965 52.747
589120 95.651 -56.975 52.721
589121 95.690 -56.983 52.694
589122 95.729 -56.991 52.668
589123 95.767 -57.000 52.638
589124 95.804 -57.006 52.604
589125 95.840 -57.014 52.566
589126 95.870 -57.021 52.524
589127 95.891 -57.028 52.482
589128 95.913 -57.034 52.446
589129 95.934 -57.041 52.411
589130 95.956 -57.048 52.375
589131 95.977 -57.054 52.339
589132 95.999 -57.061 52.300
589133 96.021 -57.068 52.262
589134 96.042 -57.075 52.231
589135 96.064 -57.081 52.202
589136 96.086 -57.088 52.177
589137 96.109 -57.099 52.153
589138 96.132 -57.109 52.127
589139 96.155 -57.119 52.100
589140 96.178 -57.128 52.077
589141 96.201 -57.138 52.054
589142 96.227 -57.146 52.035
589143 96.254 -57.151 52.016
589144 96.282 -57.156 51.996
589145 96.310 -57.160 51.975
589146 96.338 -57.163 51.961
589147 96.366 -57.167 51.947
589148 96.394 -57.171 51.931
589149 96.421 -57.174 51.912
589150 96.450 -57.177 51.890
589151 96.479 -57.180 51.868
589152 96.508 -57.183 51.845
589153 96.537 -57.185 51.820
589154 96.566 -57.188 51.794
589155 96.595 -57.194 51.768
589156 96.627 -57.198 51.740
589157 96.660 -57.203 51.711
589158 96.692 -57.210 51.684
589159 96.725 -57.217 51.661
589160 96.757 -57.223 51.643
589161 96.789 -57.230 51.630
589162 96.821 -57.237 51.617
589163 96.854 -57.243 51.604
589164 96.886 -57.250 51.595
589165 96.919 -57.257 51.584
589166 96.951 -57.263 51.571
589167 96.983 -57.268 51.558
589168 97.016 -57.259 51.547
589169 97.048 -57.250 51.540
589170 97.081 -57.241 51.538
589171 97.113 -57.232 51.529
589172 97.146 -57.223 51.517
589173 97.178 -57.214 51.506
589174 97.215 -57.212 51.498
589175 97.252 -57.210 51.491
589176 97.289 -57.207 51.488
589177 97.326 -57.205 51.486
589178 97.363 -57.203 51.483
589179 97.390 -57.200 51.475
589180 97.418 -57.197 51.464
589181 97.445 -57.194 51.446
589182 97.472 -57.190 51.424
589183 97.500 -57.187 51.398
589184 97.531 -57.194 51.372
589185 97.565 -57.207 51.347
589186 97.595 -57.214 51.319
589187 97.626 -57.222 51.288
589188 97.656 -57.229 51.252
589189 97.687 -57.236 51.216
589190 97.718 -57.244 51.180
589191 97.748 -57.250 51.142
589192 97.779 -57.258 51.101
589193 97.809 -57.266 51.061
589194 97.840 -57.272 51.024
589195 97.870 -57.280 50.989
589196 97.902 -57.290 50.956
589197 97.936 -57.303 50.926
589198 97.969 -57.311 50.896
589199 98.005 -57.315 50.867
589200 98.040 -57.317 50.839
589201 98.075 -57.319 50.813
589202 98.110 -57.322 50.788
589203 98.146 -57.326 50.762
589204 98.182 -57.330 50.736
589205 98.218 -57.334 50.708
589206 98.253 -57.338 50.677
589207 98.289 -57.342 50.644
589208 98.325 -57.347 50.611
589209 98.361 -57.351 50.578
589210 98.396 -57.355 50.546
589211 98.432 -57.359 50.515
589212 98.467 -57.363 50.485
589213 98.503 -57.367 50.454
589214 98.541 -57.372 50.422
589215 98.585 -57.374 50.390
589216 98.631 -57.377 50.360
589217 98.662 -57.373 50.330
589218 98.692 -57.366 50.302
589219 98.725 -57.363 50.279
589220 98.763 -57.366 50.255
589221 98.801 -57.369 50.230
589222 98.839 -57.372 50.205
589223 98.876 -57.376 50.179
589224 98.915 -57.380 50.153
589225 98.934 -57.385 50.126
589226 98.944 -57.389 50.101
589227 98.971 -57.397 50.076
589228 98.998 -57.404 50.050
589229 99.025 -57.412 50.023
589230 99.052 -57.421 49.995
589231 99.078 -57.433 49.965
589232 99.106 -57.446 49.932
589233 99.132 -57.459 49.897
589234 99.159 -57.471 49.860
589235 99.175 -57.479 49.823
589236 99.189 -57.485 49.787
589237 99.202 -57.492 49.751
589238 99.215 -57.499 49.714
589239 99.244 -57.504 49.677
589240 99.275 -57.508 49.638
589241 99.306 -57.514 49.598
589242 99.336 -57.518 49.559
589243 99.366 -57.523 49.522
589244 99.396 -57.529 49.486
589245 99.425 -57.536 49.448
589246 99.454 -57.541 49.411
589247 99.482 -57.547 49.375
589248 99.510 -57.552 49.337
589249 99.537 -57.558 49.301
589250 99.565 -57.563 49.263
589251 99.592 -57.568 49.226
589252 99.629 -57.574 49.189
589253 99.663 -57.583 49.153
589254 99.697 -57.592 49.120
589255 99.732 -57.601 49.089
589256 99.766 -57.610 49.059
589257 99.796 -57.625 49.031
589258 99.825 -57.642 49.004
589259 99.853 -57.660 48.979
589260 99.882 -57.677 48.958
589261 99.911 -57.694 48.937
589262 99.939 -57.711 48.917
589263 99.968 -57.728 48.897
589264 99.997 -57.746 48.878
589607 88.440 -53.818 57.541
589608 88.476 -53.833 57.543
589609 88.511 -53.848 57.544
589610 88.546 -53.864 57.545
589611 88.582 -53.879 57.547
589612 88.617 -53.894 57.546
589613 88.652 -53.910 57.544
589614 88.688 -53.925 57.542
589615 88.726 -53.944 57.539
589616 88.764 -53.964 57.536
589617 88.802 -53.983 57.533
589618 88.840 -54.002 57.527
589619 88.878 -54.022 57.520
589620 88.928 -54.041 57.512
589621 88.968 -54.064 57.503
589622 89.003 -54.088 57.492
589623 89.050 -54.098 57.479
589624 89.093 -54.110 57.461
589625 89.125 -54.132 57.440
589626 89.157 -54.155 57.416
589627 89.190 -54.177 57.390
589628 89.222 -54.200 57.360
589629 89.255 -54.222 57.327
589630 89.293 -54.239 57.289
589631 89.331 -54.255 57.247
589632 89.369 -54.273 57.203
589633 89.407 -54.289 57.159
589634 89.445 -54.306 57.114
589635 89.483 -54.323 57.069
589636 89.522 -54.339 57.022
589637 89.560 -54.356 56.974
589638 89.598 -54.373 56.928
589647 89.941 -54.524 56.393
589648 89.979 -54.541 56.357
589649 90.018 -54.562 56.322
589650 90.056 -54.585 56.288
589651 90.095 -54.608 56.257
589652 90.133 -54.631 56.228
589653 90.172 -54.655 56.202
589654 90.211 -54.678 56.176
589655 90.249 -54.702 56.150
589656 90.288 -54.725 56.126
589657 90.326 -54.749 56.105
589658 90.356 -54.766 56.088
589659 90.379 -54.781 56.072
589660 90.402 -54.794 56.058
589661 90.425 -54.809 56.045
589662 90.448 -54.823 56.033
589663 90.471 -54.837 56.021
589664 90.496 -54.851 56.009
589665 90.522 -54.865 55.991
589666 90.547 -54.878 55.969
589667 90.573 -54.892 55.942
589668 90.599 -54.906 55.914
589669 90.624 -54.919 55.893
589670 90.650 -54.933 55.876
589671 90.674 -54.948 55.864
589672 90.692 -54.966 55.855
589673 90.710 -54.984 55.849
589674 90.733 -55.001 55.846
589675 90.769 -55.016 55.845
589676 90.806 -55.032 55.844
589677 90.843 -55.047 55.844
589678 90.880 -55.062 55.845
589679 90.917 -55.078 55.846
589680 90.953 -55.093 55.846
589681 90.990 -55.108 55.845
589682 91.027 -55.123 55.840
589683 91.064 -55.138 55.833
589684 91.101 -55.154 55.828
589685 91.138 -55.169 55.825
589686 91.174 -55.185 55.824
589687 91.211 -55.200 55.818
589688 91.248 -55.215 55.802
589689 91.285 -55.231 55.778
589690 91.321 -55.246 55.752
589691 91.358 -55.261 55.733
589692 91.395 -55.276 55.717
589693 91.431 -55.292 55.701
589694 91.467 -55.304 55.675
589695 91.503 -55.314 55.648
589696 91.538 -55.326 55.629
589697 91.574 -55.336 55.616
589698 91.609 -55.347 55.608
589699 91.645 -55.358 55.596
589700 91.680 -55.369 55.567
589701 91.716 -55.379 55.529
589702 91.751 -55.390 55.490
589703 91.787 -55.401 55.459
589704 91.822 -55.412 55.431
589705 91.857 -55.423 55.406
589706 91.893 -55.433 55.382
589707 91.928 -55.444 55.360
589708 91.964 -55.455 55.341
589709 91.999 -55.466 55.325
589710 92.035 -55.477 55.314
589711 92.070 -55.490 55.305
589712 92.103 -55.507 55.300
589713 92.137 -55.524 55.302
589714 92.172 -55.540 55.309
589715 92.207 -55.557 55.322
589716 92.241 -55.574 55.331
589717 92.276 -55.590 55.336
589718 92.311 -55.607 55.336
589719 92.346 -55.624 55.331
589720 92.379 -55.641 55.320
589721 92.411 -55.661 55.304
589722 92.442 -55.680 55.279
589723 92.473 -55.699 55.253
589724 92.504 -55.719 55.231
589725 92.536 -55.738 55.216
589726 92.572 -55.758 55.210
589727 92.607 -55.777 55.207
589728 92.643 -55.796 55.208
589729 92.678 -55.815 55.211
589730 92.714 -55.835 55.207
589731 92.750 -55.854 55.198
589732 92.786 -55.874 55.187
589733 92.821 -55.893 55.180
589734 92.857 -55.912 55.182
589735 92.893 -55.931 55.191
589736 92.928 -55.951 55.205
589737 92.959 -55.969 55.220
589738 92.990 -55.987 55.234
589739 93.020 -56.006 55.249
589740 93.056 -56.027 55.260
589741 93.093 -56.036 55.266
589742 93.127 -56.038 55.271
589743 93.161 -56.041 55.279
589744 93.196 -56.044 55.290
589745 93.230 -56.047 55.304
589746 93.264 -56.049 55.313
589747 93.298 -56.051 55.318
589748 93.333 -56.054 55.318
589749 93.367 -56.056 55.313
589750 93.401 -56.059 55.303
589751 93.435 -56.062 55.285
589752 93.470 -56.065 55.261
589753 93.502 -56.076 55.236
589754 93.534 -56.098 55.212
589755 93.564 -56.121 55.193
589756 93.595 -56.143 55.177
589757 93.626 -56.165 55.159
589758 93.657 -56.188 55.143
589759 93.689 -56.208 55.125
589760 93.721 -56.225 55.105
589761 93.753 -56.242 55.084
589762 93.785 -56.259 55.061
589763 93.817 -56.276 55.040
589764 93.849 -56.293 55.017
589765 93.881 -56.310 54.995
589766 93.912 -56.326 54.972
589767 93.942 -56.342 54.948
589768 93.973 -56.357 54.924
589769 94.003 -56.373 54.894
589770 94.033 -56.389 54.861
589771 94.063 -56.404 54.824
589772 94.094 -56.420 54.785
589773 94.124 -56.436 54.744
589774 94.154 -56.451 54.700
589775 94.184 -56.467 54.653
589776 94.215 -56.483 54.604
589777 94.245 -56.498 54.552
589778 94.275 -56.514 54.500
589779 94.304 -56.526 54.447
589780 94.332 -56.535 54.394
589781 94.360 -56.544 54.338
589782 94.389 -56.553 54.287
589783 94.417 -56.562 54.240
589784 94.445 -56.571 54.195
589785 94.473 -56.580 54.150
589786 94.503 -56.589 54.106
589787 94.539 -56.599 54.058
589788 94.574 -56.608 54.012
589789 94.610 -56.618 53.966
589790 94.646 -56.627 53.919
589791 94.682 -56.637 53.873
589792 94.717 -56.646 53.828
589793 94.753 -56.656 53.785
589794 94.786 -56.668 53.747
589795 94.816 -56.683 53.713
589796 94.845 -56.697 53.679
589797 94.875 -56.712 53.643
589798 94.903 -56.726 53.608
589799 94.931 -56.739 53.569
589800 94.960 -56.752 53.524
589801 94.988 -56.765 53.473
589802 95.016 -56.777 53.416
589803 95.044 -56.790 53.356
589804 95.073 -56.802 53.301
589805 95.102 -56.813 53.251
589806 95.130 -56.824 53.208
589807 95.159 -56.836 53.168
589808 95.188 -56.846 53.130
589809 95.222 -56.858 53.093
589810 95.272 -56.860 53.057
589811 95.322 -56.857 53.021
589812 95.358 -56.862 52.984
589813 95.385 -56.871 52.944
589814 95.413 -56.881 52.904
589815 95.440 -56.890 52.867
589816 95.467 -56.899 52.834
589817 95.499 -56.909 52.804
589818 95.538 -56.917 52.775
589819 95.577 -56.925 52.746
589820 95.616 -56.934 52.719
589821 95.655 -56.943 52.692
589822 95.693 -56.951 52.664
589823 95.732 -56.959 52.637
589824 95.771 -56.968 52.607
589825 95.807 -56.975 52.574
589826 95.844 -56.982 52.536
589827 95.873 -56.989 52.493
589828 95.895 -56.996 52.449
589829 95.916 -57.003 52.411
589830 95.938 -57.009 52.375
589831 95.960 -57.015 52.339
589832 95.981 -57.022 52.301
589833 96.003 -57.029 52.263
589834 96.024 -57.036 52.226
589835 96.046 -57.043 52.196
589836 96.067 -57.050 52.168
589837 96.090 -57.059 52.142
589838 96.113 -57.069 52.118
589839 96.136 -57.079 52.094
589840 96.160 -57.089 52.067
589841 96.183 -57.099 52.046
589842 96.206 -57.109 52.024
589843 96.230 -57.118 52.005
589844 96.258 -57.123 51.986
589845 96.286 -57.127 51.967
589846 96.313 -57.130 51.945
589847 96.341 -57.134 51.930
589848 96.369 -57.138 51.917
589849 96.397 -57.141 51.904
589850 96.425 -57.144 51.888
589851 96.454 -57.147 51.869
589852 96.482 -57.150 51.847
589853 96.511 -57.153 51.823
589854 96.540 -57.156 51.799
589855 96.569 -57.159 51.776
589856 96.597 -57.160 51.750
589857 96.629 -57.166 51.721
589858 96.661 -57.174 51.688
589859 96.694 -57.180 51.659
589860 96.726 -57.187 51.633
589861 96.758 -57.194 51.612
589862 96.791 -57.200 51.595
589863 96.823 -57.206 51.579
589864 96.856 -57.213 51.562
589865 96.888 -57.219 51.548
589866 96.920 -57.226 51.535
589867 96.953 -57.233 51.523
589868 96.985 -57.239 51.511
589869 97.017 -57.230 51.501
589870 97.050 -57.222 51.495
589871 97.082 -57.213 51.493
589872 97.115 -57.203 51.485
589873 97.147 -57.194 51.473
589874 97.180 -57.186 51.461
589875 97.213 -57.177 51.453
589876 97.247 -57.171 51.449
589877 97.284 -57.169 51.446
589878 97.321 -57.167 51.442
589879 97.358 -57.165 51.438
589880 97.393 -57.163 51.429
589881 97.421 -57.159 51.417
589882 97.448 -57.156 51.400
589883 97.475 -57.153 51.378
589884 97.502 -57.150 51.354
589885 97.533 -57.156 51.329
589886 97.567 -57.167 51.302
589887 97.597 -57.175 51.274
589888 97.628 -57.182 51.244
589889 97.658 -57.189 51.210
589890 97.689 -57.197 51.175
589891 97.719 -57.203 51.140
589892 97.750 -57.211 51.104
589893 97.781 -57.219 51.067
589894 97.811 -57.225 51.030
589895 97.842 -57.233 50.996
589896 97.875 -57.247 50.963
589897 97.908 -57.262 50.931
589898 97.942 -57.271 50.900
589899 97.976 -57.279 50.871
589900 98.010 -57.288 50.845
589901 98.045 -57.291 50.816
589902 98.081 -57.294 50.791
589903 98.116 -57.296 50.766
589904 98.152 -57.297 50.741
589905 98.187 -57.300 50.716
589906 98.222 -57.302 50.690
589907 98.258 -57.304 50.661
589908 98.293 -57.308 50.631
589909 98.329 -57.312 50.599
589910 98.365 -57.316 50.568
589911 98.401 -57.320 50.535
589912 98.436 -57.325 50.504
589913 98.472 -57.329 50.473
589914 98.508 -57.333 50.442
589915 98.544 -57.337 50.412
589916 98.580 -57.341 50.381
589917 98.625 -57.344 50.351
589918 98.663 -57.342 50.322
589919 98.693 -57.336 50.295
589920 98.728 -57.336 50.272
589921 98.766 -57.339 50.249
589922 98.804 -57.342 50.226
589923 98.842 -57.346 50.201
589924 98.880 -57.351 50.176
589925 98.918 -57.354 50.149
589926 98.945 -57.359 50.123
589927 98.952 -57.363 50.098
589928 98.961 -57.368 50.074
589929 98.989 -57.376 50.050
589930 99.015 -57.383 50.023
589931 99.042 -57.395 49.996
589932 99.069 -57.408 49.965
589933 99.096 -57.421 49.931
589934 99.123 -57.433 49.895
589935 99.149 -57.446 49.859
589936 99.170 -57.456 49.821
589937 99.184 -57.463 49.785
589938 99.197 -57.470 49.748
589939 99.215 -57.476 49.710
589940 99.245 -57.480 49.673
589941 99.276 -57.485 49.634
589942 99.307 -57.489 49.594
589943 99.338 -57.494 49.556
589944 99.368 -57.499 49.519
589945 99.399 -57.504 49.483
589946 99.428 -57.509 49.447
589947 99.457 -57.515 49.410
589948 99.485 -57.520 49.373
589949 99.513 -57.526 49.337
589950 99.540 -57.532 49.301
589951 99.568 -57.537 49.264
589952 99.595 -57.542 49.228
589953 99.623 -57.548 49.192
589954 99.652 -57.560 49.155
589955 99.681 -57.577 49.120
589956 99.710 -57.595 49.087
589957 99.738 -57.611 49.057
589958 99.767 -57.629 49.029
589959 99.795 -57.646 49.001
589960 99.824 -57.663 48.976
589961 99.853 -57.680 48.954
589962 99.881 -57.698 48.933
589963 99.910 -57.714 48.913
589964 99.939 -57.732 48.893
589965 99.968 -57.749 48.874
590308 88.465 -53.783 57.535
590309 88.500 -53.798 57.537
590310 88.536 -53.813 57.538
590311 88.571 -53.828 57.539
590312 88.607 -53.844 57.539
590313 88.642 -53.860 57.538
590314 88.678 -53.875 57.536
590315 88.715 -53.894 57.535
590316 88.754 -53.914 57.532
590317 88.791 -53.933 57.528
590318 88.830 -53.953 57.524
590319 88.868 -53.972 57.518
590320 88.906 -53.991 57.510
590321 88.957 -54.011 57.502
590322 88.999 -54.033 57.492
590323 89.034 -54.057 57.481
590324 89.083 -54.064 57.468
590325 89.141 -54.061 57.450
590326 89.173 -54.083 57.431
590327 89.205 -54.106 57.409
590328 89.238 -54.129 57.383
590329 89.270 -54.151 57.355
590330 89.302 -54.173 57.323
590331 89.334 -54.196 57.286
590332 89.367 -54.219 57.246
590333 89.399 -54.241 57.203
590334 89.431 -54.264 57.159
590335 89.467 -54.283 57.115
590336 89.506 -54.299 57.070
590337 89.543 -54.316 57.023
590338 89.582 -54.333 56.976
590339 89.620 -54.350 56.930
590347 89.924 -54.484 56.433
590348 89.963 -54.501 56.397
590349 90.001 -54.518 56.361
590350 90.039 -54.535 56.327
590351 90.077 -54.553 56.295
590352 90.116 -54.576 56.264
590353 90.155 -54.599 56.235
590354 90.193 -54.623 56.209
590355 90.232 -54.646 56.183
590356 90.270 -54.670 56.158
590357 90.309 -54.693 56.136
590358 90.347 -54.717 56.116
590359 90.374 -54.733 56.102
590360 90.398 -54.747 56.088
590361 90.421 -54.762 56.075
590362 90.444 -54.775 56.064
590363 90.467 -54.790 56.051
590364 90.493 -54.803 56.038
590365 90.518 -54.817 56.023
590366 90.544 -54.831 56.002
590367 90.570 -54.844 55.976
590368 90.595 -54.858 55.948
590369 90.621 -54.872 55.920
590370 90.642 -54.888 55.899
590371 90.660 -54.906 55.881
590372 90.678 -54.924 55.868
590373 90.696 -54.941 55.859
590374 90.715 -54.960 55.852
590375 90.741 -54.976 55.848
590376 90.778 -54.991 55.846
590377 90.815 -55.007 55.844
590378 90.852 -55.022 55.844
590379 90.888 -55.038 55.845
590380 90.925 -55.053 55.843
590381 90.962 -55.068 55.839
590382 90.999 -55.083 55.834
590383 91.036 -55.098 55.827
590384 91.073 -55.114 55.819
590385 91.109 -55.129 55.812
590386 91.146 -55.144 55.809
590387 91.183 -55.160 55.808
590388 91.220 -55.175 55.799
590389 91.256 -55.191 55.782
590390 91.293 -55.206 55.757
590391 91.330 -55.221 55.730
590392 91.367 -55.236 55.710
590393 91.404 -55.251 55.696
590394 91.447 -55.260 55.681
590395 91.480 -55.264 55.656
590396 91.515 -55.275 55.631
590397 91.551 -55.286 55.613
590398 91.586 -55.297 55.601
590399 91.622 -55.307 55.594
590400 91.657 -55.318 55.581
590401 91.693 -55.329 55.553
590402 91.728 -55.340 55.515
590403 91.764 -55.351 55.477
590404 91.799 -55.361 55.447
590405 91.835 -55.373 55.420
590406 91.870 -55.383 55.396
590407 91.906 -55.394 55.374
590408 91.941 -55.405 55.353
590409 91.977 -55.416 55.336
590410 92.012 -55.426 55.321
590411 92.047 -55.439 55.311
590412 92.081 -55.456 55.304
590413 92.115 -55.473 55.302
590414 92.148 -55.489 55.306
590415 92.182 -55.506 55.316
590416 92.217 -55.524 55.332
590417 92.252 -55.540 55.341
590418 92.287 -55.557 55.349
590419 92.321 -55.574 55.353
590420 92.356 -55.590 55.352
590421 92.396 -55.605 55.345
590422 92.431 -55.623 55.332
590423 92.462 -55.643 55.310
590424 92.493 -55.662 55.282
590425 92.524 -55.681 55.259
590426 92.559 -55.701 55.242
590427 92.595 -55.720 55.233
590428 92.630 -55.740 55.234
590429 92.666 -55.759 55.237
590430 92.702 -55.778 55.235
590431 92.737 -55.797 55.230
590432 92.773 -55.817 55.220
590433 92.809 -55.836 55.206
590434 92.843 -55.855 55.195
590435 92.873 -55.873 55.191
590436 92.904 -55.891 55.192
590437 92.934 -55.909 55.201
590438 92.964 -55.928 55.215
590439 92.994 -55.946 55.229
590440 93.029 -55.966 55.245
590441 93.066 -55.987 55.259
590442 93.103 -56.006 55.268
590443 93.139 -56.005 55.275
590444 93.173 -56.008 55.281
590445 93.207 -56.010 55.290
590446 93.241 -56.013 55.303
590447 93.276 -56.016 55.313
590448 93.310 -56.019 55.319
590449 93.344 -56.021 55.318
590450 93.379 -56.024 55.312
590451 93.413 -56.026 55.299
590452 93.447 -56.028 55.282
590453 93.482 -56.031 55.260
590454 93.514 -56.043 55.238
590455 93.545 -56.065 55.218
590456 93.576 -56.088 55.200
590457 93.607 -56.110 55.184
590458 93.638 -56.132 55.163
590459 93.670 -56.154 55.142
590460 93.700 -56.176 55.121
590461 93.731 -56.198 55.104
590462 93.763 -56.219 55.088
590463 93.795 -56.235 55.073
590464 93.827 -56.252 55.056
590465 93.859 -56.269 55.037
590466 93.890 -56.285 55.016
590467 93.920 -56.301 54.994
590468 93.950 -56.317 54.971
590469 93.980 -56.332 54.947
590470 94.011 -56.348 54.917
590471 94.041 -56.364 54.885
590472 94.071 -56.379 54.849
590473 94.102 -56.395 54.810
590474 94.132 -56.411 54.768
590475 94.162 -56.426 54.724
590476 94.192 -56.442 54.675
590477 94.222 -56.458 54.625
590478 94.253 -56.473 54.573
590479 94.283 -56.489 54.519
590480 94.312 -56.500 54.464
590481 94.340 -56.509 54.407
590482 94.368 -56.518 54.349
590483 94.396 -56.527 54.296
590484 94.424 -56.536 54.246
590485 94.453 -56.545 54.199
590486 94.480 -56.553 54.154
590487 94.508 -56.561 54.108
590488 94.544 -56.571 54.057
590489 94.579 -56.582 54.005
590490 94.615 -56.592 53.955
590491 94.650 -56.602 53.905
590492 94.686 -56.611 53.857
590493 94.722 -56.620 53.810
590494 94.758 -56.630 53.766
590495 94.793 -56.640 53.727
590496 94.827 -56.652 53.692
590497 94.856 -56.666 53.658
590498 94.885 -56.680 53.622
590499 94.913 -56.693 53.588
590500 94.941 -56.705 53.552
590501 94.970 -56.717 53.509
590502 94.998 -56.727 53.460
590503 95.027 -56.739 53.406
590504 95.056 -56.750 53.348
590505 95.085 -56.762 53.286
590506 95.113 -56.772 53.230
590507 95.142 -56.783 53.181
590508 95.171 -56.795 53.137
590509 95.199 -56.806 53.096
590510 95.236 -56.817 53.059
590511 95.284 -56.825 53.024
590512 95.335 -56.822 52.988
590513 95.367 -56.830 52.951
590514 95.394 -56.839 52.912
590515 95.421 -56.849 52.873
590516 95.449 -56.858 52.835
590517 95.476 -56.868 52.802
590518 95.503 -56.877 52.772
590519 95.542 -56.886 52.743
590520 95.581 -56.894 52.714
590521 95.619 -56.903 52.686
590522 95.659 -56.911 52.658
590523 95.697 -56.919 52.629
590524 95.736 -56.928 52.601
590525 95.774 -56.936 52.571
590526 95.810 -56.943 52.537
590527 95.847 -56.950 52.500
590528 95.877 -56.957 52.459
590529 95.898 -56.963 52.417
590530 95.920 -56.970 52.379
590531 95.942 -56.977 52.344
590532 95.963 -56.984 52.308
590533 95.985 -56.990 52.271
590534 96.007 -56.997 52.232
590535 96.028 -57.004 52.193
590536 96.049 -57.011 52.160
590537 96.072 -57.019 52.131
590538 96.095 -57.029 52.105
590539 96.118 -57.040 52.080
590540 96.141 -57.050 52.056
590541 96.164 -57.059 52.031
590542 96.187 -57.069 52.011
590543 96.210 -57.079 51.993
590544 96.233 -57.089 51.975
590545 96.261 -57.093 51.955
590546 96.289 -57.097 51.933
590547 96.317 -57.100 51.913
590548 96.344 -57.103 51.898
590549 96.373 -57.107 51.886
590550 96.400 -57.111 51.874
590551 96.429 -57.114 51.861
590552 96.457 -57.118 51.844
590553 96.485 -57.120 51.823
590554 96.514 -57.123 51.800
590555 96.542 -57.123 51.778
590556 96.571 -57.124 51.755
590557 96.599 -57.125 51.728
590558 96.627 -57.125 51.697
590559 96.661 -57.135 51.663
590560 96.694 -57.144 51.631
590561 96.727 -57.153 51.603
590562 96.760 -57.163 51.579
590563 96.793 -57.170 51.559
590564 96.825 -57.176 51.542
590565 96.857 -57.183 51.522
590566 96.889 -57.190 51.504
590567 96.922 -57.196 51.491
590568 96.954 -57.203 51.478
590569 96.986 -57.210 51.466
590570 97.019 -57.202 51.457
590571 97.051 -57.193 51.452
590572 97.084 -57.184 51.449
590573 97.116 -57.175 51.442
590574 97.149 -57.166 51.430
590575 97.181 -57.157 51.419
590576 97.214 -57.149 51.411
590577 97.246 -57.140 51.409
590578 97.279 -57.131 51.405
590579 97.316 -57.128 51.400
590580 97.353 -57.126 51.394
590581 97.390 -57.124 51.384
590582 97.423 -57.122 51.370
590583 97.450 -57.119 51.353
590584 97.478 -57.115 51.332
590585 97.505 -57.112 51.310
590586 97.536 -57.116 51.286
590587 97.569 -57.128 51.260
590588 97.599 -57.135 51.232
590589 97.630 -57.142 51.202
590590 97.660 -57.150 51.170
590591 97.690 -57.157 51.137
590592 97.721 -57.164 51.103
590593 97.752 -57.172 51.069
590594 97.782 -57.178 51.035
590595 97.814 -57.190 51.002
590596 97.848 -57.204 50.969
590597 97.881 -57.219 50.938
590598 97.915 -57.231 50.907
590599 97.949 -57.239 50.877
590600 97.983 -57.247 50.847
590601 98.017 -57.257 50.821
590602 98.051 -57.265 50.793
590603 98.086 -57.267 50.769
590604 98.122 -57.269 50.745
590605 98.157 -57.272 50.720
590606 98.193 -57.274 50.695
590607 98.228 -57.276 50.671
590608 98.264 -57.278 50.645
590609 98.299 -57.280 50.616
590610 98.334 -57.282 50.586
590611 98.369 -57.285 50.555
590612 98.405 -57.287 50.524
590613 98.440 -57.290 50.492
590614 98.476 -57.294 50.461
590615 98.512 -57.298 50.429
590616 98.548 -57.302 50.399
590617 98.583 -57.307 50.369
590618 98.620 -57.310 50.339
590619 98.663 -57.313 50.312
590620 98.694 -57.307 50.287
590621 98.731 -57.309 50.263
590622 98.769 -57.313 50.241
590623 98.807 -57.316 50.219
590624 98.845 -57.321 50.194
590625 98.883 -57.325 50.169
590626 98.921 -57.329 50.143
590627 98.955 -57.333 50.118
590628 98.963 -57.338 50.093
590629 98.971 -57.342 50.070
590630 98.979 -57.347 50.047
590631 99.006 -57.357 50.020
590632 99.033 -57.370 49.992
590633 99.060 -57.383 49.960
590634 99.086 -57.395 49.925
590635 99.113 -57.408 49.889
590636 99.140 -57.420 49.854
590637 99.166 -57.432 49.817
590638 99.179 -57.439 49.780
590639 99.192 -57.446 49.742
590640 99.216 -57.451 49.704
590641 99.246 -57.456 49.666
590642 99.277 -57.461 49.626
590643 99.308 -57.466 49.586
590644 99.339 -57.470 49.547
590645 99.369 -57.475 49.510
590646 99.400 -57.479 49.475
590647 99.430 -57.485 49.441
590648 99.461 -57.489 49.406
590649 99.489 -57.495 49.370
590650 99.516 -57.501 49.335
590651 99.544 -57.508 49.300
590652 99.572 -57.514 49.264
590653 99.600 -57.520 49.228
590654 99.627 -57.526 49.192
590655 99.655 -57.539 49.154
590656 99.683 -57.557 49.119
590657 99.710 -57.575 49.085
590658 99.738 -57.592 49.055
590659 99.766 -57.611 49.025
590660 99.794 -57.629 48.997
590661 99.822 -57.646 48.972
590662 99.849 -57.664 48.949
590663 99.877 -57.682 48.927
590664 99.905 -57.700 48.907
590665 99.932 -57.718 48.887
590666 99.961 -57.736 48.868
591009 88.490 -53.748 57.529
591010 88.525 -53.763 57.531
591011 88.560 -53.778 57.531
591012 88.596 -53.794 57.532
591013 88.631 -53.809 57.532
591014 88.667 -53.825 57.530
591015 88.705 -53.844 57.529
591016 88.743 -53.864 57.527
591017 88.781 -53.883 57.524
591018 88.819 -53.903 57.520
591019 88.857 -53.922 57.516
591020 88.895 -53.941 57.509
591021 88.934 -53.960 57.501
591022 88.985 -53.980 57.493
591023 89.030 -54.002 57.483
591024 89.065 -54.026 57.471
591025 89.117 -54.030 57.458
591026 89.176 -54.025 57.441
591027 89.221 -54.035 57.421
591028 89.253 -54.057 57.400
591029 89.285 -54.079 57.376
591030 89.318 -54.102 57.349
591031 89.350 -54.125 57.319
591032 89.383 -54.147 57.283
591033 89.415 -54.170 57.244
591034 89.447 -54.192 57.202
591035 89.479 -54.214 57.160
591036 89.512 -54.237 57.116
591037 89.544 -54.260 57.071
591038 89.577 -54.283 57.025
591039 89.609 -54.305 56.978
591040 89.641 -54.327 56.932
591048 89.947 -54.461 56.437
591049 89.985 -54.478 56.401
591050 90.023 -54.495 56.366
591051 90.061 -54.511 56.332
591052 90.099 -54.528 56.300
591053 90.138 -54.545 56.270
591054 90.176 -54.568 56.242
591055 90.215 -54.591 56.216
591056 90.253 -54.615 56.190
591057 90.292 -54.638 56.166
591058 90.330 -54.662 56.144
591059 90.369 -54.685 56.127
591060 90.393 -54.699 56.114
591061 90.416 -54.714 56.102
591062 90.440 -54.728 56.090
591063 90.463 -54.742 56.079
591064 90.488 -54.756 56.067
591065 90.514 -54.770 56.053
591066 90.540 -54.784 56.034
591067 90.566 -54.797 56.010
591068 90.591 -54.811 55.982
591069 90.610 -54.828 55.953
591070 90.628 -54.846 55.925
591071 90.647 -54.864 55.904
591072 90.665 -54.881 55.887
591073 90.683 -54.900 55.873
591074 90.701 -54.917 55.863
591075 90.719 -54.935 55.856
591076 90.750 -54.951 55.851
591077 90.787 -54.966 55.848
591078 90.823 -54.981 55.844
591079 90.861 -54.997 55.841
591080 90.897 -55.013 55.839
591081 90.934 -55.028 55.837
591082 90.971 -55.043 55.831
591083 91.008 -55.058 55.823
591084 91.044 -55.074 55.814
591085 91.081 -55.089 55.804
591086 91.118 -55.104 55.796
591087 91.155 -55.119 55.792
591088 91.191 -55.135 55.788
591089 91.228 -55.151 55.778
591090 91.265 -55.166 55.761
591091 91.302 -55.181 55.737
591092 91.339 -55.195 55.710
591093 91.376 -55.208 55.689
591094 91.413 -55.222 55.675
591095 91.464 -55.228 55.659
591096 91.498 -55.232 55.636
591097 91.528 -55.235 55.614
591098 91.564 -55.247 55.597
591099 91.599 -55.257 55.588
591100 91.635 -55.268 55.582
591101 91.670 -55.279 55.568
591102 91.705 -55.290 55.540
591103 91.741 -55.301 55.502
591104 91.776 -55.311 55.465
591105 91.812 -55.322 55.438
591106 91.847 -55.333 55.413
591107 91.883 -55.344 55.390
591108 91.918 -55.354 55.369
591109 91.954 -55.365 55.349
591110 91.989 -55.376 55.332
591111 92.024 -55.389 55.319
591112 92.058 -55.405 55.310
591113 92.092 -55.422 55.304
591114 92.125 -55.439 55.305
591115 92.159 -55.456 55.311
591116 92.194 -55.473 55.324
591117 92.228 -55.489 55.340
591118 92.263 -55.506 55.349
591119 92.298 -55.523 55.359
591120 92.333 -55.539 55.366
591121 92.370 -55.555 55.369
591122 92.411 -55.571 55.365
591123 92.451 -55.586 55.354
591124 92.482 -55.605 55.335
591125 92.513 -55.624 55.311
591126 92.547 -55.644 55.287
591127 92.582 -55.663 55.268
591128 92.618 -55.683 55.253
591129 92.653 -55.702 55.253
591130 92.689 -55.721 55.258
591131 92.725 -55.740 55.259
591132 92.757 -55.759 55.256
591133 92.787 -55.777 55.250
591134 92.817 -55.795 55.237
591135 92.848 -55.813 55.223
591136 92.878 -55.831 55.207
591137 92.908 -55.850 55.200
591138 92.938 -55.868 55.203
591139 92.969 -55.886 55.212
591140 93.002 -55.906 55.227
591141 93.040 -55.927 55.240
591142 93.076 -55.949 55.254
591143 93.114 -55.970 55.267
591144 93.150 -55.975 55.277
591145 93.185 -55.975 55.284
591146 93.219 -55.978 55.291
591147 93.254 -55.980 55.301
591148 93.288 -55.983 55.312
591149 93.322 -55.985 55.317
591150 93.356 -55.987 55.316
591151 93.391 -55.991 55.310
591152 93.425 -55.993 55.298
591153 93.459 -55.996 55.284
591154 93.493 -55.998 55.266
591155 93.526 -56.009 55.248
591156 93.557 -56.031 55.230
591157 93.588 -56.054 55.214
591158 93.619 -56.076 55.197
591159 93.650 -56.098 55.175
591160 93.681 -56.121 55.153
591161 93.712 -56.143 55.133
591162 93.743 -56.165 55.117
591163 93.774 -56.188 55.104
591164 93.805 -56.210 55.091
591165 93.837 -56.229 55.072
591166 93.867 -56.245 55.055
591167 93.898 -56.260 55.035
591168 93.928 -56.276 55.013
591169 93.958 -56.292 54.990
591170 93.988 -56.307 54.964
591171 94.018 -56.323 54.938
591172 94.049 -56.339 54.904
591173 94.079 -56.355 54.868
591174 94.110 -56.370 54.831
591175 94.140 -56.386 54.791
591176 94.170 -56.402 54.749
591177 94.200 -56.418 54.698
591178 94.230 -56.433 54.649
591179 94.261 -56.449 54.598
591180 94.291 -56.465 54.543
591181 94.319 -56.474 54.487
591182 94.347 -56.483 54.428
591183 94.376 -56.491 54.368
591184 94.403 -56.498 54.312
591185 94.431 -56.507 54.258
591186 94.459 -56.514 54.208
591187 94.486 -56.522 54.161
591188 94.514 -56.530 54.113
591189 94.550 -56.540 54.059
591190 94.585 -56.550 54.003
591191 94.621 -56.561 53.949
591192 94.657 -56.570 53.897
591193 94.692 -56.581 53.847
591194 94.728 -56.591 53.796
591195 94.763 -56.602 53.750
591196 94.799 -56.612 53.709
591197 94.834 -56.622 53.672
591198 94.867 -56.628 53.635
591199 94.895 -56.642 53.599
591200 94.924 -56.653 53.563
591201 94.953 -56.664 53.527
591202 94.981 -56.676 53.487
591203 95.010 -56.687 53.441
591204 95.039 -56.698 53.390
591205 95.067 -56.709 53.335
591206 95.096 -56.721 53.277
591207 95.125 -56.732 53.216
591208 95.154 -56.743 53.161
591209 95.182 -56.754 53.112
591210 95.211 -56.765 53.067
591211 95.251 -56.777 53.025
591212 95.298 -56.787 52.989
591213 95.348 -56.788 52.953
591214 95.376 -56.797 52.917
591215 95.403 -56.807 52.879
591216 95.430 -56.816 52.841
591217 95.458 -56.826 52.801
591218 95.485 -56.835 52.765
591219 95.513 -56.845 52.736
591220 95.546 -56.854 52.707
591221 95.585 -56.862 52.678
591222 95.623 -56.871 52.648
591223 95.662 -56.879 52.620
591224 95.701 -56.887 52.590
591225 95.740 -56.896 52.561
591226 95.777 -56.903 52.530
591227 95.814 -56.911 52.497
591228 95.850 -56.918 52.461
591229 95.880 -56.924 52.423
591230 95.902 -56.931 52.383
591231 95.923 -56.938 52.347
591232 95.945 -56.945 52.311
591233 95.967 -56.952 52.276
591234 95.988 -56.959 52.239
591235 96.010 -56.965 52.199
591236 96.032 -56.972 52.159
591237 96.053 -56.980 52.125
591238 96.077 -56.990 52.095
591239 96.100 -57.000 52.068
591240 96.123 -57.010 52.043
591241 96.146 -57.020 52.019
591242 96.169 -57.030 51.996
591243 96.192 -57.040 51.976
591244 96.213 -57.049 51.960
591245 96.236 -57.058 51.941
591246 96.264 -57.063 51.921
591247 96.292 -57.067 51.900
591248 96.320 -57.070 51.880
591249 96.347 -57.074 51.866
591250 96.376 -57.078 51.855
591251 96.403 -57.081 51.844
591252 96.432 -57.084 51.831
591253 96.459 -57.087 51.816
591254 96.488 -57.087 51.795
591255 96.516 -57.087 51.774
591256 96.544 -57.088 51.751
591257 96.573 -57.089 51.725
591258 96.601 -57.090 51.697
591259 96.629 -57.091 51.667
591260 96.660 -57.094 51.634
591261 96.693 -57.103 51.604
591262 96.726 -57.113 51.577
591263 96.759 -57.122 51.553
591264 96.792 -57.131 51.530
591265 96.825 -57.141 51.510
591266 96.858 -57.150 51.490
591267 96.891 -57.160 51.472
591268 96.924 -57.166 51.454
591269 96.956 -57.172 51.438
591270 96.989 -57.179 51.425
591271 97.021 -57.173 51.415
591272 97.053 -57.164 51.410
591273 97.086 -57.156 51.405
591274 97.118 -57.147 51.398
591275 97.151 -57.138 51.388
591276 97.183 -57.129 51.377
591277 97.216 -57.120 51.371
591278 97.248 -57.111 51.370
591279 97.281 -57.103 51.364
591280 97.313 -57.094 51.358
591281 97.348 -57.088 51.350
591282 97.385 -57.086 51.339
591283 97.422 -57.084 51.324
591284 97.453 -57.081 51.305
591285 97.480 -57.078 51.286
591286 97.508 -57.075 51.266
591287 97.538 -57.078 51.244
591288 97.570 -57.088 51.219
591289 97.601 -57.095 51.192
591290 97.632 -57.103 51.163
591291 97.662 -57.110 51.132
591292 97.692 -57.117 51.100
591293 97.723 -57.125 51.067
591294 97.754 -57.133 51.036
591295 97.787 -57.147 51.006
591296 97.821 -57.162 50.975
591297 97.854 -57.176 50.943
591298 97.887 -57.191 50.912
591299 97.922 -57.199 50.883
591300 97.956 -57.207 50.854
591301 97.990 -57.216 50.824
591302 98.024 -57.225 50.798
591303 98.058 -57.233 50.771
591304 98.092 -57.241 50.747
591305 98.127 -57.243 50.724
591306 98.163 -57.245 50.699
591307 98.198 -57.247 50.674
591308 98.233 -57.250 50.652
591309 98.269 -57.252 50.628
591310 98.304 -57.254 50.600
591311 98.340 -57.257 50.571
591312 98.375 -57.258 50.541
591313 98.411 -57.260 50.511
591314 98.446 -57.263 50.480
591315 98.481 -57.265 50.448
591316 98.517 -57.267 50.415
591317 98.552 -57.269 50.385
591318 98.587 -57.272 50.355
591319 98.623 -57.275 50.327
591320 98.659 -57.280 50.302
591321 98.696 -57.279 50.278
591322 98.734 -57.283 50.254
591323 98.772 -57.287 50.232
591324 98.810 -57.291 50.210
591325 98.848 -57.295 50.187
591326 98.886 -57.299 50.162
591327 98.924 -57.304 50.136
591328 98.962 -57.307 50.111
591329 98.973 -57.312 50.087
591330 98.981 -57.316 50.064
591331 98.992 -57.322 50.040
591332 99.008 -57.334 50.014
591333 99.024 -57.345 49.985
591334 99.050 -57.357 49.952
591335 99.077 -57.370 49.918
591336 99.104 -57.382 49.882
591337 99.131 -57.395 49.847
591338 99.157 -57.408 49.811
591339 99.174 -57.416 49.773
591340 99.188 -57.423 49.735
591341 99.217 -57.428 49.696
591342 99.248 -57.432 49.656
591343 99.278 -57.437 49.616
591344 99.309 -57.442 49.576
591345 99.340 -57.447 49.538
591346 99.370 -57.451 49.501
591347 99.403 -57.454 49.466
591348 99.436 -57.457 49.433
591349 99.469 -57.460 49.400
591350 99.494 -57.472 49.365
591351 99.521 -57.481 49.330
591352 99.549 -57.487 49.297
591353 99.576 -57.493 49.262
591354 99.604 -57.499 49.226
591355 99.632 -57.505 49.189
591356 99.659 -57.517 49.152
591357 99.687 -57.534 49.116
591358 99.715 -57.552 49.082
591359 99.743 -57.570 49.051
591360 99.771 -57.588 49.021
591361 99.798 -57.606 48.992
591362 99.826 -57.623 48.966
591363 99.854 -57.642 48.943
591364 99.882 -57.659 48.921
591365 99.910 -57.677 48.900
591366 99.937 -57.695 48.880
591367 99.965 -57.713 48.861
591710 88.515 -53.712 57.523
591711 88.550 -53.728 57.524
591712 88.585 -53.744 57.525
591713 88.620 -53.759 57.525
591714 88.657 -53.775 57.524
591715 88.695 -53.794 57.523
591716 88.733 -53.814 57.521
591717 88.771 -53.833 57.518
591718 88.809 -53.853 57.515
591719 88.847 -53.872 57.511
591720 88.885 -53.891 57.507
591721 88.923 -53.910 57.501
591722 88.963 -53.930 57.492
591723 89.014 -53.950 57.484
591724 89.061 -53.971 57.473
591725 89.096 -53.994 57.461
591726 89.150 -53.996 57.448
591727 89.209 -53.991 57.431
591728 89.248 -54.005 57.413
591729 89.283 -54.024 57.393
591730 89.320 -54.044 57.369
591731 89.355 -54.063 57.344
591732 89.392 -54.082 57.314
591733 89.427 -54.101 57.280
591734 89.463 -54.121 57.242
591735 89.495 -54.143 57.202
591736 89.528 -54.166 57.160
591737 89.560 -54.189 57.116
591738 89.592 -54.211 57.071
591739 89.624 -54.233 57.026
591740 89.657 -54.256 56.980
591741 89.689 -54.278 56.934
591742 89.722 -54.301 56.889
591748 89.931 -54.421 56.478
591749 89.968 -54.438 56.441
591750 90.007 -54.455 56.405
591751 90.045 -54.471 56.371
591752 90.083 -54.488 56.336
591753 90.121 -54.505 56.304
591754 90.159 -54.522 56.275
591755 90.197 -54.539 56.248
591756 90.236 -54.559 56.222
591757 90.274 -54.583 56.198
591758 90.313 -54.606 56.174
591759 90.352 -54.630 56.153
591760 90.389 -54.652 56.137
591761 90.412 -54.667 56.125
591762 90.436 -54.681 56.114
591763 90.459 -54.695 56.103
591764 90.484 -54.709 56.091
591765 90.510 -54.722 56.076
591766 90.536 -54.736 56.059
591767 90.559 -54.751 56.038
591768 90.578 -54.768 56.013
591769 90.596 -54.786 55.984
591770 90.614 -54.804 55.956
591771 90.633 -54.822 55.929
591772 90.651 -54.840 55.908
591773 90.669 -54.857 55.891
591774 90.687 -54.875 55.878
591775 90.706 -54.893 55.868
591776 90.724 -54.911 55.860
591777 90.759 -54.926 55.854
591778 90.796 -54.941 55.850
591779 90.832 -54.957 55.844
591780 90.869 -54.972 55.838
591781 90.906 -54.988 55.834
591782 90.943 -55.003 55.830
591783 90.979 -55.018 55.823
591784 91.016 -55.034 55.813
591785 91.053 -55.049 55.802
591786 91.090 -55.064 55.790
591787 91.127 -55.079 55.779
591788 91.163 -55.094 55.773
591789 91.200 -55.110 55.765
591790 91.237 -55.126 55.753
591791 91.274 -55.139 55.736
591792 91.311 -55.153 55.714
591793 91.348 -55.166 55.691
591794 91.385 -55.180 55.672
591795 91.428 -55.191 55.656
591796 91.481 -55.195 55.637
591797 91.515 -55.199 55.617
591798 91.545 -55.203 55.597
591799 91.576 -55.207 55.582
591800 91.611 -55.218 55.574
591801 91.647 -55.229 55.570
591802 91.682 -55.239 55.556
591803 91.718 -55.251 55.527
591804 91.753 -55.261 55.490
591805 91.789 -55.272 55.455
591806 91.824 -55.282 55.430
591807 91.860 -55.294 55.407
591808 91.895 -55.304 55.385
591809 91.931 -55.315 55.364
591810 91.966 -55.326 55.346
591811 92.001 -55.338 55.330
591812 92.035 -55.355 55.319
591813 92.069 -55.372 55.310
591814 92.102 -55.389 55.306
591815 92.136 -55.405 55.306
591816 92.170 -55.422 55.313
591817 92.205 -55.439 55.327
591818 92.239 -55.455 55.343
591819 92.274 -55.472 55.355
591820 92.309 -55.489 55.365
591821 92.345 -55.505 55.374
591822 92.385 -55.520 55.380
591823 92.425 -55.536 55.380
591824 92.466 -55.551 55.370
591825 92.503 -55.568 55.353
591826 92.534 -55.587 55.337
591827 92.570 -55.606 55.317
591828 92.605 -55.626 55.293
591829 92.641 -55.645 55.271
591830 92.671 -55.663 55.268
591831 92.701 -55.681 55.274
591832 92.731 -55.699 55.279
591833 92.762 -55.718 55.279
591834 92.792 -55.735 55.276
591835 92.822 -55.753 55.266
591836 92.852 -55.771 55.250
591837 92.882 -55.790 55.231
591838 92.913 -55.808 55.218
591839 92.943 -55.826 55.213
591840 92.975 -55.845 55.216
591841 93.012 -55.866 55.227
591842 93.050 -55.888 55.241
591843 93.086 -55.909 55.255
591844 93.124 -55.931 55.269
591845 93.160 -55.946 55.280
591846 93.196 -55.944 55.288
591847 93.231 -55.944 55.295
591848 93.266 -55.947 55.302
591849 93.300 -55.950 55.312
591850 93.334 -55.952 55.317
591851 93.368 -55.955 55.317
591852 93.403 -55.957 55.315
591853 93.437 -55.960 55.307
591854 93.471 -55.962 55.296
591855 93.505 -55.965 55.281
591856 93.538 -55.976 55.267
591857 93.569 -55.998 55.251
591858 93.600 -56.021 55.233
591859 93.631 -56.043 55.216
591860 93.662 -56.065 55.193
591861 93.693 -56.087 55.170
591862 93.724 -56.110 55.151
591863 93.755 -56.132 55.134
591864 93.786 -56.154 55.119
591865 93.817 -56.176 55.105
591866 93.840 -56.204 55.085
591867 93.868 -56.216 55.066
591868 93.899 -56.232 55.046
591869 93.929 -56.248 55.024
591870 93.960 -56.264 54.999
591871 93.990 -56.280 54.975
591872 94.020 -56.295 54.944
591873 94.051 -56.311 54.912
591874 94.081 -56.327 54.876
591875 94.112 -56.343 54.839
591876 94.142 -56.359 54.800
591877 94.173 -56.375 54.759
591878 94.203 -56.391 54.711
591879 94.234 -56.406 54.663
591880 94.264 -56.422 54.614
591881 94.295 -56.436 54.561
591882 94.326 -56.445 54.507
591883 94.353 -56.452 54.451
591884 94.381 -56.460 54.389
591885 94.409 -56.467 54.329
591886 94.436 -56.475 54.272
591887 94.464 -56.483 54.218
591888 94.492 -56.490 54.167
591889 94.521 -56.498 54.117
591890 94.557 -56.508 54.062
591891 94.592 -56.519 54.005
591892 94.627 -56.529 53.948
591893 94.663 -56.539 53.893
591894 94.698 -56.550 53.841
591895 94.734 -56.560 53.787
591896 94.769 -56.570 53.737
591897 94.805 -56.580 53.693
591898 94.844 -56.574 53.652
591899 94.883 -56.570 53.613
591900 94.911 -56.583 53.575
591901 94.939 -56.596 53.537
591902 94.967 -56.609 53.500
591903 94.996 -56.623 53.461
591904 95.024 -56.636 53.417
591905 95.052 -56.649 53.369
591906 95.080 -56.662 53.317
591907 95.109 -56.675 53.262
591908 95.137 -56.688 53.202
591909 95.166 -56.702 53.143
591910 95.194 -56.713 53.090
591911 95.223 -56.724 53.042
591912 95.265 -56.736 52.997
591913 95.312 -56.747 52.958
591914 95.357 -56.755 52.920
591915 95.385 -56.765 52.883
591916 95.412 -56.774 52.846
591917 95.439 -56.784 52.808
591918 95.467 -56.793 52.769
591919 95.494 -56.803 52.732
591920 95.522 -56.812 52.698
591921 95.549 -56.822 52.667
591922 95.588 -56.830 52.638
591923 95.627 -56.839 52.608
591924 95.666 -56.847 52.578
591925 95.704 -56.855 52.547
591926 95.743 -56.865 52.516
591927 95.780 -56.871 52.486
591928 95.817 -56.879 52.453
591929 95.853 -56.886 52.419
591930 95.884 -56.893 52.385
591931 95.905 -56.899 52.349
591932 95.927 -56.906 52.312
591933 95.949 -56.913 52.276
591934 95.970 -56.920 52.239
591935 95.992 -56.927 52.202
591936 96.014 -56.934 52.164
591937 96.035 -56.940 52.124
591938 96.058 -56.950 52.089
591939 96.082 -56.960 52.059
591940 96.104 -56.971 52.032
591941 96.127 -56.981 52.007
591942 96.151 -56.990 51.984
591943 96.174 -57.000 51.960
591944 96.192 -57.009 51.941
591945 96.212 -57.016 51.924
591946 96.238 -57.027 51.905
591947 96.266 -57.034 51.886
591948 96.295 -57.037 51.865
591949 96.322 -57.040 51.846
591950 96.350 -57.044 51.834
591951 96.378 -57.047 51.823
591952 96.406 -57.051 51.812
591953 96.434 -57.054 51.799
591954 96.462 -57.056 51.784
591955 96.490 -57.055 51.764
591956 96.518 -57.053 51.744
591957 96.546 -57.053 51.720
591958 96.575 -57.054 51.694
591959 96.603 -57.055 51.666
591960 96.631 -57.056 51.636
591961 96.660 -57.056 51.605
591962 96.691 -57.062 51.578
591963 96.725 -57.072 51.552
591964 96.758 -57.081 51.528
591965 96.791 -57.091 51.505
591966 96.824 -57.100 51.484
591967 96.857 -57.109 51.463
591968 96.890 -57.119 51.443
591969 96.923 -57.128 51.424
591970 96.956 -57.138 51.406
591971 96.990 -57.147 51.390
591972 97.022 -57.144 51.376
591973 97.055 -57.136 51.366
591974 97.087 -57.127 51.361
591975 97.120 -57.118 51.353
591976 97.152 -57.109 51.345
591977 97.185 -57.100 51.338
591978 97.217 -57.091 51.333
591979 97.250 -57.083 51.332
591980 97.282 -57.074 51.325
591981 97.315 -57.065 51.315
591982 97.347 -57.056 51.304
591983 97.380 -57.047 51.291
591984 97.417 -57.045 51.276
591985 97.454 -57.043 51.259
591986 97.484 -57.040 51.240
591987 97.511 -57.037 51.220
591988 97.540 -57.039 51.199
591989 97.572 -57.049 51.176
591990 97.602 -57.056 51.151
591991 97.633 -57.063 51.123
591992 97.664 -57.070 51.093
591993 97.694 -57.078 51.063
591994 97.727 -57.090 51.033
591995 97.760 -57.105 51.005
591996 97.794 -57.119 50.977
591997 97.827 -57.134 50.948
591998 97.861 -57.148 50.917
591999 97.894 -57.159 50.887
592000 97.928 -57.167 50.858
592001 97.962 -57.176 50.831
592002 97.997 -57.185 50.803
592003 98.030 -57.193 50.776
592004 98.065 -57.201 50.750
592005 98.098 -57.210 50.725
592006 98.133 -57.216 50.702
592007 98.168 -57.219 50.678
592008 98.204 -57.221 50.654
592009 98.239 -57.223 50.632
592010 98.275 -57.225 50.609
592011 98.310 -57.228 50.582
592012 98.345 -57.230 50.553
592013 98.380 -57.232 50.525
592014 98.416 -57.235 50.496
592015 98.451 -57.237 50.466
592016 98.487 -57.239 50.433
592017 98.522 -57.241 50.400
592018 98.558 -57.243 50.369
592019 98.593 -57.245 50.340
592020 98.628 -57.247 50.313
592021 98.664 -57.250 50.291
592022 98.700 -57.256 50.266
592023 98.737 -57.257 50.243
592024 98.775 -57.261 50.222
592025 98.813 -57.266 50.200
592026 98.851 -57.269 50.177
592027 98.889 -57.274 50.152
592028 98.927 -57.278 50.127
592029 98.966 -57.282 50.101
592030 98.984 -57.286 50.077
592031 98.992 -57.291 50.054
592032 99.006 -57.301 50.030
592033 99.023 -57.312 50.005
592034 99.039 -57.323 49.976
592035 99.056 -57.335 49.943
592036 99.072 -57.345 49.909
592037 99.094 -57.357 49.874
592038 99.121 -57.370 49.839
592039 99.148 -57.382 49.802
592040 99.169 -57.393 49.764
592041 99.188 -57.399 49.725
592042 99.218 -57.404 49.685
592043 99.249 -57.408 49.645
592044 99.279 -57.413 49.605
592045 99.311 -57.416 49.565
592046 99.345 -57.420 49.527
592047 99.378 -57.423 49.490
592048 99.411 -57.426 49.456
592049 99.444 -57.428 49.424
592050 99.476 -57.432 49.391
592051 99.503 -57.439 49.357
592052 99.528 -57.453 49.324
592053 99.554 -57.466 49.291
592054 99.581 -57.472 49.256
592055 99.609 -57.478 49.220
592056 99.637 -57.484 49.184
592057 99.664 -57.494 49.146
592058 99.692 -57.511 49.111
592059 99.720 -57.529 49.077
592060 99.748 -57.548 49.045
592061 99.775 -57.565 49.015
592062 99.803 -57.583 48.986
592063 99.831 -57.601 48.959
592064 99.859 -57.619 48.935
592065 99.887 -57.636 48.913
592066 99.914 -57.655 48.892
592067 99.942 -57.673 48.871
592068 99.970 -57.690 48.851
592069 99.998 -57.708 48.831
592411 88.539 -53.678 57.516
592412 88.574 -53.693 57.517
592413 88.610 -53.709 57.518
592414 88.646 -53.725 57.518
592415 88.684 -53.744 57.517
592416 88.722 -53.764 57.515
592417 88.760 -53.783 57.513
592418 88.798 -53.803 57.511
592419 88.836 -53.822 57.508
592420 88.875 -53.841 57.503
592421 88.912 -53.860 57.499
592422 88.951 -53.880 57.492
592423 88.991 -53.899 57.484
592424 89.042 -53.919 57.474
592425 89.093 -53.939 57.464
592426 89.127 -53.963 57.452
592427 89.184 -53.962 57.438
592428 89.236 -53.963 57.422
592429 89.265 -53.984 57.405
592430 89.302 -54.003 57.385
592431 89.337 -54.022 57.362
592432 89.374 -54.041 57.338
592433 89.409 -54.060 57.310
592434 89.446 -54.079 57.277
592435 89.481 -54.098 57.240
592436 89.518 -54.118 57.201
592437 89.553 -54.137 57.159
592438 89.590 -54.156 57.115
592439 89.625 -54.175 57.071
592440 89.662 -54.195 57.027
592441 89.697 -54.214 56.981
592442 89.734 -54.232 56.934
592443 89.769 -54.252 56.890
592449 89.963 -54.387 56.481
592450 89.995 -54.409 56.444
592451 90.029 -54.431 56.409
592452 90.067 -54.448 56.375
592453 90.105 -54.464 56.341
592454 90.143 -54.482 56.310
592455 90.181 -54.499 56.281
592456 90.219 -54.515 56.253
592457 90.257 -54.532 56.228
592458 90.296 -54.551 56.204
592459 90.334 -54.574 56.181
592460 90.373 -54.598 56.160
592461 90.408 -54.619 56.145
592462 90.432 -54.634 56.135
592463 90.455 -54.647 56.126
592464 90.480 -54.661 56.114
592465 90.506 -54.674 56.099
592466 90.528 -54.691 56.081
592467 90.546 -54.709 56.062
592468 90.565 -54.726 56.039
592469 90.583 -54.744 56.013
592470 90.601 -54.762 55.984
592471 90.619 -54.780 55.957
592472 90.637 -54.797 55.932
592473 90.656 -54.815 55.912
592474 90.674 -54.833 55.896
592475 90.692 -54.851 55.883
592476 90.710 -54.869 55.872
592477 90.731 -54.886 55.865
592478 90.767 -54.901 55.859
592479 90.804 -54.917 55.853
592480 90.841 -54.932 55.846
592481 90.878 -54.947 55.838
592482 90.915 -54.963 55.831
592483 90.951 -54.978 55.822
592484 90.988 -54.994 55.814
592485 91.025 -55.009 55.803
592486 91.062 -55.024 55.789
592487 91.098 -55.039 55.775
592488 91.135 -55.054 55.762
592489 91.172 -55.070 55.752
592490 91.209 -55.084 55.742
592491 91.246 -55.097 55.729
592492 91.283 -55.111 55.713
592493 91.320 -55.125 55.694
592494 91.357 -55.138 55.676
592495 91.394 -55.152 55.661
592496 91.445 -55.158 55.646
592497 91.498 -55.163 55.625
592498 91.533 -55.167 55.604
592499 91.563 -55.170 55.585
592500 91.593 -55.174 55.569
592501 91.624 -55.179 55.560
592502 91.660 -55.189 55.555
592503 91.695 -55.200 55.540
592504 91.731 -55.211 55.513
592505 91.766 -55.222 55.478
592506 91.802 -55.232 55.447
592507 91.837 -55.243 55.424
592508 91.873 -55.254 55.401
592509 91.908 -55.265 55.380
592510 91.944 -55.276 55.361
592511 91.979 -55.288 55.343
592512 92.012 -55.304 55.328
592513 92.046 -55.321 55.319
592514 92.080 -55.338 55.312
592515 92.113 -55.354 55.309
592516 92.147 -55.371 55.309
592517 92.181 -55.388 55.315
592518 92.215 -55.405 55.328
592519 92.250 -55.421 55.343
592520 92.285 -55.439 55.356
592521 92.319 -55.455 55.369
592522 92.359 -55.470 55.379
592523 92.400 -55.486 55.387
592524 92.440 -55.501 55.390
592525 92.481 -55.516 55.386
592526 92.521 -55.531 55.373
592527 92.555 -55.549 55.358
592528 92.585 -55.567 55.337
592529 92.615 -55.585 55.310
592530 92.645 -55.603 55.285
592531 92.675 -55.621 55.277
592532 92.706 -55.640 55.282
592533 92.736 -55.658 55.290
592534 92.766 -55.675 55.295
592535 92.796 -55.693 55.295
592536 92.827 -55.712 55.290
592537 92.857 -55.730 55.275
592538 92.887 -55.748 55.259
592539 92.917 -55.766 55.243
592540 92.948 -55.784 55.231
592541 92.985 -55.806 55.225
592542 93.022 -55.828 55.229
592543 93.060 -55.849 55.241
592544 93.096 -55.871 55.255
592545 93.134 -55.892 55.269
592546 93.171 -55.914 55.281
592547 93.207 -55.914 55.290
592548 93.243 -55.912 55.298
592549 93.277 -55.914 55.304
592550 93.311 -55.916 55.312
592551 93.346 -55.919 55.317
592552 93.380 -55.922 55.318
592553 93.414 -55.924 55.319
592554 93.448 -55.927 55.314
592555 93.483 -55.929 55.304
592556 93.517 -55.932 55.293
592557 93.550 -55.943 55.281
592558 93.581 -55.965 55.267
592559 93.612 -55.987 55.250
592560 93.643 -56.009 55.234
592561 93.674 -56.031 55.212
592562 93.705 -56.053 55.190
592563 93.736 -56.076 55.169
592564 93.767 -56.098 55.150
592565 93.798 -56.120 55.132
592566 93.823 -56.147 55.114
592567 93.843 -56.176 55.093
592568 93.869 -56.193 55.072
592569 93.898 -56.205 55.050
592570 93.928 -56.219 55.027
592571 93.958 -56.235 55.001
592572 93.989 -56.250 54.978
592573 94.019 -56.266 54.943
592574 94.050 -56.282 54.910
592575 94.080 -56.298 54.876
592576 94.111 -56.313 54.839
592577 94.141 -56.329 54.801
592578 94.172 -56.345 54.762
592579 94.202 -56.360 54.715
592580 94.232 -56.376 54.669
592581 94.263 -56.392 54.622
592582 94.296 -56.404 54.572
592583 94.330 -56.413 54.520
592584 94.359 -56.420 54.466
592585 94.387 -56.428 54.406
592586 94.414 -56.436 54.345
592587 94.442 -56.443 54.286
592588 94.470 -56.451 54.230
592589 94.497 -56.458 54.176
592590 94.528 -56.467 54.122
592591 94.563 -56.477 54.065
592592 94.599 -56.488 54.007
592593 94.634 -56.498 53.948
592594 94.670 -56.508 53.891
592595 94.705 -56.518 53.837
592596 94.740 -56.529 53.782
592597 94.776 -56.538 53.729
592598 94.816 -56.531 53.681
592599 94.855 -56.524 53.636
592600 94.894 -56.518 53.594
592601 94.926 -56.524 53.552
592602 94.955 -56.538 53.512
592603 94.983 -56.551 53.472
592604 95.011 -56.564 53.432
592605 95.040 -56.577 53.390
592606 95.068 -56.590 53.344
592607 95.096 -56.604 53.295
592608 95.124 -56.617 53.243
592609 95.153 -56.630 53.187
592610 95.181 -56.644 53.127
592611 95.207 -56.670 53.070
592612 95.232 -56.695 53.019
592613 95.268 -56.709 52.972
592614 95.315 -56.721 52.929
592615 95.362 -56.732 52.889
592616 95.393 -56.733 52.851
592617 95.421 -56.743 52.813
592618 95.448 -56.752 52.776
592619 95.475 -56.762 52.738
592620 95.503 -56.771 52.701
592621 95.530 -56.780 52.663
592622 95.558 -56.790 52.629
592623 95.592 -56.799 52.597
592624 95.630 -56.807 52.567
592625 95.669 -56.816 52.537
592626 95.708 -56.824 52.504
592627 95.747 -56.833 52.473
592628 95.783 -56.840 52.442
592629 95.820 -56.847 52.411
592630 95.856 -56.854 52.379
592631 95.887 -56.861 52.345
592632 95.909 -56.868 52.310
592633 95.931 -56.874 52.275
592634 95.952 -56.881 52.238
592635 95.974 -56.888 52.200
592636 95.995 -56.895 52.163
592637 96.017 -56.902 52.125
592638 96.040 -56.911 52.088
592639 96.063 -56.921 52.055
592640 96.086 -56.931 52.023
592641 96.109 -56.941 51.995
592642 96.132 -56.951 51.970
592643 96.154 -56.960 51.947
592644 96.173 -56.968 51.923
592645 96.191 -56.976 51.904
592646 96.213 -56.984 51.886
592647 96.241 -56.997 51.869
592648 96.269 -57.004 51.850
592649 96.297 -57.008 51.831
592650 96.325 -57.011 51.812
592651 96.353 -57.015 51.799
592652 96.381 -57.018 51.788
592653 96.409 -57.022 51.777
592654 96.437 -57.024 51.765
592655 96.464 -57.025 51.750
592656 96.493 -57.024 51.729
592657 96.521 -57.022 51.709
592658 96.549 -57.022 51.686
592659 96.577 -57.020 51.661
592660 96.605 -57.020 51.635
592661 96.633 -57.020 51.606
592662 96.662 -57.021 51.577
592663 96.690 -57.022 51.552
592664 96.724 -57.031 51.528
592665 96.756 -57.040 51.505
592666 96.790 -57.050 51.483
592667 96.823 -57.059 51.460
592668 96.856 -57.069 51.437
592669 96.889 -57.078 51.417
592670 96.922 -57.087 51.398
592671 96.954 -57.102 51.378
592672 96.985 -57.115 51.358
592673 97.021 -57.113 51.340
592674 97.054 -57.104 51.325
592675 97.087 -57.097 51.318
592676 97.120 -57.088 51.311
592677 97.154 -57.080 51.304
592678 97.186 -57.072 51.300
592679 97.219 -57.063 51.298
592680 97.251 -57.054 51.296
592681 97.284 -57.045 51.287
592682 97.316 -57.037 51.275
592683 97.349 -57.028 51.261
592684 97.381 -57.018 51.246
592685 97.414 -57.010 51.231
592686 97.449 -57.005 51.214
592687 97.486 -57.003 51.196
592688 97.514 -57.000 51.175
592689 97.542 -57.000 51.155
592690 97.574 -57.009 51.133
592691 97.604 -57.016 51.109
592692 97.635 -57.023 51.082
592693 97.666 -57.033 51.054
592694 97.700 -57.047 51.026
592695 97.733 -57.062 50.998
592696 97.767 -57.077 50.972
592697 97.800 -57.091 50.946
592698 97.833 -57.106 50.919
592699 97.867 -57.119 50.889
592700 97.901 -57.127 50.861
592701 97.935 -57.136 50.835
592702 97.969 -57.144 50.810
592703 98.003 -57.153 50.785
592704 98.037 -57.161 50.759
592705 98.071 -57.169 50.731
592706 98.105 -57.178 50.704
592707 98.139 -57.187 50.679
592708 98.174 -57.193 50.656
592709 98.209 -57.195 50.633
592710 98.244 -57.197 50.611
592711 98.280 -57.199 50.588
592712 98.315 -57.201 50.562
592713 98.351 -57.203 50.535
592714 98.386 -57.206 50.508
592715 98.422 -57.208 50.480
592716 98.457 -57.210 50.450
592717 98.492 -57.213 50.417
592718 98.528 -57.215 50.385
592719 98.563 -57.217 50.355
592720 98.598 -57.219 50.325
592721 98.634 -57.221 50.299
592722 98.668 -57.222 50.276
592723 98.705 -57.235 50.252
592724 98.743 -57.241 50.230
592725 98.780 -57.241 50.210
592726 98.817 -57.242 50.188
592727 98.855 -57.244 50.166
592728 98.892 -57.248 50.141
592729 98.931 -57.252 50.115
592730 98.969 -57.257 50.090
592731 98.994 -57.260 50.066
592732 99.005 -57.267 50.043
592733 99.021 -57.279 50.019
592734 99.038 -57.290 49.993
592735 99.054 -57.301 49.964
592736 99.070 -57.312 49.932
592737 99.087 -57.323 49.899
592738 99.103 -57.335 49.864
592739 99.120 -57.346 49.829
592740 99.138 -57.357 49.792
592741 99.165 -57.369 49.755
592742 99.189 -57.376 49.715
592743 99.220 -57.379 49.675
592744 99.253 -57.382 49.635
592745 99.286 -57.385 49.594
592746 99.319 -57.388 49.555
592747 99.352 -57.391 49.518
592748 99.385 -57.394 49.482
592749 99.418 -57.397 49.448
592750 99.451 -57.400 49.416
592751 99.484 -57.403 49.383
592752 99.511 -57.410 49.351
592753 99.537 -57.420 49.318
592754 99.562 -57.435 49.285
592755 99.586 -57.450 49.250
592756 99.614 -57.457 49.214
592757 99.641 -57.463 49.178
592758 99.669 -57.473 49.141
592759 99.697 -57.489 49.105
592760 99.725 -57.507 49.071
592761 99.753 -57.525 49.039
592762 99.780 -57.542 49.008
592763 99.808 -57.561 48.979
592764 99.836 -57.578 48.951
592765 99.863 -57.596 48.927
592766 99.892 -57.614 48.904
592767 99.919 -57.632 48.882
592768 99.947 -57.650 48.862
592769 99.975 -57.667 48.841
592770 100.002 -57.686 48.821
593112 88.564 -53.643 57.509
593113 88.599 -53.658 57.510
593114 88.636 -53.675 57.511
593115 88.674 -53.694 57.510
593116 88.712 -53.714 57.510
593117 88.750 -53.733 57.508
593118 88.788 -53.753 57.506
593119 88.826 -53.772 57.503
593120 88.864 -53.791 57.499
593121 88.902 -53.810 57.495
593122 88.940 -53.830 57.490
593123 88.978 -53.849 57.483
593124 89.020 -53.869 57.475
593125 89.071 -53.888 57.465
593126 89.122 -53.909 57.454
593127 89.159 -53.932 57.443
593128 89.217 -53.928 57.430
593129 89.254 -53.941 57.414
593130 89.283 -53.962 57.396
593131 89.319 -53.981 57.377
593132 89.355 -54.001 57.355
593133 89.391 -54.019 57.332
593134 89.427 -54.038 57.305
593135 89.463 -54.058 57.274
593136 89.499 -54.077 57.237
593137 89.535 -54.096 57.199
593138 89.571 -54.115 57.158
593139 89.607 -54.135 57.115
593140 89.644 -54.154 57.071
593141 89.679 -54.173 57.027
593142 89.716 -54.192 56.981
593143 89.751 -54.211 56.935
593144 89.788 -54.230 56.891
593145 89.823 -54.249 56.849
593151 90.039 -54.364 56.446
593152 90.075 -54.383 56.412
593153 90.108 -54.405 56.378
593154 90.141 -54.428 56.345
593155 90.173 -54.451 56.314
593156 90.205 -54.473 56.285
593157 90.241 -54.492 56.258
593158 90.279 -54.508 56.232
593159 90.317 -54.526 56.209
593160 90.356 -54.543 56.186
593161 90.394 -54.566 56.167
593162 90.427 -54.586 56.153
593163 90.451 -54.600 56.143
593164 90.477 -54.614 56.131
593165 90.496 -54.631 56.118
593166 90.514 -54.649 56.100
593167 90.532 -54.666 56.081
593168 90.550 -54.684 56.059
593169 90.569 -54.702 56.035
593170 90.587 -54.720 56.009
593171 90.605 -54.737 55.981
593172 90.623 -54.756 55.955
593173 90.642 -54.773 55.933
593174 90.660 -54.791 55.915
593175 90.678 -54.809 55.900
593176 90.696 -54.827 55.887
593177 90.715 -54.844 55.877
593178 90.739 -54.861 55.869
593179 90.776 -54.877 55.863
593180 90.813 -54.892 55.857
593181 90.850 -54.907 55.848
593182 90.886 -54.922 55.838
593183 90.923 -54.938 55.828
593184 90.960 -54.953 55.817
593185 90.997 -54.969 55.804
593186 91.033 -54.984 55.790
593187 91.070 -54.999 55.776
593188 91.107 -55.014 55.760
593189 91.144 -55.028 55.744
593190 91.181 -55.042 55.732
593191 91.218 -55.056 55.720
593192 91.255 -55.069 55.707
593193 91.292 -55.083 55.693
593194 91.329 -55.097 55.678
593195 91.366 -55.110 55.664
593196 91.409 -55.121 55.654
593197 91.462 -55.126 55.641
593198 91.515 -55.131 55.618
593199 91.551 -55.135 55.594
593200 91.581 -55.138 55.573
593201 91.611 -55.141 55.556
593202 91.641 -55.145 55.546
593203 91.672 -55.150 55.538
593204 91.708 -55.160 55.523
593205 91.743 -55.172 55.499
593206 91.779 -55.182 55.469
593207 91.814 -55.193 55.440
593208 91.850 -55.204 55.418
593209 91.885 -55.215 55.397
593210 91.921 -55.226 55.377
593211 91.956 -55.237 55.359
593212 91.990 -55.254 55.342
593213 92.023 -55.270 55.327
593214 92.057 -55.287 55.319
593215 92.091 -55.304 55.315
593216 92.125 -55.320 55.315
593217 92.158 -55.337 55.316
593218 92.192 -55.354 55.318
593219 92.226 -55.371 55.326
593220 92.261 -55.388 55.341
593221 92.296 -55.405 55.356
593222 92.334 -55.420 55.370
593223 92.374 -55.436 55.383
593224 92.415 -55.451 55.393
593225 92.455 -55.466 55.397
593226 92.492 -55.480 55.397
593227 92.526 -55.492 55.388
593228 92.559 -55.507 55.373
593229 92.589 -55.525 55.352
593230 92.620 -55.543 55.325
593231 92.650 -55.561 55.300
593232 92.680 -55.580 55.288
593233 92.710 -55.598 55.286
593234 92.740 -55.615 55.290
593235 92.771 -55.633 55.297
593236 92.801 -55.652 55.302
593237 92.831 -55.670 55.301
593238 92.861 -55.688 55.294
593239 92.892 -55.706 55.285
593240 92.922 -55.724 55.272
593241 92.958 -55.745 55.259
593242 92.995 -55.767 55.246
593243 93.032 -55.788 55.242
593244 93.069 -55.810 55.243
593245 93.107 -55.831 55.253
593246 93.144 -55.853 55.267
593247 93.181 -55.875 55.280
593248 93.218 -55.884 55.290
593249 93.254 -55.883 55.298
593250 93.289 -55.881 55.305
593251 93.324 -55.883 55.310
593252 93.358 -55.886 55.314
593253 93.392 -55.888 55.316
593254 93.426 -55.891 55.318
593255 93.461 -55.893 55.314
593256 93.495 -55.897 55.306
593257 93.529 -55.899 55.298
593258 93.562 -55.909 55.289
593259 93.593 -55.931 55.276
593260 93.624 -55.953 55.260
593261 93.655 -55.976 55.244
593262 93.686 -55.998 55.223
593263 93.717 -56.020 55.201
593264 93.748 -56.043 55.180
593265 93.779 -56.065 55.159
593266 93.805 -56.090 55.138
593267 93.826 -56.119 55.115
593268 93.845 -56.149 55.091
593269 93.869 -56.170 55.066
593270 93.899 -56.182 55.042
593271 93.928 -56.193 55.017
593272 93.957 -56.204 54.989
593273 93.987 -56.220 54.961
593274 94.018 -56.236 54.930
593275 94.048 -56.252 54.897
593276 94.079 -56.268 54.863
593277 94.109 -56.284 54.828
593278 94.140 -56.300 54.792
593279 94.170 -56.316 54.755
593280 94.201 -56.331 54.711
593281 94.231 -56.347 54.667
593282 94.262 -56.361 54.623
593283 94.296 -56.370 54.576
593284 94.331 -56.379 54.526
593285 94.364 -56.389 54.474
593286 94.392 -56.396 54.416
593287 94.420 -56.404 54.356
593288 94.448 -56.412 54.297
593289 94.475 -56.420 54.241
593290 94.503 -56.427 54.185
593291 94.534 -56.436 54.128
593292 94.569 -56.446 54.070
593293 94.605 -56.456 54.012
593294 94.640 -56.467 53.954
593295 94.676 -56.476 53.896
593296 94.711 -56.487 53.840
593297 94.747 -56.495 53.783
593298 94.787 -56.488 53.726
593299 94.826 -56.481 53.673
593300 94.865 -56.475 53.625
593301 94.905 -56.468 53.581
593302 94.942 -56.466 53.536
593303 94.970 -56.479 53.494
593304 94.998 -56.492 53.451
593305 95.027 -56.505 53.408
593306 95.055 -56.519 53.365
593307 95.084 -56.532 53.318
593308 95.112 -56.546 53.269
593309 95.137 -56.572 53.218
593310 95.164 -56.599 53.166
593311 95.190 -56.624 53.110
593312 95.216 -56.651 53.053
593313 95.237 -56.669 52.999
593314 95.275 -56.684 52.950
593315 95.315 -56.699 52.905
593316 95.362 -56.711 52.860
593317 95.405 -56.718 52.820
593318 95.430 -56.714 52.781
593319 95.457 -56.720 52.744
593320 95.484 -56.729 52.709
593321 95.512 -56.739 52.671
593322 95.539 -56.749 52.633
593323 95.566 -56.758 52.594
593324 95.595 -56.767 52.558
593325 95.634 -56.776 52.526
593326 95.673 -56.784 52.496
593327 95.711 -56.793 52.462
593328 95.750 -56.801 52.431
593329 95.787 -56.808 52.400
593330 95.823 -56.815 52.369
593331 95.860 -56.822 52.338
593332 95.891 -56.829 52.304
593333 95.913 -56.836 52.269
593334 95.934 -56.843 52.235
593335 95.956 -56.849 52.198
593336 95.977 -56.856 52.160
593337 95.999 -56.863 52.123
593338 96.022 -56.871 52.086
593339 96.045 -56.881 52.051
593340 96.067 -56.891 52.019
593341 96.091 -56.901 51.988
593342 96.114 -56.912 51.959
593343 96.133 -56.920 51.932
593344 96.153 -56.928 51.910
593345 96.171 -56.936 51.887
593346 96.190 -56.943 51.868
593347 96.215 -56.954 51.849
593348 96.243 -56.965 51.831
593349 96.272 -56.975 51.813
593350 96.300 -56.978 51.794
593351 96.328 -56.981 51.776
593352 96.356 -56.985 51.764
593353 96.384 -56.988 51.752
593354 96.411 -56.990 51.742
593355 96.439 -56.993 51.729
593356 96.466 -56.994 51.713
593357 96.495 -56.993 51.692
593358 96.523 -56.991 51.672
593359 96.551 -56.990 51.650
593360 96.579 -56.989 51.627
593361 96.607 -56.987 51.602
593362 96.636 -56.987 51.576
593363 96.664 -56.986 51.548
593364 96.692 -56.987 51.525
593365 96.722 -56.990 51.504
593366 96.755 -57.000 51.484
593367 96.789 -57.009 51.461
593368 96.821 -57.018 51.437
593369 96.855 -57.028 51.414
593370 96.886 -57.043 51.393
593371 96.918 -57.057 51.372
593372 96.950 -57.071 51.351
593373 96.980 -57.079 51.329
593374 97.020 -57.080 51.309
593375 97.053 -57.072 51.292
593376 97.086 -57.064 51.280
593377 97.119 -57.056 51.273
593378 97.152 -57.048 51.268
593379 97.185 -57.040 51.265
593380 97.218 -57.032 51.265
593381 97.251 -57.024 51.262
593382 97.285 -57.015 51.252
593383 97.318 -57.008 51.239
593384 97.351 -56.999 51.224
593385 97.383 -56.990 51.207
593386 97.416 -56.981 51.188
593387 97.448 -56.972 51.171
593388 97.480 -56.964 51.153
593389 97.516 -56.962 51.132
593390 97.544 -56.961 51.112
593391 97.576 -56.969 51.091
593392 97.606 -56.976 51.066
593393 97.639 -56.990 51.041
593394 97.673 -57.005 51.016
593395 97.706 -57.019 50.990
593396 97.739 -57.034 50.963
593397 97.773 -57.048 50.937
593398 97.806 -57.063 50.913
593399 97.840 -57.078 50.888
593400 97.874 -57.087 50.861
593401 97.908 -57.096 50.837
593402 97.942 -57.104 50.813
593403 97.976 -57.113 50.790
593404 98.010 -57.121 50.768
593405 98.044 -57.129 50.743
593406 98.078 -57.138 50.713
593407 98.112 -57.147 50.684
593408 98.146 -57.155 50.658
593409 98.180 -57.163 50.634
593410 98.215 -57.169 50.612
593411 98.250 -57.171 50.590
593412 98.286 -57.173 50.567
593413 98.321 -57.175 50.541
593414 98.356 -57.178 50.515
593415 98.392 -57.179 50.490
593416 98.427 -57.181 50.463
593417 98.462 -57.184 50.433
593418 98.498 -57.186 50.401
593419 98.533 -57.188 50.370
593420 98.569 -57.191 50.340
593421 98.604 -57.193 50.311
593422 98.639 -57.195 50.283
593423 98.671 -57.194 50.260
593424 98.708 -57.206 50.238
593425 98.750 -57.225 50.217
593426 98.787 -57.226 50.196
593427 98.823 -57.227 50.176
593428 98.861 -57.228 50.154
593429 98.898 -57.228 50.129
593430 98.935 -57.229 50.103
593431 98.972 -57.231 50.079
593432 99.004 -57.235 50.055
593433 99.020 -57.245 50.031
593434 99.036 -57.257 50.007
593435 99.053 -57.268 49.982
593436 99.069 -57.279 49.952
593437 99.085 -57.290 49.921
593438 99.101 -57.301 49.888
593439 99.118 -57.312 49.854
593440 99.134 -57.322 49.818
593441 99.150 -57.333 49.781
593442 99.166 -57.344 49.744
593443 99.194 -57.348 49.704
593444 99.227 -57.351 49.664
593445 99.261 -57.354 49.624
593446 99.293 -57.357 49.584
593447 99.327 -57.360 49.545
593448 99.360 -57.362 49.510
593449 99.393 -57.365 49.475
593450 99.426 -57.368 49.441
593451 99.459 -57.371 49.408
593452 99.492 -57.374 49.376
593453 99.518 -57.382 49.344
593454 99.546 -57.389 49.311
593455 99.571 -57.401 49.278
593456 99.595 -57.416 49.244
593457 99.620 -57.431 49.208
593458 99.646 -57.442 49.172
593459 99.674 -57.451 49.135
593460 99.702 -57.466 49.098
593461 99.729 -57.484 49.065
593462 99.757 -57.502 49.032
593463 99.785 -57.520 49.002
593464 99.813 -57.538 48.972
593465 99.841 -57.555 48.944
593466 99.868 -57.573 48.919
593467 99.896 -57.591 48.895
593468 99.924 -57.609 48.873
593469 99.951 -57.627 48.852
593470 99.979 -57.645 48.831
593471 100.007 -57.663 48.811
593813 88.589 -53.608 57.504
593814 88.625 -53.625 57.504
593815 88.663 -53.644 57.503
593816 88.701 -53.664 57.503
593817 88.739 -53.683 57.502
593818 88.777 -53.703 57.501
593819 88.815 -53.722 57.498
593820 88.853 -53.741 57.495
593821 88.892 -53.760 57.491
593822 88.929 -53.780 57.487
593823 88.968 -53.799 57.482
593824 89.005 -53.819 57.475
593825 89.049 -53.838 57.466
593826 89.100 -53.858 57.456
593827 89.151 -53.878 57.446
593828 89.191 -53.899 57.434
593829 89.247 -53.897 57.421
593830 89.271 -53.919 57.406
593831 89.301 -53.941 57.388
593832 89.337 -53.960 57.370
593833 89.373 -53.979 57.348
593834 89.409 -53.998 57.325
593835 89.445 -54.017 57.300
593836 89.481 -54.036 57.271
593837 89.517 -54.055 57.235
593838 89.553 -54.075 57.196
593839 89.589 -54.094 57.156
593840 89.625 -54.113 57.114
593841 89.661 -54.132 57.071
593842 89.697 -54.151 57.027
593843 89.733 -54.170 56.980
593844 89.769 -54.189 56.935
593845 89.805 -54.209 56.892
593846 89.841 -54.228 56.849
593852 90.057 -54.343 56.449
593853 90.093 -54.362 56.414
593854 90.129 -54.381 56.381
593855 90.165 -54.400 56.348
593856 90.201 -54.420 56.316
593857 90.237 -54.439 56.288
593858 90.273 -54.458 56.261
593859 90.309 -54.477 56.236
593860 90.345 -54.496 56.213
593861 90.381 -54.515 56.191
593862 90.415 -54.536 56.173
593863 90.446 -54.553 56.160
593864 90.464 -54.571 56.147
593865 90.482 -54.589 56.134
593866 90.501 -54.607 56.118
593867 90.519 -54.624 56.097
593868 90.537 -54.642 56.076
593869 90.555 -54.660 56.053
593870 90.574 -54.678 56.027
593871 90.592 -54.696 56.001
593872 90.610 -54.713 55.976
593873 90.628 -54.731 55.952
593874 90.647 -54.749 55.934
593875 90.665 -54.767 55.918
593876 90.683 -54.784 55.904
593877 90.701 -54.802 55.892
593878 90.719 -54.820 55.881
593879 90.748 -54.837 55.872
593880 90.785 -54.852 55.865
593881 90.821 -54.867 55.859
593882 90.858 -54.882 55.849
593883 90.895 -54.897 55.836
593884 90.932 -54.913 55.823
593885 90.968 -54.928 55.810
593886 91.005 -54.944 55.793
593887 91.042 -54.959 55.777
593888 91.079 -54.973 55.761
593889 91.116 -54.987 55.745
593890 91.153 -55.000 55.728
593891 91.190 -55.014 55.715
593892 91.227 -55.028 55.705
593893 91.264 -55.041 55.695
593894 91.301 -55.055 55.686
593895 91.338 -55.069 55.675
593896 91.375 -55.082 55.663
593897 91.426 -55.088 55.651
593898 91.479 -55.093 55.637
593899 91.532 -55.098 55.614
593900 91.569 -55.102 55.587
593901 91.599 -55.106 55.563
593902 91.629 -55.109 55.544
593903 91.659 -55.113 55.531
593904 91.689 -55.116 55.520
593905 91.721 -55.121 55.506
593906 91.756 -55.132 55.486
593907 91.792 -55.143 55.461
593908 91.827 -55.154 55.435
593909 91.863 -55.164 55.414
593910 91.898 -55.176 55.394
593911 91.933 -55.186 55.375
593912 91.967 -55.203 55.358
593913 92.001 -55.220 55.341
593914 92.034 -55.236 55.328
593915 92.068 -55.253 55.323
593916 92.102 -55.270 55.322
593917 92.136 -55.287 55.326
593918 92.169 -55.304 55.334
593919 92.203 -55.320 55.338
593920 92.237 -55.337 55.337
593921 92.272 -55.354 55.346
593922 92.308 -55.370 55.359
593923 92.348 -55.386 55.373
593924 92.389 -55.401 55.386
593925 92.426 -55.414 55.395
593926 92.460 -55.427 55.400
593927 92.494 -55.441 55.401
593928 92.528 -55.454 55.396
593929 92.563 -55.467 55.382
593930 92.594 -55.483 55.361
593931 92.624 -55.502 55.337
593932 92.654 -55.520 55.315
593933 92.685 -55.538 55.299
593934 92.715 -55.556 55.291
593935 92.745 -55.574 55.292
593936 92.775 -55.592 55.299
593937 92.805 -55.610 55.307
593938 92.836 -55.628 55.313
593939 92.866 -55.646 55.313
593940 92.896 -55.665 55.308
593941 92.931 -55.685 55.298
593942 92.968 -55.706 55.285
593943 93.005 -55.727 55.271
593944 93.043 -55.749 55.266
593945 93.079 -55.771 55.263
593946 93.117 -55.793 55.263
593947 93.154 -55.814 55.266
593948 93.191 -55.836 55.272
593949 93.228 -55.856 55.279
593950 93.264 -55.853 55.286
593951 93.300 -55.851 55.290
593952 93.335 -55.850 55.298
593953 93.369 -55.853 55.303
593954 93.404 -55.856 55.304
593955 93.438 -55.858 55.307
593956 93.473 -55.860 55.303
593957 93.507 -55.863 55.299
593958 93.541 -55.865 55.295
593959 93.574 -55.876 55.288
593960 93.605 -55.898 55.278
593961 93.636 -55.920 55.264
593962 93.667 -55.943 55.248
593963 93.698 -55.965 55.227
593964 93.729 -55.987 55.205
593965 93.760 -56.009 55.184
593966 93.788 -56.033 55.162
593967 93.808 -56.063 55.140
593968 93.828 -56.092 55.115
593969 93.848 -56.122 55.087
593970 93.870 -56.147 55.059
593971 93.900 -56.158 55.031
593972 93.928 -56.169 55.003
593973 93.958 -56.181 54.974
593974 93.987 -56.193 54.944
593975 94.016 -56.207 54.910
593976 94.047 -56.222 54.877
593977 94.077 -56.238 54.843
593978 94.108 -56.254 54.808
593979 94.138 -56.270 54.773
593980 94.169 -56.285 54.737
593981 94.199 -56.301 54.696
593982 94.230 -56.317 54.653
593983 94.263 -56.329 54.610
593984 94.297 -56.338 54.566
593985 94.331 -56.347 54.519
593986 94.365 -56.356 54.469
593987 94.398 -56.365 54.414
593988 94.425 -56.373 54.358
593989 94.453 -56.380 54.302
593990 94.481 -56.388 54.247
593991 94.508 -56.395 54.192
593992 94.540 -56.404 54.136
593993 94.575 -56.414 54.078
593994 94.611 -56.425 54.020
593995 94.646 -56.435 53.963
593996 94.682 -56.445 53.906
593997 94.718 -56.451 53.846
593998 94.758 -56.445 53.788
593999 94.797 -56.439 53.726
594000 94.837 -56.432 53.669
594001 94.876 -56.425 53.617
594002 94.915 -56.419 53.570
594003 94.955 -56.412 53.524
594004 94.986 -56.420 53.480
594005 95.014 -56.434 53.434
594006 95.042 -56.448 53.390
594007 95.068 -56.474 53.345
594008 95.094 -56.501 53.298
594009 95.120 -56.526 53.249
594010 95.146 -56.553 53.199
594011 95.172 -56.580 53.147
594012 95.198 -56.605 53.093
594013 95.224 -56.631 53.037
594014 95.241 -56.643 52.983
594015 95.281 -56.658 52.932
594016 95.322 -56.674 52.884
594017 95.362 -56.689 52.838
594018 95.409 -56.700 52.795
594019 95.442 -56.702 52.754
594020 95.467 -56.697 52.715
594021 95.493 -56.697 52.678
594022 95.521 -56.707 52.640
594023 95.548 -56.716 52.602
594024 95.576 -56.726 52.565
594025 95.603 -56.735 52.527
594026 95.638 -56.744 52.490
594027 95.677 -56.752 52.455
594028 95.715 -56.761 52.422
594029 95.753 -56.768 52.391
594030 95.790 -56.776 52.359
594031 95.826 -56.783 52.327
594032 95.863 -56.790 52.296
594033 95.895 -56.797 52.261
594034 95.916 -56.804 52.227
594035 95.938 -56.811 52.192
594036 95.960 -56.818 52.156
594037 95.981 -56.824 52.120
594038 96.003 -56.832 52.084
594039 96.026 -56.842 52.047
594040 96.049 -56.852 52.011
594041 96.073 -56.862 51.981
594042 96.095 -56.871 51.953
594043 96.114 -56.880 51.923
594044 96.132 -56.887 51.895
594045 96.151 -56.896 51.872
594046 96.170 -56.903 51.851
594047 96.189 -56.912 51.832
594048 96.218 -56.923 51.813
594049 96.246 -56.934 51.794
594050 96.274 -56.945 51.775
594051 96.303 -56.949 51.757
594052 96.330 -56.952 51.740
594053 96.359 -56.956 51.727
594054 96.386 -56.958 51.716
594055 96.414 -56.960 51.705
594056 96.441 -56.962 51.691
594057 96.469 -56.963 51.675
594058 96.497 -56.962 51.654
594059 96.525 -56.961 51.634
594060 96.553 -56.959 51.612
594061 96.582 -56.958 51.590
594062 96.610 -56.956 51.567
594063 96.638 -56.956 51.544
594064 96.666 -56.954 51.520
594065 96.694 -56.953 51.499
594066 96.723 -56.953 51.481
594067 96.754 -56.959 51.460
594068 96.787 -56.969 51.438
594069 96.819 -56.984 51.412
594070 96.851 -56.997 51.390
594071 96.883 -57.012 51.369
594072 96.914 -57.026 51.347
594073 96.946 -57.040 51.324
594074 96.978 -57.044 51.301
594075 97.018 -57.045 51.279
594076 97.051 -57.040 51.259
594077 97.084 -57.032 51.247
594078 97.117 -57.024 51.242
594079 97.151 -57.016 51.238
594080 97.184 -57.008 51.232
594081 97.217 -57.000 51.227
594082 97.250 -56.992 51.223
594083 97.283 -56.984 51.215
594084 97.316 -56.975 51.203
594085 97.350 -56.968 51.187
594086 97.382 -56.959 51.169
594087 97.416 -56.951 51.150
594088 97.449 -56.943 51.130
594089 97.482 -56.935 51.111
594090 97.515 -56.926 51.092
594091 97.547 -56.922 51.073
594092 97.579 -56.933 51.053
594093 97.612 -56.947 51.030
594094 97.646 -56.962 51.007
594095 97.679 -56.977 50.982
594096 97.713 -56.991 50.957
594097 97.746 -57.006 50.930
594098 97.779 -57.020 50.905
594099 97.813 -57.035 50.881
594100 97.847 -57.047 50.858
594101 97.880 -57.056 50.833
594102 97.915 -57.064 50.811
594103 97.948 -57.072 50.791
594104 97.983 -57.081 50.769
594105 98.017 -57.089 50.748
594106 98.051 -57.097 50.721
594107 98.085 -57.106 50.695
594108 98.119 -57.115 50.666
594109 98.153 -57.123 50.639
594110 98.187 -57.131 50.614
594111 98.221 -57.140 50.592
594112 98.255 -57.144 50.570
594113 98.291 -57.147 50.546
594114 98.326 -57.149 50.520
594115 98.362 -57.151 50.495
594116 98.397 -57.153 50.471
594117 98.433 -57.156 50.445
594118 98.468 -57.158 50.416
594119 98.503 -57.160 50.386
594120 98.539 -57.162 50.355
594121 98.574 -57.164 50.325
594122 98.609 -57.166 50.296
594123 98.645 -57.169 50.268
594124 98.673 -57.164 50.244
594125 98.710 -57.177 50.223
594126 98.756 -57.207 50.203
594127 98.793 -57.210 50.183
594128 98.830 -57.212 50.162
594129 98.867 -57.213 50.140
594130 98.904 -57.213 50.117
594131 98.941 -57.214 50.091
594132 98.977 -57.213 50.066
594133 99.014 -57.209 50.042
594134 99.033 -57.219 50.019
594135 99.050 -57.229 49.995
594136 99.066 -57.240 49.971
594137 99.082 -57.250 49.940
594138 99.098 -57.261 49.910
594139 99.115 -57.272 49.877
594140 99.131 -57.282 49.843
594141 99.147 -57.293 49.807
594142 99.163 -57.304 49.770
594143 99.183 -57.313 49.732
594144 99.210 -57.317 49.693
594145 99.238 -57.321 49.653
594146 99.268 -57.325 49.613
594147 99.301 -57.328 49.575
594148 99.334 -57.331 49.537
594149 99.367 -57.333 49.502
594150 99.401 -57.336 49.467
594151 99.433 -57.339 49.434
594152 99.467 -57.342 49.401
594153 99.499 -57.345 49.369
594154 99.526 -57.353 49.336
594155 99.553 -57.360 49.304
594156 99.580 -57.368 49.271
594157 99.605 -57.383 49.237
594158 99.629 -57.398 49.202
594159 99.653 -57.411 49.165
594160 99.686 -57.424 49.129
594161 99.711 -57.442 49.093
594162 99.737 -57.460 49.059
594163 99.763 -57.478 49.026
594164 99.789 -57.496 48.995
594165 99.817 -57.514 48.964
594166 99.844 -57.532 48.937
594167 99.872 -57.549 48.910
594168 99.900 -57.567 48.886
594169 99.928 -57.586 48.864
594170 99.956 -57.603 48.842
594171 99.983 -57.621 48.820
594172 100.011 -57.639 48.800
594514 88.615 -53.575 57.497
594515 88.653 -53.595 57.497
594516 88.691 -53.614 57.496
594517 88.729 -53.633 57.496
594518 88.767 -53.653 57.495
594519 88.805 -53.672 57.493
594520 88.843 -53.691 57.491
594521 88.881 -53.710 57.487
594522 88.919 -53.730 57.483
594523 88.957 -53.749 57.478
594524 88.995 -53.769 57.473
594525 89.033 -53.787 57.466
594526 89.077 -53.807 57.457
594527 89.128 -53.827 57.447
594528 89.179 -53.847 57.437
594529 89.225 -53.866 57.425
594530 89.265 -53.875 57.411
594531 89.289 -53.898 57.398
594532 89.319 -53.919 57.380
594533 89.355 -53.938 57.362
594534 89.391 -53.957 57.341
594535 89.427 -53.976 57.318
594536 89.463 -53.995 57.294
594537 89.499 -54.015 57.265
594538 89.535 -54.034 57.231
594539 89.571 -54.053 57.192
594540 89.607 -54.073 57.153
594541 89.643 -54.091 57.111
594542 89.679 -54.110 57.069
594543 89.715 -54.129 57.025
594544 89.751 -54.149 56.979
594545 89.787 -54.168 56.934
594546 89.823 -54.187 56.891
594547 89.859 -54.206 56.849
594553 90.075 -54.321 56.451
594554 90.111 -54.340 56.416
594555 90.147 -54.360 56.384
594556 90.183 -54.379 56.350
594557 90.219 -54.398 56.319
594558 90.255 -54.417 56.290
594559 90.291 -54.436 56.263
594560 90.325 -54.454 56.239
594561 90.360 -54.471 56.217
594562 90.394 -54.489 56.197
594563 90.430 -54.506 56.179
594564 90.452 -54.524 56.164
594565 90.469 -54.544 56.149
594566 90.487 -54.565 56.133
594567 90.505 -54.582 56.115
594568 90.523 -54.600 56.092
594569 90.541 -54.618 56.070
594570 90.559 -54.636 56.045
594571 90.578 -54.653 56.019
594572 90.596 -54.671 55.994
594573 90.614 -54.689 55.972
594574 90.633 -54.707 55.951
594575 90.651 -54.724 55.935
594576 90.669 -54.742 55.921
594577 90.687 -54.760 55.908
594578 90.706 -54.778 55.896
594579 90.724 -54.796 55.884
594580 90.756 -54.812 55.873
594581 90.793 -54.827 55.864
594582 90.830 -54.842 55.856
594583 90.867 -54.857 55.845
594584 90.903 -54.873 55.830
594585 90.940 -54.888 55.815
594586 90.977 -54.903 55.800
594587 91.014 -54.918 55.782
594588 91.051 -54.931 55.764
594589 91.088 -54.945 55.746
594590 91.125 -54.959 55.729
594591 91.162 -54.972 55.714
594592 91.199 -54.986 55.701
594593 91.236 -55.000 55.693
594594 91.273 -55.013 55.688
594595 91.310 -55.027 55.682
594596 91.347 -55.041 55.675
594597 91.389 -55.051 55.663
594598 91.443 -55.056 55.649
594599 91.496 -55.061 55.632
594600 91.549 -55.066 55.609
594601 91.586 -55.069 55.581
594602 91.616 -55.073 55.554
594603 91.647 -55.076 55.534
594604 91.677 -55.080 55.519
594605 91.707 -55.084 55.506
594606 91.737 -55.087 55.492
594607 91.769 -55.093 55.476
594608 91.804 -55.104 55.455
594609 91.840 -55.114 55.434
594610 91.875 -55.125 55.412
594611 91.911 -55.136 55.392
594612 91.944 -55.152 55.374
594613 91.978 -55.169 55.357
594614 92.012 -55.186 55.342
594615 92.046 -55.203 55.330
594616 92.079 -55.220 55.327
594617 92.113 -55.236 55.329
594618 92.147 -55.253 55.338
594619 92.180 -55.270 55.350
594620 92.214 -55.286 55.357
594621 92.248 -55.303 55.355
594622 92.283 -55.320 55.358
594623 92.322 -55.335 55.365
594624 92.360 -55.349 55.376
594625 92.394 -55.362 55.386
594626 92.428 -55.376 55.393
594627 92.462 -55.389 55.399
594628 92.497 -55.401 55.400
594629 92.530 -55.415 55.394
594630 92.565 -55.428 55.381
594631 92.598 -55.442 55.364
594632 92.629 -55.460 55.347
594633 92.659 -55.478 55.328
594634 92.689 -55.496 55.310
594635 92.719 -55.514 55.300
594636 92.749 -55.532 55.299
594637 92.780 -55.550 55.306
594638 92.810 -55.568 55.314
594639 92.840 -55.586 55.325
594640 92.870 -55.605 55.331
594641 92.904 -55.624 55.329
594642 92.941 -55.646 55.321
594643 92.978 -55.667 55.310
594644 93.015 -55.689 55.297
594645 93.053 -55.710 55.292
594646 93.089 -55.732 55.288
594647 93.127 -55.753 55.280
594648 93.164 -55.775 55.272
594649 93.201 -55.796 55.268
594650 93.238 -55.818 55.268
594651 93.275 -55.824 55.271
594652 93.310 -55.821 55.274
594653 93.346 -55.820 55.280
594654 93.381 -55.819 55.286
594655 93.416 -55.822 55.289
594656 93.450 -55.824 55.292
594657 93.484 -55.828 55.294
594658 93.518 -55.830 55.291
594659 93.553 -55.833 55.286
594660 93.586 -55.843 55.278
594661 93.617 -55.865 55.268
594662 93.648 -55.887 55.258
594663 93.679 -55.909 55.243
594664 93.710 -55.931 55.222
594665 93.741 -55.953 55.198
594666 93.771 -55.976 55.177
594667 93.791 -56.006 55.159
594668 93.811 -56.035 55.137
594669 93.831 -56.065 55.112
594670 93.851 -56.094 55.082
594671 93.871 -56.123 55.051
594672 93.900 -56.135 55.019
594673 93.929 -56.146 54.988
594674 93.958 -56.157 54.958
594675 93.987 -56.169 54.928
594676 94.016 -56.181 54.892
594677 94.046 -56.193 54.857
594678 94.076 -56.209 54.824
594679 94.107 -56.225 54.790
594680 94.137 -56.241 54.755
594681 94.168 -56.256 54.720
594682 94.198 -56.272 54.679
594683 94.229 -56.287 54.637
594684 94.263 -56.296 54.595
594685 94.297 -56.305 54.552
594686 94.331 -56.314 54.507
594687 94.365 -56.323 54.458
594688 94.400 -56.332 54.405
594689 94.431 -56.341 54.351
594690 94.459 -56.348 54.297
594691 94.486 -56.356 54.244
594692 94.514 -56.364 54.191
594693 94.546 -56.373 54.136
594694 94.582 -56.383 54.081
594695 94.617 -56.394 54.025
594696 94.653 -56.404 53.968
594697 94.690 -56.408 53.911
594698 94.729 -56.401 53.852
594699 94.768 -56.395 53.792
594700 94.808 -56.389 53.730
594701 94.847 -56.382 53.669
594702 94.887 -56.376 53.613
594703 94.926 -56.369 53.561
594704 94.965 -56.363 53.514
594705 94.999 -56.376 53.468
594706 95.025 -56.403 53.422
594707 95.051 -56.429 53.377
594708 95.077 -56.455 53.330
594709 95.103 -56.482 53.283
594710 95.129 -56.508 53.234
594711 95.155 -56.534 53.183
594712 95.181 -56.561 53.131
594713 95.207 -56.586 53.078
594714 95.229 -56.605 53.024
594715 95.248 -56.618 52.971
594716 95.288 -56.633 52.918
594717 95.329 -56.648 52.867
594718 95.369 -56.663 52.819
594719 95.409 -56.678 52.773
594720 95.454 -56.689 52.729
594721 95.479 -56.684 52.688
594722 95.505 -56.680 52.648
594723 95.530 -56.675 52.610
594724 95.557 -56.684 52.572
594725 95.584 -56.693 52.535
594726 95.612 -56.703 52.497
594727 95.641 -56.712 52.458
594728 95.680 -56.721 52.420
594729 95.719 -56.729 52.385
594730 95.756 -56.736 52.352
594731 95.793 -56.744 52.320
594732 95.829 -56.751 52.287
594733 95.866 -56.758 52.254
594734 95.898 -56.765 52.219
594735 95.920 -56.772 52.184
594736 95.942 -56.779 52.150
594737 95.963 -56.786 52.114
594738 95.985 -56.793 52.078
594739 96.008 -56.802 52.043
594740 96.031 -56.812 52.008
594741 96.054 -56.822 51.974
594742 96.075 -56.831 51.945
594743 96.094 -56.839 51.916
594744 96.112 -56.847 51.887
594745 96.131 -56.855 51.860
594746 96.150 -56.863 51.834
594747 96.169 -56.871 51.813
594748 96.192 -56.881 51.795
594749 96.220 -56.892 51.776
594750 96.249 -56.903 51.756
594751 96.277 -56.915 51.737
594752 96.305 -56.918 51.722
594753 96.333 -56.922 51.706
594754 96.361 -56.925 51.692
594755 96.388 -56.927 51.678
594756 96.416 -56.929 51.665
594757 96.444 -56.931 51.651
594758 96.471 -56.932 51.636
594759 96.500 -56.931 51.615
594760 96.528 -56.930 51.594
594761 96.556 -56.928 51.572
594762 96.584 -56.927 51.551
594763 96.612 -56.926 51.530
594764 96.641 -56.924 51.510
594765 96.669 -56.923 51.490
594766 96.697 -56.921 51.474
594767 96.724 -56.921 51.456
594768 96.751 -56.924 51.435
594769 96.783 -56.938 51.412
594770 96.815 -56.953 51.387
594771 96.847 -56.967 51.363
594772 96.879 -56.981 51.342
594773 96.910 -56.995 51.319
594774 96.942 -57.009 51.295
594775 96.976 -57.009 51.272
594776 97.016 -57.010 51.247
594777 97.050 -57.008 51.225
594778 97.083 -57.000 51.215
594779 97.116 -56.992 51.212
594780 97.149 -56.984 51.207
594781 97.182 -56.975 51.198
594782 97.216 -56.968 51.189
594783 97.248 -56.959 51.182
594784 97.282 -56.951 51.175
594785 97.315 -56.943 51.163
594786 97.348 -56.935 51.147
594787 97.381 -56.927 51.129
594788 97.414 -56.919 51.109
594789 97.447 -56.911 51.088
594790 97.480 -56.903 51.069
594791 97.514 -56.895 51.051
594792 97.549 -56.893 51.033
594793 97.585 -56.905 51.013
594794 97.618 -56.919 50.992
594795 97.652 -56.934 50.973
594796 97.685 -56.949 50.950
594797 97.719 -56.963 50.924
594798 97.752 -56.978 50.899
594799 97.786 -56.992 50.873
594800 97.819 -57.007 50.850
594801 97.853 -57.015 50.828
594802 97.887 -57.024 50.805
594803 97.921 -57.032 50.785
594804 97.955 -57.040 50.766
594805 97.989 -57.049 50.746
594806 98.023 -57.057 50.724
594807 98.057 -57.066 50.699
594808 98.091 -57.075 50.674
594809 98.126 -57.083 50.647
594810 98.159 -57.091 50.620
594811 98.194 -57.100 50.594
594812 98.227 -57.108 50.572
594813 98.262 -57.117 50.549
594814 98.297 -57.120 50.525
594815 98.332 -57.122 50.499
594816 98.367 -57.125 50.474
594817 98.403 -57.127 50.450
594818 98.438 -57.129 50.424
594819 98.474 -57.131 50.397
594820 98.509 -57.134 50.367
594821 98.545 -57.136 50.338
594822 98.580 -57.138 50.308
594823 98.615 -57.141 50.279
594824 98.647 -57.139 50.252
594825 98.676 -57.134 50.227
594826 98.713 -57.147 50.207
594827 98.758 -57.178 50.187
594828 98.799 -57.195 50.166
594829 98.837 -57.196 50.146
594830 98.873 -57.197 50.125
594831 98.909 -57.194 50.102
594832 98.944 -57.190 50.077
594833 98.978 -57.185 50.053
594834 99.017 -57.184 50.030
594835 99.047 -57.189 50.007
594836 99.063 -57.200 49.983
594837 99.079 -57.211 49.957
594838 99.095 -57.222 49.927
594839 99.112 -57.232 49.896
594840 99.128 -57.243 49.864
594841 99.144 -57.254 49.829
594842 99.160 -57.264 49.794
594843 99.177 -57.275 49.758
594844 99.199 -57.282 49.719
594845 99.227 -57.286 49.680
594846 99.255 -57.290 49.642
594847 99.282 -57.294 49.604
594848 99.310 -57.299 49.567
594849 99.342 -57.302 49.531
594850 99.375 -57.305 49.495
594851 99.408 -57.307 49.461
594852 99.441 -57.310 49.428
594853 99.474 -57.313 49.395
594854 99.507 -57.316 49.363
594855 99.534 -57.324 49.331
594856 99.561 -57.332 49.299
594857 99.588 -57.338 49.265
594858 99.614 -57.350 49.231
594859 99.638 -57.364 49.195
594860 99.660 -57.373 49.159
594861 99.697 -57.401 49.122
594862 99.722 -57.418 49.087
594863 99.748 -57.436 49.052
594864 99.773 -57.454 49.020
594865 99.799 -57.472 48.988
594866 99.825 -57.490 48.957
594867 99.850 -57.507 48.928
594868 99.876 -57.526 48.903
594869 99.904 -57.543 48.878
594870 99.931 -57.561 48.855
594871 99.959 -57.579 48.832
594872 99.987 -57.597 48.811
594873 100.015 -57.615 48.789
595216 88.680 -53.564 57.491
595217 88.718 -53.584 57.490
595218 88.756 -53.603 57.489
595219 88.794 -53.621 57.487
595220 88.832 -53.641 57.485
595221 88.871 -53.660 57.482
595222 88.909 -53.680 57.478
595223 88.947 -53.699 57.474
595224 88.985 -53.719 57.469
595225 89.023 -53.737 57.463
595226 89.061 -53.757 57.456
595227 89.106 -53.777 57.448
595228 89.157 -53.797 57.438
595229 89.208 -53.816 57.427
595230 89.258 -53.831 57.415
595231 89.282 -53.853 57.403
595232 89.307 -53.876 57.390
595233 89.337 -53.897 57.373
595234 89.373 -53.916 57.355
595235 89.409 -53.935 57.333
595236 89.445 -53.955 57.310
595237 89.481 -53.974 57.287
595238 89.517 -53.993 57.259
595239 89.553 -54.013 57.226
595240 89.589 -54.032 57.189
595241 89.625 -54.051 57.150
595242 89.661 -54.070 57.110
595243 89.697 -54.089 57.068
595244 89.733 -54.108 57.025
595245 89.769 -54.127 56.980
595246 89.805 -54.146 56.935
595247 89.841 -54.166 56.892
595248 89.877 -54.185 56.849
595249 89.913 -54.204 56.808
595254 90.093 -54.300 56.453
595255 90.129 -54.319 56.418
595256 90.163 -54.336 56.386
595257 90.197 -54.354 56.352
595258 90.232 -54.371 56.321
595259 90.266 -54.389 56.292
595260 90.300 -54.407 56.265
595261 90.335 -54.424 56.241
595262 90.369 -54.442 56.218
595263 90.403 -54.459 56.199
595264 90.443 -54.473 56.181
595265 90.460 -54.493 56.163
595266 90.477 -54.513 56.146
595267 90.495 -54.533 56.127
595268 90.512 -54.553 56.108
595269 90.529 -54.574 56.087
595270 90.546 -54.593 56.064
595271 90.565 -54.612 56.038
595272 90.583 -54.629 56.011
595273 90.601 -54.647 55.988
595274 90.619 -54.665 55.968
595275 90.637 -54.682 55.951
595276 90.656 -54.700 55.937
595277 90.674 -54.718 55.925
595278 90.692 -54.736 55.912
595279 90.710 -54.753 55.900
595280 90.728 -54.772 55.887
595281 90.765 -54.787 55.873
595282 90.802 -54.802 55.860
595283 90.838 -54.817 55.848
595284 90.875 -54.833 55.835
595285 90.912 -54.848 55.819
595286 90.949 -54.862 55.803
595287 90.986 -54.876 55.788
595288 91.023 -54.890 55.770
595289 91.060 -54.903 55.751
595290 91.097 -54.917 55.733
595291 91.134 -54.931 55.715
595292 91.171 -54.944 55.701
595293 91.208 -54.957 55.692
595294 91.245 -54.971 55.688
595295 91.282 -54.985 55.685
595296 91.319 -54.998 55.682
595297 91.356 -55.012 55.678
595298 91.406 -55.019 55.665
595299 91.460 -55.023 55.647
595300 91.513 -55.028 55.628
595301 91.567 -55.033 55.604
595302 91.604 -55.037 55.576
595303 91.634 -55.041 55.546
595304 91.664 -55.044 55.525
595305 91.695 -55.047 55.508
595306 91.725 -55.051 55.494
595307 91.755 -55.054 55.480
595308 91.785 -55.058 55.465
595309 91.817 -55.064 55.448
595310 91.852 -55.075 55.432
595311 91.888 -55.085 55.410
595312 91.922 -55.102 55.390
595313 91.955 -55.119 55.372
595314 91.989 -55.135 55.356
595315 92.023 -55.152 55.343
595316 92.057 -55.169 55.332
595317 92.090 -55.185 55.330
595318 92.124 -55.202 55.335
595319 92.158 -55.219 55.347
595320 92.192 -55.235 55.363
595321 92.225 -55.253 55.374
595322 92.259 -55.270 55.377
595323 92.294 -55.284 55.375
595324 92.328 -55.298 55.375
595325 92.362 -55.311 55.379
595326 92.397 -55.323 55.385
595327 92.431 -55.336 55.391
595328 92.465 -55.350 55.396
595329 92.499 -55.363 55.396
595330 92.533 -55.376 55.389
595331 92.567 -55.389 55.377
595332 92.601 -55.402 55.363
595333 92.633 -55.418 55.350
595334 92.663 -55.436 55.333
595335 92.694 -55.454 55.319
595336 92.724 -55.472 55.312
595337 92.754 -55.490 55.314
595338 92.784 -55.508 55.322
595339 92.814 -55.527 55.329
595340 92.845 -55.545 55.338
595341 92.876 -55.564 55.344
595342 92.914 -55.585 55.345
595343 92.951 -55.607 55.341
595344 92.988 -55.628 55.333
595345 93.025 -55.649 55.321
595346 93.063 -55.671 55.318
595347 93.099 -55.693 55.315
595348 93.137 -55.715 55.305
595349 93.174 -55.736 55.288
595350 93.211 -55.758 55.268
595351 93.249 -55.779 55.259
595352 93.285 -55.795 55.258
595353 93.321 -55.793 55.258
595354 93.357 -55.790 55.262
595355 93.393 -55.788 55.268
595356 93.428 -55.789 55.272
595357 93.462 -55.792 55.277
595358 93.496 -55.794 55.284
595359 93.531 -55.796 55.284
595360 93.565 -55.799 55.280
595361 93.598 -55.809 55.272
595362 93.629 -55.831 55.262
595363 93.660 -55.853 55.252
595364 93.691 -55.876 55.235
595365 93.722 -55.898 55.212
595366 93.753 -55.920 55.188
595367 93.774 -55.949 55.165
595368 93.794 -55.978 55.144
595369 93.814 -56.007 55.125
595370 93.834 -56.037 55.103
595371 93.854 -56.066 55.076
595372 93.874 -56.096 55.045
595373 93.901 -56.111 55.012
595374 93.930 -56.122 54.979
595375 93.959 -56.135 54.947
595376 93.988 -56.146 54.916
595377 94.017 -56.157 54.882
595378 94.046 -56.169 54.847
595379 94.075 -56.180 54.812
595380 94.105 -56.195 54.776
595381 94.136 -56.210 54.741
595382 94.166 -56.226 54.705
595383 94.197 -56.242 54.663
595384 94.229 -56.254 54.622
595385 94.264 -56.263 54.580
595386 94.297 -56.273 54.537
595387 94.332 -56.282 54.492
595388 94.366 -56.291 54.445
595389 94.400 -56.300 54.393
595390 94.434 -56.309 54.340
595391 94.464 -56.317 54.289
595392 94.492 -56.325 54.237
595393 94.520 -56.332 54.186
595394 94.553 -56.342 54.134
595395 94.588 -56.352 54.080
595396 94.624 -56.362 54.024
595397 94.661 -56.365 53.968
595398 94.700 -56.359 53.910
595399 94.740 -56.352 53.851
595400 94.779 -56.345 53.790
595401 94.819 -56.339 53.727
595402 94.858 -56.332 53.666
595403 94.893 -56.338 53.609
595404 94.927 -56.345 53.557
595405 94.961 -56.353 53.509
595406 95.006 -56.358 53.463
595407 95.034 -56.384 53.417
595408 95.060 -56.410 53.370
595409 95.086 -56.436 53.324
595410 95.112 -56.463 53.278
595411 95.138 -56.489 53.229
595412 95.164 -56.515 53.178
595413 95.190 -56.541 53.125
595414 95.215 -56.567 53.072
595415 95.233 -56.580 53.019
595416 95.255 -56.592 52.966
595417 95.296 -56.608 52.909
595418 95.336 -56.623 52.856
595419 95.376 -56.637 52.805
595420 95.416 -56.652 52.756
595421 95.457 -56.668 52.709
595422 95.491 -56.672 52.664
595423 95.516 -56.668 52.621
595424 95.541 -56.663 52.581
595425 95.567 -56.658 52.543
595426 95.593 -56.661 52.505
595427 95.620 -56.671 52.467
595428 95.648 -56.680 52.429
595429 95.684 -56.689 52.389
595430 95.723 -56.697 52.351
595431 95.760 -56.705 52.316
595432 95.796 -56.712 52.282
595433 95.832 -56.719 52.248
595434 95.869 -56.726 52.213
595435 95.902 -56.733 52.178
595436 95.923 -56.740 52.142
595437 95.945 -56.746 52.109
595438 95.967 -56.753 52.075
595439 95.989 -56.763 52.040
595440 96.013 -56.773 52.005
595441 96.036 -56.783 51.970
595442 96.054 -56.791 51.937
595443 96.074 -56.799 51.908
595444 96.092 -56.807 51.880
595445 96.111 -56.815 51.854
595446 96.130 -56.823 51.828
595447 96.149 -56.830 51.803
595448 96.168 -56.839 51.778
595449 96.194 -56.849 51.757
595450 96.223 -56.861 51.738
595451 96.251 -56.873 51.718
595452 96.280 -56.884 51.700
595453 96.308 -56.889 51.685
595454 96.336 -56.892 51.669
595455 96.363 -56.894 51.655
595456 96.391 -56.896 51.642
595457 96.418 -56.898 51.628
595458 96.446 -56.899 51.611
595459 96.474 -56.901 51.593
595460 96.502 -56.900 51.573
595461 96.530 -56.899 51.551
595462 96.558 -56.897 51.530
595463 96.587 -56.896 51.510
595464 96.613 -56.896 51.489
595465 96.641 -56.895 51.471
595466 96.667 -56.894 51.453
595467 96.694 -56.893 51.438
595468 96.721 -56.893 51.421
595469 96.748 -56.893 51.402
595470 96.780 -56.907 51.380
595471 96.811 -56.921 51.355
595472 96.843 -56.936 51.330
595473 96.875 -56.950 51.308
595474 96.907 -56.964 51.284
595475 96.936 -56.975 51.260
595476 96.974 -56.973 51.236
595477 97.015 -56.975 51.211
595478 97.048 -56.975 51.190
595479 97.082 -56.968 51.182
595480 97.114 -56.959 51.180
595481 97.148 -56.952 51.176
595482 97.181 -56.943 51.165
595483 97.214 -56.935 51.151
595484 97.247 -56.928 51.139
595485 97.280 -56.919 51.131
595486 97.313 -56.911 51.118
595487 97.347 -56.903 51.102
595488 97.380 -56.895 51.085
595489 97.413 -56.887 51.066
595490 97.446 -56.879 51.043
595491 97.479 -56.871 51.026
595492 97.512 -56.862 51.009
595493 97.552 -56.868 50.990
595494 97.591 -56.877 50.971
595495 97.625 -56.891 50.953
595496 97.658 -56.906 50.936
595497 97.692 -56.921 50.917
595498 97.725 -56.935 50.892
595499 97.758 -56.950 50.867
595500 97.792 -56.964 50.842
595501 97.826 -56.975 50.820
595502 97.860 -56.984 50.799
595503 97.894 -56.992 50.779
595504 97.928 -57.000 50.759
595505 97.962 -57.009 50.740
595506 97.996 -57.018 50.722
595507 98.030 -57.026 50.701
595508 98.064 -57.034 50.678
595509 98.098 -57.043 50.653
595510 98.132 -57.051 50.627
595511 98.166 -57.059 50.600
595512 98.200 -57.068 50.576
595513 98.234 -57.077 50.553
595514 98.268 -57.085 50.529
595515 98.302 -57.094 50.505
595516 98.338 -57.097 50.480
595517 98.373 -57.099 50.454
595518 98.409 -57.100 50.429
595519 98.444 -57.103 50.402
595520 98.479 -57.105 50.377
595521 98.514 -57.107 50.348
595522 98.550 -57.109 50.320
595523 98.585 -57.112 50.291
595524 98.621 -57.114 50.262
595525 98.650 -57.110 50.235
595526 98.678 -57.106 50.211
595527 98.715 -57.119 50.190
595528 98.761 -57.149 50.169
595529 98.806 -57.180 50.149
595530 98.841 -57.177 50.129
595531 98.876 -57.172 50.108
595532 98.910 -57.168 50.086
595533 98.945 -57.163 50.062
595534 98.980 -57.159 50.039
595535 99.021 -57.158 50.016
595536 99.060 -57.160 49.993
595537 99.076 -57.171 49.968
595538 99.092 -57.181 49.941
595539 99.109 -57.192 49.912
595540 99.125 -57.203 49.880
595541 99.141 -57.213 49.848
595542 99.157 -57.224 49.815
595543 99.173 -57.235 49.779
595544 99.190 -57.246 49.744
595545 99.216 -57.250 49.706
595546 99.244 -57.255 49.668
595547 99.272 -57.259 49.631
595548 99.299 -57.263 49.594
595549 99.327 -57.267 49.558
595550 99.355 -57.272 49.523
595551 99.382 -57.276 49.488
595552 99.416 -57.279 49.454
595553 99.448 -57.282 49.422
595554 99.482 -57.285 49.389
595555 99.514 -57.288 49.357
595556 99.542 -57.295 49.325
595557 99.569 -57.303 49.292
595558 99.595 -57.310 49.258
595559 99.623 -57.317 49.224
595560 99.647 -57.332 49.188
595561 99.682 -57.359 49.152
595562 99.708 -57.376 49.115
595563 99.733 -57.395 49.080
595564 99.759 -57.412 49.045
595565 99.784 -57.430 49.013
595566 99.810 -57.448 48.981
595567 99.835 -57.466 48.950
595568 99.861 -57.484 48.920
595569 99.887 -57.501 48.894
595570 99.912 -57.520 48.869
595571 99.937 -57.537 48.845
595572 99.963 -57.555 48.822
595573 99.991 -57.573 48.800
595574 100.018 -57.591 48.779
595917 88.708 -53.534 57.484
595918 88.746 -53.552 57.483
595919 88.784 -53.572 57.482
595920 88.822 -53.591 57.479
595921 88.860 -53.610 57.477
595922 88.898 -53.630 57.473
595923 88.936 -53.649 57.470
595924 88.974 -53.668 57.465
595925 89.012 -53.687 57.460
595926 89.050 -53.707 57.454
595927 89.088 -53.726 57.448
595928 89.134 -53.746 57.439
595929 89.185 -53.766 57.429
595930 89.237 -53.786 57.418
595931 89.276 -53.809 57.406
595932 89.300 -53.832 57.394
595933 89.324 -53.855 57.381
595934 89.354 -53.875 57.363
595935 89.391 -53.895 57.345
595936 89.426 -53.914 57.325
595937 89.463 -53.933 57.303
595938 89.498 -53.953 57.279
595939 89.535 -53.972 57.254
595940 89.570 -53.991 57.221
595941 89.607 -54.010 57.185
595942 89.642 -54.029 57.148
595943 89.679 -54.048 57.108
595944 89.715 -54.067 57.067
595945 89.751 -54.086 57.025
595946 89.787 -54.106 56.980
595947 89.823 -54.125 56.936
595948 89.859 -54.144 56.892
595949 89.895 -54.164 56.850
595950 89.931 -54.182 56.809
595955 90.104 -54.271 56.455
595956 90.138 -54.289 56.420
595957 90.172 -54.307 56.386
595958 90.206 -54.324 56.354
595959 90.241 -54.342 56.322
595960 90.275 -54.359 56.293
595961 90.310 -54.377 56.266
595962 90.344 -54.394 56.241
595963 90.378 -54.412 56.219
595964 90.416 -54.427 56.199
595965 90.450 -54.441 56.180
595966 90.468 -54.461 56.160
595967 90.485 -54.481 56.141
595968 90.502 -54.502 56.122
595969 90.519 -54.521 56.103
595970 90.536 -54.542 56.081
595971 90.554 -54.562 56.057
595972 90.571 -54.582 56.030
595973 90.588 -54.602 56.005
595974 90.605 -54.622 55.984
595975 90.623 -54.640 55.965
595976 90.642 -54.658 55.952
595977 90.660 -54.676 55.940
595978 90.678 -54.693 55.928
595979 90.696 -54.712 55.915
595980 90.715 -54.729 55.903
595981 90.737 -54.746 55.887
595982 90.773 -54.762 55.870
595983 90.810 -54.777 55.854
595984 90.848 -54.793 55.838
595985 90.884 -54.807 55.822
595986 90.921 -54.821 55.806
595987 90.958 -54.834 55.790
595988 90.995 -54.847 55.775
595989 91.032 -54.861 55.756
595990 91.069 -54.875 55.737
595991 91.106 -54.888 55.720
595992 91.143 -54.902 55.704
595993 91.180 -54.916 55.692
595994 91.217 -54.929 55.685
595995 91.254 -54.943 55.684
595996 91.291 -54.956 55.685
595997 91.328 -54.970 55.683
595998 91.370 -54.981 55.679
595999 91.424 -54.986 55.666
596000 91.477 -54.991 55.645
596001 91.530 -54.996 55.623
596002 91.584 -55.000 55.599
596003 91.622 -55.005 55.570
596004 91.652 -55.008 55.539
596005 91.682 -55.012 55.516
596006 91.712 -55.015 55.499
596007 91.742 -55.019 55.484
596008 91.773 -55.022 55.469
596009 91.803 -55.025 55.455
596010 91.833 -55.029 55.441
596011 91.865 -55.035 55.428
596012 91.899 -55.051 55.408
596013 91.933 -55.068 55.388
596014 91.966 -55.085 55.370
596015 92.000 -55.102 55.355
596016 92.034 -55.119 55.343
596017 92.068 -55.137 55.333
596018 92.102 -55.155 55.331
596019 92.136 -55.173 55.337
596020 92.170 -55.190 55.351
596021 92.204 -55.207 55.368
596022 92.237 -55.225 55.382
596023 92.272 -55.242 55.387
596024 92.307 -55.248 55.388
596025 92.333 -55.259 55.386
596026 92.365 -55.272 55.384
596027 92.399 -55.285 55.385
596028 92.433 -55.298 55.392
596029 92.467 -55.311 55.392
596030 92.501 -55.324 55.389
596031 92.535 -55.337 55.381
596032 92.569 -55.350 55.370
596033 92.603 -55.364 55.359
596034 92.638 -55.376 55.349
596035 92.668 -55.394 55.335
596036 92.698 -55.412 55.328
596037 92.728 -55.430 55.324
596038 92.759 -55.448 55.328
596039 92.789 -55.467 55.340
596040 92.819 -55.485 55.346
596041 92.849 -55.503 55.352
596042 92.886 -55.524 55.357
596043 92.924 -55.546 55.359
596044 92.961 -55.568 55.357
596045 92.998 -55.589 55.351
596046 93.036 -55.611 55.343
596047 93.073 -55.632 55.341
596048 93.110 -55.654 55.337
596049 93.147 -55.675 55.325
596050 93.184 -55.697 55.308
596051 93.221 -55.718 55.291
596052 93.259 -55.740 55.277
596053 93.296 -55.762 55.268
596054 93.332 -55.763 55.260
596055 93.368 -55.761 55.254
596056 93.404 -55.759 55.253
596057 93.439 -55.757 55.256
596058 93.474 -55.759 55.264
596059 93.508 -55.761 55.273
596060 93.542 -55.764 55.277
596061 93.577 -55.766 55.275
596062 93.610 -55.776 55.269
596063 93.641 -55.798 55.260
596064 93.672 -55.820 55.245
596065 93.703 -55.843 55.226
596066 93.734 -55.865 55.203
596067 93.757 -55.892 55.178
596068 93.777 -55.922 55.153
596069 93.797 -55.950 55.129
596070 93.817 -55.980 55.109
596071 93.837 -56.009 55.089
596072 93.857 -56.039 55.064
596073 93.876 -56.069 55.034
596074 93.902 -56.088 55.002
596075 93.931 -56.100 54.967
596076 93.960 -56.111 54.933
596077 93.989 -56.122 54.901
596078 94.018 -56.134 54.869
596079 94.047 -56.145 54.837
596080 94.076 -56.157 54.804
596081 94.105 -56.169 54.771
596082 94.134 -56.181 54.739
596083 94.165 -56.197 54.704
596084 94.196 -56.212 54.662
596085 94.230 -56.221 54.618
596086 94.264 -56.230 54.574
596087 94.298 -56.239 54.527
596088 94.332 -56.248 54.479
596089 94.366 -56.257 54.433
596090 94.400 -56.266 54.383
596091 94.434 -56.276 54.330
596092 94.469 -56.285 54.279
596093 94.497 -56.293 54.230
596094 94.525 -56.301 54.180
596095 94.559 -56.310 54.128
596096 94.595 -56.320 54.074
596097 94.632 -56.322 54.019
596098 94.672 -56.316 53.962
596099 94.711 -56.309 53.904
596100 94.750 -56.302 53.845
596101 94.786 -56.305 53.784
596102 94.821 -56.313 53.722
596103 94.855 -56.320 53.662
596104 94.889 -56.328 53.606
596105 94.923 -56.335 53.555
596106 94.960 -56.342 53.507
596107 95.007 -56.348 53.461
596108 95.042 -56.364 53.415
596109 95.068 -56.391 53.368
596110 95.095 -56.417 53.322
596111 95.120 -56.443 53.276
596112 95.146 -56.470 53.227
596113 95.173 -56.496 53.175
596114 95.199 -56.522 53.123
596115 95.220 -56.541 53.070
596116 95.238 -56.554 53.017
596117 95.262 -56.567 52.963
596118 95.302 -56.582 52.905
596119 95.343 -56.597 52.848
596120 95.383 -56.612 52.794
596121 95.423 -56.627 52.743
596122 95.464 -56.642 52.693
596123 95.500 -56.652 52.646
596124 95.527 -56.652 52.600
596125 95.553 -56.650 52.557
596126 95.579 -56.646 52.517
596127 95.604 -56.641 52.477
596128 95.629 -56.639 52.438
596129 95.657 -56.648 52.400
596130 95.687 -56.658 52.360
596131 95.726 -56.666 52.320
596132 95.763 -56.673 52.282
596133 95.799 -56.680 52.246
596134 95.836 -56.687 52.210
596135 95.872 -56.694 52.174
596136 95.905 -56.701 52.138
596137 95.927 -56.708 52.103
596138 95.949 -56.714 52.071
596139 95.971 -56.724 52.038
596140 95.994 -56.733 52.004
596141 96.016 -56.743 51.970
596142 96.035 -56.751 51.933
596143 96.053 -56.758 51.900
596144 96.073 -56.767 51.871
596145 96.091 -56.774 51.845
596146 96.110 -56.783 51.822
596147 96.129 -56.790 51.798
596148 96.148 -56.799 51.774
596149 96.169 -56.807 51.744
596150 96.197 -56.818 51.720
596151 96.226 -56.830 51.699
596152 96.254 -56.842 51.680
596153 96.283 -56.853 51.663
596154 96.311 -56.859 51.646
596155 96.338 -56.862 51.629
596156 96.366 -56.863 51.617
596157 96.393 -56.865 51.604
596158 96.421 -56.867 51.590
596159 96.449 -56.868 51.572
596160 96.476 -56.871 51.553
596161 96.504 -56.869 51.532
596162 96.530 -56.869 51.510
596163 96.557 -56.868 51.489
596164 96.584 -56.868 51.469
596165 96.611 -56.867 51.450
596166 96.638 -56.867 51.432
596167 96.664 -56.866 51.416
596168 96.691 -56.865 51.402
596169 96.718 -56.865 51.385
596170 96.745 -56.865 51.366
596171 96.776 -56.877 51.345
596172 96.807 -56.890 51.322
596173 96.839 -56.905 51.299
596174 96.871 -56.919 51.277
596175 96.903 -56.934 51.253
596176 96.929 -56.937 51.228
596177 96.972 -56.938 51.202
596178 97.013 -56.940 51.176
596179 97.047 -56.940 51.153
596180 97.080 -56.936 51.142
596181 97.113 -56.928 51.137
596182 97.146 -56.919 51.132
596183 97.179 -56.912 51.123
596184 97.213 -56.903 51.108
596185 97.246 -56.895 51.094
596186 97.279 -56.887 51.083
596187 97.312 -56.879 51.069
596188 97.345 -56.871 51.054
596189 97.378 -56.863 51.038
596190 97.412 -56.855 51.019
596191 97.444 -56.846 51.000
596192 97.477 -56.838 50.983
596193 97.514 -56.837 50.965
596194 97.555 -56.843 50.947
596195 97.596 -56.850 50.929
596196 97.631 -56.863 50.913
596197 97.665 -56.877 50.897
596198 97.698 -56.893 50.880
596199 97.731 -56.907 50.858
596200 97.765 -56.921 50.835
596201 97.798 -56.935 50.813
596202 97.832 -56.943 50.791
596203 97.867 -56.952 50.774
596204 97.900 -56.960 50.756
596205 97.935 -56.968 50.737
596206 97.968 -56.978 50.719
596207 98.003 -56.986 50.702
596208 98.037 -56.994 50.685
596209 98.071 -57.003 50.660
596210 98.105 -57.011 50.632
596211 98.139 -57.019 50.606
596212 98.173 -57.028 50.582
596213 98.207 -57.037 50.559
596214 98.241 -57.045 50.536
596215 98.275 -57.053 50.515
596216 98.309 -57.062 50.490
596217 98.343 -57.070 50.463
596218 98.378 -57.072 50.436
596219 98.414 -57.075 50.409
596220 98.449 -57.077 50.381
596221 98.485 -57.079 50.355
596222 98.520 -57.081 50.327
596223 98.556 -57.083 50.299
596224 98.591 -57.085 50.271
596225 98.624 -57.085 50.244
596226 98.652 -57.081 50.219
596227 98.681 -57.076 50.195
596228 98.718 -57.089 50.171
596229 98.763 -57.120 50.149
596230 98.806 -57.146 50.129
596231 98.843 -57.150 50.109
596232 98.877 -57.146 50.089
596233 98.912 -57.141 50.069
596234 98.946 -57.137 50.045
596235 98.981 -57.132 50.022
596236 99.024 -57.132 49.999
596237 99.069 -57.133 49.976
596238 99.089 -57.142 49.951
596239 99.106 -57.153 49.923
596240 99.122 -57.163 49.895
596241 99.138 -57.174 49.863
596242 99.154 -57.185 49.832
596243 99.170 -57.195 49.799
596244 99.187 -57.206 49.764
596245 99.205 -57.215 49.729
596246 99.233 -57.219 49.694
596247 99.261 -57.223 49.657
596248 99.288 -57.228 49.621
596249 99.316 -57.232 49.586
596250 99.344 -57.236 49.549
596251 99.371 -57.241 49.515
596252 99.399 -57.245 49.481
596253 99.427 -57.249 49.449
596254 99.456 -57.253 49.416
596255 99.489 -57.256 49.383
596256 99.522 -57.259 49.351
596257 99.549 -57.266 49.319
596258 99.576 -57.274 49.284
596259 99.603 -57.281 49.250
596260 99.630 -57.288 49.216
596261 99.657 -57.310 49.180
596262 99.693 -57.335 49.143
596263 99.718 -57.353 49.107
596264 99.744 -57.370 49.072
596265 99.770 -57.388 49.038
596266 99.795 -57.406 49.005
596267 99.821 -57.424 48.973
596268 99.846 -57.442 48.941
596269 99.872 -57.460 48.911
596270 99.898 -57.478 48.885
596271 99.923 -57.495 48.858
596272 99.948 -57.514 48.834
596273 99.974 -57.531 48.811
596274 100.000 -57.549 48.789
596615 88.621 -53.445 57.479
596616 88.660 -53.464 57.478
596617 88.697 -53.483 57.478
596618 88.736 -53.502 57.477
596619 88.773 -53.522 57.476
596620 88.812 -53.541 57.474
596621 88.849 -53.561 57.472
596622 88.888 -53.580 57.469
596623 88.925 -53.599 57.465
596624 88.964 -53.618 57.461
596625 89.002 -53.637 57.457
596626 89.040 -53.657 57.451
596627 89.078 -53.676 57.445
596628 89.116 -53.696 57.440
596629 89.163 -53.715 57.430
596630 89.214 -53.735 57.420
596631 89.259 -53.757 57.409
596632 89.290 -53.784 57.398
596633 89.318 -53.810 57.385
596634 89.342 -53.833 57.373
596635 89.373 -53.854 57.354
596636 89.408 -53.873 57.336
596637 89.444 -53.893 57.317
596638 89.480 -53.912 57.295
596639 89.516 -53.931 57.272
596640 89.552 -53.950 57.247
596641 89.588 -53.969 57.216
596642 89.624 -53.988 57.182
596643 89.660 -54.007 57.145
596644 89.696 -54.026 57.107
596645 89.732 -54.046 57.066
596646 89.768 -54.065 57.025
596647 89.804 -54.084 56.980
596648 89.838 -54.101 56.937
596649 89.873 -54.119 56.893
596650 89.907 -54.136 56.851
596651 89.941 -54.154 56.810
596656 90.113 -54.242 56.457
596657 90.148 -54.259 56.422
596658 90.182 -54.276 56.388
596659 90.216 -54.295 56.355
596660 90.250 -54.312 56.323
596661 90.285 -54.330 56.293
596662 90.319 -54.347 56.266
596663 90.354 -54.364 56.240
596664 90.389 -54.381 56.219
596665 90.432 -54.392 56.197
596666 90.458 -54.410 56.176
596667 90.475 -54.430 56.155
596668 90.493 -54.450 56.136
596669 90.510 -54.470 56.117
596670 90.527 -54.490 56.098
596671 90.544 -54.511 56.075
596672 90.561 -54.530 56.049
596673 90.579 -54.551 56.023
596674 90.596 -54.571 55.999
596675 90.613 -54.591 55.979
596676 90.630 -54.612 55.965
596677 90.648 -54.631 55.953
596678 90.665 -54.652 55.941
596679 90.683 -54.669 55.929
596680 90.701 -54.687 55.915
596681 90.719 -54.705 55.902
596682 90.746 -54.721 55.884
596683 90.783 -54.737 55.864
596684 90.819 -54.751 55.844
596685 90.856 -54.765 55.825
596686 90.893 -54.778 55.807
596687 90.930 -54.792 55.792
596688 90.967 -54.806 55.776
596689 91.004 -54.819 55.761
596690 91.041 -54.833 55.742
596691 91.078 -54.847 55.723
596692 91.115 -54.860 55.706
596693 91.152 -54.874 55.693
596694 91.189 -54.887 55.683
596695 91.226 -54.901 55.680
596696 91.263 -54.915 55.680
596697 91.300 -54.928 55.682
596698 91.337 -54.942 55.681
596699 91.387 -54.949 55.677
596700 91.441 -54.953 55.664
596701 91.494 -54.959 55.643
596702 91.547 -54.963 55.618
596703 91.601 -54.968 55.591
596704 91.640 -54.972 55.562
596705 91.670 -54.975 55.532
596706 91.700 -54.979 55.508
596707 91.730 -54.983 55.490
596708 91.760 -54.986 55.475
596709 91.790 -54.990 55.461
596710 91.820 -54.993 55.448
596711 91.851 -54.997 55.435
596712 91.879 -55.004 55.422
596713 91.911 -55.024 55.405
596714 91.945 -55.041 55.386
596715 91.979 -55.059 55.368
596716 92.013 -55.076 55.353
596717 92.047 -55.094 55.341
596718 92.081 -55.112 55.333
596719 92.115 -55.129 55.331
596720 92.148 -55.147 55.337
596721 92.182 -55.164 55.351
596722 92.216 -55.182 55.368
596723 92.250 -55.199 55.384
596724 92.285 -55.213 55.391
596725 92.324 -55.216 55.394
596726 92.353 -55.225 55.392
596727 92.379 -55.236 55.388
596728 92.406 -55.247 55.386
596729 92.435 -55.259 55.388
596730 92.470 -55.272 55.384
596731 92.503 -55.285 55.379
596732 92.537 -55.298 55.371
596733 92.572 -55.311 55.363
596734 92.606 -55.324 55.353
596735 92.640 -55.338 55.344
596736 92.673 -55.352 55.339
596737 92.703 -55.370 55.336
596738 92.733 -55.389 55.334
596739 92.763 -55.407 55.338
596740 92.793 -55.425 55.356
596741 92.824 -55.443 55.365
596742 92.859 -55.464 55.369
596743 92.897 -55.486 55.371
596744 92.934 -55.507 55.371
596745 92.971 -55.529 55.370
596746 93.008 -55.550 55.366
596747 93.046 -55.571 55.363
596748 93.083 -55.593 55.361
596749 93.120 -55.614 55.356
596750 93.157 -55.636 55.343
596751 93.194 -55.658 55.325
596752 93.231 -55.680 55.308
596753 93.269 -55.701 55.292
596754 93.306 -55.723 55.279
596755 93.343 -55.734 55.269
596756 93.378 -55.732 55.262
596757 93.414 -55.730 55.256
596758 93.450 -55.727 55.252
596759 93.486 -55.725 55.254
596760 93.520 -55.728 55.260
596761 93.554 -55.730 55.264
596762 93.588 -55.733 55.264
596763 93.622 -55.743 55.259
596764 93.653 -55.765 55.251
596765 93.684 -55.787 55.238
596766 93.715 -55.809 55.220
596767 93.740 -55.835 55.198
596768 93.760 -55.864 55.173
596769 93.780 -55.893 55.146
596770 93.800 -55.923 55.119
596771 93.820 -55.953 55.091
596772 93.840 -55.982 55.070
596773 93.859 -56.012 55.047
596774 93.880 -56.041 55.019
596775 93.902 -56.064 54.988
596776 93.931 -56.076 54.954
596777 93.961 -56.088 54.918
596778 93.989 -56.099 54.886
596779 94.018 -56.110 54.855
596780 94.048 -56.122 54.825
596781 94.077 -56.134 54.796
596782 94.106 -56.145 54.766
596783 94.135 -56.157 54.735
596784 94.164 -56.168 54.701
596785 94.196 -56.179 54.661
596786 94.230 -56.188 54.617
596787 94.264 -56.197 54.571
596788 94.298 -56.207 54.523
596789 94.333 -56.216 54.475
596790 94.367 -56.225 54.428
596791 94.401 -56.234 54.377
596792 94.435 -56.243 54.325
596793 94.469 -56.252 54.273
596794 94.503 -56.261 54.222
596795 94.531 -56.269 54.172
596796 94.565 -56.279 54.120
596797 94.603 -56.279 54.065
596798 94.642 -56.273 54.008
596799 94.680 -56.273 53.949
596800 94.714 -56.280 53.889
596801 94.748 -56.288 53.830
596802 94.782 -56.295 53.772
596803 94.817 -56.303 53.714
596804 94.851 -56.310 53.658
596805 94.885 -56.318 53.605
596806 94.919 -56.326 53.555
596807 94.960 -56.332 53.509
596808 95.007 -56.337 53.464
596809 95.051 -56.345 53.418
596810 95.077 -56.372 53.370
596811 95.103 -56.398 53.323
596812 95.129 -56.424 53.277
596813 95.156 -56.451 53.228
596814 95.181 -56.477 53.174
596815 95.207 -56.502 53.121
596816 95.225 -56.515 53.069
596817 95.242 -56.528 53.015
596818 95.269 -56.542 52.961
596819 95.309 -56.557 52.903
596820 95.349 -56.572 52.844
596821 95.390 -56.586 52.787
596822 95.430 -56.602 52.733
596823 95.469 -56.614 52.681
596824 95.507 -56.623 52.631
596825 95.534 -56.624 52.583
596826 95.561 -56.624 52.538
596827 95.588 -56.625 52.495
596828 95.615 -56.625 52.453
596829 95.641 -56.624 52.412
596830 95.666 -56.620 52.373
596831 95.693 -56.625 52.333
596832 95.730 -56.634 52.292
596833 95.766 -56.641 52.253
596834 95.803 -56.648 52.215
596835 95.839 -56.655 52.176
596836 95.875 -56.662 52.138
596837 95.909 -56.669 52.101
596838 95.931 -56.676 52.065
596839 95.953 -56.684 52.033
596840 95.976 -56.694 52.001
596841 95.995 -56.702 51.969
596842 96.015 -56.710 51.936
596843 96.033 -56.718 51.900
596844 96.052 -56.726 51.866
596845 96.071 -56.734 51.836
596846 96.090 -56.743 51.812
596847 96.109 -56.750 51.790
596848 96.128 -56.758 51.765
596849 96.147 -56.766 51.740
596850 96.171 -56.777 51.711
596851 96.200 -56.788 51.685
596852 96.228 -56.799 51.661
596853 96.256 -56.811 51.640
596854 96.287 -56.822 51.623
596855 96.312 -56.830 51.606
596856 96.340 -56.832 51.589
596857 96.368 -56.834 51.575
596858 96.395 -56.837 51.563
596859 96.422 -56.839 51.550
596860 96.450 -56.840 51.535
596861 96.475 -56.840 51.517
596862 96.501 -56.841 51.492
596863 96.527 -56.841 51.470
596864 96.554 -56.840 51.448
596865 96.581 -56.840 51.428
596866 96.608 -56.839 51.412
596867 96.635 -56.839 51.394
596868 96.661 -56.838 51.378
596869 96.688 -56.837 51.364
596870 96.715 -56.837 51.348
596871 96.742 -56.837 51.328
596872 96.772 -56.846 51.307
596873 96.804 -56.860 51.288
596874 96.835 -56.874 51.267
596875 96.867 -56.888 51.246
596876 96.899 -56.903 51.223
596877 96.921 -56.901 51.197
596878 96.970 -56.903 51.168
596879 97.012 -56.904 51.141
596880 97.045 -56.905 51.118
596881 97.079 -56.903 51.102
596882 97.112 -56.895 51.092
596883 97.145 -56.887 51.085
596884 97.178 -56.879 51.076
596885 97.211 -56.871 51.063
596886 97.244 -56.863 51.048
596887 97.278 -56.855 51.035
596888 97.310 -56.846 51.023
596889 97.344 -56.839 51.009
596890 97.377 -56.830 50.995
596891 97.410 -56.822 50.977
596892 97.443 -56.815 50.959
596893 97.476 -56.806 50.942
596894 97.517 -56.812 50.924
596895 97.558 -56.819 50.906
596896 97.599 -56.826 50.889
596897 97.638 -56.835 50.872
596898 97.671 -56.849 50.856
596899 97.705 -56.864 50.839
596900 97.738 -56.879 50.820
596901 97.771 -56.893 50.800
596902 97.805 -56.903 50.781
596903 97.839 -56.912 50.762
596904 97.873 -56.920 50.747
596905 97.908 -56.928 50.730
596906 97.941 -56.937 50.714
596907 97.975 -56.946 50.698
596908 98.009 -56.954 50.682
596909 98.043 -56.962 50.667
596910 98.078 -56.971 50.643
596911 98.111 -56.979 50.614
596912 98.146 -56.988 50.588
596913 98.179 -56.997 50.564
596914 98.214 -57.005 50.542
596915 98.247 -57.013 50.521
596916 98.282 -57.022 50.500
596917 98.316 -57.030 50.475
596918 98.350 -57.038 50.448
596919 98.384 -57.046 50.420
596920 98.420 -57.048 50.392
596921 98.455 -57.050 50.364
596922 98.490 -57.053 50.336
596923 98.525 -57.055 50.307
596924 98.561 -57.057 50.280
596925 98.596 -57.059 50.253
596926 98.626 -57.056 50.227
596927 98.655 -57.051 50.203
596928 98.684 -57.047 50.178
596929 98.720 -57.060 50.154
596930 98.765 -57.090 50.131
596931 98.805 -57.111 50.109
596932 98.844 -57.124 50.088
596933 98.878 -57.119 50.068
596934 98.913 -57.115 50.047
596935 98.947 -57.110 50.026
596936 98.984 -57.106 50.003
596937 99.028 -57.106 49.980
596938 99.072 -57.107 49.957
596939 99.102 -57.113 49.931
596940 99.119 -57.123 49.902
596941 99.135 -57.134 49.874
596942 99.151 -57.144 49.844
596943 99.167 -57.156 49.813
596944 99.183 -57.166 49.781
596945 99.200 -57.177 49.747
596946 99.222 -57.184 49.713
596947 99.250 -57.188 49.680
596948 99.277 -57.192 49.645
596949 99.305 -57.197 49.609
596950 99.333 -57.201 49.574
596951 99.360 -57.205 49.539
596952 99.388 -57.210 49.504
596953 99.416 -57.213 49.471
596954 99.443 -57.218 49.439
596955 99.471 -57.222 49.407
596956 99.499 -57.226 49.375
596957 99.529 -57.230 49.344
596958 99.557 -57.238 49.311
596959 99.592 -57.256 49.276
596960 99.628 -57.273 49.241
596961 99.648 -57.267 49.206
596962 99.679 -57.293 49.170
596963 99.704 -57.311 49.134
596964 99.729 -57.329 49.097
596965 99.755 -57.347 49.062
596966 99.781 -57.364 49.027
596967 99.806 -57.382 48.994
596968 99.832 -57.401 48.962
596969 99.857 -57.418 48.931
596970 99.883 -57.436 48.901
596971 99.908 -57.454 48.874
596972 99.934 -57.472 48.848
596973 99.960 -57.489 48.823
596974 99.985 -57.507 48.799
596975 100.010 -57.525 48.776
597313 88.535 -53.356 57.474
597314 88.573 -53.375 57.474
597315 88.611 -53.395 57.474
597316 88.649 -53.414 57.473
597317 88.687 -53.433 57.472
597318 88.725 -53.452 57.472
597319 88.763 -53.472 57.470
597320 88.801 -53.491 57.468
597321 88.839 -53.511 57.467
597322 88.877 -53.530 57.464
597323 88.915 -53.549 57.461
597324 88.953 -53.568 57.457
597325 88.991 -53.587 57.452
597326 89.029 -53.607 57.448
597327 89.067 -53.626 57.442
597328 89.105 -53.646 57.436
597329 89.143 -53.665 57.430
597330 89.192 -53.684 57.421
597331 89.242 -53.705 57.411
597332 89.272 -53.732 57.401
597333 89.303 -53.760 57.390
597334 89.333 -53.787 57.377
597335 89.359 -53.812 57.363
597336 89.390 -53.833 57.347
597337 89.426 -53.852 57.329
597338 89.462 -53.871 57.309
597339 89.498 -53.890 57.288
597340 89.534 -53.910 57.266
597341 89.570 -53.928 57.241
597342 89.606 -53.947 57.212
597343 89.641 -53.966 57.179
597344 89.676 -53.984 57.144
597345 89.710 -54.001 57.106
597346 89.745 -54.019 57.066
597347 89.779 -54.036 57.025
597348 89.813 -54.054 56.981
597349 89.847 -54.072 56.937
597350 89.882 -54.089 56.894
597351 89.916 -54.107 56.852
597352 89.951 -54.124 56.811
597353 89.985 -54.142 56.773
597357 90.122 -54.212 56.458
597358 90.157 -54.229 56.423
597359 90.191 -54.247 56.390
597360 90.225 -54.264 56.356
597361 90.260 -54.282 56.324
597362 90.294 -54.300 56.293
597363 90.328 -54.317 56.265
597364 90.363 -54.335 56.239
597365 90.404 -54.347 56.216
597366 90.447 -54.358 56.193
597367 90.466 -54.378 56.171
597368 90.483 -54.399 56.151
597369 90.500 -54.418 56.133
597370 90.517 -54.439 56.116
597371 90.535 -54.458 56.095
597372 90.552 -54.479 56.070
597373 90.569 -54.499 56.043
597374 90.586 -54.519 56.017
597375 90.603 -54.540 55.995
597376 90.620 -54.559 55.976
597377 90.638 -54.580 55.965
597378 90.655 -54.600 55.954
597379 90.672 -54.620 55.941
597380 90.689 -54.640 55.928
597381 90.707 -54.660 55.913
597382 90.724 -54.681 55.899
597383 90.754 -54.696 55.880
597384 90.791 -54.709 55.858
597385 90.828 -54.723 55.835
597386 90.865 -54.737 55.813
597387 90.902 -54.750 55.793
597388 90.939 -54.764 55.777
597389 90.976 -54.778 55.762
597390 91.013 -54.791 55.746
597391 91.050 -54.805 55.729
597392 91.087 -54.818 55.710
597393 91.124 -54.832 55.694
597394 91.161 -54.846 55.682
597395 91.198 -54.859 55.677
597396 91.235 -54.873 55.674
597397 91.272 -54.887 55.674
597398 91.309 -54.900 55.675
597399 91.351 -54.911 55.675
597400 91.404 -54.916 55.672
597401 91.458 -54.921 55.660
597402 91.511 -54.926 55.640
597403 91.565 -54.931 55.615
597404 91.618 -54.936 55.585
597405 91.658 -54.940 55.554
597406 91.687 -54.944 55.525
597407 91.718 -54.947 55.501
597408 91.748 -54.950 55.483
597409 91.778 -54.953 55.467
597410 91.808 -54.957 55.453
597411 91.838 -54.960 55.439
597412 91.868 -54.966 55.427
597413 91.895 -54.977 55.414
597414 91.926 -54.994 55.398
597415 91.958 -55.014 55.381
597416 91.992 -55.033 55.365
597417 92.025 -55.050 55.351
597418 92.059 -55.069 55.339
597419 92.093 -55.086 55.331
597420 92.127 -55.104 55.329
597421 92.161 -55.121 55.335
597422 92.195 -55.138 55.347
597423 92.229 -55.156 55.363
597424 92.263 -55.173 55.380
597425 92.299 -55.182 55.390
597426 92.338 -55.185 55.394
597427 92.372 -55.191 55.392
597428 92.399 -55.202 55.388
597429 92.426 -55.213 55.385
597430 92.452 -55.224 55.381
597431 92.479 -55.235 55.374
597432 92.506 -55.246 55.367
597433 92.540 -55.259 55.360
597434 92.574 -55.273 55.353
597435 92.608 -55.285 55.345
597436 92.642 -55.298 55.343
597437 92.676 -55.312 55.343
597438 92.708 -55.329 55.340
597439 92.738 -55.347 55.341
597440 92.768 -55.365 55.350
597441 92.798 -55.383 55.366
597442 92.832 -55.403 55.374
597443 92.869 -55.425 55.377
597444 92.907 -55.446 55.378
597445 92.944 -55.468 55.379
597446 92.981 -55.489 55.379
597447 93.018 -55.511 55.378
597448 93.056 -55.533 55.380
597449 93.093 -55.554 55.378
597450 93.130 -55.576 55.371
597451 93.167 -55.597 55.357
597452 93.204 -55.619 55.339
597453 93.241 -55.640 55.319
597454 93.279 -55.662 55.303
597455 93.316 -55.683 55.290
597456 93.353 -55.705 55.281
597457 93.389 -55.702 55.274
597458 93.425 -55.700 55.265
597459 93.461 -55.698 55.255
597460 93.496 -55.696 55.249
597461 93.532 -55.695 55.249
597462 93.566 -55.697 55.249
597463 93.601 -55.700 55.249
597464 93.634 -55.709 55.245
597465 93.664 -55.731 55.236
597466 93.696 -55.753 55.222
597467 93.723 -55.778 55.206
597468 93.743 -55.807 55.185
597469 93.763 -55.837 55.160
597470 93.783 -55.866 55.133
597471 93.802 -55.896 55.104
597472 93.823 -55.925 55.076
597473 93.842 -55.955 55.051
597474 93.862 -55.984 55.025
597475 93.882 -56.013 54.997
597476 93.903 -56.041 54.967
597477 93.932 -56.053 54.937
597478 93.961 -56.064 54.906
597479 93.990 -56.075 54.876
597480 94.019 -56.087 54.844
597481 94.048 -56.099 54.814
597482 94.077 -56.110 54.786
597483 94.107 -56.122 54.758
597484 94.135 -56.133 54.730
597485 94.165 -56.144 54.698
597486 94.198 -56.152 54.659
597487 94.232 -56.160 54.616
597488 94.266 -56.169 54.571
597489 94.299 -56.177 54.523
597490 94.334 -56.185 54.475
597491 94.367 -56.194 54.426
597492 94.402 -56.202 54.373
597493 94.435 -56.211 54.320
597494 94.470 -56.219 54.269
597495 94.503 -56.229 54.217
597496 94.537 -56.238 54.165
597497 94.573 -56.240 54.112
597498 94.607 -56.248 54.055
597499 94.641 -56.255 53.996
597500 94.676 -56.262 53.936
597501 94.710 -56.269 53.876
597502 94.744 -56.277 53.819
597503 94.778 -56.285 53.764
597504 94.813 -56.292 53.708
597505 94.847 -56.300 53.654
597506 94.881 -56.307 53.604
597507 94.915 -56.315 53.558
597508 94.961 -56.321 53.514
597509 95.008 -56.326 53.470
597510 95.055 -56.331 53.424
597511 95.086 -56.353 53.376
597512 95.112 -56.379 53.328
597513 95.138 -56.405 53.279
597514 95.164 -56.432 53.227
597515 95.190 -56.458 53.173
597516 95.211 -56.477 53.119
597517 95.229 -56.489 53.065
597518 95.247 -56.502 53.013
597519 95.276 -56.517 52.959
597520 95.316 -56.531 52.902
597521 95.356 -56.546 52.843
597522 95.397 -56.561 52.784
597523 95.437 -56.575 52.727
597524 95.474 -56.584 52.672
597525 95.512 -56.593 52.620
597526 95.541 -56.595 52.569
597527 95.568 -56.596 52.522
597528 95.595 -56.597 52.477
597529 95.621 -56.598 52.432
597530 95.649 -56.598 52.388
597531 95.675 -56.599 52.346
597532 95.702 -56.599 52.307
597533 95.733 -56.602 52.265
597534 95.769 -56.609 52.227
597535 95.806 -56.616 52.187
597536 95.842 -56.623 52.146
597537 95.879 -56.630 52.106
597538 95.913 -56.637 52.067
597539 95.935 -56.644 52.030
597540 95.957 -56.654 51.997
597541 95.976 -56.662 51.965
597542 95.994 -56.670 51.934
597543 96.013 -56.678 51.902
597544 96.032 -56.686 51.866
597545 96.051 -56.694 51.832
597546 96.070 -56.702 51.803
597547 96.089 -56.710 51.778
597548 96.108 -56.718 51.755
597549 96.126 -56.726 51.731
597550 96.146 -56.734 51.704
597551 96.174 -56.746 51.675
597552 96.202 -56.757 51.648
597553 96.231 -56.768 51.624
597554 96.265 -56.779 51.602
597555 96.296 -56.790 51.584
597556 96.314 -56.802 51.565
597557 96.341 -56.805 51.547
597558 96.369 -56.807 51.534
597559 96.396 -56.809 51.521
597560 96.424 -56.811 51.507
597561 96.452 -56.813 51.489
597562 96.475 -56.809 51.470
597563 96.498 -56.812 51.446
597564 96.525 -56.812 51.426
597565 96.551 -56.812 51.406
597566 96.578 -56.812 51.389
597567 96.605 -56.811 51.373
597568 96.631 -56.810 51.357
597569 96.659 -56.810 51.341
597570 96.685 -56.809 51.325
597571 96.712 -56.809 51.307
597572 96.739 -56.808 51.288
597573 96.768 -56.815 51.269
597574 96.800 -56.829 51.251
597575 96.831 -56.843 51.230
597576 96.864 -56.857 51.209
597577 96.894 -56.870 51.186
597578 96.916 -56.865 51.161
597579 96.968 -56.868 51.133
597580 97.010 -56.869 51.107
597581 97.043 -56.869 51.087
597582 97.077 -56.870 51.068
597583 97.110 -56.863 51.049
597584 97.144 -56.855 51.034
597585 97.176 -56.847 51.023
597586 97.210 -56.839 51.013
597587 97.243 -56.830 51.001
597588 97.276 -56.823 50.988
597589 97.309 -56.815 50.978
597590 97.342 -56.806 50.968
597591 97.375 -56.799 50.956
597592 97.408 -56.790 50.940
597593 97.442 -56.782 50.921
597594 97.478 -56.781 50.903
597595 97.519 -56.788 50.885
597596 97.561 -56.795 50.868
597597 97.601 -56.802 50.850
597598 97.642 -56.809 50.831
597599 97.677 -56.821 50.815
597600 97.711 -56.836 50.798
597601 97.744 -56.851 50.781
597602 97.778 -56.863 50.764
597603 97.812 -56.871 50.749
597604 97.846 -56.880 50.735
597605 97.880 -56.888 50.720
597606 97.914 -56.897 50.705
597607 97.948 -56.906 50.691
597608 97.982 -56.914 50.677
597609 98.016 -56.922 50.662
597610 98.050 -56.931 50.648
597611 98.084 -56.939 50.625
597612 98.118 -56.948 50.597
597613 98.152 -56.956 50.571
597614 98.186 -56.965 50.548
597615 98.220 -56.973 50.526
597616 98.254 -56.981 50.506
597617 98.288 -56.990 50.485
597618 98.322 -56.998 50.461
597619 98.357 -57.007 50.433
597620 98.390 -57.015 50.404
597621 98.425 -57.022 50.376
597622 98.460 -57.024 50.346
597623 98.496 -57.026 50.316
597624 98.531 -57.028 50.288
597625 98.567 -57.031 50.262
597626 98.600 -57.031 50.236
597627 98.629 -57.027 50.211
597628 98.657 -57.022 50.186
597629 98.686 -57.017 50.160
597630 98.723 -57.031 50.137
597631 98.765 -57.055 50.112
597632 98.805 -57.075 50.089
597633 98.845 -57.096 50.068
597634 98.880 -57.092 50.046
597635 98.914 -57.087 50.025
597636 98.949 -57.083 50.005
597637 98.988 -57.081 49.983
597638 99.032 -57.081 49.960
597639 99.076 -57.081 49.936
597640 99.115 -57.084 49.909
597641 99.132 -57.094 49.881
597642 99.148 -57.105 49.853
597643 99.164 -57.116 49.824
597644 99.180 -57.126 49.793
597645 99.197 -57.137 49.761
597646 99.213 -57.147 49.729
597647 99.239 -57.153 49.697
597648 99.267 -57.157 49.665
597649 99.294 -57.161 49.631
597650 99.322 -57.166 49.596
597651 99.350 -57.169 49.562
597652 99.377 -57.174 49.527
597653 99.405 -57.178 49.494
597654 99.433 -57.182 49.461
597655 99.461 -57.187 49.430
597656 99.488 -57.191 49.399
597657 99.516 -57.195 49.367
597658 99.544 -57.206 49.335
597659 99.588 -57.237 49.301
597660 99.627 -57.260 49.267
597661 99.662 -57.278 49.231
597662 99.655 -57.247 49.195
597663 99.689 -57.269 49.159
597664 99.715 -57.287 49.123
597665 99.740 -57.305 49.086
597666 99.766 -57.322 49.051
597667 99.791 -57.341 49.017
597668 99.817 -57.358 48.983
597669 99.843 -57.376 48.952
597670 99.868 -57.395 48.921
597671 99.894 -57.412 48.890
597672 99.919 -57.430 48.863
597673 99.945 -57.448 48.837
597674 99.971 -57.466 48.812
597675 99.996 -57.483 48.787
597676 100.021 -57.501 48.764
598011 88.454 -53.275 57.467
598012 88.489 -53.290 57.467
598013 88.524 -53.306 57.468
598014 88.562 -53.325 57.469
598015 88.601 -53.345 57.469
598016 88.638 -53.364 57.469
598017 88.677 -53.383 57.468
598018 88.714 -53.402 57.467
598019 88.753 -53.422 57.466
598020 88.790 -53.441 57.464
598021 88.829 -53.461 57.462
598022 88.867 -53.480 57.459
598023 88.905 -53.499 57.456
598024 88.943 -53.518 57.452
598025 88.981 -53.537 57.448
598026 89.019 -53.557 57.444
598027 89.057 -53.576 57.439
598028 89.095 -53.596 57.433
598029 89.133 -53.615 57.427
598030 89.171 -53.634 57.419
598031 89.221 -53.654 57.411
598032 89.254 -53.680 57.402
598033 89.285 -53.708 57.392
598034 89.316 -53.735 57.381
598035 89.346 -53.762 57.368
598036 89.377 -53.790 57.355
598037 89.408 -53.811 57.339
598038 89.444 -53.830 57.320
598039 89.479 -53.849 57.301
598040 89.514 -53.866 57.281
598041 89.548 -53.884 57.259
598042 89.582 -53.901 57.236
598043 89.617 -53.919 57.208
598044 89.651 -53.936 57.176
598045 89.685 -53.954 57.142
598046 89.720 -53.972 57.105
598047 89.754 -53.989 57.066
598048 89.788 -54.007 57.026
598049 89.823 -54.024 56.983
598050 89.857 -54.041 56.938
598051 89.891 -54.059 56.894
598052 89.925 -54.077 56.852
598053 89.960 -54.095 56.812
598054 89.994 -54.112 56.773
598058 90.131 -54.182 56.458
598059 90.166 -54.200 56.423
598060 90.200 -54.217 56.390
598061 90.235 -54.235 56.356
598062 90.269 -54.252 56.323
598063 90.303 -54.270 56.293
598064 90.337 -54.287 56.264
598065 90.376 -54.302 56.236
598066 90.419 -54.313 56.213
598067 90.456 -54.327 56.189
598068 90.473 -54.347 56.166
598069 90.490 -54.367 56.147
598070 90.508 -54.387 56.131
598071 90.525 -54.407 56.112
598072 90.542 -54.427 56.089
598073 90.559 -54.448 56.063
598074 90.577 -54.467 56.035
598075 90.594 -54.488 56.010
598076 90.611 -54.508 55.989
598077 90.628 -54.528 55.973
598078 90.646 -54.549 55.963
598079 90.663 -54.568 55.952
598080 90.680 -54.589 55.939
598081 90.697 -54.609 55.924
598082 90.714 -54.629 55.910
598083 90.738 -54.644 55.894
598084 90.767 -54.665 55.876
598085 90.800 -54.681 55.852
598086 90.837 -54.695 55.826
598087 90.874 -54.709 55.801
598088 90.911 -54.722 55.779
598089 90.948 -54.736 55.763
598090 90.986 -54.750 55.748
598091 91.022 -54.763 55.732
598092 91.059 -54.777 55.715
598093 91.096 -54.790 55.698
598094 91.133 -54.804 55.684
598095 91.170 -54.818 55.675
598096 91.207 -54.831 55.673
598097 91.244 -54.845 55.674
598098 91.281 -54.859 55.674
598099 91.318 -54.872 55.674
598100 91.368 -54.879 55.674
598101 91.422 -54.884 55.670
598102 91.475 -54.888 55.658
598103 91.528 -54.894 55.641
598104 91.582 -54.898 55.615
598105 91.635 -54.903 55.583
598106 91.675 -54.907 55.548
598107 91.706 -54.911 55.519
598108 91.735 -54.914 55.495
598109 91.766 -54.918 55.476
598110 91.796 -54.921 55.460
598111 91.826 -54.925 55.446
598112 91.856 -54.928 55.431
598113 91.884 -54.938 55.419
598114 91.911 -54.950 55.405
598115 91.940 -54.963 55.391
598116 91.972 -54.984 55.377
598117 92.005 -55.004 55.362
598118 92.038 -55.025 55.349
598119 92.072 -55.043 55.337
598120 92.105 -55.060 55.329
598121 92.140 -55.078 55.326
598122 92.173 -55.095 55.330
598123 92.207 -55.113 55.340
598124 92.241 -55.130 55.356
598125 92.275 -55.148 55.372
598126 92.313 -55.151 55.381
598127 92.352 -55.155 55.386
598128 92.391 -55.158 55.385
598129 92.419 -55.168 55.381
598130 92.445 -55.179 55.378
598131 92.472 -55.190 55.372
598132 92.498 -55.201 55.362
598133 92.525 -55.212 55.354
598134 92.552 -55.223 55.346
598135 92.578 -55.234 55.340
598136 92.611 -55.247 55.335
598137 92.645 -55.260 55.342
598138 92.678 -55.273 55.343
598139 92.712 -55.287 55.341
598140 92.742 -55.305 55.345
598141 92.773 -55.323 55.357
598142 92.805 -55.342 55.369
598143 92.842 -55.364 55.376
598144 92.879 -55.386 55.379
598145 92.917 -55.408 55.381
598146 92.954 -55.429 55.381
598147 92.991 -55.451 55.382
598148 93.028 -55.472 55.383
598149 93.066 -55.493 55.387
598150 93.103 -55.515 55.385
598151 93.140 -55.536 55.378
598152 93.177 -55.558 55.366
598153 93.214 -55.580 55.347
598154 93.252 -55.602 55.325
598155 93.289 -55.623 55.305
598156 93.326 -55.645 55.292
598157 93.363 -55.666 55.286
598158 93.400 -55.674 55.283
598159 93.435 -55.671 55.276
598160 93.471 -55.669 55.265
598161 93.507 -55.667 55.253
598162 93.543 -55.665 55.241
598163 93.578 -55.665 55.233
598164 93.612 -55.667 55.229
598165 93.645 -55.675 55.226
598166 93.677 -55.698 55.215
598167 93.706 -55.721 55.201
598168 93.726 -55.750 55.187
598169 93.746 -55.780 55.167
598170 93.766 -55.809 55.144
598171 93.785 -55.839 55.118
598172 93.805 -55.868 55.091
598173 93.825 -55.898 55.066
598174 93.845 -55.927 55.040
598175 93.865 -55.956 55.015
598176 93.885 -55.986 54.988
598177 93.905 -56.016 54.961
598178 93.933 -56.029 54.934
598179 93.962 -56.041 54.904
598180 93.991 -56.052 54.874
598181 94.020 -56.064 54.839
598182 94.049 -56.075 54.803
598183 94.078 -56.087 54.773
598184 94.107 -56.098 54.744
598185 94.136 -56.110 54.716
598186 94.168 -56.119 54.685
598187 94.200 -56.125 54.648
598188 94.234 -56.134 54.607
598189 94.268 -56.142 54.564
598190 94.301 -56.150 54.518
598191 94.336 -56.159 54.472
598192 94.369 -56.168 54.423
598193 94.403 -56.176 54.370
598194 94.437 -56.185 54.317
598195 94.471 -56.193 54.265
598196 94.505 -56.201 54.214
598197 94.543 -56.210 54.160
598198 94.579 -56.219 54.103
598199 94.613 -56.226 54.044
598200 94.647 -56.235 53.983
598201 94.681 -56.242 53.924
598202 94.715 -56.250 53.865
598203 94.749 -56.258 53.811
598204 94.783 -56.266 53.758
598205 94.817 -56.273 53.704
598206 94.851 -56.281 53.653
598207 94.885 -56.289 53.605
598208 94.919 -56.295 53.560
598209 94.961 -56.310 53.518
598210 95.008 -56.315 53.475
598211 95.056 -56.320 53.430
598212 95.095 -56.334 53.383
598213 95.121 -56.360 53.333
598214 95.147 -56.386 53.283
598215 95.173 -56.413 53.230
598216 95.198 -56.439 53.176
598217 95.216 -56.451 53.120
598218 95.234 -56.464 53.064
598219 95.251 -56.476 53.009
598220 95.282 -56.491 52.954
598221 95.323 -56.506 52.898
598222 95.364 -56.521 52.840
598223 95.404 -56.536 52.782
598224 95.442 -56.545 52.725
598225 95.479 -56.555 52.668
598226 95.517 -56.564 52.614
598227 95.548 -56.568 52.561
598228 95.575 -56.568 52.511
598229 95.602 -56.569 52.463
598230 95.628 -56.570 52.416
598231 95.656 -56.570 52.370
598232 95.682 -56.570 52.326
598233 95.709 -56.571 52.283
598234 95.742 -56.575 52.242
598235 95.776 -56.581 52.202
598236 95.811 -56.586 52.161
598237 95.846 -56.592 52.120
598238 95.882 -56.599 52.078
598239 95.916 -56.605 52.037
598240 95.937 -56.614 51.998
598241 95.956 -56.621 51.962
598242 95.974 -56.630 51.929
598243 95.993 -56.638 51.898
598244 96.012 -56.646 51.866
598245 96.031 -56.654 51.831
598246 96.050 -56.661 51.800
598247 96.068 -56.670 51.771
598248 96.088 -56.677 51.744
598249 96.106 -56.686 51.718
598250 96.125 -56.693 51.694
598251 96.148 -56.703 51.666
598252 96.176 -56.714 51.638
598253 96.209 -56.725 51.611
598254 96.244 -56.736 51.586
598255 96.278 -56.746 51.563
598256 96.300 -56.760 51.543
598257 96.315 -56.775 51.523
598258 96.343 -56.777 51.505
598259 96.370 -56.780 51.491
598260 96.398 -56.782 51.478
598261 96.426 -56.783 51.462
598262 96.451 -56.783 51.444
598263 96.474 -56.777 51.425
598264 96.498 -56.780 51.403
598265 96.520 -56.787 51.384
598266 96.548 -56.785 51.366
598267 96.575 -56.783 51.349
598268 96.602 -56.783 51.332
598269 96.629 -56.782 51.315
598270 96.656 -56.782 51.298
598271 96.682 -56.781 51.281
598272 96.709 -56.780 51.264
598273 96.736 -56.780 51.245
598274 96.764 -56.783 51.227
598275 96.796 -56.798 51.209
598276 96.828 -56.812 51.188
598277 96.860 -56.827 51.167
598278 96.887 -56.834 51.147
598279 96.914 -56.829 51.123
598280 96.966 -56.832 51.096
598281 97.009 -56.834 51.074
598282 97.042 -56.834 51.054
598283 97.076 -56.834 51.034
598284 97.109 -56.831 51.013
598285 97.142 -56.823 50.994
598286 97.175 -56.815 50.977
598287 97.208 -56.807 50.969
598288 97.241 -56.799 50.958
598289 97.275 -56.790 50.948
598290 97.308 -56.783 50.938
598291 97.341 -56.774 50.929
598292 97.374 -56.766 50.918
598293 97.407 -56.758 50.902
598294 97.440 -56.750 50.884
598295 97.481 -56.757 50.868
598296 97.522 -56.764 50.852
598297 97.563 -56.771 50.834
598298 97.604 -56.777 50.815
598299 97.645 -56.784 50.797
598300 97.684 -56.793 50.781
598301 97.717 -56.808 50.764
598302 97.750 -56.823 50.749
598303 97.785 -56.831 50.734
598304 97.818 -56.840 50.721
598305 97.853 -56.848 50.709
598306 97.887 -56.857 50.695
598307 97.921 -56.865 50.681
598308 97.955 -56.874 50.668
598309 97.989 -56.882 50.655
598310 98.023 -56.890 50.642
598311 98.057 -56.899 50.629
598312 98.091 -56.908 50.607
598313 98.125 -56.916 50.581
598314 98.159 -56.924 50.555
598315 98.193 -56.933 50.532
598316 98.227 -56.941 50.510
598317 98.261 -56.950 50.489
598318 98.295 -56.958 50.469
598319 98.329 -56.967 50.445
598320 98.363 -56.975 50.417
598321 98.397 -56.984 50.388
598322 98.431 -56.992 50.359
598323 98.466 -56.998 50.328
598324 98.501 -57.000 50.296
598325 98.536 -57.002 50.269
598326 98.572 -57.004 50.244
598327 98.603 -57.002 50.218
598328 98.631 -56.997 50.192
598329 98.660 -56.993 50.168
598330 98.688 -56.988 50.144
598331 98.724 -56.999 50.119
598332 98.764 -57.019 50.094
598333 98.804 -57.040 50.071
598334 98.844 -57.061 50.048
598335 98.881 -57.065 50.024
598336 98.916 -57.061 50.004
598337 98.950 -57.056 49.985
598338 98.991 -57.056 49.964
598339 99.035 -57.056 49.940
598340 99.080 -57.056 49.914
598341 99.124 -57.056 49.887
598342 99.145 -57.065 49.859
598343 99.161 -57.076 49.831
598344 99.177 -57.087 49.803
598345 99.193 -57.097 49.773
598346 99.210 -57.108 49.741
598347 99.228 -57.117 49.709
598348 99.256 -57.122 49.678
598349 99.283 -57.125 49.647
598350 99.311 -57.130 49.615
598351 99.339 -57.134 49.581
598352 99.366 -57.138 49.548
598353 99.394 -57.143 49.516
598354 99.422 -57.147 49.484
598355 99.449 -57.151 49.451
598356 99.477 -57.156 49.421
598357 99.503 -57.161 49.390
598358 99.528 -57.167 49.358
598359 99.560 -57.183 49.326
598360 99.605 -57.213 49.291
598361 99.648 -57.244 49.257
598362 99.645 -57.203 49.221
598363 99.675 -57.228 49.185
598364 99.700 -57.245 49.148
598365 99.726 -57.263 49.111
598366 99.752 -57.281 49.075
598367 99.777 -57.299 49.041
598368 99.802 -57.316 49.006
598369 99.828 -57.335 48.973
598370 99.854 -57.352 48.941
598371 99.879 -57.370 48.910
598372 99.905 -57.388 48.879
598373 99.930 -57.406 48.851
598374 99.956 -57.424 48.825
598375 99.981 -57.442 48.800
598376 100.007 -57.460 48.776
598377 100.033 -57.477 48.752
598709 88.372 -53.194 57.461
598710 88.407 -53.210 57.462
598711 88.443 -53.225 57.462
598712 88.478 -53.240 57.463
598713 88.514 -53.256 57.464
598714 88.552 -53.276 57.464
598715 88.590 -53.295 57.465
598716 88.628 -53.314 57.465
598717 88.666 -53.333 57.464
598718 88.704 -53.352 57.463
598719 88.742 -53.372 57.462
598720 88.780 -53.391 57.460
598721 88.818 -53.411 57.458
598722 88.856 -53.430 57.455
598723 88.894 -53.449 57.452
598724 88.932 -53.468 57.448
598725 88.970 -53.488 57.444
598726 89.009 -53.507 57.440
598727 89.046 -53.526 57.435
598728 89.085 -53.546 57.429
598729 89.122 -53.565 57.423
598730 89.161 -53.584 57.417
598731 89.198 -53.603 57.410
598732 89.237 -53.628 57.402
598733 89.267 -53.655 57.393
598734 89.298 -53.683 57.382
598735 89.328 -53.711 57.372
598736 89.358 -53.739 57.359
598737 89.388 -53.767 57.346
598738 89.420 -53.784 57.330
598739 89.454 -53.801 57.312
598740 89.488 -53.819 57.293
598741 89.523 -53.837 57.274
598742 89.557 -53.854 57.253
598743 89.592 -53.872 57.230
598744 89.626 -53.889 57.204
598745 89.660 -53.907 57.174
598746 89.694 -53.924 57.140
598747 89.729 -53.941 57.104
598748 89.763 -53.960 57.066
598749 89.798 -53.977 57.027
598750 89.832 -53.994 56.984
598751 89.866 -54.012 56.940
598752 89.900 -54.029 56.896
598753 89.935 -54.047 56.853
598754 89.969 -54.064 56.814
598755 90.004 -54.082 56.774
598760 90.175 -54.170 56.423
598761 90.210 -54.187 56.389
598762 90.244 -54.205 56.355
598763 90.278 -54.223 56.321
598764 90.313 -54.240 56.290
598765 90.348 -54.257 56.261
598766 90.391 -54.268 56.234
598767 90.435 -54.279 56.208
598768 90.464 -54.295 56.184
598769 90.481 -54.315 56.161
598770 90.498 -54.336 56.142
598771 90.516 -54.355 56.129
598772 90.533 -54.376 56.108
598773 90.550 -54.395 56.082
598774 90.567 -54.416 56.054
598775 90.584 -54.436 56.028
598776 90.601 -54.456 56.003
598777 90.619 -54.477 55.983
598778 90.636 -54.496 55.969
598779 90.653 -54.517 55.959
598780 90.670 -54.537 55.947
598781 90.687 -54.557 55.931
598782 90.705 -54.577 55.916
598783 90.727 -54.593 55.904
598784 90.757 -54.608 55.888
598785 90.787 -54.628 55.868
598786 90.815 -54.648 55.843
598787 90.847 -54.667 55.816
598788 90.883 -54.681 55.789
598789 90.921 -54.694 55.767
598790 90.957 -54.708 55.749
598791 90.994 -54.721 55.733
598792 91.031 -54.735 55.717
598793 91.068 -54.749 55.701
598794 91.105 -54.762 55.686
598795 91.142 -54.776 55.675
598796 91.179 -54.790 55.670
598797 91.216 -54.803 55.671
598798 91.253 -54.817 55.673
598799 91.290 -54.830 55.675
598800 91.331 -54.841 55.674
598801 91.385 -54.847 55.674
598802 91.439 -54.851 55.669
598803 91.492 -54.856 55.659
598804 91.545 -54.861 55.641
598805 91.599 -54.866 55.616
598806 91.652 -54.871 55.583
598807 91.693 -54.875 55.546
598808 91.723 -54.878 55.514
598809 91.753 -54.882 55.489
598810 91.784 -54.885 55.470
598811 91.813 -54.889 55.454
598812 91.844 -54.892 55.439
598813 91.872 -54.900 55.425
598814 91.899 -54.911 55.411
598815 91.927 -54.922 55.397
598816 91.954 -54.933 55.384
598817 91.987 -54.953 55.372
598818 92.019 -54.974 55.359
598819 92.052 -54.994 55.346
598820 92.085 -55.015 55.335
598821 92.118 -55.035 55.327
598822 92.152 -55.052 55.323
598823 92.186 -55.069 55.325
598824 92.220 -55.087 55.332
598825 92.253 -55.105 55.347
598826 92.289 -55.118 55.362
598827 92.328 -55.121 55.370
598828 92.366 -55.125 55.373
598829 92.405 -55.128 55.373
598830 92.438 -55.135 55.369
598831 92.465 -55.145 55.365
598832 92.492 -55.157 55.358
598833 92.518 -55.167 55.348
598834 92.545 -55.178 55.340
598835 92.571 -55.189 55.333
598836 92.597 -55.200 55.328
598837 92.624 -55.211 55.329
598838 92.651 -55.222 55.335
598839 92.681 -55.234 55.335
598840 92.715 -55.247 55.336
598841 92.747 -55.263 55.344
598842 92.778 -55.282 55.357
598843 92.815 -55.304 55.365
598844 92.852 -55.325 55.371
598845 92.889 -55.347 55.376
598846 92.927 -55.368 55.378
598847 92.964 -55.390 55.379
598848 93.001 -55.411 55.382
598849 93.039 -55.433 55.385
598850 93.076 -55.455 55.387
598851 93.113 -55.476 55.387
598852 93.150 -55.498 55.382
598853 93.187 -55.519 55.370
598854 93.224 -55.541 55.353
598855 93.262 -55.562 55.330
598856 93.299 -55.584 55.307
598857 93.336 -55.605 55.291
598858 93.373 -55.627 55.285
598859 93.410 -55.644 55.284
598860 93.446 -55.642 55.279
598861 93.482 -55.640 55.269
598862 93.517 -55.637 55.255
598863 93.554 -55.635 55.242
598864 93.589 -55.633 55.230
598865 93.624 -55.633 55.219
598866 93.657 -55.642 55.205
598867 93.689 -55.665 55.191
598868 93.709 -55.693 55.177
598869 93.728 -55.723 55.163
598870 93.749 -55.752 55.146
598871 93.768 -55.782 55.124
598872 93.788 -55.812 55.103
598873 93.808 -55.841 55.084
598874 93.828 -55.870 55.061
598875 93.848 -55.900 55.039
598876 93.868 -55.929 55.015
598877 93.888 -55.959 54.989
598878 93.908 -55.988 54.963
598879 93.933 -56.006 54.936
598880 93.963 -56.017 54.907
598881 93.992 -56.028 54.876
598882 94.021 -56.041 54.838
598883 94.050 -56.052 54.797
598884 94.079 -56.063 54.761
598885 94.108 -56.075 54.728
598886 94.137 -56.086 54.697
598887 94.170 -56.093 54.665
598888 94.202 -56.100 54.629
598889 94.236 -56.107 54.590
598890 94.270 -56.116 54.549
598891 94.303 -56.124 54.506
598892 94.337 -56.133 54.461
598893 94.371 -56.141 54.413
598894 94.405 -56.150 54.361
598895 94.439 -56.158 54.306
598896 94.473 -56.166 54.253
598897 94.507 -56.175 54.201
598898 94.549 -56.185 54.148
598899 94.589 -56.194 54.091
598900 94.623 -56.202 54.032
598901 94.657 -56.210 53.973
598902 94.691 -56.217 53.916
598903 94.725 -56.226 53.861
598904 94.759 -56.233 53.808
598905 94.793 -56.241 53.757
598906 94.827 -56.249 53.706
598907 94.861 -56.257 53.655
598908 94.895 -56.264 53.608
598909 94.929 -56.269 53.562
598910 94.967 -56.288 53.521
598911 95.009 -56.304 53.479
598912 95.056 -56.309 53.435
598913 95.103 -56.315 53.389
598914 95.129 -56.341 53.340
598915 95.156 -56.367 53.290
598916 95.182 -56.394 53.236
598917 95.203 -56.413 53.181
598918 95.220 -56.426 53.124
598919 95.238 -56.439 53.067
598920 95.256 -56.451 53.009
598921 95.289 -56.466 52.952
598922 95.330 -56.480 52.896
598923 95.370 -56.495 52.839
598924 95.409 -56.507 52.782
598925 95.447 -56.516 52.724
598926 95.484 -56.525 52.666
598927 95.522 -56.534 52.610
598928 95.555 -56.540 52.555
598929 95.582 -56.540 52.502
598930 95.609 -56.541 52.451
598931 95.635 -56.542 52.402
598932 95.662 -56.542 52.354
598933 95.689 -56.542 52.308
598934 95.716 -56.543 52.263
598935 95.751 -56.549 52.219
598936 95.786 -56.555 52.178
598937 95.821 -56.560 52.137
598938 95.855 -56.566 52.095
598939 95.890 -56.571 52.052
598940 95.919 -56.580 52.010
598941 95.945 -56.583 51.970
598942 95.964 -56.590 51.933
598943 95.981 -56.598 51.898
598944 95.998 -56.605 51.864
598945 96.015 -56.614 51.831
598946 96.032 -56.621 51.797
598947 96.049 -56.630 51.764
598948 96.067 -56.637 51.736
598949 96.087 -56.646 51.710
598950 96.105 -56.653 51.683
598951 96.124 -56.661 51.654
598952 96.153 -56.672 51.626
598953 96.187 -56.683 51.597
598954 96.222 -56.693 51.572
598955 96.256 -56.704 51.546
598956 96.289 -56.715 51.522
598957 96.304 -56.730 51.502
598958 96.319 -56.746 51.480
598959 96.344 -56.750 51.461
598960 96.372 -56.752 51.447
598961 96.399 -56.755 51.432
598962 96.427 -56.756 51.415
598963 96.450 -56.752 51.398
598964 96.474 -56.747 51.381
598965 96.497 -56.747 51.364
598966 96.520 -56.754 51.345
598967 96.542 -56.760 51.326
598968 96.571 -56.757 51.309
598969 96.599 -56.755 51.291
598970 96.626 -56.754 51.271
598971 96.653 -56.753 51.251
598972 96.679 -56.753 51.235
598973 96.707 -56.752 51.217
598974 96.733 -56.752 51.199
598975 96.760 -56.753 51.180
598976 96.792 -56.767 51.162
598977 96.824 -56.781 51.142
598978 96.856 -56.796 51.123
598979 96.880 -56.796 51.104
598980 96.912 -56.794 51.082
598981 96.963 -56.797 51.059
598982 97.007 -56.799 51.039
598983 97.040 -56.799 51.018
598984 97.074 -56.799 50.997
598985 97.107 -56.799 50.975
598986 97.141 -56.790 50.953
598987 97.174 -56.783 50.933
598988 97.207 -56.774 50.924
598989 97.240 -56.766 50.916
598990 97.273 -56.758 50.908
598991 97.306 -56.750 50.899
598992 97.339 -56.742 50.890
598993 97.372 -56.734 50.879
598994 97.405 -56.726 50.864
598995 97.443 -56.726 50.849
598996 97.484 -56.733 50.834
598997 97.525 -56.740 50.819
598998 97.566 -56.746 50.802
598999 97.607 -56.753 50.784
599000 97.648 -56.760 50.766
599001 97.689 -56.767 50.749
599002 97.724 -56.780 50.733
599003 97.757 -56.791 50.718
599004 97.791 -56.799 50.705
599005 97.825 -56.808 50.693
599006 97.859 -56.817 50.683
599007 97.893 -56.825 50.672
599008 97.928 -56.834 50.659
599009 97.961 -56.842 50.647
599010 97.996 -56.850 50.634
599011 98.029 -56.859 50.621
599012 98.064 -56.868 50.608
599013 98.098 -56.876 50.589
599014 98.132 -56.884 50.563
599015 98.166 -56.893 50.538
599016 98.200 -56.901 50.515
599017 98.234 -56.909 50.494
599018 98.268 -56.918 50.474
599019 98.302 -56.927 50.453
599020 98.336 -56.935 50.426
599021 98.370 -56.943 50.399
599022 98.404 -56.952 50.369
599023 98.438 -56.960 50.339
599024 98.472 -56.968 50.307
599025 98.507 -56.974 50.276
599026 98.543 -56.976 50.249
599027 98.577 -56.977 50.223
599028 98.605 -56.972 50.197
599029 98.634 -56.968 50.172
599030 98.662 -56.963 50.147
599031 98.693 -56.965 50.124
599032 98.725 -56.967 50.098
599033 98.764 -56.984 50.074
599034 98.803 -57.005 50.050
599035 98.844 -57.025 50.026
599036 98.882 -57.039 50.003
599037 98.917 -57.034 49.981
599038 98.951 -57.030 49.960
599039 98.995 -57.030 49.939
599040 99.039 -57.030 49.916
599041 99.083 -57.031 49.890
599042 99.127 -57.031 49.864
599043 99.158 -57.036 49.837
599044 99.174 -57.047 49.809
599045 99.190 -57.057 49.780
599046 99.207 -57.068 49.750
599047 99.223 -57.078 49.719
599048 99.245 -57.086 49.688
599049 99.273 -57.091 49.658
599050 99.300 -57.094 49.628
599051 99.328 -57.099 49.596
599052 99.356 -57.103 49.564
599053 99.383 -57.107 49.533
599054 99.411 -57.112 49.502
599055 99.439 -57.116 49.472
599056 99.467 -57.120 49.440
599057 99.491 -57.126 49.410
599058 99.516 -57.133 49.379
599059 99.541 -57.139 49.346
599060 99.576 -57.160 49.313
599061 99.621 -57.191 49.279
599062 99.665 -57.222 49.244
599063 99.657 -57.185 49.208
599064 99.686 -57.203 49.172
599065 99.711 -57.222 49.136
599066 99.737 -57.239 49.099
599067 99.762 -57.257 49.063
599068 99.788 -57.275 49.029
599069 99.814 -57.293 48.996
599070 99.839 -57.310 48.963
599071 99.864 -57.329 48.931
599072 99.890 -57.347 48.900
599073 99.916 -57.364 48.869
599074 99.941 -57.382 48.841
599075 99.967 -57.400 48.814
599076 99.992 -57.418 48.789
599077 100.018 -57.435 48.765
599409 88.361 -53.144 57.456
599410 88.397 -53.159 57.457
599411 88.432 -53.175 57.458
599412 88.468 -53.190 57.458
599413 88.503 -53.206 57.459
599414 88.541 -53.226 57.459
599415 88.579 -53.245 57.460
599416 88.618 -53.264 57.460
599417 88.656 -53.283 57.460
599418 88.694 -53.302 57.459
599419 88.732 -53.322 57.458
599420 88.770 -53.341 57.456
599421 88.808 -53.361 57.454
599422 88.846 -53.380 57.452
599423 88.884 -53.399 57.448
599424 88.922 -53.418 57.445
599425 88.960 -53.438 57.441
599426 88.998 -53.457 57.437
599427 89.036 -53.477 57.432
599428 89.074 -53.496 57.426
599429 89.112 -53.515 57.420
599430 89.150 -53.534 57.414
599431 89.188 -53.553 57.408
599432 89.221 -53.577 57.400
599433 89.248 -53.606 57.392
599434 89.278 -53.634 57.383
599435 89.308 -53.662 57.372
599436 89.338 -53.690 57.361
599437 89.368 -53.719 57.349
599438 89.398 -53.747 57.336
599439 89.435 -53.758 57.321
599440 89.464 -53.772 57.304
599441 89.498 -53.789 57.285
599442 89.532 -53.806 57.267
599443 89.566 -53.824 57.247
599444 89.601 -53.842 57.225
599445 89.635 -53.860 57.200
599446 89.670 -53.877 57.171
599447 89.704 -53.894 57.138
599448 89.738 -53.912 57.103
599449 89.772 -53.929 57.065
599450 89.807 -53.947 57.026
599451 89.841 -53.965 56.984
599452 89.876 -53.982 56.941
599453 89.910 -54.000 56.897
599454 89.944 -54.017 56.855
599455 89.978 -54.035 56.814
599456 90.013 -54.052 56.774
599457 90.047 -54.070 56.735
599461 90.184 -54.140 56.424
599462 90.219 -54.157 56.389
599463 90.253 -54.175 56.354
599464 90.288 -54.192 56.320
599465 90.322 -54.211 56.288
599466 90.364 -54.223 56.258
599467 90.407 -54.234 56.230
599468 90.450 -54.245 56.204
599469 90.471 -54.264 56.179
599470 90.488 -54.284 56.156
599471 90.506 -54.304 56.135
599472 90.523 -54.324 56.119
599473 90.540 -54.344 56.098
599474 90.557 -54.364 56.073
599475 90.575 -54.385 56.046
599476 90.592 -54.405 56.020
599477 90.609 -54.425 55.996
599478 90.626 -54.445 55.976
599479 90.644 -54.465 55.962
599480 90.661 -54.486 55.951
599481 90.678 -54.505 55.938
599482 90.695 -54.526 55.922
599483 90.714 -54.544 55.907
599484 90.748 -54.551 55.896
599485 90.777 -54.571 55.879
599486 90.806 -54.592 55.858
599487 90.834 -54.612 55.833
599488 90.864 -54.632 55.805
599489 90.892 -54.652 55.777
599490 90.929 -54.666 55.755
599491 90.966 -54.680 55.736
599492 91.003 -54.693 55.719
599493 91.040 -54.706 55.704
599494 91.077 -54.720 55.690
599495 91.114 -54.734 55.677
599496 91.151 -54.747 55.670
599497 91.188 -54.761 55.668
599498 91.225 -54.775 55.669
599499 91.262 -54.788 55.672
599500 91.299 -54.802 55.675
599501 91.349 -54.809 55.675
599502 91.402 -54.814 55.675
599503 91.456 -54.818 55.670
599504 91.509 -54.824 55.659
599505 91.563 -54.828 55.641
599506 91.616 -54.834 55.615
599507 91.669 -54.838 55.582
599508 91.711 -54.843 55.543
599509 91.741 -54.846 55.509
599510 91.771 -54.850 55.483
599511 91.801 -54.853 55.463
599512 91.831 -54.856 55.447
599513 91.861 -54.861 55.433
599514 91.888 -54.872 55.419
599515 91.916 -54.884 55.405
599516 91.943 -54.894 55.392
599517 91.970 -54.906 55.380
599518 92.001 -54.923 55.369
599519 92.033 -54.944 55.356
599520 92.067 -54.964 55.343
599521 92.099 -54.985 55.333
599522 92.132 -55.005 55.324
599523 92.165 -55.025 55.319
599524 92.199 -55.044 55.319
599525 92.232 -55.062 55.324
599526 92.266 -55.079 55.338
599527 92.303 -55.088 55.351
599528 92.342 -55.091 55.357
599529 92.380 -55.094 55.358
599530 92.419 -55.097 55.358
599531 92.458 -55.101 55.356
599532 92.485 -55.111 55.350
599533 92.511 -55.123 55.344
599534 92.537 -55.133 55.336
599535 92.564 -55.144 55.328
599536 92.591 -55.155 55.323
599537 92.617 -55.166 55.321
599538 92.644 -55.177 55.324
599539 92.670 -55.188 55.328
599540 92.697 -55.199 55.329
599541 92.723 -55.210 55.332
599542 92.752 -55.222 55.340
599543 92.788 -55.243 55.349
599544 92.825 -55.264 55.356
599545 92.862 -55.286 55.361
599546 92.900 -55.307 55.366
599547 92.937 -55.329 55.371
599548 92.974 -55.351 55.376
599549 93.011 -55.373 55.381
599550 93.049 -55.394 55.380
599551 93.086 -55.415 55.380
599552 93.123 -55.437 55.383
599553 93.160 -55.458 55.381
599554 93.197 -55.480 55.370
599555 93.234 -55.502 55.354
599556 93.272 -55.524 55.333
599557 93.309 -55.545 55.308
599558 93.346 -55.567 55.291
599559 93.383 -55.588 55.284
599560 93.421 -55.609 55.283
599561 93.457 -55.613 55.279
599562 93.492 -55.611 55.269
599563 93.529 -55.608 55.256
599564 93.564 -55.606 55.244
599565 93.600 -55.604 55.231
599566 93.636 -55.602 55.215
599567 93.670 -55.608 55.196
599568 93.692 -55.636 55.179
599569 93.711 -55.666 55.161
599570 93.731 -55.696 55.146
599571 93.751 -55.725 55.130
599572 93.771 -55.755 55.112
599573 93.791 -55.784 55.094
599574 93.811 -55.813 55.078
599575 93.831 -55.843 55.057
599576 93.851 -55.872 55.037
599577 93.870 -55.902 55.014
599578 93.891 -55.931 54.987
599579 93.910 -55.961 54.961
599580 93.934 -55.982 54.935
599581 93.964 -55.994 54.908
599582 93.992 -56.006 54.879
599583 94.022 -56.017 54.843
599584 94.051 -56.028 54.800
599585 94.079 -56.040 54.754
599586 94.109 -56.051 54.711
599587 94.140 -56.060 54.676
599588 94.173 -56.067 54.641
599589 94.205 -56.074 54.605
599590 94.238 -56.081 54.568
599591 94.271 -56.090 54.529
599592 94.306 -56.098 54.489
599593 94.339 -56.107 54.446
599594 94.373 -56.115 54.399
599595 94.407 -56.123 54.347
599596 94.441 -56.132 54.292
599597 94.475 -56.140 54.237
599598 94.509 -56.149 54.186
599599 94.556 -56.160 54.135
599600 94.599 -56.169 54.081
599601 94.633 -56.177 54.024
599602 94.667 -56.185 53.967
599603 94.701 -56.193 53.913
599604 94.735 -56.201 53.863
599605 94.769 -56.208 53.813
599606 94.803 -56.216 53.764
599607 94.837 -56.224 53.714
599608 94.871 -56.232 53.664
599609 94.905 -56.238 53.614
599610 94.939 -56.241 53.565
599611 94.976 -56.259 53.522
599612 95.015 -56.282 53.481
599613 95.057 -56.298 53.437
599614 95.105 -56.303 53.392
599615 95.138 -56.322 53.345
599616 95.164 -56.348 53.297
599617 95.190 -56.374 53.243
599618 95.207 -56.387 53.188
599619 95.225 -56.400 53.132
599620 95.243 -56.413 53.074
599621 95.260 -56.426 53.015
599622 95.297 -56.440 52.957
599623 95.337 -56.455 52.901
599624 95.376 -56.468 52.844
599625 95.414 -56.477 52.785
599626 95.451 -56.486 52.725
599627 95.489 -56.495 52.665
599628 95.526 -56.505 52.609
599629 95.562 -56.512 52.551
599630 95.589 -56.512 52.496
599631 95.616 -56.513 52.442
599632 95.642 -56.514 52.389
599633 95.669 -56.514 52.340
599634 95.696 -56.514 52.291
599635 95.727 -56.517 52.244
599636 95.761 -56.523 52.199
599637 95.796 -56.528 52.156
599638 95.831 -56.534 52.114
599639 95.865 -56.539 52.071
599640 95.896 -56.550 52.028
599641 95.922 -56.561 51.985
599642 95.947 -56.564 51.943
599643 95.972 -56.567 51.905
599644 95.997 -56.570 51.868
599645 96.020 -56.574 51.832
599646 96.037 -56.583 51.797
599647 96.053 -56.590 51.763
599648 96.071 -56.599 51.731
599649 96.088 -56.606 51.702
599650 96.104 -56.614 51.674
599651 96.121 -56.622 51.646
599652 96.138 -56.630 51.617
599653 96.166 -56.639 51.589
599654 96.200 -56.650 51.559
599655 96.235 -56.661 51.532
599656 96.269 -56.671 51.507
599657 96.293 -56.685 51.483
599658 96.308 -56.700 51.460
599659 96.323 -56.715 51.439
599660 96.345 -56.723 51.419
599661 96.373 -56.725 51.403
599662 96.401 -56.727 51.388
599663 96.426 -56.726 51.371
599664 96.450 -56.721 51.354
599665 96.473 -56.716 51.336
599666 96.497 -56.714 51.319
599667 96.519 -56.721 51.304
599668 96.542 -56.727 51.290
599669 96.565 -56.734 51.275
599670 96.594 -56.730 51.256
599671 96.623 -56.726 51.234
599672 96.650 -56.725 51.210
599673 96.677 -56.725 51.187
599674 96.703 -56.724 51.167
599675 96.730 -56.724 51.147
599676 96.757 -56.723 51.128
599677 96.789 -56.736 51.110
599678 96.820 -56.750 51.091
599679 96.852 -56.765 51.074
599680 96.872 -56.760 51.055
599681 96.910 -56.758 51.037
599682 96.961 -56.762 51.021
599683 97.006 -56.764 51.000
599684 97.039 -56.764 50.978
599685 97.073 -56.765 50.955
599686 97.106 -56.765 50.931
599687 97.139 -56.758 50.908
599688 97.172 -56.750 50.888
599689 97.206 -56.743 50.879
599690 97.238 -56.734 50.874
599691 97.272 -56.726 50.868
599692 97.305 -56.718 50.861
599693 97.338 -56.710 50.852
599694 97.371 -56.702 50.840
599695 97.404 -56.695 50.828
599696 97.445 -56.702 50.816
599697 97.487 -56.708 50.802
599698 97.527 -56.715 50.788
599699 97.569 -56.722 50.772
599700 97.609 -56.729 50.755
599701 97.650 -56.736 50.736
599702 97.691 -56.743 50.719
599703 97.730 -56.751 50.704
599704 97.764 -56.759 50.689
599705 97.798 -56.768 50.676
599706 97.832 -56.777 50.665
599707 97.866 -56.785 50.656
599708 97.900 -56.793 50.648
599709 97.934 -56.802 50.637
599710 97.968 -56.810 50.625
599711 98.002 -56.818 50.612
599712 98.036 -56.827 50.599
599713 98.070 -56.836 50.585
599714 98.104 -56.844 50.568
599715 98.138 -56.852 50.544
599716 98.172 -56.861 50.520
599717 98.207 -56.869 50.497
599718 98.240 -56.878 50.477
599719 98.275 -56.887 50.457
599720 98.308 -56.895 50.435
599721 98.343 -56.903 50.407
599722 98.377 -56.912 50.379
599723 98.411 -56.920 50.349
599724 98.445 -56.928 50.317
599725 98.479 -56.937 50.285
599726 98.513 -56.946 50.255
599727 98.548 -56.950 50.228
599728 98.579 -56.948 50.201
599729 98.608 -56.943 50.175
599730 98.640 -56.946 50.149
599731 98.671 -56.949 50.124
599732 98.703 -56.951 50.101
599733 98.735 -56.953 50.076
599734 98.767 -56.956 50.051
599735 98.803 -56.969 50.027
599736 98.843 -56.990 50.004
599737 98.883 -57.010 49.981
599738 98.918 -57.008 49.958
599739 98.954 -57.004 49.935
599740 98.999 -57.004 49.914
599741 99.043 -57.005 49.891
599742 99.087 -57.005 49.866
599743 99.131 -57.005 49.840
599744 99.171 -57.007 49.813
599745 99.187 -57.018 49.784
599746 99.203 -57.028 49.755
599747 99.220 -57.039 49.725
599748 99.236 -57.050 49.695
599749 99.262 -57.055 49.666
599750 99.289 -57.059 49.636
599751 99.317 -57.063 49.608
599752 99.345 -57.068 49.579
599753 99.372 -57.072 49.549
599754 99.400 -57.076 49.519
599755 99.428 -57.081 49.489
599756 99.454 -57.086 49.459
599757 99.479 -57.092 49.428
599758 99.504 -57.098 49.397
599759 99.528 -57.105 49.366
599760 99.553 -57.111 49.333
599761 99.593 -57.136 49.300
599762 99.637 -57.167 49.265
599763 99.643 -57.140 49.230
599764 99.671 -57.162 49.195
599765 99.697 -57.180 49.159
599766 99.722 -57.197 49.123
599767 99.748 -57.216 49.087
599768 99.773 -57.233 49.051
599769 99.799 -57.251 49.017
599770 99.825 -57.269 48.985
599771 99.850 -57.287 48.952
599772 99.875 -57.304 48.920
599773 99.901 -57.322 48.889
599774 99.927 -57.341 48.859
599775 99.952 -57.358 48.831
599776 99.978 -57.376 48.804
599777 100.003 -57.394 48.779
599778 100.028 -57.412 48.755
600109 88.351 -53.094 57.452
600110 88.386 -53.109 57.452
600111 88.421 -53.124 57.452
600112 88.457 -53.139 57.454
600113 88.493 -53.156 57.455
600114 88.531 -53.176 57.455
600115 88.569 -53.195 57.455
600116 88.607 -53.214 57.455
600117 88.645 -53.233 57.456
600118 88.683 -53.252 57.455
600119 88.721 -53.272 57.454
600120 88.759 -53.291 57.453
600121 88.797 -53.311 57.451
600122 88.835 -53.330 57.448
600123 88.874 -53.349 57.445
600124 88.911 -53.368 57.442
600125 88.950 -53.388 57.438
600126 88.987 -53.407 57.434
600127 89.026 -53.427 57.430
600128 89.063 -53.446 57.424
600129 89.102 -53.465 57.417
600130 89.139 -53.484 57.411
600131 89.174 -53.506 57.405
600132 89.205 -53.529 57.398
600133 89.236 -53.553 57.390
600134 89.258 -53.586 57.381
600135 89.288 -53.615 57.372
600136 89.318 -53.643 57.361
600137 89.348 -53.671 57.350
600138 89.378 -53.699 57.338
600139 89.408 -53.728 57.325
600140 89.451 -53.733 57.310
600141 89.477 -53.744 57.294
600142 89.507 -53.759 57.277
600143 89.541 -53.777 57.260
600144 89.576 -53.794 57.240
600145 89.610 -53.812 57.221
600146 89.645 -53.829 57.197
600147 89.679 -53.847 57.168
600148 89.713 -53.865 57.136
600149 89.747 -53.882 57.101
600150 89.782 -53.900 57.065
600151 89.816 -53.917 57.026
600152 89.850 -53.935 56.984
600153 89.885 -53.952 56.942
600154 89.919 -53.969 56.899
600155 89.954 -53.988 56.856
600156 89.988 -54.005 56.815
600157 90.022 -54.022 56.775
600158 90.056 -54.040 56.735
600163 90.228 -54.128 56.389
600164 90.262 -54.145 56.354
600165 90.297 -54.163 56.319
600166 90.335 -54.177 56.287
600167 90.379 -54.189 56.255
600168 90.422 -54.200 56.227
600169 90.462 -54.212 56.200
600170 90.479 -54.232 56.175
600171 90.497 -54.252 56.150
600172 90.514 -54.273 56.128
600173 90.531 -54.292 56.109
600174 90.548 -54.313 56.088
600175 90.565 -54.333 56.064
600176 90.582 -54.353 56.037
600177 90.600 -54.373 56.011
600178 90.617 -54.393 55.989
600179 90.634 -54.414 55.969
600180 90.651 -54.433 55.953
600181 90.668 -54.454 55.940
600182 90.685 -54.474 55.925
600183 90.703 -54.494 55.910
600184 90.737 -54.497 55.897
600185 90.767 -54.515 55.885
600186 90.796 -54.535 55.868
600187 90.825 -54.555 55.846
600188 90.854 -54.575 55.820
600189 90.883 -54.596 55.794
600190 90.912 -54.615 55.766
600191 90.941 -54.636 55.742
600192 90.975 -54.651 55.723
600193 91.012 -54.665 55.706
600194 91.049 -54.678 55.692
600195 91.086 -54.692 55.680
600196 91.123 -54.706 55.670
600197 91.160 -54.719 55.666
600198 91.198 -54.733 55.666
600199 91.234 -54.746 55.669
600200 91.272 -54.760 55.673
600201 91.312 -54.772 55.676
600202 91.366 -54.777 55.676
600203 91.419 -54.781 55.674
600204 91.473 -54.786 55.667
600205 91.526 -54.791 55.653
600206 91.580 -54.796 55.634
600207 91.633 -54.801 55.607
600208 91.686 -54.806 55.574
600209 91.729 -54.810 55.536
600210 91.759 -54.813 55.504
600211 91.789 -54.817 55.477
600212 91.819 -54.820 55.457
600213 91.849 -54.824 55.441
600214 91.877 -54.834 55.426
600215 91.904 -54.845 55.412
600216 91.931 -54.856 55.400
600217 91.959 -54.867 55.387
600218 91.986 -54.878 55.376
600219 92.015 -54.893 55.366
600220 92.048 -54.913 55.353
600221 92.081 -54.934 55.341
600222 92.113 -54.954 55.330
600223 92.146 -54.975 55.321
600224 92.179 -54.995 55.315
600225 92.212 -55.016 55.313
600226 92.245 -55.036 55.316
600227 92.279 -55.054 55.327
600228 92.317 -55.057 55.336
600229 92.356 -55.060 55.340
600230 92.394 -55.063 55.341
600231 92.433 -55.067 55.342
600232 92.472 -55.070 55.341
600233 92.504 -55.078 55.336
600234 92.531 -55.088 55.332
600235 92.557 -55.100 55.325
600236 92.584 -55.110 55.318
600237 92.611 -55.121 55.315
600238 92.637 -55.132 55.317
600239 92.663 -55.143 55.321
600240 92.690 -55.154 55.324
600241 92.717 -55.165 55.328
600242 92.743 -55.176 55.334
600243 92.769 -55.194 55.342
600244 92.802 -55.212 55.350
600245 92.837 -55.230 55.357
600246 92.873 -55.248 55.365
600247 92.910 -55.269 55.375
600248 92.947 -55.290 55.383
600249 92.984 -55.312 55.389
600250 93.021 -55.333 55.392
600251 93.059 -55.354 55.387
600252 93.096 -55.376 55.382
600253 93.133 -55.398 55.382
600254 93.170 -55.420 55.379
600255 93.207 -55.441 55.369
600256 93.245 -55.463 55.351
600257 93.282 -55.484 55.328
600258 93.319 -55.506 55.307
600259 93.356 -55.527 55.292
600260 93.394 -55.549 55.284
600261 93.431 -55.571 55.280
600262 93.468 -55.583 55.275
600263 93.503 -55.581 55.267
600264 93.539 -55.579 55.256
600265 93.575 -55.577 55.246
600266 93.611 -55.574 55.233
600267 93.646 -55.572 55.217
600268 93.674 -55.583 55.197
600269 93.695 -55.609 55.174
600270 93.714 -55.639 55.154
600271 93.734 -55.668 55.136
600272 93.754 -55.698 55.120
600273 93.774 -55.727 55.102
600274 93.794 -55.756 55.086
600275 93.814 -55.786 55.070
600276 93.834 -55.815 55.050
600277 93.854 -55.845 55.030
600278 93.873 -55.875 55.007
600279 93.894 -55.903 54.982
600280 93.913 -55.933 54.958
600281 93.935 -55.959 54.933
600282 93.964 -55.970 54.908
600283 93.993 -55.982 54.878
600284 94.022 -55.994 54.841
600285 94.051 -56.005 54.797
600286 94.080 -56.016 54.750
600287 94.110 -56.028 54.703
600288 94.142 -56.034 54.666
600289 94.175 -56.041 54.631
600290 94.207 -56.048 54.594
600291 94.240 -56.055 54.554
600292 94.273 -56.063 54.511
600293 94.308 -56.072 54.469
600294 94.341 -56.080 54.426
600295 94.375 -56.088 54.380
600296 94.409 -56.097 54.329
600297 94.443 -56.106 54.274
600298 94.477 -56.114 54.221
600299 94.512 -56.122 54.170
600300 94.563 -56.134 54.120
600301 94.609 -56.145 54.072
600302 94.643 -56.153 54.019
600303 94.677 -56.160 53.965
600304 94.711 -56.168 53.915
600305 94.745 -56.176 53.869
600306 94.779 -56.184 53.823
600307 94.813 -56.191 53.776
600308 94.847 -56.199 53.726
600309 94.881 -56.207 53.674
600310 94.915 -56.212 53.622
600311 94.949 -56.215 53.570
600312 94.985 -56.230 53.525
600313 95.024 -56.254 53.482
600314 95.062 -56.277 53.439
600315 95.105 -56.292 53.394
600316 95.147 -56.303 53.348
600317 95.173 -56.329 53.302
600318 95.195 -56.348 53.250
600319 95.212 -56.361 53.196
600320 95.230 -56.374 53.140
600321 95.247 -56.387 53.082
600322 95.265 -56.400 53.022
600323 95.303 -56.414 52.963
600324 95.343 -56.429 52.905
600325 95.381 -56.439 52.848
600326 95.418 -56.448 52.790
600327 95.456 -56.457 52.731
600328 95.494 -56.466 52.672
600329 95.531 -56.475 52.611
600330 95.569 -56.484 52.551
600331 95.596 -56.484 52.492
600332 95.622 -56.485 52.436
600333 95.650 -56.486 52.381
600334 95.676 -56.486 52.328
600335 95.703 -56.486 52.278
600336 95.736 -56.491 52.229
600337 95.771 -56.496 52.181
600338 95.806 -56.502 52.136
600339 95.840 -56.508 52.093
600340 95.873 -56.516 52.048
600341 95.899 -56.533 52.005
600342 95.924 -56.541 51.961
600343 95.950 -56.544 51.917
600344 95.975 -56.547 51.877
600345 96.000 -56.550 51.839
600346 96.025 -56.553 51.801
600347 96.050 -56.556 51.765
600348 96.075 -56.559 51.731
600349 96.092 -56.567 51.698
600350 96.109 -56.575 51.667
600351 96.126 -56.583 51.638
600352 96.143 -56.591 51.610
600353 96.160 -56.599 51.582
600354 96.179 -56.607 51.552
600355 96.214 -56.617 51.522
600356 96.248 -56.628 51.494
600357 96.281 -56.639 51.468
600358 96.297 -56.655 51.444
600359 96.312 -56.670 51.420
600360 96.327 -56.685 51.398
600361 96.347 -56.696 51.377
600362 96.375 -56.698 51.361
600363 96.402 -56.699 51.346
600364 96.426 -56.695 51.330
600365 96.449 -56.690 51.315
600366 96.472 -56.685 51.299
600367 96.496 -56.682 51.281
600368 96.519 -56.689 51.265
600369 96.541 -56.695 51.249
600370 96.564 -56.702 51.232
600371 96.587 -56.708 51.212
600372 96.617 -56.702 51.190
600373 96.647 -56.697 51.165
600374 96.674 -56.696 51.142
600375 96.700 -56.696 51.120
600376 96.727 -56.696 51.098
600377 96.754 -56.695 51.076
600378 96.785 -56.705 51.057
600379 96.816 -56.720 51.037
600380 96.846 -56.730 51.021
600381 96.865 -56.723 51.005
600382 96.908 -56.724 50.989
600383 96.959 -56.727 50.974
600384 97.004 -56.728 50.952
600385 97.037 -56.729 50.930
600386 97.071 -56.729 50.907
600387 97.104 -56.730 50.883
600388 97.138 -56.727 50.861
600389 97.171 -56.718 50.842
600390 97.204 -56.710 50.832
600391 97.237 -56.702 50.828
600392 97.270 -56.694 50.823
600393 97.303 -56.686 50.818
600394 97.336 -56.677 50.811
600395 97.370 -56.670 50.802
600396 97.407 -56.670 50.793
600397 97.448 -56.677 50.782
600398 97.489 -56.684 50.771
600399 97.530 -56.691 50.758
600400 97.571 -56.698 50.742
600401 97.612 -56.705 50.724
600402 97.653 -56.711 50.707
600403 97.692 -56.716 50.691
600404 97.726 -56.717 50.676
600405 97.762 -56.724 50.662
600406 97.798 -56.730 50.649
600407 97.832 -56.739 50.638
600408 97.867 -56.747 50.629
600409 97.901 -56.756 50.622
600410 97.935 -56.765 50.615
600411 97.969 -56.774 50.604
600412 98.004 -56.782 50.589
600413 98.038 -56.791 50.574
600414 98.072 -56.799 50.559
600415 98.107 -56.809 50.544
600416 98.141 -56.817 50.522
600417 98.175 -56.826 50.500
600418 98.210 -56.834 50.479
600419 98.244 -56.843 50.458
600420 98.278 -56.852 50.438
600421 98.312 -56.860 50.415
600422 98.347 -56.869 50.387
600423 98.381 -56.877 50.358
600424 98.415 -56.887 50.327
600425 98.449 -56.895 50.294
600426 98.484 -56.904 50.260
600427 98.518 -56.912 50.232
600428 98.553 -56.921 50.205
600429 98.586 -56.928 50.178
600430 98.618 -56.930 50.151
600431 98.649 -56.933 50.124
600432 98.681 -56.935 50.098
600433 98.713 -56.938 50.075
600434 98.744 -56.940 50.050
600435 98.776 -56.943 50.025
600436 98.808 -56.946 50.002
600437 98.843 -56.955 49.980
600438 98.882 -56.975 49.957
600439 98.919 -56.981 49.933
600440 98.958 -56.978 49.911
600441 99.002 -56.978 49.889
600442 99.046 -56.979 49.866
600443 99.090 -56.979 49.841
600444 99.135 -56.980 49.814
600445 99.179 -56.980 49.787
600446 99.200 -56.988 49.758
600447 99.217 -56.999 49.729
600448 99.233 -57.010 49.700
600449 99.251 -57.019 49.671
600450 99.279 -57.024 49.642
600451 99.306 -57.028 49.614
600452 99.334 -57.032 49.589
600453 99.362 -57.037 49.563
600454 99.390 -57.040 49.535
600455 99.417 -57.045 49.506
600456 99.442 -57.051 49.476
600457 99.467 -57.057 49.444
600458 99.492 -57.064 49.414
600459 99.516 -57.070 49.382
600460 99.541 -57.077 49.350
600461 99.566 -57.083 49.318
600462 99.609 -57.113 49.285
600463 99.653 -57.144 49.250
600464 99.656 -57.120 49.215
600465 99.682 -57.138 49.180
600466 99.707 -57.156 49.145
600467 99.733 -57.174 49.110
600468 99.759 -57.191 49.074
600469 99.784 -57.210 49.039
600470 99.809 -57.227 49.006
600471 99.835 -57.245 48.973
600472 99.861 -57.263 48.941
600473 99.887 -57.281 48.910
600474 99.912 -57.298 48.879
600475 99.937 -57.316 48.849
600476 99.963 -57.335 48.821
600477 99.989 -57.352 48.794
600478 100.014 -57.370 48.769
600479 100.040 -57.388 48.745
600809 88.340 -53.043 57.447
600810 88.376 -53.059 57.447
600811 88.411 -53.074 57.447
600812 88.446 -53.089 57.448
600813 88.482 -53.106 57.449
600814 88.521 -53.126 57.450
600815 88.558 -53.145 57.451
600816 88.597 -53.164 57.450
600817 88.635 -53.183 57.451
600818 88.673 -53.203 57.451
600819 88.711 -53.222 57.450
600820 88.749 -53.241 57.448
600821 88.787 -53.261 57.447
600822 88.825 -53.280 57.444
600823 88.863 -53.299 57.442
600824 88.901 -53.318 57.439
600825 88.939 -53.338 57.435
600826 88.977 -53.357 57.430
600827 89.015 -53.377 57.426
600828 89.053 -53.396 57.421
600829 89.091 -53.415 57.415
600830 89.129 -53.434 57.408
600831 89.158 -53.460 57.402
600832 89.190 -53.483 57.395
600833 89.222 -53.505 57.387
600834 89.251 -53.531 57.379
600835 89.273 -53.563 57.370
600836 89.298 -53.594 57.360
600837 89.328 -53.623 57.350
600838 89.358 -53.651 57.339
600839 89.388 -53.679 57.327
600840 89.418 -53.707 57.314
600841 89.468 -53.707 57.301
600842 89.493 -53.719 57.285
600843 89.519 -53.731 57.269
600844 89.551 -53.747 57.253
600845 89.585 -53.765 57.234
600846 89.619 -53.782 57.214
600847 89.654 -53.800 57.192
600848 89.688 -53.817 57.166
600849 89.723 -53.834 57.134
600850 89.757 -53.852 57.101
600851 89.791 -53.870 57.064
600852 89.825 -53.888 57.026
600853 89.860 -53.905 56.986
600854 89.894 -53.922 56.943
600855 89.929 -53.940 56.901
600856 89.963 -53.957 56.858
600857 89.997 -53.975 56.816
600858 90.031 -53.993 56.776
600859 90.066 -54.010 56.736
600860 90.100 -54.028 56.697
600864 90.237 -54.098 56.390
600865 90.272 -54.116 56.354
600866 90.308 -54.132 56.319
600867 90.351 -54.143 56.286
600868 90.394 -54.154 56.254
600869 90.437 -54.166 56.225
600870 90.469 -54.180 56.197
600871 90.487 -54.201 56.170
600872 90.504 -54.221 56.144
600873 90.521 -54.241 56.121
600874 90.538 -54.261 56.100
600875 90.555 -54.281 56.078
600876 90.573 -54.301 56.054
600877 90.590 -54.322 56.028
600878 90.607 -54.342 56.004
600879 90.624 -54.362 55.982
600880 90.642 -54.382 55.962
600881 90.659 -54.402 55.945
600882 90.676 -54.423 55.930
600883 90.693 -54.442 55.915
600884 90.725 -54.449 55.900
600885 90.757 -54.458 55.888
600886 90.787 -54.478 55.874
600887 90.815 -54.499 55.856
600888 90.845 -54.518 55.831
600889 90.873 -54.539 55.804
600890 90.902 -54.558 55.778
600891 90.931 -54.579 55.752
600892 90.960 -54.599 55.729
600893 90.989 -54.619 55.711
600894 91.021 -54.637 55.695
600895 91.059 -54.650 55.682
600896 91.095 -54.664 55.672
600897 91.133 -54.678 55.665
600898 91.169 -54.691 55.663
600899 91.207 -54.705 55.664
600900 91.243 -54.718 55.668
600901 91.280 -54.732 55.671
600902 91.329 -54.739 55.673
600903 91.383 -54.744 55.672
600904 91.436 -54.749 55.669
600905 91.490 -54.754 55.660
600906 91.543 -54.759 55.643
600907 91.597 -54.763 55.621
600908 91.650 -54.768 55.594
600909 91.704 -54.773 55.561
600910 91.746 -54.778 55.527
600911 91.777 -54.781 55.496
600912 91.807 -54.784 55.470
600913 91.837 -54.788 55.450
600914 91.866 -54.795 55.434
600915 91.893 -54.806 55.419
600916 91.920 -54.817 55.406
600917 91.947 -54.828 55.394
600918 91.975 -54.840 55.382
600919 92.002 -54.850 55.372
600920 92.029 -54.862 55.363
600921 92.062 -54.883 55.351
600922 92.095 -54.903 55.339
600923 92.128 -54.924 55.328
600924 92.160 -54.944 55.318
600925 92.193 -54.965 55.311
600926 92.227 -54.985 55.308
600927 92.261 -55.004 55.309
600928 92.295 -55.022 55.316
600929 92.331 -55.027 55.322
600930 92.370 -55.030 55.325
600931 92.409 -55.033 55.327
600932 92.447 -55.037 55.328
600933 92.486 -55.040 55.326
600934 92.524 -55.044 55.323
600935 92.551 -55.054 55.321
600936 92.577 -55.066 55.318
600937 92.603 -55.076 55.317
600938 92.630 -55.088 55.319
600939 92.656 -55.098 55.324
600940 92.683 -55.110 55.329
600941 92.710 -55.120 55.334
600942 92.736 -55.132 55.340
600943 92.762 -55.150 55.347
600944 92.788 -55.167 55.355
600945 92.817 -55.185 55.364
600946 92.853 -55.204 55.373
600947 92.889 -55.222 55.385
600948 92.924 -55.239 55.400
600949 92.960 -55.257 55.411
600950 92.995 -55.276 55.416
600951 93.031 -55.295 55.413
600952 93.069 -55.316 55.405
600953 93.106 -55.337 55.393
600954 93.143 -55.359 55.386
600955 93.180 -55.380 55.379
600956 93.217 -55.402 55.366
600957 93.255 -55.423 55.348
600958 93.292 -55.445 55.327
600959 93.329 -55.467 55.310
600960 93.366 -55.489 55.300
600961 93.404 -55.510 55.292
600962 93.441 -55.531 55.285
600963 93.478 -55.553 55.276
600964 93.514 -55.552 55.266
600965 93.550 -55.550 55.254
600966 93.585 -55.548 55.242
600967 93.621 -55.546 55.229
600968 93.652 -55.551 55.215
600969 93.674 -55.571 55.198
600970 93.695 -55.592 55.179
600971 93.717 -55.613 55.159
600972 93.737 -55.640 55.139
600973 93.757 -55.670 55.117
600974 93.777 -55.699 55.095
600975 93.797 -55.729 55.076
600976 93.817 -55.759 55.059
600977 93.837 -55.788 55.039
600978 93.856 -55.818 55.020
600979 93.876 -55.846 54.998
600980 93.896 -55.876 54.974
600981 93.916 -55.906 54.953
600982 93.936 -55.935 54.929
600983 93.965 -55.947 54.903
600984 93.994 -55.959 54.873
600985 94.023 -55.970 54.835
600986 94.052 -55.981 54.790
600987 94.081 -55.993 54.744
600988 94.113 -56.002 54.697
600989 94.145 -56.009 54.660
600990 94.178 -56.016 54.627
600991 94.210 -56.022 54.590
600992 94.242 -56.029 54.549
600993 94.275 -56.037 54.505
600994 94.309 -56.045 54.461
600995 94.343 -56.053 54.416
600996 94.377 -56.063 54.369
600997 94.411 -56.071 54.318
600998 94.445 -56.079 54.267
600999 94.479 -56.088 54.219
601000 94.519 -56.097 54.170
601001 94.569 -56.109 54.121
601002 94.619 -56.120 54.074
601003 94.653 -56.128 54.025
601004 94.687 -56.136 53.975
601005 94.721 -56.144 53.928
601006 94.755 -56.151 53.882
601007 94.789 -56.159 53.837
601008 94.823 -56.167 53.790
601009 94.857 -56.175 53.739
601010 94.891 -56.182 53.686
601011 94.925 -56.185 53.633
601012 94.959 -56.188 53.579
601013 94.995 -56.201 53.529
601014 95.033 -56.225 53.484
601015 95.072 -56.248 53.440
601016 95.110 -56.272 53.395
601017 95.153 -56.286 53.350
601018 95.182 -56.310 53.305
601019 95.199 -56.323 53.256
601020 95.216 -56.335 53.204
601021 95.234 -56.348 53.148
601022 95.252 -56.361 53.089
601023 95.270 -56.374 53.030
601024 95.310 -56.389 52.970
601025 95.348 -56.400 52.910
601026 95.386 -56.409 52.852
601027 95.423 -56.418 52.794
601028 95.461 -56.427 52.737
601029 95.498 -56.436 52.679
601030 95.536 -56.445 52.616
601031 95.573 -56.454 52.553
601032 95.603 -56.457 52.491
601033 95.629 -56.457 52.431
601034 95.657 -56.458 52.374
601035 95.683 -56.458 52.319
601036 95.711 -56.459 52.267
601037 95.746 -56.464 52.215
601038 95.780 -56.470 52.165
601039 95.815 -56.476 52.117
601040 95.849 -56.482 52.072
601041 95.876 -56.499 52.027
601042 95.902 -56.517 51.981
601043 95.927 -56.520 51.937
601044 95.952 -56.523 51.893
601045 95.977 -56.527 51.851
601046 96.003 -56.530 51.811
601047 96.027 -56.533 51.773
601048 96.052 -56.536 51.737
601049 96.078 -56.539 51.701
601050 96.103 -56.542 51.667
601051 96.128 -56.545 51.635
601052 96.148 -56.552 51.605
601053 96.164 -56.560 51.576
601054 96.181 -56.567 51.546
601055 96.198 -56.576 51.515
601056 96.227 -56.586 51.486
601057 96.261 -56.596 51.457
601058 96.285 -56.609 51.431
601059 96.301 -56.624 51.406
601060 96.316 -56.639 51.384
601061 96.332 -56.655 51.360
601062 96.348 -56.668 51.338
601063 96.376 -56.671 51.321
601064 96.401 -56.668 51.306
601065 96.425 -56.664 51.292
601066 96.448 -56.659 51.281
601067 96.472 -56.655 51.268
601068 96.496 -56.650 51.249
601069 96.518 -56.656 51.226
601070 96.541 -56.662 51.204
601071 96.564 -56.669 51.183
601072 96.586 -56.675 51.164
601073 96.609 -56.682 51.142
601074 96.640 -56.675 51.118
601075 96.671 -56.668 51.096
601076 96.697 -56.668 51.074
601077 96.725 -56.668 51.050
601078 96.751 -56.667 51.025
601079 96.781 -56.674 51.004
601080 96.813 -56.689 50.984
601081 96.838 -56.693 50.967
601082 96.858 -56.686 50.952
601083 96.906 -56.688 50.938
601084 96.957 -56.691 50.923
601085 97.002 -56.693 50.902
601086 97.036 -56.693 50.879
601087 97.069 -56.694 50.857
601088 97.103 -56.694 50.835
601089 97.136 -56.694 50.815
601090 97.169 -56.686 50.796
601091 97.203 -56.678 50.784
601092 97.236 -56.670 50.780
601093 97.268 -56.661 50.777
601094 97.302 -56.654 50.773
601095 97.335 -56.646 50.768
601096 97.369 -56.639 50.762
601097 97.410 -56.646 50.755
601098 97.451 -56.652 50.745
601099 97.492 -56.659 50.737
601100 97.533 -56.666 50.725
601101 97.574 -56.673 50.710
601102 97.615 -56.680 50.693
601103 97.651 -56.683 50.677
601104 97.685 -56.683 50.662
601105 97.720 -56.683 50.648
601106 97.756 -56.689 50.634
601107 97.792 -56.696 50.622
601108 97.828 -56.702 50.612
601109 97.864 -56.708 50.603
601110 97.899 -56.716 50.594
601111 97.933 -56.725 50.586
601112 97.967 -56.733 50.575
601113 98.002 -56.743 50.560
601114 98.036 -56.751 50.545
601115 98.070 -56.760 50.529
601116 98.104 -56.768 50.515
601117 98.139 -56.777 50.496
601118 98.173 -56.786 50.475
601119 98.208 -56.794 50.455
601120 98.242 -56.803 50.434
601121 98.276 -56.812 50.414
601122 98.310 -56.821 50.390
601123 98.345 -56.829 50.363
601124 98.379 -56.838 50.333
601125 98.413 -56.846 50.301
601126 98.447 -56.855 50.268
601127 98.482 -56.864 50.235
601128 98.516 -56.873 50.207
601129 98.550 -56.881 50.179
601130 98.584 -56.889 50.151
601131 98.618 -56.896 50.124
601132 98.652 -56.904 50.098
601133 98.686 -56.912 50.072
601134 98.720 -56.920 50.048
601135 98.754 -56.928 50.024
601136 98.786 -56.930 49.999
601137 98.817 -56.932 49.976
601138 98.849 -56.935 49.955
601139 98.882 -56.940 49.932
601140 98.921 -56.954 49.909
601141 98.961 -56.953 49.884
601142 99.006 -56.953 49.861
601143 99.050 -56.953 49.836
601144 99.094 -56.953 49.811
601145 99.138 -56.954 49.785
601146 99.182 -56.954 49.758
601147 99.213 -56.959 49.729
601148 99.230 -56.970 49.700
601149 99.246 -56.981 49.673
601150 99.268 -56.988 49.646
601151 99.295 -56.993 49.618
601152 99.323 -56.997 49.592
601153 99.351 -57.001 49.568
601154 99.378 -57.006 49.545
601155 99.405 -57.010 49.519
601156 99.430 -57.017 49.490
601157 99.454 -57.023 49.460
601158 99.479 -57.029 49.427
601159 99.504 -57.036 49.395
601160 99.529 -57.042 49.364
601161 99.554 -57.049 49.332
601162 99.581 -57.059 49.300
601163 99.625 -57.090 49.267
601164 99.640 -57.075 49.232
601165 99.667 -57.096 49.199
601166 99.693 -57.114 49.164
601167 99.718 -57.131 49.129
601168 99.744 -57.150 49.095
601169 99.770 -57.168 49.059
601170 99.795 -57.185 49.025
601171 99.821 -57.203 48.993
601172 99.846 -57.221 48.960
601173 99.872 -57.239 48.929
601174 99.897 -57.257 48.898
601175 99.923 -57.275 48.868
601176 99.948 -57.293 48.838
601177 99.974 -57.310 48.810
601178 99.999 -57.329 48.783
601179 100.025 -57.346 48.758
601180 100.051 -57.364 48.734
601510 88.364 -53.008 57.443
601511 88.400 -53.023 57.443
601512 88.435 -53.039 57.443
601513 88.472 -53.056 57.444
601514 88.510 -53.076 57.445
601515 88.548 -53.095 57.446
601516 88.586 -53.114 57.446
601517 88.624 -53.133 57.446
601518 88.662 -53.153 57.446
601519 88.700 -53.172 57.446
601520 88.738 -53.192 57.444
601521 88.776 -53.211 57.444
601522 88.814 -53.230 57.441
601523 88.852 -53.249 57.438
601524 88.890 -53.268 57.435
601525 88.928 -53.288 57.432
601526 88.967 -53.307 57.428
601527 89.004 -53.327 57.423
601528 89.043 -53.346 57.418
601529 89.081 -53.365 57.413
601530 89.111 -53.390 57.406
601531 89.142 -53.413 57.399
601532 89.174 -53.436 57.392
601533 89.206 -53.458 57.385
601534 89.238 -53.481 57.377
601535 89.266 -53.508 57.369
601536 89.288 -53.540 57.359
601537 89.311 -53.572 57.349
601538 89.338 -53.603 57.339
601539 89.368 -53.631 57.329
601540 89.398 -53.659 57.317
601541 89.428 -53.686 57.304
601542 89.481 -53.684 57.291
601543 89.509 -53.693 57.277
601544 89.535 -53.706 57.262
601545 89.560 -53.717 57.246
601546 89.595 -53.734 57.228
601547 89.629 -53.753 57.209
601548 89.663 -53.770 57.189
601549 89.697 -53.787 57.166
601550 89.732 -53.805 57.134
601551 89.766 -53.822 57.100
601552 89.801 -53.840 57.064
601553 89.835 -53.857 57.027
601554 89.869 -53.875 56.988
601555 89.903 -53.893 56.945
601556 89.938 -53.910 56.902
601557 89.972 -53.928 56.859
601558 90.007 -53.945 56.818
601559 90.041 -53.963 56.778
601560 90.075 -53.980 56.738
601561 90.109 -53.998 56.699
601566 90.281 -54.085 56.355
601567 90.323 -54.098 56.319
601568 90.366 -54.109 56.285
601569 90.409 -54.120 56.254
601570 90.453 -54.132 56.223
601571 90.477 -54.149 56.195
601572 90.495 -54.170 56.167
601573 90.512 -54.189 56.140
601574 90.529 -54.210 56.114
601575 90.546 -54.229 56.090
601576 90.563 -54.250 56.067
601577 90.581 -54.270 56.044
601578 90.598 -54.290 56.021
601579 90.615 -54.311 55.996
601580 90.632 -54.330 55.976
601581 90.649 -54.351 55.956
601582 90.666 -54.370 55.938
601583 90.684 -54.391 55.921
601584 90.712 -54.400 55.906
601585 90.748 -54.402 55.893
601586 90.777 -54.421 55.879
601587 90.806 -54.442 55.862
601588 90.835 -54.461 55.842
601589 90.864 -54.482 55.816
601590 90.893 -54.502 55.788
601591 90.922 -54.522 55.762
601592 90.950 -54.543 55.738
601593 90.979 -54.562 55.717
601594 91.008 -54.583 55.700
601595 91.037 -54.603 55.685
601596 91.068 -54.622 55.672
601597 91.104 -54.636 55.664
601598 91.142 -54.649 55.659
601599 91.178 -54.663 55.658
601600 91.215 -54.677 55.660
601601 91.252 -54.690 55.662
601602 91.293 -54.702 55.664
601603 91.347 -54.706 55.666
601604 91.400 -54.712 55.664
601605 91.453 -54.716 55.661
601606 91.507 -54.721 55.650
601607 91.561 -54.726 55.631
601608 91.614 -54.731 55.606
601609 91.667 -54.736 55.578
601610 91.721 -54.740 55.548
601611 91.764 -54.745 55.517
601612 91.794 -54.749 55.487
601613 91.824 -54.752 55.462
601614 91.854 -54.756 55.442
601615 91.881 -54.768 55.427
601616 91.909 -54.779 55.413
601617 91.936 -54.790 55.400
601618 91.963 -54.801 55.388
601619 91.991 -54.812 55.377
601620 92.018 -54.823 55.368
601621 92.048 -54.834 55.360
601622 92.081 -54.849 55.350
601623 92.115 -54.869 55.337
601624 92.149 -54.888 55.326
601625 92.183 -54.908 55.316
601626 92.217 -54.928 55.308
601627 92.251 -54.947 55.304
601628 92.286 -54.966 55.303
601629 92.319 -54.986 55.306
601630 92.351 -54.994 55.309
601631 92.384 -55.000 55.312
601632 92.423 -55.003 55.314
601633 92.461 -55.006 55.315
601634 92.500 -55.010 55.314
601635 92.539 -55.013 55.313
601636 92.570 -55.021 55.313
601637 92.596 -55.032 55.315
601638 92.623 -55.042 55.320
601639 92.650 -55.054 55.328
601640 92.676 -55.064 55.340
601641 92.703 -55.075 55.349
601642 92.729 -55.087 55.357
601643 92.755 -55.105 55.367
601644 92.781 -55.123 55.379
601645 92.807 -55.141 55.390
601646 92.834 -55.160 55.399
601647 92.869 -55.177 55.415
601648 92.905 -55.195 55.432
601649 92.940 -55.213 55.447
601650 92.976 -55.231 55.458
601651 93.011 -55.249 55.460
601652 93.047 -55.267 55.454
601653 93.082 -55.285 55.439
601654 93.118 -55.303 55.419
601655 93.154 -55.321 55.399
601656 93.190 -55.342 55.381
601657 93.227 -55.363 55.363
601658 93.265 -55.385 55.345
601659 93.302 -55.406 55.328
601660 93.339 -55.428 55.317
601661 93.376 -55.449 55.309
601662 93.414 -55.471 55.301
601663 93.451 -55.492 55.292
601664 93.488 -55.514 55.280
601665 93.524 -55.523 55.268
601666 93.560 -55.520 55.254
601667 93.596 -55.518 55.240
601668 93.630 -55.518 55.227
601669 93.652 -55.539 55.213
601670 93.674 -55.560 55.197
601671 93.695 -55.580 55.180
601672 93.717 -55.601 55.162
601673 93.739 -55.621 55.142
601674 93.760 -55.643 55.119
601675 93.780 -55.672 55.095
601676 93.799 -55.702 55.072
601677 93.820 -55.731 55.050
601678 93.839 -55.760 55.028
601679 93.859 -55.790 55.007
601680 93.879 -55.819 54.985
601681 93.899 -55.849 54.961
601682 93.919 -55.878 54.938
601683 93.939 -55.908 54.915
601684 93.966 -55.924 54.888
601685 93.995 -55.935 54.857
601686 94.024 -55.947 54.820
601687 94.053 -55.958 54.777
601688 94.082 -55.969 54.735
601689 94.115 -55.976 54.694
601690 94.147 -55.983 54.659
601691 94.180 -55.990 54.627
601692 94.212 -55.997 54.591
601693 94.245 -56.003 54.549
601694 94.277 -56.010 54.504
601695 94.311 -56.019 54.459
601696 94.345 -56.028 54.412
601697 94.379 -56.036 54.363
601698 94.413 -56.044 54.313
601699 94.447 -56.053 54.266
601700 94.481 -56.061 54.222
601701 94.526 -56.072 54.175
601702 94.576 -56.083 54.128
601703 94.626 -56.095 54.083
601704 94.663 -56.103 54.037
601705 94.697 -56.111 53.991
601706 94.731 -56.119 53.946
601707 94.765 -56.127 53.898
601708 94.799 -56.135 53.852
601709 94.833 -56.142 53.804
601710 94.867 -56.150 53.754
601711 94.901 -56.154 53.701
601712 94.934 -56.158 53.647
601713 94.969 -56.161 53.592
601714 95.004 -56.172 53.538
601715 95.043 -56.196 53.489
601716 95.081 -56.219 53.442
601717 95.120 -56.243 53.395
601718 95.158 -56.266 53.351
601719 95.186 -56.285 53.305
601720 95.204 -56.298 53.257
601721 95.221 -56.310 53.207
601722 95.239 -56.323 53.154
601723 95.256 -56.335 53.097
601724 95.277 -56.349 53.038
601725 95.316 -56.361 52.977
601726 95.353 -56.370 52.917
601727 95.391 -56.379 52.858
601728 95.428 -56.389 52.800
601729 95.466 -56.398 52.743
601730 95.503 -56.407 52.686
601731 95.541 -56.416 52.625
601732 95.579 -56.425 52.558
601733 95.610 -56.429 52.493
601734 95.636 -56.429 52.429
601735 95.663 -56.429 52.369
601736 95.690 -56.430 52.312
601737 95.721 -56.432 52.257
601738 95.755 -56.439 52.203
601739 95.790 -56.444 52.150
601740 95.825 -56.449 52.101
601741 95.852 -56.465 52.053
601742 95.878 -56.483 52.007
601743 95.904 -56.498 51.960
601744 95.930 -56.501 51.914
601745 95.955 -56.504 51.869
601746 95.979 -56.507 51.827
601747 96.005 -56.510 51.786
601748 96.030 -56.513 51.747
601749 96.055 -56.517 51.710
601750 96.080 -56.520 51.672
601751 96.105 -56.523 51.638
601752 96.130 -56.526 51.605
601753 96.156 -56.529 51.574
601754 96.181 -56.532 51.544
601755 96.203 -56.537 51.513
601756 96.220 -56.545 51.482
601757 96.239 -56.553 51.451
601758 96.273 -56.564 51.423
601759 96.289 -56.579 51.397
601760 96.305 -56.594 51.373
601761 96.320 -56.609 51.350
601762 96.336 -56.624 51.325
601763 96.351 -56.639 51.303
601764 96.377 -56.642 51.284
601765 96.401 -56.638 51.269
601766 96.424 -56.633 51.256
601767 96.448 -56.628 51.245
601768 96.471 -56.624 51.231
601769 96.495 -56.619 51.211
601770 96.518 -56.624 51.191
601771 96.540 -56.630 51.170
601772 96.563 -56.636 51.147
601773 96.586 -56.642 51.119
601774 96.608 -56.649 51.093
601775 96.631 -56.655 51.069
601776 96.663 -56.648 51.046
601777 96.695 -56.639 51.023
601778 96.722 -56.639 50.999
601779 96.748 -56.639 50.975
601780 96.777 -56.643 50.953
601781 96.809 -56.658 50.932
601782 96.831 -56.656 50.915
601783 96.853 -56.650 50.899
601784 96.904 -56.653 50.883
601785 96.955 -56.656 50.867
601786 97.001 -56.658 50.849
601787 97.034 -56.658 50.828
601788 97.068 -56.658 50.807
601789 97.101 -56.659 50.787
601790 97.135 -56.659 50.769
601791 97.168 -56.654 50.751
601792 97.201 -56.646 50.737
601793 97.234 -56.638 50.731
601794 97.267 -56.630 50.729
601795 97.300 -56.621 50.726
601796 97.333 -56.614 50.724
601797 97.372 -56.614 50.722
601798 97.413 -56.621 50.715
601799 97.453 -56.628 50.707
601800 97.495 -56.635 50.700
601801 97.535 -56.642 50.690
601802 97.576 -56.649 50.677
601803 97.611 -56.649 50.662
601804 97.645 -56.649 50.647
601805 97.679 -56.650 50.632
601806 97.714 -56.650 50.619
601807 97.749 -56.656 50.606
601808 97.786 -56.662 50.594
601809 97.822 -56.668 50.585
601810 97.858 -56.674 50.577
601811 97.894 -56.680 50.566
601812 97.930 -56.686 50.556
601813 97.965 -56.693 50.543
601814 98.000 -56.702 50.529
601815 98.034 -56.711 50.514
601816 98.068 -56.720 50.498
601817 98.102 -56.728 50.484
601818 98.137 -56.737 50.466
601819 98.171 -56.746 50.447
601820 98.206 -56.755 50.428
601821 98.240 -56.763 50.407
601822 98.274 -56.772 50.388
601823 98.308 -56.780 50.363
601824 98.343 -56.790 50.335
601825 98.377 -56.798 50.305
601826 98.411 -56.806 50.274
601827 98.445 -56.815 50.241
601828 98.480 -56.824 50.208
601829 98.514 -56.833 50.180
601830 98.548 -56.840 50.152
601831 98.582 -56.849 50.124
601832 98.616 -56.856 50.098
601833 98.650 -56.864 50.071
601834 98.684 -56.871 50.045
601835 98.718 -56.880 50.020
601836 98.752 -56.887 49.995
601837 98.786 -56.895 49.971
601838 98.820 -56.903 49.948
601839 98.854 -56.911 49.927
601840 98.888 -56.918 49.906
601841 98.922 -56.926 49.882
601842 98.966 -56.928 49.858
601843 99.009 -56.928 49.833
601844 99.054 -56.928 49.807
601845 99.097 -56.928 49.781
601846 99.142 -56.928 49.755
601847 99.186 -56.929 49.727
601848 99.227 -56.931 49.699
601849 99.243 -56.941 49.671
601850 99.259 -56.952 49.645
601851 99.284 -56.957 49.619
601852 99.312 -56.962 49.593
601853 99.340 -56.965 49.567
601854 99.367 -56.970 49.544
601855 99.393 -56.976 49.522
601856 99.418 -56.982 49.498
601857 99.442 -56.989 49.470
601858 99.467 -56.995 49.440
601859 99.492 -57.001 49.407
601860 99.516 -57.008 49.375
601861 99.541 -57.014 49.344
601862 99.566 -57.020 49.312
601863 99.597 -57.035 49.280
601864 99.642 -57.066 49.247
601865 99.653 -57.054 49.213
601866 99.678 -57.072 49.181
601867 99.704 -57.090 49.147
601868 99.729 -57.108 49.113
601869 99.755 -57.126 49.079
601870 99.780 -57.144 49.044
601871 99.806 -57.162 49.010
601872 99.832 -57.179 48.979
601873 99.857 -57.197 48.947
601874 99.883 -57.215 48.916
601875 99.908 -57.233 48.885
601876 99.934 -57.250 48.855
601877 99.960 -57.269 48.827
601878 99.985 -57.287 48.799
601879 100.010 -57.304 48.772
601880 100.036 -57.322 48.747
602210 88.354 -52.958 57.438
602211 88.389 -52.973 57.438
602212 88.425 -52.988 57.438
602213 88.462 -53.006 57.438
602214 88.499 -53.026 57.439
602215 88.538 -53.045 57.440
602216 88.575 -53.064 57.441
602217 88.614 -53.083 57.441
602218 88.651 -53.103 57.441
602219 88.690 -53.122 57.441
602220 88.728 -53.142 57.440
602221 88.766 -53.161 57.439
602222 88.804 -53.180 57.438
602223 88.842 -53.199 57.436
602224 88.880 -53.218 57.432
602225 88.918 -53.238 57.428
602226 88.956 -53.257 57.424
602227 88.994 -53.277 57.420
602228 89.032 -53.295 57.415
602229 89.070 -53.315 57.410
602230 89.095 -53.344 57.404
602231 89.127 -53.367 57.398
602232 89.159 -53.389 57.391
602233 89.190 -53.412 57.383
602234 89.222 -53.435 57.375
602235 89.254 -53.458 57.367
602236 89.281 -53.485 57.359
602237 89.304 -53.518 57.349
602238 89.326 -53.549 57.339
602239 89.349 -53.582 57.329
602240 89.378 -53.611 57.318
602241 89.408 -53.639 57.307
602242 89.441 -53.663 57.295
602243 89.493 -53.660 57.283
602244 89.526 -53.668 57.269
602245 89.551 -53.680 57.254
602246 89.577 -53.692 57.239
602247 89.604 -53.705 57.222
602248 89.638 -53.722 57.206
602249 89.672 -53.740 57.188
602250 89.707 -53.757 57.167
602251 89.741 -53.775 57.134
602252 89.775 -53.793 57.100
602253 89.810 -53.810 57.064
602254 89.844 -53.828 57.028
602255 89.878 -53.845 56.989
602256 89.913 -53.863 56.946
602257 89.947 -53.880 56.904
602258 89.981 -53.898 56.862
602259 90.016 -53.916 56.821
602260 90.050 -53.933 56.781
602261 90.085 -53.950 56.741
602262 90.119 -53.968 56.700
602267 90.295 -54.053 56.356
602268 90.338 -54.063 56.320
602269 90.382 -54.075 56.285
602270 90.425 -54.086 56.253
602271 90.468 -54.098 56.222
602272 90.485 -54.117 56.193
602273 90.502 -54.138 56.164
602274 90.519 -54.158 56.136
602275 90.536 -54.178 56.109
602276 90.553 -54.198 56.083
602277 90.571 -54.218 56.060
602278 90.588 -54.239 56.037
602279 90.605 -54.259 56.014
602280 90.622 -54.279 55.991
602281 90.640 -54.299 55.969
602282 90.657 -54.319 55.950
602283 90.674 -54.339 55.933
602284 90.699 -54.352 55.916
602285 90.735 -54.353 55.902
602286 90.767 -54.364 55.886
602287 90.796 -54.385 55.869
602288 90.825 -54.405 55.849
602289 90.854 -54.425 55.825
602290 90.883 -54.446 55.799
602291 90.912 -54.465 55.772
602292 90.941 -54.486 55.747
602293 90.970 -54.506 55.725
602294 90.999 -54.526 55.707
602295 91.028 -54.546 55.690
602296 91.057 -54.566 55.676
602297 91.086 -54.587 55.665
602298 91.114 -54.607 55.657
602299 91.151 -54.621 55.655
602300 91.187 -54.635 55.656
602301 91.224 -54.649 55.657
602302 91.261 -54.662 55.658
602303 91.310 -54.669 55.658
602304 91.364 -54.674 55.659
602305 91.417 -54.679 55.659
602306 91.470 -54.684 55.653
602307 91.524 -54.689 55.639
602308 91.578 -54.693 55.617
602309 91.631 -54.699 55.590
602310 91.684 -54.703 55.564
602311 91.738 -54.709 55.536
602312 91.782 -54.712 55.506
602313 91.812 -54.716 55.478
602314 91.842 -54.720 55.454
602315 91.870 -54.729 55.435
602316 91.897 -54.740 55.419
602317 91.925 -54.751 55.406
602318 91.952 -54.762 55.395
602319 91.979 -54.774 55.383
602320 92.009 -54.784 55.372
602321 92.041 -54.795 55.363
602322 92.074 -54.806 55.356
602323 92.106 -54.816 55.346
602324 92.139 -54.831 55.335
602325 92.173 -54.850 55.323
602326 92.208 -54.870 55.313
602327 92.241 -54.890 55.305
602328 92.276 -54.909 55.300
602329 92.310 -54.929 55.298
602330 92.344 -54.949 55.299
602331 92.376 -54.960 55.301
602332 92.408 -54.966 55.303
602333 92.439 -54.972 55.305
602334 92.476 -54.975 55.307
602335 92.514 -54.979 55.307
602336 92.553 -54.982 55.307
602337 92.590 -54.987 55.311
602338 92.616 -54.997 55.316
602339 92.643 -55.009 55.326
602340 92.669 -55.019 55.343
602341 92.696 -55.031 55.361
602342 92.722 -55.042 55.375
602343 92.748 -55.060 55.387
602344 92.774 -55.079 55.402
602345 92.800 -55.097 55.419
602346 92.826 -55.115 55.434
602347 92.852 -55.133 55.442
602348 92.885 -55.151 55.457
602349 92.920 -55.169 55.473
602350 92.956 -55.187 55.487
602351 92.992 -55.205 55.494
602352 93.027 -55.223 55.493
602353 93.063 -55.241 55.486
602354 93.098 -55.259 55.471
602355 93.134 -55.277 55.453
602356 93.169 -55.295 55.432
602357 93.205 -55.313 55.411
602358 93.240 -55.331 55.389
602359 93.276 -55.348 55.364
602360 93.312 -55.367 55.341
602361 93.349 -55.389 55.323
602362 93.387 -55.411 55.311
602363 93.424 -55.432 55.305
602364 93.461 -55.453 55.298
602365 93.498 -55.475 55.288
602366 93.535 -55.494 55.274
602367 93.571 -55.492 55.258
602368 93.607 -55.489 55.243
602369 93.630 -55.507 55.228
602370 93.652 -55.527 55.213
602371 93.674 -55.548 55.198
602372 93.695 -55.568 55.183
602373 93.716 -55.589 55.168
602374 93.738 -55.609 55.150
602375 93.760 -55.630 55.129
602376 93.781 -55.650 55.102
602377 93.802 -55.674 55.075
602378 93.822 -55.703 55.048
602379 93.842 -55.733 55.023
602380 93.862 -55.762 54.999
602381 93.882 -55.792 54.974
602382 93.902 -55.821 54.948
602383 93.922 -55.851 54.922
602384 93.941 -55.880 54.896
602385 93.966 -55.900 54.866
602386 93.995 -55.912 54.835
602387 94.025 -55.923 54.801
602388 94.054 -55.934 54.762
602389 94.085 -55.943 54.722
602390 94.117 -55.950 54.685
602391 94.150 -55.957 54.651
602392 94.182 -55.964 54.618
602393 94.215 -55.971 54.582
602394 94.247 -55.978 54.543
602395 94.280 -55.984 54.500
602396 94.313 -55.993 54.457
602397 94.347 -56.001 54.411
602398 94.381 -56.009 54.363
602399 94.415 -56.018 54.316
602400 94.449 -56.026 54.272
602401 94.483 -56.034 54.229
602402 94.532 -56.047 54.183
602403 94.583 -56.058 54.140
602404 94.633 -56.069 54.098
602405 94.673 -56.078 54.055
602406 94.707 -56.086 54.011
602407 94.741 -56.094 53.966
602408 94.775 -56.102 53.916
602409 94.809 -56.110 53.868
602410 94.843 -56.118 53.820
602411 94.877 -56.125 53.769
602412 94.911 -56.128 53.716
602413 94.945 -56.131 53.662
602414 94.978 -56.135 53.605
602415 95.014 -56.143 53.550
602416 95.052 -56.166 53.498
602417 95.091 -56.190 53.447
602418 95.129 -56.213 53.398
602419 95.168 -56.237 53.352
602420 95.188 -56.248 53.305
602421 95.206 -56.263 53.258
602422 95.225 -56.278 53.209
602423 95.243 -56.294 53.157
602424 95.261 -56.309 53.102
602425 95.283 -56.323 53.045
602426 95.321 -56.332 52.984
602427 95.358 -56.341 52.924
602428 95.396 -56.349 52.864
602429 95.433 -56.358 52.805
602430 95.471 -56.367 52.746
602431 95.509 -56.376 52.688
602432 95.546 -56.385 52.627
602433 95.584 -56.395 52.564
602434 95.616 -56.401 52.497
602435 95.643 -56.401 52.431
602436 95.670 -56.401 52.370
602437 95.697 -56.402 52.309
602438 95.730 -56.407 52.250
602439 95.765 -56.412 52.195
602440 95.800 -56.417 52.140
602441 95.829 -56.431 52.088
602442 95.855 -56.448 52.038
602443 95.881 -56.466 51.989
602444 95.906 -56.478 51.942
602445 95.932 -56.481 51.894
602446 95.957 -56.484 51.848
602447 95.982 -56.487 51.804
602448 96.007 -56.490 51.763
602449 96.032 -56.493 51.722
602450 96.057 -56.496 51.683
602451 96.083 -56.499 51.645
602452 96.108 -56.503 51.609
602453 96.133 -56.506 51.576
602454 96.158 -56.509 51.544
602455 96.183 -56.512 51.513
602456 96.208 -56.515 51.482
602457 96.233 -56.518 51.451
602458 96.258 -56.521 51.420
602459 96.278 -56.533 51.391
602460 96.293 -56.548 51.366
602461 96.309 -56.564 51.341
602462 96.324 -56.579 51.316
602463 96.340 -56.594 51.292
602464 96.356 -56.609 51.269
602465 96.377 -56.611 51.250
602466 96.400 -56.607 51.234
602467 96.424 -56.602 51.219
602468 96.447 -56.598 51.205
602469 96.470 -56.592 51.190
602470 96.494 -56.588 51.170
602471 96.517 -56.591 51.151
602472 96.540 -56.597 51.132
602473 96.562 -56.604 51.108
602474 96.585 -56.610 51.081
602475 96.608 -56.617 51.054
602476 96.630 -56.623 51.026
602477 96.653 -56.630 51.001
602478 96.686 -56.620 50.976
602479 96.719 -56.611 50.951
602480 96.745 -56.611 50.928
602481 96.773 -56.613 50.906
602482 96.804 -56.626 50.883
602483 96.823 -56.619 50.864
602484 96.851 -56.614 50.847
602485 96.902 -56.617 50.831
602486 96.953 -56.620 50.814
602487 96.999 -56.623 50.798
602488 97.033 -56.623 50.778
602489 97.066 -56.624 50.759
602490 97.100 -56.624 50.741
602491 97.133 -56.624 50.723
602492 97.166 -56.622 50.707
602493 97.199 -56.614 50.692
602494 97.233 -56.605 50.684
602495 97.266 -56.598 50.679
602496 97.299 -56.589 50.678
602497 97.333 -56.583 50.676
602498 97.374 -56.590 50.673
602499 97.415 -56.597 50.667
602500 97.456 -56.604 50.662
602501 97.497 -56.611 50.658
602502 97.536 -56.615 50.651
602503 97.570 -56.616 50.641
602504 97.604 -56.616 50.629
602505 97.639 -56.617 50.616
602506 97.673 -56.617 50.602
602507 97.708 -56.617 50.589
602508 97.743 -56.622 50.576
602509 97.779 -56.628 50.563
602510 97.815 -56.634 50.552
602511 97.852 -56.640 50.546
602512 97.887 -56.646 50.537
602513 97.924 -56.653 50.525
602514 97.960 -56.659 50.511
602515 97.996 -56.665 50.497
602516 98.032 -56.671 50.482
602517 98.066 -56.680 50.467
602518 98.100 -56.689 50.451
602519 98.135 -56.697 50.434
602520 98.169 -56.706 50.416
602521 98.203 -56.714 50.398
602522 98.237 -56.724 50.378
602523 98.272 -56.732 50.357
602524 98.306 -56.740 50.332
602525 98.341 -56.749 50.305
602526 98.375 -56.758 50.276
602527 98.409 -56.767 50.245
602528 98.443 -56.775 50.212
602529 98.478 -56.784 50.179
602530 98.512 -56.793 50.152
602531 98.546 -56.800 50.124
602532 98.580 -56.808 50.097
602533 98.614 -56.816 50.071
602534 98.648 -56.824 50.044
602535 98.682 -56.831 50.018
602536 98.716 -56.839 49.991
602537 98.750 -56.847 49.965
602538 98.784 -56.855 49.941
602539 98.818 -56.862 49.919
602540 98.854 -56.870 49.898
602541 98.895 -56.877 49.877
602542 98.934 -56.884 49.855
602543 98.975 -56.892 49.831
602544 99.014 -56.899 49.805
602545 99.057 -56.903 49.777
602546 99.101 -56.903 49.751
602547 99.145 -56.903 49.724
602548 99.190 -56.903 49.696
602549 99.233 -56.903 49.669
602550 99.256 -56.912 49.642
602551 99.274 -56.921 49.617
602552 99.301 -56.926 49.592
602553 99.329 -56.931 49.567
602554 99.356 -56.935 49.542
602555 99.380 -56.941 49.519
602556 99.405 -56.948 49.496
602557 99.430 -56.954 49.472
602558 99.455 -56.960 49.445
602559 99.480 -56.967 49.417
602560 99.504 -56.973 49.385
602561 99.529 -56.980 49.354
602562 99.554 -56.986 49.323
602563 99.578 -56.992 49.291
602564 99.614 -57.012 49.258
602565 99.637 -57.012 49.226
602566 99.663 -57.031 49.193
602567 99.689 -57.048 49.161
602568 99.715 -57.066 49.129
602569 99.740 -57.084 49.096
602570 99.766 -57.102 49.062
602571 99.791 -57.120 49.029
602572 99.817 -57.138 48.995
602573 99.843 -57.156 48.964
602574 99.868 -57.173 48.933
602575 99.894 -57.191 48.901
602576 99.919 -57.209 48.871
602577 99.945 -57.227 48.842
602578 99.970 -57.244 48.814
602579 99.996 -57.263 48.788
602580 100.021 -57.281 48.762
602581 100.047 -57.298 48.737
602911 88.379 -52.923 57.433
602912 88.414 -52.938 57.433
602913 88.451 -52.956 57.433
602914 88.489 -52.976 57.433
602915 88.527 -52.995 57.435
602916 88.565 -53.014 57.435
602917 88.603 -53.033 57.436
602918 88.641 -53.053 57.436
602919 88.679 -53.072 57.436
602920 88.717 -53.092 57.436
602921 88.755 -53.111 57.435
602922 88.793 -53.130 57.434
602923 88.831 -53.149 57.432
602924 88.870 -53.168 57.430
602925 88.907 -53.188 57.426
602926 88.946 -53.207 57.422
602927 88.984 -53.227 57.418
602928 89.022 -53.245 57.413
602929 89.048 -53.274 57.407
602930 89.080 -53.297 57.402
602931 89.112 -53.320 57.397
602932 89.143 -53.342 57.390
602933 89.175 -53.365 57.383
602934 89.206 -53.388 57.375
602935 89.238 -53.411 57.366
602936 89.270 -53.433 57.357
602937 89.296 -53.462 57.349
602938 89.319 -53.494 57.340
602939 89.341 -53.527 57.330
602940 89.364 -53.559 57.320
602941 89.388 -53.591 57.309
602942 89.417 -53.619 57.299
602943 89.453 -53.640 57.288
602944 89.506 -53.637 57.276
602945 89.542 -53.643 57.263
602946 89.567 -53.655 57.248
602947 89.593 -53.667 57.233
602948 89.618 -53.678 57.218
602949 89.648 -53.693 57.202
602950 89.682 -53.710 57.184
602951 89.716 -53.728 57.164
602952 89.750 -53.745 57.133
602953 89.785 -53.762 57.100
602954 89.819 -53.781 57.065
602955 89.853 -53.798 57.029
602956 89.888 -53.816 56.990
602957 89.922 -53.833 56.948
602958 89.956 -53.850 56.906
602959 89.990 -53.868 56.864
602960 90.025 -53.885 56.823
602961 90.059 -53.903 56.783
602962 90.094 -53.921 56.744
602963 90.128 -53.938 56.703
602964 90.162 -53.956 56.665
602969 90.354 -54.029 56.322
602970 90.397 -54.041 56.286
602971 90.440 -54.052 56.253
602972 90.475 -54.066 56.221
602973 90.493 -54.086 56.192
602974 90.510 -54.107 56.163
602975 90.527 -54.126 56.133
602976 90.544 -54.147 56.105
602977 90.561 -54.167 56.079
602978 90.579 -54.187 56.054
602979 90.596 -54.207 56.031
602980 90.613 -54.227 56.008
602981 90.630 -54.248 55.985
602982 90.647 -54.267 55.964
602983 90.665 -54.288 55.945
602984 90.687 -54.303 55.928
602985 90.723 -54.305 55.913
602986 90.757 -54.308 55.898
602987 90.787 -54.328 55.880
602988 90.815 -54.348 55.857
602989 90.845 -54.368 55.833
602990 90.873 -54.389 55.808
602991 90.902 -54.409 55.783
602992 90.931 -54.429 55.758
602993 90.960 -54.449 55.735
602994 90.989 -54.469 55.714
602995 91.018 -54.489 55.696
602996 91.047 -54.510 55.681
602997 91.076 -54.530 55.668
602998 91.105 -54.550 55.658
602999 91.134 -54.570 55.652
603000 91.163 -54.590 55.652
603001 91.196 -54.606 55.654
603002 91.233 -54.620 55.655
603003 91.274 -54.632 55.655
603004 91.327 -54.637 55.654
603005 91.381 -54.642 55.654
603006 91.434 -54.646 55.653
603007 91.488 -54.652 55.643
603008 91.541 -54.656 55.625
603009 91.594 -54.661 55.601
603010 91.648 -54.666 55.574
603011 91.702 -54.671 55.550
603012 91.755 -54.676 55.523
603013 91.800 -54.681 55.495
603014 91.830 -54.684 55.469
603015 91.859 -54.690 55.446
603016 91.886 -54.702 55.427
603017 91.913 -54.713 55.412
603018 91.941 -54.724 55.400
603019 91.971 -54.734 55.389
603020 92.003 -54.745 55.378
603021 92.035 -54.756 55.367
603022 92.068 -54.766 55.358
603023 92.100 -54.777 55.352
603024 92.132 -54.787 55.342
603025 92.164 -54.798 55.331
603026 92.198 -54.812 55.321
603027 92.232 -54.832 55.310
603028 92.266 -54.852 55.302
603029 92.300 -54.872 55.296
603030 92.334 -54.891 55.294
603031 92.368 -54.911 55.293
603032 92.401 -54.925 55.294
603033 92.433 -54.931 55.296
603034 92.464 -54.937 55.298
603035 92.496 -54.943 55.300
603036 92.528 -54.949 55.303
603037 92.567 -54.952 55.307
603038 92.605 -54.955 55.314
603039 92.636 -54.964 55.324
603040 92.662 -54.975 55.338
603041 92.689 -54.985 55.359
603042 92.715 -54.998 55.381
603043 92.741 -55.016 55.403
603044 92.767 -55.034 55.423
603045 92.793 -55.052 55.445
603046 92.819 -55.070 55.467
603047 92.845 -55.088 55.485
603048 92.871 -55.107 55.492
603049 92.901 -55.125 55.498
603050 92.936 -55.143 55.509
603051 92.972 -55.160 55.520
603052 93.007 -55.179 55.524
603053 93.043 -55.197 55.520
603054 93.079 -55.214 55.513
603055 93.114 -55.232 55.501
603056 93.150 -55.251 55.484
603057 93.185 -55.269 55.464
603058 93.221 -55.286 55.441
603059 93.257 -55.304 55.416
603060 93.292 -55.323 55.387
603061 93.328 -55.340 55.359
603062 93.363 -55.358 55.334
603063 93.399 -55.376 55.316
603064 93.434 -55.394 55.307
603065 93.471 -55.414 55.301
603066 93.508 -55.436 55.292
603067 93.545 -55.458 55.281
603068 93.581 -55.462 55.271
603069 93.609 -55.474 55.259
603070 93.630 -55.495 55.246
603071 93.651 -55.515 55.231
603072 93.673 -55.536 55.216
603073 93.695 -55.556 55.200
603074 93.716 -55.577 55.187
603075 93.738 -55.597 55.172
603076 93.760 -55.618 55.151
603077 93.781 -55.639 55.121
603078 93.803 -55.659 55.089
603079 93.825 -55.680 55.056
603080 93.845 -55.705 55.024
603081 93.865 -55.735 54.994
603082 93.885 -55.765 54.966
603083 93.905 -55.794 54.936
603084 93.924 -55.823 54.906
603085 93.944 -55.853 54.877
603086 93.967 -55.877 54.846
603087 93.996 -55.888 54.815
603088 94.025 -55.900 54.782
603089 94.055 -55.910 54.746
603090 94.087 -55.917 54.709
603091 94.120 -55.925 54.675
603092 94.152 -55.931 54.641
603093 94.185 -55.938 54.608
603094 94.217 -55.945 54.574
603095 94.250 -55.952 54.537
603096 94.282 -55.959 54.497
603097 94.315 -55.966 54.456
603098 94.349 -55.975 54.410
603099 94.383 -55.983 54.363
603100 94.417 -55.991 54.318
603101 94.451 -56.000 54.277
603102 94.489 -56.009 54.234
603103 94.539 -56.021 54.194
603104 94.590 -56.032 54.154
603105 94.640 -56.044 54.115
603106 94.683 -56.054 54.076
603107 94.717 -56.062 54.035
603108 94.751 -56.069 53.990
603109 94.785 -56.077 53.939
603110 94.819 -56.085 53.888
603111 94.853 -56.093 53.837
603112 94.887 -56.097 53.785
603113 94.920 -56.101 53.730
603114 94.955 -56.104 53.676
603115 94.988 -56.107 53.619
603116 95.023 -56.114 53.563
603117 95.061 -56.138 53.509
603118 95.100 -56.161 53.456
603119 95.138 -56.185 53.404
603120 95.171 -56.204 53.354
603121 95.189 -56.213 53.306
603122 95.208 -56.224 53.258
603123 95.226 -56.239 53.209
603124 95.244 -56.255 53.159
603125 95.262 -56.270 53.106
603126 95.293 -56.285 53.050
603127 95.326 -56.301 52.990
603128 95.363 -56.310 52.929
603129 95.401 -56.320 52.869
603130 95.438 -56.329 52.809
603131 95.476 -56.338 52.750
603132 95.513 -56.347 52.691
603133 95.551 -56.356 52.631
603134 95.588 -56.365 52.570
603135 95.623 -56.373 52.504
603136 95.651 -56.373 52.437
603137 95.677 -56.373 52.372
603138 95.705 -56.375 52.310
603139 95.740 -56.380 52.249
603140 95.774 -56.385 52.191
603141 95.806 -56.397 52.134
603142 95.831 -56.414 52.080
603143 95.858 -56.432 52.029
603144 95.884 -56.450 51.977
603145 95.909 -56.458 51.927
603146 95.934 -56.461 51.879
603147 95.959 -56.464 51.831
603148 95.984 -56.467 51.785
603149 96.010 -56.470 51.742
603150 96.035 -56.473 51.701
603151 96.060 -56.476 51.660
603152 96.085 -56.479 51.621
603153 96.110 -56.483 51.584
603154 96.135 -56.486 51.550
603155 96.161 -56.489 51.516
603156 96.185 -56.492 51.484
603157 96.210 -56.495 51.453
603158 96.236 -56.498 51.422
603159 96.267 -56.500 51.392
603160 96.287 -56.501 51.364
603161 96.302 -56.503 51.337
603162 96.318 -56.511 51.311
603163 96.333 -56.526 51.285
603164 96.349 -56.541 51.260
603165 96.365 -56.555 51.237
603166 96.380 -56.570 51.218
603167 96.399 -56.576 51.200
603168 96.423 -56.571 51.183
603169 96.446 -56.567 51.164
603170 96.470 -56.562 51.147
603171 96.493 -56.558 51.127
603172 96.517 -56.558 51.108
603173 96.539 -56.564 51.089
603174 96.562 -56.571 51.066
603175 96.585 -56.577 51.043
603176 96.607 -56.584 51.017
603177 96.630 -56.590 50.987
603178 96.653 -56.597 50.957
603179 96.675 -56.603 50.931
603180 96.709 -56.592 50.906
603181 96.743 -56.583 50.883
603182 96.769 -56.582 50.860
603183 96.797 -56.589 50.836
603184 96.816 -56.582 50.816
603185 96.849 -56.580 50.798
603186 96.900 -56.583 50.780
603187 96.951 -56.585 50.764
603188 96.998 -56.587 50.748
603189 97.031 -56.588 50.729
603190 97.065 -56.588 50.712
603191 97.098 -56.589 50.695
603192 97.132 -56.589 50.678
603193 97.165 -56.589 50.663
603194 97.198 -56.581 50.648
603195 97.231 -56.573 50.639
603196 97.264 -56.565 50.632
603197 97.298 -56.557 50.629
603198 97.336 -56.559 50.627
603199 97.377 -56.566 50.623
603200 97.418 -56.573 50.618
603201 97.459 -56.580 50.614
603202 97.495 -56.582 50.612
603203 97.530 -56.583 50.607
603204 97.564 -56.583 50.599
603205 97.598 -56.583 50.589
603206 97.633 -56.583 50.579
603207 97.667 -56.584 50.567
603208 97.702 -56.584 50.554
603209 97.737 -56.589 50.541
603210 97.773 -56.595 50.527
603211 97.809 -56.601 50.515
603212 97.845 -56.607 50.508
603213 97.881 -56.613 50.501
603214 97.917 -56.619 50.490
603215 97.953 -56.625 50.477
603216 97.989 -56.631 50.463
603217 98.025 -56.637 50.448
603218 98.062 -56.643 50.433
603219 98.097 -56.650 50.418
603220 98.133 -56.657 50.400
603221 98.167 -56.666 50.383
603222 98.201 -56.674 50.364
603223 98.235 -56.683 50.344
603224 98.270 -56.692 50.323
603225 98.304 -56.701 50.300
603226 98.339 -56.709 50.273
603227 98.373 -56.718 50.244
603228 98.407 -56.727 50.214
603229 98.441 -56.736 50.183
603230 98.476 -56.744 50.150
603231 98.510 -56.752 50.122
603232 98.544 -56.760 50.095
603233 98.578 -56.768 50.069
603234 98.612 -56.775 50.042
603235 98.646 -56.783 50.015
603236 98.679 -56.791 49.989
603237 98.714 -56.799 49.962
603238 98.747 -56.807 49.936
603239 98.787 -56.814 49.912
603240 98.827 -56.821 49.889
603241 98.867 -56.828 49.868
603242 98.907 -56.836 49.847
603243 98.947 -56.843 49.826
603244 98.987 -56.850 49.801
603245 99.027 -56.858 49.775
603246 99.067 -56.865 49.749
603247 99.108 -56.872 49.722
603248 99.149 -56.877 49.695
603249 99.193 -56.877 49.668
603250 99.237 -56.878 49.641
603251 99.269 -56.883 49.614
603252 99.290 -56.890 49.589
603253 99.318 -56.895 49.565
603254 99.344 -56.900 49.541
603255 99.368 -56.907 49.515
603256 99.393 -56.913 49.492
603257 99.418 -56.920 49.468
603258 99.442 -56.926 49.443
603259 99.467 -56.932 49.417
603260 99.492 -56.939 49.390
603261 99.517 -56.945 49.361
603262 99.542 -56.952 49.332
603263 99.566 -56.958 49.302
603264 99.591 -56.964 49.270
603265 99.630 -56.989 49.239
603266 99.649 -56.989 49.207
603267 99.675 -57.006 49.174
603268 99.700 -57.025 49.143
603269 99.726 -57.042 49.110
603270 99.751 -57.060 49.078
603271 99.777 -57.078 49.046
603272 99.802 -57.096 49.013
603273 99.828 -57.114 48.980
603274 99.853 -57.131 48.949
603275 99.879 -57.150 48.918
603276 99.905 -57.167 48.888
603277 99.930 -57.185 48.858
603278 99.956 -57.203 48.830
603279 99.981 -57.221 48.803
603280 100.007 -57.239 48.777
603281 100.033 -57.257 48.751
603612 88.403 -52.888 57.428
603613 88.440 -52.907 57.428
603614 88.479 -52.926 57.428
603615 88.517 -52.944 57.428
603616 88.555 -52.964 57.430
603617 88.593 -52.983 57.430
603618 88.631 -53.003 57.431
603619 88.669 -53.022 57.431
603620 88.707 -53.042 57.430
603621 88.745 -53.060 57.430
603622 88.783 -53.080 57.430
603623 88.821 -53.099 57.429
603624 88.859 -53.119 57.426
603625 88.897 -53.138 57.423
603626 88.935 -53.157 57.419
603627 88.973 -53.176 57.415
603628 89.011 -53.195 57.411
603629 89.033 -53.228 57.405
603630 89.064 -53.251 57.400
603631 89.096 -53.273 57.394
603632 89.128 -53.296 57.388
603633 89.159 -53.319 57.381
603634 89.191 -53.342 57.374
603635 89.223 -53.364 57.366
603636 89.254 -53.386 57.358
603637 89.286 -53.409 57.349
603638 89.311 -53.439 57.340
603639 89.334 -53.471 57.332
603640 89.356 -53.504 57.322
603641 89.379 -53.537 57.312
603642 89.402 -53.568 57.301
603643 89.427 -53.599 57.291
603644 89.466 -53.616 57.281
603645 89.518 -53.614 57.271
603646 89.558 -53.618 57.257
603647 89.584 -53.630 57.242
603648 89.609 -53.641 57.228
603649 89.634 -53.653 57.213
603650 89.660 -53.665 57.197
603651 89.691 -53.681 57.179
603652 89.725 -53.698 57.159
603653 89.760 -53.715 57.130
603654 89.794 -53.733 57.099
603655 89.828 -53.750 57.065
603656 89.863 -53.768 57.029
603657 89.897 -53.786 56.992
603658 89.932 -53.803 56.950
603659 89.966 -53.821 56.909
603660 90.000 -53.838 56.867
603661 90.034 -53.856 56.826
603662 90.069 -53.873 56.786
603663 90.103 -53.891 56.747
603664 90.138 -53.909 56.706
603665 90.172 -53.926 56.668
603671 90.412 -54.007 56.287
603672 90.455 -54.018 56.254
603673 90.483 -54.035 56.221
603674 90.500 -54.054 56.191
603675 90.517 -54.075 56.161
603676 90.534 -54.095 56.131
603677 90.552 -54.115 56.102
603678 90.569 -54.135 56.075
603679 90.586 -54.155 56.050
603680 90.603 -54.176 56.026
603681 90.620 -54.196 56.003
603682 90.638 -54.216 55.980
603683 90.655 -54.236 55.960
603684 90.674 -54.254 55.940
603685 90.710 -54.256 55.925
603686 90.746 -54.258 55.911
603687 90.777 -54.271 55.894
603688 90.806 -54.292 55.873
603689 90.835 -54.312 55.846
603690 90.864 -54.332 55.820
603691 90.893 -54.352 55.794
603692 90.922 -54.372 55.769
603693 90.951 -54.392 55.746
603694 90.979 -54.413 55.724
603695 91.009 -54.433 55.705
603696 91.037 -54.453 55.688
603697 91.066 -54.473 55.673
603698 91.095 -54.493 55.661
603699 91.124 -54.513 55.652
603700 91.153 -54.533 55.647
603701 91.182 -54.554 55.649
603702 91.211 -54.574 55.652
603703 91.242 -54.592 55.655
603704 91.291 -54.599 55.656
603705 91.345 -54.604 55.655
603706 91.398 -54.609 55.653
603707 91.451 -54.614 55.648
603708 91.505 -54.619 55.633
603709 91.558 -54.624 55.612
603710 91.611 -54.629 55.585
603711 91.665 -54.634 55.558
603712 91.719 -54.638 55.536
603713 91.772 -54.643 55.511
603714 91.817 -54.648 55.485
603715 91.847 -54.652 55.460
603716 91.875 -54.663 55.439
603717 91.902 -54.674 55.420
603718 91.932 -54.685 55.406
603719 91.964 -54.696 55.394
603720 91.997 -54.706 55.383
603721 92.029 -54.717 55.373
603722 92.062 -54.728 55.362
603723 92.094 -54.738 55.353
603724 92.126 -54.749 55.346
603725 92.158 -54.759 55.337
603726 92.191 -54.769 55.327
603727 92.223 -54.780 55.317
603728 92.256 -54.795 55.308
603729 92.290 -54.815 55.299
603730 92.324 -54.834 55.294
603731 92.358 -54.853 55.290
603732 92.392 -54.873 55.289
603733 92.426 -54.891 55.290
603734 92.457 -54.897 55.291
603735 92.489 -54.903 55.292
603736 92.520 -54.909 55.296
603737 92.552 -54.915 55.302
603738 92.584 -54.920 55.309
603739 92.620 -54.925 55.321
603740 92.655 -54.930 55.336
603741 92.682 -54.941 55.354
603742 92.708 -54.953 55.378
603743 92.734 -54.972 55.403
603744 92.760 -54.990 55.430
603745 92.786 -55.008 55.455
603746 92.812 -55.026 55.480
603747 92.838 -55.044 55.504
603748 92.864 -55.062 55.521
603749 92.890 -55.080 55.530
603750 92.917 -55.098 55.536
603751 92.952 -55.116 55.543
603752 92.988 -55.135 55.548
603753 93.023 -55.152 55.547
603754 93.059 -55.170 55.540
603755 93.094 -55.188 55.534
603756 93.130 -55.206 55.523
603757 93.165 -55.224 55.508
603758 93.201 -55.242 55.491
603759 93.237 -55.260 55.470
603760 93.272 -55.278 55.444
603761 93.308 -55.296 55.414
603762 93.344 -55.314 55.381
603763 93.379 -55.332 55.348
603764 93.415 -55.350 55.323
603765 93.450 -55.368 55.308
603766 93.486 -55.386 55.303
603767 93.521 -55.404 55.295
603768 93.557 -55.422 55.289
603769 93.587 -55.442 55.283
603770 93.608 -55.462 55.274
603771 93.630 -55.483 55.262
603772 93.651 -55.503 55.247
603773 93.673 -55.524 55.232
603774 93.695 -55.545 55.218
603775 93.716 -55.565 55.207
603776 93.738 -55.586 55.194
603777 93.760 -55.606 55.174
603778 93.781 -55.627 55.146
603779 93.802 -55.647 55.113
603780 93.824 -55.668 55.078
603781 93.846 -55.688 55.042
603782 93.867 -55.709 55.006
603783 93.888 -55.737 54.970
603784 93.907 -55.766 54.933
603785 93.927 -55.796 54.899
603786 93.947 -55.825 54.867
603787 93.968 -55.853 54.834
603788 93.997 -55.865 54.801
603789 94.026 -55.877 54.766
603790 94.057 -55.884 54.729
603791 94.090 -55.891 54.695
603792 94.122 -55.899 54.662
603793 94.155 -55.906 54.630
603794 94.187 -55.912 54.598
603795 94.220 -55.919 54.566
603796 94.252 -55.926 54.531
603797 94.285 -55.933 54.493
603798 94.317 -55.940 54.452
603799 94.351 -55.948 54.407
603800 94.385 -55.957 54.363
603801 94.419 -55.965 54.321
603802 94.453 -55.974 54.282
603803 94.495 -55.984 54.244
603804 94.546 -55.995 54.211
603805 94.596 -56.007 54.176
603806 94.646 -56.019 54.141
603807 94.693 -56.029 54.104
603808 94.727 -56.037 54.062
603809 94.761 -56.045 54.019
603810 94.795 -56.053 53.970
603811 94.829 -56.060 53.920
603812 94.863 -56.067 53.869
603813 94.897 -56.071 53.816
603814 94.930 -56.074 53.761
603815 94.964 -56.077 53.704
603816 94.998 -56.081 53.649
603817 95.032 -56.085 53.592
603818 95.070 -56.109 53.533
603819 95.109 -56.132 53.473
603820 95.147 -56.156 53.415
603821 95.172 -56.169 53.359
603822 95.191 -56.179 53.309
603823 95.209 -56.188 53.259
603824 95.228 -56.201 53.209
603825 95.246 -56.216 53.158
603826 95.270 -56.232 53.105
603827 95.304 -56.247 53.051
603828 95.337 -56.262 52.993
603829 95.371 -56.277 52.933
603830 95.405 -56.290 52.873
603831 95.443 -56.299 52.814
603832 95.480 -56.308 52.754
603833 95.518 -56.317 52.696
603834 95.556 -56.326 52.636
603835 95.593 -56.335 52.576
603836 95.630 -56.345 52.512
603837 95.657 -56.345 52.444
603838 95.684 -56.345 52.377
603839 95.715 -56.348 52.312
603840 95.750 -56.354 52.249
603841 95.782 -56.363 52.188
603842 95.808 -56.380 52.130
603843 95.834 -56.398 52.074
603844 95.860 -56.415 52.020
603845 95.886 -56.433 51.967
603846 95.911 -56.438 51.914
603847 95.937 -56.441 51.865
603848 95.962 -56.445 51.815
603849 95.987 -56.448 51.768
603850 96.012 -56.451 51.724
603851 96.037 -56.454 51.681
603852 96.062 -56.457 51.639
603853 96.088 -56.460 51.599
603854 96.112 -56.463 51.561
603855 96.137 -56.466 51.525
603856 96.163 -56.469 51.491
603857 96.188 -56.472 51.458
603858 96.218 -56.474 51.426
603859 96.250 -56.476 51.395
603860 96.282 -56.479 51.364
603861 96.297 -56.479 51.336
603862 96.313 -56.481 51.309
603863 96.328 -56.482 51.283
603864 96.343 -56.483 51.256
603865 96.359 -56.484 51.231
603866 96.375 -56.498 51.206
603867 96.390 -56.514 51.185
603868 96.406 -56.528 51.165
603869 96.422 -56.541 51.145
603870 96.446 -56.536 51.125
603871 96.469 -56.531 51.106
603872 96.493 -56.526 51.086
603873 96.516 -56.526 51.066
603874 96.539 -56.532 51.045
603875 96.561 -56.539 51.022
603876 96.584 -56.545 50.996
603877 96.607 -56.552 50.970
603878 96.629 -56.558 50.942
603879 96.652 -56.564 50.915
603880 96.675 -56.570 50.889
603881 96.697 -56.577 50.864
603882 96.732 -56.565 50.840
603883 96.766 -56.554 50.816
603884 96.790 -56.552 50.792
603885 96.809 -56.545 50.770
603886 96.847 -56.544 50.751
603887 96.897 -56.547 50.732
603888 96.949 -56.550 50.715
603889 96.996 -56.552 50.699
603890 97.029 -56.552 50.680
603891 97.063 -56.553 50.665
603892 97.096 -56.553 50.650
603893 97.130 -56.554 50.634
603894 97.163 -56.554 50.619
603895 97.196 -56.549 50.605
603896 97.230 -56.541 50.595
603897 97.263 -56.533 50.586
603898 97.298 -56.528 50.580
603899 97.338 -56.535 50.576
603900 97.380 -56.542 50.572
603901 97.421 -56.548 50.568
603902 97.455 -56.548 50.564
603903 97.489 -56.549 50.562
603904 97.523 -56.549 50.559
603905 97.558 -56.550 50.553
603906 97.592 -56.550 50.545
603907 97.627 -56.551 50.538
603908 97.661 -56.551 50.528
603909 97.695 -56.552 50.517
603910 97.730 -56.555 50.503
603911 97.767 -56.561 50.489
603912 97.802 -56.567 50.476
603913 97.839 -56.573 50.468
603914 97.875 -56.579 50.462
603915 97.911 -56.586 50.453
603916 97.947 -56.592 50.442
603917 97.983 -56.598 50.428
603918 98.019 -56.604 50.414
603919 98.055 -56.610 50.399
603920 98.091 -56.616 50.384
603921 98.127 -56.622 50.367
603922 98.163 -56.628 50.349
603923 98.199 -56.635 50.330
603924 98.233 -56.643 50.309
603925 98.268 -56.652 50.288
603926 98.302 -56.661 50.266
603927 98.336 -56.670 50.239
603928 98.371 -56.678 50.210
603929 98.405 -56.686 50.180
603930 98.439 -56.696 50.150
603931 98.474 -56.704 50.119
603932 98.507 -56.712 50.091
603933 98.541 -56.720 50.064
603934 98.575 -56.727 50.039
603935 98.609 -56.735 50.012
603936 98.643 -56.743 49.985
603937 98.679 -56.751 49.959
603938 98.720 -56.758 49.932
603939 98.760 -56.765 49.906
603940 98.800 -56.772 49.882
603941 98.840 -56.780 49.859
603942 98.880 -56.787 49.837
603943 98.920 -56.794 49.816
603944 98.960 -56.802 49.794
603945 99.000 -56.809 49.769
603946 99.040 -56.816 49.744
603947 99.080 -56.824 49.719
603948 99.120 -56.830 49.693
603949 99.160 -56.838 49.666
603950 99.200 -56.846 49.639
603951 99.240 -56.852 49.612
603952 99.282 -56.853 49.586
603953 99.306 -56.860 49.561
603954 99.332 -56.866 49.537
603955 99.356 -56.872 49.513
603956 99.381 -56.879 49.488
603957 99.406 -56.885 49.465
603958 99.430 -56.892 49.440
603959 99.455 -56.898 49.415
603960 99.480 -56.904 49.389
603961 99.504 -56.911 49.363
603962 99.529 -56.917 49.336
603963 99.554 -56.923 49.309
603964 99.579 -56.930 49.280
603965 99.604 -56.936 49.250
603966 99.635 -56.948 49.219
603967 99.660 -56.965 49.188
603968 99.686 -56.983 49.156
603969 99.711 -57.000 49.125
603970 99.736 -57.018 49.093
603971 99.762 -57.036 49.061
603972 99.788 -57.054 49.029
603973 99.814 -57.072 48.997
603974 99.839 -57.090 48.965
603975 99.864 -57.108 48.935
603976 99.890 -57.125 48.905
603977 99.916 -57.144 48.876
603978 99.941 -57.161 48.847
603979 99.967 -57.179 48.819
603980 99.992 -57.197 48.793
603981 100.018 -57.215 48.767
603982 100.043 -57.233 48.741
604313 88.430 -52.857 57.423
604314 88.468 -52.876 57.422
604315 88.506 -52.895 57.422
604316 88.544 -52.914 57.423
604317 88.582 -52.933 57.425
604318 88.620 -52.953 57.425
604319 88.659 -52.972 57.426
604320 88.696 -52.991 57.425
604321 88.735 -53.010 57.425
604322 88.772 -53.030 57.424
604323 88.811 -53.049 57.424
604324 88.848 -53.069 57.422
604325 88.887 -53.088 57.419
604326 88.924 -53.107 57.416
604327 88.963 -53.126 57.412
604328 88.986 -53.158 57.407
604329 89.017 -53.181 57.402
604330 89.049 -53.204 57.398
604331 89.081 -53.227 57.392
604332 89.112 -53.249 57.386
604333 89.144 -53.272 57.379
604334 89.175 -53.295 57.372
604335 89.207 -53.317 57.364
604336 89.239 -53.340 57.357
604337 89.270 -53.363 57.349
604338 89.302 -53.386 57.340
604339 89.326 -53.416 57.332
604340 89.349 -53.449 57.324
604341 89.372 -53.481 57.315
604342 89.394 -53.514 57.304
604343 89.417 -53.546 57.294
604344 89.440 -53.578 57.284
604345 89.478 -53.593 57.274
604346 89.530 -53.590 57.264
604347 89.574 -53.593 57.250
604348 89.600 -53.604 57.235
604349 89.625 -53.616 57.221
604350 89.651 -53.628 57.207
604351 89.676 -53.640 57.190
604352 89.702 -53.652 57.173
604353 89.735 -53.668 57.152
604354 89.769 -53.686 57.126
604355 89.803 -53.703 57.097
604356 89.837 -53.721 57.065
604357 89.872 -53.738 57.030
604358 89.906 -53.756 56.993
604359 89.941 -53.773 56.953
604360 89.975 -53.791 56.911
604361 90.009 -53.809 56.870
604362 90.044 -53.826 56.829
604363 90.078 -53.844 56.789
604364 90.112 -53.861 56.750
604365 90.147 -53.878 56.709
604366 90.181 -53.896 56.671
604367 90.215 -53.914 56.634
604373 90.470 -53.984 56.255
604374 90.490 -54.003 56.222
604375 90.508 -54.023 56.190
604376 90.525 -54.044 56.160
604377 90.542 -54.063 56.129
604378 90.559 -54.084 56.101
604379 90.577 -54.104 56.073
604380 90.594 -54.124 56.047
604381 90.611 -54.145 56.023
604382 90.628 -54.164 56.000
604383 90.646 -54.185 55.977
604384 90.663 -54.204 55.956
604385 90.697 -54.207 55.938
604386 90.733 -54.209 55.923
604387 90.767 -54.215 55.909
604388 90.796 -54.235 55.892
604389 90.825 -54.255 55.868
604390 90.854 -54.275 55.838
604391 90.883 -54.295 55.809
604392 90.912 -54.316 55.782
604393 90.941 -54.336 55.758
604394 90.970 -54.356 55.735
604395 90.999 -54.376 55.715
604396 91.028 -54.396 55.697
604397 91.057 -54.417 55.680
604398 91.086 -54.436 55.666
604399 91.114 -54.457 55.654
604400 91.144 -54.477 55.645
604401 91.172 -54.497 55.642
604402 91.202 -54.517 55.643
604403 91.230 -54.537 55.647
604404 91.262 -54.557 55.650
604405 91.311 -54.566 55.650
604406 91.364 -54.571 55.648
604407 91.416 -54.576 55.646
604408 91.469 -54.581 55.637
604409 91.522 -54.587 55.621
604410 91.575 -54.591 55.598
604411 91.629 -54.596 55.572
604412 91.682 -54.601 55.546
604413 91.736 -54.606 55.523
604414 91.789 -54.611 55.500
604415 91.836 -54.615 55.476
604416 91.863 -54.624 55.452
604417 91.894 -54.635 55.432
604418 91.926 -54.646 55.414
604419 91.958 -54.656 55.400
604420 91.991 -54.667 55.388
604421 92.023 -54.678 55.377
604422 92.055 -54.688 55.367
604423 92.088 -54.699 55.357
604424 92.120 -54.709 55.348
604425 92.152 -54.720 55.340
604426 92.184 -54.731 55.332
604427 92.217 -54.741 55.323
604428 92.249 -54.752 55.313
604429 92.281 -54.762 55.304
604430 92.314 -54.777 55.296
604431 92.349 -54.797 55.291
604432 92.382 -54.816 55.288
604433 92.417 -54.836 55.287
604434 92.451 -54.856 55.286
604435 92.483 -54.862 55.287
604436 92.514 -54.868 55.289
604437 92.546 -54.874 55.295
604438 92.577 -54.880 55.304
604439 92.609 -54.886 55.314
604440 92.640 -54.892 55.331
604441 92.672 -54.897 55.352
604442 92.702 -54.909 55.374
604443 92.727 -54.927 55.399
604444 92.754 -54.945 55.426
604445 92.779 -54.963 55.455
604446 92.805 -54.981 55.482
604447 92.831 -55.000 55.509
604448 92.857 -55.018 55.532
604449 92.883 -55.036 55.548
604450 92.909 -55.054 55.557
604451 92.935 -55.072 55.563
604452 92.968 -55.090 55.567
604453 93.004 -55.108 55.568
604454 93.039 -55.126 55.563
604455 93.075 -55.144 55.553
604456 93.111 -55.162 55.545
604457 93.146 -55.180 55.533
604458 93.182 -55.198 55.518
604459 93.217 -55.216 55.500
604460 93.253 -55.234 55.481
604461 93.288 -55.252 55.457
604462 93.324 -55.270 55.427
604463 93.359 -55.288 55.393
604464 93.395 -55.306 55.359
604465 93.430 -55.323 55.329
604466 93.466 -55.342 55.310
604467 93.502 -55.360 55.303
604468 93.537 -55.377 55.297
604469 93.570 -55.395 55.296
604470 93.594 -55.408 55.294
604471 93.615 -55.433 55.287
604472 93.635 -55.457 55.275
604473 93.655 -55.480 55.260
604474 93.676 -55.505 55.246
604475 93.696 -55.529 55.234
604476 93.716 -55.553 55.224
604477 93.737 -55.574 55.212
604478 93.759 -55.594 55.194
604479 93.781 -55.614 55.168
604480 93.802 -55.635 55.135
604481 93.824 -55.655 55.098
604482 93.846 -55.676 55.059
604483 93.867 -55.696 55.021
604484 93.889 -55.718 54.982
604485 93.910 -55.739 54.943
604486 93.930 -55.768 54.907
604487 93.950 -55.798 54.872
604488 93.970 -55.828 54.838
604489 93.997 -55.841 54.803
604490 94.028 -55.852 54.767
604491 94.060 -55.859 54.728
604492 94.092 -55.865 54.692
604493 94.125 -55.873 54.657
604494 94.157 -55.880 54.622
604495 94.190 -55.887 54.587
604496 94.222 -55.893 54.553
604497 94.255 -55.900 54.517
604498 94.287 -55.907 54.480
604499 94.320 -55.915 54.440
604500 94.353 -55.922 54.400
604501 94.387 -55.931 54.362
604502 94.421 -55.939 54.327
604503 94.455 -55.947 54.287
604504 94.502 -55.959 54.260
604505 94.552 -55.970 54.234
604506 94.603 -55.981 54.206
604507 94.653 -55.994 54.174
604508 94.703 -56.005 54.137
604509 94.737 -56.013 54.094
604510 94.771 -56.020 54.051
604511 94.805 -56.028 54.005
604512 94.839 -56.036 53.957
604513 94.873 -56.041 53.906
604514 94.906 -56.044 53.852
604515 94.941 -56.047 53.795
604516 94.974 -56.050 53.736
604517 95.008 -56.053 53.680
604518 95.042 -56.056 53.621
604519 95.080 -56.079 53.559
604520 95.118 -56.103 53.494
604521 95.154 -56.125 53.430
604522 95.173 -56.134 53.369
604523 95.192 -56.144 53.314
604524 95.211 -56.154 53.261
604525 95.230 -56.163 53.210
604526 95.248 -56.178 53.157
604527 95.281 -56.193 53.102
604528 95.315 -56.208 53.048
604529 95.348 -56.223 52.990
604530 95.382 -56.238 52.930
604531 95.415 -56.254 52.870
604532 95.449 -56.268 52.811
604533 95.485 -56.279 52.752
604534 95.523 -56.288 52.695
604535 95.560 -56.297 52.637
604536 95.598 -56.306 52.577
604537 95.635 -56.315 52.516
604538 95.664 -56.317 52.452
604539 95.691 -56.317 52.386
604540 95.725 -56.322 52.315
604541 95.759 -56.329 52.250
604542 95.784 -56.346 52.188
604543 95.811 -56.364 52.127
604544 95.837 -56.381 52.068
604545 95.863 -56.399 52.013
604546 95.889 -56.415 51.958
604547 95.914 -56.418 51.904
604548 95.939 -56.421 51.852
604549 95.964 -56.424 51.801
604550 95.989 -56.427 51.752
604551 96.015 -56.430 51.706
604552 96.040 -56.434 51.662
604553 96.064 -56.437 51.619
604554 96.090 -56.440 51.578
604555 96.115 -56.443 51.538
604556 96.140 -56.446 51.501
604557 96.168 -56.448 51.466
604558 96.200 -56.451 51.433
604559 96.232 -56.453 51.401
604560 96.264 -56.454 51.368
604561 96.292 -56.457 51.337
604562 96.308 -56.458 51.308
604563 96.323 -56.459 51.282
604564 96.338 -56.460 51.255
604565 96.354 -56.461 51.228
604566 96.369 -56.462 51.202
604567 96.385 -56.463 51.176
604568 96.400 -56.464 51.152
604569 96.415 -56.471 51.131
604570 96.431 -56.486 51.110
604571 96.447 -56.501 51.089
604572 96.469 -56.501 51.067
604573 96.492 -56.495 51.046
604574 96.516 -56.493 51.024
604575 96.538 -56.500 51.003
604576 96.561 -56.506 50.981
604577 96.584 -56.512 50.955
604578 96.606 -56.519 50.929
604579 96.629 -56.525 50.901
604580 96.652 -56.532 50.875
604581 96.674 -56.538 50.850
604582 96.697 -56.545 50.824
604583 96.720 -56.551 50.800
604584 96.754 -56.534 50.775
604585 96.782 -56.517 50.751
604586 96.801 -56.508 50.728
604587 96.844 -56.509 50.707
604588 96.895 -56.512 50.688
604589 96.947 -56.514 50.670
604590 96.995 -56.517 50.653
604591 97.028 -56.517 50.636
604592 97.062 -56.517 50.621
604593 97.095 -56.518 50.607
604594 97.129 -56.518 50.592
604595 97.162 -56.519 50.578
604596 97.195 -56.517 50.564
604597 97.228 -56.509 50.552
604598 97.261 -56.501 50.542
604599 97.300 -56.504 50.535
604600 97.342 -56.511 50.529
604601 97.380 -56.515 50.524
604602 97.414 -56.515 50.519
604603 97.449 -56.516 50.515
604604 97.483 -56.516 50.512
604605 97.517 -56.517 50.508
604606 97.551 -56.517 50.504
604607 97.586 -56.517 50.497
604608 97.620 -56.517 50.491
604609 97.655 -56.518 50.483
604610 97.689 -56.519 50.473
604611 97.724 -56.521 50.461
604612 97.760 -56.527 50.448
604613 97.796 -56.533 50.436
604614 97.832 -56.539 50.426
604615 97.868 -56.545 50.420
604616 97.904 -56.552 50.414
604617 97.940 -56.558 50.404
604618 97.977 -56.564 50.392
604619 98.012 -56.570 50.378
604620 98.049 -56.576 50.365
604621 98.085 -56.583 50.349
604622 98.121 -56.589 50.332
604623 98.157 -56.595 50.314
604624 98.193 -56.601 50.294
604625 98.229 -56.607 50.274
604626 98.265 -56.613 50.252
604627 98.300 -56.620 50.228
604628 98.334 -56.630 50.202
604629 98.368 -56.638 50.173
604630 98.403 -56.647 50.142
604631 98.437 -56.655 50.113
604632 98.471 -56.664 50.085
604633 98.505 -56.671 50.059
604634 98.539 -56.680 50.033
604635 98.573 -56.687 50.007
604636 98.613 -56.695 49.980
604637 98.652 -56.702 49.953
604638 98.693 -56.709 49.927
604639 98.732 -56.716 49.902
604640 98.773 -56.724 49.875
604641 98.812 -56.731 49.850
604642 98.853 -56.738 49.828
604643 98.893 -56.746 49.806
604644 98.933 -56.753 49.784
604645 98.973 -56.760 49.761
604646 99.013 -56.768 49.735
604647 99.053 -56.774 49.711
604648 99.093 -56.782 49.688
604649 99.130 -56.790 49.663
604650 99.166 -56.798 49.637
604651 99.203 -56.805 49.610
604652 99.239 -56.814 49.584
604653 99.275 -56.821 49.558
604654 99.311 -56.830 49.532
604655 99.342 -56.837 49.508
604656 99.369 -56.844 49.485
604657 99.394 -56.851 49.461
604658 99.418 -56.857 49.437
604659 99.443 -56.863 49.412
604660 99.468 -56.870 49.388
604661 99.492 -56.876 49.363
604662 99.517 -56.883 49.337
604663 99.542 -56.889 49.311
604664 99.567 -56.895 49.286
604665 99.591 -56.902 49.258
604666 99.619 -56.911 49.230
604667 99.645 -56.923 49.200
604668 99.671 -56.941 49.169
604669 99.697 -56.959 49.139
604670 99.722 -56.977 49.107
604671 99.748 -56.994 49.075
604672 99.773 -57.012 49.044
604673 99.799 -57.030 49.013
604674 99.824 -57.048 48.982
604675 99.850 -57.066 48.951
604676 99.875 -57.084 48.922
604677 99.901 -57.102 48.893
604678 99.926 -57.119 48.864
604679 99.952 -57.138 48.836
604680 99.978 -57.155 48.809
604681 100.003 -57.173 48.783
604682 100.028 -57.191 48.757
605015 88.496 -52.845 57.417
605016 88.534 -52.864 57.417
605017 88.572 -52.883 57.417
605018 88.610 -52.903 57.419
605019 88.648 -52.922 57.420
605020 88.686 -52.941 57.420
605021 88.724 -52.960 57.420
605022 88.762 -52.980 57.419
605023 88.800 -52.999 57.418
605024 88.838 -53.019 57.418
605025 88.876 -53.038 57.415
605026 88.914 -53.057 57.412
605027 88.952 -53.076 57.408
605028 88.970 -53.112 57.404
605029 89.001 -53.135 57.399
605030 89.033 -53.158 57.393
605031 89.065 -53.180 57.388
605032 89.097 -53.203 57.383
605033 89.128 -53.226 57.376
605034 89.160 -53.248 57.370
605035 89.192 -53.270 57.363
605036 89.223 -53.293 57.355
605037 89.255 -53.316 57.347
605038 89.286 -53.339 57.339
605039 89.318 -53.361 57.331
605040 89.341 -53.393 57.323
605041 89.364 -53.426 57.317
605042 89.387 -53.458 57.307
605043 89.409 -53.490 57.296
605044 89.432 -53.523 57.286
605045 89.455 -53.556 57.277
605046 89.491 -53.569 57.267
605047 89.542 -53.568 57.257
605048 89.590 -53.567 57.244
605049 89.616 -53.579 57.229
605050 89.641 -53.591 57.215
605051 89.667 -53.603 57.200
605052 89.692 -53.615 57.183
605053 89.718 -53.626 57.166
605054 89.744 -53.638 57.145
605055 89.778 -53.656 57.123
605056 89.813 -53.673 57.095
605057 89.847 -53.691 57.064
605058 89.881 -53.709 57.031
605059 89.915 -53.726 56.995
605060 89.950 -53.744 56.955
605061 89.984 -53.761 56.914
605062 90.019 -53.778 56.873
605063 90.053 -53.796 56.832
605064 90.087 -53.814 56.793
605065 90.121 -53.831 56.754
605066 90.156 -53.849 56.713
605067 90.190 -53.866 56.675
605068 90.227 -53.882 56.638
605075 90.498 -53.972 56.224
605076 90.515 -53.992 56.192
605077 90.533 -54.012 56.162
605078 90.550 -54.032 56.129
605079 90.567 -54.052 56.100
605080 90.584 -54.073 56.072
605081 90.601 -54.092 56.045
605082 90.618 -54.113 56.021
605083 90.636 -54.133 55.998
605084 90.653 -54.153 55.974
605085 90.685 -54.159 55.953
605086 90.721 -54.160 55.935
605087 90.756 -54.162 55.920
605088 90.787 -54.178 55.906
605089 90.815 -54.198 55.888
605090 90.845 -54.219 55.861
605091 90.873 -54.239 55.830
605092 90.902 -54.259 55.801
605093 90.931 -54.279 55.774
605094 90.960 -54.299 55.750
605095 90.989 -54.320 55.728
605096 91.018 -54.339 55.708
605097 91.047 -54.360 55.691
605098 91.076 -54.380 55.674
605099 91.105 -54.400 55.660
605100 91.134 -54.420 55.648
605101 91.163 -54.440 55.639
605102 91.192 -54.461 55.635
605103 91.221 -54.480 55.636
605104 91.254 -54.500 55.640
605105 91.287 -54.521 55.642
605106 91.332 -54.531 55.642
605107 91.385 -54.536 55.640
605108 91.438 -54.542 55.637
605109 91.491 -54.546 55.627
605110 91.543 -54.552 55.610
605111 91.596 -54.557 55.586
605112 91.649 -54.562 55.560
605113 91.702 -54.568 55.535
605114 91.754 -54.573 55.514
605115 91.807 -54.577 55.492
605116 91.854 -54.585 55.468
605117 91.887 -54.596 55.445
605118 91.920 -54.607 55.425
605119 91.952 -54.618 55.407
605120 91.985 -54.628 55.395
605121 92.017 -54.638 55.383
605122 92.049 -54.649 55.373
605123 92.081 -54.659 55.362
605124 92.113 -54.670 55.352
605125 92.146 -54.681 55.343
605126 92.178 -54.691 55.335
605127 92.211 -54.702 55.327
605128 92.243 -54.712 55.319
605129 92.275 -54.723 55.310
605130 92.307 -54.734 55.301
605131 92.340 -54.744 55.294
605132 92.373 -54.759 55.289
605133 92.407 -54.778 55.287
605134 92.441 -54.798 55.285
605135 92.475 -54.818 55.285
605136 92.507 -54.828 55.286
605137 92.539 -54.834 55.289
605138 92.570 -54.840 55.297
605139 92.602 -54.846 55.308
605140 92.634 -54.851 55.322
605141 92.665 -54.857 55.341
605142 92.695 -54.868 55.366
605143 92.720 -54.883 55.390
605144 92.746 -54.900 55.417
605145 92.772 -54.919 55.446
605146 92.798 -54.937 55.474
605147 92.824 -54.955 55.502
605148 92.850 -54.973 55.529
605149 92.876 -54.991 55.550
605150 92.902 -55.010 55.564
605151 92.928 -55.028 55.573
605152 92.954 -55.046 55.580
605153 92.984 -55.063 55.583
605154 93.019 -55.082 55.582
605155 93.055 -55.100 55.574
605156 93.090 -55.118 55.561
605157 93.126 -55.135 55.552
605158 93.162 -55.154 55.538
605159 93.197 -55.172 55.520
605160 93.233 -55.189 55.503
605161 93.269 -55.207 55.487
605162 93.304 -55.226 55.465
605163 93.340 -55.244 55.435
605164 93.375 -55.261 55.401
605165 93.411 -55.279 55.366
605166 93.446 -55.298 55.336
605167 93.482 -55.315 55.319
605168 93.517 -55.333 55.312
605169 93.551 -55.351 55.305
605170 93.576 -55.364 55.300
605171 93.601 -55.378 55.296
605172 93.623 -55.399 55.289
605173 93.643 -55.423 55.278
605174 93.663 -55.448 55.263
605175 93.684 -55.472 55.251
605176 93.704 -55.496 55.241
605177 93.724 -55.520 55.232
605178 93.744 -55.544 55.222
605179 93.764 -55.568 55.207
605180 93.784 -55.593 55.182
605181 93.805 -55.617 55.150
605182 93.825 -55.641 55.113
605183 93.845 -55.665 55.074
605184 93.867 -55.685 55.037
605185 93.889 -55.705 55.000
605186 93.910 -55.726 54.960
605187 93.932 -55.746 54.923
605188 93.953 -55.771 54.886
605189 93.973 -55.799 54.849
605190 93.998 -55.818 54.814
605191 94.030 -55.826 54.778
605192 94.062 -55.833 54.737
605193 94.095 -55.840 54.698
605194 94.127 -55.847 54.659
605195 94.160 -55.854 54.622
605196 94.192 -55.861 54.586
605197 94.225 -55.868 54.550
605198 94.257 -55.875 54.511
605199 94.290 -55.881 54.474
605200 94.322 -55.889 54.436
605201 94.355 -55.896 54.400
605202 94.389 -55.904 54.365
605203 94.423 -55.912 54.332
605204 94.459 -55.922 54.298
605205 94.509 -55.933 54.278
605206 94.559 -55.944 54.258
605207 94.610 -55.956 54.235
605208 94.660 -55.968 54.208
605209 94.710 -55.979 54.174
605210 94.747 -55.987 54.133
605211 94.781 -55.996 54.089
605212 94.815 -56.003 54.043
605213 94.849 -56.010 53.994
605214 94.883 -56.014 53.943
605215 94.916 -56.017 53.888
605216 94.950 -56.020 53.829
605217 94.984 -56.023 53.769
605218 95.018 -56.027 53.711
605219 95.052 -56.030 53.651
605220 95.090 -56.050 53.586
605221 95.128 -56.074 53.518
605222 95.155 -56.090 53.449
605223 95.174 -56.099 53.381
605224 95.193 -56.109 53.321
605225 95.212 -56.119 53.264
605226 95.231 -56.128 53.210
605227 95.258 -56.142 53.155
605228 95.293 -56.154 53.098
605229 95.326 -56.169 53.042
605230 95.359 -56.184 52.984
605231 95.393 -56.199 52.924
605232 95.426 -56.214 52.864
605233 95.460 -56.229 52.805
605234 95.493 -56.244 52.747
605235 95.528 -56.258 52.689
605236 95.565 -56.267 52.633
605237 95.603 -56.276 52.575
605238 95.640 -56.285 52.515
605239 95.671 -56.289 52.455
605240 95.699 -56.290 52.391
605241 95.734 -56.296 52.321
605242 95.761 -56.312 52.253
605243 95.788 -56.329 52.189
605244 95.813 -56.347 52.126
605245 95.839 -56.365 52.065
605246 95.866 -56.382 52.007
605247 95.891 -56.395 51.950
605248 95.916 -56.398 51.895
605249 95.942 -56.401 51.841
605250 95.967 -56.404 51.788
605251 95.992 -56.407 51.738
605252 96.017 -56.410 51.691
605253 96.042 -56.414 51.645
605254 96.067 -56.417 51.601
605255 96.092 -56.420 51.557
605256 96.118 -56.423 51.516
605257 96.150 -56.425 51.478
605258 96.182 -56.427 51.442
605259 96.214 -56.429 51.408
605260 96.246 -56.431 51.374
605261 96.277 -56.433 51.342
605262 96.303 -56.435 51.311
605263 96.318 -56.436 51.282
605264 96.333 -56.437 51.255
605265 96.349 -56.438 51.229
605266 96.364 -56.439 51.202
605267 96.380 -56.440 51.174
605268 96.395 -56.441 51.146
605269 96.410 -56.442 51.121
605270 96.426 -56.443 51.099
605271 96.442 -56.445 51.076
605272 96.457 -56.445 51.054
605273 96.472 -56.459 51.031
605274 96.491 -56.465 51.007
605275 96.515 -56.461 50.985
605276 96.538 -56.467 50.964
605277 96.560 -56.473 50.942
605278 96.583 -56.479 50.917
605279 96.606 -56.486 50.890
605280 96.628 -56.492 50.862
605281 96.651 -56.498 50.836
605282 96.673 -56.504 50.812
605283 96.694 -56.508 50.787
605284 96.716 -56.514 50.761
605285 96.743 -56.507 50.735
605286 96.774 -56.489 50.710
605287 96.799 -56.478 50.687
605288 96.843 -56.474 50.666
605289 96.893 -56.476 50.646
605290 96.945 -56.479 50.626
605291 96.993 -56.482 50.609
605292 97.026 -56.483 50.592
605293 97.060 -56.483 50.578
605294 97.093 -56.483 50.564
605295 97.127 -56.483 50.550
605296 97.160 -56.484 50.536
605297 97.194 -56.484 50.522
605298 97.227 -56.476 50.509
605299 97.262 -56.473 50.499
605300 97.303 -56.479 50.490
605301 97.339 -56.482 50.483
605302 97.374 -56.482 50.477
605303 97.408 -56.483 50.471
605304 97.442 -56.483 50.466
605305 97.476 -56.483 50.461
605306 97.511 -56.483 50.457
605307 97.545 -56.484 50.453
605308 97.580 -56.484 50.448
605309 97.614 -56.485 50.444
605310 97.648 -56.485 50.438
605311 97.683 -56.486 50.430
605312 97.717 -56.487 50.418
605313 97.753 -56.493 50.405
605314 97.790 -56.499 50.393
605315 97.825 -56.505 50.382
605316 97.862 -56.511 50.376
605317 97.898 -56.518 50.370
605318 97.934 -56.524 50.361
605319 97.970 -56.530 50.351
605320 98.006 -56.536 50.341
605321 98.042 -56.542 50.327
605322 98.078 -56.548 50.312
605323 98.114 -56.555 50.296
605324 98.150 -56.561 50.277
605325 98.186 -56.567 50.258
605326 98.223 -56.573 50.237
605327 98.258 -56.580 50.214
605328 98.295 -56.586 50.189
605329 98.331 -56.592 50.162
605330 98.366 -56.599 50.134
605331 98.401 -56.607 50.105
605332 98.435 -56.616 50.077
605333 98.469 -56.624 50.049
605334 98.505 -56.631 50.022
605335 98.545 -56.639 49.996
605336 98.585 -56.646 49.969
605337 98.625 -56.653 49.943
605338 98.665 -56.660 49.918
605339 98.705 -56.668 49.892
605340 98.743 -56.675 49.866
605341 98.780 -56.683 49.840
605342 98.816 -56.691 49.816
605343 98.852 -56.699 49.792
605344 98.888 -56.707 49.770
605345 98.925 -56.714 49.749
605346 98.961 -56.723 49.725
605347 98.997 -56.730 49.700
605348 99.033 -56.739 49.676
605349 99.070 -56.746 49.653
605350 99.106 -56.755 49.629
605351 99.142 -56.762 49.603
605352 99.179 -56.771 49.575
605353 99.215 -56.778 49.550
605354 99.251 -56.787 49.526
605355 99.287 -56.794 49.501
605356 99.324 -56.802 49.477
605357 99.360 -56.810 49.454
605358 99.395 -56.818 49.431
605359 99.423 -56.826 49.408
605360 99.451 -56.834 49.384
605361 99.480 -56.842 49.361
605362 99.505 -56.848 49.338
605363 99.529 -56.855 49.313
605364 99.554 -56.861 49.288
605365 99.579 -56.867 49.263
605366 99.604 -56.874 49.236
605367 99.632 -56.884 49.209
605368 99.656 -56.899 49.181
605369 99.682 -56.917 49.151
605370 99.707 -56.935 49.120
605371 99.733 -56.953 49.089
605372 99.759 -56.971 49.058
605373 99.784 -56.988 49.028
605374 99.809 -57.006 48.998
605375 99.835 -57.024 48.968
605376 99.861 -57.042 48.937
605377 99.887 -57.060 48.908
605378 99.912 -57.078 48.880
605379 99.937 -57.096 48.852
605380 99.963 -57.113 48.825
605381 99.989 -57.131 48.799
605382 100.014 -57.149 48.773
605716 88.523 -52.814 57.412
605717 88.561 -52.834 57.412
605718 88.600 -52.853 57.411
605719 88.637 -52.872 57.413
605720 88.676 -52.891 57.414
605721 88.713 -52.910 57.414
605722 88.752 -52.930 57.414
605723 88.789 -52.949 57.413
605724 88.828 -52.969 57.411
605725 88.865 -52.988 57.410
605726 88.904 -53.007 57.408
605727 88.926 -53.039 57.405
605728 88.955 -53.065 57.401
605729 88.986 -53.088 57.396
605730 89.018 -53.111 57.391
605731 89.049 -53.133 57.385
605732 89.081 -53.156 57.379
605733 89.113 -53.179 57.373
605734 89.144 -53.201 57.367
605735 89.176 -53.224 57.361
605736 89.208 -53.247 57.355
605737 89.239 -53.270 57.347
605738 89.271 -53.292 57.339
605739 89.303 -53.315 57.331
605740 89.334 -53.338 57.324
605741 89.356 -53.371 57.316
605742 89.379 -53.403 57.308
605743 89.402 -53.435 57.299
605744 89.424 -53.468 57.290
605745 89.447 -53.500 57.280
605746 89.477 -53.523 57.270
605747 89.516 -53.534 57.259
605748 89.555 -53.544 57.250
605749 89.607 -53.542 57.237
605750 89.632 -53.554 57.223
605751 89.658 -53.565 57.209
605752 89.683 -53.577 57.194
605753 89.708 -53.589 57.177
605754 89.734 -53.601 57.161
605755 89.760 -53.612 57.141
605756 89.788 -53.626 57.121
605757 89.822 -53.643 57.097
605758 89.856 -53.661 57.067
605759 89.891 -53.678 57.034
605760 89.925 -53.697 56.998
605761 89.959 -53.714 56.958
605762 89.993 -53.731 56.918
605763 90.028 -53.749 56.877
605764 90.062 -53.766 56.837
605765 90.097 -53.784 56.798
605766 90.131 -53.801 56.759
605767 90.165 -53.819 56.718
605768 90.199 -53.837 56.680
605769 90.242 -53.848 56.644
605770 90.285 -53.860 56.608
605777 90.523 -53.960 56.195
605778 90.540 -53.981 56.164
605779 90.557 -54.001 56.132
605780 90.575 -54.021 56.101
605781 90.592 -54.041 56.073
605782 90.609 -54.061 56.046
605783 90.626 -54.082 56.020
605784 90.644 -54.101 55.997
605785 90.672 -54.110 55.973
605786 90.708 -54.112 55.951
605787 90.743 -54.113 55.932
605788 90.777 -54.122 55.917
605789 90.806 -54.142 55.901
605790 90.835 -54.162 55.882
605791 90.864 -54.182 55.854
605792 90.893 -54.202 55.823
605793 90.922 -54.223 55.793
605794 90.951 -54.242 55.767
605795 90.979 -54.263 55.744
605796 91.009 -54.283 55.723
605797 91.037 -54.303 55.703
605798 91.067 -54.323 55.685
605799 91.095 -54.343 55.669
605800 91.124 -54.364 55.655
605801 91.153 -54.383 55.644
605802 91.182 -54.404 55.635
605803 91.212 -54.424 55.632
605804 91.245 -54.444 55.631
605805 91.279 -54.464 55.633
605806 91.312 -54.483 55.634
605807 91.354 -54.496 55.632
605808 91.406 -54.502 55.631
605809 91.459 -54.507 55.626
605810 91.512 -54.512 55.616
605811 91.565 -54.518 55.598
605812 91.618 -54.522 55.575
605813 91.671 -54.527 55.549
605814 91.723 -54.533 55.525
605815 91.776 -54.538 55.505
605816 91.828 -54.543 55.483
605817 91.865 -54.555 55.459
605818 91.900 -54.565 55.438
605819 91.936 -54.577 55.419
605820 91.972 -54.588 55.402
605821 92.008 -54.599 55.390
605822 92.043 -54.610 55.378
605823 92.075 -54.621 55.368
605824 92.107 -54.631 55.357
605825 92.140 -54.642 55.347
605826 92.172 -54.652 55.337
605827 92.204 -54.663 55.330
605828 92.237 -54.673 55.323
605829 92.269 -54.684 55.315
605830 92.301 -54.694 55.307
605831 92.334 -54.705 55.298
605832 92.366 -54.715 55.291
605833 92.398 -54.726 55.288
605834 92.431 -54.740 55.286
605835 92.465 -54.760 55.284
605836 92.499 -54.780 55.285
605837 92.532 -54.793 55.287
605838 92.564 -54.799 55.291
605839 92.595 -54.805 55.301
605840 92.627 -54.811 55.313
605841 92.658 -54.817 55.329
605842 92.688 -54.828 55.352
605843 92.715 -54.846 55.379
605844 92.741 -54.861 55.403
605845 92.766 -54.875 55.432
605846 92.791 -54.893 55.462
605847 92.817 -54.911 55.490
605848 92.843 -54.928 55.517
605849 92.869 -54.947 55.540
605850 92.895 -54.965 55.557
605851 92.921 -54.983 55.571
605852 92.947 -55.001 55.582
605853 92.973 -55.019 55.588
605854 93.000 -55.038 55.590
605855 93.036 -55.055 55.588
605856 93.071 -55.073 55.578
605857 93.107 -55.091 55.564
605858 93.142 -55.110 55.553
605859 93.178 -55.127 55.537
605860 93.213 -55.145 55.516
605861 93.249 -55.163 55.498
605862 93.284 -55.181 55.485
605863 93.320 -55.199 55.467
605864 93.355 -55.217 55.439
605865 93.391 -55.235 55.405
605866 93.427 -55.253 55.371
605867 93.462 -55.271 55.345
605868 93.498 -55.289 55.335
605869 93.533 -55.307 55.329
605870 93.557 -55.320 55.319
605871 93.582 -55.334 55.310
605872 93.608 -55.348 55.300
605873 93.631 -55.367 55.290
605874 93.651 -55.390 55.279
605875 93.671 -55.414 55.266
605876 93.691 -55.439 55.253
605877 93.711 -55.463 55.243
605878 93.732 -55.487 55.234
605879 93.752 -55.511 55.224
605880 93.772 -55.536 55.209
605881 93.792 -55.559 55.184
605882 93.813 -55.583 55.153
605883 93.833 -55.608 55.117
605884 93.853 -55.632 55.082
605885 93.873 -55.656 55.048
605886 93.894 -55.680 55.015
605887 93.914 -55.704 54.978
605888 93.934 -55.728 54.942
605889 93.954 -55.752 54.906
605890 93.975 -55.775 54.871
605891 94.000 -55.793 54.836
605892 94.033 -55.800 54.799
605893 94.065 -55.807 54.757
605894 94.098 -55.815 54.712
605895 94.130 -55.821 54.670
605896 94.163 -55.828 54.630
605897 94.195 -55.835 54.592
605898 94.227 -55.842 54.555
605899 94.260 -55.849 54.513
605900 94.292 -55.856 54.475
605901 94.325 -55.863 54.440
605902 94.357 -55.870 54.407
605903 94.391 -55.878 54.374
605904 94.424 -55.886 54.342
605905 94.465 -55.896 54.312
605906 94.516 -55.908 54.295
605907 94.566 -55.919 54.280
605908 94.616 -55.931 54.262
605909 94.667 -55.942 54.239
605910 94.717 -55.954 54.209
605911 94.757 -55.963 54.171
605912 94.790 -55.971 54.128
605913 94.825 -55.978 54.081
605914 94.858 -55.984 54.031
605915 94.892 -55.987 53.979
605916 94.926 -55.990 53.922
605917 94.960 -55.994 53.860
605918 94.994 -55.997 53.797
605919 95.028 -56.000 53.735
605920 95.061 -56.003 53.672
605921 95.099 -56.022 53.608
605922 95.137 -56.045 53.539
605923 95.157 -56.055 53.468
605924 95.175 -56.064 53.397
605925 95.194 -56.074 53.330
605926 95.213 -56.088 53.268
605927 95.233 -56.106 53.210
605928 95.265 -56.118 53.153
605929 95.300 -56.129 53.094
605930 95.334 -56.141 53.036
605931 95.369 -56.152 52.976
605932 95.403 -56.164 52.916
605933 95.438 -56.175 52.855
605934 95.471 -56.191 52.794
605935 95.505 -56.206 52.737
605936 95.538 -56.221 52.681
605937 95.571 -56.236 52.626
605938 95.608 -56.247 52.568
605939 95.645 -56.256 52.511
605940 95.678 -56.261 52.452
605941 95.709 -56.264 52.391
605942 95.738 -56.278 52.326
605943 95.764 -56.295 52.256
605944 95.790 -56.313 52.190
605945 95.816 -56.331 52.126
605946 95.842 -56.348 52.063
605947 95.868 -56.366 52.003
605948 95.894 -56.376 51.945
605949 95.919 -56.379 51.888
605950 95.944 -56.382 51.832
605951 95.969 -56.385 51.777
605952 95.994 -56.388 51.725
605953 96.019 -56.391 51.676
605954 96.044 -56.394 51.629
605955 96.070 -56.397 51.584
605956 96.101 -56.399 51.539
605957 96.132 -56.401 51.496
605958 96.164 -56.403 51.456
605959 96.196 -56.405 51.420
605960 96.228 -56.407 51.385
605961 96.260 -56.409 51.350
605962 96.292 -56.411 51.317
605963 96.313 -56.413 51.286
605964 96.328 -56.414 51.256
605965 96.344 -56.415 51.229
605966 96.359 -56.416 51.202
605967 96.375 -56.417 51.175
605968 96.390 -56.418 51.147
605969 96.405 -56.419 51.119
605970 96.421 -56.420 51.093
605971 96.437 -56.421 51.069
605972 96.452 -56.423 51.045
605973 96.467 -56.423 51.022
605974 96.482 -56.425 50.998
605975 96.498 -56.426 50.973
605976 96.514 -56.429 50.949
605977 96.537 -56.435 50.927
605978 96.559 -56.439 50.904
605979 96.581 -56.445 50.878
605980 96.602 -56.449 50.853
605981 96.624 -56.455 50.826
605982 96.646 -56.461 50.801
605983 96.668 -56.467 50.778
605984 96.690 -56.472 50.753
605985 96.712 -56.477 50.726
605986 96.740 -56.470 50.698
605987 96.770 -56.458 50.672
605988 96.814 -56.454 50.648
605989 96.859 -56.452 50.627
605990 96.902 -56.450 50.605
605991 96.947 -56.448 50.585
605992 96.991 -56.447 50.568
605993 97.025 -56.447 50.550
605994 97.058 -56.448 50.535
605995 97.092 -56.448 50.522
605996 97.125 -56.448 50.508
605997 97.159 -56.448 50.495
605998 97.192 -56.449 50.482
605999 97.226 -56.445 50.468
606000 97.264 -56.448 50.457
606001 97.299 -56.448 50.447
606002 97.333 -56.448 50.439
606003 97.367 -56.449 50.431
606004 97.402 -56.449 50.424
606005 97.436 -56.450 50.418
606006 97.470 -56.450 50.412
606007 97.505 -56.451 50.407
606008 97.539 -56.451 50.402
606009 97.574 -56.451 50.398
606010 97.608 -56.451 50.396
606011 97.642 -56.452 50.392
606012 97.676 -56.453 50.385
606013 97.711 -56.454 50.374
606014 97.747 -56.460 50.362
606015 97.783 -56.466 50.349
606016 97.819 -56.472 50.338
606017 97.855 -56.478 50.331
606018 97.891 -56.484 50.325
606019 97.927 -56.490 50.317
606020 97.963 -56.496 50.309
606021 98.000 -56.502 50.300
606022 98.035 -56.508 50.288
606023 98.072 -56.515 50.273
606024 98.108 -56.521 50.257
606025 98.144 -56.527 50.239
606026 98.180 -56.533 50.220
606027 98.216 -56.539 50.199
606028 98.252 -56.545 50.176
606029 98.288 -56.552 50.150
606030 98.324 -56.558 50.123
606031 98.360 -56.564 50.095
606032 98.397 -56.570 50.067
606033 98.436 -56.579 50.039
606034 98.470 -56.586 50.010
606035 98.505 -56.593 49.983
606036 98.539 -56.601 49.957
606037 98.574 -56.608 49.929
606038 98.610 -56.616 49.903
606039 98.647 -56.624 49.877
606040 98.683 -56.632 49.852
606041 98.719 -56.640 49.825
606042 98.756 -56.648 49.801
606043 98.792 -56.656 49.776
606044 98.828 -56.664 49.753
606045 98.864 -56.672 49.731
606046 98.901 -56.680 49.710
606047 98.937 -56.687 49.686
606048 98.973 -56.696 49.661
606049 99.009 -56.703 49.637
606050 99.046 -56.711 49.614
606051 99.082 -56.719 49.591
606052 99.118 -56.727 49.565
606053 99.154 -56.735 49.537
606054 99.191 -56.743 49.514
606055 99.227 -56.751 49.491
606056 99.263 -56.759 49.468
606057 99.299 -56.767 49.445
606058 99.336 -56.775 49.423
606059 99.372 -56.783 49.402
606060 99.408 -56.791 49.380
606061 99.444 -56.799 49.357
606062 99.475 -56.806 49.335
606063 99.503 -56.815 49.312
606064 99.532 -56.822 49.288
606065 99.560 -56.830 49.264
606066 99.589 -56.838 49.239
606067 99.616 -56.846 49.214
606068 99.642 -56.857 49.188
606069 99.667 -56.875 49.161
606070 99.693 -56.893 49.132
606071 99.718 -56.911 49.101
606072 99.744 -56.929 49.071
606073 99.770 -56.946 49.041
606074 99.795 -56.965 49.011
606075 99.821 -56.982 48.982
606076 99.846 -57.000 48.953
606077 99.871 -57.018 48.923
606078 99.897 -57.036 48.894
606079 99.923 -57.054 48.867
606080 99.948 -57.072 48.841
606081 99.974 -57.090 48.814
606082 99.999 -57.107 48.788
606083 100.025 -57.125 48.763
606417 88.551 -52.784 57.407
606418 88.589 -52.803 57.407
606419 88.627 -52.822 57.405
606420 88.665 -52.841 57.407
606421 88.703 -52.860 57.408
606422 88.741 -52.880 57.408
606423 88.779 -52.899 57.407
606424 88.817 -52.919 57.406
606425 88.855 -52.938 57.404
606426 88.893 -52.957 57.402
606427 88.907 -52.996 57.400
606428 88.939 -53.019 57.396
606429 88.971 -53.042 57.392
606430 89.002 -53.064 57.387
606431 89.034 -53.087 57.381
606432 89.065 -53.110 57.375
606433 89.097 -53.132 57.370
606434 89.129 -53.154 57.364
606435 89.161 -53.177 57.358
606436 89.192 -53.200 57.352
606437 89.224 -53.223 57.346
606438 89.255 -53.245 57.339
606439 89.287 -53.268 57.331
606440 89.319 -53.291 57.324
606441 89.349 -53.315 57.316
606442 89.372 -53.348 57.308
606443 89.394 -53.380 57.301
606444 89.417 -53.412 57.293
606445 89.440 -53.445 57.284
606446 89.463 -53.477 57.275
606447 89.501 -53.488 57.265
606448 89.540 -53.499 57.255
606449 89.578 -53.510 57.244
606450 89.619 -53.518 57.232
606451 89.649 -53.528 57.218
606452 89.674 -53.540 57.203
606453 89.699 -53.552 57.188
606454 89.725 -53.564 57.173
606455 89.750 -53.575 57.157
606456 89.776 -53.587 57.139
606457 89.802 -53.599 57.120
606458 89.831 -53.614 57.098
606459 89.866 -53.631 57.071
606460 89.900 -53.649 57.038
606461 89.934 -53.666 57.001
606462 89.968 -53.684 56.963
606463 90.003 -53.702 56.923
606464 90.037 -53.719 56.883
606465 90.072 -53.737 56.843
606466 90.106 -53.754 56.804
606467 90.140 -53.772 56.766
606468 90.174 -53.789 56.726
606469 90.214 -53.803 56.688
606470 90.257 -53.814 56.650
606471 90.300 -53.825 56.615
606479 90.548 -53.949 56.168
606480 90.565 -53.969 56.136
606481 90.582 -53.989 56.104
606482 90.599 -54.010 56.075
606483 90.617 -54.029 56.047
606484 90.634 -54.050 56.021
606485 90.660 -54.062 55.997
606486 90.695 -54.063 55.972
606487 90.731 -54.065 55.949
606488 90.766 -54.066 55.929
606489 90.796 -54.085 55.911
606490 90.825 -54.105 55.894
606491 90.854 -54.126 55.873
606492 90.883 -54.145 55.845
606493 90.912 -54.166 55.815
606494 90.941 -54.185 55.786
606495 90.970 -54.206 55.761
606496 90.999 -54.226 55.739
606497 91.028 -54.246 55.718
606498 91.057 -54.267 55.698
606499 91.086 -54.286 55.681
606500 91.114 -54.307 55.665
606501 91.144 -54.327 55.651
606502 91.172 -54.347 55.641
606503 91.203 -54.367 55.632
606504 91.236 -54.387 55.629
606505 91.270 -54.407 55.630
606506 91.303 -54.427 55.632
606507 91.336 -54.446 55.632
606508 91.375 -54.462 55.632
606509 91.428 -54.467 55.627
606510 91.481 -54.472 55.618
606511 91.534 -54.477 55.604
606512 91.587 -54.483 55.585
606513 91.640 -54.488 55.562
606514 91.693 -54.493 55.538
606515 91.745 -54.498 55.515
606516 91.798 -54.503 55.495
606517 91.840 -54.512 55.473
606518 91.876 -54.524 55.451
606519 91.912 -54.535 55.430
606520 91.947 -54.546 55.412
606521 91.984 -54.557 55.397
606522 92.019 -54.568 55.384
606523 92.056 -54.579 55.373
606524 92.091 -54.590 55.363
606525 92.127 -54.602 55.353
606526 92.163 -54.612 55.342
606527 92.198 -54.624 55.332
606528 92.230 -54.634 55.325
606529 92.263 -54.645 55.318
606530 92.295 -54.656 55.311
606531 92.328 -54.666 55.303
606532 92.360 -54.677 55.295
606533 92.392 -54.687 55.288
606534 92.424 -54.697 55.286
606535 92.456 -54.708 55.284
606536 92.490 -54.723 55.284
606537 92.523 -54.743 55.285
606538 92.557 -54.759 55.289
606539 92.588 -54.765 55.294
606540 92.620 -54.771 55.305
606541 92.652 -54.777 55.318
606542 92.681 -54.789 55.338
606543 92.708 -54.806 55.362
606544 92.735 -54.824 55.390
606545 92.761 -54.840 55.415
606546 92.786 -54.853 55.444
606547 92.810 -54.867 55.474
606548 92.836 -54.884 55.502
606549 92.862 -54.903 55.526
606550 92.888 -54.921 55.546
606551 92.914 -54.938 55.561
606552 92.940 -54.956 55.577
606553 92.966 -54.975 55.589
606554 92.992 -54.993 55.594
606555 93.018 -55.011 55.594
606556 93.051 -55.029 55.590
606557 93.087 -55.047 55.577
606558 93.122 -55.065 55.563
606559 93.158 -55.083 55.551
606560 93.194 -55.101 55.534
606561 93.229 -55.119 55.512
606562 93.265 -55.137 55.491
606563 93.300 -55.155 55.481
606564 93.336 -55.173 55.462
606565 93.371 -55.191 55.435
606566 93.407 -55.209 55.404
606567 93.442 -55.227 55.377
606568 93.478 -55.245 55.363
606569 93.513 -55.263 55.364
606570 93.539 -55.276 55.356
606571 93.564 -55.290 55.342
606572 93.589 -55.304 55.326
606573 93.614 -55.318 55.308
606574 93.638 -55.333 55.292
606575 93.659 -55.358 55.282
606576 93.679 -55.382 55.270
606577 93.699 -55.406 55.257
606578 93.719 -55.430 55.247
606579 93.740 -55.454 55.237
606580 93.760 -55.478 55.227
606581 93.780 -55.502 55.213
606582 93.800 -55.527 55.188
606583 93.821 -55.551 55.157
606584 93.841 -55.574 55.124
606585 93.861 -55.599 55.091
606586 93.881 -55.623 55.059
606587 93.902 -55.647 55.028
606588 93.922 -55.671 54.993
606589 93.942 -55.696 54.957
606590 93.962 -55.720 54.923
606591 93.982 -55.743 54.888
606592 94.008 -55.761 54.853
606593 94.039 -55.771 54.817
606594 94.069 -55.780 54.774
606595 94.100 -55.789 54.729
606596 94.132 -55.796 54.689
606597 94.165 -55.803 54.651
606598 94.197 -55.809 54.613
606599 94.230 -55.816 54.575
606600 94.262 -55.823 54.534
606601 94.294 -55.830 54.493
606602 94.327 -55.837 54.457
606603 94.359 -55.844 54.423
606604 94.393 -55.851 54.391
606605 94.426 -55.860 54.360
606606 94.472 -55.871 54.329
606607 94.522 -55.882 54.312
606608 94.572 -55.893 54.299
606609 94.623 -55.906 54.285
606610 94.674 -55.917 54.266
606611 94.724 -55.928 54.240
606612 94.766 -55.938 54.206
606613 94.801 -55.946 54.165
606614 94.834 -55.953 54.118
606615 94.868 -55.956 54.067
606616 94.902 -55.960 54.013
606617 94.936 -55.963 53.954
606618 94.970 -55.966 53.889
606619 95.003 -55.969 53.823
606620 95.037 -55.973 53.757
606621 95.071 -55.976 53.692
606622 95.108 -55.992 53.626
606623 95.139 -56.010 53.558
606624 95.158 -56.020 53.485
606625 95.177 -56.030 53.411
606626 95.195 -56.047 53.342
606627 95.214 -56.066 53.276
606628 95.240 -56.081 53.212
606629 95.273 -56.094 53.152
606630 95.306 -56.105 53.090
606631 95.340 -56.116 53.030
606632 95.375 -56.128 52.969
606633 95.410 -56.140 52.908
606634 95.444 -56.151 52.846
606635 95.479 -56.163 52.784
606636 95.514 -56.174 52.724
606637 95.548 -56.185 52.666
606638 95.583 -56.197 52.611
606639 95.616 -56.213 52.555
606640 95.650 -56.226 52.499
606641 95.685 -56.233 52.443
606642 95.714 -56.244 52.386
606643 95.741 -56.261 52.323
606644 95.766 -56.279 52.258
606645 95.793 -56.297 52.190
606646 95.819 -56.314 52.125
606647 95.844 -56.332 52.061
606648 95.871 -56.349 52.000
606649 95.896 -56.355 51.942
606650 95.921 -56.358 51.882
606651 95.946 -56.361 51.824
606652 95.971 -56.365 51.767
606653 95.996 -56.368 51.713
606654 96.022 -56.371 51.662
606655 96.051 -56.373 51.614
606656 96.083 -56.376 51.567
606657 96.115 -56.377 51.521
606658 96.147 -56.379 51.477
606659 96.178 -56.381 51.436
606660 96.210 -56.383 51.399
606661 96.242 -56.385 51.363
606662 96.274 -56.387 51.327
606663 96.306 -56.389 51.294
606664 96.323 -56.391 51.261
606665 96.339 -56.392 51.231
606666 96.354 -56.393 51.204
606667 96.370 -56.394 51.177
606668 96.385 -56.395 51.149
606669 96.400 -56.396 51.121
606670 96.416 -56.397 51.092
606671 96.432 -56.398 51.066
606672 96.447 -56.399 51.041
606673 96.462 -56.401 51.016
606674 96.477 -56.401 50.992
606675 96.493 -56.403 50.967
606676 96.509 -56.404 50.940
606677 96.522 -56.405 50.916
606678 96.535 -56.407 50.893
606679 96.557 -56.412 50.868
606680 96.579 -56.417 50.842
606681 96.601 -56.423 50.816
606682 96.622 -56.429 50.789
606683 96.645 -56.435 50.764
606684 96.666 -56.440 50.741
606685 96.688 -56.445 50.717
606686 96.712 -56.447 50.690
606687 96.742 -56.434 50.663
606688 96.784 -56.429 50.637
606689 96.828 -56.427 50.613
606690 96.872 -56.425 50.591
606691 96.916 -56.423 50.570
606692 96.954 -56.421 50.550
606693 96.988 -56.420 50.530
606694 97.022 -56.420 50.512
606695 97.055 -56.419 50.497
606696 97.089 -56.418 50.483
606697 97.123 -56.417 50.470
606698 97.157 -56.417 50.457
606699 97.190 -56.416 50.442
606700 97.224 -56.415 50.429
606701 97.258 -56.415 50.417
606702 97.293 -56.415 50.406
606703 97.327 -56.416 50.396
606704 97.361 -56.416 50.387
606705 97.395 -56.417 50.379
606706 97.430 -56.417 50.371
606707 97.464 -56.417 50.365
606708 97.499 -56.417 50.359
606709 97.533 -56.418 50.353
606710 97.567 -56.418 50.349
606711 97.601 -56.419 50.345
606712 97.636 -56.420 50.339
606713 97.670 -56.420 50.335
606714 97.705 -56.420 50.328
606715 97.740 -56.426 50.317
606716 97.777 -56.432 50.304
606717 97.813 -56.438 50.294
606718 97.849 -56.444 50.285
606719 97.885 -56.451 50.279
606720 97.921 -56.457 50.272
606721 97.957 -56.463 50.264
606722 97.993 -56.469 50.255
606723 98.029 -56.475 50.246
606724 98.065 -56.481 50.232
606725 98.101 -56.487 50.216
606726 98.138 -56.493 50.200
606727 98.173 -56.499 50.181
606728 98.210 -56.506 50.161
606729 98.249 -56.514 50.138
606730 98.290 -56.522 50.111
606731 98.330 -56.531 50.084
606732 98.370 -56.539 50.056
606733 98.411 -56.548 50.028
606734 98.446 -56.555 49.999
606735 98.480 -56.563 49.970
606736 98.515 -56.570 49.942
606737 98.549 -56.577 49.914
606738 98.584 -56.585 49.887
606739 98.618 -56.592 49.859
606740 98.653 -56.599 49.832
606741 98.687 -56.607 49.806
606742 98.722 -56.614 49.780
606743 98.756 -56.621 49.756
606744 98.791 -56.629 49.733
606745 98.825 -56.636 49.710
606746 98.860 -56.643 49.687
606747 98.894 -56.651 49.667
606748 98.929 -56.658 49.645
606749 98.963 -56.665 49.620
606750 98.998 -56.673 49.596
606751 99.032 -56.680 49.572
606752 99.067 -56.687 49.548
606753 99.101 -56.695 49.523
606754 99.136 -56.702 49.497
606755 99.170 -56.709 49.475
606756 99.205 -56.717 49.453
606757 99.239 -56.724 49.433
606758 99.275 -56.732 49.413
606759 99.311 -56.740 49.392
606760 99.348 -56.748 49.372
606761 99.384 -56.755 49.351
606762 99.421 -56.763 49.329
606763 99.457 -56.771 49.307
606764 99.493 -56.779 49.285
606765 99.527 -56.787 49.263
606766 99.556 -56.795 49.239
606767 99.585 -56.805 49.215
606768 99.618 -56.821 49.191
606769 99.646 -56.836 49.166
606770 99.674 -56.852 49.140
606771 99.702 -56.870 49.111
606772 99.729 -56.887 49.082
606773 99.755 -56.905 49.053
606774 99.780 -56.923 49.024
606775 99.806 -56.940 48.995
606776 99.832 -56.959 48.966
606777 99.857 -56.976 48.938
606778 99.883 -56.994 48.909
606779 99.908 -57.012 48.881
606780 99.934 -57.030 48.855
606781 99.959 -57.048 48.829
606782 99.985 -57.065 48.804
606783 100.010 -57.084 48.778
607118 88.578 -52.753 57.401
607119 88.616 -52.772 57.401
607120 88.654 -52.791 57.399
607121 88.693 -52.811 57.400
607122 88.731 -52.830 57.401
607123 88.769 -52.849 57.400
607124 88.807 -52.869 57.399
607125 88.845 -52.888 57.398
607126 88.866 -52.921 57.396
607127 88.892 -52.949 57.394
607128 88.923 -52.972 57.391
607129 88.955 -52.995 57.387
607130 88.987 -53.017 57.383
607131 89.018 -53.040 57.377
607132 89.050 -53.063 57.372
607133 89.082 -53.086 57.366
607134 89.113 -53.108 57.361
607135 89.145 -53.131 57.355
607136 89.177 -53.154 57.349
607137 89.208 -53.176 57.342
607138 89.240 -53.199 57.337
607139 89.272 -53.222 57.331
607140 89.303 -53.245 57.324
607141 89.335 -53.267 57.317
607142 89.365 -53.292 57.310
607143 89.387 -53.325 57.302
607144 89.409 -53.357 57.295
607145 89.432 -53.389 57.287
607146 89.455 -53.422 57.279
607147 89.487 -53.442 57.270
607148 89.525 -53.452 57.261
607149 89.564 -53.464 57.250
607150 89.602 -53.475 57.240
607151 89.635 -53.488 57.227
607152 89.666 -53.505 57.213
607153 89.690 -53.515 57.198
607154 89.716 -53.527 57.183
607155 89.741 -53.538 57.169
607156 89.766 -53.550 57.154
607157 89.792 -53.562 57.136
607158 89.818 -53.574 57.117
607159 89.843 -53.586 57.097
607160 89.875 -53.602 57.074
607161 89.909 -53.619 57.043
607162 89.944 -53.637 57.006
607163 89.978 -53.654 56.968
607164 90.012 -53.672 56.928
607165 90.046 -53.689 56.889
607166 90.081 -53.706 56.850
607167 90.115 -53.725 56.811
607168 90.150 -53.742 56.773
607169 90.186 -53.758 56.735
607170 90.229 -53.769 56.696
607171 90.272 -53.780 56.659
607172 90.316 -53.791 56.623
607173 90.359 -53.803 56.590
607181 90.573 -53.938 56.140
607182 90.590 -53.958 56.108
607183 90.607 -53.978 56.078
607184 90.624 -53.998 56.049
607185 90.647 -54.013 56.023
607186 90.683 -54.015 55.998
607187 90.718 -54.016 55.973
607188 90.754 -54.018 55.949
607189 90.787 -54.029 55.927
607190 90.815 -54.048 55.906
607191 90.845 -54.069 55.887
607192 90.873 -54.088 55.864
607193 90.902 -54.109 55.836
607194 90.931 -54.129 55.808
607195 90.960 -54.149 55.781
607196 90.989 -54.170 55.756
607197 91.018 -54.189 55.734
607198 91.047 -54.210 55.714
607199 91.076 -54.229 55.695
607200 91.105 -54.250 55.677
607201 91.134 -54.270 55.661
607202 91.163 -54.290 55.647
607203 91.194 -54.311 55.637
607204 91.228 -54.330 55.629
607205 91.261 -54.350 55.628
607206 91.294 -54.370 55.630
607207 91.328 -54.389 55.632
607208 91.361 -54.410 55.633
607209 91.397 -54.427 55.632
607210 91.450 -54.433 55.625
607211 91.503 -54.438 55.612
607212 91.556 -54.443 55.595
607213 91.608 -54.448 55.574
607214 91.661 -54.453 55.551
607215 91.714 -54.458 55.528
607216 91.767 -54.464 55.507
607217 91.815 -54.471 55.486
607218 91.851 -54.481 55.465
607219 91.887 -54.493 55.444
607220 91.923 -54.504 55.424
607221 91.959 -54.515 55.408
607222 91.995 -54.526 55.392
607223 92.030 -54.537 55.380
607224 92.067 -54.549 55.368
607225 92.102 -54.559 55.358
607226 92.138 -54.571 55.348
607227 92.174 -54.582 55.337
607228 92.210 -54.593 55.326
607229 92.246 -54.604 55.320
607230 92.282 -54.615 55.313
607231 92.318 -54.626 55.306
607232 92.353 -54.637 55.299
607233 92.385 -54.648 55.293
607234 92.418 -54.659 55.286
607235 92.450 -54.669 55.284
607236 92.483 -54.680 55.284
607237 92.515 -54.690 55.284
607238 92.548 -54.705 55.286
607239 92.582 -54.724 55.291
607240 92.614 -54.730 55.297
607241 92.645 -54.736 55.310
607242 92.674 -54.749 55.326
607243 92.701 -54.767 55.348
607244 92.728 -54.784 55.373
607245 92.755 -54.802 55.400
607246 92.781 -54.818 55.425
607247 92.806 -54.831 55.451
607248 92.831 -54.845 55.478
607249 92.856 -54.859 55.503
607250 92.881 -54.876 55.525
607251 92.907 -54.894 55.543
607252 92.933 -54.913 55.562
607253 92.959 -54.931 55.583
607254 92.985 -54.948 55.593
607255 93.011 -54.966 55.596
607256 93.037 -54.985 55.595
607257 93.067 -55.003 55.589
607258 93.102 -55.021 55.574
607259 93.138 -55.038 55.558
607260 93.174 -55.057 55.545
607261 93.209 -55.075 55.528
607262 93.245 -55.093 55.506
607263 93.280 -55.110 55.483
607264 93.316 -55.129 55.472
607265 93.352 -55.147 55.454
607266 93.387 -55.164 55.430
607267 93.423 -55.182 55.408
607268 93.459 -55.201 55.394
607269 93.494 -55.219 55.389
607270 93.520 -55.232 55.389
607271 93.545 -55.247 55.380
607272 93.570 -55.260 55.364
607273 93.595 -55.274 55.345
607274 93.621 -55.288 55.326
607275 93.646 -55.302 55.308
607276 93.666 -55.324 55.296
607277 93.687 -55.348 55.284
607278 93.707 -55.373 55.272
607279 93.727 -55.397 55.261
607280 93.748 -55.421 55.251
607281 93.768 -55.445 55.239
607282 93.788 -55.469 55.222
607283 93.808 -55.493 55.194
607284 93.828 -55.517 55.164
607285 93.849 -55.542 55.133
607286 93.869 -55.566 55.103
607287 93.889 -55.590 55.072
607288 93.909 -55.614 55.041
607289 93.929 -55.638 55.008
607290 93.949 -55.662 54.974
607291 93.970 -55.687 54.941
607292 93.991 -55.709 54.907
607293 94.016 -55.725 54.871
607294 94.049 -55.737 54.834
607295 94.080 -55.746 54.791
607296 94.110 -55.755 54.748
607297 94.141 -55.764 54.710
607298 94.171 -55.773 54.673
607299 94.202 -55.781 54.635
607300 94.232 -55.790 54.597
607301 94.265 -55.797 54.555
607302 94.297 -55.804 54.515
607303 94.330 -55.812 54.477
607304 94.362 -55.818 54.442
607305 94.395 -55.825 54.409
607306 94.428 -55.834 54.377
607307 94.479 -55.845 54.346
607308 94.529 -55.856 54.328
607309 94.579 -55.868 54.314
607310 94.630 -55.880 54.301
607311 94.680 -55.891 54.286
607312 94.731 -55.903 54.267
607313 94.776 -55.914 54.237
607314 94.811 -55.922 54.199
607315 94.844 -55.926 54.152
607316 94.878 -55.930 54.100
607317 94.912 -55.933 54.044
607318 94.946 -55.936 53.984
607319 94.980 -55.940 53.917
607320 95.014 -55.943 53.850
607321 95.047 -55.946 53.779
607322 95.081 -55.950 53.710
607323 95.117 -55.963 53.642
607324 95.140 -55.975 53.574
607325 95.159 -55.988 53.501
607326 95.178 -56.007 53.427
607327 95.197 -56.026 53.356
607328 95.216 -56.045 53.286
607329 95.248 -56.057 53.218
607330 95.280 -56.069 53.153
607331 95.313 -56.081 53.089
607332 95.347 -56.093 53.026
607333 95.382 -56.104 52.964
607334 95.416 -56.116 52.901
607335 95.451 -56.127 52.839
607336 95.485 -56.138 52.777
607337 95.520 -56.150 52.715
607338 95.554 -56.161 52.655
607339 95.589 -56.172 52.598
607340 95.624 -56.184 52.542
607341 95.658 -56.195 52.486
607342 95.691 -56.209 52.430
607343 95.717 -56.227 52.376
607344 95.743 -56.244 52.318
607345 95.769 -56.262 52.256
607346 95.795 -56.280 52.193
607347 95.821 -56.298 52.128
607348 95.847 -56.315 52.063
607349 95.873 -56.332 52.001
607350 95.898 -56.335 51.940
607351 95.923 -56.338 51.879
607352 95.949 -56.342 51.819
607353 95.974 -56.345 51.761
607354 96.002 -56.348 51.705
607355 96.033 -56.350 51.652
607356 96.065 -56.351 51.602
607357 96.097 -56.354 51.554
607358 96.129 -56.355 51.506
607359 96.161 -56.357 51.461
607360 96.192 -56.360 51.419
607361 96.224 -56.361 51.380
607362 96.256 -56.364 51.343
607363 96.288 -56.366 51.306
607364 96.318 -56.367 51.272
607365 96.334 -56.369 51.238
607366 96.349 -56.370 51.207
607367 96.365 -56.371 51.180
607368 96.380 -56.372 51.153
607369 96.395 -56.373 51.125
607370 96.411 -56.374 51.097
607371 96.427 -56.375 51.068
607372 96.442 -56.376 51.040
607373 96.457 -56.377 51.015
607374 96.473 -56.379 50.989
607375 96.488 -56.379 50.963
607376 96.504 -56.381 50.937
607377 96.518 -56.382 50.910
607378 96.530 -56.383 50.884
607379 96.542 -56.384 50.860
607380 96.555 -56.385 50.834
607381 96.577 -56.392 50.807
607382 96.599 -56.397 50.780
607383 96.621 -56.402 50.753
607384 96.643 -56.408 50.728
607385 96.665 -56.414 50.704
607386 96.686 -56.420 50.681
607387 96.715 -56.410 50.655
607388 96.753 -56.404 50.629
607389 96.798 -56.402 50.604
607390 96.842 -56.400 50.580
607391 96.884 -56.398 50.557
607392 96.918 -56.397 50.536
607393 96.951 -56.396 50.515
607394 96.985 -56.395 50.495
607395 97.018 -56.395 50.476
607396 97.052 -56.394 50.461
607397 97.086 -56.393 50.446
607398 97.119 -56.392 50.433
607399 97.153 -56.391 50.419
607400 97.187 -56.390 50.404
607401 97.221 -56.389 50.391
607402 97.254 -56.388 50.378
607403 97.288 -56.386 50.367
607404 97.322 -56.385 50.356
607405 97.356 -56.385 50.345
607406 97.389 -56.383 50.336
607407 97.424 -56.384 50.327
607408 97.458 -56.384 50.319
607409 97.493 -56.385 50.312
607410 97.527 -56.385 50.306
607411 97.561 -56.385 50.301
607412 97.595 -56.385 50.296
607413 97.630 -56.386 50.289
607414 97.664 -56.387 50.283
607415 97.699 -56.387 50.276
607416 97.734 -56.392 50.267
607417 97.770 -56.398 50.255
607418 97.806 -56.404 50.247
607419 97.843 -56.410 50.240
607420 97.878 -56.417 50.233
607421 97.915 -56.423 50.226
607422 97.951 -56.429 50.218
607423 97.987 -56.435 50.208
607424 98.023 -56.441 50.200
607425 98.060 -56.448 50.188
607426 98.101 -56.457 50.173
607427 98.141 -56.465 50.156
607428 98.181 -56.473 50.138
607429 98.222 -56.483 50.117
607430 98.263 -56.491 50.094
607431 98.303 -56.499 50.067
607432 98.343 -56.508 50.040
607433 98.383 -56.517 50.012
607434 98.421 -56.524 49.983
607435 98.456 -56.532 49.954
607436 98.490 -56.539 49.925
607437 98.525 -56.546 49.897
607438 98.559 -56.554 49.868
607439 98.594 -56.561 49.840
607440 98.628 -56.568 49.812
607441 98.663 -56.576 49.784
607442 98.698 -56.583 49.757
607443 98.732 -56.590 49.730
607444 98.767 -56.598 49.705
607445 98.801 -56.605 49.682
607446 98.836 -56.612 49.659
607447 98.870 -56.620 49.637
607448 98.905 -56.627 49.619
607449 98.939 -56.634 49.598
607450 98.974 -56.642 49.573
607451 99.008 -56.649 49.548
607452 99.043 -56.656 49.525
607453 99.077 -56.664 49.502
607454 99.112 -56.671 49.477
607455 99.146 -56.678 49.453
607456 99.181 -56.686 49.431
607457 99.215 -56.693 49.412
607458 99.250 -56.700 49.395
607459 99.284 -56.708 49.377
607460 99.319 -56.715 49.357
607461 99.353 -56.722 49.339
607462 99.387 -56.730 49.319
607463 99.422 -56.737 49.299
607464 99.456 -56.744 49.278
607465 99.491 -56.752 49.258
607466 99.525 -56.759 49.237
607467 99.556 -56.771 49.213
607468 99.589 -56.787 49.190
607469 99.621 -56.802 49.167
607470 99.648 -56.815 49.143
607471 99.676 -56.831 49.117
607472 99.703 -56.849 49.090
607473 99.731 -56.866 49.062
607474 99.759 -56.883 49.034
607475 99.787 -56.900 49.006
607476 99.815 -56.918 48.978
607477 99.842 -56.934 48.951
607478 99.868 -56.953 48.923
607479 99.894 -56.970 48.895
607480 99.919 -56.988 48.869
607481 99.944 -57.006 48.843
607482 99.970 -57.024 48.818
607483 99.996 -57.042 48.793
607819 88.606 -52.722 57.396
607820 88.644 -52.741 57.395
607821 88.682 -52.761 57.394
607822 88.720 -52.780 57.393
607823 88.758 -52.799 57.393
607824 88.796 -52.819 57.393
607825 88.834 -52.838 57.392
607826 88.844 -52.880 57.390
607827 88.876 -52.903 57.387
607828 88.908 -52.926 57.385
607829 88.940 -52.948 57.382
607830 88.971 -52.971 57.378
607831 89.003 -52.994 57.373
607832 89.035 -53.016 57.368
607833 89.066 -53.038 57.363
607834 89.098 -53.061 57.357
607835 89.129 -53.084 57.352
607836 89.161 -53.107 57.347
607837 89.193 -53.129 57.341
607838 89.225 -53.152 57.334
607839 89.256 -53.175 57.329
607840 89.288 -53.198 57.323
607841 89.319 -53.220 57.317
607842 89.351 -53.243 57.310
607843 89.380 -53.270 57.303
607844 89.402 -53.302 57.295
607845 89.424 -53.334 57.289
607846 89.447 -53.367 57.282
607847 89.473 -53.396 57.274
607848 89.511 -53.406 57.266
607849 89.550 -53.418 57.256
607850 89.588 -53.429 57.246
607851 89.621 -53.442 57.236
607852 89.645 -53.459 57.223
607853 89.685 -53.486 57.209
607854 89.713 -53.499 57.194
607855 89.733 -53.502 57.179
607856 89.757 -53.513 57.165
607857 89.783 -53.525 57.150
607858 89.808 -53.537 57.133
607859 89.834 -53.549 57.113
607860 89.860 -53.560 57.093
607861 89.885 -53.572 57.071
607862 89.918 -53.589 57.042
607863 89.953 -53.607 57.009
607864 89.987 -53.625 56.972
607865 90.022 -53.642 56.934
607866 90.056 -53.659 56.896
607867 90.090 -53.677 56.858
607868 90.124 -53.694 56.819
607869 90.159 -53.712 56.782
607870 90.201 -53.724 56.743
607871 90.245 -53.735 56.705
607872 90.288 -53.746 56.669
607873 90.331 -53.757 56.633
607874 90.374 -53.769 56.599
607882 90.581 -53.907 56.145
607883 90.598 -53.926 56.113
607884 90.615 -53.947 56.083
607885 90.635 -53.964 56.053
607886 90.670 -53.966 56.026
607887 90.706 -53.968 56.000
607888 90.741 -53.969 55.974
607889 90.777 -53.972 55.948
607890 90.806 -53.991 55.925
607891 90.835 -54.012 55.902
607892 90.864 -54.032 55.880
607893 90.893 -54.052 55.855
607894 90.922 -54.073 55.828
607895 90.951 -54.092 55.801
607896 90.979 -54.113 55.776
607897 91.009 -54.132 55.752
607898 91.037 -54.153 55.730
607899 91.067 -54.173 55.710
607900 91.095 -54.193 55.691
607901 91.124 -54.214 55.673
607902 91.153 -54.233 55.657
607903 91.186 -54.254 55.644
607904 91.219 -54.273 55.634
607905 91.252 -54.293 55.626
607906 91.286 -54.313 55.625
607907 91.319 -54.333 55.627
607908 91.353 -54.353 55.630
607909 91.386 -54.373 55.631
607910 91.419 -54.392 55.630
607911 91.471 -54.398 55.623
607912 91.524 -54.403 55.608
607913 91.577 -54.408 55.587
607914 91.630 -54.414 55.564
607915 91.683 -54.419 55.542
607916 91.735 -54.424 55.519
607917 91.788 -54.429 55.499
607918 91.826 -54.439 55.478
607919 91.862 -54.451 55.457
607920 91.898 -54.461 55.437
607921 91.934 -54.473 55.419
607922 91.970 -54.484 55.403
607923 92.006 -54.495 55.388
607924 92.041 -54.506 55.375
607925 92.078 -54.518 55.364
607926 92.113 -54.529 55.354
607927 92.149 -54.540 55.343
607928 92.186 -54.551 55.333
607929 92.221 -54.562 55.323
607930 92.258 -54.573 55.316
607931 92.293 -54.584 55.309
607932 92.329 -54.596 55.302
607933 92.365 -54.606 55.296
607934 92.401 -54.618 55.290
607935 92.437 -54.629 55.285
607936 92.473 -54.640 55.284
607937 92.509 -54.651 55.284
607938 92.541 -54.662 55.285
607939 92.573 -54.672 55.288
607940 92.606 -54.687 55.293
607941 92.638 -54.696 55.301
607942 92.667 -54.709 55.314
607943 92.694 -54.727 55.334
607944 92.721 -54.745 55.357
607945 92.748 -54.762 55.382
607946 92.775 -54.780 55.406
607947 92.801 -54.797 55.428
607948 92.827 -54.810 55.452
607949 92.851 -54.824 55.475
607950 92.876 -54.837 55.498
607951 92.901 -54.851 55.519
607952 92.926 -54.868 55.538
607953 92.952 -54.886 55.558
607954 92.978 -54.904 55.577
607955 93.004 -54.922 55.587
607956 93.030 -54.941 55.589
607957 93.056 -54.958 55.585
607958 93.083 -54.976 55.577
607959 93.119 -54.994 55.563
607960 93.154 -55.013 55.548
607961 93.190 -55.030 55.536
607962 93.225 -55.048 55.520
607963 93.261 -55.066 55.498
607964 93.296 -55.085 55.473
607965 93.332 -55.102 55.459
607966 93.367 -55.120 55.443
607967 93.403 -55.138 55.428
607968 93.438 -55.156 55.419
607969 93.474 -55.174 55.417
607970 93.501 -55.189 55.417
607971 93.526 -55.203 55.413
607972 93.552 -55.217 55.402
607973 93.577 -55.230 55.383
607974 93.602 -55.244 55.363
607975 93.627 -55.258 55.343
607976 93.652 -55.272 55.325
607977 93.675 -55.292 55.312
607978 93.695 -55.316 55.300
607979 93.715 -55.339 55.288
607980 93.735 -55.364 55.277
607981 93.755 -55.388 55.265
607982 93.776 -55.412 55.252
607983 93.796 -55.436 55.234
607984 93.816 -55.461 55.205
607985 93.836 -55.485 55.178
607986 93.856 -55.508 55.151
607987 93.876 -55.533 55.123
607988 93.897 -55.557 55.093
607989 93.917 -55.581 55.065
607990 93.937 -55.605 55.032
607991 93.958 -55.630 55.001
607992 93.978 -55.653 54.968
607993 94.000 -55.673 54.931
607994 94.026 -55.690 54.891
607995 94.058 -55.701 54.852
607996 94.090 -55.712 54.809
607997 94.121 -55.721 54.770
607998 94.151 -55.730 54.733
607999 94.182 -55.739 54.696
608000 94.212 -55.748 54.658
608001 94.243 -55.757 54.619
608002 94.273 -55.765 54.576
608003 94.303 -55.774 54.536
608004 94.334 -55.784 54.497
608005 94.364 -55.793 54.461
608006 94.397 -55.799 54.426
608007 94.435 -55.808 54.395
608008 94.485 -55.820 54.363
608009 94.536 -55.831 54.341
608010 94.586 -55.843 54.325
608011 94.636 -55.855 54.314
608012 94.687 -55.866 54.302
608013 94.737 -55.878 54.287
608014 94.786 -55.889 54.260
608015 94.820 -55.897 54.225
608016 94.854 -55.900 54.180
608017 94.888 -55.903 54.128
608018 94.922 -55.906 54.071
608019 94.956 -55.909 54.009
608020 94.989 -55.912 53.941
608021 95.023 -55.916 53.871
608022 95.057 -55.919 53.800
608023 95.091 -55.922 53.730
608024 95.122 -55.931 53.660
608025 95.141 -55.948 53.592
608026 95.160 -55.967 53.520
608027 95.179 -55.986 53.446
608028 95.198 -56.005 53.372
608029 95.223 -56.021 53.299
608030 95.256 -56.033 53.228
608031 95.288 -56.044 53.160
608032 95.320 -56.056 53.091
608033 95.354 -56.069 53.027
608034 95.388 -56.080 52.961
608035 95.423 -56.091 52.897
608036 95.457 -56.103 52.835
608037 95.492 -56.114 52.772
608038 95.527 -56.125 52.710
608039 95.561 -56.137 52.648
608040 95.596 -56.148 52.588
608041 95.630 -56.160 52.530
608042 95.665 -56.172 52.474
608043 95.692 -56.190 52.418
608044 95.719 -56.209 52.364
608045 95.746 -56.228 52.310
608046 95.772 -56.246 52.253
608047 95.798 -56.263 52.194
608048 95.824 -56.281 52.130
608049 95.850 -56.298 52.065
608050 95.876 -56.313 52.002
608051 95.901 -56.316 51.940
608052 95.926 -56.319 51.878
608053 95.952 -56.322 51.817
608054 95.983 -56.324 51.757
608055 96.016 -56.326 51.699
608056 96.047 -56.328 51.645
608057 96.079 -56.329 51.593
608058 96.111 -56.332 51.543
608059 96.143 -56.334 51.494
608060 96.175 -56.335 51.448
608061 96.206 -56.338 51.404
608062 96.238 -56.340 51.364
608063 96.270 -56.342 51.326
608064 96.302 -56.344 51.289
608065 96.329 -56.345 51.253
608066 96.344 -56.347 51.219
608067 96.360 -56.348 51.187
608068 96.375 -56.348 51.158
608069 96.390 -56.350 51.130
608070 96.406 -56.351 51.101
608071 96.421 -56.352 51.073
608072 96.437 -56.353 51.044
608073 96.452 -56.354 51.015
608074 96.468 -56.355 50.989
608075 96.483 -56.357 50.964
608076 96.499 -56.357 50.937
608077 96.513 -56.359 50.909
608078 96.525 -56.360 50.882
608079 96.538 -56.361 50.855
608080 96.550 -56.363 50.830
608081 96.562 -56.364 50.803
608082 96.575 -56.365 50.776
608083 96.597 -56.370 50.748
608084 96.619 -56.376 50.720
608085 96.641 -56.382 50.694
608086 96.663 -56.388 50.670
608087 96.687 -56.388 50.646
608088 96.723 -56.379 50.621
608089 96.767 -56.377 50.596
608090 96.811 -56.374 50.571
608091 96.847 -56.373 50.547
608092 96.880 -56.373 50.525
608093 96.914 -56.372 50.504
608094 96.948 -56.371 50.482
608095 96.981 -56.370 50.462
608096 97.015 -56.370 50.443
608097 97.049 -56.368 50.426
608098 97.083 -56.367 50.411
608099 97.116 -56.366 50.396
608100 97.150 -56.365 50.382
608101 97.184 -56.364 50.367
608102 97.218 -56.363 50.354
608103 97.251 -56.362 50.341
608104 97.285 -56.360 50.328
608105 97.318 -56.360 50.317
608106 97.352 -56.358 50.306
608107 97.386 -56.357 50.295
608108 97.420 -56.357 50.285
608109 97.453 -56.355 50.276
608110 97.487 -56.354 50.268
608111 97.521 -56.353 50.261
608112 97.555 -56.352 50.254
608113 97.589 -56.353 50.248
608114 97.624 -56.354 50.241
608115 97.658 -56.354 50.233
608116 97.692 -56.354 50.226
608117 97.728 -56.358 50.217
608118 97.764 -56.364 50.207
608119 97.800 -56.370 50.199
608120 97.836 -56.376 50.194
608121 97.872 -56.383 50.187
608122 97.912 -56.391 50.180
608123 97.952 -56.400 50.171
608124 97.993 -56.408 50.162
608125 98.033 -56.417 50.154
608126 98.074 -56.425 50.142
608127 98.114 -56.434 50.128
608128 98.154 -56.442 50.112
608129 98.195 -56.451 50.093
608130 98.235 -56.460 50.072
608131 98.276 -56.468 50.048
608132 98.316 -56.476 50.023
608133 98.357 -56.485 49.996
608134 98.397 -56.494 49.968
608135 98.432 -56.501 49.939
608136 98.466 -56.508 49.910
608137 98.501 -56.516 49.881
608138 98.535 -56.523 49.852
608139 98.570 -56.530 49.823
608140 98.604 -56.538 49.794
608141 98.639 -56.545 49.765
608142 98.673 -56.552 49.737
608143 98.708 -56.560 49.710
608144 98.742 -56.567 49.682
608145 98.777 -56.574 49.657
608146 98.811 -56.582 49.634
608147 98.846 -56.589 49.611
608148 98.880 -56.596 49.588
608149 98.915 -56.604 49.570
608150 98.949 -56.611 49.549
608151 98.984 -56.618 49.525
608152 99.018 -56.626 49.500
608153 99.053 -56.633 49.477
608154 99.087 -56.640 49.454
608155 99.122 -56.648 49.431
608156 99.156 -56.655 49.409
608157 99.191 -56.662 49.388
608158 99.225 -56.670 49.371
608159 99.260 -56.677 49.355
608160 99.294 -56.684 49.339
608161 99.329 -56.692 49.321
608162 99.363 -56.699 49.303
608163 99.398 -56.706 49.286
608164 99.432 -56.714 49.267
608165 99.467 -56.721 49.247
608166 99.501 -56.728 49.228
608167 99.530 -56.736 49.208
608168 99.560 -56.752 49.186
608169 99.593 -56.768 49.165
608170 99.623 -56.781 49.143
608171 99.649 -56.793 49.119
608172 99.678 -56.810 49.095
608173 99.705 -56.827 49.069
608174 99.733 -56.845 49.043
608175 99.761 -56.862 49.016
608176 99.789 -56.879 48.989
608177 99.817 -56.896 48.962
608178 99.844 -56.913 48.935
608179 99.872 -56.931 48.909
608180 99.900 -56.948 48.882
608181 99.928 -56.965 48.858
608182 99.956 -56.982 48.832
608183 99.981 -57.000 48.807
608184 100.007 -57.018 48.783
608521 88.672 -52.711 57.389
608522 88.710 -52.730 57.388
608523 88.748 -52.750 57.387
608524 88.786 -52.769 57.386
608525 88.806 -52.803 57.385
608526 88.829 -52.833 57.385
608527 88.860 -52.856 57.382
608528 88.892 -52.879 57.378
608529 88.924 -52.901 57.376
608530 88.956 -52.924 57.373
608531 88.987 -52.947 57.369
608532 89.019 -52.970 57.364
608533 89.051 -52.992 57.359
608534 89.082 -53.015 57.353
608535 89.114 -53.038 57.348
608536 89.145 -53.060 57.343
608537 89.177 -53.083 57.338
608538 89.209 -53.106 57.332
608539 89.241 -53.129 57.326
608540 89.272 -53.151 57.320
608541 89.304 -53.173 57.315
608542 89.336 -53.196 57.310
608543 89.367 -53.219 57.304
608544 89.395 -53.246 57.297
608545 89.417 -53.279 57.290
608546 89.440 -53.311 57.285
608547 89.462 -53.344 57.278
608548 89.497 -53.360 57.271
608549 89.535 -53.371 57.263
608550 89.574 -53.383 57.254
608551 89.607 -53.396 57.243
608552 89.630 -53.413 57.232
608553 89.654 -53.430 57.218
608554 89.695 -53.458 57.204
608555 89.737 -53.484 57.191
608556 89.761 -53.494 57.176
608557 89.780 -53.496 57.163
608558 89.799 -53.499 57.147
608559 89.825 -53.512 57.129
608560 89.850 -53.523 57.110
608561 89.876 -53.535 57.089
608562 89.901 -53.547 57.068
608563 89.928 -53.559 57.041
608564 89.962 -53.577 57.012
608565 89.996 -53.594 56.977
608566 90.031 -53.612 56.941
608567 90.065 -53.630 56.904
608568 90.099 -53.647 56.866
608569 90.134 -53.665 56.829
608570 90.173 -53.678 56.792
608571 90.217 -53.690 56.754
608572 90.260 -53.701 56.716
608573 90.303 -53.712 56.679
608574 90.346 -53.723 56.643
608575 90.389 -53.734 56.609
608583 90.588 -53.875 56.152
608584 90.605 -53.895 56.119
608585 90.622 -53.915 56.088
608586 90.658 -53.917 56.057
608587 90.693 -53.919 56.029
608588 90.729 -53.921 56.002
608589 90.764 -53.922 55.975
608590 90.796 -53.935 55.949
608591 90.825 -53.955 55.923
608592 90.854 -53.976 55.899
608593 90.883 -53.995 55.875
608594 90.912 -54.016 55.849
608595 90.941 -54.035 55.823
608596 90.970 -54.056 55.797
608597 90.999 -54.076 55.772
608598 91.028 -54.096 55.749
608599 91.057 -54.117 55.727
608600 91.086 -54.136 55.707
608601 91.114 -54.157 55.688
608602 91.144 -54.177 55.670
608603 91.177 -54.197 55.655
608604 91.211 -54.217 55.642
608605 91.244 -54.236 55.632
608606 91.277 -54.256 55.624
608607 91.310 -54.276 55.622
608608 91.344 -54.296 55.624
608609 91.377 -54.316 55.627
608610 91.411 -54.336 55.628
608611 91.444 -54.355 55.626
608612 91.493 -54.364 55.619
608613 91.546 -54.368 55.604
608614 91.599 -54.373 55.584
608615 91.652 -54.379 55.559
608616 91.704 -54.384 55.534
608617 91.757 -54.389 55.511
608618 91.801 -54.398 55.490
608619 91.837 -54.408 55.471
608620 91.873 -54.420 55.451
608621 91.909 -54.431 55.432
608622 91.945 -54.442 55.415
608623 91.981 -54.453 55.400
608624 92.017 -54.464 55.385
608625 92.053 -54.475 55.372
608626 92.089 -54.486 55.361
608627 92.125 -54.498 55.351
608628 92.161 -54.508 55.340
608629 92.197 -54.520 55.330
608630 92.232 -54.531 55.320
608631 92.269 -54.543 55.313
608632 92.304 -54.553 55.306
608633 92.340 -54.565 55.300
608634 92.376 -54.576 55.294
608635 92.412 -54.587 55.288
608636 92.448 -54.598 55.283
608637 92.484 -54.609 55.283
608638 92.520 -54.620 55.284
608639 92.556 -54.631 55.285
608640 92.592 -54.643 55.289
608641 92.628 -54.654 55.296
608642 92.660 -54.670 55.304
608643 92.687 -54.687 55.320
608644 92.714 -54.705 55.342
608645 92.741 -54.723 55.365
608646 92.768 -54.740 55.387
608647 92.795 -54.758 55.409
608648 92.822 -54.775 55.427
608649 92.847 -54.788 55.447
608650 92.871 -54.802 55.468
608651 92.896 -54.815 55.489
608652 92.921 -54.829 55.509
608653 92.945 -54.843 55.528
608654 92.971 -54.859 55.548
608655 92.997 -54.878 55.564
608656 93.023 -54.896 55.574
608657 93.049 -54.914 55.575
608658 93.075 -54.932 55.569
608659 93.101 -54.950 55.560
608660 93.134 -54.968 55.545
608661 93.170 -54.986 55.531
608662 93.205 -55.004 55.519
608663 93.241 -55.022 55.504
608664 93.277 -55.040 55.486
608665 93.312 -55.058 55.464
608666 93.348 -55.076 55.451
608667 93.383 -55.094 55.444
608668 93.419 -55.112 55.444
608669 93.454 -55.130 55.449
608670 93.483 -55.145 55.453
608671 93.508 -55.159 55.446
608672 93.533 -55.173 55.434
608673 93.558 -55.186 55.418
608674 93.583 -55.201 55.398
608675 93.609 -55.214 55.378
608676 93.633 -55.228 55.359
608677 93.658 -55.242 55.343
608678 93.682 -55.258 55.329
608679 93.703 -55.282 55.316
608680 93.723 -55.307 55.305
608681 93.743 -55.331 55.292
608682 93.763 -55.355 55.279
608683 93.783 -55.379 55.265
608684 93.803 -55.403 55.247
608685 93.824 -55.427 55.220
608686 93.844 -55.452 55.196
608687 93.864 -55.476 55.173
608688 93.885 -55.500 55.148
608689 93.905 -55.524 55.120
608690 93.925 -55.548 55.093
608691 93.945 -55.572 55.062
608692 93.966 -55.596 55.031
608693 93.986 -55.620 54.997
608694 94.010 -55.636 54.958
608695 94.035 -55.653 54.918
608696 94.067 -55.665 54.877
608697 94.100 -55.677 54.834
608698 94.131 -55.687 54.795
608699 94.162 -55.696 54.757
608700 94.192 -55.705 54.719
608701 94.223 -55.714 54.679
608702 94.253 -55.723 54.640
608703 94.284 -55.732 54.595
608704 94.314 -55.741 54.554
608705 94.345 -55.749 54.514
608706 94.375 -55.759 54.477
608707 94.406 -55.768 54.443
608708 94.445 -55.778 54.412
608709 94.495 -55.791 54.378
608710 94.544 -55.804 54.351
608711 94.594 -55.816 54.333
608712 94.643 -55.829 54.323
608713 94.694 -55.840 54.315
608714 94.744 -55.853 54.302
608715 94.794 -55.864 54.278
608716 94.830 -55.869 54.247
608717 94.864 -55.873 54.204
608718 94.898 -55.876 54.152
608719 94.931 -55.879 54.095
608720 94.963 -55.884 54.032
608721 94.995 -55.890 53.961
608722 95.026 -55.895 53.890
608723 95.058 -55.900 53.818
608724 95.089 -55.906 53.748
608725 95.121 -55.912 53.679
608726 95.142 -55.928 53.612
608727 95.162 -55.946 53.543
608728 95.180 -55.965 53.470
608729 95.199 -55.984 53.394
608730 95.231 -55.997 53.318
608731 95.263 -56.008 53.244
608732 95.296 -56.020 53.171
608733 95.328 -56.032 53.100
608734 95.360 -56.044 53.032
608735 95.395 -56.056 52.964
608736 95.429 -56.067 52.898
608737 95.464 -56.078 52.833
608738 95.498 -56.090 52.769
608739 95.533 -56.101 52.706
608740 95.568 -56.113 52.645
608741 95.602 -56.125 52.585
608742 95.637 -56.136 52.526
608743 95.667 -56.151 52.469
608744 95.694 -56.170 52.412
608745 95.721 -56.189 52.356
608746 95.747 -56.208 52.302
608747 95.774 -56.228 52.246
608748 95.801 -56.247 52.189
608749 95.826 -56.264 52.129
608750 95.852 -56.282 52.066
608751 95.878 -56.292 52.002
608752 95.903 -56.296 51.942
608753 95.934 -56.298 51.878
608754 95.966 -56.300 51.816
608755 95.997 -56.302 51.755
608756 96.030 -56.304 51.695
608757 96.061 -56.306 51.639
608758 96.093 -56.308 51.586
608759 96.125 -56.310 51.534
608760 96.157 -56.312 51.484
608761 96.189 -56.314 51.436
608762 96.221 -56.316 51.391
608763 96.252 -56.318 51.350
608764 96.284 -56.320 51.310
608765 96.316 -56.322 51.273
608766 96.339 -56.323 51.236
608767 96.355 -56.325 51.200
608768 96.370 -56.326 51.167
608769 96.385 -56.326 51.137
608770 96.401 -56.328 51.108
608771 96.416 -56.329 51.078
608772 96.432 -56.330 51.049
608773 96.447 -56.331 51.020
608774 96.463 -56.332 50.990
608775 96.478 -56.333 50.964
608776 96.493 -56.335 50.938
608777 96.509 -56.335 50.910
608778 96.521 -56.337 50.881
608779 96.533 -56.338 50.853
608780 96.545 -56.339 50.826
608781 96.558 -56.341 50.801
608782 96.571 -56.342 50.774
608783 96.583 -56.343 50.748
608784 96.595 -56.345 50.719
608785 96.617 -56.350 50.690
608786 96.639 -56.355 50.663
608787 96.661 -56.361 50.638
608788 96.693 -56.354 50.613
608789 96.737 -56.351 50.590
608790 96.776 -56.350 50.565
608791 96.810 -56.349 50.539
608792 96.843 -56.348 50.516
608793 96.877 -56.348 50.494
608794 96.910 -56.347 50.472
608795 96.944 -56.345 50.450
608796 96.978 -56.345 50.430
608797 97.012 -56.344 50.410
608798 97.045 -56.342 50.392
608799 97.079 -56.342 50.376
608800 97.113 -56.340 50.361
608801 97.147 -56.339 50.346
608802 97.180 -56.338 50.331
608803 97.214 -56.337 50.318
608804 97.248 -56.335 50.305
608805 97.282 -56.335 50.292
608806 97.315 -56.334 50.279
608807 97.349 -56.332 50.268
608808 97.383 -56.332 50.256
608809 97.417 -56.330 50.245
608810 97.450 -56.329 50.235
608811 97.484 -56.328 50.226
608812 97.518 -56.327 50.216
608813 97.551 -56.326 50.208
608814 97.585 -56.325 50.201
608815 97.619 -56.324 50.195
608816 97.652 -56.323 50.188
608817 97.686 -56.322 50.181
608818 97.723 -56.326 50.172
608819 97.764 -56.334 50.160
608820 97.804 -56.342 50.153
608821 97.844 -56.351 50.148
608822 97.885 -56.360 50.141
608823 97.925 -56.368 50.134
608824 97.965 -56.377 50.125
608825 98.006 -56.385 50.115
608826 98.046 -56.394 50.107
608827 98.087 -56.402 50.095
608828 98.127 -56.411 50.081
608829 98.167 -56.420 50.065
608830 98.208 -56.428 50.047
608831 98.248 -56.436 50.026
608832 98.289 -56.445 50.002
608833 98.329 -56.454 49.978
608834 98.370 -56.462 49.952
608835 98.407 -56.470 49.924
608836 98.442 -56.477 49.896
608837 98.476 -56.485 49.867
608838 98.511 -56.492 49.838
608839 98.545 -56.499 49.808
608840 98.580 -56.507 49.779
608841 98.614 -56.514 49.749
608842 98.649 -56.521 49.719
608843 98.683 -56.529 49.692
608844 98.718 -56.536 49.664
608845 98.752 -56.543 49.637
608846 98.787 -56.551 49.612
608847 98.821 -56.558 49.587
608848 98.856 -56.565 49.564
608849 98.890 -56.573 49.541
608850 98.925 -56.580 49.521
608851 98.959 -56.587 49.500
608852 98.994 -56.595 49.476
608853 99.028 -56.602 49.452
608854 99.063 -56.609 49.428
608855 99.097 -56.617 49.406
608856 99.132 -56.624 49.385
608857 99.166 -56.631 49.365
608858 99.201 -56.639 49.346
608859 99.235 -56.646 49.329
608860 99.270 -56.653 49.314
608861 99.304 -56.661 49.299
608862 99.339 -56.668 49.282
608863 99.373 -56.675 49.267
608864 99.408 -56.683 49.251
608865 99.442 -56.690 49.233
608866 99.477 -56.697 49.215
608867 99.506 -56.700 49.198
608868 99.532 -56.718 49.179
608869 99.564 -56.734 49.159
608870 99.596 -56.750 49.140
608871 99.625 -56.761 49.118
608872 99.651 -56.772 49.096
608873 99.679 -56.789 49.073
608874 99.707 -56.806 49.048
608875 99.735 -56.824 49.024
608876 99.763 -56.840 48.998
608877 99.790 -56.858 48.972
608878 99.818 -56.875 48.946
608879 99.846 -56.892 48.920
608880 99.874 -56.909 48.895
608881 99.902 -56.927 48.868
608882 99.929 -56.943 48.846
608883 99.958 -56.961 48.822
608884 99.985 -56.978 48.797
609224 88.775 -52.719 57.380
609225 88.782 -52.764 57.379
609226 88.814 -52.787 57.378
609227 88.845 -52.810 57.376
609228 88.877 -52.832 57.374
609229 88.908 -52.855 57.369
609230 88.940 -52.878 57.366
609231 88.972 -52.900 57.363
609232 89.003 -52.923 57.359
609233 89.035 -52.945 57.354
609234 89.067 -52.968 57.349
609235 89.098 -52.991 57.344
609236 89.130 -53.013 57.339
609237 89.162 -53.036 57.334
609238 89.193 -53.059 57.329
609239 89.225 -53.082 57.323
609240 89.257 -53.104 57.318
609241 89.288 -53.127 57.313
609242 89.320 -53.150 57.309
609243 89.352 -53.173 57.304
609244 89.383 -53.195 57.298
609245 89.410 -53.223 57.292
609246 89.433 -53.256 57.286
609247 89.455 -53.289 57.281
609248 89.483 -53.314 57.274
609249 89.521 -53.325 57.268
609250 89.559 -53.336 57.262
609251 89.593 -53.349 57.252
609252 89.616 -53.367 57.241
609253 89.639 -53.384 57.228
609254 89.665 -53.402 57.215
609255 89.706 -53.430 57.201
609256 89.747 -53.456 57.188
609257 89.789 -53.483 57.174
609258 89.809 -53.488 57.161
609259 89.828 -53.491 57.145
609260 89.846 -53.494 57.127
609261 89.866 -53.498 57.107
609262 89.892 -53.510 57.087
609263 89.917 -53.521 57.066
609264 89.943 -53.534 57.041
609265 89.971 -53.547 57.014
609266 90.006 -53.565 56.983
609267 90.040 -53.582 56.948
609268 90.075 -53.599 56.912
609269 90.109 -53.617 56.875
609270 90.146 -53.633 56.839
609271 90.189 -53.644 56.803
609272 90.232 -53.656 56.765
609273 90.275 -53.667 56.727
609274 90.318 -53.678 56.690
609275 90.362 -53.689 56.655
609276 90.405 -53.700 56.620
609277 90.448 -53.712 56.587
609284 90.596 -53.844 56.160
609285 90.613 -53.863 56.126
609286 90.645 -53.869 56.093
609287 90.680 -53.871 56.062
609288 90.716 -53.872 56.033
609289 90.752 -53.874 56.006
609290 90.787 -53.878 55.978
609291 90.816 -53.898 55.950
609292 90.845 -53.919 55.924
609293 90.874 -53.939 55.898
609294 90.902 -53.959 55.873
609295 90.931 -53.979 55.847
609296 90.960 -53.999 55.821
609297 90.989 -54.019 55.795
609298 91.018 -54.039 55.771
609299 91.047 -54.060 55.747
609300 91.076 -54.080 55.725
609301 91.105 -54.100 55.705
609302 91.135 -54.120 55.686
609303 91.169 -54.140 55.668
609304 91.202 -54.160 55.654
609305 91.235 -54.179 55.641
609306 91.269 -54.200 55.630
609307 91.302 -54.220 55.622
609308 91.335 -54.239 55.620
609309 91.368 -54.259 55.621
609310 91.402 -54.279 55.624
609311 91.435 -54.299 55.624
609312 91.468 -54.319 55.621
609313 91.515 -54.329 55.615
609314 91.568 -54.334 55.602
609315 91.621 -54.339 55.582
609316 91.673 -54.344 55.557
609317 91.726 -54.349 55.530
609318 91.777 -54.355 55.505
609319 91.812 -54.367 55.483
609320 91.848 -54.377 55.464
609321 91.884 -54.389 55.446
609322 91.920 -54.400 55.429
609323 91.956 -54.411 55.413
609324 91.992 -54.422 55.397
609325 92.028 -54.433 55.382
609326 92.064 -54.445 55.370
609327 92.100 -54.455 55.359
609328 92.136 -54.467 55.349
609329 92.172 -54.478 55.338
609330 92.208 -54.489 55.327
609331 92.243 -54.500 55.317
609332 92.280 -54.511 55.310
609333 92.315 -54.522 55.304
609334 92.351 -54.533 55.298
609335 92.387 -54.545 55.292
609336 92.423 -54.556 55.286
609337 92.459 -54.567 55.282
609338 92.495 -54.578 55.282
609339 92.532 -54.590 55.284
609340 92.571 -54.602 55.286
609341 92.609 -54.615 55.291
609342 92.648 -54.627 55.298
609343 92.681 -54.646 55.308
609344 92.707 -54.665 55.325
609345 92.734 -54.684 55.345
609346 92.761 -54.701 55.366
609347 92.788 -54.718 55.385
609348 92.815 -54.736 55.403
609349 92.842 -54.753 55.419
609350 92.867 -54.767 55.439
609351 92.892 -54.781 55.460
609352 92.916 -54.794 55.478
609353 92.941 -54.808 55.496
609354 92.966 -54.822 55.513
609355 92.991 -54.834 55.531
609356 93.016 -54.851 55.545
609357 93.042 -54.869 55.554
609358 93.068 -54.887 55.555
609359 93.094 -54.906 55.550
609360 93.120 -54.924 55.539
609361 93.150 -54.942 55.524
609362 93.186 -54.960 55.511
609363 93.221 -54.978 55.500
609364 93.257 -54.996 55.487
609365 93.292 -55.013 55.473
609366 93.328 -55.032 55.459
609367 93.363 -55.050 55.453
609368 93.399 -55.068 55.456
609369 93.435 -55.085 55.465
609370 93.464 -55.101 55.474
609371 93.489 -55.115 55.476
609372 93.514 -55.129 55.468
609373 93.540 -55.143 55.452
609374 93.565 -55.157 55.432
609375 93.589 -55.170 55.412
609376 93.615 -55.184 55.392
609377 93.640 -55.198 55.372
609378 93.665 -55.212 55.355
609379 93.690 -55.226 55.344
609380 93.710 -55.249 55.331
609381 93.730 -55.273 55.319
609382 93.751 -55.298 55.307
609383 93.771 -55.322 55.294
609384 93.791 -55.346 55.278
609385 93.812 -55.370 55.260
609386 93.832 -55.395 55.238
609387 93.852 -55.418 55.219
609388 93.872 -55.442 55.199
609389 93.893 -55.467 55.177
609390 93.913 -55.491 55.151
609391 93.933 -55.515 55.123
609392 93.953 -55.539 55.093
609393 93.973 -55.563 55.060
609394 93.995 -55.583 55.023
609395 94.020 -55.600 54.983
609396 94.044 -55.617 54.942
609397 94.076 -55.630 54.901
609398 94.109 -55.642 54.859
609399 94.142 -55.653 54.819
609400 94.173 -55.662 54.779
609401 94.203 -55.671 54.738
609402 94.234 -55.680 54.697
609403 94.264 -55.690 54.656
609404 94.294 -55.698 54.610
609405 94.325 -55.707 54.566
609406 94.355 -55.716 54.525
609407 94.386 -55.725 54.487
609408 94.416 -55.734 54.453
609409 94.458 -55.745 54.423
609410 94.508 -55.758 54.389
609411 94.557 -55.771 54.362
609412 94.607 -55.783 54.344
609413 94.657 -55.796 54.332
609414 94.706 -55.809 54.323
609415 94.755 -55.821 54.313
609416 94.803 -55.834 54.289
609417 94.834 -55.844 54.259
609418 94.862 -55.854 54.220
609419 94.891 -55.864 54.171
609420 94.921 -55.871 54.114
609421 94.953 -55.876 54.052
609422 94.984 -55.882 53.978
609423 95.016 -55.887 53.906
609424 95.047 -55.893 53.834
609425 95.078 -55.898 53.764
609426 95.110 -55.903 53.697
609427 95.141 -55.909 53.632
609428 95.163 -55.925 53.566
609429 95.182 -55.944 53.495
609430 95.206 -55.960 53.418
609431 95.238 -55.972 53.340
609432 95.271 -55.984 53.263
609433 95.303 -55.996 53.186
609434 95.336 -56.008 53.112
609435 95.368 -56.020 53.040
609436 95.401 -56.031 52.970
609437 95.436 -56.043 52.902
609438 95.471 -56.054 52.835
609439 95.505 -56.066 52.770
609440 95.540 -56.078 52.707
609441 95.574 -56.089 52.645
609442 95.609 -56.100 52.584
609443 95.642 -56.113 52.525
609444 95.669 -56.132 52.467
609445 95.695 -56.151 52.409
609446 95.722 -56.170 52.353
609447 95.749 -56.189 52.297
609448 95.775 -56.208 52.241
609449 95.802 -56.227 52.185
609450 95.828 -56.246 52.127
609451 95.855 -56.265 52.068
609452 95.884 -56.272 52.007
609453 95.916 -56.274 51.943
609454 95.948 -56.276 51.881
609455 95.980 -56.278 51.817
609456 96.012 -56.280 51.753
609457 96.043 -56.282 51.692
609458 96.076 -56.284 51.635
609459 96.107 -56.286 51.580
609460 96.139 -56.288 51.527
609461 96.171 -56.290 51.475
609462 96.202 -56.292 51.425
609463 96.235 -56.294 51.378
609464 96.266 -56.296 51.336
609465 96.298 -56.298 51.296
609466 96.330 -56.300 51.257
609467 96.349 -56.301 51.219
609468 96.365 -56.303 51.182
609469 96.380 -56.304 51.148
609470 96.396 -56.304 51.118
609471 96.411 -56.306 51.087
609472 96.427 -56.307 51.057
609473 96.442 -56.308 51.026
609474 96.458 -56.309 50.995
609475 96.473 -56.310 50.965
609476 96.488 -56.311 50.938
609477 96.504 -56.313 50.910
609478 96.516 -56.313 50.882
609479 96.529 -56.315 50.853
609480 96.541 -56.317 50.823
609481 96.553 -56.317 50.796
609482 96.566 -56.319 50.771
609483 96.578 -56.320 50.746
609484 96.591 -56.321 50.719
609485 96.603 -56.323 50.690
609486 96.615 -56.323 50.661
609487 96.637 -56.329 50.633
609488 96.663 -56.329 50.607
609489 96.705 -56.327 50.582
609490 96.739 -56.326 50.558
609491 96.772 -56.325 50.534
609492 96.806 -56.324 50.508
609493 96.840 -56.323 50.484
609494 96.874 -56.322 50.462
609495 96.907 -56.321 50.440
609496 96.941 -56.320 50.419
609497 96.975 -56.319 50.398
609498 97.009 -56.317 50.377
609499 97.042 -56.317 50.358
609500 97.076 -56.315 50.341
609501 97.110 -56.314 50.326
609502 97.144 -56.313 50.311
609503 97.177 -56.312 50.295
609504 97.212 -56.311 50.281
609505 97.246 -56.310 50.268
609506 97.282 -56.309 50.255
609507 97.316 -56.308 50.242
609508 97.351 -56.307 50.230
609509 97.386 -56.306 50.217
609510 97.421 -56.305 50.205
609511 97.456 -56.304 50.195
609512 97.491 -56.303 50.185
609513 97.526 -56.301 50.174
609514 97.561 -56.301 50.165
609515 97.596 -56.300 50.157
609516 97.631 -56.298 50.149
609517 97.666 -56.298 50.141
609518 97.700 -56.297 50.134
609519 97.736 -56.303 50.125
609520 97.777 -56.311 50.114
609521 97.817 -56.320 50.106
609522 97.858 -56.329 50.100
609523 97.898 -56.337 50.094
609524 97.938 -56.345 50.087
609525 97.979 -56.354 50.078
609526 98.019 -56.363 50.070
609527 98.060 -56.371 50.061
609528 98.100 -56.379 50.049
609529 98.141 -56.389 50.035
609530 98.181 -56.397 50.019
609531 98.221 -56.405 50.001
609532 98.262 -56.414 49.980
609533 98.302 -56.423 49.957
609534 98.343 -56.431 49.933
609535 98.383 -56.439 49.907
609536 98.417 -56.447 49.880
609537 98.452 -56.454 49.852
609538 98.486 -56.461 49.823
609539 98.521 -56.469 49.794
609540 98.556 -56.476 49.764
609541 98.590 -56.483 49.735
609542 98.625 -56.491 49.705
609543 98.659 -56.498 49.675
609544 98.694 -56.505 49.647
609545 98.728 -56.513 49.619
609546 98.763 -56.520 49.592
609547 98.797 -56.527 49.567
609548 98.832 -56.535 49.543
609549 98.866 -56.542 49.519
609550 98.901 -56.549 49.496
609551 98.935 -56.557 49.474
609552 98.970 -56.564 49.452
609553 99.004 -56.571 49.429
609554 99.039 -56.579 49.406
609555 99.073 -56.586 49.384
609556 99.108 -56.593 49.362
609557 99.142 -56.601 49.341
609558 99.177 -56.608 49.321
609559 99.211 -56.615 49.304
609560 99.245 -56.623 49.289
609561 99.280 -56.630 49.274
609562 99.314 -56.637 49.258
609563 99.349 -56.645 49.244
609564 99.383 -56.652 49.229
609565 99.418 -56.659 49.213
609566 99.452 -56.667 49.199
609567 99.482 -56.669 49.183
609568 99.508 -56.682 49.166
609569 99.535 -56.700 49.149
609570 99.568 -56.716 49.131
609571 99.600 -56.731 49.113
609572 99.626 -56.740 49.093
609573 99.653 -56.751 49.072
609574 99.681 -56.768 49.050
609575 99.709 -56.785 49.027
609576 99.736 -56.802 49.003
609577 99.764 -56.820 48.979
609578 99.792 -56.837 48.955
609579 99.820 -56.854 48.930
609580 99.848 -56.871 48.905
609581 99.875 -56.888 48.881
609582 99.904 -56.906 48.855
609583 99.931 -56.923 48.833
609584 99.959 -56.940 48.810
609926 88.798 -52.740 57.372
609927 88.830 -52.763 57.370
609928 88.861 -52.785 57.368
609929 88.893 -52.808 57.365
609930 88.924 -52.831 57.361
609931 88.956 -52.854 57.358
609932 88.988 -52.876 57.353
609933 89.020 -52.899 57.349
609934 89.051 -52.922 57.344
609935 89.083 -52.944 57.339
609936 89.115 -52.967 57.334
609937 89.146 -52.990 57.329
609938 89.178 -53.013 57.324
609939 89.209 -53.035 57.319
609940 89.241 -53.057 57.315
609941 89.273 -53.080 57.311
609942 89.305 -53.103 57.306
609943 89.336 -53.126 57.302
609944 89.368 -53.148 57.297
609945 89.399 -53.171 57.292
609946 89.425 -53.201 57.287
609947 89.448 -53.233 57.282
609948 89.470 -53.266 57.277
609949 89.507 -53.279 57.271
609950 89.545 -53.290 57.265
609951 89.579 -53.303 57.260
609952 89.602 -53.321 57.250
609953 89.625 -53.338 57.238
609954 89.649 -53.355 57.225
609955 89.676 -53.374 57.212
609956 89.717 -53.402 57.198
609957 89.758 -53.429 57.185
609958 89.799 -53.455 57.172
609959 89.837 -53.480 57.158
609960 89.857 -53.483 57.143
609961 89.875 -53.485 57.126
609962 89.894 -53.488 57.106
609963 89.913 -53.491 57.086
609964 89.934 -53.496 57.065
609965 89.959 -53.508 57.043
609966 89.984 -53.520 57.019
609967 90.015 -53.535 56.990
609968 90.049 -53.552 56.957
609969 90.084 -53.570 56.922
609970 90.118 -53.587 56.885
609971 90.161 -53.599 56.849
609972 90.204 -53.610 56.814
609973 90.247 -53.621 56.778
609974 90.291 -53.633 56.739
609975 90.333 -53.644 56.702
609976 90.377 -53.655 56.666
609977 90.420 -53.666 56.632
609978 90.463 -53.678 56.600
609985 90.603 -53.812 56.169
609986 90.633 -53.820 56.134
609987 90.668 -53.822 56.101
609988 90.704 -53.824 56.070
609989 90.739 -53.825 56.040
609990 90.775 -53.828 56.011
609991 90.806 -53.842 55.982
609992 90.835 -53.862 55.954
609993 90.864 -53.882 55.926
609994 90.893 -53.902 55.899
609995 90.922 -53.922 55.872
609996 90.951 -53.942 55.846
609997 90.979 -53.963 55.821
609998 91.009 -53.983 55.795
609999 91.037 -54.003 55.771
610000 91.067 -54.023 55.747
610001 91.095 -54.043 55.726
610002 91.127 -54.063 55.705
610003 91.160 -54.083 55.686
610004 91.193 -54.103 55.669
610005 91.227 -54.123 55.654
610006 91.260 -54.143 55.640
610007 91.293 -54.163 55.630
610008 91.326 -54.182 55.622
610009 91.360 -54.202 55.617
610010 91.393 -54.222 55.616
610011 91.427 -54.242 55.618
610012 91.460 -54.262 55.617
610013 91.493 -54.282 55.616
610014 91.536 -54.294 55.611
610015 91.589 -54.299 55.600
610016 91.642 -54.305 55.582
610017 91.695 -54.310 55.557
610018 91.747 -54.315 55.527
610019 91.788 -54.324 55.500
610020 91.823 -54.336 55.478
610021 91.860 -54.347 55.459
610022 91.895 -54.358 55.442
610023 91.931 -54.369 55.426
610024 91.967 -54.380 55.411
610025 92.003 -54.391 55.396
610026 92.039 -54.402 55.381
610027 92.075 -54.414 55.369
610028 92.111 -54.425 55.358
610029 92.147 -54.436 55.347
610030 92.183 -54.447 55.336
610031 92.219 -54.458 55.325
610032 92.254 -54.469 55.315
610033 92.291 -54.480 55.308
610034 92.326 -54.492 55.302
610035 92.363 -54.502 55.296
610036 92.399 -54.514 55.291
610037 92.437 -54.527 55.286
610038 92.476 -54.539 55.282
610039 92.514 -54.552 55.282
610040 92.553 -54.564 55.284
610041 92.591 -54.576 55.287
610042 92.630 -54.589 55.292
610043 92.668 -54.601 55.299
610044 92.704 -54.617 55.311
610045 92.727 -54.643 55.328
610046 92.754 -54.662 55.346
610047 92.781 -54.679 55.364
610048 92.808 -54.696 55.380
610049 92.835 -54.714 55.396
610050 92.862 -54.732 55.412
610051 92.887 -54.745 55.430
610052 92.912 -54.759 55.449
610053 92.937 -54.772 55.467
610054 92.961 -54.786 55.483
610055 92.986 -54.800 55.497
610056 93.011 -54.813 55.509
610057 93.036 -54.827 55.521
610058 93.061 -54.843 55.529
610059 93.087 -54.861 55.531
610060 93.113 -54.879 55.525
610061 93.139 -54.897 55.514
610062 93.166 -54.916 55.499
610063 93.202 -54.934 55.486
610064 93.237 -54.951 55.477
610065 93.273 -54.969 55.469
610066 93.308 -54.988 55.459
610067 93.344 -55.005 55.455
610068 93.379 -55.023 55.463
610069 93.415 -55.041 55.478
610070 93.445 -55.057 55.490
610071 93.471 -55.071 55.495
610072 93.496 -55.085 55.492
610073 93.521 -55.099 55.484
610074 93.546 -55.113 55.466
610075 93.571 -55.126 55.445
610076 93.596 -55.140 55.425
610077 93.621 -55.154 55.406
610078 93.646 -55.168 55.387
610079 93.672 -55.182 55.370
610080 93.697 -55.195 55.358
610081 93.718 -55.217 55.346
610082 93.739 -55.241 55.335
610083 93.759 -55.265 55.323
610084 93.779 -55.289 55.310
610085 93.799 -55.313 55.296
610086 93.820 -55.337 55.280
610087 93.840 -55.361 55.261
610088 93.860 -55.386 55.242
610089 93.880 -55.410 55.223
610090 93.900 -55.433 55.202
610091 93.920 -55.458 55.178
610092 93.941 -55.482 55.151
610093 93.961 -55.506 55.116
610094 93.981 -55.530 55.081
610095 94.005 -55.547 55.044
610096 94.030 -55.564 55.003
610097 94.054 -55.580 54.960
610098 94.085 -55.594 54.922
610099 94.118 -55.605 54.881
610100 94.150 -55.618 54.840
610101 94.183 -55.629 54.798
610102 94.213 -55.638 54.755
610103 94.244 -55.646 54.713
610104 94.274 -55.655 54.668
610105 94.305 -55.665 54.622
610106 94.335 -55.674 54.577
610107 94.366 -55.682 54.535
610108 94.396 -55.691 54.497
610109 94.427 -55.700 54.462
610110 94.471 -55.712 54.433
610111 94.521 -55.724 54.402
610112 94.570 -55.737 54.376
610113 94.620 -55.750 54.357
610114 94.670 -55.762 54.342
610115 94.719 -55.775 54.331
610116 94.768 -55.788 54.320
610117 94.808 -55.801 54.294
610118 94.841 -55.812 54.265
610119 94.870 -55.821 54.226
610120 94.898 -55.831 54.180
610121 94.926 -55.841 54.125
610122 94.954 -55.851 54.062
610123 94.982 -55.860 53.989
610124 95.011 -55.870 53.917
610125 95.039 -55.880 53.848
610126 95.067 -55.890 53.782
610127 95.099 -55.895 53.717
610128 95.130 -55.900 53.652
610129 95.162 -55.906 53.586
610130 95.183 -55.922 53.517
610131 95.213 -55.935 53.443
610132 95.246 -55.947 53.365
610133 95.278 -55.959 53.285
610134 95.311 -55.972 53.204
610135 95.343 -55.984 53.126
610136 95.376 -55.996 53.052
610137 95.408 -56.007 52.980
610138 95.443 -56.019 52.909
610139 95.477 -56.031 52.841
610140 95.512 -56.042 52.774
610141 95.546 -56.053 52.710
610142 95.581 -56.065 52.648
610143 95.615 -56.076 52.586
610144 95.643 -56.094 52.525
610145 95.670 -56.113 52.466
610146 95.697 -56.132 52.407
610147 95.724 -56.151 52.351
610148 95.750 -56.170 52.294
610149 95.777 -56.189 52.238
610150 95.804 -56.208 52.182
610151 95.832 -56.224 52.127
610152 95.865 -56.238 52.071
610153 95.898 -56.251 52.014
610154 95.930 -56.252 51.948
610155 95.962 -56.254 51.883
610156 95.994 -56.257 51.818
610157 96.026 -56.258 51.753
610158 96.057 -56.260 51.691
610159 96.089 -56.262 51.631
610160 96.121 -56.264 51.575
610161 96.153 -56.266 51.520
610162 96.185 -56.268 51.467
610163 96.217 -56.270 51.416
610164 96.249 -56.273 51.367
610165 96.281 -56.274 51.323
610166 96.312 -56.276 51.282
610167 96.344 -56.279 51.242
610168 96.360 -56.279 51.202
610169 96.375 -56.280 51.164
610170 96.391 -56.282 51.128
610171 96.406 -56.282 51.096
610172 96.421 -56.284 51.065
610173 96.437 -56.285 51.033
610174 96.453 -56.286 51.001
610175 96.468 -56.287 50.969
610176 96.483 -56.288 50.939
610177 96.499 -56.289 50.910
610178 96.512 -56.291 50.882
610179 96.524 -56.291 50.854
610180 96.536 -56.293 50.824
610181 96.549 -56.295 50.794
610182 96.561 -56.295 50.765
610183 96.574 -56.297 50.739
610184 96.586 -56.298 50.714
610185 96.598 -56.299 50.686
610186 96.611 -56.301 50.658
610187 96.623 -56.302 50.631
610188 96.636 -56.303 50.604
610189 96.668 -56.302 50.578
610190 96.703 -56.301 50.551
610191 96.738 -56.301 50.527
610192 96.772 -56.299 50.502
610193 96.808 -56.298 50.477
610194 96.842 -56.298 50.453
610195 96.878 -56.296 50.430
610196 96.912 -56.295 50.409
610197 96.948 -56.295 50.387
610198 96.982 -56.293 50.366
610199 97.018 -56.292 50.344
610200 97.052 -56.291 50.325
610201 97.088 -56.290 50.308
610202 97.122 -56.289 50.291
610203 97.158 -56.288 50.275
610204 97.192 -56.287 50.260
610205 97.227 -56.286 50.244
610206 97.262 -56.285 50.231
610207 97.297 -56.284 50.218
610208 97.332 -56.282 50.205
610209 97.367 -56.282 50.192
610210 97.402 -56.281 50.179
610211 97.437 -56.279 50.167
610212 97.472 -56.279 50.155
610213 97.507 -56.278 50.144
610214 97.542 -56.276 50.133
610215 97.577 -56.276 50.124
610216 97.612 -56.275 50.114
610217 97.647 -56.273 50.104
610218 97.682 -56.273 50.095
610219 97.709 -56.271 50.088
610220 97.749 -56.280 50.079
610221 97.790 -56.288 50.070
610222 97.830 -56.297 50.062
610223 97.871 -56.305 50.055
610224 97.911 -56.314 50.048
610225 97.952 -56.323 50.041
610226 97.992 -56.331 50.033
610227 98.032 -56.340 50.025
610228 98.073 -56.348 50.015
610229 98.113 -56.357 50.004
610230 98.154 -56.365 49.990
610231 98.194 -56.374 49.974
610232 98.235 -56.382 49.955
610233 98.275 -56.391 49.935
610234 98.315 -56.399 49.913
610235 98.356 -56.408 49.889
610236 98.393 -56.416 49.864
610237 98.428 -56.423 49.837
610238 98.462 -56.430 49.808
610239 98.497 -56.438 49.780
610240 98.531 -56.445 49.751
610241 98.566 -56.452 49.721
610242 98.600 -56.460 49.691
610243 98.635 -56.467 49.661
610244 98.669 -56.474 49.632
610245 98.704 -56.482 49.603
610246 98.738 -56.489 49.575
610247 98.773 -56.496 49.548
610248 98.807 -56.504 49.522
610249 98.842 -56.511 49.498
610250 98.876 -56.518 49.475
610251 98.911 -56.526 49.451
610252 98.945 -56.533 49.428
610253 98.980 -56.540 49.406
610254 99.014 -56.548 49.383
610255 99.049 -56.555 49.361
610256 99.083 -56.562 49.339
610257 99.118 -56.570 49.318
610258 99.152 -56.577 49.298
610259 99.187 -56.584 49.279
610260 99.221 -56.592 49.262
610261 99.256 -56.599 49.247
610262 99.290 -56.606 49.233
610263 99.325 -56.614 49.219
610264 99.359 -56.621 49.205
610265 99.394 -56.628 49.191
610266 99.428 -56.636 49.177
610267 99.458 -56.639 49.163
610268 99.484 -56.646 49.148
610269 99.509 -56.664 49.133
610270 99.539 -56.682 49.118
610271 99.572 -56.698 49.101
610272 99.602 -56.711 49.084
610273 99.628 -56.720 49.066
610274 99.655 -56.730 49.046
610275 99.683 -56.747 49.026
610276 99.710 -56.764 49.005
610277 99.738 -56.781 48.982
610278 99.766 -56.799 48.960
610279 99.794 -56.815 48.937
610280 99.822 -56.833 48.914
610281 99.850 -56.850 48.890
610282 99.877 -56.867 48.866
610283 99.905 -56.884 48.842
610284 99.933 -56.902 48.820
610628 88.846 -52.739 57.362
610629 88.878 -52.762 57.360
610630 88.909 -52.784 57.357
610631 88.941 -52.807 57.352
610632 88.972 -52.829 57.348
610633 89.004 -52.852 57.344
610634 89.036 -52.875 57.339
610635 89.067 -52.897 57.335
610636 89.099 -52.920 57.330
610637 89.131 -52.943 57.325
610638 89.162 -52.966 57.320
610639 89.194 -52.988 57.315
610640 89.226 -53.011 57.310
610641 89.257 -53.034 57.306
610642 89.289 -53.057 57.302
610643 89.321 -53.079 57.298
610644 89.352 -53.102 57.294
610645 89.384 -53.125 57.290
610646 89.415 -53.147 57.286
610647 89.440 -53.178 57.282
610648 89.463 -53.211 57.277
610649 89.492 -53.233 57.272
610650 89.531 -53.244 57.266
610651 89.564 -53.257 57.259
610652 89.588 -53.274 57.254
610653 89.611 -53.292 57.245
610654 89.634 -53.309 57.233
610655 89.658 -53.327 57.220
610656 89.686 -53.347 57.207
610657 89.728 -53.374 57.194
610658 89.768 -53.401 57.182
610659 89.810 -53.427 57.169
610660 89.851 -53.455 57.155
610661 89.885 -53.474 57.141
610662 89.904 -53.477 57.125
610663 89.923 -53.480 57.106
610664 89.942 -53.483 57.086
610665 89.961 -53.485 57.066
610666 89.979 -53.488 57.046
610667 90.001 -53.495 57.024
610668 90.026 -53.506 56.998
610669 90.058 -53.523 56.967
610670 90.093 -53.540 56.931
610671 90.133 -53.554 56.896
610672 90.176 -53.565 56.861
610673 90.219 -53.576 56.827
610674 90.262 -53.587 56.791
610675 90.306 -53.599 56.753
610676 90.349 -53.610 56.716
610677 90.392 -53.621 56.679
610678 90.436 -53.632 56.645
610679 90.478 -53.643 56.613
610680 90.508 -53.659 56.580
610687 90.655 -53.773 56.144
610688 90.691 -53.775 56.110
610689 90.727 -53.777 56.078
610690 90.762 -53.778 56.048
610691 90.796 -53.785 56.019
610692 90.825 -53.805 55.989
610693 90.854 -53.825 55.959
610694 90.883 -53.845 55.931
610695 90.912 -53.866 55.903
610696 90.941 -53.886 55.876
610697 90.970 -53.906 55.850
610698 90.999 -53.926 55.823
610699 91.028 -53.946 55.797
610700 91.057 -53.966 55.772
610701 91.086 -53.987 55.749
610702 91.118 -54.007 55.727
610703 91.151 -54.026 55.707
610704 91.184 -54.046 55.688
610705 91.218 -54.066 55.671
610706 91.251 -54.086 55.655
610707 91.285 -54.106 55.641
610708 91.318 -54.126 55.630
610709 91.351 -54.145 55.622
610710 91.384 -54.166 55.615
610711 91.418 -54.185 55.612
610712 91.451 -54.205 55.611
610713 91.485 -54.225 55.609
610714 91.518 -54.245 55.607
610715 91.558 -54.260 55.602
610716 91.611 -54.264 55.591
610717 91.663 -54.270 55.574
610718 91.716 -54.275 55.550
610719 91.763 -54.283 55.522
610720 91.799 -54.294 55.496
610721 91.835 -54.305 55.474
610722 91.871 -54.316 55.456
610723 91.907 -54.327 55.440
610724 91.942 -54.338 55.424
610725 91.979 -54.349 55.409
610726 92.014 -54.361 55.395
610727 92.050 -54.371 55.381
610728 92.086 -54.383 55.368
610729 92.122 -54.394 55.357
610730 92.158 -54.405 55.346
610731 92.194 -54.416 55.335
610733 92.266 -54.439 55.314
610734 92.303 -54.450 55.307
610735 92.342 -54.463 55.301
610736 92.380 -54.475 55.295
610737 92.419 -54.488 55.291
610738 92.457 -54.500 55.286
610739 92.496 -54.513 55.282
610740 92.534 -54.525 55.282
610741 92.573 -54.537 55.285
610742 92.611 -54.550 55.288
610743 92.649 -54.562 55.293
610744 92.688 -54.575 55.300
610745 92.726 -54.587 55.311
610746 92.749 -54.616 55.325
610747 92.774 -54.640 55.341
610748 92.801 -54.657 55.358
610749 92.829 -54.674 55.373
610750 92.855 -54.692 55.387
610751 92.882 -54.710 55.404
610752 92.907 -54.724 55.421
610753 92.932 -54.737 55.438
610754 92.957 -54.751 55.454
610755 92.982 -54.765 55.469
610756 93.006 -54.778 55.481
610757 93.031 -54.791 55.491
610758 93.056 -54.805 55.500
610759 93.081 -54.818 55.506
610760 93.106 -54.835 55.507
610761 93.132 -54.853 55.499
610762 93.158 -54.871 55.489
610763 93.184 -54.889 55.477
610764 93.217 -54.907 55.467
610765 93.253 -54.925 55.463
610766 93.288 -54.943 55.461
610767 93.324 -54.961 55.462
610768 93.360 -54.979 55.472
610769 93.395 -54.997 55.488
610770 93.427 -55.013 55.502
610771 93.452 -55.027 55.510
610772 93.477 -55.041 55.510
610773 93.502 -55.055 55.502
610774 93.528 -55.069 55.489
610775 93.552 -55.082 55.473
610776 93.577 -55.097 55.456
610777 93.603 -55.110 55.438
610778 93.628 -55.124 55.419
610779 93.653 -55.138 55.402
610780 93.678 -55.151 55.385
610781 93.703 -55.166 55.372
610782 93.726 -55.183 55.360
610783 93.747 -55.207 55.348
610784 93.767 -55.232 55.335
610785 93.787 -55.256 55.323
610786 93.807 -55.280 55.310
610787 93.827 -55.304 55.296
610788 93.847 -55.328 55.278
610789 93.868 -55.352 55.260
610790 93.888 -55.376 55.241
610791 93.908 -55.401 55.220
610792 93.928 -55.425 55.198
610793 93.948 -55.449 55.169
610794 93.969 -55.473 55.134
610795 93.991 -55.495 55.096
610796 94.015 -55.511 55.056
610797 94.039 -55.527 55.014
610798 94.063 -55.544 54.972
610799 94.094 -55.558 54.935
610800 94.126 -55.570 54.893
610801 94.159 -55.582 54.852
610802 94.192 -55.593 54.810
610803 94.224 -55.604 54.767
610804 94.255 -55.613 54.722
610805 94.285 -55.622 54.677
610806 94.316 -55.630 54.630
610807 94.346 -55.640 54.586
610808 94.377 -55.649 54.544
610809 94.407 -55.658 54.505
610810 94.437 -55.666 54.472
610811 94.484 -55.679 54.444
610812 94.534 -55.691 54.416
610813 94.583 -55.704 54.393
610814 94.633 -55.717 54.373
610815 94.683 -55.729 54.355
610816 94.732 -55.742 54.339
610817 94.773 -55.755 54.323
610818 94.813 -55.768 54.296
610819 94.848 -55.779 54.266
610820 94.876 -55.789 54.228
610821 94.905 -55.799 54.182
610822 94.933 -55.809 54.129
610823 94.961 -55.818 54.068
610824 94.989 -55.828 53.996
610825 95.018 -55.837 53.927
610826 95.046 -55.847 53.862
610827 95.074 -55.857 53.798
610828 95.103 -55.867 53.733
610829 95.131 -55.877 53.667
610830 95.160 -55.887 53.600
610831 95.188 -55.899 53.531
610832 95.221 -55.911 53.459
610833 95.254 -55.923 53.382
610834 95.286 -55.935 53.303
610835 95.318 -55.947 53.221
610836 95.351 -55.959 53.141
610837 95.383 -55.971 53.066
610838 95.416 -55.983 52.992
610839 95.449 -55.995 52.921
610840 95.484 -56.006 52.851
610841 95.519 -56.018 52.781
610842 95.553 -56.029 52.717
610843 95.588 -56.041 52.653
610844 95.618 -56.056 52.589
610845 95.645 -56.075 52.527
610846 95.672 -56.094 52.466
610847 95.698 -56.113 52.407
610848 95.725 -56.132 52.350
610849 95.752 -56.150 52.293
610850 95.780 -56.167 52.237
610851 95.812 -56.181 52.182
610852 95.843 -56.194 52.129
610853 95.877 -56.208 52.077
610854 95.911 -56.222 52.023
610855 95.944 -56.231 51.956
610856 95.976 -56.232 51.888
610857 96.008 -56.235 51.820
610858 96.040 -56.236 51.754
610859 96.072 -56.238 51.690
610860 96.103 -56.241 51.629
610861 96.135 -56.242 51.570
610862 96.167 -56.244 51.514
610863 96.199 -56.247 51.460
610864 96.231 -56.248 51.407
610865 96.262 -56.251 51.357
610866 96.295 -56.252 51.311
610867 96.326 -56.254 51.268
610868 96.355 -56.257 51.227
610869 96.370 -56.257 51.186
610870 96.386 -56.258 51.146
610871 96.401 -56.260 51.109
610872 96.416 -56.260 51.075
610873 96.432 -56.262 51.043
610874 96.448 -56.263 51.010
610875 96.463 -56.264 50.977
610876 96.478 -56.265 50.945
610877 96.493 -56.266 50.913
610878 96.507 -56.267 50.883
610879 96.519 -56.269 50.854
610880 96.532 -56.270 50.825
610881 96.544 -56.271 50.795
610882 96.556 -56.273 50.764
610883 96.569 -56.273 50.735
610884 96.581 -56.275 50.708
610885 96.594 -56.276 50.681
610886 96.606 -56.277 50.653
610887 96.618 -56.279 50.625
610888 96.632 -56.279 50.598
610889 96.649 -56.279 50.571
610890 96.684 -56.277 50.546
610891 96.719 -56.276 50.521
610892 96.754 -56.276 50.496
610893 96.789 -56.274 50.471
610894 96.824 -56.273 50.446
610895 96.859 -56.272 50.422
610896 96.894 -56.271 50.400
610897 96.929 -56.270 50.378
610898 96.964 -56.269 50.356
610899 96.999 -56.268 50.334
610900 97.034 -56.267 50.312
610901 97.069 -56.266 50.292
610902 97.103 -56.265 50.275
610903 97.139 -56.264 50.258
610904 97.173 -56.263 50.241
610905 97.209 -56.262 50.225
610906 97.243 -56.260 50.209
610907 97.279 -56.260 50.195
610908 97.313 -56.259 50.182
610909 97.349 -56.257 50.168
610910 97.383 -56.257 50.154
610911 97.419 -56.256 50.140
610912 97.453 -56.254 50.128
610913 97.489 -56.254 50.116
610914 97.523 -56.253 50.105
610915 97.558 -56.251 50.093
610916 97.593 -56.251 50.083
610917 97.628 -56.249 50.072
610918 97.663 -56.248 50.061
610919 97.693 -56.248 50.053
610920 97.722 -56.248 50.044
610921 97.762 -56.257 50.035
610922 97.803 -56.266 50.027
610923 97.844 -56.274 50.019
610924 97.884 -56.282 50.011
610925 97.925 -56.291 50.004
610926 97.965 -56.300 49.997
610927 98.005 -56.308 49.989
610928 98.045 -56.317 49.980
610929 98.086 -56.326 49.970
610930 98.127 -56.334 49.958
610931 98.167 -56.342 49.944
610932 98.208 -56.351 49.929
610933 98.248 -56.360 49.911
610934 98.288 -56.368 49.891
610935 98.329 -56.376 49.869
610936 98.369 -56.385 49.846
610937 98.403 -56.392 49.820
610938 98.438 -56.400 49.794
610939 98.472 -56.407 49.766
610940 98.507 -56.414 49.737
610941 98.541 -56.422 49.708
610942 98.576 -56.429 49.679
610943 98.610 -56.436 49.648
610944 98.645 -56.444 49.618
610945 98.679 -56.451 49.589
610946 98.714 -56.458 49.560
610947 98.748 -56.466 49.532
610948 98.783 -56.473 49.505
610949 98.817 -56.480 49.479
610950 98.852 -56.488 49.454
610951 98.886 -56.495 49.430
610952 98.921 -56.502 49.407
610953 98.955 -56.510 49.383
610954 98.990 -56.517 49.362
610955 99.024 -56.524 49.338
610956 99.059 -56.532 49.316
610957 99.093 -56.539 49.295
610958 99.128 -56.546 49.274
610959 99.162 -56.554 49.256
610960 99.197 -56.561 49.238
610961 99.231 -56.568 49.221
610962 99.266 -56.576 49.206
610963 99.300 -56.583 49.192
610964 99.335 -56.590 49.179
610965 99.369 -56.598 49.166
610966 99.404 -56.605 49.153
610967 99.434 -56.608 49.140
610968 99.461 -56.610 49.127
610969 99.486 -56.628 49.112
610970 99.511 -56.647 49.099
610971 99.543 -56.664 49.085
610972 99.575 -56.680 49.069
610973 99.604 -56.690 49.054
610974 99.630 -56.699 49.038
610975 99.656 -56.709 49.020
610976 99.685 -56.726 49.001
610977 99.712 -56.743 48.981
610978 99.740 -56.760 48.960
610979 99.768 -56.777 48.940
610980 99.796 -56.794 48.919
610981 99.824 -56.812 48.896
610982 99.851 -56.829 48.874
610983 99.879 -56.846 48.852
610984 99.907 -56.863 48.829
611330 88.894 -52.738 57.351
611331 88.925 -52.760 57.347
611332 88.957 -52.783 57.343
611333 88.988 -52.806 57.339
611334 89.020 -52.828 57.335
611335 89.052 -52.851 57.330
611336 89.084 -52.874 57.325
611337 89.115 -52.897 57.321
611338 89.147 -52.919 57.315
611339 89.179 -52.941 57.310
611340 89.210 -52.964 57.306
611341 89.242 -52.987 57.301
611342 89.273 -53.010 57.297
611343 89.305 -53.032 57.293
611344 89.337 -53.055 57.289
611345 89.369 -53.078 57.285
611346 89.400 -53.101 57.282
611347 89.432 -53.123 57.280
611348 89.455 -53.155 57.276
611349 89.478 -53.186 57.271
611350 89.517 -53.198 57.265
611351 89.550 -53.211 57.259
611352 89.574 -53.228 57.252
611353 89.597 -53.245 57.245
611354 89.620 -53.263 57.237
611355 89.644 -53.280 57.226
611356 89.667 -53.298 57.214
611357 89.697 -53.319 57.202
611358 89.738 -53.346 57.190
611359 89.779 -53.373 57.180
611360 89.820 -53.399 57.167
611361 89.862 -53.427 57.154
611362 89.902 -53.447 57.140
611363 89.927 -53.454 57.124
611364 89.948 -53.461 57.106
611365 89.968 -53.468 57.087
611366 89.989 -53.475 57.069
611367 90.008 -53.480 57.050
611368 90.027 -53.483 57.029
611369 90.046 -53.486 57.005
611370 90.068 -53.493 56.976
611371 90.105 -53.509 56.941
611372 90.148 -53.520 56.908
611373 90.191 -53.531 56.874
611374 90.235 -53.542 56.840
611375 90.278 -53.553 56.805
611376 90.321 -53.565 56.768
611377 90.364 -53.576 56.731
611378 90.407 -53.587 56.696
611379 90.451 -53.598 56.662
611380 90.494 -53.609 56.628
611381 90.515 -53.628 56.595
611388 90.678 -53.727 56.155
611389 90.714 -53.728 56.120
611390 90.750 -53.730 56.088
611391 90.786 -53.732 56.057
611392 90.816 -53.749 56.028
611393 90.845 -53.769 55.998
611394 90.874 -53.789 55.966
611395 90.902 -53.809 55.938
611396 90.932 -53.829 55.909
611397 90.960 -53.849 55.881
611398 90.990 -53.869 55.855
611399 91.018 -53.890 55.829
611400 91.047 -53.910 55.802
611401 91.076 -53.930 55.777
611402 91.109 -53.950 55.754
611403 91.143 -53.969 55.732
611404 91.176 -53.989 55.711
611405 91.209 -54.010 55.692
611406 91.243 -54.029 55.674
611407 91.276 -54.049 55.658
611408 91.309 -54.069 55.642
611409 91.343 -54.088 55.631
611410 91.376 -54.109 55.622
611411 91.409 -54.129 55.614
611412 91.442 -54.148 55.608
611413 91.476 -54.168 55.604
611414 91.509 -54.188 55.601
611415 91.542 -54.208 55.596
611416 91.580 -54.225 55.590
611417 91.633 -54.230 55.579
611418 91.685 -54.236 55.563
611419 91.738 -54.240 55.540
611420 91.774 -54.251 55.514
611421 91.810 -54.263 55.491
611422 91.846 -54.273 55.470
611423 91.882 -54.285 55.453
611424 91.918 -54.296 55.438
611425 91.954 -54.308 55.423
611426 91.990 -54.318 55.408
611427 92.025 -54.330 55.394
611428 92.062 -54.341 55.380
611429 92.097 -54.352 55.368
611430 92.133 -54.363 55.356
611431 92.169 -54.374 55.345
611438 92.439 -54.461 55.290
611439 92.477 -54.474 55.286
611440 92.515 -54.486 55.283
611441 92.554 -54.499 55.283
611442 92.592 -54.511 55.285
611443 92.631 -54.524 55.289
611444 92.669 -54.536 55.294
611445 92.708 -54.549 55.300
611446 92.746 -54.561 55.309
611447 92.772 -54.587 55.321
611448 92.795 -54.615 55.336
611449 92.821 -54.635 55.352
611450 92.848 -54.652 55.366
611451 92.875 -54.670 55.380
611452 92.903 -54.688 55.395
611453 92.928 -54.702 55.411
611454 92.952 -54.715 55.426
611455 92.977 -54.729 55.441
611456 93.002 -54.743 55.455
611457 93.026 -54.756 55.468
611458 93.051 -54.770 55.476
611459 93.076 -54.784 55.483
611460 93.101 -54.797 55.486
611461 93.126 -54.811 55.483
611462 93.151 -54.827 55.473
611463 93.177 -54.845 55.465
611464 93.203 -54.862 55.457
611465 93.233 -54.881 55.454
611466 93.269 -54.899 55.456
611467 93.304 -54.917 55.462
611468 93.340 -54.934 55.473
611469 93.375 -54.953 55.492
611470 93.408 -54.969 55.511
611471 93.433 -54.984 55.522
611472 93.459 -54.997 55.525
611473 93.484 -55.011 55.521
611474 93.509 -55.025 55.511
611475 93.534 -55.038 55.498
611476 93.559 -55.053 55.486
611477 93.584 -55.066 55.473
611478 93.609 -55.080 55.456
611479 93.634 -55.094 55.438
611480 93.659 -55.108 55.419
611481 93.684 -55.122 55.398
611482 93.709 -55.135 55.383
611483 93.734 -55.151 55.367
611484 93.754 -55.175 55.353
611485 93.774 -55.198 55.341
611486 93.794 -55.223 55.330
611487 93.815 -55.247 55.319
611488 93.835 -55.271 55.306
611489 93.855 -55.295 55.290
611490 93.875 -55.320 55.273
611491 93.896 -55.343 55.255
611492 93.916 -55.367 55.233
611493 93.936 -55.392 55.210
611494 93.957 -55.416 55.181
611495 93.977 -55.440 55.146
611496 94.000 -55.458 55.106
611497 94.025 -55.475 55.064
611498 94.049 -55.492 55.021
611499 94.073 -55.508 54.979
611500 94.102 -55.522 54.942
611501 94.135 -55.534 54.901
611502 94.168 -55.546 54.859
611503 94.201 -55.558 54.816
611504 94.234 -55.570 54.773
611505 94.265 -55.579 54.730
611506 94.295 -55.588 54.687
611507 94.326 -55.597 54.643
611508 94.356 -55.606 54.600
611509 94.387 -55.614 54.558
611510 94.417 -55.624 54.521
611511 94.448 -55.633 54.488
611512 94.497 -55.646 54.460
611513 94.547 -55.658 54.434
611514 94.597 -55.671 54.410
611515 94.646 -55.683 54.389
611516 94.696 -55.696 54.368
611517 94.739 -55.709 54.344
611518 94.779 -55.721 54.323
611519 94.819 -55.734 54.293
611520 94.855 -55.746 54.260
611521 94.883 -55.756 54.221
611522 94.911 -55.765 54.176
611523 94.940 -55.775 54.124
611524 94.968 -55.785 54.067
611525 94.996 -55.795 53.998
611526 95.025 -55.805 53.935
611527 95.053 -55.815 53.874
611528 95.082 -55.824 53.813
611529 95.110 -55.834 53.747
611530 95.138 -55.843 53.678
611531 95.167 -55.856 53.608
611532 95.196 -55.872 53.539
611533 95.228 -55.885 53.468
611534 95.261 -55.898 53.393
611535 95.294 -55.911 53.316
611536 95.326 -55.923 53.236
611537 95.358 -55.934 53.156
611538 95.391 -55.947 53.080
611539 95.423 -55.959 53.007
611540 95.456 -55.971 52.935
611541 95.490 -55.982 52.863
611542 95.525 -55.994 52.792
611543 95.559 -56.005 52.726
611544 95.593 -56.017 52.660
611545 95.620 -56.036 52.595
611546 95.647 -56.055 52.532
611547 95.673 -56.074 52.469
611548 95.700 -56.094 52.409
611549 95.729 -56.110 52.350
611550 95.760 -56.124 52.293
611551 95.791 -56.137 52.236
611552 95.822 -56.150 52.181
611553 95.855 -56.165 52.127
611554 95.890 -56.179 52.076
611555 95.924 -56.192 52.024
611556 95.958 -56.206 51.963
611557 95.990 -56.210 51.894
611558 96.022 -56.213 51.824
611559 96.054 -56.215 51.756
611560 96.086 -56.216 51.689
611561 96.117 -56.219 51.627
611562 96.149 -56.221 51.567
611563 96.181 -56.222 51.509
611564 96.213 -56.225 51.453
611565 96.245 -56.226 51.399
611566 96.276 -56.229 51.348
611567 96.309 -56.231 51.301
611568 96.340 -56.232 51.256
611569 96.365 -56.235 51.212
611570 96.381 -56.235 51.170
611571 96.396 -56.236 51.129
611572 96.411 -56.238 51.090
611573 96.427 -56.238 51.056
611574 96.443 -56.240 51.021
611575 96.458 -56.241 50.988
611576 96.473 -56.242 50.954
611577 96.488 -56.243 50.920
611578 96.503 -56.244 50.888
611579 96.515 -56.245 50.857
611580 96.527 -56.247 50.827
611581 96.539 -56.248 50.796
611582 96.552 -56.249 50.766
611583 96.565 -56.251 50.735
611584 96.577 -56.252 50.705
611585 96.589 -56.253 50.678
611586 96.601 -56.254 50.650
611587 96.614 -56.256 50.621
611588 96.630 -56.255 50.593
611589 96.647 -56.254 50.565
611590 96.665 -56.253 50.539
611591 96.700 -56.252 50.515
611592 96.735 -56.251 50.490
611593 96.770 -56.250 50.465
611594 96.805 -56.249 50.441
611595 96.840 -56.248 50.416
611596 96.875 -56.247 50.392
611597 96.910 -56.246 50.369
611598 96.945 -56.245 50.347
611599 96.980 -56.244 50.326
611600 97.015 -56.243 50.303
611601 97.049 -56.241 50.281
611602 97.085 -56.241 50.261
611603 97.119 -56.240 50.243
611604 97.155 -56.238 50.226
611605 97.189 -56.238 50.208
611606 97.225 -56.237 50.192
611607 97.259 -56.235 50.175
611608 97.295 -56.235 50.160
611609 97.329 -56.234 50.147
611610 97.365 -56.232 50.132
611611 97.399 -56.232 50.118
611612 97.435 -56.230 50.104
611613 97.469 -56.229 50.091
611614 97.504 -56.229 50.078
611615 97.539 -56.227 50.066
611616 97.574 -56.226 50.054
611617 97.609 -56.226 50.044
611618 97.644 -56.224 50.033
611619 97.679 -56.223 50.022
611620 97.702 -56.222 50.012
611621 97.736 -56.226 50.004
611622 97.776 -56.234 49.995
611623 97.816 -56.243 49.986
611624 97.857 -56.251 49.978
611625 97.897 -56.260 49.970
611626 97.938 -56.269 49.962
611627 97.978 -56.277 49.954
611628 98.019 -56.285 49.946
611629 98.059 -56.294 49.937
611630 98.099 -56.303 49.927
611631 98.140 -56.311 49.915
611632 98.180 -56.320 49.901
611633 98.221 -56.328 49.885
611634 98.261 -56.337 49.867
611635 98.301 -56.345 49.848
611636 98.342 -56.354 49.826
611637 98.379 -56.362 49.803
611638 98.413 -56.369 49.778
611639 98.448 -56.376 49.752
611640 98.483 -56.384 49.725
611641 98.517 -56.391 49.696
611642 98.552 -56.398 49.667
611643 98.586 -56.406 49.638
611644 98.621 -56.413 49.608
611645 98.655 -56.420 49.578
611646 98.690 -56.428 49.548
611647 98.724 -56.435 49.519
611648 98.759 -56.442 49.491
611649 98.793 -56.450 49.464
611650 98.828 -56.457 49.437
611651 98.862 -56.464 49.412
611652 98.897 -56.472 49.388
611653 98.931 -56.479 49.364
611654 98.966 -56.486 49.341
611655 99.000 -56.494 49.318
611656 99.034 -56.501 49.295
611657 99.069 -56.508 49.273
611658 99.103 -56.516 49.253
611659 99.138 -56.523 49.233
611660 99.172 -56.530 49.214
611661 99.207 -56.538 49.196
611662 99.241 -56.545 49.180
611663 99.276 -56.552 49.165
611664 99.310 -56.559 49.152
611665 99.345 -56.567 49.139
611666 99.379 -56.574 49.126
611667 99.410 -56.577 49.112
611668 99.436 -56.578 49.101
611669 99.462 -56.592 49.088
611670 99.488 -56.611 49.076
611671 99.514 -56.629 49.063
611672 99.547 -56.646 49.049
611673 99.579 -56.661 49.036
611674 99.606 -56.670 49.022
611675 99.632 -56.680 49.007
611676 99.658 -56.689 48.991
611677 99.686 -56.705 48.974
611678 99.714 -56.721 48.956
611679 99.742 -56.739 48.937
611680 99.770 -56.756 48.918
611681 99.797 -56.773 48.898
611682 99.825 -56.790 48.878
611683 99.853 -56.808 48.857
611684 99.881 -56.825 48.836
612031 88.910 -52.713 57.343
612032 88.941 -52.736 57.339
612033 88.973 -52.759 57.334
612034 89.004 -52.782 57.330
612035 89.036 -52.804 57.326
612036 89.068 -52.827 57.320
612037 89.100 -52.850 57.315
612038 89.131 -52.872 57.310
612039 89.163 -52.895 57.305
612040 89.195 -52.918 57.300
612041 89.226 -52.941 57.296
612042 89.258 -52.963 57.291
612043 89.290 -52.986 57.287
612044 89.321 -53.009 57.283
612045 89.353 -53.031 57.279
612046 89.385 -53.054 57.276
612047 89.416 -53.076 57.273
612048 89.448 -53.100 57.271
612049 89.470 -53.132 57.266
612050 89.502 -53.151 57.260
612051 89.536 -53.164 57.254
612052 89.559 -53.182 57.247
612053 89.583 -53.199 57.241
612054 89.606 -53.217 57.235
612055 89.629 -53.234 57.227
612056 89.653 -53.252 57.218
612057 89.676 -53.269 57.208
612058 89.707 -53.291 57.198
612059 89.749 -53.317 57.186
612060 89.790 -53.345 57.177
612061 89.831 -53.371 57.165
612062 89.872 -53.396 57.152
612063 89.910 -53.402 57.139
612064 89.941 -53.408 57.124
612065 89.962 -53.415 57.106
612066 89.982 -53.422 57.088
612067 90.003 -53.430 57.071
612068 90.024 -53.436 57.053
612069 90.044 -53.444 57.033
612070 90.064 -53.452 57.011
612071 90.085 -53.458 56.983
612072 90.120 -53.474 56.952
612073 90.163 -53.486 56.920
612074 90.206 -53.497 56.887
612075 90.250 -53.508 56.854
612076 90.293 -53.519 56.820
612077 90.336 -53.531 56.784
612078 90.380 -53.542 56.748
612079 90.423 -53.552 56.713
612080 90.466 -53.564 56.679
612081 90.506 -53.577 56.646
612082 90.523 -53.596 56.613
612090 90.737 -53.681 56.134
612091 90.773 -53.683 56.100
612092 90.806 -53.692 56.068
612093 90.835 -53.712 56.038
612094 90.864 -53.732 56.007
612095 90.893 -53.752 55.976
612096 90.922 -53.772 55.947
612097 90.951 -53.793 55.917
612098 90.979 -53.813 55.888
612099 91.009 -53.833 55.862
612100 91.037 -53.853 55.836
612101 91.068 -53.873 55.809
612102 91.101 -53.893 55.783
612103 91.134 -53.913 55.760
612104 91.167 -53.933 55.737
612105 91.201 -53.953 55.716
612106 91.234 -53.972 55.696
612107 91.267 -53.992 55.678
612108 91.301 -54.012 55.662
612109 91.334 -54.032 55.646
612110 91.367 -54.052 55.632
612111 91.400 -54.072 55.621
612112 91.434 -54.091 55.612
612113 91.467 -54.111 55.604
612114 91.501 -54.132 55.599
612115 91.534 -54.151 55.593
612116 91.567 -54.171 55.585
612117 91.601 -54.190 55.575
612118 91.654 -54.195 55.563
612119 91.707 -54.201 55.548
612120 91.749 -54.210 55.528
612121 91.785 -54.220 55.506
612122 91.821 -54.232 55.486
612123 91.857 -54.243 55.467
612124 91.893 -54.254 55.450
612125 91.929 -54.265 55.436
612126 91.965 -54.276 55.422
612127 92.001 -54.288 55.408
612128 92.036 -54.298 55.394
612129 92.074 -54.311 55.380
612130 92.112 -54.323 55.367
612131 92.151 -54.336 55.356
612132 92.190 -54.348 55.345
612139 92.458 -54.435 55.289
612140 92.497 -54.448 55.285
612141 92.535 -54.460 55.283
612142 92.574 -54.473 55.283
612143 92.613 -54.485 55.286
612144 92.651 -54.498 55.290
612145 92.689 -54.510 55.295
612146 92.728 -54.522 55.300
612147 92.766 -54.535 55.308
612148 92.795 -54.558 55.318
612149 92.818 -54.586 55.332
612150 92.842 -54.613 55.346
612151 92.868 -54.630 55.360
612152 92.896 -54.649 55.373
612153 92.923 -54.666 55.385
612154 92.948 -54.681 55.400
612155 92.973 -54.694 55.415
612156 92.997 -54.708 55.430
612157 93.022 -54.721 55.446
612158 93.047 -54.735 55.460
612159 93.072 -54.749 55.467
612160 93.096 -54.762 55.471
612161 93.121 -54.775 55.471
612162 93.146 -54.789 55.464
612163 93.170 -54.803 55.451
612164 93.196 -54.818 55.446
612165 93.222 -54.837 55.446
612166 93.249 -54.855 55.452
612167 93.285 -54.872 55.462
612168 93.320 -54.891 55.476
612169 93.356 -54.909 55.493
612170 93.390 -54.925 55.515
612171 93.415 -54.940 55.531
612172 93.440 -54.953 55.538
612173 93.465 -54.967 55.537
612174 93.490 -54.981 55.529
612175 93.515 -54.995 55.519
612176 93.540 -55.009 55.508
612177 93.565 -55.022 55.497
612178 93.590 -55.036 55.486
612179 93.616 -55.050 55.471
612180 93.641 -55.064 55.452
612181 93.665 -55.078 55.430
612182 93.691 -55.091 55.407
612183 93.716 -55.105 55.388
612184 93.741 -55.119 55.369
612185 93.762 -55.141 55.352
612186 93.782 -55.166 55.337
612187 93.802 -55.190 55.324
612188 93.823 -55.214 55.314
612189 93.843 -55.238 55.301
612190 93.863 -55.262 55.285
612191 93.884 -55.286 55.269
612192 93.904 -55.311 55.252
612193 93.924 -55.335 55.232
612194 93.944 -55.359 55.211
612195 93.964 -55.383 55.183
612196 93.986 -55.405 55.147
612197 94.010 -55.422 55.108
612198 94.034 -55.439 55.067
612199 94.058 -55.455 55.023
612200 94.083 -55.472 54.982
612201 94.111 -55.486 54.944
612202 94.144 -55.498 54.905
612203 94.177 -55.510 54.863
612204 94.210 -55.522 54.820
612205 94.243 -55.534 54.778
612206 94.275 -55.546 54.738
612207 94.306 -55.555 54.700
612208 94.336 -55.563 54.658
612209 94.367 -55.572 54.615
612210 94.397 -55.581 54.574
612211 94.428 -55.590 54.537
612212 94.461 -55.599 54.504
612213 94.510 -55.612 54.477
612214 94.560 -55.625 54.451
612215 94.610 -55.637 54.428
612216 94.659 -55.650 54.405
612217 94.704 -55.663 54.380
612218 94.744 -55.676 54.351
612219 94.784 -55.688 54.323
612220 94.824 -55.701 54.288
612221 94.862 -55.714 54.250
612222 94.890 -55.723 54.210
612223 94.918 -55.733 54.165
612224 94.947 -55.743 54.114
612225 94.975 -55.752 54.059
612226 95.003 -55.762 53.999
612227 95.032 -55.772 53.940
612228 95.060 -55.782 53.880
612229 95.088 -55.792 53.817
612230 95.117 -55.801 53.751
612231 95.145 -55.812 53.681
612232 95.174 -55.828 53.610
612233 95.203 -55.843 53.540
612234 95.235 -55.858 53.469
612235 95.268 -55.871 53.396
612236 95.301 -55.884 53.321
612237 95.333 -55.898 53.245
612238 95.366 -55.910 53.169
612239 95.398 -55.922 53.096
612240 95.431 -55.934 53.024
612241 95.463 -55.947 52.950
612242 95.497 -55.958 52.877
612243 95.532 -55.969 52.806
612244 95.566 -55.981 52.737
612245 95.595 -55.998 52.669
612246 95.621 -56.017 52.603
612247 95.648 -56.036 52.539
612248 95.676 -56.053 52.475
612249 95.707 -56.066 52.412
612250 95.739 -56.080 52.352
612251 95.770 -56.094 52.294
612252 95.802 -56.107 52.237
612253 95.834 -56.121 52.180
612254 95.868 -56.135 52.126
612255 95.902 -56.149 52.074
612256 95.936 -56.163 52.022
612257 95.970 -56.176 51.965
612258 96.004 -56.189 51.899
612259 96.036 -56.191 51.829
612260 96.068 -56.193 51.758
612261 96.100 -56.195 51.690
612262 96.131 -56.197 51.627
612263 96.163 -56.199 51.566
612264 96.195 -56.201 51.507
612265 96.227 -56.203 51.450
612266 96.259 -56.205 51.394
612267 96.291 -56.207 51.341
612268 96.322 -56.209 51.293
612269 96.355 -56.211 51.247
612270 96.376 -56.212 51.202
612271 96.391 -56.213 51.158
612272 96.406 -56.214 51.116
612273 96.422 -56.216 51.075
612274 96.438 -56.216 51.039
612275 96.453 -56.218 51.003
612276 96.468 -56.219 50.968
612277 96.484 -56.220 50.933
612278 96.498 -56.221 50.898
612279 96.510 -56.222 50.864
612280 96.523 -56.224 50.833
612281 96.535 -56.225 50.802
612282 96.547 -56.226 50.771
612283 96.559 -56.227 50.739
612284 96.572 -56.229 50.708
612285 96.585 -56.230 50.677
612286 96.597 -56.231 50.649
612287 96.611 -56.232 50.621
612288 96.628 -56.231 50.593
612289 96.645 -56.230 50.564
612290 96.661 -56.229 50.536
612291 96.681 -56.228 50.509
612292 96.716 -56.227 50.485
612293 96.751 -56.226 50.459
612294 96.786 -56.225 50.435
612295 96.821 -56.224 50.410
612296 96.856 -56.223 50.385
612297 96.891 -56.222 50.361
612298 96.926 -56.221 50.340
612299 96.961 -56.219 50.318
612300 96.996 -56.219 50.296
612301 97.031 -56.218 50.273
612302 97.066 -56.216 50.252
612303 97.101 -56.216 50.231
612304 97.136 -56.215 50.212
612305 97.171 -56.213 50.193
612306 97.206 -56.213 50.176
612307 97.241 -56.211 50.159
612308 97.276 -56.210 50.142
612309 97.311 -56.210 50.127
612310 97.346 -56.208 50.112
612311 97.380 -56.207 50.098
612312 97.416 -56.207 50.082
612313 97.450 -56.205 50.069
612314 97.486 -56.204 50.055
612315 97.520 -56.204 50.041
612316 97.556 -56.202 50.029
612317 97.590 -56.201 50.016
612318 97.626 -56.200 50.005
612319 97.660 -56.199 49.995
612320 97.687 -56.198 49.983
612321 97.710 -56.195 49.973
612322 97.749 -56.203 49.964
612323 97.789 -56.211 49.956
612324 97.830 -56.220 49.948
612325 97.870 -56.229 49.939
612326 97.911 -56.237 49.931
612327 97.951 -56.245 49.922
612328 97.992 -56.254 49.914
612329 98.032 -56.263 49.904
612330 98.072 -56.271 49.895
612331 98.112 -56.280 49.884
612332 98.153 -56.288 49.872
612333 98.194 -56.297 49.858
612334 98.234 -56.305 49.842
612335 98.275 -56.314 49.825
612336 98.315 -56.323 49.806
612337 98.355 -56.331 49.784
612338 98.389 -56.338 49.761
612339 98.424 -56.345 49.737
612340 98.458 -56.353 49.712
612341 98.493 -56.360 49.685
612342 98.527 -56.367 49.657
612343 98.562 -56.375 49.627
612344 98.596 -56.382 49.599
612345 98.631 -56.389 49.569
612346 98.665 -56.397 49.539
612347 98.700 -56.404 49.509
612348 98.734 -56.411 49.480
612349 98.769 -56.419 49.452
612350 98.803 -56.426 49.425
612351 98.838 -56.433 49.398
612352 98.872 -56.441 49.372
612353 98.907 -56.448 49.348
612354 98.941 -56.455 49.323
612355 98.976 -56.463 49.300
612356 99.010 -56.470 49.276
612357 99.045 -56.477 49.254
612358 99.079 -56.485 49.232
612359 99.114 -56.492 49.212
612360 99.148 -56.499 49.192
612361 99.183 -56.507 49.173
612362 99.217 -56.514 49.156
612363 99.252 -56.521 49.139
612364 99.286 -56.529 49.124
612365 99.321 -56.536 49.110
612366 99.355 -56.543 49.097
612367 99.385 -56.547 49.085
612368 99.412 -56.547 49.072
612369 99.438 -56.556 49.061
612370 99.464 -56.574 49.049
612371 99.489 -56.593 49.038
612372 99.518 -56.611 49.026
612373 99.551 -56.627 49.013
612374 99.581 -56.640 49.001
612375 99.608 -56.649 48.988
612376 99.634 -56.659 48.975
612377 99.660 -56.668 48.960
612378 99.688 -56.683 48.945
612379 99.716 -56.701 48.929
612380 99.744 -56.718 48.911
612381 99.771 -56.735 48.894
612382 99.799 -56.752 48.875
612383 99.827 -56.769 48.857
612384 99.855 -56.787 48.837
612731 88.894 -52.667 57.339
612732 88.926 -52.690 57.335
612733 88.958 -52.713 57.330
612734 88.989 -52.735 57.325
612735 89.021 -52.758 57.320
612736 89.052 -52.780 57.316
612737 89.084 -52.803 57.311
612738 89.116 -52.825 57.306
612739 89.147 -52.848 57.301
612740 89.179 -52.871 57.296
612741 89.211 -52.894 57.291
612742 89.242 -52.916 57.286
612743 89.274 -52.939 57.281
612744 89.306 -52.962 57.277
612745 89.337 -52.985 57.272
612746 89.369 -53.007 57.269
612747 89.401 -53.030 57.266
612748 89.433 -53.053 57.264
612749 89.463 -53.077 57.261
612750 89.488 -53.105 57.255
612751 89.522 -53.118 57.249
612752 89.545 -53.136 57.243
612753 89.568 -53.153 57.236
612754 89.592 -53.170 57.230
612755 89.615 -53.188 57.224
612756 89.638 -53.205 57.218
612757 89.662 -53.223 57.210
612758 89.685 -53.240 57.201
612759 89.718 -53.263 57.192
612760 89.759 -53.289 57.182
612761 89.801 -53.317 57.173
612762 89.842 -53.343 57.162
612763 89.881 -53.352 57.150
612764 89.919 -53.357 57.138
612765 89.956 -53.362 57.124
612766 89.976 -53.369 57.107
612767 89.996 -53.377 57.090
612768 90.017 -53.384 57.074
612769 90.038 -53.391 57.056
612770 90.058 -53.399 57.037
612771 90.079 -53.405 57.017
612772 90.103 -53.416 56.990
612773 90.139 -53.436 56.962
612774 90.179 -53.452 56.932
612775 90.222 -53.463 56.901
612776 90.265 -53.474 56.870
612777 90.308 -53.485 56.837
612778 90.352 -53.496 56.802
612779 90.395 -53.508 56.765
612780 90.438 -53.519 56.730
612781 90.481 -53.530 56.697
612782 90.514 -53.545 56.665
612783 90.531 -53.565 56.632
612784 90.548 -53.585 56.597
612792 90.796 -53.637 56.114
612793 90.825 -53.655 56.081
612794 90.854 -53.675 56.050
612795 90.883 -53.696 56.019
612796 90.912 -53.715 55.987
612797 90.941 -53.736 55.956
612798 90.970 -53.756 55.927
612799 90.999 -53.776 55.898
612800 91.028 -53.797 55.870
612801 91.059 -53.816 55.844
612802 91.092 -53.836 55.817
612803 91.126 -53.856 55.790
612804 91.159 -53.876 55.766
612805 91.192 -53.896 55.744
612806 91.225 -53.916 55.722
612807 91.258 -53.935 55.702
612808 91.292 -53.955 55.683
612809 91.325 -53.976 55.666
612810 91.359 -53.995 55.650
612811 91.392 -54.015 55.635
612812 91.425 -54.035 55.623
612813 91.458 -54.054 55.612
612814 91.492 -54.075 55.602
612815 91.525 -54.095 55.594
612816 91.559 -54.114 55.586
612817 91.592 -54.134 55.577
612818 91.625 -54.154 55.566
612819 91.675 -54.161 55.552
612820 91.724 -54.167 55.537
612821 91.760 -54.179 55.519
612822 91.796 -54.190 55.500
612823 91.832 -54.201 55.482
612824 91.868 -54.212 55.465
612825 91.904 -54.223 55.448
612826 91.940 -54.235 55.433
612827 91.979 -54.247 55.421
612828 92.017 -54.259 55.407
612829 92.056 -54.272 55.394
612830 92.094 -54.284 55.380
612831 92.133 -54.297 55.367
612832 92.171 -54.309 55.356
612840 92.479 -54.409 55.290
612841 92.517 -54.421 55.287
612842 92.556 -54.434 55.284
612843 92.594 -54.446 55.284
612844 92.632 -54.458 55.287
612845 92.671 -54.471 55.291
612846 92.709 -54.483 55.296
612847 92.747 -54.496 55.301
612848 92.786 -54.508 55.307
612849 92.818 -54.528 55.316
612850 92.841 -54.557 55.329
612851 92.864 -54.586 55.341
612852 92.888 -54.609 55.354
612853 92.916 -54.627 55.366
612854 92.943 -54.644 55.376
612855 92.968 -54.659 55.389
612856 92.993 -54.672 55.405
612857 93.017 -54.686 55.422
612858 93.043 -54.699 55.440
612859 93.067 -54.713 55.453
612860 93.092 -54.727 55.460
612861 93.117 -54.740 55.462
612862 93.141 -54.754 55.459
612863 93.166 -54.768 55.449
612864 93.191 -54.781 55.438
612865 93.216 -54.795 55.437
612866 93.241 -54.810 55.446
612867 93.267 -54.828 55.462
612868 93.300 -54.847 55.478
612869 93.336 -54.864 55.492
612870 93.371 -54.882 55.507
612871 93.396 -54.896 55.520
612872 93.421 -54.909 55.530
612873 93.446 -54.923 55.537
612874 93.471 -54.938 55.538
612875 93.496 -54.951 55.533
612876 93.521 -54.965 55.527
612877 93.547 -54.978 55.518
612878 93.572 -54.992 55.506
612879 93.597 -55.007 55.494
612880 93.622 -55.020 55.480
612881 93.647 -55.034 55.462
612882 93.672 -55.047 55.437
612883 93.697 -55.062 55.411
612884 93.722 -55.075 55.391
612885 93.748 -55.089 55.370
612886 93.770 -55.108 55.350
612887 93.790 -55.132 55.332
612888 93.811 -55.157 55.316
612889 93.831 -55.181 55.304
612890 93.851 -55.205 55.290
612891 93.871 -55.229 55.274
612892 93.891 -55.254 55.257
612893 93.912 -55.277 55.241
612894 93.932 -55.301 55.223
612895 93.952 -55.326 55.202
612896 93.972 -55.350 55.174
612897 93.995 -55.369 55.140
612898 94.019 -55.386 55.103
612899 94.044 -55.402 55.064
612900 94.068 -55.419 55.023
612901 94.092 -55.436 54.983
612902 94.120 -55.450 54.945
612903 94.153 -55.462 54.905
612904 94.186 -55.474 54.864
612905 94.219 -55.486 54.826
612906 94.252 -55.498 54.789
612907 94.285 -55.510 54.755
612908 94.317 -55.520 54.718
612909 94.347 -55.530 54.675
612910 94.378 -55.539 54.631
612911 94.408 -55.547 54.589
612912 94.438 -55.556 54.553
612913 94.474 -55.566 54.521
612914 94.524 -55.579 54.495
612915 94.573 -55.592 54.469
612916 94.623 -55.604 54.445
612917 94.670 -55.617 54.420
612918 94.710 -55.630 54.392
612919 94.750 -55.643 54.360
612920 94.789 -55.655 54.325
612921 94.829 -55.668 54.285
612922 94.868 -55.680 54.243
612923 94.897 -55.690 54.201
612924 94.925 -55.700 54.156
612925 94.954 -55.710 54.107
612926 94.982 -55.720 54.055
612927 95.011 -55.730 54.000
612928 95.039 -55.740 53.942
612929 95.067 -55.749 53.881
612930 95.095 -55.759 53.817
612931 95.123 -55.768 53.750
612932 95.153 -55.784 53.679
612933 95.182 -55.800 53.608
612934 95.211 -55.816 53.539
612935 95.242 -55.831 53.470
612936 95.275 -55.844 53.398
612937 95.308 -55.857 53.326
612938 95.340 -55.871 53.255
612939 95.373 -55.884 53.181
612940 95.406 -55.897 53.109
612941 95.439 -55.910 53.037
612942 95.471 -55.922 52.966
612943 95.503 -55.934 52.894
612944 95.538 -55.945 52.821
612945 95.569 -55.959 52.751
612946 95.596 -55.978 52.682
612947 95.624 -55.996 52.614
612948 95.656 -56.009 52.548
612949 95.687 -56.023 52.483
612950 95.718 -56.037 52.421
612951 95.749 -56.050 52.358
612952 95.780 -56.063 52.298
612953 95.813 -56.077 52.239
612954 95.847 -56.091 52.181
612955 95.881 -56.105 52.125
612956 95.915 -56.119 52.071
612957 95.949 -56.133 52.019
612958 95.983 -56.147 51.963
612959 96.017 -56.160 51.901
612960 96.050 -56.169 51.832
612961 96.082 -56.171 51.761
612962 96.114 -56.173 51.692
612963 96.146 -56.175 51.629
612964 96.177 -56.177 51.568
612965 96.209 -56.179 51.508
612966 96.241 -56.181 51.449
612967 96.273 -56.183 51.392
612968 96.305 -56.185 51.337
612969 96.336 -56.187 51.287
612970 96.368 -56.189 51.240
612971 96.386 -56.190 51.194
612972 96.401 -56.191 51.149
612973 96.417 -56.192 51.105
612974 96.433 -56.194 51.062
612975 96.448 -56.194 51.024
612976 96.463 -56.196 50.987
612977 96.479 -56.197 50.950
612978 96.493 -56.198 50.914
612979 96.506 -56.199 50.878
612980 96.518 -56.201 50.844
612981 96.530 -56.202 50.812
612982 96.543 -56.203 50.780
612983 96.555 -56.204 50.748
612984 96.568 -56.206 50.715
612985 96.580 -56.207 50.682
612986 96.592 -56.208 50.651
612987 96.609 -56.207 50.622
612988 96.625 -56.207 50.595
612989 96.642 -56.206 50.566
612990 96.659 -56.204 50.537
612991 96.675 -56.204 50.508
612992 96.697 -56.203 50.481
612993 96.732 -56.202 50.457
612994 96.767 -56.201 50.431
612995 96.802 -56.200 50.407
612996 96.837 -56.199 50.382
612997 96.872 -56.197 50.358
612998 96.907 -56.197 50.334
612999 96.942 -56.196 50.313
613000 96.977 -56.194 50.290
613001 97.012 -56.194 50.267
613002 97.047 -56.193 50.245
613003 97.082 -56.191 50.223
613004 97.117 -56.191 50.202
613005 97.152 -56.189 50.182
613006 97.187 -56.188 50.165
613007 97.222 -56.188 50.145
613008 97.256 -56.186 50.128
613009 97.292 -56.185 50.110
613010 97.326 -56.185 50.094
613011 97.362 -56.183 50.079
613012 97.396 -56.182 50.064
613013 97.432 -56.182 50.048
613014 97.466 -56.180 50.034
613015 97.502 -56.179 50.019
613016 97.536 -56.178 50.005
613017 97.572 -56.177 49.992
613018 97.606 -56.176 49.981
613019 97.642 -56.175 49.968
613020 97.673 -56.174 49.958
613021 97.696 -56.173 49.947
613022 97.722 -56.172 49.936
613023 97.762 -56.180 49.927
613024 97.803 -56.188 49.919
613025 97.843 -56.197 49.910
613026 97.883 -56.206 49.900
613027 97.924 -56.214 49.893
613028 97.964 -56.222 49.884
613029 98.005 -56.232 49.875
613030 98.045 -56.240 49.864
613031 98.086 -56.248 49.854
613032 98.126 -56.257 49.843
613033 98.166 -56.266 49.831
613034 98.207 -56.274 49.817
613035 98.247 -56.282 49.801
613036 98.288 -56.291 49.784
613037 98.328 -56.300 49.765
613038 98.365 -56.307 49.744
613039 98.399 -56.315 49.722
613040 98.434 -56.322 49.698
613041 98.468 -56.329 49.673
613042 98.503 -56.337 49.646
613043 98.537 -56.344 49.618
613044 98.572 -56.351 49.589
613045 98.606 -56.359 49.561
613046 98.641 -56.366 49.531
613047 98.675 -56.373 49.501
613048 98.710 -56.381 49.472
613049 98.744 -56.388 49.442
613050 98.779 -56.395 49.415
613051 98.813 -56.403 49.386
613052 98.848 -56.410 49.359
613053 98.882 -56.417 49.334
613054 98.917 -56.425 49.309
613055 98.951 -56.432 49.284
613056 98.986 -56.439 49.260
613057 99.020 -56.447 49.236
613058 99.055 -56.454 49.214
613059 99.089 -56.461 49.192
613060 99.124 -56.469 49.172
613061 99.158 -56.476 49.152
613062 99.193 -56.483 49.133
613063 99.227 -56.491 49.116
613064 99.262 -56.498 49.099
613065 99.296 -56.505 49.083
613066 99.331 -56.513 49.069
613067 99.361 -56.516 49.056
613068 99.388 -56.517 49.043
613069 99.415 -56.520 49.032
613070 99.440 -56.539 49.021
613071 99.466 -56.557 49.010
613072 99.491 -56.576 48.999
613073 99.521 -56.592 48.987
613074 99.554 -56.609 48.976
613075 99.583 -56.620 48.965
613076 99.609 -56.629 48.954
613077 99.636 -56.639 48.941
613078 99.662 -56.648 48.928
613079 99.690 -56.662 48.914
613080 99.717 -56.680 48.900
613081 99.745 -56.696 48.884
613082 99.773 -56.714 48.868
613083 99.801 -56.731 48.851
613431 88.879 -52.620 57.335
613432 88.910 -52.643 57.331
613433 88.942 -52.666 57.326
613434 88.974 -52.688 57.321
613435 89.005 -52.711 57.315
613436 89.037 -52.734 57.310
613437 89.068 -52.756 57.306
613438 89.100 -52.779 57.301
613439 89.132 -52.802 57.296
613440 89.164 -52.825 57.291
613441 89.195 -52.847 57.285
613442 89.227 -52.870 57.280
613443 89.258 -52.893 57.275
613444 89.290 -52.915 57.271
613445 89.322 -52.938 57.267
613446 89.353 -52.960 57.262
613447 89.385 -52.983 57.258
613448 89.417 -53.006 57.255
613449 89.449 -53.029 57.254
613450 89.478 -53.054 57.250
613451 89.508 -53.072 57.244
613452 89.531 -53.089 57.237
613453 89.554 -53.107 57.231
613454 89.578 -53.124 57.225
613455 89.601 -53.142 57.219
613456 89.624 -53.159 57.214
613457 89.648 -53.176 57.208
613458 89.671 -53.194 57.201
613459 89.694 -53.211 57.193
613460 89.729 -53.235 57.185
613461 89.770 -53.261 57.176
613462 89.811 -53.289 57.167
613463 89.850 -53.301 57.158
613464 89.889 -53.306 57.148
613465 89.928 -53.311 57.138
613466 89.966 -53.317 57.125
613467 89.990 -53.324 57.108
613468 90.011 -53.331 57.092
613469 90.031 -53.338 57.075
613470 90.052 -53.346 57.058
613471 90.073 -53.352 57.041
613472 90.093 -53.360 57.022
613473 90.124 -53.376 56.997
613474 90.160 -53.396 56.971
613475 90.195 -53.416 56.944
613476 90.237 -53.429 56.915
613477 90.281 -53.440 56.886
613478 90.324 -53.451 56.855
613479 90.367 -53.462 56.821
613480 90.410 -53.474 56.785
613481 90.453 -53.485 56.751
613482 90.497 -53.496 56.718
613483 90.521 -53.514 56.686
613484 90.538 -53.534 56.653
613485 90.555 -53.554 56.618
613486 90.573 -53.574 56.586
613494 90.845 -53.618 56.095
613495 90.874 -53.639 56.063
613496 90.902 -53.659 56.031
613497 90.932 -53.679 55.999
613498 90.960 -53.700 55.968
613499 90.990 -53.719 55.938
613500 91.018 -53.740 55.908
613501 91.050 -53.759 55.880
613502 91.083 -53.779 55.853
613503 91.117 -53.800 55.826
613504 91.150 -53.819 55.799
613505 91.183 -53.839 55.775
613506 91.217 -53.859 55.751
613507 91.250 -53.878 55.730
613508 91.283 -53.899 55.709
613509 91.317 -53.919 55.689
613510 91.350 -53.938 55.671
613511 91.383 -53.958 55.654
613512 91.417 -53.978 55.639
613513 91.450 -53.998 55.625
613514 91.483 -54.018 55.614
613515 91.516 -54.038 55.601
613516 91.550 -54.057 55.591
613517 91.583 -54.077 55.582
613518 91.616 -54.098 55.572
613519 91.650 -54.117 55.559
613520 91.697 -54.126 55.545
613521 91.735 -54.136 55.530
613522 91.772 -54.148 55.513
613523 91.807 -54.159 55.496
613524 91.845 -54.171 55.479
613525 91.883 -54.183 55.463
613526 91.922 -54.196 55.447
613527 91.960 -54.208 55.433
613528 91.999 -54.220 55.420
613529 92.037 -54.233 55.407
613530 92.076 -54.245 55.393
613531 92.114 -54.258 55.380
613532 92.152 -54.270 55.367
613541 92.499 -54.383 55.291
613542 92.537 -54.395 55.288
613543 92.575 -54.408 55.286
613544 92.614 -54.420 55.286
613545 92.652 -54.433 55.289
613546 92.691 -54.445 55.292
613547 92.729 -54.458 55.297
613548 92.768 -54.470 55.302
613549 92.806 -54.483 55.308
613550 92.841 -54.499 55.315
613551 92.864 -54.527 55.327
613552 92.886 -54.556 55.337
613553 92.909 -54.585 55.349
613554 92.936 -54.605 55.359
613555 92.963 -54.622 55.370
613556 92.988 -54.637 55.383
613557 93.013 -54.651 55.399
613558 93.038 -54.665 55.419
613559 93.063 -54.678 55.436
613560 93.087 -54.692 55.448
613561 93.112 -54.706 55.453
613562 93.137 -54.718 55.454
613563 93.161 -54.732 55.449
613564 93.187 -54.746 55.437
613565 93.211 -54.759 55.429
613566 93.236 -54.773 55.433
613567 93.261 -54.787 55.449
613568 93.286 -54.802 55.471
613569 93.317 -54.820 55.490
613570 93.352 -54.838 55.501
613571 93.377 -54.852 55.510
613572 93.403 -54.866 55.516
613573 93.428 -54.879 55.522
613574 93.452 -54.894 55.528
613575 93.478 -54.907 55.532
613576 93.503 -54.921 55.532
613577 93.528 -54.934 55.527
613578 93.553 -54.949 55.520
613579 93.578 -54.963 55.507
613580 93.603 -54.976 55.492
613581 93.628 -54.990 55.477
613582 93.653 -55.004 55.460
613583 93.679 -55.018 55.437
613584 93.704 -55.032 55.413
613585 93.729 -55.045 55.393
613586 93.754 -55.059 55.374
613587 93.778 -55.075 55.353
613588 93.798 -55.100 55.330
613589 93.818 -55.124 55.307
613590 93.839 -55.148 55.288
613591 93.859 -55.172 55.272
613592 93.879 -55.196 55.255
613593 93.899 -55.220 55.239
613594 93.919 -55.245 55.224
613595 93.939 -55.269 55.209
613596 93.960 -55.292 55.187
613597 93.981 -55.316 55.161
613598 94.005 -55.332 55.130
613599 94.029 -55.349 55.095
613600 94.054 -55.366 55.059
613601 94.078 -55.383 55.022
613602 94.102 -55.399 54.984
613603 94.129 -55.414 54.948
613604 94.161 -55.426 54.910
613605 94.195 -55.439 54.874
613606 94.227 -55.450 54.841
613607 94.261 -55.462 54.809
613608 94.293 -55.474 54.776
613609 94.326 -55.486 54.738
613610 94.357 -55.495 54.694
613611 94.388 -55.505 54.649
613612 94.418 -55.514 54.607
613613 94.449 -55.523 54.570
613614 94.487 -55.533 54.539
613615 94.537 -55.546 54.514
613616 94.586 -55.558 54.487
613617 94.636 -55.571 54.461
613618 94.676 -55.583 54.434
613619 94.715 -55.596 54.404
613620 94.755 -55.609 54.369
613621 94.795 -55.622 54.332
613622 94.835 -55.635 54.291
613623 94.875 -55.648 54.246
613624 94.904 -55.658 54.200
613625 94.932 -55.668 54.153
613626 94.961 -55.677 54.105
613627 94.989 -55.687 54.057
613628 95.017 -55.697 54.002
613629 95.046 -55.706 53.942
613630 95.074 -55.716 53.879
613631 95.102 -55.726 53.812
613632 95.131 -55.740 53.745
613633 95.161 -55.756 53.676
613634 95.190 -55.772 53.607
613635 95.219 -55.789 53.540
613636 95.249 -55.804 53.472
613637 95.282 -55.817 53.404
613638 95.315 -55.831 53.334
613639 95.347 -55.843 53.266
613640 95.380 -55.857 53.194
613641 95.413 -55.870 53.123
613642 95.446 -55.884 53.053
613643 95.478 -55.897 52.984
613644 95.511 -55.909 52.913
613645 95.544 -55.922 52.840
613646 95.572 -55.939 52.767
613647 95.604 -55.953 52.697
613648 95.635 -55.965 52.628
613649 95.666 -55.979 52.560
613650 95.697 -55.993 52.494
613651 95.729 -56.006 52.430
613652 95.760 -56.020 52.367
613653 95.791 -56.034 52.306
613654 95.825 -56.047 52.245
613655 95.860 -56.061 52.185
613656 95.893 -56.075 52.127
613657 95.927 -56.089 52.071
613658 95.961 -56.103 52.016
613659 95.995 -56.117 51.959
613660 96.030 -56.131 51.898
613661 96.063 -56.144 51.833
613662 96.096 -56.149 51.765
613663 96.128 -56.151 51.697
613664 96.160 -56.154 51.632
613665 96.191 -56.155 51.570
613666 96.223 -56.157 51.509
613667 96.255 -56.160 51.450
613668 96.287 -56.161 51.392
613669 96.319 -56.163 51.336
613670 96.350 -56.165 51.284
613671 96.381 -56.167 51.236
613672 96.396 -56.168 51.188
613673 96.412 -56.169 51.141
613674 96.428 -56.170 51.096
613675 96.443 -56.172 51.052
613676 96.458 -56.172 51.012
613677 96.474 -56.174 50.973
613678 96.488 -56.175 50.935
613679 96.501 -56.176 50.897
613680 96.513 -56.177 50.860
613681 96.526 -56.179 50.825
613682 96.538 -56.180 50.791
613683 96.550 -56.181 50.759
613684 96.563 -56.182 50.725
613685 96.575 -56.184 50.691
613686 96.590 -56.184 50.658
613687 96.607 -56.183 50.626
613688 96.623 -56.182 50.596
613689 96.640 -56.181 50.568
613690 96.657 -56.180 50.539
613691 96.673 -56.179 50.510
613692 96.690 -56.179 50.481
613693 96.713 -56.178 50.454
613694 96.748 -56.177 50.429
613695 96.783 -56.175 50.404
613696 96.818 -56.175 50.379
613697 96.853 -56.174 50.355
613698 96.888 -56.172 50.330
613699 96.923 -56.172 50.306
613700 96.958 -56.170 50.285
613701 96.993 -56.169 50.261
613702 97.028 -56.169 50.239
613703 97.063 -56.167 50.217
613704 97.098 -56.166 50.195
613705 97.133 -56.166 50.173
613706 97.168 -56.164 50.153
613707 97.203 -56.163 50.135
613708 97.238 -56.163 50.115
613709 97.273 -56.161 50.097
613710 97.308 -56.160 50.079
613711 97.343 -56.159 50.061
613712 97.378 -56.158 50.046
613713 97.413 -56.157 50.030
613714 97.448 -56.156 50.014
613715 97.483 -56.155 49.999
613716 97.518 -56.154 49.984
613717 97.552 -56.153 49.969
613718 97.587 -56.152 49.956
613719 97.623 -56.151 49.944
613720 97.657 -56.150 49.933
613721 97.681 -56.149 49.921
613722 97.704 -56.147 49.910
613723 97.735 -56.149 49.899
613724 97.776 -56.157 49.890
613725 97.816 -56.166 49.881
613726 97.856 -56.174 49.873
613727 97.896 -56.183 49.863
613728 97.937 -56.191 49.854
613729 97.978 -56.200 49.846
613730 98.018 -56.208 49.837
613731 98.059 -56.217 49.826
613732 98.099 -56.226 49.816
613733 98.139 -56.234 49.804
613734 98.179 -56.243 49.791
613735 98.220 -56.251 49.776
613736 98.261 -56.260 49.761
613737 98.301 -56.268 49.744
613738 98.340 -56.276 49.724
613739 98.375 -56.284 49.705
613740 98.410 -56.291 49.683
613741 98.444 -56.298 49.660
613742 98.479 -56.306 49.635
613743 98.513 -56.313 49.609
613744 98.548 -56.320 49.582
613745 98.582 -56.328 49.554
613746 98.617 -56.335 49.525
613747 98.651 -56.342 49.495
613748 98.686 -56.350 49.466
613752 98.823 -56.379 49.351
613753 98.858 -56.386 49.323
613754 98.892 -56.394 49.297
613755 98.927 -56.401 49.271
613756 98.961 -56.408 49.246
613757 98.996 -56.416 49.222
613758 99.030 -56.423 49.199
613759 99.065 -56.430 49.176
613760 99.099 -56.438 49.154
613761 99.134 -56.445 49.133
613762 99.168 -56.452 49.113
613763 99.203 -56.460 49.094
613764 99.237 -56.467 49.077
613765 99.272 -56.474 49.059
613766 99.306 -56.482 49.044
613767 99.337 -56.486 49.029
613768 99.364 -56.486 49.016
613769 99.391 -56.486 49.004
613770 99.416 -56.502 48.992
613771 99.442 -56.521 48.981
613772 99.467 -56.539 48.970
613773 99.493 -56.558 48.959
613774 99.525 -56.574 48.948
613775 99.558 -56.590 48.937
613776 99.585 -56.599 48.928
613777 99.611 -56.608 48.916
613778 99.638 -56.618 48.905
613779 99.664 -56.627 48.893
613780 99.691 -56.641 48.880
613781 99.719 -56.658 48.867
613782 99.747 -56.675 48.853
613783 99.775 -56.693 48.839
614131 88.863 -52.574 57.331
614132 88.895 -52.597 57.327
614133 88.926 -52.619 57.323
614134 88.958 -52.642 57.318
614135 88.990 -52.664 57.312
614136 89.022 -52.687 57.307
614137 89.053 -52.709 57.301
614138 89.085 -52.732 57.296
614139 89.116 -52.755 57.290
614140 89.148 -52.778 57.285
614141 89.180 -52.800 57.280
614142 89.211 -52.823 57.275
614143 89.243 -52.846 57.269
614144 89.275 -52.869 57.264
614145 89.306 -52.891 57.260
614146 89.338 -52.914 57.256
614147 89.370 -52.937 57.252
614148 89.401 -52.960 57.248
614149 89.433 -52.982 57.244
614150 89.465 -53.005 57.241
614151 89.493 -53.026 57.237
614152 89.517 -53.043 57.232
614153 89.540 -53.060 57.226
614154 89.563 -53.078 57.220
614155 89.587 -53.095 57.215
614156 89.610 -53.113 57.209
614157 89.633 -53.130 57.204
614158 89.657 -53.148 57.198
614159 89.680 -53.165 57.192
614160 89.703 -53.183 57.185
614161 89.740 -53.207 57.178
614162 89.780 -53.233 57.170
614163 89.821 -53.250 57.162
614164 89.859 -53.255 57.153
614165 89.898 -53.261 57.145
614166 89.937 -53.266 57.136
614167 89.975 -53.271 57.126
614168 90.005 -53.278 57.111
614169 90.025 -53.285 57.094
614170 90.045 -53.292 57.078
614171 90.066 -53.299 57.061
614172 90.087 -53.307 57.046
614173 90.109 -53.315 57.028
614174 90.145 -53.336 57.004
614175 90.180 -53.355 56.980
614176 90.216 -53.376 56.955
614177 90.252 -53.395 56.928
614178 90.296 -53.405 56.902
614179 90.339 -53.417 56.872
614180 90.382 -53.428 56.840
614181 90.426 -53.440 56.806
614182 90.469 -53.451 56.773
614183 90.512 -53.462 56.740
614184 90.529 -53.482 56.708
614185 90.546 -53.502 56.676
614186 90.563 -53.522 56.641
614187 90.592 -53.531 56.608
614188 90.628 -53.532 56.576
614189 90.664 -53.534 56.545
614199 90.979 -53.662 55.980
614200 91.009 -53.683 55.950
614201 91.041 -53.703 55.920
614202 91.075 -53.722 55.891
614203 91.108 -53.743 55.864
614204 91.142 -53.762 55.837
614205 91.175 -53.782 55.810
614206 91.208 -53.802 55.785
614207 91.241 -53.822 55.761
614208 91.275 -53.842 55.739
614209 91.308 -53.862 55.719
614210 91.342 -53.882 55.698
614211 91.375 -53.901 55.680
614212 91.408 -53.921 55.663
614213 91.441 -53.941 55.647
614214 91.474 -53.961 55.631
614215 91.508 -53.981 55.618
614216 91.541 -54.001 55.605
614217 91.575 -54.020 55.593
614218 91.608 -54.041 55.582
614219 91.641 -54.060 55.570
614220 91.674 -54.080 55.556
614221 91.711 -54.095 55.540
614222 91.749 -54.107 55.526
614223 91.788 -54.120 55.509
614224 91.826 -54.132 55.494
614225 91.865 -54.145 55.478
614226 91.904 -54.157 55.462
614227 91.942 -54.170 55.447
614228 91.980 -54.182 55.434
614229 92.019 -54.195 55.420
614230 92.057 -54.207 55.407
614231 92.095 -54.220 55.394
614232 92.134 -54.232 55.380
614233 92.172 -54.244 55.368
614242 92.518 -54.356 55.293
614243 92.557 -54.369 55.290
614244 92.595 -54.381 55.288
614245 92.634 -54.394 55.288
614246 92.672 -54.406 55.291
614247 92.711 -54.419 55.294
614248 92.749 -54.431 55.298
614249 92.788 -54.443 55.303
614250 92.826 -54.456 55.309
614251 92.864 -54.470 55.316
614252 92.886 -54.498 55.326
614253 92.909 -54.527 55.335
614254 92.932 -54.555 55.346
614255 92.956 -54.583 55.355
614256 92.983 -54.600 55.366
614257 93.009 -54.616 55.379
614258 93.034 -54.629 55.397
614259 93.058 -54.643 55.416
614260 93.083 -54.656 55.430
614261 93.108 -54.670 55.438
614262 93.132 -54.684 55.442
614263 93.157 -54.697 55.441
614264 93.182 -54.711 55.436
614265 93.207 -54.724 55.423
614266 93.231 -54.738 55.420
614267 93.256 -54.752 55.431
614268 93.281 -54.765 55.453
614269 93.305 -54.778 55.477
614270 93.332 -54.793 55.493
614271 93.359 -54.808 55.499
614272 93.384 -54.822 55.502
614273 93.409 -54.836 55.503
614274 93.434 -54.850 55.507
614275 93.459 -54.863 55.514
614276 93.484 -54.877 55.523
614277 93.509 -54.891 55.528
614278 93.535 -54.905 55.525
614279 93.560 -54.919 55.518
614280 93.584 -54.932 55.504
614281 93.610 -54.946 55.487
614282 93.635 -54.960 55.470
614283 93.660 -54.974 55.454
614284 93.685 -54.988 55.433
614285 93.710 -55.001 55.409
614286 93.735 -55.016 55.390
614287 93.760 -55.029 55.371
614288 93.785 -55.043 55.350
614289 93.806 -55.066 55.326
614290 93.826 -55.091 55.301
614291 93.846 -55.115 55.279
614292 93.866 -55.139 55.258
614293 93.887 -55.163 55.238
614294 93.907 -55.187 55.219
614295 93.927 -55.211 55.204
614296 93.947 -55.235 55.187
614297 93.968 -55.260 55.168
614298 93.990 -55.279 55.144
614299 94.014 -55.296 55.116
614300 94.039 -55.313 55.085
614301 94.063 -55.329 55.053
614302 94.087 -55.346 55.020
614303 94.112 -55.363 54.986
614304 94.137 -55.379 54.953
614305 94.171 -55.391 54.921
614306 94.203 -55.402 54.890
614307 94.236 -55.414 54.861
614308 94.269 -55.426 54.832
614309 94.302 -55.438 54.798
614310 94.335 -55.450 54.759
614311 94.368 -55.462 54.713
614312 94.399 -55.471 54.667
614313 94.429 -55.480 54.626
614314 94.460 -55.489 54.590
614315 94.500 -55.500 54.559
614316 94.550 -55.512 54.533
614317 94.599 -55.525 54.505
614318 94.641 -55.538 54.476
614319 94.681 -55.551 54.446
614320 94.720 -55.564 54.414
614321 94.760 -55.576 54.380
614322 94.800 -55.589 54.342
614323 94.840 -55.602 54.300
614324 94.880 -55.614 54.255
614325 94.911 -55.625 54.207
614326 94.939 -55.635 54.158
614327 94.968 -55.645 54.108
614328 94.996 -55.654 54.056
614329 95.024 -55.664 53.998
614330 95.052 -55.674 53.935
614331 95.081 -55.683 53.871
614332 95.110 -55.697 53.806
614333 95.139 -55.713 53.741
614334 95.168 -55.729 53.676
614335 95.197 -55.745 53.611
614336 95.227 -55.761 53.547
614337 95.256 -55.777 53.482
614338 95.289 -55.790 53.415
614339 95.322 -55.803 53.348
614340 95.354 -55.817 53.280
614341 95.387 -55.830 53.209
614342 95.420 -55.843 53.139
614343 95.453 -55.856 53.071
614344 95.485 -55.870 53.003
614345 95.519 -55.883 52.933
614346 95.552 -55.895 52.860
614347 95.583 -55.909 52.785
614348 95.614 -55.922 52.713
614349 95.645 -55.936 52.643
614350 95.677 -55.950 52.574
614351 95.708 -55.963 52.507
614352 95.739 -55.977 52.440
614353 95.770 -55.990 52.377
614354 95.804 -56.004 52.314
614355 95.838 -56.018 52.252
614356 95.872 -56.031 52.190
614357 95.906 -56.045 52.130
614358 95.940 -56.059 52.072
614359 95.974 -56.073 52.015
614360 96.008 -56.087 51.957
614361 96.042 -56.101 51.897
614362 96.076 -56.115 51.834
614363 96.110 -56.128 51.769
614364 96.142 -56.129 51.701
614365 96.174 -56.132 51.638
614366 96.205 -56.133 51.575
614367 96.237 -56.135 51.513
614368 96.269 -56.138 51.452
614369 96.301 -56.139 51.394
614370 96.333 -56.141 51.336
614371 96.365 -56.144 51.283
614372 96.391 -56.145 51.232
614373 96.407 -56.146 51.183
614374 96.422 -56.147 51.135
614375 96.438 -56.148 51.087
614376 96.453 -56.150 51.042
614377 96.469 -56.150 51.001
614378 96.484 -56.152 50.960
614379 96.497 -56.153 50.920
614380 96.509 -56.154 50.881
614381 96.521 -56.156 50.842
614382 96.533 -56.157 50.805
614383 96.546 -56.158 50.770
614384 96.558 -56.160 50.737
614385 96.571 -56.160 50.703
614386 96.588 -56.160 50.668
614387 96.604 -56.159 50.634
614388 96.621 -56.158 50.602
614389 96.638 -56.157 50.571
614390 96.654 -56.156 50.542
614391 96.671 -56.155 50.512
614392 96.687 -56.154 50.483
614393 96.704 -56.154 50.454
614394 96.729 -56.153 50.426
614395 96.764 -56.151 50.401
614396 96.799 -56.150 50.376
614397 96.834 -56.150 50.352
614398 96.869 -56.148 50.327
614399 96.904 -56.147 50.301
614400 96.939 -56.147 50.278
614401 96.974 -56.145 50.257
614402 97.009 -56.144 50.234
614403 97.044 -56.144 50.211
614404 97.079 -56.142 50.189
614405 97.114 -56.141 50.166
614406 97.149 -56.140 50.144
614407 97.184 -56.139 50.124
614408 97.219 -56.138 50.105
614409 97.254 -56.137 50.085
614410 97.289 -56.136 50.067
614411 97.324 -56.135 50.047
614412 97.359 -56.134 50.029
614413 97.394 -56.133 50.013
614414 97.429 -56.132 49.997
614415 97.464 -56.131 49.981
614416 97.499 -56.130 49.964
614417 97.533 -56.128 49.948
614418 97.569 -56.128 49.934
614419 97.603 -56.127 49.920
614420 97.639 -56.125 49.908
614421 97.666 -56.125 49.897
614422 97.689 -56.124 49.885
614423 97.712 -56.120 49.873
614424 97.748 -56.125 49.863
614425 97.789 -56.135 49.854
614426 97.829 -56.143 49.844
614427 97.870 -56.151 49.836
614428 97.910 -56.160 49.827
614429 97.950 -56.169 49.817
614430 97.991 -56.177 49.808
614431 98.031 -56.185 49.799
614432 98.072 -56.194 49.788
614433 98.112 -56.203 49.777
614434 98.153 -56.211 49.764
614435 98.193 -56.219 49.751
614436 98.233 -56.229 49.737
614437 98.274 -56.237 49.721
614438 98.314 -56.245 49.704
614439 98.351 -56.253 49.685
614440 98.385 -56.260 49.666
614441 98.420 -56.268 49.645
614442 98.454 -56.275 49.622
614443 98.489 -56.282 49.598
614444 98.523 -56.290 49.572
614454 98.868 -56.363 49.289
614455 98.903 -56.370 49.262
614456 98.937 -56.378 49.236
614457 98.972 -56.385 49.211
614458 99.006 -56.392 49.187
614459 99.041 -56.400 49.163
614460 99.075 -56.407 49.140
614461 99.110 -56.414 49.118
614462 99.144 -56.422 49.097
614463 99.179 -56.429 49.076
614464 99.213 -56.436 49.057
614465 99.248 -56.444 49.039
614466 99.282 -56.451 49.022
614467 99.313 -56.455 49.006
614468 99.340 -56.455 48.991
614469 99.367 -56.456 48.977
614470 99.393 -56.467 48.964
614471 99.418 -56.485 48.952
614472 99.443 -56.503 48.941
614473 99.469 -56.522 48.930
614474 99.497 -56.540 48.918
614475 99.529 -56.556 48.907
614476 99.560 -56.570 48.897
614477 99.586 -56.579 48.887
614478 99.613 -56.589 48.878
614479 99.639 -56.598 48.867
614480 99.666 -56.607 48.856
614481 99.693 -56.620 48.845
614482 99.721 -56.637 48.833
614483 99.749 -56.654 48.821
614831 88.847 -52.527 57.332
614832 88.879 -52.550 57.324
614833 88.911 -52.572 57.320
614834 88.943 -52.595 57.315
614835 88.974 -52.618 57.309
614836 89.006 -52.640 57.303
614837 89.038 -52.663 57.297
614838 89.069 -52.686 57.291
614839 89.101 -52.709 57.286
614840 89.132 -52.731 57.281
614841 89.164 -52.754 57.275
614842 89.196 -52.777 57.269
614843 89.228 -52.799 57.264
614844 89.259 -52.822 57.259
614845 89.291 -52.844 57.254
614846 89.322 -52.867 57.249
614847 89.354 -52.890 57.245
614848 89.386 -52.913 57.241
614849 89.417 -52.935 57.237
614850 89.451 -52.955 57.233
614851 89.483 -52.974 57.229
614852 89.517 -52.989 57.225
614853 89.537 -53.007 57.221
614854 89.553 -53.029 57.216
614855 89.573 -53.049 57.210
614856 89.596 -53.067 57.206
614857 89.619 -53.084 57.201
614858 89.642 -53.101 57.195
614859 89.666 -53.119 57.190
614860 89.689 -53.136 57.185
614861 89.712 -53.154 57.179
614862 89.750 -53.179 57.173
614863 89.791 -53.200 57.165
614864 89.829 -53.205 57.157
614865 89.868 -53.211 57.149
614866 89.906 -53.216 57.141
614867 89.945 -53.221 57.133
614868 89.983 -53.227 57.126
614869 90.019 -53.232 57.112
614870 90.039 -53.239 57.096
614871 90.059 -53.246 57.080
614872 90.080 -53.254 57.065
614873 90.101 -53.261 57.050
614874 90.129 -53.275 57.033
614875 90.165 -53.295 57.011
614876 90.201 -53.315 56.990
614877 90.236 -53.336 56.966
614878 90.272 -53.355 56.942
614879 90.311 -53.371 56.917
614880 90.354 -53.383 56.890
614881 90.397 -53.394 56.860
614882 90.441 -53.405 56.828
614883 90.484 -53.417 56.796
614884 90.519 -53.430 56.764
614885 90.536 -53.451 56.731
614886 90.553 -53.471 56.700
614887 90.580 -53.482 56.666
614888 90.615 -53.483 56.632
614889 90.651 -53.485 56.599
614890 90.687 -53.487 56.567
614891 90.723 -53.489 56.537
614905 91.166 -53.725 55.849
614906 91.200 -53.745 55.822
614907 91.233 -53.766 55.797
614908 91.266 -53.785 55.773
614909 91.299 -53.805 55.750
614910 91.332 -53.825 55.730
614911 91.366 -53.844 55.709
614912 91.399 -53.865 55.690
614913 91.433 -53.885 55.672
614914 91.466 -53.904 55.655
614915 91.499 -53.924 55.639
614916 91.532 -53.944 55.624
614917 91.566 -53.964 55.610
614918 91.599 -53.984 55.597
614919 91.630 -54.004 55.584
614920 91.657 -54.025 55.571
614921 91.689 -54.048 55.557
614922 91.727 -54.066 55.540
614923 91.768 -54.080 55.526
614924 91.808 -54.093 55.509
614925 91.847 -54.106 55.494
614926 91.885 -54.118 55.478
614927 91.923 -54.131 55.463
614928 91.962 -54.143 55.447
614929 92.000 -54.156 55.435
614930 92.038 -54.168 55.421
614931 92.077 -54.181 55.408
614932 92.116 -54.193 55.394
614933 92.154 -54.205 55.381
614943 92.538 -54.330 55.294
614944 92.577 -54.342 55.292
614945 92.616 -54.355 55.290
614946 92.654 -54.367 55.290
614947 92.692 -54.380 55.293
614948 92.731 -54.392 55.297
614949 92.769 -54.405 55.301
614950 92.807 -54.417 55.306
614951 92.846 -54.430 55.311
614952 92.884 -54.442 55.317
614953 92.909 -54.469 55.326
614954 92.932 -54.498 55.335
614955 92.955 -54.527 55.344
614956 92.978 -54.555 55.354
614957 93.003 -54.578 55.364
614958 93.029 -54.594 55.377
614959 93.054 -54.608 55.395
614960 93.078 -54.621 55.412
614961 93.103 -54.635 55.421
614962 93.128 -54.649 55.426
614963 93.153 -54.662 55.427
614964 93.178 -54.675 55.425
614965 93.202 -54.689 55.419
614966 93.227 -54.703 55.409
614967 93.252 -54.716 55.410
614968 93.274 -54.727 55.423
614969 93.296 -54.736 55.444
614970 93.318 -54.748 55.465
614971 93.344 -54.765 55.478
614972 93.366 -54.778 55.481
614973 93.391 -54.792 55.481
614974 93.415 -54.806 55.484
614975 93.440 -54.819 55.492
614976 93.466 -54.833 55.504
614977 93.491 -54.847 55.516
614978 93.516 -54.861 55.523
614979 93.541 -54.875 55.522
614980 93.566 -54.888 55.513
614981 93.591 -54.903 55.496
614982 93.616 -54.916 55.476
614983 93.641 -54.930 55.456
614984 93.666 -54.944 55.444
614985 93.692 -54.958 55.423
614986 93.716 -54.972 55.401
614987 93.742 -54.985 55.382
614988 93.767 -54.999 55.364
614989 93.792 -55.013 55.344
614990 93.814 -55.034 55.320
614991 93.834 -55.057 55.296
614992 93.854 -55.082 55.271
614993 93.874 -55.106 55.248
614994 93.895 -55.130 55.225
614995 93.915 -55.154 55.203
614996 93.935 -55.179 55.185
614997 93.956 -55.202 55.167
614998 93.976 -55.226 55.149
614999 94.000 -55.243 55.130
615000 94.024 -55.260 55.108
615001 94.049 -55.276 55.083
615002 94.073 -55.293 55.057
615003 94.097 -55.310 55.030
615004 94.122 -55.326 55.000
615005 94.146 -55.343 54.970
615006 94.179 -55.354 54.940
615007 94.212 -55.367 54.912
615008 94.245 -55.379 54.886
615009 94.278 -55.390 54.856
615010 94.311 -55.402 54.821
615011 94.344 -55.414 54.779
615012 94.377 -55.426 54.732
615013 94.409 -55.437 54.685
615014 94.439 -55.446 54.645
615015 94.470 -55.455 54.611
615016 94.513 -55.467 54.580
615017 94.563 -55.479 54.552
615018 94.607 -55.492 54.521
615019 94.646 -55.505 54.489
615020 94.686 -55.517 54.457
615021 94.726 -55.530 54.423
615022 94.766 -55.543 54.389
615023 94.806 -55.556 54.351
615024 94.845 -55.568 54.309
615025 94.885 -55.581 54.262
615026 94.918 -55.593 54.212
615027 94.946 -55.602 54.160
615028 94.974 -55.611 54.108
615029 95.003 -55.621 54.053
615030 95.031 -55.631 53.992
615031 95.059 -55.641 53.930
615032 95.088 -55.653 53.867
615033 95.117 -55.670 53.805
615034 95.146 -55.686 53.740
615035 95.176 -55.702 53.679
615036 95.205 -55.718 53.618
615037 95.235 -55.734 53.558
615038 95.264 -55.749 53.496
615039 95.296 -55.763 53.431
615040 95.329 -55.776 53.362
615041 95.361 -55.790 53.291
615042 95.394 -55.803 53.220
615043 95.427 -55.816 53.149
615044 95.460 -55.829 53.080
615045 95.494 -55.841 53.013
615046 95.528 -55.853 52.943
615047 95.561 -55.865 52.873
615048 95.594 -55.879 52.801
615049 95.625 -55.892 52.730
615050 95.656 -55.906 52.660
615051 95.687 -55.919 52.591
615052 95.719 -55.933 52.523
615053 95.750 -55.947 52.454
615054 95.782 -55.960 52.387
615055 95.816 -55.974 52.323
615056 95.850 -55.988 52.259
615057 95.884 -56.002 52.195
615058 95.918 -56.016 52.133
615059 95.953 -56.029 52.074
615060 95.986 -56.044 52.016
615061 96.021 -56.057 51.957
615062 96.054 -56.071 51.898
615063 96.089 -56.085 51.835
615064 96.123 -56.099 51.771
615065 96.156 -56.107 51.706
615066 96.188 -56.110 51.643
615067 96.220 -56.112 51.580
615068 96.251 -56.113 51.517
615069 96.283 -56.116 51.456
615070 96.315 -56.118 51.396
615071 96.347 -56.119 51.336
615072 96.379 -56.122 51.282
615073 96.402 -56.123 51.230
615074 96.417 -56.124 51.179
615075 96.433 -56.125 51.129
615076 96.448 -56.126 51.080
615077 96.464 -56.128 51.032
615078 96.479 -56.128 50.990
615079 96.492 -56.130 50.948
615080 96.504 -56.131 50.906
615081 96.517 -56.132 50.865
615082 96.529 -56.134 50.825
615083 96.541 -56.135 50.786
615084 96.554 -56.136 50.750
615085 96.569 -56.136 50.716
615086 96.586 -56.135 50.681
615087 96.602 -56.135 50.646
615088 96.619 -56.134 50.611
615089 96.636 -56.132 50.578
615090 96.652 -56.132 50.547
615091 96.669 -56.131 50.517
615092 96.685 -56.130 50.487
615093 96.702 -56.129 50.457
615094 96.719 -56.128 50.428
615095 96.745 -56.128 50.399
615096 96.780 -56.126 50.374
615097 96.815 -56.125 50.350
615098 96.850 -56.125 50.325
615099 96.885 -56.123 50.299
615100 96.920 -56.122 50.274
615101 96.955 -56.122 50.250
615102 96.990 -56.120 50.228
615103 97.025 -56.119 50.206
615104 97.060 -56.118 50.183
615105 97.095 -56.117 50.160
615106 97.130 -56.116 50.137
615107 97.165 -56.115 50.115
615108 97.200 -56.114 50.095
615109 97.235 -56.113 50.074
615110 97.270 -56.112 50.055
615111 97.305 -56.111 50.035
615120 97.619 -56.102 49.885
615121 97.652 -56.100 49.872
615122 97.675 -56.100 49.860
615123 97.698 -56.099 49.848
615124 97.721 -56.094 49.837
615125 97.761 -56.103 49.826
615126 97.802 -56.112 49.817
615127 97.843 -56.120 49.808
615128 97.883 -56.128 49.799
615129 97.923 -56.137 49.790
615130 97.964 -56.146 49.781
615131 98.004 -56.154 49.772
615132 98.044 -56.163 49.761
615133 98.085 -56.171 49.749
615134 98.126 -56.180 49.738
615135 98.166 -56.188 49.725
615136 98.206 -56.197 49.712
615137 98.246 -56.206 49.698
615138 98.287 -56.214 49.683
615139 98.326 -56.222 49.666
615140 98.361 -56.229 49.648
615157 98.947 -56.354 49.203
615158 98.982 -56.361 49.177
615159 99.016 -56.369 49.152
615160 99.051 -56.376 49.129
615161 99.085 -56.383 49.105
615162 99.120 -56.391 49.083
615163 99.154 -56.398 49.062
615164 99.189 -56.405 49.041
615165 99.223 -56.413 49.021
615166 99.258 -56.420 49.002
615167 99.289 -56.424 48.985
615168 99.315 -56.425 48.969
615169 99.343 -56.425 48.954
615170 99.369 -56.430 48.939
615171 99.395 -56.448 48.926
615172 99.420 -56.467 48.913
615173 99.445 -56.486 48.902
615174 99.471 -56.505 48.890
615175 99.500 -56.522 48.878
615176 99.533 -56.538 48.867
615177 99.562 -56.549 48.856
615178 99.588 -56.558 48.846
615179 99.615 -56.568 48.838
615180 99.641 -56.577 48.828
615181 99.668 -56.586 48.818
615182 99.695 -56.599 48.809
615531 88.832 -52.481 57.331
615532 88.863 -52.503 57.324
615533 88.895 -52.526 57.316
615534 88.927 -52.548 57.311
615535 88.959 -52.571 57.306
615536 88.990 -52.593 57.300
615537 89.022 -52.616 57.294
615538 89.054 -52.639 57.287
615539 89.085 -52.662 57.281
615540 89.117 -52.684 57.276
615541 89.148 -52.707 57.271
615542 89.180 -52.730 57.265
615543 89.212 -52.753 57.260
615544 89.244 -52.775 57.254
615545 89.275 -52.798 57.248
615546 89.307 -52.821 57.244
615547 89.339 -52.844 57.239
615548 89.371 -52.863 57.234
615549 89.404 -52.883 57.230
615550 89.438 -52.902 57.226
615551 89.470 -52.922 57.222
615552 89.504 -52.941 57.218
615553 89.536 -52.956 57.214
615554 89.568 -52.964 57.210
615555 89.585 -52.986 57.206
615556 89.601 -53.008 57.201
615557 89.617 -53.030 57.197
615558 89.632 -53.053 57.192
615559 89.652 -53.073 57.187
615560 89.675 -53.090 57.182
615561 89.698 -53.107 57.178
615562 89.722 -53.125 57.173
615563 89.761 -53.149 57.167
615564 89.799 -53.154 57.160
615565 89.838 -53.160 57.152
615566 89.877 -53.165 57.145
615567 89.915 -53.170 57.137
615568 89.954 -53.176 57.131
615569 89.992 -53.181 57.123
615570 90.031 -53.186 57.110
615571 90.053 -53.193 57.095
615572 90.074 -53.201 57.081
615573 90.094 -53.208 57.068
615574 90.115 -53.215 57.054
615575 90.150 -53.235 57.037
615576 90.186 -53.255 57.019
615577 90.221 -53.275 56.999
615578 90.257 -53.295 56.978
615579 90.293 -53.315 56.956
615580 90.328 -53.336 56.933
615581 90.370 -53.349 56.909
615582 90.413 -53.360 56.881
615583 90.456 -53.371 56.851
615584 90.499 -53.383 56.821
615585 90.527 -53.399 56.790
615586 90.544 -53.419 56.758
615587 90.567 -53.433 56.727
615588 90.603 -53.435 56.693
615589 90.639 -53.437 56.658
615590 90.674 -53.439 56.624
615591 90.710 -53.440 56.592
615592 90.745 -53.442 56.560
615593 90.781 -53.444 56.531
615606 91.191 -53.688 55.865
615607 91.224 -53.709 55.837
615608 91.257 -53.728 55.811
615609 91.291 -53.748 55.787
615610 91.324 -53.768 55.764
615611 91.357 -53.787 55.743
615612 91.391 -53.808 55.721
615613 91.424 -53.828 55.701
615614 91.457 -53.847 55.682
615615 91.491 -53.867 55.664
615616 91.524 -53.887 55.647
615617 91.554 -53.907 55.631
615618 91.581 -53.928 55.617
615619 91.607 -53.949 55.602
615620 91.637 -53.971 55.588
615621 91.670 -53.996 55.574
615622 91.704 -54.020 55.559
615623 91.741 -54.038 55.543
615624 91.782 -54.051 55.527
615625 91.823 -54.065 55.511
615626 91.864 -54.079 55.495
615627 91.905 -54.091 55.480
615628 91.943 -54.104 55.464
615629 91.982 -54.117 55.449
615630 92.020 -54.129 55.435
615631 92.059 -54.142 55.423
615632 92.097 -54.154 55.409
615633 92.136 -54.167 55.396
615634 92.174 -54.179 55.382
615644 92.558 -54.304 55.296
615645 92.597 -54.317 55.293
615646 92.635 -54.329 55.292
615647 92.674 -54.341 55.292
615648 92.712 -54.354 55.295
615649 92.751 -54.366 55.299
615650 92.789 -54.379 55.304
615651 92.828 -54.391 55.309
615652 92.866 -54.404 55.314
615653 92.905 -54.416 55.321
615654 92.932 -54.439 55.329
615655 92.955 -54.468 55.338
615656 92.978 -54.497 55.348
615657 93.001 -54.526 55.357
615658 93.023 -54.555 55.366
615659 93.049 -54.573 55.376
615660 93.074 -54.586 55.390
615661 93.098 -54.599 55.402
615662 93.124 -54.613 55.409
615663 93.148 -54.627 55.410
615664 93.170 -54.637 55.409
615665 93.192 -54.646 55.405
615666 93.214 -54.656 55.400
615667 93.236 -54.666 55.395
615668 93.258 -54.676 55.399
615669 93.280 -54.686 55.413
615670 93.301 -54.703 55.431
615671 93.322 -54.727 55.447
615672 93.355 -54.736 55.458
615673 93.377 -54.749 55.461
615674 93.400 -54.762 55.463
615675 93.422 -54.775 55.471
615676 93.447 -54.790 55.486
615677 93.472 -54.803 55.502
615678 93.497 -54.817 55.515
615679 93.522 -54.831 55.519
615680 93.547 -54.845 55.515
615681 93.572 -54.859 55.503
615682 93.597 -54.872 55.486
615683 93.623 -54.886 55.464
615684 93.648 -54.900 55.440
615685 93.673 -54.914 55.427
615686 93.698 -54.928 55.408
615687 93.723 -54.941 55.385
615688 93.748 -54.955 55.367
615689 93.773 -54.969 55.351
615690 93.798 -54.983 55.334
615691 93.822 -55.000 55.312
615692 93.842 -55.025 55.288
615693 93.862 -55.049 55.263
615694 93.883 -55.073 55.239
615695 93.903 -55.097 55.215
615696 93.923 -55.121 55.191
615697 93.943 -55.145 55.168
615698 93.963 -55.169 55.149
615699 93.985 -55.190 55.133
615700 94.009 -55.207 55.119
615701 94.034 -55.223 55.104
615702 94.058 -55.240 55.086
615703 94.082 -55.257 55.066
615704 94.107 -55.273 55.043
615705 94.131 -55.290 55.017
615706 94.155 -55.307 54.989
615707 94.188 -55.319 54.961
615708 94.221 -55.331 54.934
615709 94.254 -55.342 54.909
615710 94.287 -55.354 54.877
615711 94.320 -55.367 54.841
615712 94.353 -55.379 54.797
615713 94.386 -55.390 54.749
615714 94.419 -55.402 54.703
615715 94.450 -55.412 54.665
615716 94.481 -55.421 54.632
615717 94.527 -55.433 54.600
615718 94.572 -55.446 54.569
615719 94.612 -55.458 54.535
615720 94.652 -55.471 54.500
615721 94.692 -55.484 54.465
615722 94.732 -55.497 54.429
615723 94.771 -55.510 54.393
615724 94.811 -55.523 54.356
615725 94.851 -55.536 54.312
615726 94.891 -55.549 54.264
615727 94.925 -55.559 54.212
615728 94.953 -55.569 54.158
615729 94.981 -55.579 54.104
615730 95.010 -55.589 54.047
615731 95.038 -55.599 53.986
615732 95.066 -55.610 53.927
615733 95.096 -55.626 53.868
615734 95.125 -55.642 53.806
615735 95.155 -55.658 53.742
615736 95.184 -55.674 53.682
615737 95.213 -55.690 53.625
615738 95.242 -55.705 53.569
615739 95.271 -55.722 53.510
615740 95.303 -55.736 53.445
615741 95.336 -55.749 53.374
615742 95.369 -55.762 53.302
615743 95.402 -55.775 53.229
615744 95.436 -55.787 53.157
615745 95.469 -55.799 53.087
615746 95.503 -55.812 53.019
615747 95.537 -55.824 52.951
615748 95.570 -55.837 52.884
615749 95.604 -55.849 52.815
615750 95.635 -55.862 52.745
615751 95.667 -55.876 52.674
615752 95.698 -55.890 52.605
615753 95.729 -55.903 52.536
615754 95.761 -55.917 52.466
615755 95.795 -55.931 52.398
615756 95.829 -55.944 52.331
615757 95.863 -55.958 52.266
615758 95.897 -55.972 52.202
615759 95.931 -55.986 52.139
615760 95.965 -56.000 52.078
615761 95.999 -56.014 52.019
615762 96.033 -56.028 51.961
615763 96.067 -56.041 51.901
615764 96.101 -56.055 51.839
615765 96.135 -56.069 51.775
615766 96.169 -56.083 51.711
615767 96.201 -56.088 51.649
615768 96.234 -56.090 51.585
615769 96.265 -56.092 51.522
615770 96.297 -56.094 51.461
615771 96.329 -56.096 51.399
615772 96.361 -56.097 51.338
615773 96.393 -56.100 51.283
615774 96.412 -56.101 51.230
615775 96.428 -56.102 51.177
615776 96.443 -56.103 51.125
615777 96.459 -56.104 51.074
615778 96.474 -56.106 51.024
615779 96.487 -56.107 50.980
615780 96.500 -56.108 50.937
615781 96.512 -56.110 50.893
615782 96.524 -56.110 50.850
615783 96.537 -56.112 50.808
615784 96.550 -56.113 50.767
615785 96.567 -56.112 50.730
615786 96.583 -56.111 50.694
615787 96.600 -56.110 50.659
615788 96.616 -56.109 50.623
615789 96.633 -56.108 50.589
615790 96.650 -56.107 50.554
615791 96.666 -56.107 50.523
615792 96.683 -56.106 50.492
615793 96.699 -56.105 50.462
615794 96.716 -56.104 50.432
615795 96.733 -56.103 50.402
615796 96.761 -56.103 50.373
615797 96.796 -56.101 50.348
615798 96.831 -56.100 50.323
615799 96.866 -56.099 50.298
615800 96.901 -56.098 50.272
615801 96.936 -56.097 50.247
615802 96.971 -56.096 50.223
615803 97.006 -56.095 50.200
615804 97.041 -56.094 50.177
615805 97.076 -56.093 50.155
615806 97.111 -56.092 50.132
615807 97.146 -56.091 50.109
615808 97.181 -56.090 50.086
615809 97.216 -56.089 50.066
615810 97.251 -56.088 50.045
615811 97.286 -56.087 50.025
615824 97.706 -56.073 49.812
615825 97.734 -56.072 49.801
615826 97.775 -56.080 49.790
615827 97.815 -56.088 49.780
615828 97.856 -56.097 49.772
615829 97.896 -56.106 49.763
615830 97.937 -56.114 49.754
615831 97.977 -56.123 49.744
615832 98.017 -56.132 49.734
615833 98.058 -56.140 49.724
616231 88.817 -52.434 57.330
616232 88.848 -52.456 57.323
616233 88.880 -52.479 57.316
616234 88.911 -52.502 57.308
616235 88.943 -52.525 57.303
616236 88.975 -52.547 57.297
616237 89.006 -52.570 57.291
616238 89.038 -52.593 57.285
616239 89.070 -52.615 57.279
616240 89.102 -52.638 57.272
616241 89.133 -52.660 57.266
616242 89.165 -52.683 57.261
616243 89.196 -52.706 57.255
616244 89.228 -52.728 57.250
616245 89.260 -52.751 57.245
616246 89.292 -52.772 57.238
616247 89.325 -52.792 57.233
616248 89.358 -52.811 57.229
616249 89.391 -52.831 57.224
616250 89.424 -52.850 57.219
616251 89.458 -52.869 57.215
616252 89.490 -52.888 57.211
616253 89.524 -52.908 57.208
616254 89.556 -52.923 57.204
616255 89.589 -52.930 57.200
616256 89.616 -52.943 57.196
616257 89.632 -52.965 57.192
616258 89.648 -52.987 57.188
616259 89.664 -53.010 57.184
616260 89.680 -53.032 57.179
616261 89.696 -53.054 57.175
616262 89.712 -53.076 57.170
616263 89.731 -53.096 57.166
616264 89.769 -53.104 57.161
616265 89.808 -53.109 57.155
616266 89.846 -53.114 57.148
616267 89.885 -53.120 57.141
616268 89.923 -53.125 57.133
616269 89.962 -53.130 57.127
616270 90.001 -53.136 57.120
616271 90.039 -53.141 57.109
616272 90.067 -53.148 57.096
616273 90.088 -53.154 57.084
616274 90.109 -53.162 57.071
616275 90.135 -53.174 57.058
616276 90.171 -53.195 57.043
616277 90.206 -53.214 57.027
616278 90.242 -53.235 57.009
616279 90.277 -53.255 56.990
616280 90.313 -53.275 56.970
616281 90.349 -53.295 56.950
616282 90.385 -53.314 56.929
616283 90.428 -53.326 56.903
616284 90.471 -53.337 56.876
616285 90.515 -53.349 56.847
616286 90.534 -53.368 56.818
616287 90.554 -53.385 56.787
616288 90.590 -53.386 56.756
616289 90.626 -53.388 56.722
616290 90.662 -53.390 56.687
616291 90.697 -53.392 56.652
616292 90.733 -53.393 56.619
616293 90.768 -53.395 56.586
616294 90.804 -53.397 56.557
616295 90.835 -53.412 56.524
616307 91.216 -53.652 55.881
616308 91.249 -53.672 55.854
616309 91.282 -53.691 55.828
616310 91.315 -53.711 55.803
616311 91.349 -53.731 55.779
616312 91.382 -53.751 55.758
616313 91.416 -53.771 55.736
616314 91.449 -53.791 55.714
616315 91.477 -53.811 55.694
616316 91.504 -53.832 55.676
616317 91.531 -53.853 55.659
616318 91.558 -53.873 55.642
616319 91.585 -53.894 55.624
616320 91.617 -53.918 55.610
616321 91.651 -53.943 55.595
616322 91.684 -53.967 55.578
616323 91.717 -53.991 55.562
616324 91.755 -54.009 55.546
616325 91.796 -54.022 55.530
616326 91.837 -54.036 55.514
616327 91.878 -54.050 55.498
616328 91.919 -54.063 55.482
616329 91.959 -54.076 55.466
616330 92.000 -54.090 55.451
616331 92.040 -54.103 55.437
616332 92.079 -54.116 55.424
616333 92.117 -54.128 55.410
616334 92.155 -54.141 55.396
616345 92.578 -54.277 55.297
616346 92.617 -54.290 55.295
616347 92.655 -54.302 55.294
616348 92.694 -54.315 55.294
616349 92.732 -54.327 55.298
616350 92.771 -54.340 55.302
616351 92.809 -54.352 55.307
616352 92.848 -54.364 55.312
616353 92.886 -54.377 55.318
616354 92.924 -54.389 55.326
616355 92.955 -54.410 55.334
616356 92.978 -54.439 55.344
616357 93.001 -54.467 55.353
616358 93.023 -54.496 55.361
616359 93.047 -54.525 55.369
616360 93.067 -54.548 55.375
616361 93.089 -54.557 55.384
616362 93.111 -54.567 55.391
616363 93.133 -54.577 55.393
616364 93.155 -54.587 55.390
616365 93.177 -54.596 55.386
616366 93.199 -54.605 55.381
616367 93.220 -54.615 55.376
616368 93.242 -54.625 55.375
616369 93.264 -54.635 55.384
616370 93.285 -54.658 55.397
616371 93.306 -54.681 55.412
616372 93.335 -54.697 55.425
616373 93.367 -54.706 55.434
616374 93.389 -54.720 55.440
616375 93.411 -54.733 55.448
616376 93.433 -54.746 55.464
616377 93.455 -54.759 55.485
616378 93.479 -54.773 55.504
616379 93.503 -54.787 55.513
616380 93.529 -54.801 55.514
616381 93.554 -54.815 55.505
616382 93.579 -54.828 55.492
616383 93.604 -54.842 55.475
616384 93.629 -54.856 55.451
616385 93.654 -54.870 55.424
616386 93.679 -54.884 55.408
616387 93.704 -54.897 55.391
616388 93.729 -54.912 55.370
616389 93.755 -54.925 55.352
616390 93.780 -54.939 55.339
616391 93.804 -54.953 55.323
616392 93.830 -54.967 55.302
616393 93.850 -54.991 55.279
616394 93.870 -55.016 55.256
616395 93.890 -55.040 55.233
616396 93.910 -55.064 55.210
616397 93.930 -55.088 55.185
616398 93.951 -55.112 55.160
616399 93.971 -55.136 55.138
616400 93.995 -55.154 55.124
616401 94.019 -55.170 55.117
616402 94.044 -55.187 55.110
616403 94.068 -55.204 55.097
616404 94.092 -55.220 55.080
616405 94.117 -55.237 55.059
616406 94.141 -55.254 55.035
616407 94.165 -55.270 55.010
616408 94.197 -55.283 54.984
616409 94.229 -55.295 54.956
616410 94.263 -55.307 54.928
616411 94.295 -55.319 54.896
616412 94.329 -55.331 54.858
616413 94.361 -55.342 54.813
616414 94.395 -55.354 54.765
616415 94.427 -55.367 54.720
616416 94.461 -55.378 54.683
616417 94.491 -55.388 54.650
616418 94.538 -55.400 54.617
616419 94.577 -55.413 54.584
616420 94.617 -55.426 54.547
616421 94.657 -55.439 54.509
616422 94.697 -55.451 54.472
616423 94.737 -55.464 54.434
616424 94.776 -55.477 54.396
616425 94.816 -55.489 54.357
616426 94.856 -55.502 54.311
616427 94.896 -55.515 54.260
616428 94.932 -55.527 54.208
616429 94.960 -55.536 54.152
616430 94.988 -55.546 54.098
616431 95.017 -55.556 54.043
616432 95.045 -55.567 53.986
616433 95.074 -55.583 53.927
616434 95.104 -55.599 53.868
616435 95.133 -55.614 53.806
616436 95.162 -55.630 53.743
616437 95.191 -55.646 53.686
616438 95.220 -55.662 53.632
616439 95.250 -55.678 53.578
616440 95.279 -55.694 53.520
616441 95.310 -55.709 53.456
616442 95.343 -55.721 53.385
616443 95.377 -55.734 53.311
616444 95.411 -55.746 53.237
616445 95.445 -55.758 53.163
616446 95.478 -55.771 53.091
616447 95.512 -55.783 53.024
616448 95.546 -55.795 52.957
616449 95.580 -55.807 52.893
616450 95.613 -55.819 52.827
616451 95.646 -55.832 52.756
616452 95.677 -55.846 52.685
616453 95.708 -55.859 52.616
616454 95.740 -55.873 52.547
616455 95.773 -55.887 52.477
616456 95.807 -55.901 52.407
616457 95.841 -55.915 52.339
616458 95.876 -55.928 52.273
616459 95.909 -55.942 52.207
616460 95.944 -55.956 52.144
616461 95.977 -55.970 52.081
616462 96.012 -55.984 52.022
616463 96.046 -55.998 51.963
616464 96.080 -56.012 51.903
616465 96.114 -56.025 51.841
616466 96.148 -56.039 51.779
616467 96.182 -56.053 51.716
616468 96.216 -56.066 51.655
616469 96.248 -56.068 51.592
616470 96.279 -56.070 51.529
616471 96.311 -56.072 51.466
616472 96.343 -56.074 51.403
616473 96.375 -56.076 51.343
616474 96.407 -56.078 51.286
616475 96.422 -56.079 51.231
616476 96.438 -56.080 51.176
616477 96.454 -56.081 51.122
616478 96.469 -56.082 51.069
616479 96.482 -56.084 51.017
616480 96.495 -56.085 50.971
616481 96.507 -56.086 50.926
616482 96.520 -56.088 50.881
616483 96.532 -56.088 50.836
616484 96.548 -56.088 50.793
616485 96.565 -56.088 50.750
616486 96.581 -56.087 50.713
616487 96.598 -56.086 50.675
616488 96.614 -56.085 50.638
616489 96.631 -56.084 50.602
616490 96.648 -56.083 50.566
616491 96.664 -56.082 50.531
616492 96.681 -56.081 50.500
616493 96.697 -56.081 50.468
616494 96.714 -56.080 50.437
616495 96.731 -56.079 50.407
616496 96.747 -56.078 50.377
616497 96.777 -56.077 50.348
616498 96.812 -56.076 50.322
616499 96.847 -56.075 50.297
616500 96.882 -56.074 50.271
616501 96.917 -56.073 50.246
616502 96.952 -56.072 50.222
616503 96.987 -56.071 50.196
616504 97.022 -56.070 50.172
616505 97.057 -56.069 50.149
616506 97.092 -56.068 50.128
616507 97.127 -56.067 50.104
616508 97.162 -56.066 50.080
616509 97.197 -56.065 50.057
616510 97.232 -56.064 50.037
616932 88.833 -52.410 57.324
616933 88.864 -52.432 57.316
616934 88.896 -52.455 57.308
616935 88.927 -52.477 57.301
616936 88.959 -52.500 57.294
616937 88.991 -52.523 57.288
616938 89.023 -52.546 57.282
616939 89.054 -52.568 57.275
616940 89.086 -52.591 57.269
616941 89.118 -52.614 57.262
616942 89.149 -52.637 57.256
616943 89.181 -52.659 57.251
616944 89.213 -52.681 57.245
616945 89.246 -52.700 57.239
616946 89.279 -52.720 57.234
616947 89.312 -52.739 57.228
616948 89.345 -52.759 57.222
616949 89.378 -52.778 57.218
616950 89.411 -52.797 57.213
616951 89.445 -52.816 57.208
616952 89.477 -52.836 57.205
616953 89.511 -52.855 57.200
616954 89.544 -52.875 57.197
616955 89.577 -52.889 57.194
616956 89.609 -52.897 57.190
616957 89.641 -52.905 57.187
616958 89.663 -52.922 57.183
616959 89.679 -52.944 57.180
616960 89.695 -52.966 57.176
616961 89.711 -52.988 57.172
616962 89.727 -53.010 57.167
616963 89.743 -53.032 57.164
616964 89.759 -53.054 57.160
616965 89.785 -53.064 57.156
616966 89.817 -53.064 57.151
616967 89.855 -53.070 57.143
616968 89.894 -53.075 57.136
616969 89.932 -53.080 57.130
616970 89.971 -53.086 57.123
616971 90.010 -53.091 57.118
616972 90.048 -53.096 57.108
616973 90.082 -53.102 57.097
616974 90.102 -53.109 57.086
616975 90.123 -53.117 57.074
616976 90.156 -53.134 57.063
616977 90.191 -53.154 57.050
616978 90.227 -53.174 57.035
616979 90.262 -53.195 57.019
616980 90.298 -53.214 57.002
616981 90.334 -53.235 56.985
616982 90.369 -53.255 56.969
616983 90.405 -53.275 56.950
616984 90.443 -53.292 56.928
616985 90.486 -53.303 56.902
616986 90.525 -53.316 56.875
616987 90.542 -53.336 56.848
616988 90.578 -53.338 56.817
616989 90.613 -53.339 56.787
616990 90.649 -53.341 56.752
616991 90.685 -53.343 56.717
616992 90.720 -53.345 56.681
616993 90.756 -53.346 56.647
616994 90.792 -53.349 56.614
616995 90.825 -53.355 56.583
616996 90.855 -53.375 56.551
616997 90.883 -53.396 56.519
616998 90.912 -53.416 56.488
617008 91.240 -53.615 55.900
617009 91.274 -53.634 55.872
617010 91.307 -53.654 55.845
617011 91.340 -53.675 55.820
617012 91.373 -53.694 55.795
617013 91.401 -53.715 55.773
617014 91.428 -53.735 55.751
617015 91.455 -53.756 55.729
617016 91.482 -53.777 55.708
617017 91.509 -53.798 55.689
617018 91.535 -53.819 55.671
617019 91.565 -53.841 55.652
617020 91.598 -53.865 55.634
617021 91.631 -53.890 55.618
617022 91.664 -53.914 55.601
617023 91.698 -53.938 55.585
617024 91.731 -53.963 55.568
617025 91.770 -53.980 55.551
617026 91.810 -53.994 55.534
617027 91.851 -54.007 55.518
617028 91.892 -54.020 55.502
617029 91.933 -54.034 55.486
617030 91.974 -54.048 55.469
617031 92.014 -54.061 55.453
617032 92.055 -54.075 55.439
617033 92.095 -54.088 55.426
617034 92.136 -54.101 55.411
617035 92.175 -54.114 55.397
617046 92.598 -54.251 55.298
617047 92.637 -54.264 55.296
617048 92.675 -54.276 55.295
617049 92.714 -54.289 55.295
617050 92.752 -54.301 55.300
617051 92.790 -54.314 55.304
617052 92.829 -54.326 55.309
617053 92.867 -54.339 55.315
617054 92.906 -54.351 55.322
617055 92.944 -54.364 55.330
617056 92.978 -54.381 55.341
617057 93.001 -54.410 55.351
617058 93.023 -54.439 55.359
617059 93.047 -54.467 55.366
617060 93.069 -54.496 55.372
617061 93.078 -54.508 55.375
617062 93.095 -54.517 55.378
617063 93.117 -54.526 55.379
617064 93.139 -54.536 55.377
617065 93.161 -54.546 55.371
617066 93.183 -54.555 55.365
617067 93.205 -54.565 55.360
617068 93.227 -54.574 55.356
617069 93.248 -54.589 55.359
617070 93.269 -54.613 55.369
617071 93.290 -54.637 55.382
617072 93.310 -54.660 55.396
617073 93.350 -54.667 55.408
617074 93.378 -54.678 55.417
617075 93.401 -54.690 55.425
617076 93.423 -54.704 55.438
617077 93.445 -54.717 55.458
617078 93.467 -54.731 55.480
617079 93.489 -54.744 55.498
617080 93.511 -54.757 55.505
617081 93.535 -54.771 55.501
617082 93.560 -54.784 55.492
617083 93.585 -54.799 55.480
617084 93.611 -54.812 55.463
617085 93.635 -54.826 55.438
617086 93.660 -54.840 55.407
617087 93.686 -54.853 55.390
617088 93.711 -54.868 55.374
617089 93.736 -54.881 55.356
617090 93.761 -54.895 55.341
617091 93.786 -54.909 55.328
617092 93.811 -54.923 55.312
617093 93.836 -54.937 55.292
617094 93.857 -54.959 55.270
617095 93.878 -54.983 55.248
617096 93.898 -55.007 55.228
617097 93.918 -55.031 55.205
617098 93.938 -55.055 55.180
617099 93.959 -55.079 55.153
617100 93.980 -55.101 55.130
617101 94.004 -55.117 55.121
617102 94.029 -55.134 55.119
617103 94.053 -55.151 55.117
617104 94.077 -55.167 55.108
617105 94.102 -55.184 55.095
617106 94.126 -55.201 55.077
617107 94.150 -55.217 55.053
617108 94.175 -55.234 55.029
617109 94.205 -55.247 55.003
617110 94.239 -55.259 54.974
617111 94.271 -55.271 54.943
617112 94.305 -55.283 54.912
617113 94.337 -55.295 54.872
617114 94.370 -55.307 54.827
617115 94.403 -55.319 54.779
617116 94.436 -55.330 54.734
617117 94.469 -55.342 54.697
617118 94.503 -55.354 54.664
617119 94.543 -55.367 54.631
617120 94.583 -55.379 54.596
617121 94.623 -55.392 54.557
617122 94.663 -55.405 54.517
617123 94.702 -55.418 54.478
617124 94.742 -55.431 54.438
617125 94.782 -55.443 54.398
617126 94.822 -55.456 54.357
617127 94.861 -55.469 54.309
617128 94.901 -55.482 54.257
617129 94.938 -55.494 54.204
617130 94.967 -55.504 54.148
617131 94.995 -55.514 54.095
617132 95.024 -55.524 54.043
617133 95.053 -55.539 53.987
617134 95.082 -55.555 53.929
617135 95.111 -55.571 53.868
617136 95.140 -55.586 53.805
617137 95.170 -55.602 53.743
617138 95.199 -55.619 53.687
617139 95.229 -55.635 53.633
617140 95.258 -55.650 53.579
617141 95.289 -55.665 53.522
617142 95.321 -55.678 53.458
617143 95.354 -55.691 53.388
617144 95.387 -55.704 53.314
617145 95.420 -55.717 53.240
617146 95.454 -55.729 53.166
617147 95.487 -55.741 53.094
617148 95.521 -55.753 53.024
617149 95.555 -55.765 52.957
617150 95.589 -55.778 52.895
617151 95.622 -55.790 52.829
617152 95.656 -55.803 52.760
617153 95.688 -55.816 52.691
617154 95.719 -55.830 52.623
617155 95.752 -55.843 52.553
617156 95.786 -55.857 52.484
617157 95.820 -55.871 52.414
617158 95.854 -55.885 52.345
617159 95.888 -55.899 52.278
617160 95.922 -55.912 52.212
617161 95.956 -55.926 52.147
617162 95.990 -55.940 52.085
617163 96.024 -55.954 52.024
617164 96.058 -55.968 51.964
617165 96.092 -55.982 51.904
617166 96.126 -55.996 51.844
617167 96.160 -56.009 51.782
617168 96.194 -56.023 51.721
617169 96.229 -56.038 51.661
617170 96.261 -56.046 51.599
617171 96.294 -56.048 51.536
617172 96.325 -56.050 51.472
617173 96.357 -56.052 51.409
617174 96.389 -56.054 51.348
617175 96.417 -56.056 51.290
617176 96.433 -56.057 51.233
617177 96.449 -56.058 51.177
617178 96.464 -56.059 51.121
617179 96.478 -56.060 51.066
617180 96.490 -56.062 51.013
617181 96.503 -56.063 50.964
617182 96.515 -56.064 50.917
617183 96.529 -56.065 50.870
617184 96.546 -56.064 50.824
617185 96.562 -56.063 50.779
617186 96.579 -56.063 50.736
617187 96.595 -56.061 50.697
617188 96.612 -56.060 50.658
617189 96.628 -56.059 50.621
617193 96.695 -56.056 50.479
617194 96.712 -56.056 50.448
617195 96.728 -56.055 50.417
617196 96.745 -56.054 50.386
617197 96.761 -56.053 50.355
617198 96.793 -56.052 50.326
617199 96.828 -56.051 50.300
617200 96.863 -56.050 50.274
617201 96.898 -56.049 50.248
617202 96.933 -56.048 50.222
617203 96.968 -56.047 50.196
617204 97.003 -56.046 50.171
617205 97.038 -56.045 50.149
617206 97.073 -56.044 50.126
617207 97.108 -56.043 50.103
617208 97.143 -56.042 50.079
617209 97.178 -56.041 50.055
617210 97.213 -56.040 50.033
617633 88.849 -52.386 57.317
617634 88.881 -52.409 57.308
617635 88.912 -52.431 57.300
617636 88.944 -52.454 57.292
617637 88.975 -52.477 57.285
617638 89.007 -52.499 57.279
617639 89.039 -52.522 57.273
617640 89.070 -52.544 57.266
617641 89.102 -52.567 57.259
617642 89.134 -52.590 57.252
617643 89.167 -52.609 57.246
617644 89.200 -52.629 57.241
617645 89.233 -52.648 57.235
617646 89.266 -52.668 57.229
617647 89.299 -52.687 57.223
617648 89.332 -52.706 57.217
617649 89.365 -52.725 57.212
617650 89.398 -52.744 57.207
617651 89.431 -52.764 57.202
617652 89.464 -52.783 57.199
617653 89.497 -52.803 57.195
617654 89.531 -52.822 57.190
617655 89.563 -52.841 57.187
617656 89.596 -52.856 57.183
617657 89.628 -52.863 57.180
617658 89.661 -52.872 57.178
617659 89.693 -52.879 57.175
617660 89.710 -52.901 57.172
617661 89.726 -52.923 57.168
617662 89.742 -52.945 57.164
617663 89.758 -52.967 57.161
617664 89.774 -52.989 57.157
617665 89.791 -53.011 57.154
617666 89.808 -53.032 57.151
617667 89.840 -53.032 57.146
617668 89.873 -53.032 57.139
617669 89.904 -53.032 57.133
617670 89.941 -53.035 57.127
617671 89.979 -53.040 57.121
617672 90.018 -53.045 57.116
617673 90.056 -53.051 57.108
617674 90.095 -53.056 57.098
617675 90.116 -53.064 57.088
617676 90.141 -53.073 57.078
617677 90.176 -53.094 57.068
617678 90.212 -53.114 57.057
617679 90.247 -53.134 57.044
617680 90.283 -53.154 57.030
617681 90.319 -53.174 57.015
617682 90.355 -53.195 57.001
617683 90.390 -53.214 56.987
617684 90.426 -53.235 56.972
617685 90.461 -53.255 56.952
617686 90.502 -53.269 56.930
617687 90.532 -53.285 56.905
617688 90.565 -53.289 56.878
617689 90.601 -53.291 56.849
617690 90.637 -53.292 56.819
617691 90.672 -53.295 56.785
617692 90.708 -53.296 56.748
617693 90.743 -53.298 56.712
617694 90.779 -53.300 56.677
617695 90.815 -53.302 56.643
617696 90.845 -53.319 56.612
617697 90.874 -53.339 56.579
617698 90.902 -53.359 56.547
617699 90.932 -53.379 56.516
617700 90.965 -53.399 56.486
617709 91.265 -53.577 55.920
617710 91.298 -53.598 55.891
617711 91.325 -53.618 55.864
617712 91.352 -53.639 55.837
617713 91.378 -53.660 55.813
617714 91.405 -53.681 55.790
617715 91.432 -53.701 55.766
617716 91.459 -53.722 55.744
617717 91.486 -53.743 55.723
617718 91.513 -53.764 55.702
617719 91.545 -53.787 55.683
617720 91.578 -53.812 55.663
617721 91.612 -53.837 55.645
617722 91.645 -53.861 55.627
617723 91.678 -53.885 55.610
617724 91.712 -53.910 55.592
617725 91.745 -53.935 55.574
617726 91.784 -53.951 55.557
617727 91.824 -53.965 55.540
617728 91.865 -53.979 55.524
617729 91.906 -53.991 55.507
617730 91.947 -54.005 55.490
617731 91.988 -54.019 55.473
617732 92.028 -54.032 55.456
617733 92.069 -54.046 55.442
617734 92.110 -54.059 55.428
617735 92.150 -54.073 55.413
617747 92.618 -54.225 55.298
617748 92.657 -54.238 55.297
617749 92.695 -54.250 55.297
617750 92.733 -54.262 55.296
617751 92.772 -54.275 55.301
617752 92.810 -54.287 55.305
617753 92.849 -54.300 55.311
617754 92.887 -54.312 55.318
617755 92.926 -54.325 55.325
617756 92.964 -54.337 55.335
617757 93.001 -54.352 55.348
617758 93.023 -54.380 55.359
617759 93.047 -54.409 55.366
617760 93.069 -54.438 55.371
617761 93.092 -54.467 55.375
617762 93.093 -54.471 55.375
617763 93.101 -54.476 55.373
617764 93.123 -54.485 55.368
617765 93.145 -54.495 55.362
617766 93.167 -54.505 55.354
617767 93.189 -54.515 55.346
617768 93.211 -54.524 55.340
617769 93.232 -54.544 55.340
617770 93.253 -54.568 55.345
617771 93.273 -54.591 55.355
617772 93.294 -54.615 55.369
617773 93.323 -54.631 55.383
617774 93.364 -54.637 55.393
617775 93.390 -54.648 55.403
617776 93.412 -54.662 55.413
617777 93.434 -54.674 55.429
617778 93.457 -54.688 55.451
617779 93.479 -54.702 55.472
617780 93.501 -54.715 55.486
617781 93.523 -54.728 55.491
617782 93.545 -54.741 55.484
617783 93.567 -54.755 55.474
617784 93.592 -54.768 55.461
617785 93.617 -54.782 55.443
617786 93.642 -54.796 55.420
617787 93.667 -54.810 55.392
617788 93.692 -54.824 55.375
617789 93.717 -54.837 55.364
617790 93.743 -54.851 55.350
617791 93.767 -54.866 55.335
617792 93.792 -54.879 55.321
617793 93.818 -54.893 55.302
617794 93.843 -54.906 55.282
617795 93.865 -54.925 55.262
617796 93.886 -54.950 55.242
617797 93.906 -54.974 55.223
617798 93.926 -54.998 55.201
617799 93.946 -55.022 55.176
617800 93.967 -55.046 55.147
617801 93.990 -55.064 55.125
617802 94.014 -55.081 55.118
617803 94.039 -55.097 55.121
617804 94.063 -55.114 55.122
617805 94.087 -55.131 55.117
617806 94.112 -55.148 55.107
617807 94.136 -55.164 55.090
617808 94.160 -55.181 55.069
617809 94.185 -55.198 55.046
617810 94.214 -55.211 55.019
617811 94.247 -55.223 54.989
617812 94.280 -55.235 54.955
617813 94.313 -55.247 54.922
617814 94.346 -55.259 54.881
617815 94.379 -55.271 54.836
617816 94.412 -55.282 54.788
617817 94.445 -55.295 54.741
617818 94.478 -55.307 54.704
617819 94.516 -55.318 54.669
617820 94.553 -55.332 54.637
617821 94.591 -55.345 54.602
617822 94.628 -55.359 54.563
617823 94.668 -55.372 54.522
617824 94.708 -55.385 54.483
617825 94.748 -55.398 54.441
617826 94.787 -55.411 54.399
617827 94.827 -55.423 54.357
617828 94.867 -55.436 54.309
617829 94.907 -55.448 54.255
617830 94.946 -55.461 54.202
617831 94.974 -55.471 54.149
617832 95.002 -55.481 54.096
617833 95.031 -55.495 54.045
617834 95.060 -55.511 53.990
617835 95.090 -55.527 53.930
617836 95.119 -55.543 53.868
617837 95.148 -55.559 53.805
617838 95.178 -55.575 53.742
617839 95.208 -55.589 53.686
617840 95.239 -55.605 53.631
617841 95.270 -55.620 53.576
617842 95.302 -55.633 53.518
617843 95.333 -55.646 53.455
617844 95.366 -55.660 53.386
617845 95.399 -55.672 53.314
617846 95.432 -55.685 53.241
617847 95.465 -55.698 53.167
617848 95.498 -55.711 53.094
617849 95.531 -55.724 53.023
617850 95.564 -55.737 52.956
617851 95.598 -55.749 52.891
617852 95.631 -55.761 52.826
617853 95.665 -55.773 52.760
617854 95.698 -55.786 52.695
617855 95.730 -55.799 52.629
617856 95.764 -55.813 52.561
617857 95.799 -55.828 52.490
617858 95.832 -55.841 52.420
617859 95.867 -55.855 52.350
617860 95.901 -55.869 52.282
617861 95.935 -55.883 52.215
617862 95.969 -55.897 52.150
617863 96.003 -55.910 52.088
617864 96.037 -55.925 52.027
617865 96.071 -55.938 51.967
617866 96.105 -55.952 51.907
617867 96.139 -55.966 51.846
617868 96.173 -55.980 51.785
617869 96.207 -55.994 51.725
617870 96.241 -56.007 51.665
617871 96.275 -56.022 51.604
617872 96.307 -56.026 51.542
617873 96.339 -56.028 51.479
617874 96.371 -56.031 51.416
617875 96.403 -56.032 51.354
617876 96.428 -56.034 51.295
617877 96.444 -56.035 51.237
617878 96.459 -56.036 51.179
617879 96.473 -56.038 51.121
617880 96.486 -56.038 51.065
617881 96.498 -56.040 51.010
617882 96.511 -56.041 50.958
617883 96.527 -56.041 50.910
617884 96.543 -56.040 50.861
617885 96.560 -56.039 50.813
617886 96.577 -56.038 50.768
617899 96.809 -56.027 50.305
617900 96.844 -56.026 50.279
617901 96.879 -56.025 50.252
617902 96.914 -56.024 50.227
617903 96.949 -56.023 50.200
617904 96.984 -56.022 50.174
617905 97.019 -56.021 50.149
617906 97.054 -56.020 50.127
617907 97.089 -56.019 50.104
617908 97.124 -56.018 50.081
617909 97.159 -56.016 50.057
618334 88.865 -52.362 57.310
618335 88.897 -52.384 57.300
618336 88.928 -52.407 57.292
618337 88.960 -52.430 57.284
618338 88.991 -52.452 57.277
618339 89.023 -52.475 57.270
618340 89.055 -52.498 57.264
618341 89.088 -52.518 57.256
618342 89.120 -52.537 57.249
618343 89.154 -52.557 57.242
618344 89.187 -52.576 57.236
618345 89.220 -52.596 57.231
618346 89.253 -52.615 57.225
618347 89.286 -52.634 57.219
618348 89.319 -52.653 57.213
618349 89.352 -52.673 57.207
618350 89.385 -52.692 57.201
618351 89.418 -52.712 57.197
618352 89.451 -52.731 57.192
618353 89.484 -52.750 57.189
618354 89.518 -52.769 57.185
618355 89.550 -52.789 57.181
618356 89.584 -52.808 57.177
618357 89.616 -52.822 57.174
618358 89.649 -52.830 57.172
618359 89.681 -52.838 57.169
618360 89.713 -52.846 57.167
618361 89.742 -52.858 57.164
618362 89.757 -52.880 57.160
618363 89.773 -52.902 57.157
618364 89.790 -52.924 57.155
618365 89.806 -52.946 57.151
618366 89.822 -52.969 57.149
618367 89.838 -52.991 57.146
618368 89.864 -52.999 57.142
618369 89.896 -52.998 57.136
618370 89.928 -52.998 57.131
618371 89.960 -52.998 57.126
618372 89.992 -52.998 57.121
618373 90.027 -53.001 57.116
618374 90.065 -53.006 57.109
618375 90.104 -53.011 57.101
618376 90.130 -53.017 57.092
618377 90.161 -53.033 57.083
618378 90.197 -53.054 57.073
618379 90.233 -53.073 57.063
618380 90.268 -53.094 57.052
618381 90.304 -53.114 57.041
618382 90.339 -53.134 57.029
618383 90.375 -53.154 57.017
618384 90.410 -53.174 57.006
618385 90.446 -53.195 56.993
618386 90.482 -53.214 56.977
618387 90.517 -53.235 56.958
618388 90.552 -53.241 56.935
618389 90.588 -53.242 56.909
618390 90.624 -53.244 56.881
618391 90.660 -53.246 56.850
618392 90.695 -53.248 56.816
618393 90.731 -53.249 56.780
618394 90.766 -53.251 56.743
618395 90.802 -53.253 56.707
618396 90.835 -53.262 56.675
618397 90.864 -53.282 56.642
618398 90.893 -53.302 56.609
618399 90.923 -53.323 56.577
618400 90.956 -53.342 56.545
618401 90.990 -53.362 56.515
618402 91.023 -53.382 56.486
618410 91.275 -53.543 55.941
618411 91.302 -53.564 55.912
618412 91.329 -53.584 55.884
618413 91.356 -53.605 55.857
618414 91.383 -53.626 55.831
618415 91.410 -53.646 55.807
618416 91.436 -53.667 55.784
618417 91.463 -53.688 55.760
618418 91.493 -53.710 55.738
618419 91.526 -53.734 55.717
618420 91.559 -53.759 55.696
618421 91.592 -53.784 55.676
618422 91.626 -53.808 55.657
618423 91.659 -53.833 55.639
618424 91.693 -53.857 55.621
618425 91.726 -53.882 55.602
618426 91.759 -53.907 55.582
618427 91.798 -53.922 55.564
618428 91.839 -53.936 55.548
618429 91.879 -53.949 55.530
618430 91.920 -53.963 55.513
618431 91.961 -53.976 55.495
618432 92.002 -53.990 55.477
618433 92.042 -54.004 55.460
618434 92.083 -54.016 55.445
618435 92.124 -54.030 55.430
618436 92.165 -54.044 55.415
618448 92.638 -54.198 55.299
618449 92.676 -54.211 55.298
618450 92.715 -54.223 55.298
618451 92.754 -54.236 55.298
618452 92.792 -54.248 55.301
618453 92.831 -54.261 55.307
618454 92.869 -54.273 55.313
618455 92.907 -54.286 55.320
618456 92.946 -54.298 55.328
618457 92.984 -54.311 55.340
618458 93.022 -54.323 55.355
618459 93.047 -54.351 55.366
618460 93.069 -54.380 55.373
618461 93.092 -54.408 55.376
618462 93.109 -54.430 55.377
618463 93.109 -54.433 55.373
618464 93.110 -54.435 55.366
618465 93.129 -54.444 55.356
618466 93.151 -54.454 55.346
618467 93.173 -54.464 55.336
618468 93.195 -54.475 55.327
618469 93.216 -54.499 55.323
618470 93.236 -54.523 55.326
618471 93.257 -54.546 55.333
618472 93.278 -54.570 55.343
618473 93.298 -54.593 55.357
618474 93.337 -54.601 55.372
618475 93.378 -54.606 55.382
618476 93.402 -54.619 55.391
618477 93.424 -54.632 55.403
618478 93.446 -54.646 55.421
618479 93.468 -54.659 55.443
618480 93.490 -54.672 55.461
618481 93.512 -54.686 55.469
618482 93.535 -54.699 55.473
618483 93.557 -54.712 55.463
618484 93.579 -54.725 55.452
618485 93.601 -54.739 55.439
618486 93.623 -54.753 55.422
618487 93.648 -54.766 55.404
618488 93.674 -54.780 55.385
618489 93.699 -54.793 55.375
618490 93.724 -54.807 55.365
618491 93.749 -54.822 55.351
618492 93.774 -54.835 55.334
618493 93.799 -54.849 55.315
618494 93.824 -54.862 55.293
618495 93.849 -54.877 55.272
618496 93.873 -54.893 55.253
618497 93.894 -54.916 55.234
618498 93.914 -54.941 55.216
618499 93.934 -54.965 55.195
618500 93.954 -54.989 55.169
618501 93.975 -55.011 55.139
618502 94.000 -55.028 55.119
618503 94.024 -55.044 55.116
618504 94.048 -55.061 55.120
618505 94.072 -55.078 55.124
618506 94.097 -55.094 55.122
618507 94.121 -55.111 55.114
618508 94.145 -55.128 55.098
618509 94.170 -55.144 55.081
618510 94.194 -55.161 55.059
618511 94.223 -55.176 55.032
618512 94.256 -55.187 55.001
618513 94.289 -55.199 54.965
618514 94.322 -55.211 54.928
618515 94.355 -55.223 54.886
618516 94.388 -55.235 54.842
618517 94.421 -55.247 54.794
618518 94.455 -55.259 54.746
618519 94.492 -55.270 54.708
618520 94.530 -55.282 54.673
618521 94.567 -55.295 54.639
618522 94.604 -55.309 54.605
618523 94.642 -55.323 54.566
618524 94.679 -55.336 54.524
618525 94.716 -55.350 54.483
618526 94.754 -55.364 54.441
618527 94.792 -55.377 54.397
618528 94.832 -55.390 54.354
618529 94.872 -55.403 54.306
618530 94.912 -55.416 54.254
618531 94.952 -55.428 54.203
618532 94.981 -55.439 54.150
618533 95.010 -55.452 54.096
618534 95.039 -55.467 54.043
618535 95.068 -55.483 53.986
618536 95.098 -55.499 53.925
618537 95.128 -55.514 53.862
618538 95.159 -55.530 53.798
618539 95.189 -55.545 53.737
618540 95.219 -55.559 53.680
618541 95.251 -55.574 53.624
618542 95.282 -55.588 53.567
618543 95.314 -55.601 53.509
618544 95.346 -55.614 53.447
618545 95.379 -55.628 53.381
618546 95.411 -55.641 53.311
618547 95.445 -55.654 53.239
618548 95.478 -55.666 53.165
618549 95.511 -55.679 53.092
618550 95.544 -55.692 53.020
618551 95.577 -55.705 52.953
618552 95.610 -55.718 52.886
618553 95.643 -55.730 52.820
618554 95.676 -55.743 52.757
618555 95.708 -55.756 52.696
618556 95.743 -55.770 52.633
618557 95.777 -55.784 52.566
618558 95.811 -55.797 52.494
618559 95.845 -55.812 52.423
618560 95.879 -55.825 52.353
618561 95.913 -55.839 52.284
618562 95.947 -55.853 52.218
618563 95.981 -55.867 52.153
618564 96.015 -55.881 52.091
618565 96.049 -55.894 52.031
618566 96.083 -55.909 51.971
618567 96.117 -55.922 51.910
618568 96.152 -55.936 51.850
618569 96.185 -55.950 51.790
618570 96.220 -55.964 51.729
618571 96.253 -55.978 51.670
618572 96.288 -55.991 51.610
618573 96.321 -56.005 51.549
618574 96.354 -56.006 51.487
618575 96.385 -56.009 51.424
618576 96.417 -56.010 51.361
618577 96.439 -56.012 51.301
618578 96.454 -56.013 51.242
618579 96.469 -56.014 51.182
618580 96.481 -56.016 51.124
618581 96.493 -56.017 51.066
618582 96.508 -56.017 51.009
618583 96.525 -56.016 50.956
618584 96.541 -56.016 50.906
618585 96.558 -56.015 50.855
618586 96.575 -56.013 50.806
618587 96.591 -56.013 50.759
619035 88.881 -52.338 57.302
619036 88.913 -52.361 57.292
619037 88.945 -52.384 57.284
619038 88.976 -52.406 57.275
619039 89.008 -52.427 57.268
619040 89.041 -52.446 57.261
619041 89.074 -52.466 57.254
619042 89.107 -52.485 57.246
619043 89.140 -52.505 57.239
619044 89.174 -52.524 57.232
619045 89.207 -52.543 57.226
619046 89.240 -52.562 57.220
619047 89.273 -52.581 57.214
619048 89.306 -52.601 57.209
619049 89.339 -52.620 57.203
619050 89.372 -52.640 57.196
619051 89.405 -52.659 57.191
619052 89.438 -52.678 57.187
619053 89.471 -52.697 57.183
619054 89.505 -52.717 57.179
619055 89.538 -52.736 57.176
619056 89.570 -52.756 57.171
619057 89.604 -52.775 57.168
619058 89.636 -52.789 57.165
619059 89.669 -52.797 57.163
619060 89.701 -52.805 57.161
619061 89.733 -52.813 57.158
619062 89.765 -52.821 57.156
619063 89.789 -52.837 57.153
619064 89.805 -52.859 57.151
619065 89.821 -52.881 57.149
619066 89.837 -52.903 57.146
619067 89.853 -52.926 57.143
619068 89.869 -52.948 57.141
619069 89.888 -52.966 57.138
619070 89.920 -52.966 57.134
619071 89.952 -52.966 57.129
619072 89.984 -52.966 57.125
619073 90.016 -52.966 57.120
619074 90.048 -52.966 57.116
619075 90.080 -52.966 57.111
619076 90.112 -52.966 57.104
619077 90.146 -52.973 57.097
619078 90.182 -52.993 57.088
619079 90.218 -53.013 57.080
619080 90.253 -53.033 57.072
619081 90.289 -53.054 57.063
619082 90.323 -53.074 57.053
619083 90.359 -53.095 57.043
619084 90.394 -53.116 57.033
619085 90.429 -53.136 57.024
619086 90.464 -53.157 57.014
619087 90.499 -53.177 57.002
619088 90.534 -53.198 56.986
619089 90.575 -53.195 56.965
619090 90.611 -53.195 56.939
619091 90.647 -53.197 56.912
619092 90.682 -53.199 56.882
619093 90.718 -53.201 56.848
619094 90.754 -53.202 56.811
619095 90.790 -53.205 56.774
619096 90.825 -53.206 56.738
619097 90.855 -53.226 56.706
619098 90.883 -53.245 56.673
619099 90.915 -53.266 56.639
619100 90.948 -53.286 56.607
619101 90.981 -53.305 56.576
619102 91.014 -53.325 56.545
619103 91.047 -53.345 56.516
619104 91.081 -53.365 56.488
619105 91.114 -53.385 56.458
619111 91.280 -53.509 55.963
619112 91.306 -53.530 55.933
619113 91.333 -53.550 55.905
619114 91.360 -53.571 55.879
619115 91.387 -53.592 55.851
619116 91.414 -53.612 55.826
619117 91.441 -53.633 55.802
619118 91.473 -53.657 55.778
619119 91.506 -53.681 55.755
619120 91.539 -53.706 55.733
619121 91.573 -53.731 55.712
619122 91.606 -53.755 55.691
619123 91.640 -53.780 55.670
619124 91.673 -53.804 55.651
619125 91.706 -53.828 55.632
619126 91.739 -53.853 55.612
619127 91.773 -53.878 55.592
619128 91.812 -53.894 55.573
619129 91.853 -53.907 55.555
619130 91.893 -53.920 55.537
619131 91.934 -53.934 55.519
619132 91.975 -53.947 55.500
619133 92.016 -53.961 55.482
619134 92.057 -53.975 55.465
619135 92.097 -53.988 55.448
619136 92.138 -54.001 55.433
619149 92.658 -54.173 55.299
619150 92.697 -54.185 55.298
619151 92.735 -54.198 55.298
619152 92.774 -54.210 55.298
619153 92.812 -54.223 55.302
619154 92.850 -54.235 55.307
619155 92.889 -54.247 55.314
619156 92.927 -54.260 55.321
619157 92.966 -54.272 55.329
619158 93.004 -54.285 55.344
619159 93.043 -54.297 55.360
619160 93.069 -54.322 55.372
619161 93.092 -54.351 55.379
619162 93.115 -54.380 55.380
619163 93.124 -54.392 55.378
619164 93.125 -54.395 55.369
619165 93.125 -54.398 55.356
619166 93.136 -54.404 55.342
619167 93.157 -54.413 55.329
619168 93.179 -54.430 55.317
619169 93.200 -54.454 55.310
619170 93.220 -54.477 55.308
619171 93.241 -54.502 55.313
619172 93.262 -54.525 55.321
619173 93.282 -54.549 55.332
619174 93.310 -54.565 55.346
619175 93.352 -54.571 55.360
619176 93.391 -54.577 55.371
619177 93.413 -54.590 55.380
619178 93.435 -54.603 55.391
619179 93.458 -54.616 55.410
619180 93.480 -54.630 55.430
619181 93.502 -54.643 55.444
619182 93.524 -54.656 55.449
619183 93.546 -54.670 55.447
619184 93.568 -54.683 55.437
619185 93.590 -54.696 55.426
619186 93.613 -54.709 55.413
619187 93.635 -54.723 55.399
619188 93.657 -54.737 55.387
619189 93.680 -54.750 55.384
619190 93.705 -54.764 55.383
619191 93.730 -54.778 55.373
619192 93.755 -54.791 55.355
619193 93.780 -54.805 55.333
619194 93.805 -54.819 55.307
619195 93.831 -54.833 55.283
619196 93.856 -54.847 55.260
619197 93.880 -54.860 55.242
619198 93.901 -54.884 55.225
619199 93.922 -54.908 55.207
619200 93.942 -54.932 55.186
619201 93.962 -54.956 55.158
619202 93.985 -54.975 55.128
619203 94.009 -54.991 55.113
619204 94.034 -55.008 55.111
619205 94.058 -55.025 55.115
619206 94.082 -55.041 55.118
619207 94.107 -55.058 55.117
619208 94.131 -55.075 55.112
619209 94.155 -55.091 55.101
619210 94.180 -55.108 55.084
619211 94.204 -55.125 55.064
619212 94.231 -55.139 55.038
619213 94.265 -55.151 55.004
619214 94.297 -55.163 54.968
619215 94.331 -55.176 54.930
619216 94.363 -55.187 54.888
619217 94.397 -55.199 54.844
619218 94.431 -55.211 54.798
619219 94.469 -55.223 54.750
619220 94.506 -55.234 54.711
619221 94.544 -55.246 54.674
619222 94.581 -55.258 54.639
619223 94.618 -55.273 54.604
619224 94.655 -55.286 54.565
619225 94.693 -55.300 54.523
619226 94.731 -55.314 54.480
619227 94.768 -55.327 54.437
619228 94.805 -55.341 54.392
619229 94.842 -55.355 54.346
619230 94.880 -55.369 54.299
619231 94.917 -55.383 54.249
619232 94.957 -55.395 54.198
619233 94.988 -55.408 54.147
619234 95.018 -55.423 54.092
619235 95.048 -55.439 54.036
619236 95.079 -55.454 53.977
619237 95.109 -55.469 53.915
619238 95.140 -55.484 53.852
619239 95.170 -55.499 53.788
619240 95.200 -55.514 53.726
619241 95.232 -55.529 53.666
619242 95.263 -55.542 53.608
619243 95.295 -55.555 53.551
619244 95.327 -55.569 53.492
619245 95.359 -55.583 53.431
619246 95.391 -55.596 53.366
619247 95.424 -55.609 53.300
619248 95.457 -55.622 53.231
619249 95.490 -55.635 53.159
619250 95.523 -55.648 53.086
619251 95.556 -55.660 53.015
619252 95.589 -55.673 52.946
619253 95.622 -55.686 52.879
619254 95.655 -55.699 52.814
619255 95.688 -55.712 52.753
619256 95.720 -55.723 52.693
619257 95.753 -55.736 52.631
619258 95.788 -55.751 52.565
619259 95.822 -55.765 52.494
619260 95.856 -55.780 52.423
619261 95.891 -55.795 52.353
619262 95.925 -55.809 52.285
619263 95.960 -55.823 52.220
619264 95.994 -55.837 52.156
619265 96.028 -55.851 52.094
619266 96.062 -55.865 52.034
619267 96.096 -55.878 51.974
619268 96.130 -55.893 51.913
619269 96.164 -55.906 51.854
619270 96.198 -55.920 51.795
619271 96.232 -55.934 51.733
619272 96.266 -55.948 51.674
619273 96.300 -55.962 51.615
619274 96.334 -55.975 51.556
619275 96.367 -55.984 51.495
619276 96.399 -55.987 51.432
619277 96.431 -55.989 51.368
619278 96.449 -55.990 51.308
619279 96.464 -55.991 51.247
619280 96.476 -55.992 51.187
619281 96.489 -55.994 51.127
619282 96.506 -55.993 51.068
619283 96.522 -55.992 51.008
619284 96.539 -55.991 50.954
619285 96.555 -55.991 50.901
619286 96.572 -55.989 50.850
619287 96.589 -55.988 50.800
619288 96.605 -55.987 50.751
619736 88.897 -52.314 57.294
619737 88.929 -52.336 57.284
619738 88.962 -52.355 57.275
619739 88.995 -52.374 57.266
619740 89.028 -52.394 57.258
619741 89.061 -52.413 57.251
619742 89.095 -52.433 57.245
619743 89.127 -52.452 57.237
619744 89.161 -52.471 57.230
619745 89.194 -52.490 57.222
619746 89.227 -52.510 57.215
619747 89.260 -52.529 57.210
619748 89.293 -52.549 57.204
619749 89.326 -52.568 57.198
619750 89.359 -52.587 57.192
619751 89.392 -52.606 57.186
619752 89.425 -52.626 57.181
619753 89.458 -52.645 57.177
619754 89.491 -52.665 57.173
619755 89.525 -52.684 57.168
619756 89.557 -52.703 57.166
619757 89.591 -52.722 57.163
619758 89.624 -52.741 57.159
619759 89.657 -52.756 57.157
619760 89.689 -52.763 57.155
619761 89.721 -52.772 57.153
619762 89.753 -52.779 57.150
619763 89.786 -52.788 57.148
619764 89.818 -52.795 57.146
619765 89.836 -52.816 57.144
619766 89.852 -52.838 57.143
619767 89.868 -52.860 57.140
619768 89.884 -52.882 57.138
619769 89.900 -52.904 57.137
619770 89.916 -52.926 57.135
619771 89.944 -52.933 57.130
619772 89.976 -52.933 57.127
619773 90.008 -52.933 57.124
619774 90.040 -52.933 57.120
619775 90.072 -52.933 57.117
619776 90.104 -52.933 57.113
619777 90.136 -52.933 57.106
619778 90.167 -52.935 57.101
619779 90.199 -52.957 57.094
619780 90.234 -52.977 57.087
619781 90.269 -52.998 57.081
619782 90.304 -53.019 57.074
619783 90.339 -53.039 57.065
619784 90.375 -53.060 57.057
619785 90.410 -53.080 57.049
619786 90.445 -53.101 57.042
619787 90.480 -53.122 57.036
619788 90.515 -53.142 57.029
619789 90.550 -53.163 57.018
619790 90.591 -53.161 57.000
619791 90.632 -53.153 56.969
619792 90.670 -53.151 56.941
619793 90.706 -53.152 56.911
619794 90.741 -53.154 56.878
619795 90.777 -53.156 56.842
619796 90.813 -53.158 56.805
619797 90.845 -53.169 56.770
619798 90.874 -53.189 56.736
619799 90.906 -53.209 56.704
619800 90.939 -53.229 56.671
619801 90.972 -53.248 56.637
619802 91.006 -53.268 56.607
619803 91.039 -53.289 56.576
619804 91.069 -53.308 56.548
619805 91.096 -53.330 56.519
619806 91.122 -53.350 56.489
619807 91.150 -53.371 56.459
619812 91.284 -53.474 55.989
619813 91.310 -53.496 55.958
619814 91.338 -53.516 55.929
619815 91.364 -53.537 55.901
619816 91.391 -53.558 55.873
619817 91.420 -53.580 55.848
619818 91.453 -53.604 55.823
619819 91.487 -53.628 55.797
619820 91.520 -53.653 55.773
619821 91.554 -53.678 55.750
619822 91.587 -53.702 55.728
619823 91.620 -53.727 55.707
619824 91.653 -53.751 55.685
619825 91.687 -53.775 55.665
619826 91.720 -53.800 55.645
619827 91.753 -53.825 55.624
619828 91.787 -53.849 55.604
619829 91.826 -53.865 55.583
619830 91.867 -53.878 55.565
619831 91.908 -53.891 55.546
619832 91.948 -53.905 55.526
619833 91.989 -53.919 55.507
619834 92.030 -53.932 55.489
619835 92.071 -53.945 55.470
619836 92.111 -53.959 55.453
619837 92.152 -53.972 55.437
619850 92.678 -54.146 55.299
619851 92.716 -54.159 55.297
619852 92.755 -54.171 55.297
619853 92.793 -54.183 55.298
619854 92.832 -54.196 55.301
619855 92.870 -54.208 55.307
619856 92.909 -54.221 55.313
619857 92.947 -54.233 55.320
619858 92.986 -54.246 55.327
619859 93.024 -54.258 55.341
619860 93.063 -54.270 55.358
619861 93.092 -54.292 55.370
619862 93.115 -54.321 55.376
619863 93.138 -54.350 55.377
619864 93.140 -54.355 55.372
619865 93.140 -54.357 55.359
619866 93.140 -54.360 55.343
619867 93.142 -54.363 55.326
619868 93.162 -54.386 55.310
619869 93.183 -54.409 55.298
619870 93.204 -54.433 55.293
619871 93.225 -54.456 55.293
619872 93.246 -54.480 55.298
619873 93.266 -54.504 55.308
619874 93.287 -54.527 55.321
619875 93.325 -54.535 55.335
619876 93.366 -54.540 55.348
619877 93.403 -54.547 55.359
619878 93.425 -54.561 55.367
619879 93.447 -54.574 55.378
619880 93.469 -54.587 55.395
619881 93.491 -54.601 55.411
619882 93.513 -54.614 55.422
619883 93.536 -54.627 55.426
619884 93.558 -54.640 55.420
619885 93.580 -54.654 55.411
619886 93.602 -54.667 55.401
619887 93.624 -54.681 55.392
619888 93.646 -54.694 55.387
619889 93.669 -54.707 55.390
619890 93.691 -54.721 55.391
619891 93.713 -54.734 55.388
619892 93.736 -54.747 55.377
619893 93.762 -54.761 55.356
619894 93.787 -54.775 55.328
619895 93.812 -54.789 55.297
619896 93.837 -54.803 55.268
619897 93.862 -54.816 55.247
619898 93.887 -54.831 55.229
619899 93.909 -54.850 55.213
619900 93.929 -54.875 55.196
619901 93.949 -54.899 55.174
619902 93.970 -54.922 55.145
619903 93.995 -54.939 55.115
619904 94.019 -54.956 55.105
619905 94.043 -54.972 55.104
619906 94.068 -54.988 55.107
619907 94.092 -55.005 55.110
619908 94.116 -55.022 55.109
619909 94.140 -55.038 55.106
619910 94.165 -55.055 55.098
619911 94.189 -55.072 55.083
619912 94.213 -55.088 55.064
619913 94.241 -55.104 55.039
619914 94.273 -55.116 55.005
619915 94.307 -55.128 54.969
619916 94.339 -55.139 54.930
619917 94.372 -55.151 54.889
619918 94.407 -55.163 54.847
619919 94.445 -55.175 54.802
619920 94.482 -55.186 54.754
619921 94.520 -55.198 54.711
619922 94.557 -55.210 54.671
619923 94.595 -55.222 54.634
619924 94.632 -55.235 54.597
619925 94.670 -55.249 54.558
619926 94.707 -55.263 54.514
619927 94.744 -55.277 54.471
619928 94.782 -55.291 54.425
619929 94.819 -55.304 54.378
619930 94.856 -55.318 54.332
619931 94.894 -55.332 54.286
619932 94.931 -55.346 54.238
619933 94.967 -55.361 54.187
619934 94.999 -55.378 54.138
619935 95.029 -55.393 54.082
619936 95.060 -55.408 54.025
619937 95.090 -55.423 53.964
619938 95.120 -55.439 53.901
619939 95.151 -55.454 53.839
619940 95.181 -55.469 53.776
619941 95.212 -55.483 53.712
619942 95.244 -55.497 53.649
619943 95.276 -55.510 53.589
619944 95.308 -55.524 53.531
619945 95.340 -55.537 53.471
619946 95.372 -55.551 53.409
619947 95.404 -55.564 53.346
619948 95.436 -55.577 53.284
619949 95.469 -55.590 53.219
619950 95.502 -55.603 53.150
619951 95.535 -55.616 53.077
619952 95.568 -55.629 53.007
619953 95.601 -55.642 52.937
619954 95.634 -55.654 52.870
619955 95.668 -55.667 52.804
619956 95.700 -55.679 52.741
619957 95.730 -55.690 52.680
619958 95.763 -55.702 52.618
619959 95.798 -55.717 52.552
619960 95.832 -55.732 52.484
619961 95.867 -55.746 52.417
619962 95.901 -55.761 52.350
619963 95.936 -55.775 52.285
619964 95.970 -55.790 52.222
619965 96.005 -55.805 52.160
619966 96.039 -55.819 52.098
619967 96.074 -55.834 52.038
619968 96.108 -55.849 51.978
619969 96.143 -55.862 51.918
619970 96.177 -55.877 51.858
619971 96.210 -55.890 51.798
619972 96.245 -55.904 51.736
619973 96.278 -55.919 51.677
619974 96.313 -55.932 51.619
619975 96.347 -55.946 51.560
619976 96.381 -55.959 51.501
619977 96.413 -55.965 51.440
619978 96.444 -55.967 51.376
619979 96.459 -55.968 51.315
619980 96.472 -55.969 51.253
619981 96.487 -55.969 51.192
619982 96.504 -55.969 51.130
619983 96.520 -55.968 51.069
619984 96.537 -55.967 51.008
619985 96.553 -55.966 50.951
619986 96.570 -55.965 50.897
619987 96.587 -55.964 50.845
619988 96.603 -55.963 50.793
619989 96.620 -55.962 50.744
620437 88.916 -52.283 57.285
620438 88.949 -52.303 57.275
620439 88.982 -52.322 57.266
620440 89.015 -52.342 57.257
620441 89.048 -52.361 57.249
620442 89.082 -52.380 57.242
620443 89.114 -52.399 57.235
620444 89.147 -52.418 57.227
620445 89.181 -52.438 57.219
620446 89.213 -52.457 57.212
620447 89.247 -52.477 57.205
620448 89.280 -52.496 57.199
620449 89.313 -52.515 57.193
620450 89.346 -52.534 57.188
620451 89.379 -52.554 57.182
620452 89.412 -52.573 57.176
620453 89.445 -52.593 57.171
620454 89.478 -52.612 57.167
620455 89.512 -52.631 57.163
620456 89.544 -52.650 57.159
620457 89.578 -52.670 57.157
620458 89.611 -52.689 57.154
620459 89.644 -52.709 57.151
620460 89.676 -52.722 57.149
620461 89.708 -52.730 57.147
620462 89.741 -52.738 57.144
620463 89.773 -52.746 57.143
620464 89.806 -52.754 57.141
620465 89.838 -52.762 57.139
620466 89.867 -52.772 57.137
620467 89.883 -52.794 57.137
620468 89.899 -52.817 57.135
620469 89.915 -52.839 57.134
620470 89.932 -52.861 57.133
620471 89.948 -52.883 57.131
620472 89.967 -52.901 57.128
620473 90.000 -52.901 57.126
620474 90.031 -52.901 57.124
620475 90.063 -52.901 57.121
620476 90.096 -52.900 57.118
620477 90.127 -52.900 57.115
620478 90.160 -52.900 57.111
620479 90.192 -52.900 57.105
620480 90.218 -52.921 57.100
620481 90.250 -52.941 57.095
620482 90.286 -52.963 57.090
620483 90.320 -52.983 57.084
620484 90.356 -53.004 57.077
620485 90.391 -53.024 57.071
620486 90.426 -53.045 57.065
620487 90.461 -53.066 57.060
620488 90.497 -53.086 57.057
620489 90.531 -53.107 57.055
620490 90.567 -53.127 57.048
620491 90.607 -53.125 57.030
620492 90.648 -53.117 56.998
620493 90.689 -53.111 56.968
620494 90.729 -53.105 56.938
620495 90.764 -53.107 56.906
620496 90.800 -53.109 56.872
620497 90.835 -53.112 56.837
620498 90.864 -53.132 56.802
620499 90.897 -53.152 56.769
620500 90.931 -53.172 56.737
620501 90.964 -53.192 56.704
620502 90.993 -53.212 56.671
620503 91.019 -53.233 56.640
620504 91.046 -53.254 56.610
620505 91.073 -53.274 56.580
620506 91.100 -53.295 56.551
620507 91.127 -53.316 56.521
620508 91.154 -53.336 56.490
620509 91.180 -53.358 56.460
620513 91.288 -53.440 56.015
620514 91.315 -53.462 55.984
620515 91.342 -53.482 55.954
620516 91.368 -53.502 55.925
620517 91.401 -53.527 55.897
620518 91.434 -53.551 55.871
620519 91.467 -53.576 55.845
620520 91.501 -53.600 55.819
620521 91.534 -53.625 55.794
620522 91.567 -53.650 55.769
620523 91.601 -53.674 55.747
620524 91.634 -53.698 55.724
620525 91.667 -53.723 55.702
620526 91.701 -53.747 55.680
620527 91.734 -53.772 55.658
620528 91.768 -53.797 55.637
620529 91.801 -53.821 55.616
620530 91.841 -53.836 55.595
620531 91.881 -53.849 55.575
620532 91.922 -53.863 55.555
620533 91.962 -53.876 55.535
620534 92.003 -53.890 55.515
620535 92.044 -53.903 55.497
620536 92.085 -53.916 55.477
620537 92.126 -53.930 55.459
620551 92.696 -54.119 55.298
620552 92.736 -54.132 55.297
620553 92.775 -54.145 55.296
620554 92.813 -54.157 55.297
620555 92.852 -54.170 55.300
620556 92.890 -54.182 55.305
620557 92.929 -54.195 55.311
620558 92.967 -54.207 55.317
620559 93.006 -54.220 55.324
620560 93.044 -54.232 55.336
620561 93.082 -54.245 55.353
620562 93.115 -54.263 55.364
620563 93.138 -54.292 55.369
620564 93.155 -54.314 55.367
620565 93.155 -54.317 55.359
620566 93.156 -54.320 55.342
620567 93.156 -54.323 55.325
620568 93.167 -54.345 55.306
620569 93.177 -54.366 55.291
620570 93.188 -54.388 55.281
620571 93.208 -54.411 55.276
620572 93.229 -54.435 55.277
620573 93.250 -54.458 55.284
620574 93.271 -54.482 55.296
620575 93.298 -54.499 55.310
620576 93.339 -54.505 55.323
620577 93.380 -54.510 55.335
620578 93.414 -54.518 55.346
620579 93.436 -54.531 55.353
620580 93.459 -54.545 55.363
620581 93.481 -54.558 55.377
620582 93.503 -54.571 55.390
620583 93.525 -54.585 55.397
620584 93.547 -54.598 55.400
620585 93.569 -54.612 55.392
620586 93.591 -54.624 55.385
620587 93.614 -54.638 55.379
620588 93.636 -54.651 55.376
620589 93.658 -54.665 55.378
620590 93.680 -54.678 55.388
620591 93.702 -54.691 55.391
620592 93.724 -54.705 55.386
620593 93.747 -54.718 55.372
620594 93.769 -54.731 55.348
620595 93.793 -54.745 55.317
620596 93.818 -54.759 55.281
620597 93.843 -54.773 55.248
620598 93.868 -54.787 55.228
620599 93.894 -54.800 55.212
620600 93.917 -54.818 55.198
620601 93.937 -54.842 55.181
620602 93.958 -54.866 55.160
620603 93.980 -54.886 55.130
620604 94.004 -54.903 55.102
620605 94.029 -54.919 55.094
620606 94.053 -54.936 55.094
620607 94.077 -54.953 55.095
620608 94.102 -54.969 55.098
620609 94.126 -54.985 55.098
620610 94.150 -55.002 55.096
620611 94.175 -55.019 55.089
620612 94.199 -55.035 55.077
620613 94.223 -55.052 55.059
620614 94.249 -55.068 55.035
620615 94.282 -55.080 55.003
620616 94.315 -55.091 54.968
620617 94.348 -55.104 54.929
620618 94.384 -55.116 54.890
620619 94.421 -55.127 54.850
620620 94.459 -55.138 54.807
620621 94.496 -55.151 54.759
620622 94.534 -55.162 54.713
620623 94.571 -55.173 54.669
620624 94.609 -55.185 54.628
620625 94.646 -55.199 54.588
620626 94.684 -55.213 54.546
620627 94.721 -55.226 54.501
620628 94.758 -55.240 54.454
620629 94.795 -55.254 54.406
620630 94.833 -55.268 54.358
620631 94.871 -55.282 54.310
620632 94.908 -55.295 54.262
620633 94.944 -55.311 54.213
620634 94.976 -55.328 54.165
620635 95.008 -55.346 54.115
620636 95.041 -55.363 54.062
620637 95.071 -55.378 54.006
620638 95.101 -55.393 53.947
620639 95.132 -55.408 53.885
620640 95.162 -55.423 53.822
620641 95.193 -55.438 53.758
620642 95.225 -55.452 53.693
620643 95.257 -55.464 53.628
620644 95.289 -55.478 53.566
620645 95.321 -55.492 53.506
620646 95.353 -55.505 53.445
620647 95.385 -55.519 53.383
620648 95.416 -55.533 53.322
620649 95.449 -55.546 53.263
620650 95.481 -55.559 53.202
620651 95.515 -55.571 53.136
620652 95.548 -55.584 53.067
620653 95.581 -55.597 52.997
620654 95.614 -55.610 52.927
620655 95.647 -55.623 52.860
620656 95.680 -55.636 52.793
620657 95.710 -55.646 52.729
620658 95.741 -55.657 52.666
620659 95.773 -55.669 52.604
620660 95.808 -55.683 52.540
620661 95.842 -55.698 52.476
620662 95.877 -55.712 52.412
620663 95.911 -55.727 52.349
620664 95.946 -55.742 52.285
620665 95.980 -55.756 52.224
620666 96.015 -55.771 52.163
620667 96.049 -55.786 52.102
620668 96.084 -55.800 52.042
620669 96.118 -55.815 51.982
620670 96.153 -55.830 51.922
620671 96.187 -55.844 51.863
620672 96.222 -55.859 51.802
620673 96.256 -55.874 51.739
620674 96.291 -55.888 51.681
620675 96.325 -55.903 51.622
620676 96.359 -55.916 51.563
620677 96.393 -55.930 51.504
620678 96.427 -55.943 51.444
620679 96.454 -55.945 51.380
620680 96.468 -55.946 51.318
620681 96.484 -55.945 51.256
620682 96.501 -55.944 51.194
620683 96.518 -55.944 51.131
620684 96.534 -55.943 51.069
620685 96.551 -55.941 51.008
620686 96.568 -55.940 50.950
620687 96.584 -55.940 50.895
620688 96.601 -55.939 50.840
620689 96.617 -55.938 50.787
620690 96.634 -55.937 50.736
621138 88.936 -52.250 57.277
621139 88.969 -52.270 57.267
621140 89.002 -52.289 57.258
621141 89.035 -52.308 57.249
621142 89.068 -52.327 57.240
621143 89.101 -52.347 57.232
621144 89.134 -52.366 57.225
621145 89.168 -52.386 57.217
621146 89.200 -52.405 57.209
621147 89.234 -52.424 57.201
621148 89.267 -52.443 57.194
621149 89.300 -52.463 57.188
621150 89.333 -52.482 57.183
621151 89.366 -52.501 57.178
621152 89.399 -52.521 57.173
621153 89.432 -52.540 57.167
621154 89.465 -52.559 57.161
621155 89.498 -52.578 57.158
621156 89.531 -52.598 57.154
621157 89.564 -52.617 57.151
621158 89.598 -52.637 57.148
621159 89.630 -52.656 57.145
621160 89.664 -52.675 57.142
621162 89.729 -52.697 57.139
621163 89.761 -52.705 57.137
621164 89.793 -52.713 57.136
621165 89.825 -52.721 57.135
621166 89.858 -52.728 57.133
621167 89.890 -52.737 57.132
621168 89.914 -52.751 57.132
621169 89.931 -52.774 57.131
621170 89.947 -52.796 57.130
621171 89.963 -52.818 57.129
621172 89.979 -52.840 57.128
621173 89.995 -52.862 57.126
621174 90.023 -52.868 57.125
621175 90.055 -52.867 57.124
621176 90.087 -52.867 57.123
621177 90.119 -52.867 57.121
621178 90.152 -52.867 57.119
621179 90.183 -52.867 57.116
621180 90.216 -52.867 57.111
621181 90.243 -52.884 57.107
621182 90.268 -52.906 57.103
621183 90.302 -52.927 57.100
621184 90.336 -52.948 57.097
621185 90.372 -52.968 57.091
621186 90.407 -52.988 57.085
621187 90.442 -53.010 57.082
621188 90.477 -53.030 57.078
621189 90.513 -53.051 57.077
621190 90.547 -53.071 57.078
621191 90.583 -53.092 57.075
621192 90.623 -53.089 57.053
621193 90.664 -53.082 57.024
621194 90.705 -53.075 56.994
621195 90.746 -53.068 56.964
621196 90.787 -53.061 56.933
621197 90.823 -53.062 56.900
621198 90.855 -53.076 56.867
621199 90.888 -53.095 56.834
621200 90.916 -53.116 56.801
621201 90.943 -53.137 56.770
621202 90.970 -53.158 56.739
621203 90.997 -53.178 56.706
621204 91.024 -53.199 56.675
621205 91.050 -53.220 56.645
621206 91.077 -53.240 56.614
621207 91.104 -53.261 56.585
621208 91.131 -53.282 56.556
621209 91.158 -53.302 56.525
621210 91.185 -53.324 56.494
621211 91.212 -53.344 56.465
621212 91.238 -53.365 56.436
621214 91.292 -53.406 56.043
621215 91.319 -53.427 56.011
621216 91.348 -53.449 55.980
621217 91.381 -53.474 55.950
621218 91.415 -53.498 55.922
621219 91.448 -53.523 55.895
621220 91.482 -53.547 55.868
621221 91.515 -53.571 55.841
621222 91.548 -53.596 55.816
621223 91.581 -53.621 55.790
621224 91.614 -53.645 55.767
621225 91.648 -53.670 55.743
621226 91.681 -53.694 55.720
621227 91.715 -53.719 55.697
621228 91.748 -53.744 55.674
621229 91.781 -53.768 55.652
621230 91.814 -53.792 55.630
621231 91.855 -53.807 55.608
621232 91.895 -53.820 55.587
621233 91.936 -53.834 55.566
621234 91.977 -53.847 55.545
621235 92.017 -53.861 55.525
621236 92.058 -53.874 55.504
621237 92.099 -53.888 55.484
621238 92.140 -53.901 55.464
621252 92.710 -54.090 55.296
621253 92.751 -54.104 55.294
621254 92.792 -54.117 55.294
621255 92.833 -54.131 55.294
621256 92.872 -54.144 55.297
621257 92.910 -54.156 55.301
621258 92.948 -54.168 55.307
621259 92.987 -54.181 55.312
621260 93.025 -54.193 55.318
621261 93.064 -54.206 55.329
621262 93.102 -54.218 55.345
621263 93.138 -54.234 55.355
621264 93.161 -54.263 55.358
621265 93.171 -54.276 55.354
621266 93.171 -54.279 55.342
621267 93.172 -54.284 55.323
621268 93.183 -54.306 55.304
621269 93.193 -54.327 55.286
621270 93.204 -54.349 55.272
621271 93.214 -54.370 55.263
621272 93.224 -54.392 55.260
621273 93.234 -54.414 55.262
621274 93.254 -54.437 55.270
621275 93.275 -54.461 55.283
621276 93.312 -54.469 55.296
621277 93.354 -54.474 55.308
621278 93.395 -54.480 55.320
621279 93.426 -54.489 55.329
621280 93.448 -54.502 55.336
621281 93.470 -54.515 55.344
621282 93.492 -54.529 55.355
621283 93.514 -54.543 55.365
621284 93.537 -54.555 55.371
621285 93.559 -54.569 55.371
621286 93.581 -54.582 55.364
621287 93.603 -54.596 55.361
621288 93.625 -54.609 55.362
621289 93.647 -54.622 55.365
621290 93.670 -54.636 55.373
621291 93.692 -54.649 55.380
621292 93.714 -54.662 55.384
621293 93.736 -54.675 55.378
621294 93.758 -54.689 55.361
621295 93.780 -54.702 55.334
621296 93.802 -54.715 55.300
621297 93.825 -54.729 55.261
621298 93.850 -54.743 55.228
621299 93.875 -54.756 55.210
621300 93.900 -54.770 55.197
621301 93.925 -54.784 55.183
621302 93.945 -54.809 55.166
621303 93.966 -54.833 55.144
621304 93.990 -54.850 55.112
621305 94.014 -54.866 55.090
621306 94.038 -54.883 55.084
621307 94.063 -54.900 55.084
621308 94.087 -54.916 55.085
621309 94.111 -54.933 55.088
621310 94.135 -54.950 55.087
621311 94.160 -54.966 55.084
621312 94.184 -54.982 55.076
621313 94.208 -54.999 55.064
621314 94.232 -55.016 55.047
621315 94.258 -55.032 55.026
621316 94.291 -55.044 54.998
621317 94.324 -55.056 54.963
621318 94.360 -55.067 54.927
621319 94.398 -55.079 54.890
621320 94.435 -55.091 54.852
621321 94.473 -55.103 54.812
621322 94.510 -55.114 54.767
621323 94.548 -55.126 54.718
621324 94.585 -55.138 54.669
621325 94.623 -55.149 54.622
621326 94.660 -55.163 54.577
621327 94.698 -55.176 54.532
621328 94.735 -55.190 54.484
621329 94.772 -55.204 54.434
621330 94.810 -55.217 54.385
621331 94.847 -55.231 54.335
621332 94.884 -55.245 54.285
621333 94.920 -55.260 54.235
621334 94.953 -55.278 54.184
621335 94.985 -55.295 54.137
621336 95.018 -55.314 54.088
621337 95.051 -55.332 54.037
621338 95.082 -55.348 53.983
621339 95.113 -55.363 53.925
621340 95.143 -55.378 53.864
621341 95.174 -55.392 53.801
621342 95.206 -55.406 53.736
621343 95.238 -55.419 53.670
621344 95.270 -55.433 53.605
621345 95.302 -55.446 53.540
621346 95.334 -55.460 53.477
621347 95.366 -55.474 53.415
621348 95.397 -55.487 53.353
621349 95.429 -55.501 53.293
621350 95.461 -55.514 53.236
621351 95.494 -55.527 53.176
621352 95.527 -55.540 53.115
621353 95.560 -55.553 53.052
621354 95.593 -55.565 52.987
621355 95.626 -55.578 52.918
621356 95.659 -55.591 52.851
621357 95.691 -55.603 52.784
621358 95.722 -55.614 52.719
621359 95.752 -55.624 52.656
621360 95.783 -55.635 52.594
621361 95.818 -55.650 52.533
621362 95.852 -55.665 52.470
621363 95.887 -55.679 52.409
621364 95.921 -55.693 52.349
621365 95.956 -55.709 52.287
621366 95.990 -55.723 52.227
621367 96.025 -55.737 52.166
621368 96.059 -55.752 52.107
621369 96.094 -55.767 52.047
621370 96.128 -55.781 51.986
621371 96.163 -55.796 51.926
621372 96.197 -55.811 51.866
621373 96.232 -55.825 51.806
621374 96.266 -55.840 51.744
621375 96.301 -55.855 51.685
621376 96.336 -55.869 51.626
621377 96.370 -55.884 51.566
621378 96.405 -55.899 51.507
621379 96.439 -55.913 51.448
621380 96.466 -55.922 51.384
621381 96.482 -55.921 51.322
621382 96.499 -55.920 51.259
621383 96.516 -55.919 51.195
621384 96.532 -55.919 51.132
621385 96.549 -55.917 51.069
621386 96.566 -55.916 51.007
621387 96.582 -55.915 50.948
621388 96.599 -55.915 50.891
621389 96.615 -55.914 50.836
621390 96.632 -55.913 50.782
621391 96.649 -55.912 50.729
621839 88.956 -52.217 57.269
621840 88.989 -52.236 57.259
621841 89.022 -52.255 57.249
621842 89.055 -52.275 57.240
621843 89.088 -52.294 57.230
621844 89.121 -52.314 57.223
621845 89.155 -52.333 57.215
621846 89.187 -52.352 57.207
621847 89.221 -52.371 57.199
621848 89.254 -52.391 57.191
621849 89.287 -52.410 57.183
621850 89.320 -52.430 57.178
621851 89.353 -52.449 57.173
621852 89.386 -52.468 57.168
621853 89.419 -52.487 57.163
621854 89.452 -52.507 57.157
621855 89.485 -52.526 57.153
621856 89.518 -52.546 57.148
621857 89.551 -52.565 57.145
621858 89.585 -52.584 57.142
621859 89.617 -52.603 57.139
621860 89.651 -52.622 57.137
621864 89.781 -52.672 57.129
621865 89.813 -52.679 57.129
621866 89.845 -52.687 57.129
621867 89.878 -52.695 57.128
621868 89.910 -52.703 57.127
621869 89.942 -52.712 57.127
621870 89.962 -52.731 57.127
621871 89.978 -52.753 57.127
621872 89.994 -52.775 57.127
621873 90.010 -52.797 57.127
621874 90.026 -52.819 57.126
621875 90.047 -52.835 57.126
621876 90.079 -52.835 57.127
621877 90.111 -52.835 57.128
621878 90.143 -52.835 57.127
621879 90.175 -52.835 57.125
621880 90.207 -52.835 57.122
621881 90.239 -52.835 57.119
621882 90.267 -52.847 57.116
621883 90.293 -52.869 57.113
621884 90.318 -52.891 57.111
621885 90.353 -52.912 57.110
621886 90.388 -52.933 57.104
621887 90.423 -52.954 57.101
621888 90.458 -52.974 57.099
621889 90.494 -52.995 57.096
621890 90.529 -53.015 57.095
621891 90.564 -53.036 57.095
621892 90.599 -53.057 57.090
621893 90.639 -53.054 57.069
621894 90.680 -53.047 57.042
621895 90.721 -53.040 57.014
621896 90.762 -53.032 56.986
621897 90.803 -53.026 56.958
621898 90.844 -53.022 56.927
621899 90.867 -53.040 56.895
621900 90.893 -53.061 56.865
621901 90.921 -53.082 56.834
621902 90.947 -53.103 56.804
621903 90.974 -53.123 56.774
621904 91.001 -53.144 56.742
621905 91.028 -53.165 56.711
621906 91.054 -53.186 56.681
621907 91.082 -53.206 56.651
621908 91.108 -53.227 56.621
621909 91.135 -53.248 56.592
621910 91.162 -53.268 56.561
621911 91.189 -53.289 56.530
621912 91.216 -53.310 56.499
621913 91.243 -53.331 56.470
621914 91.270 -53.352 56.441
621953 92.724 -54.061 55.294
621954 92.765 -54.075 55.292
621955 92.806 -54.088 55.291
621956 92.847 -54.102 55.290
621957 92.887 -54.115 55.292
621958 92.928 -54.129 55.296
621959 92.969 -54.142 55.300
621960 93.007 -54.154 55.304
621961 93.046 -54.167 55.311
621962 93.084 -54.179 55.322
621963 93.123 -54.192 55.336
621964 93.161 -54.204 55.344
621965 93.184 -54.233 55.344
621966 93.187 -54.239 55.338
621967 93.189 -54.245 55.322
621968 93.199 -54.267 55.300
621969 93.209 -54.289 55.280
621970 93.220 -54.311 55.264
621971 93.230 -54.332 55.253
621972 93.240 -54.354 55.246
621973 93.251 -54.375 55.243
621974 93.261 -54.397 55.246
621975 93.271 -54.418 55.255
621976 93.287 -54.434 55.268
621977 93.327 -54.439 55.280
621978 93.368 -54.444 55.291
621979 93.409 -54.449 55.303
621980 93.437 -54.460 55.310
621981 93.460 -54.473 55.317
621982 93.482 -54.486 55.324
621983 93.504 -54.500 55.332
621984 93.526 -54.513 55.340
621985 93.548 -54.527 55.345
621986 93.570 -54.540 55.344
621987 93.592 -54.553 55.340
621988 93.615 -54.566 55.342
621989 93.637 -54.580 55.346
621990 93.659 -54.593 55.352
621991 93.681 -54.606 55.360
621992 93.703 -54.620 55.365
621993 93.725 -54.633 55.363
621994 93.748 -54.646 55.354
621995 93.770 -54.659 55.335
621996 93.792 -54.673 55.308
621997 93.814 -54.687 55.275
621998 93.836 -54.699 55.238
621999 93.858 -54.713 55.207
622000 93.882 -54.727 55.194
622001 93.907 -54.740 55.184
622002 93.931 -54.754 55.170
622003 93.953 -54.775 55.151
622004 93.975 -54.797 55.125
622005 93.999 -54.813 55.095
622006 94.024 -54.830 55.079
622007 94.048 -54.847 55.075
622008 94.072 -54.863 55.077
622009 94.097 -54.880 55.078
622010 94.121 -54.897 55.080
622011 94.145 -54.913 55.077
622012 94.170 -54.930 55.070
622013 94.194 -54.947 55.060
622014 94.218 -54.963 55.047
622015 94.243 -54.979 55.031
622016 94.267 -54.996 55.012
622017 94.300 -55.008 54.983
622018 94.337 -55.019 54.950
622019 94.374 -55.032 54.917
622020 94.412 -55.043 54.884
622021 94.449 -55.055 54.849
622022 94.487 -55.066 54.809
622023 94.524 -55.078 54.765
622024 94.562 -55.090 54.717
622025 94.599 -55.101 54.667
622026 94.637 -55.113 54.616
622027 94.674 -55.126 54.566
622028 94.711 -55.139 54.516
622029 94.749 -55.154 54.462
622030 94.786 -55.167 54.411
622031 94.824 -55.181 54.360
622032 94.861 -55.195 54.309
622033 94.897 -55.209 54.258
622034 94.930 -55.227 54.205
622035 94.963 -55.245 54.151
622036 94.995 -55.263 54.105
622037 95.028 -55.281 54.056
622038 95.060 -55.298 54.007
622039 95.093 -55.317 53.956
622040 95.124 -55.332 53.900
622041 95.155 -55.347 53.840
622042 95.187 -55.361 53.777
622043 95.219 -55.373 53.712
622044 95.251 -55.387 53.645
622045 95.282 -55.401 53.579
622046 95.315 -55.414 53.513
622047 95.346 -55.428 53.448
622048 95.378 -55.442 53.386
622049 95.410 -55.455 53.325
622050 95.442 -55.469 53.265
622051 95.474 -55.483 53.207
622052 95.506 -55.495 53.149
622053 95.539 -55.508 53.089
622054 95.572 -55.521 53.031
622055 95.605 -55.534 52.972
622056 95.638 -55.547 52.907
622057 95.671 -55.559 52.842
622058 95.702 -55.570 52.777
622059 95.732 -55.580 52.714
622060 95.763 -55.591 52.652
622061 95.793 -55.602 52.592
622062 95.828 -55.616 52.532
622063 95.862 -55.630 52.471
622064 95.897 -55.646 52.411
622065 95.931 -55.660 52.351
622066 95.966 -55.674 52.291
622067 96.001 -55.690 52.232
622068 96.035 -55.704 52.172
622069 96.070 -55.718 52.112
622070 96.104 -55.734 52.052
622071 96.138 -55.748 51.991
622072 96.173 -55.762 51.930
622073 96.207 -55.777 51.870
622074 96.242 -55.792 51.809
622075 96.276 -55.806 51.749
622076 96.311 -55.821 51.690
622077 96.346 -55.836 51.630
622078 96.380 -55.850 51.570
622079 96.413 -55.862 51.511
622080 96.444 -55.873 51.449
622081 96.479 -55.894 51.388
622082 96.497 -55.896 51.325
622083 96.513 -55.895 51.261
622084 96.530 -55.893 51.197
622085 96.546 -55.893 51.132
622086 96.563 -55.892 51.068
622087 96.580 -55.891 51.005
622088 96.596 -55.890 50.946
622089 96.613 -55.890 50.888
622090 96.629 -55.889 50.832
622091 96.646 -55.888 50.777
622092 96.663 -55.887 50.723
622541 89.009 -52.203 57.251
622542 89.042 -52.223 57.241
622543 89.075 -52.242 57.231
622544 89.108 -52.261 57.220
622545 89.141 -52.280 57.213
622546 89.175 -52.300 57.204
622547 89.207 -52.319 57.197
622548 89.241 -52.338 57.190
622549 89.274 -52.358 57.181
622550 89.307 -52.377 57.173
622551 89.340 -52.396 57.169
622552 89.373 -52.415 57.164
622553 89.406 -52.435 57.158
622554 89.439 -52.454 57.153
622555 89.472 -52.474 57.148
622556 89.505 -52.493 57.143
622557 89.538 -52.512 57.139
622558 89.571 -52.531 57.136
622559 89.605 -52.551 57.134
622565 89.801 -52.638 57.124
622566 89.833 -52.646 57.123
622567 89.866 -52.654 57.122
622568 89.898 -52.662 57.122
622569 89.930 -52.670 57.122
622570 89.962 -52.678 57.122
622571 89.993 -52.687 57.123
622572 90.009 -52.709 57.124
622573 90.025 -52.731 57.125
622574 90.041 -52.753 57.126
622575 90.057 -52.775 57.126
622576 90.074 -52.798 57.128
622577 90.103 -52.802 57.131
622578 90.134 -52.802 57.134
622579 90.167 -52.802 57.134
622580 90.199 -52.802 57.133
622581 90.231 -52.802 57.131
622582 90.263 -52.802 57.129
622583 90.293 -52.810 57.127
622584 90.318 -52.832 57.126
622585 90.343 -52.854 57.124
622586 90.369 -52.877 57.123
622587 90.404 -52.897 57.118
622588 90.439 -52.918 57.117
622589 90.474 -52.938 57.116
622590 90.510 -52.959 57.114
622591 90.544 -52.980 57.112
622592 90.580 -53.001 57.110
622593 90.615 -53.021 57.101
622594 90.655 -53.019 57.082
622595 90.696 -53.011 57.058
622596 90.737 -53.004 57.032
622597 90.778 -52.998 57.007
622598 90.819 -52.990 56.982
622599 90.860 -52.995 56.953
622600 90.888 -53.007 56.924
622601 90.898 -53.027 56.894
622602 90.925 -53.048 56.866
622603 90.951 -53.069 56.839
622604 90.978 -53.089 56.810
622605 91.005 -53.110 56.779
622606 91.032 -53.131 56.748
622607 91.059 -53.151 56.719
622608 91.086 -53.172 56.689
622609 91.112 -53.193 56.659
622610 91.140 -53.214 56.629
622611 91.166 -53.234 56.597
622612 91.193 -53.255 56.566
622613 91.220 -53.276 56.534
622614 91.247 -53.297 56.504
622615 91.276 -53.319 56.475
622616 91.309 -53.343 56.445
622617 91.343 -53.368 56.413
622618 91.376 -53.392 56.383
622619 91.409 -53.417 56.355
622620 91.442 -53.441 56.326
622621 91.476 -53.466 56.298
622622 91.509 -53.490 56.270
622623 91.542 -53.515 56.242
622624 91.576 -53.540 56.215
622625 91.609 -53.564 56.189
622626 91.642 -53.588 56.163
622627 91.676 -53.613 56.137
622628 91.709 -53.637 56.111
622629 91.742 -53.662 56.086
622630 91.776 -53.687 56.060
622631 91.809 -53.711 56.033
622632 91.843 -53.735 56.007
622633 91.883 -53.749 55.981
622634 91.924 -53.762 55.955
622635 91.964 -53.776 55.929
622636 92.005 -53.790 55.902
622637 92.046 -53.803 55.876
622638 92.087 -53.816 55.850
622654 92.738 -54.032 55.292
622655 92.779 -54.046 55.289
622656 92.820 -54.060 55.286
622657 92.861 -54.073 55.285
622658 92.902 -54.086 55.286
622659 92.942 -54.100 55.288
622660 92.983 -54.113 55.291
622661 93.024 -54.127 55.295
622662 93.064 -54.140 55.301
622663 93.104 -54.153 55.313
622664 93.142 -54.166 55.324
622665 93.181 -54.178 55.329
622666 93.202 -54.198 55.327
622667 93.205 -54.207 55.316
622668 93.215 -54.229 55.298
622669 93.226 -54.250 55.276
622670 93.236 -54.272 55.258
622671 93.247 -54.293 55.244
622672 93.256 -54.314 55.235
622673 93.265 -54.333 55.229
622674 93.274 -54.353 55.227
622675 93.283 -54.373 55.230
622676 93.292 -54.393 55.237
622677 93.312 -54.407 55.248
622678 93.350 -54.411 55.261
622679 93.389 -54.415 55.272
622680 93.427 -54.420 55.284
622681 93.452 -54.431 55.291
622682 93.474 -54.444 55.296
622683 93.495 -54.458 55.302
622684 93.517 -54.471 55.308
622685 93.538 -54.484 55.315
622686 93.560 -54.497 55.319
622687 93.582 -54.511 55.318
622688 93.604 -54.524 55.318
622689 93.626 -54.537 55.322
622690 93.648 -54.551 55.329
622691 93.671 -54.564 55.336
622692 93.693 -54.577 55.342
622693 93.715 -54.590 55.341
622694 93.737 -54.604 55.334
622695 93.759 -54.617 55.320
622696 93.781 -54.630 55.300
622697 93.803 -54.644 55.273
622698 93.826 -54.657 55.242
622699 93.848 -54.671 55.212
622700 93.870 -54.684 55.191
622701 93.892 -54.697 55.180
622702 93.914 -54.710 55.169
622703 93.938 -54.724 55.153
622704 93.961 -54.743 55.132
622705 93.985 -54.760 55.105
622706 94.009 -54.777 55.079
622707 94.033 -54.793 55.072
622708 94.058 -54.810 55.073
622709 94.082 -54.827 55.075
622710 94.106 -54.844 55.076
622711 94.131 -54.860 55.076
622712 94.155 -54.877 55.067
622713 94.179 -54.894 55.054
622714 94.203 -54.910 55.041
622715 94.228 -54.927 55.027
622716 94.252 -54.943 55.013
622717 94.277 -54.960 54.994
622718 94.313 -54.972 54.963
622719 94.351 -54.984 54.933
622720 94.388 -54.995 54.902
622721 94.426 -55.007 54.871
622722 94.463 -55.019 54.838
622723 94.501 -55.030 54.800
622724 94.538 -55.042 54.758
622725 94.576 -55.054 54.711
622726 94.613 -55.065 54.660
622727 94.651 -55.077 54.607
622728 94.688 -55.089 54.554
622729 94.725 -55.103 54.501
622730 94.763 -55.116 54.444
622731 94.800 -55.130 54.389
622732 94.837 -55.144 54.334
622733 94.874 -55.159 54.281
622734 94.907 -55.176 54.227
622735 94.939 -55.195 54.173
622736 94.972 -55.212 54.119
622737 95.004 -55.230 54.072
622738 95.037 -55.248 54.023
622739 95.069 -55.266 53.975
622740 95.103 -55.284 53.926
622741 95.136 -55.300 53.873
622742 95.168 -55.314 53.814
622743 95.200 -55.328 53.751
622744 95.232 -55.342 53.684
622745 95.264 -55.355 53.617
622746 95.296 -55.369 53.551
622747 95.327 -55.383 53.487
622748 95.359 -55.396 53.422
622749 95.391 -55.410 53.360
622750 95.423 -55.423 53.299
622751 95.455 -55.437 53.237
622752 95.487 -55.451 53.178
622753 95.519 -55.464 53.120
622754 95.551 -55.477 53.063
622755 95.585 -55.489 53.008
622756 95.618 -55.502 52.954
622757 95.651 -55.515 52.897
622758 95.682 -55.527 52.835
622759 95.712 -55.537 52.773
622760 95.743 -55.548 52.713
622761 95.773 -55.558 52.653
622762 95.804 -55.569 52.594
622763 95.838 -55.583 52.535
622764 95.872 -55.597 52.475
622765 95.907 -55.611 52.416
622766 95.941 -55.627 52.357
622767 95.976 -55.641 52.298
622768 96.011 -55.655 52.239
622769 96.045 -55.671 52.179
622770 96.080 -55.685 52.119
622771 96.114 -55.699 52.059
622772 96.149 -55.714 51.999
622773 96.183 -55.729 51.938
622774 96.218 -55.743 51.877
622775 96.252 -55.758 51.815
622776 96.287 -55.773 51.755
622777 96.321 -55.787 51.695
622778 96.355 -55.801 51.634
622779 96.386 -55.811 51.574
622780 96.417 -55.821 51.514
622781 96.449 -55.833 51.452
622782 96.484 -55.857 51.389
622783 96.511 -55.871 51.327
622784 96.528 -55.869 51.263
622785 96.544 -55.868 51.199
622786 96.561 -55.868 51.135
622787 96.578 -55.867 51.070
622788 96.594 -55.866 51.006
622789 96.611 -55.865 50.946
622790 96.627 -55.865 50.887
622791 96.644 -55.864 50.830
622792 96.661 -55.863 50.774
622793 96.677 -55.862 50.720
623194 87.519 -51.108 56.283
623195 87.550 -51.131 56.295
623196 87.581 -51.155 56.306
623197 87.612 -51.178 56.317
623198 87.643 -51.202 56.322
623199 87.674 -51.224 56.326
623200 87.705 -51.248 56.328
623201 87.736 -51.271 56.328
623202 87.767 -51.295 56.327
623203 87.798 -51.318 56.325
623204 87.829 -51.342 56.319
623243 89.062 -52.189 57.232
623244 89.095 -52.208 57.222
623245 89.128 -52.228 57.211
623246 89.162 -52.247 57.203
623247 89.194 -52.267 57.195
623248 89.228 -52.286 57.186
623249 89.261 -52.305 57.179
623250 89.294 -52.324 57.171
623251 89.327 -52.344 57.163
623252 89.360 -52.363 57.158
623253 89.393 -52.383 57.154
623254 89.426 -52.402 57.148
623255 89.459 -52.421 57.143
623256 89.492 -52.440 57.139
623257 89.525 -52.459 57.133
623258 89.558 -52.479 57.131
623267 89.853 -52.612 57.118
623268 89.886 -52.621 57.117
623269 89.918 -52.628 57.118
623270 89.950 -52.637 57.118
623271 89.982 -52.644 57.119
623272 90.015 -52.653 57.120
623273 90.040 -52.666 57.122
623274 90.056 -52.688 57.123
623275 90.073 -52.710 57.126
623276 90.089 -52.733 57.127
623277 90.105 -52.755 57.131
623278 90.126 -52.769 57.135
623279 90.159 -52.769 57.138
623280 90.190 -52.769 57.140
623281 90.223 -52.769 57.141
623282 90.255 -52.769 57.140
623283 90.287 -52.769 57.140
623284 90.318 -52.772 57.139
623285 90.343 -52.794 57.139
623286 90.368 -52.817 57.138
623287 90.393 -52.840 57.137
623288 90.420 -52.862 57.134
623289 90.455 -52.882 57.134
623290 90.490 -52.904 57.134
623291 90.526 -52.924 57.133
623292 90.561 -52.944 57.130
623293 90.596 -52.965 57.126
623294 90.631 -52.985 57.111
623295 90.671 -52.983 57.092
623296 90.712 -52.976 57.072
623297 90.753 -52.969 57.049
623298 90.794 -52.962 57.027
623299 90.835 -52.962 57.005
623300 90.875 -52.970 56.978
623301 90.916 -52.977 56.951
623302 90.927 -52.994 56.924
623303 90.932 -53.013 56.898
623304 90.956 -53.035 56.872
623305 90.982 -53.055 56.846
623306 91.009 -53.076 56.816
623307 91.036 -53.097 56.786
623308 91.063 -53.117 56.757
623309 91.090 -53.138 56.727
623310 91.117 -53.159 56.698
623311 91.144 -53.180 56.667
623312 91.170 -53.200 56.635
623313 91.198 -53.221 56.604
623314 91.224 -53.242 56.572
623315 91.256 -53.266 56.541
623316 91.290 -53.290 56.510
623317 91.323 -53.314 56.479
623318 91.356 -53.339 56.446
623319 91.390 -53.364 56.416
623320 91.423 -53.388 56.385
623321 91.456 -53.413 56.355
623322 91.490 -53.437 56.326
623323 91.523 -53.462 56.297
623324 91.556 -53.487 56.268
623325 91.590 -53.511 56.240
623326 91.623 -53.535 56.212
623327 91.656 -53.560 56.185
623328 91.689 -53.584 56.159
623329 91.723 -53.609 56.131
623330 91.756 -53.634 56.104
623331 91.790 -53.658 56.077
623332 91.823 -53.682 56.050
623333 91.856 -53.706 56.022
623334 91.897 -53.720 55.994
623335 91.938 -53.734 55.967
623336 91.979 -53.747 55.939
623337 92.019 -53.761 55.911
623338 92.060 -53.774 55.883
623339 92.101 -53.787 55.856
623340 92.142 -53.801 55.829
623355 92.753 -54.004 55.287
623356 92.793 -54.017 55.283
623357 92.834 -54.031 55.280
623358 92.875 -54.044 55.278
623359 92.916 -54.057 55.278
623360 92.956 -54.071 55.279
623361 92.997 -54.085 55.281
623362 93.038 -54.098 55.284
623363 93.079 -54.111 55.289
623364 93.119 -54.125 55.299
623365 93.160 -54.138 55.308
623366 93.201 -54.152 55.312
623367 93.221 -54.168 55.307
623368 93.231 -54.189 55.292
623369 93.240 -54.208 55.273
623370 93.249 -54.229 55.253
623371 93.258 -54.248 55.237
623372 93.267 -54.268 55.224
623373 93.276 -54.289 55.216
623374 93.285 -54.308 55.212
623375 93.295 -54.328 55.211
623376 93.304 -54.348 55.212
623377 93.313 -54.368 55.219
623378 93.337 -54.380 55.229
623379 93.375 -54.384 55.240
623380 93.413 -54.389 55.252
623381 93.448 -54.395 55.262
623382 93.479 -54.402 55.268
623383 93.501 -54.415 55.273
623384 93.522 -54.429 55.278
623385 93.544 -54.442 55.283
623386 93.565 -54.455 55.289
623387 93.587 -54.468 55.293
623388 93.609 -54.482 55.295
623389 93.630 -54.496 55.298
623390 93.651 -54.508 55.303
623391 93.673 -54.522 55.309
623392 93.694 -54.535 55.316
623393 93.716 -54.549 55.318
623394 93.737 -54.562 55.311
623395 93.759 -54.575 55.298
623396 93.780 -54.588 55.280
623397 93.801 -54.602 55.259
623398 93.823 -54.615 55.235
623399 93.844 -54.628 55.209
623400 93.866 -54.641 55.189
623401 93.888 -54.655 55.177
623402 93.909 -54.668 55.167
623403 93.930 -54.681 55.155
623404 93.952 -54.695 55.136
623405 93.972 -54.707 55.113
623406 93.996 -54.723 55.086
623407 94.021 -54.740 55.067
623408 94.046 -54.756 55.064
623409 94.070 -54.773 55.067
623410 94.095 -54.790 55.070
623411 94.120 -54.806 55.069
623412 94.145 -54.822 55.066
623413 94.170 -54.839 55.052
623414 94.194 -54.856 55.036
623415 94.219 -54.872 55.019
623416 94.244 -54.889 55.003
623417 94.268 -54.906 54.988
623418 94.293 -54.922 54.968
623419 94.328 -54.935 54.939
623420 94.365 -54.947 54.909
623421 94.402 -54.959 54.879
623422 94.439 -54.971 54.850
623423 94.477 -54.982 54.819
623424 94.514 -54.994 54.786
623425 94.552 -55.006 54.746
623426 94.590 -55.017 54.698
623427 94.627 -55.029 54.648
623428 94.665 -55.041 54.595
623429 94.702 -55.053 54.540
623430 94.740 -55.066 54.485
623431 94.777 -55.080 54.427
623432 94.814 -55.094 54.368
623433 94.851 -55.108 54.309
623434 94.884 -55.126 54.252
623435 94.916 -55.144 54.197
623436 94.949 -55.162 54.142
623437 94.981 -55.179 54.087
623438 95.014 -55.198 54.038
623439 95.048 -55.215 53.990
623440 95.083 -55.233 53.941
623441 95.117 -55.251 53.892
623442 95.149 -55.266 53.840
623443 95.181 -55.280 53.782
623444 95.213 -55.295 53.718
623445 95.245 -55.309 53.653
623446 95.276 -55.323 53.587
623447 95.308 -55.337 53.522
623448 95.340 -55.351 53.459
623449 95.372 -55.364 53.396
623450 95.404 -55.378 53.335
623451 95.436 -55.392 53.275
623452 95.468 -55.405 53.216
623453 95.499 -55.418 53.159
623454 95.531 -55.432 53.104
623455 95.564 -55.445 53.050
623456 95.597 -55.458 52.997
623457 95.630 -55.470 52.944
623458 95.662 -55.483 52.888
623459 95.693 -55.493 52.832
623460 95.724 -55.504 52.774
623461 95.754 -55.514 52.716
623462 95.784 -55.525 52.659
623463 95.815 -55.536 52.601
623464 95.848 -55.549 52.542
623465 95.882 -55.564 52.484
623466 95.917 -55.578 52.426
623467 95.951 -55.593 52.367
623468 95.986 -55.608 52.307
623469 96.021 -55.622 52.248
623470 96.055 -55.636 52.188
623471 96.090 -55.652 52.128
623472 96.124 -55.666 52.067
623473 96.159 -55.680 52.007
623474 96.193 -55.695 51.946
623475 96.228 -55.710 51.885
623476 96.262 -55.724 51.824
623477 96.297 -55.739 51.762
623478 96.328 -55.749 51.701
623479 96.359 -55.759 51.640
623480 96.390 -55.769 51.580
623481 96.421 -55.779 51.518
623482 96.454 -55.795 51.456
623483 96.489 -55.819 51.392
623484 96.524 -55.843 51.330
623485 96.542 -55.844 51.266
623486 96.558 -55.843 51.202
623487 96.575 -55.843 51.138
623488 96.592 -55.842 51.073
623489 96.608 -55.841 51.009
623490 96.625 -55.840 50.948
623491 96.642 -55.840 50.889
623492 96.658 -55.839 50.831
623493 96.675 -55.838 50.775
623494 96.691 -55.837 50.721
623893 87.479 -51.032 56.282
623894 87.510 -51.055 56.295
623895 87.541 -51.078 56.307
623896 87.572 -51.102 56.320
623897 87.603 -51.125 56.331
623898 87.634 -51.149 56.342
623899 87.665 -51.172 56.348
623900 87.696 -51.195 56.351
623901 87.727 -51.218 56.353
623902 87.758 -51.242 56.353
623903 87.789 -51.265 56.351
623904 87.820 -51.289 56.347
623905 87.851 -51.312 56.340
623906 87.882 -51.335 56.331
623907 87.916 -51.352 56.321
623945 89.115 -52.175 57.213
623946 89.148 -52.195 57.202
623947 89.181 -52.214 57.193
623948 89.214 -52.233 57.185
623949 89.248 -52.252 57.176
623950 89.280 -52.272 57.168
623951 89.314 -52.291 57.161
623952 89.347 -52.311 57.153
623953 89.380 -52.330 57.148
623954 89.413 -52.349 57.143
623955 89.446 -52.368 57.138
623956 89.479 -52.388 57.133
623957 89.512 -52.407 57.129
623958 89.545 -52.427 57.124
623968 89.873 -52.579 57.113
623969 89.905 -52.587 57.113
623970 89.938 -52.595 57.114
623971 89.970 -52.603 57.115
623972 90.003 -52.611 57.116
623973 90.035 -52.619 57.119
623974 90.067 -52.628 57.122
623975 90.088 -52.645 57.124
623976 90.104 -52.667 57.128
623977 90.120 -52.690 57.130
623978 90.136 -52.712 57.134
623979 90.152 -52.734 57.139
623980 90.182 -52.737 57.143
623981 90.214 -52.736 57.146
623982 90.246 -52.736 57.151
623983 90.279 -52.736 57.150
623984 90.310 -52.736 57.150
623985 90.342 -52.736 57.151
623986 90.368 -52.757 57.151
623987 90.393 -52.780 57.152
623988 90.418 -52.803 57.153
623989 90.444 -52.825 57.150
623990 90.471 -52.847 57.150
623991 90.507 -52.868 57.151
623992 90.542 -52.888 57.151
623993 90.577 -52.909 57.149
623994 90.612 -52.929 57.145
623995 90.647 -52.951 57.125
623996 90.687 -52.948 57.108
623997 90.728 -52.941 57.089
623998 90.769 -52.934 57.069
623999 90.810 -52.928 57.049
624000 90.851 -52.935 57.028
624001 90.891 -52.943 57.004
624002 90.931 -52.951 56.978
624003 90.962 -52.961 56.953
624004 90.966 -52.981 56.929
624005 90.970 -53.001 56.906
624006 90.987 -53.021 56.882
624007 91.014 -53.042 56.852
624008 91.040 -53.063 56.823
624009 91.067 -53.083 56.795
624010 91.094 -53.104 56.765
624011 91.121 -53.125 56.736
624012 91.148 -53.145 56.706
624013 91.175 -53.166 56.674
624014 91.204 -53.188 56.641
624015 91.237 -53.213 56.610
624016 91.270 -53.237 56.577
624017 91.304 -53.261 56.545
624018 91.337 -53.286 56.514
624019 91.370 -53.311 56.480
624020 91.403 -53.335 56.448
624021 91.437 -53.360 56.417
624022 91.470 -53.384 56.385
624023 91.504 -53.408 56.354
624024 91.537 -53.433 56.325
624025 91.570 -53.458 56.295
624026 91.603 -53.482 56.265
624027 91.637 -53.507 56.236
624028 91.670 -53.531 56.208
624029 91.704 -53.556 56.181
624030 91.737 -53.581 56.153
624031 91.770 -53.605 56.124
624032 91.803 -53.629 56.096
624033 91.837 -53.654 56.068
624034 91.870 -53.678 56.038
624035 91.911 -53.691 56.009
624036 91.952 -53.705 55.980
624037 91.993 -53.719 55.950
624038 92.033 -53.732 55.922
624039 92.074 -53.745 55.892
624040 92.115 -53.759 55.864
624041 92.156 -53.772 55.835
624056 92.767 -53.975 55.283
624057 92.807 -53.988 55.278
624058 92.848 -54.001 55.274
624059 92.889 -54.015 55.271
624060 92.930 -54.029 55.269
624061 92.971 -54.042 55.269
624062 93.011 -54.056 55.270
624063 93.052 -54.069 55.271
624064 93.093 -54.082 55.273
624065 93.134 -54.096 55.280
624066 93.174 -54.110 55.287
624067 93.216 -54.120 55.289
624068 93.236 -54.142 55.281
624069 93.252 -54.164 55.263
624070 93.261 -54.183 55.247
624071 93.270 -54.204 55.231
624072 93.279 -54.223 55.216
624073 93.288 -54.243 55.205
624074 93.297 -54.264 55.198
624075 93.306 -54.283 55.195
624076 93.315 -54.303 55.194
624077 93.325 -54.323 55.194
624078 93.334 -54.343 55.201
624079 93.361 -54.353 55.210
624080 93.400 -54.358 55.219
624081 93.437 -54.362 55.229
624082 93.468 -54.370 55.238
624083 93.499 -54.377 55.243
624084 93.529 -54.387 55.248
624085 93.550 -54.400 55.252
624086 93.571 -54.414 55.256
624087 93.592 -54.427 55.263
624088 93.614 -54.440 55.268
624089 93.636 -54.453 55.273
624090 93.657 -54.467 55.278
624091 93.679 -54.480 55.283
624092 93.700 -54.493 55.287
624093 93.721 -54.506 55.287
624094 93.743 -54.520 55.283
624095 93.764 -54.533 55.271
624096 93.786 -54.546 55.253
624097 93.807 -54.560 55.233
624098 93.829 -54.573 55.214
624099 93.850 -54.587 55.198
624100 93.871 -54.599 55.184
624101 93.893 -54.613 55.175
624102 93.915 -54.626 55.168
624103 93.936 -54.640 55.157
624104 93.958 -54.652 55.140
624105 93.979 -54.666 55.119
624106 93.995 -54.674 55.094
624107 94.012 -54.686 55.069
624108 94.037 -54.702 55.058
624109 94.062 -54.718 55.057
624110 94.086 -54.735 55.061
624111 94.112 -54.752 55.062
624112 94.136 -54.768 55.059
624113 94.161 -54.784 55.053
624114 94.186 -54.801 55.035
624115 94.210 -54.818 55.015
624116 94.235 -54.834 54.995
624117 94.260 -54.851 54.978
624118 94.284 -54.868 54.962
624119 94.309 -54.884 54.941
624120 94.342 -54.897 54.912
624121 94.380 -54.909 54.881
624122 94.417 -54.922 54.852
624123 94.454 -54.934 54.824
624124 94.491 -54.946 54.795
624125 94.529 -54.958 54.764
624126 94.566 -54.969 54.724
624127 94.604 -54.981 54.677
624128 94.641 -54.993 54.630
624129 94.679 -55.004 54.579
624130 94.716 -55.016 54.525
624131 94.754 -55.029 54.471
624132 94.791 -55.044 54.410
624133 94.828 -55.057 54.348
624134 94.860 -55.075 54.287
624135 94.893 -55.093 54.227
624136 94.925 -55.111 54.170
624137 94.959 -55.129 54.113
624138 94.993 -55.147 54.056
624139 95.028 -55.164 54.005
624140 95.062 -55.182 53.958
624141 95.096 -55.200 53.908
624142 95.130 -55.217 53.858
624143 95.163 -55.232 53.804
624144 95.194 -55.246 53.745
624145 95.226 -55.260 53.683
624146 95.258 -55.275 53.620
624147 95.289 -55.289 53.556
624148 95.321 -55.304 53.494
624149 95.353 -55.318 53.433
624150 95.385 -55.332 53.372
624151 95.417 -55.346 53.312
624152 95.449 -55.359 53.255
624153 95.480 -55.373 53.200
624154 95.513 -55.386 53.147
624155 95.544 -55.400 53.097
624156 95.576 -55.414 53.046
624157 95.609 -55.426 52.995
624158 95.642 -55.439 52.944
624159 95.673 -55.450 52.892
624160 95.703 -55.461 52.839
624161 95.734 -55.471 52.784
624162 95.765 -55.482 52.729
624163 95.795 -55.492 52.673
624164 95.826 -55.503 52.616
624165 95.858 -55.515 52.558
624166 95.892 -55.530 52.499
624167 95.927 -55.545 52.440
624168 95.961 -55.559 52.381
624169 95.996 -55.574 52.321
624170 96.031 -55.589 52.260
624171 96.065 -55.603 52.200
624172 96.100 -55.618 52.139
624173 96.134 -55.632 52.078
624174 96.169 -55.647 52.017
624175 96.203 -55.662 51.956
624176 96.238 -55.676 51.894
624177 96.270 -55.687 51.833
624178 96.301 -55.697 51.770
624179 96.332 -55.708 51.708
624180 96.364 -55.718 51.647
624181 96.395 -55.727 51.587
624182 96.426 -55.738 51.525
624183 96.460 -55.758 51.461
624184 96.495 -55.782 51.397
624185 96.530 -55.806 51.333
624186 96.556 -55.819 51.270
624187 96.573 -55.818 51.207
624188 96.590 -55.818 51.142
624189 96.606 -55.817 51.078
624190 96.623 -55.816 51.013
624191 96.640 -55.815 50.951
624192 96.656 -55.815 50.892
624193 96.673 -55.814 50.834
624194 96.689 -55.812 50.777
624195 96.706 -55.812 50.723
624594 87.501 -51.002 56.304
624595 87.532 -51.025 56.319
624596 87.563 -51.048 56.331
624597 87.593 -51.072 56.345
624598 87.625 -51.096 56.357
624599 87.655 -51.119 56.368
624600 87.687 -51.142 56.374
624601 87.717 -51.165 56.377
624602 87.749 -51.189 56.378
624603 87.779 -51.212 56.378
624604 87.811 -51.236 56.375
624605 87.841 -51.259 56.371
624606 87.872 -51.282 56.363
624607 87.906 -51.302 56.352
624608 87.941 -51.317 56.340
624609 87.976 -51.332 56.326
624647 89.168 -52.161 57.192
624648 89.201 -52.181 57.183
624649 89.235 -52.200 57.175
624650 89.267 -52.220 57.166
624651 89.301 -52.239 57.158
624652 89.334 -52.258 57.151
624653 89.367 -52.277 57.142
624654 89.400 -52.296 57.138
624655 89.433 -52.316 57.133
624656 89.466 -52.335 57.128
624657 89.499 -52.355 57.123
624658 89.532 -52.374 57.120
624669 89.893 -52.546 57.108
624670 89.925 -52.554 57.110
624671 89.958 -52.562 57.111
624672 89.990 -52.570 57.112
624673 90.022 -52.578 57.114
624674 90.054 -52.586 57.117
624675 90.087 -52.593 57.122
624676 90.119 -52.602 57.125
624677 90.135 -52.624 57.130
624678 90.151 -52.647 57.134
624679 90.167 -52.669 57.138
624680 90.183 -52.691 57.143
624681 90.206 -52.703 57.149
624682 90.238 -52.703 57.154
624683 90.270 -52.703 57.159
624684 90.302 -52.703 57.161
624685 90.334 -52.703 57.162
624686 90.366 -52.703 57.164
624687 90.393 -52.720 57.165
624688 90.418 -52.743 57.166
624689 90.444 -52.766 57.172
624690 90.469 -52.788 57.167
624691 90.494 -52.810 57.164
624692 90.523 -52.832 57.165
624693 90.558 -52.853 57.167
624694 90.593 -52.873 57.166
624695 90.628 -52.894 57.161
624696 90.663 -52.915 57.140
624697 90.704 -52.912 57.124
624698 90.744 -52.905 57.108
624699 90.785 -52.898 57.091
624700 90.826 -52.901 57.073
624701 90.866 -52.909 57.055
624702 90.906 -52.916 57.031
624703 90.947 -52.924 57.006
624704 90.987 -52.932 56.982
624705 91.001 -52.948 56.960
624706 91.005 -52.968 56.939
624707 91.010 -52.988 56.915
624708 91.018 -53.007 56.887
624709 91.044 -53.029 56.860
624710 91.072 -53.049 56.832
624711 91.098 -53.070 56.803
624712 91.125 -53.091 56.773
624713 91.152 -53.111 56.744
624714 91.184 -53.135 56.712
624715 91.218 -53.160 56.680
624716 91.251 -53.184 56.647
624717 91.284 -53.208 56.614
624718 91.317 -53.233 56.581
624719 91.351 -53.258 56.549
624720 91.384 -53.282 56.515
624721 91.418 -53.307 56.482
624722 91.451 -53.331 56.449
624723 91.484 -53.355 56.417
624724 91.517 -53.380 56.385
624725 91.551 -53.405 56.354
624726 91.584 -53.429 56.322
624727 91.617 -53.454 56.292
624728 91.651 -53.478 56.263
624729 91.684 -53.502 56.234
624730 91.717 -53.527 56.205
624731 91.751 -53.552 56.175
624732 91.784 -53.576 56.146
624733 91.817 -53.601 56.117
624734 91.851 -53.625 56.086
624735 91.884 -53.649 56.056
624736 91.925 -53.662 56.025
624737 91.966 -53.676 55.994
624738 92.007 -53.690 55.963
624739 92.048 -53.703 55.933
624740 92.088 -53.716 55.902
624741 92.129 -53.730 55.872
624742 92.170 -53.744 55.842
624757 92.781 -53.946 55.279
624758 92.821 -53.960 55.272
624759 92.862 -53.972 55.267
624760 92.903 -53.986 55.262
624761 92.944 -54.000 55.259
624762 92.985 -54.013 55.258
624763 93.025 -54.027 55.257
624764 93.066 -54.040 55.256
624765 93.107 -54.054 55.256
624766 93.148 -54.067 55.259
624767 93.194 -54.075 55.263
624768 93.221 -54.092 55.261
624769 93.241 -54.114 55.252
624770 93.261 -54.136 55.236
624771 93.280 -54.158 55.222
624772 93.291 -54.179 55.208
624773 93.300 -54.198 55.196
624774 93.309 -54.218 55.186
624775 93.318 -54.239 55.181
624776 93.327 -54.258 55.178
624777 93.337 -54.278 55.178
624778 93.346 -54.298 55.180
624779 93.355 -54.318 55.185
624780 93.386 -54.327 55.192
624781 93.424 -54.330 55.199
624782 93.458 -54.337 55.204
624783 93.489 -54.345 55.212
624784 93.520 -54.353 55.217
624785 93.551 -54.361 55.221
624786 93.577 -54.371 55.224
624787 93.599 -54.385 55.228
624788 93.620 -54.398 55.237
624789 93.641 -54.411 55.245
624790 93.663 -54.424 55.251
624791 93.684 -54.438 55.257
624792 93.706 -54.452 55.261
624793 93.727 -54.464 55.261
624794 93.749 -54.478 55.255
624795 93.770 -54.491 55.245
624796 93.791 -54.505 55.233
624797 93.813 -54.518 55.217
624798 93.835 -54.531 55.201
624799 93.856 -54.544 55.187
624800 93.877 -54.558 55.175
624801 93.899 -54.571 55.166
624802 93.920 -54.584 55.161
624803 93.942 -54.598 55.152
624804 93.963 -54.611 55.138
624805 93.985 -54.624 55.119
624806 94.003 -54.634 55.096
624807 94.018 -54.643 55.072
624808 94.034 -54.652 55.056
624809 94.054 -54.664 55.052
624810 94.078 -54.681 55.053
624811 94.103 -54.697 55.053
624812 94.128 -54.714 55.052
624813 94.152 -54.731 55.046
624814 94.177 -54.747 55.036
624815 94.202 -54.763 55.015
624816 94.226 -54.780 54.993
624817 94.251 -54.797 54.971
624818 94.276 -54.813 54.952
624819 94.301 -54.830 54.936
624820 94.326 -54.847 54.914
624821 94.357 -54.860 54.884
624822 94.394 -54.872 54.853
624823 94.431 -54.884 54.824
624824 94.469 -54.897 54.796
624825 94.506 -54.909 54.768
624826 94.543 -54.921 54.739
624827 94.580 -54.933 54.700
624828 94.618 -54.945 54.655
624829 94.655 -54.956 54.609
624830 94.693 -54.969 54.560
624831 94.730 -54.980 54.508
624832 94.767 -54.993 54.456
624833 94.805 -55.007 54.395
624834 94.837 -55.025 54.333
624835 94.870 -55.043 54.270
624836 94.904 -55.060 54.209
624837 94.938 -55.078 54.149
624838 94.973 -55.096 54.091
624839 95.007 -55.113 54.032
624840 95.041 -55.131 53.979
624841 95.075 -55.149 53.930
624842 95.110 -55.166 53.879
624843 95.144 -55.184 53.826
624844 95.176 -55.198 53.771
624845 95.207 -55.212 53.711
624846 95.239 -55.226 53.648
624847 95.271 -55.241 53.587
624848 95.303 -55.255 53.526
624849 95.334 -55.270 53.467
624850 95.366 -55.284 53.409
624851 95.398 -55.298 53.352
624852 95.430 -55.312 53.294
624853 95.462 -55.326 53.239
624854 95.493 -55.341 53.190
624855 95.525 -55.354 53.142
624856 95.557 -55.368 53.096
624857 95.589 -55.382 53.049
624858 95.621 -55.395 53.000
624859 95.654 -55.407 52.950
624860 95.684 -55.417 52.901
624861 95.714 -55.428 52.851
624862 95.745 -55.439 52.798
624863 95.775 -55.449 52.744
624864 95.806 -55.460 52.689
624865 95.836 -55.470 52.633
624866 95.868 -55.482 52.574
624867 95.902 -55.496 52.515
624868 95.937 -55.511 52.456
624869 95.971 -55.526 52.397
624870 96.006 -55.540 52.336
624871 96.041 -55.555 52.275
624872 96.075 -55.569 52.214
624873 96.110 -55.584 52.151
624874 96.144 -55.599 52.090
624875 96.179 -55.613 52.028
624876 96.212 -55.626 51.966
624877 96.243 -55.636 51.905
624878 96.274 -55.646 51.843
624879 96.305 -55.655 51.780
624880 96.336 -55.666 51.718
624881 96.368 -55.676 51.656
624882 96.399 -55.686 51.594
624883 96.430 -55.696 51.532
624884 96.465 -55.721 51.468
624885 96.500 -55.745 51.403
624886 96.535 -55.769 51.339
624887 96.570 -55.793 51.276
624888 96.587 -55.793 51.212
624889 96.604 -55.793 51.148
624890 96.620 -55.792 51.082
624891 96.637 -55.791 51.017
624892 96.654 -55.790 50.955
624893 96.670 -55.790 50.895
624894 96.687 -55.788 50.837
624895 96.703 -55.787 50.780
624896 96.720 -55.787 50.724
625295 87.522 -50.973 56.327
625296 87.553 -50.995 56.342
625297 87.584 -51.019 56.356
625298 87.615 -51.042 56.370
625299 87.646 -51.066 56.383
625300 87.677 -51.089 56.393
625301 87.708 -51.112 56.401
625302 87.739 -51.136 56.403
625303 87.770 -51.159 56.404
625304 87.801 -51.183 56.403
625305 87.832 -51.206 56.400
625306 87.863 -51.230 56.394
625307 87.895 -51.252 56.386
625308 87.930 -51.267 56.373
625309 87.966 -51.282 56.360
625310 88.002 -51.299 56.345
625349 89.222 -52.148 57.173
625350 89.254 -52.167 57.165
625351 89.287 -52.186 57.155
625352 89.321 -52.205 57.147
625353 89.354 -52.225 57.140
625354 89.387 -52.244 57.132
625355 89.420 -52.264 57.126
625356 89.453 -52.283 57.122
625357 89.486 -52.302 57.117
625358 89.519 -52.321 57.113
625370 89.913 -52.512 57.104
625371 89.946 -52.521 57.107
625372 89.978 -52.528 57.109
625373 90.010 -52.537 57.112
625374 90.042 -52.544 57.115
625375 90.075 -52.553 57.119
625376 90.107 -52.560 57.123
625377 90.139 -52.568 57.127
625378 90.166 -52.581 57.133
625379 90.182 -52.603 57.138
625380 90.198 -52.625 57.143
625381 90.215 -52.647 57.148
625382 90.231 -52.669 57.155
625383 90.262 -52.671 57.162
625384 90.294 -52.671 57.167
625385 90.326 -52.671 57.171
625386 90.358 -52.671 57.174
625387 90.390 -52.671 57.176
625388 90.418 -52.683 57.178
625389 90.443 -52.706 57.183
625390 90.468 -52.728 57.188
625391 90.494 -52.750 57.181
625392 90.519 -52.773 57.176
625393 90.544 -52.796 57.174
625394 90.574 -52.818 57.174
625395 90.609 -52.838 57.174
625396 90.644 -52.859 57.167
625397 90.679 -52.879 57.153
625398 90.720 -52.877 57.140
625399 90.760 -52.869 57.128
625400 90.801 -52.868 57.115
625401 90.841 -52.875 57.099
625402 90.882 -52.882 57.083
625403 90.922 -52.890 57.059
625404 90.962 -52.897 57.034
625405 91.003 -52.905 57.010
625406 91.035 -52.915 56.990
625407 91.040 -52.935 56.970
625408 91.044 -52.955 56.947
625409 91.048 -52.975 56.921
625410 91.053 -52.995 56.894
625411 91.076 -53.015 56.867
625412 91.102 -53.035 56.839
625413 91.132 -53.057 56.810
625414 91.165 -53.082 56.782
625415 91.198 -53.107 56.751
625416 91.231 -53.131 56.717
625417 91.264 -53.156 56.684
625418 91.298 -53.180 56.651
625419 91.331 -53.205 56.618
625420 91.365 -53.230 56.585
625421 91.398 -53.254 56.550
625422 91.431 -53.278 56.516
625423 91.464 -53.303 56.482
625424 91.498 -53.327 56.449
625425 91.531 -53.352 56.416
625426 91.565 -53.377 56.383
625427 91.598 -53.401 56.351
625428 91.631 -53.425 56.320
625429 91.665 -53.450 56.290
625430 91.698 -53.474 56.260
625431 91.731 -53.499 56.229
625432 91.765 -53.524 56.198
625433 91.798 -53.548 56.168
625434 91.831 -53.572 56.137
625435 91.865 -53.597 56.106
625436 91.899 -53.620 56.074
625437 91.940 -53.634 56.042
625438 91.980 -53.647 56.009
625439 92.021 -53.661 55.977
625440 92.062 -53.674 55.945
625441 92.102 -53.687 55.913
625442 92.143 -53.701 55.881
625443 92.184 -53.715 55.849
625444 92.225 -53.728 55.818
625458 92.795 -53.917 55.274
625459 92.836 -53.930 55.267
625460 92.876 -53.944 55.261
625461 92.917 -53.957 55.254
625462 92.958 -53.971 55.248
625463 92.999 -53.985 55.245
625464 93.040 -53.997 55.243
625465 93.080 -54.011 55.239
625466 93.121 -54.025 55.236
625467 93.169 -54.031 55.236
625468 93.207 -54.042 55.236
625469 93.226 -54.064 55.232
625470 93.246 -54.086 55.221
625471 93.266 -54.108 55.209
625472 93.285 -54.130 55.196
625473 93.304 -54.152 55.185
625474 93.321 -54.173 55.175
625475 93.330 -54.193 55.167
625476 93.339 -54.214 55.163
625477 93.348 -54.233 55.162
625478 93.357 -54.253 55.164
625479 93.366 -54.273 55.166
625480 93.375 -54.293 55.169
625481 93.410 -54.299 55.174
625482 93.447 -54.304 55.178
625483 93.478 -54.312 55.180
625484 93.509 -54.320 55.185
625485 93.540 -54.328 55.189
625486 93.571 -54.336 55.193
625487 93.603 -54.344 55.197
625488 93.626 -54.356 55.202
625489 93.647 -54.370 55.210
625490 93.669 -54.383 55.221
625491 93.690 -54.396 55.230
625492 93.712 -54.409 55.236
625493 93.733 -54.423 55.237
625494 93.755 -54.436 55.231
625495 93.776 -54.449 55.220
625496 93.797 -54.462 55.206
625497 93.819 -54.476 55.198
625498 93.840 -54.489 55.189
625499 93.862 -54.502 55.179
625500 93.883 -54.516 55.170
625501 93.905 -54.529 55.161
625502 93.926 -54.543 55.152
625503 93.947 -54.555 55.145
625504 93.969 -54.569 55.133
625505 93.991 -54.582 55.115
625506 94.011 -54.595 55.094
625507 94.027 -54.603 55.073
625508 94.042 -54.612 55.056
625509 94.057 -54.621 55.050
625510 94.073 -54.629 55.050
625511 94.095 -54.643 55.048
625512 94.119 -54.659 55.045
625513 94.144 -54.676 55.039
625514 94.169 -54.693 55.030
625515 94.193 -54.709 55.017
625516 94.218 -54.725 54.994
625517 94.243 -54.742 54.970
625518 94.268 -54.759 54.948
625519 94.292 -54.775 54.929
625520 94.317 -54.792 54.910
625521 94.342 -54.809 54.888
625522 94.371 -54.823 54.857
625523 94.409 -54.835 54.826
625524 94.446 -54.847 54.796
625525 94.483 -54.859 54.767
625526 94.521 -54.872 54.740
625527 94.558 -54.884 54.711
625528 94.595 -54.896 54.672
625529 94.632 -54.908 54.631
625530 94.670 -54.920 54.586
625531 94.707 -54.932 54.539
625532 94.744 -54.944 54.490
625533 94.781 -54.956 54.439
625534 94.815 -54.974 54.380
625535 94.849 -54.991 54.317
625536 94.884 -55.010 54.254
625537 94.918 -55.027 54.191
625538 94.952 -55.044 54.131
625539 94.986 -55.063 54.073
625540 95.021 -55.080 54.015
625541 95.055 -55.097 53.960
625542 95.089 -55.116 53.906
625543 95.124 -55.133 53.853
625544 95.157 -55.150 53.799
625545 95.189 -55.164 53.743
625546 95.220 -55.179 53.683
625547 95.252 -55.192 53.623
625548 95.284 -55.207 53.565
625549 95.316 -55.221 53.507
625550 95.347 -55.235 53.451
625551 95.379 -55.250 53.396
625552 95.411 -55.264 53.340
625553 95.443 -55.278 53.284
625554 95.475 -55.292 53.234
625555 95.507 -55.307 53.187
625556 95.538 -55.321 53.143
625557 95.570 -55.336 53.100
625558 95.602 -55.350 53.056
625559 95.634 -55.363 53.010
625560 95.664 -55.373 52.961
625561 95.695 -55.384 52.914
625562 95.725 -55.395 52.865
625563 95.756 -55.405 52.814
625564 95.787 -55.416 52.762
625565 95.817 -55.426 52.708
625566 95.847 -55.437 52.651
625567 95.878 -55.448 52.593
625568 95.912 -55.463 52.533
625569 95.947 -55.477 52.474
625570 95.981 -55.492 52.415
625571 96.016 -55.507 52.353
625572 96.051 -55.521 52.291
625573 96.085 -55.536 52.229
625574 96.120 -55.550 52.165
625575 96.154 -55.564 52.102
625576 96.185 -55.574 52.040
625577 96.216 -55.584 51.977
625578 96.247 -55.594 51.916
625579 96.278 -55.604 51.852
625580 96.309 -55.614 51.789
625581 96.340 -55.624 51.727
625582 96.372 -55.634 51.664
625583 96.403 -55.644 51.602
625584 96.435 -55.658 51.539
625585 96.470 -55.683 51.474
625586 96.505 -55.707 51.410
625587 96.540 -55.731 51.345
625588 96.575 -55.756 51.282
625589 96.602 -55.768 51.218
625590 96.618 -55.768 51.154
625591 96.635 -55.767 51.088
625592 96.652 -55.766 51.023
625593 96.668 -55.765 50.961
625594 96.685 -55.764 50.901
625595 96.701 -55.763 50.842
625596 96.718 -55.762 50.784
625996 87.565 -50.960 56.349
625997 87.575 -50.966 56.365
625998 87.606 -50.989 56.380
625999 87.637 -51.013 56.395
626000 87.668 -51.036 56.408
626001 87.699 -51.060 56.419
626002 87.730 -51.083 56.427
626003 87.761 -51.106 56.430
626004 87.792 -51.130 56.430
626005 87.823 -51.153 56.428
626006 87.854 -51.177 56.424
626007 87.885 -51.200 56.418
626008 87.919 -51.216 56.408
626009 87.955 -51.232 56.395
626010 87.991 -51.249 56.380
626011 88.030 -51.268 56.363
626012 88.066 -51.289 56.345
626055 89.407 -52.211 57.121
626056 89.440 -52.230 57.115
626057 89.473 -52.249 57.111
626058 89.506 -52.269 57.107
626071 89.933 -52.479 57.102
626072 89.965 -52.487 57.105
626073 89.998 -52.495 57.108
626074 90.030 -52.503 57.112
626075 90.062 -52.511 57.116
626076 90.095 -52.519 57.121
626077 90.127 -52.527 57.125
626078 90.159 -52.535 57.130
626079 90.191 -52.543 57.137
626080 90.214 -52.560 57.144
626081 90.230 -52.582 57.149
626082 90.246 -52.604 57.156
626083 90.262 -52.626 57.163
626084 90.286 -52.638 57.169
626085 90.317 -52.638 57.176
626086 90.350 -52.638 57.181
626087 90.382 -52.638 57.184
626088 90.413 -52.638 57.187
626089 90.443 -52.646 57.190
626090 90.468 -52.669 57.197
626091 90.494 -52.691 57.200
626092 90.519 -52.713 57.193
626093 90.544 -52.736 57.184
626094 90.569 -52.759 57.179
626095 90.594 -52.782 57.176
626096 90.625 -52.803 57.176
626097 90.660 -52.823 57.170
626098 90.695 -52.844 57.161
626099 90.736 -52.841 57.153
626100 90.777 -52.835 57.145
626101 90.817 -52.841 57.133
626102 90.857 -52.849 57.121
626103 90.897 -52.857 57.108
626104 90.938 -52.864 57.086
626105 90.978 -52.871 57.060
626106 91.018 -52.879 57.038
626107 91.059 -52.886 57.019
626108 91.075 -52.902 57.000
626109 91.079 -52.922 56.979
626110 91.083 -52.941 56.953
626111 91.088 -52.962 56.927
626112 91.092 -52.982 56.901
626113 91.112 -53.004 56.873
626114 91.145 -53.029 56.845
626115 91.178 -53.054 56.818
626116 91.212 -53.078 56.788
626117 91.245 -53.103 56.756
626118 91.279 -53.127 56.722
626119 91.312 -53.151 56.689
626120 91.345 -53.176 56.656
626121 91.379 -53.201 56.622
626122 91.412 -53.225 56.586
626123 91.445 -53.250 56.551
626124 91.478 -53.274 56.516
626125 91.512 -53.299 56.482
626126 91.545 -53.324 56.447
626127 91.579 -53.348 56.414
626128 91.612 -53.372 56.380
626129 91.645 -53.397 56.349
626130 91.678 -53.421 56.317
626131 91.712 -53.446 56.286
626132 91.745 -53.471 56.255
626133 91.779 -53.495 56.223
626134 91.812 -53.519 56.191
626135 91.845 -53.544 56.159
626136 91.878 -53.568 56.126
626137 91.913 -53.591 56.093
626138 91.954 -53.605 56.059
626139 91.995 -53.618 56.025
626140 92.035 -53.631 55.991
626141 92.076 -53.645 55.958
626142 92.117 -53.659 55.925
626143 92.157 -53.672 55.891
626144 92.198 -53.686 55.857
626145 92.239 -53.699 55.825
626159 92.809 -53.888 55.270
626160 92.850 -53.901 55.261
626161 92.891 -53.915 55.253
626162 92.931 -53.928 55.244
626163 92.972 -53.942 55.237
626164 93.013 -53.956 55.232
626165 93.054 -53.969 55.228
626166 93.096 -53.981 55.222
626167 93.144 -53.987 55.216
626168 93.192 -53.993 55.211
626169 93.212 -54.015 55.206
626170 93.231 -54.036 55.200
626171 93.251 -54.058 55.193
626172 93.271 -54.080 55.181
626173 93.290 -54.102 55.171
626174 93.309 -54.124 55.162
626175 93.329 -54.146 55.154
626176 93.349 -54.168 55.149
626177 93.360 -54.189 55.146
626178 93.369 -54.208 55.147
626179 93.378 -54.228 55.150
626180 93.388 -54.248 55.152
626181 93.397 -54.268 55.153
626182 93.435 -54.272 55.155
626183 93.468 -54.279 55.157
626184 93.499 -54.287 55.157
626185 93.530 -54.295 55.160
626186 93.561 -54.303 55.163
626187 93.592 -54.311 55.167
626188 93.623 -54.319 55.172
626189 93.653 -54.327 55.178
626190 93.675 -54.341 55.186
626191 93.696 -54.354 55.197
626192 93.717 -54.367 55.207
626193 93.739 -54.380 55.211
626194 93.760 -54.394 55.208
626195 93.782 -54.408 55.200
626196 93.803 -54.421 55.189
626197 93.825 -54.434 55.181
626198 93.846 -54.447 55.181
626199 93.867 -54.461 55.181
626200 93.889 -54.474 55.176
626201 93.911 -54.487 55.167
626202 93.932 -54.500 55.156
626203 93.954 -54.514 55.140
626204 93.975 -54.527 55.128
626205 93.996 -54.540 55.110
626206 94.018 -54.553 55.088
626207 94.035 -54.563 55.066
626208 94.050 -54.571 55.049
626209 94.065 -54.580 55.043
626210 94.081 -54.589 55.047
626211 94.097 -54.598 55.046
626212 94.112 -54.606 55.042
626213 94.135 -54.621 55.034
626214 94.160 -54.638 55.023
626215 94.185 -54.655 55.012
626216 94.210 -54.671 54.997
626217 94.235 -54.688 54.973
626218 94.259 -54.704 54.949
626219 94.284 -54.721 54.927
626220 94.309 -54.737 54.908
626221 94.333 -54.754 54.888
626222 94.358 -54.771 54.865
626223 94.387 -54.786 54.835
626224 94.424 -54.798 54.803
626225 94.461 -54.810 54.772
626226 94.498 -54.822 54.741
626227 94.535 -54.834 54.711
626228 94.572 -54.847 54.682
626229 94.610 -54.858 54.643
626230 94.647 -54.870 54.601
626231 94.684 -54.882 54.558
626232 94.721 -54.894 54.513
626233 94.759 -54.906 54.466
626234 94.795 -54.925 54.416
626235 94.829 -54.941 54.358
626236 94.863 -54.959 54.296
626237 94.897 -54.976 54.233
626238 94.931 -54.994 54.172
626239 94.966 -55.012 54.114
626240 95.000 -55.029 54.062
626241 95.034 -55.047 54.006
626242 95.069 -55.065 53.946
626243 95.103 -55.082 53.887
626244 95.137 -55.100 53.830
626245 95.170 -55.116 53.775
626246 95.202 -55.130 53.721
626247 95.234 -55.144 53.663
626248 95.265 -55.159 53.605
626249 95.297 -55.173 53.550
626250 95.329 -55.187 53.494
626251 95.360 -55.201 53.440
626252 95.393 -55.216 53.386
626253 95.424 -55.230 53.330
626254 95.456 -55.244 53.278
626255 95.488 -55.258 53.232
626256 95.520 -55.273 53.189
626257 95.551 -55.288 53.147
626258 95.583 -55.303 53.107
626259 95.614 -55.318 53.066
626260 95.645 -55.330 53.022
626261 95.675 -55.341 52.975
626262 95.705 -55.351 52.928
626263 95.736 -55.362 52.879
626264 95.766 -55.373 52.829
626265 95.797 -55.383 52.779
626266 95.828 -55.394 52.726
626267 95.858 -55.405 52.669
626268 95.889 -55.415 52.611
626269 95.922 -55.429 52.553
626270 95.957 -55.444 52.495
626271 95.991 -55.458 52.434
626272 96.026 -55.473 52.372
626273 96.061 -55.487 52.308
626274 96.095 -55.502 52.244
626275 96.126 -55.512 52.178
626276 96.158 -55.522 52.114
626277 96.189 -55.532 52.052
626278 96.220 -55.542 51.988
626279 96.251 -55.552 51.926
626280 96.283 -55.562 51.862
626281 96.314 -55.573 51.798
626282 96.345 -55.583 51.735
626283 96.376 -55.593 51.673
626284 96.407 -55.602 51.610
626285 96.441 -55.621 51.546
626286 96.475 -55.646 51.482
626287 96.511 -55.670 51.416
626288 96.545 -55.694 51.352
626289 96.581 -55.718 51.288
626290 96.615 -55.743 51.224
626291 96.632 -55.743 51.160
626292 96.649 -55.742 51.094
626293 96.666 -55.740 51.030
626294 96.682 -55.740 50.967
626295 96.699 -55.739 50.907
626296 96.716 -55.738 50.848
626697 87.625 -50.963 56.371
626698 87.633 -50.966 56.389
626699 87.628 -50.960 56.405
626700 87.659 -50.983 56.420
626701 87.690 -51.007 56.433
626702 87.721 -51.030 56.445
626703 87.752 -51.053 56.453
626704 87.783 -51.077 56.457
626705 87.814 -51.100 56.456
626706 87.845 -51.124 56.454
626707 87.875 -51.147 56.449
626708 87.909 -51.166 56.442
626709 87.944 -51.181 56.431
626710 87.981 -51.199 56.415
626711 88.019 -51.218 56.399
626712 88.027 -51.263 56.381
626713 88.013 -51.325 56.360
626714 88.045 -51.348 56.339
626773 89.985 -52.453 57.105
626774 90.018 -52.462 57.109
626775 90.050 -52.470 57.113
626776 90.082 -52.477 57.117
626777 90.115 -52.486 57.123
626778 90.147 -52.493 57.128
626779 90.179 -52.502 57.134
626780 90.212 -52.510 57.142
626781 90.244 -52.518 57.149
626782 90.261 -52.539 57.155
626783 90.277 -52.561 57.162
626784 90.293 -52.583 57.170
626785 90.309 -52.606 57.177
626786 90.341 -52.606 57.185
626787 90.373 -52.605 57.190
626788 90.405 -52.605 57.193
626789 90.437 -52.605 57.196
626790 90.468 -52.609 57.199
626791 90.494 -52.631 57.204
626792 90.519 -52.654 57.205
626793 90.544 -52.676 57.198
626794 90.569 -52.699 57.189
626795 90.594 -52.722 57.181
626796 90.619 -52.744 57.176
626797 90.645 -52.766 57.174
626798 90.676 -52.788 57.170
626799 90.712 -52.809 57.165
626800 90.751 -52.806 57.161
626801 90.792 -52.807 57.156
626802 90.832 -52.815 57.147
626803 90.873 -52.822 57.136
626804 90.913 -52.830 57.124
626805 90.953 -52.838 57.105
626806 90.994 -52.845 57.084
626807 91.034 -52.853 57.064
626808 91.074 -52.860 57.048
626809 91.108 -52.870 57.030
626810 91.110 -52.895 57.008
626811 91.111 -52.919 56.982
626812 91.113 -52.944 56.957
626813 91.117 -52.967 56.932
626814 91.133 -52.981 56.906
626815 91.159 -53.001 56.879
626816 91.192 -53.025 56.853
626817 91.226 -53.050 56.824
626818 91.259 -53.074 56.793
626819 91.293 -53.098 56.761
626820 91.326 -53.123 56.727
626821 91.359 -53.148 56.693
626822 91.392 -53.172 56.659
626823 91.426 -53.197 56.623
626824 91.459 -53.221 56.586
626825 91.493 -53.245 56.550
626826 91.526 -53.270 56.515
626827 91.559 -53.295 56.479
626828 91.592 -53.319 56.446
626829 91.626 -53.344 56.411
626830 91.659 -53.368 56.378
626831 91.693 -53.393 56.345
626832 91.726 -53.418 56.313
626833 91.759 -53.442 56.281
626834 91.792 -53.466 56.248
626835 91.825 -53.491 56.215
626836 91.859 -53.515 56.181
626837 91.892 -53.540 56.147
626838 91.927 -53.562 56.112
626839 91.968 -53.576 56.077
626840 92.009 -53.590 56.042
626841 92.050 -53.603 56.006
626842 92.090 -53.616 55.971
626843 92.131 -53.630 55.936
626844 92.171 -53.643 55.901
626845 92.212 -53.657 55.866
626846 92.253 -53.670 55.832
626847 92.294 -53.684 55.801
626860 92.824 -53.860 55.265
626861 92.864 -53.872 55.255
626862 92.905 -53.886 55.245
626863 92.946 -53.900 55.234
626864 92.986 -53.913 55.225
626865 93.027 -53.926 55.219
626866 93.071 -53.937 55.213
626867 93.119 -53.943 55.205
626868 93.166 -53.949 55.197
626869 93.198 -53.965 55.189
626870 93.217 -53.987 55.181
626871 93.237 -54.009 55.174
626872 93.256 -54.030 55.167
626873 93.276 -54.052 55.157
626874 93.295 -54.074 55.149
626875 93.315 -54.096 55.141
626876 93.334 -54.118 55.134
626877 93.354 -54.140 55.130
626878 93.373 -54.162 55.129
626879 93.390 -54.183 55.129
626880 93.399 -54.203 55.132
626881 93.408 -54.223 55.133
626882 93.421 -54.241 55.134
626883 93.457 -54.246 55.136
626884 93.488 -54.254 55.136
626885 93.519 -54.262 55.135
626886 93.550 -54.270 55.137
626887 93.581 -54.278 55.139
626888 93.613 -54.286 55.143
626889 93.643 -54.294 55.149
626890 93.675 -54.302 55.155
626891 93.702 -54.312 55.163
626892 93.723 -54.326 55.173
626893 93.745 -54.339 55.181
626894 93.766 -54.352 55.183
626895 93.787 -54.365 55.179
626896 93.809 -54.379 55.171
626897 93.831 -54.392 55.162
626898 93.852 -54.405 55.160
626899 93.873 -54.418 55.165
626900 93.895 -54.432 55.169
626901 93.916 -54.445 55.167
626902 93.938 -54.458 55.157
626903 93.959 -54.472 55.143
626904 93.981 -54.485 55.124
626905 94.002 -54.499 55.105
626906 94.024 -54.511 55.083
626907 94.043 -54.523 55.062
626908 94.058 -54.531 55.047
626909 94.073 -54.540 55.039
626910 94.089 -54.549 55.037
626911 94.105 -54.558 55.038
626912 94.120 -54.566 55.035
626913 94.136 -54.575 55.027
626914 94.152 -54.583 55.017
626915 94.177 -54.600 55.005
626916 94.201 -54.617 54.991
626917 94.226 -54.634 54.975
626918 94.251 -54.650 54.953
626919 94.275 -54.666 54.931
626920 94.300 -54.683 54.910
626921 94.325 -54.699 54.890
626922 94.349 -54.716 54.870
626923 94.374 -54.733 54.847
626924 94.401 -54.749 54.816
626925 94.438 -54.761 54.784
626926 94.476 -54.773 54.750
626927 94.513 -54.784 54.717
626928 94.550 -54.797 54.684
626929 94.587 -54.809 54.653
626930 94.624 -54.821 54.614
626931 94.662 -54.833 54.572
626932 94.699 -54.845 54.529
626933 94.736 -54.857 54.485
626934 94.774 -54.873 54.439
626935 94.812 -54.895 54.389
626936 94.845 -54.912 54.331
626937 94.879 -54.928 54.270
626938 94.912 -54.946 54.211
626939 94.946 -54.963 54.156
626940 94.980 -54.979 54.104
626941 95.014 -54.996 54.053
626942 95.048 -55.014 53.998
626943 95.083 -55.032 53.940
626944 95.117 -55.049 53.881
626945 95.151 -55.067 53.824
626946 95.183 -55.082 53.767
626947 95.215 -55.096 53.713
626948 95.247 -55.110 53.658
626949 95.278 -55.125 53.599
626950 95.310 -55.138 53.543
626951 95.342 -55.153 53.488
626952 95.374 -55.168 53.434
626953 95.405 -55.183 53.379
626954 95.437 -55.198 53.324
626955 95.469 -55.213 53.277
626956 95.500 -55.228 53.235
626957 95.532 -55.242 53.194
626958 95.564 -55.257 53.154
626959 95.596 -55.272 53.116
626960 95.627 -55.287 53.077
626961 95.657 -55.298 53.034
626962 95.687 -55.308 52.989
626963 95.718 -55.319 52.941
626964 95.747 -55.329 52.891
626965 95.777 -55.340 52.842
626966 95.808 -55.350 52.794
626967 95.838 -55.361 52.742
626968 95.869 -55.371 52.686
626969 95.899 -55.382 52.630
626970 95.933 -55.395 52.573
626971 95.967 -55.410 52.514
626972 96.002 -55.424 52.452
626973 96.036 -55.439 52.389
626974 96.068 -55.450 52.323
626975 96.099 -55.461 52.258
626976 96.130 -55.470 52.192
626977 96.162 -55.480 52.127
626978 96.193 -55.491 52.064
626979 96.224 -55.501 52.000
626980 96.255 -55.511 51.937
626981 96.287 -55.520 51.872
626982 96.318 -55.531 51.808
626983 96.349 -55.541 51.744
626984 96.380 -55.551 51.681
626985 96.411 -55.561 51.618
626986 96.446 -55.583 51.555
626987 96.481 -55.608 51.490
626988 96.516 -55.632 51.424
626989 96.551 -55.656 51.359
626990 96.586 -55.680 51.296
626991 96.621 -55.705 51.231
626992 96.647 -55.717 51.167
626993 96.664 -55.716 51.101
626994 96.680 -55.715 51.037
626995 96.697 -55.715 50.974
626996 96.714 -55.714 50.914
627398 87.642 -50.933 56.394
627399 87.678 -50.958 56.412
627400 87.701 -50.972 56.428
627401 87.681 -50.954 56.444
627402 87.711 -50.977 56.458
627403 87.743 -51.001 56.470
627404 87.773 -51.023 56.479
627405 87.804 -51.047 56.482
627406 87.835 -51.070 56.481
627407 87.866 -51.094 56.477
627408 87.898 -51.116 56.471
627409 87.933 -51.131 56.463
627410 87.971 -51.149 56.452
627411 88.006 -51.171 56.435
627412 87.988 -51.236 56.416
627413 87.998 -51.279 56.396
627414 88.030 -51.302 56.375
627415 88.061 -51.324 56.351
627474 90.006 -52.420 57.104
627475 90.038 -52.428 57.109
627476 90.070 -52.437 57.114
627477 90.102 -52.444 57.120
627478 90.134 -52.452 57.125
627479 90.167 -52.460 57.132
627480 90.199 -52.468 57.138
627481 90.232 -52.476 57.146
627482 90.264 -52.484 57.153
627483 90.292 -52.496 57.160
627484 90.308 -52.518 57.168
627485 90.324 -52.540 57.177
627486 90.340 -52.562 57.184
627487 90.365 -52.572 57.191
627488 90.397 -52.572 57.196
627489 90.429 -52.572 57.200
627490 90.461 -52.572 57.202
627491 90.493 -52.572 57.204
627492 90.519 -52.594 57.206
627493 90.544 -52.617 57.205
627494 90.569 -52.639 57.198
627495 90.594 -52.662 57.190
627496 90.619 -52.684 57.182
627497 90.645 -52.707 57.175
627498 90.670 -52.729 57.171
627499 90.695 -52.752 57.169
627500 90.728 -52.773 57.166
627501 90.767 -52.773 57.163
627502 90.808 -52.781 57.160
627503 90.848 -52.788 57.155
627504 90.888 -52.796 57.147
627505 90.929 -52.803 57.135
627506 90.968 -52.812 57.120
627507 91.008 -52.820 57.104
627508 91.047 -52.828 57.090
627509 91.086 -52.836 57.076
627510 91.118 -52.851 57.056
627511 91.133 -52.882 57.034
627512 91.134 -52.907 57.009
627513 91.136 -52.932 56.984
627514 91.147 -52.949 56.960
627515 91.163 -52.963 56.935
627516 91.179 -52.976 56.911
627517 91.207 -52.997 56.887
627518 91.240 -53.021 56.859
627519 91.273 -53.046 56.830
627520 91.306 -53.070 56.798
627521 91.340 -53.095 56.764
627522 91.373 -53.120 56.728
627523 91.406 -53.144 56.695
627524 91.440 -53.168 56.659
627525 91.473 -53.193 56.622
627526 91.506 -53.217 56.584
627527 91.539 -53.242 56.548
627528 91.573 -53.267 56.512
627529 91.606 -53.291 56.478
627530 91.640 -53.315 56.442
627531 91.673 -53.340 56.408
627532 91.706 -53.364 56.374
627533 91.740 -53.389 56.340
627534 91.773 -53.414 56.307
627535 91.806 -53.438 56.273
627536 91.840 -53.462 56.238
627537 91.873 -53.487 56.203
627538 91.906 -53.512 56.167
627539 91.941 -53.534 56.132
627540 91.982 -53.547 56.095
627541 92.023 -53.561 56.058
627542 92.064 -53.574 56.021
627543 92.104 -53.587 55.984
627544 92.145 -53.601 55.948
627545 92.186 -53.615 55.911
627546 92.226 -53.628 55.875
627547 92.267 -53.641 55.840
627548 92.308 -53.655 55.807
627561 92.838 -53.830 55.260
627562 92.878 -53.844 55.248
627563 92.919 -53.857 55.237
627564 92.960 -53.871 55.225
627565 93.001 -53.885 55.214
627566 93.046 -53.894 55.206
627567 93.093 -53.900 55.198
627568 93.141 -53.906 55.189
627569 93.184 -53.915 55.179
627570 93.203 -53.937 55.169
627571 93.222 -53.959 55.159
627572 93.242 -53.981 55.151
627573 93.262 -54.003 55.143
627574 93.281 -54.024 55.135
627575 93.300 -54.046 55.127
627576 93.320 -54.068 55.120
627577 93.339 -54.090 55.114
627578 93.359 -54.112 55.112
627579 93.378 -54.134 55.110
627580 93.398 -54.156 55.111
627581 93.417 -54.178 55.113
627582 93.429 -54.198 55.115
627583 93.445 -54.214 55.116
627584 93.478 -54.222 55.116
627585 93.509 -54.229 55.116
627586 93.540 -54.238 55.115
627587 93.571 -54.245 55.116
627588 93.602 -54.254 55.118
627589 93.633 -54.261 55.123
627590 93.664 -54.269 55.128
627591 93.695 -54.277 55.133
627592 93.726 -54.285 55.139
627593 93.751 -54.297 55.148
627594 93.772 -54.310 55.153
627595 93.793 -54.323 55.154
627596 93.815 -54.336 55.150
627597 93.836 -54.350 55.143
627598 93.858 -54.364 55.139
627599 93.879 -54.377 55.142
627600 93.901 -54.390 55.148
627601 93.922 -54.403 55.153
627602 93.943 -54.417 55.151
627603 93.965 -54.430 55.141
627604 93.986 -54.443 55.125
627605 94.008 -54.456 55.104
627606 94.030 -54.470 55.082
627607 94.050 -54.483 55.060
627608 94.066 -54.491 55.043
627609 94.081 -54.500 55.034
627610 94.097 -54.508 55.029
627611 94.113 -54.518 55.027
627612 94.128 -54.526 55.025
627613 94.144 -54.535 55.019
627614 94.159 -54.543 55.010
627615 94.175 -54.552 54.999
627616 94.193 -54.562 54.987
627617 94.217 -54.579 54.975
627618 94.242 -54.596 54.962
627619 94.267 -54.612 54.944
627620 94.291 -54.629 54.924
627621 94.316 -54.645 54.902
627622 94.341 -54.662 54.881
627623 94.366 -54.678 54.860
627624 94.391 -54.695 54.834
627625 94.416 -54.711 54.802
627626 94.453 -54.723 54.769
627627 94.490 -54.735 54.734
627628 94.528 -54.747 54.697
627629 94.565 -54.759 54.659
627630 94.602 -54.772 54.625
627631 94.639 -54.784 54.585
627632 94.676 -54.796 54.543
627633 94.713 -54.808 54.500
627634 94.751 -54.820 54.456
627635 94.790 -54.844 54.410
627636 94.828 -54.866 54.360
627637 94.861 -54.882 54.303
627638 94.895 -54.899 54.245
627639 94.928 -54.916 54.191
627640 94.962 -54.933 54.142
627641 94.996 -54.950 54.094
627642 95.030 -54.966 54.044
627643 95.063 -54.983 53.988
627644 95.097 -55.000 53.931
627645 95.131 -55.016 53.874
627646 95.165 -55.033 53.818
627647 95.196 -55.048 53.762
627648 95.228 -55.063 53.709
627649 95.260 -55.078 53.652
627650 95.291 -55.093 53.593
627651 95.323 -55.107 53.536
627652 95.355 -55.123 53.481
627653 95.386 -55.137 53.427
627654 95.418 -55.152 53.376
627655 95.450 -55.167 53.328
627656 95.481 -55.182 53.285
627657 95.513 -55.197 53.245
627658 95.545 -55.212 53.205
627659 95.577 -55.226 53.164
627660 95.608 -55.242 53.125
627661 95.639 -55.254 53.085
627662 95.669 -55.265 53.043
627663 95.699 -55.276 52.996
627664 95.729 -55.286 52.949
627665 95.759 -55.297 52.902
627666 95.790 -55.307 52.855
627667 95.819 -55.317 52.807
627668 95.849 -55.328 52.757
627669 95.880 -55.339 52.704
627670 95.910 -55.349 52.649
627671 95.943 -55.362 52.592
627672 95.977 -55.376 52.532
627673 96.010 -55.389 52.468
627674 96.041 -55.398 52.404
627675 96.073 -55.408 52.337
627676 96.104 -55.419 52.272
627677 96.135 -55.429 52.205
627678 96.166 -55.439 52.140
627679 96.197 -55.449 52.076
627680 96.229 -55.459 52.013
627681 96.260 -55.469 51.949
627682 96.291 -55.480 51.884
627683 96.322 -55.489 51.819
627684 96.353 -55.499 51.755
627685 96.384 -55.509 51.691
627686 96.416 -55.522 51.628
627687 96.451 -55.546 51.564
627688 96.486 -55.571 51.500
627689 96.521 -55.595 51.433
627690 96.556 -55.619 51.369
627691 96.591 -55.643 51.305
627692 96.626 -55.668 51.241
627693 96.661 -55.692 51.177
627694 96.678 -55.691 51.111
627695 96.694 -55.690 51.046
627696 96.711 -55.690 50.984
628099 87.658 -50.903 56.415
628100 87.695 -50.928 56.433
628101 87.731 -50.952 56.450
628102 87.768 -50.977 56.465
628103 87.733 -50.948 56.480
628104 87.764 -50.970 56.494
628105 87.795 -50.994 56.503
628106 87.826 -51.017 56.506
628107 87.857 -51.041 56.505
628108 87.888 -51.064 56.500
628109 87.923 -51.081 56.493
628110 87.960 -51.099 56.483
628111 87.967 -51.144 56.470
628112 87.951 -51.209 56.452
628113 87.982 -51.232 56.431
628114 88.014 -51.255 56.410
628115 88.046 -51.277 56.387
628116 88.077 -51.300 56.362
628176 90.058 -52.395 57.110
628177 90.090 -52.403 57.116
628178 90.122 -52.411 57.122
628179 90.155 -52.419 57.128
628180 90.187 -52.427 57.135
628181 90.219 -52.435 57.142
628182 90.251 -52.443 57.151
628183 90.284 -52.451 57.158
628184 90.316 -52.459 57.165
628185 90.339 -52.474 57.173
628186 90.356 -52.497 57.182
628187 90.372 -52.519 57.190
628188 90.389 -52.540 57.197
628189 90.421 -52.540 57.201
628190 90.453 -52.540 57.204
628191 90.485 -52.540 57.206
628192 90.517 -52.540 57.207
628193 90.543 -52.557 57.207
628194 90.569 -52.580 57.204
628195 90.594 -52.602 57.198
628196 90.619 -52.625 57.190
628197 90.645 -52.647 57.182
628198 90.670 -52.670 57.175
628199 90.695 -52.692 57.170
628200 90.720 -52.715 57.168
628201 90.745 -52.738 57.165
628202 90.783 -52.747 57.162
628203 90.822 -52.756 57.160
628204 90.862 -52.764 57.156
628205 90.901 -52.772 57.150
628206 90.940 -52.780 57.142
628207 90.980 -52.788 57.132
628208 91.019 -52.797 57.123
628209 91.059 -52.805 57.115
628210 91.098 -52.813 57.104
628211 91.126 -52.832 57.080
628212 91.141 -52.863 57.057
628213 91.155 -52.893 57.034
628214 91.161 -52.917 57.010
628215 91.177 -52.931 56.986
628216 91.193 -52.944 56.964
628217 91.210 -52.958 56.941
628218 91.226 -52.972 56.919
628219 91.253 -52.993 56.892
628220 91.287 -53.017 56.864
628221 91.320 -53.042 56.832
628222 91.354 -53.067 56.799
628223 91.387 -53.091 56.764
628224 91.420 -53.115 56.730
628225 91.453 -53.140 56.694
628226 91.487 -53.164 56.656
628227 91.520 -53.189 56.619
628228 91.554 -53.214 56.583
628229 91.587 -53.238 56.545
628230 91.620 -53.262 56.510
628231 91.654 -53.287 56.474
628232 91.687 -53.311 56.438
628233 91.720 -53.336 56.403
628234 91.753 -53.361 56.368
628235 91.787 -53.385 56.334
628236 91.820 -53.409 56.298
628237 91.854 -53.434 56.262
628238 91.887 -53.458 56.225
628239 91.920 -53.483 56.188
628240 91.955 -53.505 56.151
628241 91.996 -53.518 56.113
628242 92.037 -53.531 56.074
628243 92.078 -53.545 56.035
628244 92.119 -53.559 55.996
628245 92.159 -53.572 55.959
628246 92.200 -53.586 55.921
628247 92.240 -53.599 55.883
628248 92.281 -53.612 55.847
628249 92.322 -53.626 55.813
628250 92.363 -53.640 55.781
628263 92.893 -53.815 55.242
628264 92.933 -53.828 55.229
628265 92.974 -53.842 55.215
628266 93.020 -53.850 55.203
628267 93.068 -53.856 55.193
628268 93.116 -53.862 55.183
628269 93.164 -53.868 55.173
628270 93.189 -53.887 55.162
628271 93.208 -53.909 55.151
628272 93.227 -53.931 55.140
628273 93.247 -53.953 55.130
628274 93.267 -53.975 55.123
628275 93.286 -53.997 55.115
628276 93.305 -54.018 55.108
628277 93.325 -54.040 55.101
628278 93.345 -54.062 55.095
628279 93.364 -54.084 55.093
628280 93.383 -54.106 55.091
628281 93.403 -54.128 55.092
628282 93.423 -54.150 55.094
628283 93.442 -54.172 55.096
628284 93.467 -54.189 55.097
628285 93.498 -54.197 55.096
628286 93.529 -54.204 55.096
628287 93.560 -54.213 55.095
628288 93.591 -54.220 55.095
628289 93.622 -54.229 55.098
628290 93.653 -54.236 55.102
628291 93.685 -54.245 55.106
628292 93.715 -54.252 55.111
628293 93.747 -54.261 55.116
628294 93.778 -54.268 55.123
628295 93.799 -54.282 55.127
628296 93.821 -54.295 55.128
628297 93.842 -54.308 55.127
628298 93.863 -54.321 55.123
628299 93.885 -54.335 55.121
628300 93.906 -54.348 55.123
628301 93.928 -54.361 55.126
628302 93.949 -54.374 55.128
628303 93.971 -54.388 55.124
628304 93.992 -54.401 55.114
628305 94.014 -54.414 55.097
628306 94.035 -54.428 55.077
628307 94.057 -54.441 55.058
628308 94.074 -54.451 55.042
628309 94.089 -54.460 55.031
628310 94.105 -54.468 55.026
628311 94.121 -54.477 55.021
628312 94.136 -54.486 55.017
628313 94.152 -54.495 55.013
628314 94.168 -54.503 55.006
628315 94.183 -54.512 54.997
628316 94.199 -54.521 54.987
628317 94.214 -54.529 54.976
628318 94.234 -54.541 54.967
628319 94.258 -54.558 54.955
628320 94.283 -54.574 54.939
628321 94.308 -54.591 54.918
628322 94.333 -54.607 54.895
628323 94.357 -54.624 54.873
628324 94.382 -54.640 54.852
628325 94.407 -54.657 54.824
628326 94.431 -54.674 54.790
628327 94.468 -54.686 54.754
628328 94.505 -54.698 54.717
628329 94.542 -54.710 54.678
628330 94.579 -54.722 54.638
628331 94.617 -54.734 54.601
628332 94.654 -54.746 54.561
628333 94.691 -54.759 54.518
628334 94.728 -54.771 54.473
628335 94.767 -54.788 54.427
628336 94.807 -54.815 54.379
628337 94.844 -54.836 54.331
628338 94.878 -54.853 54.278
628339 94.911 -54.869 54.227
628340 94.945 -54.886 54.178
628341 94.978 -54.903 54.131
628342 95.012 -54.920 54.085
628343 95.046 -54.937 54.034
628344 95.079 -54.953 53.976
628345 95.113 -54.970 53.919
628346 95.147 -54.987 53.864
628347 95.178 -55.003 53.811
628348 95.209 -55.018 53.758
628349 95.241 -55.032 53.703
628350 95.272 -55.047 53.646
628351 95.304 -55.062 53.587
628352 95.336 -55.077 53.530
628353 95.368 -55.091 53.475
628354 95.399 -55.107 53.422
628355 95.431 -55.122 53.374
628356 95.462 -55.136 53.333
628357 95.494 -55.151 53.293
628358 95.526 -55.166 53.255
628359 95.557 -55.181 53.215
628360 95.589 -55.196 53.174
628361 95.621 -55.210 53.133
628362 95.651 -55.222 53.092
628363 95.681 -55.232 53.050
628364 95.711 -55.243 53.006
628365 95.741 -55.253 52.962
628366 95.771 -55.263 52.917
628367 95.801 -55.274 52.871
628368 95.831 -55.285 52.823
628369 95.862 -55.295 52.775
628370 95.891 -55.306 52.723
628371 95.921 -55.317 52.667
628372 95.952 -55.326 52.608
628373 95.983 -55.337 52.545
628374 96.014 -55.347 52.481
628375 96.045 -55.357 52.416
628376 96.077 -55.367 52.349
628377 96.108 -55.377 52.285
628378 96.139 -55.387 52.220
628379 96.170 -55.397 52.155
628380 96.201 -55.408 52.091
628381 96.233 -55.417 52.027
628382 96.264 -55.427 51.963
628383 96.295 -55.438 51.898
628384 96.326 -55.448 51.833
628385 96.358 -55.458 51.768
628386 96.389 -55.467 51.704
628387 96.421 -55.484 51.640
628388 96.457 -55.508 51.575
628389 96.491 -55.533 51.510
628390 96.527 -55.557 51.445
628391 96.561 -55.582 51.380
628392 96.597 -55.606 51.316
628393 96.631 -55.630 51.252
628394 96.667 -55.655 51.188
628395 96.692 -55.666 51.122
628396 96.709 -55.665 51.058
628800 87.676 -50.873 56.434
628801 87.712 -50.898 56.452
628802 87.748 -50.923 56.469
628803 87.784 -50.947 56.485
628804 87.821 -50.972 56.501
628805 87.798 -50.951 56.514
628806 87.817 -50.964 56.525
628807 87.848 -50.988 56.528
628808 87.879 -51.011 56.527
628809 87.912 -51.030 56.522
628810 87.946 -51.052 56.514
628811 87.928 -51.118 56.501
628812 87.935 -51.163 56.487
628813 87.967 -51.186 56.468
628814 87.999 -51.208 56.445
628815 88.031 -51.228 56.421
628816 88.064 -51.248 56.396
628817 88.098 -51.267 56.370
628878 90.110 -52.370 57.117
628879 90.143 -52.377 57.123
628880 90.175 -52.386 57.130
628881 90.207 -52.393 57.138
628882 90.239 -52.402 57.146
628883 90.271 -52.409 57.155
628884 90.304 -52.418 57.163
628885 90.336 -52.426 57.169
628886 90.368 -52.434 57.178
628887 90.387 -52.454 57.186
628888 90.403 -52.476 57.193
628889 90.419 -52.498 57.200
628890 90.444 -52.507 57.204
628891 90.476 -52.507 57.206
628892 90.509 -52.507 57.208
628893 90.541 -52.507 57.208
628894 90.569 -52.520 57.206
628895 90.594 -52.543 57.203
628896 90.619 -52.565 57.198
628897 90.644 -52.587 57.190
628898 90.669 -52.610 57.182
628899 90.694 -52.633 57.175
628900 90.720 -52.655 57.170
628901 90.745 -52.678 57.167
628902 90.773 -52.700 57.164
628903 90.805 -52.719 57.161
628904 90.834 -52.732 57.160
628905 90.873 -52.741 57.157
628906 90.912 -52.749 57.154
628907 90.952 -52.757 57.152
628908 90.992 -52.765 57.148
628909 91.031 -52.773 57.143
628910 91.071 -52.782 57.139
628911 91.110 -52.790 57.125
628912 91.134 -52.813 57.104
628913 91.148 -52.843 57.080
628914 91.163 -52.873 57.056
628915 91.187 -52.897 57.033
628916 91.207 -52.913 57.011
628917 91.223 -52.926 56.990
628918 91.239 -52.940 56.970
628919 91.256 -52.954 56.950
628920 91.272 -52.967 56.926
628921 91.301 -52.988 56.897
628922 91.334 -53.013 56.865
628923 91.367 -53.038 56.832
628924 91.401 -53.062 56.797
628925 91.434 -53.087 56.764
628926 91.467 -53.111 56.728
628927 91.501 -53.136 56.690
628928 91.534 -53.161 56.654
628929 91.568 -53.185 56.617
628930 91.601 -53.209 56.579
628931 91.634 -53.234 56.543
628932 91.667 -53.258 56.505
628933 91.701 -53.283 56.468
628934 91.734 -53.308 56.432
628935 91.768 -53.332 56.396
628936 91.801 -53.356 56.360
628937 91.834 -53.381 56.323
628938 91.867 -53.405 56.286
628939 91.900 -53.430 56.248
628940 91.934 -53.455 56.210
628941 91.969 -53.476 56.170
628942 92.010 -53.490 56.130
628943 92.051 -53.502 56.089
628944 92.092 -53.516 56.048
628945 92.133 -53.530 56.009
628946 92.173 -53.543 55.970
628947 92.214 -53.557 55.931
628948 92.255 -53.570 55.892
628949 92.295 -53.584 55.855
628950 92.336 -53.597 55.820
628951 92.377 -53.611 55.786
628964 92.907 -53.786 55.235
628965 92.947 -53.800 55.221
628966 92.995 -53.806 55.207
628967 93.043 -53.812 55.192
628968 93.091 -53.818 55.181
628969 93.139 -53.824 55.170
628970 93.174 -53.837 55.158
628971 93.194 -53.859 55.146
628972 93.213 -53.881 55.134
628973 93.232 -53.903 55.122
628974 93.252 -53.925 55.113
628975 93.272 -53.947 55.104
628976 93.291 -53.969 55.097
628977 93.310 -53.991 55.089
628978 93.330 -54.012 55.082
628979 93.350 -54.034 55.077
628980 93.369 -54.056 55.075
628981 93.389 -54.078 55.073
628982 93.408 -54.100 55.073
628983 93.428 -54.122 55.074
628984 93.452 -54.140 55.076
628985 93.482 -54.154 55.076
628986 93.515 -54.169 55.075
628987 93.550 -54.179 55.075
628988 93.581 -54.188 55.075
628989 93.612 -54.195 55.075
628990 93.643 -54.204 55.077
628991 93.674 -54.211 55.081
628992 93.705 -54.220 55.085
628993 93.736 -54.227 55.089
628994 93.767 -54.236 55.095
628995 93.798 -54.243 55.100
628996 93.827 -54.253 55.103
628997 93.848 -54.266 55.106
628998 93.869 -54.279 55.107
628999 93.891 -54.292 55.106
629000 93.912 -54.306 55.103
629001 93.934 -54.320 55.101
629002 93.955 -54.333 55.101
629003 93.977 -54.346 55.098
629004 93.998 -54.359 55.092
629005 94.019 -54.373 55.083
629006 94.041 -54.386 55.069
629007 94.062 -54.399 55.054
629008 94.082 -54.411 55.041
629009 94.098 -54.420 55.032
629010 94.113 -54.428 55.026
629011 94.129 -54.437 55.022
629012 94.144 -54.446 55.017
629013 94.160 -54.455 55.013
629014 94.176 -54.463 55.009
629015 94.191 -54.472 55.003
629016 94.207 -54.480 54.996
629017 94.222 -54.489 54.988
629018 94.238 -54.498 54.979
629019 94.254 -54.506 54.969
629020 94.275 -54.520 54.955
629021 94.299 -54.536 54.937
629022 94.324 -54.553 54.913
629023 94.349 -54.570 54.888
629024 94.373 -54.586 54.865
629025 94.398 -54.602 54.842
629026 94.423 -54.619 54.813
629027 94.448 -54.636 54.778
629028 94.482 -54.649 54.739
629029 94.520 -54.661 54.700
629030 94.557 -54.673 54.660
629031 94.594 -54.685 54.618
629032 94.631 -54.697 54.580
629033 94.669 -54.709 54.538
629034 94.706 -54.721 54.494
629035 94.743 -54.734 54.448
629036 94.783 -54.759 54.402
629037 94.823 -54.786 54.355
629038 94.860 -54.806 54.310
629039 94.894 -54.823 54.261
629040 94.927 -54.840 54.213
629041 94.961 -54.856 54.166
629042 94.994 -54.874 54.118
629043 95.028 -54.891 54.068
629044 95.062 -54.907 54.015
629045 95.096 -54.924 53.957
629046 95.129 -54.941 53.900
629047 95.161 -54.957 53.846
629048 95.191 -54.974 53.794
629049 95.222 -54.990 53.742
629050 95.253 -55.003 53.691
629051 95.285 -55.018 53.636
629052 95.317 -55.032 53.579
629053 95.348 -55.046 53.523
629054 95.380 -55.061 53.469
629055 95.412 -55.075 53.419
629056 95.444 -55.091 53.375
629057 95.475 -55.106 53.338
629058 95.507 -55.120 53.302
629059 95.538 -55.135 53.264
629060 95.570 -55.151 53.224
629061 95.602 -55.165 53.183
629062 95.633 -55.179 53.142
629063 95.663 -55.189 53.099
629064 95.693 -55.200 53.058
629065 95.723 -55.210 53.018
629066 95.753 -55.220 52.977
629067 95.783 -55.231 52.933
629068 95.813 -55.242 52.887
629069 95.843 -55.252 52.838
629070 95.873 -55.263 52.790
629071 95.902 -55.275 52.738
629072 95.932 -55.287 52.682
629073 95.961 -55.297 52.622
629074 95.990 -55.301 52.560
629075 96.019 -55.306 52.495
629076 96.049 -55.315 52.430
629077 96.081 -55.326 52.365
629078 96.112 -55.336 52.302
629079 96.143 -55.345 52.237
629080 96.174 -55.356 52.173
629081 96.205 -55.366 52.109
629082 96.237 -55.376 52.045
629083 96.268 -55.386 51.981
629084 96.299 -55.396 51.916
629085 96.330 -55.406 51.850
629086 96.362 -55.416 51.785
629087 96.393 -55.426 51.720
629088 96.427 -55.447 51.656
629089 96.462 -55.471 51.591
629090 96.497 -55.495 51.525
629091 96.532 -55.520 51.460
629092 96.567 -55.544 51.395
629093 96.602 -55.568 51.331
629094 96.637 -55.593 51.266
629095 96.672 -55.617 51.203
629096 96.707 -55.641 51.138
629097 96.723 -55.640 51.072
629502 87.728 -50.868 56.469
629503 87.765 -50.892 56.487
629504 87.801 -50.917 56.504
629505 87.837 -50.942 56.519
629506 87.873 -50.967 56.532
629507 87.865 -50.957 56.548
629508 87.869 -50.958 56.554
629509 87.901 -50.980 56.550
629510 87.907 -51.026 56.541
629511 87.890 -51.092 56.528
629512 87.920 -51.116 56.515
629513 87.952 -51.137 56.500
629514 87.985 -51.156 56.479
629515 88.019 -51.176 56.456
629516 88.051 -51.195 56.431
629517 88.084 -51.215 56.404
629518 88.118 -51.233 56.376
629519 88.150 -51.253 56.348
629581 90.194 -52.352 57.132
629582 90.227 -52.360 57.141
629583 90.259 -52.368 57.149
629584 90.292 -52.376 57.158
629585 90.324 -52.384 57.167
629586 90.356 -52.392 57.172
629587 90.388 -52.400 57.181
629588 90.418 -52.411 57.189
629589 90.434 -52.433 57.195
629590 90.450 -52.455 57.202
629591 90.468 -52.474 57.206
629592 90.500 -52.474 57.207
629593 90.532 -52.474 57.208
629594 90.565 -52.474 57.207
629595 90.594 -52.483 57.204
629596 90.619 -52.506 57.202
629597 90.644 -52.528 57.197
629598 90.669 -52.550 57.189
629599 90.694 -52.573 57.183
629600 90.720 -52.596 57.175
629601 90.745 -52.618 57.170
629602 90.770 -52.641 57.167
629603 90.805 -52.660 57.164
629604 90.845 -52.678 57.161
629605 90.863 -52.701 57.159
629606 90.885 -52.717 57.158
629607 90.925 -52.725 57.160
629608 90.964 -52.734 57.165
629609 91.003 -52.742 57.167
629610 91.043 -52.750 57.164
629611 91.082 -52.758 57.161
629612 91.121 -52.766 57.145
629613 91.141 -52.793 57.124
629614 91.156 -52.823 57.101
629615 91.176 -52.850 57.078
629616 91.204 -52.870 57.057
629617 91.231 -52.891 57.037
629618 91.253 -52.907 57.021
629619 91.270 -52.922 57.001
629620 91.286 -52.935 56.985
629621 91.302 -52.949 56.962
629622 91.318 -52.963 56.931
629623 91.348 -52.985 56.895
629624 91.381 -53.009 56.864
629625 91.415 -53.034 56.829
629626 91.448 -53.058 56.797
629627 91.482 -53.082 56.761
629628 91.515 -53.107 56.724
629629 91.548 -53.132 56.687
629630 91.581 -53.156 56.650
629631 91.614 -53.181 56.612
629632 91.648 -53.205 56.575
629633 91.681 -53.230 56.537
629634 91.715 -53.255 56.499
629635 91.748 -53.279 56.461
629636 91.781 -53.303 56.424
629637 91.814 -53.328 56.385
629638 91.848 -53.352 56.347
629639 91.881 -53.377 56.308
629640 91.915 -53.402 56.270
629641 91.948 -53.426 56.230
629642 91.984 -53.447 56.187
629643 92.024 -53.460 56.146
629644 92.065 -53.474 56.104
629645 92.106 -53.487 56.062
629646 92.147 -53.501 56.021
629647 92.188 -53.515 55.981
629648 92.228 -53.527 55.940
629649 92.269 -53.541 55.901
629650 92.310 -53.555 55.863
629651 92.350 -53.568 55.826
629652 92.391 -53.582 55.791
629666 92.970 -53.762 55.215
629667 93.018 -53.768 55.199
629668 93.066 -53.774 55.183
629669 93.114 -53.780 55.170
629670 93.160 -53.787 55.157
629671 93.179 -53.809 55.144
629672 93.199 -53.831 55.132
629673 93.218 -53.853 55.119
629674 93.238 -53.875 55.106
629675 93.257 -53.897 55.096
629676 93.277 -53.919 55.087
629677 93.296 -53.941 55.079
629678 93.316 -53.963 55.072
629679 93.335 -53.985 55.064
629680 93.355 -54.006 55.058
629681 93.374 -54.028 55.057
629682 93.394 -54.050 55.054
629683 93.413 -54.072 55.054
629684 93.438 -54.090 55.054
629685 93.469 -54.104 55.055
629686 93.498 -54.119 55.055
629687 93.529 -54.134 55.054
629688 93.562 -54.148 55.055
629689 93.600 -54.161 55.055
629690 93.632 -54.171 55.056
629691 93.663 -54.179 55.059
629692 93.694 -54.187 55.062
629693 93.725 -54.195 55.066
629694 93.757 -54.203 55.070
629695 93.787 -54.211 55.074
629696 93.819 -54.219 55.078
629697 93.850 -54.226 55.083
629698 93.875 -54.238 55.087
629699 93.897 -54.251 55.090
629700 93.918 -54.264 55.090
629701 93.939 -54.277 55.084
629702 93.961 -54.291 55.079
629703 93.982 -54.304 55.075
629704 94.004 -54.317 55.069
629705 94.026 -54.330 55.062
629706 94.047 -54.344 55.055
629707 94.068 -54.357 55.044
629708 94.089 -54.370 55.035
629709 94.106 -54.380 55.027
629710 94.121 -54.388 55.022
629711 94.137 -54.397 55.019
629712 94.152 -54.405 55.018
629713 94.168 -54.414 55.014
629714 94.184 -54.423 55.009
629715 94.199 -54.432 55.006
629716 94.215 -54.440 55.002
629717 94.230 -54.449 54.996
629718 94.246 -54.458 54.988
629719 94.262 -54.466 54.979
629720 94.277 -54.475 54.965
629721 94.292 -54.483 54.949
629722 94.316 -54.499 54.928
629723 94.340 -54.515 54.904
629724 94.365 -54.532 54.878
629725 94.390 -54.548 54.854
629726 94.414 -54.565 54.829
629727 94.439 -54.581 54.800
629728 94.464 -54.598 54.765
629729 94.497 -54.612 54.725
629730 94.534 -54.624 54.684
629731 94.571 -54.636 54.642
629732 94.609 -54.648 54.599
629733 94.646 -54.660 54.559
629734 94.683 -54.672 54.517
629735 94.720 -54.684 54.471
629736 94.759 -54.703 54.425
629737 94.800 -54.730 54.380
629738 94.840 -54.756 54.335
629739 94.876 -54.777 54.292
629740 94.909 -54.793 54.245
629741 94.943 -54.810 54.198
629742 94.977 -54.827 54.150
629743 95.011 -54.844 54.101
629744 95.044 -54.861 54.049
629745 95.078 -54.878 53.995
629746 95.112 -54.894 53.937
629747 95.143 -54.911 53.880
629748 95.174 -54.928 53.826
629749 95.204 -54.944 53.774
629750 95.234 -54.960 53.722
629751 95.265 -54.975 53.673
629752 95.297 -54.989 53.620
629753 95.329 -55.003 53.567
629754 95.361 -55.017 53.514
629755 95.393 -55.032 53.465
629756 95.424 -55.046 53.420
629757 95.456 -55.060 53.381
629758 95.488 -55.075 53.346
629759 95.520 -55.090 53.309
629760 95.551 -55.104 53.271
629761 95.583 -55.119 53.233
629762 95.614 -55.135 53.193
629763 95.645 -55.146 53.154
629764 95.675 -55.157 53.112
629765 95.705 -55.166 53.072
629766 95.735 -55.177 53.031
629767 95.765 -55.188 52.990
629768 95.795 -55.200 52.947
629769 95.824 -55.211 52.902
629770 95.853 -55.223 52.852
629771 95.883 -55.235 52.802
629772 95.912 -55.247 52.748
629773 95.941 -55.259 52.692
629774 95.971 -55.269 52.635
629775 96.000 -55.273 52.575
629776 96.028 -55.278 52.510
629777 96.057 -55.282 52.447
629778 96.086 -55.286 52.385
629779 96.116 -55.294 52.321
629780 96.148 -55.304 52.257
629781 96.179 -55.314 52.193
629782 96.210 -55.324 52.129
629783 96.241 -55.334 52.065
629784 96.272 -55.344 52.000
629785 96.304 -55.354 51.934
629786 96.334 -55.364 51.868
629787 96.366 -55.374 51.803
629788 96.397 -55.385 51.738
629789 96.432 -55.409 51.673
629790 96.467 -55.434 51.607
629791 96.502 -55.458 51.541
629792 96.537 -55.483 51.476
629793 96.572 -55.507 51.411
629794 96.607 -55.531 51.347
629795 96.642 -55.555 51.282
629796 96.677 -55.580 51.219
629797 96.712 -55.604 51.153
629798 96.738 -55.615 51.088
630203 87.746 -50.838 56.483
630204 87.782 -50.863 56.501
630205 87.818 -50.887 56.518
630206 87.854 -50.912 56.534
630207 87.891 -50.936 56.548
630208 87.927 -50.961 56.570
630209 87.922 -50.972 56.577
630210 87.868 -51.000 56.570
630211 87.873 -51.046 56.555
630212 87.906 -51.065 56.540
630213 87.939 -51.085 56.525
630214 87.972 -51.104 56.509
630215 88.005 -51.123 56.487
630216 88.038 -51.142 56.462
630217 88.071 -51.161 56.436
630218 88.105 -51.181 56.408
630219 88.137 -51.200 56.380
630220 88.171 -51.220 56.350
630283 90.247 -52.327 57.142
630284 90.279 -52.335 57.150
630285 90.311 -52.343 57.159
630286 90.344 -52.351 57.168
630287 90.376 -52.358 57.176
630288 90.408 -52.367 57.183
630289 90.441 -52.375 57.190
630290 90.465 -52.390 57.197
630291 90.481 -52.412 57.202
630292 90.498 -52.434 57.206
630293 90.524 -52.441 57.207
630294 90.556 -52.441 57.207
630295 90.588 -52.441 57.205
630296 90.619 -52.446 57.202
630297 90.644 -52.468 57.200
630298 90.669 -52.490 57.196
630299 90.694 -52.513 57.189
630300 90.720 -52.536 57.183
630301 90.745 -52.559 57.176
630302 90.770 -52.581 57.171
630303 90.795 -52.603 57.169
630304 90.835 -52.622 57.166
630305 90.876 -52.640 57.162
630306 90.917 -52.657 57.163
630307 90.921 -52.683 57.166
630308 90.936 -52.702 57.173
630309 90.975 -52.710 57.183
630310 91.015 -52.719 57.189
630311 91.054 -52.727 57.188
630312 91.094 -52.735 57.180
630313 91.134 -52.743 57.163
630314 91.149 -52.773 57.142
630315 91.164 -52.803 57.121
630316 91.192 -52.824 57.101
630317 91.220 -52.844 57.081
630318 91.248 -52.865 57.064
630319 91.276 -52.885 57.051
630320 91.300 -52.903 57.035
630321 91.316 -52.917 57.020
630322 91.332 -52.931 56.996
630323 91.349 -52.944 56.962
630324 91.365 -52.958 56.925
630325 91.395 -52.981 56.892
630326 91.429 -53.005 56.860
630327 91.462 -53.030 56.827
630328 91.495 -53.054 56.792
630329 91.528 -53.079 56.756
630330 91.562 -53.104 56.720
630331 91.595 -53.128 56.683
630332 91.629 -53.152 56.644
630333 91.662 -53.177 56.607
630334 91.695 -53.201 56.569
630335 91.728 -53.226 56.529
630336 91.762 -53.251 56.490
630337 91.795 -53.275 56.451
630338 91.828 -53.299 56.411
630339 91.862 -53.324 56.371
630340 91.895 -53.349 56.331
630341 91.929 -53.373 56.290
630342 91.962 -53.398 56.248
630343 91.998 -53.418 56.205
630344 92.038 -53.431 56.162
630345 92.079 -53.445 56.118
630346 92.120 -53.458 56.075
630347 92.161 -53.472 56.033
630348 92.202 -53.486 55.991
630349 92.242 -53.499 55.949
630350 92.283 -53.512 55.910
630351 92.324 -53.526 55.871
630352 92.365 -53.540 55.833
630353 92.405 -53.553 55.797
630368 93.041 -53.730 55.192
630369 93.089 -53.736 55.175
630370 93.137 -53.742 55.161
630371 93.165 -53.759 55.147
630372 93.185 -53.781 55.133
630373 93.204 -53.803 55.119
630374 93.223 -53.825 55.105
630375 93.243 -53.847 55.092
630376 93.263 -53.869 55.081
630377 93.282 -53.891 55.071
630378 93.301 -53.913 55.063
630379 93.321 -53.935 55.055
630380 93.340 -53.957 55.047
630381 93.360 -53.979 55.041
630382 93.379 -54.000 55.039
630383 93.399 -54.022 55.036
630384 93.424 -54.040 55.035
630385 93.454 -54.054 55.035
630386 93.484 -54.069 55.034
630387 93.514 -54.084 55.034
630388 93.545 -54.098 55.034
630389 93.574 -54.113 55.036
630390 93.610 -54.127 55.036
630391 93.647 -54.141 55.038
630392 93.684 -54.154 55.041
630393 93.715 -54.162 55.043
630394 93.746 -54.170 55.048
630395 93.777 -54.178 55.051
630396 93.808 -54.185 55.055
630397 93.839 -54.194 55.059
630398 93.870 -54.201 55.064
630399 93.901 -54.210 55.068
630400 93.924 -54.222 55.071
630401 93.945 -54.236 55.072
630402 93.967 -54.248 55.067
630403 93.988 -54.262 55.063
630404 94.009 -54.276 55.058
630405 94.031 -54.289 55.053
630406 94.053 -54.302 55.044
630407 94.074 -54.315 55.037
630408 94.096 -54.329 55.029
630409 94.114 -54.339 55.022
630410 94.129 -54.348 55.017
630411 94.145 -54.357 55.013
630412 94.160 -54.365 55.012
630413 94.176 -54.374 55.012
630414 94.192 -54.383 55.010
630415 94.207 -54.392 55.006
630416 94.223 -54.400 55.003
630417 94.239 -54.408 55.000
630418 94.254 -54.417 54.995
630419 94.270 -54.426 54.987
630420 94.285 -54.435 54.976
630421 94.300 -54.443 54.960
630422 94.316 -54.452 54.942
630423 94.332 -54.461 54.920
630424 94.356 -54.477 54.894
630425 94.382 -54.494 54.869
630426 94.406 -54.511 54.844
630427 94.431 -54.527 54.817
630428 94.456 -54.543 54.789
630429 94.480 -54.560 54.753
630430 94.512 -54.574 54.713
630431 94.549 -54.587 54.673
630432 94.586 -54.598 54.632
630433 94.624 -54.610 54.590
630434 94.661 -54.622 54.548
630435 94.698 -54.634 54.501
630436 94.736 -54.647 54.452
630437 94.776 -54.674 54.405
630438 94.816 -54.701 54.360
630439 94.856 -54.728 54.318
630440 94.892 -54.747 54.275
630441 94.925 -54.764 54.228
630442 94.959 -54.781 54.179
630443 94.993 -54.797 54.129
630444 95.027 -54.814 54.079
630445 95.060 -54.831 54.027
630446 95.094 -54.848 53.975
630447 95.126 -54.865 53.917
630448 95.157 -54.881 53.860
630449 95.187 -54.897 53.806
630450 95.217 -54.914 53.754
630451 95.247 -54.931 53.703
630452 95.278 -54.946 53.655
630453 95.310 -54.960 53.606
630454 95.341 -54.975 53.556
630455 95.373 -54.988 53.509
630456 95.405 -55.003 53.465
630457 95.437 -55.017 53.424
630458 95.469 -55.031 53.388
630459 95.500 -55.045 53.353
630460 95.532 -55.060 53.317
630461 95.564 -55.074 53.281
630462 95.596 -55.089 53.243
630463 95.627 -55.103 53.205
630464 95.657 -55.113 53.167
630465 95.687 -55.124 53.126
630466 95.716 -55.135 53.085
630467 95.745 -55.148 53.044
630468 95.775 -55.160 53.002
630469 95.804 -55.171 52.958
630470 95.833 -55.183 52.911
630471 95.863 -55.195 52.862
630472 95.892 -55.207 52.812
630473 95.921 -55.218 52.759
630474 95.951 -55.230 52.704
630475 95.980 -55.241 52.648
630476 96.009 -55.245 52.590
630477 96.038 -55.249 52.530
630478 96.067 -55.254 52.472
630479 96.096 -55.258 52.411
630480 96.125 -55.263 52.348
630481 96.154 -55.267 52.283
630482 96.183 -55.273 52.218
630483 96.214 -55.282 52.152
630484 96.245 -55.292 52.087
630485 96.276 -55.303 52.021
630486 96.308 -55.313 51.955
630487 96.339 -55.323 51.888
630488 96.370 -55.332 51.822
630489 96.402 -55.348 51.756
630490 96.438 -55.372 51.691
630491 96.472 -55.396 51.625
630492 96.508 -55.420 51.559
630493 96.542 -55.445 51.493
630494 96.578 -55.469 51.428
630495 96.612 -55.493 51.364
630496 96.648 -55.517 51.299
630497 96.682 -55.542 51.235
630498 96.718 -55.567 51.169
630499 96.752 -55.590 51.105
630904 87.762 -50.807 56.493
630905 87.798 -50.832 56.511
630906 87.835 -50.857 56.529
630907 87.871 -50.882 56.547
630908 87.907 -50.907 56.560
630909 87.918 -50.945 56.573
630910 87.903 -50.997 56.581
630911 87.885 -51.041 56.577
630912 87.900 -51.026 56.563
630913 87.926 -51.032 56.548
630914 87.959 -51.052 56.533
630915 87.992 -51.070 56.514
630916 88.025 -51.090 56.492
630917 88.058 -51.109 56.466
630918 88.092 -51.129 56.438
630919 88.124 -51.148 56.409
630920 88.157 -51.168 56.380
630921 88.191 -51.186 56.349
630986 90.331 -52.309 57.161
630987 90.364 -52.318 57.169
630988 90.396 -52.325 57.177
630989 90.428 -52.333 57.184
630990 90.461 -52.342 57.190
630991 90.493 -52.349 57.196
630992 90.513 -52.368 57.202
630993 90.529 -52.390 57.205
630994 90.547 -52.409 57.205
630995 90.580 -52.409 57.205
630996 90.612 -52.409 57.203
630997 90.644 -52.409 57.200
630998 90.669 -52.431 57.198
630999 90.694 -52.453 57.195
631000 90.720 -52.476 57.189
631001 90.745 -52.499 57.184
631002 90.770 -52.521 57.177
631003 90.795 -52.543 57.174
631004 90.826 -52.565 57.173
631005 90.867 -52.583 57.171
631006 90.907 -52.600 57.166
631007 90.948 -52.619 57.168
631008 90.974 -52.639 57.174
631009 90.978 -52.665 57.184
631010 90.987 -52.687 57.194
631011 91.027 -52.695 57.198
631012 91.066 -52.703 57.194
631013 91.105 -52.712 57.185
631014 91.142 -52.722 57.172
631015 91.156 -52.753 57.158
631016 91.180 -52.777 57.143
631017 91.208 -52.797 57.127
631018 91.236 -52.818 57.109
631019 91.264 -52.838 57.092
631020 91.292 -52.859 57.082
631021 91.320 -52.879 57.069
631022 91.346 -52.898 57.054
631023 91.362 -52.913 57.028
631024 91.378 -52.926 56.991
631025 91.394 -52.940 56.954
631026 91.411 -52.954 56.921
631027 91.442 -52.977 56.889
631028 91.476 -53.001 56.857
631029 91.509 -53.026 56.823
631030 91.542 -53.051 56.789
631031 91.576 -53.075 56.753
631032 91.609 -53.099 56.716
631033 91.643 -53.124 56.677
631034 91.676 -53.148 56.639
631035 91.709 -53.173 56.599
631036 91.742 -53.198 56.559
631037 91.776 -53.222 56.518
631038 91.809 -53.246 56.477
631039 91.843 -53.271 56.436
631040 91.876 -53.295 56.394
631041 91.909 -53.320 56.352
631042 91.942 -53.345 56.310
631043 91.976 -53.369 56.266
631044 92.012 -53.389 56.222
631045 92.053 -53.402 56.177
631046 92.093 -53.416 56.132
631047 92.134 -53.430 56.088
631048 92.175 -53.443 56.044
631049 92.216 -53.456 56.002
631050 92.257 -53.470 55.959
631051 92.297 -53.483 55.918
631052 92.338 -53.497 55.878
631053 92.379 -53.511 55.840
631054 92.420 -53.524 55.803
631069 93.064 -53.692 55.185
631070 93.112 -53.698 55.167
631071 93.150 -53.709 55.151
631072 93.170 -53.731 55.136
631073 93.190 -53.753 55.121
631074 93.209 -53.775 55.106
631075 93.228 -53.797 55.091
631076 93.248 -53.819 55.077
631077 93.268 -53.841 55.066
631078 93.287 -53.863 55.056
631079 93.306 -53.885 55.047
631080 93.326 -53.907 55.039
631081 93.346 -53.928 55.031
631082 93.365 -53.950 55.024
631083 93.384 -53.972 55.021
631084 93.410 -53.990 55.019
631085 93.440 -54.004 55.018
631086 93.471 -54.019 55.017
631087 93.500 -54.033 55.016
631088 93.531 -54.048 55.015
631089 93.560 -54.063 55.017
631090 93.590 -54.077 55.018
631091 93.621 -54.091 55.019
631092 93.656 -54.106 55.021
631093 93.694 -54.120 55.023
631094 93.731 -54.133 55.026
631095 93.766 -54.145 55.030
631096 93.797 -54.153 55.033
631097 93.829 -54.160 55.037
631098 93.859 -54.169 55.042
631099 93.891 -54.176 55.046
631100 93.922 -54.185 55.050
631101 93.951 -54.194 55.053
631102 93.973 -54.207 55.055
631103 93.994 -54.220 55.052
631104 94.015 -54.233 55.049
631105 94.037 -54.247 55.045
631106 94.058 -54.260 55.040
631107 94.080 -54.273 55.033
631108 94.101 -54.286 55.027
631109 94.122 -54.299 55.022
631110 94.137 -54.308 55.018
631111 94.153 -54.317 55.013
631112 94.169 -54.325 55.008
631113 94.184 -54.334 55.004
631114 94.200 -54.342 55.004
631115 94.215 -54.352 55.001
631116 94.231 -54.360 54.998
631117 94.247 -54.368 54.997
631118 94.262 -54.377 54.994
631119 94.278 -54.386 54.989
631120 94.293 -54.395 54.981
631121 94.309 -54.403 54.970
631122 94.324 -54.412 54.953
631123 94.340 -54.421 54.935
631124 94.355 -54.430 54.912
631125 94.373 -54.439 54.886
631126 94.398 -54.456 54.861
631127 94.422 -54.473 54.836
631128 94.447 -54.489 54.809
631129 94.472 -54.505 54.781
631130 94.496 -54.522 54.746
631131 94.527 -54.536 54.706
631132 94.564 -54.549 54.665
631133 94.601 -54.561 54.624
631134 94.638 -54.573 54.582
631135 94.676 -54.585 54.538
631136 94.713 -54.597 54.489
631137 94.752 -54.618 54.438
631138 94.792 -54.645 54.389
631139 94.832 -54.671 54.343
631140 94.873 -54.698 54.300
631141 94.908 -54.718 54.258
631142 94.942 -54.734 54.210
631143 94.975 -54.751 54.160
631144 95.009 -54.768 54.109
631145 95.043 -54.784 54.058
631146 95.076 -54.801 54.007
631147 95.110 -54.818 53.955
631148 95.139 -54.834 53.899
631149 95.170 -54.851 53.844
631150 95.200 -54.868 53.791
631151 95.230 -54.884 53.740
631152 95.260 -54.901 53.691
631153 95.290 -54.917 53.646
631154 95.322 -54.931 53.599
631155 95.354 -54.946 53.553
631156 95.386 -54.960 53.511
631157 95.417 -54.974 53.471
631158 95.449 -54.988 53.432
631159 95.481 -55.002 53.396
631160 95.513 -55.016 53.361
631161 95.544 -55.031 53.327
631162 95.576 -55.044 53.292
631163 95.608 -55.059 53.256
631164 95.637 -55.072 53.219
631165 95.667 -55.084 53.180
631166 95.696 -55.095 53.140
631167 95.726 -55.107 53.099
631168 95.755 -55.119 53.057
631169 95.784 -55.131 53.014
631170 95.814 -55.143 52.969
631171 95.843 -55.154 52.922
631172 95.873 -55.166 52.874
631173 95.902 -55.178 52.824
631174 95.932 -55.190 52.773
631175 95.961 -55.202 52.719
631176 95.990 -55.213 52.664
631177 96.019 -55.217 52.608
631178 96.048 -55.221 52.554
631179 96.077 -55.226 52.502
631180 96.106 -55.230 52.443
631181 96.134 -55.235 52.381
631182 96.163 -55.239 52.315
631183 96.192 -55.244 52.248
631184 96.221 -55.248 52.181
631185 96.250 -55.252 52.114
631186 96.281 -55.261 52.046
631187 96.312 -55.271 51.979
631188 96.343 -55.281 51.911
631189 96.374 -55.292 51.843
631190 96.408 -55.311 51.777
631191 96.443 -55.335 51.712
631192 96.478 -55.359 51.645
631193 96.513 -55.383 51.579
631194 96.548 -55.408 51.512
631195 96.583 -55.432 51.447
631196 96.618 -55.456 51.382
631197 96.653 -55.480 51.317
631198 96.688 -55.505 51.252
631199 96.723 -55.529 51.187
631200 96.758 -55.553 51.122
631201 96.783 -55.564 51.059
631605 87.779 -50.778 56.501
631606 87.815 -50.802 56.518
631607 87.851 -50.827 56.535
631608 87.888 -50.851 56.553
631609 87.891 -50.895 56.564
631610 87.876 -50.947 56.573
631611 87.872 -50.992 56.578
631612 87.907 -51.013 56.574
631613 87.941 -51.033 56.562
631614 87.959 -51.022 56.548
631615 87.979 -51.018 56.532
631616 88.012 -51.037 56.514
631617 88.045 -51.057 56.491
631618 88.078 -51.076 56.466
631619 88.111 -51.096 56.438
631620 88.144 -51.114 56.408
631621 88.178 -51.134 56.377
631622 88.211 -51.153 56.345
631687 90.352 -52.276 57.161
631688 90.384 -52.284 57.170
631689 90.416 -52.292 57.176
631690 90.448 -52.300 57.183
631691 90.480 -52.308 57.189
631692 90.513 -52.316 57.195
631693 90.544 -52.325 57.200
631694 90.560 -52.347 57.203
631695 90.576 -52.370 57.202
631696 90.603 -52.376 57.201
631697 90.636 -52.376 57.200
631698 90.668 -52.376 57.197
631699 90.694 -52.394 57.194
631700 90.719 -52.416 57.192
631701 90.744 -52.439 57.187
631702 90.769 -52.462 57.182
631703 90.795 -52.484 57.179
631704 90.820 -52.506 57.178
631705 90.857 -52.526 57.179
631706 90.897 -52.543 57.179
631707 90.938 -52.562 57.173
631708 90.979 -52.580 57.175
631709 91.020 -52.597 57.181
631710 91.032 -52.621 57.192
631711 91.035 -52.647 57.200
631712 91.039 -52.672 57.202
631713 91.078 -52.680 57.198
631714 91.117 -52.688 57.189
631715 91.149 -52.703 57.183
631716 91.168 -52.731 57.178
631717 91.196 -52.751 57.170
631718 91.224 -52.771 57.158
631719 91.252 -52.791 57.141
631720 91.280 -52.812 57.120
631721 91.308 -52.832 57.111
631722 91.336 -52.853 57.098
631723 91.364 -52.873 57.079
631724 91.392 -52.894 57.050
631725 91.409 -52.908 57.015
631726 91.425 -52.922 56.980
631727 91.441 -52.935 56.948
631728 91.457 -52.949 56.918
631729 91.490 -52.973 56.886
631730 91.523 -52.998 56.854
631731 91.557 -53.022 56.822
631732 91.590 -53.046 56.787
631733 91.623 -53.071 56.749
631734 91.656 -53.095 56.710
631735 91.689 -53.120 56.671
631736 91.723 -53.145 56.630
631737 91.756 -53.169 56.588
631738 91.790 -53.193 56.546
631739 91.823 -53.218 56.503
631740 91.856 -53.242 56.460
631741 91.889 -53.267 56.416
631742 91.923 -53.292 56.372
631743 91.956 -53.316 56.327
631744 91.990 -53.340 56.283
631745 92.026 -53.360 56.236
631746 92.067 -53.374 56.190
631747 92.107 -53.387 56.145
631748 92.148 -53.401 56.099
631749 92.189 -53.415 56.055
631750 92.230 -53.427 56.012
631751 92.271 -53.441 55.969
631752 92.311 -53.455 55.926
631753 92.352 -53.468 55.885
631754 92.393 -53.482 55.846
631755 92.434 -53.495 55.808
631771 93.135 -53.660 55.160
631772 93.156 -53.681 55.142
631773 93.175 -53.703 55.125
631774 93.195 -53.725 55.109
631775 93.214 -53.747 55.094
631776 93.233 -53.769 55.079
631777 93.253 -53.791 55.064
631778 93.273 -53.813 55.052
631779 93.292 -53.834 55.042
631780 93.311 -53.856 55.033
631781 93.331 -53.878 55.024
631782 93.351 -53.900 55.016
631783 93.370 -53.922 55.008
631784 93.397 -53.940 55.005
631785 93.426 -53.954 55.003
631786 93.457 -53.969 55.001
631787 93.486 -53.983 55.000
631788 93.516 -53.997 54.999
631789 93.547 -54.013 54.998
631790 93.576 -54.027 54.999
631791 93.607 -54.041 55.000
631792 93.637 -54.056 55.002
631793 93.666 -54.071 55.004
631794 93.703 -54.085 55.007
631795 93.741 -54.098 55.009
631796 93.778 -54.112 55.013
631797 93.816 -54.126 55.017
631798 93.849 -54.136 55.022
631799 93.880 -54.144 55.026
631800 93.911 -54.152 55.030
631801 93.942 -54.160 55.034
631802 93.973 -54.168 55.039
631803 94.000 -54.178 55.040
631804 94.021 -54.192 55.039
631805 94.043 -54.204 55.037
631806 94.064 -54.218 55.034
631807 94.085 -54.232 55.031
631808 94.107 -54.245 55.026
631809 94.129 -54.258 55.023
631810 94.145 -54.267 55.020
631811 94.161 -54.276 55.018
631812 94.177 -54.285 55.016
631813 94.192 -54.294 55.011
631814 94.208 -54.302 55.003
631815 94.223 -54.311 54.998
631816 94.239 -54.320 54.993
631817 94.255 -54.328 54.989
631818 94.270 -54.337 54.987
631819 94.286 -54.345 54.984
631820 94.301 -54.355 54.979
631821 94.317 -54.363 54.972
631822 94.332 -54.372 54.962
631823 94.348 -54.380 54.947
631824 94.363 -54.389 54.930
631825 94.379 -54.398 54.909
631826 94.395 -54.406 54.885
631827 94.414 -54.418 54.862
631828 94.438 -54.435 54.838
631829 94.463 -54.452 54.812
631830 94.488 -54.467 54.782
631831 94.512 -54.484 54.747
631832 94.541 -54.499 54.706
631833 94.578 -54.511 54.663
631834 94.616 -54.524 54.619
631835 94.653 -54.536 54.575
631836 94.690 -54.548 54.530
631837 94.728 -54.563 54.481
631838 94.768 -54.590 54.428
631839 94.809 -54.616 54.377
631840 94.849 -54.643 54.330
631841 94.889 -54.669 54.286
631842 94.924 -54.688 54.243
631843 94.958 -54.705 54.195
631844 94.991 -54.721 54.144
631845 95.025 -54.738 54.092
631846 95.059 -54.755 54.041
631847 95.093 -54.772 53.990
631848 95.123 -54.788 53.940
631849 95.153 -54.805 53.886
631850 95.183 -54.822 53.834
631851 95.213 -54.838 53.784
631852 95.243 -54.855 53.737
631853 95.273 -54.872 53.692
631854 95.303 -54.887 53.649
631855 95.334 -54.903 53.605
631856 95.367 -54.917 53.561
631857 95.398 -54.931 53.520
631858 95.430 -54.945 53.480
631859 95.462 -54.960 53.440
631860 95.493 -54.973 53.405
631861 95.525 -54.988 53.370
631862 95.557 -55.002 53.336
631863 95.588 -55.016 53.302
631864 95.619 -55.031 53.268
631865 95.649 -55.044 53.232
631866 95.677 -55.055 53.193
631867 95.706 -55.067 53.154
631868 95.736 -55.079 53.114
631869 95.765 -55.091 53.071
631870 95.794 -55.103 53.027
631871 95.824 -55.114 52.982
631872 95.853 -55.126 52.936
631873 95.882 -55.138 52.890
631874 95.912 -55.150 52.841
631875 95.941 -55.162 52.790
631876 95.970 -55.173 52.736
631877 96.000 -55.184 52.682
631878 96.029 -55.188 52.629
631879 96.057 -55.193 52.581
631880 96.087 -55.198 52.534
631881 96.115 -55.202 52.479
631882 96.144 -55.207 52.418
631883 96.173 -55.211 52.353
631884 96.202 -55.215 52.284
631885 96.231 -55.220 52.214
631886 96.260 -55.224 52.146
631887 96.289 -55.229 52.076
631888 96.318 -55.233 52.006
631889 96.347 -55.239 51.937
631890 96.378 -55.250 51.868
631891 96.413 -55.273 51.801
631892 96.448 -55.297 51.734
631893 96.483 -55.321 51.667
631894 96.518 -55.345 51.600
631895 96.553 -55.370 51.533
631896 96.588 -55.394 51.467
631897 96.623 -55.419 51.402
631898 96.658 -55.443 51.337
631899 96.693 -55.467 51.271
631900 96.728 -55.492 51.206
631901 96.763 -55.516 51.140
631902 96.797 -55.539 51.077
632306 87.796 -50.748 56.503
632307 87.832 -50.773 56.520
632308 87.868 -50.797 56.536
632309 87.863 -50.844 56.550
632310 87.849 -50.896 56.562
632311 87.855 -50.936 56.571
632312 87.889 -50.956 56.572
632313 87.924 -50.976 56.566
632314 87.959 -50.997 56.558
632315 87.993 -51.017 56.544
632316 88.017 -51.019 56.528
632317 88.032 -51.005 56.509
632318 88.065 -51.023 56.486
632319 88.099 -51.042 56.460
632320 88.131 -51.062 56.432
632321 88.165 -51.081 56.403
632322 88.198 -51.101 56.371
632323 88.231 -51.120 56.339
632387 90.339 -52.235 57.150
632388 90.372 -52.242 57.159
632389 90.404 -52.251 57.167
632390 90.436 -52.258 57.174
632391 90.468 -52.267 57.181
632392 90.501 -52.274 57.187
632393 90.533 -52.283 57.192
632394 90.565 -52.291 57.196
632395 90.591 -52.304 57.199
632396 90.607 -52.327 57.198
632397 90.627 -52.343 57.197
632398 90.659 -52.343 57.196
632399 90.691 -52.343 57.193
632400 90.719 -52.357 57.191
632401 90.744 -52.379 57.189
632402 90.769 -52.402 57.185
632403 90.795 -52.424 57.181
632404 90.820 -52.447 57.180
632405 90.847 -52.469 57.181
632406 90.888 -52.487 57.183
632407 90.929 -52.505 57.184
632408 90.969 -52.523 57.179
632409 91.010 -52.540 57.179
632410 91.050 -52.559 57.185
632411 91.085 -52.578 57.193
632412 91.089 -52.603 57.200
632413 91.093 -52.628 57.202
632414 91.096 -52.653 57.199
632415 91.129 -52.665 57.193
632416 91.157 -52.684 57.195
632417 91.184 -52.704 57.197
632418 91.213 -52.725 57.197
632419 91.240 -52.745 57.189
632420 91.269 -52.766 57.172
632421 91.296 -52.786 57.149
632422 91.324 -52.807 57.139
632423 91.352 -52.827 57.125
632424 91.380 -52.847 57.104
632425 91.408 -52.868 57.074
632426 91.436 -52.888 57.038
632427 91.455 -52.903 57.009
632428 91.471 -52.917 56.978
632429 91.487 -52.931 56.947
632430 91.504 -52.944 56.917
632431 91.537 -52.969 56.888
632432 91.570 -52.994 56.856
632433 91.603 -53.018 56.821
632434 91.637 -53.042 56.782
632435 91.670 -53.067 56.742
632436 91.704 -53.092 56.702
632437 91.737 -53.116 56.661
632438 91.770 -53.141 56.618
632439 91.803 -53.165 56.574
632440 91.837 -53.189 56.529
632441 91.870 -53.214 56.484
632442 91.904 -53.239 56.438
632443 91.937 -53.263 56.391
632444 91.970 -53.288 56.345
632445 92.004 -53.312 56.298
632446 92.040 -53.331 56.251
632447 92.081 -53.345 56.204
632448 92.122 -53.358 56.157
632449 92.162 -53.372 56.111
632450 92.203 -53.385 56.066
632451 92.244 -53.399 56.021
632452 92.285 -53.412 55.977
632453 92.325 -53.426 55.934
632454 92.366 -53.440 55.892
632455 92.407 -53.452 55.852
632474 93.180 -53.675 55.117
632475 93.200 -53.697 55.100
632476 93.219 -53.719 55.083
632477 93.239 -53.741 55.067
632478 93.258 -53.763 55.052
632479 93.278 -53.784 55.040
632480 93.297 -53.806 55.029
632481 93.317 -53.828 55.019
632482 93.336 -53.850 55.011
632483 93.356 -53.872 55.002
632484 93.382 -53.889 54.994
632485 93.413 -53.904 54.990
632486 93.442 -53.919 54.987
632487 93.473 -53.933 54.985
632488 93.502 -53.947 54.983
632489 93.533 -53.963 54.982
632490 93.563 -53.977 54.981
632491 93.592 -53.991 54.983
632492 93.623 -54.006 54.984
632493 93.652 -54.021 54.987
632494 93.683 -54.035 54.989
632495 93.713 -54.050 54.991
632496 93.751 -54.063 54.994
632497 93.788 -54.077 54.997
632498 93.825 -54.091 55.002
632499 93.862 -54.104 55.007
632500 93.900 -54.118 55.011
632501 93.932 -54.127 55.016
632502 93.963 -54.135 55.021
632503 93.994 -54.143 55.025
632504 94.025 -54.151 55.028
632505 94.049 -54.163 55.029
632506 94.070 -54.176 55.029
632507 94.091 -54.189 55.027
632508 94.113 -54.203 55.025
632509 94.134 -54.216 55.023
632510 94.153 -54.227 55.023
632511 94.169 -54.236 55.025
632512 94.185 -54.245 55.025
632513 94.200 -54.254 55.025
632514 94.216 -54.262 55.021
632515 94.231 -54.271 55.011
632516 94.247 -54.279 54.997
632517 94.263 -54.288 54.987
632518 94.278 -54.297 54.981
632519 94.293 -54.305 54.978
632520 94.309 -54.314 54.974
632521 94.325 -54.323 54.969
632522 94.340 -54.332 54.965
632523 94.356 -54.340 54.956
632524 94.371 -54.349 54.943
632525 94.387 -54.358 54.927
632526 94.403 -54.366 54.908
632527 94.418 -54.375 54.887
632528 94.434 -54.383 54.866
632529 94.455 -54.397 54.843
632530 94.480 -54.414 54.818
632531 94.504 -54.430 54.788
632532 94.529 -54.446 54.752
632533 94.556 -54.462 54.709
632534 94.593 -54.474 54.663
632535 94.630 -54.486 54.617
632536 94.668 -54.499 54.569
632537 94.705 -54.511 54.523
632538 94.744 -54.533 54.474
632539 94.784 -54.560 54.423
632540 94.825 -54.587 54.375
632541 94.865 -54.614 54.327
632542 94.905 -54.640 54.283
632543 94.940 -54.659 54.236
632544 94.974 -54.675 54.185
632545 95.007 -54.692 54.133
632546 95.041 -54.709 54.081
632547 95.075 -54.725 54.028
632548 95.106 -54.742 53.978
632549 95.136 -54.759 53.929
632550 95.166 -54.775 53.878
632551 95.196 -54.792 53.829
632552 95.226 -54.809 53.783
632553 95.256 -54.825 53.742
632554 95.286 -54.841 53.700
632555 95.316 -54.858 53.658
632556 95.347 -54.874 53.614
632557 95.379 -54.888 53.571
632558 95.410 -54.902 53.529
632559 95.442 -54.916 53.488
632560 95.474 -54.931 53.447
632561 95.506 -54.945 53.411
632562 95.537 -54.959 53.377
632563 95.567 -54.973 53.344
632564 95.599 -54.988 53.311
632565 95.629 -55.002 53.278
632566 95.660 -55.016 53.243
632567 95.688 -55.028 53.207
632568 95.716 -55.038 53.167
632569 95.745 -55.050 53.127
632570 95.774 -55.063 53.085
632571 95.804 -55.074 53.041
632572 95.833 -55.086 52.996
632573 95.863 -55.097 52.950
632574 95.892 -55.110 52.903
632575 95.921 -55.121 52.851
632576 95.951 -55.133 52.802
632577 95.980 -55.145 52.754
632578 96.010 -55.156 52.706
632579 96.038 -55.160 52.659
632580 96.067 -55.165 52.614
632581 96.096 -55.169 52.567
632582 96.125 -55.174 52.518
632583 96.154 -55.179 52.462
632584 96.183 -55.182 52.398
632585 96.212 -55.187 52.329
632586 96.241 -55.191 52.256
632587 96.269 -55.196 52.185
632588 96.299 -55.201 52.113
632589 96.327 -55.205 52.039
632590 96.356 -55.210 51.966
632591 96.384 -55.217 51.895
632592 96.418 -55.235 51.827
632593 96.453 -55.260 51.758
632594 96.488 -55.284 51.690
632595 96.523 -55.308 51.622
632596 96.558 -55.332 51.555
632597 96.593 -55.357 51.489
632598 96.628 -55.381 51.423
632599 96.663 -55.405 51.358
632600 96.698 -55.430 51.292
632601 96.733 -55.454 51.226
632602 96.768 -55.478 51.161
632603 96.803 -55.503 51.097
632604 96.828 -55.513 51.035
632605 96.845 -55.512 50.974
633007 87.813 -50.718 56.502
633008 87.849 -50.742 56.517
633009 87.837 -50.793 56.531
633010 87.822 -50.845 56.544
633011 87.837 -50.879 56.554
633012 87.872 -50.899 56.561
633013 87.907 -50.920 56.562
633014 87.941 -50.940 56.557
633015 87.976 -50.961 56.548
633016 88.010 -50.982 56.535
633017 88.045 -51.002 56.520
633018 88.075 -51.015 56.501
633019 88.091 -51.001 56.479
633020 88.118 -51.010 56.453
633021 88.151 -51.029 56.424
633022 88.185 -51.048 56.394
633023 88.217 -51.067 56.362
633024 88.251 -51.087 56.330
633087 90.327 -52.193 57.137
633088 90.359 -52.202 57.147
633089 90.391 -52.209 57.156
633090 90.424 -52.217 57.164
633091 90.456 -52.225 57.171
633092 90.488 -52.233 57.177
633093 90.521 -52.241 57.183
633094 90.553 -52.249 57.187
633095 90.585 -52.258 57.190
633096 90.617 -52.265 57.194
633097 90.639 -52.283 57.194
633098 90.655 -52.305 57.192
633099 90.683 -52.310 57.191
633100 90.715 -52.310 57.189
633101 90.744 -52.320 57.187
633102 90.769 -52.342 57.186
633103 90.795 -52.365 57.182
633104 90.820 -52.387 57.180
633105 90.845 -52.410 57.180
633106 90.878 -52.430 57.182
633107 90.919 -52.448 57.185
633108 90.959 -52.466 57.187
633109 91.000 -52.484 57.184
633110 91.041 -52.502 57.184
633111 91.082 -52.520 57.189
633112 91.122 -52.537 57.195
633113 91.143 -52.559 57.201
633114 91.147 -52.585 57.205
633115 91.150 -52.610 57.204
633116 91.154 -52.635 57.203
633117 91.173 -52.657 57.207
633118 91.201 -52.678 57.213
633119 91.229 -52.698 57.218
633120 91.257 -52.719 57.217
633121 91.285 -52.739 57.201
633122 91.312 -52.760 57.178
633123 91.341 -52.780 57.165
633124 91.368 -52.800 57.150
633125 91.396 -52.821 57.127
633126 91.424 -52.841 57.096
633127 91.452 -52.862 57.060
633128 91.480 -52.882 57.038
633129 91.501 -52.898 57.007
633130 91.517 -52.913 56.974
633131 91.533 -52.926 56.946
633132 91.551 -52.941 56.920
633133 91.584 -52.965 56.890
633134 91.617 -52.989 56.855
633135 91.651 -53.014 56.814
633136 91.684 -53.038 56.775
633137 91.717 -53.063 56.737
633138 91.751 -53.088 56.695
633139 91.784 -53.112 56.650
633140 91.817 -53.136 56.602
633141 91.851 -53.161 56.555
633142 91.884 -53.186 56.507
633143 91.918 -53.210 56.459
633144 91.951 -53.235 56.410
633145 91.984 -53.259 56.362
633146 92.017 -53.283 56.314
633147 92.055 -53.302 56.265
633148 92.095 -53.316 56.218
633149 92.136 -53.330 56.170
633150 92.177 -53.343 56.123
633151 92.217 -53.356 56.076
633152 92.258 -53.370 56.031
633153 92.299 -53.383 55.986
633154 92.340 -53.397 55.942
633155 92.380 -53.411 55.900
633156 92.421 -53.424 55.859
633178 93.244 -53.713 55.057
633179 93.264 -53.735 55.042
633180 93.283 -53.757 55.028
633181 93.302 -53.778 55.016
633182 93.322 -53.800 55.006
633183 93.341 -53.822 54.996
633184 93.368 -53.839 54.988
633185 93.399 -53.853 54.980
633186 93.428 -53.869 54.976
633187 93.459 -53.883 54.973
633188 93.489 -53.897 54.970
633189 93.518 -53.912 54.968
633190 93.549 -53.927 54.968
633191 93.578 -53.941 54.966
633192 93.609 -53.956 54.968
633193 93.639 -53.970 54.970
633194 93.669 -53.985 54.972
633195 93.699 -54.000 54.974
633196 93.728 -54.014 54.977
633197 93.760 -54.029 54.979
633198 93.797 -54.042 54.983
633199 93.835 -54.056 54.988
633200 93.872 -54.070 54.993
633201 93.910 -54.084 54.998
633202 93.947 -54.098 55.003
633203 93.983 -54.110 55.009
633204 94.014 -54.118 55.013
633205 94.045 -54.126 55.018
633206 94.076 -54.134 55.020
633207 94.097 -54.148 55.022
633208 94.119 -54.160 55.023
633209 94.140 -54.174 55.025
633210 94.161 -54.187 55.028
633211 94.177 -54.196 55.034
633212 94.193 -54.204 55.042
633213 94.208 -54.214 55.046
633214 94.224 -54.222 55.046
633215 94.240 -54.231 55.040
633216 94.255 -54.239 55.026
633217 94.271 -54.248 55.009
633218 94.286 -54.257 54.992
633219 94.301 -54.265 54.979
633220 94.317 -54.274 54.972
633221 94.333 -54.283 54.968
633222 94.348 -54.292 54.962
633223 94.364 -54.300 54.959
633224 94.380 -54.309 54.952
633225 94.395 -54.317 54.941
633226 94.411 -54.326 54.927
633227 94.426 -54.335 54.911
633228 94.442 -54.343 54.891
633229 94.458 -54.352 54.872
633230 94.473 -54.361 54.849
633231 94.496 -54.376 54.823
633232 94.521 -54.392 54.792
633233 94.545 -54.408 54.754
633234 94.571 -54.425 54.710
633235 94.608 -54.437 54.663
633236 94.645 -54.449 54.614
633237 94.682 -54.461 54.565
633238 94.720 -54.478 54.519
633239 94.761 -54.505 54.471
633240 94.801 -54.531 54.424
633241 94.841 -54.558 54.379
633242 94.882 -54.584 54.334
633243 94.922 -54.611 54.288
633244 94.956 -54.629 54.241
633245 94.990 -54.646 54.191
633246 95.024 -54.662 54.139
633247 95.057 -54.679 54.088
633248 95.089 -54.696 54.036
633249 95.119 -54.712 53.985
633250 95.149 -54.729 53.936
633251 95.179 -54.746 53.885
633252 95.209 -54.762 53.838
633253 95.239 -54.778 53.795
633254 95.269 -54.795 53.753
633255 95.300 -54.812 53.711
633256 95.329 -54.828 53.669
633257 95.359 -54.845 53.624
633258 95.391 -54.859 53.580
633259 95.423 -54.874 53.535
633260 95.455 -54.887 53.491
633261 95.486 -54.902 53.451
633262 95.517 -54.916 53.416
633263 95.547 -54.931 53.383
633264 95.579 -54.945 53.351
633265 95.609 -54.960 53.319
633266 95.640 -54.974 53.287
633267 95.671 -54.988 53.254
633268 95.699 -55.000 53.220
633269 95.727 -55.011 53.180
633270 95.755 -55.022 53.139
633271 95.784 -55.034 53.097
633272 95.814 -55.046 53.053
633273 95.843 -55.057 53.009
633274 95.873 -55.069 52.964
633275 95.902 -55.081 52.915
633276 95.931 -55.093 52.865
633277 95.961 -55.105 52.818
633278 95.990 -55.116 52.775
633279 96.019 -55.128 52.732
633280 96.048 -55.132 52.688
633281 96.077 -55.137 52.645
633282 96.106 -55.141 52.599
633283 96.135 -55.146 52.550
633284 96.164 -55.150 52.494
633285 96.192 -55.154 52.431
633286 96.222 -55.159 52.365
633287 96.250 -55.163 52.297
633288 96.279 -55.168 52.228
633289 96.308 -55.173 52.155
633290 96.337 -55.177 52.079
633291 96.366 -55.187 52.003
633292 96.391 -55.195 51.930
633293 96.422 -55.203 51.859
633294 96.459 -55.222 51.789
633295 96.493 -55.246 51.719
633296 96.529 -55.271 51.651
633297 96.564 -55.295 51.583
633298 96.599 -55.320 51.516
633299 96.633 -55.344 51.449
633300 96.669 -55.368 51.383
633301 96.704 -55.392 51.316
633302 96.739 -55.417 51.250
633303 96.774 -55.441 51.184
633304 96.809 -55.465 51.120
633305 96.842 -55.488 51.057
633306 96.859 -55.487 50.995
633307 96.876 -55.486 50.933
633708 87.824 -50.691 56.497
633709 87.810 -50.743 56.511
633710 87.794 -50.795 56.523
633711 87.820 -50.822 56.534
633712 87.855 -50.842 56.542
633713 87.890 -50.863 56.548
633714 87.924 -50.883 56.549
633715 87.959 -50.904 56.543
633716 87.993 -50.925 56.534
633717 88.028 -50.945 56.522
633718 88.062 -50.966 56.506
633719 88.097 -50.986 56.488
633720 88.132 -51.008 56.468
633721 88.149 -50.997 56.443
633722 88.172 -50.995 56.412
633723 88.204 -51.015 56.382
633724 88.238 -51.034 56.350
633725 88.271 -51.054 56.318
633787 90.315 -52.152 57.123
633788 90.347 -52.160 57.134
633789 90.379 -52.167 57.143
633790 90.411 -52.176 57.152
633791 90.444 -52.184 57.160
633792 90.476 -52.192 57.166
633793 90.508 -52.200 57.172
633794 90.540 -52.208 57.177
633795 90.573 -52.216 57.182
633796 90.605 -52.224 57.184
633797 90.638 -52.232 57.188
633798 90.670 -52.240 57.187
633799 90.686 -52.262 57.186
633800 90.707 -52.277 57.185
633801 90.739 -52.277 57.184
633802 90.769 -52.283 57.183
633803 90.795 -52.305 57.183
633804 90.820 -52.327 57.181
633805 90.845 -52.350 57.179
633806 90.870 -52.373 57.179
633807 90.909 -52.391 57.181
633808 90.950 -52.409 57.185
633809 90.991 -52.427 57.189
633810 91.031 -52.445 57.188
633811 91.072 -52.463 57.190
633812 91.113 -52.481 57.194
633813 91.153 -52.499 57.199
633814 91.194 -52.517 57.205
633815 91.201 -52.541 57.211
633816 91.203 -52.568 57.213
633817 91.185 -52.601 57.216
633818 91.191 -52.628 57.222
633819 91.218 -52.651 57.229
633820 91.245 -52.672 57.235
633821 91.273 -52.693 57.234
633822 91.301 -52.713 57.219
633823 91.329 -52.734 57.197
633824 91.357 -52.754 57.180
633825 91.385 -52.774 57.161
633826 91.413 -52.794 57.138
633827 91.440 -52.815 57.112
633828 91.468 -52.835 57.084
633829 91.496 -52.856 57.065
633830 91.524 -52.876 57.035
633831 91.547 -52.894 57.001
633832 91.564 -52.908 56.973
633833 91.580 -52.922 56.952
633834 91.598 -52.936 56.925
633835 91.632 -52.961 56.888
633836 91.665 -52.985 56.845
633837 91.698 -53.010 56.806
633838 91.731 -53.035 56.770
633839 91.765 -53.059 56.731
633840 91.798 -53.083 56.683
633841 91.831 -53.108 56.630
633842 91.865 -53.132 56.579
633843 91.898 -53.157 56.529
633844 91.931 -53.182 56.479
633845 91.964 -53.206 56.429
633846 91.998 -53.230 56.378
633847 92.031 -53.255 56.329
633848 92.069 -53.273 56.280
633849 92.109 -53.287 56.230
633850 92.150 -53.301 56.181
633851 92.191 -53.314 56.133
633852 92.232 -53.327 56.085
633853 92.272 -53.341 56.040
633854 92.313 -53.355 55.994
633855 92.354 -53.368 55.950
633856 92.394 -53.382 55.907
633857 92.435 -53.395 55.865
633882 93.307 -53.751 55.005
633883 93.327 -53.772 54.994
633884 93.354 -53.789 54.983
633885 93.384 -53.803 54.975
633886 93.415 -53.819 54.967
633887 93.444 -53.833 54.962
633888 93.475 -53.847 54.959
633889 93.504 -53.862 54.956
633890 93.535 -53.877 54.954
633891 93.565 -53.891 54.953
633892 93.594 -53.906 54.952
633893 93.625 -53.920 54.953
633894 93.654 -53.935 54.956
633895 93.685 -53.950 54.959
633896 93.715 -53.964 54.961
633897 93.745 -53.979 54.964
633898 93.775 -53.993 54.966
633899 93.807 -54.007 54.970
633900 93.844 -54.021 54.975
633901 93.882 -54.035 54.981
633902 93.919 -54.049 54.986
633903 93.957 -54.063 54.992
633904 93.994 -54.076 54.997
633905 94.031 -54.090 55.002
633906 94.066 -54.101 55.008
633907 94.097 -54.109 55.012
633908 94.124 -54.119 55.017
633909 94.146 -54.132 55.022
633910 94.168 -54.145 55.028
633911 94.185 -54.156 55.037
633912 94.201 -54.164 55.049
633913 94.216 -54.173 55.061
633914 94.232 -54.182 55.066
633915 94.248 -54.191 55.064
633916 94.263 -54.199 55.056
633917 94.279 -54.208 55.040
633918 94.294 -54.217 55.019
633919 94.310 -54.225 54.999
633920 94.325 -54.234 54.984
633921 94.341 -54.242 54.975
633922 94.356 -54.251 54.970
633923 94.372 -54.260 54.965
633924 94.388 -54.269 54.964
633925 94.403 -54.277 54.958
633926 94.419 -54.286 54.948
633927 94.434 -54.295 54.935
633928 94.450 -54.303 54.918
633929 94.466 -54.312 54.896
633930 94.481 -54.320 54.874
633931 94.497 -54.330 54.849
633932 94.512 -54.338 54.821
633933 94.537 -54.355 54.790
633934 94.561 -54.370 54.752
633935 94.586 -54.387 54.708
633936 94.623 -54.400 54.661
633937 94.660 -54.412 54.612
633938 94.697 -54.424 54.564
633939 94.737 -54.449 54.518
633940 94.777 -54.475 54.473
633941 94.818 -54.502 54.429
633942 94.858 -54.529 54.387
633943 94.898 -54.555 54.345
633944 94.938 -54.582 54.299
633945 94.972 -54.599 54.252
633946 95.006 -54.616 54.203
633947 95.040 -54.633 54.151
633948 95.071 -54.649 54.099
633949 95.102 -54.666 54.046
633950 95.132 -54.683 53.995
633951 95.162 -54.699 53.946
633952 95.192 -54.715 53.897
633953 95.222 -54.732 53.852
633954 95.252 -54.749 53.810
633955 95.282 -54.765 53.769
633956 95.313 -54.782 53.728
633957 95.342 -54.799 53.685
633958 95.373 -54.815 53.639
633959 95.404 -54.831 53.592
633960 95.435 -54.845 53.545
633961 95.466 -54.859 53.500
633962 95.496 -54.874 53.460
633963 95.527 -54.888 53.426
633964 95.558 -54.903 53.393
633965 95.589 -54.917 53.362
633966 95.620 -54.931 53.331
633967 95.651 -54.946 53.299
633968 95.682 -54.960 53.266
633969 95.710 -54.972 53.231
633970 95.739 -54.983 53.191
633971 95.766 -54.994 53.148
633972 95.794 -55.006 53.105
633973 95.823 -55.017 53.062
633974 95.853 -55.029 53.019
633975 95.882 -55.041 52.974
633976 95.911 -55.053 52.927
633977 95.941 -55.065 52.882
633978 95.970 -55.076 52.840
633979 96.000 -55.088 52.799
633980 96.029 -55.100 52.759
633981 96.058 -55.104 52.717
633982 96.087 -55.109 52.674
633983 96.116 -55.113 52.629
633984 96.145 -55.117 52.578
633985 96.173 -55.122 52.522
633986 96.202 -55.126 52.461
633987 96.231 -55.131 52.399
633988 96.260 -55.135 52.337
633989 96.289 -55.140 52.270
633990 96.318 -55.145 52.196
633991 96.346 -55.155 52.118
633992 96.372 -55.165 52.040
633993 96.399 -55.172 51.965
633994 96.429 -55.180 51.893
633995 96.462 -55.189 51.822
633996 96.499 -55.209 51.751
633997 96.534 -55.233 51.682
633998 96.569 -55.257 51.614
633999 96.604 -55.282 51.546
634000 96.639 -55.306 51.478
634001 96.674 -55.330 51.411
634002 96.709 -55.354 51.344
634003 96.744 -55.379 51.277
634004 96.779 -55.404 51.210
634005 96.814 -55.428 51.146
634006 96.849 -55.452 51.083
634007 96.874 -55.462 51.021
634008 96.890 -55.461 50.958
634009 96.906 -55.461 50.897
634010 96.923 -55.459 50.836
634409 87.782 -50.692 56.489
634410 87.768 -50.744 56.502
634411 87.803 -50.765 56.512
634412 87.838 -50.785 56.521
634413 87.872 -50.806 56.528
634414 87.907 -50.826 56.532
634415 87.941 -50.848 56.533
634416 87.976 -50.868 56.526
634417 88.010 -50.889 56.517
634418 88.045 -50.909 56.504
634419 88.080 -50.930 56.488
634420 88.114 -50.951 56.471
634421 88.149 -50.971 56.451
634422 88.184 -50.992 56.425
634423 88.208 -50.993 56.396
634424 88.224 -50.982 56.366
634425 88.258 -51.001 56.335
634488 90.334 -52.118 57.118
634489 90.367 -52.127 57.128
634490 90.399 -52.134 57.138
634491 90.432 -52.142 57.146
634492 90.464 -52.151 57.154
634493 90.496 -52.158 57.160
634494 90.528 -52.167 57.165
634495 90.561 -52.174 57.170
634496 90.593 -52.183 57.175
634497 90.625 -52.190 57.177
634498 90.657 -52.199 57.180
634499 90.690 -52.207 57.180
634500 90.717 -52.219 57.180
634501 90.733 -52.241 57.179
634502 90.762 -52.245 57.178
634503 90.794 -52.246 57.179
634504 90.819 -52.268 57.180
634505 90.845 -52.290 57.178
634506 90.870 -52.313 57.178
634507 90.899 -52.334 57.179
634508 90.940 -52.352 57.182
634509 90.980 -52.370 57.186
634510 91.021 -52.388 57.191
634511 91.062 -52.406 57.194
634512 91.103 -52.424 57.197
634513 91.144 -52.442 57.202
634514 91.184 -52.460 57.208
634515 91.225 -52.478 57.215
634516 91.243 -52.503 57.224
634517 91.226 -52.536 57.229
634518 91.209 -52.570 57.234
634519 91.210 -52.599 57.239
634520 91.236 -52.621 57.247
634521 91.263 -52.644 57.254
634522 91.290 -52.666 57.251
634523 91.317 -52.687 57.234
634524 91.345 -52.707 57.213
634525 91.373 -52.728 57.191
634526 91.401 -52.748 57.169
634527 91.429 -52.769 57.146
634528 91.457 -52.789 57.122
634529 91.485 -52.810 57.102
634530 91.513 -52.830 57.084
634531 91.540 -52.850 57.055
634532 91.568 -52.871 57.026
634533 91.593 -52.889 57.001
634534 91.610 -52.904 56.979
634535 91.626 -52.917 56.960
634536 91.645 -52.932 56.920
634537 91.678 -52.957 56.877
634538 91.712 -52.982 56.836
634539 91.745 -53.006 56.801
634540 91.779 -53.030 56.766
634541 91.812 -53.055 56.716
634542 91.845 -53.079 56.658
634543 91.878 -53.104 56.602
634544 91.912 -53.129 56.550
634545 91.945 -53.153 56.499
634546 91.979 -53.177 56.447
634547 92.012 -53.202 56.395
634548 92.045 -53.227 56.344
634549 92.083 -53.245 56.294
634550 92.124 -53.258 56.243
634551 92.164 -53.272 56.193
634552 92.205 -53.285 56.144
634553 92.246 -53.299 56.096
634554 92.287 -53.312 56.049
634555 92.327 -53.326 56.002
634556 92.368 -53.339 55.957
634557 92.409 -53.352 55.914
634585 93.370 -53.753 54.973
634586 93.401 -53.768 54.962
634587 93.430 -53.783 54.955
634588 93.461 -53.797 54.949
634589 93.491 -53.812 54.945
634590 93.520 -53.826 54.942
634591 93.551 -53.841 54.941
634592 93.581 -53.856 54.940
634593 93.611 -53.870 54.939
634594 93.641 -53.885 54.940
634595 93.671 -53.900 54.943
634596 93.701 -53.914 54.946
634597 93.731 -53.928 54.949
634598 93.761 -53.943 54.952
634599 93.791 -53.958 54.954
634600 93.821 -53.972 54.958
634601 93.854 -53.987 54.963
634602 93.892 -54.001 54.969
634603 93.929 -54.014 54.975
634604 93.966 -54.028 54.981
634605 94.003 -54.041 54.987
634606 94.041 -54.055 54.994
634607 94.078 -54.069 55.001
634608 94.116 -54.082 55.007
634609 94.148 -54.092 55.014
634610 94.173 -54.104 55.023
634611 94.193 -54.116 55.034
634612 94.209 -54.124 55.050
634613 94.224 -54.133 55.067
634614 94.240 -54.142 55.079
634615 94.256 -54.151 55.083
634616 94.271 -54.159 55.078
634617 94.287 -54.168 55.069
634618 94.302 -54.176 55.051
634619 94.318 -54.185 55.027
634620 94.333 -54.194 55.006
634621 94.349 -54.202 54.990
634622 94.364 -54.211 54.981
634623 94.380 -54.220 54.975
634624 94.396 -54.229 54.970
634625 94.411 -54.237 54.969
634626 94.427 -54.245 54.963
634627 94.442 -54.254 54.954
634628 94.458 -54.263 54.940
634629 94.474 -54.272 54.921
634630 94.489 -54.280 54.896
634631 94.505 -54.289 54.872
634632 94.521 -54.298 54.844
634633 94.536 -54.307 54.816
634634 94.553 -54.317 54.784
634635 94.578 -54.333 54.747
634636 94.603 -54.349 54.705
634637 94.637 -54.362 54.660
634638 94.675 -54.374 54.614
634639 94.713 -54.393 54.568
634640 94.753 -54.420 54.524
634641 94.793 -54.446 54.482
634642 94.834 -54.473 54.441
634643 94.874 -54.499 54.400
634644 94.914 -54.526 54.358
634645 94.955 -54.553 54.314
634646 94.988 -54.570 54.268
634647 95.022 -54.587 54.218
634648 95.054 -54.603 54.165
634649 95.085 -54.620 54.113
634650 95.115 -54.637 54.060
634651 95.145 -54.653 54.007
634652 95.175 -54.669 53.957
634653 95.205 -54.686 53.912
634654 95.235 -54.703 53.870
634655 95.265 -54.719 53.829
634656 95.296 -54.736 53.786
634657 95.326 -54.753 53.744
634658 95.355 -54.769 53.700
634659 95.385 -54.786 53.653
634660 95.413 -54.803 53.604
634661 95.446 -54.816 53.556
634662 95.476 -54.831 53.512
634663 95.507 -54.845 53.471
634664 95.538 -54.859 53.436
634665 95.569 -54.874 53.403
634666 95.600 -54.888 53.372
634667 95.631 -54.903 53.342
634668 95.662 -54.917 53.310
634669 95.693 -54.931 53.277
634670 95.722 -54.944 53.242
634671 95.750 -54.956 53.202
634672 95.777 -54.966 53.161
634673 95.806 -54.978 53.119
634674 95.833 -54.989 53.078
634675 95.863 -55.000 53.035
634676 95.892 -55.013 52.992
634677 95.921 -55.025 52.948
634678 95.951 -55.036 52.905
634679 95.980 -55.048 52.864
634680 96.010 -55.060 52.824
634681 96.039 -55.072 52.784
634682 96.067 -55.076 52.743
634683 96.097 -55.081 52.700
634684 96.125 -55.085 52.654
634685 96.154 -55.089 52.603
634686 96.183 -55.094 52.549
634687 96.212 -55.098 52.494
634688 96.241 -55.103 52.436
634689 96.270 -55.107 52.377
634690 96.299 -55.114 52.309
634691 96.327 -55.124 52.235
634692 96.355 -55.135 52.156
634693 96.379 -55.142 52.077
634694 96.406 -55.150 52.002
634695 96.436 -55.157 51.930
634696 96.467 -55.166 51.858
634697 96.503 -55.176 51.787
634698 96.539 -55.196 51.716
634699 96.575 -55.220 51.647
634700 96.609 -55.245 51.578
634701 96.645 -55.269 51.511
634702 96.679 -55.293 51.442
634703 96.715 -55.317 51.374
634704 96.749 -55.342 51.306
634705 96.785 -55.366 51.239
634706 96.819 -55.390 51.173
634707 96.855 -55.414 51.110
634708 96.888 -55.437 51.047
634709 96.904 -55.436 50.984
634710 96.921 -55.435 50.922
634711 96.938 -55.434 50.861
634712 96.954 -55.433 50.802
634713 96.971 -55.433 50.745
635110 87.751 -50.687 56.475
635111 87.786 -50.708 56.487
635112 87.820 -50.729 56.495
635113 87.855 -50.749 56.502
635114 87.890 -50.770 56.508
635115 87.924 -50.791 56.512
635116 87.959 -50.811 56.513
635117 87.993 -50.832 56.506
635118 88.028 -50.852 56.496
635119 88.062 -50.873 56.483
635120 88.097 -50.894 56.467
635121 88.132 -50.914 56.450
635122 88.167 -50.935 56.431
635123 88.201 -50.955 56.404
635124 88.236 -50.976 56.376
635125 88.266 -50.990 56.347
635126 88.281 -50.975 56.316
635188 90.322 -52.077 57.102
635189 90.355 -52.085 57.112
635190 90.387 -52.093 57.121
635191 90.419 -52.101 57.131
635192 90.451 -52.109 57.139
635193 90.484 -52.117 57.147
635194 90.516 -52.125 57.153
635195 90.548 -52.133 57.158
635196 90.581 -52.141 57.163
635197 90.613 -52.149 57.167
635198 90.645 -52.157 57.170
635199 90.677 -52.165 57.173
635200 90.710 -52.174 57.173
635201 90.742 -52.181 57.172
635202 90.764 -52.198 57.173
635203 90.786 -52.212 57.173
635204 90.818 -52.212 57.174
635205 90.845 -52.230 57.176
635206 90.870 -52.253 57.176
635207 90.895 -52.276 57.177
635208 90.930 -52.296 57.180
635209 90.971 -52.313 57.184
635210 91.012 -52.331 57.189
635211 91.052 -52.349 57.196
635212 91.093 -52.367 57.202
635213 91.134 -52.385 57.208
635214 91.175 -52.403 57.215
635215 91.211 -52.421 57.224
635216 91.239 -52.445 57.232
635217 91.266 -52.471 57.239
635218 91.249 -52.505 57.250
635219 91.232 -52.539 57.256
635220 91.228 -52.568 57.262
635221 91.255 -52.591 57.269
635222 91.282 -52.613 57.275
635223 91.308 -52.636 57.271
635224 91.335 -52.658 57.250
635225 91.362 -52.681 57.226
635226 91.389 -52.701 57.201
635227 91.417 -52.722 57.175
635228 91.445 -52.742 57.151
635229 91.473 -52.763 57.129
635230 91.501 -52.783 57.115
635231 91.529 -52.803 57.099
635232 91.557 -52.824 57.075
635233 91.585 -52.844 57.050
635234 91.612 -52.865 57.026
635235 91.640 -52.885 57.003
635236 91.656 -52.899 56.983
635237 91.672 -52.913 56.947
635238 91.693 -52.929 56.905
635239 91.726 -52.953 56.864
635240 91.759 -52.978 56.829
635241 91.792 -53.002 56.792
635242 91.826 -53.026 56.742
635243 91.859 -53.051 56.685
635244 91.892 -53.076 56.626
635245 91.926 -53.100 56.570
635246 91.959 -53.125 56.517
635247 91.993 -53.149 56.465
635248 92.026 -53.173 56.411
635249 92.059 -53.198 56.359
635250 92.097 -53.216 56.308
635251 92.138 -53.230 56.256
635252 92.178 -53.243 56.205
635253 92.219 -53.256 56.155
635254 92.260 -53.270 56.106
635255 92.301 -53.283 56.058
635256 92.342 -53.297 56.010
635257 92.382 -53.311 55.964
635258 92.423 -53.324 55.920
635288 93.446 -53.747 54.944
635289 93.477 -53.762 54.938
635290 93.507 -53.776 54.934
635291 93.537 -53.791 54.931
635292 93.567 -53.806 54.929
635293 93.596 -53.820 54.927
635294 93.627 -53.834 54.927
635295 93.657 -53.849 54.928
635296 93.687 -53.864 54.930
635297 93.717 -53.878 54.933
635298 93.747 -53.893 54.937
635299 93.777 -53.907 54.940
635300 93.807 -53.922 54.942
635301 93.837 -53.937 54.947
635302 93.867 -53.951 54.952
635303 93.901 -53.966 54.958
635304 93.938 -53.979 54.964
635305 93.976 -53.993 54.972
635306 94.013 -54.007 54.978
635307 94.051 -54.020 54.987
635308 94.088 -54.034 54.994
635309 94.125 -54.048 55.002
635310 94.163 -54.061 55.012
635311 94.200 -54.076 55.024
635312 94.217 -54.084 55.040
635313 94.232 -54.093 55.059
635314 94.248 -54.101 55.075
635315 94.264 -54.110 55.084
635316 94.279 -54.119 55.084
635317 94.295 -54.128 55.079
635318 94.310 -54.136 55.070
635319 94.326 -54.145 55.050
635320 94.341 -54.154 55.029
635321 94.357 -54.162 55.011
635322 94.372 -54.171 54.997
635323 94.388 -54.179 54.987
635324 94.404 -54.189 54.981
635325 94.419 -54.197 54.975
635326 94.435 -54.205 54.972
635327 94.451 -54.214 54.967
635328 94.466 -54.223 54.957
635329 94.482 -54.232 54.941
635330 94.497 -54.240 54.920
635331 94.513 -54.249 54.894
635332 94.529 -54.258 54.866
635333 94.544 -54.267 54.837
635334 94.559 -54.275 54.808
635335 94.575 -54.283 54.778
635336 94.594 -54.295 54.742
635337 94.619 -54.311 54.701
635338 94.652 -54.325 54.660
635339 94.689 -54.337 54.616
635340 94.730 -54.364 54.573
635341 94.770 -54.391 54.532
635342 94.810 -54.417 54.493
635343 94.850 -54.444 54.453
635344 94.891 -54.471 54.415
635345 94.931 -54.497 54.373
635346 94.971 -54.524 54.331
635347 95.004 -54.540 54.285
635348 95.038 -54.557 54.235
635349 95.067 -54.574 54.180
635350 95.098 -54.590 54.125
635351 95.128 -54.606 54.070
635352 95.158 -54.623 54.018
635353 95.188 -54.640 53.973
635354 95.218 -54.656 53.931
635355 95.248 -54.673 53.889
635356 95.278 -54.690 53.846
635357 95.308 -54.706 53.802
635358 95.338 -54.723 53.756
635359 95.367 -54.740 53.709
635360 95.394 -54.759 53.662
635361 95.421 -54.775 53.615
635362 95.454 -54.789 53.569
635363 95.487 -54.803 53.526
635364 95.518 -54.817 53.484
635365 95.549 -54.831 53.447
635366 95.580 -54.846 53.413
635367 95.611 -54.860 53.382
635368 95.641 -54.875 53.352
635369 95.673 -54.889 53.321
635370 95.703 -54.903 53.288
635371 95.734 -54.916 53.254
635372 95.761 -54.928 53.215
635373 95.789 -54.939 53.176
635374 95.817 -54.950 53.136
635375 95.845 -54.961 53.095
635376 95.873 -54.972 53.054
635377 95.902 -54.984 53.013
635378 95.931 -54.996 52.971
635379 95.960 -55.008 52.930
635380 95.990 -55.019 52.889
635381 96.019 -55.032 52.849
635382 96.048 -55.043 52.809
635383 96.078 -55.048 52.768
635384 96.106 -55.052 52.725
635385 96.135 -55.057 52.681
635386 96.164 -55.061 52.632
635387 96.193 -55.066 52.581
635388 96.222 -55.070 52.527
635389 96.251 -55.075 52.468
635390 96.279 -55.083 52.407
635391 96.308 -55.093 52.337
635392 96.336 -55.104 52.263
635393 96.362 -55.112 52.190
635394 96.386 -55.120 52.118
635395 96.414 -55.128 52.046
635396 96.443 -55.135 51.974
635397 96.473 -55.143 51.901
635398 96.505 -55.151 51.828
635399 96.543 -55.163 51.755
635400 96.580 -55.182 51.684
635401 96.615 -55.207 51.613
635402 96.650 -55.231 51.544
635403 96.685 -55.256 51.475
635404 96.720 -55.280 51.406
635405 96.755 -55.304 51.338
635406 96.790 -55.329 51.270
635407 96.825 -55.353 51.203
635408 96.860 -55.377 51.138
635409 96.895 -55.401 51.074
635410 96.919 -55.411 51.011
635411 96.935 -55.410 50.948
635412 96.952 -55.409 50.886
635413 96.968 -55.408 50.826
635414 96.985 -55.408 50.768
635415 97.002 -55.407 50.712
635811 87.768 -50.651 56.459
635812 87.803 -50.672 56.468
635813 87.838 -50.692 56.475
635814 87.872 -50.713 56.481
635815 87.907 -50.734 56.486
635816 87.941 -50.754 56.488
635817 87.976 -50.775 56.488
635818 88.010 -50.795 56.480
635819 88.045 -50.817 56.470
635820 88.080 -50.837 56.457
635821 88.115 -50.857 56.441
635822 88.149 -50.878 56.425
635823 88.184 -50.899 56.406
635824 88.218 -50.920 56.380
635825 88.253 -50.940 56.354
635826 88.287 -50.961 56.325
635888 90.310 -52.036 57.084
635889 90.342 -52.044 57.094
635890 90.375 -52.051 57.104
635891 90.407 -52.060 57.114
635892 90.439 -52.067 57.123
635893 90.471 -52.076 57.131
635894 90.504 -52.084 57.139
635895 90.536 -52.092 57.145
635896 90.568 -52.100 57.150
635897 90.601 -52.108 57.154
635898 90.633 -52.116 57.158
635899 90.665 -52.123 57.162
635900 90.697 -52.132 57.165
635901 90.730 -52.140 57.166
635902 90.762 -52.148 57.166
635903 90.794 -52.156 57.166
635904 90.812 -52.177 57.168
635905 90.842 -52.179 57.169
635906 90.870 -52.193 57.173
635907 90.895 -52.216 57.175
635908 90.920 -52.239 57.178
635909 90.961 -52.256 57.182
635910 91.002 -52.274 57.188
635911 91.042 -52.293 57.194
635912 91.083 -52.310 57.202
635913 91.124 -52.328 57.211
635914 91.165 -52.346 57.219
635915 91.191 -52.366 57.231
635916 91.218 -52.389 57.244
635917 91.248 -52.415 57.251
635918 91.278 -52.442 57.257
635919 91.273 -52.474 57.271
635920 91.255 -52.507 57.280
635921 91.247 -52.539 57.285
635922 91.274 -52.561 57.290
635923 91.301 -52.584 57.290
635924 91.327 -52.606 57.282
635925 91.354 -52.628 57.261
635926 91.380 -52.650 57.237
635927 91.407 -52.672 57.210
635928 91.434 -52.695 57.181
635929 91.461 -52.716 57.154
635930 91.489 -52.737 57.136
635931 91.517 -52.757 57.124
635932 91.545 -52.777 57.110
635933 91.573 -52.797 57.093
635934 91.601 -52.818 57.072
635935 91.629 -52.838 57.047
635936 91.657 -52.859 57.024
635937 91.684 -52.879 57.001
635938 91.703 -52.894 56.968
635939 91.719 -52.908 56.927
635940 91.740 -52.925 56.887
635941 91.773 -52.949 56.849
635942 91.806 -52.973 56.809
635943 91.840 -52.998 56.761
635944 91.873 -53.023 56.704
635945 91.907 -53.047 56.644
635946 91.940 -53.072 56.587
635947 91.973 -53.096 56.534
635948 92.006 -53.120 56.481
635949 92.039 -53.145 56.428
635950 92.073 -53.170 56.375
635951 92.111 -53.187 56.323
635952 92.152 -53.201 56.270
635953 92.193 -53.214 56.218
635954 92.233 -53.227 56.166
635955 92.274 -53.241 56.116
635956 92.315 -53.255 56.067
635957 92.356 -53.268 56.018
635958 92.397 -53.281 55.971
635959 92.437 -53.295 55.926
635990 93.493 -53.726 54.927
635991 93.522 -53.740 54.923
635992 93.553 -53.756 54.920
635993 93.583 -53.770 54.918
635994 93.613 -53.784 54.916
635995 93.643 -53.799 54.916
635996 93.673 -53.814 54.917
635997 93.703 -53.828 54.919
635998 93.733 -53.843 54.922
635999 93.763 -53.857 54.926
636000 93.793 -53.872 54.930
636001 93.823 -53.887 54.933
636002 93.853 -53.901 54.937
636003 93.883 -53.916 54.942
636004 93.913 -53.930 54.949
636005 93.948 -53.944 54.955
636006 93.985 -53.958 54.963
636007 94.023 -53.972 54.971
636008 94.060 -53.985 54.979
636009 94.098 -53.999 54.988
636010 94.135 -54.013 54.997
636011 94.172 -54.027 55.009
636012 94.207 -54.046 55.025
636013 94.230 -54.059 55.044
636014 94.247 -54.067 55.063
636015 94.264 -54.075 55.077
636016 94.280 -54.083 55.081
636017 94.297 -54.091 55.079
636018 94.314 -54.099 55.074
636019 94.330 -54.107 55.065
636020 94.347 -54.115 55.045
636021 94.363 -54.123 55.029
636022 94.380 -54.131 55.017
636023 94.396 -54.139 55.008
636024 94.412 -54.148 54.997
636025 94.427 -54.157 54.988
636026 94.443 -54.166 54.980
636027 94.459 -54.174 54.975
636028 94.474 -54.182 54.968
636029 94.490 -54.192 54.955
636030 94.505 -54.200 54.937
636031 94.521 -54.209 54.914
636032 94.536 -54.217 54.885
636033 94.552 -54.226 54.858
636034 94.567 -54.235 54.830
636035 94.583 -54.243 54.801
636036 94.599 -54.252 54.771
636037 94.614 -54.261 54.737
636038 94.635 -54.274 54.698
636039 94.667 -54.288 54.658
636040 94.706 -54.308 54.617
636041 94.746 -54.335 54.578
636042 94.786 -54.361 54.542
636043 94.827 -54.388 54.505
636044 94.867 -54.414 54.466
636045 94.907 -54.442 54.429
636046 94.947 -54.468 54.389
636047 94.987 -54.494 54.348
636048 95.021 -54.511 54.302
636049 95.051 -54.527 54.250
636050 95.080 -54.543 54.194
636051 95.111 -54.560 54.137
636052 95.141 -54.577 54.081
636053 95.171 -54.593 54.031
636054 95.201 -54.610 53.989
636055 95.231 -54.627 53.947
636056 95.261 -54.643 53.904
636057 95.290 -54.660 53.860
636058 95.320 -54.677 53.815
636059 95.349 -54.694 53.768
636060 95.376 -54.712 53.721
636061 95.403 -54.731 53.675
636062 95.430 -54.749 53.628
636063 95.463 -54.762 53.582
636064 95.495 -54.776 53.537
636065 95.528 -54.789 53.495
636066 95.560 -54.803 53.459
636067 95.591 -54.818 53.426
636068 95.621 -54.831 53.396
636069 95.653 -54.846 53.367
636070 95.683 -54.860 53.336
636071 95.714 -54.875 53.304
636072 95.745 -54.889 53.269
636073 95.773 -54.900 53.232
636074 95.801 -54.911 53.194
636075 95.828 -54.922 53.154
636076 95.856 -54.934 53.115
636077 95.884 -54.944 53.076
636078 95.912 -54.956 53.037
636079 95.941 -54.968 52.996
636080 95.970 -54.979 52.956
636081 96.000 -54.991 52.915
636082 96.029 -55.003 52.874
636083 96.058 -55.015 52.833
636084 96.087 -55.019 52.792
636085 96.116 -55.024 52.748
636086 96.145 -55.028 52.706
636087 96.174 -55.033 52.660
636088 96.203 -55.038 52.610
636089 96.232 -55.042 52.557
636090 96.260 -55.052 52.496
636091 96.288 -55.062 52.433
636092 96.316 -55.072 52.363
636093 96.343 -55.082 52.292
636094 96.368 -55.090 52.225
636095 96.393 -55.098 52.160
636096 96.421 -55.105 52.093
636097 96.450 -55.112 52.022
636098 96.480 -55.120 51.948
636099 96.512 -55.129 51.873
636100 96.543 -55.138 51.798
636101 96.583 -55.150 51.725
636102 96.620 -55.169 51.652
636103 96.655 -55.194 51.582
636104 96.690 -55.218 51.511
636105 96.725 -55.242 51.441
636106 96.760 -55.267 51.372
636107 96.795 -55.291 51.304
636108 96.830 -55.315 51.236
636109 96.865 -55.340 51.170
636110 96.900 -55.364 51.105
636111 96.933 -55.386 51.040
636112 96.950 -55.385 50.976
636113 96.966 -55.384 50.913
636114 96.983 -55.383 50.852
636115 97.000 -55.383 50.794
636116 97.016 -55.382 50.737
636117 97.033 -55.381 50.681
636118 97.049 -55.380 50.628
636513 87.820 -50.635 56.445
636514 87.855 -50.656 56.451
636515 87.890 -50.677 56.457
636516 87.924 -50.698 56.461
636517 87.959 -50.718 56.462
636518 87.993 -50.738 56.461
636519 88.028 -50.760 56.452
636520 88.063 -50.780 56.441
636521 88.097 -50.801 56.428
636522 88.132 -50.821 56.413
636523 88.167 -50.842 56.396
636524 88.201 -50.863 56.378
636525 88.236 -50.883 56.353
636526 88.270 -50.904 56.327
636588 90.298 -51.994 57.065
636589 90.330 -52.002 57.075
636590 90.363 -52.011 57.085
636591 90.395 -52.018 57.095
636592 90.427 -52.026 57.105
636593 90.459 -52.034 57.114
636594 90.492 -52.042 57.122
636595 90.524 -52.050 57.130
636596 90.556 -52.058 57.136
636597 90.588 -52.067 57.141
636598 90.620 -52.074 57.146
636599 90.653 -52.083 57.149
636600 90.685 -52.090 57.153
636601 90.718 -52.098 57.156
636602 90.750 -52.107 57.158
636603 90.782 -52.114 57.158
636604 90.814 -52.123 57.160
636605 90.843 -52.134 57.162
636606 90.866 -52.146 57.165
636607 90.895 -52.156 57.170
636608 90.920 -52.179 57.174
636609 90.951 -52.199 57.179
636610 90.992 -52.217 57.185
636611 91.033 -52.236 57.195
636612 91.074 -52.253 57.205
636613 91.114 -52.271 57.216
636614 91.145 -52.291 57.225
636615 91.170 -52.311 57.235
636616 91.198 -52.333 57.250
636617 91.227 -52.360 57.267
636618 91.257 -52.387 57.272
636619 91.287 -52.413 57.277
636620 91.296 -52.443 57.293
636621 91.279 -52.476 57.303
636622 91.266 -52.509 57.308
636623 91.293 -52.531 57.308
636624 91.319 -52.553 57.301
636625 91.346 -52.575 57.289
636626 91.373 -52.598 57.266
636627 91.399 -52.620 57.240
636628 91.426 -52.643 57.213
636629 91.452 -52.665 57.185
636630 91.479 -52.687 57.157
636631 91.506 -52.709 57.142
636632 91.533 -52.731 57.131
636633 91.561 -52.751 57.122
636634 91.589 -52.772 57.110
636635 91.617 -52.792 57.090
636636 91.645 -52.813 57.061
636637 91.673 -52.833 57.038
636638 91.701 -52.854 57.010
636639 91.729 -52.874 56.979
636640 91.748 -52.889 56.942
636641 91.765 -52.904 56.902
636642 91.787 -52.920 56.861
636643 91.820 -52.945 56.819
636644 91.854 -52.970 56.774
636645 91.887 -52.994 56.718
636646 91.920 -53.019 56.660
636647 91.953 -53.043 56.603
636648 91.987 -53.067 56.550
636649 92.020 -53.092 56.498
636650 92.054 -53.117 56.445
636651 92.087 -53.141 56.390
636652 92.125 -53.158 56.337
636653 92.166 -53.172 56.284
636654 92.207 -53.185 56.231
636655 92.247 -53.198 56.177
636656 92.288 -53.212 56.127
636657 92.329 -53.226 56.076
636658 92.370 -53.239 56.027
636659 92.411 -53.252 55.979
636660 92.451 -53.266 55.932
636692 93.539 -53.705 54.912
636693 93.569 -53.720 54.910
636694 93.599 -53.734 54.907
636695 93.629 -53.749 54.906
636696 93.659 -53.763 54.905
636697 93.689 -53.778 54.907
636698 93.719 -53.793 54.909
636699 93.749 -53.807 54.912
636700 93.779 -53.822 54.916
636701 93.809 -53.837 54.920
636702 93.839 -53.851 54.923
636703 93.869 -53.866 54.929
636704 93.899 -53.880 54.935
636705 93.929 -53.895 54.941
636706 93.959 -53.910 54.948
636707 93.995 -53.923 54.956
636708 94.033 -53.937 54.965
636709 94.070 -53.950 54.973
636710 94.107 -53.964 54.981
636711 94.144 -53.978 54.993
636712 94.180 -53.997 55.007
636713 94.214 -54.018 55.025
636714 94.243 -54.035 55.043
636715 94.259 -54.043 55.059
636716 94.276 -54.051 55.068
636717 94.292 -54.059 55.069
636718 94.309 -54.067 55.066
636719 94.326 -54.075 55.060
636720 94.342 -54.083 55.053
636721 94.359 -54.091 55.038
636722 94.376 -54.099 55.028
636723 94.392 -54.107 55.023
636724 94.409 -54.115 55.017
636725 94.425 -54.123 55.007
636726 94.442 -54.131 54.995
636727 94.459 -54.139 54.985
636728 94.475 -54.147 54.977
636729 94.492 -54.155 54.967
636730 94.508 -54.163 54.951
636731 94.525 -54.171 54.931
636732 94.542 -54.179 54.907
636733 94.558 -54.187 54.877
636734 94.575 -54.195 54.850
636735 94.591 -54.203 54.823
636736 94.607 -54.212 54.796
636737 94.622 -54.220 54.767
636738 94.638 -54.229 54.736
636739 94.653 -54.238 54.699
636740 94.682 -54.252 54.662
636741 94.722 -54.279 54.624
636742 94.762 -54.306 54.588
636743 94.803 -54.333 54.555
636744 94.843 -54.359 54.520
636745 94.883 -54.386 54.482
636746 94.923 -54.412 54.445
636747 94.964 -54.439 54.407
636748 95.003 -54.464 54.364
636749 95.034 -54.480 54.317
636750 95.063 -54.498 54.264
636751 95.094 -54.515 54.206
636752 95.123 -54.531 54.146
636753 95.154 -54.548 54.091
636754 95.183 -54.565 54.046
636755 95.213 -54.581 54.005
636756 95.243 -54.598 53.963
636757 95.273 -54.615 53.919
636758 95.303 -54.631 53.874
636759 95.331 -54.649 53.827
636760 95.358 -54.667 53.780
636761 95.385 -54.685 53.735
636762 95.412 -54.703 53.689
636763 95.439 -54.721 53.642
636764 95.471 -54.735 53.595
636765 95.503 -54.749 53.549
636766 95.536 -54.762 53.508
636767 95.568 -54.776 53.474
636768 95.601 -54.789 53.442
636769 95.632 -54.803 53.411
636770 95.663 -54.818 53.382
636771 95.694 -54.832 53.352
636772 95.725 -54.847 53.320
636773 95.756 -54.861 53.287
636774 95.784 -54.872 53.251
636775 95.812 -54.884 53.214
636776 95.840 -54.895 53.176
636777 95.868 -54.906 53.139
636778 95.895 -54.917 53.100
636779 95.923 -54.928 53.062
636780 95.951 -54.939 53.021
636781 95.980 -54.951 52.980
636782 96.009 -54.963 52.939
636783 96.039 -54.975 52.897
636784 96.068 -54.986 52.855
636785 96.097 -54.991 52.813
636786 96.126 -54.996 52.769
636787 96.155 -55.000 52.725
636788 96.184 -55.005 52.678
636789 96.213 -55.010 52.629
636790 96.241 -55.021 52.577
636791 96.269 -55.031 52.518
636792 96.297 -55.041 52.454
636793 96.325 -55.051 52.386
636794 96.350 -55.060 52.320
636795 96.375 -55.068 52.260
636796 96.400 -55.075 52.203
636797 96.429 -55.083 52.143
636798 96.458 -55.090 52.075
636799 96.487 -55.097 52.000
636800 96.518 -55.106 51.923
636801 96.550 -55.115 51.846
636802 96.584 -55.125 51.769
636803 96.623 -55.137 51.694
636804 96.660 -55.157 51.622
636805 96.695 -55.181 51.550
636806 96.730 -55.205 51.479
636807 96.765 -55.229 51.409
636808 96.800 -55.254 51.339
636809 96.835 -55.278 51.270
636810 96.870 -55.302 51.203
636811 96.905 -55.326 51.138
636812 96.940 -55.351 51.071
636813 96.964 -55.360 51.007
636814 96.980 -55.359 50.943
636815 96.997 -55.358 50.880
636816 97.014 -55.358 50.821
636817 97.030 -55.357 50.764
636818 97.047 -55.356 50.707
636819 97.064 -55.354 50.652
636820 97.080 -55.354 50.598
637214 87.838 -50.599 56.419
637215 87.872 -50.620 56.425
637216 87.907 -50.641 56.429
637217 87.941 -50.661 56.431
637218 87.976 -50.682 56.432
637219 88.011 -50.703 56.431
637220 88.045 -50.723 56.423
637221 88.080 -50.744 56.411
637222 88.115 -50.764 56.398
637223 88.149 -50.785 56.383
637224 88.184 -50.806 56.367
637225 88.218 -50.826 56.348
637226 88.253 -50.847 56.324
637288 90.286 -51.953 57.046
637289 90.318 -51.960 57.057
637290 90.350 -51.969 57.067
637291 90.382 -51.977 57.076
637292 90.414 -51.985 57.085
637293 90.447 -51.993 57.095
637294 90.479 -52.001 57.104
637295 90.512 -52.009 57.112
637296 90.544 -52.017 57.119
637297 90.576 -52.025 57.125
637298 90.608 -52.033 57.130
637299 90.641 -52.041 57.134
637300 90.673 -52.049 57.138
637301 90.705 -52.057 57.142
637302 90.737 -52.065 57.146
637303 90.769 -52.073 57.148
637304 90.802 -52.081 57.150
637305 90.834 -52.089 57.153
637306 90.867 -52.097 57.155
637307 90.890 -52.113 57.161
637308 90.920 -52.119 57.166
637309 90.945 -52.142 57.173
637310 90.982 -52.161 57.181
637311 91.023 -52.179 57.191
637312 91.064 -52.197 57.204
637313 91.099 -52.215 57.218
637314 91.124 -52.235 57.231
637315 91.150 -52.255 57.242
637316 91.177 -52.278 57.254
637317 91.207 -52.305 57.268
637318 91.237 -52.331 57.286
637319 91.266 -52.358 57.293
637320 91.296 -52.384 57.296
637321 91.319 -52.412 57.305
637322 91.302 -52.445 57.314
637323 91.285 -52.478 57.317
637324 91.311 -52.501 57.311
637325 91.338 -52.523 57.305
637326 91.365 -52.546 57.289
637327 91.391 -52.568 57.265
637328 91.418 -52.590 57.238
637329 91.445 -52.612 57.212
637330 91.471 -52.635 57.187
637331 91.498 -52.657 57.160
637332 91.524 -52.680 57.147
637333 91.551 -52.702 57.136
637334 91.578 -52.724 57.129
637335 91.605 -52.745 57.119
637336 91.633 -52.766 57.099
637337 91.661 -52.786 57.069
637338 91.689 -52.807 57.043
637339 91.717 -52.827 57.012
637340 91.745 -52.847 56.979
637341 91.773 -52.868 56.942
637342 91.795 -52.885 56.902
637343 91.811 -52.899 56.860
637344 91.834 -52.916 56.819
637345 91.867 -52.941 56.775
637346 91.901 -52.966 56.724
637347 91.934 -52.990 56.670
637348 91.967 -53.015 56.618
637349 92.001 -53.039 56.566
637350 92.034 -53.064 56.514
637351 92.067 -53.089 56.460
637352 92.101 -53.113 56.406
637353 92.139 -53.129 56.352
637354 92.180 -53.142 56.298
637355 92.221 -53.156 56.243
637356 92.262 -53.170 56.189
637357 92.302 -53.183 56.138
637358 92.343 -53.197 56.086
637359 92.384 -53.210 56.035
637360 92.425 -53.223 55.986
637361 92.465 -53.237 55.938
637394 93.585 -53.684 54.900
637395 93.615 -53.699 54.897
637396 93.645 -53.713 54.897
637397 93.675 -53.728 54.895
637398 93.705 -53.743 54.897
637399 93.735 -53.757 54.900
637400 93.765 -53.772 54.903
637401 93.795 -53.786 54.907
637402 93.825 -53.801 54.910
637403 93.855 -53.816 54.914
637404 93.885 -53.830 54.921
637405 93.915 -53.844 54.928
637406 93.945 -53.860 54.935
637407 93.975 -53.874 54.942
637408 94.005 -53.888 54.949
637409 94.042 -53.902 54.959
637410 94.079 -53.916 54.968
637411 94.117 -53.929 54.976
637412 94.152 -53.947 54.988
637413 94.187 -53.969 55.003
637414 94.221 -53.990 55.020
637415 94.255 -54.011 55.035
637416 94.271 -54.019 55.047
637417 94.288 -54.027 55.051
637418 94.305 -54.035 55.050
637419 94.321 -54.043 55.048
637420 94.338 -54.051 55.045
637421 94.354 -54.059 55.042
637422 94.371 -54.067 55.032
637423 94.388 -54.076 55.029
637424 94.404 -54.083 55.026
637425 94.421 -54.091 55.021
637426 94.437 -54.099 55.013
637427 94.454 -54.107 55.003
637428 94.471 -54.115 54.990
637429 94.487 -54.123 54.978
637430 94.504 -54.131 54.964
637431 94.521 -54.139 54.946
637432 94.537 -54.147 54.924
637433 94.554 -54.155 54.900
637434 94.570 -54.163 54.870
637435 94.587 -54.171 54.844
637436 94.604 -54.179 54.818
637437 94.620 -54.187 54.793
637438 94.637 -54.195 54.766
637439 94.653 -54.203 54.737
637440 94.672 -54.212 54.702
637441 94.696 -54.225 54.667
637442 94.736 -54.251 54.633
637443 94.776 -54.277 54.600
637444 94.817 -54.304 54.569
637445 94.857 -54.330 54.535
637446 94.897 -54.357 54.496
637447 94.937 -54.383 54.458
637448 94.977 -54.411 54.420
637449 95.018 -54.437 54.376
637450 95.046 -54.452 54.327
637451 95.076 -54.468 54.273
637452 95.106 -54.485 54.213
637453 95.136 -54.502 54.154
637454 95.166 -54.518 54.105
637455 95.196 -54.535 54.062
637456 95.226 -54.552 54.022
637457 95.256 -54.568 53.980
637458 95.285 -54.585 53.935
637459 95.314 -54.602 53.889
637460 95.340 -54.621 53.841
637461 95.367 -54.639 53.795
637462 95.394 -54.657 53.750
637463 95.421 -54.675 53.703
637464 95.448 -54.693 53.654
637465 95.479 -54.708 53.607
637466 95.512 -54.721 53.561
637467 95.544 -54.735 53.522
637468 95.577 -54.749 53.491
637469 95.609 -54.762 53.459
637470 95.642 -54.776 53.427
637471 95.674 -54.790 53.397
637472 95.705 -54.804 53.368
637473 95.736 -54.818 53.339
637474 95.767 -54.833 53.307
637475 95.796 -54.845 53.271
637476 95.823 -54.856 53.236
637477 95.851 -54.867 53.200
637478 95.879 -54.878 53.163
637479 95.907 -54.889 53.125
637480 95.935 -54.900 53.087
637481 95.962 -54.912 53.047
637482 95.990 -54.923 53.005
637483 96.019 -54.934 52.963
637484 96.048 -54.946 52.920
637485 96.078 -54.958 52.877
637486 96.107 -54.963 52.834
637487 96.135 -54.968 52.788
637488 96.165 -54.972 52.741
637489 96.193 -54.979 52.693
637490 96.221 -54.990 52.644
637491 96.250 -55.000 52.592
637492 96.278 -55.010 52.535
637493 96.306 -55.020 52.475
637494 96.333 -55.029 52.412
637495 96.357 -55.038 52.353
637496 96.382 -55.046 52.299
637497 96.408 -55.054 52.247
637498 96.437 -55.060 52.191
637499 96.465 -55.067 52.127
637500 96.494 -55.075 52.056
637501 96.525 -55.083 51.980
637502 96.556 -55.091 51.901
637503 96.588 -55.101 51.821
637504 96.624 -55.111 51.741
637505 96.664 -55.124 51.667
637506 96.700 -55.143 51.592
637507 96.736 -55.167 51.521
637508 96.770 -55.191 51.449
637509 96.806 -55.217 51.378
637510 96.840 -55.241 51.308
637511 96.876 -55.265 51.240
637512 96.910 -55.289 51.172
637513 96.946 -55.314 51.105
637514 96.978 -55.335 51.040
637515 96.995 -55.334 50.974
637516 97.012 -55.333 50.910
637517 97.028 -55.332 50.850
637518 97.045 -55.332 50.792
637519 97.062 -55.330 50.733
637520 97.078 -55.329 50.677
637521 97.095 -55.329 50.623
637522 97.111 -55.328 50.570
637523 97.128 -55.327 50.518
637915 87.855 -50.563 56.393
637916 87.890 -50.584 56.397
637917 87.924 -50.604 56.400
637918 87.959 -50.625 56.401
637919 87.993 -50.646 56.401
637920 88.028 -50.666 56.398
637921 88.063 -50.687 56.390
637922 88.097 -50.707 56.379
637923 88.132 -50.729 56.366
637924 88.167 -50.749 56.351
637925 88.201 -50.770 56.335
637989 90.306 -51.920 57.036
637990 90.338 -51.927 57.045
637991 90.370 -51.935 57.055
637992 90.402 -51.944 57.065
637993 90.435 -51.951 57.075
637994 90.467 -51.960 57.085
637995 90.499 -51.967 57.092
637996 90.531 -51.976 57.100
637997 90.564 -51.983 57.107
637998 90.596 -51.992 57.114
637999 90.628 -52.000 57.119
638000 90.661 -52.007 57.122
638001 90.693 -52.016 57.127
638002 90.725 -52.023 57.131
638003 90.757 -52.032 57.135
638004 90.790 -52.039 57.138
638005 90.822 -52.048 57.141
638006 90.854 -52.056 57.144
638007 90.886 -52.064 57.149
638008 90.919 -52.072 57.155
638009 90.945 -52.082 57.163
638010 90.972 -52.104 57.172
638011 91.013 -52.122 57.184
638012 91.053 -52.140 57.200
638013 91.078 -52.160 57.217
638014 91.104 -52.180 57.233
638015 91.129 -52.199 57.246
638016 91.156 -52.223 57.260
638017 91.186 -52.249 57.273
638018 91.216 -52.276 57.285
638019 91.246 -52.302 57.301
638020 91.276 -52.328 57.307
638021 91.305 -52.355 57.308
638022 91.335 -52.381 57.313
638023 91.326 -52.414 57.319
638024 91.308 -52.447 57.319
638025 91.330 -52.471 57.309
638026 91.357 -52.493 57.298
638027 91.383 -52.515 57.282
638028 91.410 -52.538 57.259
638029 91.437 -52.560 57.231
638030 91.463 -52.583 57.205
638031 91.490 -52.605 57.182
638032 91.517 -52.627 57.161
638033 91.543 -52.650 57.147
638034 91.570 -52.672 57.139
638035 91.597 -52.694 57.131
638036 91.623 -52.716 57.120
638037 91.650 -52.739 57.102
638038 91.677 -52.760 57.074
638039 91.705 -52.780 57.044
638040 91.733 -52.800 57.011
638041 91.761 -52.821 56.976
638042 91.789 -52.841 56.938
638043 91.817 -52.862 56.898
638044 91.841 -52.880 56.857
638045 91.858 -52.894 56.817
638046 91.881 -52.913 56.774
638047 91.915 -52.937 56.727
638048 91.948 -52.962 56.677
638049 91.981 -52.986 56.628
638050 92.015 -53.010 56.580
638051 92.048 -53.035 56.532
638052 92.081 -53.060 56.478
638053 92.115 -53.084 56.422
638054 92.154 -53.101 56.367
638055 92.195 -53.114 56.312
638056 92.235 -53.127 56.256
638057 92.276 -53.141 56.201
638058 92.316 -53.154 56.148
638059 92.357 -53.168 56.095
638060 92.398 -53.181 56.043
638061 92.439 -53.195 55.993
638062 92.480 -53.208 55.944
638095 93.601 -53.649 54.892
638096 93.631 -53.663 54.889
638097 93.661 -53.678 54.888
638098 93.691 -53.693 54.887
638099 93.721 -53.707 54.888
638100 93.751 -53.722 54.890
638101 93.781 -53.736 54.894
638102 93.811 -53.750 54.898
638103 93.841 -53.766 54.902
638104 93.871 -53.780 54.907
638105 93.901 -53.794 54.913
638106 93.931 -53.809 54.920
638107 93.961 -53.824 54.928
638108 93.991 -53.838 54.935
638109 94.021 -53.853 54.944
638110 94.052 -53.867 54.953
638111 94.089 -53.881 54.963
638112 94.125 -53.897 54.972
638113 94.159 -53.919 54.983
638114 94.193 -53.941 54.998
638115 94.227 -53.962 55.014
638116 94.262 -53.983 55.024
638117 94.283 -53.995 55.032
638118 94.300 -54.004 55.034
638119 94.317 -54.011 55.034
638120 94.333 -54.019 55.034
638121 94.350 -54.027 55.036
638122 94.366 -54.035 55.035
638123 94.383 -54.043 55.029
638124 94.400 -54.051 55.026
638125 94.416 -54.059 55.024
638126 94.433 -54.067 55.022
638127 94.450 -54.076 55.019
638128 94.466 -54.083 55.010
638129 94.483 -54.091 54.994
638130 94.499 -54.099 54.978
638131 94.516 -54.107 54.960
638132 94.533 -54.115 54.940
638133 94.549 -54.123 54.918
638134 94.566 -54.131 54.895
638135 94.582 -54.139 54.867
638136 94.599 -54.147 54.841
638137 94.616 -54.155 54.817
638138 94.632 -54.163 54.793
638139 94.649 -54.171 54.769
638140 94.671 -54.182 54.742
638141 94.695 -54.195 54.708
638142 94.718 -54.207 54.675
638143 94.749 -54.223 54.643
638144 94.788 -54.249 54.613
638145 94.829 -54.276 54.583
638146 94.869 -54.302 54.549
638147 94.909 -54.329 54.509
638148 94.950 -54.355 54.470
638149 94.990 -54.382 54.430
638150 95.030 -54.408 54.385
638151 95.061 -54.426 54.334
638152 95.090 -54.441 54.279
638153 95.118 -54.455 54.220
638154 95.148 -54.472 54.163
638155 95.178 -54.489 54.119
638156 95.208 -54.505 54.079
638157 95.238 -54.522 54.039
638158 95.268 -54.540 53.996
638159 95.296 -54.557 53.950
638160 95.323 -54.575 53.903
638161 95.349 -54.593 53.854
638162 95.376 -54.612 53.807
638163 95.403 -54.630 53.760
638164 95.430 -54.648 53.713
638165 95.457 -54.666 53.664
638166 95.487 -54.681 53.617
638167 95.520 -54.694 53.574
638168 95.553 -54.708 53.538
638169 95.585 -54.721 53.507
638170 95.618 -54.735 53.476
638171 95.650 -54.749 53.446
638172 95.683 -54.762 53.417
638173 95.715 -54.776 53.389
638174 95.747 -54.790 53.360
638175 95.777 -54.804 53.329
638176 95.807 -54.817 53.295
638177 95.835 -54.828 53.260
638178 95.863 -54.839 53.225
638179 95.890 -54.850 53.189
638180 95.918 -54.862 53.153
638181 95.946 -54.873 53.114
638182 95.974 -54.884 53.072
638183 96.002 -54.895 53.029
638184 96.029 -54.906 52.986
638185 96.058 -54.918 52.944
638186 96.088 -54.930 52.900
638187 96.116 -54.935 52.854
638188 96.146 -54.940 52.807
638189 96.174 -54.948 52.758
638190 96.202 -54.959 52.708
638191 96.230 -54.969 52.658
638192 96.258 -54.979 52.606
638193 96.287 -54.989 52.552
638194 96.315 -55.000 52.496
638195 96.339 -55.007 52.439
638196 96.364 -55.016 52.387
638197 96.388 -55.023 52.339
638198 96.415 -55.031 52.291
638199 96.444 -55.038 52.238
638200 96.473 -55.045 52.178
638201 96.502 -55.052 52.110
638202 96.532 -55.060 52.036
638203 96.564 -55.069 51.958
638204 96.595 -55.078 51.877
638205 96.626 -55.086 51.794
638206 96.664 -55.098 51.717
638207 96.704 -55.111 51.640
638208 96.741 -55.130 51.567
638209 96.776 -55.154 51.494
638210 96.811 -55.179 51.422
638211 96.846 -55.203 51.350
638212 96.881 -55.227 51.281
638213 96.916 -55.251 51.211
638214 96.951 -55.276 51.144
638215 96.986 -55.300 51.077
638216 97.009 -55.309 51.010
638217 97.026 -55.308 50.945
638218 97.042 -55.307 50.884
638219 97.059 -55.306 50.824
638220 97.076 -55.305 50.765
638221 97.092 -55.304 50.708
638222 97.109 -55.304 50.652
638223 97.125 -55.303 50.598
638224 97.142 -55.302 50.546
638618 87.941 -50.568 56.369
638619 87.976 -50.589 56.370
638620 88.011 -50.610 56.369
638621 88.045 -50.630 56.365
638622 88.080 -50.650 56.357
638623 88.115 -50.672 56.346
638624 88.149 -50.692 56.333
638690 90.325 -51.886 57.026
638691 90.358 -51.894 57.035
638692 90.390 -51.902 57.044
638693 90.423 -51.910 57.054
638694 90.455 -51.918 57.063
638695 90.487 -51.926 57.072
638696 90.519 -51.934 57.080
638697 90.551 -51.942 57.087
638698 90.584 -51.950 57.095
638699 90.616 -51.958 57.101
638700 90.648 -51.967 57.106
638701 90.680 -51.974 57.110
638702 90.713 -51.982 57.114
638703 90.745 -51.990 57.119
638704 90.778 -51.998 57.124
638705 90.810 -52.006 57.128
638706 90.842 -52.014 57.131
638707 90.874 -52.023 57.135
638708 90.906 -52.030 57.141
638709 90.939 -52.039 57.149
638710 90.969 -52.048 57.158
638711 91.003 -52.065 57.171
638712 91.032 -52.084 57.187
638713 91.058 -52.105 57.207
638714 91.083 -52.124 57.224
638715 91.108 -52.144 57.241
638716 91.136 -52.167 57.260
638717 91.166 -52.194 57.278
638718 91.195 -52.220 57.291
638719 91.225 -52.246 57.300
638720 91.255 -52.273 57.309
638721 91.285 -52.299 57.311
638722 91.314 -52.326 57.312
638723 91.345 -52.352 57.316
638724 91.349 -52.383 57.318
638725 91.332 -52.416 57.316
638726 91.349 -52.441 57.304
638727 91.376 -52.463 57.289
638728 91.402 -52.486 57.274
638729 91.429 -52.508 57.254
638730 91.455 -52.530 57.228
638731 91.482 -52.553 57.201
638732 91.509 -52.575 57.178
638733 91.535 -52.597 57.158
638734 91.562 -52.619 57.145
638735 91.589 -52.642 57.135
638736 91.615 -52.664 57.126
638737 91.642 -52.687 57.116
638738 91.669 -52.709 57.103
638739 91.695 -52.731 57.077
638740 91.722 -52.753 57.044
638741 91.749 -52.775 57.008
638742 91.777 -52.795 56.970
638743 91.805 -52.816 56.932
638744 91.833 -52.836 56.893
638745 91.861 -52.857 56.852
638746 91.887 -52.876 56.813
638747 91.904 -52.890 56.772
638748 91.929 -52.909 56.728
638749 91.962 -52.933 56.683
638750 91.995 -52.957 56.639
638751 92.028 -52.982 56.595
638752 92.062 -53.007 56.549
638753 92.095 -53.031 56.495
638754 92.129 -53.056 56.439
638755 92.168 -53.072 56.381
638756 92.209 -53.085 56.324
638757 92.249 -53.098 56.267
638758 92.290 -53.112 56.211
638759 92.331 -53.126 56.157
638760 92.371 -53.139 56.103
638761 92.412 -53.152 56.049
638762 92.453 -53.166 55.998
638763 92.500 -53.173 55.948
638797 93.647 -53.628 54.881
638798 93.677 -53.642 54.879
638799 93.707 -53.657 54.879
638800 93.737 -53.672 54.880
638801 93.767 -53.686 54.882
638802 93.797 -53.700 54.886
638803 93.827 -53.715 54.891
638804 93.857 -53.730 54.895
638805 93.887 -53.744 54.900
638806 93.917 -53.759 54.907
638807 93.947 -53.774 54.914
638808 93.977 -53.788 54.922
638809 94.007 -53.803 54.930
638810 94.037 -53.817 54.940
638811 94.067 -53.832 54.951
638812 94.098 -53.847 54.962
638813 94.132 -53.869 54.970
638814 94.166 -53.891 54.978
638815 94.200 -53.913 54.991
638816 94.234 -53.934 55.003
638817 94.268 -53.955 55.012
638818 94.295 -53.972 55.017
638819 94.312 -53.979 55.019
638820 94.329 -53.988 55.021
638821 94.345 -53.995 55.025
638822 94.362 -54.004 55.031
638823 94.379 -54.011 55.032
638824 94.395 -54.019 55.027
638825 94.412 -54.027 55.021
638826 94.428 -54.035 55.020
638827 94.445 -54.043 55.023
638828 94.462 -54.051 55.024
638829 94.478 -54.059 55.016
638830 94.495 -54.067 54.998
638831 94.511 -54.076 54.978
638832 94.528 -54.083 54.956
638833 94.545 -54.091 54.935
638834 94.561 -54.099 54.913
638835 94.578 -54.107 54.892
638836 94.595 -54.115 54.867
638837 94.611 -54.123 54.844
638838 94.628 -54.131 54.823
638839 94.647 -54.141 54.801
638840 94.671 -54.153 54.779
638841 94.694 -54.166 54.755
638842 94.718 -54.178 54.722
638843 94.742 -54.190 54.688
638844 94.765 -54.202 54.656
638845 94.801 -54.221 54.625
638846 94.841 -54.247 54.596
638847 94.882 -54.274 54.563
638848 94.921 -54.301 54.522
638849 94.962 -54.327 54.481
638850 95.002 -54.354 54.436
638851 95.042 -54.380 54.389
638852 95.076 -54.401 54.337
638853 95.105 -54.415 54.282
638854 95.133 -54.430 54.225
638855 95.162 -54.445 54.175
638856 95.191 -54.460 54.134
638857 95.220 -54.477 54.095
638858 95.250 -54.493 54.054
638859 95.278 -54.511 54.011
638860 95.305 -54.529 53.964
638861 95.332 -54.547 53.914
638862 95.358 -54.565 53.865
638863 95.385 -54.583 53.816
638864 95.412 -54.602 53.769
638865 95.439 -54.620 53.721
638866 95.466 -54.638 53.673
638867 95.496 -54.654 53.628
638868 95.528 -54.668 53.589
638869 95.561 -54.681 53.555
638870 95.594 -54.694 53.525
638871 95.626 -54.708 53.496
638872 95.659 -54.721 53.467
638873 95.691 -54.735 53.439
638874 95.724 -54.749 53.412
638875 95.756 -54.762 53.384
638876 95.789 -54.776 53.355
638877 95.818 -54.790 53.322
638878 95.846 -54.800 53.288
638879 95.874 -54.812 53.252
638880 95.902 -54.823 53.216
638881 95.930 -54.834 53.178
638882 95.957 -54.845 53.139
638883 95.985 -54.856 53.097
638884 96.013 -54.867 53.055
638885 96.041 -54.878 53.013
638886 96.068 -54.890 52.971
638887 96.097 -54.901 52.928
638888 96.126 -54.907 52.880
638889 96.155 -54.917 52.830
638890 96.183 -54.928 52.779
638891 96.211 -54.938 52.726
638892 96.239 -54.948 52.674
638893 96.267 -54.958 52.623
638894 96.296 -54.969 52.568
638895 96.322 -54.977 52.517
638896 96.346 -54.985 52.469
638897 96.371 -54.994 52.423
638898 96.395 -55.001 52.380
638899 96.423 -55.009 52.335
638900 96.452 -55.016 52.284
638901 96.480 -55.022 52.228
638902 96.509 -55.030 52.161
638903 96.539 -55.038 52.090
638904 96.570 -55.046 52.016
638905 96.601 -55.055 51.936
638906 96.632 -55.063 51.851
638907 96.665 -55.072 51.770
638908 96.704 -55.085 51.693
638909 96.744 -55.097 51.618
638910 96.781 -55.117 51.543
638911 96.816 -55.141 51.469
638912 96.851 -55.166 51.396
638913 96.886 -55.190 51.326
638914 96.921 -55.214 51.254
638915 96.956 -55.238 51.185
638916 96.991 -55.263 51.117
638917 97.024 -55.284 51.049
638918 97.040 -55.282 50.984
638919 97.057 -55.282 50.921
638920 97.074 -55.281 50.860
638921 97.090 -55.280 50.800
638922 97.107 -55.279 50.741
638923 97.123 -55.279 50.684
638924 97.140 -55.278 50.629
638925 97.157 -55.277 50.575
638926 97.186 -55.276 50.523
639391 90.345 -51.853 57.016
639392 90.378 -51.860 57.025
639393 90.410 -51.869 57.034
639394 90.442 -51.876 57.043
639395 90.474 -51.885 57.051
639396 90.507 -51.893 57.059
639397 90.539 -51.901 57.067
639398 90.572 -51.909 57.074
639399 90.604 -51.917 57.081
639400 90.636 -51.925 57.088
639401 90.668 -51.932 57.093
639402 90.700 -51.941 57.097
639403 90.733 -51.949 57.102
639404 90.765 -51.957 57.107
639405 90.797 -51.965 57.113
639406 90.830 -51.973 57.118
639407 90.862 -51.981 57.121
639408 90.894 -51.989 57.125
639409 90.927 -51.997 57.132
639410 90.959 -52.005 57.142
639411 90.990 -52.014 57.154
639412 91.012 -52.029 57.169
639413 91.037 -52.049 57.189
639414 91.063 -52.069 57.209
639415 91.088 -52.089 57.227
639416 91.115 -52.111 57.247
639417 91.145 -52.138 57.267
639418 91.175 -52.164 57.284
639419 91.205 -52.191 57.294
639420 91.234 -52.217 57.302
639421 91.264 -52.244 57.306
639422 91.294 -52.271 57.306
639423 91.324 -52.297 57.305
639424 91.354 -52.323 57.305
639425 91.373 -52.351 57.308
639426 91.355 -52.385 57.307
639427 91.368 -52.411 57.296
639428 91.394 -52.433 57.277
639429 91.421 -52.456 57.266
639430 91.448 -52.477 57.252
639431 91.474 -52.500 57.229
639432 91.501 -52.522 57.200
639433 91.527 -52.545 57.176
639434 91.554 -52.567 57.154
639435 91.581 -52.590 57.140
639436 91.607 -52.612 57.129
639437 91.634 -52.634 57.119
639438 91.661 -52.656 57.110
639439 91.687 -52.679 57.100
639440 91.714 -52.701 57.075
639441 91.741 -52.723 57.039
639442 91.768 -52.746 57.004
639443 91.794 -52.768 56.967
639444 91.821 -52.789 56.928
639445 91.850 -52.810 56.888
639446 91.877 -52.830 56.850
639447 91.905 -52.850 56.811
639448 91.933 -52.871 56.771
639449 91.950 -52.885 56.730
639450 91.976 -52.905 56.690
639451 92.009 -52.929 56.650
639452 92.042 -52.954 56.609
639453 92.076 -52.978 56.565
639454 92.109 -53.003 56.513
639455 92.142 -53.027 56.455
639456 92.182 -53.042 56.395
639457 92.223 -53.056 56.335
639458 92.264 -53.070 56.278
639459 92.304 -53.083 56.221
639460 92.345 -53.097 56.165
639461 92.385 -53.110 56.110
639462 92.427 -53.123 56.055
639463 92.475 -53.129 56.003
639464 92.522 -53.136 55.952
639465 92.570 -53.142 55.901
639498 93.663 -53.592 54.874
639499 93.693 -53.606 54.873
639500 93.723 -53.621 54.872
639501 93.753 -53.636 54.874
639502 93.783 -53.650 54.876
639503 93.813 -53.665 54.879
639504 93.843 -53.680 54.884
639505 93.873 -53.694 54.889
639506 93.903 -53.709 54.894
639507 93.933 -53.723 54.900
639508 93.963 -53.738 54.908
639509 93.993 -53.753 54.916
639510 94.023 -53.767 54.925
639511 94.053 -53.784 54.935
639512 94.084 -53.803 54.947
639513 94.112 -53.821 54.957
639514 94.139 -53.841 54.965
639515 94.173 -53.863 54.972
639516 94.207 -53.884 54.983
639517 94.241 -53.906 54.992
639518 94.275 -53.927 55.000
639519 94.308 -53.947 55.006
639520 94.324 -53.955 55.009
639521 94.341 -53.963 55.012
639522 94.357 -53.972 55.017
639523 94.374 -53.979 55.021
639524 94.391 -53.988 55.022
639525 94.407 -53.995 55.018
639526 94.424 -54.004 55.015
639527 94.440 -54.011 55.019
639528 94.457 -54.019 55.025
639529 94.474 -54.027 55.026
639530 94.490 -54.035 55.019
639531 94.507 -54.043 54.999
639532 94.524 -54.051 54.976
639533 94.540 -54.060 54.953
639534 94.557 -54.067 54.931
639535 94.573 -54.076 54.912
639536 94.590 -54.083 54.892
639537 94.607 -54.091 54.869
639538 94.623 -54.099 54.850
639539 94.646 -54.111 54.831
639540 94.670 -54.123 54.812
639541 94.693 -54.136 54.792
639542 94.717 -54.148 54.771
639543 94.741 -54.160 54.740
639544 94.764 -54.173 54.705
639545 94.788 -54.185 54.671
639546 94.817 -54.201 54.639
639547 94.853 -54.220 54.608
639548 94.894 -54.245 54.575
639549 94.934 -54.272 54.532
639550 94.974 -54.299 54.487
639551 95.014 -54.326 54.439
639552 95.054 -54.352 54.390
639553 95.091 -54.375 54.338
639554 95.120 -54.389 54.285
639555 95.148 -54.405 54.233
639556 95.177 -54.419 54.188
639557 95.205 -54.433 54.148
639558 95.234 -54.449 54.109
639559 95.260 -54.465 54.067
639560 95.287 -54.483 54.021
639561 95.314 -54.502 53.973
639562 95.341 -54.520 53.922
639563 95.368 -54.538 53.872
639564 95.394 -54.556 53.823
639565 95.421 -54.574 53.776
639566 95.448 -54.593 53.729
639567 95.475 -54.611 53.683
639568 95.504 -54.627 53.640
639569 95.537 -54.640 53.604
639570 95.569 -54.654 53.573
639571 95.602 -54.668 53.545
639572 95.634 -54.681 53.517
639573 95.667 -54.694 53.489
639574 95.699 -54.708 53.462
639575 95.732 -54.721 53.436
639576 95.764 -54.735 53.409
639577 95.797 -54.749 53.381
639578 95.829 -54.762 53.350
639579 95.857 -54.775 53.315
639580 95.885 -54.786 53.279
639581 95.912 -54.797 53.241
639582 95.940 -54.809 53.202
639583 95.968 -54.819 53.164
639584 95.996 -54.831 53.123
639585 96.024 -54.842 53.081
639586 96.051 -54.853 53.040
639587 96.079 -54.864 52.997
639588 96.107 -54.875 52.955
639589 96.135 -54.886 52.906
639590 96.163 -54.896 52.856
639591 96.192 -54.906 52.804
639592 96.220 -54.916 52.752
639593 96.248 -54.927 52.699
639594 96.276 -54.938 52.649
639595 96.304 -54.947 52.598
639596 96.328 -54.955 52.550
639597 96.353 -54.963 52.505
639598 96.377 -54.971 52.462
639599 96.402 -54.979 52.419
639600 96.431 -54.986 52.375
639601 96.459 -54.994 52.327
639602 96.488 -55.000 52.272
639603 96.517 -55.007 52.209
639604 96.545 -55.015 52.140
639605 96.577 -55.023 52.068
639606 96.608 -55.032 51.991
639607 96.640 -55.041 51.910
639608 96.671 -55.050 51.832
639609 96.705 -55.060 51.756
639610 96.745 -55.072 51.678
639611 96.785 -55.085 51.599
639612 96.821 -55.104 51.522
639613 96.857 -55.128 51.447
639614 96.892 -55.152 51.374
639615 96.927 -55.176 51.302
639616 96.962 -55.201 51.230
639617 96.997 -55.226 51.160
639618 97.032 -55.250 51.092
639619 97.054 -55.257 51.025
639620 97.071 -55.257 50.960
639621 97.088 -55.256 50.898
639622 97.104 -55.255 50.837
639623 97.121 -55.254 50.776
639624 97.138 -55.254 50.718
639625 97.154 -55.253 50.661
639626 97.171 -55.252 50.607
639627 97.202 -55.251 50.554
639628 97.237 -55.250 50.501
640092 90.366 -51.819 57.005
640093 90.398 -51.827 57.013
640094 90.430 -51.835 57.021
640095 90.462 -51.843 57.030
640096 90.495 -51.851 57.038
640097 90.527 -51.860 57.046
640098 90.559 -51.867 57.054
640099 90.592 -51.876 57.061
640100 90.624 -51.883 57.068
640101 90.656 -51.891 57.075
640102 90.688 -51.899 57.080
640103 90.721 -51.907 57.084
640104 90.753 -51.916 57.089
640105 90.785 -51.923 57.094
640106 90.817 -51.932 57.101
640107 90.850 -51.939 57.107
640108 90.882 -51.948 57.111
640109 90.914 -51.955 57.117
640110 90.947 -51.964 57.125
640111 90.979 -51.972 57.136
640112 90.995 -51.986 57.150
640113 91.016 -51.995 57.165
640114 91.042 -52.014 57.183
640115 91.067 -52.033 57.200
640116 91.095 -52.056 57.223
640117 91.124 -52.083 57.248
640118 91.154 -52.109 57.268
640119 91.184 -52.136 57.281
640120 91.214 -52.162 57.289
640121 91.243 -52.189 57.297
640122 91.274 -52.214 57.296
640123 91.303 -52.241 57.296
640124 91.333 -52.268 57.293
640125 91.363 -52.294 57.292
640126 91.392 -52.321 57.295
640127 91.379 -52.354 57.296
640128 91.386 -52.380 57.286
640129 91.413 -52.403 57.267
640130 91.440 -52.425 57.256
640131 91.466 -52.448 57.245
640132 91.493 -52.470 57.227
640133 91.520 -52.493 57.203
640134 91.546 -52.515 57.176
640135 91.573 -52.537 57.151
640136 91.600 -52.559 57.133
640137 91.626 -52.582 57.120
640138 91.653 -52.604 57.110
640139 91.679 -52.626 57.102
640140 91.706 -52.649 57.093
640141 91.733 -52.671 57.069
640142 91.759 -52.694 57.034
640143 91.786 -52.716 57.000
640144 91.813 -52.738 56.964
640145 91.840 -52.760 56.925
640146 91.866 -52.783 56.884
640147 91.893 -52.803 56.847
640148 91.922 -52.824 56.810
640149 91.949 -52.844 56.772
640150 91.978 -52.865 56.733
640151 91.996 -52.880 56.695
640152 92.023 -52.901 56.657
640153 92.056 -52.926 56.617
640154 92.090 -52.950 56.573
640155 92.123 -52.974 56.523
640156 92.156 -52.999 56.466
640157 92.196 -53.013 56.405
640158 92.237 -53.027 56.344
640159 92.278 -53.041 56.287
640160 92.318 -53.054 56.229
640161 92.359 -53.068 56.172
640162 92.402 -53.079 56.116
640163 92.449 -53.086 56.061
640164 92.497 -53.092 56.008
640165 92.545 -53.098 55.956
640166 92.593 -53.104 55.904
640200 93.709 -53.571 54.867
640201 93.739 -53.586 54.866
640202 93.769 -53.600 54.867
640203 93.799 -53.615 54.869
640204 93.829 -53.630 54.873
640205 93.859 -53.644 54.877
640206 93.889 -53.659 54.882
640207 93.919 -53.673 54.887
640208 93.949 -53.687 54.894
640209 93.979 -53.703 54.902
640210 94.010 -53.722 54.910
640211 94.040 -53.740 54.919
640212 94.071 -53.759 54.930
640213 94.101 -53.778 54.941
640214 94.132 -53.797 54.952
640215 94.147 -53.813 54.959
640216 94.180 -53.834 54.965
640217 94.214 -53.856 54.974
640218 94.248 -53.878 54.981
640219 94.282 -53.899 54.988
640220 94.316 -53.921 54.995
640221 94.336 -53.932 54.999
640222 94.353 -53.939 55.003
640223 94.369 -53.947 55.008
640224 94.386 -53.955 55.010
640225 94.403 -53.963 55.011
640226 94.419 -53.972 55.009
640227 94.436 -53.979 55.009
640228 94.453 -53.988 55.014
640229 94.469 -53.995 55.021
640230 94.486 -54.004 55.021
640231 94.502 -54.011 55.012
640232 94.519 -54.019 54.993
640233 94.536 -54.027 54.971
640234 94.552 -54.035 54.949
640235 94.569 -54.043 54.930
640236 94.585 -54.051 54.912
640237 94.602 -54.060 54.896
640238 94.622 -54.070 54.875
640239 94.645 -54.082 54.859
640240 94.669 -54.094 54.843
640241 94.693 -54.107 54.825
640242 94.716 -54.119 54.808
640243 94.740 -54.131 54.790
640244 94.764 -54.143 54.762
640245 94.787 -54.156 54.726
640246 94.811 -54.168 54.689
640247 94.835 -54.180 54.654
640248 94.870 -54.198 54.621
640249 94.906 -54.217 54.586
640250 94.946 -54.244 54.541
640251 94.986 -54.270 54.492
640252 95.027 -54.297 54.442
640253 95.067 -54.324 54.392
640254 95.106 -54.350 54.341
640255 95.135 -54.364 54.291
640256 95.163 -54.379 54.243
640257 95.192 -54.393 54.202
640258 95.220 -54.408 54.161
640259 95.247 -54.423 54.120
640260 95.271 -54.439 54.076
640261 95.296 -54.455 54.030
640262 95.323 -54.474 53.979
640263 95.349 -54.492 53.928
640264 95.377 -54.510 53.877
640265 95.403 -54.528 53.828
640266 95.430 -54.546 53.782
640267 95.457 -54.565 53.739
640268 95.484 -54.583 53.696
640269 95.513 -54.599 53.658
640270 95.545 -54.613 53.623
640271 95.578 -54.627 53.593
640272 95.610 -54.640 53.566
640273 95.642 -54.654 53.538
640274 95.675 -54.668 53.511
640275 95.708 -54.681 53.484
640276 95.740 -54.694 53.459
640277 95.773 -54.708 53.432
640278 95.808 -54.718 53.404
640279 95.837 -54.732 53.373
640280 95.864 -54.750 53.339
640281 95.894 -54.764 53.303
640282 95.921 -54.776 53.264
640283 95.950 -54.787 53.225
640284 95.977 -54.798 53.189
640285 96.005 -54.809 53.149
640286 96.033 -54.821 53.108
640287 96.060 -54.832 53.066
640288 96.088 -54.843 53.023
640289 96.116 -54.854 52.978
640290 96.144 -54.865 52.931
640291 96.172 -54.875 52.882
640292 96.200 -54.885 52.831
640293 96.229 -54.896 52.779
640294 96.257 -54.906 52.728
640295 96.286 -54.916 52.678
640296 96.311 -54.925 52.631
640297 96.335 -54.933 52.585
640298 96.360 -54.941 52.541
640299 96.384 -54.949 52.498
640300 96.409 -54.957 52.455
640301 96.438 -54.964 52.412
640302 96.467 -54.971 52.366
640303 96.496 -54.978 52.312
640304 96.524 -54.985 52.250
640305 96.553 -54.992 52.183
640306 96.584 -55.000 52.112
640307 96.615 -55.009 52.039
640308 96.646 -55.018 51.965
640309 96.677 -55.027 51.893
640310 96.709 -55.035 51.821
640311 96.746 -55.047 51.744
640312 96.785 -55.059 51.665
640313 96.825 -55.072 51.585
640314 96.862 -55.091 51.505
640315 96.897 -55.115 51.427
640316 96.932 -55.139 51.353
640317 96.967 -55.163 51.279
640318 97.002 -55.188 51.207
640319 97.037 -55.212 51.137
640320 97.069 -55.232 51.069
640321 97.086 -55.232 51.003
640322 97.102 -55.231 50.940
640323 97.119 -55.230 50.877
640324 97.136 -55.229 50.815
640325 97.152 -55.229 50.755
640326 97.169 -55.228 50.697
640327 97.185 -55.227 50.642
640328 97.218 -55.226 50.586
640329 97.253 -55.225 50.532
640330 97.288 -55.223 50.479
640793 90.386 -51.786 56.995
640794 90.418 -51.794 57.002
640795 90.450 -51.802 57.009
640796 90.482 -51.810 57.017
640797 90.515 -51.818 57.025
640798 90.547 -51.826 57.032
640799 90.579 -51.834 57.040
640800 90.611 -51.842 57.047
640801 90.644 -51.850 57.054
640802 90.676 -51.858 57.061
640803 90.709 -51.866 57.066
640804 90.741 -51.874 57.070
640805 90.773 -51.882 57.075
640806 90.805 -51.890 57.081
640807 90.837 -51.898 57.088
640808 90.870 -51.906 57.096
640809 90.902 -51.914 57.101
640810 90.934 -51.922 57.109
640811 90.966 -51.930 57.119
640812 90.989 -51.942 57.132
640813 91.000 -51.958 57.144
640814 91.020 -51.968 57.158
640815 91.046 -51.978 57.173
640816 91.074 -52.001 57.189
640817 91.104 -52.027 57.211
640818 91.134 -52.054 57.233
640819 91.163 -52.080 57.252
640820 91.193 -52.107 57.267
640821 91.223 -52.133 57.275
640822 91.253 -52.159 57.280
640823 91.283 -52.186 57.279
640824 91.312 -52.212 57.279
640825 91.342 -52.239 57.278
640826 91.372 -52.265 57.277
640827 91.401 -52.292 57.280
640828 91.402 -52.322 57.282
640829 91.405 -52.351 57.274
640830 91.432 -52.373 57.257
640831 91.458 -52.396 57.244
640832 91.485 -52.418 57.235
640833 91.512 -52.440 57.222
640834 91.538 -52.462 57.201
640835 91.565 -52.485 57.172
640836 91.592 -52.507 57.146
640837 91.618 -52.529 57.126
640838 91.645 -52.552 57.110
640839 91.672 -52.574 57.098
640840 91.699 -52.597 57.087
640841 91.725 -52.619 57.075
640842 91.751 -52.641 57.056
640843 91.778 -52.663 57.027
640844 91.805 -52.686 56.995
640845 91.831 -52.708 56.961
640846 91.858 -52.731 56.923
640847 91.885 -52.753 56.882
640848 91.912 -52.775 56.844
640849 91.938 -52.797 56.808
640850 91.965 -52.819 56.772
640851 91.994 -52.839 56.735
640852 92.021 -52.860 56.698
640853 92.042 -52.876 56.662
640854 92.070 -52.897 56.621
640855 92.103 -52.921 56.577
640856 92.137 -52.946 56.528
640857 92.170 -52.970 56.472
640858 92.210 -52.985 56.411
640859 92.251 -52.998 56.351
640860 92.292 -53.012 56.294
640861 92.333 -53.026 56.235
640862 92.376 -53.035 56.177
640863 92.424 -53.042 56.121
640864 92.472 -53.048 56.065
640865 92.520 -53.054 56.011
640866 92.568 -53.060 55.958
640867 92.616 -53.066 55.906
640901 93.725 -53.536 54.862
640902 93.755 -53.550 54.861
640903 93.785 -53.565 54.862
640904 93.815 -53.579 54.864
640905 93.845 -53.594 54.867
640906 93.875 -53.609 54.871
640907 93.905 -53.623 54.876
640908 93.935 -53.640 54.882
640909 93.966 -53.659 54.889
640910 93.996 -53.678 54.897
640911 94.027 -53.697 54.905
640912 94.057 -53.715 54.916
640913 94.088 -53.734 54.927
640914 94.119 -53.753 54.937
640915 94.149 -53.772 54.946
640916 94.157 -53.784 54.952
640917 94.187 -53.806 54.958
640918 94.220 -53.828 54.965
640919 94.255 -53.850 54.970
640920 94.288 -53.871 54.978
640921 94.323 -53.892 54.984
640922 94.348 -53.907 54.989
640923 94.365 -53.916 54.995
640924 94.382 -53.923 54.999
640925 94.398 -53.932 55.000
640926 94.415 -53.939 55.001
640927 94.431 -53.947 55.001
640928 94.448 -53.956 55.003
640929 94.465 -53.963 55.008
640930 94.481 -53.972 55.014
640931 94.498 -53.979 55.013
640932 94.514 -53.988 55.004
640933 94.531 -53.995 54.988
640934 94.548 -54.004 54.969
640935 94.564 -54.011 54.951
640936 94.581 -54.019 54.934
640937 94.598 -54.027 54.918
640938 94.621 -54.040 54.904
640939 94.645 -54.052 54.885
640940 94.669 -54.064 54.870
640941 94.692 -54.077 54.856
640942 94.716 -54.089 54.843
640943 94.740 -54.101 54.829
640944 94.763 -54.113 54.815
640945 94.787 -54.126 54.785
640946 94.811 -54.138 54.748
640947 94.834 -54.151 54.710
640948 94.858 -54.164 54.673
640949 94.886 -54.178 54.637
640950 94.922 -54.196 54.598
640951 94.959 -54.216 54.550
640952 94.998 -54.242 54.499
640953 95.039 -54.269 54.447
640954 95.079 -54.295 54.395
640955 95.119 -54.322 54.346
640956 95.149 -54.339 54.302
640957 95.178 -54.354 54.257
640958 95.206 -54.368 54.215
640959 95.235 -54.383 54.173
640960 95.260 -54.399 54.129
640961 95.284 -54.414 54.084
640962 95.308 -54.430 54.036
640963 95.332 -54.446 53.984
640964 95.358 -54.464 53.933
640965 95.386 -54.483 53.882
640966 95.412 -54.501 53.834
640967 95.439 -54.519 53.791
640968 95.466 -54.537 53.750
640969 95.493 -54.555 53.712
640970 95.521 -54.572 53.676
640971 95.553 -54.586 53.643
640972 95.586 -54.599 53.614
640973 95.618 -54.613 53.587
640974 95.651 -54.627 53.560
640975 95.684 -54.640 53.533
640976 95.716 -54.654 53.506
640977 95.749 -54.668 53.480
640978 95.783 -54.678 53.453
640979 95.820 -54.684 53.426
640980 95.846 -54.703 53.396
640981 95.872 -54.721 53.362
640982 95.900 -54.737 53.327
640983 95.929 -54.751 53.291
640984 95.959 -54.766 53.253
640985 95.986 -54.777 53.215
640986 96.014 -54.788 53.175
640987 96.042 -54.800 53.132
640988 96.070 -54.811 53.089
640989 96.098 -54.822 53.045
640990 96.125 -54.833 52.999
640991 96.153 -54.844 52.954
640992 96.181 -54.854 52.906
640993 96.209 -54.865 52.857
640994 96.238 -54.875 52.807
640995 96.266 -54.885 52.757
640996 96.293 -54.894 52.712
640997 96.317 -54.903 52.666
640998 96.342 -54.911 52.619
640999 96.367 -54.919 52.576
641000 96.391 -54.927 52.533
641001 96.416 -54.934 52.489
641002 96.445 -54.941 52.445
641003 96.474 -54.949 52.400
641004 96.503 -54.956 52.348
641005 96.532 -54.963 52.287
641006 96.560 -54.970 52.221
641007 96.591 -54.978 52.153
641008 96.622 -54.986 52.084
641009 96.653 -54.995 52.015
641010 96.684 -55.003 51.952
641011 96.716 -55.013 51.884
641012 96.747 -55.022 51.809
641013 96.786 -55.034 51.730
641014 96.825 -55.046 51.648
641015 96.865 -55.059 51.564
641016 96.902 -55.078 51.482
641017 96.937 -55.102 51.405
641018 96.972 -55.126 51.330
641019 97.007 -55.151 51.256
641020 97.042 -55.175 51.184
641021 97.077 -55.199 51.115
641022 97.100 -55.207 51.048
641023 97.116 -55.206 50.983
641024 97.133 -55.205 50.918
641025 97.150 -55.204 50.855
641026 97.166 -55.204 50.793
641027 97.183 -55.203 50.734
641028 97.199 -55.201 50.676
641029 97.234 -55.201 50.620
641030 97.269 -55.200 50.565
641031 97.304 -55.198 50.510
641494 90.405 -51.753 56.985
641495 90.438 -51.760 56.991
641496 90.470 -51.769 56.997
641497 90.503 -51.776 57.004
641498 90.535 -51.785 57.011
641499 90.567 -51.792 57.018
641500 90.599 -51.801 57.026
641501 90.632 -51.809 57.033
641502 90.664 -51.816 57.040
641503 90.696 -51.825 57.047
641504 90.728 -51.832 57.052
641505 90.760 -51.841 57.057
641506 90.793 -51.848 57.063
641507 90.825 -51.857 57.069
641508 90.858 -51.865 57.076
641509 90.890 -51.873 57.085
641510 90.922 -51.881 57.092
641511 90.954 -51.888 57.101
641512 90.982 -51.898 57.115
641513 90.994 -51.914 57.129
641514 91.005 -51.930 57.139
641515 91.023 -51.942 57.148
641516 91.050 -51.947 57.160
641517 91.082 -51.972 57.176
641518 91.113 -51.998 57.196
641519 91.143 -52.024 57.216
641520 91.172 -52.051 57.233
641521 91.203 -52.077 57.248
641522 91.232 -52.104 57.254
641523 91.262 -52.130 57.257
641524 91.292 -52.157 57.258
641525 91.321 -52.183 57.259
641526 91.351 -52.210 57.259
641527 91.381 -52.236 57.260
641528 91.411 -52.262 57.262
641529 91.426 -52.291 57.264
641530 91.424 -52.321 57.258
641531 91.451 -52.343 57.244
641532 91.477 -52.365 57.231
641533 91.504 -52.388 57.223
641534 91.530 -52.410 57.212
641535 91.557 -52.432 57.193
641536 91.584 -52.455 57.165
641537 91.610 -52.477 57.139
641538 91.637 -52.499 57.117
641539 91.664 -52.521 57.100
641540 91.690 -52.544 57.084
641541 91.717 -52.566 57.070
641542 91.744 -52.589 57.056
641543 91.771 -52.611 57.038
641544 91.797 -52.634 57.013
641545 91.823 -52.656 56.983
641546 91.850 -52.678 56.951
641547 91.877 -52.700 56.915
641548 91.904 -52.723 56.876
641549 91.930 -52.745 56.840
641550 91.957 -52.767 56.806
641551 91.984 -52.790 56.770
641552 92.010 -52.812 56.734
641553 92.037 -52.833 56.699
641554 92.066 -52.854 56.661
641555 92.089 -52.871 56.622
641556 92.118 -52.893 56.579
641557 92.151 -52.917 56.530
641558 92.184 -52.941 56.475
641559 92.224 -52.956 56.415
641560 92.265 -52.970 56.357
641561 92.306 -52.983 56.298
641562 92.351 -52.991 56.239
641563 92.399 -52.998 56.181
641564 92.447 -53.004 56.124
641565 92.495 -53.010 56.068
641566 92.543 -53.016 56.012
641567 92.591 -53.022 55.959
641568 92.639 -53.028 55.907
641602 93.741 -53.500 54.858
641603 93.771 -53.515 54.856
641604 93.801 -53.529 54.857
641605 93.831 -53.543 54.859
641606 93.861 -53.559 54.862
641607 93.892 -53.577 54.866
641608 93.922 -53.596 54.871
641609 93.953 -53.615 54.876
641610 93.983 -53.634 54.883
641611 94.014 -53.653 54.891
641612 94.044 -53.672 54.901
641613 94.075 -53.690 54.912
641614 94.105 -53.709 54.922
641615 94.136 -53.728 54.932
641616 94.167 -53.747 54.940
641617 94.167 -53.756 54.945
641618 94.197 -53.778 54.950
641619 94.227 -53.800 54.958
641620 94.261 -53.822 54.962
641621 94.295 -53.843 54.969
641622 94.329 -53.864 54.975
641623 94.360 -53.884 54.981
641624 94.377 -53.891 54.988
641625 94.394 -53.900 54.993
641626 94.410 -53.907 54.995
641627 94.427 -53.916 54.997
641628 94.443 -53.923 54.997
641629 94.460 -53.932 54.998
641630 94.477 -53.939 55.002
641631 94.493 -53.947 55.005
641632 94.510 -53.956 55.003
641633 94.527 -53.963 54.995
641634 94.543 -53.972 54.984
641635 94.560 -53.979 54.971
641636 94.576 -53.988 54.957
641637 94.597 -53.997 54.942
641638 94.621 -54.010 54.928
641639 94.644 -54.022 54.914
641640 94.668 -54.035 54.896
641641 94.692 -54.048 54.881
641642 94.715 -54.060 54.870
641643 94.739 -54.072 54.860
641644 94.762 -54.084 54.849
641645 94.786 -54.097 54.836
641646 94.810 -54.109 54.805
641647 94.833 -54.121 54.768
641648 94.857 -54.134 54.730
641649 94.881 -54.146 54.692
641650 94.904 -54.158 54.654
641651 94.938 -54.176 54.613
641652 94.974 -54.195 54.563
641653 95.011 -54.214 54.511
641654 95.051 -54.241 54.458
641655 95.092 -54.267 54.407
641656 95.131 -54.294 54.359
641657 95.165 -54.314 54.316
641658 95.193 -54.328 54.271
641659 95.222 -54.342 54.228
641660 95.249 -54.358 54.183
641661 95.273 -54.373 54.137
641662 95.297 -54.389 54.089
641663 95.320 -54.405 54.039
641664 95.344 -54.421 53.987
641665 95.368 -54.436 53.936
641666 95.395 -54.455 53.888
641667 95.421 -54.473 53.842
641668 95.448 -54.491 53.801
641669 95.475 -54.509 53.764
641670 95.502 -54.527 53.729
641671 95.529 -54.546 53.695
641672 95.562 -54.559 53.664
641673 95.594 -54.572 53.635
641674 95.627 -54.586 53.608
641675 95.659 -54.599 53.582
641676 95.692 -54.613 53.556
641677 95.725 -54.627 53.529
641678 95.759 -54.637 53.502
641679 95.796 -54.643 53.475
641680 95.829 -54.654 53.449
641681 95.854 -54.672 53.420
641682 95.880 -54.690 53.388
641683 95.906 -54.709 53.355
641684 95.936 -54.723 53.321
641685 95.965 -54.738 53.283
641686 95.994 -54.753 53.242
641687 96.023 -54.767 53.200
641688 96.051 -54.778 53.156
641689 96.079 -54.790 53.111
641690 96.107 -54.800 53.066
641691 96.134 -54.812 53.020
641692 96.162 -54.823 52.975
641693 96.190 -54.834 52.929
641694 96.219 -54.844 52.881
641695 96.247 -54.854 52.834
641696 96.275 -54.864 52.787
641697 96.300 -54.872 52.743
641698 96.324 -54.881 52.699
641699 96.348 -54.888 52.653
641700 96.373 -54.897 52.608
641701 96.398 -54.905 52.565
641702 96.424 -54.913 52.521
641703 96.453 -54.919 52.476
641704 96.482 -54.926 52.429
641705 96.511 -54.934 52.377
641706 96.539 -54.941 52.320
641707 96.568 -54.947 52.259
641708 96.597 -54.955 52.196
641709 96.628 -54.963 52.130
641710 96.660 -54.972 52.061
641711 96.691 -54.981 52.000
641712 96.722 -54.990 51.938
641713 96.754 -54.998 51.868
641714 96.787 -55.008 51.792
641715 96.826 -55.020 51.710
641716 96.866 -55.033 51.626
641717 96.905 -55.045 51.542
641718 96.943 -55.064 51.459
641719 96.977 -55.088 51.381
641720 97.013 -55.113 51.306
641721 97.047 -55.137 51.234
641722 97.083 -55.162 51.163
641723 97.114 -55.182 51.094
641724 97.131 -55.181 51.028
641725 97.148 -55.180 50.962
641726 97.164 -55.179 50.897
641727 97.181 -55.178 50.834
641728 97.197 -55.177 50.771
641729 97.215 -55.176 50.712
641730 97.250 -55.176 50.655
641731 97.285 -55.175 50.598
641732 97.320 -55.173 50.542
641733 97.355 -55.173 50.487
642195 90.426 -51.719 56.974
642196 90.458 -51.727 56.979
642197 90.490 -51.735 56.985
642198 90.522 -51.743 56.992
642199 90.554 -51.751 56.998
642200 90.587 -51.759 57.005
642201 90.619 -51.767 57.012
642202 90.652 -51.776 57.018
642203 90.684 -51.783 57.026
642204 90.716 -51.791 57.032
642205 90.748 -51.799 57.039
642206 90.781 -51.807 57.044
642207 90.813 -51.815 57.050
642208 90.845 -51.823 57.057
642209 90.877 -51.832 57.064
642210 90.909 -51.839 57.073
642211 90.942 -51.848 57.082
642212 90.974 -51.855 57.092
642213 90.987 -51.871 57.106
642214 90.999 -51.886 57.118
642215 91.010 -51.902 57.128
642216 91.026 -51.915 57.137
642217 91.054 -51.920 57.147
642218 91.086 -51.945 57.161
642219 91.118 -51.970 57.178
642220 91.150 -51.996 57.196
642221 91.182 -52.022 57.212
642222 91.212 -52.048 57.225
642223 91.241 -52.075 57.230
642224 91.271 -52.101 57.234
642225 91.301 -52.127 57.237
642226 91.330 -52.154 57.239
642227 91.361 -52.180 57.239
642228 91.390 -52.207 57.241
642229 91.420 -52.233 57.244
642230 91.449 -52.260 57.245
642231 91.443 -52.291 57.239
642232 91.469 -52.313 57.227
642233 91.496 -52.335 57.214
642234 91.523 -52.358 57.205
642235 91.549 -52.380 57.195
642236 91.576 -52.402 57.177
642237 91.603 -52.424 57.154
642238 91.629 -52.447 57.129
642239 91.656 -52.469 57.106
642240 91.682 -52.492 57.087
642241 91.709 -52.514 57.068
642242 91.736 -52.537 57.052
642243 91.763 -52.559 57.036
642244 91.789 -52.581 57.017
642245 91.816 -52.603 56.995
642246 91.843 -52.626 56.969
642247 91.869 -52.648 56.939
642248 91.895 -52.670 56.904
642249 91.922 -52.693 56.869
642250 91.949 -52.715 56.834
642251 91.976 -52.738 56.801
642252 92.002 -52.760 56.767
642253 92.029 -52.782 56.732
642254 92.056 -52.804 56.697
642255 92.082 -52.827 56.660
642256 92.110 -52.847 56.621
642257 92.135 -52.866 56.577
642258 92.165 -52.889 56.529
642259 92.198 -52.913 56.475
642260 92.238 -52.927 56.417
642261 92.279 -52.941 56.360
642262 92.326 -52.948 56.301
642263 92.374 -52.954 56.242
642264 92.422 -52.960 56.184
642265 92.470 -52.966 56.126
642266 92.518 -52.972 56.069
642267 92.566 -52.978 56.014
642268 92.614 -52.984 55.960
642269 92.661 -52.990 55.908
642270 92.710 -52.996 55.858
642304 93.787 -53.479 54.853
642305 93.818 -53.496 54.853
642306 93.848 -53.515 54.855
642307 93.878 -53.534 54.857
642308 93.909 -53.552 54.861
642309 93.939 -53.571 54.866
642310 93.970 -53.590 54.871
642311 94.000 -53.609 54.878
642312 94.031 -53.628 54.886
642313 94.062 -53.646 54.896
642314 94.092 -53.665 54.905
642315 94.123 -53.684 54.916
642316 94.153 -53.703 54.925
642317 94.184 -53.721 54.933
642318 94.192 -53.733 54.938
642319 94.207 -53.750 54.943
642320 94.238 -53.772 54.950
642321 94.268 -53.794 54.955
642322 94.302 -53.815 54.961
642323 94.336 -53.836 54.967
642324 94.370 -53.858 54.974
642325 94.389 -53.868 54.982
642326 94.406 -53.875 54.988
642327 94.422 -53.884 54.992
642328 94.439 -53.891 54.995
642329 94.456 -53.900 54.996
642330 94.472 -53.907 54.998
642331 94.489 -53.916 55.001
642332 94.505 -53.923 55.004
642333 94.522 -53.932 55.004
642334 94.539 -53.939 55.000
642335 94.555 -53.947 54.991
642336 94.572 -53.956 54.981
642337 94.596 -53.969 54.969
642338 94.620 -53.981 54.957
642339 94.643 -53.993 54.944
642340 94.667 -54.005 54.928
642341 94.691 -54.018 54.908
642342 94.714 -54.030 54.894
642343 94.738 -54.042 54.885
642344 94.762 -54.055 54.877
642345 94.785 -54.067 54.868
642346 94.809 -54.079 54.852
642347 94.833 -54.091 54.822
642348 94.856 -54.104 54.787
642349 94.880 -54.117 54.751
642350 94.904 -54.129 54.714
642351 94.927 -54.142 54.674
642352 94.954 -54.155 54.629
642353 94.990 -54.173 54.577
642354 95.027 -54.192 54.524
642355 95.063 -54.213 54.471
642356 95.104 -54.239 54.420
642357 95.144 -54.266 54.372
642358 95.180 -54.288 54.330
642359 95.208 -54.302 54.285
642360 95.237 -54.317 54.239
642361 95.262 -54.333 54.191
642362 95.285 -54.348 54.143
642363 95.309 -54.364 54.093
642364 95.333 -54.380 54.042
642365 95.357 -54.396 53.990
642366 95.381 -54.412 53.940
642367 95.404 -54.427 53.894
642368 95.430 -54.446 53.853
642369 95.457 -54.464 53.816
642370 95.484 -54.482 53.782
642371 95.511 -54.500 53.749
642372 95.538 -54.518 53.717
642373 95.570 -54.532 53.686
642374 95.603 -54.546 53.659
642375 95.635 -54.559 53.633
642376 95.668 -54.572 53.607
642377 95.700 -54.586 53.581
642378 95.736 -54.596 53.554
642379 95.772 -54.602 53.527
642380 95.809 -54.610 53.501
642381 95.838 -54.624 53.474
642382 95.864 -54.643 53.447
642383 95.889 -54.661 53.418
642384 95.914 -54.679 53.387
642385 95.942 -54.696 53.353
642386 95.971 -54.710 53.313
642387 96.001 -54.725 53.270
642388 96.030 -54.740 53.225
642389 96.059 -54.755 53.179
642390 96.088 -54.768 53.134
642391 96.116 -54.779 53.088
642392 96.144 -54.790 53.041
642393 96.171 -54.802 52.997
642394 96.199 -54.812 52.951
642395 96.228 -54.823 52.906
642396 96.256 -54.833 52.861
642397 96.282 -54.842 52.816
642398 96.307 -54.850 52.773
642399 96.331 -54.859 52.730
642400 96.356 -54.866 52.684
642401 96.380 -54.875 52.639
642402 96.404 -54.882 52.594
642403 96.432 -54.890 52.550
642404 96.460 -54.897 52.504
642405 96.489 -54.904 52.455
642406 96.518 -54.911 52.403
642407 96.547 -54.918 52.349
642408 96.576 -54.925 52.292
642409 96.604 -54.932 52.233
642410 96.636 -54.941 52.170
642411 96.667 -54.950 52.101
642412 96.698 -54.958 52.039
642413 96.729 -54.967 51.978
642414 96.760 -54.975 51.911
642415 96.792 -54.985 51.838
642416 96.827 -54.995 51.759
642417 96.867 -55.007 51.679
642418 96.906 -55.020 51.598
642419 96.946 -55.032 51.515
642420 96.983 -55.051 51.437
642421 97.018 -55.075 51.362
642422 97.053 -55.100 51.287
642423 97.088 -55.124 51.214
642424 97.123 -55.148 51.144
642425 97.145 -55.156 51.076
642426 97.162 -55.155 51.008
642427 97.178 -55.154 50.941
642428 97.195 -55.153 50.876
642429 97.212 -55.152 50.812
642430 97.231 -55.151 50.751
642431 97.266 -55.151 50.692
642432 97.301 -55.149 50.633
642433 97.336 -55.148 50.575
642434 97.371 -55.148 50.518
642435 97.406 -55.146 50.463
642897 90.478 -51.694 56.967
642898 90.510 -51.702 56.973
642899 90.542 -51.710 56.979
642900 90.575 -51.718 56.985
642901 90.607 -51.725 56.991
642902 90.639 -51.734 56.998
642903 90.671 -51.742 57.005
642904 90.704 -51.750 57.011
642905 90.736 -51.758 57.018
642906 90.768 -51.766 57.024
642907 90.801 -51.774 57.030
642908 90.833 -51.782 57.037
642909 90.865 -51.790 57.044
642910 90.897 -51.798 57.052
642911 90.930 -51.806 57.061
642912 90.962 -51.814 57.071
642913 90.980 -51.827 57.081
642914 90.992 -51.843 57.093
642915 91.004 -51.859 57.106
642916 91.016 -51.875 57.116
642917 91.030 -51.888 57.126
642918 91.058 -51.894 57.135
642919 91.089 -51.917 57.147
642920 91.121 -51.943 57.163
642921 91.153 -51.969 57.178
642922 91.185 -51.995 57.192
642923 91.217 -52.020 57.205
642924 91.249 -52.046 57.209
642925 91.280 -52.072 57.213
642926 91.310 -52.098 57.217
642927 91.340 -52.125 57.220
642928 91.370 -52.152 57.221
642929 91.399 -52.178 57.223
642930 91.429 -52.205 57.225
642931 91.459 -52.230 57.227
642932 91.461 -52.261 57.221
642933 91.488 -52.283 57.210
642934 91.515 -52.305 57.197
642935 91.541 -52.327 57.187
642936 91.568 -52.350 57.177
642937 91.595 -52.372 57.161
642938 91.622 -52.395 57.142
642939 91.648 -52.417 57.120
642940 91.675 -52.440 57.098
642941 91.702 -52.462 57.078
642942 91.728 -52.484 57.057
642943 91.754 -52.506 57.038
642944 91.781 -52.529 57.020
642945 91.808 -52.551 56.999
642946 91.835 -52.573 56.977
642947 91.861 -52.596 56.953
642948 91.888 -52.618 56.926
642949 91.915 -52.641 56.893
642950 91.941 -52.662 56.860
642951 91.968 -52.685 56.828
642952 91.994 -52.707 56.796
642953 92.021 -52.730 56.763
642954 92.048 -52.752 56.729
642955 92.074 -52.775 56.694
642956 92.101 -52.797 56.657
642957 92.128 -52.819 56.618
642958 92.154 -52.841 56.575
642959 92.181 -52.862 56.527
642960 92.212 -52.885 56.474
642961 92.253 -52.898 56.417
642962 92.301 -52.904 56.360
642963 92.349 -52.910 56.302
642964 92.397 -52.916 56.243
642965 92.445 -52.923 56.186
642966 92.493 -52.929 56.128
642967 92.541 -52.934 56.070
642968 92.588 -52.940 56.015
642969 92.637 -52.946 55.961
642970 92.685 -52.952 55.908
642971 92.732 -52.958 55.858
643005 93.804 -53.452 54.850
643006 93.835 -53.471 54.849
643007 93.865 -53.490 54.851
643008 93.896 -53.509 54.854
643009 93.926 -53.527 54.857
643010 93.957 -53.546 54.862
643011 93.987 -53.565 54.867
643012 94.018 -53.584 54.874
643013 94.048 -53.603 54.882
643014 94.079 -53.621 54.891
643015 94.109 -53.640 54.900
643016 94.140 -53.659 54.910
643017 94.171 -53.677 54.919
643018 94.201 -53.696 54.925
643019 94.218 -53.710 54.932
643020 94.217 -53.721 54.938
643021 94.248 -53.743 54.943
643022 94.277 -53.765 54.949
643023 94.309 -53.787 54.955
643024 94.343 -53.808 54.961
643025 94.377 -53.830 54.968
643026 94.401 -53.844 54.977
643027 94.418 -53.852 54.984
643028 94.434 -53.860 54.990
643029 94.451 -53.868 54.994
643030 94.468 -53.875 54.997
643031 94.484 -53.884 55.000
643032 94.501 -53.891 55.003
643033 94.518 -53.900 55.007
643034 94.534 -53.907 55.009
643035 94.551 -53.916 55.007
643036 94.572 -53.926 55.001
643037 94.596 -53.939 54.993
643038 94.619 -53.951 54.985
643039 94.643 -53.963 54.974
643040 94.667 -53.976 54.959
643041 94.690 -53.988 54.941
643042 94.714 -54.001 54.920
643043 94.738 -54.013 54.904
643044 94.761 -54.026 54.894
643045 94.785 -54.038 54.888
643046 94.809 -54.050 54.881
643047 94.832 -54.062 54.864
643048 94.856 -54.075 54.837
643049 94.880 -54.087 54.806
643050 94.903 -54.099 54.774
643051 94.927 -54.112 54.738
643052 94.951 -54.124 54.693
643053 94.974 -54.136 54.646
643054 95.006 -54.153 54.591
643055 95.042 -54.172 54.536
643056 95.079 -54.191 54.483
643057 95.116 -54.211 54.433
643058 95.156 -54.238 54.388
643059 95.195 -54.262 54.345
643060 95.223 -54.277 54.298
643061 95.251 -54.292 54.249
643062 95.274 -54.308 54.199
643063 95.298 -54.323 54.148
643064 95.322 -54.339 54.097
643065 95.346 -54.355 54.046
643066 95.370 -54.371 53.995
643067 95.393 -54.387 53.946
643068 95.417 -54.403 53.903
643069 95.441 -54.419 53.866
643070 95.466 -54.436 53.833
643071 95.493 -54.454 53.802
643072 95.520 -54.472 53.770
643073 95.547 -54.490 53.738
643074 95.579 -54.505 53.709
643075 95.611 -54.518 53.683
643076 95.643 -54.532 53.659
643077 95.676 -54.546 53.634
643078 95.711 -54.555 53.608
643079 95.748 -54.562 53.582
643080 95.784 -54.569 53.556
643081 95.821 -54.576 53.530
643082 95.846 -54.594 53.505
643083 95.872 -54.612 53.480
643084 95.897 -54.630 53.452
643085 95.923 -54.649 53.420
643086 95.949 -54.667 53.384
643087 95.977 -54.683 53.344
643088 96.007 -54.697 53.300
643089 96.036 -54.712 53.253
643090 96.065 -54.728 53.206
643091 96.095 -54.742 53.159
643092 96.124 -54.757 53.111
643093 96.153 -54.769 53.065
643094 96.180 -54.781 53.019
643095 96.208 -54.792 52.974
643096 96.236 -54.802 52.931
643097 96.264 -54.812 52.888
643098 96.289 -54.820 52.845
643099 96.313 -54.828 52.802
643100 96.338 -54.836 52.758
643101 96.362 -54.844 52.713
643102 96.387 -54.853 52.668
643103 96.411 -54.860 52.622
643104 96.439 -54.868 52.576
643105 96.468 -54.875 52.529
643106 96.497 -54.881 52.477
643107 96.526 -54.889 52.426
643108 96.554 -54.896 52.373
643109 96.583 -54.903 52.320
643110 96.612 -54.909 52.264
643111 96.642 -54.918 52.204
643112 96.673 -54.926 52.136
643113 96.704 -54.935 52.072
643114 96.736 -54.944 52.010
643115 96.767 -54.953 51.945
643116 96.799 -54.962 51.876
643117 96.830 -54.970 51.803
643118 96.867 -54.981 51.726
643119 96.907 -54.994 51.649
643120 96.946 -55.007 51.569
643121 96.986 -55.019 51.495
643122 97.023 -55.038 51.421
643123 97.058 -55.063 51.345
643124 97.093 -55.087 51.270
643125 97.128 -55.111 51.197
643126 97.160 -55.130 51.126
643127 97.176 -55.129 51.057
643128 97.193 -55.129 50.989
643129 97.210 -55.128 50.922
643130 97.226 -55.127 50.856
643131 97.247 -55.126 50.793
643132 97.282 -55.126 50.732
643133 97.317 -55.124 50.672
643134 97.352 -55.123 50.612
643135 97.387 -55.122 50.554
643136 97.415 -55.121 50.497
643137 97.438 -55.120 50.443
643599 90.530 -51.669 56.961
643600 90.563 -51.676 56.966
643601 90.595 -51.685 56.972
643602 90.627 -51.692 56.978
643603 90.659 -51.700 56.984
643604 90.691 -51.708 56.991
643605 90.724 -51.716 56.997
643606 90.756 -51.725 57.004
643607 90.788 -51.732 57.010
643608 90.821 -51.741 57.017
643609 90.853 -51.748 57.023
643610 90.885 -51.757 57.030
643611 90.918 -51.764 57.039
643612 90.950 -51.772 57.049
643613 90.973 -51.784 57.058
643614 90.986 -51.800 57.068
643615 90.997 -51.815 57.079
643616 91.009 -51.831 57.091
643617 91.021 -51.847 57.103
643618 91.033 -51.862 57.114
643619 91.061 -51.867 57.124
643620 91.092 -51.891 57.136
643621 91.124 -51.916 57.150
643622 91.156 -51.942 57.164
643623 91.188 -51.967 57.177
643624 91.221 -51.993 57.189
643625 91.252 -52.019 57.193
643626 91.285 -52.045 57.198
643627 91.316 -52.070 57.202
643628 91.349 -52.096 57.205
643629 91.379 -52.123 57.207
643630 91.409 -52.148 57.208
643631 91.438 -52.175 57.209
643632 91.468 -52.202 57.208
643633 91.481 -52.230 57.204
643634 91.507 -52.253 57.195
643635 91.533 -52.275 57.182
643636 91.560 -52.298 57.170
643637 91.587 -52.320 57.159
643638 91.613 -52.343 57.145
643639 91.640 -52.365 57.129
643640 91.667 -52.387 57.111
643641 91.694 -52.409 57.091
643642 91.720 -52.432 57.070
643643 91.747 -52.454 57.048
643644 91.774 -52.476 57.026
643645 91.800 -52.499 57.007
643646 91.826 -52.521 56.986
643647 91.853 -52.543 56.964
643648 91.880 -52.565 56.940
643649 91.907 -52.588 56.914
643650 91.933 -52.610 56.883
643651 91.960 -52.633 56.852
643652 91.987 -52.655 56.821
643653 92.013 -52.678 56.790
643654 92.040 -52.700 56.758
643655 92.067 -52.722 56.724
643656 92.093 -52.744 56.689
643657 92.120 -52.766 56.652
643658 92.146 -52.789 56.614
643659 92.173 -52.811 56.572
643660 92.201 -52.832 56.525
643661 92.228 -52.853 56.472
643662 92.276 -52.860 56.416
643663 92.323 -52.866 56.359
643664 92.372 -52.872 56.301
643665 92.420 -52.879 56.244
643666 92.467 -52.885 56.186
643667 92.515 -52.891 56.129
643668 92.563 -52.897 56.071
643669 92.612 -52.903 56.016
643670 92.659 -52.908 55.961
643671 92.707 -52.914 55.909
643672 92.755 -52.920 55.858
643706 93.822 -53.427 54.847
643707 93.852 -53.446 54.847
643708 93.883 -53.465 54.848
643709 93.913 -53.483 54.850
643710 93.943 -53.502 54.853
643711 93.974 -53.521 54.858
643712 94.004 -53.540 54.863
643713 94.035 -53.559 54.870
643714 94.066 -53.577 54.877
643715 94.096 -53.596 54.885
643716 94.127 -53.615 54.894
643717 94.157 -53.633 54.904
643718 94.188 -53.652 54.912
643719 94.218 -53.671 54.919
643720 94.245 -53.688 54.926
643721 94.227 -53.692 54.932
643722 94.257 -53.714 54.938
643723 94.287 -53.736 54.945
643724 94.318 -53.759 54.951
643725 94.350 -53.780 54.957
643726 94.384 -53.802 54.964
643727 94.413 -53.819 54.973
643728 94.430 -53.828 54.981
643729 94.447 -53.836 54.987
643730 94.463 -53.844 54.993
643731 94.480 -53.852 54.998
643732 94.496 -53.860 55.002
643733 94.513 -53.868 55.007
643734 94.530 -53.875 55.015
643735 94.548 -53.885 55.018
643736 94.571 -53.897 55.015
643737 94.595 -53.910 55.011
643738 94.619 -53.922 55.007
643739 94.642 -53.934 55.002
643740 94.666 -53.946 54.991
643741 94.690 -53.959 54.974
643742 94.713 -53.971 54.954
643743 94.737 -53.983 54.930
643744 94.761 -53.996 54.911
643745 94.784 -54.008 54.901
643746 94.808 -54.020 54.896
643747 94.831 -54.032 54.889
643748 94.855 -54.045 54.874
643749 94.879 -54.057 54.852
643750 94.902 -54.070 54.825
643751 94.926 -54.082 54.794
643752 94.950 -54.095 54.757
643753 94.973 -54.107 54.711
643754 94.997 -54.119 54.660
643755 95.022 -54.132 54.603
643756 95.058 -54.151 54.547
643757 95.095 -54.170 54.495
643758 95.131 -54.189 54.448
643759 95.169 -54.209 54.403
643760 95.208 -54.236 54.357
643761 95.238 -54.251 54.308
643762 95.264 -54.267 54.256
643763 95.287 -54.283 54.203
643764 95.311 -54.298 54.151
643765 95.335 -54.314 54.100
643766 95.358 -54.330 54.050
643767 95.382 -54.346 54.000
643768 95.406 -54.362 53.956
643769 95.429 -54.378 53.918
643770 95.454 -54.394 53.883
643771 95.477 -54.410 53.852
643772 95.502 -54.427 53.822
643773 95.529 -54.445 53.791
643774 95.556 -54.463 53.760
643775 95.587 -54.477 53.733
643776 95.619 -54.491 53.708
643777 95.652 -54.505 53.685
643778 95.687 -54.514 53.661
643779 95.724 -54.521 53.636
643780 95.760 -54.528 53.611
643781 95.797 -54.535 53.586
643782 95.830 -54.546 53.561
643783 95.855 -54.565 53.537
643784 95.881 -54.583 53.512
643785 95.906 -54.601 53.485
643786 95.932 -54.619 53.452
643787 95.957 -54.637 53.416
643788 95.983 -54.655 53.375
643789 96.013 -54.670 53.330
643790 96.042 -54.684 53.282
643791 96.072 -54.699 53.233
643792 96.101 -54.715 53.185
643793 96.130 -54.729 53.136
643794 96.160 -54.744 53.089
643795 96.189 -54.759 53.044
643796 96.218 -54.771 53.000
643797 96.245 -54.781 52.957
643798 96.271 -54.790 52.915
643799 96.296 -54.798 52.873
643800 96.320 -54.806 52.829
643801 96.344 -54.814 52.784
643802 96.369 -54.822 52.738
643803 96.393 -54.830 52.691
643804 96.418 -54.838 52.643
643805 96.447 -54.845 52.595
643806 96.475 -54.853 52.547
643807 96.504 -54.859 52.493
643808 96.533 -54.866 52.439
643809 96.562 -54.873 52.387
643810 96.591 -54.881 52.336
643811 96.619 -54.887 52.281
643812 96.649 -54.895 52.222
643813 96.680 -54.903 52.158
643814 96.712 -54.913 52.094
643815 96.743 -54.921 52.031
643816 96.774 -54.930 51.969
643817 96.805 -54.938 51.906
643818 96.836 -54.947 51.840
643819 96.868 -54.956 51.766
643820 96.907 -54.969 51.692
643821 96.947 -54.981 51.621
643822 96.986 -54.994 51.554
643823 97.026 -55.006 51.484
643824 97.064 -55.025 51.408
643825 97.098 -55.049 51.329
643826 97.134 -55.073 51.253
643827 97.168 -55.097 51.179
643828 97.190 -55.104 51.109
643829 97.207 -55.104 51.039
643830 97.224 -55.103 50.970
643831 97.240 -55.102 50.903
643832 97.263 -55.101 50.838
643833 97.298 -55.100 50.774
643834 97.333 -55.099 50.713
643835 97.368 -55.098 50.651
643836 97.400 -55.097 50.592
643837 97.423 -55.096 50.533
643838 97.446 -55.095 50.477
644302 90.615 -51.651 56.959
644303 90.647 -51.659 56.965
644304 90.679 -51.667 56.970
644305 90.712 -51.675 56.976
644306 90.744 -51.683 56.983
644307 90.776 -51.691 56.990
644308 90.808 -51.699 56.997
644309 90.840 -51.707 57.003
644310 90.873 -51.715 57.010
644311 90.905 -51.723 57.017
644312 90.938 -51.731 57.026
644313 90.967 -51.740 57.036
644314 90.978 -51.756 57.045
644315 90.991 -51.772 57.055
644316 91.002 -51.788 57.066
644317 91.014 -51.804 57.077
644318 91.026 -51.819 57.090
644319 91.037 -51.835 57.102
644320 91.064 -51.841 57.115
644321 91.096 -51.863 57.128
644322 91.128 -51.889 57.140
644323 91.160 -51.915 57.152
644324 91.191 -51.940 57.164
644325 91.224 -51.966 57.176
644326 91.256 -51.992 57.182
644327 91.288 -52.017 57.186
644328 91.320 -52.043 57.189
644329 91.352 -52.069 57.192
644330 91.384 -52.095 57.195
644331 91.416 -52.120 57.195
644332 91.448 -52.146 57.194
644333 91.477 -52.173 57.193
644334 91.502 -52.200 57.189
644335 91.526 -52.223 57.180
644336 91.553 -52.246 57.168
644337 91.579 -52.268 57.156
644338 91.606 -52.290 57.143
644339 91.632 -52.312 57.129
644340 91.659 -52.335 57.115
644341 91.685 -52.357 57.100
644342 91.712 -52.379 57.083
644343 91.739 -52.402 57.063
644344 91.766 -52.424 57.040
644345 91.792 -52.446 57.017
644346 91.819 -52.468 56.997
644347 91.846 -52.490 56.975
644348 91.873 -52.510 56.953
644349 91.900 -52.531 56.929
644350 91.928 -52.552 56.903
644351 91.955 -52.572 56.873
644352 91.983 -52.593 56.843
644353 92.010 -52.613 56.813
644354 92.037 -52.634 56.783
644355 92.065 -52.655 56.751
644356 92.092 -52.675 56.718
644357 92.119 -52.696 56.683
644358 92.146 -52.716 56.646
644359 92.174 -52.737 56.608
644360 92.201 -52.758 56.566
644361 92.228 -52.778 56.520
644362 92.256 -52.799 56.468
644363 92.298 -52.822 56.413
644364 92.347 -52.828 56.357
644365 92.394 -52.835 56.300
644366 92.442 -52.841 56.243
644367 92.490 -52.847 56.186
644368 92.538 -52.853 56.129
644369 92.586 -52.859 56.072
644370 92.634 -52.865 56.017
644371 92.682 -52.871 55.962
644372 92.730 -52.877 55.909
644373 92.778 -52.882 55.859
644374 92.826 -52.888 55.809
644408 93.869 -53.421 54.845
644409 93.900 -53.440 54.845
644410 93.930 -53.458 54.847
644411 93.961 -53.477 54.851
644412 93.991 -53.496 54.856
644413 94.022 -53.515 54.860
644414 94.052 -53.533 54.866
644415 94.083 -53.552 54.874
644416 94.113 -53.571 54.881
644417 94.144 -53.590 54.890
644418 94.175 -53.608 54.898
644419 94.205 -53.627 54.906
644420 94.236 -53.646 54.914
644421 94.266 -53.664 54.922
644422 94.247 -53.667 54.928
644423 94.267 -53.686 54.935
644424 94.297 -53.708 54.941
644425 94.328 -53.730 54.947
644426 94.358 -53.752 54.954
644427 94.390 -53.774 54.961
644428 94.424 -53.795 54.969
644429 94.442 -53.803 54.977
644430 94.459 -53.812 54.985
644431 94.475 -53.819 54.992
644432 94.492 -53.828 54.999
644433 94.508 -53.836 55.004
644434 94.525 -53.844 55.009
644435 94.547 -53.855 55.015
644436 94.570 -53.867 55.019
644437 94.594 -53.880 55.019
644438 94.618 -53.892 55.019
644439 94.641 -53.904 55.017
644440 94.665 -53.916 55.011
644441 94.689 -53.929 55.000
644442 94.712 -53.941 54.984
644443 94.736 -53.954 54.961
644444 94.760 -53.966 54.936
644445 94.783 -53.979 54.915
644446 94.807 -53.991 54.905
644447 94.831 -54.003 54.899
644448 94.854 -54.016 54.892
644449 94.878 -54.028 54.881
644450 94.902 -54.040 54.867
644451 94.925 -54.053 54.845
644452 94.949 -54.065 54.813
644453 94.973 -54.077 54.772
644454 94.996 -54.089 54.726
644455 95.020 -54.102 54.671
644456 95.044 -54.114 54.614
644457 95.074 -54.130 54.559
644458 95.111 -54.149 54.508
644459 95.147 -54.168 54.462
644460 95.184 -54.187 54.415
644461 95.221 -54.207 54.368
644462 95.253 -54.226 54.315
644463 95.276 -54.242 54.262
644464 95.300 -54.258 54.208
644465 95.324 -54.273 54.154
644466 95.347 -54.289 54.104
644467 95.371 -54.305 54.056
644468 95.395 -54.321 54.010
644469 95.419 -54.337 53.969
644470 95.443 -54.353 53.933
644471 95.466 -54.369 53.901
644472 95.490 -54.385 53.871
644473 95.514 -54.401 53.842
644474 95.538 -54.417 53.812
644475 95.565 -54.435 53.785
644476 95.595 -54.451 53.760
644477 95.627 -54.464 53.736
644478 95.664 -54.473 53.713
644479 95.700 -54.480 53.690
644480 95.737 -54.487 53.665
644481 95.773 -54.494 53.643
644482 95.810 -54.502 53.618
644483 95.838 -54.516 53.593
644484 95.864 -54.534 53.569
644485 95.889 -54.552 53.544
644486 95.915 -54.571 53.516
644487 95.940 -54.589 53.483
644488 95.966 -54.607 53.446
644489 95.991 -54.625 53.404
644490 96.019 -54.642 53.359
644491 96.048 -54.657 53.311
644492 96.078 -54.671 53.261
644493 96.107 -54.687 53.212
644494 96.136 -54.702 53.162
644495 96.166 -54.716 53.114
644496 96.195 -54.731 53.070
644497 96.225 -54.746 53.027
644498 96.253 -54.759 52.984
644499 96.278 -54.768 52.941
644500 96.302 -54.776 52.898
644501 96.327 -54.784 52.855
644502 96.351 -54.792 52.809
644503 96.376 -54.800 52.762
644504 96.400 -54.808 52.714
644505 96.425 -54.816 52.664
644506 96.454 -54.823 52.613
644507 96.483 -54.830 52.564
644508 96.512 -54.837 52.509
644509 96.540 -54.844 52.455
644510 96.570 -54.851 52.401
644511 96.598 -54.858 52.348
644512 96.627 -54.865 52.294
644513 96.656 -54.872 52.237
644514 96.687 -54.881 52.176
644515 96.718 -54.890 52.115
644516 96.749 -54.898 52.054
644517 96.781 -54.907 51.994
644518 96.812 -54.916 51.933
644519 96.843 -54.925 51.871
644520 96.875 -54.933 51.807
644521 96.908 -54.943 51.740
644522 96.948 -54.956 51.676
644523 96.987 -54.968 51.615
644524 97.027 -54.981 51.549
644525 97.067 -54.993 51.476
644526 97.104 -55.012 51.399
644527 97.139 -55.036 51.318
644528 97.174 -55.060 51.239
644529 97.205 -55.079 51.165
644530 97.222 -55.079 51.093
644531 97.238 -55.078 51.022
644532 97.255 -55.077 50.953
644533 97.279 -55.076 50.886
644534 97.314 -55.075 50.821
644535 97.349 -55.074 50.757
644536 97.384 -55.073 50.693
644537 97.408 -55.072 50.631
644538 97.432 -55.071 50.572
644539 97.454 -55.070 50.514
644540 97.477 -55.069 50.458
645004 90.667 -51.625 56.951
645005 90.699 -51.634 56.956
645006 90.732 -51.641 56.962
645007 90.764 -51.650 56.969
645008 90.796 -51.658 56.977
645009 90.828 -51.666 56.983
645010 90.861 -51.674 56.989
645011 90.893 -51.682 56.996
645012 90.925 -51.690 57.004
645013 90.957 -51.697 57.014
645014 90.972 -51.713 57.024
645015 90.984 -51.729 57.033
645016 90.996 -51.744 57.043
645017 91.007 -51.760 57.054
645018 91.019 -51.776 57.066
645019 91.031 -51.791 57.079
645020 91.042 -51.807 57.094
645021 91.068 -51.814 57.111
645022 91.099 -51.836 57.126
645023 91.131 -51.862 57.138
645024 91.163 -51.888 57.148
645025 91.195 -51.913 57.158
645026 91.227 -51.939 57.168
645027 91.259 -51.964 57.173
645028 91.291 -51.990 57.177
645029 91.323 -52.016 57.180
645030 91.355 -52.042 57.182
645031 91.387 -52.067 57.184
645032 91.419 -52.093 57.183
645033 91.451 -52.119 57.181
645034 91.484 -52.145 57.178
645035 91.515 -52.170 57.173
645036 91.546 -52.189 57.166
645037 91.573 -52.210 57.155
645038 91.601 -52.230 57.143
645039 91.628 -52.251 57.129
645040 91.655 -52.271 57.115
645041 91.682 -52.292 57.101
645042 91.710 -52.313 57.088
645043 91.737 -52.333 57.073
645044 91.765 -52.354 57.058
645045 91.792 -52.374 57.034
645046 91.819 -52.395 57.010
645047 91.847 -52.416 56.988
645048 91.874 -52.437 56.966
645049 91.901 -52.457 56.944
645050 91.928 -52.477 56.921
645051 91.956 -52.498 56.892
645052 91.983 -52.519 56.863
645053 92.010 -52.540 56.834
645054 92.037 -52.560 56.804
645055 92.065 -52.581 56.775
645056 92.092 -52.601 56.745
645057 92.120 -52.622 56.712
645058 92.147 -52.643 56.677
645059 92.174 -52.663 56.640
645060 92.202 -52.684 56.601
645061 92.229 -52.704 56.560
645062 92.256 -52.725 56.514
645063 92.284 -52.746 56.464
645064 92.324 -52.775 56.411
645065 92.369 -52.791 56.356
645066 92.417 -52.797 56.300
645067 92.465 -52.803 56.244
645068 92.513 -52.809 56.187
645069 92.561 -52.815 56.129
645070 92.609 -52.821 56.073
645071 92.657 -52.827 56.018
645072 92.705 -52.833 55.963
645073 92.753 -52.839 55.910
645074 92.801 -52.845 55.859
645075 92.849 -52.851 55.810
645109 93.887 -53.396 54.844
645110 93.917 -53.415 54.844
645111 93.947 -53.433 54.846
645112 93.978 -53.452 54.849
645113 94.008 -53.471 54.854
645114 94.039 -53.490 54.858
645115 94.070 -53.508 54.864
645116 94.100 -53.527 54.872
645117 94.131 -53.546 54.879
645118 94.161 -53.564 54.886
645119 94.192 -53.583 54.895
645120 94.222 -53.602 54.903
645121 94.253 -53.620 54.911
645122 94.283 -53.639 54.918
645123 94.273 -53.645 54.926
645124 94.277 -53.657 54.932
645125 94.308 -53.679 54.938
645126 94.338 -53.701 54.944
645127 94.367 -53.723 54.951
645128 94.398 -53.746 54.958
645129 94.431 -53.767 54.966
645130 94.454 -53.780 54.974
645131 94.471 -53.787 54.982
645132 94.487 -53.796 54.991
645133 94.504 -53.803 54.998
645134 94.523 -53.813 55.003
645135 94.546 -53.825 55.008
645136 94.570 -53.838 55.012
645137 94.594 -53.850 55.017
645138 94.617 -53.863 55.021
645139 94.641 -53.875 55.023
645140 94.665 -53.887 55.021
645141 94.688 -53.900 55.015
645142 94.712 -53.912 55.005
645143 94.736 -53.924 54.988
645144 94.759 -53.937 54.964
645145 94.783 -53.949 54.937
645146 94.807 -53.961 54.916
645147 94.830 -53.974 54.905
645148 94.854 -53.986 54.899
645149 94.878 -53.998 54.897
645150 94.901 -54.010 54.893
645151 94.925 -54.023 54.877
645152 94.949 -54.035 54.855
645153 94.972 -54.048 54.824
645154 94.996 -54.060 54.782
645155 95.020 -54.073 54.734
645156 95.043 -54.085 54.679
645157 95.067 -54.097 54.623
645158 95.091 -54.110 54.570
645159 95.127 -54.128 54.521
645160 95.163 -54.147 54.474
645161 95.199 -54.166 54.425
645162 95.236 -54.185 54.378
645163 95.266 -54.202 54.321
645164 95.289 -54.218 54.266
645165 95.313 -54.233 54.211
645166 95.337 -54.249 54.158
645167 95.360 -54.265 54.109
645168 95.384 -54.280 54.064
645169 95.408 -54.296 54.021
645170 95.431 -54.312 53.983
645171 95.455 -54.328 53.949
645172 95.479 -54.344 53.919
645173 95.502 -54.360 53.890
645174 95.526 -54.376 53.862
645175 95.550 -54.392 53.834
645176 95.574 -54.408 53.810
645177 95.603 -54.424 53.786
645178 95.639 -54.431 53.763
645179 95.676 -54.439 53.741
645180 95.712 -54.446 53.720
645181 95.749 -54.453 53.696
645182 95.785 -54.461 53.673
645183 95.822 -54.468 53.648
645184 95.847 -54.486 53.624
645185 95.873 -54.505 53.600
645186 95.898 -54.523 53.574
645187 95.923 -54.541 53.545
645188 95.949 -54.559 53.511
645189 95.974 -54.577 53.474
645190 96.000 -54.596 53.431
645191 96.025 -54.614 53.386
645192 96.054 -54.629 53.339
645193 96.084 -54.644 53.290
645194 96.113 -54.659 53.240
645195 96.143 -54.674 53.189
645196 96.172 -54.689 53.141
645197 96.202 -54.703 53.096
645198 96.231 -54.718 53.051
645199 96.259 -54.732 53.008
645200 96.285 -54.745 52.966
645201 96.309 -54.753 52.923
645202 96.333 -54.762 52.880
645203 96.358 -54.770 52.835
645204 96.382 -54.778 52.786
645205 96.407 -54.786 52.735
645206 96.433 -54.793 52.683
645207 96.462 -54.800 52.630
645208 96.490 -54.808 52.579
645209 96.519 -54.815 52.525
645210 96.548 -54.822 52.472
645211 96.577 -54.828 52.417
645212 96.606 -54.836 52.362
645213 96.635 -54.843 52.309
645214 96.663 -54.850 52.256
645215 96.694 -54.858 52.200
645216 96.725 -54.867 52.141
645217 96.756 -54.875 52.084
645218 96.788 -54.884 52.025
645219 96.819 -54.893 51.967
645220 96.850 -54.902 51.908
645221 96.881 -54.910 51.850
645222 96.912 -54.919 51.790
645223 96.948 -54.930 51.730
645224 96.988 -54.942 51.671
645225 97.028 -54.955 51.608
645226 97.067 -54.967 51.538
645227 97.107 -54.980 51.460
645228 97.144 -54.999 51.381
645229 97.179 -55.023 51.302
645230 97.214 -55.047 51.225
645231 97.236 -55.054 51.149
645232 97.252 -55.053 51.075
645233 97.269 -55.052 51.004
645234 97.295 -55.050 50.935
645235 97.330 -55.050 50.868
645236 97.365 -55.049 50.802
645237 97.394 -55.048 50.737
645238 97.417 -55.047 50.673
645239 97.440 -55.046 50.611
645240 97.463 -55.045 50.552
645241 97.486 -55.044 50.495
645242 97.509 -55.044 50.438
645706 90.719 -51.600 56.943
645707 90.751 -51.608 56.949
645708 90.784 -51.616 56.956
645709 90.816 -51.624 56.963
645710 90.849 -51.632 56.970
645711 90.881 -51.641 56.977
645712 90.913 -51.648 56.984
645713 90.945 -51.656 56.992
645714 90.965 -51.669 57.001
645715 90.977 -51.685 57.012
645716 90.989 -51.700 57.022
645717 91.001 -51.716 57.031
645718 91.012 -51.732 57.042
645719 91.024 -51.748 57.056
645720 91.036 -51.763 57.070
645721 91.047 -51.779 57.088
645722 91.071 -51.788 57.109
645723 91.102 -51.809 57.127
645724 91.134 -51.835 57.140
645725 91.166 -51.860 57.147
645726 91.199 -51.886 57.156
645727 91.230 -51.912 57.163
645728 91.262 -51.938 57.168
645729 91.294 -51.964 57.172
645730 91.326 -51.989 57.175
645731 91.359 -52.014 57.177
645732 91.390 -52.040 57.177
645733 91.423 -52.066 57.175
645734 91.455 -52.092 57.169
645735 91.487 -52.117 57.163
645736 91.534 -52.130 57.157
645737 91.567 -52.145 57.150
645738 91.598 -52.161 57.142
645739 91.629 -52.177 57.130
645740 91.656 -52.198 57.115
645741 91.683 -52.218 57.102
645742 91.711 -52.239 57.089
645743 91.738 -52.259 57.075
645744 91.765 -52.280 57.061
645745 91.792 -52.301 57.049
645746 91.820 -52.321 57.026
645747 91.847 -52.342 57.003
645748 91.874 -52.362 56.980
645749 91.901 -52.384 56.958
645750 91.929 -52.404 56.935
645751 91.956 -52.424 56.913
645752 91.984 -52.445 56.885
645753 92.011 -52.465 56.856
645754 92.038 -52.486 56.827
645755 92.066 -52.507 56.798
645756 92.093 -52.528 56.769
645757 92.120 -52.548 56.740
645758 92.147 -52.568 56.705
645759 92.175 -52.589 56.670
645760 92.202 -52.610 56.632
645761 92.229 -52.631 56.594
645762 92.257 -52.651 56.553
645763 92.284 -52.672 56.508
645764 92.311 -52.693 56.459
645765 92.352 -52.722 56.408
645766 92.393 -52.750 56.354
645767 92.440 -52.759 56.300
645768 92.488 -52.765 56.244
645769 92.536 -52.771 56.187
645770 92.584 -52.777 56.131
645771 92.632 -52.783 56.074
645772 92.680 -52.789 56.019
645773 92.728 -52.795 55.965
645774 92.776 -52.801 55.912
645775 92.824 -52.807 55.861
645776 92.871 -52.813 55.811
645777 92.920 -52.819 55.763
645811 93.934 -53.389 54.843
645812 93.965 -53.408 54.845
645813 93.995 -53.427 54.848
645814 94.026 -53.446 54.852
645815 94.056 -53.464 54.856
645816 94.087 -53.483 54.862
645817 94.118 -53.502 54.869
645818 94.148 -53.520 54.876
645819 94.179 -53.539 54.884
645820 94.209 -53.558 54.892
645821 94.240 -53.576 54.900
645822 94.270 -53.595 54.908
645823 94.300 -53.614 54.915
645824 94.299 -53.622 54.922
645825 94.287 -53.628 54.929
645826 94.318 -53.650 54.936
645827 94.347 -53.673 54.942
645828 94.378 -53.695 54.948
645829 94.408 -53.717 54.954
645830 94.438 -53.739 54.961
645831 94.466 -53.756 54.969
645832 94.483 -53.764 54.978
645833 94.499 -53.772 54.986
645834 94.522 -53.784 54.992
645835 94.546 -53.796 54.998
645836 94.569 -53.808 55.001
645837 94.593 -53.821 55.006
645838 94.617 -53.833 55.012
645839 94.640 -53.845 55.019
645840 94.664 -53.858 55.022
645841 94.688 -53.870 55.019
645842 94.711 -53.882 55.013
645843 94.735 -53.894 55.006
645844 94.759 -53.907 54.987
645845 94.782 -53.919 54.963
645846 94.806 -53.932 54.936
645847 94.830 -53.944 54.915
645848 94.853 -53.957 54.905
645849 94.877 -53.969 54.900
645850 94.901 -53.981 54.901
645851 94.924 -53.994 54.900
645852 94.948 -54.006 54.883
645853 94.971 -54.018 54.858
645854 94.995 -54.031 54.825
645855 95.019 -54.043 54.782
645856 95.042 -54.055 54.734
645857 95.066 -54.067 54.682
645858 95.090 -54.080 54.629
645859 95.113 -54.092 54.577
645860 95.142 -54.107 54.528
645861 95.179 -54.126 54.478
645862 95.215 -54.145 54.427
645863 95.252 -54.164 54.377
645864 95.280 -54.181 54.321
645865 95.304 -54.197 54.266
645866 95.327 -54.212 54.213
645867 95.351 -54.228 54.164
645868 95.375 -54.243 54.119
645869 95.398 -54.259 54.078
645870 95.422 -54.275 54.036
645871 95.446 -54.290 54.000
645872 95.469 -54.306 53.967
645873 95.493 -54.322 53.937
645874 95.517 -54.337 53.908
645875 95.540 -54.353 53.882
645876 95.564 -54.369 53.857
645877 95.588 -54.381 53.834
645878 95.615 -54.390 53.811
645879 95.652 -54.398 53.790
645880 95.688 -54.405 53.771
645881 95.725 -54.412 53.751
645882 95.762 -54.420 53.728
645883 95.798 -54.427 53.704
645884 95.830 -54.439 53.678
645885 95.855 -54.457 53.654
645886 95.881 -54.475 53.630
645887 95.907 -54.493 53.604
645888 95.932 -54.511 53.574
645889 95.958 -54.530 53.539
645890 95.983 -54.548 53.502
645891 96.009 -54.566 53.460
645892 96.034 -54.584 53.415
645893 96.060 -54.601 53.368
645894 96.090 -54.616 53.318
645895 96.119 -54.631 53.267
645896 96.149 -54.646 53.215
645897 96.178 -54.661 53.167
645898 96.208 -54.675 53.120
645899 96.237 -54.690 53.075
645900 96.264 -54.704 53.031
645901 96.290 -54.717 52.988
645902 96.315 -54.730 52.945
645903 96.340 -54.740 52.901
645904 96.365 -54.747 52.855
645905 96.389 -54.756 52.807
645906 96.413 -54.764 52.756
645907 96.440 -54.772 52.705
645908 96.469 -54.778 52.653
645909 96.498 -54.785 52.600
645910 96.527 -54.792 52.549
645911 96.555 -54.800 52.496
645912 96.584 -54.806 52.441
645913 96.613 -54.813 52.387
645914 96.642 -54.821 52.337
645915 96.671 -54.828 52.290
645916 96.700 -54.835 52.237
645917 96.732 -54.844 52.180
645918 96.763 -54.853 52.124
645919 96.794 -54.862 52.068
645920 96.826 -54.870 52.012
645921 96.857 -54.879 51.955
645922 96.888 -54.887 51.899
645923 96.920 -54.897 51.842
645924 96.951 -54.905 51.782
645925 96.989 -54.917 51.723
645926 97.028 -54.929 51.661
645927 97.068 -54.942 51.593
645928 97.107 -54.954 51.516
645929 97.147 -54.967 51.441
645930 97.184 -54.985 51.365
645931 97.220 -55.010 51.285
645932 97.250 -55.028 51.206
645933 97.267 -55.028 51.129
645934 97.284 -55.027 51.056
645935 97.311 -55.025 50.985
645936 97.346 -55.025 50.916
645937 97.379 -55.024 50.848
645938 97.402 -55.023 50.781
645939 97.425 -55.022 50.716
645940 97.448 -55.021 50.652
645941 97.471 -55.020 50.591
645942 97.495 -55.019 50.532
645943 97.518 -55.019 50.474
645944 97.541 -55.018 50.418
646409 90.804 -51.583 56.943
646410 90.836 -51.591 56.950
646411 90.868 -51.599 56.957
646412 90.900 -51.607 56.964
646413 90.933 -51.615 56.972
646414 90.959 -51.625 56.980
646415 90.970 -51.641 56.989
646416 90.982 -51.657 57.000
646417 90.994 -51.672 57.011
646418 91.006 -51.688 57.021
646419 91.017 -51.704 57.033
646420 91.029 -51.720 57.047
646421 91.041 -51.736 57.062
646422 91.052 -51.752 57.083
646423 91.075 -51.760 57.106
646424 91.105 -51.782 57.128
646425 91.138 -51.807 57.145
646426 91.169 -51.833 57.150
646427 91.202 -51.859 57.157
646428 91.234 -51.885 57.162
646429 91.265 -51.910 57.165
646430 91.298 -51.936 57.168
646431 91.330 -51.962 57.171
646432 91.362 -51.988 57.173
646433 91.394 -52.013 57.172
646434 91.426 -52.039 57.168
646435 91.459 -52.064 57.160
646436 91.509 -52.075 57.151
646437 91.553 -52.089 57.143
646438 91.586 -52.105 57.137
646439 91.617 -52.120 57.130
646440 91.648 -52.136 57.119
646441 91.678 -52.152 57.105
646442 91.710 -52.167 57.091
646443 91.738 -52.186 57.077
646444 91.766 -52.206 57.064
646445 91.793 -52.227 57.049
646446 91.820 -52.247 57.037
646447 91.848 -52.268 57.016
646448 91.875 -52.289 56.995
646449 91.902 -52.309 56.973
646450 91.930 -52.330 56.950
646451 91.957 -52.350 56.927
646452 91.984 -52.371 56.906
646453 92.011 -52.392 56.879
646454 92.039 -52.412 56.851
646455 92.066 -52.433 56.823
646456 92.093 -52.453 56.795
646457 92.121 -52.474 56.765
646458 92.148 -52.495 56.735
646459 92.175 -52.515 56.701
646460 92.203 -52.536 56.665
646461 92.230 -52.556 56.627
646462 92.257 -52.578 56.588
646463 92.285 -52.598 56.546
646464 92.312 -52.619 56.503
646465 92.340 -52.639 56.456
646466 92.380 -52.668 56.406
646467 92.421 -52.697 56.353
646468 92.463 -52.721 56.299
646469 92.511 -52.727 56.244
646470 92.559 -52.733 56.188
646471 92.606 -52.739 56.133
646472 92.655 -52.745 56.077
646473 92.703 -52.751 56.021
646474 92.751 -52.757 55.967
646475 92.798 -52.763 55.915
646476 92.846 -52.769 55.864
646477 92.895 -52.775 55.814
646478 92.942 -52.782 55.766
646479 92.990 -52.788 55.719
646480 93.010 -52.809 55.672
646512 93.952 -53.364 54.844
646513 93.982 -53.383 54.845
646514 94.012 -53.402 54.847
646515 94.043 -53.420 54.850
646516 94.074 -53.439 54.854
646517 94.104 -53.458 54.860
646518 94.135 -53.476 54.867
646519 94.165 -53.495 54.875
646520 94.196 -53.514 54.882
646521 94.226 -53.532 54.890
646522 94.257 -53.551 54.898
646523 94.287 -53.570 54.906
646524 94.318 -53.588 54.912
646525 94.325 -53.599 54.919
646526 94.301 -53.602 54.927
646527 94.328 -53.622 54.934
646528 94.357 -53.644 54.940
646529 94.388 -53.666 54.945
646530 94.418 -53.688 54.951
646531 94.448 -53.710 54.957
646532 94.478 -53.732 54.964
646533 94.497 -53.742 54.972
646534 94.521 -53.754 54.979
646535 94.545 -53.766 54.985
646536 94.568 -53.778 54.990
646537 94.592 -53.791 54.994
646538 94.616 -53.803 54.999
646539 94.639 -53.816 55.006
646540 94.663 -53.828 55.012
646541 94.687 -53.841 55.015
646542 94.710 -53.853 55.012
646543 94.734 -53.865 55.005
646544 94.758 -53.878 54.995
646545 94.781 -53.890 54.978
646546 94.805 -53.902 54.956
646547 94.829 -53.915 54.931
646548 94.852 -53.927 54.912
646549 94.876 -53.939 54.905
646550 94.900 -53.951 54.904
646551 94.923 -53.964 54.906
646552 94.947 -53.976 54.903
646553 94.971 -53.988 54.885
646554 94.994 -54.001 54.857
646555 95.018 -54.013 54.820
646556 95.042 -54.026 54.777
646557 95.065 -54.038 54.730
646558 95.089 -54.051 54.682
646559 95.113 -54.063 54.632
646560 95.136 -54.075 54.581
646561 95.160 -54.088 54.531
646562 95.195 -54.105 54.479
646563 95.232 -54.124 54.427
646564 95.268 -54.143 54.375
646565 95.295 -54.160 54.321
646566 95.318 -54.176 54.269
646567 95.342 -54.191 54.220
646568 95.365 -54.207 54.174
646569 95.389 -54.223 54.132
646570 95.412 -54.238 54.093
646571 95.436 -54.254 54.054
646572 95.460 -54.269 54.018
646573 95.483 -54.285 53.985
646574 95.507 -54.301 53.955
646575 95.531 -54.316 53.927
646576 95.555 -54.326 53.903
646577 95.580 -54.336 53.879
646578 95.604 -54.346 53.858
646579 95.628 -54.357 53.837
646580 95.665 -54.364 53.817
646581 95.701 -54.371 53.800
646582 95.738 -54.378 53.781
646583 95.774 -54.386 53.758
646584 95.811 -54.393 53.734
646585 95.839 -54.408 53.709
646586 95.865 -54.427 53.685
646587 95.890 -54.445 53.661
646588 95.915 -54.463 53.634
646589 95.941 -54.481 53.602
646590 95.966 -54.499 53.565
646591 95.991 -54.518 53.529
646592 96.017 -54.536 53.489
646593 96.043 -54.554 53.444
646594 96.068 -54.572 53.396
646595 96.096 -54.588 53.345
646596 96.126 -54.603 53.292
646597 96.155 -54.618 53.240
646598 96.185 -54.633 53.192
646599 96.214 -54.648 53.145
646600 96.244 -54.662 53.099
646601 96.270 -54.676 53.054
646602 96.295 -54.689 53.010
646603 96.321 -54.701 52.967
646604 96.346 -54.714 52.922
646605 96.372 -54.725 52.877
646606 96.396 -54.734 52.829
646607 96.420 -54.741 52.780
646608 96.448 -54.749 52.729
646609 96.476 -54.756 52.678
646610 96.506 -54.763 52.625
646611 96.534 -54.770 52.573
646612 96.563 -54.777 52.521
646613 96.592 -54.784 52.467
646614 96.621 -54.791 52.415
646615 96.650 -54.798 52.369
646616 96.678 -54.805 52.325
646617 96.708 -54.812 52.276
646618 96.739 -54.821 52.222
646619 96.770 -54.830 52.167
646620 96.801 -54.838 52.112
646621 96.832 -54.847 52.056
646622 96.864 -54.856 52.000
646623 96.895 -54.865 51.946
646624 96.926 -54.874 51.890
646625 96.957 -54.882 51.830
646626 96.990 -54.891 51.769
646627 97.029 -54.903 51.706
646628 97.069 -54.916 51.640
646629 97.108 -54.928 51.568
646630 97.148 -54.941 51.497
646631 97.187 -54.953 51.422
646632 97.225 -54.972 51.344
646633 97.260 -54.997 51.267
646634 97.281 -55.003 51.189
646635 97.298 -55.001 51.110
646636 97.327 -55.000 51.037
646637 97.362 -55.000 50.965
646638 97.388 -54.999 50.895
646639 97.410 -54.998 50.826
646640 97.434 -54.997 50.760
646641 97.457 -54.996 50.695
646642 97.480 -54.995 50.631
646643 97.503 -54.994 50.571
646644 97.526 -54.994 50.512
646645 97.549 -54.993 50.454
647111 90.856 -51.557 56.938
647112 90.888 -51.566 56.945
647113 90.921 -51.574 56.951
647114 90.952 -51.582 56.961
647115 90.964 -51.597 56.968
647116 90.975 -51.613 56.978
647117 90.988 -51.629 56.989
647118 90.999 -51.645 56.999
647119 91.011 -51.661 57.010
647120 91.023 -51.676 57.022
647121 91.034 -51.692 57.037
647122 91.046 -51.708 57.054
647123 91.058 -51.724 57.076
647124 91.078 -51.734 57.100
647125 91.109 -51.754 57.124
647126 91.141 -51.780 57.143
647127 91.173 -51.806 57.153
647128 91.205 -51.832 57.160
647129 91.237 -51.857 57.163
647130 91.269 -51.883 57.165
647131 91.301 -51.909 57.166
647132 91.333 -51.935 57.169
647133 91.365 -51.960 57.172
647134 91.397 -51.986 57.168
647135 91.431 -52.011 57.162
647136 91.482 -52.021 57.152
647137 91.532 -52.032 57.140
647138 91.571 -52.048 57.130
647139 91.605 -52.064 57.125
647140 91.636 -52.080 57.119
647141 91.666 -52.095 57.109
647142 91.698 -52.111 57.096
647143 91.728 -52.127 57.082
647144 91.759 -52.142 57.067
647145 91.790 -52.158 57.053
647146 91.821 -52.174 57.039
647147 91.848 -52.194 57.024
647148 91.875 -52.214 57.005
647149 91.903 -52.236 56.985
647150 91.930 -52.256 56.964
647151 91.957 -52.277 56.943
647152 91.985 -52.297 56.919
647153 92.012 -52.318 56.898
647154 92.039 -52.339 56.873
647155 92.067 -52.359 56.848
647156 92.094 -52.380 56.821
647157 92.121 -52.400 56.792
647158 92.149 -52.421 56.762
647159 92.176 -52.442 56.731
647160 92.203 -52.462 56.698
647161 92.230 -52.483 56.661
647162 92.258 -52.503 56.623
647163 92.285 -52.524 56.583
647164 92.312 -52.545 56.542
647165 92.340 -52.565 56.500
647166 92.367 -52.586 56.454
647167 92.408 -52.615 56.404
647168 92.448 -52.644 56.352
647169 92.489 -52.672 56.299
647170 92.533 -52.689 56.245
647171 92.581 -52.695 56.190
647172 92.630 -52.701 56.135
647173 92.677 -52.707 56.080
647174 92.725 -52.713 56.025
647175 92.773 -52.719 55.972
647176 92.821 -52.725 55.920
647177 92.869 -52.731 55.868
647178 92.917 -52.738 55.818
647179 92.965 -52.744 55.771
647180 92.996 -52.760 55.723
647181 93.015 -52.782 55.676
647182 93.035 -52.803 55.629
647214 93.999 -53.358 54.845
647215 94.030 -53.376 54.846
647216 94.060 -53.395 54.849
647217 94.091 -53.414 54.853
647218 94.122 -53.432 54.857
647219 94.152 -53.451 54.865
647220 94.183 -53.470 54.872
647221 94.213 -53.488 54.880
647222 94.244 -53.507 54.888
647223 94.274 -53.526 54.896
647224 94.305 -53.544 54.903
647225 94.335 -53.563 54.911
647226 94.351 -53.577 54.917
647227 94.327 -53.579 54.925
647228 94.337 -53.593 54.931
647229 94.367 -53.615 54.937
647230 94.398 -53.637 54.943
647231 94.428 -53.660 54.949
647232 94.458 -53.682 54.955
647233 94.489 -53.703 54.960
647234 94.518 -53.719 54.966
647235 94.542 -53.733 54.973
647236 94.566 -53.747 54.978
647237 94.591 -53.760 54.982
647238 94.615 -53.774 54.986
647239 94.639 -53.786 54.992
647240 94.663 -53.799 54.998
647241 94.686 -53.811 55.003
647242 94.710 -53.823 55.004
647243 94.734 -53.835 54.999
647244 94.757 -53.848 54.990
647245 94.781 -53.860 54.979
647246 94.805 -53.872 54.965
647247 94.828 -53.885 54.946
647248 94.852 -53.897 54.926
647249 94.876 -53.910 54.913
647250 94.899 -53.922 54.910
647251 94.923 -53.935 54.910
647252 94.947 -53.947 54.910
647253 94.970 -53.959 54.902
647254 94.994 -53.972 54.882
647255 95.018 -53.984 54.850
647256 95.041 -53.996 54.812
647257 95.065 -54.008 54.768
647258 95.089 -54.021 54.723
647259 95.112 -54.033 54.678
647260 95.136 -54.045 54.631
647261 95.160 -54.058 54.581
647262 95.183 -54.070 54.530
647263 95.211 -54.085 54.476
647264 95.247 -54.104 54.423
647265 95.284 -54.123 54.372
647266 95.309 -54.138 54.321
647267 95.332 -54.154 54.274
647268 95.356 -54.170 54.230
647269 95.380 -54.185 54.187
647270 95.403 -54.201 54.146
647271 95.427 -54.217 54.109
647272 95.451 -54.232 54.072
647273 95.474 -54.248 54.036
647274 95.498 -54.261 54.003
647275 95.523 -54.271 53.973
647276 95.547 -54.281 53.946
647277 95.571 -54.292 53.923
647278 95.596 -54.301 53.901
647279 95.620 -54.312 53.881
647280 95.644 -54.322 53.861
647281 95.677 -54.330 53.843
647282 95.713 -54.337 53.827
647283 95.750 -54.345 53.809
647284 95.787 -54.352 53.785
647285 95.822 -54.361 53.761
647286 95.847 -54.379 53.736
647287 95.873 -54.397 53.712
647288 95.898 -54.415 53.687
647289 95.924 -54.433 53.660
647290 95.950 -54.452 53.628
647291 95.975 -54.470 53.595
647292 96.001 -54.488 53.559
647293 96.026 -54.506 53.519
647294 96.051 -54.524 53.473
647295 96.077 -54.543 53.423
647296 96.102 -54.561 53.370
647297 96.132 -54.575 53.316
647298 96.161 -54.590 53.265
647299 96.191 -54.605 53.216
647300 96.220 -54.620 53.169
647301 96.250 -54.635 53.122
647302 96.275 -54.647 53.077
647303 96.301 -54.660 53.033
647304 96.326 -54.673 52.990
647305 96.352 -54.686 52.944
647306 96.377 -54.699 52.899
647307 96.403 -54.711 52.854
647308 96.427 -54.719 52.806
647309 96.455 -54.727 52.755
647310 96.484 -54.734 52.702
647311 96.513 -54.740 52.651
647312 96.542 -54.747 52.598
647313 96.571 -54.755 52.546
647314 96.599 -54.762 52.494
647315 96.628 -54.768 52.445
647316 96.657 -54.776 52.402
647317 96.686 -54.783 52.360
647318 96.715 -54.790 52.314
647319 96.745 -54.798 52.264
647320 96.777 -54.807 52.211
647321 96.808 -54.815 52.157
647322 96.839 -54.825 52.102
647323 96.871 -54.833 52.046
647324 96.902 -54.842 51.991
647325 96.933 -54.850 51.935
647326 96.964 -54.859 51.874
647327 96.996 -54.869 51.812
647328 97.030 -54.878 51.749
647329 97.070 -54.891 51.683
647330 97.109 -54.903 51.614
647331 97.149 -54.916 51.545
647332 97.188 -54.928 51.471
647333 97.228 -54.941 51.396
647334 97.265 -54.959 51.320
647335 97.296 -54.977 51.243
647336 97.312 -54.976 51.164
647337 97.344 -54.975 51.088
647338 97.373 -54.975 51.015
647339 97.396 -54.974 50.943
647340 97.419 -54.973 50.873
647341 97.442 -54.972 50.804
647342 97.465 -54.972 50.738
647343 97.489 -54.970 50.674
647344 97.512 -54.969 50.611
647345 97.535 -54.969 50.551
647346 97.558 -54.968 50.492
647347 97.581 -54.967 50.433
647812 90.876 -51.524 56.927
647813 90.908 -51.532 56.934
647814 90.941 -51.541 56.940
647815 90.957 -51.554 56.948
647816 90.969 -51.570 56.957
647817 90.980 -51.586 56.966
647818 90.993 -51.601 56.977
647819 91.004 -51.617 56.988
647820 91.016 -51.633 56.999
647821 91.028 -51.649 57.012
647822 91.039 -51.665 57.026
647823 91.051 -51.681 57.044
647824 91.063 -51.696 57.067
647825 91.081 -51.707 57.092
647826 91.112 -51.728 57.116
647827 91.144 -51.753 57.138
647828 91.176 -51.779 57.151
647829 91.208 -51.804 57.159
647830 91.240 -51.830 57.162
647831 91.272 -51.856 57.161
647832 91.304 -51.882 57.160
647833 91.336 -51.907 57.160
647834 91.368 -51.933 57.161
647835 91.403 -51.957 57.159
647836 91.454 -51.967 57.153
647837 91.504 -51.979 57.142
647838 91.552 -51.991 57.129
647839 91.589 -52.007 57.119
647840 91.624 -52.024 57.114
647841 91.654 -52.039 57.107
647842 91.685 -52.054 57.099
647843 91.716 -52.070 57.088
647844 91.747 -52.086 57.074
647845 91.778 -52.101 57.058
647846 91.809 -52.117 57.042
647847 91.840 -52.133 57.028
647848 91.871 -52.148 57.013
647849 91.902 -52.164 56.995
647850 91.931 -52.182 56.976
647851 91.958 -52.202 56.956
647852 91.985 -52.224 56.935
647853 92.012 -52.244 56.913
647854 92.040 -52.264 56.892
647855 92.067 -52.285 56.867
647856 92.094 -52.305 56.842
647857 92.122 -52.327 56.817
647858 92.149 -52.347 56.791
647859 92.176 -52.368 56.762
647860 92.204 -52.388 56.730
647861 92.231 -52.409 56.695
647862 92.259 -52.430 56.657
647863 92.286 -52.450 56.619
647864 92.313 -52.471 56.580
647865 92.340 -52.491 56.539
647866 92.368 -52.512 56.498
647867 92.395 -52.533 56.452
647868 92.436 -52.562 56.403
647869 92.477 -52.590 56.352
647870 92.517 -52.619 56.300
647871 92.558 -52.647 56.247
647872 92.604 -52.657 56.193
647873 92.652 -52.663 56.138
647874 92.700 -52.669 56.084
647875 92.748 -52.675 56.031
647876 92.796 -52.681 55.978
647877 92.844 -52.687 55.926
647878 92.892 -52.694 55.875
647879 92.940 -52.700 55.825
647880 92.981 -52.709 55.778
647881 93.001 -52.731 55.729
647882 93.020 -52.753 55.681
647883 93.045 -52.772 55.634
647884 93.075 -52.786 55.589
647885 93.105 -52.800 55.545
647915 94.017 -53.332 54.845
647916 94.047 -53.351 54.846
647917 94.078 -53.370 54.848
647918 94.108 -53.388 54.852
647919 94.139 -53.407 54.856
647920 94.170 -53.426 54.863
647921 94.200 -53.445 54.870
647922 94.230 -53.463 54.878
647923 94.261 -53.482 54.886
647924 94.291 -53.501 54.893
647925 94.322 -53.519 54.901
647926 94.352 -53.538 54.909
647927 94.377 -53.555 54.916
647928 94.353 -53.556 54.923
647929 94.347 -53.565 54.930
647930 94.378 -53.587 54.937
647931 94.408 -53.609 54.944
647932 94.437 -53.631 54.951
647933 94.469 -53.652 54.956
647934 94.502 -53.672 54.960
647935 94.535 -53.691 54.964
647936 94.559 -53.705 54.969
647937 94.583 -53.719 54.973
647938 94.608 -53.733 54.975
647939 94.632 -53.747 54.978
647940 94.657 -53.760 54.982
647941 94.681 -53.774 54.987
647942 94.706 -53.787 54.991
647943 94.730 -53.801 54.990
647944 94.755 -53.815 54.983
647945 94.779 -53.828 54.973
647946 94.804 -53.842 54.963
647947 94.828 -53.856 54.951
647948 94.851 -53.868 54.937
647949 94.875 -53.880 54.924
647950 94.899 -53.892 54.917
647951 94.922 -53.905 54.917
647952 94.946 -53.917 54.917
647953 94.970 -53.929 54.913
647954 94.993 -53.942 54.900
647955 95.017 -53.954 54.876
647956 95.040 -53.966 54.843
647957 95.064 -53.979 54.803
647958 95.088 -53.991 54.761
647959 95.111 -54.004 54.718
647960 95.135 -54.016 54.676
647961 95.159 -54.029 54.628
647962 95.182 -54.041 54.577
647963 95.206 -54.053 54.526
647964 95.230 -54.066 54.473
647965 95.263 -54.082 54.423
647966 95.300 -54.101 54.377
647967 95.323 -54.117 54.331
647968 95.347 -54.133 54.287
647969 95.370 -54.148 54.244
647970 95.394 -54.164 54.201
647971 95.417 -54.180 54.161
647972 95.441 -54.195 54.126
647973 95.466 -54.206 54.089
647974 95.490 -54.216 54.054
647975 95.515 -54.226 54.022
647976 95.539 -54.236 53.992
647977 95.563 -54.246 53.965
647978 95.588 -54.257 53.944
647979 95.612 -54.267 53.922
647980 95.636 -54.277 53.903
647981 95.661 -54.287 53.885
647982 95.690 -54.296 53.868
647983 95.726 -54.304 53.852
647984 95.763 -54.311 53.834
647985 95.800 -54.318 53.810
647986 95.831 -54.330 53.785
647987 95.856 -54.348 53.760
647988 95.882 -54.367 53.735
647989 95.907 -54.385 53.710
647990 95.933 -54.403 53.683
647991 95.958 -54.421 53.652
647992 95.983 -54.439 53.620
647993 96.009 -54.458 53.585
647994 96.034 -54.476 53.545
647995 96.060 -54.494 53.498
647996 96.086 -54.512 53.447
647997 96.111 -54.530 53.392
647998 96.138 -54.547 53.339
647999 96.168 -54.562 53.288
648000 96.197 -54.577 53.238
648001 96.227 -54.592 53.191
648002 96.255 -54.606 53.145
648003 96.281 -54.619 53.099
648004 96.306 -54.632 53.055
648005 96.331 -54.645 53.010
648006 96.357 -54.658 52.964
648007 96.383 -54.670 52.920
648008 96.408 -54.683 52.874
648009 96.434 -54.696 52.826
648010 96.463 -54.704 52.776
648011 96.491 -54.711 52.724
648012 96.521 -54.718 52.672
648013 96.549 -54.725 52.621
648014 96.578 -54.732 52.569
648015 96.607 -54.740 52.521
648016 96.636 -54.746 52.475
648017 96.665 -54.753 52.433
648018 96.693 -54.760 52.392
648019 96.722 -54.768 52.351
648020 96.752 -54.775 52.306
648021 96.784 -54.784 52.256
648022 96.815 -54.793 52.204
648023 96.846 -54.802 52.148
648024 96.877 -54.810 52.092
648025 96.908 -54.819 52.035
648026 96.940 -54.828 51.977
648027 96.971 -54.837 51.916
648028 97.003 -54.845 51.854
648029 97.034 -54.854 51.790
648030 97.070 -54.865 51.725
648031 97.110 -54.878 51.657
648032 97.149 -54.890 51.589
648033 97.189 -54.903 51.518
648034 97.229 -54.915 51.446
648035 97.268 -54.928 51.371
648036 97.305 -54.946 51.296
648037 97.326 -54.951 51.219
648038 97.359 -54.950 51.142
648039 97.381 -54.950 51.066
648040 97.404 -54.949 50.993
648041 97.428 -54.948 50.921
648042 97.451 -54.947 50.850
648043 97.474 -54.947 50.782
648044 97.497 -54.945 50.718
648045 97.520 -54.944 50.653
648046 97.543 -54.944 50.592
648047 97.567 -54.943 50.531
648048 97.589 -54.942 50.471
648049 97.612 -54.941 50.413
648513 90.896 -51.490 56.916
648514 90.928 -51.499 56.922
648515 90.951 -51.510 56.928
648516 90.962 -51.526 56.936
648517 90.974 -51.542 56.945
648518 90.986 -51.558 56.954
648519 90.998 -51.574 56.965
648520 91.009 -51.590 56.977
648521 91.021 -51.605 56.987
648522 91.033 -51.621 57.000
648523 91.044 -51.637 57.014
648524 91.056 -51.653 57.032
648525 91.068 -51.669 57.055
648526 91.085 -51.681 57.081
648527 91.115 -51.700 57.105
648528 91.147 -51.726 57.128
648529 91.179 -51.752 57.143
648530 91.212 -51.777 57.155
648531 91.243 -51.803 57.158
648532 91.276 -51.829 57.155
648533 91.308 -51.854 57.150
648534 91.340 -51.880 57.149
648535 91.375 -51.903 57.148
648536 91.426 -51.914 57.146
648537 91.476 -51.925 57.139
648538 91.527 -51.936 57.128
648539 91.570 -51.950 57.115
648540 91.608 -51.967 57.107
648541 91.642 -51.983 57.101
648542 91.673 -51.999 57.095
648543 91.704 -52.014 57.087
648544 91.735 -52.029 57.076
648545 91.766 -52.045 57.063
648546 91.797 -52.061 57.049
648547 91.828 -52.076 57.033
648548 91.859 -52.092 57.019
648549 91.890 -52.108 57.004
648550 91.921 -52.123 56.986
648551 91.952 -52.139 56.967
648552 91.983 -52.154 56.949
648553 92.013 -52.170 56.928
648554 92.040 -52.191 56.907
648555 92.068 -52.211 56.887
648556 92.095 -52.232 56.862
648557 92.122 -52.252 56.838
648558 92.150 -52.273 56.814
648559 92.177 -52.294 56.788
648560 92.204 -52.315 56.760
648561 92.231 -52.335 56.727
648562 92.259 -52.355 56.692
648563 92.286 -52.376 56.656
648564 92.313 -52.397 56.618
648565 92.341 -52.418 56.578
648566 92.368 -52.438 56.537
648567 92.395 -52.459 56.497
648568 92.423 -52.479 56.451
648569 92.463 -52.508 56.403
648570 92.504 -52.537 56.353
648571 92.545 -52.565 56.302
648572 92.585 -52.594 56.249
648573 92.627 -52.619 56.196
648574 92.675 -52.625 56.142
648575 92.723 -52.631 56.090
648576 92.771 -52.637 56.037
648577 92.819 -52.644 55.985
648578 92.867 -52.650 55.934
648579 92.915 -52.656 55.883
648580 92.963 -52.662 55.833
648581 92.986 -52.681 55.786
648582 93.006 -52.703 55.738
648583 93.031 -52.722 55.690
648584 93.061 -52.736 55.642
648585 93.091 -52.750 55.595
648586 93.121 -52.765 55.551
648587 93.151 -52.780 55.508
648588 93.181 -52.794 55.465
648617 94.064 -53.326 54.846
648618 94.095 -53.345 54.849
648619 94.126 -53.363 54.852
648620 94.156 -53.382 54.856
648621 94.187 -53.401 54.862
648622 94.217 -53.419 54.868
648623 94.248 -53.438 54.875
648624 94.278 -53.457 54.883
648625 94.309 -53.475 54.891
648626 94.339 -53.494 54.899
648627 94.369 -53.513 54.907
648628 94.400 -53.531 54.915
648629 94.380 -53.534 54.922
648630 94.357 -53.537 54.929
648631 94.388 -53.559 54.937
648632 94.417 -53.581 54.945
648633 94.449 -53.602 54.952
648634 94.482 -53.621 54.957
648635 94.514 -53.641 54.960
648636 94.547 -53.660 54.963
648637 94.576 -53.678 54.966
648638 94.601 -53.691 54.969
648639 94.625 -53.705 54.968
648640 94.649 -53.719 54.970
648641 94.674 -53.732 54.973
648642 94.698 -53.747 54.975
648643 94.723 -53.760 54.976
648644 94.747 -53.774 54.975
648645 94.772 -53.787 54.968
648646 94.796 -53.801 54.958
648647 94.821 -53.815 54.950
648648 94.845 -53.828 54.942
648649 94.870 -53.842 54.933
648650 94.894 -53.856 54.927
648651 94.918 -53.870 54.926
648652 94.943 -53.884 54.928
648653 94.967 -53.897 54.925
648654 94.992 -53.911 54.916
648655 95.016 -53.925 54.896
648656 95.040 -53.937 54.869
648657 95.063 -53.950 54.835
648658 95.087 -53.962 54.796
648659 95.111 -53.974 54.757
648660 95.134 -53.986 54.716
648661 95.158 -53.999 54.675
648662 95.182 -54.011 54.627
648663 95.205 -54.023 54.575
648664 95.229 -54.036 54.524
648665 95.253 -54.048 54.473
648666 95.279 -54.062 54.428
648667 95.314 -54.080 54.386
648668 95.337 -54.096 54.343
648669 95.361 -54.112 54.300
648670 95.385 -54.127 54.257
648671 95.409 -54.141 54.216
648672 95.433 -54.151 54.176
648673 95.458 -54.160 54.141
648674 95.482 -54.171 54.106
648675 95.506 -54.181 54.072
648676 95.531 -54.192 54.039
648677 95.555 -54.201 54.010
648678 95.580 -54.212 53.985
648679 95.604 -54.222 53.964
648680 95.628 -54.232 53.942
648681 95.653 -54.242 53.924
648682 95.677 -54.253 53.907
648683 95.702 -54.263 53.890
648684 95.739 -54.270 53.873
648685 95.775 -54.277 53.855
648686 95.812 -54.284 53.830
648687 95.839 -54.301 53.804
648688 95.865 -54.319 53.778
648689 95.890 -54.337 53.754
648690 95.916 -54.355 53.730
648691 95.941 -54.373 53.704
648692 95.967 -54.392 53.675
648693 95.992 -54.410 53.643
648694 96.018 -54.428 53.607
648695 96.043 -54.446 53.566
648696 96.068 -54.464 53.520
648697 96.094 -54.483 53.468
648698 96.119 -54.501 53.413
648699 96.145 -54.519 53.360
648700 96.174 -54.534 53.309
648701 96.203 -54.549 53.259
648702 96.233 -54.565 53.212
648703 96.260 -54.578 53.165
648704 96.286 -54.591 53.120
648705 96.312 -54.604 53.075
648706 96.337 -54.617 53.029
648707 96.363 -54.629 52.983
648708 96.388 -54.642 52.938
648709 96.415 -54.655 52.891
648710 96.443 -54.667 52.841
648711 96.472 -54.679 52.792
648712 96.500 -54.687 52.742
648713 96.528 -54.695 52.692
648714 96.557 -54.703 52.642
648715 96.586 -54.710 52.593
648716 96.614 -54.717 52.547
648717 96.643 -54.724 52.505
648718 96.672 -54.731 52.464
648719 96.701 -54.738 52.424
648720 96.730 -54.745 52.385
648721 96.759 -54.753 52.343
648722 96.790 -54.761 52.297
648723 96.821 -54.770 52.245
648724 96.853 -54.779 52.190
648725 96.884 -54.787 52.133
648726 96.915 -54.797 52.075
648727 96.947 -54.805 52.016
648728 96.978 -54.814 51.955
648729 97.009 -54.822 51.894
648730 97.040 -54.831 51.832
648731 97.072 -54.840 51.770
648732 97.110 -54.852 51.705
648733 97.150 -54.865 51.638
648734 97.189 -54.877 51.569
648735 97.229 -54.890 51.498
648736 97.269 -54.902 51.425
648737 97.308 -54.914 51.350
648738 97.341 -54.926 51.277
648739 97.367 -54.925 51.201
648740 97.390 -54.925 51.119
648741 97.413 -54.924 51.042
648742 97.436 -54.923 50.969
648743 97.459 -54.922 50.898
648744 97.483 -54.922 50.829
648745 97.506 -54.920 50.763
648746 97.529 -54.919 50.697
648747 97.552 -54.919 50.634
648748 97.575 -54.918 50.572
648749 97.598 -54.917 50.511
648750 97.621 -54.916 50.452
648751 97.644 -54.916 50.395
649214 90.916 -51.457 56.904
649215 90.944 -51.467 56.909
649216 90.956 -51.483 56.916
649217 90.967 -51.499 56.924
649218 90.979 -51.515 56.933
649219 90.991 -51.530 56.943
649220 91.003 -51.546 56.953
649221 91.014 -51.562 56.965
649222 91.026 -51.578 56.975
649223 91.038 -51.594 56.988
649224 91.049 -51.609 57.002
649225 91.062 -51.625 57.019
649226 91.073 -51.641 57.041
649227 91.088 -51.654 57.066
649228 91.118 -51.673 57.090
649229 91.151 -51.699 57.115
649230 91.183 -51.725 57.129
649231 91.215 -51.750 57.139
649232 91.247 -51.776 57.144
649233 91.279 -51.801 57.142
649234 91.311 -51.827 57.135
649235 91.348 -51.850 57.134
649236 91.398 -51.860 57.132
649237 91.448 -51.872 57.129
649238 91.499 -51.882 57.122
649239 91.549 -51.893 57.112
649240 91.589 -51.910 57.101
649241 91.626 -51.926 57.094
649242 91.661 -51.943 57.089
649243 91.692 -51.958 57.083
649244 91.723 -51.973 57.076
649245 91.754 -51.989 57.066
649246 91.785 -52.004 57.052
649247 91.816 -52.020 57.038
649248 91.847 -52.036 57.023
649249 91.878 -52.051 57.009
649250 91.909 -52.067 56.994
649251 91.940 -52.083 56.977
649252 91.970 -52.098 56.960
649253 92.001 -52.114 56.942
649254 92.032 -52.129 56.922
649255 92.063 -52.145 56.900
649256 92.094 -52.160 56.881
649257 92.123 -52.179 56.858
649258 92.150 -52.199 56.834
649259 92.177 -52.220 56.810
649260 92.205 -52.240 56.785
649261 92.232 -52.261 56.757
649262 92.260 -52.282 56.725
649263 92.287 -52.302 56.691
649264 92.314 -52.323 56.655
649265 92.341 -52.343 56.618
649266 92.369 -52.365 56.578
649267 92.396 -52.385 56.538
649268 92.423 -52.405 56.498
649269 92.450 -52.426 56.453
649270 92.491 -52.455 56.404
649271 92.532 -52.484 56.354
649272 92.573 -52.512 56.303
649273 92.613 -52.541 56.252
649274 92.654 -52.569 56.200
649275 92.698 -52.587 56.148
649276 92.746 -52.593 56.096
649277 92.794 -52.600 56.045
649278 92.842 -52.606 55.994
649279 92.889 -52.612 55.943
649280 92.938 -52.618 55.893
649281 92.972 -52.631 55.843
649282 92.991 -52.653 55.796
649283 93.017 -52.671 55.747
649284 93.047 -52.686 55.699
649285 93.077 -52.700 55.651
649286 93.107 -52.715 55.604
649287 93.137 -52.729 55.559
649288 93.167 -52.744 55.515
649289 93.197 -52.759 55.472
649290 93.227 -52.777 55.430
649291 93.258 -52.795 55.388
649318 94.082 -53.301 54.848
649319 94.112 -53.319 54.849
649320 94.143 -53.338 54.851
649321 94.174 -53.357 54.856
649322 94.204 -53.375 54.861
649323 94.235 -53.394 54.867
649324 94.265 -53.413 54.874
649325 94.295 -53.431 54.882
649326 94.326 -53.450 54.890
649327 94.356 -53.469 54.897
649328 94.387 -53.487 54.905
649329 94.417 -53.506 54.913
649330 94.406 -53.512 54.921
649331 94.382 -53.513 54.928
649332 94.398 -53.530 54.936
649333 94.428 -53.551 54.946
649334 94.461 -53.571 54.953
649335 94.494 -53.590 54.958
649336 94.527 -53.610 54.962
649337 94.559 -53.630 54.963
649338 94.592 -53.650 54.964
649339 94.618 -53.664 54.965
649340 94.642 -53.678 54.963
649341 94.667 -53.691 54.963
649342 94.691 -53.705 54.964
649343 94.715 -53.719 54.965
649344 94.740 -53.732 54.965
649345 94.764 -53.746 54.964
649346 94.789 -53.760 54.960
649347 94.813 -53.774 54.954
649348 94.838 -53.787 54.951
649349 94.862 -53.801 54.944
649350 94.887 -53.815 54.939
649351 94.911 -53.828 54.936
649352 94.935 -53.842 54.936
649353 94.960 -53.856 54.933
649354 94.984 -53.870 54.924
649355 95.009 -53.884 54.908
649356 95.033 -53.897 54.889
649357 95.058 -53.911 54.864
649358 95.082 -53.925 54.832
649359 95.107 -53.938 54.796
649360 95.131 -53.952 54.759
649361 95.156 -53.966 54.720
649362 95.180 -53.979 54.678
649363 95.204 -53.994 54.630
649364 95.229 -54.007 54.577
649365 95.252 -54.019 54.525
649366 95.276 -54.031 54.477
649367 95.300 -54.043 54.437
649368 95.328 -54.059 54.397
649369 95.352 -54.075 54.356
649370 95.376 -54.085 54.314
649371 95.400 -54.095 54.271
649372 95.425 -54.106 54.230
649373 95.449 -54.116 54.191
649374 95.474 -54.126 54.157
649375 95.498 -54.136 54.122
649376 95.523 -54.147 54.089
649377 95.547 -54.157 54.057
649378 95.571 -54.167 54.028
649379 95.596 -54.177 54.003
649380 95.620 -54.188 53.982
649381 95.644 -54.198 53.961
649382 95.669 -54.207 53.944
649383 95.693 -54.218 53.928
649384 95.718 -54.228 53.911
649385 95.751 -54.236 53.893
649386 95.788 -54.243 53.871
649387 95.823 -54.252 53.844
649388 95.848 -54.270 53.816
649389 95.874 -54.289 53.790
649390 95.899 -54.307 53.766
649391 95.924 -54.325 53.742
649392 95.950 -54.343 53.717
649393 95.975 -54.361 53.689
649394 96.001 -54.380 53.658
649395 96.026 -54.398 53.623
649396 96.052 -54.416 53.582
649397 96.077 -54.434 53.535
649398 96.103 -54.452 53.482
649399 96.128 -54.471 53.427
649400 96.154 -54.489 53.376
649401 96.180 -54.507 53.326
649402 96.209 -54.521 53.279
649403 96.239 -54.536 53.231
649404 96.266 -54.550 53.184
649405 96.292 -54.563 53.138
649406 96.317 -54.576 53.093
649407 96.342 -54.588 53.048
649408 96.369 -54.601 53.000
649409 96.396 -54.613 52.953
649410 96.424 -54.626 52.904
649411 96.452 -54.638 52.854
649412 96.480 -54.651 52.805
649413 96.509 -54.662 52.757
649414 96.537 -54.670 52.710
649415 96.566 -54.678 52.663
649416 96.594 -54.686 52.618
649417 96.622 -54.694 52.574
649418 96.651 -54.702 52.534
649419 96.680 -54.709 52.495
649420 96.709 -54.715 52.456
649421 96.737 -54.723 52.417
649422 96.766 -54.730 52.377
649423 96.797 -54.738 52.333
649424 96.828 -54.747 52.283
649425 96.860 -54.756 52.229
649426 96.891 -54.765 52.172
649427 96.922 -54.773 52.113
649428 96.953 -54.782 52.055
649429 96.984 -54.791 51.996
649430 97.016 -54.800 51.937
649431 97.047 -54.809 51.878
649432 97.079 -54.817 51.819
649433 97.111 -54.826 51.756
649434 97.151 -54.839 51.690
649435 97.190 -54.851 51.622
649436 97.230 -54.864 51.551
649437 97.270 -54.876 51.477
649438 97.310 -54.887 51.403
649439 97.350 -54.897 51.331
649440 97.375 -54.900 51.255
649441 97.398 -54.900 51.172
649442 97.422 -54.899 51.094
649443 97.445 -54.898 51.020
649444 97.468 -54.897 50.947
649445 97.491 -54.897 50.877
649446 97.514 -54.896 50.811
649447 97.537 -54.894 50.744
649448 97.560 -54.894 50.679
649449 97.583 -54.893 50.617
649450 97.606 -54.892 50.554
649451 97.630 -54.891 50.494
649452 97.653 -54.891 50.436
649453 97.676 -54.890 50.377
649454 97.699 -54.889 50.320
649455 97.720 -54.882 50.265
649915 90.936 -51.424 56.893
649916 90.949 -51.439 56.900
649917 90.961 -51.455 56.906
649918 90.972 -51.471 56.913
649919 90.984 -51.487 56.921
649920 90.996 -51.503 56.931
649921 91.008 -51.519 56.942
649922 91.019 -51.534 56.952
649923 91.031 -51.550 56.963
649924 91.043 -51.566 56.975
649925 91.054 -51.581 56.989
649926 91.067 -51.597 57.006
649927 91.078 -51.613 57.026
649928 91.092 -51.628 57.050
649929 91.122 -51.646 57.073
649930 91.154 -51.672 57.097
649931 91.186 -51.697 57.111
649932 91.218 -51.723 57.120
649933 91.250 -51.749 57.125
649934 91.282 -51.775 57.124
649935 91.319 -51.796 57.116
649936 91.370 -51.807 57.116
649937 91.421 -51.818 57.114
649938 91.471 -51.829 57.111
649939 91.522 -51.840 57.105
649940 91.569 -51.852 57.097
649941 91.607 -51.869 57.087
649942 91.645 -51.885 57.081
649943 91.680 -51.902 57.078
649944 91.711 -51.918 57.073
649945 91.742 -51.933 57.066
649946 91.773 -51.948 57.055
649947 91.804 -51.964 57.042
649948 91.835 -51.979 57.027
649949 91.866 -51.995 57.013
649950 91.896 -52.011 56.999
649951 91.927 -52.026 56.985
649952 91.958 -52.042 56.969
649953 91.989 -52.058 56.952
649954 92.020 -52.073 56.935
649955 92.051 -52.089 56.916
649956 92.082 -52.104 56.896
649957 92.113 -52.120 56.876
649958 92.144 -52.135 56.854
649959 92.175 -52.150 56.830
649960 92.205 -52.167 56.806
649961 92.233 -52.187 56.781
649962 92.260 -52.208 56.753
649963 92.287 -52.228 56.723
649964 92.314 -52.249 56.690
649965 92.342 -52.270 56.655
649966 92.369 -52.290 56.618
649967 92.397 -52.311 56.580
649968 92.424 -52.331 56.540
649969 92.451 -52.352 56.499
649970 92.479 -52.373 56.455
649971 92.519 -52.402 56.406
649972 92.560 -52.430 56.355
649973 92.600 -52.459 56.305
649974 92.641 -52.487 56.255
649975 92.682 -52.516 56.204
649976 92.722 -52.545 56.153
649977 92.769 -52.556 56.104
649978 92.816 -52.562 56.054
649979 92.865 -52.568 56.005
649980 92.913 -52.574 55.954
649981 92.957 -52.581 55.905
649982 92.977 -52.603 55.855
649983 93.003 -52.621 55.807
649984 93.033 -52.636 55.758
649985 93.063 -52.650 55.710
649986 93.093 -52.665 55.661
649987 93.123 -52.679 55.614
649988 93.153 -52.695 55.567
649989 93.184 -52.714 55.523
649990 93.214 -52.733 55.480
649991 93.245 -52.751 55.437
649992 93.275 -52.770 55.395
650020 94.130 -53.294 54.850
650021 94.160 -53.313 54.851
650022 94.191 -53.331 54.856
650023 94.221 -53.350 54.860
650024 94.252 -53.369 54.866
650025 94.282 -53.387 54.873
650026 94.313 -53.406 54.880
650027 94.343 -53.425 54.888
650028 94.373 -53.443 54.896
650029 94.404 -53.462 54.903
650030 94.435 -53.481 54.911
650031 94.432 -53.490 54.919
650032 94.408 -53.491 54.928
650033 94.408 -53.501 54.937
650034 94.441 -53.520 54.947
650035 94.474 -53.540 54.955
650036 94.506 -53.559 54.960
650037 94.539 -53.579 54.962
650038 94.572 -53.599 54.963
650039 94.605 -53.618 54.964
650040 94.635 -53.636 54.962
650041 94.659 -53.650 54.959
650042 94.684 -53.664 54.959
650043 94.708 -53.678 54.958
650044 94.733 -53.691 54.958
650045 94.757 -53.705 54.959
650046 94.781 -53.719 54.958
650047 94.806 -53.732 54.957
650048 94.830 -53.746 54.958
650049 94.855 -53.759 54.957
650050 94.879 -53.774 54.951
650051 94.904 -53.787 54.945
650052 94.928 -53.801 54.942
650053 94.953 -53.815 54.941
650054 94.977 -53.828 54.934
650055 95.001 -53.842 54.918
650056 95.026 -53.856 54.899
650057 95.050 -53.869 54.882
650058 95.075 -53.884 54.860
650059 95.099 -53.897 54.831
650060 95.124 -53.911 54.800
650061 95.148 -53.925 54.765
650062 95.173 -53.938 54.728
650063 95.197 -53.952 54.686
650064 95.222 -53.966 54.639
650065 95.246 -53.979 54.587
650066 95.270 -53.994 54.534
650067 95.295 -54.007 54.486
650068 95.319 -54.022 54.449
650069 95.340 -54.037 54.411
650070 95.367 -54.044 54.370
650071 95.392 -54.051 54.327
650072 95.417 -54.061 54.285
650073 95.441 -54.071 54.244
650074 95.465 -54.082 54.206
650075 95.490 -54.091 54.173
650076 95.514 -54.102 54.139
650077 95.539 -54.112 54.105
650078 95.563 -54.122 54.074
650079 95.588 -54.132 54.045
650080 95.612 -54.142 54.018
650081 95.636 -54.153 53.998
650082 95.661 -54.163 53.979
650083 95.685 -54.173 53.962
650084 95.709 -54.183 53.947
650085 95.734 -54.194 53.931
650086 95.764 -54.202 53.912
650087 95.801 -54.210 53.888
650088 95.831 -54.223 53.857
650089 95.856 -54.241 53.827
650090 95.882 -54.259 53.800
650091 95.908 -54.277 53.775
650092 95.933 -54.295 53.752
650093 95.959 -54.314 53.728
650094 95.984 -54.332 53.700
650095 96.010 -54.350 53.671
650096 96.035 -54.368 53.635
650097 96.060 -54.386 53.594
650098 96.086 -54.405 53.547
650099 96.111 -54.423 53.495
650100 96.137 -54.441 53.440
650101 96.162 -54.459 53.390
650102 96.188 -54.477 53.342
650103 96.216 -54.494 53.296
650104 96.245 -54.508 53.250
650105 96.271 -54.522 53.203
650106 96.297 -54.535 53.154
650107 96.322 -54.547 53.106
650108 96.349 -54.560 53.059
650109 96.377 -54.572 53.012
650110 96.404 -54.585 52.965
650111 96.432 -54.597 52.916
650112 96.460 -54.609 52.868
650113 96.489 -54.622 52.821
650114 96.518 -54.635 52.774
650115 96.546 -54.644 52.729
650116 96.575 -54.652 52.686
650117 96.603 -54.660 52.644
650118 96.631 -54.668 52.602
650119 96.660 -54.677 52.562
650120 96.688 -54.685 52.524
650121 96.716 -54.693 52.486
650122 96.745 -54.700 52.447
650123 96.773 -54.708 52.407
650124 96.804 -54.715 52.364
650125 96.835 -54.724 52.317
650126 96.866 -54.733 52.265
650127 96.898 -54.742 52.208
650128 96.929 -54.750 52.151
650129 96.960 -54.759 52.094
650130 96.992 -54.768 52.037
650131 97.023 -54.777 51.982
650132 97.054 -54.786 51.926
650133 97.085 -54.794 51.869
650134 97.117 -54.803 51.808
650135 97.152 -54.812 51.743
650136 97.192 -54.823 51.674
650137 97.233 -54.834 51.602
650138 97.273 -54.845 51.528
650139 97.314 -54.856 51.454
650140 97.353 -54.866 51.382
650141 97.387 -54.862 51.306
650142 97.407 -54.875 51.226
650143 97.430 -54.874 51.147
650144 97.453 -54.873 51.070
650145 97.476 -54.872 50.997
650146 97.500 -54.872 50.925
650147 97.523 -54.871 50.859
650148 97.545 -54.869 50.792
650149 97.569 -54.869 50.726
650150 97.592 -54.868 50.662
650151 97.615 -54.867 50.598
650152 97.638 -54.866 50.537
650153 97.661 -54.866 50.478
650154 97.684 -54.865 50.418
650155 97.707 -54.862 50.360
650156 97.728 -54.855 50.302
650157 97.749 -54.849 50.246
650158 97.771 -54.842 50.191
650159 97.792 -54.836 50.138
650616 90.942 -51.396 56.883
650617 90.954 -51.412 56.890
650618 90.966 -51.428 56.896
650619 90.977 -51.443 56.902
650620 90.990 -51.459 56.910
650621 91.001 -51.475 56.919
650622 91.013 -51.490 56.930
650623 91.024 -51.506 56.940
650624 91.036 -51.522 56.951
650625 91.048 -51.538 56.962
650626 91.060 -51.553 56.976
650627 91.072 -51.569 56.991
650628 91.083 -51.585 57.009
650629 91.095 -51.601 57.032
650630 91.125 -51.619 57.054
650631 91.157 -51.644 57.072
650632 91.189 -51.670 57.088
650633 91.221 -51.696 57.098
650634 91.253 -51.722 57.104
650635 91.292 -51.742 57.104
650636 91.342 -51.754 57.098
650637 91.393 -51.764 57.098
650638 91.443 -51.776 57.097
650639 91.494 -51.786 57.094
650640 91.544 -51.797 57.088
650641 91.588 -51.811 57.082
650642 91.625 -51.829 57.074
650643 91.663 -51.845 57.071
650644 91.699 -51.862 57.069
650645 91.730 -51.877 57.065
650646 91.760 -51.892 57.058
650647 91.792 -51.908 57.047
650648 91.822 -51.923 57.032
650649 91.854 -51.939 57.017
650650 91.884 -51.954 57.003
650651 91.915 -51.970 56.990
650652 91.946 -51.986 56.976
650653 91.977 -52.001 56.961
650654 92.008 -52.017 56.945
650655 92.039 -52.032 56.928
650656 92.070 -52.048 56.910
650657 92.101 -52.064 56.891
650658 92.132 -52.079 56.871
650659 92.163 -52.095 56.851
650660 92.194 -52.110 56.828
650661 92.225 -52.125 56.805
650662 92.255 -52.141 56.780
650663 92.287 -52.156 56.753
650664 92.315 -52.175 56.723
650665 92.343 -52.195 56.690
650666 92.370 -52.217 56.655
650667 92.397 -52.237 56.618
650668 92.424 -52.258 56.580
650669 92.451 -52.278 56.540
650670 92.479 -52.299 56.500
650671 92.506 -52.320 56.456
650672 92.547 -52.348 56.407
650673 92.587 -52.377 56.356
650674 92.628 -52.405 56.306
650675 92.669 -52.434 56.257
650676 92.710 -52.463 56.208
650677 92.750 -52.492 56.160
650678 92.791 -52.518 56.112
650679 92.840 -52.524 56.064
650680 92.887 -52.530 56.016
650681 92.935 -52.536 55.967
650682 92.963 -52.554 55.917
650683 92.989 -52.571 55.868
650684 93.019 -52.585 55.820
650685 93.049 -52.600 55.771
650686 93.079 -52.615 55.722
650687 93.110 -52.633 55.673
650688 93.140 -52.651 55.624
650689 93.170 -52.670 55.578
650690 93.201 -52.689 55.533
650691 93.231 -52.707 55.489
650692 93.262 -52.726 55.445
650693 93.292 -52.745 55.403
650721 94.147 -53.269 54.851
650722 94.178 -53.287 54.853
650723 94.208 -53.306 54.856
650724 94.239 -53.325 54.860
650725 94.269 -53.343 54.866
650726 94.299 -53.362 54.872
650727 94.330 -53.381 54.878
650728 94.360 -53.399 54.885
650729 94.391 -53.418 54.894
650730 94.421 -53.437 54.901
650731 94.452 -53.456 54.909
650732 94.458 -53.467 54.918
650733 94.434 -53.468 54.927
650734 94.432 -53.475 54.937
650735 94.454 -53.490 54.946
650736 94.486 -53.509 54.955
650737 94.519 -53.529 54.958
650738 94.552 -53.549 54.960
650739 94.584 -53.568 54.960
650740 94.617 -53.587 54.960
650741 94.650 -53.607 54.959
650742 94.676 -53.622 54.956
650743 94.701 -53.636 54.955
650744 94.725 -53.650 54.955
650745 94.750 -53.664 54.956
650746 94.774 -53.678 54.958
650747 94.799 -53.691 54.960
650748 94.823 -53.705 54.963
650749 94.847 -53.719 54.969
650750 94.872 -53.732 54.965
650751 94.896 -53.746 54.956
650752 94.921 -53.759 54.946
650753 94.945 -53.774 54.940
650754 94.970 -53.787 54.934
650755 94.994 -53.801 54.929
650756 95.019 -53.815 54.908
650757 95.043 -53.828 54.889
650758 95.067 -53.842 54.873
650759 95.092 -53.856 54.855
650760 95.116 -53.869 54.832
650761 95.141 -53.883 54.806
650762 95.165 -53.897 54.772
650763 95.190 -53.911 54.736
650764 95.214 -53.925 54.695
650765 95.239 -53.938 54.648
650766 95.263 -53.952 54.596
650767 95.287 -53.966 54.544
650768 95.309 -53.985 54.496
650769 95.330 -54.004 54.459
650770 95.349 -54.019 54.422
650771 95.375 -54.026 54.382
650772 95.401 -54.033 54.340
650773 95.426 -54.040 54.298
650774 95.453 -54.047 54.258
650775 95.478 -54.054 54.222
650776 95.505 -54.060 54.188
650777 95.530 -54.067 54.154
650778 95.555 -54.077 54.120
650779 95.580 -54.088 54.089
650780 95.604 -54.098 54.059
650781 95.628 -54.108 54.033
650782 95.653 -54.118 54.011
650783 95.677 -54.129 53.991
650784 95.701 -54.138 53.975
650785 95.726 -54.149 53.961
650786 95.750 -54.159 53.947
650787 95.776 -54.169 53.929
650788 95.813 -54.176 53.908
650789 95.840 -54.193 53.875
650790 95.866 -54.211 53.840
650791 95.891 -54.229 53.808
650792 95.916 -54.248 53.781
650793 95.942 -54.266 53.758
650794 95.967 -54.284 53.735
650795 95.992 -54.302 53.710
650796 96.018 -54.320 53.680
650797 96.044 -54.339 53.645
650798 96.069 -54.357 53.603
650799 96.095 -54.375 53.555
650800 96.120 -54.393 53.505
650801 96.146 -54.411 53.451
650802 96.171 -54.430 53.402
650803 96.196 -54.448 53.355
650804 96.222 -54.466 53.311
650805 96.251 -54.481 53.266
650806 96.277 -54.494 53.220
650807 96.303 -54.506 53.170
650808 96.330 -54.519 53.120
650809 96.358 -54.531 53.072
650810 96.385 -54.543 53.025
650811 96.412 -54.556 52.978
650812 96.440 -54.568 52.931
650813 96.469 -54.581 52.884
650814 96.498 -54.593 52.839
650815 96.526 -54.606 52.794
650816 96.555 -54.618 52.751
650817 96.584 -54.627 52.712
650818 96.612 -54.635 52.672
650819 96.640 -54.643 52.632
650820 96.669 -54.651 52.593
650821 96.697 -54.659 52.555
650822 96.725 -54.668 52.518
650823 96.753 -54.675 52.479
650824 96.782 -54.684 52.439
650825 96.811 -54.691 52.396
650826 96.842 -54.699 52.349
650827 96.874 -54.707 52.298
650828 96.905 -54.715 52.245
650829 96.937 -54.724 52.189
650830 96.968 -54.732 52.133
650831 97.000 -54.740 52.078
650832 97.031 -54.749 52.023
650833 97.063 -54.757 51.966
650834 97.094 -54.765 51.909
650835 97.125 -54.774 51.848
650836 97.156 -54.781 51.786
650837 97.196 -54.792 51.721
650838 97.237 -54.803 51.651
650839 97.277 -54.814 51.576
650840 97.317 -54.825 51.501
650841 97.357 -54.834 51.428
650842 97.390 -54.830 51.356
650843 97.422 -54.829 51.279
650844 97.440 -54.844 51.199
650845 97.462 -54.848 51.121
650846 97.485 -54.847 51.046
650847 97.508 -54.847 50.974
650848 97.531 -54.846 50.906
650849 97.554 -54.845 50.840
650850 97.577 -54.844 50.773
650851 97.600 -54.843 50.708
650852 97.624 -54.842 50.644
650853 97.647 -54.841 50.581
650854 97.670 -54.840 50.521
650855 97.693 -54.840 50.460
650856 97.715 -54.834 50.400
650857 97.736 -54.828 50.341
650858 97.757 -54.822 50.284
650859 97.779 -54.815 50.228
650860 97.800 -54.809 50.173
650861 97.832 -54.810 50.118
650862 97.873 -54.819 50.065
650863 97.913 -54.828 50.013
651318 90.959 -51.384 56.879
651319 90.971 -51.399 56.885
651320 90.982 -51.415 56.891
651321 90.995 -51.431 56.898
651322 91.006 -51.447 56.908
651323 91.018 -51.462 56.918
651324 91.030 -51.478 56.928
651325 91.041 -51.494 56.938
651326 91.053 -51.510 56.949
651327 91.065 -51.526 56.962
651328 91.077 -51.542 56.976
651329 91.088 -51.557 56.992
651330 91.100 -51.573 57.012
651331 91.129 -51.591 57.031
651332 91.161 -51.617 57.048
651333 91.192 -51.643 57.061
651334 91.225 -51.669 57.071
651335 91.264 -51.689 57.078
651336 91.314 -51.700 57.081
651337 91.365 -51.710 57.079
651338 91.416 -51.722 57.082
651339 91.466 -51.732 57.081
651340 91.516 -51.744 57.077
651341 91.567 -51.754 57.072
651342 91.606 -51.771 57.067
651343 91.644 -51.788 57.061
651344 91.681 -51.804 57.060
651345 91.718 -51.821 57.058
651346 91.748 -51.837 57.055
651347 91.780 -51.852 57.049
651348 91.810 -51.867 57.038
651349 91.842 -51.883 57.021
651350 91.872 -51.898 57.007
651351 91.903 -51.914 56.992
651352 91.934 -51.929 56.978
651353 91.965 -51.945 56.965
651354 91.996 -51.960 56.951
651355 92.027 -51.976 56.935
651356 92.058 -51.992 56.918
651357 92.089 -52.007 56.902
651358 92.120 -52.023 56.884
651359 92.151 -52.039 56.867
651360 92.181 -52.054 56.849
651361 92.213 -52.069 56.828
651362 92.243 -52.085 56.805
651363 92.274 -52.100 56.780
651364 92.305 -52.116 56.753
651365 92.336 -52.131 56.723
651366 92.367 -52.147 56.689
651367 92.398 -52.163 56.654
651368 92.425 -52.183 56.617
651369 92.452 -52.205 56.579
651370 92.480 -52.225 56.540
651371 92.507 -52.246 56.499
651372 92.534 -52.266 56.454
651373 92.575 -52.295 56.407
651374 92.615 -52.324 56.357
651375 92.656 -52.352 56.309
651376 92.697 -52.381 56.261
651377 92.737 -52.409 56.213
651378 92.778 -52.438 56.166
651379 92.818 -52.467 56.120
651380 92.862 -52.487 56.074
651381 92.910 -52.492 56.029
651382 92.948 -52.504 55.981
651383 92.975 -52.521 55.931
651384 93.005 -52.535 55.883
651385 93.035 -52.551 55.834
651386 93.066 -52.570 55.785
651387 93.096 -52.589 55.735
651388 93.127 -52.607 55.686
651389 93.157 -52.626 55.636
651390 93.188 -52.645 55.589
651391 93.218 -52.663 55.544
651392 93.249 -52.682 55.499
651393 93.279 -52.701 55.455
651394 93.310 -52.719 55.413
651422 94.164 -53.243 54.853
651423 94.195 -53.262 54.853
651424 94.225 -53.281 54.856
651425 94.256 -53.300 54.859
651426 94.286 -53.318 54.864
651427 94.317 -53.337 54.870
651428 94.347 -53.356 54.876
651429 94.378 -53.374 54.882
651430 94.408 -53.393 54.890
651431 94.439 -53.412 54.897
651432 94.469 -53.430 54.906
651433 94.484 -53.444 54.916
651434 94.460 -53.446 54.926
651435 94.465 -53.455 54.937
651436 94.481 -53.466 54.944
651437 94.499 -53.478 54.951
651438 94.532 -53.498 54.955
651439 94.564 -53.518 54.955
651440 94.597 -53.537 54.955
651441 94.630 -53.557 54.953
651442 94.663 -53.577 54.954
651443 94.693 -53.595 54.951
651444 94.718 -53.609 54.950
651445 94.742 -53.622 54.950
651446 94.767 -53.636 54.953
651447 94.791 -53.650 54.956
651448 94.816 -53.663 54.960
651449 94.840 -53.678 54.966
651450 94.864 -53.691 54.969
651451 94.889 -53.705 54.963
651452 94.913 -53.719 54.950
651453 94.938 -53.732 54.937
651454 94.962 -53.746 54.926
651455 94.987 -53.759 54.917
651456 95.012 -53.773 54.905
651457 95.036 -53.787 54.887
651458 95.060 -53.801 54.874
651459 95.085 -53.815 54.864
651460 95.109 -53.828 54.851
651461 95.133 -53.842 54.832
651462 95.158 -53.856 54.807
651463 95.182 -53.869 54.777
651464 95.207 -53.883 54.741
651465 95.232 -53.897 54.701
651466 95.256 -53.911 54.655
651467 95.277 -53.930 54.604
651468 95.298 -53.948 54.553
651469 95.320 -53.967 54.505
651470 95.340 -53.985 54.466
651471 95.359 -54.000 54.432
651472 95.383 -54.009 54.394
651473 95.409 -54.016 54.353
651474 95.435 -54.022 54.311
651475 95.461 -54.029 54.272
651476 95.486 -54.036 54.236
651477 95.513 -54.044 54.203
651478 95.538 -54.051 54.168
651479 95.564 -54.057 54.134
651480 95.590 -54.064 54.103
651481 95.616 -54.071 54.073
651482 95.642 -54.078 54.046
651483 95.668 -54.085 54.023
651484 95.693 -54.094 54.000
651485 95.718 -54.104 53.982
651486 95.742 -54.114 53.969
651487 95.766 -54.124 53.957
651488 95.791 -54.135 53.941
651489 95.823 -54.145 53.922
651490 95.848 -54.163 53.889
651491 95.874 -54.181 53.854
651492 95.899 -54.199 53.818
651493 95.925 -54.217 53.786
651494 95.951 -54.236 53.758
651495 95.976 -54.254 53.735
651496 96.002 -54.272 53.710
651497 96.027 -54.290 53.680
651498 96.052 -54.308 53.645
651499 96.078 -54.327 53.603
651500 96.103 -54.345 53.557
651501 96.129 -54.363 53.508
651502 96.154 -54.381 53.456
651503 96.180 -54.399 53.410
651504 96.205 -54.417 53.366
651505 96.231 -54.436 53.324
651506 96.257 -54.453 53.281
651507 96.284 -54.465 53.235
651508 96.311 -54.477 53.187
651509 96.338 -54.490 53.135
651510 96.366 -54.502 53.087
651511 96.393 -54.515 53.040
651512 96.421 -54.527 52.994
651513 96.449 -54.540 52.948
651514 96.477 -54.552 52.901
651515 96.506 -54.565 52.858
651516 96.535 -54.577 52.815
651517 96.564 -54.590 52.774
651518 96.593 -54.601 52.738
651519 96.621 -54.609 52.702
651520 96.649 -54.618 52.663
651521 96.678 -54.626 52.625
651522 96.706 -54.634 52.588
651523 96.734 -54.642 52.550
651524 96.762 -54.650 52.511
651525 96.791 -54.658 52.470
651526 96.820 -54.664 52.426
651527 96.851 -54.670 52.378
651528 96.882 -54.678 52.329
651529 96.913 -54.687 52.277
651530 96.945 -54.695 52.223
651531 96.976 -54.703 52.167
651532 97.008 -54.712 52.111
651533 97.039 -54.720 52.055
651534 97.071 -54.728 51.998
651535 97.102 -54.736 51.940
651536 97.134 -54.744 51.878
651537 97.164 -54.752 51.817
651538 97.200 -54.761 51.753
651539 97.240 -54.772 51.685
651540 97.281 -54.783 51.615
651541 97.321 -54.794 51.543
651542 97.360 -54.802 51.471
651543 97.394 -54.798 51.400
651544 97.428 -54.793 51.325
651545 97.455 -54.798 51.248
651546 97.473 -54.813 51.170
651547 97.494 -54.822 51.095
651548 97.516 -54.822 51.023
651549 97.539 -54.821 50.954
651550 97.563 -54.820 50.887
651551 97.586 -54.818 50.821
651552 97.609 -54.818 50.756
651553 97.632 -54.817 50.691
651554 97.655 -54.816 50.627
651555 97.678 -54.815 50.566
651556 97.701 -54.814 50.504
651557 97.723 -54.808 50.444
651558 97.744 -54.801 50.383
651559 97.766 -54.794 50.324
651560 97.787 -54.788 50.265
651561 97.808 -54.781 50.209
651562 97.846 -54.787 50.154
651563 97.886 -54.796 50.098
651564 97.923 -54.804 50.045
652019 90.964 -51.356 56.870
652020 90.976 -51.372 56.875
652021 90.988 -51.387 56.881
652022 91.000 -51.403 56.888
652023 91.011 -51.419 56.896
652024 91.023 -51.435 56.906
652025 91.035 -51.451 56.916
652026 91.046 -51.466 56.926
652027 91.058 -51.482 56.936
652028 91.070 -51.498 56.948
652029 91.082 -51.514 56.961
652030 91.093 -51.530 56.976
652031 91.105 -51.546 56.993
652032 91.132 -51.564 57.009
652033 91.164 -51.590 57.023
652034 91.196 -51.616 57.035
652035 91.236 -51.635 57.044
652036 91.287 -51.646 57.052
652037 91.337 -51.657 57.057
652038 91.387 -51.668 57.060
652039 91.438 -51.679 57.065
652040 91.489 -51.690 57.064
652041 91.539 -51.701 57.060
652042 91.587 -51.713 57.056
652043 91.625 -51.730 57.052
652044 91.662 -51.747 57.046
652045 91.700 -51.764 57.046
652046 91.736 -51.781 57.044
652047 91.768 -51.796 57.042
652048 91.798 -51.811 57.037
652049 91.829 -51.827 57.027
652050 91.860 -51.842 57.010
652051 91.891 -51.858 56.995
652052 91.922 -51.873 56.980
652053 91.953 -51.889 56.966
652054 91.984 -51.904 56.954
652055 92.015 -51.920 56.941
652056 92.046 -51.935 56.926
652057 92.077 -51.951 56.910
652058 92.107 -51.967 56.895
652059 92.139 -51.982 56.880
652060 92.169 -51.998 56.865
652061 92.201 -52.014 56.848
652062 92.231 -52.029 56.828
652063 92.262 -52.044 56.805
652064 92.293 -52.060 56.780
652065 92.324 -52.075 56.751
652066 92.355 -52.091 56.721
652067 92.386 -52.106 56.688
652068 92.417 -52.122 56.653
652069 92.448 -52.137 56.616
652070 92.479 -52.153 56.577
652071 92.507 -52.172 56.540
652072 92.534 -52.192 56.497
652073 92.562 -52.213 56.452
652074 92.602 -52.242 56.405
652075 92.643 -52.271 56.357
652076 92.684 -52.299 56.311
652077 92.724 -52.327 56.264
652078 92.765 -52.356 56.218
652079 92.806 -52.385 56.173
652080 92.846 -52.414 56.129
652081 92.887 -52.443 56.087
652082 92.933 -52.455 56.042
652083 92.961 -52.471 55.995
652084 92.992 -52.489 55.946
652085 93.022 -52.507 55.898
652086 93.053 -52.526 55.850
652087 93.083 -52.545 55.800
652088 93.114 -52.563 55.749
652089 93.144 -52.582 55.699
652090 93.174 -52.601 55.650
652091 93.205 -52.620 55.601
652092 93.235 -52.638 55.556
652093 93.266 -52.657 55.510
652094 93.296 -52.676 55.465
652095 93.327 -52.694 55.423
652123 94.182 -53.218 54.854
652124 94.212 -53.237 54.854
652125 94.243 -53.256 54.856
652126 94.273 -53.274 54.858
652127 94.303 -53.293 54.863
652128 94.334 -53.312 54.868
652129 94.364 -53.330 54.874
652130 94.395 -53.349 54.879
652131 94.425 -53.368 54.886
652132 94.456 -53.386 54.893
652133 94.487 -53.405 54.902
652134 94.510 -53.422 54.911
652135 94.486 -53.424 54.921
652136 94.498 -53.433 54.930
652137 94.513 -53.445 54.936
652138 94.530 -53.456 54.941
652139 94.546 -53.468 54.945
652140 94.576 -53.487 54.946
652141 94.609 -53.506 54.945
652142 94.642 -53.526 54.944
652143 94.675 -53.546 54.944
652144 94.707 -53.565 54.943
652145 94.735 -53.581 54.943
652146 94.759 -53.595 54.944
652147 94.784 -53.609 54.947
652148 94.808 -53.622 54.951
652149 94.833 -53.636 54.956
652150 94.857 -53.650 54.965
652151 94.882 -53.663 54.962
652152 94.906 -53.678 54.952
652153 94.930 -53.691 54.938
652154 94.955 -53.705 54.922
652155 94.979 -53.719 54.908
652156 95.004 -53.732 54.896
652157 95.028 -53.746 54.881
652158 95.053 -53.759 54.869
652159 95.077 -53.773 54.862
652160 95.102 -53.787 54.855
652161 95.126 -53.801 54.845
652162 95.150 -53.815 54.828
652163 95.175 -53.828 54.803
652164 95.199 -53.842 54.775
652165 95.224 -53.856 54.740
652166 95.245 -53.875 54.701
652167 95.266 -53.893 54.658
652168 95.288 -53.912 54.609
652169 95.309 -53.930 54.559
652170 95.331 -53.949 54.513
652171 95.350 -53.965 54.476
652172 95.369 -53.980 54.442
652173 95.391 -53.992 54.405
652174 95.417 -53.999 54.364
652175 95.443 -54.006 54.323
652176 95.469 -54.013 54.284
652177 95.494 -54.019 54.249
652178 95.521 -54.026 54.217
652179 95.546 -54.033 54.182
652180 95.572 -54.040 54.147
652181 95.598 -54.047 54.116
652182 95.624 -54.054 54.088
652183 95.651 -54.060 54.060
652184 95.676 -54.067 54.034
652185 95.702 -54.074 54.007
652186 95.728 -54.081 53.986
652187 95.754 -54.088 53.972
652188 95.780 -54.095 53.961
652189 95.806 -54.102 53.948
652190 95.832 -54.115 53.930
652191 95.857 -54.133 53.900
652192 95.883 -54.151 53.866
652193 95.908 -54.170 53.829
652194 95.934 -54.188 53.792
652195 95.959 -54.206 53.759
652196 95.984 -54.225 53.732
652197 96.010 -54.243 53.705
652198 96.035 -54.261 53.674
652199 96.061 -54.279 53.638
652200 96.086 -54.298 53.596
652201 96.112 -54.316 53.551
652202 96.137 -54.334 53.505
652203 96.163 -54.353 53.457
652204 96.188 -54.371 53.413
652205 96.214 -54.389 53.372
652206 96.239 -54.408 53.332
652207 96.265 -54.425 53.290
652208 96.292 -54.436 53.245
652209 96.320 -54.449 53.197
652210 96.347 -54.461 53.148
652211 96.374 -54.474 53.101
652212 96.402 -54.486 53.055
652213 96.429 -54.499 53.009
652214 96.457 -54.511 52.963
652215 96.486 -54.524 52.918
652216 96.515 -54.536 52.874
652217 96.544 -54.549 52.833
652218 96.573 -54.562 52.797
652219 96.601 -54.574 52.765
652220 96.630 -54.583 52.730
652221 96.658 -54.592 52.693
652222 96.687 -54.600 52.657
652223 96.715 -54.609 52.621
652224 96.743 -54.616 52.584
652225 96.771 -54.624 52.543
652226 96.800 -54.632 52.500
652227 96.830 -54.637 52.453
652228 96.860 -54.643 52.405
652229 96.891 -54.649 52.356
652230 96.922 -54.658 52.305
652231 96.954 -54.666 52.252
652232 96.985 -54.674 52.197
652233 97.016 -54.682 52.140
652234 97.048 -54.690 52.082
652235 97.079 -54.699 52.024
652236 97.110 -54.707 51.965
652237 97.142 -54.715 51.904
652238 97.173 -54.723 51.842
652239 97.204 -54.730 51.779
652240 97.244 -54.741 51.715
652241 97.285 -54.753 51.650
652242 97.325 -54.763 51.582
652243 97.364 -54.770 51.512
652244 97.397 -54.766 51.441
652245 97.431 -54.762 51.368
652246 97.465 -54.758 51.293
652247 97.488 -54.767 51.218
652248 97.506 -54.782 51.144
652249 97.525 -54.797 51.073
652250 97.548 -54.796 51.003
652251 97.571 -54.795 50.936
652252 97.594 -54.793 50.869
652253 97.617 -54.793 50.804
652254 97.641 -54.792 50.738
652255 97.664 -54.791 50.673
652256 97.687 -54.790 50.611
652257 97.709 -54.787 50.550
652258 97.730 -54.781 50.488
652259 97.752 -54.774 50.426
652260 97.773 -54.768 50.365
652261 97.795 -54.761 50.305
652262 97.818 -54.756 50.247
652263 97.859 -54.765 50.191
652264 97.899 -54.773 50.134
652265 97.933 -54.781 50.078
652266 97.968 -54.788 50.025
652720 90.969 -51.328 56.861
652721 90.981 -51.344 56.866
652722 90.993 -51.360 56.871
652723 91.005 -51.376 56.877
652724 91.016 -51.391 56.885
652725 91.028 -51.407 56.894
652726 91.040 -51.423 56.904
652727 91.051 -51.439 56.913
652728 91.063 -51.455 56.923
652729 91.075 -51.471 56.933
652730 91.087 -51.486 56.945
652731 91.098 -51.502 56.958
652732 91.110 -51.518 56.973
652733 91.135 -51.537 56.987
652734 91.167 -51.563 56.998
652735 91.208 -51.581 57.009
652736 91.258 -51.593 57.018
652737 91.309 -51.603 57.026
652738 91.360 -51.615 57.033
652739 91.410 -51.625 57.040
652740 91.461 -51.636 57.047
652741 91.511 -51.647 57.047
652742 91.562 -51.658 57.043
652743 91.605 -51.672 57.039
652744 91.643 -51.689 57.035
652745 91.680 -51.707 57.030
652746 91.718 -51.723 57.028
652747 91.755 -51.740 57.027
652748 91.786 -51.756 57.025
652749 91.817 -51.771 57.020
652750 91.848 -51.786 57.013
652751 91.879 -51.802 56.997
652752 91.910 -51.817 56.982
652753 91.941 -51.833 56.967
652754 91.972 -51.848 56.954
652755 92.003 -51.864 56.943
652756 92.034 -51.879 56.932
652757 92.065 -51.895 56.918
652758 92.095 -51.910 56.903
652759 92.127 -51.926 56.889
652760 92.157 -51.942 56.875
652761 92.189 -51.957 56.861
652762 92.219 -51.973 56.846
652763 92.250 -51.988 56.827
652764 92.281 -52.004 56.803
652765 92.312 -52.019 56.777
652766 92.343 -52.035 56.748
652767 92.374 -52.050 56.718
652768 92.405 -52.066 56.685
652769 92.436 -52.081 56.650
652770 92.466 -52.097 56.614
652771 92.498 -52.112 56.577
652772 92.528 -52.128 56.537
652773 92.560 -52.143 56.495
652774 92.590 -52.160 56.450
652775 92.630 -52.188 56.403
652776 92.670 -52.217 56.357
652777 92.710 -52.248 56.313
652778 92.750 -52.278 56.267
652779 92.790 -52.308 56.223
652780 92.830 -52.339 56.179
652781 92.869 -52.368 56.138
652782 92.909 -52.399 56.098
652783 92.947 -52.427 56.056
652784 92.978 -52.445 56.010
652785 93.008 -52.464 55.962
652786 93.039 -52.483 55.915
652787 93.069 -52.501 55.866
652788 93.100 -52.520 55.816
652789 93.131 -52.538 55.765
652790 93.161 -52.557 55.715
652791 93.192 -52.576 55.665
652792 93.222 -52.594 55.616
652793 93.253 -52.613 55.568
652794 93.283 -52.632 55.522
652795 93.314 -52.650 55.478
652796 93.344 -52.669 55.435
652824 94.199 -53.193 54.856
652825 94.229 -53.212 54.856
652826 94.260 -53.230 54.856
652827 94.290 -53.249 54.857
652828 94.321 -53.268 54.861
652829 94.351 -53.286 54.866
652830 94.382 -53.305 54.871
652831 94.412 -53.324 54.876
652832 94.443 -53.342 54.882
652833 94.473 -53.361 54.889
652834 94.504 -53.380 54.897
652835 94.535 -53.399 54.906
652836 94.514 -53.402 54.914
652837 94.531 -53.413 54.922
652838 94.546 -53.424 54.926
652839 94.562 -53.436 54.931
652840 94.578 -53.447 54.933
652841 94.594 -53.458 54.934
652842 94.622 -53.475 54.934
652843 94.654 -53.495 54.934
652844 94.687 -53.515 54.933
652845 94.720 -53.534 54.932
652846 94.752 -53.553 54.931
652847 94.776 -53.567 54.932
652848 94.801 -53.581 54.934
652849 94.825 -53.595 54.939
652850 94.850 -53.609 54.944
652851 94.875 -53.622 54.948
652852 94.899 -53.636 54.944
652853 94.923 -53.650 54.933
652854 94.948 -53.663 54.918
652855 94.972 -53.677 54.901
652856 94.996 -53.691 54.885
652857 95.021 -53.705 54.870
652858 95.045 -53.719 54.859
652859 95.070 -53.732 54.853
652860 95.095 -53.746 54.851
652861 95.119 -53.759 54.847
652862 95.143 -53.773 54.837
652863 95.168 -53.787 54.821
652864 95.192 -53.801 54.795
652865 95.213 -53.819 54.768
652866 95.235 -53.838 54.735
652867 95.256 -53.856 54.698
652868 95.277 -53.875 54.658
652869 95.299 -53.894 54.612
652870 95.320 -53.912 54.563
652871 95.341 -53.930 54.520
652872 95.360 -53.946 54.486
652873 95.379 -53.961 54.454
652874 95.399 -53.975 54.418
652875 95.425 -53.982 54.377
652876 95.451 -53.988 54.336
652877 95.477 -53.995 54.297
652878 95.502 -54.002 54.262
652879 95.529 -54.009 54.229
652880 95.555 -54.016 54.193
652881 95.581 -54.022 54.158
652882 95.607 -54.029 54.125
652883 95.632 -54.036 54.095
652884 95.659 -54.043 54.067
652885 95.684 -54.051 54.039
652886 95.710 -54.057 54.011
652887 95.736 -54.064 53.987
652888 95.761 -54.071 53.969
652889 95.787 -54.078 53.956
652890 95.812 -54.085 53.944
652891 95.838 -54.094 53.927
652892 95.866 -54.106 53.905
652893 95.891 -54.124 53.874
652894 95.916 -54.142 53.838
652895 95.942 -54.160 53.799
652896 95.967 -54.179 53.761
652897 95.993 -54.197 53.728
652898 96.018 -54.215 53.695
652899 96.044 -54.233 53.662
652900 96.069 -54.251 53.625
652901 96.095 -54.270 53.584
652902 96.120 -54.289 53.541
652903 96.146 -54.307 53.499
652904 96.171 -54.325 53.456
652905 96.196 -54.343 53.414
652906 96.222 -54.361 53.375
652907 96.247 -54.380 53.336
652908 96.273 -54.398 53.295
652909 96.300 -54.409 53.252
652910 96.328 -54.421 53.205
652911 96.355 -54.433 53.158
652912 96.383 -54.445 53.113
652913 96.410 -54.458 53.068
652914 96.438 -54.470 53.023
652915 96.466 -54.483 52.979
652916 96.495 -54.495 52.935
652917 96.523 -54.508 52.892
652918 96.552 -54.520 52.853
652919 96.581 -54.533 52.820
652920 96.610 -54.545 52.789
652921 96.639 -54.558 52.756
652922 96.667 -54.567 52.721
652923 96.695 -54.574 52.686
652924 96.724 -54.583 52.650
652925 96.752 -54.591 52.613
652926 96.781 -54.599 52.571
652927 96.810 -54.605 52.526
652928 96.840 -54.610 52.478
652929 96.870 -54.615 52.428
652930 96.899 -54.621 52.377
652931 96.931 -54.628 52.326
652932 96.962 -54.637 52.273
652933 96.994 -54.645 52.219
652934 97.025 -54.653 52.163
652935 97.056 -54.662 52.105
652936 97.088 -54.670 52.048
652937 97.119 -54.678 51.990
652938 97.150 -54.687 51.929
652939 97.181 -54.693 51.868
652940 97.212 -54.701 51.807
652941 97.248 -54.710 51.746
652942 97.288 -54.721 51.683
652943 97.328 -54.732 51.617
652944 97.367 -54.738 51.548
652945 97.401 -54.734 51.478
652946 97.435 -54.730 51.407
652947 97.468 -54.726 51.335
652948 97.503 -54.721 51.263
652949 97.521 -54.736 51.193
652950 97.539 -54.751 51.122
652951 97.558 -54.766 51.052
652952 97.580 -54.770 50.984
652953 97.603 -54.769 50.918
652954 97.626 -54.768 50.852
652955 97.649 -54.767 50.787
652956 97.672 -54.766 50.721
652957 97.695 -54.765 50.657
652958 97.717 -54.760 50.596
652959 97.738 -54.753 50.533
652960 97.759 -54.746 50.471
652961 97.781 -54.740 50.409
652962 97.803 -54.734 50.347
652963 97.831 -54.734 50.288
652964 97.872 -54.742 50.229
652965 97.909 -54.750 50.171
652966 97.944 -54.757 50.114
652967 97.978 -54.765 50.058
652968 98.012 -54.771 50.004
653422 90.986 -51.316 56.856
653423 90.998 -51.332 56.861
653424 91.010 -51.348 56.866
653425 91.021 -51.364 56.873
653426 91.033 -51.380 56.882
653427 91.045 -51.395 56.891
653428 91.057 -51.411 56.900
653429 91.069 -51.427 56.909
653430 91.080 -51.443 56.919
653431 91.092 -51.459 56.930
653432 91.104 -51.475 56.942
653433 91.115 -51.490 56.956
653434 91.139 -51.510 56.967
653435 91.180 -51.528 56.978
653436 91.231 -51.539 56.986
653437 91.281 -51.550 56.994
653438 91.332 -51.561 57.003
653439 91.382 -51.572 57.013
653440 91.433 -51.583 57.022
653441 91.484 -51.594 57.027
653442 91.534 -51.605 57.027
653443 91.585 -51.616 57.024
653444 91.624 -51.632 57.020
653445 91.661 -51.649 57.017
653446 91.699 -51.666 57.012
653447 91.736 -51.682 57.010
653448 91.774 -51.700 57.009
653449 91.805 -51.715 57.007
653450 91.836 -51.730 57.002
653451 91.867 -51.746 56.996
653452 91.898 -51.761 56.981
653453 91.929 -51.777 56.967
653454 91.960 -51.792 56.954
653455 91.991 -51.808 56.943
653456 92.021 -51.823 56.933
653457 92.053 -51.839 56.923
653458 92.083 -51.854 56.910
653459 92.115 -51.870 56.897
653460 92.145 -51.885 56.883
653461 92.176 -51.901 56.871
653462 92.207 -51.917 56.858
653463 92.238 -51.932 56.842
653464 92.269 -51.948 56.824
653465 92.300 -51.963 56.800
653466 92.331 -51.979 56.774
653467 92.362 -51.994 56.745
653468 92.393 -52.010 56.715
653469 92.424 -52.025 56.683
653470 92.454 -52.041 56.648
653471 92.486 -52.056 56.613
653472 92.518 -52.074 56.575
653473 92.553 -52.093 56.535
653474 92.583 -52.111 56.494
653475 92.611 -52.131 56.449
653476 92.643 -52.160 56.403
653477 92.683 -52.190 56.359
653478 92.722 -52.221 56.316
653479 92.762 -52.251 56.271
653480 92.802 -52.280 56.227
653481 92.842 -52.311 56.185
653482 92.881 -52.341 56.146
653483 92.921 -52.371 56.108
653484 92.961 -52.402 56.067
653485 92.993 -52.421 56.024
653486 93.023 -52.440 55.978
653487 93.054 -52.459 55.932
653488 93.085 -52.478 55.882
653489 93.116 -52.496 55.831
653490 93.146 -52.515 55.782
653491 93.177 -52.534 55.730
653492 93.207 -52.552 55.679
653493 93.238 -52.571 55.630
653494 93.269 -52.589 55.582
653495 93.299 -52.608 55.535
653496 93.330 -52.626 55.490
653497 93.360 -52.645 55.447
653498 93.391 -52.663 55.404
653525 94.216 -53.168 54.857
653526 94.247 -53.186 54.857
653527 94.277 -53.205 54.857
653528 94.308 -53.224 54.857
653529 94.338 -53.242 54.859
653530 94.368 -53.261 54.863
653531 94.399 -53.280 54.868
653532 94.429 -53.299 54.872
653533 94.460 -53.317 54.878
653534 94.491 -53.336 54.885
653535 94.521 -53.355 54.893
653536 94.552 -53.374 54.900
653537 94.547 -53.381 54.907
653538 94.563 -53.392 54.913
653539 94.579 -53.403 54.915
653540 94.595 -53.415 54.919
653541 94.611 -53.426 54.921
653542 94.627 -53.437 54.922
653543 94.643 -53.449 54.922
653544 94.667 -53.465 54.922
653545 94.700 -53.484 54.920
653546 94.733 -53.503 54.918
653547 94.765 -53.523 54.917
653548 94.793 -53.540 54.918
653549 94.818 -53.553 54.920
653550 94.842 -53.567 54.922
653551 94.867 -53.581 54.927
653552 94.892 -53.595 54.926
653553 94.916 -53.609 54.919
653554 94.941 -53.622 54.908
653555 94.965 -53.636 54.894
653556 94.989 -53.650 54.879
653557 95.014 -53.663 54.863
653558 95.038 -53.677 54.850
653559 95.062 -53.690 54.844
653560 95.087 -53.705 54.841
653561 95.112 -53.719 54.838
653562 95.136 -53.732 54.833
653563 95.161 -53.746 54.823
653564 95.182 -53.764 54.805
653565 95.203 -53.783 54.782
653566 95.225 -53.801 54.757
653567 95.246 -53.820 54.727
653568 95.267 -53.838 54.691
653569 95.288 -53.857 54.651
653570 95.310 -53.875 54.609
653571 95.331 -53.894 54.564
653572 95.351 -53.911 54.525
653573 95.370 -53.926 54.496
653574 95.389 -53.941 54.467
653575 95.407 -53.957 54.432
653576 95.433 -53.964 54.392
653577 95.459 -53.971 54.351
653578 95.485 -53.978 54.311
653579 95.511 -53.985 54.274
653580 95.537 -53.991 54.239
653581 95.563 -53.999 54.202
653582 95.589 -54.006 54.165
653583 95.615 -54.013 54.130
653584 95.640 -54.019 54.099
653585 95.666 -54.026 54.070
653586 95.691 -54.033 54.041
653587 95.716 -54.041 54.011
653588 95.741 -54.048 53.984
653589 95.767 -54.054 53.962
653590 95.792 -54.061 53.946
653591 95.817 -54.068 53.933
653592 95.842 -54.076 53.917
653593 95.870 -54.088 53.896
653594 95.898 -54.100 53.869
653595 95.925 -54.114 53.836
653596 95.950 -54.132 53.799
653597 95.976 -54.151 53.760
653598 96.001 -54.170 53.723
653599 96.027 -54.188 53.686
653600 96.052 -54.206 53.649
653601 96.078 -54.224 53.610
653602 96.103 -54.242 53.570
653603 96.128 -54.261 53.530
653604 96.154 -54.279 53.491
653605 96.179 -54.297 53.450
653606 96.205 -54.316 53.411
653607 96.230 -54.334 53.374
653608 96.256 -54.352 53.336
653609 96.281 -54.370 53.297
653610 96.309 -54.382 53.256
653611 96.336 -54.393 53.210
653612 96.364 -54.405 53.166
653613 96.391 -54.417 53.123
653614 96.418 -54.429 53.080
653615 96.446 -54.441 53.037
653616 96.474 -54.454 52.995
653617 96.503 -54.466 52.952
653618 96.532 -54.479 52.912
653619 96.561 -54.491 52.873
653620 96.590 -54.504 52.841
653621 96.619 -54.517 52.810
653622 96.648 -54.529 52.779
653623 96.676 -54.541 52.747
653624 96.704 -54.549 52.713
653625 96.733 -54.557 52.677
653626 96.761 -54.565 52.639
653627 96.790 -54.573 52.596
653628 96.820 -54.578 52.550
653629 96.850 -54.583 52.501
653630 96.879 -54.589 52.450
653631 96.909 -54.594 52.399
653632 96.939 -54.599 52.348
653633 96.970 -54.608 52.295
653634 97.002 -54.616 52.242
653635 97.033 -54.624 52.186
653636 97.065 -54.633 52.129
653637 97.096 -54.641 52.073
653638 97.127 -54.649 52.014
653639 97.159 -54.657 51.955
653640 97.189 -54.665 51.896
653641 97.221 -54.671 51.836
653642 97.251 -54.679 51.776
653643 97.292 -54.690 51.715
653644 97.332 -54.701 51.651
653645 97.370 -54.706 51.582
653646 97.404 -54.702 51.514
653647 97.438 -54.698 51.446
653648 97.472 -54.694 51.378
653649 97.506 -54.690 51.311
653650 97.536 -54.690 51.244
653651 97.555 -54.705 51.174
653652 97.573 -54.720 51.104
653653 97.591 -54.735 51.035
653654 97.611 -54.744 50.968
653655 97.635 -54.743 50.902
653656 97.658 -54.742 50.837
653657 97.680 -54.741 50.772
653658 97.703 -54.739 50.707
653659 97.725 -54.733 50.646
653660 97.746 -54.726 50.581
653661 97.768 -54.720 50.519
653662 97.789 -54.713 50.456
653663 97.810 -54.707 50.393
653664 97.845 -54.711 50.332
653665 97.885 -54.719 50.271
653666 97.919 -54.726 50.212
653667 97.954 -54.734 50.153
653668 97.988 -54.740 50.095
653669 98.014 -54.740 50.040
653670 98.041 -54.741 49.985
654123 90.991 -51.289 56.846
654124 91.003 -51.305 56.850
654125 91.015 -51.320 56.856
654126 91.026 -51.336 56.862
654127 91.038 -51.352 56.870
654128 91.050 -51.368 56.879
654129 91.062 -51.384 56.887
654130 91.074 -51.399 56.895
654131 91.085 -51.415 56.905
654132 91.097 -51.431 56.916
654133 91.109 -51.446 56.927
654134 91.120 -51.462 56.939
654135 91.152 -51.475 56.949
654136 91.203 -51.485 56.958
654137 91.253 -51.497 56.966
654138 91.304 -51.507 56.974
654139 91.355 -51.519 56.982
654140 91.405 -51.529 56.994
654141 91.455 -51.540 57.002
654142 91.506 -51.551 57.006
654143 91.557 -51.562 57.005
654144 91.604 -51.575 57.003
654145 91.642 -51.591 56.999
654146 91.679 -51.608 56.998
654147 91.717 -51.625 56.994
654148 91.755 -51.642 56.991
654149 91.792 -51.659 56.990
654150 91.824 -51.674 56.988
654151 91.855 -51.690 56.984
654152 91.886 -51.705 56.977
654153 91.917 -51.721 56.965
654154 91.948 -51.736 56.953
654155 91.979 -51.752 56.942
654156 92.009 -51.767 56.933
654157 92.040 -51.783 56.924
654158 92.071 -51.798 56.915
654159 92.102 -51.814 56.903
654160 92.133 -51.829 56.890
654161 92.164 -51.845 56.878
654162 92.195 -51.860 56.866
654163 92.226 -51.876 56.852
654164 92.257 -51.891 56.837
654165 92.288 -51.907 56.818
654166 92.319 -51.923 56.796
654167 92.350 -51.938 56.769
654168 92.381 -51.954 56.741
654169 92.413 -51.971 56.712
654170 92.447 -51.990 56.681
654171 92.481 -52.010 56.647
654172 92.515 -52.029 56.610
654173 92.549 -52.048 56.574
654174 92.575 -52.065 56.534
654175 92.601 -52.083 56.493
654176 92.632 -52.105 56.449
654177 92.662 -52.133 56.404
654178 92.695 -52.163 56.361
654179 92.734 -52.193 56.318
654180 92.774 -52.223 56.274
654181 92.814 -52.253 56.232
654182 92.854 -52.283 56.191
654183 92.894 -52.314 56.152
654184 92.933 -52.344 56.114
654185 92.973 -52.374 56.075
654186 93.008 -52.398 56.035
654187 93.039 -52.416 55.993
654188 93.069 -52.435 55.947
654189 93.100 -52.454 55.899
654190 93.131 -52.472 55.848
654191 93.161 -52.491 55.797
654192 93.192 -52.509 55.746
654193 93.222 -52.528 55.694
654194 93.253 -52.546 55.645
654195 93.284 -52.565 55.597
654196 93.314 -52.584 55.549
654197 93.345 -52.603 55.503
654198 93.376 -52.622 55.460
654199 93.406 -52.640 55.416
654226 94.234 -53.142 54.859
654227 94.264 -53.161 54.858
654228 94.294 -53.180 54.857
654229 94.325 -53.198 54.857
654230 94.355 -53.217 54.858
654231 94.386 -53.236 54.861
654232 94.416 -53.255 54.865
654233 94.447 -53.273 54.869
654234 94.477 -53.292 54.874
654235 94.508 -53.311 54.881
654236 94.539 -53.330 54.889
654237 94.569 -53.349 54.894
654238 94.580 -53.360 54.897
654239 94.596 -53.371 54.901
654240 94.612 -53.383 54.903
654241 94.628 -53.394 54.905
654242 94.643 -53.405 54.907
654243 94.660 -53.417 54.908
654244 94.676 -53.428 54.908
654245 94.691 -53.440 54.907
654246 94.712 -53.453 54.906
654247 94.745 -53.473 54.903
654248 94.778 -53.493 54.901
654249 94.811 -53.512 54.902
654250 94.835 -53.526 54.903
654251 94.859 -53.540 54.903
654252 94.884 -53.553 54.906
654253 94.909 -53.567 54.901
654254 94.933 -53.581 54.893
654255 94.958 -53.595 54.883
654256 94.982 -53.609 54.871
654257 95.006 -53.622 54.859
654258 95.031 -53.636 54.847
654259 95.055 -53.650 54.836
654260 95.079 -53.663 54.832
654261 95.104 -53.677 54.830
654262 95.129 -53.690 54.826
654263 95.150 -53.709 54.819
654264 95.172 -53.728 54.809
654265 95.193 -53.746 54.791
654266 95.214 -53.765 54.769
654267 95.236 -53.783 54.743
654268 95.257 -53.802 54.712
654269 95.278 -53.820 54.678
654270 95.300 -53.839 54.641
654271 95.321 -53.857 54.603
654272 95.342 -53.876 54.563
654273 95.361 -53.891 54.531
654274 95.380 -53.907 54.505
654275 95.399 -53.922 54.480
654276 95.417 -53.938 54.449
654277 95.441 -53.947 54.411
654278 95.467 -53.954 54.369
654279 95.493 -53.961 54.326
654280 95.519 -53.968 54.285
654281 95.544 -53.975 54.247
654282 95.570 -53.982 54.208
654283 95.595 -53.988 54.168
654284 95.620 -53.996 54.130
654285 95.645 -54.003 54.098
654286 95.671 -54.010 54.069
654287 95.696 -54.016 54.040
654288 95.722 -54.023 54.009
654289 95.747 -54.031 53.979
654290 95.772 -54.038 53.954
654291 95.798 -54.044 53.934
654292 95.823 -54.051 53.917
654293 95.848 -54.059 53.902
654294 95.875 -54.070 53.882
654295 95.903 -54.082 53.857
654296 95.931 -54.094 53.827
654297 95.958 -54.107 53.794
654298 95.984 -54.123 53.756
654299 96.010 -54.142 53.717
654300 96.035 -54.160 53.678
654301 96.060 -54.178 53.640
654302 96.086 -54.196 53.601
654303 96.111 -54.214 53.562
654304 96.137 -54.233 53.524
654305 96.162 -54.251 53.486
654306 96.188 -54.270 53.447
654307 96.213 -54.288 53.410
654308 96.239 -54.306 53.373
654309 96.264 -54.324 53.336
654310 96.290 -54.342 53.299
654311 96.317 -54.355 53.260
654312 96.344 -54.366 53.216
654313 96.372 -54.377 53.175
654314 96.399 -54.389 53.134
654315 96.427 -54.400 53.094
654316 96.454 -54.412 53.055
654317 96.483 -54.425 53.014
654318 96.512 -54.438 52.973
654319 96.541 -54.450 52.930
654320 96.570 -54.463 52.891
654321 96.598 -54.475 52.857
654322 96.627 -54.488 52.826
654323 96.656 -54.500 52.797
654324 96.685 -54.513 52.768
654325 96.714 -54.524 52.738
654326 96.742 -54.532 52.702
654327 96.770 -54.540 52.662
654328 96.800 -54.546 52.619
654329 96.829 -54.552 52.573
654330 96.859 -54.557 52.523
654331 96.889 -54.562 52.472
654332 96.919 -54.568 52.421
654333 96.949 -54.573 52.371
654334 96.979 -54.579 52.319
654335 97.010 -54.587 52.265
654336 97.041 -54.595 52.210
654337 97.073 -54.603 52.154
654338 97.104 -54.612 52.097
654339 97.136 -54.620 52.037
654340 97.167 -54.627 51.981
654341 97.198 -54.635 51.924
654342 97.229 -54.643 51.865
654343 97.260 -54.650 51.805
654344 97.296 -54.659 51.744
654345 97.336 -54.670 51.682
654346 97.374 -54.674 51.616
654347 97.407 -54.670 51.551
654348 97.441 -54.666 51.485
654349 97.475 -54.662 51.421
654350 97.509 -54.658 51.357
654351 97.538 -54.662 51.292
654352 97.563 -54.669 51.225
654353 97.586 -54.677 51.156
654354 97.606 -54.689 51.088
654355 97.624 -54.704 51.021
654356 97.643 -54.718 50.954
654357 97.666 -54.717 50.888
654358 97.689 -54.716 50.823
654359 97.711 -54.712 50.759
654360 97.733 -54.706 50.696
654361 97.754 -54.699 50.632
654362 97.776 -54.693 50.568
654363 97.797 -54.686 50.505
654364 97.818 -54.680 50.441
654365 97.858 -54.687 50.379
654366 97.895 -54.696 50.316
654367 97.929 -54.703 50.256
654368 97.963 -54.709 50.195
654369 97.990 -54.710 50.135
654370 98.017 -54.710 50.077
654371 98.043 -54.711 50.021
654372 98.071 -54.711 49.966
654824 90.997 -51.261 56.838
654825 91.008 -51.277 56.841
654826 91.020 -51.293 56.845
654827 91.031 -51.309 56.851
654828 91.043 -51.324 56.858
654829 91.055 -51.340 56.866
654830 91.067 -51.356 56.874
654831 91.079 -51.371 56.882
654832 91.090 -51.387 56.892
654833 91.102 -51.403 56.901
654834 91.130 -51.418 56.912
654835 91.166 -51.433 56.923
654836 91.202 -51.447 56.932
654837 91.240 -51.459 56.940
654838 91.284 -51.462 56.948
654839 91.327 -51.465 56.956
654840 91.377 -51.475 56.965
654841 91.428 -51.487 56.977
654842 91.478 -51.497 56.984
654843 91.529 -51.509 56.986
654844 91.579 -51.519 56.985
654845 91.623 -51.534 56.983
654846 91.660 -51.550 56.980
654847 91.698 -51.568 56.979
654848 91.735 -51.584 56.976
654849 91.773 -51.601 56.974
654850 91.811 -51.618 56.973
654851 91.843 -51.634 56.970
654852 91.874 -51.649 56.966
654853 91.905 -51.665 56.961
654854 91.936 -51.680 56.950
654855 91.966 -51.696 56.940
654856 91.997 -51.711 56.931
654857 92.028 -51.727 56.923
654858 92.059 -51.742 56.916
654859 92.090 -51.758 56.908
654860 92.121 -51.773 56.897
654861 92.152 -51.789 56.883
654862 92.183 -51.804 56.872
654863 92.214 -51.820 56.860
654864 92.245 -51.835 56.847
654865 92.276 -51.851 56.832
654866 92.308 -51.868 56.814
654867 92.342 -51.888 56.791
654868 92.376 -51.907 56.765
654869 92.410 -51.926 56.738
654870 92.444 -51.945 56.710
654871 92.478 -51.964 56.680
654872 92.512 -51.983 56.647
654873 92.546 -52.002 56.610
654874 92.567 -52.018 56.574
654875 92.592 -52.036 56.535
654876 92.623 -52.055 56.494
654877 92.653 -52.078 56.450
654878 92.683 -52.107 56.405
654879 92.713 -52.136 56.362
654880 92.747 -52.165 56.320
654881 92.787 -52.195 56.277
654882 92.827 -52.226 56.236
654883 92.866 -52.256 56.197
654884 92.906 -52.286 56.158
654885 92.946 -52.316 56.120
654886 92.986 -52.346 56.082
654887 93.023 -52.374 56.044
654888 93.054 -52.393 56.004
654889 93.085 -52.412 55.961
654890 93.115 -52.430 55.913
654891 93.146 -52.449 55.863
654892 93.177 -52.467 55.812
654893 93.207 -52.486 55.761
654894 93.238 -52.505 55.710
654895 93.269 -52.523 55.661
654896 93.299 -52.542 55.612
654897 93.330 -52.560 55.564
654898 93.360 -52.579 55.518
654899 93.391 -52.597 55.473
654900 93.422 -52.616 55.430
654927 94.249 -53.119 54.862
654928 94.279 -53.138 54.860
654929 94.310 -53.156 54.857
654930 94.341 -53.175 54.856
654931 94.371 -53.193 54.857
654932 94.402 -53.212 54.859
654933 94.432 -53.230 54.862
654934 94.463 -53.249 54.866
654935 94.494 -53.267 54.870
654936 94.525 -53.286 54.877
654937 94.555 -53.305 54.883
654938 94.585 -53.324 54.886
654939 94.612 -53.339 54.887
654940 94.629 -53.351 54.889
654941 94.644 -53.361 54.891
654942 94.661 -53.373 54.892
654943 94.676 -53.384 54.894
654944 94.692 -53.396 54.895
654945 94.708 -53.407 54.894
654946 94.724 -53.418 54.893
654947 94.740 -53.430 54.891
654948 94.758 -53.442 54.888
654949 94.790 -53.462 54.886
654950 94.823 -53.481 54.884
654951 94.852 -53.499 54.883
654952 94.877 -53.512 54.883
654953 94.901 -53.526 54.883
654954 94.926 -53.540 54.878
654955 94.950 -53.553 54.871
654956 94.975 -53.567 54.862
654957 94.999 -53.581 54.853
654958 95.024 -53.594 54.844
654959 95.048 -53.609 54.837
654960 95.072 -53.622 54.829
654961 95.097 -53.636 54.824
654962 95.119 -53.654 54.820
654963 95.140 -53.672 54.814
654964 95.161 -53.691 54.807
654965 95.183 -53.709 54.797
654966 95.204 -53.728 54.778
654967 95.226 -53.747 54.755
654968 95.247 -53.765 54.728
654969 95.268 -53.784 54.696
654970 95.289 -53.802 54.664
654971 95.311 -53.821 54.633
654972 95.332 -53.839 54.598
654973 95.353 -53.857 54.564
654974 95.372 -53.872 54.536
654975 95.390 -53.888 54.512
654976 95.409 -53.903 54.488
654977 95.427 -53.918 54.459
654978 95.449 -53.930 54.422
654979 95.474 -53.937 54.382
654980 95.499 -53.944 54.337
654981 95.525 -53.951 54.292
654982 95.550 -53.958 54.250
654983 95.576 -53.965 54.209
654984 95.601 -53.972 54.167
654985 95.626 -53.979 54.128
654986 95.651 -53.986 54.091
654987 95.676 -53.993 54.062
654988 95.702 -54.000 54.034
654989 95.727 -54.007 54.003
654990 95.752 -54.014 53.973
654991 95.777 -54.021 53.945
654992 95.803 -54.028 53.921
654993 95.828 -54.035 53.899
654994 95.853 -54.042 53.881
654995 95.880 -54.051 53.860
654996 95.907 -54.063 53.837
654997 95.935 -54.076 53.812
654998 95.963 -54.088 53.781
654999 95.990 -54.101 53.747
655000 96.018 -54.113 53.710
655001 96.043 -54.132 53.671
655002 96.069 -54.151 53.633
655003 96.094 -54.169 53.594
655004 96.120 -54.187 53.557
655005 96.145 -54.205 53.522
655006 96.171 -54.223 53.484
655007 96.196 -54.242 53.446
655008 96.222 -54.260 53.409
655009 96.247 -54.279 53.373
655010 96.272 -54.297 53.338
655011 96.298 -54.315 53.301
655012 96.325 -54.328 53.264
655013 96.353 -54.339 53.222
655014 96.380 -54.351 53.185
655015 96.407 -54.361 53.147
655016 96.435 -54.373 53.109
655017 96.463 -54.384 53.072
655018 96.492 -54.396 53.032
655019 96.521 -54.409 52.990
655020 96.549 -54.421 52.948
655021 96.578 -54.434 52.908
655022 96.607 -54.446 52.873
655023 96.636 -54.459 52.842
655024 96.665 -54.472 52.813
655025 96.694 -54.484 52.785
655026 96.723 -54.497 52.757
655027 96.751 -54.506 52.723
655028 96.780 -54.515 52.683
655029 96.809 -54.520 52.640
655030 96.839 -54.525 52.594
655031 96.869 -54.530 52.545
655032 96.899 -54.536 52.494
655033 96.929 -54.541 52.443
655034 96.958 -54.546 52.391
655035 96.988 -54.552 52.339
655036 97.019 -54.558 52.286
655037 97.050 -54.566 52.231
655038 97.081 -54.574 52.175
655039 97.113 -54.583 52.119
655040 97.144 -54.591 52.061
655041 97.175 -54.599 52.005
655042 97.206 -54.606 51.947
655043 97.237 -54.613 51.889
655044 97.268 -54.621 51.830
655045 97.299 -54.628 51.770
655046 97.340 -54.640 51.710
655047 97.377 -54.643 51.650
655048 97.411 -54.638 51.587
655049 97.445 -54.634 51.525
655050 97.477 -54.632 51.463
655051 97.506 -54.637 51.402
655052 97.534 -54.642 51.339
655053 97.561 -54.648 51.273
655054 97.585 -54.655 51.209
655055 97.608 -54.662 51.143
655056 97.632 -54.669 51.075
655057 97.655 -54.677 51.007
655058 97.676 -54.688 50.941
655059 97.698 -54.691 50.875
655060 97.719 -54.685 50.812
655061 97.740 -54.678 50.748
655062 97.762 -54.672 50.684
655063 97.783 -54.665 50.619
655064 97.805 -54.659 50.555
655065 97.831 -54.656 50.491
655066 97.871 -54.665 50.427
655067 97.905 -54.672 50.363
655068 97.939 -54.679 50.301
655069 97.966 -54.679 50.238
655070 97.993 -54.680 50.177
655071 98.019 -54.680 50.117
655072 98.046 -54.681 50.059
655073 98.073 -54.681 50.002
655074 98.100 -54.681 49.947
655526 91.013 -51.249 56.831
655527 91.025 -51.265 56.835
655528 91.037 -51.280 56.840
655529 91.048 -51.296 56.846
655530 91.060 -51.312 56.854
655531 91.072 -51.327 56.862
655532 91.084 -51.343 56.869
655533 91.111 -51.359 56.878
655534 91.147 -51.373 56.887
655535 91.183 -51.388 56.898
655536 91.219 -51.403 56.908
655537 91.255 -51.417 56.917
655538 91.291 -51.431 56.925
655539 91.328 -51.444 56.932
655540 91.372 -51.447 56.941
655541 91.416 -51.450 56.951
655542 91.459 -51.453 56.963
655543 91.503 -51.457 56.968
655544 91.552 -51.465 56.969
655545 91.602 -51.477 56.967
655546 91.641 -51.493 56.965
655547 91.678 -51.510 56.963
655548 91.716 -51.527 56.962
655549 91.754 -51.544 56.960
655550 91.792 -51.560 56.958
655551 91.829 -51.578 56.956
655552 91.862 -51.593 56.954
655553 91.892 -51.609 56.950
655554 91.924 -51.624 56.946
655555 91.954 -51.640 56.936
655556 91.985 -51.655 56.928
655557 92.016 -51.671 56.921
655558 92.047 -51.686 56.915
655559 92.078 -51.702 56.908
655560 92.109 -51.717 56.902
655561 92.140 -51.733 56.891
655562 92.171 -51.748 56.878
655563 92.203 -51.766 56.867
655564 92.237 -51.785 56.855
655565 92.271 -51.804 56.841
655566 92.305 -51.823 56.827
655567 92.339 -51.842 56.809
655568 92.373 -51.861 56.786
655569 92.407 -51.880 56.761
655570 92.441 -51.900 56.736
655571 92.475 -51.919 56.710
655572 92.509 -51.938 56.680
655573 92.538 -51.956 56.647
655574 92.559 -51.972 56.611
655575 92.582 -51.988 56.575
655576 92.613 -52.007 56.536
655577 92.644 -52.028 56.495
655578 92.674 -52.052 56.452
655579 92.704 -52.081 56.408
655580 92.734 -52.110 56.365
655581 92.764 -52.139 56.323
655582 92.799 -52.168 56.282
655583 92.839 -52.199 56.241
655584 92.878 -52.228 56.202
655585 92.918 -52.258 56.164
655586 92.958 -52.289 56.127
655587 92.998 -52.319 56.089
655588 93.038 -52.349 56.051
655589 93.069 -52.369 56.013
655590 93.100 -52.387 55.972
655591 93.131 -52.406 55.926
655592 93.161 -52.425 55.878
655593 93.192 -52.443 55.826
655594 93.222 -52.462 55.776
655595 93.253 -52.481 55.726
655596 93.284 -52.499 55.676
655597 93.314 -52.518 55.627
655598 93.345 -52.537 55.579
655599 93.375 -52.555 55.533
655600 93.406 -52.574 55.487
655601 93.437 -52.593 55.445
655628 94.264 -53.095 54.864
655629 94.294 -53.114 54.861
655630 94.325 -53.132 54.858
655631 94.356 -53.151 54.857
655632 94.387 -53.170 54.857
655633 94.417 -53.189 54.858
655634 94.448 -53.207 54.860
655635 94.478 -53.226 54.862
655636 94.508 -53.244 54.866
655637 94.537 -53.262 54.870
655638 94.567 -53.280 54.875
655639 94.596 -53.299 54.876
655640 94.625 -53.317 54.876
655641 94.653 -53.325 54.877
655642 94.673 -53.338 54.879
655643 94.692 -53.352 54.881
655644 94.709 -53.364 54.882
655645 94.725 -53.375 54.882
655646 94.741 -53.386 54.881
655647 94.757 -53.398 54.879
655648 94.773 -53.409 54.877
655649 94.788 -53.421 54.874
655650 94.805 -53.431 54.871
655651 94.835 -53.450 54.868
655652 94.868 -53.470 54.866
655653 94.894 -53.484 54.866
655654 94.918 -53.499 54.865
655655 94.943 -53.512 54.862
655656 94.967 -53.526 54.857
655657 94.992 -53.540 54.851
655658 95.016 -53.553 54.845
655659 95.041 -53.567 54.841
655660 95.065 -53.581 54.839
655661 95.087 -53.599 54.833
655662 95.108 -53.617 54.824
655663 95.130 -53.636 54.815
655664 95.151 -53.654 54.806
655665 95.172 -53.673 54.797
655666 95.194 -53.691 54.783
655667 95.215 -53.710 54.763
655668 95.237 -53.728 54.740
655669 95.258 -53.747 54.711
655670 95.279 -53.766 54.680
655671 95.301 -53.784 54.650
655672 95.322 -53.803 54.624
655673 95.343 -53.822 54.595
655674 95.362 -53.838 54.566
655675 95.381 -53.853 54.541
655676 95.400 -53.868 54.518
655677 95.419 -53.884 54.493
655678 95.441 -53.899 54.463
655679 95.463 -53.912 54.427
655680 95.480 -53.920 54.389
655681 95.505 -53.927 54.344
655682 95.530 -53.934 54.296
655683 95.555 -53.941 54.252
655684 95.581 -53.948 54.209
655685 95.606 -53.955 54.166
655686 95.631 -53.962 54.125
655687 95.657 -53.969 54.086
655688 95.682 -53.976 54.055
655689 95.707 -53.983 54.026
655690 95.733 -53.990 53.996
655691 95.758 -53.997 53.966
655692 95.783 -54.004 53.937
655693 95.808 -54.011 53.912
655694 95.834 -54.018 53.888
655695 95.859 -54.025 53.867
655696 95.885 -54.033 53.844
655697 95.912 -54.045 53.821
655698 95.940 -54.057 53.796
655699 95.968 -54.070 53.769
655700 95.995 -54.082 53.737
655701 96.023 -54.095 53.703
655702 96.051 -54.107 53.664
655703 96.077 -54.123 53.627
655704 96.103 -54.141 53.591
655705 96.128 -54.159 53.556
655706 96.154 -54.177 53.523
655707 96.179 -54.196 53.488
655708 96.204 -54.214 53.449
655709 96.230 -54.232 53.412
655710 96.255 -54.251 53.376
655711 96.281 -54.269 53.340
655712 96.306 -54.287 53.304
655713 96.333 -54.301 53.268
655714 96.361 -54.312 53.230
655715 96.388 -54.323 53.196
655716 96.415 -54.335 53.162
655717 96.443 -54.345 53.126
655718 96.473 -54.356 53.088
655719 96.502 -54.368 53.048
655720 96.531 -54.381 53.007
655721 96.559 -54.393 52.965
655722 96.588 -54.405 52.924
655723 96.616 -54.418 52.888
655724 96.645 -54.430 52.856
655725 96.674 -54.443 52.828
655726 96.702 -54.455 52.800
655727 96.731 -54.468 52.773
655728 96.760 -54.480 52.741
655729 96.789 -54.488 52.702
655730 96.819 -54.493 52.659
655731 96.849 -54.499 52.613
655732 96.879 -54.504 52.565
655733 96.908 -54.509 52.515
655734 96.938 -54.515 52.463
655735 96.968 -54.520 52.410
655736 96.998 -54.525 52.357
655737 97.028 -54.530 52.304
655738 97.058 -54.537 52.251
655739 97.090 -54.546 52.197
655740 97.121 -54.554 52.141
655741 97.153 -54.562 52.085
655742 97.183 -54.569 52.029
655743 97.215 -54.577 51.972
655744 97.245 -54.584 51.914
655745 97.277 -54.592 51.857
655746 97.307 -54.599 51.800
655747 97.344 -54.609 51.743
655748 97.380 -54.611 51.685
655749 97.414 -54.606 51.624
655750 97.446 -54.606 51.563
655751 97.474 -54.612 51.503
655752 97.502 -54.617 51.444
655753 97.531 -54.622 51.383
655754 97.559 -54.627 51.319
655755 97.583 -54.634 51.258
655756 97.607 -54.641 51.194
655757 97.630 -54.648 51.129
655758 97.654 -54.656 51.063
655759 97.677 -54.662 50.996
655760 97.704 -54.662 50.930
655761 97.727 -54.658 50.866
655762 97.748 -54.652 50.802
655763 97.770 -54.645 50.738
655764 97.791 -54.638 50.673
655765 97.813 -54.632 50.608
655766 97.844 -54.634 50.542
655767 97.881 -54.641 50.477
655768 97.915 -54.648 50.411
655769 97.942 -54.649 50.347
655770 97.968 -54.649 50.283
655771 97.996 -54.649 50.220
655772 98.022 -54.649 50.158
655773 98.049 -54.650 50.098
655774 98.076 -54.650 50.039
655775 98.103 -54.651 49.983
655776 98.130 -54.651 49.928
656227 91.018 -51.221 56.822
656228 91.030 -51.237 56.825
656229 91.042 -51.252 56.829
656230 91.053 -51.268 56.836
656231 91.065 -51.284 56.843
656232 91.092 -51.299 56.850
656233 91.129 -51.314 56.857
656234 91.164 -51.328 56.865
656235 91.201 -51.343 56.873
656236 91.236 -51.358 56.884
656237 91.273 -51.372 56.894
656238 91.308 -51.387 56.902
656239 91.345 -51.401 56.910
656240 91.380 -51.416 56.918
656241 91.417 -51.429 56.928
656242 91.461 -51.433 56.938
656243 91.504 -51.436 56.948
656244 91.548 -51.439 56.952
656245 91.591 -51.442 56.951
656246 91.628 -51.448 56.950
656247 91.661 -51.455 56.948
656248 91.697 -51.469 56.947
656249 91.734 -51.486 56.947
656250 91.772 -51.503 56.946
656251 91.810 -51.520 56.943
656252 91.848 -51.537 56.941
656253 91.880 -51.553 56.939
656254 91.912 -51.568 56.937
656255 91.942 -51.584 56.933
656256 91.973 -51.599 56.924
656257 92.004 -51.615 56.917
656258 92.035 -51.630 56.912
656259 92.066 -51.646 56.907
656260 92.098 -51.663 56.901
656261 92.132 -51.682 56.895
656262 92.166 -51.701 56.884
656263 92.200 -51.720 56.872
656264 92.234 -51.739 56.862
656265 92.268 -51.758 56.850
656266 92.302 -51.778 56.837
656267 92.336 -51.797 56.823
656268 92.370 -51.816 56.805
656269 92.404 -51.835 56.783
656270 92.438 -51.854 56.759
656271 92.472 -51.873 56.735
656272 92.506 -51.893 56.710
656273 92.530 -51.909 56.681
656274 92.551 -51.925 56.648
656275 92.573 -51.941 56.613
656276 92.603 -51.960 56.576
656277 92.634 -51.980 56.538
656278 92.665 -52.000 56.497
656279 92.695 -52.026 56.455
656280 92.725 -52.054 56.412
656281 92.755 -52.083 56.369
656282 92.785 -52.112 56.327
656283 92.815 -52.141 56.287
656284 92.851 -52.170 56.247
656285 92.890 -52.201 56.209
656286 92.930 -52.231 56.171
656287 92.970 -52.261 56.133
656288 93.010 -52.291 56.095
656289 93.050 -52.321 56.057
656290 93.084 -52.346 56.019
656291 93.115 -52.364 55.979
656292 93.146 -52.383 55.936
656293 93.177 -52.401 55.890
656294 93.207 -52.420 55.840
656295 93.238 -52.438 55.790
656296 93.268 -52.457 55.740
656297 93.299 -52.476 55.690
656298 93.330 -52.494 55.642
656299 93.360 -52.513 55.594
656300 93.391 -52.531 55.548
656301 93.422 -52.550 55.502
656302 93.452 -52.568 55.459
656329 94.279 -53.071 54.866
656330 94.310 -53.090 54.862
656331 94.341 -53.109 54.859
656332 94.371 -53.127 54.857
656333 94.401 -53.145 54.857
656334 94.430 -53.164 54.855
656335 94.460 -53.183 54.856
656336 94.489 -53.201 54.858
656337 94.519 -53.219 54.860
656338 94.548 -53.237 54.862
656339 94.577 -53.255 54.865
656340 94.607 -53.273 54.865
656341 94.636 -53.292 54.866
656342 94.665 -53.307 54.867
656343 94.693 -53.313 54.868
656344 94.715 -53.324 54.869
656345 94.734 -53.337 54.869
656346 94.753 -53.350 54.869
656347 94.771 -53.364 54.867
656348 94.789 -53.377 54.866
656349 94.806 -53.388 54.864
656350 94.821 -53.399 54.860
656351 94.837 -53.411 54.857
656352 94.853 -53.422 54.854
656353 94.881 -53.439 54.852
656354 94.911 -53.457 54.851
656355 94.935 -53.471 54.851
656356 94.960 -53.484 54.850
656357 94.984 -53.498 54.847
656358 95.009 -53.512 54.843
656359 95.033 -53.526 54.841
656360 95.055 -53.543 54.840
656361 95.076 -53.562 54.841
656362 95.098 -53.581 54.834
656363 95.119 -53.599 54.824
656364 95.140 -53.618 54.811
656365 95.162 -53.636 54.798
656366 95.183 -53.655 54.785
656367 95.205 -53.673 54.768
656368 95.226 -53.692 54.746
656369 95.248 -53.710 54.721
656370 95.269 -53.729 54.694
656371 95.290 -53.747 54.667
656372 95.312 -53.766 54.642
656373 95.333 -53.784 54.623
656374 95.354 -53.803 54.598
656375 95.373 -53.818 54.572
656376 95.391 -53.834 54.546
656377 95.410 -53.849 54.521
656378 95.434 -53.864 54.495
656379 95.458 -53.879 54.464
656380 95.482 -53.894 54.427
656381 95.497 -53.902 54.384
656382 95.513 -53.910 54.342
656383 95.536 -53.917 54.297
656384 95.561 -53.925 54.252
656385 95.586 -53.932 54.207
656386 95.612 -53.938 54.164
656387 95.637 -53.945 54.122
656388 95.662 -53.953 54.082
656389 95.687 -53.960 54.048
656390 95.712 -53.966 54.018
656391 95.738 -53.973 53.990
656392 95.763 -53.980 53.960
656393 95.789 -53.988 53.932
656394 95.814 -53.994 53.905
656395 95.839 -54.001 53.882
656396 95.865 -54.008 53.858
656397 95.890 -54.016 53.834
656398 95.917 -54.027 53.809
656399 95.945 -54.039 53.784
656400 95.972 -54.052 53.758
656401 96.000 -54.064 53.728
656402 96.028 -54.076 53.696
656403 96.055 -54.088 53.660
656404 96.083 -54.101 53.624
656405 96.111 -54.113 53.589
656406 96.136 -54.132 53.556
656407 96.162 -54.150 53.524
656408 96.187 -54.168 53.490
656409 96.213 -54.186 53.453
656410 96.238 -54.204 53.417
656411 96.264 -54.223 53.381
656412 96.289 -54.242 53.346
656413 96.315 -54.260 53.311
656414 96.341 -54.273 53.277
656415 96.369 -54.285 53.242
656416 96.396 -54.296 53.211
656417 96.424 -54.308 53.177
656418 96.453 -54.318 53.141
656419 96.483 -54.328 53.103
656420 96.512 -54.340 53.062
656421 96.541 -54.352 53.022
656422 96.570 -54.365 52.981
656423 96.598 -54.377 52.941
656424 96.626 -54.390 52.903
656425 96.655 -54.402 52.870
656426 96.684 -54.415 52.839
656427 96.713 -54.427 52.811
656428 96.741 -54.440 52.786
656429 96.770 -54.452 52.755
656430 96.800 -54.459 52.718
656431 96.828 -54.466 52.677
656432 96.859 -54.471 52.632
656433 96.888 -54.477 52.584
656434 96.918 -54.482 52.534
656435 96.948 -54.487 52.481
656436 96.978 -54.493 52.427
656437 97.008 -54.499 52.373
656438 97.037 -54.504 52.322
656439 97.067 -54.509 52.271
656440 97.098 -54.516 52.218
656441 97.130 -54.524 52.163
656442 97.161 -54.533 52.108
656443 97.192 -54.540 52.053
656444 97.223 -54.548 51.997
656445 97.254 -54.555 51.940
656446 97.285 -54.562 51.884
656447 97.316 -54.570 51.830
656448 97.348 -54.577 51.776
656449 97.384 -54.579 51.719
656450 97.414 -54.580 51.659
656451 97.442 -54.586 51.600
656452 97.470 -54.591 51.542
656453 97.499 -54.596 51.484
656454 97.527 -54.602 51.425
656455 97.556 -54.606 51.364
656456 97.582 -54.612 51.306
656457 97.605 -54.620 51.244
656458 97.629 -54.627 51.179
656459 97.654 -54.630 51.115
656460 97.680 -54.629 51.050
656461 97.706 -54.628 50.985
656462 97.732 -54.627 50.921
656463 97.756 -54.624 50.858
656464 97.778 -54.618 50.794
656465 97.799 -54.612 50.729
656466 97.820 -54.605 50.662
656467 97.856 -54.611 50.594
656468 97.891 -54.618 50.528
656469 97.918 -54.618 50.461
656470 97.944 -54.618 50.395
656471 97.971 -54.618 50.329
656472 97.998 -54.619 50.265
656473 98.025 -54.619 50.201
656474 98.052 -54.620 50.139
656475 98.079 -54.620 50.079
656476 98.105 -54.621 50.021
656477 98.132 -54.621 49.964
656907 90.552 -50.904 56.851
656908 90.585 -50.923 56.836
656909 90.618 -50.942 56.821
656910 90.651 -50.961 56.805
656911 90.684 -50.978 56.789
656912 90.717 -50.986 56.771
656913 90.749 -50.994 56.752
656914 90.781 -51.001 56.733
656929 91.035 -51.209 56.816
656930 91.047 -51.225 56.820
656931 91.074 -51.240 56.825
656932 91.110 -51.255 56.832
656933 91.146 -51.269 56.839
656934 91.182 -51.284 56.846
656935 91.218 -51.298 56.853
656936 91.254 -51.313 56.861
656937 91.290 -51.327 56.871
656938 91.326 -51.342 56.880
656939 91.362 -51.356 56.888
656940 91.398 -51.371 56.896
656941 91.434 -51.386 56.905
656942 91.470 -51.400 56.914
656943 91.506 -51.415 56.924
656944 91.549 -51.418 56.932
656945 91.593 -51.421 56.935
656946 91.629 -51.428 56.934
656947 91.661 -51.434 56.933
656948 91.694 -51.442 56.932
656949 91.726 -51.449 56.931
656950 91.758 -51.456 56.932
656951 91.791 -51.463 56.931
656952 91.828 -51.479 56.929
656953 91.866 -51.496 56.928
656954 91.899 -51.512 56.926
656955 91.930 -51.528 56.923
656956 91.961 -51.543 56.919
656957 91.993 -51.559 56.914
656958 92.027 -51.579 56.908
656959 92.061 -51.598 56.903
656960 92.095 -51.617 56.899
656961 92.129 -51.637 56.894
656962 92.163 -51.656 56.888
656963 92.197 -51.675 56.876
656964 92.231 -51.694 56.866
656965 92.265 -51.713 56.856
656966 92.299 -51.732 56.845
656967 92.333 -51.751 56.832
656968 92.367 -51.771 56.818
656969 92.401 -51.790 56.799
656970 92.435 -51.809 56.778
656971 92.469 -51.829 56.756
656972 92.501 -51.847 56.733
656973 92.522 -51.863 56.711
656974 92.544 -51.878 56.683
656975 92.564 -51.894 56.650
656976 92.593 -51.913 56.614
656977 92.624 -51.933 56.578
656978 92.655 -51.953 56.540
656979 92.686 -51.973 56.501
656980 92.716 -51.999 56.460
656981 92.746 -52.028 56.416
656982 92.776 -52.057 56.374
656983 92.806 -52.086 56.333
656984 92.836 -52.114 56.293
656985 92.866 -52.143 56.254
656986 92.903 -52.174 56.215
656987 92.943 -52.204 56.178
656988 92.983 -52.233 56.139
656989 93.022 -52.264 56.102
656990 93.062 -52.294 56.064
656991 93.100 -52.321 56.026
656992 93.131 -52.340 55.986
656993 93.161 -52.358 55.944
656994 93.192 -52.377 55.901
656995 93.222 -52.396 55.852
656996 93.253 -52.415 55.803
656997 93.284 -52.434 55.754
656998 93.314 -52.452 55.705
656999 93.345 -52.471 55.657
657000 93.375 -52.489 55.609
657001 93.406 -52.508 55.563
657002 93.437 -52.526 55.518
657003 93.468 -52.545 55.474
657030 94.294 -53.048 54.868
657031 94.324 -53.066 54.863
657032 94.353 -53.084 54.860
657033 94.383 -53.102 54.857
657034 94.412 -53.121 54.856
657035 94.441 -53.139 54.853
657036 94.471 -53.158 54.853
657037 94.500 -53.176 54.853
657038 94.530 -53.194 54.853
657039 94.559 -53.212 54.854
657040 94.588 -53.230 54.855
657041 94.618 -53.248 54.856
657042 94.647 -53.267 54.856
657043 94.676 -53.285 54.856
657044 94.704 -53.295 54.857
657045 94.732 -53.302 54.858
657046 94.758 -53.309 54.858
657047 94.776 -53.323 54.858
657048 94.795 -53.336 54.855
657049 94.814 -53.349 54.854
657050 94.832 -53.363 54.852
657051 94.851 -53.377 54.849
657052 94.870 -53.389 54.846
657053 94.886 -53.401 54.843
657054 94.902 -53.412 54.842
657055 94.926 -53.428 54.841
657056 94.953 -53.443 54.843
657057 94.977 -53.457 54.842
657058 95.001 -53.471 54.840
657059 95.024 -53.488 54.839
657060 95.045 -53.507 54.840
657061 95.066 -53.525 54.841
657062 95.088 -53.544 54.842
657063 95.109 -53.562 54.835
657064 95.130 -53.581 54.824
657065 95.151 -53.599 54.809
657066 95.173 -53.618 54.790
657067 95.194 -53.637 54.772
657068 95.216 -53.655 54.753
657069 95.237 -53.674 54.728
657070 95.259 -53.692 54.702
657071 95.280 -53.711 54.678
657072 95.301 -53.729 54.657
657073 95.323 -53.748 54.640
657074 95.344 -53.766 54.625
657075 95.364 -53.784 54.604
657076 95.383 -53.799 54.580
657077 95.403 -53.814 54.553
657078 95.427 -53.830 54.526
657079 95.451 -53.845 54.496
657080 95.476 -53.860 54.463
657081 95.500 -53.875 54.424
657082 95.516 -53.883 54.379
657083 95.531 -53.891 54.337
657084 95.547 -53.900 54.294
657085 95.566 -53.908 54.250
657086 95.592 -53.915 54.207
657087 95.617 -53.922 54.164
657088 95.642 -53.928 54.122
657089 95.668 -53.936 54.082
657090 95.693 -53.943 54.049
657091 95.718 -53.950 54.019
657092 95.743 -53.957 53.990
657093 95.769 -53.963 53.960
657094 95.794 -53.971 53.931
657095 95.819 -53.978 53.905
657096 95.844 -53.985 53.882
657097 95.870 -53.991 53.858
657098 95.895 -53.999 53.832
657099 95.922 -54.009 53.806
657100 95.950 -54.021 53.779
657101 95.977 -54.033 53.751
657102 96.005 -54.046 53.721
657103 96.033 -54.058 53.691
657104 96.060 -54.070 53.656
657105 96.088 -54.082 53.622
657106 96.116 -54.095 53.589
657107 96.144 -54.107 53.557
657108 96.170 -54.122 53.526
657109 96.196 -54.140 53.493
657110 96.221 -54.159 53.457
657111 96.247 -54.177 53.422
657112 96.272 -54.195 53.388
657113 96.298 -54.214 53.353
657114 96.323 -54.232 53.320
657115 96.349 -54.246 53.287
657116 96.377 -54.258 53.256
657117 96.404 -54.269 53.226
657118 96.433 -54.279 53.193
657119 96.463 -54.290 53.158
657120 96.493 -54.301 53.120
657121 96.522 -54.312 53.079
657122 96.551 -54.324 53.040
657123 96.580 -54.337 52.999
657124 96.608 -54.349 52.958
657125 96.637 -54.362 52.920
657126 96.665 -54.374 52.885
657127 96.694 -54.387 52.853
657128 96.723 -54.399 52.823
657129 96.753 -54.409 52.799
657130 96.783 -54.419 52.768
657131 96.813 -54.427 52.732
657132 96.841 -54.433 52.691
657133 96.871 -54.440 52.648
657134 96.899 -54.447 52.600
657135 96.929 -54.454 52.549
657136 96.957 -54.461 52.497
657137 96.987 -54.466 52.442
657138 97.017 -54.471 52.389
657139 97.047 -54.477 52.338
657140 97.077 -54.482 52.289
657141 97.107 -54.487 52.240
657142 97.138 -54.496 52.187
657143 97.169 -54.504 52.133
657144 97.200 -54.511 52.076
657145 97.231 -54.518 52.021
657146 97.262 -54.526 51.966
657147 97.293 -54.533 51.911
657148 97.324 -54.540 51.859
657149 97.355 -54.548 51.806
657150 97.382 -54.554 51.751
657151 97.410 -54.560 51.693
657152 97.439 -54.565 51.636
657153 97.467 -54.571 51.579
657154 97.495 -54.576 51.522
657155 97.523 -54.581 51.465
657156 97.552 -54.587 51.406
657157 97.580 -54.591 51.352
657158 97.604 -54.599 51.291
657159 97.630 -54.598 51.227
657160 97.656 -54.596 51.164
657161 97.682 -54.596 51.101
657162 97.709 -54.595 51.038
657163 97.735 -54.594 50.976
657164 97.761 -54.593 50.914
657165 97.786 -54.591 50.850
657166 97.807 -54.584 50.785
657167 97.830 -54.579 50.717
657168 97.867 -54.587 50.647
657169 97.893 -54.587 50.579
657170 97.920 -54.588 50.512
657171 97.947 -54.588 50.444
657172 97.974 -54.589 50.376
657173 98.001 -54.589 50.309
657174 98.027 -54.589 50.244
657175 98.055 -54.590 50.181
657176 98.081 -54.590 50.119
657177 98.108 -54.590 50.059
657178 98.135 -54.590 50.000
657179 98.162 -54.591 49.943
657606 90.506 -50.832 56.873
657607 90.539 -50.851 56.861
657608 90.572 -50.870 56.848
657609 90.605 -50.890 56.834
657610 90.638 -50.909 56.820
657611 90.671 -50.929 56.804
657612 90.704 -50.945 56.788
657613 90.736 -50.952 56.771
657614 90.769 -50.961 56.753
657615 90.801 -50.968 56.734
657616 90.833 -50.976 56.715
657617 90.866 -50.985 56.697
657631 91.091 -51.196 56.811
657632 91.128 -51.210 56.815
657633 91.163 -51.224 56.821
657634 91.200 -51.239 56.829
657635 91.235 -51.253 56.835
657636 91.272 -51.268 56.841
657637 91.307 -51.283 56.849
657638 91.344 -51.297 56.857
657639 91.379 -51.312 56.866
657640 91.416 -51.326 56.875
657641 91.451 -51.341 56.882
657642 91.488 -51.356 56.891
657643 91.523 -51.370 56.899
657644 91.560 -51.384 56.907
657645 91.595 -51.399 56.915
657646 91.629 -51.407 56.918
657647 91.661 -51.414 56.917
657648 91.694 -51.421 56.916
657649 91.726 -51.428 56.916
657650 91.758 -51.435 56.916
657651 91.791 -51.442 56.918
657652 91.823 -51.450 56.917
657653 91.856 -51.456 56.915
657654 91.889 -51.463 56.914
657655 91.922 -51.476 56.912
657656 91.956 -51.495 56.909
657657 91.990 -51.515 56.906
657658 92.024 -51.534 56.903
657659 92.058 -51.553 56.898
657660 92.092 -51.572 56.895
657661 92.126 -51.591 56.891
657662 92.160 -51.610 56.885
657663 92.194 -51.629 56.878
657664 92.228 -51.649 56.869
657665 92.262 -51.668 56.859
657666 92.296 -51.687 56.850
657667 92.330 -51.707 56.840
657668 92.364 -51.725 56.827
657669 92.398 -51.744 56.812
657670 92.432 -51.764 56.793
657671 92.466 -51.783 56.772
657672 92.493 -51.801 56.751
657673 92.514 -51.816 56.730
657674 92.535 -51.832 56.709
657675 92.556 -51.848 56.683
657676 92.584 -51.866 56.649
657677 92.615 -51.885 56.614
657678 92.645 -51.905 56.578
657679 92.676 -51.925 56.541
657680 92.707 -51.945 56.504
657681 92.737 -51.973 56.464
657682 92.767 -52.001 56.423
657683 92.797 -52.030 56.381
657684 92.827 -52.059 56.341
657685 92.857 -52.088 56.301
657686 92.886 -52.117 56.261
657687 92.917 -52.146 56.223
657688 92.955 -52.176 56.185
657689 92.995 -52.206 56.147
657690 93.035 -52.236 56.109
657691 93.074 -52.267 56.071
657692 93.114 -52.296 56.032
657693 93.146 -52.317 55.993
657694 93.177 -52.335 55.953
657695 93.207 -52.354 55.910
657696 93.238 -52.372 55.864
657697 93.268 -52.391 55.816
657698 93.299 -52.409 55.767
657699 93.330 -52.428 55.719
657700 93.360 -52.447 55.671
657701 93.391 -52.465 55.624
657702 93.422 -52.484 55.578
657703 93.452 -52.503 55.533
657704 93.483 -52.521 55.489
657731 94.306 -53.023 54.869
657732 94.335 -53.041 54.864
657733 94.364 -53.059 54.859
657734 94.394 -53.078 54.855
657735 94.423 -53.096 54.852
657736 94.452 -53.114 54.850
657737 94.482 -53.132 54.848
657738 94.511 -53.151 54.847
657739 94.540 -53.169 54.847
657740 94.570 -53.187 54.846
657741 94.599 -53.205 54.846
657742 94.628 -53.223 54.847
657743 94.658 -53.242 54.846
657744 94.687 -53.261 54.846
657745 94.716 -53.277 54.847
657746 94.744 -53.283 54.847
657747 94.771 -53.289 54.847
657748 94.799 -53.295 54.846
657749 94.819 -53.308 54.844
657750 94.837 -53.322 54.842
657751 94.856 -53.335 54.841
657752 94.875 -53.349 54.838
657753 94.893 -53.362 54.835
657754 94.912 -53.376 54.833
657755 94.931 -53.389 54.832
657756 94.950 -53.402 54.833
657757 94.970 -53.416 54.836
657758 94.992 -53.433 54.837
657759 95.013 -53.452 54.836
657760 95.035 -53.470 54.836
657761 95.056 -53.489 54.838
657762 95.077 -53.507 54.837
657763 95.099 -53.526 54.836
657764 95.120 -53.544 54.830
657765 95.141 -53.563 54.819
657766 95.163 -53.581 54.802
657767 95.184 -53.600 54.780
657768 95.205 -53.618 54.757
657769 95.227 -53.637 54.734
657770 95.248 -53.656 54.710
657771 95.270 -53.675 54.685
657772 95.291 -53.693 54.665
657773 95.312 -53.711 54.652
657774 95.334 -53.730 54.643
657775 95.355 -53.748 54.633
657776 95.374 -53.764 54.613
657777 95.396 -53.779 54.589
657778 95.420 -53.795 54.561
657779 95.443 -53.810 54.531
657780 95.468 -53.825 54.497
657781 95.493 -53.841 54.460
657782 95.518 -53.856 54.419
657783 95.534 -53.865 54.373
657784 95.550 -53.873 54.331
657785 95.565 -53.882 54.289
657786 95.581 -53.890 54.248
657787 95.597 -53.898 54.207
657788 95.622 -53.905 54.164
657789 95.648 -53.912 54.123
657790 95.673 -53.919 54.086
657791 95.698 -53.926 54.053
657792 95.724 -53.933 54.024
657793 95.749 -53.940 53.995
657794 95.774 -53.947 53.964
657795 95.800 -53.954 53.936
657796 95.825 -53.961 53.909
657797 95.850 -53.968 53.886
657798 95.875 -53.975 53.863
657799 95.901 -53.982 53.836
657800 95.926 -53.991 53.808
657801 95.954 -54.003 53.779
657802 95.982 -54.015 53.749
657803 96.010 -54.028 53.719
657804 96.038 -54.040 53.689
657805 96.065 -54.052 53.656
657806 96.093 -54.064 53.621
657807 96.121 -54.077 53.588
657808 96.149 -54.089 53.558
657809 96.176 -54.101 53.527
657810 96.204 -54.113 53.495
657811 96.230 -54.131 53.462
657812 96.255 -54.149 53.428
657813 96.281 -54.167 53.395
657814 96.306 -54.185 53.363
657815 96.331 -54.204 53.332
657816 96.358 -54.219 53.300
657817 96.385 -54.230 53.269
657818 96.413 -54.242 53.241
657819 96.443 -54.252 53.211
657820 96.473 -54.262 53.177
657821 96.503 -54.273 53.138
657822 96.533 -54.284 53.098
657823 96.561 -54.296 53.060
657824 96.590 -54.309 53.018
657825 96.618 -54.321 52.975
657826 96.647 -54.334 52.936
657827 96.675 -54.346 52.901
657828 96.705 -54.357 52.868
657829 96.735 -54.367 52.837
657830 96.766 -54.377 52.811
657831 96.796 -54.386 52.780
657832 96.825 -54.395 52.744
657833 96.855 -54.402 52.704
657834 96.884 -54.408 52.661
657835 96.912 -54.415 52.613
657836 96.942 -54.422 52.562
657837 96.970 -54.429 52.511
657838 97.000 -54.436 52.456
657839 97.028 -54.442 52.404
657840 97.058 -54.449 52.354
657841 97.087 -54.455 52.306
657842 97.116 -54.461 52.258
657843 97.147 -54.467 52.207
657844 97.177 -54.474 52.155
657845 97.209 -54.482 52.099
657846 97.239 -54.489 52.044
657847 97.270 -54.497 51.990
657848 97.300 -54.505 51.936
657849 97.330 -54.512 51.884
657850 97.360 -54.520 51.832
657851 97.386 -54.523 51.779
657852 97.412 -54.527 51.723
657853 97.440 -54.533 51.668
657854 97.468 -54.540 51.613
657855 97.496 -54.546 51.558
657856 97.524 -54.552 51.502
657857 97.551 -54.558 51.445
657858 97.579 -54.564 51.393
657859 97.606 -54.565 51.334
657860 97.633 -54.564 51.271
657861 97.659 -54.563 51.209
657862 97.685 -54.562 51.148
657863 97.711 -54.562 51.086
657864 97.737 -54.561 51.026
657865 97.764 -54.560 50.965
657866 97.790 -54.559 50.903
657867 97.815 -54.557 50.839
657868 97.842 -54.556 50.773
657869 97.869 -54.557 50.703
657870 97.896 -54.557 50.631
657871 97.923 -54.558 50.562
657872 97.950 -54.558 50.492
657873 97.977 -54.558 50.423
657874 98.003 -54.558 50.355
657875 98.030 -54.559 50.288
657876 98.057 -54.559 50.222
657877 98.084 -54.559 50.159
657878 98.110 -54.560 50.098
657879 98.138 -54.560 50.037
657880 98.164 -54.561 49.979
657881 98.191 -54.561 49.924
658306 90.493 -50.779 56.876
658307 90.526 -50.798 56.867
658308 90.559 -50.818 56.857
658309 90.592 -50.837 56.845
658310 90.625 -50.857 56.832
658311 90.658 -50.876 56.818
658312 90.691 -50.895 56.804
658313 90.724 -50.911 56.789
658314 90.756 -50.919 56.772
658315 90.789 -50.927 56.754
658316 90.821 -50.935 56.735
658317 90.853 -50.943 56.716
658318 90.886 -50.951 56.697
658319 90.904 -50.964 56.679
658332 91.145 -51.165 56.802
658333 91.181 -51.180 56.805
658334 91.217 -51.194 56.811
658335 91.253 -51.208 56.817
658336 91.289 -51.224 56.823
658337 91.325 -51.238 56.829
658338 91.361 -51.252 56.836
658339 91.397 -51.267 56.844
658340 91.433 -51.281 56.853
658341 91.469 -51.296 56.861
658342 91.505 -51.311 56.868
658343 91.538 -51.325 56.875
658344 91.569 -51.340 56.882
658345 91.600 -51.354 56.890
658346 91.631 -51.368 56.896
658347 91.662 -51.383 56.899
658348 91.694 -51.396 56.899
658349 91.726 -51.405 56.899
658350 91.758 -51.412 56.900
658351 91.791 -51.419 56.901
658352 91.823 -51.426 56.903
658353 91.856 -51.434 56.903
658354 91.888 -51.440 56.902
658355 91.921 -51.447 56.902
658356 91.953 -51.455 56.898
658357 91.987 -51.469 56.895
658358 92.021 -51.488 56.894
658359 92.055 -51.507 56.891
658360 92.089 -51.527 56.888
658361 92.123 -51.546 56.885
658362 92.157 -51.565 56.880
658363 92.191 -51.584 56.875
658364 92.225 -51.603 56.868
658365 92.259 -51.622 56.860
658366 92.293 -51.642 56.852
658367 92.327 -51.661 56.842
658368 92.361 -51.680 56.831
658369 92.395 -51.700 56.819
658370 92.429 -51.719 56.804
658371 92.463 -51.738 56.785
658372 92.486 -51.754 56.766
658373 92.506 -51.769 56.746
658374 92.527 -51.785 56.725
658375 92.549 -51.801 56.705
658376 92.574 -51.818 56.679
658377 92.605 -51.838 56.647
658378 92.636 -51.858 56.612
658379 92.666 -51.878 56.577
658380 92.697 -51.898 56.542
658381 92.728 -51.917 56.506
658382 92.758 -51.946 56.468
658383 92.788 -51.975 56.430
658384 92.817 -52.004 56.390
658385 92.848 -52.033 56.349
658386 92.877 -52.062 56.309
658387 92.908 -52.091 56.271
658388 92.937 -52.120 56.232
658389 92.968 -52.148 56.194
658390 93.007 -52.179 56.156
658391 93.047 -52.208 56.118
658392 93.086 -52.239 56.079
658393 93.127 -52.269 56.041
658394 93.161 -52.293 56.002
658395 93.192 -52.311 55.962
658396 93.222 -52.330 55.920
658397 93.253 -52.349 55.874
658398 93.284 -52.368 55.827
658399 93.314 -52.386 55.779
658400 93.345 -52.405 55.732
658401 93.375 -52.423 55.685
658402 93.406 -52.442 55.638
658403 93.437 -52.460 55.592
658404 93.468 -52.479 55.548
658405 93.498 -52.497 55.504
658432 94.316 -52.998 54.870
658433 94.345 -53.016 54.864
658434 94.375 -53.035 54.858
658435 94.404 -53.053 54.853
658436 94.433 -53.071 54.850
658437 94.463 -53.089 54.846
658438 94.492 -53.107 54.844
658439 94.522 -53.126 54.842
658440 94.551 -53.144 54.840
658441 94.580 -53.162 54.839
658442 94.610 -53.180 54.838
658443 94.639 -53.198 54.838
658444 94.669 -53.217 54.837
658445 94.698 -53.236 54.836
658446 94.728 -53.254 54.837
658447 94.756 -53.265 54.837
658448 94.783 -53.271 54.836
658449 94.811 -53.277 54.835
658450 94.838 -53.283 54.834
658451 94.861 -53.294 54.832
658452 94.880 -53.308 54.831
658453 94.898 -53.321 54.829
658454 94.917 -53.334 54.826
658455 94.935 -53.348 54.824
658456 94.954 -53.361 54.823
658457 94.973 -53.374 54.826
658458 94.989 -53.390 54.830
658459 95.002 -53.408 54.832
658460 95.018 -53.427 54.833
658461 95.040 -53.446 54.834
658462 95.062 -53.465 54.835
658463 95.084 -53.484 54.832
658464 95.106 -53.503 54.829
658465 95.127 -53.522 54.823
658466 95.149 -53.541 54.810
658467 95.171 -53.561 54.791
658468 95.193 -53.580 54.768
658469 95.215 -53.599 54.740
658470 95.237 -53.618 54.715
658471 95.259 -53.637 54.694
658472 95.280 -53.656 54.674
658473 95.302 -53.675 54.659
658474 95.323 -53.693 54.651
658475 95.345 -53.712 54.646
658476 95.366 -53.729 54.635
658477 95.389 -53.745 54.615
658478 95.413 -53.760 54.591
658479 95.436 -53.775 54.562
658480 95.460 -53.791 54.531
658481 95.485 -53.806 54.496
658482 95.510 -53.822 54.456
658483 95.535 -53.837 54.412
658484 95.553 -53.846 54.368
658485 95.568 -53.854 54.325
658486 95.584 -53.863 54.285
658487 95.600 -53.871 54.246
658488 95.615 -53.879 54.206
658489 95.631 -53.888 54.164
658490 95.653 -53.895 54.125
658491 95.679 -53.902 54.091
658492 95.704 -53.910 54.060
658493 95.729 -53.916 54.032
658494 95.754 -53.923 54.002
658495 95.779 -53.930 53.971
658496 95.805 -53.938 53.943
658497 95.830 -53.944 53.916
658498 95.855 -53.951 53.892
658499 95.881 -53.958 53.868
658500 95.906 -53.965 53.841
658501 95.932 -53.972 53.811
658502 95.959 -53.985 53.781
658503 95.987 -53.997 53.750
658504 96.015 -54.010 53.721
658505 96.042 -54.022 53.690
658506 96.070 -54.034 53.657
658507 96.098 -54.046 53.623
658508 96.125 -54.059 53.590
658509 96.153 -54.071 53.560
658510 96.181 -54.083 53.531
658511 96.208 -54.095 53.500
658512 96.236 -54.107 53.468
658513 96.263 -54.122 53.436
658514 96.289 -54.140 53.404
658515 96.314 -54.158 53.374
658516 96.340 -54.176 53.345
658517 96.366 -54.192 53.315
658518 96.393 -54.203 53.286
658519 96.422 -54.214 53.258
658520 96.453 -54.224 53.228
658521 96.483 -54.235 53.195
658522 96.513 -54.245 53.159
658523 96.543 -54.256 53.119
658524 96.571 -54.268 53.080
658525 96.600 -54.281 53.036
658526 96.628 -54.293 52.992
658527 96.658 -54.305 52.952
658528 96.688 -54.314 52.917
658529 96.718 -54.324 52.885
658530 96.748 -54.334 52.854
658531 96.779 -54.344 52.824
658532 96.809 -54.354 52.791
658533 96.838 -54.362 52.755
658534 96.868 -54.369 52.715
658535 96.896 -54.376 52.670
658536 96.926 -54.383 52.623
658537 96.954 -54.389 52.573
658538 96.983 -54.396 52.523
658539 97.013 -54.403 52.470
658540 97.041 -54.410 52.418
658541 97.071 -54.417 52.369
658542 97.099 -54.424 52.321
658543 97.129 -54.430 52.275
658544 97.157 -54.437 52.226
658545 97.184 -54.446 52.174
658546 97.215 -54.453 52.120
658547 97.245 -54.461 52.066
658548 97.275 -54.468 52.012
658549 97.305 -54.477 51.960
658550 97.335 -54.484 51.908
658551 97.366 -54.492 51.858
658552 97.391 -54.495 51.806
658553 97.416 -54.496 51.753
658554 97.441 -54.500 51.699
658555 97.469 -54.507 51.646
658556 97.497 -54.513 51.592
658557 97.525 -54.519 51.537
658558 97.553 -54.525 51.482
658559 97.581 -54.531 51.430
658560 97.607 -54.531 51.373
658561 97.634 -54.530 51.313
658562 97.660 -54.530 51.252
658563 97.686 -54.529 51.192
658564 97.713 -54.528 51.132
658565 97.739 -54.527 51.073
658566 97.766 -54.527 51.013
658567 97.792 -54.526 50.953
658568 97.819 -54.525 50.891
658569 97.845 -54.526 50.826
658570 97.872 -54.527 50.756
658571 97.899 -54.527 50.684
658572 97.926 -54.527 50.612
658573 97.952 -54.527 50.542
658574 97.979 -54.528 50.472
658575 98.006 -54.528 50.402
658576 98.033 -54.529 50.334
658577 98.060 -54.529 50.267
658578 98.087 -54.529 50.202
658579 98.113 -54.530 50.139
658580 98.140 -54.530 50.078
658581 98.167 -54.530 50.018
658582 98.194 -54.530 49.961
658583 98.221 -54.531 49.906
659006 90.480 -50.727 56.877
659007 90.513 -50.746 56.869
659008 90.546 -50.765 56.861
659009 90.579 -50.785 56.852
659010 90.612 -50.804 56.842
659011 90.645 -50.823 56.829
659012 90.678 -50.842 56.816
659013 90.712 -50.862 56.802
659014 90.744 -50.878 56.788
659015 90.777 -50.886 56.771
659016 90.809 -50.894 56.754
659017 90.841 -50.901 56.736
659018 90.873 -50.910 56.716
659019 90.898 -50.920 56.698
659020 90.909 -50.936 56.679
659021 90.921 -50.952 56.660
659034 91.234 -51.149 56.796
659035 91.270 -51.164 56.800
659036 91.306 -51.179 56.806
659037 91.342 -51.193 56.811
659038 91.378 -51.208 56.817
659039 91.414 -51.222 56.824
659040 91.448 -51.236 56.831
659041 91.480 -51.251 56.839
659042 91.511 -51.265 56.848
659043 91.542 -51.280 56.854
659044 91.573 -51.294 56.860
659045 91.604 -51.309 56.866
659046 91.635 -51.323 56.872
659047 91.666 -51.337 56.878
659048 91.698 -51.352 56.881
659049 91.729 -51.366 56.882
659050 91.760 -51.380 56.883
659051 91.791 -51.394 56.885
659052 91.823 -51.403 56.886
659053 91.856 -51.410 56.888
659054 91.888 -51.418 56.889
659055 91.921 -51.425 56.888
659056 91.953 -51.431 56.886
659057 91.986 -51.439 56.884
659058 92.018 -51.446 56.882
659059 92.052 -51.462 56.882
659060 92.086 -51.481 56.880
659061 92.120 -51.500 56.877
659062 92.154 -51.520 56.874
659063 92.188 -51.539 56.868
659064 92.222 -51.558 56.864
659065 92.256 -51.578 56.858
659066 92.290 -51.597 56.851
659067 92.324 -51.616 56.843
659068 92.358 -51.635 56.834
659069 92.392 -51.654 56.822
659070 92.426 -51.673 56.809
659071 92.456 -51.691 56.794
659072 92.478 -51.707 56.775
659073 92.498 -51.722 56.757
659074 92.519 -51.738 56.738
659075 92.541 -51.754 56.718
659076 92.565 -51.771 56.697
659077 92.595 -51.791 56.671
659078 92.626 -51.810 56.641
659079 92.657 -51.830 56.609
659080 92.688 -51.850 56.576
659081 92.718 -51.870 56.542
659082 92.748 -51.891 56.508
659083 92.779 -51.920 56.471
659084 92.808 -51.949 56.436
659085 92.839 -51.978 56.398
659086 92.868 -52.007 56.359
659087 92.899 -52.036 56.319
659088 92.928 -52.064 56.281
659089 92.958 -52.093 56.242
659090 92.988 -52.122 56.203
659091 93.019 -52.151 56.166
659092 93.059 -52.181 56.128
659093 93.099 -52.211 56.089
659094 93.139 -52.242 56.050
659095 93.177 -52.269 56.011
659096 93.207 -52.288 55.971
659097 93.237 -52.306 55.930
659098 93.268 -52.325 55.885
659099 93.299 -52.343 55.839
659100 93.330 -52.362 55.792
659101 93.360 -52.380 55.745
659102 93.391 -52.399 55.697
659103 93.421 -52.418 55.652
659104 93.452 -52.437 55.606
659105 93.483 -52.456 55.562
659106 93.513 -52.474 55.519
659134 94.356 -52.991 54.863
659135 94.386 -53.010 54.857
659136 94.415 -53.028 54.851
659137 94.445 -53.046 54.847
659138 94.474 -53.064 54.842
659139 94.503 -53.082 54.839
659140 94.533 -53.101 54.837
659141 94.562 -53.119 54.833
659142 94.591 -53.137 54.831
659143 94.621 -53.155 54.829
659144 94.650 -53.174 54.828
659145 94.679 -53.192 54.827
659146 94.709 -53.211 54.827
659147 94.738 -53.229 54.826
659148 94.767 -53.247 54.826
659149 94.795 -53.253 54.825
659150 94.823 -53.259 54.824
659151 94.850 -53.265 54.824
659152 94.878 -53.272 54.822
659153 94.903 -53.280 54.821
659154 94.922 -53.293 54.819
659155 94.941 -53.307 54.817
659156 94.959 -53.320 54.815
659157 94.978 -53.333 54.815
659158 94.996 -53.347 54.819
659159 95.010 -53.365 54.824
659160 95.023 -53.383 54.826
659161 95.037 -53.402 54.828
659162 95.050 -53.420 54.829
659163 95.064 -53.438 54.829
659164 95.086 -53.457 54.826
659165 95.108 -53.476 54.821
659166 95.130 -53.495 54.813
659167 95.152 -53.514 54.798
659168 95.174 -53.534 54.777
659169 95.195 -53.552 54.753
659170 95.217 -53.571 54.723
659171 95.239 -53.590 54.698
659172 95.261 -53.609 54.679
659173 95.283 -53.629 54.664
659174 95.305 -53.648 54.654
659175 95.330 -53.669 54.649
659176 95.357 -53.691 54.645
659177 95.382 -53.710 54.634
659178 95.406 -53.725 54.615
659179 95.429 -53.741 54.589
659180 95.453 -53.756 54.558
659181 95.478 -53.772 54.525
659182 95.502 -53.787 54.489
659183 95.528 -53.802 54.448
659184 95.552 -53.817 54.403
659185 95.571 -53.828 54.358
659186 95.587 -53.836 54.317
659187 95.602 -53.844 54.278
659188 95.618 -53.853 54.241
659189 95.633 -53.861 54.204
659190 95.649 -53.869 54.165
659191 95.665 -53.878 54.129
659192 95.684 -53.885 54.097
659193 95.709 -53.893 54.069
659194 95.735 -53.900 54.041
659195 95.760 -53.907 54.011
659196 95.785 -53.913 53.980
659197 95.811 -53.920 53.951
659198 95.836 -53.928 53.923
659199 95.861 -53.935 53.898
659200 95.886 -53.941 53.873
659201 95.911 -53.948 53.846
659202 95.937 -53.956 53.816
659203 95.964 -53.966 53.785
659204 95.992 -53.979 53.754
659205 96.020 -53.991 53.726
659206 96.047 -54.004 53.695
659207 96.075 -54.016 53.661
659208 96.103 -54.028 53.626
659209 96.130 -54.041 53.593
659210 96.158 -54.053 53.563
659211 96.186 -54.065 53.536
659212 96.214 -54.077 53.506
659213 96.241 -54.089 53.475
659214 96.269 -54.102 53.444
659215 96.297 -54.114 53.414
659216 96.323 -54.130 53.386
659217 96.348 -54.148 53.359
659218 96.374 -54.165 53.332
659219 96.402 -54.176 53.303
659220 96.433 -54.186 53.275
659221 96.463 -54.196 53.244
659222 96.493 -54.207 53.211
659223 96.523 -54.217 53.175
659224 96.553 -54.228 53.138
659225 96.582 -54.240 53.097
659226 96.610 -54.252 53.052
659227 96.640 -54.263 53.008
659228 96.670 -54.273 52.968
659229 96.700 -54.282 52.932
659230 96.731 -54.292 52.900
659231 96.761 -54.301 52.868
659232 96.792 -54.311 52.835
659233 96.822 -54.321 52.800
659234 96.852 -54.330 52.763
659235 96.881 -54.336 52.723
659236 96.909 -54.343 52.678
659237 96.939 -54.350 52.630
659238 96.967 -54.357 52.581
659239 96.997 -54.364 52.532
659240 97.025 -54.370 52.482
659241 97.054 -54.377 52.432
659242 97.083 -54.384 52.383
659243 97.112 -54.391 52.335
659244 97.140 -54.399 52.288
659245 97.165 -54.410 52.239
659246 97.191 -54.417 52.190
659247 97.220 -54.425 52.138
659248 97.250 -54.433 52.085
659249 97.281 -54.440 52.033
659250 97.311 -54.448 51.982
659251 97.341 -54.455 51.931
659252 97.371 -54.464 51.882
659253 97.396 -54.466 51.832
659254 97.421 -54.467 51.782
659255 97.445 -54.470 51.730
659256 97.471 -54.474 51.677
659257 97.499 -54.480 51.623
659258 97.527 -54.486 51.569
659259 97.555 -54.493 51.516
659260 97.582 -54.498 51.464
659261 97.608 -54.497 51.410
659262 97.635 -54.496 51.352
659263 97.661 -54.496 51.293
659264 97.687 -54.495 51.234
659265 97.714 -54.494 51.176
659266 97.740 -54.493 51.118
659267 97.768 -54.493 51.058
659268 97.797 -54.494 51.000
659269 97.825 -54.495 50.940
659270 97.849 -54.496 50.876
659271 97.875 -54.496 50.807
659272 97.901 -54.496 50.737
659273 97.928 -54.497 50.664
659274 97.955 -54.497 50.593
659275 97.982 -54.498 50.521
659276 98.009 -54.498 50.450
659277 98.035 -54.499 50.381
659278 98.063 -54.499 50.313
659279 98.089 -54.499 50.247
659280 98.116 -54.499 50.182
659281 98.143 -54.499 50.119
659282 98.170 -54.500 50.058
659283 98.197 -54.500 50.000
659284 98.223 -54.501 49.944
659706 90.469 -50.680 56.875
659707 90.500 -50.694 56.868
659708 90.533 -50.713 56.862
659709 90.566 -50.732 56.855
659710 90.599 -50.751 56.847
659711 90.632 -50.771 56.838
659712 90.665 -50.790 56.826
659713 90.698 -50.810 56.814
659714 90.731 -50.829 56.801
659715 90.764 -50.845 56.786
659716 90.797 -50.852 56.771
659717 90.829 -50.860 56.754
659718 90.861 -50.868 56.736
659719 90.891 -50.877 56.717
659720 90.903 -50.893 56.699
659721 90.915 -50.909 56.680
659722 90.926 -50.924 56.661
659723 90.938 -50.940 56.643
659735 91.288 -51.119 56.786
659736 91.323 -51.134 56.790
659737 91.359 -51.149 56.795
659738 91.390 -51.163 56.800
659739 91.421 -51.177 56.805
659740 91.452 -51.191 56.811
659741 91.484 -51.205 56.817
659742 91.515 -51.220 56.825
659743 91.546 -51.234 56.833
659744 91.577 -51.249 56.838
659745 91.608 -51.263 56.844
659746 91.639 -51.277 56.849
659747 91.670 -51.292 56.855
659748 91.702 -51.306 56.860
659749 91.733 -51.321 56.865
659750 91.764 -51.335 56.866
659751 91.795 -51.349 56.867
659752 91.826 -51.363 56.869
659753 91.857 -51.378 56.870
659754 91.888 -51.392 56.872
659755 91.921 -51.402 56.874
659756 91.953 -51.409 56.874
659757 91.986 -51.415 56.872
659758 92.018 -51.423 56.870
659759 92.051 -51.430 56.869
659760 92.083 -51.437 56.869
659761 92.117 -51.456 56.868
659762 92.151 -51.475 56.864
659763 92.185 -51.494 56.860
659764 92.219 -51.512 56.856
659765 92.253 -51.532 56.852
659766 92.287 -51.551 56.848
659767 92.321 -51.570 56.842
659768 92.355 -51.590 56.834
659769 92.389 -51.609 56.825
659770 92.423 -51.628 56.812
659771 92.448 -51.645 56.798
659772 92.470 -51.660 56.782
659773 92.491 -51.676 56.764
659774 92.511 -51.691 56.746
659775 92.532 -51.707 56.728
659776 92.555 -51.723 56.709
659777 92.586 -51.744 56.688
659778 92.617 -51.763 56.663
659779 92.647 -51.783 56.634
659780 92.677 -51.803 56.604
659781 92.708 -51.823 56.573
659782 92.739 -51.842 56.542
659783 92.770 -51.865 56.510
659784 92.799 -51.894 56.476
659785 92.830 -51.923 56.441
659786 92.859 -51.951 56.405
659787 92.889 -51.980 56.368
659788 92.919 -52.009 56.330
659789 92.949 -52.038 56.292
659790 92.979 -52.067 56.253
659791 93.009 -52.095 56.214
659792 93.039 -52.124 56.177
659793 93.071 -52.154 56.139
659794 93.111 -52.184 56.100
659795 93.151 -52.214 56.061
659796 93.191 -52.244 56.022
659797 93.222 -52.264 55.980
659798 93.253 -52.283 55.940
659799 93.284 -52.301 55.896
659800 93.314 -52.320 55.851
659801 93.345 -52.339 55.804
659802 93.375 -52.357 55.758
659803 93.406 -52.376 55.710
659804 93.437 -52.394 55.666
659805 93.468 -52.413 55.620
659806 93.498 -52.431 55.576
659807 93.529 -52.450 55.533
659836 94.396 -52.985 54.856
659837 94.426 -53.003 54.849
659838 94.455 -53.021 54.844
659839 94.484 -53.039 54.838
659840 94.514 -53.057 54.835
659841 94.543 -53.076 54.831
659842 94.572 -53.094 54.827
659843 94.602 -53.112 54.824
659844 94.631 -53.131 54.821
659845 94.661 -53.149 54.819
659846 94.690 -53.167 54.818
659847 94.719 -53.186 54.817
659848 94.749 -53.204 54.817
659849 94.778 -53.222 54.815
659850 94.807 -53.235 54.814
659851 94.834 -53.241 54.814
659852 94.862 -53.248 54.814
659853 94.890 -53.254 54.813
659854 94.917 -53.260 54.811
659855 94.945 -53.266 54.810
659856 94.964 -53.279 54.808
659857 94.983 -53.292 54.807
659858 95.001 -53.306 54.807
659859 95.017 -53.322 54.810
659860 95.031 -53.340 54.815
659861 95.044 -53.358 54.818
659862 95.057 -53.377 54.818
659863 95.071 -53.394 54.818
659864 95.085 -53.412 54.816
659865 95.098 -53.430 54.811
659866 95.111 -53.449 54.805
659867 95.132 -53.468 54.796
659868 95.154 -53.487 54.783
659869 95.176 -53.506 54.764
659870 95.198 -53.525 54.738
659871 95.220 -53.544 54.706
659872 95.242 -53.563 54.685
659873 95.266 -53.584 54.667
659874 95.293 -53.606 54.655
659875 95.320 -53.629 54.648
659876 95.347 -53.651 54.644
659877 95.374 -53.674 54.640
659878 95.399 -53.691 54.630
659879 95.422 -53.706 54.611
659880 95.446 -53.722 54.585
659881 95.470 -53.737 54.551
659882 95.495 -53.752 54.515
659883 95.520 -53.767 54.477
659884 95.545 -53.782 54.436
659885 95.569 -53.797 54.391
659886 95.590 -53.809 54.347
659887 95.605 -53.818 54.307
659888 95.621 -53.826 54.271
659889 95.636 -53.834 54.236
659890 95.652 -53.843 54.201
659891 95.668 -53.851 54.166
659892 95.683 -53.860 54.133
659893 95.699 -53.868 54.103
659894 95.715 -53.875 54.076
659895 95.740 -53.883 54.049
659896 95.765 -53.890 54.019
659897 95.791 -53.897 53.987
659898 95.816 -53.903 53.956
659899 95.841 -53.911 53.929
659900 95.867 -53.918 53.902
659901 95.892 -53.925 53.876
659902 95.917 -53.932 53.849
659903 95.943 -53.939 53.819
659904 95.969 -53.948 53.789
659905 95.996 -53.960 53.759
659906 96.024 -53.973 53.731
659907 96.052 -53.985 53.700
659908 96.080 -53.997 53.666
659909 96.107 -54.010 53.631
659910 96.135 -54.022 53.596
659911 96.163 -54.035 53.565
659912 96.190 -54.047 53.539
659913 96.218 -54.059 53.512
659914 96.246 -54.071 53.482
659915 96.273 -54.084 53.453
659916 96.301 -54.096 53.425
659917 96.329 -54.108 53.398
659918 96.357 -54.121 53.373
659919 96.382 -54.138 53.350
659920 96.412 -54.148 53.321
659921 96.442 -54.158 53.291
659922 96.472 -54.169 53.259
659923 96.503 -54.179 53.224
659924 96.533 -54.189 53.190
659925 96.563 -54.200 53.154
659926 96.593 -54.211 53.112
659927 96.623 -54.220 53.068
659928 96.653 -54.230 53.024
659929 96.683 -54.240 52.983
659930 96.714 -54.249 52.947
659931 96.744 -54.259 52.913
659932 96.774 -54.269 52.879
659933 96.805 -54.279 52.844
659934 96.835 -54.289 52.807
659935 96.865 -54.297 52.770
659936 96.893 -54.304 52.730
659937 96.923 -54.311 52.685
659938 96.951 -54.317 52.638
659939 96.980 -54.324 52.591
659940 97.010 -54.331 52.542
659941 97.038 -54.338 52.493
659942 97.068 -54.345 52.445
659943 97.096 -54.352 52.397
659944 97.122 -54.361 52.349
659945 97.148 -54.372 52.300
659946 97.173 -54.381 52.253
659947 97.199 -54.389 52.205
659948 97.226 -54.397 52.154
659949 97.256 -54.405 52.103
659950 97.286 -54.412 52.052
659951 97.316 -54.420 52.003
659952 97.347 -54.427 51.953
659953 97.377 -54.435 51.906
659954 97.401 -54.437 51.857
659955 97.426 -54.439 51.806
659956 97.450 -54.441 51.755
659957 97.474 -54.443 51.703
659958 97.500 -54.447 51.649
659959 97.528 -54.453 51.596
659960 97.556 -54.460 51.545
659961 97.584 -54.464 51.495
659962 97.610 -54.463 51.442
659963 97.636 -54.462 51.387
659964 97.663 -54.461 51.332
659965 97.689 -54.461 51.275
659966 97.716 -54.461 51.218
659967 97.745 -54.461 51.160
659968 97.775 -54.462 51.102
659969 97.804 -54.463 51.044
659970 97.829 -54.464 50.984
659971 97.853 -54.465 50.922
659972 97.877 -54.466 50.856
659973 97.904 -54.466 50.790
659974 97.931 -54.467 50.716
659975 97.958 -54.467 50.642
659976 97.985 -54.467 50.570
659977 98.011 -54.467 50.499
659978 98.038 -54.468 50.429
659979 98.065 -54.468 50.361
659980 98.092 -54.469 50.293
659981 98.119 -54.469 50.227
659982 98.146 -54.470 50.163
659983 98.172 -54.470 50.100
659984 98.199 -54.470 50.040
659985 98.226 -54.471 49.983
659986 98.253 -54.471 49.926
660407 90.487 -50.644 56.864
660408 90.521 -50.665 56.859
660409 90.555 -50.685 56.854
660410 90.586 -50.699 56.848
660411 90.619 -50.718 56.842
660412 90.652 -50.738 56.832
660413 90.685 -50.757 56.823
660414 90.718 -50.776 56.811
660415 90.751 -50.795 56.799
660416 90.784 -50.811 56.785
660417 90.816 -50.819 56.770
660418 90.849 -50.827 56.754
660419 90.881 -50.835 56.735
660420 90.896 -50.849 56.717
660421 90.908 -50.865 56.699
660422 90.920 -50.881 56.680
660423 90.931 -50.897 56.661
660424 90.943 -50.913 56.642
660425 90.955 -50.929 56.625
660437 91.363 -51.103 56.779
660438 91.394 -51.117 56.785
660439 91.425 -51.132 56.789
660440 91.456 -51.146 56.793
660441 91.488 -51.161 56.798
660442 91.519 -51.174 56.804
660443 91.550 -51.189 56.811
660444 91.581 -51.203 56.819
660445 91.612 -51.218 56.823
660446 91.643 -51.232 56.829
660447 91.674 -51.246 56.834
660448 91.706 -51.261 56.839
660449 91.737 -51.275 56.844
660450 91.768 -51.290 56.849
660451 91.799 -51.304 56.851
660452 91.830 -51.318 56.852
660453 91.861 -51.333 56.853
660454 91.892 -51.346 56.855
660455 91.924 -51.361 56.857
660456 91.954 -51.375 56.859
660457 91.986 -51.390 56.859
660458 92.018 -51.399 56.858
660459 92.050 -51.407 56.856
660460 92.083 -51.414 56.855
660461 92.115 -51.421 56.856
660462 92.148 -51.429 56.854
660463 92.182 -51.448 56.851
660464 92.216 -51.468 56.847
660465 92.250 -51.487 56.844
660466 92.284 -51.506 56.841
660467 92.318 -51.525 56.837
660468 92.352 -51.544 56.831
660469 92.386 -51.563 56.824
660470 92.420 -51.582 56.815
660471 92.440 -51.598 56.802
660472 92.461 -51.614 56.785
660473 92.483 -51.629 56.769
660474 92.503 -51.645 56.752
660475 92.524 -51.660 56.734
660476 92.546 -51.676 56.717
660477 92.576 -51.696 56.698
660478 92.606 -51.716 56.677
660479 92.637 -51.735 56.656
660480 92.668 -51.755 56.628
660481 92.699 -51.776 56.599
660482 92.729 -51.795 56.570
660483 92.760 -51.815 56.541
660484 92.790 -51.838 56.511
660485 92.820 -51.867 56.480
660486 92.850 -51.896 56.447
660487 92.880 -51.925 56.412
660488 92.910 -51.954 56.377
660489 92.940 -51.982 56.341
660490 92.970 -52.011 56.303
660491 93.000 -52.040 56.266
660492 93.030 -52.069 56.228
660493 93.060 -52.098 56.191
660494 93.090 -52.127 56.153
660495 93.124 -52.156 56.113
660496 93.163 -52.186 56.072
660497 93.203 -52.217 56.033
660498 93.237 -52.240 55.992
660499 93.268 -52.259 55.951
660500 93.299 -52.277 55.908
660501 93.330 -52.296 55.862
660502 93.360 -52.315 55.816
660503 93.391 -52.333 55.771
660504 93.421 -52.352 55.724
660505 93.452 -52.371 55.679
660506 93.483 -52.389 55.634
660507 93.513 -52.408 55.590
660508 93.544 -52.427 55.547
660537 94.407 -52.960 54.854
660538 94.436 -52.978 54.846
660539 94.466 -52.996 54.840
660540 94.495 -53.014 54.834
660541 94.525 -53.032 54.829
660542 94.554 -53.051 54.825
660543 94.583 -53.069 54.821
660544 94.613 -53.088 54.816
660545 94.642 -53.106 54.813
660546 94.672 -53.124 54.811
660547 94.701 -53.142 54.809
660548 94.731 -53.161 54.808
660549 94.760 -53.179 54.807
660550 94.789 -53.197 54.806
660551 94.819 -53.215 54.805
660552 94.846 -53.223 54.804
660553 94.874 -53.230 54.804
660554 94.901 -53.236 54.803
660555 94.929 -53.242 54.801
660556 94.957 -53.248 54.800
660557 94.984 -53.254 54.800
660558 95.006 -53.264 54.801
660559 95.024 -53.279 54.801
660560 95.038 -53.297 54.802
660561 95.051 -53.315 54.805
660562 95.065 -53.333 54.807
660563 95.078 -53.351 54.806
660564 95.092 -53.369 54.804
660565 95.105 -53.387 54.800
660566 95.119 -53.405 54.793
660567 95.132 -53.424 54.786
660568 95.145 -53.442 54.777
660569 95.159 -53.460 54.765
660570 95.178 -53.478 54.746
660571 95.202 -53.499 54.722
660572 95.229 -53.521 54.693
660573 95.256 -53.543 54.673
660574 95.283 -53.566 54.659
660575 95.310 -53.589 54.650
660576 95.337 -53.611 54.645
660577 95.364 -53.634 54.641
660578 95.391 -53.656 54.634
660579 95.415 -53.672 54.627
660580 95.439 -53.687 54.606
660581 95.462 -53.703 54.579
660582 95.487 -53.718 54.544
660583 95.512 -53.733 54.505
660584 95.537 -53.748 54.466
660585 95.562 -53.763 54.422
660586 95.587 -53.778 54.377
660587 95.608 -53.791 54.334
660588 95.624 -53.799 54.296
660589 95.639 -53.807 54.262
660590 95.655 -53.816 54.230
660591 95.671 -53.824 54.199
660592 95.686 -53.832 54.167
660593 95.702 -53.841 54.138
660594 95.718 -53.849 54.109
660595 95.733 -53.857 54.082
660596 95.749 -53.866 54.055
660597 95.771 -53.873 54.026
660598 95.796 -53.880 53.993
660599 95.821 -53.887 53.959
660600 95.847 -53.894 53.933
660601 95.872 -53.901 53.905
660602 95.897 -53.908 53.878
660603 95.922 -53.915 53.850
660604 95.948 -53.922 53.820
660605 95.974 -53.930 53.791
660606 96.002 -53.942 53.761
660607 96.029 -53.955 53.733
660608 96.057 -53.967 53.703
660609 96.085 -53.979 53.669
660610 96.112 -53.991 53.633
660611 96.140 -54.004 53.598
660612 96.168 -54.016 53.567
660613 96.195 -54.029 53.541
660614 96.223 -54.041 53.516
660615 96.251 -54.053 53.488
660616 96.278 -54.066 53.460
660617 96.306 -54.078 53.434
660618 96.334 -54.090 53.409
660619 96.362 -54.102 53.385
660620 96.391 -54.111 53.363
660621 96.421 -54.122 53.334
660622 96.452 -54.132 53.303
660623 96.482 -54.143 53.270
660624 96.512 -54.153 53.236
660625 96.542 -54.164 53.201
660626 96.572 -54.172 53.166
660627 96.604 -54.179 53.126
660628 96.636 -54.188 53.082
660629 96.666 -54.198 53.038
660630 96.696 -54.207 52.997
660631 96.727 -54.217 52.959
660632 96.757 -54.226 52.923
660633 96.787 -54.236 52.887
660634 96.817 -54.246 52.850
660635 96.848 -54.256 52.813
660636 96.878 -54.265 52.776
660637 96.906 -54.272 52.736
660638 96.936 -54.279 52.692
660639 96.964 -54.286 52.646
660640 96.994 -54.292 52.600
660641 97.022 -54.299 52.553
660642 97.051 -54.306 52.505
660643 97.079 -54.314 52.458
660644 97.104 -54.324 52.410
660645 97.130 -54.334 52.362
660646 97.155 -54.345 52.314
660647 97.181 -54.353 52.268
660648 97.207 -54.361 52.219
660649 97.233 -54.369 52.170
660650 97.261 -54.377 52.120
660651 97.292 -54.384 52.071
660652 97.322 -54.392 52.023
660653 97.352 -54.399 51.975
660654 97.382 -54.407 51.929
660655 97.406 -54.409 51.880
660656 97.431 -54.411 51.829
660657 97.455 -54.413 51.778
660658 97.479 -54.414 51.726
660659 97.504 -54.417 51.674
660660 97.530 -54.421 51.622
660661 97.558 -54.427 51.573
660662 97.585 -54.430 51.524
660663 97.611 -54.429 51.473
660664 97.638 -54.429 51.420
660665 97.665 -54.428 51.366
660666 97.694 -54.429 51.311
660667 97.724 -54.430 51.255
660668 97.753 -54.430 51.198
660669 97.783 -54.432 51.140
660670 97.810 -54.433 51.083
660671 97.834 -54.433 51.025
660672 97.858 -54.434 50.964
660673 97.881 -54.436 50.901
660674 97.907 -54.436 50.836
660675 97.934 -54.436 50.765
660676 97.960 -54.436 50.692
660677 97.987 -54.437 50.619
660678 98.014 -54.437 50.547
660679 98.041 -54.438 50.477
660680 98.068 -54.438 50.407
660681 98.094 -54.439 50.339
660682 98.122 -54.439 50.272
660683 98.148 -54.439 50.207
660684 98.175 -54.439 50.143
660685 98.202 -54.440 50.081
660686 98.229 -54.440 50.022
660687 98.255 -54.440 49.965
661107 90.469 -50.588 56.857
661108 90.504 -50.608 56.854
661109 90.539 -50.629 56.849
661110 90.574 -50.649 56.844
661111 90.608 -50.670 56.841
661112 90.639 -50.685 56.836
661113 90.672 -50.704 56.827
661114 90.705 -50.723 56.818
661115 90.738 -50.743 56.808
661116 90.771 -50.762 56.796
661117 90.804 -50.777 56.783
661118 90.836 -50.785 56.768
661119 90.869 -50.794 56.753
661120 90.889 -50.806 56.736
661121 90.901 -50.822 56.718
661122 90.913 -50.838 56.699
661123 90.925 -50.853 56.680
661124 90.937 -50.869 56.661
661125 90.962 -50.884 56.642
661126 90.999 -50.898 56.625
661138 91.398 -51.072 56.770
661139 91.429 -51.086 56.775
661140 91.460 -51.101 56.778
661141 91.492 -51.115 56.781
661142 91.523 -51.130 56.786
661143 91.554 -51.144 56.791
661144 91.585 -51.158 56.798
661145 91.616 -51.172 56.805
661146 91.647 -51.186 56.809
661147 91.678 -51.201 56.814
661148 91.710 -51.215 56.818
661149 91.740 -51.230 56.823
661150 91.772 -51.244 56.829
661151 91.803 -51.258 56.833
661152 91.834 -51.273 56.835
661153 91.865 -51.287 56.836
661154 91.896 -51.302 56.838
661155 91.927 -51.315 56.841
661156 91.958 -51.330 56.843
661157 91.990 -51.344 56.845
661158 92.021 -51.359 56.844
661159 92.052 -51.373 56.843
661160 92.083 -51.387 56.842
661161 92.115 -51.398 56.842
661162 92.147 -51.405 56.842
661163 92.180 -51.412 56.841
661164 92.213 -51.422 56.838
661165 92.247 -51.441 56.835
661166 92.281 -51.461 56.833
661167 92.315 -51.480 56.829
661168 92.349 -51.499 56.825
661169 92.383 -51.519 56.819
661170 92.412 -51.536 56.811
661171 92.433 -51.552 56.803
661172 92.453 -51.567 56.789
661173 92.475 -51.583 56.772
661174 92.496 -51.598 56.755
661175 92.516 -51.614 56.738
661176 92.537 -51.629 56.721
661177 92.566 -51.649 56.703
661178 92.597 -51.669 56.686
661179 92.628 -51.688 56.667
661180 92.658 -51.708 56.645
661181 92.689 -51.728 56.620
661182 92.720 -51.747 56.593
661183 92.751 -51.767 56.566
661184 92.781 -51.788 56.539
661185 92.811 -51.812 56.511
661186 92.841 -51.841 56.483
661187 92.871 -51.870 56.453
661188 92.901 -51.898 56.420
661189 92.931 -51.927 56.386
661190 92.961 -51.956 56.352
661191 92.991 -51.985 56.315
661192 93.021 -52.014 56.279
661193 93.051 -52.043 56.243
661194 93.081 -52.072 56.207
661195 93.111 -52.101 56.168
661196 93.141 -52.130 56.127
661197 93.176 -52.159 56.085
661198 93.215 -52.189 56.045
661199 93.253 -52.217 56.004
661200 93.283 -52.235 55.962
661201 93.314 -52.254 55.919
661202 93.345 -52.272 55.875
661203 93.375 -52.291 55.828
661204 93.406 -52.310 55.782
661205 93.437 -52.328 55.737
661206 93.467 -52.347 55.693
661207 93.498 -52.365 55.648
661208 93.529 -52.384 55.604
661209 93.559 -52.402 55.561
661239 94.448 -52.953 54.843
661240 94.477 -52.971 54.836
661241 94.506 -52.989 54.830
661242 94.536 -53.007 54.824
661243 94.565 -53.026 54.819
661244 94.594 -53.045 54.814
661245 94.624 -53.063 54.809
661246 94.653 -53.081 54.805
661247 94.682 -53.099 54.802
661248 94.712 -53.117 54.800
661249 94.741 -53.136 54.798
661250 94.770 -53.154 54.797
661251 94.800 -53.172 54.797
661252 94.829 -53.190 54.795
661253 94.858 -53.205 54.794
661254 94.886 -53.211 54.794
661255 94.913 -53.217 54.793
661256 94.941 -53.223 54.792
661257 94.968 -53.230 54.791
661258 94.996 -53.236 54.791
661259 95.024 -53.242 54.792
661260 95.046 -53.255 54.792
661261 95.060 -53.273 54.793
661262 95.074 -53.292 54.793
661263 95.088 -53.311 54.793
661264 95.101 -53.329 54.791
661265 95.115 -53.347 54.786
661266 95.129 -53.366 54.780
661267 95.142 -53.384 54.774
661268 95.156 -53.402 54.766
661269 95.170 -53.421 54.757
661270 95.184 -53.440 54.743
661271 95.198 -53.458 54.725
661272 95.219 -53.481 54.704
661273 95.246 -53.503 54.682
661274 95.273 -53.526 54.665
661275 95.300 -53.548 54.655
661276 95.327 -53.571 54.650
661277 95.354 -53.593 54.645
661278 95.381 -53.615 54.640
661279 95.408 -53.637 54.634
661280 95.431 -53.653 54.624
661281 95.455 -53.668 54.602
661282 95.479 -53.683 54.573
661283 95.505 -53.698 54.538
661284 95.529 -53.713 54.499
661285 95.554 -53.728 54.457
661286 95.580 -53.744 54.414
661287 95.604 -53.759 54.371
661288 95.627 -53.772 54.328
661289 95.642 -53.781 54.290
661290 95.658 -53.789 54.258
661291 95.674 -53.797 54.231
661292 95.689 -53.806 54.203
661293 95.705 -53.814 54.173
661294 95.721 -53.822 54.145
661295 95.736 -53.831 54.114
661296 95.752 -53.839 54.086
661297 95.767 -53.847 54.059
661298 95.783 -53.856 54.030
661299 95.802 -53.863 53.996
661300 95.827 -53.870 53.962
661301 95.852 -53.877 53.934
661302 95.878 -53.885 53.906
661303 95.903 -53.891 53.879
661304 95.928 -53.898 53.849
661305 95.953 -53.905 53.819
661306 95.979 -53.913 53.789
661307 96.006 -53.924 53.761
661308 96.034 -53.937 53.732
661309 96.061 -53.949 53.704
661310 96.089 -53.961 53.670
661311 96.117 -53.973 53.634
661312 96.145 -53.986 53.601
661313 96.172 -53.998 53.569
661314 96.200 -54.010 53.543
661315 96.228 -54.022 53.519
661316 96.255 -54.035 53.493
661317 96.283 -54.048 53.467
661318 96.311 -54.060 53.442
661319 96.339 -54.070 53.420
661320 96.370 -54.077 53.397
661321 96.400 -54.086 53.374
661322 96.430 -54.097 53.346
661323 96.460 -54.107 53.314
661324 96.490 -54.118 53.280
661325 96.521 -54.129 53.246
661326 96.550 -54.136 53.210
661327 96.580 -54.143 53.174
661328 96.612 -54.151 53.133
661329 96.645 -54.158 53.092
661330 96.678 -54.166 53.050
661331 96.709 -54.175 53.009
661332 96.740 -54.185 52.970
661333 96.770 -54.194 52.932
661334 96.800 -54.204 52.894
661335 96.830 -54.214 52.859
661336 96.861 -54.223 52.822
661337 96.890 -54.232 52.784
661338 96.920 -54.239 52.745
661339 96.949 -54.246 52.703
661340 96.977 -54.253 52.657
661341 97.007 -54.260 52.611
661342 97.035 -54.267 52.566
661343 97.061 -54.276 52.520
661344 97.087 -54.286 52.472
661345 97.112 -54.297 52.425
661346 97.138 -54.307 52.377
661347 97.163 -54.317 52.329
661348 97.189 -54.325 52.282
661349 97.215 -54.333 52.233
661350 97.241 -54.340 52.185
661351 97.267 -54.348 52.137
661352 97.297 -54.356 52.089
661353 97.327 -54.364 52.042
661354 97.358 -54.371 51.995
661355 97.387 -54.378 51.949
661356 97.412 -54.380 51.900
661357 97.436 -54.383 51.850
661358 97.460 -54.384 51.800
661359 97.485 -54.386 51.748
661360 97.509 -54.388 51.698
661361 97.533 -54.390 51.647
661362 97.559 -54.394 51.599
661363 97.586 -54.396 51.551
661364 97.613 -54.395 51.502
661365 97.643 -54.396 51.450
661366 97.672 -54.398 51.396
661367 97.702 -54.399 51.343
661368 97.731 -54.399 51.288
661369 97.760 -54.400 51.232
661370 97.790 -54.401 51.175
661371 97.814 -54.402 51.119
661372 97.839 -54.403 51.062
661373 97.862 -54.404 51.003
661374 97.886 -54.405 50.942
661375 97.910 -54.406 50.878
661376 97.936 -54.406 50.810
661377 97.963 -54.406 50.739
661378 97.990 -54.407 50.666
661379 98.017 -54.407 50.594
661380 98.043 -54.408 50.523
661381 98.070 -54.408 50.453
661382 98.097 -54.408 50.384
661383 98.124 -54.408 50.317
661384 98.151 -54.409 50.250
661385 98.178 -54.409 50.185
661386 98.205 -54.410 50.123
661387 98.231 -54.410 50.062
661388 98.258 -54.410 50.004
661389 98.285 -54.411 49.948
661808 90.487 -50.551 56.844
661809 90.521 -50.572 56.841
661810 90.556 -50.592 56.838
661811 90.591 -50.613 56.835
661812 90.625 -50.634 56.833
661813 90.660 -50.654 56.828
661814 90.692 -50.671 56.822
661815 90.725 -50.691 56.814
661816 90.758 -50.710 56.804
661817 90.792 -50.729 56.793
661818 90.824 -50.744 56.780
661819 90.857 -50.752 56.766
661820 90.883 -50.763 56.752
661821 90.894 -50.778 56.735
661822 90.906 -50.794 56.717
661823 90.918 -50.810 56.699
661824 90.944 -50.825 56.680
661825 90.980 -50.839 56.661
661826 91.016 -50.854 56.642
661827 91.052 -50.869 56.624
661841 91.496 -51.070 56.767
661842 91.527 -51.084 56.770
661843 91.558 -51.099 56.774
661844 91.589 -51.113 56.779
661845 91.620 -51.127 56.785
661846 91.651 -51.141 56.791
661847 91.682 -51.155 56.796
661848 91.714 -51.170 56.800
661849 91.744 -51.184 56.804
661850 91.776 -51.199 56.808
661851 91.807 -51.213 56.814
661852 91.838 -51.227 56.819
661853 91.869 -51.242 56.819
661854 91.900 -51.256 56.821
661855 91.931 -51.271 56.823
661856 91.962 -51.285 56.826
661857 91.994 -51.299 56.829
661858 92.025 -51.313 56.831
661859 92.056 -51.327 56.829
661860 92.087 -51.342 56.828
661861 92.118 -51.356 56.827
661862 92.149 -51.371 56.827
661863 92.180 -51.385 56.827
661864 92.212 -51.396 56.827
661865 92.244 -51.403 56.824
661866 92.278 -51.415 56.820
661867 92.312 -51.434 56.818
661868 92.346 -51.453 56.815
661869 92.380 -51.473 56.811
661870 92.404 -51.489 56.805
661871 92.425 -51.505 56.796
661872 92.445 -51.521 56.786
661873 92.466 -51.536 56.773
661874 92.488 -51.552 56.758
661875 92.508 -51.567 56.742
661876 92.529 -51.583 56.725
661877 92.557 -51.601 56.707
661878 92.587 -51.621 56.690
661879 92.618 -51.641 56.674
661880 92.649 -51.660 56.655
661881 92.680 -51.681 56.635
661882 92.710 -51.700 56.612
661883 92.741 -51.720 56.587
661884 92.772 -51.740 56.561
661885 92.802 -51.760 56.536
661886 92.833 -51.785 56.511
661887 92.862 -51.814 56.485
661888 92.893 -51.843 56.456
661889 92.922 -51.872 56.425
661890 92.952 -51.901 56.393
661891 92.982 -51.930 56.360
661892 93.012 -51.959 56.327
661893 93.042 -51.988 56.294
661894 93.072 -52.017 56.259
661895 93.102 -52.045 56.222
661896 93.132 -52.074 56.183
661897 93.162 -52.103 56.142
661898 93.192 -52.132 56.099
661899 93.227 -52.161 56.057
661900 93.267 -52.192 56.016
661901 93.299 -52.211 55.973
661902 93.330 -52.230 55.930
661903 93.360 -52.249 55.886
661904 93.391 -52.268 55.840
661905 93.421 -52.286 55.794
661906 93.452 -52.305 55.749
661907 93.483 -52.323 55.705
661908 93.513 -52.342 55.662
661909 93.544 -52.360 55.618
661910 93.574 -52.379 55.575
661941 94.487 -52.946 54.833
661942 94.517 -52.964 54.826
661943 94.546 -52.982 54.819
661944 94.575 -53.001 54.813
661945 94.605 -53.020 54.807
661946 94.634 -53.038 54.801
661947 94.664 -53.056 54.796
661948 94.693 -53.074 54.794
661949 94.722 -53.092 54.791
661950 94.752 -53.111 54.789
661951 94.781 -53.129 54.787
661952 94.811 -53.147 54.786
661953 94.840 -53.166 54.785
661954 94.870 -53.184 54.784
661955 94.897 -53.193 54.783
661956 94.925 -53.199 54.782
661957 94.953 -53.205 54.782
661958 94.980 -53.211 54.781
661959 95.008 -53.218 54.780
661960 95.035 -53.224 54.781
661961 95.063 -53.235 54.782
661962 95.087 -53.255 54.782
661963 95.100 -53.273 54.781
661964 95.114 -53.292 54.779
661965 95.127 -53.311 54.774
661966 95.141 -53.329 54.767
661967 95.155 -53.348 54.759
661968 95.169 -53.366 54.751
661969 95.183 -53.384 54.743
661970 95.196 -53.402 54.732
661971 95.210 -53.421 54.718
661972 95.225 -53.441 54.702
661973 95.243 -53.463 54.685
661974 95.263 -53.486 54.671
661975 95.290 -53.508 54.665
661976 95.317 -53.531 54.662
661977 95.344 -53.553 54.657
661978 95.372 -53.575 54.650
661979 95.398 -53.598 54.643
661980 95.424 -53.618 54.639
661981 95.448 -53.634 54.623
661982 95.472 -53.649 54.598
661983 95.497 -53.664 54.568
661984 95.522 -53.679 54.533
661985 95.547 -53.694 54.494
661986 95.571 -53.709 54.451
661987 95.597 -53.725 54.410
661988 95.621 -53.739 54.368
661989 95.645 -53.753 54.327
661990 95.661 -53.762 54.288
661991 95.676 -53.770 54.257
661992 95.692 -53.778 54.233
661993 95.707 -53.787 54.209
661994 95.723 -53.795 54.180
661995 95.739 -53.803 54.151
661996 95.754 -53.812 54.120
661997 95.770 -53.820 54.090
661998 95.785 -53.828 54.062
661999 95.801 -53.837 54.033
662000 95.817 -53.845 53.999
662001 95.832 -53.853 53.966
662002 95.858 -53.860 53.938
662003 95.883 -53.868 53.910
662004 95.908 -53.875 53.882
662005 95.934 -53.882 53.852
662006 95.959 -53.888 53.820
662007 95.984 -53.896 53.788
662008 96.011 -53.906 53.758
662009 96.039 -53.919 53.730
662010 96.066 -53.931 53.701
662011 96.094 -53.943 53.669
662012 96.122 -53.955 53.635
662013 96.150 -53.968 53.602
662014 96.177 -53.980 53.570
662015 96.205 -53.992 53.544
662016 96.233 -54.004 53.521
662017 96.260 -54.016 53.497
662018 96.288 -54.029 53.473
662019 96.318 -54.037 53.450
662020 96.348 -54.044 53.430
662021 96.379 -54.051 53.409
662022 96.409 -54.062 53.384
662023 96.439 -54.073 53.356
662024 96.469 -54.082 53.325
662025 96.499 -54.093 53.290
662026 96.529 -54.101 53.255
662027 96.558 -54.107 53.218
662028 96.588 -54.115 53.180
662029 96.619 -54.123 53.139
662030 96.653 -54.130 53.100
662031 96.686 -54.138 53.060
662032 96.719 -54.145 53.020
662033 96.752 -54.152 52.980
662034 96.783 -54.162 52.942
662035 96.813 -54.171 52.903
662036 96.843 -54.181 52.868
662037 96.874 -54.191 52.830
662038 96.903 -54.200 52.793
662039 96.933 -54.207 52.754
662040 96.961 -54.214 52.713
662041 96.991 -54.220 52.668
662042 97.018 -54.229 52.622
662043 97.043 -54.239 52.577
662044 97.069 -54.249 52.531
662045 97.094 -54.259 52.484
662046 97.120 -54.270 52.437
662047 97.145 -54.279 52.391
662048 97.171 -54.289 52.343
662049 97.197 -54.297 52.296
662050 97.223 -54.305 52.249
662051 97.249 -54.312 52.201
662052 97.275 -54.320 52.153
662053 97.303 -54.328 52.106
662054 97.333 -54.336 52.059
662055 97.363 -54.343 52.013
662056 97.392 -54.350 51.966
662057 97.417 -54.352 51.917
662058 97.441 -54.354 51.869
662059 97.465 -54.355 51.819
662060 97.490 -54.358 51.769
662061 97.514 -54.360 51.719
662062 97.538 -54.361 51.669
662063 97.563 -54.364 51.622
662064 97.591 -54.364 51.575
662065 97.620 -54.365 51.525
662066 97.650 -54.366 51.474
662067 97.679 -54.367 51.422
662068 97.709 -54.367 51.370
662069 97.738 -54.368 51.315
662070 97.768 -54.370 51.261
662071 97.795 -54.370 51.205
662072 97.819 -54.371 51.150
662073 97.843 -54.372 51.094
662074 97.866 -54.373 51.037
662075 97.890 -54.374 50.977
662076 97.914 -54.375 50.914
662077 97.939 -54.376 50.850
662078 97.966 -54.376 50.781
662079 97.993 -54.377 50.712
662080 98.019 -54.377 50.640
662081 98.046 -54.377 50.568
662082 98.073 -54.377 50.498
662083 98.100 -54.378 50.430
662084 98.127 -54.378 50.362
662085 98.154 -54.379 50.294
662086 98.180 -54.379 50.228
662087 98.207 -54.380 50.165
662088 98.234 -54.380 50.103
662089 98.261 -54.380 50.044
662090 98.287 -54.383 49.987
662509 90.504 -50.515 56.829
662510 90.539 -50.535 56.828
662511 90.574 -50.556 56.827
662512 90.608 -50.577 56.825
662513 90.643 -50.597 56.823
662514 90.677 -50.618 56.820
662515 90.712 -50.639 56.815
662516 90.746 -50.659 56.808
662517 90.779 -50.676 56.799
662518 90.811 -50.696 56.789
662519 90.844 -50.710 56.778
662520 90.876 -50.719 56.764
662521 90.888 -50.735 56.750
662522 90.899 -50.751 56.734
662523 90.925 -50.766 56.716
662524 90.961 -50.780 56.698
662525 90.997 -50.795 56.679
662526 91.032 -50.809 56.661
662527 91.064 -50.823 56.642
662528 91.095 -50.838 56.624
662529 91.126 -50.852 56.608
662543 91.562 -51.053 56.759
662544 91.593 -51.067 56.763
662545 91.624 -51.082 56.767
662546 91.655 -51.096 56.772
662547 91.686 -51.110 56.778
662548 91.717 -51.124 56.782
662549 91.748 -51.139 56.786
662550 91.780 -51.153 56.790
662551 91.811 -51.168 56.794
662552 91.842 -51.182 56.799
662553 91.873 -51.196 56.804
662554 91.904 -51.211 56.806
662555 91.935 -51.225 56.807
662556 91.966 -51.240 56.809
662557 91.998 -51.254 56.811
662558 92.029 -51.268 56.814
662559 92.060 -51.282 56.817
662560 92.091 -51.296 56.815
662561 92.122 -51.311 56.813
662562 92.153 -51.325 56.812
662563 92.184 -51.340 56.812
662564 92.216 -51.354 56.812
662565 92.247 -51.368 56.812
662566 92.278 -51.383 56.809
662567 92.309 -51.395 56.805
662568 92.343 -51.409 56.803
662569 92.375 -51.427 56.800
662570 92.395 -51.443 56.795
662571 92.417 -51.458 56.788
662572 92.438 -51.474 56.778
662573 92.458 -51.490 56.768
662574 92.480 -51.505 56.756
662575 92.501 -51.521 56.741
662576 92.521 -51.536 56.726
662577 92.547 -51.554 56.710
662578 92.578 -51.574 56.694
662579 92.609 -51.594 56.677
662580 92.639 -51.613 56.661
662581 92.670 -51.633 56.643
662582 92.700 -51.653 56.625
662583 92.731 -51.672 56.604
662584 92.762 -51.693 56.580
662585 92.792 -51.713 56.555
662586 92.823 -51.732 56.532
662587 92.853 -51.759 56.509
662588 92.883 -51.788 56.486
662589 92.913 -51.816 56.459
662590 92.943 -51.846 56.430
662591 92.973 -51.875 56.399
662592 93.003 -51.904 56.368
662593 93.033 -51.932 56.337
662594 93.063 -51.961 56.305
662595 93.093 -51.990 56.272
662596 93.123 -52.019 56.237
662597 93.153 -52.048 56.200
662598 93.183 -52.076 56.159
662599 93.213 -52.105 56.117
662600 93.243 -52.134 56.072
662601 93.280 -52.164 56.028
662602 93.314 -52.188 55.985
662603 93.345 -52.206 55.941
662604 93.375 -52.225 55.898
662605 93.406 -52.243 55.852
662606 93.436 -52.262 55.806
662607 93.467 -52.281 55.762
662608 93.498 -52.299 55.718
662609 93.529 -52.318 55.675
662610 93.558 -52.336 55.632
662611 93.587 -52.355 55.590
662645 94.586 -52.976 54.807
662646 94.616 -52.995 54.800
662647 94.645 -53.013 54.794
662648 94.675 -53.031 54.789
662649 94.704 -53.049 54.785
662650 94.733 -53.067 54.782
662651 94.763 -53.086 54.779
662652 94.792 -53.104 54.777
662653 94.821 -53.123 54.775
662654 94.851 -53.141 54.774
662655 94.880 -53.159 54.772
662656 94.909 -53.175 54.771
662657 94.937 -53.181 54.770
662658 94.964 -53.187 54.770
662659 94.992 -53.194 54.769
662660 95.020 -53.200 54.769
662661 95.047 -53.206 54.770
662662 95.075 -53.214 54.771
662663 95.104 -53.236 54.770
662664 95.126 -53.255 54.768
662665 95.140 -53.274 54.763
662666 95.154 -53.292 54.756
662667 95.168 -53.311 54.748
662668 95.181 -53.330 54.738
662669 95.195 -53.348 54.729
662670 95.209 -53.366 54.720
662671 95.222 -53.384 54.709
662672 95.236 -53.403 54.697
662673 95.252 -53.423 54.684
662674 95.270 -53.446 54.672
662675 95.288 -53.468 54.667
662676 95.307 -53.490 54.669
662677 95.334 -53.512 54.666
662678 95.361 -53.535 54.661
662679 95.388 -53.558 54.655
662680 95.415 -53.580 54.647
662681 95.441 -53.599 54.634
662682 95.464 -53.614 54.617
662683 95.489 -53.629 54.594
662684 95.514 -53.644 54.563
662685 95.539 -53.659 54.527
662686 95.564 -53.675 54.487
662687 95.589 -53.690 54.447
662688 95.614 -53.705 54.410
662689 95.639 -53.720 54.371
662690 95.664 -53.735 54.331
662691 95.679 -53.744 54.292
662692 95.695 -53.752 54.260
662693 95.710 -53.760 54.238
662694 95.726 -53.769 54.217
662695 95.742 -53.777 54.188
662696 95.757 -53.785 54.157
662697 95.773 -53.794 54.126
662698 95.789 -53.802 54.095
662699 95.804 -53.810 54.064
662700 95.820 -53.819 54.034
662701 95.835 -53.827 54.003
662702 95.851 -53.835 53.971
662703 95.867 -53.844 53.944
662704 95.888 -53.851 53.916
662705 95.914 -53.858 53.888
662706 95.939 -53.865 53.857
662707 95.964 -53.872 53.823
662708 95.989 -53.879 53.788
662709 96.016 -53.888 53.757
662710 96.043 -53.900 53.728
662711 96.071 -53.913 53.699
662712 96.099 -53.925 53.669
662713 96.126 -53.937 53.637
662714 96.154 -53.949 53.604
662715 96.182 -53.962 53.572
662716 96.209 -53.974 53.545
662717 96.237 -53.986 53.522
662718 96.266 -53.996 53.500
662719 96.297 -54.004 53.478
662720 96.327 -54.010 53.456
662721 96.358 -54.018 53.436
662722 96.388 -54.026 53.415
662723 96.417 -54.037 53.390
662724 96.448 -54.048 53.363
662725 96.478 -54.057 53.332
662726 96.507 -54.064 53.298
662727 96.537 -54.072 53.262
662728 96.567 -54.079 53.224
662729 96.596 -54.087 53.185
662730 96.627 -54.095 53.145
662731 96.660 -54.102 53.108
662732 96.693 -54.109 53.069
662733 96.727 -54.117 53.030
662734 96.760 -54.124 52.991
662735 96.793 -54.132 52.951
662736 96.826 -54.138 52.914
662737 96.856 -54.148 52.878
662738 96.886 -54.158 52.840
662739 96.916 -54.167 52.802
662740 96.946 -54.174 52.763
662741 96.974 -54.181 52.722
662742 97.001 -54.191 52.678
662743 97.026 -54.201 52.633
662744 97.051 -54.211 52.587
662745 97.077 -54.221 52.542
662746 97.102 -54.232 52.496
662747 97.127 -54.242 52.450
662748 97.153 -54.252 52.404
662749 97.179 -54.261 52.357
662750 97.205 -54.268 52.310
662751 97.231 -54.276 52.264
662752 97.257 -54.284 52.216
662753 97.283 -54.292 52.168
662754 97.309 -54.300 52.121
662755 97.338 -54.308 52.075
662756 97.368 -54.315 52.029
662757 97.397 -54.321 51.982
662758 97.422 -54.323 51.934
662759 97.446 -54.325 51.886
662760 97.470 -54.327 51.837
662761 97.495 -54.329 51.788
662762 97.519 -54.331 51.739
662763 97.544 -54.333 51.689
662764 97.571 -54.333 51.643
662765 97.598 -54.333 51.596
662766 97.628 -54.334 51.546
662767 97.657 -54.335 51.496
662768 97.687 -54.336 51.445
662769 97.716 -54.337 51.393
662770 97.746 -54.338 51.341
662771 97.775 -54.339 51.287
662772 97.800 -54.339 51.233
662773 97.823 -54.341 51.178
662774 97.847 -54.342 51.123
662775 97.871 -54.342 51.067
662776 97.894 -54.343 51.009
662777 97.919 -54.345 50.948
662778 97.942 -54.345 50.884
662779 97.968 -54.346 50.819
662780 97.995 -54.346 50.752
662781 98.022 -54.346 50.683
662782 98.049 -54.347 50.612
662783 98.076 -54.347 50.542
662784 98.102 -54.348 50.473
662785 98.130 -54.348 50.405
662786 98.156 -54.348 50.338
662787 98.183 -54.348 50.271
662788 98.210 -54.349 50.207
662789 98.237 -54.349 50.145
662790 98.264 -54.350 50.085
662791 98.289 -54.366 50.026
662792 98.314 -54.384 49.969
663209 90.487 -50.458 56.812
663210 90.522 -50.478 56.814
663211 90.556 -50.500 56.814
663212 90.591 -50.520 56.814
663213 90.625 -50.541 56.814
663214 90.660 -50.561 56.813
663215 90.694 -50.582 56.811
663216 90.729 -50.603 56.807
663217 90.764 -50.623 56.802
663218 90.799 -50.644 56.794
663219 90.831 -50.663 56.784
663220 90.864 -50.677 56.774
663221 90.881 -50.691 56.761
663222 90.906 -50.706 56.747
663223 90.943 -50.721 56.732
663224 90.974 -50.735 56.715
663225 91.005 -50.750 56.697
663226 91.036 -50.763 56.678
663227 91.068 -50.778 56.660
663228 91.099 -50.792 56.642
663229 91.130 -50.807 56.624
663230 91.161 -50.821 56.608
663245 91.628 -51.036 56.752
663246 91.659 -51.051 56.755
663247 91.690 -51.065 56.760
663248 91.721 -51.080 56.765
663249 91.752 -51.093 56.769
663250 91.784 -51.108 56.773
663251 91.815 -51.122 56.776
663252 91.846 -51.136 56.780
663253 91.877 -51.151 56.784
663254 91.908 -51.165 56.789
663255 91.939 -51.180 56.792
663256 91.969 -51.194 56.793
663257 92.000 -51.208 56.794
663258 92.030 -51.223 56.797
663259 92.060 -51.237 56.799
663260 92.090 -51.252 56.801
663261 92.121 -51.266 56.800
663262 92.151 -51.280 56.799
663263 92.181 -51.295 56.798
663264 92.211 -51.309 56.797
663265 92.241 -51.324 56.797
663266 92.272 -51.338 56.796
663267 92.302 -51.352 56.793
663268 92.333 -51.367 56.790
663269 92.363 -51.381 56.786
663270 92.388 -51.396 56.782
663271 92.409 -51.412 56.777
663272 92.430 -51.428 56.770
663273 92.450 -51.443 56.760
663274 92.472 -51.459 56.750
663275 92.493 -51.474 56.738
663276 92.513 -51.490 56.724
663277 92.537 -51.506 56.710
663278 92.568 -51.526 56.695
663279 92.598 -51.546 56.680
663280 92.629 -51.566 56.665
663281 92.660 -51.586 56.648
663282 92.691 -51.606 56.631
663283 92.721 -51.625 56.613
663284 92.752 -51.645 56.594
663285 92.783 -51.665 56.572
663286 92.813 -51.685 56.550
663287 92.844 -51.704 56.529
663288 92.874 -51.732 56.507
663289 92.904 -51.762 56.486
663290 92.934 -51.791 56.461
663291 92.964 -51.819 56.433
663292 92.994 -51.848 56.404
663293 93.024 -51.877 56.375
663294 93.054 -51.906 56.347
663295 93.084 -51.935 56.316
663296 93.114 -51.964 56.284
663297 93.144 -51.992 56.251
663298 93.174 -52.021 56.215
663299 93.204 -52.051 56.175
663300 93.234 -52.080 56.134
663301 93.264 -52.108 56.087
663302 93.294 -52.137 56.042
663303 93.329 -52.164 55.998
663304 93.360 -52.183 55.954
663305 93.391 -52.202 55.910
663306 93.421 -52.220 55.864
663307 93.451 -52.239 55.819
663308 93.481 -52.257 55.775
663309 93.510 -52.275 55.731
663310 93.540 -52.293 55.688
663311 93.569 -52.311 55.646
663312 93.599 -52.330 55.605
663349 94.685 -53.006 54.782
663350 94.714 -53.024 54.777
663351 94.744 -53.042 54.773
663352 94.773 -53.060 54.770
663353 94.803 -53.079 54.767
663354 94.832 -53.098 54.765
663355 94.861 -53.116 54.763
663356 94.891 -53.134 54.761
663357 94.920 -53.152 54.760
663358 94.949 -53.163 54.759
663359 94.976 -53.169 54.758
663360 95.004 -53.176 54.757
663361 95.031 -53.182 54.755
663362 95.059 -53.188 54.757
663363 95.087 -53.194 54.758
663364 95.115 -53.214 54.757
663365 95.144 -53.236 54.753
663366 95.166 -53.256 54.746
663367 95.180 -53.274 54.737
663368 95.194 -53.292 54.728
663369 95.207 -53.311 54.716
663370 95.222 -53.330 54.706
663371 95.235 -53.348 54.697
663372 95.249 -53.366 54.688
663373 95.262 -53.385 54.678
663374 95.278 -53.405 54.669
663375 95.297 -53.427 54.663
663376 95.315 -53.450 54.665
663377 95.334 -53.472 54.671
663378 95.352 -53.495 54.669
663379 95.379 -53.518 54.663
663380 95.405 -53.540 54.657
663381 95.432 -53.562 54.649
663382 95.457 -53.580 54.630
663383 95.481 -53.595 54.609
663384 95.507 -53.610 54.584
663385 95.531 -53.625 54.553
663386 95.556 -53.640 54.518
663387 95.581 -53.655 54.481
663388 95.606 -53.670 54.444
663389 95.631 -53.685 54.411
663390 95.656 -53.700 54.374
663391 95.681 -53.715 54.334
663392 95.698 -53.725 54.295
663393 95.713 -53.733 54.264
663394 95.729 -53.741 54.243
663395 95.745 -53.750 54.224
663396 95.760 -53.758 54.196
663397 95.776 -53.766 54.166
663398 95.792 -53.775 54.132
663399 95.807 -53.783 54.098
663400 95.823 -53.791 54.065
663401 95.838 -53.800 54.035
663402 95.854 -53.808 54.006
663403 95.870 -53.816 53.978
663404 95.885 -53.825 53.953
663405 95.901 -53.833 53.926
663406 95.919 -53.841 53.897
663407 95.945 -53.848 53.865
663408 95.970 -53.855 53.828
663409 95.995 -53.862 53.790
663410 96.021 -53.869 53.756
663411 96.048 -53.882 53.727
663412 96.076 -53.894 53.699
663413 96.104 -53.907 53.670
663414 96.131 -53.919 53.639
663415 96.159 -53.931 53.608
663416 96.187 -53.944 53.576
663417 96.215 -53.956 53.548
663418 96.245 -53.963 53.524
663419 96.275 -53.970 53.502
663420 96.306 -53.978 53.482
663421 96.336 -53.985 53.461
663422 96.366 -53.992 53.441
663423 96.396 -54.002 53.420
663424 96.427 -54.012 53.395
663425 96.456 -54.022 53.367
663426 96.486 -54.029 53.336
663427 96.515 -54.036 53.303
663428 96.545 -54.044 53.267
663429 96.575 -54.051 53.229
663430 96.604 -54.059 53.191
663431 96.635 -54.066 53.152
663432 96.668 -54.073 53.114
663433 96.701 -54.081 53.077
663434 96.734 -54.088 53.039
663435 96.767 -54.096 53.001
663436 96.801 -54.103 52.961
663437 96.834 -54.110 52.923
663438 96.867 -54.118 52.886
663439 96.899 -54.126 52.850
663440 96.930 -54.135 52.812
663441 96.957 -54.143 52.772
663442 96.983 -54.153 52.732
663443 97.008 -54.164 52.688
663444 97.034 -54.173 52.644
663445 97.059 -54.184 52.599
663446 97.085 -54.194 52.554
663447 97.110 -54.204 52.509
663448 97.136 -54.214 52.463
663449 97.161 -54.224 52.417
663450 97.187 -54.233 52.371
663451 97.213 -54.240 52.324
663452 97.239 -54.248 52.278
663453 97.264 -54.256 52.230
663454 97.291 -54.264 52.183
663455 97.316 -54.271 52.136
663456 97.344 -54.279 52.090
663457 97.374 -54.287 52.043
663458 97.402 -54.293 51.996
663459 97.427 -54.295 51.949
663460 97.451 -54.297 51.902
663461 97.475 -54.298 51.855
663462 97.500 -54.301 51.807
663463 97.526 -54.301 51.757
663464 97.552 -54.301 51.708
663465 97.579 -54.302 51.663
663466 97.606 -54.303 51.615
663467 97.636 -54.304 51.566
663468 97.665 -54.305 51.516
663469 97.695 -54.305 51.466
663470 97.724 -54.306 51.415
663471 97.753 -54.308 51.365
663472 97.781 -54.308 51.311
663473 97.804 -54.309 51.258
663474 97.828 -54.310 51.204
663475 97.852 -54.311 51.150
663476 97.875 -54.312 51.096
663477 97.899 -54.313 51.039
663478 97.923 -54.314 50.981
663479 97.946 -54.315 50.918
663480 97.971 -54.316 50.855
663481 97.998 -54.316 50.791
663482 98.025 -54.317 50.723
663483 98.052 -54.317 50.654
663484 98.078 -54.317 50.584
663485 98.105 -54.317 50.516
663486 98.132 -54.317 50.448
663487 98.159 -54.318 50.381
663488 98.185 -54.318 50.314
663489 98.213 -54.319 50.249
663490 98.239 -54.319 50.186
663491 98.266 -54.330 50.125
663492 98.291 -54.348 50.066
663493 98.316 -54.367 50.008
663910 90.504 -50.422 56.795
663911 90.539 -50.443 56.798
663912 90.574 -50.463 56.800
663913 90.608 -50.484 56.801
663914 90.643 -50.504 56.802
663915 90.677 -50.525 56.802
663916 90.712 -50.546 56.801
663917 90.747 -50.566 56.799
663918 90.781 -50.587 56.795
663919 90.816 -50.608 56.788
663920 90.850 -50.629 56.779
663921 90.883 -50.647 56.770
663922 90.912 -50.661 56.758
663923 90.943 -50.676 56.744
663924 90.973 -50.690 56.729
663925 91.003 -50.704 56.713
663926 91.033 -50.719 56.696
663927 91.064 -50.733 56.678
663928 91.094 -50.748 56.660
663929 91.124 -50.762 56.642
663930 91.154 -50.776 56.625
663931 91.184 -50.791 56.608
663947 91.668 -51.020 56.745
663948 91.699 -51.035 56.749
663949 91.729 -51.049 56.754
663950 91.759 -51.064 56.757
663951 91.789 -51.078 56.759
663952 91.819 -51.092 56.763
663953 91.850 -51.107 56.767
663954 91.880 -51.121 56.770
663955 91.910 -51.136 56.776
663956 91.940 -51.150 56.778
663957 91.970 -51.164 56.779
663958 92.001 -51.179 56.781
663959 92.031 -51.193 56.783
663960 92.061 -51.208 56.785
663961 92.091 -51.222 56.786
663962 92.122 -51.236 56.786
663963 92.152 -51.251 56.784
663964 92.182 -51.265 56.783
663965 92.212 -51.280 56.782
663966 92.242 -51.294 56.781
663967 92.273 -51.309 56.780
663968 92.303 -51.323 56.775
663969 92.334 -51.337 56.771
663970 92.364 -51.352 56.767
663971 92.394 -51.366 56.762
663972 92.422 -51.381 56.758
663973 92.443 -51.396 56.752
663974 92.463 -51.412 56.742
663975 92.485 -51.428 56.731
663976 92.506 -51.443 56.720
663977 92.527 -51.459 56.707
663978 92.558 -51.479 56.694
663979 92.589 -51.499 56.680
663980 92.620 -51.519 56.666
663981 92.650 -51.538 56.651
663982 92.681 -51.558 56.635
663983 92.712 -51.578 56.619
663984 92.742 -51.597 56.602
663985 92.773 -51.618 56.584
663986 92.804 -51.638 56.564
663987 92.835 -51.657 56.545
663988 92.865 -51.678 56.525
663989 92.895 -51.707 56.505
663990 92.925 -51.735 56.484
663991 92.955 -51.764 56.461
663992 92.985 -51.793 56.435
663993 93.015 -51.822 56.408
663994 93.045 -51.851 56.381
663995 93.075 -51.879 56.355
663996 93.105 -51.908 56.326
663997 93.135 -51.937 56.296
663998 93.165 -51.967 56.263
663999 93.195 -51.995 56.227
664000 93.225 -52.024 56.190
664001 93.255 -52.053 56.149
664002 93.285 -52.082 56.102
664003 93.314 -52.111 56.056
664004 93.345 -52.139 56.011
664005 93.374 -52.158 55.967
664006 93.404 -52.177 55.922
664007 93.433 -52.195 55.877
664008 93.463 -52.214 55.832
664009 93.492 -52.232 55.788
664010 93.521 -52.250 55.745
664011 93.551 -52.268 55.702
664012 93.580 -52.286 55.660
664013 93.609 -52.305 55.619
664051 94.725 -52.999 54.769
664052 94.755 -53.017 54.764
664053 94.784 -53.035 54.761
664054 94.814 -53.054 54.757
664055 94.843 -53.073 54.755
664056 94.873 -53.091 54.752
664057 94.902 -53.109 54.750
664058 94.931 -53.127 54.749
664059 94.960 -53.145 54.747
664060 94.988 -53.151 54.745
664061 95.016 -53.158 54.745
664062 95.043 -53.164 54.742
664063 95.071 -53.170 54.743
664064 95.098 -53.176 54.743
664065 95.127 -53.193 54.742
664066 95.156 -53.214 54.737
664067 95.185 -53.236 54.728
664068 95.206 -53.256 54.718
664069 95.220 -53.274 54.707
664070 95.234 -53.292 54.696
664071 95.248 -53.311 54.688
664072 95.261 -53.330 54.681
664073 95.275 -53.348 54.674
664074 95.289 -53.367 54.668
664075 95.305 -53.387 54.661
664076 95.324 -53.410 54.657
664077 95.342 -53.432 54.662
664078 95.360 -53.455 54.665
664079 95.379 -53.477 54.665
664080 95.397 -53.499 54.663
664081 95.422 -53.522 54.657
664082 95.450 -53.544 54.646
664083 95.474 -53.560 54.627
664084 95.499 -53.575 54.599
664085 95.524 -53.590 54.570
664086 95.549 -53.606 54.540
664087 95.573 -53.621 54.508
664088 95.599 -53.636 54.475
664089 95.623 -53.651 54.442
664090 95.649 -53.666 54.411
664091 95.673 -53.681 54.375
664092 95.698 -53.697 54.336
664093 95.716 -53.706 54.298
664094 95.732 -53.715 54.267
664095 95.747 -53.723 54.249
664096 95.763 -53.731 54.234
664097 95.778 -53.740 54.206
664098 95.794 -53.748 54.174
664099 95.810 -53.756 54.138
664100 95.825 -53.765 54.101
664101 95.841 -53.773 54.065
664102 95.856 -53.781 54.034
664103 95.872 -53.790 54.007
664104 95.888 -53.798 53.984
664105 95.903 -53.806 53.962
664106 95.919 -53.815 53.936
664107 95.935 -53.823 53.905
664108 95.950 -53.831 53.869
664109 95.975 -53.838 53.830
664110 96.001 -53.845 53.791
664111 96.026 -53.853 53.759
664112 96.053 -53.864 53.732
664113 96.081 -53.876 53.704
664114 96.108 -53.888 53.675
664115 96.136 -53.900 53.645
664116 96.164 -53.913 53.614
664117 96.193 -53.922 53.583
664118 96.223 -53.930 53.552
664119 96.254 -53.937 53.526
664120 96.284 -53.944 53.504
664121 96.314 -53.951 53.483
664122 96.344 -53.959 53.464
664123 96.375 -53.966 53.446
664124 96.405 -53.977 53.424
664125 96.435 -53.985 53.397
664126 96.464 -53.993 53.369
664127 96.494 -54.001 53.338
664128 96.523 -54.008 53.305
664129 96.553 -54.016 53.270
664130 96.583 -54.023 53.234
664131 96.612 -54.030 53.197
664132 96.643 -54.038 53.158
664133 96.676 -54.045 53.120
664134 96.709 -54.053 53.084
664135 96.742 -54.060 53.048
664136 96.775 -54.067 53.010
664137 96.808 -54.075 52.971
664138 96.841 -54.082 52.932
664139 96.875 -54.090 52.895
664140 96.908 -54.097 52.859
664141 96.940 -54.106 52.821
664142 96.965 -54.116 52.782
664143 96.991 -54.126 52.742
664144 97.016 -54.136 52.699
664145 97.041 -54.146 52.656
664146 97.067 -54.157 52.612
664147 97.092 -54.167 52.567
664148 97.118 -54.177 52.522
664149 97.143 -54.187 52.477
664150 97.169 -54.196 52.431
664151 97.195 -54.204 52.385
664152 97.221 -54.212 52.338
664153 97.247 -54.220 52.291
664154 97.273 -54.227 52.243
664155 97.299 -54.236 52.196
664156 97.324 -54.243 52.149
664157 97.351 -54.251 52.103
664158 97.379 -54.259 52.056
664159 97.407 -54.264 52.009
664160 97.432 -54.266 51.964
664161 97.456 -54.268 51.916
664162 97.481 -54.270 51.869
664163 97.508 -54.270 51.821
664164 97.534 -54.270 51.773
664165 97.561 -54.271 51.725
664166 97.587 -54.271 51.679
664167 97.613 -54.272 51.633
664168 97.643 -54.273 51.585
664169 97.672 -54.274 51.537
664170 97.702 -54.275 51.487
664171 97.731 -54.276 51.436
664172 97.760 -54.276 51.386
664173 97.785 -54.277 51.334
664174 97.809 -54.279 51.283
664175 97.832 -54.279 51.230
664176 97.856 -54.280 51.178
664177 97.880 -54.281 51.125
664178 97.903 -54.283 51.070
664179 97.927 -54.283 51.011
664180 97.951 -54.284 50.951
664181 97.974 -54.285 50.889
664182 98.001 -54.286 50.827
664183 98.027 -54.286 50.762
664184 98.054 -54.286 50.694
664185 98.081 -54.287 50.625
664186 98.108 -54.287 50.557
664187 98.135 -54.287 50.491
664188 98.161 -54.288 50.424
664189 98.189 -54.288 50.357
664190 98.215 -54.289 50.291
664191 98.242 -54.294 50.228
664192 98.267 -54.312 50.166
664193 98.293 -54.331 50.106
664194 98.318 -54.349 50.048
664195 98.343 -54.367 49.992
664611 90.518 -50.390 56.778
664612 90.553 -50.412 56.782
664613 90.587 -50.432 56.785
664614 90.622 -50.453 56.787
664615 90.656 -50.474 56.789
664616 90.691 -50.494 56.790
664617 90.725 -50.516 56.790
664618 90.760 -50.536 56.789
664619 90.794 -50.557 56.786
664620 90.829 -50.578 56.781
664621 90.863 -50.599 56.773
664622 90.897 -50.620 56.765
664623 90.932 -50.641 56.753
664624 90.962 -50.659 56.741
664625 90.994 -50.677 56.727
664626 91.024 -50.694 56.711
664627 91.048 -50.707 56.695
664628 91.078 -50.720 56.677
664629 91.107 -50.733 56.660
664630 91.137 -50.747 56.642
664631 91.166 -50.760 56.625
664632 91.195 -50.773 56.609
664649 91.700 -51.005 56.739
664650 91.730 -51.020 56.743
664651 91.760 -51.034 56.746
664652 91.790 -51.049 56.748
664653 91.820 -51.063 56.751
664654 91.851 -51.077 56.754
664655 91.881 -51.092 56.758
664656 91.911 -51.106 56.763
664657 91.941 -51.121 56.765
664658 91.971 -51.135 56.766
664659 92.002 -51.149 56.767
664660 92.032 -51.164 56.769
664661 92.062 -51.178 56.770
664662 92.092 -51.193 56.773
664663 92.123 -51.207 56.771
664664 92.153 -51.221 56.770
664665 92.183 -51.236 56.769
664666 92.213 -51.250 56.767
664667 92.244 -51.265 56.765
664668 92.274 -51.279 56.763
664669 92.304 -51.293 56.759
664670 92.335 -51.308 56.754
664671 92.365 -51.322 56.749
664672 92.395 -51.337 56.744
664673 92.425 -51.351 56.740
664674 92.455 -51.365 56.734
664675 92.477 -51.381 56.723
664676 92.498 -51.396 56.714
664677 92.518 -51.412 56.703
664678 92.549 -51.431 56.691
664679 92.579 -51.451 56.679
664680 92.610 -51.471 56.666
664681 92.641 -51.491 56.653
664682 92.671 -51.511 56.638
664683 92.702 -51.531 56.623
664684 92.733 -51.550 56.607
664685 92.764 -51.570 56.590
664686 92.794 -51.590 56.573
664687 92.825 -51.609 56.555
664688 92.856 -51.630 56.538
664689 92.886 -51.651 56.521
664690 92.916 -51.680 56.502
664691 92.946 -51.709 56.482
664692 92.976 -51.738 56.460
664693 93.006 -51.766 56.437
664694 93.036 -51.795 56.412
664695 93.066 -51.824 56.388
664696 93.096 -51.853 56.362
664697 93.126 -51.882 56.335
664698 93.156 -51.911 56.305
664699 93.186 -51.940 56.272
664700 93.216 -51.969 56.238
664701 93.246 -51.998 56.201
664702 93.276 -52.026 56.160
664703 93.305 -52.055 56.114
664704 93.333 -52.086 56.069
664705 93.366 -52.114 56.024
664706 93.394 -52.133 55.979
664707 93.420 -52.152 55.934
664708 93.447 -52.170 55.889
664709 93.474 -52.189 55.845
664710 93.502 -52.207 55.802
664711 93.532 -52.225 55.758
664712 93.561 -52.243 55.716
664713 93.590 -52.261 55.674
664714 93.620 -52.280 55.632
664715 93.649 -52.298 55.591
664752 94.736 -52.974 54.761
664753 94.766 -52.992 54.756
664754 94.795 -53.011 54.752
664755 94.824 -53.029 54.748
664756 94.854 -53.048 54.745
664757 94.883 -53.066 54.742
664758 94.912 -53.084 54.739
664759 94.942 -53.102 54.737
664760 94.971 -53.120 54.735
664761 95.000 -53.133 54.733
664762 95.027 -53.139 54.733
664763 95.055 -53.145 54.730
664764 95.083 -53.151 54.728
664765 95.110 -53.158 54.727
664766 95.138 -53.172 54.725
664767 95.167 -53.193 54.720
664768 95.196 -53.214 54.711
664769 95.225 -53.236 54.699
664770 95.246 -53.256 54.688
664771 95.260 -53.274 54.678
664772 95.274 -53.293 54.672
664773 95.287 -53.311 54.668
664774 95.301 -53.330 54.663
664775 95.315 -53.349 54.659
664776 95.332 -53.369 54.654
664777 95.350 -53.392 54.651
664778 95.369 -53.415 54.655
664779 95.387 -53.437 54.660
664780 95.405 -53.459 54.663
664781 95.424 -53.482 54.663
664782 95.442 -53.504 54.656
664783 95.466 -53.525 54.638
664784 95.491 -53.540 54.614
664785 95.516 -53.556 54.585
664786 95.541 -53.571 54.555
664787 95.566 -53.586 54.527
664788 95.591 -53.601 54.499
664789 95.616 -53.616 54.469
664790 95.641 -53.631 54.441
664791 95.666 -53.646 54.411
664792 95.691 -53.662 54.374
664793 95.716 -53.677 54.336
664794 95.735 -53.687 54.299
664795 95.750 -53.696 54.272
664796 95.766 -53.704 54.256
664797 95.781 -53.712 54.243
664798 95.797 -53.721 54.217
664799 95.813 -53.729 54.183
664800 95.828 -53.737 54.145
664801 95.844 -53.746 54.104
664802 95.860 -53.754 54.065
664803 95.875 -53.762 54.033
664804 95.891 -53.771 54.007
664805 95.906 -53.779 53.987
664806 95.922 -53.787 53.966
664807 95.938 -53.796 53.940
664808 95.953 -53.804 53.908
664809 95.969 -53.813 53.871
664810 95.984 -53.821 53.830
664811 96.006 -53.828 53.792
664812 96.031 -53.836 53.763
664813 96.058 -53.846 53.737
664814 96.086 -53.858 53.711
664815 96.113 -53.870 53.682
664816 96.141 -53.882 53.652
664817 96.171 -53.890 53.621
664818 96.202 -53.897 53.588
664819 96.232 -53.904 53.558
664820 96.262 -53.911 53.529
664821 96.293 -53.919 53.504
664822 96.323 -53.925 53.484
664823 96.354 -53.933 53.466
664824 96.384 -53.941 53.447
664825 96.413 -53.950 53.424
664826 96.443 -53.957 53.397
664827 96.472 -53.965 53.368
664828 96.502 -53.972 53.337
664829 96.532 -53.980 53.305
664830 96.561 -53.988 53.272
664831 96.591 -53.994 53.238
664832 96.620 -54.002 53.201
664833 96.650 -54.010 53.164
664834 96.683 -54.017 53.126
664835 96.717 -54.025 53.089
664836 96.750 -54.032 53.054
664837 96.783 -54.039 53.018
664838 96.816 -54.047 52.981
664839 96.849 -54.054 52.940
664840 96.882 -54.062 52.903
664841 96.914 -54.070 52.867
664842 96.946 -54.078 52.830
664843 96.973 -54.091 52.792
664844 96.999 -54.101 52.752
664845 97.024 -54.111 52.710
664846 97.049 -54.120 52.668
664847 97.075 -54.130 52.624
664848 97.100 -54.140 52.580
664849 97.125 -54.150 52.535
664850 97.151 -54.160 52.490
664851 97.177 -54.168 52.444
664852 97.203 -54.176 52.398
664853 97.229 -54.183 52.352
664854 97.254 -54.192 52.303
664855 97.281 -54.199 52.255
664856 97.306 -54.207 52.209
664857 97.332 -54.215 52.162
664858 97.358 -54.223 52.116
664859 97.385 -54.230 52.069
664860 97.413 -54.236 52.022
664861 97.437 -54.238 51.977
664862 97.463 -54.239 51.929
664863 97.490 -54.239 51.882
664864 97.516 -54.239 51.836
664865 97.542 -54.240 51.788
664866 97.569 -54.240 51.742
664867 97.595 -54.241 51.696
664868 97.621 -54.242 51.650
664869 97.650 -54.242 51.603
664870 97.680 -54.243 51.555
664871 97.709 -54.244 51.506
664872 97.739 -54.245 51.456
664873 97.766 -54.246 51.407
664874 97.790 -54.247 51.357
664875 97.813 -54.248 51.307
664876 97.837 -54.248 51.255
664877 97.861 -54.250 51.202
664878 97.884 -54.251 51.150
664879 97.908 -54.251 51.096
664880 97.932 -54.252 51.039
664881 97.955 -54.254 50.980
664882 97.979 -54.254 50.921
664883 98.003 -54.255 50.860
664884 98.030 -54.256 50.797
664885 98.057 -54.256 50.732
664886 98.084 -54.257 50.665
664887 98.110 -54.257 50.599
664888 98.138 -54.257 50.532
664889 98.164 -54.258 50.466
664890 98.191 -54.258 50.400
664891 98.218 -54.258 50.335
664892 98.243 -54.276 50.271
664893 98.269 -54.295 50.209
664894 98.294 -54.313 50.148
664895 98.319 -54.332 50.089
664896 98.345 -54.350 50.032
665312 90.531 -50.361 56.760
665313 90.566 -50.381 56.765
665314 90.600 -50.403 56.769
665315 90.635 -50.423 56.772
665316 90.670 -50.444 56.775
665317 90.704 -50.465 56.778
665318 90.739 -50.485 56.779
665319 90.773 -50.506 56.778
665320 90.808 -50.527 56.777
665321 90.842 -50.548 56.773
665322 90.876 -50.570 56.767
665323 90.910 -50.591 56.759
665324 90.945 -50.613 56.749
665325 90.978 -50.634 56.737
665326 91.010 -50.653 56.724
665327 91.041 -50.670 56.709
665328 91.072 -50.688 56.693
665329 91.103 -50.705 56.677
665330 91.134 -50.723 56.659
665331 91.165 -50.740 56.643
665332 91.189 -50.753 56.626
665333 91.214 -50.765 56.609
665353 91.803 -51.033 56.737
665354 91.832 -51.047 56.740
665355 91.862 -51.060 56.743
665356 91.891 -51.074 56.747
665357 91.921 -51.087 56.750
665358 91.950 -51.101 56.753
665359 91.980 -51.114 56.754
665360 92.009 -51.127 56.755
665361 92.038 -51.141 56.756
665362 92.068 -51.154 56.757
665363 92.097 -51.168 56.759
665364 92.127 -51.181 56.758
665365 92.156 -51.194 56.756
665366 92.186 -51.208 56.754
665367 92.215 -51.221 56.752
665368 92.245 -51.235 56.749
665369 92.275 -51.249 56.747
665370 92.305 -51.264 56.742
665371 92.336 -51.278 56.737
665372 92.366 -51.293 56.733
665373 92.396 -51.307 56.727
665374 92.426 -51.321 56.722
665375 92.456 -51.336 56.717
665376 92.487 -51.350 56.706
665377 92.511 -51.365 56.697
665378 92.539 -51.384 56.686
665379 92.570 -51.404 56.675
665380 92.600 -51.424 56.663
665381 92.631 -51.443 56.652
665382 92.662 -51.463 56.639
665383 92.693 -51.483 56.625
665384 92.723 -51.503 56.610
665385 92.754 -51.523 56.594
665386 92.784 -51.543 56.579
665387 92.815 -51.562 56.564
665388 92.846 -51.582 56.548
665389 92.876 -51.602 56.531
665390 92.907 -51.625 56.515
665391 92.937 -51.653 56.498
665392 92.967 -51.682 56.479
665393 92.997 -51.711 56.459
665394 93.027 -51.740 56.437
665395 93.057 -51.769 56.414
665396 93.087 -51.798 56.393
665397 93.117 -51.827 56.369
665398 93.147 -51.856 56.343
665399 93.177 -51.885 56.313
665400 93.207 -51.913 56.280
665401 93.236 -51.942 56.247
665402 93.267 -51.971 56.211
665403 93.294 -52.001 56.169
665404 93.322 -52.032 56.123
665405 93.350 -52.062 56.079
665406 93.389 -52.088 56.036
665407 93.415 -52.107 55.992
665408 93.442 -52.125 55.947
665409 93.469 -52.144 55.902
665410 93.496 -52.162 55.857
665411 93.522 -52.181 55.815
665412 93.549 -52.199 55.771
665413 93.576 -52.218 55.729
665414 93.603 -52.236 55.687
665415 93.631 -52.255 55.646
665416 93.660 -52.274 55.604
665453 94.747 -52.949 54.755
665454 94.776 -52.968 54.749
665455 94.806 -52.986 54.744
665456 94.835 -53.004 54.740
665457 94.864 -53.023 54.736
665458 94.894 -53.041 54.732
665459 94.923 -53.059 54.728
665460 94.953 -53.077 54.726
665461 94.982 -53.095 54.723
665462 95.012 -53.114 54.721
665463 95.039 -53.121 54.720
665464 95.067 -53.127 54.716
665465 95.094 -53.133 54.713
665466 95.122 -53.139 54.710
665467 95.150 -53.151 54.707
665468 95.179 -53.172 54.702
665469 95.208 -53.193 54.693
665470 95.237 -53.215 54.683
665471 95.265 -53.236 54.671
665472 95.286 -53.256 54.664
665473 95.300 -53.275 54.660
665474 95.314 -53.293 54.657
665475 95.328 -53.311 54.654
665476 95.341 -53.330 54.650
665477 95.358 -53.352 54.646
665478 95.377 -53.374 54.646
665479 95.395 -53.396 54.649
665480 95.414 -53.419 54.656
665481 95.432 -53.441 54.662
665482 95.451 -53.464 54.663
665483 95.469 -53.486 54.653
665484 95.486 -53.505 54.630
665485 95.509 -53.521 54.602
665486 95.533 -53.537 54.572
665487 95.558 -53.552 54.541
665488 95.583 -53.567 54.513
665489 95.608 -53.582 54.488
665490 95.633 -53.597 54.462
665491 95.658 -53.612 54.436
665492 95.683 -53.627 54.405
665493 95.708 -53.642 54.368
665494 95.733 -53.657 54.329
665495 95.753 -53.669 54.293
665496 95.769 -53.678 54.269
665497 95.784 -53.686 54.259
665498 95.800 -53.694 54.247
665499 95.816 -53.703 54.220
665500 95.831 -53.711 54.188
665501 95.847 -53.719 54.150
665502 95.863 -53.728 54.108
665503 95.878 -53.736 54.065
665504 95.894 -53.744 54.029
665505 95.909 -53.753 54.004
665506 95.925 -53.761 53.983
665507 95.941 -53.769 53.962
665508 95.956 -53.778 53.937
665509 95.972 -53.786 53.907
665510 95.987 -53.794 53.869
665511 96.003 -53.803 53.828
665512 96.019 -53.811 53.792
665513 96.037 -53.819 53.765
665514 96.062 -53.828 53.741
665515 96.090 -53.840 53.716
665516 96.119 -53.849 53.689
665517 96.150 -53.856 53.658
665518 96.180 -53.863 53.626
665519 96.210 -53.871 53.593
665520 96.241 -53.878 53.563
665521 96.271 -53.885 53.533
665522 96.302 -53.892 53.505
665523 96.332 -53.900 53.483
665524 96.362 -53.907 53.466
665525 96.392 -53.914 53.447
665526 96.421 -53.922 53.423
665527 96.451 -53.929 53.395
665528 96.480 -53.937 53.367
665529 96.510 -53.944 53.336
665530 96.540 -53.951 53.305
665531 96.569 -53.959 53.273
665532 96.599 -53.966 53.240
665533 96.628 -53.974 53.205
665534 96.658 -53.982 53.168
665535 96.691 -53.989 53.131
665536 96.724 -53.996 53.094
665537 96.757 -54.004 53.058
665538 96.791 -54.011 53.023
665539 96.824 -54.019 52.986
665540 96.856 -54.026 52.946
665541 96.888 -54.035 52.911
665542 96.920 -54.043 52.876
665543 96.952 -54.051 52.840
665544 96.980 -54.063 52.802
665545 97.007 -54.076 52.763
665546 97.033 -54.086 52.721
665547 97.058 -54.096 52.679
665548 97.083 -54.106 52.635
665549 97.109 -54.116 52.591
665550 97.134 -54.125 52.546
665551 97.159 -54.135 52.502
665552 97.186 -54.133 52.456
665553 97.212 -54.142 52.410
665554 97.238 -54.151 52.364
665555 97.263 -54.160 52.315
665556 97.289 -54.168 52.267
665557 97.315 -54.177 52.221
665558 97.341 -54.185 52.175
665559 97.366 -54.194 52.128
665560 97.392 -54.202 52.081
665561 97.419 -54.207 52.035
665562 97.445 -54.207 51.989
665563 97.471 -54.207 51.941
665564 97.498 -54.208 51.894
665565 97.524 -54.209 51.849
665566 97.550 -54.209 51.803
665567 97.577 -54.210 51.759
665568 97.603 -54.211 51.714
665569 97.630 -54.211 51.668
665570 97.658 -54.211 51.620
665571 97.687 -54.213 51.573
665572 97.717 -54.214 51.526
665573 97.746 -54.214 51.476
665574 97.770 -54.215 51.428
665575 97.794 -54.216 51.379
665576 97.818 -54.217 51.329
665577 97.842 -54.218 51.278
665578 97.865 -54.219 51.226
665579 97.889 -54.220 51.174
665580 97.913 -54.221 51.121
665581 97.936 -54.222 51.065
665582 97.960 -54.223 51.008
665583 97.984 -54.224 50.951
665584 98.007 -54.225 50.892
665585 98.033 -54.226 50.831
665586 98.060 -54.226 50.768
665587 98.086 -54.226 50.703
665588 98.113 -54.226 50.638
665589 98.140 -54.227 50.573
665590 98.167 -54.227 50.507
665591 98.194 -54.227 50.441
665592 98.220 -54.240 50.377
665593 98.245 -54.258 50.313
665594 98.271 -54.277 50.250
665595 98.296 -54.295 50.189
665596 98.321 -54.314 50.130
665597 98.347 -54.333 50.073
665598 98.372 -54.351 50.018
666014 90.579 -50.352 56.749
666015 90.614 -50.372 56.754
666016 90.648 -50.393 56.758
666017 90.683 -50.414 56.762
666018 90.717 -50.434 56.765
666019 90.752 -50.456 56.767
666020 90.786 -50.476 56.767
666021 90.820 -50.498 56.767
666022 90.855 -50.519 56.764
666023 90.889 -50.541 56.759
666024 90.923 -50.562 56.752
666025 90.957 -50.583 56.743
666026 90.992 -50.605 56.732
666027 91.026 -50.626 56.720
666028 91.059 -50.646 56.706
666029 91.089 -50.663 56.691
666030 91.120 -50.681 56.675
666031 91.151 -50.698 56.659
666032 91.182 -50.716 56.643
666033 91.213 -50.733 56.627
666057 91.910 -51.065 56.737
666058 91.939 -51.079 56.740
666059 91.968 -51.092 56.742
666060 91.998 -51.105 56.743
666061 92.027 -51.119 56.744
666062 92.057 -51.133 56.744
666063 92.086 -51.146 56.745
666064 92.116 -51.159 56.747
666065 92.145 -51.173 56.745
666066 92.174 -51.186 56.742
666067 92.204 -51.199 56.740
666068 92.233 -51.213 56.738
666069 92.263 -51.226 56.735
666070 92.292 -51.240 56.733
666071 92.321 -51.253 56.727
666072 92.351 -51.266 56.722
666073 92.380 -51.280 56.717
666074 92.410 -51.293 56.712
666075 92.439 -51.307 56.705
666076 92.469 -51.320 56.700
666077 92.498 -51.334 56.691
666078 92.527 -51.347 56.681
666079 92.560 -51.357 56.670
666080 92.591 -51.376 56.660
666081 92.621 -51.396 56.648
666082 92.652 -51.416 56.638
666083 92.683 -51.436 56.625
666084 92.713 -51.456 56.612
666085 92.744 -51.475 56.597
666086 92.775 -51.495 56.583
666087 92.805 -51.515 56.569
666088 92.836 -51.535 56.556
666089 92.867 -51.555 56.541
666090 92.898 -51.575 56.524
666091 92.928 -51.598 56.509
666092 92.958 -51.627 56.493
666093 92.988 -51.656 56.475
666094 93.018 -51.685 56.455
666095 93.048 -51.714 56.435
666096 93.078 -51.743 56.414
666097 93.108 -51.772 56.395
666098 93.138 -51.801 56.373
666099 93.167 -51.829 56.348
666100 93.198 -51.858 56.318
666101 93.227 -51.887 56.287
666102 93.255 -51.917 56.254
666103 93.283 -51.948 56.218
666104 93.310 -51.978 56.177
666105 93.338 -52.008 56.132
666106 93.376 -52.039 56.089
666107 93.410 -52.062 56.047
666108 93.437 -52.080 56.003
666109 93.464 -52.099 55.959
666110 93.491 -52.117 55.915
666111 93.517 -52.136 55.870
666112 93.544 -52.155 55.827
666113 93.571 -52.174 55.784
666114 93.598 -52.192 55.741
666115 93.625 -52.211 55.700
666116 93.651 -52.229 55.658
666117 93.678 -52.248 55.618
666154 94.758 -52.925 54.748
666155 94.787 -52.943 54.742
666156 94.817 -52.961 54.737
666157 94.846 -52.979 54.731
666158 94.875 -52.998 54.727
666159 94.905 -53.016 54.722
666160 94.934 -53.034 54.718
666161 94.963 -53.052 54.714
666162 94.993 -53.070 54.711
666163 95.022 -53.089 54.708
666164 95.051 -53.103 54.705
666165 95.078 -53.109 54.701
666166 95.106 -53.115 54.697
666167 95.134 -53.122 54.694
666168 95.162 -53.129 54.690
666169 95.190 -53.151 54.684
666170 95.219 -53.172 54.676
666171 95.248 -53.194 54.667
666172 95.277 -53.215 54.659
666173 95.306 -53.237 54.653
666174 95.326 -53.257 54.649
666175 95.340 -53.275 54.646
666176 95.354 -53.293 54.643
666177 95.368 -53.312 54.640
666178 95.385 -53.333 54.638
666179 95.404 -53.356 54.641
666180 95.422 -53.379 54.648
666181 95.441 -53.401 54.658
666182 95.459 -53.424 54.665
666183 95.477 -53.443 54.662
666184 95.494 -53.462 54.647
666185 95.512 -53.481 54.622
666186 95.530 -53.499 54.589
666187 95.550 -53.517 54.559
666188 95.576 -53.532 54.527
666189 95.601 -53.547 54.498
666190 95.625 -53.562 54.475
666191 95.651 -53.577 54.451
666192 95.675 -53.593 54.427
666193 95.700 -53.608 54.397
666194 95.725 -53.623 54.360
666195 95.750 -53.638 54.319
666196 95.772 -53.651 54.283
666197 95.788 -53.659 54.261
666198 95.803 -53.668 54.255
666199 95.819 -53.676 54.241
666200 95.834 -53.684 54.218
666201 95.850 -53.693 54.186
666202 95.866 -53.701 54.148
666203 95.881 -53.709 54.106
666204 95.897 -53.718 54.062
666205 95.912 -53.726 54.022
666206 95.928 -53.734 53.992
666207 95.944 -53.743 53.968
666208 95.959 -53.751 53.946
666209 95.975 -53.759 53.923
666210 95.990 -53.768 53.894
666211 96.006 -53.776 53.861
666212 96.022 -53.784 53.822
666213 96.037 -53.793 53.789
666214 96.052 -53.801 53.763
666215 96.067 -53.809 53.739
666216 96.098 -53.816 53.715
666217 96.128 -53.824 53.687
666218 96.158 -53.831 53.657
666219 96.188 -53.839 53.627
666220 96.218 -53.847 53.595
666221 96.248 -53.854 53.565
666222 96.278 -53.861 53.535
666223 96.308 -53.869 53.507
666224 96.338 -53.876 53.482
666225 96.369 -53.883 53.465
666226 96.400 -53.888 53.445
666227 96.430 -53.895 53.420
666228 96.459 -53.901 53.394
666229 96.488 -53.909 53.366
666230 96.518 -53.916 53.335
666231 96.548 -53.923 53.304
666232 96.578 -53.931 53.274
666233 96.607 -53.938 53.242
666234 96.637 -53.946 53.208
666235 96.666 -53.954 53.173
666236 96.699 -53.960 53.136
666237 96.732 -53.968 53.099
666238 96.765 -53.976 53.063
666239 96.798 -53.983 53.028
666240 96.830 -53.991 52.991
666241 96.862 -54.000 52.953
666242 96.894 -54.008 52.919
666243 96.927 -54.016 52.885
666244 96.958 -54.024 52.850
666245 96.987 -54.036 52.813
666246 97.015 -54.049 52.773
666247 97.041 -54.061 52.732
666248 97.067 -54.071 52.689
666249 97.092 -54.081 52.646
666250 97.117 -54.091 52.602
666251 97.142 -54.101 52.557
666252 97.168 -54.107 52.512
666253 97.195 -54.104 52.467
666254 97.222 -54.107 52.420
666255 97.247 -54.116 52.373
666256 97.274 -54.124 52.325
666257 97.299 -54.132 52.279
666258 97.325 -54.142 52.232
666259 97.351 -54.150 52.185
666260 97.376 -54.159 52.138
666261 97.402 -54.167 52.092
666262 97.428 -54.171 52.046
666263 97.453 -54.176 52.000
666264 97.479 -54.177 51.953
666265 97.506 -54.178 51.907
666266 97.532 -54.178 51.863
666267 97.558 -54.179 51.818
666268 97.585 -54.179 51.774
666269 97.611 -54.179 51.730
666270 97.638 -54.180 51.683
666271 97.665 -54.181 51.637
666272 97.695 -54.182 51.591
666273 97.724 -54.182 51.544
666274 97.751 -54.184 51.495
666275 97.775 -54.185 51.447
666276 97.798 -54.185 51.399
666277 97.822 -54.186 51.350
666278 97.846 -54.188 51.300
666279 97.870 -54.189 51.249
666280 97.893 -54.189 51.198
666281 97.917 -54.190 51.145
666282 97.941 -54.192 51.089
666283 97.964 -54.192 51.034
666284 97.988 -54.193 50.978
666285 98.012 -54.194 50.921
666286 98.035 -54.195 50.863
666287 98.062 -54.195 50.803
666288 98.089 -54.196 50.741
666289 98.116 -54.196 50.676
666290 98.143 -54.197 50.612
666291 98.169 -54.197 50.547
666292 98.196 -54.204 50.482
666293 98.221 -54.223 50.418
666294 98.247 -54.241 50.354
666295 98.272 -54.260 50.292
666296 98.298 -54.278 50.231
666297 98.323 -54.296 50.171
666298 98.349 -54.315 50.113
666299 98.374 -54.333 50.058
666715 90.592 -50.321 56.733
666716 90.627 -50.342 56.739
666717 90.661 -50.363 56.743
666718 90.696 -50.384 56.747
666719 90.730 -50.405 56.751
666720 90.765 -50.425 56.754
666721 90.799 -50.447 56.756
666722 90.833 -50.469 56.757
666723 90.868 -50.490 56.755
666724 90.902 -50.511 56.750
666725 90.936 -50.533 56.744
666726 90.970 -50.554 56.737
666727 91.005 -50.575 56.727
666728 91.039 -50.597 56.716
666729 91.073 -50.619 56.703
666730 91.106 -50.639 56.689
666731 91.137 -50.657 56.674
666732 91.168 -50.674 56.658
666733 91.199 -50.691 56.643
666734 91.230 -50.709 56.628
666760 91.987 -51.084 56.732
666761 92.016 -51.097 56.733
666762 92.046 -51.111 56.733
666763 92.075 -51.124 56.733
666764 92.104 -51.137 56.734
666765 92.134 -51.151 56.734
666766 92.163 -51.164 56.733
666767 92.193 -51.177 56.730
666768 92.222 -51.191 56.727
666769 92.251 -51.205 56.724
666770 92.281 -51.218 56.721
666771 92.310 -51.231 56.718
666772 92.340 -51.245 56.713
666773 92.369 -51.258 56.707
666774 92.399 -51.271 56.702
666775 92.428 -51.285 56.696
666776 92.457 -51.299 56.689
666777 92.487 -51.312 56.684
666778 92.516 -51.325 56.674
666779 92.546 -51.339 56.665
666780 92.575 -51.352 56.656
666781 92.605 -51.365 56.645
666782 92.642 -51.368 56.634
666783 92.673 -51.388 56.623
666784 92.704 -51.408 56.611
666785 92.734 -51.428 56.598
666786 92.765 -51.448 56.585
666787 92.796 -51.468 56.572
666788 92.827 -51.487 56.559
666789 92.857 -51.507 56.548
666790 92.888 -51.527 56.533
666791 92.919 -51.547 56.517
666792 92.949 -51.572 56.502
666793 92.979 -51.600 56.486
666794 93.009 -51.630 56.469
666795 93.039 -51.659 56.451
666796 93.069 -51.688 56.433
666797 93.098 -51.716 56.415
666798 93.129 -51.745 56.395
666799 93.158 -51.774 56.374
666800 93.189 -51.803 56.349
666801 93.216 -51.833 56.322
666802 93.243 -51.863 56.293
666803 93.271 -51.894 56.260
666804 93.299 -51.924 56.225
666805 93.326 -51.954 56.186
666806 93.357 -51.985 56.143
666807 93.402 -52.014 56.099
666808 93.432 -52.036 56.056
666809 93.459 -52.054 56.014
666810 93.486 -52.073 55.970
666811 93.512 -52.092 55.926
666812 93.539 -52.110 55.882
666813 93.566 -52.129 55.838
666814 93.593 -52.147 55.796
666815 93.620 -52.166 55.754
666816 93.646 -52.184 55.712
666817 93.673 -52.203 55.671
666818 93.700 -52.221 55.630
666856 94.798 -52.918 54.735
666857 94.827 -52.936 54.729
666858 94.856 -52.954 54.723
666859 94.886 -52.973 54.718
666860 94.915 -52.991 54.712
666861 94.945 -53.009 54.707
666862 94.974 -53.027 54.702
666863 95.003 -53.046 54.698
666864 95.033 -53.064 54.695
666865 95.062 -53.082 54.690
666866 95.090 -53.091 54.685
666867 95.118 -53.098 54.681
666868 95.145 -53.104 54.678
666869 95.173 -53.110 54.674
666870 95.202 -53.129 54.667
666871 95.231 -53.151 54.661
666872 95.260 -53.173 54.654
666873 95.288 -53.194 54.648
666874 95.317 -53.216 54.643
666875 95.346 -53.237 54.639
666876 95.367 -53.257 54.636
666877 95.380 -53.275 54.633
666878 95.394 -53.294 54.631
666879 95.412 -53.316 54.631
666880 95.430 -53.339 54.635
666881 95.449 -53.361 54.644
666882 95.467 -53.382 54.655
666883 95.485 -53.401 54.660
666884 95.502 -53.420 54.653
666885 95.520 -53.439 54.636
666886 95.538 -53.458 54.607
666887 95.555 -53.476 54.578
666888 95.573 -53.495 54.549
666889 95.593 -53.512 54.517
666890 95.618 -53.527 54.485
666891 95.642 -53.543 54.459
666892 95.668 -53.558 54.437
666893 95.692 -53.573 54.415
666894 95.718 -53.588 54.385
666895 95.743 -53.603 54.348
666896 95.767 -53.618 54.307
666897 95.790 -53.632 54.271
666898 95.806 -53.640 54.249
666899 95.821 -53.649 54.244
666900 95.837 -53.657 54.229
666901 95.852 -53.665 54.210
666902 95.868 -53.674 54.180
666903 95.884 -53.682 54.142
666904 95.899 -53.690 54.100
666905 95.915 -53.699 54.057
666906 95.931 -53.707 54.013
666907 95.946 -53.715 53.978
666908 95.962 -53.724 53.951
666909 95.977 -53.732 53.928
666910 95.993 -53.740 53.904
666911 96.009 -53.749 53.878
666912 96.024 -53.757 53.848
666913 96.040 -53.766 53.813
666914 96.054 -53.774 53.781
666915 96.058 -53.781 53.756
666916 96.072 -53.787 53.732
666917 96.103 -53.794 53.707
666918 96.133 -53.801 53.681
666919 96.164 -53.809 53.652
666920 96.194 -53.816 53.624
666921 96.224 -53.823 53.594
666922 96.254 -53.831 53.565
666923 96.284 -53.838 53.536
666924 96.314 -53.846 53.507
666925 96.344 -53.853 53.481
666926 96.376 -53.859 53.462
666927 96.408 -53.864 53.442
666928 96.438 -53.870 53.419
666929 96.467 -53.877 53.393
666930 96.497 -53.883 53.365
666931 96.526 -53.890 53.335
666932 96.556 -53.896 53.305
666933 96.586 -53.903 53.275
666934 96.615 -53.910 53.245
666935 96.645 -53.917 53.212
666936 96.674 -53.925 53.178
666937 96.707 -53.932 53.142
666938 96.740 -53.940 53.105
666939 96.772 -53.948 53.069
666940 96.804 -53.956 53.033
666941 96.836 -53.964 52.996
666942 96.868 -53.972 52.960
666943 96.900 -53.981 52.928
666944 96.933 -53.989 52.895
666945 96.965 -53.997 52.860
666946 96.994 -54.008 52.822
666947 97.021 -54.022 52.783
666948 97.048 -54.035 52.742
666949 97.075 -54.047 52.700
666950 97.100 -54.056 52.657
666951 97.125 -54.066 52.613
666952 97.151 -54.076 52.568
666953 97.177 -54.080 52.523
666954 97.204 -54.077 52.477
666955 97.231 -54.075 52.430
666956 97.257 -54.080 52.382
666957 97.283 -54.088 52.335
666958 97.309 -54.098 52.289
666959 97.334 -54.106 52.243
666960 97.360 -54.115 52.195
666961 97.386 -54.123 52.148
666962 97.412 -54.129 52.102
666963 97.437 -54.133 52.056
666964 97.463 -54.138 52.011
666965 97.488 -54.142 51.965
666966 97.514 -54.146 51.920
666967 97.540 -54.148 51.876
666968 97.566 -54.148 51.833
666969 97.593 -54.148 51.790
666970 97.619 -54.149 51.746
666971 97.646 -54.150 51.700
666972 97.672 -54.151 51.653
666973 97.702 -54.151 51.607
666974 97.731 -54.152 51.560
666975 97.755 -54.153 51.514
666976 97.779 -54.154 51.466
666977 97.803 -54.155 51.418
666978 97.826 -54.156 51.370
666979 97.850 -54.157 51.320
666980 97.874 -54.157 51.270
666981 97.898 -54.159 51.218
666982 97.922 -54.160 51.166
666983 97.945 -54.160 51.112
666984 97.969 -54.162 51.058
666985 97.993 -54.163 51.003
666986 98.016 -54.164 50.948
666987 98.040 -54.164 50.892
666988 98.065 -54.165 50.834
666989 98.092 -54.166 50.775
666990 98.119 -54.166 50.713
666991 98.145 -54.167 50.649
666992 98.172 -54.168 50.586
666993 98.198 -54.186 50.522
666994 98.223 -54.205 50.458
666995 98.248 -54.223 50.396
666996 98.274 -54.242 50.334
666997 98.299 -54.261 50.272
666998 98.324 -54.279 50.213
666999 98.350 -54.297 50.154
667416 90.606 -50.291 56.715
667417 90.640 -50.312 56.723
667418 90.675 -50.333 56.728
667419 90.709 -50.354 56.733
667420 90.743 -50.375 56.737
667421 90.778 -50.397 56.741
667422 90.812 -50.418 56.744
667423 90.846 -50.439 56.746
667424 90.880 -50.461 56.744
667425 90.915 -50.482 56.741
667426 90.949 -50.503 56.736
667427 90.983 -50.525 56.730
667428 91.017 -50.546 56.722
667429 91.051 -50.568 56.711
667430 91.086 -50.589 56.700
667431 91.120 -50.610 56.686
667432 91.154 -50.632 56.672
667433 91.185 -50.650 56.657
667434 91.216 -50.667 56.643
667435 91.247 -50.685 56.629
667462 92.034 -51.089 56.724
667463 92.064 -51.102 56.724
667464 92.093 -51.116 56.724
667465 92.123 -51.129 56.724
667466 92.152 -51.142 56.724
667467 92.181 -51.156 56.721
667468 92.211 -51.169 56.717
667469 92.240 -51.183 56.714
667470 92.270 -51.196 56.710
667471 92.299 -51.209 56.708
667472 92.329 -51.223 56.704
667473 92.358 -51.236 56.699
667474 92.387 -51.249 56.692
667475 92.417 -51.263 56.686
667476 92.446 -51.277 56.680
667477 92.476 -51.290 56.674
667478 92.505 -51.303 56.668
667479 92.535 -51.317 56.659
667480 92.564 -51.331 56.650
667481 92.594 -51.343 56.641
667482 92.623 -51.357 56.631
667483 92.652 -51.371 56.620
667484 92.683 -51.384 56.610
667485 92.725 -51.381 56.598
667486 92.756 -51.400 56.586
667487 92.786 -51.420 56.573
667488 92.817 -51.440 56.561
667489 92.848 -51.460 56.551
667490 92.878 -51.480 56.539
667491 92.909 -51.500 56.525
667492 92.940 -51.519 56.511
667493 92.970 -51.546 56.495
667494 93.000 -51.575 56.479
667495 93.029 -51.603 56.463
667496 93.060 -51.632 56.447
667497 93.089 -51.661 56.430
667498 93.120 -51.690 56.413
667499 93.149 -51.719 56.395
667500 93.177 -51.749 56.374
667501 93.205 -51.779 56.351
667502 93.232 -51.810 56.325
667503 93.260 -51.840 56.297
667504 93.287 -51.870 56.266
667505 93.315 -51.901 56.232
667506 93.342 -51.931 56.195
667507 93.383 -51.961 56.153
667508 93.427 -51.991 56.110
667509 93.454 -52.010 56.066
667510 93.481 -52.028 56.024
667511 93.507 -52.047 55.981
667512 93.534 -52.065 55.936
667513 93.561 -52.084 55.893
667514 93.588 -52.102 55.850
667515 93.615 -52.121 55.808
667516 93.641 -52.139 55.766
667517 93.668 -52.158 55.724
667518 93.695 -52.177 55.682
667519 93.722 -52.195 55.642
667557 94.809 -52.893 54.729
667558 94.838 -52.911 54.722
667559 94.867 -52.929 54.715
667560 94.897 -52.948 54.709
667561 94.926 -52.966 54.703
667562 94.956 -52.984 54.697
667563 94.985 -53.003 54.692
667564 95.015 -53.021 54.687
667565 95.044 -53.039 54.683
667566 95.073 -53.057 54.677
667567 95.102 -53.073 54.672
667568 95.130 -53.079 54.667
667569 95.157 -53.086 54.663
667570 95.185 -53.092 54.658
667571 95.213 -53.108 54.653
667572 95.242 -53.129 54.650
667573 95.271 -53.151 54.645
667574 95.300 -53.173 54.639
667575 95.329 -53.195 54.634
667576 95.358 -53.216 54.630
667577 95.387 -53.237 54.627
667578 95.406 -53.257 54.624
667579 95.420 -53.276 54.623
667580 95.439 -53.298 54.625
667581 95.457 -53.321 54.629
667582 95.475 -53.339 54.638
667583 95.492 -53.358 54.648
667584 95.510 -53.377 54.650
667585 95.528 -53.396 54.640
667586 95.545 -53.415 54.622
667587 95.563 -53.433 54.593
667588 95.581 -53.452 54.569
667589 95.598 -53.471 54.541
667590 95.616 -53.490 54.508
667591 95.635 -53.509 54.473
667592 95.660 -53.524 54.442
667593 95.685 -53.539 54.420
667594 95.710 -53.554 54.399
667595 95.735 -53.569 54.371
667596 95.760 -53.584 54.335
667597 95.784 -53.599 54.295
667598 95.809 -53.614 54.258
667599 95.824 -53.622 54.233
667600 95.840 -53.631 54.225
667601 95.855 -53.639 54.212
667602 95.871 -53.647 54.193
667603 95.887 -53.656 54.166
667604 95.902 -53.664 54.131
667605 95.918 -53.672 54.091
667606 95.934 -53.681 54.047
667607 95.949 -53.689 54.003
667608 95.965 -53.697 53.964
667609 95.980 -53.706 53.933
667610 95.996 -53.714 53.906
667611 96.012 -53.722 53.880
667612 96.027 -53.731 53.857
667613 96.042 -53.739 53.833
667614 96.045 -53.746 53.801
667615 96.049 -53.753 53.771
667616 96.053 -53.761 53.746
667617 96.075 -53.766 53.722
667618 96.108 -53.772 53.697
667619 96.139 -53.778 53.672
667620 96.169 -53.785 53.646
667621 96.199 -53.793 53.620
667622 96.230 -53.800 53.593
667623 96.260 -53.808 53.565
667624 96.290 -53.816 53.536
667625 96.320 -53.823 53.506
667626 96.350 -53.830 53.479
667627 96.383 -53.835 53.460
667628 96.415 -53.840 53.440
667629 96.446 -53.845 53.417
667630 96.475 -53.852 53.392
667631 96.505 -53.858 53.365
667632 96.535 -53.865 53.336
667633 96.564 -53.871 53.306
667634 96.594 -53.878 53.277
667635 96.623 -53.884 53.248
667636 96.653 -53.891 53.216
667637 96.682 -53.897 53.184
667638 96.714 -53.904 53.148
667639 96.746 -53.913 53.111
667640 96.778 -53.921 53.074
667641 96.810 -53.929 53.039
667642 96.842 -53.938 53.002
667643 96.874 -53.945 52.968
667644 96.906 -53.954 52.935
667645 96.939 -53.962 52.903
667646 96.971 -53.970 52.868
667647 97.001 -53.980 52.832
667648 97.028 -53.994 52.794
667649 97.055 -54.007 52.754
667650 97.083 -54.021 52.710
667651 97.108 -54.032 52.667
667652 97.134 -54.041 52.623
667653 97.159 -54.051 52.578
667654 97.185 -54.053 52.533
667655 97.213 -54.051 52.486
667656 97.240 -54.048 52.440
667657 97.267 -54.044 52.392
667658 97.293 -54.054 52.345
667659 97.319 -54.062 52.298
667660 97.345 -54.071 52.252
667661 97.370 -54.079 52.204
667662 97.396 -54.085 52.156
667663 97.421 -54.090 52.112
667664 97.447 -54.095 52.066
667665 97.472 -54.099 52.021
667666 97.498 -54.104 51.976
667667 97.523 -54.108 51.933
667668 97.549 -54.113 51.890
667669 97.575 -54.117 51.848
667670 97.601 -54.118 51.806
667671 97.627 -54.119 51.762
667672 97.654 -54.119 51.716
667673 97.680 -54.120 51.670
667674 97.710 -54.120 51.624
667675 97.736 -54.121 51.577
667676 97.760 -54.123 51.532
667677 97.784 -54.123 51.484
667678 97.807 -54.124 51.436
667679 97.831 -54.126 51.389
667680 97.855 -54.126 51.340
667681 97.878 -54.127 51.289
667682 97.902 -54.128 51.238
667683 97.926 -54.129 51.186
667684 97.950 -54.130 51.135
667685 97.973 -54.131 51.083
667686 97.997 -54.132 51.030
667687 98.021 -54.133 50.976
667688 98.044 -54.134 50.921
667689 98.068 -54.135 50.865
667690 98.094 -54.135 50.809
667691 98.121 -54.135 50.748
667692 98.148 -54.136 50.686
667693 98.174 -54.151 50.624
667694 98.199 -54.169 50.561
667695 98.225 -54.187 50.499
667696 98.250 -54.206 50.437
667697 98.276 -54.224 50.375
667698 98.301 -54.243 50.314
667699 98.326 -54.261 50.255
667700 98.352 -54.279 50.196
668118 90.653 -50.282 56.706
668119 90.688 -50.303 56.713
668120 90.722 -50.325 56.718
668121 90.756 -50.346 56.723
668122 90.791 -50.367 56.728
668123 90.825 -50.388 56.732
668124 90.859 -50.410 56.734
668125 90.893 -50.431 56.734
668126 90.928 -50.453 56.732
668127 90.962 -50.474 56.728
668128 90.996 -50.496 56.723
668129 91.030 -50.517 56.716
668130 91.065 -50.538 56.706
668131 91.099 -50.560 56.696
668132 91.133 -50.582 56.684
668133 91.167 -50.603 56.671
668134 91.201 -50.624 56.658
668135 91.233 -50.643 56.645
668165 92.111 -51.108 56.715
668166 92.141 -51.121 56.714
668167 92.170 -51.134 56.713
668168 92.200 -51.148 56.711
668169 92.229 -51.161 56.706
668170 92.259 -51.174 56.702
668171 92.288 -51.188 56.698
668172 92.317 -51.201 56.694
668173 92.347 -51.215 56.690
668174 92.376 -51.228 56.685
668175 92.406 -51.241 56.678
668176 92.435 -51.255 56.672
668177 92.465 -51.268 56.665
668178 92.494 -51.281 56.659
668179 92.524 -51.295 56.653
668180 92.553 -51.309 56.645
668181 92.582 -51.322 56.636
668182 92.612 -51.335 56.627
668183 92.641 -51.349 56.618
668184 92.671 -51.362 56.608
668185 92.700 -51.375 56.598
668186 92.730 -51.389 56.586
668187 92.765 -51.395 56.574
668188 92.807 -51.393 56.562
668189 92.838 -51.412 56.551
668190 92.868 -51.432 56.541
668191 92.899 -51.452 56.530
668192 92.930 -51.472 56.517
668193 92.960 -51.492 56.503
668194 92.991 -51.519 56.488
668195 93.020 -51.548 56.473
668196 93.051 -51.577 56.457
668197 93.080 -51.606 56.442
668198 93.111 -51.635 56.426
668199 93.138 -51.665 56.410
668200 93.165 -51.695 56.392
668201 93.193 -51.725 56.374
668202 93.221 -51.756 56.352
668203 93.248 -51.786 56.327
668204 93.276 -51.816 56.299
668205 93.303 -51.847 56.270
668206 93.331 -51.877 56.238
668207 93.364 -51.907 56.203
668208 93.409 -51.937 56.164
668209 93.449 -51.965 56.121
668210 93.476 -51.983 56.077
668211 93.502 -52.002 56.034
668212 93.529 -52.020 55.991
668213 93.556 -52.039 55.947
668214 93.583 -52.058 55.904
668215 93.610 -52.076 55.861
668216 93.636 -52.095 55.819
668217 93.663 -52.114 55.777
668218 93.690 -52.132 55.734
668219 93.717 -52.151 55.694
668220 93.744 -52.169 55.653
668258 94.820 -52.868 54.723
668259 94.849 -52.886 54.716
668260 94.878 -52.904 54.708
668261 94.908 -52.923 54.700
668262 94.937 -52.941 54.695
668263 94.966 -52.960 54.689
668264 94.996 -52.978 54.682
668265 95.025 -52.996 54.677
668266 95.054 -53.014 54.672
668267 95.084 -53.032 54.666
668268 95.113 -53.051 54.660
668269 95.141 -53.061 54.655
668270 95.169 -53.067 54.651
668271 95.197 -53.073 54.646
668272 95.225 -53.087 54.642
668273 95.254 -53.108 54.640
668274 95.282 -53.130 54.637
668275 95.312 -53.151 54.633
668276 95.340 -53.173 54.629
668277 95.370 -53.195 54.624
668278 95.398 -53.216 54.620
668279 95.427 -53.238 54.617
668280 95.447 -53.258 54.615
668281 95.465 -53.279 54.618
668282 95.483 -53.297 54.622
668283 95.500 -53.316 54.628
668284 95.518 -53.335 54.634
668285 95.536 -53.354 54.634
668286 95.553 -53.373 54.622
668287 95.571 -53.392 54.602
668288 95.589 -53.410 54.579
668289 95.606 -53.429 54.561
668290 95.624 -53.448 54.535
668291 95.641 -53.467 54.499
668292 95.659 -53.486 54.461
668293 95.677 -53.504 54.426
668294 95.702 -53.519 54.402
668295 95.727 -53.534 54.381
668296 95.752 -53.549 54.355
668297 95.777 -53.565 54.321
668298 95.802 -53.580 54.282
668299 95.827 -53.595 54.244
668300 95.843 -53.603 54.216
668301 95.859 -53.612 54.205
668302 95.874 -53.620 54.193
668303 95.890 -53.628 54.174
668304 95.905 -53.637 54.148
668305 95.921 -53.645 54.116
668306 95.937 -53.653 54.078
668307 95.952 -53.662 54.036
668308 95.968 -53.670 53.992
668309 95.983 -53.678 53.951
668310 95.999 -53.687 53.918
668311 96.015 -53.695 53.890
668312 96.029 -53.703 53.864
668313 96.033 -53.711 53.839
668314 96.036 -53.718 53.817
668315 96.040 -53.725 53.790
668316 96.044 -53.733 53.761
668317 96.047 -53.740 53.735
668318 96.079 -53.745 53.710
668319 96.112 -53.750 53.686
668320 96.145 -53.755 53.663
668321 96.175 -53.762 53.640
668322 96.205 -53.770 53.615
668323 96.235 -53.778 53.591
668324 96.265 -53.785 53.565
668325 96.296 -53.793 53.536
668326 96.326 -53.800 53.506
668327 96.357 -53.806 53.478
668328 96.389 -53.811 53.459
668329 96.422 -53.816 53.438
668330 96.454 -53.821 53.416
668331 96.484 -53.827 53.392
668332 96.513 -53.834 53.366
668333 96.543 -53.840 53.337
668334 96.573 -53.847 53.308
668335 96.602 -53.853 53.279
668336 96.631 -53.859 53.250
668337 96.661 -53.866 53.220
668338 96.689 -53.872 53.189
668339 96.720 -53.878 53.154
668340 96.752 -53.886 53.117
668341 96.784 -53.894 53.081
668342 96.816 -53.902 53.045
668343 96.849 -53.910 53.008
668344 96.881 -53.919 52.974
668345 96.912 -53.927 52.942
668346 96.945 -53.935 52.910
668347 96.977 -53.943 52.877
668348 97.008 -53.953 52.842
668349 97.035 -53.966 52.805
668350 97.063 -53.979 52.764
668351 97.090 -53.993 52.720
668352 97.117 -54.007 52.678
668353 97.142 -54.016 52.634
668354 97.167 -54.026 52.589
668355 97.194 -54.026 52.543
668356 97.222 -54.023 52.497
668357 97.249 -54.020 52.450
668358 97.276 -54.017 52.402
668359 97.303 -54.018 52.355
668360 97.328 -54.027 52.309
668361 97.354 -54.035 52.263
668362 97.380 -54.043 52.215
668363 97.405 -54.048 52.168
668364 97.431 -54.052 52.123
668365 97.456 -54.057 52.077
668366 97.482 -54.060 52.032
668367 97.508 -54.065 51.989
668368 97.533 -54.070 51.946
668369 97.559 -54.074 51.905
668370 97.584 -54.079 51.863
668371 97.610 -54.083 51.821
668372 97.635 -54.088 51.778
668373 97.662 -54.088 51.733
668374 97.688 -54.089 51.688
668375 97.717 -54.090 51.642
668376 97.741 -54.091 51.595
668377 97.764 -54.091 51.549
668378 97.788 -54.093 51.501
668379 97.812 -54.094 51.453
668380 97.836 -54.095 51.405
668381 97.859 -54.095 51.357
668382 97.883 -54.097 51.308
668383 97.907 -54.098 51.258
668384 97.930 -54.098 51.206
668385 97.954 -54.099 51.158
668386 97.978 -54.101 51.107
668387 98.002 -54.101 51.056
668388 98.025 -54.102 51.004
668389 98.049 -54.104 50.951
668390 98.073 -54.104 50.897
668391 98.097 -54.105 50.842
668392 98.124 -54.105 50.783
668393 98.150 -54.114 50.722
668394 98.175 -54.133 50.661
668395 98.201 -54.151 50.600
668396 98.226 -54.170 50.539
668397 98.252 -54.189 50.477
668398 98.277 -54.207 50.416
668399 98.303 -54.225 50.356
668400 98.328 -54.244 50.297
668401 98.354 -54.262 50.239
668819 90.667 -50.253 56.690
668820 90.701 -50.274 56.697
668821 90.735 -50.295 56.703
668822 90.769 -50.316 56.709
668823 90.803 -50.338 56.714
668824 90.837 -50.359 56.719
668825 90.872 -50.381 56.722
668826 90.906 -50.402 56.722
668827 90.940 -50.423 56.722
668828 90.974 -50.445 56.719
668829 91.009 -50.466 56.715
668830 91.043 -50.488 56.710
668831 91.077 -50.509 56.703
668832 91.111 -50.531 56.693
668833 91.146 -50.552 56.682
668834 91.180 -50.573 56.671
668835 91.214 -50.594 56.659
668836 91.248 -50.616 56.647
668868 92.189 -51.126 56.704
668869 92.218 -51.139 56.701
668870 92.247 -51.152 56.695
668871 92.277 -51.166 56.690
668872 92.306 -51.180 56.686
668873 92.336 -51.193 56.681
668874 92.365 -51.206 56.677
668875 92.395 -51.220 56.671
668876 92.424 -51.233 56.665
668877 92.453 -51.246 56.658
668878 92.483 -51.260 56.650
668879 92.512 -51.273 56.645
668880 92.542 -51.287 56.640
668881 92.571 -51.300 56.632
668882 92.601 -51.314 56.623
668883 92.630 -51.327 56.614
668884 92.660 -51.340 56.606
668885 92.689 -51.354 56.596
668886 92.719 -51.367 56.587
668887 92.748 -51.381 56.576
668888 92.778 -51.394 56.564
668889 92.807 -51.407 56.552
668890 92.848 -51.406 56.542
668891 92.889 -51.405 56.530
668892 92.920 -51.425 56.519
668893 92.951 -51.444 56.508
668894 92.982 -51.464 56.495
668895 93.012 -51.493 56.481
668896 93.042 -51.522 56.467
668897 93.071 -51.550 56.452
668898 93.099 -51.581 56.438
668899 93.127 -51.611 56.422
668900 93.154 -51.641 56.406
668901 93.182 -51.672 56.389
668902 93.209 -51.702 56.371
668903 93.236 -51.732 56.350
668904 93.264 -51.763 56.327
668905 93.292 -51.793 56.300
668906 93.319 -51.823 56.274
668907 93.347 -51.854 56.244
668908 93.390 -51.884 56.210
668909 93.434 -51.913 56.173
668910 93.471 -51.939 56.132
668911 93.497 -51.957 56.087
668912 93.524 -51.976 56.043
668913 93.551 -51.995 56.001
668914 93.578 -52.013 55.958
668915 93.605 -52.032 55.915
668916 93.631 -52.050 55.872
668917 93.658 -52.069 55.829
668918 93.685 -52.087 55.787
668919 93.712 -52.106 55.745
668920 93.739 -52.124 55.704
668921 93.765 -52.143 55.663
668961 94.889 -52.879 54.702
668962 94.918 -52.897 54.694
668963 94.948 -52.916 54.687
668964 94.977 -52.935 54.681
668965 95.006 -52.953 54.674
668966 95.036 -52.971 54.667
668967 95.065 -52.989 54.662
668968 95.095 -53.007 54.656
668969 95.124 -53.026 54.651
668970 95.153 -53.043 54.645
668971 95.181 -53.049 54.640
668972 95.208 -53.055 54.636
668973 95.236 -53.066 54.632
668974 95.265 -53.087 54.631
668975 95.294 -53.109 54.629
668976 95.323 -53.130 54.626
668977 95.352 -53.152 54.622
668978 95.381 -53.173 54.618
668979 95.410 -53.195 54.614
668980 95.439 -53.217 54.611
668981 95.468 -53.239 54.611
668982 95.488 -53.257 54.612
668983 95.509 -53.273 54.615
668984 95.526 -53.292 54.617
668985 95.544 -53.311 54.618
668986 95.561 -53.330 54.614
668987 95.579 -53.349 54.601
668988 95.597 -53.368 54.581
668989 95.614 -53.386 54.569
668990 95.632 -53.405 54.554
668991 95.650 -53.424 54.529
668992 95.667 -53.443 54.493
668993 95.685 -53.462 54.451
668994 95.702 -53.480 54.411
668995 95.720 -53.499 54.383
668996 95.744 -53.515 54.361
668997 95.769 -53.530 54.337
668998 95.795 -53.545 54.306
668999 95.819 -53.560 54.268
669000 95.844 -53.575 54.230
669001 95.862 -53.585 54.200
669002 95.877 -53.593 54.183
669003 95.893 -53.602 54.172
669004 95.908 -53.610 54.154
669005 95.924 -53.618 54.130
669006 95.940 -53.627 54.100
669007 95.955 -53.635 54.065
669008 95.970 -53.643 54.025
669009 95.986 -53.652 53.982
669010 96.002 -53.660 53.940
669011 96.016 -53.668 53.905
669012 96.020 -53.675 53.876
669013 96.024 -53.683 53.849
669014 96.027 -53.690 53.825
669015 96.031 -53.697 53.805
669016 96.035 -53.705 53.780
669017 96.039 -53.712 53.752
669018 96.050 -53.719 53.726
669019 96.083 -53.724 53.701
669020 96.116 -53.729 53.677
669021 96.149 -53.734 53.655
669022 96.181 -53.740 53.634
669023 96.211 -53.747 53.611
669024 96.241 -53.754 53.588
669025 96.271 -53.762 53.562
669026 96.301 -53.769 53.534
669027 96.331 -53.777 53.505
669028 96.364 -53.782 53.477
669029 96.396 -53.787 53.457
669030 96.429 -53.791 53.437
669031 96.461 -53.797 53.416
669032 96.492 -53.802 53.393
669033 96.521 -53.809 53.366
669034 96.551 -53.815 53.338
669035 96.581 -53.822 53.309
669036 96.610 -53.828 53.282
669037 96.640 -53.834 53.253
669038 96.668 -53.841 53.223
669039 96.696 -53.847 53.193
669040 96.727 -53.853 53.159
669041 96.758 -53.860 53.124
669042 96.790 -53.867 53.087
669043 96.822 -53.875 53.050
669044 96.855 -53.883 53.012
669045 96.887 -53.891 52.980
669046 96.919 -53.900 52.949
669047 96.951 -53.908 52.918
669048 96.983 -53.916 52.886
669049 97.015 -53.925 52.852
669050 97.042 -53.938 52.814
669051 97.069 -53.952 52.774
669052 97.096 -53.966 52.731
669053 97.123 -53.979 52.688
669054 97.151 -53.991 52.645
669055 97.176 -54.001 52.600
669056 97.203 -53.998 52.554
669057 97.230 -53.996 52.507
669058 97.257 -53.993 52.459
669059 97.285 -53.990 52.412
669060 97.312 -53.988 52.365
669061 97.338 -53.991 52.320
669062 97.364 -54.000 52.274
669063 97.389 -54.004 52.226
669064 97.415 -54.009 52.178
669065 97.441 -54.013 52.133
669066 97.466 -54.018 52.088
669067 97.492 -54.022 52.044
669068 97.517 -54.027 52.001
669069 97.543 -54.032 51.960
669070 97.568 -54.036 51.919
669071 97.594 -54.040 51.879
669072 97.619 -54.044 51.837
669073 97.645 -54.049 51.794
669074 97.670 -54.054 51.749
669075 97.696 -54.058 51.705
669076 97.722 -54.059 51.659
669077 97.745 -54.060 51.612
669078 97.769 -54.061 51.565
669079 97.793 -54.062 51.517
669080 97.816 -54.063 51.470
669081 97.840 -54.064 51.423
669082 97.864 -54.065 51.375
669083 97.887 -54.066 51.327
669084 97.911 -54.067 51.278
669085 97.935 -54.068 51.228
669086 97.958 -54.069 51.182
669087 97.982 -54.070 51.133
669088 98.006 -54.071 51.083
669089 98.030 -54.072 51.031
669090 98.054 -54.073 50.978
669091 98.077 -54.073 50.926
669092 98.101 -54.075 50.872
669093 98.127 -54.079 50.815
669094 98.152 -54.097 50.756
669095 98.177 -54.115 50.697
669096 98.203 -54.134 50.638
669097 98.228 -54.152 50.577
669098 98.253 -54.170 50.516
669099 98.279 -54.189 50.457
669100 98.304 -54.207 50.396
669101 98.330 -54.226 50.338
669102 98.355 -54.245 50.280
669521 90.714 -50.244 56.682
669522 90.748 -50.265 56.688
669523 90.782 -50.287 56.694
669524 90.816 -50.309 56.700
669525 90.851 -50.330 56.705
669526 90.885 -50.351 56.709
669527 90.919 -50.373 56.711
669528 90.953 -50.394 56.712
669529 90.988 -50.415 56.710
669530 91.022 -50.437 56.708
669531 91.056 -50.459 56.703
669532 91.090 -50.480 56.698
669533 91.124 -50.501 56.690
669534 91.159 -50.522 56.681
669535 91.193 -50.544 56.671
669536 91.227 -50.566 56.660
669571 92.266 -51.144 56.686
669572 92.295 -51.158 56.680
669573 92.325 -51.171 56.675
669574 92.354 -51.184 56.670
669575 92.383 -51.198 56.665
669576 92.413 -51.211 56.659
669577 92.442 -51.224 56.652
669578 92.472 -51.238 56.645
669579 92.501 -51.252 56.637
669580 92.531 -51.265 56.631
669581 92.560 -51.278 56.627
669582 92.590 -51.292 56.619
669583 92.619 -51.305 56.610
669584 92.649 -51.318 56.602
669585 92.678 -51.332 56.594
669586 92.708 -51.346 56.585
669587 92.737 -51.359 56.576
669588 92.766 -51.372 56.566
669589 92.796 -51.386 56.555
669590 92.825 -51.399 56.543
669591 92.855 -51.412 56.533
669592 92.887 -51.421 56.523
669593 92.931 -51.418 56.511
669594 92.972 -51.417 56.499
669595 93.003 -51.437 56.487
669596 93.032 -51.466 56.475
669597 93.060 -51.497 56.462
669598 93.087 -51.527 56.447
669599 93.115 -51.557 56.435
669600 93.142 -51.587 56.418
669601 93.170 -51.618 56.401
669602 93.198 -51.648 56.385
669603 93.225 -51.678 56.367
669604 93.253 -51.709 56.347
669605 93.280 -51.739 56.326
669606 93.308 -51.769 56.302
669607 93.335 -51.800 56.276
669608 93.372 -51.830 56.247
669609 93.416 -51.860 56.215
669610 93.460 -51.890 56.179
669611 93.492 -51.913 56.140
669612 93.519 -51.931 56.098
669613 93.546 -51.950 56.052
669614 93.573 -51.968 56.009
669615 93.600 -51.987 55.966
669616 93.626 -52.005 55.923
669617 93.653 -52.024 55.880
669618 93.680 -52.042 55.838
669619 93.707 -52.061 55.796
669620 93.733 -52.080 55.755
669621 93.760 -52.098 55.714
669622 93.787 -52.117 55.673
669663 94.929 -52.872 54.689
669664 94.959 -52.891 54.682
669665 94.988 -52.910 54.675
669666 95.017 -52.928 54.667
669667 95.047 -52.946 54.660
669668 95.076 -52.964 54.655
669669 95.105 -52.982 54.649
669670 95.135 -53.001 54.643
669671 95.164 -53.019 54.637
669672 95.193 -53.031 54.632
669673 95.220 -53.037 54.627
669674 95.248 -53.045 54.624
669675 95.277 -53.066 54.621
669676 95.306 -53.088 54.620
669677 95.335 -53.109 54.618
669678 95.364 -53.131 54.615
669679 95.392 -53.152 54.612
669680 95.421 -53.173 54.609
669681 95.453 -53.195 54.607
669682 95.484 -53.218 54.608
669683 95.500 -53.240 54.608
669684 95.522 -53.257 54.608
669685 95.543 -53.273 54.607
669686 95.564 -53.290 54.603
669687 95.586 -53.308 54.596
669688 95.604 -53.325 54.581
669689 95.622 -53.344 54.564
669690 95.640 -53.363 54.557
669691 95.658 -53.382 54.547
669692 95.675 -53.401 54.525
669693 95.693 -53.420 54.491
669694 95.710 -53.438 54.445
669695 95.728 -53.457 54.400
669696 95.746 -53.476 54.365
669697 95.763 -53.495 54.339
669698 95.787 -53.511 54.315
669699 95.812 -53.526 54.287
669700 95.836 -53.541 54.253
669701 95.862 -53.556 54.217
669702 95.880 -53.566 54.184
669703 95.895 -53.574 54.165
669704 95.911 -53.583 54.152
669705 95.926 -53.591 54.134
669706 95.942 -53.599 54.112
669707 95.958 -53.608 54.084
669708 95.973 -53.616 54.051
669709 95.989 -53.625 54.013
669710 96.004 -53.633 53.972
669711 96.008 -53.640 53.930
669712 96.011 -53.648 53.894
669713 96.015 -53.655 53.863
669714 96.019 -53.662 53.836
669715 96.022 -53.670 53.813
669716 96.026 -53.677 53.795
669717 96.030 -53.684 53.772
669718 96.033 -53.692 53.745
669719 96.054 -53.698 53.719
669720 96.087 -53.703 53.693
669721 96.120 -53.708 53.669
669722 96.153 -53.712 53.648
669723 96.186 -53.718 53.628
669724 96.217 -53.724 53.606
669725 96.247 -53.731 53.584
669726 96.277 -53.739 53.558
669727 96.307 -53.747 53.530
669728 96.338 -53.753 53.502
669729 96.371 -53.758 53.477
669730 96.403 -53.762 53.456
669731 96.436 -53.768 53.436
669732 96.468 -53.772 53.415
669733 96.500 -53.778 53.393
669734 96.530 -53.784 53.367
669735 96.559 -53.791 53.339
669736 96.589 -53.797 53.310
669737 96.618 -53.803 53.283
669738 96.647 -53.809 53.253
669739 96.675 -53.816 53.225
669740 96.702 -53.822 53.195
669741 96.733 -53.828 53.161
669742 96.765 -53.835 53.126
669743 96.797 -53.842 53.090
669744 96.829 -53.849 53.053
669745 96.861 -53.856 53.017
669746 96.893 -53.865 52.984
669747 96.925 -53.872 52.953
669748 96.957 -53.881 52.923
669749 96.989 -53.889 52.891
669750 97.021 -53.897 52.857
669751 97.049 -53.910 52.821
669752 97.076 -53.924 52.782
669753 97.103 -53.938 52.740
669754 97.131 -53.951 52.698
669755 97.158 -53.965 52.655
669756 97.184 -53.974 52.611
669757 97.212 -53.971 52.565
669758 97.239 -53.969 52.516
669759 97.266 -53.966 52.469
669760 97.294 -53.963 52.421
669761 97.321 -53.960 52.375
669762 97.348 -53.957 52.331
669763 97.374 -53.962 52.285
669764 97.399 -53.966 52.237
669765 97.424 -53.972 52.189
669766 97.449 -53.979 52.143
669767 97.474 -53.984 52.099
669768 97.500 -53.990 52.056
669769 97.525 -53.996 52.014
669770 97.550 -54.001 51.972
669771 97.575 -54.007 51.933
669772 97.600 -54.013 51.894
669773 97.626 -54.019 51.853
669774 97.651 -54.025 51.809
669775 97.676 -54.031 51.764
669776 97.703 -54.033 51.720
669777 97.724 -54.038 51.674
669778 97.745 -54.040 51.628
669779 97.769 -54.041 51.582
669780 97.792 -54.042 51.534
669781 97.816 -54.043 51.488
669782 97.840 -54.044 51.441
669783 97.864 -54.044 51.394
669784 97.887 -54.045 51.347
669785 97.911 -54.046 51.299
669786 97.935 -54.047 51.251
669787 97.959 -54.048 51.205
669788 97.983 -54.048 51.157
669789 98.006 -54.049 51.108
669790 98.030 -54.050 51.057
669791 98.054 -54.051 51.005
669792 98.078 -54.051 50.953
669793 98.101 -54.052 50.901
669794 98.128 -54.060 50.846
669795 98.153 -54.079 50.789
669796 98.179 -54.098 50.731
669797 98.204 -54.116 50.673
669798 98.230 -54.135 50.614
669799 98.255 -54.153 50.554
669800 98.281 -54.172 50.496
669801 98.306 -54.190 50.437
669802 98.332 -54.208 50.379
669803 98.357 -54.227 50.323
670223 90.761 -50.237 56.674
670224 90.795 -50.258 56.680
670225 90.829 -50.279 56.686
670226 90.863 -50.301 56.692
670227 90.897 -50.322 56.697
670228 90.932 -50.343 56.700
670229 90.966 -50.365 56.702
670230 91.000 -50.386 56.702
670231 91.034 -50.408 56.700
670232 91.069 -50.429 56.697
670233 91.103 -50.450 56.693
670234 91.137 -50.472 56.687
670235 91.171 -50.494 56.680
670236 91.206 -50.515 56.672
670274 92.343 -51.163 56.665
670275 92.372 -51.176 56.659
670276 92.402 -51.190 56.653
670277 92.431 -51.203 56.647
670278 92.461 -51.216 56.639
670279 92.490 -51.230 56.632
670280 92.520 -51.243 56.624
670281 92.549 -51.256 56.618
670282 92.579 -51.270 56.613
670283 92.608 -51.284 56.606
670284 92.638 -51.297 56.598
670285 92.667 -51.310 56.590
670286 92.696 -51.324 56.583
670287 92.726 -51.337 56.574
670288 92.755 -51.350 56.565
670289 92.785 -51.364 56.556
670290 92.814 -51.378 56.546
670291 92.844 -51.390 56.535
670292 92.873 -51.404 56.525
670293 92.903 -51.418 56.515
670294 92.932 -51.431 56.504
670295 92.971 -51.433 56.492
670296 93.013 -51.429 56.480
670297 93.048 -51.443 56.469
670298 93.076 -51.473 56.456
670299 93.103 -51.503 56.443
670300 93.131 -51.534 56.430
670301 93.158 -51.564 56.414
670302 93.186 -51.594 56.397
670303 93.214 -51.625 56.380
670304 93.241 -51.655 56.363
670305 93.269 -51.685 56.344
670306 93.296 -51.716 56.323
670307 93.324 -51.746 56.303
670308 93.353 -51.776 56.276
670309 93.397 -51.807 56.248
670310 93.441 -51.836 56.217
670311 93.485 -51.866 56.183
670312 93.514 -51.886 56.146
670313 93.541 -51.905 56.105
670314 93.568 -51.923 56.059
670315 93.594 -51.942 56.016
670316 93.621 -51.960 55.973
670317 93.648 -51.979 55.930
670318 93.675 -51.998 55.887
670319 93.702 -52.017 55.845
670320 93.728 -52.035 55.803
670321 93.756 -52.054 55.763
670322 93.782 -52.072 55.722
670323 93.809 -52.091 55.681
670365 94.969 -52.866 54.676
670366 94.998 -52.885 54.669
670367 95.028 -52.903 54.661
670368 95.057 -52.921 54.655
670369 95.087 -52.939 54.649
670370 95.116 -52.957 54.643
670371 95.145 -52.976 54.637
670372 95.175 -52.994 54.631
670373 95.204 -53.013 54.626
670374 95.232 -53.019 54.621
670375 95.260 -53.026 54.618
670376 95.288 -53.045 54.615
670377 95.317 -53.067 54.613
670378 95.346 -53.088 54.611
670379 95.375 -53.110 54.609
670380 95.404 -53.131 54.608
670381 95.434 -53.152 54.608
670382 95.469 -53.175 54.609
670383 95.500 -53.198 54.610
670384 95.516 -53.220 54.608
670385 95.534 -53.241 54.606
670386 95.555 -53.258 54.600
670387 95.577 -53.274 54.592
670388 95.598 -53.291 54.580
670389 95.619 -53.308 54.564
670390 95.640 -53.324 54.550
670391 95.662 -53.341 54.545
670392 95.683 -53.358 54.538
670393 95.700 -53.377 54.521
670394 95.718 -53.396 54.489
670395 95.736 -53.415 54.442
670396 95.754 -53.433 54.392
670397 95.771 -53.452 54.350
670398 95.789 -53.471 54.320
670399 95.807 -53.490 54.293
670400 95.829 -53.506 54.266
670401 95.854 -53.521 54.236
670402 95.879 -53.537 54.204
670403 95.898 -53.548 54.171
670404 95.914 -53.556 54.149
670405 95.930 -53.565 54.133
670406 95.945 -53.573 54.116
670407 95.961 -53.581 54.095
670408 95.976 -53.590 54.069
670409 95.991 -53.598 54.038
670410 95.994 -53.605 54.002
670411 95.999 -53.612 53.963
670412 96.002 -53.620 53.922
670413 96.006 -53.628 53.884
670414 96.010 -53.634 53.851
670415 96.013 -53.642 53.824
670416 96.017 -53.650 53.801
670417 96.021 -53.656 53.786
670418 96.024 -53.664 53.763
670419 96.028 -53.672 53.738
670420 96.058 -53.677 53.712
670421 96.091 -53.681 53.687
670422 96.124 -53.687 53.663
670423 96.157 -53.691 53.641
670424 96.190 -53.697 53.619
670425 96.223 -53.701 53.597
670426 96.253 -53.709 53.576
670427 96.283 -53.716 53.552
670428 96.313 -53.724 53.525
670429 96.345 -53.729 53.498
670430 96.377 -53.734 53.476
670431 96.410 -53.739 53.454
670432 96.442 -53.744 53.435
670433 96.475 -53.748 53.415
670434 96.508 -53.753 53.393
670435 96.538 -53.759 53.367
670436 96.568 -53.766 53.339
670437 96.597 -53.772 53.311
670438 96.626 -53.778 53.282
670439 96.654 -53.784 53.253
670440 96.681 -53.791 53.225
670441 96.709 -53.797 53.195
670442 96.740 -53.803 53.163
670443 96.771 -53.810 53.127
670444 96.803 -53.817 53.092
670445 96.835 -53.824 53.055
670446 96.867 -53.831 53.019
670447 96.899 -53.838 52.986
670448 96.931 -53.846 52.956
670449 96.963 -53.854 52.926
670450 96.995 -53.862 52.894
670451 97.027 -53.870 52.860
670452 97.055 -53.883 52.825
670453 97.083 -53.896 52.787
670454 97.110 -53.910 52.746
670455 97.137 -53.923 52.705
670456 97.164 -53.935 52.662
670457 97.190 -53.946 52.618
670458 97.216 -53.955 52.573
670459 97.243 -53.952 52.524
670460 97.271 -53.950 52.477
670461 97.298 -53.947 52.429
670462 97.325 -53.944 52.384
670463 97.352 -53.941 52.340
670464 97.378 -53.944 52.293
670465 97.403 -53.950 52.246
670466 97.429 -53.955 52.198
670467 97.454 -53.961 52.153
670468 97.479 -53.967 52.108
670469 97.504 -53.972 52.066
670470 97.529 -53.979 52.024
670471 97.555 -53.985 51.983
670472 97.580 -53.991 51.944
670473 97.605 -53.996 51.907
670474 97.630 -54.002 51.866
670475 97.656 -54.006 51.823
670476 97.682 -54.008 51.778
670477 97.709 -54.010 51.733
670478 97.733 -54.014 51.688
670479 97.749 -54.021 51.642
670480 97.767 -54.024 51.598
670481 97.791 -54.025 51.552
670482 97.814 -54.026 51.506
670483 97.839 -54.026 51.461
670484 97.862 -54.027 51.415
670485 97.886 -54.028 51.367
670486 97.910 -54.029 51.321
670487 97.933 -54.029 51.274
670488 97.957 -54.030 51.227
670489 97.981 -54.031 51.180
670490 98.005 -54.032 51.132
670491 98.028 -54.032 51.083
670492 98.053 -54.033 51.031
670493 98.076 -54.034 50.979
670494 98.100 -54.035 50.929
670495 98.131 -54.044 50.875
670496 98.155 -54.062 50.821
670497 98.180 -54.080 50.765
670498 98.206 -54.098 50.707
670499 98.231 -54.117 50.649
670500 98.257 -54.135 50.591
670501 98.282 -54.154 50.534
670502 98.308 -54.173 50.477
670503 98.333 -54.191 50.420
670504 98.359 -54.209 50.365
670925 90.808 -50.228 56.667
670926 90.842 -50.250 56.675
670927 90.876 -50.271 56.681
670928 90.910 -50.293 56.686
670929 90.945 -50.314 56.690
670930 90.979 -50.336 56.692
670931 91.013 -50.357 56.693
670932 91.047 -50.378 56.693
670933 91.082 -50.400 56.692
670934 91.116 -50.422 56.689
670935 91.150 -50.443 56.685
670936 91.184 -50.464 56.679
670977 92.438 -51.184 56.643
670978 92.461 -51.196 56.636
670979 92.484 -51.208 56.628
670980 92.509 -51.221 56.620
670981 92.538 -51.235 56.612
670982 92.568 -51.248 56.606
670983 92.597 -51.262 56.601
670984 92.626 -51.275 56.595
670985 92.656 -51.288 56.587
670986 92.685 -51.302 56.579
670987 92.715 -51.315 56.572
670988 92.744 -51.329 56.565
670989 92.774 -51.342 56.556
670990 92.803 -51.356 56.548
670991 92.833 -51.369 56.539
670992 92.862 -51.382 56.529
670993 92.892 -51.396 56.519
670994 92.921 -51.409 56.506
670995 92.950 -51.422 56.497
670996 92.980 -51.436 56.486
670997 93.010 -51.448 56.474
670998 93.053 -51.444 56.462
670999 93.092 -51.450 56.450
671000 93.120 -51.480 56.438
671001 93.147 -51.510 56.425
671002 93.174 -51.541 56.409
671003 93.202 -51.571 56.393
671004 93.229 -51.601 56.375
671005 93.257 -51.631 56.357
671006 93.285 -51.662 56.339
671007 93.312 -51.692 56.319
671008 93.340 -51.722 56.298
671009 93.379 -51.753 56.274
671010 93.423 -51.782 56.247
671011 93.467 -51.813 56.218
671012 93.509 -51.842 56.185
671013 93.536 -51.860 56.149
671014 93.563 -51.879 56.108
671015 93.589 -51.898 56.064
671016 93.616 -51.916 56.021
671017 93.643 -51.935 55.979
671018 93.670 -51.953 55.936
671019 93.697 -51.972 55.893
671020 93.723 -51.990 55.851
671021 93.751 -52.009 55.810
671022 93.777 -52.027 55.770
671023 93.804 -52.046 55.729
671024 93.831 -52.064 55.688
671067 95.010 -52.860 54.664
671068 95.039 -52.878 54.657
671069 95.068 -52.896 54.651
671070 95.098 -52.914 54.644
671071 95.127 -52.932 54.638
671072 95.157 -52.951 54.632
671073 95.186 -52.970 54.626
671074 95.215 -52.988 54.621
671075 95.244 -53.001 54.616
671076 95.271 -53.007 54.614
671077 95.300 -53.023 54.610
671078 95.329 -53.045 54.608
671079 95.357 -53.067 54.606
671080 95.387 -53.088 54.606
671081 95.415 -53.110 54.606
671082 95.450 -53.132 54.608
671083 95.485 -53.154 54.611
671084 95.517 -53.177 54.611
671085 95.533 -53.200 54.609
671086 95.548 -53.223 54.604
671087 95.568 -53.242 54.596
671088 95.589 -53.258 54.585
671089 95.610 -53.275 54.573
671090 95.632 -53.292 54.551
671091 95.653 -53.308 54.534
671092 95.674 -53.325 54.528
671093 95.696 -53.342 54.522
671094 95.717 -53.358 54.508
671095 95.739 -53.375 54.482
671096 95.760 -53.392 54.440
671097 95.779 -53.410 54.390
671098 95.797 -53.429 54.342
671099 95.814 -53.448 54.303
671100 95.832 -53.466 54.271
671101 95.850 -53.485 54.243
671102 95.871 -53.502 54.217
671103 95.896 -53.517 54.190
671104 95.917 -53.529 54.161
671105 95.933 -53.537 54.133
671106 95.948 -53.546 54.116
671107 95.964 -53.554 54.099
671108 95.978 -53.562 54.079
671109 95.982 -53.570 54.055
671110 95.986 -53.577 54.025
671111 95.989 -53.584 53.990
671112 95.993 -53.592 53.953
671113 95.997 -53.599 53.913
671114 96.001 -53.606 53.874
671115 96.005 -53.614 53.840
671116 96.008 -53.621 53.811
671117 96.012 -53.629 53.789
671118 96.016 -53.636 53.774
671119 96.019 -53.643 53.754
671120 96.029 -53.650 53.729
671121 96.062 -53.656 53.703
671122 96.095 -53.660 53.678
671123 96.128 -53.665 53.655
671124 96.161 -53.670 53.632
671125 96.194 -53.675 53.610
671126 96.227 -53.680 53.589
671127 96.259 -53.685 53.567
671128 96.289 -53.693 53.544
671129 96.319 -53.700 53.518
671130 96.352 -53.705 53.493
671131 96.384 -53.710 53.471
671132 96.416 -53.715 53.452
671133 96.449 -53.719 53.432
671134 96.482 -53.724 53.413
671135 96.514 -53.729 53.391
671136 96.546 -53.734 53.366
671137 96.576 -53.740 53.338
671138 96.605 -53.747 53.310
671139 96.632 -53.753 53.281
671140 96.660 -53.759 53.251
671141 96.688 -53.766 53.224
671142 96.716 -53.772 53.195
671143 96.746 -53.778 53.162
671144 96.778 -53.785 53.128
671145 96.810 -53.792 53.092
671146 96.842 -53.799 53.056
671147 96.874 -53.806 53.019
671148 96.905 -53.813 52.988
671149 96.937 -53.820 52.958
671150 96.969 -53.827 52.927
671151 97.002 -53.835 52.896
671152 97.033 -53.844 52.862
671153 97.063 -53.855 52.828
671154 97.090 -53.869 52.790
671155 97.117 -53.882 52.751
671156 97.143 -53.893 52.710
671157 97.169 -53.903 52.668
671158 97.195 -53.915 52.623
671159 97.221 -53.925 52.579
671160 97.247 -53.936 52.530
671161 97.274 -53.935 52.483
671162 97.301 -53.932 52.437
671163 97.328 -53.929 52.392
671164 97.356 -53.927 52.347
671165 97.382 -53.926 52.301
671166 97.407 -53.932 52.254
671167 97.433 -53.938 52.207
671168 97.458 -53.944 52.163
671169 97.483 -53.950 52.118
671170 97.508 -53.956 52.076
671171 97.533 -53.961 52.034
671172 97.559 -53.967 51.994
671173 97.584 -53.973 51.954
671174 97.609 -53.978 51.916
671175 97.636 -53.980 51.877
671176 97.662 -53.982 51.835
671177 97.689 -53.985 51.791
671178 97.716 -53.987 51.746
671179 97.741 -53.990 51.702
671180 97.756 -53.998 51.657
671181 97.774 -54.004 51.614
671182 97.791 -54.007 51.570
671183 97.813 -54.008 51.526
671184 97.837 -54.009 51.482
671185 97.861 -54.010 51.436
671186 97.884 -54.010 51.389
671187 97.908 -54.011 51.344
671188 97.932 -54.012 51.296
671189 97.956 -54.013 51.250
671190 97.980 -54.013 51.204
671191 98.003 -54.014 51.156
671192 98.027 -54.015 51.106
671193 98.051 -54.016 51.056
671194 98.075 -54.017 51.005
671195 98.100 -54.019 50.954
671196 98.133 -54.028 50.903
671197 98.157 -54.044 50.850
671198 98.182 -54.063 50.796
671199 98.208 -54.081 50.740
671200 98.233 -54.099 50.683
671201 98.258 -54.118 50.628
671202 98.284 -54.136 50.571
671203 98.309 -54.155 50.515
671204 98.335 -54.173 50.460
671205 98.360 -54.192 50.405
671628 90.889 -50.242 56.670
671629 90.924 -50.264 56.675
671630 90.957 -50.285 56.680
671631 90.992 -50.306 56.684
671632 91.026 -50.328 56.686
671633 91.060 -50.349 56.687
671634 91.094 -50.371 56.686
671635 91.129 -50.392 56.685
671680 92.534 -51.206 56.617
671681 92.557 -51.218 56.608
671682 92.580 -51.230 56.602
671683 92.602 -51.243 56.595
671684 92.626 -51.255 56.589
671685 92.648 -51.268 56.584
671686 92.674 -51.280 56.576
671687 92.704 -51.293 56.569
671688 92.733 -51.307 56.562
671689 92.763 -51.321 56.555
671690 92.792 -51.334 56.548
671691 92.821 -51.347 56.540
671692 92.851 -51.361 56.531
671693 92.880 -51.374 56.522
671694 92.910 -51.387 56.512
671695 92.939 -51.401 56.501
671696 92.969 -51.414 56.492
671697 92.998 -51.428 56.481
671698 93.027 -51.441 56.469
671699 93.057 -51.454 56.457
671700 93.093 -51.459 56.444
671701 93.135 -51.456 56.433
671702 93.163 -51.487 56.419
671703 93.191 -51.517 56.403
671704 93.218 -51.547 56.387
671705 93.246 -51.578 56.370
671706 93.273 -51.608 56.351
671707 93.301 -51.638 56.334
671708 93.328 -51.669 56.314
671709 93.360 -51.699 56.293
671710 93.404 -51.729 56.270
671711 93.448 -51.759 56.244
671712 93.492 -51.789 56.217
671713 93.531 -51.816 56.185
671714 93.558 -51.835 56.148
671715 93.584 -51.853 56.108
671716 93.611 -51.871 56.066
671717 93.638 -51.890 56.025
671718 93.665 -51.908 55.984
671719 93.692 -51.927 55.941
671720 93.718 -51.945 55.899
671721 93.746 -51.964 55.857
671722 93.772 -51.982 55.816
671723 93.799 -52.001 55.776
671724 93.826 -52.020 55.735
671725 93.853 -52.039 55.694
671768 95.020 -52.835 54.661
671769 95.050 -52.853 54.653
671770 95.079 -52.871 54.647
671771 95.108 -52.889 54.640
671772 95.138 -52.907 54.634
671773 95.167 -52.926 54.627
671774 95.196 -52.944 54.622
671775 95.226 -52.963 54.617
671776 95.255 -52.981 54.613
671777 95.283 -52.989 54.610
671778 95.311 -53.002 54.607
671779 95.340 -53.024 54.605
671780 95.369 -53.045 54.605
671781 95.398 -53.067 54.604
671782 95.431 -53.089 54.605
671783 95.466 -53.111 54.608
671784 95.501 -53.134 54.611
671785 95.533 -53.157 54.611
671786 95.549 -53.180 54.608
671787 95.564 -53.202 54.601
671788 95.581 -53.226 54.592
671789 95.602 -53.242 54.580
671790 95.623 -53.259 54.567
671791 95.644 -53.276 54.541
671792 95.666 -53.292 54.520
671793 95.687 -53.309 54.509
671794 95.708 -53.326 54.501
671795 95.730 -53.342 54.488
671796 95.751 -53.359 54.466
671797 95.772 -53.376 54.431
671798 95.794 -53.393 54.384
671799 95.815 -53.409 54.336
671800 95.836 -53.426 54.293
671801 95.858 -53.443 54.256
671802 95.875 -53.462 54.225
671803 95.893 -53.480 54.198
671804 95.913 -53.498 54.173
671805 95.935 -53.511 54.147
671806 95.951 -53.519 54.119
671807 95.966 -53.527 54.100
671808 95.969 -53.535 54.083
671809 95.973 -53.542 54.064
671810 95.977 -53.549 54.042
671811 95.980 -53.557 54.012
671812 95.984 -53.564 53.978
671813 95.988 -53.571 53.942
671814 95.991 -53.579 53.902
671815 95.995 -53.586 53.863
671816 95.999 -53.593 53.828
671817 96.003 -53.601 53.798
671818 96.007 -53.609 53.776
671819 96.010 -53.615 53.762
671820 96.014 -53.623 53.742
671821 96.033 -53.629 53.717
671822 96.065 -53.634 53.692
671823 96.099 -53.639 53.668
671824 96.131 -53.643 53.645
671825 96.165 -53.649 53.623
671826 96.197 -53.653 53.600
671827 96.231 -53.659 53.579
671828 96.263 -53.663 53.558
671829 96.295 -53.670 53.535
671830 96.326 -53.676 53.510
671831 96.359 -53.681 53.487
671832 96.391 -53.686 53.467
671833 96.424 -53.690 53.448
671834 96.456 -53.695 53.429
671835 96.488 -53.700 53.410
671836 96.521 -53.705 53.389
671837 96.553 -53.709 53.364
671838 96.584 -53.715 53.336
671839 96.611 -53.722 53.307
671840 96.639 -53.728 53.279
671841 96.667 -53.734 53.249
671842 96.695 -53.740 53.221
671843 96.723 -53.747 53.193
671844 96.753 -53.753 53.162
671845 96.785 -53.760 53.129
671846 96.816 -53.767 53.093
671847 96.849 -53.774 53.057
671848 96.880 -53.781 53.020
671849 96.912 -53.788 52.988
671850 96.944 -53.795 52.958
671851 96.976 -53.802 52.928
671852 97.008 -53.809 52.897
671853 97.040 -53.816 52.864
671854 97.070 -53.827 52.831
671855 97.096 -53.840 52.792
671856 97.122 -53.850 52.752
671857 97.148 -53.862 52.712
671858 97.174 -53.872 52.670
671859 97.200 -53.883 52.627
671860 97.226 -53.894 52.583
671861 97.252 -53.904 52.535
671862 97.278 -53.915 52.488
671863 97.305 -53.917 52.442
671864 97.332 -53.915 52.398
671865 97.359 -53.912 52.353
671866 97.386 -53.910 52.307
671867 97.412 -53.915 52.260
671868 97.437 -53.921 52.217
671869 97.462 -53.927 52.173
671870 97.487 -53.932 52.129
671871 97.512 -53.938 52.086
671872 97.537 -53.944 52.044
671873 97.563 -53.950 52.004
671874 97.589 -53.953 51.964
671875 97.616 -53.955 51.925
671876 97.642 -53.957 51.886
671877 97.669 -53.960 51.846
671878 97.696 -53.962 51.803
671879 97.722 -53.964 51.759
671880 97.749 -53.966 51.716
671881 97.765 -53.974 51.672
671882 97.781 -53.982 51.631
671883 97.798 -53.986 51.590
671884 97.815 -53.989 51.546
671885 97.835 -53.991 51.502
671886 97.859 -53.992 51.457
671887 97.883 -53.993 51.412
671888 97.907 -53.994 51.366
671889 97.930 -53.994 51.319
671890 97.954 -53.995 51.273
671891 97.978 -53.996 51.227
671892 98.002 -53.997 51.179
671893 98.025 -53.998 51.130
671894 98.049 -53.999 51.081
671895 98.073 -54.000 51.030
671896 98.101 -54.004 50.981
671897 98.134 -54.013 50.931
671898 98.159 -54.026 50.879
671899 98.184 -54.045 50.827
671900 98.210 -54.063 50.772
671901 98.235 -54.082 50.717
671902 98.261 -54.101 50.663
671903 98.286 -54.119 50.607
671904 98.311 -54.137 50.553
671905 98.337 -54.156 50.498
671906 98.362 -54.174 50.444
672383 92.630 -51.227 56.592
672384 92.653 -51.240 56.585
672385 92.675 -51.252 56.578
672386 92.699 -51.265 56.573
672387 92.721 -51.277 56.566
672388 92.744 -51.289 56.559
672389 92.767 -51.302 56.553
672390 92.790 -51.314 56.547
672391 92.813 -51.326 56.541
672392 92.840 -51.339 56.533
672393 92.869 -51.352 56.524
672394 92.899 -51.365 56.516
672395 92.928 -51.379 56.505
672396 92.957 -51.393 56.496
672397 92.987 -51.406 56.487
672398 93.016 -51.419 56.476
672399 93.046 -51.433 56.464
672400 93.075 -51.446 56.452
672401 93.105 -51.459 56.438
672402 93.134 -51.473 56.426
672403 93.176 -51.471 56.412
672404 93.207 -51.494 56.396
672405 93.234 -51.524 56.380
672406 93.262 -51.554 56.363
672407 93.289 -51.584 56.345
672408 93.317 -51.615 56.328
672409 93.344 -51.645 56.308
672410 93.386 -51.675 56.288
672411 93.430 -51.705 56.265
672412 93.474 -51.735 56.240
672413 93.517 -51.766 56.214
672414 93.553 -51.790 56.184
672415 93.579 -51.808 56.147
672416 93.606 -51.826 56.107
672417 93.633 -51.845 56.067
672418 93.660 -51.863 56.027
672419 93.687 -51.882 55.986
672420 93.713 -51.901 55.945
672421 93.741 -51.920 55.903
672422 93.767 -51.938 55.861
672423 93.794 -51.957 55.820
672424 93.821 -51.975 55.780
672425 93.848 -51.994 55.740
672426 93.874 -52.012 55.700
672469 95.031 -52.810 54.658
672470 95.060 -52.828 54.650
672471 95.090 -52.846 54.643
672472 95.119 -52.864 54.636
672473 95.148 -52.883 54.629
672474 95.178 -52.901 54.623
672475 95.207 -52.919 54.618
672476 95.237 -52.938 54.613
672477 95.266 -52.956 54.609
672478 95.295 -52.971 54.606
672479 95.323 -52.981 54.605
672480 95.352 -53.003 54.603
672481 95.381 -53.024 54.603
672482 95.413 -53.046 54.602
672483 95.448 -53.069 54.603
672484 95.482 -53.091 54.605
672485 95.518 -53.114 54.607
672486 95.550 -53.136 54.606
672487 95.565 -53.159 54.602
672488 95.581 -53.182 54.594
672489 95.597 -53.205 54.585
672490 95.614 -53.227 54.574
672491 95.635 -53.243 54.559
672492 95.657 -53.260 54.534
672493 95.678 -53.277 54.507
672494 95.699 -53.293 54.489
672495 95.721 -53.310 54.476
672496 95.742 -53.327 54.463
672497 95.763 -53.343 54.444
672498 95.784 -53.360 54.417
672499 95.806 -53.377 54.374
672500 95.827 -53.393 54.329
672501 95.848 -53.410 54.284
672502 95.870 -53.427 54.243
672503 95.891 -53.443 54.209
672504 95.912 -53.460 54.180
672505 95.934 -53.477 54.156
672506 95.953 -53.492 54.131
672507 95.957 -53.499 54.105
672508 95.960 -53.507 54.086
672509 95.964 -53.515 54.069
672510 95.968 -53.521 54.051
672511 95.972 -53.529 54.029
672512 95.975 -53.537 54.000
672513 95.979 -53.543 53.966
672514 95.983 -53.551 53.930
672515 95.986 -53.559 53.892
672516 95.990 -53.565 53.854
672517 95.994 -53.573 53.818
672518 95.997 -53.581 53.787
672519 96.002 -53.587 53.763
672520 96.005 -53.595 53.748
672521 96.009 -53.603 53.727
672522 96.037 -53.608 53.703
672523 96.070 -53.612 53.679
672524 96.103 -53.618 53.656
672525 96.135 -53.622 53.634
672526 96.169 -53.628 53.613
672527 96.201 -53.632 53.590
672528 96.235 -53.637 53.567
672529 96.267 -53.642 53.547
672530 96.301 -53.647 53.525
672531 96.333 -53.652 53.502
672532 96.365 -53.657 53.480
672533 96.398 -53.662 53.462
672534 96.431 -53.665 53.443
672535 96.463 -53.670 53.424
672536 96.496 -53.675 53.407
672537 96.528 -53.679 53.386
672538 96.560 -53.684 53.360
672539 96.588 -53.690 53.332
672540 96.618 -53.697 53.304
672541 96.646 -53.703 53.276
672542 96.674 -53.709 53.245
672543 96.701 -53.715 53.218
672544 96.729 -53.722 53.191
672545 96.759 -53.728 53.162
672546 96.791 -53.735 53.129
672547 96.823 -53.742 53.093
672548 96.855 -53.749 53.056
672549 96.887 -53.756 53.020
672550 96.919 -53.763 52.987
672551 96.950 -53.770 52.957
672552 96.982 -53.777 52.928
672553 97.014 -53.784 52.897
672554 97.047 -53.790 52.865
672555 97.076 -53.798 52.832
672556 97.101 -53.809 52.792
672557 97.127 -53.819 52.751
672558 97.153 -53.830 52.711
672559 97.179 -53.841 52.670
672560 97.205 -53.851 52.628
672561 97.231 -53.862 52.585
672562 97.257 -53.873 52.539
672563 97.283 -53.884 52.492
672564 97.309 -53.894 52.447
672565 97.335 -53.900 52.402
672566 97.363 -53.897 52.359
672567 97.390 -53.894 52.314
672568 97.416 -53.898 52.268
672569 97.441 -53.903 52.225
672570 97.466 -53.910 52.181
672571 97.491 -53.916 52.139
672572 97.516 -53.922 52.096
672573 97.542 -53.925 52.054
672574 97.569 -53.928 52.014
672575 97.596 -53.930 51.974
672576 97.623 -53.932 51.935
672577 97.649 -53.934 51.896
672578 97.675 -53.936 51.856
672579 97.702 -53.938 51.814
672580 97.729 -53.941 51.772
672581 97.755 -53.943 51.730
672582 97.773 -53.950 51.689
672583 97.789 -53.958 51.649
672584 97.805 -53.966 51.610
672585 97.822 -53.969 51.567
672586 97.840 -53.972 51.523
672587 97.858 -53.975 51.479
672588 97.881 -53.976 51.434
672589 97.905 -53.976 51.389
672590 97.929 -53.977 51.342
672591 97.952 -53.979 51.296
672592 97.976 -53.979 51.249
672593 98.000 -53.980 51.203
672594 98.024 -53.981 51.155
672595 98.048 -53.982 51.106
672596 98.071 -53.982 51.056
672597 98.102 -53.989 51.008
672598 98.135 -53.998 50.958
672599 98.160 -54.009 50.907
672600 98.185 -54.027 50.856
672601 98.211 -54.046 50.804
672602 98.236 -54.064 50.750
672603 98.262 -54.082 50.696
672604 98.287 -54.101 50.643
672605 98.313 -54.120 50.589
672606 98.338 -54.138 50.536
672607 98.364 -54.157 50.483
673086 92.715 -51.243 56.570
673087 92.746 -51.260 56.565
673088 92.772 -51.274 56.558
673089 92.794 -51.287 56.551
673090 92.817 -51.299 56.545
673091 92.841 -51.311 56.539
673092 92.863 -51.323 56.533
673093 92.886 -51.335 56.527
673094 92.909 -51.348 56.518
673095 92.932 -51.360 56.510
673096 92.955 -51.372 56.500
673097 92.978 -51.384 56.492
673098 93.005 -51.397 56.482
673099 93.035 -51.411 56.472
673100 93.064 -51.425 56.460
673101 93.093 -51.437 56.447
673102 93.123 -51.451 56.433
673103 93.152 -51.465 56.419
673104 93.182 -51.478 56.405
673105 93.215 -51.486 56.389
673106 93.250 -51.500 56.373
673107 93.278 -51.531 56.357
673108 93.305 -51.561 56.339
673109 93.333 -51.591 56.322
673110 93.367 -51.622 56.302
673111 93.412 -51.652 56.282
673112 93.455 -51.682 56.260
673113 93.499 -51.712 56.236
673114 93.543 -51.742 56.211
673115 93.574 -51.763 56.181
673116 93.602 -51.782 56.144
673117 93.628 -51.801 56.106
673118 93.655 -51.819 56.068
673119 93.682 -51.838 56.028
673120 93.708 -51.856 55.988
673121 93.735 -51.875 55.947
673122 93.762 -51.893 55.905
673123 93.789 -51.912 55.864
673124 93.816 -51.930 55.823
673125 93.843 -51.949 55.783
673126 93.869 -51.967 55.744
673127 93.896 -51.986 55.704
673170 95.049 -52.784 54.655
673171 95.075 -52.803 54.648
673172 95.102 -52.821 54.640
673173 95.130 -52.840 54.632
673174 95.159 -52.858 54.625
673175 95.189 -52.876 54.619
673176 95.218 -52.894 54.614
673177 95.247 -52.913 54.609
673178 95.277 -52.931 54.605
673179 95.306 -52.949 54.603
673180 95.334 -52.960 54.603
673181 95.363 -52.982 54.602
673182 95.394 -53.003 54.601
673183 95.429 -53.026 54.600
673184 95.464 -53.048 54.599
673185 95.499 -53.070 54.599
673186 95.534 -53.093 54.600
673187 95.566 -53.115 54.600
673188 95.582 -53.139 54.595
673189 95.598 -53.161 54.587
673190 95.613 -53.184 54.577
673191 95.629 -53.208 54.563
673192 95.648 -53.227 54.546
673193 95.669 -53.244 54.522
673194 95.690 -53.261 54.496
673195 95.711 -53.277 54.470
673196 95.733 -53.294 54.449
673197 95.754 -53.311 54.432
673198 95.775 -53.327 54.416
673199 95.797 -53.344 54.392
673200 95.818 -53.361 54.359
673201 95.839 -53.377 54.320
673202 95.861 -53.394 54.277
673203 95.882 -53.411 54.234
673204 95.904 -53.427 54.197
673205 95.925 -53.444 54.164
673206 95.946 -53.461 54.137
673207 95.956 -53.470 54.115
673208 95.962 -53.475 54.092
673209 95.965 -53.483 54.073
673210 95.969 -53.490 54.056
673211 95.972 -53.498 54.038
673212 95.976 -53.505 54.016
673213 95.979 -53.513 53.989
673214 95.982 -53.520 53.955
673215 95.986 -53.527 53.920
673216 95.989 -53.535 53.883
673217 95.992 -53.543 53.845
673218 95.996 -53.550 53.809
673219 96.000 -53.558 53.777
673220 96.003 -53.565 53.750
673221 96.006 -53.572 53.732
673222 96.014 -53.577 53.710
673223 96.048 -53.580 53.687
673224 96.086 -53.583 53.665
673225 96.120 -53.587 53.643
673226 96.154 -53.592 53.622
673227 96.187 -53.596 53.601
673228 96.221 -53.600 53.578
673229 96.255 -53.605 53.555
673230 96.289 -53.609 53.534
673231 96.323 -53.613 53.513
673232 96.340 -53.621 53.492
673233 96.373 -53.626 53.473
673234 96.405 -53.631 53.454
673235 96.438 -53.635 53.436
673236 96.471 -53.640 53.418
673237 96.503 -53.643 53.400
673238 96.536 -53.648 53.379
673239 96.566 -53.654 53.354
673240 96.587 -53.665 53.327
673241 96.616 -53.672 53.299
673242 96.645 -53.678 53.270
673243 96.675 -53.684 53.241
673244 96.704 -53.690 53.215
673245 96.734 -53.697 53.190
673246 96.767 -53.702 53.161
673247 96.800 -53.708 53.128
673248 96.832 -53.714 53.094
673249 96.865 -53.720 53.056
673250 96.897 -53.726 53.019
673251 96.930 -53.732 52.986
673252 96.962 -53.737 52.956
673253 96.995 -53.744 52.928
673254 97.028 -53.750 52.897
673255 97.058 -53.757 52.866
673256 97.083 -53.769 52.832
673257 97.108 -53.779 52.791
673258 97.134 -53.790 52.749
673259 97.159 -53.800 52.709
673260 97.185 -53.809 52.670
673261 97.211 -53.820 52.629
673262 97.236 -53.831 52.587
673263 97.262 -53.841 52.542
673264 97.288 -53.852 52.496
673265 97.314 -53.863 52.452
673266 97.340 -53.873 52.408
673267 97.366 -53.882 52.365
673268 97.393 -53.879 52.321
673269 97.420 -53.881 52.277
673270 97.445 -53.887 52.233
673271 97.470 -53.892 52.190
673272 97.496 -53.897 52.149
673273 97.522 -53.900 52.107
673274 97.549 -53.902 52.065
673275 97.576 -53.904 52.024
673276 97.602 -53.907 51.984
673277 97.629 -53.909 51.946
673278 97.655 -53.911 51.907
673279 97.682 -53.913 51.867
673280 97.709 -53.916 51.827
673281 97.735 -53.918 51.786
673282 97.762 -53.920 51.745
673283 97.782 -53.926 51.707
673284 97.797 -53.934 51.668
673285 97.813 -53.942 51.628
673286 97.829 -53.948 51.586
673287 97.847 -53.951 51.543
673288 97.865 -53.954 51.499
673289 97.882 -53.957 51.454
673290 97.903 -53.960 51.409
673291 97.927 -53.960 51.364
673292 97.951 -53.961 51.319
673293 97.974 -53.962 51.274
673294 97.999 -53.963 51.227
673295 98.022 -53.963 51.179
673296 98.046 -53.964 51.131
673297 98.070 -53.966 51.083
673298 98.103 -53.974 51.034
673299 98.136 -53.983 50.985
673300 98.165 -53.992 50.935
673301 98.187 -54.010 50.885
673302 98.213 -54.028 50.833
673303 98.238 -54.047 50.780
673304 98.264 -54.065 50.728
673305 98.289 -54.084 50.675
673306 98.314 -54.102 50.623
673307 98.340 -54.120 50.571
673308 98.365 -54.139 50.520
673789 92.794 -51.253 56.550
673790 92.825 -51.271 56.543
673791 92.856 -51.288 56.537
673792 92.886 -51.305 56.532
673793 92.914 -51.321 56.526
673794 92.936 -51.333 56.520
673795 92.959 -51.345 56.512
673796 92.982 -51.357 56.503
673797 93.005 -51.369 56.495
673798 93.028 -51.382 56.486
673799 93.051 -51.394 56.477
673800 93.074 -51.406 56.467
673801 93.097 -51.418 56.454
673802 93.120 -51.431 56.441
673803 93.143 -51.443 56.428
673804 93.170 -51.456 56.414
673805 93.200 -51.469 56.398
673806 93.229 -51.483 56.382
673807 93.259 -51.497 56.366
673808 93.294 -51.507 56.350
673809 93.321 -51.537 56.332
673810 93.349 -51.568 56.315
673811 93.393 -51.598 56.295
673812 93.437 -51.628 56.275
673813 93.481 -51.658 56.254
673814 93.525 -51.688 56.231
673815 93.569 -51.718 56.206
673816 93.596 -51.738 56.176
673817 93.623 -51.756 56.140
673818 93.650 -51.775 56.104
673819 93.677 -51.793 56.066
673820 93.703 -51.811 56.027
673821 93.730 -51.830 55.988
673822 93.757 -51.848 55.947
673823 93.784 -51.867 55.906
673824 93.811 -51.885 55.865
673825 93.838 -51.904 55.825
673826 93.864 -51.923 55.785
673827 93.891 -51.942 55.746
673828 93.918 -51.960 55.707
673871 95.070 -52.758 54.653
673872 95.098 -52.776 54.645
673873 95.124 -52.795 54.638
673874 95.151 -52.813 54.629
673875 95.178 -52.832 54.622
673876 95.204 -52.850 54.616
673877 95.232 -52.869 54.610
673878 95.258 -52.888 54.605
673879 95.287 -52.906 54.602
673880 95.317 -52.924 54.601
673881 95.346 -52.941 54.601
673882 95.376 -52.960 54.600
673883 95.410 -52.982 54.599
673884 95.445 -53.005 54.596
673885 95.480 -53.027 54.594
673886 95.515 -53.050 54.592
673887 95.550 -53.073 54.592
673888 95.583 -53.095 54.593
673889 95.598 -53.118 54.590
673890 95.614 -53.141 54.582
673891 95.629 -53.164 54.568
673892 95.645 -53.187 54.551
673893 95.661 -53.210 54.532
673894 95.681 -53.227 54.509
673895 95.703 -53.244 54.482
673896 95.724 -53.261 54.453
673897 95.745 -53.278 54.426
673898 95.767 -53.295 54.405
673899 95.788 -53.311 54.386
673900 95.809 -53.328 54.365
673901 95.831 -53.345 54.339
673902 95.852 -53.361 54.307
673903 95.873 -53.378 54.269
673904 95.895 -53.395 54.228
673905 95.916 -53.411 54.187
673906 95.938 -53.428 54.152
673907 95.954 -53.442 54.123
673908 95.960 -53.447 54.099
673909 95.965 -53.453 54.077
673910 95.971 -53.459 54.060
673911 95.975 -53.466 54.044
673912 95.978 -53.474 54.024
673913 95.982 -53.481 54.003
673914 95.985 -53.488 53.976
673915 95.989 -53.496 53.945
673916 95.992 -53.503 53.911
673917 95.995 -53.511 53.874
673918 95.999 -53.518 53.837
673919 96.003 -53.526 53.801
673920 96.006 -53.533 53.769
673921 96.009 -53.540 53.739
673922 96.018 -53.545 53.715
673923 96.028 -53.549 53.692
673924 96.048 -53.552 53.671
673925 96.091 -53.554 53.651
673926 96.132 -53.556 53.630
673927 96.175 -53.557 53.609
673928 96.212 -53.561 53.588
673929 96.246 -53.565 53.565
673930 96.279 -53.569 53.541
673931 96.313 -53.574 53.521
673932 96.347 -53.577 53.500
673933 96.363 -53.587 53.481
673934 96.381 -53.596 53.463
673935 96.413 -53.599 53.445
673936 96.446 -53.604 53.428
673937 96.478 -53.609 53.410
673938 96.511 -53.613 53.392
673939 96.543 -53.618 53.371
673940 96.567 -53.627 53.347
673941 96.588 -53.637 53.320
673942 96.613 -53.646 53.292
673943 96.643 -53.653 53.263
673944 96.673 -53.659 53.234
673945 96.702 -53.665 53.208
673946 96.736 -53.669 53.184
673947 96.778 -53.670 53.157
673948 96.813 -53.674 53.125
673949 96.846 -53.680 53.090
673950 96.878 -53.686 53.053
673951 96.911 -53.692 53.015
673952 96.944 -53.698 52.983
673953 96.976 -53.704 52.955
673954 97.009 -53.709 52.927
673955 97.041 -53.715 52.897
673956 97.066 -53.728 52.865
673957 97.091 -53.739 52.829
673958 97.115 -53.750 52.788
673959 97.141 -53.761 52.747
673960 97.166 -53.771 52.708
673961 97.192 -53.781 52.669
673962 97.218 -53.791 52.629
673963 97.243 -53.801 52.589
673964 97.268 -53.811 52.544
673965 97.294 -53.822 52.501
673966 97.319 -53.831 52.458
673967 97.345 -53.842 52.415
673968 97.371 -53.853 52.371
673969 97.397 -53.863 52.328
673970 97.424 -53.863 52.286
673971 97.449 -53.869 52.242
673972 97.475 -53.872 52.200
673973 97.502 -53.875 52.159
673974 97.529 -53.877 52.118
673975 97.556 -53.879 52.075
673976 97.582 -53.882 52.034
673977 97.609 -53.884 51.995
673978 97.636 -53.885 51.959
673979 97.662 -53.888 51.919
673980 97.688 -53.890 51.880
673981 97.715 -53.892 51.841
673982 97.742 -53.894 51.800
673983 97.769 -53.897 51.761
673984 97.790 -53.902 51.724
673985 97.805 -53.910 51.685
673986 97.821 -53.918 51.645
673987 97.837 -53.926 51.604
673988 97.854 -53.932 51.561
673989 97.871 -53.935 51.518
673990 97.889 -53.937 51.474
673991 97.907 -53.940 51.429
673992 97.926 -53.943 51.385
673993 97.949 -53.944 51.342
673994 97.973 -53.944 51.297
673995 97.997 -53.945 51.252
673996 98.021 -53.946 51.205
673997 98.044 -53.947 51.157
673998 98.071 -53.950 51.110
673999 98.104 -53.959 51.061
674000 98.136 -53.968 51.011
674001 98.169 -53.977 50.962
674002 98.189 -53.992 50.913
674003 98.215 -54.010 50.862
674004 98.240 -54.029 50.811
674005 98.266 -54.048 50.760
674006 98.291 -54.066 50.708
674007 98.316 -54.085 50.657
674008 98.342 -54.103 50.606
674009 98.367 -54.121 50.555
674491 92.842 -51.246 56.536
674492 92.873 -51.264 56.530
674493 92.904 -51.281 56.525
674494 92.935 -51.299 56.520
674495 92.966 -51.316 56.514
674496 92.997 -51.334 56.506
674497 93.027 -51.351 56.497
674498 93.055 -51.367 56.489
674499 93.078 -51.379 56.481
674500 93.101 -51.391 56.472
674501 93.124 -51.403 56.462
674502 93.147 -51.415 56.449
674503 93.170 -51.428 56.436
674504 93.193 -51.440 56.423
674505 93.216 -51.453 56.408
674506 93.238 -51.465 56.393
674507 93.262 -51.477 56.377
674508 93.284 -51.489 56.360
674509 93.307 -51.502 56.342
674510 93.337 -51.514 56.325
674511 93.374 -51.544 56.308
674512 93.419 -51.575 56.289
674513 93.463 -51.604 56.269
674514 93.506 -51.635 56.247
674515 93.550 -51.665 56.224
674516 93.591 -51.693 56.199
674517 93.618 -51.711 56.169
674518 93.645 -51.730 56.134
674519 93.672 -51.748 56.099
674520 93.699 -51.767 56.063
674521 93.725 -51.785 56.025
674522 93.752 -51.804 55.985
674523 93.779 -51.823 55.945
674524 93.805 -51.841 55.905
674525 93.833 -51.860 55.865
674526 93.859 -51.878 55.825
674527 93.886 -51.897 55.785
674528 93.913 -51.915 55.746
674529 93.940 -51.934 55.707
674572 95.093 -52.731 54.651
674573 95.119 -52.750 54.643
674574 95.146 -52.769 54.635
674575 95.173 -52.788 54.626
674576 95.199 -52.806 54.619
674577 95.227 -52.825 54.612
674578 95.253 -52.843 54.606
674579 95.280 -52.862 54.600
674580 95.307 -52.881 54.597
674581 95.333 -52.900 54.596
674582 95.360 -52.919 54.596
674583 95.392 -52.940 54.595
674584 95.426 -52.962 54.592
674585 95.461 -52.985 54.588
674586 95.496 -53.007 54.585
674587 95.531 -53.029 54.584
674588 95.566 -53.052 54.588
674589 95.599 -53.074 54.590
674590 95.615 -53.098 54.588
674591 95.630 -53.120 54.578
674592 95.646 -53.143 54.560
674593 95.662 -53.167 54.538
674594 95.677 -53.189 54.517
674595 95.694 -53.211 54.493
674596 95.715 -53.228 54.467
674597 95.737 -53.245 54.437
674598 95.758 -53.261 54.407
674599 95.779 -53.278 54.380
674600 95.801 -53.295 54.358
674601 95.822 -53.311 54.339
674602 95.843 -53.329 54.317
674603 95.865 -53.346 54.292
674604 95.886 -53.362 54.260
674605 95.907 -53.379 54.221
674606 95.928 -53.396 54.179
674607 95.950 -53.412 54.141
674608 95.958 -53.419 54.109
674609 95.963 -53.425 54.084
674610 95.969 -53.431 54.062
674611 95.975 -53.437 54.044
674612 95.980 -53.443 54.028
674613 95.985 -53.449 54.010
674614 95.988 -53.456 53.990
674615 95.992 -53.464 53.965
674616 95.995 -53.471 53.934
674617 95.999 -53.479 53.900
674618 96.002 -53.487 53.865
674619 96.006 -53.494 53.828
674620 96.009 -53.501 53.792
674621 96.013 -53.509 53.759
674622 96.023 -53.512 53.727
674623 96.033 -53.517 53.700
674624 96.043 -53.521 53.675
674625 96.053 -53.524 53.654
674626 96.091 -53.527 53.634
674627 96.133 -53.528 53.614
674628 96.175 -53.530 53.593
674629 96.218 -53.531 53.573
674630 96.259 -53.533 53.550
674631 96.302 -53.534 53.527
674632 96.337 -53.538 53.505
674633 96.372 -53.543 53.486
674634 96.387 -53.552 53.469
674635 96.401 -53.562 53.452
674636 96.421 -53.569 53.435
674637 96.454 -53.574 53.418
674638 96.486 -53.578 53.401
674639 96.519 -53.582 53.383
674640 96.547 -53.589 53.362
674641 96.569 -53.599 53.338
674642 96.590 -53.610 53.311
674643 96.611 -53.621 53.283
674644 96.641 -53.628 53.255
674645 96.670 -53.634 53.227
674646 96.700 -53.640 53.201
674647 96.742 -53.640 53.176
674648 96.783 -53.641 53.151
674649 96.824 -53.641 53.120
674650 96.859 -53.646 53.085
674651 96.892 -53.652 53.049
674652 96.925 -53.658 53.011
674653 96.957 -53.664 52.979
674654 96.990 -53.670 52.950
674655 97.022 -53.676 52.922
674656 97.049 -53.686 52.893
674657 97.074 -53.697 52.859
674658 97.098 -53.709 52.822
674659 97.123 -53.721 52.782
674660 97.148 -53.732 52.742
674661 97.173 -53.742 52.704
674662 97.199 -53.752 52.666
674663 97.224 -53.762 52.628
674664 97.250 -53.772 52.589
674665 97.276 -53.782 52.546
674666 97.301 -53.793 52.503
674667 97.326 -53.803 52.461
674668 97.352 -53.813 52.419
674669 97.378 -53.823 52.377
674670 97.403 -53.833 52.335
674671 97.428 -53.843 52.294
674672 97.455 -53.847 52.252
674673 97.482 -53.849 52.210
674674 97.509 -53.851 52.169
674675 97.535 -53.853 52.128
674676 97.562 -53.856 52.085
674677 97.589 -53.858 52.044
674678 97.615 -53.860 52.006
674679 97.642 -53.863 51.971
674680 97.668 -53.865 51.933
674681 97.695 -53.867 51.895
674682 97.722 -53.869 51.855
674683 97.748 -53.872 51.814
674684 97.775 -53.874 51.775
674685 97.798 -53.878 51.739
674686 97.814 -53.886 51.700
674687 97.829 -53.894 51.660
674688 97.845 -53.902 51.619
674689 97.861 -53.910 51.578
674690 97.878 -53.914 51.535
674691 97.896 -53.917 51.493
674692 97.914 -53.920 51.451
674693 97.931 -53.922 51.408
674694 97.949 -53.925 51.365
674695 97.971 -53.927 51.321
674696 97.995 -53.928 51.276
674697 98.019 -53.928 51.231
674698 98.043 -53.929 51.183
674699 98.072 -53.935 51.136
674700 98.105 -53.944 51.087
674701 98.137 -53.953 51.038
674702 98.170 -53.962 50.989
674703 98.191 -53.975 50.940
674704 98.216 -53.993 50.891
674705 98.242 -54.011 50.841
674706 98.267 -54.030 50.791
674707 98.293 -54.048 50.740
674708 98.318 -54.067 50.690
674709 98.343 -54.085 50.640
674710 98.369 -54.104 50.589
675193 92.890 -51.240 56.524
675194 92.921 -51.257 56.519
675195 92.951 -51.274 56.513
675196 92.983 -51.292 56.508
675197 93.013 -51.309 56.500
675198 93.045 -51.327 56.491
675199 93.075 -51.344 56.483
675200 93.107 -51.362 56.475
675201 93.137 -51.380 56.466
675202 93.168 -51.397 56.456
675203 93.197 -51.413 56.444
675204 93.220 -51.425 56.430
675205 93.243 -51.437 56.417
675206 93.266 -51.450 56.403
675207 93.289 -51.462 56.387
675208 93.311 -51.475 56.372
675209 93.335 -51.487 56.354
675210 93.357 -51.499 56.337
675211 93.380 -51.511 56.320
675212 93.404 -51.523 56.300
675213 93.444 -51.550 56.281
675214 93.488 -51.581 56.261
675215 93.532 -51.611 56.240
675216 93.576 -51.641 56.217
675217 93.613 -51.666 56.190
675218 93.640 -51.685 56.161
675219 93.666 -51.703 56.127
675220 93.694 -51.722 56.093
675221 93.720 -51.741 56.058
675222 93.747 -51.760 56.021
675223 93.774 -51.778 55.981
675224 93.801 -51.796 55.942
675225 93.828 -51.815 55.902
675226 93.854 -51.833 55.863
675227 93.881 -51.852 55.823
675228 93.908 -51.870 55.784
675229 93.935 -51.889 55.745
675273 95.114 -52.709 54.650
675274 95.140 -52.728 54.641
675275 95.167 -52.747 54.631
675276 95.194 -52.767 54.623
675277 95.220 -52.786 54.615
675278 95.247 -52.805 54.608
675279 95.274 -52.824 54.601
675280 95.301 -52.843 54.595
675281 95.327 -52.862 54.591
675282 95.354 -52.881 54.589
675283 95.381 -52.900 54.588
675284 95.408 -52.919 54.587
675285 95.443 -52.941 54.584
675286 95.477 -52.964 54.579
675287 95.513 -52.986 54.575
675288 95.547 -53.009 54.578
675289 95.582 -53.031 54.583
675290 95.615 -53.054 54.585
675291 95.631 -53.076 54.582
675292 95.647 -53.100 54.571
675293 95.662 -53.123 54.552
675294 95.678 -53.146 54.528
675295 95.694 -53.169 54.503
675296 95.709 -53.192 54.478
675297 95.728 -53.212 54.452
675298 95.749 -53.229 54.423
675299 95.770 -53.245 54.392
675300 95.792 -53.262 54.361
675301 95.813 -53.279 54.335
675302 95.834 -53.295 54.314
675303 95.855 -53.312 54.294
675304 95.877 -53.329 54.272
675305 95.898 -53.346 54.245
675306 95.919 -53.362 54.212
675307 95.941 -53.379 54.173
675308 95.956 -53.391 54.131
675309 95.961 -53.397 54.097
675310 95.967 -53.403 54.069
675311 95.972 -53.408 54.046
675312 95.978 -53.415 54.027
675313 95.984 -53.421 54.011
675314 95.989 -53.426 53.994
675315 95.995 -53.432 53.976
675316 95.998 -53.440 53.952
675317 96.002 -53.447 53.923
675318 96.005 -53.455 53.890
675319 96.008 -53.462 53.854
675320 96.012 -53.469 53.817
675321 96.017 -53.476 53.780
675322 96.027 -53.480 53.746
675323 96.038 -53.483 53.713
675324 96.048 -53.488 53.685
675325 96.058 -53.492 53.659
675326 96.068 -53.496 53.636
675327 96.092 -53.499 53.616
675328 96.134 -53.501 53.595
675329 96.176 -53.502 53.575
675330 96.218 -53.504 53.554
675331 96.260 -53.505 53.532
675332 96.302 -53.507 53.509
675333 96.344 -53.509 53.487
675334 96.380 -53.512 53.468
675335 96.411 -53.516 53.451
675336 96.425 -53.526 53.436
675337 96.439 -53.536 53.418
675338 96.461 -53.543 53.402
675339 96.494 -53.547 53.387
675340 96.526 -53.552 53.370
675341 96.549 -53.562 53.349
675342 96.570 -53.572 53.325
675343 96.591 -53.583 53.298
675344 96.612 -53.594 53.271
675345 96.639 -53.602 53.245
675346 96.668 -53.609 53.219
675347 96.705 -53.611 53.193
675348 96.747 -53.611 53.167
675349 96.789 -53.612 53.141
675350 96.830 -53.612 53.113
675351 96.872 -53.613 53.079
675352 96.905 -53.618 53.044
675353 96.938 -53.624 53.007
675354 96.971 -53.630 52.973
675355 97.003 -53.636 52.944
675356 97.032 -53.645 52.916
675357 97.056 -53.656 52.885
675358 97.081 -53.668 52.851
675359 97.106 -53.679 52.813
675360 97.131 -53.690 52.774
675361 97.155 -53.703 52.736
675362 97.180 -53.713 52.699
675363 97.206 -53.723 52.663
675364 97.231 -53.734 52.626
675365 97.257 -53.744 52.588
675366 97.283 -53.753 52.548
675367 97.308 -53.764 52.506
675368 97.333 -53.774 52.464
675369 97.359 -53.784 52.423
675370 97.385 -53.794 52.382
675371 97.410 -53.804 52.341
675372 97.432 -53.815 52.300
675373 97.460 -53.820 52.259
675374 97.489 -53.826 52.218
675375 97.515 -53.828 52.177
675376 97.542 -53.831 52.135
675377 97.569 -53.833 52.094
675378 97.595 -53.835 52.054
675379 97.622 -53.838 52.018
675380 97.649 -53.839 51.982
675381 97.675 -53.841 51.946
675382 97.702 -53.844 51.908
675383 97.728 -53.846 51.868
675384 97.755 -53.848 51.828
675385 97.782 -53.850 51.789
675386 97.806 -53.853 51.752
675387 97.822 -53.862 51.713
675388 97.838 -53.870 51.674
675389 97.854 -53.878 51.634
675390 97.869 -53.886 51.594
675391 97.885 -53.894 51.553
675392 97.903 -53.897 51.512
675393 97.921 -53.900 51.471
675394 97.938 -53.903 51.429
675395 97.956 -53.906 51.388
675396 97.973 -53.909 51.345
675397 97.994 -53.910 51.301
675398 98.017 -53.911 51.257
675399 98.041 -53.912 51.211
675400 98.073 -53.920 51.164
675401 98.106 -53.929 51.114
675402 98.138 -53.938 51.064
675403 98.171 -53.947 51.016
675404 98.196 -53.958 50.967
675405 98.218 -53.976 50.919
675406 98.243 -53.994 50.870
675407 98.269 -54.013 50.820
675408 98.294 -54.031 50.771
675409 98.319 -54.049 50.721
675410 98.345 -54.068 50.672
675411 98.370 -54.086 50.623
675896 92.969 -51.250 56.507
675897 93.000 -51.268 56.502
675898 93.030 -51.285 56.493
675899 93.062 -51.302 56.486
675900 93.092 -51.320 56.477
675901 93.124 -51.338 56.469
675902 93.154 -51.356 56.460
675903 93.186 -51.373 56.450
675904 93.216 -51.390 56.438
675905 93.248 -51.408 56.424
675906 93.278 -51.425 56.412
675907 93.309 -51.443 56.398
675908 93.339 -51.459 56.383
675909 93.362 -51.472 56.367
675910 93.384 -51.484 56.349
675911 93.408 -51.496 56.331
675912 93.431 -51.509 56.313
675913 93.453 -51.521 56.294
675914 93.477 -51.533 56.273
675915 93.513 -51.557 56.253
675916 93.557 -51.587 56.232
675917 93.602 -51.617 56.208
675918 93.635 -51.641 56.181
675919 93.661 -51.659 56.152
675920 93.689 -51.678 56.119
675921 93.715 -51.696 56.085
675922 93.742 -51.715 56.051
675923 93.769 -51.733 56.014
675924 93.796 -51.752 55.975
675925 93.823 -51.770 55.936
675926 93.849 -51.788 55.897
675927 93.876 -51.807 55.859
675928 93.903 -51.826 55.820
675929 93.930 -51.845 55.781
675930 93.957 -51.863 55.743
675974 95.134 -52.691 54.647
675975 95.161 -52.710 54.638
675976 95.188 -52.729 54.628
675977 95.214 -52.748 54.619
675978 95.241 -52.767 54.611
675979 95.268 -52.786 54.603
675980 95.295 -52.805 54.596
675981 95.321 -52.824 54.590
675982 95.348 -52.843 54.584
675983 95.375 -52.863 54.579
675984 95.401 -52.882 54.578
675985 95.428 -52.901 54.577
675986 95.459 -52.921 54.574
675987 95.494 -52.944 54.568
675988 95.529 -52.966 54.565
675989 95.563 -52.988 54.571
675990 95.599 -53.010 54.575
675991 95.631 -53.033 54.576
675992 95.648 -53.056 54.573
675993 95.663 -53.079 54.562
675994 95.679 -53.102 54.542
675995 95.694 -53.125 54.518
675996 95.710 -53.148 54.492
675997 95.726 -53.171 54.464
675998 95.742 -53.195 54.438
675999 95.761 -53.213 54.409
676000 95.782 -53.230 54.378
676001 95.804 -53.246 54.346
676002 95.825 -53.263 54.315
676003 95.846 -53.280 54.291
676004 95.868 -53.296 54.270
676005 95.889 -53.313 54.248
676006 95.910 -53.330 54.223
676007 95.932 -53.346 54.194
676008 95.953 -53.363 54.160
676009 95.959 -53.369 54.117
676010 95.965 -53.375 54.083
676011 95.970 -53.380 54.053
676012 95.976 -53.386 54.028
676013 95.982 -53.393 54.008
676014 95.987 -53.398 53.992
676015 95.993 -53.404 53.977
676016 95.999 -53.409 53.960
676017 96.005 -53.415 53.937
676018 96.008 -53.423 53.911
676019 96.011 -53.430 53.880
676020 96.015 -53.437 53.844
676021 96.022 -53.443 53.804
676022 96.032 -53.447 53.766
676023 96.042 -53.451 53.731
676024 96.052 -53.455 53.697
676025 96.062 -53.459 53.669
676026 96.073 -53.463 53.642
676027 96.083 -53.468 53.618
676028 96.093 -53.471 53.596
676029 96.135 -53.473 53.575
676030 96.177 -53.474 53.555
676031 96.219 -53.476 53.534
676032 96.261 -53.478 53.512
676033 96.296 -53.481 53.489
676034 96.328 -53.486 53.468
676035 96.361 -53.490 53.449
676036 96.393 -53.493 53.431
676037 96.426 -53.498 53.417
676038 96.458 -53.502 53.399
676039 96.476 -53.510 53.383
676040 96.502 -53.516 53.370
676041 96.529 -53.524 53.354
676042 96.550 -53.534 53.333
676043 96.572 -53.545 53.309
676044 96.593 -53.556 53.282
676045 96.614 -53.567 53.258
676046 96.636 -53.577 53.233
676047 96.670 -53.581 53.208
676048 96.712 -53.581 53.181
676049 96.753 -53.582 53.155
676050 96.794 -53.583 53.129
676051 96.836 -53.583 53.101
676052 96.877 -53.584 53.068
676053 96.919 -53.584 53.033
676054 96.952 -53.590 52.998
676055 96.984 -53.596 52.965
676056 97.015 -53.603 52.935
676057 97.039 -53.615 52.906
676058 97.064 -53.627 52.874
676059 97.089 -53.638 52.840
676060 97.113 -53.650 52.803
676061 97.138 -53.661 52.765
676062 97.163 -53.672 52.728
676063 97.187 -53.684 52.693
676064 97.213 -53.695 52.658
676065 97.238 -53.705 52.623
676066 97.264 -53.715 52.587
676067 97.290 -53.725 52.548
676068 97.315 -53.735 52.508
676069 97.341 -53.745 52.467
676070 97.366 -53.756 52.427
676071 97.392 -53.766 52.386
676072 97.415 -53.775 52.347
676073 97.437 -53.786 52.306
676074 97.464 -53.791 52.266
676075 97.493 -53.797 52.224
676076 97.521 -53.803 52.182
676077 97.548 -53.807 52.141
676078 97.575 -53.809 52.101
676079 97.602 -53.812 52.063
676080 97.629 -53.814 52.029
676081 97.655 -53.816 51.993
676082 97.681 -53.819 51.957
676083 97.708 -53.821 51.920
676084 97.735 -53.823 51.881
676085 97.761 -53.825 51.841
676086 97.788 -53.828 51.802
676087 97.815 -53.830 51.764
676088 97.830 -53.838 51.727
676089 97.846 -53.846 51.688
676090 97.862 -53.854 51.649
676091 97.877 -53.862 51.610
676092 97.893 -53.870 51.571
676093 97.910 -53.876 51.531
676094 97.927 -53.879 51.490
676095 97.945 -53.882 51.449
676096 97.963 -53.885 51.409
676097 97.981 -53.888 51.367
676098 97.998 -53.891 51.325
676099 98.016 -53.894 51.282
676100 98.041 -53.897 51.237
676101 98.074 -53.905 51.190
676102 98.106 -53.914 51.140
676103 98.139 -53.923 51.091
676104 98.171 -53.932 51.042
676105 98.200 -53.941 50.995
676106 98.220 -53.958 50.946
676107 98.245 -53.976 50.898
676108 98.271 -53.994 50.850
676109 98.296 -54.013 50.801
676110 98.321 -54.032 50.752
676111 98.347 -54.050 50.704
676112 98.372 -54.069 50.656
676113 98.397 -54.087 50.608
676598 93.017 -51.243 56.496
676599 93.048 -51.261 56.487
676600 93.079 -51.278 56.479
676601 93.110 -51.296 56.471
676602 93.141 -51.314 56.462
676603 93.171 -51.331 56.453
676604 93.203 -51.349 56.444
676605 93.233 -51.366 56.432
676606 93.264 -51.384 56.419
676607 93.295 -51.401 56.405
676608 93.326 -51.418 56.392
676609 93.357 -51.436 56.377
676610 93.388 -51.453 56.361
676611 93.419 -51.471 56.343
676612 93.450 -51.488 56.326
676613 93.481 -51.506 56.307
676614 93.504 -51.518 56.287
676615 93.526 -51.530 56.267
676616 93.550 -51.542 56.246
676617 93.583 -51.563 56.224
676618 93.627 -51.594 56.200
676619 93.656 -51.614 56.173
676620 93.684 -51.633 56.143
676621 93.710 -51.651 56.110
676622 93.737 -51.670 56.077
676623 93.764 -51.688 56.042
676624 93.791 -51.707 56.006
676625 93.818 -51.725 55.969
676626 93.844 -51.744 55.930
676627 93.871 -51.763 55.892
676628 93.898 -51.781 55.853
676629 93.925 -51.800 55.815
676630 93.952 -51.818 55.776
676631 93.978 -51.837 55.738
676676 95.182 -52.691 54.635
676677 95.208 -52.710 54.625
676678 95.235 -52.729 54.615
676679 95.262 -52.748 54.607
676680 95.288 -52.767 54.598
676681 95.315 -52.787 54.590
676682 95.342 -52.806 54.583
676683 95.369 -52.825 54.576
676684 95.395 -52.844 54.570
676685 95.422 -52.863 54.566
676686 95.449 -52.882 54.564
676687 95.476 -52.901 54.561
676688 95.510 -52.923 54.556
676689 95.545 -52.945 54.555
676690 95.580 -52.968 54.559
676691 95.615 -52.990 54.562
676692 95.648 -53.013 54.563
676693 95.664 -53.035 54.561
676694 95.679 -53.059 54.550
676695 95.695 -53.082 54.531
676696 95.711 -53.104 54.509
676697 95.727 -53.128 54.482
676698 95.742 -53.151 54.453
676699 95.758 -53.174 54.424
676700 95.774 -53.197 54.396
676701 95.795 -53.214 54.365
676702 95.816 -53.230 54.331
676703 95.838 -53.247 54.297
676704 95.859 -53.264 54.269
676705 95.880 -53.280 54.246
676706 95.902 -53.297 54.223
676707 95.923 -53.314 54.199
676708 95.944 -53.330 54.172
676709 95.957 -53.341 54.141
676710 95.963 -53.347 54.100
676711 95.968 -53.352 54.066
676712 95.974 -53.358 54.035
676713 95.979 -53.364 54.009
676714 95.985 -53.370 53.988
676715 95.991 -53.376 53.972
676716 95.996 -53.381 53.959
676717 96.002 -53.387 53.943
676718 96.008 -53.393 53.921
676719 96.014 -53.399 53.895
676720 96.018 -53.405 53.865
676721 96.027 -53.411 53.828
676722 96.037 -53.415 53.788
676723 96.047 -53.418 53.747
676724 96.057 -53.423 53.712
676725 96.067 -53.427 53.680
676726 96.077 -53.430 53.650
676727 96.087 -53.435 53.622
676728 96.097 -53.439 53.598
676729 96.107 -53.443 53.575
676730 96.135 -53.446 53.552
676731 96.178 -53.447 53.532
676732 96.211 -53.451 53.512
676733 96.244 -53.455 53.490
676734 96.276 -53.459 53.468
676735 96.309 -53.463 53.448
676736 96.341 -53.467 53.429
676737 96.374 -53.471 53.411
676738 96.406 -53.475 53.395
676739 96.439 -53.479 53.378
676740 96.472 -53.483 53.362
676741 96.504 -53.487 53.350
676742 96.530 -53.494 53.335
676743 96.552 -53.505 53.315
676744 96.573 -53.515 53.291
676745 96.594 -53.525 53.265
676746 96.615 -53.536 53.243
676747 96.637 -53.546 53.221
676748 96.674 -53.550 53.196
676749 96.716 -53.552 53.169
676750 96.758 -53.553 53.141
676751 96.800 -53.553 53.116
676752 96.841 -53.554 53.087
676753 96.883 -53.555 53.055
676754 96.925 -53.556 53.020
676755 96.965 -53.556 52.987
676756 96.998 -53.562 52.955
676757 97.023 -53.574 52.924
676758 97.047 -53.585 52.894
676759 97.072 -53.596 52.862
676760 97.097 -53.608 52.828
676761 97.121 -53.620 52.793
676762 97.146 -53.631 52.756
676763 97.171 -53.643 52.721
676764 97.195 -53.654 52.687
676765 97.220 -53.666 52.654
676766 97.245 -53.676 52.620
676767 97.271 -53.686 52.586
676768 97.297 -53.697 52.549
676769 97.322 -53.706 52.510
676770 97.348 -53.716 52.470
676771 97.373 -53.727 52.430
676772 97.397 -53.737 52.391
676773 97.419 -53.747 52.352
676774 97.441 -53.757 52.311
676775 97.469 -53.763 52.271
676776 97.497 -53.769 52.229
676777 97.525 -53.775 52.187
676778 97.553 -53.780 52.147
676779 97.581 -53.785 52.109
676780 97.608 -53.789 52.072
676781 97.635 -53.791 52.037
676782 97.662 -53.793 52.003
676783 97.688 -53.795 51.967
676784 97.715 -53.797 51.930
676785 97.741 -53.800 51.891
676786 97.768 -53.802 51.851
676787 97.795 -53.804 51.813
676788 97.821 -53.806 51.776
676789 97.839 -53.814 51.740
676790 97.855 -53.822 51.703
676791 97.870 -53.830 51.665
676792 97.886 -53.838 51.627
676793 97.901 -53.846 51.589
676794 97.917 -53.854 51.549
676795 97.934 -53.859 51.510
676796 97.952 -53.862 51.469
676797 97.969 -53.865 51.430
676798 97.987 -53.868 51.389
676799 98.005 -53.871 51.348
676800 98.023 -53.875 51.306
676801 98.042 -53.882 51.262
676802 98.075 -53.891 51.215
676803 98.107 -53.899 51.167
676804 98.140 -53.908 51.117
676805 98.172 -53.916 51.069
676806 98.205 -53.925 51.021
676807 98.221 -53.940 50.973
676808 98.247 -53.959 50.926
676809 98.272 -53.977 50.879
676810 98.298 -53.996 50.831
676811 98.323 -54.014 50.783
676812 98.349 -54.032 50.735
676813 98.374 -54.051 50.688
676814 98.400 -54.070 50.641
677301 93.095 -51.254 56.472
677302 93.127 -51.271 56.464
677303 93.157 -51.289 56.456
677304 93.189 -51.307 56.447
677305 93.219 -51.324 56.437
677306 93.251 -51.342 56.425
677307 93.281 -51.359 56.412
677308 93.312 -51.377 56.399
677309 93.343 -51.394 56.385
677310 93.374 -51.412 56.371
677311 93.405 -51.429 56.355
677312 93.436 -51.446 56.337
677313 93.467 -51.464 56.320
677314 93.498 -51.481 56.300
677315 93.529 -51.499 56.281
677316 93.560 -51.516 56.260
677317 93.591 -51.534 56.238
677318 93.622 -51.552 56.215
677319 93.651 -51.569 56.192
677320 93.679 -51.588 56.164
677321 93.705 -51.606 56.134
677322 93.732 -51.625 56.101
677323 93.759 -51.644 56.068
677324 93.785 -51.663 56.033
677325 93.812 -51.682 55.998
677326 93.839 -51.701 55.961
677327 93.866 -51.720 55.923
677328 93.893 -51.739 55.885
677329 93.919 -51.758 55.847
677330 93.946 -51.778 55.808
677331 93.973 -51.797 55.771
677332 93.999 -51.816 55.733
677377 95.202 -52.672 54.633
677378 95.229 -52.691 54.622
677379 95.256 -52.711 54.611
677380 95.282 -52.730 54.602
677381 95.309 -52.749 54.593
677382 95.336 -52.768 54.584
677383 95.362 -52.787 54.576
677384 95.389 -52.806 54.568
677385 95.416 -52.825 54.561
677386 95.443 -52.844 54.556
677387 95.469 -52.863 54.551
677388 95.496 -52.882 54.547
677389 95.526 -52.902 54.542
677390 95.561 -52.925 54.542
677391 95.596 -52.948 54.545
677392 95.631 -52.970 54.547
677393 95.665 -52.992 54.547
677394 95.680 -53.015 54.544
677395 95.696 -53.038 54.534
677396 95.711 -53.061 54.516
677397 95.727 -53.084 54.494
677398 95.743 -53.107 54.471
677399 95.759 -53.130 54.440
677400 95.774 -53.153 54.408
677401 95.790 -53.176 54.378
677402 95.808 -53.198 54.348
677403 95.829 -53.214 54.314
677404 95.850 -53.231 54.279
677405 95.872 -53.248 54.249
677406 95.893 -53.264 54.222
677407 95.914 -53.281 54.197
677408 95.936 -53.298 54.173
677409 95.955 -53.313 54.146
677410 95.960 -53.319 54.115
677411 95.966 -53.324 54.080
677412 95.972 -53.330 54.046
677413 95.977 -53.336 54.016
677414 95.983 -53.342 53.989
677415 95.989 -53.348 53.968
677416 95.994 -53.353 53.951
677417 96.000 -53.359 53.940
677418 96.006 -53.365 53.923
677419 96.012 -53.371 53.901
677420 96.020 -53.378 53.877
677421 96.029 -53.386 53.846
677422 96.039 -53.389 53.809
677423 96.049 -53.393 53.768
677424 96.059 -53.396 53.727
677425 96.070 -53.400 53.692
677426 96.080 -53.404 53.660
677427 96.090 -53.408 53.630
677428 96.100 -53.411 53.602
677429 96.111 -53.415 53.577
677430 96.121 -53.418 53.554
677431 96.136 -53.422 53.531
677432 96.164 -53.424 53.510
677433 96.192 -53.426 53.490
677434 96.225 -53.430 53.468
677435 96.257 -53.433 53.447
677436 96.290 -53.437 53.427
677437 96.322 -53.441 53.409
677438 96.355 -53.445 53.390
677439 96.387 -53.449 53.375
677440 96.420 -53.452 53.357
677441 96.452 -53.456 53.341
677442 96.485 -53.459 53.327
677443 96.517 -53.463 53.311
677444 96.550 -53.467 53.292
677445 96.574 -53.476 53.270
677446 96.595 -53.487 53.247
677447 96.616 -53.497 53.225
677448 96.638 -53.507 53.205
677449 96.671 -53.512 53.183
677450 96.714 -53.514 53.156
677451 96.756 -53.515 53.127
677452 96.798 -53.517 53.100
677453 96.841 -53.518 53.071
677454 96.883 -53.520 53.040
677455 96.926 -53.521 53.007
677456 96.965 -53.526 52.976
677457 96.999 -53.534 52.944
677458 97.030 -53.544 52.912
677459 97.055 -53.556 52.881
677460 97.080 -53.567 52.849
677461 97.104 -53.578 52.816
677462 97.129 -53.590 52.782
677463 97.154 -53.602 52.748
677464 97.178 -53.613 52.714
677465 97.203 -53.625 52.681
677466 97.228 -53.636 52.649
677467 97.252 -53.647 52.616
677468 97.278 -53.657 52.584
677469 97.304 -53.668 52.548
677470 97.329 -53.678 52.511
677471 97.355 -53.687 52.471
677472 97.379 -53.698 52.433
677473 97.401 -53.709 52.396
677474 97.423 -53.719 52.358
677475 97.445 -53.728 52.318
677476 97.473 -53.734 52.277
677477 97.501 -53.740 52.235
677478 97.529 -53.746 52.193
677479 97.558 -53.751 52.154
677480 97.585 -53.757 52.116
677481 97.613 -53.762 52.079
677482 97.642 -53.768 52.045
677483 97.668 -53.770 52.010
677484 97.695 -53.772 51.975
677485 97.721 -53.775 51.938
677486 97.748 -53.777 51.900
677487 97.775 -53.779 51.860
677488 97.801 -53.781 51.822
677489 97.828 -53.784 51.788
677490 97.847 -53.790 51.752
677491 97.863 -53.798 51.717
677492 97.878 -53.806 51.680
677493 97.894 -53.814 51.643
677494 97.910 -53.822 51.605
677495 97.926 -53.830 51.567
677496 97.941 -53.838 51.529
677497 97.959 -53.841 51.490
677498 97.977 -53.844 51.450
677499 97.994 -53.847 51.410
677500 98.012 -53.854 51.370
677501 98.030 -53.860 51.328
677502 98.048 -53.867 51.285
677503 98.076 -53.875 51.240
677504 98.108 -53.884 51.192
677505 98.141 -53.893 51.144
677506 98.173 -53.901 51.095
677507 98.206 -53.910 51.048
677508 98.225 -53.924 51.000
677509 98.248 -53.941 50.953
677510 98.274 -53.960 50.907
677511 98.299 -53.978 50.860
677512 98.324 -53.997 50.813
677513 98.350 -54.015 50.766
677514 98.375 -54.033 50.720
677515 98.401 -54.052 50.674
677516 98.426 -54.070 50.630
678003 93.144 -51.247 56.457
678004 93.174 -51.265 56.449
678005 93.206 -51.283 56.441
678006 93.236 -51.300 56.431
678007 93.268 -51.318 56.417
678008 93.298 -51.335 56.405
678009 93.330 -51.352 56.393
678010 93.360 -51.370 56.379
678011 93.392 -51.387 56.364
678012 93.422 -51.405 56.349
678013 93.453 -51.422 56.331
678014 93.484 -51.440 56.313
678015 93.515 -51.457 56.294
678016 93.546 -51.475 56.275
678017 93.577 -51.492 56.253
678018 93.608 -51.509 56.231
678019 93.639 -51.528 56.207
678020 93.670 -51.545 56.182
678021 93.698 -51.564 56.153
678022 93.725 -51.584 56.123
678023 93.752 -51.603 56.091
678024 93.779 -51.623 56.057
678025 93.805 -51.643 56.023
678026 93.833 -51.663 55.988
678027 93.859 -51.682 55.952
678028 93.887 -51.702 55.914
678029 93.913 -51.721 55.876
678030 93.940 -51.740 55.839
678031 93.967 -51.759 55.802
678032 93.993 -51.778 55.765
678033 94.020 -51.797 55.727
678079 95.250 -52.673 54.617
678080 95.276 -52.692 54.606
678081 95.303 -52.711 54.596
678082 95.330 -52.730 54.587
678083 95.356 -52.749 54.577
678084 95.383 -52.768 54.569
678085 95.410 -52.787 54.560
678086 95.437 -52.807 54.552
678087 95.463 -52.825 54.545
678088 95.490 -52.844 54.539
678089 95.517 -52.863 54.532
678090 95.544 -52.882 54.528
678091 95.578 -52.904 54.528
678092 95.612 -52.927 54.531
678093 95.647 -52.949 54.531
678094 95.681 -52.972 54.531
678095 95.696 -52.995 54.526
678096 95.712 -53.017 54.515
678097 95.728 -53.041 54.497
678098 95.744 -53.064 54.475
678099 95.759 -53.087 54.453
678100 95.775 -53.110 54.423
678101 95.791 -53.132 54.389
678102 95.807 -53.156 54.357
678103 95.822 -53.179 54.327
678104 95.841 -53.198 54.293
678105 95.863 -53.214 54.258
678106 95.884 -53.232 54.228
678107 95.905 -53.248 54.200
678108 95.926 -53.265 54.173
678109 95.948 -53.282 54.146
678110 95.958 -53.291 54.118
678111 95.964 -53.296 54.088
678112 95.970 -53.302 54.055
678113 95.975 -53.308 54.023
678114 95.981 -53.314 53.995
678115 95.986 -53.320 53.970
678116 95.992 -53.325 53.948
678117 95.998 -53.331 53.930
678118 96.004 -53.337 53.917
678119 96.013 -53.346 53.899
678120 96.022 -53.353 53.878
678121 96.031 -53.361 53.854
678122 96.040 -53.369 53.823
678123 96.050 -53.373 53.785
678124 96.060 -53.377 53.744
678125 96.071 -53.380 53.705
678126 96.081 -53.383 53.670
678127 96.091 -53.387 53.638
678128 96.101 -53.391 53.610
678129 96.111 -53.395 53.582
678130 96.121 -53.399 53.557
678131 96.132 -53.402 53.533
678132 96.148 -53.405 53.510
678133 96.175 -53.407 53.489
678134 96.203 -53.409 53.469
678135 96.231 -53.411 53.447
678136 96.258 -53.414 53.426
678137 96.286 -53.415 53.407
678138 96.313 -53.418 53.388
678139 96.341 -53.420 53.370
678140 96.369 -53.421 53.354
678141 96.400 -53.425 53.336
678142 96.433 -53.429 53.319
678143 96.465 -53.433 53.304
678144 96.498 -53.437 53.287
678145 96.531 -53.440 53.269
678146 96.563 -53.444 53.250
678147 96.596 -53.448 53.228
678148 96.618 -53.458 53.207
678149 96.639 -53.468 53.188
678150 96.668 -53.475 53.167
678151 96.711 -53.477 53.141
678152 96.753 -53.478 53.112
678153 96.796 -53.480 53.084
678154 96.838 -53.481 53.056
678155 96.872 -53.490 53.025
678156 96.906 -53.498 52.994
678157 96.941 -53.506 52.962
678158 96.975 -53.515 52.931
678159 97.009 -53.524 52.899
678160 97.043 -53.532 52.867
678161 97.077 -53.540 52.836
678162 97.111 -53.549 52.804
678163 97.137 -53.560 52.773
678164 97.161 -53.571 52.740
678165 97.186 -53.584 52.707
678166 97.211 -53.595 52.675
678167 97.235 -53.606 52.643
678168 97.260 -53.618 52.613
678169 97.285 -53.628 52.580
678170 97.310 -53.639 52.547
678171 97.336 -53.649 52.510
678172 97.362 -53.659 52.473
678173 97.383 -53.669 52.437
678174 97.405 -53.680 52.402
678175 97.427 -53.690 52.366
678176 97.449 -53.700 52.324
678177 97.477 -53.706 52.283
678178 97.505 -53.711 52.241
678179 97.533 -53.717 52.199
678180 97.562 -53.722 52.160
678181 97.590 -53.728 52.123
678182 97.618 -53.734 52.086
678183 97.646 -53.740 52.051
678184 97.674 -53.745 52.016
678185 97.701 -53.749 51.981
678186 97.728 -53.751 51.945
678187 97.754 -53.753 51.907
678188 97.781 -53.756 51.868
678189 97.808 -53.758 51.832
678190 97.835 -53.760 51.798
678191 97.855 -53.766 51.764
678192 97.871 -53.774 51.730
678193 97.887 -53.781 51.694
678194 97.902 -53.790 51.658
678195 97.918 -53.798 51.621
678196 97.934 -53.806 51.584
678197 97.950 -53.814 51.547
678198 97.965 -53.822 51.510
678199 97.984 -53.827 51.470
678200 98.002 -53.833 51.430
678201 98.019 -53.840 51.390
678202 98.037 -53.847 51.349
678203 98.056 -53.853 51.307
678204 98.077 -53.860 51.262
678205 98.109 -53.869 51.216
678206 98.142 -53.878 51.169
678207 98.174 -53.887 51.122
678208 98.207 -53.895 51.074
678209 98.230 -53.907 51.027
678210 98.250 -53.923 50.981
678211 98.276 -53.942 50.935
678212 98.301 -53.960 50.890
678213 98.326 -53.979 50.843
678214 98.352 -53.997 50.797
678215 98.377 -54.016 50.751
678216 98.403 -54.035 50.706
678217 98.428 -54.053 50.662
678705 93.192 -51.240 56.442
678706 93.222 -51.258 56.433
678707 93.254 -51.276 56.423
678708 93.284 -51.293 56.411
678709 93.315 -51.311 56.398
678710 93.346 -51.328 56.385
678711 93.377 -51.346 56.372
678712 93.408 -51.363 56.357
678713 93.439 -51.381 56.342
678714 93.470 -51.398 56.325
678715 93.501 -51.415 56.307
678716 93.532 -51.433 56.288
678717 93.563 -51.450 56.268
678718 93.594 -51.468 56.247
678719 93.625 -51.485 56.224
678720 93.656 -51.503 56.199
678721 93.687 -51.521 56.172
678722 93.717 -51.538 56.143
678723 93.744 -51.559 56.112
678724 93.771 -51.578 56.080
678725 93.797 -51.598 56.047
678726 93.825 -51.618 56.013
678727 93.851 -51.638 55.978
678728 93.878 -51.657 55.942
678729 93.905 -51.677 55.904
678730 93.932 -51.697 55.867
678731 93.959 -51.716 55.830
678732 93.986 -51.736 55.794
678733 94.013 -51.756 55.757
678780 95.270 -52.654 54.613
678781 95.297 -52.673 54.602
678782 95.324 -52.692 54.591
678783 95.350 -52.711 54.581
678784 95.377 -52.731 54.570
678785 95.404 -52.750 54.562
678786 95.430 -52.769 54.552
678787 95.457 -52.788 54.543
678788 95.484 -52.807 54.535
678789 95.511 -52.825 54.527
678790 95.538 -52.844 54.520
678791 95.564 -52.863 54.514
678792 95.594 -52.884 54.512
678793 95.628 -52.906 54.514
678794 95.663 -52.929 54.515
678795 95.697 -52.951 54.513
678796 95.713 -52.974 54.507
678797 95.729 -52.997 54.493
678798 95.744 -53.020 54.474
678799 95.760 -53.043 54.454
678800 95.776 -53.066 54.432
678801 95.792 -53.089 54.403
678802 95.807 -53.112 54.370
678803 95.823 -53.136 54.336
678804 95.838 -53.158 54.306
678805 95.854 -53.181 54.271
678806 95.875 -53.198 54.237
678807 95.896 -53.215 54.207
678808 95.917 -53.232 54.178
678809 95.939 -53.248 54.149
678810 95.956 -53.263 54.121
678811 95.962 -53.268 54.093
678812 95.967 -53.274 54.062
678813 95.973 -53.280 54.031
678814 95.979 -53.286 54.002
678815 95.984 -53.292 53.975
678816 95.990 -53.297 53.949
678817 95.997 -53.305 53.926
678818 96.007 -53.312 53.907
678819 96.016 -53.321 53.890
678820 96.024 -53.328 53.870
678821 96.033 -53.336 53.847
678822 96.042 -53.344 53.821
678823 96.051 -53.352 53.790
678824 96.061 -53.356 53.754
678825 96.071 -53.360 53.715
678826 96.082 -53.364 53.679
678827 96.092 -53.367 53.647
678828 96.102 -53.371 53.616
678829 96.112 -53.374 53.588
678830 96.122 -53.378 53.562
678831 96.132 -53.382 53.536
678832 96.143 -53.386 53.512
678833 96.159 -53.389 53.490
678834 96.187 -53.390 53.469
678835 96.215 -53.393 53.448
678836 96.242 -53.395 53.427
678837 96.269 -53.397 53.406
678838 96.297 -53.399 53.385
678839 96.325 -53.401 53.366
678840 96.353 -53.403 53.349
678841 96.380 -53.405 53.332
678842 96.407 -53.407 53.315
678843 96.435 -53.409 53.297
678844 96.463 -53.411 53.281
678845 96.490 -53.413 53.264
678846 96.518 -53.415 53.247
678847 96.546 -53.418 53.229
678848 96.577 -53.421 53.209
678849 96.609 -53.424 53.189
678850 96.640 -53.429 53.171
678851 96.666 -53.437 53.149
678852 96.708 -53.439 53.125
678853 96.745 -53.445 53.098
678854 96.780 -53.453 53.069
678855 96.814 -53.462 53.041
678856 96.848 -53.470 53.012
678857 96.882 -53.479 52.981
678858 96.916 -53.487 52.950
678859 96.950 -53.496 52.920
678860 96.984 -53.504 52.889
678861 97.019 -53.512 52.857
678862 97.053 -53.521 52.826
678863 97.087 -53.529 52.795
678864 97.121 -53.538 52.765
678865 97.156 -53.546 52.733
678866 97.190 -53.555 52.701
678867 97.218 -53.565 52.669
678868 97.243 -53.577 52.639
678869 97.267 -53.588 52.608
678870 97.292 -53.599 52.577
678871 97.317 -53.610 52.544
678872 97.343 -53.620 52.508
678873 97.366 -53.631 52.473
678874 97.387 -53.641 52.439
678875 97.409 -53.651 52.406
678876 97.431 -53.662 52.371
678877 97.453 -53.672 52.330
678878 97.481 -53.677 52.288
678879 97.510 -53.683 52.246
678880 97.538 -53.688 52.205
678881 97.566 -53.694 52.165
678882 97.594 -53.700 52.129
678883 97.622 -53.706 52.093
678884 97.650 -53.711 52.056
678885 97.678 -53.717 52.019
678886 97.706 -53.722 51.984
678887 97.734 -53.728 51.949
678888 97.761 -53.731 51.912
678889 97.788 -53.733 51.875
678890 97.814 -53.735 51.840
678891 97.841 -53.737 51.806
678892 97.864 -53.741 51.774
678893 97.879 -53.750 51.741
678894 97.895 -53.758 51.707
678895 97.911 -53.766 51.673
678896 97.927 -53.774 51.638
678897 97.942 -53.781 51.602
678898 97.956 -53.791 51.565
678899 97.969 -53.801 51.529
678900 97.991 -53.813 51.490
678901 98.009 -53.819 51.450
678902 98.027 -53.826 51.410
678903 98.045 -53.832 51.369
678904 98.063 -53.839 51.327
678905 98.081 -53.846 51.283
678906 98.110 -53.854 51.239
678907 98.143 -53.863 51.194
678908 98.175 -53.872 51.148
678909 98.208 -53.880 51.101
678910 98.235 -53.891 51.054
678911 98.252 -53.906 51.009
678912 98.277 -53.925 50.963
678913 98.303 -53.943 50.919
678914 98.328 -53.961 50.873
678915 98.354 -53.980 50.828
678916 98.379 -53.998 50.783
678917 98.405 -54.016 50.739
678918 98.430 -54.035 50.695
679408 93.271 -51.252 56.416
679409 93.301 -51.269 56.403
679410 93.333 -51.287 56.391
679411 93.363 -51.304 56.378
679412 93.395 -51.321 56.365
679413 93.425 -51.339 56.350
679414 93.457 -51.356 56.335
679415 93.487 -51.374 56.318
679416 93.518 -51.391 56.300
679417 93.549 -51.409 56.281
679418 93.580 -51.426 56.260
679419 93.611 -51.443 56.238
679420 93.642 -51.461 56.215
679421 93.673 -51.478 56.189
679422 93.704 -51.497 56.163
679423 93.735 -51.514 56.133
679424 93.763 -51.533 56.102
679425 93.789 -51.553 56.070
679426 93.817 -51.572 56.036
679427 93.843 -51.592 56.002
679428 93.870 -51.612 55.968
679429 93.897 -51.631 55.932
679430 93.924 -51.651 55.894
679431 93.951 -51.672 55.857
679432 93.978 -51.691 55.821
679433 94.005 -51.711 55.785
679434 94.032 -51.731 55.749
679481 95.291 -52.635 54.609
679482 95.318 -52.654 54.597
679483 95.344 -52.674 54.586
679484 95.371 -52.693 54.575
679485 95.398 -52.712 54.564
679486 95.424 -52.731 54.554
679487 95.451 -52.750 54.544
679488 95.478 -52.769 54.534
679489 95.505 -52.788 54.525
679490 95.531 -52.807 54.517
679491 95.558 -52.826 54.509
679492 95.585 -52.845 54.501
679493 95.612 -52.864 54.497
679494 95.644 -52.885 54.497
679495 95.680 -52.908 54.497
679496 95.713 -52.930 54.495
679497 95.729 -52.954 54.487
679498 95.745 -52.976 54.472
679499 95.761 -53.000 54.452
679500 95.776 -53.023 54.434
679501 95.792 -53.045 54.411
679502 95.808 -53.069 54.383
679503 95.823 -53.092 54.351
679504 95.839 -53.115 54.318
679505 95.855 -53.138 54.287
679506 95.871 -53.161 54.251
679507 95.887 -53.183 54.217
679508 95.909 -53.199 54.186
679509 95.930 -53.216 54.157
679510 95.951 -53.233 54.127
679511 95.960 -53.240 54.098
679512 95.965 -53.246 54.069
679513 95.971 -53.252 54.039
679514 95.977 -53.258 54.010
679515 95.982 -53.264 53.982
679516 95.991 -53.271 53.955
679517 96.000 -53.280 53.928
679518 96.009 -53.287 53.904
679519 96.018 -53.295 53.883
679520 96.027 -53.303 53.861
679521 96.036 -53.311 53.839
679522 96.044 -53.319 53.814
679523 96.053 -53.327 53.787
679524 96.062 -53.335 53.756
679525 96.072 -53.339 53.722
679526 96.082 -53.343 53.687
679527 96.093 -53.347 53.654
679528 96.103 -53.351 53.624
679529 96.113 -53.354 53.595
679530 96.123 -53.358 53.568
679531 96.133 -53.361 53.541
679532 96.144 -53.365 53.516
679533 96.154 -53.369 53.492
679534 96.171 -53.372 53.470
679535 96.198 -53.374 53.448
679536 96.226 -53.376 53.428
679537 96.253 -53.378 53.406
679538 96.281 -53.380 53.385
679539 96.309 -53.382 53.366
679540 96.336 -53.384 53.346
679541 96.364 -53.386 53.328
679542 96.391 -53.389 53.313
679543 96.419 -53.390 53.294
679544 96.447 -53.393 53.277
679545 96.474 -53.395 53.260
679546 96.502 -53.397 53.244
679547 96.530 -53.399 53.227
679548 96.557 -53.401 53.210
679549 96.585 -53.403 53.191
679550 96.612 -53.405 53.171
679551 96.640 -53.407 53.152
679552 96.666 -53.410 53.132
679553 96.696 -53.418 53.109
679554 96.729 -53.426 53.083
679555 96.762 -53.434 53.055
679556 96.796 -53.443 53.027
679557 96.829 -53.451 52.999
679558 96.862 -53.459 52.969
679559 96.895 -53.468 52.940
679560 96.928 -53.477 52.910
679561 96.961 -53.485 52.879
679562 96.995 -53.493 52.849
679563 97.029 -53.502 52.818
679564 97.063 -53.510 52.788
679565 97.097 -53.518 52.758
679566 97.131 -53.527 52.727
679567 97.165 -53.535 52.696
679568 97.199 -53.543 52.664
679569 97.234 -53.552 52.634
679570 97.268 -53.561 52.603
679571 97.300 -53.570 52.573
679572 97.324 -53.581 52.540
679573 97.348 -53.591 52.506
679574 97.370 -53.602 52.472
679575 97.391 -53.612 52.439
679576 97.413 -53.623 52.407
679577 97.435 -53.633 52.373
679578 97.458 -53.643 52.334
679579 97.486 -53.648 52.291
679580 97.514 -53.654 52.249
679581 97.542 -53.659 52.208
679582 97.570 -53.665 52.170
679583 97.598 -53.671 52.133
679584 97.626 -53.677 52.096
679585 97.654 -53.682 52.059
679586 97.682 -53.688 52.023
679587 97.711 -53.693 51.987
679588 97.739 -53.700 51.952
679589 97.767 -53.705 51.916
679590 97.794 -53.709 51.881
679591 97.821 -53.712 51.847
679592 97.848 -53.714 51.814
679593 97.872 -53.718 51.783
679594 97.887 -53.725 51.752
679595 97.903 -53.734 51.719
679596 97.919 -53.742 51.686
679597 97.934 -53.750 51.651
679598 97.946 -53.760 51.616
679599 97.959 -53.771 51.581
679600 97.971 -53.781 51.546
679601 97.996 -53.796 51.508
679602 98.016 -53.805 51.469
679603 98.034 -53.812 51.430
679604 98.052 -53.819 51.388
679605 98.070 -53.825 51.347
679606 98.088 -53.831 51.305
679607 98.111 -53.839 51.263
679608 98.144 -53.848 51.219
679609 98.176 -53.856 51.174
679610 98.209 -53.866 51.127
679611 98.240 -53.875 51.080
679612 98.255 -53.889 51.035
679613 98.279 -53.907 50.991
679614 98.304 -53.925 50.947
679615 98.330 -53.944 50.902
679616 98.355 -53.962 50.858
679617 98.381 -53.981 50.814
679618 98.406 -53.999 50.770
679619 98.432 -54.018 50.728
679620 98.457 -54.036 50.686
680110 93.319 -51.245 56.395
680111 93.349 -51.262 56.382
680112 93.380 -51.280 56.370
680113 93.411 -51.297 56.358
680114 93.442 -51.315 56.343
680115 93.473 -51.332 56.327
680116 93.504 -51.349 56.310
680117 93.535 -51.367 56.292
680118 93.566 -51.384 56.273
680119 93.597 -51.402 56.252
680120 93.628 -51.419 56.229
680121 93.659 -51.437 56.205
680122 93.690 -51.454 56.180
680123 93.721 -51.472 56.152
680124 93.752 -51.490 56.122
680125 93.781 -51.508 56.091
680126 93.808 -51.528 56.058
680127 93.835 -51.547 56.025
680128 93.862 -51.567 55.991
680129 93.889 -51.587 55.957
680130 93.916 -51.606 55.921
680131 93.943 -51.626 55.884
680132 93.970 -51.646 55.847
680133 93.997 -51.666 55.811
680134 94.024 -51.685 55.776
680135 94.051 -51.705 55.741
680183 95.338 -52.636 54.592
680184 95.365 -52.655 54.580
680185 95.392 -52.674 54.569
680186 95.418 -52.693 54.557
680187 95.445 -52.712 54.547
680188 95.472 -52.731 54.536
680189 95.498 -52.750 54.526
680190 95.525 -52.769 54.516
680191 95.552 -52.788 54.507
680192 95.579 -52.807 54.497
680193 95.606 -52.826 54.489
680194 95.632 -52.845 54.483
680195 95.661 -52.865 54.480
680196 95.696 -52.888 54.479
680197 95.730 -52.910 54.476
680198 95.746 -52.933 54.467
680199 95.761 -52.956 54.451
680200 95.777 -52.979 54.431
680201 95.793 -53.002 54.412
680202 95.809 -53.025 54.391
680203 95.824 -53.048 54.365
680204 95.840 -53.071 54.334
680205 95.855 -53.094 54.299
680206 95.871 -53.117 54.267
680207 95.887 -53.140 54.231
680208 95.903 -53.164 54.197
680209 95.921 -53.183 54.166
680210 95.943 -53.200 54.136
680211 95.958 -53.212 54.106
680212 95.963 -53.218 54.077
680213 95.969 -53.224 54.049
680214 95.976 -53.230 54.019
680215 95.985 -53.239 53.990
680216 95.993 -53.246 53.962
680217 96.003 -53.255 53.934
680218 96.011 -53.262 53.907
680219 96.020 -53.270 53.882
680220 96.029 -53.278 53.858
680221 96.038 -53.286 53.832
680222 96.047 -53.294 53.806
680223 96.055 -53.302 53.779
680224 96.064 -53.310 53.751
680225 96.074 -53.317 53.721
680226 96.083 -53.323 53.690
680227 96.093 -53.327 53.659
680228 96.103 -53.330 53.630
680229 96.114 -53.334 53.600
680230 96.124 -53.338 53.573
680231 96.134 -53.341 53.547
680232 96.145 -53.345 53.521
680233 96.155 -53.349 53.495
680234 96.165 -53.352 53.472
680235 96.182 -53.355 53.450
680236 96.209 -53.358 53.428
680237 96.237 -53.359 53.407
680238 96.265 -53.361 53.385
680239 96.293 -53.364 53.365
680240 96.320 -53.366 53.345
680241 96.348 -53.368 53.326
680242 96.375 -53.370 53.308
680243 96.403 -53.372 53.292
680244 96.431 -53.374 53.274
680245 96.458 -53.376 53.256
680246 96.486 -53.378 53.240
680247 96.514 -53.380 53.224
680248 96.541 -53.382 53.207
680249 96.569 -53.384 53.190
680250 96.596 -53.386 53.172
680251 96.624 -53.389 53.153
680252 96.652 -53.390 53.134
680253 96.676 -53.395 53.114
680254 96.697 -53.401 53.091
680255 96.728 -53.408 53.067
680256 96.761 -53.417 53.041
680257 96.795 -53.426 53.014
680258 96.828 -53.434 52.986
680259 96.861 -53.443 52.958
680260 96.894 -53.451 52.929
680261 96.927 -53.459 52.901
680262 96.960 -53.468 52.871
680263 96.994 -53.476 52.841
680264 97.027 -53.484 52.811
680265 97.060 -53.493 52.781
680266 97.093 -53.501 52.752
680267 97.126 -53.509 52.722
680268 97.160 -53.518 52.691
680269 97.193 -53.526 52.660
680270 97.226 -53.534 52.629
680271 97.259 -53.543 52.598
680272 97.293 -53.551 52.568
680273 97.326 -53.559 52.536
680274 97.353 -53.568 52.504
680275 97.374 -53.575 52.471
680276 97.395 -53.584 52.438
680277 97.417 -53.594 52.406
680278 97.439 -53.605 52.374
680279 97.462 -53.614 52.335
680280 97.490 -53.619 52.293
680281 97.518 -53.625 52.252
680282 97.546 -53.631 52.211
680283 97.574 -53.637 52.173
680284 97.602 -53.642 52.136
680285 97.631 -53.648 52.098
680286 97.659 -53.653 52.062
680287 97.686 -53.659 52.026
680288 97.715 -53.665 51.990
680289 97.743 -53.671 51.955
680290 97.771 -53.676 51.920
680291 97.799 -53.682 51.886
680292 97.827 -53.687 51.853
680293 97.854 -53.691 51.822
680294 97.880 -53.693 51.792
680295 97.896 -53.702 51.761
680296 97.912 -53.709 51.730
680297 97.924 -53.720 51.697
680298 97.936 -53.730 51.664
680299 97.949 -53.740 51.629
680300 97.961 -53.751 51.596
680301 97.973 -53.761 51.561
680302 97.998 -53.776 51.525
680303 98.023 -53.791 51.486
680304 98.041 -53.797 51.447
680305 98.060 -53.804 51.407
680306 98.078 -53.811 51.367
680307 98.096 -53.818 51.327
680308 98.113 -53.824 51.286
680309 98.145 -53.833 51.243
680310 98.177 -53.841 51.198
680311 98.210 -53.850 51.152
680312 98.242 -53.859 51.105
680313 98.260 -53.872 51.061
680314 98.281 -53.889 51.018
680315 98.306 -53.908 50.974
680316 98.332 -53.926 50.931
680317 98.357 -53.944 50.888
680318 98.382 -53.963 50.846
680319 98.408 -53.982 50.802
680320 98.433 -54.000 50.761
680321 98.458 -54.019 50.720
680322 98.484 -54.037 50.679
680323 98.509 -54.055 50.639
680812 93.366 -51.238 56.374
680813 93.398 -51.255 56.362
680814 93.428 -51.273 56.349
680815 93.460 -51.290 56.335
680816 93.490 -51.308 56.319
680817 93.521 -51.325 56.301
680818 93.552 -51.343 56.283
680819 93.583 -51.360 56.264
680820 93.614 -51.378 56.243
680821 93.645 -51.395 56.219
680822 93.676 -51.412 56.195
680823 93.707 -51.430 56.170
680824 93.738 -51.448 56.141
680825 93.769 -51.465 56.111
680826 93.800 -51.483 56.080
680827 93.827 -51.503 56.047
680828 93.854 -51.522 56.014
680829 93.881 -51.542 55.980
680830 93.908 -51.562 55.946
680831 93.935 -51.581 55.910
680832 93.962 -51.601 55.874
680833 93.989 -51.621 55.837
680834 94.015 -51.641 55.801
680835 94.043 -51.660 55.766
680885 95.386 -52.636 54.574
680886 95.412 -52.655 54.562
680887 95.439 -52.674 54.550
680888 95.466 -52.694 54.539
680889 95.492 -52.713 54.528
680890 95.519 -52.731 54.517
680891 95.546 -52.750 54.506
680892 95.572 -52.769 54.496
680893 95.600 -52.788 54.486
680894 95.626 -52.807 54.477
680895 95.653 -52.826 54.469
680896 95.680 -52.846 54.464
680897 95.712 -52.867 54.461
680898 95.746 -52.889 54.456
680899 95.762 -52.913 54.447
680900 95.778 -52.935 54.430
680901 95.794 -52.958 54.410
680902 95.809 -52.982 54.390
680903 95.825 -53.004 54.371
680904 95.840 -53.028 54.348
680905 95.856 -53.051 54.317
680906 95.872 -53.073 54.280
680907 95.888 -53.097 54.246
680908 95.903 -53.120 54.211
680909 95.919 -53.142 54.178
680910 95.935 -53.166 54.146
680911 95.955 -53.184 54.116
680912 95.961 -53.190 54.086
680913 95.969 -53.198 54.057
680914 95.978 -53.205 54.029
680915 95.987 -53.214 54.000
680916 95.996 -53.221 53.971
680917 96.005 -53.230 53.942
680918 96.014 -53.237 53.914
680919 96.023 -53.245 53.887
680920 96.031 -53.253 53.861
680921 96.040 -53.261 53.835
680922 96.049 -53.269 53.808
680923 96.058 -53.277 53.780
680924 96.067 -53.285 53.752
680925 96.076 -53.292 53.724
680926 96.085 -53.300 53.695
680927 96.094 -53.306 53.666
680928 96.104 -53.310 53.636
680929 96.114 -53.314 53.607
680930 96.125 -53.317 53.578
680931 96.135 -53.321 53.552
680932 96.145 -53.324 53.526
680933 96.156 -53.328 53.500
680934 96.166 -53.332 53.475
680935 96.176 -53.336 53.450
680936 96.193 -53.339 53.429
680937 96.221 -53.341 53.408
680938 96.249 -53.342 53.387
680939 96.276 -53.345 53.365
680940 96.304 -53.347 53.344
680941 96.332 -53.349 53.325
680942 96.359 -53.351 53.305
680943 96.387 -53.353 53.287
680944 96.414 -53.355 53.270
680945 96.442 -53.358 53.253
680946 96.470 -53.359 53.236
680947 96.498 -53.361 53.219
680948 96.525 -53.364 53.203
680949 96.552 -53.365 53.187
680950 96.580 -53.368 53.171
680951 96.608 -53.370 53.154
680952 96.636 -53.372 53.134
680953 96.663 -53.374 53.116
680954 96.685 -53.380 53.096
680955 96.707 -53.386 53.074
680956 96.729 -53.392 53.051
680957 96.760 -53.400 53.027
680958 96.794 -53.408 53.001
680959 96.827 -53.417 52.973
680960 96.860 -53.425 52.946
680961 96.893 -53.433 52.918
680962 96.926 -53.442 52.890
680963 96.959 -53.450 52.861
680964 96.993 -53.458 52.832
680965 97.026 -53.467 52.803
680966 97.059 -53.475 52.774
680967 97.092 -53.483 52.745
680968 97.125 -53.492 52.716
680969 97.159 -53.500 52.686
680970 97.192 -53.509 52.655
680971 97.225 -53.517 52.625
680972 97.258 -53.525 52.594
680973 97.292 -53.534 52.564
680974 97.325 -53.542 52.532
680975 97.361 -53.549 52.500
680976 97.380 -53.558 52.467
680977 97.401 -53.565 52.434
680978 97.422 -53.573 52.402
680979 97.444 -53.581 52.369
680980 97.465 -53.589 52.332
680981 97.495 -53.596 52.292
680982 97.523 -53.600 52.252
680983 97.550 -53.606 52.212
680984 97.579 -53.611 52.174
680985 97.607 -53.616 52.138
680986 97.635 -53.621 52.099
680987 97.663 -53.626 52.063
680988 97.691 -53.631 52.027
680989 97.719 -53.637 51.992
680990 97.747 -53.642 51.958
680991 97.775 -53.648 51.924
680992 97.803 -53.653 51.892
680993 97.831 -53.659 51.861
680994 97.860 -53.665 51.831
680995 97.887 -53.670 51.801
680996 97.902 -53.679 51.771
680997 97.914 -53.690 51.740
680998 97.927 -53.700 51.708
680999 97.939 -53.710 51.676
681000 97.951 -53.721 51.642
681001 97.963 -53.731 51.610
681002 97.975 -53.741 51.576
681003 98.000 -53.756 51.540
681004 98.025 -53.771 51.502
681005 98.049 -53.784 51.464
681006 98.067 -53.791 51.424
681007 98.085 -53.797 51.386
681008 98.103 -53.803 51.348
681009 98.121 -53.810 51.307
681010 98.146 -53.818 51.265
681011 98.178 -53.826 51.221
681012 98.211 -53.835 51.175
681013 98.243 -53.844 51.130
681014 98.265 -53.856 51.086
681015 98.282 -53.872 51.044
681016 98.308 -53.890 51.001
681017 98.333 -53.909 50.959
681018 98.359 -53.927 50.918
681019 98.384 -53.945 50.876
681020 98.410 -53.964 50.834
681021 98.435 -53.982 50.793
681022 98.460 -54.001 50.752
681023 98.486 -54.019 50.711
681024 98.511 -54.038 50.672
681025 98.536 -54.057 50.634
681514 93.415 -51.231 56.354
681515 93.445 -51.249 56.341
681516 93.476 -51.266 56.326
681517 93.507 -51.284 56.310
681518 93.538 -51.301 56.293
681519 93.569 -51.318 56.274
681520 93.600 -51.336 56.254
681521 93.631 -51.353 56.233
681522 93.662 -51.371 56.209
681523 93.693 -51.388 56.184
681524 93.724 -51.406 56.158
681525 93.755 -51.424 56.130
681526 93.786 -51.441 56.100
681527 93.817 -51.459 56.068
681528 93.846 -51.477 56.035
681529 93.873 -51.497 56.002
681530 93.900 -51.516 55.969
681531 93.927 -51.536 55.934
681532 93.954 -51.556 55.899
681533 93.981 -51.575 55.863
681534 94.007 -51.595 55.827
681535 94.035 -51.615 55.791
681587 95.433 -52.637 54.555
681588 95.460 -52.656 54.543
681589 95.486 -52.675 54.531
681590 95.513 -52.694 54.520
681591 95.540 -52.713 54.508
681592 95.566 -52.731 54.497
681593 95.593 -52.750 54.485
681594 95.620 -52.770 54.475
681595 95.647 -52.789 54.465
681596 95.674 -52.808 54.455
681597 95.700 -52.827 54.447
681598 95.728 -52.847 54.442
681599 95.763 -52.869 54.435
681600 95.778 -52.892 54.424
681601 95.794 -52.915 54.406
681602 95.810 -52.938 54.387
681603 95.826 -52.961 54.369
681604 95.841 -52.984 54.350
681605 95.857 -53.007 54.326
681606 95.873 -53.030 54.294
681607 95.888 -53.053 54.258
681608 95.904 -53.076 54.222
681609 95.920 -53.099 54.188
681610 95.935 -53.122 54.157
681611 95.950 -53.145 54.128
681612 95.963 -53.164 54.098
681613 95.972 -53.173 54.067
681614 95.980 -53.180 54.037
681615 95.989 -53.189 54.010
681616 95.998 -53.196 53.981
681617 96.007 -53.204 53.952
681618 96.016 -53.212 53.923
681619 96.025 -53.220 53.895
681620 96.034 -53.228 53.868
681621 96.042 -53.236 53.840
681622 96.051 -53.244 53.814
681623 96.060 -53.252 53.785
681624 96.069 -53.259 53.757
681625 96.078 -53.267 53.728
681626 96.087 -53.275 53.699
681627 96.096 -53.283 53.670
681628 96.105 -53.289 53.642
681629 96.115 -53.293 53.613
681630 96.125 -53.297 53.585
681631 96.135 -53.301 53.557
681632 96.146 -53.305 53.530
681633 96.156 -53.308 53.505
681634 96.166 -53.311 53.479
681635 96.177 -53.315 53.454
681636 96.187 -53.319 53.430
681637 96.205 -53.322 53.408
681638 96.233 -53.324 53.387
681639 96.260 -53.326 53.366
681640 96.288 -53.328 53.344
681641 96.316 -53.330 53.323
681642 96.343 -53.333 53.303
681643 96.371 -53.334 53.284
681644 96.398 -53.336 53.266
681645 96.426 -53.339 53.249
681646 96.454 -53.341 53.232
681647 96.481 -53.342 53.215
681648 96.509 -53.345 53.198
681649 96.536 -53.347 53.183
681650 96.564 -53.349 53.167
681651 96.592 -53.351 53.151
681652 96.619 -53.353 53.134
681653 96.647 -53.355 53.115
681654 96.673 -53.358 53.096
681655 96.694 -53.364 53.078
681656 96.717 -53.371 53.058
681657 96.739 -53.377 53.036
681658 96.760 -53.383 53.011
681659 96.793 -53.391 52.987
681660 96.826 -53.399 52.961
681661 96.859 -53.408 52.934
681662 96.892 -53.416 52.907
681663 96.925 -53.424 52.880
681664 96.958 -53.433 52.852
681665 96.992 -53.441 52.824
681666 97.025 -53.449 52.795
681667 97.058 -53.458 52.767
681668 97.091 -53.466 52.739
681669 97.124 -53.474 52.711
681670 97.158 -53.483 52.682
681671 97.191 -53.491 52.651
681672 97.224 -53.499 52.622
681673 97.257 -53.508 52.591
681674 97.291 -53.516 52.561
681675 97.324 -53.524 52.529
681676 97.357 -53.533 52.497
681677 97.388 -53.540 52.464
681678 97.406 -53.549 52.430
681679 97.428 -53.556 52.398
681680 97.449 -53.564 52.364
681681 97.470 -53.571 52.328
681682 97.496 -53.580 52.290
681683 97.527 -53.590 52.252
681684 97.557 -53.596 52.213
681685 97.585 -53.602 52.176
681686 97.613 -53.607 52.140
681687 97.641 -53.612 52.103
681688 97.669 -53.618 52.066
681689 97.697 -53.622 52.031
681690 97.725 -53.628 51.997
681691 97.753 -53.633 51.964
681692 97.781 -53.638 51.932
681693 97.809 -53.643 51.901
681694 97.837 -53.648 51.870
681695 97.865 -53.653 51.840
681696 97.890 -53.656 51.810
681697 97.904 -53.659 51.780
681698 97.917 -53.669 51.749
681699 97.929 -53.680 51.718
681700 97.941 -53.690 51.686
681701 97.953 -53.700 51.654
681702 97.965 -53.711 51.622
681703 97.978 -53.722 51.588
681704 98.002 -53.736 51.553
681705 98.027 -53.751 51.516
681706 98.053 -53.766 51.478
681707 98.075 -53.776 51.439
681708 98.092 -53.783 51.402
681709 98.110 -53.790 51.364
681710 98.129 -53.796 51.325
681711 98.147 -53.803 51.284
681712 98.179 -53.812 51.242
681713 98.212 -53.820 51.198
681714 98.244 -53.829 51.153
681715 98.270 -53.840 51.111
681716 98.286 -53.855 51.069
681717 98.309 -53.872 51.028
681718 98.335 -53.891 50.987
681719 98.360 -53.910 50.947
681720 98.386 -53.928 50.906
681721 98.411 -53.947 50.865
681722 98.437 -53.965 50.824
681723 98.462 -53.983 50.784
681724 98.488 -54.002 50.744
681725 98.513 -54.020 50.705
681726 98.538 -54.038 50.667
681727 98.564 -54.057 50.631
682217 93.493 -51.242 56.317
682218 93.524 -51.259 56.301
682219 93.555 -51.277 56.284
682220 93.586 -51.294 56.265
682221 93.617 -51.312 56.245
682222 93.648 -51.329 56.223
682223 93.679 -51.346 56.199
682224 93.710 -51.364 56.174
682225 93.741 -51.381 56.147
682226 93.772 -51.399 56.118
682227 93.803 -51.417 56.088
682228 93.834 -51.434 56.057
682229 93.865 -51.452 56.024
682230 93.892 -51.472 55.991
682231 93.919 -51.491 55.956
682232 93.945 -51.511 55.922
682233 93.973 -51.531 55.887
682234 93.999 -51.550 55.851
682235 94.027 -51.570 55.815
682289 95.480 -52.637 54.536
682290 95.507 -52.656 54.525
682291 95.534 -52.675 54.512
682292 95.560 -52.694 54.499
682293 95.587 -52.713 54.487
682294 95.614 -52.732 54.476
682295 95.641 -52.751 54.464
682296 95.668 -52.770 54.453
682297 95.694 -52.789 54.442
682298 95.721 -52.808 54.432
682299 95.748 -52.827 54.423
682300 95.779 -52.848 54.415
682301 95.795 -52.872 54.401
682302 95.811 -52.894 54.382
682303 95.826 -52.917 54.363
682304 95.842 -52.941 54.345
682305 95.858 -52.963 54.324
682306 95.873 -52.986 54.299
682307 95.887 -53.009 54.268
682308 95.901 -53.031 54.233
682309 95.915 -53.054 54.199
682310 95.930 -53.076 54.167
682311 95.944 -53.098 54.137
682312 95.958 -53.121 54.107
682313 95.972 -53.143 54.078
682314 95.983 -53.155 54.048
682315 95.992 -53.164 54.020
682316 96.001 -53.171 53.991
682317 96.010 -53.179 53.964
682318 96.018 -53.187 53.935
682319 96.027 -53.195 53.905
682320 96.036 -53.203 53.877
682321 96.045 -53.211 53.849
682322 96.054 -53.219 53.821
682323 96.062 -53.227 53.794
682324 96.072 -53.234 53.764
682325 96.081 -53.242 53.735
682326 96.089 -53.250 53.705
682327 96.098 -53.258 53.676
682328 96.107 -53.266 53.647
682329 96.116 -53.273 53.620
682330 96.126 -53.277 53.592
682331 96.136 -53.280 53.563
682332 96.147 -53.284 53.535
682333 96.157 -53.288 53.509
682334 96.167 -53.292 53.484
682335 96.177 -53.295 53.459
682336 96.188 -53.299 53.434
682337 96.198 -53.302 53.410
682338 96.217 -53.305 53.388
682339 96.244 -53.308 53.366
682340 96.272 -53.310 53.345
682341 96.300 -53.311 53.324
682342 96.327 -53.314 53.302
682343 96.355 -53.316 53.282
682344 96.382 -53.317 53.263
682345 96.410 -53.320 53.246
682346 96.438 -53.322 53.229
682347 96.465 -53.324 53.211
682348 96.493 -53.326 53.194
682349 96.520 -53.328 53.178
682350 96.548 -53.330 53.162
682351 96.576 -53.333 53.147
682352 96.603 -53.334 53.132
682353 96.631 -53.336 53.114
682354 96.658 -53.339 53.097
682355 96.682 -53.343 53.079
682356 96.704 -53.349 53.061
682357 96.726 -53.355 53.042
682358 96.748 -53.361 53.021
682359 96.770 -53.368 52.997
682360 96.792 -53.374 52.973
682361 96.825 -53.382 52.949
682362 96.858 -53.390 52.924
682363 96.891 -53.399 52.898
682364 96.924 -53.407 52.871
682365 96.957 -53.415 52.843
682366 96.991 -53.424 52.817
682367 97.024 -53.432 52.789
682368 97.057 -53.440 52.762
682369 97.090 -53.449 52.735
682370 97.123 -53.457 52.707
682371 97.157 -53.465 52.677
682372 97.190 -53.474 52.648
682373 97.223 -53.482 52.619
682374 97.256 -53.490 52.588
682375 97.290 -53.499 52.558
682376 97.323 -53.507 52.527
682377 97.356 -53.515 52.494
682378 97.393 -53.522 52.461
682379 97.412 -53.531 52.428
682380 97.434 -53.539 52.395
682381 97.455 -53.546 52.361
682382 97.476 -53.554 52.325
682383 97.497 -53.562 52.289
682384 97.527 -53.571 52.252
682385 97.558 -53.581 52.214
682386 97.589 -53.590 52.178
682387 97.618 -53.598 52.143
682388 97.647 -53.603 52.107
682389 97.675 -53.609 52.070
682390 97.703 -53.613 52.036
682391 97.731 -53.618 52.003
682392 97.759 -53.624 51.971
682393 97.787 -53.629 51.941
682394 97.815 -53.634 51.911
682395 97.843 -53.639 51.880
682396 97.871 -53.644 51.849
682397 97.890 -53.642 51.819
682398 97.907 -53.639 51.789
682399 97.919 -53.650 51.758
682400 97.931 -53.660 51.727
682401 97.943 -53.670 51.695
682402 97.955 -53.681 51.664
682403 97.967 -53.691 51.632
682404 97.980 -53.702 51.600
682405 98.004 -53.716 51.566
682406 98.029 -53.731 51.529
682407 98.055 -53.746 51.491
682408 98.081 -53.761 51.453
682409 98.100 -53.769 51.416
682410 98.118 -53.775 51.378
682411 98.136 -53.782 51.340
682412 98.154 -53.789 51.301
682413 98.180 -53.797 51.260
682414 98.213 -53.805 51.218
682415 98.245 -53.814 51.175
682416 98.275 -53.824 51.134
682417 98.290 -53.838 51.094
682418 98.311 -53.855 51.055
682419 98.337 -53.873 51.015
682420 98.362 -53.892 50.976
682421 98.387 -53.910 50.936
682422 98.413 -53.928 50.896
682423 98.438 -53.947 50.855
682424 98.463 -53.966 50.815
682425 98.489 -53.984 50.775
682426 98.514 -54.003 50.736
682427 98.540 -54.021 50.698
682428 98.565 -54.040 50.662
682429 98.591 -54.058 50.627
682919 93.542 -51.235 56.293
682920 93.572 -51.252 56.275
682921 93.604 -51.270 56.256
682922 93.634 -51.287 56.236
682923 93.665 -51.305 56.214
682924 93.696 -51.322 56.189
682925 93.727 -51.340 56.164
682926 93.758 -51.357 56.136
682927 93.789 -51.374 56.107
682928 93.820 -51.393 56.076
682929 93.851 -51.410 56.045
682930 93.882 -51.428 56.012
682931 93.911 -51.446 55.979
682932 93.937 -51.466 55.944
682933 93.965 -51.486 55.910
682934 93.991 -51.506 55.875
682935 94.018 -51.525 55.840
682991 95.526 -52.631 54.518
682992 95.553 -52.650 54.505
682993 95.580 -52.670 54.492
682994 95.607 -52.690 54.479
682995 95.634 -52.709 54.467
682996 95.661 -52.730 54.454
682997 95.687 -52.750 54.441
682998 95.714 -52.769 54.429
682999 95.741 -52.789 54.417
683000 95.768 -52.809 54.406
683001 95.796 -52.828 54.395
683002 95.811 -52.850 54.378
683003 95.825 -52.873 54.358
683004 95.839 -52.895 54.338
683005 95.853 -52.918 54.318
683006 95.868 -52.940 54.296
683007 95.882 -52.963 54.270
683008 95.896 -52.985 54.240
683009 95.910 -53.007 54.208
683010 95.924 -53.030 54.176
683011 95.939 -53.052 54.146
683012 95.953 -53.075 54.117
683013 95.967 -53.098 54.088
683014 95.981 -53.120 54.059
683015 95.994 -53.139 54.030
683016 96.003 -53.146 54.002
683017 96.012 -53.154 53.973
683018 96.021 -53.162 53.946
683019 96.030 -53.170 53.917
683020 96.038 -53.178 53.887
683021 96.047 -53.186 53.858
683022 96.056 -53.193 53.830
683023 96.065 -53.201 53.802
683024 96.074 -53.209 53.774
683025 96.083 -53.217 53.744
683026 96.092 -53.225 53.715
683027 96.100 -53.233 53.685
683028 96.109 -53.241 53.656
683029 96.118 -53.248 53.627
683030 96.127 -53.257 53.599
683031 96.137 -53.261 53.571
683032 96.148 -53.264 53.543
683033 96.158 -53.267 53.515
683034 96.168 -53.271 53.489
683035 96.178 -53.275 53.464
683036 96.188 -53.279 53.438
683037 96.198 -53.283 53.413
683038 96.209 -53.286 53.390
683039 96.228 -53.289 53.368
683040 96.256 -53.291 53.345
683041 96.284 -53.293 53.324
683042 96.311 -53.295 53.303
683043 96.338 -53.297 53.283
683044 96.366 -53.299 53.262
683045 96.394 -53.302 53.244
683046 96.421 -53.303 53.226
683047 96.449 -53.305 53.208
683048 96.476 -53.308 53.191
683049 96.504 -53.309 53.174
683050 96.532 -53.311 53.158
683051 96.559 -53.314 53.142
683052 96.587 -53.316 53.128
683053 96.615 -53.317 53.113
683054 96.642 -53.320 53.096
683055 96.670 -53.322 53.079
683056 96.692 -53.328 53.062
683057 96.714 -53.334 53.045
683058 96.736 -53.340 53.026
683059 96.757 -53.346 53.006
683060 96.780 -53.352 52.984
683061 96.801 -53.358 52.961
683062 96.824 -53.364 52.938
683063 96.857 -53.373 52.914
683064 96.890 -53.381 52.889
683065 96.923 -53.389 52.863
683066 96.956 -53.398 52.837
683067 96.990 -53.406 52.810
683068 97.023 -53.415 52.783
683069 97.056 -53.423 52.757
683070 97.089 -53.431 52.731
683071 97.122 -53.440 52.703
683072 97.156 -53.448 52.674
683073 97.189 -53.456 52.645
683074 97.222 -53.465 52.617
683075 97.255 -53.474 52.587
683076 97.289 -53.482 52.557
683077 97.322 -53.490 52.527
683078 97.355 -53.499 52.494
683079 97.388 -53.507 52.461
683080 97.432 -53.510 52.428
683081 97.439 -53.521 52.395
683082 97.460 -53.530 52.361
683083 97.481 -53.537 52.325
683084 97.503 -53.545 52.289
683085 97.527 -53.553 52.253
683086 97.558 -53.562 52.217
683087 97.588 -53.571 52.182
683088 97.619 -53.581 52.146
683089 97.650 -53.590 52.110
683090 97.681 -53.599 52.075
683091 97.709 -53.604 52.041
683092 97.737 -53.609 52.009
683093 97.765 -53.615 51.980
683094 97.793 -53.620 51.951
683095 97.821 -53.625 51.921
683096 97.849 -53.630 51.890
683097 97.874 -53.632 51.858
683098 97.890 -53.628 51.827
683099 97.908 -53.623 51.797
683100 97.921 -53.630 51.766
683101 97.933 -53.640 51.735
683102 97.945 -53.650 51.705
683103 97.957 -53.661 51.673
683104 97.970 -53.672 51.642
683105 97.982 -53.681 51.611
683106 98.006 -53.696 51.578
683107 98.031 -53.711 51.542
683108 98.057 -53.726 51.504
683109 98.083 -53.740 51.467
683110 98.107 -53.755 51.429
683111 98.125 -53.761 51.393
683112 98.143 -53.768 51.355
683113 98.161 -53.775 51.316
683114 98.181 -53.781 51.277
683115 98.214 -53.791 51.237
683116 98.246 -53.799 51.197
683117 98.279 -53.808 51.157
683118 98.295 -53.822 51.119
683119 98.313 -53.838 51.081
683120 98.338 -53.856 51.042
683121 98.364 -53.874 51.004
683122 98.389 -53.893 50.965
683123 98.415 -53.911 50.925
683124 98.440 -53.930 50.886
683125 98.465 -53.948 50.846
683126 98.491 -53.966 50.806
683127 98.516 -53.985 50.767
683128 98.541 -54.004 50.729
683129 98.567 -54.022 50.692
683130 98.592 -54.041 50.657
683621 93.589 -51.228 56.266
683622 93.620 -51.246 56.247
683623 93.651 -51.263 56.226
683624 93.682 -51.280 56.203
683625 93.713 -51.298 56.179
683626 93.744 -51.315 56.153
683627 93.775 -51.333 56.125
683628 93.806 -51.350 56.095
683629 93.837 -51.368 56.064
683630 93.868 -51.386 56.032
683631 93.899 -51.403 56.000
683632 93.929 -51.421 55.966
683633 93.957 -51.440 55.933
683634 93.983 -51.460 55.898
683635 94.010 -51.480 55.863
683636 94.037 -51.500 55.828
683692 95.545 -52.606 54.512
683693 95.572 -52.625 54.497
683694 95.599 -52.645 54.483
683695 95.625 -52.665 54.470
683696 95.653 -52.684 54.457
683697 95.679 -52.704 54.444
683698 95.706 -52.724 54.430
683699 95.733 -52.744 54.416
683700 95.760 -52.763 54.402
683701 95.787 -52.783 54.389
683702 95.806 -52.805 54.376
683703 95.819 -52.827 54.357
683704 95.834 -52.850 54.335
683705 95.848 -52.872 54.314
683706 95.862 -52.894 54.292
683707 95.876 -52.916 54.268
683708 95.891 -52.939 54.242
683709 95.905 -52.962 54.214
683710 95.919 -52.984 54.185
683711 95.933 -53.007 54.156
683712 95.948 -53.029 54.126
683713 95.962 -53.051 54.097
683714 95.976 -53.073 54.070
683715 95.990 -53.096 54.042
683716 96.005 -53.119 54.012
683717 96.014 -53.129 53.984
683718 96.023 -53.137 53.956
683719 96.032 -53.145 53.927
683720 96.041 -53.153 53.898
683721 96.049 -53.161 53.869
683722 96.058 -53.168 53.840
683723 96.067 -53.176 53.811
683724 96.076 -53.184 53.783
683725 96.085 -53.192 53.754
683726 96.094 -53.200 53.725
683727 96.103 -53.208 53.695
683728 96.112 -53.216 53.666
683729 96.120 -53.223 53.637
683730 96.129 -53.232 53.608
683731 96.138 -53.239 53.580
683732 96.148 -53.244 53.551
683733 96.159 -53.248 53.523
683734 96.169 -53.251 53.496
683735 96.179 -53.255 53.470
683736 96.189 -53.258 53.445
683737 96.199 -53.262 53.419
683738 96.209 -53.266 53.394
683739 96.220 -53.270 53.371
683740 96.240 -53.272 53.348
683741 96.267 -53.274 53.326
683742 96.295 -53.277 53.305
683743 96.322 -53.278 53.284
683744 96.350 -53.280 53.264
683745 96.378 -53.283 53.243
683746 96.405 -53.285 53.225
683747 96.433 -53.286 53.206
683748 96.460 -53.289 53.189
683749 96.488 -53.291 53.172
683750 96.516 -53.293 53.155
683751 96.543 -53.295 53.138
683752 96.571 -53.297 53.122
683753 96.599 -53.299 53.108
683754 96.626 -53.301 53.093
683755 96.654 -53.303 53.078
683756 96.679 -53.307 53.062
683757 96.701 -53.313 53.046
683758 96.723 -53.319 53.029
683759 96.745 -53.325 53.012
683760 96.767 -53.331 52.993
683761 96.789 -53.337 52.971
683762 96.811 -53.343 52.949
683763 96.833 -53.349 52.927
683764 96.856 -53.355 52.904
683765 96.889 -53.364 52.880
683766 96.922 -53.372 52.856
683767 96.955 -53.380 52.830
683768 96.989 -53.389 52.804
683769 97.022 -53.397 52.778
683770 97.055 -53.405 52.753
683771 97.088 -53.415 52.727
683772 97.121 -53.423 52.700
683773 97.155 -53.431 52.672
683774 97.188 -53.440 52.643
683775 97.221 -53.448 52.615
683776 97.254 -53.456 52.586
683777 97.288 -53.465 52.557
683778 97.321 -53.473 52.527
683779 97.354 -53.481 52.495
683780 97.387 -53.490 52.462
683781 97.426 -53.496 52.429
683782 97.454 -53.502 52.395
683783 97.466 -53.512 52.362
683784 97.487 -53.520 52.327
683785 97.509 -53.527 52.291
683786 97.530 -53.536 52.255
683787 97.558 -53.544 52.220
683788 97.588 -53.553 52.186
683789 97.619 -53.563 52.151
683790 97.650 -53.572 52.115
683791 97.681 -53.581 52.080
683792 97.712 -53.591 52.047
683793 97.743 -53.600 52.016
683794 97.771 -53.606 51.987
683795 97.799 -53.611 51.958
683796 97.827 -53.616 51.929
683797 97.855 -53.621 51.898
683798 97.874 -53.618 51.866
683799 97.891 -53.614 51.835
683800 97.908 -53.609 51.804
683801 97.923 -53.609 51.774
683802 97.935 -53.620 51.744
683803 97.948 -53.631 51.714
683804 97.960 -53.641 51.683
683805 97.972 -53.651 51.652
683806 97.984 -53.662 51.621
683807 98.008 -53.676 51.588
683808 98.033 -53.691 51.552
683809 98.059 -53.706 51.516
683810 98.085 -53.721 51.480
683811 98.110 -53.735 51.445
683812 98.133 -53.747 51.408
683813 98.151 -53.754 51.371
683814 98.169 -53.760 51.333
683815 98.186 -53.767 51.295
683816 98.214 -53.775 51.257
683817 98.247 -53.784 51.218
683818 98.279 -53.793 51.180
683819 98.300 -53.806 51.144
683820 98.315 -53.820 51.107
683821 98.340 -53.838 51.068
683822 98.365 -53.856 51.031
683823 98.391 -53.875 50.993
683824 98.416 -53.894 50.953
683825 98.442 -53.912 50.915
683826 98.467 -53.931 50.876
683827 98.493 -53.949 50.837
683828 98.518 -53.967 50.797
683829 98.544 -53.986 50.758
683830 98.569 -54.004 50.721
683831 98.594 -54.023 50.686
684323 93.637 -51.221 56.238
684324 93.669 -51.239 56.216
684325 93.699 -51.256 56.193
684326 93.730 -51.274 56.168
684327 93.761 -51.291 56.141
684328 93.792 -51.309 56.113
684329 93.823 -51.326 56.083
684330 93.854 -51.344 56.052
684331 93.885 -51.362 56.020
684332 93.916 -51.379 55.988
684333 93.947 -51.396 55.954
684334 93.975 -51.415 55.920
684335 94.002 -51.435 55.886
684336 94.029 -51.455 55.851
684393 95.563 -52.581 54.504
684394 95.591 -52.600 54.490
684395 95.617 -52.620 54.474
684396 95.644 -52.640 54.461
684397 95.671 -52.659 54.447
684398 95.698 -52.679 54.432
684399 95.725 -52.699 54.418
684400 95.752 -52.719 54.403
684401 95.779 -52.738 54.387
684402 95.801 -52.759 54.372
684403 95.814 -52.782 54.356
684404 95.828 -52.803 54.336
684405 95.842 -52.825 54.314
684406 95.857 -52.848 54.292
684407 95.871 -52.871 54.269
684408 95.885 -52.893 54.244
684409 95.899 -52.916 54.218
684410 95.913 -52.938 54.191
684411 95.928 -52.960 54.164
684412 95.942 -52.983 54.136
684413 95.956 -53.005 54.108
684414 95.970 -53.028 54.080
684415 95.985 -53.050 54.052
684416 95.999 -53.073 54.024
684417 96.013 -53.095 53.995
684418 96.025 -53.112 53.968
684419 96.034 -53.120 53.938
684420 96.043 -53.127 53.910
684421 96.052 -53.136 53.881
684422 96.061 -53.143 53.851
684423 96.070 -53.151 53.822
684424 96.079 -53.159 53.793
684425 96.087 -53.167 53.764
684426 96.096 -53.175 53.735
684427 96.105 -53.183 53.705
684428 96.114 -53.191 53.675
684429 96.123 -53.198 53.646
684430 96.131 -53.207 53.618
684431 96.141 -53.214 53.589
684432 96.150 -53.223 53.561
684433 96.159 -53.227 53.532
684434 96.170 -53.231 53.503
684435 96.180 -53.235 53.477
684436 96.190 -53.238 53.451
684437 96.200 -53.242 53.426
684438 96.210 -53.245 53.401
684439 96.221 -53.249 53.376
684440 96.231 -53.253 53.353
684441 96.251 -53.255 53.329
684442 96.278 -53.258 53.307
684443 96.306 -53.260 53.286
684444 96.334 -53.262 53.266
684445 96.362 -53.264 53.245
684446 96.389 -53.266 53.224
684447 96.417 -53.268 53.206
684448 96.444 -53.270 53.187
684449 96.472 -53.272 53.170
684450 96.500 -53.274 53.153
684451 96.527 -53.277 53.136
684452 96.555 -53.278 53.120
684453 96.583 -53.280 53.105
684454 96.610 -53.283 53.090
684455 96.638 -53.285 53.077
684456 96.665 -53.286 53.062
684457 96.689 -53.292 53.047
684458 96.711 -53.298 53.031
684459 96.733 -53.304 53.015
684460 96.754 -53.310 52.998
684461 96.777 -53.316 52.981
684462 96.798 -53.322 52.960
684463 96.820 -53.328 52.939
684464 96.842 -53.334 52.917
684465 96.864 -53.340 52.895
684466 96.888 -53.346 52.873
684467 96.921 -53.355 52.849
684468 96.954 -53.364 52.823
684469 96.987 -53.372 52.798
684470 97.021 -53.380 52.774
684471 97.054 -53.389 52.749
684472 97.087 -53.397 52.724
684473 97.120 -53.405 52.698
684474 97.154 -53.414 52.670
684475 97.187 -53.422 52.642
684476 97.220 -53.430 52.615
684477 97.253 -53.439 52.585
684478 97.287 -53.447 52.557
684479 97.320 -53.455 52.527
684480 97.353 -53.464 52.495
684481 97.386 -53.472 52.463
684482 97.419 -53.480 52.430
684483 97.476 -53.480 52.397
684484 97.475 -53.494 52.364
684485 97.493 -53.502 52.329
684486 97.514 -53.511 52.294
684487 97.535 -53.518 52.258
684488 97.557 -53.526 52.224
684489 97.588 -53.535 52.190
684490 97.619 -53.545 52.156
684491 97.650 -53.554 52.121
684492 97.680 -53.563 52.086
684493 97.712 -53.573 52.053
684494 97.742 -53.582 52.022
684495 97.773 -53.592 51.993
684496 97.804 -53.601 51.964
684497 97.833 -53.607 51.936
684498 97.858 -53.609 51.905
684499 97.874 -53.605 51.872
684500 97.891 -53.599 51.841
684501 97.909 -53.595 51.811
684502 97.925 -53.590 51.782
684503 97.938 -53.600 51.752
684504 97.950 -53.611 51.722
684505 97.962 -53.621 51.692
684506 97.974 -53.631 51.661
684507 97.987 -53.642 51.630
684508 98.010 -53.656 51.597
684509 98.035 -53.671 51.564
684510 98.061 -53.686 51.529
684511 98.086 -53.700 51.494
684512 98.112 -53.715 51.460
684513 98.138 -53.731 51.423
684514 98.158 -53.740 51.387
684515 98.176 -53.747 51.350
684516 98.194 -53.753 51.313
684517 98.215 -53.760 51.276
684518 98.248 -53.769 51.239
684519 98.280 -53.778 51.202
684520 98.305 -53.789 51.166
684521 98.320 -53.803 51.131
684522 98.342 -53.821 51.094
684523 98.367 -53.839 51.057
684524 98.392 -53.857 51.020
684525 98.418 -53.876 50.981
684526 98.443 -53.894 50.943
684527 98.469 -53.913 50.904
684528 98.494 -53.932 50.864
684529 98.520 -53.950 50.825
684530 98.545 -53.969 50.787
684531 98.571 -53.987 50.749
684532 98.596 -54.005 50.713
685025 93.685 -51.215 56.206
685026 93.716 -51.232 56.183
685027 93.747 -51.249 56.156
685028 93.778 -51.267 56.129
685029 93.809 -51.284 56.101
685030 93.840 -51.302 56.071
685031 93.871 -51.319 56.040
685032 93.902 -51.337 56.008
685033 93.933 -51.355 55.975
685034 93.964 -51.372 55.942
685035 93.994 -51.390 55.908
685036 94.021 -51.410 55.874
685094 95.583 -52.555 54.497
685095 95.609 -52.575 54.483
685096 95.636 -52.594 54.467
685097 95.663 -52.614 54.453
685098 95.690 -52.634 54.437
685099 95.717 -52.653 54.421
685100 95.744 -52.674 54.406
685101 95.771 -52.694 54.390
685102 95.796 -52.713 54.372
685103 95.809 -52.736 54.355
685104 95.823 -52.758 54.337
685105 95.837 -52.780 54.317
685106 95.851 -52.802 54.295
685107 95.866 -52.825 54.272
685108 95.880 -52.847 54.248
685109 95.894 -52.869 54.223
685110 95.908 -52.892 54.198
685111 95.922 -52.914 54.171
685112 95.937 -52.937 54.144
685113 95.951 -52.959 54.117
685114 95.965 -52.982 54.090
685115 95.979 -53.004 54.062
685116 95.993 -53.026 54.035
685117 96.008 -53.049 54.006
685118 96.022 -53.071 53.977
685119 96.036 -53.094 53.950
685120 96.045 -53.102 53.920
685121 96.054 -53.111 53.892
685122 96.063 -53.118 53.863
685123 96.072 -53.126 53.834
685124 96.081 -53.134 53.804
685125 96.090 -53.142 53.775
685126 96.099 -53.150 53.747
685127 96.107 -53.158 53.717
685128 96.116 -53.166 53.688
685129 96.125 -53.173 53.658
685130 96.134 -53.182 53.629
685131 96.143 -53.189 53.600
685132 96.152 -53.198 53.572
685133 96.161 -53.205 53.543
685134 96.170 -53.211 53.514
685135 96.180 -53.214 53.486
685136 96.191 -53.218 53.459
685137 96.201 -53.222 53.433
685138 96.211 -53.225 53.408
685151 96.511 -53.258 53.135
685152 96.539 -53.260 53.118
685153 96.567 -53.261 53.103
685154 96.594 -53.264 53.088
685155 96.621 -53.266 53.074
685156 96.649 -53.268 53.060
685157 96.676 -53.270 53.046
685158 96.698 -53.277 53.032
685159 96.720 -53.283 53.017
685160 96.742 -53.289 53.002
685161 96.764 -53.295 52.985
685162 96.786 -53.301 52.969
685163 96.808 -53.307 52.949
685164 96.830 -53.313 52.929
685165 96.852 -53.319 52.908
685166 96.874 -53.325 52.887
685167 96.895 -53.331 52.865
685168 96.920 -53.338 52.843
685169 96.953 -53.346 52.817
685170 96.986 -53.355 52.794
685173 97.086 -53.380 52.721
685174 97.119 -53.388 52.695
685175 97.153 -53.396 52.668
685176 97.186 -53.405 52.641
685177 97.219 -53.413 52.614
685178 97.252 -53.421 52.585
685179 97.286 -53.430 52.557
685180 97.319 -53.438 52.527
685181 97.352 -53.446 52.496
685182 97.385 -53.455 52.464
685183 97.418 -53.463 52.432
685184 97.458 -53.469 52.398
685185 97.498 -53.472 52.365
685186 97.499 -53.486 52.332
685187 97.520 -53.493 52.297
685188 97.541 -53.501 52.262
685189 97.563 -53.509 52.227
685190 97.588 -53.517 52.193
685191 97.618 -53.527 52.161
685192 97.650 -53.536 52.126
685193 97.680 -53.546 52.092
685194 97.711 -53.555 52.059
685195 97.742 -53.564 52.028
685196 97.773 -53.574 51.999
685197 97.804 -53.583 51.970
685198 97.835 -53.592 51.941
685199 97.858 -53.595 51.910
685200 97.875 -53.590 51.878
685201 97.892 -53.586 51.848
685202 97.909 -53.581 51.819
685203 97.926 -53.576 51.789
685204 97.940 -53.581 51.759
685205 97.952 -53.590 51.730
685206 97.964 -53.601 51.700
685207 97.977 -53.612 51.670
685208 97.989 -53.622 51.639
685209 98.012 -53.636 51.607
685210 98.037 -53.651 51.575
685211 98.063 -53.665 51.542
685212 98.088 -53.681 51.508
685213 98.114 -53.696 51.474
685214 98.140 -53.710 51.439
685215 98.165 -53.725 51.402
685216 98.183 -53.732 51.367
685217 98.202 -53.739 51.331
685218 98.219 -53.746 51.295
685219 98.249 -53.754 51.259
685220 98.281 -53.762 51.224
685221 98.309 -53.773 51.188
685222 98.325 -53.787 51.154
685223 98.343 -53.803 51.118
685224 98.369 -53.822 51.081
685225 98.394 -53.840 51.045
685226 98.420 -53.859 51.008
685227 98.445 -53.877 50.969
685228 98.471 -53.895 50.930
685229 98.496 -53.914 50.891
685230 98.521 -53.932 50.852
685231 98.547 -53.950 50.813
685232 98.572 -53.969 50.775
685233 98.597 -53.988 50.738
685796 95.628 -52.550 54.475
685797 95.655 -52.569 54.459
685798 95.682 -52.589 54.444
685799 95.709 -52.609 54.427
685800 95.736 -52.628 54.410
685801 95.763 -52.648 54.393
685802 95.790 -52.668 54.376
685803 95.804 -52.690 54.358
685804 95.818 -52.713 54.340
685805 95.832 -52.735 54.321
685806 95.846 -52.756 54.300
685807 95.861 -52.778 54.278
685808 95.875 -52.801 54.255
685809 95.889 -52.823 54.230
685810 95.903 -52.846 54.204
685811 95.917 -52.869 54.179
685812 95.932 -52.891 54.153
685813 95.946 -52.913 54.127
685814 95.960 -52.935 54.100
685815 95.974 -52.958 54.072
685816 95.988 -52.981 54.044
685817 96.003 -53.003 54.018
685818 96.017 -53.026 53.990
685819 96.031 -53.048 53.960
685820 96.045 -53.070 53.931
685821 96.056 -53.086 53.903
685822 96.065 -53.093 53.874
685823 96.075 -53.101 53.846
685824 96.083 -53.109 53.816
685825 96.092 -53.117 53.787
685826 96.101 -53.125 53.757
685827 96.110 -53.132 53.729
685828 96.119 -53.141 53.700
685829 96.127 -53.148 53.671
685830 96.136 -53.157 53.641
685831 96.145 -53.164 53.612
685832 96.154 -53.172 53.583
685833 96.163 -53.180 53.555
685834 96.172 -53.188 53.526
685856 96.633 -53.249 53.058
685857 96.661 -53.252 53.044
685858 96.686 -53.255 53.030
685859 96.708 -53.261 53.016
685860 96.730 -53.267 53.002
685861 96.751 -53.273 52.988
685862 96.773 -53.280 52.973
685863 96.796 -53.286 52.957
685864 96.817 -53.292 52.938
685865 96.839 -53.298 52.919
685866 96.861 -53.304 52.899
685867 96.883 -53.310 52.879
685876 97.152 -53.379 52.666
685877 97.185 -53.387 52.640
685878 97.218 -53.396 52.613
685879 97.251 -53.404 52.585
685880 97.285 -53.412 52.557
685881 97.318 -53.421 52.528
685882 97.351 -53.429 52.497
685883 97.384 -53.437 52.465
685884 97.417 -53.446 52.434
685885 97.450 -53.454 52.401
685886 97.517 -53.451 52.368
685887 97.519 -53.464 52.334
685888 97.525 -53.476 52.300
685889 97.546 -53.483 52.265
685890 97.568 -53.492 52.231
685891 97.589 -53.499 52.198
685892 97.618 -53.509 52.165
685893 97.649 -53.518 52.130
685894 97.680 -53.527 52.098
685895 97.711 -53.537 52.065
685896 97.742 -53.546 52.034
685897 97.773 -53.556 52.004
685898 97.804 -53.565 51.974
685899 97.835 -53.574 51.944
685900 97.861 -53.581 51.914
685901 97.875 -53.577 51.883
685902 97.892 -53.572 51.853
685903 97.909 -53.567 51.825
685904 97.926 -53.562 51.797
685905 97.942 -53.560 51.768
685906 97.954 -53.571 51.738
685907 97.966 -53.581 51.708
685908 97.979 -53.592 51.678
685909 97.991 -53.602 51.648
685910 98.014 -53.616 51.617
685911 98.039 -53.631 51.586
685912 98.065 -53.646 51.555
685913 98.090 -53.661 51.522
685914 98.116 -53.675 51.489
685915 98.142 -53.690 51.454
685916 98.167 -53.706 51.419
685917 98.191 -53.719 51.384
685918 98.209 -53.725 51.348
685919 98.227 -53.731 51.313
685920 98.249 -53.739 51.277
685921 98.282 -53.747 51.244
685922 98.314 -53.756 51.209
685923 98.330 -53.771 51.175
685924 98.346 -53.786 51.140
685925 98.370 -53.804 51.104
685926 98.396 -53.822 51.069
685927 98.421 -53.841 51.032
685928 98.447 -53.860 50.994
685929 98.472 -53.878 50.956
685930 98.498 -53.896 50.917
685931 98.523 -53.915 50.878
685932 98.549 -53.933 50.838
685933 98.574 -53.952 50.799
685934 98.599 -53.970 50.763
685935 98.625 -53.988 50.727
686497 95.647 -52.525 54.467
686498 95.674 -52.544 54.451
686499 95.701 -52.564 54.435
686500 95.728 -52.584 54.417
686501 95.755 -52.603 54.399
686502 95.781 -52.623 54.381
686503 95.800 -52.644 54.363
686504 95.813 -52.666 54.345
686505 95.827 -52.689 54.325
686506 95.841 -52.711 54.306
686507 95.855 -52.733 54.284
686508 95.870 -52.756 54.262
686509 95.883 -52.778 54.238
686510 95.897 -52.800 54.213
686511 95.912 -52.822 54.188
686512 95.926 -52.844 54.163
686513 95.940 -52.867 54.137
686514 95.954 -52.890 54.111
686515 95.969 -52.912 54.084
686516 95.983 -52.935 54.057
686517 95.997 -52.957 54.029
686518 96.011 -52.979 54.002
686519 96.026 -53.002 53.973
686520 96.040 -53.024 53.945
686521 96.054 -53.047 53.916
686522 96.068 -53.068 53.887
686523 96.077 -53.076 53.858
686524 96.086 -53.084 53.829
686525 96.094 -53.092 53.800
686526 96.103 -53.100 53.770
686527 96.112 -53.107 53.741
686528 96.121 -53.116 53.712
686529 96.130 -53.123 53.684
686530 96.138 -53.132 53.654
686561 96.739 -53.252 52.990
686562 96.761 -53.258 52.976
686563 96.783 -53.264 52.962
686564 96.805 -53.270 52.946
686565 96.827 -53.277 52.928
686566 96.849 -53.283 52.910
686578 97.184 -53.370 52.638
686579 97.217 -53.378 52.612
686580 97.250 -53.386 52.586
686581 97.284 -53.395 52.557
686582 97.317 -53.403 52.528
686583 97.350 -53.411 52.498
686584 97.383 -53.420 52.467
686585 97.416 -53.428 52.436
686586 97.449 -53.437 52.404
686587 97.491 -53.443 52.372
686588 97.542 -53.442 52.337
686589 97.540 -53.456 52.303
686590 97.552 -53.467 52.270
686591 97.574 -53.474 52.236
686592 97.595 -53.482 52.204
686593 97.618 -53.490 52.170
686594 97.649 -53.499 52.136
686595 97.680 -53.509 52.103
686596 97.711 -53.518 52.071
686597 97.742 -53.527 52.040
686598 97.773 -53.537 52.010
686599 97.803 -53.546 51.980
686600 97.835 -53.556 51.948
686601 97.865 -53.565 51.920
686602 97.877 -53.562 51.889
686603 97.892 -53.558 51.860
686604 97.910 -53.553 51.833
686605 97.926 -53.549 51.805
686606 97.943 -53.543 51.776
686607 97.956 -53.551 51.747
686608 97.969 -53.562 51.718
686609 97.981 -53.571 51.689
686610 97.993 -53.582 51.659
686611 98.016 -53.596 51.629
686612 98.041 -53.611 51.599
686613 98.067 -53.626 51.568
686614 98.092 -53.640 51.536
686615 98.118 -53.656 51.504
686616 98.144 -53.671 51.470
686617 98.169 -53.685 51.435
686618 98.195 -53.700 51.400
686619 98.216 -53.711 51.366
686620 98.234 -53.718 51.332
686621 98.252 -53.724 51.296
686622 98.283 -53.733 51.263
686623 98.315 -53.741 51.230
686624 98.335 -53.755 51.196
686625 98.350 -53.769 51.162
686626 98.372 -53.786 51.127
686627 98.397 -53.805 51.092
686628 98.423 -53.823 51.055
686629 98.448 -53.842 51.018
686630 98.474 -53.860 50.980
686631 98.499 -53.878 50.941
686632 98.525 -53.897 50.902
686633 98.550 -53.916 50.862
686634 98.576 -53.934 50.823
686635 98.601 -53.953 50.787
686636 98.627 -53.971 50.750
687198 95.666 -52.499 54.461
687199 95.693 -52.518 54.444
687200 95.720 -52.538 54.426
687201 95.747 -52.558 54.408
687202 95.773 -52.578 54.390
687203 95.795 -52.599 54.371
687204 95.808 -52.621 54.352
687205 95.822 -52.644 54.332
687206 95.836 -52.666 54.311
687207 95.850 -52.687 54.291
687208 95.865 -52.709 54.268
687209 95.878 -52.732 54.245
687210 95.892 -52.754 54.222
687211 95.906 -52.776 54.197
687212 95.920 -52.799 54.173
687213 95.935 -52.821 54.148
687214 95.949 -52.844 54.121
687215 95.963 -52.866 54.095
687216 95.977 -52.888 54.069
687217 95.991 -52.911 54.042
687218 96.006 -52.933 54.015
687219 96.020 -52.956 53.987
687220 96.034 -52.978 53.958
687221 96.048 -53.001 53.930
687222 96.063 -53.023 53.901
687223 96.077 -53.045 53.872
687224 96.088 -53.059 53.843
687225 96.097 -53.067 53.814
687226 96.106 -53.075 53.784
687227 96.114 -53.082 53.754
687264 96.793 -53.249 52.951
687265 96.814 -53.255 52.936
687281 97.249 -53.369 52.587
687282 97.283 -53.377 52.560
687283 97.316 -53.386 52.530
687284 97.349 -53.394 52.500
687285 97.382 -53.402 52.470
687286 97.415 -53.411 52.440
687287 97.448 -53.420 52.407
687288 97.481 -53.428 52.375
687289 97.549 -53.424 52.342
687290 97.564 -53.434 52.308
687291 97.562 -53.449 52.275
687292 97.579 -53.457 52.242
687293 97.600 -53.465 52.209
687294 97.621 -53.473 52.176
687295 97.649 -53.481 52.143
687296 97.679 -53.491 52.110
687297 97.711 -53.500 52.078
687298 97.741 -53.509 52.047
687299 97.772 -53.519 52.017
687300 97.803 -53.528 51.987
687301 97.834 -53.537 51.955
687302 97.865 -53.547 51.926
687303 97.882 -53.547 51.897
687304 97.893 -53.544 51.869
687305 97.910 -53.539 51.842
687306 97.927 -53.534 51.813
687307 97.944 -53.530 51.785
687308 97.959 -53.531 51.757
687309 97.971 -53.541 51.728
687310 97.983 -53.552 51.700
687311 97.996 -53.562 51.671
687312 98.018 -53.576 51.642
687313 98.043 -53.591 51.611
687314 98.069 -53.606 51.581
687315 98.094 -53.621 51.550
687316 98.120 -53.636 51.518
687317 98.146 -53.650 51.486
687318 98.171 -53.665 51.452
687319 98.197 -53.681 51.417
687320 98.222 -53.695 51.384
687321 98.242 -53.703 51.350
687322 98.260 -53.710 51.316
687323 98.284 -53.718 51.283
687324 98.316 -53.726 51.250
687325 98.340 -53.738 51.217
687326 98.355 -53.753 51.184
687327 98.374 -53.769 51.149
687328 98.400 -53.787 51.114
687329 98.425 -53.806 51.078
687330 98.450 -53.824 51.041
687331 98.476 -53.843 51.003
687332 98.501 -53.861 50.964
687333 98.526 -53.879 50.926
687334 98.552 -53.898 50.886
687335 98.577 -53.916 50.847
687336 98.603 -53.935 50.810
687337 98.628 -53.954 50.773
687900 95.711 -52.493 54.438
687901 95.739 -52.513 54.418
687902 95.765 -52.533 54.400
687903 95.790 -52.553 54.381
687904 95.803 -52.575 54.361
687905 95.817 -52.597 54.341
687906 95.831 -52.620 54.321
687907 95.845 -52.642 54.300
687908 95.860 -52.664 54.278
687909 95.873 -52.686 54.256
687910 95.887 -52.709 54.232
687911 95.901 -52.731 54.209
687912 95.915 -52.753 54.184
687913 95.930 -52.775 54.159
687914 95.944 -52.797 54.135
687915 95.958 -52.820 54.107
687916 95.972 -52.842 54.082
687917 95.986 -52.865 54.055
687918 96.001 -52.888 54.028
687919 96.015 -52.910 54.001
687920 96.029 -52.932 53.973
687921 96.043 -52.954 53.943
687922 96.057 -52.977 53.915
687923 96.072 -53.000 53.886
687924 96.086 -53.022 53.857
687983 97.282 -53.361 52.562
687984 97.315 -53.369 52.533
687985 97.348 -53.377 52.503
687986 97.381 -53.386 52.473
687987 97.414 -53.394 52.443
687988 97.447 -53.402 52.412
687989 97.480 -53.411 52.380
687990 97.523 -53.415 52.347
687991 97.587 -53.411 52.314
687992 97.585 -53.426 52.281
687993 97.585 -53.440 52.249
687994 97.606 -53.448 52.216
687995 97.628 -53.455 52.183
687996 97.649 -53.463 52.151
687997 97.679 -53.473 52.118
687998 97.710 -53.482 52.086
687999 97.741 -53.491 52.056
688000 97.772 -53.501 52.026
688001 97.803 -53.510 51.996
688002 97.834 -53.519 51.965
688003 97.865 -53.529 51.935
688004 97.887 -53.532 51.907
688005 97.898 -53.529 51.879
688006 97.910 -53.525 51.851
688007 97.927 -53.521 51.824
688008 97.944 -53.516 51.796
688009 97.961 -53.511 51.767
688010 97.973 -53.521 51.740
688011 97.986 -53.532 51.712
688012 97.998 -53.542 51.683
688013 98.020 -53.556 51.654
688014 98.045 -53.571 51.625
688015 98.071 -53.586 51.595
688016 98.096 -53.601 51.563
688017 98.122 -53.615 51.533
688018 98.148 -53.631 51.502
688019 98.173 -53.646 51.469
688020 98.199 -53.660 51.346
688021 98.224 -53.675 51.311
688022 98.249 -53.690 51.277
688023 98.267 -53.696 51.243
688024 98.285 -53.703 51.208
688025 98.317 -53.712 51.173
688026 98.345 -53.722 51.139
688027 98.360 -53.737 51.104
688028 98.376 -53.751 51.068
688029 98.401 -53.769 51.031
688030 98.426 -53.788 50.993
688031 98.452 -53.806 50.954
688032 98.477 -53.825 50.913
688033 98.503 -53.844 50.873
688034 98.528 -53.862 50.833
688035 98.554 -53.881 50.792
688036 98.579 -53.899 50.751
688037 98.604 -53.917 50.710
688038 98.630 -53.936 50.672
688602 95.757 -52.488 54.412
688603 95.784 -52.508 54.392
688604 95.799 -52.530 54.373
688605 95.812 -52.552 54.352
688606 95.826 -52.574 54.331
688607 95.840 -52.597 54.310
688608 95.854 -52.619 54.289
688609 95.869 -52.641 54.266
688610 95.882 -52.663 54.243
688611 95.896 -52.685 54.220
688612 95.910 -52.707 54.196
688613 95.924 -52.729 54.171
688614 95.939 -52.752 54.146
688615 95.952 -52.774 54.122
688616 95.967 -52.797 54.096
688617 95.981 -52.819 54.069
688618 95.995 -52.841 54.043
688619 96.009 -52.863 54.015
688620 96.024 -52.886 53.988
688621 96.038 -52.909 53.960
688689 97.446 -53.385 52.417
688690 97.479 -53.393 52.385
688691 97.513 -53.402 52.353
688692 97.582 -53.397 52.320
688693 97.608 -53.404 52.288
688694 97.606 -53.418 52.256
688695 97.611 -53.430 52.224
688696 97.633 -53.438 52.191
688697 97.654 -53.446 52.159
688698 97.679 -53.455 52.127
688699 97.710 -53.464 52.095
688700 97.741 -53.473 52.065
688701 97.772 -53.483 52.035
688702 97.803 -53.492 52.005
688703 97.833 -53.501 51.975
688704 97.864 -53.511 51.946
688705 97.891 -53.518 51.917
688706 97.902 -53.514 51.889
688707 97.914 -53.511 51.862
688708 97.928 -53.506 51.835
688709 97.944 -53.502 51.807
688710 97.961 -53.497 51.779
688711 97.975 -53.502 51.752
688712 97.988 -53.512 51.724
688713 98.000 -53.522 51.696
688714 98.022 -53.536 51.668
688715 98.048 -53.551 51.639
688716 98.073 -53.566 51.609
688717 98.098 -53.581 51.579
688718 98.124 -53.596 51.462
688719 98.150 -53.610 51.429
688720 98.175 -53.625 51.396
688721 98.201 -53.640 51.362
688722 98.226 -53.655 51.327
688723 98.252 -53.670 51.293
688724 98.275 -53.682 51.259
688725 98.292 -53.688 51.225
688726 98.318 -53.697 51.191
688727 98.349 -53.706 51.157
688728 98.365 -53.720 51.122
688729 98.380 -53.734 51.086
688730 98.403 -53.752 51.050
688731 98.428 -53.771 51.013
688732 98.453 -53.789 50.973
688733 98.479 -53.807 50.934
688734 98.504 -53.826 50.894
688735 98.530 -53.844 50.854
688736 98.555 -53.863 50.813
688737 98.581 -53.882 50.772
688738 98.606 -53.900 50.732
688739 98.632 -53.918 50.693
689304 95.794 -52.484 54.386
689305 95.807 -52.506 54.365
689306 95.821 -52.528 54.344
689307 95.835 -52.551 54.322
689308 95.849 -52.573 54.301
689309 95.864 -52.595 54.279
689310 95.877 -52.617 54.255
689311 95.891 -52.640 54.232
689312 95.905 -52.662 54.208
689313 95.919 -52.684 54.184
689314 95.934 -52.706 54.160
689315 95.947 -52.728 54.135
689316 95.961 -52.750 54.109
689317 95.975 -52.773 54.084
689318 95.989 -52.795 54.058
689319 96.004 -52.818 54.031
689396 97.628 -53.411 52.232
689397 97.639 -53.421 52.201
689398 97.660 -53.429 52.169
689399 97.681 -53.437 52.136
689400 97.710 -53.446 52.105
689401 97.741 -53.455 52.074
689402 97.772 -53.465 52.045
689403 97.802 -53.474 52.016
689404 97.833 -53.483 51.986
689405 97.864 -53.493 51.957
689406 97.895 -53.502 51.928
689407 97.908 -53.499 51.900
689408 97.919 -53.496 51.874
689409 97.930 -53.492 51.847
689410 97.945 -53.488 51.819
689411 97.962 -53.483 51.791
689412 97.978 -53.481 51.765
689413 97.990 -53.492 51.737
689414 98.002 -53.502 51.710
689415 98.024 -53.516 51.682
689416 98.050 -53.531 51.653
689417 98.075 -53.546 51.624
689418 98.100 -53.561 51.506
689419 98.126 -53.575 51.475
689420 98.152 -53.590 51.442
689421 98.177 -53.606 51.410
689422 98.203 -53.620 51.376
689423 98.228 -53.635 51.343
689424 98.253 -53.650 51.309
689425 98.279 -53.665 51.275
689426 98.300 -53.675 51.242
689427 98.319 -53.681 51.208
689428 98.352 -53.690 51.175
689429 98.370 -53.704 51.140
689430 98.385 -53.719 51.105
689431 98.405 -53.734 51.069
689432 98.430 -53.753 51.031
689433 98.455 -53.772 50.992
689434 98.481 -53.790 50.953
689435 98.506 -53.808 50.914
689436 98.531 -53.827 50.874
689437 98.557 -53.845 50.833
689438 98.582 -53.864 50.793
689439 98.608 -53.882 50.753
689440 98.633 -53.900 50.714
690008 95.844 -52.528 54.314
690009 95.859 -52.550 54.292
690010 95.872 -52.572 54.269
690011 95.886 -52.594 54.246
690012 95.900 -52.616 54.222
690013 95.914 -52.638 54.198
690014 95.928 -52.660 54.174
690015 95.942 -52.683 54.150
690016 95.956 -52.705 54.125
690100 97.687 -53.419 52.057
690101 97.710 -53.427 52.027
690102 97.740 -53.437 51.997
690103 97.771 -53.446 51.968
690104 97.802 -53.455 51.940
690105 97.833 -53.465 51.910
690106 97.864 -53.474 51.881
690107 97.895 -53.483 51.853
690108 97.912 -53.484 51.826
690109 97.923 -53.480 51.799
690110 97.935 -53.477 51.772
690111 97.946 -53.474 51.745
690112 97.962 -53.469 51.717
690113 97.979 -53.465 51.690
690114 97.992 -53.472 51.663
690115 98.004 -53.483 51.635
690116 98.026 -53.496 51.607
690117 98.052 -53.511 51.577
690118 98.077 -53.526 51.548
690119 98.102 -53.540 51.518
690120 98.128 -53.556 51.487
690121 98.154 -53.571 51.455
690122 98.179 -53.585 51.423
690123 98.205 -53.600 51.391
690124 98.230 -53.615 51.358
690125 98.255 -53.630 51.325
690126 98.281 -53.645 51.291
690127 98.307 -53.660 51.258
690128 98.325 -53.667 51.225
690129 98.353 -53.675 51.192
690130 98.374 -53.687 51.159
690131 98.390 -53.702 51.125
690132 98.406 -53.717 51.088
690133 98.432 -53.735 51.049
690134 98.457 -53.754 51.011
690135 98.483 -53.772 50.972
690136 98.508 -53.791 50.932
690137 98.533 -53.809 50.892
690138 98.559 -53.828 50.852
690139 98.584 -53.846 50.812
690140 98.609 -53.865 50.773
690141 98.635 -53.883 50.734
690800 97.671 -53.395 52.096
690801 97.692 -53.402 52.065
690802 97.714 -53.410 52.035
690803 97.740 -53.418 52.005
690804 97.771 -53.428 51.976
690805 97.802 -53.437 51.948
690806 97.833 -53.446 51.919
690807 97.864 -53.456 51.890
690808 97.894 -53.465 51.863
690809 97.917 -53.469 51.836
690810 97.928 -53.465 51.809
690811 97.939 -53.462 51.783
690812 97.950 -53.458 51.756
690813 97.962 -53.455 51.728
690814 97.980 -53.451 51.701
690815 97.994 -53.452 51.674
690816 98.007 -53.462 51.647
690817 98.028 -53.476 51.618
690818 98.054 -53.491 51.589
690819 98.079 -53.505 51.560
690820 98.104 -53.521 51.529
690821 98.130 -53.536 51.499
690822 98.156 -53.550 51.468
690823 98.181 -53.565 51.435
690824 98.207 -53.581 51.403
690825 98.232 -53.595 51.371
690826 98.257 -53.610 51.338
690827 98.283 -53.625 51.306
690828 98.309 -53.640 51.274
690829 98.333 -53.653 51.241
690830 98.354 -53.660 51.209
690831 98.379 -53.672 51.176
690832 98.395 -53.686 51.141
690833 98.411 -53.700 51.105
690834 98.433 -53.718 51.065
690835 98.458 -53.736 51.028
690836 98.484 -53.755 50.989
690837 98.509 -53.773 50.949
690838 98.535 -53.791 50.910
690839 98.560 -53.810 50.871
690840 98.586 -53.828 50.831
690841 98.611 -53.847 50.791
690842 98.637 -53.866 50.752
691501 97.693 -53.373 52.103
691502 97.698 -53.385 52.072
691503 97.719 -53.393 52.042
691504 97.740 -53.400 52.013
691505 97.771 -53.410 51.985
691506 97.802 -53.419 51.957
691507 97.832 -53.428 51.928
691508 97.863 -53.438 51.900
691509 97.894 -53.447 51.873
691510 97.922 -53.454 51.846
691511 97.933 -53.451 51.819
691512 97.944 -53.447 51.793
691513 97.955 -53.444 51.767
691514 97.966 -53.440 51.740
691515 97.980 -53.437 51.712
691516 97.997 -53.432 51.685
691517 98.009 -53.443 51.658
691518 98.030 -53.456 51.630
691519 98.056 -53.471 51.601
691520 98.081 -53.486 51.572
691521 98.106 -53.501 51.541
691522 98.132 -53.515 51.512
691523 98.157 -53.531 51.480
691524 98.183 -53.546 51.448
691525 98.209 -53.560 51.416
691526 98.234 -53.575 51.384
691527 98.260 -53.590 51.352
691528 98.285 -53.605 51.320
691529 98.311 -53.620 51.289
691530 98.336 -53.634 51.257
691531 98.358 -53.646 51.226
691532 98.384 -53.655 51.193
691533 98.400 -53.669 51.158
691534 98.415 -53.684 51.121
691535 98.435 -53.700 51.082
691536 98.460 -53.719 51.045
691537 98.486 -53.737 51.006
691538 98.511 -53.756 50.967
691539 98.536 -53.774 50.928
691540 98.562 -53.793 50.889
691541 98.587 -53.811 50.849
691542 98.613 -53.829 50.810
691543 98.638 -53.848 50.771
692202 97.716 -53.350 52.112
692203 97.714 -53.364 52.081
692204 97.725 -53.376 52.052
692205 97.746 -53.383 52.023
692206 97.771 -53.392 51.995
692207 97.801 -53.401 51.967
692208 97.832 -53.410 51.939
692209 97.863 -53.420 51.911
692210 97.894 -53.429 51.884
692211 97.925 -53.438 51.858
692212 97.938 -53.436 51.831
692213 97.949 -53.433 51.805
692214 97.960 -53.429 51.778
692215 97.971 -53.426 51.751
692216 97.983 -53.422 51.725
692217 97.997 -53.418 51.698
692218 98.011 -53.423 51.670
692219 98.032 -53.436 51.642
692220 98.058 -53.451 51.614
692221 98.083 -53.466 51.584
692222 98.108 -53.480 51.555
692223 98.134 -53.496 51.524
692224 98.159 -53.511 51.493
692225 98.185 -53.525 51.462
692226 98.211 -53.540 51.430
692227 98.236 -53.556 51.398
692228 98.262 -53.570 51.365
692229 98.287 -53.585 51.334
692230 98.313 -53.599 51.303
692231 98.338 -53.615 51.273
692232 98.364 -53.630 51.242
692233 98.388 -53.639 51.210
692234 98.405 -53.653 51.175
692235 98.420 -53.668 51.136
692236 98.437 -53.683 51.098
692237 98.462 -53.701 51.061
692238 98.488 -53.719 51.023
692239 98.513 -53.738 50.984
692240 98.538 -53.756 50.945
692241 98.564 -53.775 50.906
692242 98.589 -53.794 50.868
692243 98.615 -53.812 50.829
692244 98.640 -53.830 50.790
692903 97.739 -53.328 52.121
692904 97.737 -53.342 52.090
692905 97.736 -53.357 52.061
692906 97.752 -53.366 52.033
692907 97.773 -53.374 52.006
692908 97.801 -53.383 51.979
692909 97.832 -53.392 51.951
692910 97.863 -53.402 51.923
692911 97.894 -53.411 51.897
692912 97.925 -53.420 51.871
692913 97.942 -53.421 51.845
692914 97.953 -53.418 51.819
692915 97.965 -53.414 51.792
692916 97.976 -53.411 51.765
692917 97.987 -53.407 51.738
692918 97.998 -53.404 51.712
692919 98.013 -53.403 51.685
692920 98.034 -53.416 51.657
692921 98.060 -53.431 51.628
692922 98.085 -53.446 51.598
692923 98.110 -53.461 51.567
692924 98.136 -53.476 51.537
692925 98.161 -53.490 51.506
692926 98.187 -53.505 51.475
692927 98.213 -53.521 51.443
692928 98.238 -53.535 51.411
692929 98.264 -53.550 51.379
692930 98.289 -53.565 51.348
692931 98.315 -53.580 51.318
692932 98.340 -53.595 51.288
692933 98.366 -53.609 51.258
692934 98.391 -53.624 51.226
692935 98.410 -53.637 51.191
692936 98.425 -53.651 51.153
692937 98.440 -53.665 51.115
692938 98.463 -53.684 51.077
692939 98.489 -53.702 51.039
692940 98.514 -53.720 51.002
692941 98.540 -53.739 50.964
692942 98.565 -53.757 50.925
692943 98.591 -53.776 50.886
692944 98.616 -53.794 50.848
692945 98.642 -53.813 50.810
692946 98.667 -53.831 50.772
693604 97.762 -53.306 52.130
693605 97.760 -53.321 52.101
693606 97.758 -53.335 52.072
693607 97.757 -53.349 52.044
693608 97.779 -53.357 52.016
693609 97.801 -53.364 51.991
693610 97.832 -53.374 51.962
693611 97.863 -53.383 51.937
693612 97.893 -53.393 51.911
693613 97.925 -53.402 51.885
693614 97.947 -53.406 51.861
693615 97.958 -53.402 51.836
693616 97.969 -53.399 51.809
693617 97.981 -53.396 51.781
693618 97.992 -53.393 51.754
693619 98.003 -53.389 51.726
693620 98.015 -53.385 51.700
693621 98.036 -53.396 51.672
693622 98.061 -53.411 51.642
693623 98.087 -53.426 51.612
693624 98.112 -53.441 51.581
693625 98.138 -53.455 51.551
693626 98.163 -53.471 51.520
693627 98.189 -53.485 51.488
693628 98.215 -53.500 51.456
693629 98.240 -53.515 51.425
693630 98.266 -53.530 51.393
693631 98.291 -53.545 51.362
693632 98.317 -53.561 51.333
693633 98.344 -53.576 51.303
693634 98.370 -53.592 51.273
693635 98.396 -53.606 51.242
693636 98.414 -53.620 51.207
693637 98.430 -53.635 51.170
693638 98.445 -53.650 51.132
693639 98.465 -53.665 51.094
693640 98.491 -53.684 51.057
693641 98.516 -53.703 51.019
693642 98.541 -53.722 50.982
693643 98.567 -53.740 50.943
693644 98.592 -53.758 50.906
693645 98.618 -53.777 50.868
693646 98.643 -53.795 50.829
693647 98.668 -53.814 50.792
694305 97.785 -53.283 52.140
694306 97.783 -53.298 52.111
694307 97.782 -53.312 52.083
694308 97.780 -53.327 52.056
694309 97.784 -53.339 52.029
694310 97.806 -53.347 52.003
694311 97.831 -53.355 51.976
694312 97.862 -53.365 51.951
694313 97.893 -53.374 51.925
694314 97.924 -53.383 51.901
694315 97.952 -53.391 51.879
694316 97.963 -53.387 51.856
694317 97.974 -53.384 51.828
694318 97.986 -53.380 51.801
694319 97.997 -53.377 51.771
694320 98.008 -53.374 51.743
694321 98.019 -53.371 51.716
694322 98.041 -53.380 51.688
694323 98.068 -53.396 51.658
694324 98.094 -53.411 51.626
694325 98.120 -53.427 51.596
694326 98.146 -53.442 51.565
694327 98.172 -53.457 51.535
694328 98.198 -53.473 51.503
694329 98.224 -53.488 51.470
694330 98.250 -53.503 51.439
694331 98.276 -53.519 51.408
694332 98.302 -53.534 51.377
694333 98.328 -53.550 51.348
694334 98.355 -53.565 51.319
694335 98.386 -53.574 51.289
694336 98.415 -53.584 51.257
694337 98.421 -53.602 51.223
694338 98.435 -53.618 51.186
694339 98.450 -53.633 51.149
694340 98.467 -53.648 51.111
694341 98.493 -53.667 51.074
694342 98.518 -53.685 51.037
694343 98.544 -53.703 50.999
694344 98.569 -53.722 50.962
694345 98.594 -53.740 50.925
694346 98.620 -53.759 50.887
694347 98.645 -53.778 50.849
694348 98.670 -53.797 50.812
694349 98.694 -53.817 50.774
695006 97.770 -53.272 52.151
695007 97.806 -53.276 52.122
695008 97.805 -53.290 52.095
695009 97.803 -53.305 52.068
695010 97.801 -53.319 52.042
695011 97.811 -53.330 52.016
695012 97.832 -53.338 51.990
695013 97.862 -53.347 51.965
695014 97.893 -53.356 51.940
695015 97.924 -53.365 51.918
695016 97.955 -53.375 51.897
695017 97.967 -53.373 51.875
695018 97.979 -53.369 51.848
695019 97.990 -53.366 51.819
695020 98.001 -53.362 51.788
695021 98.013 -53.359 51.759
695022 98.024 -53.355 51.730
695023 98.051 -53.370 51.701
695024 98.078 -53.385 51.671
695025 98.104 -53.400 51.640
695026 98.130 -53.415 51.609
695027 98.156 -53.431 51.578
695028 98.182 -53.446 51.546
695029 98.209 -53.462 51.515
695030 98.235 -53.477 51.483
695031 98.261 -53.493 51.451
695032 98.287 -53.509 51.420
695033 98.313 -53.524 51.390
695034 98.345 -53.532 51.360
695035 98.376 -53.541 51.332
695036 98.408 -53.550 51.302
695037 98.427 -53.562 51.270
695038 98.434 -53.579 51.237
695039 98.439 -53.603 51.202
695040 98.455 -53.617 51.164
695041 98.471 -53.631 51.127
695042 98.494 -53.649 51.091
695043 98.520 -53.668 51.054
695044 98.545 -53.686 51.017
695045 98.571 -53.705 50.980
695046 98.596 -53.723 50.942
695047 98.622 -53.741 50.904
695048 98.647 -53.760 50.867
695049 98.671 -53.780 50.830
695050 98.696 -53.800 50.792
695051 98.720 -53.820 50.755
695052 98.744 -53.840 50.718
695053 98.769 -53.860 50.681
695054 98.795 -53.879 50.646
695055 98.822 -53.888 50.612
695707 97.744 -53.264 52.162
695708 97.828 -53.254 52.134
695709 97.827 -53.268 52.106
695710 97.826 -53.283 52.080
695711 97.824 -53.297 52.054
695712 97.822 -53.311 52.030
695713 97.838 -53.321 52.005
695714 97.862 -53.329 51.980
695715 97.893 -53.338 51.956
695716 97.924 -53.347 51.935
695717 97.955 -53.357 51.916
695718 97.972 -53.358 51.894
695719 97.984 -53.354 51.867
695720 97.995 -53.351 51.837
695721 98.006 -53.347 51.805
695722 98.017 -53.344 51.774
695723 98.033 -53.346 51.744
695724 98.060 -53.360 51.714
695725 98.088 -53.375 51.683
695726 98.115 -53.389 51.652
695727 98.141 -53.405 51.620
695728 98.167 -53.421 51.590
695729 98.193 -53.436 51.558
695730 98.219 -53.451 51.526
695731 98.245 -53.467 51.495
695732 98.272 -53.482 51.463
695733 98.303 -53.491 51.432
695734 98.335 -53.499 51.402
695735 98.366 -53.509 51.372
695736 98.398 -53.518 51.344
695737 98.429 -53.526 51.314
695738 98.438 -53.541 51.283
695739 98.447 -53.556 51.251
695740 98.448 -53.583 51.216
695741 98.460 -53.600 51.179
695742 98.476 -53.615 51.143
695743 98.496 -53.631 51.106
695744 98.521 -53.650 51.070
695745 98.547 -53.668 51.033
695746 98.572 -53.687 50.996
695747 98.597 -53.706 50.959
695748 98.623 -53.724 50.921
695749 98.648 -53.743 50.884
695750 98.672 -53.762 50.846
695751 98.697 -53.783 50.809
695752 98.721 -53.803 50.772
695753 98.746 -53.823 50.735
695754 98.770 -53.843 50.698
695755 98.797 -53.859 50.663
695756 98.824 -53.868 50.628
695757 98.850 -53.877 50.594
695758 98.877 -53.886 50.561
696408 97.718 -53.255 52.172
696409 97.803 -53.245 52.145
696410 97.851 -53.245 52.118
696411 97.849 -53.260 52.092
696412 97.847 -53.274 52.066
696413 97.846 -53.289 52.043
696414 97.844 -53.303 52.018
696415 97.865 -53.311 51.994
696416 97.892 -53.320 51.972
696417 97.924 -53.329 51.952
696418 97.954 -53.339 51.934
696419 97.977 -53.342 51.912
696420 97.989 -53.339 51.884
696421 98.000 -53.336 51.853
696422 98.011 -53.333 51.821
696423 98.022 -53.329 51.790
696424 98.042 -53.336 51.758
696425 98.070 -53.350 51.725
696426 98.097 -53.365 51.694
696427 98.124 -53.380 51.662
696428 98.152 -53.394 51.631
696429 98.178 -53.410 51.601
696430 98.204 -53.425 51.570
696431 98.230 -53.440 51.538
696432 98.262 -53.449 51.506
696433 98.293 -53.458 51.474
696434 98.324 -53.467 51.443
696435 98.356 -53.476 51.413
696436 98.388 -53.485 51.383
696437 98.420 -53.493 51.355
696438 98.440 -53.505 51.325
696439 98.449 -53.520 51.295
696440 98.459 -53.535 51.264
696441 98.461 -53.559 51.230
696442 98.464 -53.584 51.194
696443 98.480 -53.599 51.157
696444 98.498 -53.614 51.121
696445 98.523 -53.632 51.085
696446 98.549 -53.651 51.049
696447 98.574 -53.669 51.012
696448 98.599 -53.688 50.975
696449 98.625 -53.706 50.938
696450 98.649 -53.726 50.900
696451 98.674 -53.746 50.863
696452 98.699 -53.766 50.825
696453 98.723 -53.786 50.788
696454 98.747 -53.806 50.752
696455 98.772 -53.826 50.715
696456 98.799 -53.838 50.679
696457 98.826 -53.847 50.644
696458 98.852 -53.856 50.609
696459 98.878 -53.866 50.577
696460 98.905 -53.875 50.545
697109 97.700 -53.244 52.184
697110 97.777 -53.237 52.156
697111 97.861 -53.227 52.130
697112 97.872 -53.238 52.104
697113 97.870 -53.252 52.079
697114 97.869 -53.267 52.057
697115 97.867 -53.281 52.033
697116 97.871 -53.294 52.012
697117 97.892 -53.302 51.992
697118 97.923 -53.311 51.973
697119 97.954 -53.321 51.954
697120 97.982 -53.327 51.932
697121 97.993 -53.324 51.903
697122 98.004 -53.321 51.871
697123 98.015 -53.317 51.837
697124 98.027 -53.314 51.804
697125 98.052 -53.326 51.771
697126 98.079 -53.340 51.737
697127 98.106 -53.355 51.705
697128 98.134 -53.370 51.673
697129 98.161 -53.384 51.642
697130 98.188 -53.399 51.612
697131 98.218 -53.409 51.581
697132 98.249 -53.418 51.549
697133 98.281 -53.427 51.516
697134 98.311 -53.437 51.485
697135 98.343 -53.446 51.454
697136 98.374 -53.455 51.425
697137 98.405 -53.465 51.395
697138 98.436 -53.474 51.366
697139 98.452 -53.484 51.336
697140 98.461 -53.499 51.305
697141 98.471 -53.514 51.274
697142 98.474 -53.536 51.242
697143 98.475 -53.563 51.207
697144 98.485 -53.582 51.171
697145 98.501 -53.596 51.135
697146 98.525 -53.615 51.099
697147 98.550 -53.634 51.064
697148 98.576 -53.652 51.028
697149 98.601 -53.670 50.991
697150 98.627 -53.689 50.955
697151 98.651 -53.709 50.917
697152 98.675 -53.728 50.880
697153 98.700 -53.749 50.842
697154 98.724 -53.769 50.805
697155 98.748 -53.789 50.769
697156 98.774 -53.808 50.732
697157 98.801 -53.818 50.696
697158 98.828 -53.827 50.660
697159 98.854 -53.836 50.626
697160 98.880 -53.845 50.592
697161 98.907 -53.854 50.559
697162 98.933 -53.863 50.528
697810 97.699 -53.227 52.196
697811 97.750 -53.228 52.168
697812 97.835 -53.219 52.141
697813 97.895 -53.216 52.116
697814 97.893 -53.230 52.093
697815 97.891 -53.245 52.071
697816 97.890 -53.259 52.048
697817 97.888 -53.273 52.030
697818 97.897 -53.284 52.014
697819 97.923 -53.292 51.996
697820 97.954 -53.302 51.976
697821 97.985 -53.311 51.953
697822 97.998 -53.309 51.922
697823 98.009 -53.306 51.888
697824 98.020 -53.302 51.853
697825 98.033 -53.301 51.818
697826 98.061 -53.316 51.783
697827 98.088 -53.330 51.749
697828 98.115 -53.346 51.716
697829 98.143 -53.360 51.684
697830 98.170 -53.374 51.653
697831 98.197 -53.386 51.622
697832 98.225 -53.396 51.591
697833 98.256 -53.406 51.559
697834 98.287 -53.415 51.527
697835 98.318 -53.424 51.495
697836 98.349 -53.433 51.465
697837 98.380 -53.443 51.436
697838 98.412 -53.452 51.405
697839 98.442 -53.462 51.376
697840 98.461 -53.465 51.346
697841 98.473 -53.478 51.315
697842 98.482 -53.493 51.285
697843 98.487 -53.513 51.254
697844 98.488 -53.540 51.221
697845 98.490 -53.566 51.185
697846 98.505 -53.581 51.149
697847 98.526 -53.597 51.115
697848 98.552 -53.615 51.080
697849 98.577 -53.634 51.044
697850 98.603 -53.653 51.008
697851 98.628 -53.672 50.971
697852 98.652 -53.692 50.934
697853 98.676 -53.712 50.897
697854 98.701 -53.731 50.860
697855 98.725 -53.752 50.823
697856 98.750 -53.772 50.786
697857 98.777 -53.788 50.749
697858 98.803 -53.797 50.713
697859 98.830 -53.806 50.677
697860 98.856 -53.816 50.642
697861 98.882 -53.825 50.608
697862 98.909 -53.834 50.575
697863 98.935 -53.843 50.544
697864 98.961 -53.852 50.513
697865 98.988 -53.862 50.482
698511 97.698 -53.209 52.208
698512 97.731 -53.217 52.181
698513 97.809 -53.210 52.156
698514 97.893 -53.201 52.131
698515 97.916 -53.208 52.108
698516 97.915 -53.222 52.087
698517 97.913 -53.236 52.067
698518 97.911 -53.251 52.052
698519 97.910 -53.265 52.038
698520 97.924 -53.275 52.020
698521 97.954 -53.284 51.999
698522 97.985 -53.293 51.975
698523 98.003 -53.295 51.941
698524 98.014 -53.291 51.905
698525 98.025 -53.288 51.869
698526 98.042 -53.291 51.831
698527 98.070 -53.306 51.795
698528 98.097 -53.321 51.760
698529 98.125 -53.336 51.727
698530 98.152 -53.350 51.694
698531 98.178 -53.361 51.663
698532 98.205 -53.373 51.631
698533 98.232 -53.384 51.600
698534 98.263 -53.393 51.569
698535 98.294 -53.403 51.537
698536 98.325 -53.412 51.506
698537 98.356 -53.421 51.476
698538 98.387 -53.430 51.446
698539 98.418 -53.440 51.415
698540 98.446 -53.448 51.386
698541 98.465 -53.452 51.356
698542 98.484 -53.457 51.326
698543 98.493 -53.471 51.297
698544 98.500 -53.490 51.266
698545 98.501 -53.517 51.234
698546 98.502 -53.543 51.199
698547 98.510 -53.564 51.164
698548 98.528 -53.580 51.130
698549 98.554 -53.598 51.095
698550 98.579 -53.617 51.060
698551 98.604 -53.635 51.024
698552 98.629 -53.655 50.987
698553 98.654 -53.675 50.950
698554 98.678 -53.694 50.914
698555 98.703 -53.715 50.877
698556 98.727 -53.734 50.840
698557 98.751 -53.755 50.803
698558 98.778 -53.768 50.766
698559 98.805 -53.777 50.730
698560 98.831 -53.786 50.694
698561 98.858 -53.795 50.659
698562 98.884 -53.804 50.624
698563 98.911 -53.813 50.591
698564 98.937 -53.822 50.559
698565 98.963 -53.832 50.527
698566 98.990 -53.841 50.495
699212 97.697 -53.192 52.221
699213 97.730 -53.200 52.194
699214 97.783 -53.201 52.169
699215 97.867 -53.192 52.146
699216 97.939 -53.186 52.123
699217 97.938 -53.200 52.103
699218 97.936 -53.214 52.085
699219 97.934 -53.229 52.072
699220 97.933 -53.243 52.059
699221 97.931 -53.258 52.042
699222 97.953 -53.266 52.020
699223 97.984 -53.275 51.993
699224 98.007 -53.280 51.957
699225 98.018 -53.276 51.920
699226 98.030 -53.273 51.881
699227 98.052 -53.281 51.844
699228 98.079 -53.296 51.807
699229 98.106 -53.311 51.771
699230 98.134 -53.326 51.737
699231 98.160 -53.337 51.705
699232 98.187 -53.349 51.672
699233 98.213 -53.360 51.641
699234 98.240 -53.371 51.610
699235 98.270 -53.381 51.579
699236 98.301 -53.390 51.547
699237 98.332 -53.399 51.516
699238 98.363 -53.409 51.486
699239 98.394 -53.418 51.457
699240 98.425 -53.427 51.427
699241 98.449 -53.434 51.397
699242 98.469 -53.439 51.367
699243 98.489 -53.443 51.338
699244 98.505 -53.450 51.309
699245 98.513 -53.467 51.278
699246 98.514 -53.493 51.247
699247 98.515 -53.521 51.214
699248 98.516 -53.547 51.180
699249 98.530 -53.562 51.145
699250 98.555 -53.581 51.110
699251 98.581 -53.599 51.076
699252 98.606 -53.618 51.040
699253 98.631 -53.637 51.005
699254 98.655 -53.658 50.967
699255 98.679 -53.678 50.931
699256 98.704 -53.697 50.894
699257 98.728 -53.718 50.857
699258 98.753 -53.737 50.820
699259 98.780 -53.747 50.783
699260 98.807 -53.756 50.747
699261 98.833 -53.766 50.712
699262 98.860 -53.775 50.676
699263 98.886 -53.784 50.641
699264 98.913 -53.793 50.607
699265 98.939 -53.802 50.573
699266 98.966 -53.812 50.541
699267 98.992 -53.821 50.509
699913 97.696 -53.174 52.235
699914 97.729 -53.183 52.208
699915 97.762 -53.191 52.183
699916 97.842 -53.183 52.161
699917 97.926 -53.173 52.139
699918 97.960 -53.178 52.120
699919 97.959 -53.192 52.102
699920 97.957 -53.207 52.090
699921 97.955 -53.221 52.077
699922 97.954 -53.236 52.060
699923 97.957 -53.248 52.038
699924 97.984 -53.257 52.009
699925 98.012 -53.264 51.972
699926 98.023 -53.261 51.933
699927 98.034 -53.258 51.893
699928 98.061 -53.271 51.854
699929 98.088 -53.286 51.817
699930 98.116 -53.301 51.781
699931 98.142 -53.312 51.747
699932 98.169 -53.324 51.714
699933 98.196 -53.335 51.681
699934 98.222 -53.346 51.650
699935 98.248 -53.358 51.619
699936 98.276 -53.369 51.588
699937 98.307 -53.378 51.557
699938 98.339 -53.387 51.526
699939 98.369 -53.396 51.496
699940 98.401 -53.406 51.467
699941 98.432 -53.415 51.438
699942 98.453 -53.421 51.408
699943 98.473 -53.425 51.378
699944 98.492 -53.430 51.349
699945 98.512 -53.434 51.319
699946 98.525 -53.444 51.290
699947 98.527 -53.471 51.259
699948 98.528 -53.497 51.226
699949 98.529 -53.524 51.192
699950 98.535 -53.546 51.158
699951 98.557 -53.563 51.124
699952 98.582 -53.581 51.090
699953 98.607 -53.600 51.055
699954 98.632 -53.621 51.019
699955 98.656 -53.640 50.983
699956 98.680 -53.661 50.946
699957 98.705 -53.681 50.910
699958 98.729 -53.700 50.873
699959 98.756 -53.718 50.836
699960 98.782 -53.727 50.799
699961 98.808 -53.736 50.762
699962 98.835 -53.745 50.727
699963 98.861 -53.754 50.691
699964 98.888 -53.763 50.656
699965 98.914 -53.772 50.621
699966 98.941 -53.781 50.587
699967 98.967 -53.791 50.553
699968 98.994 -53.800 50.521
699969 99.020 -53.809 50.488
700614 97.695 -53.157 52.247
700615 97.728 -53.165 52.222
700616 97.761 -53.173 52.198
700617 97.815 -53.175 52.176
700618 97.900 -53.165 52.155
700619 97.984 -53.156 52.139
700620 97.982 -53.170 52.119
700621 97.980 -53.184 52.103
700622 97.979 -53.198 52.089
700623 97.977 -53.213 52.071
700624 97.975 -53.227 52.048
700625 97.984 -53.239 52.020
700626 98.015 -53.248 51.984
700627 98.028 -53.246 51.943
700628 98.043 -53.247 51.903
700629 98.070 -53.261 51.863
700630 98.098 -53.277 51.825
700631 98.124 -53.288 51.790
700632 98.151 -53.299 51.756
700633 98.177 -53.311 51.723
700634 98.204 -53.322 51.690
700635 98.230 -53.333 51.659
700636 98.256 -53.345 51.628
700637 98.283 -53.356 51.597
700638 98.314 -53.366 51.567
700639 98.345 -53.375 51.536
700640 98.376 -53.384 51.505
700641 98.407 -53.393 51.476
700642 98.438 -53.402 51.448
700643 98.457 -53.407 51.419
700644 98.477 -53.411 51.389
700645 98.496 -53.416 51.359
700646 98.516 -53.421 51.330
700647 98.535 -53.425 51.301
700648 98.540 -53.447 51.271
700649 98.541 -53.474 51.238
700650 98.541 -53.501 51.204
700651 98.543 -53.527 51.171
700652 98.559 -53.546 51.138
700653 98.584 -53.564 51.103
700654 98.609 -53.584 51.068
700655 98.633 -53.603 51.033
700656 98.658 -53.624 50.997
700657 98.682 -53.643 50.961
700658 98.707 -53.663 50.925
700659 98.731 -53.684 50.888
700660 98.758 -53.697 50.851
700661 98.784 -53.706 50.814
700662 98.810 -53.715 50.776
700663 98.837 -53.725 50.741
700664 98.863 -53.734 50.705
700665 98.890 -53.743 50.669
700666 98.916 -53.752 50.633
700667 98.943 -53.761 50.599
700668 98.969 -53.771 50.565
700669 98.996 -53.780 50.531
700670 99.022 -53.789 50.497
701315 97.694 -53.140 52.260
701316 97.727 -53.148 52.235
701317 97.760 -53.157 52.212
701318 97.794 -53.165 52.190
701319 97.874 -53.157 52.173
701320 97.958 -53.147 52.158
701321 98.005 -53.148 52.136
701322 98.003 -53.162 52.117
701323 98.002 -53.176 52.101
701324 98.000 -53.191 52.081
701325 97.998 -53.205 52.055
701326 97.996 -53.220 52.026
701327 98.014 -53.230 51.991
701328 98.033 -53.231 51.951
701329 98.052 -53.237 51.910
701330 98.080 -53.252 51.871
701331 98.106 -53.264 51.833
701332 98.133 -53.275 51.798
701333 98.159 -53.286 51.763
701334 98.185 -53.298 51.731
701335 98.212 -53.309 51.699
701336 98.238 -53.321 51.667
701337 98.265 -53.332 51.637
701338 98.292 -53.343 51.606
701339 98.320 -53.353 51.575
701340 98.352 -53.362 51.544
701341 98.383 -53.372 51.515
701342 98.414 -53.381 51.486
701343 98.442 -53.389 51.457
701344 98.461 -53.393 51.428
701345 98.481 -53.398 51.398
701346 98.500 -53.402 51.369
701347 98.520 -53.407 51.340
701348 98.533 -53.419 51.311
701349 98.542 -53.437 51.281
701350 98.551 -53.455 51.248
701351 98.555 -53.477 51.215
701352 98.556 -53.505 51.183
701353 98.561 -53.528 51.151
701354 98.586 -53.546 51.116
701355 98.610 -53.566 51.080
701356 98.635 -53.587 51.045
701357 98.659 -53.606 51.009
701358 98.684 -53.627 50.973
701359 98.708 -53.646 50.937
701360 98.732 -53.666 50.901
701361 98.760 -53.677 50.864
701362 98.786 -53.686 50.827
701363 98.812 -53.695 50.789
701364 98.839 -53.704 50.753
701365 98.865 -53.713 50.716
701366 98.891 -53.722 50.680
701367 98.918 -53.731 50.644
701368 98.944 -53.741 50.609
701369 98.971 -53.750 50.574
701370 98.997 -53.759 50.540
701371 99.024 -53.769 50.505
702016 97.693 -53.123 52.272
702017 97.726 -53.131 52.247
702018 97.759 -53.139 52.224
702019 97.793 -53.148 52.205
702020 97.848 -53.148 52.191
702021 97.932 -53.139 52.175
702022 98.017 -53.129 52.154
702023 98.026 -53.140 52.132
702024 98.024 -53.154 52.112
702025 98.023 -53.169 52.090
702026 98.021 -53.183 52.061
702027 98.019 -53.198 52.030
702028 98.018 -53.212 51.995
702029 98.037 -53.216 51.957
702030 98.062 -53.227 51.917
702031 98.088 -53.239 51.877
702032 98.114 -53.250 51.841
702033 98.141 -53.261 51.806
702034 98.167 -53.273 51.772
702035 98.194 -53.284 51.740
702036 98.221 -53.295 51.708
702037 98.247 -53.307 51.677
702038 98.274 -53.318 51.646
702039 98.300 -53.330 51.615
702040 98.327 -53.341 51.584
702041 98.358 -53.350 51.554
702042 98.389 -53.359 51.524
702043 98.421 -53.369 51.496
702044 98.446 -53.376 51.467
702045 98.465 -53.380 51.438
702046 98.485 -53.384 51.409
702047 98.504 -53.389 51.380
702048 98.519 -53.399 51.351
702049 98.527 -53.418 51.321
702050 98.536 -53.435 51.291
702051 98.545 -53.452 51.261
702052 98.554 -53.471 51.228
702053 98.563 -53.488 51.196
702054 98.570 -53.509 51.164
702055 98.587 -53.529 51.130
702056 98.612 -53.549 51.093
702057 98.636 -53.569 51.057
702058 98.660 -53.590 51.021
702059 98.685 -53.609 50.985
702060 98.709 -53.629 50.950
702061 98.735 -53.647 50.913
702062 98.761 -53.656 50.876
702063 98.788 -53.665 50.839
702064 98.814 -53.675 50.802
702065 98.841 -53.684 50.766
702066 98.867 -53.693 50.728
702067 98.894 -53.702 50.692
702068 98.920 -53.711 50.655
702069 98.946 -53.721 50.620
702070 98.973 -53.730 50.584
702071 98.999 -53.739 50.549
702072 99.026 -53.748 50.514
702073 99.052 -53.757 50.480
702717 97.692 -53.105 52.284
702718 97.725 -53.114 52.259
702719 97.758 -53.122 52.236
702720 97.792 -53.130 52.217
702721 97.825 -53.139 52.205
702722 97.907 -53.130 52.192
702723 97.991 -53.120 52.173
702724 98.049 -53.118 52.149
702725 98.048 -53.132 52.125
702726 98.046 -53.146 52.098
702727 98.044 -53.161 52.064
702728 98.042 -53.175 52.031
702729 98.040 -53.189 51.996
702730 98.043 -53.203 51.960
702731 98.070 -53.214 51.921
702732 98.096 -53.226 51.884
702733 98.123 -53.237 51.849
702734 98.149 -53.248 51.815
702735 98.176 -53.260 51.782
702736 98.203 -53.271 51.750
702737 98.229 -53.283 51.719
702738 98.255 -53.294 51.688
702739 98.282 -53.305 51.656
702740 98.308 -53.317 51.625
702741 98.335 -53.328 51.594
702742 98.365 -53.338 51.564
702743 98.396 -53.347 51.534
702744 98.427 -53.356 51.506
702745 98.450 -53.362 51.478
702746 98.469 -53.367 51.449
702747 98.489 -53.371 51.419
702748 98.504 -53.380 51.391
702749 98.513 -53.398 51.362
702750 98.522 -53.415 51.333
702751 98.531 -53.433 51.303
702752 98.539 -53.451 51.273
702753 98.549 -53.468 51.242
702754 98.558 -53.486 51.209
702755 98.570 -53.496 51.176
702756 98.588 -53.512 51.143
702757 98.613 -53.532 51.106
702758 98.637 -53.552 51.069
702759 98.662 -53.572 51.033
702760 98.686 -53.593 50.998
702761 98.711 -53.612 50.962
702762 98.736 -53.627 50.926
702763 98.763 -53.636 50.888
702764 98.790 -53.645 50.851
702765 98.816 -53.654 50.814
702766 98.843 -53.663 50.778
702767 98.869 -53.672 50.740
702768 98.896 -53.681 50.703
702769 98.922 -53.690 50.666
702770 98.948 -53.700 50.630
702771 98.975 -53.709 50.594
702772 99.001 -53.719 50.558
702773 99.027 -53.728 50.523
702774 99.054 -53.737 50.487
703418 97.691 -53.088 52.296
703419 97.724 -53.096 52.271
703420 97.757 -53.104 52.248
703421 97.791 -53.113 52.229
703422 97.824 -53.121 52.215
703423 97.880 -53.121 52.203
703424 97.965 -53.111 52.184
703425 98.050 -53.102 52.160
703426 98.071 -53.110 52.133
703427 98.069 -53.124 52.103
703428 98.067 -53.139 52.067
703429 98.066 -53.153 52.033
703430 98.064 -53.167 51.998
703431 98.069 -53.182 51.963
703432 98.088 -53.197 51.927
703433 98.107 -53.211 51.892
703434 98.131 -53.223 51.860
703435 98.158 -53.235 51.827
703436 98.184 -53.246 51.795
703437 98.211 -53.258 51.762
703438 98.237 -53.269 51.731
703439 98.264 -53.280 51.699
703440 98.290 -53.292 51.668
703441 98.317 -53.304 51.637
703442 98.343 -53.315 51.606
703443 98.372 -53.325 51.576
703444 98.403 -53.335 51.546
703445 98.434 -53.344 51.517
703446 98.454 -53.349 51.489
703447 98.474 -53.353 51.461
703448 98.490 -53.361 51.432
703449 98.499 -53.379 51.403
703450 98.508 -53.396 51.375
703451 98.516 -53.414 51.346
703452 98.525 -53.431 51.316
703453 98.534 -53.448 51.286
703454 98.546 -53.455 51.255
703455 98.559 -53.467 51.222
703456 98.577 -53.483 51.189
703457 98.594 -53.498 51.154
703458 98.614 -53.515 51.118
703459 98.639 -53.535 51.081
703460 98.663 -53.556 51.046
703461 98.688 -53.575 51.010
703462 98.712 -53.595 50.973
703463 98.738 -53.606 50.937
703464 98.765 -53.615 50.900
703465 98.791 -53.625 50.863
703466 98.818 -53.634 50.825
703467 98.844 -53.643 50.789
703468 98.871 -53.652 50.752
703469 98.897 -53.661 50.714
703470 98.924 -53.670 50.677
703471 98.950 -53.680 50.640
703472 98.977 -53.689 50.603
703473 99.003 -53.698 50.567
703474 99.029 -53.707 50.530
703475 99.056 -53.716 50.495
704119 97.700 -53.073 52.307
704120 97.723 -53.079 52.282
704121 97.756 -53.087 52.259
704122 97.790 -53.095 52.239
704123 97.823 -53.104 52.224
704124 97.856 -53.112 52.210
704125 97.939 -53.103 52.190
704126 98.023 -53.093 52.166
704127 98.093 -53.088 52.138
704128 98.092 -53.102 52.106
704129 98.090 -53.117 52.070
704130 98.088 -53.131 52.036
704131 98.087 -53.145 52.002
704132 98.097 -53.160 51.968
704133 98.116 -53.174 51.934
704134 98.135 -53.189 51.902
704135 98.154 -53.204 51.871
704136 98.173 -53.219 51.841
704137 98.193 -53.233 51.810
704138 98.219 -53.245 51.777
704139 98.246 -53.256 51.745
704140 98.272 -53.267 51.713
704141 98.299 -53.279 51.681
704142 98.325 -53.290 51.650
704143 98.352 -53.302 51.619
704144 98.378 -53.313 51.589
704145 98.410 -53.322 51.558
704146 98.438 -53.330 51.529
704147 98.458 -53.335 51.501
704148 98.476 -53.342 51.473
704149 98.485 -53.359 51.444
704150 98.493 -53.377 51.416
704151 98.502 -53.395 51.388
704152 98.512 -53.409 51.360
704153 98.523 -53.418 51.331
704154 98.534 -53.426 51.299
704155 98.548 -53.437 51.267
704156 98.566 -53.453 51.234
704157 98.583 -53.468 51.199
704158 98.600 -53.484 51.164
704159 98.618 -53.499 51.129
704160 98.640 -53.518 51.093
704161 98.664 -53.538 51.058
704162 98.689 -53.559 51.022
704163 98.714 -53.577 50.985
704164 98.740 -53.586 50.949
704165 98.767 -53.595 50.912
704166 98.793 -53.604 50.874
704167 98.819 -53.613 50.837
704168 98.846 -53.622 50.800
704169 98.873 -53.631 50.762
704170 98.899 -53.640 50.725
704171 98.926 -53.650 50.687
704172 98.952 -53.659 50.650
704173 98.979 -53.668 50.613
704174 99.005 -53.678 50.576
704175 99.031 -53.687 50.539
704176 99.058 -53.696 50.502
704820 97.709 -53.057 52.318
704821 97.731 -53.064 52.293
704822 97.755 -53.070 52.268
704823 97.789 -53.078 52.247
704824 97.822 -53.086 52.229
704825 97.855 -53.095 52.212
704826 97.913 -53.095 52.191
704827 97.997 -53.085 52.165
704828 98.082 -53.075 52.137
704829 98.115 -53.080 52.106
704830 98.113 -53.094 52.074
704831 98.111 -53.108 52.042
704832 98.110 -53.123 52.009
704833 98.125 -53.138 51.977
704834 98.144 -53.152 51.945
704835 98.163 -53.167 51.913
704836 98.182 -53.182 51.884
704837 98.201 -53.197 51.857
704838 98.220 -53.211 51.827
704839 98.239 -53.227 51.794
704840 98.258 -53.241 51.761
704841 98.281 -53.255 51.729
704842 98.307 -53.266 51.697
704843 98.334 -53.277 51.664
704844 98.360 -53.289 51.633
704845 98.386 -53.300 51.603
704846 98.416 -53.310 51.572
704847 98.442 -53.317 51.543
704848 98.461 -53.322 51.515
704849 98.471 -53.339 51.486
704850 98.479 -53.358 51.457
704851 98.489 -53.371 51.429
704852 98.500 -53.379 51.402
704853 98.511 -53.387 51.373
704854 98.522 -53.396 51.344
704855 98.537 -53.408 51.312
704856 98.554 -53.424 51.278
704857 98.572 -53.440 51.244
704858 98.589 -53.455 51.209
704859 98.607 -53.471 51.174
704860 98.624 -53.486 51.139
704861 98.642 -53.502 51.103
704862 98.666 -53.521 51.069
704863 98.690 -53.541 51.034
704864 98.716 -53.556 50.997
704865 98.742 -53.565 50.960
704866 98.769 -53.574 50.922
704867 98.795 -53.584 50.884
704868 98.821 -53.593 50.847
704869 98.848 -53.602 50.810
704870 98.874 -53.611 50.772
704871 98.901 -53.620 50.734
704872 98.927 -53.630 50.697
704873 98.954 -53.639 50.659
704874 98.980 -53.648 50.621
704875 99.007 -53.657 50.584
704876 99.033 -53.666 50.546
704877 99.060 -53.675 50.508
705521 97.719 -53.042 52.328
705522 97.740 -53.048 52.302
705523 97.762 -53.054 52.278
705524 97.788 -53.060 52.255
705525 97.821 -53.069 52.234
705526 97.854 -53.077 52.214
705527 97.887 -53.086 52.191
705528 97.971 -53.076 52.165
705529 98.056 -53.067 52.138
705530 98.138 -53.057 52.109
705531 98.136 -53.072 52.078
705532 98.135 -53.086 52.049
705533 98.134 -53.101 52.017
705534 98.153 -53.116 51.986
705535 98.172 -53.130 51.955
705536 98.191 -53.145 51.926
705537 98.210 -53.160 51.899
705538 98.229 -53.174 51.874
705539 98.248 -53.189 51.845
705540 98.268 -53.204 51.813
705541 98.287 -53.219 51.779
705542 98.305 -53.233 51.746
705543 98.324 -53.248 51.713
705544 98.344 -53.263 51.681
705545 98.368 -53.275 51.648
705546 98.395 -53.286 51.617
705547 98.423 -53.298 51.587
705548 98.446 -53.304 51.557
705549 98.456 -53.321 51.527
705550 98.466 -53.333 51.498
705551 98.477 -53.341 51.470
705552 98.488 -53.349 51.443
705553 98.499 -53.357 51.415
705554 98.510 -53.365 51.386
705555 98.525 -53.379 51.354
705556 98.543 -53.394 51.322
705557 98.561 -53.410 51.288
705558 98.578 -53.425 51.253
705559 98.595 -53.441 51.218
705560 98.613 -53.456 51.182
705561 98.631 -53.472 51.147
705562 98.648 -53.487 51.113
705563 98.667 -53.504 51.079
705564 98.692 -53.524 51.045
705565 98.718 -53.536 51.008
705566 98.744 -53.545 50.970
705567 98.771 -53.554 50.931
705568 98.797 -53.563 50.893
705569 98.823 -53.572 50.855
705570 98.850 -53.581 50.818
705571 98.876 -53.590 50.780
705572 98.903 -53.599 50.742
705573 98.929 -53.609 50.705
705574 98.955 -53.618 50.667
705575 98.982 -53.628 50.628
705576 99.009 -53.637 50.591
705577 99.035 -53.646 50.552
705578 99.062 -53.655 50.513
706222 97.728 -53.026 52.337
706223 97.750 -53.032 52.311
706224 97.772 -53.039 52.288
706225 97.794 -53.045 52.265
706226 97.820 -53.052 52.240
706227 97.853 -53.060 52.217
706228 97.886 -53.069 52.193
706229 97.945 -53.068 52.167
706230 98.030 -53.058 52.141
706231 98.114 -53.048 52.113
706232 98.159 -53.050 52.084
706233 98.157 -53.064 52.057
706234 98.162 -53.079 52.027
706235 98.181 -53.093 51.996
706236 98.200 -53.108 51.967
706237 98.219 -53.123 51.939
706238 98.238 -53.138 51.913
706239 98.257 -53.152 51.889
706240 98.276 -53.167 51.862
706241 98.295 -53.182 51.831
706242 98.314 -53.197 51.797
706243 98.334 -53.211 51.763
706244 98.353 -53.227 51.730
706245 98.372 -53.241 51.697
706246 98.391 -53.256 51.664
706247 98.410 -53.270 51.633
706248 98.430 -53.285 51.603
706249 98.443 -53.294 51.572
706250 98.454 -53.302 51.540
706251 98.465 -53.311 51.510
706252 98.476 -53.319 51.481
706253 98.487 -53.327 51.454
706254 98.498 -53.335 51.425
706255 98.514 -53.349 51.395
706256 98.532 -53.365 51.362
706257 98.550 -53.380 51.329
706258 98.567 -53.396 51.295
706259 98.584 -53.411 51.260
706260 98.602 -53.427 51.224
706261 98.619 -53.443 51.189
706262 98.637 -53.458 51.154
706263 98.654 -53.474 51.121
706264 98.672 -53.489 51.087
706265 98.693 -53.506 51.052
706266 98.719 -53.515 51.015
706267 98.746 -53.524 50.977
706268 98.772 -53.534 50.938
706269 98.799 -53.543 50.900
706270 98.826 -53.552 50.861
706271 98.852 -53.561 50.824
706272 98.878 -53.570 50.786
706273 98.905 -53.580 50.748
706274 98.931 -53.589 50.710
706275 98.957 -53.598 50.671
706276 98.984 -53.607 50.633
706277 99.010 -53.616 50.594
706278 99.037 -53.625 50.555
706279 99.063 -53.634 50.517
706923 97.737 -53.011 52.346
706924 97.759 -53.017 52.322
706925 97.782 -53.023 52.297
706926 97.803 -53.029 52.273
706927 97.825 -53.035 52.248
706928 97.852 -53.043 52.224
706929 97.885 -53.051 52.200
706930 97.919 -53.060 52.174
706931 98.004 -53.050 52.147
706932 98.088 -53.040 52.120
706933 98.173 -53.030 52.094
706934 98.180 -53.042 52.067
706935 98.190 -53.057 52.037
706936 98.209 -53.071 52.007
706937 98.228 -53.086 51.979
706938 98.247 -53.101 51.952
706939 98.266 -53.116 51.926
706940 98.285 -53.130 51.903
706941 98.304 -53.145 51.878
706942 98.323 -53.160 51.849
706943 98.343 -53.175 51.815
706944 98.361 -53.189 51.780
706945 98.378 -53.206 51.745
706946 98.396 -53.222 51.712
706947 98.413 -53.238 51.681
706948 98.430 -53.254 51.650
706949 98.437 -53.265 51.619
706950 98.442 -53.272 51.587
706951 98.453 -53.280 51.554
706952 98.464 -53.289 51.521
706953 98.476 -53.297 51.491
706954 98.486 -53.305 51.462
706955 98.503 -53.320 51.432
706956 98.521 -53.336 51.402
706957 98.538 -53.351 51.369
706958 98.556 -53.367 51.335
706959 98.573 -53.382 51.300
706960 98.591 -53.398 51.266
706961 98.608 -53.413 51.230
706962 98.626 -53.429 51.195
706963 98.643 -53.444 51.161
706964 98.661 -53.459 51.127
706965 98.678 -53.472 51.093
706966 98.698 -53.484 51.058
706967 98.721 -53.495 51.021
706968 98.748 -53.504 50.983
706969 98.774 -53.513 50.944
706970 98.801 -53.522 50.905
706971 98.827 -53.531 50.867
706972 98.854 -53.540 50.830
706973 98.880 -53.549 50.792
706974 98.907 -53.559 50.754
706975 98.933 -53.568 50.715
706976 98.959 -53.577 50.676
706977 98.986 -53.587 50.638
706978 99.012 -53.596 50.599
706979 99.039 -53.605 50.560
706980 99.065 -53.614 50.521
706981 99.092 -53.623 50.483
707624 97.747 -52.996 52.356
707625 97.769 -53.002 52.332
707626 97.791 -53.008 52.307
707627 97.813 -53.014 52.282
707628 97.835 -53.020 52.258
707629 97.857 -53.026 52.234
707630 97.884 -53.034 52.209
707631 97.918 -53.042 52.184
707632 97.978 -53.041 52.158
707633 98.062 -53.032 52.132
707634 98.147 -53.022 52.106
707635 98.204 -53.020 52.079
707636 98.218 -53.035 52.050
707637 98.237 -53.049 52.019
707638 98.256 -53.064 51.991
707639 98.275 -53.079 51.964
707640 98.292 -53.095 51.939
707641 98.310 -53.111 51.917
707642 98.327 -53.127 51.891
707643 98.344 -53.143 51.862
707644 98.362 -53.159 51.829
707645 98.379 -53.175 51.795
707646 98.396 -53.192 51.761
707647 98.414 -53.208 51.730
707648 98.431 -53.223 51.698
707649 98.448 -53.239 51.666
707650 98.441 -53.244 51.633
707651 98.441 -53.250 51.600
707652 98.452 -53.258 51.567
707653 98.463 -53.267 51.534
707654 98.475 -53.275 51.502
707655 98.492 -53.290 51.471
707656 98.509 -53.306 51.440
707657 98.527 -53.321 51.408
707658 98.545 -53.337 51.373
707659 98.562 -53.352 51.340
707660 98.579 -53.368 51.306
707661 98.597 -53.383 51.271
707662 98.615 -53.396 51.237
707663 98.633 -53.409 51.201
707664 98.651 -53.422 51.168
707665 98.668 -53.435 51.132
707666 98.688 -53.446 51.097
707667 98.708 -53.458 51.062
707668 98.726 -53.474 51.026
707669 98.749 -53.483 50.989
707670 98.776 -53.493 50.949
707671 98.802 -53.502 50.911
707672 98.829 -53.511 50.872
707673 98.855 -53.520 50.834
707674 98.882 -53.529 50.797
707675 98.909 -53.539 50.759
707676 98.935 -53.548 50.720
707677 98.961 -53.557 50.681
707678 98.988 -53.566 50.642
707679 99.014 -53.575 50.603
707680 99.041 -53.584 50.565
707681 99.067 -53.593 50.525
707682 99.093 -53.603 50.486
707683 99.120 -53.612 50.446
708325 97.756 -52.981 52.365
708326 97.779 -52.987 52.341
708327 97.800 -52.993 52.316
708328 97.822 -52.999 52.291
708329 97.844 -53.005 52.268
708330 97.866 -53.011 52.244
708331 97.888 -53.017 52.219
708332 97.917 -53.025 52.194
708333 97.951 -53.032 52.170
708334 98.036 -53.023 52.146
708335 98.121 -53.013 52.120
708336 98.205 -53.004 52.093
708337 98.241 -53.016 52.063
708338 98.258 -53.032 52.032
708339 98.276 -53.048 52.003
708340 98.293 -53.064 51.976
708341 98.310 -53.080 51.951
708342 98.327 -53.096 51.928
708343 98.345 -53.113 51.902
708344 98.362 -53.129 51.873
708345 98.379 -53.145 51.841
708346 98.396 -53.161 51.808
708347 98.414 -53.177 51.776
708348 98.431 -53.193 51.745
708349 98.448 -53.209 51.713
708350 98.453 -53.220 51.680
708351 98.444 -53.223 51.647
708352 98.441 -53.228 51.614
708353 98.452 -53.236 51.581
708354 98.463 -53.245 51.546
708355 98.481 -53.261 51.513
708356 98.498 -53.277 51.480
708357 98.516 -53.292 51.448
708358 98.533 -53.308 51.415
708359 98.551 -53.321 51.380
708360 98.569 -53.334 51.346
708361 98.587 -53.347 51.312
708362 98.605 -53.360 51.278
708363 98.623 -53.373 51.243
708364 98.641 -53.386 51.208
708365 98.659 -53.399 51.173
708366 98.676 -53.411 51.137
708367 98.699 -53.417 51.101
708368 98.719 -53.433 51.065
708369 98.736 -53.449 51.029
708370 98.754 -53.462 50.992
708371 98.778 -53.472 50.954
708372 98.804 -53.481 50.915
708373 98.831 -53.490 50.877
708374 98.857 -53.499 50.840
708375 98.884 -53.509 50.801
708376 98.910 -53.518 50.763
708377 98.937 -53.527 50.725
708378 98.963 -53.537 50.687
708379 98.990 -53.546 50.648
708380 99.016 -53.555 50.609
708381 99.043 -53.564 50.569
708382 99.069 -53.573 50.529
708383 99.095 -53.582 50.490
708384 99.122 -53.591 50.449
708385 99.148 -53.600 50.409
709026 97.766 -52.966 52.375
709027 97.788 -52.972 52.351
709028 97.810 -52.978 52.327
709029 97.832 -52.984 52.302
709030 97.854 -52.990 52.278
709031 97.876 -52.996 52.255
709032 97.897 -53.002 52.231
709033 97.920 -53.008 52.207
709034 97.949 -53.016 52.184
709035 98.010 -53.014 52.161
709036 98.094 -53.004 52.136
709037 98.179 -52.998 52.108
709038 98.236 -53.003 52.078
709039 98.268 -53.018 52.046
709040 98.293 -53.034 52.016
709041 98.310 -53.050 51.988
709042 98.327 -53.066 51.962
709043 98.345 -53.082 51.938
709044 98.362 -53.098 51.910
709045 98.379 -53.114 51.881
709046 98.397 -53.130 51.851
709047 98.414 -53.147 51.820
709048 98.432 -53.163 51.790
709049 98.449 -53.179 51.759
709050 98.465 -53.195 51.725
709051 98.456 -53.198 51.692
709052 98.447 -53.202 51.659
709053 98.440 -53.207 51.628
709054 98.452 -53.216 51.594
709055 98.469 -53.232 51.558
709056 98.487 -53.246 51.523
709057 98.505 -53.259 51.489
709058 98.523 -53.272 51.456
709059 98.541 -53.285 51.423
709060 98.559 -53.298 51.388
709061 98.577 -53.311 51.352
709062 98.595 -53.324 51.318
709063 98.613 -53.336 51.284
709064 98.631 -53.349 51.249
709065 98.649 -53.362 51.214
709066 98.667 -53.375 51.178
709067 98.687 -53.384 51.141
709068 98.711 -53.387 51.105
709069 98.729 -53.406 51.069
709070 98.747 -53.425 51.033
709071 98.765 -53.438 50.996
709072 98.782 -53.451 50.958
709073 98.806 -53.461 50.920
709074 98.833 -53.470 50.883
709075 98.859 -53.479 50.845
709076 98.885 -53.489 50.807
709077 98.912 -53.498 50.768
709078 98.938 -53.507 50.730
709079 98.965 -53.516 50.692
709080 98.992 -53.525 50.653
709081 99.018 -53.534 50.614
709082 99.045 -53.543 50.574
709083 99.071 -53.552 50.534
709084 99.097 -53.562 50.494
709085 99.124 -53.571 50.453
709086 99.150 -53.580 50.412
709087 99.177 -53.589 50.372
709727 97.776 -52.951 52.385
709728 97.797 -52.957 52.361
709729 97.819 -52.963 52.337
709730 97.841 -52.969 52.312
709731 97.863 -52.975 52.289
709732 97.885 -52.981 52.267
709733 97.907 -52.987 52.243
709734 97.929 -52.993 52.220
709735 97.951 -52.999 52.198
709736 97.984 -53.006 52.176
709737 98.069 -52.996 52.151
709738 98.153 -52.991 52.122
709739 98.227 -52.990 52.091
709740 98.258 -53.006 52.059
709741 98.289 -53.021 52.029
709742 98.320 -53.036 52.000
709743 98.345 -53.051 51.973
709744 98.363 -53.068 51.947
709745 98.380 -53.084 51.918
709746 98.397 -53.100 51.889
709747 98.415 -53.116 51.859
709748 98.432 -53.132 51.830
709749 98.449 -53.148 51.801
709750 98.466 -53.164 51.768
709751 98.469 -53.173 51.733
709752 98.460 -53.177 51.700
709753 98.450 -53.181 51.668
709754 98.444 -53.188 51.636
709755 98.459 -53.197 51.603
709756 98.478 -53.210 51.568
709757 98.495 -53.223 51.532
709758 98.513 -53.236 51.497
709759 98.531 -53.248 51.464
709760 98.549 -53.261 51.429
709761 98.567 -53.274 51.394
709762 98.585 -53.287 51.359
709763 98.603 -53.300 51.324
709764 98.621 -53.313 51.289
709765 98.639 -53.326 51.255
709766 98.657 -53.339 51.219
709767 98.675 -53.352 51.183
709768 98.699 -53.355 51.147
709769 98.722 -53.361 51.110
709770 98.739 -53.380 51.073
709771 98.758 -53.399 51.037
709772 98.775 -53.414 51.001
709773 98.792 -53.427 50.963
709774 98.810 -53.440 50.926
709775 98.835 -53.449 50.889
709776 98.861 -53.458 50.850
709777 98.887 -53.468 50.813
709778 98.914 -53.477 50.774
709779 98.940 -53.487 50.735
709780 98.967 -53.496 50.697
709781 98.993 -53.505 50.659
709782 99.020 -53.514 50.619
709783 99.046 -53.523 50.579
709784 99.073 -53.532 50.538
709785 99.099 -53.541 50.498
709786 99.126 -53.550 50.457
709787 99.152 -53.559 50.416
709788 99.179 -53.568 50.376
709789 99.205 -53.577 50.335
710428 97.785 -52.935 52.395
710429 97.807 -52.941 52.371
710430 97.829 -52.948 52.348
710431 97.851 -52.954 52.323
710432 97.873 -52.960 52.301
710433 97.894 -52.966 52.278
710434 97.917 -52.972 52.255
710435 97.938 -52.978 52.232
710436 97.960 -52.984 52.211
710437 97.982 -52.990 52.189
710438 98.042 -52.988 52.164
710439 98.127 -52.985 52.135
710440 98.211 -52.980 52.103
710441 98.248 -52.993 52.071
710442 98.280 -53.008 52.040
710443 98.311 -53.023 52.011
710444 98.343 -53.038 51.983
710445 98.374 -53.054 51.956
710446 98.397 -53.070 51.927
710447 98.415 -53.086 51.897
710448 98.432 -53.102 51.869
710449 98.449 -53.118 51.840
710450 98.467 -53.134 51.809
710451 98.482 -53.149 51.775
710452 98.473 -53.153 51.740
710453 98.463 -53.157 51.706
710454 98.459 -53.167 51.674
710455 98.456 -53.176 51.643
710456 98.471 -53.182 51.610
710457 98.486 -53.186 51.575
710458 98.503 -53.198 51.539
710459 98.521 -53.211 51.504
710460 98.539 -53.224 51.471
710461 98.557 -53.237 51.435
710462 98.575 -53.250 51.401
710463 98.593 -53.263 51.365
710464 98.611 -53.276 51.330
710465 98.629 -53.289 51.295
710466 98.647 -53.302 51.259
710467 98.665 -53.314 51.223
710468 98.687 -53.322 51.188
710469 98.711 -53.325 51.151
710470 98.732 -53.334 51.114
710471 98.750 -53.354 51.077
710472 98.768 -53.374 51.041
710473 98.785 -53.389 51.004
710474 98.803 -53.402 50.967
710475 98.820 -53.415 50.930
710476 98.838 -53.429 50.892
710477 98.863 -53.438 50.854
710478 98.889 -53.448 50.817
710479 98.916 -53.457 50.779
710480 98.942 -53.466 50.740
710481 98.969 -53.475 50.701
710482 98.995 -53.484 50.663
710483 99.021 -53.493 50.623
710484 99.048 -53.502 50.583
710485 99.074 -53.512 50.542
710486 99.101 -53.521 50.501
710487 99.128 -53.530 50.459
710488 99.154 -53.539 50.418
710489 99.181 -53.548 50.378
710490 99.207 -53.557 50.336
710491 99.233 -53.566 50.296
711129 97.794 -52.920 52.405
711130 97.816 -52.926 52.382
711131 97.839 -52.932 52.358
711132 97.860 -52.938 52.334
711133 97.882 -52.944 52.312
711134 97.904 -52.951 52.290
711135 97.926 -52.957 52.267
711136 97.948 -52.963 52.244
711137 97.968 -52.968 52.223
711138 97.990 -52.973 52.201
711139 98.016 -52.980 52.175
711140 98.100 -52.978 52.145
711141 98.184 -52.974 52.113
711142 98.239 -52.980 52.081
711143 98.271 -52.995 52.051
711144 98.302 -53.010 52.022
711145 98.333 -53.026 51.993
711146 98.364 -53.042 51.965
711147 98.395 -53.057 51.936
711148 98.427 -53.072 51.907
711149 98.450 -53.088 51.879
711150 98.467 -53.104 51.850
711151 98.485 -53.120 51.818
711152 98.485 -53.128 51.782
711153 98.479 -53.135 51.747
711154 98.475 -53.145 51.713
711155 98.471 -53.155 51.680
711156 98.468 -53.165 51.648
711157 98.483 -53.170 51.615
711158 98.498 -53.175 51.581
711159 98.513 -53.180 51.545
711160 98.529 -53.188 51.510
711161 98.547 -53.201 51.477
711162 98.565 -53.214 51.441
711163 98.583 -53.227 51.407
711164 98.601 -53.239 51.371
711165 98.619 -53.252 51.335
711166 98.637 -53.265 51.300
711167 98.655 -53.278 51.263
711168 98.674 -53.289 51.227
711169 98.698 -53.292 51.192
711170 98.722 -53.295 51.154
711171 98.742 -53.308 51.117
711172 98.761 -53.327 51.080
711173 98.778 -53.347 51.043
711174 98.796 -53.366 51.007
711175 98.813 -53.379 50.970
711176 98.831 -53.392 50.933
711177 98.849 -53.405 50.895
711178 98.866 -53.418 50.857
711179 98.891 -53.427 50.820
711180 98.918 -53.437 50.782
711181 98.944 -53.446 50.744
711182 98.971 -53.455 50.705
711183 98.997 -53.464 50.666
711184 99.023 -53.473 50.626
711185 99.050 -53.482 50.585
711186 99.076 -53.491 50.544
711187 99.103 -53.500 50.503
711188 99.129 -53.509 50.461
711189 99.156 -53.518 50.420
711190 99.182 -53.527 50.379
711191 99.209 -53.537 50.337
711192 99.235 -53.546 50.296
711775 96.622 -52.584 53.564
711776 96.630 -52.593 53.537
711777 96.640 -52.600 53.511
711778 96.650 -52.603 53.483
711779 96.672 -52.606 53.459
711780 96.699 -52.608 53.434
711830 97.804 -52.905 52.415
711831 97.826 -52.911 52.391
711832 97.847 -52.916 52.369
711833 97.868 -52.922 52.345
711834 97.889 -52.927 52.322
711835 97.910 -52.932 52.301
711836 97.931 -52.938 52.278
711837 97.952 -52.943 52.255
711838 97.973 -52.948 52.232
711839 97.994 -52.954 52.210
711840 98.015 -52.959 52.184
711841 98.074 -52.972 52.155
711842 98.158 -52.967 52.124
711843 98.230 -52.968 52.092
711844 98.261 -52.983 52.062
711845 98.292 -52.998 52.032
711846 98.324 -53.013 52.003
711847 98.355 -53.029 51.976
711848 98.386 -53.044 51.947
711849 98.418 -53.059 51.918
711850 98.449 -53.074 51.890
711851 98.480 -53.089 51.861
711852 98.497 -53.103 51.829
711853 98.494 -53.114 51.792
711854 98.490 -53.123 51.756
711855 98.487 -53.133 51.721
711856 98.483 -53.144 51.686
711857 98.481 -53.154 51.653
711858 98.496 -53.158 51.620
711859 98.511 -53.164 51.586
711860 98.526 -53.169 51.551
711861 98.541 -53.173 51.515
711862 98.556 -53.179 51.482
711863 98.573 -53.190 51.446
711864 98.591 -53.203 51.412
711865 98.609 -53.216 51.376
711866 98.627 -53.229 51.340
711867 98.645 -53.242 51.303
711868 98.663 -53.255 51.267
711869 98.686 -53.259 51.231
711870 98.710 -53.263 51.194
711871 98.734 -53.266 51.157
711872 98.753 -53.282 51.119
711873 98.771 -53.302 51.082
711874 98.789 -53.321 51.045
711875 98.806 -53.340 51.009
711876 98.824 -53.355 50.973
711877 98.842 -53.368 50.935
711878 98.859 -53.380 50.897
711879 98.876 -53.393 50.859
711880 98.894 -53.406 50.822
711881 98.920 -53.416 50.785
711882 98.946 -53.425 50.746
711883 98.973 -53.434 50.707
711884 98.999 -53.443 50.668
711885 99.025 -53.452 50.628
711886 99.052 -53.462 50.587
711887 99.078 -53.471 50.546
711888 99.104 -53.480 50.504
711889 99.131 -53.489 50.462
711890 99.157 -53.498 50.421
711891 99.184 -53.507 50.379
711892 99.211 -53.516 50.336
711893 99.237 -53.525 50.295
711894 99.264 -53.534 50.255
712474 96.608 -52.533 53.635
712475 96.622 -52.556 53.608
712476 96.633 -52.568 53.582
712477 96.642 -52.575 53.555
712478 96.651 -52.583 53.528
712479 96.661 -52.587 53.501
712480 96.683 -52.589 53.476
712481 96.711 -52.591 53.452
712482 96.739 -52.593 53.426
712483 96.761 -52.599 53.402
712484 96.783 -52.605 53.378
712531 97.809 -52.886 52.425
712532 97.830 -52.891 52.401
712533 97.851 -52.897 52.379
712534 97.872 -52.902 52.356
712535 97.893 -52.907 52.333
712536 97.914 -52.913 52.311
712537 97.935 -52.918 52.289
712538 97.956 -52.923 52.265
712539 97.977 -52.929 52.241
712540 97.998 -52.934 52.219
712541 98.019 -52.939 52.193
712542 98.051 -52.948 52.165
712543 98.132 -52.961 52.134
712544 98.216 -52.957 52.103
712545 98.252 -52.970 52.073
712546 98.283 -52.985 52.043
712547 98.314 -53.001 52.014
712548 98.346 -53.016 51.987
712549 98.377 -53.031 51.959
712550 98.408 -53.047 51.930
712551 98.439 -53.062 51.902
712552 98.473 -53.074 51.873
712553 98.502 -53.089 51.842
712554 98.506 -53.102 51.804
712555 98.502 -53.112 51.766
712556 98.498 -53.122 51.730
712557 98.495 -53.132 51.694
712558 98.494 -53.142 51.661
712559 98.508 -53.147 51.627
712560 98.523 -53.152 51.593
712561 98.538 -53.157 51.557
712562 98.554 -53.162 51.522
712563 98.568 -53.167 51.488
712564 98.583 -53.172 51.453
712565 98.599 -53.180 51.418
712566 98.617 -53.192 51.382
712567 98.635 -53.205 51.345
712568 98.653 -53.218 51.309
712569 98.673 -53.227 51.273
712570 98.698 -53.230 51.236
712571 98.722 -53.233 51.198
712572 98.746 -53.236 51.160
712573 98.764 -53.255 51.123
712574 98.781 -53.275 51.086
712575 98.799 -53.295 51.049
712576 98.817 -53.314 51.012
712577 98.835 -53.330 50.975
712578 98.852 -53.343 50.938
712579 98.869 -53.356 50.900
712580 98.887 -53.369 50.862
712581 98.905 -53.382 50.825
712582 98.922 -53.395 50.788
712583 98.948 -53.405 50.749
712584 98.974 -53.414 50.710
712585 99.001 -53.423 50.670
712586 99.027 -53.432 50.631
712587 99.054 -53.441 50.590
712588 99.080 -53.450 50.549
712589 99.106 -53.459 50.507
712590 99.133 -53.468 50.465
712591 99.159 -53.477 50.423
712592 99.186 -53.487 50.381
712593 99.212 -53.496 50.339
712594 99.239 -53.505 50.296
712595 99.265 -53.514 50.256
713174 96.603 -52.487 53.679
713175 96.617 -52.509 53.653
713176 96.631 -52.532 53.626
713177 96.644 -52.550 53.600
713178 96.653 -52.558 53.573
713179 96.662 -52.566 53.547
713180 96.672 -52.570 53.520
713181 96.695 -52.572 53.495
713182 96.723 -52.575 53.470
713183 96.749 -52.578 53.444
713184 96.770 -52.584 53.420
713185 96.793 -52.590 53.395
713186 96.814 -52.596 53.371
713232 97.813 -52.867 52.434
713233 97.835 -52.872 52.412
713234 97.856 -52.877 52.389
713235 97.876 -52.882 52.366
713236 97.897 -52.888 52.343
713237 97.919 -52.893 52.321
713238 97.940 -52.898 52.299
713239 97.960 -52.904 52.274
713240 97.982 -52.909 52.251
713241 98.003 -52.914 52.227
713242 98.023 -52.919 52.202
713243 98.044 -52.925 52.175
713244 98.107 -52.941 52.146
713245 98.190 -52.951 52.115
713246 98.242 -52.957 52.084
713247 98.274 -52.973 52.054
713248 98.305 -52.988 52.025
713249 98.337 -53.004 51.999
713250 98.368 -53.018 51.971
713251 98.402 -53.031 51.942
713252 98.435 -53.044 51.915
713253 98.469 -53.056 51.887
713254 98.498 -53.070 51.856
713255 98.512 -53.089 51.818
713256 98.514 -53.101 51.778
713257 98.510 -53.111 51.740
713258 98.506 -53.121 51.704
713259 98.506 -53.130 51.670
713260 98.521 -53.136 51.636
713261 98.536 -53.140 51.601
713262 98.551 -53.145 51.565
713263 98.566 -53.151 51.530
713264 98.581 -53.155 51.495
713265 98.596 -53.161 51.460
713266 98.611 -53.166 51.425
713267 98.626 -53.170 51.389
713268 98.643 -53.181 51.352
713269 98.661 -53.194 51.315
713270 98.685 -53.197 51.278
713271 98.709 -53.200 51.241
713272 98.733 -53.203 51.203
713273 98.756 -53.210 51.164
713274 98.774 -53.230 51.127
713275 98.792 -53.249 51.090
713276 98.809 -53.268 51.052
713277 98.828 -53.288 51.015
713278 98.845 -53.306 50.978
713279 98.862 -53.319 50.940
713280 98.880 -53.332 50.903
713281 98.898 -53.345 50.865
713282 98.915 -53.358 50.828
713283 98.933 -53.371 50.791
713284 98.950 -53.383 50.752
713285 98.976 -53.393 50.713
713286 99.003 -53.402 50.673
713287 99.029 -53.411 50.634
713288 99.056 -53.421 50.593
713289 99.082 -53.430 50.552
713290 99.109 -53.439 50.510
713291 99.135 -53.448 50.467
713292 99.161 -53.457 50.425
713293 99.188 -53.466 50.383
713294 99.214 -53.475 50.340
713295 99.240 -53.484 50.298
713296 99.267 -53.493 50.257
713874 96.598 -52.441 53.720
713875 96.612 -52.464 53.694
713876 96.626 -52.486 53.669
713877 96.640 -52.508 53.643
713878 96.654 -52.531 53.617
713879 96.664 -52.541 53.590
713880 96.673 -52.549 53.564
713881 96.683 -52.553 53.537
713882 96.707 -52.556 53.512
713883 96.734 -52.558 53.487
713884 96.758 -52.562 53.462
713885 96.780 -52.568 53.437
713886 96.802 -52.575 53.412
713887 96.824 -52.581 53.387
713888 96.846 -52.587 53.363
713933 97.818 -52.847 52.445
713934 97.839 -52.853 52.421
713935 97.860 -52.858 52.399
713936 97.881 -52.863 52.376
713937 97.902 -52.868 52.353
713938 97.923 -52.873 52.331
713939 97.944 -52.879 52.309
713940 97.965 -52.884 52.284
713941 97.986 -52.889 52.261
713942 98.007 -52.894 52.237
713943 98.028 -52.900 52.211
713944 98.049 -52.905 52.185
713945 98.084 -52.914 52.156
713946 98.164 -52.935 52.124
713947 98.233 -52.945 52.095
713948 98.265 -52.960 52.065
713949 98.296 -52.975 52.036
713950 98.330 -52.988 52.010
713951 98.364 -53.000 51.982
713952 98.397 -53.013 51.954
713953 98.432 -53.026 51.928
713954 98.465 -53.038 51.901
713955 98.494 -53.052 51.869
713956 98.508 -53.070 51.831
713957 98.522 -53.088 51.789
713958 98.522 -53.100 51.751
713959 98.518 -53.110 51.715
713960 98.518 -53.119 51.680
713961 98.533 -53.124 51.645
713962 98.549 -53.129 51.609
713963 98.564 -53.134 51.572
713964 98.578 -53.139 51.537
713965 98.593 -53.144 51.502
713966 98.608 -53.149 51.467
713967 98.624 -53.154 51.433
713968 98.638 -53.159 51.397
713969 98.654 -53.163 51.359
713970 98.673 -53.164 51.322
713971 98.697 -53.167 51.284
713972 98.721 -53.170 51.246
713973 98.745 -53.173 51.208
713974 98.767 -53.183 51.168
713975 98.784 -53.203 51.131
713976 98.802 -53.223 51.094
713977 98.820 -53.242 51.056
713978 98.838 -53.262 51.019
713979 98.856 -53.282 50.982
713980 98.873 -53.295 50.943
713981 98.890 -53.308 50.906
713982 98.908 -53.321 50.868
713983 98.926 -53.334 50.831
713984 98.943 -53.347 50.794
713985 98.960 -53.360 50.756
713986 98.978 -53.373 50.716
713987 99.004 -53.382 50.677
713988 99.031 -53.391 50.636
713989 99.057 -53.400 50.597
713990 99.084 -53.409 50.555
713991 99.111 -53.418 50.513
713992 99.137 -53.427 50.470
713993 99.163 -53.437 50.427
713994 99.190 -53.446 50.385
713995 99.216 -53.455 50.342
713996 99.242 -53.464 50.301
713997 99.269 -53.473 50.260
714574 96.593 -52.396 53.759
714575 96.607 -52.418 53.734
714576 96.621 -52.440 53.708
714577 96.635 -52.462 53.683
714578 96.649 -52.484 53.658
714579 96.663 -52.507 53.632
714580 96.675 -52.524 53.607
714581 96.684 -52.532 53.580
714582 96.694 -52.537 53.554
714583 96.718 -52.539 53.528
714584 96.746 -52.541 53.503
714585 96.767 -52.547 53.478
714586 96.790 -52.553 53.453
714587 96.811 -52.559 53.427
714588 96.833 -52.565 53.402
714589 96.855 -52.572 53.377
714590 96.877 -52.578 53.353
714634 97.822 -52.828 52.455
714635 97.843 -52.833 52.431
714636 97.864 -52.838 52.409
714637 97.885 -52.844 52.386
714638 97.906 -52.849 52.363
714639 97.927 -52.854 52.342
714640 97.948 -52.860 52.319
714641 97.969 -52.864 52.295
714642 97.990 -52.869 52.271
714643 98.011 -52.875 52.247
714644 98.032 -52.880 52.222
714645 98.053 -52.885 52.196
714646 98.074 -52.891 52.166
714647 98.142 -52.909 52.135
714648 98.222 -52.930 52.106
714649 98.258 -52.944 52.077
714650 98.292 -52.957 52.049
714651 98.326 -52.970 52.022
714652 98.360 -52.982 51.995
714653 98.394 -52.995 51.969
714654 98.428 -53.007 51.943
714655 98.461 -53.020 51.915
714656 98.490 -53.034 51.883
714657 98.504 -53.052 51.844
714658 98.518 -53.070 51.802
714659 98.532 -53.088 51.763
714660 98.530 -53.098 51.728
714661 98.531 -53.107 51.692
714662 98.546 -53.112 51.655
714663 98.561 -53.117 51.618
714664 98.576 -53.123 51.581
714665 98.591 -53.127 51.545
714666 98.606 -53.132 51.510
714667 98.622 -53.135 51.474
714668 98.638 -53.136 51.440
714669 98.655 -53.136 51.403
714670 98.671 -53.136 51.365
714671 98.688 -53.136 51.327
714672 98.709 -53.138 51.290
714673 98.733 -53.141 51.252
714674 98.757 -53.144 51.213
714675 98.777 -53.158 51.174
714676 98.795 -53.177 51.136
714677 98.812 -53.196 51.098
714678 98.831 -53.216 51.060
714679 98.848 -53.236 51.023
714680 98.866 -53.255 50.986
714681 98.883 -53.271 50.947
714682 98.901 -53.284 50.910
714683 98.919 -53.297 50.872
714684 98.936 -53.310 50.834
714685 98.953 -53.323 50.796
714686 98.971 -53.336 50.759
714687 98.989 -53.349 50.721
714688 99.006 -53.361 50.681
714689 99.033 -53.371 50.640
714690 99.059 -53.380 50.600
714691 99.086 -53.389 50.559
714692 99.112 -53.398 50.517
714693 99.139 -53.407 50.474
714694 99.165 -53.416 50.431
714695 99.192 -53.425 50.389
714696 99.218 -53.434 50.346
714697 99.244 -53.443 50.305
714698 99.271 -53.452 50.263
715274 96.588 -52.350 53.794
715275 96.602 -52.372 53.770
715276 96.616 -52.395 53.746
715277 96.630 -52.417 53.721
715278 96.644 -52.439 53.696
715279 96.658 -52.461 53.672
715280 96.672 -52.484 53.647
715281 96.686 -52.506 53.622
715282 96.695 -52.515 53.596
715283 96.705 -52.520 53.569
715284 96.730 -52.522 53.543
715285 96.755 -52.526 53.519
715286 96.777 -52.532 53.494
715287 96.799 -52.538 53.468
715288 96.821 -52.544 53.442
715289 96.843 -52.550 53.416
715290 96.865 -52.556 53.391
715291 96.887 -52.562 53.365
715335 97.826 -52.808 52.465
715336 97.848 -52.813 52.441
715337 97.868 -52.819 52.419
715338 97.889 -52.824 52.396
715339 97.911 -52.829 52.374
715340 97.931 -52.835 52.352
715341 97.952 -52.840 52.329
715342 97.973 -52.845 52.305
715343 97.994 -52.850 52.281
715344 98.015 -52.856 52.258
715345 98.036 -52.860 52.235
715346 98.057 -52.866 52.207
715347 98.078 -52.871 52.178
715348 98.119 -52.882 52.148
715349 98.199 -52.903 52.118
715350 98.247 -52.919 52.088
715351 98.283 -52.934 52.062
715352 98.319 -52.948 52.036
715353 98.355 -52.963 52.010
715354 98.390 -52.976 51.984
715355 98.424 -52.989 51.957
715356 98.457 -53.002 51.929
715357 98.486 -53.016 51.897
715358 98.500 -53.034 51.856
715359 98.514 -53.051 51.815
715360 98.528 -53.070 51.775
715361 98.542 -53.087 51.740
715362 98.544 -53.095 51.703
715363 98.559 -53.101 51.665
715364 98.574 -53.106 51.627
715365 98.590 -53.107 51.589
715366 98.606 -53.107 51.552
715367 98.623 -53.107 51.517
715368 98.640 -53.107 51.480
715369 98.656 -53.108 51.445
715370 98.673 -53.108 51.408
715371 98.690 -53.108 51.370
715372 98.706 -53.109 51.332
715373 98.722 -53.109 51.295
715374 98.744 -53.111 51.257
715375 98.769 -53.114 51.218
715376 98.787 -53.131 51.179
715377 98.805 -53.151 51.140
715378 98.823 -53.170 51.102
715379 98.841 -53.190 51.064
715380 98.859 -53.210 51.027
715381 98.876 -53.230 50.989
715382 98.894 -53.247 50.951
715383 98.912 -53.260 50.913
715384 98.929 -53.273 50.875
715385 98.947 -53.286 50.837
715386 98.964 -53.299 50.799
715387 98.982 -53.311 50.762
715388 98.999 -53.324 50.723
715389 99.017 -53.337 50.683
715390 99.034 -53.350 50.644
715391 99.061 -53.359 50.603
715392 99.087 -53.368 50.561
715393 99.114 -53.377 50.519
715394 99.140 -53.386 50.477
715395 99.167 -53.396 50.434
715396 99.194 -53.405 50.392
715397 99.220 -53.414 50.350
715398 99.246 -53.423 50.308
715399 99.273 -53.432 50.266
715400 99.299 -53.441 50.225
715975 96.597 -52.327 53.804
715976 96.611 -52.349 53.781
715977 96.625 -52.371 53.757
715978 96.639 -52.393 53.733
715979 96.653 -52.415 53.709
715980 96.667 -52.438 53.685
715981 96.681 -52.460 53.661
715982 96.695 -52.482 53.636
715983 96.707 -52.497 53.610
715984 96.716 -52.503 53.583
715985 96.741 -52.506 53.558
715986 96.764 -52.511 53.533
715987 96.787 -52.517 53.507
715988 96.809 -52.523 53.481
715989 96.830 -52.529 53.455
715990 96.853 -52.535 53.429
715991 96.874 -52.541 53.403
715992 96.896 -52.547 53.377
715993 96.918 -52.553 53.352
716036 97.830 -52.788 52.475
716037 97.852 -52.794 52.451
716038 97.873 -52.799 52.429
716039 97.893 -52.804 52.406
716040 97.915 -52.810 52.384
716041 97.936 -52.815 52.362
716042 97.956 -52.820 52.339
716043 97.978 -52.825 52.315
716044 97.999 -52.831 52.291
716045 98.020 -52.836 52.268
716046 98.040 -52.841 52.246
716047 98.062 -52.847 52.219
716048 98.083 -52.851 52.189
716049 98.103 -52.857 52.159
716050 98.175 -52.876 52.129
716051 98.234 -52.894 52.100
716052 98.270 -52.908 52.074
716053 98.306 -52.923 52.049
716054 98.342 -52.938 52.023
716055 98.378 -52.952 51.997
716056 98.414 -52.966 51.969
716057 98.450 -52.981 51.940
716058 98.476 -52.997 51.909
716059 98.490 -53.015 51.870
716060 98.504 -53.032 51.828
716061 98.518 -53.051 51.789
716062 98.532 -53.068 51.751
716063 98.556 -53.082 51.713
716064 98.573 -53.081 51.674
716065 98.591 -53.080 51.635
716066 98.607 -53.079 51.597
716067 98.624 -53.080 51.559
716068 98.641 -53.080 51.521
716069 98.657 -53.080 51.485
716070 98.674 -53.081 51.448
716071 98.691 -53.081 51.411
716072 98.707 -53.081 51.374
716073 98.724 -53.081 51.336
716074 98.740 -53.082 51.299
716075 98.757 -53.082 51.260
716076 98.780 -53.086 51.222
716077 98.798 -53.105 51.183
716078 98.815 -53.124 51.144
716079 98.834 -53.144 51.106
716080 98.851 -53.164 51.067
716081 98.869 -53.183 51.030
716082 98.887 -53.203 50.992
716083 98.905 -53.223 50.954
716084 98.922 -53.236 50.917
716085 98.940 -53.248 50.879
716086 98.957 -53.261 50.841
716087 98.975 -53.274 50.803
716088 98.992 -53.287 50.765
716089 99.010 -53.300 50.725
716090 99.027 -53.313 50.685
716091 99.045 -53.326 50.646
716092 99.063 -53.339 50.605
716093 99.089 -53.348 50.563
716094 99.116 -53.357 50.521
716095 99.142 -53.366 50.480
716096 99.169 -53.375 50.437
716097 99.195 -53.384 50.394
716098 99.222 -53.393 50.353
716099 99.248 -53.402 50.310
716100 99.275 -53.411 50.268
716101 99.301 -53.421 50.227
716675 96.592 -52.281 53.833
716676 96.606 -52.303 53.812
716677 96.620 -52.326 53.790
716678 96.634 -52.348 53.768
716679 96.648 -52.370 53.744
716680 96.662 -52.392 53.721
716681 96.676 -52.415 53.697
716682 96.690 -52.437 53.673
716683 96.704 -52.459 53.649
716684 96.718 -52.481 53.623
716685 96.727 -52.487 53.597
716686 96.752 -52.490 53.572
716687 96.774 -52.496 53.547
716688 96.796 -52.502 53.521
716689 96.818 -52.508 53.494
716690 96.840 -52.514 53.468
716691 96.862 -52.520 53.441
716692 96.884 -52.526 53.415
716693 96.906 -52.532 53.389
716694 96.928 -52.538 53.364
716738 97.856 -52.774 52.461
716739 97.877 -52.779 52.439
716740 97.898 -52.785 52.416
716741 97.919 -52.790 52.393
716742 97.940 -52.795 52.371
716743 97.961 -52.800 52.349
716744 97.982 -52.806 52.324
716745 98.003 -52.811 52.301
716746 98.024 -52.816 52.279
716747 98.045 -52.822 52.256
716748 98.066 -52.827 52.229
716749 98.087 -52.832 52.200
716750 98.108 -52.838 52.170
716751 98.153 -52.849 52.140
716752 98.222 -52.869 52.112
716753 98.258 -52.883 52.086
716754 98.294 -52.897 52.060
716755 98.329 -52.912 52.034
716756 98.366 -52.926 52.007
716757 98.402 -52.941 51.979
716758 98.430 -52.957 51.951
716759 98.444 -52.975 51.919
716760 98.458 -52.992 51.881
716761 98.472 -53.010 51.839
716762 98.494 -53.027 51.798
716763 98.531 -53.044 51.758
716764 98.562 -53.057 51.719
716765 98.579 -53.062 51.681
716766 98.598 -53.067 51.642
716767 98.616 -53.065 51.603
716768 98.633 -53.064 51.564
716769 98.650 -53.064 51.527
716770 98.668 -53.063 51.490
716771 98.686 -53.062 51.452
716772 98.703 -53.060 51.414
716773 98.720 -53.060 51.376
716774 98.738 -53.059 51.339
716775 98.755 -53.058 51.301
716776 98.773 -53.057 51.263
716777 98.791 -53.059 51.225
716778 98.808 -53.079 51.186
716779 98.826 -53.098 51.147
716780 98.844 -53.118 51.108
716781 98.862 -53.138 51.069
716782 98.879 -53.157 51.032
716783 98.897 -53.176 50.994
716784 98.915 -53.196 50.956
716785 98.933 -53.211 50.919
716786 98.950 -53.224 50.881
716787 98.968 -53.237 50.843
716788 98.985 -53.250 50.805
716789 99.003 -53.263 50.766
716790 99.020 -53.276 50.728
716791 99.038 -53.289 50.688
716792 99.055 -53.302 50.648
716793 99.073 -53.315 50.607
716794 99.091 -53.327 50.565
716795 99.118 -53.336 50.524
716796 99.144 -53.346 50.482
716797 99.170 -53.355 50.439
716798 99.197 -53.364 50.396
716799 99.223 -53.373 50.354
716800 99.250 -53.382 50.312
716801 99.276 -53.391 50.271
716802 99.303 -53.401 50.229
717302 95.426 -50.641 55.421
717303 95.453 -50.661 55.406
717304 95.480 -50.681 55.391
717305 95.507 -50.701 55.375
717306 95.533 -50.720 55.359
717307 95.560 -50.740 55.342
717308 95.587 -50.760 55.325
717309 95.614 -50.779 55.307
717376 96.601 -52.258 53.841
717377 96.615 -52.280 53.821
717378 96.629 -52.302 53.799
717379 96.643 -52.324 53.777
717380 96.657 -52.346 53.755
717381 96.671 -52.368 53.731
717382 96.685 -52.391 53.708
717383 96.699 -52.413 53.686
717384 96.713 -52.435 53.662
717385 96.727 -52.458 53.637
717386 96.738 -52.470 53.611
717387 96.762 -52.474 53.585
717388 96.784 -52.481 53.559
717389 96.806 -52.487 53.533
717390 96.827 -52.493 53.507
717391 96.850 -52.499 53.481
717392 96.871 -52.505 53.452
717393 96.893 -52.511 53.426
717394 96.915 -52.517 53.400
717395 96.937 -52.522 53.375
717439 97.860 -52.754 52.471
717440 97.881 -52.760 52.447
717441 97.902 -52.765 52.425
717442 97.923 -52.770 52.402
717443 97.944 -52.775 52.380
717444 97.965 -52.781 52.357
717445 97.986 -52.786 52.334
717446 98.007 -52.791 52.310
717447 98.028 -52.797 52.289
717448 98.049 -52.802 52.265
717449 98.070 -52.807 52.239
717450 98.091 -52.813 52.210
717451 98.112 -52.818 52.180
717452 98.133 -52.823 52.151
717453 98.209 -52.843 52.123
717454 98.245 -52.857 52.098
717455 98.281 -52.872 52.071
717456 98.317 -52.887 52.044
717457 98.353 -52.901 52.017
717458 98.384 -52.916 51.989
717459 98.397 -52.934 51.960
717460 98.412 -52.952 51.927
717461 98.426 -52.970 51.890
717462 98.463 -52.986 51.847
717463 98.500 -53.003 51.805
717464 98.536 -53.020 51.764
717465 98.570 -53.035 51.724
717466 98.583 -53.035 51.686
717467 98.602 -53.043 51.648
717468 98.623 -53.051 51.609
717469 98.640 -53.051 51.569
717470 98.658 -53.050 51.532
717471 98.675 -53.049 51.494
717472 98.693 -53.048 51.456
717473 98.710 -53.047 51.418
717474 98.728 -53.046 51.380
717475 98.745 -53.045 51.341
717476 98.763 -53.045 51.303
717477 98.780 -53.044 51.266
717478 98.797 -53.043 51.227
717479 98.818 -53.055 51.188
717480 98.837 -53.072 51.150
717481 98.854 -53.092 51.111
717482 98.872 -53.111 51.072
717483 98.890 -53.131 51.034
717484 98.908 -53.151 50.997
717485 98.925 -53.170 50.959
717486 98.943 -53.187 50.922
717487 98.960 -53.200 50.883
717488 98.978 -53.213 50.845
717489 98.996 -53.226 50.807
717490 99.013 -53.239 50.769
717491 99.031 -53.252 50.730
717492 99.048 -53.265 50.691
717493 99.066 -53.278 50.651
717494 99.083 -53.291 50.609
717495 99.101 -53.304 50.567
717496 99.120 -53.316 50.526
717497 99.146 -53.325 50.484
717498 99.172 -53.334 50.441
717499 99.199 -53.343 50.398
717500 99.225 -53.352 50.356
717501 99.252 -53.361 50.315
717502 99.278 -53.371 50.273
717503 99.305 -53.380 50.231
718001 95.383 -50.581 55.473
718002 95.414 -50.598 55.458
718003 95.445 -50.616 55.443
718004 95.471 -50.636 55.428
718005 95.498 -50.656 55.412
718006 95.525 -50.676 55.396
718007 95.552 -50.695 55.380
718008 95.579 -50.715 55.362
718009 95.606 -50.735 55.344
718010 95.633 -50.754 55.325
718011 95.660 -50.774 55.306
718012 95.687 -50.794 55.288
718077 96.610 -52.234 53.850
718078 96.624 -52.256 53.829
718079 96.638 -52.279 53.808
718080 96.652 -52.301 53.787
718081 96.666 -52.323 53.764
718082 96.680 -52.346 53.741
718083 96.694 -52.368 53.719
718084 96.708 -52.390 53.697
718085 96.722 -52.412 53.673
718086 96.736 -52.434 53.649
718087 96.749 -52.454 53.623
718088 96.771 -52.459 53.597
718089 96.793 -52.465 53.572
718090 96.815 -52.471 53.546
718091 96.836 -52.476 53.519
718092 96.857 -52.481 53.492
718093 96.878 -52.487 53.465
718094 96.899 -52.492 53.438
718095 96.920 -52.497 53.411
718096 96.941 -52.503 53.385
718140 97.865 -52.735 52.480
718141 97.885 -52.740 52.456
718142 97.907 -52.745 52.433
718143 97.928 -52.750 52.411
718144 97.948 -52.756 52.388
718145 97.969 -52.761 52.365
718146 97.991 -52.766 52.342
718147 98.011 -52.772 52.319
718148 98.032 -52.777 52.296
718149 98.054 -52.782 52.273
718150 98.074 -52.788 52.247
718151 98.095 -52.793 52.218
718152 98.116 -52.798 52.189
718153 98.137 -52.803 52.161
718154 98.186 -52.816 52.133
718155 98.233 -52.832 52.108
718156 98.269 -52.847 52.082
718157 98.305 -52.861 52.055
718158 98.338 -52.876 52.027
718159 98.352 -52.894 51.999
718160 98.366 -52.912 51.968
718161 98.395 -52.929 51.932
718162 98.432 -52.945 51.894
718163 98.469 -52.962 51.852
718164 98.506 -52.979 51.812
718165 98.543 -52.995 51.771
718166 98.578 -53.012 51.729
718167 98.589 -53.010 51.690
718168 98.606 -53.016 51.653
718169 98.626 -53.023 51.615
718170 98.646 -53.031 51.575
718171 98.665 -53.036 51.536
718172 98.683 -53.035 51.498
718173 98.700 -53.035 51.460
718174 98.717 -53.033 51.422
718175 98.735 -53.032 51.384
718176 98.752 -53.032 51.345
718177 98.770 -53.031 51.306
718178 98.787 -53.030 51.269
718179 98.805 -53.029 51.231
718180 98.825 -53.039 51.191
718181 98.845 -53.051 51.153
718182 98.865 -53.066 51.114
718183 98.882 -53.085 51.075
718184 98.901 -53.104 51.038
718185 98.918 -53.124 50.999
718186 98.936 -53.144 50.962
718187 98.953 -53.163 50.923
718188 98.971 -53.176 50.884
718189 98.989 -53.189 50.846
718190 99.006 -53.202 50.809
718191 99.024 -53.215 50.771
718192 99.041 -53.228 50.733
718193 99.059 -53.241 50.694
718194 99.076 -53.254 50.654
718195 99.094 -53.267 50.611
718196 99.112 -53.280 50.569
718197 99.129 -53.292 50.527
718198 99.148 -53.305 50.486
718199 99.174 -53.314 50.443
718200 99.201 -53.323 50.401
718201 99.227 -53.332 50.358
718202 99.254 -53.341 50.317
718203 99.280 -53.350 50.276
718204 99.306 -53.360 50.234
718701 95.369 -50.539 55.507
718702 95.400 -50.556 55.493
718703 95.431 -50.574 55.479
718704 95.462 -50.591 55.464
718705 95.490 -50.610 55.449
718706 95.517 -50.630 55.433
718707 95.544 -50.650 55.418
718708 95.571 -50.670 55.400
718709 95.598 -50.690 55.382
718710 95.625 -50.710 55.363
718711 95.652 -50.729 55.344
718712 95.679 -50.749 55.324
718713 95.706 -50.769 55.305
718714 95.723 -50.790 55.286
718778 96.619 -52.211 53.857
718779 96.633 -52.233 53.837
718780 96.647 -52.255 53.817
718781 96.661 -52.277 53.795
718782 96.675 -52.299 53.774
718783 96.689 -52.322 53.751
718784 96.704 -52.343 53.730
718785 96.719 -52.365 53.708
718786 96.734 -52.386 53.685
718787 96.749 -52.407 53.660
718788 96.764 -52.428 53.635
718789 96.782 -52.442 53.610
718790 96.803 -52.446 53.584
718791 96.823 -52.452 53.558
718792 96.843 -52.456 53.531
718793 96.864 -52.462 53.503
718794 96.885 -52.467 53.476
718795 96.905 -52.472 53.448
718796 96.926 -52.478 53.422
718797 96.947 -52.483 53.395
718841 97.869 -52.715 52.489
718842 97.890 -52.720 52.464
718843 97.911 -52.725 52.441
718844 97.932 -52.731 52.419
718845 97.953 -52.736 52.396
718846 97.973 -52.741 52.373
718847 97.995 -52.747 52.349
718848 98.016 -52.752 52.326
718849 98.037 -52.757 52.303
718850 98.058 -52.763 52.280
718851 98.079 -52.768 52.255
718852 98.100 -52.773 52.227
718853 98.121 -52.778 52.198
718854 98.142 -52.784 52.170
718855 98.163 -52.789 52.142
718856 98.220 -52.807 52.116
718857 98.256 -52.822 52.091
718858 98.291 -52.836 52.063
718859 98.305 -52.854 52.035
718860 98.327 -52.871 52.005
718861 98.364 -52.888 51.972
718862 98.401 -52.904 51.935
718863 98.438 -52.921 51.896
718864 98.475 -52.938 51.855
718865 98.511 -52.954 51.816
718866 98.548 -52.971 51.775
718867 98.585 -52.988 51.735
718868 98.597 -52.988 51.697
718869 98.609 -52.988 51.658
718870 98.629 -52.996 51.620
718871 98.649 -53.004 51.580
718872 98.669 -53.011 51.542
718873 98.689 -53.020 51.503
718874 98.707 -53.021 51.465
718875 98.725 -53.020 51.427
718876 98.742 -53.019 51.388
718877 98.760 -53.018 51.349
718878 98.777 -53.017 51.311
718879 98.794 -53.017 51.272
718880 98.812 -53.015 51.234
718881 98.832 -53.023 51.195
718882 98.852 -53.035 51.156
718883 98.872 -53.048 51.117
718884 98.892 -53.060 51.078
718885 98.911 -53.079 51.040
718886 98.928 -53.098 51.002
718887 98.946 -53.118 50.963
718888 98.964 -53.137 50.925
718889 98.982 -53.152 50.887
718890 98.999 -53.165 50.849
718891 99.017 -53.178 50.812
718892 99.034 -53.191 50.774
718893 99.052 -53.204 50.736
718894 99.069 -53.217 50.696
718895 99.087 -53.230 50.656
718896 99.104 -53.242 50.614
718897 99.122 -53.255 50.572
718898 99.139 -53.268 50.530
718899 99.157 -53.281 50.488
718900 99.176 -53.293 50.446
718901 99.203 -53.302 50.404
718902 99.229 -53.311 50.362
718903 99.256 -53.321 50.320
718904 99.282 -53.330 50.278
718905 99.308 -53.339 50.238
719401 95.355 -50.497 55.542
719402 95.386 -50.515 55.528
719403 95.417 -50.532 55.514
719404 95.448 -50.550 55.500
719405 95.479 -50.567 55.486
719406 95.509 -50.585 55.470
719407 95.536 -50.605 55.455
719408 95.563 -50.625 55.439
719409 95.590 -50.644 55.421
719410 95.617 -50.664 55.402
719411 95.644 -50.684 55.384
719412 95.671 -50.704 55.364
719413 95.698 -50.723 55.344
719414 95.718 -50.745 55.324
719415 95.732 -50.766 55.303
719416 95.746 -50.789 55.283
719479 96.633 -52.183 53.863
719480 96.649 -52.204 53.844
719481 96.664 -52.225 53.824
719482 96.679 -52.246 53.803
719483 96.693 -52.268 53.781
719484 96.709 -52.289 53.760
719485 96.724 -52.310 53.740
719486 96.739 -52.331 53.717
719487 96.754 -52.352 53.694
719488 96.769 -52.374 53.670
719489 96.784 -52.396 53.645
719490 96.799 -52.417 53.621
719491 96.816 -52.432 53.596
719492 96.837 -52.435 53.569
719493 96.858 -52.438 53.543
719494 96.879 -52.442 53.515
719495 96.900 -52.447 53.487
719496 96.921 -52.452 53.458
719497 96.941 -52.458 53.431
719498 96.961 -52.463 53.405
719542 97.873 -52.696 52.497
719543 97.894 -52.701 52.473
719544 97.915 -52.706 52.449
719545 97.936 -52.711 52.425
719546 97.957 -52.716 52.403
719547 97.978 -52.722 52.381
719548 97.999 -52.727 52.356
719549 98.020 -52.732 52.333
719550 98.041 -52.738 52.310
719551 98.062 -52.743 52.286
719552 98.083 -52.748 52.261
719553 98.104 -52.753 52.235
719554 98.125 -52.759 52.206
719555 98.146 -52.764 52.178
719556 98.167 -52.769 52.150
719557 98.208 -52.782 52.124
719558 98.244 -52.796 52.098
719559 98.260 -52.813 52.070
719560 98.297 -52.830 52.040
719561 98.334 -52.847 52.010
719562 98.370 -52.863 51.975
719563 98.407 -52.880 51.937
719564 98.444 -52.897 51.898
719565 98.481 -52.913 51.857
719566 98.517 -52.930 51.819
719567 98.554 -52.947 51.780
719568 98.591 -52.963 51.742
719569 98.605 -52.965 51.704
719570 98.616 -52.963 51.665
719571 98.633 -52.969 51.626
719572 98.653 -52.976 51.587
719573 98.672 -52.985 51.547
719574 98.693 -52.992 51.508
719575 98.713 -53.000 51.470
719576 98.732 -53.007 51.431
719577 98.749 -53.005 51.392
719578 98.767 -53.004 51.353
719579 98.784 -53.004 51.315
719580 98.802 -53.003 51.276
719581 98.819 -53.001 51.237
719582 98.838 -53.007 51.198
719583 98.859 -53.020 51.159
719584 98.879 -53.032 51.120
719585 98.900 -53.044 51.081
719586 98.920 -53.056 51.043
719587 98.939 -53.072 51.005
719588 98.957 -53.092 50.966
719589 98.975 -53.111 50.928
719590 98.992 -53.128 50.890
719591 99.010 -53.141 50.853
719592 99.027 -53.154 50.815
719593 99.045 -53.167 50.777
719594 99.062 -53.180 50.738
719595 99.080 -53.192 50.699
719596 99.097 -53.205 50.658
719597 99.115 -53.218 50.617
719598 99.132 -53.231 50.575
719599 99.150 -53.244 50.533
719600 99.167 -53.257 50.491
719601 99.185 -53.270 50.449
719602 99.205 -53.282 50.407
719603 99.231 -53.291 50.366
719604 99.258 -53.300 50.324
719605 99.284 -53.310 50.282
720101 95.335 -50.449 55.576
720102 95.370 -50.470 55.563
720103 95.403 -50.491 55.549
720104 95.434 -50.508 55.536
720105 95.465 -50.525 55.523
720106 95.496 -50.543 55.509
720107 95.527 -50.560 55.493
720108 95.555 -50.580 55.477
720109 95.582 -50.600 55.461
720110 95.609 -50.619 55.443
720111 95.636 -50.639 55.424
720112 95.663 -50.659 55.404
720113 95.689 -50.679 55.384
720114 95.713 -50.699 55.363
720115 95.727 -50.721 55.343
720116 95.741 -50.743 55.321
720117 95.755 -50.766 55.300
720180 96.653 -52.149 53.868
720181 96.668 -52.171 53.849
720182 96.683 -52.192 53.830
720183 96.698 -52.214 53.809
720184 96.714 -52.235 53.788
720185 96.729 -52.256 53.767
720186 96.744 -52.277 53.746
720187 96.759 -52.299 53.724
720188 96.773 -52.320 53.702
720189 96.789 -52.341 53.678
720190 96.804 -52.362 53.654
720191 96.819 -52.384 53.629
720192 96.834 -52.405 53.604
720193 96.850 -52.423 53.578
720194 96.871 -52.426 53.551
720195 96.892 -52.428 53.524
720196 96.913 -52.431 53.495
720197 96.935 -52.434 53.467
720198 96.956 -52.438 53.440
720199 96.976 -52.443 53.413
720243 97.882 -52.676 52.504
720244 97.902 -52.681 52.481
720245 97.923 -52.687 52.457
720246 97.943 -52.691 52.433
720247 97.964 -52.697 52.410
720248 97.985 -52.702 52.387
720249 98.005 -52.707 52.362
720250 98.025 -52.713 52.338
720251 98.046 -52.718 52.315
720252 98.067 -52.723 52.291
720253 98.087 -52.728 52.267
720254 98.108 -52.734 52.241
720255 98.129 -52.739 52.214
720256 98.150 -52.744 52.186
720257 98.171 -52.750 52.157
720258 98.196 -52.756 52.130
720259 98.229 -52.772 52.103
720260 98.266 -52.789 52.073
720261 98.302 -52.806 52.043
720262 98.339 -52.822 52.011
720263 98.376 -52.839 51.975
720264 98.413 -52.856 51.938
720265 98.449 -52.872 51.900
720266 98.486 -52.889 51.860
720267 98.523 -52.906 51.824
720268 98.560 -52.923 51.787
720269 98.596 -52.939 51.751
720270 98.613 -52.942 51.714
720271 98.624 -52.941 51.674
720272 98.636 -52.941 51.634
720273 98.656 -52.949 51.593
720274 98.676 -52.957 51.553
720275 98.696 -52.965 51.514
720276 98.716 -52.973 51.475
720277 98.736 -52.980 51.436
720278 98.756 -52.988 51.397
720279 98.774 -52.991 51.358
720280 98.792 -52.990 51.319
720281 98.809 -52.989 51.280
720282 98.827 -52.988 51.241
720283 98.845 -52.991 51.202
720284 98.865 -53.004 51.163
720285 98.886 -53.016 51.123
720286 98.906 -53.028 51.084
720287 98.927 -53.040 51.046
720288 98.947 -53.052 51.009
720289 98.967 -53.065 50.970
720290 98.985 -53.085 50.933
720291 99.003 -53.104 50.894
720292 99.020 -53.117 50.856
720293 99.038 -53.129 50.818
720294 99.055 -53.142 50.780
720295 99.073 -53.155 50.740
720296 99.090 -53.168 50.701
720297 99.108 -53.181 50.660
720298 99.125 -53.194 50.619
720299 99.143 -53.207 50.578
720300 99.160 -53.220 50.537
720301 99.178 -53.233 50.495
720302 99.196 -53.246 50.453
720303 99.213 -53.259 50.411
720304 99.233 -53.270 50.370
720305 99.259 -53.280 50.329
720306 99.286 -53.289 50.287
720802 95.348 -50.419 55.594
720803 95.382 -50.440 55.581
720804 95.416 -50.462 55.568
720805 95.451 -50.484 55.556
720806 95.482 -50.501 55.544
720807 95.513 -50.519 55.529
720808 95.544 -50.536 55.515
720809 95.574 -50.554 55.499
720810 95.601 -50.574 55.482
720811 95.627 -50.594 55.464
720812 95.655 -50.614 55.445
720813 95.681 -50.634 55.424
720814 95.708 -50.654 55.404
720815 95.722 -50.676 55.382
720816 95.736 -50.698 55.362
720817 95.750 -50.720 55.340
720818 95.764 -50.742 55.317
720819 95.778 -50.764 55.294
720882 96.688 -52.138 53.854
720883 96.703 -52.159 53.834
720884 96.718 -52.180 53.814
720885 96.733 -52.202 53.794
720886 96.748 -52.223 53.772
720887 96.763 -52.244 53.751
720888 96.779 -52.265 53.730
720889 96.794 -52.286 53.708
720890 96.809 -52.308 53.685
720891 96.824 -52.329 53.661
720892 96.838 -52.350 53.636
720893 96.854 -52.372 53.612
720894 96.869 -52.393 53.587
720895 96.884 -52.413 53.559
720896 96.905 -52.416 53.533
720897 96.927 -52.419 53.504
720898 96.948 -52.422 53.476
720899 96.969 -52.424 53.449
720900 96.990 -52.427 53.422
720944 97.896 -52.656 52.512
720945 97.917 -52.661 52.488
720946 97.938 -52.666 52.464
720947 97.958 -52.672 52.440
720948 97.979 -52.677 52.416
720949 97.999 -52.682 52.393
720950 98.020 -52.687 52.368
720951 98.040 -52.693 52.343
720952 98.061 -52.698 52.319
720953 98.082 -52.703 52.295
720954 98.102 -52.709 52.271
720955 98.123 -52.714 52.246
720956 98.143 -52.719 52.219
720957 98.164 -52.725 52.192
720958 98.184 -52.730 52.163
720959 98.205 -52.735 52.135
720960 98.234 -52.748 52.107
720961 98.271 -52.765 52.076
720962 98.308 -52.782 52.043
720963 98.345 -52.798 52.010
720964 98.381 -52.815 51.975
720965 98.419 -52.832 51.939
720966 98.455 -52.848 51.903
720967 98.492 -52.865 51.866
720968 98.529 -52.882 51.832
720969 98.566 -52.898 51.797
720970 98.602 -52.915 51.762
720971 98.622 -52.919 51.725
720972 98.632 -52.918 51.685
720973 98.642 -52.916 51.642
720974 98.659 -52.923 51.599
720975 98.679 -52.930 51.558
720976 98.699 -52.938 51.518
720977 98.719 -52.945 51.479
720978 98.739 -52.954 51.439
720979 98.759 -52.961 51.400
720980 98.779 -52.969 51.361
720981 98.799 -52.976 51.321
720982 98.816 -52.976 51.283
720983 98.834 -52.975 51.245
720984 98.852 -52.976 51.205
720985 98.872 -52.988 51.166
720986 98.892 -53.000 51.127
720987 98.913 -53.012 51.088
720988 98.933 -53.024 51.048
720989 98.954 -53.036 51.010
720990 98.974 -53.049 50.973
720991 98.995 -53.061 50.934
720992 99.013 -53.079 50.897
720993 99.030 -53.092 50.859
720994 99.048 -53.105 50.821
720995 99.066 -53.118 50.782
720996 99.083 -53.131 50.743
720997 99.101 -53.144 50.702
720998 99.118 -53.157 50.662
720999 99.136 -53.170 50.621
721000 99.153 -53.183 50.580
721001 99.171 -53.196 50.540
721002 99.189 -53.209 50.499
721003 99.206 -53.222 50.457
721004 99.223 -53.235 50.415
721005 99.241 -53.248 50.374
721006 99.261 -53.259 50.332
721007 99.288 -53.269 50.291
721503 95.361 -50.390 55.611
721504 95.395 -50.412 55.598
721505 95.429 -50.433 55.586
721506 95.464 -50.454 55.575
721507 95.498 -50.475 55.562
721508 95.530 -50.494 55.549
721509 95.561 -50.512 55.535
721510 95.592 -50.529 55.519
721511 95.619 -50.549 55.502
721512 95.647 -50.569 55.484
721513 95.673 -50.588 55.464
721514 95.700 -50.608 55.443
721515 95.717 -50.630 55.422
721516 95.731 -50.652 55.400
721517 95.745 -50.674 55.378
721518 95.759 -50.697 55.356
721519 95.773 -50.719 55.332
721520 95.787 -50.741 55.309
721583 96.708 -52.105 53.855
721584 96.723 -52.126 53.835
721585 96.738 -52.148 53.816
721586 96.753 -52.169 53.796
721587 96.768 -52.190 53.775
721588 96.783 -52.211 53.756
721589 96.798 -52.233 53.735
721590 96.813 -52.254 53.713
721591 96.828 -52.275 53.691
721592 96.843 -52.296 53.667
721593 96.859 -52.318 53.643
721594 96.874 -52.339 53.619
721595 96.889 -52.360 53.595
721596 96.904 -52.381 53.568
721597 96.919 -52.402 53.541
721598 96.940 -52.406 53.513
721599 96.961 -52.409 53.485
721600 96.982 -52.412 53.458
721645 97.912 -52.636 52.520
721646 97.932 -52.641 52.495
721647 97.952 -52.647 52.470
721648 97.973 -52.652 52.446
721649 97.994 -52.657 52.422
721650 98.014 -52.662 52.397
721651 98.035 -52.668 52.373
721652 98.056 -52.673 52.348
721653 98.076 -52.678 52.323
721654 98.096 -52.684 52.298
721655 98.117 -52.689 52.273
721656 98.138 -52.694 52.249
721657 98.158 -52.700 52.223
721658 98.179 -52.705 52.196
721659 98.200 -52.710 52.168
721660 98.220 -52.716 52.139
721661 98.243 -52.724 52.110
721662 98.277 -52.741 52.077
721663 98.314 -52.757 52.043
721664 98.351 -52.774 52.009
721665 98.387 -52.791 51.974
721666 98.424 -52.807 51.940
721667 98.461 -52.824 51.906
721668 98.498 -52.841 51.873
721669 98.534 -52.857 51.841
721670 98.571 -52.874 51.807
721671 98.608 -52.891 51.772
721672 98.630 -52.897 51.735
721673 98.640 -52.896 51.694
721674 98.650 -52.894 51.649
721675 98.663 -52.895 51.604
721676 98.683 -52.903 51.562
721677 98.703 -52.910 51.523
721678 98.723 -52.918 51.483
721679 98.742 -52.926 51.443
721680 98.763 -52.934 51.403
721681 98.783 -52.941 51.363
721682 98.802 -52.949 51.324
721683 98.822 -52.957 51.285
721684 98.841 -52.961 51.248
721685 98.859 -52.960 51.209
721686 98.879 -52.972 51.169
721687 98.900 -52.984 51.130
721688 98.920 -52.996 51.090
721689 98.940 -53.008 51.051
721690 98.960 -53.021 51.012
721691 98.981 -53.033 50.975
721692 99.001 -53.045 50.936
721693 99.022 -53.056 50.899
721694 99.041 -53.068 50.861
721695 99.059 -53.081 50.823
721696 99.076 -53.094 50.784
721697 99.094 -53.107 50.744
721698 99.112 -53.120 50.703
721699 99.129 -53.133 50.664
721700 99.146 -53.146 50.624
721701 99.164 -53.159 50.582
721702 99.182 -53.172 50.542
721703 99.199 -53.185 50.501
721704 99.216 -53.198 50.460
721705 99.234 -53.211 50.418
721706 99.252 -53.223 50.377
721707 99.269 -53.236 50.335
721708 99.289 -53.248 50.294
722203 95.339 -50.339 55.637
722204 95.374 -50.361 55.624
722205 95.408 -50.382 55.612
722206 95.442 -50.403 55.601
722207 95.476 -50.425 55.590
722208 95.511 -50.447 55.579
722209 95.545 -50.468 55.567
722210 95.578 -50.488 55.553
722211 95.609 -50.505 55.538
722212 95.638 -50.524 55.522
722213 95.665 -50.544 55.504
722214 95.692 -50.563 55.483
722215 95.712 -50.585 55.461
722216 95.726 -50.607 55.438
722217 95.740 -50.629 55.416
722218 95.754 -50.650 55.393
722219 95.768 -50.673 55.370
722220 95.782 -50.695 55.346
722221 95.796 -50.717 55.322
722222 95.810 -50.740 55.299
722285 96.743 -52.093 53.837
722286 96.758 -52.114 53.817
722287 96.773 -52.136 53.797
722288 96.788 -52.157 53.777
722289 96.803 -52.178 53.759
722290 96.818 -52.199 53.738
722291 96.833 -52.221 53.717
722292 96.849 -52.242 53.695
722293 96.863 -52.263 53.672
722294 96.878 -52.284 53.649
722295 96.893 -52.305 53.625
722296 96.908 -52.327 53.601
722297 96.924 -52.349 53.575
722298 96.939 -52.370 53.548
722299 96.954 -52.391 53.521
722300 96.974 -52.396 53.493
722301 96.995 -52.399 53.466
722346 97.926 -52.616 52.527
722347 97.947 -52.622 52.501
722348 97.967 -52.627 52.476
722349 97.988 -52.632 52.451
722350 98.009 -52.637 52.426
722351 98.029 -52.643 52.403
722352 98.050 -52.648 52.378
722353 98.070 -52.653 52.351
722354 98.091 -52.659 52.326
722355 98.111 -52.664 52.301
722356 98.132 -52.669 52.277
722357 98.153 -52.675 52.251
722358 98.173 -52.680 52.224
722359 98.194 -52.685 52.198
722360 98.214 -52.691 52.169
722361 98.235 -52.696 52.140
722362 98.255 -52.701 52.110
722363 98.283 -52.716 52.078
722364 98.319 -52.733 52.044
722365 98.356 -52.750 52.009
722366 98.393 -52.766 51.975
722367 98.430 -52.783 51.942
722368 98.466 -52.800 51.911
722369 98.504 -52.816 51.881
722370 98.541 -52.833 51.850
722371 98.577 -52.850 51.816
722372 98.614 -52.866 51.780
722373 98.638 -52.875 51.741
722374 98.648 -52.873 51.699
722375 98.658 -52.872 51.653
722376 98.668 -52.870 51.609
722377 98.686 -52.876 51.566
722378 98.706 -52.883 51.526
722379 98.726 -52.891 51.487
722380 98.746 -52.899 51.446
722381 98.766 -52.907 51.406
722382 98.786 -52.914 51.366
722383 98.806 -52.922 51.327
722384 98.826 -52.930 51.288
722385 98.846 -52.938 51.250
722386 98.866 -52.945 51.213
722387 98.886 -52.955 51.173
722388 98.907 -52.966 51.132
722389 98.927 -52.977 51.092
722390 98.947 -52.988 51.053
722391 98.968 -52.999 51.015
722392 98.988 -53.010 50.977
722393 99.009 -53.020 50.938
722394 99.030 -53.030 50.901
722395 99.054 -53.042 50.863
722396 99.069 -53.057 50.824
722397 99.087 -53.070 50.785
722398 99.104 -53.083 50.745
722399 99.122 -53.096 50.705
722400 99.139 -53.109 50.666
722401 99.157 -53.122 50.626
722402 99.174 -53.135 50.585
722403 99.192 -53.148 50.544
722404 99.209 -53.161 50.503
722405 99.227 -53.173 50.462
722406 99.244 -53.186 50.421
722407 99.262 -53.199 50.380
722408 99.280 -53.212 50.339
722409 99.297 -53.225 50.297
722904 95.352 -50.310 55.648
722905 95.387 -50.331 55.635
722906 95.421 -50.353 55.624
722907 95.455 -50.374 55.614
722908 95.489 -50.396 55.603
722909 95.524 -50.417 55.592
722910 95.558 -50.438 55.580
722911 95.592 -50.459 55.567
722912 95.626 -50.481 55.552
722913 95.657 -50.499 55.536
722914 95.684 -50.518 55.518
722915 95.707 -50.538 55.498
722916 95.721 -50.561 55.476
722917 95.735 -50.583 55.454
722918 95.749 -50.605 55.430
722919 95.763 -50.628 55.406
722920 95.777 -50.650 55.384
722921 95.791 -50.672 55.359
722922 95.805 -50.694 55.335
722923 95.819 -50.716 55.310
722987 96.778 -52.081 53.818
722988 96.793 -52.102 53.799
722989 96.808 -52.124 53.780
722990 96.823 -52.145 53.761
722991 96.838 -52.167 53.741
722992 96.853 -52.188 53.721
722993 96.868 -52.209 53.699
722994 96.883 -52.230 53.676
722995 96.898 -52.252 53.654
722996 96.913 -52.273 53.631
722997 96.929 -52.294 53.607
722998 96.943 -52.315 53.583
722999 96.958 -52.336 53.556
723000 96.973 -52.358 53.529
723001 96.989 -52.379 53.502
723047 97.941 -52.597 52.534
723048 97.962 -52.602 52.508
723049 97.983 -52.607 52.482
723050 98.003 -52.612 52.456
723051 98.023 -52.618 52.432
723052 98.044 -52.623 52.407
723053 98.065 -52.628 52.382
723054 98.085 -52.634 52.355
723055 98.106 -52.639 52.330
723056 98.127 -52.644 52.304
723057 98.147 -52.650 52.279
723058 98.167 -52.655 52.253
723059 98.188 -52.660 52.225
723060 98.209 -52.666 52.198
723061 98.229 -52.671 52.169
723062 98.250 -52.676 52.140
723063 98.270 -52.681 52.110
723064 98.295 -52.692 52.078
723065 98.325 -52.709 52.045
723066 98.362 -52.725 52.011
723067 98.399 -52.742 51.977
723068 98.436 -52.759 51.945
723069 98.473 -52.775 51.915
723070 98.509 -52.792 51.887
723071 98.546 -52.809 51.855
723072 98.583 -52.825 51.820
723073 98.620 -52.842 51.783
723074 98.646 -52.852 51.744
723075 98.656 -52.850 51.701
723076 98.666 -52.849 51.656
723077 98.677 -52.847 51.612
723078 98.690 -52.848 51.570
723079 98.710 -52.856 51.530
723080 98.729 -52.864 51.490
723081 98.749 -52.872 51.449
723082 98.769 -52.879 51.408
723083 98.789 -52.887 51.368
723084 98.809 -52.895 51.330
723085 98.829 -52.903 51.291
723086 98.849 -52.910 51.255
723087 98.872 -52.919 51.216
723088 98.894 -52.931 51.176
723089 98.914 -52.941 51.135
723090 98.934 -52.952 51.095
723091 98.955 -52.963 51.056
723092 98.975 -52.973 51.018
723093 98.996 -52.985 50.979
723094 99.016 -52.995 50.941
723095 99.043 -53.003 50.903
723096 99.064 -53.017 50.865
723097 99.082 -53.029 50.826
723098 99.098 -53.045 50.788
723099 99.115 -53.059 50.748
723100 99.132 -53.072 50.708
723101 99.150 -53.085 50.668
723102 99.167 -53.098 50.629
723103 99.185 -53.111 50.588
723104 99.203 -53.123 50.547
723105 99.220 -53.136 50.507
723106 99.237 -53.149 50.466
723107 99.255 -53.162 50.425
723108 99.273 -53.175 50.384
723109 99.290 -53.188 50.343
723110 99.307 -53.201 50.302
723605 95.365 -50.281 55.656
723606 95.399 -50.302 55.644
723607 95.433 -50.324 55.634
723608 95.468 -50.345 55.625
723609 95.502 -50.366 55.614
723610 95.536 -50.387 55.603
723611 95.570 -50.409 55.591
723612 95.605 -50.431 55.579
723613 95.639 -50.452 55.564
723614 95.673 -50.473 55.548
723615 95.702 -50.493 55.531
723616 95.716 -50.516 55.510
723617 95.730 -50.538 55.488
723618 95.744 -50.560 55.465
723619 95.759 -50.581 55.441
723620 95.774 -50.602 55.418
723621 95.789 -50.623 55.394
723622 95.804 -50.645 55.368
723623 95.819 -50.666 55.343
723624 95.834 -50.688 55.319
723689 96.813 -52.070 53.800
723690 96.828 -52.091 53.780
723691 96.843 -52.112 53.762
723692 96.858 -52.133 53.743
723693 96.873 -52.155 53.723
723694 96.888 -52.176 53.702
723695 96.903 -52.197 53.680
723696 96.918 -52.218 53.658
723697 96.933 -52.239 53.637
723698 96.948 -52.261 53.614
723699 96.963 -52.282 53.590
723700 96.978 -52.303 53.565
723701 96.994 -52.325 53.538
723748 97.956 -52.577 52.541
723749 97.977 -52.582 52.514
723750 97.997 -52.587 52.488
723751 98.018 -52.593 52.462
723752 98.038 -52.598 52.437
723753 98.059 -52.603 52.412
723754 98.080 -52.609 52.385
723755 98.100 -52.614 52.359
723756 98.121 -52.619 52.334
723757 98.141 -52.625 52.307
723758 98.162 -52.630 52.281
723759 98.182 -52.635 52.255
723760 98.203 -52.641 52.226
723761 98.223 -52.646 52.198
723762 98.244 -52.651 52.169
723763 98.265 -52.656 52.140
723764 98.285 -52.662 52.110
723765 98.306 -52.668 52.079
723766 98.335 -52.684 52.046
723767 98.368 -52.701 52.013
723768 98.405 -52.718 51.980
723769 98.441 -52.735 51.948
723770 98.478 -52.751 51.919
723771 98.515 -52.768 51.890
723772 98.552 -52.785 51.857
723773 98.589 -52.801 51.821
723774 98.626 -52.818 51.784
723775 98.654 -52.829 51.745
723776 98.664 -52.828 51.702
723777 98.675 -52.826 51.658
723778 98.685 -52.825 51.614
723779 98.695 -52.823 51.573
723780 98.713 -52.828 51.533
723781 98.733 -52.837 51.493
723782 98.753 -52.844 51.453
723783 98.773 -52.852 51.411
723784 98.793 -52.860 51.371
723785 98.812 -52.868 51.332
723786 98.833 -52.876 51.294
723787 98.853 -52.883 51.258
723788 98.880 -52.894 51.219
723789 98.904 -52.906 51.179
723790 98.921 -52.916 51.138
723791 98.942 -52.927 51.098
723792 98.962 -52.938 51.058
723793 98.983 -52.948 51.021
723794 99.003 -52.960 50.982
723795 99.024 -52.970 50.944
723796 99.055 -52.975 50.907
723797 99.074 -52.991 50.868
723798 99.096 -53.002 50.830
723799 99.111 -53.018 50.791
723800 99.127 -53.033 50.751
723801 99.143 -53.048 50.711
723802 99.160 -53.060 50.672
723803 99.178 -53.073 50.632
723804 99.196 -53.086 50.592
723805 99.213 -53.099 50.551
723806 99.230 -53.112 50.511
723807 99.248 -53.125 50.470
723808 99.266 -53.139 50.429
723809 99.283 -53.151 50.388
723810 99.300 -53.164 50.348
723811 99.318 -53.177 50.307
723812 99.336 -53.190 50.266
724306 95.378 -50.252 55.662
724307 95.412 -50.273 55.651
724308 95.449 -50.294 55.641
724309 95.486 -50.315 55.632
724310 95.524 -50.337 55.622
724311 95.560 -50.358 55.611
724312 95.598 -50.380 55.599
724313 95.634 -50.401 55.588
724314 95.672 -50.422 55.574
724315 95.678 -50.460 55.558
724316 95.709 -50.468 55.541
724317 95.734 -50.484 55.520
724318 95.749 -50.506 55.497
724319 95.764 -50.527 55.474
724320 95.779 -50.548 55.449
724321 95.794 -50.569 55.426
724322 95.809 -50.591 55.402
724323 95.824 -50.612 55.376
724324 95.839 -50.633 55.350
724325 95.854 -50.654 55.325
724391 96.848 -52.058 53.780
724392 96.863 -52.079 53.761
724393 96.878 -52.101 53.742
724394 96.893 -52.122 53.723
724395 96.908 -52.143 53.703
724396 96.923 -52.164 53.682
724397 96.938 -52.186 53.660
724398 96.953 -52.207 53.640
724399 96.968 -52.228 53.617
724400 96.983 -52.249 53.595
724449 97.971 -52.557 52.548
724450 97.992 -52.562 52.520
724451 98.012 -52.568 52.493
724452 98.033 -52.573 52.467
724453 98.053 -52.578 52.441
724454 98.074 -52.584 52.417
724455 98.094 -52.589 52.389
724456 98.115 -52.594 52.363
724457 98.136 -52.600 52.336
724458 98.156 -52.605 52.310
724459 98.176 -52.610 52.284
724460 98.197 -52.615 52.258
724461 98.218 -52.621 52.229
724462 98.238 -52.626 52.200
724463 98.259 -52.631 52.170
724464 98.280 -52.637 52.141
724465 98.300 -52.641 52.112
724466 98.320 -52.647 52.081
724467 98.347 -52.660 52.048
724468 98.375 -52.677 52.014
724469 98.411 -52.694 51.983
724470 98.447 -52.710 51.951
724471 98.484 -52.727 51.922
724472 98.521 -52.744 51.891
724473 98.558 -52.760 51.857
724474 98.594 -52.777 51.820
724475 98.631 -52.794 51.783
724476 98.662 -52.807 51.744
724477 98.672 -52.805 51.703
724478 98.683 -52.803 51.660
724479 98.693 -52.802 51.617
724480 98.704 -52.800 51.575
724481 98.716 -52.802 51.535
724482 98.736 -52.810 51.496
724483 98.756 -52.817 51.457
724484 98.776 -52.825 51.415
724485 98.796 -52.833 51.375
724486 98.816 -52.841 51.335
724487 98.836 -52.848 51.297
724488 98.858 -52.857 51.260
724489 98.889 -52.869 51.222
724490 98.915 -52.882 51.182
724491 98.932 -52.892 51.141
724492 98.949 -52.902 51.101
724493 98.970 -52.913 51.061
724494 98.990 -52.923 51.023
724495 99.010 -52.935 50.985
724496 99.036 -52.942 50.948
724497 99.066 -52.947 50.910
724498 99.083 -52.968 50.871
724499 99.106 -52.976 50.833
724500 99.125 -52.990 50.794
724501 99.140 -53.006 50.755
724502 99.155 -53.021 50.715
724503 99.171 -53.036 50.676
724504 99.189 -53.049 50.636
724505 99.206 -53.062 50.596
724506 99.223 -53.076 50.556
724507 99.241 -53.089 50.515
724508 99.259 -53.101 50.474
724509 99.276 -53.114 50.434
724510 99.293 -53.127 50.393
724511 99.311 -53.140 50.353
724512 99.329 -53.153 50.312
724513 99.346 -53.166 50.271
724514 99.364 -53.179 50.231
724515 99.381 -53.192 50.190
725007 95.425 -50.221 55.665
725008 95.462 -50.243 55.654
725009 95.499 -50.264 55.644
725010 95.536 -50.286 55.635
725011 95.573 -50.307 55.626
725012 95.610 -50.328 55.616
725013 95.648 -50.350 55.605
725014 95.654 -50.389 55.592
725015 95.645 -50.434 55.578
725016 95.678 -50.435 55.562
725017 95.708 -50.445 55.545
725018 95.735 -50.460 55.525
725019 95.761 -50.476 55.503
725020 95.783 -50.494 55.480
725021 95.799 -50.515 55.455
725022 95.814 -50.536 55.432
725023 95.829 -50.557 55.407
725024 95.844 -50.578 55.382
725025 95.859 -50.600 55.355
725026 95.874 -50.622 55.330
725094 96.897 -52.067 53.741
725095 96.912 -52.089 53.722
725096 96.928 -52.110 53.702
725097 96.943 -52.131 53.682
725098 96.958 -52.152 53.661
725099 96.973 -52.174 53.641
725150 97.986 -52.537 52.555
725151 98.006 -52.543 52.527
725152 98.027 -52.548 52.499
725153 98.048 -52.553 52.473
725154 98.068 -52.559 52.446
725155 98.089 -52.564 52.421
725156 98.109 -52.569 52.393
725157 98.130 -52.575 52.366
725158 98.150 -52.580 52.340
725159 98.171 -52.585 52.313
725160 98.192 -52.590 52.288
725161 98.212 -52.596 52.261
725162 98.233 -52.601 52.232
725163 98.253 -52.606 52.203
725164 98.274 -52.611 52.173
725165 98.294 -52.616 52.144
725166 98.315 -52.622 52.114
725167 98.336 -52.627 52.083
725168 98.358 -52.635 52.050
725169 98.387 -52.653 52.017
725170 98.416 -52.669 51.985
725171 98.453 -52.686 51.953
725172 98.490 -52.703 51.922
725173 98.526 -52.719 51.890
725174 98.563 -52.736 51.854
725175 98.600 -52.753 51.817
725176 98.637 -52.769 51.780
725177 98.670 -52.784 51.742
725178 98.681 -52.782 51.702
725179 98.691 -52.781 51.661
725180 98.702 -52.780 51.619
725181 98.712 -52.778 51.578
725182 98.722 -52.777 51.538
725183 98.739 -52.782 51.499
725184 98.760 -52.790 51.459
725185 98.780 -52.797 51.419
725186 98.799 -52.806 51.379
725187 98.819 -52.813 51.340
725188 98.839 -52.821 51.301
725189 98.866 -52.832 51.263
725190 98.898 -52.845 51.224
725191 98.926 -52.857 51.184
725192 98.943 -52.867 51.144
725193 98.959 -52.878 51.104
725194 98.977 -52.888 51.065
725195 98.997 -52.898 51.028
725196 99.018 -52.910 50.989
725197 99.048 -52.915 50.952
725198 99.078 -52.919 50.913
725199 99.092 -52.944 50.875
725200 99.117 -52.951 50.838
725201 99.138 -52.963 50.799
725202 99.153 -52.978 50.759
725203 99.168 -52.994 50.720
725204 99.184 -53.010 50.681
725205 99.199 -53.025 50.641
725206 99.216 -53.038 50.600
725207 99.234 -53.051 50.561
725208 99.252 -53.064 50.520
725209 99.269 -53.077 50.480
725210 99.287 -53.090 50.439
725211 99.304 -53.103 50.399
725212 99.322 -53.116 50.358
725213 99.339 -53.129 50.317
725214 99.357 -53.142 50.276
725215 99.374 -53.154 50.236
725216 99.392 -53.167 50.196
725217 99.409 -53.180 50.155
725708 95.475 -50.191 55.666
725709 95.512 -50.213 55.655
725710 95.549 -50.234 55.644
725711 95.586 -50.256 55.636
725712 95.623 -50.277 55.627
725713 95.631 -50.315 55.619
725714 95.618 -50.366 55.608
725715 95.612 -50.409 55.595
725716 95.644 -50.409 55.580
725717 95.678 -50.410 55.564
725718 95.708 -50.422 55.547
725719 95.734 -50.437 55.527
725720 95.761 -50.453 55.504
725721 95.787 -50.469 55.481
725722 95.813 -50.484 55.457
725723 95.833 -50.503 55.433
725724 95.848 -50.525 55.409
725725 95.864 -50.546 55.384
725726 95.879 -50.567 55.358
725727 95.894 -50.588 55.333
725851 98.001 -52.518 52.562
725852 98.021 -52.523 52.533
725853 98.042 -52.528 52.505
725854 98.063 -52.534 52.478
725855 98.083 -52.539 52.451
725856 98.103 -52.544 52.427
725857 98.124 -52.550 52.398
725858 98.145 -52.555 52.370
725859 98.165 -52.560 52.343
725860 98.186 -52.565 52.317
725861 98.207 -52.571 52.291
725862 98.227 -52.575 52.265
725863 98.247 -52.581 52.235
725864 98.268 -52.586 52.206
725865 98.289 -52.591 52.176
725866 98.309 -52.597 52.147
725867 98.330 -52.602 52.117
725868 98.351 -52.607 52.085
725869 98.371 -52.612 52.053
725870 98.398 -52.628 52.020
725871 98.427 -52.645 51.988
725872 98.459 -52.662 51.955
725873 98.496 -52.678 51.922
725874 98.532 -52.695 51.889
725875 98.569 -52.712 51.852
725876 98.606 -52.728 51.815
725877 98.643 -52.745 51.777
725878 98.678 -52.762 51.740
725879 98.689 -52.760 51.701
725880 98.699 -52.759 51.662
725881 98.710 -52.757 51.621
725882 98.720 -52.756 51.581
725883 98.730 -52.754 51.541
725884 98.743 -52.755 51.502
725885 98.763 -52.763 51.462
725886 98.783 -52.770 51.422
725887 98.803 -52.778 51.383
725888 98.823 -52.786 51.343
725889 98.844 -52.794 51.304
725890 98.875 -52.807 51.265
725891 98.906 -52.820 51.226
725892 98.937 -52.833 51.187
725893 98.953 -52.843 51.147
725894 98.971 -52.854 51.107
725895 98.987 -52.863 51.070
725896 99.005 -52.873 51.032
725897 99.029 -52.882 50.994
725898 99.060 -52.887 50.956
725899 99.090 -52.891 50.917
725900 99.101 -52.919 50.879
725901 99.128 -52.926 50.842
725902 99.149 -52.937 50.803
725903 99.167 -52.951 50.764
725904 99.182 -52.966 50.725
725905 99.197 -52.982 50.685
725906 99.213 -52.998 50.645
725907 99.228 -53.013 50.605
725908 99.244 -53.027 50.565
725909 99.262 -53.040 50.525
725910 99.280 -53.053 50.485
725911 99.297 -53.066 50.445
725912 99.314 -53.079 50.404
725913 99.332 -53.092 50.363
725914 99.350 -53.104 50.322
725915 99.367 -53.117 50.282
725916 99.384 -53.130 50.242
725917 99.402 -53.143 50.201
725918 99.420 -53.156 50.160
725919 99.437 -53.169 50.119
726410 95.562 -50.183 55.654
726411 95.599 -50.204 55.644
726412 95.609 -50.241 55.635
726413 95.595 -50.293 55.626
726414 95.582 -50.343 55.618
726415 95.579 -50.383 55.607
726416 95.611 -50.384 55.594
726417 95.644 -50.384 55.580
726418 95.677 -50.385 55.564
726419 95.707 -50.398 55.546
726420 95.734 -50.415 55.527
726421 95.760 -50.431 55.504
726422 95.787 -50.446 55.481
726423 95.813 -50.462 55.458
726424 95.839 -50.478 55.433
726425 95.866 -50.493 55.409
726426 95.883 -50.513 55.385
726427 95.898 -50.534 55.359
726553 98.036 -52.503 52.540
726554 98.057 -52.509 52.512
726555 98.077 -52.514 52.484
726556 98.098 -52.519 52.458
726557 98.119 -52.525 52.431
726558 98.139 -52.530 52.403
726559 98.160 -52.535 52.375
726560 98.180 -52.540 52.348
726561 98.201 -52.545 52.321
726562 98.221 -52.550 52.295
726563 98.242 -52.556 52.268
726564 98.263 -52.561 52.238
726565 98.283 -52.566 52.209
726566 98.304 -52.572 52.179
726567 98.324 -52.577 52.150
726568 98.345 -52.582 52.120
726569 98.365 -52.587 52.088
726570 98.386 -52.593 52.056
726571 98.410 -52.603 52.022
726572 98.439 -52.621 51.989
726573 98.468 -52.637 51.957
726574 98.501 -52.654 51.923
726575 98.538 -52.671 51.888
726576 98.575 -52.687 51.851
726577 98.612 -52.704 51.814
726578 98.648 -52.721 51.777
726579 98.685 -52.738 51.740
726580 98.697 -52.738 51.702
726581 98.708 -52.736 51.663
726582 98.718 -52.735 51.622
726583 98.728 -52.733 51.583
726584 98.738 -52.731 51.545
726585 98.748 -52.730 51.506
726586 98.766 -52.735 51.466
726587 98.786 -52.744 51.426
726588 98.806 -52.751 51.386
726589 98.826 -52.759 51.346
726590 98.852 -52.769 51.307
726591 98.883 -52.782 51.268
726592 98.915 -52.795 51.228
726593 98.946 -52.808 51.189
726594 98.964 -52.819 51.151
726595 98.981 -52.828 51.112
726596 98.998 -52.839 51.073
726597 99.015 -52.849 51.037
726598 99.041 -52.854 50.998
726599 99.072 -52.859 50.960
726600 99.102 -52.863 50.922
726601 99.111 -52.895 50.884
726602 99.138 -52.901 50.846
726603 99.159 -52.912 50.807
726604 99.181 -52.923 50.768
726605 99.196 -52.938 50.729
726606 99.211 -52.954 50.690
726607 99.226 -52.970 50.649
726608 99.241 -52.985 50.610
726609 99.257 -53.001 50.570
726610 99.273 -53.016 50.530
726611 99.290 -53.029 50.490
726612 99.307 -53.042 50.450
726613 99.325 -53.054 50.409
726614 99.343 -53.067 50.368
726615 99.360 -53.080 50.328
726616 99.377 -53.093 50.287
726617 99.395 -53.106 50.247
726618 99.413 -53.120 50.206
726619 99.430 -53.132 50.165
726620 99.448 -53.145 50.124
726621 99.466 -53.158 50.084
726622 99.488 -53.169 50.043
727111 95.587 -50.168 55.652
727112 95.573 -50.219 55.642
727113 95.559 -50.270 55.633
727114 95.545 -50.321 55.624
727115 95.545 -50.357 55.616
727116 95.578 -50.358 55.604
727117 95.611 -50.359 55.591
727118 95.644 -50.359 55.577
727119 95.677 -50.360 55.560
727120 95.706 -50.375 55.543
727121 95.733 -50.392 55.524
727122 95.760 -50.408 55.502
727123 95.786 -50.423 55.479
727124 95.812 -50.439 55.456
727125 95.839 -50.454 55.431
727126 95.865 -50.470 55.407
727127 95.891 -50.486 55.384
727128 95.918 -50.501 55.359
727255 98.072 -52.489 52.519
727256 98.092 -52.494 52.491
727257 98.113 -52.499 52.464
727258 98.134 -52.505 52.436
727259 98.154 -52.509 52.408
727260 98.174 -52.515 52.380
727261 98.195 -52.520 52.353
727262 98.216 -52.525 52.326
727263 98.236 -52.531 52.299
727264 98.257 -52.536 52.272
727265 98.277 -52.541 52.242
727266 98.298 -52.546 52.212
727267 98.318 -52.552 52.183
727268 98.339 -52.557 52.153
727269 98.360 -52.562 52.123
727270 98.380 -52.568 52.091
727271 98.401 -52.573 52.058
727272 98.422 -52.579 52.025
727273 98.450 -52.596 51.992
727274 98.480 -52.613 51.959
727275 98.508 -52.630 51.925
727276 98.544 -52.647 51.889
727277 98.581 -52.663 51.853
727278 98.618 -52.680 51.815
727279 98.654 -52.697 51.778
727280 98.691 -52.713 51.741
727281 98.705 -52.715 51.703
727282 98.716 -52.713 51.665
727283 98.726 -52.712 51.625
727284 98.736 -52.710 51.586
727285 98.746 -52.709 51.547
727286 98.757 -52.707 51.509
727287 98.770 -52.708 51.470
727288 98.790 -52.716 51.429
727289 98.809 -52.724 51.389
727290 98.830 -52.731 51.349
727291 98.861 -52.744 51.310
727292 98.892 -52.757 51.270
727293 98.923 -52.770 51.232
727294 98.954 -52.784 51.193
727295 98.975 -52.794 51.154
727296 98.992 -52.804 51.116
727297 99.009 -52.815 51.078
727298 99.031 -52.822 51.040
727299 99.058 -52.826 51.003
727300 99.084 -52.832 50.965
727301 99.114 -52.837 50.927
727302 99.120 -52.871 50.888
727303 99.147 -52.877 50.849
727304 99.170 -52.886 50.811
727305 99.192 -52.897 50.772
727306 99.209 -52.911 50.733
727307 99.224 -52.926 50.694
727308 99.239 -52.942 50.654
727309 99.255 -52.958 50.614
727310 99.270 -52.973 50.575
727311 99.285 -52.989 50.535
727312 99.300 -53.004 50.494
727313 99.318 -53.017 50.454
727314 99.336 -53.030 50.413
727315 99.353 -53.043 50.373
727316 99.371 -53.057 50.333
727317 99.388 -53.070 50.292
727318 99.406 -53.082 50.251
727319 99.423 -53.095 50.210
727320 99.441 -53.108 50.169
727321 99.458 -53.121 50.128
727322 99.476 -53.134 50.088
727323 99.493 -53.147 50.047
727324 99.516 -53.158 50.007
727813 95.523 -50.247 55.640
727814 95.509 -50.298 55.631
727815 95.512 -50.331 55.621
727816 95.545 -50.332 55.612
727817 95.578 -50.333 55.600
727818 95.611 -50.334 55.586
727819 95.643 -50.334 55.571
727820 95.676 -50.335 55.555
727821 95.705 -50.351 55.538
727822 95.733 -50.369 55.520
727823 95.759 -50.384 55.498
727824 95.785 -50.400 55.474
727825 95.812 -50.416 55.451
727826 95.838 -50.431 55.428
727827 95.865 -50.447 55.404
727828 95.891 -50.462 55.380
727957 98.107 -52.474 52.498
727958 98.128 -52.479 52.470
727959 98.148 -52.484 52.441
727960 98.169 -52.490 52.413
727961 98.190 -52.495 52.386
727962 98.210 -52.500 52.358
727963 98.230 -52.506 52.330
727964 98.251 -52.511 52.303
727965 98.272 -52.516 52.276
727966 98.292 -52.521 52.246
727967 98.313 -52.527 52.216
727968 98.334 -52.532 52.186
727969 98.354 -52.537 52.156
727970 98.374 -52.543 52.126
727971 98.395 -52.548 52.094
727972 98.416 -52.553 52.061
727973 98.436 -52.559 52.028
727974 98.462 -52.572 51.995
727975 98.491 -52.588 51.961
727976 98.520 -52.606 51.927
727977 98.550 -52.622 51.891
727978 98.586 -52.639 51.855
727979 98.623 -52.656 51.818
727980 98.660 -52.672 51.780
727981 98.697 -52.689 51.744
727982 98.714 -52.692 51.705
727983 98.724 -52.691 51.667
727984 98.734 -52.689 51.629
727985 98.744 -52.687 51.589
727986 98.755 -52.686 51.551
727987 98.765 -52.684 51.512
727988 98.775 -52.683 51.473
727989 98.793 -52.689 51.433
727990 98.813 -52.697 51.393
727991 98.838 -52.706 51.352
727992 98.869 -52.720 51.313
727993 98.901 -52.733 51.274
727994 98.932 -52.746 51.235
727995 98.963 -52.759 51.197
727996 98.986 -52.770 51.158
727997 99.003 -52.780 51.119
727998 99.023 -52.788 51.081
727999 99.049 -52.794 51.045
728000 99.075 -52.798 51.007
728001 99.101 -52.803 50.970
728002 99.124 -52.812 50.931
728003 99.130 -52.847 50.892
728004 99.156 -52.853 50.854
728005 99.181 -52.861 50.816
728006 99.202 -52.872 50.777
728007 99.222 -52.883 50.738
728008 99.238 -52.899 50.698
728009 99.253 -52.915 50.658
728010 99.268 -52.930 50.619
728011 99.283 -52.946 50.579
728012 99.299 -52.962 50.539
728013 99.314 -52.977 50.499
728014 99.329 -52.993 50.459
728015 99.346 -53.007 50.418
728016 99.364 -53.020 50.378
728017 99.381 -53.032 50.338
728018 99.399 -53.045 50.297
728019 99.416 -53.058 50.256
728020 99.434 -53.071 50.215
728021 99.451 -53.084 50.174
728022 99.469 -53.097 50.133
728023 99.486 -53.110 50.093
728024 99.504 -53.123 50.052
728025 99.521 -53.136 50.012
728514 95.473 -50.276 55.638
728515 95.479 -50.306 55.628
728516 95.512 -50.306 55.617
728517 95.545 -50.307 55.607
728518 95.578 -50.308 55.595
728519 95.610 -50.309 55.580
728520 95.643 -50.309 55.564
728521 95.676 -50.310 55.549
728522 95.705 -50.328 55.533
728523 95.732 -50.346 55.514
728524 95.759 -50.362 55.492
728525 95.785 -50.378 55.469
728526 95.811 -50.394 55.446
728527 95.838 -50.409 55.423
728528 95.864 -50.425 55.399
728662 98.204 -52.475 52.391
728663 98.225 -52.481 52.363
728664 98.245 -52.486 52.335
728665 98.266 -52.491 52.308
728666 98.287 -52.496 52.281
728667 98.307 -52.502 52.251
728668 98.327 -52.507 52.220
728669 98.348 -52.512 52.190
728670 98.369 -52.518 52.160
728671 98.389 -52.523 52.130
728672 98.410 -52.528 52.098
728673 98.431 -52.534 52.065
728674 98.451 -52.539 52.032
728675 98.474 -52.547 51.999
728676 98.503 -52.564 51.965
728677 98.531 -52.581 51.931
728678 98.560 -52.598 51.894
728679 98.592 -52.615 51.858
728680 98.629 -52.631 51.821
728681 98.666 -52.648 51.784
728682 98.703 -52.665 51.747
728683 98.722 -52.669 51.709
728684 98.732 -52.668 51.670
728685 98.742 -52.666 51.632
728686 98.752 -52.665 51.593
728687 98.763 -52.663 51.555
728688 98.773 -52.662 51.516
728689 98.783 -52.661 51.477
728690 98.796 -52.662 51.437
728691 98.816 -52.669 51.397
728692 98.847 -52.682 51.356
728693 98.878 -52.695 51.317
728694 98.909 -52.708 51.277
728695 98.940 -52.721 51.239
728696 98.972 -52.734 51.200
728697 98.997 -52.746 51.162
728698 99.014 -52.756 51.123
728699 99.040 -52.760 51.085
728700 99.066 -52.766 51.049
728701 99.088 -52.775 51.011
728702 99.108 -52.786 50.974
728703 99.128 -52.797 50.936
728704 99.140 -52.821 50.897
728705 99.165 -52.828 50.859
728706 99.191 -52.835 50.820
728707 99.212 -52.847 50.782
728708 99.234 -52.857 50.742
728709 99.251 -52.872 50.703
728710 99.267 -52.887 50.663
728711 99.282 -52.903 50.624
728712 99.297 -52.919 50.584
728713 99.312 -52.934 50.544
728714 99.327 -52.950 50.504
728715 99.343 -52.965 50.464
728716 99.358 -52.981 50.423
728717 99.374 -52.995 50.383
728718 99.392 -53.008 50.343
728719 99.409 -53.021 50.303
728720 99.427 -53.034 50.261
728721 99.444 -53.047 50.220
728722 99.462 -53.060 50.179
728723 99.479 -53.073 50.138
728724 99.497 -53.086 50.098
728725 99.514 -53.098 50.057
728726 99.532 -53.111 50.016
728727 99.550 -53.124 49.976
729216 95.479 -50.281 55.623
729217 95.512 -50.281 55.614
729218 95.544 -50.282 55.603
729219 95.577 -50.283 55.589
729220 95.610 -50.284 55.574
729221 95.643 -50.284 55.559
729222 95.676 -50.285 55.542
729223 95.704 -50.304 55.526
729224 95.732 -50.324 55.507
729225 95.758 -50.339 55.485
729226 95.784 -50.355 55.463
729227 95.811 -50.370 55.440
729228 95.837 -50.386 55.417
729367 98.301 -52.477 52.286
729368 98.322 -52.482 52.255
729369 98.343 -52.487 52.225
729370 98.363 -52.493 52.194
729371 98.384 -52.498 52.164
729372 98.405 -52.503 52.134
729373 98.425 -52.509 52.103
729374 98.445 -52.514 52.069
729375 98.466 -52.519 52.036
729376 98.487 -52.525 52.002
729377 98.514 -52.540 51.969
729378 98.543 -52.556 51.934
729379 98.572 -52.574 51.899
729380 98.600 -52.590 51.862
729381 98.635 -52.607 51.825
729382 98.671 -52.624 51.788
729383 98.708 -52.641 51.751
729384 98.730 -52.647 51.714
729385 98.740 -52.645 51.675
729386 98.750 -52.644 51.637
729387 98.761 -52.643 51.599
729388 98.771 -52.641 51.560
729389 98.781 -52.640 51.522
729390 98.792 -52.638 51.483
729391 98.802 -52.637 51.442
729392 98.824 -52.644 51.401
729393 98.855 -52.657 51.361
729394 98.886 -52.670 51.321
729395 98.918 -52.683 51.282
729396 98.949 -52.696 51.243
729397 98.980 -52.709 51.204
729398 99.007 -52.721 51.166
729399 99.031 -52.728 51.128
729400 99.052 -52.738 51.090
729401 99.072 -52.749 51.052
729402 99.091 -52.760 51.016
729403 99.112 -52.772 50.978
729404 99.131 -52.783 50.941
729405 99.150 -52.795 50.903
729406 99.174 -52.804 50.864
729407 99.201 -52.810 50.826
729408 99.223 -52.821 50.787
729409 99.244 -52.832 50.747
729410 99.265 -52.844 50.709
729411 99.280 -52.860 50.669
729412 99.295 -52.876 50.629
729413 99.310 -52.891 50.589
729414 99.326 -52.907 50.549
729415 99.341 -52.922 50.509
729416 99.356 -52.938 50.469
729417 99.371 -52.954 50.428
729418 99.386 -52.969 50.389
729419 99.402 -52.984 50.348
729420 99.420 -52.997 50.307
729421 99.437 -53.010 50.266
729422 99.455 -53.023 50.226
729423 99.472 -53.035 50.185
729424 99.490 -53.048 50.143
729425 99.507 -53.061 50.103
729426 99.525 -53.074 50.062
729427 99.543 -53.088 50.020
729428 99.560 -53.101 49.980
729429 99.577 -53.114 49.941
729919 95.544 -50.257 55.598
729920 95.577 -50.258 55.584
729921 95.610 -50.259 55.569
729922 95.642 -50.259 55.553
729923 95.675 -50.260 55.536
729924 95.703 -50.281 55.519
729925 95.731 -50.300 55.501
729926 95.758 -50.316 55.479
729927 95.784 -50.332 55.457
729928 95.810 -50.347 55.434
730071 98.378 -52.473 52.200
730072 98.398 -52.478 52.169
730073 98.419 -52.484 52.139
730074 98.440 -52.489 52.107
730075 98.460 -52.494 52.073
730076 98.481 -52.499 52.040
730077 98.502 -52.505 52.006
730078 98.526 -52.515 51.973
730079 98.555 -52.532 51.939
730080 98.583 -52.549 51.904
730081 98.612 -52.566 51.867
730082 98.641 -52.583 51.830
730083 98.677 -52.600 51.793
730084 98.714 -52.616 51.757
730085 98.738 -52.625 51.720
730086 98.748 -52.623 51.681
730087 98.759 -52.622 51.643
730088 98.769 -52.620 51.604
730089 98.779 -52.619 51.566
730090 98.790 -52.617 51.528
730091 98.800 -52.615 51.489
730092 98.813 -52.616 51.448
730093 98.841 -52.627 51.408
730094 98.871 -52.639 51.367
730095 98.902 -52.652 51.326
730096 98.933 -52.665 51.287
730097 98.963 -52.677 51.248
730098 98.994 -52.690 51.209
730099 99.018 -52.698 51.171
730100 99.040 -52.706 51.133
730101 99.060 -52.717 51.095
730102 99.080 -52.728 51.056
730103 99.100 -52.739 51.020
730104 99.121 -52.750 50.982
730105 99.141 -52.760 50.945
730106 99.161 -52.772 50.907
730107 99.184 -52.780 50.869
730108 99.211 -52.786 50.831
730109 99.233 -52.796 50.792
730110 99.255 -52.807 50.753
730111 99.276 -52.818 50.714
730112 99.293 -52.832 50.674
730113 99.308 -52.847 50.634
730114 99.324 -52.863 50.595
730115 99.339 -52.879 50.555
730116 99.354 -52.894 50.514
730117 99.369 -52.910 50.475
730118 99.384 -52.926 50.434
730119 99.400 -52.941 50.394
730120 99.415 -52.957 50.353
730121 99.430 -52.973 50.312
730122 99.448 -52.985 50.271
730123 99.465 -52.998 50.230
730124 99.483 -53.011 50.189
730125 99.500 -53.025 50.148
730126 99.518 -53.038 50.107
730127 99.536 -53.051 50.066
730128 99.553 -53.064 50.025
730129 99.570 -53.076 49.985
730130 99.588 -53.089 49.945
730622 95.609 -50.234 55.564
730623 95.642 -50.234 55.547
730624 95.675 -50.237 55.530
730625 95.702 -50.256 55.512
730626 95.731 -50.277 55.495
730627 95.757 -50.293 55.473
730773 98.414 -52.459 52.174
730774 98.434 -52.464 52.144
730775 98.455 -52.469 52.111
730776 98.475 -52.474 52.078
730777 98.496 -52.480 52.046
730778 98.516 -52.485 52.011
730779 98.537 -52.490 51.977
730780 98.566 -52.508 51.944
730781 98.595 -52.525 51.909
730782 98.624 -52.541 51.872
730783 98.652 -52.559 51.836
730784 98.683 -52.575 51.800
730785 98.720 -52.592 51.763
730786 98.746 -52.602 51.726
730787 98.757 -52.600 51.688
730788 98.767 -52.599 51.650
730789 98.777 -52.597 51.610
730790 98.787 -52.596 51.572
730791 98.799 -52.595 51.533
730792 98.814 -52.597 51.495
730793 98.833 -52.603 51.454
730794 98.864 -52.615 51.415
730795 98.895 -52.628 51.374
730796 98.925 -52.640 51.333
730797 98.956 -52.653 51.293
730798 98.982 -52.661 51.253
730799 99.006 -52.669 51.214
730800 99.028 -52.677 51.175
730801 99.050 -52.684 51.137
730802 99.072 -52.693 51.099
730803 99.094 -52.701 51.062
730804 99.114 -52.712 51.024
730805 99.134 -52.722 50.986
730806 99.154 -52.733 50.948
730807 99.174 -52.744 50.911
730808 99.195 -52.755 50.874
730809 99.220 -52.762 50.835
730810 99.244 -52.770 50.797
730811 99.265 -52.782 50.758
730812 99.287 -52.792 50.719
730813 99.307 -52.804 50.679
730814 99.322 -52.820 50.639
730815 99.337 -52.835 50.600
730816 99.353 -52.851 50.560
730817 99.368 -52.867 50.520
730818 99.383 -52.882 50.480
730819 99.398 -52.898 50.440
730820 99.413 -52.913 50.399
730821 99.429 -52.929 50.358
730822 99.444 -52.945 50.317
730823 99.459 -52.960 50.276
730824 99.476 -52.975 50.235
730825 99.493 -52.988 50.194
730826 99.511 -53.001 50.153
730827 99.528 -53.013 50.111
730828 99.546 -53.026 50.070
730829 99.563 -53.039 50.030
730830 99.581 -53.052 49.990
730831 99.598 -53.065 49.950
731474 98.428 -52.439 52.180
731475 98.449 -52.444 52.149
731476 98.469 -52.449 52.117
731477 98.490 -52.455 52.084
731478 98.511 -52.459 52.050
731479 98.531 -52.465 52.017
731480 98.552 -52.470 51.983
731481 98.578 -52.483 51.950
731482 98.606 -52.500 51.914
731483 98.635 -52.517 51.878
731484 98.664 -52.534 51.842
731485 98.693 -52.551 51.806
731486 98.726 -52.568 51.771
731487 98.755 -52.579 51.734
731488 98.765 -52.578 51.696
731489 98.775 -52.576 51.656
731490 98.785 -52.575 51.617
731491 98.800 -52.577 51.578
731492 98.815 -52.579 51.539
731493 98.830 -52.581 51.500
731494 98.856 -52.590 51.461
731495 98.886 -52.603 51.421
731496 98.917 -52.615 51.379
731497 98.944 -52.625 51.339
731498 98.969 -52.631 51.298
731499 98.993 -52.639 51.258
731500 99.015 -52.647 51.219
731501 99.038 -52.656 51.180
731502 99.059 -52.664 51.141
731503 99.081 -52.672 51.103
731504 99.103 -52.680 51.066
731505 99.125 -52.688 51.028
731506 99.147 -52.697 50.990
731507 99.168 -52.706 50.952
731508 99.188 -52.716 50.915
731509 99.208 -52.728 50.878
731510 99.229 -52.738 50.840
731511 99.255 -52.745 50.802
731512 99.276 -52.756 50.762
731513 99.297 -52.767 50.723
731514 99.319 -52.778 50.684
731515 99.336 -52.792 50.644
731516 99.351 -52.808 50.604
731517 99.366 -52.824 50.565
731518 99.381 -52.839 50.525
731519 99.397 -52.855 50.484
731520 99.412 -52.870 50.444
731521 99.427 -52.886 50.403
731522 99.442 -52.902 50.362
731523 99.457 -52.917 50.321
731524 99.473 -52.933 50.281
731525 99.488 -52.948 50.240
731526 99.504 -52.963 50.198
731527 99.521 -52.976 50.157
731528 99.539 -52.989 50.115
731529 99.556 -53.002 50.075
731530 99.574 -53.015 50.034
731531 99.591 -53.028 49.994
731532 99.609 -53.041 49.954
732175 98.443 -52.419 52.185
732176 98.464 -52.424 52.154
732177 98.485 -52.429 52.123
732178 98.505 -52.434 52.090
732179 98.525 -52.440 52.056
732180 98.546 -52.445 52.024
732181 98.567 -52.450 51.990
732182 98.589 -52.459 51.956
732183 98.618 -52.475 51.920
732184 98.647 -52.493 51.884
732185 98.675 -52.509 51.849
732186 98.704 -52.527 51.813
732187 98.733 -52.543 51.778
732188 98.763 -52.556 51.744
732189 98.773 -52.555 51.704
732190 98.786 -52.556 51.664
732191 98.801 -52.558 51.624
732192 98.816 -52.560 51.585
732193 98.831 -52.562 51.546
732194 98.848 -52.566 51.507
732195 98.878 -52.578 51.467
732196 98.907 -52.589 51.426
732197 98.931 -52.596 51.385
732198 98.955 -52.603 51.345
732199 98.980 -52.609 51.305
732200 99.003 -52.618 51.264
732201 99.025 -52.626 51.224
732202 99.047 -52.634 51.185
732203 99.069 -52.643 51.146
732204 99.090 -52.651 51.108
732205 99.113 -52.659 51.071
732206 99.134 -52.667 51.032
732207 99.156 -52.675 50.994
732208 99.179 -52.683 50.957
732209 99.200 -52.691 50.919
732210 99.222 -52.700 50.882
732211 99.242 -52.711 50.843
732212 99.265 -52.719 50.805
732213 99.286 -52.731 50.766
732214 99.307 -52.741 50.727
732215 99.329 -52.753 50.688
732216 99.349 -52.765 50.648
732217 99.364 -52.781 50.608
732218 99.379 -52.796 50.568
732219 99.395 -52.812 50.529
732220 99.410 -52.827 50.488
732221 99.425 -52.843 50.448
732222 99.440 -52.859 50.408
732223 99.455 -52.874 50.367
732224 99.471 -52.890 50.326
732225 99.486 -52.905 50.285
732226 99.501 -52.921 50.244
732227 99.516 -52.937 50.203
732228 99.532 -52.952 50.161
732229 99.549 -52.965 50.120
732230 99.567 -52.978 50.079
732231 99.584 -52.991 50.038
732232 99.602 -53.004 49.998
732233 99.620 -53.017 49.958
732876 98.458 -52.399 52.191
732877 98.479 -52.404 52.160
732878 98.499 -52.409 52.128
732879 98.520 -52.415 52.096
732880 98.541 -52.420 52.063
732881 98.561 -52.425 52.030
732882 98.582 -52.431 51.996
732883 98.602 -52.436 51.963
732884 98.630 -52.451 51.927
732885 98.658 -52.468 51.891
732886 98.687 -52.485 51.856
732887 98.716 -52.502 51.821
732888 98.744 -52.519 51.787
732889 98.772 -52.534 51.753
732890 98.787 -52.537 51.712
732891 98.802 -52.540 51.672
732892 98.817 -52.542 51.631
732893 98.832 -52.544 51.591
732894 98.847 -52.546 51.553
732895 98.869 -52.553 51.515
732896 98.894 -52.559 51.474
732897 98.918 -52.566 51.431
732898 98.942 -52.573 51.391
732899 98.967 -52.580 51.351
732900 98.990 -52.588 51.311
732901 99.012 -52.597 51.270
732902 99.034 -52.605 51.230
732903 99.056 -52.613 51.191
732904 99.078 -52.622 51.152
732905 99.100 -52.630 51.114
732906 99.122 -52.637 51.076
732907 99.144 -52.646 51.037
732908 99.166 -52.654 50.999
732909 99.188 -52.662 50.961
732910 99.210 -52.670 50.924
732911 99.231 -52.678 50.886
732912 99.254 -52.687 50.847
732913 99.275 -52.695 50.809
732914 99.297 -52.705 50.771
732915 99.318 -52.716 50.732
732916 99.340 -52.728 50.692
732917 99.361 -52.738 50.652
732918 99.378 -52.753 50.613
732919 99.393 -52.769 50.572
732920 99.408 -52.784 50.532
732921 99.423 -52.800 50.492
732922 99.438 -52.816 50.453
732923 99.454 -52.831 50.412
732924 99.469 -52.847 50.371
732925 99.484 -52.862 50.330
732926 99.499 -52.878 50.289
732927 99.515 -52.894 50.248
732928 99.530 -52.909 50.207
732929 99.545 -52.925 50.165
732930 99.560 -52.941 50.124
732931 99.577 -52.954 50.083
732932 99.595 -52.966 50.042
732933 99.613 -52.979 50.002
732934 99.630 -52.992 49.962
733577 98.473 -52.379 52.197
733578 98.494 -52.384 52.166
733579 98.514 -52.390 52.134
733580 98.535 -52.395 52.102
733581 98.556 -52.400 52.069
733582 98.576 -52.405 52.036
733583 98.596 -52.411 52.003
733584 98.617 -52.416 51.969
733585 98.641 -52.427 51.935
733586 98.670 -52.443 51.899
733587 98.699 -52.461 51.864
733588 98.727 -52.478 51.829
733589 98.756 -52.494 51.795
733590 98.780 -52.508 51.761
733591 98.799 -52.515 51.720
733592 98.814 -52.517 51.679
733593 98.830 -52.520 51.638
733594 98.845 -52.522 51.599
733595 98.860 -52.525 51.560
733596 98.880 -52.530 51.521
733597 98.905 -52.537 51.480
733598 98.929 -52.543 51.439
733599 98.953 -52.550 51.398
733600 98.977 -52.559 51.357
733601 99.000 -52.568 51.316
733602 99.022 -52.576 51.276
733603 99.044 -52.584 51.236
733604 99.066 -52.592 51.197
733605 99.087 -52.600 51.158
733606 99.110 -52.609 51.119
733607 99.131 -52.617 51.081
733608 99.153 -52.625 51.042
733609 99.175 -52.633 51.004
733610 99.197 -52.641 50.966
733611 99.219 -52.650 50.927
733612 99.241 -52.657 50.890
733613 99.263 -52.666 50.852
733614 99.285 -52.673 50.812
733615 99.307 -52.681 50.775
733616 99.329 -52.691 50.735
733617 99.350 -52.702 50.696
733618 99.371 -52.713 50.657
733619 99.391 -52.725 50.617
733620 99.407 -52.741 50.577
733621 99.422 -52.756 50.537
733622 99.437 -52.772 50.497
733623 99.452 -52.788 50.457
733624 99.467 -52.803 50.417
733625 99.483 -52.819 50.375
733626 99.498 -52.835 50.334
733627 99.513 -52.850 50.293
733628 99.528 -52.866 50.252
733629 99.544 -52.882 50.211
733630 99.559 -52.897 50.169
733631 99.574 -52.913 50.128
733632 99.589 -52.929 50.087
733633 99.606 -52.943 50.046
733634 99.623 -52.956 50.005
733635 99.640 -52.969 49.965
734278 98.488 -52.361 52.204
734279 98.509 -52.365 52.172
734280 98.529 -52.370 52.141
734281 98.550 -52.375 52.108
734282 98.570 -52.380 52.075
734283 98.591 -52.386 52.042
734284 98.612 -52.391 52.009
734285 98.632 -52.396 51.976
734286 98.653 -52.402 51.942
734287 98.681 -52.419 51.906
734288 98.710 -52.436 51.871
734289 98.739 -52.453 51.837
734290 98.764 -52.467 51.803
734291 98.788 -52.481 51.768
734292 98.808 -52.490 51.727
734293 98.826 -52.495 51.686
734294 98.841 -52.498 51.645
734295 98.856 -52.500 51.606
734296 98.872 -52.503 51.567
734297 98.891 -52.508 51.527
734298 98.916 -52.515 51.486
734299 98.940 -52.521 51.445
734300 98.964 -52.529 51.405
734301 98.987 -52.538 51.364
734302 99.009 -52.546 51.323
734303 99.031 -52.555 51.283
734304 99.053 -52.563 51.242
734305 99.075 -52.572 51.202
734306 99.097 -52.579 51.164
734307 99.119 -52.587 51.124
734308 99.141 -52.596 51.086
734309 99.162 -52.604 51.047
734310 99.185 -52.612 51.008
734311 99.207 -52.620 50.970
734312 99.228 -52.628 50.932
734313 99.251 -52.636 50.894
734314 99.273 -52.644 50.856
734315 99.295 -52.651 50.817
734316 99.317 -52.659 50.778
734317 99.339 -52.666 50.739
734318 99.360 -52.677 50.700
734319 99.382 -52.687 50.661
734320 99.403 -52.699 50.621
734321 99.420 -52.713 50.581
734322 99.435 -52.729 50.542
734323 99.450 -52.744 50.501
734324 99.466 -52.760 50.461
734325 99.481 -52.775 50.421
734326 99.496 -52.791 50.380
734327 99.511 -52.807 50.338
734328 99.526 -52.822 50.297
734329 99.542 -52.838 50.255
734330 99.557 -52.854 50.214
734331 99.572 -52.869 50.172
734332 99.587 -52.885 50.131
734333 99.603 -52.901 50.091
734334 99.618 -52.916 50.050
734335 99.633 -52.932 50.010
734336 99.651 -52.944 49.970
734979 98.501 -52.348 52.210
734980 98.522 -52.351 52.179
734981 98.543 -52.354 52.148
734982 98.564 -52.357 52.115
734983 98.585 -52.361 52.081
734984 98.606 -52.366 52.048
734985 98.626 -52.371 52.016
734986 98.647 -52.377 51.983
734987 98.667 -52.382 51.949
734988 98.693 -52.394 51.913
734989 98.722 -52.412 51.878
734990 98.748 -52.427 51.845
734991 98.772 -52.440 51.811
734992 98.795 -52.453 51.774
734993 98.818 -52.466 51.733
734994 98.836 -52.470 51.692
734995 98.852 -52.474 51.652
734996 98.868 -52.478 51.612
734997 98.883 -52.481 51.573
734998 98.903 -52.485 51.533
734999 98.927 -52.492 51.492
735000 98.951 -52.499 51.452
735001 98.974 -52.509 51.411
735002 98.996 -52.518 51.371
735003 99.018 -52.526 51.331
735004 99.041 -52.534 51.290
735005 99.062 -52.542 51.249
735006 99.084 -52.550 51.209
735007 99.106 -52.558 51.169
735008 99.128 -52.566 51.130
735009 99.150 -52.575 51.091
735010 99.172 -52.583 51.052
735011 99.194 -52.590 51.013
735012 99.216 -52.598 50.974
735013 99.238 -52.606 50.937
735014 99.260 -52.614 50.898
735015 99.282 -52.622 50.860
735016 99.304 -52.629 50.821
735017 99.327 -52.637 50.782
735018 99.349 -52.644 50.743
735019 99.371 -52.652 50.704
735020 99.393 -52.662 50.664
735021 99.414 -52.673 50.625
735022 99.433 -52.686 50.585
735023 99.448 -52.701 50.546
735024 99.464 -52.717 50.505
735025 99.479 -52.732 50.465
735026 99.494 -52.748 50.425
735027 99.509 -52.764 50.384
735028 99.524 -52.779 50.342
735029 99.540 -52.795 50.300
735030 99.555 -52.810 50.259
735031 99.570 -52.826 50.218
735032 99.585 -52.842 50.175
735033 99.600 -52.857 50.134
735034 99.616 -52.873 50.095
735035 99.631 -52.889 50.054
735036 99.646 -52.904 50.014
735037 99.661 -52.920 49.974
735680 98.514 -52.336 52.218
735681 98.535 -52.339 52.186
735682 98.556 -52.342 52.155
735683 98.577 -52.344 52.122
735684 98.598 -52.347 52.088
735685 98.620 -52.350 52.055
735686 98.641 -52.353 52.023
735687 98.662 -52.357 51.990
735688 98.683 -52.362 51.956
735689 98.705 -52.370 51.920
735690 98.732 -52.386 51.886
735691 98.756 -52.399 51.851
735692 98.779 -52.412 51.816
735693 98.803 -52.426 51.779
735694 98.827 -52.439 51.738
735695 98.845 -52.446 51.698
735696 98.862 -52.450 51.658
735697 98.879 -52.455 51.619
735698 98.896 -52.459 51.580
735699 98.914 -52.462 51.540
735700 98.938 -52.470 51.500
735701 98.960 -52.479 51.459
735702 98.984 -52.487 51.417
735703 99.006 -52.496 51.378
735704 99.028 -52.505 51.338
735705 99.050 -52.512 51.297
735706 99.072 -52.521 51.256
735707 99.094 -52.529 51.215
735708 99.116 -52.537 51.175
735709 99.138 -52.545 51.135
735710 99.159 -52.553 51.096
735711 99.182 -52.561 51.057
735712 99.204 -52.568 51.018
735713 99.226 -52.576 50.979
735714 99.248 -52.584 50.941
735715 99.270 -52.591 50.903
735716 99.292 -52.600 50.864
735717 99.314 -52.607 50.824
735718 99.337 -52.615 50.787
735719 99.359 -52.622 50.747
735720 99.380 -52.630 50.708
735721 99.403 -52.637 50.668
735722 99.424 -52.648 50.629
735723 99.445 -52.659 50.590
735724 99.462 -52.674 50.551
735725 99.477 -52.689 50.510
735726 99.492 -52.705 50.470
735727 99.508 -52.721 50.429
735728 99.523 -52.736 50.387
735729 99.538 -52.752 50.346
735730 99.553 -52.767 50.304
735731 99.569 -52.783 50.262
735732 99.584 -52.799 50.221
735733 99.599 -52.814 50.179
735734 99.614 -52.830 50.138
735735 99.629 -52.846 50.098
735736 99.645 -52.861 50.058
735737 99.660 -52.877 50.018
735738 99.675 -52.892 49.978
736381 98.527 -52.323 52.226
736382 98.548 -52.326 52.194
736383 98.569 -52.329 52.162
736384 98.590 -52.332 52.129
736385 98.612 -52.335 52.095
736386 98.633 -52.338 52.061
736387 98.654 -52.340 52.029
736388 98.675 -52.343 51.996
736389 98.696 -52.346 51.962
736390 98.717 -52.349 51.927
736391 98.740 -52.358 51.892
736392 98.763 -52.372 51.857
736393 98.787 -52.385 51.821
736394 98.810 -52.399 51.784
736395 98.834 -52.412 51.744
736396 98.855 -52.421 51.704
736397 98.872 -52.426 51.665
736398 98.889 -52.431 51.626
736399 98.906 -52.435 51.587
736400 98.925 -52.440 51.548
736401 98.948 -52.449 51.507
736402 98.971 -52.458 51.466
736403 98.993 -52.467 51.425
736404 99.015 -52.475 51.384
736405 99.038 -52.484 51.344
736406 99.059 -52.492 51.304
736407 99.081 -52.499 51.262
736408 99.103 -52.508 51.222
736409 99.125 -52.516 51.181
736410 99.147 -52.524 51.141
736411 99.169 -52.531 51.102
736412 99.191 -52.539 51.062
736413 99.213 -52.546 51.022
736414 99.235 -52.554 50.983
736415 99.258 -52.562 50.945
736416 99.280 -52.569 50.907
736417 99.302 -52.577 50.868
736418 99.324 -52.585 50.828
736419 99.346 -52.593 50.791
736420 99.368 -52.600 50.752
736421 99.391 -52.608 50.712
736422 99.413 -52.615 50.672
736423 99.435 -52.623 50.633
736424 99.456 -52.634 50.593
736425 99.476 -52.646 50.554
736426 99.491 -52.662 50.514
736427 99.506 -52.678 50.474
736428 99.521 -52.693 50.433
736429 99.537 -52.709 50.392
736430 99.552 -52.724 50.350
736431 99.567 -52.740 50.308
736432 99.582 -52.756 50.266
736433 99.597 -52.771 50.225
736434 99.613 -52.787 50.183
736435 99.628 -52.803 50.142
736436 99.643 -52.818 50.101
736437 99.658 -52.834 50.062
736438 99.674 -52.849 50.022
736439 99.689 -52.865 49.982
737082 98.539 -52.311 52.233
737083 98.561 -52.313 52.201
737084 98.582 -52.316 52.169
737085 98.603 -52.319 52.136
737086 98.625 -52.322 52.101
737087 98.646 -52.325 52.068
737088 98.667 -52.328 52.035
737089 98.688 -52.331 52.002
737090 98.709 -52.333 51.968
737091 98.730 -52.336 51.933
737092 98.751 -52.340 51.899
737093 98.770 -52.346 51.863
737094 98.795 -52.358 51.826
737095 98.818 -52.371 51.789
737096 98.842 -52.384 51.750
737097 98.865 -52.397 51.712
737098 98.882 -52.402 51.673
737099 98.899 -52.406 51.633
737100 98.916 -52.412 51.594
737101 98.935 -52.420 51.555
737102 98.957 -52.428 51.514
737103 98.981 -52.437 51.473
737104 99.003 -52.446 51.432
737105 99.025 -52.454 51.391
737106 99.047 -52.462 51.351
737107 99.069 -52.471 51.310
737108 99.090 -52.478 51.269
737109 99.113 -52.487 51.228
737110 99.135 -52.494 51.187
737111 99.157 -52.502 51.146
737112 99.179 -52.509 51.108
737113 99.201 -52.517 51.067
737114 99.223 -52.525 51.027
737115 99.245 -52.532 50.987
737116 99.268 -52.540 50.948
737117 99.290 -52.547 50.911
737118 99.311 -52.555 50.872
737119 99.334 -52.562 50.832
737120 99.356 -52.571 50.794
737121 99.378 -52.578 50.756
737122 99.400 -52.586 50.716
737123 99.422 -52.593 50.676
737124 99.444 -52.601 50.637
737125 99.467 -52.609 50.597
737126 99.488 -52.619 50.558
737127 99.504 -52.634 50.518
737128 99.519 -52.650 50.477
737129 99.535 -52.666 50.436
737130 99.550 -52.681 50.395
737131 99.565 -52.697 50.354
737132 99.580 -52.713 50.312
737133 99.595 -52.728 50.270
737134 99.611 -52.744 50.228
737135 99.626 -52.760 50.187
737136 99.641 -52.775 50.145
737137 99.656 -52.791 50.104
737138 99.671 -52.806 50.065
737139 99.687 -52.822 50.025
737140 99.702 -52.838 49.985
737783 98.553 -52.298 52.242
737784 98.574 -52.301 52.210
737785 98.595 -52.304 52.177
737786 98.616 -52.307 52.144
737787 98.638 -52.309 52.109
737788 98.659 -52.312 52.075
737789 98.680 -52.315 52.041
737790 98.701 -52.318 52.008
737791 98.722 -52.321 51.974
737792 98.743 -52.324 51.939
737793 98.765 -52.327 51.903
737794 98.785 -52.331 51.867
737795 98.803 -52.338 51.831
737796 98.821 -52.344 51.794
737797 98.847 -52.356 51.756
737798 98.873 -52.371 51.718
737799 98.892 -52.377 51.680
737800 98.910 -52.384 51.640
737801 98.927 -52.392 51.601
737802 98.945 -52.399 51.563
737803 98.968 -52.408 51.521
737804 98.990 -52.417 51.481
737805 99.012 -52.425 51.439
737806 99.034 -52.433 51.398
737807 99.056 -52.441 51.358
737808 99.078 -52.449 51.316
737809 99.100 -52.457 51.274
737810 99.122 -52.465 51.232
737811 99.144 -52.472 51.192
737812 99.166 -52.480 51.152
737813 99.189 -52.487 51.113
737814 99.211 -52.495 51.072
737815 99.233 -52.503 51.031
737816 99.255 -52.510 50.992
737817 99.277 -52.518 50.952
737818 99.299 -52.525 50.914
737819 99.322 -52.533 50.876
737820 99.344 -52.540 50.836
737821 99.366 -52.548 50.797
737822 99.388 -52.556 50.760
737823 99.410 -52.564 50.720
737824 99.432 -52.572 50.680
737825 99.454 -52.579 50.641
737826 99.476 -52.587 50.601
737827 99.498 -52.594 50.562
737828 99.517 -52.606 50.521
737829 99.533 -52.622 50.482
737830 99.548 -52.637 50.440
737831 99.563 -52.653 50.399
737832 99.578 -52.669 50.357
737833 99.594 -52.684 50.315
737834 99.609 -52.700 50.273
737835 99.624 -52.716 50.231
737836 99.639 -52.731 50.190
737837 99.654 -52.747 50.148
737838 99.670 -52.763 50.108
737839 99.685 -52.778 50.067
737840 99.700 -52.794 50.028
737841 99.715 -52.810 49.987
738484 98.566 -52.285 52.251
738485 98.587 -52.288 52.219
738486 98.608 -52.291 52.185
738487 98.629 -52.294 52.151
738488 98.650 -52.297 52.117
738489 98.671 -52.300 52.082
738490 98.693 -52.303 52.048
738491 98.714 -52.305 52.015
738492 98.735 -52.308 51.980
738493 98.757 -52.311 51.944
738494 98.778 -52.315 51.909
738495 98.799 -52.318 51.873
738496 98.818 -52.323 51.837
738497 98.837 -52.330 51.800
738498 98.855 -52.336 51.762
738499 98.873 -52.343 51.724
738500 98.899 -52.353 51.686
738501 98.921 -52.364 51.647
738502 98.938 -52.371 51.608
738503 98.956 -52.378 51.569
738504 98.977 -52.387 51.529
738505 99.000 -52.396 51.488
738506 99.022 -52.404 51.446
738507 99.044 -52.412 51.405
738508 99.066 -52.420 51.364
738509 99.088 -52.427 51.323
738510 99.110 -52.435 51.280
738511 99.132 -52.443 51.239
738512 99.154 -52.450 51.198
738513 99.177 -52.458 51.157
738514 99.199 -52.465 51.118
738515 99.221 -52.473 51.076
738516 99.242 -52.481 51.036
738517 99.265 -52.488 50.996
738518 99.287 -52.496 50.957
738519 99.309 -52.503 50.918
738520 99.331 -52.511 50.879
738521 99.353 -52.518 50.840
738522 99.375 -52.526 50.801
738523 99.398 -52.534 50.762
738524 99.420 -52.541 50.723
738525 99.442 -52.550 50.684
738526 99.464 -52.557 50.644
738527 99.486 -52.565 50.605
738528 99.508 -52.572 50.566
738529 99.530 -52.580 50.525
738530 99.546 -52.594 50.486
738531 99.562 -52.610 50.444
738532 99.577 -52.626 50.403
738533 99.592 -52.641 50.361
738534 99.607 -52.657 50.319
738535 99.623 -52.672 50.277
738536 99.638 -52.688 50.235
738537 99.653 -52.704 50.193
738538 99.668 -52.719 50.152
738539 99.683 -52.735 50.111
738540 99.699 -52.751 50.071
738541 99.714 -52.766 50.031
738542 99.729 -52.782 49.990
739185 98.579 -52.273 52.260
739186 98.600 -52.276 52.228
739187 98.621 -52.279 52.194
739188 98.642 -52.281 52.158
739189 98.663 -52.284 52.123
739190 98.685 -52.287 52.090
739191 98.706 -52.290 52.056
739192 98.727 -52.293 52.022
739193 98.748 -52.296 51.986
739194 98.770 -52.299 51.950
739195 98.791 -52.302 51.915
739196 98.812 -52.305 51.879
739197 98.833 -52.308 51.843
739198 98.851 -52.315 51.806
739199 98.870 -52.321 51.768
739200 98.888 -52.327 51.730
739201 98.907 -52.334 51.692
739202 98.925 -52.340 51.654
739203 98.949 -52.351 51.615
739204 98.967 -52.358 51.575
739205 98.987 -52.366 51.537
739206 99.009 -52.374 51.494
739207 99.031 -52.383 51.453
739208 99.053 -52.390 51.411
739209 99.075 -52.398 51.370
739210 99.097 -52.405 51.329
739211 99.120 -52.413 51.287
739212 99.142 -52.421 51.245
739213 99.164 -52.428 51.203
739214 99.186 -52.436 51.163
739215 99.208 -52.443 51.122
739216 99.230 -52.451 51.080
739217 99.253 -52.459 51.040
739218 99.275 -52.466 51.000
739219 99.297 -52.474 50.961
739220 99.319 -52.481 50.921
739221 99.341 -52.490 50.882
739222 99.363 -52.498 50.843
739223 99.385 -52.506 50.804
739224 99.408 -52.514 50.766
739225 99.430 -52.521 50.726
739226 99.452 -52.530 50.687
739227 99.475 -52.538 50.648
739228 99.497 -52.546 50.608
739229 99.519 -52.554 50.569
739230 99.542 -52.562 50.529
739231 99.561 -52.571 50.489
739232 99.575 -52.583 50.447
739233 99.590 -52.598 50.406
739234 99.606 -52.614 50.364
739235 99.621 -52.629 50.322
739236 99.636 -52.645 50.280
739237 99.651 -52.661 50.238
739238 99.666 -52.676 50.196
739239 99.682 -52.692 50.155
739240 99.697 -52.708 50.114
739241 99.712 -52.723 50.074
739242 99.727 -52.739 50.033
739243 99.743 -52.754 49.993
739886 98.592 -52.260 52.269
739887 98.613 -52.263 52.237
739888 98.634 -52.266 52.203
739889 98.655 -52.269 52.167
739890 98.676 -52.272 52.133
739891 98.698 -52.274 52.098
739892 98.719 -52.277 52.063
739893 98.740 -52.280 52.028
739894 98.761 -52.283 51.992
739895 98.782 -52.286 51.956
739896 98.804 -52.290 51.920
739897 98.825 -52.293 51.884
739898 98.846 -52.295 51.849
739899 98.866 -52.299 51.812
739900 98.885 -52.306 51.774
739901 98.903 -52.312 51.736
739902 98.922 -52.319 51.697
739903 98.940 -52.325 51.659
739904 98.959 -52.332 51.621
739905 98.977 -52.339 51.581
739906 98.997 -52.346 51.542
739907 99.019 -52.353 51.500
739908 99.041 -52.362 51.459
739909 99.064 -52.369 51.418
739910 99.086 -52.377 51.376
739911 99.108 -52.386 51.334
739912 99.130 -52.393 51.292
739913 99.152 -52.402 51.251
739914 99.175 -52.410 51.209
739915 99.197 -52.418 51.167
739916 99.219 -52.426 51.127
739917 99.241 -52.434 51.085
739918 99.264 -52.442 51.044
739919 99.286 -52.450 51.004
739920 99.308 -52.458 50.965
739921 99.331 -52.466 50.926
739922 99.353 -52.474 50.886
739923 99.375 -52.482 50.847
739924 99.398 -52.490 50.809
739925 99.420 -52.498 50.769
739926 99.442 -52.506 50.729
739927 99.464 -52.515 50.689
739928 99.486 -52.522 50.650
739929 99.509 -52.531 50.611
739930 99.531 -52.538 50.572
739931 99.553 -52.546 50.531
739932 99.575 -52.555 50.491
739933 99.591 -52.565 50.450
739934 99.605 -52.576 50.408
739935 99.619 -52.587 50.366
739936 99.634 -52.602 50.324
739937 99.649 -52.618 50.282
739938 99.664 -52.633 50.240
739939 99.680 -52.649 50.198
739940 99.695 -52.665 50.157
739941 99.710 -52.680 50.116
739942 99.725 -52.696 50.076
739943 99.740 -52.711 50.035
739944 99.756 -52.727 49.994
740588 98.626 -52.250 52.247
740589 98.647 -52.253 52.212
740590 98.668 -52.256 52.176
740591 98.690 -52.259 52.142
740592 98.711 -52.262 52.106
740593 98.732 -52.265 52.070
740594 98.753 -52.268 52.035
740595 98.774 -52.271 51.999
740596 98.795 -52.274 51.962
740597 98.816 -52.277 51.926
740598 98.838 -52.280 51.890
740599 98.859 -52.283 51.854
740600 98.880 -52.286 51.818
740601 98.900 -52.291 51.780
740602 98.918 -52.297 51.742
740603 98.937 -52.304 51.703
740604 98.955 -52.311 51.664
740605 98.974 -52.317 51.626
740606 98.992 -52.324 51.587
740607 99.011 -52.330 51.547
740608 99.031 -52.338 51.506
740609 99.053 -52.346 51.465
740610 99.075 -52.354 51.424
740611 99.097 -52.362 51.383
740612 99.120 -52.370 51.341
740613 99.142 -52.378 51.298
740614 99.164 -52.386 51.255
740615 99.187 -52.394 51.213
740616 99.209 -52.402 51.171
740617 99.231 -52.410 51.131
740619 99.276 -52.426 51.048
740620 99.298 -52.434 51.008
740621 99.320 -52.443 50.968
740622 99.342 -52.450 50.929
740623 99.365 -52.459 50.890
740624 99.387 -52.467 50.850
740625 99.409 -52.474 50.811
740626 99.431 -52.483 50.772
740627 99.453 -52.490 50.732
740628 99.476 -52.499 50.692
740629 99.498 -52.507 50.653
740630 99.520 -52.515 50.613
740631 99.543 -52.523 50.573
740632 99.565 -52.531 50.533
740633 99.587 -52.539 50.492
740634 99.607 -52.548 50.451
740635 99.621 -52.559 50.410
740636 99.635 -52.570 50.368
740637 99.649 -52.581 50.325
740638 99.663 -52.592 50.283
740639 99.678 -52.606 50.241
740640 99.693 -52.622 50.198
740641 99.708 -52.637 50.158
740642 99.724 -52.653 50.118
740643 99.739 -52.668 50.078
740644 99.754 -52.684 50.037
740645 99.769 -52.700 49.997
741291 98.681 -52.244 52.187
741292 98.703 -52.246 52.151
741293 98.724 -52.249 52.114
741294 98.745 -52.252 52.078
741295 98.766 -52.255 52.042
741296 98.787 -52.258 52.006
741297 98.808 -52.261 51.970
741298 98.830 -52.264 51.933
741299 98.851 -52.267 51.896
741300 98.872 -52.270 51.860
741301 98.893 -52.273 51.824
741302 98.914 -52.276 51.785
741303 98.933 -52.283 51.747
741304 98.951 -52.289 51.709
741305 98.970 -52.296 51.670
741306 98.989 -52.302 51.632
741307 99.007 -52.308 51.593
741308 99.025 -52.315 51.552
741309 99.044 -52.321 51.512
741310 99.065 -52.330 51.471
741311 99.087 -52.338 51.430
741312 99.110 -52.346 51.388
741313 99.132 -52.355 51.347
741314 99.154 -52.362 51.304
741315 99.176 -52.371 51.260
741316 99.198 -52.378 51.218
741317 99.221 -52.387 51.176
741318 99.243 -52.395 51.135
741322 99.332 -52.427 50.972
741323 99.354 -52.435 50.933
741324 99.376 -52.443 50.893
741325 99.399 -52.451 50.853
741326 99.421 -52.459 50.814
741327 99.443 -52.467 50.774
741328 99.466 -52.475 50.735
741329 99.488 -52.483 50.695
741330 99.510 -52.491 50.656
741331 99.532 -52.499 50.616
741332 99.555 -52.507 50.575
741333 99.577 -52.515 50.534
741334 99.599 -52.524 50.494
741335 99.621 -52.531 50.453
741336 99.637 -52.542 50.411
741337 99.651 -52.553 50.369
741338 99.665 -52.564 50.327
741339 99.679 -52.575 50.284
741340 99.693 -52.586 50.242
741341 99.707 -52.597 50.200
741342 99.722 -52.609 50.160
741343 99.737 -52.625 50.119
741344 99.753 -52.641 50.079
741345 99.768 -52.656 50.039
741346 99.783 -52.672 49.999
741993 98.716 -52.234 52.159
741994 98.737 -52.237 52.122
741995 98.758 -52.240 52.086
741996 98.779 -52.242 52.051
741997 98.800 -52.246 52.015
741998 98.821 -52.249 51.977
741999 98.843 -52.252 51.939
742000 98.864 -52.255 51.902
742001 98.885 -52.258 51.866
742002 98.906 -52.260 51.829
742003 98.927 -52.263 51.791
742004 98.948 -52.268 51.753
742005 98.966 -52.274 51.714
742006 98.985 -52.280 51.675
742007 99.003 -52.287 51.637
742008 99.022 -52.293 51.599
742009 99.040 -52.300 51.558
742010 99.059 -52.306 51.517
742011 99.077 -52.315 51.476
742012 99.099 -52.323 51.435
742013 99.121 -52.330 51.393
742014 99.143 -52.339 51.351
742015 99.165 -52.347 51.308
742016 99.188 -52.355 51.266
742017 99.210 -52.363 51.223
742018 99.232 -52.371 51.182
742024 99.366 -52.419 50.937
742025 99.388 -52.427 50.896
742026 99.411 -52.435 50.856
742027 99.433 -52.443 50.817
742028 99.455 -52.452 50.777
742029 99.477 -52.459 50.738
742030 99.499 -52.468 50.698
742031 99.522 -52.476 50.658
742032 99.544 -52.484 50.619
742033 99.566 -52.492 50.577
742034 99.588 -52.499 50.536
742035 99.611 -52.508 50.496
742036 99.633 -52.516 50.454
742037 99.653 -52.525 50.412
742038 99.667 -52.536 50.370
742039 99.681 -52.546 50.328
742040 99.695 -52.558 50.285
742041 99.709 -52.568 50.243
742042 99.723 -52.580 50.202
742043 99.737 -52.590 50.161
742044 99.752 -52.602 50.121
742045 99.766 -52.613 50.080
742046 99.781 -52.628 50.040
742047 99.796 -52.644 50.000
742696 98.771 -52.227 52.095
742697 98.792 -52.230 52.059
742698 98.813 -52.233 52.022
742699 98.835 -52.236 51.984
742700 98.856 -52.239 51.946
742701 98.877 -52.242 51.908
742702 98.898 -52.245 51.871
742703 98.919 -52.248 51.835
742704 98.940 -52.251 51.797
742705 98.961 -52.254 51.759
742706 98.981 -52.259 51.720
742707 99.000 -52.265 51.682
742708 99.018 -52.272 51.643
742709 99.037 -52.278 51.605
742710 99.055 -52.285 51.563
742711 99.074 -52.291 51.523
742712 99.092 -52.298 51.481
742713 99.111 -52.307 51.440
742714 99.133 -52.315 51.398
742715 99.155 -52.324 51.356
742716 99.178 -52.331 51.314
742717 99.200 -52.340 51.271
742718 99.222 -52.347 51.228
742719 99.244 -52.355 51.187
742726 99.400 -52.412 50.899
742727 99.422 -52.420 50.859
742728 99.444 -52.428 50.820
742729 99.467 -52.436 50.780
742730 99.489 -52.444 50.741
742731 99.511 -52.452 50.701
742732 99.534 -52.460 50.660
742733 99.556 -52.468 50.620
742734 99.578 -52.476 50.579
742735 99.600 -52.484 50.537
742736 99.623 -52.492 50.497
742737 99.645 -52.500 50.455
742738 99.667 -52.509 50.413
742739 99.683 -52.518 50.371
742740 99.697 -52.530 50.329
742741 99.711 -52.540 50.287
742742 99.725 -52.552 50.245
742743 99.739 -52.562 50.204
742744 99.754 -52.574 50.163
742745 99.768 -52.584 50.122
742746 99.782 -52.595 50.082
742747 99.796 -52.606 50.042
742748 99.810 -52.617 50.001
743398 98.805 -52.217 52.067
743399 98.827 -52.221 52.029
743400 98.848 -52.224 51.991
743401 98.869 -52.227 51.953
743402 98.890 -52.230 51.914
743403 98.911 -52.232 51.878
743404 98.932 -52.235 51.841
743405 98.953 -52.238 51.803
743406 98.975 -52.241 51.765
743407 98.996 -52.244 51.727
743408 99.014 -52.250 51.689
743409 99.033 -52.257 51.649
743410 99.052 -52.263 51.610
743411 99.070 -52.270 51.569
743412 99.088 -52.277 51.529
743413 99.107 -52.283 51.487
743414 99.126 -52.290 51.445
743415 99.145 -52.299 51.404
743416 99.167 -52.308 51.362
743417 99.189 -52.316 51.319
743418 99.211 -52.324 51.277
743419 99.234 -52.332 51.234
743420 99.256 -52.340 51.192
743427 99.412 -52.396 50.903
743428 99.434 -52.404 50.863
743429 99.456 -52.412 50.823
743430 99.479 -52.421 50.783
743431 99.501 -52.428 50.743
743432 99.523 -52.437 50.703
743433 99.545 -52.444 50.662
743434 99.568 -52.452 50.622
743435 99.590 -52.461 50.581
743436 99.612 -52.468 50.539
743437 99.634 -52.477 50.498
743438 99.656 -52.485 50.456
743439 99.679 -52.493 50.414
743440 99.699 -52.502 50.372
743441 99.713 -52.512 50.330
743442 99.727 -52.524 50.288
743443 99.741 -52.534 50.246
743444 99.755 -52.546 50.205
743445 99.769 -52.556 50.164
743446 99.783 -52.568 50.123
743447 99.798 -52.578 50.083
743448 99.812 -52.589 50.042
743449 99.826 -52.600 50.002
743480 100.432 -52.988 49.143
743481 100.459 -53.001 49.118
743482 100.487 -53.019 49.092
743483 100.514 -53.035 49.065
743484 100.542 -53.053 49.037
744100 98.839 -52.208 52.037
744101 98.861 -52.211 51.998
744102 98.882 -52.214 51.960
744103 98.903 -52.217 51.921
744104 98.924 -52.220 51.884
744105 98.945 -52.223 51.848
744106 98.967 -52.226 51.809
744107 98.988 -52.228 51.771
744108 99.009 -52.231 51.732
744109 99.029 -52.236 51.694
744110 99.048 -52.242 51.655
744111 99.066 -52.249 51.616
744112 99.085 -52.255 51.575
744113 99.103 -52.261 51.534
744114 99.122 -52.268 51.493
744115 99.140 -52.274 51.451
744116 99.159 -52.283 51.409
744117 99.179 -52.292 51.367
744118 99.201 -52.300 51.325
744119 99.223 -52.308 51.283
744120 99.245 -52.316 51.239
744121 99.268 -52.324 51.197
744129 99.446 -52.389 50.867
744130 99.468 -52.396 50.826
744131 99.490 -52.405 50.786
744132 99.512 -52.413 50.745
744133 99.535 -52.421 50.704
744134 99.557 -52.429 50.664
744135 99.579 -52.437 50.624
744136 99.602 -52.445 50.582
744137 99.624 -52.453 50.541
744138 99.646 -52.461 50.498
744139 99.668 -52.469 50.457
744140 99.691 -52.478 50.414
744141 99.713 -52.485 50.372
744142 99.729 -52.496 50.331
744143 99.743 -52.506 50.289
744144 99.757 -52.518 50.248
744145 99.771 -52.528 50.206
744146 99.785 -52.540 50.165
744147 99.799 -52.550 50.124
744148 99.814 -52.561 50.083
744149 99.828 -52.572 50.043
744150 99.842 -52.583 50.003
744178 100.354 -52.941 49.209
744179 100.381 -52.950 49.183
744180 100.407 -52.959 49.160
744181 100.433 -52.968 49.135
744182 100.461 -52.980 49.111
744183 100.488 -52.998 49.085
744184 100.516 -53.014 49.058
744185 100.544 -53.032 49.031
744186 100.572 -53.049 49.003
744187 100.600 -53.066 48.972
744188 100.628 -53.083 48.942
744803 98.895 -52.201 51.967
744804 98.916 -52.204 51.929
744805 98.937 -52.207 51.891
744806 98.958 -52.210 51.855
744807 98.980 -52.213 51.816
744808 99.001 -52.216 51.777
744809 99.022 -52.219 51.738
744810 99.043 -52.221 51.699
744811 99.063 -52.227 51.660
744812 99.081 -52.233 51.621
744813 99.099 -52.239 51.581
744814 99.118 -52.246 51.539
744815 99.137 -52.253 51.498
744816 99.155 -52.259 51.457
744817 99.173 -52.266 51.415
744818 99.192 -52.276 51.373
744819 99.213 -52.285 51.331
744820 99.235 -52.293 51.288
744821 99.257 -52.301 51.245
744822 99.280 -52.308 51.202
744831 99.480 -52.381 50.828
744832 99.502 -52.390 50.787
744833 99.524 -52.397 50.747
744834 99.547 -52.405 50.706
744835 99.569 -52.413 50.666
744836 99.591 -52.421 50.625
744837 99.614 -52.430 50.583
744838 99.636 -52.437 50.542
744839 99.658 -52.446 50.499
744840 99.680 -52.453 50.457
744841 99.702 -52.462 50.415
744842 99.725 -52.470 50.373
744843 99.745 -52.478 50.331
744844 99.759 -52.490 50.289
744845 99.773 -52.500 50.248
744846 99.787 -52.512 50.206
744847 99.801 -52.522 50.165
744848 99.816 -52.534 50.125
744849 99.830 -52.544 50.084
744850 99.844 -52.555 50.044
744878 100.330 -52.911 49.227
744879 100.356 -52.920 49.201
744880 100.382 -52.929 49.176
744881 100.409 -52.938 49.152
744882 100.435 -52.948 49.128
744883 100.462 -52.959 49.104
744884 100.490 -52.976 49.079
744885 100.518 -52.994 49.053
744886 100.546 -53.010 49.026
744887 100.573 -53.028 48.998
744888 100.602 -53.045 48.969
744889 100.629 -53.062 48.938
744890 100.657 -53.079 48.906
745505 98.929 -52.192 51.936
745506 98.950 -52.195 51.898
745507 98.972 -52.197 51.861
745508 98.993 -52.200 51.822
745509 99.014 -52.203 51.783
745510 99.035 -52.206 51.744
745511 99.056 -52.209 51.705
745512 99.077 -52.212 51.666
745513 99.096 -52.218 51.626
745514 99.115 -52.225 51.586
745515 99.133 -52.231 51.545
745516 99.151 -52.238 51.503
745517 99.170 -52.244 51.462
745518 99.189 -52.251 51.420
745519 99.207 -52.258 51.379
745520 99.225 -52.269 51.336
745521 99.247 -52.277 51.293
745522 99.269 -52.285 51.250
745533 99.514 -52.374 50.789
745534 99.536 -52.382 50.749
745535 99.558 -52.390 50.708
745536 99.581 -52.398 50.667
745537 99.603 -52.405 50.625
745538 99.625 -52.414 50.584
745539 99.647 -52.422 50.542
745540 99.669 -52.430 50.499
745541 99.692 -52.438 50.456
745542 99.714 -52.446 50.414
745543 99.736 -52.454 50.373
745544 99.759 -52.462 50.331
745545 99.775 -52.472 50.290
745546 99.789 -52.484 50.248
745547 99.803 -52.494 50.207
745548 99.817 -52.506 50.166
745549 99.831 -52.516 50.125
745550 99.846 -52.527 50.084
745551 99.860 -52.538 50.044
745578 100.305 -52.882 49.246
745579 100.332 -52.891 49.220
745580 100.358 -52.900 49.193
745581 100.385 -52.909 49.168
745582 100.411 -52.918 49.145
745583 100.437 -52.927 49.122
745584 100.464 -52.938 49.098
745585 100.492 -52.955 49.073
745586 100.519 -52.973 49.048
745587 100.548 -52.989 49.021
745588 100.575 -53.007 48.993
745589 100.603 -53.024 48.964
745590 100.631 -53.041 48.933
745591 100.659 -53.058 48.902
745592 100.687 -53.076 48.868
746208 98.984 -52.185 51.867
746209 99.005 -52.188 51.828
746210 99.027 -52.191 51.788
746211 99.048 -52.193 51.750
746212 99.069 -52.196 51.711
746213 99.090 -52.199 51.671
746214 99.111 -52.203 51.632
746215 99.129 -52.210 51.592
746216 99.148 -52.216 51.551
746217 99.166 -52.223 51.509
746218 99.185 -52.230 51.468
746219 99.203 -52.236 51.426
746220 99.222 -52.242 51.384
746221 99.240 -52.252 51.341
746222 99.259 -52.261 51.297
746234 99.525 -52.358 50.791
746235 99.548 -52.366 50.750
746236 99.570 -52.374 50.709
746237 99.592 -52.382 50.668
746238 99.615 -52.390 50.626
746239 99.637 -52.399 50.584
746240 99.659 -52.406 50.542
746241 99.682 -52.415 50.498
746242 99.704 -52.422 50.456
746243 99.726 -52.431 50.414
746244 99.748 -52.439 50.373
746245 99.771 -52.446 50.332
746246 99.791 -52.456 50.290
746247 99.805 -52.466 50.248
746248 99.819 -52.478 50.207
746249 99.833 -52.488 50.166
746250 99.847 -52.499 50.125
746251 99.861 -52.510 50.084
746278 100.280 -52.852 49.269
746279 100.307 -52.861 49.240
746280 100.333 -52.870 49.213
746281 100.360 -52.879 49.186
746282 100.387 -52.888 49.160
746283 100.413 -52.897 49.138
746284 100.439 -52.907 49.116
746285 100.466 -52.916 49.093
746286 100.494 -52.934 49.068
746287 100.522 -52.951 49.043
746288 100.549 -52.968 49.017
746289 100.577 -52.985 48.989
746290 100.605 -53.003 48.960
746291 100.633 -53.020 48.930
746292 100.661 -53.037 48.898
746293 100.688 -53.054 48.864
746294 100.716 -53.071 48.829
746910 99.018 -52.175 51.835
746911 99.040 -52.178 51.795
746912 99.061 -52.181 51.756
746913 99.082 -52.184 51.717
746914 99.103 -52.186 51.677
746915 99.125 -52.189 51.638
746916 99.144 -52.195 51.598
746917 99.162 -52.202 51.557
746918 99.181 -52.208 51.515
746919 99.200 -52.214 51.473
746920 99.218 -52.221 51.431
746921 99.236 -52.227 51.388
746922 99.255 -52.234 51.345
746936 99.560 -52.351 50.753
746937 99.582 -52.358 50.711
746938 99.604 -52.367 50.669
746939 99.627 -52.374 50.627
746940 99.649 -52.383 50.586
746941 99.671 -52.391 50.543
746942 99.693 -52.399 50.499
746943 99.715 -52.407 50.456
746944 99.738 -52.415 50.414
746945 99.760 -52.423 50.373
746946 99.782 -52.431 50.332
746947 99.804 -52.439 50.291
746948 99.821 -52.449 50.248
746949 99.835 -52.460 50.207
746950 99.849 -52.471 50.166
746951 99.863 -52.482 50.126
746978 100.261 -52.820 49.293
746979 100.282 -52.832 49.263
746980 100.309 -52.841 49.234
746981 100.335 -52.850 49.207
746982 100.362 -52.859 49.181
746983 100.388 -52.868 49.156
746984 100.415 -52.877 49.135
746985 100.441 -52.886 49.112
746986 100.468 -52.895 49.088
746987 100.495 -52.913 49.064
746988 100.523 -52.930 49.039
746989 100.551 -52.947 49.012
746990 100.579 -52.964 48.985
746991 100.607 -52.982 48.956
746992 100.634 -52.998 48.926
746993 100.662 -53.016 48.894
746994 100.690 -53.033 48.861
746995 100.718 -53.050 48.827
746996 100.746 -53.067 48.791
747614 99.095 -52.171 51.724
747615 99.116 -52.174 51.683
747616 99.137 -52.177 51.645
747617 99.159 -52.180 51.604
747618 99.178 -52.186 51.562
747619 99.196 -52.192 51.520
747639 99.616 -52.351 50.671
747640 99.638 -52.359 50.629
747641 99.660 -52.367 50.587
747642 99.683 -52.375 50.544
747643 99.705 -52.384 50.501
747644 99.727 -52.391 50.458
747645 99.750 -52.399 50.416
747646 99.772 -52.408 50.374
747647 99.794 -52.415 50.333
747648 99.816 -52.424 50.291
747649 99.837 -52.432 50.249
747650 99.851 -52.443 50.208
747651 99.865 -52.454 50.167
747678 100.254 -52.783 49.322
747679 100.271 -52.796 49.290
747680 100.289 -52.809 49.259
747681 100.311 -52.820 49.231
747682 100.337 -52.829 49.204
747683 100.363 -52.838 49.178
747684 100.390 -52.847 49.154
747685 100.416 -52.857 49.132
747686 100.443 -52.866 49.109
747687 100.470 -52.875 49.085
747688 100.497 -52.891 49.060
747689 100.525 -52.909 49.035
747690 100.553 -52.926 49.009
747691 100.580 -52.943 48.981
747692 100.609 -52.960 48.953
747693 100.636 -52.978 48.923
747694 100.664 -52.995 48.891
747695 100.692 -53.012 48.859
747696 100.720 -53.029 48.824
747697 100.748 -53.046 48.789
747698 100.775 -53.064 48.753
748343 99.695 -52.360 50.546
748344 99.717 -52.368 50.502
748345 99.739 -52.376 50.460
748379 100.268 -52.755 49.320
748380 100.283 -52.771 49.287
748381 100.299 -52.785 49.256
748382 100.317 -52.798 49.228
748383 100.339 -52.809 49.202
748384 100.365 -52.818 49.176
748385 100.392 -52.827 49.151
748386 100.418 -52.836 49.129
748387 100.445 -52.845 49.107
748388 100.471 -52.854 49.083
748389 100.499 -52.870 49.058
748390 100.527 -52.888 49.033
748391 100.555 -52.905 49.007
748392 100.582 -52.922 48.979
748393 100.610 -52.939 48.950
748394 100.638 -52.957 48.920
748395 100.666 -52.973 48.889
748396 100.694 -52.991 48.856
748397 100.721 -53.008 48.823
748398 100.749 -53.025 48.788
748399 100.777 -53.042 48.752
749079 100.266 -52.712 49.353
749080 100.281 -52.728 49.318
749081 100.296 -52.744 49.285
749082 100.312 -52.759 49.255
749083 100.327 -52.774 49.227
749084 100.345 -52.787 49.200
749085 100.367 -52.797 49.175
749086 100.394 -52.807 49.149
749087 100.420 -52.816 49.127
749088 100.446 -52.825 49.105
749089 100.473 -52.834 49.081
749090 100.500 -52.849 49.057
749091 100.528 -52.866 49.030
749092 100.556 -52.884 49.004
749093 100.584 -52.901 48.977
749094 100.612 -52.918 48.949
749095 100.640 -52.935 48.919
749096 100.668 -52.952 48.887
749097 100.695 -52.970 48.855
749098 100.723 -52.987 48.821
749099 100.751 -53.004 48.787
749100 100.779 -53.021 48.752
749101 100.807 -53.038 48.715
749780 100.280 -52.687 49.352
749781 100.294 -52.700 49.317
749782 100.310 -52.716 49.284
749783 100.325 -52.731 49.254
749784 100.340 -52.747 49.226
749785 100.355 -52.763 49.199
749786 100.373 -52.775 49.173
749787 100.396 -52.786 49.148
749788 100.422 -52.795 49.126
749789 100.449 -52.804 49.103
749790 100.475 -52.813 49.080
749791 100.502 -52.828 49.055
749792 100.530 -52.845 49.029
749793 100.558 -52.863 49.002
749794 100.586 -52.879 48.976
749795 100.614 -52.897 48.947
749796 100.641 -52.914 48.916
749797 100.669 -52.931 48.885
749798 100.697 -52.948 48.852
749799 100.725 -52.966 48.819
749800 100.753 -52.982 48.786
749801 100.780 -53.000 48.751
749802 100.809 -53.017 48.715
750481 100.296 -52.671 49.352
750482 100.310 -52.681 49.316
750483 100.324 -52.693 49.283
750484 100.339 -52.703 49.253
750485 100.354 -52.719 49.225
750486 100.369 -52.735 49.197
750487 100.384 -52.750 49.172
750488 100.401 -52.764 49.147
750489 100.424 -52.775 49.125
750490 100.451 -52.784 49.102
750491 100.477 -52.794 49.079
750492 100.504 -52.807 49.054
750493 100.532 -52.824 49.028
750494 100.560 -52.841 49.001
750495 100.587 -52.859 48.974
750496 100.615 -52.876 48.944
750497 100.643 -52.893 48.914
750498 100.671 -52.910 48.883
750499 100.699 -52.927 48.849
750500 100.726 -52.944 48.817
750501 100.755 -52.961 48.784
750502 100.783 -52.977 48.750
750503 100.812 -52.993 48.715
751182 100.312 -52.653 49.352
751183 100.326 -52.665 49.316
751184 100.340 -52.675 49.283
751185 100.354 -52.687 49.253
751186 100.368 -52.697 49.224
751187 100.382 -52.708 49.197
751188 100.398 -52.723 49.171
751189 100.413 -52.738 49.147
751190 100.429 -52.753 49.124
751191 100.452 -52.763 49.101
751192 100.479 -52.773 49.080
751193 100.506 -52.786 49.055
751194 100.534 -52.803 49.027
751195 100.561 -52.820 48.999
751196 100.590 -52.838 48.971
751197 100.617 -52.854 48.942
751198 100.645 -52.872 48.911
751199 100.673 -52.889 48.880
751200 100.701 -52.905 48.847
751201 100.729 -52.921 48.814
751202 100.758 -52.937 48.782
751203 100.786 -52.952 48.748
751204 100.815 -52.968 48.713
751883 100.328 -52.637 49.352
751884 100.342 -52.647 49.317
751885 100.356 -52.659 49.284
751886 100.370 -52.669 49.253
751887 100.385 -52.681 49.223
751888 100.399 -52.691 49.196
751889 100.413 -52.702 49.170
751890 100.427 -52.713 49.146
751891 100.441 -52.727 49.122
751892 100.457 -52.742 49.101
751893 100.481 -52.753 49.080
751894 100.507 -52.765 49.055
751895 100.536 -52.782 49.026
751896 100.563 -52.799 48.996
751897 100.591 -52.816 48.968
751898 100.619 -52.833 48.938
751899 100.647 -52.848 48.908
751900 100.676 -52.864 48.876
751901 100.705 -52.880 48.844
751902 100.733 -52.896 48.812
751903 100.762 -52.912 48.780
751904 100.790 -52.927 48.747
751905 100.819 -52.943 48.713
752584 100.344 -52.619 49.354
752585 100.358 -52.631 49.318
752586 100.372 -52.641 49.285
752587 100.386 -52.653 49.253
752588 100.400 -52.663 49.223
752589 100.414 -52.674 49.195
752590 100.429 -52.685 49.169
752591 100.443 -52.696 49.144
752592 100.457 -52.707 49.121
752593 100.471 -52.718 49.099
752594 100.485 -52.731 49.078
752595 100.509 -52.744 49.054
752596 100.537 -52.760 49.025
752597 100.566 -52.776 48.994
752598 100.594 -52.792 48.965
752599 100.623 -52.808 48.935
752600 100.651 -52.823 48.906
752601 100.680 -52.839 48.873
752602 100.708 -52.855 48.842
752603 100.738 -52.872 48.811
752604 100.767 -52.888 48.779
752605 100.795 -52.904 48.747
752606 100.825 -52.919 48.714
753285 100.360 -52.603 49.356
753286 100.374 -52.613 49.320
753287 100.388 -52.625 49.286
753288 100.402 -52.635 49.253
753289 100.416 -52.647 49.223
753290 100.430 -52.657 49.194
753291 100.444 -52.668 49.167
753292 100.459 -52.679 49.142
753293 100.473 -52.690 49.119
753294 100.487 -52.701 49.095
753295 100.501 -52.712 49.074
753296 100.516 -52.722 49.052
753297 100.546 -52.738 49.023
753298 100.574 -52.754 48.992
753299 100.604 -52.770 48.964
753300 100.633 -52.787 48.934
753301 100.662 -52.803 48.903
753302 100.691 -52.819 48.872
753303 100.720 -52.835 48.841
753304 100.749 -52.850 48.810
753305 100.778 -52.866 48.779
753306 100.807 -52.883 48.748
753307 100.836 -52.899 48.715
753986 100.376 -52.585 49.358
753987 100.390 -52.597 49.320
753988 100.404 -52.607 49.285
753989 100.418 -52.619 49.253
753990 100.432 -52.629 49.222
753991 100.446 -52.640 49.192
753992 100.461 -52.651 49.164
753993 100.475 -52.662 49.138
753994 100.489 -52.673 49.115
753995 100.503 -52.684 49.091
753996 100.517 -52.695 49.068
753997 100.532 -52.706 49.044
753998 100.557 -52.718 49.017
753999 100.586 -52.734 48.988
754000 100.615 -52.750 48.960
754001 100.644 -52.766 48.931
754002 100.673 -52.782 48.901
754003 100.702 -52.798 48.871
754004 100.731 -52.814 48.840
754005 100.760 -52.830 48.811
754006 100.789 -52.846 48.780
754007 100.818 -52.862 48.749
754008 100.847 -52.878 48.718
754687 100.392 -52.568 49.359
754688 100.406 -52.579 49.321
754689 100.420 -52.590 49.285
754690 100.434 -52.601 49.252
754691 100.448 -52.612 49.220
754692 100.462 -52.623 49.190
754693 100.477 -52.634 49.161
754694 100.491 -52.645 49.134
754695 100.505 -52.656 49.109
754696 100.519 -52.667 49.086
754697 100.533 -52.678 49.062
754698 100.548 -52.688 49.036
754699 100.568 -52.698 49.009
754700 100.597 -52.713 48.982
754701 100.626 -52.729 48.955
754702 100.655 -52.745 48.927
754703 100.684 -52.761 48.899
754704 100.713 -52.777 48.869
754705 100.743 -52.793 48.840
754706 100.771 -52.810 48.811
754707 100.800 -52.825 48.782
754708 100.830 -52.841 48.751
754709 100.858 -52.857 48.720
755389 100.422 -52.562 49.321
755390 100.436 -52.573 49.285
755391 100.450 -52.584 49.251
755392 100.464 -52.595 49.218
755393 100.478 -52.606 49.187
755394 100.492 -52.617 49.158
755395 100.506 -52.628 49.130
755396 100.521 -52.639 49.104
755397 100.535 -52.650 49.079
755398 100.549 -52.661 49.055
755399 100.564 -52.672 49.029
755400 100.583 -52.681 49.003
755401 100.609 -52.692 48.976
755402 100.637 -52.708 48.950
755403 100.667 -52.725 48.924
755404 100.696 -52.741 48.895
755405 100.724 -52.756 48.868
755406 100.754 -52.772 48.840
755407 100.783 -52.788 48.811
755408 100.812 -52.804 48.782
755409 100.841 -52.821 48.753
756090 100.438 -52.545 49.321
756091 100.452 -52.556 49.284
756092 100.466 -52.567 49.249
756093 100.480 -52.578 49.215
756094 100.494 -52.589 49.184
756095 100.508 -52.600 49.153
756096 100.522 -52.611 49.125
756097 100.537 -52.622 49.098
756098 100.551 -52.633 49.073
756099 100.565 -52.644 49.048
756100 100.579 -52.654 49.022
756101 100.599 -52.664 48.997
756102 100.619 -52.673 48.971
756103 100.649 -52.687 48.946
756104 100.678 -52.703 48.920
756105 100.707 -52.719 48.893
756106 100.736 -52.736 48.866
756107 100.765 -52.752 48.839
756108 100.794 -52.768 48.811
756109 100.823 -52.784 48.783
756110 100.852 -52.800 48.756
756792 100.468 -52.539 49.283
756793 100.482 -52.550 49.247
756794 100.496 -52.561 49.213
756795 100.510 -52.572 49.180
756796 100.525 -52.583 49.149
756797 100.539 -52.593 49.120
756798 100.553 -52.605 49.092
756799 100.567 -52.615 49.067
756800 100.581 -52.627 49.042
756801 100.596 -52.637 49.016
756802 100.614 -52.647 48.991
756803 100.633 -52.656 48.966
756804 100.660 -52.667 48.942
756805 100.689 -52.683 48.916
756806 100.718 -52.699 48.892
756807 100.747 -52.715 48.865
756808 100.776 -52.731 48.838
756809 100.805 -52.747 48.811
756810 100.834 -52.763 48.785
757493 100.484 -52.522 49.282
757494 100.498 -52.533 49.246
757495 100.512 -52.544 49.210
757496 100.526 -52.555 49.176
757497 100.540 -52.565 49.145
757498 100.554 -52.577 49.115
757499 100.568 -52.587 49.087
757500 100.582 -52.599 49.061
757501 100.597 -52.609 49.037
757502 100.611 -52.620 49.011
757503 100.630 -52.630 48.986
757504 100.649 -52.639 48.962
757505 100.671 -52.649 48.938
757506 100.700 -52.662 48.914
757507 100.729 -52.678 48.890
757508 100.758 -52.694 48.864
757509 100.787 -52.710 48.839
757510 100.817 -52.726 48.814
757511 100.845 -52.743 48.788
758195 100.514 -52.516 49.244
758196 100.528 -52.527 49.208
758197 100.542 -52.537 49.174
758198 100.556 -52.549 49.143
758199 100.570 -52.559 49.111
758200 100.584 -52.571 49.083
758201 100.599 -52.581 49.057
758202 100.613 -52.593 49.032
758203 100.627 -52.603 49.007
758204 100.646 -52.612 48.983
758205 100.665 -52.622 48.960
758206 100.684 -52.631 48.937
758207 100.711 -52.642 48.913
758208 100.741 -52.658 48.890
758209 100.770 -52.674 48.866
758210 100.798 -52.690 48.841
758211 100.828 -52.706 48.817
758897 100.544 -52.510 49.206
758898 100.558 -52.521 49.172
758899 100.572 -52.531 49.140
758900 100.586 -52.543 49.108
758901 100.601 -52.553 49.080
758902 100.615 -52.565 49.054
758903 100.629 -52.575 49.029
758904 100.643 -52.587 49.004
758905 100.662 -52.596 48.980
758906 100.681 -52.605 48.958
758907 100.700 -52.614 48.936
758908 100.722 -52.624 48.913
758909 100.752 -52.637 48.890
758910 100.781 -52.653 48.868
758911 100.810 -52.669 48.844
759600 100.588 -52.515 49.137
759601 100.602 -52.525 49.105
759602 100.616 -52.537 49.078
759603 100.630 -52.547 49.050
759604 100.644 -52.559 49.025
759605 100.658 -52.569 49.001
759606 100.678 -52.578 48.978
759607 100.697 -52.587 48.955
759608 100.715 -52.597 48.934
759609 100.735 -52.606 48.912
759610 100.763 -52.617 48.891
759611 100.792 -52.632 48.869
760302 100.618 -52.509 49.102
760303 100.632 -52.519 49.074
760304 100.646 -52.531 49.047
760305 100.661 -52.541 49.022
760306 100.675 -52.553 48.998
760307 100.693 -52.562 48.974
760308 100.712 -52.571 48.953
760309 100.732 -52.580 48.933
760310 100.750 -52.589 48.912
770837 101.263 -52.462 48.788
770838 101.292 -52.478 48.770
770839 101.321 -52.493 48.750
770840 101.350 -52.509 48.731
770841 101.379 -52.525 48.712
770842 101.408 -52.542 48.692
770843 101.437 -52.558 48.671
770844 101.466 -52.574 48.650
770845 101.495 -52.590 48.628
771536 101.215 -52.413 48.833
771537 101.245 -52.424 48.821
771538 101.274 -52.440 48.805
771539 101.303 -52.457 48.786
771540 101.332 -52.473 48.767
771541 101.361 -52.489 48.748
771542 101.390 -52.505 48.728
771543 101.419 -52.521 48.709
771544 101.448 -52.537 48.688
771545 101.477 -52.553 48.666
771546 101.506 -52.569 48.645
771547 101.535 -52.585 48.623
772235 101.171 -52.374 48.871
772236 101.196 -52.384 48.861
772237 101.226 -52.396 48.849
772238 101.256 -52.407 48.836
772239 101.285 -52.420 48.821
772240 101.314 -52.436 48.803
772241 101.343 -52.452 48.785
772242 101.373 -52.468 48.766
772243 101.401 -52.484 48.746
772244 101.430 -52.500 48.726
772245 101.460 -52.516 48.705
772246 101.488 -52.532 48.683
772247 101.518 -52.549 48.661
772248 101.547 -52.565 48.641
772249 101.575 -52.581 48.618
772935 101.168 -52.347 48.892
772936 101.187 -52.356 48.886
772937 101.207 -52.366 48.877
772938 101.237 -52.377 48.866
772939 101.266 -52.389 48.853
772940 101.296 -52.400 48.838
772941 101.326 -52.415 48.822
772942 101.354 -52.431 48.803
772943 101.384 -52.447 48.784
772944 101.413 -52.464 48.765
772945 101.442 -52.480 48.745
772946 101.471 -52.496 48.724
772947 101.500 -52.512 48.702
772948 101.529 -52.528 48.678
772949 101.558 -52.543 48.657
772950 101.587 -52.560 48.634
772951 101.616 -52.576 48.611
773635 101.165 -52.321 48.910
773636 101.183 -52.330 48.907
773637 101.202 -52.340 48.902
773638 101.221 -52.349 48.894
773639 101.247 -52.359 48.884
773640 101.277 -52.371 48.871
773641 101.307 -52.382 48.856
773642 101.337 -52.395 48.841
773643 101.366 -52.411 48.822
773644 101.395 -52.427 48.804
773645 101.424 -52.443 48.785
773646 101.453 -52.459 48.765
773647 101.482 -52.475 48.743
773648 101.511 -52.491 48.720
773649 101.540 -52.507 48.696
773650 101.569 -52.523 48.674
773651 101.598 -52.539 48.651
773652 101.627 -52.555 48.627
773653 101.656 -52.572 48.603
774335 101.162 -52.295 48.925
774336 101.180 -52.304 48.924
774337 101.199 -52.313 48.923
774338 101.218 -52.323 48.918
774339 101.237 -52.332 48.911
774340 101.258 -52.341 48.902
774341 101.288 -52.352 48.891
774342 101.318 -52.364 48.877
774343 101.348 -52.375 48.861
774344 101.377 -52.390 48.842
774345 101.406 -52.406 48.825
774346 101.435 -52.422 48.806
774347 101.464 -52.438 48.785
774348 101.493 -52.454 48.762
774349 101.522 -52.470 48.739
774350 101.551 -52.487 48.715
774351 101.581 -52.503 48.692
774352 101.609 -52.518 48.669
774353 101.638 -52.534 48.644
774354 101.668 -52.550 48.619
775035 101.158 -52.269 48.936
775036 101.177 -52.278 48.939
775037 101.196 -52.287 48.940
775038 101.215 -52.296 48.938
775039 101.234 -52.305 48.935
775040 101.253 -52.315 48.929
775041 101.272 -52.324 48.921
775042 101.299 -52.335 48.912
775043 101.329 -52.346 48.898
775044 101.358 -52.358 48.882
775045 101.388 -52.369 48.864
775046 101.417 -52.385 48.846
775047 101.447 -52.402 48.827
775048 101.475 -52.418 48.805
775049 101.504 -52.434 48.782
775050 101.534 -52.449 48.758
775051 101.562 -52.465 48.734
775052 101.592 -52.481 48.710
775053 101.621 -52.498 48.686
775054 101.650 -52.514 48.660
775055 101.679 -52.530 48.635
775056 101.708 -52.546 48.609
775734 101.136 -52.233 48.944
775735 101.155 -52.242 48.946
775736 101.174 -52.252 48.949
775737 101.193 -52.261 48.953
775738 101.212 -52.270 48.955
775739 101.231 -52.280 48.955
775740 101.250 -52.289 48.952
775741 101.269 -52.298 48.947
775742 101.288 -52.307 48.941
775743 101.310 -52.317 48.932
775744 101.339 -52.328 48.919
775745 101.370 -52.340 48.903
775746 101.399 -52.351 48.886
775747 101.428 -52.365 48.868
775748 101.458 -52.380 48.848
775749 101.487 -52.396 48.826
775750 101.516 -52.413 48.803
775751 101.545 -52.429 48.778
775752 101.574 -52.445 48.753
775753 101.603 -52.461 48.728
775754 101.632 -52.477 48.703
775755 101.661 -52.493 48.677
775756 101.690 -52.509 48.651
775757 101.719 -52.525 48.624
776434 101.133 -52.207 48.953
776435 101.152 -52.216 48.955
776436 101.171 -52.226 48.957
776437 101.190 -52.235 48.962
776438 101.209 -52.244 48.966
776439 101.228 -52.253 48.969
776440 101.247 -52.262 48.971
776441 101.266 -52.271 48.970
776442 101.284 -52.280 48.966
776443 101.304 -52.290 48.962
776444 101.323 -52.299 48.953
776445 101.350 -52.310 48.940
776446 101.380 -52.321 48.926
776447 101.410 -52.333 48.907
776448 101.440 -52.344 48.889
776449 101.469 -52.360 48.869
776450 101.498 -52.376 48.848
776451 101.527 -52.392 48.824
776452 101.556 -52.408 48.799
776453 101.585 -52.424 48.773
776454 101.614 -52.440 48.747
776455 101.642 -52.457 48.721
776456 101.671 -52.473 48.695
776457 101.699 -52.490 48.667
776458 101.728 -52.506 48.640
777135 101.149 -52.190 48.964
777136 101.168 -52.199 48.965
777137 101.187 -52.208 48.969
777138 101.206 -52.217 48.975
777139 101.225 -52.227 48.980
777140 101.244 -52.236 48.985
777141 101.262 -52.246 48.988
777142 101.281 -52.255 48.988
777143 101.301 -52.264 48.986
777144 101.319 -52.273 48.982
777145 101.338 -52.282 48.974
777146 101.361 -52.292 48.964
777147 101.391 -52.303 48.949
777148 101.421 -52.315 48.931
777149 101.451 -52.326 48.913
777150 101.480 -52.340 48.892
777151 101.509 -52.355 48.870
777152 101.537 -52.372 48.846
777153 101.565 -52.389 48.820
777154 101.594 -52.405 48.792
777155 101.623 -52.421 48.766
777156 101.651 -52.437 48.739
777157 101.680 -52.454 48.711
777158 101.708 -52.470 48.683
777159 101.737 -52.487 48.654
777835 101.145 -52.164 48.975
777836 101.165 -52.173 48.975
777837 101.184 -52.183 48.977
777838 101.202 -52.192 48.981
777839 101.221 -52.201 48.988
777840 101.241 -52.210 48.995
777841 101.259 -52.219 49.002
777842 101.278 -52.228 49.005
777843 101.297 -52.237 49.007
777844 101.316 -52.247 49.006
777845 101.335 -52.256 49.004
777846 101.354 -52.265 48.996
777847 101.373 -52.274 48.987
777848 101.402 -52.286 48.973
777849 101.431 -52.297 48.955
777850 101.461 -52.308 48.937
777851 101.488 -52.321 48.916
777852 101.517 -52.336 48.892
777853 101.546 -52.353 48.867
777854 101.574 -52.369 48.840
777855 101.603 -52.386 48.812
777856 101.631 -52.402 48.785
777857 101.659 -52.418 48.756
777858 101.688 -52.435 48.727
777859 101.717 -52.451 48.698
777860 101.745 -52.468 48.669
778535 101.142 -52.138 48.988
778536 101.162 -52.147 48.986
778537 101.180 -52.156 48.987
778538 101.199 -52.165 48.990
778539 101.218 -52.174 48.994
778540 101.237 -52.183 49.002
778541 101.256 -52.193 49.011
778542 101.275 -52.202 49.019
778543 101.294 -52.211 49.024
778544 101.313 -52.221 49.026
778545 101.332 -52.230 49.027
778546 101.351 -52.239 49.026
778547 101.370 -52.248 49.020
778548 101.389 -52.257 49.011
778549 101.412 -52.268 48.998
778550 101.443 -52.279 48.980
778551 101.471 -52.291 48.961
778552 101.495 -52.308 48.940
778553 101.526 -52.318 48.916
778554 101.554 -52.334 48.889
778555 101.583 -52.350 48.861
778556 101.611 -52.367 48.833
778557 101.640 -52.383 48.805
778558 101.669 -52.399 48.774
778559 101.697 -52.415 48.744
778560 101.726 -52.432 48.714
778561 101.754 -52.449 48.685
779236 101.159 -52.120 49.000
779237 101.177 -52.130 48.998
779238 101.196 -52.139 48.999
779239 101.215 -52.148 49.003
779240 101.234 -52.158 49.008
779241 101.253 -52.167 49.017
779242 101.272 -52.176 49.027
779243 101.291 -52.185 49.035
779244 101.310 -52.194 49.042
779245 101.329 -52.204 49.046
779246 101.348 -52.213 49.049
779247 101.367 -52.222 49.050
779248 101.386 -52.231 49.045
779249 101.405 -52.240 49.035
779250 101.423 -52.249 49.023
779251 101.453 -52.261 49.006
779252 101.479 -52.275 48.987
779253 101.503 -52.291 48.964
779254 101.534 -52.299 48.939
779255 101.563 -52.315 48.912
779256 101.592 -52.331 48.883
779257 101.620 -52.348 48.854
779258 101.649 -52.364 48.825
779259 101.677 -52.380 48.793
779260 101.705 -52.396 48.762
779261 101.734 -52.413 48.731
779262 101.763 -52.429 48.700
779937 101.174 -52.104 49.012
779938 101.193 -52.113 49.011
779939 101.212 -52.122 49.012
779940 101.231 -52.131 49.016
779941 101.250 -52.140 49.022
779942 101.269 -52.150 49.031
779943 101.288 -52.159 49.042
779944 101.307 -52.168 49.052
779945 101.326 -52.177 49.059
779946 101.345 -52.186 49.065
779947 101.363 -52.195 49.071
779948 101.383 -52.205 49.075
779949 101.402 -52.214 49.071
779950 101.420 -52.224 49.060
779951 101.440 -52.233 49.047
779952 101.463 -52.243 49.032
779953 101.488 -52.259 49.012
779954 101.512 -52.275 48.989
779955 101.543 -52.280 48.963
779956 101.571 -52.296 48.935
779957 101.600 -52.312 48.906
779958 101.628 -52.328 48.876
779959 101.657 -52.345 48.845
779960 101.686 -52.361 48.813
779961 101.714 -52.377 48.781
779962 101.743 -52.394 48.749
779963 101.771 -52.410 48.716
780638 101.190 -52.086 49.025
780639 101.209 -52.096 49.023
780640 101.228 -52.105 49.024
780641 101.247 -52.114 49.029
780642 101.266 -52.123 49.035
780643 101.285 -52.133 49.045
780644 101.304 -52.142 49.055
780645 101.323 -52.151 49.067
780646 101.341 -52.161 49.077
780647 101.360 -52.170 49.084
780648 101.379 -52.179 49.093
780649 101.398 -52.188 49.101
780650 101.417 -52.197 49.097
780651 101.437 -52.206 49.086
780652 101.455 -52.215 49.072
780653 101.472 -52.227 49.059
780654 101.496 -52.243 49.038
780655 101.521 -52.259 49.014
780656 101.555 -52.261 48.987
780657 101.580 -52.277 48.958
780658 101.609 -52.293 48.927
780659 101.637 -52.309 48.897
780660 101.666 -52.326 48.866
780661 101.694 -52.342 48.832
780662 101.723 -52.358 48.799
780663 101.751 -52.374 48.766
780664 101.780 -52.391 48.732
781339 101.206 -52.070 49.039
781340 101.225 -52.079 49.036
781341 101.244 -52.088 49.037
781342 101.263 -52.097 49.040
781343 101.281 -52.107 49.047
781344 101.301 -52.116 49.057
781345 101.320 -52.125 49.068
781346 101.338 -52.134 49.079
781347 101.357 -52.143 49.090
781348 101.376 -52.152 49.099
781349 101.395 -52.161 49.110
781350 101.414 -52.170 49.121
781351 101.433 -52.180 49.119
781352 101.452 -52.189 49.108
781353 101.471 -52.199 49.096
781354 101.488 -52.214 49.081
781355 101.504 -52.227 49.061
781356 101.529 -52.243 49.037
781357 101.557 -52.255 49.010
781358 101.589 -52.258 48.980
781359 101.617 -52.274 48.949
781360 101.646 -52.290 48.918
781361 101.675 -52.306 48.886
781362 101.703 -52.323 48.852
781363 101.732 -52.340 48.818
781364 101.760 -52.355 48.784
781365 101.788 -52.372 48.749
782040 101.221 -52.053 49.052
782041 101.241 -52.062 49.048
782042 101.259 -52.071 49.049
782043 101.278 -52.080 49.052
782044 101.298 -52.089 49.058
782045 101.316 -52.098 49.068
782046 101.335 -52.108 49.080
782047 101.354 -52.117 49.091
782048 101.373 -52.127 49.102
782049 101.392 -52.136 49.112
782050 101.411 -52.145 49.124
782051 101.430 -52.154 49.135
782052 101.449 -52.163 49.136
782053 101.468 -52.172 49.128
782054 101.488 -52.184 49.117
782055 101.504 -52.201 49.103
782056 101.513 -52.211 49.083
782057 101.538 -52.227 49.060
782058 101.562 -52.243 49.032
782059 101.598 -52.239 49.003
782060 101.626 -52.255 48.971
782061 101.655 -52.271 48.939
782062 101.683 -52.287 48.906
782063 101.711 -52.304 48.872
782064 101.740 -52.320 48.837
782065 101.769 -52.336 48.801
782066 101.797 -52.353 48.766
782067 101.826 -52.369 48.733
782741 101.238 -52.036 49.065
782742 101.256 -52.045 49.060
782743 101.275 -52.054 49.061
782744 101.294 -52.064 49.064
782745 101.313 -52.073 49.069
782746 101.332 -52.082 49.078
782747 101.351 -52.091 49.090
782748 101.370 -52.100 49.102
782749 101.389 -52.109 49.113
782750 101.408 -52.118 49.124
782751 101.427 -52.127 49.135
782752 101.446 -52.137 49.144
782753 101.465 -52.146 49.149
782754 101.484 -52.155 49.145
782755 101.505 -52.170 49.136
782756 101.521 -52.187 49.123
782757 101.529 -52.198 49.105
782758 101.546 -52.211 49.081
782759 101.570 -52.227 49.054
782760 101.608 -52.220 49.024
782761 101.634 -52.236 48.993
782762 101.663 -52.252 48.960
782763 101.692 -52.268 48.927
782764 101.721 -52.285 48.892
782765 101.749 -52.301 48.857
782766 101.777 -52.318 48.820
782767 101.806 -52.333 48.784
782768 101.834 -52.350 48.749
782769 101.863 -52.366 48.716
783443 101.272 -52.028 49.074
783444 101.291 -52.037 49.074
783445 101.310 -52.046 49.076
783446 101.329 -52.055 49.082
783447 101.348 -52.064 49.090
783448 101.367 -52.074 49.101
783449 101.386 -52.083 49.114
783450 101.405 -52.092 49.125
783451 101.424 -52.101 49.135
783452 101.443 -52.111 49.146
783453 101.462 -52.120 49.156
783454 101.481 -52.129 49.161
783455 101.500 -52.141 49.159
783456 101.522 -52.157 49.156
783457 101.538 -52.174 49.145
783458 101.545 -52.184 49.128
783459 101.554 -52.195 49.105
783460 101.579 -52.211 49.077
783461 101.622 -52.203 49.046
783462 101.643 -52.217 49.014
783463 101.672 -52.233 48.980
783464 101.700 -52.249 48.947
783465 101.729 -52.265 48.911
783466 101.757 -52.282 48.875
783467 101.786 -52.299 48.838
783468 101.815 -52.315 48.802
783469 101.843 -52.331 48.766
783470 101.872 -52.347 48.732
783471 101.900 -52.364 48.699
784144 101.288 -52.011 49.089
784145 101.307 -52.020 49.087
784146 101.326 -52.029 49.089
784147 101.345 -52.039 49.094
784148 101.364 -52.048 49.103
784149 101.383 -52.057 49.113
784150 101.402 -52.066 49.125
784151 101.420 -52.075 49.136
784152 101.440 -52.084 49.147
784153 101.459 -52.094 49.157
784154 101.477 -52.103 49.167
784155 101.496 -52.112 49.174
784156 101.518 -52.127 49.175
784157 101.538 -52.144 49.176
784158 101.554 -52.161 49.168
784159 101.562 -52.171 49.152
784160 101.569 -52.181 49.127
784161 101.588 -52.195 49.099
784162 101.622 -52.199 49.068
784163 101.652 -52.197 49.034
784164 101.680 -52.214 49.000
784165 101.709 -52.230 48.966
784166 101.738 -52.246 48.930
784167 101.766 -52.263 48.893
784168 101.795 -52.279 48.857
784169 101.823 -52.296 48.820
784170 101.852 -52.311 48.784
784171 101.880 -52.328 48.749
784172 101.909 -52.345 48.715
784173 101.938 -52.361 48.682
784845 101.304 -51.994 49.104
784846 101.323 -52.003 49.102
784847 101.342 -52.012 49.103
784848 101.360 -52.021 49.107
784849 101.379 -52.031 49.116
784850 101.399 -52.040 49.126
784851 101.417 -52.049 49.137
784852 101.437 -52.058 49.148
784853 101.455 -52.067 49.158
784854 101.474 -52.076 49.168
784855 101.493 -52.086 49.178
784856 101.513 -52.097 49.186
784857 101.534 -52.112 49.190
784858 101.555 -52.131 49.196
784859 101.571 -52.148 49.192
784860 101.579 -52.158 49.175
784861 101.586 -52.168 49.149
784862 101.596 -52.179 49.120
784863 101.620 -52.195 49.088
784864 101.661 -52.179 49.055
784865 101.689 -52.195 49.021
784866 101.717 -52.211 48.986
784867 101.746 -52.227 48.949
784868 101.775 -52.244 48.912
784869 101.803 -52.260 48.875
784870 101.832 -52.277 48.837
784871 101.861 -52.293 48.801
784872 101.889 -52.309 48.766
784873 101.917 -52.325 48.732
784874 101.946 -52.342 48.698
785546 101.320 -51.977 49.120
785547 101.338 -51.986 49.117
785548 101.357 -51.995 49.117
785549 101.376 -52.004 49.121
785550 101.395 -52.014 49.129
785551 101.414 -52.023 49.138
785552 101.433 -52.032 49.148
785553 101.452 -52.041 49.158
785554 101.471 -52.051 49.168
785555 101.490 -52.060 49.179
785556 101.509 -52.069 49.189
785557 101.530 -52.083 49.197
785558 101.551 -52.098 49.203
785559 101.571 -52.117 49.209
785560 101.586 -52.130 49.210
785561 101.595 -52.145 49.195
785562 101.603 -52.155 49.168
785563 101.610 -52.165 49.138
785564 101.629 -52.179 49.107
785565 101.676 -52.161 49.073
785566 101.698 -52.176 49.039
785567 101.727 -52.192 49.004
785568 101.755 -52.208 48.967
785569 101.783 -52.224 48.930
785570 101.812 -52.241 48.893
785571 101.840 -52.257 48.856
785572 101.869 -52.274 48.819
785573 101.898 -52.290 48.783
785574 101.926 -52.306 48.749
785575 101.955 -52.323 48.714
785576 101.983 -52.339 48.680
786247 101.335 -51.960 49.136
786248 101.354 -51.969 49.132
786249 101.373 -51.978 49.132
786250 101.392 -51.987 49.134
786251 101.411 -51.997 49.142
786252 101.430 -52.006 49.150
786253 101.449 -52.015 49.159
786254 101.468 -52.024 49.169
786255 101.487 -52.033 49.179
786256 101.506 -52.042 49.190
786257 101.526 -52.054 49.200
786258 101.547 -52.069 49.208
786259 101.568 -52.084 49.215
786260 101.588 -52.104 49.221
786261 101.601 -52.111 49.223
786262 101.610 -52.127 49.211
786263 101.619 -52.141 49.184
786264 101.627 -52.152 49.154
786265 101.637 -52.163 49.123
786266 101.690 -52.145 49.091
786267 101.706 -52.156 49.057
786268 101.735 -52.173 49.022
786269 101.763 -52.189 48.985
786270 101.792 -52.205 48.948
786271 101.821 -52.222 48.911
786272 101.849 -52.238 48.873
786273 101.878 -52.255 48.836
786274 101.906 -52.271 48.801
786275 101.935 -52.287 48.766
786276 101.963 -52.304 48.731
786277 101.992 -52.320 48.696
786948 101.351 -51.943 49.153
786949 101.370 -51.952 49.148
786950 101.389 -51.961 49.147
786951 101.408 -51.970 49.148
786952 101.427 -51.979 49.154
786953 101.446 -51.989 49.162
786954 101.465 -51.998 49.170
786955 101.484 -52.007 49.180
786956 101.503 -52.017 49.190
786957 101.522 -52.026 49.200
786958 101.543 -52.039 49.210
786959 101.564 -52.054 49.219
786960 101.585 -52.071 49.226
786961 101.602 -52.086 49.231
786962 101.616 -52.092 49.233
786963 101.625 -52.108 49.223
786964 101.634 -52.123 49.196
786965 101.643 -52.137 49.168
786966 101.651 -52.149 49.139
786967 101.687 -52.143 49.107
786968 101.715 -52.137 49.074
786969 101.744 -52.154 49.039
786970 101.772 -52.170 49.002
786971 101.801 -52.186 48.965
786972 101.829 -52.202 48.928
786973 101.858 -52.219 48.890
786974 101.886 -52.236 48.854
786975 101.915 -52.252 48.818
786976 101.944 -52.268 48.783
786977 101.972 -52.284 48.748
786978 102.000 -52.301 48.712
787649 101.367 -51.926 49.171
787650 101.386 -51.935 49.165
787651 101.405 -51.944 49.163
787652 101.424 -51.954 49.163
787653 101.443 -51.963 49.167
787654 101.462 -51.972 49.175
787655 101.481 -51.981 49.182
787656 101.499 -51.990 49.191
787657 101.519 -51.999 49.200
787658 101.538 -52.011 49.210
787659 101.559 -52.025 49.219
787660 101.581 -52.040 49.227
787661 101.601 -52.058 49.235
787662 101.617 -52.068 49.240
787663 101.631 -52.074 49.244
787664 101.640 -52.089 49.231
787665 101.650 -52.104 49.208
787666 101.658 -52.119 49.182
787667 101.667 -52.133 49.153
787668 101.682 -52.142 49.122
787669 101.729 -52.120 49.089
787670 101.752 -52.135 49.054
787671 101.781 -52.151 49.019
787672 101.809 -52.167 48.982
787673 101.838 -52.183 48.945
787674 101.867 -52.200 48.907
787675 101.895 -52.216 48.871
787676 101.923 -52.233 48.836
787677 101.952 -52.249 48.800
787678 101.981 -52.265 48.764
787679 102.009 -52.282 48.728
788351 101.402 -51.918 49.183
788352 101.421 -51.927 49.179
788353 101.440 -51.936 49.178
788354 101.459 -51.945 49.181
788355 101.478 -51.954 49.188
788356 101.496 -51.964 49.194
788357 101.516 -51.973 49.202
788358 101.534 -51.982 49.211
788359 101.555 -51.996 49.219
788360 101.576 -52.011 49.227
788361 101.598 -52.026 49.235
788362 101.617 -52.043 49.244
788363 101.631 -52.050 49.249
788364 101.646 -52.056 49.253
788365 101.656 -52.070 49.241
788366 101.665 -52.085 49.221
788367 101.673 -52.100 49.197
788368 101.682 -52.115 49.168
788369 101.693 -52.127 49.136
788370 101.743 -52.103 49.103
788371 101.761 -52.115 49.070
788372 101.790 -52.132 49.034
788373 101.818 -52.148 48.998
788374 101.846 -52.164 48.961
788375 101.875 -52.181 48.924
788376 101.904 -52.197 48.888
788377 101.932 -52.214 48.853
788378 101.961 -52.230 48.817
788379 101.989 -52.246 48.781
788380 102.018 -52.262 48.745
789052 101.417 -51.901 49.201
789053 101.437 -51.910 49.196
789054 101.456 -51.920 49.195
789055 101.474 -51.929 49.197
789056 101.493 -51.938 49.201
789057 101.513 -51.947 49.206
789058 101.531 -51.956 49.213
789059 101.551 -51.967 49.221
789060 101.572 -51.982 49.228
789061 101.593 -51.997 49.235
789062 101.614 -52.011 49.243
789063 101.632 -52.026 49.252
789064 101.646 -52.032 49.258
789065 101.661 -52.039 49.261
789066 101.671 -52.051 49.251
789067 101.680 -52.067 49.235
789068 101.688 -52.081 49.212
789069 101.699 -52.095 49.183
789070 101.714 -52.104 49.150
789071 101.756 -52.088 49.117
789072 101.769 -52.096 49.085
789073 101.798 -52.113 49.049
789074 101.827 -52.129 49.013
789075 101.855 -52.145 48.977
789076 101.884 -52.161 48.940
789077 101.912 -52.178 48.905
789078 101.941 -52.195 48.870
789079 101.969 -52.211 48.834
789080 101.998 -52.227 48.798
789081 102.027 -52.243 48.761
789753 101.433 -51.884 49.220
789754 101.452 -51.893 49.214
789755 101.471 -51.902 49.211
789756 101.490 -51.911 49.212
789757 101.509 -51.921 49.216
789758 101.528 -51.930 49.220
789759 101.547 -51.939 49.225
789760 101.568 -51.953 49.230
789761 101.589 -51.967 49.237
789762 101.610 -51.982 49.243
789763 101.631 -51.998 49.250
789764 101.646 -52.007 49.257
789765 101.661 -52.014 49.263
789766 101.676 -52.020 49.268
789767 101.686 -52.032 49.261
789768 101.695 -52.048 49.250
789769 101.705 -52.061 49.227
789770 101.720 -52.070 49.196
789771 101.735 -52.080 49.162
789772 101.752 -52.087 49.130
789773 101.782 -52.079 49.098
789774 101.807 -52.093 49.063
789775 101.835 -52.110 49.030
789776 101.864 -52.127 48.994
789777 101.892 -52.142 48.957
789778 101.921 -52.159 48.921
789779 101.950 -52.175 48.886
789780 101.978 -52.192 48.851
789781 102.007 -52.208 48.814
789782 102.035 -52.224 48.778
789783 102.063 -52.241 48.741
790420 100.841 -51.537 50.618
790421 100.859 -51.544 50.572
790422 100.878 -51.550 50.526
790423 100.896 -51.561 50.479
790424 100.914 -51.571 50.432
790425 100.932 -51.581 50.384
790426 100.951 -51.591 50.336
790453 101.430 -51.857 49.248
790454 101.449 -51.867 49.239
790455 101.468 -51.876 49.232
790456 101.487 -51.885 49.228
790457 101.506 -51.895 49.228
790458 101.525 -51.904 49.230
790459 101.544 -51.913 49.233
790460 101.563 -51.923 49.236
790461 101.585 -51.939 49.241
790462 101.606 -51.953 49.246
790463 101.627 -51.968 49.252
790464 101.646 -51.983 49.257
790465 101.661 -51.989 49.263
790466 101.676 -51.996 49.268
790467 101.690 -52.002 49.274
790468 101.701 -52.014 49.272
790469 101.710 -52.029 49.262
790470 101.726 -52.038 49.240
790471 101.740 -52.047 49.207
790472 101.755 -52.056 49.173
790473 101.770 -52.065 49.142
790474 101.796 -52.062 49.110
790475 101.815 -52.074 49.077
790476 101.844 -52.091 49.044
790477 101.873 -52.107 49.009
790478 101.901 -52.123 48.973
790479 101.929 -52.140 48.937
790480 101.958 -52.156 48.903
790481 101.986 -52.173 48.867
790482 102.015 -52.189 48.831
790483 102.044 -52.205 48.794
790484 102.073 -52.221 48.758
791119 100.813 -51.516 50.735
791120 100.834 -51.519 50.690
791121 100.855 -51.522 50.645
791122 100.874 -51.528 50.600
791123 100.893 -51.534 50.554
791124 100.911 -51.544 50.507
791125 100.929 -51.553 50.460
791126 100.948 -51.564 50.413
791127 100.966 -51.574 50.364
791128 100.984 -51.584 50.316
791129 101.002 -51.594 50.266
791154 101.446 -51.841 49.268
791155 101.465 -51.850 49.258
791156 101.484 -51.859 49.250
791157 101.503 -51.868 49.244
791158 101.522 -51.878 49.243
791159 101.541 -51.887 49.243
791160 101.560 -51.896 49.245
791161 101.581 -51.909 49.247
791162 101.602 -51.924 49.251
791163 101.623 -51.939 49.256
791164 101.644 -51.954 49.260
791165 101.661 -51.965 49.264
791166 101.676 -51.972 49.269
791167 101.690 -51.978 49.273
791168 101.705 -51.985 49.282
791169 101.716 -51.995 49.283
791170 101.731 -52.004 49.273
791171 101.746 -52.014 49.250
791172 101.761 -52.023 49.217
791173 101.776 -52.032 49.184
791174 101.791 -52.041 49.152
791175 101.811 -52.045 49.121
791176 101.824 -52.055 49.090
791177 101.852 -52.072 49.058
791178 101.881 -52.088 49.023
791179 101.910 -52.105 48.986
791180 101.938 -52.120 48.953
791181 101.967 -52.137 48.919
791182 101.995 -52.153 48.883
791183 102.024 -52.170 48.848
791184 102.052 -52.186 48.811
791185 102.081 -52.202 48.773
791819 100.805 -51.501 50.806
791820 100.826 -51.503 50.762
791821 100.847 -51.506 50.717
791822 100.868 -51.509 50.673
791823 100.889 -51.513 50.629
791824 100.907 -51.520 50.583
791825 100.926 -51.526 50.537
791826 100.945 -51.537 50.489
791827 100.963 -51.547 50.441
791828 100.981 -51.556 50.392
791829 100.999 -51.567 50.344
791830 101.018 -51.577 50.295
791831 101.036 -51.587 50.244
791832 101.054 -51.597 50.194
791855 101.462 -51.824 49.287
791856 101.481 -51.833 49.276
791857 101.500 -51.842 49.267
791858 101.519 -51.851 49.260
791859 101.538 -51.860 49.257
791860 101.557 -51.870 49.257
791861 101.576 -51.880 49.257
791862 101.597 -51.895 49.258
791863 101.618 -51.910 49.261
791864 101.639 -51.925 49.264
791865 101.661 -51.939 49.268
791866 101.676 -51.948 49.271
791867 101.690 -51.954 49.274
791868 101.704 -51.961 49.278
791869 101.717 -51.969 49.286
791870 101.732 -51.976 49.289
791871 101.747 -51.986 49.277
791872 101.762 -51.994 49.254
791873 101.777 -52.003 49.224
791874 101.793 -52.011 49.192
791875 101.808 -52.020 49.162
791876 101.824 -52.029 49.132
791877 101.835 -52.037 49.101
791878 101.861 -52.052 49.070
791879 101.890 -52.069 49.037
791880 101.918 -52.086 49.002
791881 101.947 -52.101 48.968
791882 101.975 -52.118 48.934
791883 102.004 -52.134 48.899
791884 102.033 -52.151 48.863
791885 102.061 -52.167 48.826
791886 102.090 -52.183 48.788
792519 100.796 -51.485 50.877
792520 100.818 -51.488 50.833
792521 100.839 -51.491 50.790
792522 100.860 -51.494 50.746
792523 100.882 -51.497 50.702
792524 100.903 -51.500 50.658
792525 100.922 -51.505 50.613
792526 100.941 -51.511 50.566
792527 100.959 -51.519 50.519
792528 100.978 -51.529 50.471
792529 100.996 -51.540 50.422
792530 101.014 -51.550 50.373
792531 101.033 -51.559 50.322
792532 101.051 -51.570 50.272
792533 101.069 -51.580 50.221
792534 101.088 -51.590 50.171
792556 101.478 -51.807 49.306
792557 101.496 -51.816 49.294
792558 101.516 -51.825 49.285
792559 101.535 -51.835 49.277
792560 101.553 -51.844 49.273
792561 101.572 -51.853 49.271
792562 101.593 -51.866 49.270
792563 101.614 -51.881 49.270
792564 101.635 -51.895 49.271
792565 101.656 -51.910 49.273
792566 101.673 -51.923 49.276
792567 101.686 -51.934 49.278
792568 101.700 -51.943 49.280
792569 101.713 -51.951 49.283
792570 101.727 -51.958 49.290
792571 101.741 -51.966 49.293
792572 101.755 -51.973 49.281
792573 101.770 -51.982 49.260
792574 101.785 -51.991 49.232
792575 101.801 -52.000 49.201
792576 101.816 -52.008 49.171
792577 101.834 -52.014 49.142
792578 101.849 -52.020 49.112
792579 101.870 -52.033 49.082
792580 101.898 -52.050 49.049
792581 101.927 -52.066 49.016
792582 101.956 -52.083 48.982
792583 101.984 -52.098 48.948
792584 102.013 -52.115 48.914
792585 102.041 -52.132 48.878
792586 102.069 -52.148 48.841
792587 102.098 -52.164 48.803
792588 102.127 -52.180 48.767
793219 100.788 -51.470 50.947
793220 100.810 -51.473 50.904
793221 100.831 -51.475 50.861
793222 100.852 -51.478 50.819
793223 100.873 -51.481 50.774
793224 100.894 -51.484 50.731
793225 100.916 -51.487 50.687
793226 100.937 -51.490 50.642
793227 100.956 -51.497 50.596
793228 100.974 -51.503 50.549
793229 100.992 -51.512 50.501
793230 101.011 -51.522 50.452
793231 101.029 -51.532 50.402
793232 101.047 -51.543 50.351
793233 101.066 -51.553 50.301
793234 101.084 -51.562 50.250
793235 101.102 -51.572 50.198
793257 101.493 -51.790 49.326
793258 101.513 -51.799 49.313
793259 101.531 -51.808 49.303
793260 101.550 -51.817 49.294
793261 101.569 -51.826 49.289
793262 101.589 -51.837 49.285
793263 101.610 -51.851 49.283
793264 101.631 -51.866 49.283
793265 101.652 -51.881 49.282
793266 101.671 -51.895 49.283
793267 101.683 -51.905 49.283
793268 101.696 -51.916 49.284
793269 101.709 -51.927 49.285
793270 101.723 -51.938 49.287
793271 101.737 -51.948 49.292
793272 101.750 -51.955 49.297
793273 101.764 -51.963 49.285
793274 101.778 -51.970 49.266
793275 101.793 -51.979 49.240
793276 101.809 -51.988 49.210
793277 101.824 -51.997 49.180
793278 101.844 -52.000 49.152
793279 101.864 -52.004 49.123
793280 101.879 -52.014 49.093
793281 101.907 -52.031 49.060
793282 101.936 -52.047 49.028
793283 101.964 -52.064 48.995
793284 101.992 -52.080 48.961
793285 102.021 -52.096 48.927
793286 102.050 -52.112 48.892
793287 102.079 -52.129 48.855
793288 102.107 -52.145 48.818
793289 102.135 -52.161 48.782
793290 102.164 -52.178 48.744
793920 100.802 -51.457 50.975
793921 100.823 -51.460 50.932
793922 100.844 -51.463 50.890
793923 100.865 -51.466 50.847
793924 100.886 -51.469 50.804
793925 100.907 -51.472 50.761
793926 100.928 -51.475 50.717
793927 100.950 -51.478 50.673
793928 100.971 -51.481 50.628
793929 100.989 -51.487 50.580
793930 101.007 -51.495 50.531
793931 101.025 -51.505 50.482
793932 101.044 -51.515 50.432
793933 101.062 -51.525 50.380
793934 101.080 -51.535 50.331
793935 101.099 -51.545 50.280
793936 101.117 -51.556 50.227
793937 101.135 -51.566 50.176
793958 101.510 -51.772 49.346
793959 101.528 -51.782 49.332
793960 101.547 -51.791 49.321
793961 101.566 -51.801 49.311
793962 101.585 -51.810 49.306
793963 101.606 -51.823 49.300
793964 101.627 -51.837 49.298
793965 101.648 -51.852 49.295
793966 101.669 -51.866 49.294
793967 101.681 -51.877 49.293
793968 101.693 -51.888 49.292
793969 101.705 -51.898 49.292
793970 101.718 -51.909 49.292
793971 101.733 -51.920 49.292
793972 101.746 -51.931 49.294
793973 101.760 -51.942 49.298
793974 101.773 -51.951 49.289
793975 101.787 -51.959 49.271
793976 101.802 -51.967 49.245
793977 101.817 -51.976 49.216
793978 101.834 -51.983 49.187
793979 101.854 -51.986 49.160
793980 101.875 -51.989 49.132
793981 101.889 -51.995 49.103
793982 101.916 -52.011 49.072
793983 101.944 -52.028 49.039
793984 101.973 -52.044 49.007
793985 102.001 -52.061 48.976
793986 102.030 -52.077 48.941
793987 102.058 -52.093 48.906
793988 102.087 -52.110 48.870
793989 102.115 -52.126 48.833
793990 102.144 -52.142 48.797
793991 102.173 -52.158 48.758
793992 102.201 -52.175 48.719
794620 100.793 -51.442 51.043
794621 100.815 -51.445 51.002
794622 100.836 -51.447 50.960
794623 100.857 -51.450 50.918
794624 100.878 -51.453 50.876
794625 100.899 -51.456 50.833
794626 100.920 -51.459 50.791
794627 100.941 -51.462 50.747
794628 100.963 -51.465 50.704
794629 100.984 -51.468 50.659
794630 101.004 -51.473 50.612
794631 101.022 -51.479 50.563
794632 101.041 -51.487 50.515
794633 101.059 -51.498 50.464
794634 101.077 -51.508 50.412
794635 101.096 -51.518 50.361
794636 101.114 -51.528 50.309
794637 101.132 -51.538 50.256
794638 101.150 -51.548 50.204
794639 101.169 -51.559 50.153
794658 101.506 -51.747 49.382
794659 101.525 -51.756 49.365
794660 101.544 -51.765 49.351
794661 101.563 -51.774 49.339
794662 101.582 -51.783 49.329
794663 101.601 -51.793 49.322
794664 101.622 -51.808 49.316
794665 101.643 -51.823 49.312
794666 101.665 -51.838 49.308
794667 101.679 -51.849 49.305
794668 101.691 -51.860 49.303
794669 101.703 -51.870 49.302
794670 101.715 -51.880 49.300
794671 101.728 -51.891 49.299
794672 101.742 -51.901 49.297
794673 101.755 -51.913 49.297
794674 101.769 -51.924 49.301
794675 101.783 -51.935 49.294
794676 101.796 -51.947 49.277
794677 101.810 -51.956 49.251
794678 101.825 -51.964 49.223
794679 101.844 -51.969 49.195
794680 101.864 -51.972 49.168
794681 101.885 -51.975 49.141
794682 101.903 -51.979 49.112
794683 101.924 -51.992 49.083
794684 101.953 -52.009 49.051
794685 101.981 -52.025 49.020
794686 102.010 -52.042 48.990
794687 102.038 -52.058 48.956
794688 102.067 -52.074 48.920
794689 102.096 -52.091 48.885
794690 102.124 -52.107 48.849
794691 102.153 -52.123 48.812
794692 102.181 -52.139 48.773
794693 102.210 -52.156 48.734
794694 102.238 -52.172 48.692
795321 100.806 -51.429 51.069
795322 100.828 -51.432 51.028
795323 100.849 -51.435 50.987
795324 100.870 -51.438 50.946
795325 100.891 -51.440 50.904
795326 100.912 -51.443 50.862
795327 100.933 -51.446 50.821
795328 100.955 -51.450 50.778
795329 100.976 -51.453 50.735
795330 100.997 -51.456 50.691
795331 101.018 -51.459 50.644
795332 101.037 -51.464 50.596
795333 101.056 -51.471 50.547
795334 101.074 -51.481 50.496
795335 101.092 -51.490 50.444
795336 101.111 -51.501 50.391
795337 101.129 -51.511 50.338
795338 101.147 -51.521 50.285
795339 101.165 -51.531 50.232
795358 101.503 -51.720 49.421
795359 101.522 -51.729 49.401
795360 101.541 -51.738 49.384
795361 101.560 -51.748 49.370
795362 101.579 -51.757 49.357
795363 101.598 -51.766 49.346
795364 101.618 -51.779 49.338
795365 101.639 -51.794 49.331
795366 101.660 -51.808 49.326
795367 101.677 -51.821 49.322
795368 101.689 -51.832 49.317
795369 101.701 -51.841 49.314
795370 101.713 -51.852 49.311
795371 101.726 -51.862 49.309
795372 101.738 -51.873 49.306
795373 101.751 -51.884 49.303
795374 101.765 -51.895 49.302
795375 101.778 -51.906 49.304
795376 101.793 -51.917 49.297
795377 101.806 -51.929 49.282
795378 101.820 -51.940 49.257
795379 101.834 -51.951 49.229
795380 101.854 -51.955 49.202
795381 101.874 -51.957 49.176
795382 101.895 -51.960 49.150
795383 101.915 -51.963 49.122
795384 101.933 -51.973 49.093
795385 101.962 -51.989 49.064
795386 101.990 -52.006 49.034
795387 102.019 -52.023 49.003
795388 102.047 -52.039 48.971
795389 102.076 -52.055 48.936
795390 102.104 -52.071 48.900
795391 102.133 -52.088 48.864
795392 102.161 -52.104 48.827
795393 102.190 -52.120 48.788
795394 102.219 -52.137 48.749
795395 102.247 -52.153 48.706
795396 102.275 -52.170 48.664
796021 100.798 -51.412 51.136
796022 100.819 -51.416 51.095
796023 100.840 -51.419 51.054
796024 100.861 -51.422 51.014
796025 100.883 -51.425 50.975
796026 100.904 -51.428 50.933
796027 100.925 -51.431 50.892
796028 100.947 -51.434 50.852
796029 100.968 -51.437 50.810
796030 100.989 -51.440 50.767
796031 101.010 -51.443 50.724
796032 101.031 -51.446 50.677
796033 101.052 -51.450 50.629
796034 101.070 -51.456 50.579
796035 101.089 -51.463 50.528
796036 101.107 -51.474 50.476
796037 101.125 -51.484 50.421
796038 101.144 -51.494 50.368
796039 101.162 -51.504 50.314
796040 101.180 -51.514 50.260
796041 101.199 -51.524 50.207
796057 101.488 -51.682 49.486
796058 101.502 -51.694 49.463
796059 101.519 -51.703 49.440
796060 101.538 -51.713 49.421
796061 101.557 -51.722 49.403
796062 101.575 -51.731 49.387
796063 101.595 -51.740 49.374
796064 101.614 -51.750 49.362
796065 101.635 -51.764 49.355
796066 101.656 -51.779 49.346
796067 101.676 -51.793 49.341
796068 101.687 -51.804 49.335
796069 101.699 -51.813 49.330
796070 101.711 -51.824 49.325
796071 101.724 -51.834 49.321
796072 101.736 -51.845 49.317
796073 101.748 -51.854 49.313
796074 101.760 -51.866 49.308
796075 101.774 -51.876 49.304
796076 101.788 -51.888 49.302
796077 101.802 -51.899 49.297
796078 101.815 -51.910 49.284
796079 101.829 -51.922 49.261
796080 101.846 -51.930 49.235
796081 101.865 -51.938 49.209
796082 101.884 -51.943 49.184
796083 101.905 -51.946 49.159
796084 101.925 -51.949 49.133
796085 101.942 -51.954 49.105
796086 101.970 -51.970 49.076
796087 101.998 -51.987 49.047
796088 102.027 -52.003 49.016
796089 102.056 -52.020 48.984
796090 102.084 -52.036 48.951
796091 102.113 -52.052 48.916
796092 102.141 -52.069 48.880
796093 102.170 -52.085 48.842
796094 102.198 -52.101 48.803
796095 102.227 -52.117 48.764
796096 102.256 -52.134 48.721
796097 102.284 -52.151 48.678
796098 102.313 -52.167 48.636
796722 100.819 -51.379 51.162
796723 100.834 -51.400 51.121
796724 100.853 -51.407 51.081
796725 100.875 -51.410 51.041
796726 100.896 -51.412 51.002
796727 100.917 -51.415 50.962
796728 100.938 -51.418 50.922
796729 100.960 -51.421 50.882
796730 100.981 -51.425 50.842
796731 101.002 -51.428 50.800
796732 101.023 -51.431 50.757
796733 101.044 -51.433 50.712
796734 101.065 -51.436 50.663
796735 101.086 -51.440 50.612
796736 101.104 -51.447 50.561
796737 101.122 -51.456 50.508
796738 101.140 -51.466 50.453
796739 101.159 -51.477 50.399
796740 101.177 -51.487 50.344
796741 101.195 -51.497 50.289
796757 101.489 -51.654 49.534
796758 101.503 -51.666 49.507
796759 101.518 -51.676 49.482
796760 101.535 -51.686 49.460
796761 101.554 -51.695 49.440
796762 101.572 -51.705 49.422
796763 101.592 -51.714 49.405
796764 101.611 -51.723 49.392
796765 101.631 -51.735 49.380
796766 101.652 -51.750 49.371
796767 101.673 -51.765 49.362
796768 101.686 -51.776 49.355
796769 101.698 -51.786 49.349
796770 101.709 -51.796 49.342
796771 101.722 -51.807 49.337
796772 101.734 -51.816 49.332
796773 101.746 -51.827 49.326
796774 101.758 -51.837 49.321
796775 101.770 -51.848 49.314
796776 101.783 -51.858 49.308
796777 101.797 -51.870 49.303
796778 101.811 -51.881 49.300
796779 101.825 -51.892 49.288
796780 101.840 -51.902 49.267
796781 101.859 -51.910 49.241
796782 101.877 -51.917 49.216
796783 101.896 -51.926 49.192
796784 101.915 -51.932 49.168
796785 101.936 -51.935 49.143
796786 101.956 -51.938 49.115
796787 101.979 -51.951 49.089
796788 102.007 -51.968 49.061
796789 102.036 -51.984 49.031
796790 102.064 -52.001 48.999
796791 102.093 -52.017 48.966
796792 102.121 -52.033 48.931
796793 102.150 -52.049 48.895
796794 102.179 -52.066 48.858
796795 102.207 -52.083 48.819
796796 102.236 -52.098 48.779
796797 102.263 -52.114 48.737
796798 102.290 -52.130 48.693
796799 102.318 -52.144 48.650
796800 102.344 -52.159 48.607
797423 100.838 -51.346 51.187
797424 100.853 -51.368 51.146
797425 100.868 -51.389 51.107
797426 100.888 -51.397 51.069
797427 100.909 -51.400 51.030
797428 100.930 -51.403 50.991
797429 100.951 -51.406 50.952
797430 100.972 -51.409 50.913
797431 100.994 -51.412 50.873
797432 101.015 -51.415 50.833
797433 101.036 -51.418 50.793
797434 101.057 -51.421 50.748
797435 101.078 -51.424 50.697
797436 101.100 -51.426 50.645
797437 101.119 -51.432 50.593
797438 101.137 -51.439 50.539
797439 101.156 -51.450 50.484
797440 101.174 -51.459 50.429
797441 101.192 -51.469 50.373
797442 101.210 -51.480 50.318
797456 101.476 -51.616 49.617
797457 101.491 -51.626 49.585
797458 101.505 -51.638 49.555
797459 101.519 -51.648 49.527
797460 101.534 -51.660 49.502
797461 101.550 -51.669 49.480
797462 101.569 -51.678 49.459
797463 101.589 -51.688 49.441
797464 101.607 -51.697 49.424
797465 101.626 -51.707 49.410
797466 101.647 -51.721 49.398
797467 101.669 -51.736 49.388
797468 101.684 -51.747 49.379
797469 101.696 -51.758 49.370
797470 101.708 -51.768 49.363
797471 101.720 -51.779 49.355
797472 101.732 -51.788 49.348
797473 101.744 -51.799 49.342
797474 101.756 -51.809 49.336
797475 101.768 -51.819 49.329
797476 101.780 -51.829 49.321
797477 101.793 -51.840 49.313
797478 101.807 -51.851 49.307
797479 101.820 -51.863 49.304
797480 101.834 -51.874 49.294
797481 101.853 -51.882 49.273
797482 101.871 -51.889 49.248
797483 101.890 -51.898 49.223
797484 101.908 -51.905 49.200
797485 101.927 -51.913 49.177
797486 101.946 -51.920 49.152
797487 101.966 -51.923 49.126
797488 101.987 -51.932 49.104
797489 102.016 -51.948 49.076
797490 102.044 -51.965 49.046
797491 102.073 -51.982 49.015
797492 102.102 -51.998 48.982
797493 102.130 -52.014 48.947
797494 102.159 -52.030 48.911
797495 102.186 -52.045 48.875
797496 102.213 -52.061 48.835
797497 102.240 -52.076 48.795
797498 102.267 -52.090 48.751
797499 102.294 -52.105 48.708
797500 102.321 -52.120 48.665
797501 102.348 -52.135 48.621
797502 102.375 -52.150 48.578
798124 100.858 -51.313 51.211
798125 100.873 -51.334 51.172
798126 100.888 -51.356 51.134
798127 100.903 -51.377 51.095
798128 100.922 -51.387 51.057
798129 100.943 -51.390 51.019
798130 100.964 -51.393 50.980
798131 100.985 -51.396 50.942
798132 101.006 -51.399 50.904
798133 101.028 -51.402 50.868
798134 101.049 -51.405 50.830
798135 101.070 -51.408 50.785
798136 101.092 -51.411 50.733
798137 101.113 -51.414 50.678
798138 101.133 -51.417 50.624
798139 101.152 -51.424 50.571
798140 101.171 -51.432 50.515
798141 101.189 -51.442 50.459
798142 101.207 -51.453 50.402
798143 101.226 -51.462 50.346
798144 101.244 -51.472 50.292
798155 101.445 -51.584 49.709
798156 101.466 -51.592 49.672
798157 101.488 -51.600 49.637
798158 101.507 -51.609 49.605
798159 101.522 -51.620 49.575
798160 101.536 -51.631 49.547
798161 101.550 -51.642 49.521
798162 101.566 -51.652 49.499
798163 101.585 -51.662 49.479
798164 101.604 -51.671 49.461
798165 101.623 -51.680 49.444
798166 101.643 -51.692 49.429
798167 101.664 -51.707 49.416
798168 101.682 -51.719 49.405
798169 101.694 -51.730 49.395
798170 101.706 -51.740 49.386
798171 101.718 -51.750 49.378
798172 101.730 -51.760 49.369
798173 101.742 -51.771 49.361
798174 101.754 -51.781 49.353
798175 101.766 -51.791 49.345
798176 101.778 -51.801 49.337
798177 101.790 -51.812 49.328
798178 101.802 -51.822 49.318
798179 101.816 -51.833 49.313
798180 101.830 -51.845 49.308
798181 101.846 -51.854 49.299
798182 101.865 -51.861 49.281
798183 101.884 -51.869 49.257
798184 101.902 -51.877 49.232
798185 101.921 -51.885 49.209
798186 101.940 -51.892 49.185
798187 101.958 -51.900 49.162
798188 101.976 -51.908 49.140
798189 101.996 -51.913 49.121
798190 102.025 -51.929 49.093
798191 102.053 -51.946 49.061
798192 102.082 -51.962 49.029
798193 102.108 -51.977 48.996
798194 102.135 -51.992 48.962
798195 102.163 -52.007 48.926
798196 102.189 -52.022 48.889
798197 102.217 -52.036 48.850
798198 102.244 -52.051 48.809
798199 102.270 -52.067 48.767
798200 102.298 -52.081 48.722
798201 102.325 -52.096 48.679
798202 102.351 -52.111 48.636
798203 102.379 -52.126 48.592
798825 100.878 -51.280 51.235
798826 100.893 -51.301 51.195
798827 100.908 -51.322 51.157
798828 100.923 -51.344 51.119
798829 100.938 -51.365 51.082
798830 100.956 -51.378 51.044
798831 100.977 -51.381 51.006
798832 100.998 -51.384 50.969
798833 101.019 -51.387 50.933
798834 101.041 -51.390 50.901
798835 101.062 -51.393 50.864
798836 101.083 -51.396 50.818
798837 101.105 -51.398 50.765
798838 101.126 -51.401 50.709
798839 101.147 -51.404 50.655
798840 101.167 -51.409 50.600
798841 101.185 -51.415 50.544
798842 101.204 -51.425 50.488
798843 101.222 -51.435 50.430
798844 101.240 -51.445 50.374
798845 101.259 -51.455 50.318
798855 101.442 -51.556 49.769
798856 101.460 -51.566 49.729
798857 101.478 -51.576 49.692
798858 101.500 -51.584 49.658
798859 101.522 -51.593 49.625
798860 101.537 -51.603 49.594
798861 101.551 -51.614 49.567
798862 101.565 -51.625 49.541
798863 101.582 -51.635 49.518
798864 101.601 -51.644 49.498
798865 101.620 -51.653 49.480
798866 101.639 -51.663 49.463
798867 101.660 -51.678 49.448
798868 101.680 -51.691 49.435
798869 101.692 -51.702 49.423
798870 101.704 -51.712 49.412
798871 101.716 -51.722 49.402
798872 101.728 -51.732 49.393
798873 101.740 -51.743 49.383
798874 101.752 -51.753 49.374
798875 101.764 -51.763 49.365
798876 101.776 -51.773 49.355
798877 101.788 -51.784 49.345
798878 101.800 -51.794 49.336
798879 101.812 -51.804 49.325
798880 101.825 -51.815 49.319
798881 101.840 -51.826 49.313
798882 101.859 -51.833 49.304
798883 101.878 -51.841 49.288
798884 101.896 -51.849 49.265
798885 101.915 -51.857 49.241
798886 101.933 -51.864 49.217
798887 101.952 -51.872 49.193
798888 101.970 -51.880 49.170
798889 101.988 -51.888 49.150
798890 102.006 -51.899 49.132
798891 102.031 -51.910 49.107
798892 102.058 -51.925 49.075
798893 102.085 -51.939 49.043
798894 102.112 -51.954 49.011
798895 102.139 -51.968 48.977
798896 102.166 -51.982 48.941
798897 102.193 -51.998 48.904
798898 102.220 -52.013 48.865
798899 102.247 -52.027 48.823
798900 102.274 -52.042 48.783
798901 102.301 -52.058 48.737
798902 102.328 -52.072 48.694
798903 102.355 -52.087 48.650
798904 102.382 -52.102 48.606
799526 100.898 -51.247 51.257
799527 100.913 -51.268 51.217
799528 100.928 -51.290 51.179
799529 100.943 -51.311 51.142
799530 100.958 -51.332 51.105
799531 100.973 -51.353 51.067
799532 100.990 -51.368 51.031
799533 101.011 -51.371 50.994
799534 101.033 -51.374 50.960
799535 101.054 -51.377 50.929
799536 101.075 -51.380 50.892
799537 101.096 -51.383 50.845
799538 101.117 -51.386 50.793
799539 101.138 -51.389 50.738
799540 101.160 -51.392 50.684
799541 101.181 -51.394 50.628
799542 101.200 -51.400 50.573
799543 101.218 -51.408 50.516
799544 101.237 -51.418 50.458
799545 101.255 -51.428 50.401
799546 101.273 -51.438 50.345
799555 101.438 -51.529 49.831
799556 101.457 -51.539 49.789
799557 101.475 -51.549 49.749
799558 101.493 -51.559 49.712
799559 101.512 -51.569 49.678
799560 101.534 -51.577 49.645
799561 101.553 -51.586 49.614
799562 101.567 -51.597 49.587
799563 101.582 -51.608 49.561
799564 101.598 -51.619 49.538
799565 101.617 -51.628 49.518
799566 101.636 -51.637 49.500
799567 101.656 -51.648 49.483
799568 101.677 -51.663 49.468
799569 101.690 -51.674 49.454
799570 101.702 -51.685 49.442
799571 101.714 -51.694 49.430
799572 101.726 -51.705 49.419
799573 101.738 -51.715 49.408
799574 101.750 -51.725 49.397
799575 101.762 -51.735 49.387
799576 101.774 -51.745 49.376
799577 101.786 -51.756 49.365
799578 101.798 -51.766 49.354
799579 101.810 -51.776 49.344
799580 101.822 -51.786 49.333
799581 101.835 -51.797 49.326
799582 101.853 -51.805 49.319
799583 101.872 -51.813 49.309
799584 101.889 -51.823 49.294
799585 101.906 -51.834 49.273
799586 101.924 -51.844 49.250
799587 101.941 -51.854 49.226
799588 101.959 -51.865 49.201
799589 101.976 -51.875 49.178
799590 101.993 -51.885 49.157
799591 102.013 -51.885 49.139
799592 102.034 -51.896 49.118
799593 102.061 -51.910 49.087
799594 102.088 -51.925 49.056
799595 102.115 -51.939 49.025
799596 102.142 -51.954 48.992
799597 102.169 -51.969 48.956
799598 102.196 -51.983 48.918
799599 102.223 -51.998 48.879
799600 102.250 -52.012 48.838
799601 102.277 -52.026 48.797
799602 102.305 -52.042 48.751
799603 102.331 -52.056 48.709
799604 102.359 -52.070 48.665
799605 102.386 -52.085 48.620
799606 102.412 -52.099 48.575
800227 100.918 -51.214 51.278
800228 100.933 -51.235 51.239
800229 100.948 -51.256 51.200
800230 100.963 -51.277 51.163
800231 100.978 -51.299 51.126
800232 100.993 -51.321 51.089
800233 101.008 -51.342 51.053
800234 101.025 -51.359 51.017
800235 101.046 -51.362 50.983
800236 101.067 -51.365 50.950
800237 101.088 -51.367 50.912
800238 101.109 -51.370 50.866
800239 101.130 -51.373 50.817
800240 101.151 -51.376 50.764
800241 101.173 -51.379 50.710
800242 101.194 -51.382 50.655
800243 101.215 -51.385 50.600
800244 101.234 -51.392 50.543
800245 101.252 -51.400 50.485
800246 101.270 -51.411 50.428
800247 101.288 -51.421 50.371
800255 101.435 -51.502 49.896
800256 101.453 -51.512 49.852
800257 101.471 -51.522 49.811
800258 101.490 -51.532 49.771
800259 101.508 -51.542 49.733
800260 101.526 -51.552 49.698
800261 101.546 -51.562 49.666
800262 101.568 -51.569 49.636
800263 101.584 -51.580 49.607
800264 101.598 -51.591 49.582
800265 101.614 -51.601 49.560
800266 101.633 -51.610 49.540
800267 101.652 -51.619 49.521
800268 101.673 -51.634 49.504
800269 101.688 -51.646 49.489
800270 101.700 -51.656 49.475
800271 101.712 -51.666 49.462
800272 101.724 -51.677 49.449
800273 101.736 -51.687 49.436
800274 101.748 -51.697 49.424
800275 101.760 -51.707 49.411
800276 101.772 -51.718 49.400
800277 101.784 -51.728 49.388
800278 101.796 -51.738 49.377
800279 101.808 -51.748 49.365
800280 101.820 -51.760 49.352
800281 101.831 -51.774 49.342
800282 101.844 -51.788 49.335
800283 101.859 -51.800 49.325
800284 101.877 -51.810 49.314
800285 101.894 -51.820 49.300
800286 101.911 -51.831 49.282
800287 101.929 -51.841 49.260
800288 101.946 -51.851 49.235
800289 101.964 -51.862 49.209
800290 101.981 -51.872 49.185
800291 101.998 -51.882 49.163
800292 102.022 -51.870 49.144
800293 102.037 -51.882 49.123
800294 102.064 -51.896 49.096
800295 102.092 -51.911 49.066
800296 102.118 -51.926 49.036
800297 102.146 -51.940 49.004
800298 102.173 -51.954 48.968
800299 102.199 -51.969 48.931
800300 102.227 -51.983 48.892
800301 102.254 -51.998 48.852
800302 102.280 -52.013 48.810
800303 102.308 -52.027 48.765
800304 102.335 -52.042 48.722
800305 102.362 -52.056 48.678
800306 102.389 -52.070 48.634
800307 102.416 -52.086 48.590
800928 100.937 -51.181 51.298
800929 100.953 -51.202 51.258
800930 100.968 -51.224 51.220
800931 100.983 -51.245 51.183
800932 100.998 -51.266 51.147
800933 101.013 -51.287 51.109
800934 101.028 -51.309 51.074
800935 101.043 -51.330 51.038
800936 101.059 -51.349 51.003
800937 101.080 -51.352 50.967
800938 101.101 -51.355 50.928
800939 101.122 -51.358 50.884
800940 101.143 -51.361 50.836
800941 101.165 -51.363 50.786
800942 101.186 -51.366 50.734
800943 101.207 -51.369 50.680
800944 101.228 -51.372 50.625
800945 101.248 -51.377 50.568
800946 101.267 -51.384 50.511
800947 101.285 -51.393 50.454
800948 101.304 -51.403 50.396
800955 101.431 -51.475 49.964
800956 101.450 -51.484 49.917
800957 101.468 -51.494 49.873
800958 101.486 -51.505 49.832
800959 101.504 -51.515 49.792
800960 101.523 -51.525 49.754
800961 101.541 -51.534 49.719
800962 101.559 -51.545 49.687
800963 101.580 -51.554 49.657
800964 101.599 -51.563 49.629
800965 101.613 -51.574 49.605
800966 101.629 -51.584 49.582
800967 101.649 -51.594 49.562
800968 101.668 -51.605 49.544
800969 101.686 -51.618 49.527
800970 101.698 -51.628 49.512
800971 101.710 -51.638 49.497
800972 101.722 -51.649 49.483
800973 101.734 -51.659 49.470
800974 101.746 -51.669 49.454
800975 101.758 -51.679 49.440
800976 101.770 -51.690 49.426
800977 101.782 -51.700 49.413
800978 101.794 -51.712 49.401
800979 101.806 -51.725 49.388
800980 101.817 -51.740 49.375
800981 101.828 -51.754 49.362
800982 101.839 -51.769 49.351
800983 101.852 -51.782 49.343
800984 101.867 -51.794 49.331
800985 101.882 -51.807 49.319
800986 101.899 -51.817 49.305
800987 101.916 -51.828 49.288
800988 101.934 -51.838 49.266
800989 101.952 -51.848 49.242
800990 101.969 -51.858 49.215
800991 101.986 -51.869 49.191
800992 102.004 -51.875 49.168
800993 102.030 -51.854 49.148
800994 102.041 -51.868 49.127
800995 102.067 -51.882 49.102
800996 102.095 -51.897 49.073
800997 102.122 -51.911 49.045
800998 102.149 -51.926 49.013
800999 102.176 -51.940 48.978
801000 102.203 -51.955 48.942
801001 102.230 -51.970 48.904
801002 102.257 -51.984 48.864
801003 102.284 -51.998 48.822
801004 102.311 -52.013 48.779
801005 102.338 -52.027 48.735
801006 102.365 -52.042 48.692
801007 102.392 -52.057 48.648
801008 102.419 -52.071 48.604
801629 100.957 -51.148 51.316
801630 100.972 -51.169 51.277
801631 100.987 -51.190 51.239
801632 101.002 -51.211 51.202
801633 101.018 -51.233 51.166
801634 101.033 -51.254 51.129
801635 101.048 -51.275 51.093
801636 101.063 -51.297 51.057
801637 101.078 -51.318 51.021
801638 101.093 -51.339 50.984
801639 101.114 -51.342 50.944
801640 101.135 -51.345 50.902
801641 101.156 -51.348 50.856
801642 101.178 -51.351 50.807
801643 101.199 -51.354 50.756
801644 101.220 -51.357 50.703
801645 101.241 -51.359 50.648
801646 101.262 -51.362 50.592
801647 101.282 -51.368 50.535
801648 101.300 -51.376 50.478
801649 101.318 -51.387 50.421
801654 101.410 -51.437 50.086
801655 101.428 -51.447 50.035
801656 101.447 -51.457 49.986
801657 101.465 -51.467 49.939
801658 101.483 -51.477 49.895
801659 101.501 -51.487 49.853
801660 101.519 -51.497 49.813
801661 101.538 -51.507 49.776
801662 101.556 -51.518 49.741
801663 101.574 -51.528 49.709
801664 101.593 -51.537 49.679
801665 101.614 -51.547 49.653
801666 101.629 -51.557 49.628
801667 101.645 -51.567 49.605
801668 101.664 -51.576 49.585
801669 101.684 -51.590 49.567
801670 101.696 -51.600 49.550
801671 101.708 -51.610 49.534
801672 101.720 -51.621 49.519
801673 101.732 -51.631 49.505
801674 101.744 -51.641 49.491
801675 101.756 -51.651 49.473
801676 101.768 -51.663 49.457
801677 101.779 -51.678 49.441
801678 101.790 -51.692 49.426
801679 101.802 -51.707 49.413
801680 101.813 -51.720 49.400
801681 101.825 -51.735 49.385
801682 101.836 -51.749 49.372
801683 101.847 -51.763 49.360
801684 101.861 -51.776 49.351
801685 101.875 -51.789 49.338
801686 101.889 -51.802 49.324
801687 101.904 -51.814 49.309
801688 101.921 -51.824 49.291
801689 101.939 -51.835 49.271
801690 101.957 -51.845 49.246
801691 101.974 -51.855 49.220
801692 101.991 -51.866 49.196
801693 102.013 -51.859 49.173
801694 102.038 -51.838 49.151
801695 102.063 -51.816 49.129
801696 102.071 -51.868 49.106
801697 102.098 -51.882 49.079
801698 102.125 -51.897 49.050
801699 102.152 -51.912 49.019
801700 102.179 -51.926 48.985
801701 102.206 -51.941 48.950
801702 102.233 -51.955 48.913
801703 102.260 -51.970 48.874
801704 102.287 -51.984 48.833
801705 102.314 -51.999 48.791
801706 102.341 -52.014 48.748
801707 102.368 -52.028 48.705
801708 102.395 -52.042 48.661
802331 100.992 -51.136 51.294
802332 101.007 -51.158 51.256
802333 101.022 -51.179 51.220
802334 101.037 -51.200 51.183
802335 101.052 -51.221 51.147
802336 101.067 -51.243 51.111
802337 101.082 -51.264 51.075
802338 101.098 -51.285 51.039
802339 101.113 -51.306 51.001
802340 101.128 -51.327 50.961
802341 101.148 -51.333 50.919
802342 101.170 -51.335 50.875
802343 101.191 -51.338 50.827
802344 101.212 -51.341 50.777
802345 101.233 -51.344 50.725
802346 101.254 -51.347 50.671
802347 101.275 -51.350 50.615
802348 101.296 -51.353 50.558
802349 101.315 -51.359 50.501
802355 101.425 -51.420 50.109
802356 101.443 -51.430 50.057
802357 101.461 -51.440 50.008
802358 101.480 -51.450 49.963
802359 101.498 -51.460 49.918
802360 101.516 -51.470 49.875
802361 101.535 -51.480 49.835
802362 101.553 -51.490 49.798
802363 101.571 -51.500 49.764
802364 101.590 -51.510 49.732
802365 101.608 -51.521 49.703
802366 101.626 -51.531 49.677
802367 101.645 -51.540 49.652
802368 101.661 -51.550 49.629
802369 101.681 -51.562 49.608
802370 101.694 -51.572 49.590
802371 101.706 -51.582 49.573
802372 101.718 -51.593 49.557
802373 101.730 -51.603 49.542
802374 101.742 -51.615 49.527
802375 101.753 -51.629 49.511
802376 101.765 -51.644 49.492
802377 101.776 -51.658 49.475
802378 101.787 -51.672 49.457
802379 101.799 -51.687 49.441
802380 101.810 -51.701 49.426
802381 101.821 -51.715 49.411
802382 101.833 -51.729 49.396
802383 101.844 -51.744 49.382
802384 101.855 -51.758 49.368
802385 101.870 -51.771 49.358
802386 101.884 -51.783 49.344
802387 101.898 -51.796 49.327
802388 101.912 -51.809 49.311
802389 101.927 -51.821 49.294
802390 101.944 -51.832 49.274
802391 101.962 -51.841 49.250
802392 101.979 -51.852 49.225
802393 101.996 -51.863 49.200
802394 102.021 -51.843 49.177
802395 102.046 -51.822 49.153
802396 102.071 -51.801 49.130
802397 102.074 -51.854 49.106
802398 102.101 -51.869 49.081
802399 102.128 -51.883 49.053
802400 102.155 -51.898 49.022
802401 102.182 -51.912 48.990
802402 102.209 -51.926 48.956
802403 102.236 -51.941 48.920
802404 102.263 -51.956 48.883
802405 102.290 -51.970 48.842
802406 102.317 -51.985 48.800
802407 102.344 -51.999 48.759
802408 102.372 -52.014 48.717
802409 102.398 -52.028 48.673
803032 101.012 -51.103 51.310
803033 101.027 -51.124 51.273
803034 101.042 -51.146 51.236
803035 101.057 -51.167 51.200
803036 101.072 -51.188 51.162
803037 101.088 -51.209 51.127
803038 101.103 -51.230 51.090
803039 101.117 -51.252 51.054
803040 101.132 -51.274 51.016
803041 101.147 -51.295 50.976
803042 101.162 -51.316 50.935
803043 101.183 -51.323 50.891
803044 101.204 -51.326 50.844
803045 101.225 -51.328 50.796
803046 101.246 -51.331 50.745
803047 101.267 -51.334 50.691
803048 101.288 -51.337 50.637
803049 101.310 -51.340 50.580
803050 101.330 -51.344 50.524
803055 101.421 -51.392 50.186
803056 101.440 -51.403 50.133
803057 101.458 -51.412 50.082
803058 101.476 -51.422 50.032
803059 101.494 -51.433 49.985
803060 101.513 -51.443 49.941
803061 101.531 -51.453 49.899
803062 101.549 -51.463 49.859
803063 101.568 -51.473 49.823
803064 101.586 -51.483 49.788
803065 101.604 -51.494 49.757
803066 101.623 -51.503 49.728
803067 101.641 -51.513 49.701
803068 101.659 -51.524 49.677
803069 101.677 -51.533 49.654
803070 101.692 -51.544 49.634
803071 101.704 -51.555 49.615
803072 101.716 -51.567 49.598
803073 101.727 -51.581 49.581
803074 101.739 -51.595 49.564
803075 101.750 -51.609 49.548
803076 101.761 -51.624 49.530
803077 101.773 -51.638 49.511
803078 101.784 -51.653 49.493
803079 101.796 -51.667 49.473
803080 101.807 -51.682 49.456
803081 101.818 -51.696 49.439
803082 101.829 -51.710 49.423
803083 101.841 -51.724 49.408
803084 101.852 -51.738 49.392
803085 101.864 -51.752 49.377
803086 101.878 -51.765 49.366
803087 101.893 -51.778 49.351
803088 101.907 -51.791 49.332
803089 101.921 -51.803 49.315
803090 101.935 -51.816 49.296
803091 101.949 -51.829 49.276
803092 101.967 -51.838 49.253
803093 101.984 -51.849 49.227
803094 102.004 -51.848 49.202
803095 102.029 -51.827 49.179
803096 102.054 -51.806 49.154
803097 102.079 -51.785 49.131
803098 102.078 -51.840 49.108
803099 102.104 -51.854 49.082
803100 102.131 -51.869 49.055
803101 102.159 -51.883 49.025
803102 102.185 -51.898 48.994
803103 102.212 -51.913 48.961
803104 102.240 -51.927 48.926
803105 102.266 -51.942 48.890
803106 102.293 -51.956 48.850
803107 102.321 -51.970 48.809
803108 102.348 -51.985 48.769
803109 102.375 -52.000 48.728
803110 102.402 -52.014 48.685
803734 101.047 -51.092 51.287
803735 101.062 -51.113 51.251
803736 101.077 -51.134 51.215
803737 101.092 -51.155 51.177
803738 101.107 -51.177 51.141
803739 101.122 -51.198 51.105
803740 101.137 -51.219 51.068
803741 101.152 -51.240 51.031
803742 101.168 -51.262 50.991
803743 101.183 -51.283 50.950
803744 101.197 -51.304 50.907
803745 101.216 -51.313 50.861
803746 101.238 -51.316 50.814
803747 101.259 -51.319 50.764
803748 101.280 -51.322 50.712
803749 101.302 -51.324 50.658
803750 101.323 -51.327 50.602
803751 101.344 -51.331 50.546
803755 101.418 -51.365 50.265
803756 101.437 -51.375 50.209
803757 101.455 -51.385 50.157
803758 101.473 -51.395 50.105
803759 101.491 -51.406 50.055
803760 101.510 -51.415 50.009
803761 101.528 -51.425 49.964
803762 101.546 -51.436 49.923
803763 101.564 -51.446 49.884
803764 101.583 -51.456 49.847
803765 101.601 -51.466 49.813
803766 101.619 -51.476 49.783
803767 101.637 -51.486 49.754
803768 101.656 -51.497 49.727
803769 101.674 -51.506 49.703
803770 101.690 -51.518 49.680
803771 101.701 -51.531 49.659
803772 101.712 -51.545 49.640
803773 101.724 -51.559 49.622
803774 101.735 -51.572 49.604
803775 101.746 -51.587 49.586
803776 101.757 -51.600 49.568
803777 101.768 -51.614 49.548
803778 101.779 -51.628 49.530
803779 101.790 -51.642 49.511
803780 101.802 -51.656 49.489
803781 101.813 -51.669 49.471
803782 101.824 -51.683 49.453
803783 101.836 -51.697 49.436
803784 101.847 -51.711 49.419
803785 101.858 -51.725 49.403
803786 101.872 -51.735 49.387
803787 101.886 -51.748 49.373
803788 101.901 -51.766 49.358
803789 101.915 -51.783 49.337
803790 101.929 -51.797 49.319
803791 101.944 -51.810 49.298
803792 101.958 -51.823 49.277
803793 101.972 -51.835 49.254
803794 101.989 -51.846 49.229
803795 102.013 -51.832 49.204
803796 102.037 -51.811 49.180
803797 102.062 -51.790 49.155
803798 102.087 -51.769 49.132
803799 102.081 -51.826 49.109
803800 102.108 -51.840 49.083
803801 102.135 -51.854 49.056
803802 102.162 -51.870 49.027
803803 102.189 -51.884 48.996
803804 102.216 -51.898 48.965
803805 102.243 -51.913 48.931
803806 102.270 -51.927 48.897
803807 102.297 -51.942 48.858
803808 102.324 -51.957 48.819
803809 102.351 -51.971 48.779
803810 102.378 -51.986 48.738
804435 101.054 -51.065 51.300
804436 101.080 -51.080 51.264
804437 101.097 -51.101 51.228
804438 101.112 -51.122 51.191
804439 101.127 -51.143 51.154
804440 101.142 -51.164 51.117
804441 101.157 -51.186 51.080
804442 101.172 -51.207 51.044
804443 101.187 -51.228 51.005
804444 101.202 -51.250 50.964
804445 101.217 -51.271 50.922
804446 101.233 -51.293 50.878
804447 101.251 -51.303 50.831
804448 101.272 -51.306 50.783
804449 101.293 -51.309 50.732
804450 101.315 -51.312 50.678
804451 101.336 -51.315 50.623
804456 101.433 -51.348 50.287
804457 101.452 -51.358 50.233
804458 101.470 -51.368 50.180
804459 101.488 -51.378 50.128
804460 101.506 -51.388 50.079
804461 101.525 -51.398 50.032
804462 101.543 -51.409 49.988
804463 101.561 -51.418 49.947
804464 101.580 -51.428 49.909
804465 101.598 -51.439 49.873
804466 101.616 -51.449 49.840
804467 101.634 -51.459 49.809
804468 101.653 -51.469 49.781
804469 101.671 -51.479 49.755
804470 101.689 -51.489 49.730
804471 101.699 -51.506 49.708
804472 101.709 -51.518 49.685
804473 101.720 -51.530 49.664
804474 101.730 -51.542 49.645
804475 101.740 -51.553 49.626
804476 101.751 -51.567 49.606
804477 101.762 -51.581 49.587
804478 101.773 -51.594 49.567
804479 101.784 -51.609 49.548
804480 101.796 -51.622 49.528
804481 101.807 -51.636 49.506
804482 101.818 -51.650 49.486
804483 101.829 -51.663 49.468
804484 101.840 -51.678 49.449
804485 101.851 -51.691 49.431
804486 101.863 -51.705 49.413
804487 101.878 -51.713 49.396
804488 101.893 -51.720 49.382
804489 101.908 -51.732 49.366
804490 101.922 -51.750 49.344
804491 101.937 -51.767 49.322
804492 101.952 -51.784 49.300
804493 101.966 -51.801 49.277
804494 101.980 -51.819 49.254
804495 101.995 -51.835 49.231
804496 102.021 -51.816 49.206
804497 102.045 -51.795 49.182
804498 102.070 -51.774 49.157
804499 102.095 -51.753 49.134
804500 102.109 -51.760 49.110
804501 102.111 -51.826 49.085
804502 102.138 -51.841 49.058
804503 102.165 -51.855 49.029
804504 102.192 -51.870 49.000
804505 102.219 -51.884 48.969
804506 102.246 -51.898 48.936
804507 102.273 -51.913 48.901
804508 102.300 -51.928 48.865
804509 102.327 -51.942 48.826
804510 102.354 -51.957 48.788
804511 102.381 -51.971 48.749
805137 101.080 -51.058 51.275
805138 101.106 -51.074 51.238
805139 101.132 -51.089 51.202
805140 101.147 -51.111 51.165
805141 101.162 -51.132 51.129
805142 101.177 -51.153 51.092
805143 101.192 -51.174 51.055
805144 101.207 -51.196 51.016
805145 101.222 -51.217 50.975
805146 101.237 -51.238 50.934
805147 101.252 -51.259 50.892
805148 101.267 -51.280 50.846
805149 101.285 -51.294 50.799
805150 101.306 -51.296 50.748
805151 101.327 -51.299 50.696
805156 101.430 -51.321 50.366
805157 101.448 -51.331 50.311
805158 101.466 -51.340 50.255
805159 101.485 -51.351 50.203
805160 101.503 -51.361 50.151
805161 101.521 -51.371 50.103
805162 101.540 -51.381 50.056
805163 101.558 -51.391 50.012
805164 101.576 -51.401 49.972
805165 101.595 -51.412 49.934
805166 101.613 -51.421 49.899
805167 101.631 -51.431 49.867
805168 101.650 -51.442 49.837
805169 101.668 -51.452 49.809
805170 101.686 -51.462 49.783
805171 101.704 -51.472 49.758
805172 101.709 -51.493 49.733
805173 101.718 -51.507 49.710
805174 101.729 -51.519 49.688
805175 101.739 -51.531 49.666
805176 101.749 -51.543 49.645
805177 101.759 -51.554 49.624
805178 101.769 -51.566 49.605
805179 101.779 -51.578 49.586
805180 101.790 -51.590 49.566
805181 101.800 -51.603 49.546
805182 101.812 -51.616 49.523
805183 101.823 -51.630 49.503
805184 101.834 -51.644 49.484
805185 101.845 -51.658 49.465
805186 101.856 -51.672 49.446
805187 101.870 -51.682 49.426
805188 101.885 -51.690 49.407
805189 101.900 -51.698 49.390
805190 101.915 -51.706 49.374
805191 101.930 -51.716 49.351
805192 101.944 -51.734 49.327
805193 101.959 -51.751 49.303
805194 101.973 -51.768 49.280
805195 101.987 -51.785 49.256
805196 102.004 -51.801 49.233
805197 102.029 -51.801 49.208
805198 102.053 -51.779 49.184
805199 102.079 -51.758 49.158
805200 102.103 -51.737 49.135
805201 102.128 -51.716 49.111
805202 102.114 -51.812 49.086
805203 102.141 -51.826 49.060
805204 102.168 -51.841 49.032
805205 102.195 -51.855 49.003
805206 102.223 -51.870 48.973
805207 102.249 -51.885 48.941
805208 102.276 -51.899 48.907
805209 102.304 -51.913 48.872
805210 102.330 -51.928 48.835
805211 102.358 -51.943 48.798
805839 101.106 -51.050 51.249
805840 101.132 -51.066 51.212
805841 101.159 -51.082 51.175
805842 101.182 -51.099 51.139
805843 101.197 -51.120 51.102
805844 101.212 -51.141 51.065
805845 101.227 -51.162 51.027
805846 101.242 -51.183 50.987
805847 101.257 -51.205 50.947
805848 101.272 -51.227 50.905
805849 101.287 -51.248 50.861
805850 101.302 -51.269 50.815
805851 101.319 -51.284 50.765
805856 101.425 -51.299 50.445
805857 101.445 -51.304 50.389
805858 101.463 -51.313 50.333
805859 101.481 -51.324 50.278
805860 101.499 -51.334 50.226
805861 101.518 -51.343 50.175
805862 101.536 -51.354 50.127
805863 101.554 -51.364 50.081
805864 101.573 -51.374 50.037
805865 101.591 -51.384 49.997
805866 101.609 -51.394 49.960
805867 101.628 -51.404 49.926
805868 101.646 -51.415 49.895
805869 101.664 -51.425 49.866
805870 101.683 -51.434 49.838
805871 101.701 -51.444 49.812
805872 101.714 -51.458 49.785
805873 101.720 -51.479 49.760
805874 101.728 -51.496 49.735
805875 101.738 -51.508 49.711
805876 101.748 -51.519 49.688
805877 101.758 -51.531 49.665
805878 101.768 -51.543 49.643
805879 101.778 -51.555 49.623
805880 101.789 -51.567 49.604
805881 101.799 -51.578 49.583
805882 101.809 -51.591 49.563
805883 101.819 -51.603 49.540
805884 101.829 -51.614 49.519
805885 101.839 -51.626 49.500
805886 101.850 -51.638 49.481
805887 101.861 -51.652 49.462
805888 101.876 -51.660 49.439
805889 101.891 -51.668 49.418
805890 101.906 -51.675 49.400
805891 101.921 -51.683 49.382
805892 101.937 -51.691 49.359
805893 101.952 -51.700 49.333
805894 101.966 -51.718 49.307
805895 101.980 -51.735 49.283
805896 101.995 -51.752 49.259
805897 102.013 -51.767 49.235
805898 102.037 -51.779 49.210
805899 102.062 -51.763 49.186
805900 102.087 -51.742 49.160
805901 102.112 -51.721 49.137
805902 102.136 -51.700 49.111
805903 102.117 -51.797 49.088
805904 102.144 -51.812 49.062
805905 102.172 -51.827 49.035
805906 102.198 -51.841 49.007
805907 102.226 -51.856 48.977
805908 102.253 -51.870 48.945
805909 102.279 -51.885 48.913
805910 102.307 -51.900 48.879
805911 102.334 -51.914 48.843
805912 102.361 -51.929 48.808
806542 101.158 -51.059 51.185
806543 101.184 -51.074 51.149
806544 101.210 -51.090 51.112
806545 101.232 -51.108 51.076
806546 101.247 -51.130 51.038
806547 101.262 -51.151 50.999
806548 101.277 -51.172 50.959
806549 101.292 -51.193 50.918
806550 101.307 -51.215 50.876
806551 101.322 -51.236 50.830
806557 101.438 -51.286 50.466
806558 101.460 -51.289 50.411
806559 101.478 -51.296 50.356
806560 101.496 -51.306 50.301
806561 101.515 -51.316 50.250
806562 101.533 -51.327 50.200
806563 101.551 -51.337 50.152
806564 101.569 -51.346 50.106
806565 101.588 -51.357 50.063
806566 101.606 -51.367 50.022
806567 101.624 -51.377 49.986
806568 101.642 -51.387 49.953
806569 101.661 -51.397 49.923
806570 101.679 -51.407 49.895
806571 101.697 -51.417 49.867
806572 101.716 -51.426 49.841
806573 101.725 -51.444 49.813
806574 101.731 -51.465 49.786
806575 101.737 -51.484 49.760
806576 101.747 -51.497 49.736
806577 101.757 -51.508 49.711
806578 101.768 -51.520 49.687
806579 101.778 -51.532 49.664
806580 101.788 -51.544 49.643
806581 101.798 -51.556 49.622
806582 101.808 -51.568 49.600
806583 101.818 -51.579 49.580
806584 101.829 -51.591 49.557
806585 101.839 -51.603 49.535
806586 101.849 -51.615 49.515
806587 101.858 -51.623 49.497
806588 101.869 -51.627 49.478
806589 101.884 -51.636 49.454
806590 101.898 -51.645 49.430
806591 101.913 -51.653 49.410
806592 101.928 -51.661 49.389
806593 101.943 -51.669 49.366
806594 101.958 -51.676 49.338
806595 101.973 -51.685 49.312
806596 101.988 -51.702 49.286
806597 102.002 -51.719 49.262
806598 102.023 -51.733 49.238
806599 102.046 -51.745 49.213
806600 102.070 -51.747 49.188
806601 102.095 -51.726 49.163
806602 102.120 -51.705 49.138
806603 102.144 -51.684 49.114
806604 102.121 -51.784 49.090
806605 102.147 -51.798 49.064
806606 102.175 -51.813 49.037
806607 102.202 -51.827 49.009
806608 102.229 -51.841 48.981
806609 102.256 -51.857 48.951
806610 102.283 -51.871 48.919
806611 102.310 -51.885 48.886
806612 102.337 -51.900 48.851
807245 101.210 -51.067 51.122
807246 101.237 -51.083 51.086
807247 101.263 -51.099 51.049
807248 101.282 -51.117 51.010
807249 101.296 -51.139 50.971
807250 101.312 -51.160 50.931
807258 101.451 -51.274 50.488
807259 101.472 -51.277 50.433
807260 101.493 -51.280 50.378
807261 101.511 -51.289 50.325
807262 101.530 -51.299 50.274
807263 101.548 -51.309 50.225
807264 101.566 -51.319 50.177
807265 101.585 -51.330 50.132
807266 101.603 -51.340 50.089
807267 101.621 -51.349 50.050
807268 101.639 -51.359 50.014
807269 101.658 -51.370 49.982
807270 101.676 -51.380 49.952
807271 101.694 -51.390 49.925
807272 101.712 -51.400 49.899
807273 101.731 -51.409 49.871
807274 101.736 -51.431 49.843
807275 101.741 -51.452 49.814
807276 101.747 -51.473 49.786
807277 101.757 -51.485 49.760
807278 101.767 -51.497 49.734
807279 101.777 -51.509 49.709
807280 101.787 -51.521 49.686
807281 101.797 -51.533 49.662
807282 101.808 -51.544 49.639
807283 101.818 -51.556 49.616
807284 101.828 -51.569 49.594
807285 101.838 -51.580 49.571
807286 101.848 -51.592 49.549
807287 101.856 -51.593 49.530
807288 101.864 -51.593 49.512
807289 101.878 -51.600 49.493
807290 101.892 -51.609 49.471
807291 101.907 -51.618 49.444
807292 101.921 -51.627 49.421
807293 101.936 -51.636 49.396
807294 101.950 -51.645 49.372
807295 101.965 -51.653 49.344
807296 101.980 -51.662 49.317
807297 101.995 -51.669 49.290
807298 102.010 -51.686 49.265
807299 102.032 -51.699 49.241
807300 102.055 -51.711 49.216
807301 102.079 -51.724 49.190
807302 102.103 -51.710 49.165
807303 102.128 -51.689 49.140
807304 102.153 -51.668 49.116
807305 102.149 -51.717 49.091
807306 102.151 -51.784 49.065
807307 102.178 -51.798 49.038
807308 102.205 -51.813 49.012
807309 102.232 -51.828 48.984
807310 102.259 -51.842 48.955
807311 102.286 -51.857 48.924
807312 102.313 -51.871 48.892
807960 101.485 -51.264 50.454
807961 101.506 -51.267 50.401
807962 101.526 -51.272 50.349
807963 101.544 -51.282 50.298
807964 101.563 -51.292 50.249
807965 101.581 -51.302 50.202
807966 101.599 -51.312 50.157
807967 101.618 -51.322 50.115
807968 101.636 -51.332 50.076
807969 101.654 -51.343 50.041
807970 101.673 -51.352 50.010
807971 101.691 -51.362 49.982
807972 101.709 -51.373 49.955
807973 101.728 -51.382 49.929
807974 101.741 -51.395 49.900
807975 101.746 -51.417 49.871
807976 101.752 -51.438 49.842
807977 101.757 -51.460 49.812
807978 101.766 -51.474 49.784
807979 101.776 -51.486 49.757
807980 101.786 -51.498 49.731
807981 101.796 -51.509 49.706
807982 101.807 -51.522 49.682
807983 101.817 -51.533 49.656
807984 101.827 -51.545 49.633
807985 101.837 -51.557 49.609
807986 101.846 -51.562 49.585
807987 101.854 -51.562 49.563
807988 101.862 -51.562 49.545
807989 101.872 -51.565 49.528
807990 101.886 -51.574 49.509
807991 101.901 -51.583 49.487
807992 101.915 -51.591 49.459
807993 101.930 -51.600 49.432
807994 101.944 -51.609 49.405
807995 101.959 -51.618 49.380
807996 101.973 -51.627 49.351
807997 101.988 -51.636 49.322
807998 102.003 -51.645 49.295
807999 102.019 -51.657 49.270
808000 102.041 -51.665 49.245
808001 102.064 -51.677 49.219
808002 102.088 -51.690 49.193
808003 102.111 -51.694 49.167
808004 102.136 -51.673 49.143
808005 102.161 -51.652 49.117
808006 102.179 -51.649 49.093
808007 102.154 -51.770 49.067
808008 102.181 -51.785 49.040
808009 102.208 -51.799 49.014
808010 102.235 -51.813 48.987
808011 102.262 -51.828 48.959
808012 102.290 -51.842 48.929
808013 102.316 -51.857 48.898
808661 101.498 -51.251 50.476
808662 101.519 -51.254 50.424
808663 101.541 -51.257 50.372
808664 101.559 -51.265 50.322
808665 101.578 -51.274 50.273
808666 101.596 -51.285 50.227
808667 101.614 -51.295 50.181
808668 101.632 -51.305 50.141
808669 101.651 -51.315 50.103
808670 101.669 -51.325 50.068
808671 101.687 -51.335 50.037
808672 101.706 -51.346 50.011
808673 101.724 -51.356 49.985
808674 101.745 -51.362 49.958
808675 101.752 -51.381 49.929
808676 101.757 -51.403 49.899
808677 101.762 -51.425 49.869
808678 101.767 -51.446 49.838
808679 101.775 -51.462 49.808
808680 101.785 -51.475 49.781
808681 101.796 -51.487 49.754
808682 101.806 -51.498 49.726
808683 101.816 -51.510 49.700
808684 101.826 -51.522 49.674
808685 101.836 -51.531 49.650
808686 101.843 -51.531 49.624
808687 101.851 -51.531 49.600
808688 101.859 -51.531 49.579
808689 101.867 -51.531 49.561
808690 101.880 -51.538 49.545
808691 101.895 -51.547 49.526
808692 101.909 -51.556 49.502
808693 101.924 -51.565 49.474
808694 101.939 -51.574 49.443
808695 101.953 -51.583 49.414
808696 101.967 -51.591 49.390
808697 101.983 -51.603 49.359
808698 101.999 -51.620 49.328
808699 102.016 -51.638 49.300
808700 102.033 -51.650 49.274
808701 102.050 -51.633 49.248
808702 102.073 -51.643 49.221
808703 102.096 -51.656 49.196
808704 102.120 -51.668 49.170
808705 102.144 -51.657 49.144
808706 102.169 -51.636 49.119
808707 102.194 -51.615 49.093
808708 102.158 -51.756 49.068
808709 102.184 -51.770 49.041
808710 102.211 -51.785 49.016
808711 102.239 -51.799 48.990
808712 102.266 -51.814 48.962
808713 102.293 -51.829 48.933
808714 102.320 -51.843 48.902
808715 102.347 -51.857 48.870
808716 102.374 -51.872 48.838
808717 102.401 -51.886 48.805
809362 101.512 -51.239 50.496
809363 101.533 -51.242 50.445
809364 101.554 -51.245 50.394
809365 101.574 -51.249 50.344
809366 101.593 -51.258 50.296
809367 101.611 -51.268 50.249
809368 101.629 -51.277 50.205
809369 101.647 -51.288 50.164
809370 101.666 -51.298 50.127
809371 101.684 -51.308 50.092
809372 101.702 -51.318 50.063
809373 101.721 -51.328 50.037
809374 101.739 -51.338 50.011
809375 101.757 -51.346 49.984
809376 101.762 -51.368 49.954
809377 101.767 -51.390 49.923
809378 101.773 -51.411 49.892
809379 101.778 -51.433 49.861
809380 101.785 -51.452 49.830
809381 101.795 -51.463 49.801
809382 101.805 -51.475 49.774
809383 101.815 -51.487 49.745
809384 101.825 -51.499 49.717
809385 101.833 -51.500 49.691
809386 101.841 -51.500 49.666
809387 101.849 -51.500 49.640
809388 101.857 -51.500 49.616
809389 101.865 -51.500 49.595
809390 101.875 -51.503 49.576
809391 101.889 -51.512 49.558
809392 101.903 -51.520 49.538
809393 101.918 -51.529 49.515
809394 101.933 -51.538 49.486
809395 101.947 -51.547 49.453
809396 101.964 -51.564 49.424
809397 101.980 -51.581 49.397
809398 101.997 -51.599 49.366
809399 102.014 -51.616 49.335
809400 102.031 -51.634 49.306
809401 102.048 -51.643 49.279
809402 102.065 -51.626 49.252
809403 102.082 -51.609 49.225
809404 102.105 -51.622 49.198
809405 102.129 -51.634 49.171
809406 102.152 -51.641 49.146
809407 102.177 -51.620 49.120
809408 102.202 -51.599 49.095
809409 102.161 -51.741 49.069
809410 102.188 -51.756 49.043
809411 102.215 -51.771 49.018
809412 102.242 -51.785 48.993
809413 102.269 -51.800 48.965
809414 102.296 -51.814 48.937
809415 102.323 -51.829 48.907
809416 102.350 -51.843 48.876
809417 102.377 -51.858 48.845
809418 102.404 -51.873 48.812
809419 102.431 -51.887 48.778
809420 102.458 -51.901 48.742
810063 101.527 -51.220 50.515
810064 101.546 -51.229 50.464
810065 101.567 -51.232 50.414
810066 101.588 -51.235 50.365
810067 101.608 -51.240 50.317
810068 101.626 -51.250 50.271
810069 101.644 -51.261 50.227
810070 101.663 -51.271 50.186
810071 101.681 -51.280 50.149
810072 101.699 -51.291 50.114
810073 101.717 -51.301 50.085
810074 101.736 -51.311 50.060
810075 101.756 -51.318 50.034
810076 101.767 -51.333 50.006
810077 101.773 -51.354 49.976
810078 101.778 -51.376 49.945
810079 101.783 -51.397 49.913
810080 101.788 -51.419 49.882
810081 101.794 -51.440 49.851
810082 101.804 -51.453 49.821
810083 101.814 -51.464 49.793
810084 101.823 -51.470 49.764
810085 101.831 -51.470 49.735
810086 101.839 -51.470 49.708
810087 101.847 -51.469 49.682
810088 101.854 -51.469 49.655
810089 101.863 -51.469 49.632
810090 101.870 -51.469 49.609
810091 101.883 -51.476 49.589
810092 101.898 -51.485 49.569
810093 101.912 -51.494 49.549
810094 101.928 -51.508 49.526
810095 101.945 -51.525 49.498
810096 101.962 -51.543 49.464
810097 101.978 -51.560 49.434
810098 101.995 -51.578 49.405
810099 102.012 -51.595 49.374
810100 102.029 -51.613 49.343
810101 102.045 -51.630 49.313
810102 102.062 -51.636 49.284
810103 102.080 -51.619 49.256
810104 102.097 -51.602 49.228
810105 102.114 -51.587 49.201
810106 102.138 -51.600 49.172
810107 102.161 -51.613 49.147
810108 102.185 -51.604 49.121
810109 102.210 -51.583 49.096
810110 102.189 -51.675 49.071
810111 102.191 -51.742 49.045
810112 102.218 -51.757 49.021
810113 102.245 -51.771 48.995
810114 102.272 -51.785 48.969
810115 102.299 -51.800 48.941
810116 102.326 -51.815 48.912
810117 102.353 -51.829 48.882
810118 102.380 -51.844 48.851
810119 102.407 -51.858 48.819
810120 102.435 -51.873 48.786
810121 102.461 -51.887 48.752
810122 102.488 -51.902 48.716
810764 101.546 -51.186 50.533
810765 101.561 -51.208 50.483
810766 101.580 -51.219 50.433
810767 101.601 -51.222 50.385
810768 101.622 -51.225 50.338
810769 101.641 -51.233 50.292
810770 101.659 -51.243 50.248
810771 101.678 -51.253 50.207
810772 101.696 -51.264 50.169
810773 101.714 -51.274 50.135
810774 101.733 -51.284 50.105
810775 101.751 -51.293 50.080
810776 101.772 -51.299 50.054
810777 101.777 -51.320 50.026
810778 101.783 -51.341 49.995
810779 101.788 -51.362 49.963
810780 101.794 -51.384 49.932
810781 101.799 -51.406 49.901
810782 101.805 -51.427 49.871
810783 101.813 -51.439 49.841
810784 101.821 -51.439 49.812
810785 101.829 -51.439 49.782
810786 101.837 -51.439 49.752
810787 101.844 -51.439 49.724
810788 101.852 -51.439 49.698
810789 101.860 -51.439 49.670
810790 101.868 -51.438 49.646
810791 101.877 -51.440 49.622
810792 101.892 -51.452 49.601
810793 101.909 -51.469 49.579
810794 101.926 -51.487 49.558
810795 101.943 -51.504 49.536
810796 101.959 -51.522 49.508
810797 101.976 -51.539 49.476
810798 101.993 -51.556 49.445
810799 102.010 -51.574 49.415
810800 102.027 -51.591 49.385
810801 102.043 -51.608 49.354
810802 102.060 -51.625 49.322
810803 102.077 -51.629 49.292
810804 102.094 -51.612 49.261
810805 102.111 -51.595 49.230
810806 102.128 -51.578 49.201
810807 102.147 -51.566 49.173
810808 102.170 -51.578 49.148
810809 102.193 -51.588 49.122
810810 102.219 -51.567 49.096
810811 102.219 -51.607 49.071
810812 102.194 -51.728 49.046
810813 102.222 -51.742 49.022
810814 102.248 -51.757 48.997
810815 102.275 -51.772 48.971
810816 102.303 -51.786 48.945
810817 102.329 -51.801 48.916
810818 102.357 -51.815 48.887
810819 102.384 -51.829 48.857
810820 102.410 -51.844 48.826
810821 102.438 -51.859 48.794
810822 102.465 -51.873 48.760
810823 102.491 -51.888 48.724
810824 102.519 -51.902 48.690
811465 101.566 -51.154 50.549
811466 101.582 -51.175 50.500
811467 101.597 -51.196 50.451
811468 101.614 -51.210 50.404
811469 101.635 -51.212 50.357
811470 101.656 -51.216 50.312
811471 101.674 -51.226 50.268
811472 101.692 -51.236 50.227
811473 101.711 -51.246 50.188
811474 101.729 -51.256 50.153
811475 101.747 -51.267 50.122
811476 101.767 -51.274 50.094
811477 101.778 -51.290 50.066
811478 101.784 -51.310 50.038
811479 101.790 -51.331 50.008
811480 101.795 -51.352 49.977
811481 101.801 -51.374 49.947
811482 101.807 -51.395 49.917
811483 101.813 -51.414 49.888
811484 101.825 -51.418 49.860
811485 101.832 -51.418 49.831
811486 101.840 -51.417 49.801
811487 101.847 -51.417 49.769
811488 101.855 -51.417 49.739
811489 101.863 -51.416 49.712
811490 101.871 -51.416 49.684
811491 101.879 -51.416 49.658
811492 101.891 -51.428 49.634
811493 101.907 -51.448 49.611
811494 101.924 -51.465 49.589
811495 101.941 -51.483 49.567
811496 101.957 -51.500 49.546
811497 101.974 -51.517 49.520
811498 101.991 -51.534 49.489
811499 102.007 -51.552 49.456
811500 102.024 -51.569 49.425
811501 102.041 -51.587 49.395
811502 102.058 -51.604 49.364
811503 102.075 -51.622 49.332
811504 102.092 -51.622 49.300
811505 102.109 -51.605 49.267
811506 102.126 -51.588 49.234
811507 102.143 -51.571 49.203
811508 102.161 -51.554 49.174
811509 102.179 -51.544 49.148
811510 102.202 -51.556 49.123
811511 102.227 -51.551 49.097
811512 102.248 -51.538 49.072
811513 102.197 -51.713 49.047
811514 102.225 -51.729 49.023
811515 102.252 -51.743 48.999
811516 102.278 -51.757 48.974
811517 102.306 -51.772 48.948
811518 102.333 -51.786 48.919
811519 102.360 -51.801 48.891
811520 102.387 -51.816 48.862
811521 102.414 -51.830 48.832
811522 102.441 -51.845 48.800
811523 102.468 -51.859 48.767
811524 102.495 -51.873 48.732
811525 102.522 -51.888 48.698
811526 102.549 -51.903 48.664
812166 101.575 -51.126 50.565
812167 101.601 -51.142 50.517
812168 101.616 -51.163 50.469
812169 101.631 -51.184 50.422
812170 101.649 -51.200 50.376
812171 101.670 -51.203 50.330
812172 101.689 -51.209 50.287
812173 101.707 -51.219 50.245
812174 101.726 -51.229 50.205
812175 101.744 -51.240 50.169
812176 101.762 -51.249 50.137
812177 101.779 -51.258 50.106
812178 101.784 -51.280 50.078
812179 101.790 -51.301 50.049
812180 101.796 -51.322 50.019
812181 101.802 -51.343 49.990
812182 101.808 -51.363 49.961
812183 101.814 -51.381 49.932
812184 101.820 -51.399 49.904
812185 101.839 -51.401 49.877
812186 101.846 -51.401 49.848
812187 101.854 -51.401 49.817
812188 101.862 -51.400 49.785
812189 101.870 -51.400 49.754
812190 101.877 -51.400 49.726
812191 101.885 -51.399 49.697
812192 101.893 -51.399 49.670
812193 101.908 -51.417 49.645
812194 101.923 -51.440 49.621
812195 101.939 -51.461 49.599
812196 101.955 -51.478 49.577
812197 101.972 -51.496 49.556
812198 101.988 -51.513 49.532
812199 102.006 -51.531 49.500
812200 102.022 -51.548 49.467
812201 102.039 -51.566 49.435
812202 102.056 -51.583 49.405
812203 102.073 -51.600 49.373
812204 102.089 -51.618 49.340
812205 102.106 -51.615 49.306
812206 102.123 -51.598 49.272
812207 102.140 -51.581 49.238
812208 102.158 -51.564 49.205
812209 102.175 -51.547 49.175
812210 102.192 -51.530 49.148
812211 102.211 -51.522 49.123
812212 102.235 -51.534 49.098
812213 102.260 -51.514 49.072
812214 102.201 -51.700 49.048
812215 102.228 -51.714 49.024
812216 102.255 -51.729 49.000
812217 102.282 -51.743 48.976
812218 102.309 -51.757 48.950
812219 102.336 -51.772 48.922
812220 102.363 -51.787 48.894
812221 102.390 -51.801 48.866
812222 102.417 -51.816 48.837
812223 102.444 -51.830 48.806
812224 102.471 -51.845 48.773
812225 102.498 -51.860 48.738
812226 102.525 -51.874 48.705
812227 102.552 -51.888 48.672
812228 102.579 -51.903 48.639
812867 101.575 -51.103 50.581
812868 101.601 -51.119 50.534
812869 101.628 -51.134 50.487
812870 101.651 -51.152 50.440
812871 101.666 -51.173 50.394
812872 101.683 -51.190 50.348
812873 101.704 -51.193 50.304
812874 101.723 -51.202 50.261
812875 101.741 -51.212 50.222
812876 101.759 -51.222 50.185
812877 101.779 -51.229 50.151
812878 101.785 -51.249 50.119
812879 101.791 -51.271 50.089
812880 101.796 -51.292 50.059
812881 101.802 -51.312 50.030
812882 101.809 -51.330 50.002
812883 101.815 -51.348 49.973
812884 101.821 -51.366 49.945
812885 101.832 -51.382 49.918
812886 101.853 -51.384 49.891
812887 101.861 -51.384 49.862
812888 101.869 -51.384 49.832
812889 101.876 -51.384 49.800
812890 101.884 -51.384 49.769
812891 101.891 -51.384 49.739
812892 101.899 -51.383 49.709
812893 101.907 -51.384 49.683
812894 101.925 -51.405 49.657
812895 101.940 -51.428 49.632
812896 101.955 -51.450 49.610
812897 101.970 -51.473 49.588
812898 101.986 -51.492 49.567
812899 102.003 -51.509 49.542
812900 102.020 -51.527 49.510
812901 102.037 -51.544 49.477
812902 102.053 -51.562 49.445
812903 102.070 -51.579 49.413
812904 102.087 -51.597 49.378
812905 102.104 -51.613 49.345
812906 102.121 -51.608 49.310
812907 102.138 -51.591 49.277
812908 102.155 -51.574 49.243
812909 102.173 -51.557 49.209
812910 102.190 -51.540 49.177
812911 102.207 -51.523 49.149
812912 102.224 -51.506 49.122
812913 102.244 -51.500 49.098
812914 102.268 -51.498 49.073
812915 102.229 -51.633 49.048
812916 102.231 -51.700 49.025
812917 102.258 -51.714 49.001
812918 102.285 -51.729 48.976
812919 102.312 -51.744 48.951
812920 102.339 -51.758 48.924
812921 102.366 -51.772 48.896
812922 102.393 -51.787 48.869
812923 102.420 -51.802 48.840
812924 102.447 -51.816 48.810
812925 102.474 -51.831 48.778
812926 102.501 -51.845 48.745
812927 102.529 -51.860 48.713
812928 102.555 -51.874 48.680
812929 102.583 -51.889 48.647
812930 102.610 -51.904 48.613
813568 101.574 -51.080 50.596
813569 101.601 -51.096 50.549
813570 101.627 -51.111 50.503
813571 101.653 -51.127 50.456
813572 101.680 -51.142 50.411
813573 101.701 -51.161 50.365
813574 101.716 -51.180 50.321
813575 101.737 -51.184 50.278
813576 101.756 -51.195 50.239
813577 101.774 -51.205 50.201
813578 101.786 -51.219 50.166
813579 101.792 -51.240 50.133
813580 101.798 -51.261 50.101
813581 101.804 -51.279 50.072
813582 101.810 -51.297 50.042
813583 101.816 -51.315 50.014
813584 101.822 -51.334 49.985
813585 101.828 -51.352 49.958
813586 101.844 -51.365 49.930
813587 101.867 -51.368 49.902
813588 101.875 -51.368 49.873
813589 101.883 -51.368 49.843
813590 101.890 -51.367 49.812
813591 101.898 -51.367 49.781
813592 101.906 -51.367 49.752
813593 101.913 -51.366 49.722
813594 101.924 -51.373 49.694
813595 101.942 -51.393 49.667
813596 101.957 -51.416 49.643
813597 101.972 -51.439 49.621
813598 101.987 -51.462 49.599
813599 102.002 -51.484 49.576
813600 102.018 -51.505 49.549
813601 102.035 -51.522 49.517
813602 102.051 -51.540 49.484
813603 102.068 -51.557 49.451
813604 102.085 -51.575 49.418
813605 102.102 -51.592 49.383
813606 102.118 -51.609 49.348
813607 102.135 -51.601 49.311
813608 102.153 -51.584 49.276
813609 102.170 -51.567 49.243
813610 102.187 -51.550 49.209
813611 102.204 -51.533 49.177
813612 102.222 -51.516 49.150
813613 102.239 -51.499 49.123
813614 102.256 -51.482 49.098
813615 102.276 -51.479 49.074
813616 102.259 -51.565 49.049
813617 102.234 -51.686 49.027
813618 102.261 -51.700 49.002
813619 102.289 -51.715 48.977
813620 102.315 -51.729 48.952
813621 102.342 -51.744 48.926
813622 102.370 -51.759 48.899
813623 102.397 -51.773 48.871
813624 102.423 -51.788 48.843
813625 102.451 -51.802 48.814
813626 102.478 -51.816 48.783
813627 102.504 -51.831 48.750
813628 102.532 -51.846 48.719
813629 102.559 -51.860 48.687
813630 102.586 -51.875 48.654
813631 102.613 -51.889 48.620
814269 101.574 -51.058 50.609
814270 101.601 -51.073 50.563
814271 101.627 -51.089 50.518
814272 101.653 -51.104 50.472
814273 101.679 -51.120 50.427
814274 101.706 -51.136 50.381
814275 101.732 -51.151 50.337
814276 101.751 -51.171 50.295
814277 101.770 -51.177 50.255
814278 101.786 -51.189 50.217
814279 101.793 -51.210 50.180
814280 101.799 -51.228 50.146
814281 101.805 -51.246 50.114
814282 101.811 -51.265 50.083
814283 101.817 -51.283 50.053
814284 101.823 -51.301 50.025
814285 101.829 -51.319 49.996
814286 101.835 -51.337 49.969
814287 101.857 -51.348 49.941
814288 101.881 -51.352 49.913
814289 101.889 -51.351 49.884
814290 101.897 -51.351 49.854
814291 101.904 -51.351 49.824
814292 101.912 -51.350 49.793
814293 101.920 -51.350 49.763
814294 101.927 -51.349 49.733
814295 101.941 -51.362 49.705
814296 101.959 -51.382 49.678
814297 101.974 -51.405 49.653
814298 101.989 -51.428 49.630
814299 102.004 -51.450 49.607
814300 102.019 -51.473 49.583
814301 102.034 -51.496 49.555
814302 102.049 -51.519 49.522
814303 102.066 -51.536 49.489
814304 102.083 -51.553 49.456
814305 102.099 -51.571 49.422
814306 102.116 -51.588 49.386
814307 102.133 -51.606 49.350
814308 102.150 -51.594 49.312
814309 102.167 -51.577 49.276
814310 102.184 -51.559 49.243
814311 102.202 -51.543 49.210
814312 102.219 -51.526 49.178
814313 102.236 -51.509 49.150
814314 102.253 -51.492 49.124
814315 102.270 -51.475 49.098
814316 102.288 -51.458 49.075
814317 102.288 -51.497 49.051
814318 102.238 -51.672 49.028
814319 102.265 -51.686 49.004
814320 102.292 -51.700 48.979
814321 102.319 -51.716 48.954
814322 102.346 -51.730 48.928
814323 102.373 -51.744 48.901
814324 102.400 -51.759 48.873
814325 102.427 -51.773 48.846
814326 102.454 -51.788 48.818
814327 102.481 -51.803 48.787
814328 102.508 -51.817 48.755
814329 102.535 -51.832 48.724
814330 102.562 -51.846 48.692
814331 102.589 -51.860 48.660
814332 102.616 -51.875 48.628
814333 102.643 -51.890 48.595
814970 101.573 -51.034 50.621
814971 101.600 -51.050 50.577
814972 101.626 -51.066 50.532
814973 101.653 -51.081 50.486
814974 101.679 -51.097 50.442
814975 101.705 -51.112 50.396
814976 101.732 -51.128 50.352
814977 101.758 -51.144 50.310
814978 101.784 -51.159 50.270
814979 101.793 -51.177 50.231
814980 101.799 -51.196 50.194
814981 101.806 -51.214 50.158
814982 101.812 -51.232 50.125
814983 101.818 -51.250 50.094
814984 101.824 -51.268 50.064
814985 101.830 -51.287 50.035
814986 101.836 -51.305 50.007
814987 101.842 -51.323 49.978
814988 101.869 -51.331 49.950
814989 101.895 -51.335 49.922
814990 101.903 -51.334 49.893
814991 101.911 -51.334 49.864
814992 101.918 -51.334 49.835
814993 101.926 -51.334 49.804
814994 101.934 -51.334 49.773
814995 101.942 -51.334 49.743
814996 101.958 -51.350 49.716
814997 101.976 -51.371 49.689
814998 101.991 -51.393 49.663
814999 102.007 -51.415 49.638
815000 102.022 -51.438 49.614
815001 102.036 -51.462 49.589
815002 102.051 -51.484 49.560
815003 102.066 -51.507 49.526
815004 102.081 -51.530 49.492
815005 102.097 -51.550 49.458
815006 102.114 -51.567 49.424
815007 102.131 -51.584 49.388
815008 102.147 -51.601 49.351
815009 102.165 -51.587 49.314
815010 102.182 -51.570 49.278
815011 102.199 -51.553 49.244
815012 102.217 -51.536 49.211
815013 102.234 -51.519 49.180
815014 102.251 -51.502 49.152
815015 102.268 -51.485 49.126
815016 102.285 -51.468 49.100
815017 102.296 -51.462 49.078
815018 102.307 -51.456 49.055
815019 102.241 -51.657 49.031
815020 102.268 -51.672 49.006
815021 102.295 -51.687 48.981
815022 102.322 -51.701 48.956
815023 102.349 -51.716 48.930
815024 102.376 -51.730 48.904
815025 102.403 -51.744 48.876
815026 102.430 -51.760 48.849
815027 102.457 -51.774 48.821
815028 102.484 -51.788 48.791
815029 102.511 -51.803 48.760
815030 102.538 -51.817 48.729
815031 102.565 -51.832 48.698
815032 102.592 -51.847 48.666
815033 102.619 -51.861 48.634
815034 102.647 -51.876 48.602
815035 102.673 -51.890 48.569
815036 102.700 -51.904 48.537
815670 101.547 -50.996 50.676
815671 101.573 -51.011 50.633
815672 101.600 -51.027 50.589
815673 101.626 -51.043 50.544
815674 101.652 -51.058 50.500
815675 101.679 -51.074 50.455
815676 101.705 -51.089 50.410
815677 101.731 -51.105 50.366
815678 101.758 -51.121 50.324
815679 101.784 -51.141 50.283
815680 101.801 -51.163 50.245
815681 101.807 -51.181 50.207
815682 101.813 -51.199 50.171
815683 101.819 -51.218 50.136
815684 101.825 -51.236 50.104
815685 101.831 -51.254 50.073
815686 101.837 -51.272 50.044
815687 101.843 -51.290 50.015
815688 101.850 -51.308 49.986
815689 101.881 -51.313 49.958
815690 101.910 -51.318 49.930
815691 101.917 -51.318 49.901
815692 101.925 -51.318 49.872
815693 101.933 -51.318 49.842
815694 101.941 -51.317 49.813
815695 101.948 -51.317 49.784
815696 101.957 -51.318 49.755
815697 101.975 -51.339 49.725
815698 101.993 -51.359 49.698
815699 102.009 -51.381 49.670
815700 102.024 -51.404 49.644
815701 102.039 -51.427 49.619
815702 102.053 -51.450 49.593
815703 102.068 -51.472 49.563
815704 102.084 -51.495 49.529
815705 102.098 -51.518 49.494
815706 102.113 -51.541 49.460
815707 102.128 -51.562 49.426
815708 102.146 -51.580 49.389
815709 102.162 -51.597 49.354
815710 102.179 -51.580 49.317
815711 102.196 -51.562 49.281
815712 102.214 -51.546 49.248
815713 102.231 -51.528 49.215
815714 102.246 -51.515 49.184
815715 102.257 -51.510 49.156
815716 102.267 -51.505 49.129
815717 102.277 -51.500 49.104
815718 102.288 -51.495 49.081
815719 102.298 -51.490 49.058
815720 102.269 -51.591 49.034
815721 102.271 -51.658 49.009
815722 102.298 -51.672 48.984
815723 102.325 -51.687 48.958
815724 102.352 -51.701 48.932
815725 102.379 -51.716 48.905
815726 102.406 -51.731 48.879
815727 102.434 -51.745 48.852
815728 102.460 -51.760 48.824
815729 102.487 -51.774 48.794
815730 102.515 -51.788 48.764
815731 102.541 -51.804 48.734
815732 102.569 -51.818 48.704
815733 102.596 -51.832 48.672
815734 102.622 -51.847 48.641
815735 102.650 -51.861 48.609
815736 102.677 -51.876 48.576
815737 102.703 -51.891 48.545
815738 102.731 -51.905 48.514
816371 101.546 -50.973 50.686
816372 101.572 -50.989 50.644
816373 101.599 -51.005 50.600
816374 101.625 -51.020 50.557
816375 101.652 -51.036 50.512
816376 101.678 -51.052 50.468
816377 101.704 -51.067 50.424
816378 101.731 -51.083 50.380
816379 101.757 -51.098 50.338
816380 101.784 -51.124 50.296
816381 101.808 -51.148 50.258
816382 101.814 -51.166 50.220
816383 101.820 -51.184 50.183
816384 101.826 -51.203 50.147
816385 101.832 -51.221 50.115
816386 101.838 -51.240 50.083
816387 101.844 -51.258 50.053
816388 101.850 -51.276 50.024
816389 101.863 -51.291 49.994
816390 101.893 -51.296 49.966
816391 101.924 -51.302 49.938
816392 101.932 -51.302 49.909
816393 101.939 -51.301 49.880
816394 101.947 -51.301 49.851
816395 101.955 -51.300 49.823
816396 101.962 -51.300 49.795
816397 101.973 -51.307 49.767
816398 101.991 -51.327 49.737
816399 102.010 -51.348 49.707
816400 102.026 -51.370 49.678
816401 102.041 -51.393 49.650
816402 102.056 -51.415 49.624
816403 102.070 -51.438 49.597
816404 102.086 -51.461 49.566
816405 102.101 -51.484 49.533
816406 102.115 -51.506 49.497
816407 102.130 -51.529 49.463
816408 102.146 -51.552 49.429
816409 102.161 -51.575 49.393
816410 102.177 -51.590 49.359
816411 102.194 -51.572 49.322
816412 102.206 -51.564 49.287
816413 102.217 -51.559 49.254
816414 102.228 -51.554 49.221
816415 102.238 -51.549 49.190
816416 102.248 -51.544 49.161
816417 102.258 -51.539 49.133
816418 102.269 -51.534 49.107
816419 102.279 -51.529 49.084
816420 102.290 -51.525 49.061
816421 102.299 -51.523 49.037
816422 102.274 -51.644 49.012
816423 102.302 -51.658 48.986
816424 102.328 -51.673 48.960
816425 102.356 -51.688 48.933
816426 102.383 -51.702 48.907
816427 102.409 -51.716 48.881
816428 102.437 -51.731 48.855
816429 102.464 -51.745 48.827
816430 102.490 -51.760 48.798
816431 102.518 -51.775 48.768
816432 102.545 -51.789 48.739
816433 102.572 -51.804 48.710
816434 102.599 -51.818 48.679
816435 102.626 -51.832 48.648
816436 102.653 -51.848 48.616
816437 102.680 -51.862 48.584
816438 102.707 -51.876 48.553
816439 102.734 -51.891 48.522
816440 102.761 -51.905 48.490
817072 101.546 -50.950 50.696
817073 101.572 -50.966 50.654
817074 101.599 -50.982 50.611
817075 101.625 -50.997 50.569
817076 101.651 -51.013 50.524
817077 101.678 -51.028 50.481
817078 101.704 -51.044 50.438
817079 101.730 -51.060 50.395
817080 101.757 -51.081 50.351
817081 101.783 -51.106 50.310
817082 101.810 -51.132 50.271
817083 101.821 -51.152 50.232
817084 101.827 -51.170 50.194
817085 101.833 -51.188 50.159
817086 101.839 -51.206 50.125
817087 101.845 -51.224 50.093
817088 101.852 -51.243 50.063
817089 101.858 -51.261 50.033
817090 101.875 -51.274 50.003
817091 101.905 -51.279 49.975
817092 101.936 -51.284 49.946
817093 101.946 -51.285 49.917
817094 101.953 -51.284 49.888
817095 101.961 -51.284 49.860
817096 101.969 -51.284 49.833
817097 101.976 -51.284 49.808
817098 101.990 -51.296 49.780
817099 102.009 -51.316 49.749
817100 102.027 -51.336 49.718
817101 102.043 -51.358 49.687
817102 102.058 -51.381 49.657
817103 102.073 -51.403 49.628
817104 102.088 -51.427 49.599
817105 102.103 -51.450 49.569
817106 102.118 -51.472 49.537
817107 102.132 -51.495 49.502
817108 102.147 -51.518 49.469
817109 102.163 -51.541 49.435
817110 102.178 -51.563 49.399
817111 102.193 -51.581 49.365
817112 102.199 -51.586 49.329
817113 102.203 -51.590 49.294
817114 102.208 -51.593 49.261
817115 102.219 -51.588 49.230
817116 102.229 -51.583 49.198
817117 102.239 -51.578 49.167
817118 102.250 -51.573 49.137
817119 102.260 -51.569 49.110
817120 102.270 -51.563 49.086
817121 102.281 -51.559 49.062
817122 102.291 -51.554 49.039
817123 102.277 -51.630 49.014
817124 102.305 -51.644 48.988
817125 102.332 -51.659 48.961
817126 102.359 -51.673 48.934
817127 102.386 -51.688 48.908
817128 102.413 -51.702 48.883
817129 102.440 -51.717 48.856
817130 102.467 -51.732 48.830
817131 102.494 -51.746 48.802
817132 102.521 -51.760 48.773
817133 102.548 -51.775 48.744
817134 102.575 -51.789 48.715
817135 102.602 -51.804 48.685
817136 102.629 -51.819 48.654
817137 102.656 -51.833 48.622
817138 102.683 -51.848 48.590
817139 102.710 -51.862 48.559
817140 102.737 -51.876 48.528
817141 102.764 -51.891 48.497
817142 102.791 -51.906 48.465
817773 101.545 -50.927 50.706
817774 101.571 -50.943 50.665
817775 101.598 -50.958 50.623
817776 101.624 -50.974 50.580
817777 101.650 -50.990 50.537
817778 101.677 -51.005 50.494
817779 101.703 -51.021 50.451
817780 101.730 -51.038 50.408
817781 101.757 -51.064 50.365
817782 101.783 -51.089 50.324
817783 101.810 -51.114 50.284
817784 101.828 -51.137 50.245
817785 101.834 -51.155 50.207
817786 101.840 -51.174 50.172
817787 101.847 -51.192 50.137
817788 101.853 -51.210 50.105
817789 101.859 -51.228 50.073
817790 101.865 -51.246 50.043
817791 101.887 -51.257 50.013
817792 101.917 -51.262 49.983
817793 101.948 -51.268 49.954
817794 101.960 -51.268 49.925
817795 101.968 -51.268 49.896
817796 101.975 -51.268 49.869
817797 101.983 -51.268 49.843
817798 101.991 -51.267 49.816
817799 102.007 -51.284 49.790
817800 102.025 -51.304 49.759
817801 102.043 -51.324 49.727
817802 102.060 -51.346 49.696
817803 102.075 -51.369 49.666
817804 102.090 -51.392 49.637
817805 102.105 -51.415 49.608
817806 102.120 -51.437 49.578
817807 102.135 -51.460 49.545
817808 102.150 -51.483 49.509
817809 102.165 -51.506 49.476
817810 102.180 -51.528 49.441
817811 102.195 -51.552 49.406
817812 102.210 -51.562 49.371
817813 102.222 -51.572 49.335
817814 102.227 -51.575 49.300
817815 102.231 -51.580 49.267
817816 102.235 -51.584 49.237
817817 102.240 -51.588 49.206
817818 102.244 -51.592 49.174
817819 102.248 -51.597 49.142
817820 102.253 -51.600 49.114
817821 102.262 -51.597 49.088
817822 102.272 -51.593 49.063
817823 102.282 -51.587 49.039
817824 102.281 -51.616 49.015
817825 102.308 -51.630 48.990
817826 102.335 -51.644 48.963
817827 102.362 -51.659 48.936
817828 102.389 -51.674 48.910
817829 102.416 -51.688 48.885
817830 102.443 -51.703 48.859
817831 102.470 -51.717 48.833
817832 102.497 -51.732 48.806
817833 102.524 -51.746 48.777
817834 102.551 -51.761 48.749
817835 102.578 -51.776 48.720
817836 102.605 -51.790 48.691
817837 102.632 -51.804 48.660
817838 102.660 -51.819 48.628
817839 102.688 -51.834 48.597
817840 102.716 -51.848 48.566
817841 102.743 -51.863 48.535
817842 102.771 -51.878 48.503
817843 102.799 -51.892 48.471
817844 102.826 -51.907 48.442
818475 101.571 -50.920 50.676
818476 101.598 -50.936 50.634
818477 101.624 -50.951 50.592
818478 101.650 -50.967 50.550
818479 101.677 -50.983 50.508
818480 101.703 -50.998 50.465
818481 101.730 -51.021 50.422
818482 101.756 -51.046 50.379
818483 101.783 -51.072 50.338
818484 101.810 -51.098 50.298
818485 101.835 -51.123 50.259
818486 101.842 -51.141 50.221
818487 101.848 -51.159 50.185
818488 101.854 -51.177 50.149
818489 101.860 -51.196 50.116
818490 101.866 -51.214 50.085
818491 101.872 -51.232 50.053
818492 101.899 -51.240 50.023
818493 101.929 -51.245 49.993
818494 101.961 -51.250 49.964
818495 101.974 -51.252 49.935
818496 101.982 -51.251 49.906
818497 101.989 -51.251 49.878
818498 101.997 -51.251 49.852
818499 102.006 -51.252 49.826
818500 102.024 -51.272 49.799
818501 102.042 -51.293 49.768
818502 102.060 -51.313 49.737
818503 102.077 -51.335 49.705
818504 102.092 -51.358 49.675
818505 102.107 -51.381 49.646
818506 102.122 -51.403 49.617
818507 102.137 -51.426 49.587
818508 102.152 -51.449 49.553
818509 102.167 -51.472 49.516
818510 102.182 -51.494 49.482
818511 102.197 -51.517 49.447
818512 102.212 -51.534 49.412
818513 102.228 -51.545 49.377
818514 102.244 -51.556 49.341
818515 102.250 -51.561 49.305
818516 102.254 -51.566 49.272
818517 102.258 -51.569 49.242
818518 102.263 -51.574 49.212
818519 102.267 -51.578 49.180
818520 102.271 -51.582 49.147
818521 102.276 -51.586 49.118
818522 102.280 -51.591 49.091
818523 102.284 -51.594 49.064
818524 102.289 -51.599 49.039
818525 102.293 -51.603 49.016
818526 102.315 -51.616 48.991
818527 102.342 -51.631 48.965
818528 102.370 -51.646 48.938
818529 102.398 -51.660 48.911
818530 102.426 -51.675 48.887
818531 102.453 -51.690 48.861
818532 102.481 -51.704 48.835
818533 102.509 -51.719 48.809
818534 102.536 -51.734 48.781
818535 102.564 -51.748 48.754
818536 102.592 -51.763 48.726
818537 102.619 -51.778 48.696
818538 102.647 -51.792 48.665
818539 102.675 -51.807 48.634
818540 102.703 -51.822 48.603
818541 102.731 -51.836 48.571
818542 102.758 -51.851 48.540
818543 102.786 -51.866 48.509
818544 102.814 -51.880 48.477
818545 102.841 -51.895 48.448
818546 102.869 -51.910 48.419
819177 101.597 -50.913 50.646
819178 101.623 -50.929 50.605
819179 101.650 -50.944 50.562
819180 101.676 -50.960 50.521
819181 101.702 -50.978 50.478
819182 101.729 -51.004 50.435
819183 101.756 -51.030 50.393
819184 101.783 -51.055 50.351
819185 101.810 -51.080 50.312
819186 101.836 -51.106 50.273
819187 101.849 -51.127 50.234
819188 101.855 -51.145 50.198
819189 101.861 -51.163 50.162
819190 101.867 -51.181 50.127
819191 101.873 -51.199 50.096
819192 101.880 -51.217 50.064
819193 101.911 -51.223 50.033
819194 101.942 -51.228 50.003
819195 101.973 -51.233 49.974
819196 101.988 -51.235 49.945
819197 101.996 -51.235 49.916
819198 102.004 -51.234 49.888
819199 102.012 -51.234 49.861
819200 102.023 -51.240 49.835
819201 102.041 -51.261 49.808
819202 102.059 -51.281 49.777
819203 102.077 -51.302 49.746
819204 102.094 -51.323 49.713
819205 102.109 -51.346 49.684
819206 102.124 -51.368 49.656
819207 102.139 -51.392 49.627
819208 102.154 -51.415 49.596
819209 102.169 -51.437 49.561
819210 102.184 -51.460 49.522
819211 102.199 -51.483 49.487
819212 102.214 -51.505 49.453
819213 102.230 -51.516 49.417
819214 102.246 -51.527 49.382
819215 102.262 -51.538 49.345
819216 102.273 -51.547 49.309
819217 102.277 -51.550 49.276
819218 102.282 -51.555 49.245
819219 102.286 -51.559 49.215
819220 102.290 -51.563 49.183
819221 102.295 -51.568 49.152
819222 102.299 -51.572 49.122
819223 102.304 -51.576 49.094
819224 102.308 -51.580 49.067
819225 102.312 -51.584 49.040
819226 102.316 -51.592 49.015
819227 102.330 -51.605 48.989
819228 102.358 -51.619 48.964
819229 102.385 -51.634 48.938
819230 102.413 -51.649 48.912
819231 102.441 -51.663 48.888
819232 102.468 -51.678 48.863
819233 102.496 -51.693 48.837
819234 102.524 -51.707 48.811
819235 102.551 -51.722 48.785
819236 102.579 -51.737 48.758
819237 102.607 -51.751 48.731
819238 102.634 -51.766 48.701
819239 102.662 -51.781 48.671
819240 102.690 -51.795 48.640
819241 102.718 -51.810 48.609
819242 102.745 -51.825 48.577
819243 102.773 -51.839 48.545
819244 102.801 -51.854 48.515
819245 102.829 -51.869 48.483
819246 102.857 -51.883 48.453
819247 102.884 -51.898 48.425
819879 101.623 -50.906 50.616
819880 101.650 -50.921 50.574
819881 101.676 -50.937 50.533
819882 101.702 -50.961 50.490
819883 101.729 -50.986 50.447
819884 101.756 -51.012 50.406
819885 101.782 -51.038 50.365
819886 101.809 -51.063 50.326
819887 101.836 -51.089 50.286
819888 101.856 -51.112 50.247
819889 101.862 -51.130 50.210
819890 101.868 -51.149 50.174
819891 101.875 -51.167 50.140
819892 101.881 -51.185 50.108
819893 101.892 -51.200 50.076
819894 101.923 -51.205 50.045
819895 101.954 -51.211 50.015
819896 101.985 -51.216 49.986
819897 102.003 -51.218 49.957
819898 102.010 -51.218 49.928
819899 102.018 -51.218 49.900
819900 102.026 -51.218 49.873
819901 102.039 -51.229 49.845
819902 102.057 -51.249 49.816
819903 102.076 -51.270 49.787
819904 102.094 -51.290 49.755
819905 102.111 -51.312 49.722
819906 102.126 -51.334 49.691
819907 102.141 -51.357 49.663
819908 102.156 -51.380 49.634
819909 102.171 -51.403 49.603
819910 102.186 -51.425 49.566
819911 102.201 -51.448 49.526
819912 102.216 -51.471 49.489
819913 102.232 -51.487 49.453
819914 102.248 -51.498 49.419
819915 102.263 -51.509 49.383
819916 102.279 -51.520 49.345
819917 102.296 -51.531 49.310
819918 102.301 -51.536 49.278
819919 102.305 -51.541 49.247
819920 102.310 -51.544 49.217
819921 102.314 -51.549 49.187
819922 102.318 -51.553 49.156
819923 102.323 -51.557 49.127
819924 102.327 -51.562 49.097
819925 102.331 -51.566 49.070
819926 102.336 -51.570 49.043
819927 102.339 -51.582 49.017
819928 102.345 -51.593 48.991
819929 102.373 -51.607 48.965
819930 102.400 -51.622 48.940
819931 102.428 -51.637 48.914
819932 102.456 -51.651 48.890
819933 102.483 -51.666 48.866
819934 102.511 -51.681 48.840
819935 102.539 -51.695 48.815
819936 102.567 -51.710 48.788
819937 102.594 -51.725 48.763
819938 102.622 -51.739 48.736
819939 102.650 -51.754 48.707
819940 102.677 -51.769 48.677
819941 102.705 -51.783 48.646
819942 102.733 -51.798 48.615
819943 102.760 -51.813 48.583
819944 102.788 -51.827 48.551
819945 102.816 -51.842 48.521
819946 102.843 -51.857 48.489
819947 102.872 -51.871 48.459
819948 102.899 -51.886 48.432
820581 101.649 -50.898 50.586
820582 101.675 -50.918 50.545
820583 101.702 -50.944 50.501
820584 101.729 -50.970 50.460
820585 101.756 -50.995 50.418
820586 101.782 -51.020 50.378
820587 101.809 -51.046 50.338
820588 101.836 -51.071 50.299
820589 101.863 -51.097 50.260
820590 101.870 -51.115 50.223
820591 101.876 -51.134 50.187
820592 101.882 -51.152 50.152
820593 101.888 -51.171 50.120
820594 101.904 -51.183 50.088
820595 101.935 -51.188 50.057
820596 101.966 -51.194 50.026
820597 101.997 -51.199 49.997
820598 102.017 -51.202 49.969
820599 102.024 -51.202 49.940
820600 102.032 -51.201 49.912
820601 102.040 -51.201 49.885
820602 102.056 -51.218 49.858
820603 102.075 -51.238 49.828
820604 102.093 -51.258 49.797
820605 102.111 -51.279 49.765
820606 102.128 -51.300 49.730
820607 102.144 -51.323 49.697
820608 102.159 -51.346 49.667
820609 102.173 -51.368 49.638
820610 102.188 -51.391 49.606
820611 102.203 -51.414 49.569
820612 102.218 -51.437 49.530
820613 102.233 -51.459 49.492
820614 102.249 -51.469 49.454
820615 102.265 -51.481 49.419
820616 102.281 -51.491 49.383
820617 102.297 -51.503 49.346
820618 102.313 -51.513 49.312
820619 102.324 -51.522 49.281
820620 102.328 -51.526 49.250
820621 102.333 -51.530 49.220
820622 102.337 -51.534 49.191
820623 102.341 -51.538 49.161
820624 102.346 -51.543 49.131
820625 102.350 -51.547 49.101
820626 102.355 -51.551 49.074
820627 102.359 -51.556 49.046
820628 102.362 -51.572 49.020
820629 102.360 -51.581 48.994
820630 102.388 -51.596 48.967
820631 102.415 -51.610 48.942
820632 102.443 -51.625 48.917
820633 102.471 -51.640 48.892
820634 102.499 -51.654 48.869
820635 102.526 -51.669 48.843
820636 102.554 -51.684 48.819
820637 102.582 -51.698 48.792
820638 102.609 -51.713 48.767
820639 102.637 -51.728 48.741
820640 102.665 -51.742 48.712
820641 102.692 -51.757 48.683
820642 102.720 -51.772 48.652
820643 102.748 -51.786 48.620
820644 102.775 -51.801 48.589
820645 102.803 -51.816 48.558
820646 102.831 -51.830 48.527
820647 102.859 -51.845 48.495
820648 102.886 -51.860 48.466
820649 102.914 -51.874 48.440
821283 101.675 -50.901 50.556
821284 101.702 -50.926 50.513
821285 101.729 -50.952 50.472
821286 101.755 -50.978 50.431
821287 101.782 -51.003 50.391
821288 101.809 -51.029 50.351
821289 101.835 -51.055 50.311
821290 101.862 -51.080 50.273
821291 101.877 -51.101 50.236
821292 101.883 -51.119 50.199
821293 101.889 -51.137 50.165
821294 101.895 -51.155 50.133
821295 101.916 -51.166 50.100
821296 101.948 -51.171 50.068
821297 101.978 -51.177 50.037
821298 102.009 -51.182 50.009
821299 102.031 -51.185 49.981
821300 102.038 -51.185 49.953
821301 102.046 -51.185 49.925
821302 102.055 -51.186 49.898
821303 102.073 -51.206 49.872
821304 102.091 -51.227 49.842
821305 102.109 -51.247 49.809
821306 102.127 -51.268 49.775
821307 102.146 -51.288 49.738
821308 102.161 -51.311 49.702
821309 102.176 -51.334 49.669
821310 102.190 -51.357 49.638
821311 102.205 -51.380 49.606
821312 102.221 -51.403 49.571
821313 102.235 -51.425 49.534
821314 102.251 -51.440 49.494
821315 102.267 -51.452 49.455
821316 102.283 -51.462 49.419
821317 102.299 -51.474 49.383
821318 102.315 -51.484 49.347
821319 102.331 -51.496 49.314
821320 102.346 -51.506 49.284
821321 102.352 -51.512 49.254
821322 102.357 -51.515 49.225
821323 102.361 -51.520 49.195
821324 102.365 -51.524 49.166
821325 102.369 -51.528 49.136
821326 102.374 -51.532 49.106
821327 102.378 -51.537 49.077
821328 102.382 -51.544 49.050
821329 102.385 -51.562 49.024
821330 102.375 -51.569 48.997
821331 102.403 -51.584 48.970
821332 102.430 -51.598 48.944
821333 102.458 -51.613 48.920
821334 102.486 -51.628 48.895
821335 102.514 -51.642 48.871
821336 102.541 -51.657 48.847
821337 102.569 -51.672 48.822
821338 102.597 -51.686 48.796
821339 102.624 -51.701 48.772
821340 102.652 -51.716 48.745
821341 102.680 -51.730 48.717
821342 102.708 -51.745 48.688
821343 102.735 -51.760 48.657
821344 102.763 -51.774 48.627
821345 102.790 -51.789 48.595
821346 102.818 -51.804 48.564
821347 102.846 -51.818 48.534
821348 102.874 -51.833 48.503
821349 102.901 -51.848 48.475
821350 102.929 -51.862 48.448
821985 101.701 -50.910 50.527
821986 101.728 -50.935 50.486
821987 101.755 -50.961 50.444
821988 101.782 -50.986 50.405
821989 101.809 -51.011 50.365
821990 101.835 -51.037 50.325
821991 101.862 -51.063 50.286
821992 101.884 -51.086 50.249
821993 101.890 -51.105 50.212
821994 101.896 -51.123 50.179
821995 101.902 -51.141 50.146
821996 101.928 -51.149 50.112
821997 101.960 -51.154 50.082
821998 101.990 -51.159 50.051
821999 102.021 -51.165 50.022
822000 102.045 -51.168 49.994
822001 102.053 -51.168 49.966
822002 102.060 -51.168 49.939
822003 102.071 -51.174 49.913
822004 102.090 -51.195 49.887
822005 102.108 -51.215 49.857
822006 102.126 -51.236 49.822
822007 102.144 -51.256 49.785
822008 102.163 -51.277 49.745
822009 102.178 -51.299 49.707
822010 102.193 -51.322 49.671
822011 102.207 -51.345 49.637
822012 102.223 -51.368 49.604
822013 102.238 -51.390 49.570
822014 102.253 -51.412 49.534
822015 102.268 -51.423 49.497
822016 102.284 -51.434 49.460
822017 102.301 -51.445 49.421
822018 102.316 -51.456 49.385
822019 102.332 -51.467 49.350
822020 102.348 -51.478 49.317
822021 102.364 -51.489 49.287
822022 102.375 -51.497 49.258
822023 102.380 -51.501 49.229
822024 102.384 -51.506 49.201
822025 102.388 -51.509 49.172
822026 102.393 -51.514 49.142
822027 102.397 -51.518 49.111
822028 102.401 -51.522 49.083
822029 102.405 -51.534 49.054
822030 102.407 -51.553 49.027
822031 102.392 -51.558 49.001
822032 102.418 -51.572 48.974
822033 102.446 -51.586 48.948
822034 102.473 -51.601 48.923
822035 102.501 -51.616 48.898
822036 102.529 -51.630 48.874
822037 102.556 -51.645 48.850
822038 102.584 -51.660 48.825
822039 102.612 -51.674 48.800
822040 102.640 -51.689 48.776
822041 102.667 -51.703 48.750
822042 102.695 -51.719 48.722
822043 102.723 -51.733 48.693
822044 102.750 -51.747 48.663
822045 102.778 -51.763 48.632
822046 102.806 -51.777 48.601
822047 102.833 -51.791 48.571
822048 102.861 -51.807 48.541
822049 102.889 -51.821 48.511
822050 102.916 -51.835 48.483
822051 102.944 -51.851 48.457
822052 102.972 -51.865 48.432
822685 101.675 -50.867 50.586
822686 101.701 -50.892 50.542
822687 101.728 -50.918 50.501
822688 101.755 -50.943 50.460
822689 101.781 -50.969 50.420
822690 101.808 -50.995 50.381
822691 101.835 -51.020 50.341
822692 101.862 -51.045 50.301
822693 101.888 -51.071 50.264
822694 101.897 -51.090 50.228
822695 101.903 -51.108 50.194
822696 101.910 -51.127 50.160
822697 101.941 -51.132 50.127
822698 101.972 -51.137 50.096
822699 102.003 -51.142 50.065
822700 102.033 -51.148 50.036
822701 102.059 -51.152 50.007
822702 102.067 -51.152 49.979
822703 102.075 -51.152 49.952
822704 102.089 -51.163 49.926
822705 102.107 -51.183 49.900
822706 102.125 -51.204 49.870
822707 102.143 -51.224 49.835
822708 102.161 -51.245 49.796
822709 102.179 -51.265 49.754
822710 102.195 -51.288 49.713
822711 102.210 -51.311 49.674
822712 102.225 -51.334 49.637
822713 102.240 -51.356 49.602
822714 102.255 -51.379 49.567
822715 102.270 -51.393 49.533
822716 102.286 -51.405 49.498
822717 102.302 -51.416 49.462
822718 102.318 -51.427 49.425
822719 102.334 -51.438 49.389
822720 102.350 -51.449 49.354
822721 102.366 -51.460 49.321
822722 102.382 -51.471 49.291
822723 102.398 -51.482 49.262
822724 102.403 -51.487 49.235
822725 102.407 -51.491 49.207
822726 102.412 -51.495 49.178
822727 102.416 -51.500 49.148
822728 102.420 -51.503 49.117
822729 102.425 -51.508 49.088
822730 102.427 -51.524 49.059
822731 102.430 -51.543 49.031
822732 102.416 -51.549 49.004
822733 102.433 -51.559 48.977
822734 102.461 -51.575 48.951
822735 102.488 -51.589 48.926
822736 102.516 -51.603 48.902
822737 102.544 -51.619 48.878
822738 102.572 -51.633 48.854
822739 102.599 -51.647 48.829
822740 102.627 -51.663 48.804
822741 102.655 -51.677 48.780
822742 102.682 -51.691 48.755
822743 102.710 -51.707 48.727
822744 102.738 -51.721 48.698
822745 102.765 -51.735 48.668
822746 102.793 -51.750 48.637
822747 102.821 -51.765 48.607
822748 102.848 -51.780 48.577
822749 102.876 -51.794 48.548
822750 102.904 -51.809 48.519
822751 102.932 -51.824 48.490
822752 102.959 -51.838 48.465
822753 102.987 -51.853 48.441
823386 101.674 -50.850 50.602
823387 101.701 -50.875 50.559
823388 101.728 -50.901 50.518
823389 101.755 -50.926 50.476
823390 101.781 -50.952 50.436
823391 101.808 -50.977 50.397
823392 101.835 -51.003 50.358
823393 101.862 -51.029 50.318
823394 101.888 -51.054 50.279
823395 101.904 -51.076 50.245
823396 101.910 -51.094 50.210
823397 101.922 -51.109 50.176
823398 101.953 -51.114 50.143
823399 101.984 -51.120 50.111
823400 102.015 -51.125 50.080
823401 102.045 -51.130 50.051
823402 102.074 -51.136 50.022
823403 102.081 -51.135 49.993
823404 102.089 -51.135 49.965
823405 102.105 -51.152 49.939
823406 102.123 -51.172 49.913
823407 102.141 -51.193 49.883
823408 102.160 -51.213 49.848
823409 102.178 -51.233 49.807
823410 102.196 -51.254 49.764
823411 102.212 -51.276 49.720
823412 102.227 -51.299 49.678
823413 102.242 -51.322 49.638
823414 102.257 -51.345 49.600
823415 102.272 -51.365 49.564
823416 102.288 -51.376 49.531
823417 102.304 -51.387 49.497
823418 102.320 -51.398 49.463
823419 102.336 -51.409 49.429
823420 102.351 -51.420 49.394
823421 102.368 -51.431 49.359
823422 102.384 -51.442 49.325
823423 102.399 -51.453 49.295
823424 102.415 -51.464 49.267
823425 102.427 -51.472 49.240
823426 102.430 -51.476 49.212
823427 102.435 -51.481 49.184
823428 102.440 -51.485 49.154
823429 102.444 -51.489 49.123
823430 102.448 -51.494 49.094
823431 102.451 -51.514 49.065
823432 102.453 -51.533 49.036
823433 102.441 -51.541 49.009
823434 102.448 -51.548 48.981
823435 102.476 -51.562 48.955
823436 102.504 -51.577 48.929
823437 102.531 -51.592 48.905
823438 102.559 -51.606 48.881
823439 102.587 -51.621 48.857
823440 102.614 -51.636 48.832
823441 102.642 -51.650 48.808
823442 102.670 -51.665 48.783
823443 102.697 -51.680 48.757
823444 102.725 -51.694 48.730
823445 102.753 -51.709 48.702
823446 102.781 -51.724 48.672
823447 102.808 -51.738 48.642
823448 102.836 -51.753 48.612
823449 102.864 -51.768 48.583
823450 102.891 -51.782 48.554
823451 102.919 -51.797 48.525
823452 102.947 -51.812 48.497
823453 102.974 -51.826 48.474
823454 103.002 -51.841 48.450
824086 101.647 -50.807 50.661
824087 101.674 -50.832 50.620
824088 101.701 -50.858 50.578
824089 101.728 -50.884 50.536
824090 101.754 -50.909 50.495
824091 101.781 -50.935 50.454
824092 101.808 -50.961 50.415
824093 101.834 -50.986 50.376
824094 101.861 -51.011 50.336
824095 101.888 -51.037 50.298
824096 101.911 -51.061 50.262
824097 101.917 -51.080 50.227
824098 101.934 -51.092 50.193
824099 101.965 -51.097 50.160
824100 101.996 -51.103 50.127
824101 102.027 -51.108 50.096
824102 102.058 -51.114 50.066
824103 102.088 -51.119 50.037
824104 102.095 -51.118 50.008
824105 102.104 -51.120 49.980
824106 102.122 -51.140 49.953
824107 102.140 -51.161 49.926
824108 102.159 -51.181 49.895
824109 102.177 -51.202 49.859
824110 102.195 -51.222 49.818
824111 102.213 -51.243 49.774
824112 102.229 -51.265 49.729
824113 102.244 -51.287 49.684
824114 102.259 -51.310 49.642
824115 102.274 -51.333 49.603
824116 102.290 -51.347 49.568
824117 102.306 -51.358 49.534
824118 102.321 -51.369 49.500
824119 102.337 -51.380 49.467
824120 102.353 -51.391 49.432
824121 102.369 -51.403 49.398
824122 102.385 -51.413 49.363
824123 102.401 -51.425 49.331
824124 102.417 -51.435 49.301
824125 102.433 -51.446 49.273
824126 102.449 -51.457 49.245
824127 102.454 -51.462 49.218
824128 102.458 -51.466 49.190
824129 102.463 -51.470 49.160
824130 102.467 -51.475 49.129
824131 102.471 -51.484 49.099
824132 102.473 -51.504 49.070
824133 102.476 -51.524 49.042
824134 102.465 -51.532 49.014
824135 102.463 -51.536 48.986
824136 102.491 -51.550 48.959
824137 102.519 -51.566 48.934
824138 102.546 -51.580 48.909
824139 102.574 -51.594 48.885
824140 102.602 -51.609 48.861
824141 102.629 -51.624 48.836
824142 102.657 -51.638 48.811
824143 102.685 -51.653 48.786
824144 102.713 -51.668 48.760
824145 102.740 -51.682 48.733
824146 102.768 -51.697 48.705
824147 102.796 -51.712 48.677
824148 102.823 -51.726 48.647
824149 102.851 -51.741 48.618
824150 102.879 -51.756 48.589
824151 102.906 -51.770 48.561
824152 102.934 -51.785 48.532
824153 102.962 -51.800 48.505
824154 102.989 -51.814 48.482
824155 103.017 -51.829 48.459
824788 101.674 -50.816 50.638
824789 101.700 -50.841 50.597
824790 101.727 -50.867 50.554
824791 101.754 -50.892 50.513
824792 101.781 -50.917 50.472
824793 101.808 -50.943 50.433
824794 101.834 -50.969 50.394
824795 101.861 -50.994 50.354
824796 101.888 -51.020 50.316
824797 101.914 -51.045 50.279
824798 101.924 -51.065 50.245
824799 101.947 -51.075 50.211
824800 101.977 -51.080 50.176
824801 102.008 -51.086 50.143
824802 102.039 -51.091 50.111
824803 102.070 -51.096 50.081
824804 102.101 -51.102 50.051
824805 102.109 -51.102 50.022
824806 102.121 -51.108 49.993
824807 102.139 -51.129 49.966
824808 102.157 -51.149 49.938
824809 102.175 -51.170 49.907
824810 102.193 -51.190 49.871
824811 102.211 -51.211 49.830
824812 102.230 -51.231 49.786
824813 102.246 -51.253 49.738
824814 102.261 -51.276 49.692
824815 102.276 -51.299 49.648
824816 102.291 -51.318 49.609
824817 102.307 -51.329 49.573
824818 102.323 -51.340 49.538
824819 102.339 -51.351 49.505
824820 102.355 -51.362 49.471
824821 102.371 -51.373 49.436
824822 102.387 -51.384 49.402
824823 102.403 -51.396 49.369
824824 102.419 -51.406 49.338
824825 102.435 -51.418 49.308
824826 102.451 -51.428 49.280
824827 102.467 -51.440 49.252
824828 102.477 -51.447 49.224
824829 102.482 -51.452 49.196
824830 102.486 -51.456 49.166
824831 102.490 -51.460 49.135
824832 102.493 -51.475 49.105
824833 102.496 -51.494 49.076
824834 102.499 -51.514 49.047
824835 102.489 -51.524 49.019
824836 102.478 -51.524 48.991
824837 102.506 -51.539 48.964
824838 102.534 -51.553 48.938
824839 102.561 -51.568 48.913
824840 102.589 -51.583 48.889
824841 102.617 -51.597 48.865
824842 102.644 -51.612 48.840
824843 102.672 -51.627 48.814
824844 102.700 -51.641 48.789
824845 102.728 -51.656 48.764
824846 102.755 -51.671 48.737
824847 102.783 -51.685 48.709
824848 102.811 -51.700 48.681
824849 102.838 -51.715 48.652
824850 102.866 -51.729 48.623
824851 102.894 -51.744 48.595
824852 102.921 -51.759 48.567
824853 102.949 -51.773 48.539
824854 102.977 -51.788 48.513
824855 103.005 -51.803 48.490
825489 101.674 -50.798 50.656
825490 101.700 -50.824 50.615
825491 101.727 -50.849 50.572
825492 101.754 -50.875 50.530
825493 101.780 -50.901 50.490
825494 101.807 -50.926 50.450
825495 101.834 -50.951 50.411
825496 101.861 -50.977 50.373
825497 101.887 -51.002 50.334
825498 101.914 -51.028 50.297
825499 101.932 -51.050 50.262
825500 101.959 -51.058 50.228
825501 101.989 -51.063 50.192
825502 102.020 -51.068 50.159
825503 102.051 -51.074 50.126
825504 102.082 -51.080 50.095
825505 102.113 -51.085 50.065
825506 102.123 -51.085 50.035
825507 102.137 -51.097 50.006
825508 102.156 -51.117 49.978
825509 102.174 -51.138 49.950
825510 102.192 -51.158 49.919
825511 102.210 -51.179 49.882
825512 102.229 -51.199 49.842
825513 102.247 -51.220 49.797
825514 102.263 -51.241 49.748
825515 102.278 -51.264 49.700
825516 102.293 -51.287 49.656
825517 102.309 -51.300 49.616
825518 102.325 -51.312 49.579
825519 102.341 -51.322 49.544
825520 102.357 -51.334 49.511
825521 102.373 -51.344 49.475
825522 102.389 -51.356 49.440
825523 102.405 -51.366 49.408
825524 102.420 -51.378 49.376
825525 102.437 -51.389 49.345
825526 102.453 -51.399 49.316
825527 102.468 -51.411 49.288
825528 102.484 -51.421 49.260
825529 102.501 -51.433 49.232
825530 102.505 -51.437 49.202
825531 102.510 -51.441 49.173
825532 102.514 -51.446 49.142
825533 102.516 -51.465 49.113
825534 102.519 -51.484 49.083
825535 102.522 -51.504 49.053
825536 102.514 -51.515 49.025
825537 102.493 -51.512 48.997
825538 102.521 -51.527 48.970
825539 102.548 -51.541 48.943
825540 102.576 -51.556 48.918
825541 102.604 -51.571 48.893
825542 102.632 -51.585 48.869
825543 102.660 -51.600 48.844
825544 102.687 -51.615 48.818
825545 102.715 -51.629 48.794
825546 102.743 -51.644 48.767
825547 102.770 -51.659 48.740
825548 102.798 -51.673 48.713
825549 102.826 -51.688 48.685
825550 102.854 -51.703 48.657
825551 102.881 -51.717 48.629
825552 102.909 -51.732 48.601
825553 102.937 -51.747 48.574
825554 102.964 -51.761 48.547
825555 102.992 -51.776 48.522
826190 101.673 -50.781 50.675
826191 101.700 -50.807 50.633
826192 101.727 -50.832 50.590
826193 101.753 -50.857 50.547
826194 101.780 -50.883 50.508
826195 101.807 -50.909 50.468
826196 101.834 -50.934 50.429
826197 101.861 -50.960 50.391
826198 101.887 -50.986 50.352
826199 101.914 -51.011 50.315
826200 101.940 -51.036 50.279
826201 101.971 -51.041 50.246
826202 102.001 -51.046 50.208
826203 102.032 -51.052 50.174
826204 102.063 -51.057 50.141
826205 102.094 -51.062 50.110
826206 102.125 -51.067 50.079
826207 102.138 -51.069 50.048
826208 102.154 -51.086 50.018
826209 102.172 -51.106 49.989
826210 102.191 -51.127 49.962
826211 102.209 -51.147 49.930
826212 102.227 -51.168 49.894
826213 102.245 -51.188 49.853
826214 102.263 -51.208 49.808
826215 102.280 -51.230 49.759
826216 102.296 -51.252 49.711
826217 102.311 -51.271 49.665
826218 102.327 -51.283 49.623
826219 102.342 -51.293 49.585
826220 102.359 -51.305 49.550
826221 102.375 -51.315 49.517
826222 102.390 -51.327 49.482
826223 102.406 -51.337 49.446
826224 102.422 -51.349 49.414
826225 102.438 -51.359 49.382
826226 102.454 -51.371 49.353
826227 102.470 -51.382 49.324
826228 102.486 -51.393 49.296
826229 102.502 -51.404 49.269
826230 102.518 -51.415 49.241
826231 102.528 -51.423 49.211
826232 102.533 -51.427 49.181
826233 102.537 -51.436 49.151
826234 102.539 -51.456 49.120
826235 102.542 -51.475 49.091
826236 102.544 -51.494 49.061
826237 102.538 -51.507 49.032
826238 102.508 -51.500 49.004
826239 102.536 -51.515 48.976
826240 102.563 -51.530 48.950
826241 102.591 -51.544 48.924
826242 102.619 -51.559 48.899
826243 102.647 -51.574 48.874
826244 102.674 -51.588 48.848
826245 102.702 -51.603 48.823
826246 102.730 -51.618 48.798
826247 102.758 -51.632 48.772
826248 102.786 -51.647 48.745
826249 102.813 -51.662 48.719
826250 102.841 -51.676 48.690
826251 102.869 -51.691 48.662
826252 102.896 -51.706 48.635
826253 102.924 -51.720 48.608
826254 102.952 -51.735 48.581
826255 102.979 -51.750 48.555
826256 103.007 -51.764 48.531
826891 101.673 -50.764 50.694
826892 101.700 -50.789 50.652
826893 101.727 -50.815 50.609
826894 101.753 -50.841 50.567
826895 101.780 -50.866 50.527
826896 101.807 -50.892 50.487
826897 101.833 -50.917 50.448
826898 101.860 -50.942 50.409
826899 101.887 -50.968 50.370
826900 101.914 -50.994 50.333
826901 101.942 -51.014 50.298
826902 101.973 -51.028 50.263
826903 102.012 -51.030 50.226
826904 102.045 -51.034 50.191
826905 102.076 -51.039 50.157
826906 102.106 -51.045 50.126
826907 102.137 -51.051 50.095
826908 102.153 -51.054 50.062
826909 102.171 -51.074 50.031
826910 102.189 -51.095 50.000
826911 102.207 -51.115 49.972
826912 102.226 -51.136 49.940
826913 102.244 -51.156 49.904
826914 102.262 -51.177 49.864
826915 102.280 -51.197 49.819
826916 102.298 -51.218 49.770
826917 102.313 -51.241 49.721
826918 102.328 -51.254 49.675
826919 102.344 -51.265 49.632
826920 102.360 -51.276 49.593
826921 102.376 -51.287 49.556
826922 102.392 -51.298 49.522
826923 102.408 -51.309 49.488
826924 102.424 -51.320 49.454
826925 102.440 -51.331 49.422
826926 102.456 -51.342 49.391
826927 102.472 -51.352 49.363
826928 102.488 -51.364 49.335
826929 102.504 -51.375 49.307
826930 102.520 -51.386 49.279
826931 102.536 -51.397 49.251
826932 102.551 -51.408 49.221
826933 102.556 -51.412 49.191
826934 102.559 -51.426 49.160
826935 102.562 -51.446 49.129
826936 102.564 -51.465 49.099
826937 102.567 -51.484 49.069
826938 102.563 -51.499 49.040
826939 102.533 -51.492 49.011
826940 102.551 -51.503 48.983
826941 102.579 -51.518 48.957
826942 102.606 -51.532 48.931
826943 102.634 -51.547 48.905
826944 102.662 -51.562 48.880
826945 102.689 -51.576 48.854
826946 102.717 -51.591 48.828
826947 102.745 -51.606 48.803
826948 102.772 -51.620 48.777
826949 102.800 -51.635 48.751
826950 102.828 -51.650 48.724
826951 102.856 -51.664 48.696
826952 102.884 -51.679 48.668
826953 102.911 -51.694 48.641
826954 102.939 -51.708 48.615
826955 102.967 -51.723 48.589
826956 102.995 -51.738 48.563
827592 101.673 -50.747 50.713
827593 101.699 -50.773 50.671
827594 101.726 -50.798 50.628
827595 101.753 -50.823 50.587
827596 101.780 -50.849 50.546
827597 101.807 -50.874 50.506
827598 101.833 -50.900 50.466
827599 101.860 -50.926 50.427
827600 101.887 -50.951 50.388
827601 101.914 -50.975 50.350
827602 101.944 -50.992 50.315
827603 101.975 -51.006 50.280
827604 102.006 -51.020 50.244
827605 102.043 -51.025 50.208
827606 102.083 -51.026 50.174
827607 102.118 -51.028 50.143
827608 102.149 -51.033 50.111
827609 102.170 -51.042 50.077
827610 102.188 -51.063 50.044
827611 102.206 -51.083 50.012
827612 102.224 -51.104 49.982
827613 102.242 -51.124 49.950
827614 102.261 -51.145 49.913
827615 102.279 -51.165 49.875
827616 102.297 -51.186 49.830
827617 102.315 -51.206 49.781
827618 102.330 -51.225 49.732
827619 102.346 -51.236 49.685
827620 102.362 -51.247 49.641
827621 102.378 -51.258 49.601
827622 102.394 -51.269 49.564
827623 102.410 -51.280 49.529
827624 102.426 -51.291 49.495
827625 102.442 -51.302 49.462
827626 102.458 -51.313 49.431
827627 102.473 -51.324 49.401
827628 102.489 -51.335 49.374
827629 102.506 -51.346 49.347
827630 102.521 -51.357 49.318
827631 102.537 -51.368 49.290
827632 102.553 -51.379 49.262
827633 102.569 -51.390 49.232
827634 102.580 -51.398 49.201
827635 102.582 -51.416 49.169
827636 102.585 -51.436 49.138
827637 102.587 -51.456 49.108
827638 102.590 -51.475 49.079
827639 102.587 -51.490 49.050
827640 102.557 -51.484 49.021
827641 102.566 -51.491 48.992
827642 102.594 -51.506 48.964
827643 102.621 -51.520 48.938
827644 102.649 -51.535 48.912
827645 102.677 -51.550 48.887
827646 102.704 -51.565 48.861
827647 102.732 -51.579 48.835
827648 102.760 -51.594 48.810
827649 102.787 -51.609 48.784
827650 102.815 -51.623 48.758
827651 102.843 -51.638 48.731
827652 102.871 -51.653 48.703
827653 102.898 -51.667 48.675
827654 102.927 -51.682 48.649
827655 102.954 -51.697 48.623
827656 102.982 -51.711 48.597
828293 101.673 -50.729 50.733
828294 101.699 -50.755 50.691
828295 101.726 -50.781 50.648
828296 101.753 -50.806 50.607
828297 101.779 -50.832 50.566
828298 101.806 -50.857 50.526
828299 101.833 -50.882 50.485
828300 101.859 -50.908 50.445
828301 101.886 -50.934 50.406
828302 101.914 -50.954 50.368
828303 101.945 -50.970 50.332
828304 101.976 -50.984 50.296
828305 102.007 -50.998 50.258
828306 102.038 -51.012 50.223
828307 102.074 -51.020 50.190
828308 102.113 -51.020 50.158
828309 102.153 -51.021 50.126
828310 102.185 -51.029 50.092
828311 102.204 -51.050 50.058
828312 102.223 -51.071 50.026
828313 102.241 -51.092 49.994
828314 102.259 -51.113 49.959
828315 102.277 -51.133 49.922
828316 102.296 -51.154 49.884
828317 102.314 -51.174 49.840
828318 102.332 -51.195 49.791
828319 102.347 -51.207 49.742
828320 102.364 -51.218 49.695
828321 102.380 -51.229 49.651
828322 102.395 -51.240 49.610
828323 102.411 -51.251 49.572
828324 102.427 -51.262 49.538
828325 102.443 -51.273 49.504
828326 102.459 -51.284 49.472
828327 102.475 -51.295 49.442
828328 102.490 -51.305 49.415
828329 102.506 -51.317 49.388
828330 102.522 -51.327 49.361
828331 102.538 -51.338 49.332
828332 102.553 -51.349 49.304
828333 102.569 -51.360 49.276
828334 102.585 -51.371 49.246
828335 102.602 -51.387 49.214
828336 102.605 -51.407 49.181
828337 102.608 -51.426 49.148
828338 102.610 -51.446 49.120
828339 102.613 -51.465 49.090
828340 102.612 -51.481 49.060
828341 102.582 -51.475 49.032
828342 102.581 -51.479 49.003
828343 102.609 -51.494 48.975
828344 102.636 -51.509 48.948
828345 102.664 -51.523 48.922
828346 102.692 -51.537 48.895
828347 102.720 -51.553 48.869
828348 102.747 -51.567 48.843
828349 102.775 -51.581 48.818
828350 102.803 -51.597 48.792
828351 102.830 -51.611 48.765
828352 102.858 -51.626 48.737
828353 102.886 -51.641 48.709
828354 102.913 -51.655 48.682
828355 102.941 -51.670 48.656
828356 102.969 -51.685 48.631
828994 101.672 -50.713 50.755
828995 101.699 -50.738 50.711
828996 101.726 -50.763 50.669
828997 101.752 -50.789 50.627
828998 101.779 -50.814 50.586
828999 101.806 -50.840 50.545
829000 101.833 -50.866 50.504
829001 101.859 -50.891 50.463
829002 101.887 -50.914 50.424
829003 101.915 -50.933 50.385
829004 101.946 -50.948 50.348
829005 101.977 -50.961 50.311
829006 102.009 -50.976 50.273
829007 102.040 -50.990 50.238
829008 102.071 -51.005 50.204
829009 102.105 -51.014 50.171
829010 102.144 -51.015 50.139
829011 102.184 -51.016 50.105
829012 102.215 -51.025 50.073
829013 102.233 -51.046 50.043
829014 102.251 -51.067 50.010
829015 102.270 -51.089 49.973
829016 102.289 -51.110 49.933
829017 102.307 -51.131 49.894
829018 102.326 -51.152 49.848
829019 102.344 -51.174 49.801
829020 102.360 -51.185 49.753
829021 102.376 -51.196 49.706
829022 102.391 -51.207 49.661
829023 102.407 -51.218 49.620
829024 102.423 -51.228 49.582
829025 102.439 -51.240 49.548
829026 102.454 -51.250 49.514
829027 102.470 -51.261 49.483
829028 102.486 -51.271 49.455
829029 102.501 -51.283 49.428
829030 102.517 -51.293 49.403
829031 102.533 -51.304 49.375
829032 102.548 -51.315 49.347
829033 102.564 -51.326 49.320
829034 102.580 -51.337 49.292
829035 102.599 -51.357 49.261
829036 102.618 -51.378 49.228
829037 102.628 -51.397 49.194
829038 102.630 -51.416 49.161
829039 102.633 -51.436 49.132
829040 102.635 -51.455 49.102
829041 102.636 -51.473 49.072
829042 102.606 -51.466 49.043
829043 102.596 -51.467 49.014
829044 102.624 -51.482 48.986
829045 102.652 -51.497 48.959
829046 102.679 -51.511 48.933
829047 102.707 -51.526 48.906
829048 102.735 -51.541 48.879
829049 102.762 -51.555 48.852
829050 102.790 -51.570 48.826
829051 102.818 -51.584 48.801
829052 102.845 -51.599 48.774
829053 102.873 -51.614 48.745
829054 102.901 -51.628 48.717
829055 102.929 -51.643 48.690
829056 102.956 -51.658 48.664
829696 101.699 -50.721 50.731
829697 101.726 -50.746 50.689
829698 101.752 -50.772 50.646
829699 101.779 -50.798 50.604
829700 101.806 -50.823 50.563
829701 101.832 -50.848 50.522
829702 101.859 -50.873 50.481
829703 101.887 -50.893 50.442
829704 101.916 -50.911 50.402
829705 101.948 -50.926 50.363
829706 101.979 -50.939 50.326
829707 102.010 -50.954 50.287
829708 102.041 -50.968 50.252
829709 102.073 -50.982 50.217
829710 102.104 -50.996 50.183
829711 102.135 -51.009 50.149
829712 102.175 -51.010 50.116
829713 102.215 -51.011 50.087
829714 102.243 -51.022 50.060
829715 102.262 -51.043 50.027
829716 102.280 -51.064 49.988
829717 102.299 -51.086 49.946
829718 102.318 -51.107 49.904
829719 102.336 -51.128 49.857
829720 102.355 -51.149 49.810
829721 102.371 -51.162 49.762
829722 102.387 -51.173 49.717
829723 102.402 -51.183 49.672
829724 102.418 -51.195 49.631
829725 102.434 -51.205 49.593
829726 102.450 -51.216 49.558
829727 102.465 -51.227 49.525
829728 102.481 -51.238 49.494
829729 102.497 -51.249 49.468
829730 102.512 -51.259 49.442
829731 102.528 -51.271 49.418
829732 102.544 -51.281 49.391
829733 102.559 -51.292 49.364
829734 102.577 -51.307 49.338
829735 102.596 -51.327 49.308
829736 102.615 -51.347 49.276
829737 102.634 -51.367 49.242
829738 102.650 -51.387 49.207
829739 102.653 -51.406 49.175
829740 102.656 -51.426 49.144
829741 102.658 -51.446 49.115
829742 102.661 -51.465 49.084
829743 102.630 -51.458 49.056
829744 102.611 -51.456 49.026
829745 102.639 -51.470 48.998
829746 102.667 -51.484 48.971
829747 102.694 -51.500 48.944
829748 102.722 -51.514 48.917
829749 102.750 -51.528 48.890
829750 102.777 -51.544 48.863
829751 102.805 -51.558 48.837
829752 102.833 -51.572 48.811
829753 102.861 -51.587 48.783
829754 102.888 -51.602 48.755
829755 102.916 -51.616 48.727
829756 102.944 -51.631 48.700
829757 102.971 -51.646 48.673
830397 101.698 -50.704 50.748
830398 101.725 -50.729 50.706
830399 101.752 -50.754 50.664
830400 101.779 -50.780 50.621
830401 101.806 -50.806 50.581
830402 101.832 -50.831 50.539
830403 101.860 -50.852 50.497
830404 101.888 -50.873 50.457
830405 101.918 -50.889 50.416
830406 101.949 -50.903 50.377
830407 101.980 -50.917 50.339
830408 102.012 -50.932 50.299
830409 102.043 -50.945 50.264
830410 102.074 -50.960 50.228
830411 102.105 -50.974 50.193
830412 102.136 -50.989 50.157
830413 102.167 -51.002 50.126
830414 102.206 -51.005 50.100
830415 102.245 -51.006 50.073
830416 102.272 -51.018 50.040
830417 102.291 -51.039 50.003
830418 102.310 -51.061 49.960
830419 102.328 -51.082 49.915
830420 102.346 -51.103 49.866
830421 102.365 -51.124 49.819
830422 102.382 -51.139 49.771
830423 102.398 -51.150 49.725
830424 102.413 -51.161 49.682
830425 102.429 -51.171 49.641
830426 102.445 -51.182 49.603
830427 102.460 -51.193 49.568
830428 102.476 -51.204 49.536
830429 102.492 -51.215 49.506
830430 102.507 -51.226 49.480
830431 102.523 -51.236 49.458
830432 102.539 -51.247 49.433
830433 102.554 -51.258 49.407
830434 102.573 -51.277 49.382
830435 102.592 -51.297 49.355
830436 102.611 -51.317 49.323
830437 102.630 -51.337 49.290
830438 102.650 -51.357 49.255
830439 102.668 -51.378 49.221
830440 102.676 -51.397 49.189
830441 102.678 -51.416 49.158
830442 102.681 -51.436 49.128
830443 102.684 -51.455 49.098
830444 102.655 -51.450 49.069
830445 102.626 -51.443 49.040
830446 102.654 -51.458 49.012
830447 102.682 -51.473 48.985
830448 102.710 -51.487 48.958
830449 102.737 -51.502 48.930
830450 102.765 -51.517 48.903
830451 102.793 -51.531 48.876
830452 102.820 -51.546 48.849
830453 102.848 -51.561 48.822
830454 102.876 -51.575 48.793
830455 102.903 -51.590 48.766
830456 102.931 -51.605 48.738
830457 102.959 -51.619 48.710
830458 102.986 -51.634 48.684
830459 103.014 -51.649 48.658
831099 101.725 -50.712 50.721
831100 101.752 -50.738 50.679
831101 101.778 -50.763 50.637
831102 101.805 -50.788 50.596
831103 101.832 -50.812 50.553
831104 101.861 -50.832 50.511
831105 101.888 -50.852 50.469
831106 101.919 -50.867 50.429
831107 101.950 -50.881 50.388
831108 101.981 -50.895 50.350
831109 102.013 -50.910 50.309
831110 102.044 -50.923 50.273
831111 102.075 -50.938 50.236
831112 102.106 -50.952 50.199
831113 102.137 -50.966 50.163
831114 102.169 -50.980 50.131
831115 102.200 -50.995 50.104
831116 102.237 -51.000 50.077
831117 102.276 -51.001 50.045
831118 102.302 -51.015 50.009
831119 102.320 -51.036 49.967
831120 102.338 -51.058 49.922
831121 102.357 -51.079 49.874
831122 102.376 -51.100 49.827
831123 102.393 -51.116 49.780
831124 102.408 -51.127 49.735
831125 102.424 -51.137 49.692
831126 102.440 -51.149 49.651
831127 102.456 -51.159 49.613
831128 102.471 -51.170 49.579
831129 102.487 -51.181 49.546
831130 102.503 -51.192 49.517
831131 102.518 -51.202 49.492
831132 102.534 -51.213 49.471
831133 102.550 -51.227 49.446
831134 102.570 -51.246 49.422
831135 102.589 -51.267 49.397
831136 102.608 -51.287 49.370
831137 102.627 -51.307 49.338
831138 102.646 -51.327 49.304
831139 102.665 -51.347 49.269
831140 102.684 -51.367 49.235
831141 102.699 -51.387 49.203
831142 102.701 -51.406 49.173
831143 102.704 -51.426 49.142
831144 102.706 -51.445 49.112
831145 102.679 -51.441 49.083
831146 102.649 -51.434 49.054
831147 102.669 -51.446 49.028
831148 102.697 -51.461 49.000
831149 102.725 -51.475 48.972
831150 102.752 -51.490 48.944
831151 102.780 -51.505 48.917
831152 102.808 -51.519 48.889
831153 102.835 -51.534 48.862
831154 102.863 -51.549 48.834
831155 102.891 -51.563 48.805
831156 102.918 -51.578 48.777
831157 102.946 -51.593 48.749
831158 102.974 -51.607 48.722
831159 103.001 -51.622 48.695
831160 103.029 -51.637 48.669
831161 103.057 -51.651 48.644
831801 101.751 -50.720 50.694
831802 101.778 -50.746 50.652
831803 101.805 -50.771 50.610
831804 101.833 -50.791 50.567
831805 101.861 -50.811 50.524
831806 101.889 -50.830 50.481
831807 101.920 -50.845 50.440
831808 101.952 -50.859 50.399
831809 101.983 -50.873 50.360
831810 102.014 -50.887 50.319
831811 102.045 -50.901 50.280
831812 102.077 -50.916 50.242
831813 102.108 -50.929 50.204
831814 102.139 -50.944 50.167
831815 102.170 -50.958 50.134
831816 102.201 -50.973 50.105
831817 102.233 -50.986 50.077
831818 102.267 -50.995 50.046
831819 102.307 -50.995 50.010
831820 102.330 -51.012 49.970
831821 102.349 -51.033 49.927
831822 102.368 -51.055 49.881
831823 102.386 -51.076 49.835
831824 102.404 -51.092 49.789
831825 102.419 -51.104 49.744
831826 102.435 -51.114 49.701
831827 102.451 -51.125 49.660
831828 102.466 -51.136 49.623
831829 102.482 -51.147 49.589
831830 102.498 -51.158 49.556
831831 102.514 -51.168 49.528
831832 102.529 -51.180 49.504
831833 102.547 -51.196 49.483
831834 102.566 -51.216 49.459
831835 102.585 -51.236 49.435
831836 102.604 -51.256 49.410
831837 102.623 -51.277 49.384
831838 102.643 -51.296 49.353
831839 102.662 -51.317 49.318
831840 102.681 -51.337 49.283
831841 102.699 -51.357 49.249
831842 102.719 -51.378 49.217
831843 102.724 -51.396 49.187
831844 102.727 -51.416 49.157
831845 102.729 -51.435 49.127
831846 102.704 -51.433 49.099
831847 102.674 -51.426 49.071
831848 102.684 -51.434 49.043
831849 102.712 -51.449 49.015
831850 102.740 -51.464 48.987
831851 102.767 -51.478 48.960
831852 102.795 -51.493 48.932
831853 102.823 -51.508 48.902
831854 102.851 -51.522 48.876
831855 102.878 -51.537 48.848
831856 102.906 -51.552 48.819
831857 102.934 -51.566 48.791
831858 102.961 -51.581 48.762
831859 102.989 -51.596 48.734
831860 103.017 -51.610 48.708
831861 103.044 -51.625 48.682
831862 103.072 -51.640 48.656
831863 103.100 -51.654 48.633
832502 101.751 -50.703 50.708
832503 101.778 -50.729 50.665
832504 101.806 -50.751 50.623
832505 101.833 -50.770 50.580
832506 101.862 -50.791 50.537
832507 101.891 -50.808 50.494
832508 101.922 -50.823 50.452
832509 101.953 -50.837 50.410
832510 101.984 -50.851 50.369
832511 102.016 -50.865 50.328
832512 102.047 -50.879 50.287
832513 102.078 -50.894 50.247
832514 102.109 -50.907 50.208
832515 102.140 -50.922 50.169
832516 102.172 -50.936 50.134
832517 102.203 -50.950 50.103
832518 102.234 -50.964 50.074
832519 102.265 -50.979 50.043
832520 102.298 -50.989 50.008
832521 102.338 -50.990 49.971
832522 102.360 -51.008 49.931
832523 102.378 -51.030 49.888
832524 102.397 -51.051 49.842
832525 102.414 -51.070 49.797
832526 102.430 -51.080 49.752
832527 102.446 -51.091 49.709
832528 102.462 -51.102 49.668
832529 102.477 -51.113 49.631
832530 102.493 -51.124 49.598
832531 102.509 -51.135 49.566
832532 102.525 -51.146 49.538
832533 102.544 -51.166 49.514
832534 102.562 -51.186 49.493
832535 102.582 -51.206 49.470
832536 102.601 -51.227 49.446
832537 102.620 -51.246 49.422
832538 102.639 -51.267 49.396
832539 102.658 -51.287 49.366
832540 102.677 -51.307 49.332
832541 102.696 -51.327 49.298
832542 102.715 -51.347 49.264
832543 102.734 -51.367 49.232
832544 102.747 -51.387 49.201
832545 102.749 -51.406 49.172
832546 102.752 -51.426 49.143
832547 102.728 -51.425 49.115
832548 102.698 -51.418 49.087
832549 102.699 -51.422 49.060
832550 102.727 -51.437 49.032
832551 102.755 -51.452 49.004
832552 102.783 -51.466 48.976
832553 102.810 -51.481 48.948
832554 102.838 -51.496 48.919
832555 102.866 -51.510 48.891
832556 102.893 -51.525 48.863
832557 102.921 -51.540 48.834
832558 102.949 -51.554 48.805
832559 102.976 -51.569 48.776
832560 103.004 -51.584 48.748
832561 103.032 -51.598 48.720
832562 103.059 -51.613 48.694
832563 103.087 -51.628 48.668
832564 103.115 -51.642 48.644
833204 101.778 -50.710 50.678
833205 101.806 -50.730 50.636
833206 101.834 -50.750 50.592
833207 101.862 -50.770 50.548
833208 101.892 -50.786 50.505
833209 101.923 -50.801 50.462
833210 101.955 -50.814 50.420
833211 101.986 -50.829 50.378
833212 102.017 -50.843 50.336
833213 102.048 -50.857 50.295
833214 102.079 -50.871 50.253
833215 102.110 -50.886 50.213
833216 102.141 -50.900 50.174
833217 102.173 -50.914 50.137
833218 102.204 -50.928 50.103
833219 102.235 -50.942 50.072
833220 102.266 -50.957 50.040
833221 102.298 -50.970 50.007
833222 102.329 -50.985 49.973
833223 102.369 -50.986 49.934
833224 102.389 -51.005 49.893
833225 102.407 -51.027 49.848
833226 102.426 -51.046 49.803
833227 102.441 -51.058 49.759
833228 102.457 -51.068 49.717
833229 102.472 -51.079 49.677
833230 102.488 -51.090 49.641
833231 102.504 -51.101 49.607
833232 102.521 -51.116 49.575
833233 102.540 -51.136 49.548
833234 102.559 -51.156 49.523
833235 102.578 -51.176 49.502
833236 102.597 -51.196 49.480
833237 102.616 -51.216 49.457
833238 102.635 -51.236 49.433
833239 102.655 -51.256 49.407
833240 102.674 -51.277 49.377
833241 102.693 -51.296 49.346
833242 102.712 -51.317 49.312
833243 102.731 -51.337 49.279
833244 102.750 -51.357 49.247
833245 102.769 -51.378 49.217
833246 102.772 -51.396 49.188
833247 102.775 -51.416 49.160
833248 102.753 -51.415 49.132
833249 102.723 -51.409 49.104
833250 102.715 -51.410 49.078
833251 102.742 -51.425 49.050
833252 102.770 -51.440 49.022
833253 102.798 -51.455 48.994
833254 102.825 -51.469 48.966
833255 102.853 -51.484 48.937
833256 102.881 -51.499 48.908
833257 102.908 -51.513 48.879
833258 102.936 -51.528 48.849
833259 102.964 -51.543 48.820
833260 102.992 -51.557 48.790
833261 103.019 -51.572 48.761
833262 103.047 -51.587 48.733
833263 103.075 -51.601 48.706
833264 103.102 -51.616 48.680
833265 103.130 -51.631 48.655
833906 101.807 -50.709 50.648
833907 101.835 -50.729 50.604
833908 101.863 -50.749 50.559
833909 101.893 -50.764 50.516
833910 101.924 -50.778 50.473
833911 101.956 -50.792 50.430
833912 101.987 -50.807 50.388
833913 102.018 -50.821 50.344
833914 102.049 -50.835 50.302
833915 102.081 -50.849 50.260
833916 102.112 -50.864 50.219
833917 102.143 -50.878 50.180
833918 102.174 -50.892 50.140
833919 102.205 -50.906 50.104
833920 102.237 -50.920 50.071
833921 102.268 -50.934 50.039
833922 102.299 -50.948 50.008
833923 102.330 -50.963 49.975
833924 102.362 -50.977 49.938
833925 102.399 -50.980 49.898
833926 102.417 -51.001 49.853
833927 102.436 -51.023 49.809
833928 102.452 -51.034 49.766
833929 102.468 -51.045 49.725
833930 102.483 -51.056 49.686
833931 102.499 -51.067 49.650
833932 102.518 -51.086 49.616
833933 102.537 -51.105 49.585
833934 102.556 -51.126 49.558
833935 102.575 -51.146 49.533
833936 102.594 -51.166 49.511
833937 102.613 -51.186 49.489
833938 102.632 -51.206 49.468
833939 102.651 -51.227 49.443
833940 102.670 -51.246 49.417
833941 102.689 -51.267 49.389
833942 102.709 -51.287 49.359
833943 102.728 -51.307 49.326
833944 102.746 -51.327 49.293
833945 102.765 -51.347 49.262
833946 102.784 -51.367 49.233
833947 102.795 -51.387 49.203
833948 102.798 -51.406 49.177
833949 102.777 -51.407 49.150
833950 102.747 -51.400 49.123
833951 102.729 -51.399 49.098
833952 102.757 -51.413 49.070
833953 102.785 -51.428 49.041
833954 102.813 -51.443 49.013
833955 102.840 -51.457 48.984
833956 102.868 -51.472 48.956
833957 102.896 -51.487 48.926
833958 102.924 -51.501 48.896
833959 102.951 -51.516 48.865
833960 102.979 -51.531 48.835
833961 103.007 -51.545 48.804
833962 103.034 -51.560 48.776
833963 103.062 -51.575 48.747
833964 103.090 -51.589 48.719
833965 103.117 -51.604 48.693
833966 103.145 -51.619 48.667
834608 101.835 -50.709 50.615
834609 101.864 -50.728 50.571
834610 101.895 -50.741 50.527
834611 101.926 -50.756 50.484
834612 101.957 -50.770 50.440
834613 101.988 -50.785 50.397
834614 102.020 -50.798 50.353
834615 102.051 -50.813 50.309
834616 102.082 -50.827 50.268
834617 102.113 -50.842 50.226
834618 102.144 -50.855 50.185
834619 102.176 -50.870 50.145
834620 102.207 -50.884 50.108
834621 102.238 -50.898 50.072
834622 102.269 -50.912 50.040
834623 102.301 -50.926 50.010
834624 102.332 -50.941 49.978
834625 102.363 -50.954 49.942
834626 102.394 -50.969 49.902
834627 102.428 -50.977 49.858
834628 102.447 -50.998 49.816
834629 102.463 -51.011 49.774
834630 102.478 -51.022 49.734
834631 102.495 -51.035 49.695
834632 102.514 -51.055 49.659
834633 102.533 -51.075 49.626
834634 102.552 -51.096 49.595
834635 102.572 -51.116 49.569
834636 102.591 -51.136 49.545
834637 102.609 -51.156 49.522
834638 102.628 -51.176 49.499
834639 102.647 -51.196 49.476
834640 102.667 -51.216 49.451
834641 102.686 -51.236 49.426
834642 102.705 -51.256 49.398
834643 102.724 -51.277 49.370
834644 102.743 -51.296 49.339
834645 102.762 -51.317 49.307
834646 102.781 -51.337 49.276
834647 102.800 -51.357 49.248
834648 102.818 -51.377 49.220
834649 102.820 -51.396 49.194
834650 102.802 -51.399 49.169
834651 102.771 -51.392 49.143
834652 102.744 -51.387 49.117
834653 102.772 -51.401 49.089
834654 102.800 -51.416 49.060
834655 102.827 -51.431 49.032
834656 102.855 -51.445 49.004
834657 102.883 -51.460 48.975
834658 102.911 -51.475 48.945
834659 102.939 -51.489 48.914
834660 102.966 -51.504 48.882
834661 102.994 -51.519 48.851
834662 103.022 -51.533 48.820
834663 103.049 -51.548 48.790
834664 103.077 -51.562 48.761
834665 103.105 -51.578 48.732
834666 103.132 -51.592 48.705
834667 103.160 -51.606 48.680
834668 103.188 -51.622 48.655
835310 101.865 -50.706 50.580
835311 101.896 -50.719 50.536
835312 101.927 -50.734 50.492
835313 101.959 -50.748 50.448
835314 101.990 -50.762 50.404
835315 102.021 -50.776 50.359
835316 102.052 -50.791 50.316
835317 102.084 -50.805 50.274
835318 102.115 -50.819 50.232
835319 102.146 -50.833 50.191
835320 102.177 -50.848 50.151
835321 102.208 -50.861 50.112
835322 102.239 -50.876 50.075
835323 102.270 -50.890 50.042
835324 102.302 -50.904 50.011
835325 102.333 -50.918 49.979
835326 102.364 -50.933 49.943
835327 102.395 -50.947 49.904
835328 102.427 -50.961 49.863
835329 102.457 -50.974 49.822
835330 102.474 -50.988 49.781
835331 102.491 -51.005 49.742
835332 102.511 -51.025 49.704
835333 102.530 -51.045 49.668
835334 102.549 -51.065 49.635
835335 102.568 -51.086 49.605
835336 102.587 -51.105 49.579
835337 102.606 -51.126 49.555
835338 102.625 -51.146 49.532
835339 102.644 -51.166 49.508
835340 102.663 -51.186 49.484
835341 102.682 -51.206 49.459
835342 102.701 -51.226 49.434
835343 102.721 -51.246 49.407
835344 102.740 -51.267 49.380
835345 102.759 -51.287 49.351
835346 102.777 -51.307 49.320
835347 102.797 -51.327 49.291
835348 102.816 -51.347 49.264
835349 102.835 -51.367 49.238
835350 102.843 -51.387 49.211
835351 102.826 -51.390 49.188
835352 102.796 -51.384 49.163
835353 102.765 -51.377 49.136
835354 102.787 -51.390 49.108
835355 102.815 -51.404 49.080
835356 102.842 -51.418 49.051
835357 102.870 -51.434 49.023
835358 102.898 -51.448 48.994
835359 102.926 -51.462 48.964
835360 102.953 -51.478 48.931
835361 102.981 -51.492 48.900
835362 103.009 -51.506 48.868
835363 103.037 -51.522 48.836
835364 103.065 -51.536 48.806
835365 103.092 -51.550 48.776
835366 103.120 -51.566 48.747
835367 103.148 -51.580 48.718
835368 103.175 -51.594 48.692
835369 103.203 -51.609 48.668
836012 101.898 -50.698 50.543
836013 101.929 -50.712 50.499
836014 101.960 -50.726 50.455
836015 101.991 -50.740 50.411
836016 102.022 -50.754 50.365
836017 102.053 -50.769 50.323
836018 102.085 -50.782 50.280
836019 102.116 -50.797 50.238
836020 102.147 -50.811 50.196
836021 102.178 -50.826 50.156
836022 102.209 -50.839 50.116
836023 102.241 -50.854 50.079
836024 102.272 -50.868 50.045
836025 102.303 -50.882 50.011
836026 102.334 -50.896 49.978
836027 102.366 -50.911 49.942
836028 102.397 -50.925 49.904
836029 102.428 -50.939 49.866
836030 102.459 -50.953 49.828
836031 102.484 -50.972 49.788
836032 102.503 -50.992 49.749
836033 102.524 -51.014 49.711
836034 102.545 -51.035 49.676
836035 102.564 -51.055 49.643
836036 102.584 -51.075 49.614
836037 102.603 -51.096 49.589
836038 102.622 -51.115 49.565
836039 102.641 -51.136 49.541
836040 102.660 -51.156 49.517
836041 102.679 -51.176 49.493
836042 102.698 -51.196 49.468
836043 102.717 -51.216 49.442
836044 102.736 -51.236 49.416
836045 102.755 -51.256 49.390
836046 102.774 -51.277 49.362
836047 102.793 -51.296 49.333
836048 102.812 -51.317 49.306
836049 102.831 -51.337 49.281
836050 102.851 -51.357 49.254
836051 102.866 -51.377 49.229
836052 102.851 -51.382 49.208
836053 102.820 -51.375 49.182
836054 102.790 -51.368 49.155
836055 102.802 -51.378 49.128
836056 102.830 -51.392 49.100
836057 102.858 -51.407 49.071
836058 102.885 -51.421 49.042
836059 102.913 -51.436 49.012
836060 102.941 -51.451 48.982
836061 102.968 -51.465 48.949
836062 102.996 -51.480 48.917
836063 103.024 -51.495 48.885
836064 103.051 -51.509 48.852
836065 103.080 -51.524 48.821
836066 103.107 -51.539 48.791
836067 103.135 -51.553 48.762
836068 103.163 -51.568 48.734
836069 103.190 -51.583 48.706
836070 103.218 -51.597 48.681
836714 101.930 -50.689 50.504
836715 101.961 -50.704 50.461
836716 101.992 -50.718 50.416
836717 102.024 -50.732 50.372
836718 102.055 -50.746 50.328
836719 102.086 -50.760 50.284
836720 102.117 -50.775 50.243
836721 102.149 -50.789 50.201
836722 102.180 -50.803 50.160
836723 102.211 -50.817 50.121
836724 102.242 -50.832 50.083
836725 102.273 -50.845 50.047
836726 102.305 -50.860 50.012
836727 102.336 -50.874 49.976
836728 102.367 -50.889 49.940
836729 102.398 -50.902 49.903
836730 102.430 -50.917 49.867
836731 102.460 -50.931 49.831
836732 102.485 -50.954 49.792
836733 102.510 -50.976 49.754
836734 102.526 -50.994 49.717
836735 102.540 -51.011 49.683
836736 102.560 -51.031 49.651
836737 102.581 -51.053 49.623
836738 102.602 -51.074 49.598
836739 102.624 -51.096 49.574
836740 102.645 -51.117 49.551
836741 102.666 -51.139 49.526
836742 102.687 -51.161 49.501
836743 102.709 -51.183 49.478
836744 102.730 -51.205 49.454
836745 102.751 -51.226 49.429
836746 102.771 -51.246 49.402
836747 102.790 -51.266 49.375
836748 102.809 -51.287 49.348
836749 102.828 -51.307 49.321
836750 102.847 -51.327 49.296
836751 102.866 -51.347 49.271
836752 102.885 -51.367 49.246
836753 102.875 -51.374 49.224
836754 102.845 -51.367 49.199
836755 102.815 -51.360 49.174
836756 102.817 -51.365 49.147
836757 102.845 -51.381 49.120
836758 102.873 -51.395 49.089
836759 102.900 -51.409 49.057
836760 102.928 -51.425 49.027
836761 102.956 -51.439 48.998
836762 102.983 -51.453 48.966
836763 103.011 -51.468 48.934
836764 103.039 -51.483 48.902
836765 103.067 -51.497 48.869
836766 103.094 -51.512 48.838
836767 103.122 -51.527 48.807
836768 103.150 -51.541 48.778
836769 103.178 -51.556 48.749
836770 103.206 -51.571 48.720
836771 103.233 -51.585 48.695
837415 101.932 -50.667 50.510
837416 101.963 -50.682 50.466
837417 101.994 -50.696 50.422
837418 102.025 -50.710 50.378
837419 102.056 -50.724 50.334
837420 102.088 -50.738 50.290
837421 102.119 -50.753 50.248
837422 102.150 -50.766 50.207
837423 102.181 -50.781 50.166
837424 102.212 -50.795 50.126
837425 102.244 -50.810 50.088
837426 102.275 -50.823 50.050
837427 102.306 -50.838 50.014
837428 102.337 -50.852 49.978
837429 102.368 -50.866 49.941
837430 102.399 -50.880 49.904
837431 102.430 -50.895 49.869
837432 102.459 -50.912 49.833
837433 102.484 -50.934 49.796
837434 102.509 -50.957 49.759
837435 102.533 -50.979 49.724
837436 102.554 -50.999 49.690
837437 102.568 -51.016 49.659
837438 102.582 -51.033 49.631
837439 102.596 -51.049 49.606
837440 102.617 -51.070 49.583
837441 102.638 -51.092 49.559
837442 102.659 -51.114 49.534
837443 102.681 -51.136 49.510
837444 102.702 -51.157 49.489
837445 102.723 -51.179 49.466
837446 102.744 -51.200 49.442
837447 102.766 -51.222 49.415
837448 102.787 -51.243 49.388
837449 102.808 -51.265 49.362
837450 102.829 -51.287 49.335
837451 102.851 -51.309 49.311
837452 102.872 -51.331 49.286
837453 102.893 -51.352 49.263
837454 102.899 -51.365 49.240
837455 102.869 -51.359 49.215
837456 102.839 -51.352 49.190
837457 102.832 -51.354 49.165
837458 102.860 -51.368 49.137
837459 102.888 -51.383 49.106
837460 102.915 -51.398 49.073
837461 102.943 -51.412 49.043
837462 102.971 -51.427 49.014
837463 102.999 -51.442 48.982
837464 103.026 -51.456 48.950
837465 103.054 -51.471 48.918
837466 103.082 -51.486 48.885
837467 103.109 -51.500 48.855
837468 103.137 -51.515 48.823
837469 103.165 -51.530 48.793
837470 103.192 -51.544 48.764
837471 103.220 -51.559 48.735
837472 103.248 -51.574 48.708
838117 101.964 -50.660 50.473
838118 101.995 -50.673 50.429
838119 102.027 -50.688 50.384
838120 102.058 -50.702 50.340
838121 102.089 -50.716 50.297
838122 102.120 -50.730 50.255
838123 102.151 -50.745 50.213
838124 102.182 -50.759 50.173
838125 102.214 -50.773 50.132
838126 102.245 -50.787 50.093
838127 102.276 -50.801 50.054
838128 102.307 -50.816 50.017
838129 102.338 -50.829 49.980
838130 102.370 -50.844 49.944
838131 102.401 -50.858 49.907
838132 102.432 -50.873 49.871
838133 102.459 -50.892 49.836
838134 102.483 -50.915 49.800
838135 102.508 -50.937 49.764
838136 102.533 -50.960 49.730
838137 102.557 -50.982 49.697
838138 102.582 -51.005 49.667
838139 102.596 -51.021 49.639
838140 102.610 -51.038 49.614
838141 102.624 -51.055 49.591
838142 102.638 -51.071 49.568
838143 102.653 -51.088 49.544
838144 102.674 -51.110 49.521
838145 102.695 -51.131 49.500
838146 102.716 -51.153 49.479
838147 102.737 -51.174 49.454
838148 102.759 -51.196 49.428
838149 102.780 -51.218 49.401
838150 102.801 -51.240 49.375
838151 102.823 -51.261 49.349
838152 102.844 -51.283 49.325
838153 102.865 -51.305 49.302
838154 102.886 -51.326 49.279
838155 102.908 -51.348 49.255
838156 102.894 -51.350 49.230
838157 102.864 -51.343 49.205
838158 102.847 -51.342 49.180
838159 102.875 -51.356 49.152
838160 102.903 -51.371 49.121
838161 102.930 -51.386 49.089
838162 102.958 -51.400 49.059
838163 102.986 -51.415 49.030
838164 103.014 -51.430 48.998
838165 103.041 -51.444 48.966
838166 103.069 -51.459 48.934
838167 103.097 -51.474 48.902
838168 103.124 -51.488 48.871
838169 103.152 -51.503 48.839
838170 103.180 -51.518 48.808
838171 103.208 -51.532 48.779
838172 103.235 -51.547 48.750
838173 103.263 -51.562 48.722
838819 101.996 -50.651 50.436
838820 102.028 -50.666 50.391
838821 102.059 -50.680 50.346
838822 102.090 -50.694 50.304
838823 102.121 -50.708 50.262
838824 102.153 -50.723 50.220
838825 102.184 -50.737 50.179
838826 102.215 -50.751 50.139
838827 102.246 -50.765 50.099
838828 102.277 -50.779 50.059
838829 102.309 -50.794 50.022
838830 102.340 -50.807 49.985
838831 102.371 -50.822 49.949
838832 102.402 -50.836 49.912
838833 102.433 -50.851 49.876
838834 102.458 -50.873 49.840
838835 102.482 -50.895 49.804
838836 102.507 -50.918 49.769
838837 102.532 -50.940 49.736
838838 102.556 -50.963 49.704
838839 102.581 -50.985 49.674
838840 102.606 -51.008 49.647
838841 102.623 -51.026 49.622
838842 102.638 -51.043 49.599
838843 102.652 -51.060 49.577
838844 102.666 -51.077 49.555
838845 102.680 -51.093 49.534
838846 102.694 -51.110 49.512
838847 102.709 -51.127 49.489
838848 102.730 -51.149 49.465
838849 102.752 -51.171 49.438
838850 102.773 -51.192 49.413
838851 102.794 -51.214 49.388
838852 102.816 -51.235 49.364
838853 102.837 -51.257 49.340
838854 102.858 -51.279 49.317
838855 102.879 -51.300 49.294
838856 102.901 -51.322 49.271
838857 102.918 -51.341 49.246
838858 102.888 -51.335 49.221
838859 102.863 -51.330 49.194
838860 102.890 -51.345 49.166
838861 102.918 -51.359 49.136
838862 102.946 -51.374 49.106
838863 102.973 -51.389 49.076
838864 103.001 -51.403 49.047
838865 103.029 -51.418 49.016
838866 103.056 -51.433 48.984
838867 103.084 -51.447 48.951
838868 103.112 -51.462 48.918
838869 103.140 -51.477 48.887
838870 103.167 -51.491 48.855
838871 103.195 -51.506 48.824
838872 103.223 -51.521 48.795
838873 103.250 -51.535 48.766
838874 103.278 -51.550 48.738
839520 101.998 -50.629 50.443
839521 102.029 -50.644 50.399
839522 102.060 -50.657 50.355
839523 102.092 -50.672 50.311
839524 102.123 -50.686 50.269
839525 102.154 -50.701 50.227
839526 102.185 -50.714 50.186
839527 102.217 -50.729 50.146
839528 102.248 -50.743 50.106
839529 102.279 -50.757 50.067
839530 102.310 -50.771 50.029
839531 102.341 -50.785 49.992
839532 102.373 -50.800 49.955
839533 102.404 -50.813 49.917
839534 102.432 -50.831 49.881
839535 102.457 -50.854 49.845
839536 102.481 -50.876 49.810
839537 102.506 -50.898 49.775
839538 102.531 -50.921 49.743
839539 102.555 -50.943 49.712
839540 102.580 -50.966 49.683
839541 102.605 -50.988 49.655
839542 102.629 -51.011 49.630
839543 102.650 -51.031 49.607
839544 102.665 -51.048 49.585
839545 102.679 -51.064 49.565
839546 102.693 -51.081 49.544
839547 102.708 -51.099 49.522
839548 102.722 -51.115 49.499
839549 102.736 -51.132 49.475
839550 102.750 -51.149 49.450
839551 102.766 -51.166 49.426
839552 102.787 -51.188 49.403
839553 102.809 -51.209 49.379
839554 102.830 -51.231 49.356
839555 102.851 -51.253 49.333
839556 102.872 -51.274 49.311
839557 102.894 -51.296 49.288
839558 102.915 -51.318 49.263
839559 102.912 -51.326 49.237
839560 102.882 -51.320 49.210
839561 102.905 -51.333 49.182
839562 102.933 -51.347 49.153
839563 102.961 -51.362 49.124
839564 102.988 -51.377 49.094
839565 103.016 -51.391 49.065
839566 103.044 -51.406 49.034
839567 103.072 -51.421 49.002
839568 103.099 -51.435 48.969
839569 103.127 -51.450 48.935
839570 103.155 -51.465 48.903
839571 103.182 -51.479 48.872
839572 103.210 -51.494 48.841
839573 103.238 -51.509 48.811
839574 103.265 -51.523 48.782
839575 103.293 -51.538 48.754
840222 102.031 -50.622 50.406
840223 102.062 -50.635 50.362
840224 102.093 -50.650 50.319
840225 102.124 -50.664 50.276
840226 102.156 -50.678 50.235
840227 102.187 -50.692 50.193
840228 102.218 -50.707 50.153
840229 102.249 -50.721 50.113
840230 102.280 -50.735 50.073
840231 102.311 -50.749 50.035
840232 102.342 -50.763 49.998
840233 102.374 -50.778 49.961
840234 102.405 -50.791 49.923
840235 102.432 -50.812 49.887
840236 102.456 -50.834 49.851
840237 102.481 -50.857 49.816
840238 102.506 -50.879 49.782
840239 102.530 -50.901 49.750
840240 102.554 -50.924 49.720
840241 102.579 -50.946 49.691
840242 102.604 -50.969 49.664
840243 102.628 -50.991 49.638
840244 102.653 -51.014 49.615
840245 102.678 -51.036 49.594
840246 102.692 -51.053 49.573
840247 102.706 -51.070 49.553
840248 102.721 -51.086 49.532
840249 102.735 -51.103 49.510
840250 102.749 -51.120 49.486
840251 102.763 -51.136 49.463
840252 102.777 -51.153 49.441
840253 102.792 -51.171 49.418
840254 102.806 -51.187 49.394
840255 102.823 -51.205 49.372
840256 102.844 -51.227 49.349
840257 102.866 -51.249 49.328
840258 102.887 -51.271 49.305
840259 102.908 -51.292 49.280
840260 102.929 -51.314 49.255
840261 102.906 -51.311 49.228
840262 102.921 -51.321 49.200
840263 102.948 -51.335 49.172
840264 102.976 -51.350 49.142
840265 103.004 -51.365 49.113
840266 103.031 -51.379 49.084
840267 103.059 -51.394 49.053
840268 103.087 -51.409 49.020
840269 103.114 -51.423 48.986
840270 103.142 -51.438 48.952
840271 103.170 -51.453 48.921
840272 103.197 -51.468 48.890
840273 103.225 -51.482 48.858
840274 103.253 -51.497 48.828
840275 103.281 -51.512 48.799
840276 103.308 -51.526 48.771
840923 102.032 -50.599 50.413
840924 102.063 -50.613 50.369
840925 102.094 -50.628 50.326
840926 102.125 -50.641 50.283
840927 102.157 -50.656 50.241
840928 102.188 -50.670 50.200
840929 102.219 -50.685 50.159
840930 102.250 -50.698 50.119
840931 102.281 -50.713 50.079
840932 102.313 -50.727 50.042
840933 102.344 -50.741 50.003
840934 102.375 -50.755 49.965
840935 102.406 -50.770 49.929
840936 102.430 -50.792 49.893
840937 102.455 -50.815 49.857
840938 102.480 -50.837 49.823
840939 102.504 -50.860 49.789
840940 102.529 -50.882 49.757
840941 102.554 -50.904 49.727
840942 102.579 -50.927 49.699
840943 102.603 -50.949 49.672
840944 102.628 -50.972 49.646
840945 102.653 -50.994 49.625
840946 102.677 -51.017 49.603
840947 102.701 -51.039 49.582
840948 102.720 -51.058 49.563
840949 102.734 -51.075 49.543
840950 102.748 -51.092 49.521
840951 102.763 -51.108 49.499
840952 102.777 -51.125 49.478
840953 102.791 -51.142 49.456
840954 102.805 -51.158 49.433
840955 102.819 -51.175 49.409
840956 102.833 -51.192 49.387
840957 102.848 -51.208 49.365
840958 102.862 -51.226 49.344
840959 102.880 -51.245 49.323
840960 102.901 -51.266 49.299
840961 102.922 -51.288 49.275
840962 102.931 -51.302 49.249
840963 102.936 -51.309 49.221
840964 102.963 -51.324 49.192
840965 102.991 -51.338 49.163
840966 103.019 -51.352 49.133
840967 103.046 -51.368 49.103
840968 103.074 -51.382 49.071
840969 103.102 -51.396 49.038
840970 103.129 -51.412 49.004
840971 103.157 -51.426 48.971
840972 103.184 -51.441 48.939
840973 103.211 -51.456 48.908
840974 103.237 -51.471 48.876
840975 103.264 -51.485 48.845
841625 102.064 -50.591 50.374
841626 102.096 -50.606 50.331
841627 102.127 -50.619 50.289
841628 102.158 -50.634 50.247
841629 102.189 -50.648 50.206
841630 102.221 -50.662 50.165
841631 102.252 -50.676 50.125
841632 102.283 -50.691 50.086
841633 102.314 -50.705 50.047
841634 102.345 -50.719 50.009
841635 102.377 -50.733 49.971
841636 102.405 -50.751 49.934
841637 102.430 -50.773 49.898
841638 102.454 -50.795 49.863
841639 102.479 -50.818 49.828
841640 102.504 -50.840 49.796
841641 102.528 -50.863 49.764
841642 102.553 -50.886 49.734
841643 102.578 -50.907 49.706
841644 102.602 -50.930 49.680
841645 102.627 -50.952 49.655
841646 102.652 -50.975 49.633
841647 102.676 -50.997 49.611
841648 102.701 -51.020 49.591
841649 102.726 -51.042 49.574
841650 102.748 -51.063 49.554
841651 102.762 -51.080 49.533
841652 102.776 -51.097 49.512
841653 102.790 -51.114 49.492
841654 102.805 -51.130 49.470
841655 102.819 -51.147 49.447
841656 102.833 -51.164 49.424
841657 102.847 -51.180 49.402
841658 102.861 -51.197 49.381
841659 102.875 -51.214 49.361
841660 102.890 -51.230 49.340
841661 102.904 -51.247 49.318
841662 102.918 -51.264 49.295
841663 102.937 -51.284 49.270
841664 102.951 -51.297 49.243
841665 102.978 -51.312 49.214
841666 103.006 -51.326 49.184
841667 103.033 -51.341 49.154
841668 103.059 -51.356 49.123
841669 103.086 -51.371 49.090
841670 103.113 -51.385 49.057
841671 103.140 -51.400 49.023
841672 103.167 -51.415 48.990
841673 103.193 -51.430 48.957
841674 103.220 -51.445 48.925
841675 103.247 -51.459 48.894
842327 102.097 -50.583 50.338
842328 102.128 -50.597 50.295
842329 102.160 -50.612 50.254
842330 102.191 -50.625 50.213
842331 102.222 -50.640 50.172
842332 102.253 -50.654 50.132
842333 102.284 -50.669 50.092
842334 102.316 -50.682 50.054
842335 102.347 -50.697 50.015
842336 102.378 -50.711 49.977
842337 102.404 -50.732 49.940
842338 102.429 -50.754 49.903
842339 102.454 -50.776 49.869
842340 102.478 -50.798 49.835
842341 102.503 -50.821 49.803
842342 102.527 -50.843 49.771
842343 102.552 -50.866 49.741
842344 102.577 -50.889 49.714
842345 102.601 -50.911 49.688
842346 102.626 -50.933 49.664
842347 102.650 -50.955 49.641
842348 102.675 -50.978 49.619
842349 102.700 -51.000 49.600
842350 102.725 -51.023 49.584
842351 102.749 -51.045 49.566
842352 102.774 -51.067 49.545
842353 102.790 -51.085 49.525
842354 102.804 -51.102 49.505
842355 102.818 -51.118 49.484
842356 102.832 -51.135 49.461
842357 102.846 -51.152 49.438
842358 102.861 -51.169 49.417
842359 102.875 -51.186 49.398
842360 102.889 -51.202 49.379
842361 102.903 -51.219 49.357
842362 102.917 -51.236 49.338
842363 102.931 -51.252 49.317
842364 102.946 -51.269 49.294
842365 102.962 -51.286 49.266
842366 102.989 -51.300 49.236
842367 103.016 -51.315 49.206
842368 103.043 -51.330 49.175
842369 103.070 -51.345 49.143
842370 103.096 -51.359 49.110
842371 103.123 -51.374 49.077
842372 103.150 -51.389 49.042
842373 103.176 -51.404 49.009
842374 103.204 -51.418 48.976
842375 103.230 -51.434 48.944
843029 102.130 -50.575 50.303
843030 102.161 -50.589 50.261
843031 102.192 -50.603 50.220
843032 102.223 -50.618 50.179
843033 102.254 -50.632 50.138
843034 102.286 -50.646 50.099
843035 102.317 -50.660 50.060
843036 102.348 -50.675 50.021
843037 102.379 -50.689 49.984
843038 102.403 -50.712 49.945
843039 102.428 -50.735 49.910
843040 102.453 -50.757 49.876
843041 102.477 -50.779 49.843
843042 102.502 -50.801 49.810
843043 102.527 -50.824 49.779
843044 102.551 -50.846 49.750
843045 102.576 -50.869 49.722
843046 102.601 -50.892 49.697
843047 102.625 -50.914 49.673
843048 102.650 -50.936 49.650
843049 102.674 -50.958 49.628
843050 102.699 -50.981 49.610
843051 102.724 -51.003 49.594
843052 102.748 -51.026 49.577
843053 102.773 -51.049 49.557
843054 102.798 -51.070 49.538
843055 102.817 -51.090 49.519
843056 102.831 -51.107 49.498
843057 102.845 -51.124 49.475
843058 102.860 -51.140 49.452
843059 102.874 -51.157 49.432
843060 102.888 -51.174 49.414
843061 102.902 -51.190 49.396
843062 102.916 -51.207 49.377
843063 102.930 -51.224 49.359
843064 102.945 -51.241 49.340
843065 102.959 -51.258 49.317
843066 102.973 -51.274 49.290
843067 102.999 -51.290 49.261
843068 103.026 -51.304 49.230
843069 103.052 -51.319 49.198
843070 103.079 -51.334 49.165
843071 103.106 -51.349 49.131
843072 103.133 -51.363 49.097
843073 103.160 -51.378 49.062
843074 103.186 -51.393 49.029
843075 103.213 -51.408 48.996
843076 103.240 -51.422 48.963
843077 103.267 -51.437 48.932
843733 102.225 -50.596 50.187
843734 102.256 -50.610 50.147
843735 102.287 -50.624 50.107
843736 102.318 -50.638 50.068
843737 102.349 -50.653 50.030
843738 102.378 -50.670 49.992
843739 102.402 -50.692 49.954
843740 102.427 -50.715 49.919
843741 102.452 -50.738 49.884
843742 102.476 -50.760 49.851
843743 102.501 -50.782 49.819
843744 102.526 -50.804 49.788
843745 102.550 -50.827 49.759
843746 102.575 -50.849 49.731
843747 102.600 -50.872 49.706
843748 102.624 -50.895 49.682
843749 102.649 -50.917 49.659
843750 102.674 -50.939 49.638
843751 102.698 -50.961 49.619
843752 102.723 -50.984 49.604
843753 102.748 -51.006 49.587
843754 102.772 -51.029 49.568
843755 102.797 -51.052 49.549
843756 102.821 -51.074 49.531
843757 102.844 -51.096 49.510
843758 102.859 -51.112 49.487
843759 102.873 -51.129 49.465
843760 102.887 -51.146 49.446
843761 102.901 -51.162 49.430
843762 102.915 -51.179 49.413
843763 102.930 -51.196 49.397
843764 102.944 -51.212 49.380
843765 102.958 -51.229 49.364
843766 102.972 -51.246 49.340
843767 102.986 -51.262 49.313
843768 103.009 -51.278 49.285
843769 103.035 -51.293 49.254
843770 103.062 -51.308 49.221
843771 103.089 -51.322 49.188
843772 103.116 -51.337 49.153
843773 103.143 -51.352 49.118
843774 103.169 -51.367 49.084
843775 103.196 -51.382 49.049
843776 103.223 -51.396 49.016
843777 103.250 -51.412 48.983
843778 103.277 -51.426 48.952
844439 102.377 -50.650 50.001
844440 102.402 -50.673 49.963
844441 102.427 -50.695 49.928
844442 102.451 -50.718 49.893
844443 102.476 -50.741 49.859
844444 102.501 -50.763 49.828
844445 102.525 -50.785 49.797
844446 102.549 -50.807 49.768
844447 102.574 -50.830 49.741
844448 102.599 -50.852 49.715
844449 102.623 -50.875 49.690
844450 102.648 -50.898 49.667
844451 102.673 -50.920 49.647
844452 102.697 -50.942 49.628
844453 102.722 -50.964 49.612
844454 102.747 -50.987 49.595
844455 102.771 -51.009 49.577
844456 102.796 -51.032 49.559
844457 102.821 -51.055 49.541
844458 102.845 -51.077 49.520
844459 102.870 -51.099 49.498
844460 102.886 -51.117 49.478
844461 102.900 -51.134 49.461
844462 102.915 -51.151 49.446
844463 102.929 -51.168 49.430
844464 102.943 -51.184 49.415
844465 102.957 -51.201 49.401
844466 102.971 -51.218 49.386
844467 102.985 -51.234 49.363
844468 103.000 -51.251 49.336
844469 103.019 -51.267 49.308
844470 103.045 -51.282 49.277
844471 103.072 -51.296 49.245
844472 103.099 -51.312 49.211
844473 103.125 -51.326 49.176
844474 103.152 -51.341 49.141
844475 103.179 -51.356 49.105
844476 103.206 -51.371 49.071
844477 103.233 -51.385 49.037
844478 103.259 -51.400 49.004
844479 103.286 -51.415 48.972
845143 102.450 -50.698 49.903
845144 102.475 -50.721 49.869
845145 102.500 -50.744 49.837
845146 102.524 -50.766 49.805
845147 102.549 -50.788 49.777
845148 102.574 -50.810 49.749
845149 102.598 -50.833 49.723
845150 102.623 -50.855 49.698
845151 102.647 -50.878 49.676
845152 102.672 -50.901 49.655
845153 102.696 -50.923 49.636
845154 102.721 -50.945 49.619
845155 102.746 -50.967 49.601
845156 102.770 -50.990 49.583
845157 102.795 -51.012 49.565
845158 102.820 -51.035 49.548
845159 102.844 -51.058 49.530
845160 102.869 -51.080 49.510
845161 102.894 -51.102 49.491
845162 102.914 -51.122 49.475
845163 102.928 -51.139 49.461
845164 102.942 -51.155 49.448
845165 102.957 -51.172 49.434
845166 102.971 -51.189 49.419
845167 102.985 -51.206 49.406
845168 102.999 -51.223 49.384
845169 103.013 -51.240 49.359
845170 103.028 -51.256 49.331
845171 103.055 -51.271 49.300
845172 103.082 -51.286 49.267
845173 103.109 -51.300 49.234
845174 103.135 -51.315 49.199
845175 103.162 -51.330 49.163
845176 103.189 -51.345 49.127
845177 103.216 -51.359 49.093
845178 103.242 -51.374 49.058
845179 103.269 -51.389 49.025
845180 103.296 -51.404 48.993
845181 103.323 -51.418 48.962
845844 102.449 -50.679 49.912
845845 102.474 -50.701 49.878
845846 102.499 -50.724 49.846
845847 102.523 -50.747 49.815
845848 102.548 -50.769 49.786
845849 102.573 -50.791 49.758
845850 102.597 -50.813 49.732
845851 102.622 -50.836 49.707
845852 102.647 -50.858 49.684
845853 102.671 -50.881 49.662
845854 102.696 -50.904 49.643
845855 102.721 -50.926 49.624
845856 102.745 -50.948 49.606
845857 102.769 -50.970 49.588
845858 102.794 -50.993 49.571
845859 102.819 -51.015 49.554
845860 102.843 -51.038 49.537
845861 102.868 -51.061 49.519
845862 102.893 -51.083 49.502
845863 102.917 -51.105 49.488
845864 102.942 -51.127 49.476
845865 102.956 -51.144 49.464
845866 102.970 -51.161 49.452
845867 102.984 -51.177 49.435
845868 102.998 -51.194 49.422
845869 103.013 -51.211 49.402
845870 103.027 -51.227 49.378
845871 103.041 -51.244 49.352
845872 103.065 -51.260 49.322
845873 103.092 -51.274 49.290
845874 103.118 -51.290 49.256
845875 103.145 -51.304 49.221
845876 103.172 -51.319 49.185
845877 103.199 -51.334 49.150
845878 103.226 -51.349 49.115
845879 103.252 -51.363 49.079
845880 103.279 -51.378 49.047
845881 103.306 -51.393 49.014
845882 103.333 -51.408 48.983
846546 102.473 -50.682 49.887
846547 102.498 -50.704 49.854
846548 102.522 -50.727 49.823
846549 102.547 -50.750 49.794
846550 102.572 -50.772 49.766
846551 102.596 -50.795 49.740
846552 102.621 -50.817 49.715
846553 102.646 -50.839 49.691
846554 102.670 -50.861 49.669
846555 102.695 -50.884 49.648
846556 102.720 -50.907 49.629
846557 102.744 -50.929 49.610
846558 102.769 -50.951 49.593
846559 102.794 -50.973 49.576
846560 102.818 -50.996 49.559
846561 102.843 -51.018 49.543
846562 102.868 -51.041 49.527
846563 102.892 -51.064 49.512
846564 102.916 -51.086 49.500
846565 102.941 -51.108 49.490
846566 102.966 -51.130 49.479
846567 102.983 -51.149 49.470
846568 102.998 -51.166 49.450
846569 103.012 -51.183 49.434
846570 103.026 -51.199 49.416
846571 103.040 -51.216 49.395
846572 103.054 -51.233 49.370
846573 103.075 -51.249 49.342
846574 103.101 -51.263 49.311
846575 103.128 -51.278 49.278
846576 103.155 -51.293 49.243
846577 103.182 -51.308 49.207
846578 103.209 -51.323 49.172
846579 103.235 -51.337 49.136
846580 103.262 -51.352 49.100
846581 103.289 -51.367 49.068
846582 103.316 -51.382 49.035
846583 103.342 -51.396 49.004
847247 102.472 -50.663 49.895
847248 102.497 -50.685 49.862
847249 102.522 -50.707 49.831
847250 102.546 -50.730 49.801
847251 102.571 -50.753 49.773
847252 102.596 -50.775 49.747
847253 102.620 -50.798 49.722
847254 102.644 -50.820 49.698
847255 102.669 -50.842 49.675
847256 102.694 -50.864 49.653
847257 102.719 -50.887 49.633
847258 102.743 -50.910 49.614
847259 102.768 -50.932 49.597
847260 102.793 -50.954 49.581
847261 102.817 -50.976 49.565
847262 102.842 -50.999 49.550
847263 102.867 -51.021 49.535
847264 102.891 -51.044 49.522
847265 102.916 -51.067 49.511
847266 102.941 -51.089 49.500
847267 102.965 -51.111 49.492
847268 102.990 -51.133 49.481
847269 103.011 -51.154 49.461
847270 103.025 -51.171 49.443
847271 103.039 -51.188 49.425
847272 103.053 -51.205 49.405
847273 103.068 -51.221 49.382
847274 103.085 -51.237 49.356
847275 103.111 -51.252 49.327
847276 103.138 -51.267 49.296
847277 103.165 -51.282 49.262
847278 103.191 -51.296 49.227
847279 103.218 -51.312 49.191
847280 103.245 -51.326 49.156
847281 103.272 -51.341 49.120
847282 103.299 -51.356 49.088
847283 103.325 -51.371 49.055
847284 103.352 -51.386 49.024
847949 102.496 -50.666 49.871
847950 102.521 -50.688 49.840
847951 102.545 -50.710 49.809
847952 102.570 -50.733 49.781
847953 102.595 -50.756 49.754
847954 102.619 -50.778 49.729
847955 102.644 -50.801 49.704
847956 102.669 -50.823 49.680
847957 102.693 -50.845 49.658
847958 102.718 -50.867 49.638
847959 102.743 -50.890 49.620
847960 102.767 -50.913 49.602
847961 102.792 -50.935 49.586
847962 102.816 -50.958 49.571
847963 102.841 -50.980 49.556
847964 102.866 -51.002 49.543
847965 102.890 -51.024 49.531
847966 102.915 -51.047 49.520
847967 102.940 -51.070 49.510
847968 102.964 -51.092 49.502
847969 102.989 -51.114 49.491
847970 103.014 -51.136 49.472
847971 103.038 -51.159 49.453
847972 103.052 -51.176 49.434
847973 103.067 -51.193 49.415
847974 103.081 -51.209 49.392
847975 103.095 -51.226 49.368
847976 103.121 -51.241 49.341
847977 103.148 -51.256 49.312
847978 103.175 -51.271 49.279
847979 103.201 -51.286 49.245
847980 103.228 -51.300 49.210
847981 103.255 -51.315 49.175
847982 103.281 -51.330 49.140
847983 103.308 -51.345 49.108
847984 103.335 -51.359 49.075
847985 103.362 -51.374 49.044
847986 103.389 -51.389 49.014
848650 102.495 -50.647 49.880
848651 102.520 -50.669 49.849
848652 102.544 -50.691 49.818
848653 102.569 -50.714 49.789
848654 102.594 -50.736 49.761
848655 102.618 -50.759 49.735
848656 102.643 -50.781 49.710
848657 102.668 -50.804 49.686
848658 102.692 -50.826 49.664
848659 102.717 -50.848 49.644
848660 102.742 -50.870 49.625
848661 102.766 -50.893 49.608
848662 102.791 -50.916 49.592
848663 102.816 -50.938 49.577
848664 102.840 -50.961 49.563
848665 102.865 -50.983 49.550
848666 102.889 -51.005 49.539
848667 102.914 -51.027 49.528
848668 102.939 -51.050 49.519
848669 102.963 -51.073 49.511
848670 102.988 -51.095 49.501
848671 103.013 -51.117 49.482
848672 103.037 -51.139 49.463
848673 103.062 -51.162 49.444
848674 103.080 -51.181 49.424
848675 103.094 -51.198 49.402
848676 103.109 -51.215 49.379
848677 103.131 -51.230 49.354
848678 103.158 -51.245 49.326
848679 103.184 -51.260 49.295
848680 103.211 -51.274 49.262
848681 103.238 -51.290 49.229
848682 103.265 -51.304 49.193
848683 103.292 -51.319 49.159
848684 103.318 -51.334 49.127
848685 103.345 -51.349 49.093
848686 103.372 -51.363 49.062
848687 103.399 -51.378 49.032
849352 102.519 -50.650 49.857
849353 102.544 -50.672 49.827
849354 102.569 -50.694 49.797
849355 102.593 -50.717 49.768
849356 102.617 -50.739 49.742
849357 102.642 -50.762 49.716
849358 102.667 -50.784 49.692
849359 102.691 -50.807 49.670
849360 102.716 -50.829 49.649
849361 102.741 -50.851 49.630
849362 102.765 -50.873 49.614
849363 102.790 -50.896 49.597
849364 102.815 -50.919 49.583
849365 102.839 -50.941 49.570
849366 102.864 -50.964 49.558
849367 102.889 -50.986 49.547
849368 102.913 -51.008 49.537
849369 102.938 -51.030 49.527
849370 102.963 -51.053 49.519
849371 102.987 -51.076 49.509
849372 103.012 -51.098 49.491
849373 103.036 -51.121 49.475
849374 103.061 -51.142 49.456
849375 103.086 -51.165 49.435
849376 103.108 -51.186 49.413
849377 103.122 -51.203 49.390
849378 103.141 -51.219 49.366
849379 103.167 -51.233 49.338
849380 103.194 -51.249 49.308
849381 103.221 -51.263 49.276
849382 103.248 -51.278 49.243
849383 103.275 -51.293 49.208
849384 103.301 -51.308 49.175
849385 103.328 -51.323 49.143
849386 103.355 -51.337 49.111
849387 103.380 -51.349 49.080
849388 103.406 -51.362 49.050
850054 102.543 -50.653 49.835
850055 102.568 -50.675 49.805
850056 102.592 -50.698 49.777
850057 102.617 -50.720 49.749
850058 102.641 -50.742 49.723
850059 102.666 -50.765 49.698
850060 102.691 -50.787 49.676
850061 102.716 -50.810 49.655
850062 102.740 -50.832 49.636
850063 102.764 -50.854 49.619
850064 102.789 -50.876 49.603
850065 102.814 -50.899 49.589
850066 102.838 -50.922 49.576
850067 102.863 -50.944 49.565
850068 102.888 -50.967 49.554
850069 102.912 -50.989 49.544
850070 102.937 -51.011 49.535
850071 102.962 -51.033 49.528
850072 102.986 -51.056 49.518
850073 103.011 -51.079 49.501
850074 103.036 -51.101 49.484
850075 103.061 -51.124 49.466
850076 103.085 -51.146 49.445
850077 103.110 -51.168 49.423
850078 103.134 -51.190 49.402
850079 103.151 -51.208 49.377
850080 103.177 -51.223 49.350
850081 103.204 -51.237 49.320
850082 103.231 -51.252 49.289
850083 103.257 -51.266 49.257
850084 103.283 -51.277 49.223
850085 103.308 -51.290 49.192
850086 103.333 -51.301 49.160
850087 103.359 -51.313 49.129
850088 103.384 -51.324 49.098
850089 103.409 -51.337 49.068
850756 102.567 -50.656 49.814
850757 102.591 -50.678 49.785
850758 102.616 -50.701 49.757
850759 102.641 -50.723 49.730
850760 102.665 -50.745 49.705
850761 102.690 -50.768 49.682
850762 102.715 -50.790 49.661
850763 102.739 -50.813 49.641
850764 102.764 -50.835 49.624
850765 102.789 -50.857 49.608
850766 102.813 -50.879 49.595
850767 102.838 -50.902 49.582
850768 102.862 -50.925 49.571
850769 102.887 -50.947 49.560
850770 102.911 -50.970 49.551
850771 102.936 -50.992 49.543
850772 102.961 -51.014 49.536
850773 102.985 -51.036 49.526
850774 103.010 -51.059 49.511
850775 103.035 -51.082 49.493
850776 103.059 -51.104 49.475
850777 103.084 -51.127 49.454
850778 103.109 -51.149 49.433
850779 103.134 -51.171 49.413
850780 103.158 -51.193 49.389
850781 103.184 -51.207 49.361
850782 103.210 -51.218 49.333
850783 103.235 -51.229 49.302
850784 103.260 -51.241 49.271
850785 103.286 -51.252 49.239
850786 103.311 -51.265 49.208
850787 103.336 -51.276 49.176
850788 103.362 -51.288 49.145
850789 103.387 -51.299 49.115
850790 103.412 -51.312 49.085
851458 102.591 -50.659 49.793
851459 102.615 -50.681 49.764
851460 102.640 -50.704 49.737
851461 102.664 -50.726 49.712
851462 102.689 -50.748 49.688
851463 102.714 -50.771 49.666
851464 102.738 -50.793 49.646
851465 102.763 -50.816 49.629
851466 102.787 -50.838 49.613
851467 102.812 -50.860 49.601
851468 102.837 -50.883 49.587
851469 102.862 -50.905 49.576
851470 102.886 -50.928 49.566
851471 102.911 -50.950 49.557
851472 102.936 -50.973 49.549
851473 102.960 -50.995 49.542
851474 102.984 -51.017 49.534
851475 103.009 -51.039 49.520
851476 103.034 -51.062 49.502
851477 103.058 -51.085 49.484
851478 103.083 -51.107 49.468
851479 103.108 -51.130 49.448
851480 103.132 -51.152 49.426
851481 103.158 -51.173 49.400
851482 103.185 -51.190 49.371
851483 103.211 -51.200 49.343
851484 103.237 -51.211 49.313
851485 103.262 -51.221 49.282
851486 103.288 -51.232 49.252
851487 103.313 -51.243 49.221
851488 103.339 -51.254 49.190
851489 103.365 -51.265 49.159
851490 103.390 -51.275 49.130
851491 103.416 -51.287 49.100
851492 103.441 -51.298 49.069
852161 102.639 -50.684 49.744
852162 102.664 -50.707 49.719
852163 102.688 -50.729 49.694
852164 102.713 -50.751 49.672
852165 102.737 -50.774 49.652
852166 102.762 -50.796 49.634
852167 102.787 -50.819 49.618
852168 102.811 -50.841 49.604
852169 102.836 -50.864 49.591
852170 102.861 -50.886 49.580
852171 102.885 -50.908 49.570
852172 102.910 -50.931 49.561
852173 102.935 -50.953 49.553
852174 102.959 -50.976 49.546
852175 102.984 -50.998 49.539
852176 103.009 -51.020 49.527
852177 103.033 -51.042 49.511
852178 103.058 -51.065 49.494
852179 103.083 -51.088 49.482
852180 103.107 -51.110 49.462
852181 103.132 -51.133 49.438
852182 103.159 -51.152 49.410
852183 103.186 -51.172 49.381
852184 103.212 -51.183 49.352
852185 103.238 -51.193 49.323
852186 103.263 -51.204 49.293
852187 103.289 -51.215 49.264
852188 103.315 -51.226 49.234
852189 103.340 -51.236 49.203
852190 103.366 -51.247 49.173
852191 103.392 -51.258 49.144
852192 103.418 -51.268 49.115
852193 103.443 -51.279 49.083
852865 102.712 -50.732 49.678
852866 102.737 -50.754 49.658
852867 102.761 -50.777 49.639
852868 102.786 -50.799 49.622
852869 102.811 -50.822 49.607
852870 102.835 -50.844 49.594
852871 102.860 -50.867 49.584
852872 102.884 -50.889 49.574
852873 102.909 -50.911 49.564
852874 102.934 -50.934 49.556
852875 102.958 -50.956 49.548
852876 102.983 -50.979 49.542
852877 103.008 -51.001 49.533
852878 103.032 -51.023 49.520
852879 103.057 -51.045 49.503
852880 103.082 -51.068 49.494
852881 103.106 -51.091 49.475
852882 103.132 -51.112 49.448
852883 103.160 -51.132 49.420
852884 103.187 -51.152 49.391
852885 103.213 -51.165 49.362
852886 103.239 -51.177 49.333
852887 103.264 -51.187 49.304
852888 103.290 -51.198 49.275
852889 103.316 -51.208 49.246
852890 103.341 -51.219 49.216
852891 103.367 -51.230 49.186
852892 103.393 -51.240 49.158
852893 103.418 -51.251 49.128
853566 102.711 -50.713 49.684
853567 102.736 -50.735 49.661
853568 102.760 -50.757 49.643
853569 102.785 -50.780 49.626
853570 102.810 -50.802 49.611
853571 102.834 -50.825 49.598
853572 102.859 -50.847 49.588
853573 102.884 -50.870 49.577
853574 102.908 -50.892 49.567
853575 102.933 -50.914 49.559
853576 102.958 -50.937 49.551
853577 102.982 -50.959 49.543
853578 103.007 -50.982 49.536
853579 103.031 -51.004 49.526
853580 103.056 -51.027 49.509
853581 103.081 -51.049 49.498
853582 103.105 -51.071 49.478
853583 103.132 -51.092 49.453
853584 103.160 -51.111 49.426
853585 103.187 -51.131 49.398
853586 103.215 -51.149 49.370
853587 103.240 -51.159 49.342
853588 103.266 -51.170 49.314
853589 103.292 -51.180 49.285
853590 103.317 -51.191 49.256
853591 103.343 -51.202 49.227
853592 103.369 -51.212 49.198
853593 103.394 -51.224 49.169
853594 103.420 -51.234 49.140
854268 102.735 -50.716 49.668
854269 102.759 -50.738 49.648
854270 102.784 -50.760 49.631
854271 102.809 -50.783 49.616
854272 102.833 -50.805 49.602
854273 102.858 -50.828 49.591
854274 102.883 -50.850 49.580
854275 102.907 -50.873 49.571
854276 102.932 -50.895 49.562
854277 102.957 -50.917 49.553
854278 102.981 -50.940 49.545
854279 103.006 -50.962 49.537
854280 103.031 -50.985 49.528
854281 103.055 -51.007 49.511
854282 103.080 -51.030 49.498
854283 103.106 -51.051 49.478
854284 103.133 -51.071 49.455
854285 103.161 -51.091 49.430
854286 103.188 -51.111 49.404
854287 103.216 -51.131 49.377
854288 103.241 -51.142 49.350
854289 103.267 -51.152 49.322
854290 103.293 -51.163 49.294
854291 103.318 -51.174 49.265
854292 103.344 -51.184 49.237
854293 103.370 -51.196 49.208
854294 103.395 -51.206 49.180
854295 103.421 -51.217 49.151
854969 102.734 -50.696 49.674
854970 102.759 -50.719 49.655
854971 102.784 -50.741 49.636
854972 102.808 -50.763 49.621
854973 102.832 -50.786 49.607
854974 102.857 -50.808 49.594
854975 102.882 -50.831 49.584
854976 102.906 -50.853 49.574
854977 102.931 -50.876 49.565
854978 102.956 -50.898 49.556
854979 102.980 -50.920 49.546
854980 103.005 -50.943 49.538
854981 103.030 -50.965 49.528
854982 103.054 -50.988 49.512
854983 103.079 -51.010 49.497
854984 103.106 -51.030 49.477
854985 103.134 -51.050 49.456
854986 103.161 -51.070 49.432
854987 103.188 -51.090 49.408
854988 103.216 -51.111 49.383
854989 103.243 -51.124 49.356
854990 103.268 -51.135 49.329
854991 103.294 -51.146 49.301
854992 103.319 -51.157 49.273
854993 103.345 -51.168 49.245
854994 103.371 -51.178 49.217
854995 103.396 -51.189 49.188
854996 103.422 -51.199 49.161
855671 102.758 -50.699 49.661
855672 102.783 -50.722 49.642
855673 102.807 -50.745 49.626
855674 102.832 -50.766 49.612
855675 102.857 -50.789 49.599
855676 102.881 -50.811 49.588
855677 102.906 -50.834 49.577
855678 102.930 -50.856 49.567
855679 102.955 -50.879 49.557
855680 102.979 -50.901 49.547
855681 103.004 -50.923 49.537
855682 103.029 -50.946 49.526
855683 103.053 -50.968 49.511
855684 103.079 -50.990 49.496
855685 103.107 -51.010 49.478
855686 103.134 -51.030 49.457
855687 103.162 -51.050 49.434
855688 103.189 -51.070 49.411
855689 103.217 -51.089 49.387
855690 103.244 -51.107 49.361
855691 103.269 -51.117 49.334
855692 103.295 -51.129 49.307
855693 103.321 -51.139 49.280
855694 103.347 -51.150 49.252
855695 103.372 -51.161 49.224
855696 103.398 -51.171 49.196
856372 102.757 -50.680 49.667
856373 102.781 -50.702 49.648
856374 102.806 -50.725 49.631
856375 102.831 -50.748 49.617
856376 102.856 -50.770 49.604
856377 102.880 -50.792 49.591
856378 102.905 -50.814 49.580
856379 102.930 -50.837 49.570
856380 102.954 -50.859 49.560
856381 102.979 -50.882 49.549
856382 103.004 -50.904 49.538
856383 103.028 -50.926 49.526
856384 103.053 -50.949 49.512
856385 103.080 -50.969 49.496
856386 103.107 -50.989 49.478
856387 103.135 -51.009 49.458
856388 103.162 -51.030 49.436
856389 103.190 -51.049 49.414
856390 103.217 -51.069 49.390
856391 103.245 -51.089 49.365
856392 103.270 -51.101 49.339
856393 103.296 -51.111 49.312
856394 103.322 -51.122 49.285
856395 103.348 -51.133 49.258
856396 103.373 -51.143 49.230
856397 103.399 -51.154 49.203
857074 102.781 -50.683 49.655
857075 102.806 -50.705 49.637
857076 102.830 -50.728 49.622
857077 102.855 -50.751 49.608
857078 102.879 -50.773 49.595
857079 102.904 -50.795 49.584
857080 102.929 -50.817 49.573
857081 102.953 -50.840 49.562
857082 102.978 -50.862 49.551
857083 103.003 -50.885 49.539
857084 103.027 -50.907 49.526
857085 103.053 -50.929 49.513
857086 103.080 -50.948 49.497
857087 103.108 -50.969 49.479
857088 103.135 -50.989 49.459
857089 103.163 -51.008 49.439
857090 103.190 -51.029 49.416
857091 103.218 -51.049 49.393
857092 103.245 -51.069 49.369
857093 103.272 -51.083 49.343
857094 103.298 -51.094 49.317
857095 103.323 -51.105 49.289
857096 103.349 -51.115 49.262
857097 103.375 -51.126 49.236
857098 103.400 -51.137 49.210
857775 102.780 -50.663 49.662
857776 102.805 -50.686 49.643
857777 102.829 -50.708 49.627
857778 102.854 -50.731 49.613
857779 102.879 -50.754 49.601
857780 102.903 -50.776 49.589
857781 102.928 -50.798 49.577
857782 102.952 -50.820 49.565
857783 102.977 -50.843 49.553
857784 103.001 -50.865 49.542
857785 103.026 -50.888 49.529
857786 103.053 -50.908 49.514
857787 103.081 -50.928 49.497
857788 103.108 -50.948 49.480
857789 103.135 -50.968 49.461
857790 103.163 -50.988 49.441
857791 103.191 -51.008 49.418
857792 103.218 -51.028 49.395
857793 103.246 -51.048 49.370
857794 103.273 -51.066 49.345
857795 103.299 -51.077 49.319
857796 103.324 -51.087 49.292
857797 103.350 -51.098 49.266
857798 103.376 -51.109 49.240
858477 102.804 -50.666 49.651
858478 102.828 -50.689 49.634
858479 102.853 -50.711 49.620
858480 102.878 -50.734 49.607
858481 102.902 -50.757 49.594
858482 102.927 -50.779 49.582
858483 102.952 -50.801 49.570
858484 102.976 -50.823 49.558
858485 103.001 -50.846 49.545
858486 103.026 -50.867 49.532
858487 103.054 -50.888 49.517
858488 103.081 -50.907 49.500
858489 103.109 -50.927 49.482
858490 103.136 -50.948 49.464
858491 103.164 -50.967 49.442
858492 103.191 -50.988 49.420
858493 103.219 -51.008 49.397
858494 103.246 -51.027 49.372
858495 103.274 -51.048 49.347
858496 103.300 -51.059 49.322
858497 103.325 -51.070 49.296
858498 103.351 -51.081 49.270
859179 102.827 -50.669 49.642
859180 102.852 -50.692 49.627
859181 102.877 -50.714 49.613
859182 102.901 -50.737 49.600
859183 102.926 -50.760 49.587
859184 102.951 -50.782 49.575
859185 102.975 -50.804 49.563
859186 103.000 -50.826 49.549
859187 103.027 -50.847 49.536
859188 103.054 -50.867 49.520
859189 103.082 -50.887 49.504
859190 103.109 -50.907 49.485
859191 103.137 -50.927 49.467
859192 103.164 -50.947 49.444
859193 103.192 -50.967 49.423
859194 103.219 -50.987 49.399
859195 103.247 -51.007 49.375
859196 103.275 -51.027 49.351
859197 103.301 -51.042 49.326
859880 102.827 -50.650 49.650
859881 102.852 -50.672 49.634
859882 102.876 -50.695 49.620
859883 102.901 -50.717 49.606
859884 102.926 -50.740 49.593
859885 102.950 -50.763 49.581
859886 102.974 -50.785 49.568
859887 103.000 -50.807 49.554
859888 103.028 -50.826 49.540
859889 103.055 -50.846 49.524
859890 103.082 -50.867 49.508
859891 103.110 -50.886 49.490
859892 103.137 -50.907 49.470
859893 103.165 -50.926 49.449
859894 103.192 -50.946 49.427
859895 103.220 -50.967 49.404
859896 103.247 -50.986 49.379
859897 103.275 -51.007 49.355
860582 102.851 -50.654 49.643
860583 102.875 -50.676 49.628
860584 102.900 -50.698 49.613
860585 102.924 -50.720 49.600
860586 102.949 -50.743 49.587
860587 102.974 -50.766 49.574
860588 103.001 -50.786 49.560
860589 103.028 -50.806 49.546
860590 103.055 -50.826 49.530
860591 103.083 -50.846 49.512
860592 103.110 -50.866 49.495
860593 103.138 -50.886 49.475
860594 103.165 -50.906 49.454
860595 103.193 -50.926 49.432
860596 103.221 -50.946 49.408
860597 103.248 -50.966 49.384
861284 102.874 -50.657 49.635
861285 102.899 -50.679 49.620
861286 102.924 -50.701 49.607
861287 102.948 -50.723 49.594
861288 102.974 -50.745 49.580
861289 103.001 -50.765 49.566
861290 103.029 -50.785 49.552
861291 103.056 -50.805 49.536
861292 103.083 -50.826 49.518
